#
# Copyright (c) 2013 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Base for API Sets Eseries commands
## @author dl-nacl-dev@netapp.com
## @status Shared
## @pod here

package NACL::APISet::Eseries;

#
# Compiler Directives
#
use strict;
use warnings;

# NATE imports
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use NACL::APISet::_Mixins::CLI qw(_load_version_xml);
use Data::Dumper;
use NATE::BaseException qw(:try);
use NATE::Exceptions::Argument;

# Base class
use base qw(NACL::APISet);

###############################################################################
# Method:    find_cdef_directory
# Objective: To find the cdef directory based on various version related
#            attributes obtained from the service.
# Details:   See POD documentation at the end of this file.
###############################################################################

sub find_cdef_directory {
    $Log->enter() if $may_enter;
    my $self = shift();

    # Load CDEFs from the CDEF directories which we maintain

    my $interface = lc( $self->{interface} );
    
    my $libxml_obj = $self->_load_version_xml();
    my $cdef_dir
        = $libxml_obj->find("/version/eseries/$interface/default_cdef_dir")->string_value;

    my @cdef_paths;
    push( @cdef_paths, $cdef_dir );
    
    $Log->debug("CDEF directories found are:\n" . Dumper(\@cdef_paths));

    $Log->exit() if $may_exit;
    
    return @cdef_paths;
    
}

1;

=head1 NAME

NACL::APISet::Eseries - Base class for API Sets for Eseriess

=head1 SYNOPSIS

 use NACL::APISet::Eseries;

=head1 DESCRIPTION

A C<NACL::APISet::Eseries> is used to create API Set objects for Eseries. This acts
as a factory class for creation of API Set objects for the interface CLI.

=cut

=head1 EXCEPTIONS

=over

=item NATE::BaseException

=item NATE::Exceptions::Argument

=back

=cut

=head1 METHODS

=head2 find_cdef_directory

This method gets the information of various version related attributes and uses
them to autodiscover the CDEF path. The data about the CDEF directory to
be loaded is maintained in a separate XML file from which it finds the CDEF path
based on the discovered version attributes. The version attributes to be
discovered will vary depending on the category and set.

=over

=item Synopsis

my $cdef_dir = $Api_Set_Obj->find_cdef_directory();

=item Arguments

Nothing.

=item Returns

Returns the relative path to CDEF directory.

=back

=head1 AUTHOR/MAINTAINER

=over

=item NACL Development (dl-nacl-dev@netapp.com)

=back

=cut

