#
# Copyright (c) 2001-2009 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Base for API Sets CLI commands
## @author dl-nacl-dev@netapp.com
## @status Shared
## @pod here

package NACL::APISet::Host;

#
# Compiler Directives
#
use strict;
use warnings;

# NATE imports
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Data::Dumper;
use NATE::BaseException qw(:try);
use NATE::Exceptions::Argument qw(:try);

# Base class
use base qw(NACL::APISet);

sub _new {
    $Log->enter() if $may_enter;

    my ( $class, %args ) = @_;

    #my $self  = {};
    my $self = \%args;

    # Bless the hash returned which will make it a Perl object
    bless $self, $class;

    # And give the caller the reference to the bless object.
    $Log->exit() if $may_exit;
    return $self;
}

##############################################################################
#                               Private Methods
##############################################################################
sub _read_default_cdef_dir {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift();
    
    my %args = $pkg_or_obj->validate_params(
        params => \@_,
        spec   => $pkg_or_obj->_read_default_cdef_dir_spec,
    );
    
    my @cdef_paths = $pkg_or_obj->_read_shared_cdef_dir(%args);
    
    $Log->exit() if $may_exit;
    
    return @cdef_paths;
}

###############################################################################
# Method:    find_cdef_directory
# Objective: To find the cdef directory based on various version related
#            attributes obtained from the service.
# Details:   See POD documentation at the end of this file.
###############################################################################

sub find_cdef_directory {
    $Log->enter() if $may_enter;
    my $self = shift();

    # Load CDEFs from the CDEF directories which we maintain based on branch
    # and release information.

    my $interface = lc( $self->{interface} );
    my $set       = lc( $self->{set} );

    my %cdef_args = ();

    # For Solaris, HPUX and AIX, the internal implementation makes use of Linux APISet
    # altogether and only the foll minor adjustment is required to handle
    # CDEF dir path. Since we don't have NACL::APISet::Host::CLI::Solaris,
    # NACL::APISet::Host::CLI::HPUX and NACL::APISet::Host::CLI::AIX we have
    # performed the subclass specific handling in the base class.
    $cdef_args{set} = 'solaris' if $self->{actual_set} eq 'Solaris';
    $cdef_args{set} = 'hpux' if $self->{actual_set} eq 'HPUX';
    $cdef_args{set} = 'aix' if $self->{actual_set} eq 'AIX';

    my @cdef_paths = $self->get_default_cdef_dir_list(%cdef_args);

    $Log->debug("CDEF directories found are:\n" . Dumper(\@cdef_paths));
    $Log->exit() if $may_exit;
    
    return @cdef_paths;
    
}

1;

=head1 NAME

NACL::APISet::Host - Base class for API Sets for Hosts

=head1 SYNOPSIS

 use NACL::APISet::Host;

=head1 DESCRIPTION

A C<NACL::APISet::Host> is used to create API Set objects for Host. This acts
as a factory class for creation of API Set objects for the interface CLI.

=cut

=head1 EXCEPTIONS

=over

=item NATE::BaseException

=item NATE::Exceptions::Argument

=back

=cut

=head1 METHODS

=head2 find_cdef_directory

This method gets the information of various version related attributes and uses
them to autodiscover the CDEF path. The data about the CDEF directory to
be loaded is maintained in a separate XML file from which it finds the CDEF path
based on the discovered version attributes. The version attributes to be
discovered will vary depending on the category and set.

=over

=item Synopsis

my $cdef_dir = $Api_Set_Obj->find_cdef_directory();

=item Arguments

Nothing.

=item Returns

Returns the relative path to CDEF directory.

=back

=head1 AUTHOR/MAINTAINER

=over

=item NACL Development (dl-nacl-dev@netapp.com)

=back

=cut

