#
# Copyright (c) 2001-2009 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Base for API Sets CLI commands
## @author pressley@netapp.com, dl-nacl-dev@netapp.com
## @status Shared
## @pod here

package NACL::APISet::Node;

#
# Compiler Directives
#
use strict;
use warnings;

# NATE imports
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Data::Dumper;
use Params::Validate qw(validate_with SCALAR SCALARREF);
use NACL::APISet::_Mixins::CLI qw(_load_version_xml);

use NATE::BaseException qw(:try);
use NATE::Exceptions::Argument;

# Base class
use base qw(NACL::APISet);

sub _new {
    $Log->enter() if $may_enter;

    my ( $class, %args ) = @_;
    my $self = {};

    # Bless the hash returned which will make it a Perl object
    bless $self, $class;

    # And give the caller the reference to the bless object.
    $Log->exit() if $may_exit;
    return $self;
}

###############################################################################
# Method:    find_cdef_directory
# Objective: To find the cdef directory based on various version related
#            attributes obtained from the service.
# Details:   See POD documentation at the end of this file.
###############################################################################

sub find_cdef_directory {
    $Log->enter() if $may_enter;
    my $self = shift();

    # Load CDEFs from the CDEF directories which we maintain based on release
    # information.
    my %required_attributes = (
        "CMode" => [ "release" ],
        "Vserver" => ["release"],
        "DBlade" => [ "release" ],
        "7Mode" => [ "release" ],
        "Nodescope" => [ "release" ],
        "NodescopeNGSH" => [ "release" ],
    );

    my $interface = lc( $self->{interface} );

    my $cdef_dir;
    my $attr_info = {};
    my $type = $self->{actual_set};
    if ( defined( $self->{version_manager} ) ) {
        $type = "7G" if ( $self->{version_manager}->{is_classic} );
        $attr_info = $self->_find_cdef_directory(
            required_attributes => $required_attributes{$type} )
            if (exists $required_attributes{$type} &&
                (($self->{set} ne "7Mode") || ($self->{interface} ne "CLI")));
    }
    my $release = $attr_info->{release};
    # Normalize the internal and external release strings into internal
    # codename.
    if ( defined $release ) {

        # release should be normalized in lowercase so that its case
        # insensitive when we try to map it to its codename.
        $release = lc($release);
        try {
            $release = $self->{version_manager}
                ->map_release_to_codename( release => $release );
        }
        catch NATE::BaseException with {
            my $ex = shift;
            $Log->debug( "Caught an exception while trying to map $release "
                    . "to its codename: "
                    . $ex->text() );

            # Emit a warning, set the normalized release name to undef
            # and then proceed.
            $Log->warn( "Versioning had hit the error: "
                    . $ex->text()
                    . "\nRelease mapping may be missing for release $release."
                    . "\nPlease contact dl-nacl-dev about this warning." );
            $release = undef;
        };
    }

    # Discover the CDEF path
    my ( @nodeset, $search_str );
    my $version_set = lc( $self->{set} );
    $version_set = "sevenmode" if ( $version_set eq "7mode" );
    $search_str = "/version/node/$interface/$version_set/default_cdef_dir";

    # Load version.xml file
    my $libxml_obj = $self->_load_version_xml();

    # Load the cdef_dir with the defaults of that version_set
    $cdef_dir = $libxml_obj->find($search_str)->string_value;

    # If the release information couldn't be obtained, then simply
    # return the default CDEF directory corresponding to the version_set
    unless ( defined $release ) {
        $Log->debug( "Release name is undef and hence loading default "
                    . "cdef dir." );
        $Log->exit() if $may_exit;
        return $cdef_dir;
    }

    my $release_cdef_dir = $self->_get_release_cdef_dir(
        interface => $interface,
        version   => $version_set,
        release   => $release,
        libxml_obj => $libxml_obj,
    );

    if (! defined $release_cdef_dir) {
        $Log->debug("CDEF directory is not yet supported for $release");

        # All releases which have a mapping in release hash should have
        # the corresponding default CDEF dir. For 7Mode CLI, we have only
        # one CDEF dir at the moment and a warning shouldn't be emitted
        # in that case unnecessarily.
        $Log->warn( "CDEF directory corresponding to release '$release' "
                . "doesn't exist. Please contact dl-nacl-dev for this warning."
                . " Loading CDEFs from default CDEF directory defined for the"
                . " set '$self->{set}'." );
    } else {
        $cdef_dir = $release_cdef_dir;
    }

    $Log->exit() if $may_exit;
    return $cdef_dir;
}

##############################################################################
# Method:    _is_vserver_version_manager
# Objective: To discover whether the version manager corresponds to
#            Node or Vserver
# Details:  This method is used to find whether the version manager is of
#           type 'NACL::APISet::Version::Node::Vserver' or not and
#           this helps the caller to decide whether it can invoke certain
#           methods on it or not.
##############################################################################

sub _is_vserver_version_manager {
    $Log->enter() if $may_enter;
    my ($self, $version_manager) = @_;
    my $res = $version_manager->isa('NACL::APISet::Version::Node::Vserver');
    $Log->debug("isa vserver version manager: $res");
    $Log->exit;
    return $res;
}

sub _get_release_cdef_dir_spec {

    return {
        interface  => {type => SCALAR},
        version    => {type => SCALAR},
        release    => {type => SCALAR},
        libxml_obj => {type => SCALARREF},
    };
}

sub _get_release_cdef_dir {
    $Log->enter() if $may_enter;
    my ($self, %args) = @_;
    my %opts = validate_with(
        params  => \%args,
        spec    => $self->_get_release_cdef_dir_spec(),
    );

    my $interface = $opts{interface};
    my $version   = $opts{version};
    my $release   = $opts{release};
    my $libxml_obj = $opts{libxml_obj};

    my $cdef_dir = $libxml_obj->find("/version/node/$interface/$version/release[\@value='" .
                                    $release . "']/release_cdef_dir")->string_value();
    if (! defined $cdef_dir or $cdef_dir eq "") {
        if ($release =~ s/(\w+)?_\d/$1/) {
            $cdef_dir = $libxml_obj->find("/version/node/$interface/$version/release[\@value='" .
                        $release . "']/release_cdef_dir")->string_value();
        }
    }

    if ($cdef_dir eq "") {
        undef $cdef_dir;
    }

    $Log->exit() if $may_exit;
    return $cdef_dir;
}

sub _record_connection_into_keyval {
    my $self = shift;
    my $hostobj = $self->hostobj;
    require Connectrec;
    # Careful with the "mixed stack" problem
    # Even though Connectrec is being updated at the same time as the
    # lines below are being added, there is not guarantee that
    # testers wont run a newer NACL stack with an older NATE stack
    if (Connectrec->can('_record_connected_host_into_keyval')){
        Connectrec->_record_connected_host_into_keyval(
            hostrec => $hostobj,
            from    => __PACKAGE__,
        );
    }
    return;
}

1;

=head1 NAME

NACL::APISet::Node - Base class for API Sets for Nodes

=head1 SYNOPSIS

 use NACL::APISet::Node;

=head1 DESCRIPTION

A C<NACL::APISet::Node> is used to create API Set objects for Nodes. This acts
as a factory class for creation of API Set objects for various Node interfaces
such as CLI, ZAPI, SNMP, NDMP.

=cut

=head1 EXCEPTIONS

=over

=item NATE::BaseException

=item NATE::Exceptions::Argument

=back

=cut

=head1 METHODS

=head2 find_cdef_directory

This method gets the information of various version related attributes and uses
them to autodiscover the CDEF path. The data about the CDEF directory to
be loaded is maintained in a separate XML file from which it finds the CDEF path
based on the discovered version attributes. The version attributes to be
discovered will vary depending on the category and set.

=over

=item Synopsis

my $cdef_dir = $Api_Set_Obj->find_cdef_directory();

=item Arguments

Nothing.

=item Returns

Returns the relative path to CDEF directory.

=back

=head1 AUTHOR/MAINTAINER

=over

=item NACL Development (dl-nacl-dev@netapp.com)

=back

 $Id: //depot/prod/DOT/dev/test/lib/NACL/APISet/Node.pm#4 $
 Last revised $Date: 2013/01/28 $ by $Author: rawat $

=cut
