#$Id$
#
#Copyright (c) 2012-2015 Network Appliance, Inc.
#All rights reserved.
#
## @summary Routines to parse loader prompt commands of a filer.
##
## @description
##    This is a library to parse the raw output of loader prompt commands of a filer.
##
## @synopsis
##  use NACL::APISet::Node::CLI::Firmware::default::firmware
##
##  firmware(%args)
##
## @author jais@netapp.com
##
## @status public
#
package NACL::APISet::Node::CLI::Firmware::default::firmware;

use strict;
use NACL::APISet::Response::CLI::ParserUtils;

#Include exceptions which will be thrown
use NACL::APISet::Exceptions::CommandFailedException qw(:try);
use NACL::APISet::Exceptions::MissingRequiredParamException;
use NACL::APISet::Exceptions::InvalidParamException;
use NACL::APISet::Exceptions::InvalidParamValueException;
use NACL::APISet::Exceptions::MissingValueException;

sub _check_error {
    my %args = @_;

    my $raw_output = $args{raw_output};
    my $resp       = $args{"response_object"};
    my $error_message;
    my $error;

    # Check for the error
    if ( $raw_output =~ /([iI]nvalid switch)/ ) {
        $error         = "InvalidParam";
        $error_message = $1;
    }
    elsif ( $raw_output =~ /(No program name specified)/ ) {
        $error         = "MissingRequiredParam";
        $error_message = $1;
    }
    elsif ($raw_output =~ /Could not load.*:(.*)/
        || $raw_output =~ /Invalid file name.*:(.*)/
        || $raw_output =~ /(Device.*is not valid)/
        || $raw_output =~ /(Invalid IP.*)/
        || $raw_output =~ /(Invalid hardware.*)/ )
    {
        $error         = "InvalidParamValue";
        $error_message = $raw_output;
    }
    elsif ($raw_output =~ /(Loading.*Failed)/
        || $raw_output =~ /(No program has been loaded)/
        || $raw_output =~ /(Could not resolve IP.*:.*)/
        || $raw_output =~ /(Could not transmit.*)/
        || $raw_output =~ /(Network interface has not been configured)/
        || $raw_output =~ /(Could not activate.*)/
        || $raw_output =~ /(No ARP entries)/ )
    {
        $error         = "CommandFailed";
        $error_message = $1;
    }
    elsif ( $raw_output =~ /command status = -2/ ) {
        $error         = "InvalidParam";
        $error_message = "Improper usage of command";
    }
    elsif ( $raw_output =~ /[0-9]+ is not a valid/
          ||$raw_output =~ /Invalid\s+parameter/
          ||$raw_output =~ /Invalid\s+command/ ) {
        $error         = "InvalidParamValue";
        $error_message = "Input value is not valid";
    }
    elsif ( $raw_output =~ /(only available in IPV6)/ ) {
        $error         = "CommandFailed";
        $error_message = $raw_output;
    }
    elsif ( $raw_output =~ /No such file or directory/ ) {
        $error         = "CommandFailed";
        $error_message = $raw_output;
    }
    elsif ( $raw_output =~ /ssram: insufficent write data/ ) {
        $error         = "MissingValue";
        $error_message = $raw_output;
    }
    elsif ($raw_output =~ /ssram: invalid data offset./ ) {
        $error         = "InvalidParam";
        $error_message = $raw_output;
    }
    if ($error) {
        $error = 'NACL::APISet::Exceptions::' . $error . 'Exception';
        $error->throw( text => $error_message, response_object => $resp );
    }
}

## @name ::version
##
## @summary Parses the raw output from the "version" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from version with any of its options as described in the
##    corresponding cdef file 'firmware.cdefs'
## @synopsis
##      version(%args);
## INPUTS TO METHOD CALL 'version' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

##&nbsp;Command = version
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;          {
##&nbsp;            'loader_version' => '4.1X1',
##&nbsp;            'bios_version' => '8.1X1'
##&nbsp;          }
##&nbsp;        ];
##&nbsp;
sub version {
    my %args = @_;
    my $resp       = $args{response_object};
    my $raw_output = $resp->get_processed_output();
    my $command     = $resp->get_command();
    _check_error( raw_output => $raw_output, response_object => $resp );

    my @lines = get_lines(
        "raw_output"      => $raw_output,
        "can_throw"       => '1',
        "response_object" => $resp,
    );
    my ( @result, $i );
    if ( $command =~ /-v/ ) {
	for ( my $count = 0 ; $count < @lines ; $count++ ) {
	    next if ( $lines[$count] =~ /Index/i );
	    next if ( $lines[$count] =~ /-/ );
	    my @values = split(/\s+/, $lines[$count]);
	    my @keys = qw(index flags variable_name value default);
	    for ( $i = 0 ; $i < @keys ; $i++ ) {
		$result[$count]->{$keys[$i]} = $values[$i];
	    }
	}
	@result = grep defined, @result;
    }else {
        for ( $i = 2 ; $i <= $#lines ; $i++ ) {
            if ( $lines[$i] =~ /(\S+)\s+(.*)/ ) {
                $result[0]->{ to_key_name( item => $1 ) } = $2;
            }
        }
    }
    return ( \@result );
}

## @name ::update_flash
##
## @summary Parses the raw output from the "update_flash" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from update_flash with any of its options as described in the
##    corresponding cdef file 'firmware.cdefs'
## @synopsis
##      update_flash(%args);
## INPUTS TO METHOD CALL 'update_flash' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

sub update_flash {
    my %args = @_;

    my $resp       = $args{response_object};
    my $raw_output = $resp->get_processed_output();

    _check_error( raw_output => $raw_output, response_object => $resp );

    return [];
}

## @name ::netboot
##
## @summary Parses the raw output from the "netboot" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from netboot with any of its options as described in the
##    corresponding cdef file 'firmware.cdefs'
## @synopsis
##      netboot(%args);
## INPUTS TO METHOD CALL 'netboot' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## <tr> <td> No switch for this key </td> <td> 'url' </td> <td> '0' </td> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

sub netboot {
    my %args = @_;

    my $resp       = $args{response_object};
    my $raw_output = $resp->get_processed_output();

    _check_error( raw_output => $raw_output, response_object => $resp );

    return [];
}

## @name ::boot_diags
##
## @summary Parses the raw output from the "boot_diags" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from boot_diags with any of its options as described in the
##    corresponding cdef file 'firmware.cdefs'
## @synopsis
##      boot_diags(%args);
## INPUTS TO METHOD CALL 'boot_diags' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

sub boot_diags {
    my %args = @_;

    my $resp       = $args{response_object};
    my $raw_output = $resp->get_processed_output();

    _check_error( raw_output => $raw_output, response_object => $resp );

    return [];
}

## @name ::boot_backup
##
## @summary Parses the raw output from the "boot_backup" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from boot_backup with any of its options as described in the
##    corresponding cdef file 'firmware.cdefs'
## @synopsis
##      boot_backup(%args);
## INPUTS TO METHOD CALL 'boot_backup' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

sub boot_backup {
    my %args = @_;

    my $resp       = $args{response_object};
    my $raw_output = $resp->get_processed_output();

    _check_error( raw_output => $raw_output, response_object => $resp );

    return [];
}

## @name ::boot_primary
##
## @summary Parses the raw output from the "boot_primary" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from boot_primary with any of its options as described in the
##    corresponding cdef file 'firmware.cdefs'
## @synopsis
##      boot_primary(%args);
## INPUTS TO METHOD CALL 'boot_primary' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

sub boot_primary {
    my %args = @_;

    my $resp       = $args{response_object};
    my $raw_output = $resp->get_processed_output();

    _check_error( raw_output => $raw_output, response_object => $resp );

    return [];
}

## @name ::boot_ontap
##
## @summary Parses the raw output from the "boot_ontap" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from boot_ontap with any of its options as described in the
##    corresponding cdef file 'firmware.cdefs'
## @synopsis
##      boot_ontap(%args);
## INPUTS TO METHOD CALL 'boot_ontap' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

sub boot_ontap {
    my %args = @_;

    my $resp       = $args{response_object};
    my $raw_output = $resp->get_processed_output();

    _check_error( raw_output => $raw_output, response_object => $resp );

    return [];
}

## @name ::flash
##
## @summary Parses the raw output from the "flash" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from flash with any of its options as described in the
##    corresponding cdef file 'firmware.cdefs'
## @synopsis
##      flash(%args);
## INPUTS TO METHOD CALL 'flash' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## <tr> <td> '-boot' </td> <td> 'boot-image' </td> <td> '0' </td> </tr>
## <tr> <td> No switch for this key </td> <td> 'filename' </td> <td> '0' </td> </tr>
## <tr> <td> No switch for this key </td> <td> 'flashdevice' </td> <td> '0' </td> </tr>
## <tr> <td> '-noversion' </td> <td> 'noversion' </td> <td> '1' </td> </tr>
## <tr> <td> '-all' </td> <td> 'all' </td> <td> '1' </td> </tr>
## <tr> <td> '-backup' </td> <td> 'backup' </td> <td> '1' </td> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

sub flash {
    my %args = @_;

    my $resp       = $args{response_object};
    my $raw_output = $resp->get_processed_output();

    _check_error( raw_output => $raw_output, response_object => $resp );

    return [];
}

## @name ::bye
##
## @summary Parses the raw output from the "bye" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from bye with any of its options as described in the
##    corresponding cdef file 'firmware.cdefs'
## @synopsis
##      bye(%args);
## INPUTS TO METHOD CALL 'bye' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

sub bye {
    my %args = @_;

    my $resp       = $args{response_object};
    my $raw_output = $resp->get_processed_output();

    _check_error( raw_output => $raw_output, response_object => $resp );

    return [];
}

## @name ::set_date
##
## @summary Parses the raw output from the "set date" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from set_date with any of its options as described in the
##    corresponding cdef file 'firmware.cdefs'
## @synopsis
##      set_date(%args);
## INPUTS TO METHOD CALL 'set_date' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## <tr> <td> No switch for this key </td> <td> 'date' </td> <td> '0' </td> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

sub set_date {
    my %args = @_;

    my $resp       = $args{response_object};
    my $raw_output = $resp->get_processed_output();

    _check_error( raw_output => $raw_output, response_object => $resp );

    return [];
}

## @name ::set_time
##
## @summary Parses the raw output from the "set time" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from set_time with any of its options as described in the
##    corresponding cdef file 'firmware.cdefs'
## @synopsis
##      set_time(%args);
## INPUTS TO METHOD CALL 'set_time' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## <tr> <td> No switch for this key </td> <td> 'time' </td> <td> '0' </td> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

sub set_time {
    my %args = @_;

    my $resp       = $args{response_object};
    my $raw_output = $resp->get_processed_output();

    _check_error( raw_output => $raw_output, response_object => $resp );

    return [];
}

## @name ::set_console
##
## @summary Parses the raw output from the "set console" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from set_console with any of its options as described in the
##    corresponding cdef file 'firmware.cdefs'
## @synopsis
##      set_console(%args);
## INPUTS TO METHOD CALL 'set_console' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## <tr> <td> No switch for this key </td> <td> 'device-name' </td> <td> '0' </td> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

sub set_console {
    my %args = @_;

    my $resp       = $args{response_object};
    my $raw_output = $resp->get_processed_output();

    _check_error( raw_output => $raw_output, response_object => $resp );

    return [];
}

## @name ::lsmod
##
## @summary Parses the raw output from the "lsmod" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from lsmod with any of its options as described in the
##    corresponding cdef file 'firmware.cdefs'
## @synopsis
##      lsmod(%args);
## INPUTS TO METHOD CALL 'lsmod' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## <tr> <td> '-delete' </td> <td> 'del' </td> <td> '1' </td> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

sub lsmod {
    my %args = @_;

    my $resp       = $args{response_object};
    my $raw_output = $resp->get_processed_output();

    _check_error( raw_output => $raw_output, response_object => $resp );

    return [];
}

## @name ::autoboot
##
## @summary Parses the raw output from the "autoboot" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from autoboot with any of its options as described in the
##    corresponding cdef file 'firmware.cdefs'
## @synopsis
##      autoboot(%args);
## INPUTS TO METHOD CALL 'autoboot' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## <tr> <td> No switch for this key </td> <td> 'device-name' </td> <td> '0' </td> </tr>
## <tr> <td> '-t' </td> <td> 'forever' </td> <td> '1' </td> </tr>
## <tr> <td> '-interruptible' </td> <td> 'interruptible' </td> <td> '1' </td> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

sub autoboot {
    my %args = @_;

    my $resp       = $args{response_object};
    my $raw_output = $resp->get_processed_output();

    _check_error( raw_output => $raw_output, response_object => $resp );

    return [];
}

## @name ::go
##
## @summary Parses the raw output from the "go" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from go with any of its options as described in the
##    corresponding cdef file 'firmware.cdefs'
## @synopsis
##      go(%args);
## INPUTS TO METHOD CALL 'go' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## <tr> <td> '-v' </td> <td> 'verbose' </td> <td> '1' </td> </tr>
## <tr> <td> '-s' </td> <td> 'single-user' </td> <td> '1' </td> </tr>
## <tr> <td> '-m' </td> <td> 'mute' </td> <td> '1' </td> </tr>
## <tr> <td> '-h' </td> <td> 'serial' </td> <td> '1' </td> </tr>
## <tr> <td> '-d' </td> <td> 'kdb' </td> <td> '1' </td> </tr>
## <tr> <td> '-a' </td> <td> 'askme' </td> <td> '1' </td> </tr>
## <tr> <td> '-g' </td> <td> 'invoke-gdb' </td> <td> '1' </td> </tr>
## <tr> <td> '-noclose' </td> <td> 'noclose' </td> <td> '1' </td> </tr>
## <tr> <td> No switch for this key </td> <td> 'address' </td> <td> '0' </td> </tr>
## <tr> <td> '-D' </td> <td> 'multiple' </td> <td> '1' </td> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

sub go {
    my %args = @_;

    my $resp       = $args{response_object};
    my $raw_output = $resp->get_processed_output();

    _check_error( raw_output => $raw_output, response_object => $resp );

    return [];
}

## @name ::boot
##
## @summary Parses the raw output from the "boot" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from boot with any of its options as described in the
##    corresponding cdef file 'firmware.cdefs'
## @synopsis
##      boot(%args);
## INPUTS TO METHOD CALL 'boot' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## <tr> <td> No switch for this key </td> <td> 'protocol' </td> <td> '0' </td> </tr>
## <tr> <td> '-v' </td> <td> 'verbose' </td> <td> '1' </td> </tr>
## <tr> <td> '-s' </td> <td> 'single-user' </td> <td> '1' </td> </tr>
## <tr> <td> '-d' </td> <td> 'kdb' </td> <td> '1' </td> </tr>
## <tr> <td> '-max' </td> <td> 'max' </td> <td> '0' </td> </tr>
## <tr> <td> '-g' </td> <td> 'invoke-gdb' </td> <td> '1' </td> </tr>
## <tr> <td> '-noxsum' </td> <td> 'noxsum' </td> <td> '1' </td> </tr>
## <tr> <td> '-elf' </td> <td> 'elf' </td> <td> '1' </td> </tr>
## <tr> <td> '-elf64' </td> <td> 'elf64' </td> <td> '1' </td> </tr>
## <tr> <td> '-m' </td> <td> 'mute' </td> <td> '1' </td> </tr>
## <tr> <td> '-a' </td> <td> 'askme' </td> <td> '1' </td> </tr>
## <tr> <td> '-h' </td> <td> 'serial' </td> <td> '1' </td> </tr>
## <tr> <td> No switch for this key </td> <td> 'path' </td> <td> '0' </td> </tr>
## <tr> <td> '-z' </td> <td> 'compressd-file' </td> <td> '1' </td> </tr>
## <tr> <td> '-noclose' </td> <td> 'noclose' </td> <td> '1' </td> </tr>
## <tr> <td> '-D' </td> <td> 'multiple' </td> <td> '1' </td> </tr>
## <tr> <td> '-srec' </td> <td> 's-records' </td> <td> '1' </td> </tr>
## <tr> <td> '-loader' </td> <td> 'loader' </td> <td> '0' </td> </tr>
## <tr> <td> '-addr' </td> <td> 'addr' </td> <td> '0' </td> </tr>
## <tr> <td> '-raw' </td> <td> 'raw' </td> <td> '1' </td> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

sub boot {
    my %args = @_;

    my $resp       = $args{response_object};
    my $raw_output = $resp->get_processed_output();

    _check_error( raw_output => $raw_output, response_object => $resp );

    return [];
}

## @name ::load
##
## @summary Parses the raw output from the "load" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from load with any of its options as described in the
##    corresponding cdef file 'firmware.cdefs'
## @synopsis
##      load(%args);
## INPUTS TO METHOD CALL 'load' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## <tr> <td> No switch for this key </td> <td> 'protocol' </td> <td> '0' </td> </tr>
## <tr> <td> '-v' </td> <td> 'verbose' </td> <td> '1' </td> </tr>
## <tr> <td> '-s' </td> <td> 'single-user' </td> <td> '1' </td> </tr>
## <tr> <td> '-d' </td> <td> 'kdb' </td> <td> '1' </td> </tr>
## <tr> <td> '-max' </td> <td> 'max' </td> <td> '0' </td> </tr>
## <tr> <td> '-g' </td> <td> 'invoke-gdb' </td> <td> '1' </td> </tr>
## <tr> <td> '-noxsum' </td> <td> 'noxsum' </td> <td> '1' </td> </tr>
## <tr> <td> '-elf' </td> <td> 'elf' </td> <td> '1' </td> </tr>
## <tr> <td> '-elf64' </td> <td> 'elf64' </td> <td> '1' </td> </tr>
## <tr> <td> '-m' </td> <td> 'mute' </td> <td> '1' </td> </tr>
## <tr> <td> '-a' </td> <td> 'askme' </td> <td> '1' </td> </tr>
## <tr> <td> '-h' </td> <td> 'serial' </td> <td> '1' </td> </tr>
## <tr> <td> No switch for this key </td> <td> 'path' </td> <td> '0' </td> </tr>
## <tr> <td> '-z' </td> <td> 'compressd-file' </td> <td> '1' </td> </tr>
## <tr> <td> '-type' </td> <td> 'module-type' </td> <td> '0' </td> </tr>
## <tr> <td> '-D' </td> <td> 'multiple' </td> <td> '1' </td> </tr>
## <tr> <td> '-srec' </td> <td> 's-records' </td> <td> '1' </td> </tr>
## <tr> <td> '-loader' </td> <td> 'loader' </td> <td> '0' </td> </tr>
## <tr> <td> '-addr' </td> <td> 'addr' </td> <td> '0' </td> </tr>
## <tr> <td> '-raw' </td> <td> 'raw' </td> <td> '1' </td> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

sub load {
    my %args = @_;

    my $resp       = $args{response_object};
    my $raw_output = $resp->get_processed_output();

    _check_error( raw_output => $raw_output, response_object => $resp );

    return [];
}

## @name ::ndp
##
## @summary Parses the raw output from the "ndp" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from ndp with any of its options as described in the
##    corresponding cdef file 'firmware.cdefs'
## @synopsis
##      ndp(%args);
## INPUTS TO METHOD CALL 'ndp' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## <tr> <td> '-d' </td> <td> 'delete' </td> <td> '1' </td> </tr>
## <tr> <td> No switch for this key </td> <td> 'address' </td> <td> '0' </td> </tr>
## <tr> <td> No switch for this key </td> <td> 'ip-address' </td> <td> '0' </td> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

sub ndp {
    my %args = @_;

    my $resp       = $args{response_object};
    my $raw_output = $resp->get_processed_output();

    _check_error( raw_output => $raw_output, response_object => $resp );

    return [];
}

## @name ::ping
##
## @summary Parses the raw output from the "ping" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from ping with any of its options as described in the
##    corresponding cdef file 'firmware.cdefs'
## @synopsis
##      ping(%args);
## INPUTS TO METHOD CALL 'ping' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## <tr> <td> '-c' </td> <td> 'packets' </td> <td> '0' </td> </tr>
## <tr> <td> '-x' </td> <td> 'exit' </td> <td> '1' </td> </tr>
## <tr> <td> No switch for this key </td> <td> 'host' </td> <td> '0' </td> </tr>
## <tr> <td> '-E' </td> <td> 'return' </td> <td> '1' </td> </tr>
## <tr> <td> '-A' </td> <td> 'no-abort' </td> <td> '1' </td> </tr>
## <tr> <td> '-f' </td> <td> 'flood-ping' </td> <td> '1' </td> </tr>
## <tr> <td> '-t' </td> <td> 'forever' </td> <td> '1' </td> </tr>
## <tr> <td> '-s' </td> <td> 'icmp-data' </td> <td> '0' </td> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

sub ping {
    my %args = @_;

    my $resp       = $args{response_object};
    my $raw_output = $resp->get_processed_output();

    _check_error( raw_output => $raw_output, response_object => $resp );

    my @lines = get_lines(
        "raw_output"      => $raw_output,
        "can_throw"       => '1',
        "response_object" => $resp,
    );

    my @result ;

    if(@lines){
       $result[0]->{status} = $lines[0];
       $result[0]->{response} = $lines[1];
    }

    return(\@result);
}

## @name ::arp
##
## @summary Parses the raw output from the "arp" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from arp with any of its options as described in the
##    corresponding cdef file 'firmware.cdefs'
## @synopsis
##      arp(%args);
## INPUTS TO METHOD CALL 'arp' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## <tr> <td> '-d' </td> <td> 'delete' </td> <td> '1' </td> </tr>
## <tr> <td> No switch for this key </td> <td> 'address' </td> <td> '0' </td> </tr>
## <tr> <td> No switch for this key </td> <td> 'ip-address' </td> <td> '0' </td> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

sub arp {
    my %args = @_;

    my $resp       = $args{response_object};
    my $raw_output = $resp->get_processed_output();

    _check_error( raw_output => $raw_output, response_object => $resp );

    my @lines = get_lines(
        "raw_output"      => $raw_output,
        "can_throw"       => '1',
        "response_object" => $resp,
    );

    my @keys;
    my @result;
    my $j = 0;
    while(@lines){
      my $line = shift @lines;
      next if($line =~ /----/);
      my @split = split(/\s\s+/, $line);
      if(!@keys){
         @keys=@split;
         next;
      }
      for (my $i=0; $i<=$#split; $i++ ) {
         $result[$j]->{ to_key_name( item => $keys[$i] ) } = $split[$i];
      }
      $j++ ;
    }

    return(\@result);
}

## @name ::ifconfig
##
## @summary Parses the raw output from the "ifconfig" command
##
## @description 
##    This method can be used for parsing the raw output obtained after calling 
##    the method from ifconfig with any of its options as described in the 
##    corresponding cdef file 'firmware.cdefs'
## @synopsis
##      ifconfig(%args);
## INPUTS TO METHOD CALL 'ifconfig' ON APISET OBJECT 
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## <tr> <td> '-dns' </td> <td> 'dns' </td> <td> '0' </td> </tr> 
## <tr> <td> '-mask' </td> <td> 'subnet-mask' </td> <td> '0' </td> </tr> 
## <tr> <td> '-v6' </td> <td> 'ipv6' </td> <td> '1' </td> </tr> 
## <tr> <td> No switch for this key </td> <td> 'device-name' </td> <td> '0' </td> </tr> 
## <tr> <td> '-gw' </td> <td> 'gateway' </td> <td> '0' </td> </tr> 
## <tr> <td> '-off' </td> <td> 'deactivate' </td> <td> '1' </td> </tr> 
## <tr> <td> '-domain' </td> <td> 'domain' </td> <td> '0' </td> </tr> 
## <tr> <td> '-auto' </td> <td> 'auto' </td> <td> '1' </td> </tr> 
## <tr> <td> '-speed' </td> <td> 'speed' </td> <td> '0' </td> </tr> 
## <tr> <td> No switch for this key </td> <td> 'address' </td> <td> '0' </td> </tr> 
## <tr> <td> '-hwaddr' </td> <td> 'hardware-addr' </td> <td> '0' </td> </tr> 
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

##&nbsp;Command = ifconfig e0M
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;Device e0M: 
##&nbsp;    hwaddr 00-A0-98-54-E5-22, ipaddr not set, mask not set
##&nbsp;    gateway not set, nameserver not set
##&nbsp;    speed 1000 full duplex, loopback not supported
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;          {
##&nbsp;            'gateway' => 'not set',
##&nbsp;            'device' => 'e0M',
##&nbsp;            'nameserver' => 'not set',
##&nbsp;            'hwaddr' => '00-A0-98-54-E5-22',
##&nbsp;            'speed' => '1000 full duplex',
##&nbsp;            'loopback' => 'not supported',
##&nbsp;            'ipaddr' => 'not set',
##&nbsp;            'mask' => 'not set'
##&nbsp;          }
##&nbsp;        ];
##&nbsp;
sub ifconfig {
    my %args = @_;

    my $resp       = $args{response_object};
    my $raw_output = $resp->get_processed_output();

    _check_error( raw_output => $raw_output, response_object => $resp );
    my @lines = get_lines(
        "raw_output"      => $raw_output,
        "can_throw"       => '1',
        "response_object" => $resp,
    );
    my (@result,$line,@arr,@new,$key,$value);
    while(@lines) {
        $line =  shift(@lines);
        @arr  =   split(/[:,]/,$line);
        map(s/^\s+//,@arr);

        foreach(@arr) {
            @new = split(/\s/,$_);
            $key = shift @new;
            $result[0]->{to_key_name( item =>$key)} = join (' ', @new) ;
        }
    }
    return ( \@result );
}

## @name ::show_date
##
## @summary Parses the raw output from the "show date" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from show_date with any of its options as described in the
##    corresponding cdef file 'firmware.cdefs'
## @synopsis
##      show_date(%args);
## INPUTS TO METHOD CALL 'show_date' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

##&nbsp;Command = show date
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;          {
##&nbsp;            'time' => '05:26:07',
##&nbsp;            'date' => '01/31/2012'
##&nbsp;          }
##&nbsp;        ];
##&nbsp;
sub show_date {
    my %args = @_;

    my $resp       = $args{response_object};
    my $raw_output = $resp->get_processed_output();

    _check_error( raw_output => $raw_output, response_object => $resp );

    my @result;
    if ( $raw_output =~ /(date).*(time).*:\s+(\S+)\s+(\S+)/ ) {
        $result[0]->{$1} = $3;
        $result[0]->{$2} = $4;
    }
    return ( \@result );
}

## @name ::show_time
##
## @summary Parses the raw output from the "show time" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from show_time with any of its options as described in the
##    corresponding cdef file 'firmware.cdefs'
## @synopsis
##      show_time(%args);
## INPUTS TO METHOD CALL 'show_time' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

##&nbsp;Command = show time
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;          {
##&nbsp;            'time' => '05:26:08',
##&nbsp;            'date' => '01/31/2012'
##&nbsp;          }
##&nbsp;        ];
##&nbsp;
sub show_time {
    my %args = @_;

    my $resp       = $args{response_object};
    my $raw_output = $resp->get_processed_output();

    _check_error( raw_output => $raw_output, response_object => $resp );

    my @result;
    if ( $raw_output =~ /(date).*(time).*:\s+(\S+)\s+(\S+)/ ) {
        $result[0]->{$1} = $3;
        $result[0]->{$2} = $4;
    }
    return ( \@result );
}

## @name ::show_devices
##
## @summary Parses the raw output from the "show devices" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from show_devices with any of its options as described in the
##    corresponding cdef file 'firmware.cdefs'
## @synopsis
##      show_devices(%args);
## INPUTS TO METHOD CALL 'show_devices' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

##&nbsp;Command = show devices
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;          {
##&nbsp;            'e0a' => 'IBA GE Slot 0402 v1353 (00-A0-98-18-30-5C)',
##&nbsp;            'kcs0a' => 'KCS at 0xca3 (command) and 0xca2 (data)',
##&nbsp;            'e0p' => 'IBA GE Slot 0700 v1353 (00-A0-98-18-30-60)',
##&nbsp;            'e0c' => 'IBA GE Slot 0400 v1353 (00-A0-98-18-30-5E)',
##&nbsp;            'e0d' => 'IBA GE Slot 0401 v1353 (00-A0-98-18-30-5F)',
##&nbsp;            'e0b' => 'IBA GE Slot 0403 v1353 (00-A0-98-18-30-5D)',
##&nbsp;            'u0a.0' => 'MICRON eUSB DISK-(USB 2.0)',
##&nbsp;            'boot_i' => 'u0a.0 alias boot device',
##&nbsp;            'sp0a' => 'Service Processor: Console at 0x3F8, PSI at 0x2F8',
##&nbsp;            'boot0' => 'u0a.0 alias boot device',
##&nbsp;            'e0m' => 'IBA GE Slot 0500 v1353 (00-A0-98-18-30-61)',
##&nbsp;            'clock0a' => 'ISA RTC at 0x70 (index) and 0x71 (target)'
##&nbsp;          }
##&nbsp;        ];
##&nbsp;
sub show_devices {
    my %args = @_;

    my $resp       = $args{response_object};
    my $raw_output = $resp->get_processed_output();

    _check_error( raw_output => $raw_output, response_object => $resp );

    my @lines = get_lines(
        "raw_output"      => $raw_output,
        "can_throw"       => '1',
        "response_object" => $resp,
    );
    my ( @result, $i );
    for ( $i = 2 ; $i <= $#lines ; $i++ ) {
        if ( $lines[$i] =~ /(\S+)\s+(.*)/ ) {
            $result[0]->{ to_key_name( item => $1 ) } = $2;
        }
    }

    return ( \@result );
}

## @name ::savenv
##
## @summary Parses the raw output from the "savenv" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from savenv with any of its options as described in the
##    corresponding cdef file 'firmware.cdefs'
## @synopsis
##      savenv(%args);
## INPUTS TO METHOD CALL 'savenv' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

sub savenv {
    my %args = @_;

    my $resp       = $args{response_object};
    my $raw_output = $resp->get_processed_output();

    _check_error( raw_output => $raw_output, response_object => $resp );

    return [];
}

## @name ::saveenv
##
## @summary Parses the raw output from the "saveenv" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from saveenv with any of its options as described in the
##    corresponding cdef file 'firmware.cdefs'
## @synopsis
##      saveenv(%args);
## INPUTS TO METHOD CALL 'saveenv' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

sub saveenv {
    my %args = @_;

    my $resp       = $args{response_object};
    my $raw_output = $resp->get_processed_output();

    _check_error( raw_output => $raw_output, response_object => $resp );

    return [];
}

## @name ::unsetenv
##
## @summary Parses the raw output from the "unsetenv" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from unsetenv with any of its options as described in the
##    corresponding cdef file 'firmware.cdefs'
## @synopsis
##      unsetenv(%args);
## INPUTS TO METHOD CALL 'unsetenv' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## <tr> <td> No switch for this key </td> <td> 'varname' </td> <td> '0' </td> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

sub unsetenv {
    my %args = @_;

    my $resp       = $args{response_object};
    my $raw_output = $resp->get_processed_output();

    _check_error( raw_output => $raw_output, response_object => $resp );

    return [];
}

## @name ::set_defaults
##
## @summary Parses the raw output from the "set-defaults" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from set_defaults with any of its options as described in the
##    corresponding cdef file 'firmware.cdefs'
## @synopsis
##      set_defaults(%args);
## INPUTS TO METHOD CALL 'set_defaults' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

sub set_defaults {
    my %args = @_;

    my $resp       = $args{response_object};
    my $raw_output = $resp->get_processed_output();

    _check_error( raw_output => $raw_output, response_object => $resp );

    return [];
}

## @name ::setenv
##
## @summary Parses the raw output from the "setenv" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from setenv with any of its options as described in the
##    corresponding cdef file 'firmware.cdefs'
## @synopsis
##      setenv(%args);
## INPUTS TO METHOD CALL 'setenv' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## <tr> <td> No switch for this key </td> <td> 'value' </td> <td> '0' </td> </tr>
## <tr> <td> No switch for this key </td> <td> 'varname' </td> <td> '0' </td> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

sub setenv {
    my %args = @_;

    my $resp       = $args{response_object};
    my $raw_output = $resp->get_processed_output();

    _check_error( raw_output => $raw_output, response_object => $resp );

    return [];
}

## @name ::printenv
##
## @summary Parses the raw output from the "printenv" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from printenv with any of its options as described in the
##    corresponding cdef file 'firmware.cdefs'
## @synopsis
##      printenv(%args);
## INPUTS TO METHOD CALL 'printenv' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## <tr> <td> No switch for this key </td> <td> 'verbose varname' </td> <td> '0' </td> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

##&nbsp;Command = printenv
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;          {
##&nbsp;            'boot_flash' => 'flash0a',
##&nbsp;            'gx_diag_url' => 'fat://boot0/X86_64/freebsd/image2/kernel',
##&nbsp;            'last_os_booted_wafl_ver' => undef,
##&nbsp;            'arch' => 'X86_64',
##&nbsp;            'sys_model' => 'FAS2240-4',
##&nbsp;            'bios_version' => '8.1X1',
##&nbsp;            'net_nameserver' => '10.60.132.40',
##&nbsp;            'autoboot' => 'true',
##&nbsp;            'cpu_speed' => '1730',
##&nbsp;            'mobo_serial_num' => '5006167379',
##&nbsp;            'bt_bios_version' => '8.1X1',
##&nbsp;            'boot_primary' => 'setenv BOOTED_FROM PRIMARY; boot -elf64 $GX_PRIMARY_KERNEL_URL $PRIMARY_KERNEL_URL',
##&nbsp;            'firmware_url' => 'fat://boot0/X86_64/firmware/SB_XIX/firmware.img',
##&nbsp;            'boot_diags' => 'boot -sld -elf64 $GX_DIAG_URL $PRIMARY_KERNEL_URL',
##&nbsp;            'net_ipaddr' => '10.60.172.221',
##&nbsp;            'last_os_booted_ver' => undef,
##&nbsp;            'bios_interface' => '9FC3',
##&nbsp;            'bootarg_init_clearvarfsnvram' => undef,
##&nbsp;            'net_netmask' => '255.255.255.0',
##&nbsp;            'savenv' => 'saveenv',
##&nbsp;            'boot_backup' => 'setenv BOOTED_FROM BACKUP; boot -elf64 $GX_BACKUP_KERNEL_URL $BACKUP_KERNEL_URL',
##&nbsp;            'fmmbx_lkg_0' => undef,
##&nbsp;            'bios_date' => '10/06/2011',
##&nbsp;            'boot_ontap' => 'autoboot boot0',
##&nbsp;            'auto_fw_update' => 'true',
##&nbsp;            'bootarg_kcs_clam_partnerstate' => undef,
##&nbsp;            'boot_server' => '10.60.132.162',
##&nbsp;            'ntap_init_kernelname' => undef,
##&nbsp;            'bootarg_init_boot_clustered' => undef,
##&nbsp;            'bootarg_bsdportname' => undef,
##&nbsp;            'primary_kernel_url' => 'fat://boot0/X86_64/kernel/primary.krn',
##&nbsp;            'gx_backup_kernel_url' => 'fat://boot0/X86_64/freebsd/image1/kernel',
##&nbsp;            'ldkern' => 'load -elf64 $GX_PRIMARY_KERNEL_URL $PRIMARY_KERNEL_URL',
##&nbsp;            'netboot' => 'setenv BOOTED_FROM NETWORK; boot -elf64',
##&nbsp;            'cpu_type' => 'Intel(R) Xeon(R) CPU           C3528  @ 1.73GHz',
##&nbsp;            'boardname' => 'SB_XIX',
##&nbsp;            'reboot_reason' => 'REBOOT_HALT_CMD',
##&nbsp;            'use_secondary' => 'true',
##&nbsp;            'net_gateway' => '10.60.172.1',
##&nbsp;            'booted_from' => 'OTHER',
##&nbsp;            'backup_kernel_url' => 'fat://boot0/backup/X86_64/kernel/primary.krn',
##&nbsp;            'net_domain' => 'rtp.netapp.com',
##&nbsp;            'sys_rev' => 'A1',
##&nbsp;            'boot_console' => 'sp0a',
##&nbsp;            'fmmbx_lkg_0b' => undef,
##&nbsp;            'version' => 'printenv BIOS_VERSION LOADER_VERSION',
##&nbsp;            'mobo_model' => 'FAS',
##&nbsp;            'loader_version' => '4.1X1',
##&nbsp;            'update_flash' => 'flash -backup $FIRMWARE_URL flash0a',
##&nbsp;            'env_version' => '1',
##&nbsp;            'mobo_rev' => 'A1',
##&nbsp;            'gx_primary_kernel_url' => 'fat://boot0/X86_64/freebsd/image2/kernel',
##&nbsp;            'net_device' => 'e0M',
##&nbsp;            'sys_serial_num' => '500000243604',
##&nbsp;            'partner_sysid' => undef,
##&nbsp;            'bt_loader_version' => '4.1X1',
##&nbsp;            'autoboot_from' => 'PRIMARY',
##&nbsp;            'cpu_num_cores' => '4',
##&nbsp;            'kk' => '00',
##&nbsp;            'last_os_booted_raid_ver' => undef
##&nbsp;          }
##&nbsp;        ];
##&nbsp;
sub printenv {
    my %args = @_;

    my $resp       = $args{response_object};
    my $raw_output = $resp->get_processed_output();

    _check_error( raw_output => $raw_output, response_object => $resp );

    my @lines = get_lines(
        "raw_output"      => $raw_output,
        "can_throw"       => '1',
        "response_object" => $resp,
    );
    my ( @result, $i );
    for ( $i = 2 ; $i <= $#lines ; $i++ ) {
        if ( $lines[$i] =~ /(\S+)\s+(.*)/ ) {
            my $jey = $1;
            my $val = $2;
            $jey =~ s/[\.-]/_/g;
            $result[0]->{ to_key_name( item => $jey ) } = $val;
        }
    }

    return ( \@result );
}

## @name ::sp_status
##
## @summary Parses the raw output from the "sp status" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from sp_status with any of its options as described in the
##    corresponding cdef file 'firmware.cdefs'
## @synopsis
##      sp_status(%args);
## INPUTS TO METHOD CALL 'sp_status' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

##&nbsp;Command = sp status
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;          {
##&nbsp;            'ipv4_settings' => [
##&nbsp;                                 {
##&nbsp;                                   'gateway' => '172.16.52.1',
##&nbsp;                                   'netmask' => '255.255.254.0',
##&nbsp;                                   'ip_address' => '172.16.52.139',
##&nbsp;                                   'using_dhcp' => 'NO'
##&nbsp;                                 }
##&nbsp;                               ],
##&nbsp;            'ipv6' => 'Disabled',
##&nbsp;            'ethernet_link' => 'up, 100 Mb, full duplex, auto-neg complete',
##&nbsp;            'firmware_version' => '1.4X6',
##&nbsp;            'mgmt_mac_address' => '00:A0:98:1A:50:CC'
##&nbsp;          }
##&nbsp;        ];
##&nbsp;
sub sp_status {
    my %args = @_;

    my $resp       = $args{response_object};
    my $raw_output = $resp->get_processed_output();

    _check_error( raw_output => $raw_output, response_object => $resp );

    my @array_lines = get_lines(
        "raw_output"      => $raw_output,
        "can_throw"       => '1',
        "response_object" => $resp,
    );
    my @result;
    my %hash;
    my ( $line, $key, $value, $key1, $value1 );

    for ( $line = 0 ; $line <= $#array_lines ; $line++ ) {
        if (   $array_lines[$line] =~ /(Firmware\s+Version):\s+(.*)/
            || $array_lines[$line] =~ /(IPv4\s+\S+)/
            || $array_lines[$line] =~ /(IPv6\s+\S+)/
            || $array_lines[$line] =~ /(Link\s+Local\s+Address)/
            || $array_lines[$line] =~ /(Router\s+Advertisement\s+Address)/ )
        {
            $key   = $1;
            $value = $2;
            $key   = to_key_name( item => $key );
            if ( $key =~ /firmware.*/ ) {
                $hash{$key} = $value;
            }
            elsif ( $key =~ /ipv6_settings/ ) {
                $line++;
            }
            $line++;
            if ( $array_lines[$line] ) {    #skipping blank lines in output
                while ( $array_lines[$line] =~ /\s*(.*):\s+(.*)/ ) {
                    if ( $2 eq "Disabled" ) {
                        $hash{ to_key_name( item => $1 ) } = $2;
                        last;
                    }
                    elsif ( $key eq "firmware_version" ) {
                        $key1   = $1;
                        $value1 = $2;
                        $key1 =~ /(Ethernet\s+Link)/
                          ? $hash{ to_key_name( item => $1 ) }    = $value1
                          : $hash{ to_key_name( item => $key1 ) } = $value1;
                    }
                    else {
                        $hash{$key}->[0]->{ to_key_name( item => $1 ) } = $2;
                    }
                    $line++;
                }
            }
            $line--;
        }
    }
    push( @result, \%hash );
    return \@result;
}

## @name ::sp_setup
##
## @summary Parses the raw output from the "sp setup" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from sp_setup with any of its options as described in the
##    corresponding cdef file 'firmware.cdefs'
## @synopsis
##      sp_setup(%args);
## INPUTS TO METHOD CALL 'sp_setup' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

sub sp_setup {
    my %args = @_;

    my $resp       = $args{response_object};
    my $raw_output = $resp->get_processed_output();

    _check_error( raw_output => $raw_output, response_object => $resp );

    return [];
}

## @name ::pelog
##
## @summary Parses the raw output from the "pelog" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from pelog with any of its options as described in the
##    corresponding cdef file 'firmware.cdefs'
## @synopsis
##      pelog(%args);
## INPUTS TO METHOD CALL 'pelog' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## <tr> <td> '-t' </td> <td> 'record-type' </td> <td> '0' </td> </tr>
## <tr> <td> '-R' </td> <td> 'raw-read' </td> <td> '1' </td> </tr>
## <tr> <td> '-m' </td> <td> 'set-mode' </td> <td> '1' </td> </tr>
## <tr> <td> '-g' </td> <td> 'region-type' </td> <td> '0' </td> </tr>
## <tr> <td> '-v' </td> <td> 'validate-region' </td> <td> '1' </td> </tr>
## <tr> <td> '-i' </td> <td> 'initialize-region' </td> <td> '1' </td> </tr>
## <tr> <td> '-o' </td> <td> 'offset' </td> <td> '0' </td> </tr>
## <tr> <td> '-r' </td> <td> 'read-region' </td> <td> '1' </td> </tr>
## <tr> <td> '-a' </td> <td> 'region' </td> <td> '1' </td> </tr>
## <tr> <td> '-x' </td> <td> 'data-size' </td> <td> '0' </td> </tr>
## <tr> <td> '-h' </td> <td> 'region-header' </td> <td> '1' </td> </tr>
## <tr> <td> '-W' </td> <td> 'raw-write' </td> <td> '1' </td> </tr>
## <tr> <td> '-l' </td> <td> 'log-string' </td> <td> '0' </td> </tr>
## <tr> <td> '-z' </td> <td> 'access-mode' </td> <td> '0' </td> </tr>
## <tr> <td> '-w' </td> <td> 'write-region' </td> <td> '1' </td> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

##&nbsp;Command = pelog -v -g=2 -o=20 -x=2 -z=1
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;Region 2 is valid
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [];
##&nbsp;

sub pelog {
    my %args = @_;

    my $resp       = $args{response_object};
    my $raw_output = $resp->get_processed_output();

    _check_error( raw_output => $raw_output, response_object => $resp );

    return [];
}
## @name ::mptdump
##
## @summary Parses the raw output from the "mptdump" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from pelog with any of its options as described in the
##    corresponding cdef file 'firmware.cdefs'
## @synopsis
##     mptdump(%args);
sub mpt_dump {
    my %args       = @_;
    my $resp       = $args{response_object};
    my $raw_output = $resp->get_processed_output();

    _check_error( raw_output => $raw_output, response_object => $resp );

    my @lines = get_lines(
        "raw_output"      => $raw_output,
        "response_object" => $resp,
    );
    my @result;
    my $j = 0;
    my $k = 0;
    my $m = 0;

    if ( $lines[$j] =~ /MP Floating Pointer.*/ ) {
        shift(@lines);
        my $str = "mp_floating_pointer_structure";
        $result[0]->{$str}->{loc} = "0xF00E0/0xF00E0";
		while($lines[$j] !~ /MP Confi?guration Table.*/ and scalar(@lines)){
            my @values = split( '\s+=', $lines[$j] );
            $values[1] =~ s/^\s+//;
			$values[0]=~tr/.//d;
            $result[0]->{$str}->{ to_key_name( item => $values[0] ) } =
              $values[1];
            $j++;
        }
    }
    if ( $lines[$j] =~ /MP Confi?guration Table.*/ ) {
        my $str = "mp_configuaration_table";
        $result[1]->{$str}->{header} = "0xE0840/0xE0840";
        $j++;
        while ( $lines[$j] !~ /^=/ and scalar(@lines)) {
            my @values = split( '\s+=', $lines[$j] );
            $values[1] =~ s/^\s+//;
			$values[0]=~tr/.//d;
            $result[1]->{$str}->{ to_key_name( item => $values[0] ) } =
              $values[1];
            $j++;

        }
        $k = $j;
        while ( $k <= $#lines ) {
            if ( $lines[$j] =~ /^=/ ) {
                if ( $lines[$j] =~ /NetApp OEM Table.*/ ) {
                    $j++;
                    $str = "netapp_oem_table";
                    $result[1]->{$str}->{loc} = "0xE10A0/0xE10A0";
                    while ( $lines[$j] !~ /^=/ and scalar(@lines)) {
                        my @values = split( '\s+=', $lines[$j] );
                        $values[1] =~ s/^\s+//;
						$values[0]=~tr/.//d;
                        $result[1]->{$str}
                          ->{ to_key_name( item => $values[0] ) } = $values[1];
                        $j++;

                    }
                }

                my $str1 = $lines[$j];
                $str1 =~ s/\=//g;
                $str1 =~ s/^\s+//;
                $j++;
                while ( $lines[$j] =~ /\w.*\s+=\.*/ and scalar(@lines)) {

                    my @values = split( '\s+=', $lines[$j] );
                    $values[1] =~ s/^\s+//;
					$values[0]=~tr/.//d;
                    $result[1]->{$str}->{ to_key_name( item => $str1 ) }
                      ->{ to_key_name( item => $values[0] ) } = $values[1];
                    $j++;

                }

                if ( $lines[$j] =~ /\w+:\s.*/ ) {
                    my @val1 = split( ':\s+', $lines[$j] );
                    $val1[1] =~ s/^\s+//;
					$val1[0]=~tr/.//d;
                    $result[1]->{$str}->{ to_key_name( item => $str1 ) }
                      ->{ to_key_name( item => $val1[0] ) } = $val1[1];
                    $j++;

                }

            }
            $k++;
        }

    }
    return \@result;

}
## @name ::show_pci
##
## @summary Parses the raw output from the "show_pci" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from show pci with any of its options as described in the
##    corresponding cdef file 'firmware.cdefs'
## @synopsis
##     show_pic(%args);
sub show_pci {
    my %args       = @_;
    my $resp       = $args{response_object};
    my $raw_output = $resp->get_processed_output();

    _check_error( raw_output => $raw_output, response_object => $resp );

    my @lines = get_lines(
        "raw_output"      => $raw_output,
        "response_object" => $resp,
    );
    my @result;
    for(my $j=0;$j<=$#lines;$j++){
	my @values1=split(":", $lines[$j]);
        my @values2=split('\s+',$values1[0]);
        $values1[1]=~ s/^\s+//;
        if($lines[$j] =~ /^PCI/) {
            $result[$j]->{pci_name } = $values2[1];
        }else {
            $result[$j]->{pci_name } = $values2[0];
        }
	$result[$j]->{dev_name } = $values1[1];
    }
    return \@result;
}

## @name ::bios_status
##
## @summary Parses the raw output from the "bios status" command
##
## @description 
##    This method can be used for parsing the raw output obtained after calling 
##    the method from bios_status with any of its options as described in the 
##    corresponding cdef file 'firmware.cdefs'
## @synopsis
##      bios_status(%args);
## INPUTS TO METHOD CALL 'bios_status' ON APISET OBJECT 
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples
##&nbsp;Command = bios status
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;BIOS boot time: 37469 ms 
##&nbsp;Current boot BIOS image: Primary 
##&nbsp;Previous boot BIOS image: Primary 
##&nbsp;Console redirection enabled: Yes 
##&nbsp;SP present: Yes 
##&nbsp;Profile: 00 
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;          {
##&nbsp;            'profile' => '00',
##&nbsp;            'previous_boot_bios_image' => 'Primary',
##&nbsp;            'console_redirection_enabled' => 'Yes',
##&nbsp;            'current_boot_bios_image' => 'Primary',
##&nbsp;            'sp_present' => 'Yes',
##&nbsp;            'bios_boot_time' => '37469 ms'
##&nbsp;          }
##&nbsp;        ];
##&nbsp;
sub bios_status  {
    my %args = @_;
    my $resp        = $args{response_object};
    my $output      = $resp->get_processed_output();
    my $command     = $resp->get_actual_command();
    my $error_level = $resp->get_execution_status();
    _check_error(
        "response_object" => $resp,
        "raw_output"      => $output,
        "command"         => $command,
    );
    my ( %hash,   @result_array );
    my @lines = get_lines( raw_output => $output );
    foreach my $line (@lines) {
        if ($line =~ /^(.*)\s*[:]+\s*(.*)$/) {
            my ($key, $value) = ($1, $2);
            $hash{to_key_name( item =>$key)} = $value;
        }
    }
    push( @result_array, \%hash );
    return \@result_array;
}

## @name ::sp_reboot
##
## @summary Parses the raw output from the "sp reboot" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from sp_reboot with any of its options as described in the
##    corresponding cdef file 'sp.cdefs'
## @synopsis
##      sp_reboot(%args);
## INPUTS TO METHOD CALL 'sp_reboot' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## <tr> <td> 'primary' </td> <td> 'primary' </td> <td> '1' </td> </tr>
## <tr> <td> 'backup' </td> <td> 'backup' </td> <td> '1' </td> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

sub sp_reboot {
    my %args = @_;
    my $resp        = $args{response_object};
    my $output      = $resp->get_processed_output();
    my $command     = $resp->get_actual_command();
    _check_error(
        "response_object" => $resp,
        "raw_output"      => $output,
        "command"         => $command,
    );
    return [];
}

sub platform_info {
    my %args = @_;

    my $resp       = $args{response_object};
    my $raw_output = $resp->get_processed_output();

    _check_error( raw_output => $raw_output, response_object => $resp );

    my @array_lines = get_lines(
        "raw_output"      => $raw_output,
        "can_throw"       => '1',
        "response_object" => $resp,
    );

    my @result;
    my %hash;
    my %hash1;
    my ( $line, @values1, $key );

    for ( $line = 0; $line <= $#array_lines; $line++ ) {

        my @values = split( /\s*:\s*/, $array_lines[$line] );

        if ( $values[0] =~ /DIMM\s+Slot\s+information/i ) {
            $line++;
            while ( $line <= $#array_lines ) {
                @values1 = split( /\s*:\s+/, $array_lines[$line] );
	        $values1[0] =~ s/\[|\]//g;
                if ($values1[0] =~ /Channel/i) {
                    $key = $values1[0];
                    $hash1{ to_key_name( item => $key ) }->[0] =    $values1[1];
                    $line++;

                }
                else {
                    $hash1{ to_key_name( item => $key ) }->[1]->{ to_key_name( item => $values1[0] ) } = $values1[1];
                    $line++;
                }
            }
            $hash{ to_key_name( item => $values[0] ) } = \%hash1;
        } else {
            $hash{ to_key_name( item => $values[0] ) } = $values[1];
       }
    }
    push( @result, \%hash );
    return \@result;
}

## @name ::dir
##
## @summary Parses the raw output from the "dir" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from dir with any of its options as described in the
##    corresponding cdef file 'firmware.cdefs'
## @synopsis
##      dir(%args);
## INPUTS TO METHOD CALL 'dir' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

sub dir{
    my %args = @_;

    my $resp       = $args{response_object};
    my $raw_output = $resp->get_processed_output();

    _check_error( raw_output => $raw_output, response_object => $resp );

    my @array_lines = get_lines(
        "raw_output"      => $raw_output,
        "can_throw"       => '1',
        "response_object" => $resp,
    );
    my $i = 0;
    my $line;
    my %hash;
    my @array;
    while ( $line = shift(@array_lines)){
        my @split = split(/\s+is\s+/, $line);
        $hash{ to_key_name( item => $split[0] ) } = $split[1];
	if ( $line =~ /File\s+System\s+type/i ) {
	    $line = shift @array_lines;
	    while ( $line = shift(@array_lines) ) {
                my @split1 = split(/\s{2,}/, $line);
                $array[$i]->{'directory'} = $split1[0];
                $array[$i]->{'info'} = $split1[1];
                $array[$i]->{'date'} = $split1[2];
                $array[$i]->{'time'} = $split1[3];
                $array[$i]->{'size'} = $split1[4];
	        $i++ ;
	    }
	}
    }
    push @array , \%hash;
    return \@array ;
}

## @name ::show_heap
##
## @summary Parses the raw output from the "show heap" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from show_heap with any of its options as described in the
##    corresponding cdef file 'firmware.cdefs'
## @synopsis
##      show_heap(%args);
## INPUTS TO METHOD CALL 'show_heap' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

sub show_heap{
 my %args = @_;

    my $resp       = $args{response_object};
    my $raw_output = $resp->get_processed_output();

    _check_error( raw_output => $raw_output, response_object => $resp );

    my @array_lines = get_lines(
        "raw_output"      => $raw_output,
        "can_throw"       => '1',
        "response_object" => $resp,
    );
    my (@array,$line,%hash,%hash2);
    while ( $line = shift(@array_lines)){
	my @split = split(/:\s+/,$line);
	$hash{ to_key_name( item => $split[0] ) } = $split[1];
	if ( $line =~ /Heap/i ) {
	    last if ( !$line );
	    while ( $line = shift(@array_lines) ) {
		my @split1 = split(/:\s+/,$line);
		$hash2{ to_key_name( item => $split1[0] ) } = $split1[1];
	    }
	push @array , \%hash2;
	}
    }
    push @array , \%hash;
    return \@array ;
}

## @name ::help
##
## @summary parses the raw output from the "help" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from help with any of its options as described in the
##    corresponding cdef file 'firmware.cdefs'
## @synopsis
##      help(%args);
## INPUTS TO METHOD CALL 'help' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

sub help{
    my %args = @_;
    my @result;
    my $resp       = $args{response_object};
    my $raw_output = $resp->get_processed_output();

    _check_error( raw_output => $raw_output, response_object => $resp );

    my @array_lines = get_lines(
        "raw_output"      => $raw_output,
        "can_throw"       => '1',
        "response_object" => $resp,
    );

    my $i;
    my %hash;
    my @values;
    my $line;
    while( $line = shift(@array_lines) ){
        @values = split(/\s{2,}/,$line);
        $hash{ to_key_name( item => $values[0] ) } = $values[1];
    }
    push( @result, \%hash );
    return(\@result);
}

## @name ::priv_show
##
## @summary Parses the raw output from the "priv show" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from priv_show with any of its options as described in the
##    corresponding cdef file 'firmware.cdefs'
## @synopsis
##      priv_show(%args);
## INPUTS TO METHOD CALL 'priv_show' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

sub priv_show {
    my %args = @_;

    my $resp       = $args{response_object};
    my $raw_output = $resp->get_processed_output();

    _check_error( raw_output => $raw_output, response_object => $resp );

    my @lines = get_lines(
        "raw_output"      => $raw_output,
        "can_throw"       => '1',
        "response_object" => $resp,
    );
    my @result ;

    if(@lines){
       $result[0]->{result} = $lines[0];
    }
    return(\@result);

}
## @name ::priv_set
##
## @summary Parses the raw output from the "priv set" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from priv_set with any of its options as described in the
##    corresponding cdef file 'firmware.cdefs'
## @synopsis
##      priv_set(%args);
## INPUTS TO METHOD CALL 'priv_set' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

sub priv_set {
    my %args = @_;

    my $resp       = $args{response_object};
    my $raw_output = $resp->get_processed_output();

    _check_error( raw_output => $raw_output, response_object => $resp );

    my @lines = get_lines(
        "raw_output"      => $raw_output,
        "can_throw"       => '1',
        "response_object" => $resp,
    );

    my @result ;

    if(@lines){
       $result[0]->{result} = $lines[0];
    }
    return(\@result);
}

1;
