#
# Copyright (c) 2017 NetApp, Inc., All Rights Reserved
# Any use, modification, or distribution is prohibited
# without prior written consent from NetApp, Inc.
#
##
## @synopsis
##    use bmc;
##
##    bmc_status(%args);
##    bmc_version(%args);
##    priv_set(%args);
##    priv_show(%args);
##    bmc_reboot(%args);
##    system_power(%args);
##    system_reset(%args);
##    system_console(%args);
##    bmc_events(%args);
##    bmc_update(%args);
##    bmc_shell(%args);
##    bmc_uptime(%args);
##    bmc_fw_test(%args);
##    bmc_switch_dump(%args);
##    bmc_switch_stat(%args);
##    bmc_switch_read(%args);
##    bmc_switch_write(%args);
##    bmc_switch_read_phy_88E6131(%args);
##    bmc_switch_write_phy_88E6131(%args);
##    bmc_switch_read_phyextnreg_88E6131(%args);
##    bmc_switch_write_phyextnreg_88E6131(%args);
##    bmc_switch_read_switchphy_BMC53101M(%args);
##    bmc_switch_write_switchphy_BMC53101M(%args);
##    bmc_switch_version(%args);
##    bmc_switch_update(%args);
##    bmc_ddr_version(%args);
##    bmc_ddr_update(%args);
##    bmc_hw_test(%args);
##    bmc_loopback_test(%args);
##    bmc_extended_self_test(%args);
##    bmc_self_test(%args);
##    bmc_log_history_show(%args);
##    bmc_switch_get_all_regs_tst(%args);
##    bmc_devices_version(%args);
##
## @summary Routines to parse BMC command output
## @auther dl-nacl-dev@netapp.com

package NACL::APISet::Node::CLI::SP::default::bmc;

###################
# Everytime
###################
use strict;
use warnings;
use NACL::APISet::Response::CLI::ParserUtils;

# Include exceptions which will be thrown

use NACL::APISet::Exceptions::InvalidParamException qw(:try);
use NACL::APISet::Exceptions::InvalidParamValueException qw(:try);
use NACL::APISet::Exceptions::CommandNotFoundException qw(:try);
use NACL::APISet::Exceptions::CommandFailedException qw(:try);
use Tharn;
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();

# Description

#  This Parser is used to run bmc status command on the BMC filer. The options -d and -v is taken care of.

## @name ::bmc_status
##
## @summary Parses the raw output from the "bmc status" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from bmc_status with any of its options as described in the
##    corresponding cdef file 'bmc.cdefs'
## @synopsis
##      bmc_status(%args);
## INPUTS TO METHOD CALL 'bmc_status' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## <tr> <td> '-v' </td> <td> 'verbose' </td> <td> '1' </td> </tr>
## <tr> <td> '-d' </td> <td> 'debug' </td> <td> '1' </td> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

##&nbsp;Command = bmc status
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;Firmware Version:     2.0
##&nbsp;Mgmt MAC Address:     00:A0:98:18:31:52
##&nbsp;Ethernet Link:        up
##&nbsp;  Using DHCP:         no
##&nbsp;IPv4 configuration:
##&nbsp;  IP Address:         172.16.53.23
##&nbsp;  Netmask:            255.255.254.0
##&nbsp;  Gateway:            172.16.52.1
##&nbsp;IPv6 configuration:   Disabled
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;          {
##&nbsp;            'ipv6_configuration' => 'Disabled',
##&nbsp;            'ethernet_link' => 'up',
##&nbsp;            'firmware_version' => '2.0',
##&nbsp;            'ipv4_configuration' => [
##&nbsp;                                      {
##&nbsp;                                        'gateway' => '172.16.52.1',
##&nbsp;                                        'netmask' => '255.255.254.0',
##&nbsp;                                        'ip_address' => '172.16.53.23'
##&nbsp;                                      }
##&nbsp;                                    ],
##&nbsp;            'using_dhcp' => 'no',
##&nbsp;            'mgmt_mac_address' => '00:A0:98:18:31:52'
##&nbsp;          }
##&nbsp;        ];
##&nbsp;
sub bmc_status {
    my %args = @_;
    my $error_message;
    my $resp    = $args{response_object};
    my $output  = $resp->get_processed_output();
    my $command = $resp->get_actual_command();
    my %info    = ();
    my $error   = _parse_for_errors(
        "raw_output"    => $output,
        "command"       => $command,
        "error_message" => \$error_message
    );
    if ($error) {
        $error = 'NACL::APISet::Exceptions::' . $error . 'Exception';
        $error->throw(text => $error_message, response_object => $resp);
    }

    #Now we can parse the output
    my @lines
        = get_lines("raw_output" => $output, "response_object" => $resp);
    my ($key1, $key2, @Res, %Output, %Hash_Flag_Out, @Arr_2, $key3, $value);
    my $i           = 0;
    my $Flag_ipv4        = 0;
    my $Flag_ipv6       = 0;
    my $flag_global = 0;
    my $flag_local  = 0;
    my $flag_router = 0;
    my @hist_arr;
    my $k = 0;
    foreach $value (@lines) {

        if ($value =~ /Firmware Version/) {
            my @head = split(/:\s+/, $value, 2);
            $Res[$i]->{to_key_name(item => $head[0])} = $head[1];
        }
        if ($value =~ /^([\w\s]+):\s+([0-9a-zA-Z\:\.\,\-\s]+)$/) {
            if ($Flag_ipv4 == 0) {
                $Res[$i]->{to_key_name(item => $1)} = $2;
            } elsif ($value =~ /^(IPv6 configuration):\s+(\w+)$/) {
                $Res[$i]->{to_key_name(item => $1)} = $2;
            }
        } elsif ($value =~ /^(IPv4 configuration):$/) {
            $Flag_ipv4 = 1;
            $key1 = to_key_name(item => $1);
            next;
        } elsif ($value =~ /^(IPv6 configuration):$/) {
            $Flag_ipv6 = 1;
            $key1 = to_key_name(item => $1);
            next;
        } elsif ( ($value =~ /^(\w+.*[Dd]aemon)$/) || ($value =~ /^(\w+:)$/) ) {
            $key2 = to_key_name(item => $1);
            next;
        } elsif ($value
            =~ /------------------------------------------------------------------/
            ) {
            next;
        } elsif ( ($value =~ /^(.*)\s+(\d+|NA)$/)  || ($value =~ /^(Link Status word|Speed and Duplex Bits|Partner Speed and Duplex Bits|Partner Flow Control Bits)\s+(.*)$/) ) {
	    my $sub_key = $1;
	    my $sub_value = $2;
	    $sub_key =~ s/^\s+|\s+$//g;
	    $sub_value =~ s/^\s+|\s+$//g;
            $Res[$i]->{"$key2"}->[0]->{to_key_name(item => $sub_key)} = $sub_value;
        }
        if ($Flag_ipv4 == 1) {
	    if (($value =~ /^(IP Address|Netmask|Gateway)(.*):(.*)$/) && ($value !~ /^Gateway Address/)) {
		my ($sub_key1,$sub_key2,$sub_value) = ($1,$2,$3);
		$sub_key1 =~ s/^\s+|\s+$//g;
		$sub_key2  =~ s/^\s+|\s+$//g;
		$sub_value  =~ s/^\s+|\s+$//g if (defined $sub_value);
		if (!grep {$_ eq $sub_key2} @hist_arr){
		    $k++ if (@hist_arr);
		    push (@hist_arr,$sub_key2);
		}

		$Res[$i]->{"$key1"}->[$k]->{to_key_name(item => $sub_key1)} = $sub_value;
	    }
        }
        if ($Flag_ipv6 == 1) {
            if ($value
                =~ /^(Global IP|Prefix Length|Gateway Address|Router Assigned IP|Link Local IP|Router IP):\s+([A-Za-z0-9\.\:]+)$/
                ) {
                if ($value =~ /^(Global IP):\s+([A-Za-z0-9\.\:]+)$/) {
                    $Res[$i]->{"$key1"}->[0]->{global}->{ip_address} = $2;
                    $flag_global = 1;
                } elsif ($value =~ /^(Gateway Address):\s+([A-Za-z0-9\.\:]+)$/) {
                    $Res[$i]->{"$key1"}->[0]->{gateway} = $2;
                    $flag_local = 1;
                } elsif ($value =~ /^(Link Local IP):\s+([A-Za-z0-9\.\:]+)$/) {
                    $Res[$i]->{"$key1"}->[0]->{local}->{ip_address} = $2;
                    $flag_local = 1;
                } elsif ($value =~ /^(Router IP):\s+([A-Za-z0-9\.\:]+)$/) {
                    $Res[$i]->{"$key1"}->[0]->{'router'}->{ip_address} = $2 ;
                    $flag_router = 1 ;
                } elsif ($value =~ /(Prefix Length):\s+([A-Za-z0-9\.\:]+)$/) {
                    if ($flag_global == 1) {
                        $Res[$i]->{"$key1"}->[0]->{global}->{prefixlen} = $2;
                        $flag_global = 0;
                    }
                    if ($flag_local == 1) {
                        $Res[$i]->{"$key1"}->[0]->{local}->{prefixlen} = $2;
                        $flag_local = 0;
                    }
                    if($flag_router == 1) {
                        $Res[$i]->{"$key1"}->[0]->{'router'}->{prefixlen} = $2;
                        $flag_router = 0;
                    }
                } else {
                    $Res[$i]->{"$key1"}->[0]->{to_key_name(item => $1)} = $2;
                }
            }
        }
    }
 return \@Res;
}

## @name ::bmc_version
##
## @summary Parses the raw output from the "bmc version" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from bmc version with any of its options as described in the
##    corresponding cdef file 'bmc.cdefs'
## @synopsis
##      bmc_version(%args);
## INPUTS TO METHOD CALL 'bmc version' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples
##&nbsp;Command = version; result -l
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;version
##&nbsp;Booted primary firmware version 2.2X1
##&nbsp;Booted primary ACP version 01.10
##&nbsp;Primary firmware version 2.2X1
##&nbsp;Primary ACP version 01.10
##&nbsp;Backup firmware version 2.1X6
##&nbsp;Backup ACP version 01.10
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;          {
##&nbsp;            'primary_acp_version' => '01.10',
##&nbsp;            'primary_firmware_version' => '2.2X1',
##&nbsp;            'booted_primary_acp_version' => '01.10',
##&nbsp;            'booted_primary_firmware_version' => '2.2X1',
##&nbsp;            'backup_acp_version' => '01.10',
##&nbsp;            'backup_firmware_version' => '2.1X6'
##&nbsp;          }
##&nbsp;        ];
##&nbsp;
sub bmc_version {
    my %args = @_;
    my $error_message;
    my $resp    = $args{response_object};
    my $output  = $resp->get_processed_output();
    my $command = $resp->get_actual_command();
    my @result_array;
    my $error = _parse_for_errors(
        "raw_output"    => $output,
        "command"       => $command,
        "error_message" => \$error_message
    );

    if ($error) {
        $error = 'NACL::APISet::Exceptions::' . $error . 'Exception';
        $error->throw( text => $error_message, response_object => $resp );
    }

    #Now we can parse the output
    my @arraylines =
      get_lines( raw_output => $output, response_object => $resp );
    foreach my $lines (@arraylines) {
        if ( $lines =~ /(.*)\s+(\d+.\S+)/ ) {
            $result_array[0]->{ to_key_name( item => $1 ) } = $2;
        }
    }
    return \@result_array;
}#end of bmc_version

## @name :: priv_set
##
## @summary Parses the raw output from the "priv set" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from priv set with any of its options as described in the
##    corresponding cdef file 'bmc.cdefs'
## @synopsis
##      priv_set(%args);
## INPUTS TO METHOD CALL 'priv set' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## <tr> <td> '-q' </td> <td> 'quiet' </td> <td> '1' </td> </tr>
## <tr> <td> 'diag' </td> <td> 'diag' </td> <td> '1' </td> </tr>
## <tr> <td> 'admin' </td> <td> 'admin' </td> <td> '1' </td> </tr>
## <tr> <td> 'advanced' </td> <td> 'advanced' </td> <td> '1' </td> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples
##&nbsp;Command = priv set admin; result -l
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;priv set admin
##&nbsp;Administrative commands
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;         {
##&nbsp;           'privilege' => 'Administrative commands'
##&nbsp;         }
##&nbsp;        ];
##&nbsp;
sub priv_set {
    my %args = @_;
    my $error_message;
    my $resp    = $args{response_object};
    my $output  = $resp->get_processed_output();
    my $command = $resp->get_actual_command();
    my @result_array;
    my $error = _parse_for_errors(
        "raw_output"    => $output,
        "command"       => $command,
        "error_message" => \$error_message
    );

    if ($error) {
        $error = 'NACL::APISet::Exceptions::' . $error . 'Exception';
        $error->throw( text => $error_message, response_object => $resp );
    }

    #Now we can parse the output
    my @arraylines =
      get_lines( raw_output => $output, response_object => $resp );
    foreach my $lines (@arraylines) {
        if ( $lines =~ /(.*)/ ) {
            $result_array[0]->{"privilege"} = $1;
        }
    }
    return \@result_array;
}#end of priv_set

## @name ::priv_show
##
## @summary Parses the raw output from the "priv show" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from priv_show with any of its options as described in the
##    corresponding cdef file 'bmc.cdefs'
## @synopsis
##      priv_show(%args);
## INPUTS TO METHOD CALL 'priv_show' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples
##&nbsp;Command = priv show
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;admin
##&nbsp;
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;          {
##&nbsp;            'privilege' => 'admin'
##&nbsp;          }
##&nbsp;        ];
##&nbsp;
##&nbsp;Command = priv show
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;admin
##&nbsp;
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;          {
##&nbsp;            'privilege' => 'admin'
##&nbsp;          }
##&nbsp;        ];
##&nbsp;
sub priv_show {
    my %args = @_;
    my $error_message;
    my $resp    = $args{response_object};
    my $output  = $resp->get_processed_output();
    my $command = $resp->get_actual_command();
    my @result_array;
    my $error = _parse_for_errors(
        "raw_output"    => $output,
        "command"       => $command,
        "error_message" => \$error_message
    );

    if ($error) {
        $error = 'NACL::APISet::Exceptions::' . $error . 'Exception';
        $error->throw( text => $error_message, response_object => $resp );
    }

    #Now we can parse the output
    my @arraylines =
      get_lines( raw_output => $output, response_object => $resp );
    foreach my $lines (@arraylines) {
        if ( $lines =~ /(.*)/ ) {
            $result_array[0]->{"privilege"} = $1;
        }
    }
    return \@result_array;
}#end of priv_show

## @name :: bmc_reboot
##
## @summary Parses the raw output from the "bmc reboot" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from bmc reboot with any of its options as described in the
##    corresponding cdef file 'bmc.cdefs'
## @synopsis
##      bmc_reboot(%args);
## INPUTS TO METHOD CALL 'bmc reboot' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## <tr> <td> '-f' </td> <td> 'force' </td> <td> '1' </td> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples
##&nbsp;Command = bmc reboot; result -l
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;bmc reboot
##&nbsp;The BMC will be rebooted.  Continue? [y/n]
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [];
sub bmc_reboot {
    my %args        = @_;
    my $resp        = $args{response_object};
    my $output      = $resp->get_processed_output();
    my $error_level = $resp->get_execution_status();
    my $error;
    if ($error_level) {
        $error = _parse_for_errors(
            raw_output      => $output,
            response_object => $resp,
        );
    }
    return [];
}#end of bmc_reboot

## @name :: system_power
##
## @summary Parses the raw output from the "system power" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from system power with any of its options as described in the
##    corresponding cdef file 'bmc.cdefs'
## @synopsis
##      system_power(%args);
## INPUTS TO METHOD CALL 'system power' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## <tr> <td> '-f' </td> <td> 'force' </td> <td> '1' </td> </tr>
## <tr> <td> 'status' </td> <td> 'status' </td> <td> '1' </td> </tr>
## <tr> <td> 'cycle' </td> <td> 'cycle' </td> <td> '1' </td> </tr>
## <tr> <td> 'off' </td> <td> 'off' </td> <td> '1' </td> </tr>
## <tr> <td> 'on' </td> <td> 'on' </td> <td> '1' </td> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples
##&nbsp;Command = system power status; result -l
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;system power status
##&nbsp;Chassis Power is on
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp; $VAR1 = [
##&nbsp;           {
##&nbsp;             'chassis_power' => 'on'
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp; $VAR1 = [
##&nbsp;           {
##&nbsp;             'chassis_power' => 'on'
##&nbsp;           }
##&nbsp;         ];
sub system_power {
    my %args = @_;
    my $error_message;
    my $resp    = $args{response_object};
    my $output  = $resp->get_processed_output();
    my $command = $resp->get_actual_command();
    my @result_array;
    my $error = _parse_for_errors(
        "raw_output"    => $output,
        "command"       => $command,
        "error_message" => \$error_message
    );

    if ($error) {
        $error = 'NACL::APISet::Exceptions::' . $error . 'Exception';
        $error->throw( text => $error_message, response_object => $resp );
    }

    #Now we can parse the output
    my @arraylines =
      get_lines( raw_output => $output, response_object => $resp );
    foreach my $lines (@arraylines) {
        if ( $lines =~ /^\S+\s+\S+\s+\S+\s+(\S+)$/ ) {
            $result_array[0]->{"chassis_power"} = $1;
        }
    }
        return \@result_array;

}#end of system_power

## @name :: system_reset
##
## @summary Parses the raw output from the "system reset" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from system reset with any of its options as described in the
##    corresponding cdef file 'bmc.cdefs'
## @synopsis
##      system_reset(%args);
## INPUTS TO METHOD CALL 'system reset' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## <tr> <td> '-f' </td> <td> 'force' </td> <td> '1' </td> </tr>
## <tr> <td> 'backup' </td> <td> 'backup' </td> <td> '0' </td> </tr>
## <tr> <td> 'current' </td> <td> 'current' </td> <td> '0' </td> </tr>
## <tr> <td> 'primary' </td> <td> 'primary' </td> <td> '0' </td> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples
##&nbsp;Command = system reset; result -l
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;system reset
##&nbsp;This will cause a dirty shutdown of your appliance.  Continue? [y/n]
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [];
sub system_reset {
    my %args        = @_;
    my $resp        = $args{response_object};
    my $output      = $resp->get_processed_output();
    my $error_level = $resp->get_execution_status();
    my $error;
    if ($error_level) {
        $error = _parse_for_errors(
            raw_output      => $output,
            response_object => $resp,
        );
    }
    return [];
}#end of system_reset

## @name :: system_console
##
## @summary Parses the raw output from the "system console" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from system console  with any of its options as described in the
##    corresponding cdef file 'bmc.cdefs'
## @synopsis
##      system_console(%args);
## INPUTS TO METHOD CALL 'system console' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples
##&nbsp;Command = system console; result -l
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;system console
##&nbsp;Type Ctrl-D to exit.
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [];
sub system_console {
    my %args        = @_;
    my $resp        = $args{response_object};
    my $output      = $resp->get_processed_output();
    my $error_level = $resp->get_execution_status();
    my $error;
    if ($error_level) {
        $error = _parse_for_errors(
            raw_output      => $output,
            response_object => $resp,
        );
    }
    return [];
}#end of system_console

## @name ::bmc_events
##
## @summary Parses the raw output from the "events" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from bmc_events with any of its options as described in the
##    corresponding cdef file 'bmc.cdefs'
## @synopsis
##      bmc_events(%args);
## INPUTS TO METHOD CALL 'bmc_events' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## <tr> <td> 'info' </td> <td> 'info' </td> <td> '1' </td> </tr>
## <tr> <td> 'search' </td> <td> 'search' </td> <td> '0' </td> </tr>
## <tr> <td> 'oldest' </td> <td> 'oldest' </td> <td> '0' </td> </tr>
## <tr> <td> 'newest' </td> <td> 'newest' </td> <td> '0' </td> </tr>
## <tr> <td> 'all' </td> <td> 'all' </td> <td> '1' </td> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

##&nbsp;Command = events  newest 5
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;Record 542: Fri Apr 12 11:58:06 2013 [BMC CLI.notice]: admin "log in from 10.72.8.245"
##&nbsp;Record 543: Fri Apr 12 11:58:43 2013 [BMC CLI.notice]: admin "log in from 10.72.8.245"
##&nbsp;Record 544: Fri Apr 12 12:00:23 2013 [BMC CLI.notice]: admin "log in from 10.72.8.245"
##&nbsp;Record 545: Fri Apr 12 12:07:31 2013 [BMC CLI.notice]: admin "log in from 10.72.8.245"
##&nbsp;Record 546: Fri Apr 12 12:08:22 2013 [BMC CLI.notice]: admin "log in from 10.72.8.245"
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;          {
##&nbsp;            'record' => '542',
##&nbsp;            'time' => '11:58:06',
##&nbsp;            'date' => 'Fri Apr 12',
##&nbsp;            'message' => '[BMC CLI.notice]: admin "log in from 10.72.8.245"',
##&nbsp;            'year' => '2013'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'record' => '543',
##&nbsp;            'time' => '11:58:43',
##&nbsp;            'date' => 'Fri Apr 12',
##&nbsp;            'message' => '[BMC CLI.notice]: admin "log in from 10.72.8.245"',
##&nbsp;            'year' => '2013'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'record' => '544',
##&nbsp;            'time' => '12:00:23',
##&nbsp;            'date' => 'Fri Apr 12',
##&nbsp;            'message' => '[BMC CLI.notice]: admin "log in from 10.72.8.245"',
##&nbsp;            'year' => '2013'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'record' => '545',
##&nbsp;            'time' => '12:07:31',
##&nbsp;            'date' => 'Fri Apr 12',
##&nbsp;            'message' => '[BMC CLI.notice]: admin "log in from 10.72.8.245"',
##&nbsp;            'year' => '2013'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'record' => '546',
##&nbsp;            'time' => '12:08:22',
##&nbsp;            'date' => 'Fri Apr 12',
##&nbsp;            'message' => '[BMC CLI.notice]: admin "log in from 10.72.8.245"',
##&nbsp;            'year' => '2013'
##&nbsp;          }
##&nbsp;        ];
##&nbsp;
sub bmc_events {
    my %args = @_;
    my $error_message;
    my $resp    = $args{response_object};
    my $output  = $resp->get_processed_output();
    my $command = $resp->get_actual_command();
    my @result_array;
	my $error;
     if ( $output =~ /^usage.+/si ) {
        $error          = "InvalidParam";
        $$error_message = $output;
    }

    if ($error) {
        $error = 'NACL::APISet::Exceptions::' . $error . 'Exception';
        $error->throw( text => $error_message, response_object => $resp );
    }

    #Now we can parse the output
    my @arraylines =
      get_lines( raw_output => $output, response_object => $resp );
    my $i = 0;

    foreach my $lines (@arraylines) {
        if ( $lines =~
            /\S+\s+(\d+):\s+(\S+\s+\S+\s+\d+)\s+(\S+)\s+(\d+)\s+(.*)/ )
        {
            $result_array[$i]->{"record"}  = $1;
            $result_array[$i]->{"date"}    = $2;
            $result_array[$i]->{"time"}    = $3;
            $result_array[$i]->{"year"}    = $4;
            $result_array[$i]->{"message"} = $5;
        }
        elsif ( $lines =~ /(.*):\s+(\d+)/ ) {
            $result_array[$i]->{"$1"} = $2;
        }
        $i++;
    }
    return \@result_array;
}#end of bmc_events

## @name :: bmc_update
##
## @summary Parses the raw output from the "bmc update" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from bmc update with any of its options as described in the
##    corresponding cdef file 'bmc.cdefs'
## @synopsis
##      bmc_update(%args);
## INPUTS TO METHOD CALL 'bmc update' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## <tr> <td> No switch for this key </td> <td> 'url' </td> <td> '0' </td> </tr>
## <tr> <td> '-f' </td> <td> 'force' </td> <td> '1' </td> </tr>
## <tr> <td> '-p' </td> <td> 'update-primary' </td> <td> '1' </td> </tr>
## <tr> <td> '-b' </td> <td> 'update-backup' </td> <td> '1' </td> </tr>
## <tr> <td> '-c' </td> <td> 'clear' </td> <td> '1' </td> </tr>
## <tr> <td> '-i' </td> <td> 'ignore' </td> <td> '1' </td> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples
sub bmc_update {
    my %args = @_;
    my $error_message;
    my $resp    = $args{response_object};
    my $output  = $resp->get_processed_output();
    my $command = $resp->get_actual_command();
    my $key;
    my @result_array;
    my $error = _parse_for_errors(
        "raw_output"    => $output,
        "command"       => $command,
        "error_message" => \$error_message
    );

    if ($error) {
        $error = 'NACL::APISet::Exceptions::' . $error . 'Exception';
        $error->throw( text => $error_message, response_object => $resp );
    }

    my @arraylines = get_lines( raw_output => $output, response_object => $resp );
    foreach my $lines (@arraylines) {
        if ( $lines =~ /BMC FW Update (Successful)/ ) {
            $result_array[0]->{"status"}  = $1;
        }
    }
    return \@result_array;

} #end of bmc_update

## @name :: bmc_shell
##
## @summary Parses the raw output from the "bmc shell" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from bmc shell with any of its options as described in the
##    corresponding cdef file 'bmc.cdefs'
## @synopsis
##      bmc_shell(%args);
## INPUTS TO METHOD CALL 'bmc shell' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples
sub bmc_shell {
    my %args        = @_;
    my $resp        = $args{response_object};
    my $output      = $resp->get_processed_output();
    my $error_level = $resp->get_execution_status();
    my $error;
    if ($error_level) {
        $error = _parse_for_errors(
            raw_output      => $output,
            response_object => $resp,
        );
    }
    return [];
} ##end of bmc_shell

## @name ::bmc_uptime
##
## @summary Parses the raw output from the "bmc uptime" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from bmc uptime with any of its options as described in the
##    corresponding cdef file 'bmc.cdefs'
## @synopsis
##      bmc_uptime(%args);
## INPUTS TO METHOD CALL 'bmc_uptime' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples
##&nbsp;Command = bmc uptime; result -l
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;priv set diag
##&nbsp;Warning: These diagnostic commands are for use by support personnel only.
##&nbsp;
##&nbsp;bmc uptime
##&nbsp;07:17:39 up 12 days, 8:16, load average: 1.04, 1.03, 1.00
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;           {
##&nbsp;             'load_average' => '1.04, 1.03, 1.00',
##&nbsp;             'uptime' => '07:17:39 up 12 days',
##&nbsp;             'current_time' => '8:16'
##&nbsp;           }
##&nbsp;         ];
sub bmc_uptime {
    my %args = @_;
    my $error_message;
    my $resp    = $args{response_object};
    my $output  = $resp->get_processed_output();
    my $command = $resp->get_actual_command();
    my @result_array;
    my $i=0;
    my $error = _parse_for_errors(
        "raw_output"    => $output,
        "command"       => $command,
        "error_message" => \$error_message
    );
    if ($error) {
        $error = 'NACL::APISet::Exceptions::' . $error . 'Exception';
        $error->throw( text => $error_message, response_object => $resp );
    }

    #Now we can parse the output
    my @arraylines =
      get_lines( raw_output => $output, response_object => $resp );
    foreach my $lines (@arraylines) {
        if ( $lines =~ /(\S+\s+\S+\s+\d+\s+\S+),\s+(\S+),\s+\S+\s+\S+\s+(\S+\s+\S+\s+\S+)/ ) {
            $result_array[$i]->{uptime} = $1;
            $result_array[$i]->{current_time} = $2;
            $result_array[$i]->{load_average} = $3;
        }
    }
        return \@result_array;
}    # end of bmc_uptime

## @name :bmc_fw_test
##
## @summary Parses the raw output from the "bmc fw-test" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from bmc fw-test with any of its options as described in the
##    corresponding cdef file 'bmc.cdefs'
## @synopsis
##      bmc_fw_test(%args);
## INPUTS TO METHOD CALL 'bmc_fw_test' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples
##&nbsp;Command = bmc fw-test; result -l
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;priv set diag
##&nbsp;Warning: These diagnostic commands are for use by support personnel only.
##&nbsp;
##&nbsp;bmc fw-test
##&nbsp;BMC FW Test Passed
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;           {
##&nbsp;             'bmc_fw_test' => 'Passed'
##&nbsp;           }
##&nbsp;         ];
sub bmc_fw_test {
    my %args = @_;
    my $error_message;
    my $resp    = $args{response_object};
    my $output  = $resp->get_processed_output();
    my $command = $resp->get_actual_command();
    my @result_array;
	 my $i=0;
    my $error = _parse_for_errors(
        "raw_output"    => $output,
        "command"       => $command,
        "error_message" => \$error_message
    );

    if ($error) {
        $error = 'NACL::APISet::Exceptions::' . $error . 'Exception';
        $error->throw( text => $error_message, response_object => $resp );
    }

    #Now we can parse the output
    my @arraylines =
      get_lines( raw_output => $output, response_object => $resp );
    foreach my $lines (@arraylines) {
        if ( $lines =~ /\S+\s+\S+\s+\S+\s+(\S+)/ ) {
            $result_array[$i]->{bmc_fw_test} = $1;
        }
        return \@result_array;
    }
}    # end of bmc_fw_test

## @name ::bmc_switch_dump
##
## @summary Parses the raw output from the "bmc switch dump" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from bmc switch dump with any of its options as described in the
##    corresponding cdef file 'bmc.cdefs'
## @synopsis
##      bmc_switch_dump(%args);
## INPUTS TO METHOD CALL 'bmc_switch_dump' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples
##&nbsp;Command = bmc switch dump; result -l
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;PHY0    PHY1    PHY2    PHY3    PHY4    POR0    POR1    POR2    POR3    POR4    POR5    POR6    GLO1    GLO2
##&nbsp;0x0000 :  0x1140  0x1140  0x1940  0x1140  0x1140  0x500f  0xde0f  0x500f  0xde0f  0x500f  0x4d05  0x4d05  0xc800  0x0000
##&nbsp;0x0001 :  0x7949  0x796d  0x7949  0x796d  0x7949  0x0003  0x0003  0x0003  0x0003  0x0003  0x0003  0x0003  0x0000  0x0003
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp; $VAR1 = [
##&nbsp;           {
##&nbsp;             '0x0000' => {
##&nbsp;                           'POR2' => '0x500f',
##&nbsp;                           'POR4' => '0x500f',
##&nbsp;                           'PHY1' => '0x1140',
##&nbsp;                           'PHY2' => '0x1940',
##&nbsp;                           'PHY0' => '0x1140',
##&nbsp;                           'PHY3' => '0x1140',
##&nbsp;                           'GLO1' => '0xc800',
##&nbsp;                           'PHY4' => '0x1140',
##&nbsp;                           'GLO2' => '0x0000',
##&nbsp;                           'POR6' => '0x4d05',
##&nbsp;                           'POR3' => '0xde0f',
##&nbsp;                           'POR0' => '0x500f',
##&nbsp;                           'POR1' => '0xde0f',
##&nbsp;                           'POR5' => '0x4d05'
##&nbsp;                         }
##&nbsp;           },
##&nbsp;           {
##&nbsp;             '0x0001' => {
##&nbsp;                           'POR2' => '0x0003',
##&nbsp;                           'POR4' => '0x0003',
##&nbsp;                           'PHY1' => '0x796d',
##&nbsp;                           'PHY2' => '0x7949',
##&nbsp;                           'PHY0' => '0x7949',
##&nbsp;                           'PHY3' => '0x796d',
##&nbsp;                           'GLO1' => '0x0000',
##&nbsp;                           'PHY4' => '0x7949',
##&nbsp;                           'GLO2' => '0x0003',
##&nbsp;                           'POR6' => '0x0003',
##&nbsp;                           'POR3' => '0x0003',
##&nbsp;                           'POR0' => '0x0003',
##&nbsp;                           'POR1' => '0x0003',
##&nbsp;                           'POR5' => '0x0003'
##&nbsp;                         }
##&nbsp;           },
##&nbsp;          ];

sub bmc_switch_dump {
    my %args = @_;
    my $error_message;
    my $resp    = $args{response_object};
    my $output  = $resp->get_processed_output();
    my $command = $resp->get_actual_command();
    my @result_array;
    my $error = _parse_for_errors(
        "raw_output"    => $output,
        "command"       => $command,
        "error_message" => \$error_message
    );

    if ($error) {
        $error = 'NACL::APISet::Exceptions::' . $error . 'Exception';
        $error->throw( text => $error_message, response_object => $resp );
    }
    my @lines = get_lines( raw_output => $output, response_object => $resp );
    my $count = 0;
    my @head = split( /\s+/, $lines[0] );
    foreach my $line (@lines) {
        if($line =~ /(\S+)\s+[:]\s+(.*)/){
            my $head1 = $1 ;
            my $temp = $2 ;
            my @values =  split( /\s+/, $temp );
            for ( my $i = 0 ; $i <= $#head ; $i++ ) {
               $result_array[$count]->{$head1}->{ $head[$i] } = $values[$i];
            }
         $count++;
        }
    }
    return \@result_array;
}    # end of bmc_switch_dump

## @name ::bmc_switch_stat
##
## @summary Parses the raw output from the "bmc switch stat" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from bmc switch stat with any of its options as described in the
##    corresponding cdef file 'bmc.cdefs'
## @synopsis
##      bmc_switch_stat(%args);
## INPUTS TO METHOD CALL 'bmc_switch_stat' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## <tr> <td> No switch for this key </td> <td> 'channel' </td> <td> '0' </td> </tr>
## <tr> <td> No switch for this key </td> <td> 'addr-range' </td> <td> '0' </td> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples
##&nbsp;Command = bmc switch stat; result -l
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;priv set diag
##&nbsp;Warning: These diagnostic commands are for use by support personnel only.
##&nbsp;
##&nbsp;bmc switch stat 0 1
##&nbsp;Switch Port 0 SMI Dev Addr 18
##&nbsp;Link :  Down
##&nbsp;Rx Count : 0
##&nbsp;Tx Count : 0
##&nbsp;Switch Port 1 SMI Dev Addr 19
##&nbsp;Link :  Up
##&nbsp;Rx Count : 1bc7
##&nbsp;Tx Count : 78b
##&nbsp;Switch Port 2 SMI Dev Addr 1a
##&nbsp;Link :  Up
##&nbsp;Rx Count : 10
##&nbsp;Tx Count : 8
##&nbsp;Switch Port 3 SMI Dev Addr 1b
##&nbsp;Link :  Up
##&nbsp;Rx Count : 14
##&nbsp;Tx Count : 1bde
##&nbsp;Switch Port 4 SMI Dev Addr 1c
##&nbsp;Link :  Up
##&nbsp;Rx Count : 0
##&nbsp;Tx Count : 18
##&nbsp;Switch Port 5 SMI Dev Addr 1d
##&nbsp;Link :  Up
##&nbsp;Rx Count : 788
##&nbsp;Tx Count : 1be2
##&nbsp;Switch Port 6 SMI Dev Addr 1e
##&nbsp;Link :  Up
##&nbsp;Rx Count : 8
##&nbsp;Tx Count : 10
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp; $VAR1 = [
##&nbsp;           {
##&nbsp;             'tx_count' => '0',
##&nbsp;             'link' => 'Down',
##&nbsp;             'switch_port' => '0',
##&nbsp;             'smi_device_address' => '18',
##&nbsp;             'rx_count' => '0'
##&nbsp;           },
##&nbsp;           {
##&nbsp;             'tx_count' => '78b',
##&nbsp;             'link' => 'Up',
##&nbsp;             'switch_port' => '1',
##&nbsp;             'smi_device_address' => '19',
##&nbsp;             'rx_count' => '1bc7'
##&nbsp;           },
##&nbsp;           {
##&nbsp;             'tx_count' => '8',
##&nbsp;             'link' => 'Up',
##&nbsp;             'switch_port' => '2',
##&nbsp;             'smi_device_address' => '1a',
##&nbsp;             'rx_count' => '10'
##&nbsp;           },
##&nbsp;           {
##&nbsp;             'tx_count' => '1bde',
##&nbsp;             'link' => 'Up',
##&nbsp;             'switch_port' => '3',
##&nbsp;             'smi_device_address' => '1b',
##&nbsp;             'rx_count' => '14'
##&nbsp;           },
##&nbsp;           {
##&nbsp;             'tx_count' => '18',
##&nbsp;             'link' => 'Up',
##&nbsp;             'switch_port' => '4',
##&nbsp;             'smi_device_address' => '1c',
##&nbsp;             'rx_count' => '0'
##&nbsp;           },
##&nbsp;           {
##&nbsp;             'tx_count' => '1be2',
##&nbsp;             'link' => 'Up',
##&nbsp;             'switch_port' => '5',
##&nbsp;             'smi_device_address' => '1d',
##&nbsp;             'rx_count' => '788'
##&nbsp;           },
##&nbsp;           {
##&nbsp;             'tx_count' => '10',
##&nbsp;             'link' => 'Up',
##&nbsp;             'switch_port' => '6',
##&nbsp;             'smi_device_address' => '1e',
##&nbsp;             'rx_count' => '8'
##&nbsp;           }
##&nbsp;         ];
sub bmc_switch_stat {
    my %args = @_;
    my $error_message;
    my $resp    = $args{response_object};
    my $output  = $resp->get_processed_output();
    my $command = $resp->get_actual_command();
    my @result_array;
    my $i=-1;
    my $error = _parse_for_errors(
        "raw_output"    => $output,
        "command"       => $command,
        "error_message" => \$error_message
    );

    if ($error) {
        $error = 'NACL::APISet::Exceptions::' . $error . 'Exception';
        $error->throw( text => $error_message, response_object => $resp );
    }

    #Now we can parse the output
    my @arraylines =
      get_lines( raw_output => $output, response_object => $resp );
    foreach my $lines (@arraylines) {
        if ( $lines =~ /\S+\s+\S+\s+(\S+)\s+\S+\s+\S+\s+\S+\s+(\S+)/ ) {
            $i++;
            $result_array[$i]->{switch_port} = $1;
            $result_array[$i]->{smi_device_address} = $2;
        } elsif ( $lines =~ /^\s*(\w*|\w*\s*\w*)\s*:\s*(\S+)/ ) {
            $result_array[$i]{to_key_name(item =>$1)} = $2;
        }
    }
        return \@result_array;
}    # end of bmc_switch_stat

## @name ::bmc_switch_read
##
## @summary Parses the raw output from the "bmc switch read" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from bmc switch read with any of its options as described in the
##    corresponding cdef file 'bmc.cdefs'
## @synopsis
##      bmc_switch_read(%args);
## INPUTS TO METHOD CALL 'bmc_switch_read' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## <tr> <td> No switch for this key </td> <td> 'channel' </td> <td> '0' </td> </tr>
## <tr> <td> No switch for this key </td> <td> 'smi-device-address' </td> <td> '0' </td> </tr>
## <tr> <td> No switch for this key </td> <td> 'smi-register-address' </td> <td> '0' </td> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples
##&nbsp;Command = bmc switch read; result -l
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;priv set diag
##&nbsp;Warning: These diagnostic commands are for use by support personnel only.
##&nbsp;
##&nbsp;bmc switch read 0 0-0x1F 0-0x1F
##&nbsp;ffffffff
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp; $VAR1 = [
##&nbsp;           {
##&nbsp;             'read_value' => 'ffffffff'
##&nbsp;           }
##&nbsp;         ];
sub bmc_switch_read {
    my %args = @_;
    my $error_message;
    my $resp    = $args{response_object};
    my $output  = $resp->get_processed_output();
	 my $command = $resp->get_actual_command();
    my @result_array;
    my $error = _parse_for_errors(
        "raw_output"    => $output,
        "command"       => $command,
        "error_message" => \$error_message
    );

    if ($error) {
        $error = 'NACL::APISet::Exceptions::' . $error . 'Exception';
        $error->throw( text => $error_message, response_object => $resp );
    }

    #Now we can parse the output
    my @arraylines =
      get_lines( raw_output => $output, response_object => $resp );
    foreach my $lines (@arraylines) {
        if ( $lines =~ /(\S+)/ ) {
            $result_array[0]->{read_value} = $1;
        }
    }
        return \@result_array;
}    # end of bmc_switch_read

## @name ::bmc_switch_write
##
## @summary Parses the raw output from the "bmc switch write" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from bmc switch write with any of its options as described in the
##    corresponding cdef file 'bmc.cdefs'
## @synopsis
##      bmc_switch_write(%args);
## INPUTS TO METHOD CALL 'bmc_switch_write' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## <tr> <td> No switch for this key </td> <td> 'channel' </td> <td> '0' </td> </tr>
## <tr> <td> No switch for this key </td> <td> 'smi-device-address' </td> <td> '0' </td> </tr>
## <tr> <td> No switch for this key </td> <td> 'smi-register-address' </td> <td> '0' </td> </tr>
## <tr> <td> No switch for this key </td> <td> 'value' </td> <td> '0' </td> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples
##&nbsp;Command = bmc switch write; result -l
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;priv set diag
##&nbsp;Warning: These diagnostic commands are for use by support personnel only.
##&nbsp;
##&nbsp;bmc switch write_phy_88E6131 0 0-0x1F 0-0x1F write
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp; $VAR1 = [];
sub bmc_switch_write {
    my %args = @_;
    my $error_message;
    my $resp    = $args{response_object};
    my $output  = $resp->get_processed_output();
    my $command = $resp->get_actual_command();
    my @result_array;
    my $error = _parse_for_errors(
        "raw_output"    => $output,
        "command"       => $command,
        "error_message" => \$error_message
    );

    if ($error) {
        $error = 'NACL::APISet::Exceptions::' . $error . 'Exception';
        $error->throw( text => $error_message, response_object => $resp );
    }
        return [];
}    # end of bmc_switch_write

## @name ::bmc_switch_read_phy_88E6131
##
## @summary Parses the raw output from the "bmc switch read_phy_88E6131" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from bmc switch read_phy_88E6131 with any of its options as described in the
##    corresponding cdef file 'bmc.cdefs'
## @synopsis
##      bmc_switch_read_phy_88E6131(%args);
## INPUTS TO METHOD CALL 'bmc_switch_read_phy_88E6131' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## <tr> <td> No switch for this key </td> <td> 'channel' </td> <td> '0' </td> </tr>
## <tr> <td> No switch for this key </td> <td> 'phy-device-address' </td> <td> '0' </td> </tr>
## <tr> <td> No switch for this key </td> <td> 'smi-register-address' </td> <td> '0' </td> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples
##&nbsp;Command = bmc switch read_phy_88E6131; result -l
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;priv set diag
##&nbsp;Warning: These diagnostic commands are for use by support personnel only.
##&nbsp;
##&nbsp;bmc switch read_phy_88E6131 0 0-0x1F 0-0x1F
##&nbsp;0
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp; $VAR1 = [
##&nbsp;           {
##&nbsp;             'read_phy_88E6131_value' => '0'
##&nbsp;           }
##&nbsp;         ];
sub bmc_switch_read_phy_88E6131 {
    my %args = @_;
    my $error_message;
    my $resp    = $args{response_object};
    my $output  = $resp->get_processed_output();
	 my $command = $resp->get_actual_command();
    my @result_array;
    my $error = _parse_for_errors(
        "raw_output"    => $output,
        "command"       => $command,
        "error_message" => \$error_message
    );

    if ($error) {
        $error = 'NACL::APISet::Exceptions::' . $error . 'Exception';
        $error->throw( text => $error_message, response_object => $resp );
    }

    #Now we can parse the output
    my @arraylines =
      get_lines( raw_output => $output, response_object => $resp );
    foreach my $lines (@arraylines) {
        if ( $lines =~ /(\S+)/ ) {
            $result_array[0]->{read_phy_88E6131_value} = $1;
        }
    }
        return \@result_array;
}    # end of bmc_switch_read_phy_88E6131

## @name ::bmc_switch_write_phy_88E6131
##
## @summary Parses the raw output from the "bmc switch write_phy_88E6131" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from bmc switch write_phy_88E6131 with any of its options as described in the
##    corresponding cdef file 'bmc.cdefs'
## @synopsis
##      bmc_switch_write_phy_88E6131(%args);
## INPUTS TO METHOD CALL 'bmc_switch_write_phy_88E6131' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## <tr> <td> No switch for this key </td> <td> 'channel' </td> <td> '0' </td> </tr>
## <tr> <td> No switch for this key </td> <td> 'phy-device-address' </td> <td> '0' </td> </tr>
## <tr> <td> No switch for this key </td> <td> 'smi-register-address' </td> <td> '0' </td> </tr>
## <tr> <td> No switch for this key </td> <td> 'value' </td> <td> '0' </td> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples
##&nbsp;Command = bmc switch write_phy_88E6131; result -l
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;priv set diag
##&nbsp;Warning: These diagnostic commands are for use by support personnel only.
##&nbsp;
##&nbsp;sp switch write_phy_88E6131 0 0-0x1F 0-0x1F write
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp; $VAR1 = [];
sub bmc_switch_write_phy_88E6131 {
    my %args = @_;
    my $error_message;
    my $resp    = $args{response_object};
    my $output  = $resp->get_processed_output();
    my $command = $resp->get_actual_command();
    my @result_array;
    my $error = _parse_for_errors(
        "raw_output"    => $output,
        "command"       => $command,
        "error_message" => \$error_message
    );

    if ($error) {
        $error = 'NACL::APISet::Exceptions::' . $error . 'Exception';
        $error->throw( text => $error_message, response_object => $resp );
    }
        return [];
}    # end of bmc_switch_write_phy_88E6131

## @name ::bmc_switch_read_phyextnreg_88E6131
##
## @summary Parses the raw output from the "bmc switch read_phyextnreg_88E6131" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from bmc switch read_phyextnreg_88E6131 with any of its options as described in the
##    corresponding cdef file 'bmc.cdefs'
## @synopsis
##    bmc_switch_read_phyextnreg_88E6131(%args);
## INPUTS TO METHOD CALL 'bmc_switch_read_phyextnreg_88E6131' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## <tr> <td> No switch for this key </td> <td> 'channel' </td> <td> '0' </td> </tr>
## <tr> <td> No switch for this key </td> <td> 'phy-device-address' </td> <td> '0' </td> </tr>
## <tr> <td> No switch for this key </td> <td> 'smi-register-address' </td> <td> '0' </td> </tr>
## <tr> <td> No switch for this key </td> <td> 'page-number' </td> <td> '0' </td> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples
##&nbsp;Command = bmc switch read_phyextnreg_88E6131; result -l
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;priv set diag
##&nbsp;Warning: These diagnostic commands are for use by support personnel only.
##&nbsp;
##&nbsp;bmc switch read_phyextnreg_88E6131 0 0-0x1F 0-0x1F 1
##&nbsp;0
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp; $VAR1 = [
##&nbsp;           {
##&nbsp;             'read_phyextnreg_88E6131_value' => '0'
##&nbsp;           }
##&nbsp;         ];
sub bmc_switch_read_phyextnreg_88E6131 {
    my %args = @_;
    my $error_message;
    my $resp    = $args{response_object};
    my $output  = $resp->get_processed_output();
    my $command = $resp->get_actual_command();
    my @result_array;
    my $error = _parse_for_errors(
        "raw_output"    => $output,
        "command"       => $command,
        "error_message" => \$error_message
		 );

    if ($error) {
        $error = 'NACL::APISet::Exceptions::' . $error . 'Exception';
        $error->throw( text => $error_message, response_object => $resp );
    }

    #Now we can parse the output
    my @arraylines =
      get_lines( raw_output => $output, response_object => $resp );
    foreach my $lines (@arraylines) {
        if ( $lines =~ /(\S+)/ ) {
            $result_array[0]->{read_phyextnreg_88E6131_value} = $1;
        }
    }
        return \@result_array;
}    # end of bmc_switch_read_phyextnreg_88E6131

## @name ::bmc_switch_write_phyextnreg_88E6131
##
## @summary Parses the raw output from the "bmc switch write_phyextnreg_88E6131" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from bmc switch write_phyextnreg_88E6131 with any of its options as described in the
##    corresponding cdef file 'bmc.cdefs'
## @synopsis
##    bmc_switch_write_phyextnreg_88E6131(%args);
## INPUTS TO METHOD CALL 'bmc_switch_write_phyextnreg_88E6131' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## <tr> <td> No switch for this key </td> <td> 'channel' </td> <td> '0' </td> </tr>
## <tr> <td> No switch for this key </td> <td> 'phy-device-address' </td> <td> '0' </td> </tr>
## <tr> <td> No switch for this key </td> <td> 'smi-register-address' </td> <td> '0' </td> </tr>
## <tr> <td> No switch for this key </td> <td> 'page-number' </td> <td> '0' </td> </tr>
## <tr> <td> No switch for this key </td> <td> 'value' </td> <td> '0' </td> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples
##&nbsp;Command = bmc switch write_phyextnreg_88E6131; result -l
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;priv set diag
##&nbsp;Warning: These diagnostic commands are for use by support personnel only.
##&nbsp;
##&nbsp;bmc switch write_phyextnreg_88E6131 0 0-0x1F 0-0x1F 1 write
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp; $VAR1 = [];
sub bmc_switch_write_phyextnreg_88E6131 {
    my %args = @_;
    my $error_message;
    my $resp    = $args{response_object};
    my $output  = $resp->get_processed_output();
    my $command = $resp->get_actual_command();
    my @result_array;
    my $error = _parse_for_errors(
        "raw_output"    => $output,
        "command"       => $command,
        "error_message" => \$error_message
    );

    if ($error) {
        $error = 'NACL::APISet::Exceptions::' . $error . 'Exception';
        $error->throw( text => $error_message, response_object => $resp );
    }
        return [];
}    # end of bmc_switch_write_phyextnreg_88E6131

## @name ::bmc_switch_read_switchphy_BMC53101M
##
## @summary Parses the raw output from the "bmc switch read_switchphy_BMC53101M" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from bmc switch read_switchphy_BMC53101M with any of its options as described in the
##    corresponding cdef file 'bmc.cdefs'
## @synopsis
##      bmc_switch_read_switchphy_BMC53101M(%args);
## INPUTS TO METHOD CALL 'bmc_switch_read_switchphy_BMC53101M' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## <tr> <td> No switch for this key </td> <td> 'channel' </td> <td> '0' </td> </tr>
## <tr> <td> No switch for this key </td> <td> 'register-address' </td> <td> '0' </td> </tr>
## <tr> <td> No switch for this key </td> <td> 'page' </td> <td> '0' </td> </tr>
## <tr> <td> No switch for this key </td> <td> 'word-count' </td> <td> '0' </td> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples
##&nbsp;Command = bmc switch read_switchphy_BMC53101M; result -l
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;priv set diag
##&nbsp;Warning: These diagnostic commands are for use by support personnel only.
##&nbsp;
##&nbsp;bmc switch read_switchphy_BMC53101M 0 0 0 1
##&nbsp;43
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp; $VAR1 = [
##&nbsp;           {
##&nbsp;             'read_switchphy_BMC53101M_value' => '43'
##&nbsp;           }
##&nbsp;         ];
sub bmc_switch_read_switchphy_BMC53101M {
    my %args = @_;
    my $error_message;
    my $resp    = $args{response_object};
    my $output  = $resp->get_processed_output();
    my $command = $resp->get_actual_command();
    my @result_array;
    my $error = _parse_for_errors(
        "raw_output"    => $output,
        "command"       => $command,
        "error_message" => \$error_message
    );

    if ($error) {
        $error = 'NACL::APISet::Exceptions::' . $error . 'Exception';
        $error->throw( text => $error_message, response_object => $resp );
    }

    #Now we can parse the output
    my @arraylines =
      get_lines( raw_output => $output, response_object => $resp );
    foreach my $lines (@arraylines) {
        if ( $lines =~ /(\S+)/ ) {
            $result_array[0]->{read_switchphy_BMC53101M_value}  = $1;
        }
    }
        return \@result_array;
}    # end of bmc_switch_read_switchphy_BMC53101M

## @name ::bmc_switch_write_switchphy_BMC53101M
##
## @summary Parses the raw output from the "bmc switch write_switchphy_BMC53101M" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from bmc switch write_switchphy_BMC53101M with any of its options as described in the
##    corresponding cdef file 'bmc.cdefs'
## @synopsis
##      bmc_switch_write_switchphy_BMC53101M(%args);
## INPUTS TO METHOD CALL 'bmc_switch_write_switchphy_BMC53101M' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## <tr> <td> No switch for this key </td> <td> 'channel' </td> <td> '0' </td> </tr>
## <tr> <td> No switch for this key </td> <td> 'register-address' </td> <td> '0' </td> </tr>
## <tr> <td> No switch for this key </td> <td> 'page' </td> <td> '0' </td> </tr>
## <tr> <td> No switch for this key </td> <td> 'word-count' </td> <td> '0' </td> </tr>
## <tr> <td> No switch for this key </td> <td> 'wdata1' </td> <td> '0' </td> </tr>
## <tr> <td> No switch for this key </td> <td> 'wdata2' </td> <td> '0' </td> </tr>
## <tr> <td> No switch for this key </td> <td> 'wdata3' </td> <td> '0' </td> </tr>
## <tr> <td> No switch for this key </td> <td> 'wdata4' </td> <td> '0' </td> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples
##&nbsp;Command = bmc switch write_switchphy_BMC53101M; result -l
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;priv set diag
##&nbsp;Warning: These diagnostic commands are for use by support personnel only.
##&nbsp;
##&nbsp;bmc switch write_switchphy_BMC53101M 0 0 0 1 43
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp; $VAR1 = [];
sub bmc_switch_write_switchphy_BMC53101M {
    my %args = @_;
    my $error_message;
    my $resp    = $args{response_object};
    my $output  = $resp->get_processed_output();
    my $command = $resp->get_actual_command();
    my @result_array;
    my $error = _parse_for_errors(
        "raw_output"    => $output,
        "command"       => $command,
        "error_message" => \$error_message
    );

    if ($error) {
        $error = 'NACL::APISet::Exceptions::' . $error . 'Exception';
        $error->throw( text => $error_message, response_object => $resp );
    }
        return [];
}    # end of bmc_switch_write_switchphy_BMC53101M

## @name ::bmc_switch_version
##
## @summary Parses the raw output from the "bmc switch-version" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from bmc switch-version with any of its options as described in the
##    corresponding cdef file 'bmc.cdefs'
## @synopsis
##    bmc_switch_version(%args);
## INPUTS TO METHOD CALL 'bmc_switch_version' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## <tr> <td> No switch for this key </td> <td> 'controller' </td> <td> '0' </td> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples
##&nbsp;Command = bmc switch-version; result -l
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;priv set diag
##&nbsp;Warning: These diagnostic commands are for use by support personnel only.
##&nbsp;sp switch-version 0
##&nbsp;Version2 SB_XVI Switch0
#3&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp; $VAR1 = [
##&nbsp;           {
##&nbsp;             'switch_version' => 'Version2 SB_XVI Switch0'
##&nbsp;           }
##&nbsp;         ];
sub bmc_switch_version {
    my %args = @_;
    my $error_message;
    my $resp    = $args{response_object};
    my $output  = $resp->get_processed_output();
    my $command = $resp->get_actual_command();
    my @result_array;
    my $i = 0;
    my $error = _parse_for_errors(
        "raw_output"    => $output,
        "command"       => $command,
        "error_message" => \$error_message
    );

    if ($error) {
        $error = 'NACL::APISet::Exceptions::' . $error . 'Exception';
        $error->throw( text => $error_message, response_object => $resp );
    }

    #Now we can parse the output
    my @arraylines =
      get_lines( raw_output => $output, response_object => $resp );
    foreach my $lines (@arraylines) {
        if ( $lines =~ /(.*)/ ) {
            $result_array[$i]->{switch_version}  = $1;
            $i++;
        }
    }
        return \@result_array;
}    # end of bmc_switch_version

## @name ::bmc_switch_update
##
## @summary Parses the raw output from the "bmc switch-update" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from bmc switch-update with any of its options as described in the
##    corresponding cdef file 'bmc.cdefs'
## @synopsis
##    bmc_switch_update(%args);
## INPUTS TO METHOD CALL 'bmc_switch_update' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## <tr> <td> No switch for this key </td> <td> 'controller' </td> <td> '0' </td> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples
##&nbsp;Command = bmc switch-update; result -l
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;priv set diag
##&nbsp;Warning: These diagnostic commands are for use by support personnel only.
##&nbsp;
##&nbsp;bmc switch-update 0
##&nbsp;This will disconnect your SSH session.  Continue? [y/n]
##&nbsp;n
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp; $VAR1 = [];
sub bmc_switch_update {
    my %args = @_;
    my $error_message;
    my $resp    = $args{response_object};
    my $output  = $resp->get_processed_output();
    my $command = $resp->get_actual_command();
    my @result_array;
    my $error = _parse_for_errors(
        "raw_output"    => $output,
        "command"       => $command,
        "error_message" => \$error_message
    );

    if ($error) {
        $error = 'NACL::APISet::Exceptions::' . $error . 'Exception';
        $error->throw( text => $error_message, response_object => $resp );
    }
    return [];
}    # end of bmc_switch_update

## @name ::bmc_ddr_version
##
## @summary Parses the raw output from the "bmc ddr-version" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from bmc ddr-version with any of its options as described in the
##    corresponding cdef file 'bmc.cdefs'
## @synopsis
##    bmc_ddr_version(%args);
## INPUTS TO METHOD CALL 'bmc_ddr_version' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples
##&nbsp;Command = bmc ddr-version; result -l
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;priv set diag
##&nbsp;Warning: These diagnostic commands are for use by support personnel only.
##&nbsp;
##&nbsp;bmc ddr-version
##&nbsp;Version A2 SB_XVI BMC DDR2
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp; $VAR1 = [
##&nbsp;           {
##&nbsp;             'ddr_version' => 'Version A2 SB_XVI BMC DDR2'
##&nbsp;           }
##&nbsp;         ];
sub bmc_ddr_version {
    my %args = @_;
    my $error_message;
    my $resp    = $args{response_object};
    my $output  = $resp->get_processed_output();
    my $command = $resp->get_actual_command();
    my @result_array;
    my $error = _parse_for_errors(
        "raw_output"    => $output,
        "command"       => $command,
        "error_message" => \$error_message
    );

    if ($error) {
        $error = 'NACL::APISet::Exceptions::' . $error . 'Exception';
        $error->throw( text => $error_message, response_object => $resp );
    }

    #Now we can parse the output
    my @arraylines =
      get_lines( raw_output => $output, response_object => $resp );
    foreach my $lines (@arraylines) {
        if ( $lines =~ /(Version.*)/ ) {
            $result_array[0]->{ddr_version}  = $1;
        }
    }
        return \@result_array;
}    # end of bmc_ddr_version

## @name ::bmc_ddr_update
##
## @summary Parses the raw output from the "bmc ddr-update" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from bmc ddr-update with any of its options as described in the
##    corresponding cdef file 'bmc.cdefs'
## @synopsis
##    bmc_ddr_update(%args);
## INPUTS TO METHOD CALL 'bmc_ddr_update' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples
##&nbsp;Command = bmc ddr-update; result -l
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;priv set diag
##&nbsp;Warning: These diagnostic commands are for use by support personnel only.
##&nbsp;
##&nbsp;bmc ddr-update
##&nbsp;DDR PROM already programmed
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp; $VAR1 = [
##&nbsp;           {
##&nbsp;             'ddr_update' => 'DDR PROM already programmed'
##&nbsp;           }
##&nbsp;         ];
sub bmc_ddr_update {
    my %args = @_;
    my $error_message;
    my $resp    = $args{response_object};
    my $output  = $resp->get_processed_output();
    my $command = $resp->get_actual_command();
    my @result_array;
    my $error = _parse_for_errors(
        "raw_output"    => $output,
        "command"       => $command,
        "error_message" => \$error_message
    );

    if ($error) {
        $error = 'NACL::APISet::Exceptions::' . $error . 'Exception';
        $error->throw( text => $error_message, response_object => $resp );
    }
  #Now we can parse the output
    my @arraylines =
      get_lines( raw_output => $output, response_object => $resp );
    foreach my $lines (@arraylines) {
        if ( $lines =~ /(.*)/ ) {
            $result_array[0]->{ddr_update}  = $1;
        }
    }
        return \@result_array;
}    # end of bmc_ddr_update

## @name ::bmc_hw_test
##
## @summary Parses the raw output from the "bmc hw-test" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from bmc hw-test with any of its options as described in the
##    corresponding cdef file 'bmc.cdefs'
## @synopsis
##    bmc_hw_test(%args);
## INPUTS TO METHOD CALL 'bmc_hw_test' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples
##&nbsp;Command = bmc hw-test; result -l
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;priv set diag
##&nbsp;Warning: These diagnostic commands are for use by support personnel only.
##&nbsp;
##&nbsp;bmc hw-test
##&nbsp;BMC HW Test Passed
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp; $VAR1 = [
##&nbsp;           {
##&nbsp;             'bmc_hw_test' => 'Passed''
##&nbsp;           }
##&nbsp;         ];
sub bmc_hw_test {
    my %args = @_;
    my $error_message;
    my $resp    = $args{response_object};
    my $output  = $resp->get_processed_output();
    my $command = $resp->get_actual_command();
    my @result_array;
    my $error = _parse_for_errors(
        "raw_output"    => $output,
        "command"       => $command,
        "error_message" => \$error_message
    );

    if ($error) {
        $error = 'NACL::APISet::Exceptions::' . $error . 'Exception';
        $error->throw( text => $error_message, response_object => $resp );
    }

    #Now we can parse the output
    my @arraylines =
      get_lines( raw_output => $output, response_object => $resp );
    foreach my $lines (@arraylines) {
        if ( $lines =~ /(\S+)$/ ) {
            $result_array[0]->{bmc_hw_test}  = $1;
        }
    }
        return \@result_array;
}    # end of bmc_hw_test

## @name ::bmc_loopback_test
##
## @summary Parses the raw output from the "bmc loopback-test" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from bmc loopback-test with any of its options as described in the
##    corresponding cdef file 'bmc.cdefs'
## @synopsis
##    bmc_loopback_test(%args);
## INPUTS TO METHOD CALL 'bmc_loopback_test' ON APISET OBJECT
## <tr> <td> No switch for this key </td> <td> 'port-id' </td> <td> '0' </td> </tr>
## <tr> <td> No switch for this key </td> <td> 'mode' </td> <td> '0' </td> </tr>
## <tr> <td> No switch for this key </td> <td> 'type' </td> <td> '0' </td> </tr>
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples
##&nbsp;Command = bmc loopback-test; result -l
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;priv set diag
##&nbsp;Warning: These diagnostic commands are for use by support personnel only.
##&nbsp;
##&nbsp;bmc loopback-test 1 0 0
##&nbsp;BMC Loopback Test PASSED command 0x00 result: 0x00
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp; $VAR1 = [
##&nbsp;           {
##&nbsp;             'bmc_loopback_test' => 'PASSED',
##&nbsp;             'result' => '0x00',
##&nbsp;             'command' => '0x00'
##&nbsp;           }
##&nbsp;         ];
sub bmc_loopback_test {
    my %args = @_;
    my $error_message;
    my $resp    = $args{response_object};
    my $output  = $resp->get_processed_output();
    my $command = $resp->get_actual_command();
    my @result_array;
    my $error = _parse_for_errors(
        "raw_output"    => $output,
        "command"       => $command,
        "error_message" => \$error_message
    );

    if ($error) {
        $error = 'NACL::APISet::Exceptions::' . $error . 'Exception';
        $error->throw( text => $error_message, response_object => $resp );
    }

    #Now we can parse the output
    my @arraylines =
      get_lines( raw_output => $output, response_object => $resp );
    foreach my $lines (@arraylines) {
        if ( $lines =~ /\S+\s+\S+\s+\S+\s+(\S+)\s+\S+\s+(\S+)\s+\S+\s+(\S+)/ ) {
            $result_array[0]->{bmc_loopback_test}  = $1;
            $result_array[0]->{command}  = $2;
            $result_array[0]->{result}  = $3;
        }
    }
        return \@result_array;
}    # end of bmc_loopback_test

## @name ::bmc_extended_self_test
##
## @summary Parses the raw output from the "bmc extended-self-test" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from bmc extended-self-test with any of its options as described in the
##    corresponding cdef file 'bmc.cdefs'
## @synopsis
##    bmc_extended_self_test(%args);
## INPUTS TO METHOD CALL 'bmc_extended_self_test' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples
##&nbsp;Command = bmc extended-self-test; result -l
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;priv set diag
##&nbsp;Warning: These diagnostic commands are for use by support personnel only.
##&nbsp;
##&nbsp;bmc extended-self-test
##&nbsp;Extended Test
##&nbsp;Self Test Passed status 0x0
##&nbsp;Firmware Test Passed status 0x0
##&nbsp;Loopback COM Test Passed status 0x0
##&nbsp;Loopback Ethernet Test Passed status 0x0
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp; $VAR1 = [
##&nbsp;           {
##&nbsp;             'loopback_ethernet_test_passed_status' => '0x0',
##&nbsp;             'self_test_passed_status' => '0x0',
##&nbsp;             'firmware_test_passed_status' => '0x0',
##&nbsp;             'loopback_com_test_passed_status' => '0x0'
##&nbsp;           }
##&nbsp;         ];
sub bmc_extended_self_test {
    my %args = @_;
    my $error_message;
    my $resp    = $args{response_object};
    my $output  = $resp->get_processed_output();
    my $command = $resp->get_actual_command();
    my @result_array;
    my $error = _parse_for_errors(
        "raw_output"    => $output,
        "command"       => $command,
        "error_message" => \$error_message
    );

    if ($error) {
        $error = 'NACL::APISet::Exceptions::' . $error . 'Exception';
        $error->throw( text => $error_message, response_object => $resp );
    }

    #Now we can parse the output
    my @arraylines =
      get_lines( raw_output => $output, response_object => $resp );
    foreach my $lines (@arraylines) {
        if ( $lines =~ /(.*)\s+(\dx\S+)/ ) {
        $result_array[0]{to_key_name( item => $1)} = $2;
        }
    }
        return \@result_array;
}    # end of bmc_extended_self_test

## @name ::bmc_self_test
##
## @summary Parses the raw output from the "bmc self-test" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from bmc self-test with any of its options as described in the
##    corresponding cdef file 'bmc.cdefs'
## @synopsis
##    bmc_self_test(%args);
## INPUTS TO METHOD CALL 'bmc_self_test' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples
##&nbsp;Command = bmc self-test; result -l
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;priv set diag
##&nbsp;Warning: These diagnostic commands are for use by support personnel only.
##&nbsp;
##&nbsp;bmc self-test
##&nbsp;Self Test Passed status: 55 00
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp; $VAR1 = [
##&nbsp;           {
##&nbsp;             'self_test' => 'Passed',
##&nbsp;             'status' => '55 00'
##&nbsp;           }
##&nbsp;         ];
sub bmc_self_test {
    my %args = @_;
    my $error_message;
    my $resp    = $args{response_object};
    my $output  = $resp->get_processed_output();
    my $command = $resp->get_actual_command();
    my @result_array;
    my $error = _parse_for_errors(
        "raw_output"    => $output,
        "command"       => $command,
        "error_message" => \$error_message
    );

    if ($error) {
        $error = 'NACL::APISet::Exceptions::' . $error . 'Exception';
        $error->throw( text => $error_message, response_object => $resp );
    }

    #Now we can parse the output
    my @arraylines =
      get_lines( raw_output => $output, response_object => $resp );
    foreach my $lines (@arraylines) {
        if ( $lines =~ /\S+\s+\S+\s+(\S+)\s+\S+\s+(\S+\s+\S+)/ ) {
        $result_array[0]->{self_test} = $1;
        $result_array[0]->{status} = $2;
        }
    }
        return \@result_array;
}    # end of bmc_self_test

## @name ::bmc_log_history_show
##
## @summary Parses the raw output from the "bmc_log_history_show" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from bmc log history show with any of its options as described in the
##    corresponding cdef file 'bmc.cdefs'
## @synopsis
##      bmc_log_history_show(%args);
## INPUTS TO METHOD CALL 'bmc_log_history_show' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## <tr> <td> '-archive' </td> <td> 'archive' </td> <td> '0' </td> </tr>
## <tr> <td> '-dump' </td> <td> 'dump' </td> <td> '0' </td> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples
##&nbsp;Command = bmc log history show; result -l
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;priv set diag
##&nbsp;Warning: These diagnostic commands are for use by support personnel only.
##&nbsp;
##&nbsp;bmc log history show
##&nbsp;Archive                       Archive-Size    Archive-Creation-Time
##&nbsp;------------------------------------------------------------------------------
##&nbsp;bmclog_downfiler.1.tgz          23782      Thu Jan  1 00:03:12 2009
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;  $VAR1 = [
##&nbsp;           {
##&nbsp;             'archive' => 'bmclog_downfiler.1.tgz',
##&nbsp;             'archive_size' => '23782',
##&nbsp;             'archive_creation_time' => 'Thu Jan  1 00:03:12 2009'
##&nbsp;           }
##&nbsp;         ];
sub bmc_log_history_show {
    my %args = @_;
    my $error_message;
    my $resp    = $args{response_object};
    my $output  = $resp->get_processed_output();
    my $command = $resp->get_actual_command();
    my @result_array;
    my $i=0;
    my $error = _parse_for_errors(
        "raw_output"    => $output,
        "command"       => $command,
        "error_message" => \$error_message
    );

    if ($error) {
        $error = 'NACL::APISet::Exceptions::' . $error . 'Exception';
        $error->throw( text => $error_message, response_object => $resp );
    }

    #Now we can parse the output
    my @arraylines =
      get_lines( raw_output => $output, response_object => $resp );
    foreach my $lines (@arraylines) {
        if ( $lines =~ /(\S+)\s+(\S+)\s+(\S+)?\s+(\S+\s+\S+\s+\d+\s+\S+\s+\S+)/ ) {
            $result_array[$i]->{"archive"} = $1;
            $result_array[$i]->{"archive_size"} = $2;
            $result_array[$i]->{"sequence_num"} = $3 if (defined $3);
            $result_array[$i]->{"archive_creation_time"} = $4;
            $i++;
        }
    }
    return \@result_array;
}    # end of bmc_log_history_show

## @name ::bmc_switch_get_all_regs_tst
##
## @summary Parses the raw output from the "bmc switch get_all_regs_tst" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from bmc switch get_all_regs_tst with any of its options as described in the
##    corresponding cdef file 'bmc.cdefs'
## @synopsis
##      bmc_switch_get_all_regs_tst(%args);
## INPUTS TO METHOD CALL 'bmc_switch_get_all_regs_tst' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples
##&nbsp;Command = bmc switch get_all_regs_tst; result -l
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;DA = 0x0000 , PN = 0x0000, RA = 0x0000, data = 0x1140
##&nbsp;DA = 0x0001 , PN = 0x0000, RA = 0x0000, data = 0x1140
##&nbsp;DA = 0x0002 , PN = 0x0000, RA = 0x0000, data = 0x1940
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp; $VAR1 = [
##&nbsp;           {
##&nbsp;             'ra' => '0x0000',
##&nbsp;             'da' => '0x0000',
##&nbsp;             'data' => '0x1140',
##&nbsp;             'pn' => '0x0000'
##&nbsp;           },
##&nbsp;           {
##&nbsp;             'ra' => '0x0000',
##&nbsp;             'da' => '0x0001',
##&nbsp;             'data' => '0x1140',
##&nbsp;             'pn' => '0x0000'
##&nbsp;           },
##&nbsp;           {
##&nbsp;             'ra' => '0x0000',
##&nbsp;             'da' => '0x0002',
##&nbsp;             'data' => '0x1940',
##&nbsp;             'pn' => '0x0000'
##&nbsp;           },
##&nbsp;         ];

sub bmc_switch_get_all_regs_tst{
    my %args = @_;
    my $error_message;
    my $resp    = $args{response_object};
    my $output  = $resp->get_processed_output();
    my $command = $resp->get_actual_command();
    my @result_array;
    my $i=0;
    my $error = _parse_for_errors(
        "raw_output"    => $output,
        "command"       => $command,
        "error_message" => \$error_message
    );

    if ($error) {
        $error = 'NACL::APISet::Exceptions::' . $error . 'Exception';
        $error->throw( text => $error_message, response_object => $resp );
    }

    my @lines = get_lines( raw_output => $output, response_object => $resp );
    my $count = 0;

    foreach my $line (@lines){
        my @head = split( /\s*,\s*/, $line );

        for ( my $i = 0 ; $i <= $#head ; $i++ ) {
            my @values = split( /\s*=\s*/, $head[$i] );
            $result_array[$count]->{ to_key_name( item =>$values[0]) } = $values[1] if ($values[1]);
        }
        $count++;
    }
    return \@result_array;
} # end of bmc_switch_get_all_regs_tst

## @name ::bmc_devices_version
##
## @summary Parses the raw output from the "bmc devices-version" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from bmc devices-version with any of its options as described in the
##    corresponding cdef file 'bmc.cdefs'
## @synopsis
##      bmc_devices_version(%args);
## INPUTS TO METHOD CALL 'bmc devices-version' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples
##&nbsp;Command = bmc devices-version; result -l
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;smart_battery:
##&nbsp;Smart Battery version get is still being worked on.
##&nbsp;
##&nbsp;seq_cpld:
##&nbsp;Retrieving version of sequencing CPLD
##&nbsp;2.3
##&nbsp;
##&nbsp;fan:
##&nbsp;  FAN1: 802a
##&nbsp;  FAN2: 802a
##&nbsp;  FAN3: 802a
##&nbsp;
##&nbsp;battery:
##&nbsp;8020
##&nbsp;
##&nbsp;pcie_switch:
##&nbsp;  slot 0(PCM) id0:
##&nbsp;2.01
##&nbsp;  slot 0(PCM) id1:
##&nbsp;2.01
##&nbsp;  slot 6(NVS) id0:
##&nbsp;2.02
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;          {
##&nbsp;            'pcie_switch' => {
##&nbsp;                               'slot_0pcm_id0' => '2.01',
##&nbsp;                               'slot_6nvs_id0' => '2.02',
##&nbsp;                               'slot_0pcm_id1' => '2.01',
##&nbsp;                             },
##&nbsp;            'battery' => '8020',
##&nbsp;            'fan' => {
##&nbsp;                       'fan3' => '802a',
##&nbsp;                       'fan1' => '802a',
##&nbsp;                       'fan2' => '802a',
##&nbsp;                     },
##&nbsp;            'seq_cpld' => 'Retrieving version of sequencing CPLD 2.3',
##&nbsp;            'smart_battery' => 'Smart Battery version get is still being worked on.'
##&nbsp;          }
##&nbsp;        ];
##&nbsp;

sub bmc_devices_version {
    my %args = @_;
    my $error_message;
    my $resp    = $args{response_object};
    my $output  = $resp->get_processed_output();
    my $command = $resp->get_actual_command();
    my @result_array;
    my $error = _parse_for_errors(
        "raw_output"    => $output,
        "command"       => $command,
        "error_message" => \$error_message
    );

    if ($error) {
        $error = 'NACL::APISet::Exceptions::' . $error . 'Exception';
        $error->throw( text => $error_message, response_object => $resp );
    }
    my @lines = get_lines( raw_output => $output, response_object => $resp );
    my ($value,$field,$previous_line,$sub_field);
    foreach my $line (@lines) {
        if ($line =~ /:/) {
            my @values = split( /\s*:\s*/, $line ,2 );
            if ($values[0] =~ /smart*|seq*|support*|ddr*|fan|psu|battery|nvs*|^iocard$|^iocard_cpld$|pcie*|ucd*/i) {
                $field = $values[0];
		$sub_field = '';
		$value ='';
	    }
	    else {
                $sub_field = $values[0] if ($field) ;
	    }
            if ($values[1] && !$sub_field) {
                 $result_array[0]->{ to_key_name( item =>$field) } = $values[1];
	         $field = '';
            }
	    else {
	         $result_array[0]->{to_key_name(item =>$field)}->{to_key_name(item => $sub_field)} = $values[1] if ($values[1]);
	    }
        }
        else {
            if ($value) {
                 $result_array[0]->{ to_key_name( item =>$field) } .= ' '.$line;
            }
            if (!$value && !$sub_field) {
                $result_array[0]->{ to_key_name( item =>$field) } = $line;
                $value = 1;
            }
            if ($sub_field) {
                $result_array[0]->{to_key_name(item =>$field)}->{to_key_name(item =>$sub_field)} = $line;
                $sub_field = '';
            }
        }
    }
    return \@result_array;
} # end of bmc_devices_version

sub _parse_for_errors {
    my %args          = @_;
    my $output        = $args{raw_output};
    my $error_message = $args{error_message};
    my $command       = $args{command};
    my $error;
    if ( $output =~ /^usage.+/si ) {
        $error          = "InvalidParam";
        $$error_message = $output;
    }
    elsif ( $output =~ /bmc:\s+(unrecognized\s+command)\s*(.*)/ ) {
        $error = "CommandNotFound", $$error_message = $1;
    }
    elsif ( $output =~ /bmc\s*:\s*ignoring\s*(.*)/ ) {
        $error = "InvalidParam", $$error_message = $output;
    }
    elsif ( $output =~ /.*(Error.*)/i ) {
        $$error_message = $1;
        $error          = "CommandNotFound";
    }elsif ( $output =~ /(.*Invalid WordCnt.*)/i ) {
        $$error_message = $1;
        $error          = "CommandFailed";
    }elsif ( $output =~ /(request not applicable.*)/i ) {
        $error = "InvalidParam",
        $$error_message = $output;
    }elsif (( $output =~ /(.*not supported.*)/i )&& ($output !~ /TCP offload is not supported in IPv6/i) && ( $output !~ /System\s+DIMM\s+configuration\s+is\s+not\s+supported\s+by\s+FAS.*/) && ($output !~ /loopback not supported/i) && ($output !~ /Forensics not supported on this appliance/i )) {
        $$error_message = $output;
        $error          = "InvalidParam";
    }elsif ( $output =~ /(FW package install failed)/ ) {
        $$error_message = $output;
        $error          = "CommandFailed";
    }
    return $error;
}
1;
