#
# Copyright (c) 2014 NetApp, Inc., All Rights Reserved
# Any use, modification, or distribution is prohibited
# without prior written consent from NetApp, Inc.
#
##
## @synopsis
##    use gdb;
##
##    gdb_ssh(%args);
##    gdb_telnet(%args);
##    gdb_status(%args);
##    gdb_stop(%args);
##
##
## @summary Routines to parse gdb command output
## @author dl-nacl-dev

package NACL::APISet::Node::CLI::SP::default::gdb;

###################
# Everytime
###################
use strict;
use warnings;
use NACL::APISet::Response::CLI::ParserUtils;
# Include exceptions which will be thrown

use NACL::APISet::Exceptions::InvalidParamException qw(:try);
use NACL::APISet::Exceptions::CommandFailedException qw(:try);
use Tharn;
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();


## @name ::gdb_ssh
##
## @summary Parses the raw output from the "gdb ssh" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from gdb_ssh with any of its options as described in the
##    corresponding cdef file 'gdb.cdefs'
## @synopsis
##      gdb_ssh(%args);
## INPUTS TO METHOD CALL 'gdb_ssh' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

##&nbsp;Command = gdb ssh
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [];
##&nbsp;
sub gdb_ssh {
    my %args = @_;
    my $resp    = $args{response_object};
    my $output  = $resp->get_processed_output();
    my $command = $resp->get_actual_command();
    my @result_array;
    _check_error(
        "raw_output"      => $output,
        "response_object" => $resp
    );

    # If there is no exception thrown and this part of the code is reached,
    # then it implies that the command had succeeded.
    return [];
}

## @name ::gdb_telnet
##
## @summary Parses the raw output from the "gdb telnet" command
##
## @description 
##    This method can be used for parsing the raw output obtained after calling 
##    the method from gdb_telnet with any of its options as described in the 
##    corresponding cdef file 'gdb.cdefs'
## @synopsis
##      gdb_telnet(%args);
## INPUTS TO METHOD CALL 'gdb_telnet' ON APISET OBJECT 
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples
##&nbsp;
##&nbsp;Command = gdb telnet
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [];
##&nbsp;
sub gdb_telnet {
    my %args = @_;
    my $resp    = $args{response_object};
    my $output  = $resp->get_processed_output();
    my $command = $resp->get_actual_command();
    my @result_array;
    _check_error(
        "raw_output"      => $output,
        "response_object" => $resp
    );

    # If there is no exception thrown and this part of the code is reached,
    # then it implies that the command had succeeded.
    return [];
}

## @name ::gdb_status
##
## @summary Parses the raw output from the "gdb status" command
##
## @description 
##    This method can be used for parsing the raw output obtained after calling 
##    the method from gdb_status with any of its options as described in the 
##    corresponding cdef file 'gdb.cdefs'
## @synopsis
##      gdb_status(%args);
## INPUTS TO METHOD CALL 'gdb_status' ON APISET OBJECT 
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples
##&nbsp;Command = gdb status
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;GDB service is not active.
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;          {
##&nbsp;            'status' => 'GDB service is not active.'
##&nbsp;          }
##&nbsp;        ];
##&nbsp;
sub gdb_status {
    my %args = @_;
    my $resp    = $args{response_object};
    my $output  = $resp->get_processed_output();
    my $command = $resp->get_actual_command();
    my @result_array;
	_check_error(
        "raw_output"      => $output,
        "response_object" => $resp
    );
    # Now we can parse the output
    my %hash;
	$output =~ s/\s+$//g;
    $hash{"status"} = $output;
    return [ \%hash ];
}    ##



## @name ::gdb_stop
##
## @summary Parses the raw output from the "gdb stop" command
##
## @description 
##    This method can be used for parsing the raw output obtained after calling 
##    the method from gdb_stop with any of its options as described in the 
##    corresponding cdef file 'gdb.cdefs'
## @synopsis
##      gdb_stop(%args);
## INPUTS TO METHOD CALL 'gdb_stop' ON APISET OBJECT 
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples
##&nbsp;Command = gdb stop
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;GDB service stopped.
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;          {
##&nbsp;            'status' => 'GDB service stopped.'
##&nbsp;          }
##&nbsp;        ];
##&nbsp;
sub gdb_stop {
    my %args = @_;
    my $resp    = $args{response_object};
    my $output  = $resp->get_processed_output();
    my $command = $resp->get_actual_command();
    my @result_array;
    _check_error(
        "raw_output"      => $output,
        "response_object" => $resp
    );
    my %hash;
	$output =~ s/\s+$//g;
    $hash{"status"} = $output;
    return [ \%hash ];
}    ##

sub _check_error {
    my %args   = @_;
    my $output = $args{raw_output};
    my $resp   = $args{"response_object"};
    my $error;
    my $error_message;
    if ( $output =~ /(usage.*|Invalid\s+command\s+usage)/i ) {
        $error         = "InvalidParam";
        $error_message = $output;
    }elsif ( $output =~ /gdb ssh \-/i ) {
        $error         = "CommandFailed";
        $error_message = $output;
    }
    if ($error) {
        $error = 'NACL::APISet::Exceptions::' . $error . 'Exception';
        $error->throw( text => $error_message, response_object => $resp );
    }
}
1;
