#
# Copyright (c) 2012-13 NetApp, Inc., All Rights Reserved
# Any use, modification, or distribution is prohibited
# without prior written consent from NetApp, Inc.
#
##
## @synopsis
##  system_acp_show(%args);
##  system_acp_model_show(%args);
##  system_acp_sensors_show(%args);
##  system_acp_model_refresh(%args);
##  system_fru_show(%args);
##  system_fru_list(%args);
##  system_fru_log_show(%args);
##  system_fru_log_clear(%args);
##  system_fru_led(%args);
##  system_battery_show(%args);
##  system_battery_flash(%args);
##  system_battery_verify(%args);
##  system_battery_auto_update(%args);
##  system_bbsp(%args);
##  system_log(%args);
##  system_fan_version(%args);
##  system_fan_update(%args);
##  system_pci_slot_ctrl(%args);
##  system_pci_slot_status(%args);
##  system_pci_slot_interval_show(%args);
##  system_pci_slot_interval_set(%args);
##  system_pci_dev_status(%args);
##  system_pci_dev_ctrl(%args);
##  system_battery_charger(%args);
##  system_scratchpad_read(%args);
##  system_forensics_enable(%args);
##  system_forensics_disable(%args);
##  system_forensics_set_time_limit(%args);
##  system_forensics_show(%args);
##  system_forensics_log(%args);
##  system_forensics_config(%args);
##  system_cpld_version(%args);
##  system_cpld_update(%args);
##
## @summary Routines to parse SP command output
## @author Akshay Mohit(makshay@netapp.com), dl-nacl-dev@netapp.com

package NACL::APISet::Node::CLI::SP::default::system;
###################
# Everytime
###################
use strict;
use warnings;
use NACL::APISet::Response::CLI::ParserUtils;

# Include exceptions which will be thrown

use NACL::APISet::Exceptions::InvalidParamException qw(:try);
use NACL::APISet::Exceptions::CommandNotFoundException qw(:try);
use NACL::APISet::Exceptions::CommandFailedException;
use Tharn;
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();

sub _check_error {
    my %args   = @_;
    my $output = $args{raw_output};
    my $resp   = $args{"response_object"};
    my $error;
    my $error_message;
    if ( $output =~ /[Uu]sage.*|Invalid\s+command\s+usage/ ) {
        $error         = "InvalidParam";
        $error_message = $output;
    }
    elsif ( $output =~ /(.*)not supported(.*)/i ) {
        $error         = "CommandNotFound";
        $error_message = $output;
    }
    elsif ( $output =~ /(Invalid.*)/i ) {
        $error         = "InvalidParam";
        $error_message = $1;
    }
    elsif ( $output =~ /(^Error.*)/i ) {
        $error         = "InvalidParam";
        $error_message = $1;
    }
    elsif ( $output =~ /(Invalid URL.*|not\s+an\s+http|bad\s+address)/i ) {
        $error         = "InvalidParam";
        $error_message = $1;
    }
    elsif ( $output =~ /.*Failure programming firmware/i ) {
        $error         = "CommandFailed";
        $error_message = $output;
    }
    if ($error) {
        $error = 'NACL::APISet::Exceptions::' . $error . 'Exception';
        $error->throw( text => $error_message, response_object => $resp );
    }
}

## @name ::system_acp_show
##
## @summary Parses the raw output from the "system acp show" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from system acp show with any of its options as described in the
##    corresponding cdef file 'system.cdefs'
## @synopsis
##      system_acp_show(%args);
## INPUTS TO METHOD CALL 'system_acp_show' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples
##&nbsp;Command = system acp show; result -l
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;system acp show
##&nbsp;VPD
##&nbsp;chassis serial number  : SHJ000000002B3D\r\n
##&nbsp;chassis part number    : 116-00160+\r\n
##&nbsp;midplane product id    : DS424\r\n
##&nbsp;midplane serial number : BPS0961627G00B4\r\n
##&nbsp;enclosure WWN          : 50050CC10201D41D\r\n
##&nbsp;smp target address     : 500A09800016E03F\r\n
##&nbsp;board product id       : IOM6E\r\n
##&nbsp;board serial number    : 5006168062\r\n
##&nbsp;acp mac address        : 00:A0:98:18:32:EA\r\n
##&nbsp;ebod mac address       : 00:A0:98:18:32:EB\r\n
##&nbsp;ebod ip address        : 10.1.1.2\r\n
##&nbsp;netmask                : 255.255.255.0\r\n
##&nbsp;default gateway        : 0.0.0.0\r\n
##&nbsp;shelf generation code  : 1\r\n
##&nbsp;
##&nbsp;ACP Details\r\n
##&nbsp;acpp ip address                : 192.168.2.235\r\n
##&nbsp;acpp netmask                   : 255.0.0.0\r\n
##&nbsp;locked-wrench port link status : down\r\n
##&nbsp;
##&nbsp;Partner Module \r\n
##&nbsp;status         : valid partner detected\r\n
##&nbsp;familyId       : 1\r\n
##&nbsp;familyMemberId : 4\r\n
##&nbsp;
##&nbsp;Self-test status \r\n
##&nbsp;cpld interface          : passed\r\n
##&nbsp;expander eth interface  : passed\r\n
##&nbsp;
##&nbsp;Post code status \r\n
##&nbsp;code  : 0x8000\r\n
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp; $VAR1 = [
##&nbsp;           {
##&nbsp;             'locked-wrench_port_link_status' => ' down',
##&nbsp;             'cpld_interface' => ' passed',
##&nbsp;             'acpp_netmask' => ' 255.0.0.0',
##&nbsp;             'status' => ' valid partner detected',
##&nbsp;             'midplane_serial_number' => ' BPS0961627G00B4',
##&nbsp;             'chassis_part_number' => ' 116-00160+',
##&nbsp;             'board_product_id' => ' IOM6E',
##&nbsp;             'familymemberid' => ' 4',
##&nbsp;             'smp_target_address' => ' 500A09800016E03F',
##&nbsp;             'board_serial_number' => ' 5006168062',
##&nbsp;             'enclosure_wwn' => ' 50050CC10201D41D',
##&nbsp;             'ebod_mac_address' => 'EB',
##&nbsp;             'code' => ' 0x8000',
##&nbsp;             'familyid' => ' 1',
##&nbsp;             'shelf_generation_code' => ' 1',
##&nbsp;             'default_gateway' => ' 0.0.0.0',
##&nbsp;             'acpp_ip_address' => ' 192.168.2.235',
##&nbsp;             'midplane_product_id' => ' DS424',
##&nbsp;             'netmask' => ' 255.255.255.0',
##&nbsp;             'ebod_ip_address' => ' 10.1.1.2',
##&nbsp;             'expander_eth_interface' => ' passed',
##&nbsp;             'acp_mac_address' => 'EA',
##&nbsp;             'chassis_serial_number' => ' SHJ000000002B3D'
##&nbsp;           }
##&nbsp;         ];
##&nbsp;
sub system_acp_show {
    my %args = @_;
    my $error_message;
    my $resp   = $args{response_object};
    my $output = $resp->get_processed_output();
    my @result_array;
    my $error = _check_error(
        "raw_output"      => $output,
        "response_object" => $resp
    );

    #Now we can parse the output
    my @arraylines =
      get_lines( raw_output => $output, response_object => $resp );
    foreach my $lines (@arraylines) {
        if ( $lines =~
            /(\S+|\S+\s+\S+|\S+\s+\S+\s+\S+|\S+\s+\S+\s+\S+\s+\S+)\s+:\s+(\S+)/
          )
        {
            $result_array[0]->{ to_key_name( item => $1 ) } = $2;
        }
    }
    return \@result_array;
}    #end of system_acp_show

## @name ::system_acp_model_show
##
## @summary Parses the raw output from the "system acp model show" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from system acp model show with any of its options as described in the
##    corresponding cdef file 'system.cdefs'
## @synopsis
##      system_acp_model_show(%args);
## INPUTS TO METHOD CALL 'system_acp_model_show' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples
##&nbsp;Command = system acp model show; result -l
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;priv set diag
##&nbsp;Warning: These diagnostic commands are for use by support personnel only.
##&nbsp;
##&nbsp;system acp model show
##&nbsp;System model name status
##&nbsp;PCM FRU
##&nbsp;branding  : FAS
##&nbsp;PCM type  : SB_XIX
##&nbsp;
##&nbsp;Shelf type
##&nbsp;midplane product id  : DS424
##&nbsp;board product id     : IOM6E
##&nbsp;generation code      : 1
##&nbsp;
##&nbsp;Model  : FAS2240-4
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp; $VAR1 = [
##&nbsp;           {
##&nbsp;             'branding' => 'FAS',
##&nbsp;             'pcm_type' => 'SB_XIX',
##&nbsp;             'model' => 'FAS2240-4',
##&nbsp;             'midplane_product_id' => 'DS424',
##&nbsp;             'generation_code' => '1',
##&nbsp;             'board_product_id' => 'IOM6E'
##&nbsp;           }
##&nbsp;         ];
##&nbsp;
sub system_acp_model_show {
    my %args = @_;
    my $error_message;
    my $resp   = $args{response_object};
    my $output = $resp->get_processed_output();
    my @result_array;
    my $error = _check_error(
        "raw_output"      => $output,
        "response_object" => $resp
    );

    #Now we can parse the output
    my @arraylines =
      get_lines( raw_output => $output, response_object => $resp );
    foreach my $lines (@arraylines) {
        if ( $lines =~ /(\S+|\S+\s+\S+|\S+\s+\S+\s+\S+)\s+:\s+(\S+)/ ) {
            $result_array[0]->{ to_key_name( item => $1 ) } = $2;
        }
    }
    return \@result_array;
}    #end of system_acp_model_show

## @name ::system_acp_sensor_show
##
## @summary Parses the raw output from the "system acp sensor show" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from system acp sensors show with any of its options as described in the
##    corresponding cdef file 'system.cdefs'
## @synopsis
##      system_acp_sensor_show(%args);
## INPUTS TO METHOD CALL 'system_acp_sensor_show' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples
##&nbsp;Command = system acp sensor show; result -l
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;priv set diag
##&nbsp;Warning: These diagnostic commands are for use by support personnel only.\r\n
##&nbsp;
##&nbsp;system acp sensors show
##&nbsp;Disk device installed element list: 1 2 3 4 5 6 7 8 9 10 11 12 13 14 15 16 17 18 19 20 21 22 23 24 ; with error: none
##&nbsp;Power Supply installed element list: 1 2 3 4 ; with error: none
##&nbsp;Voltage Sensor installed element list: 1 2 3 4 5 6 7 8 ; with error: none
##&nbsp;[1] 5.11 Volts
##&nbsp;[2] 12.26 Volts
##&nbsp;[3] 5.15 Volts
##&nbsp;[4] 12.14 Volts
##&nbsp;[5] 5.11 Volts
##&nbsp;[6] 12.14 Volts
##&nbsp;[7] 5.15 Volts
##&nbsp;[8] 12.26 Volts
##&nbsp;Current Sensor installed element list: 1 2 3 4 5 6 7 8 ; with error: none
##&nbsp;[1] 3430 mA
##&nbsp;[2] 6560 mA
##&nbsp;[3] 3430 mA
##&nbsp;[4] 5820 mA
##&nbsp;[5] 3390 mA
##&nbsp;[6] 6050 mA
##&nbsp;[7] 5850 mA
##&nbsp;[8] 7770 mA
##&nbsp;Cooling Unit installed element list: 1 2 3 4 5 6 7 8 ; with error: none
##&nbsp;[1] 3450 RPM
##&nbsp;[2] 3070 RPM
##&nbsp;[3] 3750 RPM
##&nbsp;[4] 3000 RPM
##&nbsp;[5] 3450 RPM
##&nbsp;[6] 3070 RPM
##&nbsp;[7] 3300 RPM
##&nbsp;[8] 3070 RPM
##&nbsp;Temperature Sensor installed element list: 1 2 3 4 5 6 7 8 9 10 11 12 ; with error: none
##&nbsp;[1] 17 C (62 F)
##&nbsp;[2] 24 C (75 F)
##&nbsp;[3] 25 C (77 F)
##&nbsp;[4] 32 C (89 F)
##&nbsp;[5] 25 C (77 F)
##&nbsp;[6] 33 C (91 F)
##&nbsp;[7] 25 C (77 F)
##&nbsp;[8] 32 C (89 F)
##&nbsp;[9] 25 C (77 F)
##&nbsp;[10] 34 C (93 F)
##&nbsp;[11] 37 C (98 F)
##&nbsp;[12] 38 C (100 F)
##&nbsp;ES Electronics installed element list: 1 2 ; with error: none
##&nbsp;SAS connector installed element list: 1 3 ; with error: none
##&nbsp;ACP installed element list: 1 2 ; with error: none
##&nbsp;Processor Complex installed element list: 1 2 ; with error: none
##&nbsp;SAS Expander Module installed element list: 1 2 ; with error: none
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp; $VAR1 = [
##&nbsp;           {
##&nbsp;             'element' => [
##&nbsp;                            {
##&nbsp;                              'error' => 'none',
##&nbsp;                              'sensor_name' => 'Disk device'
##&nbsp;                            },
##&nbsp;                            {
##&nbsp;                              'error' => 'none',
##&nbsp;                              'sensor_name' => 'Power Supply'
##&nbsp;                            },
##&nbsp;                            {
##&nbsp;                              'sensor_values' => {
##&nbsp;                                                   '6' => '12.14 Volts',
##&nbsp;                                                   '3' => '5.15 Volts',
##&nbsp;                                                   '7' => '5.15 Volts',
##&nbsp;                                                   '2' => '12.26 Volts',
##&nbsp;                                                   '8' => '12.26 Volts',
##&nbsp;                                                   '1' => '5.11 Volts',
##&nbsp;                                                   '4' => '12.14 Volts',
##&nbsp;                                                   '5' => '5.11 Volts'
##&nbsp;                                                 },
##&nbsp;                              'error' => 'none',
##&nbsp;                              'sensor_name' => 'Voltage Sensor'
##&nbsp;                            },
##&nbsp;                            {
##&nbsp;                              'sensor_values' => {
##&nbsp;                                                   '6' => '6050 mA',
##&nbsp;                                                   '3' => '3430 mA',
##&nbsp;                                                   '7' => '5850 mA',
##&nbsp;                                                   '2' => '6560 mA',
##&nbsp;                                                   '8' => '7770 mA',
##&nbsp;                                                   '1' => '3430 mA',
##&nbsp;                                                   '4' => '5820 mA',
##&nbsp;                                                   '5' => '3390 mA'
##&nbsp;                                                 },
##&nbsp;                              'error' => 'none',
##&nbsp;                              'sensor_name' => 'Current Sensor'
##&nbsp;                            },
##&nbsp;                            {
##&nbsp;                              'sensor_values' => {
##&nbsp;                                                   '6' => '3070 RPM',
##&nbsp;                                                   '3' => '3750 RPM',
##&nbsp;                                                   '7' => '3300 RPM',
##&nbsp;                                                   '2' => '3070 RPM',
##&nbsp;                                                   '8' => '3070 RPM',
##&nbsp;                                                   '1' => '3450 RPM',
##&nbsp;                                                   '4' => '3000 RPM',
##&nbsp;                                                   '5' => '3450 RPM'
##&nbsp;                                                 },
##&nbsp;                              'error' => 'none',
##&nbsp;                              'sensor_name' => 'Cooling Unit'
##&nbsp;                            },
##&nbsp;                            {
##&nbsp;                              'sensor_values' => {
##&nbsp;                                                   '6' => '33 C (91 F)',
##&nbsp;                                                   '11' => '37 C (98 F)',
##&nbsp;                                                   '3' => '25 C (77 F)',
##&nbsp;                                                   '7' => '25 C (77 F)',
##&nbsp;                                                   '9' => '25 C (77 F)',
##&nbsp;                                                   '12' => '38 C (100 F)',
##&nbsp;                                                   '2' => '24 C (75 F)',
##&nbsp;                                                   '8' => '32 C (89 F)',
##&nbsp;                                                   '1' => '17 C (62 F)',
##&nbsp;                                                   '4' => '32 C (89 F)',
##&nbsp;                                                   '10' => '34 C (93 F)',
##&nbsp;                                                   '5' => '25 C (77 F)'
##&nbsp;                                                 },
##&nbsp;                              'error' => 'none',
##&nbsp;                              'sensor_name' => 'Temperature Sensor'
##&nbsp;                            },
##&nbsp;                            {
##&nbsp;                              'error' => 'none',
##&nbsp;                              'sensor_name' => 'ES Electronics'
##&nbsp;                            },
##&nbsp;                            {
##&nbsp;                              'error' => 'none',
##&nbsp;                              'sensor_name' => 'SAS connector'
##&nbsp;                            },
##&nbsp;                            {
##&nbsp;                              'error' => 'none',
##&nbsp;                              'sensor_name' => 'ACP'
##&nbsp;                            },
##&nbsp;                            {
##&nbsp;                              'error' => 'none',
##&nbsp;                              'sensor_name' => 'Processor Complex'
##&nbsp;                            },
##&nbsp;                            {
##&nbsp;                              'error' => 'none',
##&nbsp;                              'sensor_name' => 'SAS Expander'
##&nbsp;                            }
##&nbsp;                          ]
##&nbsp;           }
##&nbsp;         ];
##&nbsp;
sub system_acp_sensors_show {
    my %args = @_;
    my $error_message;
    my $resp   = $args{response_object};
    my $output = $resp->get_processed_output();
    my $i      = -1;
    my @result_array;
    my $error = _check_error(
        "raw_output"      => $output,
        "response_object" => $resp
    );

    #Now we can parse the output
    my @arraylines =
      get_lines( raw_output => $output, response_object => $resp );
    foreach my $lines (@arraylines) {
        if ( $lines =~
/^(\w*\s*\w+)\s+\w+\s+\w+\s+\w+\s*\w*:\s+(\d+\s+){1,};\s+\w+\s+\w+:\s+(.*)/
          )
        {
            $i++;
            $result_array[0]->{"element"}[$i]{"sensor_name"} = $1;
            $result_array[0]->{"element"}[$i]{"with_error"}  = $3;
        }
        elsif ( $lines =~ /\s*\[(\w+)\]\s+(.*)/ ) {
            $result_array[0]->{"element"}[$i]{"sensor_values"}{$1} = $2;
        }
    }
    return \@result_array;
}    #end of system_acp_sensor_show

## @name ::system_acp_model_refresh
##
## @summary Parses the raw output from the "system acp model refresh" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from system acp model refresh with any of its options as described in the
##    corresponding cdef file 'system.cdefs'
## @synopsis
##      system_acp_model_refresh(%args);
## INPUTS TO METHOD CALL 'system acp model refresh' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples
##&nbsp;Command = system acp model refresh; result -l
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;priv set diag
##&nbsp;Warning: These diagnostic commands are for use by support personnel only.
##&nbsp;
##&nbsp;system acp model refresh
##&nbsp;Refreshing midplane ...  successful
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;          {
##&nbsp;             'status' => 'successful'
##&nbsp;          }
##&nbsp;        ];
##&nbsp;
sub system_acp_model_refresh {
    my %args = @_;
    my $error_message;
    my $resp   = $args{response_object};
    my $output = $resp->get_processed_output();
    my @result_array;
    my $error = _check_error(
        "raw_output"      => $output,
        "response_object" => $resp
    );

    #Now we can parse the output
    my @arraylines =
      get_lines( raw_output => $output, response_object => $resp );
    foreach my $lines (@arraylines) {
        if ( $lines =~ /\S+\s\S+\s...\s+(\S+)/ ) {
            $result_array[0]->{"status"} = $1;
        }
    }
    return \@result_array;
}    #end of system_acp_model_refresh

## @name ::system_fru_show
##
## @summary Parses the raw output from the "system fru show" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from system fru show with any of its options as described in the
##    corresponding cdef file 'system.cdefs'
## @synopsis
##      system_fru_show(%args);
## INPUTS TO METHOD CALL 'system fru show' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## <tr> <td> No switch for this key </td> <td> 'fru-id' </td> <td> '0' </td> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples
##&nbsp;Command = system fru show; result -l
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp; system fru show 0
##&nbsp; 'Controller' inventory data:
##&nbsp;
##&nbsp; *** Product Info Area ***
##&nbsp;Mfg name            NetApp, Inc.
##&nbsp;Product Name        FAS
##&nbsp;Product PartNum     111-00846
##&nbsp;Product Version     A1
##&nbsp;Product SerNum      5006168062
##&nbsp;Product AssetTag    500000243941
##&nbsp;Custom              MAC1:00:A0:98:18:32:EA;MAC2:00:A0:98:18:32:EA;PLAT:SB_XIX;HA:H;
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp; $VAR1 = [
##&nbsp;           {
##&nbsp;             'controller' => [
##&nbsp;                               {
##&nbsp;                                 'product_sernum' => '5006168062',
##&nbsp;                                 'product_partnum' => '111-00846',
##&nbsp;                                 'product_name' => 'FAS',
##&nbsp;                                 'custom' => 'MAC1:00:A0:98:18:32:EA;MAC2:00:A0:98:18:32:EA;PLAT:SB_XIX;HA:H;',
##&nbsp;                                 'product_assettag' => '500000243941',
##&nbsp;                                 'product_version' => 'A1',
##&nbsp;                                 'mfg_name' => 'NetApp, Inc.'
##&nbsp;                               }
##&nbsp;                             ]
##&nbsp;           }
##&nbsp;         ];
#&nbsp;
sub system_fru_show {
    my %args = @_;
    my $error_message;
    my $resp   = $args{response_object};
    my $output = $resp->get_processed_output();
    my @result_array;
    my $i     = 0;
    my $j     = 0;
    my $key   = "";
    my $key1  = "";
    my $error = _check_error(
        "raw_output"      => $output,
        "response_object" => $resp
    );

    #Now we can parse the output
    my @arraylines = get_lines(
        raw_output      => $output,
        first_line      => '\S+\s+Product Info Area\s+\S+',
        response_object => $resp
    );

    #Shift to get rid of the line Product Info Area
    shift @arraylines;
    foreach my $line (@arraylines) {
        if ( $line =~ /Mfg name\s+(.*)$/i ) {
            $result_array[0]->{mfg_name} = $1;
        }
        elsif ( $line =~ /(Product Name)\s+(\w+\s+\w+)/i ) {
            $result_array[0]->{ to_key_name( item => $1 ) } = $2;
        }
        else {
            my @items = split( /\s+/, $line );
            my $value = pop(@items);
            my $key   = to_key_name( item => join( " ", @items ) );
            $result_array[0]->{$key} = $value;
        }
    }

    return \@result_array;
}    #end of system_fru_show

## @name ::system_fru_set
##
## @summary Parses the raw output from the "system fru set" command
##
## @description 
##    This method can be used for parsing the raw output obtained after calling 
##    the method from system_fru_set with any of its options as described in the 
##    corresponding cdef file 'system.cdefs'
## @synopsis
##      system_fru_set(%args);
## INPUTS TO METHOD CALL 'system_fru_set' ON APISET OBJECT 
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## <tr> <td> No switch for this key </td> <td> 'fru-id' </td> <td> '0' </td> </tr> 
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

##&nbsp;Command = system fru set 1
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;
##&nbsp;Warning: This is for developers use only! Please follow FRU data format programming specification.
##&nbsp;Backspace not supported!
##&nbsp;
##&nbsp;Enter all required 'Midplane' data fields:
##&nbsp;  - Use '.' to clear field
##&nbsp;  - Hit enter to accept default
##&nbsp;  - Enter new data
##&nbsp;
##&nbsp;Mfg name [NetApp, Inc.]: 
##&nbsp;
##&nbsp;Product Name [FAS2554]: 
##&nbsp;
##&nbsp;Product PartNum [DS424]: 
##&nbsp;
##&nbsp;Product Version []: 
##&nbsp;
##&nbsp;Product SerNum [BPS0961627G4CMV]: 
##&nbsp;
##&nbsp;Product AssetTag []: 
##&nbsp;
##&nbsp;Product FRU File ID[]: 
##&nbsp;
##&nbsp;Custom [HA:H;]: 
##&nbsp;
##&nbsp;
##&nbsp;New 'Midplane' inventory data:
##&nbsp;
##&nbsp;*** Product Info Area ***
##&nbsp;Mfg name            NetApp, Inc.
##&nbsp;Product Name        FAS2554
##&nbsp;Product PartNum     DS424
##&nbsp;Product SerNum      BPS0961627G4CMV
##&nbsp;Custom              HA:H;
##&nbsp;
##&nbsp;Commit new data to 'Midplane'('yes' or 'no')? 
##&nbsp;yes
##&nbsp;
##&nbsp;'Midplane' inventory data programming completed
##&nbsp;
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;          {
##&nbsp;            'controller_info' => [
##&nbsp;                                   {
##&nbsp;                                    'product_partnum' => 'DS424',
##&nbsp;                                    'custom' => 'HA:H;',
##&nbsp;                                    'mfg_name' => 'NetApp, Inc.',
##&nbsp;                                    'product_sernum' => 'BPS0961627G4CMV',
##&nbsp;                                    'product_name' => 'FAS2554'
##&nbsp;                                   }
##&nbsp;                                 ]
##&nbsp;          }
##&nbsp;        ];
##&nbsp;
sub system_fru_set {
    my %args = @_;
    my $error_message;
    my $resp   = $args{response_object};
    my $output = $resp->get_processed_output();
    my @result_array;

    if($output !~ /This is for developers use only/i){
	my $error = _check_error(
	   "raw_output"      => $output,
	   "response_object" => $resp
	);
    }

    #Now we can parse the output
    my @arraylines = get_lines(
        raw_output      => $output,
        first_line      => '\S+\s+Product Info Area\s+\S+',
        response_object => $resp
    );

    #Shift to get rid of the line Product Info Area
    shift @arraylines;
    foreach my $line (@arraylines) {
        last if($line =~ /commit/i);
	my @items = split( /\s\s+/, $line );
	my $value = pop(@items);
	my $key   = to_key_name( item => join( " ", @items ) );
        $result_array[0]->{'controller_info'}->[0]{$key} = $value;
    }

    return \@result_array;
}    #end of system_fru_set

## @name ::system_fru_list
##
## @summary Parses the raw output from the "system fru list" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from system fru list with any of its options as described in the
##    corresponding cdef file 'system.cdefs'
## @synopsis
##      system_fru_list(%args);
## INPUTS TO METHOD CALL 'system fru list' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples
##&nbsp;Command = system fru list; result -l
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;system fru list
##&nbsp;
##&nbsp;  FRU ID      Name
##&nbsp;==============================================
##&nbsp;    0         Controller
##&nbsp;    1         Midplane
##&nbsp;    2         NVMEM Bat
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp; $VAR1 = [
##&nbsp;           {
##&nbsp;             'fru_list' => [
##&nbsp;                             {
##&nbsp;                               'fru_id' => '0',
##&nbsp;                               'name' => 'Controller'
##&nbsp;                             },
##&nbsp;                             {
##&nbsp;                               'fru_id' => '1',
##&nbsp;                               'name' => 'Midplane'
##&nbsp;                             },
##&nbsp;                             {
##&nbsp;                               'fru_id' => '2',
##&nbsp;                               'name' => 'NVMEM Bat'
##&nbsp;                             }
##&nbsp;                           ]
##&nbsp;           }
##&nbsp;         ];
##&nbsp;
sub system_fru_list {
    my %args = @_;
    my $error_message;
    my $resp   = $args{response_object};
    my $output = $resp->get_processed_output();
    my @result_array;
    my $i     = 0;
    my $error = _check_error(
        "raw_output"      => $output,
        "response_object" => $resp
    );

    #Now we can parse the output
    my @arraylines =
      get_lines( raw_output => $output, response_object => $resp );
    foreach my $lines (@arraylines) {
        if ( $lines =~ /(\d+)\s+(\S+\s+\S+|\S+)/ ) {
            $result_array[0]->{"fru_list"}[$i]{"fru_id"} = $1;
            $result_array[0]->{"fru_list"}[$i]{"name"}   = $2;
            $i++;
        }
    }
    return \@result_array;
}    #end of system_fru_list

## @name ::system_fru_log_show
##
## @summary Parses the raw output from the "system fru log show" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from system fru log show with any of its options as described in the
##    corresponding cdef file 'system.cdefs'
## @synopsis
##      system_fru_log_show(%args);
## INPUTS TO METHOD CALL 'system fru log show' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples
##&nbsp;Command = system fru log show; result -l
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;priv set diag
##&nbsp;Warning: These diagnostic commands are for use by support personnel only.
##&nbsp;
##&nbsp;system fru log show
##&nbsp;
##&nbsp;Controller
##&nbsp;
##&nbsp;install:0001-date:2012-02-02-08:46:00
##&nbsp;*** Product Info Area ***
##&nbsp;Mfg name            NetApp, Inc.
##&nbsp;Product Name        FAS
##&nbsp;Product PartNum     111-00846
##&nbsp;Product Version     A1
##&nbsp;Product SerNum      5006168062
##&nbsp;Product AssetTag    500000243941
##&nbsp;Custom              MAC1:00:A0:98:18:32:EA;MAC2:00:A0:98:18:32:EA;PLAT:SB_XIX;HA:H;
##&nbsp;
##&nbsp;Midplane
##&nbsp;
##&nbsp;install:0001-date:2012-02-02-08:46:00
##&nbsp;*** Product Info Area ***
##&nbsp;Mfg name            NetApp, Inc.
##&nbsp;Product Name        FAS2240-4
##&nbsp;Product PartNum     DS424
##&nbsp;Product SerNum      BPS0961627G00B4
##&nbsp;Custom              HA:H;
##&nbsp;
##&nbsp;NVMEM Bat
##&nbsp;
##&nbsp;install:0001-date:2012-02-02-08:46:00
##&nbsp;*** Product Info Area ***
##&nbsp;Mfg name            Nexergy:6/17/2011
##&nbsp;Product Name        bq20z80
##&nbsp;Product PartNum     27100029
##&nbsp;Product Version     0x0000:0x0210:0x00a1
##&nbsp;Product SerNum      0x6b23
##&nbsp;Custom              CHEMISTRY:LION
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp; $VAR1 = [
##&nbsp;           {
##&nbsp;             'controller' => {
##&nbsp;                               'install_0001' => [
##&nbsp;                                                   {
##&nbsp;                                                     'product_sernum' => '5006168062',
##&nbsp;                                                     'product_partnum' => '111-00846',
##&nbsp;                                                     'product_name' => 'FAS',
##&nbsp;                                                     'custom' => 'MAC1:00:A0:98:18:32:EA;MAC2:00:A0:98:18:32:EA;PLAT:SB_XIX;HA:H;',
##&nbsp;                                                     'product_assettag' => '500000243941',
##&nbsp;                                                     'product_version' => 'A1',
##&nbsp;                                                     'mfg_name' => 'NetApp, Inc.'
##&nbsp;                                                   }
##&nbsp;                                                 ]
##&nbsp;                             },
##&nbsp;             'nvem_bat' => {
##&nbsp;                              'install_0001' => [
##&nbsp;                                                  {
##&nbsp;                                                    'product_sernum' => '0x6b23',
##&nbsp;                                                    'product_partnum' => '27100029',
##&nbsp;                                                    'product_name' => 'bq20z80',
##&nbsp;                                                    'custom' => 'CHEMISTRY:LION',
##&nbsp;                                                    'product_version' => '0x0000:0x0210:0x00a1',
##&nbsp;                                                    'mfg_name' => 'Nexergy:6/17/2011'
##&nbsp;                                                  }
##&nbsp;                                                ]
##&nbsp;                            },
##&nbsp;             'midplane' => {
##&nbsp;                             'install_0001' => [
##&nbsp;                                                 {
##&nbsp;                                                   'product_sernum' => 'BPS0961627G00B4',
##&nbsp;                                                   'product_partnum' => 'DS424',
##&nbsp;                                                   'product_name' => 'FAS2240-4',
##&nbsp;                                                   'custom' => 'HA:H;',
##&nbsp;                                                   'mfg_name' => 'NetApp, Inc.'
##&nbsp;                                                 }
##&nbsp;                                               ]
##&nbsp;                           }
##&nbsp;           }
##&nbsp;         ];
##&nbsp;
sub system_fru_log_show {
    my %args = @_;
    my $error_message;
    my $resp   = $args{response_object};
    my $output = $resp->get_processed_output();
    my @result_array;
    my $key   = "";
    my $key1  = "";
    my $error = _check_error(
        "raw_output"      => $output,
        "response_object" => $resp
    );
    my @lines = split( '\n', $output );

    #Getting rid of first blank line after system fru log show shift line
    shift @lines;
    while ( my $line = shift(@lines) ) {
        $key = $line;

        #Getting rid of line *** Product Info Area *** shift line
        shift(@lines);
        $line = shift(@lines);
        if ( $line =~ /(install)\:(\d+).*/ ) {
            $key1 = $1 . "_" . $2;
        }

        #Getting rid of last blank line before next key starts.
        shift @lines;
        while ( $line = shift @lines ) {
            if ( $line =~ /^(\S+|\S+\s\S+)\s+(\S+|\w+,\s+\w+\.)$/ ) {
                $result_array[0]->{ to_key_name( item => $key ) }{$key1}->[0]
                  ->{ to_key_name( item => $1 ) } = $2;
            }
        }
    }
    return \@result_array;
}    #end of system_fru_log_show

## @name ::system_fru_log_clear
##
## @summary Parses the raw output from the "system fru log clear" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from system fru log clear with any of its options as described in the
##    corresponding cdef file 'system.cdefs'
## @synopsis
##      system_fru_log_clear(%args);
## INPUTS TO METHOD CALL 'system fru log clear' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples
##&nbsp;Command = system fru log clear; result -l
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;priv set diag
##&nbsp;Warning: These diagnostic commands are for use by support personnel only.
##&nbsp;
##&nbsp;system fru log clear
##&nbsp;All FRU logs have been cleared. All FRUs have been re-scanned and the new FRU logs have been created
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp; $VAR1 = [
##&nbsp;           {
##&nbsp;             'log' => 'All FRU logs have been cleared. All FRUs have been re-scanned and the new FRU logs have been created'
##&nbsp;           }
##&nbsp;         ];
##&nbsp;
sub system_fru_log_clear {
    my %args = @_;
    my $error_message;
    my $resp   = $args{response_object};
    my $output = $resp->get_processed_output();
    my @result_array;
    my $error = _check_error(
        "raw_output"      => $output,
        "response_object" => $resp
    );

    #Now we can parse the output
    my @arraylines =
      get_lines( raw_output => $output, response_object => $resp );
    foreach my $lines (@arraylines) {
        if ( $lines =~ /(.*)/ ) {
            $result_array[0]->{"log"} = $1;
        }
    }
    return \@result_array;
}    #end of system_fru_log_clear

## @name ::system_fru_led
##
## @summary Parses the raw output from the "system fru led" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from system fru led with any of its options as described in the
##    corresponding cdef file 'system.cdefs'
## @synopsis
##      system_fru_led(%args);
## INPUTS TO METHOD CALL 'system fru led' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## <tr> <td> No switch for this key </td> <td> 'show' </td> <td> '0' </td> </tr>
## <tr> <td> No switch for this key </td> <td> 'set' </td> <td> '0' </td> </tr>
## <tr> <td> No switch for this key </td> <td> 'fru-id' </td> <td> '0' </td> </tr>
## <tr> <td> No switch for this key </td> <td> 'power' </td> <td> '0' </td> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples
##&nbsp;Command = system fru led; result -l
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;priv set diag
##&nbsp;Warning: These diagnostic commands are for use by support personnel only.
##&nbsp;
##&nbsp;system fru led show all 
##&nbsp;FRU LED ID 1 is on
##&nbsp;FRU LED ID 2 is on
##&nbsp;FRU LED ID 3 is off
##&nbsp;FRU LED ID 4 is on
##&nbsp;FRU LED ID 5 show Unsuccessful
##&nbsp;FRU LED ID 6 is off
##&nbsp;FRU LED ID 7 show Unsuccessful
##&nbsp;FRU LED ID 8 show Unsuccessful
##&nbsp;FRU LED ID 9 is off
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp; $VAR1 = [
##&nbsp;           {
##&nbsp;             'FRU LED ID 1' => 'on',
##&nbsp;             'FRU LED ID 2' => 'on', 
##&nbsp;             'FRU LED ID 3' => 'off',
##&nbsp;             'FRU LED ID 4' => 'on',
##&nbsp;             'FRU LED ID 5' => 'Unsuccessful',
##&nbsp;             'FRU LED ID 6' => 'off',
##&nbsp;             'FRU LED ID 7' => 'Unsuccessful',
##&nbsp;             'FRU LED ID 8' => 'Unsuccessful',
##&nbsp;             'FRU LED ID 9' => 'off',
##&nbsp;           }
##&nbsp;         ];
##&nbsp;
sub system_fru_led {
    my %args = @_;
    my $error_message;
    my $resp   = $args{response_object};
    my $output = $resp->get_processed_output();
    my @result_array;
    my $error = _check_error(
        "raw_output"      => $output,
        "response_object" => $resp
    );

    #Now we can parse the output
    my @arraylines =
      get_lines( raw_output => $output, response_object => $resp );
    foreach my $lines (@arraylines) {
        if ( $lines =~ /(.*)\s+(is|show)\s+(on|off|Unsuccessful)/i ) {
            $result_array[0]->{to_key_name( item => $1) } = $3;
        }
    }
    return \@result_array;
}    #end of system_fru_led

## @name ::system_battery_show
##
## @summary Parses the raw output from the "system battery show" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from system battery show with any of its options as described in the
##    corresponding cdef file 'system.cdefs'
## @synopsis
##      system_battery_show(%args);
## INPUTS TO METHOD CALL 'system battery show' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples
##&nbsp;Command = system battery show; result -l
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;system battery show
##&nbsp;
##&nbsp;chemistry         : LION
##&nbsp;device-name       : bq20z80
##&nbsp;expected-load-mw  : 133
##&nbsp;id                : 27100029
##&nbsp;manufacturer      : Nexergy
##&nbsp;manufacturer-date : 6/17/2011
##&nbsp;rev_cell          : 0x0000
##&nbsp;rev_firmware      : 0x0210
##&nbsp;rev_hardware      : 0x00a1
##&nbsp;TI_fw_version     : 0x0110
##&nbsp;TI_hw_version     : 0xa1
##&nbsp;serial            : 0x6b23
##&nbsp;status            : ready
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp; $VAR1 = [
##&nbsp;           {
##&nbsp;             'manufacturer-date' => '6/17/2011',
##&nbsp;             'chemistry' => 'LION',
##&nbsp;             'device-name' => 'bq20z80',
##&nbsp;             'status' => 'ready',
##&nbsp;             'serial' => '0x6b23',
##&nbsp;             'rev_firmware' => '0x0210',
##&nbsp;             'rev_hardware' => '0x00a1',
##&nbsp;             'ti_hw_version' => '0xa1',
##&nbsp;             'expected-load-mw' => '133',
##&nbsp;             'manufacturer' => 'Nexergy',
##&nbsp;             'id' => '27100029',
##&nbsp;             'rev_cell' => '0x0000',
##&nbsp;             'ti_fw_version' => '0x0110'
##&nbsp;           }
##&nbsp;         ];
##&nbsp;
sub system_battery_show {
    my %args = @_;
    my $error_message;
    my $resp   = $args{response_object};
    my $output = $resp->get_processed_output();
    my @result_array;
    my $error = _check_error(
        "raw_output"      => $output,
        "response_object" => $resp
    );

    #Now we can parse the output
    my @arraylines =
      get_lines( raw_output => $output, response_object => $resp );
    foreach my $lines (@arraylines) {
        if ( $lines =~ /(.*):\s+(.*)/ ) {
            $result_array[0]->{ to_key_name( item => $1 ) } = $2;
        }
    }
    return \@result_array;
}    #end of system_battery_show

## @name ::system_battery_flash
##
## @summary Parses the raw output from the "system battery flash" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from system battery flash with any of its options as described in the
##    corresponding cdef file 'system.cdefs'
## @synopsis
##      system_battery_flash(%args);
## INPUTS TO METHOD CALL 'system battery flash' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## <tr> <td> No switch for this key </td> <td> 'url' </td> <td> '0' </td> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples
sub system_battery_flash {
    my %args = @_;
    my $error_message;
    my $resp   = $args{response_object};
    my $output = $resp->get_processed_output();
    my @result_array;
    my $error = _check_error(
        "raw_output"      => $output,
        "response_object" => $resp
    );

    return [];
}    #end of system_battery_flash

## @name ::system_battery_verify
##
## @summary Parses the raw output from the "system battery verify" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from system battery verify with any of its options as described in the
##    corresponding cdef file 'system.cdefs'
## @synopsis
##      system_battery_verify(%args);
## INPUTS TO METHOD CALL 'system battery verify' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## <tr> <td> No switch for this key </td> <td> 'url' </td> <td> '0' </td> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples
##&nbsp;Command = system battery verify; result -l
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;priv set diag
##&nbsp;Warning: These diagnostic commands are for use by support personnel only.
##&nbsp;
##&nbsp;system battery verify
##&nbsp;Comparing battery firmware image from SP build
##&nbsp;Verify test passed
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp; $VAR1 = [
##&nbsp;           {
##&nbsp;             'verify_status' => 'Verify test passed'
##&nbsp;           }
##&nbsp;         ];
##&nbsp;
sub system_battery_verify {
    my %args = @_;
    my $error_message;
    my $resp   = $args{response_object};
    my $output = $resp->get_processed_output();
    my @result_array;
    my $error = _check_error(
        "raw_output"      => $output,
        "response_object" => $resp
    );

    #Now we can parse the output
    my @arraylines =
      get_lines( raw_output => $output, response_object => $resp );
    foreach my $lines (@arraylines) {
        if ( $lines =~ /(Verify.*)/ ) {
            $result_array[0]->{"verify_status"} = $1;
        }
    }
    return \@result_array;
}    #end of system_battery_verify

## @name ::system_battery_auto_update
##
## @summary Parses the raw output from the "system battery auto_update" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from system battery auto_update with any of its options as described in the
##    corresponding cdef file 'system.cdefs'
## @synopsis
##      system_battery_auto_update(%args);
## INPUTS TO METHOD CALL 'system battery verify' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## <tr> <td> No switch for this key </td> <td> 'status' </td> <td> '0' </td> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples
sub system_battery_auto_update {
    my %args = @_;
    my $error_message;
    my $resp   = $args{response_object};
    my $output = $resp->get_processed_output();
    my @result_array;
    my $error = _check_error(
        "raw_output"      => $output,
        "response_object" => $resp
    );
    return [];
}    #end of system_battery_auto_update

## @name ::system_bbsp
##
## @summary Parses the raw output from the "system bbsp" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from system bbsp with any of its options as described in the
##    corresponding cdef file 'system.cdefs'
## @synopsis
##      system_bbsp(%args);
## INPUTS TO METHOD CALL 'system bbsp' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples
##&nbsp;Command = system bbsp; result -l
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;priv set diag
##&nbsp;Warning: These diagnostic commands are for use by support personnel only.
##&nbsp;
##&nbsp;system bbsp
##&nbsp;BBSP  = 0x2f -- ONTAP is running
##&nbsp;
##&nbsp; -------- Parsed Output ------
##&nbsp; $VAR1 = [
##&nbsp;           {
##&nbsp;             'verify_status' => '0x2f -- ONTAP is running'
##&nbsp;           }
##&nbsp;         ];
##&nbsp;
sub system_bbsp {
    my %args = @_;
    my $error_message;
    my $resp   = $args{response_object};
    my $output = $resp->get_processed_output();
    my @result_array;
    my $error = _check_error(
        "raw_output"      => $output,
        "response_object" => $resp
    );

    #Now we can parse the output
    my @arraylines =
      get_lines( raw_output => $output, response_object => $resp );
    foreach my $lines (@arraylines) {
        if ( $lines =~ /(\S+)\s+=\s+(\S+\s+..\s+\S+\s+\S+\s+\S+)/ ) {
            $result_array[0]->{"verify_status"} = $2;
        }
    }
    return \@result_array;
}    #end of system_bbsp

## @name ::system_log
##
## @summary Parses the raw output from the "system log" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from system log with any of its options as described in the
##    corresponding cdef file 'system.cdefs'
## @synopsis
##      system_log(%args);
## INPUTS TO METHOD CALL 'system log' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples
sub system_log {
    my %args   = @_;
    my $resp   = $args{response_object};
    my $output = $resp->get_processed_output();
    my $error  = _check_error(
        "raw_output"      => $output,
        "response_object" => $resp
    );

    return [];
} ## end of system log

## @name ::system_fan_version
##
## @summary Parses the raw output from the "system fan version" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from system fan version with any of its options as described in the
##    corresponding cdef file 'system.cdefs'
## @synopsis
##      system_fan_version(%args);
## INPUTS TO METHOD CALL 'system_fan_version' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## <tr> <td> No switch for this key </td> <td> 'fan-controller' </td> <td> '0' </td> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples
##&nbsp;Command = system fan version; result -l
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;priv set diag
##&nbsp;Warning: These diagnostic commands are for use by support personnel only.
##&nbsp;
##&nbsp;system fan version 11
##&nbsp;Version: FC1.7
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp; $VAR1 = [
##&nbsp;           {
##&nbsp;             'fan_version' => 'FC1.7',
##&nbsp;           }
##&nbsp;         ];
sub system_fan_version {
    my %args = @_;
    my $error_message;
    my $resp    = $args{response_object};
    my $output  = $resp->get_processed_output();
    my $command = $resp->get_actual_command();
    my @result_array;
    my $error = _check_error(
        "raw_output"    => $output,
        "command"       => $command,
        "error_message" => \$error_message
    );

    if ($error) {
        $error = 'NACL::APISet::Exceptions::' . $error . 'Exception';
        $error->throw( text => $error_message, response_object => $resp );
    }

    #Now we can parse the output
    my @arraylines =
      get_lines( raw_output => $output, response_object => $resp );
    foreach my $lines (@arraylines) {
        if ( $lines =~ /Version:\s+(\S+)/ ) {
            $result_array[0]->{fan_version} = $1;
        }
    }
    return \@result_array;
} ## end of system fan version

## @name ::system_fan_update
##
## @summary Parses the raw output from the "system fan update" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from system fan update with any of its options as described in the
##    corresponding cdef file 'system.cdefs'
## @synopsis
##      system_fan_update(%args);
## INPUTS TO METHOD CALL 'system_fan_update' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## <tr> <td> No switch for this key </td> <td> 'fan-module' </td> <td> '0' </td> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples
sub system_fan_update {
    my %args   = @_;
    my $resp   = $args{response_object};
    my $output = $resp->get_processed_output();
    _check_error( raw_output => $output, response_object => $resp );

    return [];
} ## end of system fan update

## @name ::system_pci_slot_ctrl
##
## @summary Parses the raw output from the "system pci-slot-ctrl" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from system pci-slot-ctrl with any of its options as described in the
##    corresponding cdef file 'system.cdefs'
## @synopsis
##      system_pci_slot_ctrl(%args);
## INPUTS TO METHOD CALL 'system_pci_slot_ctrl' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## <tr> <td> 'No switch for this key' </td> <td> 'slot-num' </td> <td> '0' </td> </tr>
## <tr> <td> 'No switch for this key' </td> <td> 'control' </td> <td> '0' </td> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples
##&nbsp;Command = system pci-slot-ctrl 1 0 ; result -l
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;SP PCI slot 1(0x01) control command 0 completed successfully
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp; $VAR1 = [
##&nbsp;           {
##&nbsp;             'status' => 'completed successfully'
##&nbsp;           }
##&nbsp;         ];

sub system_pci_slot_ctrl {
    my %args   = @_;
    my $resp   = $args{response_object};
    my $output = $resp->get_processed_output();
    _check_error( raw_output => $output, response_object => $resp );
    my @lines = get_lines( raw_output => $output, response_object => $resp );
    my @result_array;
    if ( $lines[0] =~ /SP PCI slot.*command\s+\d+\s+(.*)/ ) {
        $result_array[0]->{status} = $1;
    }

    return \@result_array;
} ## end sub system_pci_slot_ctrl

## @name ::system_pci_slot_status
##
## @summary Parses the raw output from the "system pci-slot-status" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from system pci-slot-status with any of its options as described in the
##    corresponding cdef file 'system.cdefs'
## @synopsis
##      system_pci_slot_status(%args);
## INPUTS TO METHOD CALL 'system_pci_slot_status' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## <tr> <td> 'No switch for this key' </td> <td> 'slot-num' </td> <td> '0' </td> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples
##&nbsp;Command = system pci-slot-status;
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;SP PCI slot 1(0x01) status 0x00
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp; $VAR1 = [
##&nbsp;           {
##&nbsp;             'pci_slot' => '0x01',
##&nbsp;             'status' => '0x00'
##&nbsp;           }
##&nbsp;         ];

sub system_pci_slot_status {
    my %args   = @_;
    my $resp   = $args{response_object};
    my $output = $resp->get_processed_output();
    _check_error( raw_output => $output, response_object => $resp );
    my @lines = get_lines( raw_output => $output, response_object => $resp );
    my @result_array;
    if ( $lines[0] =~ /.*\((.*)\)\s+status\s+(.*)/ ) {
        $result_array[0]->{pci_slot} = $1;
        $result_array[0]->{status}   = $2;
    }
    return \@result_array;

} ## end sub system_pci_slot_status

## @name ::system_pci_slot_interval_show
##
## @summary Parses the raw output from the "system pci-slot-interval show" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from system pci-slot-interval show with any of its options as described in the
##    corresponding cdef file 'system.cdefs'
## @synopsis
##      system_pci_slot_interval_show(%args);
## INPUTS TO METHOD CALL 'system_pci_slot_interval_show' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## <tr> <td> 'No switch for this key' </td> <td> 'slot-num' </td> <td> '0' </td> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples
##&nbsp;Command = system pci-slot-interval show 1;
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;PCI slot 1 power cycle delay interval is 2 second(s).
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp; $VAR1 = [
##&nbsp;           {
##&nbsp;             'power_cycle_delay' => '2',
##&nbsp;             'pci_slot' => '1'
##&nbsp;           }
##&nbsp;         ];

sub system_pci_slot_interval_show {
    my %args   = @_;
    my $resp   = $args{response_object};
    my $output = $resp->get_processed_output();
    _check_error( raw_output => $output, response_object => $resp );
    my @lines = get_lines( raw_output => $output, response_object => $resp );
    my @result_array;
    if ( $lines[0] =~ /PCI\s+slot\s+(\d+).*(\d+).*/ ) {
        $result_array[0]->{pci_slot}          = $1;
        $result_array[0]->{power_cycle_delay} = $2;
    }
    return \@result_array;

} ## end sub system_pci_slot_interval_show

## @name ::system_pci_slot_interval_set
##
## @summary Parses the raw output from the "system pci-slot-interval set" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from system pci-slot-interval set with any of its options as described in the
##    corresponding cdef file 'system.cdefs'
## @synopsis
##      system_pci_slot_interval_set(%args);
## INPUTS TO METHOD CALL 'system_pci_slot_interval_set' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## <tr> <td> 'No switch for this key' </td> <td> 'slot-num' </td> <td> '0' </td> </tr>
## <tr> <td> 'No switch for this key' </td> <td> 'interval' </td> <td> '0' </td> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples
##&nbsp;Command = system pci-slot-interval set 1 2;
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;PCI slot 1 power cycle delay interval is 2 second(s).
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp; $VAR1 = [
##&nbsp;           {
##&nbsp;             'power_cycle_delay' => '2',
##&nbsp;             'pci_slot' => '1'
##&nbsp;           }
##&nbsp;         ];

sub system_pci_slot_interval_set {
    my %args   = @_;
    my $resp   = $args{response_object};
    my $output = $resp->get_processed_output();
    _check_error( raw_output => $output, response_object => $resp );
    my @lines = get_lines( raw_output => $output, response_object => $resp );
    my @result_array;
    if ( $lines[0] =~ /PCI\s+slot\s+(\d+).*(\d+).*/ ) {
        $result_array[0]->{pci_slot}          = $1;
        $result_array[0]->{power_cycle_delay} = $2;
    }
    return \@result_array;

} ## end sub system_pci_slot_interval_set

## @name ::system_pci_dev_status
##
## @summary Parses the raw output from the "system pci-dev-status" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from system pci-dev-status with any of its options as described in the
##    corresponding cdef file 'system.cdefs'
## @synopsis
##      system_pci_dev_status(%args);
## INPUTS TO METHOD CALL 'system_pci_dev_status' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## <tr> <td> 'No switch for this key' </td> <td> 'dev-num' </td> <td> '0' </td> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples
##&nbsp;Command = system pci-dev-status 0x1;
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;SP PCI dev 1(0x01) status 0x00
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp; $VAR1 = [
##&nbsp;           {
##&nbsp;             'pci_dev' => '0x01',
##&nbsp;             'status' => '0x00'
##&nbsp;           }
##&nbsp;         ];

sub system_pci_dev_status {
    my %args   = @_;
    my $resp   = $args{response_object};
    my $output = $resp->get_processed_output();
    _check_error( raw_output => $output, response_object => $resp );
    my @lines = get_lines( raw_output => $output, response_object => $resp );
    my @result_array;
    if ( $lines[0] =~ /.*\((.*)\)\s+status\s+(.*)/ ) {
        $result_array[0]->{pci_dev} = $1;
        $result_array[0]->{status}  = $2;
    }
    return \@result_array;

} ## end sub system_pci_dev_status

## @name ::system_pci_dev_ctrl
##
## @summary Parses the raw output from the "system pci-dev-ctrl" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from system pci-dev-ctrl with any of its options as described in the
##    corresponding cdef file 'system.cdefs'
## @synopsis
##      system_pci_dev_ctrl(%args);
## INPUTS TO METHOD CALL 'system_pci_dev_ctrl' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## <tr> <td> 'No switch for this key' </td> <td> 'dev-num' </td> <td> '0' </td> </tr>
## <tr> <td> 'No switch for this key' </td> <td> 'interval' </td> <td> '0' </td> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples
##&nbsp;Command = system pci-dev-ctrl 0x1 1;
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;SP PCI dev 1(0x01) control command completed successfully
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp; $VAR1 = [
##&nbsp;           {
##&nbsp;             'status' => 'completed successfully'
##&nbsp;           }
##&nbsp;         ];

sub system_pci_dev_ctrl {
    my %args   = @_;
    my $resp   = $args{response_object};
    my $output = $resp->get_processed_output();
    _check_error( raw_output => $output, response_object => $resp );
    my @lines = get_lines( raw_output => $output, response_object => $resp );
    my @result_array;
    if ( $lines[0] =~ /.*\s+command\s+(.*)/ ) {
        $result_array[0]->{status} = $1;
    }
    return \@result_array;

} ## end sub system_pci_dev_ctrl

## @name ::system_battery_charger
##
## @summary Parses the raw output from the "system battery charger" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from system battery charger with any of its options as described in the
##    corresponding cdef file 'system.cdefs'
## @synopsis
##      system_battery_charger(%args);
## INPUTS TO METHOD CALL 'system_battery_charger' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## <tr> <td> 'status' </td> <td> 'status' </td> <td> '1' </td> </tr>
## <tr> <td> 'enable' </td> <td> 'enable' </td> <td> '1' </td> </tr>
## <tr> <td> 'disable' </td> <td> 'disable' </td> <td> '1' </td> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples
##&nbsp;Command = system battery charger status;
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;Battery charger is Enabled
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp; $VAR1 = [
##&nbsp;           {
##&nbsp;             'status' => 'Enabled'
##&nbsp;           }
##&nbsp;         ];

sub system_battery_charger {
    my %args   = @_;
    my $resp   = $args{response_object};
    my $output = $resp->get_processed_output();
    _check_error( raw_output => $output, response_object => $resp );
    my @lines = get_lines( raw_output => $output, response_object => $resp );
    my @result_array;
    if ( $lines[0] =~ /\w+\s+\w+\s+\w+\s+(\S+)/ ) {
        $result_array[0]->{status} = $1;
    }
    return \@result_array;

} ## end sub system_battery_charger

## @name ::system_scratchpad_read
##
## @summary Parses the raw output from the "system scratchpad read" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from system_scratchpad_read with any of its options as described in the
##    corresponding cdef file 'system.cdefs'
## @synopsis
##      system_scratchpad_read(%args);
## INPUTS TO METHOD CALL 'system_scratchpad_read' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

##&nbsp;Command = system scratchpad read
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;region 1 header
##&nbsp;    Sig(0x5342414e)
##&nbsp;    Size(4096)
##&nbsp;    Ver(1)
##&nbsp;    DataLen(0)
##&nbsp;    DataCRC(0x0)
##&nbsp;    HdrCRC(0x50e4)
##&nbsp;
##&nbsp;region 2 header
##&nbsp;    Sig(0x544f414e)
##&nbsp;    Size(8192)
##&nbsp;    Ver(1)
##&nbsp;    DataLen(0)
##&nbsp;    DataCRC(0x0)
##&nbsp;    HdrCRC(0x2adc)
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;          {
##&nbsp;            'region_2_header' => [
##&nbsp;                                   {
##&nbsp;                                     'sig' => '0x544f414e',
##&nbsp;                                     'ver' => '1',
##&nbsp;                                     'datacrc' => '0x0',
##&nbsp;                                     'hdrcrc' => '0x2adc',
##&nbsp;                                     'datalen' => '0',
##&nbsp;                                     'size' => '8192'
##&nbsp;                                   }
##&nbsp;                                 ],
##&nbsp;            'region_1_header' => [
##&nbsp;                                   {
##&nbsp;                                     'sig' => '0x5342414e',
##&nbsp;                                     'ver' => '1',
##&nbsp;                                     'datacrc' => '0x0',
##&nbsp;                                     'hdrcrc' => '0x50e4',
##&nbsp;                                     'datalen' => '0',
##&nbsp;                                     'size' => '4096'
##&nbsp;                                   }
##&nbsp;                                 ]
##&nbsp;          }
##&nbsp;        ];
##&nbsp;

sub system_scratchpad_read {
    my %args   = @_;
    my $resp   = $args{response_object};
    my $output = $resp->get_processed_output();
    _check_error( raw_output => $output, response_object => $resp );
    my @lines = get_lines( raw_output => $output, response_object => $resp );
    my @result_array;
    my $count = 0;
    my $head;

    foreach my $line (@lines) {
        if ( $line =~ /region.+header/ ) {
            $head = to_key_name( item => $line );
        }
        elsif ( $line =~ /(\w+)\((.+)\)/ ) {
            $result_array[0]->{$head}->[$count]->{ to_key_name( item => $1 ) } =
              $2;
        }
    }
    return \@result_array;
} ## end sub system_scratchpad_read

## @name ::system_forensics_enable
##
## @summary Parses the raw output from the "system forensics enable" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from system_forensics_enable with any of its options as described in the
##    corresponding cdef file 'system.cdefs'
## @synopsis
##      system_forensics_enable(%args);
## INPUTS TO METHOD CALL 'system_forensics_enable' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an empty array reference.
sub system_forensics_enable {
    my %args   = @_;
    my $resp   = $args{response_object};
    my $output = $resp->get_processed_output();
    _check_error( raw_output => $output, response_object => $resp );
    return [];
} ## end sub system_forensics_enable

## @name ::system_forensics_disable
##
## @summary Parses the raw output from the "system forensics disable" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from system_forensics_disable with any of its options as described in the
##    corresponding cdef file 'system.cdefs'
## @synopsis
##      system_forensics_disable(%args);
## INPUTS TO METHOD CALL 'system_forensics_disable' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an empty array reference.

sub system_forensics_disable {
    my %args   = @_;
    my $resp   = $args{response_object};
    my $output = $resp->get_processed_output();
    _check_error( raw_output => $output, response_object => $resp );
    return [];
} ## end sub system_forensics_disable

## @name ::system_forensics_set_time_limit
##
## @summary Parses the raw output from the "system forensics set-time-limit" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from system_forensics_set_time_limit with any of its options as described in the
##    corresponding cdef file 'system.cdefs'
## @synopsis
##      system_forensics_set_time_limit(%args);
## INPUTS TO METHOD CALL 'system_forensics_set_time_limit' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## <tr> <td> No switch for this key </td> <td> 'time-limit' </td> <td> '0' </td> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an empty array reference.

sub system_forensics_set_time_limit {
    my %args   = @_;
    my $resp   = $args{response_object};
    my $output = $resp->get_processed_output();
    _check_error( raw_output => $output, response_object => $resp );
    return [];
} ## end sub system_forensics_set_time_limit

## @name ::system_forensics_show
##
## @summary Parses the raw output from the "system forensics show" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from system_forensics_show with any of its options as described in the
##    corresponding cdef file 'system.cdefs'
## @synopsis
##      system_forensics_show(%args);
## INPUTS TO METHOD CALL 'system_forensics_show' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

##&nbsp;Command = system forensics show
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;Forensics enabled: false
##&nbsp;Forensics time limit: 1
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;          {
##&nbsp;            'forensics_enabled' => 'false',
##&nbsp;            'forensics_time_limit' => '1'
##&nbsp;          }
##&nbsp;        ];
##&nbsp;

sub system_forensics_show {
    my %args   = @_;
    my $resp   = $args{response_object};
    my $output = $resp->get_processed_output();
    _check_error( raw_output => $output, response_object => $resp );
    my @lines = get_lines( raw_output => $output, response_object => $resp );
    my @result_array;
    my $count = 0;
    foreach my $line (@lines) {
        my @values = split( /:\s+/, $line );
        $result_array[0]->{ to_key_name( item => $values[0] ) } = $values[1];
    }
    return \@result_array;
}

## @name ::system_forensics_log
##
## @summary Parses the raw output from the "system forensics log" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from system_forensics_log with any of its options as described in the
##    corresponding cdef file 'system.cdefs'
## @synopsis
##      system_forensics_log(%args);
## INPUTS TO METHOD CALL 'system_forensics_log' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## <tr> <td> 'dump' </td> <td> 'dump' </td> <td> '1' </td> </tr>
## <tr> <td> 'clear' </td> <td> 'clear' </td> <td> '1' </td> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

##&nbsp;Command = system forensics log dump
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;Context Switched: CPU: 1, Core 1: Thread: 0
##&nbsp;GPR: EAX: 1
##&nbsp;GPR: EBX: 0
##&nbsp;GPR: ECX: 9
##&nbsp;GPR: EDX: ee0a7db8
##&nbsp;GPR: EDI: d772848
##&nbsp;GPR: ESI: ff82973d
##&nbsp;GPR: EBP: 1e88
##&nbsp;GPR: ESP: 1e78
##&nbsp;GPR: EFLAGS: 10296
##&nbsp;GPR: EIP: 842e48c8
##&nbsp;GPR: RAX: 1
##&nbsp;GPR: RBX: 0
##&nbsp;GPR: RCX: ffffffff00000009
##&nbsp;GPR: RDX: ffffff01ee0a7db8
##&nbsp;GPR: RDI: d772848
##&nbsp;GPR: RSI: ff82973d
##&nbsp;GPR: RBP: 1e88
##&nbsp;GPR: RSP: 1e78
##&nbsp;GPR: RFLAGS: 10296
##&nbsp;GPR: RIP: ffffffff842e48c8
##&nbsp;GPR: R8: ffffffff863e4d60
##&nbsp;GPR: R9: 2636a8
##&nbsp;GPR: R10: ffffffff842d45de
##&nbsp;GPR: R11: ffffffff80a57b20
##&nbsp;GPR: R12: 0
##&nbsp;GPR: R13: 0
##&nbsp;GPR: R14: 0
##&nbsp;GPR: R15: 0
##&nbsp;CR: CR0: 80050033
##&nbsp;CR: CR2: 8b00
##&nbsp;CR: CR3: 8b00
##&nbsp;CR: CR4: 8b00
##&nbsp;CR: CR8: 20000001
##&nbsp;DR: DR0: 8b00
##&nbsp;DR: DR1: 8b00
##&nbsp;DR: DR2: 8b00
##&nbsp;DR: DR3: 8b00
##&nbsp;DR: DR6: 8b00
##&nbsp;DR: DR7: 8b00
##&nbsp;SR: CS: base = 0, limit = ffffffff, selector = 0, attr = af9b00
##&nbsp;SR: DS: base = 0, limit = ffffffff, selector = 0, attr = cff300
##&nbsp;SR: SS: base = 0, limit = ffffffff, selector = 0, attr = af9300
##&nbsp;SR: ES: base = 0, limit = ffffffff, selector = 0, attr = cff300
##&nbsp;SR: FS: base = fddd60, limit = ffffffff, selector = fddd60, attr = cff300
##&nbsp;SR: GS: base = 80b2da40, limit = ffffffff, selector = 80b2da40, attr = cff300
##&nbsp;DTR: GDTR: base = 80b2abc0, limit = 47, selector = 80b2abc0, attr = 8200
##&nbsp;DTR: LDTR: base = 0, limit = ffff, selector = 0, attr = 8200
##&nbsp;DTR: IDTR: base = 80b74120, limit = fff, selector = 80b74120, attr = 8200
##&nbsp;DTR: TR: base = 80b2be20, limit = 67, selector = 80b2be20, attr = 8b00
##&nbsp;Context Switched: CPU: 1, Core 1: Thread: 1
##&nbsp;GPR: EAX: 0
##&nbsp;GPR: EBX: 0
##&nbsp;GPR: ECX: 888c0a80
##&nbsp;GPR: EDX: 70
##&nbsp;GPR: EDI: 0
##&nbsp;GPR: ESI: c0a
##&nbsp;GPR: EBP: 3f58
##&nbsp;GPR: ESP: 3f08
##&nbsp;GPR: EFLAGS: 10282
##&nbsp;GPR: EIP: 863e4d79
##&nbsp;GPR: RAX: 0
##&nbsp;GPR: RBX: 0
##&nbsp;GPR: RCX: ffffffff888c0a80
##&nbsp;GPR: RDX: 70
##&nbsp;GPR: RDI: 0
##&nbsp;GPR: RSI: c0a
##&nbsp;GPR: RBP: 3f58
##&nbsp;GPR: RSP: 3f08
##&nbsp;GPR: RFLAGS: 10282
##&nbsp;GPR: RIP: ffffffff863e4d79
##&nbsp;GPR: R8: ffffffff863e4d60
##&nbsp;GPR: R9: a
##&nbsp;GPR: R10: ffffffff8659ed04
##&nbsp;GPR: R11: ffffffff80a57b20
##&nbsp;GPR: R12: 0
##&nbsp;GPR: R13: 0
##&nbsp;GPR: R14: 0
##&nbsp;GPR: R15: 0
##&nbsp;CR: CR0: 80050033
##&nbsp;CR: CR2: 8b00
##&nbsp;CR: CR3: 8b00
##&nbsp;CR: CR4: 8b00
##&nbsp;CR: CR8: 20000000
##&nbsp;DR: DR0: 8b00
##&nbsp;DR: DR1: 8b00
##&nbsp;DR: DR2: 8b00
##&nbsp;DR: DR3: 8b00
##&nbsp;DR: DR6: 8b00
##&nbsp;DR: DR7: 8b00
##&nbsp;SR: CS: base = 0, limit = ffffffff, selector = 0, attr = af9b00
##&nbsp;SR: DS: base = 0, limit = ffffffff, selector = 0, attr = cff300
##&nbsp;SR: SS: base = 0, limit = ffffffff, selector = 0, attr = af9300
##&nbsp;SR: ES: base = 0, limit = ffffffff, selector = 0, attr = cff300
##&nbsp;SR: FS: base = 63e2e0, limit = ffffffff, selector = 63e2e0, attr = cff300
##&nbsp;SR: GS: base = 80b2ebd8, limit = ffffffff, selector = 80b2ebd8, attr = cff300
##&nbsp;DTR: GDTR: base = 80b2ac08, limit = 47, selector = 80b2ac08, attr = 8200
##&nbsp;DTR: LDTR: base = 0, limit = ffff, selector = 0, attr = 8200
##&nbsp;DTR: IDTR: base = 80b74120, limit = fff, selector = 80b74120, attr = 8200
##&nbsp;DTR: TR: base = 80b2be88, limit = 67, selector = 80b2be88, attr = 8b00
##&nbsp;Context Switched: CPU: 1, Core 2: Thread: 0
##&nbsp;GPR: EAX: 805458b0
##&nbsp;GPR: EBX: 0
##&nbsp;GPR: ECX: 80546290
##&nbsp;GPR: EDX: d772848
##&nbsp;GPR: EDI: d772848
##&nbsp;GPR: ESI: d772848
##&nbsp;GPR: EBP: 5e18
##&nbsp;GPR: ESP: 5dd8
##&nbsp;GPR: EFLAGS: 10207
##&nbsp;GPR: EIP: 842e4582
##&nbsp;GPR: RAX: fffffffe805458b0
##&nbsp;GPR: RBX: 0
##&nbsp;GPR: RCX: ffffffff80546290
##&nbsp;GPR: RDX: ffffff020d772848
##&nbsp;GPR: RDI: d772848
##&nbsp;GPR: RSI: d772848
##&nbsp;GPR: RBP: 5e18
##&nbsp;GPR: RSP: 5dd8
##&nbsp;GPR: RFLAGS: 10207
##&nbsp;GPR: RIP: ffffffff842e4582
##&nbsp;GPR: R8: ffffffff863e4d60
##&nbsp;GPR: R9: 2636a8
##&nbsp;GPR: R10: ffffffff8659ed04
##&nbsp;GPR: R11: ffffffff80a57b20
##&nbsp;GPR: R12: 0
##&nbsp;GPR: R13: 0
##&nbsp;GPR: R14: 0
##&nbsp;GPR: R15: 0
##&nbsp;CR: CR0: 80050033
##&nbsp;CR: CR2: 8b00
##&nbsp;CR: CR3: 8b00
##&nbsp;CR: CR4: 8b00
##&nbsp;CR: CR8: a05458b0
##&nbsp;DR: DR0: 8b00
##&nbsp;DR: DR1: 8b00
##&nbsp;DR: DR2: 8b00
##&nbsp;DR: DR3: 8b00
##&nbsp;DR: DR6: 8b00
##&nbsp;DR: DR7: 8b00
##&nbsp;SR: CS: base = 0, limit = ffffffff, selector = 0, attr = af9b00
##&nbsp;SR: DS: base = 0, limit = ffffffff, selector = 0, attr = cff300
##&nbsp;SR: SS: base = 0, limit = ffffffff, selector = 0, attr = af9300
##&nbsp;SR: ES: base = 0, limit = ffffffff, selector = 0, attr = cff300
##&nbsp;SR: FS: base = fddd60, limit = ffffffff, selector = fddd60, attr = cff300
##&nbsp;SR: GS: base = 80b2fd70, limit = ffffffff, selector = 80b2fd70, attr = cff300
##&nbsp;DTR: GDTR: base = 80b2ac50, limit = 47, selector = 80b2ac50, attr = 8200
##&nbsp;DTR: LDTR: base = 0, limit = ffff, selector = 0, attr = 8200
##&nbsp;DTR: IDTR: base = 80b74120, limit = fff, selector = 80b74120, attr = 8200
##&nbsp;DTR: TR: base = 80b2bef0, limit = 67, selector = 80b2bef0, attr = 8b00
##&nbsp;Context Switched: CPU: 1, Core 2: Thread: 1
##&nbsp;GPR: EAX: ff
##&nbsp;GPR: EBX: 0
##&nbsp;GPR: ECX: 18
##&nbsp;GPR: EDX: 8
##&nbsp;GPR: EDI: ecfccb8
##&nbsp;GPR: ESI: 187
##&nbsp;GPR: EBP: 7ea8
##&nbsp;GPR: ESP: 7e78
##&nbsp;GPR: EFLAGS: 10217
##&nbsp;GPR: EIP: 863eb82a
##&nbsp;GPR: RAX: ff
##&nbsp;GPR: RBX: 0
##&nbsp;GPR: RCX: ffffffff00000018
##&nbsp;GPR: RDX: 8
##&nbsp;GPR: RDI: ecfccb8
##&nbsp;GPR: RSI: 187
##&nbsp;GPR: RBP: 7ea8
##&nbsp;GPR: RSP: 7e78
##&nbsp;GPR: RFLAGS: 10217
##&nbsp;GPR: RIP: ffffffff863eb82a
##&nbsp;GPR: R8: ffffffff863e4d60
##&nbsp;GPR: R9: 1e
##&nbsp;GPR: R10: ffffffff840fbf02
##&nbsp;GPR: R11: ffffffff88735780
##&nbsp;GPR: R12: 0
##&nbsp;GPR: R13: 0
##&nbsp;GPR: R14: 0
##&nbsp;GPR: R15: 0
##&nbsp;CR: CR0: 80050033
##&nbsp;CR: CR2: 8b00
##&nbsp;CR: CR3: 8b00
##&nbsp;CR: CR4: 8b00
##&nbsp;CR: CR8: 200000ff
##&nbsp;DR: DR0: 8b00
##&nbsp;DR: DR1: 8b00
##&nbsp;DR: DR2: 8b00
##&nbsp;DR: DR3: 8b00
##&nbsp;DR: DR6: 8b00
##&nbsp;DR: DR7: 8b00
##&nbsp;SR: CS: base = 0, limit = ffffffff, selector = 0, attr = af9b00
##&nbsp;SR: DS: base = 0, limit = ffffffff, selector = 0, attr = cff300
##&nbsp;SR: SS: base = 0, limit = ffffffff, selector = 0, attr = af9300
##&nbsp;SR: ES: base = 0, limit = ffffffff, selector = 0, attr = cff300
##&nbsp;SR: FS: base = fddd60, limit = ffffffff, selector = fddd60, attr = cff300
##&nbsp;SR: GS: base = 80b30f08, limit = ffffffff, selector = 80b30f08, attr = cff300
##&nbsp;DTR: GDTR: base = 80b2ac98, limit = 47, selector = 80b2ac98, attr = 8200
##&nbsp;DTR: LDTR: base = 0, limit = ffff, selector = 0, attr = 8200
##&nbsp;DTR: IDTR: base = 80b74120, limit = fff, selector = 80b74120, attr = 8200
##&nbsp;DTR: TR: base = 80b2bf58, limit = 67, selector = 80b2bf58, attr = 8b00
##&nbsp;MSR (10): 9c705dec90d
##&nbsp;MSR (420): 7ff
##&nbsp;MSR (421): 0
##&nbsp;MSR (422): 0
##&nbsp;MSR (423): 0
##&nbsp;IO(128): ba
##&nbsp;PCI Config Dump (0/8/2/776: 1):
##&nbsp;	0000000000000308: 00000000
##&nbsp;PCI Config Dump (0/10/1/0: 128):
##&nbsp;	0000000000000000: 37118086
##&nbsp;	0000000000000004: 00100006
##&nbsp;	0000000000000008: 08800010
##&nbsp;	000000000000000c: 00800000
##&nbsp;	0000000000000010: fe004004
##&nbsp;	0000000000000014: 00000000
##&nbsp;	0000000000000018: 00000000
##&nbsp;	000000000000001c: 00000000
##&nbsp;	0000000000000020: 00000000
##&nbsp;	0000000000000024: 00000000
##&nbsp;	0000000000000028: 00000000
##&nbsp;	000000000000002c: 00008086
##&nbsp;	0000000000000030: 00000000
##&nbsp;	0000000000000034: 00000080
##&nbsp;	0000000000000038: 00000000
##&nbsp;	000000000000003c: 000002ff
##&nbsp;	0000000000000040: 00000000
##&nbsp;	0000000000000044: 00000000
##&nbsp;	0000000000000048: 00000000
##&nbsp;	000000000000004c: 00000000
##&nbsp;	0000000000000050: 00000000
##&nbsp;	0000000000000054: 00000000
##&nbsp;	0000000000000058: 00000000
##&nbsp;	000000000000005c: 00000000
##&nbsp;	0000000000000060: 00000000
##&nbsp;	0000000000000064: 00000000
##&nbsp;	0000000000000068: 00000000
##&nbsp;	000000000000006c: 00000000
##&nbsp;	0000000000000070: 00000000
##&nbsp;	0000000000000074: 00000000
##&nbsp;	0000000000000078: 00000000
##&nbsp;	000000000000007c: 00000000
##&nbsp;	0000000000000080: 00009011
##&nbsp;	0000000000000084: 00002000
##&nbsp;	0000000000000088: 00003000
##&nbsp;	000000000000008c: 00000000
##&nbsp;	0000000000000090: 0092e010
##&nbsp;	0000000000000094: 00008000
##&nbsp;	0000000000000098: 00000000
##&nbsp;	000000000000009c: 00000000
##&nbsp;	00000000000000a0: 00000000
##&nbsp;	00000000000000a4: 00000000
##&nbsp;	00000000000000a8: 00000000
##&nbsp;	00000000000000ac: 00000000
##&nbsp;	00000000000000b0: 00000000
##&nbsp;	00000000000000b4: 00000010
##&nbsp;	00000000000000b8: 00000000
##&nbsp;	00000000000000bc: 00000000
##&nbsp;	00000000000000c0: 00000000
##&nbsp;	00000000000000c4: 00000000
##&nbsp;	00000000000000c8: 00000000
##&nbsp;	00000000000000cc: 00000000
##&nbsp;	00000000000000d0: 00000000
##&nbsp;	00000000000000d4: 00000000
##&nbsp;	00000000000000d8: 00000000
##&nbsp;	00000000000000dc: 00000000
##&nbsp;	00000000000000e0: 00030001
##&nbsp;	00000000000000e4: 00000008
##&nbsp;	00000000000000e8: 00000000
##&nbsp;	00000000000000ec: 00000000
##&nbsp;	00000000000000f0: 00000000
##&nbsp;	00000000000000f4: 00000000
##&nbsp;	00000000000000f8: 00000000
##&nbsp;	00000000000000fc: 00000000
##&nbsp;	0000000000000100: 00000000
##&nbsp;	0000000000000104: 00000000
##&nbsp;	0000000000000108: 00000000
##&nbsp;	000000000000010c: 00000000
##&nbsp;	0000000000000110: 00000000
##&nbsp;	0000000000000114: 00000000
##&nbsp;	0000000000000118: 00000000
##&nbsp;	000000000000011c: 00000000
##&nbsp;	0000000000000120: 00000000
##&nbsp;	0000000000000124: 00000000
##&nbsp;	0000000000000128: 00000000
##&nbsp;	000000000000012c: 00000000
##&nbsp;	0000000000000130: 00000000
##&nbsp;	0000000000000134: 00000000
##&nbsp;	0000000000000138: 00000000
##&nbsp;	000000000000013c: 00000000
##&nbsp;	0000000000000140: 00000000
##&nbsp;	0000000000000144: 00000000
##&nbsp;	0000000000000148: 00000000
##&nbsp;	000000000000014c: 00000000
##&nbsp;	0000000000000150: 00000000
##&nbsp;	0000000000000154: 00000000
##&nbsp;	0000000000000158: 00000000
##&nbsp;	000000000000015c: 00000000
##&nbsp;	0000000000000160: 00000000
##&nbsp;	0000000000000164: 00000000
##&nbsp;	0000000000000168: 00000000
##&nbsp;	000000000000016c: 00000000
##&nbsp;	0000000000000170: 00000000
##&nbsp;	0000000000000174: 00000000
##&nbsp;	0000000000000178: 00000000
##&nbsp;	000000000000017c: 00000000
##&nbsp;	0000000000000180: 00000000
##&nbsp;	0000000000000184: 00003e07
##&nbsp;	0000000000000188: 00000000
##&nbsp;	000000000000018c: 00000000
##&nbsp;	0000000000000190: 00000000
##&nbsp;	0000000000000194: 00000000
##&nbsp;	0000000000000198: 00000000
##&nbsp;	000000000000019c: 00000000
##&nbsp;	00000000000001a0: 00000000
##&nbsp;	00000000000001a4: 00000000
##&nbsp;	00000000000001a8: 00000000
##&nbsp;	00000000000001ac: 00000000
##&nbsp;	00000000000001b0: 00000000
##&nbsp;	00000000000001b4: 00000000
##&nbsp;	00000000000001b8: 00000000
##&nbsp;	00000000000001bc: 00000000
##&nbsp;	00000000000001c0: 00000000
##&nbsp;	00000000000001c4: 00000000
##&nbsp;	00000000000001c8: 00000000
##&nbsp;	00000000000001cc: 00000000
##&nbsp;	00000000000001d0: 00000000
##&nbsp;	00000000000001d4: 00000000
##&nbsp;	00000000000001d8: 00000000
##&nbsp;	00000000000001dc: 00000000
##&nbsp;	00000000000001e0: 00000000
##&nbsp;	00000000000001e4: 00000000
##&nbsp;	00000000000001e8: 00000000
##&nbsp;	00000000000001ec: 00000000
##&nbsp;	00000000000001f0: 00000000
##&nbsp;	00000000000001f4: 00000000
##&nbsp;	00000000000001f8: 00000000
##&nbsp;	00000000000001fc: 00000000
##&nbsp;End PCI Config Dump
##&nbsp;
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;          {
##&nbsp;            'context_switched4' => {
##&nbsp;                                     'dumped_values' => {
##&nbsp;                                                          'cr' => {
##&nbsp;                                                                    'cr3' => '8b00',
##&nbsp;                                                                    'cr2' => '8b00',
##&nbsp;                                                                    'cr8' => '200000ff',
##&nbsp;                                                                    'cr0' => 80050033,
##&nbsp;                                                                    'cr4' => '8b00'
##&nbsp;                                                                  },
##&nbsp;                                                          'dtr' => {
##&nbsp;                                                                     'tr' => {
##&nbsp;                                                                               'base' => '80b2bf58',
##&nbsp;                                                                               'selector' => '80b2bf58',
##&nbsp;                                                                               'limit' => '67',
##&nbsp;                                                                               'attr' => '8b00'
##&nbsp;                                                                             },
##&nbsp;                                                                     'ldtr' => {
##&nbsp;                                                                                 'base' => '0',
##&nbsp;                                                                                 'selector' => '0',
##&nbsp;                                                                                 'limit' => 'ffff',
##&nbsp;                                                                                 'attr' => '8200'
##&nbsp;                                                                               },
##&nbsp;                                                                     'gdtr' => {
##&nbsp;                                                                                 'base' => '80b2ac98',
##&nbsp;                                                                                 'selector' => '80b2ac98',
##&nbsp;                                                                                 'limit' => '47',
##&nbsp;                                                                                 'attr' => '8200'
##&nbsp;                                                                               },
##&nbsp;                                                                     'idtr' => {
##&nbsp;                                                                                 'base' => '80b74120',
##&nbsp;                                                                                 'selector' => '80b74120',
##&nbsp;                                                                                 'limit' => 'fff',
##&nbsp;                                                                                 'attr' => '8200'
##&nbsp;                                                                               }
##&nbsp;                                                                   },
##&nbsp;                                                          'dr' => {
##&nbsp;                                                                    'dr3' => '8b00',
##&nbsp;                                                                    'dr0' => '8b00',
##&nbsp;                                                                    'dr2' => '8b00',
##&nbsp;                                                                    'dr7' => '8b00',
##&nbsp;                                                                    'dr1' => '8b00',
##&nbsp;                                                                    'dr6' => '8b00'
##&nbsp;                                                                  },
##&nbsp;                                                          'gpr' => {
##&nbsp;                                                                     'eflags' => 10217,
##&nbsp;                                                                     'ebp' => '7ea8',
##&nbsp;                                                                     'rdx' => 8,
##&nbsp;                                                                     'r9' => '1e',
##&nbsp;                                                                     'esp' => '7e78',
##&nbsp;                                                                     'ebx' => 0,
##&nbsp;                                                                     'r8' => 'ffffffff863e4d60',
##&nbsp;                                                                     'edi' => 'ecfccb8',
##&nbsp;                                                                     'rflags' => 10217,
##&nbsp;                                                                     'r11' => 'ffffffff88735780',
##&nbsp;                                                                     'rsi' => 187,
##&nbsp;                                                                     'eip' => '863eb82a',
##&nbsp;                                                                     'rax' => 'ff',
##&nbsp;                                                                     'r12' => 0,
##&nbsp;                                                                     'r13' => 0,
##&nbsp;                                                                     'rdi' => 'ecfccb8',
##&nbsp;                                                                     'r14' => 0,
##&nbsp;                                                                     'rcx' => 'ffffffff00000018',
##&nbsp;                                                                     'rip' => 'ffffffff863eb82a',
##&nbsp;                                                                     'edx' => 8,
##&nbsp;                                                                     'rbx' => 0,
##&nbsp;                                                                     'esi' => 187,
##&nbsp;                                                                     'r15' => 0,
##&nbsp;                                                                     'rsp' => '7e78',
##&nbsp;                                                                     'ecx' => 18,
##&nbsp;                                                                     'rbp' => '7ea8',
##&nbsp;                                                                     'r10' => 'ffffffff840fbf02',
##&nbsp;                                                                     'eax' => 'ff'
##&nbsp;                                                                   },
##&nbsp;                                                          'sr' => {
##&nbsp;                                                                    'gs' => {
##&nbsp;                                                                              'base' => '80b30f08',
##&nbsp;                                                                              'selector' => '80b30f08',
##&nbsp;                                                                              'limit' => 'ffffffff',
##&nbsp;                                                                              'attr' => 'cff300'
##&nbsp;                                                                            },
##&nbsp;                                                                    'ss' => {
##&nbsp;                                                                              'base' => '0',
##&nbsp;                                                                              'selector' => '0',
##&nbsp;                                                                              'limit' => 'ffffffff',
##&nbsp;                                                                              'attr' => 'af9300'
##&nbsp;                                                                            },
##&nbsp;                                                                    'es' => {
##&nbsp;                                                                              'base' => '0',
##&nbsp;                                                                              'selector' => '0',
##&nbsp;                                                                              'limit' => 'ffffffff',
##&nbsp;                                                                              'attr' => 'cff300'
##&nbsp;                                                                            },
##&nbsp;                                                                    'cs' => {
##&nbsp;                                                                              'base' => '0',
##&nbsp;                                                                              'selector' => '0',
##&nbsp;                                                                              'limit' => 'ffffffff',
##&nbsp;                                                                              'attr' => 'af9b00'
##&nbsp;                                                                            },
##&nbsp;                                                                    'fs' => {
##&nbsp;                                                                              'base' => 'fddd60',
##&nbsp;                                                                              'selector' => 'fddd60',
##&nbsp;                                                                              'limit' => 'ffffffff',
##&nbsp;                                                                              'attr' => 'cff300'
##&nbsp;                                                                            },
##&nbsp;                                                                    'ds' => {
##&nbsp;                                                                              'base' => '0',
##&nbsp;                                                                              'selector' => '0',
##&nbsp;                                                                              'limit' => 'ffffffff',
##&nbsp;                                                                              'attr' => 'cff300'
##&nbsp;                                                                            }
##&nbsp;                                                                  }
##&nbsp;                                                        },
##&nbsp;                                     'Thread' => '1',
##&nbsp;                                     'CPU' => '1',
##&nbsp;                                     'Core' => '2'
##&nbsp;                                   },
##&nbsp;            'context_switched1' => {
##&nbsp;                                     'dumped_values' => {
##&nbsp;                                                          'cr' => {
##&nbsp;                                                                    'cr3' => '8b00',
##&nbsp;                                                                    'cr2' => '8b00',
##&nbsp;                                                                    'cr8' => 20000001,
##&nbsp;                                                                    'cr0' => 80050033,
##&nbsp;                                                                    'cr4' => '8b00'
##&nbsp;                                                                  },
##&nbsp;                                                          'dtr' => {
##&nbsp;                                                                     'tr' => {
##&nbsp;                                                                               'base' => '80b2be20',
##&nbsp;                                                                               'selector' => '80b2be20',
##&nbsp;                                                                               'limit' => '67',
##&nbsp;                                                                               'attr' => '8b00'
##&nbsp;                                                                             },
##&nbsp;                                                                     'ldtr' => {
##&nbsp;                                                                                 'base' => '0',
##&nbsp;                                                                                 'selector' => '0',
##&nbsp;                                                                                 'limit' => 'ffff',
##&nbsp;                                                                                 'attr' => '8200'
##&nbsp;                                                                               },
##&nbsp;                                                                     'gdtr' => {
##&nbsp;                                                                                 'base' => '80b2abc0',
##&nbsp;                                                                                 'selector' => '80b2abc0',
##&nbsp;                                                                                 'limit' => '47',
##&nbsp;                                                                                 'attr' => '8200'
##&nbsp;                                                                               },
##&nbsp;                                                                     'idtr' => {
##&nbsp;                                                                                 'base' => '80b74120',
##&nbsp;                                                                                 'selector' => '80b74120',
##&nbsp;                                                                                 'limit' => 'fff',
##&nbsp;                                                                                 'attr' => '8200'
##&nbsp;                                                                               }
##&nbsp;                                                                   },
##&nbsp;                                                          'dr' => {
##&nbsp;                                                                    'dr3' => '8b00',
##&nbsp;                                                                    'dr0' => '8b00',
##&nbsp;                                                                    'dr2' => '8b00',
##&nbsp;                                                                    'dr7' => '8b00',
##&nbsp;                                                                    'dr1' => '8b00',
##&nbsp;                                                                    'dr6' => '8b00'
##&nbsp;                                                                  },
##&nbsp;                                                          'gpr' => {
##&nbsp;                                                                     'eflags' => 10296,
##&nbsp;                                                                     'ebp' => '1e88',
##&nbsp;                                                                     'rdx' => 'ffffff01ee0a7db8',
##&nbsp;                                                                     'r9' => '2636a8',
##&nbsp;                                                                     'esp' => '1e78',
##&nbsp;                                                                     'ebx' => 0,
##&nbsp;                                                                     'r8' => 'ffffffff863e4d60',
##&nbsp;                                                                     'edi' => 'd772848',
##&nbsp;                                                                     'rflags' => 10296,
##&nbsp;                                                                     'r11' => 'ffffffff80a57b20',
##&nbsp;                                                                     'rsi' => 'ff82973d',
##&nbsp;                                                                     'eip' => '842e48c8',
##&nbsp;                                                                     'rax' => 1,
##&nbsp;                                                                     'r12' => 0,
##&nbsp;                                                                     'r13' => 0,
##&nbsp;                                                                     'rdi' => 'd772848',
##&nbsp;                                                                     'r14' => 0,
##&nbsp;                                                                     'rcx' => 'ffffffff00000009',
##&nbsp;                                                                     'rip' => 'ffffffff842e48c8',
##&nbsp;                                                                     'edx' => 'ee0a7db8',
##&nbsp;                                                                     'rbx' => 0,
##&nbsp;                                                                     'esi' => 'ff82973d',
##&nbsp;                                                                     'r15' => 0,
##&nbsp;                                                                     'rsp' => '1e78',
##&nbsp;                                                                     'ecx' => 9,
##&nbsp;                                                                     'rbp' => '1e88',
##&nbsp;                                                                     'r10' => 'ffffffff842d45de',
##&nbsp;                                                                     'eax' => 1
##&nbsp;                                                                   },
##&nbsp;                                                          'sr' => {
##&nbsp;                                                                    'gs' => {
##&nbsp;                                                                              'base' => '80b2da40',
##&nbsp;                                                                              'selector' => '80b2da40',
##&nbsp;                                                                              'limit' => 'ffffffff',
##&nbsp;                                                                              'attr' => 'cff300'
##&nbsp;                                                                            },
##&nbsp;                                                                    'ss' => {
##&nbsp;                                                                              'base' => '0',
##&nbsp;                                                                              'selector' => '0',
##&nbsp;                                                                              'limit' => 'ffffffff',
##&nbsp;                                                                              'attr' => 'af9300'
##&nbsp;                                                                            },
##&nbsp;                                                                    'es' => {
##&nbsp;                                                                              'base' => '0',
##&nbsp;                                                                              'selector' => '0',
##&nbsp;                                                                              'limit' => 'ffffffff',
##&nbsp;                                                                              'attr' => 'cff300'
##&nbsp;                                                                            },
##&nbsp;                                                                    'cs' => {
##&nbsp;                                                                              'base' => '0',
##&nbsp;                                                                              'selector' => '0',
##&nbsp;                                                                              'limit' => 'ffffffff',
##&nbsp;                                                                              'attr' => 'af9b00'
##&nbsp;                                                                            },
##&nbsp;                                                                    'fs' => {
##&nbsp;                                                                              'base' => 'fddd60',
##&nbsp;                                                                              'selector' => 'fddd60',
##&nbsp;                                                                              'limit' => 'ffffffff',
##&nbsp;                                                                              'attr' => 'cff300'
##&nbsp;                                                                            },
##&nbsp;                                                                    'ds' => {
##&nbsp;                                                                              'base' => '0',
##&nbsp;                                                                              'selector' => '0',
##&nbsp;                                                                              'limit' => 'ffffffff',
##&nbsp;                                                                              'attr' => 'cff300'
##&nbsp;                                                                            }
##&nbsp;                                                                  }
##&nbsp;                                                        },
##&nbsp;                                     'Thread' => '0',
##&nbsp;                                     'CPU' => '1',
##&nbsp;                                     'Core' => '1'
##&nbsp;                                   },
##&nbsp;            'context_switched2' => {
##&nbsp;                                     'dumped_values' => {
##&nbsp;                                                          'cr' => {
##&nbsp;                                                                    'cr3' => '8b00',
##&nbsp;                                                                    'cr2' => '8b00',
##&nbsp;                                                                    'cr8' => 20000000,
##&nbsp;                                                                    'cr0' => 80050033,
##&nbsp;                                                                    'cr4' => '8b00'
##&nbsp;                                                                  },
##&nbsp;                                                          'dtr' => {
##&nbsp;                                                                     'tr' => {
##&nbsp;                                                                               'base' => '80b2be88',
##&nbsp;                                                                               'selector' => '80b2be88',
##&nbsp;                                                                               'limit' => '67',
##&nbsp;                                                                               'attr' => '8b00'
##&nbsp;                                                                             },
##&nbsp;                                                                     'ldtr' => {
##&nbsp;                                                                                 'base' => '0',
##&nbsp;                                                                                 'selector' => '0',
##&nbsp;                                                                                 'limit' => 'ffff',
##&nbsp;                                                                                 'attr' => '8200'
##&nbsp;                                                                               },
##&nbsp;                                                                     'gdtr' => {
##&nbsp;                                                                                 'base' => '80b2ac08',
##&nbsp;                                                                                 'selector' => '80b2ac08',
##&nbsp;                                                                                 'limit' => '47',
##&nbsp;                                                                                 'attr' => '8200'
##&nbsp;                                                                               },
##&nbsp;                                                                     'idtr' => {
##&nbsp;                                                                                 'base' => '80b74120',
##&nbsp;                                                                                 'selector' => '80b74120',
##&nbsp;                                                                                 'limit' => 'fff',
##&nbsp;                                                                                 'attr' => '8200'
##&nbsp;                                                                               }
##&nbsp;                                                                   },
##&nbsp;                                                          'dr' => {
##&nbsp;                                                                    'dr3' => '8b00',
##&nbsp;                                                                    'dr0' => '8b00',
##&nbsp;                                                                    'dr2' => '8b00',
##&nbsp;                                                                    'dr7' => '8b00',
##&nbsp;                                                                    'dr1' => '8b00',
##&nbsp;                                                                    'dr6' => '8b00'
##&nbsp;                                                                  },
##&nbsp;                                                          'gpr' => {
##&nbsp;                                                                     'eflags' => 10282,
##&nbsp;                                                                     'ebp' => '3f58',
##&nbsp;                                                                     'rdx' => 70,
##&nbsp;                                                                     'r9' => 'a',
##&nbsp;                                                                     'esp' => '3f08',
##&nbsp;                                                                     'ebx' => 0,
##&nbsp;                                                                     'r8' => 'ffffffff863e4d60',
##&nbsp;                                                                     'edi' => 0,
##&nbsp;                                                                     'rflags' => 10282,
##&nbsp;                                                                     'r11' => 'ffffffff80a57b20',
##&nbsp;                                                                     'rsi' => 'c0a',
##&nbsp;                                                                     'eip' => '863e4d79',
##&nbsp;                                                                     'rax' => 0,
##&nbsp;                                                                     'r12' => 0,
##&nbsp;                                                                     'r13' => 0,
##&nbsp;                                                                     'rdi' => 0,
##&nbsp;                                                                     'r14' => 0,
##&nbsp;                                                                     'rcx' => 'ffffffff888c0a80',
##&nbsp;                                                                     'rip' => 'ffffffff863e4d79',
##&nbsp;                                                                     'edx' => 70,
##&nbsp;                                                                     'rbx' => 0,
##&nbsp;                                                                     'esi' => 'c0a',
##&nbsp;                                                                     'r15' => 0,
##&nbsp;                                                                     'rsp' => '3f08',
##&nbsp;                                                                     'ecx' => '888c0a80',
##&nbsp;                                                                     'rbp' => '3f58',
##&nbsp;                                                                     'r10' => 'ffffffff8659ed04',
##&nbsp;                                                                     'eax' => 0
##&nbsp;                                                                   },
##&nbsp;                                                          'sr' => {
##&nbsp;                                                                    'gs' => {
##&nbsp;                                                                              'base' => '80b2ebd8',
##&nbsp;                                                                              'selector' => '80b2ebd8',
##&nbsp;                                                                              'limit' => 'ffffffff',
##&nbsp;                                                                              'attr' => 'cff300'
##&nbsp;                                                                            },
##&nbsp;                                                                    'ss' => {
##&nbsp;                                                                              'base' => '0',
##&nbsp;                                                                              'selector' => '0',
##&nbsp;                                                                              'limit' => 'ffffffff',
##&nbsp;                                                                              'attr' => 'af9300'
##&nbsp;                                                                            },
##&nbsp;                                                                    'es' => {
##&nbsp;                                                                              'base' => '0',
##&nbsp;                                                                              'selector' => '0',
##&nbsp;                                                                              'limit' => 'ffffffff',
##&nbsp;                                                                              'attr' => 'cff300'
##&nbsp;                                                                            },
##&nbsp;                                                                    'cs' => {
##&nbsp;                                                                              'base' => '0',
##&nbsp;                                                                              'selector' => '0',
##&nbsp;                                                                              'limit' => 'ffffffff',
##&nbsp;                                                                              'attr' => 'af9b00'
##&nbsp;                                                                            },
##&nbsp;                                                                    'fs' => {
##&nbsp;                                                                              'base' => '63e2e0',
##&nbsp;                                                                              'selector' => '63e2e0',
##&nbsp;                                                                              'limit' => 'ffffffff',
##&nbsp;                                                                              'attr' => 'cff300'
##&nbsp;                                                                            },
##&nbsp;                                                                    'ds' => {
##&nbsp;                                                                              'base' => '0',
##&nbsp;                                                                              'selector' => '0',
##&nbsp;                                                                              'limit' => 'ffffffff',
##&nbsp;                                                                              'attr' => 'cff300'
##&nbsp;                                                                            }
##&nbsp;                                                                  }
##&nbsp;                                                        },
##&nbsp;                                     'Thread' => '1',
##&nbsp;                                     'CPU' => '1',
##&nbsp;                                     'Core' => '1'
##&nbsp;                                   },
##&nbsp;            'context_switched3' => {
##&nbsp;                                     'dumped_values' => {
##&nbsp;                                                          'cr' => {
##&nbsp;                                                                    'cr3' => '8b00',
##&nbsp;                                                                    'cr2' => '8b00',
##&nbsp;                                                                    'cr8' => 'a05458b0',
##&nbsp;                                                                    'cr0' => 80050033,
##&nbsp;                                                                    'cr4' => '8b00'
##&nbsp;                                                                  },
##&nbsp;                                                          'dtr' => {
##&nbsp;                                                                     'tr' => {
##&nbsp;                                                                               'base' => '80b2bef0',
##&nbsp;                                                                               'selector' => '80b2bef0',
##&nbsp;                                                                               'limit' => '67',
##&nbsp;                                                                               'attr' => '8b00'
##&nbsp;                                                                             },
##&nbsp;                                                                     'ldtr' => {
##&nbsp;                                                                                 'base' => '0',
##&nbsp;                                                                                 'selector' => '0',
##&nbsp;                                                                                 'limit' => 'ffff',
##&nbsp;                                                                                 'attr' => '8200'
##&nbsp;                                                                               },
##&nbsp;                                                                     'gdtr' => {
##&nbsp;                                                                                 'base' => '80b2ac50',
##&nbsp;                                                                                 'selector' => '80b2ac50',
##&nbsp;                                                                                 'limit' => '47',
##&nbsp;                                                                                 'attr' => '8200'
##&nbsp;                                                                               },
##&nbsp;                                                                     'idtr' => {
##&nbsp;                                                                                 'base' => '80b74120',
##&nbsp;                                                                                 'selector' => '80b74120',
##&nbsp;                                                                                 'limit' => 'fff',
##&nbsp;                                                                                 'attr' => '8200'
##&nbsp;                                                                               }
##&nbsp;                                                                   },
##&nbsp;                                                          'dr' => {
##&nbsp;                                                                    'dr3' => '8b00',
##&nbsp;                                                                    'dr0' => '8b00',
##&nbsp;                                                                    'dr2' => '8b00',
##&nbsp;                                                                    'dr7' => '8b00',
##&nbsp;                                                                    'dr1' => '8b00',
##&nbsp;                                                                    'dr6' => '8b00'
##&nbsp;                                                                  },
##&nbsp;                                                          'gpr' => {
##&nbsp;                                                                     'eflags' => 10207,
##&nbsp;                                                                     'ebp' => '5e18',
##&nbsp;                                                                     'rdx' => 'ffffff020d772848',
##&nbsp;                                                                     'r9' => '2636a8',
##&nbsp;                                                                     'esp' => '5dd8',
##&nbsp;                                                                     'ebx' => 0,
##&nbsp;                                                                     'r8' => 'ffffffff863e4d60',
##&nbsp;                                                                     'edi' => 'd772848',
##&nbsp;                                                                     'rflags' => 10207,
##&nbsp;                                                                     'r11' => 'ffffffff80a57b20',
##&nbsp;                                                                     'rsi' => 'd772848',
##&nbsp;                                                                     'eip' => '842e4582',
##&nbsp;                                                                     'rax' => 'fffffffe805458b0',
##&nbsp;                                                                     'r12' => 0,
##&nbsp;                                                                     'r13' => 0,
##&nbsp;                                                                     'rdi' => 'd772848',
##&nbsp;                                                                     'r14' => 0,
##&nbsp;                                                                     'rcx' => 'ffffffff80546290',
##&nbsp;                                                                     'rip' => 'ffffffff842e4582',
##&nbsp;                                                                     'edx' => 'd772848',
##&nbsp;                                                                     'rbx' => 0,
##&nbsp;                                                                     'esi' => 'd772848',
##&nbsp;                                                                     'r15' => 0,
##&nbsp;                                                                     'rsp' => '5dd8',
##&nbsp;                                                                     'ecx' => 80546290,
##&nbsp;                                                                     'rbp' => '5e18',
##&nbsp;                                                                     'r10' => 'ffffffff8659ed04',
##&nbsp;                                                                     'eax' => '805458b0'
##&nbsp;                                                                   },
##&nbsp;                                                          'sr' => {
##&nbsp;                                                                    'gs' => {
##&nbsp;                                                                              'base' => '80b2fd70',
##&nbsp;                                                                              'selector' => '80b2fd70',
##&nbsp;                                                                              'limit' => 'ffffffff',
##&nbsp;                                                                              'attr' => 'cff300'
##&nbsp;                                                                            },
##&nbsp;                                                                    'ss' => {
##&nbsp;                                                                              'base' => '0',
##&nbsp;                                                                              'selector' => '0',
##&nbsp;                                                                              'limit' => 'ffffffff',
##&nbsp;                                                                              'attr' => 'af9300'
##&nbsp;                                                                            },
##&nbsp;                                                                    'es' => {
##&nbsp;                                                                              'base' => '0',
##&nbsp;                                                                              'selector' => '0',
##&nbsp;                                                                              'limit' => 'ffffffff',
##&nbsp;                                                                              'attr' => 'cff300'
##&nbsp;                                                                            },
##&nbsp;                                                                    'cs' => {
##&nbsp;                                                                              'base' => '0',
##&nbsp;                                                                              'selector' => '0',
##&nbsp;                                                                              'limit' => 'ffffffff',
##&nbsp;                                                                              'attr' => 'af9b00'
##&nbsp;                                                                            },
##&nbsp;                                                                    'fs' => {
##&nbsp;                                                                              'base' => 'fddd60',
##&nbsp;                                                                              'selector' => 'fddd60',
##&nbsp;                                                                              'limit' => 'ffffffff',
##&nbsp;                                                                              'attr' => 'cff300'
##&nbsp;                                                                            },
##&nbsp;                                                                    'ds' => {
##&nbsp;                                                                              'base' => '0',
##&nbsp;                                                                              'selector' => '0',
##&nbsp;                                                                              'limit' => 'ffffffff',
##&nbsp;                                                                              'attr' => 'cff300'
##&nbsp;                                                                            }
##&nbsp;                                                                  }
##&nbsp;                                                        },
##&nbsp;                                     'Thread' => '0',
##&nbsp;                                     'CPU' => '1',
##&nbsp;                                     'Core' => '2'
##&nbsp;                                   }
##&nbsp;          },
##&nbsp;          {
##&nbsp;
##&nbsp;            'pci_config_dump_0_8_2_776_1' => {
##&nbsp;                                               '0000000000000308' => '00000000'
##&nbsp;                                             },
##&nbsp;            'pci_config_dump_0_10_1_0_128' => {
##&nbsp;                                                '000000000000012c' => '00000000',
##&nbsp;                                                '000000000000015c' => '00000000',
##&nbsp;                                                '000000000000004c' => '00000000',
##&nbsp;                                                '00000000000000c8' => '00000000',
##&nbsp;                                                '00000000000001b8' => '00000000',
##&nbsp;                                                '00000000000001cc' => '00000000',
##&nbsp;                                                '00000000000000a0' => '00000000',
##&nbsp;                                                '0000000000000180' => '00000000',
##&nbsp;                                                '0000000000000144' => '00000000',
##&nbsp;                                                '0000000000000058' => '00000000',
##&nbsp;                                                '0000000000000008' => '08800010',
##&nbsp;                                                '0000000000000080' => '00009011',
##&nbsp;                                                '00000000000000f0' => '00000000',
##&nbsp;                                                '0000000000000174' => '00000000',
##&nbsp;                                                '000000000000001c' => '00000000',
##&nbsp;                                                '0000000000000118' => '00000000',
##&nbsp;                                                '0000000000000158' => '00000000',
##&nbsp;                                                '00000000000000d8' => '00000000',
##&nbsp;                                                '00000000000000d0' => '00000000',
##&nbsp;                                                '000000000000014c' => '00000000',
##&nbsp;                                                '00000000000001a8' => '00000000',
##&nbsp;                                                '0000000000000110' => '00000000',
##&nbsp;                                                '00000000000000ec' => '00000000',
##&nbsp;                                                '0000000000000084' => '00002000',
##&nbsp;                                                '0000000000000038' => '00000000',
##&nbsp;                                                '00000000000000fc' => '00000000',
##&nbsp;                                                '0000000000000078' => '00000000',
##&nbsp;                                                '0000000000000034' => '00000080',
##&nbsp;                                                '0000000000000064' => '00000000',
##&nbsp;                                                '0000000000000108' => '00000000',
##&nbsp;                                                '00000000000000f4' => '00000000',
##&nbsp;                                                '00000000000000a4' => '00000000',
##&nbsp;                                                '000000000000005c' => '00000000',
##&nbsp;                                                '00000000000001d4' => '00000000',
##&nbsp;                                                '0000000000000170' => '00000000',
##&nbsp;                                                '0000000000000068' => '00000000',
##&nbsp;                                                '00000000000001b4' => '00000000',
##&nbsp;                                                '0000000000000138' => '00000000',
##&nbsp;                                                '00000000000001d0' => '00000000',
##&nbsp;                                                '00000000000000b8' => '00000000',
##&nbsp;                                                '0000000000000048' => '00000000',
##&nbsp;                                                '0000000000000090' => '0092e010',
##&nbsp;                                                '0000000000000074' => '00000000',
##&nbsp;                                                '00000000000001c8' => '00000000',
##&nbsp;                                                '0000000000000010' => 'fe004004',
##&nbsp;                                                '000000000000011c' => '00000000',
##&nbsp;                                                '0000000000000120' => '00000000',
##&nbsp;                                                '0000000000000028' => '00000000',
##&nbsp;                                                '0000000000000054' => '00000000',
##&nbsp;                                                '000000000000007c' => '00000000',
##&nbsp;                                                '00000000000000e4' => '00000008',
##&nbsp;                                                '0000000000000000' => '37118086',
##&nbsp;                                                '00000000000001a4' => '00000000',
##&nbsp;                                                '000000000000003c' => '000002ff',
##&nbsp;                                                '0000000000000148' => '00000000',
##&nbsp;                                                '00000000000000a8' => '00000000',
##&nbsp;                                                '0000000000000014' => '00000000',
##&nbsp;                                                '00000000000001bc' => '00000000',
##&nbsp;                                                '0000000000000018' => '00000000',
##&nbsp;                                                '00000000000000dc' => '00000000',
##&nbsp;                                                '0000000000000024' => '00000000',
##&nbsp;                                                '00000000000001f8' => '00000000',
##&nbsp;                                                '0000000000000154' => '00000000',
##&nbsp;                                                '0000000000000128' => '00000000',
##&nbsp;                                                '000000000000006c' => '00000000',
##&nbsp;                                                '00000000000000f8' => '00000000',
##&nbsp;                                                '000000000000010c' => '00000000',
##&nbsp;                                                '0000000000000194' => '00000000',
##&nbsp;                                                '0000000000000104' => '00000000',
##&nbsp;                                                '0000000000000184' => '00003e07',
##&nbsp;                                                '00000000000000cc' => '00000000',
##&nbsp;                                                '00000000000001b0' => '00000000',
##&nbsp;                                                '000000000000013c' => '00000000',
##&nbsp;                                                '00000000000000c4' => '00000000',
##&nbsp;                                                '00000000000001e8' => '00000000',
##&nbsp;                                                '000000000000018c' => '00000000',
##&nbsp;                                                '000000000000017c' => '00000000',
##&nbsp;                                                '0000000000000130' => '00000000',
##&nbsp;                                                '00000000000000d4' => '00000000',
##&nbsp;                                                '0000000000000190' => '00000000',
##&nbsp;                                                '0000000000000164' => '00000000',
##&nbsp;                                                '0000000000000198' => '00000000',
##&nbsp;                                                '0000000000000188' => '00000000',
##&nbsp;                                                '000000000000009c' => '00000000',
##&nbsp;                                                '00000000000001dc' => '00000000',
##&nbsp;                                                '00000000000001fc' => '00000000',
##&nbsp;                                                '00000000000001f4' => '00000000',
##&nbsp;                                                '0000000000000044' => '00000000',
##&nbsp;                                                '00000000000001f0' => '00000000',
##&nbsp;                                                '00000000000001ac' => '00000000',
##&nbsp;                                                '00000000000001e0' => '00000000',
##&nbsp;                                                '00000000000001e4' => '00000000',
##&nbsp;                                                '0000000000000020' => '00000000',
##&nbsp;                                                '0000000000000168' => '00000000',
##&nbsp;                                                '00000000000000b4' => '00000010',
##&nbsp;                                                '0000000000000124' => '00000000',
##&nbsp;                                                '0000000000000050' => '00000000',
##&nbsp;                                                '0000000000000070' => '00000000',
##&nbsp;                                                '0000000000000098' => '00000000',
##&nbsp;                                                '0000000000000088' => '00003000',
##&nbsp;                                                '00000000000001c0' => '00000000',
##&nbsp;                                                '000000000000016c' => '00000000',
##&nbsp;                                                '0000000000000140' => '00000000',
##&nbsp;                                                '000000000000008c' => '00000000',
##&nbsp;                                                '0000000000000094' => '00008000',
##&nbsp;                                                '0000000000000150' => '00000000',
##&nbsp;                                                '00000000000000c0' => '00000000',
##&nbsp;                                                '000000000000019c' => '00000000',
##&nbsp;                                                '0000000000000114' => '00000000',
##&nbsp;                                                '00000000000000bc' => '00000000',
##&nbsp;                                                '0000000000000004' => '00100006',
##&nbsp;                                                '00000000000000b0' => '00000000',
##&nbsp;                                                '00000000000000e8' => '00000000',
##&nbsp;                                                '00000000000000e0' => '00030001',
##&nbsp;                                                '00000000000001d8' => '00000000',
##&nbsp;                                                '00000000000001ec' => '00000000',
##&nbsp;                                                '0000000000000040' => '00000000',
##&nbsp;                                                '0000000000000100' => '00000000',
##&nbsp;                                                '0000000000000060' => '00000000',
##&nbsp;                                                '000000000000002c' => '00008086',
##&nbsp;                                                '00000000000001a0' => '00000000',
##&nbsp;                                                '0000000000000178' => '00000000',
##&nbsp;                                                '00000000000001c4' => '00000000',
##&nbsp;                                                '00000000000000ac' => '00000000',
##&nbsp;                                                '0000000000000160' => '00000000',
##&nbsp;                                                '0000000000000134' => '00000000',
##&nbsp;                                                '000000000000000c' => '00800000',
##&nbsp;                                                '0000000000000030' => '00000000'
##&nbsp;                                              },
##&nbsp;          }
##&nbsp;        ];
##&nbsp;
sub system_forensics_log {
    my %args   = @_;
    my $resp   = $args{response_object};
    my $output = $resp->get_processed_output();
    _check_error( raw_output => $output, response_object => $resp );
    my @lines = get_lines( raw_output => $output );
    my @result_array;
    my ( %hash, %hash_arr2, %hash_arr );
    my $line;
    chomp($line);
    my $uniq = 0;
    my $i    = 0;
    unshift( @lines, $line );

    while (@lines) {
        if ( $line =~
/(Context\sSwitched)\s*[:]\s+(\w+)[:]\s+(\d+)[,]*\s+(\w+)\s+(\d+)\s*[:]\s+(\w+)\s*[:]\s+(\d+)/
          )
        {
            my %hash;
            $uniq++;
            my $key = $1 . "$uniq";
            $hash{$2} = $3;
            $hash{$4} = $5;
            $hash{$6} = $7;
            $line     = shift(@lines);
            my @split_arr = ();
            my @temp_arr  = ();
            my $key_val;

            while (!( $line =~ /Context\sSwitched/ )
                && !( $line =~ /PCI\s+Config\s+Dump/ ) )
            {
                if ( $line =~
/\s*(\w+\s*[\(\w*\d+\w*\)]*)[:]\s*(\d*\w*\d*)\s*[:]*\s*(.*)/i
                  )
                {
                    my $value1 = $1;
                    my $value2 = $2;
                    my $value3 = $3;
                    $value1 =~ s/^\s+//;
                    $value1 =~ s/\(/_/g;
                    $value1 =~ s/\)//g;
                    push( @temp_arr, $value1 );
                    if ( $value3 >= 0 ) {
                        @split_arr = split( /[=|,]/, $value3 );
                        $key_val = $value1;
                        if ( grep( $value1, @temp_arr ) ) {
                            $key_val = $value1;
                        }
                        if ( !$#split_arr ) {
                            $hash{'dumped_values'}
                              { to_key_name( item => $key_val ) }
                              { to_key_name( item => $value2 ) } = $value3;
                        }
                        else {
                            for ( my $i = 0 ; $i < $#split_arr ; $i += 2 ) {
                                print "value is : $split_arr[$i] ";
                                $split_arr[$i]       =~ s/^\s+//;
                                $split_arr[$i]       =~ s/\s+$//;
                                $split_arr[$i]       =~ s/_//g;
                                $split_arr[ $i + 1 ] =~ s/^\s+//;
                                $split_arr[ $i + 1 ] =~ s/_//g;
                                $hash{'dumped_values'}
                                  { to_key_name( item => $key_val ) }
                                  { to_key_name( item => $value2 ) }
                                  { $split_arr[$i] } = $split_arr[ $i + 1 ];
                            }
                        }
                    }
                    else {
                        $hash{'dumped_values'}
                          { to_key_name( item => $value1 ) } = $value2;
                    }

                }
                $line = shift(@lines);
            }
            unshift( @lines, $line );
            $hash_arr{ to_key_name( item => $key ) } = \%hash;
        }
        elsif ( $line =~
            /(PCI\s+Config\s+Dump\s+\([\d+[\/]*]*\s*:\s+\d+\s*\))[:].*/ )
        {
            my ( $key, $value, $key_value );
            $key_value = $1;
            $key_value =~ s/^\s+/_/g;
            $key_value =~ s/\s*\(/_/g;
            $key_value =~ s/:\s*/_/g;
            $key_value =~ s/\s*\)//g;
            $key_value =~ s/\//_/g;
            my %hash2;

            while ( !( $line =~ /End\s+PCI\s+Config\s+Dump.*/i ) ) {
                $line = shift(@lines);
                my @split_arr = split( /[:]/, $line );
                for ( my $i = 0 ; $i < $#split_arr ; $i += 2 ) {
                    $key   = $split_arr[$i];
                    $value = $split_arr[ $i + 1 ];
                    $key   =~ s/^\s+//;
                    $key   =~ s/\(/_/g;
                    $key   =~ s/\)//g;
                    $value =~ s/^\s+//;
                    $value =~ s/\s+$//;
                    $value =~ s/\(/_/g;
                    $value =~ s/\)//g;
                    $hash2{ to_key_name( item => $key ) } = $value;
                }
            }
            $hash_arr2{ to_key_name( item => $key_value ) } = \%hash2;
            unshift( @lines, $line );
        }
        elsif ( $line =~ /\s*([\w+\s+]*)[-]\s*([\w+\s+]*)/ ) {
            $hash{ to_key_name( item => $1 ) } = $2;
        }
        elsif ( $line =~ /No\s+[fF]orensics/ ) {
            $hash{'dumped_values'} = "none";
            $hash{'status'}        = "no logs available";
        }
        $line = shift(@lines);
    }
    push( @result_array, \%hash_arr )  if (%hash_arr);
    push( @result_array, \%hash_arr2 ) if (%hash_arr2);
    push( @result_array, \%hash )      if (%hash);
    return \@result_array;
} ## end system_forensics_log

## @name ::system_forensics_config
##
## @summary Parses the raw output from the "system forensics config" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from system_forensics_config with any of its options as described in the
##    corresponding cdef file 'system.cdefs'
## @synopsis
##      system_forensics_config(%args);
## INPUTS TO METHOD CALL 'system_forensics_config' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## <tr> <td> 'dump' </td> <td> 'dump' </td> <td> '1' </td> </tr>
## <tr> <td> No switch for this key </td> <td> 'url' </td> <td> '0' </td> </tr>
## <tr> <td> 'clear' </td> <td> 'clear' </td> <td> '1' </td> </tr>
## <tr> <td> 'update' </td> <td> 'update' </td> <td> '1' </td> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

##&nbsp;Command = system forensics config dump
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;system forensics show - show system forensics configuration settings
##&nbsp;system forensics log - display or clear system forensics logs
##&nbsp;
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;          {
##&nbsp;            'system_forensics_show' => 'show system forensics configuration settings',
##&nbsp;            'system_forensics_log' => 'display or clear system forensics logs'
##&nbsp;          }
##&nbsp;        ];
##&nbsp;
sub system_forensics_config {
    my %args   = @_;
    my $resp   = $args{response_object};
    my $output = $resp->get_processed_output();
    _check_error( raw_output => $output, response_object => $resp );
    my @lines = get_lines( raw_output => $output );
    my %hash;
    my %hash2;
    my @result_array;
    my ( $key, $value, $main_key, $inner_main_key, $outer_main_key,
        $inner_most_key );
    $inner_main_key = 0;
    my $line;

    while (@lines) {
        $line = shift(@lines);
        if ( $line =~ /system\sforensics/i ) {
            my @split_arr = split( "-", $line );
            for ( my $i = 0 ; $i < $#split_arr ; $i += 2 ) {
                $key   = $split_arr[$i];
                $value = $split_arr[ $i + 1 ];
                $key   =~ s/_$//;
                $value =~ s/^\s+//;
                $hash{ to_key_name( item => $key ) } = $value;
            }
        }
        elsif ( $line =~ /Connecting/ ) {
            $hash{'connecting_url'}  = $1;
            $hash{'connecting_port'} = $2;
        }
        elsif ( $line =~ /(forensics.xml)\s+(\d+[%])\|[*]*\|\s+(\d+[\w])/ ) {
            $hash{'file_name'}           = $1;
            $hash{'download_percentage'} = $2;
            $hash{'file_size'}           = $3;
        }
        elsif ( $line =~ /([\w+\s+\w+]*)\s*\([\d+[.]*]*\)[:]\s*(\w+)/ ) {
            $hash{'status'} = $1;
            $hash{'reason'} = $2;
        }
        elsif ( $line =~ /^[<](Forensics\s*Input)/i ) {
            $outer_main_key = $1;
            while (@lines) {
                $line = shift(@lines);
                if ( $line =~ /^[<][!][-][-]\s*([\w*\s*\w*\s*]*)/i ) {
                    $key = $1;
                    if ( $key =~ /[sS]tart|[eE]nd/i ) {
                        $key =~ s/\s*[sS]tart|\s*[eE]nd//g;
                        $main_key = $key;
                    }
                    else {
                        $inner_main_key = $key;
                    }
                }
                if ( ( $line =~ /^[<][\w+]/ )
                    && !( $line =~ /[sS]tart|[eE]nd/ ) )
                {
                    if ( $line =~ /[<][!][-][-]([\w*\s*\w*\s*]*) [-][-][>]$/ ) {
                        $inner_most_key = $1;
                    }
                    else { $inner_most_key = 0; }
                    my @split_arr2 = split( /[>]/, $line );
                    if ($#split_arr2) {
                        for ( my $i = 0 ; $i < $#split_arr2 ; $i += 2 ) {
                            $split_arr2[$i]       =~ s/^\s+//g;
                            $split_arr2[$i]       =~ s/^\s+$//g;
                            $split_arr2[ $i + 1 ] =~ s/\s+$//g;
                            $split_arr2[ $i + 1 ] =~ s/^\s+//g;
                            $split_arr2[ $i + 1 ] =~ s/[<\!-\/]//g;
                            my @split_arr3 = split( /['|"]/, $split_arr2[$i] );
                            for ( my $i = 0 ; $i < $#split_arr3 ; $i += 2 ) {
                                $split_arr3[$i] =~ s/^\s+//g;
                                $split_arr3[$i] =~ s/[<=]//g;
                                if ($inner_main_key) {
                                    $hash{$main_key}{$inner_main_key}
                                      {$inner_most_key}{ $split_arr3[$i] } =
                                      $split_arr3[ $i + 1 ];
                                }
                                else {
                                    $hash{$main_key}{$inner_most_key}
                                      { $split_arr3[$i] } =
                                      $split_arr3[ $i + 1 ];
                                }
                            }
                        }
                    }
                    else {
                        my @split_arr4 = split( /['|"]/, $line );
                        for ( my $i = 0 ; $i < $#split_arr4 ; $i += 2 ) {
                            $split_arr4[$i] =~ s/^\s+//g;
                            $split_arr4[$i] =~ s/[<=]//g;
                            $hash{$main_key}{ $split_arr4[$i] } =
                              $split_arr4[ $i + 1 ];
                        }
                    }
                }
            }
            $hash2{$outer_main_key} = \%hash;
        }
    }
    push( @result_array, \%hash2 ) if (%hash2);
    push( @result_array, \%hash )  if (%hash);
    return \@result_array;
} ## end system_forensics_config

## @name ::system_cpld_version
##
## @summary Parses the raw output from the "system cpld version" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from system_cpld_version with any of its options as described in the
##    corresponding cdef file 'system.cdefs'
## @synopsis
##      system_cpld_version(%args);
## INPUTS TO METHOD CALL 'system_cpld_version' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp; sequencer : rev 9.5
##&nbsp; ebod      : rev 1.4
##&nbsp; acp       : rev 12.8
##&nbsp; nvmem     : rev 10.3
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;          {
##&nbsp;            'sequencer' => 'rev 9.5',
##&nbsp;            'ebod'      => 'rev 1.4',
##&nbsp;            'acp'       => 'rev 12.8',
##&nbsp;            'nvmem'     => 'rev 10.3',
##&nbsp;          }
##&nbsp;        ];
##&nbsp;

sub system_cpld_version {
    my %args   = @_;
    my $resp   = $args{response_object};
    my $output = $resp->get_processed_output();
    _check_error( raw_output => $output, response_object => $resp );
    my @lines = get_lines( raw_output => $output, response_object => $resp );
    my @result_array;
    foreach my $line (@lines) {
        my @values = split( /:|\s+/, $line );
        $result_array[0]->{ to_key_name( item => $values[0] ) } = $values[1];
    }
    return \@result_array;
}

## @name ::system_cpld_update
##
## @summary Parses the raw output from the "system cpld update" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from system_cpld_update with any of its options as described in the
##    corresponding cdef file 'system.cdefs'
## @synopsis
##      system_cpld_update(%args);
## INPUTS TO METHOD CALL 'system_cpld_update' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an empty array reference.

sub system_cpld_update {
    my %args   = @_;
    my $resp   = $args{response_object};
    my $output = $resp->get_processed_output();
    _check_error( raw_output => $output, response_object => $resp );
    return [];
}

1;
