#
# Copyright (c) 2012 NetApp, Inc., All Rights Reserved
# Any use, modification, or distribution is prohibited
# without prior written consent from NetApp, Inc.
#
##
## @synopsis
##    gpiotool(%args);
##
## @summary Routines to parse gpiotool command output
## @author Shakeela Shaik(sshaik@netapp.com), dl-nacl-dev@netapp.com

package NACL::APISet::Node::CLI::SPShell::default::gpiotool;
###################
# Everytime
###################
use strict;
use warnings;
use NACL::APISet::Response::CLI::ParserUtils;

# Include exceptions which will be thrown

use NACL::APISet::Exceptions::InvalidParamException qw(:try);

# Description

#  This Parser is used to run gpiotool command on the SP filer.

sub _parse_for_errors {
    my %args          = @_;
    my $output        = $args{raw_output};
    my $error_message = $args{error_message};
    my $error;
    if ( $output =~ /[Uu]sage.*/ ) {
        $error          = "InvalidParam";
        $$error_message = $output;
    }
    return $error;
}

## @name ::gpiotool
##
## @summary Parses the raw output from the "gpiotool" command
##
## @description 
##    This method can be used for parsing the raw output obtained after calling 
##    the method from gpiotool with any of its options as described in the 
##    corresponding cdef file 'gpiotool.cdefs'
## @synopsis
##      gpiotool(%args);
## INPUTS TO METHOD CALL 'gpiotool' ON APISET OBJECT 
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## <tr> <td> No switch for this key </td> <td> 'pin' </td> <td> '0' </td> </tr> 
## <tr> <td> '--verbose' </td> <td> 'verbose' </td> <td> '1' </td> </tr> 
## <tr> <td> '--set-dir-input' </td> <td> 'set-pol-high' </td> <td> '1' </td> </tr> 
## <tr> <td> '--set-gpio-ints' </td> <td> 'set-gpio-interrupts' </td> <td> '1' </td> </tr> 
## <tr> <td> '--set-dir-output' </td> <td> 'set-dir-output' </td> <td> '1' </td> </tr> 
## <tr> <td> '--set-data-low' </td> <td> 'set-data-low' </td> <td> '1' </td> </tr> 
## <tr> <td> '--set-dir-input' </td> <td> 'set-dir-input' </td> <td> '1' </td> </tr> 
## <tr> <td> '--set-od-high' </td> <td> 'set-open-drain-high' </td> <td> '1' </td> </tr> 
## <tr> <td> '--set-od-low' </td> <td> 'set-open-drain-low' </td> <td> '1' </td> </tr> 
## <tr> <td> '--set-data-high' </td> <td> 'set-data-high' </td> <td> '1' </td> </tr> 
## <tr> <td> No switch for this key </td> <td> 'port' </td> <td> '0' </td> </tr> 
## <tr> <td> '--get-data' </td> <td> 'get-data' </td> <td> '1' </td> </tr> 
## <tr> <td> '--set-dir-input' </td> <td> 'set-pol-low' </td> <td> '1' </td> </tr> 
## <tr> <td> '--reset-gpio-ints' </td> <td> 'reset-gpio-interrupts' </td> <td> '1' </td> </tr> 
## <tr> <td> '--wait-for-ints' </td> <td> 'wait-for-interrupts' </td> <td> '0' </td> </tr> 
## <tr> <td> '--get-dir' </td> <td> 'get-dir' </td> <td> '1' </td> </tr> 
## <tr> <td> '--get-pol' </td> <td> 'get-polarity' </td> <td> '1' </td> </tr> 
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

##&nbsp;Command = gpiotool --get-dir 5
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;Input Pin
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;          {
##&nbsp;            'type' => 'Input'
##&nbsp;          }
##&nbsp;        ];
##&nbsp;
##&nbsp;Command = gpiotool --set-dir-input 5
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [];
##&nbsp;
sub gpiotool {
    my %args = @_;
    my $error_message;
    my $resp    = $args{response_object};
    my $output  = $resp->get_processed_output();
    my $command = $resp->get_actual_command();
    my @result_array;
    my $error = _parse_for_errors(
        "raw_output"    => $output,
        "command"       => $command,
        "error_message" => \$error_message
    );
    if ($error) {
        $error = 'NACL::APISet::Exceptions::' . $error . 'Exception';
        $error->throw( text => $error_message, response_object => $resp );
    }

    #Now we can parse the output
    my @arraylines =
      get_lines( raw_output => $output, response_object => $resp );
    foreach my $lines (@arraylines) {
        if ( $lines =~ /Pin\s+is\s+(.*)/ ) {
            $result_array[0]->{"state"} = $1;
        } elsif ($lines =~ /(Input|Output)\s+pin/i) {
            $result_array[0]->{"type"} = $1;
        }
    }
    return \@result_array;
}    ##end of gpiotool

1;
