#
# Copyright (c) 2001-2009 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Base for API Sets related to GUI operations
## @author neelkant@netapp.com, dl-nacl-dev@netapp.com
## @status Shared
## @pod here

package NACL::APISet::Node::GUI;

#
# Compiler Directives
#
use strict;
use warnings;

#
# NATE imports
#
use Tharn qw( host);
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
my $nate_params = NATE::ParamSet->new(global => 1);

#Module Imports
use NACL::APISet::Response::GUI;
use Params::Validate qw(:all);

# Base class
use base qw(NACL::APISet::Node);

our $AUTOLOAD;

###############################################################################
# Method Name: AUTOLOAD
# Objective:  This function will be called when method is missing
###############################################################################

sub AUTOLOAD {
    my ($self, %args) = @_;
    $Log->enter() if $may_enter;
    my ($response, $auto);
    my $method = $AUTOLOAD;

    my $session = $self->{session};

    $method =~ s/.*://;    # strip fully-qualified portion

    # Call Base functions with the arguments given to execute the command
    eval {
        $auto = $session->$method(%args);
        unless (ref $auto) {
            $auto = {};
        }
        $auto->{'method_name'} = $method;
        $auto->{execution_status} = "PASS";
    };
    if ($@) {
        unless (ref $auto) {
            $auto = {};
        }
        $auto->{'method_name'}    = $method;
        $auto->{execution_status} = "FAIL";
        $auto->{output}           = $@;
    }

    $response = NACL::APISet::Response::GUI->new($auto);

    $Log->exit() if $may_exit;
    return $response;
} ## end sub AUTOLOAD

sub clear_privilege {
    $Log->enter() if $may_enter;
    $Log->exit() if $may_exit;
}

sub maximum_privilege {
    $Log->enter() if $may_enter;
    $Log->exit() if $may_exit;
    return 'admin';
}

sub DESTROY{
    my $self = shift;
    $self->quit() if ($self and $self->can("quit"));
} ## end sub DESTROY

##############################################################################
#                               Private Methods
##############################################################################

##############################################################################
# Method:    _init_API
# Objective: Initialize the Node API Set
# Details:   Initializes the API Set.  This method drives the workflow needed
#            to make the API Set ready for work.  This is accomplished by
#            creating a new schema object and loading it with CDEFs from the
#            appropriate path.
##############################################################################

sub _init_API {
    my ( $self, %args ) = @_;
    $Log->enter() if $may_enter;
    # GUI implementations do not require Schema or cdef object.
    $Log->exit() if $may_exit;
    return 0;
}

##############################################################################
# Method:    _init_Session
# Objective: Initialize the Session object
# Details:   This method establishes the Session and initializes the APISet
#            object with the session object.
##############################################################################
sub _init_Session {

    my ($self, %args) = @_;

    $Log->enter() if $may_enter;

    # Read all the parameters from Tharnhost
    my $hostobj = $self->{hostobj};
    my ( $hostname, $result, $cluster_mgmt_ip );

    # Set argument hash with the current System details
    # ex: Ip,username , password etc.
    # Get hostname and cluster managment ip form CLI connection

    my $cliApisetObj = NACL::APISet->new(
            hostobj   => $hostobj,
            category  => "Node",
            interface => "CLI",
            set       => "CMode",
    );

    $args{'apiset_cli'} = $cliApisetObj;

    $Log->comment("Getting the cluster name of the Node");
    $result     = $cliApisetObj->hostname();
    ($hostname) = $result->get_raw_output() =~ /(\S+)/;
    $Log->comment("Getting the cluster-management-ip to open SM");
    $result     = $cliApisetObj->network_interface_show(
                                     vserver => $hostname,
                                     role    => "cluster-mgmt" );

    my @parsed_output = @{ $result->get_parsed_output() };
    $cluster_mgmt_ip  = $parsed_output[0]->{'address'};

    $args{remote_server_addr} //= $nate_params->get('SMSERVERIP');

    unless ( $args{remote_server_addr} ) {
        if ( $nate_params->get('WINCLIENT') ) {
            my $c_hostobj = host($nate_params->get('WINCLIENT'));
            $args{remote_server_addr} = $c_hostobj->{'default_ip'};
        } elsif ( $nate_params->get('CLIENT') ) {
            my $c_hostobj = host($nate_params->get('CLIENT'));
            $args{remote_server_addr} = $c_hostobj->{'default_ip'};
        } else {
            $args{remote_server_addr} = $nate_params->get('SMSERVERIP') ||
                NACL::APISet::Exceptions::MissingRequiredParamException->throw(
                "Param 'WINCLIENT' or 'CLIENT' or SMSERVERIP'"
              . " is required to use GUI interface" );
        }
    }

    $args{browser_url} //= $nate_params->get('SMURL');
    $args{port} //= $nate_params->get('SELPORT') || "4444";
    $args{browser_name} //= $nate_params->get('BROWSER')
        ||  $nate_params->get("SM_BROWSER") || "firefox";
    $args{vserver} //= $nate_params->get('VSERVER')
          || $nate_params->get("DEFAULT_VSR");
    $args{'systemIp'}       = $args{ipaddress} || $cluster_mgmt_ip ;
    $args{'systemName'}     = $hostname || $hostobj->{hostname};
    $args{'systemUsername'} = $args{username}
                                    || $hostobj->{default_username};
    $args{'systemPassword'} = $args{password}
                                    || $hostobj->{default_password};

    unless ( defined($args{'browser_url'})
            and  $args{'browser_url'} =~ /http/ ) {
        $args{'browser_url'} = "https://".$args{'systemIp'};
    }

    $self->_record_connection_into_keyval();

    # Create a Session object.
    # The Session object automatically takes care of
    # connecting to the service and preparing the browser for use
    delete $args{context_definition};
    delete $args{prompt_password};
    delete $args{prompt_username};

    my $session_pkg = 'NACL::APISet::Session::GUI::CMode';
    $self->{session} = $session_pkg->new(%args);

    $Log->comment("Opening system for Managing.");
    $self->manage_system();

    $Log->exit() if $may_exit;
} ## end sub _init_Session

sub _get_highest_possible_privilege {
    $Log->enter() if $may_enter;
    my $self = $_[0];
    $Log->exit() if $may_exit;
    return "admin";
}

sub _is_versioning_needed {
    my ($self) = @_;
    $Log->enter() if $may_enter;
    $Log->exit() if $may_exit;
    return 0;
}

sub _need_to_load_cdefs {
    my ($self) = @_;
    $Log->enter() if $may_enter;
    $Log->exit() if $may_exit;
    return 0;
}

1;

##############################################################################
#                                    POD here
##############################################################################

=head1 NAME

NACL::APISet::Host::GUI::SysManager - Base class for APISets for GUI operations.

=head1 SYNOPSIS

use NACL::APISet::Host::GUI::SysManager;

=head1 DESCRIPTION

Contains Common routines for libraries under NACL/APISet/Host/GUI/SysManager

=head1 METHODS
=head2 new

This method is used to create a NACL::APISet::Host::GUI::SysManager object.

=over

=item Synopsis

$output = $GuiObj->new(%args);;

=item Arguments

new function accepts all arguments.

=item Returns

(Scalar)
GUI object

=back

=head2 parse_system_details

This function parses input parameter and creates hash with System IP, System Name, Username and Password.

=over

=item Synopsis

%ouput = $GuiObj->parse_system_details($host_detail_string);

=item Arguments

accepts input string containing the detail of a host.

=item Returns

(Hash)
A hash containing System IP, System Name, Username and Password as keys.

=back

=head1 AUTHOR/MAINTAINER

=over

=item NACL Development (neelkant@netapp.com, dl-nacl-dev@netapp.com)

=back

=cut
