# Copyright (c) 2001-2010 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Base for API Sets SNMP commands
## @author Isshwarya.M@netapp.com, dl-nacl-dev@netapp.com
## @status Shared
## @pod here

package NACL::APISet::Node::SNMP;

#
# Compiler Directives
#
use strict;
use warnings;

#
# NATE imports
#
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Data::Dumper;
use Params::Validate qw(:all);
use Tie::IxHash;

#
# Module import - APISets
#
use NACL::APISet::Schema::SnmpSchema;

#
# Exceptions imported
#
use NACL::APISet::Exceptions::CDEFNotFoundException qw(:try);
use NATE::BaseException qw(:try);
use NATE::Exceptions::Argument qw(:try);
use NACL::APISet::Exceptions::InvalidInputRequest qw(:try);
use NACL::Exceptions::InvalidFilterField qw(:try);

# global variables
use constant RETRIES => 0;
use constant MSG_SIZE => 1472; # UDP/IPv4
use constant TIMEOUT => 5;

# Base class
use base qw(NACL::APISet::Node);

our $AUTOLOAD;

###############################################################################
# Method Name: AUTOLOAD
# Objective:  Catch all commands not specifically implemented and call _invoke
###############################################################################

sub AUTOLOAD
{
    my ($self, %args) = @_;
    $Log->enter();

    my $method = $AUTOLOAD;
    $method =~ s/.*://; # strip fully-qualified portion
    $Log->debug(sub { "Options to '$method' are:\n".Dumper(\%args)});
    if ($method eq "DESTROY") {
        $Log->exit();
        return;
    }
    # Check if the method is supported to execute on SNMP APISet object.
    my $actual_method = $method;
    $actual_method =~ s/^xlate_//;
    if ($actual_method !~ /(get|get_next|set|get_bulk|snmp_walk)/) {
        NACL::APISet::Exceptions::InvalidInputRequest->throw(
            "Invalid SNMP PDU specified: $method");
    }

    if ($method =~ /^xlate_/) {
        if ($method =~ /^xlate_execute_raw_command/) {
            # Explicitly prevent xlate_ being used with execute_raw_command
            $Log->exit();
            NATE::BaseException->throw("xlate cannot be used " .
                                       "with execute_raw_command");
        } else {
            $Log->exit();
            if (defined(&$actual_method)) {
                $Log->debug("$actual_method is defined");
                return $self->$actual_method(cdef_method => $method, %args);
            } else {
                return $self->_invoke(cdef_method => $method, %args);
            }
        }
    } else {
        $Log->comment("In AUTOLOAD executing SNMP $method");
        $self->{cdef_method} = $method;

        # Call invoke with the arguments given to execute the command
        $Log->exit();
        return $self->_invoke(cdef_method => $method, %args);
    }
} ## end sub AUTOLOAD

##############################################################################
# Method:    snmp_walk
# Objective: To implement snmpwalk
# Details:   Refer to POD documentation at the end of the file.
##############################################################################

sub snmp_walk {
    my $self = shift();
    $Log->enter();

    my %args = NACL::APISet::validate_params(
        $self,
        params => \@_,
        spec   => { "baseoid" => { type => SCALAR, optional => 0 },
                    # If snmp_walk is invoked as method call, then AUTOLOAD
                    # won't be hit and hence assign defaults for cdef_method.
                    "cdef_method" => { type => SCALAR, optional => 1,
                        default=>'snmp_walk' },
                  }
    );
    my $schema_obj = $self->get_schema_obj();
    my $only_translate = 0;
    my $method = delete $args{cdef_method};
    if ($method =~ /^xlate_/) {
        $method =~ s/^xlate_//;
        $only_translate = 1;
    }

    # obtain the numeric OID
    my $baseoid = delete($args{baseoid});
    # Throw exception if an invalid parameter is passed.
    if (scalar(keys %args)) {
        $Log->exit();
        NATE::Exceptions::Argument->throw("Invalid Argument(s) used: " .
         join(', ', keys(%args)));
    }
    my $varbind = $schema_obj->translate(varbind => $baseoid);
    my %input;
    $input{method} = $method;
    $input{baseoid} = $varbind->{oid};

    $Log->debug("Translated command to session class is:\n". Dumper(\%input));

    # Invoke
    my $result;
    if ($only_translate) {
        $result = \%input;
    } else {
        # Passing schema object for response class parsing. Adding it after
        # debug line purposely to avoid returning them in the output for
        # xlate_*() unnecessarily and also to prevent logs getting bloated.
        $result = $self->{session}->send_command(%input, schema => $schema_obj);
    }

    $Log->exit();
    return $result;
}

##############################################################################
# Method:    translate_name_to_oid
# Objective: To translate the snmp object name to numeric OID format
# Details:   Refer to POD documentation at the end of the file.
##############################################################################

sub translate_name_to_oid {
    my $self = shift();
    $Log->enter();

    my %args = NACL::APISet::validate_params (
        $self,
        params => \@_,
        spec   => { "object-name" => { type => SCALAR, optional => 0 },
                    "need-cdef-entry" => {type => SCALAR, optional => 1,
                        default => 0}
                  }
    );
    my $schema_obj = $self->get_schema_obj();
    my $need_cdef_entry = delete $args{'need-cdef-entry'};

    # obtain the object name
    my $name = delete($args{'object-name'});
    # Throw exception if an invalid parameter is passed.
    if (scalar(keys %args)) {
        $Log->exit();
        NATE::Exceptions::Argument->throw("Invalid Argument(s) used: " .
         join(', ', keys(%args)));
    }
    my $result = $schema_obj->convert_to_numericOID(object => $name);

    my $output;
    if ($need_cdef_entry) {
        $output = $result;
    } else {
        $output = $result->{oid};
    }
    $Log->exit();
    return $output;
}

##############################################################################
# Method:    translate_oid_to_name
# Objective: To translate the numeric OID to its object name
# Details:   Refer to POD documentation at the end of the file.
##############################################################################

sub translate_oid_to_name {
    my $self = shift();
    $Log->enter();

    my %args = NACL::APISet::validate_params(
        $self,
        params => \@_,
        spec   => { "oid" => { type => SCALAR, optional => 0 },
                    "need-cdef-entry" => {type => SCALAR, optional => 1,
                        default => 0}
                  }
    );
    my $schema_obj = $self->get_schema_obj();
    my $need_cdef_entry = delete $args{'need-cdef-entry'};

    # obtain the numeric OID
    my $oid = delete($args{'oid'});
    # Throw exception if an invalid parameter is passed.
    if (scalar(keys %args)) {
        $Log->exit();
        NATE::Exceptions::Argument->throw("Invalid Argument(s) used: " .
         join(', ', keys(%args)));
    }
    my $result = $schema_obj->convert_numericOID_to_name(OID => $oid);

    my $output;
    if ($need_cdef_entry) {
        $output = $result;
    } else {
        $output = $result->{object};
    }
    $Log->exit();
    return $output;
}

##############################################################################
# Method:    execute_raw_command
# Objective: Execute the raw SNMP command on the filer and return the raw
#            output
# Details:   See POD documentation at the end of this file.
##############################################################################

sub execute_raw_command {
    my ($self, %args) = @_;
    $Log->enter();

    # If users had requested for a response object, then return a response
    # object.
    my $need_resp_obj;
    if (defined $args{'return-response-object'}) {
        $need_resp_obj = $args{'return-response-object'};
    } elsif (defined $self->{'return-response-object'}) {
        $need_resp_obj = $self->{'return-response-object'};
    }
    $Log->debug("Response object is requested") if ($need_resp_obj);

    if (defined $need_resp_obj) {
        $args{'return-response-object'} = $need_resp_obj;
        $args{schema} = $self->get_schema_obj();
    }

    my $output = $self->{session}->execute_raw_command(%args);

    $Log->exit();
    return $output;
}

# This operation can only be invoked on tables or columns
# Eg :1
# Fetch information about all aggregates in aggrTable whose state is online and
# of type 'flex' - equivalent to specifing the query option
# $self->get_objects(baseoid => 'aggrTable', aggrState => 'online',
#   aggrType=>'flex');
# Eg :2
# Same as above but in the output, show only aggrName column -equivalent
# to specifying the desired attributes along with query.
# $self->get_objects(baseoid => 'aggrName', aggrState => 'online',
#   aggrType=>'flex');
# Eg :3
# Same as Eg#1 but in the output, show aggrName,aggrUUID columns -equivalent
# to specifying the desired attributes along with query.
# $self->get_objects(baseoid => 'aggrTable', aggrState => 'online',
#   aggrType=>'flex',requested_columns => [ qw(aggrName aggrUUID) ] );

# TODO: Specifying instance as string is not yet supported

##############################################################################
# Method:    get_matching_objects
# Objective: Get the rows of the table which matches the query. If
# the baseoid is a column name, then only the respective column will be
# displayed in the output for each of the matching row. This is equivalent to
# specifying desired attributes for ZAPIs.
# Details:   See POD documentation at the end of this file.
##############################################################################

sub get_matching_objects {
    my $self = shift();
    $Log->enter();

    my %args = NACL::APISet::validate_params(
        $self,
        params => \@_,
        spec   => {
            "baseoid"           => { type => SCALAR,   optional => 0 },
            "requested_columns" => { type => ARRAYREF, optional => 1 }
        }
    );
    $Log->debug("Arguments to get_matching_objects". Dumper(\%args ));
    my (%grouped, @result_array);
    my $baseoid = delete $args{baseoid};
    my $requested_columns = delete $args{requested_columns} ;
    my $result = $self->translate_name_to_oid( 'object-name' => $baseoid,
        'need-cdef-entry' => 1);
    my $need_cdef = $self->get_cdef_entry_option();
    my $schema_obj = $self->get_schema_obj();
    $self->set_cdef_entry_option('need-cdef-entry' => 1);

    # Throw exception if the query is missing
    NATE::Exceptions::Argument->throw("Missing query")
        unless (scalar(keys %args));

    try {
        foreach my $filter_object (keys %args) {
            # If we couldn't obtain the information for any of the filter object,
            # then the exception thrown would be propogated up and get_objects()
            # will also fail.
            my $response =  $self->snmp_walk(baseoid => $filter_object);
            # The parsed output will be of the form
            # [{filter_object => {}}, {filter_object => {}},..]
            my $parsed_output = $response->get_parsed_output();
            # If the parsed output has information about more than one column
            # object, then the specified filter doesn't correspond to a column
            # name. Since the operation is snmp_walk, all the rows would be
            #  symmetrical and it's enough to examine one row.
            if (scalar (keys %{$parsed_output->[0]}) > 1) {
                NATE::Exceptions::Argument->throw("Invalid filter object ".
                  "specified: $filter_object is not a column object in any ".
                  "of the known tables");
            }
            # Check if the specified filter is a column object under the specified
            # baseoid table object. If baseoid corresponds to a column object, then
            # both baseoid and filter object should come under same table.
            my @names = keys %{$parsed_output->[0]};
            my $filter_base_oid =
                $parsed_output->[0]->{$names[0]}->{cdef_entry}->{oid};
            my $walk_oid = quotemeta($result->{node}->{oid});
            if ($result->{node}->{type} eq 'table') {
                NATE::Exceptions::Argument->throw("Specified filter ".
                    "$filter_object is not a column under the table $baseoid")
                    if ($filter_base_oid !~ /^$walk_oid\.\d+\.\d+$/);
            } elsif ($result->{node}->{type} eq 'column') {
                my $row_entry = $filter_base_oid;
                $row_entry =~ s/\.(\d+)$//;
                $row_entry = quotemeta($row_entry);
                if ($result->{node}->{oid} !~ /^$row_entry\.\d+$/) {
                    NATE::Exceptions::Argument->throw("Filter object ".
                        "$filter_object and baseoid $baseoid belongs to ".
                        "different tables");
                }
            } else {
                NATE::Exceptions::Argument->throw("Invalid baseoid object ".
                  "specified: $baseoid is of type $result->{node}->{type}. You".
                  " can only specify an object of type 'table' or 'column'.");
            }

            # Group hash entries in the parsed output having same instances
            # together.
            foreach my $entry (@$parsed_output) {
                @names = keys %$entry;
                $grouped{$entry->{$names[0]}->{instance}}{$entry->{$names[0]}->{object_name}} =
                    $entry->{$names[0]};
            }
        }
    } finally {
        # set the 'cdef_entry_in_response' back to what user had set already
        $self->set_cdef_entry_option('need-cdef-entry' => $need_cdef);
    };

    # Get the instance(s) which matches the specified query
    my @matching_rows = ();
    foreach my $index (keys %grouped) {
        my $flag = 1;
        foreach my $query (keys %args) {

            my $value_obtained = $grouped{$index}{$query}{value};
            if ($value_obtained ne $args{$query}) {
                $flag = 0;
                last;
            }
        }
        push (@matching_rows, $index) if ($flag);
    }
    unless (scalar @matching_rows) {
        # No matching query.
        $Log->exit();
        return [];
    }
    # Now get the information for the matching rows.

    if ( $result->{node}->{type} eq 'table' ) {

        # Read all columns for the matching rows
        my $column_nodes = $schema_obj->read_column_entries(
            baseoid => $result->{node}->{oid},
            module  => $result->{node}->{mib}
        );

        foreach my $row (@matching_rows) {

            my ( @desired_column_nodes, $column_found, @invalid_columns );
            $column_found    = 0;
            @invalid_columns = ();
            if ( defined @$requested_columns ) {

                my @read_requested_columns = @$requested_columns;
                my $final_index            = $#read_requested_columns;
                my $loop_index             = 0;

                ##Iterate through each entry in the requested_columns array
                ##If the name matches an entry in @$column_nodes returned
                ##by the schema object, store the hash entry in
                ##desired_column_nodes array. Later the oid from the entry
                ##will be used in the varbindlist of snmp-get command.

                for ( my $i = $loop_index; $i <= $final_index; $i++ ) {
                    $column_found = 0;
                    foreach (@$column_nodes) {
                        if ( $_->{name} eq $read_requested_columns[$i] ) {
                            push @desired_column_nodes, $_;
                            $column_found = 1;
                            last;
                        } ## end if  { $_->{name}
                    }    ##end foreach @$column_nodes
                    if ( !$column_found ) {
                        push @invalid_columns, $read_requested_columns[$i];
                    }
                }    #end for loop

            }    ## End if ( defined @$requested_columns )
            else {
                @desired_column_nodes = @$column_nodes;
            }

            if (@invalid_columns) {

                my $message = join( ",", @invalid_columns );

                $Log->exit();
                NACL::APISet::Exceptions::MethodNotFoundException->throw(
                    "\n CDEF entry doesn't exist (in CDEF Path: ". $self->{cdef_path}
                        . ") for the following invalid column(s) "
                        . "requested by the user:\n$message\nThis is because,\n "
                        . "1) These are invalid object names which are not "
                        . "supported by the service.\n"
                        . "2) These are valid object names but the CDEF file needs "
                        . "to be updated.\n"
                        . "If it is case #2, then raise a burt with type as nacl "
                        . "and subtype as nacl_core" );
            }

            $Log->debug( "This is the desired column nodes "
                    . Dumper( \@desired_column_nodes ) );

            my %row_entry;
            my @oids = ();
            map { push( @oids, $_->{oid} . ".$row" ) } @desired_column_nodes;
            my $resp = $self->execute_raw_command(
                varbind                  => \@oids,
                method                   => 'get',
                'return-response-object' => 1,
                'on-error'               => 'ignore'
            );
            $Log->debug( "This is the response " . Dumper($resp) );

            foreach my $entry ( @{ $resp->get_parsed_output() } ) {
                $row_entry{ $entry->{object_name} } = $entry
                    if ( $entry->{asn1_class_tag} !~ /noSuchObject/ );
            }
            push( @result_array, \%row_entry );
        }
        $Log->debug(
            "This is the output of API Call " . Dumper( \@result_array ) );
        my %keys_result_columns = %{ $result_array[0] };
        my @result_columns      = keys %keys_result_columns;

        if ( @$requested_columns
            && scalar @result_columns != scalar @$requested_columns ) {

            my @search_column   = @$requested_columns;
            my @invalid_columns = ();
            my $column_found;

            for ( my $i = 0; $i <= $#search_column; $i++ ) {
                $column_found = 0;
                for ( my $j = 0; $j <= $#result_columns; $j++ ) {
                    if ( $result_columns[$j] eq $search_column[$i] ) {
                        $column_found = 1;
                        last;
                    }
                }    ## inner j loop ends here
                if ( !$column_found ) {
                        push @invalid_columns, $search_column[$i];
                    }

            }    ## Outer for loop ends here

            $Log->exit();
            if (scalar @invalid_columns ){
            NACL::Exceptions::InvalidFilterField->throw( "One or more Fields "
                    . "requested are not valid . Invalid fields are  : "
                    . Dumper( \@invalid_columns ) );
            }
        }    ## if scalar @result_columns ends here

    }
    elsif ( $result->{node}->{type} eq 'column' ) {
    if ( defined @$requested_columns ) {

        NATE::Exceptions::Argument->throw( "The 'baseoid' input is  "
                . "provided to be the column name $baseoid \n and also the "
                . "requested_columns is specified to be "
                . join( ",", @$requested_columns )
                . "\n This is a "
                . "conflicting input scenario that you cannot request  "
                . "other columns, when the view of the table (baseoid)  "
                . " is limited to be a single column of that table " );
    }
        # Read only that column for the matching rows
        my @oids = ();
        map { push( @oids, $result->{node}->{oid} . ".$_" ) } @matching_rows;
        my $resp = $self->execute_raw_command(
            varbind                  => \@oids,
            method                   => 'get',
            'return-response-object' => 1,
            'on-error'               => 'ignore'
        );
        foreach my $entry ( @{ $resp->get_parsed_output() } ) {
            push( @result_array, { $entry->{object_name} => $entry } )
                if ( $entry->{asn1_class_tag} !~ /noSuchObject/ );
        }
    }

    $Log->exit();
    return \@result_array;
}

##############################################################################
# Method:    set_error_handler
# Objective: Set the error handler for ObjectRetrival failure.
# Details:   Refer to POD documentation at the end of the file.
##############################################################################

sub set_error_handler {
    $Log->enter();

    my $self = shift();

    my %args = NACL::APISet::validate_params(
        $self,
        params => \@_,
        spec =>
            { 'on-error' => { type => SCALAR|CODEREF, optional => 0},
            }
    );
    $self->{session}->{error_handler} = $args{'on-error'};
    $Log->exit();
}

##############################################################################
# Method:    get_error_handler
# Objective: Get the error handler stored in APISet object for ObjectRetrival
# failure.
# Details:   Refer to POD documentation at the end of the file.
##############################################################################

sub get_error_handler {
    $Log->enter();

    my $self = shift();

    $Log->exit();
    return $self->{session}->{error_handler};
}

##############################################################################
# Method:    clear_error_handler
# Objective: clear the error handler for ObjectRetrival failure. Default is
# to throw exception
# Details:   Refer to POD documentation at the end of the file.
##############################################################################

sub clear_error_handler {
    $Log->enter();

    my $self = shift();

    $self->set_error_handler('on-error' => 'exception');
    $Log->exit();
}

##############################################################################
# Method:    set_session_attributes
# Objective: set the attributes of the SNMP session such as retries, timeout
# and maximum message size
# Details:   Refer to POD documentation at the end of the file.
##############################################################################

sub set_session_attributes {
    $Log->enter();

    my $self = shift();

    my %args = NACL::APISet::validate_params(
        $self,
        params => \@_,
        spec =>
            { 'retries' => { type => SCALAR, optional => 1},
              'timeout' => { type => SCALAR, optional => 1},
              'max-msg-size' => { type => SCALAR, optional => 1},
            }
    );

    my $netsnmp = $self->get_connection();
    foreach my $attribute (keys %args) {
        my $value = $args{$attribute};
        $self->{session}->{session_attributes}->{$attribute} = $value;
        $attribute =~ s/-/_/g;
        my $output = $netsnmp->$attribute($value);
        unless (defined $output) {
            NATE::BaseException->throw("Failed to set $attribute and the".
                "reason is: ". $netsnmp->error());
        }
    }
    $Log->exit();
}

##############################################################################
# Method:    get_session_attributes
# Objective: Get the values for various attributes of the SNMP session such as
# retries, timeout and maximum message size
# Details:   Refer to POD documentation at the end of the file.
##############################################################################

sub get_session_attributes {
    $Log->enter();

    my $self = shift();
    my %status;

    my $netsnmp = $self->get_connection();
    foreach my $attribute (keys %{$self->{session}->{session_attributes}}) {
        my $attr = $attribute;
        $attr =~ s/-/_/g;
        $status{$attribute} = $netsnmp->$attr();
    }

    $Log->exit();
    return \%status;
}

##############################################################################
# Method:    clear_session_attributes
# Objective: set the defaults for various attributes of the SNMP session such as
# retries, timeout and maximum message size
# Details:   Refer to POD documentation at the end of the file.
##############################################################################

sub clear_session_attributes {
    $Log->enter();

    my $self = shift();

    $self->set_session_attributes('retries' => RETRIES, timeout => TIMEOUT,
        'max-msg-size' => MSG_SIZE);
    $Log->exit();
}

##############################################################################
# Method:    set_cdef_entry_option
# Objective: To set whether CDEF entry should be returned in parsed output or
# not for each of the SNMP object
# Details:   Refer to POD documentation at the end of the file.
##############################################################################

sub set_cdef_entry_option {
    $Log->enter();

    my $self = shift();

    my %args = NACL::APISet::validate_params(
        $self,
        params => \@_,
        spec =>
            { 'need-cdef-entry' => { type => SCALAR, optional => 0 } }
    );

    $self->{session}->{'cdef_entry_in_response'}
        = ( $args{'need-cdef-entry'} ) ? 1 : 0;

    $Log->exit();
}

##############################################################################
# Method:    get_cdef_entry_option
# Objective: To get the status of cdef_entry option which determines whether
# cdef_entries should be returned in parsed output or not.
# Details:   Refer to POD documentation at the end of the file.
##############################################################################

sub get_cdef_entry_option {
    $Log->enter();

    my $self = shift();
    $Log->exit();
    return $self->{session}->{'cdef_entry_in_response'};
}

##############################################################################
# Method:    clear_cdef_entry_option
# Objective: To clear the cdef_entry option to its default
# Details:   Refer to POD documentation at the end of the file.
##############################################################################

sub clear_cdef_entry_option {
    $Log->enter();
    my $self = shift();
    $self->{session}->{'cdef_entry_in_response'} = 0;

    $Log->exit();
}


##############################################################################
#                               Private Methods
##############################################################################

##############################################################################
# Method:    _init_API
# Objective: Initialize the SNMP API Set
# Details:   Initializes the API Set.  This method drives the workflow needed
#            to make the API Set ready for work.  This is accomplished by
#            creating a new schema object and loading it with CDEFs from the
#            appropriate path
##############################################################################

sub _init_API {
    my ($self, %args) = @_;
    $Log->enter();

    # Get the absolute path for CDEF directory
    my $path = $self->{cdef_path};
    $self->{cdef_path} = NATE::Inc::find_file($path);

    unless (defined $self->{cdef_path}) {
        NACL::APISet::Exceptions::CDEFNotFoundException->throw(
            "Error in finding CDEF directory $path and the reason: \n$@");
    }

    # Create the schema object and load the CDEFs from CDEF path discovered by
    # versioning or the user specified path. Load CDEFS after the check for
    # shell_type is done so that we need not unnessarily load CDEFs even in
    # case of invalid shell_type.
    $self->{schema} = NACL::APISet::Schema::SnmpSchema->new();
    $self->{schema}->{searched_cdef_path} = $self->{cdef_path};
    $self->{schema}->load(path => $self->{cdef_path});

    $Log->exit();
}

##############################################################################
# Method:    _init_Session
# Objective: Initialize the SNMP Session object
# Details:   This method establishes the Session and initializes the APISet
#            object with the session object.
##############################################################################

sub _init_Session {
    my ( $self, %args ) = @_;
    $Log->enter();

    $self->_get_connectrec_or_hostrec(\%args);

    my $interface = $self->{interface};
    my $api_set = $self->{set};

    if($api_set ne $self->{shell_type}) {
        $Log->exit();
        NATE::Exceptions::Argument->throw("Invalid hostobj or shell_type "
            . "is sent for set '$self->{actual_set}'"
            . " and the shell_type obtained is ".$self->{shell_type});
    }

    # Establish the session if not provided
    my $session_pkg = "NACL::APISet::Session::" . $interface . "::"
        . $api_set;

    # We havent technically connected to the node yet
    # but it's imminent. Doing it here avoids having to do work
    # on every SNMP call
    $self->_record_connection_into_keyval();

    # Create a Session object
    $self->{session} = $session_pkg->new(%args);

    # Set error handler
    $self->clear_error_handler();

    # Clear session attributes
    $self->clear_session_attributes();

    $Log->exit();
}

###############################################################################
# Method Name: _invoke
# Objective:   translate object identifier into numeric OID and execute
# Details:     The user specified partially qualified object path is translated
# to its corresponding numeric OID using the schema object which has loaded
# MIB files in it.
###############################################################################

sub _invoke
{
    my $self = shift();
    $Log->enter();

    my %args = NACL::APISet::validate_params(
        $self,
        params => \@_,
        spec   => {
            "cdef_method" => { type => SCALAR, optional => 0 },
            "varbind" => { type => SCALAR|ARRAYREF, optional => 0 },
            "maxrepetitions" => { type => SCALAR|UNDEF, optional => 1 },
            "nonrepeaters" => { type => SCALAR|UNDEF, optional => 1 },
            }
    );

    # verify mandatory cdef argument
    unless (length $args{cdef_method}) {
        $Log->exit();
        NATE::Exceptions::Argument->throw(
            "Mandatory argument \"cdef_method\" not specified ");
    }
    my $only_translate = 0;
    if ($args{cdef_method} =~ /^xlate_/) {
        $args{cdef_method} =~ s/^xlate_//;
        $only_translate = 1;
    }

    my $schema_obj = $self->get_schema_obj();

    # Handle both array and scalar inputs for varbind
    my @oids = (!ref $args{varbind})? (delete $args{varbind}) :
        @{delete ($args{varbind})};
    my %set_hash;
    if ($args{cdef_method} eq 'set') {
        %set_hash = @oids;
        @oids = keys %set_hash;
    }

    # translate object instances to numeric OIDs.
    my $varbind = $schema_obj->translate(varbind => \@oids);

    # Construct the command to be sent for session
    my %input;
    # For GetBulk PDU, order of the varbindlist matters and so tie the hash and
    # preserve the order. Also order need to be preserved so that user can
    # compare the error index field with the object order specified.
    my %varbindhash;
    tie %varbindhash, 'Tie::IxHash';
    foreach my $object_in_input (@oids) {
        $varbindhash{$object_in_input} = $varbind->{oid}->{$object_in_input};
    }
    $input{varbind} = \%varbindhash;
    $input{method} = delete($args{cdef_method});
    if ($input{method} eq "get_bulk") {
        $input{'maxrepetitions'} = delete($args{'maxrepetitions'});
        $input{'nonrepeaters'} = delete($args{'nonrepeaters'});
    } elsif ($input{method} eq "set") {
        # Reconstruct the varbind entry in the command to meet the
        # needs of 'Set' PDU.
        my %sethash;
        tie %sethash, 'Tie::IxHash';
        foreach my $object (@oids) {
            my $type;
            my $node = $varbind->{node}->{$object};
            if (exists $node->{'syntax'}->{'type'}->{'basetype'}) {
                $type = $node->{'syntax'}->{'type'}->{'basetype'};
            } elsif (exists $node->{'syntax'}->{'type'}->{'name'}) {
                $type = $node->{'syntax'}->{'type'}->{'name'};
            } else {
                NATE::Exceptions::Argument->throw(
                    "Internal error: Missing datatype information in the CDEF entry")
                    if ($node->{type} =~ /(column|scalar)/);
                $type = '';
            }
            $sethash{$object} = {type => $type,
                value => $set_hash{$object},
                oid => $varbind->{oid}->{$object}};
        }
        $input{varbind} = \%sethash;
    }
    $input{'on-error'} = delete $args{'on-error'}
        if(defined $args{'on-error'});

    # Throw exception if an invalid parameter is passed. For SNMP, we cannot
    # place the invalid parameters also in the PDU and expect the service to
    # react, like how we do in other interfaces.
    if (scalar(keys %args)) {
        $Log->exit();
        NATE::Exceptions::Argument->throw("Invalid Argument(s) used: " .
         join(', ', keys(%args)));
    }

    $Log->debug("Translated command to session class is:\n". Dumper(\%input));

    my $result;
    if ($only_translate) {
        $result = \%input;
    } else {
        # Passing schema object for response class parsing. Adding it after
        # debug line purposely to avoid returning them in the output for
        # xlate_*() unnecessarily and also to prevent logs getting bloated.
        $result = $self->{session}->send_command(%input, schema =>$schema_obj);
    }
    $Log->exit();
    return $result;
} ## end sub _invoke

1;

##############################################################################
#                                    POD here
##############################################################################

=head1 NAME

NACL::APISet::Node::SNMP - Base for API Sets for SNMP commands

=head1 SYNOPSIS

 use NACL::APISet::Node::SNMP;

=head1 DESCRIPTION

A C<NACL::APISet::Node::SNMP> is used to invoke SNMP PDUs and other SNMP related
operations like snmpwalk, translate OID to name, viceversa. This
is further subclassed to send SNMP commands specifically to CMode, 7Mode etc.

=cut

=head1 EXCEPTIONS

=over

=item NATE::BaseException

=item NATE::Exceptions::Argument

=item NACL::APISet::Exceptions::CDEFNotFoundException

=item NACL::APISet::Exceptions::InvalidInputRequest

=back

=cut

=head1 METHODS

=head2 set_error_handler

This method sets the error handler to be performed on obtaining SNMPv2c
and SNMPv3 specific errors such as noSuchInstance, noSuchObject and
endOfMibView in which cases object retrieval fails but SNMP protocol
will not poulate error status and error index fields.

=over

=item Synopsis

$Api_Set_Obj->set_error_handler('on-error' => 'ignore');

=item Arguments

=over

=item on-error

Required
It accepts either scalar value or code reference. Valid scalar
values are 'exception' (default) which will throw exception on obtaining
'noSuchObject', 'noSuchInstance', 'endOfMibView' and 'ignore' will
ignore these errors and return a response object which will have partially
parsed output (parsed output entries for objects whose retrieval is failed
will not have all the fields present). If a code reference is specified,
then the subroutine will be invoked by passing exception object as the
first argument t oit. Also a response object will be returned finally
which will have partially parsed output.

=back

=item Returns

Nothing.

=back

=head2 get_error_handler

Returns the error handler value which is set in APISet object.

=over

=item Synopsis

my $handler = $Api_Set_Obj->get_error_handler();

=item Arguments

Nothing

=item Returns

Returns the error handler which can be a scalar value or code reference.

=back

=head2 clear_error_handler

Sets the error handler to the default value 'exception'. If in
case SNMP object retrival fails for noSuchInstance, noSuchObject,
endOfMibView, then 'NACL::APISet::Exceptions::ObjectRetrievalFailed' exception
will be thrown.

=over

=item Synopsis

$Api_Set_Obj->clear_error_handler();

=item Arguments

Nothing

=item Returns

Nothing

=back

=head2 set_session_attributes

This method sets the attributes of the SNMP session to user specified value.
NATE::BaseException will be thrown if the attribute cannot be set to the
specified value.

=over

=item Synopsis

$Api_Set_Obj->set_session_attributes(
    ['retries' => $retries],
    ['timeout' => $timeout],
    ['max-msg-size' => $max_msg_size],
    );

=item Arguments

=over

=item retries

Optional
The number of times to retry sending a SNMP message to the remote host.
Default is 1.

=item timeout

Optional
The maximum number of seconds that the object will wait for a response from
the agent on the remote host
Default is 5 seconds.

=item max-msg-size

Optional
The maximum message size in octets that can be prepared or processed by the
object. Default is 1472 octets corresponding to UDP/IPv4

=back

=item Returns

Nothing.

=back

=head2 get_session_attributes

This method gets the current values for various attributes of SNMP session.

=over

=item Synopsis

my $optionHash = $Api_Set_Obj->get_session_attributes();

=item Arguments

Nothing

=item Returns

Returns a hash reference whose keys are session attributes such as retries,
timeout, max-msg-size and values are the current session values for those
attributes.

=back

=head2 clear_session_attributes

This method resets the current values for various attributes of SNMP session
to its corresponding defaults.

=over

=item Synopsis

$Api_Set_Obj->clear_session_attributes();

=item Arguments

Nothing

=item Returns

Nothing

=back

=head2 set_cdef_entry_option

This method enables/disables the cdef_entry option which determines whether or
not to return the corresponding cdef_entry in the parsed output hash for each
OID object returned in the output.
Default value for this option is 0 (disabled) so that parsed output entries will
be light weight in response object.

=over

=item Synopsis

$Api_Set_Obj->set_cdef_entry_option('need-cdef-entry' => 1);

=item Arguments

=over

=item need-cdef-entry

Required
Takes a value 0/1.

=back

=item Returns

Nothing.

=back

=head2 get_cdef_entry_option

This method gets the current status of cdef_entry option.

=over

=item Synopsis

my $is_cdef_entry_set = $Api_Set_Obj->get_cdef_entry_option();

=item Arguments

Nothing

=item Returns

Returns 1/0 based on whether the option is set or not.

=back

=head2 clear_cdef_entry_option

This method disables the cdef_entry option which is the default behavior.

=over

=item Synopsis

$Api_Set_Obj->clear_cdef_entry_option();

=item Arguments

Nothing

=item Returns

Nothing

=back

=head2 get_matching_objects

When invoked on a table, this method returns the parsed output for the rows
of the table which matches the query. When invoked against a column object of
a table, it will return only the specified column information of the rows
which matches the query. The second use case can be considered to be equivalent
of specifying 'desired-attributes' in CMode ZAPI or -fields option in CMode CLI.
If none of the rows matches the query, then
NACL::APISet::Exceptions::NoMatchingEntriesException will be thrown.

=over

=item Synopsis

my $output = $Api_Set_Obj->get_matching_objects(baseoid => $objectname,
              [columnName => value],...);

=item Arguments

=over

=item baseoid
Name or partially qualified object path of the table or column

=item query
key-value pairs representing query in which keys are the objects and the
value are corresponding values which the user is interested at.

=item requested_columns
(Optional)
Reference to an array of column names which the user is interested in the
parsed output. If this parameter is provided, then only these columns with
matching rows if  a query is provided or otherwise all the rows in this column,
are returned in the output .

=back

=item Returns

Returns a hash reference which is similar in structure to the parsed output
of snmpwalk. If the baseoid corresponds to table, then all columns will be
returned for matching rows. Its an array of hashes structure where each hash
represents a conceptual row and hash keys are column names and values are
hash references representing the corresponding instance of the column object.
If the baseoid corresponds to a column, then only the specified column
information will be returned for the matching rows. Here the hashes in array
of hashes structure will have only one entry or key-value pair representing
the column specified.

=back

=head2 execute_raw_command

This method is used to execute SNMP PDUs and snmpwalk by directly specifying the
operation to be performed and the numeric OID. So this involves no translation
of object name to numeric OID.

=over

=item Synopsis

$output = $Api_Set_Obj->execute_raw_command(%args);

=item Arguments

=over

=item method

(Mandatory)
The operation to be invoked. Valid values are get, set, get_next, get_bulk and
snmp_walk.

=item varbind

(Optional)
Array reference specifying the list of numeric OIDs
This is a mandatory option for all operations other than snmp_walk.

=item baseoid

(Optional)
BaseOID to be used for performing snmp walk. This can be specified
only if the 'method' name is snmp_walk.

=item maxrepetitions

(Optional)
maxrepetition value for get_bulk PDU.

=item nonrepeaters

(Optional)
nonrepeaters value for get_bulk PDU.

=item on-error
Error handler override if the object retrival fails with reasons such as
noSuchObject, noSuchInstance, endOfMibView. This option should be specified
only if 'return-response-object' is set to 1.

=item return-response-object

(Optional)
Boolean option. If set, a response object will be returned else raw output
(hash reference) will be returned. By default, this option will be set to 0.

=back

=item Returns

Raw output or Response object depending on 'return-response-object' option.

=back

=head2 snmp_walk

Implements snmp walk operation. Reads all rows and all columns in a table
when the baseoid corresponds to column name and reads all instances of the
specified column if the baseoid corresponds to a column name.

=over

=item Synopsis

$response = $Api_Set_Obj->snmp_walk(baseoid => $baseoid);

=item Arguments

=over

=item baseoid

(Mandatory)
The object name which either corresponds to name of the table or column name.

=back

=item Returns

Response object.

=back

=head2 translate_name_to_oid

Translates the object name to numeric OID format

=over

=item Synopsis

$output = $Api_Set_Obj->translate_name_to_oid(
       'object-name' => $name,
       ['need-cdef-entry' => 0/1]
       );

=item Arguments

=over

=item object-name

(Mandatory)
It can be a simple object name (Eg: aggrName) or the object name specified along
with partially qualified MIB path (Eg: fileSys_aggrTable_aggrEntry_aggrName).

=item need-cdef-entry

(Optional)
If set to 1, then a hash reference will be returned as output in which cdef
entry will be returned. Default is 0.

=back

=item Returns

Numeric OID if 'need-cdef-entry' is not set (default) or a hash reference
whose keys are 'oid' and 'node' and the correspomding values are numeric OID
and CDEF entry.

=back

=head2 translate_oid_to_name

Translates the numeric OID to object name.

=over

=item Synopsis

$output = $Api_Set_Obj->translate_oid_to_name(
       'oid' => $oid,
       ['need-cdef-entry' => 0/1]
       );

=item Arguments

=over

=item oid

(Mandatory)
Numeric OID which needs to be translated of the form
MIB_NAME::Objectname.Instance

=item need-cdef-entry

(Optional)
If set to 1, then a hash reference will be returned as output in which cdef
entry will be returned. Default is 0.

=back

=item Returns

Object name if 'need-cdef-entry' is not set (default) or a hash reference
whose keys are 'object' and 'node' and the correspomding values are object name
and CDEF entry.

=back

=cut

=head1 AUTHOR/MAINTAINER

=over

=item NACL Development (dl-nacl-dev@netapp.com)

=back

=cut




