
#
# Copyright (c) 2001-2010 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary API Set for cluster mode ZAPI commands
## @author nsesha@netapp.com, dl-nacl-dev@netapp.com
## @status Shared
## @pod here

package NACL::APISet::Node::ZAPI::CMode;

#
# Compiler Directives
#
use strict;
use warnings;

# Module import - APISets
use NACL::APISet::Session::ZAPI::CMode;
use NACL::APISet::_Mixins::ZAPI qw(_get_xpathcontext);

# Exceptions imported
use NACL::APISet::Exceptions::CDEFNotFoundException qw(:try);

#
# NATE imports
#
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Params::Validate qw(:all);
use Data::Dumper;

# Base class
use base qw(NACL::APISet::Node::ZAPI);

# Constructor
sub new {
    $Log->enter() if $may_enter;

    # an instance of "this" is always the first arg
    # the second arg is a set of named arguments used to init
    # a hash.
    my $class = shift;
    my %args  = @_;
    my $self  = {};

    $self = NACL::APISet::validate_params(
        $class,
        params => \%args,
        spec   =>
          { version => { type => SCALAR, optional => 1, default => 99.99 }, }
    );

    # Bless the hash returned which will make it a Perl object
    bless $self, $class;

    # And give the caller the reference to the bless object.
    $Log->exit() if $may_exit;
    return $self;
}

##############################################################################
# Method:    set_vserver
# Objective: To set the current vserver context
# Details:   Refer to POD documentation at the end of the file.
##############################################################################

sub set_vserver 
{
    $Log->enter() if $may_enter;
    my $self = shift();
    my %args = NACL::APISet::validate_params
      ($self,
       params => \@_, 
       spec   => {"vserver-name" => {type => SCALAR}}
      );
    $self->{session}->set_vfiler_or_vserver(server_name => $args{"vserver-name"});    
    $Log->exit() if $may_exit;
}

##############################################################################
# Method:    get_vserver
# Objective: Get the current vserver context
# Details:   Refer to POD documentation at the end of the file.
##############################################################################

sub get_vserver 
{
    $Log->enter() if $may_enter;
    my ($self, %args) = @_;
    $Log->exit() if $may_exit;
   return $self->{session}->get_vfiler_or_vserver();
}

##############################################################################
# Method:    clear_vserver
# Objective: To set the default vserver context
# Details:   Refer to POD documentation at the end of the file.
##############################################################################

sub clear_vserver 
{
    $Log->enter() if $may_enter;
    my ($self, %args) = @_;
    $self->{session}->clear_vfiler_or_vserver();
    $Log->exit() if $may_exit;
}

##############################################################################
#                               Private Methods
##############################################################################

##############################################################################
# Method:    _init_API
# Objective: Initialize the CMode API Set
# Details:   Initializes the API Set.  This method drives the workflow needed 
# to make the API Set ready for work.  This is accomplished by calling the 
# parents' init_API method and performing work specific to this class afterwards.
##############################################################################

sub _init_API {
    $Log->enter() if $may_enter;

    my ( $self, %args ) = @_;
    
    my $path = $self->{cdef_path};
    $self->{cdef_path} = NATE::Inc::find_file($self->{cdef_path});

    unless (defined $self->{cdef_path}) {
        NACL::APISet::Exceptions::CDEFNotFoundException->throw(
            "Error in finding CDEF directory $path and the reason: \n$@");
    }
    
    $self->SUPER::_init_API( %args, set => $self->{actual_set});
    
    # Set the vserver name if provided irrespective of the shell_type
    if ($self->{actual_set} eq "Vserver") {
        my $vserver_name = defined ($self->{context_definition}->{name})? 
            $self->{context_definition}->{name}:
            (defined $self->{context_definition}->{schema_hostobj}?
            $self->{context_definition}->{schema_hostobj}->hostname(): "");
        $self->set_vserver("vserver-name" => $vserver_name)
            if(length $vserver_name);
    }
    
    $Log->exit() if $may_exit;
}

##############################################################################
# Method:    fpolicy_policy_scope_modify 
# Objective: Overriden method for when the "fpolicy_policy_scope_modify" 
#            method is invoked
# Details:   The cdefs for "fpolicy-policy-scope-modify" has multiple entries
#            for "file-extensions-to-include" and  "volumes-to-include" options.
#            Based on the build used, this method removes the duplicate entry 
#            from the cdef.
##############################################################################
# NOTE: This is not required. This is just left behind as an example of how
# to use _set_correct_option() should it be needed in the future.
# sub fpolicy_policy_scope_modify {
    # my ( $self, %args ) = @_;
    # $Log->enter() if $may_enter;

    # my %opts;
    # Pass all the options required to select the right option. 
    # my $zapi_command = "fpolicy-policy-scope-modify";
    # my @old_typedef = ("file-extension", "volume-name");
    # my @new_typedef = ("string", "string");
    # my @option_name = ("file-extensions-to-include", "volumes-to-include");
    
    # Call helper routine to remove duplicate options.
    # $self->_set_correct_option( zapi_command => $zapi_command,
                                # old_typedef  => \@old_typedef,
                                # new_typedef  => \@new_typedef,
                                # option_name  => \@option_name,
                                # %args,
                            # );
    # my $obj = $self->SUPER::fpolicy_policy_scope_modify(%args);
    
    # $Log->exit() if $may_exit;

    # return $obj;

# }

# Helper routine to remove duplicate entries
# All duplicate entries have been cleaned up at the moment, but this method
# is deliberately left behind should it be needed in the future again.
sub _set_correct_option {
    my ( $self, %args ) = @_;
	$Log->enter() if $may_enter;
    my %opts = NACL::APISet->validate_params(
        params => \%args,
        spec   => { "zapi_command" => { type => SCALAR, },
                    "old_typedef" => { type => ARRAYREF, },
                    "new_typedef" => { type => ARRAYREF, },
                    "option_name" => { type => ARRAYREF, },
                    "vserver-override-name" => {type => SCALAR, optional =>1},
                    # Some commands work only in vserver context
                    # For these commands, system_api_get_elements() should be
                    # invoked in vserver context too. Hence passing the
                    # 'vserver-override-name' which is passed via Components'
        }
    );
    my $zapi_command = $opts{zapi_command};
    my @old_typedef  = @{$opts{old_typedef}};
    my @new_typedef  = @{$opts{new_typedef}};
    my @option_name  = @{$opts{option_name}};
    my ($flag, $system_api_get);
    my  @result;
	
    
    my $cmd = $self->get_command_definition('command' => $zapi_command); 
    
    my $node = $self->_get_xpathcontext(xmlString => $cmd->zapi_command_str());
    
    # Only if both old option and new option 
    # is present, we will do further processing of removing the duplicates.
    for (my $i=0; $i < @option_name; $i++) {
    
        if ( scalar($node->findnodes("//$old_typedef[$i]")) and 
             scalar ($node->findnodes("//$new_typedef[$i]")) ) {
            my $command;
    
            # Execute 'system-api-get-elements' to figure out which data type is
            # supported by the build.   
            if (!$system_api_get) {
	            try {
                    my %overrides;
                    $overrides{"vserver-override-name"} = $opts{"vserver-override-name"} if (defined $opts{"vserver-override-name"});
                    my $res1 = $self->system_api_get_elements('api-list' => [$zapi_command], %overrides);
                    my @res = @{$res1->get_parsed_output()};
                    @result = @{$res[0]->{'api-entries'}[0]->{'system-api-entry-info'}[0]->{'api-elements'}[0]->{'system-api-element-info'}};
		
                } catch NACL::APISet::Exceptions::ResponseException with {
            
                    # Catch any exception thrown while execution of 'system-api-get-elements'
                    # Default to the latest option name. 
            
                    my $exception = shift;
                    $Log->debug( "Error in executing 'system-api-get-elements': "
                        . $exception->text() );
                    $flag = 1;
                };
	            $system_api_get = 1;
            } 	
            if ( $flag ) {
                $Log->debug( "Using the default option $new_typedef[$i]\n'");
                $command = $old_typedef[$i];
            } else {
                foreach my $result (@result) {    
                    if ( $result->{name} eq $option_name[$i]) {
                        $command = $result->{type};
                        last;
                     }
                }

                # The data type obtained should be retained in the cdef, and
                # the duplicate should be removed.
                $command = $command =~ /$old_typedef[$i]/ ? $new_typedef[$i] : $old_typedef[$i];
            }
            my @nodes = $node->findnodes("//$command");
            my $p_parent;
            foreach my $node (@nodes) {
                my $parent = $node->getParentNode();
                $p_parent = $parent->getParentNode();
                $p_parent->removeChild($parent);
            }
            while ($p_parent->toString() !~ /^\s*<$zapi_command>/) {
                $p_parent = $p_parent->getParentNode();
            }

            my $updated_zapi_str = $p_parent->toString();
            $cmd->zapi_command_str($updated_zapi_str);
            $node = $self->_get_xpathcontext(xmlString => $updated_zapi_str);
        }  		
    }
    
    $Log->exit() if $may_exit;
}
=head1 NAME

NACL::APISet::Node::ZAPI::CMode - API Set for cluster mode ZAPI commands

=head1 SYNOPSIS

 use NACL::APISet::Node::ZAPI::CMode;

=head1 DESCRIPTION

A C<NACL::APISet::Node::ZAPI::CMode> is used to connect to and send ZAPI commands to the
cluster interface of a node.

=cut

=head1 EXCEPTIONS

=over

=item NACL::APISet::Exceptions::CDEFNotFoundException

=back

=head1 CONSTRUCTORS

=head2 new

Returns a reference to a newly created C<NACL::APISet::Node::ZAPI::CMode>.

=over 

=item Synopsis

 $api_set_obj = NACL::APISet::Node::ZAPI::CMode->new( %args );

=item Arguments

=over 

=item %args

(Optional)
Parameters passed to parent classes.  See NACL::APISet and NACL::APISet::Node::ZAPI for details.

=back

=item Returns

A reference to a new API Set for Cluster Mode.

=back

=cut

=head1 METHODS

=head2 set_vserver

Specify which vserver in which to run the commands following the call

=over

=item Synopsis

$Api_Set_Obj->set_vserver('vserver-name' => $vserver_name);

=item Arguments

=over

=item vserver-name

Name of the vserver

=back

=item Returns

Nothing.

=back

=head2 get_vserver

Get current vserver context in which we are executing commands

=over

=item Synopsis

$Api_Set_Obj->get_vserver();

=item Arguments

None.

=item Returns

(String)
Current vserver context

=back

=head2 clear_vserver

Reset to executing commands in default vserver context

=over

=item Synopsis

$Api_Set_Obj->clear_vserver();

=item Arguments

None.

=item Returns

None.

=back

=head1 AUTHOR/MAINTAINER

=over

=item NACL Development (dl-nacl-dev@netapp.com)

=back

=cut

1;
