#
# Copyright (c) 2001-2009 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Base Resonse class for all CMode API Sets
## @author olalde@netapp.com, dl-nacl-dev@netapp.com
## @status Shared
## @pod here

package NACL::APISet::Response::CLI::7Mode;

use strict;
use warnings;

use base qw(NACL::APISet::Response::CLI);

use NACL::APISet::Exceptions::ResponseException qw(:try);
use NACL::APISet::Exceptions::ParserNotLoadedException qw(:try);
use NACL::APISet::Exceptions::CommandFailedException qw(:try);
use NACL::APISet::Exceptions::MissingRequiredParamException qw(:try);
use NACL::APISet::Exceptions::InvalidParamException qw(:try);
use NACL::APISet::Exceptions::InvalidParamValueException qw(:try);
use NACL::APISet::Exceptions::FeatureNotLicensedException qw(:try);
use NACL::APISet::Exceptions::InvalidNodescopeNodeException;
use NACL::APISet::Exceptions::FailedAndNotFoundException ();
use NACL::APISet::Exceptions::CommandNotFoundException ();

use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Params::Validate qw(SCALAR);
use Data::Dumper;

sub new {
    $Log->enter() if $may_enter;

    my ($class, %args) = @_;

    # Fill the member vars for this instance
    my $self = NACL::APISet::validate_params(
        $class,
        params => \%args,
        spec   => {
            output           => { type => SCALAR, required => 1},
            command          => { type => SCALAR, required => 1},
            method_name      => { type => SCALAR, required => 1}, 
            cdef_path        => { type => SCALAR, required => 1}, 
            execution_time   => { type => SCALAR, required => 1},
            parser_file      => { type => SCALAR, required => 1},           
            actual_command   => { type => SCALAR, required => 1},
            execution_status => { type => SCALAR, required => 1},
        }
    );

    bless $self, $class;
    
    # Check for any errors due to prefix. Parsers of the actual command
    # are not aware of these errors since in all these errors, the actual
    # command wouldn't have run at all.
    $self->_check_prefix_errors();

    # Populate parsed output at time of invoking API
    $self->_parse_output();
    $self->_throw_exception_based_on_execution_status();
    $Log->debug( Dumper($self) );
    $Log->exit() if $may_exit;
    return $self;
}

my %cmds_hash = (file => {'([Uu]sage:\s(.*)|file: Can\'t find path)' => 'CommandFailed',
                     "Invalid algorithm:" => 'InvalidParamValue'
                    },
            setflag => {'([Uu]sage:\s(.*)|no such flag)' => 'CommandFailed'},
            rm => {'Invalid argument' => 'InvalidParam',
                   '(Is a directory| No such file or directory)'=> 'CommandFailed'
                  },
            sync => {'[Uu]sage:\s(.*)' => 'CommandFailed'},
            slist => {'([Uu]sage:\s(.*)|Invalid path)'=> 'CommandFailed'}, 
            wrfile => {'[Uu]sage:\s(.*)'=> 'CommandFailed', 
                       'Invalid argument' => 'InvalidParam'},
            restore => {'[Uu]sage:\s(.*)' => 'CommandFailed',
                        "RESTORE: Must specify" => 'MissingRequiredParam'},
            mkfile => {'[Uu]sage:\s(.*)' => 'CommandFailed'},
           );

## Internal method - Error check for commands that don't have parser files defined
sub _check_error {
    my $self = shift;
    my $raw_output = $self->get_raw_output();
    my $command = $self->{parser_file};
    $command =~ s/.+\///;
    $command =~ s/\.pm//;
    my $error;
    
    foreach (keys %{$cmds_hash{$command}}) {
        if($raw_output =~ /$_/) {
            $error = $cmds_hash{$command}->{$_};
            $error = 'NACL::APISet::Exceptions::'.$error.'Exception';
            $error->throw(text => $raw_output, response_object => $self);
        }
    }
}


##############################################################################
# Method:    _check_prefix_errors
# Objective: Throws exceptions for errors arising due to prefixing and context
#            support
# Details:   Looks for the error messages which the bruteforce parsers are not
#            aware of. This case will occur when we do prefixing to support
#            vfiler context, modify privilege level, partner context and
#            nodescope context for CMode
##############################################################################

sub _check_prefix_errors {
    $Log->enter() if $may_enter;

    my ($self, %args) = @_;
    my $raw_output = $self->{"output"};
    my $command = $self->{"command"};
    my ($error, $error_msg);
    
    # NOTE: When adding any error messages below, if the error will affect
    # getting the execution status, then the same should be added to the error
    # messages list in _find_execution_status() of Session/CLI/7Mode.pm    
    if ($raw_output =~ /([Vv]filer not found.+|unrecognized privilege.+)/) {
        $error = "InvalidParamValue";
        $error_msg = $1;
    } elsif ($raw_output =~ /(.+requires a multistore license)/) {
        $error = "FeatureNotLicensed";
        $error_msg = $1;
    } elsif ($raw_output =~ /(Not in takeover mode)/) {
        $error = "CommandFailed";
        $error_msg = $1;
    } elsif ($raw_output =~ /(controller failover is not licensed)/) {
        # This catches the error message when controller failover
        # is not licensed
        $error = "FeatureNotLicensed";
        $error_msg = $1;
    } elsif ($raw_output =~ /(partner:\s*Not.*|command not supported.*)/i) {
        # This catches the error message associated with prefixing
        # of partner
        $error = "CommandFailed";
        $error_msg = $1;
    } elsif ($raw_output =~ /Type \'\?\' for a list of commands/) {
        # These are the error messages noticed when certain commands are not
        # supported for vfiler context but available from normal vfiler0 
        # context.
        if ($raw_output !~ /result not found/) {
            # When connecting to vfiler shell directly through persay
            # connection, result -l is not a supported command.
            $error = "FailedAndNotFound";
            $error_msg = "No such command exist";
        }
    } elsif (($raw_output =~ /The following commands are available;/) && 
        ($command !~ /\bhelp\b/)) {
        # _check_prefix_errors considers a command as a help command if it
        # finds the word help anywhere in the command. There are a few commands
        # which displays the message "The following commands are available;"
        # for help command. So we want to give the choice of determining whether 
        # the command is a help command or not to the individual parsers.
        $error = "FailedAndNotFound";
        $error_msg = "No such command exist";
    } elsif ($raw_output =~ /E(?:(?i)RROR):\s(.*)/) {
        # Error message obtained when an invalid value is sent for nodename
        # in Nodescope context.
        $error_msg = $1;
        if($error_msg =~ /invalid value for field "-node <nodename>"/) {
            $error = 'InvalidNodescopeNode';
        }
    }
    
    # Exception will be thrown here itself for all errors arising
    # due to prefixing commands.
    if ($error) {
        $error_msg =~ s/\cM//g;
        $Log->exit() if $may_exit;
        $error = 'NACL::APISet::Exceptions::' . $error . 'Exception';
        $error->throw( text => $error_msg, response_object => $self );
    }
    
    $Log->exit() if $may_exit;
}
##############################################################################
# Method:    _throw_exception_based_on_execution_status
# Objective: Throw exceptions for errors arising due to non-zero and non-255
#            execution status
# Error classification is based on:
# #http://web.netapp.com/engineering/rlse/DOT/mainN/ontap/bedrock/obj/common/files/man/http/man/man1h/na_result.1h.#html
##############################################################################
sub _throw_exception_based_on_execution_status {
    $Log->enter() if $may_enter;
    my $self = $_[0];
    my %error_status_exception_map = (
        1 => {text      => 'Unspecified Failure',
              exception => 'NACL::APISet::Exceptions::CommandFailedException',
        },
        2 => {text      => 'Usage Error',
              exception => 'NACL::APISet::Exceptions::CommandFailedException',
        },
        3 => { text      => 'Improper command argument',
               exception => 'NACL::APISet::Exceptions::InvalidParamException',
        },
        127 => {
            text      => 'Command not found',
            exception => 'NACL::APISet::Exceptions::CommandNotFoundException',
        },
    );
    my $exec_status = $self->get_execution_status();
    my $len = length($exec_status);
    my $raw_output  = $self->get_raw_output();
    if ( exists $error_status_exception_map{$exec_status} ) {
        $Log->exit() if $may_exit;
        $error_status_exception_map{$exec_status}->{exception}->throw(
               text => $error_status_exception_map{$exec_status}->{text} . "\n"
                   . $raw_output,
               response_object => $self
        );
    } elsif ( $exec_status !~ /^0$|^255$/ && $len > 0  ) {
        $Log->exit() if $may_exit;

      #Throwing a generic exception if error status is not 0,1,2,3,127 or 255.
        NACL::APISet::Exceptions::CommandFailedException->throw(
                                              text => $raw_output,
                                              response_object => $self );
    }
    $Log->exit() if $may_exit;
}

1;

=head1 NAME

NACL::APISet::Response::CLI::7Mode - 7mode CLI Response class

=head1 SYNOPSIS

 use NACL::APISet::Response::CLI::7Mode;

=head1 DESCRIPTION

A C<NACL::APISet::Response::CLI::CMode> object is used to recieve, parse and hold 
the output from 7mode CLI API methods.

=head1 CONSTRUCTORS

=head2 new

Returns a reference to a newly created C<NACL::APISet::Response::CLI::7Mode>.  

=over 

=item Synopsis

 $response = NACL::APISet::Response::CLI::7Mode->new( %args );

=item Arguments

=over 

=item output

The raw output from the command.  This can be parsed into an array of key value pairs.

=item command

The command invoked on the service

=item method_name

Name of the method invoked.

=item cdef_path

Path to the directory containing the Freebsd CDEF files

=item execution_time

Time taken to execute the command

=item execution_status

Execution status of the command

=item parser_file

The name of the parser file

=item actual command

The command as translated by the schema.


=back

=back

=cut

