# $Id$

#
# Copyright (c) 2001-2009 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Response class for BARD API Sets
## @author dl-mei-nacl-dev@netapp.com
## @status Shared
## @pod here

package NACL::APISet::Response::CLI::DFM::DFM6;

#
# Compiler Directives
#
use strict;
use warnings;

use base qw(NACL::APISet::Response::CLI::DFM);

#
# NATE imports
#
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Params::Validate qw(:all);
use Data::Dumper;

sub new {
    $Log->enter() if $may_enter;
    my ( $class, %args ) = @_;

    # Fill the member vars for this instance
    my $self = NACL::APISet::validate_params(
        $class,
        params => \%args,
        spec   => {
            output           => { type => SCALAR },
            command          => { type => SCALAR },
            execution_time   => { type => SCALAR },
            execution_status => { type => SCALAR },
            method_name      => { type => SCALAR },
        }
    );

    # This tags the hash pointed to by $self as a
    # type of DFM
    bless $self, $class;

    # Creating a new key "processed_output" for methods that need to modify
    # the command output while retaining the original output in the "output"
    # key
    my @output_array = split( "\n", $self->{output} );

    # Removing warning messages from the output before copying
    for ( my $i = 0; $i < @output_array; $i++ ) {

        if ( $output_array[$i] =~ /^#Warning/ ) {
            $Log->debug(
                "Removing from the output, line: " . shift(@output_array) );
            $i--;
        } else {
            last;
        }
    }
    $self->{processed_output} = join( "\n", @output_array );

    # Strip command echo from raw output.
    $self->SUPER::_remove_echo();

    # Check the raw output for errors. In the case of an error, an exception
    # is thrown.
    $self->SUPER::_check_for_errors() if ( $self->{execution_status} );

# DFM contains a couple of generic parsers, which are capable of parsing
# most commands' output. However, they cannot parse (or incorrectly)
# parse output of certain commands. Command-Specific parsers for those
# commands have been written here (whose method names are the same as the
# name of the command). If it does not find an implementation for the
# method, it means that the generic parsers can be used for the command. In
# such case, it hits AUTOLOAD, which in turn calls corresponding parser methods
# based on the command
#

    my $method_name = "_" . $self->{method_name};

    # Do not parse if we determine it's a help command. If it's a help command
    # then the parsed output should be a hash with a single key (help_output)
    # containing the entire output  verbatim. Hence, parsing should not be
    # done if it's a help command.

   # The above guideline is currently followed for CMode. However , the output
   # of the help command in DFM is structured and can be parsed into a
   # meaningful data structure facilitating easy access of relevant fields
   # in the help.

    $self->$method_name() unless $self->SUPER::_is_help_command();

    $Log->debug( Dumper($self) );

    $Log->exit() if $may_exit;
    return $self;
}

##############################################################################
# Method:    AUTOLOAD
# Objective: Call corresponding parser methods based on the command type
#            if a command specific parser is not found
#
##############################################################################
sub AUTOLOAD {
    my $self    = shift;
    my $command = $self->{command};
    my $output  = $self->{processed_output};
    $Log->enter() if $may_enter;

# On the basis of the "type" of a command , a corresponding parser is invoked.
# By default , the command output is returned as the parsed output with the
# corresponding key "output" .

    if ( $command =~ /list|view|status/i ) {
        ( $command =~ /list.*-l/ )
            ? $self->SUPER::_parse_get_output()
            : $self->SUPER::_parse_table_output();

    } elsif ( $command =~ /detail|get|summary/i ) {
        $self->SUPER::_parse_get_output();
    } elsif (
        $command =~ /add|delete|discover|set|create|modify|pull|import/i )
    {
        $self->SUPER::_parse_field_output();
    } elsif ( $command =~ /set$/i ) {

        #remove the first line of output

        my @line_output = split( /\n/, $output, 2 );
        $self->{processed_output} = $line_output[1];
        $self->SUPER::_parse_get_output();

    } elsif ( $command =~ /service/i ) {
        $self->SUPER::_process_service_output();
    } else {

        $self->{parsed_output} = [ { 'output' => $output } ];
    }

    $Log->exit() if $may_exit;
}

# This is to prevent scripts from failing with ugly error messages like this:
# Method 'NACL::APISet::Response::CLI::DFM:DESTROY' not defined on instance
#'NACL::APISet::Response::CLI::DFM=HASH(0x3c36760)' at ....
# The issue is essentially that if an AUTOLOAD is written, it will end up
# catching DESTROY (which is Perl's built-in implicitly called destructor).
# So, at runtime, when the class tries to define itself completely, including
# DESTROY, it fails.
# Note that this is a runtime error and occurs intermittently (basically as and
# when it pleases) so this error could be difficult to replicate. But,
# for any class with an AUTOLOAD, this error _could_ happen, so it's best
# we follow this convention of defining an empty DESTROY for such classes

# Note that an alternative mechanism of handling this is to have a condition
# in the AUTOLOAD like
# if ($AUTOLOAD eq "DESTROY") {
#     return; }

sub DESTROY { }

##############################################################################
#                               Private Methods
##############################################################################

# Command specific parsers come here
# These are for commands whose outputs cannot be parsed using the generic
# parsers
##############################################################################
# Method:    _dfm_run_status
# Objective: Command specific parser for "dfm run status" command.
# Details:   This is the parser for the 'dfm run status' command.
#	         This command process the output for run status which is quite
#			 different
#
##############################################################################
sub _dfm_run_status {
    $Log->enter() if $may_enter;
    my $self = shift;
	
	my ( $found, $taskArray, $eachTask, @jobsArray, $eachJob );
	my @rawIDs = (
		'Job ID',
		'Status',
		'Submitted',
		'Completed',
		'Task',
	);
	$found = 0;
	$taskArray = [];
	$eachJob = {};
	undef @jobsArray;
	my $processedOutput = $self->{'processed_output'};
	my @lines = split( /\n/ , $processedOutput );
	for ( my $i = 0 ; $i <= $#lines; $i++ ){
	
		#Check if there is a new line or end of the raw output... So we process and add the data
		if ( $lines[$i] =~ /^\s*$/ || $i == $#lines){
			my @jobKeys = keys( %{$eachJob} );
			if( $#jobKeys >= 0 ){
				if ( $#{@{$taskArray}} >= 0 ){
					$eachJob->{ 'Task' } = $taskArray;
					$taskArray = [];
				}
				push ( @jobsArray, $eachJob );
				$eachJob = {};
			}
		}
		else{
			foreach my $rawID ( @rawIDs ){
				if ( $lines[$i] =~ /^\s*$rawID\s*:\s*([^\r\n]*)/i ){
					if ( $rawID =~ /task/i ){
						$i++;
						$eachTask = {};
						for ( my $j = $i ; $j <= $i+2; $j++ ){
							$eachTask->{ $1 } = $2 if ( $lines[$j] =~ /^\s*(.*)\s*:\s*([^\r\n]*)/i );
						}
						push( @{$taskArray}, $eachTask );
					}
					else{
						$eachJob -> { $rawID } = $1;
					}
				}
			}
		}		
	}
	
	$self->{parsed_output} = \@jobsArray;
	$Log->exit() if $may_exit;
}

# Command specific parsers come here
# These are for commands whose outputs cannot be parsed using the generic
# parsers
##############################################################################
# Method:    _dfm_role_details
# Objective: Command specific parser for "dfm role details" command.
# Details:   This is the parser for the 'dfm role details' command.
#	         This command gives a tabular output by default; but on specifying the
#            role-name at the end, it gives the output as a partial key-value pair, partila list
#            and partial table output
#
##############################################################################

sub _dfm_role_details {
    $Log->enter() if $may_enter;
    my $self = shift;

    $self->_parse_role_list_type_output(

        'list_output_type'  => ["Inherited Roles"],
        'table_output_type' => ["Capabilities"],
    );

    $Log->exit() if $may_exit;
}
#####################################################################################################
# Method:    _parse_role_list_type_output
# Objective: Command specific parser for output of type "dfm role list <role-name>" command.
# Details:   The output contains partial get type output, partial list type, partial table output,
# Input:   Two array references:
#          'list_output_type' contains keys which are followed by list type of output
#          'table_output_type' contains keys which are followed by table type of output
#
#####################################################################################################

sub _parse_role_list_type_output {

    $Log->enter() if $may_enter;

    my $self = shift;
    my %args = @_;

    my ( $output, $temp, $i );
    my ( @output_array, @hash_array, @output_line, @partial_output_array )
        = ();
    my $temp_hash = {};
    my @arr;
    $output = $self->{'processed_output'};

    @output_array = split( /\n/, $output );

    my @list_output_sections = @{ $args{'list_output_type'} };

    my @table_output_sections = @{ $args{'table_output_type'} };

    for ( $i = 0; $i < @output_array; $i++ ) {

        $temp = $output_array[$i];
        @output_line = split( ':', $temp );

        unless ( grep ( /$output_line[0]/, @table_output_sections ) ) {

            if ( grep ( /$output_line[0]/, @list_output_sections ) ) {

                for ( $i++; $i < @output_array; $i++ )
                {
					if ($output_array[$i] =~ /(.*):(.*)/) {
					last;
					}
					else {
					$output_array[$i] =~ s/^\s*(.*)\s*$/$1/;
                    push( @arr, $output_array[$i] );
					}

                }
                $i--;
                $temp_hash->{ $output_line[0] } = \@arr;

            } else {
                $output_line[1] =~ s/^\s*(.*)\s*$/$1/
                    ;    # Remove leading and trailing spaces
                $temp_hash->{ $output_line[0] } = $output_line[1];
            }
        } else {
            @partial_output_array = ();
            for ( $i++;$i < @output_array; $i++)
            {
				if ($output_array[$i] =~ /(.*):(.*)/) {
				last;
				}
				else {
                push( @partial_output_array, $output_array[$i] );
				}

            }
            $self->{processed_output} = join( "\n", @partial_output_array );
            $self->_parse_table_output();

            $temp_hash->{ $output_line[0] } = $self->{parsed_output};
            $i--;
        }
    }

    push( @hash_array, $temp_hash );

    $self->{parsed_output} = \@hash_array;
    $Log->exit() if $may_exit;

}

1;
##############################################################################
#                                    POD here
##############################################################################

=head1 NAME

NACL::APISet::Response::CLI::DFM::DFM6 - DFM6 CLI Response class

=head1 SYNOPSIS

 use NACL::APISet::Response::CLI::DFM;

=head1 DESCRIPTION

A C<NACL::APISet::Response::CLI::DFM::DFM6> object is used to receive, parse and hold 
the output from DFM CLI API methods.

=head1 ATTRIBUTES

=over

=item output

Raw command output

=item parsed_output

Output parsed into a standardized array of hashes structure

=item not_parsed

Lines of output not parsed - these lines would include info lines

=item command

Entire command line invoked on the DFM filer

=item execution_time

Time taken to execute the command (in seconds)

=item method_name

The base command invoked

=back

=head1 EXCEPTIONS

=over

=item NACL::APISet::Exceptions::ResponseException

=item NACL::APISet::Exceptions::MissingRequiredParamException

=item NACL::APISet::Exceptions::InvalidParamException

=item NACL::APISet::Exceptions::InvalidParamValueException

=item NACL::APISet::Exceptions::UnsupportedOptionException 

=back

=item 

=head1 CONSTRUCTORS

=head2 new

Returns a reference to a newly created C<NACL::APISet::Response::CLI::DFM::DFM6>.  

=over 

=item Synopsis

 $response = NACL::APISet::Response::CLI::DFM::DFM6->new( %args );

=item Arguments

=over 

=item output

The raw output from the command.  This can be parsed into an array of key value pairs.

=item CLI_separator

The ngsh string used to separate the fields of the output.

=item command

This is the full command line invoked

=item execution_time

The time taken to execute the command (in seconds)

=item method_name

The name of the method invoked.
E.g: system_node_show

=item parsed_output

A reference to parsed output - which is an array of key-value pairs.

=item not parsed

Reference to an array of lines not parsed

=back

=back

=head1 METHODS



=item Synopsis

$raw_output = $response->get_raw_output();

=back



=item Synopsis

$raw_time = $response->get_raw_execution_time();

=back

=item 

=head1 AUTHOR/MAINTAINER

=over

=item NACL Development (dl-mei-nacl-dev@netapp.com)

=back

 $Id$
 Last revised $Date$ by $Author$

=cut

