#
# Copyright (c) 2015,.. NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Base Response class for MarsMaint APISet
## @author nsesha@netapp.com, dl-nacl-dev@netapp.com
## @status Shared
## @pod here

package NACL::APISet::Response::CLI::MarsMaint;

use strict;
use warnings;

use base qw(NACL::APISet::Response::CLI);

use NACL::APISet::Response::CLI::ParserUtils;

# NATE imports
use NATE::Log qw(log_global);
my $log = log_global();
my $may_enter = $log->may_enter();
my $may_exit  = $log->may_exit();

# Exceptions
use NACL::APISet::Exceptions::ResponseException qw(:try);
use NACL::APISet::Exceptions::CommandFailedException qw(:try);

# CPAN stuff
use Data::Dumper;
use Params::Validate qw(:all);

sub new {
    $log->enter() if $may_enter;

    my ( $class, %args ) = @_;

    # Fill the member vars for this instance
    my $self = NACL::APISet::validate_params(
        $class,
        params => \%args,
        spec   => {
            output         => { type => SCALAR|HASHREF, },
            execution_time => { type => SCALAR, },
        }
    );
    bless $self, $class;

    my $output = $self->get_raw_output();
    $self->{processed_output} = $output;
    $self->{complete_raw_output} = $output;

    # check for errors - will die if there is an error
    # TODO: Add more error handling for MarsMaint commands
    #$self->_check_for_errors();

    # parse the output - will be returned in Perl data structure
    $self->_parse_output();

    $log->exit() if $may_exit;
    return $self;
} ## end sub new


##############################################################################
## Method:    get_error_number
## Objective: To get the error number
## Details:   Refer to POD documentation at the end of the file.
###############################################################################
sub get_error_number{
    $log->enter() if $may_enter;

    my $self = shift;

    $log->exit() if $may_exit;

    return defined $self->{error_number} ? $self->{error_number} : "" ;
}


##############################################################################
## Method:    get_error_reason
## Objective: To get the reason for error associated with the execution of CLI
## Details:   Refer to POD documentation at the end of the file.
###############################################################################
sub get_error_reason{
    $log->enter() if $may_enter;

    my $self = shift;

    $log->exit() if $may_exit;

    return defined $self->{error_reason} ? $self->{error_reason} : "" ;
}

##############################################################################
# Method:  _check_for_errors
# Purpose: Check for errors and throw a CommandFailed exception with the
#          relevant error type if an error is found.
##############################################################################

sub _check_for_errors {
    $log->enter() if $may_enter;
    my $self = shift;
    my $output = $self->{complete_raw_output} || '';

    $log->debug( sub { "Raw output is:\n" . Dumper( $output ) } );

    $log->exit() if $may_exit;
} ## end sub _check_for_errors

##############################################################################
# Method:    _parse_output
# Objective: To decode the output which is in JSON format
# Details:  This function will be invoked in the constructor call of the
#           response object. This parses the JSON output into perl data
#           structure.
##############################################################################

sub _parse_output {
    $log->enter() if $may_enter;

    my ($self, %params) = @_;

    my $raw_output = trim_string( string => $self->{output} );
    if ( $raw_output =~ /\{\}$/ ) {
        my @arr = split( /\{\}/, $raw_output);
        $raw_output = @arr[0];
    }

    my $processed_output = $raw_output;

    $self->{complete_parsed_output} = $processed_output;
    $self->{parsed_output} = $processed_output;

    # TODO: Add parsers to output based on command
    $log->exit() if $may_exit;
}

1;

##############################################################################
#                                    POD here
##############################################################################

=head1 NAME

NACL::APISet::Response::CLI::MarsMaint - MarsMaint Response class

=head1 SYNOPSIS

use NACL::APISet::Response::CLI::MarsMaint;

=head1 DESCRIPTION

A C<NACL::APISet::Response::CLI::MarsMaint> object is used to provide parsing and other
methods for MarsMaint class.

=head1 EXCEPTIONS

=over

=item NACL::APISet::Exceptions::ResponseException

=item NACL::APISet::Exceptions::CommandFailedException

=back

=head1 CONSTRUCTOR

Returns a reference to a newly created C<NACL::APISet::Response::CLI::MarsMaint>.

=head1 AUTHOR/MAINTAINER

=over

=item NACL Development (dl-nacl-dev@netapp.com)

=back

=cut

1;
