# Copyright (c) 2016-2017 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Base Resonse class for all ONTAPSelectDeploy API Sets
## @author dl-nacl-dev@netapp.com
## @status Shared
## @pod here

package NACL::APISet::Response::CLI::ONTAPSelectDeploy;

use strict;
use warnings;

use base qw(NACL::APISet::Response::CLI);
use NATE::Log qw(log_global);
use NACL::APISet::Response::CLI::ParserUtils;
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Params::Validate qw(:all);
use Data::Dumper;
use NACL::APISet::Exceptions::InvalidParamException;
use NACL::APISet::Exceptions::NoMatchingEntriesException;
use NACL::APISet::Exceptions::ServiceException;
use JSON::PP;


sub new {
    $Log->enter();

    # an instance of "this" is always the first arg
    # the second arg is a set of named arguments used to init
    # a hash.

    my $class = shift;
    my (%args) = @_;

    # Fill the member vars for this instance
    my $self = NACL::APISet::validate_params(
        $class,
        params => \%args,
        spec   => {
            output           => { type => SCALAR, required => 1},
            command          => { type => SCALAR, required => 1},
            method_name      => { type => SCALAR, required => 1}, 
            cdef_path        => { type => SCALAR, required => 1}, 
            execution_time   => { type => SCALAR, required => 1},
            parser_file      => { type => SCALAR, required => 1},           
            actual_command   => { type => SCALAR, required => 1},
            execution_status => { type => SCALAR, required => 1},
        }
    );

    # This tags the hash pointed to by $object as a
    # type of CMode
    bless $self, $class;

    # Fetching the raw output
    my $output = $self->get_raw_output();
    $self->{"processed_output"} = $output;

    # Check the raw output for errors. In the case of an error, an exception
    # is thrown.
    $self->_check_for_errors();

    # Add more variables that are sub-class specific.

    # Populate parsed output at time of invoking API
    $self->_parse_output();
	
	# check error and throw an exception if the parser hasn't already defined.
	if(!defined($self->{parsed_output})) {
		NACL::APISet::Exceptions::CommandFailedException->throw( 
        text => $self->get_raw_output(), response_object => $self)
        if ($self->get_execution_status());
	}
	
    # And give the caller the referece to the bless object.
    $Log->debug( Dumper($self) );
    $Log->exit();
    return $self;
}

##############################################################################
# Method:  _check_for_errors
# Purpose: Check for errors and throw a relevant exception if an error is
#          found.
# Input:   None.
# Output:  None.
# Notes:   If an error occurs while running a command on a ONTAP Deploy client
#          the output will be in one of two forms:
#          1) Output contains Usage: <usage details>
#             This occurs when a command is invoked with none of the required
#             parameters. Though the output itself might not appear like an
#             error message, this will show up ONLY if none of the required
#             parameters are provided. Hence, it is a
#             MissingRequiredParamException.
#          2) The output is "ERROR:(or)Error: <error message>" The "ERROR (or)
#             Error" indicates
#             the occurence of an error (obviously). The error message
#             itself indicates the kind of error that has occured. The logic
#             here is to analyse the error message to determine what kind of
#             error it is and throw an appropriate exception
# Exceptions: The following are the exceptions that this method could throw:
#             1) MissingRequiredParamException
#             2) InvalidParamException
#             3) InvalidParamValueException
#             4) UnsupportedOptionException
#             5) ResponseException
#             6) CommandFailedException
#             7) CommandNotFoundException
##############################################################################

sub _check_for_errors {
    $Log->enter() if $may_enter;
    my $self = shift;
    my $output = $self->{'processed_output'} || '';
    if ( $output =~ /.*\:\s+error\:.*is required/i ) {
        NACL::APISet::Exceptions::MissingRequiredParamException->throw(
            text            => "Required parameters were not passed : $output",
            response_object => $self
        );
    } elsif ( $output =~ /Error\:.*not found|^Error\:.*is invalid|is too short|(Invalid|No such) file or directory|invalid choice|invalid int value|Error\:.*iSCSI target name.*invalid|Error\:.*ErrorGettingISCSITargetStatus/i ) {
	NACL::APISet::Exceptions::InvalidParamValueException->throw(
            text            => "$output",
            response_object => $self
        );
    } elsif ( $output =~ /unrecognized arguments/i ) {
        NACL::APISet::Exceptions::InvalidParamException->throw(
            text            => "$output",
            response_object => $self
        );
    } elsif ( $output =~ /Cannot find connectivity check run with run_id.*|No connectivity checks run yet|No (licenses|clusters|hosts) found|No mediator iSCSI targets found|Error\:.*FileNotFound\: No such file found in the directory/i ) {
	NACL::APISet::Exceptions::NoMatchingEntriesException->throw(
	    text            => "$output",
	    response_object => $self
	);
    } elsif ( $output =~ /^Error:.*HostOpsLimit/i ) {
        NACL::APISet::Exceptions::ServiceException->throw(
            text            => "$output",
            response_object => $self
        );
    }

    $Log->exit() if $may_exit;
}

##############################################################################
# Method:  _parse_output
# Purpose: Parse the provided raw output into a standard structure - array
#          of hashes. The keys of the hashes being the keys and each entry
#          in the array for one row in the output.
##############################################################################

sub _parse_output {
    $Log->enter() if $may_enter;
    my ( $self, %args ) = @_;
    return $self->{parsed_output} = [] if ( $self->{command} =~ /--help|user password modify/ );
    if ( $self->{command} =~ /-f json/ ) {
        $self->_parse_json();
    } else {
   	my $output = $self->{processed_output};
    	my @result_arr;
   	my @arr = get_lines(
                        "raw_output"      => $output,
                        "response_object" => $self
                      );
    	@arr = grep ! /----/, @arr;
    	foreach my $line (@arr) {
            next if ( $line =~ /Field.+Value/i );
            $line =~ s/^(\|\s+)|(\|)$//g;
            my @values = split(/\|/,$line);
            $values[1] =~ s/^\s+|\s+$//g;
            $result_arr[0]->{to_key_name(item => $values[0])} = $values[1];
    	}
    	$self->{parsed_output} = \@result_arr;
   }

   $Log->exit() if $may_exit;
}

sub _parse_json {
        $Log->enter() if $may_enter;
        my ( $self, %args ) = @_;
	my @result;
	my $output_json = $self->{processed_output};
	my $output = decode_json $output_json ;
	if ( ref $output eq 'ARRAY' ) {
	    $self->{parsed_output} = $output ;
	} else {
	    push (@result, $output);
            $self->{parsed_output} = \@result;
	}
	$Log->exit() if $may_exit;
} ##end of sub _parse_json

1;

##############################################################################
#                                    POD here
##############################################################################

=head1 NAME

NACL::APISet::Response::CLI::ONTAPSelectDeploy - ONTAPSelectDeploy CLI Response class

=head1 SYNOPSIS

 use NACL::APISet::Response::CLI::ONTAPSelectDeploy;

=head1 DESCRIPTION

A C<NACL::APISet::Response::CLI::ONTAPSelectDeploy> object is used to recieve, parse and hold 
the output from ONTAPSelectDeploy CLI API methods.

=head1 EXCEPTIONS

None.

=head1 CONSTRUCTORS

=head2 new

Returns a reference to a newly created C<NACL::APISet::Response::CLI::ONTAPSelectDeploy>.  

=over 

=item Synopsis

 $response = NACL::APISet::Response::CLI::ONTAPSelectDeploy->new( %args );

=item Arguments

=over 

=item output

The raw output from the command.  This can be parsed into an array of key value pairs.

=item command

The command invoked on the service

=item method_name

Name of the method invoked.

=item cdef_path

Path to the directory containing the ONTAPSelectDeploy CDEF files

=item execution_time

Time taken to execute the command

=item execution_status

Execution status of the command

=item parser_file

The name of the parser file

=item actual command

The command as translated by the schema.

=back

=back

=head1 AUTHOR/MAINTAINER

=over

=item NACL Development (dl-nacl-dev@netapp.com)

=back

=cut

