# Copyright (c) 2001-2009 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Base Response class for all GUI API Sets.
## @author nraju@netapp.com, dl-nacl-dev@netapp.com
## @status Shared
## @pod here

package NACL::APISet::Response::GUI;

use strict;
use warnings;

use base qw(NACL::APISet::Response);

use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Params::Validate qw(SCALAR HASHREF ARRAYREF UNDEF);

use NACL::APISet::Exceptions::ResponseException qw(:try);
use NACL::APISet::Exceptions::CommandFailedException qw(:try);
use NACL::APISet::Exceptions::MissingRequiredParamException qw(:try);
use NACL::APISet::Exceptions::InvalidParamException qw(:try);
use NACL::APISet::Exceptions::InvalidParamValueException qw(:try);

sub new {
    $Log->enter() if $may_enter;

    my ($class, $api_result) = @_;

    # Fill the member vars for this instance
    my $self =
      NACL::APISet::validate_params(
        $class,
        params => $api_result,
        spec   => {
            output => {type     => SCALAR|HASHREF,
                       optional => 1,
                       default  => "Command completed successfully"},
            method_name   => {type => SCALAR,  required => 1},
            loading_time  => {type => SCALAR|UNDEF,  optional => 1},
            commit_time   => {type => SCALAR|UNDEF,  optional => 1},
            parsed_output => {type => ARRAYREF|HASHREF, default => undef},
            execution_status => {type => SCALAR, required => 1},
            execution_time   => {type => SCALAR, optional => 1, default => 0},}
      );

    if (exists $self->{loading_time}) {
        if (exists $self->{commit_time}) {
            $self->{execution_time} =
              $self->{loading_time} + $self->{commit_time};
        } else {
            $self->{execution_time} = $self->{loading_time};
        }
    }

    bless $self, $class;

    # Check for any errors in the output.
    # catch the errors thrown by base pm files.
    $self->_check_errors();
    $Log->exit() if $may_exit;
    return $self;
} ## end sub new

# this command hash contaning common gui error strings.
my %error_hash = ('Element not found'   => 'CommandFailed',
                  'Invalid entry:'      => 'InvalidParamValue',
                  'Invalid input in'    => 'InvalidParamValue',
                  'Invalid value'       => 'InvalidParamValue',
                  'Invalid argument'    => 'InvalidParam',
                  'not present in grid' => 'CommandFailed',);

## Internal method - Error check for commands that don't have parser files defined
sub _check_errors {
    my $self       = shift;
    my $raw_output = $self->get_raw_output();

    unless ($raw_output =~ /Command completed successfully/) {
        my $error;
        if(ref $raw_output) {
            $raw_output = $raw_output->{'-stacktrace'};
        }
        foreach (keys(%error_hash)) {
            if ($raw_output =~ /$_/) {
                $error =
                  'NACL::APISet::Exceptions::' . $error_hash{$_} . 'Exception';
                $error->throw(text => $raw_output, response_object => $self);
            }
        }
        # if Response string did not match any known exception then
        # throw a Response Exception
        NACL::APISet::Exceptions::ResponseException->throw( text => $raw_output,
            , response_object => $self);
    } ## end unless ($raw_output =~ /Command completed successfully/)
} ## end sub _check_errors

sub DESTROY { }

1;

=head1 NAME

NACL::APISet::Response::GUI - mode agnostic GUI Response class

=head1 SYNOPSIS

 use NACL::APISet::Response::GUI;

=head1 DESCRIPTION

A C<NACL::APISet::Response::GUI> object is used to recieve, parse and hold
the output from GUI methods.

=head1 CONSTRUCTORS

=head2 new

Returns a reference to a newly created C<NACL::APISet::Response::GUI>.

=over

=item Synopsis

 $response = NACL::APISet::Response::GUI->new( %args );

=item Arguments

=over

=item output

The raw output from the command.  This can be parsed into an array of key value pairs.

=item command

The command invoked on the service

=item method_name

Name of the GUI method invoked.

=item execution_time

Time taken to execute the GUI operation.

=item execution_status

Execution status of the command

=item actual command

The command as translated by the schema.

=item loading_time

Time taken to load GUI object.

=item commit_time

Time taken to load and execute the GUI Operations.

=back

=back

=cut

