#
# Copyright (c) 2001-2010 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Base Response class for SNMP APISet
## @author Isshwarya.M@netapp.com, dl-nacl-dev@netapp.com
## @status Shared
## @pod here

package NACL::APISet::Response::SNMP;

use strict;
use warnings;

#
# NATE imports
#
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Params::Validate qw(:all);
use Data::Dumper;
use Tie::IxHash;
use Net::SNMP qw(:asn1 :snmp);

use base  qw(NACL::APISet::Response);

#
# Exceptions
#
use NATE::BaseException qw(:try);
use NACL::APISet::Exceptions::ResponseException qw(:try);
use NACL::APISet::Exceptions::NoMatchingEntriesException qw(:try);
use NACL::APISet::Exceptions::ObjectRetrievalFailed qw(:try);

# constructor
sub new {
    $Log->enter();

    # an instance of "this" is always the first arg
    # the second arg is a set of named arguments used to init
    # a hash.
    my ($class, %args) = @_;

    # Fill the member vars for this instance
    my $self = NACL::APISet::validate_params
       ($class,
        params => \%args,
        spec   => {
            output           => { type => HASHREF|UNDEF, required => 1,},
            asn1_type        => { type => HASHREF|UNDEF, required => 1,},
            command          => { type => HASHREF, required => 1,},
            execution_time   => { type => SCALAR, required => 1,},
            execution_status => { type => SCALAR, required => 1,},
            error_index      => { type => SCALAR, required => 1,},
            error_reason     => { type => SCALAR, required => 1,},
            need_cdef_entry  => { type => SCALAR, required => 0, default => 0},
            schema           => { isa => 'NACL::APISet::Schema::SnmpSchema', 
                required => 1,},
        }
    );
    
    # schema object won't be stored in response object
    my $schema = delete $self->{schema};
    bless $self, $class;
    
    # check for errors
    $self->_check_error();
    
    # Populate parsed output at time of invoking API
    $self->_parse_output(schema => $schema);
    
    # check for SNMP v2c and v3 specific errors like noSuchObject, 
    # noSuchInstance and endOfMibView. This is invoked after
    # parsing the output so that we can get partially parsed output.
    $self->_check_if_object_retrieval_failed();

    # And give the caller the referece to the bless object.
    $Log->debug( Dumper($self) );
    $Log->exit();
    return $self;
}

##############################################################################
# Method:  get_output_not_parsed
# Objective: Return all those lines which have not been parsed.
# Details:   Refer to POD documentation at the end of the file.
##############################################################################

sub get_output_not_parsed {
    $Log->enter();

    my $self = shift;

    $Log->exit();
    return $self->{oids_not_parsed};
}

##############################################################################
# Method:    get_error_index
# Objective: To get the error index associated with the execution of SNMP PDU
# Details:   Refer to POD documentation at the end of the file.
##############################################################################

sub get_error_index {
    $Log->enter();    
    my $self = shift;    
    $Log->exit();
    return $self->{error_index};
}

##############################################################################
# Method:    get_error_reason
# Objective: To get the reason for error associated with the execution of SNMP
# PDU
# Details:   Refer to POD documentation at the end of the file.
##############################################################################

sub get_error_reason{
    $Log->enter();
    
    my $self = shift;
    
    $Log->exit();

    return defined $self->{error_reason} ? $self->{error_reason} : "" ;
}


##############################################################################
#                               Private Methods
##############################################################################

##############################################################################
# Method:    _check_error
# Objective: To check for the presence of error and throw exception
# Details:  This function tries to classify the error and throw a specific
# exception if possible, else ResponseException will be thrown.
##############################################################################

sub _check_error {
    $Log->enter();
    my $self = shift;
    
    # Net::SNMP populates error_reason field from error_index and
    # execution_status fields in all cases where a Response PDU is obtained
    # back. For cases such as where the table is found to be empty (this
    # might occur when user specify invalid instance information or the
    # objects might be 64 bit objects which would not be visible for v1),
    # SNMP populates the error message saying that the table is empty.
    # Hence we use error_reason as the basis to check whether
    # an exception should be thrown or not.
    my $error = $self->{error_reason};
    if (( defined $error) && (length $error)) {
        # Try to classify the error if possible, else throw ResponseException
        if ($error =~ /Requested\s+table\s+is\s+empty\s+or\s+does\s+not\s+exist/i) {
            NACL::APISet::Exceptions::NoMatchingEntriesException->throw(
                text => $error, response_object => $self);
        } else {
            NACL::APISet::Exceptions::ResponseException->throw(
                text => $error, response_object => $self);
        }
    }   
    $Log->exit();
}

##############################################################################
# Method:    _check_if_object_retrieval_failed
# Objective: To check if any of the object retrieval failed
# Details:  Looks for OIDs whose ASN.1 types is obtained to be noSuchObject, 
# noSuchInstance, endOfMibView and throw exception.
##############################################################################

sub _check_if_object_retrieval_failed {
    $Log->enter();

    my ($self, %params) = @_;
    # Check for noSuchObject, noSuchInstance, endOfMibView error cases and
    # throw exception. If user wants to ignore these errors, then based
    # on user input, catch the exception in Node/SNMP.pm later
    my %failed_objects;
    foreach my $oid (keys(%{$self->{asn1_type}})) {
        my $type = $self->{asn1_type}->{$oid};
        if (($type eq NOSUCHOBJECT) || ($type eq NOSUCHINSTANCE) || 
            ($type eq ENDOFMIBVIEW)) {
            $failed_objects{$oid} = snmp_type_ntop($type);
        }
    }
    if (scalar (keys %failed_objects)) {
        my $error_text = "Object retrival failed for these objects:\n".
            Dumper(\%failed_objects);
        NACL::APISet::Exceptions::ObjectRetrievalFailed->throw(
                text => $error_text, response_object => $self);
    }
    $Log->exit();
}

##############################################################################
# Method:    _parse_output
# Objective: To parse the raw SNMP output.
# Details:  This function will be invoked in the constructor call of the
# response object. This parses the variable bindings field in the response PDU
# and for each numeric OID, it places a hash ref in the parsed output having
# information about instance, object_name, processed value, raw value,
# cdef_entry, ASN.1 type information. 
##############################################################################

sub _parse_output {
    $Log->enter();

    my ($self, %params) = @_;
    my $schema = $params{schema};
    
    my $output = $self->get_raw_output();  
    my $command = $self->get_command();
    
    my @result_array = ();
    my @unknown_entries = ();
    if ($command->{command} =~ /^(get|get_next|get_bulk)$/) {
        # The hash entries are not related to each other.       
        foreach my $oid (keys %$output) {
            my $entry;
            try {
                $entry = $self->_parse_entry(oid => $oid, schema => $schema);
            }  catch NATE::BaseException with {
                my $ex = shift;
                $Log->debug("Error in finding the cdef entry: ". $ex->text());
                $entry = $self->_get_base_entry(oid => $oid);
                my $value = $self->_get_processed_value( 
                    value => $output->{$oid}, 
                    asn1_class_tag => $self->{asn1_type}->{$oid});
                $entry->{value} = $value if(defined $value);
                push(@unknown_entries, $oid);
            };
            push (@result_array, $entry);
        }
    } else {
        # snmp_walk. The hash entries are inter-related and so parse the output
        # as a tabular structure.
        my %table;
        tie %table, 'Tie::IxHash';
        my $entry;
        foreach my $oid (keys %$output) {           
            try {
                $entry = $self->_parse_entry(oid => $oid, schema => $schema);
                $table{$entry->{instance}}{$entry->{object_name}} = $entry;
            }  catch NATE::BaseException with {
                my $ex = shift;
                $Log->debug("Error in finding the cdef entry: ". $ex->text());
                push(@unknown_entries, $oid);
                # CDEF entry is not present. So the parsed output of this node
                # will have few fields missing.
                $entry = $self->_get_base_entry(oid => $oid);
                my $value = $self->_get_processed_value( 
                    value => $output->{$oid}, 
                    asn1_class_tag => $self->{asn1_type}->{$oid});
                $entry->{value} = $value if(defined $value);
                # The loaded CDEFs might be outdated from the build which the
                # service is running and there might be newly added columns to
                # the end of the table. We compare the row information 
                # (instance) obtained so far as the raw output is sorted in a
                # columnar fashion to the columns for which CDEF entry doesn't
                # exist and insert them in appropriate hash in the parsed output
                my $row_to_insert; 
                my @known_rows = keys %table;
                map { if($entry->{oid} =~ /$_$/) {$row_to_insert = $_;last;}} @known_rows;
                $table{$row_to_insert}{$entry->{oid}} = $entry;         
            };
        }
        push (@result_array, values(%table));
    }
    $self->{oids_not_parsed} = \@unknown_entries;
    $Log->warn("CDEF entries couldn't be found for the OID(s):\n". 
        Dumper(\@unknown_entries).
        "Hence entries for these OID(s) in parsed output will have only ".
        "'numericOID', 'raw_value' and 'asn1_class_tag' fields defined") 
        if(scalar @unknown_entries);
    
    $self->{parsed_output} = \@result_array;
    $Log->exit();
}

##############################################################################
# Method:    _get_base_entry
# Objective: Get the basic information for each OID in parsed output
# Details: Populate information about numeric OID, raw value, ASN.1 type
# which doesn't require CDEF entry.
##############################################################################

sub _get_base_entry {
    $Log->enter();

    my ($self, %params) = @_;
    my $output = $self->get_raw_output();
    my $oid = $params{oid};
    my %hash;
    $hash{numericOID} = $oid;
    $hash{raw_value} = $output->{$oid};
    $hash{asn1_class_tag} = $self->{asn1_type}->{$oid}.'('.
        snmp_type_ntop($self->{asn1_type}->{$oid}).')';
    $Log->exit();
    return \%hash;
}

##############################################################################
# Method:    _parse_entry
# Objective: To parse each of the OID in the raw output
# Details: Invokes _get_base_entry, and then tries to obtain the cdef entry
# for the OID and populates other fields which needs CDEF support
##############################################################################

sub _parse_entry {
    $Log->enter();
    my ($self, %params) = @_;
    my $oid = $params{oid};
    my $schema = $params{schema};
    my %hash = %{$self->_get_base_entry(oid => $oid)};
    
    my $output = $self->get_raw_output();
    
    # If the asn1_type for this OID says 'noSuchObject, then we don't
    # have to search for the matching node in the loaded MIB.
    if (($self->{asn1_type}->{$oid}) eq NOSUCHOBJECT) {
        # Don't throw exception which will result in logging a warning
        # meassage for unparsed entries. Since it is expected that CDEF
        # entry will not exist for an OID whos value is received as noSuchObject
        # we simply return base entry for this object.
        $Log->exit();
        $hash{value} = $self->_get_processed_value( value => $output->{$oid}, 
            asn1_class_tag => $self->{asn1_type}->{$oid});
        return \%hash;
    }
        
    # Get the CDEF entry. A new look up in loaded CDEFs becomes necessary
    # for all methods other than Get request.
    my $result = $schema->convert_numericOID_to_name(OID => $oid);
    my $node = $result->{node};
    # Process the information
    $hash{object_name} = $node->{name};
    $hash{descriptive_name} = $result->{object};
    $hash{instance} = $self->_get_instance(node => $node, 
        oid => $oid, schema =>$schema );    
    my $processed_value = $self->_get_processed_value(node => $node, 
        value => $output->{$oid}, 
        asn1_class_tag => $self->{asn1_type}->{$oid});
    $hash{value} = $processed_value if (defined $processed_value);
    $hash{cdef_entry} = $node if ($self->{need_cdef_entry});
    $Log->exit();
    return \%hash;
}

##############################################################################
# Method:    _get_instance
# Objective: Get the instance information
# Details: Obtain the instance information by comparing the OID obtained in
# the output and the OID in the cdef_entry obtained for that object.
##############################################################################

sub _get_instance {
    $Log->enter();

    my ($self, %params) = @_;
    my $node = $params{node};
    my $oid = $params{oid};
    my $schema = $params{schema};
    
    # Remove the leading dot in the OID if any
    $oid =~ s/^\.//;
    
    # Remove the actual OID to get the instance part alone
    my $oid_from_cdef = $node->{oid};
    $oid =~ s/^$oid_from_cdef\.?//;
    
    $Log->exit();
    return $oid;
}

##############################################################################
# Method:    _get_processed_value
# Objective: Get the processed value
# Details: Convert the raw value to processed value using the ASN.1 datatype
# information or in some cases using the cdef_entry
##############################################################################

sub _get_processed_value {
    $Log->enter();

    my ($self, %params) = @_;
    my $node = $params{node};
    my $value = $params{value};
    my $type = $params{asn1_class_tag};
    
    my $ret;
    if ($type eq NULL) {
        # 5 represents NULL data type
        $ret = 'NULL';
    } elsif ($type eq TIMETICKS) {
        # Timeticks
        $ret = ticks_to_time($value);
    } elsif ($type eq NOSUCHOBJECT) {
        $ret = 'noSuchObject';
    } elsif ($type eq NOSUCHINSTANCE) {
        $ret = 'noSuchInstance';
    } elsif ($type eq ENDOFMIBVIEW) {
        $ret = 'endOfMibView';
    } elsif (defined $node) {
        # CDEF entry exists
        if ((exists $node->{'syntax'}->{'type'}->{'basetype'}) &&
            ($node->{'syntax'}->{'type'}->{'basetype'} eq 'Enumeration')) {
            # Enumeration datatype
            my %infoHash = %{$node->{'syntax'}->{'type'}};
            delete $infoHash{basetype};
            %infoHash = reverse %infoHash;
            # If the obtained enumeration value is not defined in CDEFs, then this
            # will result in undef and we are not throwing an exception because the
            # CDEF might be as well outdated.
            $ret = $infoHash{$value};
        } else {
            # No procesing is required.
            $ret = $value;
        }   
    } else {
        # CDEF entry doesn't exist and from the asn1_type_tag information also,
        # we don't know how to process the value. Hence return undef as the
        # processed value.
        $ret = undef;
    }
    
    $Log->exit();
    return $ret;
}

##############################################################################
#                                    POD here
##############################################################################

=head1 NAME

NACL::APISet::Response::SNMP - Base SNMP Response class

=head1 SYNOPSIS

 use NACL::APISet::Response::SNMP;

=head1 DESCRIPTION

A C<NACL::APISet::Response::SNMP> object is used to provide parsing and other
methods common to both 7Mode and CMode SNMP response classes. This includes
accessors for getting unparsed output and error index.

=head1 EXCEPTIONS

=over 

=item NACL::APISet::Exceptions::ResponseException

=item NACL::APISet::Exceptions::NoMatchingEntriesException

=item NACL::APISet::Exceptions::ObjectRetrievalFailed

=item NATE::BaseException

=back

=head1 CONSTRUCTORS

=head2 new

Returns a reference to a newly created C<NACL::APISet::Response::SNMP::CMode>
or C<NACL::APISet::Response::SNMP::7Mode>. 

=over 

=item Synopsis

NACL::APISet::Response::SNMP should not be directly invoked.

 $response = NACL::APISet::Response::SNMP::CMode->new( %args );
 or
 $response = NACL::APISet::Response::SNMP::7Mode->new( %args );

=item Arguments

=over 

=item output

Raw SNMP output which is a simple hashref of numeric OID and value pairs.

=item command

This is a hash ref having information about the SNMP method executed. The keys
in the hash ref are 'command' and 'varbindlist' for get, get_nect, get_bulk and
set PDUs, 'maxrepetitions' and 'nonrepeaters' for get_bulk, 'baseoid' for
snmp_walk. Except for the 'varbindlist', all other keys hold scalar value
whereas, 'varbindlist' holds reference to an array having the list of OIDs
specified in the SNMP input request. 

=item asn1_type

It takes hash ref whose keys are OIDs obtained in the response PDU and the
values are ASN.1 datatypes obtained for those OIDs.

=item execution_time

The time taken to execute the SNMP PDU.

=item execution_status

The status returned for the execution of SNMP command or PDU.

=item error_index

The error index associated obtained in the SNMP respanse PDU.

=item error_reason

The reason for error in case if any

=item need-cdef-entry

A boolean option which determines whether cdef_entry should be placed
in the parsed output corresponding to each of the numeric OID obtained in
the raw output.

=item schema

A reference to schema object to parse the output.

=back

=item Returns

A reference to a CMode or 7Mode Response object.

=back

=head1 METHODS

=head2 get_error_index

Returns the error index associated with the invocation of SNMP operation.

=over 

=over

=item Synopsis

 $error_index = $response->get_error_index();

=back
 
=item Returns

The error index obtained on invoking the SNMP command

=back

=head2 get_output_not_parsed

Any SNMP operation would return a hash ref as the raw output. While
parsing, for each OID in the output, we will try to obtain the CDEF entry.
If we couldn't, then the entry (hash ref) corresponding to these OIDs in
the parsed output will have limited fields defined. This method will
return the list of numeric OIDs for which the cdef_entry could not be found
and so the parsing is partial.

=over 

=item Synopsis

 my $oid_list = $response->get_output_not_parsed();

=item Returns

Reference to an array having the list of OIDs which are not parsed.

=back

=head2 get_error_reason

Returns the reason for error in case if any. Orelse it will be an empty string

=over 

=item Synopsis

 $error_reason = $response->get_error_reason();

=item Returns

The reason for error associated with the invocation of SNMP method.

=back

=head1 AUTHOR/MAINTAINER

=over

=item NACL Development (dl-nacl-dev@netapp.com)

=back

=cut


1;
