#
# Copyright (c) 2014 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary APISet Response object for SOAP interfaces
## @author dl-nacl-dev@netapp.com
## @status Shared
## @pod here

package NACL::APISet::Response::SOAP;

use strict;
use warnings;

use base  qw(NACL::APISet::Response);

use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();

use Params::Validate qw(:all);

use Class::MethodMaker [
    scalar  => ['command'],
    scalar  => ['output'],
    scalar  => ['execution_status'],
    scalar  => ['execution_time'],
];

sub new {
    my ($class, %args) = @_;

    my $self = {};
    bless $self, $class;

    foreach my $key (%args) {
        if ($self->can("$key")) {
            $self->$key($args{$key});
        }
    }

    $self->init(%args);

    return $self;
}

sub init {
    # No initialization needed for general case
    # Included so no error thrown
}

##############################################################################
# Method:    _parse_output
# Objective: Returns the parsed output
# Details:   See POD documentation at the end of this file.
##############################################################################
sub _parse_output {
    $Log->enter() if $may_enter;
    my $self = shift;

    my $output = $self->get_raw_output();
    # the call to 'result' may not contain all elements that need to be deserialized from the output
    # we need to call 'paramsout' to get any remaining elements if they exist.
    my @total_parsed_output;
    push(@total_parsed_output,$output->result);

    if ($output->paramsout)
    {
        push(@total_parsed_output, $output->paramsout);
    }

    if ($output->result) {
        $self->_set_parsed_output(parsed_output => \@total_parsed_output);
    } else {
        $self->_set_parsed_output(parsed_output => []);
    }
    $Log->exit() if $may_exit;
}

##############################################################################
# Method:    get_actual_command
# Objective: Returns the actual command
# Details:   See POD documentation at the end of this file.
##############################################################################
sub get_actual_command {
    $Log->enter() if $may_enter;

    my $self = shift;

    $Log->exit() if $may_exit;
    return $self->{command};
}

##############################################################################
# Method:    _set_parsed_output
# Objective: Sets the parsed output to the specified value
# Details:   See POD documentation at the end of this file.
##############################################################################
sub _set_parsed_output {
    my ($self, %args) = @_;

    $Log->enter() if $may_enter;
    $self->{parsed_output} = $args{parsed_output};
    $Log->exit() if $may_exit;
}

##############################################################################
# Method:    get_parsed_output
# Objective: Returns the parsed output
# Details:   See POD documentation at the end of this file.
##############################################################################
sub get_parsed_output {
    my $self = shift;
    $Log->enter() if $may_enter;

    $Log->exit() if $may_exit;
    return $self->{parsed_output};
}

##############################################################################
# Method:    get_raw_output
# Objective: Returns the raw output
# Details:   See POD documentation at the end of this file.
##############################################################################
sub get_raw_output {
    my $self = shift;
    $Log->enter() if $may_enter;

    $Log->exit() if $may_exit;
    return $self->output;
}
1;

##############################################################################
#                                    POD here
##############################################################################

=head1 NAME

NACL::APISet::Response::SOAP - API Set Response base class for CLIs

=head1 SYNOPSIS

use NACL::APISet::Response::SOAP;

=head1 DESCRIPTION

A L<NACL::APISet::Response::SOAP|lib-NACL-APISet-Response-SOAP-pm> contains data and methods common to all API Set
Response SOAP classes. This includes accessors for raw output, actual command invoked,
getting one row of output and logic to parse output for APISets which will have
multiple parsers divided into different packages.

Inherit from this class when creating new API Set Response classes
for SOAP.

=head1 CONSTRUCTOR

=head2 new

Returns a reference to a newly created C<NACL::APISet::Session::CLI::Linux>.

=over

=item Synopsis

my $response_obj = NACL::APISet::Response::SOAP->new( %args );

=item Arguments

=over

=item output

The ouput of the executed command.

=item command

The command to execute.

=item execution_status

The execution_status of the command. Upon command failure, the execution_status
will be the SOAP fault code. Otherwise, it will be '0'.

=item execution_time

The time it took to execute the command.

=back

=item Returns

A reference to a new VPserver SOAP Response class. This is designed to provide common functionality.
An object of this kind should never be instantiated directly.

=back

=head1 Methods

=head2 get_actual_command

This method is used to return the actual command that was executed.

=over

=item Synopsis

$actual_command = $response_object->get_actual_command();

=item Arguments

None.

=item Returns

The actual command that was executed.

=back

=head2 get_parsed_output

This method is used to return the parsed output.

=over

=item Synopsis

$processed_output = $response_object->get_parsed_output();

=item Arguments

None.

=item Returns

The parsed output

=back

=head2 get_raw_output

This method is used to return the raw output of the command executed.

=over

=item Synopsis

$raw_output = $response_object->get_raw_output();

=item Arguments

None.

=item Returns

The raw output of the command executed.

=back


