#
# Copyright (c) 2014 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary APISet Response object for VPserver SOAP interfaces
## @author dl-nacl-dev@netapp.com
## @status Shared
## @pod here

package NACL::APISet::Response::SOAP::VPserver;

use strict;
use warnings;

use base qw(NACL::APISet::Response::SOAP);
use Params::Validate qw(:all);

# NACL Exceptions
use NACL::APISet::Exceptions::InvalidParamException;
use NACL::APISet::Exceptions::CommandFailedException;
use NACL::APISet::Exceptions::CommandNotFoundException;
use NACL::APISet::Exceptions::MissingRequiredParamException;

use NATE::Log qw(log_global);
my $Log = log_global;
my $may_enter = $Log->may_enter();
my $may_exit = $Log->may_exit();

use Class::MethodMaker [
    scalar  => ['command'],
    scalar  => ['execution_time'],
    scalar  => ['execution_status'],
    scalar  => [ { -type => 'SOAP::SOM' }, 'output'],
    scalar  => [ { -type => 'HASH' }, 'parsed_output'],
];

sub init {
    $Log ||= log_global;
    $Log->enter() if $may_enter;

    my ($self, @params) = @_;

    # Validate Params
    my %args = validate_with (
        params => \@params,
        spec   => {
            output           => { type => OBJECT, required => 1},
            command          => { type => SCALAR, required => 1},
            execution_time   => { type => SCALAR, required => 1},
            execution_status => { type => SCALAR, required => 1},
        }
    );

    # check error and throw an exception if there is a fault in the response.
    if($self->output->fault() and $self->get_execution_status) {
        my $output = $self->output;
        if ($output->faultstring =~ /Does it exist in service WSDL/i) {
            NACL::APISet::Exceptions::CommandNotFoundException->throw(
                text => "The command (" . $self->command . ") was not found. Please refer to the VPserver documentation " .
                        "for a list of valid commands. " . "(Fault: " .  $output->faultstring. ")",
                response_object => $self);
        } elsif ($output->faultstring =~ /unmarshalling Error: unexpected element/i or
                 $output->faultstring =~ /null while invoking public/i) {
            NACL::APISet::Exceptions::InvalidParamException->throw(
                text => "One of the parameters for command (" . $self->command. ") was not " .
                        "formatted correctly: \n" . "\t(Fault: " .  $output->faultstring . ")",
                response_object => $self);
        } elsif ($output->faultstring =~ /VPCredential object expected in argument position zero/i) {
            NACL::APISet::Exceptions::MissingRequiredParamException->throw(
                text => "Command, " . $self->command . " failed due to: " . $output->faultstring,
                response_object => $self);
        } else {
            NACL::APISet::Exceptions::CommandFailedException->throw(
                text => "Command, " . $self->command . " failed due to: " . $output->faultstring,
                response_object => $self);
        }
    }

    # Populate parsed output at time of invoking API
    $self->_parse_output();

    $Log->exit() if $may_exit;

    return $self;
}

1;

##############################################################################
#                                    POD here
##############################################################################

=head1 NAME

NACL::APISet::Response::SOAP::VPserver - VPserver SOAP Response class

=head1 SYNOPSIS

 use NACL::APISet::Response::SOAP::VPserver;

=head1 DESCRIPTION

A C<NACL::APISet::Response::SOAP::VPserver> object is used to receive, parse and hold
the output from VPserver SOAP API methods.

=head1 EXCEPTIONS

=over

=item  NACL::APISet::Exceptions::CommandFailedException;

=item  NACL::APISet::Exceptions::CommandNotFoundException;

=item  NACL::APISet::Exceptions::InvalidParamException;

=item  NACL::APISet::Exceptions::MissingRequiredParamException;

=back

=head1 CONSTRUCTORS

=head2 new

Returns a reference to a newly created C<NACL::APISet::Response::SOAP::VPserver>.

=over

=item Synopsis

 $response = NACL::APISet::Response::SOAP::VPserver->new( %args );

=item Arguments

=over

=item command

The command invoked on the service

=item execution_time

Time taken to execute the command

=item execution_status

Execution status of the command

=item output

A SOAP::SOM object representing the raw output from executing the command.
Information for the library can be found here: L<SOAP::SOM|http://search.cpan.org/dist/SOAP-Lite/lib/SOAP/SOM.pod>.

=back

=back

=head1 AUTHOR/MAINTAINER

=over

=item NACL Development (dl-nacl-dev@netapp.com)

=back

=cut

