
# Copyright (c) 2001-2010 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Base Response class for ZAPI APISet
## @author nsesha@netapp.com, dl-nacl-dev@netapp.com
## @status Shared
## @pod here

package NACL::APISet::Response::ZAPI;

use strict;
use warnings;

use base  qw(NACL::APISet::Response);

use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use NACL::APISet::Schema;
use NACL::APISet::_Mixins::ZAPI qw(_get_xpathcontext _xml2perl);
use Params::Validate qw(validate_with SCALAR OBJECT HASHREF);

use NACL::APISet::Exceptions::ResponseException qw(:try);
use NACL::APISet::Exceptions::CommandFailedException;
use NACL::APISet::Exceptions::CommandNotFoundException;
use NACL::APISet::Exceptions::UnexpectedOutputException;
use NACL::APISet::Exceptions::InvalidXMLException;
use NACL::APISet::Exceptions::InvalidParamValueException;

use Data::Dumper;

# Constructor
sub new {
    $Log->enter() if $may_enter;
    my ($class, %args) = @_;

    # Fill the member vars for this instance
    my $self = validate_with(
        params => \%args,
        spec   => {
            output           => {type => SCALAR},
            command          => {type => SCALAR},
            execution_time   => {type => SCALAR},
            execution_status => {type => SCALAR},
            error_number     => {type => SCALAR},
            error_reason     => {type => SCALAR},
            complete_raw_output => {type => SCALAR},
            
            # Certain calls don't provide this, in which case _parse_output
            # calculates this value
            complete_parsed_output => {type => HASHREF, optional => 1}
        }
	);

    # This tags the hash pointed to by $object as a
    # type of CMode
    bless $self, $class;
    
    # Populate parsed output at time of invoking API
    $self->_parse_output();

    # And give the caller the referece to the bless object.
    $Log->debug( sub { "Dumper of $class object:\n" . Dumper($self) } );
    $Log->exit() if $may_exit;
    return $self;
}

##############################################################################
# Method:    _parse_output
# Objective: To parse the raw xml output
# Details:  This function will be invoked in the constructor call of the
# response object. This parses the raw xml output into perl data structure
# similar to the parsed output obtained from using CLI interface.
##############################################################################

sub _parse_output {
    $Log->enter() if $may_enter;

    my ($self, %params) = @_;
    
    my $complete_raw_output = $self->get_complete_raw_output();
    
    # First step is to check whether the result status is failed.
    if ($self->get_execution_status() ne "passed") {
        # Classify the error if possible
        my $error = $self->_classify_error();
        $Log->exit() if $may_exit;
        $error = "NACL::APISet::Exceptions::" . $error;
        $error->throw( text => $self->get_error_reason(),
            response_object => $self);
    }
    
    # If Response->new() is called by Session->send_command, then it
    # would have passed in the complete_parsed_output. However, there are
    # instances of Response->new() being directly called -- in these
    # cases, the parsed output would not have been supplied, so we need
    # to parse the output in that case.
    my $complete_parsed_output = $self->get_complete_parsed_output();
    if (!defined $complete_parsed_output) {
        $complete_parsed_output = $self->_xml2perl(
            xmlString  => $self->get_complete_raw_output(),
            ForceArray => 1,
            KeyAttr    => [],
        );
    }

    my $parsed_output = [];
    my $trace_output;
    
    eval {
        
        if (!$self->isa("NACL::APISet::Response::ZAPI::DBlade")) {
            # get Trace output
            $trace_output = $complete_parsed_output->{trace_output}[0] if (defined $complete_parsed_output->{trace_output});
        }
        
        # To always return a reference to an array so that consistency with
        # CLI response class can be achieved
        if (ref($complete_parsed_output) ne "ARRAY") {
            $complete_parsed_output=[$complete_parsed_output];
        }
        
    };
    _convert($complete_parsed_output);
    
    if($@){
        NACL::APISet::Exceptions::UnexpectedOutputException->throw(
            text => $@, response_object => $self);
    }
    
    if ($self->isa("NACL::APISet::Response::ZAPI::DBlade")) {
        $parsed_output = $complete_parsed_output;
    } else {
        # Set complete output
        $self->{complete_parsed_output} = $complete_parsed_output;
        
        # Set trace output
        $self->{trace_output} = $trace_output;
        
        $Log->debug(sub {Dumper($complete_parsed_output)});
        
        # Get the results
        $parsed_output = $complete_parsed_output->[0]->{results}[0];
    }
    
    if (ref($parsed_output) ne "ARRAY") {
            $parsed_output=[$parsed_output];
    }
    
    # The result tag has the attribute called "status" and that should
    # be removed from parsed output
    # Do this only if "status" is a scalar, else it signifies that the ZAPI
    # output also had a "status" tag. Example: quota-status
    if((exists $parsed_output->[0]->{"status"}) && (ref($parsed_output->[0]->{"status"}) eq "")) {
        delete $parsed_output->[0]->{"status"};
    }
    
    $self->{parsed_output} = $parsed_output;
}

##############################################################################
# Method:  _classify_error
# Purpose: Classify the error if possible orelse throw a Response Exception
# Input:   None.
# Output:  None.
# Notes: This is an attempt to classify the errors in ZAPI if the error message
#        is consistent so that we can throw a more specific exception
# Exceptions: The following are the exceptions that this method could throw:
#             1) ResponseException
#             2) CommandNotFoundException
#             3) InvalidParamValueException
##############################################################################
sub _classify_error {
    $Log->enter() if $may_enter;

    my $self = shift;
    my $error = $self->get_error_reason();
    my $error_no = $self->get_error_number();
    my $exception = "CommandFailedException"; 
    if ($error =~ /Unable to find API:/) {
        $exception = "CommandNotFoundException";
    } elsif ($error_no == 2) { # Element does not exist
        $exception = "InvalidParamValueException";
    } elsif ($error_no == 17) { # Element already exists
        $exception = "InvalidParamValueException";
    }
    
    $Log->exit() if $may_exit;
    return $exception;    
}

##############################################################################
# Method:    _convert
# Objective: Convert only the leaf nodes in the input xml into scalar
# Details:  This function recursively traverses through each of the node till
# a leaf node is recahed and checks whether the leaf node is scalar or an array.
# If it is an array, then it will store the value in leaf node as simple scalar.
##############################################################################
sub _convert {
    my $arr_ref = shift;
    foreach my $hash_ref (@$arr_ref) {
        my @keys = keys(%{$hash_ref});
        foreach my $key (@keys) {
            if (ref($hash_ref->{$key}) eq 'ARRAY') {
                if (
                    # ...the first element is a scalar
                    (!ref($hash_ref->{$key}->[0]) )
                ) {
                    # if the array contains only 1 element, and...
                    if (scalar(@{$hash_ref->{$key}}) == 1) {
                        # convert leaf to a scalar
                        $hash_ref->{$key} = $hash_ref->{$key}->[0];
                    }
                } else {
                    # not a scalar leaf node, convert it
                    _convert($hash_ref->{$key});
                }
            }
        }
    }
}

##############################################################################
# Method:    _register_ns
# Objective: Will register Namespace URI if uri is present in the xml
# Details:  Namespace URI needs to be register before we can search for a node
# in the xml
##############################################################################
sub _register_ns {
    $Log->enter() if $may_enter;
    my ($self, %args) = @_;
    validate_with(
        params => \%args,
        spec   => {
            xpc => {  type => OBJECT, isa => 'XML::LibXML::XPathContext' },
        }
    );
    my $prefix;
    my $xpc = $args{xpc};
    my @nodes = $xpc->findnodes('*');

    my $uri = $nodes[0]->namespaceURI();
    if (defined $uri) {
        $xpc->registerNs('NetAppUriNs', $uri);
        $prefix = "NetAppUriNs";
    }
    
    $Log->exit() if $may_exit;
    return $prefix;
}

##############################################################################
# Method:    get_error_number
# Objective: To get the error number associated with the execution of ZAPI
# Details:   Refer to POD documentation at the end of the file.
##############################################################################

sub get_error_number{
    $Log->enter() if $may_enter;
    
    my $self = shift;
    
    $Log->exit() if $may_exit;
    return defined $self->{error_number} ? $self->{error_number} : "" ;
}

##############################################################################
# Method:    get_error_reason
# Objective: To get the reason for error associated with the execution of ZAPI
# Details:   Refer to POD documentation at the end of the file.
##############################################################################

sub get_error_reason{
    $Log->enter() if $may_enter;
    
    my $self = shift;
    
    $Log->exit() if $may_exit;

    return defined $self->{error_reason} ? $self->{error_reason} : "" ;
}

##############################################################################
# Method:    get_complete_raw_output
# Objective: Returns complete raw command output
# Details:   See POD documentation at the end of this file.
##############################################################################
sub get_complete_raw_output {
    $Log->enter() if $may_enter;
    my $self = shift;

    $Log->exit() if $may_exit;
    return $self->{complete_raw_output};
}

##############################################################################
# Method:    get_complete_parsed_output
# Objective: Returns the complete parsed output
# Details:   See POD documentation at the end of this file.
##############################################################################
sub get_complete_parsed_output {
    $Log->enter() if $may_enter;
    my $self = shift;
    
    $Log->exit() if $may_exit;
    return $self->{complete_parsed_output};
}

##############################################################################
# Method:    get_trace_output
# Objective: Returns the trace output
# Details:   See POD documentation at the end of this file.
##############################################################################
sub get_trace_output {
    $Log->enter() if $may_enter;
    my $self = shift;

    $Log->exit() if $may_exit;
    return $self->{trace_output};
}

1;

##############################################################################
#                                    POD here
##############################################################################

=head1 NAME

NACL::APISet::Response::ZAPI - Base ZAPI Response class

=head1 SYNOPSIS

 use NACL::APISet::Response::ZAPI;

=head1 DESCRIPTION

A C<NACL::APISet::Response::ZAPI> object is used to provide parsing and other
methods common to both 7Mode and CMode ZAPI response classes. This includes
accessors for raw output, error number and error reason.

=head1 EXCEPTIONS

=over 

=item NACL::APISet::Exceptions::ResponseException

=item NACL::APISet::Exceptions::UnexpectedOutputException

=item NACL::APISet::Exceptions::InvalidXMLException

=back

=head1 CONSTRUCTOR

None. This class is present only to provide common functionality, an object of 
this class should never be instantiated. We should only instantiate objects
of Response sub-classes.

=head1 METHODS

=head2 get_error_number

Returns the error number obtained on invoking the ZAPI command.

=over 

=over

=item Synopsis

 $error_no = $response->get_error_number();

=back
 
=item Returns

The error number returned on invoking the ZAPI command

=back

=head2 get_error_reason

Returns the reason for error in case if any. Or else it will be an empty string

=over 

=over

=item Synopsis

 $error_reason = $response->get_error_reason();

=back
 
=item Returns

The reason for error in the invoked ZAPI command

=back

=head2 get_complete_parsed_output

Returns the complete parsed output. Or else it will be empty array ref. This method is not aplicable for DBlade ZAPI

=over 

=over

=item Synopsis

 $complete_parsed_output = $response->get_complete_parsed_output();

=back
 
=item Returns

The complete parsed output of invoked ZAPI command

=back

=head2 get_trace_output

Returns the trace if requested. Or else it will be undef. This method is not aplicable for DBlade ZAPI

=over 

=over

=item Synopsis

 $trace_output = $response->get_trace_output();

=back
 
=item Returns

The trace

=back

=head2 get_complete_raw_output

Returns the complete raw output. Or else it will be undef. This method is not aplicable for DBlade ZAPI

=over 

=over

=item Synopsis

 $complete_raw_output = $response->get_complete_raw_output();

=back
 
=item Returns

The complete raw output of invoked ZAPI command

=back

=head1 AUTHOR/MAINTAINER

=over

=item NACL Development (dl-nacl-dev@netapp.com)

=back

=cut


