# $Id$
#
# Copyright (c) 2007-2008 NetApp Inc.
# All rights reserved.
#
## @summary Provides an interface to get required and optional fields
##          from command definitions
##
## @author dl-nacl-dev, korry@netapp.com, dl-ssan-test
## @status Shared

package NACL::APISet::Schema::Cdef;

# compiler directives
use strict;
use warnings;

# imported modules
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Params::Validate qw( validate_with SCALAR HASHREF ARRAYREF BOOLEAN UNDEF);
use Data::Dumper;
use Tie::IxHash;
use File::Spec::Functions;

# Exceptions
use NATE::Exceptions::Argument qw(:try);

sub _new_validate_spec {
    return {
        command                 => { type => SCALAR },
        option_order            => { type => ARRAYREF },
        concat_limits           => { type => ARRAYREF },
        alias_params            => { type => HASHREF },
        file                    => { type => SCALAR },
        switch_identifier       => { type => SCALAR, optional => 1 },
        is_windows              => { type => BOOLEAN, default => 0 },
        command_sep             => { type => SCALAR, default => ' ' },
        options_sep             => { type => HASHREF, optional => 1 },
        boolean_options         => { type => ARRAYREF, default => [] },
        repeated_options        => { type => ARRAYREF, default => [] },
        special_properties      => { type => HASHREF, optional => 1 },
        alias_map               => { type => HASHREF|UNDEF, optional => 1 },
    };
}

BEGIN {
    my $spec = _new_validate_spec();
    delete $spec->{alias_map};
    my @fields = keys %{ $spec };
    no strict 'refs';
    foreach my $field (@fields) {
        *{"get_$field"} = sub {
            $Log->enter() if $may_enter;
            my $self = shift;
            $Log->exit() if $may_exit;
            return $self->{$field};
        };
    }
    use strict 'refs';
}


sub get_alias_map {
    $Log->enter() if $may_enter;
    my $self = shift;
    my %alias_map;
    if( $self->{alias_map} ) {
        %alias_map =  %{ $self->{alias_map} };
        map { if( length($_) == 1 ) {
                delete $alias_map{$_}
              }
            } keys %alias_map;
    }
    $Log->exit() if $may_exit;
    return \%alias_map;
}

###############################################################################
# Method:    new()
# Objective: Create cdef object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub new {
    $Log->enter() if $may_enter;
    my $class = shift;

    my $self = validate_with(
        params => \@_,
        spec   => _new_validate_spec()
    );

    # bless instance into class
    bless $self, $class;

    $Log->exit() if $may_exit;
    return $self;

} ## end sub new
###############################################################################
# Method:    translate()
# Objective: return cmd syntax for requested system operation
# Details:   See POD documentation at the end of this file.
###############################################################################
sub translate {
    $Log->enter() if $may_enter;
    my ( $self, %args );

    # preserve the order of input fields
    tie %args, 'Tie::IxHash';

    # Tie::IxHash implements Perl hashes that preserve the order in which the
    # hash elements were added. This is necessary for 7Mode since the option
    # order might matter
    ( $self, %args ) = @_;

    #
    # This is the "pass-through" option, which allows a user to append
    # a string to the end of the command built. It should be used for
    # negative testing
    #
    my $pass_through;
    $pass_through = ' ' . $args{'pass_through'}
        if ( defined $args{pass_through} );

    #
    # We delete this entry in args, so that it is not included in the
    # following logic which checks to see what type of parameter
    # an argument is.
    #
    delete $args{'pass_through'};

    my %ordered_args;
    tie %ordered_args, 'Tie::IxHash';

    # Handle processing w.r.t special properties
    my $spl_props    = $self->get_special_properties();
    my $derived_hash = {};
    if ( scalar( keys %$spl_props ) ) {
        %args = $self->_handle_processing_for_special_properties(%args);

        # Delete it so that 'cdef_trans_derived_info' added by
        # _handle_processing_for_special_properties
        # will not end up wrongly translated as command option.
        $derived_hash = delete $args{cdef_trans_derived_info};
    }

    my ( $param, $value, @cmd_inputs, @option_order, @concat_limits );

    my ( $i, $start, $end, $start_flag, $lower_limit, $upper_limit ) = (0);

    # for options to be concatenated if any
    @concat_limits = @{ $self->{concat_limits} };

    # If the option order has been defined, we reorder the args.  This
    # is for 7Mode since the order can matter
    @option_order = @{ $self->get_option_order() };

    if (@option_order) {

        if (@concat_limits) {
            foreach (@option_order) {
                if ( $_ eq $concat_limits[0] ) {
                    $lower_limit = $i;
                } elsif ( $_ eq $concat_limits[1] ) {
                    $upper_limit = $i;
                    last;
                }
                $i++;
            }
        }

        foreach my $key ( keys %args ) {

            # If the option order doesn't have the current key, add it
            unless ( grep /^$key$/, @option_order ) {
                push( @option_order, $key );
            }
        }

        # At this point @option_order should hold the args in the order
        # we want them in.  So we will set the ordered_args hash in
        # this order then set args to ordered args.

        ( $i, $value ) = ( 0, 0 );
        foreach my $key (@option_order) {

            # This conditional is to build the command with only the options
            # users had passed
            if ( defined $args{$key} ) {

                $ordered_args{$key} = $args{$key};
                if (   @concat_limits
                    && $i >= $lower_limit
                    && $i <= $upper_limit )
                {

                    if ( !$start_flag ) {
                        $start      = $value;
                        $start_flag = 1;
                    }
                    if ($start_flag) {

                        $end = $value;
                    }
                }

                $value += 2;
            }
            $i++;
        }
        %args = %ordered_args;
    }

    my $extended_query = delete $args{extended_query};

    # parse command attributes/parameters
    foreach my $key ( keys %args ) {

        # resolve param alias to proper param name.
        # (ex. volume will become -volume)
        $param = $self->_get_alias_root( alias => $key);

        $value = $args{$key};

        # Ignore keys for which value provided is undef
        if ( !defined $value ) {
            $Log->debug( "Key '$key' was specified as undef. Hence it is "
                    . 'being ignored' );
            next;
        }

        if ( $self->_is_boolean_option( key => $key ) ) {

            # If the option is a boolean, then there is no "value"
            # Instead, if the "value" provided is true, then the boolean
            # option must be present in the constructed command
            if ( $value eq "" ) {
                $Log->warn( "Specifying Boolean options as <boolopt> => \"\" "
                        . "is deprecated. The suggested syntax is "
                        . "<boolopt> => 1" );
            } elsif ( $value and $value !~ /false/i ) {
                $value = '';
            } else {

                # when $value is '0' or 0
                $param = '';
                $value = '';
            }
        } elsif ( $self->_is_repeated_option( key => $key ) ) {

            # If the option is repeated, then each of the values will be
            # prefixed by the same switch
            my @values = split( /,/, $value );
            foreach my $val (@values) {
                my $insert_char = $self->_get_insert_char(
                    key          => $val,
                    system_param => $param
                );
                $val = $insert_char . $val;
                push @cmd_inputs, $param, $val;
            }
            next;
        } else {

            # Determine the insert character for that option
            my $insert_char = $self->_get_insert_char(
                key          => $key,
                system_param => $param
            );
            $value = $insert_char . $value;
        }
        push @cmd_inputs, $param, $value;
    }

    # reset param and value for use in the following loop
    $param = $value = '';

    if ( @concat_limits && $end ) {

        # concatenate the switches and the values
        for ( $i = $start; $i <= $end; $i += 2 ) {

         # if the switch is empty and the option still has to be concatenated,
         # the value will be concatenated
            if ( $cmd_inputs[$i] ) {
                $param .= $cmd_inputs[$i];
            } else {
                $param .= $cmd_inputs[ $i + 1 ];
                $cmd_inputs[ $i + 1 ] = '';
            }

            # concatenate the current value to the value string only if it is
            # not an empty string
            $value .= $cmd_inputs[ $i + 1 ];

            # reset these elements in cmd_inputs. When all cmd_inputs are
            # finally joined, these switches and values must not be considered
            # again
            $cmd_inputs[$i] = $cmd_inputs[ $i + 1 ] = '';
        }
        $cmd_inputs[$start] = $param;
        $cmd_inputs[ $start + 1 ] = $value;
    }

    # return command and its parameters

    my $ret_cmd = $derived_hash->{command} || $self->get_command();

    # The extended query needs to be supplied after the command
    # and before the other options. For example, this works:
    # volume modify { -state online } -state offline
    # But this does not:
    # volume modify -state offline { -state online }
    if ( keys %$extended_query ) {

        # This is CMode-only, so simply obtaining the key mapped to
        # and putting a space between the pairs of options. We don't
        # need to worry about insert characters and boolean options.
        $ret_cmd .= ' { ';
        while ( my ( $key, $value ) = each %$extended_query ) {
            $param = $self->_get_alias_root( alias => $key );
            $ret_cmd .= "$param $value ";
        }
        $ret_cmd .= '} ';
    }

   # join all cmd_inputs to the command line to be sent (ignore empty strings)
   # items with even indices are the switches
   # items with odd indices are the values (including the insert_char)
   # a space is used after each switch-value pair
    for ( $i = 0; $i < $#cmd_inputs; $i += 2 ) {
        $value = $cmd_inputs[$i] . $cmd_inputs[ $i + 1 ];
        $ret_cmd .= ' ' . $value if $value ne '';
    }

    $ret_cmd .= $pass_through if ( defined $pass_through );

    # The following regex is to handle hourly-snapshot-num and hour-list
    # options which cannot be handled in the cdef.
    if ( $ret_cmd
        =~ /(snap sched .*)hourly-snapshot-num (\S+) (hour-list (\S+))?(.*)/ )
    {
        if ($3) {
            $ret_cmd = $1 . $2 . "@" . $4 . $5;
        } else {
            $ret_cmd = $1 . $2 . " " . $5;
        }
    }

    #
    # The regex is to remove additional spaces at the end of the returned
    # command. This case will be encountered for commands having switches with
    # no value at the end
    $ret_cmd =~ s/\s+$//;
    $Log->exit() if $may_exit;
    return $ret_cmd;
} ## end sub translate

###############################################################################
#  P R I V A T E
###############################################################################

###############################################################################
# Method:    _handle_processing_for_special_properties
# Objective: To perform the additional processing for the special properties
#             defined, if any
# Details: Users can define special properties (which are common) in the
#          .cdef file and handle the processing here.
#          Eg: A property called tool => 1 is added to the cdef_entries which
#          denote tools and this entry helps us to perform additional
#          processing needed such as handling the absolute path of the tool
#          and its version
###############################################################################
sub _handle_processing_for_special_properties {
    my ( $self, %args ) = @_;
    $Log->enter() if $may_enter;
    my %spl_props = %{ $self->get_special_properties() };
    my $command = $self->get_command();
    if ( exists $spl_props{tool} && $spl_props{tool} ) {

        # This cdef entry represents a tool.
        if (defined $args{tool_full_path}
            && (   defined( $args{tool_dir_path} )
                || defined( $args{tool_version} ) )
            )
        {
            $Log->exit() if $may_exit;
            NATE::Exceptions::Argument( "option tool_full_path is mutually "
                    . "exclusive with the options tool_dir_path and tool_version"
            );
        }
        if ( $args{tool_full_path} ) {
            $command = delete $args{tool_full_path};
        } else {
            my $path_chosen_by_cdef
                = ( $self->get_is_windows() )
                ? $spl_props{windows_tool_path}
                : $spl_props{unix_tool_path};
            my $path = delete $args{tool_dir_path} || $path_chosen_by_cdef;
            my $version = delete $args{tool_version}
                || $spl_props{tool_version};
                if ( defined $path ) {
                    if($self->get_is_windows()) {
                        chop($path) if ($path =~ /(.*)\\$/);
                        $command = join( '\\',$path,$command );
                    } else {
                        $command = catfile( $path, $command );
                    }
                }
            if ( defined $version ) {
                my $sep = $spl_props{tool_version_sep};
                unless ( defined $sep ) {
                    $Log->exit() if $may_exit;
                    NATE::Exceptions::Argument->throw(
                              "CDEF for '$command' is "
                            . "missing information about 'tool_version_sep' in "
                            . "special_properties attribute. Please submit a burt "
                            . "with type as nacl and subtype as nacl_core" );
                }
                $command .= $sep . $version;
            }
        }
        $args{cdef_trans_derived_info}->{command} = $command;
    }
    # For commands that run on Unix based operating systems whose path is not set as
    # an environment variable, we need to execute the command using the absolute path.
    # command_full_path can be used to provide this information programatically.
    # Otherwise,
    if ( $args{command_full_path} )
    {
        $args{cdef_trans_derived_info}->{command} = delete $args{command_full_path}
    }
    # If you define any new special properties in the cdef, you can
    # place the corresponding handling code here.
    # if ($spl_props{my_prop}) {
    #    ...
    # }
    $Log->exit;
    return %args;
}

###############################################################################
# Method:    _get_alias_root()
# Objective: Return system cmd parameter corresponding to common alias
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _get_alias_root {
    my ( $self, %args ) = @_;
    $Log->enter() if $may_enter;
    my $alias = $args{alias};

    my $system_cmd_param;

    # get system command parameter corresponding to alias from cdef
    $system_cmd_param = $self->get_alias_params()->{$alias};

    #
    # If the alias could not be mapped into a system command parameter
    # we just prepend a hyphen before it and let it fall through to the
    # service.
    # In the case of Windows however, switches begin with a '\' not a '-'.
    # Therefore, for Windows commands we need to prepend a '\' instead.
    #
    unless ( defined $system_cmd_param ) {
        $system_cmd_param = $self->get_switch_identifier() . $alias;
    }

    $Log->exit() if $may_exit;
    return $system_cmd_param;
} ## end sub _get_alias_root

###############################################################################
# Method:    _get_insert_char()
# Objective: Return the "insert character" for that programmatic parameter
# Details:   The "insert character" is the character to be inserted between
#            the switch and its value. Generally it's a space, but for certain
#            commands it could be an = or some other character. Example
#            dfm options set volOverDeduplicatedThreshold=150
#            To determine this, we look at two fields:
#            1) "options_sep" - This is a hash which contains the options
#                separator for individual options. This is defined only if
#                there are individual options which have separators different
#                from the default.
#            2) "command_sep" - This is a scalar, which defines the separator
#                for all options of a particular command. If the CDEF does
#                not have any entry for this, it defaults to space. The
#                "command_sep" field is only looked at if an entry
#                 corresponding to that option is not found in the
#                "options_sep" hash.
#            Refer BURT 401694.
###############################################################################
sub _get_insert_char {
    my ( $self, %args ) = @_;
    $Log->enter() if $may_enter;
    my $key          = $args{key};
    my $system_param = $args{system_param};

    my $insert_char;

    if ( $system_param eq "" ) {
        $Log->exit() if $may_exit;
        return "";
    }

    if ( defined $self->get_options_sep()->{$key} ) {
        $insert_char = $self->get_options_sep()->{$key};
    } else {
        $insert_char = $self->get_command_sep();
    }

    $Log->exit() if $may_exit;
    return $insert_char;
}

###############################################################################
# Method:    _is_boolean_option()
# Objective: Determine whether a given option is a boolean or not.
# Details:   This method returns 1 if the option is boolean, 0 else.
#            This is done by searching for the option in the list of boolean
#            options (which is present in $self->{boolean_options})
###############################################################################
sub _is_boolean_option {
    my ( $self, %args ) = @_;
    $Log->enter() if $may_enter;
    my $key = $args{key};

    $Log->exit() if $may_exit;
    map { return 1 if ( $key eq $_ ) } @{ $self->{boolean_options} };

    return 0;
}

###############################################################################
# Method:    _is_repeated_option()
# Objective: Determine whether a given option is a repeated or not.
# Details:   This method returns 1 if the option is repeated, 0 else.
#            This is done by searching for the option in the list of repeated
#            options (which is present in $self->{repeated_options})
###############################################################################
sub _is_repeated_option {
    my ( $self, %args ) = @_;
    $Log->enter() if $may_enter;
    my $key = $args{key};

    $Log->exit() if $may_exit;
    map { return 1 if ( $key eq $_ ) } @{ $self->{repeated_options} };

    return 0;
}

1;

#############################################################################
#                             POD Documentation
#############################################################################
## @pod here

=pod

=head1 NAME

NACL::APISet::Schema::Cdef - interface to get command definition

=head1 Synopsis

use NACL::APISet::Schema::Cdef;

=head1 Description

Cdef.pm provides an interface to get command definition.

=head1 EXCEPTIONS

None.

=head1 ATTRIBUTES

=over

=item command

(String)
command definition (cdef) action

=item alias_params

(Hashref)
hash of alias parameters that mapped to additional options of the cdef action

=item file

(String)
cdef file that contains data structures for cdef command

=back

=head1 CONSTRUCTOR

=head2 new

Creates a CDEF object for the given command.

=over

=item Synopsis

$cdef_obj = NACL::APISet::Schema::Cdef->new(command      => $command,
                                            option_order => $option_order,
                                            alias_params => $alias_params,
                                            file         => $file
                                            [is_windows  => 0|1]);

=item Arguments

=over

=item command

(Mandatory)
Command action (eg. volume create, volume rename, lun create, ...)

=item alias_params

(Mandatory)
[Hashref]
List of aliases that mapped to attributes of a particular cdef file

=item file

(Mandatory)
Cdef file that contained data structures which pertain to cdef action 

=item is_windows

(Optional)
[Boolean]
Specifies whether the resource is Windows or not. This is necessary for
deciding whether to prepend a hyphen (for switches, clients and filers)
or a slash (for Windows) options

=back

=item Returns

A reference to the created Cdef object.

=back

=head2 get_command

Get command action 

=over

=item Synopsis

$command = $cdef_obj->get_command();

=item Arguments

None.

=item Returns

Command corresponding to the Cdef object

=back

=head2 get_alias_params

Get an alias list of attributes from a particular cdef file

=over

=item Synopsis

my $alias_href = $cdef_obj->get_alias_params();

=item Arguments

None.

=item Returns

Hashref list of attributes
(ex. dsid => "-dsid", volume-type => "-volume-type")

=back

=head2 B<get_file()>

Get cdef file name

=over 4

=item Inputs

None

=item Outputs

Return a string 
(ex. volume)

=item Examples

 # get cdef file name
 #
 my $cdef_file = $cdef_obj->get_file();

=back

=head2 B<is_param_mapped(%args)>

Verify input parameter is mapped to ONTAP command parameter 

=over 4 

=item Inputs

=back

=over

=item * I<-param> => <optional | required parameter>

Command parameter

=back

=over

=item Outputs

Return TRUE | FALSE (1 | 0)

=item Examples

 # verify the required param (newname) for "volume rename" is mapped to ONTAP 
 # command parameter
 #
 if (! $cdef_obj->is_param_mapped(-param => "newname") {
     $self->_set_err("\"newname\" param is not mapped to command in cdef file");
     return $FALSE;
 }

=back

=head2 B<translate(%args)>

Convert command definition into actual command

=over 4 

=item Inputs

=back

=over

=item * I<-filer> => <$filer_name>

Name of filer or controller

=item * I<-cdef> => <command_definition>

Command definition (ei. volume_transition, lun_create, ...) 

=item * I<%args>

Other attributes based on each command definition.
(See cdef files for additional attributes per command definition)

=back

=over

=item Outputs

Return formated command that is ready to be executed on the filer

=item Examples

 # Convert cdef into formated command that can be executed on a filer
 #
 my $cmd = $cdef_obj->translate(-cdef   => "volume_create",
                                -vserver => $vserver_name,
                                -aggregate => $aggr_name,
                                -volume => $volume_name,
                                -size => "100m");
 if ($cdef_obj->error()) {
     $self->_set_err("failed to translate cdef: volume_create", 
                     $cdef_obj->errmsg());
     return $FALSE;
 }

=back

=cut
