## $Id$
#
# Copyright (c) 2013 NetApp Inc.
# All rights reserved.
#

package NACL::APISet::Schema::SymbolapiSchema;

# compiler directives
use strict;
use warnings;

# NATE imports
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Data::Dumper;
use Params::Validate qw(:all);
use NATE;

my $Classpath;

BEGIN {
    $Classpath .= Tharn::param('SYMBOL_API_PATH') || "/usr/software/test/share/symbol_API/current:/usr/software/test/share/symbol_API/current/lib/SYMsdk.jar:/usr/software/test/share/symbol_API/current/lib/gson-2.2.2.jar";
    $ENV{CLASSPATH} = $Classpath;
}


# Exceptions imported

use NATE::Exceptions::Argument qw(:try);
use NACL::APISet::Exceptions::CDEFNotFoundException qw(:try);
use NACL::APISet::Exceptions::MethodNotFoundException qw(:try);

our %nacl_apiset_symbolapi_hash = ();

###############################################################################
# Method:    new()
# Objective: To Create schema object for command translation
# Details:   See POD documentation at the end of this file.
###############################################################################
sub new {
    my ($class, %args) = @_;
    $Log->enter() if $may_enter;

	my $self = {};

    # bless instance into class
    my $schema = bless $self, $class;

    $Log->exit() if $may_exit;
    return $schema;
} ## end sub new


###############################################################################
# Method:    load()
# Objective: Load command definitions
# Details:   See POD documentation at the end of this file.
###############################################################################

sub load {
    my ($self, %args) = @_;
    $Log->enter() if $may_enter;

    my $path;
    my @files = ();
    $path = $args{path};
    $self->{path} = $path;
    if (defined $nacl_apiset_symbolapi_hash{$path}) {
        # Already loaded
        $Log->debug("CDEF dir is already loaded for '$path' and cached");
        $Log->exit() if $may_exit;
        return;
    }

    # process parameters
    while (my ($key, $value) =  each(%args)) {
        if ($key =~ /path$/) {
            if (! length($path = $value)) {
                $Log->exit() if $may_exit;
                NATE::Exceptions::Argument->throw(
                    "Invalid value for \"$key\": " . $value);
            }
        }
    }

    unless (defined $path and length $path) {
        $Log->exit() if $may_exit;
        NATE::Exceptions::Argument->throw(
            "Mandatory argument \"path\" not specified ");
    }

    # check for valid cdef path
    if (-d $path) {
        # retrieve cdef files. non HTTP path
        opendir(DIR, $path) or do {
            $Log->exit() if $may_exit;
            NACL::APISet::Exceptions::CDEFNotFoundException->throw(
                "Cannot open dir: $path ");
        };
        my $sep = ($path =~ /^\//)? "/": "\\";
        foreach my $file (readdir(DIR)) {
            # skip files that does not have cdefs extension
            next if ($file !~ /\.xml$/);
            push @files, ($path. $sep .$file);
        }
        closedir(DIR);
    } elsif (-f $path) {
        push @files, $path;
    
    } else {
            NACL::APISet::Exceptions::CDEFNotFoundException->throw(
                "Could not open $path: $@");
    }

    # if files list is empty, cdef files have not been consolidated
    if (scalar(@files) == 0) {
        $Log->exit() if $may_exit;
        NACL::APISet::Exceptions::CDEFNotFoundException->throw(
            "No cdef data files found in: $path ");
    }

    # Load each of the XML file and get an XPATH object for each of the
    # ZAPI and store them directly in the name of ZAPI in schema object
    foreach my $cdef_file (sort @files) {
        $Log->debug("Loading cdef: " . $cdef_file);
        $cdef_file =~ /([\w-]+)\.xml$/;
        my $file = $1;
        my $content = NATE::Inc::get_file($cdef_file, register_for_s2t => 0) or
            NATE::BaseException->throw("Error reading file '$cdef_file': $@");
        
        require XML::Simple;
        $self->{cdef_contents} = XML::Simple::XMLin($content);

    }
    $Log->exit() if $may_exit;
} ## end sub load	

###############################################################################
# Method:    translate()
# Objective: return cmd syntax for requested system operation
# Details:   See POD documentation at the end of this file.
###############################################################################

sub translate {

    my ($self, %opts) = @_;
    $Log->enter() if $may_enter;

	my @array;
    my $command = $opts{cdef_method};
    my $hash_input = $self->{cdef_contents};
	
    $Log->debug("Translating procedure call $command");

    if (exists $hash_input->{proc}->{$command}) {
        my $input = $hash_input->{proc}->{$command}->{input}->{ref};
        my $output = $hash_input->{proc}->{$command}->{output}->{ref};
        my $value = $hash_input->{proc}->{$command}->{value};

        # Convert the data type unsigned <int|hyper|long> etc to just <int|hyper|long>,
        # because this has to be converted to XDRtype which is understood by SYMbolAPI call
        $input =~ s/unsigned (.*)/$1/g;
        $output =~ s/unsigned (.*)/$1/g;

        # Provide mapping for the primitive data type to its equivalent XDRtype
        my %primitive_to_XDR = (
            "void"   => "XDRvoid",
            "int"    => "XDRint",
            "string" => "XDRstring",
            "long"   => "XDRlong",
            "hyper"  => "XDRlong",
        );

        if ( exists $primitive_to_XDR{$input} ) {	
            $input = $primitive_to_XDR{$input};
        }
    	
        if ( exists $primitive_to_XDR{$output} ) {	
            $output = $primitive_to_XDR{$output};
        }

        # Form an array of the found input, output and procedure number of the 
        # procedure call to be executed
        @array = ($value, $input, $output);

        # We will have to convert the enum strings into enum values
        # looking into the original cdef contents
        $self->_process_enums($opts{input}, \&_convert_enum_string_to_value, $hash_input);

        # Storing the hash structure in a string format 
        my $str = Dumper ($opts{input});

        # Process and convert the input hash into Json string
        # Json string is of the form:
        # {"key":"val","key2":"val2",etc....}
        $str =~ s/=>/:/g;
        $str =~ s/.*VAR.*\s*=//g;
        $str =~ s/\'/\"/g;
        $str =~ s/;$//g;
        
        $Log->debug("Input options in the form of Json string formed is:\n $str");
        push(@array, $str);

        $Log->debug("Translated command $command to procedure value $value with input"
        . " object of type $input, and response object of type $output");

    } else {
        NACL::APISet::Exceptions::MethodNotFoundException->throw(
              "Failed to get command definition for "
            . "command $command" )
    }

    $Log->exit() if $may_exit;
    return @array;

}

###############################################################################
# Method:    _process_enums()
# Objective: internal method to process the user given enums to numeric value
###############################################################################

sub _process_enums {
    my $self = shift;
    $Log->enter() if $may_enter;
    my ($hash, $callback, $hash_input) = @_;
    while (my ($k, $v) = each %$hash) {
        # Keep track of the hierarchy of keys, in case
        # our callback needs it.

        if (ref($v) eq 'HASH') {
            # Recurse.
            $self->_process_enums($v, $callback, $hash_input);
        } else {
            # Otherwise, invoke our callback, passing it
            # the current key and value.
            $callback->($k, \$v, $hash_input);
        }

        # Replace old hash values with the new ones
        $hash->{$k} = $v;
    }
    $Log->exit() if $may_exit;
}

# Helper method for _process_enums
sub _convert_enum_string_to_value {
    my ($k, $v, $hash_input) = @_;
    
    # For enum values which are of type CandidateSelectionType.CANDIDATE_SEL_COUNT
    # we have to convert it to its equivalent value 3
    
    my $original_enum = $$v;
    
    if ( $$v =~ /([A-Za-z]+)\.([A-Z_0-9]+)/) {
        my $enum_type = $1;
        my $enum_string = $2;
        $$v = $hash_input->{enum}->{$enum_type}{EnumValue}{$enum_string}{value};
        $Log->debug("Converted enum $original_enum to value $$v");
    }
}

###############################################################################
# Method:    input_output_object_creation()
# Objective: method to create input objects and response objects before passing
#            it to Connectrec's execute method
# Details:   Uses Inline Java code because SYMbol API takes in JAVA objects
###############################################################################

sub input_output_object_creation {
    my ($self, %args) = @_;
    $Log->enter() if $may_enter;
    
    my @cmd_input_output = @{$args{command_input_output}};
    my $command          = $cmd_input_output[0]; 
    my $input            = $cmd_input_output[1]; 
    my $output           = $cmd_input_output[2];
    my $input_data       =  $cmd_input_output[3];

    use Inline Java => <<'END_OF_JAVA_CODE', CLASSPATH => $Classpath, ;
    import devmgr.versioned.symbol.*;
    import devmgr.versioned.jrpc.*;
    import demo.Common.*;
    import demo.Common.Utility;
    import java.io.*;
    import java.util.*;
    import java.lang.reflect.Method;
    import java.lang.reflect.*;
    import java.awt.Rectangle;
    import com.google.gson.*;

    public class I_O_obj_creation {
   
        // Constructor - does nothing
        public I_O_obj_creation() { }

        // This method actually forms the objects of the desired type
        public Object[] form_objects(int command, String input, String output, String json1) {

        System.out.println("Executing procedure number " + command + " with input object type " + input + " and output object type " + output);

        // The input object type is either a symbol object of type "devmgr.versioned.symbol.<input>"
        // or jrpc object of type "devmgr.versioned.jrpc.<input>". We will try with symbol object
        // first, and if that fails we will fall back to jrpc object. 
        // During runtime, it is not possible to find out which object type the input
        // belongs to. Similarly we do for the output object too.
           
        int flag = 0;
        String actual_input = input;
        input = "devmgr.versioned.symbol." + input;
        Object argument = null;
        Gson gson = new Gson();

        try {
            Class classDefinition = Class.forName(input);
       	    argument = classDefinition.newInstance();

            try {
                String theType = argument.getClass().getName();
                argument = gson.fromJson(json1,Class.forName(theType) );
            } catch (ClassNotFoundException e) {
          		System.out.println(e);
      	   }
        } catch (InstantiationException e) {
            System.out.println(e);
		} catch (IllegalAccessException e) {
          	System.out.println(e);
      	} catch (ClassNotFoundException e) {
          	System.out.println(e);
            flag = 1;
   	    }
      
        if (flag==1) {
            input = "devmgr.versioned.jrpc." + actual_input;
            argument = null;

            try {
                Class classDefinition = Class.forName(input);
                argument = classDefinition.newInstance();

                try {
                    String theType = argument.getClass().getName();
                    argument = gson.fromJson(json1,Class.forName(theType) );
                } catch (ClassNotFoundException e) {
      	            System.out.println(e);
                }

            } catch (InstantiationException e) {
                System.out.println(e);
	        } catch (IllegalAccessException e) {
                System.out.println(e);
            } catch (ClassNotFoundException e) {
      	        System.out.println(e);
            }
      	   
        } // end if (flag==1)
      
        flag = 0;
        String actual_output = output;
  	    output = "devmgr.versioned.symbol." + output;
   	    Object returncode = null;
        try {
            Class classDefinition = Class.forName(output);
       	    returncode = classDefinition.newInstance();
      	} catch (InstantiationException e) {
            System.out.println(e);
      	} catch (IllegalAccessException e) {
            System.out.println(e);
      	} catch (ClassNotFoundException e) {
            flag = 1;
            System.out.println(e);
        }
      	   
        if (flag==1) {
            output = "devmgr.versioned.jrpc." + actual_output;
            returncode = null;
            try {
                Class classDefinition = Class.forName(output);
                returncode = classDefinition.newInstance();    
      	    } catch (InstantiationException e) {
                System.out.println(e);
      	    } catch (IllegalAccessException e) {
          	    System.out.println(e);
      	    } catch (ClassNotFoundException e) {
          	    System.out.println(e);
      	    }

        } // end if (flag==1)

        // Populate the object array with the procedure number,
        // the input object created and the response object created.
        // This is then packaged and sent to Connectrec's execute,
        // where the response object is populated with the output.

        Object[] arr = new Object[3];
        arr[0]=command;
        arr[1]=argument;
        arr[2]=returncode;
        return arr;
    }
}

END_OF_JAVA_CODE

    # Create an I_O_obj_creation object to form the input object and response
    # object
    my $NACL = new NACL::APISet::Schema::SymbolapiSchema::I_O_obj_creation();

    # Procedure calls that do not take any argument, should be sent {}
    # for the Json call to work
    $input_data = "{}" if ($input_data =~ /undef/);
    
    # Procedure calls which take a single input i.e all the XDR types
    # should convert the 'value' to 'm_Value' which is understood by Json
    $input_data =~ s/value/m_Value/g if ($input =~ /XDR/i);
    
    # Form the objects with the input JSon string formed
    my $return = $NACL->form_objects($command, $input, $output, $input_data);
    
    $Log->exit() if $may_exit;
    return $return; 
}

1;

#############################################################################
#                             POD Documentation
#############################################################################

=pod

=head1 NAME

NACL::APISet::Schema::SymbolapiSchema - interface to get SymbolAPI schema

=head1 Synopsis

use NACL::APISet::Schema::SymbolapiSchema;

=head1 Description

Provides a mechanism to load, retrieve, and translate schema of
SYMbolAPI procedures into its corresponding procedure values,
and create objects for input and output

=head1 EXCEPTIONS

=over

=item NATE::Exceptions::Argument

=item NACL::APISet::Exceptions::CDEFNotFoundException

=item NACL::APISet::Exceptions::MethodNotFoundException

=back

=head1 CONSTRUCTOR

=head2 new

Creates a SymbolapiSchema object for the given procedure call.

=over

=item Synopsis

$cdef_obj = NACL::APISet::Schema::SymbolapiSchema->new();

=item Arguments

No input

=item Returns

A reference to the created SymbolapiSchema object.

=back

=head1 METHODS

=head2 load

Loads SYMbol.xml file present in the specified 'path' input into
the schema object itself.

=over

=item Synopsis

$SymbolapiSchema_obj->load(path=>$path);

=item Arguments

=over

=item path

(Required)
The path to CDEF file

=back

=item Returns

Nothing

=back

=head2 translate

Translate the SYMbolAPI procedure to its corresponding procedure number
and find out the input and output object types for the same.

=over

=item Synopsis

$command = $SymbolapiSchema_obj->translate(cdef_method=>"getVolumeCandidates");

=item Arguments

=over

=item cdef_method

(Required)
SYMbolAPI procedure call to be invoked

=back

=item Returns

Returns an array of procedure call number, input and output object type,
and JSon string formed from the input data which was provided.

=back

=head1 AUTHOR/MAINTAINER

=over

=item (dl-nacl-dev@netapp.com)

=back

 $Id$
 Last revised $Date$ by $Author$

=cut


1;

