
## @summary Class to hold all connection information.
## @author olalde@netapp.com, dl-nacl-dev@netapp.com
## @status Shared
## @pod here

# Not all 'connections' to services will be in the form of NATE::Connectrec
# The Session class will define all data needed for connections, whether
# NATE Connection or not.
package NACL::APISet::Session;

#
# Compiler Directives
#
use strict;
use warnings;

#
# NATE imports
#
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Data::Dumper;

use Class::MethodMaker
[ scalar => [qw / timeout_scaling_factor/] ];

###############################################################################
# Method:    cleanup
# Objective: Cleanup the Session object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub cleanup {
    $Log->enter() if $may_enter;

    my ($self, %args) = @_;

    delete $self->{connection} ;

    $Log->exit() if $may_exit;
}


###############################################################################
# Method:    get_timeout
# Objective: Retrieves the current timeout value
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_timeout {
    my $self = shift;
    $Log->enter() if $may_enter;
    $Log->debug("Timeout values is " . $self->{timeout});
    $Log->exit() if $may_exit;
    
    return $self->{timeout};
}


1;

##############################################################################
#                                    POD here
##############################################################################

=head1 NAME

NACL::APISet::Session - API Set Session base class

=head1 SYNOPSIS

 use NACL::APISet::Session;
 
=head1 DESCRIPTION

A C<NACL::APISet::Session> contains data and methods common to all API Set 
Session classes. This includes functionality for cleanup, retrieving, setting
and clearing the timeout value.
Inherit from this class when creating new API Set Session classes.
This class provides functionality common to all Session classes, regardless of APISet.

=head1 EXCEPTIONS

None.

=head1 CONSTRUCTOR

None. This class is present only to provide common functionality, an object of 
this class should never be instantiated. We should only instantiate objects
of Session sub-classes.

=head1 Methods

=head2 cleanup

This method is used to clean up the Session after use. This involves 
deleting the connection object stored within the Session object.

=over

=item Synopsis

$self->{session}->cleanup();

=item Arguments

None.

=item Returns

Nothing.

=back


=head2 get_timeout

This method is used to retrieve the current timeout value (in seconds).

=over

=item Synopsis

$timeout = $self->{session}->get_timeout();

=item Arguments

None.

=item Returns

Current timeout value.

=back







