## @summary Class to hold all CLI connection information.
## @author madhav.shekar@netapp.com, dl-nacl-dev@netapp.com
## @status Shared
## @pod here
# Not all 'connections' to services will be in the form of NATE::Connectrec
# The Session class will define all data needed for connections, whether
# NATE Connection or not.
package NACL::APISet::Session::CLI;

#
# Compiler Directives
#
use strict;
use warnings;

#
# NATE imports
#
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Params::Validate qw(:all);
use Data::Dumper;

#
# Exceptions imported
#
use NACL::APISet::Exceptions::ConnectionFailedException qw(:try);
use NATE::Exceptions::Argument qw(:try);

use base qw(NACL::APISet::Session);

sub new {
    $Log->enter() if $may_enter;
    my ($class, @opts) = @_;

    my %args = validate_with(
        params => \@opts,
        spec   => { additional_spec => { type => HASHREF, default => {} } },
        allow_extra => 1
    );
    my $additional_spec = delete $args{additional_spec};

    # Fill the member vars for this instance
    my $self = NACL::APISet->validate_params(
        params => \%args,
        spec   => {
            username => { type => SCALAR, optional => 1},
            password => { type => SCALAR, optional => 1 },
            conntype => { type => SCALAR, optional => 1 },
            ipaddress => { type => SCALAR, optional => 1 },
            connection => { type => OBJECT, isa => 'Connectrec', optional => 1 },
            hostobj => { type => OBJECT, isa => 'Hostrec', optional => 1 },
            %$additional_spec
        },
    );

    bless $self, $class;

    # Connect to the service
    $self->connect();
    # Prepare it for use
    $self->prep_service();

    # And give the caller the referece to the bless object.
    $Log->exit() if $may_exit;
    return $self;
}

###############################################################################
# Method:    connect
# Objective: Connect to the service
# Details:   See POD documentation at the end of this file.
###############################################################################
sub connect {
    $Log->enter() if $may_enter;

    # %args is usually empty. But the individual Session classes can send in
    # something that has to be passed to Hostrec->connect() additionally

    my ( $self, %args ) = @_;

    try {
        if ( defined $self->{connection} ) {
            if ($self->{connection}->is_persist() && $self->{connection}->is_disconnected()) {
                $self->_reconnect();
            }
            $self->set_connection( connection => $self->{connection} );
        } else {
            # Check for user-defined overrides. If any of these are
            # passed we add them to the connspec
            #
            $self->{'connhost'} = $self->{'ipaddress'};
            my @overrides = (
                qw (conntype username password prompt_username
                    prompt_password connhost)
            );
            my %conn_overrides;
            foreach my $param (@overrides) {
                if ( defined $self->{$param} ) {
                    $conn_overrides{$param} = $self->{$param};
                }
            }
            if (defined $self->{connid}) {
                if ($self->{connid} eq 'console') {
                    # only for console connections, we will try to reuse any
                    # existing connections and we assume that the connection
                    # id would be 'console', an expected standard name which
                    # users should adhere to for console connections.
                    $conn_overrides{id} = $conn_overrides{template} =
                        $self->{connid};
                    $conn_overrides{login} = 1;
                } else {
                    $conn_overrides{template} = $self->{connid};
                }
            }
            $Log->trace( Dumper( \%conn_overrides ) );
            #The workaround adds RLM to the list of recognized prompts by the Hostrec object.
            #This code will go away when burt577429 (support for RLM prompts) are fixed.

            if ( ref $self eq 'NACL::APISet::Session::CLI::SP' ) {
                my $p      = $self->{hostobj}->prompt_norm();
                my @prompt = @{$p};
                push( @prompt, 'RLM.*>' );
                $self->{hostobj}->prompt_norm( \@prompt );
            }
            
            require NATE::ParamSet;
            my $idle_timeout = NATE::ParamSet::param_global()
                ->get('NACL_CONNECTION_IDLE_TIMEOUT');
            
            # Note that we don't need to pass a callback for what to do on
            # reconnect, because the APISet code already handles reapplying
            # all the original settings such as privilege, showallfields, rows=0
            # and so on.
            my %idle_timeout_args;
            $idle_timeout_args{idle_timeout} = $idle_timeout if defined $idle_timeout;

            ## extra logging for burt 999020
            my $connect_obj = $self->{hostobj}->connect(
                    %conn_overrides, %args, %idle_timeout_args,
            );

            try{
              $self->set_connection(
                 connection => $connect_obj,
              );
            }otherwise {
              my $exception = shift ;
              if(defined NATE::ParamSet::param_global()->get('NACL_BURT_999020')) {
                 $Log->comment( sub { "The connect obj returned is " . Dumper( $connect_obj ) }) ;
                 $Log->comment( sub { "The self obj is ". Dumper($self) });
              }
              $Log->exit() if $may_exit;
              $exception->throw();
            }
            $Log->debug( sub { Dumper( $self->{connection} ) });
        }
    }
    otherwise {
        NACL::APISet::Exceptions::ConnectionFailedException->throw(
            "Could not create connection for: #" . $self->{connid} . "\n$@" );
    };
    $Log->exit() if $may_exit;
}

###############################################################################
# Method:    set_connection
# Objective: Set the Session object to internally hold the connection object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub set_connection {
    $Log->enter() if $may_enter;
     my ( $self, %args ) = @_;
    $self->{connection} = $args{connection};
    # storing the Default value so that we can use this in clear_timeout()
    # Only if default_timeout is not defined, store the values in the
    # session object
    if ( !defined $self->{default_timeout} ) {
        $self->{default_timeout} = $self->{connection}->timeout();
        $self->{timeout}         = $self->{default_timeout};
    }

    $Log->exit() if $may_exit;
}

###############################################################################
# Method:    prep_service
# Objective: Prepare the service for use. We define it to default to not
#            do anything.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub prep_service {
    $Log->enter() if $may_enter;

    my ( $self, %args ) = @_;
    if($self->{connection}->conntype() =~ /^(serial|pm_cons|rconsole)$/){
	   $self->{connection}->execute(command => "");
    }

    $Log->exit() if $may_exit;
}

###############################################################################
# Method:    set_timeout
# Objective: Set the timeout to a new value
# Details:   See POD documentation at the end of this file.
###############################################################################
sub set_timeout {
    my $self = shift;
    $Log->enter() if $may_enter;
    my %args = validate_with(
        params => \@_,
        spec   => {
            'connectrec-timeout' => { type => SCALAR, optional => 1 },
            timeout              => { type => SCALAR, optional => 1 },
        }
    );
    my $timeout;
    if (defined $args{'connectrec-timeout'}) {
        $timeout = $args{'connectrec-timeout'};
    } elsif (defined $args{timeout}) {
        $timeout = $args{timeout};
    } else {
        NATE::Exceptions::Argument->throw( 'Either of the parameters '
                . "'connectrec-timeout' or 'timeout' should have been "
                . 'passed but neither were' );
    }
    $self->{connection}->timeout($timeout);
    $self->{timeout} = $self->{connection}->timeout();
    $Log->debug( 'Timeout set to ' . $self->{timeout} );
    $Log->exit() if $may_exit;
}

###############################################################################
# Method:    clear_timeout
# Objective: Set the timeout to the default value
# Details:   See POD documentation at the end of this file.
###############################################################################
sub clear_timeout {
    my $self = shift;
    $Log->enter() if $may_enter;

    $self->{connection}->timeout( $self->{default_timeout} );
    $self->{timeout} = $self->{connection}->timeout();
    $Log->debug( "Timeout has been reset to " . $self->{timeout} );

    $Log->exit() if $may_exit;
}
###############################################################################
# Method:    _get_scaled_timeout
# Objective: Set the timeout to a value scaled by the timeout scaling factor
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _get_scaled_timeout_value {
    $Log->enter() if $may_enter;
    my ($self, %opts) = @_;
    validate_with(
        params => \%opts,
        spec   => {
            timeout => { type => SCALARREF },
        }
    );
    my $tsf = $self->timeout_scaling_factor()||1.0;
    my $timeout = defined ${$opts{timeout}} ? ${$opts{timeout}} : $self->get_timeout();
    $tsf = 1 if ($timeout == -1);
    ${ $opts{timeout} } = $timeout * $tsf;
    $Log->exit() if $may_exit;
}
###############################################################################
# Method:    _add_prompts
# Objective: A wrapper to add prompts to the Connectrec object
###############################################################################
sub _add_prompts {
    $Log->enter() if $may_enter;
    my ( $self, %args ) = @_;
    validate_with(
        params => \%args,
        spec   => {
            prompts => { type => ARRAYREF },
        }
    );
    my @prompts = $self->{connection}->prompts();
    foreach my $prompt (@{$args{prompts}} ) {
        push( @prompts, $prompt );
    }
    $self->{connection}->prompts(@prompts);
    $Log->exit() if $may_exit;
}

sub _reconnect {
    $Log->enter() if $may_enter;
    my ($self, %opts) = @_;
    my $conn = $self->{connection};
    if ($conn->is_persist()) {
        if ($conn->is_console()) {
            # No sense in disconnecting/reconnecting console connections
            # However, the connection might be off by a prompt, so just resync
            $conn->resync();
        } else {
            $conn->disconnect();
            $Log->comment(
                sub {
                    "Reestablishing the connection to "
                        . $conn->hostp()->id()
                        . " with connid = "
                        . $conn->connid();
                }
            );
            my $max_reconnect          = $opts{max_reconnect};
            my $max_reconnect_timewait = $opts{max_reconnect_timewait};
            $conn->max_reconnect($max_reconnect) if ($max_reconnect);
            $conn->max_reconnect_timewait($max_reconnect_timewait)
                if ($max_reconnect_timewait);

            # This is the hack for burt482653
            if ($conn->{hostp}->{hosttype} =~ /switch/){
                $conn->{shell_type} = "none";
            }

	    try {
		$conn->connect();
	    } catch NATE::BaseException with {
		my $ex = shift;
		# as per burt#1056183
		# ensure the connection isn't left in an odd state
		eval {$conn->disconnect();};
		$ex->throw();
	    };
        }

        $self->_get_privilege_to_set_on_reconnect();
    }

    $Log->exit() if $may_exit;
}

sub _get_privilege_to_set_on_reconnect {
    $Log->enter() if $may_enter;
    my ($self) =  @_;

    my $conn = $self->{connection};
    my $output;

    # Depending on 'CMode/7Mode', appropriate commands are executed 
    if ($self->isa('NACL::APISet::Session::CLI::CMode')
        && !$self->isa('NACL::APISet::Session::CLI::CMode::Shell')) {
        $output = $conn->execute(command => "set -confirmations off; set -privilege test");
        $conn->execute(command => "set -confirmations off; set -privilege diagnostic") if($output =~ /Error/i);
    } elsif ($self->isa('NACL::APISet::Session::CLI::7Mode')) {
        $output = $conn->execute(command => "priv set -q test");
        $conn->execute(command => "priv set -q diag") if ($output =~ /unrecognized privilege: test/);
    }

    $Log->exit() if $may_exit;
}

1;

##############################################################################
#                                    POD here
##############################################################################



=head1 NAME

NACL::APISet::Session::CLI - API Set Session base class for CLIs

=head1 SYNOPSIS

 use NACL::APISet::Session::CLI;

=head1 DESCRIPTION

A C<NACL::APISet::Session::CLI> contains data and methods common to all API Set
Session CLI classes. This includes implementations for connect(), set_connection()
and prep_service().
Inherit from this class when creating new API Set Session classes
for CLI. This class provides functionality common to all Session classes for CLI.

=head1 EXCEPTIONS

=over

=item NACL::APISet::Exceptions::ConnectionFailedException

=back

=head1 CONSTRUCTOR

None. This class is present only to provide common functionality, an object of
this class should never be instantiated. We should only instantiate objects
of Session sub-classes.

=head1 Methods

=head2 connect

This method is used to connect to the service. It does this by either using the
connection object provided, or by constructing a full connection spec and
connecting using this full connection spec.

If the connection cannot be established, a ConnectionFailedException is thrown.

=over

=item Synopsis

$self->{session}->connect();

=item Arguments

=over

=item %Additional_Connection_Parameters

A hash containing any additional arguments to Hostrec->connect()

=back

=item Returns

Nothing.

=back

=head2 set_connection

This method is used to set the Session object to internally hold the connection
object created.

=over

=item Synopsis

$self->{session}->set_connection(connection => $conn_object);

=item Arguments

=over

=item connection

This is the connection object created.

=back

=item Returns

Nothing.

=back

=head2 prep_service

This method is used to take all the steps needed to prepare the service
for use. It is defined here to do nothing, since for most services nothing
needs to be done. For any commands that do have preparatory steps, we should
override this method in the appropriate Session sub-class.

=over

=item Synopsis

$self->{session}->prep_service();

=item Arguments

None.

=item Returns

Nothing.

=back

=head2 clear_timeout

This method is used to set the timeout to the default value.
Default = Timeout value that was present when the connectrec object was created

=over

=item Synopsis

$timeout = $self->{session}->clear_timeout();

=item Arguments

None.

=item Returns

Nothing.

=back

=head2 set_timeout

This method is used to set the timeout to a new value.

=over

=item Synopsis

    $self->{session}->set_timeout('connectrec-timeout' => $new_timeout);

                                 or

    $self->{session}->set_timeout(timeout => $new_timeout);

=item Arguments

Either of timeout or connectrec-timeout should be provided. If both are
provided, then connectrec-timeout takes precedence.

=over

=item connectrec-timeout

The new timeout value (in seconds).
If connectrec-timeout is specified as -1 then connection will
wait untill the command execution completes.

If connectrec-timeout is specified as 0 then connection will execute
the command on resource and come back immediately (applicable only for Firmware
commands at present), in such scenario
there will be no output in created response object.

=item timeout

The new timeout value (in seconds)

=back

=item Returns

Nothing.

=back

=head2 _get_scaled_timeout_value

This method sets the timeout value scaled by the specified timeout scaling
factor on the hash that is passed to Connectrec->execute().

