#
# Copyright (c) 2001-2009 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Class to hold all connection information for 7Mode Systemshell Sessions
## @author Isshwarya.M@netapp.com, anerudh@netapp.com, dl-nacl-dev@netapp.com
## @status Shared
## @pod here

# Not all 'connections' to services will be in the form of NATE::Connectionrec
# The Session class will define all data needed for connections, whether
# NATE Connection or not.

package NACL::APISet::Session::CLI::7Mode::Shell;

#
# Compiler Directives
#
use strict;
use warnings;

use base qw(NACL::APISet::Session::CLI::7Mode);

# Module import - APISets
use NACL::APISet::Response::CLI::Freebsd;

#
# NATE imports
#
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Params::Validate qw(:all);
use Data::Dumper;
use Time::HiRes qw(time); # This module is for measuring time in high precision

#
# Exceptions imported
#
use NATE::Exceptions::Argument qw(:try);
use NACL::APISet::Exceptions::ConnectionFailedException qw(:try);

sub new {
    $Log->enter() if $may_enter;
    $Log->debug( Dumper(@_) );

    # an instance of "this" is always the first arg
    # the second arg is a set of named arguments used to init
    # a hash.
    my ($class, %args) = @_;

    #
    # The conntype parameter should be specified ONLY if connid is also
    # provided
    #
    if ((defined $args{conntype}) and (not defined $args{connid})) {
        NATE::Exceptions::Argument->throw(
        "Conntype provided but connid not provided");
    }

    # Fill the member vars for this instance

    my $self = NACL::APISet->validate_params(
        params => \%args,
        spec   => {
            hostobj            => { isa => "Hostrec", optional => 1 },
            username           => { type => SCALAR, optional => 1 },
            password           => { type => SCALAR, optional => 1 },
            context_definition => { type => HASHREF, optional => 1 },
            prompt_username    => { type => SCALAR|UNDEF, optional => 1 },
            prompt_password    => { type => SCALAR|UNDEF, optional => 1 },
            conntype           => { type => SCALAR, optional => 1 },
            ipaddress          => { type => SCALAR, optional => 1 },
            connid             => { type     => SCALAR,
                                    optional => 1,
                                    default  => "7-system" },
            connection         => { isa => "Connectrec", optional => 1 },
        },
    );

    # Add more variables that are sub-class specific.


    # This tags the hash pointed to by $object as a
    # type of 7Mode/Shell
    bless $self, $class;

    # Connect to the service
    $self->connect();

    # And give the caller the referece to the bless object.
    $Log->exit() if $may_exit;
    return $self;
}

###############################################################################
# Method:    send_command
# Objective: Execute the command in the systemshell and gather the raw output
# Details:   See POD documentation at the end of this file.
###############################################################################
sub send_command {
    $Log->enter() if $may_enter;

    my ($self, %opts) = @_;
    my %args = validate_with(
        params      => \%opts,
        spec        => { connectrec => { type => HASHREF, default => {} } },
        allow_extra => 1
    );
    
    my %connectrec = %{delete $args{connectrec}};
    $self->_get_scaled_timeout_value(timeout => \$connectrec{timeout});
    my $parser_file = $args{parser_file};

    $Log->debug( sub { "Session/CLI/7Mode/Shell:send_command: " . Dumper( \%args ) } );

    my $do_not_answer;
    my @password_match_table = ('.*[Pp]assword.*:\s?'=>
                                 sub {
                                    return undef if $do_not_answer;
                                    $do_not_answer = 1;
                                    return $self->{connection}->{password};
                                 }
                               ) if delete $args{answer_password};

    
    unless ($args{method_name}) {
        # Caller is execute_command, all the options passed
        # should be considered as connectrec options
        %connectrec = ( %connectrec, %args );
        $args{command} ||= $connectrec{command};
        delete $connectrec{command};
    }


    my $execution_status;

    #this match table is used only during sudo

    $connectrec{'match_table'} = [@password_match_table,
                                  exists($connectrec{'match_table'}) ?
                                  @{$connectrec{match_table}} : () ];

    my $time_start = time();

    $connectrec{command_username} = $args{command_username} if(defined $args{command_username});
    my $output = $self->{connection}->execute(command=> $args{command},
                                                # shell type should be explicitly specified else connections will treat
                                                # it as of type "node".
                                                "shell_type" => "bash",
                                                get_status=>\$execution_status,
                                                %connectrec );
    my $time_end = time();
    my $execution_time = $time_end - $time_start;

    if (!defined $execution_status) {
        # Execution status will be undef only if there is some problem with
        # establishing the connection itself.
        NACL::APISet::Exceptions::ConnectionFailedException->throw($output);
    }

    $output =~ s/.*[Pp]assword.*// if $do_not_answer; # if password was answered
    $output =~ s/__status__=(\d+)\s+//;

    my ($package, $filename, $line) = caller;
    my $response_object;
    if (not $package->isa("NACL::APISet::Node::ZAPI::DBlade")) {

        if( !$args{method_name} ) {
            $Log->exit() if $may_exit;
            return $output;
        }

        # The response object should be of type Freebsd in this session where
        # we had connected to systemshell.
        $response_object =  NACL::APISet::Response::CLI::Freebsd->new(
                output            => $output,
                actual_command    => $args{actual_command},
                command           => $args{command},
                cdef_path         => $args{cdef_path},
                method_name       => $args{method_name},
                execution_status  => $execution_status,
                execution_time    => $execution_time,
                parser_file       => $parser_file
            );
    } else {
        # DBlade(zsmcli) uses systemshell connection type but it returns xml as
        # output. Freebsd response object cant be used so return here whatever is
        # needed to create DBlade response object.
        $Log->debug("Session/CLI/7Mode/Shell:send_command output:\n $output\n execution_time: $execution_time");
        $Log->exit() if $may_exit;
        return ($output, $execution_time, $execution_status);
    }

    $Log->exit() if $may_exit;
    return $response_object;

}

1;

##############################################################################
#                                    POD here
##############################################################################

=head1 NAME

NACL::APISet::Session::CLI::7Mode::Shell - 7Mode Systemshell CLI Session class

=head1 SYNOPSIS

 use NACL::APISet::Session::CLI::7Mode::Shell;

=head1 DESCRIPTION

A C<NACL::APISet::Session::CLI::7Mode::Shell> object is used to establish a
connection and send systemshell commands for the Freebsd CLI APISet.

=head1 EXCEPTIONS

=over

=item NATE::Exceptions::Argument

=item NACL::APISet::Exceptions::ConnectionFailedException

=back

=head1 CONSTRUCTORS

=head2 new

Returns a reference to a newly created C<NACL::APISet::Session::CLI::7Mode::Shell>.

=over

=item Synopsis

 $self->{session} = NACL::APISet::Session::CLI::7Mode::Shell->new( %args );

=item Arguments

=over

=item hostobj

The hostrec object of the resource

=item username

Username with which to log into the resource

=item password

Password with which to log into the resource

=item prompt_username

Username with which to log into the systemshell

=item prompt_password

Username with which to log into the systemshell

=item connid

The connection ID

=item conntype

Connection type with which to connect

=item connection

connectrec object for an existing connection to the resource

=back

=item Returns

A reference to a new 7Mode CLI Session class.

=back

=head1 METHODS

=head2 send_command

Send the command to the systemshell.  The output of the command is returned in a
NACL::APISet::Response::CLI::Freebsd object.

=over

=item Synopsis

$response = $session->send_command( command => "raw command string" );

=item Arguments

=over

=item command

The text of the command to send.

=item command_username

For commands to be run as that user

=item answer_password

Boolean value indicating whether password prompt must be answered or not.

=back

=back

=head1 AUTHOR/MAINTAINER

=over

=item NACL Development (dl-nacl-dev@netapp.com)

=back

=cut

