#
# Copyright (c) 2016,... NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Class to hold all connection information for AVAAdmin Sessions
## @author nsesha@netapp.com, dl-nacl-dev@netapp.com
## @status Shared
## @pod here

package NACL::APISet::Session::CLI::AVAAdmin;

use strict;
use warnings;

use base qw(NACL::APISet::Session::CLI);

use NACL::APISet::Response::CLI::AVAAdmin;
use NACL::APISet::_Mixins::CLI qw(_throw_timeout_exception);
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Params::Validate qw(:all);
use Data::Dumper;
use Time::HiRes qw(time);

use NATE::Exceptions::Argument qw(:try);

sub new {
    $Log->enter() if $may_enter;
    $Log->debug(Dumper(@_));

    # an instance of "this" is always the first arg
    # the second arg is a set of named arguments used to init
    # a hash.
    my ($class, %args) = @_;

    #
    # The conntype parameter should be specified ONLY if connid is also
    # provided
    #
    if ((defined $args{conntype}) and (not defined $args{connid})) {
        NATE::Exceptions::Argument->throw(
            "Conntype provided but connid not provided");
    }

    # Fill the member vars for this instance
    #my $self = validate(@_, {  });
    my $self = NACL::APISet->validate_params(
        params => \%args,
        spec   => {
            username => {type => SCALAR, optional => 1, default => "admin"},
            password =>
                {type => SCALAR, optional => 1, default => "netapp123"},
            conntype => {
                type     => SCALAR,
                optional => 1,
                default  => 'ssh-persist',
            },
            ipaddress => {type => SCALAR, optional => 1},
            connid    => {
                type     => SCALAR,
                optional => 1,
                default  => 'ssh-persist',
            },
            connection => {isa => "Connectrec", optional => 1},
            hostobj    => {isa => "Hostrec",    optional => 1},
        },
    );

    #get password from natehost file
    if (defined $self->{hostobj} &&
        defined $self->{hostobj}->{default_password}) {

        my $pwd = $self->{hostobj}->{default_password};
        $Log->debug("Getting password from natehost file:$pwd");
        $self->{password} = $pwd;
    }

    # Add more variables that are sub-class specific.

    # This tags the hash pointed to by $self as a
    # type of AVAAdmin
    bless $self, $class;

    # Connect to the service
    $self->connect();
    # Prepare it for use:
    $self->prep_service;

    # And give the caller the referece to the bless object.
    $Log->exit() if $may_exit;
    return $self;
} ## end sub new

##############################################################################
# Method:    prep_service
# Objective: Prepare the service for commands
# Details:   Send any initial-connection commands to the AVA appliance
##############################################################################
sub prep_service {
    $Log->enter;
    my ($self, %args) = @_;

    $args{on_timeout} = $self->_throw_timeout_exception;
    $args{on_idle}    = $self->_throw_timeout_exception;

    my $connection = $self->{connection};
    my @commands = (
        'show version',
        'enable',
        'configure terminal',
        'no cli session auto-logout',
        'cli session terminal type dumb',
        'cli session terminal length 0',
    );

    for my $cmd (@commands) {
        $connection->execute(command => $cmd, %args);
    }

    $Log->exit;
}

##############################################################################
# Method:    send_command
# Objective: Invoke the command on the AVAAdmin system
# Details:   See POD documentation at the end of this file.
##############################################################################
sub send_command {
    $Log->enter() if $may_enter;

    my ($self, %args) = @_;
    my %connectrec;

    if ($args{connectrec}) {
        %connectrec = %{$args{connectrec}};
        $Log->debug("Arguments passed to Connectrec" . Dumper(\%connectrec));
    }
    $self->_get_scaled_timeout_value(timeout => \$connectrec{timeout});

    my ($actual_command, $raw_execution_time);

    my $time_start = time();
    my $execution_status;
    my $output = $self->{connection}->execute(
        command => $args{command},
        %connectrec,
    );
    my $time_end       = time();
    my $execution_time = $time_end - $time_start;

    my $parser_path = "NACL::APISet::Response::CLI::AVAAdmin";

    eval "use $parser_path";
    if ($@) {
        NATE::BaseException->throw("\nInternal Error in APISet Layer: $@");
    }

    $Log->exit() if $may_exit;

    return $parser_path->new(
        output             => $output,
        command            => $args{command},
        method_name        => $args{method_name},
        execution_time     => $execution_time,
        execution_status   => $execution_status,
        raw_execution_time => $raw_execution_time,
    );

} ## end sub send_command

1;

##############################################################################
#                                    POD here
##############################################################################

=head1 NAME

NACL::APISet::Session::CLI::AVAAdmin - Session class for AVAAdmin CLI connections

=head1 SYNOPSIS

 use NACL::APISet::Session::CLI::AVAAdmin;

=head1 DESCRIPTION

A C<NACL::APISet::Session::CLI::AVAAdmin> provides the methods needed to control the connection
to a AVAAdmin system.

=head1 EXCEPTIONS

=over

=item NATE::Exceptions::Argument

=back

=head1 CONSTRUCTORS

=head2 new

Returns a reference to a newly created C<NACL::APISet::Session::CLI::AVAAdmin>.

=over

=item Synopsis

 $session = NACL::APISet::Session::CLI::AVAAdmin->new(name=>$name, %args);

=item Arguments

=over

=item name

The AVAAdmin host for this session.

=item username

(Optional)
Default "admin".  The AVA shell user.

=item password

(Optional)
Default "netapp123".  The AVA shell user's password.

=item conntype

(Optional)
Default "ssh-persist".  The Connectrec conntype to use for this session.

=item connid

(Optional)
Default "ssh-persist".  The Connectrec connid to use for this session.

=item connection

(Optional)
An established NATE::Connectrec object to be use for this session.

=back

=item Returns

A reference to a new Session.

=back

=head1 METHODS

=head2 send_command

Send the command to the node.  The output of the command is returned in a NACL::APISet::Response::CLI::AVAAdmin
object.

=over

=item Synopsis

$response = $session->send_command(command => "raw command string" );

=item Arguments

=over

=item command

The text of the command to send.

=item Connectrec options

Options can be passed to Connectrec->execute. To

=back

=back

=head1 AUTHOR/MAINTAINER

=over

=item NACL Development (dl-nacl-dev@netapp.com)

=back

=cut
