#
# Copyright (c) 2001-2017 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Class to hold all connection information for CMode Sessions
## @author olalde@netapp.com, dl-nacl-dev@netapp.com
## @status Shared
## @pod here

# Not all 'connections' to services will be in the form of NATE::Connectionrec
# The Session class will define all data needed for connections, whether
# NATE Connection or not.

package NACL::APISet::Session::CLI::CMode;


use strict;
use warnings;
use feature 'state';

use NACL::GeneralUtils qw(_log_profiling_data);

use base qw(NACL::APISet::Session::CLI);

use NACL::APISet::Response::CLI::CMode;

use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Params::Validate qw(:all);
use Data::Dumper;
use Time::HiRes qw(time);
use NATE::Exceptions::Argument qw(:try);
use NACL::APISet::_Mixins::CLI qw(_throw_timeout_exception _is_contain_question_mark);
use NACL::APISet::_Mixins::CLIZAPI qw(_handle_baseline_errors :baseline_opts);

sub new {
    $Log->enter() if $may_enter;
    
    # an instance of "this" is always the first arg
    # the second arg is a set of named arguments used to init
    # a hash.
    my ($class, %args) = @_;
    
    $Log->debug( sub { Dumper(\%args) } );
    
    my $self = $class->SUPER::new(
        %args,
        additional_spec => {
            connid        => { type => SCALAR, default => 'c-login' },
            CLI_separator => { type => SCALAR, default => 'X;X' },
        }
    );

    $Log->exit() if $may_exit;
    return $self;
}

##############################################################################
# Method:    prep_service
# Objective: Prepare the service for commands
# Details:   This may be an initial stanza needed to establish a session or some
# up front command needed to prep the session, whatever.  In Cluster Mode
# this is the act of turning off prompts and requesting all fields.
##############################################################################
sub prep_service {
    $Log->enter() if $may_enter;

    my ($self, %args) = @_;
    my $CLI_separator = $self->{CLI_separator};
    $args{"on_timeout"}
        = $self->_throw_timeout_exception();
    $args{"on_idle"}
        = $self->_throw_timeout_exception();

    my $connection = $self->{"connection"};
    $connection->execute( command => "set -confirmations off", %args);
    $connection->execute(
            command => "set -showallfields true -showseparator \"$CLI_separator\" "
          . "; system timeout modify -timeout 0 \; rows 0", %args);
    $Log->exit() if $may_exit;
}

##############################################################################
# Method:    send_command
# Objective: Invoke the command on the CMode filer
# Details:   See POD documentation at the end of this file.
##############################################################################
sub send_command {
    $Log->enter() if $may_enter;
    my ($self, %opts) = @_;
    state $spec = {connectrec => {type => HASHREF, default => {}}};
    my %args = validate_with(
        params      => \%opts,
        spec        => $spec,
        allow_extra => 1
    );

    my %connectrec    = %{delete $args{connectrec}};
    my %baseline_args = $self->_move_baseline_args(source => \%args);
    
    my ($actual_set);
    if (exists $connectrec{actual_set} and $connectrec{actual_set} eq "SystemshellNGSH") {
        $actual_set = delete $connectrec{actual_set};
    }
    
    $self->_get_scaled_timeout_value(timeout => \$connectrec{timeout});
    unless ($args{method_name}) {
        # Caller is execute_command, all the options passed
        # should be considered as connectrec options
        %connectrec = ( %connectrec, %args );
        $args{command} ||= $connectrec{command};
        delete $connectrec{command};
    }
    
    $Log->debug( sub { "Arguments that are being passed to execute(): ". "command = '".
                $args{command}. "'\n". Dumper(\%connectrec) } );
    
    my $resp;
    my $execute_command_form_response = sub {
        my ( $time_start, $output, $time_end );
        my $contain_question_mark = 0;
        if ($self->_is_contain_question_mark(command => $args{command})) {
            $contain_question_mark = 1;
            $self->{connection}->execute(
                command => "set -active-help false",
                %connectrec,
            );
        }
    my $retry_count = 3;
AGAIN: {
        try {
            $time_start = time();
            $output = $self->{connection}->execute(
                command => $args{command},
                %connectrec,
            );
            $time_end = time();
        } finally {
            if($contain_question_mark == 1) {
                $self->{connection}->execute(
                    command => "set -active-help true",
                    %connectrec,
                );
            }
        };
        
        my $execution_time = $time_end - $time_start;
        
        # for execute_raw_command, just return the raw output
        if(!$args{method_name}) {
            $Log->exit() if $may_exit;
            return $output;
        }
        
        # With a connection to Cmode shell, the context of the commands can
        # be nodescope, cluster shell, vserver or systemshell. 
        # If the context is cluster shell or vserver then response object 
        # should be of type CMode. For systemshell and nodescope context, we will
        # create a appropriate session subclass of CMode and hence from this session
        # class, the response object returned will always be of type CMode.
        if ( Tharn::param('PROFILING_ENABLED') ) {
            my $node = $self->{connection}->hostp()->hostname();
            $self->_log_profiling_data(
                node           => $node,
                start_time     => $time_start,
                end_time       => $time_end,
                elapsed_time   => $execution_time,
                command        => $args{command},
                interface      => 'cli'
            );
        }

        
        if (defined $actual_set and $actual_set eq "SystemshellNGSH") {
            $resp = NACL::APISet::Response::CLI::Freebsd->new(
                        output            => $output,
                        actual_command    => $args{actual_command},
                        command           => $args{command},
                        cdef_path         => $args{cdef_path},
                        method_name       => $args{method_name},
                        execution_status  => 0,
                        execution_time    => $execution_time,
                        parser_file       => $args{parser_file},
                    );
            # Setting execution_status to undef so that user doesn't have wrong idea that
            # the execution status is ZERO
            $resp->{execution_status} = undef;  
        } else {
	    try {
		    $resp = NACL::APISet::Response::CLI::CMode->new(
			output         => $output,
			command        => $args{command},
			CLI_separator  => $self->{CLI_separator},
			execution_time => $execution_time,
			method_name    => $args{method_name},
			output_compat_mirror => $args{output_compat_mirror},
		    );
	    } catch NACL::APISet::Exceptions::ResponseException with {
		my $ex = shift;
		if ( ( $ex->text() =~ /Error while parsing CMode CLI output!\nFailed to handle embedded newlines in the CLI output/i) && ( $retry_count > 0 ) ) {
		    # as per burt#1014604, when we report FAIL: 'Error while parsing CMode CLI output!
		    # * Failed to handle embedded newlines in the CLI output' retry the command
		    # up to 3 times first before just Failing
		    $retry_count--;
		    redo AGAIN;
		}
		$ex->throw();
	    }
        }
} # AGAIN
        return;
    };
    
    my $return_value = $self->_handle_baseline_errors(
        code => $execute_command_form_response, %baseline_args);
    if (defined $return_value) {
        # This will be populated for execute_raw_command
        $Log->exit() if $may_exit;
        return $return_value;
    }
    
    $Log->exit() if $may_exit;
    return $resp;
}

### Connect to the service (the service provider really)
# Use the existing passed ConnectRec reference and verify connection.
# Build the connection using passed parameters
# Build the connection using natehost file and passed overrides
sub connect {
    $Log->enter() if $may_enter;
    
    my ($self, %args) = @_;

    #
    # When a connid is provided, the connection layer automatically
    # determines a conntype to use. For rsh and ssh, it defaults the 
    # conntype to "rsh-persay" and "ssh-persay" respectively. Since
    # CMode APISets do not work for persay connections, we need to 
    # choose conntype as "rsh-persist" for connid "rsh" and  
    # "ssh-persist" for connid "ssh" to ensure it works correctly.
    # For all other connections, we leave it to the connection layer
    # to determine the appropriate conntype.
    #       
    unless ($self->{conntype}) {
        if($self->{connid} eq "ssh") {
            $self->{conntype}= "ssh-persist";
        } elsif($self->{connid} eq "rsh") {
            $self->{conntype}= "rsh-persist";
        }
    }
    $self->SUPER::connect(%args);
    $Log->exit() if $may_exit;
}

##############################################################################
#                                    POD here
##############################################################################

=head1 NAME

NACL::APISet::Session::CLI::CMode - Session class for cluster mode CLI connections

=head1 SYNOPSIS

 use NACL::APISet::Session::CLI::CMode;

=head1 DESCRIPTION

A C<NACL::APISet::Session::CLI::CMode> provides the methods needed to control the connection
to a cluster mode node.  

=head1 EXCEPTIONS

=over

=item NATE::Exceptions::Argument

=back

=head1 CONSTRUCTORS

=head2 new

Returns a reference to a newly created C<NACL::APISet::Session::CLI::CMode>.  

=over 

=item Synopsis

 $session = NACL::APISet::Session::CLI::CMode->new(name=>$name, %args);

=item Arguments

=over 

=item name

The cluster mode node for this session.

=item username

(Optional)
Default "admin".  The ngsh shell user.

=item password

(Optional)
Default "netapp1!".  The ngsh shell user's password.

=item conntype

(Optional)
Default "ssh-persist".  The Connectrec conntype to use for this session.

=item connid

(Optional)
Default "ssh".  The Connectrec connid to use for this session.

=item CLI_separator

(Optional)
Default "X;X".  The string to use for setting the ngsh -showseparator option.

=item connection

(Optional)
An established NATE::Connectrec object to be use for this session.

=back

=item Returns

A reference to a new Session.

=back

=head1 METHODS

=head2 prep_service

After connecting, initialize the connection.  In the cluster mode CLI case, this means 
preparing the ngsh session the show all fields and separate fields in a predictable way.

=over
 
=item Synopsis

$session->prep_service();

=back

=head2 send_command

Send the command to the node.  The output of the command is returned in a NACL::APISet::Response::CLI::CMode
object.

=over 

=item Synopsis

$response = $session->send_command(command => "raw command string" );

=item Arguments

=over 

=item command

The text of the command to send.

=item Connectrec options

Options can be passed to Connectrec->execute. To 

=back

=back

=head1 AUTHOR/MAINTAINER

=over

=item NACL Development (dl-nacl-dev@netapp.com)

=back

=cut

1;
