#
# Copyright (c) 2001-2009 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Class to hold all connection information for DFM Sessions
## @author tanzim.rahmathullah, dl-sm-nacl-dev@netapp.com
## @status Shared
## @pod here

# Not all 'connections' to services will be in the form of NATE::Connectionrec
# The Session class will define all data needed for connections, whether
# NATE Connection or not.

package NACL::APISet::Session::CLI::DFM;

use strict;
use warnings;

use base qw(NACL::APISet::Session::CLI);

use NACL::APISet::Response::CLI::DFM;

use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Params::Validate qw(:all);
use Data::Dumper;
use Time::HiRes qw(time);

use NATE::Exceptions::Argument qw(:try);

sub new {
    $Log->enter() if $may_enter;
    $Log->debug( Dumper(@_) );

    # an instance of "this" is always the first arg
    # the second arg is a set of named arguments used to init
    # a hash.
    my ( $class, %args ) = @_;

    #
    # The conntype parameter should be specified ONLY if connid is also
    # provided
    #
    if ( ( defined $args{conntype} ) and ( not defined $args{connid} ) ) {
        NATE::Exceptions::Argument->throw(
            "Conntype provided but connid not provided");
    }

    # Fill the member vars for this instance
    #my $self = validate(@_, {  });
    my $self = NACL::APISet->validate_params(
        params => \%args,
        spec   => {
            username         => { type => SCALAR,         optional => 1 },
            password         => { type => SCALAR,         optional => 1 },
            conntype         => { type => SCALAR,         optional => 1 },
            ipaddress        => { type => SCALAR,         optional => 1 },
            dfm_version_info => { type => SCALAR | UNDEF, optional => 1 },
            connid           => {
                type     => SCALAR,
                optional => 1,
                default  => "libssh-persist"
            },
            connection => { isa => "Connectrec", optional => 1 },
            hostobj    => { isa => "Hostrec",    optional => 1 },
        },
    );

    # Add more variables that are sub-class specific.

    # This tags the hash pointed to by $self as a
    # type of DFM
    bless $self, $class;

    # Connect to the service
    $self->connect();

    # And give the caller the referece to the bless object.
    $Log->exit() if $may_exit;
    return $self;
}
##############################################################################
# Method:    send_command
# Objective: Invoke the command on the DFM system
# Details:   See POD documentation at the end of this file.
##############################################################################
sub send_command {
    $Log->enter() if $may_enter;

    my ( $self, %args ) = @_;
    my %connectrec;

    if ( $args{connectrec} ) {
         %connectrec = %{$args{connectrec}};
         $Log->debug("Arguments passed to Connectrec" . Dumper(\%connectrec));
    }
    $self->_get_scaled_timeout_value(timeout => \$connectrec{timeout});
    if (!$args{method_name} ) {
        $args{command} = delete $connectrec{command};
        delete $connectrec{answer_password};
    }

    my ( $actual_command, $raw_execution_time );

    if ( defined( $ENV{DFM_API_PERF_FLAG} )
        && $ENV{DFM_API_PERF_FLAG} =~ /TRUE/i )
    {

        $Log->debug("The DFM_API_PERF_FLAG is set to TRUE.");

        # Append and prepend the perl time command
        $actual_command = $args{command};

        #Escaping the quotes present in the command, if any
        $args{command} =~ s/"/\\"/g;

        $args{command}
            = "perl -e 'use Time::HiRes  qw( gettimeofday ); my \$start = gettimeofday;"
            . "print \"<<<<<\"; system(\"$args{command}\"); print \">>>\$?\";"
            . "my \$end = gettimeofday; print \"<<<\".(\$end-\$start).\">>>\"; '";

    }
    my $time_start = time();
    my $execution_status;
    my $output = $self->{connection}->execute(
        command => $args{command},
        %connectrec, get_status => \$execution_status
    );
    my $time_end       = time();
    my $execution_time = $time_end - $time_start;

    $output =~ s/__status__=(\d+)\s+//;

    if ( defined( $ENV{DFM_API_PERF_FLAG} )
        && $ENV{DFM_API_PERF_FLAG} =~ /TRUE/i )
    {
        $output =~ s/.*<<<<<(.*)>>>(.+?)<<<(.+?)>>>/$1/s;

        # Extract the execution status
        $execution_status = $2;

        # Extract the time taken
        $raw_execution_time = $3;

        #Put back the original command
        $args{command} = $actual_command;

    } else {

        $raw_execution_time = undef;
    }
    if ( !defined $self->{dfm_version_info} ) {
        
        $self->{dfm_version_info} = "dfm3";
    }
    
    if (!$args{method_name} ) {
        return $output;
    }
    
    my $release     = uc( $self->{dfm_version_info} );
    my $parser_path = "NACL::APISet::Response::CLI::DFM::" . $release;

    eval "use $parser_path";
    if ($@) {
        NATE::BaseException->throw("\nInternal Error in APISet Layer:$@");
    }

    $Log->exit() if $may_exit;

    return $parser_path->new(
        output             => $output,
        command            => $args{command},
        cdef_path          => $args{cdef_path},
        method_name        => $args{method_name},
        execution_time     => $execution_time,
        execution_status   => $execution_status,
        raw_execution_time => $raw_execution_time,
    );

}

1;

##############################################################################
#                                    POD here
##############################################################################

=head1 NAME

NACL::APISet::Session::CLI::DFM - Session class for DFM CLI connections

=head1 SYNOPSIS

 use NACL::APISet::Session::CLI::DFM;

=head1 DESCRIPTION

A C<NACL::APISet::Session::CLI::DFM> provides the methods needed to control the connection
to a DFM system.  

=head1 EXCEPTIONS

=over

=item NATE::Exceptions::Argument

=back

=head1 CONSTRUCTORS

=head2 new

Returns a reference to a newly created C<NACL::APISet::Session::CLI::DFM>.  

=over 

=item Synopsis

 $session = NACL::APISet::Session::CLI::DFM->new(name=>$name, %args);

=item Arguments

=over 

=item name

The DFM host for this session.

=item username

(Optional)
Default "admin".  The ngsh shell user.

=item password

(Optional)
Default "netapp1!".  The ngsh shell user's password.

=item conntype

(Optional)
Default "ssh-persist".  The Connectrec conntype to use for this session.

=item connid

(Optional)
Default "ssh".  The Connectrec connid to use for this session.

=item CLI_separator

(Optional)
Default "X;X".  The string to use for setting the ngsh -showseparator option.

=item connection

(Optional)
An established NATE::Connectrec object to be use for this session.

=back

=item Returns

A reference to a new Session.

=back

=head1 METHODS

=head2 send_command

Send the command to the node.  The output of the command is returned in a NACL::APISet::Response::CLI::DFM
object.

=over 

=item Synopsis

$response = $session->send_command(command => "raw command string" );

=item Arguments

=over 

=item command

The text of the command to send.

=item Connectrec options

Options can be passed to Connectrec->execute. To 

=back

=back

=head1 AUTHOR/MAINTAINER

=over

=item NACL Development (dl-sm-nacl-dev@netapp.com)

=back

=cut
