#
# Copyright (c) 2001-2016 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Class to hold all connection information for Firmware Sessions
## @author olalde@netapp.com, anerudh@netapp.com, dl-nacl-dev@netapp.com
## @status Shared
## @pod here

# Not all 'connections' to services will be in the form of NATE::Connectionrec
# The Session class will define all data needed for connections, whether
# NATE Connection or not.

package NACL::APISet::Session::CLI::Firmware;

use NACL::APISet::Response::CLI::Firmware;

#
# Compiler Directives
#
use strict;
use warnings;

use base qw(NACL::APISet::Session::CLI);

# Module import - APISets
#use NACL::APISet::Response::CLI::Firmware;

#
# NATE imports
#
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Params::Validate qw(:all);
use Data::Dumper;
use Time::HiRes qw(time); # This module is for measuring time in high precision

#
# Exceptions imported
#
use NATE::Exceptions::Argument qw(:try);
use NACL::APISet::Exceptions::ConnectionFailedException qw(:try);

sub new {
    $Log->enter() if $may_enter;
    $Log->debug( Dumper(@_) );

    # an instance of "this" is always the first arg
    # the second arg is a set of named arguments used to init
    # a hash.
    my ($class, %args) = @_;

    #
    # The conntype parameter should be specified ONLY if connid is also
    # provided
    #
    if ((defined $args{conntype}) and (not defined $args{connid})) {
        NATE::Exceptions::Argument->throw(
        "Conntype provided but connid not provided");
    }

    # Fill the member vars for this instance

    my $self = NACL::APISet->validate_params(
        params => \%args,
        spec   => {
            hostobj   => { isa => "Hostrec", optional => 1 },
            conntype => { type => SCALAR, optional => 1 },
            connid   => { type => SCALAR, optional => 1, default => "console" },
            connection => { isa => "Connectrec", optional => 1 },
        },
    );

    # Add more variables that are sub-class specific.

    # This tags the hash pointed to by $object as a
    # type of Firmware
    bless $self, $class;

    # Pass the firmware prompt as an additional parameter to connect() in
    # Session/CLI.pm but only if the hostobj is defined (when user does not
    # pass in Connectrec object directly).

    my %additional_parameters;

    if ( ref ($self->{hostobj}->{prompt_fw}) eq 'ARRAY'  &&
         ref($self->{hostobj}->{prompt_norm}) eq 'ARRAY' ) {
        %additional_parameters = ( prompts => [
                                        @{ $self->{hostobj}->{prompt_fw} },
                                        @{ $self->{hostobj}->{prompt_norm} }
                                    ]
                                 );
    }

    # Connect to the service

    $self->connect(%additional_parameters);

    # If the user has passed in a Connection object, then we have to manually
    # add the FW prompt also to the list of prompts in the Connectrec object.
    # We do this by extrcting it from the Hostrec object within the Connectrec
    # object. Adding a duplicate entry for this prompt is not an issue.
    # So we do this irrespective of how the Connectrec object was created.

    $self->{connection}->prompts($self->{connection}->prompts(),
                                 $self->{connection}->hostp()->prompt_fw());

    # Prepare the service for use
    $self->prep_service();

    # And give the caller the referece to the bless object.
    $Log->exit() if $may_exit;
    return $self;
}

###############################################################################
# Method:    prep_service
# Objective: Prepare the service for sending commands and reading output
# Details:   See POD documentation at the end of this file.
###############################################################################
sub prep_service {
    $Log->enter() if $may_enter;
    my $self = shift;
    # Lines indicating the establishment of the connection should not be read
    # while reading the output of the first command sent. So send a carriage
    # return which causes a prompt to appear which ensures that the output of
    # the next command sent is read from here onwards
    foreach (1..2) {
        # probe the connection (keep poking every 5 seconds if it is not responding)
        $self->{connection}->execute(command => "",
                                     max_idle => 5,
                                     on_idle_wakeup => sub {""});
        $self->{connection}->discardbuff();
        sleep 1;
    }

    # Burt856301, set env LINES to large number
    $self->{connection}->execute(command => "setenv LINES 9999");

    $Log->exit() if $may_exit;
}

###############################################################################
# Method:    send_command
# Objective: Execute the Firmware command to the filer and gather the raw output
# Details:   See POD documentation at the end of this file.
###############################################################################
sub send_command {
    $Log->enter() if $may_enter;

    my ($self, %opts) = @_;
    my %args = validate_with(
        params      => \%opts,
        spec        => { connectrec => { type => HASHREF, default => {} } },
        allow_extra => 1
    );

    my %connectrec = %{delete $args{connectrec}};
    $self->_get_scaled_timeout_value(timeout => \$connectrec{timeout});
    my $parser_file = $args{parser_file};
    $Log->debug( sub { "Session/CLI/Firmware:send_command: " . Dumper( \%args ) } );

    unless ($args{method_name}) {
        # Caller is execute_command, all the options passed
        # should be considered as connectrec options
        %connectrec = ( %connectrec, %args );
        $args{command} ||= $connectrec{command};
        delete $connectrec{command};
    }

    $Log->debug( sub { "Arguments that are being passed to execute(): ". "command = '".
                $args{command}. "'\n". Dumper(\%connectrec) } );

    my $time_start = time();
    my $output;
    if ($connectrec{timeout} == 0){
        $self->{connection}->execute_start(command => $args{command},);
	$output = "";
    } else{
	$output = $self->{connection}->execute(command => $args{command},
    	                  %connectrec);
    }

    my $time_end = time();

    my $execution_time = $time_end - $time_start;

    # for execute_raw_command and execute_command, just return the raw output
    if( !$args{method_name} ) {
        $Log->exit() if $may_exit;
        return $output;
    }

    my $response_object =  NACL::APISet::Response::CLI::Firmware->new(
            output            => $output,
            command           => $args{command},
            cdef_path         => $args{cdef_path},
            method_name       => $args{method_name},
            execution_time    => $execution_time,
            parser_file       => $parser_file
        );

    $Log->exit() if $may_exit;
    return $response_object;
}

1;

##############################################################################
#                                    POD here
##############################################################################

=head1 NAME

NACL::APISet::Session::CLI::Firmware - Firmware CLI Session class

=head1 SYNOPSIS

use NACL::APISet::Session::CLI::Firmware;

=head1 DESCRIPTION

A C<NACL::APISet::Session::CLI::Firmware> object is used to establish a connection
and send commands for the Firmware CLI APISet.

=head1 EXCEPTIONS

=over

=item NATE::Exceptions::Argument

=item NACL::APISet::Exceptions::ConnectionFailedException

=back

=head1 CONSTRUCTORS

=head2 new

Returns a reference to a newly created C<NACL::APISet::Session::CLI::Firmware>.

=over

=item Synopsis

 $self->{session} = NACL::APISet::Session::CLI::Firmware->new( %args );

=item Arguments

=over

=item hostobj

The hostrec object of the resource

=item connid

The connection ID

=item conntype

Connection type with which to connect

=item connection

connectrec object for an existing connection to the resource

=back

=item Returns

A reference to a new Firmware CLI Session class.

=back

=head1 METHODS

=head2 prep_service

Prepare the service for use - Send a carriage return and make the Firmware
prompt appear so that the output of the first command sent is read from that
prompt onwards and does not contain the messages indicating the establishment of
the console connection

=over

=item Synopsis

$session->prep_service();

=item Arguments

None

=item Returns

Nothing

=back

=head2 send_command

Send the command to the node.  The output of the command is returned in a NACL::APISet::Response::CLI::Firmware
object.

=over

=item Synopsis

$response = $session->send_command( command => "raw command string", %other_connection_overrides );

=item Arguments

=over

=item command

The text of the command to send.

=back

=back

=head1 AUTHOR/MAINTAINER

=over

=item NACL Development (dl-nacl-dev@netapp.com)

=back

=cut
