#
# Copyright (c) 2001-2009 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Class to hold all connection information for Freebsd Sessions
## @author Isshwarya.M@netapp.com, anerudh@netapp.com, dl-nacl-dev@netapp.com
## @status Shared
## @pod here

# Not all 'connections' to services will be in the form of NATE::Connectionrec
# The Session class will define all data needed for connections, whether
# NATE Connection or not.

package NACL::APISet::Session::CLI::Freebsd;

use strict;
use warnings;

use base qw(NACL::APISet::Session::CLI);

use NACL::APISet::Response::CLI::Freebsd;

use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Time::HiRes qw(time);
use Params::Validate qw(:all);
use Data::Dumper;

use NATE::Exceptions::Argument qw(:try);
use NACL::APISet::Exceptions::ConnectionFailedException qw(:try);

sub new {
    $Log->enter() if $may_enter;
    $Log->debug( Dumper(@_) );

    # an instance of "this" is always the first arg
    # the second arg is a set of named arguments used to init
    # a hash.
    my ($class, %args) = @_;

    if ((defined $args{conntype}) and (not defined $args{connid})) {
        NATE::Exceptions::Argument->throw(
        "Conntype provided but connid not provided");
    }

    # Fill the member vars for this instance
    my $self = NACL::APISet->validate_params(
        params => \%args,
        spec   => {
            hostobj  => { isa => "Hostrec", optional => 1 },
            username => { type => SCALAR, optional => 1 },
            password => { type => SCALAR, optional => 1 },
            conntype => { type => SCALAR, optional => 1 },
            connid   => { type => SCALAR, optional => 1, default => "login" },
            ipaddress => { type => SCALAR, optional => 1 },
            connection => { isa => "Connectrec", optional => 1 },
        },
    );

    # Add more variables that are sub-class specific.

    # This tags the hash pointed to by $object as a
    # type of Freebsd
    bless $self, $class;

    # Connect to the service
    $self->connect();

    # Prepare the service for use
    $self->prep_service();

    # And give the caller the referece to the bless object.
    $Log->exit() if $may_exit;
    return $self;
}

### Execute the command and gather the raw output
sub send_command {
    $Log->enter() if $may_enter;

    my ( $self, %args ) = @_;
    $Log->debug( "Session/CLI/Freebsd:send_command: " . Dumper( \%args ) );
    
    my $do_not_answer; # for answering the password prompt
    my @password_match_table = ('.*[Pp]assword.*:\s?'=>
                                 sub {
                                    return undef if $do_not_answer;
                                    $do_not_answer = 1;
                                    return $self->{connection}->{password};
                                 }) if $args{answer_password};

    my %connectrec = %{delete $args{connectrec}};
    $self->_get_scaled_timeout_value(timeout => \$connectrec{timeout});
    my $parser_file = $args{parser_file};

    my $execution_status;
    #this match table is used only during sudo
    

    $connectrec{'match_table'} = [@password_match_table,
                                  exists($connectrec{'match_table'}) ?
                                  @{$connectrec{match_table}} : () ];
                                  
    my $time_start = time();

    my $output = $self->{connection}->execute(command=> $args{command},
                                              command_username=>
                                                        $args{command_username},
                                              get_status=>\$execution_status,
                                              %connectrec);
    my $time_end = time();

    my $execution_time = $time_end - $time_start;

    if (!defined $execution_status) {
        # Execution status will be undef only if there is some problem with
        # establishing the connection itself.
        NACL::APISet::Exceptions::ConnectionFailedException->throw($output);
    }

    $output =~ s/.*[Pp]assword.*// if $do_not_answer; # if password was answered
    $output =~ s/__status__=(\d+)\s+//;
    
    if( !$args{method_name} ) {
        $Log->exit() if $may_exit;
        return $output;
    }

    my $response_object = NACL::APISet::Response::CLI::Freebsd->new(
            output            => $output,
            actual_command    => $args{actual_command},
            command           => $args{command},
            cdef_path         => $args{cdef_path},
            method_name       => $args{method_name},
            execution_status  => $execution_status,
            execution_time    => $execution_time,
            parser_file       => $parser_file
    );
        
    $Log->exit() if $may_exit;
    return $response_object;
}

1;

##############################################################################
#                                    POD here
##############################################################################

=head1 NAME

NACL::APISet::Session::CLI::Freebsd - Freebsd CLI Session class

=head1 SYNOPSIS

 use NACL::APISet::Session::CLI::Freebsd;

=head1 DESCRIPTION

A C<NACL::APISet::Session::CLI::Freebsd> object is used to establish a connection
and send commands for the Freebsd CLI APISet.

=head1 EXCEPTIONS

=over

=item NATE::Exceptions::Argument

=item NACL::APISet::Exceptions::ConnectionFailedException

=back

=head1 CONSTRUCTORS

=head2 new

Returns a reference to a newly created C<NACL::APISet::Session::CLI::Freebsd>.

=over

=item Synopsis

 $self->{session} = NACL::APISet::Session::CLI::Freebsd->new( %args );

=item Arguments

=over

=item hostobj

The hostrec object of the resource

=item username

Username with which to log into the resource

=item password

Password with which to log into the resource

=item connid

The connection ID

=item conntype

Connection type with which to connect

=item connection

connectrec object for an existing connection to the resource

=back

=item Returns

A reference to a new Freebsd CLI Session class.

=back

=head1 METHODS

=head2 send_command

Send the command to the Freebsd host. The output of the command is returned in
a NACL::APISet::Response::CLI::Freebsd object.

=over

=item Synopsis

$response = $session->send_command( command => "raw command string" );

=item Arguments

=over

=item command

The text of the command to send.

=item command_username

For commands to be run as that user

=item answer_password

Boolean value indicating whether password prompt must be answered or not.

=back

=back

=head1 AUTHOR/MAINTAINER

=over

=item NACL Development (dl-nacl-dev@netapp.com)

=back

=cut

1;
