#
# Copyright (c) 2001-2012 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#
## @summary Class to hold all connection information for Kerberos Sessions
## @author dl-nacl-dev@netapp.com
## @status Shared
## @pod here
# Not all 'connections' to services will be in the form of NATE::Connectionrec
# The Session class will define all data needed for connections, whether
# NATE Connection or not.
package NACL::APISet::Session::CLI::Kerberos;
use NACL::APISet::Exceptions::ConnectionFailedException;
use NATE::Exceptions::Argument qw(:try);

#use NATE::BaseException;
use strict;
use warnings;
use base qw(NACL::APISet::Session::CLI);
use NACL::APISet::Response::CLI::Kerberos;
use Data::Dumper;
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Time::HiRes qw(time);
use Params::Validate qw(:all);
###############################################################################
## Method:    new
## Objective: Creates Kerberos session class object. This also handles the
##            arguments to be passed to Connectrec::new which determines how
##            the connection should be started
###############################################################################
sub new {
    $Log->enter() if $may_enter;
    $Log->debug( Dumper( \@_ ) ) if ( $Log->may_debug() );

    # an instance of "this" is always the first arg
    # the second arg is a set of named arguments used to init
    # a hash.
    my ( $class, %args ) = @_;
    if ( ( defined $args{conntype} ) and ( not defined $args{connid} ) ) {
        NATE::Exceptions::Argument->throw(
                              "Conntype provided but connid is not provided");
    }

    # Fill the member vars for this instance
    my $self =
        NACL::APISet->validate_params(
          params => \%args,
          spec   => {
              hostobj   => { isa  => "Hostrec", optional => 1 },
              username  => { type => SCALAR,    optional => 1 },
              password  => { type => SCALAR,    optional => 1 },
              ipaddress => { type => SCALAR,    optional => 1 },
              conntype =>
                  { type => SCALAR, default => "ssh-persist" },
              connid => { type => SCALAR, default => 'login' },
              context_definition => { type => HASHREF, optional => 1 },
          },
        );
    bless $self, $class;
    if ( exists $self->{connection} ) {
        NATE::Argument->Exception( "Using Connectrec objects is not allowed. "
                                   . "Please use Hostrec objects instead" );
    }
    $self->_login();
    return $self;
}

sub _login {
    $Log->enter() if $may_enter;
    my $self    = $_[0];
    my $context = $self->{context_definition}->{name}
        if defined $self->{context_definition}->{name};
    $Log->debug("Context to be logged into: $context") if $context;
    try {
        if ($context) {
            if ( $context eq "kadmin" ) {
                $self->_login_to_kadmin();
            } elsif ( $context eq "kadmin.local" ) {
                $context = 'kadmin.local';
                $self->_login_to_kadmin_local();
            }
        } else {
            $context = 'kadmin.local';
            $self->_login_to_kadmin_local();
        }
    }
    catch NACL::APISet::Exceptions::ConnectionFailedException with {
        my $exception_object = shift;
        $Log->exit() if $may_exit;
        NACL::APISet::Exceptions::ConnectionFailedException->throw(
                       "\nError: Connection could not be established."
                     . "\nPlease verify that your machine is up and"
                     . "running and $context is working on your system."
                     . " \nAlso, verify that you are able to manually login"
                     . " to the system using $self->{connid}."
                     . "If none of this works, please contact dl-nacl-dev " );
    };
    $Log->exit() if $may_exit;
}

sub _login_to_kadmin {
    $Log->enter() if $may_enter;
    my $self = $_[0];
    my %connection_overrides;
    $connection_overrides{'initial_command'}
        = $self->{context_definition}->{name};
    $connection_overrides{'prompts'} = [ 'Password.+:', ];
    $connection_overrides{shell_type} = "none";
    $self->connect(%connection_overrides);
    my $password
        = $self->{prompt_password}
        || $self->{password}
        || $self->{connection}->hostp()->default_password();
    $self->_add_prompts( prompts => [qr/kadmin:/] );
    $self->{connection}->execute( command => $password, );
    $Log->exit() if $may_exit;
}

sub _login_to_kadmin_local {
    $Log->enter() if $may_enter;
    my $self = $_[0];
    my %connection_overrides;
    $connection_overrides{shell_type} = "none";
    $connection_overrides{'initial_command'}
        = $self->{context_definition}->{name} || 'kadmin.local';

# While logging into kadmin.local, we get this prompt:
# kadmin.local:  Authenticating as principal root/admin@MITKDC01.NFSDEV.NB.NETAPP.COM with password.
# Hence, we need to push this prompt intially.
    $connection_overrides{prompts} = [ qr/kadmin.*:\s+Authenticating/i, ];
    $self->connect(%connection_overrides);
    my $password = $self->{password}
        || $self->{connection}->hostp()->default_password();
    $self->_add_prompts( prompts => [qr/kadmin.local:/] );
    $Log->exit() if $may_exit;
}

sub send_command {
    $Log->enter() if $may_enter;
    my ( $self, %args ) = @_;
    my %connectrec = %{ delete $args{connectrec} };
    $self->_get_scaled_timeout_value( timeout => \$connectrec{timeout} );
    my $parser_file = $args{parser_file};
    my $command = $args{command} || delete $connectrec{command};

    # For commands that throw an error, the prompt looks like this:
    # kadmin: <execute invalid command>
    # kadmin: kadmin: <error text>
    # So, the <error text> gets prepended to the output of the next command.
    # Hence, we need to discard the contents of the buffer before reading the
    # output.
    $self->{connection}->discardbuff();
    my $time_start = time();
    my $output =
        $self->{connection}->execute( command => $command,
                                      %connectrec, );
    my $time_end       = time();
    my $execution_time = $time_end - $time_start;

    # kadmin prompts look like this:
    # kadmin : <invalid command>
    # <invalid comamnd>
    # kadmin: kadmin: Usage ...
    # The problem is that since are matching for the prompt as kadmin:
    # or kadmin.local:, we need to capture the error message that comes
    # after the prompt. This error message will be stored in $after.
    my $after;
    my $response =
        $self->{connection}->expect( match   => [qr/kadmin.*:|/],
                                     after   => \$after,
                                     timeout => 0,
        );
    $output =~ s/^$after//;

    # Removing the command name from the raw output
    $output =~ s/$command//;
    $output = "$output" . "$after" if $after;
    $Log->debug( sub { Dumper( \$output ); } );
    if ( !$args{method_name} ) {
        $Log->exit() if $may_exit;
        return $output;
    }
    my $response_object =
        NACL::APISet::Response::CLI::Kerberos->new(
                                      output         => $output,
                                      actual_command => $args{actual_command},
                                      command        => $args{command},
                                      cdef_path      => $args{cdef_path},
                                      method_name    => $args{method_name},
                                      execution_time => $execution_time,
                                      parser_file    => $parser_file
        );
    $Log->exit() if $may_exit;
    return $response_object;
}
1;

=head1 NAME

NACL::APISet::Session::CLI::Kerberos - Kerberos CLI Session class

=head1 SYNOPSIS

 use NACL::APISet::Session::CLI::Kerberos;

=head1 DESCRIPTION

A C<NACL::APISet::Session::CLI::Kerberos> object is used to establish a
connection and send commands for the Kerberos CLI APISet.

=head1 EXCEPTIONS

=over

=item NATE::Exceptions::Argument

=item NACL::APISet::Exceptions::ConnectionFailedException

=back

=head1 CONSTRUCTOR

=head2 new

Returns a reference to a newly created C<NACL::APISet::Session::CLI::Kerberos>.

=over

=item Synopsis

 $self->{session} = NACL::APISet::Session::CLI::Kerberos->new( %args );

=item Arguments

=over

=item hostobj

The hostrec object of the resource

=item username

Username with which to log into the resource

=item password

Password with which to log into the resource

=item connid

The connection ID. The default connid is 'login'.

=item conntype

Connection type with which to connect. The default conntype is 'ssh-persist'

=item connection

connectrec object for an existing connection to the resource

=item context_definition

context_definition defines the context (kadmin or kadmin.local) to use for
executing commands. Context_definition is a HASHREF. The HASHREF accepts the
the following parameters as key-value pairs:

=over

=item name

Context to log into. Possible values are kadmin and kadmin.local.
The default is kadmin.local. If the context (kadmin or kadmin.local) is not
specified in the path, then absolute path of the context (Eg: /usr/sbin/kadmin)
needs to be specified.

=item password

Password to be used to log into the B<kadmin> shell (Password is not required
to log into the kadmin.local shell).
If no password is specified, the 'password' override provided in the APISet
constructor will be used. Otherwise, the password from the Hostrec object
will be used.

Example:
If the APISet constructor looks like this:

C< my $Api_Set_Obj = NACL::APISet-E<gt>new ( hostobj =E<gt> $host,>

                                      category => Host,

                                      interface => CLI,

                                      set => Kerberos,

                                      username => "username",

                                      password => "password",

                                      context_definiton => { name => "kadmin",

                                                             password => "password1"
                                                           },
                                            );

Then, 'password' will be used to log into the client and 'password1' will be
used to log into the kadmin shell.

=back

=back

=item Returns

A reference to a new Kerberos CLI Session class.

=back

B<NOTE:> Usage of 'Connectrec' objects is not allowed.

=head1 METHODS

=head2 send_command

Returns the Kerberos response object (NACL::APISet::Response::CLI::Kerberos).

=over

=item Arguments

=over

=item command

The text of the command to send.

=item command_username

For commands to be run as that user

=item answer_password

Boolean value indicating whether password prompt must be answered or not.

=item parser_file

The parser file to be loaded for parsing the raw output

=back

=item Synopsis

$response = $session->send_command(command => " raw command string ");

=item Returns

Returns a Kerberos  response object

=back

=head1 AUTHOR/MAINTAINER

=over

=item NACL Development (dl-nacl-dev@netapp.com)

=back

=cut
