#
# Copyright (c) 2015,... NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Class to hold all connection information for MarsMaint Sessions
## @author nsesha@netapp.com, dl-nacl-dev@netapp.com
## @status Shared
## @pod here

package NACL::APISet::Session::CLI::MarsMaint;

use strict;
use warnings;

use base qw(NACL::APISet::Session::CLI);

use NACL::APISet::Response::CLI::MarsMaint;

use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Params::Validate qw(:all);
use Data::Dumper;
use Time::HiRes qw(time);

use NATE::Exceptions::Argument qw(:try);

sub new {
    $Log->enter() if $may_enter;
    $Log->debug( Dumper(@_) );

    # an instance of "this" is always the first arg
    # the second arg is a set of named arguments used to init
    # a hash.
    my ( $class, %args ) = @_;

    #
    # The conntype parameter should be specified ONLY if connid is also
    # provided
    #
    if ( ( defined $args{conntype} ) and ( not defined $args{connid} ) ) {
        NATE::Exceptions::Argument->throw(
            "Conntype provided but connid not provided");
    }

    # Fill the member vars for this instance
    #my $self = validate(@_, {  });
    my $self = NACL::APISet->validate_params(
        params => \%args,
        spec   => {
            username         => { type => SCALAR,         optional => 1 },
            password         => { type => SCALAR,         optional => 1 },
            conntype         => {
                type => SCALAR,
                optional => 1,
                default => "pm_cons",
            },
            ipaddress        => { type => SCALAR,         optional => 1 },
            connid           => {
                type     => SCALAR,
                optional => 1,
                default  => "console",
            },
            connection => { isa => "Connectrec", optional => 1 },
            hostobj    => { isa => "Hostrec",    optional => 1 },
        },
    );

    # This tags the hash pointed to by $self as a
    # type of MarsMaint
    bless $self, $class;

    # Pass the Maintenance prompt as an additional parameter to connect() in
    # Session/CLI.pm but only if the hostobj is defined (when user does not
    # pass in Connectrec object directly).

    my %additional_parameters;

    # Add prompts specific to MarsMaint mode
    if ( ref ($self->{hostobj}->{prompt_fw}) eq 'ARRAY'  &&
         ref($self->{hostobj}->{prompt_norm}) eq 'ARRAY' ) {
        %additional_parameters = ( prompts =>
                               [@{ $self->{hostobj}->{prompt_norm}} ,'(mars|marsdev)\ \-\ maintenance\(?(test|diag)?\)?> ']);
    }

    # Connect to the service
    $self->connect(%additional_parameters);

    # If the user has passed in a Connection object, then we have to manually
    # add the Maint mode prompt also to the list of prompts in the Connectrec object.
    # We do this by extracting it from the Hostrec object within the Connectrec
    # object. Adding a duplicate entry for this prompt is not an issue.
    # So we do this irrespective of how the Connectrec object was created.

    $self->{connection}->prompts($self->{connection}->prompts(), '(mars|marsdev)\ \-\ maintenance\(?(test|diag)?\)?> ');

    # And give the caller the referece to the bless object.
    $Log->exit() if $may_exit;
    return $self;
}
##############################################################################
# Method:    send_command
# Objective: Invoke the command on the MarsMaint system
# Details:   See POD documentation at the end of this file.
##############################################################################
sub send_command {
    $Log->enter() if $may_enter;

    my ( $self, %args ) = @_;
    my %connectrec;

    if ( $args{connectrec} ) {
         %connectrec = %{$args{connectrec}};
         $Log->debug("Arguments passed to Connectrec" . Dumper(\%connectrec));
    }
    $self->_get_scaled_timeout_value(timeout => \$connectrec{timeout});

    my ( $actual_command, $raw_execution_time );

    my $time_start = time();
    my $execution_status;
    my $output = $self->{connection}->execute(
        command => $args{command},
        %connectrec,
    );
    my $time_end       = time();
    my $execution_time = $time_end - $time_start;

    my $parser_path = "NACL::APISet::Response::CLI::MarsMaint";

    eval "use $parser_path";
    if ($@) {
        NATE::BaseException->throw("\nInternal Error in APISet Layer:$@");
    }

    $Log->exit() if $may_exit;

    return $parser_path->new(
        output             => $output,
        command            => $args{command},
        method_name        => $args{method_name},
        execution_time     => $execution_time,
        execution_status   => $execution_status,
        raw_execution_time => $raw_execution_time,
    );

}

1;

##############################################################################
#                                    POD here
##############################################################################

=head1 NAME

NACL::APISet::Session::CLI::MarsMaint - Session class for MarsMaint CLI connections

=head1 SYNOPSIS

 use NACL::APISet::Session::CLI::MarsMaint;

=head1 DESCRIPTION

A C<NACL::APISet::Session::CLI::MarsMaint> provides the methods needed to control the connection
to a MarsMaint system.

=head1 EXCEPTIONS

=over

=item NATE::Exceptions::Argument

=back

=head1 CONSTRUCTORS

=head2 new

Returns a reference to a newly created C<NACL::APISet::Session::CLI::MarsMaint>.

=over

=item Synopsis

 $session = NACL::APISet::Session::CLI::MarsMaint->new(name=>$name, %args);

=item Arguments

=over

=item name

The MarsMaint host for this session.

=item conntype

(Optional)
Default "pm_cons".  The Connectrec conntype to use for this session.

=item connid

(Optional)
Default "console".  The Connectrec connid to use for this session.

=item connection

(Optional)
An established NATE::Connectrec object to be use for this session.

=back

=item Returns

A reference to a new Session.

=back

=head1 METHODS

=head2 send_command

Send the command to the node.  The output of the command is returned in a NACL::APISet::Response::CLI::MarsMaint
object.

=over

=item Synopsis

$response = $session->send_command(command => "raw command string" );

=item Arguments

=over

=item command

The text of the command to send.

=item Connectrec options

Options can be passed to Connectrec->execute. To

=back

=back

=head1 AUTHOR/MAINTAINER

=over

=item NACL Development (dl-nacl-dev@netapp.com)

=back

=cut
