#
# Copyright (c) 2015,... NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Class to hold all connection information for MarsNormal Sessions
## @author nsesha@netapp.com, dl-nacl-dev@netapp.com
## @status Shared
## @pod here

package NACL::APISet::Session::CLI::MarsNormal;

use strict;
use warnings;

use base qw(NACL::APISet::Session::CLI);

use NACL::APISet::Response::CLI::MarsNormal;

use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Params::Validate qw(:all);
use Data::Dumper;
use Time::HiRes qw(time);

use NATE::Exceptions::Argument qw(:try);

sub new {
    $Log->enter() if $may_enter;
    $Log->debug( Dumper(@_) );

    # an instance of "this" is always the first arg
    # the second arg is a set of named arguments used to init
    # a hash.
    my ( $class, %args ) = @_;

    #
    # The conntype parameter should be specified ONLY if connid is also
    # provided
    #
    if ( ( defined $args{conntype} ) and ( not defined $args{connid} ) ) {
        NATE::Exceptions::Argument->throw(
            "Conntype provided but connid not provided");
    }

    # Fill the member vars for this instance
    #my $self = validate(@_, {  });
    my $self = NACL::APISet->validate_params(
        params => \%args,
        spec   => {
            username         => { type => SCALAR,         optional => 1 , default => "admin" },
            password         => { type => SCALAR,         optional => 1 , password => "changeme" },
            conntype         => { type => SCALAR,         optional => 1 },
            ipaddress        => { type => SCALAR,         optional => 1 },
            connid           => {
                type     => SCALAR,
                optional => 1,
                default  => "libssh-persay",
            },
            connection => { isa => "Connectrec", optional => 1 },
            hostobj    => { isa => "Hostrec",    optional => 1 },
        },
    );

    # Add more variables that are sub-class specific.

    # This tags the hash pointed to by $self as a
    # type of MarsNormal
    bless $self, $class;

    # Connect to the service
    $self->connect();

    # And give the caller the referece to the bless object.
    $Log->exit() if $may_exit;
    return $self;
}
##############################################################################
# Method:    send_command
# Objective: Invoke the command on the MarsNormal system
# Details:   See POD documentation at the end of this file.
##############################################################################
sub send_command {
    $Log->enter() if $may_enter;

    my ( $self, %args ) = @_;
    my %connectrec;

    if ( $args{connectrec} ) {
         %connectrec = %{$args{connectrec}};
         $Log->debug("Arguments passed to Connectrec" . Dumper(\%connectrec));
    }
    $self->_get_scaled_timeout_value(timeout => \$connectrec{timeout});

    my ( $actual_command, $raw_execution_time );

    my $time_start = time();
    my $execution_status;
    my $output = $self->{connection}->execute(
        command => $args{command},
        %connectrec,
    );
    my $time_end       = time();
    my $execution_time = $time_end - $time_start;

    my $parser_path = "NACL::APISet::Response::CLI::MarsNormal";

    eval "use $parser_path";
    if ($@) {
        NATE::BaseException->throw("\nInternal Error in APISet Layer:$@");
    }

    $Log->exit() if $may_exit;

    return $parser_path->new(
        output             => $output,
        command            => $args{command},
        method_name        => $args{method_name},
        execution_time     => $execution_time,
        execution_status   => $execution_status,
        raw_execution_time => $raw_execution_time,
    );

}

1;

##############################################################################
#                                    POD here
##############################################################################

=head1 NAME

NACL::APISet::Session::CLI::MarsNormal - Session class for MarsNormal CLI connections

=head1 SYNOPSIS

 use NACL::APISet::Session::CLI::MarsNormal;

=head1 DESCRIPTION

A C<NACL::APISet::Session::CLI::MarsNormal> provides the methods needed to control the connection
to a MarsNormal system.

=head1 EXCEPTIONS

=over

=item NATE::Exceptions::Argument

=back

=head1 CONSTRUCTORS

=head2 new

Returns a reference to a newly created C<NACL::APISet::Session::CLI::MarsNormal>.

=over

=item Synopsis

 $session = NACL::APISet::Session::CLI::MarsNormal->new(name=>$name, %args);

=item Arguments

=over

=item name

The MarsNormal host for this session.

=item username

(Optional)
Default "admin".  The Mars shell user.

=item password

(Optional)
Default "changeme".  The Mars shell user's password.

=item conntype

(Optional)
Default "libssh-persay".  The Connectrec conntype to use for this session.

=item connid

(Optional)
Default "libssh-persay".  The Connectrec connid to use for this session.

=item connection

(Optional)
An established NATE::Connectrec object to be use for this session.

=back

=item Returns

A reference to a new Session.

=back

=head1 METHODS

=head2 send_command

Send the command to the node.  The output of the command is returned in a NACL::APISet::Response::CLI::MarsNormal
object.

=over

=item Synopsis

$response = $session->send_command(command => "raw command string" );

=item Arguments

=over

=item command

The text of the command to send.

=item Connectrec options

Options can be passed to Connectrec->execute. To

=back

=back

=head1 AUTHOR/MAINTAINER

=over

=item NACL Development (dl-nacl-dev@netapp.com)

=back

=cut
