#
# Copyright (c) 2001-2011 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Class to hold all connection information for SP Sessions
## @author kabhishe@netapp.com, dl-nacl-dev@netapp.com
## @status Shared
## @pod here

# Not all 'connections' to services will be in the form of NATE::Connectionrec
# The Session class will define all data needed for connections, whether
# NATE Connection or not.

# For SP sessions, you can override the answer for the prompt
# "Would you like to disconnect that session, and start yours [y/n]?" using
# <NATE_FORCE_SP_LOGIN> parameter
# Value for NATE_FORCE_SP_LOGIN can be y/n

package NACL::APISet::Session::CLI::SP;

#
# Compiler Directives
#
use strict;
use warnings;

use base qw(NACL::APISet::Session::CLI);

# Module import - APISets
use NACL::APISet::Response::CLI::SP;

#
# NATE imports
#
use NATE::Log qw(log_global);
use NATE::ParamSet qw(param_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();

use Params::Validate qw(:all);
use Data::Dumper;
use Time::HiRes qw(time); # This module is for measuring time in high precision

#
# Exceptions imported
#
use NATE::Exceptions::Argument qw(:try);
use NACL::APISet::Exceptions::ConnectionFailedException qw(:try);

sub new {
    $Log->enter() if $may_enter;

    # an instance of "this" is always the first arg
    # the second arg is a set of named arguments used to init
    # a hash.
    my ($class, %args) = @_;
    #
    # The conntype parameter should be specified ONLY if connid is also
    # provided
    #
    if ((defined $args{conntype}) and (not defined $args{connid})) {
        NATE::Exceptions::Argument->throw(
        "Conntype provided but connid not provided");
    }

    # Fill the member vars for this instance

    my $self = NACL::APISet->validate_params(
        params => \%args,
        spec   => {
            hostobj   => { isa => "Hostrec", optional => 1 },
            username => { type => SCALAR, optional => 1, default => "admin"},
            password => { type => SCALAR, optional => 1, default => "netapp1!"},
            conntype => { type => SCALAR, optional => 1 },
            ipaddress => { type => SCALAR, optional => 1 },
            connid   => { type => SCALAR, optional => 1, default => "sp_login" },
            connection => { isa => "Connectrec", optional => 1 },
            login_to_console => {type => SCALAR, default => 0},
        },
    );

    # Add more variables that are sub-class specific.

    # This tags the hash pointed to by $object as a
    # type of SP
    bless $self, $class;

    # boot whoever is on the SP console if we plan on logging in
    # to the console
    if ($self->{login_to_console}) {
        param_global()->set(key => "NATE_FORCE_SP_LOGIN",
                            value => "y");
    }

    # Connect to the service
    $self->connect();

    # Prepare the service for use
    $self->prep_service();

    # And give the caller the referece to the bless object.
    $Log->exit() if $may_exit;
    return $self;
}

###############################################################################
# Method:    prep_service
# Objective: Prep the SP session (login to the console if needed)
# Details:   See POD documentation at the end of this file.
###############################################################################
sub prep_service {
    $Log->enter() if $may_enter;
    my ($self) = @_;

    $Log->comment("Prepping Service!");

    # log in to the console from the service processor if requested
    if ($self->{login_to_console}) {
        my $c = $self->{connection};
        $c->execute(
            command => "system console",
            "additional_prompts" => ["Type Ctrl-D to exit"],
            "match_table" => ["and start yours [y/n]\?" => "y"]
        );
    
        $Log->comment("Waiting for login");
    
        $c->execute(command => "", max_idle => 5, on_idle_wakeup => sub {""}, "match_table" => ["SP-login:" => $self->{username},"Password:" => $self->{password}]);
        $c->discardbuff();
        sleep 1;
    }

    $Log->exit() if $may_exit;
}

###############################################################################
# Method:    send_command
# Objective: Execute the command on a SP filer and gather the raw output
# Details:   See POD documentation at the end of this file.
###############################################################################
sub send_command {
    $Log->enter() if $may_enter;

    my ($self, %opts) = @_;
    my %args = validate_with(
        params      => \%opts,
        spec        => { connectrec => { type => HASHREF, default => {} } },
        allow_extra => 1
    );

    my %connectrec = %{delete $args{connectrec}};
    $self->_get_scaled_timeout_value(timeout => \$connectrec{timeout});
    my $parser_file = $args{parser_file};
    $Log->debug( sub { "Session/CLI/SP:send_command: " . Dumper( \%args ) } );

    unless ($args{method_name}) {
        # Caller is execute_command, all the options passed
        # should be considered as connectrec options
        %connectrec = ( %connectrec, %args );
        $args{command} ||= $connectrec{command};
        delete $connectrec{command}
    }

    $Log->debug( sub { "Arguments that are being passed to execute(): ". "command = '".
                $args{command}. "'\n". Dumper(\%connectrec) } );

    my $time_start = time();
    my $output = $self->{connection}->execute(command => $args{command},
                            %connectrec);
    my $time_end = time();

    # Find total Execution time
    my $execution_time = $time_end - $time_start;

    # for execute_raw_command and execute_command just return the raw output
    if(!$args{method_name}) {
        $Log->exit() if $may_exit;
        return $output;
    }

    my $response_object =  NACL::APISet::Response::CLI::SP->new(
            output            => $output,
            actual_command    => $args{actual_command},
            command           => $args{command},
            cdef_path         => $args{cdef_path},
            method_name       => $args{method_name},
            execution_time    => $execution_time,
            parser_file       => $parser_file
        );

    $Log->exit() if $may_exit;
    return $response_object;
}

1;

##############################################################################
#                                    POD here
##############################################################################

=head1 NAME

NACL::APISet::Session::CLI::SP - SP CLI Session class

=head1 SYNOPSIS

 use NACL::APISet::Session::CLI::SP;

=head1 DESCRIPTION

A C<NACL::APISet::Session::CLI::SP> object is used to establish a connection
and send commands for the SP CLI APISet.

=head1 EXCEPTIONS

=over

=item NATE::Exceptions::Argument

=item NACL::APISet::Exceptions::ConnectionFailedException

=back

=head1 CONSTRUCTORS

=head2 new

Returns a reference to a newly created C<NACL::APISet::Session::CLI::SP>.

=over

=item Synopsis

 $self->{session} = NACL::APISet::Session::CLI::SP->new( %args );

=item Arguments

=over

=item hostobj

The hostrec object of the resource

=item username

Username with which to log into the resource

=item password

Password with which to log into the resource

=item connid

The connection ID

=item conntype

Connection type with which to connect. We don't recommend
overriding default conntype until its necessary. sp_login template has libssh-persist as default conntype


=item connection

connectrec object for an existing connection to the resource

=back

=item Returns

A reference to a new SP CLI Session class.

=back

=head1 METHODS

=head2 send_command

Send the command to the node.  The output of the command is returned in a NACL::APISet::Response::CLI::SP
object.

=over

=item Synopsis

$response = $session->send_command( command => "raw command string" );

=item Arguments

=over

=item command

The text of the command to send.

=back

=back

=head1 AUTHOR/MAINTAINER

=over

=item NACL Development (dl-nacl-dev@netapp.com)

=back

=cut
