## @summary Class to hold all selenium GUI connection information.
## @author madhav.shekar@netapp.com, sindhe@netapp.com, dl-nacl-dev@netapp.com
## @status Shared
## @pod here

package NACL::APISet::Session::GUI::SysManager;

#
# Compiler Directives
#
use strict;
use warnings;

#
# NATE imports
#
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Params::Validate qw(:all);
use NATE::BaseException qw(:try);
use NATE::Exceptions::Argument qw(:try);

use Selenium::Remote::Driver;

##############################################################################
# Method:    get_timeout
# Objective: Retrieve the current timeout value
# Details:   See POD documentation at the end of this file.
##############################################################################
sub get_timeout {
    my $self = shift;
    $Log->enter() if $may_enter;
    $Log->exit() if $may_exit;

    return $self->{timeout};
}

##############################################################################
# Method:    cleanup
# Objective: close the conection of GUI object
# Details:   See POD documentation at the end of this file.
##############################################################################
sub cleanup {
    my $self = shift;
    $Log->enter() if $may_enter;

    $self->stop();
    $self->DESTROY();

    $Log->exit() if $may_exit;
    return 1;
} ## end sub cleanup

###############################################################################
# Method:    set_timeout
# Objective: Set the timeout to a new value
# Details:   See POD documentation at the end of this file.
###############################################################################
sub set_timeout {
    my $self = shift;
    $Log->enter() if $may_enter;
    my %args = validate_with(
                        params => \@_,
                        spec   => {
                            'method-timeout' => {type => SCALAR, optional => 1},
                            timeout          => {type => SCALAR, optional => 1},
                                });
    my $timeout;
    if (defined $args{'method-timeout'}) {
        $timeout = $args{'method-timeout'};
    } elsif (defined $args{timeout}) {
        $timeout = $args{timeout};
    } else {
        NATE::Exceptions::Argument->throw('Either of the parameters '
                         . "'method-timeout' or 'timeout' should have been "
                         . 'passed but neither were');
    }
    $self->{timeout} = $timeout;

    # Setting the GUI sellinium timeout
    my $res = {'command' => 'setTimeout'};

    foreach my $type ('script','page load') {
        my $params = {'type' => $type, 'ms' => 30000};
        eval { $self->_execute_command( $res, $params); };
    }

    $Log->debug('Timeout set to ' . $self->{timeout});
    $Log->exit() if $may_exit;
} ## end sub set_timeout

1;

##############################################################################
#                                    POD here
##############################################################################

=head1 NAME

NACL::APISet::Session::GUI::SysManager - APISet Session base class. 

=head1 SYNOPSIS

 use NACL::APISet::Session::GUI::SysManager;

=head1 DESCRIPTION

Contains Common routines for libraries under NACL/APISet/Session/GUI/SysManager folder.

=head2 set_timeout

This method is used to set the timeout to a new value.

=over

=item Synopsis

    $self->{session}->set_timeout('method-timeout' => $new_timeout);

                                 or

    $self->{session}->set_timeout(timeout => $new_timeout);

=item Arguments

Either of timeout or method-timeout should be provided. If both are
provided, then method-timeout takes precedence.

=over

=item method-timeout

The new timeout value (in seconds)

=item timeout

The new timeout value (in seconds)

=back

=item Returns

Nothing.

=back

=head2 cleanup

This method is used to clean up the Session after use. This involves 
deleting the connection object stored within the Session object.

=over

=item Synopsis

$self->{session}->cleanup();

=item Arguments

None.

=item Returns

Nothing.

=back

=head2 get_timeout

This method is used to get the timeout of the GUI session object

=over

=item Synopsis

    $self->{session}->get_timeout();

=over

=item method-timeout

The new timeout value (in seconds)

=item Returns

Session timeout value

=back

=head1 AUTHOR/MAINTAINER

=over

=item NACL Development madhav.shekar@netapp.com, sindhe@netapp.com, dl-nacl-dev@netapp.com

=back

=cut
