#
# Copyright (c) 2001-2010 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Base Session class for SNMP APISet
## @author Isshwarya.M@netapp.com, dl-nacl-dev@netapp.com
## @status Shared
## @pod here

package NACL::APISet::Session::SNMP;

use strict;
use warnings;

use base  qw(NACL::APISet::Session);

# Importing Exceptions
use NACL::APISet::Exceptions::ConnectionFailedException qw(:try);
use NATE::Exceptions::Argument qw(:try);
use NACL::APISet::Exceptions::TimeoutException qw(:try);
use NACL::APISet::Exceptions::InvalidInputRequest qw(:try);

#
# NATE imports
#
use Params::Validate qw(:all);
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Data::Dumper;
use Net::SNMP qw(:asn1 :snmp);
use Net::IP qw(:PROC);
use Tie::IxHash;
use Time::HiRes qw(time); # This module is for measuring time in high precision

# global variables
use constant SNMP_VERSION => "2c";
use constant SNMP_COMMUNITY => "public";
use constant SNMP_SECURITY_LEVEL => 'authNoPriv';
use constant SNMP_AUTH_PROTOCOL => "md5";
use constant SNMP_PRIV_PROTOCOL => "des";

# Default log level SNMP request is 'comment'.
use constant DEFAULT_LOG_LEVEL => 'comment';

# Default log level for an SNMP request on error is 'comment'.
use constant DEFAULT_LOG_LEVEL_ON_ERROR => 'comment';

# Constructor
sub new {
    $Log->enter();
    $Log->debug( Dumper(@_) );

    # an instance of "this" is always the first arg
    # the second arg is a set of named arguments used to init
    # a hash.
    my $class = shift;

    # Fill the member vars for this instance
    my $self = NACL::APISet->validate_params(
        params => \@_,
        spec   => {
            hostobj   => { isa => "Hostrec", optional => 1 },
            connection => { isa => "Connectrec", optional => 1 },
            ipaddress => { type => SCALAR, optional => 1 },            
            snmp_auth  => { type => HASHREF, optional => 1 },
        },
    );
    
    # This tags the hash pointed to by $object as a
    # type of 7Mode
    bless $self, $class;
    
    # Initialize the cdef_entry_in_response option
    $self->{cdef_entry_in_response} = 0;
    
    # Obtain the SNMP related inputs to create the sesssion. This will look
    # into the constructor overrides or tharnhost else provides default values.
    $self->_get_session_inputs();
    
    #Set test parameters
    $self->set_test_parameters();
    
    # connect to the service
    $self->connect();   

    # And give the caller the referece to the bless object.
    $Log->exit();
    return $self;
}

###############################################################################
# Method:    connect
# Objective: Connect to the service
# Details:   See POD documentation at the end of this file.
###############################################################################

sub connect {
    $Log->enter();
    my ($self, %args) = @_;
    
    # Get the relevant SNMP constructor inputs based on version
    # and security level.
    my %snmp_auth = %{$self->{session}};
    $Log->debug("SNMP auth inputs read from various sources: ". 
        Dumper \%snmp_auth);
    my @inputs;
    if ($snmp_auth{version} ne '3') {
        # For v1 and v2c, pass 'community' string
        @inputs = qw(hostname community version domain);
    } else {
        # For v3, pass username along with other authentication and privacy
        # related inputs based on the security level.
        @inputs = qw(hostname username version domain);
        if ($snmp_auth{security_level} eq 'authnopriv') {
            push(@inputs, 'authprotocol', 'authkey', 'authpassword');
        } elsif ($snmp_auth{security_level} eq 'authpriv') {
            push(@inputs, 'authprotocol', 'authkey', 'authpassword');
            push(@inputs, 'privprotocol', 'privkey', 'privpassword');
        }
    }
    my $opts = $self->_populate_hash(@inputs);
    $Log->debug("SNMP session inputs are \n". Dumper($opts));
    my ($session, $error) = Net::SNMP->session(%$opts);
    unless (defined $session) {
        $Log->exit();
        NACL::APISet::Exceptions::ConnectionFailedException->throw(
            "Could not create SNMP session and the error is:". "\n$error");
    }
    # set the translation.
    $Log->debug("Translation status immediately after object creation: " . 
        $session->translate());
    my $new_translation = $session->translate([
        # To translate non printable characters in Octet string to Hex value.
        # Disabled now and translation routines will be added if someone burts.
        '-octetstring' => 0,
        # For NULL values, SNMP returns 0 bytes as contents and so the raw
        # value would be empty string and processed value would be 'NULL'.
        # We would later identify NELL values from ASN.1 varbind type info.
        # So turning off the translation.       
        '-null' => 0,
        # Raw value will be processed later using ticks_to_time() routine later
        # to get the readable format.
        '-timeticks' => 0,
        # There are no objects of type 'opaque' datatype in netapp mibs. So
        # turning off this translation for now.
        '-opaque' => 0,
        # Turn off the translation for wrongly encode unsigned values as signed
        # values so that QA can fix those product bugs.
        '-unsigned' => 0,
        # For all these three, we will turn off the translation so that we get
        # empty string as raw value (as the length field in TLV format followed
        # in BER encoding will say 0 bytes) and 'noSuchObject' , 
        # 'noSuchInstance', 'endOfMiBView' as processed values.
        '-nosuchinstance' => 0,
        '-nosuchobject' => 0,
        '-endofmibview' => 0,       
        ]);
    unless (defined $new_translation) {
        NATE::BaseException->throw("Failed to set translation: ". 
            $session->error());
    }
    $Log->debug("Translation status after setting the desired translation: ".
        $new_translation);
    $self->{connection} = $session;
    $Log->exit();
}

###############################################################################
# Method:    set_timeout
# Objective: Set the timeout to a new value
# Details:   See POD documentation at the end of this file.
###############################################################################
sub set_timeout {
    my $self = shift;
    $Log->enter();
    my %args = validate_with(
        params => \@_,
        spec   => {
            timeout  => { type => SCALAR, optional => 0 },
        }
    );
    my $outp = $self->{connection}->timeout($args{timeout});
    unless (defined $outp) {
        NATE::BaseException->throw("Failed to set timeout and the".
            "reason is: ". $self->{connection}->error());
    }
    $Log->debug( 'Timeout set to ' . $self->{connection}->timeout() );
    $Log->exit();
}

###############################################################################
# Method:    get_timeout
# Objective: Get the current timeout
# Details:   See POD documentation at the end of this file.
###############################################################################

sub get_timeout {
    my $self = shift;
    $Log->enter();
    
    my $timeout = $self->{connection}->timeout() ;
    $Log->exit();
    return $timeout;
}

###############################################################################
# Method:    clear_timeout
# Objective: Clear the timeout to its defaults
# Details:   See POD documentation at the end of this file.
###############################################################################

sub clear_timeout {
    my $self = shift;
    $Log->enter();
    
    $self->set_timeout(timeout => &NACL::APISet::Node::SNMP::TIMEOUT());
    $Log->exit();
}

###############################################################################
# Method:    execute_snmp_method
# Objective: To execute SNMP PDUS and snmpwalk
# Details:   See POD documentation at the end of this file.
###############################################################################

sub execute_snmp_method {
    my ($self, %args) = @_;
    $Log->enter();
    my $output;
    
    if ($args{method} eq 'snmp_walk') {
        $output = $self->_snmp_walk(%args);
    } else {
        $output = $self->_executeRequest(%args);
    }
    $Log->exit();
    return $output;
}

###############################################################################
# Method:    cleanup
# Objective: Cleanup the Session object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub cleanup {
    $Log->enter();
    my ($self, %args) = @_;

    # Close the Net::SNMP session
    $self->{connection}->close;
    
    delete $self->{connection} ;
    $Log->exit();
}

###############################################################################
# Method:    execute_raw_command
# Objective: execute the raw SNMP command and return back either raw output or
#            response object depending on what user had asked for.
# Details:   See POD documentation at the end of this file.
###############################################################################

sub execute_raw_command {
    $Log->enter();
    my ($self, %args) = @_;
    
    # Get the method name and pass all other options to Net::SNMP method call
    my $method = $args{method};
    my $return_response_object = (defined $args{'return-response-object'})?
        delete $args{'return-response-object'} : 0;
    
    # Support all the basic PDUs + snmp_walk
    if ($method !~ /^(set|get|get_next|get_bulk|snmp_walk)$/) {
        NATE::Exceptions::Argument->throw("Invalid method specified: $method. ".
            "Supported PDUs are set, get, get_next or get_bulk.")
    }
    
    my $result;
    if ($return_response_object) {
        $result = $self->send_command(%args);
    } else {
        my $output = $self->execute_snmp_method(%args);
        $result = $output->{output};
    }
    
    $Log->exit();
    return $result;
}

##############################################################################
#                               Private Methods
##############################################################################

###############################################################################
# Method:    _executeRequest
# Objective: Executes the specified SNMP PDU using Net::SNMP connection object
# Details:   Returns a hash reference as output having information about
#            raw output, execution time, execution status, error index, error
#            reason, asn1_type information, command executed which can be
#            used for creating response object.
###############################################################################

sub _executeRequest {
    $Log->enter();

    my $self = shift();
    my %args = NACL::APISet::validate_params(
        $self,
        params => \@_,
        spec   => { "varbind" => { type => HASHREF|ARRAYREF, optional => 0 },
                    "method"  => { type => SCALAR, optional => 0 },
                    "schema"  => { isa => 'NACL::APISet::Schema::SnmpSchema', 
                        optional => 1 },
                    "maxrepetitions" => { type => SCALAR|UNDEF, optional => 1 },
                    "nonrepeaters" => { type => SCALAR|UNDEF, optional => 1 },
                    "log_level"        => { type => SCALAR | UNDEF, optional => 1 },
                    "input_log_level"  => { type => SCALAR | UNDEF, optional => 1 },
                    "output_log_level"  => { type => SCALAR | UNDEF, optional => 1 },
                  }
    );
    my (%output, $netsnmp, $result, $varbind, %input); 
    $varbind = delete $args{varbind}; 
    my $pdu = delete($args{method});
    my $method = $pdu .'_request';  
    if ($method eq 'get_bulk_request') {    
        $input{maxrepetitions} = delete $args{maxrepetitions} 
            if (defined $args{maxrepetitions});
        $input{nonrepeaters} = delete $args{nonrepeaters}
            if (defined $args{nonrepeaters});
    }
    my $schema = delete $args{schema};
    
    # Throw exception if an invalid parameter is passed.
    if (scalar(keys %args)) {
        $Log->exit();
        NATE::Exceptions::Argument->throw("Invalid Argument(s) used: " .
         join(', ', keys(%args)));
    }
    
    $netsnmp = $self->{connection};
    
    # Get the immediate caller.
    my @arr = caller(2);
    my $is_execute_raw = ($arr[3] =~ /execute_raw_command$/)? 1 : 0;
    
    if (ref($varbind) eq 'ARRAY') {
        # For execute_raw_command
        $input{varbindlist} = $varbind;
    } elsif ($method eq 'set_request') {
        # For Set PDU, construct varbindlist as trios of numeric OID,
        # data type value and the value to be set.
        foreach my $object (keys %$varbind) {
            my $type = $self->_translate_datatypes_to_asn(
                type => $varbind->{$object}->{type});
            push(@{$input{varbindlist}}, $varbind->{$object}->{oid},
                $type, $varbind->{$object}->{value});
        }
    } else {
        # Covers Get, GetNext and GetBulk PDU method calls.     
        $input{varbindlist} = [values %{$varbind}];
    }
    
    my $inputstr = "Input is:\n" . 
        Dumper({%input, command => $pdu});

    my $start_time = time();
    $result = $netsnmp->$method(%input);
    my $end_time = time();              
    if ( ref $result eq 'HASH' ) {
        require Encode;
        while (my ($key, $value) = each %$result) {
            if (!ref $value ) {
                $result->{$key} = Encode::decode_utf8($value);
            }
        }
    }
    $output{output} = $result;
    $output{execution_time} = $end_time - $start_time;  
    $output{command} = {%input, command => $pdu};   
    $self->_post_execution( output => \%output, 
                            input => $inputstr,
                            log_level => $args{log_level},
                            input_log_level => $args{input_log_level},
                            output_log_level => $args{output_log_level},
                        );
    $Log->debug("Output returned by _executeRequest\n". Dumper(\%output));
    # schema object is added after debug statement to avoid dumping the
    # schema object for every method call in debug logs.
    $output{schema} = $schema;
    
    $Log->exit();
    return \%output;
}

###############################################################################
# Method:    _post_execution
# Objective: Performs the post processing after executing any SNMP PDU or
#            method using connection object.
# Details:   Populates the hash reference obtained as input with post 
#            processing details such as error index, execution status, error
#            reason, asn1_type and sorts the OIDs in the output in lexical
#            order. If the SNMP request had failed because of hitting timeout
#            or invalid PDU (in which cases there is no response PDU obtained),
#            then throws appropriate exception.
###############################################################################

sub _post_execution {
    my ($self, %args) = @_;
    $Log->enter();
    
    # Pointer to output hash to be filled
    my $output    = $args{output};
    my $inputstr  = $args{input};
    
    my  (   $input_log_level_on_success,
            $output_log_level_on_success,
            $input_log_level_on_error,   
            $output_log_level_on_error,  
            $outputstr  );
	
    $input_log_level_on_success
        = $args{input_log_level}
        || $args{log_level}
	|| $self->{snmp_log_level}
        || DEFAULT_LOG_LEVEL;
    $output_log_level_on_success
        = $args{output_log_level}
        || $args{log_level}
	|| $self->{snmp_log_level}
        || DEFAULT_LOG_LEVEL;
    $input_log_level_on_error
        = $args{input_log_level}
        || $args{log_level}
	|| $self->{snmp_log_level_on_error}
        || $self->{snmp_log_level}
	|| DEFAULT_LOG_LEVEL_ON_ERROR;
    $output_log_level_on_error
        = $args{output_log_level}
        || $args{log_level}
	|| $self->{snmp_log_level_on_error}
	|| $self->{snmp_log_level}
        || DEFAULT_LOG_LEVEL_ON_ERROR;
    
    my $netsnmp = $self->{connection};
    
    $output->{asn1_type} = $netsnmp->var_bind_types;
    $output->{error_index} = $netsnmp->error_index;
    $output->{execution_status} = $netsnmp->error_status;
    $output->{error_reason} = $netsnmp->error;
    
    if(defined $output->{output}) {
        # Sort the OIDs in the raw output
        my %sorted_output;
        tie %sorted_output, 'Tie::IxHash';
        my @sorted_oids = oid_lex_sort(keys %{$output->{output}});
        map {$sorted_output{$_} = $output->{output}->{$_}} @sorted_oids;
        $output->{output} = \%sorted_output;    
        
	    $self->_log_data(
            data      => $inputstr,
            log_level => $input_log_level_on_success
        );

        $outputstr .= "Raw output is:\n". Dumper($output->{output});
        $self->_log_data(
            data      => $outputstr,
            log_level => $output_log_level_on_success
        );
        
    } elsif ((defined $output->{error_reason}) && 
        (length $output->{error_reason})) {
        # Log the error message
        $outputstr .= "Raw output is undef and the error reason is:" . 
            $output->{error_reason};
        
        $self->_log_data(
            data      => $inputstr,
            log_level => $input_log_level_on_error
        );
        
        $self->_log_data(
            data      => $outputstr,
            log_level => $output_log_level_on_error
        );
    }   
    
    # Throw exceptions for cases where a response PDU is not obtained.
    if ($output->{error_reason} =~ /No response from remote host .+/) {
        # Timeout is hit.
        NACL::APISet::Exceptions::TimeoutException->throw(
            "Connection timed out");
    } elsif ($output->{error_reason} =~ /PDU not supported.+/) {
        # Invalid PDU
        # When a invalid PDU like get_bulk is specified for v1, then request
        # itself won't be sent in which case no response PDU is obtained. It
        # just populates error_reason field and sets 0 or '' to other
        # error related fields.
        NACL::APISet::Exceptions::InvalidInputRequest->throw(
            $output->{error_reason});
    } elsif ($output->{error_reason} =~ 
        /Non-repeaters greater than the number of variable-bindings/) {
        # Request PDU itself is not sent when specifying a nonrepeaters greater
        # than the number of objects in variable bindings. Hence a response
        # exception cannot be thrown.
        NACL::APISet::Exceptions::InvalidInputRequest->throw(
            $output->{error_reason});
    }
    $Log->exit();
}

###############################################################################
# Method:    _snmp_walk
# Objective: Implement snmp walk operation
# Details:   Tries to perform snmp walk using get_bulk PDU for SNMPv2c and v3.
#            If it fails for some reason, then it will fall back to get_next
#            PDU. For SNMPv1, it will always use get_next as get_bulk is not
#            supported in SNNPv1. 
###############################################################################

sub _snmp_walk {    
    my ($self, %args) = @_;
    $Log->enter();
    
    my $netsnmp = $self->{connection};
    my $baseoid;
            
    if (ref($args{baseoid}) eq 'HASH') {
        # method call
        my $varbind = delete $args{baseoid};
        my @keys = keys %$varbind;
        $baseoid = $varbind->{$keys[0]};
    } else {
        # execute_raw_command
        $baseoid = delete $args{baseoid};
    }
    my $comm = delete $args {method};
    my $schema = delete $args {schema};
    # Throw exception if an invalid parameter is passed.
    if (scalar(keys %args)) {
        $Log->exit();
        NATE::Exceptions::Argument->throw("Invalid Argument(s) used: " .
         join(', ', keys(%args)));
    }
    my ($start_time, $end_time, $result, $asn1_type, $inputstr);
    
    my $command = {command => $comm, baseoid => $baseoid};
    $inputstr = "Input is:\n". Dumper($command);
    $start_time = time();
	
    # $result = $netsnmp->get_table(baseoid => $baseoid);
	# The above line of code was commented due to a issue with Net::SNMP's
	# GetBULK not getting information properly, the information that considered
	# was related to lunTable.  Refer BURT #614626 for more details.  This code
	# will be reused when we find a permanent solution in the form of a different
	# version of Net::SNMP  
	
    my %additional_option;
    if ($self->{session}->{version} !~ /1/) {
        $additional_option{maxrepetitions} = 1;	
    }
    $result = $netsnmp->get_table(baseoid => $baseoid, %additional_option);
    $end_time = time();
    # For v2c and v3, if we hit maxMsgSize error, then fall back to 
    # get_next_request approach because Net::SNMP uses scaling factor of 0.017
    # (which is obtained by dividing 1472(default max msg size for UDP/IPv4) by
    # 25(hardcoded max-repetitions)) and it may
    # fail for some cases with "maxMsgSize" error. We couldn't select the
    # max-repetitions value dynamcially eventhough if we know about the 
    # datatype of each column object in the table. This is because we don't
    # know how many instances (or rows) exist in the table and the values are
    # read columnar fashion rather than row by row.
    # For eg: aggrName.0, aggrName.1, aggrStatus.0, aggrStatus.1 rather than
    # aggrName.0, aggrStatus.0, aggrName.1, aggrStatus.1
    $Log->debug("Error in invoking get_table: ". $netsnmp->error()) 
        unless(defined $result);
    
	# if ((!defined $result) && ($self->{session}->{version} ne '1')) {
        # $Log->debug("GetBulk returned the error: ". $netsnmp->error());
        # $Log->debug("Fall back to GetNext");
        # $start_time = time();
        # $result = $netsnmp->get_table(baseoid => $baseoid, maxrepetitions =>1);
        # $end_time = time();
    # }
    # The above code is commented for the same reason as mentioned in line #474
    if ( ref $result eq 'HASH' ) {
        require Encode;
        while (my ($key, $value) = each %$result) {
            if (!ref $value){
                $result->{$key} = Encode::decode_utf8($value);
            }
        }
    }
		
    my $output = {output =>$result, execution_time => ($end_time - $start_time),
        command => $command}; 
    $self->_post_execution( output => $output, 
                            input => $inputstr,
                            log_level => $args{log_level},
                            input_log_level => $args{input_log_level},
                            output_log_level => $args{output_log_level},
                        );
    $Log->debug("Output returned by _snmp_walk\n". Dumper($output));
    # schema object is added after debug statement to avoid dumping the
    # schema object for every method call in debug logs.
    $output->{schema} = $schema;
    
    $Log->exit();
    return $output;   
}

##############################################################################
# Method:    _populate_hash
# Objective: Populate hash with specific arguments
# Details:   Provided as input is the list of keys which we'd like to store
#            in another hash. If any of these are defined in $self->{session},
#            we copy them into the hash and return this hash reference.
##############################################################################

sub _populate_hash {
    my ( $self, @keys ) = @_;
    $Log->enter();
    my %hash;

    foreach my $key (@keys) {
        if ( defined( $self->{session}->{$key} ) ) {
            $hash{$key} = $self->{session}->{$key};
        }
    }
    $Log->exit();
    return \%hash;
}

##############################################################################
# Method:    _get_session_inputs
# Objective: consolidate the inputs required for creating SNMP session from
#            various sources.
# Details:   Session inputs such as version, community, IP address, username,
#            security level, authprotocol, authpassword, authkey, privprotocol,
#            privkey and privpassword are looked up in various sources like
#            APISet constructor, tharnhost in the order of preference. If
#            it couldn't find the information in both these sources, then
#            assign the defaults defined for them.
##############################################################################

sub _get_session_inputs {
    $Log->enter();
    my $self = shift();
    
    my %snmp_auth = defined ($self->{snmp_auth})? %{$self->{snmp_auth}}: ();
    my $host = ($self->{hostobj})? 
        $self->{hostobj}: $self->{connection}->hostp();
    
    # Version
    $self->{session}->{version} = $snmp_auth{version} || SNMP_VERSION;
    if ($self->{session}->{version} !~ /^(1|2c|2|3)$/i) {
        $Log->exit();
        NATE::Exceptions::Argument->throw("Invalid SNMP version '". 
            $self->{session}->{version}. "'. Valid values are 1, 2(or 2c), 3");
    }
    
    # IP address
    unless ((defined $self->{ipaddress}) && ($self->{ipaddress})) {
        $self->{ipaddress} = ($self->{hostobj})?
            $self->{hostobj}->default_ip(): $self->{connection}->connhost();
    }
    $self->{session}->{hostname} = $self->{ipaddress};

    if (ip_is_ipv6($self->{session}->{hostname})) {
        $self->{session}->{domain} = 'udp/ipv6';
    }
    
    # Community - v1, v2c. Tharnhost parameter names are SNMP_V1_COMMUNITY
    # and SNMP_V2C_COMMUNITY depending on the version.
    my $community = "";
    if ($self->{session}->{version} eq '1') {
        $community = $host->{'SNMP_V1_COMMUNITY'};
    } elsif ($self->{session}->{version} =~ /^2c?$/i) {
        $community = $host->{'SNMP_V2C_COMMUNITY'};
    }
    $self->{session}->{community} = $snmp_auth{community} || 
        $community || SNMP_COMMUNITY;
    
    # Username - v3. Tharnhost parameter name is SNMP_V3_USERNAME
    $self->{session}->{username} = $snmp_auth{username} || 
        $host->{'SNMP_V3_USERNAME'} || $host->default_username();
    
    # security level - v3. Tharnhost parameter name is SNMP_V3_SECURITY_LEVEL
    $self->{session}->{security_level} = $snmp_auth{security_level} || 
        $host->{'SNMP_V3_SECURITY_LEVEL'} || SNMP_SECURITY_LEVEL ;
    $self->{session}->{security_level} = lc($self->{session}->{security_level});
    if ($self->{session}->{security_level} !~
         /^(noAuthNoPriv|authNoPriv|authPriv)$/i) {
        $Log->exit();
        NATE::Exceptions::Argument->throw("Invalid SNMP security level '". 
            $self->{session}->{security level}. "' is specified. Valid values ".
            "are noAuthNoPriv, authNoPriv, authPriv");
    }
    
    # Authentication related inputs - v3
    # authprotocol - v3. Tharnhost parameter name is SNMP_V3_AUTHPROTOCOL
    $self->{session}->{authprotocol} = $snmp_auth{authprotocol} || 
        $host->{'SNMP_V3_AUTHPROTOCOL'} || SNMP_AUTH_PROTOCOL;
    $self->{session}->{authprotocol} = lc($self->{session}->{authprotocol});
    # authkey - v3. Tharnhost parameter name is SNMP_V3_AUTHKEY
    $self->{session}->{authkey} = $snmp_auth{authkey} || 
        $host->{'SNMP_V3_AUTHKEY'};
    # authpassword - v3. Tharnhost parameter name is SNMP_V3_AUTHPASSWORD
    $self->{session}->{authpassword} = defined ($snmp_auth{authpassword})?
        $snmp_auth{authpassword} : 
        (defined ($host->{'SNMP_V3_AUTHPASSWORD'})? 
            $host->{'SNMP_V3_AUTHPASSWORD'}: $host->default_password());
    
    # Privacy related inputs - v3
    # privprotocol - v3. Tharnhost parameter name is SNMP_V3_PRIVPROTOCOL
    $self->{session}->{privprotocol} = $snmp_auth{privprotocol} || 
        $host->{'SNMP_V3_PRIVPROTOCOL'} || SNMP_PRIV_PROTOCOL;
    $self->{session}->{privprotocol} = lc($self->{session}->{privprotocol});
    # privkey - v3. Tharnhost parameter name is SNMP_V3_PRIVKEY
    $self->{session}->{privkey} = $snmp_auth{privkey} || 
        $host->{'SNMP_V3_PRIVKEY'};
    # privpassword - v3. Tharnhost parameter name is SNMP_V3_PRIVPASSWORD
    $self->{session}->{privpassword} = defined ($snmp_auth{privpassword})?
        $snmp_auth{privpassword} : 
        (defined ($host->{'SNMP_V3_PRIVPASSWORD'})? 
            $host->{'SNMP_V3_PRIVPASSWORD'}: $host->default_password());    
        
    $Log->exit();
}

##############################################################################
# Method:    set_test_parameters
# Objective: Set the test parameters globally
# Details:   Refer to POD documentation at the end of the file.
##############################################################################
sub set_test_parameters {
    $Log->enter();
    my $self = shift;
    
    #set Log level for SNMP on success
	my $snmp_log_level_param = 'NACL_SNMP_LOG_LEVEL';
	my $snmp_log_level_value = Tharn::param($snmp_log_level_param);
    if ( defined $snmp_log_level_value ) {
        $self->{snmp_log_level} = $snmp_log_level_value;
        $Log->debug(
            " SNMP log level override obtained from NACL_SNMP_LOG_LEVEL parameter is "
                . $self->{snmp_log_level});
    }
    
    #set Log level for SNMP on error
	my $snmp_log_level_on_error_param = 'NACL_SNMP_LOG_LEVEL_ON_ERROR';
	my $snmp_log_level_on_error_value = Tharn::param($snmp_log_level_on_error_param);
	if ( defined $snmp_log_level_on_error_value ) {
        $self->{snmp_log_level_on_error} = $snmp_log_level_on_error_value;
        $Log->debug(
            " SNMP error log level override obtained from NACL_SNMP_LOG_LEVEL_ON_ERROR parameter is "
                . $self->{snmp_log_level_on_error});
    }
    $Log->exit();
}

###############################################################################
# Method :    _log_data
# Objective:  Log the data in the specified log level
# Details:    This method uses Tharn log object ($Log) to log the specified
#             data in the given log level. Throws exception if the specified
#             log level is invalid
###############################################################################
sub _log_data {
    $Log->enter();
    my $self      = shift();
    my %args      = @_;
    my $log_level = $args{log_level};
    try {
        $Log->$log_level( $args{data} );
    }
    otherwise {
        my $ex = shift;
        if ( $ex->text() =~ /Can't locate object method/ ) {
            NATE::Exceptions::Argument->throw(
                'Invalid log level: ' . $log_level );
        } else {
            $ex->throw();
        }
    };
    $Log->exit();
}


##############################################################################
# Method:    _translate_datatypes_to_asn
# Objective: Translate datatypes to their respective ASN.1 format
# Details:   The datatypes obtained from CDEF entry is translated to the ASN.1
#            type value which can be used for specifying the type information
#            in the varbind field of 'set' request.
##############################################################################

sub _translate_datatypes_to_asn {
    $Log->enter();
    my ($self, %args) = @_;
    my %table = (
        # Datatype       # ASN.1 type
        OctetString   => 'OCTET_STRING',
        Unsigned32    => 'UNSIGNED32',
        Integer32     => 'INTEGER32',
        Enumeration   => 'INTEGER',
        Boolean       => 'INTEGER',
        Counter32     => 'COUNTER32',
        Counter64     => 'COUNTER64',
        TimeTicks     => 'TIMETICKS',
        DisplayString => 'OCTET_STRING',
        Uuid          => 'OCTET_STRING',
        PhysAddress   => 'OCTET_STRING',
        IpAddress     => 'IPADDRESS',
        Gauge32       => 'GAUGE32',      
    );
    no strict 'refs';
    # Convert the datatypes obtained from the CDEFs to the standard ASN.1
    # types as supported by Net::SNMP
    my $method = $table{$args{type}};
    my $res = &$method;
    use strict;
    $Log->debug("Datatype translation : $args{type} => $res");
    $Log->exit();
    return $res;
}

1;

##############################################################################
#                                    POD here
##############################################################################

=head1 NAME

NACL::APISet::Session::SNMP - Base session class for SNMP APISet

=head1 SYNOPSIS

 use NACL::APISet::Session::SNMP;

=head1 DESCRIPTION

A C<NACL::APISet::Session::SNMP> is used to establish a connection with the
filer/node and invoke SNMP PDUs. This is further subclassed to execute SNMP
PDUs specifically to CMode, 7Mode.

=cut

=head1 EXCEPTIONS

=over 

=item NACL::APISet::Exceptions::ConnectionFailedException

=item NACL::APISet::Exceptions::InvalidInputRequest

=item NACL::APISet::Exceptions::TimeoutException

=item NATE::Exceptions::Argument

=back

=cut

=head1 CONSTRUCTORS

=head2 new

Returns a reference to a newly created C<NACL::APISet::Session::SNMP::CMode> or
C<NACL::APISet::Session::SNMP::7Mode> object. This method should not be
directly called.

=over 

=item Synopsis

 $snmp_session = NACL::APISet::Node::SNMP::CMode->new( %args );
 or
 $snmp_session = NACL::APISet::Node::SNMP::7Mode->new( %args );

=item Arguments

=over 

=item hostobj

(Optional)
Hostrec object for the node from which all other SNMP related information
like Ipaddress, username, community, security level, privprotocol,
privpassword, privkey, authprotocol, authpassword, authkey can be obtained.
hostobj and connection are mutually exclusive parameters.

=item connection

(Optional)
Connectrec object for the node. This connection object will not be used to
execute SNMP PDUs. This will just be used to obtain hostrec from which all
other SNMP session related information can be obtained.
hostobj and connection are mutually exclusive parameters.

=item ipaddress

(Optional)
IpAddress override of the host

=item snmp_auth

(Optional)
This takes the input as a reference to hash structure which contains information
about version (possible values are 1, 2c ,3), community (for SNMP v1 and SNMPv2c),
username, security_level, authprotocol, authkey, authpassword, privprotocol,
privkey and privpassword for SNMPv3. All these except version will act as 
overrides to the values specified in tharnhost if any.

=back

=item Returns

A reference to a new SNMP session object either for 7Mode or CMode
depending on whom the caller is.

=back

=cut

=head1 METHODS

=head2 clear_timeout

This method is used to set the timeout to the default value (5 seconds).

=over

=item Synopsis

$snmp_session->clear_timeout();

=item Arguments

None.

=item Returns

Nothing.

=back

=head2 set_timeout

This method is used to set the timeout to a new value. It throws
NATE::BaseException if it couldn't set the timeout to the specified value.

=over

=item Synopsis

$snmp_session->set_timeout(timeout => $new_timeout);

=item Arguments

=over

=item timeout

The new timeout value in seconds

=back

=item Returns

Nothing.

=back

=head2 get_timeout

This method returns the current timeout value associated with SNMP session.

=over

=item Synopsis

my $timeout = $snmp_session->get_timeout();

=item Arguments

Nothing.

=item Returns

The current timeout value in seconds.

=back

=head2 connect

This method is used to establish a SNMP session to the service. It creates a
NET::SNMP session object which will be used as a connection object to
execute SNMP PDUs. This sets the created connection object in SNMP session 
object itself.

If the Net::SNMP object cannot be created, a ConnectionFailedException is 
thrown.

=over

=item Synopsis

$snmp_session->connect();

=item Arguments

Nothing.

=item Returns

Nothing.

=back

=head2 cleanup

This method clears the transport Domain and any errors associated with the 
NET::SNMP object created. Once closed, the Net::SNMP object can no longer be
used to send or receive SNMP messages.

=over

=item Synopsis

$snmp_session->cleanup();

=item Arguments

Nothing.

=item Returns

Nothing.

=back

=head2 execute_snmp_method

This method can be used to execute any SNMP request PDU and snmp_walk.

=over

=item Synopsis

my $output = $snmp_session->execute_snmp_method(%args);

=item Arguments

=over

=item method

(Mandatory)
The operation to be invoked. Valid values are get, set, get_next, get_bulk and
snmp_walk.

=item varbind

(Optional)
Hash reference whose keys are object names and values are the corresponding numeric
OIDs. This is a mandatory option for all operations other than snmp_walk.

=item baseoid

(Optional)
BaseOID to be used for performing snmp walk. This can be specified
only if the 'method' name is snmp_walk.

=item maxrepetitions

(Optional)
maxrepetition value for get_bulk PDU.

=item nonrepeaters

(Optional)
nonrepeaters value for get_bulk PDU.

=item on-error

(Optional)
Error handler override if the object retrival fails with reasons such as 
noSuchObject, noSuchInstance, endOfMibView. The default is 'exception', throw
exception on failure.

=item schema

(Optional)
schema object which is required by response class to parse the output if response
object is requested. 

=back

=item Returns

Returns a hash reference as output having information about raw output, 
execution time, execution status, error index, error reason, asn1_type 
information, command executed.

=back

=head2 execute_raw_command

This method can be used to execute any SNMP request PDU and snmp_walk by
directly specifying the numeric OIDs. Also user can customize whether
this method should return raw output or a response object

=over

=item Synopsis

my $output = $snmp_session->execute_raw_command(%args);

=item Arguments

=over

=item method

(Mandatory)
The operation to be invoked. Valid values are get, set, get_next, get_bulk and
snmp_walk.

=item varbind

(Optional)
Array reference specifying the list of numeric OIDs
This is a mandatory option for all operations other than snmp_walk.

=item baseoid

(Optional)
BaseOID to be used for performing snmp walk. This can be specified
only if the 'method' name is snmp_walk.

=item maxrepetitions

(Optional)
maxrepetition value for get_bulk PDU.

=item nonrepeaters

(Optional)
nonrepeaters value for get_bulk PDU.

=item schema

(Optional)
schema object which is required by response class to parse the output if response
object is requested. 

=item return-response-object

(Optional)
If set to 1, a response object will be returned else raw output (hash ref whose
keys are numeric OIDs and values are values for the corresponding OIDs) will
be returned by default.

=back

=item Returns

Returns either response object or raw output based on 'return-response-object'
option.

=back

=head2 set_test_parameters

To set the test parameters globally

=head3 Parameters

=over

=item NACL_SNMP_LOG_LEVEL

Takes a values comment | trace | debug and defaults to comment
This parameter decides at what level the logging of SNMP Input and 
Output to be set at, irrespective of the SNMP request succeeded
or failed

=item NACL_SNMP_LOG_LEVEL_ON_ERROR

Takes a values comment | trace | debug and defaults to comment
This parameter decides at what level the logging of SNMP Input and 
Output to be set at, when there is an error in the SNMP request

=back

=head1 AUTHOR/MAINTAINER

=over

=item NACL Development (dl-nacl-dev@netapp.com)

=back

=cut
