#
# Copyright (c) 2001-2010 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Class to hold all connection information for CMode Sessions
## @author Isshwarya.M@netapp.com, dl-nacl-dev@netapp.com
## @status Shared
## @pod here

# Not all 'connections' to services will be in the form of NATE::Connectionrec
# The Session class will define all data needed for connections, whether
# NATE Connection or not.

package NACL::APISet::Session::SNMP::CMode;

#
# Compiler Directives
#
use strict;
use warnings;

use base qw(NACL::APISet::Session::SNMP);

# Module import - APISets
use NACL::APISet::Response::SNMP::CMode;

#
# NATE imports
#
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Data::Dumper;

#
# Exceptions
#
use NACL::APISet::Exceptions::ObjectRetrievalFailed qw(:try);
use NATE::Exceptions::Argument qw(:try);


##############################################################################
# Method:    send_command
# Objective: Send the SNMP PDU request to the service.
# Details:   Refer to POD documentation at the end of the file.
##############################################################################

sub send_command {
    $Log->enter();
    my ($self, %args) = @_;
    
    my $error_handler = (defined $args{'on-error'})?
        delete $args{'on-error'} : $self->{error_handler};
    $Log->debug("Error handler: " . $error_handler);
    
    my $output = $self->execute_snmp_method(%args);
    my $response;
    # Return a CMode SNMP response object
    try {
        $response = NACL::APISet::Response::SNMP::CMode->new(%{$output},
            need_cdef_entry => $self->{cdef_entry_in_response});
    } catch NACL::APISet::Exceptions::ObjectRetrievalFailed with {
        my $ex = shift;
        $Log->debug("Error: ". $ex->text());
        if (ref($error_handler) eq 'CODE') {
            # Return the response object as well
            $response = $ex->get_response_object();
            &$error_handler(exception_object => $ex);            
        } elsif ($error_handler eq 'exception') {
            $ex->throw();
        } elsif ($error_handler eq 'ignore') {
            # Already we caught exception. Set the response object
            $response = $ex->get_response_object();
        } else {
            NATE::Exceptions::Argument->throw(
                "Invalid error handler specified: $error_handler");
        }
    };
    $Log->exit();
    return $response;
}
1;

#############################################################################
#                             POD Documentation
#############################################################################

=pod

=head1 NAME

NACL::APISet::Session::SNMP::CMode - Session class for sending SNMP requests
to CMode nodes.

=head1 SYNOPSIS

 use NACL::APISet::Session::SNMP::CMode;

=head1 DESCRIPTION

A C<NACL::APISet::Session::SNMP::CMode> provides the methods needs to control
the connection to a CMode node.  

=cut

=head1 METHODS

=head2 send_command

Send the SNMP request to the node.  The output of the request is returned in a 
NACL::APISet::Response::SNMP::CMode object.

=over 

=item Synopsis

$response = $session->send_command( %args, 'on-error' => $handler );

=item Arguments

=over 

=item on-error

(Optional)
The error handler to be invoked when ObjectRetrievalFailed exception is thrown.

= item %args

All the other arguments taken by this method are described in 
execute_snmp_command() of C<NACL::APISet::Session::SNMP>.

=back

=item Returns

Response object of type NACL::APISet::Response::SNMP::CMode

=back

=head1 AUTHOR/MAINTAINER

=over

=item NACL Development (dl-nacl-dev@netapp.com)

=back

=cut



