## @summary Class to hold all Symbolapi connection information.
## @author dl-nacl-dev@netapp.com
## @status Shared

package NACL::APISet::Session::Symbolapi;

#
# Compiler Directives
#
use strict;
use warnings;

#
# NATE imports
#
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Params::Validate qw(:all);
use Data::Dumper;

#
# Exceptions imported
#
use NACL::APISet::Exceptions::ConnectionFailedException qw(:try);
use NATE::Exceptions::Argument qw(:try);
use NACL::APISet::Response::Symbolapi;

use base qw(NACL::APISet::Session);

sub new {
    $Log->enter() if $may_enter;
    my ($class, @opts) = @_;
    
    my %args = validate_with(
        params => \@opts,
        spec   => { additional_spec => { type => HASHREF, default => {} } },
        allow_extra => 1
    );
    my $additional_spec = delete $args{additional_spec};

    # Fill the member vars for this instance
    my $self = NACL::APISet->validate_params(
        params => \%args,
        spec   => {
            username => { type => SCALAR, optional => 1},
            password => { type => SCALAR, optional => 1 },
            conntype => { type => SCALAR, optional => 1, default => "api" },
            ipaddress => { type => SCALAR, optional => 1 },
            connection => { type => OBJECT, isa => 'Connectrec', optional => 1 },
            hostobj => { type => OBJECT, isa => 'Hostrec', optional => 1 },
            %$additional_spec
        },
    );

    bless $self, $class;
    
    # Connect to the service
    $self->connect();

    # And give the caller the referece to the bless object.
    $Log->exit() if $may_exit;
    return $self;
}

###############################################################################
# Method:    connect
# Objective: Connect to the service
# Details:   See POD documentation at the end of this file.
###############################################################################
sub connect {
    $Log->enter() if $may_enter;

    # %args is usually empty. But the individual Session classes can send in
    # something that has to be passed to Hostrec->connect() additionally

    my ( $self, %args ) = @_;

    try {
        if ( defined $self->{connection} ) {
            $self->set_connection( connection => $self->{connection} );
        } else {
            # Check for user-defined overrides. If any of these are
            # passed we add them to the connspec
            
            my @overrides = (
                qw (conntype username password prompt_username
                    prompt_password connhost)
            );
            my %conn_overrides;
            foreach my $param (@overrides) {
                if ( defined $self->{$param} ) {
                    $conn_overrides{$param} = $self->{$param};
                }
            }
            if (defined $self->{connid}) {
                $conn_overrides{template} = $self->{connid};
            } else {
                $conn_overrides{template} = "symbolapi";
            }
            
            my $connection = $self->{hostobj}->connect(%conn_overrides);
	    	$self->set_connection(
                connection => $connection
            );
	    	$Log->debug( Dumper( $self->{connection} ) );
        }
    }
    otherwise {
        NACL::APISet::Exceptions::ConnectionFailedException->throw(
            "Could not create connection for #symbolapi $@" );
    };
    $Log->exit() if $may_exit;
}

###############################################################################
# Method:    set_connection
# Objective: Set the Session object to internally hold the connection object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub set_connection {
    $Log->enter() if $may_enter;
     my ( $self, %args ) = @_;
    $self->{connection} = $args{connection};
    # storing the Default value so that we can use this in clear_timeout()
    # Only if default_timeout is not defined, store the values in the
    # session object
    if ( !defined $self->{default_timeout} ) {
        $self->{default_timeout} = $self->{connection}->timeout();
        $self->{timeout}         = $self->{default_timeout};
    }

    $Log->exit() if $may_exit;
}

###############################################################################
# Method:    set_timeout
# Objective: Set the timeout to a new value
# Details:   See POD documentation at the end of this file.
###############################################################################
sub set_timeout {
    my $self = shift;
    $Log->enter() if $may_enter;
    my %args = validate_with(
        params => \@_,
        spec   => {
            timeout              => { type => SCALAR, optional => 1 },
        }
    );
    my $timeout;
    if (defined $args{timeout}) {
        $timeout = $args{timeout};
    } else {
        NATE::Exceptions::Argument->throw(" 'timeout' should have been "
                . 'passed but it wasnt' );
    }
    $self->{connection}->timeout($timeout);
    $self->{timeout} = $self->{connection}->timeout();
    $Log->debug( 'Timeout set to ' . $self->{timeout} );
    $Log->exit() if $may_exit;
}

###############################################################################
# Method:    clear_timeout
# Objective: Set the timeout to the default value
# Details:   See POD documentation at the end of this file.
###############################################################################
sub clear_timeout {
    my $self = shift;
    $Log->enter() if $may_enter;

    $self->{connection}->timeout( $self->{default_timeout} );
    $self->{timeout} = $self->{connection}->timeout();
    $Log->debug( "Timeout has been reset to " . $self->{timeout} );

    $Log->exit() if $may_exit;
}

##############################################################################
# Method:    send_command
# Objective: Invoke the command on the Eseries host
# Details:   See POD documentation at the end of this file.
##############################################################################
sub send_command {
    $Log->enter() if $may_enter;

    my ($self, %opts) = @_;
    my %args = validate_with(
        params      => \%opts,
        spec        => { connectrec => { type => HASHREF, default => {} } },
        allow_extra => 1
    );
 
    my %connectrec = %{delete $args{connectrec}};
    
    unless ($args{method_name}) {
        # Caller is execute_command, all the options passed
        # should be considered as connectrec options
        %connectrec = ( %connectrec, %args );
        $args{command} ||= $connectrec{command};
        delete $connectrec{command};
    }
    
    $Log->debug( sub { "Arguments that are being passed to execute(): ". "command = '".
                $args{command}. "'\n". Dumper(\%connectrec) } );
    
    my ( $time_start, $output, $time_end );
    
        $time_start = time();
        $output = $self->{connection}->execute(command => $args{command},
                                               %connectrec);
        $time_end = time();

    
    my $execution_time = $time_end - $time_start;
    
    # for execute_raw_command, just return the raw output
    if(!$args{method_name}) {
         $Log->exit() if $may_exit;
         return $output;
    }

    my $resp;
    
    $resp = NACL::APISet::Response::Symbolapi->new(
                output         => $output,
                command        => $args{command},
                execution_time => $execution_time,
                method_name    => $args{method_name},
                actual_command => $args{actual_command},
        );
    
    $Log->exit() if $may_exit;
    return $resp;
}

1;

##############################################################################
#                                    POD here
##############################################################################



=head1 NAME

NACL::APISet::Session::Symbolapi - API Set Session class for Symbolapi

=head1 SYNOPSIS

 use NACL::APISet::Session::Symbolapi;

=head1 DESCRIPTION

A C<NACL::APISet::Session::Symbolapi> contains data and methods common to all API Set
Session Symbolapi classes. This includes implementations for connect(), set_connection()
and send_command().
This class provides functionality to all Session classes for Symbolapi.

=head1 EXCEPTIONS

=over

=item NACL::APISet::Exceptions::ConnectionFailedException

=back

=head1 CONSTRUCTOR

=head2 new

Returns a reference to a newly created C<NACL::APISet::Session::Symbolapi>.  

=head1 Methods

=head2 connect

This method is used to connect to the service. It does this by either using the
connection object provided, or by constructing a full connection spec and
connecting using this full connection spec.

If the connection cannot be established, a ConnectionFailedException is thrown.

=over

=item Synopsis

$self->{session}->connect();

=item Arguments

=over

=item %Additional_Connection_Parameters

A hash containing any additional arguments to Hostrec->connect()

=back

=item Returns

Nothing.

=back

=head2 set_connection

This method is used to set the Session object to internally hold the connection
object created.

=over

=item Synopsis

$self->{session}->set_connection(connection => $conn_object);

=item Arguments

=over

=item connection

This is the connection object created.

=back

=item Returns

Nothing.

=back

=head2 clear_timeout

This method is used to set the timeout to the default value.
Default = Timeout value that was present when the connectrec object was created

=over

=item Synopsis

$timeout = $self->{session}->clear_timeout();

=item Arguments

None.

=item Returns

Nothing.

=back

=head2 set_timeout

This method is used to set the timeout to a new value.

=over

=item Synopsis

    $self->{session}->set_timeout('connectrec-timeout' => $new_timeout);

                                 or

    $self->{session}->set_timeout(timeout => $new_timeout);

=item Arguments

=over

=item timeout

The new timeout value (in seconds)

=back

=item Returns

Nothing.

=back
