# $Id: //depot/prod/test/nacldev/lib/NACL/APISet/Session/VMAPI.pm $
#
# Copyright (c) 2001-2011 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Construction of VIM object connected to the specified ESX server
## @author bkprasha@netapp.com, dl-nacl-dev@netapp.com
## @status Shared
## @pod here

package NACL::APISet::Session::VMAPI;

#
# Compiler Directives
#
use strict;
use warnings;

#
# NATE imports
#
use Tharn;
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Params::Validate qw(:all);
use Data::Dumper;

#
# VMWare imports
#
use VMware::VIRuntime;
use VMware::VILib;
use VMware::VIExt;

#
# Exceptions imported
#
use NACL::APISet::Exceptions::ConnectionFailedException qw(:try);

use base qw(NACL::APISet::Session);

###############################################################################
# Method:    esx_connect
# Objective: Connect to the VIM SDK service on the ESX server
# Details:   See POD documentation at the end of this file.
###############################################################################
sub esx_connect {
    $Log->enter() if $may_enter;

    my ( $self, %args ) = @_;

    my $service_url = "https://" . $self->{ipaddress} . "/sdk/vimService";
    $Log->trace($service_url);

    try {
        my $vim_obj = Vim->new( service_url => $service_url );
        my $vim_obj_session = $vim_obj->login(
            user_name => $self->{username},
            password  => $self->{password}
        );

        $self->set_connection( connection => $vim_obj_session );
        $Log->debug( Dumper( $self->{connection} ) );
    }
    otherwise {
        NACL::APISet::Exceptions::ConnectionFailedException->throw(
                  text => "Could not create a connection to: "
                . $service_url
                . "\n$@", );
    };

} ## end sub esx_connect

###############################################################################
# Method:    set_connection
# Objective: Set the Session object to internally hold the VIM connection object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub set_connection {
    $Log->enter() if $may_enter;

    my ( $self, %args ) = @_;
    $self->{connection} = $args{connection};

    $Log->exit() if $may_exit;
} ## end sub set_connection

###############################################################################
# Method:    set_timeout
# Objective: Set the timeout to a new value
# Details:   See POD documentation at the end of this file.
###############################################################################
sub set_timeout {
    my $self = shift;
    $Log->enter() if $may_enter;
    my %args = validate_with(
        params => \@_,
        spec   => { timeout => { type => SCALAR, optional => 0 }, }
    );
    my $outp = $self->{connection}->{vim_service}->{vim_soap}->{user_agent}
        ->timeout( $args{timeout} );
    unless ( defined $outp ) {
        NATE::BaseException->throw( "Failed to set timeout and the"
                . "reason is: "
                . $self->{connection}->error() );
    }
    $Log->exit() if $may_exit;
} ## end sub set_timeout

###############################################################################
# Method:    get_timeout
# Objective: Get the current timeout
# Details:   See POD documentation at the end of this file.
###############################################################################

sub get_timeout {
    my $self = shift;
    $Log->enter() if $may_enter;
    my $timeout = $self->{connection}->{vim_service}->{vim_soap}->{user_agent}
        ->timeout();
    $Log->exit() if $may_exit;
    return $timeout;
} ## end sub get_timeout

=head1 NAME

NACL::APISet::Session::VMAPI - Construction of VIM object connected to the specified ESX server

=head1 SYNOPSIS

 use NACL::APISet::Session::VMAPI;

=head1 DESCRIPTION

A C<NACL::APISet::Session::VMAPI> is used to construct the VIM object connected to the specified 
ESX server and include it as a value to the APISet session connection parameter

=head1 EXCEPTIONS

=over

=item NACL::APISet::Exceptions::ConnectionFailedException

=back

=head1 Methods

=head2 esx_connect

Connect to the VIM SDK service on the ESX server specified by the user in APISet constructor.
Uses the credentials provided by the user to connect

=over

=item Synopsis

$self->esx_connect();

where $self is a NACL::APISet::Session::VMAPI::ESX object

=item Arguments

Accepts all the arguments passed on to it from the APISet constructor. Uses the ipaddress, 
username, password to connect and login

=back

=head2 set_connection

Set the Session object to internally hold the VIM connection object

=over

=item Synopsis

$self->set_connection()

=item Arguments

A connection object to the VIM SDK service

=back

# $Id: //depot/prod/test/nacldev/lib/NACL/APISet/Session/VMAPI.pm $

=cut

1;
