#
# Copyright (c) 2012-* NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Construction of Xen Session Object to the specified Xen server
## @author rawat@netapp.com, dl-nacl-dev@netapp.com
## @status Shared
## @pod here

package NACL::APISet::Session::XAPI;

# Compiler Directives
use strict;
use warnings;

# NATE imports
use Tharn;
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Params::Validate qw(:all);
use Data::Dumper;

# Import Xen RCP API
use RPC::XML::Client;

# Exceptions imported
use NATE::BaseException qw(:try);
use NACL::APISet::Exceptions::ConnectionFailedException;

use base qw(NACL::APISet::Session);

###############################################################################
# Method:    xen_connect
# Objective: Connect to Xen server
# Details:   See POD documentation at the end of this file.
###############################################################################
sub xen_connect {
    $Log->enter() if $may_enter;

    my ( $self, %args ) = @_;

    my $service_url = "http://" . $self->{ipaddress};
    $Log->comment("Connecting to service url " . $service_url);

    my $rpc_clientobj = RPC::XML::Client->new($service_url);
    my $session;

    my $req =  $rpc_clientobj->simple_request(
                                        "session.login_with_password",
                                        $self->{username},
                                        $self->{password},
                                );
    $Log->debug(Dumper($req));
    
    unless ($req) {
        NACL::APISet::Exceptions::ConnectionFailedException
                ->throw("Could not create a connection to: "
                        . $service_url . "\n error is $RPC::XML::ERROR");
    }

    $session = $req->{Value};

    $self->{xen_session} = $session;
    $self->{rpc_clientobj} = $rpc_clientobj;
    $self->{xen_host} = $self->_extract_value($rpc_clientobj->simple_request("session.get_this_host", $session, $session));

} ## end sub xen_connect

###############################################################################
# Method:    set_timeout
# Objective: Set the timeout to a new value
# Details:   See POD documentation at the end of this file.
###############################################################################
sub set_timeout {
    my ($self, @opts) = @_;
    $Log->enter() if $may_enter;
    my %args = validate_with(
        params => \@opts,
        spec   => { timeout => { type => SCALAR, default => 60 }, }
    );

    my $outp = $self->{rpc_clientobj}->timeout( $args{timeout} );
    unless ( defined $outp ) {
        NATE::BaseException->throw( "Failed to set timeout and the"
                . "reason is: "
                . $self->{connection}->error() );
    }
    $Log->exit() if $may_exit;
} ## end sub set_timeout

###############################################################################
# Method:    get_timeout
# Objective: Get the current timeout
# Details:   See POD documentation at the end of this file.
###############################################################################

sub get_timeout {
    my ($self) = @_;
    $Log->enter() if $may_enter;
    my $timeout = $self->{rpc_clientobj}->timeout();
    $Log->exit() if $may_exit;
    return $timeout;
} ## end sub get_timeout


sub _extract_value {
    my ($self, $val) = @_;
    $Log->enter() if $may_enter;
    
    if ($val->{'Status'} eq "Success"){
        $Log->exit() if $may_exit;
        return $val->{'Value'};
    } else {
        my $error = $val->{ErrorDescription}->[0];
        $Log->exit() if $may_exit;
        NATE::BaseException->throw($error);
    }
}

=head1 NAME

NACL::APISet::Session::XAPI - Construction of Xen session object connected
to the specified Xen server

=head1 SYNOPSIS

 use NACL::APISet::Session::XAPI;

=head1 DESCRIPTION

A C<NACL::APISet::Session::XAPI> is used to construct the Xen session object
connected to the specified Xen server and include it as a value to the
APISet session connection parameter

=head1 EXCEPTIONS

=over

=item NACL::APISet::Exceptions::ConnectionFailedException

=back

=head1 Methods

=over

=back

=head2 xen_connect

Will create RCP::XML::Client object which will refer to Xen server. Will use this
object to create Xen server session object. This session will be used
to get Xen Server Object

=over

=item Synopsis

$self->xen_connect();

where $self is a NACL::APISet::Session::XAPI::Xen object

=item Arguments

Accepts all the arguments passed on to it from the APISet constructor. Uses the ipaddress,
username, password to connect and login

=back

=head2 set_timeout

Set the timeout on RPC::XML::Client object for Xen server

=over

=item Arguments

=item timeout

Timeout to be set on RPC::XML::Client object

=back

=head2 get_timeout

Returns the timeout set on RPC::XML::Client object for Xen server

=over

=item Arguments

=item Returns

timeout set on RPC::XML::Client object for Xen server

=back

=cut

1;
