#
# Copyright (c) 2001-2012 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Base for API Sets CLI commands for Switch
## @author dl-nacl-dev@netapp.com
## @status Shared
## @pod here

package NACL::APISet::Switch::CLI;

#
# Compiler Directives
#
use strict;
use warnings;
use Params::Validate qw(SCALAR BOOLEAN);

# NATE imports
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Data::Dumper;

# Module import - APISets
use NACL::APISet::Schema;
use NACL::APISet::_Mixins::CLI qw(:all);

#
# Exceptions imported
#
use NATE::Exceptions::Argument qw(:try);
use NACL::APISet::Exceptions::CDEFNotFoundException;

# Base class
use base qw(NACL::APISet::Switch);

our $AUTOLOAD;

###############################################################################
# Method Name: AUTOLOAD
# Objective:  Catch all commands not specifically implemented and call invoke
#             If the API call is prefixed with an "xlate_" prefix, it should
#             only be translated and not invoked, so we should call
#             _only_translate_method
###############################################################################

sub AUTOLOAD {
    my ($self,@opts) = @_;
    $Log->enter() if $may_enter;
    my $method = $AUTOLOAD;
    $method =~ s/.*://;    # strip fully-qualified portion
    my %args = $self->_check_for_unnamed_parameters(options => \@opts,
                                                    method  => $method);
    $Log->debug(sub { "Options to '$method' are:\n".Dumper(\%args)});

    if ($method =~ /^xlate_/) {
        if ($method =~ /^xlate_execute(_raw)?_command/) {
            my $name = $1 || "";
            # Explicitly prevent xlate_ being used with execute_raw_command
            $Log->exit() if $may_exit;
            NATE::BaseException->throw(
                          "xlate cannot be used with execute$name" .
                          "_command");
        } else {
            $Log->exit() if $may_exit;
            return $self->_only_translate_method(method => $method, %args);
        }
    } else {
        # Call invoke with the arguments given to execute the command
        $Log->exit() if $may_exit;
        return $self->_invoke(method => $method, %args);
    }

} ## end sub AUTOLOAD

##############################################################################
# Method:    set_privilege
# Objective: For running the command as the specified user
# Details:   Refer to POD documentation at the end of the file.
##############################################################################

sub set_privilege {
    $Log->enter() if $may_enter;

    my $self = shift();
    my %args =
      NACL::APISet::validate_params(
                       $self,
                       params => \@_,
                       spec   => {
                           "privilege-level" => {type => SCALAR, optional => 1},
                           "enable_password" => {type => SCALAR, optional => 1},
                       });
    my $privilege = delete ($args{'privilege-level'});
    my $password = delete ($args{'enable_password'});
    
    # If enable password is not passed by user then use PROMPT_PASSWORD.
    # If PROMPT_PASSWORD is also not set then use normal password used to
    # connect to the machine. PROMPT_PASSWORD can be set by passing it inside
    # conn in tharnhost files
    # PROMPT_PASSWORD=<prompt_password>
    if (! defined $password ) {
        if (defined $self->{session}->{connection}->{hostp}->{PROMPT_PASSWORD}) {
            $password = $self->{session}->{connection}->{hostp}->{PROMPT_PASSWORD};
        } else {
            $password = $self->{session}->{connection}->{password};
        }
    }
    
    if ($privilege =~ /enable/) {
        $self->enable('connectrec-match_table' => ["\[Pp\]assword:" => $password]);
    } elsif ($privilege =~ /configure/) {
        $self->configure();
    }

    # For QlogicFC to go to the admin mode 'admin start' is issued	
    if ($privilege =~ /admin/ && $self->{set} =~ /QlogicFC/) {
        $self->admin_start();
    }
    $Log->exit() if $may_exit;
} ## end sub set_privilege

##############################################################################
#                               Private Methods
##############################################################################

##############################################################################
# Method:    _init_API
# Objective: Initialize the Switch API Set
# Details:   Initializes the API Set.  This method drives the workflow needed
#            to make the API Set ready for work.  This is accomplished by
#            creating a new schema object and loading it with CDEFs from the
#            appropriate path.
##############################################################################

sub _init_API {
    my ($self, %args) = @_;
    $Log->enter() if $may_enter;

    # set default privilege level
    $self->{'privilege-level'} = "";

    # Create the schema object and load the CDEFs from CDEF path discovered by
    # versioning or the user specified path. Load CDEFS after the check for
    # shell_type is done so that we need not unnessarily load CDEFs even in
    # case of invalid shell_type.
    my $cdef_path = $self->{cdef_path};
    $self->{schema} =
      NACL::APISet::Schema->new();

    # Get the list of th CDEF files to be loaded. The implementation of getting
    # the CDEF files list will go into the specific APISet class.
    # Handle linux, windows and HTTP paths as well.
    my @files = @{$self->{schema}->read_cdef_dir(directory => $cdef_path)};

    # if files list is empty, cdef files have not been consolidated
    if (scalar(@files) == 0) {
        $Log->exit() if $may_exit;
        NACL::APISet::Exceptions::CDEFNotFoundException->throw(
                                "No cdef data files found in: $cdef_path ");
    }

    # Load the cdef files
    $self->{schema}->load(files => \@files, path => $cdef_path);

    # For Network Cisco IOS switch all the command need to be
    # run in 'enable' mode
    my $major = $self->{version_manager}->{os_revision}->{major};
    if (($self->{set} =~ /NetAppNetwork/) or ($self->{set} =~ /CiscoNetwork/ and $major =~ /IOS/i)) {
        $self->set_privilege('privilege-level' => "enable");
    }

    $Log->exit() if $may_exit;
} ## end sub _init_API

##############################################################################
# Method:    _init_Session
# Objective: Initialize the Switch Session object
# Details:   This method establishes the Session and initializes the APISet
#            object with the session object.
##############################################################################

sub _init_Session {
    my ( $self, %args ) = @_;
    $Log->enter() if $may_enter;

    $self->_get_connectrec_or_hostrec(\%args);

    my $interface = $self->{interface};

    # Establish the session if not provided
    # First check whether a valid shell_host (hostobj) is passed
    # for the context
    $self->_check_session_type();

    # A normal Switch APISet is requested
    my $session_type = $self->{set};
    
        # Establish the session if not provided
    my $session_pkg =
      "NACL::APISet::Session::" . $interface . "::" . $session_type;

    # Create a Session object. This takes care of connecting to the service
    # and preparing it for use.
    $self->{session} = $session_pkg->new(%args);
    
    $Log->exit() if $may_exit;
}   
###############################################################################
# Method Name: _check_session_type
# Objective:   Check whether the shell_type passed by the user is appropriate
#              for the shell_switch
###############################################################################
sub _check_session_type {
    $Log->enter() if $may_enter;
    my ($self, %args) = @_;

    # Valid combinations for shell_type and set passed by the user
    my %valid_hash = (

        #actual_set         shell_type
        "BrocadeFC"      => "Switch",
        "CiscoFC"        => "Switch",
        "CiscoNetwork"   => "Switch",
        "NetAppNetwork"  => "Switch",
        "Armadillo"      => "Switch",
        "QlogicFC"       => "Switch",
        "DellEthernet"   => "Switch",
        "Mellanox"       => "Switch",
    );
    if ($valid_hash{$self->{actual_set}} !~ /$self->{shell_type}/) {
        $Log->exit() if $may_exit;
        NATE::Exceptions::Argument->throw("Invalid hostobj or shell_type "
                                       . "is sent for set '$self->{actual_set}'"
                                       . " and the shell_type obtained is "
                                       . $self->{shell_type});
    }
    $Log->exit() if $may_exit;
} ## end sub _check_session_type

###############################################################################
# Method Name: _invoke
# Objective:   translate cdef schema into actual command and execute
# Details:     This method invokes the command along with its options and returns
#              its output. It does this by getting the CDEF schema, translating
#              this CDEF schema into the actual command to be invoked on the
#              service, and then invoking that on the filer using the Session
#              object.
###############################################################################

sub _invoke {
    my ($self) = shift();
    $Log->enter() if $may_enter;

    $self->api_reconnect_session() if
                   $self->get_connection()->is_disconnected(verify => 0);

    my (%connectrec, $actual_command);
    my %args =
      NACL::APISet->validate_params(
                       params => \@_,
                       spec   => {
                           "privilege-level" => {type => SCALAR, optional => 1},
                               });

    my $privilege       = delete $args{'privilege-level'};

    %args = @_;

    my $method = $args{method};

    # process parameters
    while (my ($key, $value) = each(%args)) {
        if ($key =~ /-?method$/) {
            if (!length($value)) {
                $Log->exit() if $may_exit;
                NATE::Exceptions::Argument->throw(
                                       "Invalid value for \"$key\": " . $value);
            }
        } elsif ($key =~ /-?debug_level$/) {
            if (!length($value) || $value !~ /\d+/) {
                $Log->exit() if $may_exit;
                NATE::Exceptions::Argument->throw(
                                       "Invalid value for \"$key\": " . $value);
            }

            # delete args that is not needed for translation
            delete $args{$key};
            next;
        } elsif ($key =~ /connectrec\-(.*)/ || $method eq 'execute_command' ) {
            # All options prefixed with a connectrec are meant as options for
            # Connectrec and should not be translated by the CDEFs. These
            # are moved into a separate hash and delete from the args hash
            # so that they are not sent for translation
            $connectrec{$key} = $value;
            delete $args{$key};
        }
    } ## end while (my ($key, $value) ...
    
    # verify mandatory cdef argument
    unless (defined $args{method} and length $args{method}) {
        $Log->exit() if $may_exit;
        NATE::Exceptions::Argument->throw(
                                  "Mandatory argument \"cdef\" not specified ");
    }


    my ($command, $file) = ($connectrec{command}, undef);
    if ($method ne 'execute_command') {
        
        # Find parser file
        $file = $self->_find_parser_file(method => $method);
        # retrieve schema handle
        my $schema_obj = $self->get_schema_obj();

        # translate cdef schema into command
        $command = $schema_obj->translate(%args);

    }

    $connectrec{"connectrec-on_timeout"} = $self->_throw_timeout_exception(
                                before => $connectrec{"connectrec-on_timeout"} || delete $connectrec{'on_timeout'} );
    $connectrec{"connectrec-on_idle"} = $self->_throw_timeout_exception(
                                   before => $connectrec{"connectrec-on_idle"}   || delete $connectrec{'on_idle'} );
    my %connectrec_ptr;
    $self->_move_connectrec_options(
        source => \%connectrec,
        target => \%connectrec_ptr
    );
    # execute command
    my $result;
    if ($method ne 'execute_command' ) {
        $result =
            $self->{session}->send_command(
                     command        => $command,
                     method_name    => $method,
                     cdef_path      => $self->{cdef_path},
                     connectrec     => \%connectrec_ptr,
                     parser_file    => $file
            );
    } else {
        delete $connectrec{method};
        $result =
            $self->{session}->send_command(
                connectrec      => \%connectrec_ptr,
                command 		=> $command,
            );
    }

    $Log->exit() if $may_exit;
    return $result;

} ## end sub _invoke

###############################################################################
# Method Name: _only_translate_method
# Objective:   Translate the API call into the command invoked on the service
#              and return this without executing it.
# Details:     When the API is prefixed by an "xlate_", it must only translate
#              the API call without executing it. This will come in handy
#              while developing new APISets and for cases where undoing
#              API calls can be cumbersome.
#              This method strips the leading xlate_ prefix, translates
#              the API and returns the command
###############################################################################
sub _only_translate_method {
    my ($self, %args) = @_;
    $Log->enter() if $may_enter;

    $args{method} =~ s/^xlate_//;
    my $schema_obj = $self->get_schema_obj();
    my $command    = $schema_obj->translate(%args);
    $Log->exit() if $may_exit;
    return $command;
} ## end sub _only_translate_method

1;

##############################################################################
#                                    POD here
##############################################################################

=head1 NAME

NACL::APISet::Switch::CLI - Base for API Sets for CLI commands

=head1 SYNOPSIS

 use NACL::APISet::Switch::CLI;

=head1 DESCRIPTION

A C<NACL::APISet::Switch::CLI> is used to connect to and send CLI commands. This
is further subclassed to send CLI commands specifically to BrocadeFC, CicsoFC,
BrocadeNetWork etc.

=head1 EXCEPTIONS

=over

=item NATE::Exceptions::Argument

=item NACL::APISet::Exceptions::CDEFNotFoundException

=back

=head1 METHODS

=head2 execute_raw_command

This method is used to execute raw CLI commands on the switch. The entire command
line is to be provided exactly as it would be on the switch. This command is
executed and the raw output is returned as a scalar. This is useful for cases
where we want to invoke a command without parsing the output, or for cases
where the particular parser has not been defined yet but we want to see the
raw output.

=over

=item Synopsis

$raw_output = $Api_Set_Obj->execute_raw_command(command => "portshow -i 1-5");

=item Arguments

execute_raw_command accepts all arguments accepted by Connectrec->execute().
Of these, only command is mandatory.

=over

=item command

(Mandatory)
The command to be invoked. This is the full command line as it will be entered on the switch.

=back

=item Returns

(Scalar)
Raw output

=back

=head2 get_allowed_options_for_command

See
L<NACL::APISet::_Mixins::CLI::get_allowed_options_for_command|lib-NACL-APISet-_Mixins-CLI-pm/get_allowed_options_for_command>.

=head1 AUTHOR/MAINTAINER

=over

=item NACL Development (dl-nacl-dev@netapp.com)

=back

 $Id: $

=cut
