#$Id: //depot/prod/test/nacldev/lib/NACL/APISet/Switch/CLI/Armadillo/default/info.pm#1 $
#
# Copyright (c) 2012 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#
## @summary Routines to parse Armadillo bridge commands
##
## @description
##    This is a library to parse the raw output of Armadillo bridge commands.
##
## @synopsis
##  use NACL::APISet::Switch::CLI::Armadillo::default::info
##
##  info(%args)
##
## @author sangavai@netapp.com, dl-nacl-dev@netapp.com
##
## @status public


package NACL::APISet::Switch::CLI::Armadillo::default::info;

use NACL::APISet::Response::CLI::ParserUtils;

use NACL::APISet::Exceptions::InvalidParamValueException;

use strict;

sub _check_error {
    my %args       = @_;
    my $resp       = $args{response_object};
    my $raw_output = $args{raw_output};
    my $error_message;
    my $error = undef;

    if ( $raw_output =~ /Wrong\/Missing Parameters/i ) {
        $error         = "InvalidParamValue";
        $error_message = $raw_output;
    }
    if ($error) {
        $error = 'NACL::APISet::Exceptions::' . $error . 'Exception';
        $error->throw( text => $error_message, response_object => $resp );
    }
}

## @name ::info
##
## @summary Parses the raw output from the "info" command
##
## @description 
##    This method can be used for parsing the raw output obtained after calling 
##    the method from info with any of its options as described in the 
##    corresponding cdef file 'info.cdefs'
## @synopsis
##      info(%args);
## INPUTS TO METHOD CALL 'info' ON APISET OBJECT 
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## <tr> <td> No switch for this key </td> <td> 'brief' </td> <td> '0' </td> </tr> 
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

##&nbsp;Command = info
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;Device Status         = Good
##&nbsp;Device                = "FibreBridge 6500N"
##&nbsp;Serial Number         = FB6500N101111
##&nbsp;Device Version        = 1.50
##&nbsp;Build Number          = 064Y
##&nbsp;Build Date            = "Jun  1 2012" 09:34:06
##&nbsp;Flash Revision        = 1
##&nbsp;CLI Revision          = 1.75
##&nbsp;Base version          = 51.01  
##&nbsp;Version Number        = 1.50
##&nbsp;User-defined name     = "        "
##&nbsp;World Wide Name       = 20 00 00 10 86 60 2C 62 
##&nbsp;FC1 Node Name         = 20 00 00 10 86 60 2C 62 
##&nbsp;FC1 Port Name         = 21 00 00 10 86 60 2C 62 
##&nbsp;FC1 Data Rate         = 8Gb
##&nbsp;FC1 Connection Mode   = ptp
##&nbsp;FC2 Node Name         = 20 00 00 10 86 60 2C 62 
##&nbsp;FC2 Port Name         = 22 00 00 10 86 60 2C 62 
##&nbsp;FC2 Data Rate         = N/A
##&nbsp;FC2 Connection Mode   = N/A
##&nbsp;MP1 MAC Address       = 00 10 86 60 2C 62 
##&nbsp;MP1 IP Address        = 10.227.184.101
##&nbsp;MP1 IP Subnet Mask    = 255.255.254.0
##&nbsp;MP1 IP Gateway        = 10.227.184.1
##&nbsp;MP1 IP DHCP           = disabled
##&nbsp;MP2 MAC Address       = 00 10 86 60 2C 63 
##&nbsp;MP2 IP Address        = 192.168.10.39
##&nbsp;MP2 IP Subnet Mask    = 255.255.0.0
##&nbsp;MP2 IP Gateway        = 0.0.0.0
##&nbsp;MP2 IP DHCP           = disabled
##&nbsp;Active Configuration  = ATTO
##&nbsp;
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;          {
##&nbsp;            'serial_number' => 'FB6500N101111',
##&nbsp;            'fc1_node_name' => '20 00 00 10 86 60 2C 62',
##&nbsp;            'device_version' => '1.50',
##&nbsp;            'mp1_mac_address' => '00 10 86 60 2C 62',
##&nbsp;            'mp2_mac_address' => '00 10 86 60 2C 63',
##&nbsp;            'mp1_ip_subnet_mask' => '255.255.254.0',
##&nbsp;            'fc1_connection_mode' => 'ptp',
##&nbsp;            'flash_revision' => '1',
##&nbsp;            'build_date' => 'Jun  1 2012 09:34:06',
##&nbsp;            'mp2_ip_dhcp' => 'disabled',
##&nbsp;            'mp2_ip_gateway' => '0.0.0.0',
##&nbsp;            'fc2_node_name' => '20 00 00 10 86 60 2C 62',
##&nbsp;            'user-defined_name' => '        ',
##&nbsp;            'mp2_ip_subnet_mask' => '255.255.0.0',
##&nbsp;            'mp1_ip_address' => '10.227.184.101',
##&nbsp;            'mp2_ip_address' => '192.168.10.39',
##&nbsp;            'device_status' => 'Good',
##&nbsp;            'fc2_data_rate' => 'N/A',
##&nbsp;            'active_configuration' => 'ATTO',
##&nbsp;            'version_number' => '1.50',
##&nbsp;            'world_wide_name' => '20 00 00 10 86 60 2C 62',
##&nbsp;            'fc2_connection_mode' => 'N/A',
##&nbsp;            'cli_revision' => '1.75',
##&nbsp;            'base_version' => '51.01',
##&nbsp;            'device' => 'FibreBridge 6500N',
##&nbsp;            'mp1_ip_dhcp' => 'disabled',
##&nbsp;            'fc2_port_name' => '22 00 00 10 86 60 2C 62',
##&nbsp;            'build_number' => '064Y',
##&nbsp;            'mp1_ip_gateway' => '10.227.184.1',
##&nbsp;            'fc1_data_rate' => '8Gb',
##&nbsp;            'fc1_port_name' => '21 00 00 10 86 60 2C 62'
##&nbsp;          }
##&nbsp;        ];
##&nbsp;
##&nbsp;Command = info brief
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;Device Status         = Good
##&nbsp;Device                = "FibreBridge 6500N"
##&nbsp;Build Date            = "Jun  1 2012" 09:34:06
##&nbsp;Version Number        = 1.50
##&nbsp;World Wide Name       = 20 00 00 10 86 60 2C 62 
##&nbsp;FC1 Node Name         = 20 00 00 10 86 60 2C 62 
##&nbsp;FC1 Port Name         = 21 00 00 10 86 60 2C 62 
##&nbsp;FC1 Data Rate         = 8Gb
##&nbsp;FC1 Connection Mode   = ptp
##&nbsp;FC2 Node Name         = 20 00 00 10 86 60 2C 62 
##&nbsp;FC2 Port Name         = 22 00 00 10 86 60 2C 62 
##&nbsp;FC2 Data Rate         = N/A
##&nbsp;FC2 Connection Mode   = N/A
##&nbsp;MP1 MAC Address       = 00 10 86 60 2C 62 
##&nbsp;MP1 IP Address        = 10.227.184.101
##&nbsp;MP1 IP Subnet Mask    = 255.255.254.0
##&nbsp;MP1 IP Gateway        = 10.227.184.1
##&nbsp;MP1 IP DHCP           = disabled
##&nbsp;MP2 MAC Address       = 00 10 86 60 2C 63 
##&nbsp;MP2 IP Address        = 192.168.10.39
##&nbsp;MP2 IP Subnet Mask    = 255.255.0.0
##&nbsp;MP2 IP Gateway        = 0.0.0.0
##&nbsp;MP2 IP DHCP           = disabled
##&nbsp;Active Configuration  = ATTO
##&nbsp;
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;          {
##&nbsp;            'fc1_node_name' => '20 00 00 10 86 60 2C 62',
##&nbsp;            'mp1_mac_address' => '00 10 86 60 2C 62',
##&nbsp;            'mp2_mac_address' => '00 10 86 60 2C 63',
##&nbsp;            'mp1_ip_subnet_mask' => '255.255.254.0',
##&nbsp;            'fc1_connection_mode' => 'ptp',
##&nbsp;            'build_date' => 'Jun  1 2012 09:34:06',
##&nbsp;            'mp2_ip_dhcp' => 'disabled',
##&nbsp;            'mp2_ip_gateway' => '0.0.0.0',
##&nbsp;            'fc2_node_name' => '20 00 00 10 86 60 2C 62',
##&nbsp;            'mp2_ip_subnet_mask' => '255.255.0.0',
##&nbsp;            'mp1_ip_address' => '10.227.184.101',
##&nbsp;            'mp2_ip_address' => '192.168.10.39',
##&nbsp;            'device_status' => 'Good',
##&nbsp;            'fc2_data_rate' => 'N/A',
##&nbsp;            'active_configuration' => 'ATTO',
##&nbsp;            'version_number' => '1.50',
##&nbsp;            'world_wide_name' => '20 00 00 10 86 60 2C 62',
##&nbsp;            'fc2_connection_mode' => 'N/A',
##&nbsp;            'device' => 'FibreBridge 6500N',
##&nbsp;            'mp1_ip_dhcp' => 'disabled',
##&nbsp;            'fc2_port_name' => '22 00 00 10 86 60 2C 62',
##&nbsp;            'mp1_ip_gateway' => '10.227.184.1',
##&nbsp;            'fc1_data_rate' => '8Gb',
##&nbsp;            'fc1_port_name' => '21 00 00 10 86 60 2C 62'
##&nbsp;          }
##&nbsp;        ];
##&nbsp;
sub info {
    my %args       = @_;
    my $resp       = $args{response_object};
    my $raw_output = $resp->get_raw_output();
    my @result_array;
    _check_error(
        raw_output      => $raw_output,
        response_object => $resp
    );
    my @lines = get_lines( response_object => $resp, raw_output => $raw_output);
    for(my $i = 0; $i <= $#lines; $i++) {
        if($lines[$i] =~ /\s+=\s+/) {
            my @temp = split(/\s+=\s+/, $lines[$i]);
            $temp[1] =~ s/\"//g;
            $result_array[0]->{to_key_name( item => $temp[0])} = $temp[1];
        }
    }
    return \@result_array;
}

1;
