#$Id: //depot/prod/test/nacldev/lib/NACL/APISet/Switch/CLI/Armadillo/default/routedisplay.pm#1 $
#
# Copyright (c) 2012 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#
## @summary Routines to parse Armadillo bridge commands
##
## @description
##    This is a library to parse the raw output of Armadillo bridge commands.
##
## @synopsis
##  use NACL::APISet::Switch::CLI::Armadillo::default::routedisplay
##
##  routedisplay(%args)
##
## @author sangavai@netapp.com, dl-nacl-dev@netapp.com
##
## @status public

package NACL::APISet::Switch::CLI::Armadillo::default::routedisplay;

use NACL::APISet::Response::CLI::ParserUtils;

use NACL::APISet::Exceptions::InvalidParamValueException;

use strict;

sub _check_error {
    my %args       = @_;
    my $resp       = $args{response_object};
    my $raw_output = $args{raw_output};
    my $error_message;
    my $error = undef;

    if ( $raw_output =~ /Wrong\/Missing Parameters/i ) {
        $error         = "InvalidParamValue";
        $error_message = $raw_output;
    }
    if ($error) {
        $error = 'NACL::APISet::Exceptions::' . $error . 'Exception';
        $error->throw( text => $error_message, response_object => $resp );
    }
} ## end sub _check_error

## @name ::routedisplay
##
## @summary Parses the raw output from the "Routedisplay FC" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from routedisplay with any of its options as described in the
##    corresponding cdef file 'routedisplay.cdefs'
## @synopsis
##      routedisplay(%args);
## INPUTS TO METHOD CALL 'routedisplay' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

##&nbsp;Command = Routedisplay FC
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;41
##&nbsp;; FL   Device Address   LUN VendorID ProductID        SerialNumber
##&nbsp;;=================================================================
##&nbsp;   0   Bridge
##&nbsp;   1   5000cca01227b1a9   0 NETAPP   X421_HCOBD450A10 PNGPURXZ
##&nbsp;   2   5000cca01227b3b9   0 NETAPP   X421_HCOBD450A10 PNGPUW5Z
##&nbsp;   3   5000cca0122d3745   0 NETAPP   X421_HCOBD450A10 PNGTVWAZ
##&nbsp;   4   5000cca012297f1d   0 NETAPP   X421_HCOBD450A10 PNGRUGLZ
##&nbsp;   5   5000cca01227ac99   0 NETAPP   X421_HCOBD450A10 PNGPUDGZ
##&nbsp;
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;          {
##&nbsp;            'lun' => undef,
##&nbsp;            'vendorid' => undef,
##&nbsp;            'device_address' => 'Bridge',
##&nbsp;            'fl' => '0',
##&nbsp;            'serialnumber' => undef,
##&nbsp;            'productid' => undef
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'lun' => '0',
##&nbsp;            'vendorid' => 'NETAPP',
##&nbsp;            'device_address' => '5000cca01227b1a9',
##&nbsp;            'fl' => '1',
##&nbsp;            'serialnumber' => 'PNGPURXZ',
##&nbsp;            'productid' => 'X421_HCOBD450A10'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'lun' => '0',
##&nbsp;            'vendorid' => 'NETAPP',
##&nbsp;            'device_address' => '5000cca01227b3b9',
##&nbsp;            'fl' => '2',
##&nbsp;            'serialnumber' => 'PNGPUW5Z',
##&nbsp;            'productid' => 'X421_HCOBD450A10'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'lun' => '0',
##&nbsp;            'vendorid' => 'NETAPP',
##&nbsp;            'device_address' => '5000cca0122d3745',
##&nbsp;            'fl' => '3',
##&nbsp;            'serialnumber' => 'PNGTVWAZ',
##&nbsp;            'productid' => 'X421_HCOBD450A10'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'lun' => '0',
##&nbsp;            'vendorid' => 'NETAPP',
##&nbsp;            'device_address' => '5000cca012297f1d',
##&nbsp;            'fl' => '4',
##&nbsp;            'serialnumber' => 'PNGRUGLZ',
##&nbsp;            'productid' => 'X421_HCOBD450A10'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'lun' => '0',
##&nbsp;            'vendorid' => 'NETAPP',
##&nbsp;            'device_address' => '5000cca01227ac99',
##&nbsp;            'fl' => '5',
##&nbsp;            'serialnumber' => 'PNGPUDGZ',
##&nbsp;            'productid' => 'X421_HCOBD450A10'
##&nbsp;          },
##&nbsp;        ];
##&nbsp;
sub routedisplay_fc {
    my %args       = @_;
    my $resp       = $args{response_object};
    my $raw_output = $resp->get_raw_output();
    my ( @result_array, @head, @values, $count );
    _check_error(
        raw_output      => $raw_output,
        response_object => $resp
    );
    my @lines
        = get_lines( response_object => $resp, raw_output => $raw_output );
    for ( my $i = 1; $i <= $#lines; $i++ ) {
        if ( $lines[$i] =~ /FL\s+Device Address/ ) {
            $lines[$i] =~ s/Device Address/device_address/;
            $lines[$i] =~ s/\;\s+//;
            @head = split( /\s+/, $lines[$i] );
        }
        if ( $lines[$i] =~ /\d+\s+\w+/ ) {
            @values = split( /\s+/, $lines[$i] );
            for ( my $j = 0; $j <= $#head; $j++ ) {
                $result_array[$count]->{ to_key_name( item => $head[$j] ) }
                    = $values[$j];
            }
            $count++;
        }
    }
    return \@result_array;
} ## end sub routedisplay_fc

1;
