# Copyright (c) 2012 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#
## @summary Routines to parse Armadillo bridge commands
##
## @description
##    This is a library to parse the raw output of Armadillo bridge commands.
##
## @synopsis
##  use NACL::APISet::Switch::CLI::Armadillo::default::sastargets
##
##  sastargets(%args)
## @author dl-nacl-dev@netapp.com
##
## @status public
#
package NACL::APISet::Switch::CLI::Armadillo::default::sastargets;

use NACL::APISet::Response::CLI::ParserUtils;

use NACL::APISet::Exceptions::InvalidParamValueException;
use NACL::APISet::Exceptions::UnexpectedOutputException;
use strict;
use warnings;

sub _check_error {
    my %args       = @_;
    my $resp       = $args{response_object};
    my $raw_output = $args{raw_output};
    my $error_message;
    my $error = undef;
    
    if ( $raw_output =~ /Wrong\/Missing Parameters/i ) {
        $error         = "InvalidParamValue";
        $error_message = $raw_output;
    }
    if ($error) {
        $error = 'NACL::APISet::Exceptions::' . $error . 'Exception';
        $error->throw( text => $error_message, response_object => $resp );
    }

    return $error;
}

## @name ::sastargets
##
## @summary Parses the raw output from the "sastargets" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from sastargets with any of its options as described in the
##    corresponding cdef file 'sastargets.cdefs'
## @synopsis
##      sastargets(%args);
## INPUTS TO METHOD CALL 'sastargets' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

##&nbsp;Command = sastargets
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;51
##&nbsp;; Tgt VendorID ProductID        Type          SerialNumber
##&nbsp;    0 NETAPP   X410_S15K7288A15 Disk          3SJ13WSE000090406HZ6
##&nbsp;    1 NETAPP   X410_S15K7288A15 Disk          3SJ13VEV00009040S609
##&nbsp;    2 NETAPP   X410_S15K7288A15 Disk          3SJ14HNL0000904074NP
##&nbsp;    3 NETAPP   X410_S15K7288A15 Disk          3SJ14HHX00009040UZET
##&nbsp;    4 NETAPP   X410_S15K7288A15 Disk          3SJ14HTJ0000904073FS
##&nbsp;    5 NETAPP   X410_S15K7288A15 Disk          3SJ14HTX0000904074QG
##&nbsp;    6 NETAPP   X410_S15K7288A15 Disk          3SJ14HVZ0000904073C4
##&nbsp;    7 NETAPP   X410_S15K7288A15 Disk          3SJ14GRY00009041WT13
##&nbsp;    8 NETAPP   X410_S15K7288A15 Disk          3SJ14GS400009041WT02
##&nbsp;    9 NETAPP   X410_S15K7288A15 Disk          3SJ14GVD00009041WS7B
##&nbsp;   10 NETAPP   X410_S15K7288A15 Disk          3SJ14GP900009041WP5W
##&nbsp;   11 NETAPP   X410_S15K7288A15 Disk          3SJ14GMX00009041WTJG
##&nbsp;   12 NETAPP   X410_S15K7288A15 Disk          3SJ14GMK00009041WTGU
##&nbsp;   13 NETAPP   X410_S15K7288A15 Disk          3SJ12JJD00009040T7VK
##&nbsp;   14 NETAPP   X410_S15K7288A15 Disk          3SJ13ZN70000904067CZ
##&nbsp;   15 NETAPP   X410_S15K7288A15 Disk          3SJ13KEV000090405EGL
##&nbsp;   16 NETAPP   X410_S15K7288A15 Disk          3SJ13HSF000090193WP3
##&nbsp;   17 NETAPP   X410_S15K7288A15 Disk          3SJ13HSM00009041WPMF
##&nbsp;   18 NETAPP   X410_S15K7288A15 Disk          3SJ14F0C000090407TQQ
##&nbsp;   19 NETAPP   X410_S15K7288A15 Disk          3SJ14F1H00009040W1R4
##&nbsp;   20 NETAPP   X410_S15K7288A15 Disk          3SJ141VV000090407S39
##&nbsp;   21 NETAPP   X410_S15K7288A15 Disk          3SJ14F8C0000904073ST
##&nbsp;   22 NETAPP   X410_S15K7288A15 Disk          3SJ13HB2000090407PBG
##&nbsp;   23 NETAPP   X410_S15K7288A15 Disk          3SJ13G8P000090407SX7
##&nbsp;   24 NETAPP   DS424IOM3        SES           SHU0946862G0395
##&nbsp;   25 NETAPP   DS424IOM3        SES           SHU0946862G0DTK
##&nbsp;   26 SEAGATE  ST31000340NS  SM Disk          9QJ7V5WT
##&nbsp;   27 SEAGATE  ST31000340NS  SM Disk          9QJ7T8DM
##&nbsp;   28 SEAGATE  ST31000340NS  SM Disk          9QJ7V5CX
##&nbsp;   29 SEAGATE  ST31000340NS  SM Disk          9QJ7T9EP
##&nbsp;   30 SEAGATE  ST31000340NS  SM Disk          9QJ7V6M5
##&nbsp;   31 SEAGATE  ST31000340NS  SM Disk          9QJ7TR8S
##&nbsp;   32 SEAGATE  ST31000340NS  SM Disk          9QJ7TR6P
##&nbsp;   33 SEAGATE  ST31000340NS  SM Disk          9QJ7TQBB
##&nbsp;   34 SEAGATE  ST31000340NS  SM Disk          9QJ7PYE2
##&nbsp;   35 SEAGATE  ST31000340NS  SM Disk          9QJ7TSYR
##&nbsp;   36 SEAGATE  ST31000340NS  SM Disk          9QJ7TTY2
##&nbsp;   37 SEAGATE  ST31000340NS  SM Disk          9QJ7PYHS
##&nbsp;   38 SEAGATE  ST31000340NS  SM Disk          9QJ7TQ8X
##&nbsp;   39 SEAGATE  ST31000340NS  SM Disk          9QJ7TQMS
##&nbsp;   40 SEAGATE  ST31000340NS  SM Disk          9QJ7TS53
##&nbsp;   41 SEAGATE  ST31000340NS  SM Disk          9QJ7TQLF
##&nbsp;   42 SEAGATE  ST31000340NS  SM Disk          9QJ7S155
##&nbsp;   43 SEAGATE  ST31000340NS  SM Disk          9QJ7TQ38
##&nbsp;   44 HITACHI  HDE721010SLA33SM Disk          MQ0KBX9F
##&nbsp;   45 WDC      WD1002FBYS-05ASM Disk          WD-WMATV8162416
##&nbsp;   46 WDC      WD1002FBYS-05ASM Disk          WD-WMATV8119551
##&nbsp;   47 WDC      WD1002FBYS-05ASM Disk          WD-WMATV8757414
##&nbsp;   48 WDC      WD1002FBYS-05ASM Disk          WD-WMATV8747789
##&nbsp;   49 WDC      WD1002FBYS-05ASM Disk          WD-WMATV8721577
##&nbsp;
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;          {
##&nbsp;            'serial_number' => '3SJ13WSE000090406HZ6',
##&nbsp;            'product_id' => 'X410_S15K7288A15',
##&nbsp;            'tgt' => '0',
##&nbsp;            'vendor_id' => 'NETAPP',
##&nbsp;            'type' => 'Disk'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'serial_number' => '3SJ13VEV00009040S609',
##&nbsp;            'product_id' => 'X410_S15K7288A15',
##&nbsp;            'tgt' => '1',
##&nbsp;            'vendor_id' => 'NETAPP',
##&nbsp;            'type' => 'Disk'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'serial_number' => '3SJ14HNL0000904074NP',
##&nbsp;            'product_id' => 'X410_S15K7288A15',
##&nbsp;            'tgt' => '2',
##&nbsp;            'vendor_id' => 'NETAPP',
##&nbsp;            'type' => 'Disk'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'serial_number' => '3SJ14HHX00009040UZET',
##&nbsp;            'product_id' => 'X410_S15K7288A15',
##&nbsp;            'tgt' => '3',
##&nbsp;            'vendor_id' => 'NETAPP',
##&nbsp;            'type' => 'Disk'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'serial_number' => '3SJ14HTJ0000904073FS',
##&nbsp;            'product_id' => 'X410_S15K7288A15',
##&nbsp;            'tgt' => '4',
##&nbsp;            'vendor_id' => 'NETAPP',
##&nbsp;            'type' => 'Disk'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'serial_number' => '3SJ14HTX0000904074QG',
##&nbsp;            'product_id' => 'X410_S15K7288A15',
##&nbsp;            'tgt' => '5',
##&nbsp;            'vendor_id' => 'NETAPP',
##&nbsp;            'type' => 'Disk'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'serial_number' => '3SJ14HVZ0000904073C4',
##&nbsp;            'product_id' => 'X410_S15K7288A15',
##&nbsp;            'tgt' => '6',
##&nbsp;            'vendor_id' => 'NETAPP',
##&nbsp;            'type' => 'Disk'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'serial_number' => '3SJ14GRY00009041WT13',
##&nbsp;            'product_id' => 'X410_S15K7288A15',
##&nbsp;            'tgt' => '7',
##&nbsp;            'vendor_id' => 'NETAPP',
##&nbsp;            'type' => 'Disk'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'serial_number' => '3SJ14GS400009041WT02',
##&nbsp;            'product_id' => 'X410_S15K7288A15',
##&nbsp;            'tgt' => '8',
##&nbsp;            'vendor_id' => 'NETAPP',
##&nbsp;            'type' => 'Disk'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'serial_number' => '3SJ14GVD00009041WS7B',
##&nbsp;            'product_id' => 'X410_S15K7288A15',
##&nbsp;            'tgt' => '9',
##&nbsp;            'vendor_id' => 'NETAPP',
##&nbsp;            'type' => 'Disk'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'serial_number' => '3SJ14GP900009041WP5W',
##&nbsp;            'product_id' => 'X410_S15K7288A15',
##&nbsp;            'tgt' => '10',
##&nbsp;            'vendor_id' => 'NETAPP',
##&nbsp;            'type' => 'Disk'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'serial_number' => '3SJ14GMX00009041WTJG',
##&nbsp;            'product_id' => 'X410_S15K7288A15',
##&nbsp;            'tgt' => '11',
##&nbsp;            'vendor_id' => 'NETAPP',
##&nbsp;            'type' => 'Disk'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'serial_number' => '3SJ14GMK00009041WTGU',
##&nbsp;            'product_id' => 'X410_S15K7288A15',
##&nbsp;            'tgt' => '12',
##&nbsp;            'vendor_id' => 'NETAPP',
##&nbsp;            'type' => 'Disk'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'serial_number' => '3SJ12JJD00009040T7VK',
##&nbsp;            'product_id' => 'X410_S15K7288A15',
##&nbsp;            'tgt' => '13',
##&nbsp;            'vendor_id' => 'NETAPP',
##&nbsp;            'type' => 'Disk'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'serial_number' => '3SJ13ZN70000904067CZ',
##&nbsp;            'product_id' => 'X410_S15K7288A15',
##&nbsp;            'tgt' => '14',
##&nbsp;            'vendor_id' => 'NETAPP',
##&nbsp;            'type' => 'Disk'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'serial_number' => '3SJ13KEV000090405EGL',
##&nbsp;            'product_id' => 'X410_S15K7288A15',
##&nbsp;            'tgt' => '15',
##&nbsp;            'vendor_id' => 'NETAPP',
##&nbsp;            'type' => 'Disk'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'serial_number' => '3SJ13HSF000090193WP3',
##&nbsp;            'product_id' => 'X410_S15K7288A15',
##&nbsp;            'tgt' => '16',
##&nbsp;            'vendor_id' => 'NETAPP',
##&nbsp;            'type' => 'Disk'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'serial_number' => '3SJ13HSM00009041WPMF',
##&nbsp;            'product_id' => 'X410_S15K7288A15',
##&nbsp;            'tgt' => '17',
##&nbsp;            'vendor_id' => 'NETAPP',
##&nbsp;            'type' => 'Disk'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'serial_number' => '3SJ14F0C000090407TQQ',
##&nbsp;            'product_id' => 'X410_S15K7288A15',
##&nbsp;            'tgt' => '18',
##&nbsp;            'vendor_id' => 'NETAPP',
##&nbsp;            'type' => 'Disk'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'serial_number' => '3SJ14F1H00009040W1R4',
##&nbsp;            'product_id' => 'X410_S15K7288A15',
##&nbsp;            'tgt' => '19',
##&nbsp;            'vendor_id' => 'NETAPP',
##&nbsp;            'type' => 'Disk'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'serial_number' => '3SJ141VV000090407S39',
##&nbsp;            'product_id' => 'X410_S15K7288A15',
##&nbsp;            'tgt' => '20',
##&nbsp;            'vendor_id' => 'NETAPP',
##&nbsp;            'type' => 'Disk'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'serial_number' => '3SJ14F8C0000904073ST',
##&nbsp;            'product_id' => 'X410_S15K7288A15',
##&nbsp;            'tgt' => '21',
##&nbsp;            'vendor_id' => 'NETAPP',
##&nbsp;            'type' => 'Disk'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'serial_number' => '3SJ13HB2000090407PBG',
##&nbsp;            'product_id' => 'X410_S15K7288A15',
##&nbsp;            'tgt' => '22',
##&nbsp;            'vendor_id' => 'NETAPP',
##&nbsp;            'type' => 'Disk'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'serial_number' => '3SJ13G8P000090407SX7',
##&nbsp;            'product_id' => 'X410_S15K7288A15',
##&nbsp;            'tgt' => '23',
##&nbsp;            'vendor_id' => 'NETAPP',
##&nbsp;            'type' => 'Disk'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'serial_number' => 'SHU0946862G0395',
##&nbsp;            'product_id' => 'DS424IOM3       ',
##&nbsp;            'tgt' => '24',
##&nbsp;            'vendor_id' => 'NETAPP',
##&nbsp;            'type' => 'SES'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'serial_number' => 'SHU0946862G0DTK',
##&nbsp;            'product_id' => 'DS424IOM3       ',
##&nbsp;            'tgt' => '25',
##&nbsp;            'vendor_id' => 'NETAPP',
##&nbsp;            'type' => 'SES'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'serial_number' => '9QJ7V5WT',
##&nbsp;            'product_id' => 'ST31000340NS  SM',
##&nbsp;            'tgt' => '26',
##&nbsp;            'vendor_id' => 'SEAGATE',
##&nbsp;            'type' => 'Disk'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'serial_number' => '9QJ7T8DM',
##&nbsp;            'product_id' => 'ST31000340NS  SM',
##&nbsp;            'tgt' => '27',
##&nbsp;            'vendor_id' => 'SEAGATE',
##&nbsp;            'type' => 'Disk'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'serial_number' => '9QJ7V5CX',
##&nbsp;            'product_id' => 'ST31000340NS  SM',
##&nbsp;            'tgt' => '28',
##&nbsp;            'vendor_id' => 'SEAGATE',
##&nbsp;            'type' => 'Disk'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'serial_number' => '9QJ7T9EP',
##&nbsp;            'product_id' => 'ST31000340NS  SM',
##&nbsp;            'tgt' => '29',
##&nbsp;            'vendor_id' => 'SEAGATE',
##&nbsp;            'type' => 'Disk'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'serial_number' => '9QJ7V6M5',
##&nbsp;            'product_id' => 'ST31000340NS  SM',
##&nbsp;            'tgt' => '30',
##&nbsp;            'vendor_id' => 'SEAGATE',
##&nbsp;            'type' => 'Disk'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'serial_number' => '9QJ7TR8S',
##&nbsp;            'product_id' => 'ST31000340NS  SM',
##&nbsp;            'tgt' => '31',
##&nbsp;            'vendor_id' => 'SEAGATE',
##&nbsp;            'type' => 'Disk'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'serial_number' => '9QJ7TR6P',
##&nbsp;            'product_id' => 'ST31000340NS  SM',
##&nbsp;            'tgt' => '32',
##&nbsp;            'vendor_id' => 'SEAGATE',
##&nbsp;            'type' => 'Disk'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'serial_number' => '9QJ7TQBB',
##&nbsp;            'product_id' => 'ST31000340NS  SM',
##&nbsp;            'tgt' => '33',
##&nbsp;            'vendor_id' => 'SEAGATE',
##&nbsp;            'type' => 'Disk'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'serial_number' => '9QJ7PYE2',
##&nbsp;            'product_id' => 'ST31000340NS  SM',
##&nbsp;            'tgt' => '34',
##&nbsp;            'vendor_id' => 'SEAGATE',
##&nbsp;            'type' => 'Disk'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'serial_number' => '9QJ7TSYR',
##&nbsp;            'product_id' => 'ST31000340NS  SM',
##&nbsp;            'tgt' => '35',
##&nbsp;            'vendor_id' => 'SEAGATE',
##&nbsp;            'type' => 'Disk'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'serial_number' => '9QJ7TTY2',
##&nbsp;            'product_id' => 'ST31000340NS  SM',
##&nbsp;            'tgt' => '36',
##&nbsp;            'vendor_id' => 'SEAGATE',
##&nbsp;            'type' => 'Disk'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'serial_number' => '9QJ7PYHS',
##&nbsp;            'product_id' => 'ST31000340NS  SM',
##&nbsp;            'tgt' => '37',
##&nbsp;            'vendor_id' => 'SEAGATE',
##&nbsp;            'type' => 'Disk'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'serial_number' => '9QJ7TQ8X',
##&nbsp;            'product_id' => 'ST31000340NS  SM',
##&nbsp;            'tgt' => '38',
##&nbsp;            'vendor_id' => 'SEAGATE',
##&nbsp;            'type' => 'Disk'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'serial_number' => '9QJ7TQMS',
##&nbsp;            'product_id' => 'ST31000340NS  SM',
##&nbsp;            'tgt' => '39',
##&nbsp;            'vendor_id' => 'SEAGATE',
##&nbsp;            'type' => 'Disk'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'serial_number' => '9QJ7TS53',
##&nbsp;            'product_id' => 'ST31000340NS  SM',
##&nbsp;            'tgt' => '40',
##&nbsp;            'vendor_id' => 'SEAGATE',
##&nbsp;            'type' => 'Disk'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'serial_number' => '9QJ7TQLF',
##&nbsp;            'product_id' => 'ST31000340NS  SM',
##&nbsp;            'tgt' => '41',
##&nbsp;            'vendor_id' => 'SEAGATE',
##&nbsp;            'type' => 'Disk'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'serial_number' => '9QJ7S155',
##&nbsp;            'product_id' => 'ST31000340NS  SM',
##&nbsp;            'tgt' => '42',
##&nbsp;            'vendor_id' => 'SEAGATE',
##&nbsp;            'type' => 'Disk'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'serial_number' => '9QJ7TQ38',
##&nbsp;            'product_id' => 'ST31000340NS  SM',
##&nbsp;            'tgt' => '43',
##&nbsp;            'vendor_id' => 'SEAGATE',
##&nbsp;            'type' => 'Disk'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'serial_number' => 'MQ0KBX9F',
##&nbsp;            'product_id' => 'HDE721010SLA33SM',
##&nbsp;            'tgt' => '44',
##&nbsp;            'vendor_id' => 'HITACHI',
##&nbsp;            'type' => 'Disk'
##&nbsp;          }
##&nbsp;        ];
##&nbsp;
sub sastargets {
    my %args = @_;
    my @result_array;
    my $resp       = $args{response_object};
    my $output     = $resp->get_processed_output();
    my $command    = $resp->get_actual_command();
    my $raw_output = $resp->get_raw_output();
    my $error      = _check_error(
        raw_output      => $raw_output,
        response_object => $resp
    );
    my %hash;

    if ( $output =~ /^(\d)/ ) {
        my $first_line = $1;
        my @arr_lines  = get_lines(
            "raw_output"      => $raw_output,
            "first_line"      => $first_line,
            "can_throw"       => 1,
            "response_object" => $resp
        );
        my @header;
        foreach my $lines (@arr_lines) {
            my %hash = ();
            next if ( $lines =~ /^;/ );
            if ( $lines =~ /(\d+)\s+(\S+)\s+(\w+\s*\w*)\s+(\S+)\s+(\S+)/ ) {
                $hash{'tgt'}           = $1;
                $hash{'vendor_id'}     = $2;
                $hash{'product_id'}    = $3;
                $hash{'type'}          = $4;
                $hash{'serial_number'} = $5;

            }
            push( @result_array, \%hash ) if %hash;
        }
    } else {
        NACL::APISet::Exceptions::UnexpectedOutputException->throw(
            'text'            => "Output was not in the expected format",
            'response_object' => $resp
        );

    }
    return \@result_array;
}

1;
