# Copyright (c) 2012 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#
## @summary Routines to parse Armadillo bridge commands
##
## @description
##    This is a library to parse the raw output of Armadillo bridge commands.
##
## @synopsis
##  use NACL::APISet::Switch::CLI::Armadillo::default::set
##
##  set_bridgename (%args)
##  set_date(%args)
##  set_fcconnmode(%args)
##  set_fcdatarate (%args)
##  set_ipaddress(%args)
##  set_ipgateway (%args)
##  set_ipsubnetmask(%args)
##  set_sastargetport(%args)
##
## @author dl-nacl-dev@netapp.com
## @change 2012-06-27:
##  Added apiset for set sastargetport device enabled/disabled
##
## @status public
#
package NACL::APISet::Switch::CLI::Armadillo::default::set;

use NACL::APISet::Response::CLI::ParserUtils;

use NACL::APISet::Exceptions::InvalidParamValueException;

use strict;

sub _check_error {
    my %args       = @_;
    my $resp       = $args{response_object};
    my $raw_output = $args{raw_output};
    my $error_message;
    my $error = undef;

    if ( $raw_output =~ /Wrong\/Missing Parameters|Invalid/i ) {
        $error         = "InvalidParamValue";
        $error_message = $raw_output;
    }
    if ($error) {
        $error = 'NACL::APISet::Exceptions::' . $error . 'Exception';
        $error->throw( text => $error_message, response_object => $resp );
    }

}

## @name ::set_bridgename
##
## @summary Parses the raw output from the "set bridgename" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from set_bridgename with any of its options as described in the
##    corresponding cdef file 'set.cdefs'
## @synopsis
##      set_bridgename(%args);
## INPUTS TO METHOD CALL 'set_bridgename' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## <tr> <td> No switch for this key </td> <td> 'name' </td> <td> '0' </td> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

##&nbsp;Command = set bridgename fcs01
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp; *
##&nbsp;set bridgename fcs01
##&nbsp;
##&nbsp;
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [];
##&nbsp;
sub set_bridgename {
    my %args    = @_;
    my $resp    = $args{response_object};
    my $output  = $resp->get_processed_output();
    my $command = $resp->get_actual_command();
    my $error   = _check_error(
        raw_output      => $output,
        response_object => $resp
    );
    return [];
}

## @name ::set_date
##
## @summary Parses the raw output from the "set date" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from set_date with any of its options as described in the
##    corresponding cdef file 'set.cdefs'
## @synopsis
##      set_date(%args);
## INPUTS TO METHOD CALL 'set_date' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## <tr> <td> No switch for this key </td> <td> 'date' </td> <td> '0' </td> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

##&nbsp;Command = set date 02/07/2012
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp; *
##&nbsp;set date 02/07/2012
##&nbsp;
##&nbsp;
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [];
##&nbsp;
sub set_date {
    my %args    = @_;
    my $resp    = $args{response_object};
    my $output  = $resp->get_processed_output();
    my $command = $resp->get_actual_command();
    my $error   = _check_error(
        raw_output      => $output,
        response_object => $resp
    );
    return [];
}

## @name ::set_ipaddress
##
## @summary Parses the raw output from the "set ipaddress" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from set_ipaddress with any of its options as described in the
##    corresponding cdef file 'set.cdefs'
## @synopsis
##      set_ipaddress(%args);
## INPUTS TO METHOD CALL 'set_ipaddress' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## <tr> <td> No switch for this key </td> <td> 'mp' </td> <td> '0' </td> </tr>
## <tr> <td> No switch for this key </td> <td> 'ipaddress' </td> <td> '0' </td> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

##&nbsp;Command = set ipaddress mp1 10.228.213.94
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [];
##&nbsp;
sub set_ipaddress {
    my %args    = @_;
    my $resp    = $args{response_object};
    my $output  = $resp->get_processed_output();
    my $command = $resp->get_actual_command();
    my $error   = _check_error(
        raw_output      => $output,
        response_object => $resp
    );
    return [];
}

## @name ::set_ipgateway
##
## @summary Parses the raw output from the "set ipgateway" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from set_ipgateway with any of its options as described in the
##    corresponding cdef file 'set.cdefs'
## @synopsis
##      set_ipgateway(%args);
## INPUTS TO METHOD CALL 'set_ipgateway' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## <tr> <td> No switch for this key </td> <td> 'ipgateway' </td> <td> '0' </td> </tr>
## <tr> <td> No switch for this key </td> <td> 'mp' </td> <td> '0' </td> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

##&nbsp;Command = set ipgateway mp1 10.228.213.1
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp; *
##&nbsp;set ipgateway mp1 10.228.213.1
##&nbsp;
##&nbsp;
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [];
##&nbsp;
sub set_ipgateway {
    my %args    = @_;
    my $resp    = $args{response_object};
    my $output  = $resp->get_processed_output();
    my $command = $resp->get_actual_command();
    my $error   = _check_error(
        raw_output      => $output,
        response_object => $resp
    );
    return [];
}

## @name ::set_ipsubnetmask
##
## @summary Parses the raw output from the "set ipsubnetmask" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from set_ipsubnetmask with any of its options as described in the
##    corresponding cdef file 'set.cdefs'
## @synopsis
##      set_ipsubnetmask(%args);
## INPUTS TO METHOD CALL 'set_ipsubnetmask' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## <tr> <td> No switch for this key </td> <td> 'ipsubnetmask' </td> <td> '0' </td> </tr>
## <tr> <td> No switch for this key </td> <td> 'mp' </td> <td> '0' </td> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

##&nbsp;Command = set ipsubnetmask mp1 255.255.255.0
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp; *
##&nbsp;set ipsubnetmask mp1 255.255.255.0
##&nbsp;
##&nbsp;
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [];
##&nbsp;
sub set_ipsubnetmask {
    my %args    = @_;
    my $resp    = $args{response_object};
    my $output  = $resp->get_processed_output();
    my $command = $resp->get_actual_command();
    my $error   = _check_error(
        raw_output      => $output,
        response_object => $resp
    );
    return [];
}

## @name ::set_fcconnmode
##
## @summary Parses the raw output from the "set fcconnmode" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from set_fcconnmode with any of its options as described in the
##    corresponding cdef file 'set.cdefs'
## @synopsis
##      set_fcconnmode(%args);
## INPUTS TO METHOD CALL 'set_fcconnmode' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## <tr> <td> No switch for this key </td> <td> 'connection' </td> <td> '0' </td> </tr>
## <tr> <td> 'fp' </td> <td> 'fp' </td> <td> '1' </td> </tr>
## <tr> <td> 'all' </td> <td> 'all' </td> <td> '1' </td> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

##&nbsp;Command = set fcconnmode all ptp
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp; *
##&nbsp;set fcconnmode all ptp
##&nbsp;
##&nbsp;
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [];
##&nbsp;
sub set_fcconnmode {
    my %args       = @_;
    my $resp       = $args{response_object};
    my $output     = $resp->get_processed_output();
    my $command    = $resp->get_actual_command();
    my $raw_output = $resp->get_raw_output();
    my $error      = _check_error(
        raw_output      => $raw_output,
        response_object => $resp
    );
    return [];
}
## @name ::set_fcdatarate
##
## @summary Parses the raw output from the "set fcdatarate" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from set_fcdatarate with any of its options as described in the
##    corresponding cdef file 'set.cdefs'
## @synopsis
##      set_fcdatarate(%args);
## INPUTS TO METHOD CALL 'set_fcdatarate' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## <tr> <td> No switch for this key </td> <td> 'rate' </td> <td> '0' </td> </tr>
## <tr> <td> 'fp' </td> <td> 'fp' </td> <td> '1' </td> </tr>
## <tr> <td> 'all' </td> <td> 'all' </td> <td> '1' </td> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

##&nbsp;Command = set fcdatarate all 8gb
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp; *
##&nbsp;set fcdatarate all 8gb
##&nbsp;
##&nbsp;
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [];
##&nbsp;
sub set_fcdatarate {
    my %args       = @_;
    my $resp       = $args{response_object};
    my $output     = $resp->get_processed_output();
    my $command    = $resp->get_actual_command();
    my $raw_output = $resp->get_raw_output();
    my $error      = _check_error(
        raw_output      => $raw_output,
        response_object => $resp
    );
    return [];
}

## @name ::set_sastargetport
##
## @summary Parses the raw output from the "set SASTargetPort" command
##
## @description 
##    This method can be used for parsing the raw output obtained after calling 
##    the method from set_sastargetport with any of its options as described in the 
##    corresponding cdef file 'set.cdefs'
## @synopsis
##      set_sastargetport(%args);
## INPUTS TO METHOD CALL 'set_sastargetport' ON APISET OBJECT 
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## <tr> <td> No switch for this key </td> <td> 'device' </td> <td> '0' </td> </tr> 
## <tr> <td> No switch for this key </td> <td> 'state' </td> <td> '0' </td> </tr> 
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

##&nbsp;Command = set SASTargetPort A enabled
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [];
##&nbsp;
sub set_sastargetport {
    my %args       = @_;
    my $resp       = $args{response_object};
    my $raw_output = $resp->get_raw_output();
    _check_error(
        raw_output      => $raw_output,
        response_object => $resp
    );
    return [];
}
1;
