#$Id: //depot/prod/test/nacldev/lib/NACL/APISet/Switch/CLI/BrocadeFC/default/cfg.pm#1 $
#
#Copyright (c) 2012 Network Appliance, Inc.
#All rights reserved.
#
## @summary Routines to parse BrocadeFC cfg commands
##
## @description
##    This is a library to parse the raw output of BrocadeFC switch cfg commands.
##
## @synopsis
##  use NACL::APISet::Switch::CLI::BrocadeFC::default::cfg
##
##  cfgadd(%args)
##  cfgcreate(%args)
##  cfgdelete(%args)
##  cfgdisable(%args)
##  cfgremove(%args)
##
## @author udaykk@netapp.com dl-nacl-dev@netapp.com
##
## @status public
#
package NACL::APISet::Switch::CLI::BrocadeFC::default::cfg;

use strict;
use NACL::APISet::Response::CLI::ParserUtils;

#Include exceptions which will be thrown
use NACL::APISet::Exceptions::InvalidParamException;

sub _check_error {
    my %args = @_;
    my $raw_output    = $args{raw_output};
    my $command       = $args{command};
    my $error_message = $args{error_message};
    my $error;

    # Check for the error
    if ( $raw_output =~ /Usage:(.*)/ ) {
        $$error_message = $1;
        $error          = "InvalidParam";
    }
    return $error;
} ## end sub _check_error

## @name ::cfgadd
##
## @summary Parses the raw output from the "cfgadd" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from cfgadd with any of its options as described in the
##    corresponding cdef file 'cfgadd.cdefs'
## @synopsis
##      cfgadd(%args);
## INPUTS TO METHOD CALL 'cfgadd' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> </tr>
## <tr> <td> No switch for this key </td> <td> 'arg2' </td> </tr>
## <tr> <td> No switch for this key </td> <td> 'arg1' </td> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

##&nbsp;Command =
##&nbsp;$VAR1 = [];
##&nbsp;
sub cfgadd {
    my %args = @_;
    my $error_message;
    my $resp        = $args{response_object};
    my $output      = $resp->get_processed_output();
    my $command     = $resp->get_command();
    my $error_level = $resp->get_execution_status();
    my $error;

    # Status returned by switch for this command is
    # is ZERO even if command failed
    $error = _check_error(
        raw_output    => $output,
        command       => "cfgadd",
        error_message => \$error_message
    );
    if ($error) {
        $error = 'NACL::APISet::Exceptions::' . $error . 'Exception';
        $error->throw( text => $error_message, response_object => $resp );
    }

    return [];
} ## end sub cfgadd

## @name ::cfgcreate
##
## @summary Parses the raw output from the "cfgcreate" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from cfgcreate with any of its options as described in the
##    corresponding cdef file 'cfgcreate.cdefs'
## @synopsis
##      cfgcreate(%args);
## INPUTS TO METHOD CALL 'cfgcreate' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> </tr>
## <tr> <td> No switch for this key </td> <td> 'arg2' </td> </tr>
## <tr> <td> No switch for this key </td> <td> 'arg1' </td> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @example

##&nbsp;Command = cfgcreate "d1", "d2"
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [];
##&nbsp;
sub cfgcreate {
    my %args = @_;

    my $error_message;
    my $resp        = $args{response_object};
    my $output      = $resp->get_processed_output();
    my $command     = $resp->get_command();
    my $error_level = $resp->get_execution_status();
    my $error;

    # Status returned by switch for this command is
    # is ZERO even if command failed
    $error = _check_error(
        raw_output    => $output,
        command       => "cfgcreate",
        error_message => \$error_message
    );
    if ($error) {
        $error = 'NACL::APISet::Exceptions::' . $error . 'Exception';
        $error->throw( text => $error_message, response_object => $resp );
    }

    return [];
} ## end sub cfgcreate

## @name ::cfgdelete
##
## @summary Parses the raw output from the "cfgdelete" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from cfgdelete with any of its options as described in the
##    corresponding cdef file 'cfgdelete.cdefs'
## @synopsis
##      cfgdelete(%args);
## INPUTS TO METHOD CALL 'cfgdelete' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> </tr>
## <tr> <td> No switch for this key </td> <td> 'arg1' </td> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

##&nbsp;Command = cfgdelete
##&nbsp;$VAR1 = [];
##&nbsp;
sub cfgdelete {
    my %args = @_;

    my $error_message;
    my $resp        = $args{response_object};
    my $output      = $resp->get_processed_output();
    my $command     = $resp->get_command();
    my $error_level = $resp->get_execution_status();
    my $error;

    $error = _check_error(
        raw_output    => $output,
        command       => "cfgdelete",
        error_message => \$error_message
    );
    if ($error) {
        $error = 'NACL::APISet::Exceptions::' . $error . 'Exception';
        $error->throw( text => $error_message, response_object => $resp );
    }

    return [];
} ## end sub cfgdelete

## @name ::cfgdisable
##
## @summary Parses the raw output from the "cfgdisable" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from cfgdisable with any of its options as described in the
##    corresponding cdef file 'cfgdisable.cdefs'
## @synopsis
##      cfgdisable(%args);
## INPUTS TO METHOD CALL 'cfgdisable' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

##&nbsp;Command = cfgdisable
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;      {
##&nbsp;        'status' => 'No effective configuration found.'
##&nbsp;      }
##&nbsp; ];
##&nbsp;
sub cfgdisable {
    my %args = @_;

    my ( $error_message, @result_array );
    my $resp        = $args{response_object};
    my $output      = $resp->get_processed_output();
    my $command     = $resp->get_command();
    my $error_level = $resp->get_execution_status();
    my $error;

    # Status returned by switch for this command is
    # is ZERO even if command failed
    $error = _check_error(
        raw_output    => $output,
        command       => "cfgdisable",
        error_message => \$error_message
    );
    if ($error) {
        $error = 'NACL::APISet::Exceptions::' . $error . 'Exception';
        $error->throw( text => $error_message, response_object => $resp );
    }

    my %hash;

    #To match with the multiple patterns like 'Openration cancelled,
    #Nothing changed, Updating flash'.
    if ( $output
        =~ /((No effective configuration found.*)|(Updating flash.*))/i
        ) {
        $hash{'status'} = $1;
    }

    push( @result_array, \%hash );
    return \@result_array;
} ## end sub cfgdisable

## @name ::cfgremove
##
## @summary Parses the raw output from the "cfgremove" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from cfgremove with any of its options as described in the
##    corresponding cdef file 'cfgremove.cdefs'
## @synopsis
##      cfgremove(%args);
## INPUTS TO METHOD CALL 'cfgremove' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> </tr>
## <tr> <td> No switch for this key </td> <td> 'arg2' </td> </tr>
## <tr> <td> No switch for this key </td> <td> 'arg1' </td> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

##&nbsp;Command = cfgremove "d1", "d2"
##&nbsp;$VAR1 = [];
##&nbsp;
sub cfgremove {
    my %args = @_;

    my $error_message;
    my $resp        = $args{response_object};
    my $output      = $resp->get_processed_output();
    my $command     = $resp->get_command();
    my $error_level = $resp->get_execution_status();
    my $error;

    $error = _check_error(
        raw_output    => $output,
        command       => "cfgremove",
        error_message => \$error_message
    );
    if ($error) {
        $error = 'NACL::APISet::Exceptions::' . $error . 'Exception';
        $error->throw( text => $error_message, response_object => $resp );
    }

    return [];
} ## end sub cfgremove
1;
