#
#Copyright (c) 2014 NetApp, Inc.
#All rights reserved.
#
## @summary Routines to parse BrocadeFC lscfg
##
## @description
##    This is a library to parse the raw output of BrocadeFC switch lscfg.
##
## @synopsis
##  use NACL::APISet::Switch::CLI::BrocadeFC::default::lscfg
##
##  lscfg(%args)
##
## @author sangavai@netapp.com, dl-nacl-dev@netapp.com
##
## @status public
#
package NACL::APISet::Switch::CLI::BrocadeFC::default::lscfg;

use NACL::APISet::Response::CLI::ParserUtils;

## @name ::lscfg_show
##
## @summary Parses the raw output from the "lscfg --show" command
##
## @description 
##    This method can be used for parsing the raw output obtained after calling 
##    the method from lscfg_show with any of its options as described in the 
##    corresponding cdef file 'lscfg.cdefs'
## @synopsis
##      lscfg_show(%args);
## INPUTS TO METHOD CALL 'lscfg_show' ON APISET OBJECT 
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples
##&nbsp;Command = lscfg --show
##&nbsp;
##&nbsp;-------- Raw Output 6.2.0---------
##&nbsp;; echo __status__=$?
##&nbsp;
##&nbsp;Created switches:  128(ds) 127
##&nbsp;
##&nbsp;Port      0     1     2     3     4     5     6     7     8     9
##&nbsp;-------------------------------------------------------------------
##&nbsp;FID     128 | 128 | 128 | 128 | 128 | 128 | 128 | 128 | 128 | 128 |
##&nbsp;
##&nbsp;Port     10    11    12    13    14    15    16    17    18    19
##&nbsp;-------------------------------------------------------------------
##&nbsp;FID     128 | 128 | 128 | 128 | 128 | 128 | 128 | 128 | 128 | 128 |
##&nbsp;
##&nbsp;Port     20    21    22    23    24    25    26    27    28    29
##&nbsp;-------------------------------------------------------------------
##&nbsp;FID     128 | 128 | 128 | 128 | 128 | 128 | 128 | 128 | 128 | 128 |
##&nbsp;
##&nbsp;Port     30    31    32    33    34    35    36    37    38    39
##&nbsp;-------------------------------------------------------------------
##&nbsp;FID     128 | 128 | 128 | 128 | 128 | 128 | 128 | 128 | 128 | 128 |
##&nbsp;
##&nbsp;Port     40    41    42    43    44    45    46    47
##&nbsp;-------------------------------------------------------
##&nbsp;FID     128 | 128 | 128 | 128 | 128 | 128 | 128 | 128 |
##&nbsp;
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;         {
##&nbsp;           'fabrics_info' => [
##&nbsp;                               {
##&nbsp;                                 'fid' => 128,
##&nbsp;                                 'port' => [
##&nbsp;                                             '0',
##&nbsp;                                             '1',
##&nbsp;                                             '2',
##&nbsp;                                             '3',
##&nbsp;                                             '4',
##&nbsp;                                             '5',
##&nbsp;                                             '6',
##&nbsp;                                             '7',
##&nbsp;                                             '8',
##&nbsp;                                             '9',
##&nbsp;                                             '10',
##&nbsp;                                             '11',
##&nbsp;                                             '12',
##&nbsp;                                             '13',
##&nbsp;                                             '14',
##&nbsp;                                             '15',
##&nbsp;                                             '16',
##&nbsp;                                             '17',
##&nbsp;                                             '18',
##&nbsp;                                             '19',
##&nbsp;                                             '20',
##&nbsp;                                             '21',
##&nbsp;                                             '22',
##&nbsp;                                             '23',
##&nbsp;                                             '24',
##&nbsp;                                             '25',
##&nbsp;                                             '26',
##&nbsp;                                             '27',
##&nbsp;                                             '28',
##&nbsp;                                             '29',
##&nbsp;                                             '30',
##&nbsp;                                             '31',
##&nbsp;                                             '32',
##&nbsp;                                             '33',
##&nbsp;                                             '34',
##&nbsp;                                             '35',
##&nbsp;                                             '36',
##&nbsp;                                             '37',
##&nbsp;                                             '38',
##&nbsp;                                             '39',
##&nbsp;                                             '40',
##&nbsp;                                             '41',
##&nbsp;                                             '42',
##&nbsp;                                             '43',
##&nbsp;                                             '44',
##&nbsp;                                             '45',
##&nbsp;                                             '46',
##&nbsp;                                             '47'
##&nbsp;                                           ]
##&nbsp;                               }
##&nbsp;                             ],
##&nbsp;             'created_switches' => [
##&nbsp;                                     '128(ds)',
##&nbsp;                                     '127'
##&nbsp;                                   ]
##&nbsp;           }
##&nbsp;       ];
##&nbsp;
##&nbsp;
##&nbsp;-------- Raw Output v7.2.1---------
##&nbsp;; echo __status__=$?
##&nbsp;
##&nbsp;Created switches FIDs(Domain IDs): 128(ds)(10) 1(bs)(1) 2(1)
##&nbsp;
##&nbsp;Port      0     1     2     3     4     5     6     7     8     9
##&nbsp;-------------------------------------------------------------------
##&nbsp;FID      1 | 1 | 1 | 1 | 1 | 1 | 1 | 1 | 2 | 2
##&nbsp;
##&nbsp;Port     10    11    12    13    14    15    16    17    18    19
##&nbsp;-------------------------------------------------------------------
##&nbsp;FID     128 | 128 | 128 | 128 | 128 | 128 | 128 | 128 | 128 | 128 |
##&nbsp;
##&nbsp;Port     20    21    22    23    24    25    26    27    28    29
##&nbsp;-------------------------------------------------------------------
##&nbsp;FID     128 | 128 | 128 | 128 | 128 | 128 | 128 | 128 | 128 | 128 |
##&nbsp;
##&nbsp;Port     30    31    32    33    34    35    36    37    38    39
##&nbsp;-------------------------------------------------------------------
##&nbsp;FID     128 | 128 | 128 | 2 | 2 | 2 | 128 | 128 | 128 | 128
##&nbsp;
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;         {
##&nbsp;           'fabrics_info' => [
##&nbsp;                               {
##&nbsp;                                 'fid' => 128,
##&nbsp;                                 'port' => [
##&nbsp;                                             '10',
##&nbsp;                                             '11',
##&nbsp;                                             '12',
##&nbsp;                                             '13',
##&nbsp;                                             '14',
##&nbsp;                                             '15',
##&nbsp;                                             '16',
##&nbsp;                                             '17',
##&nbsp;                                             '18',
##&nbsp;                                             '19',
##&nbsp;                                             '20',
##&nbsp;                                             '21',
##&nbsp;                                             '22',
##&nbsp;                                             '23',
##&nbsp;                                             '24',
##&nbsp;                                             '25',
##&nbsp;                                             '26',
##&nbsp;                                             '27',
##&nbsp;                                             '28',
##&nbsp;                                             '29',
##&nbsp;                                             '30',
##&nbsp;                                             '31',
##&nbsp;                                             '32',
##&nbsp;                                             '36',
##&nbsp;                                             '37',
##&nbsp;                                             '38',
##&nbsp;                                             '39',
##&nbsp;                                           ]
##&nbsp;                               },
##&nbsp;                                 'fid' => 1,
##&nbsp;                                 'port' => [
##&nbsp;                                             '0',
##&nbsp;                                             '1',
##&nbsp;                                             '2',
##&nbsp;                                             '3',
##&nbsp;                                             '4',
##&nbsp;                                             '5',
##&nbsp;                                             '6',
##&nbsp;                                             '7',
##&nbsp;                                           ]
##&nbsp;                               },
##&nbsp;                                 'fid' => 2,
##&nbsp;                                 'port' => [
##&nbsp;                                             '8',
##&nbsp;                                             '9',
##&nbsp;                                             '33',
##&nbsp;                                             '34',
##&nbsp;                                             '35',
##&nbsp;                                           ]
##&nbsp;                               }
##&nbsp;                             ],
##&nbsp;             'created_switches' => [
##&nbsp;                                     '128(ds)(10)',
##&nbsp;                                     '1(bs)(1)',
##&nbsp;                                     '2(1)'
##&nbsp;                                   ]
##&nbsp;           }
##&nbsp;       ];
##&nbsp;

sub lscfg_show {
    my %args = @_;
    my (@result_array);
    my $resp       = $args{response_object};
    my $output     = $resp->get_processed_output();
    my @lines = get_lines( raw_output => $output );
    my $m = 0;
    my ( @ports, @switches, @inc );

    foreach my $line (@lines) {
        if ( $line =~ /(Created\s+switches)(?: FIDs\(Domain IDs\))?:\s+(.*)/ ) {
            @switches = split( /\s+/, $2 );
            for ( my $pos = 0; $pos <= $#switches; $pos++ ) {
                $result_array[0]->{ to_key_name(item => $1) }[$pos] = $switches[$pos];
                push( @inc, 0 );
            }
        } elsif ( $line =~ /Port\s+/ ) {
            @ports = split( /\s+/, $' );
        } elsif ( $line =~ /FID\s+/ ) {
            my @fids = split( /\s+\|\s+/, $' );
            for ( my $sw = 0; $sw <= $#fids; $sw++ ) {
                for ( my $f = 0; $f <= $#switches; $f++ ) {
                    $fids[$sw] =~ s/\s+\|//g;
                    $switches[$f] =~ s/\W+\w+\W+\W*\w*\W*//g;
                    if ( $switches[$f] == $fids[$sw] ) {
                        $m = $f;
                        $result_array[0]->{'fabrics_info'}[$m]->{fid}
                            = $fids[$sw];
                        $result_array[0]->{'fabrics_info'}[$m]
                            ->{'port'}[ $inc[$m]++ ] = $ports[$sw];
                    }
                }
            }
        }
    }
    return \@result_array;
} ## end sub lscfg_show

1;
