#
#Copyright (c) 2011 Network Appliance, Inc.
#All rights reserved.
#
## @summary Routines to parse BrocadeFC  nsshow command
##
## @description
##    This is a library to parse the raw output of BrocadeFC switch  nsshow command.
##
## @synopsis
##  use NACL::APISet::Switch::CLI::BrocadeFC::default::nsshow
##
##   nsshow(%args)
##
## @author Govardhan Yadav (yadag@netapp.com),dl-nacl-dev@netapp.com
##
## @status public
#
package NACL::APISet::Switch::CLI::BrocadeFC::default::nsshow;

use strict;
use NACL::APISet::Response::CLI::ParserUtils;

#Include exceptions which will be thrown
use NACL::APISet::Exceptions::CommandFailedException qw(:try);
use NACL::APISet::Exceptions::MissingRequiredParamException;
use NACL::APISet::Exceptions::InvalidParamValueException;

sub _check_error {
    my %args = @_;

    my $raw_output = $args{raw_output};
    my $command = $args{command};
    my $error_message = $args{error_message};
    my $error;

    if($raw_output =~ /usage.+/si) {
        $error = "InvalidParam";
        $$error_message = "Invalid options are used in the command";
    } elsif($raw_output =~ /^(nsshow: invalid option .*)/si) {
        $error = "InvalidParamValue";
        $$error_message = $1;
    } elsif($raw_output =~ /^(nsshow: Invalid input.*)/si) {
        $error = "InvalidParamValue";
        $$error_message = $1;
    }
    return $error;
}


## @name ::nsshow
##
## @summary Parses the raw output from the "nsshow" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from nsshow with any of its options as described in the
##    corresponding cdef file 'nsshow.cdefs'
## @synopsis
##      nsshow(%args);
## INPUTS TO METHOD CALL 'nsshow' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> </tr>
## <tr> <td> '-r' </td> <td> 'replace' </td> </tr>
## <tr> <td> '-t' </td> <td> 'device-type' </td> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples
##&nbsp;Command = nsshow
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;{
##&nbsp; Type Pid    COS     PortName                NodeName                 TTL(sec)
##&nbsp; N    010000;      3;50:0a:09:80:00:14:1b:a0;50:0a:09:80:00:14:1b:a0; na
##&nbsp;    PortSymb: [29] "NetApp FC Storage Adapter 0c."
##&nbsp;    Fabric Port Name: 20:00:00:05:1e:03:aa:e8
##&nbsp;    Permanent Port Name: 50:0a:09:80:00:14:1b:a0
##&nbsp;    Port Index: 0
##&nbsp;    Share Area: No
##&nbsp;    Device Shared in Other AD: No
##&nbsp;    Redirect: No
##&nbsp; N    010400;      3;50:0a:09:80:00:14:25:90;50:0a:09:80:00:14:25:90; na
##&nbsp;    PortSymb: [29] "NetApp FC Storage Adapter 0c."
##&nbsp;    Fabric Port Name: 20:04:00:05:1e:03:aa:e8
##&nbsp;    Permanent Port Name: 50:0a:09:80:00:14:25:90
##&nbsp;    Port Index: 4
##&nbsp;    Share Area: No
##&nbsp;    Device Shared in Other AD: No
##&nbsp;    Redirect: No
##&nbsp; N    010f00;      3;50:06:01:68:30:22:7a:7f;50:06:01:60:b0:22:7a:7f; na
##&nbsp;    FC4s: FCP [DGC     LUNZ            0226]
##&nbsp;    Fabric Port Name: 20:0f:00:05:1e:03:aa:e8
##&nbsp;    Permanent Port Name: 50:06:01:68:30:22:7a:7f
##&nbsp;    Port Index: 15
##&nbsp;    Share Area: No
##&nbsp;    Device Shared in Other AD: No
##&nbsp;    Redirect: No
##&nbsp;The Local Name Server has 3 entries }
##&nbsp;vgbr200es2:admin
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;          {
##&nbsp;            'share_area' => 'No',
##&nbsp;            'fabric_port_name' => '20:00:00:05:1e:03:aa:e8',
##&nbsp;            'nodename' => '50:0a:09:80:00:14:1b:a0',
##&nbsp;            'portsymb' => '[29] "NetApp FC Storage Adapter 0c."',
##&nbsp;            'portname' => '50:0a:09:80:00:14:1b:a0',
##&nbsp;            'device_shared_in_other_ad' => 'No',
##&nbsp;            'ttlsec' => 'na',
##&nbsp;            'port_index' => '0',
##&nbsp;            'cos' => '3',
##&nbsp;            'redirect' => 'No',
##&nbsp;            'pid' => '010000',
##&nbsp;            'permanent_port_name' => '50:0a:09:80:00:14:1b:a0',
##&nbsp;            'type' => 'N'
##&nbsp;          },
##&nbsp;          ...................
##&nbsp;          ...................
##&nbsp;          ...................
##&nbsp;          {
##&nbsp;            'no_of_entries' => '3 entries '
##&nbsp;          }
##&nbsp;        ];
##&nbsp;
sub  nsshow {
    my %args = @_;
    my ($error_message, @result_array);
    my $resp = $args{response_object};
    my $output = $resp->get_processed_output();
    my $command = $resp->get_command();
    my $error_level = $resp->get_execution_status();
    my ($error, $key);

    $error = _check_error(raw_output=>$output,command=>" nsshow",
                             error_message=>\$error_message);
    if($error) {
        $error = 'NACL::APISet::Exceptions::'.$error.'Exception';
        $error->throw(text=>$error_message,response_object => $resp);
    }
    # Now we can parse the output
    my @lines = get_lines("raw_output"=>$output,"response_object"=>$resp);

    my (@Res,$value);
    my $i = -1;
    my @H_ar;
    foreach $value (@lines){
        if($value =~/[\{\>]/){
            next;
        } elsif($value =~/Type\s+Pid\s+COS\s+PortName\s+NodeName\s+?(TTL.sec.|SCR)/){
            @H_ar = split(/\s+/,$value);
        } elsif($value =~/^(\w+)\s+(\d+|\w+);\s+(\d+);([0-9a-z\:]+);([0-9a-z\:]+);\s+(\d+|\w+)$/){
            $i++;
            $Res[$i]->{to_key_name(item =>$H_ar[0])} = $1;
            $Res[$i]->{to_key_name(item =>$H_ar[1])} = $2;
            $Res[$i]->{to_key_name(item =>$H_ar[2])} = $3;
            $Res[$i]->{to_key_name(item =>$H_ar[3])} = $4;
            $Res[$i]->{to_key_name(item =>$H_ar[4])} = $5;
            $Res[$i]->{to_key_name(item =>$H_ar[5])} = $6;
        } elsif($value =~/^([\w\s]+):\s+(.*)$/){
            $Res[$i]->{to_key_name(item => $1)} = $2;
        } elsif($value =~/^The Local Name Server has (.*)}/){
            $Res[$i+1]->{"no_of_entries"} = $1;
        }
    }
    return \@Res;
}

1;
