#
#Copyright (c) 2012 Network Appliance, Inc.
#All rights reserved.
#
## @summary Routines to parse BrocadeFC portenable, portdisable commands
##
## @description
##    This is a library to parse the raw output of BrocadeFC switch portenable, portdisable commands.
##
## @synopsis
##  use NACL::APISet::Switch::CLI::BrocadeFC::default::port
##
##  portenable(%args), portdisable(%args)
##
## @author jais@netapp.com
##
## @status public
#
package NACL::APISet::Switch::CLI::BrocadeFC::default::port;

use strict;

#Include exceptions which will be thrown
use NACL::APISet::Exceptions::CommandFailedException qw(:try);
use NACL::APISet::Exceptions::InvalidParamException;
use NACL::APISet::Exceptions::InvalidParamValueException;

sub _check_error {
    my %args = @_;

    my $raw_output = $args{raw_output};
    my $resp = $args{"response_object"};
    my $error_message;
    my $error;

    # Check for the error
    if ($raw_output =~ /(failed.*license)/) {
        $error = "CommandFailed";
        $error_message = "Portenable".$1;
    } elsif ($raw_output =~ /(Invalid.*)/) {
        $error = "InvalidParam";
        $error_message = $1;
    } elsif ($raw_output =~ /(Failed.*range)/) {
        $error = "InvalidParamValue";
        $error_message = $1;
    } elsif ($raw_output =~ /(command not found)/) {
        $error = "InvalidParam";
        $error_message = $1;
    } elsif ($raw_output =~ /(Error.*options)/) {
        $error = "InvalidParam";
        $error_message = $1;
    } elsif ($raw_output =~ /(syntax error.*)/) {
        $error = "InvalidParam";
        $error_message = $1;
    } elsif ($raw_output =~ /slot.*supported/) {
        $error = "CommandFailed";
        $error_message = "-s options supported only in slot based system";
    }
    if ($error) {
        $error = 'NACL::APISet::Exceptions::'.$error.'Exception';
        $error->throw(text => $error_message, response_object => $resp);
    }
}

## @name ::portenable
##
## @summary Parses the raw output from the "portenable" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from portenable with any of its options as described in the
##    corresponding cdef file 'port.cdefs'
## @synopsis
##      portenable(%args);
## INPUTS TO METHOD CALL 'portenable' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## <tr> <td> '-i' </td> <td> 'index' </td> <td> '0' </td> </tr>
## <tr> <td> '-slot' </td> <td> 'slot-number' </td> <td> '0' </td> </tr>
## <tr> <td> '-f' </td> <td> 'ignore-nonexisting-port' </td> <td> '1' </td> </tr>
## <tr> <td> No switch for this key </td> <td> 'port' </td> <td> '0' </td> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

##&nbsp;Command = portenable -i 5
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;dpg-mcc-6510-01:FID128:admin
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [];
##&nbsp;
sub portenable {
    my %args = @_;

    my $resp         = $args{response_object};
    my $raw_output   = $resp->get_processed_output();

    _check_error(raw_output=>$raw_output, response_object=>$resp);

    return [];
}

## @name ::portdisable
##
## @summary Parses the raw output from the "portdisable" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from portdisable with any of its options as described in the
##    corresponding cdef file 'port.cdefs'
## @synopsis
##      portdisable(%args);
## INPUTS TO METHOD CALL 'portdisable' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## <tr> <td> '-i' </td> <td> 'index' </td> <td> '0' </td> </tr>
## <tr> <td> '-slot' </td> <td> 'slot-number' </td> <td> '0' </td> </tr>
# <tr> <td> '-f' </td> <td> 'ignore-nonexisting-port' </td> <td> '1' </td> </tr>
## <tr> <td> No switch for this key </td> <td> 'port' </td> <td> '0' </td> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

##&nbsp;Command = portdisable -i 5
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;dpg-mcc-6510-01:FID128:admin
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [];
##&nbsp;
sub portdisable {
    my %args = @_;

    my $resp         = $args{response_object};
    my $raw_output   = $resp->get_processed_output();

    _check_error(raw_output=>$raw_output, response_object=>$resp);

    return [];
}

1;

