#
#Copyright (c) 2011 Network Appliance, Inc.
#All rights reserved.
#
## @summary Routines to parse BrocadeFC portbuffershow command
##
## @description
##    This is a library to parse the raw output of BrocadeFC switch portbuffershow command.
##
## @synopsis
##  use NACL::APISet::Switch::CLI::BrocadeFC::default::portbuffershow
##
##  portbuffershow(%args)
##
## @author rroshan@netapp.com, dl-nacl-dev@netapp.com 
##
## @status
#
package NACL::APISet::Switch::CLI::BrocadeFC::default::portbuffershow;

use strict;
use NACL::APISet::Response::CLI::ParserUtils;

#Include exceptions which will be thrown
use NACL::APISet::Exceptions::CommandFailedException qw(:try);
use NACL::APISet::Exceptions::InvalidParamValueException qw(:try);

sub _check_error {
    my %args = @_;

    my $raw_output = $args{raw_output};
    my $command = $args{command};
    my $error_message = $args{error_message};
    my $error ;

    # Check for the error
    if($raw_output =~ /(portbuffershow: Invalid input.* )/) {
        $error = "InvalidParamValue";
        $$error_message = $1;
    }
    return $error;
}

## @name ::portbuffershow
##
## @summary Parses the raw output from the "portbuffershow" command
##
## @description 
##    This method can be used for parsing the raw output obtained after calling 
##    the method from portbuffershow with any of its options as described in the 
##    corresponding cdef file 'portbuffershow.cdefs'
## @synopsis
##      portbuffershow(%args);
## INPUTS TO METHOD CALL 'portbuffershow' ON APISET OBJECT 
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> </tr>
## <tr> <td> No switch for this key </td> <td> 'slot-num' </td> </tr> 
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples
##&nbsp;
##&nbsp;Command = portbuffershow
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;          {
##&nbsp;            'lx_mode' => '-',
##&nbsp;            'user_port' => '0',
##&nbsp;            'link_distance' => '-',
##&nbsp;            'max_resv_buffers' => '8',
##&nbsp;            'needed_buffer' => '-',
##&nbsp;            'buffer_usage' => '8',
##&nbsp;            'port_type' => 'F'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'lx_mode' => '-',
##&nbsp;            'user_port' => '1',
##&nbsp;            'link_distance' => '-',
##&nbsp;            'max_resv_buffers' => '8',
##&nbsp;            'needed_buffer' => '-',
##&nbsp;            'buffer_usage' => '0',
##&nbsp;            'port_type' => ''
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'lx_mode' => '-',
##&nbsp;            'user_port' => '2',
##&nbsp;            'link_distance' => '-',
##&nbsp;            'max_resv_buffers' => '8',
##&nbsp;            'needed_buffer' => '-',
##&nbsp;            'buffer_usage' => '0',
##&nbsp;            'port_type' => ''
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'lx_mode' => '-',
##&nbsp;            'user_port' => '3',
##&nbsp;            'link_distance' => '-',
##&nbsp;            'max_resv_buffers' => '8',
##&nbsp;            'needed_buffer' => '-',
##&nbsp;            'buffer_usage' => '0',
##&nbsp;            'port_type' => ''
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'lx_mode' => '-',
##&nbsp;            'user_port' => '4',
##&nbsp;            'link_distance' => '-',
##&nbsp;            'max_resv_buffers' => '8',
##&nbsp;            'needed_buffer' => '-',
##&nbsp;            'buffer_usage' => '8',
##&nbsp;            'port_type' => 'F'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'lx_mode' => '-',
##&nbsp;            'user_port' => '5',
##&nbsp;            'link_distance' => '-',
##&nbsp;            'max_resv_buffers' => '8',
##&nbsp;            'needed_buffer' => '-',
##&nbsp;            'buffer_usage' => '0',
##&nbsp;            'port_type' => ''
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'lx_mode' => '-',
##&nbsp;            'user_port' => '6',
##&nbsp;            'link_distance' => '-',
##&nbsp;            'max_resv_buffers' => '8',
##&nbsp;            'needed_buffer' => '-',
##&nbsp;            'buffer_usage' => '0',
##&nbsp;            'port_type' => ''
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'lx_mode' => '-',
##&nbsp;            'user_port' => '7',
##&nbsp;            'link_distance' => '-',
##&nbsp;            'max_resv_buffers' => '8',
##&nbsp;            'needed_buffer' => '-',
##&nbsp;            'buffer_usage' => '0',
##&nbsp;            'port_type' => ''
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'lx_mode' => '-',
##&nbsp;            'user_port' => '8',
##&nbsp;            'link_distance' => '-',
##&nbsp;            'max_resv_buffers' => '8',
##&nbsp;            'needed_buffer' => '-',
##&nbsp;            'buffer_usage' => '0',
##&nbsp;            'port_type' => ''
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'lx_mode' => '-',
##&nbsp;            'user_port' => '9',
##&nbsp;            'link_distance' => '-',
##&nbsp;            'max_resv_buffers' => '8',
##&nbsp;            'needed_buffer' => '-',
##&nbsp;            'buffer_usage' => '0',
##&nbsp;            'port_type' => ''
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'lx_mode' => '-',
##&nbsp;            'user_port' => '10',
##&nbsp;            'link_distance' => '-',
##&nbsp;            'max_resv_buffers' => '8',
##&nbsp;            'needed_buffer' => '-',
##&nbsp;            'buffer_usage' => '0',
##&nbsp;            'port_type' => ''
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'lx_mode' => '-',
##&nbsp;            'user_port' => '11',
##&nbsp;            'link_distance' => '-',
##&nbsp;            'max_resv_buffers' => '8',
##&nbsp;            'needed_buffer' => '-',
##&nbsp;            'buffer_usage' => '0',
##&nbsp;            'port_type' => ''
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'lx_mode' => '-',
##&nbsp;            'user_port' => '12',
##&nbsp;            'link_distance' => '-',
##&nbsp;            'max_resv_buffers' => '8',
##&nbsp;            'needed_buffer' => '-',
##&nbsp;            'buffer_usage' => '0',
##&nbsp;            'port_type' => ''
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'lx_mode' => '-',
##&nbsp;            'user_port' => '13',
##&nbsp;            'link_distance' => '-',
##&nbsp;            'max_resv_buffers' => '8',
##&nbsp;            'needed_buffer' => '-',
##&nbsp;            'buffer_usage' => '0',
##&nbsp;            'port_type' => ''
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'lx_mode' => '-',
##&nbsp;            'user_port' => '14',
##&nbsp;            'link_distance' => '-',
##&nbsp;            'max_resv_buffers' => '8',
##&nbsp;            'needed_buffer' => '-',
##&nbsp;            'buffer_usage' => '0',
##&nbsp;            'port_type' => ''
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'lx_mode' => '-',
##&nbsp;            'user_port' => '15',
##&nbsp;            'remaining_buffers' => '144',
##&nbsp;            'link_distance' => '-',
##&nbsp;            'max_resv_buffers' => '8',
##&nbsp;            'needed_buffer' => '-',
##&nbsp;            'buffer_usage' => '8',
##&nbsp;            'port_type' => 'F'
##&nbsp;          }
##&nbsp;        ];
##&nbsp;
##&nbsp;Command = portbuffershow "2"
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;          {
##&nbsp;            'lx_mode' => '-',
##&nbsp;            'user_port' => '0',
##&nbsp;            'link_distance' => '-',
##&nbsp;            'max_resv_buffers' => '8',
##&nbsp;            'needed_buffer' => '-',
##&nbsp;            'buffer_usage' => '8',
##&nbsp;            'port_type' => 'F'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'lx_mode' => '-',
##&nbsp;            'user_port' => '1',
##&nbsp;            'link_distance' => '-',
##&nbsp;            'max_resv_buffers' => '8',
##&nbsp;            'needed_buffer' => '-',
##&nbsp;            'buffer_usage' => '0',
##&nbsp;            'port_type' => ''
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'lx_mode' => '-',
##&nbsp;            'user_port' => '2',
##&nbsp;            'link_distance' => '-',
##&nbsp;            'max_resv_buffers' => '8',
##&nbsp;            'needed_buffer' => '-',
##&nbsp;            'buffer_usage' => '0',
##&nbsp;            'port_type' => ''
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'lx_mode' => '-',
##&nbsp;            'user_port' => '3',
##&nbsp;            'link_distance' => '-',
##&nbsp;            'max_resv_buffers' => '8',
##&nbsp;            'needed_buffer' => '-',
##&nbsp;            'buffer_usage' => '0',
##&nbsp;            'port_type' => ''
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'lx_mode' => '-',
##&nbsp;            'user_port' => '4',
##&nbsp;            'link_distance' => '-',
##&nbsp;            'max_resv_buffers' => '8',
##&nbsp;            'needed_buffer' => '-',
##&nbsp;            'buffer_usage' => '8',
##&nbsp;            'port_type' => 'F'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'lx_mode' => '-',
##&nbsp;            'user_port' => '5',
##&nbsp;            'link_distance' => '-',
##&nbsp;            'max_resv_buffers' => '8',
##&nbsp;            'needed_buffer' => '-',
##&nbsp;            'buffer_usage' => '0',
##&nbsp;            'port_type' => ''
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'lx_mode' => '-',
##&nbsp;            'user_port' => '6',
##&nbsp;            'link_distance' => '-',
##&nbsp;            'max_resv_buffers' => '8',
##&nbsp;            'needed_buffer' => '-',
##&nbsp;            'buffer_usage' => '0',
##&nbsp;            'port_type' => ''
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'lx_mode' => '-',
##&nbsp;            'user_port' => '7',
##&nbsp;            'link_distance' => '-',
##&nbsp;            'max_resv_buffers' => '8',
##&nbsp;            'needed_buffer' => '-',
##&nbsp;            'buffer_usage' => '0',
##&nbsp;            'port_type' => ''
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'lx_mode' => '-',
##&nbsp;            'user_port' => '8',
##&nbsp;            'link_distance' => '-',
##&nbsp;            'max_resv_buffers' => '8',
##&nbsp;            'needed_buffer' => '-',
##&nbsp;            'buffer_usage' => '0',
##&nbsp;            'port_type' => ''
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'lx_mode' => '-',
##&nbsp;            'user_port' => '9',
##&nbsp;            'link_distance' => '-',
##&nbsp;            'max_resv_buffers' => '8',
##&nbsp;            'needed_buffer' => '-',
##&nbsp;            'buffer_usage' => '0',
##&nbsp;            'port_type' => ''
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'lx_mode' => '-',
##&nbsp;            'user_port' => '10',
##&nbsp;            'link_distance' => '-',
##&nbsp;            'max_resv_buffers' => '8',
##&nbsp;            'needed_buffer' => '-',
##&nbsp;            'buffer_usage' => '0',
##&nbsp;            'port_type' => ''
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'lx_mode' => '-',
##&nbsp;            'user_port' => '11',
##&nbsp;            'link_distance' => '-',
##&nbsp;            'max_resv_buffers' => '8',
##&nbsp;            'needed_buffer' => '-',
##&nbsp;            'buffer_usage' => '0',
##&nbsp;            'port_type' => ''
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'lx_mode' => '-',
##&nbsp;            'user_port' => '12',
##&nbsp;            'link_distance' => '-',
##&nbsp;            'max_resv_buffers' => '8',
##&nbsp;            'needed_buffer' => '-',
##&nbsp;            'buffer_usage' => '0',
##&nbsp;            'port_type' => ''
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'lx_mode' => '-',
##&nbsp;            'user_port' => '13',
##&nbsp;            'link_distance' => '-',
##&nbsp;            'max_resv_buffers' => '8',
##&nbsp;            'needed_buffer' => '-',
##&nbsp;            'buffer_usage' => '0',
##&nbsp;            'port_type' => ''
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'lx_mode' => '-',
##&nbsp;            'user_port' => '14',
##&nbsp;            'link_distance' => '-',
##&nbsp;            'max_resv_buffers' => '8',
##&nbsp;            'needed_buffer' => '-',
##&nbsp;            'buffer_usage' => '0',
##&nbsp;            'port_type' => ''
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'lx_mode' => '-',
##&nbsp;            'user_port' => '15',
##&nbsp;            'remaining_buffers' => '144',
##&nbsp;            'link_distance' => '-',
##&nbsp;            'max_resv_buffers' => '8',
##&nbsp;            'needed_buffer' => '-',
##&nbsp;            'buffer_usage' => '8',
##&nbsp;            'port_type' => 'F'
##&nbsp;          }
##&nbsp;        ];
##&nbsp;
sub portbuffershow {
    my %args = @_;

    my ($error_message, @result_array);
    my $resp = $args{response_object};
    my $output = $resp->get_processed_output();
    my $command = $resp->get_command();
    my $error_level = $resp->get_execution_status();
    my ($error, $key);
    # Status returned by switch for this command is
    # is ZERO even if command failed
    $error = _check_error(raw_output=>$output,command=>$command,
                             error_message=>\$error_message);
    if($error) {
        $error = 'NACL::APISet::Exceptions::'.$error.'Exception';
        $error->throw(text=>$error_message,response_object => $resp);
    }
    my $first_line = "Port     Type    Mode          Buffers     Usage  Buffers     Distance  Buffers";
    #Now we can parse the output
    my @array_lines = get_lines(raw_output=>$output,first_line=>$first_line);
    shift (@array_lines);
    shift (@array_lines);
    my (@Res,$value);
    my $i = 0;
    foreach $value(@array_lines){
        if($value =~/^(\w+)\s+(\w+|)\s+([\w\-]+)\s+(\d+)\s+(\d+)\s+([\d\-]+)\s+([\w\-]+)$/) {
            $Res[$i]->{"user_port"} = $1;
            $Res[$i]->{"port_type"} = $2;
            $Res[$i]->{"lx_mode"} = $3;
            $Res[$i]->{"max_resv_buffers"} = $4;
            $Res[$i]->{"buffer_usage"} = $5;
            $Res[$i]->{"needed_buffer"} = $6;
            $Res[$i]->{"link_distance"} = $7;
        } elsif($value =~ /^(\w+)\s+(\w+|)\s+([\w\-]+)\s+(\d+)\s+(\d+)\s+([\d\-]+)\s+([\w\-]+)\s+(\w+)$/) {
            $Res[$i]->{"user_port"} = $1;
            $Res[$i]->{"port_type"} = $2;
            $Res[$i]->{"lx_mode"} = $3;
            $Res[$i]->{"max_resv_buffers"} = $4;
            $Res[$i]->{"buffer_usage"} = $5;
            $Res[$i]->{"needed_buffer"} = $6;
            $Res[$i]->{"link_distance"} = $7;
            $Res[$i]->{"remaining_buffers"} = $8;
        }
        $i++;
    }
    return \@Res;
}
1;
