#
#Copyright (c) 2011 Network Appliance, Inc.
#All rights reserved.
#
## @summary Routines to parse BrocadeFC portcfgspeed command
##
## @description
##    This is a library to parse the raw output of BrocadeFC switch portcfgspeed command.
##
## @synopsis
##  use NACL::APISet::Switch::CLI::BrocadeFC::default::portcfgspeed
##
##  portcfgspeed(%args)
##
## @author rroshan@netapp.com ,dl-nacl-dev@netapp.com
##
## @status
#
package NACL::APISet::Switch::CLI::BrocadeFC::default::portcfgspeed;

use strict;
use NACL::APISet::Response::CLI::ParserUtils;

#Include exceptions which will be thrown
use NACL::APISet::Exceptions::CommandFailedException qw(:try);
use NACL::APISet::Exceptions::InvalidParamException;
use NACL::APISet::Exceptions::InvalidParamValueException qw(:try);

sub _check_error {
    my %args = @_;
    my $raw_output = $args{raw_output};
    my $command = $args{command};
    my $error_message = $args{error_message};
    my $error ;

    # Check for the error
    if ($raw_output =~ /.*:\s(.*)/) {
        $$error_message = $1;
        if($$error_message =~ /(Invalid input.*)|(portCfgSpeed.*)|(Failed to get.*)/) {
            $error = "InvalidParam";
        } elsif($$error_message =~ /([Ii]nvalid port)|([Ii]nvalid speed)/) {
            $error = "InvalidParamValue";
        }
    }
    return $error;
}

## @name ::portcfgspeed
##
## @summary Parses the raw output from the "portcfgspeed" command
##
## @description 
##    This method can be used for parsing the raw output obtained after calling 
##    the method from portcfgspeed with any of its options as described in the 
##    corresponding cdef file 'portcfgspeed.cdefs'
## @synopsis
##      portcfgspeed(%args);
## INPUTS TO METHOD CALL 'portcfgspeed' ON APISET OBJECT 
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> </tr>
## <tr> <td> No switch for this key </td> <td> 'speed-level' </td> </tr> 
## <tr> <td> No switch for this key </td> <td> 'port-num' </td> </tr> 
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples
##&nbsp;Command = portcfgspeed "0/15","2"
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;          {
##&nbsp;            'Configuration' => 'same for port 15'
##&nbsp;          }
##&nbsp;        ];
##&nbsp;
##&nbsp;Command = portcfgspeed "0/15","ax"
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [];
##&nbsp;
sub portcfgspeed {
    my %args = @_;
    my ($error_message, %result_array, @res);
    my $resp = $args{response_object};
    my $output = $resp->get_processed_output();
    my $command = $resp->get_command();
    my $error_level = $resp->get_execution_status();
    my ($error, $key);
    # Status returned by switch for this command is
    # is ZERO even if command failed
    $error = _check_error(raw_output=>$output,command=>$command,
                             error_message=>\$error_message);
    if($error) {
        $error = 'NACL::APISet::Exceptions::'.$error.'Exception';
        $error->throw(text=>$error_message,response_object => $resp);
    }
    if($output =~ /Same configuration for port\s\((\d+)\)\./) {
        $result_array{"Configuration"} = "same for port $1";
    } else {
        return \@res;
    }
    push (@res, \%result_array);
    return \@res;
}
1;
