#
#Copyright (c) 2011 Network Appliance, Inc.
#All rights reserved.
#
## @summary Routines to parse BrocadeFC porterrshow command
##
## @description
##    This is a library to parse the raw output of BrocadeFC switch porterrshow command.
##
## @synopsis
##  use NACL::APISet::Switch::CLI::BrocadeFC::default::porterrshow
##
##  porterrshow(%args)
##
## @author rroshan@netapp.com
##
## @status
#
package NACL::APISet::Switch::CLI::BrocadeFC::default::porterrshow;

use strict;
use NACL::APISet::Response::CLI::ParserUtils;

#Include exceptions which will be thrown
use NACL::APISet::Exceptions::CommandFailedException qw(:try);
use NACL::APISet::Exceptions::InvalidParamException;

sub _check_error {
    my %args = @_;

    my $raw_output = $args{raw_output};
    my $command = $args{command};
    my $error_message = $args{error_message};
    my $error ;

    # Check for the error
        if($raw_output =~ /(porterrshow: Invalid input.* )/) {
            $error = $1;
        }
    return $error;
}

## @name ::porterrshow
##
## @summary Parses the raw output from the "porterrshow" command
##
## @description 
##    This method can be used for parsing the raw output obtained after calling 
##    the method from porterrshow with any of its options as described in the 
##    corresponding cdef file 'porterrshow.cdefs'
## @synopsis
##      porterrshow(%args);
## INPUTS TO METHOD CALL 'porterrshow' ON APISET OBJECT 
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

##&nbsp;Command = porterrshow
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;          {
##&nbsp;            'frames_transmitted' => '34m',
##&nbsp;            'signal_loss' => '148',
##&nbsp;            'frames_received' => '59m',
##&nbsp;            'fbsy_frames' => '0',
##&nbsp;            'sync_loss' => '117',
##&nbsp;            'c3_frames_disc' => '44',
##&nbsp;            'bad_eof' => '0',
##&nbsp;            'crc_err_eof' => '0',
##&nbsp;            'frjt_frames' => '0',
##&nbsp;            'encoded_errors_out' => '905',
##&nbsp;            'too_short' => '0',
##&nbsp;            'crc_errors' => '0',
##&nbsp;            'too_long' => '0',
##&nbsp;            'link_failures' => '0',
##&nbsp;            'encoded_errors_in' => '0'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'frames_transmitted' => '0',
##&nbsp;            'signal_loss' => '44',
##&nbsp;            'frames_received' => '0',
##&nbsp;            'fbsy_frames' => '0',
##&nbsp;            'sync_loss' => '0',
##&nbsp;            'c3_frames_disc' => '0',
##&nbsp;            'bad_eof' => '0',
##&nbsp;            'crc_err_eof' => '0',
##&nbsp;            'frjt_frames' => '0',
##&nbsp;            'encoded_errors_out' => '0',
##&nbsp;            'too_short' => '0',
##&nbsp;            'crc_errors' => '0',
##&nbsp;            'too_long' => '0',
##&nbsp;            'link_failures' => '0',
##&nbsp;            'encoded_errors_in' => '0'
##&nbsp;          }
##&nbsp;        ];
##&nbsp;
sub porterrshow {
    my %args = @_;

    my ($error_message, @result_array);
    my $resp = $args{response_object};
    my $output = $resp->get_processed_output();
    my $command = $resp->get_command();
    my $error_level = $resp->get_execution_status();
    my ($error, $key);

    # Status returned by switch for this command is
    # is ZERO even if command failed
    $error = _check_error(raw_output=>$output,command=>"porterrshow",
                             error_message=>\$error_message);
    if($error) {
        $error = 'NACL::APISet::Exceptions::'.$error.'Exception';
        $error->throw(text=>$error_message,response_object => $resp);
    }
    my $first_line = "============================================================================";
    #Now we can parse the output
    my @array_lines = get_lines(raw_output=>$output,first_line=>$first_line);
    my (@Res,$value);
    my $i = 0;
    foreach $value(@array_lines) {
        if($value =~/.*\:\s+(\w+)\s+(\w+)\s+(\w+)\s+(\w+)\s+(\w+)\s+(\w+)\s+(\w+)\s+(\w+)\s+(\w+)\s+(\w+)\s+(\w+)\s+(\w+)\s+(\w+)\s+(\w+)\s+(\w+)/) {
            $Res[$i]->{"frames_transmitted"}    = $1;
            $Res[$i]->{"frames_received"}       = $2;
            $Res[$i]->{"encoded_errors_in"}     = $3;
            $Res[$i]->{"crc_errors"}            = $4;
            $Res[$i]->{"crc_err_eof"}           = $5;
            $Res[$i]->{"too_short"}             = $6;
            $Res[$i]->{"too_long"}              = $7;
            $Res[$i]->{"bad_eof"}               = $8;
            $Res[$i]->{"encoded_errors_out"}    = $9;
            $Res[$i]->{"c3_frames_disc"}        = $10;
            $Res[$i]->{"link_failures"}         = $11;
            $Res[$i]->{"sync_loss"}             = $12;
            $Res[$i]->{"signal_loss"}           = $13;
            $Res[$i]->{"frjt_frames"}           = $14;
            $Res[$i]->{"fbsy_frames"}           = $15;
            $i++;
        }
    }
    return \@Res;
}
1;
