#$Id$
#
#Copyright (c) 2011 Network Appliance, Inc.
#All rights reserved.
#
## @summary Routines to parse BrocadeFC portshow command
##
## @description
##    This is a library to parse the raw output of BrocadeFC portshow  command.
##
## @synopsis
##  use NACL::APISet::Switch::CLI::BrocadeFC::default::portshow
##  
##  portshow(%args)
##
## @author dl-nacl-dev@netapp.com
##
## @status public
#
package NACL::APISet::Switch::CLI::BrocadeFC::default::portshow;

use strict;
use NACL::APISet::Response::CLI::ParserUtils;

#Include exceptions which will be thrown
use NACL::APISet::Exceptions::CommandFailedException qw(:try);
use NACL::APISet::Exceptions::MissingRequiredParamException qw(:try);
use NACL::APISet::Exceptions::InvalidParamException qw(:try);
use NACL::APISet::Exceptions::InvalidParamValueException qw(:try);

sub _check_error {
    my %args = @_;
    my $raw_output = $args{raw_output};
    my $command = $args{command};
    my $error_message = $args{error_message};
    my $error = "CommandFailed"; # we'll try to classify it if possible
    $$error_message = "Command failed - Check command line and dependencies";
    if ($raw_output =~ /$command.*:\s(.*)/) {
        $$error_message = $1;
        if($$error_message =~ /(Missing)/) {
            $error = "MissingRequiredParam";
        } elsif($$error_message =~ /([Ii]nvalid option)|([Uu]nrecognized option)/) {
            $error = "InvalidParam";
        } elsif($$error_message =~ /(argument to)|(can't find)/) {
            $error = "InvalidParamValue";
        }
    }
    return $error;
}

## @name ::portshow
##
## @summary Parses the raw output from the "portshow" command
##
## @description 
##    This method can be used for parsing the raw output obtained after calling 
##    the method from portshow with any of its options as described in the 
##    corresponding cdef file 'portshow.cdefs'
## @synopsis
##      portshow(%args);
## INPUTS TO METHOD CALL 'portshow' ON APISET OBJECT 
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> </tr>
## <tr> <td> No switch for this key </td> <td> 'port-num' </td> </tr> 
## <tr> <td> '-i' </td> <td> 'port-index' </td> </tr> 
## <tr> <td> No switch for this key </td> <td> 'slot-num' </td> </tr> 
## <tr> <td> '-f' </td> <td> 'ignore-invalid' </td> </tr> 
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples
##&nbsp;
##&nbsp;Command = portshow 1
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;portIndex:   1
##&nbsp;portName: 
##&nbsp;portHealth: No Fabric Watch License
##&nbsp;
##&nbsp;Authentication: None
##&nbsp;portDisableReason: None
##&nbsp;portCFlags: 0x1
##&nbsp;portFlags: 0x24b03	 PRESENT ACTIVE F_PORT G_PORT LOGICAL_ONLINE LOGIN NOELP LED ACCEPT FLOGI
##&nbsp;portType:  18.0
##&nbsp;POD Port: Port is licensed
##&nbsp;portState: 1	Online   
##&nbsp;Protocol: FC
##&nbsp;portPhys:  6	In_Sync  	portScn:   32	F_Port    
##&nbsp;port generation number:    19060
##&nbsp;state transition count:    1089       
##&nbsp;
##&nbsp;portId:    640100
##&nbsp;portIfId:    43020015
##&nbsp;portWwn:   20:01:00:05:1e:0a:c0:c0
##&nbsp;portWwn of device(s) connected:
##&nbsp;	50:0a:09:82:87:59:82:bb
##&nbsp;Distance:  normal
##&nbsp;portSpeed: N8Gbps
##&nbsp;
##&nbsp;LE domain: 0
##&nbsp;FC Fastwrite: OFF
##&nbsp;Interrupts:        0          Link_failure: 265        Frjt:         0          
##&nbsp;Unknown:           0          Loss_of_sync: 555        Fbsy:         0          
##&nbsp;Lli:               6800       Loss_of_sig:  658        
##&nbsp;Proc_rqrd:         18748      Protocol_err: 0          
##&nbsp;Timed_out:         0          Invalid_word: 534218226  
##&nbsp;Rx_flushed:        0          Invalid_crc:  413        
##&nbsp;Tx_unavail:        0          Delim_err:    14         
##&nbsp;Free_buffer:       0          Address_err:  0          
##&nbsp;Overrun:           0          Lr_in:        369        
##&nbsp;Suspended:         0          Lr_out:       273        
##&nbsp;Parity_err:        0          Ols_in:       272        
##&nbsp;2_parity_err:      0          Ols_out:      369        
##&nbsp;CMI_bus_err:       0          
##&nbsp;
##&nbsp;Port part of other ADs: No
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;          {
##&nbsp;            'lli' => '6800',
##&nbsp;            'portwwn' => '20:01:00:05:1e:0a:c0:c0',
##&nbsp;            'protocol' => 'FC',
##&nbsp;            'proc_rqrd' => '18748',
##&nbsp;            'distance' => 'normal',
##&nbsp;            'lr_out' => '273',
##&nbsp;            'authentication' => 'None',
##&nbsp;            'port_part_of_other_ads' => 'No',
##&nbsp;            'porthealth' => 'No Fabric Watch License',
##&nbsp;            'frjt' => '0',
##&nbsp;            'pod_port' => 'Port is licensed',
##&nbsp;            'le_domain' => '0',
##&nbsp;            'invalid_word' => '534218226',
##&nbsp;            'parity_err' => '0',
##&nbsp;            'portflags' => '0x24b03	 PRESENT ACTIVE F_PORT G_PORT LOGICAL_ONLINE LOGIN NOELP LED ACCEPT FLOGI',
##&nbsp;            'state_transition_count' => '1089',
##&nbsp;            'protocol_err' => '0',
##&nbsp;            'overrun' => '0',
##&nbsp;            'port_generation_number' => '19060',
##&nbsp;            'portifid' => '43020015',
##&nbsp;            'loss_of_sync' => '555',
##&nbsp;            'cmi_bus_err' => '0',
##&nbsp;            'portphys' => '6 In_Sync',
##&nbsp;            'lr_in' => '369',
##&nbsp;            'address_err' => '0',
##&nbsp;            'invalid_crc' => '413',
##&nbsp;            'portdisablereason' => 'None',
##&nbsp;            'rx_flushed' => '0',
##&nbsp;            'portwwn_of_devices_connected' => [
##&nbsp;                                                '50:0a:09:82:87:59:82:bb'
##&nbsp;                                              ],
##&nbsp;            'timed_out' => '0',
##&nbsp;            'link_failure' => '265',
##&nbsp;            'interrupts' => '0',
##&nbsp;            'unknown' => '0',
##&nbsp;            'porttype' => '18.0',
##&nbsp;            'loss_of_sig' => '658',
##&nbsp;            '2_parity_err' => '0',
##&nbsp;            'fc_fastwrite' => 'OFF',
##&nbsp;            'portindex' => '1',
##&nbsp;            'portscn' => '32	F_Port',
##&nbsp;            'ols_out' => '369',
##&nbsp;            'portspeed' => 'N8Gbps',
##&nbsp;            'fbsy' => '0',
##&nbsp;            'free_buffer' => '0',
##&nbsp;            'ols_in' => '272',
##&nbsp;            'delim_err' => '14',
##&nbsp;            'portcflags' => '0x1',
##&nbsp;            'suspended' => '0',
##&nbsp;            'portid' => '640100',
##&nbsp;            'tx_unavail' => '0',
##&nbsp;            'portstate' => '1	Online'
##&nbsp;          }
##&nbsp;        ];
##&nbsp;
##&nbsp;Command = portshow -i 1-2
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;portIndex:   1
##&nbsp;portName: 
##&nbsp;portHealth: No Fabric Watch License
##&nbsp;
##&nbsp;Authentication: None
##&nbsp;portDisableReason: None
##&nbsp;portCFlags: 0x1
##&nbsp;portFlags: 0x24b03	 PRESENT ACTIVE F_PORT G_PORT LOGICAL_ONLINE LOGIN NOELP LED ACCEPT FLOGI
##&nbsp;portType:  18.0
##&nbsp;POD Port: Port is licensed
##&nbsp;portState: 1	Online   
##&nbsp;Protocol: FC
##&nbsp;portPhys:  6	In_Sync  	portScn:   32	F_Port    
##&nbsp;port generation number:    19060
##&nbsp;state transition count:    1089       
##&nbsp;
##&nbsp;portId:    640100
##&nbsp;portIfId:    43020015
##&nbsp;portWwn:   20:01:00:05:1e:0a:c0:c0
##&nbsp;portWwn of device(s) connected:
##&nbsp;	50:0a:09:82:87:59:82:bb
##&nbsp;Distance:  normal
##&nbsp;portSpeed: N8Gbps
##&nbsp;
##&nbsp;LE domain: 0
##&nbsp;FC Fastwrite: OFF
##&nbsp;Interrupts:        0          Link_failure: 265        Frjt:         0          
##&nbsp;Unknown:           0          Loss_of_sync: 555        Fbsy:         0          
##&nbsp;Lli:               6800       Loss_of_sig:  658        
##&nbsp;Proc_rqrd:         18748      Protocol_err: 0          
##&nbsp;Timed_out:         0          Invalid_word: 534218226  
##&nbsp;Rx_flushed:        0          Invalid_crc:  413        
##&nbsp;Tx_unavail:        0          Delim_err:    14         
##&nbsp;Free_buffer:       0          Address_err:  0          
##&nbsp;Overrun:           0          Lr_in:        369        
##&nbsp;Suspended:         0          Lr_out:       273        
##&nbsp;Parity_err:        0          Ols_in:       272        
##&nbsp;2_parity_err:      0          Ols_out:      369        
##&nbsp;CMI_bus_err:       0          
##&nbsp;
##&nbsp;Port part of other ADs: No
##&nbsp;
##&nbsp;portIndex:   2
##&nbsp;portName: 
##&nbsp;portHealth: No Fabric Watch License
##&nbsp;
##&nbsp;Authentication: None
##&nbsp;portDisableReason: None
##&nbsp;portCFlags: 0x1
##&nbsp;portFlags: 0x4001	 PRESENT U_PORT LED
##&nbsp;portType:  18.0
##&nbsp;POD Port: Port is licensed
##&nbsp;portState: 2	Offline  
##&nbsp;Protocol: FC
##&nbsp;portPhys:  4	No_Light 	portScn:   2	Offline   
##&nbsp;port generation number:    19174
##&nbsp;state transition count:    1370       
##&nbsp;
##&nbsp;portId:    640200
##&nbsp;portIfId:    43020013
##&nbsp;portWwn:   20:02:00:05:1e:0a:c0:c0
##&nbsp;portWwn of device(s) connected:
##&nbsp;
##&nbsp;Distance:  normal
##&nbsp;portSpeed: N8Gbps
##&nbsp;
##&nbsp;LE domain: 0
##&nbsp;FC Fastwrite: OFF
##&nbsp;Interrupts:        0          Link_failure: 0          Frjt:         0          
##&nbsp;Unknown:           0          Loss_of_sync: 696        Fbsy:         0          
##&nbsp;Lli:               10738      Loss_of_sig:  829        
##&nbsp;Proc_rqrd:         15821      Protocol_err: 0          
##&nbsp;Timed_out:         0          Invalid_word: 204452270  
##&nbsp;Rx_flushed:        0          Invalid_crc:  0          
##&nbsp;Tx_unavail:        0          Delim_err:    1          
##&nbsp;Free_buffer:       0          Address_err:  0          
##&nbsp;Overrun:           0          Lr_in:        685        
##&nbsp;Suspended:         0          Lr_out:       0          
##&nbsp;Parity_err:        0          Ols_in:       0          
##&nbsp;2_parity_err:      0          Ols_out:      685        
##&nbsp;CMI_bus_err:       0          
##&nbsp;
##&nbsp;Port part of other ADs: No
##&nbsp;
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;          {
##&nbsp;            'protocol' => 'FC',
##&nbsp;            'portwwn' => '20:01:00:05:1e:0a:c0:c0',
##&nbsp;            'lli' => '6800',
##&nbsp;            'distance' => 'normal',
##&nbsp;            'proc_rqrd' => '18748',
##&nbsp;            'authentication' => 'None',
##&nbsp;            'lr_out' => '273',
##&nbsp;            'port_part_of_other_ads' => 'No',
##&nbsp;            'porthealth' => 'No Fabric Watch License',
##&nbsp;            'frjt' => '0',
##&nbsp;            'pod_port' => 'Port is licensed',
##&nbsp;            'le_domain' => '0',
##&nbsp;            'invalid_word' => '534218226',
##&nbsp;            'parity_err' => '0',
##&nbsp;            'portflags' => '0x24b03	 PRESENT ACTIVE F_PORT G_PORT LOGICAL_ONLINE LOGIN NOELP LED ACCEPT FLOGI',
##&nbsp;            'state_transition_count' => '1089',
##&nbsp;            'protocol_err' => '0',
##&nbsp;            'port_generation_number' => '19060',
##&nbsp;            'overrun' => '0',
##&nbsp;            'portifid' => '43020015',
##&nbsp;            'loss_of_sync' => '555',
##&nbsp;            'cmi_bus_err' => '0',
##&nbsp;            'portphys' => '6 In_Sync',
##&nbsp;            'lr_in' => '369',
##&nbsp;            'address_err' => '0',
##&nbsp;            'invalid_crc' => '413',
##&nbsp;            'portdisablereason' => 'None',
##&nbsp;            'rx_flushed' => '0',
##&nbsp;            'portwwn_of_devices_connected' => [
##&nbsp;                                                '50:0a:09:82:87:59:82:bb'
##&nbsp;                                              ],
##&nbsp;            'timed_out' => '0',
##&nbsp;            'link_failure' => '265',
##&nbsp;            'interrupts' => '0',
##&nbsp;            'unknown' => '0',
##&nbsp;            'porttype' => '18.0',
##&nbsp;            'loss_of_sig' => '658',
##&nbsp;            '2_parity_err' => '0',
##&nbsp;            'fc_fastwrite' => 'OFF',
##&nbsp;            'portindex' => '1',
##&nbsp;            'portscn' => '32	F_Port',
##&nbsp;            'portspeed' => 'N8Gbps',
##&nbsp;            'ols_out' => '369',
##&nbsp;            'fbsy' => '0',
##&nbsp;            'free_buffer' => '0',
##&nbsp;            'portcflags' => '0x1',
##&nbsp;            'delim_err' => '14',
##&nbsp;            'ols_in' => '272',
##&nbsp;            'suspended' => '0',
##&nbsp;            'portid' => '640100',
##&nbsp;            'portstate' => '1	Online',
##&nbsp;            'tx_unavail' => '0'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'lli' => '10738',
##&nbsp;            'portwwn' => '20:02:00:05:1e:0a:c0:c0',
##&nbsp;            'protocol' => 'FC',
##&nbsp;            'proc_rqrd' => '15821',
##&nbsp;            'distance' => 'normal',
##&nbsp;            'lr_out' => '0',
##&nbsp;            'authentication' => 'None',
##&nbsp;            'port_part_of_other_ads' => 'No',
##&nbsp;            'porthealth' => 'No Fabric Watch License',
##&nbsp;            'frjt' => '0',
##&nbsp;            'pod_port' => 'Port is licensed',
##&nbsp;            'le_domain' => '0',
##&nbsp;            'invalid_word' => '204452270',
##&nbsp;            'parity_err' => '0',
##&nbsp;            'portflags' => '0x4001	 PRESENT U_PORT LED',
##&nbsp;            'state_transition_count' => '1370',
##&nbsp;            'protocol_err' => '0',
##&nbsp;            'overrun' => '0',
##&nbsp;            'port_generation_number' => '19174',
##&nbsp;            'portifid' => '43020013',
##&nbsp;            'loss_of_sync' => '696',
##&nbsp;            'cmi_bus_err' => '0',
##&nbsp;            'portphys' => '4 No_Light',
##&nbsp;            'lr_in' => '685',
##&nbsp;            'address_err' => '0',
##&nbsp;            'invalid_crc' => '0',
##&nbsp;            'portdisablereason' => 'None',
##&nbsp;            'HASH(0x2154ee30)' => undef,
##&nbsp;            'rx_flushed' => '0',
##&nbsp;            'timed_out' => '0',
##&nbsp;            'link_failure' => '0',
##&nbsp;            'unknown' => '0',
##&nbsp;            'interrupts' => '0',
##&nbsp;            'porttype' => '18.0',
##&nbsp;            'loss_of_sig' => '829',
##&nbsp;            '2_parity_err' => '0',
##&nbsp;            'fc_fastwrite' => 'OFF',
##&nbsp;            'portindex' => '2',
##&nbsp;            'portscn' => '2	Offline',
##&nbsp;            'ols_out' => '685',
##&nbsp;            'portspeed' => 'N8Gbps',
##&nbsp;            'fbsy' => '0',
##&nbsp;            'free_buffer' => '0',
##&nbsp;            'ols_in' => '0',
##&nbsp;            'delim_err' => '1',
##&nbsp;            'portcflags' => '0x1',
##&nbsp;            'suspended' => '0',
##&nbsp;            'portid' => '640200',
##&nbsp;            'tx_unavail' => '0',
##&nbsp;            'portstate' => '2	Offline'
##&nbsp;          }
##&nbsp;        ];
##&nbsp;
sub portshow {
    my %args = @_;

    my ($error_message, @result_array);
    my $resp = $args{response_object};
    my $output = $resp->get_processed_output();
    my $command = $resp->get_actual_command();
    my $error_level = $resp->get_execution_status();
    my $error;

    if($error_level) {
        $error = _check_error(raw_output=>$output,command=>"portshow",
                              error_message=>\$error_message);
        $error = 'NACL::APISet::Exceptions::'.$error.'Exception';
        $error->throw(text=>$error_message,response_object => $resp);
    }

    my @lines = get_lines(raw_output=>$output,first_line=>'portIndex:\s');

    return [] if !@lines;

    my %hash;
    my $count = 0;
    for( my $i = 0 ; $i <= $#lines ; $i++) {
        my ($key, $value);
        my $line = $lines[$i];
        chomp($line);
        if ($line =~ /^portIndex/ and $count != 0) {
            # Need to make local copy of the hash because
            # hashref is pushed into the array and hash is
            # cleared after pushing its ref into the array
            my %tmp_hash = %hash;
            push (@result_array, \%tmp_hash);
            %hash = {};
        }

        # handling the case where value for this key is provided in next line
        if ($line =~ /^(portWwn of device\(s\) connected):$/) {
            $key = to_key_name(item => $1);
            while ($lines[++$i] =~ /\w\w\:\w\w\:\w\w\:\w\w\:\w\w\:\w\w\:\w\w\:\w\w/) {
                $value = $lines[$i];
                push(@{$hash{$key}}, $value);
            }
            $i--;
        } else {
            # This will handle both single and multiple key,value pair per line
            my @tmp_arr = split(/:\s+/, $line);
            $key = to_key_name(item => shift(@tmp_arr));
            for(my $j = 0; $j <= $#tmp_arr; $j++) {
                if ($j == $#tmp_arr) {
                    $value = pop(@tmp_arr);
                    $hash{$key} = $value;
                    last;
                }
                my @tmp_arr2 = split(/\s+/, $tmp_arr[$j]);
                my $next_key = to_key_name(item => pop(@tmp_arr2));
                $value = join(" ", @tmp_arr2);
                $hash{$key} = $value;
                $key = $next_key;
            }
        }
        $count = 1;
    }
    push (@result_array, \%hash);

    return \@result_array;
}

1;
