#$Id$
#
#Copyright (c) 2012 Network Appliance, Inc.
#All rights reserved.
#
## @summary Routines to parse BrocadeFC switchenable, switchdisable commands
##
## @description
##    This is a library to parse the raw output of BrocadeFC switch switchenable, switchdisable commands.
##
## @synopsis
##  use NACL::APISet::Switch::CLI::BrocadeFC::default::switch
##
##  switchenable(%args), switchdisable(%args)
##
## @author jais@netapp.com, dl-nacl-dev@netapp.com
##
## @status public
#
package NACL::APISet::Switch::CLI::BrocadeFC::default::switch;

use strict;

#Include exceptions which will be thrown
use NACL::APISet::Exceptions::InvalidParamException;

use NACL::APISet::Response::CLI::ParserUtils;

sub _check_error {
    my %args = @_;

    my $raw_output = $args{raw_output};
    my $resp = $args{"response_object"};
    my $error_message;
    my $error;

    # Check for the error
    if ($raw_output =~ /([Ii]nvalid input parameters)/) {
        $error = "InvalidParam";
        $error_message = $1;
    }
    if ($error) {
        $error = 'NACL::APISet::Exceptions::'.$error.'Exception';
        $error->throw(text => $error_message, response_object => $resp);
    }
}

## @name ::switchenable
##
## @summary Parses the raw output from the "switchenable" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from switchenable with any of its options as described in the
##    corresponding cdef file 'switch.cdefs'
## @synopsis
##      switchenable(%args);
## INPUTS TO METHOD CALL 'switchenable' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

##&nbsp;Command = switchenable
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;dpg-mcc-6510-01:FID128:adminw
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [];
##&nbsp;
sub switchenable {
    my %args = @_;

    my $resp         = $args{response_object};
    my $raw_output   = $resp->get_processed_output();

    _check_error(raw_output=>$raw_output, response_object=>$resp);

    return [];
}

## @name ::switchdisable
##
## @summary Parses the raw output from the "switchdisable" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from switchdisable with any of its options as described in the
##    corresponding cdef file 'switch.cdefs'
## @synopsis
##      switchdisable(%args);
## INPUTS TO METHOD CALL 'switchdisable' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

##&nbsp;Command = switchdisable
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;dpg-mcc-6510-01:FID128:admin>
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [];
##&nbsp;
sub switchdisable {
    my %args = @_;

    my $resp         = $args{response_object};
    my $raw_output   = $resp->get_processed_output();

    _check_error(raw_output=>$raw_output, response_object=>$resp);

    return [];
}

## @name ::switchname
##
## @summary Parses the raw output from the "switchname" command
##
## @description 
##    This method can be used for parsing the raw output obtained after calling 
##    the method from switchname with any of its options as described in the 
##    corresponding cdef file 'switch.cdefs'
## @synopsis
##      switchname(%args);
## INPUTS TO METHOD CALL 'switchname' ON APISET OBJECT 
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

##&nbsp;Command = switchname
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;; echo __status__=$?
##&nbsp;dpg-mcc-brcd6510-01-A-fabX
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;          {
##&nbsp;            'switchname' => 'dpg-mcc-brcd6510-01-A-fabX'
##&nbsp;          }
##&nbsp;        ];
##&nbsp;
sub switchname {
    my %args = @_;

    my $resp = $args{response_object};
    my $output = $resp->get_processed_output();
    my $command = $resp->get_command();

    # Status returned by switch for this command is
    # is ZERO even if command failed
    _check_error(raw_output=>$output, response_object=>$resp);
    
    my @lines = get_lines(raw_output => $output, can_throw => 1);
    my @result_array = {switchname => $lines[0]};
    
    return \@result_array;
}
1;
