#
#Copyright (c) 2013 NetApp, Inc.
#All rights reserved.
#
## @summary Routines to parse BrocadeFC switchshow command
##
## @description
##    This is a library to parse the raw output of BrocadeFC switch switchshow command.
##
## @synopsis
##  use NACL::APISet::Switch::CLI::BrocadeFC::default::switchshow
##
##  switchshow(%args)
##
## @author dl-nacl-dev@netapp.com
##
## @status public
#
package NACL::APISet::Switch::CLI::BrocadeFC::default::switchshow;

use strict;
use NACL::APISet::Response::CLI::ParserUtils;

#Include exceptions which will be thrown
use NACL::APISet::Exceptions::CommandFailedException qw(:try);
use NACL::APISet::Exceptions::MissingRequiredParamException;
use NACL::APISet::Exceptions::InvalidParamException;
use NACL::APISet::Exceptions::InvalidParamValueException;

sub _check_error {
    my %args = @_;

    my $raw_output    = $args{raw_output};
    my $command       = $args{command};
    my $error_message = $args{error_message};
    my $error;

    # Check for the error
    if (   ( $raw_output =~ /$command.*:\s(.*)/ )
        or ( $raw_output =~ /([Ii]nvalid\s.*)$/ ) )
    {
        $$error_message = $1;
        if ( $$error_message =~ /(Missing)/ ) {
            $error = "MissingRequiredParam";
        } elsif ( $$error_message
            =~ /([Ii]nvalid option)|([Uu]nrecognized option)|([Ii]nvalid input)/
            )
        {
            $error = "InvalidParam";
        } elsif ( $$error_message =~ /(argument to)|(can't find)/ ) {
            $error = "InvalidParamValue";
        }
    }
    return $error;
}

## @name ::switchshow
##
## @summary Parses the raw output from the "switchshow" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from switchshow with any of its options as described in the
##    corresponding cdef file 'switchshow.cdefs'
## @synopsis
##      switchshow(%args);
## INPUTS TO METHOD CALL 'switchshow' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## <tr> <td> '-portname' </td> <td> 'port-name' </td> <td> '1' </td> </tr>
## <tr> <td> '-slot' </td> <td> 'slot-number' </td> <td> '0' </td> </tr>
## <tr> <td> '-portcount' </td> <td> 'port-count' </td> <td> '1' </td> </tr>
## <tr> <td> '-iscsi' </td> <td> 'iscsi-sessions' </td> <td> '1' </td> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples
##&nbsp;
##&nbsp;Command = switchshow
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;switchName:	dpg-mcc-brcd6510-01-A-fabY
##&nbsp;switchType:	109.1
##&nbsp;switchState:	Online
##&nbsp;switchMode:	Native
##&nbsp;switchRole:	Subordinate
##&nbsp;switchDomain:	2
##&nbsp;switchId:	fffc02
##&nbsp;switchWwn:	10:00:00:05:33:91:27:8e
##&nbsp;zoning:		ON (fcvi_storage_cfg_fabY)
##&nbsp;switchBeacon:	OFF
##&nbsp;FC Router:	OFF
##&nbsp;Allow XISL Use:	OFF
##&nbsp;LS Attributes:	[FID: 128, Base Switch: No, Default Switch: Yes, Address Mode 0]
##&nbsp;
##&nbsp;Index Port Address Media Speed State     Proto
##&nbsp;==============================================
##&nbsp;   0   0   020000   id     N8   Online      FC  F-Port  21:00:00:24:ff:32:06:90
##&nbsp;   1   1   020100   id     N4   Online      FC  F-Port  21:00:00:1b:32:8e:ba:8d
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;          {
##&nbsp;            'switchmode' => 'Native',
##&nbsp;            'zoning' => 'ON (fcvi_storage_cfg_fabY)',
##&nbsp;            'base_switch' => 'No',
##&nbsp;            'ports-information' => [
##&nbsp;                                     {
##&nbsp;                                       'proto' => 'FC',
##&nbsp;                                       'wwpn' => '21:00:00:24:ff:32:06:90',
##&nbsp;                                       'port' => '0',
##&nbsp;                                       'state' => 'Online',
##&nbsp;                                       'index' => '0',
##&nbsp;                                       'media' => 'id',
##&nbsp;                                       'speed' => 'N8',
##&nbsp;                                       'porttype' => 'F-Port',
##&nbsp;                                       'address' => '020000'
##&nbsp;                                     },
##&nbsp;                                     {
##&nbsp;                                       'proto' => 'FC',
##&nbsp;                                       'wwpn' => '21:00:00:1b:32:8e:ba:8d',
##&nbsp;                                       'port' => '1',
##&nbsp;                                       'state' => 'Online',
##&nbsp;                                       'index' => '1',
##&nbsp;                                       'media' => 'id',
##&nbsp;                                       'speed' => 'N4',
##&nbsp;                                       'porttype' => 'F-Port',
##&nbsp;                                       'address' => '020100'
##&nbsp;                                     }
##&nbsp;                                   ],
##&nbsp;            'address_mode' => '0',
##&nbsp;            'use' => 'OFF',
##&nbsp;            'fid' => '128',
##&nbsp;            'switchid' => 'fffc02',
##&nbsp;            'switchwwn' => '10:00:00:05:33:91:27:8e',
##&nbsp;            'switchname' => 'dpg-mcc-brcd6510-01-A-fabY',
##&nbsp;            'switchrole' => 'Subordinate',
##&nbsp;            'router' => 'OFF',
##&nbsp;            'switchbeacon' => 'OFF',
##&nbsp;            'switchtype' => '109.1',
##&nbsp;            'switchdomain' => '2',
##&nbsp;            'switchstate' => 'Online',
##&nbsp;            'attributes' => '[FID: 128, Base Switch: No, Default Switch: Yes, Address Mode 0]',
##&nbsp;            'default_switch' => 'Yes'
##&nbsp;          }
##&nbsp;        ];
##&nbsp;
##&nbsp;Command = switchshow -portcount
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;FC ports =  24
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;          {
##&nbsp;            'fc_ports' => '24'
##&nbsp;          }
##&nbsp;        ];
##&nbsp;
sub switchshow {
    my %args = @_;

    my ( $error_message, @result_array );
    my $resp        = $args{response_object};
    my $output      = $resp->get_processed_output();
    my $command     = $resp->get_command();
    my $error_level = $resp->get_execution_status();
    my ( $error, $key, $value );

    # Status returned by switch for this command is
    # is ZERO even if command failed
    $error = _check_error(
        raw_output    => $output,
        command       => "switchshow",
        error_message => \$error_message
    );
    if ($error) {
        $error = 'NACL::APISet::Exceptions::' . $error . 'Exception';
        $error->throw( text => $error_message, response_object => $resp );
    }

    my %hash;
    if ( $command =~ /\-portcount|\-iscsi/ ) {
        my @lines
            = get_lines( raw_output => $output, first_line => 'FC ports' );
        return [] if !@lines;
        foreach my $line (@lines) {
            chomp($line);
            my @tmp_arr = split( /\s+=\s+/, $line );
            $key = to_key_name( item => $tmp_arr[0] );
            $hash{$key} = $tmp_arr[1];
        }
        push( @result_array, \%hash );
    } else {
        my @lines = get_lines( raw_output => $output,
            first_line => 'switchName:\s' );
        return [] if !@lines;

        foreach my $line (@lines) {
            chomp($line);
            if ( $line =~ /(\w+):\s+(\S.*)$/ ) {
                $key        = to_key_name( item => $1 );
                $value      = $2;
                $hash{$key} = $value;
                if ( $key eq 'attributes' ) {
                    my @values = split( /, /, $value );
                    $values[0] =~ s/\[//;
                    $values[$#values] =~ s/\]//;
                    foreach my $var (@values) {
                        if ( $var =~ /.*[:].*/ ) {
                            my @values = split( /: /, $var );
                            $hash{ to_key_name( item => $values[0] ) }
                                = $values[1];
                        } elsif ( $var =~ /(.*)\s+(\d+)/ ) {
                            $hash{ to_key_name( item => $1 ) } = $2;
                        }
                    }
                }
            } else {
                last;
            }
        }

        @lines = get_lines( raw_output => $output, first_line => 'Index' );

        # First line will have all the headings.
        my $headings = shift(@lines);

        # Getting rid of line unwanted line
        shift(@lines);
        my $count = 0;
        my %port_info;
        my @port_infos;

        my @headings_arr = split( /\s+/, $headings );

        #Comparing is speed heading is there, if not there then it will
        #parse the output of show only port command
        if ( $headings =~ /Speed/i ) {

            # Find out number of head keys till Speed key.This is done because
            #the head keys are not at fixed position always
            foreach my $hdkey (@headings_arr) {
                if ( $hdkey =~ /Speed/i ) {
                    last;
                } else {
                    $count++;
                }
            }
            foreach my $line (@lines) {
                my %port_info;

                #To get rid of the last line in the raw output.
                if ( $line =~ /__status__/ ) {
                    next;
                }

                #whereever may the distance position find it and clear
                #that string poisition
                if ( $line =~ /LS|LE|LD|L\d+|L\d+\.\d+/ ) {
                    $line =~ s/(L\S+)//;
                    $port_info{'distance'} = $1;
                }
                my @port_info = split( '\s+', $line );

                # Join the splited term for the last column "proto".
                my $protoline = $port_info[ $count + 2 ];
                for ( my $k = $count + 3; $k < @port_info; $k++ ) {
                    $protoline = join( ' ', $protoline, $port_info[$k] );
                }

               # For all the head keys except porto key create hash structure.
                for ( my $k = 0; $k <= ( $count + 1 ); $k++ ) {
                    my $key = to_key_name( item => $headings_arr[$k] );
                    $port_info{$key} = $port_info[$k];

                    # Check is the first term FC is there in porto column.
                    if ( $protoline =~ /FC/ ) {
                        $protoline =~ s/(FC)//;
                        $port_info{'proto'} = $1;
                    }

                    #parse remaining lines
                    if ( $protoline =~ /\((No POD.*)\)\s+(\S+)$/ ) {
                        $port_info{'porttype'} = $1;
                        $port_info{'wwpn'}     = $2;
                    } elsif ( $protoline =~ /(\w+\-\w+)\s+(\w+\:.*\:\w+)/ ) {
                        $port_info{'porttype'} = $1;
                        $port_info{'wwpn'}     = $2;
                    } elsif ( $protoline =~ /^\s*(\w+\-\w+)/ || $protoline =~ /(\w+)/ ) {
                        $port_info{'porttype'} = $1;
                    }
                }
                push( @port_infos, \%port_info );
            }
            $hash{'ports-information'} = \@port_infos;
            push( @result_array, \%hash );
        } else {

        #This Section of code is for parsing output for show port only command
            foreach my $line (@lines) {
                chomp($line);
                my %port_info;
                my @port_info = split( /\s+/, $line );
                for ( my $i; $i <= $#headings_arr; $i++ ) {
                    my $key = to_key_name( item => $headings_arr[$i] );
                    $port_info{$key} = shift @port_info;
                }
                push( @port_infos, \%port_info );
            }
            $hash{'ports-information'} = \@port_infos;
            push( @result_array, \%hash );
        }
    }

    return \@result_array;
}

1;
