#
#Copyright (c) 2011 Network Appliance, Inc.
#All rights reserved.
#
## @summary Routines to parse BrocadeFC version command
##
## @description
##    This is a library to parse the raw output of BrocadeFC switch version command.
##
## @synopsis
##  use NACL::APISet::Switch::CLI::BrocadeFC::default::version
##
##  version(%args)
##
## @author Govardhan Yadav (yadag@netapp.com),dl-nacl-dev@netapp.com
##
## @status public
#
package NACL::APISet::Switch::CLI::BrocadeFC::default::version;

use strict;
use NACL::APISet::Response::CLI::ParserUtils;

#Include exceptions which will be thrown
use NACL::APISet::Exceptions::CommandFailedException qw(:try);
use NACL::APISet::Exceptions::MissingRequiredParamException;
use NACL::APISet::Exceptions::InvalidParamValueException;

sub _check_error {
    my %args = @_;

    my $raw_output = $args{raw_output};
    my $command = $args{command};
    my $error_message = $args{error_message};

    return undef;
}


## @name ::version
##
## @summary Parses the raw output from the "version" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from version with any of its options as described in the
##    corresponding cdef file 'version.cdefs'
## @synopsis
##      version(%args);
## INPUTS TO METHOD CALL 'version' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

##&nbsp;Command = version
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;          {
##&nbsp;            'bootprom' => '4.6.6',
##&nbsp;            'fabric_os' => 'v6.2.2',
##&nbsp;            'kernel' => '2.6.14.2',
##&nbsp;            'made_on' => 'Mon Nov 16 17:54:35 2009',
##&nbsp;            'flash' => 'Tue Feb 9 10:20:52 2010'
##&nbsp;          }
##&nbsp;        ];
##&nbsp;
sub version {
    my %args = @_;
    my ($error_message, @result_array);
    my $resp = $args{response_object};
    my $output = $resp->get_processed_output();
    my $command = $resp->get_command();
    my $error_level = $resp->get_execution_status();
    my ($error, $key);

    $error = _check_error(raw_output=>$output,command=>"version",
                             error_message=>\$error_message);
    if($error) {
        $error = 'NACL::APISet::Exceptions::'.$error.'Exception';
        $error->throw(text=>$error_message,response_object => $resp);
    }
    #Now we can parse the output
    my @lines = get_lines("raw_output"=>$output,"response_object"=>$resp);
    my (@Res,$value);
    my $i = 0;
    foreach $value (@lines){
        if($value =~/[\#\>]/){
            next;
        } elsif($value =~/^([\w\s]+):\s+(.*)$/){
            $Res[$i]->{to_key_name(item => $1)} = $2;
        }
    }
    return \@Res;
}

1;
