#
#Copyright (c) 2011 Network Appliance, Inc.
#All rights reserved.
#
## @summary Routines to parse BrocadeFC zone command
##
## @description
##    This is a library to parse the raw output of BrocadeFC switch zone command.
##
## @synopsis
##  use NACL::APISet::Switch::CLI::BrocadeFC::default::zone
##
##  zone_create(%args);
##  zone_add(%args);
##  zone_show(%args);
##  zone_delete(%args);
##
## @author dl-nacl-dev@netapp.com,veerr@netapp.com
##
## @status public
#
package NACL::APISet::Switch::CLI::BrocadeFC::default::zone;

use strict;
use warnings;
use NACL::APISet::Response::CLI::ParserUtils;

#Include exceptions which will be thrown
use NACL::APISet::Exceptions::CommandFailedException qw(:try);
use NACL::APISet::Exceptions::InvalidParamValueException;
use NACL::APISet::Exceptions::InvalidParamException;

## @name ::zone_create
##
## @summary Parses the raw output from the "zone --create" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from zone_create with any of its options as described in the
##    corresponding cdef file 'zone.cdefs'
## @synopsis
##      zone_create(%args);
## INPUTS TO METHOD CALL 'zone_create' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> </tr>
## <tr> <td> '-t' </td> <td> 'object-type' </td> </tr>
## <tr> <td> '-p' </td> <td> 'port-list' </td> </tr>
## <tr> <td> No switch for this key </td> <td> 'zone-name' </td> </tr>
## <tr> <td> '-o' </td> <td> 'optlist' </td> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

##&nbsp;Command = zone --create -t ti purplezone1 -p " 1,1; 2,4; 1,8; 2,6"
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [];
##&nbsp;
##&nbsp;Command = zone --create -t ti purplezone12 -p "3,3;3,4"
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [];
##&nbsp;
sub zone_create {
    my %args = @_;
    my $error_message;
    my $resp       = $args{response_object};
    my $raw_output = $resp->get_processed_output();
    my $error = _check_error("raw_output"        => $raw_output,
                             "command"           => 'zone create',
                             "ptr_error_message" => \$error_message,);
    if ($error) {
        $error = 'NACL::APISet::Exceptions::' . $error . 'Exception';
        $error->throw(text => $error_message, response_object => $resp);
    }
    return [];
} ## end sub zone_create

## @name ::zone_add
##
## @summary Parses the raw output from the "zone --add" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from zone_add with any of its options as described in the
##    corresponding cdef file 'zone.cdefs'
## @synopsis
##      zone_add(%args);
## INPUTS TO METHOD CALL 'zone_add' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> </tr>
## <tr> <td> '-p' </td> <td> 'port-list' </td> </tr>
## <tr> <td> No switch for this key </td> <td> 'zone-name' </td> </tr>
## <tr> <td> '-o' </td> <td> 'optlist' </td> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

##&nbsp;Command = zone --add purplezone1 -p "1,4; 1,6"
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [];
##&nbsp;
sub zone_add {
    my %args = @_;
    my $error_message;
    my $resp       = $args{response_object};
    my $raw_output = $resp->get_processed_output();
    my $error = _check_error("raw_output"        => $raw_output,
                             "command"           => 'zone add',
                             "ptr_error_message" => \$error_message,);
    if ($error) {
        $error = 'NACL::APISet::Exceptions::' . $error . 'Exception';
        $error->throw(text => $error_message, response_object => $resp);
    }
    return [];
} ## end sub zone_add

## @name ::zone_show
##
## @summary Parses the raw output from the "zone --show" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from zone_show with any of its options as described in the
##    corresponding cdef file 'zone.cdefs'
## @synopsis
##      zone_show(%args);
## INPUTS TO METHOD CALL 'zone_show' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> </tr>
## <tr> <td> No switch for this key </td> <td> 'zone-name' </td> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

##&nbsp;Command = zone --show purplezone
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;             {
##&nbsp;                 'port_list' => '1,4; 1,6; 1,1; 2,4; 1,8; 2,6',
##&nbsp;                 'configured_status' => 'Activated / Failover-Enabled',
##&nbsp;                 'enabled_status' => 'Deactivated',
##&nbsp;                 'ti_zone_name' => 'purplezone1'
##&nbsp;             },
##&nbsp;             {
##&nbsp;                 'port_list' => '3,3; 3,4',
##&nbsp;                 'configured_status' => 'Activated / Failover-Enabled',
##&nbsp;                 'enabled_status' => 'Deactivated',
##&nbsp;                 'ti_zone_name' => 'purplezone12'
##&nbsp;             }
##&nbsp;        ];
##&nbsp;
sub zone_show {
    my %args = @_;
    my $error_message;
    my $resp       = $args{response_object};
    my $raw_output = $resp->get_processed_output();
    my $error = _check_error("raw_output"        => $raw_output,
                             "command"           => 'zone show',
                             "ptr_error_message" => \$error_message,);
    if ($error) {
        $error = 'NACL::APISet::Exceptions::' . $error . 'Exception';
        $error->throw(text => $error_message, response_object => $resp);
    }
    my @result_array;
    my ($i, $j, $name);
    my @lines = get_lines(raw_output      => $raw_output,
                          can_throw       => 1,
                          response_object => $resp);
    $j    = -1;
    foreach my $line (@lines) {
        if ($line =~ /(TI Zone Name):\s+(\w+)/) {
            $j++;
            $result_array[$j]->{to_key_name(item => $1)} = $2;
        } elsif ($line =~ /(Port List):\s+([0-9\,\;\s]+)/) {
            $result_array[$j]->{to_key_name(item => $1)} = $2;
        } elsif ($line =~ /(Configured Status):\s+([A-Za-z\s\/\-]+)/) {
            $result_array[$j]->{to_key_name(item => $1)} = $2;
        } elsif ($line =~ /(Enabled Status):\s+(\w+)/) {
            $result_array[$j]->{to_key_name(item => $1)} = $2;
        } elsif ($line =~ /vgbr200es2:admin/) {
            next;
        } elsif ($line =~ /(no TI zone configuration defined)/) {
            $result_array[$j] = $1;
        }
    } ## end foreach my $line (@lines)
    return \@result_array;
} ## end sub zone_show

## @name ::zone_delete
##
## @summary Parses the raw output from the "zone --delete" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from zone_delete with any of its options as described in the
##    corresponding cdef file 'zone.cdefs'
## @synopsis
##      zone_delete(%args);
## INPUTS TO METHOD CALL 'zone_delete' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> </tr>
## <tr> <td> No switch for this key </td> <td> 'zone-name' </td> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

##&nbsp;Command = zone --delete purplezone1
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [];
##&nbsp;
sub zone_delete {
    my %args = @_;
    my $error_message;
    my $resp       = $args{response_object};
    my $raw_output = $resp->get_processed_output();
    my $error = _check_error("raw_output"        => $raw_output,
                             "command"           => 'zone delete',
                             "ptr_error_message" => \$error_message,);
    if ($error) {
        $error = 'NACL::APISet::Exceptions::' . $error . 'Exception';
        $error->throw(text => $error_message, response_object => $resp);
    }
    return [];
} ## end sub zone_delete

sub _check_error {
    my %args          = @_;
    my $raw_output    = $args{"raw_output"};
    my $command       = $args{"command"};
    my $error_message = $args{"ptr_error_message"};
    my $error         = undef;
    if ($raw_output =~ /[Uu]sage:/) {
        $$error_message = $raw_output;
        $error          = "CommandFailed";
    } elsif ($raw_output =~
             /(Name \w+ not found|Name exists error|TI Zone create Failed)/) {
        $$error_message = $1;
        $error          = "InvalidParamValue";
    } elsif ($raw_output =~ /(AD invalid parameter)/) {
        $$error_message = $1;
        $error          = "InvalidParam";
    }
    return $error;
} ## end sub _check_error
1;
