#
#Copyright (c) 2011-2012 Network Appliance, Inc.
#All rights reserved.
#
## @summary Routines to parse CiscoFC ntp command
##
## @description
##    This is a library to parse the raw output of CiscoFC ntp command.
##
## @synopsis
##  use NACL::APISet::Switch::CLI::CiscoFC::default::ntp
##
##  ntp(%args)
##
## @author skotagir@netapp.com,dl-nacl-dev@netapp.com
##
## @status public
#
package NACL::APISet::Switch::CLI::CiscoFC::default::ntp;

use strict;
use NACL::APISet::Response::CLI::ParserUtils;

#Include exceptions which will be thrown
use NACL::APISet::Exceptions::CommandFailedException qw(:try);
use NACL::APISet::Exceptions::MissingRequiredParamException qw(:try);
use NACL::APISet::Exceptions::InvalidParamException qw(:try);
use NACL::APISet::Exceptions::InvalidParamValueException qw(:try);

sub _check_error {
    my $error;
    my %args          = @_;
    my $raw_output    = $args{raw_output};
    my $command       = $args{command};
    my $error_message = $args{error_message};
    if ( $raw_output =~ /Invalid\s+command/ ) {
        $error          = "InvalidParam";
        $error_message = $raw_output;
    } elsif ( $raw_output =~ /Cannot\s+resolve\s+the\s+domain\s+name/ ) {
        $error          = "InvalidParamValue";
        $error_message = $raw_output;
    } elsif ( $raw_output
        =~ /ERROR:\s+Maximum\s+polling\s+interval\s+is\s+less\s+than\s+minimum./
        ) {
        $error          = "InvalidParamValue";
        $error_message = $raw_output;
    }
    return $error;
} ## end sub _check_error

## @name ::ntp_server
##
## @summary Parses the raw output from the "ntp server" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from ntp_server with any of its options as described in the
##    corresponding cdef file 'ntp.cdefs'
## @synopsis
##      ntp_server(%args);
## INPUTS TO METHOD CALL 'ntp_server' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> </tr>
## <tr> <td> No switch for this key </td> <td> 'minpoll' </td> </tr>
## <tr> <td> No switch for this key </td> <td> 'maxpoll' </td> </tr>
## <tr> <td> 'prefer' </td> <td> 'prefer' </td> </tr>
## <tr> <td> No switch for this key </td> <td> 'host-name-ip' </td> </tr>
## <tr> <td> No switch for this key </td> <td> 'key' </td> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

sub ntp_server {
    my %args = @_;
    my ( $error_message, @result_array );
    my $resp        = $args{response_object};
    my $output      = $resp->get_processed_output();
    my $command     = $resp->get_command();
    my $error_level = $resp->get_execution_status();
    my $error       = _check_error(
        raw_output    => $output,
        error_message => \$error_message
    );
    if ($error) {
        $error = 'NACL::APISet::Exceptions::' . $error . 'Exception';
        $error->throw( text => $error_message, response_object => $resp );
    }
    return [];
} ## end sub ntp_server

## @name ::ntp_sync_retry
##
## @summary Parses the raw output from the "ntp sync-retry" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from ntp_sync_retry with any of its options as described in the
##    corresponding cdef file 'ntp.cdefs'
## @synopsis
##      ntp_sync_retry(%args);
## INPUTS TO METHOD CALL 'ntp_sync_retry' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an empty array reference.

sub ntp_sync_retry {
    my %args = @_;
    my ( $error_message, @result_array );
    my $resp        = $args{response_object};
    my $output      = $resp->get_processed_output();
    my $error_level = $resp->get_execution_status();
    my $error       = _check_error(
        raw_output    => $output,
        error_message => \$error_message
    );
    if ($error) {
        $error = 'NACL::APISet::Exceptions::' . $error . 'Exception';
        $error->throw( text => $error_message, response_object => $resp );
    }
    if ($error_level) {
        NACL::APISet::Exceptions::CommandFailedException->throw(
            text => "Execution status was $error_level "
                . "indicating that the command has failed",
            response_object => $resp
        );
    }
    return [];
} ## end sub ntp_sync_retry
1;

