#
#Copyright (c) 2011 Network Appliance, Inc.
#All rights reserved.
#
## @summary Routines to parse CiscoFC switch show_fcs_database command
##
## @description
##    This is a library to parse the raw output of CiscoFC switch show_fcs_database command.
##    The function name will be the same as command name with underscores being
##    used instead of spaces.
##
## @synopsis
##  use NACL::APISet::Switch::CLI::CiscoFC::default::show_fcs_database
##
##  show_fcs_database(%args)
##
## @author rroshan@netapp.com
##
## @status
#
package NACL::APISet::Switch::CLI::CiscoFC::default::show_fcs_database;

use NACL::APISet::Response::CLI::ParserUtils;

#Include exceptions which will be thrown
use NACL::APISet::Exceptions::CommandFailedException qw(:try);
use NACL::APISet::Exceptions::InvalidParamException qw(:try);

sub _check_error {
    my %args = @_;
    my $raw_output = $args{raw_output};
    my $command = $args{command};
    my $error_message = $args{error_message};
    my $error ;

    #Check for the error
    if($raw_output =~ /(Erro.* )/) {
        $error = "CommandFailed";
        $$error_message = $1;
    } elsif ($raw_output =~ /(.*Invalid.* )/) {
        $error = "InvalidParam";
        $$error_message = $1;
    } elsif ($raw_output =~ /(.*Incomplete.* )/) {
        $error = "CommandFailed";
        $$error_message = $1;
    }
    return $error;
}

## @name ::show_fcs_database
##
## @summary Parses the raw output from the "show fcs database" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from show device-alias database with any of its options as described in the
##    corresponding cdef file 'show_fcns_database_local.cdefs'
## @synopsis
##      show fcalias(%args);
## INPUTS TO METHOD CALL 'show_fcs_database' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples
##&nbsp;
##&nbsp;Command = show fcs database
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;          {
##&nbsp;            'switch_ports' => [
##&nbsp;                                {
##&nbsp;                                  'interface' => 'fc1/1',
##&nbsp;                                  'type' => 'Unknown',
##&nbsp;                                  'attached_pwwn' => 'None',
##&nbsp;                                  'fwwn' => '20:01:00:0d:ec:3e:c0:80'
##&nbsp;                                },
##&nbsp;                                {
##&nbsp;                                  'interface' => 'fc1/2',
##&nbsp;                                  'type' => 'F',
##&nbsp;                                  'attached_pwwn' => '21:00:00:e0:8b:9b:7a:f1',
##&nbsp;                                  'fwwn' => '20:02:00:0d:ec:3e:c0:80'
##&nbsp;                                },
##&nbsp;                              ],
##&nbsp;            'fcs_local_database' => [
##&nbsp;                                      {
##&nbsp;                                        'switch_wwn_' => '20:01:00:0d:ec:3e:c0:81',
##&nbsp;                                        'switch_domain_id_' => '0x1b(27)',
##&nbsp;                                        'switch_logical-name_' => 'vgci9124s38',
##&nbsp;                                        'switch_mgmt_addresses' => [
##&nbsp;                                                                     'snmp://10.227.195.58/eth-ip',
##&nbsp;                                                                     'http://10.227.195.58/eth-ip'
##&nbsp;                                                                   ],
##&nbsp;                                        'switch_information_list_' => '[Cisco Systems, Inc.*DS-C9124-K9*5.0(4c)*20:00:00:0d:',
##&nbsp;                                        'fabric-name_' => '20:01:00:0d:ec:3e:c0:81'
##&nbsp;                                      }
##&nbsp;                                    ],
##&nbsp;            'vsan' => '1'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'fcs_local_database' => [
##&nbsp;                                      {
##&nbsp;                                        'switch_wwn_' => '20:03:00:0d:ec:3e:c0:81',
##&nbsp;                                        'switch_domain_id_' => '0x3d(61)',
##&nbsp;                                        'switch_logical-name_' => 'vgci9124s38',
##&nbsp;                                        'switch_mgmt_addresses' => [
##&nbsp;                                                                     'http://10.227.195.58/eth-ip',
##&nbsp;                                                                     'snmp://10.227.195.58/eth-ip'
##&nbsp;                                                                   ],
##&nbsp;                                        'switch_information_list_' => '[Cisco Systems, Inc.*DS-C9124-K9*5.0(4c)*20:00:00:0d:',
##&nbsp;                                        'fabric-name_' => '20:03:00:0d:ec:3e:c0:81'
##&nbsp;                                      }
##&nbsp;                                    ],
##&nbsp;            'vsan' => '3'
##&nbsp;          }
##&nbsp;        ];
##&nbsp;                       
##&nbsp;Command = show fcs database vsan 1
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;          {
##&nbsp;            'switch_ports' => [
##&nbsp;                                {
##&nbsp;                                  'interface' => 'fc1/1',
##&nbsp;                                  'type' => 'Unknown',
##&nbsp;                                  'attached_pwwn' => 'None',
##&nbsp;                                  'fwwn' => '20:01:00:0d:ec:3e:c0:80'
##&nbsp;                                },
##&nbsp;                                {
##&nbsp;                                  'interface' => 'fc1/2',
##&nbsp;                                  'type' => 'F',
##&nbsp;                                  'attached_pwwn' => '21:00:00:e0:8b:9b:7a:f1',
##&nbsp;                                  'fwwn' => '20:02:00:0d:ec:3e:c0:80'
##&nbsp;                                },
##&nbsp;                              ],
##&nbsp;            'fcs_local_database' => [
##&nbsp;                                      {
##&nbsp;                                        'switch_wwn_' => '20:01:00:0d:ec:3e:c0:81',
##&nbsp;                                        'switch_domain_id_' => '0x1b(27)',
##&nbsp;                                        'switch_logical-name_' => 'vgci9124s38',
##&nbsp;                                        'switch_mgmt_addresses' => [
##&nbsp;                                                                     'snmp://10.227.195.58/eth-ip',
##&nbsp;                                                                     'http://10.227.195.58/eth-ip'
##&nbsp;                                                                   ],
##&nbsp;                                        'switch_information_list_' => '[Cisco Systems, Inc.*DS-C9124-K9*5.0(4c)*20:00:00:0d:',
##&nbsp;                                        'fabric-name_' => '20:01:00:0d:ec:3e:c0:81'
##&nbsp;                                      }
##&nbsp;                                    ],
##&nbsp;            'vsan' => '1'
##&nbsp;          }
##&nbsp;        ];
##&nbsp; 

sub show_fcs_database {
    my %args = @_;
    my ($error_message, @result_array);
    my $resp = $args{response_object};
    my $output = $resp->get_processed_output();
    my $command = $resp->get_actual_command();
    my $error_level = $resp->get_execution_status();
    my ($error, $key);

    # Status returned by switch for this command is
    # is ZERO even if command failed
    $error = _check_error(raw_output=>$output,command=>$command,
                             error_message=>\$error_message);
    if($error) {
        $error = 'NACL::APISet::Exceptions::'.$error.'Exception';
        $error->throw(text=>$error_message,response_object => $resp);
    }
    my $i=-1;
    my $j= 0;
    my $k= 0;
    my ($key1 ,$key2 ,$value);
    my @arr = get_lines(raw_output=>$output,response_object=>$resp);
    foreach my $line (@arr){
        if ($line =~ /(FCS Local Database).*(VSAN)\:\s+(\d+)/){
            $i++;
            $key = to_key_name(item=>$1);
            $result_array[$i]->{to_key_name(item=>$2)} = $3;
        } elsif ($line =~ /.*fWWN                     Type.*/) {
            next;
        } elsif ($line =~ /.*Device.*/) {
            next;
        } elsif ($line =~ /((?:Switch|Fabric)[\s+-][^Mgmt].*)\s+\:\s+(.*)/) {
            $j=0;
            $value = $2 ;
            $key2 = to_key_name(item=>$1);
            $key2 =~ s/_$// ;
            $key2 =~ s/-/_/ ;
            $result_array[$i]->{$key}->[$j]->{$key2} = $value ;
        } elsif ($line =~ /((?:snmp|http).*)$/) {
            $result_array[$i]->{$key}->[$j]->{'switch_mgmt_addresses'}->[$k] = $1;
            $k++;
            if ($k == 2) {
                $k = 0;
            }
        } elsif ($line =~ /.*-----------.*/) {
            next;
        } elsif ($line =~ /(Switch.*)\:/) {
            $key1 = to_key_name(item=>$1);
        } elsif ($line =~ /(fc\S+)\s+(\S+)\s+(\S+)\s+(\S+)/) {
            $result_array[$i]->{$key1}->[$j]->{'interface'} = $1;
            $result_array[$i]->{$key1}->[$j]->{'fwwn'} = $2;
            $result_array[$i]->{$key1}->[$j]->{'type'} = $3;
            $result_array[$i]->{$key1}->[$j]->{'attached_pwwn'} = $4;
            $j++;
        }
    }
    return \@result_array;
};
1;
