#$Id: //depot/prod/test/nacldev/lib/NACL/APISet/Switch/CLI/CiscoFC/default/show_topology.pm#1 $
#
# Copyright (c) 2012 NetApp, Inc., All Rights Reserved
# Any use, modification, or distribution is prohibited
# without prior written consent from NetApp, Inc.
#
## @summary Routines to parse CiscoFC switch show_topology command
##
## @description
##    This is a library to parse the raw output of CiscoFC switch show_topology command.
##    The function name will be the same as command name with underscores being
##    used instead of spaces.
##
## @synopsis
##  use NACL::APISet::Switch::CLI::CiscoFC::default::show_topology
##
##  show_topology(%args)
##
## @author skotagir@netapp.com, dl-nacl-dev@netapp.com
##
## @status public
#

package NACL::APISet::Switch::CLI::CiscoFC::default::show_topology;

use NACL::APISet::Response::CLI::ParserUtils;

#Include exceptions which will be thrown
use NACL::APISet::Exceptions::CommandFailedException;
use NACL::APISet::Exceptions::InvalidParamException qw(:try);
use NACL::APISet::Exceptions::MissingRequiredParamException qw(:try);
use NACL::APISet::Exceptions::InvalidParamValueException qw(:try);

sub _check_error {
    my %args        = @_;
    my $raw_output  = $args{raw_output};
    my $resp        = $args{response_object};
    my $error_level = $resp->get_execution_status();
    my ( $error_message, $error );
    if ( $raw_output =~ /Invalid command/i ) {
        $error         = "InvalidParam";
        $error_message = $raw_output;
    } elsif ( $raw_output =~ /Incomplete command/i ) {
        $error         = "MissingRequiredParam";
        $error_message = $raw_output;
    } elsif ( $raw_output =~ /Invalid number/i ) {
        $error         = "InvalidParamValue";
        $error_message = $raw_output;
    }
    if ($error) {
        $error = 'NACL::APISet::Exceptions::' . $error . 'Exception';
        $error->throw( text => $error_message, response_object => $resp );
    }
    if ($error_level) {
        NACL::APISet::Exceptions::CommandFailedException->throw(
            text => "Execution status was $error_level "
                . "indicating that the command has failed",
            response_object => $resp
        );
    }

} ## end sub _check_error

## @name ::show_topology
##
## @summary Parses the raw output from the "show topology" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from show_topology with any of its options as described in the
##    corresponding cdef file 'show_topology.cdefs'
## @synopsis
##      show_topology(%args);
## INPUTS TO METHOD CALL 'show_topology' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## <tr> <td> 'vsan' </td> <td> 'vsan' </td> <td> '0' </td> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

##&nbsp;Command = show topology vsan 22
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;# show topology vsan 22
##&nbsp;
##&nbsp;FC Topology for VSAN 22 :
##&nbsp;--------------------------------------------------------------------------------
##&nbsp;       Interface  Peer Domain Peer Interface     Peer IP Address
##&nbsp;--------------------------------------------------------------------------------
##&nbsp;           fc1/15 0xd7(215)           fc1/15  10.228.160.212
##&nbsp;           fc1/16 0xd7(215)           fc1/16  10.228.160.212
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;          {
##&nbsp;            'vsan_id' => '22',
##&nbsp;            'fc_topology' => [
##&nbsp;                               {
##&nbsp;                                 'peer_domain' => '0xd7(215)',
##&nbsp;                                 'interface' => 'fc1/15',
##&nbsp;                                 'peer_ip_address' => '10.228.160.212',
##&nbsp;                                 'peer_interface' => 'fc1/15'
##&nbsp;                               },
##&nbsp;                               {
##&nbsp;                                 'peer_domain' => '0xd7(215)',
##&nbsp;                                 'interface' => 'fc1/16',
##&nbsp;                                 'peer_ip_address' => '10.228.160.212',
##&nbsp;                                 'peer_interface' => 'fc1/16'
##&nbsp;                               }
##&nbsp;                             ]
##&nbsp;          }
##&nbsp;        ];
##&nbsp;

sub show_topology {
    my %args   = @_;
    my $resp   = $args{response_object};
    my $output = $resp->get_processed_output();
    _check_error( raw_output => $output, response_object => $resp );
    my @lines = get_lines( raw_output => $output, response_object => $resp );
    my @result_array;
    my ( @head, @keys );
    my ( $c, $j ) = ( 0, -1 );

    foreach my $line (@lines) {
        if ( $line =~ /^FC Topology for VSAN (\d+) :/ ) {
            $j++;
            $c = 0;
            $result_array[$j]->{vsan_id} = $1;
        } elsif ( $line
            =~ /(Interface)\s+(Peer Domain)\s+(Peer Interface)\s+(Peer IP Address)/
            ) {
            @head = ( $1, $2, $3, $4 );
        } elsif ( $line =~ /\S+\s+\S+\(\d+\)/ ) {
            my @keys = split( /\s+/, $line );
            for ( my $i = 0; $i <= $#head; $i++ ) {
                $result_array[$j]->{fc_topology}->[$c]
                    ->{ to_key_name( item => $head[$i] ) } = $keys[$i];
            }
            $c++;
        }
    }
    return \@result_array;
} ## end sub show_topology
1;
