#
#Copyright (c) 2012 Network Appliance, Inc.
#All rights reserved.
#
## @summary Routines to parse CiscoFC showhardware
##
## @description
##    This is a library to parse the raw output of CiscoFC switch showhardware.
##
## @synopsis
##  use NACL::APISet::Switch::CLI::CiscoFC::default::showhardware
##
##  show_hardware(%args)
##
## @author dl-nacl-dev@netapp.com, veerap@netapp.com
##
## @status public
#
package NACL::APISet::Switch::CLI::CiscoFC::default::showhardware;

use strict;

use NACL::APISet::Exceptions::InvalidParamException qw(:try);

use NACL::APISet::Response::CLI::ParserUtils;

sub _check_error {
    my %args       = @_;
    my $raw_output = $args{raw_output};
    my $resp       = $args{response_object};
    my $command    = $args{command};
    my $error_message;
    my $error = undef;
    if ( $raw_output =~ /Invalid./ ) {
        $error         = "InvalidParamException";
        $error_message = "Invalid parameter";
    }
    if ($error) {
        $error = 'NACL::APISet::Exceptions::' . $error . 'Exception';
        $error->throw( text => $error_message, response_object => $resp );
    }
} ## end sub _check_error

## @name ::show_hardware
##
## @summary Parses the raw output from the "show hardware" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from show_hardware with any of its options as described in the
##    corresponding cdef file 'showhardware.cdefs'
## @synopsis
##      show_hardware(%args);
## INPUTS TO METHOD CALL 'show_hardware' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples
##&nbsp;Command = show hardware
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;Cisco Nexus Operating System (NX-OS) Software
##&nbsp;TAC support: http://www.cisco.com/tac
##&nbsp;Documents: http://www.cisco.com/en/US/products/ps9372/tsd_products_support_serie
##&nbsp;s_home.html
##&nbsp;Copyright (c) 2002-2011, Cisco Systems, Inc. All rights reserved.
##&nbsp;The copyrights to certain works contained herein are owned by
##&nbsp;other third parties and are used and distributed under license.
##&nbsp;Some parts of this software are covered under the GNU Public
##&nbsp;License. A copy of the license is available at
##&nbsp;http://www.gnu.org/licenses/gpl.html.
##&nbsp;
##&nbsp;Software
##&nbsp;  BIOS:      version 1.0.19
##&nbsp;  loader:    version N/A
##&nbsp;  kickstart: version 5.0(7)
##&nbsp;  system:    version 5.0(7)
##&nbsp;  BIOS compile time:       02/01/10
##&nbsp;  kickstart image file is: bootflash:/m9100-s2ek9-kickstart-mz.5.0.7.bin
##&nbsp;  kickstart compile time:  9/20/2011 15:00:00 [10/05/2011 06:17:51]
##&nbsp;  system image file is:    bootflash:/m9100-s2ek9-mz.5.0.7.bin
##&nbsp;  system compile time:     9/20/2011 15:00:00 [10/05/2011 08:09:39]
##&nbsp;
##&nbsp;Hardware
##&nbsp;  cisco MDS 9124 (1 Slot) Chassis ("1/2/4 Gbps FC/Supervisor-2")
##&nbsp;  Motorola, e500  with 516128 kB of memory.
##&nbsp;  Processor Board ID JAE1053JPHH
##&nbsp;
##&nbsp;  Device name: vgci9124s38
##&nbsp;  bootflash:     250368 kB
##&nbsp;Kernel uptime is 19 day(s), 18 hour(s), 36 minute(s), 32 second(s)
##&nbsp;
##&nbsp;Last reset
##&nbsp;  Reason: Unknown
##&nbsp;  System version: 5.0(7)
##&nbsp;  Service:
##&nbsp;--------------------------------
##&nbsp;Switch hardware ID information
##&nbsp;--------------------------------
##&nbsp;
##&nbsp;Switch is booted up
##&nbsp;  Switch type is : MDS 9124 (1 Slot) Chassis
##&nbsp;  Model number is DS-C9124-K9
##&nbsp;  H/W version is 1.0
##&nbsp;  Part Number is 73-10583-03
##&nbsp;  Part Revision is A2
##&nbsp;  Manufacture Date is Year 10 Week 44
##&nbsp;  Serial number is FOX10440B29
##&nbsp;  CLEI code is COM7600ARA
##&nbsp;
##&nbsp;--------------------------------
##&nbsp;Chassis has 1 Module slots
##&nbsp;--------------------------------
##&nbsp;
##&nbsp;Module1  ok
##&nbsp;  Module type is : 1/2/4 Gbps FC/Supervisor-2
##&nbsp;  0 submodules are present
##&nbsp;  Model number is DS-C9124-K9-SUP
##&nbsp;  H/W version is 1.0
##&nbsp;  Part Number is 73-10583-03
##&nbsp;  Part Revision is A2
##&nbsp;  Manufacture Date is Year 10 Week 53
##&nbsp;  Serial number is JAE1053JPHH
##&nbsp;  CLEI code is COM7600ARA
##&nbsp;
##&nbsp;---------------------------------------
##&nbsp;Chassis has 2 PowerSupply Slots
##&nbsp;---------------------------------------
##&nbsp;
##&nbsp;PS1 ok
##&nbsp;  Power supply type is: 299.88W 110v/220v AC
##&nbsp;  Model number is PWR-C49-300AC
##&nbsp;  H/W version is 208.0
##&nbsp;  Part Number is 341-0103-02
##&nbsp;  Part Revision is D0
##&nbsp;  Manufacture Date is Year 11 Week 10
##&nbsp;  Serial number is PAC11100LF4
##&nbsp;  CLEI code is
##&nbsp;
##&nbsp;PS2 ok
##&nbsp;  Power supply type is: 299.88W 110v/220v AC
##&nbsp;  Model number is PWR-C49-300AC
##&nbsp;  H/W version is 208.0
##&nbsp;  Part Number is 341-0103-02
##&nbsp;  Part Revision is D0
##&nbsp;  Manufacture Date is Year 11 Week 15
##&nbsp;  Serial number is QCS1115B0MZ
##&nbsp;  CLEI code is
##&nbsp;
##&nbsp;----------------------------------
##&nbsp;Chassis has 3 Fan slots
##&nbsp;----------------------------------
##&nbsp;
##&nbsp;Fan1 ok
##&nbsp;
##&nbsp;Fan2 ok
##&nbsp;
##&nbsp;Fan3 ok
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;           {
##&nbsp;             'operating_system' => 'Nexus',
##&nbsp;             'kernel_uptime' => '19 day(s), 18 hour(s), 36 minute(s), 32 second(s)',
##&nbsp;             'software' => [
##&nbsp;                             {
##&nbsp;                               'system_compile_time' => '920/2011 15:00:00 [10/05/2011 08:09:39]',
##&nbsp;                               'system_image_file_is' => 'm9100-s2ek9-mz.5.0.7.bin',
##&nbsp;                               'system' => 'version 5.0(7)',
##&nbsp;                               'kickstart' => 'version 5.0(7)',
##&nbsp;                               'bios' => 'version 1.0.19',
##&nbsp;                               'bios_compile_time' => '0201/10',
##&nbsp;                               'kickstart_image_file_is' => 'm9100-s2ek9-kickstart-mz.5.0.7.bin',
##&nbsp;                               'kickstart_compile_time' => '920/2011 15:00:00 [10/05/2011 06:17:51]',
##&nbsp;                               'loader' => 'version NA'
##&nbsp;                             }
##&nbsp;                           ],
##&nbsp;             'hardware' => [
##&nbsp;                             {
##&nbsp;                               'bootflash' => '250368 kB',
##&nbsp;                               'device_name' => 'vgci9124s38',
##&nbsp;                               'processor_board_id' => 'JAE1053JPHH',
##&nbsp;                               'chassis' => 'cisco MDS 9124 (1 Slot) Chassis ("12/4 Gbps FC/Supervisor-2")'
##&nbsp;                             }
##&nbsp;                           ],
##&nbsp;             'switch_hardware_id_information' => [
##&nbsp;                                                   {
##&nbsp;                                                     'serial_number' => 'FOX10440B29',
##&nbsp;                                                     'manufacture_date' => 'Year 10 Week 44',
##&nbsp;                                                     'part_revision' => 'A2',
##&nbsp;                                                     'switch_type' => 'MDS 9124 (1 Slot) Chassis',
##&nbsp;                                                     'hw_version' => '1.0',
##&nbsp;                                                     'part_number' => '73-10583-03',
##&nbsp;                                                     'boot_state' => 'booted up',
##&nbsp;                                                     'clei_code' => 'COM7600ARA',
##&nbsp;                                                     'model_number' => 'DS-C9124-K9'
##&nbsp;                                                   }
##&nbsp;                                                 ],
##&nbsp;             'last_reset' => [
##&nbsp;                               {
##&nbsp;                                 'system_version' => '5.0(7)',
##&nbsp;                                 'reason' => 'Unknown',
##&nbsp;                                 'service' => ''
##&nbsp;                               }
##&nbsp;                             ],
##&nbsp;             'chassis' => [
##&nbsp;                            {
##&nbsp;                              'fan' => [
##&nbsp;                                         {
##&nbsp;                                           'fan3' => [
##&nbsp;                                                       {
##&nbsp;                                                         'status' => 'ok'
##&nbsp;                                                       }
##&nbsp;                                                     ],
##&nbsp;                                           'fan1' => [
##&nbsp;                                                       {
##&nbsp;                                                         'status' => 'ok'
##&nbsp;                                                       }
##&nbsp;                                                     ],
##&nbsp;                                           'slots' => '3',
##&nbsp;                                           'fan2' => [
##&nbsp;                                                       {
##&nbsp;                                                         'status' => 'ok'
##&nbsp;                                                       }
##&nbsp;                                                     ]
##&nbsp;                                         }
##&nbsp;                                       ],
##&nbsp;                              'powersupply' => [
##&nbsp;                                                 {
##&nbsp;                                                   'ps2' => [
##&nbsp;                                                              {
##&nbsp;                                                                'serial_number' => 'QCS1115B0MZ',
##&nbsp;                                                                'manufacture_date' => 'Year 11 Week 15',
##&nbsp;                                                                'status' => 'ok',
##&nbsp;                                                                'part_revision' => 'D0',
##&nbsp;                                                                'power_supply_type' => '299.88W 110v220v AC',
##&nbsp;                                                                'hw_version' => '208.0',
##&nbsp;                                                                'part_number' => '341-0103-02',
##&nbsp;                                                                'clei_code' => '',
##&nbsp;                                                                'model_number' => 'PWR-C49-300AC'
##&nbsp;                                                              }
##&nbsp;                                                            ],
##&nbsp;                                                   'ps1' => [
##&nbsp;                                                              {
##&nbsp;                                                                'serial_number' => 'PAC11100LF4',
##&nbsp;                                                                'manufacture_date' => 'Year 11 Week 10',
##&nbsp;                                                                'status' => 'ok',
##&nbsp;                                                                'part_revision' => 'D0',
##&nbsp;                                                                'power_supply_type' => '299.88W 110v220v AC',
##&nbsp;                                                                'hw_version' => '208.0',
##&nbsp;                                                                'part_number' => '341-0103-02',
##&nbsp;                                                                'clei_code' => '',
##&nbsp;                                                                'model_number' => 'PWR-C49-300AC'
##&nbsp;                                                              }
##&nbsp;                                                            ],
##&nbsp;                                                   'slots' => '2'
##&nbsp;                                                 }
##&nbsp;                                               ],
##&nbsp;                              'module' => [
##&nbsp;                                            {
##&nbsp;                                              'module1' => [
##&nbsp;                                                             {
##&nbsp;                                                               'serial_number' => 'JAE1053JPHH',
##&nbsp;                                                               'manufacture_date' => 'Year 10 Week 53',
##&nbsp;                                                               'status' => 'ok',
##&nbsp;                                                               'part_revision' => 'A2',
##&nbsp;                                                               'hw_version' => '1.0',
##&nbsp;                                                               'part_number' => '73-10583-03',
##&nbsp;                                                               'module_type' => '12/4 Gbps FC/Supervisor-2',
##&nbsp;                                                               'submodules_are_present' => '0',
##&nbsp;                                                               'clei_code' => 'COM7600ARA',
##&nbsp;                                                               'model_number' => 'DS-C9124-K9-SUP'
##&nbsp;                                                             }
##&nbsp;                                                           ],
##&nbsp;                                              'slots' => '1'
##&nbsp;                                            }
##&nbsp;                                          ]
##&nbsp;                            }
##&nbsp;                          ]
##&nbsp;           }
##&nbsp;         ];

sub show_hardware {
    my %args = @_;
    my (@result_array);
    my $resp        = $args{response_object};
    my $output      = $resp->get_processed_output();
    my $error_level = $resp->get_execution_status();

    if ($error_level) {
        _check_error(
            raw_output      => $output,
            response_object => $resp,
        );
    }
    my @lines = get_lines( raw_output => $output );
    my $key;
    my $slots;
    my $key1;
    my $key_1;
    my $i;
    my $status;

    while ( my $line = shift(@lines) ) {
        $line =~ s/\///;
        if ( $line =~ /^(Software|Hardware|Last\s+reset)/ ) {
            $key = to_key_name( item => "$1" );
            $i = 0;
        } elsif ( $line =~ /(^http:\/\/.*)/ ) {
            $result_array[0]->{gnu_license_info} = $1;
        } elsif ( $line =~ /^Cisco\s+(.*)\s+Operating\s+System\s+.*\s+Software/ )
        {
            $result_array[0]->{operating_system} = "$1";
        } elsif ( $line =~ /(.*):\s+(.*)/ && defined($key) && $line !~ /type/ ) {
            my $key1 = $1;
            my $val1 = $2;
            $key1 =~ s/-/_/g;
            $result_array[0]->{$key}->[$i]->{ to_key_name( item => "$key1" ) } =
              $val1;
        } elsif ( $line =~ /(Processor\s+Board\s+ID)\s+(.*)/ ) {
            $result_array[0]->{$key}->[$i]->{ to_key_name( item => "$1" ) } =
              $2;
        } elsif ( $line =~ /(.*)\s+CPU\s+with\s+(.*)\s+of memory/ ) {
            $result_array[0]->{$key}->[$i]->{cpu_memory} = $2;
            $result_array[0]->{$key}->[$i]->{cpu}        = $1;
        } elsif ( $line =~ /(.*\s+Chassis\s+.*)/ ) {
            $result_array[0]->{$key}->[$i]->{chassis} = $1;
        } elsif($line =~ /(.*):\s*(.*)/ && defined($key) && $line !~ /type/ ){
            my $key1 = $1;
            my $val1 = $2;
            $key1 =~ s/-/_/g;
            $result_array[0]->{$key}->[$i]->{to_key_name(item => "$key1")} = $val1;
        } elsif ( $line =~ /(Kernel\s+uptime)\s+is\s+(.*)/ ) {
            $result_array[0]->{ to_key_name( item => "$1" ) } = $2;
        } elsif ( $line =~ /Switch hardware ID information/ ) {
            $line = shift(@lines);
            $line = shift(@lines);
            $line =~ /.*is\s+(.*)/;
            $result_array[0]->{switch_hardware_id_information}->[0]
              ->{boot_state} = $1;
            while ( $line !~ /----/ ) {
                $line = shift(@lines);
                $line =~ s/\///;
                $line =~ /(\S+\s+\w+)\s+is\s*:?\s*(.*)/;
                $result_array[0]->{switch_hardware_id_information}->[0]
                  ->{ to_key_name( item => $1 ) } = $2;
            }
        } elsif ( $line =~ /Chassis\s+has\s+(\d+)\s+(\S+)\s+[sS]lots/ ) {
            $slots = $1;
            $key1  = $2;
        } elsif ( $line =~ /^(\w+)\s+(\w+)$/ ) {
            $key_1  = $1;
            $status = $2;
            $result_array[0]->{chassis}->[0]
              ->{ to_key_name( item => $key1 ) }[0]->{slots} = $slots;
            $result_array[0]->{chassis}->[0]
              ->{ to_key_name( item => $key1 ) }[0]
              ->{ to_key_name( item => $key_1 ) }[0]->{status} = $status;
        } elsif ($line =~ /(\S+\s+\w+\s+\w+)\s+is\s*:?\s*(.*)/
            || $line =~ /(\S+\s+\w+)\s+is\s*:?\s*(.*)/ ) {
            if ( $line =~ /[Ll]icense/ ) { }
            else {
                $result_array[0]->{chassis}->[0]
                  ->{ to_key_name( item => $key1 ) }[0]
                  ->{ to_key_name( item => $key_1 ) }[0]
                  ->{ to_key_name( item => $1 ) } = $2;
            }
        } elsif ( $line =~ /\s*(\d+)\s+(submodules.*)/ ) {
            $result_array[0]->{chassis}->[0]
              ->{ to_key_name( item => $key1 ) }[0]
              ->{ to_key_name( item => $key_1 ) }[0]
              ->{ to_key_name( item => $2 ) } = $1;
        }
    }
    return \@result_array;
} ## end sub show_hardware

1;

