#
#Copyright (c) 2010 Network Appliance, Inc.
#All rights reserved.
#
## @summary Routines to parse CiscoNetwork configure terminal commands
##
## @description
##    This is a library to parse the raw output of CiscoNetwork switch configure interface subcommands.
##
## @synopsis
##  use NACL::APISet::Switch::CLI::CiscoNetwork::IOS::configure_default
##  
##  configure_mac_address_table(%args)
##
## @author dl-nacl-dev@netapp.com
##
## @status public
#
package NACL::APISet::Switch::CLI::CiscoNetwork::IOS::configure_default;

use strict;
use NACL::APISet::Response::CLI::ParserUtils;

#Include exceptions which will be thrown
use NACL::APISet::Exceptions::MissingRequiredParamException qw(:try);
use NACL::APISet::Exceptions::InvalidParamException;

sub _check_error {
    my %args = @_;

    my $raw_output = $args{raw_output};
    my $command = $args{command};
    my $error_message = $args{error_message};
    my $error ;

    # Check for the error
    if ($raw_output =~ /(Invalid\s+input\s+detected)|(not\s+a\s+switching\s+port)|(Invalid\s+command)|(Configuration\s+does\s+not\s+match\s+the\s+port\s+capability)/i) {
        $$error_message = $raw_output;
        $error = "InvalidParam";
    } elsif ($raw_output =~ /(Incomplete\s+command)/i) {
        $$error_message = $raw_output;
        $error = "MissingRequiredParam"
    }

    return $error;
}

## @name ::default
##
## @summary Parses the raw output from the "default" command
##
## @description 
##    This method can be used for parsing the raw output obtained after calling 
##    the method from default with any of its options as described in the 
##    corresponding cdef file 'configure_default.cdefs'
## @synopsis
##      default(%args);
## INPUTS TO METHOD CALL 'default' ON APISET OBJECT 
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> </tr>
## <tr> <td> No switch for this key </td> <td> 'entity-type' </td> </tr> 
## <tr> <td> No switch for this key </td> <td> 'entity' </td> </tr> 
## <tr> <td> No switch for this key </td> <td> 'entity-name' </td> </tr> 
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

##&nbsp;Command = default interface Gi2/1
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;Interface GigabitEthernet2/1 set to default configuration
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [];
##&nbsp;
sub default {
    my %args = @_;
    
    my ($error_message);

    my $resp = $args{response_object};
    my $output = $resp->get_processed_output();
    my $command = $resp->get_command();
    my $error_level = $resp->get_execution_status();
    my ($error, $key);

    # Status returned by switch for this command is  
    # is ZERO even if command failed
    if (not defined($error)) {
        $error = _check_error(raw_output    => $output,
                              error_message => \$error_message);
    }
    if($error) {
        $error = 'NACL::APISet::Exceptions::'.$error.'Exception';
        $error->throw(text=>$error_message,response_object => $resp);
    }

    return [];
}

1;
