#
#Copyright (c) 2010 Network Appliance, Inc.
#All rights reserved.
#
## @summary Routines to parse CiscoNetwork configure terminal commands
##
## @description
##    This is a library to parse the raw output of CiscoNetwork switch configure terminal commands.
##
## @synopsis
##  use NACL::APISet::Switch::CLI::CiscoNetwork::IOS::configure
##  
##  configure_mac_address_table(%args)
##
## @author dl-nacl-dev@netapp.com
##
## @status public
#
package NACL::APISet::Switch::CLI::CiscoNetwork::IOS::configure_interface;

use strict;
use NACL::APISet::Response::CLI::ParserUtils;

#Include exceptions which will be thrown
use NACL::APISet::Exceptions::MissingRequiredParamException qw(:try);
use NACL::APISet::Exceptions::InvalidParamException;
use NACL::APISet::Exceptions::InvalidParamValueException;

sub _check_error {
    my %args = @_;

    my $raw_output = $args{raw_output};
    my $command = $args{command};
    my $error_message = $args{error_message};
    my $error ;

    # Check for the error
    if ($raw_output =~ /(Invalid\s+input\s+detected)|(not\s+a\s+switching\s+port)|(Invalid\s+command)|(Configuration\s+is\s+not\s+allowed)/i) {
        $$error_message = $raw_output;
        $error = "InvalidParam";
    } elsif ($raw_output =~ /(Configuration\s+does\s+not\s+match\s+the\s+port\s+capability)|(Cannot\s+set\s+speed)/i) {
        $$error_message = $raw_output;
        $error = "InvalidParamValue"
    } elsif ($raw_output =~ /(Incomplete\s+command)/i) {
        $$error_message = $raw_output;
        $error = "MissingRequiredParam"
    }

    return $error;
}

## @name ::channel_protocol
##
## @summary Parses the raw output from the "channel-protocol" command
##
## @description 
##    This method can be used for parsing the raw output obtained after calling 
##    the method from channel_group with any of its options as described in the 
##    corresponding cdef file 'configure_interface.cdefs'
## @synopsis
##      channel_group(%args);
## INPUTS TO METHOD CALL 'channel_protocol' ON APISET OBJECT 
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> </tr>
## <tr> <td> No switch for this key </td> <td> 'channel-protocol' </td> </tr> 
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

##&nbsp;Command = channel-protocol lacp
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [];
##&nbsp;
sub channel_protocol {
    my %args = @_;
    
    configure_interface_options(%args);
    return [];
}

## @name ::channel_group
##
## @summary Parses the raw output from the "channel-group" command
##
## @description 
##    This method can be used for parsing the raw output obtained after calling 
##    the method from channel_group with any of its options as described in the 
##    corresponding cdef file 'configure_interface.cdefs'
## @synopsis
##      channel_group(%args);
## INPUTS TO METHOD CALL 'channel_group' ON APISET OBJECT 
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> </tr>
## <tr> <td> No switch for this key </td> <td> 'channel-mode' </td> </tr> 
## <tr> <td> No switch for this key </td> <td> 'channel-num' </td> </tr> 
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

##&nbsp;Command = channel-group 5 mode active
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [];
##&nbsp;
sub channel_group {
    my %args = @_;
    
    configure_interface_options(%args);
    return [];
}

## @name ::spanning_tree
##
## @summary Parses the raw output from the "spanning-tree" command
##
## @description 
##    This method can be used for parsing the raw output obtained after calling 
##    the method from spanning_tree with any of its options as described in the 
##    corresponding cdef file 'configure_interface.cdefs'
## @synopsis
##      spanning_tree(%args);
## INPUTS TO METHOD CALL 'spanning_tree' ON APISET OBJECT 
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> </tr>
## <tr> <td> No switch for this key </td> <td> 'st-option' </td> </tr> 
## <tr> <td> No switch for this key </td> <td> 'st-option-value' </td> </tr> 
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

##&nbsp;Command = spanning-tree portfast network
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [];
##&nbsp;
sub spanning_tree {
    my %args = @_;
    
    configure_interface_options(%args);
    return [];
}

## @name ::mtu
##
## @summary Parses the raw output from the "mtu" command
##
## @description 
##    This method can be used for parsing the raw output obtained after calling 
##    the method from mtu with any of its options as described in the 
##    corresponding cdef file 'configure_interface.cdefs'
## @synopsis
##      mtu(%args);
## INPUTS TO METHOD CALL 'mtu' ON APISET OBJECT 
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> </tr>
## <tr> <td> No switch for this key </td> <td> 'mtu' </td> </tr> 
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

##&nbsp;Command = mtu 9216
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [];
##&nbsp;
sub mtu {
    my %args = @_;
    
    configure_interface_options(%args);
    return [];
}

## @name ::speed
##
## @summary Parses the raw output from the "speed" command
##
## @description 
##    This method can be used for parsing the raw output obtained after calling 
##    the method from speed with any of its options as described in the 
##    corresponding cdef file 'configure_interface.cdefs'
## @synopsis
##      speed(%args);
## INPUTS TO METHOD CALL 'speed' ON APISET OBJECT 
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> </tr>
## <tr> <td> No switch for this key </td> <td> 'speed' </td> </tr> 
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

##&nbsp;Command = speed 1000
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [];
##&nbsp;
sub speed {
    my %args = @_;
    
    configure_interface_options(%args);
    return [];
}

sub duplex {
    my %args = @_;
    
    configure_interface_options(%args);
    return [];
}

## @name ::flowcontrol
##
## @summary Parses the raw output from the "flowcontrol" command
##
## @description 
##    This method can be used for parsing the raw output obtained after calling 
##    the method from flowcontrol with any of its options as described in the 
##    corresponding cdef file 'configure_interface.cdefs'
## @synopsis
##      flowcontrol(%args);
## INPUTS TO METHOD CALL 'flowcontrol' ON APISET OBJECT 
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> </tr>
## <tr> <td> No switch for this key </td> <td> 'flow-operation' </td> </tr> 
## <tr> <td> No switch for this key </td> <td> 'flow-direction' </td> </tr> 
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

##&nbsp;Command = flowcontrol send on
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [];
##&nbsp;
sub flowcontrol {
    my %args = @_;
    
    configure_interface_options(%args);
    return [];
}

## @name ::shutdown
##
## @summary Parses the raw output from the "shutdown" command
##
## @description 
##    This method can be used for parsing the raw output obtained after calling 
##    the method from switchport with any of its options as described in the 
##    corresponding cdef file 'configure_interface.cdefs'
## @synopsis
##      shutdown(%args);
## INPUTS TO METHOD CALL 'shutdown' ON APISET OBJECT 
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples
##&nbsp;Command = shutdown
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [];
##&nbsp;
sub shutdown {
    my %args = @_;
    
    configure_interface_options(%args);
    return [];
}

## @name ::switchport
##
## @summary Parses the raw output from the "switchport" command
##
## @description 
##    This method can be used for parsing the raw output obtained after calling 
##    the method from switchport with any of its options as described in the 
##    corresponding cdef file 'configure_interface.cdefs'
## @synopsis
##      switchport(%args);
## INPUTS TO METHOD CALL 'switchport' ON APISET OBJECT 
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> </tr>
## <tr> <td> No switch for this key </td> <td> 'vlan-list' </td> </tr> 
## <tr> <td> 'mode' </td> <td> 'mode' </td> </tr> 
## <tr> <td> 'access vlan' </td> <td> 'access-vlan' </td> </tr> 
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples
##&nbsp;Command = switchport mode trunk
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;GigabitEthernet3/2: Configuration is not allowed on Port members. Remove the interface from the Port Channel to modify its config
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [];
##&nbsp;
##&nbsp;Command = switchport mode dynamic auto
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [];
##&nbsp;
sub switchport {
    my %args = @_;
    
    configure_interface_options(%args);
    return [];
}

## @name ::switchport_trunk
##
## @summary Parses the raw output from the "switchport trunk" command
##
## @description 
##    This method can be used for parsing the raw output obtained after calling 
##    the method from switchport_trunk with any of its options as described in the 
##    corresponding cdef file 'configure_interface.cdefs'
## @synopsis
##      switchport_trunk(%args);
## INPUTS TO METHOD CALL 'switchport_trunk' ON APISET OBJECT 
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> </tr>
## <tr> <td> 'remove' </td> <td> 'remove' </td> </tr> 
## <tr> <td> 'native vlan' </td> <td> 'native-vlan' </td> </tr> 
## <tr> <td> 'allowed vlan' </td> <td> 'allowed-vlan' </td> </tr> 
## <tr> <td> 'encapsulation' </td> <td> 'encapsulation' </td> </tr> 
## <tr> <td> No switch for this key </td> <td> 'vlan-list' </td> </tr> 
## <tr> <td> 'add' </td> <td> 'add' </td> </tr> 
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples
##&nbsp;Command = switchport trunk allowed vlan add 777
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [];
##&nbsp;
sub switchport_trunk {
    my %args = @_;
    
    configure_interface_options(%args);
    return [];
}

sub configure_interface_options {
    my %args = @_;

    my ($error_message);

    my $resp = $args{response_object};
    my $output = $resp->get_processed_output();
    my $command = $resp->get_command();
    my $error_level = $resp->get_execution_status();
    my ($error, $key);

    # Status returned by switch for this command is  
    # is ZERO even if command failed
    if (not defined($error)) {
        $error = _check_error(raw_output    => $output,
                              error_message => \$error_message);
    }
    if($error) {
        $error = 'NACL::APISet::Exceptions::'.$error.'Exception';
        $error->throw(text=>$error_message,response_object => $resp);
    }

    return [];

}


1;
