#
#Copyright (c) 2010-2012 Network Appliance, Inc.
#All rights reserved.
#
## @summary Routines to parse CiscoNetwork show and its subcommands
##
## @description
##    This is a library to parse the raw output of CiscoNetwork switch show and its subcommands.
##
## @synopsis
##  use NACL::APISet::Switch::CLI::CiscoNetwork::default::show
##
##  show_mac_address_table(%args)
##  show_version(%args)
##
## @author dl-nacl-dev@netapp.com
##
## @status public
#
package NACL::APISet::Switch::CLI::CiscoNetwork::IOS::show;

use strict;
use NACL::APISet::Response::CLI::ParserUtils;

#Include exceptions which will be thrown
use NACL::APISet::Exceptions::CommandFailedException qw(:try);
use NACL::APISet::Exceptions::MissingRequiredParamException;
use NACL::APISet::Exceptions::InvalidParamException;
use NACL::APISet::Exceptions::InvalidParamValueException;

sub _check_error {
    my %args = @_;

    my $raw_output    = $args{raw_output};
    my $command       = $args{command};
    my $error_message = $args{error_message};
    my $error;

    # Check for the error
    if ($raw_output =~ /(Invalid\s+input\s+detected)/i) {
        $$error_message = $raw_output;
        $error          = "InvalidParam";
    } elsif ($raw_output =~ /(Incomplete\s+command)/i) {
        $$error_message = $raw_output;
        $error          = "MissingRequiredParam";
    }

    return $error;
} ## end sub _check_error

## @name ::show_mac_address_table
##
## @summary Parses the raw output from the "show mac-address-table" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from show_mac_address_table with any of its options as described in the
##    corresponding cdef file 'show.cdefs'
## @synopsis
##      show_mac_address_table(%args);
## INPUTS TO METHOD CALL 'show_mac_address_table' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## <tr> <td> 'address' </td> <td> 'address' </td> <td> '0' </td> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples
##&nbsp;Command = show mac-address-table address 3333.0000.000d
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;
##&nbsp;Legend: * - primary entry
##&nbsp;       age - seconds since last seen
##&nbsp;       n/a - not available
##&nbsp;
##&nbsp; vlan   mac address     type    learn     age              ports
##&nbsp;------+----------------+--------+-----+----------+--------------------------
##&nbsp;Active Supervisor:
##&nbsp;*  352  3333.0000.000d    static  Yes          -   Te1/1,Te1/2,Te1/4,Gi2/17
##&nbsp;                                                  Gi3/19,Gi3/23,Gi3/35,Gi3/36
##&nbsp;                                                  Gi3/39,Gi3/40,Gi3/41,Gi3/42
##&nbsp;                                                  Gi3/43,Gi3/44,Gi3/45,Gi3/46
##&nbsp;                                                  Gi3/47,Gi3/48,Gi5/2,Gi6/1
##&nbsp;                                                  Gi6/2,Gi6/3,Gi6/4,Gi6/13
##&nbsp;                                                  Gi6/15,Gi6/41,Gi6/43,Gi6/44
##&nbsp;                                                  Gi6/45,Gi6/46,Gi6/47,Gi6/48
##&nbsp;                                                  Te7/1,Te7/2,Te8/1,Te8/2
##&nbsp;                                                  Gi9/1,Gi9/2,Gi9/3,Gi9/4
##&nbsp;                                                  Gi9/5,Gi9/6,Gi9/7,Gi9/9
##&nbsp;                                                  Gi9/10,Gi9/11,Gi9/12,Gi9/15
##&nbsp;                                                  Gi9/16,Gi9/17,Gi9/19,Gi9/42
##&nbsp;                                                  Gi9/45,Gi9/47,Gi9/48,Po255
##&nbsp;                                                  Router,Switch
##&nbsp;*  353  3333.0000.000d    static  Yes          -   Te1/1,Te1/2,Te1/4,Gi2/17
##&nbsp;                                                  Gi3/19,Gi3/23,Gi3/35,Gi3/36
##&nbsp;                                                  Gi3/39,Gi3/40,Gi3/41,Gi3/42
##&nbsp;                                                  Gi3/43,Gi3/44,Gi3/45,Gi3/46
##&nbsp;                                                  Gi3/47,Gi3/48,Gi5/2,Gi6/1
##&nbsp;                                                  Gi6/2,Gi6/3,Gi6/4,Gi6/13
##&nbsp;                                                  Gi6/15,Gi6/41,Gi6/43,Gi6/44
##&nbsp;                                                  Gi6/45,Gi6/46,Gi6/47,Gi6/48
##&nbsp;                                                  Te7/1,Te7/2,Te8/1,Te8/2
##&nbsp;                                                  Gi9/1,Gi9/2,Gi9/3,Gi9/4
##&nbsp;                                                  Gi9/5,Gi9/6,Gi9/7,Gi9/9
##&nbsp;                                                  Gi9/10,Gi9/11,Gi9/12,Gi9/15
##&nbsp;                                                  Gi9/16,Gi9/17,Gi9/19,Gi9/42
##&nbsp;                                                  Gi9/45,Gi9/47,Gi9/48,Po255
##&nbsp;                                                  Router,Switch
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;          {
##&nbsp;            'vlans' => [
##&nbsp;                         '352',
##&nbsp;                         '353',
##&nbsp;                       ],
##&nbsp;            'ports' => [
##&nbsp;                         'Te1/1',
##&nbsp;                         'Te1/2',
##&nbsp;                         'Te1/4',
##&nbsp;                         'Gi2/17',
##&nbsp;                         'Gi3/19',
##&nbsp;                         'Gi3/23',
##&nbsp;                         'Gi3/35',
##&nbsp;                         'Gi3/36',
##&nbsp;                         'Gi3/39',
##&nbsp;                         'Gi3/40',
##&nbsp;                         'Gi3/41',
##&nbsp;                         'Gi3/42',
##&nbsp;                         'Gi3/43',
##&nbsp;                         'Gi3/44',
##&nbsp;                         'Gi3/45',
##&nbsp;                         'Gi3/46',
##&nbsp;                         'Gi3/47',
##&nbsp;                         'Gi3/48',
##&nbsp;                         'Gi5/2',
##&nbsp;                         'Gi6/1',
##&nbsp;                         'Gi6/2',
##&nbsp;                         'Gi6/3',
##&nbsp;                         'Gi6/4',
##&nbsp;                         'Gi6/13',
##&nbsp;                         'Gi6/15',
##&nbsp;                         'Gi6/41',
##&nbsp;                         'Gi6/43',
##&nbsp;                         'Gi6/44',
##&nbsp;                         'Gi6/45',
##&nbsp;                         'Gi6/46',
##&nbsp;                         'Gi6/47',
##&nbsp;                         'Gi6/48',
##&nbsp;                         'Te7/1',
##&nbsp;                         'Te7/2',
##&nbsp;                         'Te8/1',
##&nbsp;                         'Te8/2',
##&nbsp;                         'Gi9/1',
##&nbsp;                         'Gi9/2',
##&nbsp;                         'Gi9/3',
##&nbsp;                         'Gi9/4',
##&nbsp;                         'Gi9/5',
##&nbsp;                         'Gi9/6',
##&nbsp;                         'Gi9/7',
##&nbsp;                         'Gi9/9',
##&nbsp;                         'Gi9/10',
##&nbsp;                         'Gi9/11',
##&nbsp;                         'Gi9/12',
##&nbsp;                         'Gi9/15',
##&nbsp;                         'Gi9/16',
##&nbsp;                         'Gi9/17',
##&nbsp;                         'Gi9/19',
##&nbsp;                         'Gi9/42',
##&nbsp;                         'Gi9/45',
##&nbsp;                         'Gi9/47',
##&nbsp;                         'Gi9/48',
##&nbsp;                         'Po255',
##&nbsp;                         'Router',
##&nbsp;                         'Switch'
##&nbsp;                       ],
##&nbsp;            'mac_address' => '3333.0000.000d',
##&nbsp;            'type' => 'static',
##&nbsp;            'age' => '-',
##&nbsp;            'learn' => 'Yes'
##&nbsp;          }
##&nbsp;        ];
##&nbsp;
##&nbsp;Command = show mac-address-table address 00:04:23:AD:BC:5E
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;Unicast Entries
##&nbsp; vlan   mac address     type        protocols               port
##&nbsp;-------+---------------+--------+---------------------+--------------------
##&nbsp;   1    0004.23ad.bc5e   dynamic ip                    GigabitEthernet1/2
##&nbsp;
##&nbsp;
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;          {
##&nbsp;            'vlans' => [
##&nbsp;                         '1'
##&nbsp;                       ],
##&nbsp;            'ports' => [
##&nbsp;                         {
##&nbsp;                           'interface_type' => 'GigabitEthernet',
##&nbsp;                           'interface_num' => '1/2',
##&nbsp;                           'port' => 'GigabitEthernet1/2'
##&nbsp;                         }
##&nbsp;                       ],
##&nbsp;            'mac_address' => '0004.23ad.bc5e',
##&nbsp;            'protocols' => 'ip',
##&nbsp;            'type' => 'dynamic'
##&nbsp;          }
##&nbsp;        ];
##&nbsp;
sub show_mac_address_table {
    my %args = @_;

    my ($error_message, @result_array);
    my $resp        = $args{response_object};
    my $output      = $resp->get_processed_output();
    my $command     = $resp->get_command();
    my $error_level = $resp->get_execution_status();
    my ($error);

    # Status returned by switch for this command is
    # is ZERO even if command failed
    $error = _check_error(
        raw_output    => $output,
        command       => "show mac-address-table",
        error_message => \$error_message
    );
    if ($error) {
        $error = 'NACL::APISet::Exceptions::' . $error . 'Exception';
        $error->throw(text => $error_message, response_object => $resp);
    }
    if ($output
        =~ /Vlan\s+Mac Address\s+Type\s+Ports|Mac Entries for Vlan.*|VLAN\s+Learning Status/i
        ) {
        my $count = 0;
        my %internal_hash;
        my @lines
            = get_lines(raw_output => $output, response_object => $resp);
        my $first_line = shift(@lines);
        if ($first_line =~ /Mac Address Table/) {
            for (my $i = 3; $i < @lines; $i++) {
                if ($lines[$i] =~ /(\w+)\s+(\w+\.\w+\.\w+)\s+(\w+)\s+(\S+)/) {
                    $result_array[$count]->{'vlan'}        = $1;
                    $result_array[$count]->{'mac_address'} = $2;
                    $result_array[$count]->{'type'}        = $3;
                    $result_array[$count]->{'ports'}       = $4;
                    $count++;
                }
            }
        } elsif ($first_line =~ /Learning Status/) {
            foreach my $line (@lines) {
                if ($line =~ /(\w+)\s+(\w+)/) {
                    $result_array[$count]->{'valn'}            = $1;
                    $result_array[$count]->{'learning_status'} = $2;
                    $count++;
                }
            }
        } elsif ($first_line =~ /Mac Entries for Vlan/) {
            my $key;
            unshift @lines, $first_line;
            foreach my $line (@lines) {
                if ($line =~ /Total Mac Address Space.*\:\s+(\d+)/) {
                    $internal_hash{total_address_apace} = $1;
                } elsif ($line =~ /Mac Entries for Vlan\s*(\d+)/) {
                    $key = to_key_name(item => 'mac entry for ' . $1);

                } elsif ($line =~ /(\w+\s+\w+\s+\w+)\s+\:\s(\d+)/) {
                    $internal_hash{$key}->[0]{to_key_name(item => $1)} = $2;
                }
            }
            push(@result_array, \%internal_hash);
        }

    } else {
        my @lines = get_lines(raw_output => $output);
        my $line;
        my %legends;
        while ($line !~ /vlan/i and (@lines)) {
            $line = shift(@lines);
            chomp($line);
            if ($line =~ /^legend:\s+(.*)$/i) {
                $line = $1;
            }
            if ($line =~ /(.*)\s*\-\s*(.*)/i) {
                $legends{$1} = $2;
            }
        }

        @lines = grep (!/^-----/,             @lines);
        @lines = grep (!/^Active Supervisor/, @lines);
        @lines = grep (!/^Module \d+/,        @lines);

        my %hash;
        my $id;
        my (%mac_addresses, @mac_address);
        while (@lines) {
            $line = shift(@lines);
            chomp($line);
            if ($line =~ /^\*\s+/) {
                @mac_address = split(/\s+/, $line);
                if (defined $mac_addresses{$mac_address[2]}{mac_address}) {
                    push(
                        @{$mac_addresses{$mac_address[2]}{'vlans'}},
                        $mac_address[1]
                    );
                    my $already_met = 0;
                    $line = shift(@lines);
                    while ($line !~ /^\*\s+/ and (@lines)) {
                        $line        = shift(@lines);
                        $already_met = 1;
                    }
                    if ($already_met == 1) {
                        if (scalar(@lines) > 0) {
                            unshift(@lines, $line);
                        }
                        next;
                    }
                }
                push(
                    @{$mac_addresses{$mac_address[2]}{'vlans'}},
                    $mac_address[1]
                );
                $mac_addresses{$mac_address[2]}{'entry_type'}
                    = $legends{$mac_address[0]};
                $mac_addresses{$mac_address[2]}{'type'} = $mac_address[3];
                $mac_addresses{$mac_address[2]}{'mac_address'}
                    = $mac_address[2];
                $mac_addresses{$mac_address[2]}{'learn'} = $mac_address[4];
                $mac_addresses{$mac_address[2]}{'age'}   = $mac_address[5];
                $line                                    = $mac_address[6];
                $id                                      = 2;
            } elsif ($line =~ /^\d+/) {
                @mac_address = split(/\s+/, $line);
                if (defined $mac_addresses{$mac_address[1]}{mac_address}) {
                    push(
                        @{$mac_addresses{$mac_address[1]}{'vlans'}},
                        $mac_address[0]
                    );
                    my $already_met = 0;
                    $line = shift(@lines);
                    while ($line !~ /^\d+/ and (@lines)) {
                        $line        = shift(@lines);
                        $already_met = 1;
                    }
                    if ($already_met == 1) {
                        if (scalar(@lines) > 0) {
                            unshift(@lines, $line);
                        }
                        next;
                    }
                }
                push(
                    @{$mac_addresses{$mac_address[1]}{'vlans'}},
                    $mac_address[0]
                );
                $mac_addresses{$mac_address[1]}{'type'} = $mac_address[2];
                $mac_addresses{$mac_address[1]}{'mac_address'}
                    = $mac_address[1];
                $mac_addresses{$mac_address[1]}{'protocols'}
                    = $mac_address[3];
                $line = $mac_address[4];
                $id   = 1;
            }
            my @ports = split(/\,/, $line);
            if (not defined($mac_addresses{$mac_address[$id]}{'ports'})) {
                foreach my $port (@ports) {
                    my %ports;
                    if ($port =~ /([a-zA-Z]+)(\d+)(\/\d+)?/) {
                        $ports{port}           = $port;
                        $ports{interface_type} = $1;
                        $ports{interface_num}  = $2 . $3;
                        push(
                            @{$mac_addresses{$mac_address[$id]}{'ports'}},
                            \%ports
                        );
                    }
                }
            }
        }

        foreach my $key (keys %mac_addresses) {
            my %tmp;
            $tmp{'mac_address'} = $mac_addresses{$key}{'mac_address'};
            $tmp{'age'}         = $mac_addresses{$key}{'age'}
                if (defined($mac_addresses{$key}{'age'}));
            $tmp{'entry_type'} = $mac_addresses{$key}{'entry_type'}
                if ($mac_addresses{$key}{'entry_type'});
            $tmp{'learn'} = $mac_addresses{$key}{'learn'}
                if (defined($mac_addresses{$key}{'learn'}));
            $tmp{'protocols'} = $mac_addresses{$key}{'protocols'}
                if ($mac_addresses{$key}{'protocols'});
            $tmp{'type'}  = $mac_addresses{$key}{'type'};
            $tmp{'vlans'} = $mac_addresses{$key}{'vlans'};
            $tmp{'ports'} = $mac_addresses{$key}{'ports'};
            push(@result_array, \%tmp);
        }
    }

    return \@result_array;
} ## end sub show_mac_address_table

## @name ::show_vlan
##
## @summary Parses the raw output from the "show vlan" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from show_vlan with any of its options as described in the
##    corresponding cdef file 'show.cdefs'
## @synopsis
##      show_vlan(%args);
## INPUTS TO METHOD CALL 'show_vlan' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> </tr>
## <tr> <td> 'id' </td> <td> 'vlan-ids' </td> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

##&nbsp;Command = show vlan id 1-2
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;
##&nbsp;VLAN Name                             Status    Ports
##&nbsp;---- -------------------------------- --------- -------------------------------
##&nbsp;1    default                          active    Te1/1, Te1/2, Te1/3, Te1/4
##&nbsp;                                                Gi2/2, Gi2/3, Gi2/4, Gi2/5
##&nbsp;                                                Gi2/6, Gi2/7, Gi2/8, Gi2/9
##&nbsp;                                                Gi2/10, Gi2/11, Gi2/12, Gi2/13
##&nbsp;                                                Gi2/14, Gi2/15, Gi2/16, Gi2/17
##&nbsp;                                                Gi2/18, Gi2/19, Gi2/20, Gi2/21
##&nbsp;                                                Gi2/22, Gi2/23, Gi2/24, Gi3/1
##&nbsp;                                                Gi3/2, Gi3/3, Gi3/5, Gi3/6
##&nbsp;                                                Gi3/7, Gi3/8, Gi3/9, Gi3/10
##&nbsp;                                                Gi3/11, Gi3/12, Gi3/13, Gi3/14
##&nbsp;                                                Gi3/15, Gi3/16, Gi3/17, Gi3/18
##&nbsp;                                                Gi3/19, Gi3/20, Gi3/21, Gi3/22
##&nbsp;                                                Gi3/23, Gi3/24, Gi3/25, Gi3/26
##&nbsp;                                                Gi3/27, Gi3/28, Gi3/29, Gi3/30
##&nbsp;                                                Gi3/31, Gi3/32, Gi3/33, Gi3/34
##&nbsp;                                                Gi3/35, Gi3/36, Gi3/37, Gi3/38
##&nbsp;                                                Gi3/39, Gi3/40, Gi3/41, Gi3/42
##&nbsp;                                                Gi3/43, Gi3/44, Gi3/45, Gi3/46
##&nbsp;                                                Gi3/47, Gi3/48, Te4/1, Te4/2
##&nbsp;                                                Te4/3, Te4/4, Gi5/1, Gi6/1
##&nbsp;                                                Gi6/2, Gi6/3, Gi6/4, Gi6/5
##&nbsp;                                                Gi6/6, Gi6/7, Gi6/8, Gi6/9
##&nbsp;                                                Gi6/10, Gi6/11, Gi6/12, Gi6/13
##&nbsp;                                                Gi6/14, Gi6/15, Gi6/16, Gi6/17
##&nbsp;                                                Gi6/18, Gi6/19, Gi6/20, Gi6/21
##&nbsp;                                                Gi6/22, Gi6/23, Gi6/24, Gi6/25
##&nbsp;                                                Gi6/26, Gi6/27, Gi6/28, Gi6/29
##&nbsp;                                                Gi6/30, Gi6/31, Gi6/32, Gi6/33
##&nbsp;                                                Gi6/34, Gi6/35, Gi6/36, Gi6/37
##&nbsp;                                                Gi6/38, Gi6/39, Gi6/40, Gi6/41
##&nbsp;                                                Gi6/42, Gi6/43, Gi6/44, Gi6/45
##&nbsp;                                                Gi6/46, Gi6/47, Gi6/48, Te7/3
##&nbsp;                                                Te7/4, Te8/1, Te8/2, Te8/3
##&nbsp;                                                Te8/4, Gi9/1, Gi9/2, Gi9/3
##&nbsp;                                                Gi9/4, Gi9/5, Gi9/6, Gi9/7
##&nbsp;                                                Gi9/8, Gi9/9, Gi9/10, Gi9/11
##&nbsp;                                                Gi9/12, Gi9/13, Gi9/14, Gi9/15
##&nbsp;                                                Gi9/16, Gi9/17, Gi9/18, Gi9/19
##&nbsp;                                                Gi9/20, Gi9/21, Gi9/22, Gi9/23
##&nbsp;                                                Gi9/24, Gi9/26, Gi9/28, Gi9/29
##&nbsp;                                                Gi9/30, Gi9/31, Gi9/32, Gi9/33
##&nbsp;                                                Gi9/34, Gi9/35, Gi9/36, Gi9/37
##&nbsp;                                                Gi9/38, Gi9/39, Gi9/40, Gi9/41
##&nbsp;                                                Gi9/42, Gi9/43, Gi9/44, Gi9/45
##&nbsp;                                                Gi9/46, Gi9/47, Gi9/48
##&nbsp;2    VLAN0002                         act/lshut
##&nbsp;
##&nbsp;VLAN Type  SAID       MTU   Parent RingNo BridgeNo Stp  BrdgMode Trans1 Trans2
##&nbsp;---- ----- ---------- ----- ------ ------ -------- ---- -------- ------ ------
##&nbsp;1    enet  100001     1500  -      -      -        -    -        0      0
##&nbsp;2    enet  100002     9216  -      -      -        -    -        0      0
##&nbsp;
##&nbsp;Remote SPAN VLANs
##&nbsp;------------------------------------------------------------------------------
##&nbsp;
##&nbsp;
##&nbsp;Primary Secondary Type              Ports
##&nbsp;------- --------- ----------------- ------------------------------------------
##&nbsp;
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;          undef,
##&nbsp;          {
##&nbsp;            'status' => 'active',
##&nbsp;            'name' => 'default',
##&nbsp;            'Trans1' => '0',
##&nbsp;            'Parent' => '-',
##&nbsp;            'MTU' => '1500',
##&nbsp;            'ports' => [
##&nbsp;                         'Te1/1',
##&nbsp;                         'Te1/2',
##&nbsp;                         'Te1/3',
##&nbsp;                         'Te1/4',
##&nbsp;                         'Gi2/2',
##&nbsp;                         'Gi2/3',
##&nbsp;                         'Gi2/4',
##&nbsp;                         'Gi2/5',
##&nbsp;                         'Gi2/6',
##&nbsp;                         'Gi2/7',
##&nbsp;                         'Gi2/8',
##&nbsp;                         'Gi2/9',
##&nbsp;                         'Gi2/10',
##&nbsp;                         'Gi2/11',
##&nbsp;                         'Gi2/12',
##&nbsp;                         'Gi2/13',
##&nbsp;                         'Gi2/14',
##&nbsp;                         'Gi2/15',
##&nbsp;                         'Gi2/16',
##&nbsp;                         'Gi2/17',
##&nbsp;                         'Gi2/18',
##&nbsp;                         'Gi2/19',
##&nbsp;                         'Gi2/20',
##&nbsp;                         'Gi2/21',
##&nbsp;                         'Gi2/22',
##&nbsp;                         'Gi2/23',
##&nbsp;                         'Gi2/24',
##&nbsp;                         'Gi3/1',
##&nbsp;                         'Gi3/2',
##&nbsp;                         'Gi3/3',
##&nbsp;                         'Gi3/5',
##&nbsp;                         'Gi3/6',
##&nbsp;                         'Gi3/7',
##&nbsp;                         'Gi3/8',
##&nbsp;                         'Gi3/9',
##&nbsp;                         'Gi3/10',
##&nbsp;                         'Gi3/11',
##&nbsp;                         'Gi3/12',
##&nbsp;                         'Gi3/13',
##&nbsp;                         'Gi3/14',
##&nbsp;                         'Gi3/15',
##&nbsp;                         'Gi3/16',
##&nbsp;                         'Gi3/17',
##&nbsp;                         'Gi3/18',
##&nbsp;                         'Gi3/19',
##&nbsp;                         'Gi3/20',
##&nbsp;                         'Gi3/21',
##&nbsp;                         'Gi3/22',
##&nbsp;                         'Gi3/23',
##&nbsp;                         'Gi3/24',
##&nbsp;                         'Gi3/25',
##&nbsp;                         'Gi3/26',
##&nbsp;                         'Gi3/27',
##&nbsp;                         'Gi3/28',
##&nbsp;                         'Gi3/29',
##&nbsp;                         'Gi3/30',
##&nbsp;                         'Gi3/31',
##&nbsp;                         'Gi3/32',
##&nbsp;                         'Gi3/33',
##&nbsp;                         'Gi3/34',
##&nbsp;                         'Gi3/35',
##&nbsp;                         'Gi3/36',
##&nbsp;                         'Gi3/37',
##&nbsp;                         'Gi3/38',
##&nbsp;                         'Gi3/39',
##&nbsp;                         'Gi3/40',
##&nbsp;                         'Gi3/41',
##&nbsp;                         'Gi3/42',
##&nbsp;                         'Gi3/43',
##&nbsp;                         'Gi3/44',
##&nbsp;                         'Gi3/45',
##&nbsp;                         'Gi3/46',
##&nbsp;                         'Gi3/47',
##&nbsp;                         'Gi3/48',
##&nbsp;                         'Te4/1',
##&nbsp;                         'Te4/2',
##&nbsp;                         'Te4/3',
##&nbsp;                         'Te4/4',
##&nbsp;                         'Gi5/1',
##&nbsp;                         'Gi6/1',
##&nbsp;                         'Gi6/2',
##&nbsp;                         'Gi6/3',
##&nbsp;                         'Gi6/4',
##&nbsp;                         'Gi6/5',
##&nbsp;                         'Gi6/6',
##&nbsp;                         'Gi6/7',
##&nbsp;                         'Gi6/8',
##&nbsp;                         'Gi6/9',
##&nbsp;                         'Gi6/10',
##&nbsp;                         'Gi6/11',
##&nbsp;                         'Gi6/12',
##&nbsp;                         'Gi6/13',
##&nbsp;                         'Gi6/14',
##&nbsp;                         'Gi6/15',
##&nbsp;                         'Gi6/16',
##&nbsp;                         'Gi6/17',
##&nbsp;                         'Gi6/18',
##&nbsp;                         'Gi6/19',
##&nbsp;                         'Gi6/20',
##&nbsp;                         'Gi6/21',
##&nbsp;                         'Gi6/22',
##&nbsp;                         'Gi6/23',
##&nbsp;                         'Gi6/24',
##&nbsp;                         'Gi6/25',
##&nbsp;                         'Gi6/26',
##&nbsp;                         'Gi6/27',
##&nbsp;                         'Gi6/28',
##&nbsp;                         'Gi6/29',
##&nbsp;                         'Gi6/30',
##&nbsp;                         'Gi6/31',
##&nbsp;                         'Gi6/32',
##&nbsp;                         'Gi6/33',
##&nbsp;                         'Gi6/34',
##&nbsp;                         'Gi6/35',
##&nbsp;                         'Gi6/36',
##&nbsp;                         'Gi6/37',
##&nbsp;                         'Gi6/38',
##&nbsp;                         'Gi6/39',
##&nbsp;                         'Gi6/40',
##&nbsp;                         'Gi6/41',
##&nbsp;                         'Gi6/42',
##&nbsp;                         'Gi6/43',
##&nbsp;                         'Gi6/44',
##&nbsp;                         'Gi6/45',
##&nbsp;                         'Gi6/46',
##&nbsp;                         'Gi6/47',
##&nbsp;                         'Gi6/48',
##&nbsp;                         'Te7/3',
##&nbsp;                         'Te7/4',
##&nbsp;                         'Te8/1',
##&nbsp;                         'Te8/2',
##&nbsp;                         'Te8/3',
##&nbsp;                         'Te8/4',
##&nbsp;                         'Gi9/1',
##&nbsp;                         'Gi9/2',
##&nbsp;                         'Gi9/3',
##&nbsp;                         'Gi9/4',
##&nbsp;                         'Gi9/5',
##&nbsp;                         'Gi9/6',
##&nbsp;                         'Gi9/7',
##&nbsp;                         'Gi9/8',
##&nbsp;                         'Gi9/9',
##&nbsp;                         'Gi9/10',
##&nbsp;                         'Gi9/11',
##&nbsp;                         'Gi9/12',
##&nbsp;                         'Gi9/13',
##&nbsp;                         'Gi9/14',
##&nbsp;                         'Gi9/15',
##&nbsp;                         'Gi9/16',
##&nbsp;                         'Gi9/17',
##&nbsp;                         'Gi9/18',
##&nbsp;                         'Gi9/19',
##&nbsp;                         'Gi9/20',
##&nbsp;                         'Gi9/21',
##&nbsp;                         'Gi9/22',
##&nbsp;                         'Gi9/23',
##&nbsp;                         'Gi9/24',
##&nbsp;                         'Gi9/26',
##&nbsp;                         'Gi9/28',
##&nbsp;                         'Gi9/29',
##&nbsp;                         'Gi9/30',
##&nbsp;                         'Gi9/31',
##&nbsp;                         'Gi9/32',
##&nbsp;                         'Gi9/33',
##&nbsp;                         'Gi9/34',
##&nbsp;                         'Gi9/35',
##&nbsp;                         'Gi9/36',
##&nbsp;                         'Gi9/37',
##&nbsp;                         'Gi9/38',
##&nbsp;                         'Gi9/39',
##&nbsp;                         'Gi9/40',
##&nbsp;                         'Gi9/41',
##&nbsp;                         'Gi9/42',
##&nbsp;                         'Gi9/43',
##&nbsp;                         'Gi9/44',
##&nbsp;                         'Gi9/45',
##&nbsp;                         'Gi9/46',
##&nbsp;                         'Gi9/47',
##&nbsp;                         'Gi9/48'
##&nbsp;                       ],
##&nbsp;            'Type' => 'enet',
##&nbsp;            'SAID' => '100001',
##&nbsp;            'Trans2' => '0',
##&nbsp;            'VLAN' => 1,
##&nbsp;            'BridgeNo' => '-',
##&nbsp;            'BrdgMode' => '-',
##&nbsp;            'id' => 1,
##&nbsp;            'RingNo' => '-',
##&nbsp;            'Stp' => '-'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'status' => 'act/lshut',
##&nbsp;            'name' => 'VLAN0002',
##&nbsp;            'Trans1' => '0',
##&nbsp;            'Parent' => '-',
##&nbsp;            'MTU' => '9216',
##&nbsp;            'ports' => [],
##&nbsp;            'Type' => 'enet',
##&nbsp;            'SAID' => '100002',
##&nbsp;            'Trans2' => '0',
##&nbsp;            'VLAN' => 2,
##&nbsp;            'BridgeNo' => '-',
##&nbsp;            'BrdgMode' => '-',
##&nbsp;            'id' => 2,
##&nbsp;            'RingNo' => '-',
##&nbsp;            'Stp' => '-'
##&nbsp;          }
##&nbsp;        ];
##&nbsp;
sub show_vlan {
    my %args = @_;

    my ($error_message, @result_array);
    my $resp        = $args{response_object};
    my $output      = $resp->get_processed_output();
    my $command     = $resp->get_command();
    my $error_level = $resp->get_execution_status();
    my ($error);

    # Status returned by switch for this command is
    # is ZERO even if command failed
    $error = _check_error(
        raw_output    => $output,
        command       => "show mac-address-table",
        error_message => \$error_message
    );
    if ($error) {
        $error = 'NACL::APISet::Exceptions::' . $error . 'Exception';
        $error->throw(text => $error_message, response_object => $resp);
    }

    my @lines = get_lines(raw_output => $output, first_line => 'VLAN');

    @lines = grep (!/^--/, @lines);

    # Get the headings
    my $line = shift(@lines);
    my @headings = split(/\s+/, $line);

    my ($vlan_count, %vlans, @vlan_info);
    $vlan_count = -1;
    $line       = shift(@lines);
    while ($line !~ /VLAN Type/i and @lines) {
        @vlan_info = split(/\s+/, $line);
        if ($line =~ /^\d+/) {
            $vlan_count++;
            $result_array[$vlan_count]->{'id'}     = shift(@vlan_info);
            $result_array[$vlan_count]->{'name'}   = shift(@vlan_info);
            $result_array[$vlan_count]->{'status'} = shift(@vlan_info);
            $line = join(" ", @vlan_info);
        }
        my @ports = split(/\,\s+/, $line);
        push(@{$result_array[$vlan_count]->{'ports'}}, @ports);
        $line = shift(@lines);
        chomp($line);
    }

    # get the second headings
    @headings = split(/\s+/, $line);

    $line = shift(@lines);
    while ($line !~ /Remote\s+/ and @lines) {
        my $j = 0;
        @vlan_info = split(/\s+/, $line);
        for (my $i; $i < scalar(@headings); $i++) {
            $result_array[$j]->{$headings[$i]} = $vlan_info[$i];
        }
        $j++;
        $line = shift(@lines);
    }

    return \@result_array;
} ## end sub show_vlan

## @name ::show_interfaces
##
## @summary Parses the raw output from the "show interfaces" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from show_interfaces with any of its options as described in the
##    corresponding cdef file 'show.cdefs'
## @synopsis
##      show_interfaces(%args);
## INPUTS TO METHOD CALL 'show_interfaces' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## <tr> <td> No switch for this key </td> <td> 'entity-type' </td> <td> '0' </td> </tr>
## <tr> <td> No switch for this key </td> <td> 'field' </td> <td> '0' </td> </tr>
## <tr> <td> No switch for this key </td> <td> 'entity-name' </td> <td> '0' </td> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples
##&nbsp;Command = show interfaces GigabitEthernet3/4 mtu
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;
##&nbsp;Port       Name               MTU
##&nbsp;Gi3/4                         9216
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;          {
##&nbsp;            'interface_type' => 'Gi',
##&nbsp;            'name' => '--',
##&nbsp;            'interface_num' => '3/4',
##&nbsp;            'port' => 'Gi3/4',
##&nbsp;            'mtu' => '9216'
##&nbsp;          }
##&nbsp;        ];
##&nbsp;
##&nbsp;Command = show interfaces GigabitEthernet3/4 status
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;
##&nbsp;Port         Name               Status       Vlan       Duplex  Speed Type
##&nbsp;Gi3/4                           notconnect   routed       full   1000 10/100/1000BaseT
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;          {
##&nbsp;            'status' => 'notconnect',
##&nbsp;            'name' => '--',
##&nbsp;            'vlan' => 'routed',
##&nbsp;            'port' => 'Gi3/4',
##&nbsp;            'duplex' => 'full',
##&nbsp;            'interface_type' => 'Gi',
##&nbsp;            'speed' => '1000',
##&nbsp;            'type' => '10/100/1000BaseT',
##&nbsp;            'interface_num' => '3/4'
##&nbsp;          }
##&nbsp;        ];
##&nbsp;
##&nbsp;Command = show interfaces GigabitEthernet3/4 flowcontrol
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;Port          Send FlowControl  Receive FlowControl  RxPause TxPause
##&nbsp;              admin    oper     admin    oper
##&nbsp;-----------   -------- -------- -------- --------    ------- -------
##&nbsp;Gi3/4         on       on       desired  desired     0       0
##&nbsp;
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;          {
##&nbsp;            'tx_pause' => '0',
##&nbsp;            'interface_type' => 'Gi',
##&nbsp;            'rx_pause' => '0',
##&nbsp;            'fc_send' => {
##&nbsp;                           'admin' => 'on',
##&nbsp;                           'oper' => 'on'
##&nbsp;                         },
##&nbsp;            'fc_receive' => {
##&nbsp;                              'admin' => 'desired',
##&nbsp;                              'oper' => 'desired'
##&nbsp;                            },
##&nbsp;            'interface_num' => '3/4',
##&nbsp;            'port' => 'Gi3/4'
##&nbsp;          }
##&nbsp;        ];
##&nbsp;
##&nbsp;Command = show interfaces gi3/9
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;GigabitEthernet3/9 is up, line protocol is up (connected)
##&nbsp;  Hardware is C6k 1000Mb 802.3, address is 0013.1962.b808 (bia 0013.1962.b808)
##&nbsp;  MTU 9216 bytes, BW 1000000 Kbit, DLY 10 usec,
##&nbsp;     reliability 255/255, txload 1/255, rxload 1/255
##&nbsp;  Encapsulation ARPA, loopback not set
##&nbsp;  Keepalive set (10 sec)
##&nbsp;  Full-duplex, 1000Mb/s, media type is 10/100/1000BaseT
##&nbsp;  input flow-control is on, output flow-control is on
##&nbsp;  Clock mode is auto
##&nbsp;  ARP type: ARPA, ARP Timeout 04:00:00
##&nbsp;  Last input 05:24:10, output 00:00:16, output hang never
##&nbsp;  Last clearing of "show interface" counters never
##&nbsp;  Input queue: 0/2000/1512/0 (size/max/drops/flushes); Total output drops: 682055611
##&nbsp;  Queueing strategy: fifo
##&nbsp;  Output queue: 0/40 (size/max)
##&nbsp;  5 minute input rate 0 bits/sec, 0 packets/sec
##&nbsp;  5 minute output rate 14000 bits/sec, 20 packets/sec
##&nbsp;     104301616 packets input, 173294545252 bytes, 0 no buffer
##&nbsp;     Received 52660852 broadcasts (51918302 multicasts)
##&nbsp;     6 runts, 19683640 giants, 0 throttles
##&nbsp;     678 input errors, 275 CRC, 96 frame, 1512 overrun, 0 ignored
##&nbsp;     0 watchdog, 0 multicast, 51699876 pause input
##&nbsp;     0 input packets with dribble condition detected
##&nbsp;     101429031731 packets output, 149408332327461 bytes, 0 underruns
##&nbsp;     0 output errors, 1 collisions, 138 interface resets
##&nbsp;     0 babbles, 0 late collision, 0 deferred
##&nbsp;     0 lost carrier, 0 no carrier, 0 PAUSE output
##&nbsp;     0 output buffer failures, 0 output buffers swapped out
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;          {
##&nbsp;            'arp_timeout' => '04:00:00',
##&nbsp;            'reliability' => '255/255',
##&nbsp;            'underruns' => '0',
##&nbsp;            'loopback' => 'not set',
##&nbsp;            'pause_output' => '0',
##&nbsp;            'clock_mode' => 'auto',
##&nbsp;            'output' => '00:00:16',
##&nbsp;            'duplex' => 'Full',
##&nbsp;            'overrun' => '1512',
##&nbsp;            'output_errors' => '0',
##&nbsp;            'input_queue' => [
##&nbsp;                               {
##&nbsp;                                 'drops' => '1512',
##&nbsp;                                 'max' => '2000',
##&nbsp;                                 'flushes' => '0',
##&nbsp;                                 'size' => '0'
##&nbsp;                               }
##&nbsp;                             ],
##&nbsp;            'output_hang' => 'never',
##&nbsp;            'output_queue' => [
##&nbsp;                                {
##&nbsp;                                  'max' => '40',
##&nbsp;                                  'size' => '0'
##&nbsp;                                }
##&nbsp;                              ],
##&nbsp;            'frame' => '96',
##&nbsp;            '5_minute_input_rate' => [
##&nbsp;                                       {
##&nbsp;                                         'in_packets_per_sec' => '0',
##&nbsp;                                         'in_bits_per_sec' => '0'
##&nbsp;                                       }
##&nbsp;                                     ],
##&nbsp;            'giants' => '19683640',
##&nbsp;            'bia' => '0013.1962.b808',
##&nbsp;            'speed' => '1000Mb/s',
##&nbsp;            'media_type' => '10/100/1000BaseT',
##&nbsp;            '5_minute_output_rate' => [
##&nbsp;                                        {
##&nbsp;                                          'in_packets_per_sec' => '20',
##&nbsp;                                          'in_bits_per_sec' => '14000'
##&nbsp;                                        }
##&nbsp;                                      ],
##&nbsp;            'input_errors' => '678',
##&nbsp;            'txload' => '1/255',
##&nbsp;            'input_bytes' => '173294545252',
##&nbsp;            'output_buffer_failures' => '0',
##&nbsp;            'pause_input' => '51699876',
##&nbsp;            'watchdog' => '0',
##&nbsp;            'late_collision' => '0',
##&nbsp;            'interface_resets' => '138',
##&nbsp;            'runts' => '6',
##&nbsp;            'last_input' => '05:24:10',
##&nbsp;            'input_packets_with_dribble_condition_detected' => '0',
##&nbsp;            'throttles' => '0',
##&nbsp;            'multicast' => '0',
##&nbsp;            'collisions' => '1',
##&nbsp;            'dly' => '10 usec',
##&nbsp;            'hardware' => 'C6k 1000Mb 802.3',
##&nbsp;            'output_buffers_swapped_out' => '0',
##&nbsp;            'interface' => 'GigabitEthernet3/9',
##&nbsp;            'address' => '0013.1962.b808',
##&nbsp;            'mtu' => '9216',
##&nbsp;            'no_buffer' => '0',
##&nbsp;            'no_carrier' => '0',
##&nbsp;            'lost_carrier' => '0',
##&nbsp;            'multicasts_received' => '51918302',
##&nbsp;            'output_packets' => '101429031731',
##&nbsp;            'broadcasts_received' => '52660852',
##&nbsp;            'ignored' => '0',
##&nbsp;            'rxload' => '1/255',
##&nbsp;            'input_flow_control' => 'on',
##&nbsp;            'encapsulation' => 'ARPA',
##&nbsp;            'deferred' => '0',
##&nbsp;            'arp_type' => 'ARPA',
##&nbsp;            'status' => 'up',
##&nbsp;            'output_bytes' => '149408332327461',
##&nbsp;            'total_output_drops' => '682055611',
##&nbsp;            'queueing_strategy' => 'fifo',
##&nbsp;            'input_packets' => '104301616',
##&nbsp;            'keepalive' => 'set (10 sec)',
##&nbsp;            'interface_num' => '3/9',
##&nbsp;            'line_protocol' => 'up',
##&nbsp;            'output_flow_control' => 'on',
##&nbsp;            'babbles' => '0',
##&nbsp;            'crc' => '275',
##&nbsp;            'last_clearing_of_show_interface_counters' => 'never',
##&nbsp;            'bw' => '1000000 Kbit',
##&nbsp;            'interface_type' => 'GigabitEthernet'
##&nbsp;          }
##&nbsp;        ];
##&nbsp;
sub show_interfaces {
    my %args = @_;

    my ($error_message, @result_array);
    my $resp        = $args{response_object};
    my $output      = $resp->get_processed_output();
    my $command     = $resp->get_command();
    my $error_level = $resp->get_execution_status();
    my ($error, $key);

    # Status returned by switch for this command is
    # is ZERO even if command failed
    if (not defined($error)) {
        $error = _check_error(
            raw_output    => $output,
            command       => "show interfaces",
            error_message => \$error_message
        );
    }
    if ($error) {
        $error = 'NACL::APISet::Exceptions::' . $error . 'Exception';
        $error->throw(text => $error_message, response_object => $resp);
    }

    chomp($command);
    my @lines;
    my $line;
    if ($command =~ /mtu$/) {
        @lines = get_lines(raw_output => $output, first_line => 'Port');
        $line = shift(@lines);
        foreach $line (@lines) {
            my %hash;
            chomp($line);
            my @array = split(/\s+/, $line);
            if (scalar(@array) == 2) {
                $hash{'mtu'} = $array[1];

                # In case Name field is not defined making it "--" thats what
                # comes for Nexus switch
                $hash{'name'} = "--";
            } else {
                $hash{'mtu'}  = $array[2];
                $hash{'name'} = $array[1];
            }
            my $if = $array[0];
            if ($if =~ /(\w+)(\d+(\/\d+)?)/) {
                $hash{'interface_type'} = $1;
                $hash{'interface_num'}  = $2;
            }
            $hash{'port'} = $array[0];
            push(@result_array, \%hash);
        }
    } elsif ($command =~ /status$/) {
        @lines = get_lines(raw_output => $output, first_line => 'Port');
        $line = shift(@lines);
        foreach $line (@lines) {
            my %hash;
            chomp($line);
            my @array = split(/\s+/, $line);
            my $if = shift(@array);
            if ($if =~ /(\w+)(\d+(\/\d+)?)/) {
                $hash{'interface_type'} = $1;
                $hash{'interface_num'}  = $2;
            }
            $hash{'port'} = $if;
            if (scalar(@array) == 6) {
                $hash{'name'} = shift(@array);
            } else {

                # In case Name field is NULL making it "--" thats what
                # comes for Nexus switch
                $hash{'name'}   = "--";
                $hash{'status'} = $array[0];
                $hash{'vlan'}   = $array[1];
                $hash{'duplex'} = $array[2];
                $hash{'speed'}  = $array[3];
                $hash{'type'}   = $array[4];
            }
            push(@result_array, \%hash);
        }
    } elsif ($command =~ /flowcontrol$/) {
        @lines = get_lines(raw_output => $output, first_line => 'Port');
        $line  = shift(@lines);
        $line  = shift(@lines);
        foreach $line (@lines) {
            next if ($line =~ /^-----/);
            my %hash;
            chomp($line);
            my @array = split(/\s+/, $line);
            my $if = $array[0];

            if ($if =~ /(\w+)(\d+(\/\d+)?)/) {
                $hash{'interface_type'} = $1;
                $hash{'interface_num'}  = $2;
            }
            $hash{'port'}                = $array[0];
            $hash{'fc_send'}{'admin'}    = $array[1];
            $hash{'fc_send'}{'oper'}     = $array[2];
            $hash{'fc_receive'}{'admin'} = $array[3];
            $hash{'fc_receive'}{'oper'}  = $array[4];
            $hash{'rx_pause'}            = $array[5];
            $hash{'tx_pause'}            = $array[6];
            push(@result_array, \%hash);
        }
    } else {
        @lines = get_lines(raw_output => $output);
        my $i = -1;
        foreach $line (@lines) {
            next if (!$line || !defined($line));
            if ($line
                =~ /(([a-zA-Z\-]+)(\d+(\/\d+)?))\s+is\s+(.*)\,\s+line\sprotocol\sis\s(\w+)/
                ) {
                $i++;
                $result_array[$i]->{interface}      = $1;
                $result_array[$i]->{interface_type} = $2;
                $result_array[$i]->{interface_num}  = $3;
                $result_array[$i]->{status}         = $5;
                $result_array[$i]->{line_protocol}  = $6;
            } elsif ($line
                =~ /Hardware\sis\s*(.+)\,\s*address\sis\s*(.+)\s+\(bia\s*(.+)\)/
                ) {
                $result_array[$i]->{hardware} = $1;
                $result_array[$i]->{address}  = $2;
                $result_array[$i]->{bia}      = $3;
            } elsif ($line =~ /Description\:\s*(.*)/) {
                $result_array[$i]->{description} = $1;
            } elsif (
                $line =~ /MTU\s*(\d+)\s*bytes\,\s*BW\s*(.+)\,\s*DLY\s+(.+)\,/)
            {
                $result_array[$i]->{mtu} = $1;
                $result_array[$i]->{bw}  = $2;
                $result_array[$i]->{dly} = $3;
            } elsif ($line
                =~ /reliability\s*(.+)\,\s*txload\s*(.+)\,\s*rxload\s+(.+)/) {
                $result_array[$i]->{reliability} = $1;
                $result_array[$i]->{txload}      = $2;
                $result_array[$i]->{rxload}      = $3;
            } elsif ($line =~ /Encapsulation\s*(.+)\,\s+loopback\s+(.+)/) {
                $result_array[$i]->{encapsulation} = $1;
                $result_array[$i]->{loopback}      = $2;
            } elsif ($line =~ /Keepalive\s+(.+)/) {
                $result_array[$i]->{keepalive} = $1;
            } elsif (
                $line =~ /(\w+)\-duplex\,\s*(.+)\,\s*media\stype\sis\s(.+)/) {
                $result_array[$i]->{duplex}     = $1;
                $result_array[$i]->{speed}      = $2;
                $result_array[$i]->{media_type} = $3;
            } elsif ($line =~ /(\w+)\-duplex\,\s*(.+)/) {
                $result_array[$i]->{duplex} = $1;
                $result_array[$i]->{speed}  = $2;
            } elsif ($line
                =~ /input\sflow\-control\sis\s(\w+)\,\soutput\sflow\-control\sis\s(\w+)/
                ) {
                $result_array[$i]->{input_flow_control}  = $1;
                $result_array[$i]->{output_flow_control} = $2;
            } elsif ($line =~ /ARP\stype\:\s+(\w+)\,\s+ARP\sTimeout\s(.+)/) {
                $result_array[$i]->{arp_type}    = $1;
                $result_array[$i]->{arp_timeout} = $2;
            } elsif ($line
                =~ /Last\sinput\s(.+)\,\s*output\s(.+)\,\s+output\shang\s(.+)/
                ) {
                $result_array[$i]->{last_input}  = $1;
                $result_array[$i]->{output}      = $2;
                $result_array[$i]->{output_hang} = $3;

            } elsif ($line =~ /Clock\s+mode\s+is\s+(.+)/) {
                $result_array[$i]->{clock_mode} = $1;
            } elsif ($line
                =~ /Last\sclearing\sof\s\"show\sinterface\"\scounters\s(.+)/)
            {
                $result_array[$i]->{last_clearing_of_show_interface_counters}
                    = $1;
            } elsif ($line
                =~ /Input\squeue\:\s+(.+)\s+\((.+)\)\;\s+Total\soutput\sdrops\:\s*(\d+)/
                ) {
                $result_array[$i]->{total_output_drops} = $3;
                my @key = split(/\//, $2);
                my @val = split(/\//, $1);
                for (my $k = 0; $k < scalar(@key); $k++) {
                    $result_array[$i]->{input_queue}->[0]->{$key[$k]}
                        = $val[$k];
                }
            } elsif ($line =~ /Queueing\sstrategy\:\s+(.+)/) {
                $result_array[$i]->{queueing_strategy} = $1;
            } elsif ($line =~ /Output\squeue\:\s+(.+)\s+\((.+)\)/) {
                my @key = split(/\//, $2);
                my @val = split(/\//, $1);
                for (my $k = 0; $k < scalar(@key); $k++) {
                    $result_array[$i]->{output_queue}->[0]->{$key[$k]}
                        = $val[$k];
                }
            } elsif ($line
                =~ /(\d+)\sminute\sinput\srate\s(\d+)\sbits\/sec\,\s(\d+)\spackets\/sec/
                ) {
                $result_array[$i]->{$1 . "_minute_input_rate"}->[0]
                    ->{in_bits_per_sec} = $2;
                $result_array[$i]->{$1 . "_minute_input_rate"}->[0]
                    ->{in_packets_per_sec} = $3;
            } elsif ($line
                =~ /(\d+)\sminute\soutput\srate\s(\d+)\sbits\/sec\,\s(\d+)\spackets\/sec/
                ) {
                $result_array[$i]->{$1 . "_minute_output_rate"}->[0]
                    ->{in_bits_per_sec} = $2;
                $result_array[$i]->{$1 . "_minute_output_rate"}->[0]
                    ->{in_packets_per_sec} = $3;
            } elsif ($line =~ /Members\s+in\s+this\s+channel\s*\:\s*(.*)/) {
                my @interfaces = split(/\s+/, $1);
                foreach (@interfaces) {
                    if ($_ =~ /((\w+)(\d+(\/\d+)?))/) {
                        my $temp = {
                            interface      => $1,
                            interface_type => $2,
                            interface_num  => $3,
                        };
                        push(
                            @{  $result_array[$i]->{members_in_this_channel}
                                },
                            $temp
                        );
                    }
                }
            } elsif ($line
                =~ /Received\s+(\d+)\s+broadcasts\s+\((\d+)\s+multicasts\)/) {
                $result_array[$i]->{broadcasts_received} = $1;
                $result_array[$i]->{multicasts_received} = $2;
            } elsif ($line
                =~ /(\d+)\s+packets\s+input\,\s+(\d+)\s+bytes\,\s+(\d+)\s+no\sbuffer/
                ) {
                $result_array[$i]->{input_packets} = $1;
                $result_array[$i]->{input_bytes}   = $2;
                $result_array[$i]->{no_buffer}     = $3;
            } elsif ($line
                =~ /(\d+)\s+packets\s+output\,\s+(\d+)\s+bytes\,\s+(\d+)\s+underruns/
                ) {
                $result_array[$i]->{output_packets} = $1;
                $result_array[$i]->{output_bytes}   = $2;
                $result_array[$i]->{underruns}      = $3;
            } elsif (my @counter = $line =~ /(\d+\s+[a-zA-Z\s]+)\,?/g) {
                foreach (@counter) {
                    $_ =~ s/\s*$//;
                    $_ =~ s/^\s*//;
                    if ($_ =~ /(\d+)\s+(.+)/) {
                        $result_array[$i]->{to_key_name(item => $2)} = $1;
                    }
                }
            }
        }
    }

    return \@result_array;

} ## end sub show_interfaces
## @name ::show_running_config
##
## @summary Parses the raw output from the "show running-config" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from show_running_config with any of its options as described in the
##    corresponding cdef file 'show.cdefs'
## @synopsis
##      show_running_config(%args);
## INPUTS TO METHOD CALL 'show_running_config' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> </tr>
## <tr> <td> No switch for this key </td> <td> 'entity' </td> </tr>
## <tr> <td> No switch for this key </td> <td> 'name' </td> </tr>
## <tr> <td> No switch for this key </td> <td> 'type' </td> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

##&nbsp;Command = show running-config interface Gi3/4
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;Building configuration...
##&nbsp;
##&nbsp;Current configuration : 187 bytes
##&nbsp;!
##&nbsp;interface GigabitEthernet3/4
##&nbsp; no ip address
##&nbsp; speed 1000
##&nbsp; duplex full
##&nbsp; flowcontrol receive desired
##&nbsp; flowcontrol send on
##&nbsp; spanning-tree portfast edge
##&nbsp; spanning-tree bpduguard enable
##&nbsp;end
##&nbsp;
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;          {
##&nbsp;            'interface_num' => '3/4',
##&nbsp;            'interface_type' => 'GigabitEthernet',
##&nbsp;            'running_conf' => [
##&nbsp;                                'no ip address',
##&nbsp;                                'speed 1000',
##&nbsp;                                'duplex full',
##&nbsp;                                'flowcontrol receive desired',
##&nbsp;                                'flowcontrol send on',
##&nbsp;                                'spanning-tree portfast edge',
##&nbsp;                                'spanning-tree bpduguard enable'
##&nbsp;                              ]
##&nbsp;          }
##&nbsp;        ];
##&nbsp;
sub show_running_config {
    my %args = @_;

    my ($error_message, @result_array);
    my $resp        = $args{response_object};
    my $output      = $resp->get_processed_output();
    my $command     = $resp->get_command();
    my $error_level = $resp->get_execution_status();
    my ($error, $key);
    my $i = -1;

    # Status returned by switch for this command is
    # is ZERO even if command failed
    $error = _check_error(
        raw_output    => $output,
        command       => "show running-config",
        error_message => \$error_message
    );
    if ($error) {
        $error = 'NACL::APISet::Exceptions::' . $error . 'Exception';
        $error->throw(text => $error_message, response_object => $resp);
    }
    my $option = $command;
    $option =~ s/show running-config//;

    my @lines = get_lines(raw_output => $output, first_line => '\!');

    if ($option =~ /interface/) {

        # Get rid of line starting with !
        my $line = shift(@lines);
        my $j    = -1;
        my ($interface_type, $interface_num, $interface);
        foreach $line (@lines) {
            chomp($line);
            if ($line =~ /^interface\s+([a-zA-Z]+)(\d+\/\d+)/i) {
                $j++;
                $result_array[$j]->{interface}      = $1 . $2;
                $result_array[$j]->{interface_type} = $1;
                $result_array[$j]->{interface_num}  = $2;
            } elsif ($j >= 0) {
                push(@{$result_array[$j]->{running_conf}}, $line);
            } else {
                next;
            }
        }
    } elsif ($option eq "") {
        while (my $line = shift(@lines)) {
            if ($line =~ /^interface\s+([a-zA-Z]+)(\d+\/\d+|\d+)/i) {
                my $j = 0;
                $result_array[0]->{interfaces}->[++$i]->{'interface'}
                    = $1 . $2;
                $result_array[0]->{interfaces}->[$i]->{interface_type} = $1;
                $result_array[0]->{interfaces}->[$i]->{interface_num}  = $2;
                while (($line = shift(@lines)) ne "!") {
                    $result_array[0]->{interfaces}->[$i]->{running_conf}
                        ->[$j++] = $line;
                }
            } elsif ($line
                =~ /banner\s+motd\s+\^C(NetApp Management Switch configuration version)\s+(.*)\s+-\s+(Data\s+ONTAP\s+version)\s+(.*)\^C/
                ) {
                $result_array[0]->{to_key_name(item => "$1")} = $2;
                $result_array[0]->{to_key_name(item => "$3")} = $4;
            }    #To handle newly changed banner for net-2960-3
            elsif ($line
                =~ /(banner\s+motd\s+\^C)?Catalyst\s+\d+\s+Reference.*version\s+(\S+).*/
                ) {
                $result_array[0]->{'rcf_version'} = $2;
            }
        }
    }
    return \@result_array;
} ## end sub show_running_config

## @name ::show_etherchannel
##
## @summary Parses the raw output from the "show etherchannel" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from show_etherchannel with any of its options as described in the
##    corresponding cdef file 'show.cdefs'
## @synopsis
##      show_etherchannel(%args);
## INPUTS TO METHOD CALL 'show_etherchannel' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> </tr>
## <tr> <td> No switch for this key </td> <td> 'etherchannel-num' </td> </tr>
## <tr> <td> 'summary' </td> <td> 'summary' </td> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

##&nbsp;Command = show etherchannel 255 summary
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;Flags:  D - down        P - bundled in port-channel
##&nbsp;        I - stand-alone s - suspended
##&nbsp;        H - Hot-standby (LACP only)
##&nbsp;        R - Layer3      S - Layer2
##&nbsp;        U - in use      N - not in use, no aggregation
##&nbsp;        f - failed to allocate aggregator
##&nbsp;
##&nbsp;        M - not in use, no aggregation due to minimum links not met
##&nbsp;        m - not in use, port not aggregated due to minimum links not met
##&nbsp;        u - unsuitable for bundling
##&nbsp;        d - default port
##&nbsp;
##&nbsp;        w - waiting to be aggregated
##&nbsp;Number of channel-groups in use: 4
##&nbsp;Number of aggregators:           4
##&nbsp;
##&nbsp;Group  Port-channel  Protocol    Ports
##&nbsp;------+-------------+-----------+-----------------------------------------------
##&nbsp;255    Po255(SU)       LACP      Te7/1(P)       Te7/2(P)
##&nbsp;
##&nbsp;Last applied Hash Distribution Algorithm: Fixed
##&nbsp;
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;          {
##&nbsp;
##&nbsp;            'protocol' => 'LACP',
##&nbsp;            'ports' => [
##&nbsp;                         {
##&nbsp;                           'interface_type' => 'Te',
##&nbsp;                           'interface' => 'Te7/1(P)',
##&nbsp;                           'if_flag' => 'P',
##&nbsp;                           'interface_num' => '7/1'
##&nbsp;                         },
##&nbsp;                         {
##&nbsp;                           'interface_type' => 'Te',
##&nbsp;                           'interface' => 'Te7/2(P)',
##&nbsp;                           'if_flag' => 'P',
##&nbsp;                           'interface_num' => '7/2'
##&nbsp;                         }
##&nbsp;                       ],
##&nbsp;            'port-channel' => 'Po255(SU)',
##&nbsp;            'group' => '255',
##&nbsp;            'flags' => {
##&nbsp;                         'w' => 'waiting to be aggregated',
##&nbsp;                         'S' => 'Layer2',
##&nbsp;                         'N' => 'not in use, no aggregation',
##&nbsp;                         'P' => 'bundled in port-channel',
##&nbsp;                         'd' => 'default port',
##&nbsp;                         'm' => 'not in use, port not aggregated due to minimum links not met',
##&nbsp;                         's' => 'suspended',
##&nbsp;                         'H' => 'Hot-standby (LACP only)',
##&nbsp;                         'u' => 'unsuitable for bundling',
##&nbsp;                         'M' => 'not in use, no aggregation due to minimum links not met',
##&nbsp;                         'D' => 'down',
##&nbsp;                         'I' => 'stand-alone',
##&nbsp;                         'R' => 'Layer3',
##&nbsp;                         'f' => 'failed to allocate aggregator',
##&nbsp;                         'U' => 'in use'
##&nbsp;                       },
##&nbsp;            'no_of_channel_group_in_use' => '4',
##&nbsp;            'no_of_aggregators' => '4'
##&nbsp;          }
##&nbsp;        ];
##&nbsp;
sub show_etherchannel {
    my %args = @_;

    my ($error_message, @result_array);
    my $resp        = $args{response_object};
    my $output      = $resp->get_processed_output();
    my $command     = $resp->get_command();
    my $error_level = $resp->get_execution_status();
    my ($error, $key);

    # Status returned by switch for this command is
    # is ZERO even if command failed
    $error = _check_error(
        raw_output    => $output,
        command       => "show etherchannel",
        error_message => \$error_message
    );
    if ($error) {
        $error = 'NACL::APISet::Exceptions::' . $error . 'Exception';
        $error->throw(text => $error_message, response_object => $resp);
    }

    if ($command =~ /summary/) {
        my @lines
            = get_lines(raw_output => $output, first_line => "\s*Flags:");
        my $line;
        my %flags;
        $line = shift(@lines);

        my ($no_of_channel_group_is_use, $no_of_aggregators);

        # Get rid of 'Flags:\s+' from the line
        $line =~ s/Flags:\s+//;
        while ($line !~ /Number\s+of\s+channel-groups/ and @lines) {
            chomp($line);
            my @tmp = split(/\s+\-\s+/, $line);
            my $key = shift(@tmp);
            my ($value, $next_key);
            while ($value = shift(@tmp)) {
                last if (!@tmp);
                my @tmp_arr = split(/\s+/, $value);
                $next_key    = pop(@tmp_arr);
                $value       = join(" ", @tmp_arr);
                $flags{$key} = $value;
                $key         = $next_key;
            }
            $flags{$key} = $value;
            $line = shift(@lines);
        }
        if ($line =~ /Number\s+of\s+channel\-groups\s+in\s+use:\s*(\d+)/) {
            $no_of_channel_group_is_use = $1;
        }
        $line = shift(@lines);
        if ($line =~ /Number\s+of\s+aggregators:\s*(\d+)/) {
            $no_of_aggregators = $1;
        }

        # get the headings
        $line = shift(@lines);
        my @headings = split(/\s+/, $line);

        # get rid of line
        ## ------+-------------+-----------+---------------------
        shift(@lines) if ($lines[0] =~ /^-----/);

        $line = shift(@lines);
        my @portchannel_info = split(/\s+/, $line);
        my %hash;
        foreach (my $i = 0; $i <= $#headings; $i++) {
            my $key = to_key_name('item' => $headings[$i]);
            if ($i == $#headings) {
                foreach my $port (@portchannel_info) {
                    my %port;
                    if ($port =~ /(\w+)(\d+\/\d+)\((\w+)\)/) {
                        $port{'interface'}      = $port;
                        $port{'interface_num'}  = $2;
                        $port{'interface_type'} = $1;
                        $port{'if_flag'}        = $3;
                        push(@{$hash{$key}}, \%port);
                    }
                }
                last;
            }
            $hash{$key} = shift(@portchannel_info);
        }

        $hash{'no_of_channel_group_in_use'} = $no_of_channel_group_is_use;
        $hash{'no_of_aggregators'}          = $no_of_aggregators;
        $hash{'flags'}                      = \%flags;

        @result_array = \%hash;

        return \@result_array;
    }
} ## end sub show_etherchannel

## @name ::show_snmp_community
##
## @summary Parses the raw output from the "show snmp community" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from show_snmp_community with any of its options as described in the
##    corresponding cdef file 'show.cdefs'
## @synopsis
##      show_snmp_community(%args);
## INPUTS TO METHOD CALL 'show_snmp_community' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

sub show_snmp_community {
    my %args = @_;

    my ($error_message, @result_array);
    my $resp        = $args{response_object};
    my $output      = $resp->get_processed_output();
    my $command     = $resp->get_command();
    my $error_level = $resp->get_execution_status();
    my ($error, $key);

    # Status returned by switch for this command is
    # is ZERO even if command failed
    $error = _check_error(
        raw_output    => $output,
        command       => $command,
        error_message => \$error_message
    );
    if ($error) {
        $error = 'NACL::APISet::Exceptions::' . $error . 'Exception';
        $error->throw(text => $error_message, response_object => $resp);
    }

    my @lines = get_lines(raw_output => $output);
    my $i = -1;
    my $val;
    foreach my $line (@lines) {
        $i++ if ($line =~ /^Community\sname/i);
        ($key, $val) = split(/\s*\:\s*/, $line);
        $result_array[$i]->{to_key_name('item' => $key)} = $val;
    }

    return \@result_array;
} ## end sub show_snmp_community

## @name ::show_version
##
## @summary Parses the raw output from the "show version" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from show_version with any of its options as described in the
##    corresponding cdef file 'show.cdefs'
## @synopsis
##      show_version(%args);
## INPUTS TO METHOD CALL 'show_version' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples
##&nbsp;Command = show version
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;Cisco IOS Software, C2960 Software (C2960-LANBASEK9-M), Version 12.2(55)SE, RELEASE SOFTWARE (fc2)
##&nbsp;Technical Support: http://www.cisco.com/techsupport
##&nbsp;Copyright (c) 1986-2010 by Cisco Systems, Inc.
##&nbsp;Compiled Sat 07-Aug-10 23:04 by prod_rel_team
##&nbsp;Image text-base: 0x00003000, data-base: 0x01800000
##&nbsp;
##&nbsp;ROM: Bootstrap program is C2960 boot loader
##&nbsp;BOOTLDR: C2960 Boot Loader (C2960-HBOOT-M) Version 12.2(53r)SEY3, RELEASE SOFTWARE (fc1)
##&nbsp;
##&nbsp;net-2960-3 uptime is 4 weeks, 4 days, 12 hours, 17 minutes
##&nbsp;System returned to ROM by power-on
##&nbsp;System image file is "flash:c2960-lanbasek9-mz.122-55.SE.bin"
##&nbsp;
##&nbsp;
##&nbsp;This product contains cryptographic features and is subject to United
##&nbsp;States and local country laws governing import, export, transfer and
##&nbsp;use. Delivery of Cisco cryptographic products does not imply
##&nbsp;third-party authority to import, export, distribute or use encryption.
##&nbsp;Importers, exporters, distributors and users are responsible for
##&nbsp;compliance with U.S. and local country laws. By using this product you
##&nbsp;agree to comply with applicable laws and regulations. If you are unable
##&nbsp;to comply with U.S. and local laws, return this product immediately.
##&nbsp;
##&nbsp;A summary of U.S. laws governing Cisco cryptographic products may be found at:
##&nbsp;http://www.cisco.com/wwl/export/crypto/tool/stqrg.html
##&nbsp;
##&nbsp;If you require further assistance please contact us by sending email to
##&nbsp;export@cisco.com.
##&nbsp;
##&nbsp;cisco WS-C2960-24TT-L (PowerPC405) processor (revision J0) with 65536K bytes of memory.
##&nbsp;Processor board ID FOC1505W4L6
##&nbsp;Last reset from power-on
##&nbsp;1 Virtual Ethernet interface
##&nbsp;24 FastEthernet interfaces
##&nbsp;2 Gigabit Ethernet interfaces
##&nbsp;The password-recovery mechanism is enabled.
##&nbsp;
##&nbsp;64K bytes of flash-simulated non-volatile configuration memory.
##&nbsp;Base ethernet MAC Address       : 44:E4:D9:DD:58:80
##&nbsp;Motherboard assembly number     : 73-12600-05
##&nbsp;Power supply part number        : 341-0097-03
##&nbsp;Motherboard serial number       : FOC15032RGG
##&nbsp;Power supply serial number      : DCA15028TAM
##&nbsp;Model revision number           : J0
##&nbsp;Motherboard revision number     : A0
##&nbsp;Model number                    : WS-C2960-24TT-L
##&nbsp;System serial number            : FOC1505W4L6
##&nbsp;Top Assembly Part Number        : 800-32797-01
##&nbsp;Top Assembly Revision Number    : F0
##&nbsp;Version ID                      : V09
##&nbsp;CLEI Code Number                : COM3L00BRE
##&nbsp;Hardware Board Revision Number  : 0x0A
##&nbsp;
##&nbsp;
##&nbsp;Switch Ports Model              SW Version            SW Image
##&nbsp;------ ----- -----              ----------            ----------
##&nbsp;*    1 26    WS-C2960-24TT-L    12.2(55)SE            C2960-LANBASEK9-M
##&nbsp;
##&nbsp;
##&nbsp;Configuration register is 0xF
##&nbsp;
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;          {
##&nbsp;            'compiled_by' => 'prod_rel_team',
##&nbsp;            'top_assembly_part_number' => '800-32797-01',
##&nbsp;            'model' => 'WS-C2960-24TT-L',
##&nbsp;            'boot_loader_version' => '12.2(53r)SEY3',
##&nbsp;            'boot_loader' => 'C2960-HBOOT-M',
##&nbsp;            'sw_image' => 'C2960-LANBASEK9-M',
##&nbsp;            'system_image_file' => 'flash:c2960-lanbasek9-mz.122-55.SE.bin',
##&nbsp;            'clei_code_number' => 'COM3L00BRE',
##&nbsp;            'uptime' => '4 weeks, 4 days, 12 hours, 17 minutes',
##&nbsp;            'rom' => 'Bootstrap program is C2960 boot loader',
##&nbsp;            'data_base' => '0x01800000',
##&nbsp;            'image_text_base' => '0x00003000',
##&nbsp;            'base_ethernet_mac_address' => '44:E4:D9:DD:58:80',
##&nbsp;            'cpu' => 'cisco WS-C2960-24TT-L (PowerPC405) processor (revision J0)',
##&nbsp;            'motherboard_serial_number' => 'FOC15032RGG',
##&nbsp;            'motherboard_assembly_number' => '73-12600-05',
##&nbsp;            'top_assembly_revision_number' => 'F0',
##&nbsp;            'version_id' => 'V09',
##&nbsp;            'power_supply_serial_number' => 'DCA15028TAM',
##&nbsp;            'ports' => '26',
##&nbsp;            'motherboard_revision_number' => 'A0',
##&nbsp;            'flash_simulated_non_volatile_configuration_memory' => '64K bytes',
##&nbsp;            'power_supply_part_number' => '341-0097-03',
##&nbsp;            'fastethernet_interface' => '24',
##&nbsp;            'configuration_register' => '0xF',
##&nbsp;            'hardware_board_revision_number' => '0x0A',
##&nbsp;            'processor_board_id' => 'FOC1505W4L6',
##&nbsp;            'sw_version' => '12.2(55)SE',
##&nbsp;            'compiled_date' => 'Sat 07-Aug-10 23:04',
##&nbsp;            'gigabit_ethernet_interface' => '2',
##&nbsp;            'model_revision_number' => 'J0',
##&nbsp;            'model_number' => 'WS-C2960-24TT-L',
##&nbsp;            'cisco_software' => 'IOS',
##&nbsp;            'virtual_ethernet_interface' => '1',
##&nbsp;            'technical_support' => 'http://www.cisco.com/techsupport',
##&nbsp;            'last_reset_from' => 'power-on',
##&nbsp;            'cpu_memory' => '65536K bytes',
##&nbsp;            'release_software' => 'fc2',
##&nbsp;            'system_serial_number' => 'FOC1505W4L6',
##&nbsp;            'password_recovery_mechanism' => 'enabled.',
##&nbsp;            'switch' => '*    1',
##&nbsp;            'system_returned_to_rom_by' => 'power-on'
##&nbsp;          }
##&nbsp;        ];
##&nbsp;
sub show_version {
    my %args = @_;

    my ($error_message, @result_array);
    my $resp        = $args{response_object};
    my $output      = $resp->get_processed_output();
    my $command     = $resp->get_command();
    my $error_level = $resp->get_execution_status();
    my ($error, $key, $i, $pos);

    $error = _check_error(
        raw_output    => $output,
        command       => $command,
        error_message => \$error_message
    );
    if ($error) {
        $error = 'NACL::APISet::Exceptions::' . $error . 'Exception';
        $error->throw(text => $error_message, response_object => $resp);
    }
    my @lines = get_lines(raw_output => $output);
    while (my $line = shift(@lines)) {
        if ($line =~ /(.*)\s+:(.*)/) {
            $key = $1;
            my $val = $2;
            $key = trim_string(string => "$key");
            $val = trim_string(string => "$val");
            $result_array[0]->{to_key_name('item' => $key)} = $val;
        } elsif ($line
            =~ /Cisco\s+(.*)\s+Software,\s+\w+\s+Software\s+\((.*)\),\s+Version\s+(.*),\s+RELEASE\s+SOFTWARE\s+\((.*)\)/
            ) {
            $result_array[0]->{cisco_software}   = $1;
            $result_array[0]->{sw_image}         = $2;
            $result_array[0]->{sw_version}       = $3;
            $result_array[0]->{release_software} = $4;
        } elsif ($line
            =~ /BOOTLDR:\s+.*\s+Boot\s+Loader\s+\((.*)\)\s+Version\s+(.*),\s+RELEASE SOFTWARE/i
            ) {
            $result_array[0]->{boot_loader}         = $1;
            $result_array[0]->{boot_loader_version} = $2;
        } elsif ($line
            =~ /Time\s+since\s+(.*)\s+switched\s+to\s+active\s+is\s+(.*)/) {
            $result_array[0]->{time_since_machine_switched_to_active} = $2;
        } elsif ($line
            =~ /System\s+returned to\s+ROM\s+by\s+(.*)\s+at\s+([0-9:A_Za-z\s]+)\s+/i
            ) {
            $result_array[0]->{system_returned_to_rom_by} = $1;
            $result_array[0]->{system_returned_to_rom_at} = $2;
        } elsif ($line =~ /(ROM|Technical\s+Support):\s+(.*)/) {
            $result_array[0]->{to_key_name('item' => $1)} = $2;
        } elsif ($line
            =~ /(.*)\s+CPU\s+at\s+(\d+Mhz),\s+Implementation\s+(.*),\s+Rev\s+(.*),\s+(\d+KB)\s+L2\s+Cache/
            ) {
            $result_array[0]->{cpu_type}       = $1;
            $result_array[0]->{cpu_speed}      = $2;
            $result_array[0]->{implementation} = $3;
            $result_array[0]->{revision}       = $4;
            $result_array[0]->{l2_cache}       = $5;
        } elsif ($line
            =~ /(Uptime\s+for\s+this\s+control\s+processor|password-recovery\s+mechanism|Configuration\s+register|System\s+image\s+file|uptime)\s+is\s+(.*)/
            ) {
            my $key = $1;
            my $val = $2;
            $key =~ s/-/_/g;
            $val =~ s/"//g;
            $result_array[0]->{to_key_name('item' => $key)} = $val;
        } elsif ($line =~ /(\d+)\s+(.*\s+interface)/
            || $line
            =~ /(.*)\s+of\s+(flash-simulated\s+non-volatile\s+configuration\s+memory)/
            ) {
            my $key = $2;
            my $val = $1;
            $key =~ s/-/_/g;
            $result_array[0]->{to_key_name('item' => $key)} = $val;
        } elsif ($line
            =~ /(Processor\s+Board\s+ID|Last\s+reset\s+from|System\s+returned\s+to\s+ROM\s+by|System\s+restarted\s+at)\s+(.*)/i
            ) {
            $result_array[0]->{to_key_name(item => "$1")} = $2;
        } elsif ($line =~ /Compiled\s+(.*)\s+by\s+(.*)/) {
            $result_array[0]->{compiled_date} = $1;
            $result_array[0]->{compiled_by}   = $2;
        } elsif ($line =~ /Image\s+text-base:\s+(.*),\s+data-base:\s+(.*)/) {
            $result_array[0]->{image_text_base} = $1;
            $result_array[0]->{data_base}       = $2;
        } elsif ($line =~ /(.*)\s+with\s+(.*)\s+of memory/) {
            $result_array[0]->{cpu_memory} = $2;
            $result_array[0]->{cpu}        = $1;
        } elsif ($line
            =~ /^(\d+K\s+bytes)\s+of\s+(([a-zA-Z-\s]+)(\((Sector\s+size)\s+(\d+K)\))?)./
            ) {
            $result_array[0]->{to_key_name('item' => "$5")} = $6
                if (defined($4));
            my $key_string = $3;
            $key_string = trim_string('string' => "$key_string");
            $result_array[0]->{to_key_name('item' => "$key_string")} = $1;
        } elsif ($line
            =~ /(Switch)\s+(Ports)\s+(Model)\s+(SW\s+Version)\s+(SW\s+Image)/)
        {
            my @keys = ('switch', 'ports', 'model', 'sw_version', 'sw_image');

            my $string = shift(@lines);
            $line = shift(@lines);
            my @vals = split(/\s+/, $line);
            my $k = 1;
            if ($string =~ /\s+/g) {
                $pos = pos($string);
                my $len = $pos - 0;
                my $val = substr($line, 0, $len);
                $val = trim_string('string' => $val);
                $result_array[0]->{$keys[0]} = $val;
            }
            shift(@vals);    #removing the * from line
            shift(@vals);    #removing te switch number from line
            while (my $val = shift(@vals)) {
                $result_array[0]->{$keys[$k++]} = $val;
            }
        }
    }
    return \@result_array;
} ## end sub show_version

## @name ::show_cdp_neighbors
##
## @summary Parses the raw output from the "show cdp neighbors" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from show_cdp_neighbors with any of its options as described in the
##    corresponding cdef file 'show.cdefs'
## @synopsis
##      show_cdp_neighbors(%args);
## INPUTS TO METHOD CALL 'show_cdp_neighbors' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

##&nbsp;Command = show cdp neighbors
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;Capability Codes: R - Router, T - Trans Bridge, B - Source Route Bridge
##&nbsp;                  S - Switch, H - Host, I - IGMP, r - Repeater, P - Phone,
##&nbsp;                  D - Remote, C - CVTA, M - Two-port Mac Relay
##&nbsp;
##&nbsp;Device ID        Local Intrfce     Holdtme    Capability  Platform  Port ID
##&nbsp;net-5020-2(SSI140106DF)
##&nbsp;                 Fas 0/22          179             S I C  N5K-C5020 mgmt0
##&nbsp;net-2960-4       Gig 0/1           120              S I   WS-C2960- Gig 0/1
##&nbsp;net-car-9a       Fas 0/2           121               H    FAS3240   e0M
##&nbsp;net-abs-5a       Fas 0/1           158               H    FAS6240   e0M
##&nbsp;net-car-3a       Fas 0/5           166               H    FAS3270   e0M
##&nbsp;net-car-3a       Fas 0/7           166               H    FAS3270   e0b
##&nbsp;net-car-3b       Fas 0/3           133               H    FAS3270   e0b
##&nbsp;f13-edge-sw1.lab.netapp.com
##&nbsp;                 Gig 0/2           172              S I   WS-C4948- Gig 1/1
##&nbsp;net-car-10a      Fas 0/15          162               H    FAS3210   e0M
##&nbsp;net-5010-3(SSI144906J5)
##&nbsp;                 Fas 0/24          179             S I C  N5K-C5010 mgmt0
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;          {
##&nbsp;            'device_id' => 'net-5020-2(SSI140106DF)',
##&nbsp;            'holdtime' => '179',
##&nbsp;            'port_id' => 'mgmt0',
##&nbsp;            'interface' => 'Fas 0/22',
##&nbsp;            'platform' => 'N5K-C5020',
##&nbsp;            'capability' => [
##&nbsp;                              'Switch',
##&nbsp;                              'IGMP',
##&nbsp;                              'CVTA'
##&nbsp;                            ]
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'device_id' => 'net-2960-4',
##&nbsp;            'holdtime' => '120',
##&nbsp;            'port_id' => 'Gig 0/1',
##&nbsp;            'interface' => 'Gig 0/1',
##&nbsp;            'platform' => 'WS-C2960-',
##&nbsp;            'capability' => [
##&nbsp;                              'Switch',
##&nbsp;                              'IGMP'
##&nbsp;                            ]
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'device_id' => 'net-car-9a',
##&nbsp;            'holdtime' => '121',
##&nbsp;            'port_id' => 'e0M',
##&nbsp;            'interface' => 'Fas 0/2',
##&nbsp;            'platform' => 'FAS3240',
##&nbsp;            'capability' => [
##&nbsp;                              'Host'
##&nbsp;                            ]
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'device_id' => 'net-abs-5a',
##&nbsp;            'holdtime' => '158',
##&nbsp;            'port_id' => 'e0M',
##&nbsp;            'interface' => 'Fas 0/1',
##&nbsp;            'platform' => 'FAS6240',
##&nbsp;            'capability' => [
##&nbsp;                              'Host'
##&nbsp;                            ]
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'device_id' => 'net-car-3a',
##&nbsp;            'holdtime' => '166',
##&nbsp;            'port_id' => 'e0M',
##&nbsp;            'interface' => 'Fas 0/5',
##&nbsp;            'platform' => 'FAS3270',
##&nbsp;            'capability' => [
##&nbsp;                              'Host'
##&nbsp;                            ]
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'device_id' => 'net-car-3a',
##&nbsp;            'holdtime' => '166',
##&nbsp;            'port_id' => 'e0b',
##&nbsp;            'interface' => 'Fas 0/7',
##&nbsp;            'platform' => 'FAS3270',
##&nbsp;            'capability' => [
##&nbsp;                              'Host'
##&nbsp;                            ]
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'device_id' => 'net-car-3b',
##&nbsp;            'holdtime' => '133',
##&nbsp;            'port_id' => 'e0b',
##&nbsp;            'interface' => 'Fas 0/3',
##&nbsp;            'platform' => 'FAS3270',
##&nbsp;            'capability' => [
##&nbsp;                              'Host'
##&nbsp;                            ]
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'device_id' => 'f13-edge-sw1.lab.netapp.com',
##&nbsp;            'holdtime' => '172',
##&nbsp;            'port_id' => 'Gig 1/1',
##&nbsp;            'interface' => 'Gig 0/2',
##&nbsp;            'platform' => 'WS-C4948-',
##&nbsp;            'capability' => [
##&nbsp;                              'Switch',
##&nbsp;                              'IGMP'
##&nbsp;                            ]
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'device_id' => 'net-car-10a',
##&nbsp;            'holdtime' => '162',
##&nbsp;            'port_id' => 'e0M',
##&nbsp;            'interface' => 'Fas 0/15',
##&nbsp;            'platform' => 'FAS3210',
##&nbsp;            'capability' => [
##&nbsp;                              'Host'
##&nbsp;                            ]
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'device_id' => 'net-5010-3(SSI144906J5)',
##&nbsp;            'holdtime' => '179',
##&nbsp;            'port_id' => 'mgmt0',
##&nbsp;            'interface' => 'Fas 0/24',
##&nbsp;            'platform' => 'N5K-C5010',
##&nbsp;            'capability' => [
##&nbsp;                              'Switch',
##&nbsp;                              'IGMP',
##&nbsp;                              'CVTA'
##&nbsp;                            ]
##&nbsp;          }
##&nbsp;        ];
##&nbsp;
sub show_cdp_neighbors {
    my %args = @_;
    my ($error_message, @result_array);
    my $resp    = $args{response_object};
    my $output  = $resp->get_processed_output();
    my $command = $resp->get_command();
    my $error   = _check_error(
        raw_output    => $output,
        error_message => \$error_message
    );
    if ($error) {
        $error = 'NACL::APISet::Exceptions::' . $error . 'Exception';
        $error->throw(text => $error_message, response_object => $resp);
    }
    my @lines = get_lines(raw_output => $output, response_object => $resp);
    my (@values, @codes, %capab_codes, @temp);
    my $count = 0;
    my $pos   = -1;
    my $j     = 0;
    my @capabs;
    my @head = qw(device_id interface holdtime capability platform port_id );

    for (my $i = 0; $i <= $#lines; $i++) {

        if ($lines[$i] =~ /Capability Codes: /) {
            until ($lines[$i + $j] =~ /Device ID/) {
                @temp = split(/,\s+/, $lines[$i + $j++]);
                push(@codes, @temp);
            }
            foreach my $cod (@codes) {
                @temp = split(/\s+-\s+/, $cod);
                $capab_codes{$temp[0]} = $temp[1];
            }
            $i = $i + $j;
        }
        if ($lines[$i] =~ /Device ID/) {
            $i++;
        }
        if ($lines[$i] =~ /\(/ || $lines[$i] =~ /\.com/) {
            $result_array[$count]->{to_key_name(item => $head[0])}
                = $lines[$i];
            $i++;
            if ($lines[$i]
                =~ /(\S+\s\S+)\s+(\d+)\s+([\S\s]+)\s\s+(\S+)\s+(\S+\s*\S*)+/)
            {
                push(@values, $1, $2, $3, $4, $5);
                @capabs = split(/\s/, $3);
                for (my $j = 1; $j <= $#head; $j++) {
                    if ($head[$j] eq "capability") {
                        for (my $k = 0; $k <= $#capabs; $k++) {
                            foreach my $key (keys(%capab_codes)) {
                                if ($capabs[$k] eq $key) {
                                    $capabs[$k] = $capab_codes{$key};
                                }
                            }
                        }
                        push(@{$result_array[$count]->{capability}}, @capabs);
                        $j++;
                    }
                    $result_array[$count]->{$head[$j]} = $values[$j - 1]

                }
                $count++;
            }
        }
        if ($lines[$i]
            =~ /(\S+)\s+(\S+\s\S+)\s+(\d+)\s+([\S\s]+)\s\s+(\S+)\s+(\S+\s*\S*)+/
            ) {
            push(@values, $1, $2, $3, $4, $5, $6);
            @capabs = split(/\s/, $4);
            for (my $j = 0; $j <= $#head; $j++) {
                if ($head[$j] eq "capability") {
                    for (my $k = 0; $k <= $#capabs; $k++) {
                        foreach my $key (keys(%capab_codes)) {
                            if ($capabs[$k] eq $key) {
                                $capabs[$k] = $capab_codes{$key};
                            }
                        }
                    }
                    push(@{$result_array[$count]->{capability}}, @capabs);
                    $j++;
                }
                $result_array[$count]->{$head[$j]} = $values[$j];
            }
            $count++;
        }
        $#values = -1;
    }
    return \@result_array;
} ## end sub show_cdp_neighbors
## @name ::show_ip_interface
##
## @summary Parses the raw output from the "show ip interface" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from show_ip_interface with any of its options as described in the
##    corresponding cdef file 'show.cdefs'
## @synopsis
##      show_ip_interface(%args);
## INPUTS TO METHOD CALL 'show_ip_interface' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## <tr> <td> No switch for this key </td> <td> 'field' </td> <td> '0' </td> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

##&nbsp;Command = show ip interface brief
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;Interface              IP-Address      OK? Method Status                Protocol
##&nbsp;Vlan1                  172.22.200.110  YES NVRAM  up                    up
##&nbsp;FastEthernet0/1        unassigned      YES unset  up                    up
##&nbsp;FastEthernet0/2        unassigned      YES unset  up                    up
##&nbsp;FastEthernet0/3        unassigned      YES unset  up                    up
##&nbsp;FastEthernet0/4        unassigned      YES unset  down                  down
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;          {
##&nbsp;            'protocol' => 'up',
##&nbsp;            'ok' => 'YES',
##&nbsp;            'status' => 'up',
##&nbsp;            'interface' => 'Vlan1',
##&nbsp;            'method' => 'NVRAM',
##&nbsp;            'ip_address' => '172.22.200.110'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'protocol' => 'up',
##&nbsp;            'ok' => 'YES',
##&nbsp;            'status' => 'up',
##&nbsp;            'interface' => 'FastEthernet0/1',
##&nbsp;            'method' => 'unset',
##&nbsp;            'ip_address' => 'unassigned'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'protocol' => 'up',
##&nbsp;            'ok' => 'YES',
##&nbsp;            'status' => 'up',
##&nbsp;            'interface' => 'FastEthernet0/2',
##&nbsp;            'method' => 'unset',
##&nbsp;            'ip_address' => 'unassigned'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'protocol' => 'up',
##&nbsp;            'ok' => 'YES',
##&nbsp;            'status' => 'up',
##&nbsp;            'interface' => 'FastEthernet0/3',
##&nbsp;            'method' => 'unset',
##&nbsp;            'ip_address' => 'unassigned'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'protocol' => 'down',
##&nbsp;            'ok' => 'YES',
##&nbsp;            'status' => 'down',
##&nbsp;            'interface' => 'FastEthernet0/4',
##&nbsp;            'method' => 'unset',
##&nbsp;            'ip_address' => 'unassigned'
##&nbsp;          },
##&nbsp;        ];
##&nbsp;
sub show_ip_interface {
    my %args = @_;
    my ($error_message, @result_array);
    my $resp   = $args{response_object};
    my $output = $resp->get_processed_output();
    my $error  = _check_error(
        raw_output    => $output,
        error_message => \$error_message
    );
    if ($error) {
        $error = 'NACL::APISet::Exceptions::' . $error . 'Exception';
        $error->throw(text => $error_message, response_object => $resp);
    }
    my @lines = get_lines(raw_output => $output, response_object => $resp);
    $lines[0] =~ tr/-?/_ /;
    my @headings = split(/\s+/, $lines[0]);
    my $count = 0;
    for (my $k = 1; $k <= $#lines; $k++) {
        my @values = split(/\s+/, $lines[$k]);
        for (my $i = 0; $i <= $#headings; $i++) {
            $result_array[$count]->{to_key_name(item => $headings[$i])}
                = $values[$i];
        }
        $count++;
    }
    return \@result_array;
} ## end sub show_ip_interface

## @name ::show_boot
##
## @summary Parses the raw output from the "show boot" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from show_boot with any of its options as described in the
##    corresponding cdef file 'show.cdefs'
## @synopsis
##      show_boot(%args);
## INPUTS TO METHOD CALL 'show_boot' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

##&nbsp;Command = show boot
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;BOOT path-list      : flash:c2960-lanbasek9-mz.122-55.SE.bin
##&nbsp;Config file         : flash:/config.text
##&nbsp;Private Config file : flash:/private-config.text
##&nbsp;Enable Break        : no
##&nbsp;Manual Boot         : no
##&nbsp;HELPER path-list    :
##&nbsp;Auto upgrade        : yes
##&nbsp;Auto upgrade path   :
##&nbsp;NVRAM/Config file
##&nbsp;      buffer size:   65536
##&nbsp;Timeout for Config
##&nbsp;          Download:    0 seconds
##&nbsp;Config Download
##&nbsp;       via DHCP:       disabled (next boot: disabled)
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;          {
##&nbsp;            'private_config_file' => 'flash:/private-config.text',
##&nbsp;            'auto_upgrade_path' => '--',
##&nbsp;            'enable_break' => 'no',
##&nbsp;            'timeout_for_config' => [
##&nbsp;                                      {
##&nbsp;                                        'download' => '0 seconds'
##&nbsp;                                      }
##&nbsp;                                    ],
##&nbsp;            'nvram_config_file' => [
##&nbsp;                                     {
##&nbsp;                                       'buffer_size' => '65536'
##&nbsp;                                     }
##&nbsp;                                   ],
##&nbsp;            'auto_upgrade' => 'yes',
##&nbsp;            'config_file' => 'flash:/config.text',
##&nbsp;            'boot_path_list' => 'flash:c2960-lanbasek9-mz.122-55.SE.bin',
##&nbsp;            'manual_boot' => 'no',
##&nbsp;            'config_download' => [
##&nbsp;                                   {
##&nbsp;                                     'via_dhcp' => 'disabled',
##&nbsp;                                     'next_boot' => 'disabled'
##&nbsp;                                   }
##&nbsp;                                 ],
##&nbsp;            'helper_path_list' => '--'
##&nbsp;          }
##&nbsp;        ];
##&nbsp;
sub show_boot {
    my %args = @_;
    my ($error_message, @result_array);
    my $resp   = $args{response_object};
    my $output = $resp->get_processed_output();
    my $error  = _check_error(
        raw_output    => $output,
        error_message => \$error_message
    );
    if ($error) {
        $error = 'NACL::APISet::Exceptions::' . $error . 'Exception';
        $error->throw(text => $error_message, response_object => $resp);
    }
    my @lines
        = get_lines_tree(raw_output => $output, response_object => $resp);
    my $heading2;
    my (@values, @head, @result_array);
    for (my $i = 0; $i <= $#lines; $i++) {
        if (   $lines[$i] =~ /(NVRAM.*file)/
            || $lines[$i] =~ /(Timeout.*Config)/
            || $lines[$i] =~ /(Con.*oad)/) {
            $heading2 = $1;
            $heading2 =~ s/\//_/g;
        } elsif (defined($heading2)
            && $lines[$i] =~ /\s\s+(.*):\s+(\w+)\s+\((.*)\)/) {
            my $value = $3;
            @values = split(/:\s+/, $value);
            $result_array[0]->{to_key_name(item => $heading2)}[0]
                {to_key_name(item => $values[0])} = $values[1];
            $result_array[0]
                ->{to_key_name(item => $heading2)}[0]{to_key_name(item => $1)}
                = $2;
        } elsif (defined($heading2) && $lines[$i] =~ /\s\s+(.*):\s\s+(.*)/) {
            $result_array[0]
                ->{to_key_name(item => $heading2)}[0]{to_key_name(item => $1)}
                = $2;
        } else {
            @values = split(/\s+:\s*/, $lines[$i]);
            $values[0] =~ s/-/_/g;
            $result_array[0]->{to_key_name(item => $values[0])}
                = $values[1] ? $values[1] : '--';
        }
    }
    return \@result_array;
} ## end sub show_boot

sub get_lines_tree(%) {
    my %args       = @_;
    my $raw_output = $args{'raw_output'};
    my $first_line = $args{'first_line'};
    my $can_throw  = $args{'can_throw'};
    my @lines;

    if ($raw_output = snip_only_required_output(
            'raw_output' => $raw_output,
            'delimiter'  => $first_line
        )
        ) {
        @lines = split(/\r*\n+/, $raw_output);
        map(s/\s+$//, @lines);
        map(s/\cM//g, @lines);
        @lines = grep (/\S+/, @lines);
    } elsif ($can_throw) {
        my $response_object = $args{'response_object'};
        NACL::APISet::Exceptions::UnexpectedOutputException->throw(
            text            => "output is not in the expected format",
            response_object => $response_object
        );
    }
    return @lines;
} ## end sub get_lines_tree(%)

1;
