#
#Copyright (c) 2010-2016 Network Appliance, Inc.
#All rights reserved.
#
## @summary Routines to parse CiscoNetwork show and its subcommands
##
## @description
##    This is a library to parse the raw output of CiscoNetwork switch show and its subcommands.
##
## @synopsis
##  use NACL::APISet::Switch::CLI::CiscoNetwork::default::show
##
##  show_mac_address_table(%args)
##  show_version(%args)
##
## @author dl-nacl-dev@netapp.com
##
## @status public
#
package NACL::APISet::Switch::CLI::CiscoNetwork::Nexus::show;

use strict;
use NACL::APISet::Response::CLI::ParserUtils;
use Data::Dumper;

#Include exceptions which will be thrown
use NACL::APISet::Exceptions::CommandFailedException qw(:try);
use NACL::APISet::Exceptions::MissingRequiredParamException;
use NACL::APISet::Exceptions::InvalidParamException;
use NACL::APISet::Exceptions::InvalidParamValueException;

sub _check_error {
    my %args = @_;

    my $raw_output    = $args{raw_output};
    my $command       = $args{command};
    my $error_message = $args{error_message};
    my $error;
    # Check for the error
    if ( $raw_output =~ /(Invalid\s+input\s+detected)|(Invalid\s+command)/i )
    {
        $$error_message = $raw_output;
        $error          = "InvalidParam";
    } elsif ( $raw_output =~ /(Incomplete\s+command)/i ) {
        $$error_message = $raw_output;
        $error          = "MissingRequiredParam";
    } elsif ( $raw_output =~ /(Error.*\s+SUP\s+slot)|(Invalid\s+range\s+at\s+\S+\s+marker)|(VLAN\s+\d+\s+not\s+found\s+in\s+current\s+VLAN\s+database)|(Invalid\s+number)|(absent)/i ) {
        $$error_message = $raw_output;
        $error          = "InvalidParamValue";
    }

    return $error;
}

## @name ::show_interface
##
## @summary Parses the raw output from the "show interface" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from show_interface with any of its options as described in the
##    corresponding cdef file 'show.cdefs'
## @synopsis
##      show_interface(%args);
## INPUTS TO METHOD CALL 'show_interface' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## <tr> <td> No switch for this key </td> <td> 'entity-type' </td> <td> '0' </td> </tr>
## <tr> <td> No switch for this key </td> <td> 'field' </td> <td> '0' </td> </tr>
## <tr> <td> No switch for this key </td> <td> 'entity-name' </td> <td> '0' </td> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples
##&nbsp;Command = show interface Ethernet1/4 brief
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;--------------------------------------------------------------------------------
##&nbsp;Ethernet      VLAN   Type Mode   Status  Reason                   Speed     Port
##&nbsp;Interface                                                                   Ch #
##&nbsp;--------------------------------------------------------------------------------
##&nbsp;Eth1/4        1      eth  access down    Link not connected          10G(D) --
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;          {
##&nbsp;            'port_ch_no' => '--',
##&nbsp;            'status' => 'down',
##&nbsp;            'mode' => 'access',
##&nbsp;            'reason' => 'link_not_connected',
##&nbsp;            'vlan' => '1',
##&nbsp;            'interface_type' => 'Eth',
##&nbsp;            'speed' => '10G(D)',
##&nbsp;            'interface' => 'Eth1/4',
##&nbsp;            'interface_num' => '1/4',
##&nbsp;            'type' => 'eth'
##&nbsp;          }
##&nbsp;        ];
##&nbsp;
##&nbsp;Command = show interface Ethernet1/4 flowcontrol
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;--------------------------------------------------------------------------------
##&nbsp;Port       Send FlowControl  Receive FlowControl  RxPause TxPause
##&nbsp;           admin    oper     admin    oper
##&nbsp;--------------------------------------------------------------------------------
##&nbsp;Eth1/4       off      off      off      off         0                 0
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;          {
##&nbsp;            'tx_pause' => '0',
##&nbsp;            'interface_type' => 'Eth',
##&nbsp;            'rx_pause' => '0',
##&nbsp;            'fc_send' => {
##&nbsp;                           'admin' => 'off',
##&nbsp;                           'oper' => 'off'
##&nbsp;                         },
##&nbsp;            'fc_receive' => {
##&nbsp;                              'admin' => 'off',
##&nbsp;                              'oper' => 'off'
##&nbsp;                            },
##&nbsp;            'interface_num' => '1/4',
##&nbsp;            'port' => 'Eth1/4'
##&nbsp;          }
##&nbsp;        ];
##&nbsp;
##&nbsp;Command = show interface Ethernet1/4 status
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;--------------------------------------------------------------------------------
##&nbsp;Port          Name               Status    Vlan      Duplex  Speed   Type
##&nbsp;--------------------------------------------------------------------------------
##&nbsp;Eth1/4        --                 notconnec 1         full    10G     1/10g
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;          {
##&nbsp;            'status' => 'notconnec',
##&nbsp;            'name' => '--',
##&nbsp;            'vlan' => '1',
##&nbsp;            'port' => 'Eth1/4',
##&nbsp;            'duplex' => 'full',
##&nbsp;            'interface_type' => 'Eth',
##&nbsp;            'speed' => '10G',
##&nbsp;            'type' => '1/10g',
##&nbsp;            'interface_num' => '1/4'
##&nbsp;          }
##&nbsp;        ];
##&nbsp;
##&nbsp;Command = show interface ethernet1/1
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;Ethernet1/1 is up
##&nbsp;  Hardware: 1000/10000 Ethernet, address: 0005.73d2.8448 (bia 0005.73d2.8448)
##&nbsp;  Description: Cluster Node 1
##&nbsp;  MTU 1500 bytes, BW 10000000 Kbit, DLY 10 usec,
##&nbsp;     reliability 255/255, txload 1/255, rxload 1/255
##&nbsp;  Encapsulation ARPA
##&nbsp;  Port mode is access
##&nbsp;  full-duplex, 10 Gb/s, media type is 10G
##&nbsp;  Beacon is turned off
##&nbsp;  Input flow-control is off, output flow-control is off
##&nbsp;  Rate mode is dedicated
##&nbsp;  Switchport monitor is off
##&nbsp;  EtherType is 0x8100
##&nbsp;  Last link flapped 04:03:33
##&nbsp;  Last clearing of "show interface" counters never
##&nbsp;  30 seconds input rate 63872 bits/sec, 7984 bytes/sec, 4 packets/sec
##&nbsp;  30 seconds output rate 23768 bits/sec, 2971 bytes/sec, 4 packets/sec
##&nbsp;  Load-Interval #2: 5 minute (300 seconds)
##&nbsp;    input rate 66.74 Kbps, 4 pps; output rate 24.03 Kbps, 3 pps
##&nbsp;  RX
##&nbsp;    521740 unicast packets  435 multicast packets  22 broadcast packets
##&nbsp;    522197 input packets  3563218246 bytes
##&nbsp;    432108 jumbo packets  0 storm suppression packets
##&nbsp;    0 runts  0 giants  0 CRC  0 no buffer
##&nbsp;    0 input error  0 short frame  0 overrun   0 underrun  0 ignored
##&nbsp;    0 watchdog  0 bad etype drop  0 bad proto drop  0 if down drop
##&nbsp;    0 input with dribble  0 input discard
##&nbsp;    0 Rx pause
##&nbsp;  TX
##&nbsp;    488466 unicast packets  9150 multicast packets  400 broadcast packets
##&nbsp;    498016 output packets  74488950 bytes
##&nbsp;    2791 jumbo packets
##&nbsp;    0 output errors  0 collision  0 deferred  0 late collision
##&nbsp;    0 lost carrier  0 no carrier  0 babble
##&nbsp;    0 Tx pause
##&nbsp;  111 interface resets
##&nbsp;
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;          {
##&nbsp;            'reliability' => '255/255',
##&nbsp;            '30_seconds_output_rate' => [
##&nbsp;                                          {
##&nbsp;                                            'in_bytes_per_sec' => '2971',
##&nbsp;                                            'in_packets_per_sec' => '4',
##&nbsp;                                            'in_bits_per_sec' => '23768'
##&nbsp;                                          }
##&nbsp;                                        ],
##&nbsp;            'ethertype' => '0x8100',
##&nbsp;            'last_link_flapped' => '04:03:33',
##&nbsp;            'dly' => '10 usec',
##&nbsp;            'rate_mode' => 'dedicated',
##&nbsp;            'hardware' => '1000/10000 Ethernet',
##&nbsp;            'switchport_monitor' => 'off',
##&nbsp;            'interface' => 'Ethernet1/1',
##&nbsp;            'address' => '0005.73d2.8448',
##&nbsp;            'mtu' => '1500',
##&nbsp;            'description' => 'Cluster Node 1',
##&nbsp;            'duplex' => 'full',
##&nbsp;            'rxload' => '1/255',
##&nbsp;            'input_flow_control' => 'off',
##&nbsp;            'encapsulation' => 'ARPA',
##&nbsp;            'output_rate' => [
##&nbsp;                               {
##&nbsp;                                 'in_pps' => '3',
##&nbsp;                                 'in_Kbps' => '24.03'
##&nbsp;                               }
##&nbsp;                             ],
##&nbsp;            'beacon' => 'off',
##&nbsp;            '30_seconds_input_rate' => [
##&nbsp;                                         {
##&nbsp;                                           'in_bytes_per_sec' => '7984',
##&nbsp;                                           'in_packets_per_sec' => '4',
##&nbsp;                                           'in_bits_per_sec' => '63872'
##&nbsp;                                         }
##&nbsp;                                       ],
##&nbsp;            'status' => 'up',
##&nbsp;            'load_interval' => '5 minute (300 seconds)',
##&nbsp;            'bia' => '0005.73d2.8448',
##&nbsp;            'speed' => '10 Gb/s',
##&nbsp;            'media_type' => '10G',
##&nbsp;            'interface_num' => '1/1',
##&nbsp;            'output_flow_control' => 'off',
##&nbsp;            'port_mode' => 'access',
##&nbsp;            'tx' => [
##&nbsp;                      {
##&nbsp;                        'collision' => '0',
##&nbsp;                        'bytes' => '74488950',
##&nbsp;                        'lost_carrier' => '0',
##&nbsp;                        'output_packets' => '498016',
##&nbsp;                        'unicast_packets' => '488466',
##&nbsp;                        'tx_pause' => '0',
##&nbsp;                        'multicast_packets' => '9150',
##&nbsp;                        'late_collision' => '0',
##&nbsp;                        'deferred' => '0',
##&nbsp;                        'output_errors' => '0',
##&nbsp;                        'babble' => '0',
##&nbsp;                        'jumbo_packets' => '2791',
##&nbsp;                        'no_carrier' => '0',
##&nbsp;                        'broadcast_packets' => '400'
##&nbsp;                      }
##&nbsp;                    ],
##&nbsp;            'rx' => [
##&nbsp;                      {
##&nbsp;                        'bad_proto_drop' => '0',
##&nbsp;                        'bytes' => '3563218246',
##&nbsp;                        'rx_pause' => '0',
##&nbsp;                        'input_discard' => '0',
##&nbsp;                        'unicast_packets' => '521740',
##&nbsp;                        'underrun' => '0',
##&nbsp;                        'giants' => '0',
##&nbsp;                        'multicast_packets' => '435',
##&nbsp;                        'input_packets' => '522197',
##&nbsp;                        'jumbo_packets' => '432108',
##&nbsp;                        'short_frame' => '0',
##&nbsp;                        'no_buffer' => '0',
##&nbsp;                        'input_with_dribble' => '0',
##&nbsp;                        'if_down_drop' => '0',
##&nbsp;                        'crc' => '0',
##&nbsp;                        'ignored' => '0',
##&nbsp;                        'overrun' => '0',
##&nbsp;                        'watchdog' => '0',
##&nbsp;                        'input_error' => '0',
##&nbsp;                        'bad_etype_drop' => '0',
##&nbsp;                        'storm_suppression_packets' => '0',
##&nbsp;                        'runts' => '0',
##&nbsp;                        'broadcast_packets' => '22'
##&nbsp;                      }
##&nbsp;                    ],
##&nbsp;            'txload' => '1/255',
##&nbsp;            'last_clearing_of_show_interface_counters' => 'never',
##&nbsp;            'bw' => '10000000 Kbit',
##&nbsp;            'interface_type' => 'Ethernet',
##&nbsp;            'input_rate' => [
##&nbsp;                              {
##&nbsp;                                'in_pps' => '4',
##&nbsp;                                'in_Kbps' => '66.74'
##&nbsp;                              }
##&nbsp;                            ],
##&nbsp;            'interface_resets' => '111'
##&nbsp;          }
##&nbsp;        ];
##&nbsp;
sub show_interface {
    my %args = @_;

    my ( $error_message, @result_array );
    my $resp        = $args{response_object};
    my $output      = $resp->get_processed_output();
    my $command     = $resp->get_command();
    my $error_level = $resp->get_execution_status();
    my ( $error, $key );

    # Status returned by switch for this command is
    # is ZERO even if command failed
    if ( not defined($error) ) {
        $error = _check_error(
            raw_output    => $output,
            command       => "show interface",
            error_message => \$error_message
        );
    }
    if ($error) {
        $error = 'NACL::APISet::Exceptions::' . $error . 'Exception';
        $error->throw( text => $error_message, response_object => $resp );
    }

    chomp($command);
    my $line;
    my @lines;
    if ( $command =~ /brief$/ ) {
        @lines = get_lines( raw_output => $output );
        my @headings
            = qw(interface vlan type mode status reason speed port_ch_no interface_type interface_num);
        foreach $line (@lines) {
            chomp($line);
            next if ( $line !~ /^Eth\d+/i );
            my %hash;
            my @array = split( /\s+/, $line );

            if ( $line =~ /SFP\s+not\s+inserted/i ) {
                splice( @array, 5, 3, "SFP_not_inserted" );
            } elsif ( $line =~ /Link\s+not\s+connected/i ) {
                splice( @array, 5, 3, "link_not_connected" );
            } elsif ( $line =~ /SFP validation failed/i ) {
                splice( @array, 5, 3, "SFP_validation_failed" );
            }

            my $if = $array[0];
            if ( $if =~ /([a-zA-Z]+)([\d|\/]+)/ ) {
                push( @array, $1, $2 );
            }
            for ( my $i = 0; $i <= $#array; $i++ ) {
                $hash{ $headings[$i] } = $array[$i];
            }
            push( @result_array, \%hash );
        }
    } elsif ( $command =~ /status$/ ) {
        @lines = get_lines( raw_output => $output, first_line => 'Port' );
        my $head = shift(@lines);
        my @key = $head =~ /(\w+\s+|\w+)/g;
        shift(@lines);
        my @headings
            = qw(port name status vlan duplex speed type interface_type interface_num);
        foreach $line (@lines) {
            chomp($line);
            my %hash;
            my $l = 0;
            for ( my $i = 0; $i < scalar(@key); $i++ ) {
                if ( $i != scalar(@key) - 1 ) {
                    $hash{ $headings[$i] }
                        = substr( $line, $l, length( $key[$i] ) );
                } else {
                    $hash{ $headings[$i] } = substr( $line, $l );
                }
                $l = $l + length( $key[$i] );
                $hash{ $headings[$i] } =~ s/^\s+//;
                $hash{ $headings[$i] } =~ s/\s+$//;
            }

            # separate interface_type and interface_num from port
            if ( $hash{port} =~ /([a-zA-Z]+)([\d|\/]+)/ ) {
                $hash{interface_type} = $1;
                $hash{interface_num}  = $2;
            }
            push( @result_array, \%hash );
        }
    } elsif ( $command =~ /flowcontrol$/ ) {
        @lines = get_lines( raw_output => $output, first_line => 'Port' );
        shift(@lines);
        shift(@lines);
        shift(@lines);
        foreach $line (@lines) {
            my %hash;
            chomp($line);
            my @array = split( /\s+/, $line );
            my $if = $array[0];
            if ( $if =~ /([a-zA-Z]+)([\d|\/]+)/ ) {
                $hash{'interface_type'} = $1;
                $hash{'interface_num'}  = $2;
            }
            $hash{'port'}                = $array[0];
            $hash{'fc_send'}{'admin'}    = $array[1];
            $hash{'fc_send'}{'oper'}     = $array[2];
            $hash{'fc_receive'}{'admin'} = $array[3];
            $hash{'fc_receive'}{'oper'}  = $array[4];
            $hash{'rx_pause'}            = $array[5];
            $hash{'tx_pause'}            = $array[6];
            push( @result_array, \%hash );
        }
    } else {
        @lines = get_lines( raw_output => $output );
        my $i = -1;
        my $direction;
        foreach $line (@lines) {
            next if ( !$line || !defined($line) );
            if ( $line =~ /(([a-zA-Z\-]+)([\d|\/]+))\s+is\s+(\w+)/ ) {
                $i++;
                $result_array[$i]->{interface}      = $1;
                $result_array[$i]->{interface_type} = $2;
                $result_array[$i]->{interface_num}  = $3;
                $result_array[$i]->{status}         = $4;
            } elsif ( $line
                =~ /Hardware\:\s*(.+)\,\s*address\:\s*(.+)\s+\(bia\s*(.+)\)/ )
            {
                $result_array[$i]->{hardware} = $1;
                $result_array[$i]->{address}  = $2;
                $result_array[$i]->{bia}      = $3;
            } elsif ( $line =~ /Description\:\s*(.*)/ ) {
                $result_array[$i]->{description} = $1;
            } elsif ( $line
                =~ /MTU\s*(\d+)\s*bytes\,\s*BW\s*(.+)\,\s*DLY\s+(.+)\,?/ )
            {
                $result_array[$i]->{mtu} = $1;
                $result_array[$i]->{bw}  = $2;
                $result_array[$i]->{dly} = $3;
            } elsif ( $line
                =~ /reliability\s*(.+)\,\s*txload\s*(.+)\,\s*rxload\s+(.+)/ )
            {
                $result_array[$i]->{reliability} = $1;
                $result_array[$i]->{txload}      = $2;
                $result_array[$i]->{rxload}      = $3;
            } elsif ( $line =~ /Encapsulation\s*(.+)/ ) {
                $result_array[$i]->{encapsulation} = $1;
            } elsif ( $line =~ /Port\s+mode\s+is\s+(.+)/ ) {
                $result_array[$i]->{port_mode} = $1;
            } elsif (
                $line =~ /(\w+)\-duplex\,\s*(.+)\,\s*media\stype\sis\s(\w+)/ )
            {
                $result_array[$i]->{duplex}     = $1;
                $result_array[$i]->{speed}      = $2;
                $result_array[$i]->{media_type} = $3;
            } elsif ( $line =~ /(\w+)\-duplex\,\s*(.+)/ ) {
                $result_array[$i]->{duplex} = $1;
                $result_array[$i]->{speed}  = $2;
            } elsif ( $line =~ /Beacon\s+is\s+turned\s+(.+)/ ) {
                $result_array[$i]->{beacon} = $1;
            } elsif ( $line
                =~ /Input\sflow\-control\sis\s(\w+)\,\soutput\sflow\-control\sis\s(\w+)/
                )
            {
                $result_array[$i]->{input_flow_control}  = $1;
                $result_array[$i]->{output_flow_control} = $2;
            } elsif ( $line =~ /Rate\s+mode\s+is\s+(.+)/ ) {
                $result_array[$i]->{rate_mode} = $1;
            } elsif ( $line =~ /Switchport\smonitor\sis\s(\w+)/ ) {
                $result_array[$i]->{switchport_monitor} = $1;
            } elsif ( $line =~ /EtherType\sis\s(\w+)/ ) {
                $result_array[$i]->{ethertype} = $1;
            } elsif ( $line =~ /Last\slink\sflapped\s(.+)/ ) {
                $result_array[$i]->{last_link_flapped} = $1;
            } elsif ( $line
                =~ /Last\sclearing\sof\s\"show\sinterface\"\scounters\s(.+)/ )
            {
                $result_array[$i]->{last_clearing_of_show_interface_counters}
                    = $1;
            } elsif ( $line
                =~ /(\d+)\sseconds\sinput\srate\s(\d+)\sbits\/sec\,\s(\d+)\sbytes\/sec\,\s(\d+)\spackets\/sec/
                )
            {
                $result_array[$i]->{ $1 . "_seconds_input_rate" }->[0]
                    ->{in_bits_per_sec} = $2;
                $result_array[$i]->{ $1 . "_seconds_input_rate" }->[0]
                    ->{in_bytes_per_sec} = $3;
                $result_array[$i]->{ $1 . "_seconds_input_rate" }->[0]
                    ->{in_packets_per_sec} = $4;
            } elsif ( $line
                =~ /(\d+)\sseconds\soutput\srate\s(\d+)\sbits\/sec\,\s(\d+)\sbytes\/sec\,\s(\d+)\spackets\/sec/
                )
            {
                $result_array[$i]->{ $1 . "_seconds_output_rate" }->[0]
                    ->{in_bits_per_sec} = $2;
                $result_array[$i]->{ $1 . "_seconds_output_rate" }->[0]
                    ->{in_bytes_per_sec} = $3;
                $result_array[$i]->{ $1 . "_seconds_output_rate" }->[0]
                    ->{in_packets_per_sec} = $4;
            } elsif ( $line
                =~ /(\d+)\sseconds\sinput\srate\s(\d+)\sbits\/sec\,\s(\d+)\spackets\/sec/
                )
            {
                $result_array[$i]->{ $1 . "_seconds_input_rate" }->[0]
                    ->{in_bits_per_sec} = $2;
                $result_array[$i]->{ $1 . "_seconds_input_rate" }->[0]
                    ->{in_bytes_per_sec} = $3;
            } elsif ( $line
                =~ /(\d+)\sseconds\soutput\srate\s(\d+)\sbits\/sec\,\s(\d+)\spackets\/sec/
                )
            {
                $result_array[$i]->{ $1 . "_seconds_output_rate" }->[0]
                    ->{in_bits_per_sec} = $2;
                $result_array[$i]->{ $1 . "_seconds_output_rate" }->[0]
                    ->{in_bytes_per_sec} = $3;
            } elsif ( $line =~ /Load\-Interval\s\#\d+\:\s+(.+)/ ) {
                $result_array[$i]->{load_interval} = $1;
            } elsif ( $line
                =~ /input\s+rate\s+(.+)\s+(\w+)\,\s+(.+)\s+(\w+)\;\s+output\s+rate\s+(.+)\s+(\w+)\,\s+(.+)\s+(\w+)/
                )
            {
                $result_array[$i]->{input_rate}->[0]->{ "in_" . $2 }  = $1;
                $result_array[$i]->{input_rate}->[0]->{ "in_" . $4 }  = $3;
                $result_array[$i]->{output_rate}->[0]->{ "in_" . $6 } = $5;
                $result_array[$i]->{output_rate}->[0]->{ "in_" . $8 } = $7;
            } elsif ( $line =~ /^RX$/ ) {
                $direction = "rx";
            } elsif ( $line =~ /^TX$/ ) {
                $direction = "tx";
            } elsif ( $line =~ /(\d+)\s+interface\s+resets/ ) {
                $result_array[$i]->{interface_resets} = $1;
            } elsif ( $line =~ /Members\s+in\s+this\s+channel\s*\:\s*(.*)/ ) {
                my @interfaces = split( /\s*\,\s*/, $1 );
                foreach (@interfaces) {
                    if ( $_ =~ /((\w+)(\d+(\/\d+)?))/ ) {
                        my $temp = {
                            interface      => $1,
                            interface_type => $2,
                            interface_num  => $3,
                        };
                        push(
                            @{  $result_array[$i]->{members_in_this_channel}
                                },
                            $temp
                        );
                    }
                }

            } else {
                if ( my @counter = $line =~ /(\d+\s+[a-zA-Z\s]+)/g ) {
                    foreach (@counter) {
                        $_ =~ s/\s*$//;
                        $_ =~ s/^\s*//;
                        if ( $_ =~ /(\d+)\s+(.+)/ ) {
                            $result_array[$i]->{$direction}->[0]
                                ->{ to_key_name( item => $2 ) } = $1;
                        }
                    }
                }
            }
        }
    }

    return \@result_array;
} ## end sub show_interface

## @name ::show_running_config
##
## @summary Parses the raw output from the "show running-config" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from show_running_config with any of its options as described in the
##    corresponding cdef file 'show.cdefs'
## @synopsis
##      show_running_config(%args);
## INPUTS TO METHOD CALL 'show_running_config' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> </tr>
## <tr> <td> No switch for this key </td> <td> 'entity' </td> </tr>
## <tr> <td> No switch for this key </td> <td> 'name' </td> </tr>
## <tr> <td> No switch for this key </td> <td> 'type' </td> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples
##&nbsp;Command = show running-config interface Ethernet1/4
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;!Command: show running-config interface Ethernet1/4
##&nbsp;!Time: Mon Jul 25 10:37:49 2011
##&nbsp;
##&nbsp;version 5.0(2)N1(1)
##&nbsp;
##&nbsp;interface Ethernet1/4
##&nbsp;  spanning-tree port type edge
##&nbsp;  spanning-tree bpduguard enable
##&nbsp;
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;          {
##&nbsp;            'interface_num' => 'Ethernet',
##&nbsp;            'interface' => 'Ethernet1/4',
##&nbsp;            'running_conf' => [
##&nbsp;                                   'spanning-tree port type edge',
##&nbsp;                                   'spanning-tree bpduguard enable'
##&nbsp;                              ],
##&nbsp;            'interface_type' => '1/4'
##&nbsp;          }
##&nbsp;        ];
##&nbsp;
sub show_running_config {
    my %args = @_;

    my ( $error_message, @result_array );
    my $resp        = $args{response_object};
    my $output      = $resp->get_processed_output();
    my $command     = $resp->get_command();
    my $error_level = $resp->get_execution_status();
    my ( $error, $key );
    my $i = -1;

    # Status returned by switch for this command is
    # is ZERO even if command failed
    $error = _check_error(
        raw_output    => $output,
        command       => "show running-config",
        error_message => \$error_message
    );
    if ($error) {
        $error = 'NACL::APISet::Exceptions::' . $error . 'Exception';
        $error->throw( text => $error_message, response_object => $resp );
    }
    my $option = $command;
    $option =~ s/show running-config//;

    my @lines = get_lines( raw_output => $output );

    if ( $option =~ /interface/ ) {
        my $line = shift(@lines);
        $line = shift(@lines) while ( $line !~ /^interface\s+/ );
        my %hash;
        while (@lines) {
            chomp($line);
            my %interface;
            if ( $line =~ /^interface\s+([a-zA-Z\-]+)([\d|\/]+)/i ) {
                $interface{'interface'}      = $1 . $2;
                $interface{'interface_type'} = $1;
                $interface{'interface_num'}  = $2;
            }
            $line = shift(@lines);
            while ( $line !~ /interface\s+/i ) {
                chomp($line);
                push( @{ $interface{'running_conf'} }, $line );
                last if ( !@lines );
                $line = shift(@lines);
            }
            push( @result_array, \%interface );
        }
    } elsif ( $option eq "" ) {
        while ( my $line = shift(@lines) ) {
            if ( $line =~ /^interface\s+([a-zA-Z]+)([\d|\/]+)/i ) {
                my $j = 0;
                $result_array[0]->{interfaces}->[ ++$i ]->{'interface'}
                    = $1 . $2;
                $result_array[0]->{interfaces}->[$i]->{'interface_type'} = $1;
                $result_array[0]->{interfaces}->[$i]->{'interface_num'}  = $2;
                $line = shift(@lines);
                while (($line)
                    && ( $line !~ /^interface\s+([a-zA-Z]+)([\d|\/]+)/ )
                    && ( $line !~ /line/ ) )
                {
                    $result_array[0]->{interfaces}->[$i]->{running_conf}
                        ->[ $j++ ] = $line;
                    $line = shift(@lines);
                }
                unshift( @lines, $line );
            } elsif ( $line
                =~ /banner\s+motd\s+\#(NetApp Cluster Switch configuration version)\s+(.*)\s+-\s+(Data\s+ONTAP\s+version)\s+(.*)/
                )
            {
                $result_array[0]->{ to_key_name( item => "$1" ) } = $2;
                $result_array[0]->{ to_key_name( item => "$3" ) } = $4;
            }    #to handle the change in banner for net-5010-3 and net-5020-2
            elsif ( $line
                =~ /banner\s+motd\s*#?\s*\S+\s+\d+(?:\s+\(\d+\s\w+\))?\s+NetApp Reference Configuration File\s+\S+\s+version\s+(\S+)\s+.*/
                )
            {
                $result_array[0]->{'rcf_version'} = $1;
            }
            elsif ($line =~ /\*\s+\S+\s+NetApp Reference Configuration File\s+\S+\s+version\s+(\d+.\d+)-.*/ ){
                $result_array[0]->{'rcf_version'} = $1;
            }    #to handle the change in banner for Velvet (3132) series switches
        }
    }

    return \@result_array;
}

## @name ::show_vlan
##
## @summary Parses the raw output from the "show vlan" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from show_vlan with any of its options as described in the
##    corresponding cdef file 'show.cdefs'
## @synopsis
##      show_vlan(%args);
## INPUTS TO METHOD CALL 'show_vlan' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> </tr>
## <tr> <td> 'id' </td> <td> 'vlan-ids' </td> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

##&nbsp;Command = show vlan id 1-2
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;VLAN Name                             Status    Ports
##&nbsp;---- -------------------------------- --------- -------------------------------
##&nbsp;1    default                          active    Po1, Eth1/1, Eth1/2, Eth1/3
##&nbsp;                                                Eth1/4, Eth1/5, Eth1/6, Eth1/7
##&nbsp;                                                Eth1/8, Eth1/9, Eth1/10, Eth1/11
##&nbsp;                                                Eth1/12, Eth1/13, Eth1/14
##&nbsp;                                                Eth1/15, Eth1/16, Eth1/17
##&nbsp;                                                Eth1/18, Eth1/19, Eth1/20
##&nbsp;                                                Eth1/21, Eth1/22, Eth1/23
##&nbsp;                                                Eth1/24, Eth1/25, Eth1/26
##&nbsp;                                                Eth1/27, Eth1/28, Eth1/29
##&nbsp;                                                Eth1/30, Eth1/31, Eth1/32
##&nbsp;                                                Eth1/33, Eth1/34, Eth1/35
##&nbsp;                                                Eth1/36, Eth1/37, Eth1/38
##&nbsp;                                                Eth2/1, Eth2/2, Eth2/3, Eth2/4
##&nbsp;                                                Eth2/5, Eth2/6, Eth3/1, Eth3/2
##&nbsp;                                                Eth3/3, Eth3/4, Eth3/5, Eth3/6
##&nbsp;2    VLAN0002                         active    Po1
##&nbsp;VLAN Name                             Status    Ports
##&nbsp;---- -------------------------------- --------- -------------------------------
##&nbsp;
##&nbsp;Remote SPAN VLANs
##&nbsp;-------------------------------------------------------------------------------
##&nbsp;
##&nbsp;Primary  Secondary  Type             Ports
##&nbsp;-------  ---------  ---------------  -------------------------------------------
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;          undef,
##&nbsp;          {
##&nbsp;            'ports' => [
##&nbsp;                         'Po1',
##&nbsp;                         'Eth1/1',
##&nbsp;                         'Eth1/2',
##&nbsp;                         'Eth1/3',
##&nbsp;                         'Eth1/4',
##&nbsp;                         'Eth1/5',
##&nbsp;                         'Eth1/6',
##&nbsp;                         'Eth1/7',
##&nbsp;                         'Eth1/8',
##&nbsp;                         'Eth1/9',
##&nbsp;                         'Eth1/10',
##&nbsp;                         'Eth1/11',
##&nbsp;                         'Eth1/12',
##&nbsp;                         'Eth1/13',
##&nbsp;                         'Eth1/14',
##&nbsp;                         'Eth1/15',
##&nbsp;                         'Eth1/16',
##&nbsp;                         'Eth1/17',
##&nbsp;                         'Eth1/18',
##&nbsp;                         'Eth1/19',
##&nbsp;                         'Eth1/20',
##&nbsp;                         'Eth1/21',
##&nbsp;                         'Eth1/22',
##&nbsp;                         'Eth1/23',
##&nbsp;                         'Eth1/24',
##&nbsp;                         'Eth1/25',
##&nbsp;                         'Eth1/26',
##&nbsp;                         'Eth1/27',
##&nbsp;                         'Eth1/28',
##&nbsp;                         'Eth1/29',
##&nbsp;                         'Eth1/30',
##&nbsp;                         'Eth1/31',
##&nbsp;                         'Eth1/32',
##&nbsp;                         'Eth1/33',
##&nbsp;                         'Eth1/34',
##&nbsp;                         'Eth1/35',
##&nbsp;                         'Eth1/36',
##&nbsp;                         'Eth1/37',
##&nbsp;                         'Eth1/38',
##&nbsp;                         'Eth2/1',
##&nbsp;                         'Eth2/2',
##&nbsp;                         'Eth2/3',
##&nbsp;                         'Eth2/4',
##&nbsp;                         'Eth2/5',
##&nbsp;                         'Eth2/6',
##&nbsp;                         'Eth3/1',
##&nbsp;                         'Eth3/2',
##&nbsp;                         'Eth3/3',
##&nbsp;                         'Eth3/4',
##&nbsp;                         'Eth3/5',
##&nbsp;                         'Eth3/6'
##&nbsp;                       ],
##&nbsp;            'status' => 'active',
##&nbsp;            'name' => 'default',
##&nbsp;            'id' => 1
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'ports' => [
##&nbsp;                         'Po1'
##&nbsp;                       ],
##&nbsp;            'status' => 'active',
##&nbsp;            'name' => 'VLAN0002',
##&nbsp;            'id' => 2
##&nbsp;          }
##&nbsp;        ];
##&nbsp;
sub show_vlan {
    my %args = @_;

    my ( $error_message, @result_array );
    my $resp        = $args{response_object};
    my $output      = $resp->get_processed_output();
    my $command     = $resp->get_command();
    my $error_level = $resp->get_execution_status();
    my ($error);

    # Status returned by switch for this command is
    # is ZERO even if command failed
    $error = _check_error(
        raw_output    => $output,
        command       => "show mac-address-table",
        error_message => \$error_message
    );
    if ($error) {
        $error = 'NACL::APISet::Exceptions::' . $error . 'Exception';
        $error->throw( text => $error_message, response_object => $resp );
    }

    my @lines = get_lines( raw_output => $output );

    @lines = grep ( !/^--/, @lines );

    # Get the headings
    my $line = shift(@lines);
    my @headings = split( /\s+/, $line );

    my (@vlan_info);
    $line = shift(@lines);
    my $i = 0;
    while ( ( $line !~ /^(VLAN\s+|Remote\s+|Primary)/i ) and @lines ) {
        chomp($line);
        @vlan_info = split( /\s+/, $line );
        if ( $line =~ /^\d+/ ) {
            $result_array[$i]->{'id'}     = shift(@vlan_info);
            $result_array[$i]->{'name'}   = shift(@vlan_info);
            $result_array[$i]->{'status'} = shift(@vlan_info);
            $line = join( " ", @vlan_info );
        }
        my @ports = split( /\,\s+/, $line );
        push( @{ $result_array[$i]->{'ports'} }, @ports );
        $line = shift(@lines);
        if ( $line =~ /^\d+/ ) {
            $i++;
        }
    }

    return \@result_array;
}

## @name ::show_mac_address_table
##
## @summary Parses the raw output from the "show mac address-table" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from show_mac_address_table with any of its options as described in the
##    corresponding cdef file 'show.cdefs'
## @synopsis
##      show_mac_address_table(%args);
## INPUTS TO METHOD CALL 'show_mac_address_table' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## <tr> <td> 'address' </td> <td> 'address' </td> <td> '0' </td> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples
##&nbsp;Command = show mac address-table address 0004.23ba.a4c0
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;Legend:
##&nbsp;        * - primary entry, G - Gateway MAC, (R) - Routed MAC, O - Overlay MAC
##&nbsp;        age - seconds since last seen,+ - primary entry using vPC Peer-Link
##&nbsp;   VLAN     MAC Address      Type      age     Secure NTFY    Ports
##&nbsp;---------+-----------------+--------+---------+------+----+------------------
##&nbsp;* 1        0004.23ba.a4c0    dynamic   70         F    F  Po1
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;          {
##&nbsp;            'ntfy' => 'F',
##&nbsp;            'secure' => 'F',
##&nbsp;            'entry_type' => 'primary_entry',
##&nbsp;            'vlans' => '1',
##&nbsp;            'mac_address' => '0004.23ba.a4c0',
##&nbsp;            'ports' => [
##&nbsp;                         {
##&nbsp;                           'interface_type' => 'Po',
##&nbsp;                           'interface_num' => '1',
##&nbsp;                           'port' => 'Po1'
##&nbsp;                         }
##&nbsp;                       ],
##&nbsp;            'seconds_since_last_seen' => '70',
##&nbsp;            'type' => 'dynamic'
##&nbsp;          }
##&nbsp;        ];
##&nbsp;
sub show_mac_address_table {
    my %args = @_;

    my ( $error_message, @result_array );
    my $resp        = $args{response_object};
    my $output      = $resp->get_processed_output();
    my $command     = $resp->get_command();
    my $error_level = $resp->get_execution_status();
    my ( $error, $key );

    # Status returned by switch for this command is
    # is ZERO even if command failed
    $error = _check_error(
        raw_output    => $output,
        command       => "show mac address-table",
        error_message => \$error_message
    );
    if ($error) {
        $error = 'NACL::APISet::Exceptions::' . $error . 'Exception';
        $error->throw( text => $error_message, response_object => $resp );
    }

    if ( $output =~ /VLAN\s+MAC\s+Address\s+Type\s+Age\s+Port/ ) {
        my @lines = get_lines( raw_output => $output );
        my @headings = qw(vlans mac_address type age ports);
        shift @lines if ( $lines[0] =~ /VLAN\s+MAC\s+Address/ );

        #shift line to get rid of ---------+---
        shift @lines;
        foreach my $line (@lines) {
            my %hash;
            if ( $line =~ /Total MAC Addresses/ ) {
                next;
            } else {
                my @values = split( '\s+', $line );
                foreach ( my $i = 0; $i < @headings; $i++ ) {
                    if ( $headings[$i] =~ /ports/ ) {
                        if ( $values[$i] =~ /([a-zA-Z]+)([\d|\/]+)/ ) {
                            my %ports;
                            $ports{port}           = $values[$i];
                            $ports{interface_type} = $1;
                            $ports{interface_num}  = $2;
                            push( @{ $hash{ $headings[$i] } }, \%ports );
                        }
                    } else {
                        $hash{ $headings[$i] } = $values[$i];
                    }
                }
                push( @result_array, \%hash );
            }
        }
    } else {
        my @lines
            = get_lines( raw_output => $output, first_line => 'Legend:' );
        my $line = "blah";
        my %legends;
        my @legends;
        while ( $line !~ /vlan/i and (@lines) ) {
            $line = shift(@lines);
            chomp($line);
            if ( $line =~ /^legend:/i ) {
                next;
            }
            push( @legends, split( /\,/, $line ) );
        }

        foreach my $legend (@legends) {
            my ( $key, $value ) = split( /\s+\-\s+/, $legend );
            $value = to_key_name( 'item' => $value );
            $legends{$key} = $value;
        }

        @lines = grep ( !/^-----/, @lines );

        my @headings
            = qw(entry_type vlans mac_address type seconds_since_last_seen secure ntfy ports);
        my %hash;
        my ( %mac_addresses, @mac_address, @mac_addresses );
        while (@lines) {
            $line = shift(@lines);
            chomp($line);
            @mac_address = split( /\s+/, $line );
            my $entry_type = shift(@mac_address);
            unshift( @mac_address, $legends{$entry_type} );
            my %mac_address;
            for ( my $i = 0; $i <= $#mac_address; $i++ ) {
                if ( $headings[$i] =~ /ports/ ) {
                    if ( $mac_address[$i] =~ /([a-zA-Z]+)([\d|\/]+)/ ) {
                        my %ports;
                        $ports{port}           = $mac_address[$i];
                        $ports{interface_type} = $1;
                        $ports{interface_num}  = $2;
                        push( @{ $mac_address{ $headings[$i] } }, \%ports );
                    }
                } else {
                    $mac_address{ $headings[$i] } = $mac_address[$i];
                }
            }
            push( @result_array, \%mac_address );
        }

        $hash{'mac_addresses'} = \@mac_addresses;
    }

    return \@result_array;
}

## @name ::show_port_channel
##
## @summary Parses the raw output from the "show port-channel" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from show_port_channel with any of its options as described in the
##    corresponding cdef file 'show.cdefs'
## @synopsis
##      show_port_channel(%args);
## INPUTS TO METHOD CALL 'show_port_channel' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> </tr>
## <tr> <td> 'port-channel' </td> <td> 'port-channel' </td> </tr>
## <tr> <td> No switch for this key </td> <td> 'entity' </td> </tr>
## <tr> <td> No switch for this key </td> <td> 'field' </td> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples
##&nbsp;Command = show port-channel database interface port-channel 1
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;port-channel1
##&nbsp;    Last membership update is successful
##&nbsp;    2 ports in total, 2 ports up
##&nbsp;    First operational port is Ethernet1/39
##&nbsp;    Age of the port-channel is 86d:12h:32m:07s
##&nbsp;    Time since last bundle is 86d:12h:33m:56s
##&nbsp;    Last bundled member is
##&nbsp;    Ports:   Ethernet1/39    [active ] [up] *
##&nbsp;             Ethernet1/40    [active ] [up]
##&nbsp;
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;          {
##&nbsp;            'time_since_last_bundle' => '86d:12h:33m:56s',
##&nbsp;            'ports_in_total' => '2',
##&nbsp;            'ports_up' => '2',
##&nbsp;            'last_membership_update' => 'successful',
##&nbsp;            'last_bundled_member' => '',
##&nbsp;            'ports' => [
##&nbsp;                         {
##&nbsp;                           'pc_status' => 'active ',
##&nbsp;                           'port_status' => 'up',
##&nbsp;                           'port' => 'Ethernet1/39'
##&nbsp;                         },
##&nbsp;                         {
##&nbsp;                           'pc_status' => 'active ',
##&nbsp;                           'port_status' => 'up',
##&nbsp;                           'port' => 'Ethernet1/40'
##&nbsp;                         }
##&nbsp;                       ],
##&nbsp;            'age_of_the_port-channel' => '86d:12h:32m:07s',
##&nbsp;            'first_operational_port' => 'Ethernet1/39'
##&nbsp;          }
##&nbsp;        ];
##&nbsp;
sub show_port_channel {
    my %args = @_;

    my ( $error_message, @result_array );
    my $resp        = $args{response_object};
    my $output      = $resp->get_processed_output();
    my $command     = $resp->get_command();
    my $error_level = $resp->get_execution_status();
    my ( $error, $key );

    # Status returned by switch for this command is
    # is ZERO even if command failed
    $error = _check_error(
        raw_output    => $output,
        command       => "show port-channel",
        error_message => \$error_message
    );
    if ($error) {
        $error = 'NACL::APISet::Exceptions::' . $error . 'Exception';
        $error->throw( text => $error_message, response_object => $resp );
    }

    my @lines = get_lines( raw_output => $output );
    my $line = shift(@lines);
    $line =~ s/Ports:\s+//;

    my @tmp;
    my %hash;
    while (1) {
        chomp($line);
        if ( $line
            =~ /(\d+)\s+ports?\s+in\s+total\s*\,\s+(\d+)\s+ports?\s+up/i )
        {
            $hash{'ports_in_total'} = $1;
            $hash{'ports_up'}       = $2;
        } elsif ( $line =~ /([a-zA-Z]+[\d|\/]+)\s+\[(.+)\s+\]\s+\[(.*)\]/ ) {
            my %tmp;
            $tmp{'port'}        = $1;
            $tmp{'pc_status'}   = $2;
            $tmp{'port_status'} = $3;
            push( @tmp, \%tmp );
        } elsif ( $line =~ /(.+)\s+is\s*(.*)/ ) {
            my $key = to_key_name( item => $1 );
            $hash{$key} = $2;
        }
        last if ( !@lines );
        $line = shift(@lines);
    }
    $hash{'ports'} = \@tmp;
    @result_array = \%hash;

    return \@result_array;

}

## @name ::show_snmp_community
##
## @summary Parses the raw output from the "show snmp community" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from show_snmp_community with any of its options as described in the
##    corresponding cdef file 'show.cdefs'
## @synopsis
##      show_snmp_community(%args);
## INPUTS TO METHOD CALL 'show_snmp_community' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

##&nbsp; command = show snmp community
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;show snmp community
##&nbsp; Community            Group / Access      context    acl_filter
##&nbsp; ---------            --------------      -------    ----------
##&nbsp; netapp                network-operator
##&nbsp; public                network-operator
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp; $VAR1 = [
##&nbsp;           {
##&nbsp;             'community' => 'netapp',
##&nbsp;             'context' => '',
##&nbsp;             'group_access' => 'network-operator',
##&nbsp;             'acl_filter' => ''
##&nbsp;           },
##&nbsp;           {
##&nbsp;             'community' => 'public',
##&nbsp;             'context' => '',
##&nbsp;             'group_access' => 'network-operator',
##&nbsp;             'acl_filter' => ''
##&nbsp;           }
##&nbsp;         ];
sub show_snmp_community {
    my %args = @_;

    my ( $error_message, @result_array );
    my $resp        = $args{response_object};
    my $output      = $resp->get_processed_output();
    my $command     = $resp->get_command();
    my $error_level = $resp->get_execution_status();
    my ( $error, $key );

    # Status returned by switch for this command is
    # ZERO even if command failed
    $error = _check_error(
        raw_output    => $output,
        command       => $command,
        error_message => \$error_message
    );
    if ($error) {
        $error = 'NACL::APISet::Exceptions::' . $error . 'Exception';
        $error->throw( text => $error_message, response_object => $resp );
    }

    my @lines = get_lines( raw_output => $output );
    my %hash;

    my @headings = qw(community group_access context acl_filter);
    my $j        = 0;
    for ( my $k = 2; $k <= $#lines; $k++ ) {
        my @values = split( /\s+/, $lines[$k] );
        for ( my $i = 0; $i < scalar(@headings); $i++ ) {
            if ( $values[$i] ) {
                $result_array[$j]->{ $headings[$i] } = $values[$i];
            } else {
                $result_array[$j]->{ $headings[$i] } = '';
            }
        }
        $j++;
    }
    return \@result_array;
}

## @name ::show_version
##
## @summary Parses the raw output from the "show version" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from show_version with any of its options as described in the
##    corresponding cdef file 'show.cdefs'
## @synopsis
##      show_version(%args);
## INPUTS TO METHOD CALL 'show_version' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

##&nbsp;Command = show version
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;Cisco Nexus Operating System (NX-OS) Software
##&nbsp;TAC support: http://www.cisco.com/tac
##&nbsp;Copyright (c) 2002-2010, Cisco Systems, Inc. All rights reserved.
##&nbsp;The copyrights to certain works contained herein are owned by
##&nbsp;other third parties and are used and distributed under license.
##&nbsp;Some parts of this software are covered under the GNU Public
##&nbsp;License. A copy of the license is available at
##&nbsp;http://www.gnu.org/licenses/gpl.html.
##&nbsp;
##&nbsp;Software
##&nbsp;  BIOS:      version 1.3.0
##&nbsp;  loader:    version N/A
##&nbsp;  kickstart: version 5.0(2)N1(1)
##&nbsp;  system:    version 5.0(2)N1(1)
##&nbsp;  power-seq: version v1.2
##&nbsp;  BIOS compile time:       09/08/09
##&nbsp;  kickstart image file is: bootflash:/n5000-uk9-kickstart.5.0.2.N1.1.bin
##&nbsp;  kickstart compile time:  10/15/2010 0:00:00 [10/15/2010 08:00:43]
##&nbsp;  system image file is:    bootflash:/n5000-uk9.5.0.2.N1.1.bin
##&nbsp;  system compile time:     10/15/2010 0:00:00 [10/15/2010 09:34:05]
##&nbsp;
##&nbsp;
##&nbsp;Hardware
##&nbsp;  cisco Nexus5010 Chassis ("20x10GE/Supervisor")
##&nbsp;  Intel(R) Celeron(R) M CPU    with 2073416 kB of memory.
##&nbsp;  Processor Board ID JAF1505CCNP
##&nbsp;
##&nbsp;  Device name: net-5010-3
##&nbsp;  bootflash:    1003520 kB
##&nbsp;
##&nbsp;Kernel uptime is 2 day(s), 12 hour(s), 18 minute(s), 4 second(s)
##&nbsp;
##&nbsp;Last reset at 255640 usecs after  Fri Mar 30 22:14:39 2012
##&nbsp;
##&nbsp;  Reason: Reset Requested by CLI command reload
##&nbsp;  System version: 5.0(2)N1(1)
##&nbsp;  Service:
##&nbsp;
##&nbsp;plugin
##&nbsp;  Core Plugin, Ethernet Plugin
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;          {
##&nbsp;            'plugin' => [
##&nbsp;                          'Core Plugin',
##&nbsp;                          'Ethernet Plugin'
##&nbsp;                        ],
##&nbsp;            'cisco_software' => 'Nexus',
##&nbsp;            'kernel_uptime' => '2 day(s), 12 hour(s), 18 minute(s), 4 second(s)',
##&nbsp;            'software' => [
##&nbsp;                            {
##&nbsp;                              'system_compile_time' => '10/15/2010 0:00:00 [10/15/2010 09:34:05]',
##&nbsp;                              'system_image_file_is' => 'bootflash:/n5000-uk9.5.0.2.N1.1.bin',
##&nbsp;                              'system' => 'version 5.0(2)N1(1)',
##&nbsp;                              'kickstart' => 'version 5.0(2)N1(1)',
##&nbsp;                              'bios' => 'version 1.3.0',
##&nbsp;                              'bios_compile_time' => '09/08/09',
##&nbsp;                              'power_seq' => 'version v1.2',
##&nbsp;                              'kickstart_image_file_is' => 'bootflash:/n5000-uk9-kickstart.5.0.2.N1.1.bin',
##&nbsp;                              'kickstart_compile_time' => '10/15/2010 0:00:00 [10/15/2010 08:00:43]',
##&nbsp;                              'loader' => 'version N/A'
##&nbsp;                            }
##&nbsp;                          ],
##&nbsp;            'gnu_license_info' => 'http://www.gnu.org/licenses/gpl.html.',
##&nbsp;            'hardware' => [
##&nbsp;                            {
##&nbsp;                              'bootflash' => '1003520 kB',
##&nbsp;                              'device_name' => 'net-5010-3',
##&nbsp;                              'cpu' => 'Intel(R) Celeron(R) M',
##&nbsp;                              'cpu_memory' => '2073416 kB',
##&nbsp;                              'processor_board_id' => 'JAF1505CCNP',
##&nbsp;                              'chassis' => 'cisco Nexus5010 Chassis ("20x10GE/Supervisor")'
##&nbsp;                            }
##&nbsp;                          ],
##&nbsp;            'last_reset' => [
##&nbsp;                              {
##&nbsp;                                'system_version' => '5.0(2)N1(1)',
##&nbsp;                                'reason' => 'Reset Requested by CLI command reload',
##&nbsp;                                'service' => ''
##&nbsp;                              }
##&nbsp;                            ]
##&nbsp;          }
##&nbsp;        ];
##&nbsp;

sub show_version {
    my %args = @_;

    my ( $error_message, @result_array );
    my $resp        = $args{response_object};
    my $output      = $resp->get_processed_output();
    my $command     = $resp->get_command();
    my $error_level = $resp->get_execution_status();
    my ( $error, $plugins, $i );
    my $key = undef;
    my $j   = 0;

    $error = _check_error(
        raw_output    => $output,
        command       => $command,
        error_message => \$error_message
    );
    if ($error) {
        $error = 'NACL::APISet::Exceptions::' . $error . 'Exception';
        $error->throw( text => $error_message, response_object => $resp );
    }
    my @lines = get_lines( raw_output => $output );
    while ( my $line = shift(@lines) ) {
        if ( $line =~ /^(Software|Hardware|Last\s+reset)/ ) {
            $key = to_key_name( item => "$1" );
            $i = 0;
        } elsif ( $line =~ /(^http:\/\/.*)/ ) {
            $result_array[0]->{gnu_license_info} = $1;
        } elsif (
            $line =~ /^Cisco\s+(.*)\s+Operating\s+System\s+.*\s+Software/ )
        {
            $result_array[0]->{cisco_software} = "$1";
        } elsif ( $line =~ /(.*):\s+(.*)/ && defined($key) ) {
            my $key1 = $1;
            my $val1 = $2;
            $key1 =~ s/-/_/g;
            $result_array[0]->{$key}->[$i]->{ to_key_name( item => "$key1" ) }
                = $val1;
        } elsif ( $line =~ /(Processor\s+Board\s+ID)\s+(.*)/ ) {
            $result_array[0]->{$key}->[$i]->{ to_key_name( item => "$1" ) }
                = $2;
        } elsif ( $line =~ /(.*)\s+CPU\s+with\s+(.*)\s+of memory/ ) {
            $result_array[0]->{$key}->[$i]->{cpu_memory} = $2;
            $result_array[0]->{$key}->[$i]->{cpu}        = $1;
        } elsif ( $line =~ /(.*\s+Chassis\s*.*)/i ) {
            $result_array[0]->{$key}->[$i]->{chassis} = $1;
        } elsif ( $line =~ /(.*):\s*(.*)/ && defined($key) ) {
            my $key1 = $1;
            my $val1 = $2;
            $key1 =~ s/-/_/g;
            $result_array[0]->{$key}->[$i]->{ to_key_name( item => "$key1" ) }
                = $val1;
        } elsif ( $line =~ /(Kernel\s+uptime)\s+is\s+(.*)/ ) {
            $result_array[0]->{ to_key_name( item => "$1" ) } = $2;
        } elsif ( $line =~ /plugin/ ) {
            $key = "plugin";
            while ( $line = shift(@lines) ) {
                $plugins = "$plugins" . "$line";
            }
            my @pluggins = split( /,\s*/, $plugins );
            $result_array[0]->{plugin} = \@pluggins;
        }
    }
    return \@result_array;
}

## @name ::show_cdp_global
##
## @summary Parses the raw output from the "show cdp global" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from show_cdp_global with any of its options as described in the
##    corresponding cdef file 'show.cdefs'
## @synopsis
##      show_cdp_global(%args);
## INPUTS TO METHOD CALL 'show_cdp_global' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

##&nbsp;Command = show cdp global
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;Global CDP information:
##&nbsp;    CDP enabled globally
##&nbsp;    Refresh time is 5 seconds
##&nbsp;    Hold time is 180 seconds
##&nbsp;    CDPv2 advertisements is enabled
##&nbsp;    DeviceID TLV in System-Name(Default) Format
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;          {
##&nbsp;            'deviceid_tlv' => 'Default Format',
##&nbsp;            'cdpv2_advertisements' => 'enabled',
##&nbsp;            'refresh_time' => '5 seconds',
##&nbsp;            'cdp_global_status' => 'enabled',
##&nbsp;            'hold_time' => '180 seconds'
##&nbsp;          }
##&nbsp;        ];
##&nbsp;
sub show_cdp_global {
    my %args = @_;

    my ( $error_message, @result_array );
    my $resp    = $args{response_object};
    my $output  = $resp->get_processed_output();
    my $command = $resp->get_command();

    my $error = _check_error(
        raw_output    => $output,
        command       => $command,
        error_message => \$error_message
    );
    if ($error) {
        $error = 'NACL::APISet::Exceptions::' . $error . 'Exception';
        $error->throw( text => $error_message, response_object => $resp );
    }

    my @lines = get_lines( raw_output => $output );
    foreach my $line (@lines) {
        if ( $line =~ /^CDP\s(\w+)\sglobally/ ) {
            $result_array[0]->{'cdp_global_status'} = $1;
        } elsif ( $line =~ /^(Refresh time) is (\d+\s\w+)/ ) {
            $result_array[0]->{ to_key_name( item => $1 ) } = $2;
        } elsif ( $line =~ /^(Hold time) is (\d+\s\w+)/ ) {
            $result_array[0]->{ to_key_name( item => $1 ) } = $2;
        } elsif ( $line =~ /^(CDP\w+ advertisements) is (\w+)/ ) {
            $result_array[0]->{ to_key_name( item => $1 ) } = $2;
        } elsif ( $line =~ /^(DeviceID TLV)\s.*\((\w+)\)\s(\w+)/ ) {
            $result_array[0]->{ to_key_name( item => $1 ) } = $2 . ' ' . $3;
        }
    }
    return \@result_array;
}    ##end sub show_cdp_global

## @name ::show_cdp_all
##
## @summary Parses the raw output from the "show cdp all" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from show_cdp_all with any of its options as described in the
##    corresponding cdef file 'show.cdefs'
## @synopsis
##      show_cdp_all(%args);
## INPUTS TO METHOD CALL 'show_cdp_all' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

##&nbsp;Command = show cdp all
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;mgmt0 is up
##&nbsp;    CDP enabled on interface
##&nbsp;    Refresh time is 5 seconds
##&nbsp;    Hold time is 180 seconds
##&nbsp;Ethernet1/1 is down
##&nbsp;    CDP enabled on interface
##&nbsp;    Refresh time is 5 seconds
##&nbsp;    Hold time is 180 seconds
##&nbsp;Ethernet1/2 is down
##&nbsp;    CDP enabled on interface
##&nbsp;    Refresh time is 5 seconds
##&nbsp;    Hold time is 180 seconds
##&nbsp;Ethernet1/3 is down
##&nbsp;    CDP enabled on interface
##&nbsp;    Refresh time is 5 seconds
##&nbsp;    Hold time is 180 seconds
##&nbsp;Ethernet1/4 is down
##&nbsp;    CDP enabled on interface
##&nbsp;    Refresh time is 5 seconds
##&nbsp;    Hold time is 180 seconds
##&nbsp;Ethernet1/5 is up
##&nbsp;    CDP enabled on interface
##&nbsp;    Refresh time is 5 seconds
##&nbsp;    Hold time is 180 seconds
##&nbsp;Ethernet1/6 is up
##&nbsp;    CDP enabled on interface
##&nbsp;    Refresh time is 5 seconds
##&nbsp;    Hold time is 180 seconds
##&nbsp;Ethernet1/7 is down
##&nbsp;    CDP enabled on interface
##&nbsp;    Refresh time is 5 seconds
##&nbsp;    Hold time is 180 seconds
##&nbsp;Ethernet1/8 is down
##&nbsp;    CDP enabled on interface
##&nbsp;    Refresh time is 5 seconds
##&nbsp;    Hold time is 180 seconds
##&nbsp;Ethernet1/9 is down
##&nbsp;    CDP enabled on interface
##&nbsp;    Refresh time is 5 seconds
##&nbsp;    Hold time is 180 seconds
##&nbsp;Ethernet1/10 is down
##&nbsp;    CDP enabled on interface
##&nbsp;    Refresh time is 5 seconds
##&nbsp;    Hold time is 180 seconds
##&nbsp;Ethernet1/11 is down
##&nbsp;    CDP enabled on interface
##&nbsp;    Refresh time is 5 seconds
##&nbsp;    Hold time is 180 seconds
##&nbsp;Ethernet1/12 is down
##&nbsp;    CDP enabled on interface
##&nbsp;    Refresh time is 5 seconds
##&nbsp;    Hold time is 180 seconds
##&nbsp;Ethernet1/13 is up
##&nbsp;    CDP enabled on interface
##&nbsp;    Refresh time is 5 seconds
##&nbsp;    Hold time is 180 seconds
##&nbsp;Ethernet1/14 is up
##&nbsp;    CDP enabled on interface
##&nbsp;    Refresh time is 5 seconds
##&nbsp;    Hold time is 180 seconds
##&nbsp;Ethernet1/15 is down
##&nbsp;    CDP enabled on interface
##&nbsp;    Refresh time is 5 seconds
##&nbsp;    Hold time is 180 seconds
##&nbsp;Ethernet1/16 is up
##&nbsp;    CDP enabled on interface
##&nbsp;    Refresh time is 5 seconds
##&nbsp;    Hold time is 180 seconds
##&nbsp;Ethernet1/17 is up
##&nbsp;    CDP enabled on interface
##&nbsp;    Refresh time is 5 seconds
##&nbsp;    Hold time is 180 seconds
##&nbsp;Ethernet1/18 is up
##&nbsp;    CDP enabled on interface
##&nbsp;    Refresh time is 5 seconds
##&nbsp;    Hold time is 180 seconds
##&nbsp;Ethernet1/19 is up
##&nbsp;    CDP enabled on interface
##&nbsp;    Refresh time is 5 seconds
##&nbsp;    Hold time is 180 seconds
##&nbsp;Ethernet1/20 is up
##&nbsp;    CDP enabled on interface
##&nbsp;    Refresh time is 5 seconds
##&nbsp;    Hold time is 180 seconds
##&nbsp;Ethernet2/1 is up
##&nbsp;    CDP enabled on interface
##&nbsp;    Refresh time is 5 seconds
##&nbsp;    Hold time is 180 seconds
##&nbsp;Ethernet2/2 is up
##&nbsp;    CDP enabled on interface
##&nbsp;    Refresh time is 5 seconds
##&nbsp;    Hold time is 180 seconds
##&nbsp;Ethernet2/3 is down
##&nbsp;    CDP enabled on interface
##&nbsp;    Refresh time is 5 seconds
##&nbsp;    Hold time is 180 seconds
##&nbsp;Ethernet2/4 is down
##&nbsp;    CDP enabled on interface
##&nbsp;    Refresh time is 5 seconds
##&nbsp;    Hold time is 180 seconds
##&nbsp;Ethernet2/5 is down
##&nbsp;    CDP enabled on interface
##&nbsp;    Refresh time is 5 seconds
##&nbsp;    Hold time is 180 seconds
##&nbsp;Ethernet2/6 is down
##&nbsp;    CDP enabled on interface
##&nbsp;    Refresh time is 5 seconds
##&nbsp;    Hold time is 180 seconds
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;          {
##&nbsp;            'cdp_status' => 'enabled',
##&nbsp;            'refresh_time' => '5 seconds',
##&nbsp;            'status' => 'up',
##&nbsp;            'interface' => 'mgmt0',
##&nbsp;            'hold_time' => '180 seconds'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'cdp_status' => 'enabled',
##&nbsp;            'refresh_time' => '5 seconds',
##&nbsp;            'status' => 'down',
##&nbsp;            'interface' => 'Ethernet1/1',
##&nbsp;            'hold_time' => '180 seconds'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'cdp_status' => 'enabled',
##&nbsp;            'refresh_time' => '5 seconds',
##&nbsp;            'status' => 'down',
##&nbsp;            'interface' => 'Ethernet1/2',
##&nbsp;            'hold_time' => '180 seconds'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'cdp_status' => 'enabled',
##&nbsp;            'refresh_time' => '5 seconds',
##&nbsp;            'status' => 'down',
##&nbsp;            'interface' => 'Ethernet1/3',
##&nbsp;            'hold_time' => '180 seconds'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'cdp_status' => 'enabled',
##&nbsp;            'refresh_time' => '5 seconds',
##&nbsp;            'status' => 'down',
##&nbsp;            'interface' => 'Ethernet1/4',
##&nbsp;            'hold_time' => '180 seconds'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'cdp_status' => 'enabled',
##&nbsp;            'refresh_time' => '5 seconds',
##&nbsp;            'status' => 'up',
##&nbsp;            'interface' => 'Ethernet1/5',
##&nbsp;            'hold_time' => '180 seconds'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'cdp_status' => 'enabled',
##&nbsp;            'refresh_time' => '5 seconds',
##&nbsp;            'status' => 'up',
##&nbsp;            'interface' => 'Ethernet1/6',
##&nbsp;            'hold_time' => '180 seconds'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'cdp_status' => 'enabled',
##&nbsp;            'refresh_time' => '5 seconds',
##&nbsp;            'status' => 'down',
##&nbsp;            'interface' => 'Ethernet1/7',
##&nbsp;            'hold_time' => '180 seconds'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'cdp_status' => 'enabled',
##&nbsp;            'refresh_time' => '5 seconds',
##&nbsp;            'status' => 'down',
##&nbsp;            'interface' => 'Ethernet1/8',
##&nbsp;            'hold_time' => '180 seconds'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'cdp_status' => 'enabled',
##&nbsp;            'refresh_time' => '5 seconds',
##&nbsp;            'status' => 'down',
##&nbsp;            'interface' => 'Ethernet1/9',
##&nbsp;            'hold_time' => '180 seconds'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'cdp_status' => 'enabled',
##&nbsp;            'refresh_time' => '5 seconds',
##&nbsp;            'status' => 'down',
##&nbsp;            'interface' => 'Ethernet1/10',
##&nbsp;            'hold_time' => '180 seconds'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'cdp_status' => 'enabled',
##&nbsp;            'refresh_time' => '5 seconds',
##&nbsp;            'status' => 'down',
##&nbsp;            'interface' => 'Ethernet1/11',
##&nbsp;            'hold_time' => '180 seconds'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'cdp_status' => 'enabled',
##&nbsp;            'refresh_time' => '5 seconds',
##&nbsp;            'status' => 'down',
##&nbsp;            'interface' => 'Ethernet1/12',
##&nbsp;            'hold_time' => '180 seconds'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'cdp_status' => 'enabled',
##&nbsp;            'refresh_time' => '5 seconds',
##&nbsp;            'status' => 'up',
##&nbsp;            'interface' => 'Ethernet1/13',
##&nbsp;            'hold_time' => '180 seconds'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'cdp_status' => 'enabled',
##&nbsp;            'refresh_time' => '5 seconds',
##&nbsp;            'status' => 'up',
##&nbsp;            'interface' => 'Ethernet1/14',
##&nbsp;            'hold_time' => '180 seconds'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'cdp_status' => 'enabled',
##&nbsp;            'refresh_time' => '5 seconds',
##&nbsp;            'status' => 'down',
##&nbsp;            'interface' => 'Ethernet1/15',
##&nbsp;            'hold_time' => '180 seconds'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'cdp_status' => 'enabled',
##&nbsp;            'refresh_time' => '5 seconds',
##&nbsp;            'status' => 'up',
##&nbsp;            'interface' => 'Ethernet1/16',
##&nbsp;            'hold_time' => '180 seconds'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'cdp_status' => 'enabled',
##&nbsp;            'refresh_time' => '5 seconds',
##&nbsp;            'status' => 'up',
##&nbsp;            'interface' => 'Ethernet1/17',
##&nbsp;            'hold_time' => '180 seconds'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'cdp_status' => 'enabled',
##&nbsp;            'refresh_time' => '5 seconds',
##&nbsp;            'status' => 'up',
##&nbsp;            'interface' => 'Ethernet1/18',
##&nbsp;            'hold_time' => '180 seconds'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'cdp_status' => 'enabled',
##&nbsp;            'refresh_time' => '5 seconds',
##&nbsp;            'status' => 'up',
##&nbsp;            'interface' => 'Ethernet1/19',
##&nbsp;            'hold_time' => '180 seconds'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'cdp_status' => 'enabled',
##&nbsp;            'refresh_time' => '5 seconds',
##&nbsp;            'status' => 'up',
##&nbsp;            'interface' => 'Ethernet1/20',
##&nbsp;            'hold_time' => '180 seconds'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'cdp_status' => 'enabled',
##&nbsp;            'refresh_time' => '5 seconds',
##&nbsp;            'status' => 'up',
##&nbsp;            'interface' => 'Ethernet2/1',
##&nbsp;            'hold_time' => '180 seconds'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'cdp_status' => 'enabled',
##&nbsp;            'refresh_time' => '5 seconds',
##&nbsp;            'status' => 'up',
##&nbsp;            'interface' => 'Ethernet2/2',
##&nbsp;            'hold_time' => '180 seconds'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'cdp_status' => 'enabled',
##&nbsp;            'refresh_time' => '5 seconds',
##&nbsp;            'status' => 'down',
##&nbsp;            'interface' => 'Ethernet2/3',
##&nbsp;            'hold_time' => '180 seconds'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'cdp_status' => 'enabled',
##&nbsp;            'refresh_time' => '5 seconds',
##&nbsp;            'status' => 'down',
##&nbsp;            'interface' => 'Ethernet2/4',
##&nbsp;            'hold_time' => '180 seconds'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'cdp_status' => 'enabled',
##&nbsp;            'refresh_time' => '5 seconds',
##&nbsp;            'status' => 'down',
##&nbsp;            'interface' => 'Ethernet2/5',
##&nbsp;            'hold_time' => '180 seconds'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'cdp_status' => 'enabled',
##&nbsp;            'refresh_time' => '5 seconds',
##&nbsp;            'status' => 'down',
##&nbsp;            'interface' => 'Ethernet2/6',
##&nbsp;            'hold_time' => '180 seconds'
##&nbsp;          }
##&nbsp;        ];
##&nbsp;
sub show_cdp_all {
    my %args = @_;

    my ( $error_message, @result_array );
    my $resp    = $args{response_object};
    my $output  = $resp->get_processed_output();
    my $command = $resp->get_command();

    my $error = _check_error(
        raw_output    => $output,
        command       => $command,
        error_message => \$error_message
    );
    if ($error) {
        $error = 'NACL::APISet::Exceptions::' . $error . 'Exception';
        $error->throw( text => $error_message, response_object => $resp );
    }
    my @lines = get_lines( raw_output => $output );
    my $i = 0;
    my $line;
    while (@lines) {
        $line = shift @lines;
        if ( $line =~ /([a-zA-Z]+[\d|\/]+)\s+is\s+(\w+)/ ) {
            $result_array[$i]->{'interface'} = $1;
            $result_array[$i]->{'status'}    = $2;
        }
        $line = shift @lines;
        if ( $line =~ /CDP\s+.*(enabled|disabled)\s+.*/ ) {
            $result_array[$i]->{'cdp_status'} = $1;
        }
        $line = shift @lines;
        if ( $line =~ /^(Refresh time) is (\d+\s\w+)/ ) {
            $result_array[$i]->{ to_key_name( item => $1 ) } = $2;
        }
        $line = shift @lines;
        if ( $line =~ /^(Hold time) is (\d+\s\w+)/ ) {
            $result_array[$i]->{ to_key_name( item => $1 ) } = $2;
        }
        $i++;
    }
    return \@result_array;
}    ##end sub show_cdp_all

## @name ::show_cdp_neighbors
##
## @summary Parses the raw output from the "show cdp neighbors" command
##
## @description 
##    This method can be used for parsing the raw output obtained after calling 
##    the method from show_cdp_neighbors with any of its options as described in the 
##    corresponding cdef file 'show.cdefs'
## @synopsis
##      show_cdp_neighbors(%args);
## INPUTS TO METHOD CALL 'show_cdp_neighbors' ON APISET OBJECT 
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## <tr> <td> 'detail' </td> <td> 'detail' </td> <td> '1' </td> </tr> 
## <tr> <td> 'interface' </td> <td> 'interface' </td> <td> '0' </td> </tr> 
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples
##&nbsp;Command = show cdp neighbors
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;Capability Codes: R - Router, T - Trans-Bridge, B - Source-Route-Bridge
##&nbsp;                  S - Switch, H - Host, I - IGMP, r - Repeater,
##&nbsp;                  V - VoIP-Phone, D - Remotely-Managed-Device,
##&nbsp;                  s - Supports-STP-Dispute
##&nbsp;
##&nbsp;
##&nbsp;Device-ID             Local Intrfce Hldtme Capability  Platform      Port ID
##&nbsp;g10-edge-sw1.lab.netapp.com
##&nbsp;                       mgmt0         170    S I       WS-C4948-10GE Gig1/17      
##&nbsp;net-delp-5a            Eth1/19       131    H         FAS3250       e2b          
##&nbsp;net-delp-5a            Eth1/20       131    H         FAS3250       e1b          
##&nbsp;NIC-Central-Switch-net-5010-5(SSI15100CBM)
##&nbsp;                       Eth2/5        172    S I s     N5K-C5010P-BF Eth101/1/1   
##&nbsp;NIC-Central-Switch-net-5010-5(SSI15100CBM)
##&nbsp;                       Eth2/6        172    S I s     N5K-C5010P-BF Eth101/1/2   
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;          {
##&nbsp;            'device_id' => 'g10-edge-sw1.lab.netapp.com',
##&nbsp;            'holdtime' => '170',
##&nbsp;            'port_id' => 'Gig1/17',
##&nbsp;            'interface' => 'mgmt0',
##&nbsp;            'platform' => 'WS-C4948-10GE',
##&nbsp;            'capability' => [
##&nbsp;                              'Switch',
##&nbsp;                              'IGMP'
##&nbsp;                            ]
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'device_id' => 'net-delp-5a',
##&nbsp;            'holdtime' => '131',
##&nbsp;            'port_id' => 'e2b',
##&nbsp;            'interface' => 'Eth1/19',
##&nbsp;            'platform' => 'FAS3250',
##&nbsp;            'capability' => [
##&nbsp;                              'Host'
##&nbsp;                            ]
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'device_id' => 'net-delp-5a',
##&nbsp;            'holdtime' => '131',
##&nbsp;            'port_id' => 'e1b',
##&nbsp;            'interface' => 'Eth1/20',
##&nbsp;            'platform' => 'FAS3250',
##&nbsp;            'capability' => [
##&nbsp;                              'Host'
##&nbsp;                            ]
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'device_id' => 'NIC-Central-Switch-net-5010-5(SSI15100CBM)',
##&nbsp;            'holdtime' => '172',
##&nbsp;            'port_id' => 'Eth101/1/1',
##&nbsp;            'interface' => 'Eth2/5',
##&nbsp;            'platform' => 'N5K-C5010P-BF',
##&nbsp;            'capability' => [
##&nbsp;                              'Switch',
##&nbsp;                              'IGMP',
##&nbsp;                              'Supports-STP-Dispute'
##&nbsp;                            ]
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'device_id' => 'NIC-Central-Switch-net-5010-5(SSI15100CBM)',
##&nbsp;            'holdtime' => '172',
##&nbsp;            'port_id' => 'Eth101/1/2',
##&nbsp;            'interface' => 'Eth2/6',
##&nbsp;            'platform' => 'N5K-C5010P-BF',
##&nbsp;            'capability' => [
##&nbsp;                              'Switch',
##&nbsp;                              'IGMP',
##&nbsp;                              'Supports-STP-Dispute'
##&nbsp;                            ]
##&nbsp;          }
##&nbsp;        ];
##&nbsp;
sub show_cdp_neighbors {
    my %args = @_;
    my ( $error_message, @result_array );
    my $resp    = $args{response_object};
    my $output  = $resp->get_processed_output();
    my $command = $resp->get_command();
    my $error   = _check_error(
        raw_output    => $output,
        error_message => \$error_message
    );
    if ($error) {
        $error = 'NACL::APISet::Exceptions::' . $error . 'Exception';
        $error->throw( text => $error_message, response_object => $resp );
    }
    my @lines = get_lines( raw_output => $output, response_object => $resp );
    my ( @values, @codes, %capab_codes, @temp );
    my $count = 0;
    my $pos   = -1;
    my $j     = 0;
    my @capabs;
    my @head = qw(device_id interface holdtime capability platform port_id );

    for ( my $i = 0; $i <= $#lines; $i++ ) {

        if ( $lines[$i] =~ /Capability Codes: / ) {
            until ( $lines[ $i + $j ] =~ /Device\-ID/ ) {
                @temp = split( /,\s+/, $lines[ $i + $j++ ] );
                push( @codes, @temp );
            }
            foreach my $cod (@codes) {
                @temp = split( /\s+-\s+/, $cod );
                $capab_codes{ $temp[0] } = $temp[1];
            }
            $i = $i + $j;
        }
        if ( $lines[$i] =~ /Device\-ID/ ) {
            $i++;
        }
        if ( $lines[$i] =~ /[\(\.]/ ) {
            $result_array[$count]->{ to_key_name( item => $head[0] ) }
                = $lines[$i];
            $i++;
            if ( $lines[$i]
                =~ /(\S+)\s+(\d+)\s+([\S\s]+)\s\s+(\S+)\s+(\S+\s*\S*)+/ )
            {
                push( @values, $1, $2, $3, $4, $5 );
                @capabs = split( /\s/, $3 );
                for ( my $j = 1; $j <= $#head; $j++ ) {
                    if ( $head[$j] eq "capability" ) {
                        for ( my $k = 0; $k <= $#capabs; $k++ ) {
                            foreach my $key ( keys(%capab_codes) ) {
                                if ( $capabs[$k] eq $key ) {
                                    $capabs[$k] = $capab_codes{$key};
                                }
                            }
                        }
                        push(
                            @{ $result_array[$count]->{capability} },
                            @capabs
                        );
                        $j++;
                    }
                    $result_array[$count]->{ $head[$j] } = $values[ $j - 1 ]

                }
                $count++;
            }
        }
        if ( $lines[$i]
            =~ /(\S+)\s+(\S+)\s+(\d+)\s+([\S\s]+)\s\s+(\S+)\s+(\S+\s*\S*)+/ )
        {
            push( @values, $1, $2, $3, $4, $5, $6 );
            @capabs = split( /\s/, $4 );
            for ( my $j = 0; $j <= $#head; $j++ ) {
                if ( $head[$j] eq "capability" ) {
                    for ( my $k = 0; $k <= $#capabs; $k++ ) {
                        foreach my $key ( keys(%capab_codes) ) {
                            if ( $capabs[$k] eq $key ) {
                                $capabs[$k] = $capab_codes{$key};
                            }
                        }
                    }
                    push( @{ $result_array[$count]->{capability} }, @capabs );
                    $j++;
                }
                $result_array[$count]->{ $head[$j] } = $values[$j];
            }
            $count++;
        }
        $#values = -1;
    }
    return \@result_array;
} ## end sub show_cdp_neighbors

## @name ::show_boot
##
## @summary Parses the raw output from the "show boot" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from show_boot with any of its options as described in the
##    corresponding cdef file 'show.cdefs'
## @synopsis
##      show_boot(%args);
## INPUTS TO METHOD CALL 'show_boot' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## <tr> <td> 'variables' </td> <td> 'variables' </td> <td> '1' </td> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples
##&nbsp;Command = show boot variables
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;List of boot variables are:
##&nbsp;    system
##&nbsp;    kickstart
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;          {
##&nbsp;            'list_of_boot_variables_are' => [
##&nbsp;                                              'system',
##&nbsp;                                              'kickstart'
##&nbsp;                                            ]
##&nbsp;          }
##&nbsp;        ];
##&nbsp;
##&nbsp;Command = show boot
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;Current Boot Variables:
##&nbsp;
##&nbsp;
##&nbsp;kickstart variable = bootflash:/n5000-uk9-kickstart.5.2.1.N1.1.bin
##&nbsp;system variable = bootflash:/n5000-uk9.5.2.1.N1.1.bin
##&nbsp;
##&nbsp;Boot Variables on next reload:
##&nbsp;
##&nbsp;
##&nbsp;kickstart variable = bootflash:/n5000-uk9-kickstart.5.2.1.N1.1.bin
##&nbsp;system variable = bootflash:/n5000-uk9.5.2.1.N1.1.bin
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;          {
##&nbsp;            'boot_variables_on_next_reload' => [
##&nbsp;                                                 {
##&nbsp;                                                   'kickstart_variable' => 'bootflash:/n5000-uk9-kickstart.5.2.1.N1.1.bin',
##&nbsp;                                                   'system_variable' => 'bootflash:/n5000-uk9.5.2.1.N1.1.bin'
##&nbsp;                                                 }
##&nbsp;                                               ],
##&nbsp;            'current_boot_variables' => [
##&nbsp;                                          {
##&nbsp;                                            'kickstart_variable' => 'bootflash:/n5000-uk9-kickstart.5.2.1.N1.1.bin',
##&nbsp;                                            'system_variable' => 'bootflash:/n5000-uk9.5.2.1.N1.1.bin'
##&nbsp;                                          }
##&nbsp;                                        ]
##&nbsp;          }
##&nbsp;        ];
##&nbsp;

sub show_boot {
    my %args = @_;
    my ($error_message);
    my $resp    = $args{response_object};
    my $output  = $resp->get_processed_output();
    my $command = $resp->get_command();
    my $error   = _check_error(
        raw_output    => $output,
        error_message => \$error_message
    );
    if ($error) {
        $error = 'NACL::APISet::Exceptions::' . $error . 'Exception';
        $error->throw( text => $error_message, response_object => $resp );
    }
    my @lines = get_lines( raw_output => $output, response_object => $resp );
    my $heading;
    my ( @values, @result_array );
    if ( $command =~ /variables$/ ) {
        if ( $lines[0] =~ /(List.*are)/ ) {
            $heading = $1;
        }
        for ( my $i = 1; $i <= $#lines; $i++ ) {
            $values[ $i - 1 ] = $lines[$i];
        }
        $result_array[0]->{ to_key_name( item => $heading ) } = \@values;
    } else {
        for ( my $i = 0; $i <= $#lines; $i++ ) {
            if (   $lines[$i] =~ /(Current.*Variables)/
                || $lines[$i] =~ /(Boot.*reload)/ )
            {
                $heading = $1;
            } elsif ( defined($heading) && $lines[$i] ne "" ) {
                @values = split( /\s+=\s+/, $lines[$i] );
                $result_array[0]->{ to_key_name( item => $heading ) }[0]
                    ->{ to_key_name( item => $values[0] ) } = $values[1];
            }
        }
    }
    return \@result_array;
}

## @name ::show_tech_support_module
##
## @summary Parses the raw output from the "show tech-support module" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from show_tech_support_module with any of its options as described in the
##    corresponding cdef file 'show.cdefs'
## @synopsis
##      show_tech_support_module(%args);
## INPUTS TO METHOD CALL 'show_tech_support_module' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## <tr> <td> No switch for this key </td> <td> 'module-number' </td> <td> '0' </td> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

##&nbsp;Command = show tech-support module 1
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;`show module`
##&nbsp;Mod Ports Module-Type                       Model                  Status
##&nbsp;--- ----- --------------------------------- ---------------------- ------------
##&nbsp;1   48    O2 48X10GE/Modular Supervisor     N5K-C5596UP-SUP        active *
##&nbsp;2   16    O2 16 port flexible GEM           N55-M16UP              ok
##&nbsp;3   16    O2 16 port flexible GEM           N55-M16UP              ok
##&nbsp;
##&nbsp;Mod  Sw              Hw      World-Wide-Name(s) (WWN)
##&nbsp;---  --------------  ------  --------------------------------------------------
##&nbsp;1    5.2(1)N1(8)     1.0     --
##&nbsp;2    5.2(1)N1(8)     1.0     --
##&nbsp;3    5.2(1)N1(8)     1.0     --
##&nbsp;
##&nbsp;Mod  MAC-Address(es)                         Serial-Num
##&nbsp;---  --------------------------------------  ----------
##&nbsp;1    547f.ee92.ff88 to 547f.ee92.ffb7         FOC16156EXG
##&nbsp;2    547f.ee99.8550 to 547f.ee99.855f         FOC16192DS7
##&nbsp;3    002a.6a0b.39e0 to 002a.6a0b.39ef         FOC170626UF
##&nbsp;`show environment`
##&nbsp;
##&nbsp;
##&nbsp;Fan:
##&nbsp;------------------------------------------------------
##&nbsp;Fan             Model                Hw         Status
##&nbsp;------------------------------------------------------
##&nbsp;Chassis-1       N5596UP-FAN-B        --         ok
##&nbsp;Chassis-2       N5596UP-FAN-B        --         ok
##&nbsp;Chassis-3       N5596UP-FAN-B        --         ok
##&nbsp;Chassis-4       N5596UP-FAN-B        --         ok
##&nbsp;PS-1            N55-PAC-1100W-B      --         ok
##&nbsp;PS-2            N55-PAC-1100W-B      --         ok
##&nbsp;
##&nbsp;
##&nbsp;Temperature
##&nbsp;-----------------------------------------------------------------
##&nbsp;Module   Sensor     MajorThresh   MinorThres   CurTemp     Status
##&nbsp;                    (Celsius)     (Celsius)    (Celsius)
##&nbsp;-----------------------------------------------------------------
##&nbsp;
##&nbsp;
##&nbsp;Power Supply:
##&nbsp;Voltage: 12 Volts
##&nbsp;-----------------------------------------------------------
##&nbsp;PS  Model                Input Power       Power     Status
##&nbsp;                         Type  (Watts)     (Amp)
##&nbsp;-----------------------------------------------------------
##&nbsp;1   N55-PAC-1100W-B      AC    1050.00     87.50     ok
##&nbsp;2   N55-PAC-1100W-B      AC    1050.00     87.50     ok
##&nbsp;
##&nbsp;
##&nbsp;Mod Model                   Power     Power       Power     Power       Status
##&nbsp;                            Requested Requested   Allocated Allocated
##&nbsp;                            (Watts)   (Amp)       (Watts)   (Amp)
##&nbsp;--- ----------------------  -------   ----------  --------- ----------  ----------
##&nbsp;1    N5K-C5596UP-SUP        648.00    54.00       648.00    54.00       powered-up
##&nbsp;2    N55-M16UP              90.00     7.50        90.00     7.50        powered-up
##&nbsp;3    N55-M16UP              90.00     7.50        90.00     7.50        powered-up
##&nbsp;
##&nbsp;
##&nbsp;Power Usage Summary:
##&nbsp;--------------------
##&nbsp;Power Supply redundancy mode:                 Redundant
##&nbsp;Power Supply redundancy operational mode:     Redundant
##&nbsp;
##&nbsp;Total Power Capacity                             2100.00 W
##&nbsp;
##&nbsp;Power reserved for Supervisor(s)                  648.00 W
##&nbsp;Power currently used by Modules                   180.00 W
##&nbsp;
##&nbsp;                                                -------------
##&nbsp;Total Power Available                            1272.00 W
##&nbsp;                                                -------------
##&nbsp;`show sprom sup`
##&nbsp;DISPLAY supervisor sprom contents:
##&nbsp;Common block:
##&nbsp; Block Signature : 0xabab
##&nbsp; Block Version   : 3
##&nbsp; Block Length    : 160
##&nbsp; Block Checksum  : 0x18e0
##&nbsp; EEPROM Size     : 65535
##&nbsp; Block Count     : 3
##&nbsp; FRU Major Type  : 0x6002
##&nbsp; FRU Minor Type  : 0x0
##&nbsp; OEM String      : Cisco Systems, Inc.
##&nbsp; Product Number  : N5K-C5596UP
##&nbsp; Serial Number   : FOC16156EXG
##&nbsp; Part Number     : 73-13374-05
##&nbsp; Part Revision   : D0
##&nbsp; Mfg Deviation   : 0
##&nbsp; H/W Version     : 1.0
##&nbsp; Mfg Bits        : 0
##&nbsp; Engineer Use    : 0
##&nbsp; snmpOID         : 9.12.3.1.9.72.17.0
##&nbsp; Power Consump   : -5400
##&nbsp; RMA Code        : 0-0-0-0
##&nbsp; CLEI Code       : CMMDC00ARA
##&nbsp; VID             : V01
##&nbsp;Supervisor Module specific block:
##&nbsp; Block Signature : 0x6002
##&nbsp; Block Version   : 2
##&nbsp; Block Length    : 103
##&nbsp; Block Checksum  : 0x902
##&nbsp; Feature Bits    : 0x0
##&nbsp; HW Changes Bits : 0x3
##&nbsp; Card Index      : 11044
##&nbsp; MAC Addresses   : 00-00-00-00-00-00
##&nbsp; Number of MACs  : 0
##&nbsp; Number of EPLD  : 0
##&nbsp; Port Type-Num   : 2-48
##&nbsp; Sensor #1       : 70,60
##&nbsp; Sensor #2       : -128,-128
##&nbsp; Sensor #3       : -128,-128
##&nbsp; Sensor #4       : -128,-128
##&nbsp; Sensor #5       : 65,55
##&nbsp; Sensor #6       : -128,-128
##&nbsp; Sensor #7       : -128,-128
##&nbsp; Sensor #8       : -128,-128
##&nbsp; Max Connector Power: 8000
##&nbsp; Cooling Requirement: 50
##&nbsp; Ambient Temperature: 40
##&nbsp;Temperature Sensor block:
##&nbsp; Block Signature : 0x6008
##&nbsp; Block Version   : 1
##&nbsp; Block Length    : 71
##&nbsp; Block Checksum  : 0x133
##&nbsp; Number of Valid Sensors : 1
##&nbsp; Sensor #9       : 70,60
##&nbsp; Sensor #10       : 0,0
##&nbsp; Sensor #11       : 0,0
##&nbsp; Sensor #12       : 0,0
##&nbsp; Sensor #13       : 0,0
##&nbsp; Sensor #14       : 0,0
##&nbsp; Sensor #15       : 0,0
##&nbsp; Sensor #16       : 0,0
##&nbsp; Sensor #17       : 0,0
##&nbsp; Sensor #18       : 0,0
##&nbsp; Sensor #19       : 0,0
##&nbsp; Sensor #20       : 0,0
##&nbsp; Sensor #21       : 0,0
##&nbsp; Sensor #22       : 0,0
##&nbsp; Sensor #23       : 0,0
##&nbsp; Sensor #24       : 0,0
##&nbsp; Sensor #25       : 0,0
##&nbsp; Sensor #26       : 0,0
##&nbsp; Sensor #27       : 0,0
##&nbsp; Sensor #28       : 0,0
##&nbsp; Sensor #29       : 0,0
##&nbsp; Sensor #30       : 0,0
##&nbsp; Sensor #31       : 0,0
##&nbsp; Sensor #32       : 0,0
##&nbsp; Sensor #33       : 0,0
##&nbsp; Sensor #34       : 0,0
##&nbsp; Sensor #35       : 0,0
##&nbsp; Sensor #36       : 0,0
##&nbsp; Sensor #37       : 0,0
##&nbsp; Sensor #38       : 0,0
##&nbsp; Sensor #39       : 0,0
##&nbsp; Sensor #40       : 0,0
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;          {
##&nbsp;            'module' => [
##&nbsp;                          {
##&nbsp;                            'module-type' => 'O2 48 10GE/Modular Supervisor',
##&nbsp;                            'ports' => '48',
##&nbsp;                            'status' => 'active *',
##&nbsp;                            'model' => 'N5K-C5596UP-SUP',
##&nbsp;                            'mod' => '1'
##&nbsp;                          },
##&nbsp;                          {
##&nbsp;                            'module-type' => 'O2 16 port flexible GEM',
##&nbsp;                            'ports' => '16',
##&nbsp;                            'status' => 'ok',
##&nbsp;                            'model' => 'N55-M16UP',
##&nbsp;                            'mod' => '2'
##&nbsp;                          },
##&nbsp;                          {
##&nbsp;                            'module-type' => 'O2 16 port flexible GEM',
##&nbsp;                            'ports' => '16',
##&nbsp;                            'status' => 'ok',
##&nbsp;                            'model' => 'N55-M16UP',
##&nbsp;                            'mod' => '3'
##&nbsp;                          },
##&nbsp;                          {
##&nbsp;                            'world-wide-names' => '--',
##&nbsp;                            'hw' => '1.0',
##&nbsp;                            'sw' => '5.2(1)N1(8)',
##&nbsp;                            'mod' => '1'
##&nbsp;                          },
##&nbsp;                          {
##&nbsp;                            'world-wide-names' => '--',
##&nbsp;                            'hw' => '1.0',
##&nbsp;                            'sw' => '5.2(1)N1(8)',
##&nbsp;                            'mod' => '2'
##&nbsp;                          },
##&nbsp;                          {
##&nbsp;                            'world-wide-names' => '--',
##&nbsp;                            'hw' => '1.0',
##&nbsp;                            'sw' => '5.2(1)N1(8)',
##&nbsp;                            'mod' => '3'
##&nbsp;                          },
##&nbsp;                          {
##&nbsp;                            'serial-num' => 'FOC16156E G',
##&nbsp;                            'mac-addresses' => '547f.ee92.ff88 to 547f.ee92.ffb7',
##&nbsp;                            'mod' => '1'
##&nbsp;                          },
##&nbsp;                          {
##&nbsp;                            'serial-num' => 'FOC16192DS7',
##&nbsp;                            'mac-addresses' => '547f.ee99.8550 to 547f.ee99.855f',
##&nbsp;                            'mod' => '2'
##&nbsp;                          },
##&nbsp;                          {
##&nbsp;                            'serial-num' => 'FOC170626UF',
##&nbsp;                            'mac-addresses' => '002a.6a0b.39e0 to 002a.6a0b.39ef',
##&nbsp;                            'mod' => '3'
##&nbsp;                          }
##&nbsp;                        ]
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'environment' => [
##&nbsp;                               {
##&nbsp;                                 'fan' => [
##&nbsp;                                            {
##&nbsp;                                              'fan' => 'Chassis-1',
##&nbsp;                                              'hw' => '--',
##&nbsp;                                              'status' => 'ok',
##&nbsp;                                              'model' => 'N5596UP-FAN-B'
##&nbsp;                                            },
##&nbsp;                                            {
##&nbsp;                                              'fan' => 'Chassis-2',
##&nbsp;                                              'hw' => '--',
##&nbsp;                                              'status' => 'ok',
##&nbsp;                                              'model' => 'N5596UP-FAN-B'
##&nbsp;                                            },
##&nbsp;                                            {
##&nbsp;                                              'fan' => 'Chassis-3',
##&nbsp;                                              'hw' => '--',
##&nbsp;                                              'status' => 'ok',
##&nbsp;                                              'model' => 'N5596UP-FAN-B'
##&nbsp;                                            },
##&nbsp;                                            {
##&nbsp;                                              'fan' => 'Chassis-4',
##&nbsp;                                              'hw' => '--',
##&nbsp;                                              'status' => 'ok',
##&nbsp;                                              'model' => 'N5596UP-FAN-B'
##&nbsp;                                            },
##&nbsp;                                            {
##&nbsp;                                              'fan' => 'PS-1',
##&nbsp;                                              'hw' => '--',
##&nbsp;                                              'status' => 'ok',
##&nbsp;                                              'model' => 'N55-PAC-1100W-B'
##&nbsp;                                            },
##&nbsp;                                            {
##&nbsp;                                              'fan' => 'PS-2',
##&nbsp;                                              'hw' => '--',
##&nbsp;                                              'status' => 'ok',
##&nbsp;                                              'model' => 'N55-PAC-1100W-B'
##&nbsp;                                            }
##&nbsp;                                          ]
##&nbsp;                               },
##&nbsp;                               {
##&nbsp;                                 'power_supply' => [
##&nbsp;                                                     {
##&nbsp;                                                       'voltage' => '12 Volts'
##&nbsp;                                                     },
##&nbsp;                                                     {
##&nbsp;                                                       'status' => 'ok',
##&nbsp;                                                       'power_amp' => '87.50',
##&nbsp;                                                       'input_type' => 'AC',
##&nbsp;                                                       'model' => 'N55-PAC-1100W-B',
##&nbsp;                                                       'power_watts' => '1050.00',
##&nbsp;                                                       'ps' => '1'
##&nbsp;                                                     },
##&nbsp;                                                     {
##&nbsp;                                                       'status' => 'ok',
##&nbsp;                                                       'power_amp' => '87.50',
##&nbsp;                                                       'input_type' => 'AC',
##&nbsp;                                                       'model' => 'N55-PAC-1100W-B',
##&nbsp;                                                       'power_watts' => '1050.00',
##&nbsp;                                                       'ps' => '2'
##&nbsp;                                                     },
##&nbsp;                                                     {
##&nbsp;                                                       'power_requested_watts' => '648.00',
##&nbsp;                                                       'power_allocated_amp' => '54.00',
##&nbsp;                                                       'status' => 'powered-up',
##&nbsp;                                                       'power_requested_amp' => '54.00',
##&nbsp;                                                       'power_allocated_watts' => '648.00',
##&nbsp;                                                       'model' => 'N5K-C5596UP-SUP',
##&nbsp;                                                       'mod' => '1'
##&nbsp;                                                     },
##&nbsp;                                                     {
##&nbsp;                                                       'power_requested_watts' => '90.00',
##&nbsp;                                                       'power_allocated_amp' => '7.50',
##&nbsp;                                                       'status' => 'powered-up',
##&nbsp;                                                       'power_requested_amp' => '7.50',
##&nbsp;                                                       'power_allocated_watts' => '90.00',
##&nbsp;                                                       'model' => 'N55-M16UP',
##&nbsp;                                                       'mod' => '2'
##&nbsp;                                                     },
##&nbsp;                                                     {
##&nbsp;                                                       'power_requested_watts' => '90.00',
##&nbsp;                                                       'power_allocated_amp' => '7.50',
##&nbsp;                                                       'status' => 'powered-up',
##&nbsp;                                                       'power_requested_amp' => '7.50',
##&nbsp;                                                       'power_allocated_watts' => '90.00',
##&nbsp;                                                       'model' => 'N55-M16UP',
##&nbsp;                                                       'mod' => '3'
##&nbsp;                                                     }
##&nbsp;                                                   ]
##&nbsp;                               }
##&nbsp;                             ]
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'power_usage_summary' => {
##&nbsp;                                       'power_currently_used_by_modules' => '180.00 W',
##&nbsp;                                       'total_power_capacity' => '2100.00 W',
##&nbsp;                                       'power_supply_redundancy_mode' => 'Redundant',
##&nbsp;                                       'power_reserved_for_supervisors' => '648.00 W',
##&nbsp;                                       'total_power_available' => '1272.00 W',
##&nbsp;                                       'power_supply_redundancy_operational_mode' => 'Redundant'
##&nbsp;                                     }
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'sprom_sup' => [
##&nbsp;                             {
##&nbsp;                               'common_block' => {
##&nbsp;                                                   'serial_number' => 'FOC16156EXG',
##&nbsp;                                                   'vid' => 'V01',
##&nbsp;                                                   'part_revision' => 'D0',
##&nbsp;                                                   'oem_string' => 'Cisco Systems, Inc.',
##&nbsp;                                                   'fru_major_type' => '0x6002',
##&nbsp;                                                   'snmpoid' => '9.12.3.1.9.72.17.0',
##&nbsp;                                                   'block_count' => '3',
##&nbsp;                                                   'power_consump' => '-5400',
##&nbsp;                                                   'h/w_version' => '1.0',
##&nbsp;                                                   'clei_code' => 'CMMDC00ARA',
##&nbsp;                                                   'product_number' => 'N5K-C5596UP',
##&nbsp;                                                   'block_length' => '160',
##&nbsp;                                                   'engineer_use' => '0',
##&nbsp;                                                   'fru_minor_type' => '0x0',
##&nbsp;                                                   'block_version' => '3',
##&nbsp;                                                   'rma_code' => '0-0-0-0',
##&nbsp;                                                   'block_checksum' => '0x18e0',
##&nbsp;                                                   'block_signature' => '0xabab',
##&nbsp;                                                   'part_number' => '73-13374-05',
##&nbsp;                                                   'mfg_deviation' => '0',
##&nbsp;                                                   'mfg_bits' => '0',
##&nbsp;                                                   'eeprom_size' => '65535'
##&nbsp;                                                 }
##&nbsp;                             },
##&nbsp;                             {
##&nbsp;                               'supervisor_module_specific_block' => {
##&nbsp;                                                                       'mac_addresses' => '00-00-00-00-00-00',
##&nbsp;                                                                       'ambient_temperature' => '40',
##&nbsp;                                                                       'feature_bits' => '0x0',
##&nbsp;                                                                       'sensor_#7' => '-128,-128',
##&nbsp;                                                                       'sensor_#1' => '70,60',
##&nbsp;                                                                       'max_connector_power' => '8000',
##&nbsp;                                                                       'sensor_#6' => '-128,-128',
##&nbsp;                                                                       'number_of_macs' => '0',
##&nbsp;                                                                       'sensor_#2' => '-128,-128',
##&nbsp;                                                                       'card_index' => '11044',
##&nbsp;                                                                       'number_of_epld' => '0',
##&nbsp;                                                                       'cooling_requirement' => '50',
##&nbsp;                                                                       'sensor_#4' => '-128,-128',
##&nbsp;                                                                       'block_length' => '103',
##&nbsp;                                                                       'hw_changes_bits' => '0x3',
##&nbsp;                                                                       'sensor_#3' => '-128,-128',
##&nbsp;                                                                       'block_version' => '2',
##&nbsp;                                                                       'sensor_#8' => '-128,-128',
##&nbsp;                                                                       'block_checksum' => '0x902',
##&nbsp;                                                                       'block_signature' => '0x6002',
##&nbsp;                                                                       'port_type-num' => '2-48',
##&nbsp;                                                                       'sensor_#5' => '65,55'
##&nbsp;                                                                     }
##&nbsp;                             },
##&nbsp;                             {
##&nbsp;                               'temperature_sensor_block' => {
##&nbsp;                                                               'sensor_#11' => '0,0',
##&nbsp;                                                               'sensor_#24' => '0,0',
##&nbsp;                                                               'sensor_#12' => '0,0',
##&nbsp;                                                               'number_of_valid_sensors' => '1',
##&nbsp;                                                               'sensor_#26' => '0,0',
##&nbsp;                                                               'sensor_#23' => '0,0',
##&nbsp;                                                               'sensor_#17' => '0,0',
##&nbsp;                                                               'sensor_#9' => '70,60',
##&nbsp;                                                               'block_length' => '71',
##&nbsp;                                                               'sensor_#33' => '0,0',
##&nbsp;                                                               'sensor_#18' => '0,0',
##&nbsp;                                                               'sensor_#37' => '0,0',
##&nbsp;                                                               'block_version' => '1',
##&nbsp;                                                               'sensor_#20' => '0,0',
##&nbsp;                                                               'sensor_#15' => '0,0',
##&nbsp;                                                               'sensor_#16' => '0,0',
##&nbsp;                                                               'block_checksum' => '0x133',
##&nbsp;                                                               'block_signature' => '0x6008',
##&nbsp;                                                               'sensor_#27' => '0,0',
##&nbsp;                                                               'sensor_#13' => '0,0',
##&nbsp;                                                               'sensor_#36' => '0,0',
##&nbsp;                                                               'sensor_#25' => '0,0',
##&nbsp;                                                               'sensor_#40' => '0,0',
##&nbsp;                                                               'sensor_#14' => '0,0',
##&nbsp;                                                               'sensor_#31' => '0,0',
##&nbsp;                                                               'sensor_#34' => '0,0',
##&nbsp;                                                               'sensor_#39' => '0,0',
##&nbsp;                                                               'sensor_#28' => '0,0',
##&nbsp;                                                               'sensor_#21' => '0,0',
##&nbsp;                                                               'sensor_#30' => '0,0',
##&nbsp;                                                               'sensor_#19' => '0,0',
##&nbsp;                                                               'sensor_#29' => '0,0',
##&nbsp;                                                               'sensor_#32' => '0,0',
##&nbsp;                                                               'sensor_#38' => '0,0',
##&nbsp;                                                               'sensor_#10' => '0,0',
##&nbsp;                                                               'sensor_#22' => '0,0',
##&nbsp;                                                               'sensor_#35' => '0,0'
##&nbsp;                                                             }
##&nbsp;                             }
##&nbsp;                           ]
##&nbsp;          }
##&nbsp;        ];
##&nbsp;
sub show_tech_support_module {
    my %args     = @_;
    my $response = $args{response_object};
    my $command  = $response->get_actual_command();
    my $output   = $response->get_processed_output();
    my $result;
    my $error_message;
    my $error   = _check_error(
        raw_output    => $output,
        error_message => \$error_message
    );
    if ($error) {
        $error = 'NACL::APISet::Exceptions::' . $error . 'Exception';
        $error->throw( text => $error_message, response_object => $response );
    }
    my @lines =
      get_lines( raw_output => $output, response_object => $response );
    my ( $line, $next_line, $i, $j, $k, $main_heading, $sub_heading );
    my ( @result_array, @result_array1, @result_array3, @result_array2 );
    my ( @headings, @headings2, @headings3, $flag, $flag1, $flag2, $flag7 );
    my ( $count, @column_headings, @main_headings, @sub_headings );
    $count = -1;
    my $count1 = 0;
    my $count3 = 0;
    $k           = 0;
    $sub_heading = 0;

    while (@lines) {
        $line = shift(@lines);
        my $temp;
        ## collect the headings from here
        if ( $line =~ /(.+)/ ) {
            $temp      = $line;
            $next_line = shift(@lines);
            if ( $temp =~ /((.+)[:]+$|Temperature|show (.+))/i ) {
                if ( $next_line =~ /(.+)[:]+$/ ) {
                    $sub_heading = $1;
                    $flag1       = 1;
                }
                elsif ( $flag1 && $temp =~ /(.+)[:]+$/ ) {
                    $sub_heading = $1;
                    unshift @lines, $next_line;
                }
                elsif ( $temp =~ /(Temperature)/i ) {
                    $sub_heading = $1;
                }
                if ( !$flag1 ) {
                    $main_heading = $1;
                    $main_heading =~ s/show\s+//g;
                    $main_heading =~ s/[`]//g;

                }
                elsif ( $flag1 && ( $temp =~ /show (.+)/ ) ) {
                    $main_heading = $1;
                    $main_heading =~ s/[`]//g;
                    unshift @lines, $next_line;
                }
                if ( $sub_heading =~ /Summary/i ) {
                    if ($count) {
                        $result_array3[ $count3++ ]->{
                            to_key_name(
                                'item' => $sub_headings[$#sub_headings]
                            )
                          }
                          = \@result_array
                          if (@result_array);
                        $result_array2[ $k++ ]->{
                            to_key_name(
                                'item' => $main_headings[$#main_headings]
                            )
                          }
                          = \@result_array3;
                        $count = 0;
                    }
                    $line = shift(@lines);
                    while ( $line !~ /[`]+/i ) {
                        $line = shift(@lines);
                        if ( $line =~ /---/ ) {
                            $line = shift(@lines);
                        }
                        $line =~ s/\s\s+/Y/g;
                        $line =~ s/\s/X/g;
                        if ( $line !~ /[`]+/i ) {
                            my @values = split( /Y/, $line );
                            $values[0] =~ s/X/ /g;
                            $values[1] =~ s/X/ /g;
                            $result_array2[$k]
                              ->{ to_key_name( 'item' => $sub_heading ) }
                              ->{ to_key_name( 'item' => $values[0] ) } =
                              $values[1];
                        }
                    }
                    $line =~ s/X/ /g;
                    $k++;
                    unshift @lines, $line;
                }
                elsif ( ( $sub_heading !~ /Temperature$/i ) ) {
                    $next_line = shift(@lines) if ($sub_heading);
                    if (   ( $next_line !~ /[---]+/ )
                        && ( $next_line !~ /Voltage/ ) )
                    {
                        while ($next_line
                            && ( $next_line !~ /(.+)[:]+$/ )
                            && $sub_heading )
                        {
                            my $temp = $next_line;
                            my @temp_arr = split( /[:]+/, $temp );
                            $temp_arr[0] =~ s/\s+$//g;
                            $temp_arr[1] =~ s/^\s+//g;
                            $result_array1[$count1]
                              ->{ to_key_name( 'item' => $sub_heading ) }
                              ->{ to_key_name( 'item' => $temp_arr[0] ) } =
                              $temp_arr[1];
                            $flag2++;
                            $next_line = shift(@lines);
                        }
                        $count1++ if ($flag2);
                        $result_array2[ $k++ ]
                          ->{ to_key_name( 'item' => $main_heading ) } =
                          \@result_array1
                          if ( $count1 == 3 );
                        unshift @lines, $next_line if ( !$flag2 );
                    }
                    unshift @lines, $next_line;
                }
            }
            elsif ( $next_line =~ /(Celsius|Watts|Allocated)/ ) {
                @headings = split( /\s+/, $temp );
                $next_line =~ s/\(//g;
                $next_line =~ s/\)//g;
                @headings2 = split( /\s+/, $next_line );
                $headings[2] = $headings[2] . ' ' . $headings2[0];
                $headings[3] = $headings[3] . ' ' . $headings2[1]
                  if ( $headings2[1] );
                $headings[4] = $headings[4] . ' ' . $headings2[2]
                  if ( $headings2[2] );
                $headings[5] = $headings[5] . ' ' . $headings2[3]
                  if ( $headings2[3] );
                @headings3 = @headings;
                $flag++;
            }
            elsif ( ( $next_line =~ /^---/ ) && ( $temp !~ /^Voltage/i ) ) {
                if ( !$flag ) {
                    @headings = split( /\s+/, $temp );
                    @headings = grep { $_ !~ /WWN/ } @headings;
                    @headings3 = @headings;
                }
                elsif ( $flag >= 2 ) {
                    @headings = split( /\s+/, $temp );
                    $headings3[2] = $headings3[2] . ' ' . $headings[0];
                    $headings3[3] = $headings3[3] . ' ' . $headings[1]
                      if ( $headings[1] );
                    $headings3[4] = $headings3[4] . ' ' . $headings[2]
                      if ( $headings[2] );
                    $headings3[5] = $headings3[5] . ' ' . $headings[3]
                      if ( $headings[3] );
                    $flag = 0;
                }
            }
            else {
                unshift @lines, $next_line;
            }
            my $flag_while = 0;
            my $pattern1   = $headings3[0];
            if (   ( $line !~ /^$pattern1/i )
                && ( $line !~ /$main_heading/i )
                && ( $line !~ /^$sub_heading/i )
                && ( !$flag2 ) )
            {
                if ( $line !~ /(^[`]+|^[---]+|^\(Watts\))/i ) {
                    my @values;
                    my $flag8 = 0;
                    if ( $line =~ /[:]+/ ) {
                        @values = split( /[:]+/, $line );
                        $values[0] =~ s/\s+$//g;
                        $values[1] =~ s/^\s+//g;
                        $flag8 = 1;
                    }
                    else {
                        $line =~ s/\s\s+/Y/g;
                        $line =~ s/\s/X/g;
                        @values = split( /Y/, $line );
                    }
                    if (   grep( /$headings3[2]/, @column_headings )
                        && grep( /$main_heading/, @main_headings ) )
                    {
                        $count++;
                    }
                    else {
                        if (   !grep( /$headings3[2]/, @column_headings )
                            && ( grep( /$main_heading/, @main_headings ) )
                            && ( !$sub_heading ) )
                        {
                            $count++;
                        }
                        elsif ( !grep( /$main_heading/, @main_headings )
                            && @main_headings )
                        {
                            my @back_up_arr_1 = @result_array;
                            $result_array2[ $k++ ]->{
                                to_key_name(
                                    'item' => $main_headings[$#main_headings]
                                )
                              }
                              = \@back_up_arr_1
                              if ( @back_up_arr_1 && !$flag7 );
                            $count        = 0;
                            @result_array = ();
                        }
                        elsif (!grep( /$headings3[2]/, @column_headings )
                            && ( grep( /$main_heading/, @main_headings ) )
                            && ( !grep( /$sub_heading/, @sub_headings ) ) )
                        {
                            my @back_up_arr_2 = @result_array;
                            $result_array3[ $count3++ ]->{
                                to_key_name(
                                    'item' => $sub_headings[$#sub_headings]
                                )
                              }
                              = \@back_up_arr_2
                              if (@back_up_arr_2);
                            $flag7        = 1;
                            $count        = 0;
                            @result_array = ();
                        }
                        else {
                            $count++;
                        }
                    }
                    push( @main_headings,   $main_heading );
                    push( @sub_headings,    $sub_heading );
                    push( @column_headings, $headings3[2] );
                    if ( !$flag8 ) {
                        for ( $i = 0 ; $i <= $#headings3 ; $i++ ) {
                            $values[$i] =~ s/[X]+/ /g;
                            $result_array[$count]
                              ->{ to_key_name( 'item' => $headings3[$i] ) } =
                              $values[$i];

                        }
                    }
                    else {
                        $result_array[$count]
                          ->{ to_key_name( 'item' => $values[0] ) } =
                          $values[1];
                    }

                }
            }
        }
    }
    return \@result_array2;
}
1;
