#
#Copyright (c) 20155555 NetAPP Appliance, Inc.
#All rights reserved.
#
## @summary Routines to parse Mellanox interface commands
##
## @description
##    This is a library to parse the raw output of interface commands.
##
## @synopsis
##  use NACL::APISet::Switch::CLI::Mellanox::default::interface
##
##  interface(%args)
##
## @author dl-nacl-dev@netapp.com
##
## @status public
#
package NACL::APISet::Switch::CLI::Mellanox::default::interface;

use strict;

#Include exceptions which will be thrown
use NACL::APISet::Exceptions::InvalidParamValueException;
use NACL::APISet::Exceptions::CommandNotFoundException;

sub _check_error {
    my %args = @_;

    my $raw_output = $args{raw_output};
    my $resp = $args{"response_object"};
    my $error_message;
    my $error;

    # Check for the errors
    if ($raw_output =~ /[iI]nvalid command/) {
        $error = "InvalidParamValue";
        $error_message = "Invalid command option";
    } elsif ($raw_output =~ /Unrecognized\s+command/) {
       $error_message = $raw_output;
       $error          = "CommandNotFound";
    } elsif ( $raw_output =~ /Unknown\s+interface.*/ ) {
        $error_message = $raw_output;
        $error         = "InvalidParamValue";
    }

    if ($error) {
        $error = 'NACL::APISet::Exceptions::'.$error.'Exception';
        $error->throw(text => $error_message, response_object => $resp);
    }
}

## @name ::interface
##
## @summary Parses the raw output from the "interface" command
##
## @description 
##    This method can be used for parsing the raw output obtained after calling 
##    the method from interface with any of its options as described in the 
##    corresponding cdef file 'interface.cdefs'
## @synopsis
##      interface(%args);
## INPUTS TO METHOD CALL 'interface' ON APISET OBJECT 
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## <tr> <td> No switch for this key </td> <td> 'field' </td> <td> '0' </td> </tr> 
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

##&nbsp;Command = interface mgmt0
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [];
##&nbsp;
sub interface {
    my %args = @_;

    my $resp         = $args{response_object};
    my $raw_output   = $resp->get_processed_output();

    _check_error(raw_output=>$raw_output, response_object=>$resp);

    return [];
}

## @name ::interface_eth
##
## @summary Parses the raw output from the "interface eth" command
##
## @description 
##    This method can be used for parsing the raw output obtained after calling 
##    the method from interface_eth with any of its options as described in the 
##    corresponding cdef file 'interface.cdefs'
## @synopsis
##      interface_eth(%args);
## INPUTS TO METHOD CALL 'interface_eth' ON APISET OBJECT 
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## <tr> <td> No switch for this key </td> <td> 'port' </td> <td> '0' </td> </tr> 
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

##&nbsp;Command = interface eth 1/9
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [];
##&nbsp;
sub interface_eth {
    my %args = @_;

    my $resp          = $args{response_object};
    my $raw_output   = $resp->get_processed_output();

    _check_error(raw_output=>$raw_output, response_object=>$resp);

    return [];
}

sub interface_ib {
    my %args = @_;

    my $resp          = $args{response_object};
    my $raw_output   = $resp->get_processed_output();

    _check_error(raw_output=>$raw_output, response_object=>$resp);

    return [];
}

1;
