#Copyright (c)2015 NetApp, Inc.
#All rights reserved.
#
## @summary Routines to parse Mellanox show and its subcommands
##
## @description
##    This is a library to parse the raw output of Mellanox switch show and its subcommands.
##
## @synopsis
##  use NACL::APISet::Switch::CLI::Mellanox::default::show
##
## show_version(%args)
## show_bootvar(%args)
## show_voltage(%args)
## show_mac_addr_table(%args)
## show_vlan(%args)
## show_interface(%args)
##
## @status public

package NACL::APISet::Switch::CLI::Mellanox::default::show;

use strict;
use NACL::APISet::Response::CLI::ParserUtils;

#Include exceptions which will be thrown
use NACL::APISet::Exceptions::MissingRequiredParamException qw(:try);
use NACL::APISet::Exceptions::InvalidParamException;
use NACL::APISet::Exceptions::InvalidParamValueException;
use NACL::APISet::Exceptions::CommandNotFoundException;

sub _check_error {
    my %args = @_;

    my $raw_output    = $args{raw_output};
    my $command       = $args{command};
    my $error_message = $args{error_message};
    my $error;

    # Check for the error
    if ( $raw_output =~ /Invalid\s+input\s+detected/i ) {
        $$error_message = $raw_output;
        $error          = "InvalidParam";
    } elsif ( $raw_output =~ /(Incomplete\s+command)/i ) {
        $$error_message = $raw_output;
        $error          = "MissingRequiredParam";
    } elsif ( $raw_output
        =~ /An\s+invalid\s+interface\s+has\s+been\s+used\s+for\s+this\s+command/
        ) {
        $$error_message = $raw_output;
        $error          = "InvalidParamValue";
    } elsif ( $raw_output
        =~ /The\s+group\s+does\s+not\s+exist|The\s+specified\s+interface\s+does\s+not\s+exist./
        ) {
        $$error_message = $raw_output;
        $error          = "InvalidParamValue";
    } elsif ($raw_output
        =~ /Unrecognized\s+command/
       ) {
       $$error_message = $raw_output;
       $error          = "CommandNotFound";
    }

    return $error;
} ## end sub _check_error

## @name ::show_version
##
## @summary Parses the raw output from the "show version" command
##
## @description 
##    This method can be used for parsing the raw output obtained after calling 
##    the method from show_version with any of its options as described in the 
##    corresponding cdef file 'show.cdefs'
## @synopsis
##      show_version(%args);
## INPUTS TO METHOD CALL 'show_version' ON APISET OBJECT 
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

##&nbsp;Command = show version
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;Product name:      SX_PPC_M460EX
##&nbsp;Product release:   SX_3.3.3504
##&nbsp;Build ID:          #1-dev
##&nbsp;Build date:        2013-05-31 10:59:20
##&nbsp;Target arch:       ppc
##&nbsp;Target hw:         m460ex
##&nbsp;Built by:          jenkins@fit-dhcp192
##&nbsp;Version summary:   SX_PPC_M460EX SX_3.3.3504 2013-05-31 10:59:20 ppc
##&nbsp;
##&nbsp;Product model:     ppc
##&nbsp;Host ID:           0002C95EC450
##&nbsp;
##&nbsp;Uptime:            106d 3h 17m 58.400s
##&nbsp;CPU load averages: 1.03 / 1.02 / 1.06
##&nbsp;Number of CPUs:    1
##&nbsp;System memory:     206 MB used / 1821 MB free / 2027 MB total
##&nbsp;Swap:              0 MB used / 0 MB free / 0 MB total
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;          {
##&nbsp;            'build_id' => '#1-dev',
##&nbsp;            'target_hw' => 'm460ex',
##&nbsp;            'host_id' => '0002C95EC450',
##&nbsp;            'product_model' => 'ppc',
##&nbsp;            'target_arch' => 'ppc',
##&nbsp;            'build_date' => '2013-05-31 10:59:20',
##&nbsp;            'cpu_load_averages' => '1.03 / 1.02 / 1.06',
##&nbsp;            'product_release' => 'SX_3.3.3504',
##&nbsp;            'swap' => '0 MB used / 0 MB free / 0 MB total',
##&nbsp;            'uptime' => '106d 3h 17m 58.400s',
##&nbsp;            'number_of_cpus' => '1',
##&nbsp;            'version_summary' => 'SX_PPC_M460EX SX_3.3.3504 2013-05-31 10:59:20 ppc',
##&nbsp;            'system_memory' => '206 MB used / 1821 MB free / 2027 MB total',
##&nbsp;            'product_name' => 'SX_PPC_M460EX',
##&nbsp;            'built_by' => 'jenkins@fit-dhcp192'
##&nbsp;          }
##&nbsp;        ];
##&nbsp;
sub show_version {
    my %args = @_;
    my ($error_message, @result_array);
    my $resp = $args{response_object};
    my $output = $resp->get_processed_output();
    my $command = $resp->get_command();
    my $error_level = $resp->get_execution_status();
    my ($error, $key);

    $error = _check_error(raw_output=>$output,command=>"version",
                             error_message=>\$error_message);
    if($error) {
        $error = 'NACL::APISet::Exceptions::'.$error.'Exception';
        $error->throw(text=>$error_message,response_object => $resp);
    }
    my @lines = get_lines("raw_output"=>$output,"response_object"=>$resp);
    my (@res,$value);
    my $i = 0;
    foreach $value (@lines){
        if($value =~/^([\w\s]+):\s+(.*)$/){
            $res[$i]->{to_key_name(item => $1)} = $2;
        }
    }
    return \@res;
} ## end sub show_version

## @name ::show_bootvar
##
## @summary Parses the raw output from the "show bootvar" command
##
## @description 
##    This method can be used for parsing the raw output obtained after calling 
##    the method from show_bootvar with any of its options as described in the 
##    corresponding cdef file 'show.cdefs'
## @synopsis
##      show_bootvar(%args);
## INPUTS TO METHOD CALL 'show_bootvar' ON APISET OBJECT 
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

##&nbsp;Command = show bootvar
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;Installed images:
##&nbsp;
##&nbsp;  Partition 1:
##&nbsp;  SX_PPC_M460EX SX_3.3.4100 2013-09-16 18:27:29 ppc
##&nbsp;
##&nbsp;  Partition 2:
##&nbsp;  SX_PPC_M460EX SX_3.3.4100 2013-09-16 18:27:29 ppc
##&nbsp;
##&nbsp;Last boot partition: 1
##&nbsp;Next boot partition: 1
##&nbsp;
##&nbsp;Serve image files via HTTP/HTTPS: no
##&nbsp;
##&nbsp;Boot manager password is set.
##&nbsp;
##&nbsp;Image signing: trusted signature always required
##&nbsp;Admin require signed images: yes
##&nbsp;
##&nbsp;Settings for next boot only:
##&nbsp;   Fallback reboot on configuration failure: yes (default)
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;          {
##&nbsp;            'next_boot_partition' => '1',
##&nbsp;            'installed_images' => [
##&nbsp;                                    'SX_PPC_M460EX SX_3.3.4100 2013-09-16 18:27:29 ppc',
##&nbsp;                                    'SX_PPC_M460EX SX_3.3.4100 2013-09-16 18:27:29 ppc'
##&nbsp;                                  ],
##&nbsp;            'fallback_reboot_on_configuration_failure' => 'yes (default)',
##&nbsp;            'serve_image_files_via_http_https' => 'no',
##&nbsp;            'admin_require_signed_images' => 'yes',
##&nbsp;            'image_signing' => 'trusted signature always required',
##&nbsp;            'last_boot_partition' => '1'
##&nbsp;          }
##&nbsp;        ];
##&nbsp;
sub show_bootvar {
    my %args = @_;

    my ($error_message);

    my $resp        = $args{response_object};
    my $output      = $resp->get_processed_output();
    my $command     = $resp->get_command();
    my $error_level = $resp->get_execution_status();
    my ( @values, @key_names, $key, $k );

    my $error = _check_error(
        raw_output    => $output,
        error_message => \$error_message
    );
    if ($error) {
        $error = 'NACL::APISet::Exceptions::' . $error . 'Exception';
        $error->throw( text => $error_message, response_object => $resp );
    }
    my $k = 0 ;
    my (@result_array, $key);
    my @lines = get_lines( raw_output => $output, response_object => $resp );
    my $line = shift(@lines);
    while (@lines) {
        if ( $line=~ /(Installed\s+images)/ ){
            $key = to_key_name( item => $1 );
        }elsif ( $line =~ /(Partition\s\d):/ ) {
            my $key1 = $1;
            $line = shift(@lines) ;
            $result_array[0]->{$key}->[$k] = $line;
            $k++;
        }elsif($line =~/^([\s+\w\s\/]+):\s+(.*)$/){
            my $key = $1 ;
            my $value = $2 ;
            $key =~ s|/|_|g;
            $result_array[0]->{to_key_name(item => $key)} = $value;
        }elsif($line =~ /Settings for next/) {
             $line =shift(@lines);
             my @arr = split(/:/,$line);
             $arr[1] =~ s/^\s+//;
             $result_array[0]->{to_key_name(item => $arr[0])} = $arr[1] ;
        }
        $line = shift(@lines);
    }
    return \@result_array;
} ## end sub show_bootvar

## @name ::show_voltage
##
## @summary Parses the raw output from the "show voltage" command
##
## @description 
##    This method can be used for parsing the raw output obtained after calling 
##    the method from show_voltage with any of its options as described in the 
##    corresponding cdef file 'show.cdefs'
## @synopsis
##      show_voltage(%args);
## INPUTS TO METHOD CALL 'show_voltage' ON APISET OBJECT 
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

##&nbsp;Command = show voltage
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;==========================================================================
##&nbsp;Module  Power Meter           Reg  Expected  Actual   Status  High   Low  
##&nbsp;                                   Voltage   Voltage          Range  Range
##&nbsp;==========================================================================
##&nbsp;MGMT    BOARD_MONITOR         V1   5.00      5.02     OK      5.75   4.25 
##&nbsp;MGMT    BOARD_MONITOR         V2   2.27      2.12     OK      2.61   1.93 
##&nbsp;MGMT    BOARD_MONITOR         V3   1.80      1.81     OK      2.07   1.53 
##&nbsp;MGMT    BOARD_MONITOR         V4   3.30      3.26     OK      3.79   2.80 
##&nbsp;MGMT    BOARD_MONITOR         V5   0.90      0.93     OK      1.10   0.81 
##&nbsp;MGMT    BOARD_MONITOR         V6   1.20      1.19     OK      1.38   1.02 
##&nbsp;MGMT    CPU_BOARD_MONITOR     V1   12.00     11.67    OK      13.80  10.20
##&nbsp;MGMT    CPU_BOARD_MONITOR     V2   2.50      2.47     OK      2.88   2.12 
##&nbsp;MGMT    CPU_BOARD_MONITOR     V3   3.30      3.28     OK      3.79   2.80 
##&nbsp;MGMT    CPU_BOARD_MONITOR     V4   3.30      3.26     OK      3.79   2.80 
##&nbsp;MGMT    CPU_BOARD_MONITOR     V5   1.80      1.79     OK      2.07   1.53 
##&nbsp;MGMT    CPU_BOARD_MONITOR     V6   1.20      1.23     OK      1.38   1.02 
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;          {
##&nbsp;            'low_range' => '4.25',
##&nbsp;            'high_range' => '5.75',
##&nbsp;            'actual_voltage' => '5.02',
##&nbsp;            'status' => 'OK',
##&nbsp;            'expected_voltage' => '5.00',
##&nbsp;            'power_meter' => 'BOARD_MONITOR',
##&nbsp;            'reg' => 'V1',
##&nbsp;            'module' => 'MGMT'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'low_range' => '1.93',
##&nbsp;            'high_range' => '2.61',
##&nbsp;            'actual_voltage' => '2.12',
##&nbsp;            'status' => 'OK',
##&nbsp;            'expected_voltage' => '2.27',
##&nbsp;            'power_meter' => 'BOARD_MONITOR',
##&nbsp;            'reg' => 'V2',
##&nbsp;            'module' => 'MGMT'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'low_range' => '1.53',
##&nbsp;            'high_range' => '2.07',
##&nbsp;            'actual_voltage' => '1.81',
##&nbsp;            'status' => 'OK',
##&nbsp;            'expected_voltage' => '1.80',
##&nbsp;            'power_meter' => 'BOARD_MONITOR',
##&nbsp;            'reg' => 'V3',
##&nbsp;            'module' => 'MGMT'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'low_range' => '2.80',
##&nbsp;            'high_range' => '3.79',
##&nbsp;            'actual_voltage' => '3.26',
##&nbsp;            'status' => 'OK',
##&nbsp;            'expected_voltage' => '3.30',
##&nbsp;            'power_meter' => 'BOARD_MONITOR',
##&nbsp;            'reg' => 'V4',
##&nbsp;            'module' => 'MGMT'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'low_range' => '0.81',
##&nbsp;            'high_range' => '1.10',
##&nbsp;            'actual_voltage' => '0.93',
##&nbsp;            'status' => 'OK',
##&nbsp;            'expected_voltage' => '0.90',
##&nbsp;            'power_meter' => 'BOARD_MONITOR',
##&nbsp;            'reg' => 'V5',
##&nbsp;            'module' => 'MGMT'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'low_range' => '1.02',
##&nbsp;            'high_range' => '1.38',
##&nbsp;            'actual_voltage' => '1.19',
##&nbsp;            'status' => 'OK',
##&nbsp;            'expected_voltage' => '1.20',
##&nbsp;            'power_meter' => 'BOARD_MONITOR',
##&nbsp;            'reg' => 'V6',
##&nbsp;            'module' => 'MGMT'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'low_range' => '10.20',
##&nbsp;            'high_range' => '13.80',
##&nbsp;            'actual_voltage' => '11.67',
##&nbsp;            'status' => 'OK',
##&nbsp;            'expected_voltage' => '12.00',
##&nbsp;            'power_meter' => 'CPU_BOARD_MONITOR',
##&nbsp;            'reg' => 'V1',
##&nbsp;            'module' => 'MGMT'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'low_range' => '2.12',
##&nbsp;            'high_range' => '2.88',
##&nbsp;            'actual_voltage' => '2.47',
##&nbsp;            'status' => 'OK',
##&nbsp;            'expected_voltage' => '2.50',
##&nbsp;            'power_meter' => 'CPU_BOARD_MONITOR',
##&nbsp;            'reg' => 'V2',
##&nbsp;            'module' => 'MGMT'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'low_range' => '2.80',
##&nbsp;            'high_range' => '3.79',
##&nbsp;            'actual_voltage' => '3.28',
##&nbsp;            'status' => 'OK',
##&nbsp;            'expected_voltage' => '3.30',
##&nbsp;            'power_meter' => 'CPU_BOARD_MONITOR',
##&nbsp;            'reg' => 'V3',
##&nbsp;            'module' => 'MGMT'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'low_range' => '2.80',
##&nbsp;            'high_range' => '3.79',
##&nbsp;            'actual_voltage' => '3.26',
##&nbsp;            'status' => 'OK',
##&nbsp;            'expected_voltage' => '3.30',
##&nbsp;            'power_meter' => 'CPU_BOARD_MONITOR',
##&nbsp;            'reg' => 'V4',
##&nbsp;            'module' => 'MGMT'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'low_range' => '1.53',
##&nbsp;            'high_range' => '2.07',
##&nbsp;            'actual_voltage' => '1.79',
##&nbsp;            'status' => 'OK',
##&nbsp;            'expected_voltage' => '1.80',
##&nbsp;            'power_meter' => 'CPU_BOARD_MONITOR',
##&nbsp;            'reg' => 'V5',
##&nbsp;            'module' => 'MGMT'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'low_range' => '1.02',
##&nbsp;            'high_range' => '1.38',
##&nbsp;            'actual_voltage' => '1.23',
##&nbsp;            'status' => 'OK',
##&nbsp;            'expected_voltage' => '1.20',
##&nbsp;            'power_meter' => 'CPU_BOARD_MONITOR',
##&nbsp;            'reg' => 'V6',
##&nbsp;            'module' => 'MGMT'
##&nbsp;          }
##&nbsp;        ];
##&nbsp;
sub show_voltage {
    my %args = @_;

    my ($error_message);

    my $resp        = $args{response_object};
    my $output      = $resp->get_processed_output();
    my $command     = $resp->get_command();
    my $error_level = $resp->get_execution_status();

    my $error = _check_error(
        raw_output    => $output,
        error_message => \$error_message
    );
    if ($error) {
        $error = 'NACL::APISet::Exceptions::' . $error . 'Exception';
        $error->throw( text => $error_message, response_object => $resp );
    }
    my @lines = get_lines( raw_output => $output, response_object => $resp );
    my  @headings = ("module","power_meter","reg","expected_voltage","actual_voltage","status","high_range","low_range");
    my (@result_array);
    splice @lines, 0, 4;
    my $j = 0;
    for ( my $k = 0; $k <= $#lines; $k++ ) {
        my @values = split( /\s+/, $lines[$k] );
        for ( my $i = 0; $i < scalar(@headings); $i++ ) {
            if ( $values[$i] ) {
                $result_array[$j]->{$headings[$i]} = $values[$i];
            }
        }
        $j++;
    }
    return \@result_array;
} ## end show_voltage 

## @name ::show_mac_address_table
##
## @summary Parses the raw output from the "show mac-address-table" command
##
## @description 
##    This method can be used for parsing the raw output obtained after calling 
##    the method from show_mac_address_table with any of its options as described in the 
##    corresponding cdef file 'show.cdefs'
## @synopsis
##      show_mac_address_table(%args);
## INPUTS TO METHOD CALL 'show_mac_address_table' ON APISET OBJECT 
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

##&nbsp;Command = show mac-address-table
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;
##&nbsp;Switch ethernet-default 
##&nbsp;
##&nbsp;Vlan    Mac Address         Type       Interface
##&nbsp;----    -----------         ----       ------------
##&nbsp;1       00:10:94:00:00:01   dynamic    Eth1/9
##&nbsp;1       00:10:94:00:00:02   dynamic    Eth1/9
##&nbsp;1       00:10:94:00:00:03   dynamic    Eth1/9
##&nbsp;1       00:10:94:00:00:04   dynamic    Eth1/9
##&nbsp;Number of unicast:      233
##&nbsp;Number of multicast:    0
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;          {
##&nbsp;            'switch' => 'ethernet-default',
##&nbsp;            'number_of_multicast' => '0',
##&nbsp;            'number_of_unicast' => '233',
##&nbsp;            'switch_info' => [
##&nbsp;                               {
##&nbsp;                                 'mac_address' => '00:10:94:00:00:01',
##&nbsp;                                 'interface' => 'Eth1/9',
##&nbsp;                                 'type' => 'dynamic',
##&nbsp;                                 'vlan' => '1'
##&nbsp;                               },
##&nbsp;                               {
##&nbsp;                                 'mac_address' => '00:10:94:00:00:02',
##&nbsp;                                 'interface' => 'Eth1/9',
##&nbsp;                                 'type' => 'dynamic',
##&nbsp;                                 'vlan' => '1'
##&nbsp;                               },
##&nbsp;                               {
##&nbsp;                                 'mac_address' => '00:10:94:00:00:03',
##&nbsp;                                 'interface' => 'Eth1/9',
##&nbsp;                                 'type' => 'dynamic',
##&nbsp;                                 'vlan' => '1'
##&nbsp;                               },
##&nbsp;                               {
##&nbsp;                                 'mac_address' => '54:7f:ee:fe:d4:a2',
##&nbsp;                                 'interface' => 'Eth1/2',
##&nbsp;                                 'type' => 'dynamic',
##&nbsp;                                 'vlan' => '1'
##&nbsp;                               }
##&nbsp;                             ]
##&nbsp;          }
##&nbsp;        ];
##&nbsp;
sub show_mac_address_table {
    my %args = @_;

    my ($error_message);

    my $resp   = $args{response_object};
    my $output = $resp->get_processed_output();
    my $error  = _check_error(
        raw_output    => $output,
        error_message => \$error_message
    );
    if ($error) {
        $error = 'NACL::APISet::Exceptions::' . $error . 'Exception';
        $error->throw( text => $error_message, response_object => $resp );
    }
    my @lines = get_lines( raw_output => $output, response_object => $resp );
    my @result_array ;
    my  @headings = ("vlan","mac_address","type","interface");
    my $j = 0;

    my $row = shift(@lines);
    my @array = split(/\s+/,$row);
    $result_array[0]->{to_key_name(item => $array[0])} = $array[1] ;

    for ( my $k = 1; $k <= $#lines; $k++ ) {
        next if($lines[$k] =~ /----+/) ;
        if($lines[$k] =~ /Number of/) {
            my @arr = split(/:/,$lines[$k]);
            $arr[1] =~ s/^\s+//;
            $result_array[0]->{ to_key_name(item => $arr[0])} = $arr[1];
            $j--;
        }else {
            my @values = split( /\s\s+/, $lines[$k] );
            for ( my $i = 0; $i < scalar(@headings); $i++ ) {
                if ( $values[$i] ) {
                    $result_array[0]->{switch_info}->[$j]->{ to_key_name(item => $headings[$i]) } = $values[$i];
                }
            }
         }
        $j++;
    }
    return \@result_array;
} ## end sub show_mac_addr_table

## @name ::show_vlan
##
## @summary Parses the raw output from the "show vlan" command
##
## @description 
##    This method can be used for parsing the raw output obtained after calling 
##    the method from show_vlan with any of its options as described in the 
##    corresponding cdef file 'show.cdefs'
## @synopsis
##      show_vlan(%args);
## INPUTS TO METHOD CALL 'show_vlan' ON APISET OBJECT 
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

##&nbsp;Command = show vlan
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;
##&nbsp;VLAN    Name               	Ports
##&nbsp;----    -----------        	--------------------------------------
##&nbsp;1       default         	Eth1/1, Eth1/2, Eth1/3, Eth1/4, Eth1/5,
##&nbsp;				Eth1/6, Eth1/7, Eth1/8, Eth1/9, Eth1/10,
##&nbsp;				Eth1/11, Eth1/12
##&nbsp;100                     	
##&nbsp;888                     	
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;          {
##&nbsp;            'ports' => [
##&nbsp;                         'Eth1/1',
##&nbsp;                         'Eth1/2',
##&nbsp;                         'Eth1/3',
##&nbsp;                         'Eth1/4',
##&nbsp;                         'Eth1/5,',
##&nbsp;                         'Eth1/6',
##&nbsp;                         'Eth1/7',
##&nbsp;                         'Eth1/8',
##&nbsp;                         'Eth1/9',
##&nbsp;                         'Eth1/10,',
##&nbsp;                         'Eth1/11',
##&nbsp;                         'Eth1/12'
##&nbsp;                       ],
##&nbsp;            'name' => 'default',
##&nbsp;            'id' => '1'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'ports' => [],
##&nbsp;            'name' => undef,
##&nbsp;            'id' => '100'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'ports' => [],
##&nbsp;            'name' => undef,
##&nbsp;            'id' => '888'
##&nbsp;          }
##&nbsp;        ];
##&nbsp;
sub show_vlan {
    my %args = @_;
    my ($error_message);
    my $resp   = $args{response_object};
    my $output = $resp->get_processed_output();
    my $error  = _check_error(
        raw_output    => $output,
        error_message => \$error_message
    );
    if ($error) {
        $error = 'NACL::APISet::Exceptions::' . $error . 'Exception';
        $error->throw( text => $error_message, response_object => $resp );
    }
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
    my (@result_array ,$line);
    my @lines = get_lines( raw_output => $output );
    @lines = grep (!/^--/, @lines);
    # Get the headings
    my $line = shift(@lines);
    my @headings = split(/\s+/, $line);
    my ($vlan_count , @vlan_info);
    $vlan_count = -1;
    $line       = shift(@lines);
    while ($line !~ /VLAN/i and $line) {
        print "line is $line" ;
        @vlan_info = split(/\s+/, $line);
        if ($line =~ /^\d+/) {
            $vlan_count++;
            $result_array[$vlan_count]->{'id'}     = shift(@vlan_info);
            $result_array[$vlan_count]->{'name'}   = shift(@vlan_info);
            $line = join(" ", @vlan_info);
        }
        my @ports = split(/\,\s+/, $line);
        push(@{$result_array[$vlan_count]->{'ports'}}, @ports);
        $line = shift(@lines);
    }
   return \@result_array;
}  ## end sub show_vlan

## @name ::show_interface_ethernet
##
## @summary Parses the raw output from the "show interface ethernet" command
##
## @description 
##    This method can be used for parsing the raw output obtained after calling 
##    the method from show_interface_ethernet with any of its options as described in the 
##    corresponding cdef file 'show.cdefs'
## @synopsis
##      show_interface_ethernet(%args);
## INPUTS TO METHOD CALL 'show_interface_ethernet' ON APISET OBJECT 
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## <tr> <td> No switch for this key </td> <td> 'port' </td> <td> '0' </td> </tr> 
## <tr> <td> No switch for this key </td> <td> 'field' </td> <td> '0' </td> </tr> 
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

##&nbsp;Command = show interface ethernet 1/9 status
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;
##&nbsp;Port                   Operational state           Speed                  Negotiation              
##&nbsp;----                   -----------------           -----                  -----------              
##&nbsp;Eth1/9                 Up                          40 Gbps             No-Negotiation           
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;          {
##&nbsp;            'negotiation' => 'Gbps',
##&nbsp;            'speed' => '40',
##&nbsp;            'port' => 'Eth1/9',
##&nbsp;            'operational_state' => 'Up'
##&nbsp;          }
##&nbsp;        ];
##&nbsp;
sub show_interface_ethernet {
    my %args = @_;
    my ($error_message);
    my $resp   = $args{response_object};
    my $output = $resp->get_processed_output();
    my $error  = _check_error(
        raw_output    => $output,
        error_message => \$error_message
    );
    if ($error) {
        $error = 'NACL::APISet::Exceptions::' . $error . 'Exception';
        $error->throw( text => $error_message, response_object => $resp );
    }
    my (@result_array ,$line);
    my @lines = get_lines( raw_output => $output );
    @lines = grep (!/^--/, @lines);
    # Get the headings
    my @headings = split(/\s\s+/, $lines[0]);
    my @values    = split(/\s+/,$lines[1]);

    for( my $i=0; $i<scalar(@headings); $i++) {
         $result_array[0]->{to_key_name( item => $headings[$i])} = $values[$i];
    }
    return \@result_array;
}  ## end sub show_interface_ethernet
1;
