#
#
#Copyright (c)2012-2013 NetApp, Inc.
#All rights reserved.
#
## @summary Routines to parse NetAppNetwork show and its subcommands
##
## @description
##    This is a library to parse the raw output of NetAppNetwork switch show and its subcommands.
##
## @synopsis
##  use NACL::APISet::Switch::CLI::NetAppNetwork::CN1610::show
##
##  show_version(%args)
##  show_bootvar(%args)
##  show_port(%args)
##  show_running_config(%args)
##  show_mac_addr_table(%args)
##  show_mac_addr_table_interface(%args)
##  show_interface(%args)
##  show_interface_ethernet(%args)
##  show_storm_control(%args)
##  show_process_cpu(%args)
##  show_isdp()
##  show_isdp_neighbors(%args)
##  show_tech_support
##  show_vlan
##  show_snmpcommunity
##
## @author skotagir@netapp.com,dl-nacl-dev@netapp.com
##
## @status public
#
package NACL::APISet::Switch::CLI::NetAppNetwork::CN1610::show;

use strict;
use NACL::APISet::Response::CLI::ParserUtils;

#Include exceptions which will be thrown
use NACL::APISet::Exceptions::MissingRequiredParamException qw(:try);
use NACL::APISet::Exceptions::InvalidParamException;
use NACL::APISet::Exceptions::InvalidParamValueException;

sub _check_error {
    my %args = @_;

    my $raw_output    = $args{raw_output};
    my $command       = $args{command};
    my $error_message = $args{error_message};
    my $error;

    # Check for the error
    if ( $raw_output =~ /Invalid\s+input\s+detected/i ) {
        $$error_message = $raw_output;
        $error          = "InvalidParam";
    } elsif ( $raw_output =~ /(Incomplete\s+command)/i ) {
        $$error_message = $raw_output;
        $error          = "MissingRequiredParam";
    } elsif ( $raw_output
        =~ /An\s+invalid\s+interface\s+has\s+been\s+used\s+for\s+this\s+command|The\s+specified\s+interface\s+does\s+not\s+exist./ )
    {
        $$error_message = $raw_output;
        $error          = "InvalidParamValue";
    }

    return $error;
} ## end sub _check_error

## @name ::show_version
##
## @summary Parses the raw output from the "show version" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from show_version with any of its options as described in the
##    corresponding cdef file 'show.cdefs'
## @synopsis
##      show_version(%args);
## INPUTS TO METHOD CALL 'show_version' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

##&nbsp;Command = show version
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;
##&nbsp;Switch: 1
##&nbsp;
##&nbsp;System Description............................. NetApp CN1610, 1.0.0.0, Linux
##&nbsp;                                                2.6.21.7
##&nbsp;Machine Type................................... NetApp CN1610
##&nbsp;Machine Model.................................. CN1610
##&nbsp;Serial Number.................................. 11111200069
##&nbsp;FRU Number..................................... 111-00893
##&nbsp;Part Number.................................... BCM56820
##&nbsp;Maintenance Level.............................. A
##&nbsp;Manufacturer................................... 0xbc00
##&nbsp;Burned In MAC Address.......................... 00:A0:98:4B:B7:F1
##&nbsp;Software Version............................... 1.0.0.0
##&nbsp;Operating System............................... Linux 2.6.21.7
##&nbsp;Network Processing Device...................... BCM56820_B0
##&nbsp;
##&nbsp;--More-- or (q)uit
##&nbsp;
##&nbsp;
##&nbsp;Additional Packages............................ FASTPATH QOS
##&nbsp;                                                FASTPATH IPv6 Management
##&nbsp;
##&nbsp;
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;          {
##&nbsp;            'serial_number' => '11111200069',
##&nbsp;            'maintenance_level' => 'A',
##&nbsp;            'operating_system' => 'Linux 2.6.21.7',
##&nbsp;            'network_processing_device' => 'BCM56820_B0',
##&nbsp;            'machine_type' => 'NetApp CN1610',
##&nbsp;            'system_description' => 'NetApp CN1610, 1.0.0.0, Linux 2.6.21.7',
##&nbsp;            'part_number' => 'BCM56820',
##&nbsp;            'software_version' => '1.0.0.0',
##&nbsp;            'burned_in_mac_address' => '00:A0:98:4B:B7:F1',
##&nbsp;            'manufacturer' => '0xbc00',
##&nbsp;            'additional_packages' => [
##&nbsp;                                       'FASTPATH QOS',
##&nbsp;                                       'FASTPATH IPv6 Management'
##&nbsp;                                     ],
##&nbsp;            'fru_number' => '111-00893',
##&nbsp;            'machine_model' => 'CN1610'
##&nbsp;          }
##&nbsp;        ];
##&nbsp;
sub show_version {
    my %args = @_;

    my ($error_message);

    my $resp        = $args{response_object};
    my $output      = $resp->get_processed_output();
    my $command     = $resp->get_command();
    my $error_level = $resp->get_execution_status();
    my ( $error, $key );

    # Status returned by switch for this command is
    # is ZERO even if command failed
    if ( not defined($error) ) {
        $error = _check_error(
            raw_output    => $output,
            command       => "show version",
            error_message => \$error_message
        );
    }
    if ($error) {
        $error = 'NACL::APISet::Exceptions::' . $error . 'Exception';
        $error->throw( text => $error_message, response_object => $resp );
    }

    my @result_array;
    my @lines = get_lines( raw_output => $output, first_line => 'Switch:\s+\d+' );
    shift(@lines) if ( $lines[0] =~ /Switch:\s+\d+/ );
    while ( my $line = shift(@lines) ) {
        next if ( $line =~ /--More--\s+or\s+\(q\)uit/ );
        my %hash;
        while ( $line !~ /Switch:\s+\d+/ and ( scalar(@lines) > 0 ) ) {
            if ( $line =~ /^Additional Packages/ ) {
                my @tmp_arr;
                my ( $key, $value ) = ( $line =~ /^(.*\w)\.\.+\s*(.*)/ );
                push( @tmp_arr, $value );
                while ( $lines[0] !~ /^(.*)\.\.+\s*(.*)/
                    and ( scalar(@lines) > 0 ) )
                {
                    $line = shift(@lines);
                    next if ( $line =~ /--More--\s+or\s+\(q\)uit/ );
                    push( @tmp_arr, $line );
                }
                $hash{ to_key_name( item => $key ) } = \@tmp_arr;
            } else {
                my ( $key, $value ) = ( $line =~ /^(.*\w)\.\.+\s*(.*)/ );
                $hash{ to_key_name( item => $key ) } = $value;
                while ( $lines[0] !~ /^(.*)\.\.+\s*(.*)/
                    and ( scalar(@lines) > 0 ) )
                {
                    $line = shift(@lines);
                    next if ( $line =~ /--More--\s+or\s+\(q\)uit/ );
                    $hash{ to_key_name( item => $key ) } .= " " . $line;
                }
            }
            $line = shift(@lines);
        }
        push( @result_array, \%hash );
    }

    return \@result_array;

} ## end sub show_version

## @name ::show_bootvar
##
## @summary Parses the raw output from the "show bootvar" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from show_bootvar with any of its options as described in the
##    corresponding cdef file 'show.cdefs'
## @synopsis
##      show_bootvar(%args);
## INPUTS TO METHOD CALL 'show_bootvar' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

##&nbsp;Command = show bootvar
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;
##&nbsp;Image Descriptions
##&nbsp;
##&nbsp; active :
##&nbsp; backup :
##&nbsp;
##&nbsp;
##&nbsp; Images currently available on Flash
##&nbsp;
##&nbsp;--------------------------------------------------------------------
##&nbsp; unit      active      backup     current-active        next-active
##&nbsp;--------------------------------------------------------------------
##&nbsp;
##&nbsp;    1     1.0.0.2     1.0.0.0            1.0.0.2            1.0.0.2
##&nbsp;
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;          {
##&nbsp;            'image_descriptions' => [
##&nbsp;                                      {
##&nbsp;                                        'active' => '',
##&nbsp;                                        'backup' => ''
##&nbsp;                                      }
##&nbsp;                                    ],
##&nbsp;            'images_currently_available_on_flash' => [
##&nbsp;                                                       {
##&nbsp;                                                         'unit' => '1',
##&nbsp;                                                         'active' => '1.0.0.2',
##&nbsp;                                                         'next_active' => '1.0.0.2',
##&nbsp;                                                         'current_active' => '1.0.0.2',
##&nbsp;                                                         'backup' => '1.0.0.0'
##&nbsp;                                                       }
##&nbsp;                                                     ]
##&nbsp;          }
##&nbsp;        ];
##&nbsp;
sub show_bootvar {
    my %args = @_;

    my ($error_message);

    my $resp        = $args{response_object};
    my $output      = $resp->get_processed_output();
    my $command     = $resp->get_command();
    my $error_level = $resp->get_execution_status();
    my ( @values, @key_names, $key, $k );

    my $error = _check_error(
        raw_output    => $output,
        error_message => \$error_message
    );
    if ($error) {
        $error = 'NACL::APISet::Exceptions::' . $error . 'Exception';
        $error->throw( text => $error_message, response_object => $resp );
    }
    my @result_array;
    my @lines = get_lines( raw_output => $output, response_object => $resp );
    foreach my $line (@lines) {
        if ( $line =~ /(Image\s+Descriptions|Images\s+currently\s+available\s+on\s+Flash)/ ) {
            $key = to_key_name( item => $1 );
        } elsif ( $line =~ /(active|backup)\s+:/ ) {
            my $key1 = $1;
            $line =~ s/$key1\s+://;
            trim( 'string_ptr' => \$line );
            $result_array[0]->{$key}->[0]->{$key1} = $line;
        } elsif ( $line =~ /unit\s+active\s+backup\s+current-active\s+next-active/ ) {
            @key_names = split( /\s+/, $line );
        } elsif ( $line =~ /.*/ && $line !~ /-----/ ) {
            @values = split( /\s+/, $line );
            foreach my $key_name (@key_names) {
                $key_name =~ s/-/_/;
                $result_array[$k]->{$key}->[0]->{ to_key_name( item => "$key_name" ) } = $values[0];
                shift @values;
            }
            $k++;
        }
    }
    return \@result_array;
} ## end sub show_bootvar

## @name ::show_port
##
## @summary Parses the raw output from the "show port" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from show_port with any of its options as described in the
##    corresponding cdef file 'show.cdefs'
## @synopsis
##      show_port(%args);
## INPUTS TO METHOD CALL 'show_port' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## <tr> <td> No switch for this key </td> <td> 'interface' </td> <td> '0' </td> </tr>
## <tr> <td> No switch for this key </td> <td> 'port' </td> <td> '0' </td> </tr>
## <tr> <td> 'description' </td> <td> 'description' </td> <td> '1' </td> </tr>
## <tr> <td> 'all' </td> <td> 'all' </td> <td> '1' </td> </tr>
## <tr> <td> No switch for this key </td> <td> 'group-id' </td> <td> '0' </td> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

##&nbsp;Command = show port all
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;
##&nbsp;                  Admin    Physical   Physical   Link   Link    LACP    Flow
##&nbsp; Intf      Type   Mode     Mode       Status     Status Trap    Mode    Mode
##&nbsp;--------- ------ --------- ---------- ---------- ------ ------- ------ -------
##&nbsp;0/1              Enable    10G Full              Down   Enable  Enable Disable
##&nbsp;0/2              Enable    10G Full              Down   Enable  Enable Disable
##&nbsp;0/3              Enable    10G Full   10G Full   Up     Enable  Enable Disable
##&nbsp;0/4              Enable    10G Full   10G Full   Up     Enable  Enable Disable
##&nbsp;
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;          {
##&nbsp;            'link_trap' => 'Enable',
##&nbsp;            'flow_mode' => 'Disable',
##&nbsp;            'lacp_mode' => 'Enable',
##&nbsp;            'physical_mode' => '10G Full',
##&nbsp;            'interface' => '0/1',
##&nbsp;            'physical_status' => '',
##&nbsp;            'type' => '',
##&nbsp;            'link_status' => 'Down',
##&nbsp;            'admin_mode' => 'Enable'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'link_trap' => 'Enable',
##&nbsp;            'flow_mode' => 'Disable',
##&nbsp;            'lacp_mode' => 'Enable',
##&nbsp;            'physical_mode' => '10G Full',
##&nbsp;            'interface' => '0/2',
##&nbsp;            'physical_status' => '',
##&nbsp;            'type' => '',
##&nbsp;            'link_status' => 'Down',
##&nbsp;            'admin_mode' => 'Enable'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'link_trap' => 'Enable',
##&nbsp;            'flow_mode' => 'Disable',
##&nbsp;            'lacp_mode' => 'Enable',
##&nbsp;            'physical_mode' => '10G Full',
##&nbsp;            'interface' => '0/3',
##&nbsp;            'physical_status' => '10G Full',
##&nbsp;            'type' => '',
##&nbsp;            'link_status' => 'Up',
##&nbsp;            'admin_mode' => 'Enable'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'link_trap' => 'Enable',
##&nbsp;            'flow_mode' => 'Disable',
##&nbsp;            'lacp_mode' => 'Enable',
##&nbsp;            'physical_mode' => '10G Full',
##&nbsp;            'interface' => '0/4',
##&nbsp;            'physical_status' => '10G Full',
##&nbsp;            'type' => '',
##&nbsp;            'link_status' => 'Up',
##&nbsp;            'admin_mode' => 'Enable'
##&nbsp;          }
##&nbsp;        ];
##&nbsp;

sub show_port {
    my %args = @_;

    my ($error_message);
    my $resp        = $args{response_object};
    my $output      = $resp->get_processed_output();
    my $command     = $resp->get_command();
    my $error_level = $resp->get_execution_status();

    my $error = _check_error(
        raw_output    => $output,
        error_message => \$error_message
    );
    if ($error) {
        $error = 'NACL::APISet::Exceptions::' . $error . 'Exception';
        $error->throw( text => $error_message, response_object => $resp );
    }
    my @lines = get_lines( raw_output => $output, response_object => $resp );
    my ( $pos1, $j, $k ) = ( 0, 0, 0 );
    my ( @result_array, @headings, $pos2, $length, $value, $line, $string );
    my $option = $command;
    $option =~ s/show port//;

    if ( $option =~ /description/ ) {
        foreach my $line (@lines) {
            if ( $line =~ /(.*)\.\s+(.*)/ ) {
                my $decs_key = $1;
                my $desc_val = $2;
                $decs_key =~ s/[.]/ /g;
                trim( 'string_ptr' => \$decs_key );
                $result_array[0]->{ to_key_name( item => "$decs_key" ) } = $desc_val;
            }
        }
    } else {
        foreach my $line (@lines) {
            if ( $line =~ /(Intf\s+Type\s+Mode\s+.*\s+Mode)|(Admin\s+Physical\s+.*\s+Flow)/ ) {
                @headings = (
                    "interface", "type",      "admin_mode", "physical_mode", "physical_status", "link_status",
                    "link_trap", "lacp_mode", "flow_mode"
                );
            } elsif ( $line =~ /-------/ ) {
                $string = "$line" . " ";
            } elsif ( $line =~ /(\d+\/\d+)\s+.*/ ) {
                while ( $string =~ /\s+/g ) {
                    $pos2 = pos($string);
                    my $len = $pos2 - $pos1;
                    my $val = substr( $line, $pos1, $len );
                    $pos1 = $pos2;
                    trim( 'string_ptr' => \$val );
                    $result_array[$k]->{ $headings[ $j++ ] } = $val;
                }
                $pos1 = 0;
                $j    = 0;
                $k++;
            }
        }
    }
    return \@result_array;
} ## end sub show_port

## @name ::show_running_config
##
## @summary Parses the raw output from the "show running-config" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from show_running_config with any of its options as described in the
##    corresponding cdef file 'show.cdefs'
## @synopsis
##      show_running_config(%args);
## INPUTS TO METHOD CALL 'show_running_config' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

##&nbsp;Command = show running-config
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;!Current Configuration:
##&nbsp;!
##&nbsp;!System Description "NetApp CN1610, 1.0.0.2, Linux 2.6.21.7"
##&nbsp;!System Software Version "1.0.0.2"
##&nbsp;!System Up Time          "1 days 18 hrs 1 mins 22 secs"
##&nbsp;!Additional Packages     FASTPATH QOS
##&nbsp;!Current SNTP Synchronized Time: Mar 15 04:22:55 201 UTC
##&nbsp;!
##&nbsp;set prompt "CS:net-1960-1"
##&nbsp;telnetcon timeout 60
##&nbsp;enable password 54efa5bc56b147ac920e6a5cbba5dfe76bd81744dd12ffe9141907f7a6f87a9db9fa55b14573ebe19c8654d0b98a1c393c5635b5ced4d40d1e16f5cdd6e3a991 encrypted
##&nbsp;serviceport protocol none
##&nbsp;serviceport ip 172.22.200.93 255.255.254.0 172.22.200.1
##&nbsp;vlan database
##&nbsp;exit
##&nbsp;ip ssh server enable
##&nbsp;sshcon timeout 60
##&nbsp;configure
##&nbsp;sntp client mode unicast
##&nbsp;! sntp server status is Success
##&nbsp;sntp server "172.19.2.16"
##&nbsp;ip domain name "lab.netapp.com"
##&nbsp;ip name server 172.22.201.46
##&nbsp;username "admin" password 54efa5bc56b147ac920e6a5cbba5dfe76bd81744dd12ffe9141907f7a6f87a9db9fa55b14573ebe19c8654d0b98a1c393c5635b5ced4d40d1e16f5cdd6e3a991 level 15 encrypted
##&nbsp;
##&nbsp;--More-- or (q)uit
##&nbsp;
##&nbsp;line console
##&nbsp;serial timeout 60
##&nbsp;exit
##&nbsp;line telnet
##&nbsp;exit
##&nbsp;line ssh
##&nbsp;password 54efa5bc56b147ac920e6a5cbba5dfe76bd81744dd12ffe9141907f7a6f87a9db9fa55b14573ebe19c8654d0b98a1c393c5635b5ced4d40d1e16f5cdd6e3a991 encrypted
##&nbsp;exit
##&nbsp;spanning-tree configuration name "00-A0-98-4B-B7-F1"
##&nbsp;spanning-tree bpduguard
##&nbsp;port-channel name 3/1 ISL-LAG
##&nbsp;interface 0/13
##&nbsp;no auto-negotiate
##&nbsp;addport 3/1
##&nbsp;exit
##&nbsp;interface 0/14
##&nbsp;no auto-negotiate
##&nbsp;addport 3/1
##&nbsp;exit
##&nbsp;interface 0/15
##&nbsp;no auto-negotiate
##&nbsp;addport 3/1
##&nbsp;exit
##&nbsp;interface 0/16
##&nbsp;
##&nbsp;--More-- or (q)uit
##&nbsp;
##&nbsp;no auto-negotiate
##&nbsp;addport 3/1
##&nbsp;exit
##&nbsp;snmp-server sysname "CS:net-1960-1"
##&nbsp;!
##&nbsp;snmp-server community "netapp"
##&nbsp;interface 0/1
##&nbsp;no auto-negotiate
##&nbsp;description 'Cluster Node 1'
##&nbsp;spanning-tree edgeport
##&nbsp;mtu 9216
##&nbsp;exit
##&nbsp;interface 0/2
##&nbsp;no auto-negotiate
##&nbsp;description 'Cluster Node 2'
##&nbsp;spanning-tree edgeport
##&nbsp;mtu 9216
##&nbsp;exit
##&nbsp;interface 0/15
##&nbsp;description 'ISL member port 15'
##&nbsp;mtu 9216
##&nbsp;exit
##&nbsp;interface 0/16
##&nbsp;description 'ISL member port 16'
##&nbsp;mtu 9216
##&nbsp;exit
##&nbsp;interface 3/1
##&nbsp;description 'ISL LAG - ports 13-16'
##&nbsp;port-channel load-balance 7
##&nbsp;mtu 9216
##&nbsp;exit
##&nbsp;interface 3/64
##&nbsp;description 'NetApp CN1610 Cluster Switch RCF v1'
##&nbsp;exit
##&nbsp;
##&nbsp;--More-- or (q)uit
##&nbsp;
##&nbsp;isdp timer 5
##&nbsp;exit
##&nbsp;
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;          {
##&nbsp;            'ip_ssh_server' => 'enable',
##&nbsp;            'spanning_tree_configuration' => '00-A0-98-4B-B7-F1',
##&nbsp;            'snmp_server_community' => 'netapp',
##&nbsp;            'snmp_server_sysname' => 'CS:net-1960-1',
##&nbsp;            'lines' => [
##&nbsp;                         {
##&nbsp;                           'serial_timeout' => '60',
##&nbsp;                           'line' => 'console'
##&nbsp;                         },
##&nbsp;                         {
##&nbsp;                           'line' => 'telnet'
##&nbsp;                         },
##&nbsp;                         {
##&nbsp;                           'password' => '54efa5bc56b147ac920e6a5cbba5dfe76bd81744dd12ffe9141907f7a6f87a9db9fa55b14573ebe19c8654d0b98a1c393c5635b5ced4d40d1e16f5cdd6e3a991 encrypted',
##&nbsp;                           'line' => 'ssh'
##&nbsp;                         }
##&nbsp;                       ],
##&nbsp;            'current_sntp_synchronized_time' => 'Mar 15 04:22:55 201 UTC',
##&nbsp;            'password' => '54efa5bc56b147ac920e6a5cbba5dfe76bd81744dd12ffe9141907f7a6f87a9db9fa55b14573ebe19c8654d0b98a1c393c5635b5ced4d40d1e16f5cdd6e3a991',
##&nbsp;            'interfaces' => [
##&nbsp;                              {
##&nbsp;                                'interface' => '0/13',
##&nbsp;                                'running_config' => [
##&nbsp;                                                      'no auto-negotiate',
##&nbsp;                                                      'addport 3/1',
##&nbsp;                                                      'exit'
##&nbsp;                                                    ]
##&nbsp;                              },
##&nbsp;                              {
##&nbsp;                                'interface' => '0/14',
##&nbsp;                                'running_config' => [
##&nbsp;                                                      'no auto-negotiate',
##&nbsp;                                                      'addport 3/1',
##&nbsp;                                                      'exit'
##&nbsp;                                                    ]
##&nbsp;                              },
##&nbsp;                              {
##&nbsp;                                'interface' => '0/15',
##&nbsp;                                'running_config' => [
##&nbsp;                                                      'no auto-negotiate',
##&nbsp;                                                      'addport 3/1',
##&nbsp;                                                      'exit'
##&nbsp;                                                    ]
##&nbsp;                              },
##&nbsp;                              {
##&nbsp;                                'interface' => '0/16',
##&nbsp;                                'running_config' => [
##&nbsp;                                                      'no auto-negotiate',
##&nbsp;                                                      'addport 3/1',
##&nbsp;                                                      'exit'
##&nbsp;                                                    ]
##&nbsp;                              },
##&nbsp;                              {
##&nbsp;                                'interface' => '0/1',
##&nbsp;                                'running_config' => [
##&nbsp;                                                      'no auto-negotiate',
##&nbsp;                                                      'description \'Cluster Node 1\'',
##&nbsp;                                                      'spanning-tree edgeport',
##&nbsp;                                                      'mtu 9216',
##&nbsp;                                                      'exit'
##&nbsp;                                                    ]
##&nbsp;                              },
##&nbsp;                              {
##&nbsp;                                'interface' => '0/2',
##&nbsp;                                'running_config' => [
##&nbsp;                                                      'no auto-negotiate',
##&nbsp;                                                      'description \'Cluster Node 2\'',
##&nbsp;                                                      'spanning-tree edgeport',
##&nbsp;                                                      'mtu 9216',
##&nbsp;                                                      'exit'
##&nbsp;                                                    ]
##&nbsp;                              },
##&nbsp;                              {
##&nbsp;                                'interface' => '0/15',
##&nbsp;                                'running_config' => [
##&nbsp;                                                      'description \'ISL member port 15\'',
##&nbsp;                                                      'mtu 9216',
##&nbsp;                                                      'exit'
##&nbsp;                                                    ]
##&nbsp;                              },
##&nbsp;                              {
##&nbsp;                                'interface' => '0/16',
##&nbsp;                                'running_config' => [
##&nbsp;                                                      'description \'ISL member port 16\'',
##&nbsp;                                                      'mtu 9216',
##&nbsp;                                                      'exit'
##&nbsp;                                                    ]
##&nbsp;                              },
##&nbsp;                              {
##&nbsp;                                'interface' => '3/1',
##&nbsp;                                'running_config' => [
##&nbsp;                                                      'description \'ISL LAG - ports 13-16\'',
##&nbsp;                                                      'port-channel load-balance 7',
##&nbsp;                                                      'mtu 9216',
##&nbsp;                                                      'exit'
##&nbsp;                                                    ]
##&nbsp;                              },
##&nbsp;                              {
##&nbsp;                                'interface' => '3/64',
##&nbsp;                                'running_config' => [
##&nbsp;                                                      'description \'NetApp CN1610 Cluster Switch RCF v1\'',
##&nbsp;                                                      'exit'
##&nbsp;                                                    ]
##&nbsp;                              }
##&nbsp;                            ],
##&nbsp;            'sntp_server' => '172.19.2.16',
##&nbsp;            'additional_packages' => [
##&nbsp;                                       'FASTPATH',
##&nbsp;                                       'QOS'
##&nbsp;                                     ],
##&nbsp;            'sshcon_timeout' => '60',
##&nbsp;            'port_channel' => '3/1 ISL-LAG',
##&nbsp;            'system_up_time' => '1 days 18 hrs 1 mins 22 secs',
##&nbsp;            'serviceport_protocol' => 'none',
##&nbsp;            'set_prompt' => 'CS:net-1960-1',
##&nbsp;            'enable_password' => '54efa5bc56b147ac920e6a5cbba5dfe76bd81744dd12ffe9141907f7a6f87a9db9fa55b14573ebe19c8654d0b98a1c393c5635b5ced4d40d1e16f5cdd6e3a991',
##&nbsp;            'serviceport_ip' => [
##&nbsp;                                  '172.22.200.93',
##&nbsp;                                  '255.255.254.0',
##&nbsp;                                  '172.22.200.1'
##&nbsp;                                ],
##&nbsp;            'username' => 'admin',
##&nbsp;            'system_description' => 'NetApp CN1610, 1.0.0.2, Linux 2.6.21.7',
##&nbsp;            'system_software_version' => '1.0.0.2',
##&nbsp;            'sntp_client_mode' => 'unicast',
##&nbsp;            'level' => '15',
##&nbsp;            'ip_name_server' => '172.22.201.46',
##&nbsp;            'sntp_server_status' => 'Success',
##&nbsp;            'ip_domain_name' => 'lab.netapp.com',
##&nbsp;            'isdp_timer' => '5',
##&nbsp;            'telnetcon_timeout' => '60',
##&nbsp;            'spanning_tree' => 'bpduguard'
##&nbsp;          }
##&nbsp;        ];
##&nbsp;
sub show_running_config {
    my %args = @_;

    my ( @result_array, $error_message );

    my $resp        = $args{response_object};
    my $output      = $resp->get_processed_output();
    my $command     = $resp->get_command();
    my $error_level = $resp->get_execution_status();

    my $error = _check_error(
        raw_output    => $output,
        command       => $command,
        error_message => \$error_message
    );

    if ($error) {
        $error = 'NACL::APISet::Exceptions::' . $error . 'Exception';
        $error->throw( text => $error_message, response_object => $resp );
    }

    my $int = -1;
    my $key;
    my $j     = 0;
    my $lines = -1;
    my $con   = 0;
    my @lines;

    #To handle the show tech-support command
    if ( $output =~ /Show Running Config/ ) {
        @lines = get_lines( raw_output => $output, response_object => $resp, first_line => "Show Running Config" );
    } else {
        @lines = get_lines( raw_output => $output, response_object => $resp );
    }
    if ( $output =~ /description 'NetApp CN1610 Cluster Switch RCF\s+(\S+).*/i ) {
        $result_array[0]->{'rcf_version'} = $1;
    }
    while ( my $line = shift(@lines) ) {
        next if ( $line =~ /--More--\s+or\s+\(q\)uit/ );
        if ( $line =~ /interface\s+(.*)/ ) {
            $key                                              = "interfaces";
            $result_array[0]->{$key}->[ ++$int ]->{interface} = $1;
            $con                                              = 0;
        } elsif ( $key eq "interfaces" ) {
            $result_array[0]->{interfaces}->[$int]->{'running_config'}->[ $con++ ] = $line;
            $key = undef if ( $line eq "exit" );
        } elsif ( $line =~ /(serial\s+timeout|password)\s+(.*)/
            && $key eq "lines" )
        {
            my $key3 = $1;
            my $val3 = $2;
            $key3 =~ s/-/_/g;
            $result_array[0]->{$key}->[$lines]->{ to_key_name( item => "$key3" ) } = $val3;
        } elsif ( $line =~ /line\s+(\w+)/ ) {
            $key = "lines";
            $result_array[0]->{$key}->[ ++$lines ]->{line} = $1;
        } elsif ( $line =~ /(System\s+Up\s+Time|System\s+Description|Current\s+SNTP\s+Synchronized\s+Time:)\s+"?(.*)"?/ ) {
            my $value2 = $2;
            my $key2   = $1;
            $value2 =~ s/"//g;
            $result_array[0]->{ to_key_name( item => "$key2" ) } = $value2;
        } elsif ( $line =~ /(sntp\s+server\s+status)\s+is\s+(.*)/ ) {
            $result_array[0]->{ to_key_name( item => "$1" ) } = $2;
        } elsif ( $line =~ /((spanning-tree\s+configuration|port-channel)\s+name)\s*"?(.*)"?/ ) {
            my $val4 = $3;
            my $key4 = $2;
            $val4 =~ s/"//g;
            $key4 =~ s/-/_/g;
            $result_array[0]->{ to_key_name( item => "$key4" ) } = $val4;
        } elsif ( $line =~ /^!?((ip|snmp-server|System|sntp|serviceport\s+protocol|sshcon|telnetcon|set\s+prompt|hostname)\s*.*)\s+(.*)/ ) {
            my $value1 = $3;
            my $key1   = $1;
            $key1   =~ s/-/_/;
            $value1 =~ s/"//g;
            $result_array[0]->{ to_key_name( item => "$key1" ) } = $value1;
        } elsif ( $line =~ /(Additional\s+Packages|serviceport\s+ip|network parms)\s+(.*)/ ) {
            my @entries = split( /\s+/, $2 );
            $result_array[0]->{ to_key_name( item => $1 ) } = \@entries;
        } elsif ( $line =~ /(isdp\s+timer|spanning-tree)\s+(.*)/ ) {
            my $key5 = $1;
            my $val5 = $2;
            $key5 =~ s/-/_/g;
            $result_array[0]->{ to_key_name( item => "$key5" ) } = $val5;
        } elsif ( $line =~ /(username)\s+"(\w+)"\s+(password)\s+(.*)\s+(level)\s+(.*)\s+encrypted/ ) {
            $result_array[0]->{"$1"} = $2;
            $result_array[0]->{"$3"} = $4;
            $result_array[0]->{"$5"} = $6;
        } elsif ( $line =~ /((enable)?\s+password)\s+(.*)\s+encrypted/ ) {
            $result_array[0]->{ to_key_name( item => "$1" ) } = $3;
        } elsif ( $line eq "exit" ) {
            $key = undef;
        }
    }
    return \@result_array;
} ## end sub show_running_config

## @name ::show_mac_addr_table
##
## @summary Parses the raw output from the "show mac-addr-table" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from show_mac_addr_table with any of its options as described in the
##    corresponding cdef file 'show.cdefs'
## @synopsis
##      show_mac_addr_table(%args);
## INPUTS TO METHOD CALL 'show_mac_addr_table' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

##&nbsp;Command = show mac-addr-table
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;
##&nbsp;VLAN ID  MAC Address         Interface  IfIndex  Status
##&nbsp;-------  ------------------  ---------  -------  ------------
##&nbsp;1        00:07:43:05:51:2C   0/1        1        Learned
##&nbsp;1        00:07:43:06:8E:43   0/7        7        Learned
##&nbsp;1        00:07:43:06:8E:D1   3/1        18       Learned
##&nbsp;1        00:07:43:06:E5:2A   0/2        2        Learned
##&nbsp;1        00:07:43:07:02:87   3/1        18       Learned
##&nbsp;1        00:07:43:07:8A:DC   3/1        18       Learned
##&nbsp;1        00:1B:21:57:BC:1C   3/1        18       Learned
##&nbsp;1        00:1B:21:65:40:10   0/8        8        Learned
##&nbsp;1        00:A0:98:18:39:8C   0/4        4        Learned
##&nbsp;1        00:A0:98:18:39:8D   3/1        18       Learned
##&nbsp;1        00:A0:98:18:65:40   0/3        3        Learned
##&nbsp;1        00:A0:98:18:65:41   3/1        18       Learned
##&nbsp;1        00:A0:98:23:13:C9   0/6        6        Learned
##&nbsp;1        00:A0:98:23:13:CB   3/1        18       Learned
##&nbsp;1        00:A0:98:23:14:C5   0/5        5        Learned
##&nbsp;1        00:A0:98:23:14:C7   3/1        18       Learned
##&nbsp;1        00:A0:98:4B:B3:A2   3/1        18       Learned
##&nbsp;1        00:A0:98:4B:B7:F1   5/1        17       Management
##&nbsp;
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp; $VAR1 = [
##&nbsp;           {
##&nbsp;             'ifindex' => '1',
##&nbsp;             'mac_address' => '00:07:43:05:51:2C',
##&nbsp;             'status' => 'Learned',
##&nbsp;             'interface' => '0/1',
##&nbsp;             'vlan_id' => '1'
##&nbsp;           },
##&nbsp;           {
##&nbsp;             'ifindex' => '7',
##&nbsp;             'mac_address' => '00:07:43:06:8E:43',
##&nbsp;             'status' => 'Learned',
##&nbsp;             'interface' => '0/7',
##&nbsp;             'vlan_id' => '1'
##&nbsp;           },
##&nbsp;           {
##&nbsp;             'ifindex' => '18',
##&nbsp;             'mac_address' => '00:07:43:06:8E:D1',
##&nbsp;             'status' => 'Learned',
##&nbsp;             'interface' => '3/1',
##&nbsp;             'vlan_id' => '1'
##&nbsp;           },
##&nbsp;           {
##&nbsp;             'ifindex' => '2',
##&nbsp;             'mac_address' => '00:07:43:06:E5:2A',
##&nbsp;             'status' => 'Learned',
##&nbsp;             'interface' => '0/2',
##&nbsp;             'vlan_id' => '1'
##&nbsp;           },
##&nbsp;           {
##&nbsp;             'ifindex' => '18',
##&nbsp;             'mac_address' => '00:07:43:07:02:87',
##&nbsp;             'status' => 'Learned',
##&nbsp;             'interface' => '3/1',
##&nbsp;             'vlan_id' => '1'
##&nbsp;           },
##&nbsp;           {
##&nbsp;             'ifindex' => '18',
##&nbsp;             'mac_address' => '00:07:43:07:8A:DC',
##&nbsp;             'status' => 'Learned',
##&nbsp;             'interface' => '3/1',
##&nbsp;             'vlan_id' => '1'
##&nbsp;           },
##&nbsp;           {
##&nbsp;             'ifindex' => '18',
##&nbsp;             'mac_address' => '00:1B:21:57:BC:1C',
##&nbsp;             'status' => 'Learned',
##&nbsp;             'interface' => '3/1',
##&nbsp;             'vlan_id' => '1'
##&nbsp;           },
##&nbsp;           {
##&nbsp;             'ifindex' => '8',
##&nbsp;             'mac_address' => '00:1B:21:65:40:10',
##&nbsp;             'status' => 'Learned',
##&nbsp;             'interface' => '0/8',
##&nbsp;             'vlan_id' => '1'
##&nbsp;           },
##&nbsp;           {
##&nbsp;             'ifindex' => '4',
##&nbsp;             'mac_address' => '00:A0:98:18:39:8C',
##&nbsp;             'status' => 'Learned',
##&nbsp;             'interface' => '0/4',
##&nbsp;             'vlan_id' => '1'
##&nbsp;           },
##&nbsp;           {
##&nbsp;             'ifindex' => '18',
##&nbsp;             'mac_address' => '00:A0:98:18:39:8D',
##&nbsp;             'status' => 'Learned',
##&nbsp;             'interface' => '3/1',
##&nbsp;             'vlan_id' => '1'
##&nbsp;           },
##&nbsp;           {
##&nbsp;             'ifindex' => '3',
##&nbsp;             'mac_address' => '00:A0:98:18:65:40',
##&nbsp;             'status' => 'Learned',
##&nbsp;             'interface' => '0/3',
##&nbsp;             'vlan_id' => '1'
##&nbsp;           },
##&nbsp;           {
##&nbsp;             'ifindex' => '18',
##&nbsp;             'mac_address' => '00:A0:98:18:65:41',
##&nbsp;             'status' => 'Learned',
##&nbsp;             'interface' => '3/1',
##&nbsp;             'vlan_id' => '1'
##&nbsp;           },
##&nbsp;           {
##&nbsp;             'ifindex' => '6',
##&nbsp;             'mac_address' => '00:A0:98:23:13:C9',
##&nbsp;             'status' => 'Learned',
##&nbsp;             'interface' => '0/6',
##&nbsp;             'vlan_id' => '1'
##&nbsp;           },
##&nbsp;           {
##&nbsp;             'ifindex' => '18',
##&nbsp;             'mac_address' => '00:A0:98:23:13:CB',
##&nbsp;             'status' => 'Learned',
##&nbsp;             'interface' => '3/1',
##&nbsp;             'vlan_id' => '1'
##&nbsp;           },
##&nbsp;           {
##&nbsp;             'ifindex' => '5',
##&nbsp;             'mac_address' => '00:A0:98:23:14:C5',
##&nbsp;             'status' => 'Learned',
##&nbsp;             'interface' => '0/5',
##&nbsp;             'vlan_id' => '1'
##&nbsp;           },
##&nbsp;           {
##&nbsp;             'ifindex' => '18',
##&nbsp;             'mac_address' => '00:A0:98:23:14:C7',
##&nbsp;             'status' => 'Learned',
##&nbsp;             'interface' => '3/1',
##&nbsp;             'vlan_id' => '1'
##&nbsp;           },
##&nbsp;           {
##&nbsp;             'ifindex' => '18',
##&nbsp;             'mac_address' => '00:A0:98:23:13:CB',
##&nbsp;             'status' => 'Learned',
##&nbsp;             'interface' => '3/1',
##&nbsp;             'vlan_id' => '1'
##&nbsp;           },
##&nbsp;           {
##&nbsp;             'ifindex' => '5',
##&nbsp;             'mac_address' => '00:A0:98:23:14:C5',
##&nbsp;             'status' => 'Learned',
##&nbsp;             'interface' => '0/5',
##&nbsp;             'vlan_id' => '1'
##&nbsp;           },
##&nbsp;           {
##&nbsp;             'ifindex' => '18',
##&nbsp;             'mac_address' => '00:A0:98:23:14:C7',
##&nbsp;             'status' => 'Learned',
##&nbsp;             'interface' => '3/1',
##&nbsp;             'vlan_id' => '1'
##&nbsp;           },
##&nbsp;           {
##&nbsp;             'ifindex' => '18',
##&nbsp;             'mac_address' => '00:A0:98:4B:B3:A2',
##&nbsp;             'status' => 'Learned',
##&nbsp;             'interface' => '3/1',
##&nbsp;             'vlan_id' => '1'
##&nbsp;           },
##&nbsp;           {
##&nbsp;             'ifindex' => '17',
##&nbsp;             'mac_address' => '00:A0:98:4B:B7:F1',
##&nbsp;             'status' => 'Management',
##&nbsp;             'interface' => '5/1',
##&nbsp;             'vlan_id' => '1'
##&nbsp;           }
##&nbsp;         ];

sub show_mac_addr_table {
    my %args = @_;

    my ($error_message);

    my $resp   = $args{response_object};
    my $output = $resp->get_processed_output();
    my $error  = _check_error(
        raw_output    => $output,
        error_message => \$error_message
    );
    if ($error) {
        $error = 'NACL::APISet::Exceptions::' . $error . 'Exception';
        $error->throw( text => $error_message, response_object => $resp );
    }
    my @lines = get_lines( raw_output => $output, response_object => $resp );
    my @result_array;
    my $i = 0;
    foreach my $line (@lines) {
        if ( $line =~ /(\d+)\s+(\S+)\s+(\d+\/\d+)\s+(\d+)\s+(\w+)/ ) {
            $result_array[$i]->{'vlan_id'}     = $1;
            $result_array[$i]->{'mac_address'} = $2;
            $result_array[$i]->{'interface'}   = $3;
            $result_array[$i]->{'ifindex'}     = $4;
            $result_array[$i]->{'status'}      = $5;
            $i++;
        }
    }
    return \@result_array;
} ## end sub show_mac_addr_table

## @name ::show_mac_addr_table_interface
##
## @summary Parses the raw output from the "show mac-addr-table interface" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from show_mac_addr_table_interface with any of its options as described in the
##    corresponding cdef file 'show.cdefs'
## @synopsis
##      show_mac_addr_table_interface(%args);
## INPUTS TO METHOD CALL 'show_mac_addr_table_interface' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## <tr> <td> No switch for this key </td> <td> 'interface' </td> <td> '0' </td> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

##&nbsp;Command = show mac-addr-table interface 0/1
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;
##&nbsp;   MAC Address      VLAN ID      Status
##&nbsp;-----------------  ---------  ------------
##&nbsp;00:07:43:05:51:2C     1        Learned
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;
##&nbsp; $VAR1 = [
##&nbsp;           {
##&nbsp;             'mac_address' => '00:07:43:05:51:2C',
##&nbsp;             'status' => 'Learned',
##&nbsp;             'vlan_id' => '1'
##&nbsp;           }
##&nbsp;         ];

sub show_mac_addr_table_interface {
    my %args = @_;

    my ($error_message);

    my $resp   = $args{response_object};
    my $output = $resp->get_processed_output();
    my $error  = _check_error(
        raw_output    => $output,
        error_message => \$error_message
    );

    if ($error) {
        $error = 'NACL::APISet::Exceptions::' . $error . 'Exception';
        $error->throw( text => $error_message, response_object => $resp );
    }
    my @lines = get_lines( raw_output => $output, response_object => $resp );
    my @result_array;
    my $i = 0;
    foreach my $line (@lines) {
        if ( $line =~ /(\S+)\s+(\d+)\s+(\w+)/ ) {
            $result_array[$i]->{'mac_address'} = $1;
            $result_array[$i]->{'vlan_id'}     = $2;
            $result_array[$i]->{'status'}      = $3;
            $i++;

        }
    }
    return \@result_array;
} ## end sub show_mac_addr_table_interface

## @name ::show_hardware
##
## @summary Parses the raw output from the "show hardware" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from show_hardware with any of its options as described in the
##    corresponding cdef file 'show.cdefs'
## @synopsis
##      show_hardware(%args);
## INPUTS TO METHOD CALL 'show_hardware' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

##&nbsp;Command = show hardware
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;
##&nbsp;Switch: 1
##&nbsp;
##&nbsp;System Description............................. NetApp CN1610, 1.0.0.4, Linux
##&nbsp;                                                2.6.21.7
##&nbsp;Machine Type................................... NetApp CN1610
##&nbsp;Machine Model.................................. CN1610
##&nbsp;Serial Number.................................. 10811200002
##&nbsp;Burned In MAC Address.......................... 00:A0:98:4B:AF:4F
##&nbsp;Software Version............................... 1.0.0.4
##&nbsp;Operating System............................... Linux 2.6.21.7
##&nbsp;Network Processing Device...................... BCM56820_B0
##&nbsp;Part Number.................................... 111-00893
##&nbsp;Manufacturer Name.............................. Delta Networks
##&nbsp;Revision....................................... 20
##&nbsp;Date Code...................................... 20110822
##&nbsp;
##&nbsp;--More-- or (q)uit
##&nbsp;
##&nbsp;
##&nbsp;Additional Packages............................ FASTPATH QOS
##&nbsp;                                                FASTPATH IPv6 Management
##&nbsp;
##&nbsp;Power Supply Hardware Data:
##&nbsp;
##&nbsp;      Power                Part Number       Mfg Part Number                     Date
##&nbsp;                                                 Unit  supply  State        & Revision        & Revision        Serial Number     Code
##&nbsp;                                                 ----  ------  -----------  ----------------  ----------------  ----------------  --------
##&nbsp;1     1       Present                        DPSN-300DB H.S0X  DHUD1125000813XX  20110624
##&nbsp;1     2       Present                        DPSN-300DB H.S0X  DHUD1125000765XX  20110624
##&nbsp;
##&nbsp;Fan Tray Hardware Data:
##&nbsp;
##&nbsp;Unit  Fan Tray  State        Part Number & Rev
##&nbsp;                                                 ----  --------  -----------  -------------------
##&nbsp;1     1         Present      441-00033
##&nbsp;1     2         Present      441-00033
##&nbsp;
##&nbsp;SFP Module Hardware Data:
##&nbsp;
##&nbsp;       SFP
##&nbsp;                                                  Unit  Module  State        Vendor            Part Number       Serial Number
##&nbsp;                                                  ----  ------  -----------  ----------------  ----------------  ----------------
##&nbsp;1     1       Present      Molex Inc.        74752-1301        115130105
##&nbsp;1     2       Present      FINISAR CORP.     FTLX8571D3BCL     AKH027F
##&nbsp;1     3       Not present  -                 -                 -
##&nbsp;1     4       Not present  -                 -                 -
##&nbsp;1     5       Not present  -                 -                 -
##&nbsp;1     6       Not present  -                 -                 -
##&nbsp;1     7       Not present  -                 -                 -
##&nbsp;1     8       Not present  -                 -                 -
##&nbsp;1     9       Not present  -                 -                 -
##&nbsp;1     10      Not present  -                 -                 -
##&nbsp;1     11      Not present  -                 -                 -
##&nbsp;1     12      Not present  -                 -                 -
##&nbsp;1     13      Present      FINISAR CORP.     FTLX8571D3BCL     AKH01RB
##&nbsp;1     14      Present      FINISAR CORP.     FTLX8571D3BCL     AKF0L8Q
##&nbsp;1     15      Present      FINISAR CORP.     FTLX8571D3BCL     AKH027E
##&nbsp;1     16      Present      FINISAR CORP.     FTLX8571D3BCL     AKH01XT
##&nbsp;
##&nbsp;
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;          {
##&nbsp;            'serial_number' => '10811200002',
##&nbsp;            'power_supply_hardware_data' => [
##&nbsp;                                              {
##&nbsp;                                                'date_code' => '20110624',
##&nbsp;                                                'serial_number' => 'DHUD1125000813XX',
##&nbsp;                                                'mfg_part_number_and_revision' => 'DPSN-300DB H.S0X',
##&nbsp;                                                'power_supply' => '1',
##&nbsp;                                                'unit' => '1',
##&nbsp;                                                'part_number_and_revision' => '',
##&nbsp;                                                'state' => 'Present'
##&nbsp;                                              },
##&nbsp;                                              {
##&nbsp;                                                'date_code' => '20110624',
##&nbsp;                                                'serial_number' => 'DHUD1125000765XX',
##&nbsp;                                                'mfg_part_number_and_revision' => 'DPSN-300DB H.S0X',
##&nbsp;                                                'power_supply' => '2',
##&nbsp;                                                'unit' => '1',
##&nbsp;                                                'part_number_and_revision' => '',
##&nbsp;                                                'state' => 'Present'
##&nbsp;                                              }
##&nbsp;                                            ],
##&nbsp;            'operating_system' => 'Linux 2.6.21.7',
##&nbsp;            'network_processing_device' => 'BCM56820_B0',
##&nbsp;            'machine_type' => 'NetApp CN1610',
##&nbsp;            'date_code' => '20110822',
##&nbsp;            'burned_in_mac_address' => '00:A0:98:4B:AF:4F',
##&nbsp;            'additional_packages' => [
##&nbsp;                                       'FASTPATH QOS',
##&nbsp;                                       'FASTPATH IPv6 Management'
##&nbsp;                                     ],
##&nbsp;            'fan_tray_hardware_data' => [
##&nbsp;                                          {
##&nbsp;                                            'unit' => '1',
##&nbsp;                                            'part_number_and_revision' => '441-00033',
##&nbsp;                                            'fan_tray' => '1',
##&nbsp;                                            'state' => 'Present'
##&nbsp;                                          },
##&nbsp;                                          {
##&nbsp;                                            'unit' => '1',
##&nbsp;                                            'part_number_and_revision' => '441-00033',
##&nbsp;                                            'fan_tray' => '2',
##&nbsp;                                            'state' => 'Present'
##&nbsp;                                          }
##&nbsp;                                        ],
##&nbsp;            'sfp_module_hardware_data' => [
##&nbsp;                                            {
##&nbsp;                                              'serial_number' => '115130105',
##&nbsp;                                              'sfp_module' => '1',
##&nbsp;                                              'part_number' => '74752-1301',
##&nbsp;                                              'unit' => '1',
##&nbsp;                                              'vendor' => 'Molex Inc.',
##&nbsp;                                              'state' => 'Present'
##&nbsp;                                            },
##&nbsp;                                            {
##&nbsp;                                              'serial_number' => 'AKH027F',
##&nbsp;                                              'sfp_module' => '2',
##&nbsp;                                              'part_number' => 'FTLX8571D3BCL',
##&nbsp;                                              'unit' => '1',
##&nbsp;                                              'vendor' => 'FINISAR CORP.',
##&nbsp;                                              'state' => 'Present'
##&nbsp;                                            },
##&nbsp;                                            {
##&nbsp;                                              'serial_number' => '-',
##&nbsp;                                              'sfp_module' => '3',
##&nbsp;                                              'part_number' => '-',
##&nbsp;                                              'unit' => '1',
##&nbsp;                                              'vendor' => '-',
##&nbsp;                                              'state' => 'Not present'
##&nbsp;                                            },
##&nbsp;                                            {
##&nbsp;                                              'serial_number' => '-',
##&nbsp;                                              'sfp_module' => '4',
##&nbsp;                                              'part_number' => '-',
##&nbsp;                                              'unit' => '1',
##&nbsp;                                              'vendor' => '-',
##&nbsp;                                              'state' => 'Not present'
##&nbsp;                                            },
##&nbsp;                                            {
##&nbsp;                                              'serial_number' => '-',
##&nbsp;                                              'sfp_module' => '5',
##&nbsp;                                              'part_number' => '-',
##&nbsp;                                              'unit' => '1',
##&nbsp;                                              'vendor' => '-',
##&nbsp;                                              'state' => 'Not present'
##&nbsp;                                            },
##&nbsp;                                            {
##&nbsp;                                              'serial_number' => '-',
##&nbsp;                                              'sfp_module' => '6',
##&nbsp;                                              'part_number' => '-',
##&nbsp;                                              'unit' => '1',
##&nbsp;                                              'vendor' => '-',
##&nbsp;                                              'state' => 'Not present'
##&nbsp;                                            },
##&nbsp;                                            {
##&nbsp;                                              'serial_number' => '-',
##&nbsp;                                              'sfp_module' => '7',
##&nbsp;                                              'part_number' => '-',
##&nbsp;                                              'unit' => '1',
##&nbsp;                                              'vendor' => '-',
##&nbsp;                                              'state' => 'Not present'
##&nbsp;                                            },
##&nbsp;                                            {
##&nbsp;                                              'serial_number' => '-',
##&nbsp;                                              'sfp_module' => '8',
##&nbsp;                                              'part_number' => '-',
##&nbsp;                                              'unit' => '1',
##&nbsp;                                              'vendor' => '-',
##&nbsp;                                              'state' => 'Not present'
##&nbsp;                                            },
##&nbsp;                                            {
##&nbsp;                                              'serial_number' => '-',
##&nbsp;                                              'sfp_module' => '9',
##&nbsp;                                              'part_number' => '-',
##&nbsp;                                              'unit' => '1',
##&nbsp;                                              'vendor' => '-',
##&nbsp;                                              'state' => 'Not present'
##&nbsp;                                            },
##&nbsp;                                            {
##&nbsp;                                              'serial_number' => '-',
##&nbsp;                                              'sfp_module' => '10',
##&nbsp;                                              'part_number' => '-',
##&nbsp;                                              'unit' => '1',
##&nbsp;                                              'vendor' => '-',
##&nbsp;                                              'state' => 'Not present'
##&nbsp;                                            },
##&nbsp;                                            {
##&nbsp;                                              'serial_number' => '-',
##&nbsp;                                              'sfp_module' => '11',
##&nbsp;                                              'part_number' => '-',
##&nbsp;                                              'unit' => '1',
##&nbsp;                                              'vendor' => '-',
##&nbsp;                                              'state' => 'Not present'
##&nbsp;                                            },
##&nbsp;                                            {
##&nbsp;                                              'serial_number' => '-',
##&nbsp;                                              'sfp_module' => '12',
##&nbsp;                                              'part_number' => '-',
##&nbsp;                                              'unit' => '1',
##&nbsp;                                              'vendor' => '-',
##&nbsp;                                              'state' => 'Not present'
##&nbsp;                                            },
##&nbsp;                                            {
##&nbsp;                                              'serial_number' => 'AKH01RB',
##&nbsp;                                              'sfp_module' => '13',
##&nbsp;                                              'part_number' => 'FTLX8571D3BCL',
##&nbsp;                                              'unit' => '1',
##&nbsp;                                              'vendor' => 'FINISAR CORP.',
##&nbsp;                                              'state' => 'Present'
##&nbsp;                                            },
##&nbsp;                                            {
##&nbsp;                                              'serial_number' => 'AKF0L8Q',
##&nbsp;                                              'sfp_module' => '14',
##&nbsp;                                              'part_number' => 'FTLX8571D3BCL',
##&nbsp;                                              'unit' => '1',
##&nbsp;                                              'vendor' => 'FINISAR CORP.',
##&nbsp;                                              'state' => 'Present'
##&nbsp;                                            },
##&nbsp;                                            {
##&nbsp;                                              'serial_number' => 'AKH027E',
##&nbsp;                                              'sfp_module' => '15',
##&nbsp;                                              'part_number' => 'FTLX8571D3BCL',
##&nbsp;                                              'unit' => '1',
##&nbsp;                                              'vendor' => 'FINISAR CORP.',
##&nbsp;                                              'state' => 'Present'
##&nbsp;                                            },
##&nbsp;                                            {
##&nbsp;                                              'serial_number' => 'AKH01XT',
##&nbsp;                                              'sfp_module' => '16',
##&nbsp;                                              'part_number' => 'FTLX8571D3BCL',
##&nbsp;                                              'unit' => '1',
##&nbsp;                                              'vendor' => 'FINISAR CORP.',
##&nbsp;                                              'state' => 'Present'
##&nbsp;                                            }
##&nbsp;                                          ],
##&nbsp;            'system_description' => 'NetApp CN1610, 1.0.0.4, Linux 2.6.21.7',
##&nbsp;            'part_number' => '111-00893',
##&nbsp;            'revision' => '20',
##&nbsp;            'software_version' => '1.0.0.4',
##&nbsp;            'switch' => '1',
##&nbsp;            'machine_model' => 'CN1610',
##&nbsp;            'manufacturer_name' => 'Delta Networks'
##&nbsp;          }
##&nbsp;        ];
##&nbsp;
sub show_hardware {
    my %args = @_;

    my ($error_message);

    my $resp        = $args{response_object};
    my $output      = $resp->get_processed_output();
    my $command     = $resp->get_command();
    my $error_level = $resp->get_execution_status();
    my ( $error, $key );

    # Status returned by switch for this command is
    # is ZERO even if command failed
    if ( not defined($error) ) {
        $error = _check_error(
            raw_output    => $output,
            command       => $command,
            error_message => \$error_message
        );
    }
    if ($error) {
        $error = 'NACL::APISet::Exceptions::' . $error . 'Exception';
        $error->throw( text => $error_message, response_object => $resp );
    }

    my @result_array;
    my @lines = get_lines( raw_output => $output, first_line => 'Switch:\s+\d+' );
    my $header = undef;
    my ( @header_length, @header_keys );
    my @power_supply_header = (
        qw(unit power_supply state part_number_and_revision
            mfg_part_number_and_revision serial_number date_code)
    );
    my @fan_tray_header   = (qw(unit fan_tray state part_number_and_revision));
    my @sfp_module_header = (qw(unit sfp_module state vendor part_number serial_number));
    my $switch_id         = -1;
    my ( $line, $j );
    for ( my $line_id = 0; $line_id < scalar(@lines); $line_id++ ) {
        $line = $lines[$line_id];
        next if ( $line =~ /--More--\s+or\s+\(q\)uit/ );
        next if ( !$line || !defined($line) );
        if ( $line =~ /Switch:\s+(\d+)/ ) {
            $switch_id++;
            $result_array[$switch_id]->{switch} = $1;
            next;
        }

        if ( $line =~ /^(.*\w)\.\.+\s*(.*)/ ) {
            $result_array[$switch_id]->{ to_key_name( item => $1 ) } = $2;
        } elsif ( $line =~ /Power\sSupply\sHardware\sData/ ) {
            $header      = to_key_name( item => "Power Supply Hardware Data" );
            $j           = 0;
            @header_keys = @power_supply_header;
        } elsif ( $line =~ /Fan\sTray\sHardware\sData/ ) {
            $header      = to_key_name( item => "Fan Tray Hardware Data" );
            $j           = 0;
            @header_keys = @fan_tray_header;
        } elsif ( $line =~ /SFP\sModule\sHardware\sData/ ) {
            $header      = to_key_name( item => "SFP Module Hardware Data" );
            $j           = 0;
            @header_keys = @sfp_module_header;
        } elsif ( $line =~ /(Unit|Part\sNumber|SFP)/ ) {
            next;
        } elsif ( $line =~ /^----/ ) {
            @header_length = split( /\s+/, $line );
        } else {
            if ( !defined($header) ) {
                if ( $lines[ $line_id - 1 ] =~ /Additional\sPackages/ ) {
                    my @temp_arr;
                    my $temp = $result_array[$switch_id]->{additional_packages};
                    if ( ref($temp) ) {
                        @temp_arr = ( @$temp, $line );
                    } else {
                        @temp_arr = ( $temp, $line );
                    }
                    $result_array[$switch_id]->{additional_packages} = \@temp_arr;
                } else {
                    $lines[ $line_id - 1 ] =~ m/^(.*\w)\.\.+\s*(.*)/;
                    $result_array[$switch_id]->{ to_key_name( item => $1 ) } .= " " . $line;
                }
            } else {
                my $l = 0;
                for ( my $i = 0; $i < scalar(@header_length); $i++ ) {
                    if ( $i != scalar(@header_length) - 1 ) {
                        $result_array[$switch_id]->{$header}->[$j]->{ $header_keys[$i] }
                            = substr( $line, $l, length( $header_length[$i] ) );
                    } else {
                        $result_array[$switch_id]->{$header}->[$j]->{ $header_keys[$i] } = substr( $line, $l );
                    }
                    $l += ( length( $header_length[$i] ) + 2 );
                    $result_array[$switch_id]->{$header}->[$j]->{ $header_keys[$i] } =~ s/^\s+//;
                    $result_array[$switch_id]->{$header}->[$j]->{ $header_keys[$i] } =~ s/\s+$//;
                }
                $j++;
            }
        }
    }

    return \@result_array;
} ## end sub show_hardware

## @name ::show_interface
##
## @summary Parses the raw output from the "show interface" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from show_interface with any of its options as described in the
##    corresponding cdef file 'show.cdefs'
## @synopsis
##      show_interface(%args);
## INPUTS TO METHOD CALL 'show_interface' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## <tr> <td> No switch for this key </td> <td> 'interface' </td> <td> '0' </td> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

##&nbsp;Command = show interface 0/1
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;
##&nbsp;
##&nbsp;Packets Received Without Error................. 3412009
##&nbsp;Packets Received With Error.................... 0
##&nbsp;Broadcast Packets Received..................... 1140
##&nbsp;Packets Transmitted Without Errors............. 4104069
##&nbsp;Transmit Packet Errors......................... 0
##&nbsp;Collision Frames............................... 0
##&nbsp;Time Since Counters Last Cleared............... 6 day 8 hr 34 min 12 sec
##&nbsp;
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;          {
##&nbsp;            'time_since_counters_last_cleared' => '6 day 8 hr 34 min 12 sec',
##&nbsp;            'packets_transmitted_without_errors' => '4104069',
##&nbsp;            'collision_frames' => '0',
##&nbsp;            'packets_received_with_error' => '0',
##&nbsp;            'broadcast_packets_received' => '1140',
##&nbsp;            'transmit_packet_errors' => '0',
##&nbsp;            'packets_received_without_error' => '3412009'
##&nbsp;          }
##&nbsp;        ];
##&nbsp;
sub show_interface {
    my %args = @_;

    my ($error_message);

    my $resp   = $args{response_object};
    my $output = $resp->get_processed_output();
    my $error  = _check_error(
        raw_output    => $output,
        error_message => \$error_message
    );
    if ($error) {
        $error = 'NACL::APISet::Exceptions::' . $error . 'Exception';
        $error->throw( text => $error_message, response_object => $resp );
    }
    my @lines = get_lines( raw_output => $output, response_object => $resp );
    my @result_array;
    foreach my $line (@lines) {
        if ( $line =~ /(.*\w+)\)?\.+\s+(.*)$/ ) {
            $result_array[0]->{ to_key_name( item => $1 ) } = $2;
        }
    }
    return \@result_array;
} ## end sub show_interface

## @name ::show_interface_ethernet
##
## @summary Parses the raw output from the "show interface ethernet" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from show_interface_ethernet with any of its options as described in the
##    corresponding cdef file 'show.cdefs'
## @synopsis
##      show_interface_ethernet(%args);
## INPUTS TO METHOD CALL 'show_interface_ethernet' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## <tr> <td> No switch for this key </td> <td> 'interface' </td> <td> '0' </td> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

##&nbsp;Command = show interface ethernet 0/1
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;
##&nbsp;
##&nbsp;Total Packets Received (Octets)................ 3255569764
##&nbsp;Packets Received 64 Octets..................... 2735
##&nbsp;Packets Received 65-127 Octets................. 1699869
##&nbsp;Packets Received 128-255 Octets................ 1096628
##&nbsp;Packets Received 256-511 Octets................ 18948
##&nbsp;Packets Received 512-1023 Octets............... 31518
##&nbsp;Packets Received 1024-1518 Octets.............. 158924
##&nbsp;Packets Received > 1518 Octets................. 341368
##&nbsp;Packets RX and TX 64 Octets.................... 28508
##&nbsp;Packets RX and TX 65-127 Octets................ 4622800
##&nbsp;Packets RX and TX 128-255 Octets............... 1227538
##&nbsp;Packets RX and TX 256-511 Octets............... 313217
##&nbsp;Packets RX and TX 512-1023 Octets.............. 252309
##&nbsp;Packets RX and TX 1024-1518 Octets............. 456488
##&nbsp;Packets RX and TX 1519-2047 Octets............. 1490
##&nbsp;Packets RX and TX 2048-4095 Octets............. 29479
##&nbsp;Packets RX and TX 4096-9216 Octets............. 454438
##&nbsp;
##&nbsp;Total Packets Received Without Errors.......... 3349990
##&nbsp;Unicast Packets Received....................... 3339907
##&nbsp;Multicast Packets Received..................... 8984
##&nbsp;Broadcast Packets Received..................... 1099
##&nbsp;Total Packets Received with MAC Errors......... 0
##&nbsp;Jabbers Received............................... 0
##&nbsp;Fragments Received............................. 0
##&nbsp;Undersize Received............................. 0
##&nbsp;Alignment Errors............................... 0
##&nbsp;FCS Errors..................................... 0
##&nbsp;Overruns....................................... 0
##&nbsp;
##&nbsp;Total Received Packets Not Forwarded........... 0
##&nbsp;802.3x Pause Frames Received................... 0
##&nbsp;Unacceptable Frame Type........................ 0
##&nbsp;
##&nbsp;Total Packets Transmitted (Octets)............. 2190513425
##&nbsp;Packets Transmitted 64 Octets.................. 25773
##&nbsp;Packets Transmitted 65-127 Octets.............. 2922931
##&nbsp;Packets Transmitted 128-255 Octets............. 130910
##&nbsp;Packets Transmitted 256-511 Octets............. 294269
##&nbsp;Packets Transmitted 512-1023 Octets............ 220791
##&nbsp;Packets Transmitted 1024-1518 Octets........... 297564
##&nbsp;Packets Transmitted > 1518 Octets.............. 144039
##&nbsp;Max Frame Size................................. 9216
##&nbsp;
##&nbsp;--More-- or (q)uit
##&nbsp;Total Packets Transmitted Successfully......... 4036277
##&nbsp;Unicast Packets Transmitted.................... 3515323
##&nbsp;Multicast Packets Transmitted.................. 496248
##&nbsp;Broadcast Packets Transmitted.................. 24706
##&nbsp;
##&nbsp;Total Transmit Errors.......................... 0
##&nbsp;
##&nbsp;Total Transmit Packets Discarded............... 0
##&nbsp;Single Collision Frames........................ 0
##&nbsp;Multiple Collision Frames...................... 0
##&nbsp;Excessive Collision Frames..................... 0
##&nbsp;
##&nbsp;802.3x Pause Frames Transmitted................ 0
##&nbsp;GVRP PDUs received............................. 0
##&nbsp;GVRP PDUs Transmitted.......................... 0
##&nbsp;GVRP Failed Registrations...................... 0
##&nbsp;GMRP PDUs Received............................. 0
##&nbsp;GMRP PDUs Transmitted.......................... 0
##&nbsp;GMRP Failed Registrations...................... 0
##&nbsp;
##&nbsp;STP BPDUs Transmitted.......................... 0
##&nbsp;STP BPDUs Received............................. 0
##&nbsp;RSTP BPDUs Transmitted......................... 0
##&nbsp;--More-- or (q)uit
##&nbsp;
##&nbsp;RSTP BPDUs Received............................ 0
##&nbsp;MSTP BPDUs Transmitted......................... 269530
##&nbsp;MSTP BPDUs Received............................ 0
##&nbsp;
##&nbsp;EAPOL Frames Transmitted....................... 0
##&nbsp;EAPOL Start Frames Received.................... 0
##&nbsp;
##&nbsp;Time Since Counters Last Cleared............... 6 day 5 hr 46 min 47 sec
##&nbsp;
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;          {
##&nbsp;            'packets_received_65_127_octets' => '36590',
##&nbsp;            'unacceptable_frame_type' => '0',
##&nbsp;            'total_packets_received_without_errors' => '203977',
##&nbsp;            'undersize_received' => '0',
##&nbsp;            'broadcast_packets_received' => '408',
##&nbsp;            'stp_bpdus_received' => '0',
##&nbsp;            'excessive_collision_frames' => '0',
##&nbsp;            'gvrp_pdus_transmitted' => '0',
##&nbsp;            'total_packets_received_with_mac_errors' => '0',
##&nbsp;            'packets_received_1024_1518_octets' => '22480',
##&nbsp;            'mstp_bpdus_transmitted' => '799400',
##&nbsp;            'multicast_packets_transmitted' => '1874968',
##&nbsp;            'max_frame_size' => '1518',
##&nbsp;            'total_received_packets_not_forwarded' => '0',
##&nbsp;            'packets_rx_and_tx_1024_1518_octets' => '169240',
##&nbsp;            'unicast_packets_transmitted' => '356389',
##&nbsp;            'packets_rx_and_tx_2048_4095_octets' => '0',
##&nbsp;            'packets_transmitted_>_1518_octets' => '0',
##&nbsp;            'packets_rx_and_tx_65_127_octets' => '2640061',
##&nbsp;            'packets_transmitted_1024_1518_octets' => '146760',
##&nbsp;            'multiple_collision_frames' => '0',
##&nbsp;            '802.3x_pause_frames_transmitted' => '0',
##&nbsp;            'packets_received_256_511_octets' => '386',
##&nbsp;            'multicast_packets_received' => '26583',
##&nbsp;            'broadcast_packets_transmitted' => '1516924',
##&nbsp;            'packets_received_64_octets' => '95444',
##&nbsp;            'packets_received_128_255_octets' => '48972',
##&nbsp;            'overruns' => '0',
##&nbsp;            'packets_received_>_1518_octets' => '0',
##&nbsp;            'single_collision_frames' => '0',
##&nbsp;            'time_since_counters_last_cleared' => '18 day 12 hr 17 min 7 sec',
##&nbsp;            'jabbers_received' => '0',
##&nbsp;            'mstp_bpdus_received' => '0',
##&nbsp;            '802.3x_pause_frames_received' => '0',
##&nbsp;            'gvrp_failed_registrations' => '0',
##&nbsp;            'total_packets_transmitted_successfully' => '3748281',
##&nbsp;            'packets_rx_and_tx_256_511_octets' => '35623',
##&nbsp;            'rstp_bpdus_received' => '0',
##&nbsp;            'fragments_received' => '0',
##&nbsp;            'packets_transmitted_128_255_octets' => '75253',
##&nbsp;            'fcs_errors' => '0',
##&nbsp;            'packets_transmitted_512_1023_octets' => '327534',
##&nbsp;            'stp_bpdus_transmitted' => '0',
##&nbsp;            'packets_rx_and_tx_1519_2047_octets' => '0',
##&nbsp;            'packets_rx_and_tx_128_255_octets' => '124225',
##&nbsp;            'packets_transmitted_64_octets' => '560026',
##&nbsp;            'gmrp_pdus_transmitted' => '0',
##&nbsp;            'rstp_bpdus_transmitted' => '0',
##&nbsp;            'eapol_frames_transmitted' => '0',
##&nbsp;            'unicast_packets_received' => '176986',
##&nbsp;            'gvrp_pdus_received' => '0',
##&nbsp;            'packets_rx_and_tx_4096_9216_octets' => '0',
##&nbsp;            'gmrp_pdus_received' => '0',
##&nbsp;            'total_packets_received_octets' => '52292432',
##&nbsp;            'packets_received_512_1023_octets' => '105',
##&nbsp;            'total_transmit_packets_discarded' => '0',
##&nbsp;            'packets_rx_and_tx_64_octets' => '655470',
##&nbsp;            'packets_rx_and_tx_512_1023_octets' => '327639',
##&nbsp;            'gmrp_failed_registrations' => '0',
##&nbsp;            'total_transmit_errors' => '0',
##&nbsp;            'packets_transmitted_256_511_octets' => '35237',
##&nbsp;            'total_packets_transmitted_octets' => '756944845',
##&nbsp;            'alignment_errors' => '0',
##&nbsp;            'packets_transmitted_65_127_octets' => '2603471',
##&nbsp;            'eapol_start_frames_received' => '0'
##&nbsp;          }
##&nbsp;        ];
##&nbsp;

sub show_interface_ethernet {
    my %args = @_;

    my ($error_message);

    my $resp   = $args{response_object};
    my $output = $resp->get_processed_output();
    my $error  = _check_error(
        raw_output    => $output,
        error_message => \$error_message
    );
    if ($error) {
        $error = 'NACL::APISet::Exceptions::' . $error . 'Exception';
        $error->throw( text => $error_message, response_object => $resp );
    }
    my @lines = get_lines( raw_output => $output, response_object => $resp );
    my @result_array;
    foreach my $line (@lines) {
        if ( $line =~ /(.*\w+)\)?\.+\s+(.*)$/ ) {
            my $temp1 = $1;
            my $temp2 = $2;
            $temp1 =~ s/-/_/;
            $result_array[0]->{ to_key_name( item => $temp1 ) } = $temp2;
        }
    }
    return \@result_array;
} ## end sub show_interface_ethernet

## @name ::show_storm_control
##
## @summary Parses the raw output from the "show storm-control" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from show_storm_control with any of its options as described in the
##    corresponding cdef file 'show.cdefs'
## @synopsis
##      show_storm_control(%args);
## INPUTS TO METHOD CALL 'show_storm_control' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## <tr> <td> No switch for this key </td> <td> 'interface' </td> <td> '0' </td> </tr>
## <tr> <td> 'all' </td> <td> 'all' </td> <td> '1' </td> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples
##&nbsp;Command = show storm-control all
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;
##&nbsp;        Bcast   Bcast    Mcast   Mcast    Ucast   Ucast    Flow
##&nbsp; Intf   Mode    Level    Mode    Level    Mode    Level    Mode
##&nbsp;------ ------- -------- ------- -------- ------- -------- ------
##&nbsp;0/1    Disable       5% Disable       5% Disable       5% Enable
##&nbsp;0/2    Disable       5% Disable       5% Disable       5% Enable
##&nbsp;0/3    Disable       5% Disable       5% Disable       5% Disable
##&nbsp;0/4    Disable       5% Disable       5% Disable       5% Disable
##&nbsp;0/5    Disable       5% Disable       5% Disable       5% Disable
##&nbsp;0/6    Disable       5% Disable       5% Disable       5% Disable
##&nbsp;0/7    Disable       5% Disable       5% Disable       5% Disable
##&nbsp;0/8    Disable       5% Disable       5% Disable       5% Disable
##&nbsp;0/9    Disable       5% Disable       5% Disable       5% Disable
##&nbsp;0/10   Disable       5% Disable       5% Disable       5% Disable
##&nbsp;0/11   Disable       5% Disable       5% Disable       5% Disable
##&nbsp;0/12   Disable       5% Disable       5% Disable       5% Disable
##&nbsp;0/13   Disable       5% Disable       5% Disable       5% Disable
##&nbsp;0/14   Disable       5% Disable       5% Disable       5% Disable
##&nbsp;0/15   Disable       5% Disable       5% Disable       5% Disable
##&nbsp;0/16   Disable       5% Disable       5% Disable       5% Disable
##&nbsp;
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;          {
##&nbsp;            'bcast_mode' => 'Disable',
##&nbsp;            'ucast_level' => '5%',
##&nbsp;            'interface' => '0/1',
##&nbsp;            'mcast_level' => '5%',
##&nbsp;            'bcast_level' => '5%',
##&nbsp;            'mcast_mode' => 'Disable',
##&nbsp;            'ucast_mode' => 'Disable',
##&nbsp;            'flow_mode' => 'Enable'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'bcast_mode' => 'Disable',
##&nbsp;            'ucast_level' => '5%',
##&nbsp;            'interface' => '0/2',
##&nbsp;            'mcast_level' => '5%',
##&nbsp;            'bcast_level' => '5%',
##&nbsp;            'mcast_mode' => 'Disable',
##&nbsp;            'ucast_mode' => 'Disable',
##&nbsp;            'flow_mode' => 'Enable'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'bcast_mode' => 'Disable',
##&nbsp;            'ucast_level' => '5%',
##&nbsp;            'interface' => '0/3',
##&nbsp;            'mcast_level' => '5%',
##&nbsp;            'bcast_level' => '5%',
##&nbsp;            'mcast_mode' => 'Disable',
##&nbsp;            'ucast_mode' => 'Disable',
##&nbsp;            'flow_mode' => 'Disable'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'bcast_mode' => 'Disable',
##&nbsp;            'ucast_level' => '5%',
##&nbsp;            'interface' => '0/4',
##&nbsp;            'mcast_level' => '5%',
##&nbsp;            'bcast_level' => '5%',
##&nbsp;            'mcast_mode' => 'Disable',
##&nbsp;            'ucast_mode' => 'Disable',
##&nbsp;            'flow_mode' => 'Disable'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'bcast_mode' => 'Disable',
##&nbsp;            'ucast_level' => '5%',
##&nbsp;            'interface' => '0/5',
##&nbsp;            'mcast_level' => '5%',
##&nbsp;            'bcast_level' => '5%',
##&nbsp;            'mcast_mode' => 'Disable',
##&nbsp;            'ucast_mode' => 'Disable',
##&nbsp;            'flow_mode' => 'Disable'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'bcast_mode' => 'Disable',
##&nbsp;            'ucast_level' => '5%',
##&nbsp;            'interface' => '0/6',
##&nbsp;            'mcast_level' => '5%',
##&nbsp;            'bcast_level' => '5%',
##&nbsp;            'mcast_mode' => 'Disable',
##&nbsp;            'ucast_mode' => 'Disable',
##&nbsp;            'flow_mode' => 'Disable'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'bcast_mode' => 'Disable',
##&nbsp;            'ucast_level' => '5%',
##&nbsp;            'interface' => '0/7',
##&nbsp;            'mcast_level' => '5%',
##&nbsp;            'bcast_level' => '5%',
##&nbsp;            'mcast_mode' => 'Disable',
##&nbsp;            'ucast_mode' => 'Disable',
##&nbsp;            'flow_mode' => 'Disable'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'bcast_mode' => 'Disable',
##&nbsp;            'ucast_level' => '5%',
##&nbsp;            'interface' => '0/8',
##&nbsp;            'mcast_level' => '5%',
##&nbsp;            'bcast_level' => '5%',
##&nbsp;            'mcast_mode' => 'Disable',
##&nbsp;            'ucast_mode' => 'Disable',
##&nbsp;            'flow_mode' => 'Disable'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'bcast_mode' => 'Disable',
##&nbsp;            'ucast_level' => '5%',
##&nbsp;            'interface' => '0/9',
##&nbsp;            'mcast_level' => '5%',
##&nbsp;            'bcast_level' => '5%',
##&nbsp;            'mcast_mode' => 'Disable',
##&nbsp;            'ucast_mode' => 'Disable',
##&nbsp;            'flow_mode' => 'Disable'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'bcast_mode' => 'Disable',
##&nbsp;            'ucast_level' => '5%',
##&nbsp;            'interface' => '0/10',
##&nbsp;            'mcast_level' => '5%',
##&nbsp;            'bcast_level' => '5%',
##&nbsp;            'mcast_mode' => 'Disable',
##&nbsp;            'ucast_mode' => 'Disable',
##&nbsp;            'flow_mode' => 'Disable'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'bcast_mode' => 'Disable',
##&nbsp;            'ucast_level' => '5%',
##&nbsp;            'interface' => '0/11',
##&nbsp;            'mcast_level' => '5%',
##&nbsp;            'bcast_level' => '5%',
##&nbsp;            'mcast_mode' => 'Disable',
##&nbsp;            'ucast_mode' => 'Disable',
##&nbsp;            'flow_mode' => 'Disable'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'bcast_mode' => 'Disable',
##&nbsp;            'ucast_level' => '5%',
##&nbsp;            'interface' => '0/12',
##&nbsp;            'mcast_level' => '5%',
##&nbsp;            'bcast_level' => '5%',
##&nbsp;            'mcast_mode' => 'Disable',
##&nbsp;            'ucast_mode' => 'Disable',
##&nbsp;            'flow_mode' => 'Disable'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'bcast_mode' => 'Disable',
##&nbsp;            'ucast_level' => '5%',
##&nbsp;            'interface' => '0/13',
##&nbsp;            'mcast_level' => '5%',
##&nbsp;            'bcast_level' => '5%',
##&nbsp;            'mcast_mode' => 'Disable',
##&nbsp;            'ucast_mode' => 'Disable',
##&nbsp;            'flow_mode' => 'Disable'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'bcast_mode' => 'Disable',
##&nbsp;            'ucast_level' => '5%',
##&nbsp;            'interface' => '0/14',
##&nbsp;            'mcast_level' => '5%',
##&nbsp;            'bcast_level' => '5%',
##&nbsp;            'mcast_mode' => 'Disable',
##&nbsp;            'ucast_mode' => 'Disable',
##&nbsp;            'flow_mode' => 'Disable'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'bcast_mode' => 'Disable',
##&nbsp;            'ucast_level' => '5%',
##&nbsp;            'interface' => '0/15',
##&nbsp;            'mcast_level' => '5%',
##&nbsp;            'bcast_level' => '5%',
##&nbsp;            'mcast_mode' => 'Disable',
##&nbsp;            'ucast_mode' => 'Disable',
##&nbsp;            'flow_mode' => 'Disable'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'bcast_mode' => 'Disable',
##&nbsp;            'ucast_level' => '5%',
##&nbsp;            'interface' => '0/16',
##&nbsp;            'mcast_level' => '5%',
##&nbsp;            'bcast_level' => '5%',
##&nbsp;            'mcast_mode' => 'Disable',
##&nbsp;            'ucast_mode' => 'Disable',
##&nbsp;            'flow_mode' => 'Disable'
##&nbsp;          }
##&nbsp;        ];
##&nbsp;
##&nbsp;Command = show storm-control 0/1
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;
##&nbsp;        Bcast   Bcast    Mcast   Mcast    Ucast   Ucast    Flow
##&nbsp; Intf   Mode    Level    Mode    Level    Mode    Level    Mode
##&nbsp;------ ------- -------- ------- -------- ------- -------- ------
##&nbsp;0/1    Disable       5% Disable       5% Disable       5% Enable
##&nbsp;
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;          {
##&nbsp;            'bcast_mode' => 'Disable',
##&nbsp;            'ucast_level' => '5%',
##&nbsp;            'interface' => '0/1',
##&nbsp;            'mcast_level' => '5%',
##&nbsp;            'bcast_level' => '5%',
##&nbsp;            'mcast_mode' => 'Disable',
##&nbsp;            'ucast_mode' => 'Disable',
##&nbsp;            'flow_mode' => 'Enable'
##&nbsp;          }
##&nbsp;        ];
##&nbsp;
##&nbsp;Command = show storm-control
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;
##&nbsp;Broadcast Storm Control Mode................... Disable
##&nbsp;Broadcast Storm Control Level.................. 5 percent
##&nbsp;Multicast Storm Control Mode................... Disable
##&nbsp;Multicast Storm Control Level.................. 5 percent
##&nbsp;Unicast Storm Control Mode..................... Disable
##&nbsp;Unicast Storm Control Level.................... 5 percent
##&nbsp;
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;          {
##&nbsp;            'unicast_storm_control_mode' => 'Disable',
##&nbsp;            'unicast_storm_control_level' => '5 percent',
##&nbsp;            'multicast_storm_control_level' => '5 percent',
##&nbsp;            'multicast_storm_control_mode' => 'Disable',
##&nbsp;            'broadcast_storm_control_mode' => 'Disable',
##&nbsp;            'broadcast_storm_control_level' => '5 percent'
##&nbsp;          }
##&nbsp;        ];
##&nbsp;
sub show_storm_control {
    my %args = @_;
    my ($error_message);
    my $resp   = $args{response_object};
    my $output = $resp->get_processed_output();
    my $error  = _check_error(
        raw_output    => $output,
        error_message => \$error_message
    );
    if ($error) {
        $error = 'NACL::APISet::Exceptions::' . $error . 'Exception';
        $error->throw( text => $error_message, response_object => $resp );
    }

    my @result_array;
    my @lines = get_lines( raw_output => $output, response_object => $resp );
    for ( my $i = 0; $i <= $#lines; $i++ ) {
        my @stats;
        if ( $lines[$i] =~ /(.*\w+)\)?\.+\s+(.*)$/ ) {
            my $temp1 = $1;
            $temp1 =~ s/-/_/;
            $result_array[0]->{ to_key_name( item => $temp1 ) } = $2;
        }
        my ( @all_fields, @values );
        if (   ( $lines[$i] =~ /Bcast/ )
            && ( $lines[ $i + 1 ] =~ /Intf/ ) )
        {
            my @fields1 = split( /\s+/, $lines[$i] );
            my @fields2 = split( /\s+/, $lines[ $i + 1 ] );
            $fields2[0] = "interface" if ( $fields2[0] =~ /Intf/ );
            push( @all_fields, lc( $fields2[0] ) );

            for ( my $field = 0; $field < $#fields2; $field++ ) {
                my $field_name = lc( $fields1[$field] ) . "_" . lc( $fields2[ $field + 1 ] );
                push( @all_fields, $field_name );
            }
        }
        my $line_count = 0;
        if ( $output =~ /Bcast/ ) {
            while ( $i <= $#lines ) {
                if ( $lines[$i] =~ /^\d/ ) {
                    @values = split( /\s+/, $lines[$i] );
                    for ( my $j = 0; $j <= $#all_fields; $j++ ) {
                        $result_array[$line_count]->{ $all_fields[$j] } = $values[$j];
                    }
                    $line_count++;
                }
                $i++;
            }
        }
    }
    return \@result_array;
} ## end sub show_storm_control

## @name ::show_process_cpu
##
## @summary Parses the raw output from the "show process cpu" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from show_process_cpu with any of its options as described in the
##    corresponding cdef file 'show.cdefs'
## @synopsis
##      show_process_cpu(%args);
## INPUTS TO METHOD CALL 'show_process_cpu' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## <tr> <td> 'threshold' </td> <td> 'threshold' </td> <td> '1' </td> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples
##&nbsp;Command = show process cpu
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;
##&nbsp;Memory Utilization Report
##&nbsp;status      bytes
##&nbsp;------ ----------
##&nbsp;free    392675328
##&nbsp;alloc   137015296
##&nbsp;
##&nbsp; CPU Utilization:
##&nbsp;
##&nbsp;  PID      Name                    5 Secs     60 Secs    300 Secs
##&nbsp;-----------------------------------------------------------------
##&nbsp;  770    _interrupt_thread          0.00%       0.06%       0.04%
##&nbsp;  772    bcmL2X.0                   4.10%       4.02%       4.00%
##&nbsp;  773    bcmCNTR.0                  0.78%       0.77%       0.77%
##&nbsp;  774    bcmTX                      0.00%       0.00%       0.01%
##&nbsp;  776    bcmLINK.0                  3.12%       3.06%       3.05%
##&nbsp;  778    bcmRX                      0.00%       0.09%       0.05%
##&nbsp;  779    cpuUtilMonitorTask         0.19%       0.15%       0.15%
##&nbsp;  849    boxs Req                   0.00%       0.02%       0.01%
##&nbsp;  848    RMONTask                   4.49%       1.07%       0.79%
##&nbsp;  847    isdpTask                   0.19%       0.13%       0.14%
##&nbsp;  829    snoopTask                  0.00%       0.00%       0.01%
##&nbsp;  821    dot1s_task                 0.00%       0.00%       0.01%
##&nbsp;  820    dot1s_timer_task           0.00%       0.01%       0.01%
##&nbsp;  799    hapiRxTask                 0.00%       0.01%       0.01%
##&nbsp;  795    dtlAddrTask                0.19%       0.03%       0.00%
##&nbsp;  794    dtlTask                    0.19%       0.04%       0.02%
##&nbsp;  852    sshd                      15.62%       2.41%       0.60%
##&nbsp;  866    sshd[0]                    4.88%       0.75%       0.16%
##&nbsp;-----------------------------------------------------------------
##&nbsp; Total CPU Utilization             33.78%      12.69%       9.98%
##&nbsp;
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;          {
##&nbsp;            'total_cpu_utilization' => [
##&nbsp;                                         {
##&nbsp;                                           '300_secs' => '9.98%',
##&nbsp;                                           '5_secs' => '33.78%',
##&nbsp;                                           '60_secs' => '12.69%'
##&nbsp;                                         }
##&nbsp;                                       ],
##&nbsp;            'memory_utilization_report' => [
##&nbsp;                                             {
##&nbsp;                                               'alloc_in_bytes' => '137015296',
##&nbsp;                                               'free_in_bytes' => '392675328'
##&nbsp;                                             }
##&nbsp;                                           ],
##&nbsp;            'cpu_utilization' => [
##&nbsp;                                   {
##&nbsp;                                     'pid' => '770',
##&nbsp;                                     'name' => '_interrupt_thread',
##&nbsp;                                     '300_secs' => '0.04%',
##&nbsp;                                     '5_secs' => '0.00%',
##&nbsp;                                     '60_secs' => '0.06%'
##&nbsp;                                   },
##&nbsp;                                   {
##&nbsp;                                     'pid' => '772',
##&nbsp;                                     'name' => 'bcmL2X.0',
##&nbsp;                                     '300_secs' => '4.00%',
##&nbsp;                                     '5_secs' => '4.10%',
##&nbsp;                                     '60_secs' => '4.02%'
##&nbsp;                                   },
##&nbsp;                                   {
##&nbsp;                                     'pid' => '773',
##&nbsp;                                     'name' => 'bcmCNTR.0',
##&nbsp;                                     '300_secs' => '0.77%',
##&nbsp;                                     '5_secs' => '0.78%',
##&nbsp;                                     '60_secs' => '0.77%'
##&nbsp;                                   },
##&nbsp;                                   {
##&nbsp;                                     'pid' => '774',
##&nbsp;                                     'name' => 'bcmTX',
##&nbsp;                                     '300_secs' => '0.01%',
##&nbsp;                                     '5_secs' => '0.00%',
##&nbsp;                                     '60_secs' => '0.00%'
##&nbsp;                                   },
##&nbsp;                                   {
##&nbsp;                                     'pid' => '776',
##&nbsp;                                     'name' => 'bcmLINK.0',
##&nbsp;                                     '300_secs' => '3.05%',
##&nbsp;                                     '5_secs' => '3.12%',
##&nbsp;                                     '60_secs' => '3.06%'
##&nbsp;                                   },
##&nbsp;                                   {
##&nbsp;                                     'pid' => '778',
##&nbsp;                                     'name' => 'bcmRX',
##&nbsp;                                     '300_secs' => '0.05%',
##&nbsp;                                     '5_secs' => '0.00%',
##&nbsp;                                     '60_secs' => '0.09%'
##&nbsp;                                   },
##&nbsp;                                   {
##&nbsp;                                     'pid' => '779',
##&nbsp;                                     'name' => 'cpuUtilMonitorTask',
##&nbsp;                                     '300_secs' => '0.15%',
##&nbsp;                                     '5_secs' => '0.19%',
##&nbsp;                                     '60_secs' => '0.15%'
##&nbsp;                                   },
##&nbsp;                                   {
##&nbsp;                                     'pid' => '849',
##&nbsp;                                     'name' => 'boxs',
##&nbsp;                                     '300_secs' => '0.02%',
##&nbsp;                                     '5_secs' => 'Req',
##&nbsp;                                     '60_secs' => '0.00%'
##&nbsp;                                   },
##&nbsp;                                   {
##&nbsp;                                     'pid' => '848',
##&nbsp;                                     'name' => 'RMONTask',
##&nbsp;                                     '300_secs' => '0.79%',
##&nbsp;                                     '5_secs' => '4.49%',
##&nbsp;                                     '60_secs' => '1.07%'
##&nbsp;                                   },
##&nbsp;                                   {
##&nbsp;                                     'pid' => '847',
##&nbsp;                                     'name' => 'isdpTask',
##&nbsp;                                     '300_secs' => '0.14%',
##&nbsp;                                     '5_secs' => '0.19%',
##&nbsp;                                     '60_secs' => '0.13%'
##&nbsp;                                   },
##&nbsp;                                   {
##&nbsp;                                     'pid' => '829',
##&nbsp;                                     'name' => 'snoopTask',
##&nbsp;                                     '300_secs' => '0.01%',
##&nbsp;                                     '5_secs' => '0.00%',
##&nbsp;                                     '60_secs' => '0.00%'
##&nbsp;                                   },
##&nbsp;                                   {
##&nbsp;                                     'pid' => '821',
##&nbsp;                                     'name' => 'dot1s_task',
##&nbsp;                                     '300_secs' => '0.01%',
##&nbsp;                                     '5_secs' => '0.00%',
##&nbsp;                                     '60_secs' => '0.00%'
##&nbsp;                                   },
##&nbsp;                                   {
##&nbsp;                                     'pid' => '820',
##&nbsp;                                     'name' => 'dot1s_timer_task',
##&nbsp;                                     '300_secs' => '0.01%',
##&nbsp;                                     '5_secs' => '0.00%',
##&nbsp;                                     '60_secs' => '0.01%'
##&nbsp;                                   },
##&nbsp;                                   {
##&nbsp;                                     'pid' => '799',
##&nbsp;                                     'name' => 'hapiRxTask',
##&nbsp;                                     '300_secs' => '0.01%',
##&nbsp;                                     '5_secs' => '0.00%',
##&nbsp;                                     '60_secs' => '0.01%'
##&nbsp;                                   },
##&nbsp;                                   {
##&nbsp;                                     'pid' => '795',
##&nbsp;                                     'name' => 'dtlAddrTask',
##&nbsp;                                     '300_secs' => '0.00%',
##&nbsp;                                     '5_secs' => '0.19%',
##&nbsp;                                     '60_secs' => '0.03%'
##&nbsp;                                   },
##&nbsp;                                   {
##&nbsp;                                     'pid' => '794',
##&nbsp;                                     'name' => 'dtlTask',
##&nbsp;                                     '300_secs' => '0.02%',
##&nbsp;                                     '5_secs' => '0.19%',
##&nbsp;                                     '60_secs' => '0.04%'
##&nbsp;                                   },
##&nbsp;                                   {
##&nbsp;                                     'pid' => '852',
##&nbsp;                                     'name' => 'sshd',
##&nbsp;                                     '300_secs' => '0.60%',
##&nbsp;                                     '5_secs' => '15.62%',
##&nbsp;                                     '60_secs' => '2.41%'
##&nbsp;                                   },
##&nbsp;                                   {
##&nbsp;                                     'pid' => '866',
##&nbsp;                                     'name' => 'sshd[0]',
##&nbsp;                                     '300_secs' => '0.16%',
##&nbsp;                                     '5_secs' => '4.88%',
##&nbsp;                                     '60_secs' => '0.75%'
##&nbsp;                                   }
##&nbsp;                                 ]
##&nbsp;          }
##&nbsp;        ];
##&nbsp;
##&nbsp;Command = show process cpu threshold
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;
##&nbsp;CPU Utilization Monitoring Parameters
##&nbsp;Rising Threshold............................... 0 %
##&nbsp;Rising Interval................................ 0 secs
##&nbsp;Falling Threshold.............................. 0 %
##&nbsp;Falling Interval............................... 0 secs
##&nbsp;
##&nbsp;CPU Free Memory Monitoring Threshold........... 0 KB
##&nbsp;
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;          {
##&nbsp;            'cpu_utilization_monitoring_falling_threshold' => '0 %',
##&nbsp;            'cpu_utilization_monitoring_rising_interval' => '0 secs',
##&nbsp;            'cpu_utilization_monitoring_falling_interval' => '0 secs',
##&nbsp;            'cpu_free_memory_monitoring_threshold' => '0 KB',
##&nbsp;            'cpu_utilization_monitoring_rising_threshold' => '0 %'
##&nbsp;          }
##&nbsp;        ];
##&nbsp;
sub show_process_cpu {
    my %args = @_;

    my ($error_message);

    my $resp   = $args{response_object};
    my $output = $resp->get_processed_output();

    my $error = _check_error(
        raw_output    => $output,
        error_message => \$error_message
    );
    if ($error) {
        $error = 'NACL::APISet::Exceptions::' . $error . 'Exception';
        $error->throw( text => $error_message, response_object => $resp );
    }

    my @result_array;
    my ( $mem_util, @fields, @time_fields );
    my @lines = get_lines( raw_output => $output, response_object => $resp );
    for ( my $i = 0; $i <= $#lines; $i++ ) {
        if ( $lines[$i] =~ /(.*\w+)\)?\.+\s+(.*)$/ ) {
            my $temp1 = $1;
            my $temp2 = $2;
            $temp1 =~ s/-/_/;
            my $var = to_key_name( item => $temp1 );
            if ( $lines[$i] !~ /CPU Free Memory Monitoring/ ) {
                $var = 'cpu_utilization_monitoring_' . $var;
            }
            $result_array[0]->{$var} = $temp2;
        }
        if ( $lines[$i] =~ /(Memory Utilization Report)/ ) {
            @fields = split( /\s+/, $lines[ $i + 1 ] );
            $mem_util = to_key_name( item => $1 );
        }
        if ( $lines[$i] =~ /^free\s+([0-9]+)$/ ) {
            $result_array[0]->{$mem_util}->[0]->{"free_in_$fields[1]"} = "$1";
        }
        if ( $lines[$i] =~ /^alloc\s+([0-9]+)$/ ) {
            $result_array[0]->{$mem_util}->[0]->{"alloc_in_$fields[1]"} = "$1";
        }
        my ( $cpu_util, @all_fields, $pos );
        if ( $lines[$i] =~ /(CPU Utilization):/ ) {
            $cpu_util = to_key_name( item => $1 );
            $pos = 0;

            while ( $i <= $#lines ) {
                if ( $lines[$i] =~ /PID\s+Name\s+/ ) {
                    @all_fields  = split( /\s\s+/, $lines[$i] );
                    @time_fields = split( /\s\s+/, $' );
                }
                my %temp;
                if ( $lines[$i] =~ /^[0-9]/ ) {
                    @fields = split( /\s+/, $lines[$i] );
                    for ( my $j = 0; $j <= $#all_fields; $j++ ) {
                        $temp{ to_key_name( item => $all_fields[$j] ) } = $fields[$j];
                    }
                    $result_array[0]->{$cpu_util}->[$pos] = \%temp;
                    $pos++;
                }
                if ( $lines[$i] =~ /(Total CPU Utilization)\s+(\d.*%)\s+(\d.*%)\s+(\d.*%)/ ) {
                    %temp = ();
                    $cpu_util = to_key_name( item => $1 );
                    my @details = ( $2, $3, $4 );
                    for ( my $k = 0; $k <= $#time_fields; $k++ ) {
                        $temp{ to_key_name( item => $time_fields[$k] ) } = $details[$k];
                    }
                    $result_array[0]->{$cpu_util}->[0] = \%temp;
                }
                $i++;
            }
        }
    }
    return \@result_array;
} ## end sub show_process_cpu

## @name ::show_isdp
##
## @summary Parses the raw output from the "show isdp" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from show_isdp with any of its options as described in the
##    corresponding cdef file 'show.cdefs'
## @synopsis
##      show_isdp(%args);
## INPUTS TO METHOD CALL 'show_isdp' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

##&nbsp;Command = show isdp
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;
##&nbsp;Timer.......................................... 5
##&nbsp;Hold Time...................................... 180
##&nbsp;Version 2 Advertisements....................... Enabled
##&nbsp;Neighbors table time since last change......... 39 days 14:29:10
##&nbsp;Device ID...................................... net-1960-1
##&nbsp;Device ID format capability.................... Serial Number, Host Name
##&nbsp;Device ID format............................... Host Name
##&nbsp;
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;          {
##&nbsp;            'device_id_format_capability' => 'Serial Number, Host Name',
##&nbsp;            'device_id' => 'net-1960-1',
##&nbsp;            'device_id_format' => 'Host Name',
##&nbsp;            'version_2_advertisements' => 'Enabled',
##&nbsp;            'neighbors_table_time_since_last_change' => '39 days 14:29:10',
##&nbsp;            'timer' => '5',
##&nbsp;            'hold_time' => '180'
##&nbsp;          }
##&nbsp;        ];
##&nbsp;

sub show_isdp {
    my %args = @_;
    my ( $error_message, @result_array );

    my $resp   = $args{response_object};
    my $output = $resp->get_processed_output();

    my $error = _check_error(
        raw_output    => $output,
        error_message => \$error_message
    );
    if ($error) {
        $error = 'NACL::APISet::Exceptions::' . $error . 'Exception';
        $error->throw( text => $error_message, response_object => $resp );
    }
    my @lines = get_lines( raw_output => $output, response_object => $resp );
    for ( my $i = 0; $i <= $#lines; $i++ ) {
        if ( $lines[$i] =~ /(.*\w+)\)?\.+\s+(.*)$/ ) {
            my $temp1 = $1;
            $temp1 =~ s/-/_/;
            $result_array[0]->{ to_key_name( item => $temp1 ) } = $2;
        }
    }
    return \@result_array;
} ## end sub show_isdp

## @name ::show_isdp_neighbors
##
## @summary Parses the raw output from the "show isdp neighbors" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from show_isdp_neighbors with any of its options as described in the
##    corresponding cdef file 'show.cdefs'
## @synopsis
##      show_isdp_neighbors(%args);
## INPUTS TO METHOD CALL 'show_isdp_neighbors' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## <tr> <td> 'detail' </td> <td> 'detail' </td> <td> '1' </td> </tr>
## <tr> <td> No switch for this key </td> <td> 'interface' </td> <td> '0' </td> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

##&nbsp;Command = show isdp neighbors
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;
##&nbsp;Capability Codes: R - Router, T - Trans Bridge, B - Source Route Bridge,
##&nbsp;                  S - Switch, H - Host, I - IGMP, r - Repeater
##&nbsp;Device ID              Intf         Holdtime  Capability   Platform             Port ID
##&nbsp;---------------------- ------------ --------- ------------ -------------------- ------------
##&nbsp;net-spec-8a            0/1          150       H            FAS3140              e1a
##&nbsp;net-spec-8b            0/2          166       H            FAS3140              e1a
##&nbsp;net-vespa-5a           0/4          152       H            FAS2240-2            e1a
##&nbsp;net-absm-1b            0/5          142       H            FAS6280              e0c
##&nbsp;net-absm-1a            0/6          149       H            FAS6280              e0c
##&nbsp;net-car-1a             0/7          144       H            FAS3210              e1a
##&nbsp;net-car-1b             0/8          139       H            FAS3210              e1a
##&nbsp;net-car-11a            0/10         142       H            FAS3210              e1a
##&nbsp;net-car-11b            0/11         126       H            FAS3210              e2a
##&nbsp;
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;          {
##&nbsp;            'device_id' => 'net-spec-8a',
##&nbsp;            'holdtime' => '150',
##&nbsp;            'port_id' => 'e1a',
##&nbsp;            'interface' => '0/1',
##&nbsp;            'platform' => 'FAS3140',
##&nbsp;            'capability' => [
##&nbsp;                              'Host'
##&nbsp;                            ]
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'device_id' => 'net-spec-8b',
##&nbsp;            'holdtime' => '166',
##&nbsp;            'port_id' => 'e1a',
##&nbsp;            'interface' => '0/2',
##&nbsp;            'platform' => 'FAS3140',
##&nbsp;            'capability' => [
##&nbsp;                              'Host'
##&nbsp;                            ]
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'device_id' => 'net-vespa-5a',
##&nbsp;            'holdtime' => '152',
##&nbsp;            'port_id' => 'e1a',
##&nbsp;            'interface' => '0/4',
##&nbsp;            'platform' => 'FAS2240-2',
##&nbsp;            'capability' => [
##&nbsp;                              'Host'
##&nbsp;                            ]
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'device_id' => 'net-absm-1b',
##&nbsp;            'holdtime' => '142',
##&nbsp;            'port_id' => 'e0c',
##&nbsp;            'interface' => '0/5',
##&nbsp;            'platform' => 'FAS6280',
##&nbsp;            'capability' => [
##&nbsp;                              'Host'
##&nbsp;                            ]
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'device_id' => 'net-absm-1a',
##&nbsp;            'holdtime' => '149',
##&nbsp;            'port_id' => 'e0c',
##&nbsp;            'interface' => '0/6',
##&nbsp;            'platform' => 'FAS6280',
##&nbsp;            'capability' => [
##&nbsp;                              'Host'
##&nbsp;                            ]
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'device_id' => 'net-car-1a',
##&nbsp;            'holdtime' => '144',
##&nbsp;            'port_id' => 'e1a',
##&nbsp;            'interface' => '0/7',
##&nbsp;            'platform' => 'FAS3210',
##&nbsp;            'capability' => [
##&nbsp;                              'Host'
##&nbsp;                            ]
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'device_id' => 'net-car-1b',
##&nbsp;            'holdtime' => '139',
##&nbsp;            'port_id' => 'e1a',
##&nbsp;            'interface' => '0/8',
##&nbsp;            'platform' => 'FAS3210',
##&nbsp;            'capability' => [
##&nbsp;                              'Host'
##&nbsp;                            ]
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'device_id' => 'net-car-11a',
##&nbsp;            'holdtime' => '142',
##&nbsp;            'port_id' => 'e1a',
##&nbsp;            'interface' => '0/10',
##&nbsp;            'platform' => 'FAS3210',
##&nbsp;            'capability' => [
##&nbsp;                              'Host'
##&nbsp;                            ]
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'device_id' => 'net-car-11b',
##&nbsp;            'holdtime' => '126',
##&nbsp;            'port_id' => 'e2a',
##&nbsp;            'interface' => '0/11',
##&nbsp;            'platform' => 'FAS3210',
##&nbsp;            'capability' => [
##&nbsp;                              'Host'
##&nbsp;                            ]
##&nbsp;          }
##&nbsp;        ];
##&nbsp;
##&nbsp;Command = show isdp neighbors detail
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;
##&nbsp;Device ID                       net-spec-8a
##&nbsp;Address(es):
##&nbsp;Capability                      Host
##&nbsp;Platform                        FAS3140
##&nbsp;Interface                       0/1
##&nbsp;Port ID                         e1a
##&nbsp;Holdtime                        149
##&nbsp;Advertisement Version           1
##&nbsp;Time since entry last changed   48 days 06:45:00
##&nbsp;Version :
##&nbsp;NetApp Release mainN_120531_0030 Cluster-Mode: Thu May 31 00:58:46 PDT 2012
##&nbsp;
##&nbsp;Device ID                       net-spec-8b
##&nbsp;Address(es):
##&nbsp;    IP Address:                 169.254.155.162
##&nbsp;    IP Address:                 169.254.48.27
##&nbsp;Capability                      Host
##&nbsp;Platform                        FAS3140
##&nbsp;Interface                       0/2
##&nbsp;Port ID                         e1a
##&nbsp;Holdtime                        165
##&nbsp;Advertisement Version           1
##&nbsp;Time since entry last changed   48 days 06:45:17
##&nbsp;Version :
##&nbsp;NetApp Release mainN_120531_0030 Cluster-Mode: Thu May 31 00:58:46 PDT 2012
##&nbsp;
##&nbsp;
##&nbsp;--More-- or (q)uit
##&nbsp;
##&nbsp;
##&nbsp;Device ID                       net-vespa-5a
##&nbsp;Address(es):
##&nbsp;    IP Address:                 169.254.148.56
##&nbsp;    IP Address:                 169.254.3.156
##&nbsp;Capability                      Host
##&nbsp;Platform                        FAS2240-2
##&nbsp;Interface                       0/4
##&nbsp;Port ID                         e1a
##&nbsp;Holdtime                        151
##&nbsp;Advertisement Version           1
##&nbsp;Time since entry last changed   48 days 06:45:03
##&nbsp;Version :
##&nbsp;NetApp Release mainN_120531_0030 Cluster-Mode: Thu May 31 00:58:46 PDT 2012
##&nbsp;
##&nbsp;Device ID                       net-absm-1b
##&nbsp;Address(es):
##&nbsp;    IP Address:                 169.254.156.211
##&nbsp;    IP Address:                 169.254.228.174
##&nbsp;Capability                      Host
##&nbsp;Platform                        FAS6280
##&nbsp;Interface                       0/5
##&nbsp;Port ID                         e0c
##&nbsp;Holdtime                        141
##&nbsp;Advertisement Version           1
##&nbsp;Time since entry last changed   48 days 06:44:53
##&nbsp;Version :
##&nbsp;NetApp Release mainN_120531_0030 Cluster-Mode: Thu May 31 00:58:46 PDT 2012
##&nbsp;
##&nbsp;
##&nbsp;--More-- or (q)uit
##&nbsp;
##&nbsp;
##&nbsp;Device ID                       net-absm-1a
##&nbsp;Address(es):
##&nbsp;    IP Address:                 169.254.21.130
##&nbsp;    IP Address:                 169.254.78.190
##&nbsp;Capability                      Host
##&nbsp;Platform                        FAS6280
##&nbsp;Interface                       0/6
##&nbsp;Port ID                         e0c
##&nbsp;Holdtime                        148
##&nbsp;Advertisement Version           1
##&nbsp;Time since entry last changed   48 days 06:44:59
##&nbsp;Version :
##&nbsp;NetApp Release mainN_120531_0030 Cluster-Mode: Thu May 31 00:58:46 PDT 2012
##&nbsp;
##&nbsp;Device ID                       net-car-1a
##&nbsp;Address(es):
##&nbsp;    IP Address:                 169.254.144.69
##&nbsp;    IP Address:                 169.254.46.181
##&nbsp;Capability                      Host
##&nbsp;Platform                        FAS3210
##&nbsp;Interface                       0/7
##&nbsp;Port ID                         e1a
##&nbsp;Holdtime                        143
##&nbsp;Advertisement Version           1
##&nbsp;Time since entry last changed   48 days 06:44:55
##&nbsp;Version :
##&nbsp;NetApp Release mainN_120531_0030 Cluster-Mode: Thu May 31 00:58:46 PDT 2012
##&nbsp;
##&nbsp;
##&nbsp;--More-- or (q)uit
##&nbsp;
##&nbsp;
##&nbsp;Device ID                       net-car-1b
##&nbsp;Address(es):
##&nbsp;    IP Address:                 169.254.185.198
##&nbsp;    IP Address:                 169.254.159.100
##&nbsp;Capability                      Host
##&nbsp;Platform                        FAS3210
##&nbsp;Interface                       0/8
##&nbsp;Port ID                         e1a
##&nbsp;Holdtime                        137
##&nbsp;Advertisement Version           1
##&nbsp;Time since entry last changed   48 days 06:44:49
##&nbsp;Version :
##&nbsp;NetApp Release mainN_120531_0030 Cluster-Mode: Thu May 31 00:58:46 PDT 2012
##&nbsp;
##&nbsp;Device ID                       net-car-11a
##&nbsp;Address(es):
##&nbsp;    IP Address:                 169.254.97.235
##&nbsp;Capability                      Host
##&nbsp;Platform                        FAS3210
##&nbsp;Interface                       0/10
##&nbsp;Port ID                         e1a
##&nbsp;Holdtime                        140
##&nbsp;Advertisement Version           1
##&nbsp;Time since entry last changed   48 days 06:44:52
##&nbsp;Version :
##&nbsp;NetApp Release mainN_120625_0835 Cluster-Mode: Mon Jun 25 08:56:36 PDT 2012
##&nbsp;
##&nbsp;
##&nbsp;--More-- or (q)uit
##&nbsp;
##&nbsp;
##&nbsp;Device ID                       net-car-11b
##&nbsp;Address(es):
##&nbsp;    IP Address:                 169.254.78.34
##&nbsp;    IP Address:                 169.254.222.66
##&nbsp;Capability                      Host
##&nbsp;Platform                        FAS3210
##&nbsp;Interface                       0/11
##&nbsp;Port ID                         e2a
##&nbsp;Holdtime                        124
##&nbsp;Advertisement Version           1
##&nbsp;Time since entry last changed   48 days 06:44:37
##&nbsp;Version :
##&nbsp;NetApp Release mainN_120625_0835 Cluster-Mode: Mon Jun 25 08:56:36 PDT 2012
##&nbsp;
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;          {
##&nbsp;            'holdtime' => '149',
##&nbsp;            'version' => 'NetApp Release mainN_120531_0030 Cluster-Mode: Thu May 31 00:58:46 PDT 2012',
##&nbsp;            'advertisement_version' => '1',
##&nbsp;            'port_id' => 'e1a',
##&nbsp;            'time_since_entry_last_changed' => '48 days 06:45:00',
##&nbsp;            'capability' => [
##&nbsp;                              'Host'
##&nbsp;                            ],
##&nbsp;            'device_id' => 'net-spec-8a',
##&nbsp;            'interface' => '0/1',
##&nbsp;            'platform' => 'FAS3140'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'holdtime' => '165',
##&nbsp;            'version' => 'NetApp Release mainN_120531_0030 Cluster-Mode: Thu May 31 00:58:46 PDT 2012',
##&nbsp;            'advertisement_version' => '1',
##&nbsp;            'port_id' => 'e1a',
##&nbsp;            'time_since_entry_last_changed' => '48 days 06:45:17',
##&nbsp;            'capability' => [
##&nbsp;                              'Host'
##&nbsp;                            ],
##&nbsp;            'device_id' => 'net-spec-8b',
##&nbsp;            'addresses' => [
##&nbsp;                             '169.254.155.162',
##&nbsp;                             '169.254.48.27'
##&nbsp;                           ],
##&nbsp;            'interface' => '0/2',
##&nbsp;            'platform' => 'FAS3140'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'holdtime' => '151',
##&nbsp;            'version' => 'NetApp Release mainN_120531_0030 Cluster-Mode: Thu May 31 00:58:46 PDT 2012',
##&nbsp;            'advertisement_version' => '1',
##&nbsp;            'port_id' => 'e1a',
##&nbsp;            'time_since_entry_last_changed' => '48 days 06:45:03',
##&nbsp;            'capability' => [
##&nbsp;                              'Host'
##&nbsp;                            ],
##&nbsp;            'device_id' => 'net-vespa-5a',
##&nbsp;            'addresses' => [
##&nbsp;                             '169.254.148.56',
##&nbsp;                             '169.254.3.156'
##&nbsp;                           ],
##&nbsp;            'interface' => '0/4',
##&nbsp;            'platform' => 'FAS2240-2'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'holdtime' => '141',
##&nbsp;            'version' => 'NetApp Release mainN_120531_0030 Cluster-Mode: Thu May 31 00:58:46 PDT 2012',
##&nbsp;            'advertisement_version' => '1',
##&nbsp;            'port_id' => 'e0c',
##&nbsp;            'time_since_entry_last_changed' => '48 days 06:44:53',
##&nbsp;            'capability' => [
##&nbsp;                              'Host'
##&nbsp;                            ],
##&nbsp;            'device_id' => 'net-absm-1b',
##&nbsp;            'addresses' => [
##&nbsp;                             '169.254.156.211',
##&nbsp;                             '169.254.228.174'
##&nbsp;                           ],
##&nbsp;            'interface' => '0/5',
##&nbsp;            'platform' => 'FAS6280'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'holdtime' => '148',
##&nbsp;            'version' => 'NetApp Release mainN_120531_0030 Cluster-Mode: Thu May 31 00:58:46 PDT 2012',
##&nbsp;            'advertisement_version' => '1',
##&nbsp;            'port_id' => 'e0c',
##&nbsp;            'time_since_entry_last_changed' => '48 days 06:44:59',
##&nbsp;            'capability' => [
##&nbsp;                              'Host'
##&nbsp;                            ],
##&nbsp;            'device_id' => 'net-absm-1a',
##&nbsp;            'addresses' => [
##&nbsp;                             '169.254.21.130',
##&nbsp;                             '169.254.78.190'
##&nbsp;                           ],
##&nbsp;            'interface' => '0/6',
##&nbsp;            'platform' => 'FAS6280'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'holdtime' => '143',
##&nbsp;            'version' => 'NetApp Release mainN_120531_0030 Cluster-Mode: Thu May 31 00:58:46 PDT 2012',
##&nbsp;            'advertisement_version' => '1',
##&nbsp;            'port_id' => 'e1a',
##&nbsp;            'time_since_entry_last_changed' => '48 days 06:44:55',
##&nbsp;            'capability' => [
##&nbsp;                              'Host'
##&nbsp;                            ],
##&nbsp;            'device_id' => 'net-car-1a',
##&nbsp;            'addresses' => [
##&nbsp;                             '169.254.144.69',
##&nbsp;                             '169.254.46.181'
##&nbsp;                           ],
##&nbsp;            'interface' => '0/7',
##&nbsp;            'platform' => 'FAS3210'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'holdtime' => '137',
##&nbsp;            'version' => 'NetApp Release mainN_120531_0030 Cluster-Mode: Thu May 31 00:58:46 PDT 2012',
##&nbsp;            'advertisement_version' => '1',
##&nbsp;            'port_id' => 'e1a',
##&nbsp;            'time_since_entry_last_changed' => '48 days 06:44:49',
##&nbsp;            'capability' => [
##&nbsp;                              'Host'
##&nbsp;                            ],
##&nbsp;            'device_id' => 'net-car-1b',
##&nbsp;            'addresses' => [
##&nbsp;                             '169.254.185.198',
##&nbsp;                             '169.254.159.100'
##&nbsp;                           ],
##&nbsp;            'interface' => '0/8',
##&nbsp;            'platform' => 'FAS3210'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'holdtime' => '140',
##&nbsp;            'version' => 'NetApp Release mainN_120625_0835 Cluster-Mode: Mon Jun 25 08:56:36 PDT 2012',
##&nbsp;            'advertisement_version' => '1',
##&nbsp;            'port_id' => 'e1a',
##&nbsp;            'time_since_entry_last_changed' => '48 days 06:44:52',
##&nbsp;            'capability' => [
##&nbsp;                              'Host'
##&nbsp;                            ],
##&nbsp;            'device_id' => 'net-car-11a',
##&nbsp;            'addresses' => [
##&nbsp;                             '169.254.97.235'
##&nbsp;                           ],
##&nbsp;            'interface' => '0/10',
##&nbsp;            'platform' => 'FAS3210'
##&nbsp;          },
##&nbsp;          {
##&nbsp;            'holdtime' => '124',
##&nbsp;            'version' => 'NetApp Release mainN_120625_0835 Cluster-Mode: Mon Jun 25 08:56:36 PDT 2012',
##&nbsp;            'advertisement_version' => '1',
##&nbsp;            'port_id' => 'e2a',
##&nbsp;            'time_since_entry_last_changed' => '48 days 06:44:37',
##&nbsp;            'capability' => [
##&nbsp;                              'Host'
##&nbsp;                            ],
##&nbsp;            'device_id' => 'net-car-11b',
##&nbsp;            'addresses' => [
##&nbsp;                             '169.254.78.34',
##&nbsp;                             '169.254.222.66'
##&nbsp;                           ],
##&nbsp;            'interface' => '0/11',
##&nbsp;            'platform' => 'FAS3210'
##&nbsp;          }
##&nbsp;        ];
##&nbsp;
##&nbsp;Command = show isdp neighbors 0/1
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;
##&nbsp;Capability Codes: R - Router, T - Trans Bridge, B - Source Route Bridge,
##&nbsp;                  S - Switch, H - Host, I - IGMP, r - Repeater
##&nbsp;Device ID              Intf         Holdtime  Capability   Platform             Port ID
##&nbsp;---------------------- ------------ --------- ------------ -------------------- ------------
##&nbsp;net-spec-8a            0/1          148       H            FAS3140              e1a
##&nbsp;
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;          {
##&nbsp;            'device_id' => 'net-spec-8a',
##&nbsp;            'holdtime' => '148',
##&nbsp;            'port_id' => 'e1a',
##&nbsp;            'interface' => '0/1',
##&nbsp;            'platform' => 'FAS3140',
##&nbsp;            'capability' => [
##&nbsp;                              'Host'
##&nbsp;                            ]
##&nbsp;          }
##&nbsp;        ];
##&nbsp;
##&nbsp;Command = show isdp neighbors 0/1 detail
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;
##&nbsp;Device ID                       net-spec-8a
##&nbsp;Address(es):
##&nbsp;Capability                      Host
##&nbsp;Platform                        FAS3140
##&nbsp;Interface                       0/1
##&nbsp;Port ID                         e1a
##&nbsp;Holdtime                        147
##&nbsp;Advertisement Version           1
##&nbsp;Time since entry last changed   48 days 06:45:00
##&nbsp;Version :
##&nbsp;NetApp Release mainN_120531_0030 Cluster-Mode: Thu May 31 00:58:46 PDT 2012
##&nbsp;
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;          {
##&nbsp;            'holdtime' => '147',
##&nbsp;            'version' => 'NetApp Release mainN_120531_0030 Cluster-Mode: Thu May 31 00:58:46 PDT 2012',
##&nbsp;            'advertisement_version' => '1',
##&nbsp;            'port_id' => 'e1a',
##&nbsp;            'time_since_entry_last_changed' => '48 days 06:45:00',
##&nbsp;            'capability' => [
##&nbsp;                              'Host'
##&nbsp;                            ],
##&nbsp;            'device_id' => 'net-spec-8a',
##&nbsp;            'interface' => '0/1',
##&nbsp;            'platform' => 'FAS3140'
##&nbsp;          }
##&nbsp;        ];
##&nbsp;
sub show_isdp_neighbors {
    my %args = @_;
    my ( $error_message, @result_array );
    my $resp    = $args{response_object};
    my $output  = $resp->get_processed_output();
    my $command = $resp->get_command();
    my $error   = _check_error(
        raw_output    => $output,
        error_message => \$error_message
    );
    if ($error) {
        $error = 'NACL::APISet::Exceptions::' . $error . 'Exception';
        $error->throw( text => $error_message, response_object => $resp );
    }
    my @lines = get_lines( raw_output => $output, response_object => $resp );
    my ( @head, @values, @codes, %capab_codes );
    my $count = 0;
    my $pos   = -1;
    for ( my $i = 0; $i <= $#lines; $i++ ) {
        if ( $lines[$i] =~ /Capability Codes: / ) {
            my @temp = split( /,\s+/, $' );
            push( @codes, @temp );
            @temp = split( /,\s+/, $lines[ $i + 1 ] );
            push( @codes, @temp );
            foreach my $cod (@codes) {
                @temp = split( /\s+-\s+/, $cod );
                $capab_codes{ $temp[0] } = $temp[1];
            }
            $i = $i + 2;
        }
        if ( $lines[$i] =~ /Device ID\s+Intf/ ) {
            $lines[$i] =~ s/(\w+)\s+ID/$1_id/g;
            $lines[$i] =~ s/Intf/interface/g;
            @head = split( /\s+/, $lines[$i] );
            $i++;
        }
        if ( $lines[$i] =~ /\w+\s+\d\/\d+\s+/ ) {
            $lines[$i] =~ s/cisco\s+\w+/cisco_\w/g;
            $lines[$i] =~ s/([R|T|B|S|H|I|r])(\s+){1,7}/$1_/g;
            my @capabs;
            if ( $lines[$i] =~ /(([R|T|B|S|H|I|r]_){1,7})/ ) {
                @capabs = split( /_/, $1 );
                $lines[$i] =~ s/(([R|T|B|S|H|I|r]_){1,7})/-- /g;
            }
            @values = split( /\s+/, $lines[$i] );
            for ( my $j = 0; $j <= $#head; $j++ ) {
                $result_array[$count]->{ to_key_name( item => $head[$j] ) } = $values[$j]
                    if ( $head[$j] ne "Capability" );
            }
            for ( my $k = 0; $k <= $#capabs; $k++ ) {
                foreach my $key ( keys(%capab_codes) ) {
                    if ( $capabs[$k] eq $key ) {
                        $capabs[$k] = $capab_codes{$key};
                    }
                }
            }
            push( @{ $result_array[$count]->{capability} }, @capabs );
            $count++;
        }
        if ( $command =~ /detail/ ) {
            if ( $lines[$i] =~ /:\s+/ ) {
                @head = split( /:\s+/, $lines[$i] );
            } else {
                @head = split( /\s\s+/, $lines[$i] );
            }
            $pos++ if ( $head[0] =~ /Device ID/ );
            if ( $lines[$i] =~ /Version :/ ) {
                $result_array[$pos]->{version} = $lines[ $i + 1 ];
                $i++;
            }
            my $c = 0;
            if (   $lines[$i] =~ /Address\(es\)/
                && $lines[ $i + 1 ] =~ /IP Address/ )
            {
                $i++;
                while ( $lines[$i] =~ /IP Address:\s+(.*)/ ) {
                    $i++;
                    $result_array[$pos]->{addresses}->[$c] = $1;
                    $c++;
                }
                $i--;
            }
            my @cap;
            if ( $lines[$i] =~ /Capability/ ) {
                @cap = split( /\s+/, $head[1] );
            }
            chomp( $head[0] );
            $result_array[$pos]->{ to_key_name( item => $head[0] ) } = $head[1]
                if ( $head[1] && $head[0] ne "Capability" );
            push( @{ $result_array[$pos]->{capability} }, @cap );
        }
    }
    return \@result_array;
} ## end sub show_isdp_neighbors

## @name ::show_tech_support
##
## @summary Parses the raw output from the "show tech-support" command
##
## @description
##    This method can be used for parsing the raw output obtained after calling
##    the method from show_tech_support with any of its options as described in the
##    corresponding cdef file 'show.cdefs'
## @synopsis
##      show_tech_support(%args);
## INPUTS TO METHOD CALL 'show_tech_support' ON APISET OBJECT
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples

##&nbsp;Command = show tech-support
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;
##&nbsp; ***************** Show Version ******************
##&nbsp;
##&nbsp;Switch: 1
##&nbsp;
##&nbsp;System Description............................. NetApp CN1610, 1.1.0.1, Linux
##&nbsp;                                                2.6.21.7
##&nbsp;Machine Type................................... NetApp CN1610
##&nbsp;Machine Model.................................. CN1610
##&nbsp;Serial Number.................................. 11111200069
##&nbsp;FRU Number..................................... 111-00893
##&nbsp;Part Number.................................... BCM56820
##&nbsp;Maintenance Level.............................. A
##&nbsp;Manufacturer................................... 0xbc00
##&nbsp;Burned In MAC Address.......................... 00:A0:98:4B:B7:F1
##&nbsp;Software Version............................... 1.1.0.1
##&nbsp;Operating System............................... Linux 2.6.21.7
##&nbsp;Network Processing Device...................... BCM56820_B0
##&nbsp;Additional Packages............................ FASTPATH QOS
##&nbsp;--More-- or (q)uit
##&nbsp;
##&nbsp;
##&nbsp; ***************** Show SysInfo ******************
##&nbsp;
##&nbsp;
##&nbsp; ***************** Show Running Config ***********
##&nbsp;!Current Configuration:
##&nbsp;!
##&nbsp;!System Description "NetApp CN1610, 1.1.0.1, Linux 2.6.21.7"
##&nbsp;!System Software Version "1.1.0.1"
##&nbsp;!System Up Time          "62 days 9 hrs 8 mins 51 secs"
##&nbsp;!Additional Packages     FASTPATH QOS
##&nbsp;!Current SNTP Synchronized Time: Apr 25 09:21:29 201 UTC
##&nbsp;!
##&nbsp;set prompt "net-1960-1"
##&nbsp;telnetcon timeout 60
##&nbsp;serviceport protocol none
##&nbsp;serviceport ip 172.22.200.93 255.255.254.0 172.22.200.1
##&nbsp;vlan database
##&nbsp;exit
##&nbsp;ip ssh server enable
##&nbsp;sshcon timeout 60
##&nbsp;configure
##&nbsp;sntp client mode unicast
##&nbsp;! sntp server status is Success
##&nbsp;sntp server "172.19.2.16"
##&nbsp;clock timezone -8 minutes 0 zone PST
##&nbsp;
##&nbsp;--More-- or (q)uit
##&nbsp;
##&nbsp;ip domain name "lab.netapp.com"
##&nbsp;ip name server 172.22.201.46
##&nbsp;username "admin" password 54efa5bc56b147ac920e6a5cbba5dfe76bd81744dd12ffe9141907f7a6f87a9db9fa55b14573ebe19c8654d0b98a1c393c5635b5ced4d40d1e16f5cdd6e3a991 level 15 encrypted
##&nbsp;line console
##&nbsp;serial timeout 60
##&nbsp;exit
##&nbsp;line telnet
##&nbsp;exit
##&nbsp;line ssh
##&nbsp;exit
##&nbsp;spanning-tree configuration name "00-A0-98-4B-B7-F1"
##&nbsp;spanning-tree bpduguard
##&nbsp;port-channel name 3/1 ISL-LAG
##&nbsp;interface 0/13
##&nbsp;no auto-negotiate
##&nbsp;addport 3/1
##&nbsp;exit
##&nbsp;interface 0/14
##&nbsp;no auto-negotiate
##&nbsp;addport 3/1
##&nbsp;exit
##&nbsp;
##&nbsp;--More-- or (q)uit
##&nbsp;
##&nbsp;interface 0/15
##&nbsp;no auto-negotiate
##&nbsp;addport 3/1
##&nbsp;exit
##&nbsp;interface 0/16
##&nbsp;no auto-negotiate
##&nbsp;addport 3/1
##&nbsp;exit
##&nbsp;snmp-server sysname "net-1960-1"
##&nbsp;!
##&nbsp;snmp-server community "cshm1!"
##&nbsp;snmp-server community "netapp"
##&nbsp;snmp-server community "abc1"
##&nbsp;interface 0/1
##&nbsp;no auto-negotiate
##&nbsp;storm-control broadcast level 3
##&nbsp;storm-control multicast level 3
##&nbsp;storm-control flowcontrol
##&nbsp;storm-control unicast level 3
##&nbsp;description 'Cluster Node 1'
##&nbsp;spanning-tree edgeport
##&nbsp;
##&nbsp;--More-- or (q)uit
##&nbsp;
##&nbsp;mtu 9216
##&nbsp;exit
##&nbsp;interface 0/2
##&nbsp;no auto-negotiate
##&nbsp;storm-control flowcontrol
##&nbsp;description 'Cluster Node 2'
##&nbsp;spanning-tree edgeport
##&nbsp;mtu 9216
##&nbsp;exit
##&nbsp;interface 0/3
##&nbsp;no auto-negotiate
##&nbsp;description 'Cluster Node 3'
##&nbsp;spanning-tree edgeport
##&nbsp;mtu 9216
##&nbsp;exit
##&nbsp;interface 0/4
##&nbsp;no auto-negotiate
##&nbsp;description 'Cluster Node 4'
##&nbsp;spanning-tree edgeport
##&nbsp;mtu 9216
##&nbsp;exit
##&nbsp;
##&nbsp;--More-- or (q)uit
##&nbsp;
##&nbsp;interface 0/5
##&nbsp;no auto-negotiate
##&nbsp;description 'Cluster Node 5'
##&nbsp;spanning-tree edgeport
##&nbsp;mtu 9216
##&nbsp;exit
##&nbsp;interface 0/6
##&nbsp;no auto-negotiate
##&nbsp;description 'Cluster Node 6'
##&nbsp;spanning-tree edgeport
##&nbsp;mtu 9216
##&nbsp;exit
##&nbsp;interface 0/7
##&nbsp;no auto-negotiate
##&nbsp;description 'Cluster Node 7'
##&nbsp;spanning-tree edgeport
##&nbsp;mtu 9216
##&nbsp;exit
##&nbsp;interface 0/8
##&nbsp;no auto-negotiate
##&nbsp;description 'Cluster Node 8'
##&nbsp;
##&nbsp;--More-- or (q)uit
##&nbsp;
##&nbsp;spanning-tree edgeport
##&nbsp;mtu 9216
##&nbsp;exit
##&nbsp;interface 0/9
##&nbsp;no auto-negotiate
##&nbsp;description 'Cluster Node 9'
##&nbsp;spanning-tree edgeport
##&nbsp;mtu 9216
##&nbsp;exit
##&nbsp;interface 0/10
##&nbsp;no auto-negotiate
##&nbsp;description 'Cluster Node 10'
##&nbsp;spanning-tree edgeport
##&nbsp;mtu 9216
##&nbsp;exit
##&nbsp;interface 0/11
##&nbsp;no auto-negotiate
##&nbsp;description 'Cluster Node 11'
##&nbsp;spanning-tree edgeport
##&nbsp;mtu 9216
##&nbsp;exit
##&nbsp;
##&nbsp;--More-- or (q)uit
##&nbsp;
##&nbsp;interface 0/12
##&nbsp;no auto-negotiate
##&nbsp;description 'Cluster Node 12'
##&nbsp;spanning-tree edgeport
##&nbsp;mtu 9216
##&nbsp;exit
##&nbsp;interface 0/13
##&nbsp;description 'ISL member port 13'
##&nbsp;mtu 9216
##&nbsp;exit
##&nbsp;interface 0/14
##&nbsp;description 'ISL member port 14'
##&nbsp;mtu 9216
##&nbsp;exit
##&nbsp;interface 0/15
##&nbsp;description 'ISL member port 15'
##&nbsp;mtu 9216
##&nbsp;exit
##&nbsp;interface 0/16
##&nbsp;description 'ISL member port 16'
##&nbsp;mtu 9216
##&nbsp;
##&nbsp;--More-- or (q)uit
##&nbsp;
##&nbsp;exit
##&nbsp;interface 3/1
##&nbsp;description 'ISL LAG - ports 13-16'
##&nbsp;port-channel load-balance 7
##&nbsp;mtu 9216
##&nbsp;exit
##&nbsp;interface 3/64
##&nbsp;description 'NetApp CN1610 Cluster Switch RCF v1.0 (2012-01-27)'
##&nbsp;exit
##&nbsp;isdp timer 5
##&nbsp;exit
##&nbsp;terminal length 0
##&nbsp;--More-- or (q)uit
##&nbsp;
##&nbsp;
##&nbsp;Console Break-in Status........................ Enable
##&nbsp;
##&nbsp;--More-- or (q)uit
##&nbsp
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;          {
##&nbsp;            'show_running_config_info' => [
##&nbsp;                                            {
##&nbsp;                                              'ip_ssh_server' => 'enable',
##&nbsp;                                              'spanning_tree_configuration' => '00-A0-98-4B-B7-F1',
##&nbsp;                                              'snmp_server_community' => 'abc1',
##&nbsp;                                              'snmp_server_sysname' => 'net-1960-1',
##&nbsp;                                              'lines' => [
##&nbsp;                                                           {
##&nbsp;                                                             'serial_timeout' => '60',
##&nbsp;                                                             'line' => 'console'
##&nbsp;                                                           },
##&nbsp;                                                           {
##&nbsp;                                                             'line' => 'telnet'
##&nbsp;                                                           },
##&nbsp;                                                           {
##&nbsp;                                                             'line' => 'ssh'
##&nbsp;                                                           }
##&nbsp;                                                         ],
##&nbsp;                                              'current_sntp_synchronized_time' => 'Apr 25 09:21:29 201 UTC',
##&nbsp;                                              'password' => '54efa5bc56b147ac920e6a5cbba5dfe76bd81744dd12ffe9141907f7a6f87a9db9fa55b14573ebe19c8654d0b98a1c393c5635b5ced4d40d1e16f5cdd6e3a991',
##&nbsp;                                              'interfaces' => [
##&nbsp;                                                                {
##&nbsp;                                                                  'interface' => '0/13',
##&nbsp;                                                                  'running_config' => [
##&nbsp;                                                                                        'no auto-negotiate',
##&nbsp;                                                                                        'addport 3/1',
##&nbsp;                                                                                        'exit'
##&nbsp;                                                                                      ]
##&nbsp;                                                                },
##&nbsp;                                                                {
##&nbsp;                                                                  'interface' => '0/14',
##&nbsp;                                                                  'running_config' => [
##&nbsp;                                                                                        'no auto-negotiate',
##&nbsp;                                                                                        'addport 3/1',
##&nbsp;                                                                                        'exit'
##&nbsp;                                                                                      ]
##&nbsp;                                                                },
##&nbsp;                                                                {
##&nbsp;                                                                  'interface' => '0/15',
##&nbsp;                                                                  'running_config' => [
##&nbsp;                                                                                        'no auto-negotiate',
##&nbsp;                                                                                        'addport 3/1',
##&nbsp;                                                                                        'exit'
##&nbsp;                                                                                      ]
##&nbsp;                                                                },
##&nbsp;                                                                {
##&nbsp;                                                                  'interface' => '0/16',
##&nbsp;                                                                  'running_config' => [
##&nbsp;                                                                                        'no auto-negotiate',
##&nbsp;                                                                                        'addport 3/1',
##&nbsp;                                                                                        'exit'
##&nbsp;                                                                                      ]
##&nbsp;                                                                },
##&nbsp;                                                                {
##&nbsp;                                                                  'interface' => '0/1',
##&nbsp;                                                                  'running_config' => [
##&nbsp;                                                                                        'no auto-negotiate',
##&nbsp;                                                                                        'storm-control broadcast level 3',
##&nbsp;                                                                                        'storm-control multicast level 3',
##&nbsp;                                                                                        'storm-control flowcontrol',
##&nbsp;                                                                                        'storm-control unicast level 3',
##&nbsp;                                                                                        'description \'Cluster Node 1\'',
##&nbsp;                                                                                        'spanning-tree edgeport',
##&nbsp;                                                                                        'mtu 9216',
##&nbsp;                                                                                        'exit'
##&nbsp;                                                                                      ]
##&nbsp;                                                                },
##&nbsp;                                                                {
##&nbsp;                                                                  'interface' => '0/2',
##&nbsp;                                                                  'running_config' => [
##&nbsp;                                                                                        'no auto-negotiate',
##&nbsp;                                                                                        'storm-control flowcontrol',
##&nbsp;                                                                                        'description \'Cluster Node 2\'',
##&nbsp;                                                                                        'spanning-tree edgeport',
##&nbsp;                                                                                        'mtu 9216',
##&nbsp;                                                                                        'exit'
##&nbsp;                                                                                      ]
##&nbsp;                                                                },
##&nbsp;                                                                {
##&nbsp;                                                                  'interface' => '0/3',
##&nbsp;                                                                  'running_config' => [
##&nbsp;                                                                                        'no auto-negotiate',
##&nbsp;                                                                                        'description \'Cluster Node 3\'',
##&nbsp;                                                                                        'spanning-tree edgeport',
##&nbsp;                                                                                        'mtu 9216',
##&nbsp;                                                                                        'exit'
##&nbsp;                                                                                      ]
##&nbsp;                                                                },
##&nbsp;                                                                {
##&nbsp;                                                                  'interface' => '0/4',
##&nbsp;                                                                  'running_config' => [
##&nbsp;                                                                                        'no auto-negotiate',
##&nbsp;                                                                                        'description \'Cluster Node 4\'',
##&nbsp;                                                                                        'spanning-tree edgeport',
##&nbsp;                                                                                        'mtu 9216',
##&nbsp;                                                                                        'exit'
##&nbsp;                                                                                      ]
##&nbsp;                                                                },
##&nbsp;                                                                {
##&nbsp;                                                                  'interface' => '0/5',
##&nbsp;                                                                  'running_config' => [
##&nbsp;                                                                                        'no auto-negotiate',
##&nbsp;                                                                                        'description \'Cluster Node 5\'',
##&nbsp;                                                                                        'spanning-tree edgeport',
##&nbsp;                                                                                        'mtu 9216',
##&nbsp;                                                                                        'exit'
##&nbsp;                                                                                      ]
##&nbsp;                                                                },
##&nbsp;                                                                {
##&nbsp;                                                                  'interface' => '0/6',
##&nbsp;                                                                  'running_config' => [
##&nbsp;                                                                                        'no auto-negotiate',
##&nbsp;                                                                                        'description \'Cluster Node 6\'',
##&nbsp;                                                                                        'spanning-tree edgeport',
##&nbsp;                                                                                        'mtu 9216',
##&nbsp;                                                                                        'exit'
##&nbsp;                                                                                      ]
##&nbsp;                                                                },
##&nbsp;                                                                {
##&nbsp;                                                                  'interface' => '0/7',
##&nbsp;                                                                  'running_config' => [
##&nbsp;                                                                                        'no auto-negotiate',
##&nbsp;                                                                                        'description \'Cluster Node 7\'',
##&nbsp;                                                                                        'spanning-tree edgeport',
##&nbsp;                                                                                        'mtu 9216',
##&nbsp;                                                                                        'exit'
##&nbsp;                                                                                      ]
##&nbsp;                                                                },
##&nbsp;                                                                {
##&nbsp;                                                                  'interface' => '0/8',
##&nbsp;                                                                  'running_config' => [
##&nbsp;                                                                                        'no auto-negotiate',
##&nbsp;                                                                                        'description \'Cluster Node 8\'',
##&nbsp;                                                                                        'spanning-tree edgeport',
##&nbsp;                                                                                        'mtu 9216',
##&nbsp;                                                                                        'exit'
##&nbsp;                                                                                      ]
##&nbsp;                                                                },
##&nbsp;                                                                {
##&nbsp;                                                                  'interface' => '0/9',
##&nbsp;                                                                  'running_config' => [
##&nbsp;                                                                                        'no auto-negotiate',
##&nbsp;                                                                                        'description \'Cluster Node 9\'',
##&nbsp;                                                                                        'spanning-tree edgeport',
##&nbsp;                                                                                        'mtu 9216',
##&nbsp;                                                                                        'exit'
##&nbsp;                                                                                      ]
##&nbsp;                                                                },
##&nbsp;                                                                {
##&nbsp;                                                                  'interface' => '0/10',
##&nbsp;                                                                  'running_config' => [
##&nbsp;                                                                                        'no auto-negotiate',
##&nbsp;                                                                                        'description \'Cluster Node 10\'',
##&nbsp;                                                                                        'spanning-tree edgeport',
##&nbsp;                                                                                        'mtu 9216',
##&nbsp;                                                                                        'exit'
##&nbsp;                                                                                      ]
##&nbsp;                                                                },
##&nbsp;                                                                {
##&nbsp;                                                                  'interface' => '0/11',
##&nbsp;                                                                  'running_config' => [
##&nbsp;                                                                                        'no auto-negotiate',
##&nbsp;                                                                                        'description \'Cluster Node 11\'',
##&nbsp;                                                                                        'spanning-tree edgeport',
##&nbsp;                                                                                        'mtu 9216',
##&nbsp;                                                                                        'exit'
##&nbsp;                                                                                      ]
##&nbsp;                                                                },
##&nbsp;                                                                {
##&nbsp;                                                                  'interface' => '0/12',
##&nbsp;                                                                  'running_config' => [
##&nbsp;                                                                                        'no auto-negotiate',
##&nbsp;                                                                                        'description \'Cluster Node 12\'',
##&nbsp;                                                                                        'spanning-tree edgeport',
##&nbsp;                                                                                        'mtu 9216',
##&nbsp;                                                                                        'exit'
##&nbsp;                                                                                      ]
##&nbsp;                                                                },
##&nbsp;                                                                {
##&nbsp;                                                                  'interface' => '0/13',
##&nbsp;                                                                  'running_config' => [
##&nbsp;                                                                                        'description \'ISL member port 13\'',
##&nbsp;                                                                                        'mtu 9216',
##&nbsp;                                                                                        'exit'
##&nbsp;                                                                                      ]
##&nbsp;                                                                },
##&nbsp;                                                                {
##&nbsp;                                                                  'interface' => '0/14',
##&nbsp;                                                                  'running_config' => [
##&nbsp;                                                                                        'description \'ISL member port 14\'',
##&nbsp;                                                                                        'mtu 9216',
##&nbsp;                                                                                        'exit'
##&nbsp;                                                                                      ]
##&nbsp;                                                                },
##&nbsp;                                                                {
##&nbsp;                                                                  'interface' => '0/15',
##&nbsp;                                                                  'running_config' => [
##&nbsp;                                                                                        'description \'ISL member port 15\'',
##&nbsp;                                                                                        'mtu 9216',
##&nbsp;                                                                                        'exit'
##&nbsp;                                                                                      ]
##&nbsp;                                                                },
##&nbsp;                                                                {
##&nbsp;                                                                  'interface' => '0/16',
##&nbsp;                                                                  'running_config' => [
##&nbsp;                                                                                        'description \'ISL member port 16\'',
##&nbsp;                                                                                        'mtu 9216',
##&nbsp;                                                                                        'exit'
##&nbsp;                                                                                      ]
##&nbsp;                                                                },
##&nbsp;                                                                {
##&nbsp;                                                                  'interface' => '3/1',
##&nbsp;                                                                  'running_config' => [
##&nbsp;                                                                                        'description \'ISL LAG - ports 13-16\'',
##&nbsp;                                                                                        'port-channel load-balance 7',
##&nbsp;                                                                                        'mtu 9216',
##&nbsp;                                                                                        'exit'
##&nbsp;                                                                                      ]
##&nbsp;                                                                },
##&nbsp;                                                                {
##&nbsp;                                                                  'interface' => '3/64',
##&nbsp;                                                                  'running_config' => [
##&nbsp;                                                                                        'description \'NetApp CN1610 Cluster Switch RCF v1.0 (2012-01-27)\'',
##&nbsp;                                                                                        'exit'
##&nbsp;                                                                                      ]
##&nbsp;                                                                }
##&nbsp;                                                              ],
##&nbsp;                                              'rcf_version' => 'v1.0',
##&nbsp;                                              'sntp_server' => '172.19.2.16',
##&nbsp;                                              'additional_packages' => [
##&nbsp;                                                                         'FASTPATH',
##&nbsp;                                                                         'QOS'
##&nbsp;                                                                       ],
##&nbsp;                                              'sshcon_timeout' => '60',
##&nbsp;                                              'port_channel' => '3/1 ISL-LAG',
##&nbsp;                                              'system_up_time' => '62 days 9 hrs 8 mins 51 secs',
##&nbsp;                                              'serviceport_protocol' => 'none',
##&nbsp;                                              'set_prompt' => 'net-1960-1',
##&nbsp;                                              'serviceport_ip' => [
##&nbsp;                                                                    '172.22.200.93',
##&nbsp;                                                                    '255.255.254.0',
##&nbsp;                                                                    '172.22.200.1'
##&nbsp;                                                                  ],
##&nbsp;                                              'username' => 'admin',
##&nbsp;                                              'system_description' => 'NetApp CN1610, 1.1.0.1, Linux 2.6.21.7',
##&nbsp;                                              'system_software_version' => '1.1.0.1',
##&nbsp;                                              'sntp_client_mode' => 'unicast',
##&nbsp;                                              'level' => '15',
##&nbsp;                                              'ip_name_server' => '172.22.201.46',
##&nbsp;                                              'sntp_server_status' => 'Success',
##&nbsp;                                              'ip_domain_name' => 'lab.netapp.com',
##&nbsp;                                              'isdp_timer' => '5',
##&nbsp;                                              'telnetcon_timeout' => '60',
##&nbsp;                                              'spanning_tree' => 'bpduguard'
##&nbsp;                                            }
##&nbsp;                                          ],
##&nbsp;            'show_version_info' => [
##&nbsp;                                     {
##&nbsp;                                       'serial_number' => '11111200069',
##&nbsp;                                       'maintenance_level' => 'A',
##&nbsp;                                       'operating_system' => 'Linux 2.6.21.7',
##&nbsp;                                       'network_processing_device' => 'BCM56820_B0',
##&nbsp;                                       'machine_type' => 'NetApp CN1610',
##&nbsp;                                       'system_description' => 'NetApp CN1610, 1.1.0.1, Linux 2.6.21.7',
##&nbsp;                                       'part_number' => 'BCM56820',
##&nbsp;                                       'software_version' => '1.1.0.1',
##&nbsp;                                       'burned_in_mac_address' => '00:A0:98:4B:B7:F1',
##&nbsp;                                       'manufacturer' => '0xbc00',
##&nbsp;                                       'additional_packages' => [
##&nbsp;                                                                  'FASTPATH QOS'
##&nbsp;                                                                ],
##&nbsp;                                       'fru_number' => '111-00893',
##&nbsp;                                       'machine_model' => 'CN1610'
##&nbsp;                                     }
##&nbsp;                                   ]
##&nbsp;          }
##&nbsp;        ];
##&nbsp;
sub show_tech_support {
    my %args = @_;
    my ( $error_message, @result_array, $result );
    my $resp    = $args{response_object};
    my $output  = $resp->get_processed_output();
    my $command = $resp->get_command();
    my $line;
    my %hash;
    my $error = _check_error(
        raw_output    => $output,
        error_message => \$error_message
    );

    if ($error) {
        $error = 'NACL::APISet::Exceptions::' . $error . 'Exception';
        $error->throw( text => $error_message, response_object => $resp );
    }
    my @lines = get_lines( raw_output => $output, response_object => $resp, first_line => 'Switch:\s+\d+' );
    shift(@lines) if ( $lines[0] =~ /Switch:\s+\d+/ );
    while ( $line = shift(@lines) ) {
        next if ( $line =~ /--More--\s+or\s+\(q\)uit/ );
        last if ( $line =~ /Show SysInfo/ );
        if ( $line =~ /^Additional Packages/ ) {
            my @values;
            my ( $key, $value ) = ( $line =~ /^(.*\w)\.\.+\s*(.*)/ );
            push( @values, $value );
            if ( $lines[0] =~ /^(\w+.*\w)/ ) {
                $line = shift(@lines);
                push( @values, $line );
            }
            $hash{ to_key_name( item => $key ) } = \@values;
        } else {
            my ( $key, $value ) = ( $line =~ /^(.*\w)\.\.+\s*(.*)/ );
            if ( $lines[0] =~ /^(\d+\.\d+\.*\d+)/ ) {
                $line = shift(@lines);
                $hash{ to_key_name( item => $key ) } = join( ' ', $value, $line );
            } else {
                $hash{ to_key_name( item => $key ) } = $value;
            }
        }
    }
    $result = show_running_config(@_);
    push( @{ $result_array[0]->{'show_version_info'} }, \%hash );
    $result_array[0]->{'show_running_config_info'} = $result;
    return \@result_array;
}

## @name ::show_vlan
##
## @summary Parses the raw output from the "show vlan" command
##
## @description 
##    This method can be used for parsing the raw output obtained after calling 
##    the method from show_vlan with any of its options as described in the 
##    corresponding cdef file 'show.cdefs'
## @synopsis
##      show_vlan(%args);
## INPUTS TO METHOD CALL 'show_vlan' ON APISET OBJECT 
## <table border = "1" >
## <tr> <th> OPTION </th> <th> PROGRAMMATIC NAME </th> <th> BOOLEAN OPTION </th> </tr>
## </table>
## @arg %args Required:
## response_object => The response object that was constructed (containing the
##                    raw output, command line constructed and execution status)
## @return
## Returns an array reference which contains the parsed output.
## @examples
##&nbsp;Command = show vlan
##&nbsp;
##&nbsp;-------- Raw Output ---------
##&nbsp;
##&nbsp;
##&nbsp;VLAN ID VLAN Name                        VLAN Type
##&nbsp;------- -------------------------------- -------------------
##&nbsp;1       default                           Default  
##&nbsp;
##&nbsp;
##&nbsp;-------- Parsed Output ------
##&nbsp;$VAR1 = [
##&nbsp;          {
##&nbsp;            'vlan_type' => 'Default',
##&nbsp;            'vlan_name' => 'default',
##&nbsp;            'vlan_id' => '1'
##&nbsp;          }
##&nbsp;        ];
sub show_vlan {
    my %args = @_;
    my ( $error_message, @result_array, $result );
    my $resp    = $args{response_object};
    my $output  = $resp->get_processed_output();
    my $command = $resp->get_command();
    my $line;
    my %hash;
    my $error = _check_error(
        raw_output    => $output,
        error_message => \$error_message
    );

    if ($error) {
        $error = 'NACL::APISet::Exceptions::' . $error . 'Exception';
        $error->throw( text => $error_message, response_object => $resp );
    }
    my @lines = get_lines( raw_output => $output, response_object => $resp );
    my (@result_array, @joinedHeadings, @headings, @headingsValues, $count, $line, $iteration);
    while (@lines) {
        $line = shift(@lines);
        if ( $line =~ /^VLAN/ ) {
            @headings = split( /\s+/, $line );
            for ( $count = 0; $count <= $#headings; $count = $count + 2 ) {
                $joinedHeadings[$iteration] = $headings[$count] . '_' . $headings[ $count + 1 ];
                $iteration++;
            }
        } else {
            $line = shift(@lines);
            while ($line) {
                @headingsValues = split( /\s+/, $line );
                for ( $count = 0; $count <= $#joinedHeadings; $count = $count + 1 ) {
                    $result_array[0]->{ to_key_name( item => $joinedHeadings[$count] ) } = $headingsValues[$count];
                }
                $line = shift(@lines);
            }
        }
    }
    return \@result_array;
}

sub show_snmpcommunity {
    my %args = @_;
    my ($error_message);
    my $resp    = $args{response_object};
    my $output  = $resp->get_processed_output();
    my $command = $resp->get_command();
    my $error   = _check_error(
        raw_output    => $output,
        error_message => \$error_message
    );
    if ($error) {
        $error = 'NACL::APISet::Exceptions::' . $error . 'Exception';
        $error->throw( text => $error_message, response_object => $resp );
    }
    my  @headings = ("SNMP Community Name","Client IP Address","Client IP Mask","Access Mode","Status");
    my (@result_array,$newkey,$key);
    my @lines = get_lines( raw_output => $output );
    my $j = 0;
    for ( my $k = 2; $k <= $#lines; $k++ ) {
        my @values = split( /\s\s+/, $lines[$k] );
        for ( my $i = 0; $i < scalar(@headings); $i++ ) {
            if ( $values[$i] ) {
                $result_array[$j]->{ to_key_name(item => $headings[$i]) } = $values[$i];
            }
        }
        $j++;
    }
    return \@result_array;
}

1;

