#
# Copyright (c) 2001-2010 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#
## @summary Top level base class for Versioning subsystem. APISet will make
## call to the constructor of this class which will in turn create a version
## object of appropriate type.
## @author rajivg@netapp.com, Isshwarya.M@netapp.com, dl-nacl-dev@netapp.com
## @status Shared
## @pod here

package NACL::APISet::Version;

#
# Compiler Directives
#
use strict;
use warnings;

#
# Module imports
#
use NACL::APISet;

#
# NATE imports
#
use Data::Dumper;
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Params::Validate qw(:all);
use Storable qw(freeze thaw);

#
# Exceptions imported
#
use NATE::BaseException qw(:try);
use NACL::APISet::Exceptions::AttributeDiscoveryFailed qw(:try);

use Class::MethodMaker [
    hash   => 'features',
    scalar => [ { -type => 'Connectrec' }, 'cli_connection' ],
    scalar => [ { -type => 'Connectrec' }, 'console_connection' ],
    scalar => [ { -type => 'Hostrec' }, 'hostobj' ],
    hash => 'session_object',
    hash => 'conn_overrides',
];

###############################################################################
# Method:    new
# Objective: Initialized the version object.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub new {
    $Log->enter() if $may_enter;

    my ( $proto, %args ) = @_;
    my $class = ref($proto) || $proto;

    # Validate parameters and populate member variables
    my $self = NACL::APISet::validate_params(
        $class,
        params => \%args,
        spec   => {
            hostobj            => { isa  => "Hostrec",    optional => 1 },
            category           => { type => SCALAR,       optional => 0 },
            shell_type         => { type => SCALAR,       optional => 0 },
            conn_overrides     => { type => HASHREF,      optional => 1 },
            console_connection => { isa  => "Connectrec", optional => 1 },
            cli_connection     => { isa  => "Connectrec", optional => 1 },
            session_object     => { type  => HASHREF, optional => 1},
        },
        allow_extra => 0
    );
    bless $self, $class;

    if (   !$self->hostobj()
        && !$self->cli_connection()
        && !$self->console_connection() )
    {
        $Log->exit() if $may_exit;
        NATE::Exceptions::Argument->throw(
                  "Method 'NACL::APISet::Version->new' "
                . "requires parameter 'hostobj' or 'cli_connection' or 'console_connection',"
                . " but none were specified" );
    }

    if ( !$self->hostobj() && $self->cli_connection() ) {
        $self->hostobj( $self->cli_connection()->hostp() );
    } elsif ( !$self->hostobj() && $self->console_connection() ) {
        $self->hostobj( $self->console_connection()->hostp() );
    }

    my $version_set;

    # NOTE:
    # Mapping for different shell_types Vs version subclass to be created
    # for that. An entry should be added in this mapping table whenever
    # the shell to which the user wants the higher layers to connect to is
    # different from the service type. This happens when adding support for
    # various contexts in a service type.
    # Otherwise it will default to shell_type itself as the mapped value.
    my %mapping = (
        "7Mode"  => "7Mode",
        "Vfiler" => "7Mode"
    );

    # Determine the type of version object to be created
    $version_set
        = ( defined $mapping{ $self->{shell_type} } )
        ? $mapping{ $self->{shell_type} }
        : $self->{shell_type};

    my $version_pkg
        = "NACL::APISet::Version::" . $self->{category} . "::" . $version_set;
    eval "use $version_pkg";
    $Log->debug("Creating version object of type: $version_pkg");
    if ($@) {
        NATE::BaseException->throw( "Error in loading $version_pkg:\n$@\n"
                . "It could mostly be, either the package $version_pkg doesn't"
                . " exist or you might have forgot to load it" );
    }

    # create the versioning object of appropriate type.
    my $version_manager = $version_pkg->new(%args);

    $Log->exit() if $may_exit;
    return $version_manager;

}

##############################################################################
# Method:    get_version_attribute
# Objective: Return the value of the specified version attribute
# Details:   See POD documentation at the end of this file.
##############################################################################

sub get_version_attribute {
    $Log->enter() if $may_enter;

    my $self = shift();
    my %args = NACL::APISet::validate_params(
        $self,
        params => \@_,
        spec   => { "attribute" => { type => SCALAR } }
    );
    my $attribute = $args{attribute};

   # If the information about the attribute is already present in version
   # object, then simply return it. Need not do discovery.
   # Purposely doing this before verifying whether the attribute is a
   # valid attribute or not  This gives the flexibility of setting any version
   # related attributes and retrieving them by the users though versioning
   # doesn't know how to do autodiscovery for that attribute.
    if ( defined $self->{$attribute} ) {
        $Log->debug( "$attribute is already cached to be "
                . Dumper( $self->{$attribute} ) );
        $Log->exit() if $may_exit;
        return $self->{$attribute};
    }

    # Call the corresponding find method

    # NOTE: The attribute method added for its discovery should follow the
    # naming syntax of _find_<attribute>
    my $method_name = "_find_" . $attribute;
    my $is_valid    = 0;
    $is_valid = 1 if ( $self->can($method_name) );

   # For 7G systems, valid_attributes will be explictly defined in $self.
   # Because not all the attributes which can be discovered for a 7Mode system
   # can be discovered in 7G system. This check is required as we are using
   # the same class for both 7Mode and 7G
    if ( ( $self->{is_classic} ) && ( defined $self->{valid_attributes} ) ) {
        $is_valid = 0;
        map { $is_valid = 1 if ( $attribute eq $_ ) }
            @{ $self->{valid_attributes} };
    }
    unless ($is_valid) {
        $Log->exit() if $may_exit;
        NACL::APISet::Exceptions::AttributeDiscoveryFailed->throw(
            "Version subsystem has no support for finding " . $attribute );
    }
    $Log->debug( "The version information for "
            . $attribute
            . " is not found. Hence versioning will perform a discovery" );
    no strict 'refs';

    # We are not catching exceptions thrown by the find_* method and we are
    # leaving it to callers of versioning. The expected exceptions are
    # ConnectionFailedException, AttributeDiscoveryFailed.
    $self->$method_name();

    # By this time $self should have been populated with the information.
    # If its still undef, then we are not able to obtain the information
    # and throw an exception.
    unless ( defined $self->{$attribute} ) {
        $Log->exit() if $may_exit;
        NACL::APISet::Exceptions::AttributeDiscoveryFailed->throw(
            "Unable to obtain " . $attribute . " information" );
    }
    $Log->debug(
        "$attribute is discovered as " . Dumper( $self->{$attribute} ) );
    $Log->exit() if $may_exit;
    return $self->{$attribute};
}
##############################################################################
# Method:    execute_CLI_command
# Objective: Execute the sent command and gathers raw output
# Details:   Refer to POD documentation at the end of the file.
##############################################################################

sub execute_CLI_command {
    my ($self, @args) = @_;
    $Log->enter() if $may_enter;
    my %args = NACL::APISet::validate_params(
        $self,
        params => \@args,
        spec   => {
            command    => { type => SCALAR,       optional => 0 },
            connection => { isa  => "Connectrec", optional => 0 }
        }
    );

    my %opts = %args;
    foreach ( keys %opts ) {
        if ( $_ =~ /^connectrec\-(.*)/ ) {
            $args{$1} = delete $args{$_};
        }
    }

    my $connection = delete $args{connection};

    # Any wrapper to write around connectrec::execute can be done here
    # like if we want to catch timeout exceptions and proceed with next
    # possible CLI command to obtain version related information.
    my $raw_output = "";
    try {
        $raw_output = $connection->execute(
            %args,
            on_timeout =>
                sub { $connection->send_intr(); $connection->resync(); },
            on_idle =>
                sub { $connection->send_intr(); $connection->resync(); },
        );
    }
    otherwise {
        my $ex = shift;
        $Log->comment( "Error in command execution by versioning is: "
                . $ex->text() );
    };
    
    $Log->exit() if $may_exit;
    return $raw_output;
}

##############################################################################
# Method:    execute_ZAPI
# Objective: Execute the input Zapi XML
# Details:   Refer to POD documentation at the end of the file.
##############################################################################

sub execute_ZAPI {
    my $self = shift @_;
    $Log->enter() if $may_enter;
    my %opts = NACL::APISet::validate_params(
        $self,
        params => \@_,
        spec   => {
            command => { type => SCALAR, optional => 0 },
            session => { isa => "NACL::APISet::Session::ZAPI", optional => 0 }
        }
    );
    my $response;
    try {
        $response = $opts{session}->execute_raw_ZAPI(
            command                  => $opts{command},
            'return-response-object' => 1
        );
    }
    otherwise {
        my $ex = shift;
        $Log->debug(
            "Error in ZAPI Execution by versioning is:\n" . $ex->text() );
    };

    $Log->exit() if $may_exit;
    return $response;
}

##############################################################################
# Method:    map_release_to_codename
# Objective: Map release strings to internal codenames
# Details:   See POD documentation at the end of this file.
##############################################################################

sub map_release_to_codename {
    $Log->enter() if $may_enter;
    my ( $self, %opts ) = @_;
    my $release_table = $opts{release_table};
    my %release_table = %$release_table;
    my $release       = $opts{release};

    my $codename;

    my @mapped = grep { $_ if ( $release eq $_ ) } keys(%release_table);
    if ( scalar(@mapped) == 0 ) {
        $Log->debug("Doing regex match with release table hash keys");
        @mapped = grep { $_ if ( $release =~ quotemeta($_) ) }
            keys(%release_table);
        my %hash;
        map { $hash{ $release_table{$_} } = 1 } @mapped;
        @mapped = keys %hash;

        # If @mapped is 0(no match) or >1 (more than one matched, this
        # shouldn't be but), then throw an exception as we failed in
        # mapping the release to its codename.
        if ( scalar @mapped == 1 ) {

            # Regexp match
            $codename = $mapped[0];
        } elsif ( scalar @mapped == 0 ) {
            # Lets try to guess the codename for standard release names
            if ($release =~ /(\w+?)(__\d\.\d\.(\d)|\.(\d))$/) {
                $codename = (defined $3)? $1."_".$3 : $1."_".$4;
                $Log->debug( "No matching entry found for $release. Codename has been intelligently formed as $codename");
            } else {
                $Log->exit() if $may_exit;
                NATE::BaseException->throw( 'No matching entry found to map '
                    . "release '$release' to its codename" );
            }
        } else {
            $Log->exit() if $may_exit;

            # More than one match!!!
            NATE::BaseException->throw(
                      'Multiple matching entries found while'
                    . " trying to map release '$release' to its codename. "
                    . "They are:\n"
                    . Dumper( \@mapped ) );

        }
    } else {

        # There can be only one exact match because hash structure won't
        # allow duplicate keys.
        $codename = $release_table{ $mapped[0] };
    }
    $codename =~ s/\./_/g;
    $codename = lc($codename);

    $Log->debug("codename is found as $codename");
    $Log->exit() if $may_exit;
    return $codename;
}

##############################################################################
# Method:    set_version_attribute
# Objective: Set the value of the specified version attribute
# Details:   See POD documentation at the end of this file.
##############################################################################

sub set_version_attribute {
    $Log->enter() if $may_enter;

    my $self = shift();
    my %args = NACL::APISet::validate_params(
        $self,
        params => \@_,
        spec   => {
            "attribute" => { type => SCALAR },
            "value"     => { type => SCALAR | ARRAYREF }
        }
    );

    # User can use this method to set any version attribute in version object
    # Check for valid attributes is not done so as to give the flexibility of
    # setting any version related attribute which versioning doesn't know how
    # to autodiscover.
    $self->{ $args{"attribute"} } = $args{"value"};
    $Log->exit() if $may_exit;
}

##############################################################################
# Method:    clear_version_attribute
# Objective: Clear the value of the specified version attribute
# Details:   See POD documentation at the end of this file.
##############################################################################

sub clear_version_attribute {
    $Log->enter() if $may_enter;

    my $self = shift();
    my %args = NACL::APISet::validate_params(
        $self,
        params => \@_,
        spec   => { "attribute" => { type => SCALAR } }
    );

   # User can use this method to clear any version attribute in version object
   # so that the cached result can be cleared and the next time, if the
   # same version attribute is queried, then versioning will perform a new
   # discovery.

    $self->{ $args{"attribute"} } = undef;
    $Log->exit() if $may_exit;
}

##############################################################################
# Method:    set_conn_overrides
# Objective: Set overrides required to establish a connection to the resource
# Details:   This method is deprecated
##############################################################################

sub set_conn_overrides {
    my $self = shift();
    $Log->enter() if $may_enter;
    my %args = NACL::APISet::validate_params(
        $self,
        params => \@_,
        spec   => {
            username  => { type => SCALAR, optional => 1 },
            password  => { type => SCALAR, optional => 1 },
            ipaddress => { type => SCALAR, optional => 1 },
            conntype  => { type => SCALAR, optional => 1 },

            # Defaults for connid will be given in the respective session
            # classes
            connid          => { type => SCALAR,         optional => 1 },
            prompt_username => { type => SCALAR | UNDEF, optional => 1 },
            prompt_password => { type => SCALAR | UNDEF, optional => 1 },
        },
        allow_extra => 0
    );

    my ( $key, $value );
    while ( ( $key, $value ) = each(%args) ) {
        $self->{'conn_overrides'}{$key} = $value;
    }
    $Log->exit() if $may_exit;
}

###############################################################################
# Method:    get_session
# Objective: Get session object based on interface and subtype.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_session {
    $Log->enter() if $may_enter;

    my $self = shift();
    my %args = NACL::APISet::validate_params(
        $self,
        params => \@_,
        spec   => {
            "interface" => { type => SCALAR,         optional => 0 },
            "subtype"   => { type => SCALAR,         optional => 0 },
            "connid"    => { type => SCALAR,         optional => 1 },
            "conntype"  => { type => SCALAR | UNDEF, optional => 1 }
        }
    );
    my $session;
    my @interface_subtype = keys $self->session_object();
    my $key = lc $args{interface} . "_" . lc $args{subtype};
    $key =~ s/::/_/g;
    my $has_conn_overrides = defined $args{connid} || defined $args{conntype};
    if ($self->session_object_exists($key) && !$has_conn_overrides) {
        my $existing_session = $self->session_object_index($key);
        if ($existing_session->{connection}->is_persist() && $existing_session->{connection}->is_disconnected()) {
             $existing_session->_reconnect();
             $existing_session->prep_service();
        }
        $self->session_object_set($key => $existing_session);
        $session = $self->session_object_index($key);
    } else {

    # These interface and set values depends on what information
    # the version class wants to obtain and they need not be the same as the
    # user defined interface and set values (what commands they want to run)
    my $interface = $args{interface};
    my $api_set   = $args{subtype};

    # Establish the session if not provided
    my $session_pkg
        = "NACL::APISet::Session::" . $interface . "::" . $api_set;
    eval "use $session_pkg";
    if ($@) {
        $Log->exit() if $may_exit;
        NATE::BaseException->throw(
            "Error in loading '$session_pkg': " . $@ );
    }

  # By default, consider the overrides what user had provided. This is
  # important because consider they had given rsh-persay as conntype but if we
  # try running the command with ssh and if ssh is not enabled in the service,
  # then the command will fail. So try using the same overrides provided by
  # the user wherever possible.
    my %session_overrides;
    unless ( $self->{shell_type} eq 'Vfiler' ) {

        # If the hostobj represents 'vfiler', then we internally
        # connect to the PFILER(parent filer), so those overrides
        # provided for vfiler, shouldn't be considered for pfiler.
        %session_overrides = %{ $self->{'conn_overrides'} }
            if ( defined $self->{'conn_overrides'} );
    }

    # If versioning has passed any overrides for connid and conntype,
    # then it will override the user specified connection overrides.
    if ( defined $args{connid} ) {
        $session_overrides{connid}   = $args{connid};
        $session_overrides{conntype} = $args{conntype};
    }
    delete $session_overrides{conntype}
        unless ( defined $session_overrides{conntype} );

    # set the prompt_username as "diag" for CMode and 7Mode systemshell.
    if (   $args{subtype} eq "CMode::Shell"
        or $args{subtype} eq "7Mode::Shell" )
    {
        $session_overrides{"prompt_username"} = "diag";
    }

    # Only if the requested session by individual version class is
    # 7M/CM CLI, then consider using console connection object
    my %session_param = ();
    if ( $self->cli_connection() && ( $api_set !~ /Shell$/ ) ) {
        $session_param{"connection"} = $self->cli_connection();
    } elsif ( ( $self->console_connection() )
        && ( $interface eq 'CLI' )
        && ( $api_set =~ /^(CMode|7Mode)$/ ) )
    {
        $session_param{"connection"} = $self->console_connection();
    } else {
        $session_param{"hostobj"} = $self->hostobj();
    }

    if ( $args{interface} eq 'ZAPI' ) {
        $session_param{"set"} = $args{subtype};
    }

    # We create the session class just to obtain the connection object and
    # nothing else we will use from session class. Versioning will have its
    # own implementation for sending command to the service and parsing the
    # output.
    try {
        $session = $session_pkg->new( %session_param, %session_overrides );
        $self->session_object_set($key => $session) if (!$has_conn_overrides);
    }
    catch NACL::APISet::Exceptions::ConnectionFailedException with {
        my $ex = shift;

       # Print the reason for exception and then rethrow. The reason for error
       # is printed here as session class prints the reason in 'debug' level.
        my $text = $ex->text();
        $text = $1 if ( $text =~ /(Could.*\r*\n*.*)/ );

        $Log->comment( "Versioning " . $text );
        $Log->debug( "stacktrace: " . $ex->stacktrace() );

       # Rethrow it and leave it to the calling subclass function whether or
       # not to catch this exception and proceed with the rest. If the calling
       # function is not catching this exception, then this exception will be
       # caught in the APISet class and the corresponding version attribute
       # will be undef then as discovery is not successful.
        $ex->throw();
    };
    }
    $Log->exit() if $may_exit;
    return $session;
}

###############################################################################
# Method:    destroy_session
# Objective: Destroy the created session.
# Details:   See POD documentation at the end of this file.
###############################################################################

sub destroy_session {
    $Log->enter() if $may_enter;
    # Since we are trying to reuse connections, it makes sense not to 
    # destroy the session.
    # Making this a no-op 
    $Log->exit() if $may_exit;
}

##############################################################################
# Method:    _find_codename
# Objective: Get the internal codename of the different releases
# Details:   This method executes commands so as to collect information about
#            internal codename and set the value in version object itself
##############################################################################

sub _find_codename {
    my ( $self, %args ) = @_;
    $Log->enter() if $may_enter;

    my $release = $self->get_version_attribute( attribute => 'release' );
    my $codename = $self->map_release_to_codename( release => $release );

    unless ( defined $codename ) {
        $Log->exit() if $may_exit;
        NACL::APISet::Exceptions::AttributeDiscoveryFailed->throw(
            "Error in getting the codename")

    }

    $Log->exit() if $may_exit;
    $self->{codename} = $codename;
    return;
}

sub STORABLE_freeze {
    my ( $self, $cloning ) = @_;
    return if $cloning;    # Regular default serialization
    my $data = {%$self};
    delete $data->{cli_connection};
    delete $data->{console_connection};
    delete $data->{session_object};
    my $dummy = "";
    return ( $dummy, $data );
} ## end sub STORABLE_freeze

sub STORABLE_thaw {
    my ( $self, $cloning, $dummy, $data ) = @_;
    %$self = %$data;
}

1;

=head1 NAME

NACL::APISet::Version - Base Class for APISet Version

=head1 SYNOPSIS

use NACL::APISet::Version;

=head1 DESCRIPTION

A C<NACL::APISet::Version> Top level base class for all Version related
implementation. The APISet base class will always make a call to the constructor
of this class which will in turn create an appropriate version object based
on category and shell_type

=cut

=head1 EXCEPTIONS

=over

=item NATE::BaseException

=item NATE::Exceptions::Argument

=item NACL::APISet::Exceptions::AttributeDiscoveryFailed

=back

=head1 ATTRIBUTES

=head2 features

All the verified features are stored in this Hash where the feature name
is the key and boolean value is the hash key value. Currently supported
features are L<Supported Features|lib-NACL-C-_Version-pm/FEATURES>.

=head2 hostobj

The hostrec object used by the Version Manager

=head2 cli_connection

The cli connection used by the Version manager is stored in this attribute.
The different operations which can be done on this attribute are:
Set the connection: $version_manager->cli_connection($connection);
Get the connection: $connection = $version_manager->cli_connection();
Clear the connection: $version_manager->cli_connection_reset();

=head2 console_connection

The console connection used by the Version manager is stored in this attribute.
The different operations which can be done on this attribute are:
Set the connection: $version_manager->console_connection($connection);
Get the connection: $connection = $version_manager->console_connection();
Clear the connection: $version_manager->console_connection_reset();

=head1 CONSTRUCTORS

=head2 new

    Create the appropriate version object based on input parameters supplied.

=over

=item Synopsis

# Creating a version manager by specifying basic inputs
my $version_mgr = NACL::APISet::Version->new(hostobj =>$hostobj,
                                              category => "Node",
                                              shell_type => "CMode");


# To share the console connection with version manager
my $version_mgr = NACL::APISet::Version->new(hostobj =>$hostobj,
                                              category => "Node",
                                              shell_type => "CMode",
                                              console_connection => $console);
											  
# To share the cli connection with version manager											  
my $version_mgr = NACL::APISet::Version->new(hostobj =>$hostobj,
                                              category => "Node",
                                              shell_type => "CMode",
                                              cli_connection => $cli_connection);


=item Arguments

=over

=item category

(Mandatory)
The category to use.  Valid values include "Node", "Host" and "Switch".

=item shell_type

(Mandatory)
The refers to the type of the shell host sent through 'hostobj'.
Valid values include "CMode", "7Mode", "Vfiler", "Vserver" for 'Node',
"DFM" for 'Host', "CiscoNetwork" for 'Switch'.
If hostobj refers to direct connection to vfiler/vserver then shell_type
should be "Vfiler" or "Vserver".

=item hostobj

(Optional)
The hostrec object created for the resource.
Either one of hostobj, console_connection, cli_connection is required.

=item conn_overrides

(Optional)
Hash reference to the list of connection overrides to be considered when
establishing the connection to the service.
Valid connection overrides are username, password, connid, conntype, ipaddress,
prompt_username and prompt_password.

=item console_connection

(Optional)
This paramter is applicable only to 'Node' category, especially if the
shell_type is either 7Mode or CMode. If you already have a console
connection and if you want the version manger to reuse the same connection for
whatever attributes that could be discovered, then you
can specify the existing console connection object to this option.
Either one of hostobj, console_connection, cli_connection is required.

=item cli_connection

(Optional)
If you already have a cli connection for Node/Host, and if you want the version 
manger to reuse the same connection for whatever attributes that could be 
discovered, then you can specify the existing cli connection object to this option.
Either one of hostobj, console_connection, cli_connection is required.

=back

=item Returns

A reference to a new version object.

=back

=head1 Methods

=head2 get_session

(Only for developers)
Create an ad hoc session for each piece of Versioning information
needed. We cannot reuse the same APISet session since different pieces
of versioning information require different connections - and
these might differ from the connection established for the APISet
object. For example, node version for CMode requires a CLI
connection, while WAFL version requires a ZAPI connection. Hence,
connections established for Versioning are separate from those
used for the APISet object.
Only connection object from the returned session object will be used by
Version classes.

=over

=item Synopsis

$session = $version_object->get_session(interface => "CLI", "subtype" => "7Mode");

=item Arguments

=over

=item interface
(Mandatory)

The type of interface. Valid values are CLI and ZAPI

=item subtype
(Mandatory)

The type of the session to be created.
For Eg: "7Mode", "CMode::Shell".

=item conntype

(Optional)
Connection type with which to connect to the resource

=item connid

(Optional)
Connection ID with which to connect to the resource

=back

=item Returns

Returns the session object of appropriate interface and subtype

=back

=head2 get_version_attribute

This method returns the information about the version attribute asked for.
If the information is not there in version object, then it will try to discover
and if discovery fails, then it will throw an exception.
Refer to L<http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/VersioningUsersGuide/index.html#Obtain_version_attributes_from_the_version_manager> 
for all the different version attributes that can be obtained.

=over

=item Synopsis

$branch = $version_object->get_version_attribute(attribute => "branch");

=item Arguments

=over

=item attribute

attribute whose information is required

=back

=item Returns

version information corresponding to the attribute.

=back

=head2 set_version_attribute

This method sets the version attribute to the specified value.

=over

=item Synopsis

$version_object->set_version_attribute(attribute => "branch", value => "mainN");

=item Arguments

=over

=item attribute

attribute whose information to be set

=item value

value to be set in the version object corresponding to the attribute.

=back

=item Returns

Nothing.

=back

=head2 execute_CLI_command

(Only for developers)
This method will execute the command using the connection object sent and
gathers raw output.

=over

=item Synopsis

$output = $version_object->execute_CLI_command( command => "version",
    connection => $connection );

=item Arguments

=over

=item command

The input command line to be executed on the service

=item connection

The connection object to be used for executing the command.

=back

=item Returns

This method will returns the raw output returned by the service without any
processing.

=back

=head2 execute_ZAPI

(Only for developers)
This method calls the execute_raw_ZAPI present in the Session class

=over

=item Synopsis

$output =
$version_object->execute_ZAPI( command => "<system-get-version/>" );

=item Arguments

=over

=item command

The input raw xml to be sent

=item session

session object of type NACL::APISet::Session::ZAPI::7Mode|CMode|DFM which
contain information about ip address, username and password

=back

=item Returns

This method will return the response object

=back

=head2 map_release_to_codename

For a same release, we can have different release strings like internal and
external release strings and hence map all release strings referring a
particular release to its internal codename.

=over

=item Synopsis

$codename = $version_object->map_release_to_codename(release => $release);

=item Arguments

=over

=item release

The release string.

=back

=item Returns

This method will return the codename corresponding to the specified release.
It throws NATE::BaseException if it couldn't obtain the codename.

=back

=head2 clear_version_attribute

This method clears the specified version attribute so that the cached result is
cleared and versioning will perform a fresh discovery for the next time rather
than returning the cached result.

=over

=item Synopsis

$version_object->clear_version_attribute(attribute => "branch");

=item Arguments

=over

=item attribute

attribute to be cleared

=back

=item Returns

Nothing.

=back

=head2 destroy_session

(Only for developers)
This method destroys the created session by disconnecting the associated
connection and cleanup the related objects.

=over

=item Synopsis

$version_object->destroy_session(session => $session);

=item Arguments

=over

=item session

The session object which needs to be destroyed.

=back

=item Returns

Nothing.

=back

=cut

=head1 AUTHOR/MAINTAINER

=over

=item NACL Development (dl-nacl-dev@netapp.com)

=back

=cut
