# $Id: //depot/prod/test/nacldev/lib/NACL/APISet/Version/Host/DFM.pm

#
# Copyright (c) 2001-2012 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Version class for DFM.
## @author Ramya.R@netapp.com, dl-mpo-nacl-dev@netapp.com
## @change 17-06-2012 deepthis Changing default interface of obtaining version to ZAPI 
## @status Shared
## @pod here

package NACL::APISet::Version::Host::DFM;

#
# Compiler Directives
#
use strict;
use warnings;

#
# Module imports: APISet
#
#
use NACL::APISet::Response::CLI::ParserUtils;
use base qw(NACL::APISet::Version);

#
# Exceptions imported
#

use NATE::BaseException qw(:try);

#
# NATE imports
#
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Data::Dumper;
use Params::Validate qw(:all);

#
# Base class
#
use base qw(NACL::APISet::Version);

our %dfm_release_table = (

    "3.8" => "DFM3",
    "4.0" => "DFM3",
    "5.0" => "DFM3",
    "5.1" => "DFM3",
    "5.2" => "DFM3",
    "6.0" => "DFM6",

);
###############################################################################
# Method:    new
# Objective: Constructor to create DFM version object
# Details:   See POD documentation at the end of this file.
###############################################################################

sub new {
    $Log->enter() if $may_enter;
    my ( $proto, %args ) = @_;
    my $class = ref($proto) || $proto;

    # Validate parameters and populate member variables
    my $self = NACL::APISet::validate_params(
        $class,
        params => \%args,
        spec   => {
            hostobj        => { isa  => "Hostrec", optional => 0 },
            category       => { type => SCALAR,    optional => 0 },
            shell_type     => { type => SCALAR,    optional => 0 },
            conn_overrides => { type => HASHREF,   optional => 1 },
        }
    );
    bless $self, $class;

    $Log->exit() if $may_exit;
    return $self;
} ## end sub new()

##############################################################################
#                               Private Methods
##############################################################################

##############################################################################
# Method:    map_release_to_codename
# Objective: Map release strings to internal codenames
# Details:   See POD documentation at the end of this file.
##############################################################################

sub map_release_to_codename {
    $Log->enter() if $may_enter;
    my $self = shift();
    my %args = NACL::APISet::validate_params(
        $self,
        params => \@_,
        spec   => { "release" => { type => SCALAR, optional => 0 } }
    );

    my $release  = lc( $args{release} );
    my $codename = $self->SUPER::map_release_to_codename(
        release_table => \%dfm_release_table,
        release       => $release
    );
    $Log->exit() if $may_exit;
    return $codename;
}

1;
##############################################################################
# Method:    _find_release
# Objective: Perform the discovery of release information
# Details:   See POD documentation at the end of this file.
##############################################################################

sub _find_release {
    my ( $self, %args ) = @_;

    $Log->enter() if $may_enter;

    # Get the connection. Using session class itself so that it will
    # do prep_service and it helps us to get a connetion object with
    # the connection overrides specified, if any.
    my ( $session, $conn );

    # ZAPI
    my $result;
    try {
        $session = $self->get_session(
            interface => "ZAPI",
            subtype   => "DFM"
        );
        my $command = "<dfm-about/>";
        my $release_info;
        my $response
            = $self->execute_ZAPI( command => $command, session => $session );
        if ( defined $response ) {
            my $parsed_output = $response->get_parsed_output();
            $Log->debug( Dumper $parsed_output);
            $release_info = $parsed_output->[0]->{version};

        }
        if ( defined $release_info ) {
            $self->{release} = trim_string( string => $release_info );

        }

    }
    catch NATE::BaseException with {

        my $exception = shift;
        $Log->debug(
            "Error in obtaining DFM version  " . $exception->text() );
	$Log->debug("Stack trace: " . $exception->stacktrace());
    }
    finally {
        $self->destroy_session( session => $session ) if ( defined $session );
    };

    if ( defined $self->{release} ) {
        $Log->exit() if $may_exit;
        return;
    }

    # CLI
    my $output;
    try {
        $session = $self->get_session(
            interface => "CLI",
            subtype   => "DFM"
        );

        # Find the information
        $conn   = $session->{connection};
        $output = $self->execute_CLI_command(
            command    => "dfm version",
            connection => $conn
        );
        if ( $output =~ /dfm.*\d+/ ) {
            $self->{release} = trim_string( string => $output );
        }
    }
    catch NATE::BaseException with {

        my $exception = shift;
        $Log->debug(
            "Error in obtaining DFM version : " . $exception->text() );
        $Log->debug("Stack trace: " . $exception->stacktrace()); 
    }

    finally {

        # Destroy the session
        $self->destroy_session( session => $session ) if ( defined $session );
    };

}

1;

=head1 NAME

NACL::APISet::Version::Host::DFM - Version library for DFM

=head1 SYNOPSIS

 use NACL::APISet::Version::Host::DFM;

=head1 DESCRIPTION

A C<NACL::APISet::Version::Host::DFM> is used to connect to a DFM Host and 
collect version related attributes. All version related attributes exposed from
this libary can also be set or get by the user. 

=cut

=head1 EXCEPTIONS

=over

=item NACL::APISet::Exceptions::ConnectionFailedException

=item NATE::BaseException

=back

=head1 CONSTRUCTORS

=head2 new

Returns a reference to a newly created C<NACL::APISet::Version::Host::DFM>.

=over 

=item Synopsis

$version = NACL::APISet::Version::Host::DFM->new(%args)

=item Arguments

=over

=item category

(Mandatory)
The category to use.  Valid values include "Host".

=item shell_type

(Mandatory)
The refers to the type of the shell host sent through 'hostobj'.
Valid values include "Windows" and "Linux".


=item hostobj

(Mandatory)
The hostrec object created for the resource.

=item conn_overrides

(Optional)
Hash reference to the list of connection overrides to be considered when
establishing the connection to the service.
Valid connection overrides are username, password, connid, conntype, ipaddress,
prompt_username and prompt_password.

=back

=item Returns

A reference to a DFM Version object

=back

=cut

=head1 METHODS


=head2 _find_release

This method discovers release information by creating the appropriate session
and execute various commands like "sysctl", "cat /BUILD", "system image show" 
to get information about release.

=over 

=item Synopsis

$release = $version_object->get_version_attribute(attribute => "release");

The above method will internally call $self->_find_release() if the information
about release atribute is not already present in $self. This method shouldn't be
directly invoked by the user.

=item Arguments

Nothing

=item Returns

Nothing. Set the release information in version object itself. 

=back


=head1 AUTHOR/MAINTAINER

=over

=item NACL Development (dl-nacl-dev@netapp.com)

=back

 $Id$
 Last revised $Date$ by $Author$

=cut

