#
# Copyright (c) 2011-13 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#
## @summary Version class for finding Version attributes specific to Linux.
## @author rawat@netapp.com, dl-nacl-dev@netapp.com
## @status Shared


package NACL::APISet::Version::Host::Linux;

#
# Compiler Directives
#
use strict;
use warnings;

#
# Module imports
#
use NACL::APISet::Session::CLI::Linux;
use NACL::APISet::Response::CLI::ParserUtils;

#
# NATE imports
#
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Params::Validate qw(:all);
use NATE::Exceptions::Argument qw(:try);

#
# base class
#
use base qw(NACL::APISet::Version::Host);

###############################################################################
# Method:    _get_os_revision
# Objective: Will return Operating System revision
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _find_os_revision {
    my ($self) = @_;
    $Log->enter() if $may_enter;

    my $session;
    my ($major, $minor);

    try {
        my $get_linux_session = sub {
            return $self->get_session(
                interface  => "CLI",
                subtype    => "Linux",
            );
        };


        my $hosttype = $self->{hostobj}->{hosttype};
        if ($hosttype =~ /redhat|rhel/i) {
            # Outout of the following command looks like this
            ##-------------------------------------------------------------
            # Red Hat Enterprise Linux Server release 5.4 (Tikanga)
            ##-------------------------------------------------------------
            $session = $get_linux_session->();
            my $output = $self->execute_CLI_command(
                connection  => $session->{connection},
                command     => "cat /etc/issue",
            );

            # We have separated getting OL major/minor revision from RHEL because
            # Oracle ships with three kernels now.
            # 1. RHEL kernel
            # 2. Kernel from upstream
            # 3. RHEL kernel with some patches
            # if '/etc/issue' file has Oracle in it that means the kernel is
            # upstream kernel and lots of packages changes(including multipath)
            # from RHEL kernel
            if ($output =~ /^\s*Oracle\s+Linux/i) {
                ($major, $minor) = $self->_get_ol_major_and_minor_revisions($session);
            } elsif ($output =~ /Red\s+Hat\s+Enterprise\s+Linux|Enterprise\s+Linux\s+Enterprise\s+Linux\s+Server\s+release/i) {
                ($major, $minor) = $self->_get_redhat_major_and_minor_revisions($session);
            }

            # OVM also has /etc/redhat-release file but content of the
            # file looks like this
            ##------------------------------------------------------------
            # Oracle VM server release 2.2.1
            ##------------------------------------------------------------
            elsif ($output =~ /Oracle\s+VM\s+server/i) {
                ($major, $minor) = $self->_get_ovm_major_and_minor_revisions($session);
            }
        } elsif ($hosttype =~ /suse|sles/i) {
            $session = $get_linux_session->();
            ($major, $minor) = $self->_get_suse_major_and_minor_revisions($session);
        }
    } finally {
        # Destroy the session
        $self->destroy_session(session => $session) if (defined $session);
    };

    $self->{os_revision}->{major} = $major;
    $self->{os_revision}->{minor} = $minor;

    $Log->exit() if $may_exit;
    return $major;
}

################################################################################
# Subroutine Name  _get_ol_major_and_minor_revisions
# Objective:       Will store major and minor revision in self if OS is Oracle Linux
# Details:         See POD documentation at the end of this file
################################################################################
sub _get_ol_major_and_minor_revisions {
    my ($self, $session) = @_;

    my $major;
    my $minor = 0;

    # The output of cat /etc/issue on Oracle Linux box looks like
    ##-------------------------------------------------------------
    # Oracle Linux Server release 5.6
    ##-------------------------------------------------------------
    # We are not using 'cat /etc/redhat-release' because this has
    # different kernel and lots of thing are different from RHEL
    my $output = $self->execute_CLI_command(
        connection  => $session->{connection},
        command     => "cat /etc/issue",
    );

    if ($output =~ /^\s*Oracle\s+Linux\s+Server\s+release\s+\d+/i) {
        ($major, $minor) = ($output =~ /release\s+(\d)\.?(\d*)/i);
    }

    $major = "OL" . $major;

    return ($major, $minor);
}
################################################################################
# Subroutine Name  _get_redhat_major_and_minor_revisions
# Objective:       Will store major and minor revision in self if OS is RedHat
# Details:         See POD documentation at the end of this file
################################################################################
sub _get_redhat_major_and_minor_revisions {
    my ($self, $session) = @_;

    my $major;
    my $minor = 0;

    my $cmd = "cat /etc/redhat-release";
    # The output of cat /etc/redhat-release on RHEL4 box looks like
    ##-------------------------------------------------------------
    # Red Hat Enterprise Linux AS release 4 (Nahant Update 8)
    ##-------------------------------------------------------------
    # And on RHEL5/RHEL6 it looks like
    ##-------------------------------------------------------------
    # Red Hat Enterprise Linux Server release 5.4 (Tikanga)
    ##-------------------------------------------------------------
    my $output = $self->execute_CLI_command(
        connection  => $session->{connection},
        command     => $cmd,
    );

    # Case 1: If OS is RHEL5 or RHEL6
    if ($output =~ /release\s+5|release\s+6/i) {
        ($major, $minor) = ($output =~ /release\s+(\d)\.?(\d*)/i);
    } else { # Case 2: If OS is RHEL4
        if ($output =~ /Update/i) {
            ($major, $minor) = ($output =~ /release\s+(\d+)\s+\(\w+\s+Update\s*(\d+)\)/i);
        } else {
            ($major) = ($output =~ /release\s+(\d+)\s+\(\w+\)/i);
        }
    }

    $major = "RHEL" . $major;

    return ($major, $minor);
}

################################################################################
# Subroutine Name  _get_ovm_major_and_minor_revisions
# Objective:       Will store major and minor revision in self if OS is OVM
# Details:         See POD documentation at the end of this file
################################################################################
sub _get_ovm_major_and_minor_revisions {
    my ($self, $session) = @_;

    my $major;
    my $minor = 0;

    # The output of cat /etc/ovs-release looks like
    ##-------------------------------------------------------------
    # Oracle VM server release 2.2.1
    ##-------------------------------------------------------------
    my $cmd = "cat /etc/ovs-release";
    my $output = $self->execute_CLI_command(
        connection  => $session->{connection},
        command     => $cmd,
    );
    ($major, $minor) = ($output =~ /Oracle\s+VM\s+server\s+release\s+(\d+\.\d+)\.?(\d*)/i);

    $major = "OVM" . $major;

    return ($major, $minor);
}

################################################################################
# Subroutine Name  _get_suse_major_and_minor_revisions
# Objective:       Will store major and minor revision in self if OS is SuSe
# Details:         See POD documentation at the end of this file
################################################################################
sub _get_suse_major_and_minor_revisions {
    my ($self, $session) = @_;

    my $major;
    my $minor = 0;

    my $cmd = "cat /etc/SuSE-release";
    # The output of cat /etc/SuSE-release looks like
    # SUSE Linux Enterprise Server 11 (x86_64)
    # VERSION = 11
    # PATCHLEVEL = 0
    my $output = $self->execute_CLI_command(
        connection  => $session->{connection},
        command     => $cmd,
    );
    my @output = split(/\n/, $output);
    ($major) = ($output[1] =~ /(\d+)/i);
    ($minor) = ($output[2] =~ /(\d+)/i);

    unless ($minor) {
        $minor = 0;
    }
    $major = "SLES" . $major;

    return ($major, $minor);
}

1;

=head1 NAME

NACL::APISet::Version::Host::Linux - Version library for Linux

=head1 SYNOPSIS

 use NACL::APISet::Version::Host::Linux;

=head1 DESCRIPTION

A C<NACL::APISet::Version::Host::Linux> is used to connect to Linux Host and
collect version related attributes. All version related attributes exposed from
this libary can also be set or get by the user.

=cut

=head1 CONSTRUCTORS

=head2 new

Returns a reference to a newly created C<NACL::APISet::Version::Host::Linux>.

=over

=item Synopsis

$version = NACL::APISet::Version::Host::Linux->new(%args)

=item Arguments

=over

=item category

(Mandatory)
The category to use.  Valid values include "Host".

=item shell_type

(Mandatory)
The refers to the type of the shell host sent through 'hostobj'.
Valid values include "Host"

=item hostobj

(Mandatory)
The hostrec object created for the resource.

=item conn_overrides

(Optional)
Hash reference to the list of connection overrides to be considered when
establishing the connection to the service.
Valid connection overrides are username, password, connid, conntype, ipaddress,
prompt_username and prompt_password.

=back

=item Returns

A reference to a new Linux/Solaris Version object

=back

=cut

