#
# Copyright (c) 2001-2018 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#
## @summary Base version class for 'Node' category
## @author Isshwarya.M@netapp.com, dl-nacl-dev@netapp.com
## @status Shared
## @pod here

package NACL::APISet::Version::Node;

#
# Compiler Directives
#
use strict;
use warnings;

#
# NATE imports
#
use Data::Dumper;
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Params::Validate qw(:all);
use NACL::APISet::Response::CLI::ParserUtils;
use NACL::APISet::_Mixins::CLI qw(_is_contain_question_mark);

#
# Base class
#
use base qw(NACL::APISet::Version);

#
# Module imports
#
use NACL::APISet::Version::Node::7Mode;
use NACL::APISet::Version::Node::CMode;
use NACL::APISet;

#
# Constants
#
use constant PRIVATE_VERSION => 'private_version';

#
# Exceptions imported
#
use NACL::APISet::Exceptions::ConnectionFailedException qw(:try);
use NACL::APISet::Exceptions::AttributeDiscoveryFailed qw(:try);
use NATE::Exceptions::Argument;
use NATE::BaseException qw(:try);
use Tie::IxHash;
use NACL::Global;

our %release_table;
tie %release_table, 'Tie::IxHash';

# Release table
# This release table will map the known internal and external release strings
# to internal codenames for use by versioning.

# NOTE: When adding new entries, please make sure that the strings are
#       written in lowercase. The values in the hash should have underscores
#       instead of using '.'
#       Also when a new release is added make sure the first string that is
#       added for a particular release is of the format release-name__release-number,
#       something like 'sierranevada__8.2.0'. This ordering is necessary for
#       functions which compares two releases.
%release_table = (
    "tricky.0" => "tricky_0",
    "10.0rc"   => "tricky_0",

    "10.0.1"   => "tricky_1",
    "tricky.1" => "tricky_1",

    "10.0.2"   => "tricky_2",
    "tricky.2" => "tricky_2",

    "10.0.3"   => "tricky_3",
    "tricky.3" => "tricky_3",

    "10.0.4"   => "tricky_4",
    "tricky.4" => "tricky_4",

    "boilermaker__8.0.0" => "boilermaker_0",
    "boilermaker.0"      => "boilermaker_0",
    "8.0x"               => "boilermaker_0",
    "8.0rc"              => "boilermaker_0",

    "boilermaker__8.0.1" => "boilermaker_1",
    "boilermaker.1"      => "boilermaker_1",
    "8.0.1"              => "boilermaker_1",
    
    "boilermaker__8.0.2" => "boilermaker_2",
    "boilermaker.2"      => "boilermaker_2",
    "8.0.2"              => "boilermaker_2",
    
    "boilermaker__8.0.3" => "boilermaker_3",    
    "boilermaker.3"      => "boilermaker_3",
    "8.0.3"              => "boilermaker_3",

    "boilermaker__8.0.4" => "boilermaker_4",    
    "boilermaker.4"      => "boilermaker_4",
    "8.0.4"              => "boilermaker_4",    

    "boilermaker__8.0.5" => "boilermaker_5",
    "boilermaker.5"      => "boilermaker_5",
    "8.0.5"              => "boilermaker_5",
    
    "rollingrock__8.1.0" => "rollingrock_0",
    "mainng__10.0.99"    => "rollingrock_0",
    "rollingrock.0"      => "rollingrock_0",
    "8.1.0"              => "rollingrock_0",
    
    "rollingrock__8.1.1" => "rollingrock_1",    
    "rollingrock.1"      => "rollingrock_1",
    "8.1.1"              => "rollingrock_1",
    
    "rollingrock__8.1.2" => "rollingrock_2",    
    "rollingrock.2"      => "rollingrock_2",
    "8.1.2"              => "rollingrock_2",

    "rollingrock__8.1.3" => "rollingrock_3",    
    "rollingrock.3"      => "rollingrock_3",
    "8.1.3"              => "rollingrock_3",

    "rollingrock__8.1.4" => "rollingrock_4",
    "rollingrock.4"      => "rollingrock_4",
    "8.1.4"              => "rollingrock_4",
	
    "sierranevada__8.2.0" => "sierranevada_0",
    "sierranevada.0"      => "sierranevada_0",
    "8.2.0"               => "sierranevada_0",

    "sierranevada__8.2.1" => "sierranevada_1",
    "sierranevada.1"      => "sierranevada_1",
    "8.2.1"               => "sierranevada_1",

    "sierranevada__8.2.2" => "sierranevada_2",
    "sierranevada.2"      => "sierranevada_2",
    "8.2.2"               => "sierranevada_2",

    # SN3 version is appended in the table to fix BURT 867193.
    "sierranevada__8.2.3" => "sierranevada_3",
    "sierranevada.3"      => "sierranevada_3",
    "8.2.3"               => "sierranevada_3",

    "sierranevada__8.2.4" => "sierranevada_4",
    "sierranevada.4"      => "sierranevada_4",
    "8.2.4"               => "sierranevada_4",

    "fullsteam__8.3.0"    => "fullsteam_0",
    "fullsteam.0"         => "fullsteam_0",
    "8.3.0"               => "fullsteam_0",

    "fullsteam__8.3.1"    => "fullsteam_1",
    "fullsteam.1"         => "fullsteam_1",
    "8.3.1"               => "fullsteam_1",

    "fullsteam__8.3.2"    => "fullsteam_2",
    "fullsteam.2"         => "fullsteam_2",
    "8.3.2"               => "fullsteam_2",


    "longboard__${NACL::Global::ONTAP_Release_Longboard_0}"    => "longboard_0",
    "longboard.0"                                              => "longboard_0",
    "${NACL::Global::ONTAP_Release_Longboard_0}"               => "longboard_0",

    "longboard__${NACL::Global::ONTAP_Release_Longboard_1}"    => "longboard_1",
    "longboard.1"                                              => "longboard_1",
    "${NACL::Global::ONTAP_Release_Longboard_1}"               => "longboard_1",

    "harpoon__${NACL::Global::ONTAP_Release_Harpoon_0}"        => "harpoon_0",
    "harpoon.0"                                                => "harpoon_0",
    "${NACL::Global::ONTAP_Release_Harpoon_0}"                 => "harpoon_0",

    "fattire__${NACL::Global::ONTAP_Release_FatTire_0}"        => "fattire_0",
    "fattire.0"                                                => "fattire_0",
    "${NACL::Global::ONTAP_Release_FatTire_0}"                 => "fattire_0",
);

use Class::MethodMaker [
    scalar => [
        {
            '-default_ctor' => sub {
                $Log->enter() if $may_enter;
                my $self    = $_[0];
                my $is_sksim = $self->is_sksim();
                my $is_debug;
                if ($is_sksim) {
                    # Cannot get build options for sksim, but sksim has
                    # properties of a debug build
                    $is_debug = 1;
                } else {
                    my $build_opt = $self->get_version_attribute(attribute => 'build_options');
                    $is_debug = ($build_opt =~ /debug/i) ? 1 : 0;
                }

                # debug builds are guaranteed to be non-promoted
                $self->is_promoted(0) if ($is_debug);
                $Log->exit($is_debug);
                return $is_debug;
            },
        },
        'is_debug',
    ],

    scalar => [
        {
            '-default_ctor' => sub {
                $Log->enter() if $may_enter;
                my $self    = $_[0];
                my $build_opt = $self->get_version_attribute(attribute => 'build_options');
                my $is_gcov = ($build_opt =~ /gcov/i) ? 1 : 0;
                $Log->exit($is_gcov);
                return $is_gcov;
            },
        },
        'is_gcov',
    ],

    # Enable using existing NACL scripts on an sksim, which is like
    # a stripped down ONTAP, with a 7Mode UI. Certain features typically
    # available in ONTAP are not available in sksim, which can be disabled
    # by reading this attribute
    scalar => [
        {
            '-default_ctor' => sub {
                $Log->enter() if $may_enter;
                my $self = $_[0];
                my $is_sksim = ($self->hostobj()->hosttype() =~ /sksim/) ? 1 : 0;
                $Log->exit($is_sksim);
                return $is_sksim;
            },
        },
        'is_sksim',
    ],
];

###############################################################################
# Method:    new
# Objective: Constructor to create CMode version object
# Details:   See POD documentation at the end of this file.
###############################################################################

sub new {
    $Log->enter() if $may_enter;
    my ( $proto, %args ) = @_;
    my $class = ref($proto) || $proto;

    # Validate parameters and populate member variables
    my $self = NACL::APISet::validate_params(
        $class,
        params => \%args,
        spec   => {
            hostobj        => { isa  => "Hostrec", optional => 1 },
            category       => { type => SCALAR,    optional => 0 },
            shell_type     => { type => SCALAR,    optional => 0 },
            conn_overrides => { type => HASHREF,   optional => 1 },
            console_connection  =>
                { isa  => "Connectrec", optional => 1 },
            cli_connection => { isa  => "Connectrec", optional => 1 },
        }
    );
    bless $self, $class;

    $Log->exit() if $may_exit;
    return $self;
} ## end sub new()


##############################################################################
# Method:    execute_CLI_command
# Objective: Execute the sent command and gathers raw output
# Details:   Refer to POD documentation at the end of the file.
##############################################################################

sub execute_CLI_command {
    my ($self, @args) = @_;
    $Log->enter() if $may_enter;
    my %args = NACL::APISet::validate_params(
        $self,
        params => \@args,
        spec   => {
            command    => { type => SCALAR,       optional => 0 },
            connection => { isa  => "Connectrec", optional => 0 }
        }
    );
    # Fix for burt698232
    my $command = $args{command};
    my $contain_question_mark = 0;
    if ($self->_is_contain_question_mark(command => $command)) {
        $contain_question_mark = 1;
        $self->SUPER::execute_CLI_command (
            connection => $args{connection},
            command    => "set -active-help false",
        );
    }
    
    my $raw_output = $self->SUPER::execute_CLI_command(%args);
    
    if ($contain_question_mark == 1) {
        $self->SUPER::execute_CLI_command (
            connection => $args{connection},
            command    => "set -active-help true"
        );
    }
	
    $Log->exit() if $may_exit;
    return $raw_output;	
}

##############################################################################
# Method:    parse_ngsh_output
# Objective: Parser for parsing CMode type of output
# Details:   See POD documentation at the end of this file.
##############################################################################

sub parse_ngsh_output {
    my $self = shift;

    $Log->enter() if $may_enter;
    my %args = NACL::APISet::validate_params(
        $self,
        params => \@_,
        spec   => {
            command    => { type => SCALAR, optional => 0 },
            raw_output => { type => SCALAR, optional => 0 }
        }
    );
    my $output = $args{raw_output};

    # Remove echo if any there
    remove_echo( "output_ptr" => \$output, "command" => $args{command} );

    $Log->exit() if $may_exit;

    # Parse the output
    return $self->_parse_ngsh_output( raw_output => $output, command => $args{command} );
}

##############################################################################
# Method:    is_classic
# Objective: Checks whether a build is a 7G/Classic build or not
# Details:   See POD documentation at the end of this file.
##############################################################################

sub is_classic {
    $Log->enter() if $may_enter;
    my $self = shift();
    if( defined $self->{is_classic} ) {
        $Log->debug("Cached information is available and it is is_classic" . " => "
                . $self->{is_classic} );
        $Log->exit() if $may_exit;
        return $self->{is_classic};
    }
    
    # else return 0 if it not a classic build   
    $Log->exit() if $may_exit;
    return 0;
}

##############################################################################
# Method:    map_release_to_codename
# Objective: Map release strings to internal codenames
# Details:   See POD documentation at the end of this file.
##############################################################################

sub map_release_to_codename {
    $Log->enter() if $may_enter;
    my $self = shift();
    my %args = NACL::APISet::validate_params(
        $self,
        params => \@_,
        spec   => { "release" => { type => SCALAR, optional => 0 } }
    );
    my $release = lc( $args{release} );
    if ( $self->{is_classic} ) {

        # For 7g systems, there is no mapping for the release names because
        # those release names are defined by us based on wafl version.
        # Hence the mapped codename is same as obtained release name.
        $Log->debug( "There is no mapping of release names for 7G systems."
                . " Returning release name $release itself" );
        $Log->exit() if $may_exit;
        return $release;
    }

    # Check for patch releases starting with __, example: __8.2p5d4 and 
    # convert it to readable format: 8.2p5d4
    if ( $release =~ /^__(\S+)$/ ) {
        $release = $1;	
    }   
    # If the release string is obtained as 8.1, 8.1rc, 8.1p1 etc,
    # we modify it to the standard release number 8.1.0
    if ( $release =~ s/^(\d\.\d(\.*\d*))\w*$/$1/ ) {
        # if we get strings like 8.2.1x20 then map it to 8.2.1 and not 8.2.1.0
        $release = $release . '.0' if (!$2);
    }
    my $codename = $self->SUPER::map_release_to_codename(
        release_table => \%release_table,
        release       => $release
    );
    $Log->exit() if $may_exit;
    return $codename;

}

##############################################################################
#                               Private Methods
##############################################################################

##############################################################################
# Method:    _find_wafl_version
# Objective: Perform the discovery of wafl version information
# Details:   It runs the relevant commands either in Nodescope or 7M shell
#            and set the wafl version information in version manager object
##############################################################################

sub _find_wafl_version {
    my ( $self, %args ) = @_;

    $Log->enter() if $may_enter;

    # Try running the "stats" command from 7Mode shell.
    my $output;
    my $session = $args{session};

    my $command
        = "priv set -q diag; stats show volume:"
        . "*"
        . ":wv_fsinfo_fs_version; result -l";
    $command = "system node run local -command \"$command\""
        if ( $self->{shell_type} eq 'CMode' );
    $output = $self->execute_CLI_command(
        command    => $command,
        connection => $session->{connection}
    );
    my @lines = get_lines( raw_output => $output,first_line => "volume:");
    if ( scalar(@lines) == 0 ) {
        $Log->exit() if $may_exit;
        NACL::APISet::Exceptions::AttributeDiscoveryFailed->throw(
            "Error in discovering wafl version: stats show command "
                . "returned unexpected output and the raw output is:\n$output"
        );
    }
    $lines[0] =~ /volume:\S+:wv_fsinfo_fs_version:(\d+)/;
    $self->{wafl_version} = $1 if ( defined $1 );
    $Log->exit() if $may_exit;
}

##############################################################################
# Method:    _find_release
# Objective: Perform the discovery of release information
# Details:   This method executes commands in systemshell so as to collect
#            information about release and set the value in version object
#            itself
##############################################################################

sub _find_release {
    my ( $self, %args ) = @_;

    $Log->enter() if $may_enter;
    my $conn = $args{connection};
    my $output;

    # Try reading BUILD file
    
    try {
        $self->_command_cat_build(connection => $conn);

    } catch NATE::BaseException with {

        # Catch any exception so as to try the possibility of running
        # commands from CMode shell.
        my $exception = shift;
        $Log->debug( "Error in obtaining branch from systemshell: "
                . $exception->text() );
    };  
    
    if ( defined $self->{release} ) {
        $Log->exit() if $may_exit;
        return;
    }
    
    # Try "sysctl" command
    $output = $self->execute_CLI_command(
        command    => "sysctl kern.osrelease",
        connection => $conn
    );
    $output =~ /kern\.osrelease:(.*)/;
    $output = $1;
    if ( defined $output ) {
        $self->{release} = trim_string( string => $output );
        $Log->exit() if $may_exit;
        return;
    }
    
    $Log->exit() if $may_exit;
    return;
}

##############################################################################
# Method:    _find_change_number
# Objective: Perform the discovery of change number information
# Details:   This method executes commands in systemshell and clustershell
#            so as to collect information about change number and set the value
#            in version object itself
##############################################################################

sub _find_change_number {
    my ( $self, %args ) = @_;

    $Log->enter() if $may_enter;
    my $conn = $args{connection};
    my $output;

    # Try reading BUILD file
    
    try {
    $self->_command_cat_build(connection => $conn);

    } catch NATE::BaseException with {

        # Catch any exception so as to try the possibility of running
        # commands from CMode shell.
        my $exception = shift;
        $Log->debug( "Error in obtaining branch from systemshell: "
                . $exception->text() );
    };
    
    if ( defined $self->{change_number} ) {
        $Log->exit() if $may_exit;
        return;
    }

    # Try running the ident command
    $output = $self->execute_CLI_command(
        connection => $conn,
        command    => "ident /boot/modules/maytag.ko|grep NtapChangeNumber"
    );
    $output =~ /\$NtapChangeNumber:(.*)\s*\$/;
    $output = $1;
    if ( defined $output ) {
        $self->{change_number} = trim_string( string => $output );
        $Log->exit() if $may_exit;
        return;
    }
    $Log->exit() if $may_exit;
    return;
}

##############################################################################
# Method:    _find_client
# Objective: Perform the discovery of client information
# Details:   This method executes commands in systemshell so as to collect 
#            information about client and set the value in version 
#            object itself
##############################################################################

sub _find_client {
    my ($self, %args) = @_;
    
    $Log->enter() if $may_enter;
    my $conn = $args{connection};
    my $output;
    # Try reading BUILD file
   
    $self->_command_cat_build(connection => $conn);
   
    $Log->exit() if $may_exit;

}

##############################################################################
# Method:    _find_build_options
# Objective: Perform the discovery of build options information
# Details:   This method executes commands in systemshell
#            so as to collect information about build options and set the value
#            in version object itself
##############################################################################

sub _find_build_options {
    my ( $self, %args ) = @_;

    $Log->enter() if $may_enter;
    my $conn = $args{connection};
    my $output;

    try {
        $self->_command_cat_build(connection => $conn);

    } catch NATE::BaseException with {

        # Catch any exception so as to try the possibility of running
        # other commands to get build options.
        my $exception = shift;
        $Log->debug( "Error in obtaining branch from systemshell: "
                . $exception->text() );
    };
    
    if ( defined $self->{build_options} ) {
        $Log->exit() if $may_exit;
        return;
    }

    # Try running the ident command
    $output = $self->execute_CLI_command(
        connection => $conn,
        command    => "ident /boot/modules/maytag.ko|grep NtapBuildOpts"
    );
    $output =~ /\$NtapBuildOpts:(.*)\s*\$/;
    $output = $1;
    if ( defined $output ) {
        $self->{build_options} = trim_string( string => $output );
        $Log->exit() if $may_exit;
        return;
    }

    # Unable to find build options information. Just return.
    $Log->exit() if $may_exit;
    return;
}

##############################################################################
# Method:    _find_build_path
# Objective: Perform the discovery of build path information
# Details:   This method executes commands in systemshell and nodescope
#            so as to collect information about build path and set the value
#            in version object itself
##############################################################################

sub _find_build_path {
    my ( $self, %args ) = @_;

    $Log->enter() if $may_enter;
    my $conn = $args{connection};
    my $output;

    # Try sysctl command
    $output = $self->execute_CLI_command(
        command    => "sysctl kern.ontap_os_root",
        connection => $conn
    );
    $output =~ /kern\.ontap_os_root:(.*)/;
    $output = $1;
    if ( defined $output ) {
        $self->{build_path} = trim_string( string => $output );
        $self->{build_path} =~ s/(?:NetApp|Data ONTAP) Release (.*)/$1/;
        $Log->exit() if $may_exit;
        return;
    }

    # Try running the ident command
    $output = $self->execute_CLI_command(
        connection => $conn,
        command    => "ident /boot/modules/maytag.ko|grep NtapBuildPath"
    );
    $output =~ /\$NtapBuildPath:(.*)\s*\$/;
    $output = $1;
    if ( defined $output ) {
        $self->{build_path} = trim_string( string => $output );
        $Log->exit() if $may_exit;
        return;
    }
    $Log->exit() if $may_exit;
    return;
}

##############################################################################
# Method:    _find_branch
# Objective: Perform the discovery of branch information
# Details:   This method executes commands in systemshell so as to collect
#            information about branch and set the value in version object itself
##############################################################################

sub _find_branch {
    my ( $self, %args ) = @_;

    $Log->enter() if $may_enter;
    my $conn = $args{connection};
    my $output;

    try {
    $self->_command_cat_build(connection => $conn);

    } catch NATE::BaseException with {

        # Catch any exception so as to try the possibility of running
        # commands from other commands.
        my $exception = shift;
        $Log->debug( "Error in obtaining branch from systemshell: "
                . $exception->text() );
    };
    
    if ( defined $self->{branch} ) {
        $Log->exit() if $may_exit;
        return;
    }
    
    # Try running the ident command
    $output = $self->execute_CLI_command(
        connection => $conn,
        command    => "ident /boot/modules/maytag.ko|grep NtapBranchName"
    );
    $output =~ /\$NtapBranchName:(.*)\s*\$/;
    $output = $1;
    $output =~ s/DOT\///;
    if ( defined $output ) {
        $self->{branch} = trim_string( string => $output );
        $Log->exit() if $may_exit;
        return;
    }
    $Log->exit() if $may_exit;
    return;
}

##############################################################################
# Method:    _find_node_version
# Objective: Perform the discovery of node version
# Details:   This method executes commands in systemshell so as to collect
#            information about node version and set the value in version object
#            itself
##############################################################################

sub _find_node_version {
    my ( $self, %args ) = @_;

    $Log->enter() if $may_enter;
    my $conn = $args{connection};
    my $output;

    # Try "sysctl" command
    $output = $self->execute_CLI_command(
        connection => $conn,
        command    => "sysctl kern.os.vendorinfo.ontap_version"
    );
    $output =~ /kern\.os\.vendorinfo\.ontap_version:(.*)/;
    $output = $1;
    if ( defined $output ) {
        $self->{node_version} = $self->_extract_node_version(
            version => trim_string( string => $output ) );
        $Log->exit() if $may_exit;
        return;
    }
    $Log->exit() if $may_exit;
    return;
}

##############################################################################
# Method:    _extract_node_version
# Objective: Get the node version alone by striping out the extra portion
##############################################################################
sub _extract_node_version {
    my ( $self, %args ) = @_;

    $Log->enter() if $may_enter;
    my $text = $args{version};
    my $version;
    # test for the regex is present in lib/NACL/UnitTest/versioning_basic.thpl
    # TC: ut_18_check_regex_in_extract_node_version
    if (    $text
        and $text =~ /(?:NetApp|Data ONTAP) Release (\d+(\.\d)+\w*)\s*\S*:?/ )
    {
        $version = $1;
    } elsif ( $text
        and $text =~ /(?:NetApp|Data ONTAP) Release ([\w\.]+)\s*\S*:?/ )
    {
        $version = $1;
    } elsif ( $text and $text =~ /(?:NetApp|Data ONTAP) Release/ ) {
        $version = PRIVATE_VERSION;
    }

    $Log->exit() if $may_exit;
    return $version;
}

##############################################################################
# Method:    _find_build_arch_type
# Objective: Perform the discovery of build architecture and type
# Details:   This method executes ident /boot/modules/maytag.ko|grep NtapBuildOpts
#            in systemshell.
#            so as to collect information about build architecture and set the value
#            in version object itself
#            typical values would be like x86_64.debug or x86_64
##############################################################################

sub _find_build_arch_type {
    my ( $self ) = @_;
    $Log->enter() if $may_enter;
    my ( $output, $session );
    my %session_opt = ( interface => "CLI" );
    if ( $self->isa('NACL::APISet::Version::Node::CMode') ) {
       $session_opt{subtype} = "CMode::Shell";
       $session_opt{connid} = "c-system";
    } elsif ( $self->isa('NACL::APISet::Version::Node::7Mode') ) {
       $session_opt{subtype} = "7Mode::Shell";
       $session_opt{connid} = "7-system";
    } else {
       $Log->exit() if $may_exit;
       NACL::APISet::Exceptions::AttributeDiscoveryFailed->throw(
            "Unable to discover build arch and type");
    }
    try {
        $session = $self->get_session(
            %session_opt
        );
        my $conn =  $session->{connection};
        $output = $self->execute_CLI_command(
            connection => $conn,
            command    => "ident /boot/modules/maytag.ko|grep NtapBuildOpts"
        );
    }
    finally {
        # Destroy the session
        $self->destroy_session( session => $session ) if ( defined $session );
    };


    $output =~ /\$NtapBuildOpts:(.*)\s*\$/;
    $output = $1;
    if ( defined $output ) {
        $self->{build_arch_type} = trim_string( string => $output );
        $Log->exit() if $may_exit;
        return;
    }

    # Unable to find build architecture and type information. Just return.
    $Log->exit() if $may_exit;
    return;
}

##############################################################################
# Method:    _parse_ngsh_output
# Objective: Internsl parser routine for parsing CMode type of output
# Details:   This method is a generic parser for parsing CMode type of output
#            where parsing is done based on CLI separator
##############################################################################

sub _parse_ngsh_output {
    $Log->enter() if $may_enter;

    my ( $self, %args ) = @_;

    my $want_array = [];

    my $output        = $args{raw_output};
    my $CLI_separator = "X;X";

    # just skip blank lines
    my @lines = split( /\n/, $output );
    @lines = grep {/\w/} @lines;

    # If the output has nothing than new lines (no output), we can return back
    # here itself
    if ( !scalar(@lines) ) {
        $self->{parsed_output} = [];
        $Log->exit() if $may_exit;
        return;
    }

    my ( $fields, $heading, @tmp_output ) = @lines;
    my ( @fields, @heading );
    if ( defined $fields ) {
        @fields = split( $CLI_separator, $fields );
    }
    if ( defined $heading ) {
        @heading = split( $CLI_separator, $heading );
    }

    my @not_parsed = ();

    $Log->debug( "Fields:\n" . Dumper(@fields) );
    $Log->debug( "Heading:\n" . Dumper(@heading) );

 # Some output still contains embedded newlines in the fields so we cannot
 #  totally depend on the @tmp_output containing a single line for each row in
 #  in the CLI. So, loop through it once and find any elements of @array
 #  (the row) which do not contain the right number of elements (values
 #  for the row). The right number should be the number of elements of
 #  @heading.
 #
 # For any discrepencies, try to reform the row by combining multiple
 #  elements of @tmp_output into a new @output.
    my @output;
    my $partial_line = "";

    foreach my $line (@tmp_output) {

# We're here to parse the output into key/value structures, save off skipped lines
        if ( $line =~ /^I((?i)NFO)/ ) {
            $Log->debug("Found INFO or Info line. Ignoring: \"$line\"");
            push( @not_parsed, $line );
            next;
        } elsif ( $line =~ /-showseparator \"$CLI_separator\"/ ) {

       #Ignore any separator entry that is returned via the debug_log_show cmd
            $Log->debug(
                "Found output separator config line. Ignoring: \"$line\"");
            push( @not_parsed, $line );
            next;
        }
        $self->{not_parsed} = \@not_parsed;

        my @values = split( $CLI_separator, $line );
        if ( @values != @heading ) {
            $partial_line .= $line;

            # make sure we have seen the end of the row
            next if ( $partial_line !~ /$CLI_separator$/ );

            my @partial = split( $CLI_separator, $partial_line );
            if ( @partial == @heading ) {
                $Log->debug("done with merge: $partial_line");
                push( @output, $partial_line );
                $partial_line = "";
            } elsif ( @partial > @heading ) {

                # ERROR condition
                my $error_text;
                $error_text = "Error while parsing CMode CLI output!\n";
                $error_text .= "Failed to handle embedded newlines in the "
                    . "CLI output\n";
                $error_text .= "line=$line\n";
                $error_text .= "partial_line=$partial_line";
                NACL::APISet::Exceptions::ArgumentDiscoveryFailed->throw(
                    $error_text);
            }
        } else {
            push( @output, $line );
        }
    }

    # as per burt#895084, specific handling for command 'system node uichanges..'
    # to process left out partial output for parsing.
    if ( (defined $args{command}) && ($args{command} =~ /system node uichanges/) && ($partial_line =~/$CLI_separator$/) ) {
	$output[$#output] .= $partial_line;
    }

    foreach my $line (@output) {
        my %output = ();
        my $index  = 0;
        foreach my $field ( split( $CLI_separator, $line ) ) {
            ##################################################################
            # NOTE: The splitting of a list field on a comma will break
            # if the field is of type list AND accepts a list of text values
            # (or some other field type) that supports a comma.
            # We have not seen instances of this type of field yet but will
            # evaluate if problems occur
            ##################################################################

            ##################################################################
            # If the field starts with quotes, split the individual strings
            # separated by comma-space. Otherwise, treat entire thing as one
            # string. Some job output is missing the trialing ".
            ##################################################################
            if ( $field =~ /^\".+\"?$/ || $field =~ /^\{.+\}$/ ) {
                if ( $field =~ /, \"/ ) {
                    my $array_ref = [ split( /, /, $field ) ];
                    $Log->debug("array formed because of '$field'");
                    $output{ $fields[$index] } = $array_ref;
                } else {
                    $output{ $fields[$index] } = $field;
                }
            } elsif ( $field =~ /, ?/ ) {
                my $array_ref = [ split( /, ?/, $field ) ];
                $Log->debug("array formed because of '$field'");
                $output{ $fields[$index] } = $array_ref;
            } else {
                $output{ $fields[$index] } = $field;
            }
            $index++;
        }

        # Below line removes hash entry with key as "\r" that is left
        # after parsing all other lines for key value pairs.
        delete $output{"\r"} if exists( $output{"\r"} );

        push( @$want_array, \%output );
    }
    $Log->debug( "Parsed output is:\n" . Dumper($want_array) );
    $Log->exit() if $may_exit;

    return $want_array;
}

##############################################################################
# Method:    _convert
# Objective: Convert only the leaf nodes in the input xml into scalar
# Details:  This function recursively traverses through each of the node till
# a leaf node is recahed and checks whether the leaf node is scalar or an array.
# If it is an array, then it will store the value in leaf node as simple scalar.
##############################################################################

sub _convert {
    my $arr_ref = shift;
    foreach my $hash_ref (@$arr_ref) {
        my @keys = keys( %{$hash_ref} );
        foreach my $key (@keys) {
            if ( ref( $hash_ref->{$key} ) eq 'ARRAY'
                && !ref( $hash_ref->{$key}->[0] ) )
            {
                if ( scalar( @{ $hash_ref->{$key} } ) == 1 ) {
                    $hash_ref->{$key} = $hash_ref->{$key}->[0];
                }
            } elsif ( ref( $hash_ref->{$key}->[0] ) ) {
                _convert( $hash_ref->{$key} );
            }
        }
    }
}

sub _command_cat_build {

    my ( $self, %args ) = @_;
    $Log->enter() if $may_enter;
    
    my $conn = $args{connection};
    my $output;
    
    my %attribute_to_key = (
        'BRANCH' => 'branch',
        'CLIENT' => 'client',
        'CHANGE' => 'change_number',
        'TYPE'   => 'build_options',
    );

    # Try reading BUILD file
    $output = $self->execute_CLI_command(
        command    => "cat /BUILD",
        connection => $conn
    );
        
    foreach my $attribute (keys %attribute_to_key) {
        $output =~ /$attribute:(.*)/;
        my $value = $1;
        $value =~ s/DOT\/// if ($attribute eq 'BRANCH');
        if ( defined $value ) {
            $self->{$attribute_to_key{$attribute}} = trim_string( string => $value );
        }
    }
    
    $Log->exit() if $may_exit;
    return;
}

##############################################################################
# Method:    is_version_eq
# Objective: Checks whether the version running on the build is equal to the 
#            given version string
##############################################################################
sub is_version_eq {
    my ( $self, %args ) = @_;
    $Log->enter() if $may_enter;
    my $result = $self->_compare(release => $args{release}, operation => "eq");
    $Log->exit();   
    return $result;
}

##############################################################################
# Method:    is_version_ne
# Objective: Checks whether the version running on the build is not equal to the 
#            given version string
##############################################################################
sub is_version_ne {
    my ( $self, %args ) = @_;
    $Log->enter() if $may_enter;
    my $result = $self->_compare(release => $args{release}, operation => "ne");
    $Log->exit();   
    return $result;
}

##############################################################################
# Method:    is_version_gt
# Objective: Checks whether the version running on the build is greater than 
#            the given version string
##############################################################################
sub is_version_gt {
    my ( $self, %args ) = @_;
    $Log->enter() if $may_enter;
    my $result = $self->_compare(release => $args{release}, operation => "gt");
    $Log->exit();   
    return $result;
}

##############################################################################
# Method:    is_version_ge
# Objective: Checks whether the version running on the build is greater than 
#            or equal to the given version string
##############################################################################
sub is_version_ge {
    my ( $self, %args ) = @_;
    $Log->enter() if $may_enter;
    my $result = $self->_compare(release => $args{release}, operation => "ge");
    $Log->exit();   
    return $result;
}

##############################################################################
# Method:    is_version_lt
# Objective: Checks whether the version running on the build is less than the 
#            given version string
##############################################################################
sub is_version_lt {
    my ( $self, %args ) = @_;
    $Log->enter() if $may_enter;
    my $result = $self->_compare(release => $args{release}, operation => "lt");
    $Log->exit();   
    return $result;
}

##############################################################################
# Method:    is_version_le
# Objective: Checks whether the version running on the build is less than or
#            equal to the given version string
##############################################################################
sub is_version_le {
    my ( $self, %args ) = @_;
    $Log->enter() if $may_enter;
    my $result = $self->_compare(release => $args{release}, operation => "le");
    $Log->exit();   
    return $result;
}

# Helper routine to compare two releases.   
sub _compare {
    my $self = shift ;
    my %args = NACL::APISet::validate_params(
        $self,
        params => \@_,
        spec   => { "release" => { type => SCALAR, optional => 0 },
                    "operation" => { type => SCALAR, optional => 0 },
        },
    );
    $Log->enter() if $may_enter;

    my $release = $args{release};
    my $operation = $args{operation};
    
    unless ($release =~ /^(\d\.\d)$|^(\d\.\d\.\d)$/) {
        $Log->exit() if $may_exit;
        NATE::Exceptions::Argument->throw( "An invalid value was provided to the release "
        . "parameter. Value obtained is $release. Expected format is "
        . "major.minor or major.minor.sub" 
        );
    }
    if ($release =~ /^\d\.\d$/) {
        # This is for cases when the user provides just the major and minor
        # versions of the release. Eg: "8.1".
        # For such cases append a ".0" at the end of the release string.
        
        $release = $release . '.0'; 
    }       
    my $ref_release = $self->get_version_attribute(attribute => 'release');
    my $release_code_in_numeric = $self->map_release_to_numeric_version(release => $ref_release);
    
    $Log->debug("Performing operation $release_code_in_numeric $operation $release");
 
    my $result = eval "$release_code_in_numeric $operation $release" ? 1 : 0;   
    
    if ($@) {
        $Log->exit() if $may_exit;
        NATE::BaseException->throw( "Error while computing "
        . "$release_code_in_numeric $operation $release : $@"
        );
    }
    
    $Log->exit() if $may_exit;
    return $result;
}

sub map_release_to_numeric_version {
    my ( $self, %args ) = @_;
    my $release = $args{release};
    $Log->enter() if $may_enter;
    my $not_found;
    my $ref_release =$self->map_release_to_codename(release => $release);

    # Special handling needed for IC. See burt 694033.
    if ($ref_release =~ /Ironcity/i) {
        $Log->exit() if $may_exit;
        return '7.3.0';
    }

    my %reverse_hash = reverse %release_table;
    my $release_code_in_numeric = $reverse_hash{$ref_release};
    #Process the string
    if (defined $reverse_hash{$ref_release}) {
        if ($release_code_in_numeric =~ /.*__(.+)/) {
            $release_code_in_numeric =~ s/.*__(.+)/$1/;
        } else {
            $not_found = 1;
        }
    } else {
        # Trying to find the numeric release intelligently
        if ($release =~ /^(\w+?)__(\d+\.\d+\.\d+)$|^(\d+\.\d+\.\d+)$/) {
            $release_code_in_numeric = (defined $2)? $2 : $3;
            $Log->debug( "No matching entry found for $release. Numeric release has " 
            . "been intelligently formed as $release_code_in_numeric");
        } elsif ($release =~ /^(\d+\.\d+[\.\d+]*)[(P|p)\d+]*(X|x)\d+$/) {
            # This is for releases of the form 8.2.3X3 or 9.2X17 or 9.3P1X4
            $release_code_in_numeric = $1;
        } else {
            $not_found = 1;
        }
    }

    if ($not_found) {
        $Log->exit() if $may_exit;
        NATE::BaseException->throw("Internal Error: Expected release code of the "
        . "form releasename__releasenumber, but found $release_code_in_numeric");	
    }

    $Log->exit() if $may_exit;
    return $release_code_in_numeric;
}

# Helper routine for is_promoted. Used in CMode and Vserver context
sub _is_promoted_cmode_priv {
    $Log->enter() if $may_enter;
    my $self = $_[0];
    my $is_promoted;
     
    # Try the possibility by getting the possible privilege levels.
    my $session;
    try {
        $session = $self->get_session(
            interface => "CLI",
            subtype   => "CMode"
        );

        my $outp = $self->execute_CLI_command(
            connection => $session->{connection},
            command    => "set -privilege ?"
        );

        # Check if output has 'test' as one of the privileges
        if ($outp =~ /test/i) {
            $is_promoted = 0;
        } else {
            $is_promoted = 1;
        }
    } finally {
        # Destroy the session
        $self->destroy_session( session => $session ) if ( defined $session );
    };
    
    unless (defined $is_promoted) {
        $Log->exit() if $may_exit;
        NACL::APISet::Exceptions::AttributeDiscoveryFailed->throw(
            "Unable to discover whether the build is promoted or not");
    }

    $Log->exit($is_promoted);
    return $is_promoted;
}

sub _is_promoted_systemshell {
    $Log->enter() if $may_enter;
    my $self = $_[0];
    my ($is_promoted, $session);
    try {
        if ( ref($self) =~ /CMode/ ) {
            $session = $self->get_session(
                interface => "CLI",
                subtype   => "CMode::Shell",
                connid    => "c-system"
            );
        } else {
            $session = $self->get_session(
                interface => "CLI",
                subtype   => "7Mode::Shell",
                connid    => "7-system"
            );
        }

        # Not catching any exceptions as this is the only possible way to
        # obtain the information.
        my $output = $self->execute_CLI_command(
            command    => "sysctl kern.ontap_os_release",
            connection => $session->{connection}
        );
        $output =~ /kern\.ontap_os_release:(.*)/;
        my $info = $1;
        $info = trim_string(string => $info);
        $Log->debug("Build type: $info");
        $is_promoted = ($info =~ /Development/i) ? 0 : 1;
    }
    finally {
        # Destroy the session
        $self->destroy_session(session => $session) if (defined $session);
    };
    
    unless (defined $is_promoted) {
        $Log->exit() if $may_exit;
        NACL::APISet::Exceptions::AttributeDiscoveryFailed->throw(
            "Unable to discover whether the build is promoted or not");
    }

    $Log->exit($is_promoted);
    return $is_promoted;
}

##############################################################################
# Method:    is_domain_crossing
# Objective: Checks whether a build is a domain crossing build or not
# Details:   This method executes commands so as to collect information about
#            whether the build is a domain crossing build or not and set the
#            value in version object itself
##############################################################################

sub is_domain_crossing {
    $Log->enter() if $may_enter;
    my ( $self, %args ) = @_;
    my $is_domain_crossing;
    my $session;

    if(defined $self->{is_domain_crossing}) {
        $Log->debug( "is_domain_crossing returned: " . $self->{is_domain_crossing} );
        $Log->exit() if $may_exit;
        return $self->{is_domain_crossing};
    }

    try {
        if ( ref($self) =~ /CMode/ ) {
            $session = $self->get_session(
            interface => "CLI",
            subtype   => "CMode",
            );
        } else {
            $session = $self->get_session(
            interface => "CLI",
            subtype   => "7Mode",
            );
        }
        my $output = $self->execute_CLI_command(
            connection => $session->{connection},
            command    => "version -v"
        );
        if ($output =~ /<1A>|<1dA>/) {
            $self->{is_domain_crossing} = 1;
        } else {
            $self->{is_domain_crossing} = 0;
        }
    } finally {
        # Destroy the session
        $self->destroy_session( session => $session ) if ( defined $session );
    };

    $Log->debug( "is_domain_crossing returned: " . $self->{is_domain_crossing} );
    $Log->exit() if $may_exit;
    return $self->{is_domain_crossing};
}

1;

=head1 NAME

NACL::APISet::Version::Node - Base version class for Node category

=head1 SYNOPSIS

 use NACL::APISet::Version::Node;

=head1 DESCRIPTION

A C<NACL::APISet::Version::Node> will act as a factory class for methods
common to various set like CMode and 7Mode.

=cut

=head1 EXCEPTIONS

=over 

=item NACL::APISet::Exceptions::ConnectionFailedException

=item NACL::APISet::Exceptions::AttributeDiscoveryFailed

=item NATE::BaseException

=back

=cut

=head1 METHODS

=head2 execute_CLI_command

This method will execute the command using the connection object sent and
gathers raw output.

=over 

=item Synopsis

$output = $version_object->execute_CLI_command( command => "version", 
    connection => $connection );

=item Arguments

=over

=item command

The input command line to be executed on the service

=item connection

The connection object to be used for executing the command.

=back

=item Returns

This method will returns the raw output returned by the service without any
processing. 

=back


=head2 parse_ngsh_output

This method parses the CMode type of output into a Perl data structure
so that version class can access various information from the output.
This is a generic parser which makes use of CLI seprator for parsing the
raw output into array of hashes structure.

=over 

=item Synopsis

$parsed_output = $version_object->parse_ngsh_output(raw_output => $output);

=item Arguments

=over

=item raw_output

The raw output to be parsed into a perl data structure

=back

=item Returns

This method will returns the parsed output as a reference to an array 

=back

=head2 map_release_to_codename

For a same release, we can have different release strings like internal and
external release strings and hence map all release strings referring a 
particular release to its internal codename.

=over 

=item Synopsis

$codename = $version_object->map_release_to_codename(release => $release);

=item Arguments

=over

=item release

The ONTAP release string.

=back

=item Returns

This method will return the codename corresponding to the specified release.
It throws NATE::BaseException if it couldn't obtain the codename. 

=back

=head2 map_release_to_numeric_version

For a same release, we can have different release strings like internal and
external release strings and hence map all release strings referring a
particular release to its numeric release version.

=over

=item Synopsis

$numeric_version = $version_object->map_release_to_numeric_version(release => $release);

=item Arguments

=over

=item release

The release string.

=back

=item Returns

This method will return the numeric release like '8.1.0', '8.2.1' etc
corresponding to the specified release.
It throws NATE::BaseException if it couldn't obtain the numeric version.

=back

=head2 is_promoted

Checks whether the service is running a promoted(release) or
development build.

=over 

=item Synopsis

$is_promoted = $version_object->is_promoted();

=item Arguments

Nothing

=item Returns

This method will return 1 for promoted builds and 0 for
development builds.

=back

=head2 is_classic

Checks whether the service is running a 7G/Classic(release) build.

=over 

=item Synopsis

$is_classic = $version_object->is_classic();

=item Arguments

Nothing

=item Returns

This method will return 1 for 7G/classic builds and 0 for other 
builds.

=back

=head2 is_debug

Checks whether the service is running a debug or
non-debug build.

=over 

=item Synopsis

$is_debug = $version_object->is_debug();

=item Arguments

Nothing

=item Returns

This method will return 1 for debug builds and 0 for
non-debug builds.

=back

=head2 is_gcov

Checks whether the service is running a gcov or
non-gcov build.

=over

=item Synopsis

$is_gcov = $version_object->is_gcov();

=item Arguments

Nothing

=item Returns

This method will return 1 for gcov builds and 0 for
non-gcov builds.

=back

=head2 is_version_eq

Checks whether the version running on the build is equal to
the given version string

=over 

=item Synopsis

$result = $version_object->is_version_eq(release => "8.1.0");

=item Arguments

=over

=item release

The ONTAP release string. This should be of the format major.minor.sub
or major.minor. For example "8.1.0" or "8.1".
For any other format of release string, a NATE::Exceptions::Argument will be thrown.

=back

=item Returns

This method will return 1 if the release on the build is equal to the 
given release and 0 otherwise.

=back

=head2 is_version_ne

Checks whether the version running on the build is not equal to
the given version string

=over 

=item Synopsis

$result = $version_object->is_version_ne(release => "8.1.0");

=item Arguments

=over

=item release

The ONTAP release string. This should be of the format major.minor.sub
or major.minor. For example "8.1.0" or "8.1".
For any other format of release string, a NATE::Exceptions::Argument will be thrown.

=back

=item Returns

This method will return 1 if the release on the build is not equal to 
the given release and 0 otherwise.

=back

=head2 is_version_lt

Checks whether the version running on the build is less than
the given version string

=over 

=item Synopsis

$result = $version_object->is_version_lt(release => "8.1.0");

=item Arguments

=over

=item release

The ONTAP release string. This should be of the format major.minor.sub
or major.minor. For example "8.1.0" or "8.1".
For any other format of release string, a NATE::Exceptions::Argument will be thrown.

=back

=item Returns

This method will return 1 if the release on the build is less than the 
given release and 0 otherwise.

=back

=head2 is_version_le

Checks whether the version running on the build is less than or equal to
the given version string

=over 

=item Synopsis

$result = $version_object->is_version_le(release => "8.1.0");

=item Arguments

=over

=item release

The ONTAP release string. This should be of the format major.minor.sub
or major.minor. For example "8.1.0" or "8.1".
For any other format of release string, a NATE::Exceptions::Argument will be thrown.

=back

=item Returns

This method will return 1 if the release on the build is less than or 
equal to the given release and 0 otherwise.

=back

=head2 is_version_gt

Checks whether the version running on the build is greater than 
the given version string

=over 

=item Synopsis

$result = $version_object->is_version_gt(release => "8.1.0");

=item Arguments

=over

=item release

The ONTAP release string. This should be of the format major.minor.sub
or major.minor. For example "8.1.0" or "8.1".
For any other format of release string, a NATE::Exceptions::Argument will be thrown.

=back

=item Returns

This method will return 1 if the release on the build is greater than 
the given release and 0 otherwise.

=back

=head2 is_version_ge

Checks whether the version running on the build is greater than or equal to
the given version string

=over 

=item Synopsis

$result = $version_object->is_version_ge(release => "8.1.0");

=item Arguments

=over

=item release

The ONTAP release string. This should be of the format major.minor.sub
or major.minor. For example "8.1.0" or "8.1".
For any other format of release string, a NATE::Exceptions::Argument will be thrown.

=back

=item Returns

This method will return 1 if the release on the build is greater than or
equal to the given release and 0 otherwise.

=back

=head2 is_domain_crossing

Checks whether the service is running a domain crossing or
non domain crossing build.

=over

=item Synopsis

$is_domain_crossing = $version_object->is_domain_crossing();

=item Arguments

Nothing

=item Returns

This method will return 1 for domain crossing builds and 0 for
non domain crossing builds.

=back

=head1 AUTHOR/MAINTAINER

=over

=item NACL Development (dl-nacl-dev@netapp.com)

=back

=cut
