#
# Copyright (c) 2001-2010 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#
## @summary Version class for finding Version attributes specific to 7Mode.
## @author Isshwarya.M@netapp.com, dl-nacl-dev@netapp.com
## @status Shared
## @pod here

package NACL::APISet::Version::Node::7Mode;

#
# Compiler Directives
#
use strict;
use warnings;

#
# Module imports
#
use NACL::APISet::Session::CLI::7Mode;
use NACL::APISet::Session::CLI::7Mode::Shell;
use NACL::APISet::Session::ZAPI::7Mode;
use NACL::APISet::Response::CLI::ParserUtils;

#
# Exceptions imported
#
use NACL::APISet::Exceptions::ConnectionFailedException qw(:try);
use NACL::APISet::Exceptions::AttributeDiscoveryFailed qw(:try);

#
# NATE imports
#
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Params::Validate qw(:all);
use Data::Dumper;

# This block is a hack to make "perl -wc" work despite the circular dependency
# on this module:
BEGIN {
    use File::Spec;
    $INC{'NACL/APISet/Version/Node/7Mode.pm'} ||= File::Spec->rel2abs(__FILE__);
}


#
# Base class
#
use base qw(NACL::APISet::Version::Node);

# Wafl magic number mapping to releases.
# Mostly this hash doesn't require any updation as we are going to use the
# release information given in this hash only for IC and other older
# releases. For all new releases (> IC), we can find the release information
# directly from the service by running the appropriate commands.
our %wafl_version = (
    'wafl_release' => [
        { 'min' => '0',  'release' => 'Oldest',        'max' => '9' },
        { 'min' => '10', 'release' => 'Guinness',      'max' => '12' },
        { 'min' => '13', 'release' => 'Vb',            'max' => '13' },
        { 'min' => '14', 'release' => 'Redhook',       'max' => '27' },
        { 'min' => '28', 'release' => 'Krakus',        'max' => '28' },
        { 'min' => '29', 'release' => 'Fullsail',      'max' => '35' },
        { 'min' => '36', 'release' => 'Scrimshaw',     'max' => '38' },
        { 'min' => '39', 'release' => 'Anchorsteam',   'max' => '55' },
        { 'min' => '56', 'release' => 'Tsingtao',      'max' => '57' },
        { 'min' => '64', 'release' => 'Gordonbiersch', 'max' => '73' },
        { 'min' => '74', 'release' => 'Tricky',        'max' => '75' },
        { 'min' => '76', 'release' => 'Kestrel',       'max' => '77' },

        # New version format was started from IC onwards.
        # IC min = (77 << 8), IC max = (78 << 8) | 255
        { 'min' => '19712', 'release' => 'Ironcity', 'max' => '20223' },
    ],

    # Upper limit for 7G will be IC maximum wafl version which is
    # (78 << 8) | 255. If the found wafl version is less than or equal to
    # 20223, then it is a 7G system.
    'upper_limit_for_7G' => '20223',
);

use Class::MethodMaker [
    scalar => [
        {
            '-default_ctor' => sub {
                $Log->enter() if $may_enter;
                my $self    = $_[0];
                my $is_sksim = $self->is_sksim();
                my $is_promoted;
                
                if ($is_sksim) {
                    # Queries through normal shell/dblade might fail for sksim,
                    # but sksim is like a non-promoted build
                    $is_promoted = 0;
                } else {
                    # Try to obtain the information from dblade shell.
                    my $session;
                    try {
                        $session = $self->get_session(
                            interface => "CLI",
                            subtype   => "7Mode"
                        );
                        my $old_priv;
                        my $is_console = ( $session->{connection}->conntype()
                                =~ /serial|pm_cons|rconsole|console/ ) ? 1 : 0;
                        if ($is_console) {

                            # Get the current privilege level before modifying it
                            # only for console connections, the privilege level
                            # set is saved even after a console logout.
                            my $outp = $self->execute_CLI_command(
                                connection => $session->{connection},
                                command    => "priv"
                            );
                            $old_priv = $1 if ( $outp =~ /(admin|advanced|diag|test)/ );
                        }
                        my $outp = $self->execute_CLI_command(
                            connection => $session->{connection},
                            command    => "priv set -q test;result -l"
                        );
                        if ( $outp !~ /result=\d+/ ) {

                            # connection is failed.
                            $Log->exit() if $may_exit;
                            NACL::APISet::Exceptions::ConnectionFailedException->throw(
                                "Failed to establish the connection: $outp");
                        }

                        # "priv set -q" usually returns no output when the command succeeds.
                        # We can take this decision that the absence of expected error message
                        # is a success in command execution only when the connection is
                        # established successfully. Checking for the connection failure is
                        # required in the case of persay connections especially. Hence we
                        # performed the above step which checks whether the connection was
                        # established successful or not.
                        if ( $outp =~ /unrecognized privilege: test/i ) {
                            $is_promoted = 1;
                        } else {
                            $is_promoted = 0;
                        }
                        if ($is_console) {

                           # This is a console connection which would be shared with the users
                           # So we should set the privilege back to its previous value.
                            $self->execute_CLI_command(
                                connection => $session->{connection},
                                command    => "priv set -q $old_priv"
                            ) if ( defined $old_priv );
                        }
                    } catch NATE::BaseException with {
                        my $exception = shift;
                        $Log->debug('Failed to discover whether the build is promoted '
                            . 'or not: '
                            . $exception->text()
                        );

                        $is_promoted = $self->_is_promoted_systemshell();
                    } finally {
                        # Destroy the session
                        $self->destroy_session( session => $session ) if ( defined $session );
                    };
                }

                # promoted builds are guaranteed to be non-debug
                $self->is_debug(0) if ($is_promoted);
                $Log->exit($is_promoted);
                return $is_promoted;
            },
        },
        'is_promoted',
    ],
];

###############################################################################
# Method:    new
# Objective: Constructor to create new Version 7Mode object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub new {
    $Log->enter() if $may_enter;

    my ( $proto, %args ) = @_;
    my $class = ref($proto) || $proto;

    # Validate parameters and populate member variables
    my $self = NACL::APISet::validate_params(
        $class,
        params => \%args,
        spec   => {
            hostobj        => { isa  => "Hostrec", optional => 0 },
            category       => { type => SCALAR,    optional => 0 },
            shell_type     => { type => SCALAR,    optional => 0 },
            conn_overrides => { type => HASHREF,   optional => 1 },
        }
    );
    bless $self, $class;

    # If the shell_type is vfiler, then the provided hostobj or connection
    # object will establish a connection directly to vfiler. From a vfiler
    # shell, we cannot find version related attributes and we should connect
    # to parent filer and this information can be obtained from PFILER
    # attribute present in the hostrec object.
    if ( $self->{shell_type} eq "Vfiler" ) {
        my $parent          = $self->{hostobj}->{"PFILER"};
        my $parent_hosttype = $self->{hostobj}->hosttype();
        $self->{hostobj}
            = Tharn::host( $parent . ";hosttype=$parent_hosttype" )
            if ( defined $parent );
    }

    # Discover whether it is a 7G/7Mode system as certain version related
    # attributes cannot be found in 7G system.
    try {
        $self->{is_classic} = $self->_is_pure_classic();
    }
    catch NATE::BaseException with {

       # Error occurred in determining whether the system is 7G
       # or 7Mode. By default, we will consider the system to be 7Mode because
       # preference is given to 7Mode. Another reason for this error might be
       # not providing PFILER attribute in vfiler hostobj when directly
       # connected to vfiler shell
        $self->{is_classic} = 0;
    };
    if ( $self->{is_classic} ) {
        $self->{valid_attributes}
            = [ "wafl_version", "release", "node_version", "modelinfo", "serialnumber" ];
    }

    $Log->exit() if $may_exit;
    return $self;
} ## end sub new()

##############################################################################
# Method:    _find_release
# Objective: Perform the discovery of release information
# Details:   See POD documentation at the end of this file.
##############################################################################

sub _find_release {
    my ( $self, %args ) = @_;

    $Log->enter() if $may_enter;

    # If it is a 7G system, then use wafl_version hash to find the release.
    if ( $self->{is_classic} ) {
        my @entries = @{ $wafl_version{'wafl_release'} };
        my $wafl_version
            = $self->get_version_attribute( "attribute" => "wafl_version" );
        foreach my $hash (@entries) {
            if (   ( $wafl_version >= $hash->{min} )
                && ( $wafl_version <= $hash->{max} ) )
            {
                $self->{release} = $hash->{release};
                last;
            }
        }
        $Log->debug("Release for this 7G system is : $self->{release}");
        $Log->exit() if $may_exit;
        return;
    }
    
    # Normally gets this through system-shell, but this is not accessible
    # for sksim, so just hard-code to LB
    if ($self->is_sksim()) {
        $self->{release} = 'Longboard__8.4.0';
        $Log->exit() if $may_exit;
        return;
    }

    # Get the connection. Using session class itself so that it will
    # do prep_service and it helps us to get a connetion object with
    # the connection overrides specified, if any.
    my $session;
    try {
        $session = $self->get_session(
            interface => "CLI",
            subtype   => "7Mode::Shell",
            connid    => "7-system"
        );

        # Find the information
        $self->SUPER::_find_release( connection => $session->{connection} );
    }
    finally {

        # Destroy the session
        $self->destroy_session( session => $session ) if ( defined $session );
    };
    $Log->exit() if $may_exit;
}

##############################################################################
# Method:    _find_wafl_version
# Objective: Perform the discovery of wafl version information
# Details:   See POD documentation at the end of this file.
##############################################################################

sub _find_wafl_version {
    my ( $self, %args ) = @_;

    $Log->enter() if $may_enter;

    my $session;
    try {
        $session = $self->get_session(
            interface => "ZAPI",
            subtype   => "7Mode"
        );

        # Find the information
        $self->_find_wafl_version_through_zapi( session => $session );
    }
    catch NATE::BaseException with {

        # Catching all exceptions so as to try the possibility of finding
        # wafl version from CLI.
        my $exception = shift;
        $Log->debug( "Failed to get wafl version through ZAPI: "
                . $exception->text() );
    }
    finally {

        # Destroy the session
        $self->destroy_session( session => $session ) if ( defined $session );
    };
    if ( defined $self->{wafl_version} ) {
        $Log->exit() if $may_exit;
        return;
    }

    # Try running the "stats" command from 7Mode shell.
    # We are not catching any exceptions as we had tried all possibilities to
    # obtain the information and we allow it to fail at this point.
    my $output;
    try {
        $session = $self->get_session(
            interface => "CLI",
            subtype   => "7Mode"
        );

        # Find the information
        $self->SUPER::_find_wafl_version( session => $session );
    }
    finally {

        # Destroy the session
        $self->destroy_session( session => $session ) if ( defined $session );
    };
    $Log->exit() if $may_exit;
}

##############################################################################
# Method:    _find_branch
# Objective: Perform the discovery of branch information
# Details:   See POD documentation at the end of this file.
##############################################################################

sub _find_branch {
    my ( $self, %args ) = @_;

    $Log->enter() if $may_enter;
    my $session;
    try {
        $session = $self->get_session(
            interface => "CLI",
            subtype   => "7Mode::Shell",
            connid    => "7-system"
        );

        # Find the information
        $self->SUPER::_find_branch( connection => $session->{connection} );
    }
    finally {

        # Destroy the session
        $self->destroy_session( session => $session ) if ( defined $session );
    };
    $Log->exit() if $may_exit;
}

##############################################################################
# Method:    _find_modelinfo
# Objective: Perform the discovery of model information
# Details:   See POD documentation at the end of this file.
##############################################################################

sub _find_modelinfo {
    my ( $self, %args ) = @_;

    $Log->enter() if $may_enter;

    my $output;
    my $session;

    # Try running the 7Mode command "sysconfig -v 0".
    # We are not catching any exceptions as we had tried all possibilities to
    # obtain the information and we allow it to fail at this point.

    try {
        $session = $self->get_session(
            interface => "CLI",
            subtype   => "7Mode"
        );

        # Find the information
        $output = $self->execute_CLI_command(
            command    => "sysconfig -v 0",
            connection => $session->{connection}
        );
        $output =~ /\s*Model Name:\s+(.+)/;
        $output = $1;
        if ( defined $output ) {
            $self->{modelinfo} = trim_string( string => $output );
        }
    }
    finally {

        # Destroy the session
        $self->destroy_session( session => $session ) if ( defined $session );
    };

    $Log->exit() if $may_exit;
}

##############################################################################
# Method:    _find_serialnumber
# Objective: Perform the discovery of serialnumber information
# Details:   See POD documentation at the end of this file.
##############################################################################

sub _find_serialnumber {
    my ( $self, %args ) = @_;

    $Log->enter() if $may_enter;
    my $session;

    # Try running the 7Mode command "sysconfig".
    # We are not catching any exceptions as we had tried all possibilities to
    # obtain the information and we allow it to fail at this point.

    try {
        $session = $self->get_session(
            interface => "CLI",
            subtype   => "7Mode"
        );

        # Find the information
        my $output = $self->execute_CLI_command(
            connection => $session->{connection},
            command    => "sysconfig"
        );
        $output =~ /System Serial Number:\s+([\d|-]+)\s+.*/;
        $self->{serialnumber} = $1;
    }
    finally {

        # Destroy the session
        $self->destroy_session( session => $session ) if ( defined $session );
    };
    $Log->exit() if $may_exit;
}

##############################################################################
# Method:    _find_change_number
# Objective: Perform the discovery of change number information
# Details:   See POD documentation at the end of this file.
##############################################################################

sub _find_change_number {
    my ( $self, %args ) = @_;

    $Log->enter() if $may_enter;

    my $session;
    try {
        $session = $self->get_session(
            interface => "CLI",
            subtype   => "7Mode::Shell",
            connid    => "7-system"
        );

        # Find the information
        $self->SUPER::_find_change_number(
            connection => $session->{connection} );
    }
    finally {

        # Destroy the session
        $self->destroy_session( session => $session ) if ( defined $session );
    };
    $Log->exit() if $may_exit;
}

##############################################################################
# Method:    _find_client
# Objective: Perform the discovery of client information
# Details:   See POD documentation at the end of this file.
##############################################################################

sub _find_client {
    my ($self, %args) = @_;
    
    $Log->enter() if $may_enter;
    
    my $session;
    try {
        $session = $self->get_session(interface => "CLI", 
            subtype => "7Mode::Shell", connid => "7-system");
        # Find the information
        $self->SUPER::_find_client(connection=>$session->{connection});
    } finally {
        # Destroy the session
        $self->destroy_session(session => $session) if (defined $session);
    };
    $Log->exit() if $may_exit;
}

##############################################################################
# Method:    _find_build_options
# Objective: Perform the discovery of build options information
# Details:   See POD documentation at the end of this file.
##############################################################################

sub _find_build_options {
    my ( $self, %args ) = @_;

    $Log->enter() if $may_enter;

    my $session;
    try {
        $session = $self->get_session(
            interface => "CLI",
            subtype   => "7Mode::Shell",
            connid    => "7-system"
        );

        # Find the information
        $self->SUPER::_find_build_options(
            connection => $session->{connection} );
    }
    finally {

        # Destroy the session
        $self->destroy_session( session => $session ) if ( defined $session );
    };
    $Log->exit() if $may_exit;
}

##############################################################################
# Method:    _find_build_path
# Objective: Perform the discovery of build path information
# Details:   See POD documentation at the end of this file.
##############################################################################

sub _find_build_path {
    my ( $self, %args ) = @_;

    $Log->enter() if $may_enter;

    # Get the connection. Using session class itself so that it will
    # do prep_service
    my $session;
    try {
        $session = $self->get_session(
            interface => "CLI",
            subtype   => "7Mode::Shell",
            connid    => "7-system"
        );

        # Find the information
        $self->SUPER::_find_build_path(
            connection => $session->{connection} );
    }
    catch NATE::BaseException with {

        # Catch any exception so as to try the possibility of running
        # commands from 7Mode shell.
        my $exception = shift;
        $Log->debug( "Error in obtaining build path from systemshell: "
                . $exception->text() );
    }
    finally {

        # Destroy the session
        $self->destroy_session( session => $session ) if ( defined $session );
    };
    if ( defined $self->{build_path} ) {
        $Log->exit() if $may_exit;
        return;
    }

    # Try running the version command from 7Mode shell.
    # We are not catching any exceptions as we had tried all possibilities to
    # obtain the information and we allow it to fail at this point.
    try {
        my $session = $self->get_session(
            interface => "CLI",
            subtype   => "7Mode"
        );

        # Find the information
        my $output = $self->execute_CLI_command(
            command    => "version -b",
            connection => $session->{connection}
        );
        $output =~ /.*kernel: OS (.+)/;
        $output = $1;
        if ($output =~ /(\w+\/)+/) {
            $self->{build_path} = $output;
        } else {
            # For SN and beyond the version -b output displays only the 
            # release name and not the build path. Throw an exception if such
            # is the case.
            $Log->exit();       
            NACL::APISet::Exceptions::AttributeDiscoveryFailed->throw(
                "Could not get build path from the service. " .
                "Output is $output" );       
        }
    }
    finally {

        # Destroy the session
        $self->destroy_session( session => $session ) if ( defined $session );
    };
    $Log->exit() if $may_exit;
}

##############################################################################
# Method:    _find_node_version
# Objective: Perform the discovery of node version information
# Details:   See POD documentation at the end of this file.
##############################################################################

sub _find_node_version {
    my ( $self, %args ) = @_;

    $Log->enter() if $may_enter;

    my $session;
    unless ( $self->{is_classic} ) {

        # Systemshell commands
        try {
            $session = $self->get_session(
                interface => "CLI",
                subtype   => "7Mode::Shell",
                connid    => "7-system"
            );
            $self->SUPER::_find_node_version(
                connection => $session->{connection} );
        }
        catch NATE::BaseException with {

            # Catching any exception so as to try the possibility of running
            # commands from 7Mode shell or ZAPI.
            my $exception = shift;
            $Log->debug( "Error in obtaining node version from systemshell: "
                    . $exception->text() );
        }
        finally {
            $self->destroy_session( session => $session )
                if ( defined $session );
        };
        if ( defined $self->{node_version} ) {
            $Log->exit() if $may_exit;
            return;
        }
    }

    # ZAPI
    my $output;
    try {
        $session = $self->get_session(
            interface => "ZAPI",
            subtype   => "7Mode"
        );
        my $command = "<system-get-version/>";
        my $response
            = $self->execute_ZAPI( command => $command, session => $session );
        if ( defined $response ) {
            my $parsed_output = $response->get_parsed_output();
            $self->{node_version} = $parsed_output->[0]->{version};
        }
    }
    catch NATE::BaseException with {

        # Catching this exception so as to try the possibility of running
        # commands from 7Mode shell.
        my $exception = shift;
        $Log->debug( "Error in obtaining node version from systemshell: "
                . $exception->text() );
    }
    finally {
        $self->destroy_session( session => $session ) if ( defined $session );
    };
    if ( defined $self->{node_version} ) {
        $self->{node_version} = $self->_extract_node_version(
            version => trim_string( string => $output ) );
        $Log->exit() if $may_exit;
        return;
    }

    # 7Mode CLI
    try {
        $session = $self->get_session(
            interface => "CLI",
            subtype   => "7Mode"
        );

        # Find the information
        my $output = $self->execute_CLI_command(
            command    => "version",
            connection => $session->{connection}
        );
        $self->{node_version} = $self->_extract_node_version(
            version => trim_string( string => $output ) );
    }
    finally {

        # Destroy the session
        $self->destroy_session( session => $session ) if ( defined $session );
    };

    $Log->exit() if $may_exit;
}

##############################################################################
# Method:    _find_wafl_version_through_zapi
# Objective: Perform the discovery of wafl version information using ZAPI
#            interface
# Details:   This method will discover wafl version from the root volume using
#            ZAPI interface and set the wafl_version attribute to the obtained
#            wafl magic number.
##############################################################################

sub _find_wafl_version_through_zapi {
    my ( $self, %args ) = @_;
    $Log->enter() if $may_enter;
    my $session = $args{session};

    # Get root volume name
    my $command = "<volume-get-root-name/>";
    my $parsed_output;
    my $response
        = $self->execute_ZAPI( command => $command, session => $session );
    if ( defined $response ) {
        $parsed_output = $response->get_parsed_output();
        my $volname = $parsed_output->[0]->{volume};
        unless ($volname) {
            $Log->exit() if $may_exit;
            NACL::APISet::Exceptions::AttributeDiscoveryFailed->throw(
                "Could not get root volume name and the output obtained from the"
                    . " service is $response" );
        }
        $Log->debug("Root volume is found as: $volname");

        # Get wafl magic number
        $command = "<perf-object-get-instances>
            <objectname>volume</objectname>
            <instances> <instance>$volname</instance> </instances>
            <counters> <counter>wv_fsinfo_fs_version</counter> </counters>
        </perf-object-get-instances>";
    }

    $response
        = $self->execute_ZAPI( command => $command, session => $session );
    if ( defined $response ) {
        $parsed_output = $response->get_parsed_output();
        my $version
            = $parsed_output->[0]->{instances}->[0]->{"instance-data"}->[0]
            ->{counters}->[0]->{"counter-data"}->[0]->{value};
        unless ($version) {
            $Log->exit() if $may_exit;
            NACL::APISet::Exceptions::AttributeDiscoveryFailed->throw(
                "Couldn't discover wafl magic number and the result is $response"
            );
        }
        $Log->debug("Wafl version found through ZAPI: $version");
        $self->{wafl_version} = $version;
    }

    $Log->exit() if $may_exit;
}

##############################################################################
# Method:    _is_pure_classic
# Objective: Determine whether the service is running 7G/ 7Mode.
# Details:   This method will identify whether the given system is running
#            7G (IC and earlier versions) or 7Mode (BR and later releases).
#            When finding version related information, we execute certain
#            commands from systemshell which is not possible in 7G systems and
#            hence we need to differentiate between these two.
##############################################################################

sub _is_pure_classic {
    my ( $self, %args ) = @_;

    $Log->enter() if $may_enter;

    # Get the wafl version of the service.
    my $wafl_version;
    my $is_sksim = $self->is_sksim();
    if ($is_sksim) {
        $Log->exit() if $may_exit;
        return 0;
    }
    try {
        $wafl_version
            = $self->get_version_attribute( "attribute" => "wafl_version" );
    }
    catch NACL::APISet::Exceptions::AttributeDiscoveryFailed with {
        my $ex = shift;
        my $warning
            = "Couldn't discover whether it is a 7G/7Mode system"
            . " Possible reasons include: 1)Service is not pingable"
            . " 2)Specifying wrong shell_type";
        if (   ( $ex->text() !~ /result=\d+/ )
            && ( !defined $self->{hostobj}->{"PFILER"} )
            && ( $self->{shell_type} eq "Vfiler" ) )
        {
            $warning .= " 3) The host might be a vfiler for which PFILER"
                . " attribute need to be supplied in hostrec object";
        }
        $Log->warn($warning);
        $ex->throw();
    };

    # Check whether the discovered wafl version is less than or equal to the
    # max IC wafl version
    my $upper_limit = $wafl_version{'upper_limit_for_7G'};
    if ( $wafl_version <= $upper_limit ) {
        $Log->debug("This is found to be a 7G system");
        return 1;
        $Log->exit() if $may_exit;
    }
    $Log->debug("The service is running 7Mode");
    return 0;
    $Log->exit() if $may_exit;
}

1;

=head1 NAME

NACL::APISet::Version::Node::7Mode - Version library for 7Mode

=head1 SYNOPSIS

 use NACL::APISet::Version::Node::7Mode;

=head1 DESCRIPTION

A C<NACL::APISet::Version::Node::7Mode> is used to connect to 7Mode node and 
collect version related attributes. All version related attributes exposed from
this libary can also be set or get by the user. 

=cut

=head1 EXCEPTIONS

=over

=item NACL::APISet::Exceptions::ConnectionFailedException

=item NACL::APISet::Exceptions::AttributeDiscoveryFailed

=back

=head1 CONSTRUCTORS

=head2 new

Returns a reference to a newly created C<NACL::APISet::Version::Node::7Mode>.

=over 

=item Synopsis

$version = NACL::APISet::Version::Node::7Mode->new(%args)

=item Arguments

=over

=item category

(Mandatory)
The category to use.  Valid values include "Node".

=item shell_type

(Mandatory)
The refers to the type of the shell host sent through 'hostobj'.
Valid values include "7Mode", "Vfiler". 
If hostobj refers to vfiler host, then shell_type should be "Vfiler".

=item hostobj

(Mandatory)
The hostrec object created for the resource.

=item conn_overrides

(Optional)
Hash reference to the list of connection overrides to be considered when
establishing the connection to the service.
Valid connection overrides are username, password, connid, conntype, ipaddress,
prompt_username and prompt_password. Note that these overrides would not be
considered for shell_type being Vfiler as version manger connects to parent
filer (PFILER) in such case and these overrides specified for Vfiler wouldn't
be considered.

=item console_connection

(Optional)
This paramter is applicable only to 'Node' category, especially if the
shell_type is either 7Mode or CMode. If you already have a console
connection and if you want the version manger to reuse the same connection for
whatever attributes that could be discovered from ngshell/7Mode shell, then you
can specify the existing console connection object to this option.

=back

=item Returns

A reference to a new 7Mode Version object

=back

=cut

=head1 METHODS

=head2 _find_branch

This method discovers branch information by creating the appropriate session
and execute various commands like "ident", "cat /BUILD" to get information
about branch.

=over 

=item Synopsis

$branch = $version_object->get_version_attribute(attribute => "branch");

The above method will internally call $self->_find_branch() if the information
about branch atribute is not already present in $self. This method shouldn't be
directly invoked by the user.

=item Arguments

Nothing

=item Returns

Nothing. Set the branch information in version object itself.

=back

=head2 _find_release

This method discovers release information by creating the appropriate session
and execute various commands like "sysctl", "cat /BUILD", "system image show" 
to get information about release.

=over 

=item Synopsis

$release = $version_object->get_version_attribute(attribute => "release");

The above method will internally call $self->_find_release() if the information
about release atribute is not already present in $self. This method shouldn't be
directly invoked by the user.

=item Arguments

Nothing

=item Returns

Nothing. Set the release information in version object itself.

=back

=head2 _find_change_number

This method discovers change number associated with the build by creating the 
appropriate session and execute various commands like "ident", "cat /BUILD",
"system image show" to get information about change number.

=over 

=item Synopsis

$change = $version_object->get_version_attribute(attribute => "change_number");

The above method will internally call $self->_find_change_number() if the 
information about change_number atribute is not already present in $self. This
method shouldn't be directly invoked by the user.

=item Arguments

Nothing

=item Returns

Nothing. Set the change number information in version object itself.

=back

=head2 _find_build_options

This method discovers build options by creating the 
appropriate session and execute various commands like "ident", "cat /BUILD",
to get information about build options.

=over 

=item Synopsis

$build_options = 
    $version_object->get_version_attribute(attribute => "build_options");

The above method will internally call $self->_find_build_options() if the 
information about build_options atribute is not already present in $self. This
method shouldn't be directly invoked by the user.

=item Arguments

Nothing

=item Returns

Nothing. Set the build options information in version object itself.

=back

=head2 _find_build_path

This method discovers build path by creating the 
appropriate session and execute various commands like "ident", "sysctl",
or "version" command in Nodescope to get information about build path.

=over 

=item Synopsis

$build_path = 
    $version_object->get_version_attribute(attribute => "build_path");

The above method will internally call $self->_find_build_path() if the 
information about build_path atribute is not already present in $self. This
method shouldn't be directly invoked by the user.

=item Arguments

Nothing

=item Returns

Nothing. Set the build path information in version object itself.

=back

=head2 _find_wafl_version

This method discovers wafl version information by first executing the relevant
ZAPIs and if it fails, then it falls back to running CLI commands in 7Mode 
shell.

=over 

=item Synopsis

$wafl_version = 
    $version_object->get_version_attribute(attribute => "wafl_version");

The above method will internally call $self->_find_wafl_version() if the 
information about wafl_version atribute is not already present in $self. This
method shouldn't be directly invoked by the user.

=item Arguments

Nothing

=item Returns

Nothing. Set the wafl version information in version object itself. 

=back

=head2 _find_modelinfo

This method finds the model information by creating the appropriate session
and execute commands like "sysconfig -v 0" to get information about system 
model.

=over 

=item Synopsis

$release = $version_object->get_version_attribute(attribute => "modelinfo");

The above method will internally call $self->_find_modelinfo() if the information
about modelinfo atribute is not already present in $self. This method shouldn't be
directly invoked by the user.

=item Arguments

Nothing

=item Returns

Nothing. Set the filer model information in version object itself. 

=back

=head2 _find_serialnumber

This method finds the serialnumber information by creating the appropriate session
and execute commands like "sysconfig" to get information about system 
serial number.

=over 

=item Synopsis

$serialnumber = $version_object->get_version_attribute(attribute => "serialnumber");

The above method will internally call $self->_find_serialnumber() if the information
about serialnumber attribute is not already present in $self. This method shouldn't be
directly invoked by the user.

=item Arguments

Nothing

=item Returns

Nothing. Set the system serial number in version object itself. 

=back

=head2 _find_client

This method finds the client information by creating the appropriate session
and execute commands like " cat /BUILD" to get information about CLIENT

=over 

=item Synopsis

$release = $version_object->get_version_attribute(attribute => "client");

The above method will internally call $self->_find_client() if the information
about client atribute is not already present in $self. This method shouldn't be
directly invoked by the user.

=item Arguments

Nothing

=item Returns

Nothing. Set the client information in version object itself. 

=back

=head2 _find_node_version

This method discovers node version information by creating the appropriate session
and execute various commands like "sysctl", "version" in Nodescope context
to get information about node version.

=over 

=item Synopsis

$node_version = 
    $version_object->get_version_attribute(attribute => "node_version");

The above method will internally call $self->_find_node_version() if the 
information about node_version atribute is not already present in $self. This
method shouldn't be directly invoked by the user.

=item Arguments

Nothing

=item Returns

Nothing. Set the node version information in version object itself. 

=back

=head2 is_promoted

Checks whether the service is running a promoted(release) or
development build.

=over 

=item Synopsis

$is_promoted = $version_object->is_promoted();

=item Arguments

Nothing

=item Returns

This method will return 1 for promoted builds and 0 for
development builds.

=back

=head1 AUTHOR/MAINTAINER

=over

=item NACL Development (dl-nacl-dev@netapp.com)

=back

=cut
