#
# Copyright (c) 2001-2010 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#
## @summary This package aids us to support forward compatibility for the UI
## changes happened in the service. At the moment, compatibility is supported
## only for CMode CLI.
## @author Isshwarya.M@netapp.com, dl-nacl-dev@netapp.com
## @status Shared
## @see-also lib/Node/NG/Compatibility.pm
## @pod here

package NACL::APISet::Version::Node::Compatibility;

#
# Compiler Directives
# 
use strict;
use warnings;

#
# NATE imports
#
use Data::Dumper;
use Tie::IxHash;
use NATE::Log qw(log_global);
my $Log = log_global();
my $log_may_debug = $Log->may_debug();

#
# Exceptions
#
use NATE::BaseException qw(:try);

# NOTE: The method names for building compatibility hash should be named as
# build_compatibility_<interface>

##############################################################################
# Method:    build_compatibility_cli
# Objective: To build the compatibility hash for CLI
# Details:   Refer to POD documentation at the end of the file.
##############################################################################

sub build_compatibility_cli {
    my ($self, %args) = @_;
    $Log->enter();
    
    # This array of uichange strings will be in chronological order
    my @uichanges = @{$args{uichanges}};
    
    # Read the module which defines the uichange strings
    $Log->debug("Read the uichange string mapping");
    my (%compatibility, %all_changes);
    my $file = "NACL/APISet/Version/VersionData/CompatibilityDataCli.pm";
    my $str = NATE::Inc::get_file($file) 
        or NATE::BaseException->throw("Error in reading file '$file': $@");
    eval $str;
    NATE::BaseException->throw("Error in loading $file: $@") if ($@);
    
    foreach my $uichange (@uichanges) {
        # If we don't have an entry for the corresponding uichange, then
        # continue with next iteration
        next unless (defined $all_changes{$uichange});
        $Log->debug("Processing uichange string $uichange");
        # The uichange string might be associated with changes in more than one
        # command. Hence process each and every command pattern.
        foreach my $cmd_pattern (keys %{$all_changes{$uichange}}) {
            my $value = $all_changes{$uichange}{$cmd_pattern};
            
            # NOTE: The conversion hash entries are order preserved hashes
            # to handle the case where A=>B, B=>C, C=>A. Here we preserve the
            # order in which the hashes are merged as a group and not to
            # the granularity of conversion hash entries in each individual 
            # group which is sufficent enough.      
            my (%existing_hash, %new_hash);
            tie %existing_hash, 'Tie::IxHash';
            tie %new_hash, 'Tie::IxHash';
            
            # Merge if the key (command pattern) already exists.
            if (exists $compatibility{$cmd_pattern}) {
                # Merge the entries.
                $Log->debug("Command $cmd_pattern already exists in ".
                    "compatibility hash. Hence merging");
                %existing_hash = 
                    (exists $compatibility{$cmd_pattern}{conversion})?
                    %{$compatibility{$cmd_pattern}{conversion}}:();
                %new_hash = (exists $value->{conversion}) ?
                    %{$value->{conversion}}:(); 
                my %hash;
                tie %hash, 'Tie::IxHash';
                %hash = (%existing_hash, %new_hash);        
                $value->{conversion} = \%hash ;
                $Log->debug("The conversion hash is now:\n". 
                    Dumper($value->{conversion})) if ($log_may_debug);
            }            
            
            # Add each command pattern which we come across to compatibility hash.
            $compatibility{$cmd_pattern} = $value;
            $Log->debug("For compatibility{$cmd_pattern}: " . Dumper($value))
                if ($log_may_debug);;
        }
    }
    $Log->debug("The \%compatibility hash is: " . Dumper(\%compatibility))
        if ($log_may_debug);
    $Log->exit();
    return \%compatibility; 
}

1;

##############################################################################
#                                    POD here
##############################################################################

=head1 NAME

NACL::APISet::Version::Node::Compatibility.pm - Package to support forward
compatibility in the scripts/libraries for the option and command name changes
(syntax changes or modifcations) happened in the service.

=head1 SYNOPSIS

 use NACL::APISet::Version::Node::Compatibility;

=head1 DESCRIPTION

This package provides methods which will enable the callers to build the
compatibility hash dynamically only for the uichange strings present in the
service.

=head1 EXCEPTIONS

=over

=item NATE::BaseException

=back

=head1 METHODS

=head2 build_compatibility_cli

This method takes the list of uichange strings and builds the compatibility
hash, also accounting pre-uichanges (which is not associated with uichange 
strings) dynamically. The keys of the compatibility hash are method patterns
and the corresponding value points to the method and option name changes
happened. While building up the compatibility hash, if an entry already exists,
it merges the new change with the previous change entry.

=over

=item Synopsis

$compatibility_hash = 
    NACL::APISet::Version::Node::Compatibility->build_compatibility_cli (
        uichanges => \@uichange_strings)

=item Arguments

=over

=item uichanges

(Mandatory)
Array reference to the list of uichange strings in chronological order.

=back

=item Returns

Hash reference to the compatibility hash.

=back

=cut
