# $Id: //depot/prod/test/nacldev/lib/NACL/APISet/Version/Node/Vserver.pm#1 $

#
# Copyright (c) 2001-2011 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Version class for Cluster mode.
## @author Isshwarya.M@netapp.com, dl-nacl-dev@netapp.com
## @status Shared
## @pod here

package NACL::APISet::Version::Node::Vserver;

#
# Compiler Directives
#
use strict;
use warnings;

#
# Module imports: APISet
#
use NACL::APISet::Session::CLI::CMode;
use NACL::APISet::Response::CLI::ParserUtils;

#
# Exceptions imported
#
use NATE::BaseException qw(:try);

#
# NATE imports
#
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Data::Dumper;
use Params::Validate qw(:all);

#
# Base class
#
use base qw(NACL::APISet::Version::Node);

use Class::MethodMaker [
    hash   => 'additional_uichanges',
    scalar => [
        {
            '-default_ctor' => sub {
                $Log->enter();
                my $self = $_[0];
                my $is_promoted = $self->_is_promoted_cmode_priv();

                # promoted builds are guaranteed to be non-debug
                $self->is_debug(0) if ($is_promoted);
                $Log->exit($is_promoted);
                return $is_promoted;
            },
        },
        'is_promoted',
    ],
];

##############################################################################
# Method:    _find_release
# Objective: Perform the discovery of release information
# Details:   See POD documentation at the end of this file.
##############################################################################

sub _find_release {
    my ( $self, %args ) = @_;

    $Log->enter(); 
    my $session;   

    # Try running the ngshell command "version".
    # We are not catching any exceptions as we had tried all possibilities to
    # obtain the information and we allow it to fail at this point.
    try {
        $session = $self->get_session(
            interface => "CLI",
            subtype   => "CMode"
        );

        # Find the information
        my $command = "version";
        my $output  = $self->execute_CLI_command(
            connection => $session->{connection},
            command    => $command
        );
        
        if ( defined $output ) {
            # If this regex is updated, be sure to update the
            # "vserver_31_version_test" test-case of
            # api_cmode_contextsupport.thpl. That test-script also lists the
            # various version strings that need to be handled by this regex.
            if ($output =~ /(?:NetApp|Data ONTAP) Release (.+?)[\s:]/) {
                $self->{release} = trim_string( string => $1 );
            }
        }
    }
    finally {

        # Destroy the session
        $self->destroy_session( session => $session ) if ( defined $session );
    };
    $Log->exit();
}

1;

=head1 NAME

NACL::APISet::Version::Node::Vserver - Version library for Vservers

=head1 SYNOPSIS

 use NACL::APISet::Version::Node::Vserver;

=head1 DESCRIPTION

A C<NACL::APISet::Version::Node::Vserver> is used to connect to Vserver and 
collect version related attributes. All version related attributes exposed from
this libary can also be set or get by the user. 

=cut

=head1 CONSTRUCTORS

=head2 new

Returns a reference to a newly created C<NACL::APISet::Version::Node::Vserver>.

=over 

=item Synopsis

$version = NACL::APISet::Version::Node::Vserver->new(%args)

=item Arguments

=over

=item category

(Mandatory)
The category to use.  Valid values include "Node".

=item shell_type

(Mandatory)
The refers to the type of the shell host sent through 'hostobj'.
Valid values include "Vserver".

=item hostobj

(Mandatory)
The hostrec object created for the resource.

=item conn_overrides

(Optional)
Hash reference to the list of connection overrides to be considered when
establishing the connection to the service.
Valid connection overrides are username, password, connid, conntype, ipaddress.

=back

=item Returns

A reference to a new Vserver Version object

=back

=cut

=head1 METHODS

=head2 _find_release

This method discovers release information by executing the ngshell 'version'
command to get information about release.

=over 

=item Synopsis

$release = $version_object->get_version_attribute(attribute => "release");

The above method will internally call $self->_find_release() if the information
about release atribute is not already cached in $self. This method shouldn't be
directly invoked by the user.

=item Arguments

Nothing

=item Returns

Nothing. Set the release information in version object itself. 

=back

=head2 is_promoted

Checks whether the service is running a promoted(release) or
development build.

=over 

=item Synopsis

$is_promoted = $version_object->is_promoted();

=item Arguments

Nothing

=item Returns

This method will return 1 for promoted builds and 0 for
development builds.

=back

=head1 AUTHOR/MAINTAINER

=over

=item NACL Development (dl-nacl-dev@netapp.com)

=back

 $Id: //depot/prod/test/nacldev/lib/NACL/APISet/Version/Node/Vserver.pm#1 $
 Last revised $Date: $ by $Author: isshwar $

=cut
