
#
# Copyright (c) 2011 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#
## @summary Version class for finding Version attributes specific to NetAppNetwork.
## @author rawat@netapp.com, dl-nacl-dev@netapp.com
## @status Shared
## @pod here

package NACL::APISet::Version::Switch::NetAppNetwork;

#
# Compiler Directives
#
use strict;
use warnings;

#
# Module imports
#
use NACL::APISet::Session::CLI::NetAppNetwork;
use NACL::APISet::Response::CLI::ParserUtils;

#
# NATE imports
#
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Params::Validate qw(:all);
use NATE::Exceptions::Argument qw(:try);

#
# base class
#
use base qw(NACL::APISet::Version::Switch);

###############################################################################
# Method:    _get_os_revision
# Objective: Will return Operating System revision
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _find_os_revision {
    my ($self) = @_;
    $Log->enter() if $may_enter;

    my $os_version;
    my $os_revision;
    my $session;

    try {
        $session = $self->get_session(interface  => "CLI",
                                      subtype    => "NetAppNetwork");
        
        my $password = $session->{connection}->{hostp}->{PROMPT_PASSWORD};
        if (! defined $password) {
            $password = $session->{connection}->{password};
        }
        
        my $output = $self->execute_CLI_command(connection  => $session->{connection},
                                                command     => "enable",
                                                'connectrec-match_table' => ["\[Pp\]assword:" => $password]
                                                );

        $output = $self->execute_CLI_command(connection  => $session->{connection},
                                                command     => "show version",
                                                'connectrec-match_table' => [ "--More-- or \\(q\\)uit"  => sub { $session->{connection}->send_intr(); undef } ],
                                                );

        my @lines = split(/\n/, $output);

        @lines = grep(/Machine Model/, @lines);
        if ($lines[0] =~ /^.*\w\.\.+\s+(.*)/i) {
            $os_revision = $1;
        }
        if ($output  =~ /Software Version\.\.+\s+(\S+)/i) {
            $os_version = $1;
        }

    } finally {
        # Destroy the session
        $self->destroy_session(session => $session) if (defined $session);
    };

    $self->{os_revision}->{major} = $os_revision;
    $self->{os_version}->{version} = $os_version;
    $Log->exit() if $may_exit;
    return $os_revision
}

1;


=head1 NAME

NACL::APISet::Version::Switch::NetAppNetwork - Version library for NetAppNetwork

=head1 SYNOPSIS

 use NACL::APISet::Version::Switch::NetAppNetwork;

=head1 DESCRIPTION

A C<NACL::APISet::Version::Switch::NetAppNetwork> is used to connect to NetAppNetwork Switch and 
collect version related attributes. All version related attributes exposed from
this libary can also be set or get by the user. 

=cut

=head1 CONSTRUCTORS

=head2 new

Returns a reference to a newly created C<NACL::APISet::Version::Switch::NetAppNetwork>.

=over 

=item Synopsis

$version = NACL::APISet::Version::Switch::NetAppNetwork->new(%args)

=item Arguments

=over

=item category

(Mandatory)
The category to use.  Valid values include "Switch".

=item shell_type

(Mandatory)
The refers to the type of the shell host sent through 'hostobj'.
Valid values include "Switch"

=item hostobj

(Mandatory)
The hostrec object created for the resource.

=item conn_overrides

(Optional)
Hash reference to the list of connection overrides to be considered when
establishing the connection to the service.
Valid connection overrides are username, password, connid, conntype, ipaddress,
prompt_username and prompt_password.

=back

=item Returns

A reference to a new NetAppNetwork Version object

=back

=cut

