#
# Copyright (c) 2001-2017 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#
## @summary This module maps the uichange string to its corresponding
## command and option name changes.
## @author Isshwarya.M@netapp.com, dl-central-libs, dl-nacl-dev
## @status Shared
## @see-also lib/Node/NG/Compatibility.pm

#
# NATE Imports
#
use Tie::IxHash;

#
# GUIDELINES TO ADD NEW UICHANGE STRING ENTRIES TO COMPATIBILITY HASH
# -------------------------------------------------------------------
#
# Please refer to the link
# http://wikid.netapp.com/w/QA/projects/Libraries_Initiative/Project/Common_Infrastructure/Functional_Spec/API_Func_Spec/Versioning_Design_Spec#CLI
#

# All pre-uichanges type changes
tie %compatibility, 'Tie::IxHash';
%compatibility = (
    "network_interfaces_routes_" => {
        new_command => "network_interface_route_",
    },
    "network_interfaces_" => {
        new_command => "network_interface_",
    },
    "network_ports_aggregates_" => {
       new_command => "network_port_aggregate_",
    },
    "network_ports_" => {
        new_command => "network_port_",
    },
    "protocols_" => {
        new_command => "protocol_",
    },
    "security_certificates_" => {
        new_command => "security_certificate_",
    },
    "services_nis_domain_" => {
        new_command => "vserver_services_nis_domain_",
    },
    "services_kerberos_realm_"  => {
        new_command => "vserver_services_kerberos_realm_",
    },
    "services_unix_user_"   => {
        new_command => "vserver_services_unix_user_",
    },
    "services_unix_group_"   => {
        new_command => "vserver_services_unix_group_",
    },
    "service_processor_"   => {
        new_command => "system_service_processor_",
    },
);

#
# Mapping for uichange strings
#
# Structure of each entry
# ------------------------
#
# uichange_string => {
#   command_pattern => {
#       new_command => new_command_pattern,
#       conversion => {
#           old_option => new_option,
#       },
#   },
# }
# NOTES:
# 1) 'new_command' is a required entry whether the uichange is associated
# with a command change or not. If there is no change, then command_pattern and
# new_command_pattern are one and the same. 'conversion' is an optional field
# and can be provided if the uichange string is associated with an option name
# change.
#
# 2) If a option change happened in all commands, then the command pattern
# "_ALL_COMMANDS" should be used in which case the 'new_command' entry is not
# required, the only exceptional case.
#
# 3)Entries for new ucihange strings can be added in any order to the
# %all_changes hash. The chronological order is maintained in the list of
# uichange strings obtained from the service.

%all_changes = (
    "group-migrate" => {
        "network_interface_migrate" => {
            new_command => "network_interface_migrate",
            conversion  => {
                "node"  => "dest-node",
                "port"  => "dest-port",
                "filer" => "dest-node",
            },
        },
    },
    "typos" => {
        "protocol_" => {
            new_command => "protocol_",
            conversion  => {
                "v2-enable"  => "v2-enabled",
                "v3-enable"  => "v3-enabled",
                "udp-enable" => "udp-enabled",
                "tcp-enable" => "tcp-enabled",
            },
        },
    },
    "filer-to-node" => {
        "filer_"        => { new_command => "node_", },
        "_ALL_COMMANDS" => {
            conversion => {
                "filer"       => "node",
                "filers"      => "nodes",
                "tofiler"     => "tonode",
                "byfiler"     => "bynode",
                "fromfiler"   => "fromnode",
                "panic-filer" => "panic-node",
                "filer-uuid"  => "node-uuid",
                "filername"   => "nodename",
                "home-filer"  => "home-node",
                "curr-filer"  => "curr-node",
            },
        },
    },
    "gx-license" => {
        "cluster_create" => {
            new_command => "cluster_create",
            conversion  => {
                "serialnumber" => "license",
            },
        },
    },
    "min-snapshot-space" => {
        "volume_" => {
            new_command => "volume_",
            conversion  => {
                "reserve-percent" => "min-snapshot-space",
            },
        },
    },
    "vol-rename-mirror-type" => {
        "volume_rename" => {
            new_command => "volume_rename",
            conversion  => {
                "new-name" => "newname",
            },
        },
    },
    "sfo-ui-inspection" => {
        "storage_failover_" => {
            new_command => "storage_failover_",
            conversion  => {
                "nodename"                 => "node-name",
                "partnername"              => "partner-name",
                "takeover-by-partner"      => "partner-reason",
                "takeover-by-partner-enum" => "partner-reason-enum",
            },
        },
        "storage_failover_interconnect_statistics_error_zeroout" => {
            new_command =>
              "storage_failover_interconnect_statistics_error_zero_out",
        },
    },
    "snapshot-ui-inspection" => {
        "volume_snapshot_" => {
            new_command => "volume_snapshot_",
            conversion  => {
                "crtime"      => "create-time",
                "description" => "comment",
            },
        },
    },
    "export-policy-default" => {
        "protocol_exportruleset_" => {
            new_command => "protocol_export_policy_",
        },
    },
    "cli-directories" => {
        "cluster_" => {
            new_command => "cluster_",
            conversion  => {
                "node-uuid" => "uuid",
            },
        },
        "node_" => {
            new_command => "system_",
            conversion  => {
                "node-uuid" => "uuid",
            },
        },
        "license_" => {
            new_command => "system_license_",
        },
        "protocol_cifs_server_"  => {
            new_command => "vserver_cifs_",
        },
        "protocol_nfs_server_"  => {
            new_command => "vserver_nfs_",
        },
        "protocol_"  => {
            new_command => "vserver_",
        },

        # This uichange-string is very old (added sometime in RR). It mapped
        # all system_service_ commands to system_services_. We needed to
        # remove this because in FS.0 a system service-processor set of
        # commands was added and this entry was stepping in and wrongly
        # converting it to system_services_processor. Hence, we're changing
        # it to perform the compatibility only for the required
        # sub-directories.
        "system_service_ntp_" => {
            new_command => "system_services_ntp_",
        },
        "system_service_firewall_" => {
            new_command => "system_services_firewall_",
        },
    },
    "export-policy-default-redux" => {
        "vserver_export_policy_" => {
            new_command => "vserver_export_policy_",
            conversion  => {
                "rulesetname"    => "policyname",
                "rulesetid"      => "policyid",
                "newrulesetname" => "newpolicyname",
                "newrulesetname" => "newpolicyname",
            },
        },
    },
    "cifs-nfs-ui-inspection" => {
        "vserver_cifs_" => {
            new_command => "vserver_cifs_",
        },
        "_ALL_COMMANDS" => {
            conversion => {
                "kerbconfig" => "kerberos-config",
                "dc1-ip"     => "dc1-address",
                "dc2-ip"     => "dc2-address",
                "dc3-ip"     => "dc3-address",
                "dc4-ip"     => "dc4-address",
                "dc5-ip"     => "dc5-address",
            },
        },
        "vserver_nfs_kerberos_config_" => {
            new_command => "vserver_nfs_kerberos_config_",
            conversion  => {
                "username" => "admin-username",
                "password" => "admin-password",
            },
        },
    },
    "vol-policy-router-metric" => {
        "network_interface_route_" => {
            new_command => "network_interface_route_",
            conversion  => {
                "hopcount" => "metric",
            },
        },
        "volume_" => {
            new_command => "volume_",
            conversion  => {
                "ruleset"          => "policy",
                "ruleset-override" => "policy-override",
            },
        },
    },
    "nfs-enabled-fields" => {
        "vserver_nfs_" => {
            new_command => "vserver_nfs_",
            conversion  => {
                "v2-enabled"  => "v2",
                "v3-enabled"  => "v3",
                "udp-enabled" => "udp",
                "tcp-enabled" => "tcp",
            },
        },
    },
    "v4.0" => {
        "vserver_nfs_" => {
            new_command => "vserver_nfs_",
            conversion  => {
                "v4"                  => "v4.0",
                "v4-acl"              => "v4.0-acl",
                "v4-read-delegation"  => "v4.0-read-delegation",
                "v4-write-delegation" => "v4.0-write-delegation",
                "v4-req-open-confirm" => "v4.0-req-open-confirm",
            },
        },
    },
    "dashboard" => {
        "statistics_" => {
            new_command => "statistics_",
            conversion  => {
                "category-instance" => "object",
            },
        },
    },
    "restriping-lite" => {
        "storage_aggregate_add" => {
            new_command => "storage_aggregate_add_disks",
        },
        "storage_aggregate_add_member" => {
            new_command => "storage_aggregate_add_members",
        },
        "storage_aggregate_member_add" => {
            new_command => "storage_aggregate_member_add_disks",
        },
    },
    "vif-to-lif" => {
        "network_interface_migrate" => {
            new_command => "network_interface_migrate",
            conversion  => {
                "vif"             => "lif",
                "vif_role"        => "lif_role",
                "vif-total"       => "lif-total",
                "vif-online"      => "lif-online",
                "vif-nothome"     => "lif-nothome",
                "vif-no-failover" => "lif-no-failover",
                "vif-nothosted"   => "lif-nothosted",
                "vif-name"        => "lif-name",
                "vif-id"          => "lif-id",
            },
        },
        "_ALL_COMMANDS" => {
            conversion => {
                "vif"             => "lif",
                "vif_role"        => "lif_role",
                "vif-total"       => "lif-total",
                "vif-online"      => "lif-online",
                "vif-offline"     => "lif-offline",
                "vif-nothome"     => "lif-nothome",
                "vif-no-failover" => "lif-no-failover",
                "vif-nothosted"   => "lif-nothosted",
                "vif-name"        => "lif-name",
                "vif-id"          => "lif-id",
            },
        },
    },
    "vserver-role-change" => {
        "vserver_show" => {
            new_command => "vserver_show",
            conversion  => {
                "role" => "type",
            },
        },
    },
    "cifs-server" => {
        "vserver_cifs_share_ACL_create" => {
            new_command => "vserver_cifs_share_access_control_create",
        },
        "vserver_cifs_share_ACL_delete" => {
            new_command => "vserver_cifs_share_access_control_delete",
        },
        "vserver_cifs_share_ACL_modify" => {
            new_command => "vserver_cifs_share_access_control_modify",
        },
        "vserver_cifs_share_ACL_show" => {
            new_command => "vserver_cifs_share_access_control_show",
        },
    },
    "volops-directory" => {
        "volume_move"        => {
            new_command => "volume_move_start",
        },
        "volume_member_move" => {
            new_command => "volume_member_move_start",
        },
        "volume_copy"        => {
            new_command => "volume_copy_start",
        },

    },
    "server_to_vserver_network" => {
        "network_interface_migrate" => {
            new_command => "network_interface_migrate",
            conversion  => {
                "server" => "vserver",
            },
        },
        "network_interface_revert" => {
            new_command => "network_interface_revert",
            conversion  => {
                "server" => "vserver",
            },
        },
        "network_interface_rename" => {
            new_command => "network_interface_rename",
            conversion  => {
                "server" => "vserver",
            },
        },
        "network_interface_enable_readonly" => {
            new_command => "network_interface_enable_readonly",
            conversion  => {
                "server" => "vserver",
            },
        },
        "network_interface_disable_readonly" => {
            new_command => "network_interface_disable_readonly",
            conversion  => {
                "server" => "vserver",
            },
        },
        "network_interface_delete" => {
            new_command => "network_interface_delete",
            conversion  => {
                "server" => "vserver",
            },
        },
        "network_interface_new" => {
            new_command => "network_interface_new",
            conversion  => {
                "server" => "vserver",
            },
        },
        "network_interface_create" => {
            new_command => "network_interface_create",
            conversion  => {
                "server" => "vserver",
            },
        },
        "network_interface_modify" => {
            new_command => "network_interface_modify",
            conversion  => {
                "server" => "vserver",
            },
        },
        "network_interface_show" => {
            new_command => "network_interface_show",
            conversion  => {
                "server" => "vserver",
            },
        },
        "network_interface_failover_" => {
            new_command => "network_interface_failover_",
            conversion  => {
                "server" => "vserver",
            },
        },
        "network_interface_show_" => {
           new_command => "network_interface_show_",
           conversion  => {
               "server" => "vserver",
           },
        },
        "network_connections_" => {
            new_command => "network_connections_",
            conversion  => {
                "server" => "vserver",
            },
        },
        "network_routing_groups_" => {
            new_command => "network_routing_groups_",
            conversion  => {
                "server" => "vserver",
            },
        },
    },
    "vserver-lun-igroup-portset-rr-0" => {
        "lun_" => {
            new_command => "lun_",
            conversion => {
                "space-resv" => "space-reserve",
                "space-resv-honored" => "space-reserve-honored",
                "space-alloc" => "space-allocation",
                "isOnline" => "is-online",
                "isMapped" => "is-mapped",
            },
        },
        "lun_igroup_" => {
            new_command => "lun_igroup_",
            conversion => {
                "protocol_type" => "protocol",
            },
        },
        "lun_portset_" => {
            new_command => "lun_portset_",
            conversion => {
                "protocol_type" => "protocol",
            },
        },
    },
    "vserver-iscsi-rr-0" => {
        "vserver_iscsi_new" => {
            new_command => "vserver_iscsi_new",
            conversion => {
                "alias" => "target-alias",
                "nodename" => "target-name",
            },
        },
        "vserver_iscsi_create" => {
            new_command => "vserver_iscsi_create",
            conversion => {
                "alias" => "target-alias",
                "nodename" => "target-name",
            },
        },
        "vserver_iscsi_modify" => {
            new_command => "vserver_iscsi_modify",
            conversion => {
                "alias" => "target-alias",
                "nodename" => "target-name",
            },
        },
        "vserver_iscsi_show" => {
            new_command => "vserver_iscsi_show",
            conversion => {
                "alias" => "target-alias",
                "nodename" => "target-name",
            },
        },
        "vserver_iscsi_nodename" => {
            new_command => "vserver_iscsi_nodename",
            conversion => {
                "nodename" => "target-name",
            },
        },
        "vserver_iscsi_alias" => {
            new_command => "vserver_iscsi_alias",
            conversion => {
                "alias" => "target-alias",
            },
        },
        "vserver_iscsi_initiator_show" => {
            new_command => "vserver_iscsi_initiator_show",
            conversion => {
                "alias" => "initiator-alias",
            },
        },
    },
    "qos-policy-read-ahead-name-change-rr-0" => {
        "qos_policy_cache_" => {
            new_command => "qos_policy_cache_",
            conversion => {
                "name" => "cache-policy",
            },
        },
        "qos_policy_read_ahead_" => {
            new_command => "qos_policy_read_ahead_",
            conversion => {
                "name" => "read-ahead-policy",
            },
        },
        "qos_policy_" => {
            new_command => "qos_policy_",
            conversion => {
                "cache" => "cache-policy",
                "read-ahead" => "read-ahead-policy",
            },
        },
    },
    "vserver-cifs-domain-preferred-dc-v2" => {
        "vserver_cifs_domain_preferred_dc_create" => {
            new_command => "vserver_cifs_domain_preferred_dc_add",
        },
        "vserver_cifs_domain_preferred_dc_delete" => {
            new_command => "vserver_cifs_domain_preferred_dc_remove",
        },
    },
    "remove-background-from-bg-compression" => {
        "volume_efficiency_" => {
            new_command => "volume_efficiency_",
            conversion => {
                "background-compression" => "compression",
            },
        },
    },
    "rdb-quorum-health-ui-timestamp" => {
        "statistics_rdb_quorum_events_show" => {
            new_command => "statistics_rdb_quorum_events_show",
            conversion => {
                "timestamp" => "eventtime",
            },
        },
    },
    "object-volume-group" => {
        "repository_balance_capacity_" => {
            new_command => "repository_balance_capacity_",
            conversion => {
                "object-volume-group" => "data-constituent-group",
            },
        },
    },
    "repository-balance-capacity-modify-data-constituent-group" => {
        "repository_balance_capacity_" => {
            new_command => "repository_balance_capacity_",
            conversion => {
                "data-constituent-group" => 'constituent-group-id'
            },
        },
    },
    "repository-placement-file-create-modify-data-constituent-group" => {
        "repository_placement_file_create_" => {
            new_command => "repository_placement_file_create_",
            conversion => {
                "data-constituent-group" => 'constituent-group-id'
            },
        },
    },
    "cifs-session-file-show" => {
        "vserver_cifs_utilities_session_show" => {
            new_command => "vserver_cifs_session_show",
        },
        "vserver_cifs_utilities_file_show" => {
            new_command => "vserver_cifs_session_file_show",
        }
    },
    "volume-check-volume-name" => {
        "volume_check_show" => {
            new_command => "volume_check_show",
            conversion => {
                "volume-name" => "volume",
            },
        },
    },
    "vserver-peer-transition" => {
        "vserver_peer_transition_" => {
            new_command => "vserver_peer_transition_",
            conversion => {
                "localVserver" => "local-vserver",
                "srcFilerName" => "src-filer-name",
                "srcFilerUuid" => "src-filer-uuid",
                "localVsUuid" => "local-vs-uuid",
            },
        },
    },
    "vserver-locks-filter-name" => {
        "vserver_locks_show" => {
            new_command => "vserver_locks_show",
            conversion => {
                "smb2-open-type" => "smb-open-type",
                "smb2-connect-state" => "smb-connect-state",
                "smb2-expiration-time" => "smb-expiration-time",
                "smb2-open-group-id" => "smb-open-group-id",
            }
        },
        "debug_locks_mirrored_show" => {
            new_command => "debug_locks_mirrored_show",
            conversion => {
                "smb2-open-type" => "smb-open-type",
                "smb2-connect-state" => "smb-connect-state",
                "smb2-expiration-time" => "smb-expiration-time",
                "smb2-open-group-id" => "smb-open-group-id",
            }
        }
    },
    "deprecate-enable-wipeconfig" => {
        "metrocluster_enable" => {
            new_command => "metrocluster_configure",
        },
        "metrocluster_wipeconfig" => {
            new_command => "metrocluster_unconfigure",
        },
    },
    "mcc-interconnect-commands-review" => {
        "metrocluster_interconnect_show" => {
            new_command => "metrocluster_interconnect_mirror_show",
        },
        "metrocluster_interconnect_modify" => {
            new_command => "metrocluster_interconnect_mirror_modify",
        },
    },
    "lun-unmap-wildcard" => {
        "lun_map" => {
            new_command => "lun_mapping_create",
        },
        "lun_unmap" => {
            new_command => "lun_mapping_delete",
        },
        "lun_mapped_show" => {
            new_command => "lun_mapping_show",
        },
    },
    "vserver-services-name-service-mapping" => {
        "vserver_services_netgroup_" => {
            new_command => "vserver_services_name_service_netgroup_",
        },
        "vserver_services_dns_" => {
            new_command => "vserver_services_name_service_dns_",
        },
        "vserver_services_nis_domain_" => {
            new_command => "vserver_services_name_service_nis_domain_",
        },
        "vserver_services_ldap_" => {
            new_command => "vserver_services_name_service_ldap_",
        },
        "vserver_services_unix_user_" => {
            new_command => "vserver_services_name_service_unix_user_",
        },
        "vserver_services_unix_group_" => {
            new_command => "vserver_services_name_service_unix_group_",
        },
        "vserver_services_dns_hosts_" => {
            new_command => "vserver_services_name_service_dns_hosts_",
        },
    },
        "mcc-heal-bypass-change" => {
        "metrocluster_heal" => {
            new_command => "metrocluster_heal",
            conversion => {
                "bypass-nodes-not-switched-over" => "force-if-partial-switchover",
            },
        },
    },
    "lun-move-renamed-to-lun-move-in-volume" => {
        "lun_relocation_" => {
            new_command => "lun_move_",
        },
        "lun_move" => {
            new_command => "lun_move_in_volume",
        },
    },
    "vs-failover-ports-renamed-to-targets" => {
        "network_interface_failover_groups_add_ports" => {
            new_command => "network_interface_failover_groups_add_targets",
            conversion => {
                "ports" => "targets",
            }
        },
        "network_interface_failover_groups_remove_ports" => {
            new_command => "network_interface_failover_groups_remove_targets",
            conversion => {
                "ports" => "targets",
            }
        },
    },
    "storage-pool-ui-update" => {
        "storage_pool_delete" => {
            new_command => "storage_pool_delete",
            conversion => {
                "force" => "quite",
            },
        },
    },
    "storage-switch-add-change" => {
        "storage_switch_add" => {
            new_command => "storage_switch_add",
            conversion => {
                "switch-ipaddress" => "address",
            },
        },
    },
    "storage-pool-spare-ui-update" => {
        "storage_pool_show_spare_capacity" => {
            new_command => "storage_pool_show_available_capacity",
            conversion => {
                "spare-usable-size" => "available-size",
            },
        },
        "storage_pool_add" => {
            new_command => "storage_pool_add",
            conversion => {
                "spare-capacity" => "available-capacity",
                "new-spare-capacity" => "new-available-capacity",
            },
        },
    },
);
1;
