#
# Copyright (c) 2001-2014 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#
## @summary Common routines for CLI and ZAPI APISets
## @author dl-nacl-dev@netapp.com
## @status shared
## @pod here

package NACL::APISet::_Mixins::CLIZAPI;

use strict;
use warnings;
use feature 'state';

use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Params::Validate qw(validate_with validate_pos SCALAR CODEREF OBJECT);
use NATE::BaseException qw(:try);
use NACL::ComponentUtils;

BEGIN {
    use Exporter qw(import);
    my @baseline_methods = qw(_validate_and_set_baseline_args _move_baseline_args);
    our @EXPORT_OK = (
        qw(
            _handle_baseline_errors
            _vserver_locked_recording_baseline_transfer_regex
            _internal_mcc_or_vdr_error_regex
        ),
        @baseline_methods
    );
    our %EXPORT_TAGS = (
        'all'           => \@EXPORT_OK,
        'baseline_opts' => \@baseline_methods,
    );
}

=head1 NAME

NACL::APISet::_Mixins::CLIZAPI

=head1 DESCRIPTION

This package is a mixin containing methods which are common to the
various CLI and ZAPI APISets.

=head1 METHODS

=head2 _handle_baseline_errors

    $apiset->_handle_baseline_errors(code => $coderef);
    
For a MetroCluster or VDR-relationship, if a baseline transfer (which is
basically a configuration transfer from source -> destination cluster/vserver)
is ongoing while an operation is performed, then that operation will fail
because locks are held on those elements during the period of the
baseline transfer. Since the timing of baseline transfers is non-deterministic,
it is impossible to predict when it will happen, so scripts cannot be
coded to handle this error. Hence, the CMode APISet (both CLI and ZAPI)
is coded to catch these specific errors and retry for upto 2 minutes.

This behaviour can be configured either through the script or with a
command-line parameter:

=over

=item Through the script

Passing C<< nacl_baseline_retry => 0 >> in the script will disable retrying on
these errors. The C<nacl_baseline_time> argument can be used to configure the
time to retry for. Examples:

    # Don't retry on errors
    $apiset->volume_create(..., nacl_baseline_retry => 0);
    
    # Retry for 180 seconds
    $apiset->volume_create(..., nacl_baseline_time => 180);

=item Command-line parameters

The equivalent command-line parameters are C<NACL_BASELINE_RETRY> and
<NACL_BASELINE_TIME>, respectively.

=back

=cut

sub _handle_baseline_errors {
    $Log->enter();
    my ($self, @args) = @_;
    state $spec = {code => {type => CODEREF}};
    validate_with(
        params      => \@args,
        spec        => $spec,
        allow_extra => 1,
    );
    my %opts = $self->_validate_and_set_baseline_args(@args);
    my $code = $opts{code};

    my $time_to_wait = $opts{nacl_baseline_time};
    my $server_name;
    if ($self->{server_name_override}) {
        $server_name = $self->{server_name_override};
    }
    my $wait_or_throw_exception = sub {
        my ($exception, $time_to_reduce) = validate_pos(@_, {type => OBJECT},
            {type => SCALAR});
        
        # The command itself has internally tried to execute for this much
        # time, so reduce the time to wait by that amount
        $time_to_wait -= $time_to_reduce;
        if ($opts{nacl_baseline_retry}) {
            if ($time_to_wait <=0) {
                $exception->set_text('Even after waiting for 120 seconds, the '
                    . 'command did not succeed due to a MetroCluster/VDR operation.'
                    . " Original error:\n" . $exception->text());
                $Log->exit();
                $exception->throw();
            } else {
                $Log->warn('Hit error due to MetroCluster/VDR. Will snooze '
                    . "for $time_to_reduce seconds before retrying. (See burt 750933)");
                
                # Snooze and reduce the time to wait
                Tharn::snooze($time_to_reduce);
                $time_to_wait -= $time_to_reduce;
            }
        } else {
            $Log->exit();
            $exception->throw();
        }
    };

    my ($done, $ret);
    while (1) {
        try {
            $ret = $code->();
            $done = 1;
        } catch NACL::APISet::Exceptions::VserverLockedRecordingBaselineTransfer with {
            # In this case, the command has internally tried for 60 seconds
            $wait_or_throw_exception->($_[0], 60);
            $self->{server_name_override}=$server_name
                if ( $server_name );
        } catch NACL::APISet::Exceptions::InternalMCCOrVDRError with {
            # In this case, the command has internally tried for 10 seconds
            $wait_or_throw_exception->($_[0], 10);
            $self->{server_name_override}=$server_name
                if ( $server_name );
        };
        last if ( $done );
    }
    
    $Log->exit();
    return $ret;
}

sub _vserver_locked_recording_baseline_transfer_regex {
    return qr/This Vserver is temporarily locked from changes while recording the configuration baseline/i;
}

sub _internal_mcc_or_vdr_error_regex {
    return qr/Another operation on this object is currently in progress|A background task is currently collecting the configuration|The MetroCluster or Vserver DR Configuration Replication Service is recovering/i;
}

sub _validate_and_set_baseline_args {
    $Log->enter();
    my ($self, @opts) = @_;
    state $baseline_spec = {
        nacl_baseline_retry => {type => SCALAR, optional => 1},
        nacl_baseline_time  => {type => SCALAR, optional => 1},
    };
    my %args = validate_with(
        params      => \@opts,
        spec        => $baseline_spec,
        allow_extra => 1,
    );
    # Make sure to check if (defined) rather than if() because if() would return
    # false even if the value was explicitly provided as 0.
    if (!defined $args{nacl_baseline_retry}) {
        my $baseline_retry_param = Tharn::param('NACL_BASELINE_RETRY');
        if (defined $baseline_retry_param) {
            $args{nacl_baseline_retry} = $baseline_retry_param;
        } else {
            $args{nacl_baseline_retry} = 1;
        }
    }
    if (!defined $args{nacl_baseline_time}) {
        my $baseline_time_param = Tharn::param('NACL_BASELINE_TIME');
        if (defined $baseline_time_param) {
            $args{nacl_baseline_time} = $baseline_time_param;
        } else {
            $args{nacl_baseline_time} = 120;
        }
    }
    $Log->exit();
    return %args;
}

sub _move_baseline_args {
    $Log->enter();
    my ($self, @opts) = @_;
    my %baseline_args;
    NACL::ComponentUtils->_hash_move(
        @opts,
        target => \%baseline_args,
        move   => [qw(nacl_baseline_retry nacl_baseline_time)],
    );
    $Log->exit();
    return %baseline_args;
}


1;
