
#
# Copyright (c) 2001-2012 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Common routines for ZAPI APISets
## @author dl-nacl-dev@netapp.com
## @status shared
## @pod here

package NACL::APISet::_Mixins::ZAPI;

use strict;
use warnings;

use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Data::Dumper qw(Dumper);
use Params::Validate qw(:all);
use XML::LibXML;
use XML::Simple;

# Import exceptions
use NACL::APISet::Exceptions::InvalidXMLException qw(:try);

BEGIN {
    use Exporter qw(import);
    our @EXPORT_OK = qw(
        get_cdef_method
        AUTOLOAD
        set_port
        _get_xpathcontext
        _xml2perl
        _perl2xml
    );
    our %EXPORT_TAGS = ('all' => \@EXPORT_OK);
}

=head1 NAME

NACL::APISet::_Mixins::ZAPI

=head1 DESCRIPTION

This package is a mixin containing methods which are common to the
various ZAPI APISets.

=cut

sub get_cdef_method {
    $Log->enter() if $may_enter;
    my $self = $_[0];
    $Log->exit() if $may_exit;
    return $self->{cdef_method};
}

sub AUTOLOAD {
    my ( $self, @opts ) = @_;
    $Log->enter() if $may_enter;
    our $AUTOLOAD;

    my $method = $AUTOLOAD;
    $method =~ s/.*://;    # strip fully-qualified portion
    $Log->debug( sub { "Options to '$method' are:\n" . Dumper( {@opts} ) } );

    # Let unknown methods hit AUTOLOAD only the first time. After that,
    # install it into the symbol table of the package so that it doesn't
    # need to go through AUTOLOAD again.
    my $sub;
    if ( $method =~ /^xlate_/ ) {
        $sub = sub {
            $Log->enter() if $may_enter;
            my ( $self, @opts ) = @_;
                my %args = $self->validate_params(
                params      => \@opts,
                spec        => {},
                allow_extra => 1
            );

            $Log->exit() if $may_exit;
            return $self->_only_translate_method( method => $method, %args );
        };
    } else {
        $method =~ s/\_/\-/g;
        $sub = sub {
            $Log->enter() if $may_enter;
            my ( $self, @opts ) = @_;
                my %args = $self->validate_params(
                params      => \@opts,
                spec        => {},
                allow_extra => 1
            );

            # Not really in AUTOLOAD after the first time, but there are
            # log parsers that look for this particular message to determine
            # which ZAPIs were invoked in a script, so we're leaving the
            # log message the same as it was earlier.
            $Log->comment("In AUTOLOAD executing ZAPI $method");
            $self->{cdef_method} = $method;
            $Log->exit() if $may_exit;
            return $self->_invoke( cdef_method => $method, %args );
        };
    }

    # Symbolic references are disabled by default. Allow them temporarily
    # by disabling strict refs.
    no strict 'refs';
    *{$AUTOLOAD} = $sub;
    use strict 'refs';
    $Log->exit() if $may_exit;
    return $sub->(@_);
}

sub set_port {
    $Log->enter() if $may_enter;

    my $self = shift();

    NATE::BaseException->throw("Port number can only be set for ZAPI")
        if ( $self->{interface} ne "ZAPI" );

    my %args = NACL::APISet::validate_params(
        $self,
        params => \@_,
        spec   => { "port" => { type => SCALAR, optional => 0 } }
    );

    NATE::BaseException->throw( "Invalid port number: " . $args{port} )
        if ( $args{port} !~ /^\d+$/ );
    $self->{session}->{port} = $args{port};

    $Log->exit() if $may_exit;
} ## end sub set_port

sub _get_xpathcontext {
    $Log->enter() if $may_enter;
    my ($self, %args) =  @_;
    
    validate_with(
        params => \%args,
        spec   => {
            xmlString => { type => SCALAR },
        }
    );
    
    my $xml = $args{xmlString};
    my $libxml;
    if (! defined $xml or $xml eq "") {
        $libxml = XML::LibXML->new();
    } else {
        $libxml = XML::LibXML->load_xml(string => $xml);
    }
    
    $Log->exit() if $may_exit;
    return XML::LibXML::XPathContext->new($libxml);
}

#################################################################
## Will return object of type XML::Simple
#################################################################
sub _xml2perl {
    my ($self, %args) =  @_;
    
    validate_with(
        params => \%args,
        spec   => {
            xmlString => { type => SCALAR },
        },
        allow_extra => 1,
    );
    # Setting preferred parser as XML::Parser for XML::Simple
    # it seems to be the quickest
    # For more details see burt694608 and burt773746
    local $XML::Simple::PREFERRED_PARSER="XML::Parser";
    my $xml = delete $args{xmlString};
    my $perldata;
    eval {
        $perldata = XML::Simple::XMLin($xml, %args);
    };
    if($@) {
        ### [CSI]: DUMP RAW XML ###
        $Log->comment("\n" . ('*' x 80) . "\n*** Raw XML is:\n" . ('*' x 80));
        $Log->comment($xml);
        $Log->comment(('*' x 80) . "\n");

        $Log->exit() if $may_exit;
        NACL::APISet::Exceptions::InvalidXMLException->throw(
            "Invalid XML is specified. Error is $@");
    }
    
    $Log->exit() if $may_exit;
    return $perldata;
}

sub _perl2xml {
    my ($self, %args) =  @_;
    
    validate_with(
        params => \%args,
        spec   => {
            perldata => { type => HASHREF },
        },
        allow_extra => 1,
    );
    # Setting preferred parser as XML::Parser for XML::Simple
    # it seems to be the quickest
    # For more details see burt694608 and burt773746
    local $XML::Simple::PREFERRED_PARSER="XML::Parser";
    my $data = delete $args{perldata};
    my $xml;
    eval {
        $xml = XML::Simple::XMLout($data, %args);
    };
    if($@) {
        $Log->exit() if $may_exit;
        NACL::APISet::Exceptions::InvalidXMLException->throw(
            "The translated xml structure cannot be formatted. Error is $@");
    }
    
    $Log->exit() if $may_exit;
    return $xml;
}

1;
