# Copyright (c) 2015 NetApp, Inc., All Rights Reserved.
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#
## @summary Library of core events for the EventWorker to invoke
## @author fujii@netapp.com,rjray@netapp.com
## @status shared
## @pod here

package NACL::AppBuilder::DpgWorkflows::CoreEvents;

use 5.014;
use strict;
use warnings;
use namespace::autoclean;

use Module::Load;

use Const::Fast;
use Moose;
use Params::Validate qw(validate_with validation_options
                        OBJECT HASHREF SCALAR ARRAYREF BOOLEAN);
use Scalar::Util qw(blessed);

use NATE::Log qw(log_global);
use NACL::GeneralUtils qw(random_name_generator);
use NACL::STask::Node;
use NACL::STask::Volume;

with 'NACL::AppBuilder::DpgWorkflows::EventBase';

# Global handle on the NATE::Log logging interface
const my $Log => log_global;

__PACKAGE__->meta->make_immutable;

###############################################################################
# NOTE: Because I am using Moose for the object model, I have moved the usage
# of any NATE or NACL exception classes down here, after an explicit
# "no Moose;" statement. Do not try to move this higher in the code, or
# remove the "no" statement, or compilation of the module will break.
###############################################################################

no Moose;
use NATE::BaseException qw(:try);
use NATE::Exceptions::Argument;

# Set the on_fail behavior for Params::Validate to throw an exception (of the
# NATE::Exceptions::Argument class) on errors instead of a plain croak().
validation_options(
    on_fail => sub {
        NATE::Exceptions::Argument->throw(shift);
    }
);

=head1 NAME

NACL::AppBuilder::DpgWorkflows::CoreEvents

=head1 SYNOPSIS

This class is not used directly within scripts, or even within other classes.
It is loaded dynamically by the B<NACL::AppBuilder::DpgWorkflows::EventWorker>
class in code sent to remote clients.

=head1 DESCRIPTION

The B<NACL::AppBuilder::DpgWorkflows::CoreEvents> class provides
implementations of core/basic events for the EventManager system. These are the
basic events that are expected to cover the majority of DMG needs for the
events subsystem. It utilizes the B<NACL::AppBuilder::DpgWorkflows::EventBase>
role, which provides the basis of the interface with the EventWorker that is
dispatching this event.

Note that in the context of the B<NACL::AppBuilder::DpgWorkflows::EventWorker>
(and B<NACL::AppBuilder::DpgWorkflows::EventManager>), the term I<event> is
used differently than it is used in "event-based programming". Here, event
refers to code that runs parallel and independently from the main test. It is
expected that the code will perform actions that can have an effect on the
main test (such as enabling or disabling ports, taking a volume offline, etc.).
As such, this code is referred to a an event, as it may cause "events" to occur
within the main test.

This class is not intended to be directly used in a THPL file or other Perl
library. It is intended only to be used on remote clients by the remote code
that B<NACL::AppBuilder::DpgWorkflows::EventWorker> generates. Because it only
executes on remote clients (and/or processes) that are different than the Perl
process that is executing the EventWorker/EventManager code, an event object
has no direct connection or linkage back to the worker that spawned it other
than what is provided by STAF. In general, events should not need to directly
communicate with the parent worker.

The controlling EventWorker will run a specified event in a loop that is
governed by the C<iterations> and C<runlength> parameters passed to the
worker when it is created. See the section L</EVENT ITERATION LOOP>, below.

=head1 ATTRIBUTES

The following attributes are provided by the
B<NACL::AppBuilder::DpgWorkflows::EventBase> role. Each attribute value is
obtained by calling the attribute name as an accessor. All but one of the
attributes are read-only (they cannot be changed after the object is created).

=over

=item client

The name of the client that this event is executing on. This may or may not
be the fully-qualified domain name (FQDN) of the machine, depending on how it
was specified to either the EventManager or EventWorker.

=item interface

A string, one of C<CLI> or C<ZAPI>, that defines the interface method that
this event will use for any node or cluster connections that it makes to
ONTAP.

=item event

The name of the event that is being executed by this object handle. This is
set because an event may call other methods in the class as part of its
processing. However, this value will never change. This means that any method
may look at this attribute to see what actual event is being processed, event
if that event is different than the currently-executing method.

=item iteration

The number of the current iteration, starting at 1. Unlike the rest of the
attributes, this attribute is writable. However, the code that controls the
event object will set the value each time through (so any changes made by the
event itself will be lost).

=item max_iterations

The number of iterations of the event that will be done before the worker
considers itself complete and exits. If this value is zero, then the event
is being run for a specified period of time rather than a specific number of
iterations. See C<runlength>, next.

=item runlength

If set, the number of seconds that the iteration loop will run. If set, then
the C<max_iterations> value will always be zero. Note that this does not mean
that the event will be interrupted at this precise number of seconds; if the
C<runlength> is reached during the execution of an iteration, then that
iteration will finish before the worker completes on the client.

=item start_time

The start-time (in UNIX seconds format) when the remote code started performing
the event's iteration-loop.

=item state

A hash-reference made available to all event methods in order to pass or
preserve information between iterations of the event. Note that the attribute
itself is read-only in that the reference cannot be overwritten or otherwise
changed. However, you can add/delete/modify keys on the hash-reference itself
without any conditions.

=back

=head1 METHODS

In addition to the attribute accessors, the class provides the following
methods. No subroutines are exported by this package.

=head2 new

    my $handle = NACL::AppBuilder::DpgWorkflows::CoreEvents->new(
        client         => $client,
        interface      => $interface,
        event          => $event,
        max_iterations => $iterations,
        runlength      => $runlength,
        start_time     => $start_time
    );

May only be called as a class method.

This is the object constructor, and is the only non-private method that is not
available as an event method. The parameters it takes all correspond to
attributes given in L</ATTRIBUTES>, above. Note that the attribute C<iteration>
is not a valid parameter to B<new>, since it is explicitly set before each
iteration call to the event. Likewise, the C<state> attribute is not passable
in the constructor since it is created internally for each object.

The constructor is provided by the Moose object system.

=over

=item Options

The following options are recognized by B<new>. All are required, none are
optional.

=over

=item C<< client => $string >>

A string that contains the name of the client this event is being run on.

=item C<< interface => 'CLI'|'ZAPI' >>

A string whose value is one of C<CLI> or C<ZAPI>. Any other value passed in for
this parameter will result in an exception being thrown.

=item C<< event => $string >>

A string that contains the name of the event that is being remotely invoked.

=item C<< max_iterations => $integer >>

An integer that specifies the maximum number of iterations that will be
performed on this event. If this value is zero, then the number of iterations
is being bounded by C<runlength>, instead.

=item C<< runlength => $integer >>

An integer that specifies the length of time that will be spent on iterations
of this event. When an iteration completes, the current time is compared to
the starting time and C<runlength>. If these indicate that the C<runlength>
has been reached or exceeded, the client code will finish and return to the
worker.

If this value is zero, then the number of iterations is being bounded by
C<max_iterations>, instead.

=item C<< start_time => $integer >>

An integer value, from the built-in B<time> function, that indicates when the
execution of the iteration loop started. This will be the same value for all
iterations of the event, it is not reset on each iteration.

=back

=item Exceptions

The only exceptions thrown by this method are thrown by Moose itself for a
parameter that is unknown, or whose value does not match the type of the
attribute.

=back

Upon success, the return value of this method is the new object.

=head2 dummyOp

A dummy event call. This event does nothing except return.

It takes no options and throws no exceptions.

=cut

sub dummyOp {
    return 1;
}

=head2 delay

An event that simply sleeps for a specified period of time before returning.

=over

=item Options

This method takes only one option:

=over

=item C<< event_delay => $integer >>

(Optional) The length of time, in seconds, to delay before returning. Defaults
to 60 if not given.

=back

=item Exceptions

This method throws no exceptions.

=back

=cut

sub delay {
    my $self = shift;
    my %opts = validate_with(
        params      => \@_,
        allow_extra => 1,
        spec        => {
            event_delay => { type => SCALAR, default => 60 },
        },
    );

    Tharn::snooze($opts{event_delay});

    return 1;
}

=head2 takeover

Perform a storage failover takeover.

=over

=item Options

The following options are recognized by B<takeover>:

=over

=item C<< node => $string|$node_obj >>

(Required) The node that will perform the takeover for its partner. The value
may be either a string with the name of the node or an object of the
B<NACL::STask::Node> class.

=item C<< node_partner => $string|$node_obj >>

(Required) The partner node that will be taken over. Like C<node>, above,
may be either a string with the name of the node or an object of the
B<NACL::STask::Node> class.

=item C<< takeovertype => $string >>

(Optional) Defaults to C<takeover_command>. See the documentation for
B<takeover> in the B<NACL::STask::StorageFailover> manual page for other
possible values for this parameter.

=item C<< timeout => $integer >>

(Optional) The timeout value to pass when performing the takeover operation,
passed to the B<takeover> method as the C<method-timeout> parameter. Defaults
to 600 seconds.

=back

Any other arguments passed will be given directly to the B<takeover> method of
B<NACL::STask::StorageFailover> when it is called.

=item Exceptions

Invalid parameters will trigger a B<NATE::Exceptions::Argument> exception. Any
exceptions that may be thrown by the B<takeover> method of
B<NACL::STask::StorageFailover> will be passed through here unchanged.

=back

The return value upon success is a "true" value (C<1>, in this case).

=cut

sub takeover {
    my $self = shift;
    my %opts = validate_with(
        params      => \@_,
        allow_extra => 1,
        spec        => {
            node         => { type => SCALAR | OBJECT },
            node_partner => { type => SCALAR | OBJECT },
            timeout      => { type => SCALAR, default => 600 },
            takeovertype => {
                type    => SCALAR,
                default => 'takeover_command',
            },
        },
    );

    require NACL::STask::StorageFailover;

    my ($cluster, $node, $partner) =
        $self->_getClusterAndNodes(delete @opts{qw(node node_partner)});

    NACL::STask::StorageFailover->takeover(
        command_interface            => $cluster,
        node                         => $node,
        partner                      => $partner,
        takeovertype                 => delete $opts{takeovertype},
        nacltask_verify              => 1,
        nacltask_check_partner_state => 1,
        'method-timeout'             => delete $opts{timeout},
        %opts
    );

    return 1;
}

=head2 giveback

Perform a storage failover giveback.

=over

=item Options

The following options are recognized by B<takeover>:

=over

=item C<< node => $string|$node_obj >>

(Required) The node that will perform the giveback to its partner. The value
may be either a string with the name of the node or an object of the
B<NACL::STask::Node> class.

=item C<< node_partner => $string|$node_obj >>

(Required) The partner node that will be given back control. Like C<node>,
above, may be either a string with the name of the node or an object of the
B<NACL::STask::Node> class.

=item C<< require_partner_waiting => 'true'|'false' >>

(Optional) String value, one of C<true> or C<false>. The default, if not given,
is C<true>. Any value not matching C<true> or C<false> will be changed to
C<false>. This value is passed to the B<giveback> call as the parameter
C<require-partner-waiting>.

=item C<< override_vetoes => 'true'|'false' >>

(Optional) String value, one of C<true> or C<false>. The default, if not given,
is C<true>. Any value not matching C<true> or C<false> will be changed to
C<false>. This value is passed to the B<giveback> call as the parameter
C<override-vetoes>.

=item C<< timeout => $integer >>

(Optional) The timeout value to pass when performing the giveback operation,
passed to the B<giveback> method as the C<method-timeout> parameter. Defaults
to 600 seconds.

=back

Any other arguments passed will be given directly to the B<giveback> method of
B<NACL::STask::StorageFailover> when it is called. See that manual page for
details of other parameters.

=item Exceptions

Invalid parameters will trigger a B<NATE::Exceptions::Argument> exception. Any
exceptions that may be thrown by the B<giveback> method of
B<NACL::STask::StorageFailover> will be passed through here unchanged.

=back

The return value upon success is a "true" value (C<1>, in this case).

=cut

sub giveback {
    my $self = shift;
    my %opts = validate_with(
        params      => \@_,
        allow_extra => 1,
        spec        => {
            node                    => { type => SCALAR | OBJECT },
            node_partner            => { type => SCALAR | OBJECT },
            require_partner_waiting => { type => SCALAR, default => 'true' },
            override_vetoes         => { type => SCALAR, default => 'true' },
            timeout                 => { type => SCALAR, default => 300 },
        },
    );

    require NACL::STask::StorageFailover;

    my ($cluster, $node, $partner) =
        $self->_getClusterAndNodes(delete @opts{qw(node node_partner)});

    NACL::STask::StorageFailover->giveback(
        command_interface         => $cluster,
        node                      => $node,
        partner                   => $partner,
        nacltask_verify           => 1,
        'require-partner-waiting' => delete $opts{require_partner_waiting},
        'override-vetoes'         => delete $opts{override_vetoes},
        'method-timeout'          => delete $opts{timeout},
        %opts
    );

    return 1;
}

=head2 takeoverGiveback

Provides the ability to perform a series of storage failover takeover/giveback
sequences.

=over

=item Options

The following options are recognized by B<takeoverGiveback>:

=over

=item C<< node => $string|$node_obj >>

(Required) The node that will perform the giveback to its partner. The value
may be either a string with the name of the node or an object of the
B<NACL::STask::Node> class.

=item C<< node_partner => $string|$node_obj >>

(Required) The partner node that will be given back control. Like C<node>,
above, may be either a string with the name of the node or an object of the
B<NACL::STask::Node> class.

=item C<< event_delay => $integer >>

(Optional) The length of time, in seconds, to delay between the takeover and
the giveback. Defaults to 60 if not given.

=item C<< to_args => $hashref >>

(Optional) If given, a hash-reference of additional arguments that will be
passed to the call to B<takeover>. Note that this is the call to I<this>
class's B<takeover> method, not a direct call to the method from
B<NACL::STask::StorageFailover>. See that method for definition of the
allowable parameters. It is not necessary to specify either C<node> or
C<node_partner> here, as they will be provided automatically from the options
defined above.

=item C<< gb_args => $hashref >>

(Optional) If given, a hash-reference of additional arguments that will be
passed to the call to B<giveback>. Note that this is the call to I<this>
class's B<giveback> method, not a direct call to the method from
B<NACL::STask::StorageFailover>. See that method for definition of the
allowable parameters. It is not necessary to specify either C<node> or
C<node_partner> here, as they will be provided automatically from the options
defined above.

=item C<< timeout => $integer >>

(Optional) The timeout value to pass when performing the giveback operation,
passed to the B<takeover> and B<giveback> methods as the C<method-timeout> parameter. Defaults
to 600 seconds.

=back

Unlike most event methods, B<takeoverGiveback> does not accept arbitrary
parameters (since it calls multiple other methods). Instead, extra parameters
should be given in the C<to_args> or C<gb_args> options, depending on which
method they should be given to.

=item Exceptions

Invalid parameters will trigger a B<NATE::Exceptions::Argument> exception. Any
exception thrown by either B<takeover> or B<giveback> will be passed through
here unchanged.

=back

The return value upon success is a "true" value (C<1>, in this case).

=cut

sub takeoverGiveback {
    my $self = shift;
    my %opts = validate_with(
        params      => \@_,
        allow_extra => 0,
        spec        => {
            node            => { type => SCALAR | OBJECT },
            node_partner    => { type => SCALAR | OBJECT },
            event_delay     => { type => SCALAR, default => 60 },
            to_args         => { type => HASHREF, default => {} },
            gb_args         => { type => HASHREF, default => {} },
            timeout         => { type => SCALAR, default => 600 },
        },
    );

    my $to_args     = delete $opts{to_args};
    my $gb_args     = delete $opts{gb_args};
    my $event_delay = delete $opts{event_delay};

    # Here, the passing of %opts is not because of extra, arbitrary params. It
    # is to pass along "node" and "node_partner" to each call.
    $self->takeover(%opts, %{$to_args});
    Tharn::snooze($event_delay);
    $self->giveback(%opts, %{$gb_args});

    return 1;
}

=head2 volumeClone

Create a set of volume clones. Clone names will be autogenerated.

=over

=item Options

The following options are recognized by B<volumeClone>:

=over

=item C<< node => $string|$node_obj >>

(Required) The node that will perform the volume-clone operations. The value
may be either a string with the name of the node or an object of the
B<NACL::STask::Node> class.

=item C<< vserver => $string >>

(Required) The name of the vserver that the clones are created in.

=item C<< flexclone => $string >>

(Optional) The name of the clone to create. If not given, a random name will
be generated. If this value I<is> given, and C<quantity> is greater than 1,
then each clone will have this string as its name with the sequence number
appended to it.

=item C<< parent_volume => $string|$volume_object >>

(Required) The parent volume from which to create the clones. The value may be
either a string with the name of the volume or an object of the
B<NACL::STask::Volume> class. This will become the C<parent-volume> parameter
to the clone-creation calls.

=item C<< quantity => $integer >>

(Optional) The number of clones to create. Defaults to 1 if not explicitly
given.

=back

Any other arguments passed will be given directly to the B<create> method of
B<NACL::STask::VolumeClone> when it is called. Note that any extra parameters
will be passed to I<all> calls to B<create>.

=item Exceptions

Invalid parameters will trigger a B<NATE::Exceptions::Argument> exception. Any
exceptions that may be thrown by the B<create> method of
B<NACL::STask::VolumeClone> will be passed through here unchanged.

=back

The return value from this call is an array-reference containing the
B<NACL::STask::VolumeClone> objects that were created, in the order in which
they were created.

=cut

sub volumeClone {
    my $self = shift;
    my %opts = validate_with(
        params      => \@_,
        allow_extra => 1,
        spec        => {
            node          => { type => SCALAR | OBJECT },
            vserver       => { type => SCALAR },
            parent_volume => { type => SCALAR | OBJECT },
            quantity      => { type => SCALAR, default => 1 },
        }
    );

    require NACL::STask::VolumeClone;

    my ($cluster, $node) =
        $self->_getClusterAndNodes(delete $opts{node});
    my $quantity      = delete $opts{quantity};
    my $parent_volume = delete $opts{parent_volume};
    my $name_base     = delete $opts{flexclone};
    my @vol_clones;

    if (ref $parent_volume) {
        $parent_volume = $parent_volume->volume;
    }

    for my $idx (1 .. $quantity) {
        my $flexclone;
        if ($name_base) {
            $flexclone = $name_base;
            if ($quantity > 1) {
                # Only append the seq number if we're making more than 1
                $flexclone .= "_$idx";
            }
        } else {
            $flexclone = random_name_generator;
        }

        push @vol_clones,
            NACL::STask::VolumeClone->create(
                %opts,
                command_interface => $cluster,
                'parent-volume'   => $parent_volume,
                flexclone         => $flexclone
            )
    }

    return \@vol_clones;
}

=head2 volumeFileClone

Create a set of volume file/LUN clones, then delete them after a given delay.

=over

=item Options

The following options are recognized by B<volumeClone>:

=over

=item C<< node => $string|$node_obj >>

(Required) The node that will perform the clone operations. The value may be
either a string with the name of the node or an object of the
B<NACL::STask::Node> class.

=item C<< event_delay => $integer >>

(Optional) The length of time, in seconds, to delay between the creation of
the SIS clones and their deletion. Defaults to 60 if not given.

=item C<< vserver => $string >>

(Required) The name of the vserver that the clones are created in.

=item C<< volume => $string|$vol_obj >>

(Optional) The volume that will contain the source file and the clones. May
be a name or a B<NACL::STask::Volume> object.

One of C<volume> or C<volume_list> must be given. If both are given, the
C<volume> parameter takes precedence (it is considered to be more specific).

=item C<< volume_list => $arrayref >>

(Optional) An array reference of one or more volumes that should be used for
the source file(s) and clones. Each element of the array should be a volume
name or an object.

Currently, B<volumeFileClone> can only operate on one volume at a time.
If there are two or more volumes in the list, then only the first one in the
list will be used.

One of C<volume> or C<volume_list> must be given. If both are given, the
C<volume> parameter takes precedence (it is considered to be more specific).

=item C<< volume_pattern => $string|$regex_obj >>

(Optional) If passed, must be a string or regular expression object. If given,
it is used to filter the list of volumes that will be operated on. This allows
the user to pass a more-general list of volumes while still controlling what
exactly gets operated on. Only volumes whose name matches the regex pattern
will be operated on.

Note that this option is applied even if C<volume> is used over C<volume_list>.
It is also possible to completely filter out all volumes, which will result in
an exception being thrown.

=item C<< create_LUN => $boolean >>

(Optional) A boolean value that specifies whether the source (if it does not
already exist) should be created as a LUN. If this is set and the source has
to be created, it will be a 4 megabyte LUN with a random name, created for a
Linux O/S.

=item C<< quantity => $integer >>

(Optional) An integer value specifying the number of clones to create. This is
only applied if C<destination_path> is B<not> given.

=item C<< source_path => $string >>

(Optional) The source-path of an existing file or LUN to clone. If this is not
given, then one is generated for you with a random path name.

=item C<< destination_path => $string >>

(Optional) The destination-path of the clone to create. If this is not passed,
then a random one (or more than one, if C<quanitity> is greater than 1) is
created.

=back

Any additional options passed are sent through to the B<create> method of
B<NACL::STask::VolumeFileClone> unchanged.

=item Exceptions

Invalid parameters will trigger a B<NATE::Exceptions::Argument> exception.
Calls are made to several NACL methods, each of which may throw an exception.
Any exceptions thrown are passed through unchanged.

=back

The return value of this method is 1 on success, or a thrown exception if
an error occurs.

=cut

sub volumeFileClone {
    my $self = shift;
    my %opts = validate_with(
        params      => \@_,
        allow_extra => 1,
        spec        => {
            node             => { type => SCALAR | OBJECT },
            vserver          => { type => SCALAR },
            volume           => { type => SCALAR | OBJECT, optional => 1 },
            volume_list      => { type => ARRAYREF, default => [] },
            volume_pattern   => { type => SCALAR | OBJECT, optional => 1 },
            create_LUN       => { type => SCALAR, default  => 0 },
            quantity         => {
                type    => SCALAR,
                regex   => qw/^\d+$/,
                default => 1,
            },
            event_delay      => {
                type    => SCALAR,
                regex   => qw/^\d+$/,
                default => 60,
            },
            source_path      => { type => SCALAR, optional => 1 },
            destination_path => { type => SCALAR, optional => 1 },
        },
    );

    require NACL::STask::VolumeFileClone;

    my ($cluster) = $self->_getClusterAndNodes(delete $opts{node});
    my $one_volume     = delete $opts{volume};
    my $volume_list    = delete $opts{volume_list};
    my $vserver        = delete $opts{vserver};
    my $volume_pattern = delete $opts{volume_pattern};
    my $quantity       = delete $opts{quantity};
    my $create_LUN     = delete $opts{create_LUN};
    my $source_path    = delete $opts{source_path};
    my $event_delay    = delete $opts{event_delay};

    # The singular argument "volume" takes precedence over "volume_list"
    my @volumes = $one_volume ? ($one_volume) : @{$volume_list};
    if (@volumes) {
        @volumes = $self->_filtered_volume_array(
            command_interface => $cluster,
            volume_list       => \@volumes,
            volume_pattern    => $volume_pattern,
            vserver           => $vserver
        );
        if (! @volumes) {
            NATE::Exceptions::Argument->throw(
                'No volumes were left after applying volume_pattern filter'
            );
        }
    } else {
        NATE::Exceptions::Argument->throw(
            'One of "volume" or "volume_list" must be given'
        );
    }

    # For now, we are limited to operating on just one volume, so use whatever
    # is the first element of the list. Also, we only need the name at this
    # point.
    my $volume = $volumes[0]->volume;

    my @volume_files;
    # if no source path is declared, create a zero byte source first.
    if (! $source_path) {
        my $name = random_name_generator;
        $source_path = "/vol/$volume/$name";

        my $obj;
        if ($create_LUN) {
            require NACL::STask::Lun;

            $obj = NACL::STask::Lun->create(
                command_interface => $cluster,
                vserver           => $vserver,
                volume            => $volume,
                lun               => $name,
                size              => '4m',
                ostype            => 'linux',
            );
        } else {
            require NACL::STask::VolumeFile;

            $obj = NACL::STask::VolumeFile->create(
                command_interface => $cluster,
                path              => $source_path,
                vserver           => $vserver,
                size              => '4m',
                nacltask_wait     => 1,
            );
        }

        push @volume_files, $obj;
    }

    # if no destination path is declared, randomize one.
    if (! $opts{destination_path}) {
        for (1 .. $quantity) {
            my $name = "/vol/$volume/" . random_name_generator;

            push @volume_files,
                NACL::STask::VolumeFileClone->create(
                    command_interface  => $cluster,
                    vserver            => $vserver,
                    'destination-path' => $name,
                    'source-path'      => $source_path,
                    %opts
                );
        }
    } else {
        push @volume_files,
            NACL::STask::VolumeFileClone->create(
                command_interface  => $cluster,
                vserver            => $vserver,
                'destination-path' => delete $opts{destination_path},
                'source-path'      => $source_path,
                %opts
            );
    }

    # Sleep for the requested delay:
    Tharn::snooze($event_delay);

    # Because the first element of @volume_files is the file that was cloned,
    # we should purge things in reverse order.
    for my $obj (reverse @volume_files) {
        $obj->purge;
    }

    return 1;
}

=head2 volumeCreate

Create a volume.

=over

=item Options

The following options are recognized by B<volumeCreate>:

=over

=item C<< node => $string|$node_obj >>

(Required) The node that will perform the operation. The value may be either a
string with the name of the node or an object of the B<NACL::STask::Node>
class.

=item C<< aggregate => $string|$aggr_obj >>

(Optional) The name of the aggregate to use in creating the volume, or an
existing B<NACL::STask::Aggregate> object. One of C<aggregate> or
C<candidate_aggregates> must be given.

=item C<< candidate_aggregates => $arrayref >>

(Optional) A list of one or more aggregates to consider for use in creating the
volume. Each element should be a name or an existing B<NACL::STask::Aggregate>
object. One of C<aggregate> or C<candidate_aggregates> must be given.

=item C<< vserver => $string >>

(Required) The name of the vserver that the volume is created in.

=item C<< volume => $string >>

(Optional) The name of the volume to create. If not passed, a name is
randomly generated.

=back

Any additional arguments are passed (unchanged) to the B<create> method of
B<NACL::STask::Volume>.

=item Exceptions

Invalid parameters will trigger a B<NATE::Exceptions::Argument> exception. Any
exceptions that the B<create> method of B<NACL::STask::Volume> may throw could
appear here. All exceptions are passed through unchanged.

=back

The return value is a B<NACL::STask::Volume> object representing the new
volume.

=cut

sub volumeCreate {
    my $self = shift;
    my %opts = validate_with(
        params      => \@_,
        allow_extra => 1,
        spec        => {
            node                 => { type => SCALAR | OBJECT },
            aggregate            => { type => SCALAR | OBJECT, optional => 1 },
            candidate_aggregates => { type => ARRAYREF, optional => 1 },
            vserver              => { type => SCALAR },
            volume               => { type => SCALAR, optional => 1 },
        },
    );

    my ($cluster) = $self->_getClusterAndNodes(delete $opts{node});

    # Aggregates can be given either by "aggregate" or by a list in
    # "candidate_aggregates".
    if ($opts{aggregate}) {
        # This being more specific, takes precedence over candidate_aggregates:
        delete $opts{candidate_aggregates};
    } elsif (! $opts{candidate_aggregates}) {
        # We must have at least one of them:
        NATE::Exceptions::Argument->throw(
            'One of "aggregate" or "candidate_aggregates" must be given'
        );
    }

    # If there is no name, create a random one
    $opts{volume} ||= random_name_generator;

    my $vol = NACL::STask::Volume->create(
        command_interface => $cluster,
        %opts
    );

    return $vol;
}

=head2 volumeDelete

Delete a volume.

=over

=item Options

The following options are recognized by B<volumeDelete>:

=over

=item C<< node => $string|$node_obj >>

(Required) The node that will perform the operation. The value may be either a
string with the name of the node or an object of the B<NACL::STask::Node>
class.

=item C<< vserver => $string >>

(Required) The name of the vserver that the volume exists in.

=item C<< volume => $string|$volume_obj >>

(Required) The name of the volume to delete.

=back

Any additional arguments are passed (unchanged) to the B<purge> method of
B<NACL::STask::Volume>.

=item Exceptions

Invalid parameters will trigger a B<NATE::Exceptions::Argument> exception. Any
exceptions that the B<purge> method of B<NACL::STask::Volume> may throw could
appear here. All exceptions are passed through unchanged.

=back

The return value is 1 upon success (or an exception on error).

=cut

sub volumeDelete {
    my $self = shift;
    my %opts = validate_with(
        params      => \@_,
        allow_extra => 1,
        spec        => {
            node    => { type => SCALAR | OBJECT },
            vserver => { type => SCALAR },
            volume  => { type => SCALAR | OBJECT },
        },
    );

    my ($cluster) = $self->_getClusterAndNodes(delete $opts{node});

    NACL::STask::Volume->purge(
        command_interface => $cluster,
        %opts
    );

    return 1;
}

=head2 volumeCreateDelete

Create a volume, wait a specificed number of seconds, and then delete the
volume.

=over

=item Options

The following options are recognized by B<volumeCreateDelete>:

=over

=item C<< node => $string|$node_obj >>

(Required) The node that will perform the operations. The value may be either a
string with the name of the node or an object of the B<NACL::STask::Node>
class.

=item C<< aggregate => $string|$aggr_obj >>

(Optional) The name of the aggregate to use in creating the volume, or an
existing B<NACL::STask::Aggregate> object. One of C<aggregate> or
C<candidate_aggregates> must be given, but because these options are
specific to the creation step of this event, they can be given in the
C<create_opts> parameter as an alternative.

=item C<< candidate_aggregates => $arrayref >>

(Optional) A list of one or more aggregates to consider for use in creating the
volume. Each element should be a name or an existing B<NACL::STask::Aggregate>
object. One of C<aggregate> or C<candidate_aggregates> must be given, but
because these options are specific to the creation step of this event, they can
be given in the C<create_opts> parameter as an alternative.

=item C<< vserver => $string >>

(Required) The name of the vserver that the volume is created in.

=item C<< volume => $string >>

(Optional) The name of the volume to create. If not passed, a name is
randomly generated.

=item C<< event_delay => $integer >>

(Optional) The delay in seconds between creation of the volume and the
deletion. Default is 30 seconds.

=item C<< create_opts => $hashref >>

(Optional) If passed, the contents of the hash-reference are passed to the
B<volumeCreate> method along with the general options.

=item C<< delete_opts => $hashref >>

(Optional) If passed, the contents of the hash-reference are passed to the
B<volumeDelete> method along with the general options.

=back

Any additional arguments are passed (unchanged) to both the B<volumeCreate>
and B<volumeDelete> methods. If an option should go to just one of these
methods, use the appropriate option specified above.

=item Exceptions

Invalid parameters will trigger a B<NATE::Exceptions::Argument> exception. Any
exceptions that the B<create> or B<purge> methods of B<NACL::STask::Volume> may
throw could appear here. All exceptions are passed through unchanged.

=back

The return value is 1 on success, or an exception otherwise.

=cut

sub volumeCreateDelete {
    my $self = shift;
    my %opts = validate_with(
        params      => \@_,
        allow_extra => 1,
        spec        => {
            node                 => { type => SCALAR | OBJECT },
            event_delay          => {
                type    => SCALAR,
                regex   => qw/^\d+$/,
                default => 30,
            },
            aggregate            => { type => SCALAR | OBJECT, optional => 1 },
            candidate_aggregates => { type => ARRAYREF, optional => 1 },
            vserver              => { type => SCALAR },
            volume               => { type => SCALAR, optional => 1 },
            create_opts          => { type => HASHREF, default => {} },
            delete_opts          => { type => HASHREF, default => {} },
        },
    );

    my $event_delay     = delete $opts{event_delay};
    my $create_opts     = delete $opts{create_opts};
    my $delete_opts     = delete $opts{delete_opts};

    # Aggregates can be given either by "aggregate" or by a list in
    # "candidate_aggregates". But those may be given in "create_opts" rather
    # than at this level. If given at this level, push down to "create_opts"
    # so that they don't interfere with the call to volumeDelete().
    if ($opts{aggregate} || $opts{candidate_aggregates}) {
        # However, if specified at *both* levels, the lowest-level should win
        # because it is presumed to be the most-specific.
        for my $opt (qw(aggregate candidate_aggregates)) {
            if ($opts{$opt} && ! $create_opts->{$opt}) {
                $create_opts->{$opt} = $opts{$opt};
            }
            delete $opts{$opt};
        }
    }

    my $volume = $self->volumeCreate(%opts, %{$create_opts});
    Tharn::snooze($event_delay);
    $self->volumeDelete(
        %opts,
        %{$delete_opts},
        volume => $volume->volume
    );

    return 1;
}

=head2 lunOffline

Offline a LUN.

=over

=item Options

The following options are recognized by B<lunOffline>:

=over

=item C<< node => $string|$node_obj >>

(Required) The node that will perform the operation. The value may be either a
string with the name of the node or an object of the B<NACL::STask::Node>
class.

=item C<< vserver => $string >>

(Required) The name of the vserver that the LUN exists in.

=item C<< path => $string >>

(Required) The name of the LUN to offline.

=back

No additional parameters are allowed by this method. Any additional parameters
will trigger an exception.

=item Exceptions

Invalid or unknown parameters will trigger a B<NATE::Exceptions::Argument>
exception. Any exception that can be thrown by the B<modify> method of
B<NACL::STask::Lun> will be passed through unchanged.

=back

The return value is 1 on success, an exception otherwise.

=cut

sub lunOffline {
    my $self = shift;
    my %opts = validate_with(
        params      => \@_,
        allow_extra => 0,
        spec        => {
            node    => { type => SCALAR | OBJECT },
            path    => { type => SCALAR },
            vserver => { type => SCALAR },
        },
    );

    require NACL::STask::Lun;

    my ($cluster) = $self->_getClusterAndNodes($opts{node});

    NACL::STask::Lun->modify(
        command_interface => $cluster,
        path              => $opts{path},
        vserver           => $opts{vserver},
        state             => 'offline',
    );

    return 1;
}

=head2 lunOnline

Online a LUN.

=over

=item Options

The following options are recognized by B<lunOnline>:

=over

=item C<< node => $string|$node_obj >>

(Required) The node that will perform the operation. The value may be either a
string with the name of the node or an object of the B<NACL::STask::Node>
class.

=item C<< vserver => $string >>

(Required) The name of the vserver that the LUN exists in.

=item C<< path => $string >>

(Required) The name of the LUN to online.

=back

No additional parameters are allowed by this method. Any additional parameters
will trigger an exception.

=item Exceptions

Invalid or unknown parameters will trigger a B<NATE::Exceptions::Argument>
exception. Any exception that can be thrown by the B<modify> method of
B<NACL::STask::Lun> will be passed through unchanged.

=back

The return value is 1 on success, an exception otherwise.

=cut

sub lunOnline {
    my $self = shift;
    my %opts = validate_with(
        params      => \@_,
        allow_extra => 0,
        spec        => {
            node    => { type => SCALAR | OBJECT },
            path    => { type => SCALAR },
            vserver => { type => SCALAR },
        },
    );

    require NACL::STask::Lun;

    my ($cluster) = $self->_getClusterAndNodes($opts{node});

    NACL::STask::Lun->modify(
        command_interface => $cluster,
        path              => $opts{path},
        vserver           => $opts{vserver},
        state             => 'online',
    );

    return 1;
}

=head2 lunOfflineOnline

Offline a LUN, wait a specificed number of seconds, and then online the LUN.

=over

=item Options

The following options are recognized by B<lunOfflineOnline>:

=over

=item C<< node => $string|$node_obj >>

(Required) The node that will perform the operations. The value may be either a
string with the name of the node or an object of the B<NACL::STask::Node>
class.

=item C<< vserver => $string >>

(Required) The name of the vserver that the LUN exists in.

=item C<< path => $string >>

(Required) The name of the LUN to operate on.

=item C<< event_delay => $integer >>

(Optional) The number of seconds to wait between offlining and onlining.
Defaults to 30.

=back

No additional parameters are allowed by this method. Any additional parameters
will trigger an exception.

=item Exceptions

Invalid or unknown parameters will trigger a B<NATE::Exceptions::Argument>
exception. Any exception that can be thrown by the B<modify> method of
B<NACL::STask::Lun> will be passed through unchanged.

=back

The return value is 1 on success, an exception otherwise.

=cut

sub lunOfflineOnline {
    my $self = shift;
    my %opts = validate_with(
        params      => \@_,
        allow_extra => 1,
        spec => {
            node        => { type => SCALAR | OBJECT },
            path        => { type => SCALAR },
            vserver     => { type => SCALAR },
            event_delay => {
                type    => SCALAR,
                regex   => qw/^\d+$/,
                default => 30,
            },
        },
    );

    my $event_delay = delete $opts{event_delay};

    $self->lunOffline(%opts);
    Tharn::snooze($event_delay);
    $self->lunOnline(%opts);

    return 1;
}

=head2 snapmirrorShow

Execute a snapmirror show on a simple schedule.

=over

=item Options

The following options are recognized by B<snapmirrorShow>:

=over

=item C<< node => $string|$node_obj >>

(Required) The node that will perform the operations. The value may be either a
string with the name of the node or an object of the B<NACL::STask::Node>
class.

=item C<< vserver => $string >>

(Optional) The name of the vserver to restrict the listing to. If given, this
parameter is just added to the filter. It is provided here for consistency
with other methods that utilize C<vserver> as a potential parameter.

If this parameter is given, it will override any C<vserver> key already present
in the the <Cfilter> hash reference.

=item C<< filter => $hashref >>

(Optional) The filter options to be passed to snapmirror show.

=item C<< expand => $boolean >>

(Optional) Whether to pass the C<expand> option in the call made on the node.
This can actually be done with the C<filter> option, but it is very strict
in what it accepts. Using this option allows you to pass a typical Perl value
that will be treated as a boolean and encoded properly. The default value is
false.

If this parameter is given, it will override any C<expand> key already present
in the the <Cfilter> hash reference.

=back

Additional arguments are not accepted by this method.

=item Exceptions

Invalid or unknown parameters will trigger a B<NATE::Exceptions::Argument>
exception. No other exceptions should occur.

=back

The return value is 1 on success, an exception otherwise.

=cut

sub snapmirrorShow {
    my $self = shift;
    my %opts = validate_with(
        params      => \@_,
        allow_extra => 0,
        spec => {
            node    => { type => SCALAR | OBJECT },
            vserver => { type => SCALAR,  optional => 1 },
            filter  => { type => HASHREF, optional => 1 },
            expand  => { type => BOOLEAN, optional => 1 },
        },
    );

    require NACL::CS::Snapmirror;

    my ($cluster) = $self->_getClusterAndNodes($opts{node});

    $opts{filter} ||= {};

    if ($opts{vserver}) {
        $opts{filter}->{vserver} = $opts{vserver};
    }
    if (exists $opts{expand}) {
        $opts{filter}->{expand} = $opts{expand} ? 'true' : 'false';
    }

    # For now, if the "expand" option is being passed in the call, we need
    # to force it to CLI.
    my @apiset =
        ($opts{filter}->{expand} && $opts{filter}->{expand} eq 'true') ?
        (apiset_should => { interface => 'CLI' }) : ();

    NACL::CS::Snapmirror->fetch(
        command_interface => $cluster,
        filter            => $opts{filter},
        allow_empty       => 1,
        @apiset
    );

    return 1;
}

=head2 snapmirrorShowHistory

Execute a snapmirror show history on a simple schedule.

=over

=item Options

The following options are recognized by B<snapmirrorShowHistory>:

=over

=item C<< node => $string|$node_obj >>

(Required) The node that will perform the operations. The value may be either a
string with the name of the node or an object of the B<NACL::STask::Node>
class.

=item C<< vserver => $string >>

(Optional) The name of the vserver to restrict the listing to. If given, this
parameter is just added to the filter. It is provided here for consistency
with other methods that utilize C<vserver> as a potential parameter.

=item C<< filter => $hashref >>

(Optional) The filter options to be passed to snapmirror show history.

=back

Additional arguments are not accepted by this method.

=item Exceptions

Invalid or unknown parameters will trigger a B<NATE::Exceptions::Argument>
exception. No other exceptions should occur.

=back

The return value is 1 on success, an exception otherwise.

=cut

sub snapmirrorShowHistory {
    my $self = shift;
    my %opts = validate_with(
        params      => \@_,
        allow_extra => 1,
        spec => {
            node    => { type => SCALAR | OBJECT },
            vserver => { type => SCALAR, optional => 1 },
            filter  => { type => HASHREF, default => {} },
        },
    );

    require NACL::CS::SnapmirrorShowHistory;

    my ($cluster) = $self->_getClusterAndNodes($opts{node});

    if ($opts{vserver}) {
        $opts{filter}->{vserver} = $opts{vserver};
    }

    NACL::CS::SnapmirrorShowHistory->fetch(
        command_interface => $cluster,
        filter            => $opts{filter},
        allow_empty       => 1,
    );

    return 1;
}

=head2 volumeOfflineOnline

Perform a volume offline and wait for a specified amount of time before
onlining the volume.

=over

=item Options

The following options are recognized by B<volumeOfflineOnline>:

=over

=item C<< node => $string|$node_obj >>

(Required) The node that will perform the operations. The value may be either a
string with the name of the node or an object of the B<NACL::STask::Node>
class.

=item C<< vserver => $string >>

(Required) The vserver that the volume (or volumes) are on.

=item C<< volume => $string|$volume_obj >>

(Optional) A volume to be taken offline and back online. May be a string or
a B<NACL::STask::Volume> object. If this is passed, it takes precedence over
C<volume_list>.

=item C<< volume_list => $arrayref >>

(Optional) An array-reference of volumes to offline/online. Each element may
be a string or B<NACL::STask::Volume> object. If C<volume> is passed, this
option is ignored.

=item C<< volume_pattern => $string|$regex_obj >>

(Optional) If passed, must be a string or regular expression object. If given,
it is used to filter the list of volumes that will be operated on. This allows
the user to pass a more-general list of volumes while still controlling what
exactly gets operated on. Only volumes whose name matches the regex pattern
will be operated on.

Note that this option is applied even if C<volume> is used over C<volume_list>.
It is also possible to completely filter out all volumes, resulting in none
being offlined/onlined.

=item C<< event_delay => $integer >>

(Optional) The number of seconds to wait between offlining and onlining.
Defaults to 30.

=item C<< offline_opts => $hashref >>

(Optional) If given, the contents of the hash-reference are passed as options
to the call to B<volumeOffline>. Options are not altered in any way. These
options are passed I<after> the set of additional (extra) options, so that
something specific in this set will override a general option in the other
set.

=item C<< online_opts => $hashref >>

(Optional) If given, the contents of the hash-reference are passed as options
to the call to B<volumeOnline>. Options are not altered in any way. These
options are passed I<after> the set of additional (extra) options, so that
something specific in this set will override a general option in the other
set.

=back

=item C<< is_flexgroup => $boolean >>

(Optional) If set to 1 or true, and fg_constituent_pattern is also specified
volume offline/online will be done on flexgroup constituents.
Defaults to 0. If false, and volume is a flexgroup volume
entire flexgroup will be offline and online.

=back

=item C<< fg_constituent_pattern => $string|$regex_obj >>

(Optional) If passed, must be a string or regular expression object. If given,
it is used to filter flexgroup constituent volumes that will be operated on.
Only flexgroup volumes constituents whose name matches the regex pattern
will be operated.

=back

One of C<volume> or C<volume_list> must be passed or an exception will be
thrown.

Any additional arguments are passed unchanged to both the B<volumeOffline>
and B<volumeOnline> methods. If there are parameters that should go to just
one or the other, use the appropriate option described above.

=item Exceptions

Invalid parameters will trigger a B<NATE::Exceptions::Argument> exception. Any
exceptions thrown by B<volumeOnline> or B<volumeOffline> will be passed through
unchanged.

=back

The return value is 1 on success, an exception otherwise.

=cut

sub volumeOfflineOnline {
    my $self = shift;
    my %opts = validate_with(
        params      => \@_,
        allow_extra => 0,
        spec        => {
            node           => { type => SCALAR | OBJECT },
            volume         => { type => SCALAR | OBJECT, optional => 1 },
            volume_list    => { type => ARRAYREF, default => [] },
            volume_pattern => { type => SCALAR | OBJECT, optional => 1 },
            vserver        => { type => SCALAR },
            event_delay    => {
                type    => SCALAR,
                regex   => qw/^\d+$/,
                default => 30,
            },
            offline_opts   => { type => HASHREF, default => {} },
            online_opts    => { type => HASHREF, default => {} },
            is_flexgroup     => { type => SCALAR,  optional => 1 },
            fg_constituent_pattern => { type => SCALAR | OBJECT, optional => 1 },
        },
    );

    my ($cluster, $node) = $self->_getClusterAndNodes(delete $opts{node});
    my $one_volume       = delete $opts{volume};
    my $volume_list      = delete $opts{volume_list};
    my $vserver          = delete $opts{vserver};
    my $volume_pattern   = delete $opts{volume_pattern};
    my $offline_opts     = delete $opts{offline_opts};
    my $online_opts      = delete $opts{online_opts};
    my $event_delay      = delete $opts{event_delay};
    
    # FlexGroup related params
    my $is_flexgroup     = delete $opts{is_flexgroup};
    my $fg_constituent_pattern = delete $opts{fg_constituent_pattern};
    if ($fg_constituent_pattern && (! ref $fg_constituent_pattern)) {
        $fg_constituent_pattern = qr/$fg_constituent_pattern/;
    }

    # The singular argument "volume" takes precedence over "volume_list"
    my @volumes = $one_volume ? ($one_volume) : @{$volume_list};
    if (@volumes) {
        @volumes = $self->_filtered_volume_array(
            command_interface => $cluster,
            volume_list       => \@volumes,
            volume_pattern    => $volume_pattern,
            vserver           => $vserver
        );

        # When is_flexgroup is set and fg_constituent_pattern is given 
        # Look for flexgroup constituents matching given pattern 
        if ($is_flexgroup && ($is_flexgroup !~ /false/i) 
            && $fg_constituent_pattern) {

            my @new_volumes;
            foreach my $volume (@volumes) {
                # Non-flexgroup volumes 
                if ($volume->get_one_state_attribute('is-flexgroup') !~ /true/i) {
                    push @new_volumes, $volume;
                    next;
                }
                # FlexGroup volumes
                my @fg_constituents = $volume->get_constituents();
                # include FG constituents matching given pattern
                # default all constituents
                if ($fg_constituent_pattern) {
                    @fg_constituents = grep { $_->volume =~ $fg_constituent_pattern } @fg_constituents;
                }
                push @new_volumes, @fg_constituents;
                $Log->debug("New count of volumes : " . scalar(@new_volumes));
            }
            @volumes = @new_volumes;
        }

    } else {
        NATE::Exceptions::Argument->throw(
            'One of "volume" or "volume_list" must be given'
        );
    }

    my $junction_paths = $self->volumeOffline(
        node        => $node,
        volume_list => \@volumes,
        vserver     => $vserver,
        %opts,
        %{$online_opts}
    );
    Tharn::snooze($event_delay);
    $self->volumeOnline(
        node               => $node,
        volume_list        => \@volumes,
        vserver            => $vserver,
        junction_path_list => $junction_paths,
        %opts,
        %{$offline_opts}
    );

    return 1;
}

=head2 volumeOffline

Offline a volume.

=over

=item Options

The following options are recognized by B<volumeOffline>:

=over

=item C<< node => $string|$node_obj >>

(Required) The node that will perform the operation. The value may be either a
string with the name of the node or an object of the B<NACL::STask::Node>
class.

=item C<< vserver => $string >>

(Required) The vserver that the volume (or volumes) are on.

=item C<< volume => $string|$volume_obj >>

(Optional) A volume to be taken offline. May be a string or a
B<NACL::STask::Volume> object. If this is passed, it takes precedence over
C<volume_list>.

=item C<< volume_list => $arrayref >>

(Optional) An array-reference of volumes to offline. Each element may be a
string or B<NACL::STask::Volume> object. If C<volume> is passed, this option is
ignored.

=back

One of C<volume> or C<volume_list> must be passed or an exception will be
thrown.

Any extra options will be passed through to the B<offline> method of
B<NACL::STask::Volume> unchanged.

=item Exceptions

Invalid parameters will trigger a B<NATE::Exceptions::Argument> exception. Any
exceptions thrown by B<offline> will be passed through unchanged.

=back

The return value is an array-reference of the volume(s) junction path(s). This
is so that the list may be passed to B<volumeOnline> later for moving the
volumes back to an online state. Any errors trigger exceptions.

=cut

sub volumeOffline {
    my $self = shift;
    my %opts = validate_with(
        params      => \@_,
        allow_extra => 1,
        spec        => {
            node        => { type => SCALAR | OBJECT },
            volume      => { type => SCALAR | OBJECT, optional => 1 },
            volume_list => { type => ARRAYREF, default => [] },
            vserver     => { type => SCALAR },
        }
    );

    my ($cluster)   = $self->_getClusterAndNodes(delete $opts{node});
    my $one_volume  = delete $opts{volume};
    my $volume_list = delete $opts{volume_list};
    my $vserver     = delete $opts{vserver};

    # The singular argument "volume" takes precedence over "volume_list"
    my @volumes = $one_volume ? ($one_volume) : @{$volume_list};
    if (@volumes) {
        for my $volume (@volumes) {
            next if ((blessed $volume) && $volume->isa('NACL::C::Volume'));

            $volume = NACL::STask::Volume->find(
                command_interface => $cluster,
                filter            => {
                    volume  => $volume,
                    vserver => $vserver,
                }
            );
        }
    } else {
        NATE::Exceptions::Argument->throw(
            'One of "volume" or "volume_list" must be given'
        );
    }

    my @junction_paths;
    for my $volume (@volumes) {
        my $volstate = $volume->state(
            requested_fields => [ qw(junction-active junction-path) ]
        );
        if (lc $volstate->junction_active eq 'true') {
            # If mounted, unmount and store junction-path for later
            # mounting
            push @junction_paths, $volstate->junction_path;
            $volume->unmount();
        } else {
            push @junction_paths, q{};
        }

        $volume->offline(%opts);
    }

    return \@junction_paths;
}

=head2 volumeOnline

Perform a volume online.

=over

=item Options

The following options are recognized by B<volumeOnline>:

=over

=item C<< node => $string|$node_obj >>

(Required) The node that will perform the operation. The value may be either a
string with the name of the node or an object of the B<NACL::STask::Node>
class.

=item C<< vserver => $string >>

(Required) The vserver that the volume (or volumes) are on.

=item C<< volume => $string|$volume_obj >>

(Optional) A volume to be taken offline. May be a string or a
B<NACL::STask::Volume> object. If this is passed, it takes precedence over
C<volume_list>.

=item C<< volume_list => $arrayref >>

(Optional) An array-reference of volumes to offline. Each element may be a
string or B<NACL::STask::Volume> object. If C<volume> is passed, this option is
ignored.

=item C<< junction_path => $string >>

(Optional) The junction-path to use when mounting the volume after it has
been made online. If mounting is not desired, then pass either an empty
string or C<undef> as the value for this option.

=item C<< junction_path_list => $arrayref >>

(Optional) An array-reference of junction-paths to use with volumes in the
C<volume_list> array, for the same purpose as C<junction_path>. As with
C<junction_path>, an entry that is null or C<undef> will prevent the
corresponding volume from being mounted.

=back

One of C<volume> or C<volume_list> must be passed or an exception will be
thrown. Additionally, whether C<junction_path> or C<junction_path_list> is
passed, the number of junction-paths must match the number of volumes. If
the C<volume> parameter is used to specify the volume, then C<junction_path>
will be used. If C<volume_list> is used, then C<junction_path_list> will be
used.

Any extra options will be passed through to the B<online> method of
B<NACL::STask::Volume> unchanged.

=item Exceptions

Invalid parameters will trigger a B<NATE::Exceptions::Argument> exception. Any
exceptions thrown by B<online> will be passed through unchanged.

=back

The return value is 1 on success, or an exception otherwise.

=cut

sub volumeOnline {
    my $self = shift;
    my %opts = validate_with(
        params      => \@_,
        allow_extra => 1,
        spec        => {
            node               => { type => SCALAR | OBJECT },
            volume             => { type => SCALAR | OBJECT, optional => 1 },
            volume_list        => { type => ARRAYREF, default => [] },
            vserver            => { type => SCALAR },
            junction_path      => { type => SCALAR, optional => 1 },
            junction_path_list => { type => ARRAYREF, default => [] },
        }
    );

    my ($cluster) = $self->_getClusterAndNodes(delete $opts{node});

    # The singular argument "volume" takes precedence over "volume_list". In the
    # case of putting volumes online, this also affects the consideration of the
    # "junction_path" and "junction_path_list" arguments.
    my (@volumes, @junction_paths);
    if ($opts{volume}) {
        @volumes        = (delete $opts{volume});
        @junction_paths = (delete $opts{junction_path});
    } else {
        @volumes        = @{delete $opts{volume_list}};
        @junction_paths = @{delete $opts{junction_path_list}};
    }

    if (@volumes != @junction_paths) {
        NATE::Exceptions::Argument->throw(
            'The number of volumes and number of junction-paths must match.'
        );
    }

    if (@volumes) {
        for my $volume (@volumes) {
            next if ((blessed $volume) && $volume->isa('NACL::C::Volume'));

            $volume = NACL::STask::Volume->find(
                command_interface => $cluster,
                filter            => {
                    volume  => $volume,
                    vserver => $opts{vserver},
                }
            );
        }
    } else {
        NATE::Exceptions::Argument->throw(
            'One of "volume" or "volume_list" must be given'
        );
    }

    for my $idx (0 .. $#volumes) {
        $volumes[$idx]->online(%opts);
        # A null or undef junction-path indicates that this volume wasn't
        # previously mounted.
        if ($junction_paths[$idx]) {
            $volumes[$idx]->mount('junction-path' => $junction_paths[$idx]);
        }
    }

    return 1;
}

=head2 interclusterFailure

Simulate an intercluster network failure by downing the intercluster LIFs
and then bringing them back up after a specified time interval.

=over

=item Options

The following options are recognized by B<interclusterFailure>:

=over

=item C<< node => $string|$node_obj >>

(Required) The node that will perform the operations. The value may be either a
string with the name of the node or an object of the B<NACL::STask::Node>
class.

=item C<< event_delay => $integer >>

(Optional) The number of seconds to wait between offlining and onlining LIFs.
Defaults to 30.

=back

No additional parameters are allowed by this method.

=item Exceptions

Invalid or unknown parameters will trigger a B<NATE::Exceptions::Argument>
exception. Any other exceptions thrown will be passed through unchanged.

=back

The return value is 1 on success, an exception otherwise.

=cut

sub interclusterFailure {
    my $self = shift;
    my %opts = validate_with(
        params      => \@_,
        allow_extra => 0,
        spec        => {
            node        => { type => SCALAR | OBJECT },
            timeout     => { type => SCALAR, default => 60 },
            event_delay => {
                type    => SCALAR,
                regex   => qw/^\d+$/,
                default => 30,
            },
        },
    );

    my $event_delay = $opts{event_delay};
    my ($cluster)   = $self->_getClusterAndNodes($opts{node});
    my $timeout     = delete $opts{timeout};


    my @lif_objs = NACL::STask::NetworkInterface->find(
        command_interface => $cluster,
        filter            => { role => 'intercluster' }
    );
    foreach my $lif_obj (@lif_objs) {
        $lif_obj->modify('status-admin' => 'down',
                         'method-timeout' => $timeout);
    }

    Tharn::snooze($event_delay);

    foreach my $lif_obj (@lif_objs) {
        $lif_obj->modify('status-admin' => 'up',
                         'method-timeout' => $timeout);
    }

    return 1;
}

=head2 volumeMove

Perform a volume move, wait for a specified amount of time, and move the volume
back to the original aggregate.

=over

=item Options

The following options are recognized by B<volumeMove>:

=over

=item C<< node => $string|$node_obj >>

(Required) The node that will perform the operations. The value may be either a
string with the name of the node or an object of the B<NACL::STask::Node>
class.

=item C<< vserver => $string >>

(Required) The vserver that the volume (or volumes) are on.

=item C<< volume => $string|$volume_obj >>

(Optional) A volume to be moved. May be a string or a B<NACL::STask::Volume>
object. If this is passed, it takes precedence over C<volume_list>.

=item C<< volume_list => $arrayref >>

(Optional) An array-reference of volumes to move. Each element may be a string
or B<NACL::STask::Volume> object. If C<volume> is passed, this option is
ignored.

=item C<< volume_pattern => $string|$regex_obj >>

(Optional) If passed, must be a string or regular expression object. If given,
it is used to filter the list of volumes that will be operated on. This allows
the user to pass a more-general list of volumes while still controlling what
exactly gets operated on. Only volumes whose name matches the regex pattern
will be operated on.

Note that this option is applied even if C<volume> is used over C<volume_list>.
It is also possible to completely filter out all volumes, resulting in none
being moved.

=item C<< destination_aggr => $string >>

(Optional) If given, specifies the aggregate to move the volume(s) to. If
not given, then the B<move_start> method chooses one.

=item C<< event_delay => $integer >>

(Optional) The number of seconds to wait between offlining and onlining.
Defaults to 30.

=item C<< move_opts => $hashref >>

(Optional) If given, the contents of the hash-reference are passed as options
to the initial call to B<move_start>. Options are not altered in any way. These
options are passed I<after> the set of additional (extra) options, so that
something specific in this set will override a general option in the other set.

=item C<< return_opts => $hashref >>

(Optional) If given, the contents of the hash-reference are passed as options
to the second (returning) call to B<move_start>. Options are not altered in any
way. These options are passed I<after> the set of additional (extra) options,
so that something specific in this set will override a general option in the
other set.

=item C<< is_flexgroup => $boolean >>

(Optional) If set to 1 or true, and if the volumes specified in volume_list
or volume is flexgroup, volume move operation will be done on
flexgroup constituents.  Defaults to 0.

=item C<< fg_constituent_pattern => $string|$regex_obj >>

(Optional) If passed, must be a string or regular expression object. If given,
it is used to filter flexgroup constituents volumes that will be operated on.
Only flexgroup volumes constituents whose name matches the regex pattern
will be operated.

=back

One of C<volume> or C<volume_list> must be passed or an exception will be
thrown.

Any additional arguments are passed unchanged to both B<move_start> method
calls. If there are parameters that should go to just one or the other, use the
appropriate option described above.

=item Exceptions

Invalid parameters will trigger a B<NATE::Exceptions::Argument> exception. Any
exceptions thrown by NACL will be logged, but will not stop operation.

=back

The return value is 1 on success or 0 if a NACL exception occurred during a
move. Option errors will result in an exception being thrown.

=cut

sub volumeMove {
    my $self = shift;
    my %opts = validate_with(
        params      => \@_,
        allow_extra => 1,
        spec        => {
            node             => { type => SCALAR | OBJECT },
            volume           => { type => SCALAR | OBJECT, optional => 1 },
            volume_list      => { type => ARRAYREF, default => [] },
            volume_pattern   => { type => SCALAR | OBJECT, optional => 1 },
            vserver          => { type => SCALAR },
            destination_aggr => { type => SCALAR, optional => 1 },
            event_delay      => {
                type    => SCALAR,
                regex   => qw/^\d+$/,
                default => 30,
            },
            move_opts        => { type => HASHREF, default => {} },
            return_opts      => { type => HASHREF, default => {} },
            is_flexgroup     => { type => SCALAR,  optional => 1 },
            fg_constituent_pattern => { type => SCALAR | OBJECT, optional => 1 },
        },
    );

    my ($cluster) = $self->_getClusterAndNodes(delete $opts{node});

    my $one_volume     = delete $opts{volume};
    my $volume_list    = delete $opts{volume_list};
    my $vserver        = delete $opts{vserver};
    my $volume_pattern = delete $opts{volume_pattern};
    my $event_delay    = delete $opts{event_delay};
    my $move_opts      = delete $opts{move_opts};
    my $return_opts    = delete $opts{return_opts};
    my $is_flexgroup   = delete $opts{is_flexgroup};
    my $fg_constituent_pattern = delete $opts{fg_constituent_pattern};
    if ($fg_constituent_pattern && (! ref $fg_constituent_pattern)) {
        $fg_constituent_pattern = qr/$fg_constituent_pattern/;
    }
    # The return value for the method. Set to 0 if any NACL exceptions occur.
    my $retval = 1;

    # The singular argument "volume" takes precedence over "volume_list"
    my @volumes = $one_volume ? ($one_volume) : @{$volume_list};
    if (@volumes) {
        @volumes = $self->_filtered_volume_array(
            command_interface => $cluster,
            volume_list       => \@volumes,
            volume_pattern    => $volume_pattern,
            vserver           => $vserver
        );
        if ($is_flexgroup && ($is_flexgroup !~ /false/i)) {
            my @new_volumes;
            foreach my $volume (@volumes) {
                if ($volume->get_one_state_attribute('is-flexgroup') !~ /true/i) {
                    push @new_volumes, $volume;
                    next;
                }
                # For FlexGroup volumes, start volume move on constituents
                my @fg_constituents = $volume->get_constituents();
                $Log->debug("Number of constituents " . scalar(@fg_constituents));
                # filter constituent matching given pattern, defaults all constituents
                if ($fg_constituent_pattern) {
                    @fg_constituents =
                        grep { ref $_ ? $_->volume : $_ =~ $fg_constituent_pattern }
                            @fg_constituents;
                }
                push @new_volumes, @fg_constituents;
            }
            @volumes = @new_volumes;
            $Log->debug("New list of volumes " . scalar(@volumes));
        }

    } else {
        NATE::Exceptions::Argument->throw(
            'One of "volume" or "volume_list" must be given'
        );
    }

    # destination_aggr is optional as NACL vol move doesn't need
    # destination-aggregate
    if ($opts{destination_aggr}) {
        $move_opts->{'destination-aggregate'} = delete $opts{destination_aggr};
    }

    my %original_aggrs;
    for my $vol (@volumes) {
        my $vol_name = $vol->volume;
        # Save the original aggregate:
        $original_aggrs{$vol_name} =
            $vol->get_one_state_attribute('aggregate');
        try {
            $vol->move_start(
                %opts,
                %{$move_opts},
                nacltask_wait => 1
            );
        } catch NATE::BaseException with {
            my $e = shift;
            my $class = ref $e;
            $retval = 0;

            $Log->comment(
                "Caught an exception of class $class while starting move " .
                "for volume $vol_name. Continuing."
            );
            $Log->comment("$e");
        };
    }

    Tharn::snooze($event_delay);

    for my $vol (@volumes) {
        my $vol_name      = $vol->volume;
        my $original_aggr = $original_aggrs{$vol_name};
        try {
            $vol->move_start(
                %opts,
                %{$return_opts},
                nacltask_wait           => 1,
                'destination-aggregate' => $original_aggr
            );
        } catch NATE::BaseException with {
            my $e = shift;
            my $class = ref $e;
            $retval = 0;

            $Log->comment(
                "Caught an exception of class $class while starting " .
                "move-back for volume $vol_name. Continuing."
            );
            $Log->comment("$e");
        };
    }

    return $retval;
}

=head2 reboot

Perform a node reboot.

=over

=item Options

The following options are recognized by B<reboot>:

=over

=item C<< node => $string|$node_obj >>

(Required) The node that will be rebooted. The value may be either a string
with the name of the node or an object of the B<NACL::STask::Node> class.

=item C<< timeout => $integer >>

(Optional) A timeout value for the reboot operation. Passed to the method as
the parameter C<method-timeout>. No default value.

=back

No additional parameters are allowed by this method.

=item Exceptions

Invalid or unknown parameters will trigger a B<NATE::Exceptions::Argument>
exception. Any exceptions thrown by NACL will be passed back unchanged.

=back

The return value is 1 on success, an exception otherwise.

=cut

sub reboot {
    my $self = shift;
    my %opts = validate_with(
        params      => \@_,
        allow_extra => 0,
        spec        => {
            node    => { type => SCALAR | OBJECT },
            timeout => { type => SCALAR, optional => 1 },
        },
    );

    # Note that this isn't using _getClusterAndNodes(), because we explicitly
    # use the node interface rather than the cluster for this.
    my $node = delete $opts{node};
    if (! ref($node)) {
        $node = NACL::STask::Node->new(name => $node);
    }

    # If they asked for a timeout, rename it to method-timeout
    if ($opts{timeout}) {
        $opts{'method-timeout'} = delete $opts{timeout};
    }

    $node->reboot(%opts);

    return 1;
}

=head2 aggregateRelocate

Perform an aggregrate relocation. Verifies that the aggregate relocation has
successfully completed.

=over

=item Options

The following options are recognized by B<volumeMove>:

=over

=item C<< node => $string|$node_obj >>

(Required) The node that the aggregates are currently on. The value may be
either a string with the name of the node or an object of the
B<NACL::STask::Node> class.

=item C<< node_dest => $string|$node_obj >>

(Required) The node to which the aggegates will be relocated. The value may be
either a string with the name of the node or an object of the
B<NACL::STask::Node> class.

=item C<< aggregate_list => $string|$arrayref >>

(Required) Either a single aggregate name, or an array-reference of aggregate
names. Each aggregate specified will be relocated.

=item C<< event_delay => $integer >>

(Optional) The delay to wait between initially relocating the aggregates, and
moving them back. Defaults to 60 seconds.

=back

Any extra options are passed through unchanged to both calls to the B<start>
method of B<NACL::STask::StorageAggregateRelocation>.

=item Exceptions

Invalid parameters will trigger a B<NATE::Exceptions::Argument> exception. Any
exceptions thrown by NACL will be passed back unchanged.

=back

The return value is 1 on success, an exception otherwise.

=cut

sub aggregateRelocate {
    my $self = shift;
    my %opts = validate_with(
        params      => \@_,
        allow_extra => 1,
        spec        => {
            node           => { type => SCALAR | OBJECT },
            node_dest      => { type => SCALAR | OBJECT },
            aggregate_list => { type => SCALAR | ARRAYREF },
            event_delay    => {
                type    => SCALAR,
                regex   => qw/^\d+$/,
                default => 60,
            },
        },
    );

    require NACL::STask::StorageAggregateRelocation;

    my $iteration_delay = delete $opts{iteration_delay};
    my $event_delay     = delete $opts{event_delay};
    my $iterations      = delete $opts{iterations};
    my $aggregate_list  = delete $opts{aggregate_list};

    # If they only passed one aggregrate, make it a list:
    if (! ref $aggregate_list) {
        $aggregate_list = [ $aggregate_list ];
    }

    my ($cluster, $node, $dest) =
        $self->_getClusterAndNodes(delete @opts{qw(node node_dest)});

    NACL::STask::StorageAggregateRelocation->start(
        command_interface => $cluster,
        node              => $node->node,
        destination       => $dest->node,
        'aggregate-list'  => $aggregate_list,
        nacltask_wait     => 1,
        %opts
    );
    Tharn::snooze($event_delay);
    NACL::STask::StorageAggregateRelocation->start(
        command_interface => $cluster,
        node              => $dest->node,
        destination       => $node->node,
        'aggregate-list'  => $aggregate_list,
        nacltask_wait     => 1,
        %opts
    );

    return 1;
}

=head2 snapshotCreateDelete

Create a snapshot, wait a specificed number of seconds, and then delete the
snapshot.

=over

=item Options

The following options are recognized by B<snapshotCreateDelete>:

=over

=item C<< node => $string|$node_obj >>

(Required) The node that the aggregates are currently on. The value may be
either a string with the name of the node or an object of the
B<NACL::STask::Node> class.

=item C<< vserver => $string >>

(Required) The name of the vserver that has access to the target volume that
the snapshot will be created on.

=item C<< volume => $string|$volume_obj >>

=item C<< volume => $string|$vol_obj >>

(Optional) The volume that the snapshot will be created on. May be the
volume name or an object of the B<NACL::STask::Volume> class.

One of C<volume> or C<volume_list> must be given. If both are given, the
C<volume> parameter takes precedence (it is considered to be more specific).

=item C<< volume_list => $arrayref >>

(Optional) An array reference of one or more volumes that should be used for
creating the snapshots. Each element of the array should be a volume name or an
object. Each volume will have a snapshot created on it.

One of C<volume> or C<volume_list> must be given. If both are given, the
C<volume> parameter takes precedence (it is considered to be more specific).

=item C<< volume_pattern => $string|$regex_obj >>

(Optional) If passed, must be a string or regular expression object. If given,
it is used to filter the list of volumes that will be operated on. This allows
the user to pass a more-general list of volumes while still controlling what
exactly gets operated on. Only volumes whose name matches the regex pattern
will be operated on.

Note that this option is applied even if C<volume> is used over C<volume_list>.
It is also possible to completely filter out all volumes, which will result in
an exception being thrown.

=item C<< snapshot => $string >>

(Optional) The name of the snapshot that will be created. If not specified, a
random name will be chosen.

=item C<< event_delay => $integer >>

(Optional) The delay to wait between creating the snapshot and deleting it.
Defaults to 30 seconds.

=back

Any extra options are passed through unchanged to snapshot creation call.
Extra options are not passed to the deletion call.

=item Exceptions

Invalid parameters will trigger a B<NATE::Exceptions::Argument> exception. Any
exceptions thrown by NACL will be passed back unchanged.

=back

The return value is 1 on success, an exception otherwise.

=cut

sub snapshotCreateDelete {
    my $self = shift;
    my %opts = validate_with(
        params      => \@_,
        allow_extra => 1,
        spec        => {
            node           => { type => SCALAR | OBJECT },
            vserver        => { type => SCALAR },
            volume         => { type => SCALAR | OBJECT, optional => 1 },
            volume_list    => { type => ARRAYREF, default => [] },
            volume_pattern => { type => SCALAR | OBJECT, optional => 1 },
            snapshot       => { type => SCALAR, optional => 1 },
            event_delay    => {
                type    => SCALAR,
                regex   => qw/^\d+$/,
                default => 30,
            },
        },
    );

    my $event_delay    = delete $opts{event_delay};
    my $one_volume     = delete $opts{volume};
    my $volume_list    = delete $opts{volume_list};
    my $vserver        = delete $opts{vserver};
    my $volume_pattern = delete $opts{volume_pattern};
    my $snap_name      = delete $opts{snapshot};

    my ($cluster) = $self->_getClusterAndNodes(delete $opts{node});

    # The singular argument "volume" takes precedence over "volume_list"
    my @volumes = $one_volume ? ($one_volume) : @{$volume_list};
    if (@volumes) {
        @volumes = $self->_filtered_volume_array(
            command_interface => $cluster,
            volume_list       => \@volumes,
            volume_pattern    => $volume_pattern,
            vserver           => $vserver
        );
        if (! @volumes) {
            NATE::Exceptions::Argument->throw(
                'No volumes were left after applying volume_pattern filter'
            );
        }
    } else {
        NATE::Exceptions::Argument->throw(
            'One of "volume" or "volume_list" must be given'
        );
    }

    my @snapshot_list;
    for my $vol (@volumes) {
        my $snapshot_name = $snap_name || random_name_generator;
        my $snapshot = $vol->snapshot_create(
            snapshot => $snapshot_name,
            %opts
        );
        push @snapshot_list, $snapshot;
    }

    Tharn::snooze($event_delay);

    for my $snapshot (@snapshot_list) {
        # Retry as FlexGroup snapshots get fenced 
        for (1..10) {
            try {
                $snapshot->purge;
                last;
            } catch NACL::C::Exceptions::VolumeSnapshot::SnapshotFenced with {
                my $e = shift;
                my $class = ref $e;
                $Log->warn(
                    "$class exception while deleting $snapshot," .
                    "will wait for $event_delay and retry."
                );
                $Log->comment("$e");
                Tharn::snooze($event_delay);
            };
        }

    }

    return 1;
}

=head2 sfcod

Perform a single file copy on demand (SFCOD).

=over

=item Options

The following options are recognized by B<sfcod>:

=over

=item C<< node => $string|$node_obj >>

(Required) The node that will perform the operations. The value may be either a
string with the name of the node or an object of the B<NACL::STask::Node>
class.

=item C<< quantity => $integer >>

(Optional) An integer specifying how many files to create before the copy
operation is performed. If given, then the C<source_vserver>,
C<source_volume>, C<source_volume_list>, C<source_volume_pattern>,
C<destination_vserver>, C<destination_volume>, C<destination_volume_list> and
C<destination_volume_pattern> will be used to create the paths for the
requested number of files (on both the source side and destination side). If
there is not enough information to generate the path names, an exception
will be thrown.

=item C<< source_vserver => $string >>

(Optional) A string with the name of the source vserver to use for creating
and specifying source files.

=item C<< source_volume => $string|$volume_obj >>

(Optional) The volume to use for creating and specifying source files. May be
either the name or a B<NACL::STask::Volume> object.

=item C<< source_volume_list => $arrayref >>

(Optional) An array reference of one or more volumes that should be used for
creating the files. Each element of the array should be a volume name or an
object. Currently, only one volume may be used for B<sfcod> at a time. In the
case where there are multiple volumes, the first element of the list will be
used.

One of C<source_volume> or C<source_volume_list> must be given. If both are
given, the C<source_volume> parameter takes precedence (it is considered to be
more specific).

=item C<< source_volume_pattern => $string|$regex_obj >>

(Optional) If passed, must be a string or regular expression object. If given,
it is used to filter the list of source volumes that will be operated on. This
allows the user to pass a more-general list of volumes while still controlling
what exactly gets operated on. Only volumes whose name matches the regex
pattern will be operated on.

Note that this option is applied even if C<source_volume> is used over
C<source_volume_list>. It is also possible to completely filter out all source
volumes, which will result in an exception being thrown.

=item C<< destination_vserver => $string >>

(Optional) A string with the name of the destination vserver to use for
creating and specifying target files.

=item C<< destination_volume => $string|$volume_obj >>

(Optional) The volume to use for creating and specifying destination files. May
be either the name or a B<NACL::STask::Volume> object.

=item C<< destination_volume_list => $arrayref >>

(Optional) An array reference of one or more volumes that should be used for
the destination files. Each element of the array should be a volume name or an
object. Currently, only one volume may be used for B<sfcod> at a time. In the
case where there are multiple volumes, the first element of the list will be
used.

One of C<destination_volume> or C<destination_volume_list> must be given. If
both are given, the C<destination_volume> parameter takes precedence (it is
considered to be more specific).

=item C<< destination_volume_pattern => $string|$regex_obj >>

(Optional) If passed, must be a string or regular expression object. If given,
it is used to filter the list of volumes that will be operated on. This allows
the user to pass a more-general list of volumes while still controlling what
exactly gets operated on. Only volumes whose name matches the regex pattern
will be operated on.

Note that this option is applied even if C<destination_volume> is used over
C<destination_volume_list>. It is also possible to completely filter out all
destination volumes, which will result in an exception being thrown.

=item C<< size => $integer >>

(Optional) If creating the files, this value specifies the size to create the
files as. Defaults to 4096.

=item C<< source_paths => $arrayref >>

(Optional) An array-reference of source-paths to use in the copy operation. If
this is given, it supercedes C<quantity> (and the other related parameters).

=item C<< destination_paths => $arrayref >>

(Optional) An array-reference of destination-paths to use in the copy
operation. If this is given, it supercedes C<quantity> et al.

=back

Any additional options are passed through unchanged to the B<start> method
of B<NACL::STask::VolumeFileCopy>.

Note that if either of C<source_paths> or C<destination_paths> are given, then
both must be passed, and both must contain the same number of elements. If
they differ in number, an exception will be thrown.

If C<source_paths> or C<destination_paths> are given, then the C<quantity>,
C<source_vserver>, C<source_volume>, C<source_volume_list>,
C<source_volume_pattern>, C<destination_vserver>, C<destination_volume>,
C<destination_volume_list> and C<destination_volume_pattern> parameters are
all ignored.

=item Exceptions

Invalid parameters will trigger a B<NATE::Exceptions::Argument> exception. Any
exceptions thrown by NACL will be passed back unchanged.

=back

The return value is 1 on success or an exception otherwise.

=cut

sub sfcod {
    my $self = shift;

    # The "event" attribute on $self will identify this as a copy operation:
    return $self->_sfXod(@_);
}

=head2 sfmod

Perform a single file move on demand (SFMOD).

=over

=item Options

The following options are recognized by B<sfmod>:

=over

=item C<< node => $string|$node_obj >>

(Required) The node that will perform the operations. The value may be either a
string with the name of the node or an object of the B<NACL::STask::Node>
class.

=item C<< quantity => $integer >>

(Optional) An integer specifying how many files to create before the move
operation is performed. If given, then the C<source_vserver>,
C<source_volume>, C<source_volume_list>, C<source_volume_pattern>,
C<destination_vserver>, C<destination_volume>, C<destination_volume_list> and
C<destination_volume_pattern> will be used to create the paths for the
requested number of files (on both the source side and destination side). If
there is not enough information to generate the path names, an exception
will be thrown.

=item C<< source_vserver => $string >>

(Optional) A string with the name of the source vserver to use for creating
and specifying source files.

=item C<< source_volume => $string|$volume_obj >>

(Optional) The volume to use for creating and specifying source files. May be
either the name or a B<NACL::STask::Volume> object.

=item C<< source_volume_list => $arrayref >>

(Optional) An array reference of one or more volumes that should be used for
creating the files. Each element of the array should be a volume name or an
object. Currently, only one volume may be used for B<sfmod> at a time. In the
case where there are multiple volumes, the first element of the list will be
used.

One of C<source_volume> or C<source_volume_list> must be given. If both are
given, the C<source_volume> parameter takes precedence (it is considered to be
more specific).

=item C<< source_volume_pattern => $string|$regex_obj >>

(Optional) If passed, must be a string or regular expression object. If given,
it is used to filter the list of source volumes that will be operated on. This
allows the user to pass a more-general list of volumes while still controlling
what exactly gets operated on. Only volumes whose name matches the regex
pattern will be operated on.

Note that this option is applied even if C<source_volume> is used over
C<source_volume_list>. It is also possible to completely filter out all source
volumes, which will result in an exception being thrown.

=item C<< destination_vserver => $string >>

(Optional) A string with the name of the destination vserver to use for
creating and specifying target files.

=item C<< destination_volume => $string|$volume_obj >>

(Optional) The volume to use for creating and specifying destination files. May
be either the name or a B<NACL::STask::Volume> object.

=item C<< destination_volume_list => $arrayref >>

(Optional) An array reference of one or more volumes that should be used for
the destination files. Each element of the array should be a volume name or an
object. Currently, only one volume may be used for B<sfmod> at a time. In the
case where there are multiple volumes, the first element of the list will be
used.

One of C<destination_volume> or C<destination_volume_list> must be given. If
both are given, the C<destination_volume> parameter takes precedence (it is
considered to be more specific).

=item C<< destination_volume_pattern => $string|$regex_obj >>

(Optional) If passed, must be a string or regular expression object. If given,
it is used to filter the list of volumes that will be operated on. This allows
the user to pass a more-general list of volumes while still controlling what
exactly gets operated on. Only volumes whose name matches the regex pattern
will be operated on.

Note that this option is applied even if C<destination_volume> is used over
C<destination_volume_list>. It is also possible to completely filter out all
destination volumes, which will result in an exception being thrown.

=item C<< size => $integer >>

(Optional) If creating the files, this value specifies the size to create the
files as. Defaults to 4096.

=item C<< source_paths => $arrayref >>

(Optional) An array-reference of source-paths to use in the move operation. If
this is given, it supercedes C<quantity> et al.

=item C<< destination_paths => $arrayref >>

(Optional) An array-reference of destination-paths to use in the move
operation. If this is given, it supercedes C<quantity> et al.

=back

Any additional options are passed through unchanged to the B<start> method
of B<NACL::STask::VolumeFileMove>.

Note that if either of C<source_paths> or C<destination_paths> are given, then
both must be passed, and both must contain the same number of elements. If
they differ in number, an exception will be thrown.

If C<source_paths> or C<destination_paths> are given, then the C<quantity>,
C<source_vserver>, C<source_volume>, C<source_volume_list>,
C<source_volume_pattern>, C<destination_vserver>, C<destination_volume>,
C<destination_volume_list> and C<destination_volume_pattern> parameters are
all ignored.

=item Exceptions

Invalid parameters will trigger a B<NATE::Exceptions::Argument> exception. Any
exceptions thrown by NACL will be passed back unchanged.

=back

The return value is 1 on success or an exception otherwise.

=cut

sub sfmod {
    my $self = shift;

    # The "event" attribute on $self will identify this as a move operation:
    return $self->_sfXod(@_);
}

=head2 portToggle

Simulate a port failure by bringing the port down administratively and then
bringing it back up after a specified time interval.

=over

=item Options

The following options are recognized by B<portToggle>:

=over

=item C<< node => $string|$node_obj >>

(Required) The node that the ports are on. The value may be either a string
with the name of the node or an object of the B<NACL::STask::Node> class.

=item C<< ports => $arrayref >>

(Required) An array reference of the individual ports to be toggled.

=item C<< event_delay => $integer >>

(Optional) The number of seconds to wait after downing the ports before
bringing them back up. Defaults to 30.

=back

No additional parameters are recognized by this method.

=item Exceptions

Invalid or unkown parameters will trigger a B<NATE::Exceptions::Argument>
exception. Any exceptions thrown by NACL will be passed back unchanged.

=back

The return value is 1 on success, an exception otherwise.

=cut

sub portToggle {
    my $self = shift;
    my %opts = validate_with(
        params      => \@_,
        allow_extra => 0,
        spec        => {
            node        => { type => SCALAR | OBJECT },
            ports       => { type => SCALAR | ARRAYREF },
            event_delay => {
                type    => SCALAR,
                regex   => qw/^\d+$/,
                default => 30,
            },
        },
    );

    require NACL::C::NetworkPort;

    my $event_delay     = delete $opts{event_delay};
    my $iteration_delay = delete $opts{iteration_delay};
    my $iterations      = delete $opts{iterations};
    my $ports           = delete $opts{ports};

    my @ports = (ref $ports) ? @{$ports} : ($ports);

    my ($cluster, $node) = $self->_getClusterAndNodes(delete $opts{node});

    foreach my $port (@ports) {
        NACL::C::NetworkPort->modify(
            command_interface => $cluster,
            node              => $node,
            port              => $port,
            'up-admin'        => 'false'
        );
    }

    Tharn::snooze($event_delay);

    foreach my $port (@ports) {
        NACL::C::NetworkPort->modify(
            command_interface => $cluster,
            node              => $node,
            port              => $port,
            'up-admin'        => 'true'
        );
    }

    return 1;
}

=head1 EVENT ITERATION LOOP

(This is a summary of the section by the same name in the documentation for
B<NACL::AppBuilder::DpgWorkflows::EventWorker>.)

When the C<start> method of B<NACL::AppBuilder::DpgWorkflows::EventWorker> is
called, the worker instance sends code to the chosen client machine via STAF.
This code contains the processing loop that will be used to invoke the chosen
remote event. This processing loop is called the I<event iteration loop>.

The event iteration loop utilizes the C<initial_delay>, C<iterations>,
C<iteration_delay> and C<runlength> attributes of the worker instance. These
fulfill the following roles:

=over

=item B<initial_delay>

Before the first iteration of the event (and only the first), if this value is
non-zero the worker will wait that many seconds.

=item B<iterations>

If this value is not zero, then it dictates the number of iterations the event
will run before the remote process stops execution and signals the worker
instance that it is completed. If C<runlength> is set, this will always be
zero.

=item B<runlength>

If this value is not zero, then it dictates the length of time for which the
event will run in a continuous loop. If this value is non-zero, then
B<iterations> will always be zero.

Note that if an initial delay is set with C<initial_delay>, this counts as part
of the overall run-time; something configured to run for 5 minutes with a 2
minute initial delay would have an effective available run-time of only 3
minutes.

=item B<iteration_delay>

If this value is not zero, then each iteration of the event will wait for this
many seconds before starting the next iteration. This delay will not be applied
on the last iteration.

If the run is governed by C<runlength> rather than by C<iterations>, then the
delay is applied after every iteration I<until> an iteration takes the process
past the end-time dictated by C<runlength>. That will be the last iteration,
and no additional delay is applied.

=back

In general, individual events should not need to care about the iteration loop
as it is managed and controlled at a higher level.

=head1 SEE ALSO

B<NACL::AppBuilder::DpgWorkflows::EventBase>,
B<NACL::AppBuilder::DpgWorkflows::EventManager>,
B<NACL::AppBuilder::DpgWorkflows::EventWorker>

=head1 AUTHOR

Randy J. Ray E<lt>rjray@netapp.comE<gt>

=cut

###############################################################################
# Private methods, not documented to the user.
###############################################################################

sub _sfXod {
    my $self = shift;
    my %opts = validate_with(
        params      => \@_,
        allow_extra => 1,
        spec => {
            node                       => { type => SCALAR | OBJECT },
            source_vserver             => { type => SCALAR, optional => 1, },
            source_volume              => {
                type     => SCALAR | OBJECT,
                optional => 1,
            },
            source_volume_list         => { type => ARRAYREF, default => [] },
            source_volume_pattern      => {
                type     => SCALAR | OBJECT,
                optional => 1,
            },
            destination_vserver        => { type => SCALAR, optional => 1, },
            destination_volume         => {
                type     => SCALAR | OBJECT,
                optional => 1,
            },
            destination_volume_list    => { type => ARRAYREF, default => [] },
            destination_volume_pattern => {
                type     => SCALAR | OBJECT,
                optional => 1,
            },
            destination_volume         => {
                type     => SCALAR | OBJECT,
                optional => 1,
            },
            size                       => { type => SCALAR, default => 4096 },
            source_paths               => { type => ARRAYREF, default => [] },
            destination_paths          => { type => ARRAYREF, default => [] },
            quantity                   => { type => SCALAR, optional => 1 },
        },
    );

    my %action_classes = (
        sfcod => 'NACL::STask::VolumeFileCopy',
        sfmod => 'NACL::STask::VolumeFileMove',
    );
    my $actionclass = $action_classes{$self->event};
    if ($actionclass) {
        load $actionclass;
    } else {
        NATE::Exceptions::Argument->throw(
            'Inside _sfXod: Unknown front-end event "' . $self->event .
            '". Please configure a corresponding class for this event.'
        );
    }

    require NACL::STask::VolumeFile;

    my ($cluster) = $self->_getClusterAndNodes(delete $opts{node});

    # These parameters are all for the creation of files if nothing is given
    # in the source_paths/destination_paths params:
    my ($quantity,
        $source_vserver, $source_volume,
        $source_volume_list, $source_volume_pattern,
        $destination_vserver, $destination_volume,
        $destination_volume_list, $destination_volume_pattern,
        $size) =
            delete @opts{qw(quantity
                            source_vserver source_volume
                            source_volume_list source_volume_pattern
                            destination_vserver destination_volume
                            destination_volume_list destination_volume_pattern
                            size)};
    # These will take precedence over the previous set of params, when present:
    my @source_paths      = @{delete $opts{source_paths}};
    my @destination_paths = @{delete $opts{destination_paths}};

    # Either both of these must be set, or neither:
    if (@source_paths && @destination_paths) {
        if (@source_paths != @destination_paths) {
            NATE::Exceptions::Argument->throw(
                'The number of elements given in source_paths and ' .
                'destination_paths must be equal.'
            );
        }
    } elsif ((! @source_paths) && (! @destination_paths)) {
        # Check that $quantity is given and non-zero
        if (! $quantity) {
            NATE::Exceptions::Argument->throw(
                'If the source_paths and destination_paths parameters are ' .
                'not given, then quantity (and the supporting parameters) ' .
                'must be.'
            );
        }

        # Derive a value for $source_volume based on what was given by the
        # caller. The singular argument "source_volume" takes precedence over
        # "source_volume_list".
        my @volumes =
            $source_volume ? ($source_volume) : @{$source_volume_list};
        if (@volumes) {
            @volumes = $self->_filtered_volume_array(
                command_interface => $cluster,
                volume_list       => \@volumes,
                volume_pattern    => $source_volume_pattern,
                vserver           => $source_vserver
            );
            if (! @volumes) {
                NATE::Exceptions::Argument->throw(
                    'No volumes were left after applying ' .
                    'source_volume_pattern filter'
                );
            }
        } else {
            NATE::Exceptions::Argument->throw(
                'One of source_volume or source_volume_list must be given'
            );
        }

        # For now, we can only operate on one volume at a time, so just use
        # the first volume in the now-filtered list. Also, at this point we
        # only need the name.
        $source_volume = $volumes[0]->volume;

        # Derive a value for $destination_volume based on what was given by the
        # caller. The singular argument "destination_volume" takes precedence
        # over "destination_volume_list".
        @volumes = $destination_volume ?
            ($destination_volume) : @{$destination_volume_list};
        if (@volumes) {
            @volumes = $self->_filtered_volume_array(
                command_interface => $cluster,
                volume_list       => \@volumes,
                volume_pattern    => $destination_volume_pattern,
                vserver           => $destination_vserver
            );
            if (! @volumes) {
                NATE::Exceptions::Argument->throw(
                    'No volumes were left after applying ' .
                    'destination_volume_pattern filter'
                );
            }
        } else {
            NATE::Exceptions::Argument->throw(
                'One of destination_volume or destination_volume_list ' .
                'must be given'
            );
        }

        # For now, we can only operate on one volume at a time, so just use
        # the first volume in the now-filtered list. Also, at this point we
        # only need the name.
        $destination_volume = $volumes[0]->volume;

        for (1 .. $quantity) {
            my $filename = random_name_generator;
            NACL::STask::VolumeFile->create(
                command_interface => $cluster,
                path              => "/vol/$source_volume/$filename",
                vserver           => $source_vserver,
                nacltask_wait     => 1,
                size              => $size,
            );
            push @source_paths, "$source_vserver:$source_volume/$filename";

            $filename = random_name_generator;
            push @destination_paths,
                "$destination_vserver:$destination_volume/$filename";
        }
    } else {
        # One is set, but the other is not. We can't work with that yet.
        NATE::Exceptions::Argument->throw(
            'Either both of source_paths and destination_paths must be ' .
            'given, or neither can be given.'
        );
    }

    $actionclass->start(
        command_interface   => $cluster,
        'source-paths'      => \@source_paths,
        'destination-paths' => \@destination_paths,
        'report-completed'  => 'true',
        nacltask_wait       => 1,
        %opts
    );

    return 1;
}

1;
