# Copyright (c) 2015 NetApp, Inc., All Rights Reserved.
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#
## @summary Base Moose role for event classes to utilize
## @author rjray@netapp.com
## @status shared
## @pod here

package NACL::AppBuilder::DpgWorkflows::EventBase;

use 5.014;
use strict;
use warnings;
use namespace::autoclean;

use Const::Fast;
use Moose::Role;
use Params::Validate qw(validate validation_options
                        OBJECT SCALAR ARRAYREF UNDEF);
use Scalar::Util qw(blessed);

use NATE::Log qw(log_global);
use NACL::STask::Node;
use NACL::STask::Cluster;
use NACL::STask::Volume;

# Used to track/cache node objects and cluster objects, by name
my %NODE_MAP;
my %CLUS_MAP;

# client: The name of the client this process is executing on.
# event: The name of the event being invoked.
# interface: One of 'CLI' or 'ZAPI', used in creating node/cluster objects.
has [ qw(client event interface) ] => (
    is       => 'ro',
    isa      => 'Str',
    required => 1,
);

# iteration: The current iteration being done. Set by EventWorker before the
#            call is made.
has 'iteration' => (
    is       => 'rw',
    isa      => 'Int',
    init_arg => undef,
);

# max_iterations: The maximum number of iterations planned. Will be zero if
#                 "runlength" is set.
# runlength: The total number of seconds (over all iterations) that the event
#            will be allowed to run. If set to zero, then the event is being
#            governed by number of iterations, instead.
# start_time: The UNIX time value of when the processing of this worker/event
#             started.
has [ qw(max_iterations runlength start_time) ] => (
    is  => 'ro',
    isa => 'Int'
);

# state: An open-ended hashref made available to each instance to allow data
#        to be retained across iterations.
has state => (
    is       => 'ro',
    isa      => 'HashRef',
    default  => sub { {} },
    init_arg => undef,
);

# Global handle on the NATE::Log logging interface
const my $Log => log_global;

###############################################################################
# NOTE: Because I am using Moose for the object model, I have moved the usage
# of any NATE or NACL exception classes down here, after an explicit
# "no Moose;" statement. Do not try to move this higher in the code or
# remove the "no" statement, or compilation of the module may break.
###############################################################################

no Moose::Role;
use NATE::BaseException qw(:try);
use NATE::Exceptions::Argument;

# Set the on_fail behavior for Params::Validate to throw an exception (of the
# NATE::Exceptions::Argument class) on errors instead of a plain croak().
validation_options(
    on_fail => sub {
        NATE::Exceptions::Argument->throw(shift);
    }
);

=head1 NAME

NACL::AppBuilder::DpgWorkflows::EventBase

=head1 SYNOPSIS

    package NewEvents;
    
    use 5.014;
    use strict;
    use warnings;
    use namespace::autoclean;
    
    use Moose;
    
    with 'NACL::AppBuilder::DpgWorkflows::EventBase';
    
    __PACKAGE__->meta->make_immutable;
    
    1;

=head1 DESCRIPTION

This is a Moose role that B<NACL::AppBuilder::DpgWorkflows::EventWorker>
expects any event class to be using. It provides the common attributes and
elements that the worker instances use to communicate data down to the remote
events being executed. It also provides a number of common semi-private methods
that consumers of the role can utilize.

=head1 ATTRIBUTES

The following attributes are defined for consumers of this role. Each
attribute value is obtained by calling the attribute name as an accessor. All
but one of the attributes are read-only (they cannot be changed after the
object is created).

=over

=item client

The name of the client that this event is executing on. This may or may not
be the fully-qualified domain name (FQDN) of the machine, depending on how it
was specified to either the EventManager or EventWorker.

=item interface

A string, one of C<CLI> or C<ZAPI>, that defines the interface method that
this event will use for any node or cluster connections that it makes to
ONTAP.

=item event

The name of the event that is being executed by the object handle. This is set
because an event may call other methods in the class as part of it's
processing. However, this value will never change. This means that any method
may look at this attribute to see what actual event is being processed, even if
that event is different than the currently-executing method.

=item iteration

The number of the current iteration, starting at 1. Unlike the rest of the
attributes, this attribute is writable. However, the code that controls the
event object will set the value each time through (so any changes made by the
event itself will be lost).

=item max_iterations

The number of iterations of the event that will be done before the worker
considers itself complete and exits. If this value is zero, then the event
is being run for a specified period of time rather than a specific number of
iterations. See C<runlength>, next.

=item runlength

If set, the number of seconds that the iteration loop will run. If set, then
the C<max_iterations> value will always be zero. Note that this does not mean
that the event will be interrupted at this precise number of seconds; if the
C<runlength> is reached during the execution of an iteration, then that
iteration will finish before the worker completes on the client.

=item start_time

The start-time (in UNIX seconds format) when the remote code started performing
the event's iteration-loop.

=item state

A hash-reference made available to all event methods in order to pass or
preserve information between iterations of the event. Note that the attribute
itself is read-only in that the reference cannot be overwritten or otherwise
changed. However, you can add, delete and modify keys on the hash-reference
itself without any conditions.

=back

=head1 METHODS

In addition to the attribute accessors, the role provides the following
methods. No subroutines are exported by this package. Note that all but one of
the methods here are named so as to be considered "private". This prevents them
from being accidentally called as events. Only the first method is considered
"public", as it is part of the public interface this role provides.

=head2 run

    $return_val = $self->run(%opts);

Runs the event-method that was specified when this object was created. Returns
the value that the event itself returns, or allows through any exceptions the
event throws. This is the preferred means of running the event, as it includes
some logging, etc. that may be used in deeper analysis.

All parameters passed to this method are passed through unchanged to the
event that gets invoked.

=cut

sub run {
    my ($self, %opts) = @_;
    my $event = $self->event;

    $Log->debug(
        sub {
            require Data::Dumper;

            return __PACKAGE__ . "::run: Running $event with args:\n" .
                Data::Dumper::Dumper(\%opts);
        }
    );

    return $self->$event(%opts);
}

=head2 _filtered_volume_array

    @filtered_volumes = $self->_filtered_volume_array(
        command_interface => $node,
        vserver           => $vserver,
        volume_list       => $vol_list_ref,
        volume_pattern    => $regex
    );

Takes a list of volumes and/or volume names, converts them all to
B<NACL::STask::Volume> objects, then filters them according to the given
pattern. Returns a list of B<NACL::STask::Volume> objects.

If the pattern is not given, then no matching is done and the list is not
filtered.

=over

=item Options

The following options are recognized by _filtered_volume_array:

=over

=item C<< command_interface => $node_obj >>

(Required) An object that represents a node or cluster interface. Needed for
converting volume names to objects.

=item C<< vserver => $string >>

(Required) A string that represents a vserver name. Used in converting volume
names to objects.

=item C<< volume_list => $arrayref >>

(Required) An array-reference of zero or more volumes to be converted to objects
and filtered. Each element may be either a string or a B<NACL::STask::Volume>
object. Strings are converted in-place to objects.

=item C<< volume_pattern => $string|$regex_obj >>

(Optional) A string or a regular expression object. If given, the pattern is
matched against all volume names and only those that match are kept in the
list. If the value given is a plain string it will be converted to a regular
expression object. (This will result in a substring search of the string within
each volume name.)

=back

=item Exceptions

Any exceptions that can be thrown by the B<create> method of
B<NACL::STask::Volume> can occur here. Such exceptions will be passed through
unchanged.

=back

=cut

sub _filtered_volume_array {
    my $self = shift;
    my %opts = validate(
        @_,
        {
            command_interface => { type => OBJECT | SCALAR },
            volume_list       => { type => ARRAYREF },
            volume_pattern    => {
                type     => SCALAR | OBJECT | UNDEF,
                optional => 1,
            },
            vserver           => { type => SCALAR },
        },
    );

    my @vol_list    = @{$opts{volume_list}};
    my $vol_pattern = $opts{volume_pattern};
    my $vserver     = $opts{vserver};
    my $ci          = $opts{command_interface};

    # If the pattern isn't already a Regexp object, make it into one:
    if ($vol_pattern && (! ref $vol_pattern)) {
        $vol_pattern = qr/$vol_pattern/;
    }

    if ($vol_pattern) {
        @vol_list =
            grep { (ref $_ ? $_->volume : $_) =~ $vol_pattern } @vol_list;
    }

    foreach my $vol (@vol_list) {
        next if (ref $vol);

        $vol = NACL::STask::Volume->find(
            command_interface => $ci,
            filter            => {
                volume  => $vol,
                vserver => $vserver,
            }
        );
    }

    return @vol_list;
}

=head2 _getCluster

    my $cluster = $self->_getCluster($node);

Takes a B<NACL::STask::Node> object and finds the cluster
it belongs to. The resulting B<NACL::STask::Cluster> object is returned.

=over

=item Options

The following option is recognized by _getCluster:

=over

=item B<$node>

(Required) A positional argument that is a B<NACL::STask::Node> object. This
object is used to find and create the cluster object.

=back

=item Exceptions

Any exceptions that can be thrown by the B<find_unique_clusters> method of
B<NACL::STask::Cluster> can occur here. Such exceptions will be passed through
unchanged.

If no node is passed in, a B<NATE::Exceptions::Argument> exception will be
thrown.

=back

=cut

sub _getCluster {
    my $self = shift;
    my $node = shift;

    if (! ((blessed $node) && ($node->isa('NACL::STask::Node')))) {
        NATE::Exceptions::Argument->throw('Must pass in a NACL::STask::Node');
    }

    my @clusters = NACL::STask::Cluster->find_unique_clusters(
        nodes => [ $node ]
    );

    return $clusters[0];
}

=head2 _getClusterAndNodes

    my ($cluster, @nodes) = $self->_getClusterAndNodes(@nodes);

Converts nodes that are strings into B<NACL::STask::Node> objects, and also
finds and returns the cluster object associated with the first node in the
list.

The list of nodes may contain elements that are already B<NACL::STask::Node>
objects. If so, they are returned unchanged.

The return value is a list whose first element is a B<NACL::STask::Cluster>
object, and whose remaining elements are the B<NACL::STask::Node> objects
corresponding to the inputs.

=over

=item Options

The following option is recognized by _getClusterAndNodes:

=over

=item B<@nodes>

(Required) All values passed to this method are treated as a list of node names
and/or objects. Those that are names will be converted to B<NACL::STask::Node>
objects for the return list.

=back

=item Exceptions

Any exceptions that can be thrown by the B<find_unique_clusters> method of
B<NACL::STask::Cluster> or the B<new> method of B<NACL::STask::Node> can occur
here. Such exceptions will be passed through unchanged.

If no nodes are passed in, a B<NATE::Exceptions::Argument> exception will be
thrown.

=back

=cut

sub _getClusterAndNodes {
    my $self = shift;
    my @nodes = @_;

    if (! @nodes) {
        NATE::Exceptions::Argument->throw('Must pass in at least one node');
    }

    for my $node (@nodes) {
        next if (ref($node) && $node->isa('NACL::C::Node'));

        $node = $NODE_MAP{$node} ||=
            NACL::STask::Node->new(
                name          => $node,
                apiset_should => { interface => $self->interface }
            );
    }

    my $cluster = $CLUS_MAP{$nodes[0]->node} ||= $self->_getCluster($nodes[0]);

    return ($cluster, @nodes);
}

1;

=head1 EVENT ITERATION LOOP

(This is a summary of the section by the same name in the documentation for
B<NACL::AppBuilder::DpgWorkflows::EventWorker>.)

When the C<start> method of B<NACL::AppBuilder::DpgWorkflows::EventWorker> is
called, the worker instance sends code to the chosen client machine via STAF.
This code contains the processing loop that will be used to invoke the chosen
remote event. This processing loop is called the I<event iteration loop>.

The event iteration loop utilizes the C<initial_delay>, C<iterations>,
C<iteration_delay> and C<runlength> attributes of the worker instance. These
fulfill the following roles:

=over

=item B<initial_delay>

Before the first iteration of the event (and only the first), if this value is
non-zero the worker will wait that many seconds.

=item B<iterations>

If this value is not zero, then it dictates the number of iterations the event
will run before the remote process stops execution and signals the worker
instance that it is completed. If C<runlength> is set, this will always be
zero.

=item B<runlength>

If this value is not zero, then it dictates the length of time for which the
event will run in a continuous loop. If this value is non-zero, then
B<iterations> will always be zero.

Note that if an initial delay is set with C<initial_delay>, this counts as part
of the overall run-time; something configured to run for 5 minutes with a 2
minute initial delay would have an effective available run-time of only 3
minutes.

=item B<iteration_delay>

If this value is not zero, then each iteration of the event will wait for this
many seconds before starting the next iteration. This delay will not be applied
on the last iteration.

If the run is governed by C<runlength> rather than by C<iterations>, then the
delay is applied after every iteration I<until> an iteration takes the process
past the end-time dictated by C<runlength>. That will be the last iteration,
and no additional delay is applied.

=back

In general, individual events should not need to care about the iteration loop
as it is managed and controlled at a higher level.

=head1 DEVELOPING EVENTS AND EVENT CLASSES

As the EventManager system gains more use, it will need the development of more
events. Some of these events may not fit into the scope of the
B<NACL::AppBuilder::DpgWorkflows::CoreEvents> class, and thus require new Perl
modules be developed.

=head2 Developing Event Classes

This guide will start with the development of classes. As more and more events
are developed, performance will benefit from having them broken down into
related groups and kept in different classes.

=head3 Classes are based on Moose

All of the EventManager classes are written using the B<Moose> meta-object
protocol for Perl. This may seem overwhelming at first, but it will soon be
easy to understand and extend. Start by reading the Moose manual page,
B<Moose::Manual>, which will introduce the basics of the Moose system. You will
not need to become an expert at the Moose system, as much of the work is done
for you in this module you are currently looking at.

This module, B<NACL::AppBuilder::DpgWorkflows::EventBase>, is a type of Moose
class called a I<Role>. It cannot be instantiated like a class, but it provides
the basic attributes and a few methods that are common to all the actual event
classes. These attributes and methods were documented above.

=head3 Starting the new class

The following is a complete event class, called
B<NACL::AppBuilder::DpgWorkflows::NewEvents>, that can act as a template for
creating the new class:

    package NACL::AppBuilder::DpgWorkflows::NewEvents;
    
    use 5.014;
    use strict;
    use warnings;
    use namespace::autoclean;
    
    use Moose;
    
    with 'NACL::AppBuilder::DpgWorkflows::EventBase';
    
    __PACKAGE__->meta->make_immutable;
    
    1;

This is even longer than it needs to be, since C<use Moose> actually turns on
C<strict> and C<warnings> for you automatically. The key line is this one:

    with 'NACL::AppBuilder::DpgWorkflows::EventBase';

This line has the new class use this role as part of its interface. This will
provide the new class with everything it needs to correctly interact with the
remote code that EventWorker generates and sends to client machines.

You will likely need to add some number of NATE and/or NACL libraries as well.
For example, if you plan to log information using the B<NATE::Log> system, you
may want to add:

    use Const::Fast;
    use NATE::Log qw(log_global);
    
    const my $Log => log_global;

This will declare a package-wide global called C<$Log> that cannot be changed
accidentally. Its value will be the global logging object from the B<NATE::Log>
package.

Do not add any exception class from which you plan to import the
try/catch/with/etc. keywords. This causes a conflict with Moose, and a
work-around is explained in a later section.

=head3 Adding attributes

If your event class is going to need any additional attributes beyond what this
role provides, you can add them after the C<with> statement, but before the
call to C<< __PACKAGE__->meta->make_immutable >>.

Remember that the EventWorker will not know about these attributes as it does
with the core attributes provided here. There will be no way to inform the
worker-loop to populate any data in the new attributes. However, the event code
itself can use the attributes to pass information between iterations of the
event, or between calls to other events.

In most cases, you will not need to do this, as the C<state> attribute is
persistent between iterations and is more open-ended than creating specific
new attributes.

=head3 Using NACL/NATE exceptions with Moose

One area that can create problems for users developing events within the Moose
structure, is that both Moose and the B<NATE::Error> class define a subroutine
by the name of C<with>. If you try to use NACL or NATE exceptions and import
the C<try>, C<catch>, C<with>, etc. pseudo-keywords (including C<:try>) you will
get a compile-time error because of this conflict.

To work around this, do not rely on B<namespace::autoclean> to clean up the
Moose elements at the end of compilation. Instead, explicitly clean them up
with C<no Moose> before you attempt to import the exception keywords:

    no Moose;
    use NATE::BaseException qw(:try);

Do this after the C<make_immutable> call, but before any of your event methods
are defined. Any exception classes you plan to use solely for throwing can be
loaded earlier (with the rest of the C<use> statements). This only applies to
the class from which you are going to import C<try>, C<with>, etc.

=head2 Developing Events

Whether you are writing a new events class or adding to an existing one, the
process of developing new events should be generally the same.

=head3 Events are methods

An event in the context of the EventManager system is just a method in a class
that can be invoked by a worker instance. So writing events is just like writing
methods.

Each event will always be called as an instance method. There will be no need to
check whether it is called statically. Because each call is on an instance, each
event call will have an object with a full set of the attributes (described
earlier) available on it. The object can also be used to call other events, if
you have designed an event to be a composition of other events (such as the
B<volumeOffline>, B<volumeOnline> and B<volumeOfflineOnline> methods in the
B<CoreEvents> class).

Events do not need to worry about scheduling themselves with regards to the
number of iterations or run-length. The EventWorker-generated code will already
be handling this.

=head3 Event parameters and return values

As methods, events will always receive at least one parameter, the object
referent. Any parameters given to the event by the EventWorker will be passed
in as key/value pairs in a hash table. The standard for parameter validation
in NACL is to use the B<Params::Validate> package:

    sub delay {
        my $self = shift;
        my %opts = validate_with(
            params      => \@_,
            allow_extra => 1,
            spec        => {
                event_delay => { type => SCALAR, default => 60 },
            },
        );

Any approach you wish to use for your parameter validation is fine. Using
B<Params::Validate> simply means that your code will more closely resemble
other NACL code (as well as the elements of EventManager), plus
B<Params::Validate> will already be loaded into memory (since it is used by
this role).

The return value of an event is generally not very important, since it cannot
be propagated back to the manager itself. However, it may be useful to other
events if you are composing events together, and it can also be used to signal
the iteration loop that it should stop iterating regardless of the number of
iteration or current time.

If the return value from an event is C<undef>, the iteration loop will stop
after that iteration. A return value of zero (C<0>) will not trigger this, as
it is a defined (albeit false) value. Only C<undef> will trigger this (and, by
extension, a bare C<return> with no return value). Make certain your events
always return a defined value, even if that value is just C<1> or C<$self>.

=head3 Events calling events

Lastly, events may call other events within the same class simply by calling the
method on the object referent and passing any needed parameters. There are
several examples of this in the B<CoreEvents> class. These examples also
illustrate the value of returning actual data from events, even if it is not
used by the iteration loop.

For example, the B<volumeOfflineOnline> event is a composition of the
B<volumeOnline> and B<volumeOffline> events. However, B<volumeOnline> requires
a set of junction-path strings in order to re-mount volumes, data which is not
passed in calls to B<volumeOfflineOnline>. The return value of B<volumeOffline>
provides exactly this piece of information. The implementation of
B<volumeOfflineOnline> calls the offline event first, saving the return value
and passing that as an additional parameter when the online event is then
called.

At present, events cannot call events from other classes.

=head1 SEE ALSO

B<NACL::AppBuilder::DpgWorkflows::EventManager>,
B<NACL::AppBuilder::DpgWorkflows::EventWorker>,
B<NACL::AppBuilder::DpgWorkflows::CoreEvents>

=head1 AUTHOR

Randy J. Ray E<lt>rjray@netapp.comE<gt>

=cut
