# Copyright (c) 2015 NetApp, Inc., All Rights Reserved.
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#
## @summary Event Manager class
## @author rjray@netapp.com
## @status shared
## @pod here

package NACL::AppBuilder::DpgWorkflows::EventManager;

use 5.014;
use strict;
use warnings;
use namespace::autoclean;

use Clone qw(clone);
use Const::Fast;
use List::Util qw(max);
use Moose;
use Moose::Util::TypeConstraints;
use Params::Validate qw(validate validate_with validation_options
                        UNDEF OBJECT HASHREF SCALAR SCALARREF ARRAYREF BOOLEAN);
use Scalar::Util qw(blessed refaddr);

use NATE::Log qw(log_global);
use NACL::AppBuilder::DpgWorkflows::EventWorker;

# This pattern is used in two places, so pre-define it:
const my $CLASS_RE => qr/[[:alpha:]]\w*(?:::\w+)*/;

# id: A unique ID for this manager, unique only while the object exists.
has id => (
    is      => 'ro',
    isa     => 'Int',
    lazy    => 1,
    builder => '_build_id',
);

# client_list: The list of clients to use for dispatching worker jobs.
has client_list => (
    is      => 'ro',
    isa     => subtype(
        'ArrayRef[Str]',
        where   { @{$_} > 0 },
        message { 'The client_list parameter must have at least one element' }
    ),
    reader  => '_client_list_copy',
    builder => '_build_client_list',
    traits  => [ 'Array' ],
    handles => {
        client_list        => 'elements',
        _client_list_push  => 'push',
        _client_list_shift => 'shift',
    },
);

# workers: The table of currently-alive EventWorker objects.
has workers => (
    is       => 'ro',
    isa      => 'HashRef[NACL::AppBuilder::DpgWorkflows::EventWorker]',
    default  => sub { {} },
    init_arg => undef,
    reader   => '_get_workers',
    traits   => [ 'Hash' ],
    handles  => {
        workers        => 'elements',
        workers_set    => 'set',
        workers_get    => 'get',
        workers_count  => 'count',
        workers_delete => 'delete',
    },
);

# active_workers: A list of the workers (IDs) that are currently running
has active_workers => (
    is       => 'ro',
    isa      => 'ArrayRef[Str]',
    reader   => '_active_workers',
    init_arg => undef,
    default  => sub { [] },
    traits   => [ 'Array' ],
    handles  => {
        active_workers              => 'elements',
        active_workers_count        => 'count',
        _active_workers_push        => 'push',
        _active_workers_splice      => 'splice',
        _active_workers_first_index => 'first_index',
    },
);

# default_class: The default class an event is assumed to be in, if the
#                value for that event doesn't include a class name.
has default_class => (
    is      => 'ro',
    isa     => subtype(
        'Str',
        where   { /^$CLASS_RE$/ },
        message { "Invalid Perl class name '$_'" }
    ),
    default => 'NACL::AppBuilder::DpgWorkflows::CoreEvents',
);

# runid_base: The "base directory" for the run-ID, if any. If present, is
#             prefixed to the run-ID with a "/" separator.
# runid_prefix: Any desired text prefix for the run-ID. Comes after the
#               sequence number but before the event name.
has [ qw(runid_base runid_prefix) ] => (
    is  => 'ro',
    isa => 'Str',
);

# sequence_width: The min width of numbers used as the sequence counter.
has sequence_width => (
    is      => 'ro',
    isa     => subtype(
        'Int',
        where { $_ > 0 },
        message { 'Integer value must be greater than zero' }
    ),
    default => 2,
);

# sequence_format: The sprintf string used to generate sequence numbers
has sequence_format => (
    is       => 'ro',
    isa      => 'Str',
    lazy     => 1,
    builder  => '_build_sequence_format',
    init_arg => undef,
);

# keep_finished_workers: Whether to keep finished EventWorker objects around
#                        or to discard them.
has keep_finished_workers => (
    is      => 'rw',
    isa     => 'Bool',
    default => 0,
);

# process_on_destroy: What EventWorker objects should do with their process
#                     objects when they are destroyed.
has process_on_destroy => (
    is      => 'ro',
    isa     => 'Str',
    default => 'stop',
);

# default_worker_callback: A coderef that is used as a default for on_start and
#                          on_finish when workers are created.
has default_worker_callback => (
    is       => 'ro',
    isa      => 'CodeRef',
    lazy     => 1,
    builder  => '_build_default_worker_callback',
    init_arg => undef,
);

# _next_client_index: Private index value into the client_list array.
# _next_interface_index: Private index value into the interfaces array.
has [ qw(_next_client_index _next_interface_index) ] => (
    is       => 'rw',
    isa      => 'Int',
    init_arg => undef,
    default  => 0,
);

# Global handle on the NATE::Log logging interface:
const my $Log => log_global;
# The interface options we have:
const my @INTERFACES => qw(ZAPI CLI);
# Used in create_many_workers() to separate worker options from event params:
const my %WORKER_NEW_OPTIONS =>
    map { $_ => 1 } (qw(client
                        interface
                        username
                        password
                        runid
                        class
                        event
                        event_params
                        on_start
                        on_finish
                        initial_delay
                        iterations
                        iteration_delay
                        runlength
                        process_on_destroy));
# The regular expression used to pull apart event strings into a triple of
# ($class, $name, $decorator):
const my $EVENT_REGEX =>
    qr{
          ^
          # Match an arbitrary-length Perl class name followed by "->":
          (?:($CLASS_RE)->)?
          # Match a name that must start with [a-zA-Z]:
          ([[:alpha:]]\w*)
          # Match anything inside "()", but don't match the parens:
          (?:[(](.*)[)])?
          $
      }x;


# Not a constant, used for tracking the sequence counters:
my %SequenceCounter = ();

__PACKAGE__->meta->make_immutable;

###############################################################################
# NOTE: Because I am using Moose for the object model, I have moved the usage
# of any NATE or NACL exception classes down here, after an explicit
# "no Moose;" statement. Do not try to move this higher in the code, or
# remove the "no" statement, or compilation of the module will break.
###############################################################################

no Moose;
use NATE::BaseException qw(:try);
use NATE::Exceptions::Argument;

# Set the on_fail behavior for Params::Validate to throw an exception (of the
# NATE::Exceptions::Argument class) on errors instead of a plain croak().
validation_options(
    on_fail => sub {
        NATE::Exceptions::Argument->throw(shift);
    }
);

=head1 NAME

NACL::AppBuilder::DpgWorkflows::EventManager - Events manager/scheduler

=head1 SYNOPSIS

    use NACL::AppBuilder::DpgWorkflows::EventManager;
    
    # Create a manager
    my $manager = NACL::AppBuilder::DpgWorkflows::EventManager->new(
        client_list           => \@clients,
        runid_base            => 'events',
        keep_finished_workers => 1
    );
    
    # @tasks is an array of hash-references from YAML data:
    for my $task (@tasks) {
        $manager->create_worker(
            event        => $task->{testcase},
            event_params => $task->{param}
        );
    }
    
    $manager->worker_start_all;
    $manager->worker_wait_all;

=head1 DESCRIPTION

The B<NACL::AppBuilder::DpgWorkflows::EventManager> class is a scheduler and
dispatcher for small units of work, referred to as "events", which run in the
background. The EventManager creates and manages a number of instances of the
B<NACL::AppBuilder::DpgWorkflows::EventWorker> class, which handle the actual
remote communication of individual events.

Each manager instance (there can be more than one EventManager active at a time)
will keep track of all workers that it creates, as well as managing such things
as the run-ID strings for spawned processes.

=head1 ATTRIBUTES

The following attributes are defined for instances of this class. Each
attribute value is obtained by calling the attribute name as an accessor. Most
attributes are read-only (they cannot be changed after the object is created).

=over

=item id

A numerical ID that identifies this object. Guaranteed to be unique only as
long as the object itself exists.

=item client_list

An array of the clients that were provided to manager when it was created. When
called as an accessor, this returns a list of the clients, not an
array-reference. The returned list will contain the same elements in the same
order as was given to the constructor for the object. If no clients are given
in the constructor, the EventManager will attempt to find clients from the
NATE parameters passed in to the running test.

=item workers

A hash table of the current B<NACL::AppBuilder::DpgWorkflows::EventWorker>
objects the manager is controlling. The keys of the hash are the numerical ID
values of the workers, and the value of each key is the worker object itself.
When called as an accessor, this returns a hash table, not a hash-reference.

=item active_workers

An array of the workers that are currently running. Each element of the list
will be a worker ID value that can be used to retrieve the worker object. The
list is not in any specific order. When called as an accessor, returns a list
of the IDs, not an array-reference.

=item default_class

A string that is a valid Perl class name. This is used as the default class in
which to locate an event if the call to create the worker doesn't provide an
explicit class. It defaults to B<NACL::AppBuilder::DpgWorkflows::CoreEvents>.

=item runid_base

A string that will be used as the base for all run-ID strings that are
generated by this manager. If this is set, then it will be prepended along
with a C</> character to each run-ID.

=item runid_prefix

A string that, if set, is inserted into each created run-ID string between
the sequence number and the event name.

=item sequence_width

An integer that determines the minimum width of the sequence numbers in run-ID
strings. Defaults to 2 if not explicitly given.

=item sequence_format

A string used with B<sprintf> to create a zero-padded sequence number of at
least C<sequence_width> digits in length.

=item keep_finished_workers

A boolean value that specifies whether the manager will hold on to the worker
objects that have completed, or free them. Worker objects include some amount
of data (such as total run time), so there may be a desire to keep them in
some cases.

This attribute is writable; it may be changed by the user after it is set
during object construction:

    # Explicitly choose to keep worker objects:
    $manager->keep_finished_workers(1);

If not specified during construction, the value defaults to C<0> (false).

=item process_on_destroy

A string value that specifies what the created worker objects should do with
the encapsulated B<NATE::Process> objects they have, when the workers are being
destroyed (and the process is still running). Should be one of three values:

=over

=item C<stop>

This is the default value. The B<NATE::Process> object is sent the B<stop>
message, and the worker does a B<wait> to wait for it to finish. If it does not
finish within 60 seconds, B<NATE::Process> will send the remote process a hard
KILL signal.

=item C<kill>

Immediately send the remote process a hard KILL signal, bypassing the usual
B<stop> process.

=item C<detach>

Allow the remote process (and local controlling child process) to continue to
exist after the worker object is destroyed. There will be no way for the event
manager to get at the exit status of the remote process, if this is done.

=back

Any value other than the above will be treated as C<stop>.

=item default_worker_callback

A code-reference that is used for both the C<on_start> and C<on_finish>
callbacks in worker creation, if either is not otherwise specified. This
single callback handles both the "start" and "finish" roles. It manages the
C<active_workers> list and the deletion of worker objects when they complete
(unless C<keep_finished_workers> is "true" at the time).

If you provide your own callbacks for either (or both) of the worker callback
values, you should consider calling into this callback as the last step of your
code. See L</CALLBACKS>, below.

=back

Some of the attributes also marshall additional methods on the manager object.
These will be detailed in the next section.

=head1 METHODS

In addition to the attribute accessors, the class provides the following
methods. No subroutines are exported by this package.

=head2 new

    $manager = NACL::AppBuilder::DpgWorkflows::EventManager->new(
        client_list           => \@clients,
        runid_base            => $base,
        runid_prefix          => $prefix,
        sequence_width        => $width,
        keep_finished_workers => $boolean,
        process_on_destroy    => $string
    );

(Must be called as a class method)

Create a new manager object and return it to the caller. No worker objects
are created at this time.

All failures in B<new> throw an exception. See C<Exceptions>, below.

=over

=item Options

The following options are recognized by B<new>:

=over

=item C<< client_list => $arrayref >>

(Optional) A list of one or more client names that will have events distributed
out among them. All names must have locatable NATE hostrec files, or worker
objects may throw an exception at creation-time if a client hostrec file cannot
be found. Once specified, this list cannot be added to or shortened. If the
array-reference has no elements, the constructor will fail.

If this parameter is not passed, the B<find> method of B<NACL::C::Client> will
be used to search for any clients specified to the execution environment (via
NATE params such as CLIENTS, etc.). If there were none given, the constructor
will fail.

=item C<< default_class => $string >>

(Optional) A string that specifies a Perl class name. This is the package in
which any event that does not include an explicit class will be looked for. If
not specified, the default is B<NACL::AppBuilder::DpgWorkflows::CoreEvents>.

=item C<< runid_base => $string >>

(Optional) A string that is prepended, along with a C</> character, to all
run-ID strings that are created. See L</CREATING RUNID STRINGS>, below.

=item C<< runid_prefix => $string >>

(Optional) A string that is included in all run-ID strings that are created. It
will appear after the sequence number but before the event name. See
L</CREATING RUNID STRINGS>, below. No default.

=item C<< sequence_width => $integer >>

(Optional) An integer that specifies the width (in digits, with zero padding)
of the sequence numbers used in run-ID strings. See L</CREATING RUNID STRINGS>,
below. Defaults to 2.

=item C<< keep_finished_workers => $boolean >>

(Optional) A boolean value that specifies whether worker objects should be
immediately destroyed when their events complete. Worker objects carry some
additional data that can be accessed after the event is finished, so some users
may wish to keep them around. The default is C<0> (false), to not retain the
objects.

=item C<< process_on_destroy => $string >>

(Optional) A string that determines how all workers this manager creates will
handle their B<NATE::Process> object during their own destruction. Defaults
to C<stop> (see above for the recognized values of this parameter).

=back

=item Exceptions

The following exceptions may be triggered by this method:

=over

=item Moose Exceptions

The only exceptions that can occur inside of B<new> are type-verification
exceptions or exceptions that stem from an unknown parameter. These
exceptions will be thrown by the Moose system, so they will not match
B<NATE::BaseException>. You can catch them with the C<otherwise> block of
a C<try> construct.

=back

=back

=head2 id

    printf "Event Manager created with ID #%d\n", $manager->id;

Returns the object's unique ID.

=head2 client_list

    my @clients = $manager->client_list;

Accessor; returns a list of the clients that the object was created with.
Cannot be changed by the user.

=head2 workers

    my %all_workers = $manager->workers;

Returns a hash table of the worker objects currently under management by this
object. Each key is a worker object ID, and each value is the corresponding
worker object. Takes no arguments.

=head2 workers_set

    $manager->workers_set($worker->id, $worker);
    # or
    $manager->workers_set(
        $worker1->id => $worker1,
        $worker2->id => $worker2,
        # ...
    );

Adds a new worker to table of managed worker objects. If you manipulate the
table directly like this, you are not required to use the ID of the worker
object as a key, but it is a good practice. The value (second argument) must be
an object of the B<NACL::AppBuilder::DpgWorkflows::EventWorker> class, or an
exception will be thrown.

=head2 workers_get

    $worker = $manager->workers_get($worker_id);
    # or
    @some_workers = $manager->workers_get($id1, $id2, $id3);

Get one or more worker objects by their keys (IDs) in the table. Each returned
value will be a B<NACL::AppBuilder::DpgWorkflows::EventWorker> object, as the
manager class type-checks all values going into this table.

=head2 workers_delete

    $manager->workers_delete($worker_id);
    # or
    $manager->workers_delete($id1, $id2, $id3);

Delete one or more workers from the table. Unless other references to the
deleted worker objects exist, they will also be freed by Perl.

=head2 workers_count

    if (my $count = $manager->workers_count) {
        $Log->comment("$count worker objects still alive.");
    } else {
        $Log->comment('No worker objects currently allocated.');
    }

Returns an integer count of the number of workers currently in the table. Does
not distinguish between workers that are or are not currently running (active).

=head2 active_workers

    for my $w_id ($manager->active_workers) {
        my $worker = $manager->workers_get($w_id);
        # Do something with $worker
    }

Returns a list of zero or more worker IDs that refer to workers that have been
started and are currently still running. Note that the management of this list
is handled by the callbacks that are sent to each worker as it is created. If
you send your own callbacks, or use something other than the worker ID as a key
for storing the worker objects, then the list returned by this method may be
incomplete.

=head2 active_workers_count

    while ($manager->active_workers_count) {
        # Do some other things, while we wait for active workers
        # to complete...
    }

Returns the number of active workers, as understood by the manager object.
Useful as a predicate test because it will always returned a defined value,
even if the value is 0.

=head2 runid_base

    my $runid = $manager->runid_base . q{/} . $event_name;

Accessor; returns the value of the run ID base string that the manager was
given when B<new> was called. Cannot be changed.

=head2 runid_prefix

    my $runid = $manager->runid_base . q{/} . $manager->runid_prefix .
        q{_} . $event_name;

Accessor; returns the value of the run ID prefix string that the manager was
given when B<new> was called. Cannot be changed.

=head2 sequence_width

    printf "Sequences in run IDs will be %d digits in width\n",
        $manager->sequence_width;

Accessor; returns the value of the run ID sequence-width integer that the
manager was given when B<new> was called. Cannot be changed.

=head2 sequence_format

    my $seq_number = sprintf $manager->sequence_format, $counter;

Returns the B<sprintf> format-string that is used in run ID creation to create
the zero-padded sequence number. Cannot be changed.

=head2 keep_finished_workers

    # Get the current value:
    my $keep = $manager->keep_finished_workers;
    # Explicitly set the value:
    $manager->keep_finished_workers(1);

Accessor; returns the current value of the C<keep_finished_workers> attribute.
Also allows for changing the value, as this attribute is writable.

=head2 default_worker_callback

    my $newcallback = sub {
        my ($worker, $stage) = @_;
        # Do stuff...
        # Now make sure the default callback is invoked:
        $manager->default_worker_callback->($worker, $stage);
        return;
    }

Returns a code-reference that is the default callback used when worker objects
are created. Cannot be altered, but may be useful in your own callbacks if you
want to make sure that the manager keeps all of its internal tables in order.
See L</CALLBACKS>, below.

=head2 create_worker

    my $worker_id = $manager->create_worker(%worker_options);

Create a new B<NACL::AppBuilder::DpgWorkflows::EventWorker> object and store it
for later management.

All the parameters (except one) are passed through to the C<new> method of the
B<NACL::AppBuilder::DpgWorkflows::EventWorker> class. See the documentation of
that class for a detailed list of recognized parameters.

=over

=item Options

The following options are treated specially or given defaults before the call
to create the worker:

=over

=item client

If this parameter is not explicitly passed, the manager will default it to the
next available client from the manager's C<client_list> pool.

=item event

This parameter is required, but the event manager allows some additional syntax
on the value. The event manager will recognize the following pattern:

    (?:([[:alpha:]]\w*(?:::\w+)*)->)?([[:alpha:]]\w*)(?:[(](.*)[)])?

This is a super-set of the pattern that EventWorker recognizes. It adds an
optional class-name with a Perl arrow operator at the beginning and an optional
string in parentheses at the end. If the class-name is present it is used to
qualify the name of the event. If the optional string at the end is present, it
is removed from the name and used as the "decorator" in the run ID string (see
L</CREATING RUNID STRINGS>, below).

For example, if the value given for C<event> is:

    volumeOfflineOnline(cluster1group3)

then the actual event will be C<volumeOfflineOnline>, and the string
C<cluster1group3> will be added to the end of the run ID string. The event will
be considered to be in the default class for events
(B<NACL::AppBuilder::DpgWorkflows::CoreEvents>).

If the value given is something like this:

    NACL::AppBuilder::DpgWorkflows::WAFLEvents->wreck

then the event name will be C<wreck>, but it will be in the class
B<NACL::AppBuilder::DpgWorkflows::WAFLEvents> which will be loaded by the
remote client when the remote code is executed. There will be no "decorator"
added to the run ID string (as none was provided).

=item interface

If this parameter is not explicitly passed, the manager will default it to one
of C<CLI> or C<ZAPI>. The two options alternate each time they are filled in,
in an effort to balance the use of ZAPI and command-line in the events.

=item on_start

If this parameter is not explicitly passed, the value of the manager's
C<default_worker_callback> attribute will be used as a default.

=item on_finish

If this parameter is not explicitly passed, the value of the manager's
C<default_worker_callback> attribute will be used as a default.

=item runid

If this parameters is not explicitly passed, the manager will construct a
run ID string and pass it as the value for this parameter. See the section
L</CREATING RUNID STRINGS>, below, for details on how this is created.

=item no_sequence

This is a special option that is removed from the set of parameters before
the worker object is constructed. The value of this option is treated as a
boolean value that dictates whether the sequence number in a constructed
run ID is present or not. If this option has a "true" value, then no sequence
ID will be present in the run ID if it is generated by the manager. This has
no effect on a run ID that is explicitly provided by the C<runid> parameter.

=back

=item Exceptions

The following exceptions may be thrown by B<create_worker>:

=over

=item B<NATE::Exceptions::Argument>

If an unknown or invalid parameter is passed, this exception will be thrown.

=item B<NATE::BaseException>

If the construction of the worker object fails, any exception it throws will be
passed through. Any Moose exceptions will be converted to this exception class,
as well. This exception may also signal that a value passed through to the
B<new> method of B<NACL::AppBuilder::DpgWorkflows::EventWorker> was invalid, as
the exception from that would be a Moose exception that was converted to this
class.

=back

=back

The return value of this method is the ID of the newly-created worker.

=cut

sub create_worker {
    my ($self, %opts) = @_;

    # Make sure these don't accidentally fall through to new()...
    my $no_sequence = delete $opts{no_sequence};
    my $runid_base  = delete $opts{runid_base};

    $opts{client}             ||= $self->_next_client;
    $opts{interface}          ||= $self->_next_interface;
    $opts{on_start}           ||= $self->default_worker_callback;
    $opts{on_finish}          ||= $self->default_worker_callback;
    $opts{process_on_destroy} ||= $self->process_on_destroy;

    my ($class, $event, $decorator) = $self->_parse_event($opts{event});

    if (! $opts{runid}) {
        my @runid_base = $runid_base ? (base => $runid_base) : ();
        $opts{runid} = $self->_create_runid(
            @runid_base,
            event       => $event,
            decorator   => $decorator,
            no_sequence => $no_sequence
        );
    }

    # Push $event (and possibly $class) back into %opts for the call to
    # EventWorker->new:
    $opts{event} = $event;
    # Start by defaulting the class:
    $opts{class} ||= $self->default_class;
    if ($class) {
        $opts{class} = $class;
    }

    my $worker;
    try {
        $worker = NACL::AppBuilder::DpgWorkflows::EventWorker->new(%opts);
    } catch NATE::BaseException with {
        # If we got a NATE-class exception, just re-throw it.
        my $e = shift;
        $e->throw;
    } otherwise {
        my $e = shift;

        if ($e->isa('Moose::Exception')) {
            # I don't know of a way to configure the exceptions that Moose
            # throws when new() fails, so I can't auto-coerce them into NATE
            # exceptions. So do it manually.

            my $class   = ref $e;
            my $message = $e->message;
            NATE::BaseException->throw(
                "Caught Moose exception of class $class: $message"
            );
        } elsif ($e->isa('NATE::Result')) {
            # Apparently there are circumstances where a NATE::Result-based
            # object can be thrown. This is probably a bug in NATE::Error but
            # for now try to accomodate it.

            my $class   = ref $e;
            my $message = $e->message_string;
            NATE::BaseException->throw(
                "Caught $class result object:\n$message"
            );
        } else {
            # No idea what we got here, so be verbose about it:

            require Data::Dumper;

            my $unknown = Data::Dumper->Dump([ $e ], [ 'e' ]);
            NATE::BaseException->throw(
                "Unknown exception/object caught in create_worker:\n$unknown"
            );
        }
    };

    $self->workers_set($worker->id, $worker);

    return $worker->id;
}

=head2 worker_start

    $manager->worker_start(worker => $w_id);

Start the execution of the given worker.

=over

=item Options

The following option is recognized:

=over

=item C<< worker => $scalar >>

(Required) The ID of the worker to start execution on.

=back

=item Exceptions

The following exception may be thrown by this method:

=over

=item B<NATE::Exceptions::Argument>

If the value of C<worker> does not exist in the table of workers, this
exception will be thrown.

=back

=back

The return value of this method is the manager object itself.

=cut

sub worker_start {
    my $self = shift;
    my %opts = validate(
        @_,
        {
            worker => { type => SCALAR },
        }
    );

    my $worker = $self->workers_get($opts{worker});
    if (! $worker) {
        NATE::Exceptions::Argument->throw(
            "Unknown worker-ID ($opts{worker}) passed"
        );
    }

    $worker->start;

    return $self;
}

=head2 worker_start_all

    $manager->worker_start_all;

Start the execution of all non-active workers. Takes no options, and throws
no exceptions (unless the B<start> method on B<NATE::Process> fails).

The return value of this method is the manager object itself.

=cut

sub worker_start_all {
    my $self = shift;

    my %workers = $self->workers;
    for my $worker (values %workers) {
        next if ($worker->is_running || $worker->finished);

        $worker->start;
    }

    return $self;
}

=head2 worker_wait

    $manager->worker_wait(worker => $w_id);

Wait for the specified worker to complete its remote execution.

=over

=item Options

The following options are recognized:

=over

=item C<< worker => $scalar >>

(Required) The ID of the worker to wait on.

=item C<< timeout => $integer >>

(Optional) If given, the maximum time to wait for the remote process to
complete before throwing an exception. If not given, wait indefinitely.

=back

=item Exceptions

The following exception may be thrown by this method:

=over

=item B<NATE::Exceptions::Argument>

If the value of C<worker> does not exist in the table of workers, this
exception will be thrown.

=back

=back

The return value of this method is the manager object itself.

=cut

sub worker_wait {
    my $self = shift;
    my %opts = validate(
        @_,
        {
            worker  => { type => SCALAR },
            timeout => { type => SCALAR, regex => qr/^\d+$/, optional => 1 },
        }
    );

    my $worker = $self->workers_get($opts{worker});
    if (! $worker) {
        NATE::Exceptions::Argument->throw(
            "Unknown worker-ID ($opts{worker}) passed"
        );
    }
    delete $opts{worker};

    $worker->wait(%opts);

    return $self;
}

=head2 worker_wait_all

    $manager->worker_wait_all;

Wait on the completion of all active workers.

=over

=item Options

The following option is recognized:

=over

=item C<< timeout => $integer >>

(Optional) If given, the maximum time to wait for the remote processes to
complete before throwing an exception. If not given, wait indefinitely.

=back

=item Exceptions

If any of the workers hit the timeout on their B<wait>, an exception will be
thrown. This may be an ordinary B<die>, depending on the nature of
B<NATE::Process>.

=back

The return value of this method is the manager object itself.

=cut

sub worker_wait_all {
    my $self = shift;
    my %opts = validate(
        @_,
        {
            timeout => { type => SCALAR, regex => qr/^\d+$/, optional => 1 },
        }
    );

    my %workers = $self->workers;
    for my $worker (values %workers) {
        next if (! $worker->is_running);

        $worker->wait(%opts);
    }

    return $self;
}

=head2 worker_stop

    $manager->worker_stop(worker => $w_id);

Stop the execution of the given worker. Interrupts the remote process if it
has not already completed. Does nothing if the remote process has already
finished.

=over

=item Options

The following option is recognized:

=over

=item C<< worker => $scalar >>

(Required) The ID of the worker to stop execution of.

=back

=item Exceptions

The following exception may be thrown by this method:

=over

=item B<NATE::Exceptions::Argument>

If the value of C<worker> does not exist in the table of workers, this
exception will be thrown.

=back

=back

The return value of this method is the manager object itself.

=cut

sub worker_stop {
    my $self = shift;
    my %opts = validate(
        @_,
        {
            worker => { type => SCALAR },
        }
    );

    my $worker = $self->workers_get($opts{worker});
    if (! $worker) {
        NATE::Exceptions::Argument->throw(
            "Unknown worker-ID ($opts{worker}) passed"
        );
    }

    $worker->stop;

    return $self;
}

=head2 worker_stop_all

    $manager->worker_stop_all;

Stop the execution of all active workers. Takes no options, and throws no
exceptions (unless the B<stop> method on B<NATE::Process> fails).

The return value of this method is the manager object itself.

=cut

sub worker_stop_all {
    my $self = shift;

    my %workers = $self->workers;
    for my $worker (values %workers) {
        next if (! $worker->is_running);

        $worker->stop;
    }

    return $self;
}

=head2 worker_is_running

    if ($manager->worker_is_running(worker => $w_id)) {
        $Log->comment("Worker $w_id is still running...");
    }

Check whether the given worker is currently running.

=over

=item Options

The following option is recognized:

=over

=item C<< worker => $scalar >>

(Required) The ID of the worker to check.

=back

=item Exceptions

The following exception may be thrown by this method:

=over

=item B<NATE::Exceptions::Argument>

If the value of C<worker> does not exist in the table of workers, this
exception will be thrown.

=back

=back

The return value of this method is a boolean that indicates whether the worker
is running (true) or is not running (false).

=cut

sub worker_is_running {
    my $self = shift;
    my %opts = validate(
        @_,
        {
            worker => { type => SCALAR },
        }
    );

    my $worker = $self->workers_get($opts{worker});
    if (! $worker) {
        NATE::Exceptions::Argument->throw(
            "Unknown worker-ID ($opts{worker}) passed"
        );
    }

    return $worker->is_running;
}

=head2 worker_is_stop_completed

    if (! $manager->worker_is_stop_completed(worker => $w_id)) {
        $Log->comment("Worker $w_id is still stopping...");
    }

Checks whether the given worker has completed the stop process for the remote
process.

=over

=item Options

The following option is recognized:

=over

=item C<< worker => $scalar >>

(Required) The ID of the worker to check.

=back

=item Exceptions

The following exception may be thrown by this method:

=over

=item B<NATE::Exceptions::Argument>

If the value of C<worker> does not exist in the table of workers, this
exception will be thrown.

=back

=back

The return value of this method is a boolean indicating whether the process of
stopping is completed (true) or not completed (false).

=cut

sub worker_is_stop_completed {
    my $self = shift;
    my %opts = validate(
        @_,
        {
            worker => { type => SCALAR },
        }
    );

    my $worker = $self->workers_get($opts{worker});
    if (! $worker) {
        NATE::Exceptions::Argument->throw(
            "Unknown worker-ID ($opts{worker}) passed"
        );
    }

    return $worker->is_stop_completed;
}

=head2 create_many_workers

    my $workers = $manager->create_many_workers(
        worker_count   => 5,
        option_mapping => {
            vol_lists => volume_list
        },
        # The list-of-lists here will be mapped to each individual
        # worker-creation call as "volume_list" in the "event_params"
        # option:
        vol_lists => [
            [ @vol_list1 ],
            [ @vol_list2 ],
            [ @vol_list3 ],
            [ @vol_list4 ],
            [ @vol_list5 ],
        ],
        # Other options...
    );

B<create_many_workers> is a convenience method that simplifies (sort of) the
creation of several workers with identical or nearly-identical parameters.

The method takes all the parameters that B<create_worker> does, as well as
two additional options. It then makes an appropriate number of calls to
B<create_worker> with the general creation parameters as well as any values
that have been sliced out of the extra options.

=over

=item Options

In addition to the typical B<create_worker> options, the following options are
recognized:

=over

=item C<< option_mapping => $hashref >>

(Optional) If given, this is a hash-reference that specifies the mapping of
options currently in the set of parameters to names that are recognized by
the B<new> method of B<NACL::AppBuilder::DpgWorkflows::EventWorker>. See the
section on L</Option Mapping>, below.

=item C<< worker_count => $integer >>

(Optional) If given, this is an integer indicating the number of worker objects
that should be created. If this is not given, then the value is derived from
the size of any lists in C<option_mapping>, above.

=back

=item Exceptions

The following exception may be thrown by this method:

=over

=item B<NATE::Exceptions::Argument>

If there are errors in the option-mapping table, or a mismatch in size between
the mapping table and the requested number of workers, this exception will be
thrown.

=back

Additionally, if any of the calls to worker creation fail, the exception from
that will be passed through.

=back

The return value of this method is an array-reference of the worker ID values
for all the newly-created worker objects.

=head3 Option Mapping

The C<option_mapping> parameter may seem confusing at first, but once it is
understood it can be very useful.

Suppose you wish to create 5 workers that do the B<volumeOfflineOnline> event.
Each of the workers will have the same parameters, except for the
C<volume_list> value which should be different for each one. You can make 5
calls to B<create_worker>:

    $manager->create_worker(
        event        => 'volumeOfflineOnline',
        event_params => {
            node        => $src_node,
            vserver     => $src_vserver,
            volume_list => \@volume_list_1,
        }
    );
    $manager->create_worker(
        event        => 'volumeOfflineOnline',
        event_params => {
            node        => $src_node,
            vserver     => $src_vserver,
            volume_list => \@volume_list_2,
        }
    );
    $manager->create_worker(
        event        => 'volumeOfflineOnline',
        event_params => {
            node        => $src_node,
            vserver     => $src_vserver,
            volume_list => \@volume_list_3,
        }
    );
    # Etc. Good luck with the cut/paste errors...

However, since it is only the C<volume_list> element of the C<event_params>
option that changes, this can be done more compactly with the following:

    $manager->create_many_workers(
        option_mapping => { vol_lists => 'volume_list' },
        event          => 'volumeOfflineOnline',
        event_params   => {
            node    => $src_node,
            vserver => $src_vserver,
        }
        vol_lists      => [
            \@volume_list_1,
            \@volume_list_2,
            \@volume_list_3,
            \@volume_list_4,
            \@volume_list_5,
        ]
    );

Any option in the C<option_mapping> table that is a known option to the B<new>
method of B<NACL::AppBuilder::DpgWorkflows::EventWorker> is mapped at that
level, and any options that are not known are pushed down into the
C<event_params> level. Since C<vol_lists> maps to C<volume_list>, and
C<volume_list> is not known to the B<new> method, it will be added to the
C<event_params> table, one value per iteration.

Note that you can map I<both> C<event_params> and an element that will be put
into it. The mapping of C<event_params> is done first, so that any additional
parameter values are only added to the proper set of C<event_params>.

If the C<worker_count> parameter is also given, then every mapping-set must
have at least that many elements (or an exception is thrown). For example, if
the above call to B<create_many_workers> include a C<worker_count> of 6, it
would fail as there are only 5 elements in C<vol_lists>.

All operations on mappings and mapped values are done on copies, so it is safe
to pass direct references to your own lists and/or hash tables.

=cut

sub create_many_workers {
    my ($self, @opts) = @_;
    my %opts = validate_with(
        params      => \@opts,
        allow_extra => 1,
        spec        => {
            option_mapping => { type => HASHREF, optional => 1 },
            worker_count   => { type => SCALAR, optional => 1 },
        }
    );

    my ($mapping, $worker_count, %options, @workers);

    $worker_count = delete $opts{worker_count};
    if ($mapping = delete $opts{option_mapping}) {
        # Convert the option-mapping table to something that can be used and
        # traversed in place as we create the series of workers. Also do some
        # basic sanity-checks.
        for my $mapkey (keys %{$mapping}) {
            my $targetkey = $mapping->{$mapkey};
            if (my $contents = delete $opts{$mapkey}) {
                if (ref($contents) ne 'ARRAY') {
                    NATE::Exceptions::Argument->throw(
                        "Option-mapping key '$mapkey' did not point to an " .
                        'array reference.'
                    );
                }
                $options{$targetkey} = clone $contents;
            } else {
                NATE::Exceptions::Argument->throw(
                    "Option-mapping key '$mapkey' did not exist in " .
                    'passed-in parameters.'
                );
            }
        }

        # OK, now %options is a table of arrayrefs, each of which will be
        # added to a create_worker() call one at a time. Make sure that we
        # have enough of them to satisfy $worker_count.

        # If $worker_count is not set, set it to the length of the longest
        # arrayref.
        if (! $worker_count) {
            $worker_count = max map { scalar @{$_} } values %options;
        }

        # If "clients" is in the list, the user wants to round-robin a
        # different set of clients than what the manager already has
        # stored. This one can be short. If so, just keep doubling it
        # until it is >= $worker_count (then splice it down).
        if ($options{client}) {
            while ($worker_count > @{$options{client}}) {
                push @{$options{client}}, @{$options{client}};
            }
            splice @{$options{client}}, $worker_count;
        }

        # Finally, make sure everything in %options is at least $worker_count
        # in length.
        my @short = grep { @{$options{$_}} < $worker_count } keys %options;
        if (@short) {
            NATE::Exceptions::Argument->throw(
                "Some mapped options (@short) do not have enough elements " .
                "to create $worker_count workers."
            );
        }
    }

    # Now create some workers:
    for my $idx (0 .. ($worker_count - 1)) {
        my %options_to_map  = map { $_ => $options{$_}->[$idx] } keys %options;
        my %options_to_use = %{clone \%opts};

        # If we're mapping event_params, it must go first so that anything else
        # that would go into it doesn't get lost when it is mapped at a later
        # point.
        if ($options_to_map{event_params}) {
            $options_to_use{event_params} =
                delete $options_to_map{event_params};
        }

        # Now copy over things one at a time, checking each one against the
        # %WORKER_NEW_OPTIONS table to see if it should be top-level or if it
        # should be pushed down to event_params.
        for my $opt (keys %options_to_map) {
            if (exists $WORKER_NEW_OPTIONS{$opt}) {
                # This belongs at the top level
                $options_to_use{$opt} = $options_to_map{$opt};
            } else {
                # This belongs in event_params
                $options_to_use{event_params}->{$opt} = $options_to_map{$opt};
            }
        }

        push @workers, $self->create_worker(%options_to_use);
    }

    return \@workers;
}

=head1 CREATING RUNID STRINGS

One of the primary things the event manager does when creating worker objects,
is create/manage the run ID strings that are eventually passed to the
B<NATE::Process> objects. These strings, which resemble file pathnames, control
the placement of logs when NATE/ntest executes your tests.

When the event manager creates a run ID, it creates it from several elements:

    base '/' sequence '_' prefix '_' event '_' decorator

=over

=item 1. Base

If the manager has a value for the C<runid_base> attribute, then the string
starts with that value followed by a C</> character.

=item 2. Sequence

The sequence number is added, unless a non-false C<no_sequence> parameter was
passed to the worker creation. If the sequence is suppressed, so is the C<_>
that follows it.

=item 3. Prefix

If the manager has a value for the C<runid_prefix> attribute, then that is
added to the string, with a trailing C<_>. If there is no prefix, there will
be no additional C<_>.

=item 4. Event

The event name is then added to the string. Since the event name is a required
value for worker creation, it will always be present in the run ID.

=item 5. Decorator

If the event name passed in looked like this:

    someEvent(additional text)

then the portion within the parentheses is considered a "decorator", and is
generally used when the same event is going to be run several times in
parallel. It is added at the end of the run ID to distinguish it from other
instances of the same event-name. A C<_> character is added between the
event name and the decorator (if there is a decorator).

=back

Remember that a run ID string is ultimately a pathname, and it is relative
to the current run ID of the test script that is executing the event manager.

To prevent invalid log-file names, any characters in the run ID that are
not alphabetic, numeric, C<_>, C<.> or C<-> will be converted to the C<_>
character. This does not apply to the C</> that immediately follows the
C<runid_base>.

=head1 CALLBACKS

The event manager enables the worker objects it creates to communicate back to
the manager by the use of callbacks. The event manager can provide these
callbacks when the workers are created, and in doing so pass in code blocks
that have the manager itself within their lexical scope. Unless a call to
B<create_worker> includes these callbacks, the manager will provide its own
default for either or both of them.

When a callback is invoked, it will be passed two positional parameters:

=over

=item $worker

This will be the B<NACL::AppBuilder::DpgWorkflows::EventWorker> object that is
invoking that callback. Through this object, you can get the ID to manage it
in the internal tables, get the B<NATE::Process> handle to check data on it,
etc.

=item $stage

This will be a string, with one of a fixed set of values, that indicates what
stage the worker is currently in:

=over

=item C<start>

The worker is current processing a B<start> command. Specifically, it is is
the C<oninit> callback of the B<NATE::Process> object it encapsulates.

=item C<finish>

The worker is currently processing the completion of the remote code.
Specifically, it is in the C<onexit> callback of the encapsulated
B<NATE::Process> object. Note that this could be either from the normal
completion of the event, or from a B<stop> call on the worker.

=back

=back

Since the stage is passed in as one of the parameters, a callback may be
designed to handle both C<start> and C<finish>. The callback stored on the
C<default_worker_callback> attribute is designed to do this. On C<start>, it
adds the worker ID to the list of C<active_workers>. On C<finish> it removes
the worker ID from that list, and deletes the worker from the C<workers> table
(unless C<keep_finished_workers> is currently set to a non-false value).

If you design and pass in your own callbacks, it is strongly recommended that
you end any such callback with an invocation of the default callback, so that
the internal tables are kept in sync. An example of this might be:

    # This call is being made within a THPL using TCD.
    # The global TCD object is $Test.
    $manager->create_worker(
        # other params...
        on_finish => sub {
            my ($worker, $stage) = @_;
            
            my $id      = $worker->id;
            my $event   = $worker->event;
            my $elapsed = $worker->elapsed;
            # Log only to the screen, not to ntest logs:
            $Test->log(
                "Worker #$id has completed event '$event'. Elapsed time: " .
                "${elapsed}s."
            );
            
            return $manager->default_worker_callback->($worker, $stage);
        }
    );

In this example, the callback logs (only to the user's terminal, not the ntest
logs) a message that the specific worker has finished its event, and uses the
B<elapsed> method on the worker to get the total time spent. It then calls the
default callback with the exact same arguments it was given.

=head1 SEE ALSO

B<NACL::AppBuilder::DpgWorkflows::CoreEvents>,
B<NACL::AppBuilder::DpgWorkflows::EventWorker>, B<NATE::Process>

=head1 AUTHOR

Randy J. Ray E<lt>rjray@netapp.comE<gt>

=cut

###############################################################################
# Private methods, not documented to the user.
###############################################################################

###############################################################################
#
#   Sub Name:       _build_client_list
#
#   Description:    Create a list of clients by doing NACL::C::Client->find
#                   and pulling out the names. This only gets called if the
#                   user doesn't specify a list of clients explicitly.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#
#   Returns:        arrayref of strings
#
###############################################################################
sub _build_client_list {
    my $self = shift;
    state @clients;

    # I'm using a state-variable to keep the list of NACL-located clients over
    # all constructor calls. Any call to NACL::C::Client->find after the first
    # call would yield the same list anyway.
    if (! @clients) {
        require NACL::C::Client;

        @clients = map { $_->client } (NACL::C::Client->find);
        if (! @clients) {
            $Log->warn(
                __PACKAGE__ . '::new: ' .
                'No client-list specified, and none found by NACL. ' .
                'Defaulting to using "localhost", which is generally ' .
                'not recommended.'
            );
            @clients = qw(localhost);
        }
    }

    return [ @clients ];
}

###############################################################################
#
#   Sub Name:       _next_client
#
#   Description:    Return one or more client names from the internal list,
#                   in a round-robin fashion.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#                   $count    in      scalar    Number of clients to return
#
#   Returns:        one or more strings
#
###############################################################################
sub _next_client {
    my $self  = shift;
    my $count = shift || 1;

    my @all_clients = $self->client_list;
    my $size = scalar @all_clients;
    my $start_idx = $self->_next_client_index;

    # Build up @all_clients until it is long-enough to splice out $count
    # elements starting at $start_idx:
    while (@all_clients < ($count + $start_idx)) {
        push @all_clients, @all_clients;
    }

    my @clients = splice @all_clients, $start_idx, $count;
    $self->_next_client_index(($start_idx + $count) % $size);

    return ($count == 1) ? $clients[0] : @clients;
}

###############################################################################
#
#   Sub Name:       _next_interface
#
#   Description:    Return one or more strings for use as the "interface"
#                   parameter(s) to worker-creation calls. Flip-flops between
#                   the values in @INTERFACES.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#                   $count    in      scalar    Number of strings to return
#
#   Returns:        one or more strings
#
###############################################################################
sub _next_interface {
    my $self  = shift;
    my $count = shift || 1;

    my $interface_index = $self->_next_interface_index;

    my @ifaces;
    for (1 .. $count) {
        push @ifaces, $INTERFACES[$interface_index];
        $interface_index = 1 - $interface_index;
    }
    # Whatever number we ended on here, will be the next index on the next
    # call:
    $self->_next_interface_index($interface_index);

    return ($count == 1) ? $ifaces[0] : @ifaces;
}

###############################################################################
#
#   Sub Name:       _create_runid
#
#   Description:    Return a string to be used as the run ID for a worker
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#
#   Returns:        string
#
###############################################################################
sub _create_runid {
    my ($self, @opts) = @_;
    my %opts = validate(
        @opts,
        {
            base        => { type => SCALAR, optional => 1 },
            prefix      => { type => SCALAR, optional => 1 },
            event       => { type => SCALAR },
            decorator   => { type => SCALAR, optional => 1 },
            no_sequence => { type => BOOLEAN, default => 0 },
        }
    );

    my $base      = $opts{base}   || $self->runid_base;
    my $prefix    = $opts{prefix} || $self->runid_prefix;
    my $decorator = $opts{decorator};
    my $event     = $opts{event};
    my $sequence  = $opts{no_sequence} ? undef : $self->_runid_sequence;

    my $string = join q{} =>
        ($sequence  ? "${sequence}_" : q{}),
        ($prefix    ? "${prefix}_"   : q{}),
        $event,
        ($decorator ? "_$decorator"  : q{});

    # Normalize any non-viable characters out of $string, so that it doesn't
    # result in an invalid file name:
    $string =~ s/[^-.\w]/_/g;

    if ($base) {
        # $base needs the same character normalization as the rest of the ID,
        # but it is allowed to have "/" characters:
        $base =~ s{[^-/.\w]}{_}g;
        $base =~ s/_+$//; # Remove any trailing _ characters
        $string = "$base/$string";
    }

    return $string;
}

###############################################################################
#
#   Sub Name:       _build_sequence_format
#
#   Description:    Return a value that can be used as this object's
#                   sequence_format attribute.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#
#   Returns:        string
#
###############################################################################
sub _build_sequence_format {
    my $self = shift;

    my $width = $self->sequence_width;

    return "%0${width}d";
}

###############################################################################
#
#   Sub Name:       _runid_sequence
#
#   Description:    Return the next runid sequence number, based on an internal
#                   table keyed by $self's ID.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#
#   Returns:        int value
#
###############################################################################
sub _runid_sequence {
    my $self = shift;

    return sprintf($self->sequence_format, ++$SequenceCounter{$self->id});
}

###############################################################################
#
#   Sub Name:       _build_default_worker_callback
#
#   Description:    Return a code ref that can be used as the callback for
#                   both the "on_start" and "on_finish" callbacks in a
#                   worker.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#
#   Returns:        code ref
#
###############################################################################
sub _build_default_worker_callback {
    my $self = shift;

    my $callback = sub {
        my ($worker, $stage) = @_;

        my $wid    = $worker->id;
        my $client = $worker->client;

        if ($stage eq 'start') {
            # Mark the start of the worker, and add the worker's ID to the
            # active_workers list.
            $Log->comment("Worker (ID $wid) started on client $client.");
            $self->_active_workers_push($wid);
        } elsif ($stage eq 'finish') {
            # Mark the completion of the worker, and remove the ID from the
            # active_workers list.
            $Log->comment("Worker (ID $wid) on client $client has finished.");
            my $idx = $self->_active_workers_first_index(sub { $_ eq $wid });
            $self->_active_workers_splice($idx, 1);
            # Remove the worker from the internal table unless the user
            # specifically requested that they be kept.
            if (! $self->keep_finished_workers) {
                $self->workers_delete($wid);
            }
        } else {
            NATE::Exceptions::Argument->throw(
                "Unknown stage value passed to callback: '$stage'"
            );
        }

        return;
    };

    return $callback;
}

###############################################################################
#
#   Sub Name:       _build_id
#
#   Description:    Return a value that can be used as this object's ID
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#
#   Returns:        int value
#
###############################################################################
sub _build_id {
    my $self = shift;

    return refaddr $self;
}

###############################################################################
#
#   Sub Name:       _parse_event
#
#   Description:    Parse the event-string into the (optional) class, the
#                   event name, and the (optional) decorator
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#                   $eventstr in      scalar    String to parse
#
#   Globals:        $EVENT_REGEX
#
#   Returns:        ($class, $event, $decorator)
#
###############################################################################
sub _parse_event {
    my ($self, $eventstr) = @_;

    my ($class, $event, $decorator);
    # This regex will set $1, $2, $3 if it matches:
    if ($eventstr =~ $EVENT_REGEX) {
        $class     = $1 || q{};
        $event     = $2;
        $decorator = $3 || q{};
    } else {
        NATE::Exceptions::Argument->throw(
            "Event specification '$eventstr' is not valid"
        );
    }

    return ($class, $event, $decorator);
}

1;
