# Copyright (c) 2013-2015 NetApp Inc.
# All rights reserved.
#
## @summary Provide a common set of libraries to support setup and cleanup on controllers, clients, switches, etc...
## @author ng-ptefit-automation-core@netapp.com
## @status private
## @pod here

=head1 NAME

NACL::AppBuilder::EnGenX

=head1 DESCRIPTION

EnGenX - Environment Generator X is a framework used to Config, Setup,
Cleanup and Deconfig your testing environment.  It has the ability
to take in XML that describes your environment and setup/cleanup
that environment for your testing purposes.

=cut

package NACL::AppBuilder::EnGenX;

# Compiler directives.
use strict;
use warnings;

# NATE and Standard Module import
use NATE 5;
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Data::Dumper;
use File::Basename;
use NATE::Util::DataTypes qw(TRUE FALSE);
use Cwd;
use Carp qw(confess);
use Tie::IxHash;
use NATE::ProcManager;
use NATE::Events qw(eventloop_drain);
use NATE::Messenger qw(messenger_global);
use NATE::ParamSet qw(param_global);
use NATE::Inc qw(find_file);
use Digest::MD5;
use NATE::Encode qw(encode_global);
use Time::HiRes;
use File::Temp qw(tempfile);

# Exceptions
use NATE::BaseException qw(:try);
use NATE::Exceptions::Argument qw(:try);
use Params::Validate
    qw(OBJECT SCALAR SCALARREF ARRAYREF HASHREF UNDEF GLOBREF);

# Load EnGenX and standard modules
use NACL::AppBuilder::EnGenX::Util;
use NACL::AppBuilder::EnGenX::Fibrebridge;
use NACL::AppBuilder::EnGenX::Resource;
use NACL::AppBuilder::EnGenX::ResourceList;
use NACL::AppBuilder::EnGenX::Controller ();
use NACL::AppBuilder::EnGenX::VPserver;
use ParamsMgr;
use NACL::AppBuilder::EnGenX::ClientCluster;
use NACL::AppBuilder::EnGenX::Client;
use NACL::AppBuilder::EnGenX::HyperVisor;
use NACL::AppBuilder::EnGenX::Switch;
use NACL::AppBuilder::EnGenX::Fabric;
use NACL::AppBuilder::EnGenX::EventMan;

# NACL libraries in use
use NACL::C::Node;
use NACL::C::Volume;

# Load San specific modules
use San::SanClient;
use San::SanSwitch;

use vars qw(@ISA @DTD_SCHEMA_FILES $ENGENX_DTD_PATH $SANMGR_DTD_PATH);

@ISA = qw(NACL::AppBuilder::EnGenX::Util);

# Exported global variables
###########################
our $TRUE  = 1;
our $FALSE = 0;

###############################################################################
#                               Public Methods
###############################################################################

=head1 METHODS

=head2 new

The following examples are the three different ways to instantiate
and EnGenX object.

(1) Create an EnGenX object and parse the given XML into a ResourceList.

    my $Egx_obj = new NACL::AppBuilder::EnGenX(file           => $Xml_file,
                             load_resources => 0);
    $Egx_obj->setup();


(2) Create an EnGenX object and use the provided resource list as input

    my $engenx_resource_list = NACL::AppBuilder::EnGenX::ResourceList->new();
    # populate the resource list object with NACL::AppBuilder::EnGenX::Resources...

    my $Egx_obj = new NACL::AppBuilder::EnGenX(resource_list => $engenx_resource_list);
    $Egx_obj->setup();

(3) Create an EnGenX object and use the Xgen configuration module to
generate a resource list dynamically based on the config.

    my $Egx_obj = new NACL::AppBuilder::EnGenX(config_module => "My::Xgen::Config::Library",
                             config_module_args => {param1 => 1, param2 => 2, ...});
    $Egx_obj->setup();

=over

=item Options

=over

=item C<< xml_file => $xml_file >>

(Optional) An XML file that describes the environment to generate.
Defaults to ""

=item C<< schema_files => \@my_dtd_files >>

(Optional) Optional ARRAY of paths to DTD files to be used to validate the XML
Defaults to files in the lib/NACL/AppBuilder/EnGenX/DTDFiles directory.

=item C<< resource_list => L<NACL::AppBuilder::EnGenX::ResourceList|lib-NACL-AppBuilder-EnGenX-ResourceList-pm> object >>

(Optional) If provided, this resource list will be used.
Defaults to undef

=item C<< config_module => "My::Config" >>

(Optional) If set, this EnGenX object will the Xgen service to generate a resource list based
upon the provided configuration module. The configuration module provided must be a derived class of L<NACL::Service::Xgen::Config|lib-NACL-Service-Xgen-Config-pm>.
The generated resource list will be stored on the engenx object.
Defaults to ""

=item C<< config_module_args => \%hashref >>

(Optional) If set with 'config_module', the provided hash reference will be fed into the Xgen
service when building the configuration defined by 'config_module'.
Defaults to {}

=item C<< load_resources => 0|1 >>

(Optional) If set will create objects early.  This is used for SanClient and
SanSwitch, etc. since creating these objects will load the resources.
Eventually as we migrate to NACL this argument will be unused.
Defaults to 1

=item C<< run_id_prefix=>"my_test_string" >>

(Optional) String that will be be used as a prefix for subtest run id strings.
Defaults to an empty string.

=item C<< enable_subtest_count=>0|1 >>

(Optional) If set, subtest run id's have an ever increasing integer suffix
Defaults to 1

=item C<< display_xml=>0|1 >>

(Optional) If set, the XML file's contents are logged.
Defaults to 1

=item C<< reconnect_client_objects_after_load=>0|1 >>

(Optional) If set, will create fresh connections to client objects after they load
Defaults to 0

=item C<< reconnect_hypervisor_objects_after_load=>0|1 >>

(Optional) If set, will create fresh connections to hypervisor objects after they load
Defaults to 0

=item C<< reconnect_switch_objects_after_load=>0|1 >>

(Optional) If set, will create fresh connections to switch objects after they load
Defaults to 0

=item C<< private_paramset=>0|1 >>

(Optional) If set, this EnGenX object will use a private paramset instead of a public paramset.
This is used for when multiple EnGenX objects are used in a single test.
Defaults to 0.

=back

=item Return

The EnGenX object created

=item Exceptions

NATE::BaseException
NATE::Exceptions::Argument

=item Environment Parameters

=over

=item C<ENGENX_ENABLE_PARAMSET_CACHING>

When enabled, will cache the paramset in the testbed's logs directory to speed
up future paramset loading of the XML file. If a cache file exists for a given
XML File, it will load the cache rather than load the XML file. Enable this
feature when using large XML Files (>10MB) with EnGenX.

Default: 0 (off)

=back

=back

=cut

########################################################################
sub new {
    $Log->enter() if $may_enter;
    my ($proto, @opts) = @_;
    my %args = ();
    my $self = {};
    my $class = ref($proto) || $proto;
    bless($self, $class);

    # Process arguments.
    try {
        %args = Params::Validate::validate( @opts, {
            file          => {type => SCALAR | GLOBREF,
                              optional => TRUE,
                              default=> undef,
                            },
            nodes          => {type => ARRAYREF,
                               optional => TRUE,
                               default=> undef,
                            },
            clusters          => {type => ARRAYREF,
                  optional => TRUE,
                  default=> undef,
                },
            xml_file      => {type => SCALAR | GLOBREF,
                              optional => TRUE,
                              default=> "",
                            },
            xml_string      => {type => SCALAR,
                                optional => TRUE,
                                default=> undef,
                            },
            resource_list => {type => HASHREF,
                              optional => TRUE,
                              default=> undef,
                            },
            configbroker_group => {type     => SCALAR,
                                   optional => TRUE,
                                   default  => "",
                                },

            # Common Arguments
            schema_files => {type     => ARRAYREF,
                             optional => TRUE,
                             default  => [],
                            },
            load_resources => {type     => SCALAR,
                               optional => TRUE,
                               default  => TRUE,
                            },
            clear_paramsmgr => {type     => SCALAR,
                               optional => TRUE,
                               default  => FALSE,
                            },
            private_paramset => {type     => SCALAR,
                               optional => TRUE,
                               default  => FALSE,
                            },
            run_id_prefix => {type     => SCALAR,
                              optional => TRUE,
                              default  => "",
                            },

            enable_subtest_count => {type     => SCALAR,
                                     optional => TRUE,
                                     default  => TRUE,
                                    },
            display_xml => {type     => SCALAR,
                            optional => TRUE,
                            default  => TRUE,
                                    },
            display_xml_runid => {type     => SCALAR,
                                  optional => TRUE,
                                  default  => "XML-environment",
                                    },
            obj_creation_timeout => { type     => SCALAR,
                                      optional => TRUE,
                                      default  => "300"},
            reconnect_client_objects_after_load => { type     => SCALAR,
                                                     optional => TRUE,
                                                     default  => FALSE },
            reconnect_hypervisor_objects_after_load => { type     => SCALAR,
                                                         optional => TRUE,
                                                         default  => FALSE },
            reconnect_switch_objects_after_load => { type     => SCALAR,
                                                     optional => TRUE,
                                                     default  => FALSE },
            reconnect_fibrebridge_objects_after_load => { type     => SCALAR,
                                                          optional => TRUE,
                                                          default  => FALSE },
            find_nodes => { type     => SCALAR,
                            optional => TRUE,
                            default  => '1' },

            # NEW ARGUMENTs
            debug_level => {type     => SCALAR,
                            optional => TRUE,
                            default  => '1',
                        },

            # Xgen integration
            config_module => { type => SCALAR,
                               optional => TRUE,
                               default => "" },
            config_module_args => {type => HASHREF,
                                   optional => TRUE,
                                   default => {}},

        });
    } otherwise {
        my $xcptn = shift;
        $Log->exit() if $may_exit;
        NATE::Exceptions::Argument->throw($xcptn->text());
    };

    # if nodes is passed in, save for load_resources reuse
    $self->{nodes} = delete $args{nodes} if defined $args{nodes};

    # if clusters is passed in, save for load_resources reuse
    $self->{clusters} = delete $args{clusters} if defined $args{clusters};

    # replace -file with -xml_file
    if (defined $args{file}) {
        $args{xml_file} = delete $args{file};
    }

    #Assign Values
    my $environment = {};
    my $xml_file = $args{xml_file};
    my $xml_string = $args{xml_string};
    my $cb_group = $args{configbroker_group};
    my $load_resources = $args{load_resources};

    # Set the run_id_prefix
    $self->{run_id_prefix}        = $args{run_id_prefix};
    $self->{subtest_count}        = 1;
    $self->{runid_prefix_count}   = 1;
    $self->{enable_subtest_count} = $args{enable_subtest_count};
    $self->{obj_creation_timeout} = $args{obj_creation_timeout};
    $self->{resources_loaded}     = FALSE;

    # allow for error stack build ups
    $self->{error_dont_overwrite_errors} = TRUE;
    $self->{error_stack_style}           = "normal";

    # get a handle to the resource group from cb_group(config broker db)
    if ((defined $cb_group) && ($cb_group ne "")) {
       $self->{rgname} = $cb_group;
       require ConfigBroker;
       $self->{cbh} = new ConfigBroker();
       $self->{rgh} = $self->{cbh}->get_resource_group_object($self->{rgname});
    }

    if($args{private_paramset}) {
        # clone the paramset
        $self->{nate_paramset} = &ParamsMgr::clone();
    }

    # Obtain the EnGenX DTD Path
    $ENGENX_DTD_PATH = $self->_get_dtd_path();
    $SANMGR_DTD_PATH = $self->_get_sanmgr_dtd_path();

    # Internal DTD XML Schema files for EnGenX
    @DTD_SCHEMA_FILES = (
        $ENGENX_DTD_PATH . "/scenario.dtd",
        $ENGENX_DTD_PATH . "/clients.dtd",
        $ENGENX_DTD_PATH . "/clusters.dtd",
        $ENGENX_DTD_PATH . "/client_clusters.dtd",
        $ENGENX_DTD_PATH . "/hypervisors.dtd",
        $ENGENX_DTD_PATH . "/fabrics.dtd",
        $ENGENX_DTD_PATH . "/eventman.dtd",
        $ENGENX_DTD_PATH . "/fibrebridges.dtd",
        $ENGENX_DTD_PATH . "/vpservers.dtd"
    );

    my ($tmp_xml_fh, $tmp_xml_file_name);
    if (defined $xml_string and $xml_string ne "") {
        $Log->comment("Creating a tempory file and filehandle from xml_string provided.");
        # create a temporary xml file to work with
        ($tmp_xml_fh, $tmp_xml_file_name) = tempfile();
        print $tmp_xml_fh $xml_string;
        $xml_file = $tmp_xml_fh;
    }

    if (!$xml_file and !$args{resource_list} and !$args{config_module}) {
        NATE::Exceptions::Argument->throw(
            "EnGenX::new(): You must provide either a path to an XML file, an ".
            "NACL::AppBuilder::EnGenX::ResourceList object, or an Xgen config_module. Could not find either from the " .
            "provided params \n". Dumper(%args)
        );
    }

    # If user provided an XML file
    if (ref($xml_file) =~ /GLOB/i or $xml_file ne "") {
        $Log->comment("Creating EnGenX: Using provided xml file and ENGENX_DTD_PATH : "
                      . &ParamsMgr::get("ENGENX_DTD_PATH",
                                        paramset => $self->NACL::AppBuilder::EnGenX::Util::get_paramset()));

        unless (ref($xml_file) =~ /GLOB/i) {
            # get the filepath of the xml_file
            $xml_file = $self->_get_xml_path($xml_file);
        }


        # Initialize the resource list.
        $self->{resource_list} = new NACL::AppBuilder::EnGenX::ResourceList();
        if (ref($xml_file) =~ /GLOB/i) {
            $Log->comment("Loading XML filehandle");
            seek($xml_file,0,0);
        } else {
            $Log->comment("Loading XML file $xml_file");
        }

        # Before we load the new XML files, check to see if we need to clear
        # out the ParamSet on ParamsMgr
        if ($args{clear_paramsmgr}) {
            &ParamsMgr::clear_global_paramset();
            if (&ParamsMgr::error()) {
                $Log->exit() if $may_exit;
                my $errmsg = join "\n", &ParamsMgr::get_error();
                NATE::BaseException->throw("Failed to clear ParamsMgr: $errmsg");
            }
        }

        # check for existance of a paramset cache for this xml file, load that
        # instead of going through the trouble of loading the XML
        my $hexdigest;
        # by default the paramset caching algorithm is disabled
        my $paramset_caching = param_global()->get("ENGENX_ENABLE_PARAMSET_CACHING", default => 0);
        if (!$paramset_caching or !$self->_load_cached_paramset(xml_file => $xml_file,
                                                                schema_files => $args{schema_files},
                                                                hexdigest => \$hexdigest)) {
            # Process and Validate the xml file and any override schema files
            $self->load_xml(
                -xml_files         => [$xml_file],
                -schema_files      => $args{schema_files},
                -display_xml       => $args{display_xml},
                -display_xml_runid => $self->_get_next_runid_prefix() . "_" . $args{display_xml_runid},
            );

            if ($paramset_caching) {
                # cache the paramset for future EnGenX constructor calls
                # use the hex digest that we already calculated
                $self->_cache_paramset(xml_file => $xml_file,
                                       schema_files => $args{schema_files},
                                       hexdigest => $hexdigest);
            }
        }

        # TODO: Try to change the following logic for something equivalent using
        #       Params::Validate.
        # Get the set of environment parameters from ParamsMgr
        &ParamsMgr::process_args(%args);
        $environment =
            &ParamsMgr::search(search_path => [], key => "ENVIRONMENT",
                               paramset => $self->NACL::AppBuilder::EnGenX::Util::get_paramset());
        if (&ParamsMgr::error()) {
           $Log->exit() if $may_exit;
           my $errmsg = join "\n", &ParamsMgr::get_error();
            NATE::Exceptions::Argument->throw($errmsg);
        }
        #
        # End TODO
    }

    close($xml_file) if (ref($xml_file) =~ /GLOB/i);

    # see if there is a config module provided
    if ($args{config_module}) {
        $Log->comment("Generating resource list from Xgen config module " . $args{config_module});
        require NACL::ServiceAPI::Xgen;
        my $xgen = NACL::ServiceAPI::Xgen->new();
        my $list = $xgen->build_config(config_module => $args{config_module},
                                       %{$args{config_module_args}},
                                       client_formatter_module => 'NACL::AppBuilder::EnGenX::XgenFormatter');
        # update the resource list field
        $args{resource_list} = $list;
    }

    # If user provided a ResourceList Object
    if ( defined($args{resource_list}) ) {
        $Log->comment("Creating EnGenX: Using provided NACL::AppBuilder::EnGenX::ResourceList");

        if ($self->{resource_list} and $self->{resource_list}->isa('NACL::AppBuilder::EnGenX::ResourceList')) {
            $Log->comment("Merging resource lists...");
            $self->{resource_list}->merge_with_resource_list(resource_list => $args{resource_list});
        } else {
            $self->{resource_list} = $args{resource_list};
        }
        $environment->{DEBUG_LEVEL} = $args{debug_level};
    }

    # TODO: In the future when we are only using Params::Validate, we can
    #       stop setting the ParamsMgr::debug_level parameter, we could delete
    #       that line all together, because there will be no class or methos
    #       using ParamasMgr.
    #       Also, verify if SubIpc is even used. We are using instead
    #       NACL::AppBuilder::EnGenX:: Subprocess, but SanClient or other SanMgr-inherited classes
    #       could be using it.

    # Save the debug level and set it in the objects
    my $dbg_level = defined($environment->{DEBUG_LEVEL}) ?
                                $environment->{DEBUG_LEVEL} : $args{debug_level};

    # Set the $Log accodingly.
    if ($Log->may_trace()) {
        $dbg_level = 3;
    }
    elsif ($Log->may_debug()) {
        $dbg_level = 7;
    }

    # Set the internal variable value.
    $self->debug_level(-debug_level => $dbg_level);

    # Set the debug level on ParamsMgr
    &ParamsMgr::debug_level(-debug_level => $dbg_level);
    if (&ParamsMgr::error()) {
        $Log->exit() if $may_exit;
        my $errmsg = join "\n", &ParamsMgr::get_error();
        NATE::BaseException->throw(
            "Failed to set debug level for ParamsMgr: $errmsg");
    }

    #
    # End TODO

    # Create local objects for each of the resource types
    # these are to be used later for config, setup, validate and cleanup work.
    $self->{egx_controller_object}     = NACL::AppBuilder::EnGenX::Controller->_new(-parent => $self);
    $self->{egx_fibrebridge_object}    = NACL::AppBuilder::EnGenX::Fibrebridge->_new(-parent => $self);
    $self->{egx_fabric_object}         = NACL::AppBuilder::EnGenX::Fabric->new(parent => $self);
    $self->{egx_eventman_object}       = NACL::AppBuilder::EnGenX::EventMan->new(parent => $self);
    $self->{egx_client_cluster_object} = NACL::AppBuilder::EnGenX::ClientCluster->_new(-parent => $self);
    $self->{egx_client_object}         = NACL::AppBuilder::EnGenX::Client->new(parent => $self);
    $self->{egx_switch_object}         = NACL::AppBuilder::EnGenX::Switch->new(parent => $self);
    $self->{egx_hypervisor_object}     = NACL::AppBuilder::EnGenX::HyperVisor->new(parent => $self);
    $self->{egx_vpserver_object}       = NACL::AppBuilder::EnGenX::VPserver->new(parent => $self);

    # Translate the XML stored by the ParamsMgr to a Resource list
    if (ref($xml_file) =~ /GLOB/i or length($xml_file)) {
        $self->xml_to_list();
    }

    # If load_resources is set, load up objects for the resources
    if ($load_resources == 1) {
        $self->load_resources(%args);
    } else {
        $Log->comment("No resource objects loaded load_resources => 0");
    }

    # start the vs_client_relationship_svc
    $self->_init_vs_client_relationship_svc();

    # load the fabric module's resources
    if ($load_resources == 1) {
        # fabric resource has to happen after the vs client relationship svc has initialized
        $self->{egx_fabric_object}->load_resource();
    }

    # Setup storage hash for used ips for lifs.
    $self->{used_ips} = ();

    $Log->exit() if $may_exit;
    return $self;
} ## end sub new

=head2 config

Configure each resource type (controllers, clients, switches, etc.)
This method is usually not called in scripts because setup calls config.

    $Egx_obj->config();

=over

=item Options

TBD

=over

=item C<< TBD >>

TBD

=back

=item Return

TBD

=item Exceptions

BaseException

=back

=cut

###############################################################################
sub config {
    my ($self) = @_;

    my ($st_status, $subtest);

    # Get the set of actions parameters from ParamsMgr
    my $actions = &ParamsMgr::search(search_path => [], key => "ACTIONS",
                                     paramset => $self->NACL::AppBuilder::EnGenX::Util::get_paramset());
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return FALSE;
    }

    # Check to see if user wants to run config at all
    if (defined($actions->{CONFIG}) and $actions->{CONFIG} == 0) {
        $Log->comment("Skipping config due to actions flag set to \"0\"");
        return TRUE;
    }

    # Verify resources have been loaded.
    $self->_verify_resources_loaded();

    ##############################################
    # Controller setup
    ##############################################
    $self->{egx_controller_object}->setup(
        run_id_prefix => $self->{run_id_prefix},
        resource_list => \%{$self->{resource_list}}
    );

    ##############################################
    # Hypervisor config
    ##############################################
    $self->{egx_hypervisor_object}->config(run_id_prefix => $self->{run_id_prefix});

    ##############################################
    # Client config
    ##############################################
    $self->{egx_client_object}->config(run_id_prefix => $self->{run_id_prefix});

    ##############################################
    # Fabric config
    ##############################################
    $self->{egx_fabric_object}->config(run_id_prefix => $self->{run_id_prefix});

    return TRUE;
} ## end sub config

=head2 setup

Setup each resource type (controllers, clients, switches, etc.)
Setup calls config().

    $Egx_obj->setup();

=over

=item Options

TBD

=over

=item C<< TBD >>

TBD

=back

=item Return

TBD

=item Exceptions

BaseException

=back

=cut

###############################################################################
sub setup {
    my ($self, %args) = @_;

    my ($st_status, $subtest, $controller_objs);
    my (@messages) = ();

    my $resource_list = $self->{resource_list};

    # first look for actions within the resource list
    my $actions;
    my $actions_arr = $resource_list->get_resources(category => qr/^actions/i,
                                                    type     => qr/^attributes$/i);

    if (!$actions_arr or !scalar(@{$actions_arr})) {
        # Get the set of actions parameters from ParamsMgr
        $actions = &ParamsMgr::search(search_path => [],
                                      key => "ACTIONS",
                                      paramset => $self->NACL::AppBuilder::EnGenX::Util::get_paramset());
        if (&ParamsMgr::error()) {
            my $errmsg = &ParamsMgr::get_error();
            NATE::BaseException->throw("Call to ParamsMgr::search failed: $errmsg");
        }
    } else {
        $actions = $actions_arr->[0];
    }

    # Check to see if user wants to run setup at all
    if (defined($actions->{SETUP}) and $actions->{SETUP} == 0) {
        $Log->comment("Skipping setup due to actions flag set to \"0\"");
        return $NACL::AppBuilder::EnGenX::TRUE;
    }

    $Log->comment("Starting EnGenX setup");

    # Verify resources have been loaded.
    $self->_verify_resources_loaded();

    ##############################################
    # Switch setup
    ##############################################
    $self->{egx_switch_object}->setup(-run_id_prefix => $self->{run_id_prefix});

    my $start;
    my $difference;
    try{
        $EnGenXUtil::OBJ_COUNT = 0;
        $start = [Time::HiRes::gettimeofday()];
        ##############################################
        # Cluster setup
        ##############################################
        $self->{egx_controller_object}->setup(
            run_id_prefix => $self->{run_id_prefix},
            resource_list => \%{$self->{resource_list}}
        );
    } finally {
        if($EnGenXUtil::OBJ_COUNT) {
            $difference = Time::HiRes::tv_interval($start);
            $self->_log_performance(time_difference => $difference,
                                    action          => 'setup');
        }
    };

    ###########################################################################
    # TODO: Remove CONFIG code once controller config is properly implemented #
    ###########################################################################

    ##############################################
    # Hypervisor config
    ##############################################
    $self->{egx_hypervisor_object}->config(run_id_prefix => $self->{run_id_prefix});

    ##############################################
    # Client config
    ##############################################
    $self->{egx_client_object}->config(run_id_prefix => $self->{run_id_prefix});

    ##############################################
    # Fabric config
    ##############################################
    $self->{egx_fabric_object}->config(run_id_prefix => $self->{run_id_prefix});

    ##############################################
    # VPserver setup
    ##############################################
    $self->{egx_vpserver_object}->setup(run_id_prefix => $self->{run_id_prefix});

    ##############################################
    # HyperVisor setup
    ##############################################
    $self->{egx_hypervisor_object}->setup(run_id_prefix => $self->{run_id_prefix});

    ##############################################
    # Client setup
    ##############################################
    $self->{egx_client_object}->setup(run_id_prefix => $self->{run_id_prefix});

    ##############################################
    # Client cluster setup
    ##############################################
    $self->{egx_client_cluster_object}
        ->setup(-run_id_prefix => $self->{run_id_prefix});
    if ($self->{egx_client_cluster_object}->error()) {
        $self->_set_err("Client Cluster Setup failed",
            $self->{egx_client_cluster_object}->errmsg());
        return $NACL::AppBuilder::EnGenX::FALSE;
    }

    return $NACL::AppBuilder::EnGenX::TRUE;
} ## end sub setup

=head2 cleanup

Cleanup each resource type (controllers, clients, switches, etc.)
Cleanup calls deconfig()

    $Egx_obj->cleanup();

=over

=item Options

TBD

=over

=item C<< TBD >>

TBD

=back

=item Return

TBD

=item Exceptions

BaseException

=back

=cut

###############################################################################
sub cleanup {
    my ($self) = @_;

    my ($st_status, $subtest);

    my $resource_list = $self->{resource_list};

    # first look for actions within the resource list
    my $actions;
    my $actions_arr = $resource_list->get_resources(category => qr/^actions/i,
                                                    type     => qr/^attributes$/i);

    if (!$actions_arr or !scalar(@{$actions_arr})) {
        # Get the set of actions parameters from ParamsMgr
        $actions = &ParamsMgr::search(search_path => [],
                                      key => "ACTIONS",
                                      paramset => $self->NACL::AppBuilder::EnGenX::Util::get_paramset());
        if (&ParamsMgr::error()) {
            $self->_set_err(&ParamsMgr::get_error());
            return $NACL::AppBuilder::EnGenX::FALSE;
        }
    } else {
        $actions = $actions_arr->[0];
    }

    # Check to see if user wants to run cleanup at all
    if (defined($actions->{CLEANUP}) and $actions->{CLEANUP} == 0) {
        $Log->comment("Skipping cleanup due to actions flag set to \"0\"");
        return $NACL::AppBuilder::EnGenX::TRUE;
    }

    # Verify resources have been loaded.
    $self->_verify_resources_loaded();

    ##############################################
    # Client Cluster cleanup
    ##############################################
    $self->{egx_client_cluster_object}
        ->cleanup(-run_id_prefix => $self->{run_id_prefix});
    if ($self->{egx_client_cluster_object}->error()) {
        $self->_set_err("Client Cluster Cleanup failed",
            $self->{egx_client_cluster_object}->errmsg());
        return $NACL::AppBuilder::EnGenX::FALSE;
    }
    ##############################################
    # Client cleanup
    ##############################################
    $self->{egx_client_object}->cleanup(run_id_prefix => $self->{run_id_prefix});

    ##############################################
    # HyperVisor cleanup
    ##############################################
    $self->{egx_hypervisor_object}->cleanup(run_id_prefix => $self->{run_id_prefix});

    ##############################################
    # VPserver cleanup
    ##############################################
    $self->{egx_vpserver_object}->cleanup(run_id_prefix => $self->{run_id_prefix});

    ##############################################
    # Fabric deconfig (hacked for now)
    ##############################################
    $self->{egx_fabric_object}->deconfig(run_id_prefix => $self->{run_id_prefix});

    ######################################################################
    # TODO: Remove DECONFIG code once controller deconfig is implemented #
    ######################################################################

    ##############################################
    # Client deconfig
    ##############################################
    $self->{egx_client_object}->deconfig(run_id_prefix => $self->{run_id_prefix});

    ##############################################
    # HyperVisor deconfig
    ##############################################
    $self->{egx_hypervisor_object}->deconfig(run_id_prefix => $self->{run_id_prefix});

    my $start;
    my $difference;
    try{
        $EnGenXUtil::OBJ_COUNT = 0;
        $start = [Time::HiRes::gettimeofday()];
        ##############################################
        # Controller cleanup
        ##############################################
        $self->{egx_controller_object}->cleanup(
            run_id_prefix => $self->{run_id_prefix},
            resource_list => \%{$self->{resource_list}}
        );
    } finally {
        if($EnGenXUtil::OBJ_COUNT) {
            $difference = Time::HiRes::tv_interval($start);
            $self->_log_performance(time_difference => $difference,
                                    action          => 'cleanup');
        }
    };
    return $NACL::AppBuilder::EnGenX::TRUE;
} ## end sub cleanup

=head2 deconfig

Deconfiguure each resource type (controllers, clients, switches, etc.)
This method is usually not called in scripts because cleanup calls deconfig.

    $Egx_obj->deconfig();

=over

=item Options

TBD

=over

=item C<< TBD >>

TBD

=back

=item Return

TBD

=item Exceptions

BaseException

=back

=cut

###############################################################################
sub deconfig {
    my ($self) = @_;

    # Get the set of actions parameters from ParamsMgr
    my $actions = &ParamsMgr::search(search_path => [], key => "ACTIONS",
                                     paramset => $self->NACL::AppBuilder::EnGenX::Util::get_paramset());
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $NACL::AppBuilder::EnGenX::FALSE;
    }

    # Check to see if user wants to run deconfig at all
    if (defined($actions->{DECONFIG}) and $actions->{DECONFIG} == 0) {
        $Log->comment("Skipping deconfig due to actions flag set to \"0\"");
        return TRUE;
    }

    # Verify resources have been loaded.
    $self->_verify_resources_loaded();

    ##############################################
    # Fabric deconfig
    ##############################################
    $self->{egx_fabric_object}->deconfig(run_id_prefix => $self->{run_id_prefix});

    ##############################################
    # Client deconfig
    ##############################################
    $self->{egx_client_object}->deconfig(run_id_prefix => $self->{run_id_prefix});

    ##############################################
    # HyperVisor deconfig
    ##############################################
    $self->{egx_hypervisor_object}->deconfig(run_id_prefix => $self->{run_id_prefix});

    ##############################################
    # Cluster deconfig
    ##############################################
    $self->{egx_controller_object}->cleanup(
        run_id_prefix => $self->{run_id_prefix},
        resource_list => \%{$self->{resource_list}}
    );

    return TRUE;
} ## end sub deconfig

=head2 validate

Currently not implemented.

=over

=item Options

TBD

=over

=item C<< TBD >>

TBD

=back

=item Return

TBD

=item Exceptions

BaseException

=back

=cut

###############################################################################
sub validate {
    my ($self) = @_;

    return TRUE;
} ## end sub validate

=head2 load_resources

Go through all resources provided by caller (XML file) and create
necessary objects for each to be stored on EnGenX for later use

    $Egx_obj->load_resources();

=over

=item Options

TBD

=over

=item C<< TBD >>

TBD

=back

=item Return

TBD

=item Exceptions

BaseException

=back

=cut

###############################################################################
sub load_resources {
    my ($self, %args) = @_;
    $Log->enter() if $may_enter;

    my $reconnect_clients         = $args{"reconnect_client_objects_after_load"}         // TRUE;
    my $reconnect_client_clusters = $args{"reconnect_client_cluster_objects_after_load"} // TRUE;
    my $reconnect_hypervisors     = $args{"reconnect_hypervisor_objects_after_load"}     // FALSE;
    my $reconnect_switches        = $args{"reconnect_switch_objects_after_load"}         // FALSE;
    my $reconnect_fibrebridges    = $args{"reconnect_fibrebridge_objects_after_load"}    // FALSE;
    my $find_nodes                = $args{"find_nodes"}                                  // 1;

    # load client/client clusters/hypervisor/switch resources in the background
    my ($client_listener, $client_flistener, @client_procs) = $self->{egx_client_object}->load_resource(%args);

    my ($hypervisor_listener, $hypervisor_flistener, @hypervisor_procs) = $self->{egx_hypervisor_object}->load_resource(%args);

    my ($switch_listener, $switch_flistener, @switch_procs) = $self->{egx_switch_object}->load_resource(%args);

    my ($atto_listener, $atto_flistener, @atto_procs) = $self->{egx_fibrebridge_object}->load_resource(%args);
    if ($self->{egx_fibrebridge_object}->error()) {
        my $errmsg = $self->{egx_fibrebridge_object}->errmsg();
        NATE::BaseException->throw("Failed to load fibrebridge module resources\n$errmsg");
    }

    # initialize the listeners complete count to 0
    $self->{listeners_complete_count} = 0;

    # install the listeners for each resource type
    NATE::Events::listener_add($client_listener);
    NATE::Events::listener_add($client_flistener);

    NATE::Events::listener_add($hypervisor_listener);
    NATE::Events::listener_add($hypervisor_flistener);

    NATE::Events::listener_add($switch_listener);
    NATE::Events::listener_add($switch_flistener);

    NATE::Events::listener_add($atto_listener);
    NATE::Events::listener_add($atto_flistener);

    # set up the proc manager to spawn off parallel processes to load each individual module
    my @all_procs = (@client_procs, @hypervisor_procs, @switch_procs, @atto_procs);
    my $pm;
    if (scalar(@all_procs)) {
        my $runid_prefix = $self->_get_next_runid_prefix();
        $pm = NATE::ProcManager->new(proc_info => \@all_procs,
                                     runid => "${runid_prefix}_load_objects");
        # start the processes
        $pm->start();
    }

    # load the controller modules resources
    if ($find_nodes) {
        try {
            $self->{egx_controller_object}->_load_resource();
        } otherwise {
            my $e = shift;
            push(@{$self->{load_errors}}, "Exception caught during controller load resource call: " .
                     $e->text());
        };
    } else {
        $Log->comment('Skipping loading nodes');
    }

    try {
        $self->{egx_vpserver_object}->load_resource();
    } otherwise {
         my $e = shift;
         $self->_set_err("Exception caught during vpserver load resource call: " .
                         $e->text());
    };

    # reap and cleanup the processes
    if (scalar(@all_procs)) {
        $pm->wait();
        $pm->destroy();
    }

    # burt656995: keep processing listeners until all the listeners have completed
    # it might happen that we remove the listeners prematurely before all the resources are loaded.
    my $total_proc_count = scalar(@all_procs);
    my $object_creation_timeout_time = time + $self->{obj_creation_timeout};
    while ($total_proc_count > $self->{listeners_complete_count}) {
        NATE::Events::eventloop_drain(timeout => 1);

        # break out in case all the listeners completed after the event loop
        last if ($self->{listeners_complete_count} >= $total_proc_count);

        if (time > $object_creation_timeout_time) {
            NATE::BaseException->throw("The Load Objects task exceeded the timeout '$self->{obj_creation_timeout}'");
            last;
        }
    }

    # clean up the listeners
    NATE::Events::listener_remove($client_listener);
    NATE::Events::listener_remove($client_flistener);

    NATE::Events::listener_remove($hypervisor_listener);
    NATE::Events::listener_remove($hypervisor_flistener);

    NATE::Events::listener_remove($switch_listener);
    NATE::Events::listener_remove($switch_flistener);

    NATE::Events::listener_remove($atto_listener);
    NATE::Events::listener_remove($atto_listener);

    # check if any of the individual modules hit any errors
    my $fail_count = 0;

    # if the number of registered clients and procs dont match something bad happened inside a load objects
    # process
    if ($self->{egx_client_obj}->{load_errors}) {
        push(@{$self->{load_errors}}, @{$self->{egx_client_obj}->{load_errors}});
    }
    my $num_client_procs = @client_procs;
    my $num_client_objs = scalar(keys %{$self->{_objects}->{client}});
    $fail_count = $num_client_procs - $num_client_objs;
    if ($fail_count > 0) {
        push(@{$self->{load_errors}}, "Detected that $fail_count out of $num_client_procs client " .
                                      "load object processes failed to initialize.");
    }

    # if the number of registered hypervisors and procs dont match something bad happened inside a load objects
    # process
    if ($self->{egx_hypervisor_object}->{load_errors}) {
        push(@{$self->{load_errors}}, @{$self->{egx_hypervisor_object}->{load_errors}});
    }
    my $num_hypervisor_procs = @hypervisor_procs;
    my $num_hypervisor_objs = scalar(keys %{$self->{_objects}->{hypervisor}});
    $fail_count = $num_hypervisor_procs - $num_hypervisor_objs;
    if ($fail_count > 0) {
        push(@{$self->{load_errors}}, "Detected that $fail_count out of $num_hypervisor_procs " .
                                      "hypervisor load object processes failed to initialize.");
    }

    # if the number of registered switchs and procs dont match something bad happened inside a load objects
    # process
    if ($self->{egx_switch_obj}->{load_errors}) {
        push(@{$self->{load_errors}}, @{$self->{egx_switch_obj}->{load_errors}});
    }
    my $num_switch_procs = @switch_procs;
    my $num_switch_objs = scalar(keys %{$self->{_objects}->{switch}});
    $fail_count = $num_switch_procs - $num_switch_objs;
    if ($fail_count > 0) {
        push(@{$self->{load_errors}},"Detected that $fail_count out of $num_switch_procs switch " .
                 "load object processes failed to initialize.");
    }

    # if the number of registered fibrebridges and procs dont match something bad happened inside a load
    # objects process
    if ($self->{egx_fibrebridge_object}->error()) {
        push(@{$self->{load_errors}}, "Encountered an error while loading fibrebridge objects",
             $self->{egx_fibrebridge_object}->errmsg());
    }
    my $num_atto_procs = @atto_procs;
    my $num_atto_objs = scalar(keys %{$self->{_objects}->{fibrebridge}});
    $fail_count = $num_atto_procs - $num_atto_objs;
    if ($fail_count > 0) {
        push(@{$self->{load_errors}}, "Detected that $fail_count out of $num_atto_procs fibrebridge " .
                 "load object processes failed to initialize.");
    }

    # catch all errors here and throw an exception here
    if ($self->{load_errors}) {
        my $errmsg = join("\n", @{$self->{load_errors}});
        NATE::BaseException->throw("The load_objects task failed, please investigate\n$errmsg");
    }

    # reconnect all client/hypervisor/switch objects if requested
    if ($reconnect_clients) {
        $self->reconnect_all_clients();
    }
    if ($reconnect_hypervisors) {
        my @objs = $self->get_objects(-type => "hypervisor");
        foreach my $hypervisor_obj (@objs) {
            $hypervisor_obj->reconnect();
            if ($hypervisor_obj->error()) {
                NATE::BaseException->throw("Failed to reconnect to hypervisor " . $hypervisor_obj->hostname(),
                                           $hypervisor_obj->errmsg());
            }
        }
    }
    if ($reconnect_switches) {
        my @objs = $self->get_objects(-type => "switches");
        foreach my $switch_obj (@objs) {
            $switch_obj->reconnect();
            if ($switch_obj->error()) {
                NATE::BaseException->throw("Failed to reconnect to switch " . $switch_obj->hostname(),
                                           $switch_obj->errmsg());
            }
        }
    }
    if ($reconnect_fibrebridges) {
        my @objs = $self->get_objects(-type => "fibrebridges");
        foreach my $atto_obj (@objs) {
            $atto_obj->reconnect();
            if ($atto_obj->error()) {
                NATE::BaseException->throw("Failed to reconnect to fibrebridge " . $atto_obj->get_hostname(),
                                           $atto_obj->errmsg());
            }
        }
    }

    $self->{egx_client_cluster_object}->load_resource(%args);
    if ($self->{egx_client_cluster_object}->error()) {
        my $errmsg = $self->{client_cluster_object}->errmsg();
        NATE::BaseException->throw("Failed to load client cluster module resources\n$errmsg");
    }

    # finally, if no error was encountered, mark LOAD_RESOURCES as executed
    $self->{resources_loaded} = TRUE;
    $Log->exit() if $may_exit;

    return TRUE;
} ## end sub load_resources

=head2 initialize_eventman

Initialize the eventman object via the XML file/schema specified

    $Egx_obj->initialize_eventman();

=over

=item Options

TBD

=over

=item C<< TBD >>

TBD

=back

=item Return

TBD

=item Exceptions

BaseException

=back

=cut

###############################################################################
sub initialize_eventman {
    my ($self, %args) = @_;

    # initialize the eventman module
    # this will dynamically build the eventman object
    # using elements from the ParamSet that will
    # define the api calls to invoke
    $self->{egx_eventman_object}->initialize();

    # unset the eventman tags from the paramset
    &ParamsMgr::unset("EVENTMAN");
    if (&ParamsMgr::error()) {
        NATE::BaseException->throw("Failed to remove EVENTMAN from the NATE ParamSet " .
            &ParamsMgr::get_error());
    }
}

=head2 load_xml

Load XML file(s) into ParamSet via ParamsMgr

    $Egx_obj->load_xml();

=over

=item Options

TBD

=over

=item C<< TBD >>

TBD

=back

=item Return

TBD

=item Exceptions

BaseException

=back

=cut

###############################################################################
sub load_xml {
    $Log->enter() if $may_enter;
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);
    my @xml_files = &ParamsMgr::get_arg(
        -arg      => "xml_files",
        -required => 1,
        -ref_type => "ARRAY"
    );
    my @override_schema_files = &ParamsMgr::get_arg(
        -arg      => "schema_files",
        -ref_type => "ARRAY",
        -default  => []
    );
    my $display_xml = &ParamsMgr::get_arg(
        -arg     => "display_xml",
        -default => 1
    );
    my $runid = &ParamsMgr::get_arg(
        -arg     => "display_xml_runid",
        -default => "XML-environment"
    );

    if (&ParamsMgr::error()) {
        $Log->exit() if $may_exit;
        my $errmsg = join "\n", &ParamsMgr::get_error();
        NATE::BaseException->throw("Argument parsing failed: $errmsg");
    }

    # Use the EnGenX internal schema files for each module
    # if requested to (default)
    my @schema_files = @DTD_SCHEMA_FILES;

    # if any override schema files were specified, push them on the list
    if (scalar(@override_schema_files)) {
        push(@schema_files, @override_schema_files);
    }

    # validate & load each xml_file into ParamsMgr
    foreach my $xml_file (@xml_files) {

        # Validate the XML file against the provided schema file list
        my $xml_string = &ParamsMgr::validate_and_process_xml(
            -schema_files => \@schema_files,
            -xml_file     => $xml_file
        );
        if (&ParamsMgr::error()) {
            $Log->exit() if $may_exit;
            my $errmsg = join "\n", &ParamsMgr::get_error();
            NATE::BaseException->throw(
                "Failed to validate XML file: " . "$xml_file: $errmsg");
        }

       # Load the xml string that was generated by validate_xml into paramsmgr
        &ParamsMgr::load_xml(
            -xml_string        => $xml_string,
            -display_xml       => $display_xml,
            -display_xml_runid => $runid,
            paramset           => $self->NACL::AppBuilder::EnGenX::Util::get_paramset()
        );
        if (&ParamsMgr::error()) {
            $Log->comment(
                "Failed to load XML String" . &ParamsMgr::get_error());
            $Log->exit() if $may_exit;
            my $errmsg = join "\n", &ParamsMgr::get_error();
            NATE::BaseException->throw("Failed to load XML string: $errmsg");
        }
    }

    $Log->exit() if $may_exit;
    return TRUE;
} ## end sub load_xml

=head2 xml_to_list

Translate the XML stored by ParamsMgr into a ResourceList

    $Egx_obj->xml_to_list();

=over

=item Options

None

=item Return

Nothing, resource_list attribure is populated

=item Exceptions

BaseException

=back

=cut

###############################################################################
sub xml_to_list {
    my ($self) = @_;
    $Log->enter() if $may_enter;

    $Log->comment("Translating XML to Resource List");

    # store the environment as a resource
    my $environment = &ParamsMgr::search(search_path => [], key => "ENVIRONMENT",
                                         paramset => $self->NACL::AppBuilder::EnGenX::Util::get_paramset());
    if (&ParamsMgr::error()) {
        NATE::BaseException->throw("Could not obtain Environent element from the paramset");
    }
    $self->resource_list()->add_resource(category => "ENVIRONMENT",
                                         type => "environment",
                                         data => $environment);

    # process each modules xml to list routines
    $self->{egx_controller_object}->xml_to_list();
    $self->{egx_client_object}->xml_to_list();
    $self->{egx_client_cluster_object}->xml_to_list();
    $self->{egx_hypervisor_object}->xml_to_list();
    $self->{egx_switch_object}->xml_to_list();
    $self->{egx_fabric_object}->xml_to_list();
    $self->{egx_vpserver_object}->xml_to_list();

    $Log->comment("Completed building the Resource List from the XML");

    $Log->exit() if $may_exit;
} ## end sub xml_to_list

###############################################################################
# Method:    _get_dtd_path
# Objective: Helper method that generates the DTD File path based on
#            environment variables.
# Details:   None. Private method.
###############################################################################
sub _get_dtd_path {
    my ($self) = @_;

    # (1) default to the non null value of SANMGR_DTD_PATH
    &Tharn::param("ENGENX_DTD_PATH", "-default", "");
    my $dtd_path = &ParamsMgr::get("ENGENX_DTD_PATH", paramset => $self->NACL::AppBuilder::EnGenX::Util::get_paramset());
    if (defined($dtd_path) and $dtd_path !~ /^\s*$/) {
        return $dtd_path;
    }

    # (2) try to determine from NATE_LIB/NATE_LIB_UNIX
    my @natelib;
    if (exists($ENV{"NATE_LIB"})) {
        @natelib = split(/;/, $ENV{"NATE_LIB"});
    }
    if (not scalar(@natelib) and exists($ENV{"NATE_LIB_UNIX"})) {
        @natelib = split(/;/, $ENV{"NATE_LIB_UNIX"});
    }

    # walk through all the natelib paths until we find one that works
    foreach my $natelib_path (@natelib) {
        $dtd_path = $natelib_path . "/lib/NACL/AppBuilder/EnGenX/DTDFiles";
        if (-d $dtd_path) {
            &ParamsMgr::set(key => "ENGENX_DTD_PATH", value => $dtd_path, paramset => $self->NACL::AppBuilder::EnGenX::Util::get_paramset());
            return $dtd_path;
        }
    }

    # (3) try to determine the path based on the TESTNAME variable
    my $testname = $ENV{"TESTNAME"};
    if ($testname =~ /^(\/.+?\/p4\/.+?\/)/) {
        $dtd_path = $1 . "lib/NACL/AppBuilder/EnGenX/DTDFiles";

        # make sure this path exists, otherwise dont use it
        if (-d $dtd_path) {
            &ParamsMgr::set(key => "ENGENX_DTD_PATH", value => $dtd_path, paramset => $self->NACL::AppBuilder::EnGenX::Util::get_paramset());
            return $dtd_path;
        }
    }

    # (4) otherwise, default to sanrtest's version
    $dtd_path = "/u/sanrtest/p4/test/lib/NACL/AppBuilder/EnGenX/DTDFiles";
    &ParamsMgr::set(key => "ENGENX_DTD_PATH", value => $dtd_path, paramset => $self->NACL::AppBuilder::EnGenX::Util::get_paramset());

    return $dtd_path;
} ## end sub _get_dtd_path

###############################################################################
# Method:    _get_sanmgr_dtd_path
# Objective: Helper method that generates the DTD File path based on
#            environment variables.
# Details:   None. Private method.
###############################################################################
sub _get_sanmgr_dtd_path {
    my ($self) = @_;

    # (1) default to the non null value of SANMGR_DTD_PATH
    &Tharn::param("SANMGR_DTD_PATH", "-default", "");
    my $dtd_path = &ParamsMgr::get("SANMGR_DTD_PATH", paramset => $self->NACL::AppBuilder::EnGenX::Util::get_paramset());
    if (defined($dtd_path) and $dtd_path !~ /^\s*$/) {
        return $dtd_path;
    }

    # (2) try to determine from NATE_LIB/NATE_LIB_UNIX
    my @natelib;
    if (exists($ENV{"NATE_LIB"})) {
        @natelib = split(/;/, $ENV{"NATE_LIB"});
    }
    if (not scalar(@natelib) and exists($ENV{"NATE_LIB_UNIX"})) {
        @natelib = split(/;/, $ENV{"NATE_LIB_UNIX"});
    }

    # walk through all the natelib paths until we find one that works
    foreach my $natelib_path (@natelib) {
        $dtd_path = $natelib_path . "/lib/San/SanMgr/DTDFiles";
        if (-d $dtd_path) {
            &ParamsMgr::set(key => "SANMGR_DTD_PATH", value => $dtd_path, paramset => $self->NACL::AppBuilder::EnGenX::Util::get_paramset());
            return $dtd_path;
        }
    }

    # (3) try to determine the path based on the TESTNAME variable
    my $testname = $ENV{"TESTNAME"};
    if ($testname =~ /^(\/.+?\/p4\/.+?\/)/) {
        $dtd_path = $1 . "lib/San/SanMgr/DTDFiles";

        # make sure this path exists, otherwise dont use it
        if (-d $dtd_path) {
            &ParamsMgr::set(key => "SANMGR_DTD_PATH", value => $dtd_path, paramset => $self->NACL::AppBuilder::EnGenX::Util::get_paramset());
            return $dtd_path;
        }
    }

    # (4) otherwise, default to sanrtest's version
    $dtd_path = "/u/sanrtest/p4/test/lib/San/SanMgr/DTDFiles";
    &ParamsMgr::set(key => "SANMGR_DTD_PATH", value => $dtd_path, paramset => $self->NACL::AppBuilder::EnGenX::Util::get_paramset());

    return $dtd_path;
}

###############################################################################
# Method:    _get_xml_path
# Objective: to get the full path name with extension (xml)
# Parameter:  $xml_file (Optional)   -> user specified xml name
# Expected Output if script location is in
# "/u/muthug/p4/test/san/lun_mgmt/portset/basic_func.thpl";
# and all xmls exist.
#
#  _get_xml_filename() =
#     "/u/muthug/p4/test/san/lun_mgmt/portset/basic_func.xml";
#  _get_xml_filename("") =
#     "/u/muthug/p4/test/san/lun_mgmt/portset/basic_func.xml";
#  _get_xml_filename("gmk1.xml.xml") =
#     "/u/muthug/p4/test/san/lun_mgmt/portset/gmk1.xml.xml";
#  _get_xml_filename("justfile") =
#     "/u/muthug/p4/test/san/lun_mgmt/portset/justfile.xml";
#  _get_xml_filename("p4/test/san/lun_mgmt/portset/fullpath.xml") =
#     "p4/test/san/lun_mgmt/portset/fullpath.xml";
#  _get_xml_filename("/san/lun_mgmt/portset/fullpathwitoutext") =
#     "/san/lun_mgmt/portset/fullpathwitoutext.xml";
#  _get_xml_filename("/u/any/junk/folder/name/only/portset/") =
#     "/u/any/junk/folder/name/only/portset/basic_func.xml";
#
# Return:  returns a string with full path with filename and extension(.xml)
###############################################################################
sub _get_xml_path {
    my ($self, $xml_file) = @_;
    my $old_xml_file = $xml_file;
    my ($sc_file, $sc_dir, $sc_ext);
    my ($file,    $dir,    $ext);

    # If user passed nothing or empty ("") check the script location with name
    # the same as the script
    unless (length $xml_file) {
        $xml_file = $Tharn::Scriptname;

        # get the last .thpl(or any ext) name and replace with .xml
        $xml_file =~ s/\.[^.]*/.xml/;

        return $xml_file if (-f $xml_file);
        NATE::BaseException->throw("No Xml File was passed in and $xml_file "
                                 . "does not exist");
    }

    # always ensure that a file ends with .xml
    if ($xml_file !~ /\.xml$/) {
        $xml_file .= ".xml";
    }

    # -f checks to see if the file exists and is a plain file
    return $xml_file if (-f $xml_file);

    # second get the script location: dir:filename:ext
    # this regex will split something like
    # /u/mcclella/p4/test/san/sanity/testXml_path.xml
    # into
    # $sc_dir = /u/mcclella/p4/test/san/sanity
    # $sc_file = testXml_path
    # $sc_ext = xml
    ($sc_dir, $sc_file, $sc_ext) = $Tharn::Full_Scriptname =~ m#(\S+)/(\S+)\.(\S+)$#;

    # Get file location info
    ($file, $dir, $ext) = &File::Basename::fileparse($xml_file, qr/\.[^.]*/);

    # if dir is not passed, grab it from script location
    if ($dir =~ /^.\/$|\.$|^\/$|""/) {
        $xml_file = "$sc_dir/";
    } else {
        $xml_file = $dir;
    }

    # if file is not passed, grab it from script name
    unless (length $file) {
        $xml_file .= $sc_file;
    } else {
        $xml_file .= $file;
    }

    # add the extension back
    if (length $ext) {
        $xml_file .= $ext;
    }

    # -f checks to see if the file exists and is a plain file
    # If the file does not exist, this was the final location to check
    # and an error is returned.
    return $xml_file if (-f $xml_file);

    # Still have not found the file.  Seems like we hit this when
    # $sc_dir is a relative path (smokes).
    # Next, call NATE::Inc::find_file() with the current
    # $xml_file and see if it can find it in the nate lib path.
    my @natelib;
    if (exists($ENV{"NATE_LIB"})) {
        @natelib = split(/;/, $ENV{"NATE_LIB"});
    }
    if (not scalar(@natelib) and exists($ENV{"NATE_LIB_UNIX"})) {
        @natelib = split(/;/, $ENV{"NATE_LIB_UNIX"});
    }
    $xml_file = NATE::Inc::find_file($xml_file, search => [@natelib]);

    # we found the XML file
    return $xml_file if ($xml_file and -f $xml_file);

    # If we get here, we didn't find the file anywhere we know to look
    # get the cwd for the error message
    my $cwd = &Cwd::getcwd();

    NATE::BaseException->throw("$old_xml_file does not exist as a direct path "
                        . "and $file"."$ext"." can not be found in the following:\n"
                        . "  - $cwd\n"
                        . "  - $sc_dir\n"
                        . "NATELIB below\n"
                        . "  - " . join("\n  - ",@natelib)
                        );
}    # end of sub _get_xml_path

###############################################################################
# Method:    _verify_resources_loaded
# Objective: Checks that the load resource flag has been set, meaning that
#            the subruotine load_resources has been executed, meaning that
#            objects are in memory.
# Input:     No input.
# Returns:   TRUE or propagates an exception.
###############################################################################
sub _verify_resources_loaded{
    my ($self) = @_;
    $Log->enter() if $may_enter;

    # If resources have not been loaded, load them before executing.
    if (!$self->{resources_loaded}) {
        $self->load_resources();
    }

    $Log->exit() if $may_exit;
    return TRUE;
}

###############################################################################
# Method:    _log_performance
# Objective: Logs performance data in the performance log
# Input:     time_difference - The time difference from start to finish
#            action          - The action taken(setup,cleanup)
# Returns:   TRUE
###############################################################################
sub _log_performance{
    my ($self, @opts) = @_;
    $Log->enter() if $may_enter;

    my %args = ();

    %args = Params::Validate::validate( @opts, {
            time_difference  => {type => SCALAR, optional => FALSE},
            action           => {type => SCALAR, optional => FALSE},
            }
        );

    # Redirect the $Log calls to write to performance.log
    $Log->redirect(path=>(Tharn::param('LOGDIR').'/performance.log'));
    $EnGenXUtil::TOTAL_OBJ_COUNT += $EnGenXUtil::OBJ_COUNT;
    $EnGenXUtil::TOTAL_TIME += $args{time_difference};
    $Log->comment($EnGenXUtil::OBJ_COUNT . " objects for controller ".$args{action}.
                  " took ".$args{time_difference}." seconds with an average of ".
                  $args{time_difference}/$EnGenXUtil::OBJ_COUNT . " seconds per object    ");
    $Log->comment("A total of $EnGenXUtil::TOTAL_OBJ_COUNT objects for Controller ".
                  "Provisioning so far took $EnGenXUtil::TOTAL_TIME seconds ".
                  "with an average of ". $EnGenXUtil::TOTAL_TIME/$EnGenXUtil::TOTAL_OBJ_COUNT.
                  " seconds per object    ");
    $Log->comment("Logging the following so that the performance log will PASS");
    $Log->comment("END OF TEST \/ RESULTS SUMMARY\nNo results reported\n");
    # Direct $Log calls to write back to the original log file
    $Log->reset_path();

    $Log->exit() if $may_exit;
    return TRUE;
}

###############################################################################
# Method:    _load_cached_paramset
# Objective: Attempt to a load the cached paramset for a given XML file
# Input:     xml_file        - xml file to load cache against
#            schema_files    - array ref of dtd files to load cache against
#            hexdigest       - scalarref to update when the hex is generated
# Returns:   TRUE - if the cache was loaded
#            FALSE - if the cache does not exist
###############################################################################
sub _load_cached_paramset {
    my ($self, %args) = @_;
    $Log->enter() if $may_enter;

    # calculate the hex digest of the paramset
    my $hexdigest = $self->_calculate_hex_digest(%args);

    # update the hexdigest scalar ref
    ${$args{hexdigest}} = $hexdigest;
    $Log->comment("Cache id: $hexdigest");

    my $cache_file_name = param_global()->get('TOP_LOGDIR') .
        "/../.engenx_cache/" . "engenx_${hexdigest}.cache";
    my $success = open(my $fh, "<", $cache_file_name);
    if (!$success) {
        # return 0 if there is no cached paramset
        $Log->debug("Cached file does not exist, need to load XML");
        $Log->exit() if $may_exit;
        return FALSE;
    } else {
        $Log->debug("Loading cache: $cache_file_name");
    }

    # read in the file
    my $encoded_ps;
    while (<$fh>) {
        $encoded_ps .= $_;
    }
    close($fh);

    # decode and load the paramset
    $Log->debug("Decoding paramset");
    my $decoded_ps = encode_global()->decode($encoded_ps);
    $Log->debug("Loading paramset");
    $self->NACL::AppBuilder::EnGenX::Util::get_paramset()->load(string => $decoded_ps);

    $Log->exit() if $may_exit;
    # successfully loaded;
    return TRUE;
}

###############################################################################
# Method:    _cache_paramset
# Objective: Cache a given paramset
# Input:     xml_file        - xml file to cache against
#            schema_files    - array ref of dtd files to cache against
#            hexdigest       - optional: hexdigest to use when caching the paramset
#                              if not specified will generated a digest.
# Returns:  Nothing
###############################################################################
sub _cache_paramset {
    my ($self, %args) = @_;
    $Log->enter() if $may_enter;

    my $hexdigest = delete($args{hexdigest});

    if (!$hexdigest) {
        # calculate the hex digest of the paramset
        $hexdigest = $self->_calculate_hex_digest(%args);
    }
    $Log->comment("Cache id: $hexdigest");

    # check if the cache file already exists
    my $dir = param_global()->get('TOP_LOGDIR') . "/../.engenx_cache";
    my $cache_file_name = $dir . "/engenx_${hexdigest}.cache";
    if (!-e $cache_file_name) {
        $Log->debug("Cache file does not exist");
        # dump and encode the paramset
        my $ps = $self->NACL::AppBuilder::EnGenX::Util::get_paramset();
        my $encoded_ps = encode_global()->encode($ps->dump());
        $Log->debug("Writing cached file: $cache_file_name");
        mkdir($dir);
        open(my $fh, ">", $cache_file_name);
        print $fh $encoded_ps;
        close($fh);
    } else {
        $Log->debug("Cache file already exists");
    }

    $Log->debug("Cleaning up old cache files.");
    # only keep the last four engenx cache files
    my $logdir = param_global()->get('TOP_LOGDIR') . '/../.engenx_cache';
    opendir(my $dh, $logdir);
    my @cache_files = grep {/engenx_.*\.cache/} readdir($dh);
    # sort the cache files by time
    my @sorted_cache_files = sort {
        my @stata = stat $dir . "/" . $a;
        my @statb = stat $dir . "/" . $b;
        $stata[10] <=> $statb[10]
    } @cache_files;

    my @keep_cache;
    for (my $i = 0; $i < 4; $i++) {
        push(@keep_cache, pop(@sorted_cache_files));
    }
    close($dh);

    # delete the old cache files if there are any left
    @sorted_cache_files = map { $logdir . "/" . $_ } @sorted_cache_files;
    if (@sorted_cache_files) {
        unlink(@sorted_cache_files);
    }
    $Log->exit() if $may_exit;
    return;
}

###############################################################################
# Method:    _calculate_hex_digest
# Objective: Calculate the hex digest for a given XML/DTD set
# Input:     xml_file        - Required: xml file to cache against
#            schema_files    - array ref of dtd files to cache against
# Returns:  a hex digest string
###############################################################################
sub _calculate_hex_digest {
    my ($self, %args) = @_;
    $Log->enter() if $may_enter;

    my $xml_file = delete($args{xml_file}) // confess "Missing required parameter 'xml_file'";
    my $override_schema_files = delete($args{schema_files}) // [];

    my @schema_files = @DTD_SCHEMA_FILES;
    if (scalar(@{$override_schema_files})) {
        push(@schema_files, @{$override_schema_files});
    }

    # generate a unique key based on the xml and schema file inputs
    # and the parameters used in the XML file
    # insertion order for the hash is important
    my $md5 = Digest::MD5->new();
    my %params = ();
    tie %params, 'Tie::IxHash';

    my @file_list = ($xml_file, @schema_files);

    while (my $file = shift(@file_list)) {
        my $fh;
        if (ref($xml_file) =~ /GLOB/i) {
            $fh = $xml_file;
        } else {
            open($fh, "<", $file) or confess "Cannot open $file for reading";
        }
        # read the file and find all parameters being used by the XML/DTD files
        while (<$fh>) {
            # skip documentation sections for DTD
            last if (/\=start_doc/i);

            # parameters can look like:
            #  "$PARAM_NAME"
            #  "${PARAM_NAME}"
            #  "$PARAM_NAME,something"
            #  "${PARAM_NAME,something}"
            # we only want to extract the "PARAM_NAME" part
            if (/\$(\w+)/ or /\$\{(\w+)(\}|,)/) {
                my $p = $1;
                $params{$p}++;
            }

            # we need to find any XML Files that are being
            # included via the <include_xml/> element.
            # those files need to be opened and added to the md5
            # as well.
            if (/<include_xml\s+file=\"(.+?)\"\s*\/>/) {
                my $include_xml_file = $1;
                # if it's a relative path, take the dirname of the xml file
                # and prepend it.
                if ($include_xml_file !~ /^\s*\//) {
                    # remove any preceeding ./ just in case
                    $include_xml_file =~ s/^\.\///;
                    $include_xml_file = dirname($xml_file) . "/" . $include_xml_file;
                }
                # add this file to the file list for further processing
                $Log->debug("Adding include xml $include_xml_file to md5");
                push(@file_list, $include_xml_file);
            }
        }

        # seek back to the beginning of the file
        # because i dont trust this MD5 library
        seek($fh, 0, 0);

        # add the file to the md5
        $md5->addfile($fh);
        close($fh);
    }

    # find all the parameters and their respective values
    # that are used in the XML/DTD files
    foreach my $param (keys %params) {
        $params{$param} = param_global()->get($param);
    }

    # encode the parameter hash and add it to the md5
    $Log->debug("Encoding and adding params to md5");
    $md5->add(encode_global()->encode(\%params));

    # calculate the hexdigest, this will be our cache ID
    my $digest = $md5->hexdigest();
    $Log->exit() if $may_exit;
    return $digest;
}

sub _get_next_runid_prefix {
    my ($self) = @_;

    my $prefix_count = $self->{runid_prefix_count} // 0;
    my $return_string = $prefix_count;

    # Pad the subtest count with zero's for values less than 100
    $return_string = "00" . $prefix_count if (length($prefix_count) == 1);
    $return_string = "0"  . $prefix_count if (length($prefix_count) == 2);

    # Increment the subtest count
    $self->{runid_prefix_count}++;
    return $return_string;
}

1;
