#
# Copyright (c) 2014 NetApp, Inc., All Rights Reserved.
#   Any other use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

package NACL::AppBuilder::EnGenX::Client::Deconfig;

use strict;
use warnings;

use NATE::Process;
use NATE::Log qw(log_global);
use NATE::Messenger qw(messenger_global);
use NATE::BaseException qw(:try);

sub deconfig {
    my ($self, %args) = @_;
    my $run_id_prefix = delete($args{run_id_prefix}) // "";

    my $client_groups = $self->resource_list()->get_resources(category => "CLIENTS",
                                                              type => "client_groups");

    if (!scalar(@{$client_groups})) {
        log_global()->log("No Clients to deconfig");
    }

    my @procs = ();
    my @required_clients = ();
    my $groups = {};
    foreach my $group (@{$client_groups}) {
        if ($group->{DECONFIG} == 0) {
            log_global()->log("Skipping client group " . $group->{GROUP_ID} . " because deconfig is disabled on it");
            next;
        }

        my $group_id = $group->{GROUP_ID};
        $groups->{$group_id}->{max_failures} = $group->{MAX_FAILURE_COUNT};
        my $required_clients = $group->{REQUIRED_CLIENTS};
        push(@required_clients, split(/\s*,\s*/, $required_clients));

        my $client_list = $self->resource_list()->get_resources(category => "CLIENTS",
                                                                type => "clients",
                                                                data => {GROUP_ID => $group_id});
        foreach my $client (@{$client_list}) {
            if ($client->{DECONFIG} == 0) {
                log_global()->log("Skipping client " . $client->{NAME} . " because deconfig is disabled on it.");
                next;
            }
            my $p = NATE::Process->new(runid    => $client->{NAME},
                                       codespec => sub {
                                           try {
                                               $self->deconfig_client(client => $client,
                                                                      group_id => $group_id);
                                               $self->_deconfig_complete();
                                           } otherwise {
                                               my $e = shift;
                                               $self->_deconfig_failure(error => $e->text(),
                                                                        client => $client->{NAME},
                                                                        group_id => $group_id);
                                               $e->throw();
                                           };
                                       },
                                       onexit   => \&NATE::Process::on_exit_propagate_worst_result,);

            push(@procs, $p);
        }
    }

    # install a listener to handle a deconfig completion
    my $success_listener = {
        completed => 0,
        filter => sub { shift->type() eq 'NACL/AppBuilder/EnGenX/Client/Deconfig/Complete' },
        callback => sub {
            my ($msg, $l) = @_;
            $l->{completed}++;
        }
    };

    # install a listener to handle a deconfig failure
    my $someone_failed = 0;
    my @failed_clients;
    my $fail_listener = {
        completed => 0,
        filter => sub { shift->type() eq 'NACL/AppBuilder/EnGenX/Client/Deconfig/Failure'},
        callback => sub {
            my ($msg, $l) = @_;
            $msg = $msg->message();
            $groups->{$msg->{group_id}}->{failed}++;
            push(@failed_clients, $msg->{client});
            $someone_failed = 1;
            $l->{completed}++;
        }
    };

    $self->_spawn_routines(procs => \@procs,
                           fail_listener => $fail_listener,
                           success_listener => $success_listener,
                           runid => "$run_id_prefix" . "Client-deconfig");

    # check if we had any failures
    if ($someone_failed) {
        foreach my $group (values %{$groups}) {
            # if the number of clients that failed in this group is beyond the max, then that
            # is a failure
            if ($group->{failed} > $group->{max_failures}) {
                NATE::BaseException->throw("A client group has hit its maximum deconfig " .
                                           "failure threshold");
            }
        }
    }

    # make sure that none of the clients that failed were in the required clients list
    foreach my $client (@failed_clients) {
        if (grep(/^$client$/, @required_clients)) {
            NATE::BaseException->throw("The client $client failed deconfig. Aborting since it " .
                                       "was on the required clients list.");
        }
    }

    return;
}

sub deconfig_client {
    my ($self, %args) = @_;

    my $client   = delete($args{client})   // NATE::BaseException->throw("Missing required parameter 'client'");
    my $group_id = delete($args{group_id}) // NATE::BaseException->throw("Missing required parameter 'group_id'");

    #############################################
    ### General client parameters
    #############################################
    my $client_name = $client->{NAME};
    my $protocol = $client->{PROTOCOL};
    my $hypervisor = $client->{HYPERVISOR};

    # Get client object
    my ($hostobj) = $self->get_objects(-type => "client", -names => [$client_name]);
    if ($self->error()) {
        NATE::BaseException->throw("Failed to get client ($client_name) object: " . $self->errmsg());
    }

    # if this is a sanblaze host, then no deconfig should be done
    if (ref($hostobj) =~ /Sanblaze/i) {
        log_global()->log("Bypassing deconfig because this is a sanblaze host");
        return;
    }

    # reconnect to the client
    $hostobj->reconnect(-disconnect_existing_conn => 0);
    if ($hostobj->error()) {
        NATE::BaseException->throw("Failed to reconnect to client ($client_name) object: " . $self->errmsg());
    }

    # kill off any running I/O program on the client systems.
    $self->_kill_io_instances_subtest(client => $client_name,
                                      resource => $client);

    # skip deconfig for virtual clients that are under hypervisor control
    if ($hypervisor !~ /^none$/i) {
        log_global()->log("Skipping client deconfig for client '$client_name' because it is a virtual machine");
        return;
    }

    my ($data) = $self->{parent}->get_target_list(-client_name => $client_name,
                                                  -filter_iscsi_targets_by_portset => $hostobj->{iscsi}->{DECONFIG_ISCSI_FILTER_TARGETS_BY_PORTSET},
                                                  -filter_fcp_targets_by_portset   => $hostobj->{fcp}->{DECONFIG_FCP_FILTER_TARGETS_BY_PORTSET});
    if ($self->{parent}->error()) {
        NATE::BaseException->throw("Failed to get iSCSI/FCP target lists : " . $self->{parent}->errmsg());
    }

    # Run the deconfigure routine within the client library for iscsi
    my @iSCSITargets = @{$data->{iscsi_targets}};

    if ($protocol =~ /iscsi|fcp/i and scalar(@iSCSITargets)) {
        $hostobj->deconfigure(-targets => \@iSCSITargets,
                              -clean_level => $hostobj->{iscsi}->{CONFIG_ISCSI_DISCOVERY_PRECLEANUP});
        if ($hostobj->error()) {
            NATE::BaseException->throw("Failed to deconfigure protocol $protocol for client ($client_name) : " . $hostobj->errmsg());
        }
    }

    if ($protocol =~ /iscsi|fcp/i) {
        # clean up orphaned devices on the host
        $hostobj->orphaned_device_cleanup();
        if ($hostobj->error()) {
            NATE::BaseException->throw("Failed to cleanup orphaned devices on client ($client_name) : " . $hostobj->errmsg());
        }
    }

    return;
}

sub _deconfig_failure {
    my ($self, %data) = @_;
    messenger_global()->put(endpoint => $self->{endpoint},
                            type     => "NACL/AppBuilder/EnGenX/Client/Deconfig/Failure",
                            message  => \%data);
}

sub _deconfig_complete {
    my ($self, %data) = @_;
    messenger_global()->put(endpoint => $self->{endpoint},
                            type     => "NACL/AppBuilder/EnGenX/Client/Deconfig/Complete",
                            message  => \%data);
}

1;
