#
# Copyright (c) 2013 NetApp Inc.
# All rights reserved.
#
## @summary Provide a common set of libraries to work on clusters
## @author dalmond@netapp.com

package NACL::AppBuilder::EnGenX::ClientCluster;

# Compiler directives.
use strict;

# Module import.
use NATE::Log qw(log_global);
use Error;
use San::SanMgr::Util;
use San::SanClientCluster;
use Data::Dumper;

# Load resource specific modules
use NACL::AppBuilder::EnGenX::ClientCluster::setup;
use NACL::AppBuilder::EnGenX::ClientCluster::cleanup;
use NACL::AppBuilder::EnGenX::Util;


use vars qw(@ISA);
@ISA = qw(NACL::AppBuilder::EnGenX::Util);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();

###############################################################################
#                               Public Methods
###############################################################################

###############################################################################
# Method:    _new
# Objective: Create new resource object.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _new {
    my ( $proto, %args ) = @_;

    my $self = {};

    my $class = ref($proto) || $proto;
    bless( $self, $class );

    # allow for error stack build ups
    $self->{error_dont_overwrite_errors} = 1;
    $self->{error_stack_style}           = "normal";

    &ParamsMgr::process_args(%args);
    my $parent = &ParamsMgr::get_arg(
        -arg      => "parent",
        -required => 1,
        -default  => ""
    );
    if ( &ParamsMgr::error() ) {
        $self->_set_err( &ParamsMgr::get_error() );
        return $self;
    }

    # Save parent
    $self->{parent} = $parent;

    return $self;
} ## end sub _new

sub xml_to_list {
    my ($self, %args) = @_;

    my $multi_client_clusters = &NACL::AppBuilder::EnGenX::Util::_get_all_objects_paramsets($self, "CLIENT_CLUSTER");
    my $group_id = 0;

    # process all groups
    foreach my $multi_client_cluster (@{$multi_client_clusters}) {
        # remove all the CLIENT_CLUSTER\d elements from the data structure
        my $client_elm = {};
        foreach my $key (keys(%{$multi_client_cluster})) {
            # register each client cluster as an individual resource
            if ($key =~ /^CLIENT_CLUSTER\d*$/) {
                $client_elm = $multi_client_cluster->{$key};
                $client_elm->{GROUP_ID} = $group_id;
                # change CLUSTER_NAME to NAME
                if (defined $client_elm->{CLUSTER_NAME}) {
                    $client_elm->{NAME} = delete $client_elm->{CLUSTER_NAME};
                }

                # register each client cluster as a resource
                $self->resource_list()->add_resource(
                    category => "CLIENT_CLUSTERS",
                    type     => "client_clusters",
                    data     => $client_elm,
                );
                # remove it from the set
                delete($multi_client_cluster->{$key});
            }
        }
        # store the group id on the group obj
        $multi_client_cluster->{GROUP_ID} = $group_id;

        # create a resource for the global options of the group
        $self->resource_list()->add_resource(
            category => "CLIENT_CLUSTERS",
            type => "client_cluster_groups",
            data => $multi_client_cluster);
        $group_id++;
    }

    return;
}

##############################################################################
# Method:    setup
# Objective: Launch subtest for each cluster to perform setup
# Details:   See POD documentation at the end of this file.
###############################################################################
sub setup {
    my ( $self, %args ) = @_;

    $self->clear_error();

    my $client_cluster_groups = $self->resource_list()->get_resources(category => "CLIENT_CLUSTERS",
                                                              type => "client_cluster_groups");

    if (!scalar(@{$client_cluster_groups})) {
        log_global()->log("No Client Clusters to setup");
    }

    my ( @messages, @subtests );
    foreach my $group (@{$client_cluster_groups}) {
        if ($group->{SETUP} == 0) {
            log_global()->log("Skipping client cluster group " . $group->{GROUP_ID} . " because setup is disabled on it");
            next;
        }

        my $group_id = $group->{GROUP_ID};
        my $max_setup_failure_count  = $group->{MAX_SETUP_FAILURE_COUNT};
        my $required_client_clusters = $group->{REQUIRED_CLIENT_CLUSTERS};

        my $client_cluster_resources = $self->resource_list()->get_resources(category => "CLIENT_CLUSTERS",
                                                                            type => "client_clusters",
                                                                            data => {GROUP_ID => $group_id});

        if (scalar(@$client_cluster_resources)) {
            &ParamsMgr::process_args(%args);
            my $run_id_prefix = &ParamsMgr::get_arg(
                -arg      => "run_id_prefix",
                -required => 0,
                -default  => ""
            );
            if ( &ParamsMgr::error() ) {
                $self->_set_err( "Argument parsing failed", &ParamsMgr::get_error() );
                return $SanMgr::FALSE;
            }

            my $subtest = &SubIPC::create(
                -id  => "$run_id_prefix" . "Client-cluster-setup",
                -esc => $self->{parent}->{enable_subtest_count},
                -run => sub {
                    $self->_setup_all_client_clusters(
                        -max_setup_failure_count         => $max_setup_failure_count,
                        -required_client_clusters        => $required_client_clusters,
                        -client_clusters                 => $client_cluster_resources
                    );
                },
                -ipc => 1,
                -bg  => 1
            );
            if ( &SubIPC::error() ) {
                $self->_set_err( &SubIPC::get_error() );
                return $SanMgr::FALSE;
            }
            push( @subtests, $subtest );
        }
    }

    # reap the setup subtests if they exist
    if ( scalar(@subtests) ) {

        # Wait for subtests to finish
        &SubIPC::wait_finish( -subtest => \@subtests );
        if ( &SubIPC::error() ) {
            $self->_set_err( &SubIPC::get_error() );
            return $SanMgr::FALSE;
        }

        # Get messages to process
        @messages = &SubIPC::get_messages(
            -subtest => \@subtests,
            -type    => "client_cluster_setup_complete"
        );

        if ( &SubIPC::error() ) {
            $self->_set_err( &SubIPC::get_error() );
            return $SanMgr::FALSE;
        }

    }
    
     # check to see if the client setup subtests failed
    if ( scalar(@subtests) ) {
        my $st_status = &SubIPC::get_subtest_status( -subtest => \@subtests );
        if ( $st_status->{failed} != 0 ) {
            $self->_set_err(
                      "A subtest failure has been reported:\n\t\tFailed: "
                    . $st_status->{failed}
                    . "\n\t\tPassed: "
                    . $st_status->{passed} );
            return $SanMgr::FALSE;
        } else {
            $Log->comment( -msg => "Client Cluster setup has passed\n\t\t "
                    . "\n\t\tPassed: "
                    . $st_status->{passed} );
        }
    }

    return $SanMgr::TRUE;
} ## end sub setup

###############################################################################
# Method:    _setup_all_client_clusters
# Objective: Launch subtest for each client cluster to perform setup
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _setup_all_client_clusters {

    my ( $self, %args ) = @_;

    $self->clear_error();
    my ( $st_status, @subtests );
    &ParamsMgr::process_args(%args);

    if ( &ParamsMgr::error() ) {
        $self->_set_err( "Argument parsing failed", &ParamsMgr::get_error() );
        &SubIPC::fail( -msg => $self->errmsg() );
        return $SanMgr::FALSE;
    }
    my $max_setup_failure_count = &ParamsMgr::get_arg(
        -arg      => "max_setup_failure_count",
        -required => 0,
        -default  => "1"
    );
    my $required_client_clusters = &ParamsMgr::get_arg(
        -arg      => "required_client_clusters",
        -required => 0,
        -default  => ""
    );
    my $client_clusters = &ParamsMgr::get_arg( -arg => "client_clusters", -required => 1 );

    if ( &ParamsMgr::error() ) {
        $self->_set_err( "Argument getting failed", &ParamsMgr::get_error() );
        &SubIPC::fail( -msg => $self->errmsg() );
        return $SanMgr::FALSE;
    }

    if ( not scalar( @{$client_clusters} ) ) {
        &SubIPC::pass( -msg => "There are no Clusters to perform setup for" );
        return $SanMgr::TRUE;
    }

    # Go through client list and start setup
    foreach my $client_cluster ( @{$client_clusters} ) {
        my $client_cluster_name    = $client_cluster->{NAME};
        $client_cluster->{SETUP} = 1
            if ( not defined( $client_cluster->{SETUP} ) );

        # skip setup on this client if it is not running setup
        if (   $client_cluster->{SETUP} == 0
            or $client_cluster->{DEV_LIST} ne "" ) {
            $Log->comment(
                -msg => "Skipping client cluster "
                    . $client_cluster_name
                    . " because setup tag set to \"0\" or a dev list is defined",
                -msg_level => 0
            );
            next;
        }

        my $subtest = &SubIPC::create(
            -id  => $client_cluster_name,
            -run => sub {
                &NACL::AppBuilder::EnGenX::ClientCluster::setup::setup_single( $self,
                    -resource => $client_cluster );
            },
            -parent => $self,
            -ipc    => 1,
            -bg     => 1
        );
        if ( &SubIPC::error() ) {
            $self->_set_err( &SubIPC::get_error() );
            &SubIPC::fail( -msg => &SubIPC::get_error() );
            return $SanMgr::FALSE;
        }

        push( @subtests, $subtest );
    }

    # return if there were no client cluster setup subtests
    if ( not scalar(@subtests) ) {
        &SubIPC::pass( -msg => "No Client Clusters have setup enabled" );
        return $SanMgr::TRUE;
    }

    # Wait for subtests to finish
    &SubIPC::wait_finish( -subtest => \@subtests );
    if ( &SubIPC::error() ) {
        $self->_set_err( &SubIPC::get_error() );
        &SubIPC::fail( -msg => &SubIPC::get_error() );
        return $SanMgr::FALSE;
    }


    # Go through the subtests and see if any clients in the required client cluster list
    # have failed.
    foreach my $subtest (@subtests) {
        my $client_cluster_name = $subtest->{id};
        if (    ( $subtest->{exit_status} =~ /fail/i )
            and ( $required_client_clusters =~ /$client_cluster_name/ ) ) {
            $self->_set_err(
                "A subtest failure has been reported:\n\t\tThe required client "
                    . $subtest->{id}
                    . "has failed setup.\n"
                    . "Failed: "
                    . $st_status->{failed}
                    . "\n\t\tPassed: "
                    . $st_status->{passed} );
            &SubIPC::fail( -msg => $self->errmsg() );
            return $SanMgr::FALSE;
        } elsif ($subtest->{exit_status} =~ /fail/i) {
            $st_status->{failed}++;
        } else{
            $st_status->{passed}++;
        }
    }
    if ( $st_status->{failed} > $max_setup_failure_count ) {
        $self->_set_err( "A subtest failure has been reported:\n\t\tFailed: "
                . $st_status->{failed}
                . "\n\t\tPassed: "
                . $st_status->{passed} );
        &SubIPC::fail( -msg => $self->errmsg() );
        return $SanMgr::FALSE;
    }

    if ( $st_status->{failed} > 0 ) {
        $Log->comment( -msg => "Setup passed on enough client clusters to continue "
                . "running the test.\n\t\tFailed: "
                . $st_status->{failed}
                . "\n\t\tPassed: "
                . $st_status->{passed} );
    }
    &SubIPC::pass( -msg => "Client clusters have been setup" );
    return $SanMgr::TRUE;
} ## end sub _setup_all_client_clusters

###############################################################################
# Method:    cleanup
# Objective: Launch subtest for each client to perform cleanup
# Details:   See POD documentation at the end of this file.
###############################################################################
sub cleanup {
    my ( $self, %args ) = @_;
    my @subtests;
    $self->clear_error();

    my $client_cluster_groups = $self->resource_list()->get_resources(category => "CLIENT_CLUSTERS",
                                                              type => "client_cluster_groups");

    if (!scalar(@{$client_cluster_groups})) {
        log_global()->log("No Client Clusters to cleanup");
    }

    foreach my $group (@{$client_cluster_groups}) {
        if ($group->{CLEANUP} == 0) {
            log_global()->log("Skipping client cluster group " . $group->{GROUP_ID} . " because cleanup is disabled on it");
            next;
        }

        my $group_id = $group->{GROUP_ID};

        my $client_cluster_resources = $self->resource_list()->get_resources(category => "CLIENT_CLUSTERS",
                                                                    type => "client_clusters",
                                                                    data => {GROUP_ID => $group_id});

        if (scalar(@$client_cluster_resources)) {
            &ParamsMgr::process_args(%args);
            my $run_id_prefix = &ParamsMgr::get_arg(
                -arg      => "run_id_prefix",
                -required => 0,
                -default  => ""
            );
            if ( &ParamsMgr::error() ) {
                $self->_set_err( "Argument parsing failed",
                    &ParamsMgr::get_error() );
                return $SanMgr::FALSE;
            }

            my $subtest = &SubIPC::create(
                -id  => "$run_id_prefix" . "Client-Cluster-cleanup",
                -esc => $self->{parent}->{enable_subtest_count},
                -run => sub { $self->_cleanup_all_client_clusters(
                                  -client_clusters => $client_cluster_resources); },
                -ipc => 1,
                -bg  => 1
            );
            if ( &SubIPC::error() ) {
                $self->_set_err( &SubIPC::get_error() );
                return $SanMgr::FALSE;
            }
            push( @subtests, $subtest );
        }
    }

    # return if there are no subtests running client cleanup
    if ( not scalar(@subtests) ) {
        return $SanMgr::TRUE;
    }

    &SubIPC::wait_finish( -subtest => \@subtests );
    if ( &SubIPC::error() ) {
        $self->_set_err( &SubIPC::get_error() );
        return $SanMgr::FALSE;
    }

    my $st_status = &SubIPC::get_subtest_status( -subtest => \@subtests );
    if ( $st_status->{failed} ) {
        $self->_set_err(
                  "A Client Cluster cleanup subtest failure has been reported:\n\t\tFailed: "
                . $st_status->{failed}
                . "\n\t\tPassed: "
                . $st_status->{passed} );
        return $SanMgr::FALSE;
    }
    $Log->comment( -msg => "Client Cluster cleanup has passed\n\t\tFailed: "
            . $st_status->{failed}
            . "\n\t\tPassed: "
            . $st_status->{passed} );

    return $SanMgr::TRUE;
} ## end sub cleanup

###############################################################################
# Method:    _cleanup_all_client_clusters
# Objective: Launch subtest for each client cluster to perform cleanup
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _cleanup_all_client_clusters {
    my ($self, %args) = @_;

    my ( $st_status, @subtests );
    
    &ParamsMgr::process_args(%args);
    my $client_clusters = &ParamsMgr::get_arg( -arg => "client_clusters", -required => 1 );

    if ( &ParamsMgr::error() ) {
        $self->_set_err( "Argument getting failed", &ParamsMgr::get_error() );
        &SubIPC::fail( -msg => $self->errmsg() );
        return $SanMgr::FALSE;
    }

    if ( not scalar( @{$client_clusters} ) ) {
        &SubIPC::pass( -msg => "There are no Clusters to perform cleanup for" );
        return $SanMgr::TRUE;
    }
    # Go through client cluster list and start cleanup
    foreach my $client_cluster ( @{$client_clusters} ) {
            my $client_cluster_name    = $client_cluster->{NAME};
            $client_cluster->{CLEANUP} = 1
                if ( not defined( $client_cluster->{CLEANUP} ) );

            # Check to see if resource is to be skipped
            if ( $client_cluster->{CLEANUP} == 0 ) {
                $Log->comment(
                    -msg => "Skipping client cluster "
                        . $client_cluster_name
                        . " because cleanup tag set to \"0\"",
                    -msg_level => 0
                );
                next;
            }
            $Log->comment(
                    -msg => "Cleaning client cluster "
                        . $client_cluster_name
                        . " because cleanup tag set to \"0\"",
                    -msg_level => 0
            );

            my $subtest = &SubIPC::create(
                -id  => $client_cluster_name,
                -run => sub {
                    &NACL::AppBuilder::EnGenX::ClientCluster::cleanup::cleanup_single( $self,
                        -resource => $client_cluster);
                },
                -parent => $self,
                -ipc    => 1,
                -bg     => 1
            );
            if ( &SubIPC::error() ) {
                $self->_set_err( &SubIPC::get_error() );
                &SubIPC::fail( -msg => &SubIPC::get_error() );
                return $SanMgr::FALSE;
            }

            push( @subtests, $subtest );
    }
    if ( not scalar(@subtests) ) {
        &SubIPC::pass( -msg => "No Client Clusters have cleanup enabled" );
        return $SanMgr::TRUE;
    }

    # Wait for subtests to finish
    &SubIPC::wait_finish( -subtest => \@subtests );
    if ( &SubIPC::error() ) {
        $self->_set_err( &SubIPC::get_error() );
        &SubIPC::fail( -msg => &SubIPC::get_error() );
        return $SanMgr::FALSE;
    }

    $st_status = &SubIPC::get_subtest_status( -subtest => \@subtests );

    if ( $st_status->{failed} != 0 ) {
        &SubIPC::fail(
                  -msg => "A subtest failure has been reported:\n\t\tFailed: "
                . $st_status->{failed}
                . "\n\t\tPassed: "
                . $st_status->{passed} );
        return $SanMgr::FALSE;
    } else {
        &SubIPC::pass( -msg => "All subtests have passed:\n\t\tFailed: "
                . $st_status->{failed}
                . "\n\t\tPassed: "
                . $st_status->{passed} );
        return $SanMgr::TRUE;
    }
} ## end sub _cleanup_all_client_clusters

###############################################################################
# Method:    load_resource
# Objective: Gather list of clients and create objects for each one
# Details:   See POD documentation at the end of this file.
###############################################################################
sub load_resource {
    my ($self, %args) = @_;

    $self->clear_error();

    # get the environment from the resource list
    my $environment = $self->resource_list()->get_resources(category => "ENVIRONMENT",
                                                            type => "environment");
    $environment = $environment->[0];

    # obtain all the client resources from the list and prepare the processes that will load each one
    my $client_cluster_resources = $self->resource_list()->get_resources(category => "CLIENT_CLUSTERS",
                                                                 type => "client_clusters");

    # Go through the list and create an object for each one
    foreach my $client_cluster_resource (@$client_cluster_resources) {

        # Check to see if we don't already have a client cluster object
        if ( defined( $self->{parent}->{_objects}->{client_clusters}->{$client_cluster_resource->{NAME}}->{NAME} ) )
        {
            next;
        }

        my $conn_type = $client_cluster_resource->{CONN_TYPE} // "THARNHOST";
        my $debug_level = $client_cluster_resource->{DEBUG_LEVEL} // 0;

        # Create client cluster object
        my $cl_obj = SanClientCluster->new(
            -host               => $client_cluster_resource->{PRIMARY_NODE},
            -conn_type          => $conn_type,
            -debug_level        => $debug_level,
            -pnate_environment  => $environment->{PNATE},
            -cluster_type       => $client_cluster_resource->{CLUSTER_TYPE},
            -cluster_ip         => $client_cluster_resource->{CLUSTER_IP},
            -cluster_subnet     => $client_cluster_resource->{CLUSTER_SUBNET},
            -cluster_name       => $client_cluster_resource->{NAME},
            -domain_name        => $client_cluster_resource->{WINDOWS_DOMAIN_NAME},
            -domain_admin       => $client_cluster_resource->{WINDOWS_DOMAIN_USERNAME},
            -domain_adminpass   => $client_cluster_resource->{WINDOWS_DOMAIN_PASSWORD},
            -noDiscovery        => "1",
            -domain_controller_name => $client_cluster_resource->{WINDOWS_DOMAIN_CONTROLLER},
            -skip_device_format => "1",
            -sanclient_object   => $self->{parent}->{_objects}->{client}
            ->{ $client_cluster_resource->{PRIMARY_NODE} },
            -secondary_nodes    => $client_cluster_resource->{SECONDARY_NODES},
            -cluster_lun_regex  => $client_cluster_resource->{CLUSTER_LUN_REGEX},
        );

        if ( $cl_obj->error() ) {
            $self->_set_err( $cl_obj->errmsg() );
            return $SanMgr::FALSE;
        }
        # Save the object for future use
        $self->{parent}->{_objects}->{client_cluster}
            ->{ $client_cluster_resource->{NAME} } = $cl_obj;
    }

    return $SanMgr::TRUE;
} ## end sub load_resource

###############################################################################
# Method:    _subtest_event_handler
# Objective: Example subtest event handler
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _subtest_event_handler {
    my ( $subtest, %args ) = @_;

    &ParamsMgr::process_args(%args);
    my $message = &ParamsMgr::get_arg(
        -arg      => "message",
        -required => 1,
        -default  => ""
    );
    my $parent = &ParamsMgr::get_arg(
        -arg      => "parent",
        -required => 0,
        -default  => {}
    );
    if ( &ParamsMgr::error() ) {
        $subtest->_set_err( &ParamsMgr::get_error() );
        return $SanMgr::FALSE;
    }

    $Log->comment(
        -msg       => "Read from subtest: \n" . Dumper($message) . "\n",
        -msg_level => 3
    );

    return $SanMgr::TRUE;
} ## end sub _subtest_event_handler

################################################################################
# Subroutine Name: _kill_io_instances_subtest
# Objective:       Kill all instances of the client program on any client
################################################################################
sub _kill_io_instances_subtest {
    my ( $self, %args ) = @_;

    &ParamsMgr::process_args(%args);
    my $client = &ParamsMgr::get_arg(
        -arg      => "client",
        -required => 1,
        -default  => ""
    );
    my $resource = &ParamsMgr::get_arg( -arg => "resource", -required => 1 );
    my $client_io_program = &ParamsMgr::get_arg(
        -arg      => "client_io_program",
        -required => 1,
        -default  => ""
    );
    if ( &ParamsMgr::error() ) {
        $self->_set_err( &ParamsMgr::get_error() );
        return $SanMgr::FALSE;
    }

    my ( $exe_to_kill, @exes_to_kill );
    my (@subtests);

    # skip kill io processes if the flag is disabled
    if ( $resource->{KILL_IO_PROCESSES} == 0 ) {
        $Log->comment(
            -msg => "Skipping kill_io_processes because it "
                . "is disabled for client '$client'",
            -msg_level => "0"
        );
        return $SanMgr::TRUE;
    }

    # determine the process application to kill (all instances)
    if ( $client_io_program =~ /dt/ or $client_io_program =~ /all/i ) {
        push( @exes_to_kill, "dt" );
    }

    if ( $client_io_program =~ /sio/ or $client_io_program =~ /all/i ) {
        push( @exes_to_kill, "sio" );
    }

    if ( $client_io_program =~ /blast/ or $client_io_program =~ /all/i ) {
        push( @exes_to_kill, "blast" );
    }

    if (    $client_io_program !~ /sio/
        and $client_io_program !~ /dt/
        and $client_io_program !~ /blast/
        and $client_io_program !~ /all/i ) {
        $self->_set_err(
            "No valid client_io_program passed in ($client_io_program)");
        return $SanMgr::FALSE;
    }

    foreach $exe_to_kill (@exes_to_kill) {

        # kill the I/O process still on the host (all platforms)
        my $subtest = &SubIPC::create(
            -id  => "io_task_kill_" . $exe_to_kill . "_" . $client,
            -run => sub {
                &ClientUtil::kill_process_instances( $client, $exe_to_kill );
                &SubIPC::pass( -msg =>
                        "I/O Tasks for tool '$exe_to_kill' have been cleaned up on $client"
                );
                return $SanMgr::TRUE;
            },
            -parent => $self,
            -ipc    => 1,
            -bg     => 1
        );

        if ( &SubIPC::error() ) {
            $self->_set_err( &SubIPC::get_error() );
            return $SanMgr::FALSE;
        }
        push( @subtests, $subtest );
    }

    # Wait for subtests to finish
    &SubIPC::wait_finish( -subtest => [@subtests] );

    return $SanMgr::TRUE;
} ## end sub _kill_io_instances_subtest

1;
