# $Id: //depot/prod/DOT/dev/test/lib/EnGenX/Controller/Aggregate.pm#1 $

#
# Copyright (c) 2001-2012 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Aggregate Controller Layer Module
## @author dl-san-automation-core
## @status shared
## @pod here

=head1 NAME

NACL::AppBuilder::EnGenX::Controller::Aggregate

=head1 DESCRIPTION

NACL::AppBuilder::EnGenX::Controller::Aggregate is a derived class of
NACL::AppBuilder::EnGenX::Controller::Task.  It is used to Create and
Delete ONTAP Aggregates. All methods are static.

=cut

package NACL::AppBuilder::EnGenX::Controller::Aggregate;

# Compiler directives.
use strict;
use warnings;

# NATE and Standard Module import.
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use NATE::Util::DataTypes qw(TRUE FALSE);
use Params::Validate
    qw(validate_with validate SCALAR ARRAYREF BOOLEAN UNDEF SCALARREF HASHREF);

# NACL and EnGenX Libs
use NACL::STask::Aggregate;
use NACL::C::Job;
use NACL::AppBuilder::EnGenX::SubProcess;
use NACL::AppBuilder::EnGenX::EnGenXUtil;
use NACL::AppBuilder::EnGenX::Controller::Volume;
# Exceptions
use NATE::BaseException qw(:try);
use NATE::Exceptions::Argument qw(:try);

# Inherit base class
use base qw( NACL::AppBuilder::EnGenX::Controller::Task );

use vars qw(%DTD_To_NACL_Translate
    %DTD_To_NACL_Ignore
    %Delete_Args);

# Arguments after converting from xml to a
# resource list that need their keys translated to
# keys that NACL methods will accept.
%DTD_To_NACL_Translate = (
    name       => 'aggregate',
    num_disks  => 'diskcount',
    _reuse     => 'if_exists',
    disk_owner => 'node',
    'storage-pool' => 'nacltask_storage_pool',
    'allocation-units' => 'nacltask_allocation_unit',
    'enable-hybrid'    => 'nacltask_enable_hybrid',
    'use_partitions'   => 'nacltask_use_partitions',
    'no_ssd'           => 'nacltask_no_ssd',
);

# Arguments that are used by EnGenX but cannot be
# passed into the NACL methods
%DTD_To_NACL_Ignore = (
    wafliron => '',
    state    => '',
    parallel => '',
);

# Arguments that can be passed into the NACL delete
# command
%Delete_Args = (
    command_interface      => '',
    apiset_must            => '',
    apiset_should          => '',
    aggregate              => '',
    force                  => '',
    job_component          => '',
    'ignore-vldb'          => '',
    plex                   => '',
    'dblade-only'          => '',
    'hidden-plex'          => '',
    'preserve-config-data' => '',
);

=head1 METHODS

=head2 create

(Static method) Request creation of one or multiple aggregates serially
or in parallel and return whether the aggregate(s) was created successfully.
This method passes all user arguments through to NACL::AppBuilder::EnGenX::Controller::Task
and adds specific arguments used by the EnGenX framework.

  NACL::AppBuilder::EnGenX::Controller::Aggregate->create(
            command_interface => $Node,
            aggregate         => $Aggr_name,
            diskcount         => 3,
            job_component     => \$job,
            'method-timeout'  => 3600,
            );

  NACL::AppBuilder::EnGenX::Controller::Aggregate->create(
            command_interface => $Node,
            aggregate         => $Aggr_name,
            diskcount         => 3,
            job_component     => \$job,
            'method-timeout'  => 3600,
            count             => 5
            );

  NACL::AppBuilder::EnGenX::Controller::Aggregate->create(
            command_interfac  => $Node,
            resource_list      => @Aggregates,
            'method-timeout'  => 3600,
            );

  NACL::AppBuilder::EnGenX::Controller::Aggregate->create(
            command_interface => $Node,
            resource_list      => @Aggregates,
            );

  NACL::AppBuilder::EnGenX::Controller::Aggregate->create(
            resource_list      => @Aggregates,
            'method-timeout'  => 3600,
            );

  NACL::AppBuilder::EnGenX::Controller::Aggregate->create(
            resource_list      => @Aggregates,
            );

=over

=item Options

Either the resource_list is required or the NACL Aggregate Creation Arguments
are required but both should NOT be used in the same call.

=over

=item C<< resource_list => @resource_list >>

(Optional)
If this is passed in the aggregates in the resource list will be created
based on the information that is supplied in the resource list. If
the resource list is passed in the NACL Aggregate Creation Arguments should
NOT be passed in.

=item C<< NACL Aggregate Creation Arguments >>

(Optional)
These are arguments that can be passed into NACL::STask::Aggregate->Create.
If these arguments are passed in the resource_list argument should NOT
be passed in

=item C<< count => $aggregate_count >>

(Optional)
This arguments is used to create multiple aggregates from the same NACL Aggregate
Creation Arguments.  To create the names for the aggregates, the aggregate name
passed in is appended with the index that it was create at starting at 1.

=item C<< command_interface => $NACL_command_interface >>

(Optional)
This is the command_interface that will be used during NACL calls.  If a
command_interface is not passed in then one will be created.  Passing in a
command_interface helps performance because for every command_interface created
the version of the Filer must be determined by creating a connection to the filer
and then sending commands to get the version.

=item C<< method_timeout => $NACL_method_timeout >>

(Optional)
This is the method_timeout that will be used during NACL calls. If a
method_timeout is not passed in then the method_timeout will be defaulted.

=back

=item Return

This method returns a True or False based on whether the method succeeds or fails
in the creation of the Aggregate(s) based on the Options passed in.

=item Exceptions

This method produces and throws no exceptions of its own but pass through any
uncaught exceptions from other library calls.

=cut

sub create {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %args = @_;

    # Locals
    my $parent_answer;

    # Globals that will also be set as globals in the base class
    $args{'Name'} = "aggregate";

    # Super::_create is in Task.pm
    $parent_answer = $pkg->SUPER::_create(%args);

    return $parent_answer;
} ## end sub create

=head2 delete

(Static method) Request deletion of one or multiple aggregates serially
or in parallel and return whether the aggregate(s) was deleted successfully.
This method passes all user arguments through to NACL::AppBuilder::EnGenX::Controller::Task
and adds specific arguments used by the EnGenX framework.

  NACL::AppBuilder::EnGenX::Controller::Aggregate->delete(
            command_interface => $Node,
            aggregate           => $Aggr_name,
            'method-timeout'  => 3600,
            );

  NACL::AppBuilder::EnGenX::Controller::Aggregate->delete(
            command_interface => $Node,
            aggregate           => $Aggr_name,
            count             => 5
            );

  NACL::AppBuilder::EnGenX::Controller::Aggregate->delete(
            command_interface => $Node,
            resource_list      => @Aggregates,
            'method-timeout'  => 3600,
            );

  NACL::AppBuilder::EnGenX::Controller::Aggregate->delete(
            command_interface => $Node,
            resource_list      => @Aggregates,
            );

  NACL::AppBuilder::EnGenX::Controller::Aggregate->delete(
            resource_list      => @Aggregates,
            'method-timeout'  => 3600,
            );

  NACL::AppBuilder::EnGenX::Controller::Aggregate->delete(
            resource_list      => @Aggregates,
            );

=over

=item Options

Either the resource_list is required or the NACL Aggregate Deletion Arguments
are required but both should NOT be used in the same call.

=over

=item C<< resource_list => @resource_list >>

(Required)
If this is passed in the aggregates in the resource list will be deleted
based on the information that is supplied in the resource list. If
the resource list is passed in the NACL Aggregate Deletion Arguments should
NOT be passed in.

=item C<< NACL Aggregate Deletion Arguments >>

(Required)
These are arguments that can be passed into NACL::STask::Aggregate->Delete.
If these arguments are passed in the resource_list argument should NOT
be passed in.

=item C<< count => $aggregate_count >>

(Optional)
This arguments is used to delete multiple aggregates from the same NACL Aggregate
Deletion Arguments.  To delete the names for the aggregates, the aggregate name
passed in is appended with the index that it was deleted at starting at 1.

=item C<< command_interface => $NACL_command_interface >>

(Optional)
This is the command_interface that will be used during NACL calls.  If a
command_interface is not passed in then one will be created.  Passing in a
command_interface helps performance because for every command_interface created
the version of the Filer must be determined by creating a connection to the filer
and then sending commands to get the version.

=item C<< method_timeout => $NACL_method_timeout >>

(Optional)
This is the method_timeout that will be used during NACL calls. If a
method_timeout is not passed in then the method_timeout will be defaulted.

=back

=item Return

This method returns a True or False based on whether the method succeeds or fails
in the deletion of the Aggregate(s) based on the Options passed in.

=item Exceptions

This method produces and throws no exceptions of its own but pass through any
uncaught exceptions from other library calls.

=cut

sub delete {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %args = @_;

    # Locals
    my $parent_answer;

    # Globals that will also be set as globals in the base class
    $args{'Name'} = "aggregate";

    # Super::_delete is in Task.pm
    $parent_answer = $pkg->SUPER::_delete(%args);

    return $parent_answer;
} ## end sub delete


#==============================================================================
# Method:     _nacl_create_call
# Objective:  Makes the raw NACL calls for creation of a aggregate.
# Details:    Makes the NACL::STask::Aggregate->create() call and
#             makes the NACL::STask::Aggregate->wait_for_creation call. Also
#             works around some common non-fatal errors that can be return from
#             aggregate creation and then has a retry mechanism to retry
#             aggregate creation.
# TODO:       Need to update to NACL::STask::Aggregate->create()
# Inputs:     NACL::STask::Aggregate->create arguments
#             retry => integer, is the current retry the method is on.
#                    This is used to count how many times the _nacl_create_call
#                    has been retried so that the retries can be limited.
# Outputs:    Returns whether the create was successful or a hash with success
#             and whether the aggregate was a duplicate if a hash is requested
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
#==============================================================================
sub _nacl_create_call {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %args = @_;

    # Locals
    my $was_created = 0;

    my $if_exists      = exists($args{if_exists}) ? delete($args{if_exists}) : 'die';
    my @add_disks = [split(/,/ , delete $args{'add-disks'})] if (exists($args{'add-disks'}));

    if (scalar(@add_disks)) {
        $args{nacltask_add_disks} = {'disk-list' => @add_disks,
                                     'disktype' => 'SSD'};
        # Override disktype if it is defined in the args
        $args{nacltask_add_disks}->{disktype} = $args{'add-disks-disktype'} if (exists($args{'add-disks-disktype'}));
        delete $args{'add-disks-disktype'};
    } elsif (defined $args{'add-disks-count'}) {
        # the case where count is mentioned and not the disk names.
        my $diskcount = delete($args{'add-disks-count'});
        $args{nacltask_add_disks} = {'diskcount' => $diskcount,
                                     'disktype' => 'SSD'};
        # Override disktype if it is defined in the args
        $args{nacltask_add_disks}->{disktype} = $args{'add-disks-disktype'} if (exists($args{'add-disks-disktype'}));
        delete $args{'add-disks-disktype'};
    }


    if (!exists($args{command_interface})) {
        my @Nodes = NACL::C::Node->find();
        $args{command_interface} = $Nodes[0];
    }


    # Default diskcount to 5 if user did not pass a diskcount or disklist.
    $args{diskcount} = 5 unless (exists($args{diskcount}) or exists($args{disklist}));

    # For fullsteam and later, use node instead of nodes.
    # only if nodes was specified and node is wasnt set already
    if ($args{command_interface}->version_manager()->is_version_ge(release => '8.3')) {
        if (exists($args{nodes})) {
            my $nodes = delete $args{nodes};
            if (not exists($args{node})) {
                # update node if it wasnt set
                $args{node} = $nodes->[0];
            }
        }
    }

    if ($args{command_interface}->version_manager()->is_version_ge(release => '8.3') && $args{command_interface}->version_manager()->is_version_lt(release => '8.4')) {
        # burt854829: in FS and later block type is now deprecated since all aggrs are now 
        # 64 bit
        delete($args{'block-type'});
    }

    NACL::STask::Aggregate->create(%args,
                                   nacltask_if_exists => $if_exists,
                                   _was_created => \$was_created);
    if($was_created) {
        $Log->comment("Using newly Created Aggregate:" . $args{aggregate});
    } else {
        $Log->comment("Using already Created Aggregate:" . $args{aggregate});
    }

    $Log->exit() if $may_exit;
    # Returning either a list (hash) or scalar TRUE
    return (wantarray()) ? ('success'=>TRUE, 'duplicated'=>!$was_created) : TRUE;

} ## end sub _nacl_create_call

#==============================================================================
# Method:     _nacl_delete_call
# Objective:  Makes the raw NACL calls for deletion of a aggregate.
# Details:    Makes the NACL::STask::Aggregate->purge() calls. Also works around
#             some common non-fatal errors that can be return from aggregate
#             deletion and then has a retry mechanism to retry aggregate deletion.
# Inputs:     NACL::STask::Aggregate->purge arguments
#             retry => integer, is the current retry the method is on.
#                    This is used to count how many times the _nacl_delete_call
#                    has been retried so that the retries can be limited.
# Outputs:    Returns whether the create was successful or a hash with success
#             if a hash is requested
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
#==============================================================================
sub _nacl_delete_call {
    $Log->enter() if $may_enter;
    my $pkg   = shift;
    my %args = @_;
    # Locals
    my $was_deleted = 0;
    $args{'method-timeout'} = 3600 unless exists($args{'method-timeout'});
    my $end_time = time() + $args{'method-timeout'};

    if (!exists($args{command_interface})) {
        my @Nodes = NACL::C::Node->find();
        $args{command_interface} = $Nodes[0];
    }

    while (1) {
        my $delete_suceeded = TRUE;
        try {
            NACL::STask::Aggregate->purge(%args,
                                          nacltask_if_purged => 'pass',
                                          _was_deleted       => \$was_deleted);
        }
        catch NACL::APISet::Exceptions::ResponseException with {
            my $exception = shift;
            if (time() < $end_time
                && $exception->text() =~ /aggregate is currently busy/)
            {
                $delete_suceeded = FALSE;
            } else {
                $Log->exit() if $may_exit;

                # Rethrowing caught exception
                $exception->throw();
            }
        };
        last if ($delete_suceeded);
        $Log->trace("Snoozing for 10 seconds");
        Tharn::snooze(10);
    }

    if($was_deleted) {
        $Log->comment("Aggregate:" . $args{aggregate} . " was deleted");
    } else {
        $Log->comment("Cannot cleanup Aggregate:" . $args{aggregate} .
                      ". It does not exist in the cluster.\n");
    }

    $Log->exit() if $may_exit;
    return TRUE;

} ## end sub _nacl_delete_call

#==============================================================================
# Method:     _nacl_wait_call
# Objective:  Makes a NACL call to 'wait_for_creation.
# Details:    Makes the call to NACL::STask::Aggregate->wait_for_creation();
#             that will wait for the node to finish creating the aggr.
# Inputs:     NACL::STask::Aggregate->wait arguments
# Outputs:    Returns whether the create was successful TRUE or not FALSE
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
#==============================================================================
sub _nacl_wait_call {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %args = @_;

    $Log->exit() if $may_exit;
    return TRUE;
}

#==============================================================================
# Method:     _get_dtd_nacl_hashes (translate, ignore and delete)
# Objective:  Getter methods for the variables of the same name.
# Details:    Retrieves the hash or a reference to one of the following:
#             - DTD_To_NACL_Translate.
#             - DTD_To_NACL_Ignore.
#             - Delete_Args.
# Inputs:     none.
# Outputs:    The reference or hash variable.
# Exceptions:
#==============================================================================
sub _get_dtd_nacl_translate {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = @_;

    $Log->exit() if $may_exit;
    return (wantarray()) ? %DTD_To_NACL_Translate : \%DTD_To_NACL_Translate;
}
#==============================================================================
sub _get_dtd_nacl_ignore {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = @_;

    $Log->exit() if $may_exit;
    return (wantarray()) ? %DTD_To_NACL_Ignore : \%DTD_To_NACL_Ignore;
}
#==============================================================================
sub _get_dtd_nacl_delete {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = @_;

    $Log->exit() if $may_exit;
    return (wantarray()) ? %Delete_Args : \%Delete_Args;
}

1;
