# Copyright (c) 2015 NetApp Inc.
# All rights reserved.
#
## @summary Provide a common set of libraries to work on controllers
## @author dl-san-automation
## @status private
## @pod here

=head1 NAME

NACL::AppBuilder::EnGenX::Controller::CIFS::Privilege

=head1 DESCRIPTION

NACL::AppBuilder::EnGenX::Controller::CIFS::Privilege is used to Config, Setup, Cleanup and
Deconfig CIFS Privileges on a vserver.

=cut

package NACL::AppBuilder::EnGenX::Controller::CIFS::Privilege;

# Compiler directives
use strict;
use warnings;

# NATE and Standard Module import
use NATE::Log qw(log_global);
use NATE::Util::DataTypes qw(TRUE FALSE);
use Params::Validate
    qw(validate_with validate SCALAR ARRAYREF BOOLEAN UNDEF SCALARREF HASHREF);

# NACL libs
use NACL::STask::VserverCifsUsersAndGroupsPrivilege;

# Inherit base class
use base qw( NACL::AppBuilder::EnGenX::Controller::Task );

use vars qw(%DTD_To_NACL_Translate
            %DTD_To_NACL_Ignore
            %Delete_Args
            $Log);

$Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();

# Arguments after converting from xml to a
# resource list that need their keys translated to
# keys that NACL methods will accept.
%DTD_To_NACL_Translate = (
    _reuse => 'if_exists'
);

# Arguments that are used by EnGenX but cannot be
# passed into the NACL methods
%DTD_To_NACL_Ignore = (
    cifs_service_username =>'',
    cifs_service_password =>'',
    cifs_service_domain   =>'',
    if_exists             =>'',
    'member-names'        =>'',
);

# Arguments that can be passed into the NACL delete
# command
%Delete_Args = (
    vserver     => '',
    'user-or-group-name' => ''
);

=head1 METHODS

=head2 create

(Static method) Request to create a cifs privilege with the following options

  NACL::AppBuilder::EnGenX::Controller::Privilege->create(
            command_interface               => $node,
            vserver                         => $vserver_name,
            'user-or-group-name'            => $name,
            privileges                      => $privileges
            );
=over

=item Options

=over

=item C<< command_interface => $node >>

(Required)
This is the command_interface that will be used during NACL calls.  If a
command_interface is not passed in then one will be created.  Passing in a
command_interface helps performance because for every command_interface created
the version of the Filer must be determined by creating a connection to the filer
and then sending commands to get the version.

=item C<< vserver => $vserver >>

(Required)
The name of the vserver.

=item C<< 'user-or-group-name' => $name>>

(Required)
The name of the local or Active Directory user or group.

=item C<< privileges => $privileges>>

(Required)
The privileges of the local-user or local group.

=back

=item Return

This method returns a True or False based on whether the method succeeds or fails
in the creation of the Local_User based on the Options passed in.

=item Exceptions

This method produces and throws no exceptions of its own but pass through any
uncaught exceptions from other library calls.

=cut
sub create {
    $Log->enter() if $may_enter;
    my ($pkg, %args) = @_;

    # Locals
    my $parent_answer;

    # Globals that will also be set as globals in the base class
    $args{'Name'} = "cifs_privilege";

    # Super::_create is in Task.pm
    $parent_answer = $pkg->SUPER::_create(%args);

    $Log->exit() if $may_exit;

    return $parent_answer;
}

=head2 delete

(Static method) Request to delete a cifs local-user with the following options

  NACL::AppBuilder::EnGenX::Controller::Local-User->delete(
            command_interface               => $node,
            vserver                         => $vserver_name,
            'user-or-group-name'            => $name
            );

=over

=item Options

=over

=item C<< command_interface => $node >>

(Required)
This is the command_interface that will be used during NACL calls.  If a
command_interface is not passed in then one will be created.  Passing in a
command_interface helps performance because for every command_interface created
the version of the Filer must be determined by creating a connection to the filer
and then sending commands to get the version.

=item C<< vserver => $vserver >>

(Required)
The vserver the cifs service is on.

=item C<< 'user-or-group-name' => $user_name >>

(Required)
The user-name of the local-user.

=back

=item Return

This method returns a True or False based on whether the method succeeds or fails
in the delete of the Local_User based on the Options passed in.

=item Exceptions

This method produces and throws no exceptions of its own but pass through any
uncaught exceptions from other library calls.

=cut
sub delete {
    $Log->enter() if $may_enter;
    my ($pkg, %args) = @_;

    # Locals
    my $parent_answer;

    # Globals that will also be set as globals in the base class
    $args{'Name'} = "cifs_privileges";

    # Super::_create is in Task.pm
    #no-op
    $Log->exit() if $may_exit;

    return $parent_answer;
}

#==============================================================================
# Method:     _nacl_create_call
# Objective:  Makes the raw NACL calls for creation of the cifs Privilege.
# Details:    Makes the NACL::STask::VserverCIFSUsersAndGroupsPrivilege->create().
#
# Inputs:     NACL::Stask::VserverCIFSUsersAndGroupsPrivilege->create arguments.
# Outputs:    Returns whether the create was successful.
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
#==============================================================================
sub _nacl_create_call {
    $Log->enter() if $may_enter;
    my ($pkg, %args) = @_;

    # Locals
    my $was_created = 0;
    $args{'method-timeout'} = 3600 unless exists($args{'method-timeout'});

    if (!exists($args{command_interface})) {
        my @Nodes = NACL::C::Node->find();
        $args{command_interface} = $Nodes[0];
    }

    # Arguments for add-privilege
    my $privileges = exists($args{privileges}) ? delete($args{privileges}) : undef;
    if ($privileges) {
        if(ref($privileges) eq 'ARRAY') {
            $args{privileges} = $privileges;
        } else {
            my @privileges;
            @privileges = split(",", $privileges);
            # privileges must be an ARRAYREF
            $args{privileges} = \@privileges;
        }
    }
    
    NACL::STask::VserverCifsUsersAndGroupsPrivilege->add_privilege(%args);
                                                                    
    $Log->comment("Added new privilege to CIFS Local-User: ". $args{'user-or-group-name'});

    return (wantarray()) ? ('success' => TRUE, 'duplicated' => !$was_created) : TRUE;
}

#==============================================================================
# Method:     _nacl_delete_call
# Objective:  Makes the raw NACL calls for deleting the cifs local-user.
# Details:    Makes the NACL::STask::VserverCifsUsersAndGroupsLocalUser->delete().
#
# Inputs:     NACL::STask::VserverCifsUsersAndGroupsLocalUser->delete arguments.
# Outputs:    Returns whether the delete was successful.
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
#==============================================================================
sub _nacl_delete_call {
    $Log->enter() if $may_enter;
    my ($pkg, %args) = @_;

    $Log->comment("Privileges get deleted on deleting the user/group");

    $Log->exit() if $may_exit;
    return TRUE;
}

#==============================================================================
# Method:     _nacl_wait_call
# Objective:  NOOP for CIFS
# Details:
# Inputs:     None
# Outputs:    None
# Exceptions: None
#==============================================================================
sub _nacl_wait_call {
    $Log->enter() if $may_enter;
    my ($pkg, %args) = @_;

    $Log->exit() if $may_exit;
    return TRUE;
}

#==============================================================================
# Method:     _get_dtd_nacl_hashes (translate, ignore and delete)
# Objective:  Getter methods for the variables of the same name.
# Details:    Retrieves the hash or a reference to one of the following:
#             - DTD_To_NACL_Translate.
#             - DTD_To_NACL_Ignore.
#             - Delete_Args.
# Inputs:     none.
# Outputs:    The reference or hash variable.
# Exceptions:
#==============================================================================
sub _get_dtd_nacl_translate {
    $Log->enter() if $may_enter;
    my ($pkg, %args) = @_;

    $Log->exit() if $may_exit;
    return (wantarray()) ? %DTD_To_NACL_Translate : \%DTD_To_NACL_Translate;
}
#==============================================================================
sub _get_dtd_nacl_ignore {
    $Log->enter() if $may_enter;
    my ($pkg, %args) = @_;

    $Log->exit() if $may_exit;
    return (wantarray()) ? %DTD_To_NACL_Ignore : \%DTD_To_NACL_Ignore;
}
#==============================================================================
sub _get_dtd_nacl_delete {
    $Log->enter() if $may_enter;
    my ($pkg, %args) = @_;

    $Log->exit() if $may_exit;
    return (wantarray()) ? %Delete_Args : \%Delete_Args;
}

1;
