# Copyright (c) 2012-2014 NetApp Inc.
# All rights reserved.
#
## @summary Provide a common set of libraries to work on controllers
## @author dl-san-automation
## @status private
## @pod here

=head1 NAME

NACL::AppBuilder::EnGenX::Controller::CIFS::Security

=head1 DESCRIPTION

NACL::AppBuilder::EnGenX::Controller::CIFS::Security is used to Config, Setup, Cleanup and
Deconfig CIFS Security settings on a vserver.

=cut

package NACL::AppBuilder::EnGenX::Controller::CIFS::Security;

# Compiler directives
use strict;
use warnings;

# NATE and Standard Module import
use NATE::Log qw(log_global);
use NATE::Util::DataTypes qw(TRUE FALSE);
use Params::Validate
    qw(validate_with validate SCALAR ARRAYREF BOOLEAN UNDEF SCALARREF HASHREF);

# NACL libs
use NACL::STask::VserverCifsSecurity;

# Inherit base class
use base qw( NACL::AppBuilder::EnGenX::Controller::Task );

use vars qw(%DTD_To_NACL_Translate
            %DTD_To_NACL_Ignore
            %Delete_Args
            $Log);

$Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();

# Arguments after converting from xml to a
# resource list that need their keys translated to
# keys that NACL methods will accept.
%DTD_To_NACL_Translate = (
    _reuse => 'if_exists',
    cifs_service_username => 'username',
    cifs_service_password => 'password'
);

# Arguments that are used by EnGenX but cannot be
# passed into the NACL methods
%DTD_To_NACL_Ignore = (
    cifs_service_domain => '',
);

# Arguments that can be passed into the NACL delete
# command
%Delete_Args = (
);

=head1 METHODS

=head2 create

(Static method) Request to create/modify a Security settings for a CIFS service with the following options

  NACL::AppBuilder::EnGenX::Controller::CIFS::Security->create(
            command_interface               => $node,
            vserver                         => $vserver_name,
            'kerberos-clock-skew            => $clockSkew,
            'kerberos-ticket-age            => $ticketAge,
            'kerberos-renew-age'            => $renewAge,
            'is-signing-required'           => $signingSetting,
            'use-start-tls-for-ad-ldap'     => $tlsSetting,
            'is-aes-encryption-enabled'     => $encryptionSetting
            );
=over

=item Options

=over

=item C<< command_interface => $node >>

(Required)
This is the command_interface that will be used during NACL calls.  If a
command_interface is not passed in then one will be created.  Passing in a
command_interface helps performance because for every command_interface created
the version of the Filer must be determined by creating a connection to the filer
and then sending commands to get the version.

=item C<< vserver => $vserver >>

(Required)
The vserver of the cifs service for Security settings to be changed on.

=item C<< 'kerberos-clock-skew' => $clockSkew>>

(Optional)
The the maximum kerberos clock skew amount.

=item C<< 'kerberos-ticket-age' => $ticketAge>>

(Optional)
The maximum Kerberos ticket life.

=item C<< 'kerberos-renew-age' => $renewAge>>

(Optional)
The maximum kerberos ticket renewal days.

=item C<< is-signing-required => $signingSetting>>

(Optional)
The option to require sigining for CIFS traffic.

=item C<< use-start-tls-for-ad-ldap => $tlsSettign>>

(Optional)
The option to use start_tls for AD LDAP connections.

=item C<< is-aes-encryption-enabled => $encryptionSetting>>

(Optional)
The option to use aes encryption.

=back

=item Return

This method returns a True or False based on whether the method succeeds or fails
in the creation of the Security based on the Options passed in.

=item Exceptions

This method produces and throws no exceptions of its own but pass through any
uncaught exceptions from other library calls.

=cut
sub create {
    $Log->enter() if $may_enter;
    my ($pkg, %args) = @_;

    # Locals
    my $parent_answer;

    # Globals that will also be set as globals in the base class
    $args{'Name'} = "cifs_security";

    # Super::_create is in Task.pm
    $parent_answer = $pkg->SUPER::_create(%args);

    $Log->exit() if $may_exit;

    return $parent_answer;
}

=head2 delete

This method does nothing but display a log message. The settings that are created will be deleted when the CIFS vserver is deleted.

This method produces and throws no exceptions of its own but pass through any
uncaught exceptions from other library calls.

=cut
sub delete {
    $Log->enter() if $may_enter;
    my ($pkg, %args) = @_;

    # Locals
    my $parent_answer;

    # Globals that will also be set as globals in the base class
    $args{'Name'} = "cifs_security";

    $Log->comment("The delete subroutine was called, but it does nothing. Setting will be removed when the CIFS vserver is deleted.");
    $Log->exit() if $may_exit;

    return;
}

#==============================================================================
# Method:     _nacl_create_call
# Objective:  Makes the raw NACL calls for creation of the cifs Security.
# Details:    Makes the NACL::STask::VserverCIFSsecurity->modify().
#
# Inputs:     NACL::STask::VserverCIFSSecurity->modify arguments.
# Outputs:    Returns whether the create was successful.
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
#==============================================================================
sub _nacl_create_call {
    $Log->enter() if $may_enter;
    my ($pkg, %args) = @_;
    # Locals
    delete $args{if_exists};
    $args{'method-timeout'} = 3600 unless exists($args{'method-timeout'});

    if (!exists($args{command_interface})) {
        my @Nodes = NACL::C::Node->find();
        $args{command_interface} = $Nodes[0];
    }

    # create the cifs vserver security settings
    NACL::STask::VserverCifsSecurity->modify(%args);

    $Log->comment("Modified CIFS Security settings for the CIFS vserver: "
                      . $args{vserver} . ".");

    $Log->exit() if $may_exit;
    return (wantarray()) ? ('success'=>TRUE) : TRUE;
}

#==============================================================================
# Method:     _nacl_wait_call
# Objective:  NOOP for CIFS
# Details:
# Inputs:     None
# Outputs:    None
# Exceptions: None
#==============================================================================
sub _nacl_wait_call {
    $Log->enter() if $may_enter;
    my ($pkg, %args) = @_;

    $Log->exit() if $may_exit;
    return TRUE;
}

#==============================================================================
# Method:     _get_dtd_nacl_hashes (translate, ignore and delete)
# Objective:  Getter methods for the variables of the same name.
# Details:    Retrieves the hash or a reference to one of the following:
#             - DTD_To_NACL_Translate.
#             - DTD_To_NACL_Ignore.
#             - Delete_Args.
# Inputs:     none.
# Outputs:    The reference or hash variable.
# Exceptions:
#==============================================================================
sub _get_dtd_nacl_translate {
    $Log->enter() if $may_enter;
    my ($pkg, %args) = @_;

    $Log->exit() if $may_exit;
    return (wantarray()) ? %DTD_To_NACL_Translate : \%DTD_To_NACL_Translate;
}
#==============================================================================
sub _get_dtd_nacl_ignore {
    $Log->enter() if $may_enter;
    my ($pkg, %args) = @_;

    $Log->exit() if $may_exit;
    return (wantarray()) ? %DTD_To_NACL_Ignore : \%DTD_To_NACL_Ignore;
}
#==============================================================================
sub _get_dtd_nacl_delete {
    $Log->enter() if $may_enter;
    my ($pkg, %args) = @_;

    $Log->exit() if $may_exit;
    return (wantarray()) ? %Delete_Args : \%Delete_Args;
}

1;
