# Copyright (c) 2001-2015 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Vserver Cifs Domain Preferred-dc Controller Layer Module
## @author ng-ptefit-automation-core <ng-ptefit-automation-core@netapp.com>
## @status shared
## @pod here

=head1 NAME

NACL::AppBuilder::EnGenX::Controller::VserverCifsDomainPreferredDc

=head1 DESCRIPTION

NACL::AppBuilder::EnGenX::Controller::VserverCifsDomainPreferredDc is a derived class of
NACL::AppBuilder::EnGenX::Controller::Task.  It is used to Create and
Delete ONTAP Domain controllers to be used in preference to all others by the specified Vserver.
All methods are static.

=cut

package NACL::AppBuilder::EnGenX::Controller::CIFS::VserverCifsDomainPreferredDc;

# Compiler directives
use strict;
use warnings;

# NATE and Standard Module import
use NATE::Log qw( log_global );
my $Log = log_global();
use NATE::Util::DataTypes qw(TRUE FALSE);

# NACL and EnGenX Libs
use NACL::STask::VserverCifsDomainPreferredDc;

#Inherit base class
use base qw( NACL::AppBuilder::EnGenX::Controller::Task );

use vars qw(%DTD_To_NACL_Translate
    %DTD_To_NACL_Ignore
    %Delete_Args);

# Arguments after converting from xml to a
# resource list that need their keys translated to
# keys that NACL methods will accept.
%DTD_To_NACL_Translate = (
    _reuse => 'if_exists'
);

# Arguments that are used by EnGenX but cannot be
# passed into the NACL methods
%DTD_To_NACL_Ignore = (
    cifs_service_username =>'',
    cifs_service_password =>'',
    parallel => '',
);

# Arguments that can be passed into the NACL delete
# command
%Delete_Args = (
    command_interface => '',
    apiset_must       => '',
    apiset_should     => '',
    vserver           => '',
    'preferred-dc'    => '',
    cifs_service_domain => '',
    domain              => '',
);

=head1 METHODS

=head2 create

(Static method) Request creation of one or multiple domain controllers serially
or in parallel and return whether the domain controller(s) was created successfully.
This method passes all user arguments through to NACL::AppBuilder::EnGenX::Controller::Task
and adds specific arguments used by the EnGenX framework.

  NACL::AppBuilder::EnGenX::Controller::CIFS::VserverCifsDomainPreferredDc->create(
            command_interface   => $command_interface,
            vserver             => $vserver,
            domain              => $domain,
            'preferred-dc'      => $ip_addr,
            'method-timeout'    => 3600,
            );

  NACL::AppBuilder::EnGenX::Controller::CIFS::VserverCifsDomainPreferredDc->create(
            command_interface  => $command_interface,
            resource_list      => @domains,
            'method-timeout'  => 3600,
            );

  NACL::AppBuilder::EnGenX::Controller::CIFS::VserverCifsDomainPreferredDc->create(
            command_interface => $command_interface,
            resource_list      => @domains,
            );

  NACL::AppBuilder::EnGenX::Controller::CIFS::VserverCifsDomainPreferredDc->create(
            resource_list      => @domains,
            'method-timeout'  => 3600,
            );

  NACL::AppBuilder::EnGenX::Controller::CIFS::VserverCifsDomainPreferredDc->create(
            resource_list      => @domains,
            );

=over

=item Options

Either the resource_list is required or the NACL VserverCifsDomainPreferredDc Creation Arguments
are required but both should NOT be used in the same call.

=over

=item C<< resource_list => @resource_list >>

(Optional)
If this is passed in the domains, the resource list will be created
based on the information that is supplied in the resource list. If
the resource list is passed in the NACL VserverCifsDomainPreferredDc Creation Arguments should
NOT be passed in.

=item C<< NACL VserverCifsDomainPreferredDc Creation Arguments >>

(Required)
These are arguments that can be passed into NACL::STask::VserverCifsDomainPreferredDc->add.
If these arguments are passed in the resource_list argument should NOT
be passed in

=item C<< vserver => $vserver >>

(Required)
the name of the vserver for which you want to add preferred domain controllers.

=item C<< domain => $domain >>

(Required)
the fully-qualified name of the domain that the domain controllers belong to.

=item C<< preferred-dc => $ip_addr >>

(Required)
An arrayref of the IP addresses for domain controllers that belong to the domain.

=item C<< command_interface => $NACL_command_interface >>

(Optional)
This is the command_interface that will be used during NACL calls.  If a
command_interface is not passed in then one will be created.  Passing in a
command_interface helps performance because for every command_interface created
the version of the Filer must be determined by creating a connection to the filer
and then sending commands to get the version.

=item C<< method_timeout => $NACL_method_timeout >>

(Optional)
This is the method_timeout that will be used during NACL calls. If a
method_timeout is not passed in then the method_timeout will be defaulted.

=back

=item Return

This method returns a True or False based on whether the method succeeds or fails
in the creation of the VserverCifsDomainPreferredDc(s) based on the Options passed in.

=item Exceptions

This method produces and throws no exceptions of its own but pass through any
uncaught exceptions from other library calls.

=cut

sub create {
    $Log->enter();
    my $pkg  = shift;
    my %args = @_;

    # Locals
    my $parent_answer;

    # Globals that will also be set as globals in the base class
    $args{'Name'} = "cifs-domain-preferred-dc";

    # Super::_create is in Task.pm
    $parent_answer = $pkg->SUPER::_create(%args);

    $Log->exit();

    return $parent_answer;
} ## end sub create

=head2 delete

(Static method) Request deletion of one or multiple domain controllers serially
or in parallel and return whether the domain controller(s) was deleted successfully.
This method passes all user arguments through to NACL::AppBuilder::EnGenX::Controller::Task
and adds specific arguments used by the EnGenX framework.


  NACL::AppBuilder::EnGenX::Controller::CIFS::VserverCifsDomainPreferredDc->delete(
            command_interface   => $command_interface,
            vserver             => $vserver,
            domain              => $domain,
            'preferred-dc'      => $ip_addr,
            'method-timeout'    => 3600,
            );

  NACL::AppBuilder::EnGenX::Controller::CIFS::VserverCifsDomainPreferredDc->delete(
            command_interface  => $command_interface,
            resource_list      => @domains,
            'method-timeout'  => 3600,
            );

  NACL::AppBuilder::EnGenX::Controller::CIFS::VserverCifsDomainPreferredDc->delete(
            command_interface => $command_interface,
            resource_list      => @domains,
            );

  NACL::AppBuilder::EnGenX::Controller::CIFS::VserverCifsDomainPreferredDc->delete(
            resource_list      => @domains,
            'method-timeout'  => 3600,
            );

  NACL::AppBuilder::EnGenX::Controller::CIFS::VserverCifsDomainPreferredDc->delete(
            resource_list      => @domains,
            );

=over

=item Options

Either the resource_list is required or the NACL VserverCifsDomainPreferredDc Deletion Arguments
are required but both should NOT be used in the same call.

=over

=item C<< resource_list => @resource_list >>

(Required)
If this is passed in the domain controllers, the resource list will be deleted
based on the information that is supplied in the resource list. If
the resource list is passed in the NACL VserverCifsDomainPreferredDc Deletion Arguments should
NOT be passed in.

=item C<< NACL VserverCifsDomainPreferredDc Deletion Arguments >>

(Required)
These are arguments that can be passed into NACL::STask::VserverCifsDomainPreferredDc->purge.
If these arguments are passed in the resource_list argument should NOT
be passed in.

=item C<< vserver => $vserver >>

(Required)
the name of the vserver for which you want to add preferred domain controllers.

=item C<< domain => $domain >>

(Required)
the fully-qualified name of the domain that the domain controllers belong to.

=item C<< preferred-dc => $ip_addr >>

(Optional)
An arrayref of the IP addresses for domain controllers that belong to the domain.
If this is not provided, all preferred domain controllers are removed for the vserver.

=item C<< command_interface => $NACL_command_interface >>

(Optional)
This is the command_interface that will be used during NACL calls.  If a
command_interface is not passed in then one will be created.  Passing in a
command_interface helps performance because for every command_interface created
the version of the Filer must be determined by creating a connection to the filer
and then sending commands to get the version.

=item C<< method_timeout => $NACL_method_timeout >>

(Optional)
This is the method_timeout that will be used during NACL calls. If a
method_timeout is not passed in then the method_timeout will be defaulted.

=back

=item Return

This method returns a True or False based on whether the method succeeds or fails
in the deletion of the Controller(s) based on the Options passed in.

=item Exceptions

This method produces and throws no exceptions of its own but pass through any
uncaught exceptions from other library calls.

=cut

sub delete {
    $Log->enter();
    my $pkg  = shift;
    my %args = @_;

    # Locals
    my $parent_answer;

    # Globals that will also be set as globals in the base class
    $args{'Name'} = "cifs-domain-preferred-dc";

    # Super::_delete is in Task.pm
    $parent_answer = $pkg->SUPER::_delete(%args);
    $Log->exit();
    return $parent_answer;
} ## end sub delete


#==============================================================================
# Method:     _nacl_create_call
# Objective:  Makes the raw NACL calls for creation of domain controllers.
# Details:    Makes the NACL::STask::VserverCifsDomainPreferredDc->add() call. Also works
#             around some common non-fatal errors that can be return from
#             controller creation.
# Inputs:     NACL::STask::VserverCifsDomainPreferredDc->create arguments
# Outputs:    Returns whether the create was successful or a hash with success
#             if a hash is requested
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
#==============================================================================
sub _nacl_create_call {
    $Log->enter();
    my $pkg  = shift;
    my %args = @_;
    my $domain;
    # Locals
    my $if_exists
        = exists($args{if_exists}) ? delete($args{if_exists}) : 'die';
    $args{'method-timeout'} = 3600 unless exists($args{'method-timeout'});

    if (!exists($args{command_interface})) {
        my @Nodes = NACL::C::Node->find();
        $args{command_interface} = $Nodes[0];
    }

    if (exists($args{'preferred-dc'})){
         my $preferred_dc = $args{'preferred-dc'};
         my @preferred_dc = split(/,/, $preferred_dc);
         $args{'preferred-dc'} = \@preferred_dc;
    }
    
    if (exists($args{domain})){
        delete ($args{cifs_service_domain});
        $domain = $args{domain};
    } else {
        $domain = $args{cifs_service_domain};
        delete ($args{cifs_service_domain});
    }
    # Execute Create
    NACL::STask::VserverCifsDomainPreferredDc->add(%args,
                                                   domain => $domain);

    $Log->exit();
    # Returning either a list (hash) or scalar TRUE
    return TRUE;


} ## end sub _nacl_create_call

#========================================================================================
# Method:     _nacl_delete_call
# Objective:  Makes the raw NACL calls for deletion of a domain controller.
# Details:    Makes the NACL::STask::VserverCifsDomainPreferredDc->purge() calls. Also works around
#             some common non-fatal errors that can be return from controller
#             deletion.
# Inputs:     NACL::STask::VserverCifsDomainPreferredDc->purge arguments
# Outputs:    Returns whether the delete was successful or a hash with success
#             if a hash is requested
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
#==============================================================================
sub _nacl_delete_call {
    $Log->enter();
    my $pkg  = shift;
    my %args = @_;
    my $domain;

    # Locals
    my $was_deleted = 0;
    $args{'method-timeout'} = 3600 unless exists($args{'method-timeout'});
    if (!exists($args{command_interface})) {
        my @Nodes = NACL::C::Node->find();
        $args{command_interface} = $Nodes[0];
    }

    if (exists($args{'preferred-dc'})){
         my $preferred_dc = $args{'preferred-dc'};
         my @preferred_dc = split(/,/, $preferred_dc);
         $args{'preferred-dc'} = \@preferred_dc;
    }

    if (exists($args{domain})){
        delete ($args{cifs_service_domain});
        $domain = $args{domain};
    } else {
        $domain = $args{cifs_service_domain};
        delete ($args{cifs_service_domain});
    }

    NACL::STask::VserverCifsDomainPreferredDc->purge(%args,
                            domain => $domain,
                            nacltask_if_purged => 'pass',
                            _was_deleted => \$was_deleted);

    if($was_deleted) {
        $Log->comment("Vserver Cifs Domain Preferred-dc was deleted");
    } else {
        $Log->comment("Vserver Cifs Domain Preferred-dc was not deleted");
    }

    $Log->exit();
    return TRUE;

} ## end sub _nacl_delete_call
#==============================================================================
# Method:     _nacl_wait_call
# Objective:  Makes a NACL call to 'wait_for_creation.
# Details:    Makes the call to NACL::STask::VserverCifsDomainPreferredDc->wait_for_creation();
#             that will wait for the node to finish creating the aggr.
# Inputs:     NACL::STask::VserverCifsDomainPreferredDc->wait arguments
# Outputs:    Returns whether the create was successful TRUE or not FALSE
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
# TODO: Implement this method when needed
#==============================================================================
sub _nacl_wait_call {
    $Log->enter();
    my $pkg  = shift;
    my %opts = @_;

    $Log->exit();
    return TRUE;
}

#==============================================================================
# Method:     _get_dtd_nacl_hashes (translate, ignore and delete)
# Objective:  Getter methods for the variables of the same name.
# Details:    Retrieves the hash or a reference to one of the following:
#             - DTD_To_NACL_Translate.
#             - DTD_To_NACL_Ignore.
#             - Delete_Args.
# Inputs:     none.
# Outputs:    The reference or hash variable.
# Exceptions:
#==============================================================================
sub _get_dtd_nacl_translate {
    $Log->enter();
    my $pkg  = shift;
    my %opts = @_;

    $Log->exit();
    return (wantarray()) ? %DTD_To_NACL_Translate : \%DTD_To_NACL_Translate;
}
#==============================================================================
sub _get_dtd_nacl_ignore {
    $Log->enter();
    my $pkg  = shift;
    my %opts = @_;

    $Log->exit();
    return (wantarray()) ? %DTD_To_NACL_Ignore : \%DTD_To_NACL_Ignore;
}
#==============================================================================
sub _get_dtd_nacl_delete {
    $Log->enter();
    my $pkg  = shift;
    my %opts = @_;

    $Log->exit();
    return (wantarray()) ? %Delete_Args : \%Delete_Args;
}

1;
