# Copyright (c) 2012-2014 NetApp Inc.
# All rights reserved.
#
## @summary Provide a common set of libraries to work on controllers
## @author dl-san-automation
## @status private
## @pod here

=head1 NAME

NACL::AppBuilder::EnGenX::Controller::CIFSShare

=head1 DESCRIPTION

NACL::AppBuilder::EnGenX::Controller::CIFSSshare is used to Config, Setup, Cleanup and
Deconfig CIFS Share objects on a vserver.

=cut

package NACL::AppBuilder::EnGenX::Controller::CIFSShare;

# Compiler directives
use strict;
use warnings;

# NATE and Standard Module import
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use NATE::Util::DataTypes qw(TRUE FALSE);
use Data::Dumper;
use Params::Validate qw(SCALAR ARRAYREF BOOLEAN UNDEF SCALARREF HASHREF);

# NACL and EnGenX Libs
use NACL::AppBuilder::EnGenX::SubProcess;
use NACL::AppBuilder::EnGenX::EnGenXUtil;
use NACL::STask::VserverCifsShare;

# Exceptions
use NATE::BaseException qw(:try);
use NATE::Exceptions::Argument qw(:try);

# Inherit base class
use base qw( NACL::AppBuilder::EnGenX::Controller::Task );

use vars qw(%DTD_To_NACL_Translate
            %DTD_To_NACL_Ignore
            %Delete_Args);

# Arguments after converting from xml to a
# resource list that need their keys translated to
# keys that NACL methods will accept.
%DTD_To_NACL_Translate = (
    name   => 'share-name',
    _reuse => 'if_exists'
);

# Arguments that are used by EnGenX but cannot be
# passed into the NACL methods
%DTD_To_NACL_Ignore = (
    parallel      => '',
);

# Arguments that can be passed into the NACL delete
# command
%Delete_Args = (
    command_interface => '',
    apiset_must       => '',
    apiset_should     => '',
    'share-name'      => '',
    'vserver'         => '',
);

=head1 METHODS

=head2 create

(Static method) Request to create a CIFS Share based on the options passed in.

  NACL::AppBuilder::EnGenX::Controller::CIFSShare->create(
            command_interface               => $node,
            vserver                         => $vserver_name,
            'share-name'                    => "my_cifs_share",
            path                            => "/my_junction_path",
            );

=over

=item Options

The NACL component only requires the user to specify the
command interface, vserver, share name, and junction path.
The ontap defaults will be used for any option that is not stated.

=over

=item C<< command_interface => $command_interface >>

(Required)
This is the command_interface that will be used during NACL calls.

=item C<< vserver => $vserver >>

(Required)
This is vserver that the cifs share will be created on

=item C<< 'share-name' => $name >>

(Required)
This is the name to give the cifs share.

=item C<< path => $path >>

(Required)
This is the junction path that the cifs share will be created for.
Junction paths are associated to mounted flexvols.

=back

=item Return

This method returns a True or False based on whether the method succeeds or fails
in the creation of the CIFS shares based on the Options passed in.

=item Exceptions

This method produces and throws no exceptions of its own but pass through any
uncaught exceptions from other library calls.

=cut
sub create {
    $Log->enter() if $may_enter;
    my ($pkg, %args) = @_;

    # Locals
    my $parent_answer;

    # Globals that will also be set as globals in the base class
    $args{'Name'} = "cifs-share";

    # Super::_create is in Task.pm
    $parent_answer = $pkg->SUPER::_create(%args);

    $Log->exit() if $may_exit;
    return $parent_answer;
}

=head2 delete

(Static method) Request to delete a CIFS Share based on the options passed in.

  NACL::AppBuilder::EnGenX::Controller::CIFSShare->delete(
            command_interface               => $node,
            vserver                         => $vserver_name,
            'share-name'                    => "my_cifs_share",
            );

=over

=item Options

The NACL component only requires the user to specify the
command interface, vserver, and share name. The ontap defaults will
be used for any option that is not stated.

=over

=item C<< command_interface => $command_interface >>

(Required)
This is the command_interface that will be used during NACL calls.

=item C<< vserver => $vserver >>

(Required)
This is vserver whose cifs share will be deleted

=item C<< 'share-name' => $name >>

(Required)
The share to delete

=back

=item Return

This method returns a True or False based on whether the method succeeds or fails
in the creation of the CIFS shares based on the Options passed in.

=item Exceptions

This method produces and throws no exceptions of its own but pass through any
uncaught exceptions from other library calls.

=cut
sub delete {
    $Log->enter() if $may_enter;
    my ($pkg, %args) = @_;

    # Locals
    my $parent_answer;

    # Globals that will also be set as globals in the base class
    $args{'Name'} = "cifs-share";

    # Super::_delete is in Task.pm
    $parent_answer = $pkg->SUPER::_delete(%args);

    $Log->exit() if $may_exit;
    return $parent_answer;
}

#==============================================================================
# Method:     _nacl_create_call
# Objective:  Makes the raw NACL calls for creation of the CIFS Share
# Details:    Makes the NACL::STask::VserverCifsShare->create(). Also works
#             around some common non-fatal errors that can be return from
#             volume creation.
# Inputs:     NACL::Stask::VserverCifsShare->create arguments
# Outputs:    Returns whether the create was successful or a hash with success
#             and whether the share was a duplicate if a hash is requested
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
#==============================================================================
sub _nacl_create_call {
    $Log->enter() if $may_enter;
    my ($pkg, %args) = @_;

    # Locals
    my $was_created = 0;
    my $if_exists = exists($args{if_exists}) ? delete($args{if_exists}) : 'die';
    $args{'method-timeout'} = 3600 unless exists($args{'method-timeout'});

    if (!exists($args{command_interface})) {
        my @Nodes = NACL::C::Node->find();
        $args{command_interface} = $Nodes[0];
    }

    # convert share-properties and symlink-properties to array references
    # if they were passed in as a scalar
    if (exists($args{'share-properties'}) and ref($args{'share-properties'}) ne "ARRAY") {
        my @share_properties = split(",", $args{'share-properties'});
        $args{'share-properties'} = \@share_properties;
    }
    if (exists($args{'symlink-properties'}) and ref($args{'symlink-properties'}) ne "ARRAY") {
        my @symlink_properties = split(",", $args{'symlink-properties'});
        $args{'symlink-properties'} = \@symlink_properties;
    }

    # create the share
    NACL::STask::VserverCifsShare->create(%args,
                                          nacltask_if_exists => $if_exists,
                                          nacltask_replay_timeout => 0, # dont wait for replay
                                          _was_created => \$was_created);

    if($was_created) {
        $Log->comment("Using newly created CIFS Share " . $args{'share-name'} .
                            " for vserver " . $args{'vserver'});
    } else {
        $Log->comment("Reusing CIFS Share " . $args{'share-name'} .
                            " for vserver " . $args{'vserver'});
    }

    $Log->exit() if $may_exit;
    return (wantarray()) ? ('success' => TRUE, 'duplicated' => !$was_created) : TRUE;
}

#==============================================================================
# Method:     _nacl_delete_call
# Objective:  Makes the raw NACL calls for the deletion of a CIFS Share
# Details:    Makes the NACL::STask::VserverCifsShare->delete(). Also works
#             around some common non-fatal errors that can be return from
#             volume creation.
# Inputs:     NACL::STask::VserverCifsShare->delete arguments
# Outputs:    Returns whether the delete was successful or a hash with success
#             and whether the share was a duplicate if a hash is requested
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
#==============================================================================
sub _nacl_delete_call {
    $Log->enter() if $may_enter;
    my ($pkg, %args) = @_;

    # Locals
    my $was_deleted = 0;
    $args{'method-timeout'} = 3600 unless exists($args{'method-timeout'});
    if (!exists($args{command_interface})) {
        my @Nodes = NACL::C::Node->find();
        $args{command_interface} = $Nodes[0];
    }

    NACL::STask::VserverCifsShare->purge(%args,
                            nacltask_if_purged => 'pass',
                            nacltask_replay_timeout => 0, # dont wait for replay
                            _was_deleted => \$was_deleted);

    if($was_deleted) {
        $Log->comment("CIFS Share was deleted");
    } else {
        $Log->comment("CIFS Share was not deleted");
    }

    $Log->exit() if $may_exit;
    return;
}

#==============================================================================
# Method:     _nacl_wait_call
# Objective:  Makes a NACL call to 'wait_for_creation.
# Details:    Makes the call to NACL::STask::Vserver->wait_for_creation();
#             that will wait for the node to finish creating the aggr.
# Inputs:     NACL::STask::Vserver->wait arguments
# Outputs:    Returns whether the create was successful TRUE or not FALSE
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
# TODO: Implement this method when needed
#==============================================================================
sub _nacl_wait_call {
    $Log->enter() if $may_enter;
    my ($pkg, %args) = @_;

    $Log->exit() if $may_exit;
    return TRUE;
}

#==============================================================================
# Method:     _get_dtd_nacl_hashes (translate, ignore and delete)
# Objective:  Getter methods for the variables of the same name.
# Details:    Retrieves the hash or a reference to one of the following:
#             - DTD_To_NACL_Translate.
#             - DTD_To_NACL_Ignore.
#             - Delete_Args.
# Inputs:     none.
# Outputs:    The reference or hash variable.
# Exceptions:
#==============================================================================
sub _get_dtd_nacl_translate {
    $Log->enter() if $may_enter;
    my ($pkg, %args) = @_;

    $Log->exit() if $may_exit;
    return (wantarray()) ? %DTD_To_NACL_Translate : \%DTD_To_NACL_Translate;
}
#==============================================================================
sub _get_dtd_nacl_ignore {
    $Log->enter() if $may_enter;
    my ($pkg, %args) = @_;

    $Log->exit() if $may_exit;
    return (wantarray()) ? %DTD_To_NACL_Ignore : \%DTD_To_NACL_Ignore;
}
#==============================================================================
sub _get_dtd_nacl_delete {
    $Log->enter() if $may_enter;
    my ($pkg, %args) = @_;

    $Log->exit() if $may_exit;
    return (wantarray()) ? %Delete_Args : \%Delete_Args;
}

1;
