# $Id$#
#
# Copyright (c) 2001-2013 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary ConstituentVolume Controller Layer Module
## @author dl-san-automation
## @status shared
## @pod here

=head1 NAME

EnGenX::Controller::ConstituentVolume

=head1 DESCRIPTION

EnGenX::Controller::ConstituentVolume is a derived class of
NACL::AppBuilder::EnGenX::Controller::Task.  It is used to Create and
Delete ONTAP Volumes. All methods are static.

=cut

package NACL::AppBuilder::EnGenX::Controller::ConstituentVolume;

# Compiler directives
use strict;
use warnings;

# NATE and Standard Module import
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use NATE::Util::DataTypes qw(TRUE FALSE);
use Data::Dumper;
use Params::Validate
    qw(validate_with validate SCALAR ARRAYREF BOOLEAN UNDEF SCALARREF HASHREF);

# NACL and EnGenX Libs
use NACL::C::Node;
use NACL::CS::Volume;
use NACL::STask::Volume;
use NACL::C::Job;
use NACL::AppBuilder::EnGenX::SubProcess;
use NACL::AppBuilder::EnGenX::EnGenXUtil;

use NATE::Log qw(log_global);

$Log = log_global();
$may_enter = $Log->may_enter();
$may_exit  = $Log->may_exit();

# Exceptions
use NATE::BaseException qw(:try);
use NATE::Exceptions::Argument qw(:try);
use NACL::C::Exceptions::Volume::AlreadyMounted;
use NACL::C::Exceptions::Volume::InvalidDataConstituentName;

# Inherit base class
use base qw( NACL::AppBuilder::EnGenX::Controller::Task );

use vars qw(%DTD_To_NACL_Translate
    %DTD_To_NACL_Ignore
    %Delete_Args
);

# Arguments after converting from xml to a
# resource list that need their keys translated to
# keys that NACL methods will accept.
%DTD_To_NACL_Translate = (
    aggr   => 'aggregate',
    _reuse => 'if_exists',
);

# Arguments that are used by EnGenX but cannot be
# passed into the NACL methods.
%DTD_To_NACL_Ignore = (
    parallel               => ''
);

# Arguments that can be passed into the NACL delete
# command.
%Delete_Args = (
    command_interface       => '',
    apiset_must             => '',
    apiset_should           => '',
    volume                  => '',
    vserver                 => '',
    foreground              => '',
    'disable-offline-check' => '',
    force                   => '',
    job_component           => '',
    'is-repository'         => '',
    enable_flexcache        => ''
);

=head1 METHODS

=head2 create

(Static method) Request creation of one or multiple constituent volumes serially
or in parallel and return whether the volume(s) was created successfully.
This method passes all user arguments through to NACL::AppBuilder::EnGenX::Controller::Task
and adds specific arguments used by the EnGenX framework.

  EnGenX::Controller::ConstituentVolume->create(
            command_interface => $Node,
            volume            => ( $Volume_name),
            aggregate         => $Aggr_name,
            vserver           => $Vserver_name,
            size              => "20M",
            job_component     => \$job,
            'method-timeout'  => 3600,
            );

  EnGenX::Controller::ConstituentVolume->create(
            command_interface => $Node,
            volume            => ( $Volume_name),
            aggregate         => $Aggr_name,
            vserver           => $Vserver_name,
            size              => "20M",
            job_component     => \$job,
            'method-timeout'  => 3600,
            count             => 5,
            );

  EnGenX::Controller::ConstituentVolume->create(
            command_interface  => $Node,
            resource_list      => @Vols,
            'method-timeout'  => 3600,
            );

  EnGenX::Controller::ConstituentVolume->create(
            command_interface => $Node,
            resource_list      => @Vols,
            );

  EnGenX::Controller::ConstituentVolume->create(
            resource_list      => @Vols,
            'method-timeout'  => 3600,
            );

  EnGenX::Controller::ConstituentVolume->create(
            resource_list      => @Vols,
            );

=over

=item Options

Either the resource_list is required or the NACL Volume Creation Arguments
are required but both should NOT be used in the same call.

=over

=item C<< resource_list => @resource_list >>

(Optional)
If this is passed in the volumes in the resource list will be created
based on the information that is supplied in the resource list. If
the resource list is passed in the NACL Volume Creation Arguments should
NOT be passed in.

=item C<< NACL Volume Creation Arguments >>

(Optional)
These are arguments that can be passed into NACL::C::Volume->Create.
If these arguments are passed in the resource_list argument should NOT
be passed in

=item C<< count => $volume_count >>

(Optional)
This arguments is used to create multiple volumes from the same NACL Volume
Creation Arguments.  To create the names for the volumes, the volume name
passed in is appended with the index that it was create at starting at 1.

=item C<< command_interface => $NACL_command_interface >>

(Optional)
This is the command_interface that will be used during NACL calls.  If a
command_interface is not passed in then one will be created.  Passing in a
command_interface helps performance because for every command_interface created
the version of the Filer must be determined by creating a connection to the filer
and then sending commands to get the version.

=item C<< method_timeout => $NACL_method_timeout >>

(Optional)
This is the method_timeout that will be used during NACL calls. If a
method_timeout is not passed in then the method_timeout will be defaulted to 3600.

=back

=item Return

This method returns a True or False based on whether the method succeeds or fails
in the creation of the Volume(s) based on the Options passed in.

=item Exceptions

This method produces and throws no exceptions of its own but pass through any
uncaught exceptions from other library calls.

=cut

sub create {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %args = @_;

    # Locals
    my $parent_answer;

    # TODO Remove $args{'Name'} from everywhere. And ensure
    # no errors occur.
    # NOTE: Can't remove the NAME, it is been used to create objects with a
    # counter. We could get the name from package by asking for the "caller"
    $args{'Name'} = "constituent_volume";

    # Super::_create is in Task.pm
    $parent_answer = $pkg->SUPER::_create(%args);

    return $parent_answer;
} ## end sub create

=head2 delete

(Static method) Request deletion of one or multiple volumes serially
or in parallel and return whether the volume(s) was deleted successfully.
This method passes all user arguments through to NACL::AppBuilder::EnGenX::Controller::Task
and adds specific arguments used by the EnGenX framework.

  EnGenX::Controller::Volume->delete(
            command_interface => $Node,
            volume            => ( $Volume_name),
            vserver           => $Vserver_name,
            'method-timeout'  => 3600,
            );

  EnGenX::Controller::Volume->delete(
            command_interface => $Node,
            volume            => ( $Volume_name),
            vserver           => $Vserver_name,
            'method-timeout'  => 3600,
            count             => 5
            );

  EnGenX::Controller::Volume->delete(
            command_interface => $Node,
            resource_list      => @Vols,
            'method-timeout'  => 3600,
            );

  EnGenX::Controller::Volume->delete(
            command_interface => $Node,
            resource_list      => @Vols,
            );

  EnGenX::Controller::Volume->delete(
            resource_list      => @Vols,
            'method-timeout'  => 3600,
            );

  EnGenX::Controller::Volume->delete(
            resource_list      => @Vols,
            );

=over

=item Options

Either the resource_list is required or the NACL Volume Deletion Arguments
are required but both should NOT be used in the same call.

=over

=item C<< resource_list => @resource_list >>

(Required)
If this is passed in the volumes in the resource list will be deleted
based on the information that is supplied in the resource list. If
the resource list is passed in the NACL Volume Deletion Arguments should
NOT be passed in.

=item C<< NACL Volume Deletion Arguments >>

(Required)
These are arguments that can be passed into NACL::STask::Volume->Delete.
If these arguments are passed in the resource_list argument should NOT
be passed in.

=item C<< count => $volume_count >>

(Optional)
This arguments is used to delete multiple volumes from the same NACL Volume
Deletion Arguments.  To delete the names for the volumes, the volume name
passed in is appended with the index that it was deleted at starting at 1.

=item C<< command_interface => $NACL_command_interface >>

(Optional)
This is the command_interface that will be used during NACL calls.  If a
command_interface is not passed in then one will be created.  Passing in a
command_interface helps performance because for every command_interface created
the version of the Filer must be determined by creating a connection to the filer
and then sending commands to get the version.

=item C<< method_timeout => $NACL_method_timeout >>

(Optional)
This is the method_timeout that will be used during NACL calls. If a
method_timeout is not passed in then the method_timeout will be defaulted to 3600.

=back

=item Return

This method returns a True or False based on whether the method succeeds or fails
in the deletion of the Volume(s) based on the Options passed in.

=item Exceptions

This method produces and throws no exceptions of its own but pass through any
uncaught exceptions from other library calls.

=cut

sub delete {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %args = @_;

    # Locals
    my $parent_answer;

    # TODO Remove $args{'Name'} from everywhere. And ensure
    # no errors occur.
    $args{'Name'} = "constituent_volume";

    # Super::_delete is in Task.pm
    $parent_answer = $pkg->SUPER::_delete(%args);
    $Log->comment("The delete subroutine was called, but it does nothing. Constituent-Volumes will be deleted upon parent InfiniVol deletion.");
    return;
} ## end sub delete

#==============================================================================
# Method:     _nacl_create_call
# Objective:  Makes the raw NACL calls for creation of a volume.
# Details:    Makes the NACL::STask::Volume->create() call and optionaly could
#             make a the NACL::STask::Volume->wait_for_creation. Also works
#             around some common non-fatal errors that can be return from
#             volume creation and then has a retry mechanism to retry volume
#             creation.
# TODO:       Need to update to NACL::STask::Volume->create()
# Inputs:     NACL::C::Volume->create arguments
#             retry => integer, is the current retry the method is on.
#                    This is used to count how many times the _nacl_create_call
#                    has been retried so that the retries can be limited.
# Outputs:    Returns whether the create was successful or a hash with success
#             and whether the volume was a duplicate if a hash is requested
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
#==============================================================================
sub _nacl_create_call {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %args = @_;

    # Locals
    my $was_created = 0;
    my $wait = delete $args{wait} // 0;
    my $if_exists = delete $args{if_exists} // 'die';

    # If the size argument has a percent then calculate the size.
    if ($args{size} =~ s/%//) {
        my $infinivol_cs_obj = NACL::CS::Volume->fetch(
                                        filter => {volume  => $args{volume},
                                                   vserver => $args{vserver}},
                                        command_interface => $args{command_interface},
                                        requested_fields => ['volume','vserver','size']);
        my $percent = $args{size};
        $args{size} = ($args{size}/100)*($infinivol_cs_obj->size());
        $Log->comment("Setting Volume size to " . $args{size} . "b which is " .
                      "$percent\% of parent Infinite Volume " . $infinivol_cs_obj->volume() .
                      " which is size " .     $infinivol_cs_obj->size() . "b");
    }

    # Delete 'volume' attrib and replace it with a user specified name for the constituent volume, if it exists
    delete $args{'volume'};
    if(exists $args{'name'}){
        $args{'volume'}= delete $args{'name'};
    }

    # This creates a NACL command_interface if one is not passed in.
    if (!exists($args{command_interface})) {
        my @Nodes = NACL::C::Node->find();
        $args{command_interface} = $Nodes[0];
    }

    # Execute Create
    my $volume = NACL::STask::Volume->create(%args,
                                    nacltask_if_exists => $if_exists,
                                    _was_created => \$was_created);
    if($was_created) {
        $Log->comment("Using newly Created Volume:" . $args{volume} . " in vserver:" . $args{vserver});
    } else {
        $Log->comment("Using already Created Volume:" . $args{volume} . " in vserver:" . $args{vserver});
        # need to verify that the volume is online, online it
        $volume->modify(state => 'online',
                        nacltask_if_modified => 'pass');
    }

    # Create the arguments for the volume to wait for creation if no
    # other volumes are going to be passed in.
    if ( ($was_created) && ($wait) ) {
        $pkg->_nacl_wait_call(%args);
    }

    $Log->exit() if $may_exit;
    # Returning either a list (hash) or scalar TRUE
    return (wantarray()) ? ('success'=>TRUE, 'duplicated'=>!$was_created) : TRUE;

} ## end sub _nacl_create_call

#==============================================================================
# Method:     _nacl_wait_call
# Objective:  Makes a NACL call to 'wait_for_creation.
# Details:    Makes the call to NACL::STask::Volume->wait_for_creation();
#             that will wait for the node to finish creating the aggr.
# Inputs:     NACL::STask::Volume->wait arguments
# Outputs:    Returns whether the create was successful TRUE or not FALSE
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
#==============================================================================
sub _nacl_wait_call {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %args = @_;
    # Assign values

    if (!exists($args{command_interface})) {
        my @Nodes = NACL::C::Node->find();
        $args{command_interface} = $Nodes[0];
    }

    # Locals
    my %wait_args = (
            'command_interface' => $args{'command_interface'},
            'volume'            => $args{'volume'},
            'vserver'           => $args{'vserver'},
            'method-timeout'    => $args{'method-timeout'}
        );

    #Special Cases
    $wait_args{apiset_should} = $args{apiset_should}
            if exists $args{apiset_should};

    if ($Log->may_debug()) {
        $Log->debug("Waiting args are being created: " . Dumper(\%wait_args));
    }
    #Call
    NACL::STask::Volume->wait_for_creation(%wait_args);

    $Log->exit() if $may_exit;
    return TRUE;
}

#==============================================================================
# Method:     _get_dtd_nacl_hashes (translate, ignore and delete)
# Objective:  Getter methods for the variables of the same name.
# Details:    Retrieves the hash or a reference to one of the following:
#             - DTD_To_NACL_Translate.
#             - DTD_To_NACL_Ignore.
#             - Delete_Args.
# Inputs:     none.
# Outputs:    The reference or hash variable.
# Exceptions:
#==============================================================================
sub _get_dtd_nacl_translate {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = @_;

    $Log->exit() if $may_exit;
    return (wantarray()) ? %DTD_To_NACL_Translate : \%DTD_To_NACL_Translate;
}
#==============================================================================
sub _get_dtd_nacl_ignore {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = @_;

    $Log->exit() if $may_exit;
    return (wantarray()) ? %DTD_To_NACL_Ignore : \%DTD_To_NACL_Ignore;
}
#==============================================================================
sub _get_dtd_nacl_delete {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = @_;

    $Log->exit() if $may_exit;
    return (wantarray()) ? %Delete_Args : \%Delete_Args;
}

1;
