# Copyright (c) 2012-2015 NetApp Inc.
# All rights reserved.
#
## @summary Provide a common set of libraries to work on controllers
## @author dl-san-automation
## @status private
## @pod here

=head1 NAME

NACL::AppBuilder::EnGenX::Controller::FCP

=head1 DESCRIPTION

NACL::AppBuilder::EnGenX::Controller::FCP is used to Config, Setup, Cleanup and
Deconfig FCP services on a vserver.

=cut

package NACL::AppBuilder::EnGenX::Controller::FCP;

# Compiler directives
use strict;
use warnings;

# NATE and Standard Module import
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use NATE::Util::DataTypes qw(TRUE FALSE);
use Data::Dumper;
use Params::Validate
    qw(validate_with validate SCALAR ARRAYREF BOOLEAN UNDEF SCALARREF HASHREF);

# NACL and EnGenX Libs
use NACL::STask::VserverFcp;

# Exceptions
use NATE::BaseException qw(:try);
use NATE::Exceptions::Argument qw(:try);

# Inherit base class
use base qw( NACL::AppBuilder::EnGenX::Controller::Task );

use vars qw(%DTD_To_NACL_Translate
    %DTD_To_NACL_Ignore
    %Delete_Args);

# Arguments after converting from xml to a
# resource list that need their keys translated to
# keys that NACL methods will accept.
%DTD_To_NACL_Translate = (
    _reuse   => 'if_exists',
    nodename => 'target-name'
);

# Arguments that are used by EnGenX but cannot be
# passed into the NACL methods
%DTD_To_NACL_Ignore = (
    name   => '',
    parallel      => '',
    service_enabled   => ''
);

# Arguments that can be passed into the NACL delete
# command
%Delete_Args = ();

=head1 METHODS

=head2 create

(Static method) Request to add the iscsi service to a list
of vservers the command interface and vserver name are the
only required perameters.

  NACL::AppBuilder::EnGenX::Controller::FCP->create(
            command_interface               => $node,
            vserver                         => $vserver_name,
            status-admin                    => $status_admin,
            force                           => $force,
            target-name                     => $target_name
            );

  NACL::AppBuilder::EnGenX::Controller::FCP->create(
            command_interface               => $node,
            vserver                         => $vserver_name,
            );

=over

=item Options

The NACL component only requires the user to specify the
command interface and the vserver. The ontap defaults will
be used for any option that is not stated.

=over

=item C<< target_name => $target_name >>

(Optional)
This is the target WWPN that must follow the standard WWPN
notation. If none is specified Ontap will use a defualt
WWPN.

=item C<< command_interface => $Node >>

(Required)
This is the command_interface that will be used during NACL calls.  If a
command_interface is not passed in then one will be created.  Passing in a
command_interface helps performance because for every command_interface created
the version of the Filer must be determined by creating a connection to the filer
and then sending commands to get the version.

=item C<< status-admin => $status_admin >>

(Optional)
The defualt is to have the service up and running. The user can specify
the service be down on creation.

=item C<< force => $force >>

(Optional)
This option will force the operation to complete. It is by default
set to not force the creation.

=item C<< vserver => $vserver >>

(Required)
This is the required vserver perameter.

=back

=item Return

This method returns a True or False based on whether the method succeeds or fails
in the creation of the Iscsi Targets based on the Options passed in.

=item Exceptions

This method produces and throws no exceptions of its own but pass through any
uncaught exceptions from other library calls.

=cut
sub create {

    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %args = @_;

    # Locals
    my $parent_answer;

    # Globals that will also be set as globals in the base class
    $args{'Name'} = "fcp";
    # Super::_create is in Task.pm
    $parent_answer = $pkg->SUPER::_create(%args);

    return $parent_answer;
} ## end sub create

#==============================================================================
#The delete function is not going to be implemented
#due to the options being deleted upon vserver deletion.
#==============================================================================
sub delete {
    #TODO: no-op for now. May implement in the future
    return;
}

#==============================================================================
# Method:     _nacl_create_call
# Objective:  Makes the raw NACL calls for creation of the fcp service.
# Details:    Makes the NACL::STask::VserverFCP->create(). Also works
#             around some common non-fatal errors that can be return from
#             volume creation and then has a retry mechanism to retry volume
#             creation.
# Inputs:     NACL::STask::VserverFCP->create arguments
#             retry => integer, is the current retry the method is on.
#                    This is used to count how many times the _nacl_create_call
#                    has been retried so that the retries can be limited.
# Outputs:    Returns whether the create was successful or a hash with success
#             and whether the volume was a duplicate if a hash is requested
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
#==============================================================================
sub _nacl_create_call {

    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %args = @_;

    # Locals
    my $was_created = 0;
    my $if_exists = exists($args{if_exists}) ? delete($args{if_exists}) : 'die';
    $args{'method-timeout'} = 3600 unless exists($args{'method-timeout'});

    # Default status-admin to up the ontap default if not defined.
    $args{'status-admin'}   = 'up' unless exists($args{'status-admin'});

    # Clusters.dtd has 1 or 0 for true or false and Ontap has
    # true or false.
    if (exists($args{force})) {
        if ($args{force} == 1) {
            $args{force} = "true";
        } else { #if ($args{force} == 0)
            $args{force} = "false";
        }
    }

    # This creates a NACL command_interface if one is not passed in.
    if (!exists($args{command_interface})) {
        my @Nodes = NACL::STask::Node->find();
        $args{command_interface} = $Nodes[0];
    }

    NACL::STask::VserverFcp->create(%args,
                                    nacltask_if_exists => $if_exists,
                                    _was_created       => \$was_created);

    if($was_created) {
        $Log->comment("Using newly Created FCP Service on Vserver:" . $args{vserver});
    } else {
        $Log->comment("Using already Created FCP Service on Vserver:" . $args{vserver});
        # Stop the fcp service if target-name passed in so it can be modified.
        # Also target-name and status-admin cannot be modified at the same time.
        if (exists $args{'target-name'}) {
		    $Log->comment("FCP Service on Vserver:" . $args{vserver} . " going down.");
            NACL::STask::VserverFcp->modify(
                                    command_interface => $args{command_interface},
                                    vserver           => $args{vserver},
                                    'status-admin'    => 'down',
                                    'method-timeout'  => $args{'method-timeout'},
                                    'apiset_should'   => $args{apiset_should},
                                    );

		    $Log->comment("FCP Service on Vserver:" . $args{vserver} . " modifying target-name.");
            NACL::STask::VserverFcp->modify(
                                    command_interface => $args{command_interface},
                                    vserver           => $args{vserver},
                                    'target-name'     => $args{'target-name'},
                                    'method-timeout'  => $args{'method-timeout'},
                                    'apiset_should'   => $args{apiset_should},
                                    );
        }

        # If target-name is not defines then this needs to be ran either way
        # If target-name is defined and status-admin up then this needs to be ran
        if (not exists $args{'target-name'} or $args{'status-admin'} =~ /up/i) {
		    $Log->comment("FCP Service on Vserver:" . $args{vserver} . " going " . $args{'status-admin'});
            NACL::STask::VserverFcp->modify(
                                    command_interface => $args{command_interface},
                                    vserver           => $args{vserver},
                                    'status-admin'      => $args{'status-admin'},
                                    'method-timeout'  => $args{'method-timeout'},
                                    'apiset_should'   => $args{apiset_should},
                                    );
        }

    }

    $Log->exit() if $may_exit;
    return (wantarray()) ? ('success' => TRUE) : TRUE;
}#sub _nacl_create_call

#==============================================================================
#The _nacl_delete_call function is not going to be implemented
#due to the options being deleted upon vserver deletion.
#==============================================================================
sub _nacl_delete_call {
    #TODO: no-op for now. May implement in the future
    return;
}#sub _nacl_delete_call

#==============================================================================
# Method:     _nacl_wait_call
# Objective:  Makes a NACL call to 'wait_for_creation.
# Details:    Makes the call to NACL::STask::Vserver->wait_for_creation();
#             that will wait for the node to finish creating the aggr.
# Inputs:     NACL::STask::Vserver->wait arguments
# Outputs:    Returns whether the create was successful TRUE or not FALSE
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
# TODO: Implement this method when needed
#==============================================================================
sub _nacl_wait_call {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = @_;

    $Log->exit() if $may_exit;
    return TRUE;
}

#==============================================================================
# Method:     _get_dtd_nacl_hashes (translate, ignore and delete)
# Objective:  Getter methods for the variables of the same name.
# Details:    Retrieves the hash or a reference to one of the following:
#             - DTD_To_NACL_Translate.
#             - DTD_To_NACL_Ignore.
#             - Delete_Args.
# Inputs:     none.
# Outputs:    The reference or hash variable.
# Exceptions:
#==============================================================================
sub _get_dtd_nacl_translate {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = @_;

    $Log->exit() if $may_exit;
    return (wantarray()) ? %DTD_To_NACL_Translate : \%DTD_To_NACL_Translate;
}
#==============================================================================
sub _get_dtd_nacl_ignore {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = @_;

    $Log->exit() if $may_exit;
    return (wantarray()) ? %DTD_To_NACL_Ignore : \%DTD_To_NACL_Ignore;
}
#==============================================================================
sub _get_dtd_nacl_delete {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = @_;

    $Log->exit() if $may_exit;
    return (wantarray()) ? %Delete_Args : \%Delete_Args;
}

1;
