# $Id$

#
# Copyright (c) 2001-2012 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary File Controller Layer Module
## @author dl-san-automation-core
## @status shared
## @pod here

=head1 NAME

NACL::AppBuilder::EnGenX::Controller::File

=head1 DESCRIPTION

NACL::AppBuilder::EnGenX::Controller::File is a derived class of
NACL::AppBuilder::EnGenX::Controller::Task.  It is used to Create and
Delete ONTAP Files. All methods are static.

=cut

package NACL::AppBuilder::EnGenX::Controller::File;

# Compiler directives
use strict;
use warnings;

# NATE and Standard Module import
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use NATE::Util::DataTypes qw(TRUE FALSE);
use Params::Validate
    qw(validate_with validate SCALAR ARRAYREF BOOLEAN UNDEF SCALARREF HASHREF);

# NACL and EnGenX Libs
use NACL::STask::VolumeFile;
use NACL::AppBuilder::EnGenX::SubProcess;
use NACL::AppBuilder::EnGenX::EnGenXUtil;

# Exceptions
use NATE::BaseException qw(:try);
use NATE::Exceptions::Argument qw(:try);

#Inherit base class
use base qw( NACL::AppBuilder::EnGenX::Controller::Task );

use vars qw(%DTD_To_NACL_Translate
    %DTD_To_NACL_Ignore
    %Delete_Args);

# Arguments after converting from xml to a
# resource list that need their keys translated to
# keys that NACL methods will accept.
%DTD_To_NACL_Translate = (
    _reuse => 'if_exists'
);

# Arguments that are used by EnGenX but cannot be
# passed into the NACL methods
%DTD_To_NACL_Ignore = (parallel => '');

# Arguments that can be passed into the NACL delete
# command
%Delete_Args = (
    command_interface => '',
    apiset_must       => '',
    apiset_should     => '',
    vserver           => '',
    volume            => '',
    qtree             => '',
    lun               => '',
    file              => '',
);

=head1 METHODS

=head2 create

(Static method) Request creation of one or multiple files serially
or in parallel and return whether the file(s) was created successfully.
This method passes all user arguments through to NACL::AppBuilder::EnGenX::Controller::Task
and adds specific arguments used by the EnGenX framework.

  NACL::AppBuilder::EnGenX::Controller::File->create(
            command_interface => $Node,
            file                  => $File_name,
            vserver           => $Vserver_name,
            size              => "40M",
            'method-timeout'  => 3600,
            );

  NACL::AppBuilder::EnGenX::Controller::File->create(
            command_interface => $Node,
            file                  => $File_name,
            vserver           => $Vserver_name,
            size              => "40M",
            'method-timeout'  => 3600,
            count             => 5
            );

  NACL::AppBuilder::EnGenX::Controller::File->create(
            command_interfac  => $Node,
            resource_list      => @Files,
            'method-timeout'  => 3600,
            );

  NACL::AppBuilder::EnGenX::Controller::File->create(
            command_interface => $Node,
            resource_list      => @Files,
            );

  NACL::AppBuilder::EnGenX::Controller::File->create(
            resource_list      => @Files,
            'method-timeout'  => 3600,
            );

  NACL::AppBuilder::EnGenX::Controller::File->create(
            resource_list      => @Files,
            );

=over

=item Options

Either the resource_list is required or the NACL File Creation Arguments
are required but both should NOT be used in the same call.

=over

=item C<< resource_list => @resource_list >>

(Optional)
If this is passed in the files in the resource list will be created
based on the information that is supplied in the resource list. If
the resource list is passed in the NACL File Creation Arguments should
NOT be passed in.

=item C<< NACL File Creation Arguments >>

(Optional)
These are arguments that can be passed into NACL::C::File->Create.
If these arguments are passed in the resource_list argument should NOT
be passed in

=item C<< count => $file_count >>

(Optional)
This arguments is used to create multiple files from the same NACL File
Creation Arguments.  To create the names for the files, the file name
passed in is appended with the index that it was create at starting at 1.

=item C<< command_interface => $NACL_command_interface >>

(Optional)
This is the command_interface that will be used during NACL calls.  If a
command_interface is not passed in then one will be created.  Passing in a
command_interface helps performance because for every command_interface created
the version of the Filer must be determined by creating a connection to the filer
and then sending commands to get the version.

=item C<< method_timeout => $NACL_method_timeout >>

(Optional)
This is the method_timeout that will be used during NACL calls. If a
method_timeout is not passed in then the method_timeout will be defaulted.

=back

=item Return

This method returns a True or False based on whether the method succeeds or fails
in the creation of the File(s) based on the Options passed in.

=item Exceptions

This method produces and throws no exceptions of its own but pass through any
uncaught exceptions from other library calls.

=cut

sub create {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %args = @_;

    # Locals
    my $parent_answer;

    # Globals that will also be set as globals in the base class
    $args{'Name'} = "file";

    # Super::_create is in Task.pm
    $parent_answer = $pkg->SUPER::_create(%args);

    return $parent_answer;
} ## end sub create

=head2 delete

(Static method) Request deletion of one or multiple files serially
or in parallel and return whether the file(s) was deleted successfully.
This method passes all user arguments through to NACL::AppBuilder::EnGenX::Controller::Task
and adds specific arguments used by the EnGenX framework.

  NACL::AppBuilder::EnGenX::Controller::File->delete(
            command_interface => $Node,
            file                  => $File_name,
            'method-timeout'  => 3600,
            );

  NACL::AppBuilder::EnGenX::Controller::File->delete(
            command_interface => $Node,
            file                  => $File_name,
            count             => 5
            );

  NACL::AppBuilder::EnGenX::Controller::File->delete(
            command_interface => $Node,
            resource_list      => @Files,
            'method-timeout'  => 3600,
            );

  NACL::AppBuilder::EnGenX::Controller::File->delete(
            command_interface => $Node,
            resource_list      => @Files,
            );

  NACL::AppBuilder::EnGenX::Controller::File->delete(
            resource_list      => @Files,
            'method-timeout'  => 3600,
            );

  NACL::AppBuilder::EnGenX::Controller::File->delete(
            resource_list      => @Files,
            );

=over

=item Options

Either the resource_list is required or the NACL File Deletion Arguments
are required but both should NOT be used in the same call.

=over

=item C<< resource_list => @resource_list >>

(Required)
If this is passed in the files in the resource list will be deleted
based on the information that is supplied in the resource list. If
the resource list is passed in the NACL File Deletion Arguments should
NOT be passed in.

=item C<< NACL File Deletion Arguments >>

(Required)
These are arguments that can be passed into NACL::C::File->Delete.
If these arguments are passed in the resource_list argument should NOT
be passed in.

=item C<< count => $file_count >>

(Optional)
This arguments is used to delete multiple files from the same NACL File
Deletion Arguments.  To delete the names for the files, the file name
passed in is appended with the index that it was deleted at starting at 1.

=item C<< command_interface => $NACL_command_interface >>

(Optional)
This is the command_interface that will be used during NACL calls.  If a
command_interface is not passed in then one will be created.  Passing in a
command_interface helps performance because for every command_interface created
the version of the Filer must be determined by creating a connection to the filer
and then sending commands to get the version.

=item C<< method_timeout => $NACL_method_timeout >>

(Optional)
This is the method_timeout that will be used during NACL calls. If a
method_timeout is not passed in then the method_timeout will be defaulted.

=back

=item Return

This method returns a True or False based on whether the method succeeds or fails
in the deletion of the File(s) based on the Options passed in.

=item Exceptions

This method produces and throws no exceptions of its own but pass through any
uncaught exceptions from other library calls.

=cut

sub delete {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %args = @_;

    # Locals
    my $parent_answer;

    # Globals that will also be set as globals in the base class
    $args{'Name'} = "file";

    # Super::_delete is in Task.pm
    $parent_answer = $pkg->SUPER::_delete(%args);

    return $parent_answer;
} ## end sub delete

#==============================================================================
# Method:     _nacl_create_call
# Objective:  Makes the raw NACL calls for creation of a file.
# Details:    Makes the NACL::C::File->create() call. Also works
#             around some common non-fatal errors that can be return from
#             file creation and then has a retry mechanism to retry file
#             creation.
# Inputs:     NACL::C::File->create arguments
#             retry => integer, is the current retry the method is on.
#                    This is used to count how many times the _nacl_create_call
#                    has been retried so that the retries can be limited.
# Outputs:    Returns whether the create was successful or a hash with success
#             and whether the file was a duplicate if a hash is requested
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
#==============================================================================
sub _nacl_create_call {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %args = @_;
    # Locals
    my $duplicated = FALSE;
    my $if_exists
        = exists($args{if_exists}) ? delete($args{if_exists}) : 'die';
    $args{'method-timeout'} = 3600 unless exists($args{'method-timeout'});
    $args{'determine_appropriate_nodescope_node'} = 1 unless exists($args{'determine_appropriate_nodescope_node'});

    if (!exists($args{command_interface})) {
        my @Nodes = NACL::C::Node->find();
        $args{command_interface} = $Nodes[0];
    }

    my $name;
    if(exists ($args{name})) {
        $name = delete $args{name};
    } elsif (exists ($args{lun})) {
        $name = "file_$args{lun}";
    } else {
        $name = "file_$args{volume}";
    }

    unless(defined($args{path})) {
        if(defined($args{qtree}) and length($args{qtree})){
            $args{'path'} = "/vol/$args{volume}/$args{qtree}/$name";
        } else {
            $args{'path'} = "/vol/$args{volume}/$name";
        }
    }

    # If -f is being passed then need to do an ls first to see if file exists because of burt786371
    if($args{'fragment-file'}) {
        my $found_file = NACL::STask::VolumeFile->find(
            command_interface => $args{command_interface},
            filter => {
                vserver => $args{vserver},
                path       => $args{path},
            },
            allow_empty => 1,
        );

        if (defined($found_file)) {
            $Log->comment("Using already Created File:" . $args{path});
            return (wantarray()) ? ('success'=>TRUE, 'duplicated'=>TRUE) : TRUE;
        }
    }

    # Delete volume and qtree because they are not part of the file
    # args and are contained in the file path
    delete $args{volume};
    delete $args{qtree};
    delete $args{lun};

    # Execute Create
    try {
        NACL::STask::VolumeFile->create(%args);
        $Log->comment("Comment: Using newly Created File:" . $args{path});
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = shift;
        if (   $if_exists eq 'reuse'
            && $exception->text() =~ /write failed/)
        {
            $duplicated = TRUE;
            $Log->comment("Using already Created File:" . $args{path});
        } else {
            $Log->exit() if $may_exit;

            # Rethrowing caught exception
            $exception->throw();
        }
    };

    $Log->exit() if $may_exit;
    # Returning either a list (hash) or scalar TRUE
    return (wantarray()) ? ('success'=>TRUE, 'duplicated'=>$duplicated) : TRUE;


} ## end sub _nacl_create_call

#========================================================================================
# Method:     _nacl_delete_call
# Objective:  Makes the raw NACL calls for deletion of a File.
# Details:    Makes the NACL::C::File->delete() calls. Also works around
#             some common non-fatal errors that can be return from File
#             deletion and then has a retry mechanism to retry File deletion.
# Inputs:     NACL::C::File->delete arguments
#             retry => integer, is the current retry the method is on.
#                    This is used to count how many times the _nacl_delete_call
#                    has been retried so that the retries can be limited.
# Outputs:    Returns whether the create was successful or a hash with success
#             if a hash is requested
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
#==============================================================================
sub _nacl_delete_call {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %args = @_;
    # Locals
    my $if_exists;
    my $retry = exists $args{retry} ? delete $args{retry} : 0;
    $args{'method-timeout'} = 3600 unless exists($args{'method-timeout'});
    if (!exists($args{command_interface})) {
        my @Nodes = NACL::C::Node->find();
        $args{command_interface} = $Nodes[0];
    }

    my $name;
    if(exists ($args{name})) {
        $name = delete $args{name};
    } elsif (exists ($args{lun})) {
        $name = "file_$args{lun}";
    } else {
        $name = "file_$args{volume}";
    }

    if(defined($args{qtree}) and length($args{qtree})){
        $args{'path'} = "/vol/$args{volume}/$args{qtree}/$name";
    } else {
        $args{'path'} = "/vol/$args{volume}/$name";
    }

    # Delete volume and qtree because they are not part of the file
    # args and are contained in the file path
    delete $args{volume};
    delete $args{qtree};
    delete $args{lun};

    try {
        NACL::STask::VolumeFile->delete_file(%args);
        $Log->comment("File:" . $args{path} . " was deleted");
    }
    catch NACL::Exceptions::NoElementsFound with {
        # This means the file does exists so we don't care.
        $Log->comment("File:" . $args{path} . " was already deleted");
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = shift;

        if ($exception->text() =~ /No such file/)
        {
            $Log->comment("File:" . $args{path} . " was already deleted");
        } else {
            # Rethrowing caught exception
            $Log->exit() if $may_exit;
            $exception->throw();
        }
    };

    $Log->exit() if $may_exit;
    return TRUE;

} ## end sub _nacl_delete_call
#==============================================================================
# Method:     _nacl_wait_call
# Objective:  Makes a NACL call to 'wait_for_creation.
# Details:    Makes the call to NACL::STask::File->wait_for_creation();
#             that will wait for the node to finish creating the aggr.
# Inputs:     NACL::STask::File->wait arguments
# Outputs:    Returns whether the create was successful TRUE or not FALSE
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
# TODO: Implement this method when needed
#==============================================================================
sub _nacl_wait_call {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = @_;

    $Log->exit() if $may_exit;
    return TRUE;
}

#==============================================================================
# Method:     _get_dtd_nacl_hashes (translate, ignore and delete)
# Objective:  Getter methods for the variables of the same name.
# Details:    Retrieves the hash or a reference to one of the following:
#             - DTD_To_NACL_Translate.
#             - DTD_To_NACL_Ignore.
#             - Delete_Args.
# Inputs:     none.
# Outputs:    The reference or hash variable.
# Exceptions:
#==============================================================================
sub _get_dtd_nacl_translate {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = @_;

    $Log->exit() if $may_exit;
    return (wantarray()) ? %DTD_To_NACL_Translate : \%DTD_To_NACL_Translate;
}
#==============================================================================
sub _get_dtd_nacl_ignore {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = @_;

    $Log->exit() if $may_exit;
    return (wantarray()) ? %DTD_To_NACL_Ignore : \%DTD_To_NACL_Ignore;
}
#==============================================================================
sub _get_dtd_nacl_delete {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = @_;

    $Log->exit() if $may_exit;
    return (wantarray()) ? %Delete_Args : \%Delete_Args;
}

1;
