#
# Copyright (c) 2001-2013 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary BroadcastDomain Controller Layer Module
## @author dl-san-automation-core
## @status shared
## @pod here

=head1 NAME

NACL::AppBuilder::EnGenX::Controller::IPspace::BroadcastDomain

=head1 DESCRIPTION

NACL::AppBuilder::EnGenX::Controller::IPspace::BroadcastDomain is a derived class of
NACL::AppBuilder::EnGenX::Controller::Task.  It is used to Create and
Delete ONTAP IPspace BroadcastDomains. All methods are static.

=cut

package NACL::AppBuilder::EnGenX::Controller::IPspace::BroadcastDomain;

# Compiler directives
use strict;
use warnings;

# NATE and Standard Module import
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use NATE::Util::DataTypes qw(TRUE FALSE);
use Params::Validate
    qw(validate_with validate SCALAR ARRAYREF BOOLEAN UNDEF SCALARREF HASHREF);

# NACL and EnGenX Libs
use NACL::STask::NetworkPortBroadcastDomain;

# Exceptions
use NATE::BaseException qw(:try);
use NATE::Exceptions::Argument qw(:try);

# Inherit base class
use base qw( NACL::AppBuilder::EnGenX::Controller::Task );

use vars qw(%DTD_To_NACL_Translate
    %DTD_To_NACL_Ignore
    %Delete_Args
);

# Arguments after converting from xml to a
# resource list that need their keys translated to
# keys that NACL methods will accept.
%DTD_To_NACL_Translate = (
    _reuse => 'if_exists',
    name   => 'broadcast-domain'
);

# Arguments that are used by EnGenX but cannot be
# passed into the NACL create method
%DTD_To_NACL_Ignore = (
    parallel      => '',
    delete_policy => ''
);

# Arguments that can be passed into the NACL delete
# command
%Delete_Args = (
    command_interface   => '',
    apiset_must         => '',
    apiset_should       => '',
    ipspace             => '',
    'broadcast-domain'  => '',
    delete_policy       => ''
);

=head1 METHODS

=head2 create

(Static method) Request creation of one ipspace BroadcastDomain serially
or in parallel and return whether the ipspace was created successfully.
This method passes all user arguments through to NACL::AppBuilder::EnGenX::Controller::Task
and adds specific arguments used by the EnGenX framework.

  NACL::AppBuilder::EnGenX::Controller::IPspace::BroadcastDomain->create(
            command_interface => $Node,
            node              => $Node_name,
            port              => $Port,
            ipspace           => $Ipspace,
            );

=over

=item Options

Either the resource_list is required or the NACL IPspace BroadcastDomain
Creation Arguments are required but both should NOT be used in the
same call.

=over

=item C<< resource_list => @resource_list >>

(Optional)
If this is passed in the ipspace BroadcastDomain in the resource list will be
created based on the information that is supplied in the resource
list. If the resource list is passed in the NACL IPspace BroadcastDomain
Creation Arguments should NOT be passed in.

=item C<< NACL IPspace BroadcastDomain Creation Arguments >>

(Optional)
These are arguments that can be passed into NACL::STask::NetworkPortBroadcastDomain->create.
If these arguments are passed in the resource_list argument should NOT
be passed in

=item C<< command_interface => $NACL_command_interface >>

(Optional)
This is the command_interface that will be used during NACL calls.  If a
command_interface is not passed in then one will be created.  Passing in a
command_interface helps performance because for every command_interface created
the version of the Filer must be determined by creating a connection to the filer
and then sending commands to get the version.

=item C<< method_timeout => $NACL_method_timeout >>

(Optional)
This is the method_timeout that will be used during NACL calls. If a
method_timeout is not passed in then the method_timeout will be defaulted.

=back

=item Return

This method returns a True or False based on whether the method succeeds or fails
in the creation of the IPspace BroadcastDomain based on the Options passed in.

=item Exceptions

This method produces and throws no exceptions of its own but pass through any
uncaught exceptions from other library calls.

=cut

sub create {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %args = @_;

    # Locals
    my $parent_answer;

    # Globals that will also be set as globals in the base class
    $args{'Name'} = "ipspace_broadcastdomain";

    # Super::_create is in Task.pm
    $parent_answer = $pkg->SUPER::_create(%args);

    return $parent_answer;
} ## end sub create

=head2 delete

(Static method) Request deletion of one ipspace BroadcastDomain serially or in
parallel and return whether the ipspace BroadcastDomain was deleted successfully.
This method passes all user arguments through to NACL::AppBuilder::EnGenX::Controller::Task
and adds specific arguments used by the EnGenX framework.

  NACL::AppBuilder::EnGenX::Controller::IPspace::BroadcastDomain->delete(
            command_interface => $Node,
            node              => $Node_name,
            port              => $Port,
            ipspace           => $Ipspace,
            );

=over

=item Options

Either the resource_list is required or the NACL IPspace BroadcastDomain
Deletion Arguments are required but both should NOT be used in the
same call.

=over

=item C<< resource_list => @resource_list >>

(Required)
If this is passed in the ipspace BroadcastDomain in the resource list will be
deleted based on the information that is supplied in the resource
list. If the resource list is passed in the NACL IPspace Deletion
Arguments should NOT be passed in.

=item C<< NACL IPspace BroadcastDomain Deletion Arguments >>

(Required)
These are arguments that can be passed into NACL::STask::NetworkPortBroadcastDomain->purge.
If these arguments are passed in the resource_list argument should NOT
be passed in.

=item C<< command_interface => $NACL_command_interface >>

(Optional)
This is the command_interface that will be used during NACL calls.  If a
command_interface is not passed in then one will be created.  Passing in a
command_interface helps performance because for every command_interface created
the version of the Filer must be determined by creating a connection to the filer
and then sending commands to get the version.

=item C<< method_timeout => $NACL_method_timeout >>

(Optional)
This is the method_timeout that will be used during NACL calls. If a
method_timeout is not passed in then the method_timeout will be defaulted.

=back

=item Return

This method returns a True or False based on whether the method succeeds or fails
in the deletion of the IPspace BroadcastDomain based on the Options passed in.

=item Exceptions

This method produces and throws no exceptions of its own but pass through any
uncaught exceptions from other library calls.

=cut

sub delete {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %args = @_;

    # Locals
    my $parent_answer;

    # Globals that will also be set as globals in the base class
    $args{'Name'} = "ipspace_BroadcastDomain";

    # Super::_delete is in Task.pm
    $parent_answer = $pkg->SUPER::_delete(%args);

    return $parent_answer;
} ## end sub delete


#==============================================================================
# Method:     _nacl_create_call
# Objective:  Makes the raw NACL calls for creation of a ipspace BroadcastDomain.
# Details:    Makes the NACL::STask::NetworkPortBroadcastDomain->create() call. The STask used works
#             around some common non-fatal errors that can be return from
#             ipspace BroadcastDomain creation and then has a retry mechanism to retry
#             ipspace creation.
# Inputs:     NACL::STask::NetworkPortBroadcastDomain->create arguments
# Outputs:    Returns whether the create was successful or a hash with success
#             and whether the ipspace BroadcastDomain was a duplicate if a hash is
#             requested
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
#==============================================================================
sub _nacl_create_call {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %args = @_;

    # Locals
    my $was_created = 0;
    my $if_exists = exists($args{if_exists})
                  ? delete($args{if_exists})
                  : 'die';
    if (!exists($args{command_interface})) {
        my @Nodes = NACL::C::Node->find();
        $args{command_interface} = $Nodes[0];
    }

    NACL::STask::NetworkPortBroadcastDomain->create(%args,
                                                    nacltask_if_exists => $if_exists,
                                                    _was_created => \$was_created);
    if($was_created) {
        $Log->comment("Using newly Created Broadcast Domain: " . $args{'broadcast-domain'} .
                      "in IPspace:" . $args{ipspace});
    } else {
        $Log->comment("Using already Created Broadcast Domain: " . $args{'broadcast-domain'} .
                      "in IPspace:" . $args{ipspace});
    }

    $Log->exit() if $may_exit;
    # Returning either a list (hash) or scalar TRUE
    return (wantarray()) ? ('success'=>TRUE, 'duplicated'=>!$was_created) : TRUE;

} ## end sub _nacl_create_call

#==============================================================================
# Method:     _nacl_delete_call
# Objective:  Makes the raw NACL calls for deletion of a ipspace BroadcastDomain.
# Details:    Makes the NACL::STask::NetworkPortBroadcastDomain->purge() calls. The STask used works around
#             some common non-fatal errors that can be return from ipspace
#             BroadcastDomain deletion and then has a retry mechanism to retry ipspace
#             BroadcastDomain deletion.
# Inputs:     NACL::STask::NetworkPortBroadcastDomain->purge arguments
# Outputs:    Returns whether the create was successful or a hash with success
#             if a hash is requested
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
#==============================================================================
sub _nacl_delete_call {
    $Log->enter() if $may_enter;
    my $pkg   = shift;
    my %args = @_;

    # Locals
    my $was_deleted = 0;
    if (!exists($args{command_interface})) {
        my @Nodes = NACL::C::Node->find();
        $args{command_interface} = $Nodes[0];
    }

    my $delete_policy = delete $args{delete_policy};

    try {
        NACL::STask::NetworkPortBroadcastDomain->purge(%args,
                                                        nacltask_if_purged => 'pass',
                                                        _was_deleted       => \$was_deleted);
    } otherwise {
        my $exception = shift;
        if ($delete_policy eq "warn") {
            $Log->warn("Attempted to Delete broadcast domain " .
                        $args{'broadcast-domain'} .
                       " for ipspace " .
                        $args{ipspace} .
                       "Warning and Continuing Test\n" .
                       $exception->text());
        } elsif ($delete_policy eq "fail") {
            $Log->exit() if $may_exit;
            $exception->throw();
        }
    };

    if($was_deleted) {
        $Log->comment("Broadcast domain " .
                       $args{'broadcast-domain'} .
                      " for ipspace " .
                       $args{ipspace} .
                      ". was deleted.\n");
    } else {
        $Log->comment("Cannot cleanup broadcast domain " .
                       $args{'broadcast-domain'} .
                      " for ipspace " .
                       $args{ipspace} .
                      ". It does not exist in the cluster.\n");
    }
    $Log->exit() if $may_exit;
    return TRUE;

} ## end sub _nacl_delete_call

#==============================================================================
# Method:     _nacl_wait_call
# Objective:  Makes a NACL call to 'wait_for_creation.
# Details:    Makes the call to NACL::STask::NetworkPortBrodcastDomain->wait_for_creation();
#             that will wait for the node to finish creating the aggr.
# Inputs:     NACL::STask::NetworkPortBroadcastDomain->wait arguments
# Outputs:    Returns whether the create was successful TRUE or not FALSE
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
# TODO: Implement this method when needed
#==============================================================================
sub _nacl_wait_call {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = @_;

    $Log->exit() if $may_exit;
    return TRUE;
}

#==============================================================================
# Method:     _get_dtd_nacl_hashes (translate, ignore and delete)
# Objective:  Getter methods for the variables of the same name.
# Details:    Retrieves the hash or a reference to one of the following:
#             - DTD_To_NACL_Translate.
#             - DTD_To_NACL_Ignore.
#             - Delete_Args.
# Inputs:     none.
# Outputs:    The reference or hash variable.
# Exceptions:
#==============================================================================
sub _get_dtd_nacl_translate {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = @_;

    $Log->exit() if $may_exit;
    return (wantarray()) ? %DTD_To_NACL_Translate : \%DTD_To_NACL_Translate;
}
#==============================================================================
sub _get_dtd_nacl_ignore {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = @_;

    $Log->exit() if $may_exit;
    return (wantarray()) ? %DTD_To_NACL_Ignore : \%DTD_To_NACL_Ignore;
}
#==============================================================================
sub _get_dtd_nacl_delete {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = @_;

    $Log->exit() if $may_exit;
    return (wantarray()) ? %Delete_Args : \%Delete_Args;
}

1;
