# $Id: //depot/prod/DOT/dev/test/lib/EnGenX/Controller/IPspace/BD_Port.pm#1 $

#
# Copyright (c) 2001-2013 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary IPspace BD_Port Controller Layer Module
## @author dl-san-automation-core
## @status shared
## @pod here

=head1 NAME

EnGenX::Controller::IPspace::Port

=head1 DESCRIPTION

NACL::AppBuilder::EnGenX::Controller::IPspace::BroadcastDomain::Port is a derived class of
NACL::AppBuilder::EnGenX::Controller::Task.  It is used to Create and
Delete ONTAP IPspace BD_Ports. All methods are static.

=cut

package NACL::AppBuilder::EnGenX::Controller::IPspace::BroadcastDomain::Port;

# Compiler directives
use strict;
use warnings;

# NATE and Standard Module import
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use NATE::Util::DataTypes qw(TRUE FALSE);
use Params::Validate
    qw(validate_with validate SCALAR ARRAYREF BOOLEAN UNDEF SCALARREF HASHREF);

# NACL and EnGenX Libs
use NACL::AppBuilder::EnGenX::SubProcess;
use NACL::AppBuilder::EnGenX::EnGenXUtil;
use NACL::STask::NetworkPortBroadcastDomain;

# Exceptions
use NATE::BaseException qw(:try);
use NATE::Exceptions::Argument qw(:try);

# Inherit base class
use base qw( NACL::AppBuilder::EnGenX::Controller::Task );

use vars qw(
    %DTD_To_NACL_Translate
    %DTD_To_NACL_Ignore
    %Delete_Args
);

# Arguments after converting from xml to a
# resource list that need their keys translated to
# keys that NACL methods will accept.
%DTD_To_NACL_Translate = (
    _reuse             => 'if_exists',
    'broadcast_domain' => 'broadcast-domain'
);

# Arguments that are used by EnGenX but cannot be
# passed into the NACL create method
%DTD_To_NACL_Ignore = (
    parallel      => '',
    delete_policy => ''
);

# Arguments that can be passed into the NACL delete
# command
%Delete_Args = (
    command_interface => '',
    apiset_must       => '',
    apiset_should     => '',
    vserver           => '',
    node              => '',
    port              => '',
    mtu               => '',
    delete_policy     => ''
);

=head1 METHODS

=head2 create

(Static method) Request creation of one ipspace BD_Port serially
or in parallel and return whether the ipspace was created successfully.
This method passes all user arguments through to NACL::AppBuilder::EnGenX::Controller::Task
and adds specific arguments used by the EnGenX framework.

  NACL::AppBuilder::EnGenX::Controller::IPspace::BroadcastDomain::Port->assign_port(
            command_interface => $Node,
            node              => $Node_name,
            port              => $Port,
            ipspace           => $Ipspace,
            );

=over

=item Options

Either the resource_list is required or the NACL IPspace BD_Port
Creation Arguments are required but both should NOT be used in the
same call.

=over

=item C<< resource_list => @resource_list >>

(Optional)
If this is passed in the ipspace BD_Port in the resource list will be
created based on the information that is supplied in the resource
list. If the resource list is passed in the NACL IPspace BD_Port
Creation Arguments should NOT be passed in.

=item C<< NACL IPspace BD_Port Creation Arguments >>

(Optional)
These are arguments that can be passed into NACL::C::NetworkPortBroadcastDomain->assign_port.
If these arguments are passed in the resource_list argument should NOT
be passed in

=item C<< command_interface => $NACL_command_interface >>

(Optional)
This is the command_interface that will be used during NACL calls.  If a
command_interface is not passed in then one will be created.  Passing in a
command_interface helps performance because for every command_interface created
the version of the Filer must be determined by creating a connection to the filer
and then sending commands to get the version.

=item C<< method_timeout => $NACL_method_timeout >>

(Optional)
This is the method_timeout that will be used during NACL calls. If a
method_timeout is not passed in then the method_timeout will be defaulted.

=back

=item Return

This method returns a True or False based on whether the method succeeds or fails
in the creation of the IPspace BD_Port based on the Options passed in.

=item Exceptions

This method produces and throws no exceptions of its own but pass through any
uncaught exceptions from other library calls.

=cut

sub create {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %args = @_;

    # Locals
    my $parent_answer;

    # Globals that will also be set as globals in the base class
    $args{'Name'} = "ipspace_BD_Port";

    # Super::_create is in Task.pm
    $parent_answer = $pkg->SUPER::_create(%args);

    return $parent_answer;
} ## end sub create

=head2 delete

(Static method) Request deletion of one ipspace BD_Port serially or in
parallel and return whether the ipspace BD_Port was deleted successfully.
This method passes all user arguments through to NACL::AppBuilder::EnGenX::Controller::Task
and adds specific arguments used by the EnGenX framework.

  NACL::AppBuilder::EnGenX::Controller::IPspace::BroadcastDomain::Port->assign_port(
            command_interface => $Node,
            node              => $Node_name,
            port              => $Port,
            ipspace           => $Ipspace,
            );

=over

=item Options

Either the resource_list is required or the NACL IPspace BD_Port
Deletion Arguments are required but both should NOT be used in the
same call.

=over

=item C<< resource_list => @resource_list >>

(Required)
If this is passed in the ipspace BD_Port in the resource list will be
deleted based on the information that is supplied in the resource
list. If the resource list is passed in the NACL IPspace Deletion
Arguments should NOT be passed in.

=item C<< NACL IPspace BD_Port Deletion Arguments >>

(Required)
These are arguments that can be passed into NACL::C::NetworkPortBroadcastDomain->assign_port.
If these arguments are passed in the resource_list argument should NOT
be passed in.

=item C<< command_interface => $NACL_command_interface >>

(Optional)
This is the command_interface that will be used during NACL calls.  If a
command_interface is not passed in then one will be created.  Passing in a
command_interface helps performance because for every command_interface created
the version of the Filer must be determined by creating a connection to the filer
and then sending commands to get the version.

=item C<< method_timeout => $NACL_method_timeout >>

(Optional)
This is the method_timeout that will be used during NACL calls. If a
method_timeout is not passed in then the method_timeout will be defaulted.

=back

=item Return

This method returns a True or False based on whether the method succeeds or fails
in the deletion of the IPspace BD_Port based on the Options passed in.

=item Exceptions

This method produces and throws no exceptions of its own but pass through any
uncaught exceptions from other library calls.

=cut

sub delete {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %args = @_;

    # Locals
    my $parent_answer;

    # Globals that will also be set as globals in the base class
    $args{'Name'} = "ipspace_BD_Port";

    # Super::_delete is in Task.pm
    $parent_answer = $pkg->SUPER::_delete(%args);

    return $parent_answer;
} ## end sub delete


#==============================================================================
# Method:     _nacl_create_call
# Objective:  Makes the raw NACL calls for creation of a ipspace BD_Port.
# Details:    Makes the NACL::C::NetworkPortBroadcastDomain->assign_port() call. Also works
#             around some common non-fatal errors that can be return from
#             ipspace BD_Port creation and then has a retry mechanism to retry
#             ipspace creation.
# Inputs:     NACL::C::NetworkPortBroadcastDomain->assign_port arguments
#             retry => integer, is the current retry the method is on.
#                    This is used to count how many times the _nacl_create_call
#                    has been retried so that the retries can be limited.
# Outputs:    Returns whether the create was successful or a hash with success
#             and whether the ipspace BD_Port was a duplicate if a hash is
#             requested
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
#==============================================================================
sub _nacl_create_call {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %args = @_;

    # Locals
    my $if_exists
        = exists($args{if_exists}) ? delete($args{if_exists}) : 'die';
    my $was_created = 0;
    my $was_deleted = 0;
    if (!exists($args{command_interface})) {
        my @Nodes = NACL::C::Node->find();
        $args{command_interface} = $Nodes[0];
    }

    my $port = delete($args{node}).":".delete($args{port});
    my $mtu  = delete($args{mtu});
    # Create the name of the default bd based on the MTU.
    my $default_broadcast_domain;
    my $retry_with_default_bd;
    if ($mtu == 1500) {
        $default_broadcast_domain = 'Default';
        $retry_with_default_bd = 0;
    } else {
        $default_broadcast_domain = 'bd_' . $mtu;
        $retry_with_default_bd = 1;
    }

    # Remove the port from the default IPspace and BDs.
    NACL::STask::NetworkPortBroadcastDomain->remove_ports(
        'command_interface' => $args{command_interface},
        'ports'             => [$port],
        'ipspace'           => 'Default',
        'broadcast-domain'  => $default_broadcast_domain,
        nacltask_if_purged  => 'pass',
        _was_deleted        => \$was_deleted,
        );

    # If we didn't remove it from something in the previous call
    # and we didn't use the Default BD then try to remove from Default
    # BD.
    if(not $was_deleted and $retry_with_default_bd) {
        NACL::STask::NetworkPortBroadcastDomain->remove_ports(
            'command_interface' => $args{command_interface},
            'ports'             => [$port],
            'ipspace'           => 'Default',
            'broadcast-domain'  => 'Default',
            nacltask_if_purged  => 'pass',
            _was_deleted        => \$was_deleted,
            );
    }

    # Execute Add ports
    NACL::STask::NetworkPortBroadcastDomain->add_ports(ports => [$port],
                                                       nacltask_if_exists => $if_exists,
                                                       _was_created       => \$was_created,
                                                       %args);
    if($was_created) {
        $Log->comment("Assigned port: " . $port .
                      " to ipspace: "    . $args{ipspace} .
                      " and broadcast domain: " . $args{'broadcast-domain'});
    } else {
        $Log->comment("Port: " . $port .
                      " to ipspace: "    . $args{ipspace} .
                      " and broadcast domain: " . $args{'broadcast-domain'} . " already assigned");
    }
    $Log->exit() if $may_exit;
    # Returning either a list (hash) or scalar TRUE
    return (wantarray()) ? ('success'=>TRUE, 'duplicated'=>!$was_created) : TRUE;

} ## end sub _nacl_create_call

#==============================================================================
# Method:     _nacl_delete_call
# Objective:  Makes the raw NACL calls for deletion of a ipspace BD_Port.
# Details:    Makes the NACL::C::NetworkPortBroadcastDomain->assign_port() calls. Also works around
#             some common non-fatal errors that can be return from ipspace
#             BD_Port deletion and then has a retry mechanism to retry ipspace
#             BD_Port deletion.
# Inputs:     NACL::C::NetworkPortBroadcastDomain->assign_port arguments
#             retry => integer, is the current retry the method is on.
#                    This is used to count how many times the _nacl_delete_call
#                    has been retried so that the retries can be limited.
# Outputs:    Returns whether the create was successful or a hash with success
#             if a hash is requested
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
#==============================================================================
sub _nacl_delete_call {
    $Log->enter() if $may_enter;
    my $pkg   = shift;
    my %args = @_;

    # Locals
    $args{'method-timeout'} = 3600 unless exists($args{'method-timeout'});

    if (!exists($args{command_interface})) {
        my @Nodes = NACL::C::Node->find();
        $args{command_interface} = $Nodes[0];
    }

    my $delete_policy = delete $args{delete_policy} // 'warn';
    my $was_created = 0;
    my $port = delete($args{node}).":".delete($args{port});
    my $mtu  = delete($args{mtu});
    # Create the name of the default bd based on the MTU.
    my $default_broadcast_domain;
    my $retry_with_default_bd;
    if ($mtu == 1500) {
        $default_broadcast_domain = 'Default';
        $retry_with_default_bd = 0;
    } else {
        $default_broadcast_domain = 'bd_' . $mtu;
        $retry_with_default_bd = 1;
    }

    try {
        NACL::STask::NetworkPortBroadcastDomain->add_ports(command_interface => $args{command_interface},
                                                       ports              => [$port],
                                                       ipspace           => 'Default',
                                                       'broadcast-domain'=> $default_broadcast_domain,
                                                       nacltask_if_exists => 'reuse',
                                                       _was_created       => \$was_created,
                                                       );
        $Log->comment("Network Port: " . $port . " was assigned to IPSpace: Default " .
                      "and Broadcast Domain: $default_broadcast_domain.");
    } catch NACL::C::Exceptions::NetworkPortBroadcastDomain::DoesNotExist with {
        my $exception = shift;
        if(not $retry_with_default_bd) {
            if ($delete_policy eq "warn") {
                $Log->warn("Cannot Assign port: " . $port . "to" .
                           "Default IPSpace's $default_broadcast_domain Broadcast Domain. " .
                           "It does not exist in the cluster." .
                           "Warning and Continuing Test.\n" .
                           $exception->text());
            } elsif ($delete_policy eq "fail") {
                $exception->throw();
            }
        } # else do nothing we will retry with default bd
    } catch NACL::C::Exceptions::NetworkPortBroadcastDomain::PortAlreadyAssigned with {
        # This should only happen if the user is cleaning up an already cleaned up config.
        # or if the user moved this port to a BD that was not in the XMl.
        # Because of the above we assume this should pass because this reassignment is really
        # a safety net so that manual intervention is not needed after a test with BDs in the XML.
        $Log->comment("Attempted to assign port: " . $port .
                      "to the Default IP Space's $default_broadcast_domain BD. " .
                      "but it is already assigned to a different BD and or IPspace. " .
                      "Since this is cleanup I assume the user wanted to leave the port where it is currently.");
        $retry_with_default_bd = 0;
    } otherwise {
        my $exception = shift;
        if ($delete_policy eq "warn") {
            $Log->warn("Cannot Assign port: " . $port . "to" .
                       "Default IPSpace's $default_broadcast_domain Broadcast Domain. " .
                       "It does not exist in the cluster." .
                       "Warning and Continuing Test.\n" .
                       $exception->text());
            $retry_with_default_bd = 0;
        } elsif ($delete_policy eq "fail") {
            $exception->throw();
        }
    };

    if($retry_with_default_bd and not $was_created) {
        try {
            NACL::STask::NetworkPortBroadcastDomain->add_ports(command_interface => $args{command_interface},
                                                           ports              => [$port],
                                                           ipspace           => 'Default',
                                                           'broadcast-domain'=> 'Default',
                                                           nacltask_if_exists => 'reuse',
                                                           );
            $Log->comment("Network Port: " . $port . " was assigned to IPSpace: Default " .
                          "and Broadcast Domain: Default.");
        } otherwise {
            my $exception = shift;
            if ($delete_policy eq "warn") {
                $Log->warn("Cannot Assign port: " . $port . "to" .
                           "Default IPSpace's Default Broadcast Domain. " .
                           "It does not exist in the cluster." .
                           "Warning and Continuing Test.\n" .
                           $exception->text());
                $retry_with_default_bd = 0;
            } elsif ($delete_policy eq "fail") {
                $exception->throw();
            }
        };
    }

    $Log->exit() if $may_exit;
    return TRUE;

} ## end sub _nacl_delete_call

#==============================================================================
# Method:     _nacl_wait_call
# Objective:  Makes a NACL call to 'wait_for_creation.
# Details:    Makes the call to NACL::STask::NetworkPortBroadcastDomain->wait_for_creation();
#             that will wait for the node to finish creating the aggr.
# Inputs:     NACL::STask::NetworkPortBroadcastDomain->wait arguments
# Outputs:    Returns whether the create was successful TRUE or not FALSE
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
# TODO: Implement this method when needed
#==============================================================================
sub _nacl_wait_call {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = @_;

    $Log->exit() if $may_exit;
    return TRUE;
}

#==============================================================================
# Method:     _get_dtd_nacl_hashes (translate, ignore and delete)
# Objective:  Getter methods for the variables of the same name.
# Details:    Retrieves the hash or a reference to one of the following:
#             - DTD_To_NACL_Translate.
#             - DTD_To_NACL_Ignore.
#             - Delete_Args.
# Inputs:     none.
# Outputs:    The reference or hash variable.
# Exceptions:
#==============================================================================
sub _get_dtd_nacl_translate {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = @_;

    $Log->exit() if $may_exit;
    return (wantarray()) ? %DTD_To_NACL_Translate : \%DTD_To_NACL_Translate;
}
#==============================================================================
sub _get_dtd_nacl_ignore {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = @_;

    $Log->exit() if $may_exit;
    return (wantarray()) ? %DTD_To_NACL_Ignore : \%DTD_To_NACL_Ignore;
}
#==============================================================================
sub _get_dtd_nacl_delete {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = @_;

    $Log->exit() if $may_exit;
    return (wantarray()) ? %Delete_Args : \%Delete_Args;
}

1;
