# Copyright (c) 2001-2015 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Ifgrp Controller layer module
## @author dl-san-automation-core
## @status shared
## @pod here

=head1 NAME

NACL::AppBuilder::EnGenX::Controller::Ifgrp

=head1 DESCRIPTION

NACL::AppBuilder::EnGenX::Controller::Ifgrp is a derived class of
NACL::AppBuilder::EnGenX::Controller::Task.  It is used to Create and
Delete ONTAP interface groups. All methods are static.

=cut

package NACL::AppBuilder::EnGenX::Controller::Ifgrp;

# Compiler directives
use strict;
use warnings;

# NATE and Standard Module import
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use NATE::Util::DataTypes qw(TRUE FALSE);
use NATE::STAF::Peer qw($Parent);
use Data::Dumper;
use Params::Validate
    qw(validate_with validate SCALAR ARRAYREF BOOLEAN UNDEF SCALARREF HASHREF);

# NACL and EnGenX Libs
use NACL::STask::NetworkPortIfgrp;

#Inherit base class
use base qw( NACL::AppBuilder::EnGenX::Controller::Task );

use vars qw(%DTD_To_NACL_Translate
    %DTD_To_NACL_Ignore
    %Delete_Args
);

# Arguments after converting from xml to a
# resource list that need their keys translated to
# keys that NACL methods will accept.
%DTD_To_NACL_Translate = (
    name          => 'ifgrp',
    'port-list'   => 'ports',
    _reuse        => 'if_exists'
);

# Arguments that are used by EnGenX but cannot be
# passed into the NACL methods
%DTD_To_NACL_Ignore = (
        'port-list'     => '',
);

# Arguments that can be passed into the NACL delete
# command
%Delete_Args = (
    command_interface => '',
    ifgrp             => '',
    node              => '',
);

=head1 METHODS

=head2 create

(Static method) Request creation of one or multiple Ifgrps serially
or in parallel and return whether the Ifgrp(s) was created successfully.
This method passes all user arguments through to NACL::AppBuilder::EnGenX::Controller::Task
and adds specific arguments used by the EnGenX framework.

  NACL::AppBuilder::EnGenX::Controller::Ifgrp->create(
            command_interface => $Node,
            ifgrp             => $Ifgrp_name,
            node              => $Node_name,
            'distr-func'      => $Distribution_algorithm
            mode              => $Mode_name
            'port-list'       => $port_list
            'method-timeout'  => 3600,
            );


  NACL::AppBuilder::EnGenX::Controller::Ifgrp->create(
            command_interface  => $Node,
            resource_list      => @Ifgrps,
            'method-timeout'   => 3600,
            );

=over

=item Options

Either the resource_list is required or the NACL Ifgrp Creation Arguments
are required but both should NOT be used in the same call.

=over

=item C<< resource_list => @resource_list >>

(Optional)
If this is passed in the Ifgrps in the resource list will be created
based on the information that is supplied in the resource list. If
the resource list is passed in the NACL Ifgrp Creation Arguments should
NOT be passed in.

=item C<< NACL Ifgrp Creation Arguments >>

(Optional)
These are arguments that can be passed into NACL::STask::NetworkPortIfgrp->Create.
If these arguments are passed in the resource_list argument should NOT
be passed in

=item C<< command_interface => $NACL_command_interface >>

(Optional)
This is the command_interface that will be used during NACL calls.  If a
command_interface is not passed in then one will be created.  Passing in a
command_interface helps performance because for every command_interface created
the version of the Filer must be determined by creating a connection to the filer
and then sending commands to get the version.

=item C<< method_timeout => $NACL_method_timeout >>

(Optional)
This is the method_timeout that will be used during NACL calls. If a
method_timeout is not passed in then the method_timeout will be defaulted.

=back

=item Return

This method returns a True or False based on whether the method succeeds or fails
in the creation of the Ifgrp(s) based on the Options passed in.

=item Exceptions

This method produces and throws no exceptions of its own but pass through any
uncaught exceptions from other library calls.

=cut

sub create {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %args = @_;

    # Locals
    my $parent_answer;

    # Globals that will also be set as globals in the base class
    $args{'Name'} = "ifgrp";

    # Super::_create is in Task.pm
    $parent_answer = $pkg->SUPER::_create(%args);

    $Log->exit() if $may_exit;
    return $parent_answer;
} ## end sub create

=head2 delete

(Static method) Request deletion of one or multiple ifgrps serially
or in parallel and return whether the ifgrp(s) was deleted successfully.
This method passes all user arguments through to NACL::AppBuilder::EnGenX::Controller::Task
and adds specific arguments used by the EnGenX framework.

  NACL::AppBuilder::EnGenX::Controller::Ifgrp->delete(
            command_interface => $Node,
            ifgrp             => $Ifgrp_name,
            node              => $Node_name,
            'method-timeout'  => 3600,
            );

  NACL::AppBuilder::EnGenX::Controller::Ifgrp->delete(
            command_interface => $Node,
            resource_list      => @Ifgrps,
            'method-timeout'  => 3600,
            );

=over

=item Options

Either the resource_list is required or the NACL Ifgrp Deletion Arguments
are required but both should NOT be used in the same call.

=over

=item C<< resource_list => @resource_list >>

(Required)
If this is passed in the Ifgrps in the resource list will be deleted
based on the information that is supplied in the resource list. If
the resource list is passed in the NACL Ifgrp Deletion Arguments should
NOT be passed in.

=item C<< NACL Ifgrp Deletion Arguments >>

(Required)
These are arguments that can be passed into NACL::STask::NetworkPortIfgrp->Delete.
If these arguments are passed in the resource_list argument should NOT
be passed in.

=item C<< command_interface => $NACL_command_interface >>

(Optional)
This is the command_interface that will be used during NACL calls.  If a
command_interface is not passed in then one will be created.  Passing in a
command_interface helps performance because for every command_interface created
the version of the Filer must be determined by creating a connection to the filer
and then sending commands to get the version.

=item C<< method_timeout => $NACL_method_timeout >>

(Optional)
This is the method_timeout that will be used during NACL calls. If a
method_timeout is not passed in then the method_timeout will be defaulted.

=back

=item Return

This method returns a True or False based on whether the method succeeds or fails
in the deletion of the Ifgrp(s) based on the Options passed in.

=item Exceptions

This method produces and throws no exceptions of its own but pass through any
uncaught exceptions from other library calls.

=cut

sub delete {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %args = @_;

    # Locals
    my $parent_answer;

    # Globals that will also be set as globals in the base class
    $args{'Name'} = "ifgrp";

    # Super::_delete is in Task.pm
    $parent_answer = $pkg->SUPER::_delete(%args);

    $Log->exit() if $may_exit;
    return $parent_answer;
} ## end sub delete


#==============================================================================
# Method:     _nacl_create_call
# Objective:  Makes the raw NACL calls for creation of a Ifgrp.
# Details:    Makes the NACL::STask::NetworkPortIfgrp->create() call. 
# Inputs:     NACL::STask::NetworkPortIfgrp->create arguments
# Outputs:    Returns whether the create was successful or a hash with success
#             and whether the Ifgrp was a duplicate if a hash is requested
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
#==============================================================================
sub _nacl_create_call {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %args = @_;
    # Locals
    my $was_created = 0;
    my $if_exists
        = exists($args{if_exists}) ? delete($args{if_exists}) : 'die';
    $args{'method-timeout'} = 3600 unless exists($args{'method-timeout'});
    
    if(!(ref($args{ports}) eq 'ARRAY')) {
        # split comma separated string ports into an array for NACL call
        my @port_array = split(',' , $args{ports});
        $args{ports} = \@port_array;
    }

    if (!exists($args{command_interface})) {
        my @Nodes = NACL::C::Node->find();
        $args{command_interface} = $Nodes[0];
    }

    # Execute Create
    NACL::STask::NetworkPortIfgrp->create(%args,
                                    nacltask_if_exists => $if_exists,
                                    _was_created => \$was_created);
    if($was_created) {
        $Log->comment("Using newly created Ifgrp:" . $args{ifgrp});
    } else {
        $Log->comment("Using already created Ifgrp:" . $args{ifgrp});
    }

    $Log->exit() if $may_exit;
    # Returning either a list (hash) or scalar TRUE
    return (wantarray()) ? ('success'=>TRUE, 'duplicated'=>!$was_created) : TRUE;


} ## end sub _nacl_create_call

#========================================================================================
# Method:     _nacl_delete_call
# Objective:  Makes the raw NACL calls for deletion of a Ifgrp.
# Details:    Makes the NACL::STask::NetworkPortIfgrp->delete() calls.
# Inputs:     NACL::STask::NetworkPortIfgrp->delete arguments
# Outputs:    Returns whether the delete was successful or a hash with success
#             if a hash is requested
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
#==============================================================================
sub _nacl_delete_call {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %args = @_;
    # Locals
    my $was_deleted = 0;
    $args{'method-timeout'} = 3600 unless exists($args{'method-timeout'});
    if (!exists($args{command_interface})) {
        my @Nodes = NACL::C::Node->find();
        $args{command_interface} = $Nodes[0];
    }

    NACL::STask::NetworkPortIfgrp->purge(%args,
                                        nacltask_if_purged => 'pass',
                                        _was_deleted => \$was_deleted);

    if($was_deleted) {
        $Log->comment("Ifgrp:" . $args{ifgrp} . " was deleted");
    } else {
        $Log->comment("Ifgrp:" . $args{ifgrp} . " was already deleted");
    }

    $Log->exit() if $may_exit;
    return TRUE;

} ## end sub _nacl_delete_call

#==============================================================================
# Method:     _get_dtd_nacl_hashes (translate, ignore and delete)
# Objective:  Getter methods for the variables of the same name.
# Details:    Retrieves the hash or a reference to one of the following:
#             - DTD_To_NACL_Translate.
#             - DTD_To_NACL_Ignore.
#             - Delete_Args.
# Inputs:     none.
# Outputs:    The reference or hash variable.
# Exceptions:
#==============================================================================
sub _get_dtd_nacl_translate {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = @_;

    $Log->exit() if $may_exit;
    return (wantarray()) ? %DTD_To_NACL_Translate : \%DTD_To_NACL_Translate;
}
#==============================================================================
sub _get_dtd_nacl_ignore {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = @_;

    $Log->exit() if $may_exit;
    return (wantarray()) ? %DTD_To_NACL_Ignore : \%DTD_To_NACL_Ignore;
}
#==============================================================================
sub _get_dtd_nacl_delete {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = @_;

    $Log->exit() if $may_exit;
    return (wantarray()) ? %Delete_Args : \%Delete_Args;
}

1;
