# $Id$

#
# Copyright (c) 2001-2012 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Initiator Controller Layer Module
## @author dl-san-automation
## @status public
## @pod here

=head1 NAME

NACL::AppBuilder::EnGenX::Controller::Initiator

=head1 DESCRIPTION

NACL::AppBuilder::EnGenX::Controller::Initiator is a derived class of
NACL::AppBuilder::EnGenX::Controller::Task.  It is used to Create and
Delete ONTAP Initiators. All methods are static.

=cut

package NACL::AppBuilder::EnGenX::Controller::Initiator;

# Compiler directives
use strict;
use warnings;

# NATE and Standard Module import
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use NATE::Util::DataTypes qw(TRUE FALSE);
use Data::Dumper;
use Params::Validate
    qw(validate_with validate SCALAR ARRAYREF BOOLEAN UNDEF SCALARREF HASHREF);

# NACL and EnGenX Libs
use NACL::STask::LunIgroup;
use NACL::AppBuilder::EnGenX::SubProcess;
use NACL::AppBuilder::EnGenX::EnGenXUtil;


# Inherit base class
use base qw( NACL::AppBuilder::EnGenX::Controller::Task );

use vars qw(%DTD_To_NACL_Translate
    %DTD_To_NACL_Ignore
    %Delete_Args
);

# Arguments after converting from xml to a
# resource list that need their keys translated to
# keys that NACL methods will accept.
%DTD_To_NACL_Translate = (
    _reuse => 'if_exists',
    name   => 'initiator'
);

# Arguments that are used by EnGenX but cannot be
# passed into the NACL methods
%DTD_To_NACL_Ignore = (parallel => '',);

# Arguments that can be passed into the NACL delete
# command
%Delete_Args = (
    command_interface => '',
    apiset_must       => '',
    apiset_should     => '',
    igroup            => '',
    initiator         => '',
    vserver           => '',
    force             => '',
);

=head1 METHODS

=head2 create

(Static method) Request creation of one or multiple initiators serially
or in parallel and return whether the initiator(s) was created successfully.
This method passes all user arguments through to NACL::AppBuilder::EnGenX::Controller::Task
and adds specific arguments used by the EnGenX framework.

  NACL::AppBuilder::EnGenX::Controller::Initiator->create(
            command_interface => $Node,
            initiator         => ( $Initiator_name),
            vserver           => $Vserver_name,
            igroup            => $Igroup name,
            'method-timeout'  => 3600,
            );

  NACL::AppBuilder::EnGenX::Controller::Initiator->create(
            command_interface => $Node,
            initiator         => ( $Initiator_name),
            vserver           => $Vserver_name,
            igroup            => $Igroup name,
            'method-timeout'  => 3600,
            count             => 5
            );

  NACL::AppBuilder::EnGenX::Controller::Initiator->create(
            command_interfac  => $Node,
            resource_list     => @Initiators,
            'method-timeout'  => 3600,
            );

  NACL::AppBuilder::EnGenX::Controller::Initiator->create(
            command_interface => $Node,
            resource_list     => @Initiators,
            );

  NACL::AppBuilder::EnGenX::Controller::Initiator->create(
            resource_list     => @Initiators,
            'method-timeout'  => 3600,
            );

  NACL::AppBuilder::EnGenX::Controller::Initiator->create(
            resource_list     => @Initiators,
            );

=over

=item Options

Either the resource_list is required or the NACL Initiator Creation Arguments
are required but both should NOT be used in the same call.

=over

=item C<< resource_list => @resource_list >>

(Optional)
If this is passed in the initiators in the resource list will be created
based on the information that is supplied in the resource list. If
the resource list is passed in the NACL Initiator Creation Arguments should
NOT be passed in. (This is an array of EnGenX::Resource() objects not
an EnGenX::Resource_List() currently)

=item C<< NACL Initiator Creation Arguments >>

(Optional)
These are arguments that can be passed into NACL::C::Igroup->Add.
If these arguments are passed in the resource_list argument should NOT
be passed in

=item C<< count => $initiator_count >>

(Optional)
This arguments is used to create multiple initiators from the same NACL Initiator
Creation Arguments.  To create the names for the initiators, the initiator name
passed in is appended with the index that it was create at starting at 1.

=item C<< command_interface => $NACL_command_interface >>

(Optional)
This is the command_interface that will be used during NACL calls.  If a
command_interface is not passed in then one will be created.  Passing in a
command_interface helps performance because for every command_interface created
the version of the Filer must be determined by creating a connection to the filer
and then sending commands to get the version.

=item C<< method_timeout => $NACL_method_timeout >>

(Optional)
This is the method_timeout that will be used during NACL calls. If a
method_timeout is not passed in then the method_timeout will be defaulted.

=back

=item Return

This method returns a True or False based on whether the method succeeds or fails
in the creation of the Initiator(s) based on the Options passed in.

=item Exceptions

This method produces and throws no exceptions of its own but pass through any
uncaught exceptions from other library calls.

=cut

sub create {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %args = @_;

    # Locals
    my $parent_answer;

    # Globals that will also be set as globals in the base class
    $args{'Name'} = "initiator";

    # Super::_create is in Task.pm
    $parent_answer = $pkg->SUPER::_create(%args);

    return $parent_answer;
} ## end sub create

=head2 delete

(Static method) Request deletion of one or multiple initiators serially
or in parallel and return whether the initiator(s) was deleted successfully.
This method passes all user arguments through to NACL::AppBuilder::EnGenX::Controller::Task
and adds specific arguments used by the EnGenX framework.

  NACL::AppBuilder::EnGenX::Controller::Initiator->delete(
            command_interface => $Node,
            initiator            => ( $Initiator_name),
            vserver           => $Vserver_name,
            igroup            => $Igroup name,
            'method-timeout'  => 3600,
            );

  NACL::AppBuilder::EnGenX::Controller::Initiator->delete(
            command_interface => $Node,
            initiator            => ( $Initiator_name),
            vserver           => $Vserver_name,
            igroup            => $Igroup name,
            'method-timeout'  => 3600,
            count             => 5
            );

  NACL::AppBuilder::EnGenX::Controller::Initiator->delete(
            command_interface => $Node,
            resource_list      => @Initiators,
            'method-timeout'  => 3600,
            );

  NACL::AppBuilder::EnGenX::Controller::Initiator->delete(
            command_interface => $Node,
            resource_list      => @Initiators,
            );

  NACL::AppBuilder::EnGenX::Controller::Initiator->delete(
            resource_list      => @Initiators,
            'method-timeout'  => 3600,
            );

  NACL::AppBuilder::EnGenX::Controller::Initiator->delete(
            resource_list      => @Initiators,
            );

=over

=item Options

Either the resource_list is required or the NACL Initiator Deletion Arguments
are required but both should NOT be used in the same call.

=over

=item C<< resource_list => @resource_list >>

(Required)
If this is passed in the initiators in the resource list will be deleted
based on the information that is supplied in the resource list. If
the resource list is passed in the NACL Initiator Deletion Arguments should
NOT be passed in.

=item C<< NACL Initiator Deletion Arguments >>

(Required)
These are arguments that can be passed into NACL::C::Igroup->Remove.
If these arguments are passed in the resource_list argument should NOT
be passed in.

=item C<< count => $initiator_count >>

(Optional)
This arguments is used to delete multiple initiators from the same NACL Initiator
Deletion Arguments.  To delete the names for the initiators, the initiator name
passed in is appended with the index that it was deleted at starting at 1.

=item C<< command_interface => $NACL_command_interface >>

(Optional)
This is the command_interface that will be used during NACL calls.  If a
command_interface is not passed in then one will be created.  Passing in a
command_interface helps performance because for every command_interface created
the version of the Filer must be determined by creating a connection to the filer
and then sending commands to get the version.

=item C<< method_timeout => $NACL_method_timeout >>

(Optional)
This is the method_timeout that will be used during NACL calls. If a
method_timeout is not passed in then the method_timeout will be defaulted.

=back

=item Return

This method returns a True or False based on whether the method succeeds or fails
in the deletion of the Initiator(s) based on the Options passed in.

=item Exceptions

This method produces and throws no exceptions of its own but pass through any
uncaught exceptions from other library calls.

=cut

sub delete {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %args = @_;

    # Locals
    my $parent_answer;

    # Globals that will also be set as globals in the base class
    $args{'Name'} = "initiator";

    # Super::_delete is in Task.pm
    $parent_answer = $pkg->SUPER::_delete(%args);

    return $parent_answer;
} ## end sub delete


#==============================================================================
# Method:     _nacl_create_call
# Objective:  Makes the raw NACL calls for creation of a initiator.
# Details:    Makes the NACL::C::Igroup->add() call. Also works
#             around some common non-fatal errors that can be return from
#             initiator creation and then has a retry mechanism to retry initiator
#             creation.
# Inputs:     NACL::C::Igroup->add arguments
#             retry => integer, is the current retry the method is on.
#                    This is used to count how many times the _nacl_create_call
#                    has been retried so that the retries can be limited.
# Outputs:    Returns whether the create was successful or a hash with success
#             and whether the initiator was a duplicate if a hash is requested
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
#==============================================================================
sub _nacl_create_call {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %args = @_;

    # Locals
    my $was_created = 0;
    my $if_exists      = exists($args{if_exists}) ? delete($args{if_exists}) : 'die';
    my $retry = exists $args{retry} ? delete $args{retry} : 0;
    if (!exists($args{command_interface})) {
        my @Nodes = NACL::C::Node->find();
        $args{command_interface} = $Nodes[0];
    }
    if (ref(($args{initiator})) ne 'ARRAY') {
        $Log->comment('initiator was "' . ref($args{initiator}) . '"and not an '
                      . 'arrayref and was converted to an arrayref');
        $args{initiator} = [$args{initiator}];
    }
    # Loop through the initiators and if a wwpn without colons then add colons
    foreach my $initiator ($args{initiator}) {
        if ($initiator =~ /^[A-Fa-f0-9]{16}$/) {
            $initiator = NACL::AppBuilder::EnGenX::Util->normalize_fcp_wwn(-wwn => $initiator);
        }
    }
    NACL::STask::LunIgroup->add(%args, nacltask_if_exists => $if_exists, _was_created => \$was_created);
    if(!$was_created) {
        $Log->comment("Using already created initiator:" . $args{initiator}.
                      "for vserver:". $args{vserver}.
                      "igroup:". Dumper($args{igroup}));    
    } else {
        $Log->comment("Using newly created initiator:" . $args{initiator} .
                      "for vserver: ". $args{vserver}.
                      "igroup:". Dumper($args{igroup}));
    }

    $Log->exit() if $may_exit;
    # Returning either a list (hash) or scalar TRUE
    return (wantarray()) ? ('success'=>TRUE, 'duplicated'=>$was_created) : TRUE;

} ## end sub _nacl_create_call

#==============================================================================
# Method:     _nacl_delete_call
# Objective:  Makes the raw NACL calls for deletion of a initiator.
# Details:    Makes the NACL::C::Igroup->remove() calls. Also works around
#             some common non-fatal errors that can be return from initiator
#             deletion and then has a retry mechanism to retry initiator deletion.
# Inputs:     NACL::C::Igroup->remove arguments
#             retry => integer, is the current retry the method is on.
#                    This is used to count how many times the _nacl_delete_call
#                    has been retried so that the retries can be limited.
# Outputs:    Returns whether the create was successful or a hash with success
#             if a hash is requested
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
#==============================================================================
sub _nacl_delete_call {
    $Log->enter() if $may_enter;
    my $pkg   = shift;
    my %args  = @_;
    my $was_deleted = 0;
    # Locals
    my $retry = exists $args{retry} ? delete $args{retry} : 0;
    $args{'method-timeout'} = 3600 unless exists($args{'method-timeout'});

    if (!exists($args{command_interface})) {
        my @Nodes = NACL::C::Node->find();
        $args{command_interface} = $Nodes[0];
    }
    if (ref(($args{initiator})) ne 'ARRAY') {
        $Log->comment('initiator was "' . ref($args{initiator}) . '"and not an '
                      . 'arrayref and was converted to an arrayref');
        $args{initiator} = [$args{initiator}];
    }
    # Loop through the initiators and if a wwpn without colons then add colons
    foreach my $initiator ($args{initiator}) {
        if ($initiator =~ /^[A-Fa-f0-9]{16}$/) {
            $initiator = NACL::AppBuilder::EnGenX::Util->normalize_fcp_wwn(-wwn => $initiator);
        }
    }
    NACL::STask::LunIgroup->remove(%args, nacltask_if_purged => 'pass', _was_deleted => \$was_deleted);
    if (!$was_deleted) {
        $Log->comment("Initiator:" . $args{initiator} . " was already deleted".
                      "for vserver: ". $args{vserver}. " igroup:". Dumper($args{igroup}));
    } else {
        $Log->comment("Initiator:" . $args{initiator} . " was newly deleted".
                      "for vserver: ". $args{vserver}. " igroup:". Dumper($args{igroup}));
    }
    $Log->exit() if $may_exit;
    return TRUE;

} ## end sub _nacl_delete_call

#==============================================================================
# Method:     _nacl_wait_call
# Objective:  Makes a NACL call to 'wait_for_creation.
# Details:    Makes the call to NACL::STask::Initiator->wait_for_creation();
#             that will wait for the node to finish creating the aggr.
# Inputs:     NACL::STask::Initiator->wait arguments
# Outputs:    Returns whether the create was successful TRUE or not FALSE
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
# TODO: Implement this method when needed
#==============================================================================
sub _nacl_wait_call {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = @_;

    $Log->exit() if $may_exit;
    return TRUE;
}

#==============================================================================
# Method:     _get_dtd_nacl_hashes (translate, ignore and delete)
# Objective:  Getter methods for the variables of the same name.
# Details:    Retrieves the hash or a reference to one of the following:
#             - DTD_To_NACL_Translate.
#             - DTD_To_NACL_Ignore.
#             - Delete_Args.
# Inputs:     none.
# Outputs:    The reference or hash variable.
# Exceptions:
#==============================================================================
sub _get_dtd_nacl_translate {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = @_;

    $Log->exit() if $may_exit;
    return (wantarray()) ? %DTD_To_NACL_Translate : \%DTD_To_NACL_Translate;
}
#==============================================================================
sub _get_dtd_nacl_ignore {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = @_;

    $Log->exit() if $may_exit;
    return (wantarray()) ? %DTD_To_NACL_Ignore : \%DTD_To_NACL_Ignore;
}
#==============================================================================
sub _get_dtd_nacl_delete {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = @_;

    $Log->exit() if $may_exit;
    return (wantarray()) ? %Delete_Args : \%Delete_Args;
}

1;
