# $Id: //depot/prod/test/main/lib/EnGenX/Controller/Lif.pm#2 $

#
# Copyright (c) 2015 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Lif Controller layer module
## @author dl-san-automation-core
## @status shared
## @pod here

=head1 NAME

NACL::AppBuilder::EnGenX::Controller::Lif

=head1 DESCRIPTION

EnGenX::Controller::Lif is a derived class of
NACL::AppBuilder::EnGenX::Controller::Task.  It is used to Create and
Delete ONTAP network interfaces. All methods are static.

=cut

package NACL::AppBuilder::EnGenX::Controller::Lif;

# Compiler directives
use strict;
use warnings;

# NATE and Standard Module import
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use NATE::Util::DataTypes qw(TRUE FALSE);
use NATE::STAF::Peer qw($Parent);
use Data::Dumper;
use Params::Validate
    qw(validate_with validate SCALAR ARRAYREF BOOLEAN UNDEF SCALARREF HASHREF);
use NATE::Util::Sockets qw(get_fqdn);

# NACL and EnGenX Libs
use NACL::C::NetworkInterface;
use NACL::CS::NetworkInterface;
use NACL::STask::NetworkInterface;
use NACL::CS::NetworkInterface;
use NACL::STask::NetworkRoutingGroupsRoute;
use NACL::C::Job;
use NACL::C::VserverIscsiInterface;
use Network::SCON_DB::IPReservation;
use IPManager;
use IPManager::Utils;

# Exceptions
use NATE::BaseException qw(:try);
use NATE::Exceptions::Argument qw(:try);

#Inherit base class
use base qw( NACL::AppBuilder::EnGenX::Controller::Task );

use vars qw(%DTD_To_NACL_Translate
    %DTD_To_NACL_Ignore
    %Delete_Args
    $Mem_Resource
);

# Arguments after converting from xml to a
# resource list that need their keys translated to
# keys that NACL methods will accept.
%DTD_To_NACL_Translate = (
    name   => 'lif',
    _reuse => 'nacltask_if_exists'
);

# Arguments that are used by EnGenX but cannot be
# passed into the NACL methods
%DTD_To_NACL_Ignore = (
        ips     => '',
        ips_key => '',
        gateway => '',
);

# Arguments that can be passed into the NACL delete
# command
%Delete_Args = (
    command_interface => '',
    lif               => '',
    vserver           => '',
    use_ipm           => '',
    use_ipam          => '',
);

=head1 METHODS

=head2 create

(Static method) Request creation of one or multiple Lifs serially
or in parallel and return whether the lif(s) was created successfully.
This method passes all user arguments through to NACL::AppBuilder::EnGenX::Controller::Task
and adds specific arguments used by the EnGenX framework.

  EnGenX::Controller::Lif->create(
            command_interface => $Node,
            lif               => $Lif_name,
            vserver           => $Vserver_name,
            'data-protocol'   => iscsi
            'home-port'       => e0c
            netmask           =>"255.255.192.0"
            address           => "10.63.9.163"
            role              => "data"
            'method-timeout'  => 3600,
            );

  EnGenX::Controller::Lif->create(
            command_interface => $Node,
            lif               => $Lif_name,
            vserver           => $Vserver_name,
            'data-protocol'   => iscsi
            'home-port'       => e0c
            netmask           =>"255.255.192.0"
            address           => "10.63.9.163"
            role              => "data"
            count             => 5
            );

  EnGenX::Controller::Lif->create(
            command_interfac  => $Node,
            resource_list      => @Lifs,
            'method-timeout'  => 3600,
            );

=over

=item Options

Either the resource_list is required or the NACL Lif Creation Arguments
are required but both should NOT be used in the same call.

=over

=item C<< resource_list => @resource_list >>

(Optional)
If this is passed in the lifs in the resource list will be created
based on the information that is supplied in the resource list. If
the resource list is passed in the NACL Lif Creation Arguments should
NOT be passed in.

=item C<< NACL Lif Creation Arguments >>

(Optional)
These are arguments that can be passed into NACL::C::NetworkInterface->Create.
If these arguments are passed in the resource_list argument should NOT
be passed in

=item C<< count => $lif_count >>

(Optional)
This arguments is used to create multiple lifs from the same NACL NetworkInterface
Creation Arguments.  To create the names for the lifs, the lif name
passed in is appended with the index that it was create at starting at 1.

=item C<< command_interface => $NACL_command_interface >>

(Optional)
This is the command_interface that will be used during NACL calls.  If a
command_interface is not passed in then one will be created.  Passing in a
command_interface helps performance because for every command_interface created
the version of the Filer must be determined by creating a connection to the filer
and then sending commands to get the version.

=item C<< method_timeout => $NACL_method_timeout >>

(Optional)
This is the method_timeout that will be used during NACL calls. If a
method_timeout is not passed in then the method_timeout will be defaulted.

=back

=item Return

This method returns a True or False based on whether the method succeeds or fails
in the creation of the Lif(s) based on the Options passed in.

=item Exceptions

This method produces and throws no exceptions of its own but pass through any
uncaught exceptions from other library calls.

TODO: The Task::_create_mulptiple sub will fail to create multiple LIFs because
      the method needs to also provide a unique IP ADDRESS. This unique address
      (or Array of usable address) would be passed by the user. None the less,
      the method needs to accept the array of IPs and use it when making the
      NACL call.

=cut

sub create {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %args = @_;

    # Locals
    my $parent_answer;

    # Globals that will also be set as globals in the base class
    $args{'Name'} = "lif";

    # Super::_create is in Task.pm
    $parent_answer = $pkg->SUPER::_create(%args);

    $Log->exit() if $may_exit;
    return $parent_answer;
} ## end sub create

=head2 delete

(Static method) Request deletion of one or multiple lifs serially
or in parallel and return whether the lif(s) was deleted successfully.
This method passes all user arguments through to NACL::AppBuilder::EnGenX::Controller::Task
and adds specific arguments used by the EnGenX framework.

  EnGenX::Controller::Lif->delete(
            command_interface => $Node,
            lif               => $Lif_name,
            vserver           => $Vserver_name,
            'method-timeout'  => 3600,
            );

  EnGenX::Controller::Lif->delete(
            command_interface => $Node,
            lif               => $Lif_name,
            vserver           => $Vserver_name,
            count             => 5
            );

  EnGenX::Controller::Lif->delete(
            command_interface => $Node,
            resource_list      => @Lifs,
            'method-timeout'  => 3600,
            );

=over

=item Options

Either the resource_list is required or the NACL Lif Deletion Arguments
are required but both should NOT be used in the same call.

=over

=item C<< resource_list => @resource_list >>

(Required)
If this is passed in the lifs in the resource list will be deleted
based on the information that is supplied in the resource list. If
the resource list is passed in the NACL Lif Deletion Arguments should
NOT be passed in.

=item C<< NACL Lif Deletion Arguments >>

(Required)
These are arguments that can be passed into NACL::C::NetworkInterface->Delete.
If these arguments are passed in the resource_list argument should NOT
be passed in.

=item C<< count => $lif_count >>

(Optional)
This arguments is used to delete multiple lifs from the same NACL NetworkInterface
Deletion Arguments.  To delete the names for the lifs, the lif name
passed in is appended with the index that it was deleted at starting at 1.

=item C<< command_interface => $NACL_command_interface >>

(Optional)
This is the command_interface that will be used during NACL calls.  If a
command_interface is not passed in then one will be created.  Passing in a
command_interface helps performance because for every command_interface created
the version of the Filer must be determined by creating a connection to the filer
and then sending commands to get the version.

=item C<< method_timeout => $NACL_method_timeout >>

(Optional)
This is the method_timeout that will be used during NACL calls. If a
method_timeout is not passed in then the method_timeout will be defaulted.

=back

=item Return

This method returns a True or False based on whether the method succeeds or fails
in the deletion of the Lif(s) based on the Options passed in.

=item Exceptions

This method produces and throws no exceptions of its own but pass through any
uncaught exceptions from other library calls.

=cut

sub delete {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %args = @_;

    # Locals
    my $parent_answer;

    # Globals that will also be set as globals in the base class
    $args{'Name'} = "lif";

    # Super::_delete is in Task.pm
    $parent_answer = $pkg->SUPER::_delete(%args);

    $Log->exit() if $may_exit;
    return $parent_answer;
} ## end sub delete


#==============================================================================
# Method:     _nacl_create_call
# Objective:  Makes the raw NACL calls for creation of a lif.
# Details:    Makes the NACL::C::Lif->create() call. Also works
#             around some common non-fatal errors that can be return from
#             lif creation and then has a retry mechanism to retry lif
#             creation.
# Inputs:     NACL::C::lif->create arguments
#             retry => integer, is the current retry the method is on.
#                    This is used to count how many times the _nacl_create_call
#                    has been retried so that the retries can be limited.
# Outputs:    Returns whether the create was successful or a hash with success
#             and whether the lif was a duplicate if a hash is requested
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
#==============================================================================
sub _nacl_create_call {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %args = @_;
    my $fqdn;

    # Locals
    my $was_created = 0;
    my $if_exists
        = exists($args{if_exists}) ? delete($args{if_exists}) : 'die';
    my $retry = exists $args{retry} ? delete $args{retry} : 0;
    if (exists($args{'auto-configuration'})) {
        delete($args{address});
        delete($args{netmask});
        delete($args{'netmask-length'});
    }

    # If subnet-name exists then we don't want address,netmask, or netmask-length
    if (exists($args{'subnet-name'})) {
        delete($args{address});
        delete($args{netmask});
        delete($args{'netmask-length'});
    }

    my $dup_ip = 0;
    $args{'method-timeout'} = 3600 unless exists($args{'method-timeout'});
    if (!exists($args{command_interface})) {
        my @Nodes = NACL::C::Node->find();
        $args{command_interface} = $Nodes[0];
    }

    my $ipspace = delete($args{ipspace});
    my $use_ipm = delete($args{use_ipm});
    my $use_ipam = delete($args{use_ipam});
    my $retry_count = delete($args{retry_count});

    my %ipm;
    if (exists($args{ipm})) {
        my $ipm_ref = delete($args{ipm});
        %ipm = %$ipm_ref;
    }

    my %ipam;
    if (exists($args{ipam})) {
        my $ipam_ref = delete($args{ipam});
        %ipam = %$ipam_ref;
    }
 

    if ($use_ipm) {

        delete($args{address});
        delete($args{'netmask-length'});
        delete($args{netmask});

        my $role = $args{role};

        my ($_IPM, $Error);
RESERVE: {
        $Log->comment("Reserve IP addresses from IP Manager");
        ($_IPM, $Error) = Network::SCON_DB::IPReservation->ip_addresses(
                                    'hostrec' => $args{command_interface}->hostrec,
                                    'role' => $role,
                                    'count' => 1,
                                    'job_id' => $args{lif},
                                    'query' => 'any',
                                    'version' => $ipm{'version'} // $ipm{VERSION},
                        );
        unless($_IPM) {
            $Log->comment("Failed to reserve IP addresses for IPManager DB: $Error");
            NATE::BaseException->throw("$Error");
        }
        if (($Error =~ /.*returned\s+0\s+rows.*/i) && ($retry_count != 0)){
            $retry_count--;
            redo RESERVE;
        }
    }
        $Log->comment("Got addresses for $args{command_interface}->hostrec");
        my $ips;
        @{$ips} = $_IPM->ip_address_list;
        my $netmask = $_IPM->ip_address_netmask;
        foreach my $_addr ( @$ips ) {
                $Log->comment("$_addr");
                $args{address} = $_addr;
                $args{netmask} = $netmask;
        }
    }
    elsif ($use_ipam) 
    {
        delete($args{address});
        delete($args{'netmask-length'});
        delete($args{netmask});

        my $role = $args{role};

        my $version = $ipm{'version'} // $ipm{VERSION};
 
        if ($version =~ /v4/i)
        {
            $version = 4;
        }
        elsif ($version =~ /v6/i)
        {
            $version = 6;
        }

        my $ipm;
        my %ipm_args;

        $Log->comment("Reserve IP addresses from IP Manager (IPAM)");

        %ipm_args = (
            'hostrec' => $args{command_interface}->hostrec,
            'count'   => 1,
            'version' => $version,
            'role'    => $role,
        );

        $ipm = new IPManager(%ipm_args);
        $ipm->reserve_addresses();

        my $hostname = $args{command_interface}->hostrec->hostname;

        $Log->comment("Got addresses for $hostname");
        my @ips;
        @ips = $ipm->get_ip_list();
        my $netmask = $ipm->get_address_netmask();
        foreach my $_addr ( @ips ) {
                $Log->comment("$_addr");
                $args{address} = $_addr;
                $args{netmask} = $netmask;
        }
    }

    $fqdn = delete($args{fqdn});

    # Execute Create
    try {
        NACL::STask::NetworkInterface->create(%args,
                                              _was_created       => \$was_created,
                                              nacltask_if_exists => $if_exists);
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = shift;
        if ($exception->text() =~ /Duplicate ip address|BCOMKA contains a duplicate entry/i) {
            # Delete the Address that didn't work
            my $old_address = $args{address};
            $args{address} = '';
            $args{retry} = $retry;
            $args{if_exists} = $if_exists;
            # Try to create again the lif, before executing a new ip address
            # will be requested.
            %args = $pkg->_get_lif_params(args        => \%args,
                                          old_address => $old_address,
                                          ipspace     => $ipspace);
            $dup_ip = 1;
        } else {
            # Rethrowing caught exception
            $Log->exit() if $may_exit;
            $exception->throw();
        }
    };

    if($dup_ip) {
        return $pkg->_nacl_create_call(%args);
    }

    # $iscsi will be set to 1 if the data-protocol on the lif is iscsi
    my $iscsi = 0;
    if (exists($args{'data-protocol'})) {
        foreach my $protocol (@{$args{'data-protocol'}}){
            if ($protocol =~ /iscsi/i){
                $iscsi = 1;
            }
        }
    }

    # if fqdn=1 and the data-protocol on the lif is iscsi, then set 'sendtargets-fqdn' in NACL::C::VserverIscsiInterface->modify
    if ($fqdn && $iscsi) {
        my $fqdn_ref = get_fqdn($args{address});
        if ( $fqdn_ref->[0]->{error} == '1') {
           NATE::BaseException->throw("FQDN does not exist for $args{address}");
        }
        try {
            NACL::C::VserverIscsiInterface->modify(command_interface => $args{command_interface},
                                                   vserver => $args{vserver},
                                                   lif => $args{lif},
                                                   'sendtargets-fqdn' => $fqdn_ref->[0]->{fqdn});
        } catch NACL::APISet::Exceptions::ResponseException with {
            # If already send targets fqdn is set, then do nothing.
            $Log->comment("sendtargets-fqdn is already set to ".$fqdn_ref->[0]->{fqdn});
        };
    }

    # if the iscsi lif was already created and fdqn=0, then set 'sendtargets-fqdn' => "". This is the reuse case.
    if (defined($fqdn) && !$fqdn && !$was_created && $iscsi) {
        try {
            NACL::C::VserverIscsiInterface->modify(command_interface => $args{command_interface},
                                                   vserver => $args{vserver},
                                                   lif => $args{lif},
                                                   'sendtargets-fqdn' => "\"\"");
        } catch NACL::APISet::Exceptions::ResponseException with {
            # If already send targets fqdn is set, then do nothing.
            $Log->comment("sendtargets-fqdn is already set to \"\" ");
        };
    }
    
    # Returning the ip that you are not going to use if the lif is already there.
    if(!$was_created && $use_ipm) {
        my $lif_state = NACL::CS::NetworkInterface->fetch(command_interface => $args{command_interface},
                                                            filter => { vserver => $args{vserver}, lif =>  $args{lif}},
                                                            requested_fields => ['address']);
        if (($lif_state) && ($lif_state->address() ne $args{address})) {
            my $number_ips_returned = Network::SCON_DB::IPReservation->ip_address_return( 'address' => $args{address} );
            if ($number_ips_returned) {
                $Log->comment("Ip address ". $args{address}." returned successfully");
            } else {
                NATE::BaseException->throw("Failed to return Ip address". $args{address});
            }
        }
    }

    if(!$was_created && $use_ipam) {
        my $lif_state = NACL::CS::NetworkInterface->fetch(command_interface => $args{command_interface},
                                                            filter => { vserver => $args{vserver}, lif =>  $args{lif}},
                                                            requested_fields => ['address']);
        if (($lif_state) && ($lif_state->address() ne $args{address})) {
            my $return_val = return_addresses([$args{address}]);
            if ($return_val) {
                $Log->comment("Ip address ". $args{address}." returned successfully");
            } else {
                NATE::BaseException->throw("Failed to return Ip address". $args{address});
            }
        }
    }
 
    if($was_created) {
        $Log->comment("Using newly Created Lif:" . $args{lif} . " in vserver " . $args{vserver});
    } else {
        $Log->comment("Using already Created Lif:" . $args{lif} . " in vserver " . $args{vserver});
    }

    $Log->exit() if $may_exit;
    # Returning either a list (hash) or scalar TRUE
    return (wantarray()) ? ('success'=>TRUE, 'duplicated'=>!$was_created, used_address=>$args{address}) : TRUE;


} ## end sub _nacl_create_call

#==============================================================================
# Method:     _nacl_delete_call
# Objective:  Makes the raw NACL calls for deletion of a lif.
# Details:    Makes the NACL::C::lif->delete() calls. Also works around
#             some common non-fatal errors that can be return from lif
#             deletion and then has a retry mechanism to retry lif deletion.
# Inputs:     NACL::C::lif->delete arguments
#             retry => integer, is the current retry the method is on.
#                    This is used to count how many times the _nacl_delete_call
#                    has been retried so that the retries can be limited.
# Outputs:    Returns whether the create was successful or a hash with success
#             if a hash is requested
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
#==============================================================================
sub _nacl_delete_call {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %args = @_;
    # Locals
    my $was_deleted = 0;
    $args{'method-timeout'} = 3600 unless exists($args{'method-timeout'});
    if (!exists($args{command_interface})) {
        my @Nodes = NACL::C::Node->find();
        $args{command_interface} = $Nodes[0];
    }

    my $use_ipm = delete($args{use_ipm});
    my $use_ipam = delete($args{use_ipam});

    my $lif;

    if ($use_ipm || $use_ipam) {
        $lif = NACL::CS::NetworkInterface->fetch( command_interface => $args{command_interface},
                                                    filter => {
                                                                'lif' => $args{lif},
                                                                'vserver' => $args{vserver},
                                                            },
                                                    allow_empty => "true",
                                                );
    }

    NACL::STask::NetworkInterface->purge(%args,
                                          nacltask_if_purged => 'pass',
                                          _was_deleted       => \$was_deleted);

    if ($use_ipm) {
        if($lif) {
            my $number_ips_returned = Network::SCON_DB::IPReservation->ip_address_return( 'address' => $lif->address() );
            if ($number_ips_returned) {
                $Log->comment("Ip address ". $lif->address()." returned successfully");
            } else {
                NATE::BaseException->throw("Failed to return Ip address". $lif->address());
            }
        }
    }

    if ($use_ipam) {
        if($lif) {
            my $return_val = return_addresses([$lif->address()]);
            if ($return_val) {
                $Log->comment("Ip address ". $lif->address()." returned successfully");
            } else {
                NATE::BaseException->throw("Failed to return Ip address". $lif->address());
            }
        }
    }


    if($was_deleted) {
        $Log->comment("Lif:" . $args{lif} . " in vserver " . $args{vserver} . " was deleted");
    } else {
        $Log->comment("Cannot cleanup  Lif:" . $args{lif} . " in vserver " . $args{vserver} .
                      ". It does not exist in the cluster.\n");
    }

    $Log->exit() if $may_exit;
    return TRUE;

} ## end sub _nacl_delete_call

#===============================================================================
# Method:     _get_lif_params
# Objective:  Add or Delete the needed arguments that NACL create call needs.
# Details:    Depending of the Data-Protocol that the lif has, will either do:
#             1) ISCSI - Will populate the 'address' parameter, by either
#                        retrieve the ip from the resource or just passthrough
#                        an existing one
#             2) FCP - Will delete the NETMASK and ADDRESS parameters that are
#                      not needed by NACL.
# Inputs:     args:  The resource ARGUMENTS that will be send to NACL call
#             respirces : The resource itself.
# Outputs:    Returns an update version of the arguments.
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
#==============================================================================
sub _get_lif_params {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = @_;
    # Locals
    my %args = %{$opts{args}};
    my $address_flag = TRUE;
    my $old_address = (defined $opts{old_address}) ? $opts{old_address} : '';
    my $ipspace = (defined $opts{ipspace}) ? $opts{ipspace} : 'Default';

    # User is not required to provide attribute 'data-protocol'. So first
    # verify if it even exists. If it does, verify it is an array reference as
    # requested by NACL.
    if (exists($args{'data-protocol'})){
        $Log->trace("Lif::_get_lif_params: DATA PROTOCOL defined.");
        if( !(ref($args{'data-protocol'}) =~ /ARRAY/i) ) {
            $Log->exit() if $may_exit;
            NATE::Exceptions::Argument->throw(
                               "Lif::_get_lif_params: DATA-PROTOCOL must be an".
                               " ARRAY reference."
                            );
        }
        # Since the DATA-PROTOCOLs attribute is an array, we need to find out if
        # FCP is one of the requested protocols. If it was requested we need to
        # delete the unecessary arguments that all other protocols need.
        foreach my $protocol (@{$args{'data-protocol'}}){
            if ($protocol =~ /fcp/i) {
                $address_flag = FALSE;
            }
        }
    }

    #if netmask-length exists, netmask isn't needed.
    if (exists($args{'netmask-length'})){
        delete($args{netmask});
    }

    if (exists($args{'use_ipam'}) && ($args{'use_ipam'})) {
        delete $args{address};
    }

    if (exists($args{'use_ipm'}) && ($args{'use_ipm'})) {
        delete $args{address};
    } elsif ($address_flag) {
        # If the protocol is not FCP, get an IP ADDRESS if needed.
        #Get the resource to work with.
        my $resource;
        if (exists($opts{resource})) {
            # Set the global $Resource variable
            $resource = $opts{resource};
            $Mem_Resource = $opts{resource};
        } else {
            if (defined($Mem_Resource)){
                $resource = $Mem_Resource;
            } else {
                $Log->exit() if $may_exit;
                NATE::Exceptions::Argument->throw(
                "Lif::_get_lif_params: Missing Resource reference to work with");
            }
        }

        # Verify that an IP address is needed
        if( defined($resource->{ips}) && $args{address} eq '' ){
            # If key IPS is present in the resource, means that user passed
            # an IP address range from where to chose the address. So, the
            # Resource needs to pass the array to the parent for him to
            # manage the addresses and avoid collisions.
            #
            # Request and IP Address
REPEAT:     $Parent->message_put(
                type    => "Response",
                message => {status    => "request_ip",
                            resources => $resource,
                            old_address => $old_address,
                            ipspace     => $ipspace
                           }
            );
            # GET THE MESSAGE
            my $ip_mssg = $Parent->message_get(type=>"ip_address");
            #IF THE MESSAGE IS AN ARRAY REF, THEN A RESOURCE WITH AN IP ADDRESS
            # WAS RETURNED
            if ( ref($ip_mssg) =~ /ARRAY/ ) {
                # Save the value in the resource that will be
                # returned to the ResourceList
                $resource = $ip_mssg->[0];
                $args{address} = $resource->{ADDRESS};
            } elsif ($ip_mssg =~ /wait/) {
                # If the message says wait then repeat because an ip is
                # in the used list that has not been verified as used
                # Set used_address to the empty string since we didn't
                # get back an ip
                $old_address = '';
                goto REPEAT;
            }
            # IF THE MESSAGE IS A SCALAR IS AN ERROR
            elsif ($ip_mssg =~ /ERROR: Could not find an IP address/) {
                NACL::AppBuilder::EnGenX::Exceptions::SubProcessFailure->throw(
                                                $ip_mssg,
                                                resource => $resource
                                            );
            }
        }#if ips
    }#if address
    else {
        # If this is a FCP lif, delete unecessary keys.
        delete $args{address};
        delete $args{netmask};
    }

    $Log->exit() if $may_exit;
    return (wantarray()) ? %args : \%args;
}# sub _get_lif_params

#==============================================================================
# Method:     _get_dtd_nacl_hashes (translate, ignore and delete)
# Objective:  Getter methods for the variables of the same name.
# Details:    Retrieves the hash or a reference to one of the following:
#             - DTD_To_NACL_Translate.
#             - DTD_To_NACL_Ignore.
#             - Delete_Args.
# Inputs:     none.
# Outputs:    The reference or hash variable.
# Exceptions:
#==============================================================================
sub _get_dtd_nacl_translate {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = @_;

    $Log->exit() if $may_exit;
    return (wantarray()) ? %DTD_To_NACL_Translate : \%DTD_To_NACL_Translate;
}
#==============================================================================
sub _get_dtd_nacl_ignore {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = @_;

    $Log->exit() if $may_exit;
    return (wantarray()) ? %DTD_To_NACL_Ignore : \%DTD_To_NACL_Ignore;
}
#==============================================================================
sub _get_dtd_nacl_delete {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = @_;

    $Log->exit() if $may_exit;
    return (wantarray()) ? %Delete_Args : \%Delete_Args;
}

1;
