# Copyright (c) 2001-2015 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Lun Controller Layer Module
## @author dl-san-automation-core
## @status shared
## @pod here

=head1 NAME

NACL::AppBuilder::EnGenX::Controller::Lun

=head1 DESCRIPTION

NACL::AppBuilder::EnGenX::Controller::Lun is a derived class of
NACL::AppBuilder::EnGenX::Controller::Task.  It is used to Create and
Delete ONTAP Luns. All methods are static.

=cut

package NACL::AppBuilder::EnGenX::Controller::Lun;

# Compiler directives
use strict;
use warnings;

# NATE and Standard Module import
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use NATE::Util::DataTypes qw(TRUE FALSE);
use Params::Validate
    qw(validate_with validate SCALAR ARRAYREF BOOLEAN UNDEF SCALARREF HASHREF);

# NACL and EnGenX Libs
use NACL::CS::Volume;
use NACL::STask::Lun;
use NACL::C::Job;
use NACL::AppBuilder::EnGenX::SubProcess;
use NACL::AppBuilder::EnGenX::EnGenXUtil;

# Exceptions
use NATE::BaseException qw(:try);
use NATE::Exceptions::Argument qw(:try);

#Inherit base class
use base qw( NACL::AppBuilder::EnGenX::Controller::Task );

use vars qw(%DTD_To_NACL_Translate
    %DTD_To_NACL_Ignore
    %Delete_Args);

# Arguments after converting from xml to a
# resource list that need their keys translated to
# keys that NACL methods will accept.
%DTD_To_NACL_Translate = (
    name   => 'lun',
    _reuse => 'if_exists'
);

# Arguments that are used by EnGenX but cannot be
# passed into the NACL methods
%DTD_To_NACL_Ignore = (
    parallel => '',
    reporting_nodes => '',
    report_all_nodes => '',
    report_local_nodes => '',
    report_dest_aggr => '',
    report_dest_vol => '',
    remove_reporting_nodes => '',
    remove_remote_nodes => '',
    map_list => '',
    require_same_lun_id_per_igroup => '',
    'lun-id' => '',
    lun_id_range => '',
);

# Arguments that can be passed into the NACL delete
# command
%Delete_Args = (
    command_interface => '',
    apiset_must       => '',
    apiset_should     => '',
    lun               => '',
    qtree             => '',
    volume            => '',
    vserver           => '',
    path              => '',
    force             => '',
);

=head1 METHODS

=head2 create

(Static method) Request creation of one or multiple luns serially
or in parallel and return whether the lun(s) was created successfully.
This method passes all user arguments through to NACL::AppBuilder::EnGenX::Controller::Task
and adds specific arguments used by the EnGenX framework.

  NACL::AppBuilder::EnGenX::Controller::Lun->create(
            command_interface => $Node,
            volume            => ( $Volume_name),
            lun                  => $Lun_name,
            vserver           => $Vserver_name,
            size              => "40M",
            ostype              => 'windows',
            'method-timeout'  => 3600,
            );

  NACL::AppBuilder::EnGenX::Controller::Lun->create(
            command_interface => $Node,
            volume            => ( $Volume_name),
            lun                  => $Lun_name,
            vserver           => $Vserver_name,
            size              => "40M",
            ostype              => 'windows',
            'method-timeout'  => 3600,
            count             => 5
            );

  NACL::AppBuilder::EnGenX::Controller::Lun->create(
            command_interfac  => $Node,
            resource_list      => @Luns,
            'method-timeout'  => 3600,
            );

  NACL::AppBuilder::EnGenX::Controller::Lun->create(
            command_interface => $Node,
            resource_list      => @Luns,
            );

  NACL::AppBuilder::EnGenX::Controller::Lun->create(
            resource_list      => @Luns,
            'method-timeout'  => 3600,
            );

  NACL::AppBuilder::EnGenX::Controller::Lun->create(
            resource_list      => @Luns,
            );

=over

=item Options

Either the resource_list is required or the NACL Lun Creation Arguments
are required but both should NOT be used in the same call.

=over

=item C<< resource_list => @resource_list >>

(Optional)
If this is passed in the luns in the resource list will be created
based on the information that is supplied in the resource list. If
the resource list is passed in the NACL Lun Creation Arguments should
NOT be passed in.

=item C<< NACL Lun Creation Arguments >>

(Optional)
These are arguments that can be passed into NACL::STask::Lun->Create.
If these arguments are passed in the resource_list argument should NOT
be passed in

=item C<< count => $lun_count >>

(Optional)
This arguments is used to create multiple luns from the same NACL Lun
Creation Arguments.  To create the names for the luns, the lun name
passed in is appended with the index that it was create at starting at 1.

=item C<< command_interface => $NACL_command_interface >>

(Optional)
This is the command_interface that will be used during NACL calls.  If a
command_interface is not passed in then one will be created.  Passing in a
command_interface helps performance because for every command_interface created
the version of the Filer must be determined by creating a connection to the filer
and then sending commands to get the version.

=item C<< method_timeout => $NACL_method_timeout >>

(Optional)
This is the method_timeout that will be used during NACL calls. If a
method_timeout is not passed in then the method_timeout will be defaulted.

=back

=item Return

This method returns a True or False based on whether the method succeeds or fails
in the creation of the Lun(s) based on the Options passed in.

=item Exceptions

This method produces and throws no exceptions of its own but pass through any
uncaught exceptions from other library calls.

=cut

sub create {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %args = @_;

    # Locals
    my $parent_answer;

    # Globals that will also be set as globals in the base class
    $args{'Name'} = "lun";

    # Super::_create is in Task.pm
    $parent_answer = $pkg->SUPER::_create(%args);

    return $parent_answer;
} ## end sub create

=head2 delete

(Static method) Request deletion of one or multiple luns serially
or in parallel and return whether the lun(s) was deleted successfully.
This method passes all user arguments through to NACL::AppBuilder::EnGenX::Controller::Task
and adds specific arguments used by the EnGenX framework.

  NACL::AppBuilder::EnGenX::Controller::Lun->delete(
            command_interface => $Node,
            volume            => ( $Volume_name),
            lun                  => $Lun_name,
            vserver           => $Vserver_name,
            'method-timeout'  => 3600,
            );

  NACL::AppBuilder::EnGenX::Controller::Lun->delete(
            command_interface => $Node,
            volume            => ( $Volume_name),
            lun                  => $Lun_name,
            vserver           => $Vserver_name,
            count             => 5
            );

  NACL::AppBuilder::EnGenX::Controller::Lun->delete(
            command_interface => $Node,
            resource_list      => @Luns,
            'method-timeout'  => 3600,
            );

  NACL::AppBuilder::EnGenX::Controller::Lun->delete(
            command_interface => $Node,
            resource_list      => @Luns,
            );

  NACL::AppBuilder::EnGenX::Controller::Lun->delete(
            resource_list      => @Luns,
            'method-timeout'  => 3600,
            );

  NACL::AppBuilder::EnGenX::Controller::Lun->delete(
            resource_list      => @Luns,
            );

=over

=item Options

Either the resource_list is required or the NACL Lun Deletion Arguments
are required but both should NOT be used in the same call.

=over

=item C<< resource_list => @resource_list >>

(Required)
If this is passed in the luns in the resource list will be deleted
based on the information that is supplied in the resource list. If
the resource list is passed in the NACL Lun Deletion Arguments should
NOT be passed in.

=item C<< NACL Lun Deletion Arguments >>

(Required)
These are arguments that can be passed into NACL::STask::Lun->purge.
If these arguments are passed in the resource_list argument should NOT
be passed in.

=item C<< count => $lun_count >>

(Optional)
This arguments is used to delete multiple luns from the same NACL Lun
Deletion Arguments. To delete the names for the luns, the lun name
passed in is appended with the index that it was deleted at starting at 1.

=item C<< command_interface => $NACL_command_interface >>

(Optional)
This is the command_interface that will be used during NACL calls.  If a
command_interface is not passed in then one will be created.  Passing in a
command_interface helps performance because for every command_interface created
the version of the Filer must be determined by creating a connection to the filer
and then sending commands to get the version.

=item C<< method_timeout => $NACL_method_timeout >>

(Optional)
This is the method_timeout that will be used during NACL calls. If a
method_timeout is not passed in then the method_timeout will be defaulted.

=back

=item Return

This method returns a True or False based on whether the method succeeds or fails
in the deletion of the Lun(s) based on the Options passed in.

=item Exceptions

This method produces and throws no exceptions of its own but pass through any
uncaught exceptions from other library calls.

=cut

sub delete {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %args = @_;

    # Locals
    my $parent_answer;

    # Globals that will also be set as globals in the base class
    $args{'Name'} = "lun";

    # Super::_delete is in Task.pm
    $parent_answer = $pkg->SUPER::_delete(%args);

    return $parent_answer;
} ## end sub delete


#==============================================================================
# Method:     _nacl_create_call
# Objective:  Makes the raw NACL calls for creation of a lun.
# Details:    Makes the NACL::STask::Lun->create() call. Also works
#             around some common non-fatal errors that can be return from
#             lun creation.
# Inputs:     NACL::STask::Lun->create arguments
# Outputs:    Returns whether the create was successful or a hash with success
#             and whether the lun was a duplicate if a hash is requested
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
#==============================================================================
sub _nacl_create_call {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %args = @_;
    # Locals
    my $was_created = 0;
    my $path = "";
    my $state = exists($args{'state'}) ? delete($args{'state'}) : 'online';
    my $allocation
        = exists($args{'space-allocation'}) ? $args{'space-allocation'} : 'disabled';
    my $if_exists
        = exists($args{if_exists}) ? delete($args{if_exists}) : 'die';
    $args{'method-timeout'} = 3600 unless exists($args{'method-timeout'});

    if (!exists($args{command_interface})) {
        my @Nodes = NACL::C::Node->find();
        $args{command_interface} = $Nodes[0];
    }

    # If create from file then create the file-path argument.
    if($args{create_from_file}) {
        if(defined($args{qtree}) and length($args{qtree})){
            $args{'file-path'} = "/vol/$args{volume}/$args{qtree}/file_$args{lun}";
        } else {
            $args{'file-path'} = "/vol/$args{volume}/file_$args{lun}";
        }
    }

    delete $args{create_from_file};

    # Check if size arguments have been defined and calculate percentage if necessary
    $pkg->_compute_size(\%args);

    # Put quotes around the lun  comment, if it exists.
    if (exists $args{comment}) {
        $args{comment} = "\"" . $args{comment} . "\"";
    }

    # Execute Create
    my $lun = NACL::STask::Lun->create(%args,
                             nacltask_if_exists => $if_exists,
                             _was_created => \$was_created);
    $path = $lun->path();

    if($was_created) {
        $Log->comment("Using newly Created Lun:" . $path . " " .
                      "in vserver:" . $args{vserver} );
    } else {
        $Log->comment("Using already Created Lun:" . $path . " " .
                      "in vserver:" . $args{vserver} );
    }

    if (!$was_created) {
        my %extra = ();

        # Modification of the allocation setting or the lun state must be sent seperately (burt#783657)
        # We need to build a path to turn on allocation
        $extra{'space-allocation'} = $allocation;

        # if a qos policy assignment is to be set, apply it here
        if (exists($args{'qos-policy-group'})) {
            $extra{'qos-policy-group'} = delete($args{'qos-policy-group'});
        }

        NACL::STask::Lun->modify(command_interface => $args{command_interface},
                                 vserver => $args{vserver},
                                 path => $path,
                                 %extra);
    } 
    if (!$was_created or ($state eq 'offline')) {
        NACL::STask::Lun->modify(command_interface => $args{command_interface},
                                 vserver => $args{vserver},
                                 path => $path,
                                 'state' => $state,);
    }

    $Log->exit() if $may_exit;
    # Returning either a list (hash) or scalar TRUE
    return (wantarray()) ? ('success'=>TRUE, 'duplicated'=>!$was_created) : TRUE;


} ## end sub _nacl_create_call

#========================================================================================
# Method:     _nacl_delete_call
# Objective:  Makes the raw NACL calls for deletion of a Lun.
# Details:    Makes the NACL::STask::Lun->purge() calls. Also works around
#             some common non-fatal errors that can be return from Lun
#             deletion.
# Inputs:     NACL::STask::Lun->purge arguments
# Outputs:    Returns whether the create was successful or a hash with success
#             if a hash is requested
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
#==============================================================================
sub _nacl_delete_call {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %args = @_;

    # Locals
    my $was_deleted = 0;
    $args{'method-timeout'} = 3600 unless exists($args{'method-timeout'});
    if (!exists($args{command_interface})) {
        my @Nodes = NACL::C::Node->find();
        $args{command_interface} = $Nodes[0];
    }

    NACL::STask::Lun->purge(%args,
                            nacltask_if_purged => 'pass',
                            _was_deleted => \$was_deleted);

    # Build up the lun path.
    my $path = "";
    if (!defined($args{qtree})) {
        $path = "/vol/$args{volume}/$args{lun}";
    } else {
        $path = "/vol/$args{volume}/$args{qtree}/$args{lun}";
    }

    if($was_deleted) {
        $Log->comment("Lun:" . $path . " " .
                      "in vserver:" . $args{vserver} . " " .
                      "was deleted.\n");
    } else {
        $Log->comment("Cannot cleanup Lun:" . $path . " " .
                      "in vserver:" . $args{vserver} . ". " .
                      "It does not exist in the cluster.\n");
    }

    $Log->exit() if $may_exit;
    return TRUE;

} ## end sub _nacl_delete_call
#==============================================================================
# Method:     _nacl_wait_call
# Objective:  Makes a NACL call to 'wait_for_creation.
# Details:    Makes the call to NACL::STask::Lun->wait_for_creation();
#             that will wait for the node to finish creating the aggr.
# Inputs:     NACL::STask::Lun->wait arguments
# Outputs:    Returns whether the create was successful TRUE or not FALSE
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
# TODO: Implement this method when needed
#==============================================================================
sub _nacl_wait_call {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = @_;

    $Log->exit() if $may_exit;
    return TRUE;
}

#==============================================================================
# Method:     _get_dtd_nacl_hashes (translate, ignore and delete)
# Objective:  Getter methods for the variables of the same name.
# Details:    Retrieves the hash or a reference to one of the following:
#             - DTD_To_NACL_Translate.
#             - DTD_To_NACL_Ignore.
#             - Delete_Args.
# Inputs:     none.
# Outputs:    The reference or hash variable.
# Exceptions:
#==============================================================================
sub _get_dtd_nacl_translate {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = @_;

    $Log->exit() if $may_exit;
    return (wantarray()) ? %DTD_To_NACL_Translate : \%DTD_To_NACL_Translate;
}
#==============================================================================
sub _get_dtd_nacl_ignore {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = @_;

    $Log->exit() if $may_exit;
    return (wantarray()) ? %DTD_To_NACL_Ignore : \%DTD_To_NACL_Ignore;
}
#==============================================================================
sub _get_dtd_nacl_delete {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = @_;

    $Log->exit() if $may_exit;
    return (wantarray()) ? %Delete_Args : \%Delete_Args;
}
#==============================================================================
# Method:     _compute_size (args,size_param)
# Objective:  Compute/Process the size parameter if it exists.
# Details:    Retrieves the hash or a reference to one of the following:
# Inputs:     $args, $size_param`
# Outputs:    none
# Exceptions:
#==============================================================================
sub _compute_size {
    $Log->enter() if $may_enter;

    my ($pkg, $args) = @_;
    my @size_params = ('size','prefix-size', 'extent-size', 'suffix-size');
    my $modulus;
    my $volume_cs_obj;
    my $percent;

    foreach my $size_param (@size_params) {
        # If size paramter is anything other than 'size' make sure calculated percent is a multiple of 512.
        # In this case, the size is rounded down to to the nearest multiple of 512.
        if (exists $args->{$size_param} and $args->{$size_param} =~ s/%//) {
            if($volume_cs_obj eq undef) {
                 $volume_cs_obj = NACL::CS::Volume->fetch(command_interface => $args->{command_interface},
                                                          requested_fields => ['size'],
                                                          filter => { vserver => $args->{vserver},
                                                                      volume  => $args->{volume}});
             }

             $percent = $args->{$size_param};
             $args->{$size_param} = int(($args->{$size_param}/100)*($volume_cs_obj->size()));

             if($size_param ne 'size'){
                 # Remove Decimal
                 $args->{$size_param} =~ s/\.\d+$//;

                 # Make sure the value is a multiple of 512
                 $modulus = $args->{$size_param} % 512;
                 $args->{$size_param} -= $modulus;
             }

             $Log->comment("Setting Lun " . $size_param . " to " . $args->{$size_param} . "b which is " .
                           "$percent\% of volume " . $args->{volume} . " in vserver " .
                           $args->{vserver} . " which is size " . $volume_cs_obj->size() . "b.");
        }
    }

    $Log->exit() if $may_exit;
}

1;
