# Copyright (c) 2012-2015 NetApp Inc.
# All rights reserved.
#
## @summary Provide a common set of libraries to work on controllers
## @author dl-san-automation
## @status private
## @pod here

=head1 NAME

NACL::AppBuilder::EnGenX::Controller::NFS

=head1 DESCRIPTION

NACL::AppBuilder::EnGenX::Controller::NFS is used to Config, Setup, Cleanup and
Deconfig NFS services on a vserver.

=cut

package NACL::AppBuilder::EnGenX::Controller::NFS;

# Compiler directives
use strict;
use warnings;

# NATE and Standard Module import
use NATE::Log qw(log_global);
use NATE::Util::DataTypes qw(TRUE FALSE);
use Data::Dumper;
use Params::Validate
    qw(validate_with validate SCALAR ARRAYREF BOOLEAN UNDEF SCALARREF HASHREF);

# NACL libs
use NACL::STask::VserverNfs;

# Inherit base class
use base qw( NACL::AppBuilder::EnGenX::Controller::Task );

use vars qw(%DTD_To_NACL_Translate
            %DTD_To_NACL_Ignore
            %Delete_Args
            $Log);

$Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();

# Arguments after converting from xml to a
# resource list that need their keys translated to
# keys that NACL methods will accept.
%DTD_To_NACL_Translate = (
    _reuse => 'if_exists',
    'v4_0' => 'v4.0',
    'v4_0-acl' => 'v4.0-acl',
    'v4_0-read-delegation' => 'v4.0-read-delegation',
    'v4_0-write-delegation' => 'v4.0-write-delegation',
    'v4_1' => 'v4.1',
    'v4_1-read-delegation' => 'v4.1-read-delegation',
    'v4_1-write-delegation' => 'v4.1-write-delegation',
    'v4_1-pnfs' => 'v4.1-pnfs',
    'v4_1-acl' => 'v4.1-acl',
    'v4_1-migration' => 'v4.1-migration',
    'v4_0-migration' => 'v4.0-migration',
    'v4_1-referrals' => 'v4.1-referrals'
);

# Arguments that are used by EnGenX but cannot be
# passed into the NACL methods
%DTD_To_NACL_Ignore = (
    name => '',
    parallel      => '',
    service_enabled => ''
);

# Arguments that can be passed into the NACL delete
# command
%Delete_Args = (
    vserver => '',
);

=head1 METHODS

=head2 create

(Static method) Request to create a nfs service with the following options

  NACL::AppBuilder::EnGenX::Controller::NFS->create(
            command_interface               => $node,
            vserver                         => $vserver_name,
            %nacl_c_vserver_create_args
            );

=over

=item Options

=over

=item C<< command_interface => $node >>

(Required)
This is the command_interface that will be used during NACL calls.

=item C<< vserver => $vserver >>

(Required)
The vserver to create the nfs service for.

=back

=item Return

This method returns a True or False based on whether the method succeeds or fails
in the creation of the NFS services based on the Options passed in.

=item Exceptions

This method produces and throws no exceptions of its own but pass through any
uncaught exceptions from other library calls.

=cut
sub create {
    $Log->enter() if $may_enter;
    my ($pkg, %args) = @_;

    # Locals
    my $parent_answer;

    # Globals that will also be set as globals in the base class
    $args{'Name'} = "nfs";

    # Super::_create is in Task.pm
    $parent_answer = $pkg->SUPER::_create(%args);

    $Log->exit() if $may_exit;

    return $parent_answer;
}

=head2 delete

(Static method) Request to delete a nfs service with the following options

  NACL::AppBuilder::EnGenX::Controller::NFS->delete(
            command_interface               => $node,
            vserver                         => $vserver_name,
            );

=over

=item Options

=over

=item C<< command_interface => $node >>

(Required)
This is the command_interface that will be used during NACL calls.

=item C<< vserver => $vserver >>

(Required)
The vserver to delete the nfs service for.

=item Return

This method returns a True or False based on whether the method succeeds or fails
in the deletion of the NFS Service based on the options passed in.

=item Exceptions

This method produces and throws no exceptions of its own but pass through any
uncaught exceptions from other library calls.

=cut
sub delete {
    $Log->enter() if $may_enter;
    my ($pkg, %args) = @_;

    # Locals
    my $parent_answer;

    # Globals that will also be set as globals in the base class
    $args{'Name'} = "nfs";

    # Super::_create is in Task.pm
    $parent_answer = $pkg->SUPER::_delete(%args);

    $Log->exit() if $may_exit;

    return $parent_answer;
}

#==============================================================================
# Method:     _nacl_create_call
# Objective:  Makes the raw NACL calls for creation of the nfs service.
# Details:    Makes the NACL::STask::VserverNfs->create().
#
# Inputs:     NACL::STask::VserverNfs->create arguments
# Outputs:    Returns whether the create was successful
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
#==============================================================================
sub _nacl_create_call {
    $Log->enter() if $may_enter;
    my ($pkg, %args) = @_;

    # Locals
    my $was_created = 0;
    my $if_exists
        = exists($args{if_exists}) ? delete($args{if_exists}) : 'die';

    # create the NFS service
    NACL::STask::VserverNfs->create(%args,
                                    nacltask_if_exists => $if_exists,
                                    _was_created       => \$was_created);

    if($was_created) {
        $Log->comment("Using newly Created NFS: in vserver:" . $args{vserver} );
    } else {
        $Log->comment("Using already Created NFS: in vserver:" . $args{vserver} );
    }

    if(!$was_created) {
        NACL::STask::VserverNfs->modify(%args,
                                        nacltask_if_no_values => 'pass');
    }

    $Log->exit() if $may_exit;
    return (wantarray()) ? ('success' => TRUE, 'duplicated' => !$was_created) : TRUE;
}

#==============================================================================
# Method:     _nacl_delete_call
# Objective:  Makes the raw NACL calls for deleting the nfs service
# Details:    Makes the NACL::STask::VserverNfs->purge().
#
# Inputs:     NACL::STask::VserverNfs->purge arguments
# Outputs:    Returns whether the purge was successful
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
#==============================================================================
sub _nacl_delete_call {
    $Log->enter() if $may_enter;
    my ($pkg, %args) = @_;

    my $was_deleted = 0;

    # delete the nfs service
    NACL::STask::VserverNfs->purge(%args,
                                   nacltask_if_purged => 'pass',
                                   _was_deleted       => \$was_deleted);

    if($was_deleted) {
        $Log->comment("NFS: in vserver:" . $args{vserver} . " " .
                      "was deleted.\n");
    } else {
        $Log->comment("Cannot cleanup NFS: in vserver:" . $args{vserver} . ". " .
                      "It does not exist in the cluster.\n");
    }

    $Log->exit() if $may_exit;
    return;
}

#==============================================================================
# Method:     _nacl_wait_call
# Objective:  NOOP for NFS
# Details:
# Inputs:     None
# Outputs:    None
# Exceptions: None
#==============================================================================
sub _nacl_wait_call {
    $Log->enter() if $may_enter;
    my ($pkg, %args) = @_;

    $Log->exit() if $may_exit;
    return TRUE;
}

#==============================================================================
# Method:     _get_dtd_nacl_hashes (translate, ignore and delete)
# Objective:  Getter methods for the variables of the same name.
# Details:    Retrieves the hash or a reference to one of the following:
#             - DTD_To_NACL_Translate.
#             - DTD_To_NACL_Ignore.
#             - Delete_Args.
# Inputs:     none.
# Outputs:    The reference or hash variable.
# Exceptions:
#==============================================================================
sub _get_dtd_nacl_translate {
    $Log->enter() if $may_enter;
    my ($pkg, %args) = @_;

    $Log->exit() if $may_exit;
    return (wantarray()) ? %DTD_To_NACL_Translate : \%DTD_To_NACL_Translate;
}
#==============================================================================
sub _get_dtd_nacl_ignore {
    $Log->enter() if $may_enter;
    my ($pkg, %args) = @_;

    $Log->exit() if $may_exit;
    return (wantarray()) ? %DTD_To_NACL_Ignore : \%DTD_To_NACL_Ignore;
}
#==============================================================================
sub _get_dtd_nacl_delete {
    $Log->enter() if $may_enter;
    my ($pkg, %args) = @_;

    $Log->exit() if $may_exit;
    return (wantarray()) ? %Delete_Args : \%Delete_Args;
}

1;
