# Copyright (c) 2012-2015 NetApp Inc.
# All rights reserved.
#
## @summary Provide a common set of libraries to work on controllers
## @author dl-san-automation
## @status private
## @pod here

=head1 NAME

NACL::AppBuilder::EnGenX::Controller::NIS

=head1 DESCRIPTION

NACL::AppBuilder::EnGenX::Controller::NIS is used to Config, Setup, Cleanup and
Deconfig NIS services on a vserver.

=cut

package NACL::AppBuilder::EnGenX::Controller::NIS;

# Compiler directives
use strict;
use warnings;

# NATE and Standard Module import
use NATE::Log qw(log_global);
use NATE::Util::DataTypes qw(TRUE FALSE);
use Params::Validate
    qw(validate_with validate SCALAR ARRAYREF BOOLEAN UNDEF SCALARREF HASHREF);

# NACL libs
use NACL::STask::VserverServicesNisDomain;

# Inherit base class
use base qw( NACL::AppBuilder::EnGenX::Controller::Task );

use vars qw(%DTD_To_NACL_Translate
            %DTD_To_NACL_Ignore
            %Delete_Args
            $Log);

$Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();

# Arguments after converting from xml to a
# resource list that need their keys translated to
# keys that NACL methods will accept.
%DTD_To_NACL_Translate = (
    _reuse => 'if_exists',
    nis_servers => 'servers'
);

# Arguments that are used by EnGenX but cannot be
# passed into the NACL methods
%DTD_To_NACL_Ignore = (
    name => '',
    parallel      => '',
    service_enabled => '',
);

# Arguments that can be passed into the NACL delete
# command
%Delete_Args = (
    vserver => '',
    domain => '',
);

=head1 METHODS

=head2 create

(Static method) Request to create a nis service with the following options

  NACL::AppBuilder::EnGenX::Controller::NIS->create(
            command_interface               => $node,
            vserver                         => $vserver_name,
            domain                          => $domain,
            active                          => $active,
            'nis_servers'                   => $servers,
            );

=over

=item Options

=over

=item C<< command_interface => $node >>

(Required)
This is the command_interface that will be used during NACL calls.  If a
command_interface is not passed in then one will be created.  Passing in a
command_interface helps performance because for every command_interface created
the version of the Filer must be determined by creating a connection to the filer
and then sending commands to get the version.

=item C<< vserver => $vserver >>

(Required)
The vserver to create the nis service for.

=item C<< domain => $domain >>

(Required)
This is the nis domain to use for this NIS service.

=item C<< active => $active >>

(Required)
This specifies if the nis server is active.

=item C<< nis_servers => $servers >>

(Required)
This is the server list for the NIS service.

=back

=item Return

This method returns a True or False based on whether the method succeeds or fails
in the creation of the NIS Targets based on the Options passed in.

=item Exceptions

This method produces and throws no exceptions of its own but pass through any
uncaught exceptions from other library calls.

=cut
sub create {
    $Log->enter() if $may_enter;
    my ($pkg, %args) = @_;

    # Locals
    my $parent_answer;

    # Globals that will also be set as globals in the base class
    $args{'Name'} = "nis";

    # Super::_create is in Task.pm
    $parent_answer = $pkg->SUPER::_create(%args);

    $Log->exit() if $may_exit;

    return $parent_answer;
}

=head2 delete

(Static method) Request to delete a nis service with the following options

  NACL::AppBuilder::EnGenX::Controller::NIS->delete(
            command_interface               => $node,
            vserver                         => $vserver_name,
            domain                          => $domain,
            );

=over

=item Options

=over

=item C<< command_interface => $node >>

(Required)
This is the command_interface that will be used during NACL calls.  If a
command_interface is not passed in then one will be created.  Passing in a
command_interface helps performance because for every command_interface created
the version of the Filer must be determined by creating a connection to the filer
and then sending commands to get the version.

=item C<< vserver => $vserver >>

(Required)
The vserver to delete the nis service for.

=item C<< domain => $domain >>

(Required)
This is the nis domain.

=back

=item Return

This method returns a True or False based on whether the method succeeds or fails
in the delete of the NIS Targets based on the Options passed in.

=item Exceptions

This method produces and throws no exceptions of its own but pass through any
uncaught exceptions from other library calls.

=cut
sub delete {
    $Log->enter() if $may_enter;
    my ($pkg, %args) = @_;

    # Locals
    my $parent_answer;

    # Globals that will also be set as globals in the base class
    $args{'Name'} = "nis";

    # Super::_create is in Task.pm
    $parent_answer = $pkg->SUPER::_delete(%args);

    $Log->exit() if $may_exit;

    return $parent_answer;
}

#==============================================================================
# Method:     _nacl_create_call
# Objective:  Makes the raw NACL calls for creation of the nis service.
# Details:    Makes the NACL::STask::VserverServicesNisDomain->create().
#
# Inputs:     NACL::STask::VserverServicesNisDomain->create arguments
# Outputs:    Returns whether the create was successful
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
#==============================================================================
sub _nacl_create_call {
    $Log->enter() if $may_enter;
    my ($pkg, %args) = @_;

    # Locals
    my $was_created = 0;
    $args{nacltask_if_exists} = exists($args{if_exists}) ? delete($args{if_exists}) : 'die';
    $args{'servers'} = [split(/,/, $args{'servers'})] if (exists $args{'servers'});
    $args{'method-timeout'} = 3600 unless exists($args{'method-timeout'});

    if (!exists($args{command_interface})) {
        my @Nodes = NACL::C::Node->find();
        $args{command_interface} = $Nodes[0];
    }

    # create the NIS service
    NACL::STask::VserverServicesNisDomain->create(%args,
                                                  _was_created => \$was_created);

    if($was_created) {
        $Log->comment("Using newly added NIS Service on domain: " . $args{domain}." on vserver " .$args{vserver});
    } else {
        $Log->comment("Using already created NIS Service on domain: " . $args{domain}." on vserver " .$args{vserver});
    }

    $Log->exit() if $may_exit;
    return (wantarray()) ? ('success' => TRUE, 'duplicated' => !$was_created) : TRUE;
}

#==============================================================================
# Method:     _nacl_delete_call
# Objective:  Makes the raw NACL calls for deleting the cifs service
# Details:    Makes the NACL::STask::VserverServicesNisDomain->purge().
#
# Inputs:     NACL::STask::VserverServicesNisDomain->purge arguments
# Outputs:    Returns whether the delete was successful
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
#==============================================================================
sub _nacl_delete_call {
    $Log->enter() if $may_enter;
    my ($pkg, %args) = @_;

    # Locals
    my $was_deleted = 0;
    $args{'method-timeout'} = 3600 unless exists($args{'method-timeout'});

    # delete the nis service
    NACL::STask::VserverServicesNisDomain->purge(%args,
                                                 nacltask_if_purged => 'pass',
                                                 _was_deleted => \$was_deleted);

    if($was_deleted) {
        $Log->comment("NIS service for domain: " . $args{domain} . " was removed");
    } else {
        $Log->comment("Cannot cleanup NIS service for domain: " . $args{domain} .
                      ". It does not exist in the cluster.\n");
    }

    $Log->exit() if $may_exit;
    return;
}

#==============================================================================
# Method:     _nacl_wait_call
# Objective:  NOOP for NIS
# Details:
# Inputs:     None
# Outputs:    None
# Exceptions: None
#==============================================================================
sub _nacl_wait_call {
    $Log->enter() if $may_enter;
    my ($pkg, %args) = @_;

    $Log->exit() if $may_exit;
    return TRUE;
}

#==============================================================================
# Method:     _get_dtd_nacl_hashes (translate, ignore and delete)
# Objective:  Getter methods for the variables of the same name.
# Details:    Retrieves the hash or a reference to one of the following:
#             - DTD_To_NACL_Translate.
#             - DTD_To_NACL_Ignore.
#             - Delete_Args.
# Inputs:     none.
# Outputs:    The reference or hash variable.
# Exceptions:
#==============================================================================
sub _get_dtd_nacl_translate {
    $Log->enter() if $may_enter;
    my ($pkg, %args) = @_;

    $Log->exit() if $may_exit;
    return (wantarray()) ? %DTD_To_NACL_Translate : \%DTD_To_NACL_Translate;
}
#==============================================================================
sub _get_dtd_nacl_ignore {
    $Log->enter() if $may_enter;
    my ($pkg, %args) = @_;

    $Log->exit() if $may_exit;
    return (wantarray()) ? %DTD_To_NACL_Ignore : \%DTD_To_NACL_Ignore;
}
#==============================================================================
sub _get_dtd_nacl_delete {
    $Log->enter() if $may_enter;
    my ($pkg, %args) = @_;

    $Log->exit() if $may_exit;
    return (wantarray()) ? %Delete_Args : \%Delete_Args;
}

1;
