
#
# Copyright (c) 2001-2012 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Portset Controller Layer Module
## @author dl-san-automation-core
## @status shared
## @pod here

=head1 NAME

NACL::AppBuilder::EnGenX::Controller::Portset

=head1 DESCRIPTION

NACL::AppBuilder::EnGenX::Controller::Portset is a derived class of
NACL::AppBuilder::EnGenX::Controller::Task.  It is used to Create and
Delete ONTAP Portsets. All methods are static.

=cut

package NACL::AppBuilder::EnGenX::Controller::Portset;

# Compiler directives
use strict;
use warnings;

# NATE and Standard Module import
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use NATE::Util::DataTypes qw(TRUE FALSE);
use Params::Validate
    qw(validate_with validate SCALAR ARRAYREF BOOLEAN UNDEF SCALARREF HASHREF);

# NACL and EnGenX Libs
use NACL::STask::LunPortset;
use NACL::STask::LunIgroup;
use NACL::AppBuilder::EnGenX::SubProcess;
use NACL::AppBuilder::EnGenX::EnGenXUtil;
use NACL::C::Exceptions::LunIgroup::AlreadyBound qw(:try);
use NACL::CS::LunPortset;

# Exceptions
use NATE::BaseException qw(:try);

# Inherit base class
use base qw( NACL::AppBuilder::EnGenX::Controller::Task );

use vars qw(%DTD_To_NACL_Translate
    %DTD_To_NACL_Ignore
    %Delete_Args
);

# Arguments after converting from xml to a
# resource list that need their keys translated to
# keys that NACL methods will accept.
%DTD_To_NACL_Translate = (
    name   => 'portset',
    _reuse => 'if_exists'
);

# Arguments that are used by EnGenX but cannot be
# passed into the NACL methods
%DTD_To_NACL_Ignore = (parallel => '',);

# Arguments that can be passed into the NACL delete
# command
%Delete_Args = (
    command_interface => '',
    apiset_must       => '',
    apiset_should     => '',
    portset           => '',
    vserver           => '',
    igroups           => '',
    force             => '',
);

=head1 METHODS

=head2 create

(Static method) Request creation of one or multiple portsets serially
or in parallel and return whether the portset(s) was created successfully.
This method passes all user arguments through to NACL::AppBuilder::EnGenX::Controller::Task
and adds specific arguments used by the EnGenX framework.

  NACL::AppBuilder::EnGenX::Controller::Portset->create(
            command_interface => $Node,
            vserver           => $Vserver_name,
            portset           => $portset,
            protocol          => "iscsi",
            'method-timeout'  => 3600,
            );

  NACL::AppBuilder::EnGenX::Controller::Portset->create(
            command_interface => $Node,
            vserver           => $Vserver_name,
            portset           => $portset,
            protocol          => "iscsi",
            'method-timeout'  => 3600,
            count             => 5
            );

  NACL::AppBuilder::EnGenX::Controller::Portset->create(
            command_interfac => $Node,
            resource_list    => @Portsets,
            'method-timeout' => 3600,
            );

  NACL::AppBuilder::EnGenX::Controller::Portset->create(
            command_interface => $Node,
            resource_list      => @Portsets,
            );

  NACL::AppBuilder::EnGenX::Controller::Portset->create(
            resource_list      => @Portsets,
            'method-timeout'  => 3600,
            );

  NACL::AppBuilder::EnGenX::Controller::Portset->create(
            resource_list      => @Portsets,
            );

=over

=item Options

Either the resource_list is required or the NACL Portset Creation Arguments
are required but both should NOT be used in the same call.

=over

=item C<< resource_list => @resource_list >>

(Optional)
If this is passed in the portsets in the resource list will be created
based on the information that is supplied in the resource list. If
the resource list is passed in the NACL Protset Creation Arguments should
NOT be passed in.

=item C<< NACL Portset Creation Arguments >>

(Optional)
These are arguments that can be passed into NACL::C::LunPortset->Create.
If these arguments are passed in the resource_list argument should NOT
be passed in

=item C<< count => $portset_count >>

(Optional)
This arguments is used to create multiple Protsets from the same NACL Protset
Creation Arguments.  To create the names for the Protsets, the Protset name
passed in is appended with the index that it was create at starting at 1.

=item C<< command_interface => $NACL_command_interface >>

(Optional)
This is the command_interface that will be used during NACL calls.  If a
command_interface is not passed in then one will be created.  Passing in a
command_interface helps performance because for every command_interface created
the version of the Filer must be determined by creating a connection to the filer
and then sending commands to get the version.

=item C<< method_timeout => $NACL_method_timeout >>

(Optional)
This is the method_timeout that will be used during NACL calls. If a
method_timeout is not passed in then the method_timeout will be defaulted.

=back

=item Return

This method returns a True or False based on whether the method succeeds or fails
in the creation of the Protset(s) based on the Options passed in.

=item Exceptions

This method produces and throws no exceptions of its own but pass through any
uncaught exceptions from other library calls.

=cut

sub create {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %args = @_;

    # Locals
    my $parent_answer;

    # Globals that will also be set as globals in the base class
    $args{'Name'} = "portset";

    # Super::_create is in Task.pm
    $parent_answer = $pkg->SUPER::_create(%args);

    return $parent_answer;
} ## end sub create

=head2 delete

(Static method) Request deletion of one or multiple Protsets serially
or in parallel and return whether the Protset(s) was deleted successfully.
This method passes all user arguments through to NACL::AppBuilder::EnGenX::Controller::Task
and adds specific arguments used by the EnGenX framework.

  NACL::AppBuilder::EnGenX::Controller::Portset->delete(
            command_interface => $Node,
            vserver           => $Vserver_name,
            portset           => $portset,
            'method-timeout'  => 3600,
            );

  NACL::AppBuilder::EnGenX::Controller::Portset->delete(
            command_interface => $Node,
            vserver           => $Vserver_name,
            portset           => $portset,
            'method-timeout'  => 3600,
            count             => 5
            );

  NACL::AppBuilder::EnGenX::Controller::Portset->delete(
            command_interface => $Node,
            resource_list      => @Portsets,
            'method-timeout'  => 3600,
            );

  NACL::AppBuilder::EnGenX::Controller::Portset->delete(
            command_interface => $Node,
            resource_list      => @Portsets,
            );

  NACL::AppBuilder::EnGenX::Controller::Portset->delete(
            resource_list      => @Portsets,
            'method-timeout'  => 3600,
            );

  NACL::AppBuilder::EnGenX::Controller::Portset->delete(
            resource_list      => @Portsets,
            );

=over

=item Options

Either the resource_list is required or the NACL Portset Deletion Arguments
are required but both should NOT be used in the same call.

=over

=item C<< resource_list => @resource_list >>

(Required)
If this is passed in the portsets in the resource list will be deleted
based on the information that is supplied in the resource list. If
the resource list is passed in the NACL Porset Deletion Arguments should
NOT be passed in.

=item C<< NACL Protset Deletion Arguments >>

(Required)
These are arguments that can be passed into NACL::C::LunPortset->Delete.
If these arguments are passed in the resource_list argument should NOT
be passed in.

=item C<< count => $protset_count >>

(Optional)
This arguments is used to delete multiple portsets from the same NACL Protset
Deletion Arguments.  To delete the names for the Protsets, the Protset name
passed in is appended with the index that it was deleted at starting at 1.

=item C<< command_interface => $NACL_command_interface >>

(Optional)
This is the command_interface that will be used during NACL calls.  If a
command_interface is not passed in then one will be created.  Passing in a
command_interface helps performance because for every command_interface created
the version of the Filer must be determined by creating a connection to the filer
and then sending commands to get the version.

=item C<< method_timeout => $NACL_method_timeout >>

(Optional)
This is the method_timeout that will be used during NACL calls. If a
method_timeout is not passed in then the method_timeout will be defaulted.

=back

=item Return

This method returns a True or False based on whether the method succeeds or fails
in the deletion of the Protset(s) based on the Options passed in.

=item Exceptions

This method produces and throws no exceptions of its own but pass through any
uncaught exceptions from other library calls.

=cut

sub delete {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %args = @_;

    # Locals
    my $parent_answer;

    # Globals that will also be set as globals in the base class
    $args{'Name'} = "portset";

    # Super::_delete is in Task.pm
    $parent_answer = $pkg->SUPER::_delete(%args);

    return $parent_answer;
} ## end sub delete

#==============================================================================
# Method:     _nacl_create_call
# Objective:  Makes the raw NACL calls for creation of a Protset.
# Details:    Makes the NACL::C::LunPortset->create() call. Also works
#             around some common non-fatal errors that can be return from
#             portset creation and then has a retry mechanism to retry porset
#             creation.
# Inputs:     NACL::C::LunPortset->create arguments
#             retry => integer, is the current retry the method is on.
#                    This is used to count how many times the _nacl_create_call
#                    has been retried so that the retries can be limited.
# Outputs:    Returns whether the create was successful or a hash with success
#             and whether the Protset was a duplicate if a hash is requested
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
#==============================================================================
sub _nacl_create_call {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %args = @_;

    # Locals
    my $duplicated = FALSE;
    my $if_exists
        = exists($args{if_exists}) ? delete($args{if_exists}) : 'die';
    if (!exists($args{command_interface})) {
        my @Nodes = NACL::C::Node->find();
        $args{command_interface} = $Nodes[0];
    }

    # Get the igroups that will be bound to this portset
    my $igroups = exists($args{igroups}) ? delete($args{igroups}) : [];

    if (exists($args{'port-name'})) {
        if (!(ref($args{'port-name'}) =~ /ARRAYREF/i)) {

            # Create an array of the provided string and store it as an
            # ArrayRef
            $args{'port-name'} = [split(/,/, $args{'port-name'})];
        }
    }

    my ($portset_cs, $was_created);

    # Execute Create
    NACL::STask::LunPortset->create(
        %args,
        nacltask_if_exists => $if_exists,
        _was_created       => \$was_created
    );
    if (  !$was_created
        && $if_exists eq 'reuse') {
        $duplicated = TRUE;
        $Log->comment("Using already Created PortSet:" . $args{portset});
        $portset_cs = NACL::CS::LunPortset->fetch(
            command_interface => $args{command_interface},
            filter            => {
                portset => $args{portset},
                vserver => $args{vserver}
            },
        );
    }
    my $was_igroup_bound;

    # Bind the PortSet to whatever igroup was specified.
    foreach my $igroup (@{$igroups}) {
        $was_igroup_bound = 0;
        NACL::STask::LunIgroup->bind(
            command_interface  => $args{command_interface},
            igroup             => $igroup,
            vserver            => $args{vserver},
            portset            => $args{portset},
            nacltask_if_exists => $if_exists,
            _was_created       => \$was_igroup_bound
        );
        if ($if_exists eq 'reuse'
            && !$was_igroup_bound) {

            # Gather the list of igroups this portset is bound to in an
            # Array, transform that array into a hash to make it simpler
            # to see it is defined or not.
            my @bound_to_igroups = @{$portset_cs->{igroups}};
            my %igroups_hash = map { $_ => 1 } @bound_to_igroups;

            if (exists($igroups_hash{$igroup})) {
                $Log->comment("Portset '"
                        . $args{portset}
                        . "' already bound "
                        . "to igroup "
                        . $igroup);
            } else {
                $Log->exit() if $may_exit;
                 NACL::C::Exceptions::LunIgroup::AlreadyBound->throw("Igroup '$igroup' is bound to a different "
                        . " portset: \n");
            }
        }
    }
    $Log->exit() if $may_exit;

    # Returning either a list (hash) or scalar TRUE
    return (wantarray())
        ? ('success' => TRUE, 'duplicated' => $duplicated)
        : TRUE;
} ## end sub _nacl_create_call

#==============================================================================
# Method:     _nacl_delete_call
# Objective:  Makes the raw NACL calls for deletion of a Protset.
# Details:    Makes the NACL::C::LunPortSet->delete() calls. Also works around
#             some common non-fatal errors that can be return from portset
#             deletion and then has a retry mechanism to retry portset deletion.
# Inputs:     NACL::C::LunPortSet->delete arguments
#             retry => integer, is the current retry the method is on.
#                    This is used to count how many times the _nacl_delete_call
#                    has been retried so that the retries can be limited.
# Outputs:    Returns whether the create was successful or a hash with success
#             if a hash is requested
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
#==============================================================================
sub _nacl_delete_call {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %args = @_;

    if (!exists($args{command_interface})) {
        my @Nodes = NACL::C::Node->find();
        $args{command_interface} = $Nodes[0];
    }

    # Get the igroups that will be unbound from this portset
    my $igroups = exists($args{igroups}) ? delete($args{igroups}) : [];

    # find all igroups that are bound to this portset
    my @igroup_cs_objs = ();
    if (scalar(@{$igroups})) {
        @igroup_cs_objs = NACL::CS::LunIgroup->fetch(
            command_interface => $args{command_interface},
            allow_empty       => 1,
            filter            => {
                vserver => $args{vserver},
                portset => $args{portset}
            }
        );
    }

    # unbind the igroups from this portset
    foreach my $igroup (@{$igroups}) {

        # first check if this igroup is bound to this portset
        my $bound = 0;
        foreach my $bigroup (@igroup_cs_objs) {
            if ($igroup eq $bigroup->igroup()) {
                $bound = 1;
                last;
            }
        }
        next if (not $bound);
        NACL::STask::LunIgroup->unbind(
            command_interface  => $args{command_interface},
            igroup             => $igroup,
            vserver            => $args{vserver},
            nacltask_if_purged => 'pass'
        );

    }

    # delete the portset
    NACL::STask::LunPortset->purge(%args, nacltask_if_purged => 'pass');

    $Log->exit() if $may_exit;
    return TRUE;
} ## end sub _nacl_delete_call

#==============================================================================
# Method:     _nacl_wait_call
# Objective:  Makes a NACL call to 'wait_for_creation.
# Details:    Will make the call to NACL::STask::LunPortset->wait_for_creation();
#             Right now, this method has not been implemented in NACL
# Inputs:
# Outputs:    Returns TRUE
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
# TODO: Implement this method when needed
#==============================================================================
sub _nacl_wait_call {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = @_;

    $Log->exit() if $may_exit;
    return TRUE;
} ## end sub _nacl_wait_call

#==============================================================================
# Method:     _get_dtd_nacl_hashes (translate, ignore and delete)
# Objective:  Getter methods for the variables of the same name.
# Details:    Retrieves the hash or a reference to one of the following:
#             - DTD_To_NACL_Translate.
#             - DTD_To_NACL_Ignore.
#             - Delete_Args.
# Inputs:     none.
# Outputs:    The reference or hash variable.
# Exceptions:
#==============================================================================
sub _get_dtd_nacl_translate {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = @_;

    $Log->exit() if $may_exit;
    return (wantarray()) ? %DTD_To_NACL_Translate : \%DTD_To_NACL_Translate;
} ## end sub _get_dtd_nacl_translate

#==============================================================================
sub _get_dtd_nacl_ignore {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = @_;

    $Log->exit() if $may_exit;
    return (wantarray()) ? %DTD_To_NACL_Ignore : \%DTD_To_NACL_Ignore;
} ## end sub _get_dtd_nacl_ignore

#==============================================================================
sub _get_dtd_nacl_delete {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = @_;

    $Log->exit() if $may_exit;
    return (wantarray()) ? %Delete_Args : \%Delete_Args;
} ## end sub _get_dtd_nacl_delete

1;
