# $Id$

#
# Copyright (c) 2001-2013 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Volume Controller Layer Module
## @author dl-san-automation
## @status shared
## @pod here

=head1 NAME

NACL::AppBuilder::EnGenX::Controller::Volume

=head1 DESCRIPTION

NACL::AppBuilder::EnGenX::Controller::Volume is a derived class of
NACL::AppBuilder::EnGenX::Controller::Task.  It is used to Create and
Delete ONTAP Volumes. All methods are static.

=cut

package NACL::AppBuilder::EnGenX::Controller::Volume;

# Compiler directives
use strict;
use warnings;

# NATE and Standard Module import
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use NATE::Util::DataTypes qw(TRUE FALSE);
use Data::Dumper;
use Params::Validate
    qw(validate_with validate SCALAR ARRAYREF BOOLEAN UNDEF SCALARREF HASHREF);

# NACL and EnGenX Libs
use NACL::CS::StorageAggregate;
use NACL::C::VolumeFlexcache;
use NACL::CS::Volume;
use NACL::STask::Volume;
use NACL::C::Job;
use NACL::AppBuilder::EnGenX::SubProcess;
use NACL::AppBuilder::EnGenX::EnGenXUtil;
use NACL::Global;

# Exceptions
use NATE::BaseException qw(:try);
use NATE::Exceptions::Argument qw(:try);
use NACL::C::Exceptions::Volume::AlreadyMounted;
use NACL::C::Exceptions::Volume::AlreadyModified;

# Inherit base class
use base qw( NACL::AppBuilder::EnGenX::Controller::Task );

use vars qw(%DTD_To_NACL_Translate
    %DTD_To_NACL_Ignore
    %Delete_Args
);

# Arguments after converting from xml to a
# resource list that need their keys translated to
# keys that NACL methods will accept.
%DTD_To_NACL_Translate = (
    name   => 'volume',
    aggr   => 'aggregate',
    _reuse => 'if_exists',
    export_policy => 'policy'
);

# Arguments that are used by EnGenX but cannot be
# passed into the NACL methods.
%DTD_To_NACL_Ignore = (
    efficiency               => '',
    parallel                 => '',
    engenx_relationship_id   => '',
    engenx_relationship_type => '',
);

# Arguments that can be passed into the NACL delete
# command.
%Delete_Args = (
    command_interface       => '',
    apiset_must             => '',
    apiset_should           => '',
    volume                  => '',
    vserver                 => '',
    foreground              => '',
    'disable-offline-check' => '',
    force                   => '',
    job_component           => '',
    'is-repository'         => '',
    enable_flexcache        => ''
);

=head1 METHODS

=head2 create

(Static method) Request creation of one or multiple volumes serially
or in parallel and return whether the volume(s) was created successfully.
This method passes all user arguments through to NACL::AppBuilder::EnGenX::Controller::Task
and adds specific arguments used by the EnGenX framework.

  NACL::AppBuilder::EnGenX::Controller::Volume->create(
            command_interface => $Node,
            volume            => ( $Volume_name),
            aggregate         => $Aggr_name,
            vserver           => $Vserver_name,
            size              => "20M",
            job_component     => \$job,
            'method-timeout'  => 3600,
            );

  NACL::AppBuilder::EnGenX::Controller::Volume->create(
            command_interface => $Node,
            volume            => ( $Volume_name),
            aggregate         => $Aggr_name,
            vserver           => $Vserver_name,
            size              => "20M",
            job_component     => \$job,
            'method-timeout'  => 3600,
            count             => 5,
            );

  NACL::AppBuilder::EnGenX::Controller::Volume->create(
            command_interface  => $Node,
            resource_list      => @Vols,
            'method-timeout'  => 3600,
            );

  NACL::AppBuilder::EnGenX::Controller::Volume->create(
            command_interface => $Node,
            resource_list      => @Vols,
            );

  NACL::AppBuilder::EnGenX::Controller::Volume->create(
            resource_list      => @Vols,
            'method-timeout'  => 3600,
            );

  NACL::AppBuilder::EnGenX::Controller::Volume->create(
            resource_list      => @Vols,
            );

=over

=item Options

Either the resource_list is required or the NACL Volume Creation Arguments
are required but both should NOT be used in the same call.

=over

=item C<< resource_list => @resource_list >>

(Optional)
If this is passed in the volumes in the resource list will be created
based on the information that is supplied in the resource list. If
the resource list is passed in the NACL Volume Creation Arguments should
NOT be passed in.

=item C<< NACL Volume Creation Arguments >>

(Optional)
These are arguments that can be passed into NACL::STask::Volume->Create.
If these arguments are passed in the resource_list argument should NOT
be passed in

=item C<< count => $volume_count >>

(Optional)
This arguments is used to create multiple volumes from the same NACL Volume
Creation Arguments.  To create the names for the volumes, the volume name
passed in is appended with the index that it was create at starting at 1.

=item C<< command_interface => $NACL_command_interface >>

(Optional)
This is the command_interface that will be used during NACL calls.  If a
command_interface is not passed in then one will be created.  Passing in a
command_interface helps performance because for every command_interface created
the version of the Filer must be determined by creating a connection to the filer
and then sending commands to get the version.

=item C<< method-timeout => $NACL_method_timeout >>

(Optional)
This is the method-timeout that will be used during NACL calls. If a
method-timeout is not passed in then the method-timeout will be defaulted to 3600.

=back

=item Return

This method returns a True or False based on whether the method succeeds or fails
in the creation of the Volume(s) based on the Options passed in.

=item Exceptions

This method produces and throws no exceptions of its own but pass through any
uncaught exceptions from other library calls.

=cut

sub create {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %args = @_;

    # Locals
    my $parent_answer;

    # TODO Remove $args{'Name'} from everywhere. And ensure
    # no errors occur.
    # NOTE: Can't remove the NAME, it is been used to create objects with a
    # counter. We could get the name from package by asking the for the "caller"
    $args{'Name'} = "volume";

    # Super::_create is in Task.pm
    $parent_answer = $pkg->SUPER::_create(%args);
    return $parent_answer;
} ## end sub create

=head2 delete

(Static method) Request deletion of one or multiple volumes serially
or in parallel and return whether the volume(s) was deleted successfully.
This method passes all user arguments through to NACL::AppBuilder::EnGenX::Controller::Task
and adds specific arguments used by the EnGenX framework.

  NACL::AppBuilder::EnGenX::Controller::Volume->delete(
            command_interface => $Node,
            volume            => ( $Volume_name),
            vserver           => $Vserver_name,
            'method-timeout'  => 3600,
            );

  NACL::AppBuilder::EnGenX::Controller::Volume->delete(
            command_interface => $Node,
            volume            => ( $Volume_name),
            vserver           => $Vserver_name,
            'method-timeout'  => 3600,
            count             => 5
            );

  NACL::AppBuilder::EnGenX::Controller::Volume->delete(
            command_interface => $Node,
            resource_list      => @Vols,
            'method-timeout'  => 3600,
            );

  NACL::AppBuilder::EnGenX::Controller::Volume->delete(
            command_interface => $Node,
            resource_list      => @Vols,
            );

  NACL::AppBuilder::EnGenX::Controller::Volume->delete(
            resource_list      => @Vols,
            'method-timeout'  => 3600,
            );

  NACL::AppBuilder::EnGenX::Controller::Volume->delete(
            resource_list      => @Vols,
            );

=over

=item Options

Either the resource_list is required or the NACL Volume Deletion Arguments
are required but both should NOT be used in the same call.

=over

=item C<< resource_list => @resource_list >>

(Required)
If this is passed in the volumes in the resource list will be deleted
based on the information that is supplied in the resource list. If
the resource list is passed in the NACL Volume Deletion Arguments should
NOT be passed in.

=item C<< NACL Volume Deletion Arguments >>

(Required)
These are arguments that can be passed into NACL::STask::Volume->Delete.
If these arguments are passed in the resource_list argument should NOT
be passed in.

=item C<< count => $volume_count >>

(Optional)
This arguments is used to delete multiple volumes from the same NACL Volume
Deletion Arguments.  To delete the names for the volumes, the volume name
passed in is appended with the index that it was deleted at starting at 1.

=item C<< command_interface => $NACL_command_interface >>

(Optional)
This is the command_interface that will be used during NACL calls.  If a
command_interface is not passed in then one will be created.  Passing in a
command_interface helps performance because for every command_interface created
the version of the Filer must be determined by creating a connection to the filer
and then sending commands to get the version.

=item C<< method-timeout => $NACL_method_timeout >>

(Optional)
This is the method-timeout that will be used during NACL calls. If a
method-timeout is not passed in then the method-timeout will be defaulted to 3600.

=back

=item Return

This method returns a True or False based on whether the method succeeds or fails
in the deletion of the Volume(s) based on the Options passed in.

=item Exceptions

This method produces and throws no exceptions of its own but pass through any
uncaught exceptions from other library calls.

=cut

sub delete {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %args = @_;

    # Locals
    my $parent_answer;

    # TODO Remove $args{'Name'} from everywhere. And ensure
    # no errors occur.
    $args{'Name'} = "volume";

    # Super::_delete is in Task.pm
    $parent_answer = $pkg->SUPER::_delete(%args);

    return $parent_answer;
} ## end sub delete


#==============================================================================
# Method:     _nacl_create_call
# Objective:  Makes the raw NACL calls for creation of a volume.
# Details:    Makes the NACL::STask::Volume->create() call and optionaly could
#             make a the NACL::STask::Volume->wait_for_creation. Also works
#             around some common non-fatal errors that can be return from
#             volume creation and then has a retry mechanism to retry volume
#             creation.
# TODO:       Need to update to NACL::STask::Volume->create()
# Inputs:     NACL::STask::Volume->create arguments
#             retry => integer, is the current retry the method is on.
#                    This is used to count how many times the _nacl_create_call
#                    has been retried so that the retries can be limited.
# Outputs:    Returns whether the create was successful or a hash with success
#             and whether the volume was a duplicate if a hash is requested
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
#==============================================================================
sub _nacl_create_call {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %args = @_;

    # If snapshot-policy is set for constituent vols delete the attribute.
    if(exists $args{'constituent-role'} and exists $args{'snapshot-policy'}) {
       delete $args{'snapshot-policy'};
       $Log->comment("Snapshot-policy cannot be set when creating a constituent volume" .
                     " The attribute has been removed from one of your [" .
                      $args{'volume'} . "] volumes on the vserver: " .
                      $args{'vserver'} . "./n");
    }

    # Locals
    my $was_created = 0;
    my $is_infinite_vol = FALSE;
    my $is_managed_by_service = (exists $args{'is-managed-by-service'}) ? $args{'is-managed-by-service'} : FALSE;
    my ($percent, $sizer, $max_size) = "";
    $args{nacltask_wait} = exists($args{wait}) ? delete($args{wait}) : FALSE;
    my $if_exists
        = exists($args{if_exists}) ? delete($args{if_exists}) : 'die';
    $args{'method-timeout'} = 3600 unless exists($args{'method-timeout'});
    my $files = delete $args{files};
    # account for the depricated XML percent_snapshot_space
    if(defined($args{"percent_snapshot_space"})) {
        $Log->warn("XML option percent_snapshot_space has been deprecated. " .
                   "Use XML option percent-snapshot-space.");
        # set the variables correctly.
        $args{"percent-snapshot-space"} = delete($args{"percent_snapshot_space"});
    }

    # If the size argument has a percent then calculate the size.
    if ($args{size} =~ s/%//) {
        my $aggr_cs_obj = NACL::CS::StorageAggregate->fetch(
                                            command_interface => $args{command_interface},
                                            requested_fields => ['size'],
                                            filter => { aggregate => $args{aggregate},
                                                      },
                                            );
        my $percent = $args{size};
        $args{size} = ($args{size}/100)*($aggr_cs_obj->size());
        $Log->comment("Setting Volume size to " . $args{size} . "b which is " .
                      "$percent\% of aggregate " . $args{aggregate} .
                      " which is size " .     $aggr_cs_obj->size() . "b");
    }

    # Delete the vol autosize-percent option if set and the volume is
    # reused autosize is set to false or the maxsize is set
    if (exists($args{'autosize-percent'})) {
        # remove any percentage signs on the autosize-percent flag
        $args{'autosize-percent'} =~ s/%//g;

        # Delete the autosize-percent if any of these are set
        if (exists($args{'max-autosize'}) or (exists($args{autosize}) && $args{autosize} =~ /false/i)) {
            delete $args{'autosize-percent'};
        } elsif (exists($args{autosize}) && $args{autosize} =~ /true/i) {
            # Else we need to do the math and fill in the values.
            # Find the percentage used ex 50% = .5
            $percent = sprintf("%f",.01 * $args{'autosize-percent'});
            delete $args{'autosize-percent'};

            # Look for the sizer
            # The variable $1 holds the last pattern matched.  If there
            # is a sizer, it will hold that sizer (k, m, g, t, p).  If there
            # is not a sizer it will hold "true" from where we checked
            # the autosize in the elsif above.
            $sizer = "";
            if ($args{size} =~ /\d+(\D)$/) {
                $sizer = $1;
            }

            # Do the autosize percent calculations and set the values
            if ($percent <= 1) {
                # the user wants a percentage less than 100

                # set the max size to be the specified size of the volume
                $args{'max-autosize'} = $args{size};

                # remove the unit of measurement from the size value
                # so that we can do the math calculation
                $args{size} =~ s/\D$//;

                # adjust the size to be the percentage specified against
                # the volume size
                $args{size} = $args{size} * $percent;

                # if a unit of measurement was provided, put it back on
                if ($sizer ne "") {
                    $args{size} = $args{size} . $sizer;
                }
            } elsif ($percent > 1) {
                # the user wants the percentage greater than 100

                # remove the unit of measurement from the size value
                # so that we can do the math calculation
                $args{size} =~ s/\D$//;

                # adjust the max size
                $args{'max-autosize'} = $args{size} * $percent;

                # assign the max size with the unit of measurement
                # if it was provided
                if ($sizer ne "") {
                    $args{'max-autosize'} .= $sizer;
                }
            }
        } else {
            # catch every other case and delete autosize-percent
            delete $args{'autosize-percent'};
        }
    }
    
    if (delete($args{use_deep_junction})) {
        delete($args{'junction-path'});
    }

    # parse out the use default junction path flag
    my $use_default_junction_path = delete($args{use_default_junction_path});

    # if the user specified to use the default junction path,
    # then default the junction path to be the name of the volume
    # only do this if the user did not explicitly provide a junction path
    if ((defined($use_default_junction_path) and $use_default_junction_path eq 1)
        and
        (not exists($args{'junction-path'}) or not length($args{'junction-path'})) ) {
        # set the junction path to be "/$volname"
        $args{'junction-path'} = "/" . $args{volume};
    }

    my $enable_flexcache =
        exists($args{enable_flexcache}) ? delete($args{enable_flexcache}) : '0';

    # This creates a NACL command_interface if one is not passed in.
    if (!exists($args{command_interface})) {
        my @Nodes = NACL::C::Node->find();
        $args{command_interface} = $Nodes[0];
    }

    # if we are running longboard or later, then we need to remove the deprecated autosize parameters
    if ($args{command_interface}->version_manager()->is_version_ge(
        release => $NACL::Global::ONTAP_Release_Longboard_0)) {
        # if autosize is enabled and a mode was not explicitly set, set the mode to grow_shrink
        # we do this for backwards compatibility with users that are not using the autosize-mode option yet
        if (exists $args{'autosize'} and $args{autosize} =~ /true/i and not exists $args{'autosize-mode'}) {
            $args{'autosize-mode'} = "grow_shrink";
        }
        # clean up the deprecated options, they will not work if they get passed down to the nacl stask
        delete($args{autosize});
        delete($args{'autosize-increment'});
    }

    # if no junction path exists, we need to set it to undef
    # because NACL::STask::Volume->create() will mount the volume to the default junction path
    # if it is not specified. It shouldnt be doing this, but oh well.
    if (not exists($args{'junction-path'})) {
        $args{'junction-path'} = undef;
    }

    # Execute Create
    my $vol_obj = NACL::STask::Volume->create(%args,
                                              nacltask_if_exists => $if_exists,
                                              _was_created => \$was_created);

    if($was_created) {
        if (defined($args{"aggr-list"})) {
            # We need to check to see if there is an aggr list, this is a stop gap solution
            # to identifying flexgroups. Call wait_for_creation on volume object to make
            # sure that it has completed.
            $vol_obj->wait_for_creation();
        }
            
        $Log->comment("Using newly Created Volume:" . $args{volume} .
                      " in Vserver:" . $args{vserver});
    } else {
        $Log->comment("Using already Created Volume:" . $args{volume} .
                      " in Vserver:" . $args{vserver});

        # need to verify that the volume is online, online it
        try {
            NACL::STask::Volume->online(command_interface => $args{command_interface},
                                        vserver => $args{vserver},
                                        volume => $args{volume});
        } catch NACL::APISet::Exceptions::ResponseException with {
            my $e = shift;
            # rethrow the exception if the exception was anything other than "is already online"
            if ($e->text() !~ /is already online/i) {
                $e->throw();
            }
        };

        # if a junction path was specified and if the volume was reused,
        # then make sure the volume is mounted on the proper junction-path
        if ($args{'junction-path'}) {
            # load up some extra parameters for the mount command
            my %mount_args = ();
            $mount_args{apiset_should} = $args{apiset_should} if (exists $args{apiset_should});
            $mount_args{'method-timeout'} = $args{'method-timeout'} if (exists $args{'method-timeout'});

            # The following defaults to remount if the mount is already mounted.
            NACL::STask::Volume->mount('command_interface' => $args{'command_interface'},
                                       'junction-path'     => delete $args{'junction-path'},
                                       'vserver'           => $args{'vserver'},
                                       'volume'            => $args{'volume'},
                                       %mount_args);
        }

        # modify the volume with additional options
        my %extra = ();

        # apply the qos policy group if one was specified
        if ($args{'qos-policy-group'}) {
            $extra{'qos-policy-group'} = delete($args{'qos-policy-group'});
        }

        # Modify the infinite volume with new parameters if it already exists
        if ($is_infinite_vol) {
            $extra{'is-managed-by-service'} = $is_managed_by_service;
            $extra{nacltask_if_modified} = 'pass';
        }

        # perform the vol modify operation
        if (scalar(keys %extra)) {
            NACL::STask::Volume->modify('command_interface' => $args{'command_interface'},
                                        'vserver'           => $args{'vserver'},
                                        'volume'            => $args{'volume'},
                                        %extra);
        }
    }

    if ($enable_flexcache) {
        try {
            NACL::C::VolumeFlexcache->create('command_interface' => $args{'command_interface'},
                                             'vserver'           => $args{'vserver'},
                                             'origin-volume'     => $args{'volume'});
            $Log->comment("Using newly Created Volume Flexcache: " . $args{volume});
        } otherwise {
            my $e = shift;
            if ($e->text() =~ /Error: command failed: Duplicate volume name $args{'volume'}."/i) {
                $Log->comment("Continuing setup, Volume Flexcache already exists for volume:$args{'volume'}.");
            } else {
                # otherwise rethrow the exception
                $Log->exit() if $may_exit;
                $e->throw();
            }
        };
    }

    # If the user defined the files parameter then modify the volume.
    if(defined($files) and length($files)) {
        NACL::STask::Volume->modify('command_interface' => $args{'command_interface'},
                                'vserver'           => $args{'vserver'},
                                'volume'            => $args{'volume'},
                                'files'             => $files,
                                nacltask_if_modified => 'pass');
    }

    $Log->exit() if $may_exit;
    # Returning either a list (hash) or scalar TRUE
    return (wantarray()) ? ('success'=>TRUE, 'duplicated'=>!$was_created) : TRUE;

} ## end sub _nacl_create_call

#==============================================================================
# Method:     _nacl_delete_call
# Objective:  Makes the raw NACL calls for deletion of a volume.
# Details:    Makes the NACL::STask::Volume->purge() call.
# Inputs:     NACL::STask::Volume->purge arguments
# Outputs:    Returns whether the delete was successful.
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
#==============================================================================
sub _nacl_delete_call {
    $Log->enter() if $may_enter;
    my $pkg   = shift;
    my %args = @_;

    # Locals
    my $was_deleted = 0;
    $args{'method-timeout'} = 3600 unless exists($args{'method-timeout'});

    if (!exists($args{command_interface})) {
        my @Nodes = NACL::C::Node->find();
        $args{command_interface} = $Nodes[0];
    }

    my $enable_flexcache =
        exists($args{enable_flexcache}) ? delete($args{enable_flexcache}) : '0';

    if ($enable_flexcache) {
        try {
            NACL::C::VolumeFlexcache->delete('command_interface' => $args{'command_interface'},
                                             'vserver'           => $args{'vserver'},
                                             'origin-volume'     => $args{'volume'},
                                             'method-timeout'    => $args{'method-timeout'});
            $Log->comment("Deleted Volume Flexcache: " . $args{'volume'});
        } otherwise {
            my $e = shift;
            if ($e->text() =~ /Error: Vserver $args{'vserver'} does not exist."/i) {
                $Log->comment("Continuing cleanup, Volume Flexcache was already deleted for volume:$args{'volume'}.");
            } else {
                # otherwise rethrow the exception
                $Log->exit() if $may_exit;
                $e->throw();
            }
        };
    }

    NACL::STask::Volume->purge(%args,
                               _was_deleted => \$was_deleted,
                               nacltask_if_purged => 'pass');

    if($was_deleted) {
        $Log->comment("Volume:" . $args{volume} .
                      " in Vserver:" . $args{vserver} . " was deleted");
    } else {
        $Log->comment("Cannot cleanup Volume:" . $args{volume} .
                      " in Vserver:" . $args{vserver} .
                      ". It does not exist in the cluster.\n");
    }
    $Log->exit() if $may_exit;
    return TRUE;

} ## end sub _nacl_delete_call

#==============================================================================
# Method:     _nacl_wait_call
# Objective:  Makes a NACL call to 'wait_for_creation.
# Details:    Makes the call to NACL::STask::Volume->wait_for_creation();
#             that will wait for the node to finish creating the aggr.
# Inputs:     NACL::STask::Volume->wait arguments
# Outputs:    Returns whether the create was successful TRUE or not FALSE
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
#==============================================================================
sub _nacl_wait_call {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %args = @_;
    # Assign values
    $args{'method-timeout'} = 3600 unless exists($args{'method-timeout'});
    if (!exists($args{command_interface})) {
        my @Nodes = NACL::C::Node->find();
        $args{command_interface} = $Nodes[0];
    }

    # Locals
    my %wait_args = (
            'command_interface' => $args{'command_interface'},
            'volume'            => $args{'volume'},
            'vserver'           => $args{'vserver'},
            'method-timeout'    => $args{'method-timeout'}
        );

    #Special Cases
    $wait_args{apiset_should} = $args{apiset_should}
            if exists $args{apiset_should};

    if ($Log->may_debug()) {
        $Log->debug("Waiting args are being created: " . Dumper(\%wait_args));
    }
    #Call
    NACL::STask::Volume->wait_for_creation(%wait_args);

    $Log->exit() if $may_exit;
    return TRUE;
}

#==============================================================================
# Method:     _get_dtd_nacl_hashes (translate, ignore and delete)
# Objective:  Getter methods for the variables of the same name.
# Details:    Retrieves the hash or a reference to one of the following:
#             - DTD_To_NACL_Translate.
#             - DTD_To_NACL_Ignore.
#             - Delete_Args.
# Inputs:     none.
# Outputs:    The reference or hash variable.
# Exceptions:
#==============================================================================
sub _get_dtd_nacl_translate {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = @_;

    $Log->exit() if $may_exit;
    return (wantarray()) ? %DTD_To_NACL_Translate : \%DTD_To_NACL_Translate;
}
#==============================================================================
sub _get_dtd_nacl_ignore {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = @_;

    $Log->exit() if $may_exit;
    return (wantarray()) ? %DTD_To_NACL_Ignore : \%DTD_To_NACL_Ignore;
}
#==============================================================================
sub _get_dtd_nacl_delete {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = @_;

    $Log->exit() if $may_exit;
    return (wantarray()) ? %Delete_Args : \%Delete_Args;
}

1;
