# $Id: //depot/prod/test/main/lib/EnGenX/Controller/Volume_Efficiency$

#
# Copyright (c) 2001-2012 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#


## @author dl-san-automation-core
## @status shared
## @pod here

=head1 NAME

EnGenX::Controller::Volume_Efficiency

=head1 DESCRIPTION

NACL::AppBuilder::EnGenX::Controller::Volume_Efficiency is a derived class of
NACL::AppBuiler::NACL::AppBuilder::EnGenX::Controller::Task.  It is used to Create and
Delete ONTAP Luns. All methods are static.

=cut

package NACL::AppBuilder::EnGenX::Controller::Volume_Efficiency;

# Compiler directives
use strict;
use warnings;

# NATE and Standard Module import
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use NATE::Util::DataTypes qw(TRUE FALSE);
use Data::Dumper;
use Params::Validate
    qw(validate_with validate SCALAR ARRAYREF BOOLEAN UNDEF SCALARREF HASHREF);

# NACL and EnGenX Libs
use NACL::STask::VolumeEfficiency;
use NACL::C::VolumeEfficiencyPolicy;
use NACL::C::Job;
use NACL::AppBuilder::EnGenX::SubProcess;
use NACL::AppBuilder::EnGenX::EnGenXUtil;

# Exceptions
use NATE::BaseException qw(:try);
use NATE::Exceptions::Argument qw(:try);

#Inherit base class
use base qw( NACL::AppBuilder::EnGenX::Controller::Task );

use vars qw(%DTD_To_NACL_Translate
    %DTD_To_NACL_Ignore
    %Delete_Args);

# Arguments after converting from xml to a
# resource list that need their keys translated to
# keys that NACL methods will accept.
%DTD_To_NACL_Translate = (
    name   => 'volume',
    _reuse => 'if_exists'
);

# Arguments that are used by EnGenX but cannot be
# passed into the NACL methods
%DTD_To_NACL_Ignore = (parallel => '',);

# Arguments that can be passed into the NACL delete
# command.
%Delete_Args = (
);

=head1 METHODS

=head2 create

(Static method) Request creation of one or multiple volumes serially
or in parallel and return whether the volume(s) was created successfully.
This method passes all user arguments through to NACL::AppBuilder::EnGenX::Controller::Task
and adds specific arguments used by the EnGenX framework.

  NACL::AppBuilder::EnGenX::Controller::Volume_Efficiency->create(
            command_interface => $Node,
            volume            => ( $Volume_name),
            vserver           => $Vserver_name,
            );

=over

=item Options

If the efficiency policy is already in place no error is thrown. If a new policy
is specified then it will be added and applied to the target volume.

=over

=item C<< NACL Volume Creation Arguments >>

(Optional)
These are arguments that can be passed into NACL::STask::VolumeEfficiency->Config.
If these arguments are passed in the resource_list argument should NOT
be passed in

=item C<< command_interface => $NACL_command_interface >>

(Optional)
This is the command_interface that will be used during NACL calls.  If a
command_interface is not passed in then one will be created.  Passing in a
command_interface helps performance because for every command_interface created
the version of the Filer must be determined by creating a connection to the filer
and then sending commands to get the version.

=back

=item Return

This method returns a True or False based on whether the method succeeds or fails
in the creation of the Volume(s) based on the Options passed in.

=item Exceptions

This method produces and throws no exceptions of its own but pass through any
uncaught exceptions from other library calls.

=cut

sub create {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %args = @_;

    # Locals
    my $parent_answer;

    # TODO Remove $args{'Name'} from everywhere. And ensure
    # no errors occur.
    $args{'Name'} = "volume_efficiency";

    # Super::_create is in Task.pm
    $parent_answer = $pkg->SUPER::_create(%args);

    return $parent_answer;
}

#==============================================================================
#The delete function is not going to be implemented
#due to the options being deleted upon volume deletion.
#==============================================================================
sub delete {
    #TODO: no-op for now. May implement in the future
    return;
}


#==============================================================================
# Method:     _nacl_create_call
# Objective:  Makes the raw NACL calls to add volume efficiency options and add
#             the policy if it is not added.
# Details:    Makes the NACL::STask::VolumeEfficiency->modify() call and does all
#             of the neccessary steps to make sure policies are enabled.
# Inputs:     NACL::STask::VolumeEfficiency->modify arguments
#             retry => integer, is the current retry the method is on.
#                    This is used to count how many times the _nacl_create_call
#                    has been retried so that the retries can be limited.
# Outputs:    Returns whether the create was successful or a hash with success
#             and whether the volume was a duplicate if a hash is requested
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
#==============================================================================
sub _nacl_create_call {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %args = @_;

    # Locals
    my @policies = "";
    my $policy = "";
    my $wait = exists($args{wait}) ? delete($args{wait}) : FALSE;
    my $if_exists
        = exists($args{if_exists}) ? delete($args{if_exists}) : 'die';
    my $retry = exists $args{retry} ? delete $args{retry} : 0;
    $args{'method-timeout'} = 3600 unless exists($args{'method-timeout'});

    my $start = delete $args{start};
    # This creates a NACL command_interface if one is not passed in.
    if (!exists($args{command_interface})) {
        my @Nodes = NACL::C::Node->find();
        $args{command_interface} = $Nodes[0];
    }

    #Turn on volume efficiency for the volume
    try {
        $Log->comment("Turning on volume efficiency for volume: " . $args{volume} .
                      " on vserver " . $args{vserver});
        NACL::STask::VolumeEfficiency->on( command_interface => $args{command_interface},
                                           volume => $args{volume},
                                           vserver => $args{vserver},
                                           nacltask_if_enabled => 'pass'
                                         );
    } catch NACL::APISet::Exceptions::ConnectionFailedException with {
        if ($retry < 3) {
            sleep 30 * ($retry++);
            $pkg->_nacl_create_call(
                %args,
                retry     => $retry,
                if_exists => $if_exists,
                wait      => $wait
            );
        } else {
            my $exception = shift;
            $Log->exit() if $may_exit;

            # Rethrowing caught exception
            $exception->throw();
        }
    } otherwise {
        my $exception = shift;
            $Log->exit() if $may_exit;
            # Rethrowing caught exception
            $exception->throw();
    };

    # Modify the options
    try {
        $Log->comment("Adding efficiency options to volume: " . $args{volume} .
                      " on vserver " . $args{vserver});
        NACL::STask::VolumeEfficiency->modify(%args);
    } otherwise{
        my $exception = shift;
        $Log->comment("Failed to modify volume efficiency for volume " . $args{volume} .
                      " on vserver " . $args{vserver});
        $Log->exit() if $may_exit;
        $exception->throw();
    };

    if($start) {
        # Start the efficiency options up
        try {
            $Log->comment("Starting efficiency options on volume: " . $args{volume} .
                          " on vserver " . $args{vserver});
            NACL::STask::VolumeEfficiency->start( command_interface => $args{command_interface},
                                              volume => $args{volume},
                                              vserver => $args{vserver},
                                             );
        } otherwise {
            my $exception = shift;
            $Log->comment("Failed to start volume efficiency for volume " . $args{volume} .
                          " on vserver " . $args{vserver});
            $Log->exit() if $may_exit;
            $exception->throw();
        };
    }
    $Log->exit() if $may_exit;
    # Returning either a list (hash) or scalar TRUE
    return (wantarray()) ? ('success'=>TRUE, 'duplicated'=>FALSE) : TRUE;
}

#==============================================================================
#The _nacl_delete_call function is not going to be implemented
#due to the options being deleted upon volume deletion.
#==============================================================================
sub _nacl_delete_call {
    #TODO: no-op for now. May implement in the future
    return;
}

#==============================================================================
# Method:     _nacl_wait_call
# Objective:  Makes a NACL call to 'wait_for_creation.
# Details:    Makes the call to NACL::STask::VolumeEfficiency->wait_for_creation();
#             that will wait for the node to finish creating the aggr.
# Inputs:     NACL::STask::VolumeEfficiency->wait arguments
# Outputs:    Returns whether the create was successful TRUE or not FALSE
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
# TODO: Verify if this will ever be needed.
#==============================================================================
sub _nacl_wait_call {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = @_;

    $Log->exit() if $may_exit;
    return TRUE;
}

#==============================================================================
# Method:     _get_dtd_nacl_hashes (translate, ignore and delete)
# Objective:  Getter methods for the variables of the same name.
# Details:    Retrieves the hash or a reference to one of the following:
#             - DTD_To_NACL_Translate.
#             - DTD_To_NACL_Ignore.
#             - Delete_Args.
# Inputs:     none.
# Outputs:    The reference or hash variable.
# Exceptions:
#==============================================================================
sub _get_dtd_nacl_translate {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = @_;

    $Log->exit() if $may_exit;
    return (wantarray()) ? %DTD_To_NACL_Translate : \%DTD_To_NACL_Translate;
}
#==============================================================================
sub _get_dtd_nacl_ignore {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = @_;

    $Log->exit() if $may_exit;
    return (wantarray()) ? %DTD_To_NACL_Ignore : \%DTD_To_NACL_Ignore;
}
#==============================================================================
sub _get_dtd_nacl_delete {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = @_;

    $Log->exit() if $may_exit;
    return (wantarray()) ? %Delete_Args : \%Delete_Args;
}

1;
