# Copyright (c) 2012 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Utility Layer Module
## @author dl-san-automation
## @status shared
## @pod here

=head1 NAME

NACL::AppBuilder::EnGenX::EnGenXUtil

=head1 DESCRIPTION

NACL::AppBuilder::EnGenX::EnGenXUtil Hosts several utility static methods that are
used across several classes.

=cut

package NACL::AppBuilder::EnGenX::EnGenXUtil;

# Compiler directives.
use strict;

# NetApp's module imports
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Error;

# Load  NATE resource
use NATE::BaseException qw(:try);
use NATE::Exceptions::Argument qw(:try);

# Perl's public module imports
use Data::Dumper;

# Global Variables.
use vars qw(
    @EXPORT
    @ISA
);

# Constants for EnGenX
use constant ZAPI_TIMEOUT     => 600;
use constant DEFAULT_PASSWD   => 'netapp1!';
use constant DEFAULT_CONNTYPE => 'ssh-persist';

# Global Provisioning variable
our $TOTAL_TIME = 0;
our $TOTAL_OBJ_COUNT = 0;
our $OBJ_COUNT = 0;

###############################################################################
# Method:    strip_and_translate_excess_args
# Objective: An entry in a resource list contains data that is needed by
#            NACL and data that's needed by EnGenX.  We are going to
#            strip the stuff needed by EnGenX for NACL calls.  By convention
#            EnGenX data will have keys with a underscore prefix: _type
# Input:     %args - argument hash
# Return: %stripped_args - argument hash without EnGenX data
###############################################################################
sub strip_and_translate_excess_args {
    my $self = shift;
    my %args = @_;

    $Log->enter() if $may_enter;

    # The translation hash is what will be used to translate
    #     xml variable names to NACL variable names.
    #     The translation hash should translate all lower case
    #     variables to whatever NACL excepts.  This is done so
    #     the translation hash can be consistent and there is no
    #     guessing what the correct case is for the xml.
    my $translation_hash = delete $args{translation_hash} || {};
    my $ignore_hash      = delete $args{ignore_hash}      || {};
    my $delete_hash      = delete $args{delete_hash}      || undef;

    my %stripped_args;
    foreach my $key (keys(%args)) {
        if ($key =~ /^_interface$/i) {
            $stripped_args{apiset_should}{interface} = uc($args{$key});
        } elsif (($key !~ /^_/ and (!defined $ignore_hash->{lc($key)}))
            || (defined $translation_hash->{lc($key)}))
        {    # Check to see if the key is prefaced by _
                # Add the correct key value pairs to the stripped_args hash
            if ($translation_hash->{lc($key)} eq 'nodes') {
                $stripped_args{($translation_hash->{lc($key)} || lc($key))}
                    = [$args{$key}];
            } else {
                $stripped_args{($translation_hash->{lc($key)} || lc($key))}
                    = $args{$key};
            }
        }
    }

    if (defined $delete_hash) {
        foreach my $key (keys(%stripped_args)) {
            unless (defined($delete_hash->{$key})) {
                delete $stripped_args{$key};
            }
        }
    }
    $Log->debug("The Striped Args Are: \n" . Dumper(\%stripped_args));
    $Log->exit() if $may_exit;
    return %stripped_args;
} ## end sub strip_and_translate_excess_args

=head2 list_parallel_sorter

(Static method) Sorts a resource list (of one type) into a hash table
where the keys are the PARALLEL indexes found in each resource. There
will be as many keys as many different PARALLEL indexes are foundo.

  $sorted_lists=NACL::AppBuilder::EnGenX::EnGenXUtil->list_parallel_sorter(
                                    'resource_list' => $resource_list);

=over

=item C<< resource_list => $resource_list >>

(Required)
This must be a subset of the complete resource list. This list should
hold only resources of the same type (e.g. lun, vol, vserver, lif, etc)
An must have the parallel facto in it's attributes. (it is by default)

=back

=item Return

This method returns a hash of arrays with the resources sorted by the
parallel index that is the key of each group (array).

=item Exceptions

This method deos not produces nor throws exceptions
TODO: Error handling, need to verify the input arguments.

=cut

sub list_parallel_sorter {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %args = @_;

    # LOCALS
    my $resource_list = $args{'resource_list'};
    my %sorted_lists  = ();

    foreach my $resource (@{$resource_list}) {

        # Get the number of subprocesses that will be created in parallel for
        # this resource.
        my $parallel_index = $resource->{'_parallel'};
        $parallel_index = 1 if ($parallel_index == 0);

        # If the parallel index used as a key already exists add the
        # resource to that list.
        if (exists $sorted_lists{$parallel_index}) {
            my $list_ref = $sorted_lists{$parallel_index};
            push(@{$list_ref}, $resource);
        }

        # If the key (index) is not found create an array, add the resurce and
        # dd the array to the hash under that key.
        else {
            my @list = ($resource,);
            $sorted_lists{$parallel_index} = \@list;
        }
    }

    $Log->debug("Sorted List: \n " . Dumper(\%sorted_lists));

    $Log->exit() if $may_exit;
    return (wantarray()) ? %sorted_lists : \%sorted_lists;

} ## end sub list_parallel_sorter

=head2 split_into_buckets

(Static method) Divides the provided list into an array of
arrays. The number of items per array (size of the array) is
specified by the user.
The last array could have less items if the number of the items
in the list is not evenly divisable by the size of the array.

  @list = NACL::AppBuilder::EnGenX::EnGenXUtil->split_into_buckets('list'=> $list,
                                                 'size'=> 4);

=over

=item C<< 'list' => $list >>

(Required)
It is the list that wants to be sub-devided into an array of arrays.
The list could hold anything (resources, objects, strings, etc)

=item C<< 'size' => int >>

(Required)
It is the size of the array that will be use to divide the list.
Then each array of scalar 'size' will be pushed into a container
array, that will be returned to the user.
The number must be a positive integer.

=back

=item Return

This method returns an array of arryas or the reference to it.

=item Exceptions

This method deos not produces nor throws exceptions
TODO: Error handling: Need to verify the input arguments.
      Defaults: For safety, we could default the size of the
                bucket (array).

=cut

sub split_into_buckets {
    $Log->enter() if $may_enter;
    my %args = @_;

    # Locals
    my @list       = @{$args{'list'}};
    my $size       = $args{'size'};
    my @array_list = ();

    $Log->debug("Bucket size: $size");

    # IF requested bucket size is one, no need to split...
    if ($size == 1) {
        return (wantarray()) ? @list : \@list;
    } else {
        while (@list) {
            my @bucket = splice(@list, 0, $size);
            push(@array_list, \@bucket);
        }

        #Return the array
        $Log->debug(
            "Array of arrays tha is returning: \n" . Dumper(\@array_list));
        return (wantarray()) ? @array_list : \@array_list;
    }

    $Log->exit() if $may_exit;

} ## end sub split_into_buckets
1;
