#
# Copyright (c) 2015 NetApp, Inc., All Rights Reserved.
#   Any other use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

package NACL::AppBuilder::EnGenX::EventMan;

use strict;
use warnings;

use NATE::BaseException qw(:try);
use NATE::Log qw(log_global);
use ParamsMgr;

use parent qw(NACL::AppBuilder::EnGenX::Util);

sub new {
    my ($class, %args) = @_;
    my $self = \%args;
    bless ($self, $class);
    return $self;
}

sub initialize {
    my ($self) = @_;

    require EventMan;

    my $key_filter = qr/^EVENTSET\d*$|
                        ^EVENTSETGROUP\d*$|
                        ^SEQUENCE_STYLE$|
                        ^SEQUENCE_COUNT$|
                        ^SEQUENCE_RUNLENGTH$|
                        ^RUNID$/x;

    # parse out all the "non" EventSet\d* & EventSetGroup\d* tags
    # these will be the EventParams at the EventMan Heirarchy Level
    my $eventman_params = &ParamsMgr::search(-search_path => [],
                                             -key => "EVENTMAN",
                                             -key_filter => $key_filter,
                                             -filter_match => 0,
                                             -lowercase_keys => 1,
                                             -eval_values => 1,
                                             paramset => $self->get_paramset());
    if (&ParamsMgr::error()) {
        NATE::BaseException->throw(&ParamsMgr::errmsg());
    }

    # parse out all environment EventParams (deprecated)
    my $environment_eventparams = &ParamsMgr::search(-key => "ENVIRONMENT.EVENTPARAMS",
                                                     -exact_key => 1,
                                                     -lowercase_keys => 1,
                                                     -eval_values => 1,
                                                     paramset => $self->get_paramset());
    if (&ParamsMgr::error()) {
        NATE::BaseException->throw(&ParamsMgr::errmsg());
    }

    # eventparams as a child tag of environment is deprecated, warn the user and tell them
    # use eventman_params instead
    if (keys %{$environment_eventparams}) {
        log_global()->warn("The XML element <eventparams/> as a child of <environment/> is deprecated, " .
                           "please use <eventman_params/> instead. It will continue to work but it will " .
                           "eventually be removed.");
    }

    my $environment_eventman_params = &ParamsMgr::search(-key => "ENVIRONMENT",
                                                         -key_filter => qr/EVENTMAN_PARAMS\d*/,
                                                         paramset => $self->get_paramset());
    if (&ParamsMgr::error()) {
        NATE::BaseException->throw(&ParamsMgr::errmsg());
    }

    my %sum = ();
    foreach my $key (keys %$environment_eventman_params) {
        my $eventman_params = &ParamsMgr::search(-key => "ENVIRONMENT." . $key,
                                                 -exact_key => 1,
                                                 -lowercase_keys => 1,
                                                 -eval_values => 1,
                                                 paramset => $self->get_paramset());
        if (&ParamsMgr::error()) {
            NATE::BaseException->throw(&ParamsMgr::errmsg());
        }
        foreach my $k2 (keys %{$eventman_params}) {
            $sum{$k2} = $eventman_params->{$k2};
        }
    }
    $environment_eventman_params = \%sum;

    # one issue, all event parameters are lowercased at this point, sanfilersCm, needs to get
    # corrected
    if (exists($eventman_params->{sanfilerscm})) {
        $eventman_params->{sanfilersCm} = delete($eventman_params->{sanfilerscm});
    }

    # create the eventman object & store it
    my $eventman_obj = new EventMan(%{$eventman_params},
                                    %{$environment_eventparams},
                                    %{$environment_eventman_params});
    if ($eventman_obj->error()) {
        NATE::BaseException->throw($eventman_obj->errmsg());
    }

    $self->_store_object(type => "eventman",
                         name => "default_eventman_obj",
                         obj  => $eventman_obj);

    # parse out the parameters that will indicate how to execute eventman
    # and store them for later
    my $eventman_run_events_params = &ParamsMgr::search(-search_path => [],
                                                        -key => "EVENTMAN",
                                                        -key_filter => qr/^SEQUENCE_STYLE$|^SEQUENCE_COUNT$|^SEQUENCE_RUNLENGTH$|^RUNID$/,
                                                        -lowercase_keys => 1,
                                                        paramset => $self->get_paramset());
    if (&ParamsMgr::error()) {
        NATE::BaseException->throw(&ParamsMgr::errmsg());
    }

    # register the ControllerClusters, Switches, and SanClient objects with EventMan
    my $clusters = &ParamsMgr::search(-search_path => [],
                                       -key => "CLUSTERS",
                                       -key_filter => qr/^CLUSTER\d*$/,
                                       paramset => $self->get_paramset());
    if (&ParamsMgr::error()) {
        NATE::BaseException->throw(&ParamsMgr::errmsg());
    }

    # register each Controller cluster resource with eventman
    foreach my $cluster (keys %{$clusters}) {
        my $cluster_name = $clusters->{$cluster}->{NAME};

        my ($cluster_obj);
        try {
            ($cluster_obj) = $self->get_objects(-type => "cluster",
                                                -names => [$cluster_name]);
        } otherwise {
            my $e = shift;
            # if the error message from get_objects is something other than (resource not found), error out
            if ($e->text() !~ /not defined on object|not defined in SanMgr/) {
                NATE::BaseException->throw("Failed to obtain object for $cluster_name");
            } else {
                # otherwise skip this resource and go to the next;
            }
        };
        if (!$cluster_obj) {
            # otherwise skip this resource and go to the next;
            next;
        }

        my $resource_event_params = &ParamsMgr::search(-key => "CLUSTERS." . $cluster . ".EVENTPARAMS",
                                                       -exact_key => 1,
                                                       -lowercase_keys => 1,
                                                       -eval_values => 1,
                                                       paramset => $self->get_paramset());
        if (&ParamsMgr::error()) {
            NATE::BaseException->throw(&ParamsMgr::errmsg());
        }

        # remove vserver resources that were dynamically added that do not have any "proper"
        # client/vserver relationships. The purpose of this is to not include unnecessary & unused
        # vservers into EventMan's processing routines.
        my @vserver_list = $self->get_vserver_names(-cluster_name => $cluster_name,
                                                    -type => "regular");
        if ($self->error()) {
            NATE::BaseException->throw("Failed to obtain the regular vserver list for cluster $cluster_name");
        }

        my $object_type = "";

        if ($cluster_obj->isa("San::ControllerCluster")) {
            $object_type = "controllerclusters";
        } elsif ($cluster_obj->isa("NACL::C::Cluster")) {
            $object_type = "nacl_clusters_cm";
        } else {
            NATE::BaseException->throw("Unable to get object type for cluster object." .
                                       " A " . ref($cluster_obj) . " was passed in.");
        }

        my %extra_args = ();
        if (scalar(@vserver_list)) {
            $extra_args{-vservers} = \@vserver_list;
        } else {
            log_global()->comment("No vserver declarations (non 'dummy' vservers) were specified in the XML. " .
                                  "This may cause performance problems during events since it will default " .
                                  "to operating against potentially all the vservers in the cluster.");
        }

        # now add the cluster object as an eventman resource with any vserver scope list we may have
        $eventman_obj->add_resource(-name => $cluster_name,
                                    -obj => $cluster_obj,
                                    -type => $object_type,
                                    %extra_args,
                                    %{$resource_event_params});
        if ($eventman_obj->error()) {
            NATE::BaseException->throw("Unable to add $cluster_name resource to eventman " .
                                           $eventman_obj->errmsg());
        }
    }

    # register all client resources with eventman
    my $client_paramsets = $self->_get_all_client_paramsets();

    # parse out the eventparams tag that may or may not exist on each client
    foreach my $client (@{$client_paramsets}) {
        my $client_name = $client->{NAME};

        # skip clients that are under hypervisor control
        next if ($client->{HYPERVISOR} !~ /^none$/i);
        my ($client_obj);
        try {
            ($client_obj) = $self->get_objects(-type => "client",
                                               -names => [$client_name]);
        } otherwise {
            my $e = shift;
            # if the error message from get_objects is something other than (resource not found), error out
            if ($e->text() !~ /not defined on object|not defined in SanMgr/) {
                NATE::BaseException->throw("Failed to obtain object for $client_name");
            } else {
                # skip this resource and go to the next
            }
        };
        if (!$client_obj) {
            # skip this resource and go to the next
            next;
        }

        # get all resource event parameters for this client
        my $resource_event_params = $client->{EVENTPARAMS};
        if ($resource_event_params) {
            # lower case all keys in the hash
            ParamsMgr::lc_keys($resource_event_params);

            # eval all the values of the keys in the hash
            ParamsMgr::eval_key_values($resource_event_params);
        } else {
            $resource_event_params = {};
        }

        # obtain the vservers that are associated to this client
        my @vserver_list = $self->{parent}->get_vservers_associated_to_client(-client_name => $client_name);

        my %extra_args = ();
        if (scalar(@vserver_list)) {
            $extra_args{-vservers} = \@vserver_list;
        } else {
            log_global()->comment("Could not find any vservers in the XML that are associated to client '$client_name.' " .
                                  "This may cause performance problems during path validation since it will default " .
                                  "to operating against all the vservers in the cluster.");
        }

        # register this resource with eventman
        $eventman_obj->add_resource(-name => $client_name,
                                    -obj => $client_obj,
                                    -type => "sanclients",
                                    %extra_args,
                                    %{$resource_event_params});
        if ($eventman_obj->error()) {
            NATE::BaseException->throw("Unable to add $client_name resource to eventman " . $eventman_obj->errmsg());
        }
    }

    # register all hypervisor resources with eventman
    my $hypervisors = &ParamsMgr::search(-search_path => [],
                                         -key => "HYPERVISORS",
                                         -key_filter => qr/^ESX\d*$|^HYPERV\d*$|^CXS\d*$/,
                                         paramset => $self->get_paramset());
    if (&ParamsMgr::error()) {
        NATE::BaseException->throw(&ParamsMgr::errmsg());
    }

    # parse out the eventparams tag they may or may not exist on each hypervisor
    foreach my $hypervisor (keys %{$hypervisors}) {
        my $hypervisor_name = $hypervisors->{$hypervisor}->{NAME};
        my ($hypervisor_obj);
        try {
            ($hypervisor_obj) = $self->get_objects(-type => "hypervisor",
                                                   -names => [$hypervisor_name]);
        } otherwise {
            my $e = shift;
            # if the error message from get_objects is something other than (resource not found), error out
            if ($e->text() !~ /not defined on object|not defined in SanMgr/) {
                NATE::BaseException->throw("Failed to obtain object for $hypervisor_name");
            } else {
                # otherwise skip this resource and go to the next;
            }
        };
        # skip this resource and go to the next
        if (!$hypervisor_obj) {
            next;
        }

        my $resource_event_params = &ParamsMgr::search(-key => "HYPERVISORS." . $hypervisor . ".EVENTPARAMS",
                                                       -exact_key => 1,
                                                       -lowercase_keys => 1,
                                                       -eval_values => 1,
                                                       paramset => $self->get_paramset());
        if (&ParamsMgr::error()) {
            NATE::BaseException->throw(&ParamsMgr::errmsg());
        }

        # obtain the vservers that are associated to this client
        my @vserver_list = $self->{parent}->get_vservers_associated_to_client(-client_name => $hypervisor_name);

        # obtain the virtual machines that are associated to this hypervisor
        my @vm_names = $self->get_vms_associated_to_hypervisor(hypervisor_name => $hypervisor_name);

        my %extra_args = ();
        if (scalar(@vserver_list)) {
            $extra_args{-vservers} = \@vserver_list;
        } else {
            log_global()->comment("Could not find any vservers in the XML that are associated to hypervisor '$hypervisor_name.' " .
                                  "This may cause performance problems during path validation since it will default " .
                                  "to operating against all the vservers in the cluster.");
        }

        # register this hypervisor with eventman
        $eventman_obj->add_resource(-name => $hypervisor_name,
                                    -obj => $hypervisor_obj,
                                    -type => "sanclients",
                                    %extra_args,
                                    -virtual_machines => \@vm_names,
                                    %{$resource_event_params});
        if ($eventman_obj->error()) {
            NATE::BaseException->throw("Unable to add $hypervisor_name resource to eventman " . $eventman_obj->errmsg());
        }
    }

    # register all switch resources with eventman
    my $switches = &ParamsMgr::search(-search_path => [],
                                      -key => "SWITCHES",
                                      -key_filter => qr/^SWITCH\d*$/,
                                      paramset => $self->get_paramset());
    if (&ParamsMgr::error()) {
        NATE::BaseException->throw(&ParamsMgr::errmsg());
    }


    # add each fabric as a resource into eventman
    my $fabrics = &ParamsMgr::search(-search_path => [],
                                     -key => "FABRICS",
                                     -key_filter => qr/^FABRIC\d*$/,
                                     paramset => $self->get_paramset());
    if (&ParamsMgr::error()) {
        NATE::BaseException->throw(&ParamsMgr::errmsg());
    }

    foreach my $fabric (keys %{$fabrics}) {
        my $fabric_name = $fabrics->{$fabric}->{NAME};
        my @switch_list;

        # build a list of all the switch names in this fabric
        foreach my $fab_switch (keys %{$fabrics->{$fabric}}) {
            # skip all non switch elements of the fabric element
            next if ($fab_switch !~ /^SWITCH\d*$/);
            push(@switch_list, $fabrics->{$fabric}->{$fab_switch}->{NAME});
        }
        # add the resource
        $eventman_obj->add_resource(-name => $fabric_name,
                                    -obj => $fabrics->{$fabric},
                                    -type => "fabrics",
                                    -sanswitches => \@switch_list);
        if ($eventman_obj->error()) {
            NATE::BaseException->throw("Unable to add $fabric_name resource to eventman " . $eventman_obj->errmsg());
        }
    }

    # parse out the eventparams tag they may or may not exist on each switch
    foreach my $switch (keys %{$switches}) {
        my $switch_name = $switches->{$switch}->{NAME};

        my ($switch_obj);
        try {
            ($switch_obj) = $self->get_objects(-type => "switch",
                                               -names => [$switch_name]);
        } otherwise {
            my $e = shift;

            # if the error message from get_objects is something other than (resource not found), error out
            if ($e->text() !~ /not defined on object|not defined in SanMgr/) {
                NATE::BaseException->throw("Failed to obtain object for $switch_name");
            } else {
                # otherwise skip this resource and go to the next;
            }
        };
        if (!$switch_obj) {
            # otherwise skip this resource and go to the next;
            next;
        }

        my $resource_event_params = &ParamsMgr::search(-key => "SWITCHES." . $switch . ".EVENTPARAMS",
                                                       -exact_key => 1,
                                                       -lowercase_keys => 1,
                                                       -eval_values => 1,
                                                       paramset => $self->get_paramset());
        if (&ParamsMgr::error()) {
            NATE::BaseException->throw(&ParamsMgr::errmsg());
        }

        # find out if this switch belongs to a fabric, if it does, then add the fabric name as
        # an event parameter for this switch
        # only perform this routine if the fabrics event parameter does not already exist for this switch
        if (not exists($resource_event_params->{fabrics})) {
            foreach my $fabric (keys %{$fabrics}) {
                foreach my $fab_switch (keys %{$fabrics->{$fabric}}) {
                    # skip all non switch elements of the fabric element
                    next if ($fab_switch !~ /^SWITCH\d*$/);
                    if ($fabrics->{$fabric}->{$fab_switch}->{NAME} eq $switch_name) {
                        # store the fabric name in this switches fabrics array
                        push(@{$resource_event_params->{fabrics}}, $fabrics->{$fabric}->{NAME});
                        last;
                    }
                }
            }
        }

        # register this switch with eventman
        $eventman_obj->add_resource(-name => $switch_name,
                                    -obj => $switch_obj,
                                    -type => "sanswitches",
                                    %{$resource_event_params});
        if ($eventman_obj->error()) {
            NATE::BaseException->throw("Unable to add $switch_name resource to eventman " . $eventman_obj->errmsg());
        }
    }

    # register all fibrebridge resources with eventman
    my $fibrebridges = &ParamsMgr::search(-search_path => [],
                                          -key => "FIBREBRIDGES",
                                          -key_filter => qr/^FIBREBRIDGE\d*$/);
    if (&ParamsMgr::error()) {
        NATE::BaseException->throw(&ParamsMgr::errmsg());
    }

    foreach my $fibrebridge (keys %{$fibrebridges}) {
        my $fibrebridge_name = $fibrebridges->{$fibrebridge}->{NAME};
        my ($fibrebridge_obj);
        try {
            ($fibrebridge_obj) = $self->get_objects(-type => "fibrebridge",
                                                    -names => [$fibrebridge_name]);
        } otherwise {
            my $e = shift;
            if ($e->text() !~ /not defined on object|not defined in SanMgr/) {
                NATE::BaseException->throw("Failed to obtain object for $fibrebridge_name");
            } else {
                # otherwise skip this resource and go to the next
            }
        };
        if (!$fibrebridge_obj) {
                # otherwise skip this resource and go to the next
            next;
        }

        # register this switch with eventman
        $eventman_obj->add_resource(-name => $fibrebridge_name,
                                    -obj  => $fibrebridge_obj,
                                    -type => 'fibrebridges');
        if ($eventman_obj->error()) {
            NATE::BaseException->throw("Unable to add $fibrebridge_name resource to eventman " .
                                           $eventman_obj->errmsg());
        }
    }

    # parse out the eventsets
    my $eventsets_and_groups = &ParamsMgr::search(-key => "EVENTMAN",
                                                  -key_filter => qr/^EVENTSET\d*$|^EVENTSETGROUP\d*$/,
                                                  paramset => $self->get_paramset());
    if (&ParamsMgr::error()) {
        NATE::BaseException->throw(&ParamsMgr::errmsg());
    }

    my @eventset_sequence = ();

    # create an eventset object for each one
    foreach my $element (keys %{$eventsets_and_groups}) {
        if ($element =~ /^EVENTSET\d*$/) {
            my $eventset_obj = $self->_process_eventset($eventsets_and_groups, $element, "EVENTMAN");

            # store the eventset on the sequence
            push(@eventset_sequence, $eventset_obj);
        } elsif ($element =~ /^EVENTSETGROUP\d*$/) {
            my $eventsetgroup_obj = $self->_process_eventsetgroup($eventsets_and_groups, $element, "EVENTMAN");

            # store the eventset on the sequence
            push(@eventset_sequence, $eventsetgroup_obj);
        }
    }

    # store this eventset sequence as the default sequence on the eventman object
    # this way, someone can call eventman_obj->run_events() with an implicit event set sequence
    $eventman_obj->set_default_event_set_sequence(-event_set_sequence => \@eventset_sequence,
                                                  -sequence_params => $eventman_run_events_params);
    if ($eventman_obj->error()) {
        NATE::BaseException->throw("Unable to set default event set sequence " .
                                       $eventman_obj->errmsg());
    }

    return;
}

sub _process_eventset {
    my ($self, $parent_resource, $resource_name, $parent_resource_name) = @_;

    my ($eventman_obj) = $self->get_objects(-type => "eventman");

    my $eventset_name = $parent_resource->{$resource_name}->{NAME};

    # parse out the eventset Event Parameters
    my $eventset_params = &ParamsMgr::search(-key => $parent_resource_name . "." . $resource_name,
                                             -exact_key => 1,
                                             -lowercase_keys => 1,
                                             -eval_values => 1,
                                             -no_private_keys => 1,
                                             -key_filter => qr/^NAME$|^EVENT\d*$/,
                                             -filter_match => 0,
                                             paramset => $self->get_paramset());
    if (&ParamsMgr::error()) {
        NATE::BaseException->throw(&ParamsMgr::errmsg());
    }

    # one issue, all event parameters are lowercased at this point, sanfilersCm, needs to get
    # corrected
    if (exists($eventset_params->{sanfilerscm})) {
        $eventset_params->{sanfilersCm} = delete($eventset_params->{sanfilerscm});
    }

    # create this eventset object
    my $eventset_obj = $eventman_obj->create_event_set(-event_set_name => $eventset_name,
                                                       %{$eventset_params});
    if ($eventman_obj->error()) {
        NATE::BaseException->throw("Unable to create event set $eventset_name " . $eventman_obj->errmsg());
    }

    # add each event to this eventset
    my $events = &ParamsMgr::search(-key => $parent_resource_name . "." . $resource_name,
                                    -key_filter => qr/^EVENT\d*$/,
                                    -exact_key => 1,
                                    paramset => $self->get_paramset());
    if (&ParamsMgr::error()) {
        NATE::BaseException->throw(&ParamsMgr::errmsg());
    }

    # process each event element under this eventset
    foreach my $event_element (keys %{$events}) {
        my $event_name = $events->{$event_element}->{NAME};
        my $event_params = &ParamsMgr::search(-key => $parent_resource_name . "." . $resource_name . "." . $event_element,
                                              -exact_key => 1,
                                              -lowercase_keys => 1,
                                              -eval_values => 1,
                                              -no_private_keys => 1,
                                              -key_filter => qr/^NAME$/,
                                              -filter_match => 0,
                                              paramset => $self->get_paramset());
        if (&ParamsMgr::error()) {
            NATE::BaseException->throw(&ParamsMgr::errmsg());
        }

        # one issue, all event parameters are lowercased at this point, sanfilersCm, needs to get
        # corrected
        if (exists($event_params->{sanfilerscm})) {
            $event_params->{sanfilersCm} = delete($event_params->{sanfilerscm});
        }

        # add this event to the eventset
        $eventman_obj->add_event(-event_name => $event_name,
                                 -event_set_name => $eventset_name,
                                 %{$event_params});
        if ($eventman_obj->error()) {
            NATE::BaseException->throw("Unable to add event $event_name to eventset $eventset_name " . $eventman_obj->errmsg());
        }
    }

    return $eventset_obj;
}

sub _process_eventsetgroup {
    my ($self, $parent_resource, $resource_name, $parent_resource_name) = @_;

    my ($eventman_obj) = $self->get_objects(-type => "eventman");

    my $eventsetgroup_name = $parent_resource->{$resource_name}->{NAME};

    # parse out parameters for this eventset group
    my $eventsetgroup_params = &ParamsMgr::search(-key => $parent_resource_name . "." . $resource_name,
                                                  -exact_key => 1,
                                                  -lowercase_keys => 1,
                                                  -eval_values => 1,
                                                  -no_private_keys => 1,
                                                  -key_filter => qr/^EVENTSET\d*$|^NAME$/,
                                                  -filter_match => 0,
                                                  paramset => $self->get_paramset());
    if (&ParamsMgr::error()) {
        NATE::BaseException->throw(&ParamsMgr::errmsg());
    }

    # Create this eventset group object
    my $eventsetgroup_obj = $eventman_obj->create_event_set_group(-group_name => $eventsetgroup_name,
                                                                  %{$eventsetgroup_params});
    if ($eventman_obj->error()) {
        NATE::BaseException->throw("Unable to create event set $eventsetgroup_name " . $eventman_obj->errmsg());
    }

    # add each eventset to this group
    my $eventsets = &ParamsMgr::search(-key => $parent_resource_name . "." . $resource_name,
                                       -key_filter => qr/^EVENTSET\d*$/,
                                       -exact_key => 1,
                                       paramset => $self->get_paramset());
    if (&ParamsMgr::error()) {
        NATE::BaseException->throw(&ParamsMgr::errmsg());
    }

    # process each event set element in this event set group
    foreach my $eventset_element (keys %{$eventsets}) {
        my $eventset_obj = $self->_process_eventset($eventsets, $eventset_element, $parent_resource_name . "." . $resource_name);

        # add this event set object to the group
        $eventsetgroup_obj->add_event_set_to_group(-event_set_obj => $eventset_obj);
        if ($eventsetgroup_obj->error()) {
            NATE::BaseException->throw("Failed to add eventset to group " . $eventsetgroup_obj->errmsg());
        }
    }

    return $eventsetgroup_obj;
}

1;
