#
# Copyright (c) 2015 NetApp, Inc., All Rights Reserved.
#   Any other use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

package NACL::AppBuilder::EnGenX::Fabric::Config;

use strict;
use warnings;

use NATE::Log qw(log_global);
use NATE::Messenger qw(messenger_global);
use NATE::Process;
use NATE::BaseException qw(:try);
use Data::Dump qw(dump);
use Data::Dumper;

sub config {
    my ($self, %args) = @_;

    my $run_id_prefix = delete($args{run_id_prefix}) // "";

    # check to see if config is disabled for all fabrics
    my $fabric_groups = $self->resource_list()->get_resources(category => "FABRICS",
                                                              type => "fabric_groups",
                                                              data => {CONFIG => 1});
    if (!scalar(@{$fabric_groups})) {
        log_global()->comment("No Fabrics to config");
        return;
    }

    # lazily load the fabric resources only once, and only if
    # fabric config is enabled at the global level
    if (!$self->{loaded}) {
        $self->load_resource();
    }

    # configure all fabric resources that have fabric config enabled
    my $fabrics = $self->resource_list()->get_resources(category => "FABRICS",
                                                        type => "fabrics",
                                                        data => {CONFIG => 1});
    if (!scalar(@{$fabrics})) {
        log_global()->comment("Skipping fabric config (disabled on each fabric)");
        return;
    }
    my @procs = ();
    foreach my $fabric (@{$fabrics}) {
        my $fabric_name = $fabric->get_field("NAME");
        my $proc = NATE::Process->new(runid => $fabric_name,
                                      codespec => sub {
                                          try {
                                              my $data = $self->_config_single(fabric => $fabric);
                                              $self->_config_complete(name => $fabric_name,
                                                                      %{$data});
                                          } otherwise {
                                              my $e = shift;
                                              my $msg = Dumper($e);
                                              $msg = $e->message_string() if $e->isa("NATE::Result");
                                              $msg = $e->text() if $e->isa("NATE::BaseException");
                                              $self->_config_failure(msg  => $msg,
                                                                     name => $fabric_name);
                                              $e->throw() if $e->isa("NATE::BaseException");
                                          };
                                      },
                                      onexit => \&NATE::Process::on_exit_propagate_worst_result);
        push(@procs, $proc);
    }

    # build the success lsitener
    my $success_filter = sub {
        shift->type() eq 'NACL/AppBuilder/EnGenX/Fabric/Config/Complete';
    };
    my $success_callback = sub {
        my ($msg, $listener) = @_;

        # parse out the message information
        my ($fabric_name) = $msg->message()->{name};

        $self->{summary_info}->{$fabric_name} = $msg->message()->{summary_info};
        $self->{missing_wwpns}->{$fabric_name} = $msg->message()->{missing_wwpns};
        $listener->{completed}++;
    };
    my $success_listener = {filter => $success_filter,
                            callback => $success_callback,
                            completed => 0};

    # build the fail listener
    my @failed_fabrics = ();
    my $fail_filter = sub {
        shift->type() eq 'NACL/AppBuilder/EnGenX/Fabric/Config/Failure';
    };
    my $fail_callback = sub {
        my ($msg, $listener) = @_;

        # parse out the message information
        my ($fabric_name) = $msg->message()->{name};
        push(@failed_fabrics, $fabric_name);
        $listener->{completed}++;
    };

    my $fail_listener = {filter => $fail_filter,
                         callback => $fail_callback,
                         completed => 0};

    # start up all fabric config routines
    $self->_spawn_routines(procs => \@procs,
                           fail_listener => $fail_listener,
                           success_listener => $success_listener,
                           runid => $run_id_prefix . "Fabric-config");

    # after all fabrics have finished (successfully or unsucessfully), print out the summary reports
    $self->_display_missing_wwpns(fabric_count => scalar(@{$fabrics}));
    $self->_create_summary();

    if (@failed_fabrics) {
        NATE::BaseException->throw("The following fabrics failed to configure properly (@failed_fabrics)");
    }

    return;
}

sub _config_single {
    my ($self, %args) = @_;

    my $fabric = delete($args{fabric});
    if (!$fabric) {
        my $err = "Missing required parameter 'fabric'";
        NATE::BaseException->throw($err);
    }

    my $fabric_name = $fabric->get_field("NAME");

    # get the fabric object
    my ($fabric_obj) = $self->get_objects(-type => "fabric", -names => [$fabric_name]);
    if ($self->error()) {
        my $err = "Failed to get fabric object for fabric '$fabric_name'";
        NATE::BaseException->throw($err);
    }

    # Grab the switch objects and reconnect them
    my $switches_hash = $fabric_obj->get_switch_hash();
    if ($fabric_obj->error()) {
        my $err = "Failed to get switch hash for fabric ($fabric_name) " .
            $fabric_obj->errmsg();
        NATE::BaseException->throw($err);
    }

    foreach my $switch (sort keys %$switches_hash) {
        # Reconnect the switch object in case we are using the same switch
        # in multiple fabrics
        my $obj = $switches_hash->{$switch};
        $obj->reconnect(-disconnect_existing_conn => $SanUtil::FALSE);
        if ($obj->error()) {
            my $err = "Unable to reconnect switch '$switch' " . $obj->errmsg();
            NATE::BaseException->throw($err);
        }
    }

    # reconnect to all nacl objs
    my @nodes = $self->get_objects(type => 'node', 'obj_class' => 'nacl');
    foreach (@nodes) {
        $_->refresh_command_interface();
    }

    # Build out all of the hashes. Normally you don't have to do this as they are
    # build on demand, however, we want to subtest out the dump of the hashes and
    # we don't want the dump to go throught the build out phase.
    # Dump out the hashes
    $fabric_obj->build_all_hashes();
    if ($fabric_obj->error) {
        # Subtest off the dump of the fabric hashes
        $self->dump_all_hashes(fabric => $fabric_name);
        my $err = "Failed to build the hashes for fabric ($fabric_name) " . $fabric_obj->errmsg();
             NATE::BaseException->throw($err);
    }

    # Subtest off the dump of the fabric hashes
    $self->dump_all_hashes(fabric => $fabric_name);

    # Run the "create" routines within the fabric library
    # Create the aliases
    $fabric_obj->create_aliases();
    if ($fabric_obj->error) {
        my $err = "Failed to create the aliases for fabric ($fabric_name) " . $fabric_obj->errmsg();
        NATE::BaseException->throw($err);
    }

    # Create the zones
    $fabric_obj->create_zones();
    if ($fabric_obj->error) {
        my $err = "Failed to create the zones for fabric ($fabric_name) " . $fabric_obj->errmsg();
        NATE::BaseException->throw($err);
    }

    # Activate the changes
    $fabric_obj->activate_fabric_changes();
    if ($fabric_obj->error) {
        my $err = "Failed to activate the changes for fabric ($fabric_name) " . $fabric_obj->errmsg();
        NATE::BaseException->throw($err);
    }

    $fabric_obj->dump_aliases_and_zones_created_hash();

    # report back a brief summary of what we configured and what we didn't
    my $data = {
        summary_info =>
            {aliases_and_zones_info_hash =>
                 $fabric_obj->get_aliases_and_zones_created_hash(),
             deleted_wwpns_info_hash     =>
                 $fabric_obj->get_deleted_wwpns_info_hash()},
        missing_wwpns => $fabric_obj->get_missing_wwpns_macs()
    };

    return $data;
}

sub _config_failure {
    my ($self, %data) = @_;
    messenger_global()->put(endpoint => $self->{endpoint},
                            type     => "NACL/AppBuilder/EnGenX/Fabric/Config/Failure",
                            message  => \%data);
}

sub _config_complete {
    my ($self, %data) = @_;
    messenger_global()->put(endpoint => $self->{endpoint},
                            type     => "NACL/AppBuilder/EnGenX/Fabric/Config/Complete",
                            message  => \%data);
}

1;
