#
# Copyright (c) 2015 NetApp, Inc., All Rights Reserved.
#   Any other use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

package NACL::AppBuilder::EnGenX::HyperVisor::CXS::Cleanup;

use strict;
use warnings;

use NATE::Process;
use NATE::BaseException qw(:try);
use NATE::Log qw(log_global);
use NATE::Messenger qw(messenger_global);

sub cleanup {
    my ($self, %args) = @_;

    # get all cxs cxsisors with cleanup enabled on them
    my $cxs_resources = $self->resource_list()->get_resources(category => "HYPERVISORS",
                                                                 type => "cxs",
                                                                 data => {CLEANUP => 1});

    # create proc objects for each CXS host that will perform cleanup and return
    # the list of them to the caller
    my $balance_count = 0;
    my @procs = ();
    foreach my $cxs_resource (@{$cxs_resources}) {
        my $host = $cxs_resource->get_field("NAME");
        my $proc = NATE::Process->new(codespec => sub {
                                          try {
                                              $self->_cleanup_single(resource => $cxs_resource,
                                                                   balance => $balance_count++);
                                              $self->_cleanup_complete(name => $host);
                                          } otherwise {
                                              my $e = shift;
                                              $self->_cleanup_failure(name => $host);
                                              $e->throw();
                                          };
                                      },
                                      runid => "CXS-cleanup" . "/" . $host . '-cleanup',
                                      onexit => \&NATE::Process::on_exit_propagate_worst_result);
        push(@procs, $proc);
    }

    return @procs;
}

sub _cleanup_single {
    my ($self, %args) = @_;

    # given a single cxsisor, cleanup this cxsisor
    my $cxs_host = delete($args{resource}) // NATE::BaseException->throw("Missing required parameter 'resource'");
    my $balance = delete($args{balance}) // 0;

    my $cxs_host_name = $cxs_host->get_field("NAME");
    my $protocol = $cxs_host->("PROTOCOL");

    # get all VM resources for this cxs host
    my $virtual_machines = $self->resource_list()->get_resources(category => "HYPERVISORS",
                                                                 type => "virtual_machines",
                                                                 data => {HYPERVISOR_RESOURCE => $cxs_host_name});

    my @vm_names = map { $_->get_field("NAME") } @{$virtual_machines};

    my ($cxs_host_obj) = $self->get_objects(-type => "hypervisor",
                                            -names => [$cxs_host_name]);

    # we need to obtain the list of cluster objects that this client is related to as well
    my $hash = $self->{parent}->get_vserver_client_relationship_hash();
    my @clusters = keys %{$hash->{$cxs_host_name}};

    # get list of vservers that this client will talk to
    my @vserver_list = $self->{parent}->get_vservers_associated_to_client(-client_name => $cxs_host_name);

    # reconnect to this host
    $cxs_host_obj->reconnect(-disconnect_existing_conn => $SanMgr::FALSE);
    if ($cxs_host_obj->error()) {
        NATE::BaseException->throw($cxs_host_obj->errmsg());
    }

    # fetch the NACL cluster objects that this client is related to
    my @nacl_cluster_objs = $self->get_objects('type' => "node",
                                               'names' => \@clusters,
                                               'obj_class' => 'nacl');

    # create the mpdevmgr (this will discover luns on the esx host)
    my $mpdevmgr_obj = $cxs_host_obj->create_mpdevmgr(-nacl_cluster_objs => \@nacl_cluster_objs,
                                                      -vservers => \@vserver_list);
    if ($cxs_host_obj->error()) {
        NATE::BaseException->throw($cxs_host_obj->errmsg());
    }

    foreach my $virtual_machine (values %{$virtual_machines}) {
        # get all the devices attached to this VM
        my @vdi_devices = $cxs_host_obj->get_devices_from_vm(-vm_name => $virtual_machine->{NAME});
        if ($cxs_host_obj->error()) {
            NATE::BaseException->throw("Unable to obtain the devices attched to the VM " . $cxs_host_obj->errmsg());
        }

        if (scalar(@vdi_devices) > 0) {
            $cxs_host_obj->remove_devices_from_vm(-vm_name => $virtual_machine->{NAME},
                                                  -devices => \@vdi_devices);
            if ($cxs_host_obj->error()) {
                NATE::BaseException->throw("Failed to remove the devices from VM " . $cxs_host_obj->errmsg());
            }
        }
    }

    # remove empty storage repositories
    $cxs_host_obj->destroy_storage_repos();
    if ($cxs_host_obj->error()) {
        NATE::BaseException->throw("Failed to destroy SRs " . $cxs_host_obj->errmsg());
    }

    return;
}

sub _cleanup_failure {
    my ($self, %data) = @_;
    messenger_global()->put(endpoint => $self->{endpoint},
                            type     => "NACL/AppBuilder/EnGenX/HyperVisor/CXS/Cleanup/Failure",
                            message  => \%data);
}

sub _cleanup_complete {
    my ($self, %data) = @_;
    messenger_global()->put(endpoint => $self->{endpoint},
                            type     => "NACL/AppBuilder/EnGenX/HyperVisor/CXS/Cleanup/Complete",
                            message  => \%data);
}

1;
