#
# Copyright (c) 2015 NetApp, Inc., All Rights Reserved.
#   Any other use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

package NACL::AppBuilder::EnGenX::HyperVisor::CXS::Deconfig;

use strict;
use warnings;

use NATE::Process;
use NATE::BaseException qw(:try);
use NATE::Log qw(log_global);
use NATE::Messenger qw(messenger_global);

sub deconfig {
    my ($self, %args) = @_;

    # get all cxs cxsisors with deconfig enabled on them
    my $cxs_resources = $self->resource_list()->get_resources(category => "HYPERVISORS",
                                                                 type => "cxs",
                                                                 data => {DECONFIG => 1});

    # create proc objects for each CXS host that will perform deconfig and return
    # the list of them to the caller
    my $balance_count = 0;
    my @procs = ();
    foreach my $cxs_resource (@{$cxs_resources}) {
        my $host = $cxs_resource->get_field("NAME");
        my $proc = NATE::Process->new(codespec => sub {
                                          try {
                                              $self->_deconfig_single(resource => $cxs_resource,
                                                                   balance => $balance_count++);
                                              $self->_deconfig_complete(name => $host);
                                          } otherwise {
                                              my $e = shift;
                                              $self->_deconfig_failure(name => $host);
                                              $e->throw();
                                          };
                                      },
                                      runid => "CXS-deconfig" . "/" . $host . '-deconfig',
                                      onexit => \&NATE::Process::on_exit_propagate_worst_result);
        push(@procs, $proc);
    }

    return @procs;
}

sub _deconfig_single {
    my ($self, %args) = @_;

    # given a single cxsisor, deconfig this cxsisor
    my $resource = delete($args{resource}) // NATE::BaseException->throw("Missing required parameter 'resource'");
    my $balance = delete($args{balance}) // 0;

    my $client_name = $resource->{NAME};

    # get the sanclient object for this CXS host
    my ($cxs_host_obj) = $self->get_objects(-type => "hypervisor",
                                            -names => [$client_name]);

    # reconnect to this host
    $cxs_host_obj->reconnect(-disconnect_existing_conn => $SanMgr::FALSE);
    if ($cxs_host_obj->error()) {
        NATE::BaseException->throw($cxs_host_obj->errmsg());
    }

    # obtain the vserver targets that are associated with this host
    my ($data) = $self->{parent}->get_target_list(-client_name => $client_name,
                                                  -filter_iscsi_targets_by_portset => $cxs_host_obj->{iscsi}->{CONFIG_ISCSI_FILTER_TARGETS_BY_PORTSET},
                                                  -filter_fcp_targets_by_portset   => $cxs_host_obj->{fcp}->{CONFIG_FCP_FILTER_TARGETS_BY_PORTSET});
    if ($self->{parent}->error()) {
        NATE::BaseException->throw($self->{parent}->errmsg());
    }

    my @iSCSITargets = @{$data->{iscsi_targets}};
    my @FcpTargets = @{$data->{fcp_targets}};

    my $protocol = $cxs_host_obj->get_protocol();

    # deconfigure iSCSI/FCP
    if ($protocol =~ /iscsi|fcp/i) {
        $cxs_host_obj->deconfigure(-targets => \@iSCSITargets);
        if ($cxs_host_obj->error()) {
            NATE::BaseException->throw("Failed to deconfigure $protocol for the client ($client_name) " . $cxs_host_obj->errmsg());
        }
    }

    # clean up any orphaned devices
    if ($protocol =~ /iscsi|fcp/i) {
        $cxs_host_obj->orphaned_device_cleanup(-targets => \@iSCSITargets);
        if ($cxs_host_obj->error()) {
            NATE::BaseException->throw("Failed cleanup orphaned devices " . $cxs_host_obj->errmsg());
        }
    }

    return;
}

sub _deconfig_failure {
    my ($self, %data) = @_;
    messenger_global()->put(endpoint => $self->{endpoint},
                            type     => "NACL/AppBuilder/EnGenX/HyperVisor/CXS/Deconfig/Failure",
                            message  => \%data);
}

sub _deconfig_complete {
    my ($self, %data) = @_;
    messenger_global()->put(endpoint => $self->{endpoint},
                            type     => "NACL/AppBuilder/EnGenX/HyperVisor/CXS/Deconfig/Complete",
                            message  => \%data);
}

1;
