#
# Copyright (c) 2015 NetApp, Inc., All Rights Reserved.
#   Any other use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

package NACL::AppBuilder::EnGenX::HyperVisor::CXS::Setup;

use strict;
use warnings;

use NATE::Process;
use NATE::BaseException qw(:try);
use NATE::Log qw(log_global);
use NATE::Messenger qw(messenger_global);
use Hostrec;

sub setup {
    my ($self, %args) = @_;

    # get all cxs hypervisors with setup enabled on them
    my $cxs_resources = $self->resource_list()->get_resources(category => "HYPERVISORS",
                                                                 type => "cxs",
                                                                 data => {SETUP => 1});

    # create proc objects for each CXS host that will perform setup and return
    # the list of them to the caller
    my $balance_count = 0;
    my @procs = ();
    foreach my $cxs_resource (@{$cxs_resources}) {
        my $host = $cxs_resource->get_field("NAME");
        my $proc = NATE::Process->new(codespec => sub {
                                          try {
                                              $self->_setup_single(resource => $cxs_resource,
                                                                   balance => $balance_count++);
                                              $self->_setup_complete(name => $host);
                                          } otherwise {
                                              my $e = shift;
                                              $self->_setup_failure(name => $host);
                                              $e->throw();
                                          };
                                      },
                                      runid => "CXS-setup" . "/" . $host . '-setup',
                                      onexit => \&NATE::Process::on_exit_propagate_worst_result);
        push(@procs, $proc);
    }

    return @procs;
}

sub _setup_single {
    my ($self, %args) = @_;

    # given a single cxsisor, setup this cxsisor
    my $cxs_host = delete($args{resource}) // NATE::BaseException->throw("Missing required parameter 'resource'");
    my $balance = delete($args{balance}) // 0;

    my $cxs_host_name = $cxs_host->get_field("NAME");

    my $protocol = $cxs_host->get_field("PROTOCOL");
    my @protocols;
    if ($protocol =~ /iscsi/i) {
        push(@protocols, "iscsi");
    }
    if ($protocol =~ /fcp/i) {
        push(@protocols, "fcp");
    }

    my $vdi_lun_path_regex_filter = undef;
    if ($cxs_host->field_defined("VDI_LUN_PATH_REGEX") and defined($cxs_host->get_field("VDI_LUN_PATH_REGEX"))) {
        $vdi_lun_path_regex_filter = eval($cxs_host->get_field("VDI_LUN_PATH_REGEX"));
    }

    # get list of vservers that this client will talk to
    my @vserver_list = $self->{parent}->get_vservers_associated_to_client(-client_name => $cxs_host_name);

    # get all VM resources for this cxs host
    my $virtual_machines = $self->resource_list()->get_resources(category => "HYPERVISORS",
                                                                 type => "virtual_machines",
                                                                 data => {HYPERVISOR_RESOURCE => $cxs_host_name});

    my @vm_names = map { $_->get_field("NAME") } @{$virtual_machines};

    my ($cxs_host_obj) = $self->get_objects(-type => "hypervisor",
                                            -names => [$cxs_host_name]);

    # reconnect to this host
    $cxs_host_obj->reconnect(-disconnect_existing_conn => 0);
    if ($cxs_host_obj->error()) {
        NATE::BaseException->throw("Unable to reconnect to CXS host '$cxs_host_name' " . $cxs_host_obj->errmsg());
    }

    # cleanly shutdown all virtual machines
    $cxs_host_obj->domain_management(-action => "shutdown",
                                     -vm_list => \@vm_names);
    if ($cxs_host_obj->error()) {
        NATE::BaseException->throw("Unable to power off VMs " . $cxs_host_obj->errmsg());
    }

    # get the list of vserver targets that this initiator is associated with
    my ($data) = $self->{parent}->get_target_list(-client_name => $cxs_host_name,
                                                  -filter_iscsi_targets_by_portset => $cxs_host_obj->{iscsi}->{CONFIG_ISCSI_FILTER_TARGETS_BY_PORTSET},
                                                  -filter_fcp_targets_by_portset   => $cxs_host_obj->{fcp}->{CONFIG_FCP_FILTER_TARGETS_BY_PORTSET});
    if ($self->{parent}->error()) {
        NATE::BaseException->throw("Failed to get iSCSI/FCP target lists " . $self->{parent}->errmsg());
    }

    my @iSCSITargets = @{$data->{iscsi_targets}};
    my @FcpTargets = @{$data->{fcp_targets}};

    # discover new luns
    $cxs_host_obj->discover(-targets => \@iSCSITargets);
    if ($cxs_host_obj->error()) {
        NATE::BaseException->throw("Unable to discover devices " . $cxs_host_obj->errmsg());
    }

    # we also need to obtain the list of cluster objects that this client is related to as well
    my $hash = $self->{parent}->get_vserver_client_relationship_hash();

    my @clusters = keys %{$hash->{$cxs_host_name}};

    # fetch the NACL cluster objects
    my @nacl_cluster_objs = $self->get_objects('type' => "node",
                                               'names' => \@clusters,
                                               'obj_class' => 'nacl');

    # create mpdevmgr object
    my $cxs_dev_mgr = $cxs_host_obj->create_mpdevmgr(-nacl_cluster_objs => \@nacl_cluster_objs,
                                                     -vservers => \@vserver_list);
    if ($cxs_host_obj->error()) {
        NATE::BaseException->throw("Unable to create MPDev manager object " . $cxs_host_obj->errmsg());
    }

    my @vdi_devices = $cxs_dev_mgr->get_filtered_devices(-vservers => \@vserver_list,
                                                         -device_index => $balance++,
                                                         -ontap_lun_ostype_regex => qr/linux/i,
                                                         -ontap_lun_path_regex => $vdi_lun_path_regex_filter,
                                                         -protocols => \@protocols,
                                                         -ontap_mode => "cmode",
                                                         -device_state => "free");
    if ($cxs_dev_mgr->error()) {
        NATE::BaseException->throw("Failed to obtain VDI devices " . $cxs_dev_mgr->errmsg());
    }

    my $devices_list = "";
    foreach my $vdi_device (@vdi_devices) {
        $devices_list .= "  " . $vdi_device->{mp_device_name} . "\n";
    }

    log_global()->comment("List of filtered devices:\n$devices_list");

    my $num_vdi_per_lun = $cxs_host->get_field("VDIS_PER_LUN") // 1;
    my ($total_requested_vdi_count) = $self->_verify_vdi_count($virtual_machines,
                                                               \@vdi_devices,
                                                               $num_vdi_per_lun);

    my @storage_repos = $cxs_host_obj->create_storage_repos(-devices => \@vdi_devices);
    if ($cxs_host_obj->error()) {
        NATE::BaseException->throw("Failed to create storage repositories " . $cxs_host_obj->errmsg());
    }

    my @all_vdi_disks = $cxs_host_obj->create_vdi_disks(-storage_repos => \@storage_repos,
                                                        -num_vdi_per_lun => $num_vdi_per_lun);
    if ($cxs_host_obj->error()) {
        NATE::BaseException->throw("Failed to create VDIs " . $cxs_host_obj->errmsg());
    }

    foreach my $virtual_machine (@{$virtual_machines}) {
        my $virtual_machine_name = $virtual_machine->get_field("NAME");
        my $vdi_count = $virtual_machine->get_field("VIRTUAL_DISKS") // 0;
        my @vdi_devices = ();

        # figure out the OStype of this virtual machine
        my $hostrec = Hostrec->new(id => $virtual_machine_name);
        my $ostype = $hostrec->hosttype();
        $ostype = "linux" if ($ostype =~ /linux/i);
        $ostype = "windows" if ($ostype =~ /windows/i);
        if (not length($ostype) or $ostype !~ /windows|linux/i) {
            NATE::BaseException->throw("The hosttype '$ostype' of virtual machine '$virtual_machine_name' is invalid, only linux and windows hosttypes are supported");
        }

        for (my $i = 0; $i < $vdi_count; $i++) {
            push(@vdi_devices, shift(@all_vdi_disks));
        }

        # attach all devices to this guest
        if (scalar(@vdi_devices) > 0) {
            log_global()->comment("Attaching " . @vdi_devices . " devices to ($virtual_machine_name)");

            $cxs_host_obj->add_devices_to_vm(-devices => \@vdi_devices,
                                             -vm_name => $virtual_machine_name);
            if ($cxs_host_obj->error()) {
                NATE::BaseException->throw("Failed to attach devices to VM ($virtual_machine_name)");
            }
        } else {
            log_global()->comment("No devices requested to be attached to VM ($virtual_machine_name)");
        }
    }

    # turn all VMs back on
    $cxs_host_obj->domain_management(-action => "power_on",
                                     -vm_list => \@vm_names);
    if ($cxs_host_obj->error()) {
        NATE::BaseException->throw("Unable to power on VMs " . $cxs_host_obj->errmsg());
    }

    return;
}

sub _verify_vdi_count {
    my ($self, $virtual_machines, $vdi_devices_lun_list, $num_vdi_per_lun) = @_;

    my $total_requested_vdi_count = 0;

    # calculate the total vdi disk request counts
    foreach my $vm (values %{$virtual_machines}) {
        $total_requested_vdi_count += $vm->get_field("VIRTUAL_DISKS");
    }

    # calculate available vdi count
    my $actual_vdi_count = scalar(@$vdi_devices_lun_list) * $num_vdi_per_lun;

    if ($actual_vdi_count < $total_requested_vdi_count) {
        NATE::BaseException->throw("Requested '$total_requested_vdi_count' VDIs when only '$actual_vdi_count' VDIs are available");
    }

    return ($total_requested_vdi_count);
}

sub _setup_failure {
    my ($self, %data) = @_;
    messenger_global()->put(endpoint => $self->{endpoint},
                            type     => "NACL/AppBuilder/EnGenX/HyperVisor/CXS/Setup/Failure",
                            message  => \%data);
}

sub _setup_complete {
    my ($self, %data) = @_;
    messenger_global()->put(endpoint => $self->{endpoint},
                            type     => "NACL/AppBuilder/EnGenX/HyperVisor/CXS/Setup/Complete",
                            message  => \%data);
}

1;
