#
# Copyright (c) 2015 NetApp, Inc., All Rights Reserved.
#   Any other use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

package NACL::AppBuilder::EnGenX::HyperVisor::ESX::Cleanup;

use strict;
use warnings;

use NATE::Process;
use NATE::BaseException qw(:try);
use NATE::Log qw(log_global);
use NATE::Messenger qw(messenger_global);

sub cleanup {
    my ($self, %args) = @_;

    # get all esx hypervisors with cleanup enabled on them
    my $esx_resources = $self->resource_list()->get_resources(category => "HYPERVISORS",
                                                              type => "esx",
                                                              data => {CLEANUP => 1});

    # create proc objects for each ESX host that will perform cleanup and return
    # the list of them to the caller
    my @procs = ();
    foreach my $esx_resource (@{$esx_resources}) {
        my $host = $esx_resource->get_field("NAME");
        my $proc = NATE::Process->new(codespec => sub {
                                          try {
                                              $self->_cleanup_single(resource => $esx_resource);
                                              $self->_cleanup_complete(name => $host);
                                          } otherwise {
                                              my $e = shift;
                                              $self->_cleanup_failure(name => $host);
                                              $e->throw();
                                          };
                                      },
                                      runid => "ESX-cleanup" . "/" . $host . '-cleanup',
                                      onexit => \&NATE::Process::on_exit_propagate_worst_result);
        push(@procs, $proc);
    }

    return @procs;
}

sub _cleanup_single {
    my ($self, %args) = @_;

    # given a single hypervisor, setup this hypervisor
    my $esx_host = delete($args{resource}) // NATE::BaseException->throw("Missing required parameter 'resource'");

    my $esx_host_name = $esx_host->get_field("NAME");
    my $protocol = $esx_host->get_field("PROTOCOL");

    my $virtual_machines = $self->resource_list()->get_resources(category => "HYPERVISORS",
                                                                 type => "virtual_machines",
                                                                 data => {HYPERVISOR_RESOURCE => $esx_host_name});

    # we need to obtain the list of cluster objects that this client is related to as well
    my $hash = $self->{parent}->get_vserver_client_relationship_hash();
    my @clusters = keys %{$hash->{$esx_host_name}};

    # get list of vservers that this client will talk to
    my @vserver_list = $self->{parent}->get_vservers_associated_to_client(-client_name => $esx_host_name);

    # obtain an array of all the virtual machine names
    my @vm_names = map { $_->get_field("NAME") } @{$virtual_machines};

    my ($esx_host_obj) = $self->get_objects(-type => "hypervisor",
                                            -names => [$esx_host_name]);

    # reconnect to this host
    $esx_host_obj->reconnect(-disconnect_existing_conn => 0);
    if ($esx_host_obj->error()) {
        NATE::BaseException->throw($esx_host_obj->errmsg());
    }

    if (scalar(@vm_names)) {
        # power off all the virtual machines on this host
        $esx_host_obj->power_management(-action => "power_off",
                                        -vm_list => \@vm_names);
        if ($esx_host_obj->error()) {
            NATE::BaseException->throw($esx_host_obj->errmsg());
        }
    }

    my $vmfs_lvm_obj = $esx_host_obj->create_lvm(-lvm => 'vmfs');
    if ($esx_host_obj->error()) {
        NATE::BaseException->throw($esx_host_obj->errmsg());
    }

    my $rdm_lvm_obj = $esx_host_obj->create_lvm(-lvm => 'rdm');
    if ($esx_host_obj->error()) {
        NATE::BaseException->throw($esx_host_obj->errmsg());
    }

    my $nfs_lvm_obj = $esx_host_obj->create_lvm(-lvm => 'nfs');
    if ($esx_host_obj->error()) {
        NATE::BaseException->throw($esx_host_obj->errmsg());
    }

    my $vvol_lvm_obj = $esx_host_obj->create_lvm(-lvm => 'vvol');
    if ($esx_host_obj->error()) {
        NATE::BaseException->throw($esx_host_obj->errmsg());
    }

    # fetch the NACL cluster objects that this client is related to
    my @nacl_cluster_objs = $self->get_objects('type' => "node",
                                              'names' => \@clusters,
                                              'obj_class' => 'nacl');

    my $mpdevmgr_obj;
    # create the mpdevmgr (this will discover luns on the esx host)
    if ($protocol =~ /iscsi|fcp/i) {
        $mpdevmgr_obj = $esx_host_obj->create_mpdevmgr(-nacl_cluster_objs => \@nacl_cluster_objs,
                                                       -vservers => \@vserver_list);
        if ($esx_host_obj->error()) {
            NATE::BaseException->throw($esx_host_obj->errmsg());
        }
    }

    if (scalar(@vm_names)) {
        my (@subtests, @all_subtests);

        # spawn off subtests to add all vmdks to each guest
        my $parallel_limit = $esx_host->get_field("VMDK_ADD_REMOVE_PARALLEL_LIMIT");

        foreach my $virtual_machine (@{$virtual_machines}) {
            my $virtual_machine_name = $virtual_machine->{NAME};

            my $cleanup_vm_subroutine = sub {
                # open a new connection to the esx client or else they would interfere with eachother
                $esx_host_obj->reconnect(-disconnect_existing_conn => 0);
                if ($esx_host_obj->error()) {
                    NATE::BaseException->throw($esx_host_obj->errmsg());
                }

                if ($protocol =~ /iscsi|fcp/i) {
                    # destroy all the vmdks (rdms/vmfs) associated to this virtual machine
                    $vmfs_lvm_obj->destroy_vmdks(-vm_name => $virtual_machine_name);
                    if ($vmfs_lvm_obj->error()) {
                        NATE::BaseException->throw($vmfs_lvm_obj->errmsg());
                    }

                    $rdm_lvm_obj->destroy_vmdks(-vm_name => $virtual_machine_name);
                    if ($rdm_lvm_obj->error()) {
                        NATE::BaseException->throw($rdm_lvm_obj->errmsg());
                    }
                }

                if ($protocol =~ /nfs/i) {
                    $nfs_lvm_obj->destroy_vmdks(vm_name => $virtual_machine_name);
                    if ($nfs_lvm_obj->error()) {
                        NATE::BaseException->throw($nfs_lvm_obj->errmsg());
                    }
                }

                # delete any vvol vmdks if we had any for this vm
                $vvol_lvm_obj->destroy_vmdks(vm_name => $virtual_machine_name,
                                             vmdk_filepath_regex_filter => qr/nate_(iscsi|nfs|fcp)_container/i);
                if ($vvol_lvm_obj->error()) {
                    NATE::BaseException->throw($vvol_lvm_obj->errmsg());
                }

                # cleanup this vm's config file
                $esx_host_obj->cleanup_vm_config_file(-vm_list => [$virtual_machine_name],
                                                      -vmdk_filepath_regex_filter => qr/nate_rdm_|nate_datastore_|nate_nfs_datastore|nate_(iscsi|nfs|fcp)_container/i);
                if ($esx_host_obj->error()) {
                    NATE::BaseException->throw($esx_host_obj->errmsg());
                }

                &Tharn::disconnect($esx_host_obj->{conn}, '-ok');
            };

            # wait for subtests to finish before we spawn off the next batch of subtests
            if ($parallel_limit !~ /unlimited/i and
                scalar(@subtests) >= $parallel_limit and
                $esx_host_obj->has_vma_host()) {
                # wait for these subtests to finish
                foreach (@subtests) {
                    $_->wait();
                    # if this failed, boot all the vms back up and abort
                    if ($_->worst_result()) {
                        NATE::BaseException->throw("Failed to remove disks from a virtual machine");
                    }
                    $_->destroy();
                }

                # reset the subtest list
                @subtests = ();
            }


            my $proc = NATE::Process->new(codespec => $cleanup_vm_subroutine,
                                          runid => "${virtual_machine_name}-remove_vmdks",
                                          onexit => \&NATE::Process::on_exit_propagate_worst_result);
            $proc->start();

            # push onto the working subtest list
            push(@subtests, $proc);
        }

        # if there are any subtests left in the list, reap them
        foreach (@subtests) {
            $_->wait();
            if ($_->worst_result()) {
                NATE::BaseException->throw("Failed to remove disks from a virtual machine");
            }
            $_->destroy();
        }

        # turn on all virtual machines
        $esx_host_obj->power_management(-action => "power_on",
                                        -vm_list => \@vm_names);

        if ($esx_host_obj->error()) {
            NATE::BaseException->throw($esx_host_obj->errmsg());
        }
    }

    if ($protocol =~ /iscsi|fcp/i) {
        # destroy datastores on this esx host
        $vmfs_lvm_obj->destroy_datastores(-datastore_path_regex => qr/nate_datastore_${esx_host_name}/,
                                          -delete_non_empty_datastores => $esx_host->{DELETE_NON_EMPTY_DATASTORES});
        if ($vmfs_lvm_obj->error()) {
            NATE::BaseException->throw($vmfs_lvm_obj->errmsg());
        }
    }

    # need to unmount nfs datastores
    if ($protocol =~ /nfs/i) {
        $nfs_lvm_obj->unmount_datastores(datastore_path_regex => qr/nate_nfs_datastore/);
        if ($nfs_lvm_obj->error()) {
            NATE::BaseException->throw($nfs_lvm_obj->errmsg());
        }
    }

    $self->synchronize();

    # cleanup anything left over in any vvol datastores
    $vvol_lvm_obj->cleanup_datastores(datastore_regex_filter => qr/nate_(iscsi|fcp|nfs)_container/i,
                                      delete_non_empty_datastores => $esx_host->{DELETE_NON_EMPTY_DATASTORES});
    if ($vvol_lvm_obj->error()) {
        NATE::BaseException->throw($vvol_lvm_obj->errmsg());
    }

    # unmount vvol datastores from this host
    use NACL::AppBuilder::EnGenX::VPserver::Cleanup;
    $self->NACL::AppBuilder::EnGenX::VPserver::Cleanup::unmount_vvol_datastores_from_host(
        esx_host_obj => $esx_host_obj,
        datastore_regex_filter => qr/nate_(iscsi|fcp|nfs)_container/i
    );

    if ($protocol =~ /iscsi|fcp/i) {
        # offline devices that we are cleaning up
        my @mpdev_objs = $mpdevmgr_obj->get_filtered_devices(-vservers => \@vserver_list);
        if ($mpdevmgr_obj->error()) {
            # ignore error if there are no devices to offline
            $mpdevmgr_obj->_clear_err();
        } else {
            # offline the devices that were returned
            $mpdevmgr_obj->offline_devices(-mpdev_objs => \@mpdev_objs);
            if ($mpdevmgr_obj->error()) {
                NATE::BaseException->throw($mpdevmgr_obj->errmsg());
            }
        }
    }

    # refresh each esx host
    $esx_host_obj->restart_mgmt_vmware_service();
    if ($esx_host_obj->error()) {
        NATE::BaseException->throw($esx_host_obj->errmsg());
    }

    $self->synchronize();

    sleep 10;

    $esx_host_obj->vmfs_rescan();
    if ($esx_host_obj->error()) {
        NATE::BaseException->throw($esx_host_obj->errmsg());
    }
}

sub _cleanup_failure {
    my ($self, %data) = @_;
    messenger_global()->put(endpoint => $self->{endpoint},
                            type     => "NACL/AppBuilder/EnGenX/HyperVisor/ESX/Cleanup/Failure",
                            message  => \%data);

    $self->sync_rollback();
}

sub _cleanup_complete {
    my ($self, %data) = @_;
    messenger_global()->put(endpoint => $self->{endpoint},
                            type     => "NACL/AppBuilder/EnGenX/HyperVisor/ESX/Cleanup/Complete",
                            message  => \%data);
}

1;
