#
# Copyright (c) 2015 NetApp, Inc., All Rights Reserved.
#   Any other use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

package NACL::AppBuilder::EnGenX::HyperVisor::ESX::Config;

use strict;
use warnings;

use NATE::Process;
use NATE::BaseException qw(:try);
use NATE::Log qw(log_global);
use NATE::Messenger qw(messenger_global);

sub config {
    my ($self, %args) = @_;

    # get all esx hypervisors with config enabled on them
    my $esx_resources = $self->resource_list()->get_resources(category => "HYPERVISORS",
                                                              type => "esx",
                                                              data => {CONFIG => 1});

    # create proc objects for each ESX host that will perform config and return
    # the list of them to the caller
    my @procs = ();
    foreach my $esx_resource (@{$esx_resources}) {
        my $host = $esx_resource->get_field("NAME");
        my $proc = NATE::Process->new(codespec => sub {
                                          try {
                                              $self->_config_single(resource => $esx_resource);
                                              $self->_config_complete(name => $host);
                                          } otherwise {
                                              my $e = shift;
                                              $self->_config_failure(name => $host);
                                              $e->throw();
                                          };
                                      },
                                      runid => "ESX-config" . "/" . $host . '-config',
                                      onexit => \&NATE::Process::on_exit_propagate_worst_result);
        push(@procs, $proc);
    }

    return @procs;
}

sub _config_single {
    my ($self, %args) = @_;

    # given a single hypervisor, setup this hypervisor
    my $esx_host = delete($args{resource}) // NATE::BaseException->throw("Missing required parameter 'resource'");

    my $client_name = $esx_host->get_field("NAME");

    # given a single hypervisor, configure this hypervisor
    # By the time this function is done, this client should be
    # ready for fc zoning, or iscsi logins

    # obtain the sanclient object
    my ($esx_host_obj) = $self->get_objects(-type => "hypervisor",
                                            -names => [$client_name]);

    # get the list of vserver targets that this initiator is associated with
    my ($data) = $self->{parent}->get_target_list(-client_name => $client_name,
                                                  -filter_iscsi_targets_by_portset => $esx_host_obj->{iscsi}->{CONFIG_ISCSI_FILTER_TARGETS_BY_PORTSET},
                                                  -filter_fcp_targets_by_portset   => $esx_host_obj->{fcp}->{CONFIG_FCP_FILTER_TARGETS_BY_PORTSET});
    if ($self->{parent}->error()) {
        NATE::BaseException->throw("Failed to get iSCSI/FCP target lists");
    }

    my @iSCSITargets = @{$data->{iscsi_targets}};
    my @FcpTargets = @{$data->{fcp_targets}};

    # reconnect to this host
    $esx_host_obj->reconnect(-disconnect_existing_conn => 0);
    if ($esx_host_obj->error()) {
        NATE::BaseException->throw($esx_host_obj->errmsg());
    }

    my $protocol = $esx_host_obj->get_protocol();

    # configure iSCSI/FCP
    if ($protocol =~ /iscsi|fcp/i) {
        $esx_host_obj->configure(-iscsi_targets => \@iSCSITargets,
                                 -fcp_targets => \@FcpTargets,
                                 configure_syslog => $esx_host->get_field("CONFIGURE_SYSLOG"));
        if ($esx_host_obj->error()) {
            NATE::BaseException->throw("Failed to configure $protocol for the client ($client_name) " . $esx_host_obj->errmsg());
        }
    }

    # NTP sync the hypervisors
    # TODO need to have this pull defaults from the env
    my $ntp_sync =        $esx_host->field_defined("NTP_SYNC")        ? $esx_host->get_field("NTP_SYNC") : 0;
    my $ntp_time_server = $esx_host->field_defined("NTP_TIME_SERVER") ? $esx_host->get_field("NTP_TIME_SERVER") : "ntp0.eng.netapp.com";
    my $ntp_time_zone   = $esx_host->field_defined("NTP_TIME_SERVER") ? $esx_host->get_field("NTP_TIME_ZONE") : "EST5EDT";
    if ($ntp_sync == 1) {
        $esx_host_obj->ntp_sync(-ntp_server => $ntp_time_server,
                                -ntp_time_zone => $ntp_time_zone);
        if ($esx_host_obj->error()) {
            my @error = $esx_host_obj->errmsg();
            log_global()->warn("Failed to NTP time sync the $client_name (@error)");
            $esx_host_obj->_clear_err();
        }
    }

    return;
}

sub _config_failure {
    my ($self, %data) = @_;
    messenger_global()->put(endpoint => $self->{endpoint},
                            type     => "NACL/AppBuilder/EnGenX/HyperVisor/ESX/Config/Failure",
                            message  => \%data);
}

sub _config_complete {
    my ($self, %data) = @_;
    messenger_global()->put(endpoint => $self->{endpoint},
                            type     => "NACL/AppBuilder/EnGenX/HyperVisor/ESX/Config/Complete",
                            message  => \%data);
}

1;
