#
# Copyright (c) 2015 NetApp, Inc., All Rights Reserved.
#   Any other use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

package NACL::AppBuilder::EnGenX::HyperVisor::ESX::Deconfig;

use strict;
use warnings;

use NATE::Process;
use NATE::BaseException qw(:try);
use NATE::Log qw(log_global);
use NATE::Messenger qw(messenger_global);

sub deconfig {
    my ($self, %args) = @_;

    # get all esx hypervisors with deconfig enabled on them
    my $esx_resources = $self->resource_list()->get_resources(category => "HYPERVISORS",
                                                              type => "esx",
                                                              data => {DECONFIG => 1});

    # create proc objects for each ESX host that will perform deconfig and return
    # the list of them to the caller
    my @procs = ();
    foreach my $esx_resource (@{$esx_resources}) {
        my $host = $esx_resource->get_field("NAME");
        my $proc = NATE::Process->new(codespec => sub {
                                          try {
                                              $self->_deconfig_single(resource => $esx_resource);
                                              $self->_deconfig_complete(name => $host);
                                          } otherwise {
                                              my $e = shift;
                                              $self->_deconfig_failure(name => $host);
                                              $e->throw();
                                          };
                                      },
                                      runid => "ESX-deconfig" . "/" . $host . '-deconfig',
                                      onexit => \&NATE::Process::on_exit_propagate_worst_result);
        push(@procs, $proc);
    }

    return @procs;
}

sub _deconfig_single {
    my ($self, %args) = @_;

    # given a single hypervisor, setup this hypervisor
    my $esx_host = delete($args{resource}) // NATE::BaseException->throw("Missing required parameter 'resource'");

    my $client_name = $esx_host->get_field("NAME");

    # get the sanclient object for this esx host
    my ($esx_host_obj) = $self->get_objects(-type => "hypervisor",
                                            -names => [$client_name]);

    # reconnect to this host
    $esx_host_obj->reconnect(-disconnect_existing_conn => 0);
    if ($esx_host_obj->error()) {
        NATE::BaseException->throw($esx_host_obj->errmsg());
    }

    # obtain the vserver targets that are associated with this host
    my ($data) = $self->{parent}->get_target_list(-client_name => $client_name,
                                                  -filter_iscsi_targets_by_portset => $esx_host_obj->{iscsi}->{CONFIG_ISCSI_FILTER_TARGETS_BY_PORTSET},
                                                  -filter_fcp_targets_by_portset   => $esx_host_obj->{fcp}->{CONFIG_FCP_FILTER_TARGETS_BY_PORTSET});
    if ($self->{parent}->error()) {
        NATE::BaseException->throw("Failed to get iSCSI/FCP target lists");
    }

    my @iSCSITargets = @{$data->{iscsi_targets}};
    my @FcpTargets = @{$data->{fcp_targets}};

    my $protocol = $esx_host_obj->get_protocol();

    # deconfigure iSCSI/FCP
    if ($protocol =~ /iscsi|fcp/i and scalar(@iSCSITargets)) {
        $esx_host_obj->deconfigure(-iscsi_targets => \@iSCSITargets,
                                   -clean_level => $esx_host_obj->{iscsi}->{CONFIG_ISCSI_DISCOVERY_PRECLEANUP});
        if ($esx_host_obj->error()) {
            NATE::BaseException->throw($esx_host_obj->errmsg());
        }
    }

    # clean up any orphaned devices
    $esx_host_obj->orphaned_device_cleanup();
    if ($esx_host_obj->error()) {
        NATE::BaseException->throw($esx_host_obj->errmsg());
    }

    # restart the host agent
    $esx_host_obj->restart_mgmt_vmware_service();
    if ($esx_host_obj->error()) {
        NATE::BaseException->throw($esx_host_obj->errmsg());
    }

    return;
}

sub _deconfig_failure {
    my ($self, %data) = @_;
    messenger_global()->put(endpoint => $self->{endpoint},
                            type     => "NACL/AppBuilder/EnGenX/HyperVisor/ESX/Deconfig/Failure",
                            message  => \%data);
}

sub _deconfig_complete {
    my ($self, %data) = @_;
    messenger_global()->put(endpoint => $self->{endpoint},
                            type     => "NACL/AppBuilder/EnGenX/HyperVisor/ESX/Deconfig/Complete",
                            message  => \%data);
}


1;
