#
# Copyright (c) 2015 NetApp, Inc., All Rights Reserved.
#   Any other use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

package NACL::AppBuilder::EnGenX::HyperVisor::ESX::Setup;

use strict;
use warnings;

use NATE::Process;
use NATE::BaseException qw(:try);
use NATE::Log qw(log_global);
use Hostrec qw();
use San::SanClient::ClientUtil;
use NATE::Messenger qw(messenger_global);
use Digest::MD5 qw(md5_hex);

sub setup {
    my ($self, %args) = @_;

    # get all esx hypervisors with setup enabled on them
    my $esx_resources = $self->resource_list()->get_resources(category => "HYPERVISORS",
                                                              type => "esx",
                                                              data => {SETUP => 1});

    # create proc objects for each ESX host that will perform setup and return
    # the list of them to the caller
    my $balance_count = 0;
    my @procs = ();
    foreach my $esx_resource (@{$esx_resources}) {
        my $host = $esx_resource->get_field("NAME");
        my $proc = NATE::Process->new(codespec => sub {
                                          try {
                                              $self->_setup_single(resource => $esx_resource,
                                                                   balance => $balance_count++);
                                              $self->_setup_complete(name => $host);
                                          } otherwise {
                                              my $e = shift;
                                              $self->_setup_failure(name => $host);
                                              $e->throw();
                                          };
                                      },
                                      runid => "ESX-setup" . "/" . $host . '-setup',
                                      onexit => \&NATE::Process::on_exit_propagate_worst_result);
        push(@procs, $proc);
    }

    return @procs;
}

sub _setup_single {
    my ($self, %args) = @_;

    # given a single hypervisor, setup this hypervisor
    my $esx_host = delete($args{resource}) // NATE::BaseException->throw("Missing required parameter 'resource'");
    my $balance = delete($args{balance}) // 0;

    my $esx_host_name = $esx_host->get_field("NAME");
    my $protocol = $esx_host->get_field("PROTOCOL");
    my @protocols = ();
    if ($protocol =~ /iscsi/i) {
        push(@protocols, "iscsi");
    }
    if ($protocol =~ /fcp/i) {
        push(@protocols, "fcp");
    }

    my $num_vmdks_per_datastore = $esx_host->get_field("VMDKS_PER_DATASTORE");
    $num_vmdks_per_datastore = 1 if (not defined($num_vmdks_per_datastore));

    my $scsi_bus = $esx_host->get_field("SCSI_BUS");
    $scsi_bus = "lsilogic" if (not defined($scsi_bus));

    my $vmfs_datastore_percentage = $esx_host->get_field("VMFS_DATASTORE_PERCENTAGE");
    $vmfs_datastore_percentage = "100" if (not defined($vmfs_datastore_percentage));

    my $vmfs_lun_path_regex_filter = $esx_host->get_field("VMFS_LUN_PATH_REGEX");
    if (defined($vmfs_lun_path_regex_filter)) {
        $vmfs_lun_path_regex_filter = eval($vmfs_lun_path_regex_filter);
    } else {
        $vmfs_lun_path_regex_filter = undef;
    }

    # This filter controls which VNFS Datastores have VMDKs created in them and which dont
    my $vmfs_vmdk_lun_path_regex_filter = $esx_host->get_field("VMFS_VMDK_LUN_PATH_REGEX");
    if (defined($vmfs_vmdk_lun_path_regex_filter)) {
        $vmfs_vmdk_lun_path_regex_filter = eval($vmfs_vmdk_lun_path_regex_filter);
    } else {
        $vmfs_vmdk_lun_path_regex_filter = undef;
    }

    # get list of vservers that this client will talk to
    my @vserver_list = $self->{parent}->get_vservers_associated_to_client(-client_name => $esx_host_name);

    # we also need to obtain the list of cluster objects that this client is related to as well
    my $hash = $self->{parent}->get_vserver_client_relationship_hash();

    my @clusters = keys %{$hash->{$esx_host_name}};

    # get all VM resources for this esx host
    my $virtual_machines = $self->resource_list()->get_resources(category => "HYPERVISORS",
                                                                 type => "virtual_machines",
                                                                 data => {HYPERVISOR_RESOURCE => $esx_host_name});

    my @vm_names = map { $_->get_field("NAME") } @{$virtual_machines};

    my ($esx_host_obj) = $self->get_objects(type => "hypervisor",
                                            names => [$esx_host_name]);

    # reconnect to this host
    $esx_host_obj->reconnect(-disconnect_existing_conn => 0);
    if ($esx_host_obj->error()) {
        NATE::BaseException->throw("Unable to reconnect to ESX host '$esx_host_name'. " . $esx_host_obj->errmsg());
    }

    # default VMFS type to vmfs3 (ESX4.*) or vmfs5 (ESXi5.* later)
    my $vmfs_type = $esx_host->get_field("VMFS_TYPE");
    if (not length($vmfs_type)) {
        if ($esx_host_obj->get_os_version() =~ /^(5|6)\./) {
            $vmfs_type = "vmfs5";
        } else {
            $vmfs_type = "vmfs3";
        }
    }

    if (scalar(@vm_names)) {
        # power off all the virtual machines on this host
        $esx_host_obj->power_management(-action => "power_off",
                                        -vm_list => \@vm_names);

        if ($esx_host_obj->error()) {
            my $err = "Unable to power off VMs '@vm_names'. " . $esx_host_obj->errmsg();
            $esx_host_obj->power_management(-action => "power_on",
                                            -vm_list => \@vm_names,
                                            -dont_wait => 1);
            NATE::BaseException->throw($err);
        }
    }

    # rotate syslogs
    $esx_host_obj->rotate_logs();
    if ($esx_host_obj->error()) {
        my $err = "Unable to rotate logs for ESX host '$esx_host_name'. " . $esx_host_obj->errmsg();
        if (scalar(@vm_names)) {
            $esx_host_obj->power_management(-action => "power_on",
                                            -vm_list => \@vm_names,
                                            -dont_wait => 1);
        }
        NATE::BaseException->throw($err);
    }

    # If doing SAN, discover LUNs,
    my $esx_dev_mgr;
    my ($rdm_vmdk_requested_total, $vmfs_vmdk_requested_total) = (0,0);
    if (grep(/iscsi|fcp/, @protocols)) {
        # rescan for luns
        $esx_host_obj->discover(-num_expected_luns => 1);
        if ($esx_host_obj->error()) {
            my $err = "Unable to discover LUNs for ESX host '$esx_host_name' " . $esx_host_obj->errmsg();
            if (scalar(@vm_names)) {
                $esx_host_obj->power_management(-action => "power_on",
                                                -vm_list => \@vm_names,
                                                -dont_wait => 1);
            }
            NATE::BaseException->throw($err);
        }

        # fetch the NACL cluster objects
        my @nacl_cluster_objs = $self->get_objects(type => "node",
                                                   names => \@clusters,
                                                   obj_class => 'nacl');
        if ($self->error() or not scalar(@nacl_cluster_objs)) {
            my $err = "Failed to fetch NACL Cluster objects for clusters (@clusters)";
            if (scalar(@vm_names)) {
                $esx_host_obj->power_management(-action => "power_on",
                                                -vm_list => \@vm_names,
                                                -dont_wait => 1);
            }
            NATE::BaseException->throw($err);
        }

        # create an MPDevMgr (this will find all LUNs on the system)
        $esx_dev_mgr = $esx_host_obj->create_mpdevmgr(-nacl_cluster_objs => \@nacl_cluster_objs,
                                                      -vservers => \@vserver_list);
        if ($esx_host_obj->error()) {
            my $err = "Failed to create MPDevMgr " . $esx_host_obj->errmsg();
            if (scalar(@vm_names)) {
                $esx_host_obj->power_management(-action => "power_on",
                                                -vm_list => \@vm_names,
                                                -dont_wait => 1);
            }
            NATE::BaseException->throw($err);
        }

        # perform os specific validation on the discovered luns
        my ($os_specific_validation_errors, $os_specific_validation_warnings) = $esx_dev_mgr->os_specific_validation();
        if ($esx_dev_mgr->error()) {
            my $err = "Failed to perform os specific validation " . $esx_dev_mgr->errmsg();
            NATE::BaseException->throw($err);
        }
        if (scalar(@{$os_specific_validation_warnings})) {
            log_global()->warn("Detected SAN LUN Configuration warnings on $esx_host_name:\n\t" .
                                   join("\n", @{$os_specific_validation_warnings}));
        }
        if (scalar(@{$os_specific_validation_errors})) {
            my $err = "Detected SAN LUN Configuration errors on $esx_host_name:\n\t" .
                join("\n", @{$os_specific_validation_errors});
            NATE::BaseException->throw($err);
        }

        # verify that there are enough vmfs/rdm luns to go around
        try {
            ($rdm_vmdk_requested_total, $vmfs_vmdk_requested_total) =
                $self->_verify_vmfs_rdm_counts($esx_dev_mgr,
                                               $vmfs_lun_path_regex_filter,
                                               $virtual_machines,
                                               $num_vmdks_per_datastore,
                                               \@vserver_list);
        } otherwise {
            my $e = shift;
            if (scalar(@vm_names)) {
                $esx_host_obj->power_management(-action => "power_on",
                                                -vm_list => \@vm_names,
                                                -dont_wait => 1);
            }
            $e->throw();
        };

        # setup esx psp policies
        if (defined($esx_host->get_field("PSP_LIST")) and $esx_host->get_field("PSP_LIST") ne "") {
            try {
                &_set_psp_policies($self, $esx_host_obj, $esx_host, $balance);
            } otherwise {
                my $e = shift;
                if (scalar(@vm_names)) {
                    $esx_host_obj->power_management(-action => "power_on",
                                                    -vm_list => \@vm_names,
                                                    -dont_wait => 1);
                }
                $e->throw();
            };
        }

        # setup VxDMP IO policies if the object has a vx_io_policy established
        if ($esx_host_obj->{vx_io_policy}) {
            try {
                &_set_vxdmp_io_policies($self, $esx_host_obj, $esx_host);
            } otherwise {
                my $e = shift;
                if (scalar(@vm_names)) {
                    $esx_host_obj->power_management(-action => "power_on",
                                                    -vm_list => \@vm_names,
                                                    -dont_wait => 1);
                }
                $e->throw();
            };
        }

        # configure qfull settings after all luns have been discovered
        if ($esx_host->get_field("CONFIGURE_QFULL")) {
            $esx_host_obj->qfull_configure();
            if ($esx_host_obj->error()) {
                my $err = "Failed to configure qfull for host $esx_host_name " . $esx_host_obj->errmsg();
                NATE::BaseException->throw($err);
            }
        }
    }

    # create nfs datastores
    my @nfs_vmdks = ();
    if ($protocol =~ /nfs/i) {
        my $nfs_target_hash;
        try {
            $nfs_target_hash = $self->{parent}->NACL::AppBuilder::EnGenX::Util::get_nfs_target_relationship_hash(client_name => $esx_host_name);
        } otherwise {
            my $e = shift;
            if (scalar(@vm_names)) {
                $esx_host_obj->power_management(-action => "power_on",
                                                -vm_list => \@vm_names,
                                                -dont_wait => 1);
            }
            $e->throw();
        };

        # build out the nfs options datastructure up front
        # this will contain a sweet lookup table that we can search through when trying to
        # find matches for nfs_options
        my $all_nfs_options;
        try {
            $all_nfs_options = $self->NACL::AppBuilder::EnGenX::Util::generate_nfs_options(resource => $esx_host);
        } otherwise {
            my $e = shift;
            if (scalar(@vm_names)) {
                $esx_host_obj->power_management(-action => "power_on",
                                                -vm_list => \@vm_names,
                                                -dont_wait => 1);
            }
            $e->throw();
        };

        my @nfs_datastores = ();
        my %mount_itr_hash = ();
        # we should now have a list of volumes to mount and a list of lifs to mount each volume
        foreach my $cluster_name (keys %{$nfs_target_hash}) {
            # skip if this cluster does not match the filter
            next if (!$self->NACL::AppBuilder::EnGenX::Util::find_nfs_options_match(all_nfs_options => $all_nfs_options,
                                                                  cluster_name => $cluster_name));
            foreach my $vserver_name (keys %{$nfs_target_hash->{$cluster_name}->{vservers}}) {
                # skip if this vserver does not match the filter
                next if (!$self->NACL::AppBuilder::EnGenX::Util::find_nfs_options_match(all_nfs_options => $all_nfs_options,
                                                                      cluster_name => $cluster_name,
                                                                      vserver_name => $vserver_name));

                foreach my $lif_obj (@{$nfs_target_hash->{$cluster_name}->{vservers}->{$vserver_name}->{lif_objs}}) {
                    my $lif_address = $lif_obj->address();
                    my $lif_name = $lif_obj->lif();
                    my $lif_curr_node = $lif_obj->curr_node();

                    # skip if the lif address / lif name don't match
                    next if (!$self->NACL::AppBuilder::EnGenX::Util::find_nfs_options_match(all_nfs_options => $all_nfs_options,
                                                                          cluster_name => $cluster_name,
                                                                          vserver_name => $vserver_name,
                                                                          lif_address => $lif_address,
                                                                          lif_name => $lif_name));

                    foreach my $volume_name (keys %{$nfs_target_hash->{$cluster_name}->{vservers}->{$vserver_name}->{volumes}}) {
                        my $vol_obj = $nfs_target_hash->{$cluster_name}->{vservers}->{$vserver_name}->{volumes}->{$volume_name}->{vol_obj};
                        my $junction_path = $vol_obj->junction_path();
                        my $aggr_obj = $nfs_target_hash->{$cluster_name}->{aggr_objs}->{$vol_obj->aggregate()};
                        my $vol_curr_node = $aggr_obj->nodes()->[0];

                        # discover the affinity between the volume and the lif (local|remote) based
                        # on the node
                        my $actual_affinity;
                        if ($lif_curr_node eq $vol_curr_node) {
                            $actual_affinity = "local";
                        } else {
                            $actual_affinity = "remote";
                        }

                        # now that we have all our fields, lets do one final lookup
                        # to find the nfs options that we can use for this volume
                        my $nfs_options;
                        try {
                            $nfs_options = $self->NACL::AppBuilder::EnGenX::Util::find_nfs_options_match(all_nfs_options => $all_nfs_options,
                                                                                       cluster_name => $cluster_name,
                                                                                       vserver_name => $vserver_name,
                                                                                       lif_address => $lif_address,
                                                                                       lif_name => $lif_name,
                                                                                       volume_name => $volume_name,
                                                                                       junction_path => $junction_path,
                                                                                       volume_lif_affinity => $actual_affinity);
                        } otherwise {
                            my $e = shift;
                            if (scalar(@vm_names)) {
                                $esx_host_obj->power_management(-action => "power_on",
                                                                -vm_list => \@vm_names,
                                                                -dont_wait => 1);
                            }
                            $e->throw();
                        };

                        # skip this junction path if we could not find an nfs options match
                        next if (!$nfs_options);

                        # come up with a datastore name for this mount point
                        my $uuid = md5_hex("${cluster_name}__${vserver_name}__${volume_name}__${lif_name}");
                        $uuid = lc($uuid);
                        my $mount_point = "nate_nfs_datastore_${uuid}";
                        try {
                            $esx_host_obj->nfs_mount(server => $lif_address,
                                                     share_name => $junction_path,
                                                     datastore_name => $mount_point);
                        } otherwise {
                            my $e = shift;
                            if (scalar(@vm_names)) {
                                $esx_host_obj->power_management(-action => "power_on",
                                                                -vm_list => \@vm_names,
                                                                -dont_wait => 1);
                            }
                            $e->throw();
                        };

                        # create a unique sub directory in this mount point
                        if ($nfs_options->{create_unique_subdir}) {
                            $uuid = lc($uuid);
                            # chop off the first 24 chars
                            $uuid =~ s/^\S{24}//g;
                            my $unique_dir_mount_point = "[${mount_point}] ${uuid}";
                            $mount_point = $mount_point . "/" . $uuid;
                            $esx_host_obj->make_datastore_dir(dir => $unique_dir_mount_point);
                            if ($esx_host_obj->error()) {
                                my $err = "Failed to make unique subdir " . $esx_host_obj->errmsg();
                                if (scalar(@vm_names)) {
                                    $esx_host_obj->power_management(-action => "power_on",
                                                                    -vm_list => \@vm_names,
                                                                    -dont_wait => 1);
                                }
                                NATE::BaseException->throw($err);
                            }
                        }

                        # add this nfs datastore to a list of datastores to use later
                        push(@nfs_datastores, $mount_point);

                        # skip qtrees if this nfs option explicitly specified to not mount qtrees
                        next if (!$nfs_options->{mount_qtrees});

                        # iterate through all qtrees in this volume and attempt to mount each one as well
                        foreach my $qtree_obj (@{$nfs_target_hash->{$cluster_name}->{vservers}->{$vserver_name}->{volumes}->{$volume_name}->{qtree_objs}}) {
                            my $qtree_name = $qtree_obj->qtree();
                            try {
                                $nfs_options = undef;
                                $nfs_options = $self->NACL::AppBuilder::EnGenX::Util::find_nfs_options_match(all_nfs_options => $all_nfs_options,
                                                                                           cluster_name => $cluster_name,
                                                                                           vserver_name => $vserver_name,
                                                                                           lif_address => $lif_address,
                                                                                           lif_name => $lif_name,
                                                                                           volume_name => $volume_name,
                                                                                           qtree_name => $qtree_name,
                                                                                           junction_path => $junction_path,
                                                                                           volume_lif_affinity => $actual_affinity);
                            } otherwise {
                                my $e = shift;
                                if (scalar(@vm_names)) {
                                    $esx_host_obj->power_management(-action => "power_on",
                                                                    -vm_list => \@vm_names,
                                                                    -dont_wait => 1);
                                }
                                $e->throw();
                            };

                            # skip this qtree if we could not find an nfs options match
                            next if (!$nfs_options);

                            # come up with a datastore name for this mount point
                            my $uuid = md5_hex("${cluster_name}__${vserver_name}__${volume_name}__${qtree_name}__${lif_name}");
                            $uuid = lc($uuid);
                            $mount_point = "nate_nfs_datastore_${uuid}";
                            try {
                                $esx_host_obj->nfs_mount(server => $lif_address,
                                                         share_name => $junction_path . "/" . $qtree_name,
                                                         datastore_name => $mount_point);
                            } otherwise {
                                my $e = shift;
                                if (scalar(@vm_names)) {
                                    $esx_host_obj->power_management(-action => "power_on",
                                                                    -vm_list => \@vm_names,
                                                                    -dont_wait => 1);
                                }
                                $e->throw();
                            };

                            # create a unique sub directory in this mount point
                            if ($nfs_options->{create_unique_subdir}) {
                                $uuid = lc($uuid);
                                # chop off the first 24 chars
                                $uuid =~ s/^\S{24}//g;
                                my $unique_dir_mount_point = "[${mount_point}] ${uuid}";
                                $mount_point = $mount_point . "/" . $uuid;
                                $esx_host_obj->make_datastore_dir(dir => $unique_dir_mount_point);
                                if ($esx_host_obj->error()) {
                                    my $err = "Failed to make unique subdir " . $esx_host_obj->errmsg();
                                    if (scalar(@vm_names)) {
                                        $esx_host_obj->power_management(-action => "power_on",
                                                                        -vm_list => \@vm_names,
                                                                        -dont_wait => 1);
                                    }
                                    NATE::BaseException->throw($err);
                                }
                            }

                            # add this nfs datastore to a list of datastores to use later
                            push(@nfs_datastores, $mount_point);

                        }

                    }
                }
            }
        }

        # Create NFS vmdks if we have datastores to work with
        if (scalar(@nfs_datastores)) {
            my $nfs_vmdk_disk_type = $esx_host->get_field("NFS_VMDK_DISK_TYPE");
            my $vmdks_per_datastore = $esx_host->get_field("VMDKS_PER_DATASTORE");
            my $nfs_datastore_percentage = $esx_host->get_field("NFS_DATASTORE_PERCENTAGE");
            my $overprovision_nfs_datastores = $esx_host->get_field("OVERPROVISION_NFS_DATASTORES");
            $nfs_datastore_percentage = "50" if (not defined($nfs_datastore_percentage));

            my $nfs_lvm_obj = $esx_host_obj->create_lvm(-lvm_type => "nfs");
            if ($esx_host_obj->error()) {
                my $err = "Failed to create NFS LVM object " . $esx_host_obj->errmsg();
                if (scalar(@vm_names)) {
                    $esx_host_obj->power_management(-action => "power_on",
                                                    -vm_list => \@vm_names,
                                                    -dont_wait => 1);
                }
                NATE::BaseException->throw($err);
            }

            # create the nfs VMDK files in the datastores
            @nfs_vmdks = $nfs_lvm_obj->create_vmdks(datastores => \@nfs_datastores,
                                                    vmdks_per_datastore => $vmdks_per_datastore,
                                                    disk_format => $nfs_vmdk_disk_type,
                                                    over_provision_space => $overprovision_nfs_datastores,
                                                    datastore_usage_percentage => $nfs_datastore_percentage);
            if ($nfs_lvm_obj->error()) {
                my $err = "Failed to create NFS vmdks " . $nfs_lvm_obj->errmsg();
                if (scalar(@vm_names)) {
                    $esx_host_obj->power_management(-action => "power_on",
                                                    -vm_list => \@vm_names,
                                                    -dont_wait => 1);
                }
                NATE::BaseException->throw($err);
            }
        }
    }

    my @all_vmfs_vmdks;

    my $vmfs_lvm_obj = $esx_host_obj->create_lvm(-lvm => 'vmfs');
    if ($esx_host_obj->error()) {
        my $err = "Failed to create VMFS LVM object for host '$esx_host_name' " . $esx_host_obj->errmsg();
        if (scalar(@vm_names)) {
            $esx_host_obj->power_management(-action => "power_on",
                                            -vm_list => \@vm_names,
                                            -dont_wait => 1);
        }
        NATE::BaseException->throw($err);
    }

    my $rdm_lvm_obj = $esx_host_obj->create_lvm(-lvm => "rdm");
    if ($esx_host_obj->error()) {
        my $err = "Failed to create RDM LVM object for host '$esx_host_name' " . $esx_host_obj->errmsg();
        if (scalar(@vm_names)) {
            $esx_host_obj->power_management(-action => "power_on",
                                            -vm_list => \@vm_names,
                                            -dont_wait => 1);
        }
        NATE::BaseException->throw($err);
    }

    my $vvol_lvm_obj = $esx_host_obj->create_lvm(-lvm => 'vvol');
    if ($esx_host_obj->error()) {
        my $err = "Failed to create VVOL LVM object for host '$esx_host_name' " . $esx_host_obj->errmsg();
        if (scalar(@vm_names)) {
            $esx_host_obj->power_management(-action => "power_on",
                                            -vm_list => \@vm_names,
                                            -dont_wait => 1);
        }
        NATE::BaseException->throw($err);
    }

    # create datastores & vmfs vmdks if doing SAN
    if (grep(/iscsi|fcp/, @protocols)) {
        # find devices that we can use for VMFS datastore creation
        my @vmfs_devices = $esx_dev_mgr->get_filtered_devices(-vservers => \@vserver_list,
                                                              -device_index => $balance++,
                                                              -ontap_lun_ostype_regex => qr/vmware/i,
                                                              -ontap_lun_path_regex => $vmfs_lun_path_regex_filter,
                                                              # dont touch Protocol endpoint LUNs (vvolPE)
                                                              # The VPServer names all PE luns with the prefix "vvolPE-"
                                                              -ontap_lun_path_regex_exclude => qr/\/vvolPE\-|\/naa\./i,
                                                              -protocols => \@protocols,
                                                              -ontap_mode => "cmode",
                                                              -reserve_devices => 1,
                                                              -device_state => "free");
        if ($esx_dev_mgr->error() and $vmfs_vmdk_requested_total > 0) {
            # we did not find any luns to use for VMFS datastores and the user requested that we
            # create a certain number of vmfs vmdks, set err and abort
            my $err = "Failed to obtain VMFS devices for VMFS datastore creation " . $esx_dev_mgr->errmsg();
            if (scalar(@vm_names)) {
                $esx_host_obj->power_management(-action => "power_on",
                                                -vm_list => \@vm_names,
                                                -dont_wait => 1);
            }
            NATE::BaseException->throw($err);
        } else {
            # clear the possible error on the mpdevmgr
            $esx_dev_mgr->_clear_err();
        }

        # only make some datastores and vmdks if we have VMFS devices to play with
        if (scalar(@vmfs_devices)) {
            my @datastores = $vmfs_lvm_obj->create_datastores(-vmfs_devices => \@vmfs_devices,
                                                              -num_luns_per_dstore => 1,
                                                              -datastore_type => "one-to-one",
                                                              -vmfs_type => $vmfs_type,
                                                              -block_size => "auto",
                                                              -hostname => $esx_host_name);
            if ($vmfs_lvm_obj->error()) {
                my $err = "Failed to create VMFS datastores " . $vmfs_lvm_obj->errmsg();
                if (scalar(@vm_names)) {
                    $esx_host_obj->power_management(-action => "power_on",
                                                    -vm_list => \@vm_names,
                                                    -dont_wait => 1);
                }
                NATE::BaseException->throw($err);
            }

            # only create VMDKs in these datastores if we have virtual machines
            if (scalar(@vm_names)) {
                # filter out the vmfs_devices that we dont want to create VMDKs on
                # using $vmfs_vmdk_lun_path_regex_filter
                foreach my $vmfs_device (@vmfs_devices) {
                    # load the current state of this vmfs device
                    $vmfs_device->load();

                    # get the name of the datastore on this device
                    my $datastore_label = $vmfs_device->get_os_specific_data("datastore_label");

                    # determine the lun path
                    my $lun_path = $vmfs_device->ontap_lun_path();
                    if ($lun_path !~ $vmfs_vmdk_lun_path_regex_filter) {
                        # remove this datastore_label from @datastores
                        @datastores = grep {
                            $_ !~ /^$datastore_label$/;
                        } @datastores;
                    }
                }

                my $num_vmfs_vmdks_total;
                if ($num_vmdks_per_datastore <= 0) {
                    $num_vmfs_vmdks_total = scalar(@datastores);
                } else {
                    $num_vmfs_vmdks_total = scalar(@datastores) * $num_vmdks_per_datastore;
                }

                my $vmfs_vmdk_disk_type = $esx_host->get_field("VMFS_VMDK_DISK_TYPE");
                my $vmfs_vmdk_divisibility_byte_size = $esx_host->get_field("VMFS_VMDK_DIVISIBILITY_BYTE_SIZE") // 0;

                @all_vmfs_vmdks = $vmfs_lvm_obj->create_vmdks(-datastores => \@datastores,
                                                              -controller => $scsi_bus,
                                                              -datastore_type => "one-to-one",
                                                              -num_vmfs_vmdks => $num_vmfs_vmdks_total,
                                                              -datastore_usage_percentage => $vmfs_datastore_percentage,
                                                              -vmfs_vmdk_disk_type => $vmfs_vmdk_disk_type,
                                                              -reuse => 1, # reuse vmdks that may already exist in datastores
                                                              -vmfs_vmdk_divisibility_byte_size => $vmfs_vmdk_divisibility_byte_size
                                                          );
                if ($vmfs_lvm_obj->error()) {
                    my $err = "Failed to create VMFS VMDKs " . $vmfs_lvm_obj->errmsg();
                    if (scalar(@vm_names)) {
                        $esx_host_obj->power_management(-action => "power_on",
                                                        -vm_list => \@vm_names,
                                                        -dont_wait => 1);
                    }
                    NATE::BaseException->throw($err);
                }
            }
        }
    }

    # get and filter all of the vvol datastores
    my $vvol_datastores = $self->_get_vvol_datastores(esx_host_obj => $esx_host_obj);

    # create all of the vmdk files for vvol datastores
    my $all_vvol_vmdks = $self->_create_all_vvol_vmdks(esx_host => $esx_host,
                                                       esx_host_obj => $esx_host_obj,
                                                       vvol_lvm_obj => $vvol_lvm_obj,
                                                       vvol_datastores => $vvol_datastores,
                                                       controller => $scsi_bus,
                                                       virtual_machines => $virtual_machines);

    # spawn off subtests to add all vmdks to each guest
    my (@subtests);
    my $parallel_limit = $esx_host->get_field("VMDK_ADD_REMOVE_PARALLEL_LIMIT");

    foreach my $virtual_machine (@{$virtual_machines}) {
        my $virtual_machine_name = $virtual_machine->get_field("NAME");
        my $vmfs_count = $virtual_machine->get_field("VIRTUAL_DISKS") // 0;
        my $nas_count = $virtual_machine->get_field("NAS_DISKS") // 0;
        my $rdm_count = $virtual_machine->get_field("RAW_DISKS") // 0;
        my $iscsi_vvol_count = $virtual_machine->get_field("ISCSI_VVOL_DISKS") // 0;
        my $fcp_vvol_count   = $virtual_machine->get_field("FCP_VVOL_DISKS")   // 0;
        my $nfs_vvol_count   = $virtual_machine->get_field("NFS_VVOL_DISKS")   // 0;

        my $vm_scsi_bus = $scsi_bus;
        if ($virtual_machine->field_defined("SCSI_BUS") and
            $virtual_machine->get_field("SCSI_BUS") !~ /^\s*$/) {
            $vm_scsi_bus = $virtual_machine->get_field("SCSI_BUS");
        }

        my $rdm_lun_path_regex_filter = $virtual_machine->get_field("RAW_LUN_PATH_REGEX");
        if (defined($rdm_lun_path_regex_filter)) {
            $rdm_lun_path_regex_filter = eval($rdm_lun_path_regex_filter);
        } else {
            $rdm_lun_path_regex_filter = undef;
        }

        my @rdm_vmdks_for_this_guest;
        my @vmfs_vmdks_for_this_guest;
        my @nas_vmdks_for_this_guest;

        for (my $i = 0; $i < $vmfs_count; $i++) {
            push(@vmfs_vmdks_for_this_guest, shift(@all_vmfs_vmdks));
        }

        if (scalar(@nfs_vmdks)) {
            for (my $i = 0; $i < $nas_count; $i++) {
                push(@nas_vmdks_for_this_guest, shift(@nfs_vmdks));
            }
        }

        my (@iscsi_vvol_vmdks_for_this_guest, @fcp_vvol_vmdks_for_this_guest, @nfs_vvol_vmdks_for_this_guest);
        for (my $i = 0; $i < $iscsi_vvol_count; $i++) {
            push(@iscsi_vvol_vmdks_for_this_guest, shift(@{$all_vvol_vmdks->{iscsi}}));
        }

        for (my $i = 0; $i < $fcp_vvol_count; $i++) {
            push(@fcp_vvol_vmdks_for_this_guest, shift(@{$all_vvol_vmdks->{fcp}}));
        }

        for (my $i = 0; $i < $nfs_vvol_count; $i++) {
            push(@nfs_vvol_vmdks_for_this_guest, shift(@{$all_vvol_vmdks->{nfs}}));
        }

        my @rdm_devices;

        if ($rdm_count > 0 and $esx_dev_mgr) {
            # we need to parse out the hosttype for this virtual machine
            my $hostrec = Hostrec->new(id => $virtual_machine_name);
            my $hosttype = $hostrec->hosttype();
            my $ostype;
            # need to convert linux host type to 'linux'
            if ($hosttype =~ /linux/) {
                $ostype = "linux";
            } elsif ($hosttype =~ /solaris/i) {
                $ostype = "solaris";
            } else {
                $ostype = $hosttype;
            }

            # get a list of RDM devices that match these specifications
            # for this virtual machine
            @rdm_devices = $esx_dev_mgr->get_filtered_devices(-vservers => \@vserver_list,
                                                              -device_index => $balance++,
                                                              -protocols => \@protocols,
                                                              -ontap_mode => "cmode",
                                                              -ontap_lun_path_regex => $rdm_lun_path_regex_filter,
                                                              -device_count => $rdm_count,
                                                              -reserve_devices => 1,
                                                              -ontap_lun_ostype_regex => qr/$ostype/i,
                                                              -device_state => "free");
            if ($esx_dev_mgr->error()) {
                my $err = "Failed to get RDM devices for $virtual_machine_name " . $esx_dev_mgr->errmsg();
                if (scalar(@vm_names)) {
                    $esx_host_obj->power_management(-action => "power_on",
                                                    -vm_list => \@vm_names,
                                                    -dont_wait => 1);
                }
                NATE::BaseException->throw($err);
            }
        }

        my $add_vmdks_to_vm_subroutine = sub {
            # open a new connection to the esx client or else they would interfere with eachother
            $esx_host_obj->reconnect(-disconnect_existing_conn => 0);
            if ($esx_host_obj->error()) {
                NATE::BaseException->throw($esx_host_obj->errmsg());
            }

            # create the RDM vmdks if there are any devices in the array
            if (scalar(@rdm_devices) > 0) {
                @rdm_vmdks_for_this_guest = $rdm_lvm_obj->create_vmdks(-rdm_devices => \@rdm_devices,
                                                                       -controller => $vm_scsi_bus =~ /pvscsi|lsisas1068/ ? "lsilogic" : $vm_scsi_bus,
                                                                       -rdm_type => "physical",
                                                                       -vm_name => $virtual_machine_name);
                if ($rdm_lvm_obj->error()) {
                    NATE::BaseException->throw("Failed to create RDM VMDKs for $virtual_machine_name " . $rdm_lvm_obj->errmsg());
                }
            }

            # apply all vmdks to this guest's config file
            $esx_host_obj->add_vmdks_to_vm(-vm_name => $virtual_machine_name,
                                           -vmfs_vmdk_files => \@vmfs_vmdks_for_this_guest,
                                           -nfs_vmdk_files => \@nas_vmdks_for_this_guest,
                                           -rdm_vmdk_files => \@rdm_vmdks_for_this_guest,
                                           -vvol_vmdk_files => [@iscsi_vvol_vmdks_for_this_guest, @fcp_vvol_vmdks_for_this_guest, @nfs_vvol_vmdks_for_this_guest],
                                           -controller => $vm_scsi_bus,
                                           -preclean => 1);
            if ($esx_host_obj->error()) {
                NATE::BaseException->throw("Failed to add VMDKs to $virtual_machine_name " . $esx_host_obj->errmsg());
            }

            &Tharn::disconnect($esx_host_obj->{conn}, '-ok');
        };

        # wait for subtests to finish before we spawn off the next batch of subtests
        if ($parallel_limit !~ /unlimited/i and
            scalar(@subtests) >= $parallel_limit and
            $esx_host_obj->has_vma_host()) {
            # wait for these subtests to finish
            foreach (@subtests) {
                $_->wait();
                # if this failed, boot all the vms back up and abort
                if ($_->worst_result()) {
                    $esx_host_obj->power_management(-action => "power_on",
                                                    -vm_list => \@vm_names,
                                                    -dont_wait => 1);
                    NATE::BaseException->throw("Failed to add disks to a virtual machine");
                }
                $_->destroy();
            }

            # reset the subtest list
            @subtests = ();
        }

        my $proc = NATE::Process->new(codespec => $add_vmdks_to_vm_subroutine,
                                      runid => "${virtual_machine_name}-add_vmdks");
                                      #onexit => \&NATE::Process::on_exit_propagate_worst_result);
        $proc->start();

        # push onto the working subtest list
        push(@subtests, $proc);
    }

    # remove spare VMFS vmdks not in use, ie: all elements left in @all_vmfs_vmdks
    if (defined($vmfs_lvm_obj) and $vmfs_vmdk_requested_total > 0) {
        $vmfs_lvm_obj->destroy_spare_vmdks(-vmfs_vmdk_files => \@all_vmfs_vmdks);
        if ($vmfs_lvm_obj->error()) {
            my $err = "Failed to destroy spare VMFS VMDKs " . $vmfs_lvm_obj->errmsg();
            if (scalar(@vm_names)) {
                $esx_host_obj->power_management(-action => "power_on",
                                                -vm_list => \@vm_names,
                                                -dont_wait => 1);
            }
            NATE::BaseException->throw($err);
        }
    }

    if (scalar(@vm_names)) {
        # if there are any subtests still in the list, take care of them now
        my $error = 0;
        foreach (@subtests) {
            $_->wait();
            # if this failed, boot all the vms back up and abort
            if ($_->worst_result()) {
                $error = 1;
            }
            $_->destroy();
        }

        if ($error) {
            $esx_host_obj->power_management(-action => "power_on",
                                            -vm_list => \@vm_names,
                                            -dont_wait => 1);
            NATE::BaseException->throw("Failed to add disks to a virtual machine");
        }

        # turn on all virtual machines
        $esx_host_obj->power_management(-action => "power_on",
                                        -vm_list => \@vm_names);
        if ($esx_host_obj->error()) {
            NATE::BaseException->throw("Failed to power on VMs '@vm_names' " . $esx_host_obj->errmsg());
        }
    }

    $self->synchronize();

    # refresh each esx host
    $esx_host_obj->restart_mgmt_vmware_service();
    if ($esx_host_obj->error()) {
        NATE::BaseException->throw("Failed to restart mgmt-vmware service " . $esx_host_obj->errmsg());
    }

    $self->synchronize();

    sleep 15;

    # rescan vmfs datastores
    $esx_host_obj->vmfs_rescan();
    if ($esx_host_obj->error()) {
        NATE::BaseException->throw("Failed to rescan VMFS datastores " . $esx_host_obj->errmsg());
    }

    return;
}

sub _set_psp_policies {
    my ($self, $esx_client_obj, $esx_host_hash, $balance) = @_;

    # only operate on 4.0 and later machines
    if ($esx_client_obj->get_os_version() !~ /^[4-9]\./) {
        return;
    }

    # obtain the devmgr object
    my $esx_dev_mgr = $esx_client_obj->{obj}->{mpdevmgr};

    # get every device
    my @all_devices = $esx_dev_mgr->get_device_objs();

    # initialize the VM_PSP_LIST_RANDOM_SEED
    # set a rseed if one isnt specified
    if (not $esx_host_hash->field_defined("PSP_LIST_RANDOM_SEED") or
            $esx_host_hash->get_field("PSP_LIST_RANDOM_SEED") =~ /^\s*$/) {
        # generate one using the balance value for this esx host and the current time
        $esx_host_hash->{PSP_LIST_RANDOM_SEED} = time() + ($balance * $balance);
    }

    # set the seed for all rand() calls
    srand($esx_host_hash->get_field("PSP_LIST_RANDOM_SEED"));

    my @psp_list = split(/,/, $esx_host_hash->get_field("PSP_LIST"));
    my @tmp_psp_list = @psp_list;
    my $print_msg = 1;

    # shuffle the array
    &ClientUtil::randomize_array(\@tmp_psp_list);

    # foreach device, apply a randomized psp setting
    foreach my $device_obj (@all_devices) {
        # skip this lun if it is not controlled by NMP
        next if ($device_obj->mp_type() ne "NMP");

        if ($print_msg) {
            log_global()->comment("Setting PSP Policies on all LUNs with the following properties.\n" .
                                  "\tPSP_LIST = " . $esx_host_hash->get_field("PSP_LIST") . "\n".
                                  "\tPSP_LIST_RANDOM_SEED = " . $esx_host_hash->get_field("PSP_LIST_RANDOM_SEED"));
        }
        $print_msg = 0;

        # if the temp psp list is empty, populate it again
        if (scalar(@tmp_psp_list) == 0) {
            # update the empty temp list to be fully populated again
            @tmp_psp_list = @psp_list;

            # shuffle the array
            &ClientUtil::randomize_array(\@tmp_psp_list);
        }

        # shift off a psp policy to use for this lun
        my $psp_policy = shift(@tmp_psp_list);

        # obtain the naa handle for this lun
        my $naa_device = $device_obj->mp_device_name();

        # set the policy
        $esx_client_obj->{mp}->set_psp_policy(-naa_device => $naa_device,
                                              -psp_policy => $psp_policy,
                                              -native_cmd => 1);
        if ($esx_client_obj->{mp}->error()) {
            NATE::BaseException->throw("Failed to set PSP Policy on $naa_device " . $esx_client_obj->{mp}->error());
        }
    }

    return;
}

sub _set_vxdmp_io_policies {
    my ($self, $esx_client_obj, $esx_host) = @_;

    # only operate on ESXi systems
    if (not $esx_client_obj->has_vma_host()) {
        return;
    }

    # get every device rdm/vmfs/other
    # get devices from the mpdevmgr object
    my $esx_dev_mgr = $esx_client_obj->{obj}->{mpdevmgr};
    my @all_devices = $esx_dev_mgr->get_device_objs();

    my $io_policy = $esx_client_obj->{vx_io_policy};

    my $enclosures = {};
    # foreach device, apply a randomized psp setting
    foreach my $device_obj (@all_devices) {
        # skip this lun if it is not controlled by VxDMP
        next if ($device_obj->mp_type() !~ /VxDMP/i);

        # get the enclosure name from the object
        my $mp_device_name = $device_obj->mp_device_name();
        my $enclosure_name = $device_obj->get_os_specific_data("vxdmp_enclosure_name");
        if (not $enclosure_name) {
            log_global()->warn("Could not determine VxDMP enclosure name for " .
                               "$mp_device_name, skipping setting IO policy");
            next;
        }

        # store the enclosures in a hash to get rid of dups
        $enclosures->{$enclosure_name}++;
    }

    # if we found any enclosures, then apply the io policy on each one we found
    # this eliminates the possibility of duplicate calls to setting the io policy
    if (keys %{$enclosures}) {
        log_global()->comment("Setting VxDMP I/O Policy $io_policy on all enclosures");

        foreach my $enclosure_name (keys %{$enclosures}) {
            # set the vxdmp io policy
            $esx_client_obj->{mp}->set_vxdmp_io_policy(-enclosure => $enclosure_name,
                                                       -io_policy => $io_policy);
            if ($esx_client_obj->{mp}->error()) {
                NATE::BaseException->throw($esx_client_obj->{mp}->errmsg());
            }
        }
    }

    return;
}

sub _verify_vmfs_rdm_counts {
    my ($self, $esx_dev_mgr, $vmfs_lun_path_regex_filter, $virtual_machines, $num_vmdks_per_datastore, $vserver_list) = @_;

    my ($vmfs_vmdk_requested_total, $rdm_vmdk_requested_total);
    $vmfs_vmdk_requested_total = 0;
    $rdm_vmdk_requested_total = 0;

    my @all_rdm_devices;

    foreach my $virtual_machine (@{$virtual_machines}) {
        my $vmfs_vmdk_count = $virtual_machine->get_field("VIRTUAL_DISKS") // 0;
        my $rdm_vmdk_count = $virtual_machine->get_field("RAW_DISKS") // 0;

        $vmfs_vmdk_requested_total += $vmfs_vmdk_count;
        $rdm_vmdk_requested_total  += $rdm_vmdk_count;

        if ($rdm_vmdk_count > 0) {
            my $hostrec = Hostrec->new(id => $virtual_machine->get_field("NAME"));
            my $hosttype = $hostrec->hosttype();
            my $ostype;
            if ($hosttype =~ /linux/i) {
                $ostype = "linux";
            } elsif ($hosttype =~ /solaris/i) {
                $ostype = "solaris";
            } else {
                $ostype = $hosttype;
            }

            my $rdm_lun_path_regex_filter = $virtual_machine->get_field("RAW_LUN_PATH_REGEX");
            if (defined($rdm_lun_path_regex_filter)) {
                $rdm_lun_path_regex_filter = eval($rdm_lun_path_regex_filter);
            } else {
                $rdm_lun_path_regex_filter = undef;
            }

            my @rdm_devices = $esx_dev_mgr->get_filtered_devices(-ontap_mode => "cmode",
                                                                 -reserve_devices => 0,
                                                                 -device_count => $rdm_vmdk_count,
                                                                 -ontap_lun_ostype_regex => qr/$ostype/,
                                                                 -ontap_lun_path_regex => $rdm_lun_path_regex_filter,
                                                                 -vservers => $vserver_list);
            if ($esx_dev_mgr->error()) {
                NATE::BaseException->throw("Failed to get RDM devices " . $esx_dev_mgr->errmsg());
            }
            push(@all_rdm_devices, @rdm_devices);
        }
    }
    my @all_vmfs_devices;

    # if any vmfs vmdks were requested, then obtain them
    if ($vmfs_vmdk_requested_total > 0) {
        # obtain all vmfs devices on the devmgr
        @all_vmfs_devices = $esx_dev_mgr->get_filtered_devices(-ontap_mode => "cmode",
                                                               -reserve_devices => 0,
                                                               -ontap_lun_ostype_regex => qr/vmware/i,
                                                               -ontap_lun_path_regex => $vmfs_lun_path_regex_filter,
                                                               # dont touch Protocol endpoint LUNs (vvolPE) or vvol LUNs (naa.),
                                                               # these naming conventions are used by the VPServer
                                                               -ontap_lun_path_regex_exclude => qr/\/vvolPE\-|\/naa\./i,
                                                               -vservers => $vserver_list);
        if ($esx_dev_mgr->error()) {
            NATE::BaseException->throw("Failed to get VMFS devices " . $esx_dev_mgr->errmsg());
        }
    }

    my $vmfs_vmdk_actual_count = scalar(@all_vmfs_devices) * $num_vmdks_per_datastore;
    my $rdm_vmdk_actual_count = scalar(@all_rdm_devices);

    if ($vmfs_vmdk_actual_count < $vmfs_vmdk_requested_total) {
        NATE::BaseException->throw("Requested '$vmfs_vmdk_requested_total' VMFS VMDKs " .
                                   "when only '$vmfs_vmdk_actual_count' VMFS VMDKs are available");
    }

    if ($rdm_vmdk_actual_count < $rdm_vmdk_requested_total) {
        NATE::BaseException->throw("Requested '$rdm_vmdk_requested_total' RDM VMDKs " .
                                   "when only '$rdm_vmdk_actual_count' RDM VMDKs are available");
    }

    return ($rdm_vmdk_requested_total, $vmfs_vmdk_requested_total);
}

sub _get_vvol_datastores {
    my ($self, %args) = @_;

    my $esx_host_obj = $args{esx_host_obj};

    my @vvol_datastores;
    my $datastore_info = $esx_host_obj->list_all_datastores();
    foreach my $ds_info (@{$datastore_info->{vvol}}) {
        push(@vvol_datastores, $ds_info->{label});
    }

    # filter for vvol datastores we created
    my $datastores_by_protocol;
    foreach my $datastore (@vvol_datastores) {
        if ($datastore =~ /nate_iscsi_container_\d+_ds_\S+/) {
            push(@{$datastores_by_protocol->{iscsi}}, $datastore);
        } elsif ($datastore =~ /nate_fcp_container_\d+_ds_\S+/) {
            push(@{$datastores_by_protocol->{fcp}}, $datastore);
        } elsif ($datastore =~ /nate_nfs_container_\d+_ds_\S+/) {
            push(@{$datastores_by_protocol->{nfs}}, $datastore);
        }
    }
    return $datastores_by_protocol;
}

sub _create_all_vvol_vmdks {
    my ($self, %args) = @_;

    my $esx_host = $args{esx_host};
    my $esx_host_obj = $args{esx_host_obj};
    my $vvol_lvm_obj = $args{vvol_lvm_obj};
    my $virtual_machines = $args{virtual_machines};
    my $vvol_datastores = $args{vvol_datastores};
    my $path_prefix = $args{path_prefix} // "/vmfs/volumes/";
    my $scsi_bus = $args{scsi_bus} // "lsilogic";

    my $all_vvol_vmdks;
    my $vvol_vmdk_disk_type = $esx_host->get_field("VVOL_VMDK_DISK_TYPE");
    my $vvol_vmdk_divisibility_byte_size = $esx_host->get_field("VVOL_VMDK_DIVISIBILITY_BYTE_SIZE") // 0;
    my $vvol_datastore_percentage = $esx_host->get_field("VVOL_DATASTORE_PERCENTAGE") // 25;
    my @vm_names = map { $_->get_field("NAME") } (@{$virtual_machines});

    # create the directories for each vvol datastore
    # these top level directories are required to be able to create VMDK files in this datastore
    foreach my $protocol (keys %{$vvol_datastores}) {
        foreach my $datastore (@{$vvol_datastores->{$protocol}}) {
            #/vmfs/volumes/" .$datastore_uuid . "/" . $datastore . "_${esx_hostname}"
            $esx_host_obj->make_datastore_dir(dir => "[${datastore}] nate_vvol_vmdks");
            if ($esx_host_obj->error()) {
                if (scalar(@vm_names)) {
                    $esx_host_obj->power_management(-action => "power_on",
                                                    -vm_list => \@vm_names,
                                                    -dont_wait => 1);
                }
                NATE::BaseException->throw($esx_host_obj->errmsg());
            }
        }
    }

    # get the total number of vvol vmdks for this host
    my %vvol_disk_count;
    foreach my $virtual_machine (@{$virtual_machines}) {
        $vvol_disk_count{iscsi} += $virtual_machine->get_field("ISCSI_VVOL_DISKS");
        $vvol_disk_count{fcp} += $virtual_machine->get_field("FCP_VVOL_DISKS");
        $vvol_disk_count{nfs} += $virtual_machine->get_field("NFS_VVOL_DISKS");
    }

    # create all of the vvol vmdks for each protocol
    foreach my $protocol (keys %{$vvol_datastores}) {
        my $protocol_vvol_count = $vvol_disk_count{$protocol} // 0;
        my $protocol_datastores = $vvol_datastores->{$protocol};
        if ($protocol_datastores and $protocol_vvol_count) {
            my %create_vmdk_args = (datastores => $protocol_datastores,
                                    controller => $scsi_bus,
                                    num_vvol_vmdks => $protocol_vvol_count,
                                    datastore_usage_percentage => $vvol_datastore_percentage,
                                    vmdk_folder => "nate_vvol_vmdks",
                                    vvol_vmdk_disk_type => $vvol_vmdk_disk_type,
                                    reuse => 1, # reuse vmdks that may already exist in datastores
                                    vvol_vmdk_divisibility_byte_size => $vvol_vmdk_divisibility_byte_size);

            # create $protocol vvol vmdks and push it onto the list of $protocol vvol vmdks
            my @vvol_vmdks = $vvol_lvm_obj->create_vmdks(%create_vmdk_args);
            $all_vvol_vmdks->{$protocol} = \@vvol_vmdks;
            if ($vvol_lvm_obj->error()) {
                if (scalar(@vm_names)) {
                    $esx_host_obj->power_management(-action => "power_on",
                                                    -vm_list => \@vm_names,
                                                    -dont_wait => 1);
                }
                NATE::BaseException->throw("Failed to create $protocol VVOL VMDKs " . $vvol_lvm_obj->errmsg());
            }
        }
    }

    return $all_vvol_vmdks;
}

sub _setup_failure {
    my ($self, %data) = @_;
    messenger_global()->put(endpoint => $self->{endpoint},
                            type     => "NACL/AppBuilder/EnGenX/HyperVisor/ESX/Setup/Failure",
                            message  => \%data);

    $self->sync_rollback();
}

sub _setup_complete {
    my ($self, %data) = @_;
    messenger_global()->put(endpoint => $self->{endpoint},
                            type     => "NACL/AppBuilder/EnGenX/HyperVisor/ESX/Setup/Complete",
                            message  => \%data);
}

1;
