#
# Copyright (c) 2015 NetApp, Inc., All Rights Reserved.
#   Any other use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

package NACL::AppBuilder::EnGenX::HyperVisor::HyperV::Cleanup;

use strict;
use warnings;

use NATE::Process;
use NATE::BaseException qw(:try);
use NATE::Log qw(log_global);
use NATE::Messenger qw(messenger_global);

sub cleanup {
    my ($self, %args) = @_;

    # get all hyperv hypervisors with cleanup enabled on them
    my $hyperv_resources = $self->resource_list()->get_resources(category => "HYPERVISORS",
                                                                 type => '^hyperv$',
                                                                 data => {CLEANUP => 1});

    # create proc objects for each HYPERV host that will perform cleanup and return
    # the list of them to the caller
    my $balance_count = 0;
    my @procs = ();
    foreach my $hyperv_resource (@{$hyperv_resources}) {
        my $host = $hyperv_resource->get_field("NAME");
        my $proc = NATE::Process->new(codespec => sub {
                                          try {
                                              $self->_cleanup_single(resource => $hyperv_resource,
                                                                   balance => $balance_count++);
                                              $self->_cleanup_complete(name => $host);
                                          } otherwise {
                                              my $e = shift;
                                              $self->_cleanup_failure(name => $host);
                                              $e->throw();
                                          };
                                      },
                                      runid => "HyperV-cleanup" . "/" . $host . '-cleanup',
                                      onexit => \&NATE::Process::on_exit_propagate_worst_result);
        push(@procs, $proc);
    }

    return @procs;
}

sub _cleanup_single {
    my ($self, %args) = @_;

    # given a single hypervisor, cleanup this hypervisor
    my $hyperv_host = delete($args{resource}) // NATE::BaseException->throw("Missing required parameter 'resource'");
    my $balance = delete($args{balance}) // 0;

    my $hyperv_host_name = $hyperv_host->get_field("NAME");
    my $protocol = $hyperv_host->get_field("PROTOCOL");

    # get all VM resources for this esx host
    my $virtual_machines = $self->resource_list()->get_resources(category => "HYPERVISORS",
                                                                 type => "virtual_machines",
                                                                 data => {HYPERVISOR_RESOURCE => $hyperv_host_name});

    my @vm_names = map { $_->get_field("NAME") } @{$virtual_machines};

    my ($hyperv_host_obj) = $self->get_objects(-type => "hypervisor",
                                               -names => [$hyperv_host_name]);

    # get list of vservers that this client will talk to
    my @vserver_list = $self->{parent}->get_vservers_associated_to_client(-client_name => $hyperv_host_name);

    # reconnect to this host
    $hyperv_host_obj->reconnect(-disconnect_existing_conn => 0);
    if ($hyperv_host_obj->error()) {
        NATE::BaseException->throw("Unable to reconnect to HyperV host '$hyperv_host_name' " . $hyperv_host_obj->errmsg());
    }

    # power off all the virtual machines on this host
    $hyperv_host_obj->power_management(-action => "shutdown",
                                       -vm_list => \@vm_names);
    if ($hyperv_host_obj->error()) {
        NATE::BaseException->throw("Unable to power off vms " . $hyperv_host_obj->errmsg());
    }

    # need to rescan for devices just in case
    $hyperv_host_obj->discover();
    if ($hyperv_host_obj->error()) {
        NATE::BaseException->throw("Unable discover devices " . $hyperv_host_obj->errmsg());
    }

    # make a devmgr (scope it to only the vservers that are a part of this test)
    my $devmgr_obj = $hyperv_host_obj->create_devmgr(-target_list => \@vserver_list,
                                                     -lun_path_regex => undef,
                                                     -vhd_lun_path_regex => undef);
    if ($hyperv_host_obj->error()) {
        NATE::BaseException->throw("Unable to create device manager object " . $hyperv_host_obj->errmsg());
    }

    # foreach vm
    ## get passthrough disks from vms
    ## remove passthrough disks from vms
    ## get vhd devices from vm
    ## remove vhd files from vm
    ## for each vhd lun, cleanup vhds, and destroy partition

    foreach my $virtual_machine (@{$virtual_machines}) {
        # get passthrough disks from the this VM
        my $vm_name = $virtual_machine->get_field("NAME");
        my @pt_devices = $hyperv_host_obj->get_passthrough_devices_from_vm(-vm_name => $vm_name);
        if ($hyperv_host_obj->error()) {
            NATE::BaseException->throw("Unable to obtain passthrough devices from VM " . $hyperv_host_obj->errmsg());
        }

        if (scalar(@pt_devices) > 0) {
            $hyperv_host_obj->remove_passthrough_devices_from_vm(-vm_name => $vm_name,
                                                                 -passthrough_devices => \@pt_devices);
            if ($hyperv_host_obj->error()) {
                NATE::BaseException->throw("Failed to remove passthrough devices from VM " . $hyperv_host_obj->errmsg());
            }
        }

        my $vhd_relationship_hash = $hyperv_host_obj->get_vhd_devices_from_vm(-vm_name => $vm_name);
        if ($hyperv_host_obj->error()) {
            NATE::BaseException->throw("Failed to obtain VHD devices from VM " . $hyperv_host_obj->errmsg());
        }

        my @vhd_files = keys %{$vhd_relationship_hash};
        if (scalar(@vhd_files) > 0) {
            $hyperv_host_obj->remove_vhd_files_from_vm(-vm_name => $vm_name,
                                                       -vhd_files => \@vhd_files);
            if ($hyperv_host_obj->error()) {
                NATE::BaseException->throw("Failed to remove VHD devices from VM " . $hyperv_host_obj->errmsg());
            }

            foreach my $vhd_device (values %{$vhd_relationship_hash}) {
                # skip vhd devices that dont have any more vhds
                if ($vhd_device->{vhd_count} == 0) {
                    next;
                }

                # destroy all vhds on this device
                $vhd_device->destroy_all_vhds();
                if ($vhd_device->error()) {
                    NATE::BaseException->throw("Failed to destroy VHD files from VHD device " . $vhd_device->errmsg());
                }

                # destroy the partition
                $vhd_device->destroy_partition();
                if ($vhd_device->error()) {
                    NATE::BaseException->throw("Failed to destroy partition on VHD device " . $vhd_device->errmsg());
                }
            }
        }
    }

    # remove empty drive slots
    $hyperv_host_obj->remove_empty_drive_slots_from_vms(-vm_list => \@vm_names);
    if ($hyperv_host_obj->error()) {
        NATE::BaseException->throw("Failed to remove empty drive slots " . $hyperv_host_obj->errmsg());
    }

    # power on all the virtual machines that we shut down
    $hyperv_host_obj->power_management(-action => "power_on",
                                       -vm_list => \@vm_names);
    if ($hyperv_host_obj->error()) {
        NATE::BaseException->throw("Unable to power on VMs" . $hyperv_host_obj->errmsg());
    }

    return;
}


sub _cleanup_failure {
    my ($self, %data) = @_;
    messenger_global()->put(endpoint => $self->{endpoint},
                            type     => "NACL/AppBuilder/EnGenX/HyperVisor/HyperV/Cleanup/Failure",
                            message  => \%data);
}

sub _cleanup_complete {
    my ($self, %data) = @_;
    messenger_global()->put(endpoint => $self->{endpoint},
                            type     => "NACL/AppBuilder/EnGenX/HyperVisor/HyperV/Cleanup/Complete",
                            message  => \%data);
}

1;
