# $Id: //depot/prod/test/main/lib/NACL/AppBuilder/EnGenX/UnitTest/Utils.pm#1 $

#
# Copyright (c) 2012 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary UnitTest helper repository
## @author dl-san-automation
## @status private
## @pod here

=head1 NAME

NACL::AppBuilder::EnGenX::UnitTest_Utils

=head1 DESCRIPTION

NACL::AppBuilder::EnGenX::UnitTest_Utils is a common library to host
procedures that are commonly used by unit tests.

The code was originaly hosted in several unit test thpls for
EnGenX but need to be moved outside of each for cleanness and
for re-usability.

=cut

# Package
package NACL::AppBuilder::EnGenX::UnitTest::Utils;

# Compiler directives.
use strict;

# NetApp's module imports
use Tharn;
use Error;
use ParamsMgr;
use NATE::Util::DataTypes qw(TRUE FALSE);

# EnGenX Libraries
use NACL::AppBuilder::EnGenX::Resource;
use NACL::AppBuilder::EnGenX::ResourceList;
use NACL::AppBuilder::EnGenX::EnGenXUtil;
use NACL::AppBuilder::EnGenX::Exceptions::SubProcessFailure qw(:try);


# Perl's public module imports
use Data::Dumper;
use DateTime;

# Global Variables.
use vars qw(
    @ISA
);

#
# Inherit from the Error class
#
@ISA = qw( Error );

#==============================================================================
# Method:    create_resource_list
# Objective:  Creates an instance of a ResourceList object that holds a Resource
#             object that will be build with the default EnGenX flags and
#             input key-value pairs that the user passes to make it usefull for
#             different types.
#
# Details:    Steps: 1) Get the input hash.
#                    2) Define the default data structure.
#                    3) Merge the sent information with the default data.
#                    4) Create a resource with the information
#                    5) Return the ResourceList object ref.
# Inputs:     count
#                 The number of resources the list will hold
#             category
#                 The key where the resources will be held. The category of a
#                 resource.
#             data_add
#                 The particular information, data that this Resource list will
#                  have. I should be particular data for the type of resources
#                  that are been created. e.g. Luns need VOLUME, SIZE, etc or
#                  Lifs need ADRESS, NETAMASK, etc.
# Outputs:    An instance of a ResourceList containing 'cout' Resources with
#             specific keys-values provided by add_data.
# Exceptions: No exceptions have been implemented.
#==============================================================================
sub create_resource_list {
    $Log->enter();
    my %args = @_;

    # LOCALS
    my $rl_obj = defined($args{'resource_list'}) ? $args{'resource_list'} :
                                new NACL::AppBuilder::EnGenX::ResourceList();;
    my $count = $args{'count'};
    my $add_data = $args{'add_data'};
    my $name = defined($args{'add_data'}->{NAME}) ?
                            delete($args{'add_data'}->{NAME}) : "rsrc";

    my $data_struct = {
            '_reuse'          => 'reuse',
            '_cleanup_policy' => "new",
            '_interface'      => "zapi",
            '_load_balance'   => 0,
            '_setup'          => 1,
            '_cleanup'        => 1,
            '_config'         => 1,
            '_deconfig'       => 1,
            '_validate'       => 1,
            '_precleanup'     => 1,
            '_validate_done'  => 0,
            '_config_done'    => 0,
            '_setup_done'     => 0,
            '_deconfig_done'  => 0,
            '_cleanup_done'   => 0,
            '_precleanup_done'=> 0,
        };

    for (my $i = 1; $i <= $count; $i++) {
        my $rsrc = Clone::clone($data_struct);
        $rsrc->{NAME} = $name."_".$i;
        while (my ($add_key, $add_value) = each(%$add_data)) {
            if (!defined($rsrc->{$add_key})){
                $rsrc->{$add_key} = $add_value;
            }
        }
        $rl_obj->add_resource(category => $args{category},
                               type     => $rsrc->{_type},
                               data     => $rsrc
                        );
    }

    $Log->exit();
    return $rl_obj;
}
#==============================================================================
# Method:    subprocess_call
# Objective:  This method is used only for Unit Testing.
#             This is the CodeString that the unit test will be calling when
#             there is the need to test the functionality of this library.
#             Mimics the behavior of the subprocess_create_call or
#             subprocess_delete_call found in any EnGenX::Controller component
# Details:    Steps: 1)   Tell Parent waiting on resource
#                    2.a) if parent sends new resource(s) continue to step 3.
#                    2.b) if parent sends 'exit' then create pass result and
#                         die.
#                    3)   Simulates that is doing something (nothing happens).
#                    4)   send message to parent with updated resource(s) and
#                         tell parent waiting on resource again.
#                    5)   Repeat steps 2-4 until 2.b occurs or exceptions out.
# Inputs:     Does NOT take inputs.
# Outputs:    Does NOT return anything.  Instead creates a result object based
#             on whether an error occurs while creating the volumes in the
#             list. If uncaught errors occur then a Fail is made otherwise a
#             Pass is made and on die the result is passed back to the parent
#             process.
# Exceptions: This method throws an exceptions when a conditions is meet.
#             This exception is thrown just to verify the correctnes of the
#             try and catch in the public methods of this library.
#==============================================================================
sub subprocess_call {
    my (%args) = @_;

    use NATE::STAF::Peer qw($Parent);
    use NATE::Error qw(:try);
    my $message = '';

    # Tell parent proc is ready for execution.
    $Log->trace("---------------------------------------------------");
    $Log->trace("Letting the parent know the process is ready for execution");
    my $entry = $Parent->message_put(
        type    => "Response",
        message => {'status' => "waiting"}
    );

    my $exit_flag = 1;
    try {
        while ($exit_flag) {
            $Log->trace("Waiting for message...");
            $message = '';
            $message = $Parent->message_get(type => "WhatToDo");

            if ($message =~ /exit/) {
                $Log->trace("Breaking from the loop...");
                $exit_flag = 0;
            } elsif (ref($message) =~ /ARRAY/) {
                $Log->trace("GOT AN ARRAY REF, PROCESSING... ");
                # Locals
                # LOOP THROUGH THE COMPLETE LIST OF RESOURCES
                foreach my $resource (@{$message}) {
                    # Move on if no setup was requested for this resource.
                    next if ($resource->{'_setup'} == 0);
                    # EXECUTE COMMAND
                    try {
                        my $name = $resource->{'NAME'};
                        $Log->comment("DOING SOMETHING WITH $name");
                    }
                    catch NATE::Error::Simple with {
                        my $exception = shift;
                        my $error_msg = "ERROR thrown while processing EACH ".
                                "resource at code string: EnGenX::SubProcess-".
                                ">_subprocess_call: \n$exception";
                        $Log->error($error_msg);
                        $Parent->message_put(
                                   type    => "Response",
                                   message => {
                                       'status' => "failed",
                                       'resources' => $message,
                                       'error_msg' => $error_msg
                                    }
                        );
                        my $result
                            = NATE::Result::Fail->new("Subprocess Fail");
                        die $result;
                    };

                    # UPDATE THE SET_UP FLAG
                    #$resource->set_field('setup_done', TRUE);
                    # TODO: Not yet supported, getting this error:
                    #      '-stacktrace' => 'Can\'t call method "set_field" on
                    #      unblessed reference at
                    # /u/jcoca/p4/test/lib/EnGenX/UnitTest_Utils.pm line 217.
                    $resource->{_setup_done}= 1;
                }

                # Return result and get more work
                my $entry = $Parent->message_put(
                    type    => "Response",
                    message => {
                        'status'    => "waiting",
                        'resources' => $message
                    }
                );
            } else {
                $Log->trace("FOUND AN ELSE MESSAGE: \n" . Dumper($message));
            }
        }
    }
    otherwise {
        my $exception = shift;
        my $error_msg = "Exception thrown WHILE waiting for parent process to send ".
                        "information to process to code string: EnGenX::SubProcess".
                        "->_subprocess_call: \n". $exception->text();
        $Log->trace($error_msg);
        $Parent->message_put(
            type    => "Response",
            message => {
                'status'    => "failed",
                'resources' => $message,
                'error_msg' => $error_msg
            }
        );
        my $result = NATE::Result::Fail->new("Subprocess Fail");
        die $result;
    };

    # RESULT MESSAGE
    my $result = NATE::Result::Pass->new("Subprocess Info");
    die $result;

}    # _subprocess_call

#==============================================================================
# Method:    subprocess_exception_call
# Objective:  This method is used only for Unit Testing. It will throw an
#             predetermined exception to verify that is catch.
# Details:    Steps: 1)   Tell Parent waiting on resource
#                    2.a) if parent sends new resource(s) continue to step 3.
#                    2.b) if parent sends 'exit' then create pass result and
#                         die.
#                    3)   Simulates that is doing something and randomly throw
#                         an exception.
#                    4)  Catch the exception and inform parent that an error has
#                        been encountered.
# Inputs:     Does NOT take inputs.
# Outputs:    A random exception will be thrown.
# Exceptions: This method throws an exceptions when a conditions is meet.
#             This exception is thrown just to verify the correctnes of the
#             try and catch in the public methods of this library.
#==============================================================================
sub subprocess_exception_call {
    my (%args) = @_;

    use NATE::STAF::Peer qw($Parent);
    use NATE::Error qw(:try);
    my $message = '';

    # Tell parent proc is ready for execution.
    $Log->trace("---------------------------------------------------");
    $Log->trace("Letting the parent know the process is ready for execution");
    my $entry = $Parent->message_put(
        type    => "Response",
        message => {'status' => "waiting"}
    );

    my $exit_flag = 1;
    try {
        while ($exit_flag) {
            $Log->trace("Waiting for message...");
            $message = '';
            $message = $Parent->message_get(type => "WhatToDo");

            if ($message =~ /exit/) {
                $Log->trace("Breaking from the loop...");
                $exit_flag = 0;
            } elsif (ref($message) =~ /ARRAY/) {
                $Log->trace("GOT AN ARRAY REF, PROCESSING... ");

                # Locals
                # LOOP THROUGH THE COMPLETE LIST OF RESOURCES
                foreach my $resource (@{$message}) {
                    # Move on if no setup was requested for this resource.
                    next if ($resource->{'_setup'} == 0);
                    # EXECUTE COMMAND
                    try {
                        my $name = $resource->{'NAME'};
                        $Log->comment("DOING SOMETHING WITH $name");

                        # Throw a random Exception to verify that it's caught
                        if ($name =~ /lun_9/) {
#                         if ($name eq "lun_1"|$name eq "lun_"|$name eq "lun_6"|$name eq "lun_9") {
                             throw NATE::Error::Simple("Error Oops!");
                        }
                    }
                    catch NATE::Error::Simple with {
                        my $exception = shift;
                        my $error_msg = "ERROR thrown while processing EACH".
                                " resource at code string: NACL::AppBuilder::EnGenX::SubProcess-".
                                ">_subprocess_call: \n$exception";
                        $Log->error($error_msg);
                        $Parent->message_put(
                                   type    => "Response",
                                   message => {
                                       'status' => "failed",
                                       'resources' => $message,
                                       'error_msg' => $error_msg,
                                    }
                        );
                        my $result
                            = NATE::Result::Fail->new("Subprocess Fail");
                        die $result;
                    };

                    # UPDATE THE SET_UP FLAG
                    $resource->{'_setup_done'} = 1;
                }

                # Return result and get more work
                my $entry = $Parent->message_put(
                    type    => "Response",
                    message => {
                        'status'    => "waiting",
                        'resources' => $message,
                    }
                );
            } else {
                $Log->trace("FOUND AN ELSE MESSAGE: \n" . Dumper($message));
            }
        }
    }
    otherwise {
        my $exception = shift;
        my $error_msg = "Exception thrown WHILE waiting for parent process to send ".
                        "information to process to code string: NACL::AppBuilder::EnGenX::SubProcess".
                        "->_subprocess_call: \n". $exception->text();
        $Log->trace($error_msg);
        $Parent->message_put(
            type    => "Response",
            message => {
                'status'    => "failed",
                'resources' => $message,
                'error_msg' => $error_msg,
            }
        );
        my $result = NATE::Result::Fail->new("Subprocess Fail");
        die $result;
    };

    # RESULT MESSAGE
    my $result = NATE::Result::Pass->new("Subprocess Info");
    die $result;

}# subprocess_exception_call

#==============================================================================
# Method:    subprocess_exception_call
# Objective:  This method is used only for Unit Testing. It will throw an
#             predetermined exception to verify that is catch.
# Details:    Steps: 1)   Tell Parent waiting on resource
#                    2.a) if parent sends new resource(s) continue to step 3.
#                    2.b) if parent sends 'exit' then create pass result and
#                         die.
#                    3)   Simulates that is doing something and randomly throw
#                         an exception.
#                    4)  Catch the exception and inform parent that an error has
#                        been encountered.
# Inputs:     Does NOT take inputs.
# Outputs:    A random exception will be thrown.
# Exceptions: This method throws an exceptions when a conditions is meet.
#             This exception is thrown just to verify the correctnes of the
#             try and catch in the public methods of this library.
#==============================================================================
sub subprocess_ips_call {
    my (%args) = @_;

    # Validate Global Params

    use NATE::STAF::Peer qw($Parent);
    use NATE::Error qw(:try);
    my $message = '';

    # Tell parent proc is ready for execution.
    $Log->trace("---------------------------------------------------");
    $Log->trace("Letting the parent know the process is ready for execution");
    my $entry = $Parent->message_put(
        type    => "Response",
        message => {'status' => "waiting"}
    );

    my $exit_flag = 1;
    try {
        while ($exit_flag) {
            $Log->trace("Waiting for message...");
            $message = '';
            $message = $Parent->message_get(type => "WhatToDo");

            if ($message =~ /exit/) {
                $Log->trace("Breaking from the loop...");
                $exit_flag = 0;
            } elsif (ref($message) =~ /ARRAY/) {
                $Log->comment("GOT AN ARRAY REF, PROCESSING... ");

                # Locals
                # LOOP THROUGH THE COMPLETE LIST OF RESOURCES
                foreach my $resource (@{$message}) {
                    # Move on if no setup was requested for this resource.
                    if (   ($resource->{'_setup'} == 0)
                        || ($resource->{'_setup_done'} == 1))
                    {
                        $Log->comment("Setup was skipped because _setup="
                                . $resource->{'_setup'}
                                . " _setup_done="
                                . $resource->{'_setup_done'});
                        next;
                    }
                    # If key IPS is present in the resource, means that user passed
                    # an IP address range from where to chose the address. So, the
                    # Resource needs to pass the array to the parent for him to
                    # manage the addresses and avoid collisions.

                    if (defined($resource->{ips}) && ($resource->{ADDRESS} eq''))
                    {
                        $Log->comment("COKE, FOUND IPS");
                        # Request and IP Address
                        $Parent->message_put(
                            type    => "Response",
                            message => {
                                'status'    => "request_ip",
                                'resources' => $resource
                            }
                        );

                        while(1){
                            $Log->comment("COKE, WAITING FOR ANSWER:");
                            if (my $my_message = $Parent->message_get(type => "ip_address")){
                                $Log->comment("COKE, MESSAGE:" . $my_message->[0]->{NAME});
                                $resource = $my_message->[0];
                                last;
                            }
                        }
                    }

                    # EXECUTE COMMAND
                    try {
                        my $name = $resource->{'NAME'};
                        my $ip = $resource->{'ADDRESS'};
                        $Log->comment("DOING SOMETHING WITH $name and IP address: $ip");

                        # Throw a random Exception to verify that it's caught
                        #if ($name eq "lun_10"|$name eq "lun_9") {{
                        #    throw NATE::Error::Simple("Error Oops!");
                        # }
                    }
                    catch NATE::Error::Simple with {
                        my $exception = shift;
                        my $error_msg = "ERROR thrown while processing EACH".
                                " resource at code string: NACL::AppBuilder::EnGenX::SubProcess-".
                                ">_subprocess_call: \n$exception";
                        $Log->error($error_msg);
                        $Parent->message_put(
                                   type    => "Response",
                                   message => {
                                       'status' => "failed",
                                       'resources' => $message,
                                       'error_msg' => $error_msg
                                       }
                        );
                        my $result
                            = NATE::Result::Fail->new("Subprocess Fail");
                        die $result;
                    };

                    # UPDATE THE SET_UP FLAG
                    $resource->{'_setup_done'} = 1;
                }

                # Return result and get more work
                my $entry = $Parent->message_put(
                    type    => "Response",
                    message => {
                        'status'    => "waiting",
                        'resources' => $message
                    }
                );
            } else {
                $Log->trace("FOUND AN ELSE MESSAGE: \n" . Dumper($message));
            }
        }
    }
    otherwise {
        my $exception = shift;
        my $error_msg = "Exception thrown WHILE waiting for parent process to send ".
                        "information to process to code string: NACL::AppBuilder::EnGenX::SubProcess".
                        "->_subprocess_call: \n". $exception->text();
        $Log->trace($error_msg);
        $Parent->message_put(
            type    => "Response",
            message => {
                'status'    => "failed",
                'resources' => $message,
                'error_msg' => $error_msg
            }
        );
        my $result = NATE::Result::Fail->new("Subprocess Fail");
        die $result;
    };

    # RESULT MESSAGE
    my $result = NATE::Result::Pass->new("Subprocess Info");
    die $result;

}    # subprocess_ips_call

1;
