#
# Copyright (c) 2015 NetApp, Inc., All Rights Reserved.
#   Any other use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

package NACL::AppBuilder::EnGenX::XgenFormatter;

use strict;
use warnings;

use NATE::Log qw(log_global);
use NATE::ParamSet qw(param_global);

use Data::Dumper;

use parent qw( NACL::Service::Xgen::Formatter );

my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();

sub new {
    my ($class, %args) = @_;

    my $self = $class->SUPER::new(%args);

    $self->generate_resource_translation_hash();

    $self->generate_resource_list_translation_hash();

    return $self;
}

sub apply_format_to_resource {
    my ($self, $resource) = @_;

    # Grab the metadata from the resource and delete it off the resource.
    my $metadata = delete $resource->{metadata};

    # based on engenx object output decide what parameters need to be
    # modified and which missing parameters are generated by engenx

    foreach my $key (keys %$resource) {
        # store the EnGenX translation of the current key
        my $key_translation = $self->{resource_translation}->{ref($resource)}->{$key};
        my $common_translation = $self->{resource_translation}->{common}->{$key};

        if (defined($key_translation)) {
            # store the value for the current key as the value for the translated key
            # and delete the original key from the resource
            $resource->{$key_translation} = delete($resource->{$key});
        } elsif (defined($common_translation)) {
            $resource->{$common_translation} = delete($resource->{$key});
        }
    }

    # EnGenX controller specific settings
    # Allows users to pass per objects setting in the format type.attribute.
    # Example: NACL.AppBuilder.Xgen.Resource.Lun.SETUP=0
    my $type = ref($resource);
    $type =~ s/::/\./g;
    unless (defined $self->{engenx_controller_settings}->{$type}) {
        $self->{engenx_controller_settings}->{$type}->{_cleanup_policy} = param_global()->get($type.".CLEANUP_POLICY")  // param_global()->get("ENGENX.CLEANUP_POLICY") //'noaggr';
        $self->{engenx_controller_settings}->{$type}->{if_exists}       = param_global()->get($type.".IF_EXISTS")       // param_global()->get("ENGENX.IF_EXISTS")      // 'reuse';
        $self->{engenx_controller_settings}->{$type}->{_parallel}       = param_global()->get($type.".PARALLEL")        // param_global()->get("ENGENX.PARALLEL")       // 8;
        $self->{engenx_controller_settings}->{$type}->{_interface}      = param_global()->get($type.".INTERFACE")       // param_global()->get("ENGENX.INTERFACE")      // 'zapi';
        $self->{engenx_controller_settings}->{$type}->{_setup}          = param_global()->get($type.".SETUP")           // param_global()->get("ENGENX.SETUP")          // 1;
        $self->{engenx_controller_settings}->{$type}->{_cleanup}        = param_global()->get($type.".CLEANUP")         // param_global()->get("ENGENX.CLEANUP")        // 1;
        $self->{engenx_controller_settings}->{$type}->{_setup_done}     = param_global()->get($type.".SETUP_DONE")      // param_global()->get("ENGENX.SETUP_DONE")     // 0;
        $self->{engenx_controller_settings}->{$type}->{_cleanup_done}   = param_global()->get($type.".CLEANUP_DONE")    // param_global()->get("ENGENX.CLEANUP_DONE")   // 0;
        # Also allow retry error strings to passed in.
        $self->{engenx_controller_settings}->{$type}->{_cleanup_done}   = param_global()->get($type.".CLEANUP_DONE")    // param_global()->get("ENGENX.CLEANUP_DONE")   // 0;
        my $retry_strings = param_global()->get($type.".RETRY_STRINGS")    // param_global()->get("ENGENX.RETRY_STRINGS");
        @{$self->{engenx_controller_settings}->{$type}->{_retry_strings}}  = split(',',$retry_strings) if defined $retry_strings;
    }

    # Loop through all of the defaults generated above and add them to the
    # resource if the do not already exist.
    foreach my $key (keys %{$self->{engenx_controller_settings}->{$type}}) {
        $resource->{$key} = $self->{engenx_controller_settings}->{$type}->{$key} unless (defined $resource->{$key});
    }

    # Loop through and make all the metadata keys private keys of the resource.
    # These will overwrite any other private keys that may have been set.
    foreach my $key (keys %{$metadata}) {
        # Since EnGenX needs if_exists to not have the _ then we do not want to
        # add an underscore to if_exists if it is in the metadata arguments.
        if($key eq 'if_exists') {
            $resource->{$key} = $metadata->{$key};
        } else {
            $resource->{"_$key"} = $metadata->{$key};
        }
    }

    if (ref($resource) eq 'NACL::AppBuilder::Xgen::Resource::Volume::Mount' and not defined $resource->{'volumes'}) {
        $resource->{'volumes'} = [Clone::clone($resource)];
    }

    return ;
}

sub apply_format_to_resource_list {
    my ($self, $xgen_resource_list)  = @_;

    my $cluster_name;

    require NACL::AppBuilder::EnGenX::Resource;

    foreach my $key (keys %$xgen_resource_list) {

        # 'CLUSTER_' + $cluster_name is needed for the engenx resource list
        if ($key eq 'NACL::AppBuilder::Xgen::Resource::Cluster') {
            $cluster_name = $xgen_resource_list->{$key}->[0]->get_attribute('name');
        }

        foreach my $resource (@{$xgen_resource_list->{$key}}) {
            if ($key eq 'NACL::AppBuilder::Xgen::Resource::Lif') {
                # disable the address for lifs, since the "ips" field is a better way to do it
                $resource->{address} = '';
            } elsif ($key eq 'NACL::AppBuilder::Xgen::Resource::Fcp') {
                $resource->{service_enabled} = 1;
            } elsif ($key eq 'NACL::AppBuilder::Xgen::Resource::Iscsi') {
                $resource->{service_enabled} = 1;
            } elsif ($key eq 'NACL::AppBuilder::Xgen::Resource::VserverIscsiIsns') {
                $resource->{service_enabled} = 1;
            } elsif ($key eq 'NACL::AppBuilder::Xgen::Resource::Volume::Efficiency') {
                $resource->{_efficiency} = 1;
            } elsif ($key eq 'NACL::AppBuilder::Xgen::Resource::Nfs') {
                $resource->{service_enabled} = 1;
            } elsif ($key eq 'NACL::AppBuilder::Xgen::Resource::Cifs') {
                $resource->{service_enabled} = 1;
            } elsif ($key eq 'NACL::AppBuilder::Xgen::Resource::Audit') {
                # audit resources can not be done in parallel
                $resource->{_parallel} = 1;
            } elsif ($key eq 'NACL::AppBuilder::Xgen::Resource::SnapmirrorPolicyRule') {
                # Snapmirror Policy Rule resources can not be done in parallel
                $resource->{_parallel} = 1;
            }elsif ($key eq 'NACL::AppBuilder::Xgen::Resource::Snapmirror') {
                    # Check if this is the lun/file level, volume level or vserver level snapmirror.
                    # The _snapmirror_object_level is used to determine at which level the snapmirror is being created.
                    # This will be used to determine when the snapmirror can be cleaned up.
                    if(exists $resource->{'cg-item-mappings'} or
                       exists $resource->{'CG-ITEM-MAPPINGS'}) {
                        $resource->{'_snapmirror_object_level'} = 'lun';
                    } elsif(exists $resource->{'destination-volume'} or
                            exists $resource->{'DESTINATION-VOLUME'} or
                            (exists $resource->{'destination-path'} and
                            $resource->{'destination-path'} =~ /^((\S+):(\S+))|((\S+):\/\/(\S+)\/(\S+))$/) or
                            (exists $resource->{'DESTINATION-PATH'} and
                            $resource->{'DESTINATION-PATH'} =~ /^((\S+):(\S+))|((\S+):\/\/(\S+)\/(\S+))$/)) {
                        $resource->{'_snapmirror_object_level'} = 'volume';
                    } else {
                        $resource->{'_snapmirror_object_level'} = 'vserver';
                    }

            }

            bless($resource, 'NACL::AppBuilder::EnGenX::Resource');
        }

        my $key_translation = $self->{resource_list_translation}->{$key};

        if (defined($key_translation)) {
            $xgen_resource_list->{$key_translation} = delete $xgen_resource_list->{$key};
        }
    }

    # remove resources that won't be used by EnGenX
    my @resources_to_delete = ('NACL::AppBuilder::Xgen::Resource::Port::Eth',
                               'NACL::AppBuilder::Xgen::Resource::Port::Fcoe',
                               'NACL::AppBuilder::Xgen::Resource::Port::Fcp');

    foreach my $resource_to_delete (@resources_to_delete) {
        delete $xgen_resource_list->{$resource_to_delete};
    }

    my $list_data = {};

    # Create default environment resource
    my $environment_data = {
                'GLOBAL_VX_OPTIONS' => {
                    'DMP_FAST_RECOVERY' => 'on',
                    'IO_POLICY' => 'adaptive',
                    'POWER_FAIL_TIMEOUT' => '30',
                    'DMP_RESTORE_INTERVAL' => '60',
                    'DMP_MONITOR_FABRIC' => 'NOT_SET',
                    'DMP_QUEUE_DEPTH' => 'NOT_SET',
                    'RECOVERY_OPTION' => 'default',
                    'DMP_RESTORE_POLICY' => 'check_disabled',
                    'DMP_LOG_LEVEL' => '4',
                    'DMP_PATH_AGE' => 'NOT_SET',
                    'DMP_LUN_RETRY_TIMEOUT' => 'NOT_SET',
                    'DMP_PROBE_IDLE_LUN' => 'NOT_SET'
                },
                'GLOBAL_DMMP_OPTIONS' => {
                    'PATH_GROUPING_POLICY' => 'NOT_SET',
                    'PRIO_CALLOUT' => 'NOT_SET',
                    'PATH_SELECTOR' => 'NOT_SET',
                    'USE_USER_FRIENDLY_NAMES' => '0',
                    'QUEUE_WITHOUT_DAEMON' => 'NOT_SET',
                    'HARDWARE_HANDLER' => 'NOT_SET',
                    'MAX_FDS' => 'NOT_SET',
                    'PG_PRIO_CALC' => 'NOT_SET',
                    'PATH_CHECKER' => 'NOT_SET',
                    'RR_WEIGHT' => 'NOT_SET',
                    'FEATURES' => 'NOT_SET',
                    'RR_MIN_IO' => 'NOT_SET',
                    'FAILBACK' => 'NOT_SET',
                    'PRIO' => 'NOT_SET',
                    'FAST_IO_FAIL_TMO' => 'NOT_SET',
                    'FLUSH_ON_LAST_DEL' => 'NOT_SET',
                    'GETUID_CALLOUT' => 'NOT_SET',
                    'PRODUCT' => 'NOT_SET',
                    'DEV_LOSS_TMO' => 'NOT_SET',
                    'VENDOR' => 'NOT_SET'
                },
                'GLOBAL_FCP_OPTIONS' => {
                    'DRIVER_TYPE' => '',
                    'MULTI_TARGET_ZONING' => param_global()->get("GLOBAL_FCP_OPTIONS.MULTI_TARGET_ZONING") // '1',
                    'FILTER_TARGETS_BY_PORTSET' => '0',
                    'PERFORM_ZONING' => '1',
                    'DONT_CREATE_TARGET_ALIASES' => param_global()->get("GLOBAL_FCP_OPTIONS.DONT_CREATE_TARGET_ALIASES") // '1',
                    'MAX_MEMBERS_PER_MULTI_TARGET_ZONE' => '8',
                    'AIX_DYNAMIC_TRACKING' => '',
                    'AIX_ERROR_RECOVERY' => '',
                    'AIX_NUM_CMD_ELEMS' => ''
                },
                'GLOBAL_NFS_OPTIONS' => {
                    'LIF_ADDRESS_FILTER' => 'qr/.*/',
                    'VOLUME_FILTER' => 'qr/.*/',
                    'CREATE_UNIQUE_SUBDIR' => '1',
                    'PERSISTENT' => 'no',
                    'VSERVER_FILTER' => 'qr/.*/',
                    'MOUNT_QTREES' => '0',
                    'CLUSTER_FILTER' => 'qr/.*/',
                    'VOLUME_LIF_AFFINITY' => 'any',
                    'QTREE_FILTER' => 'qr/.*/',
                    'LIF_NAME_FILTER' => 'qr/.*/',
                    'JUNCTION_PATH_FILTER' => 'qr/.*/'
                },
                'GLOBAL_ISCSI_OPTIONS' => {
                    'HEADER_DIGEST' => '0',
                    'INITIAL_RT' => '0',
                    'WIN_MPIO' => '1',
                    'CHAP_OUT_PASSWORD' => '87654321ETAN',
                    'IP_VERSION' => 'ipv4',
                    'WIN_KEEP_TARGET_PORTALS' => '0',
                    'FILTER_TARGETS_BY_PORTSET' => '0',
                    'WIN_NUM_SESSIONS' => '1',
                    'DISCOVERY_TYPE' => param_global()->get("GLOBAL_ISCSI_OPTIONS.DISCOVERY_TYPE") // 'sendtargets',
                    'WIN_PERSISTENT' => param_global()->get("GLOBAL_ISCSI_OPTIONS.WIN_PERSISTENT") // '0',
                    'IMMEDIATE_DATA' => '0',
                    'CHAP_IN_USERNAME' => '',
                    'WIN_NUM_CONNS_PER_SESSION' => '1',
                    'HARDWARE_VENDOR' => '',
                    'DEFAULT_TIMERETAIN' => '',
                    'INITIATOR_TYPE' => 'software',
                    'ISNS_SERVER' => param_global()->get("GLOBAL_ISCSI_OPTIONS.ISNS_SERVER") // '',
                    'DISCOVERY_PRECLEANUP' => '2',
                    'SUBNET_MATCHING' => param_global()->get("GLOBAL_ISCSI_OPTIONS.SUBNET_MATCHING") // 1,
                    'WIN_MAX_CONNECTIONS' => '',
                    'DEFAULT_TIMEWAIT' => '',
                    'CHAP_OUT_USERNAME' => '',
                    'DISCOVERY' => '1',
                    'DATA_DIGEST' => '0',
                    'USE_MGMT_PORT' => param_global()->get("GLOBAL_ISCSI_OPTIONS.USE_MGMT_PORT") // 0,
                    'IPV_EXCLUDE_LINK_LOCAL' => '1',
                    'ERL' => '0',
                    'CHAP_IN_PASSWORD' => 'NATE',
                    'CHAP_MODE' => 'none'
                },
                'OEM_REBRAND' => 'netapp',
                'PNATE' => '0',
                'DEBUG_LEVEL' => '0',
                'NTP_TIME_SERVER' => 'ntp.eng.netapp.com',
                'NTP_TIME_ZONE' => 'ESTEDT',
                'NTP_SYNC' => '1'
            };
    bless($environment_data, 'NACL::AppBuilder::EnGenX::Resource');
    $list_data->{'ENVIRONMENT'}->{'environment'}->[0] = $environment_data;

    # Create a default client group.
    my $client_group_data = {
        'CONFIG' => '1',
        'SETUP' => '1',
        'CLEANUP' => '1',
        'DECONFIG' => '1',
        'REQUIRED_CLIENTS' => '',
        'MAX_FAILURE_COUNT' => '0',
        'MAX_SETUP_FAILURE_COUNT' => '0',
        'VALIDATE' => '1',
        'GROUP_ID' => 0
    };
    bless($client_group_data, 'NACL::AppBuilder::EnGenX::Resource');
    $list_data->{CLIENTS}->{client_groups}->[0] = $client_group_data;

    # take out all the client resources and put them in their own resource list
    # category: "CLIENTS", type: "client"

    my $client_list = delete $xgen_resource_list->{'client'};
    my @client_list_arr;
    my @vmware_list_arr;
    my @xen_list_arr;
    foreach my $client_res (@{$client_list}) {
        foreach my $key (keys %$client_res) {
            $client_res->{uc($key)} = delete $client_res->{$key};
        }

        if ($client_res->{OSTYPE} =~ /vmware/i) {
            push(@vmware_list_arr, $client_res);
            next;
        }

        if ($client_res->{OSTYPE} =~ /xen/i) {
            push(@xen_list_arr, $client_res);
            next;
        }

        my $type = ref($client_res);
        $type =~ s/::/\./g;
        $client_res->{'GROUP_ID'} = 0;
        $client_res->{'PROTOCOL'} = join('|', @{delete$client_res->{'PROTOCOLS'}});
        $client_res->{'HYPERVISOR'} = 'none';
        $client_res->{KILL_IO_PROCESSES} = 1;
        $client_res->{CLEANUP} = param_global()->get($type.".CLEANUP") // param_global()->get("ENGENX.CLEANUP") // 1;
        $client_res->{SETUP} = param_global()->get($type.".SETUP") // param_global()->get("ENGENX.SETUP") // 1;
        $client_res->{DECONFIG} = param_global()->get($type.".DECONFIG") // param_global()->get("ENGENX.DECONFIG") // 1;
        $client_res->{CONFIG} = param_global()->get($type.".CONFIG") // param_global()->get("ENGENX.CONFIG") // 1;
        $client_res->{'MP_TYPE'} = 'none' unless (defined $client_res->{'MP_TYPE'});
        $client_res->{'LVM_TYPE'} = 'none';
        $client_res->{'IO_TYPE'} = param_global()->get("CLIENT_IO_TYPE") // 'raw';
        $client_res->{'LUN_PATH_REGEX'} = 'qr/.*/';

        # if linux and fcp, default the driver type to be native
        if ($client_res->{OSTYPE} eq 'linux' and
            $client_res->{PROTOCOL} =~ /fcp/i) {
            $client_res->{'FCP_OPTIONS'} = {
                DRIVER_TYPE => 'native'
            };
        }
        push(@client_list_arr, $client_res);
    }

    $list_data->{CLIENTS}->{clients} = \@client_list_arr;

    if (@vmware_list_arr or @xen_list_arr) {
        # Create a default client group.
        my $hypervisor_group_data = {
            'CONFIG' => '1',
            'SETUP' => '1',
            'CLEANUP' => '1',
            'DECONFIG' => '1',
        };
        bless($hypervisor_group_data, 'NACL::AppBuilder::EnGenX::Resource');

        $list_data->{HYPERVISORS}->{hypervisor_groups}->[0] = $hypervisor_group_data;

        # setup hypervisor attributes as resources
        foreach my $client_res (@vmware_list_arr) {
            foreach my $key (keys %$client_res) {
                $client_res->{uc($key)} = delete $client_res->{$key};
            }

            my $type = ref($client_res);
            $type =~ s/::/\./g;
            $client_res->{PROTOCOL} = join('|', @{delete $client_res->{PROTOCOLS}});
            $client_res->{LVM_TYPE} = 'none';
            $client_res->{PSP_LIST} = 'rr';
            $client_res->{PSP_LIST_RANDOM_SEED} = "";
            $client_res->{VMDKS_PER_DATASTORE} = 2;
            $client_res->{VMFS_LUN_PATH_REGEX} = "qr/.*/";
            $client_res->{VMFS_VMDK_LUN_PATH_REGEX} = "qr/.*/";
            $client_res->{VMFS_DATASTORE_PERCENTAGE} = "100";
            $client_res->{NFS_DATASTORE_PERCENTAGE} = "50";
            $client_res->{VVOL_DATASTORE_PERCENTAGE} = "25";
            $client_res->{OVERPROVISION_NFS_DATASTORES} = "1";
            $client_res->{SCSI_BUS} = "lsilogic";
            $client_res->{VMFS_TYPE} = "";
            $client_res->{VMDK_ADD_REMOVE_PARALLEL_LIMIT} = "3";
            $client_res->{DELETE_NON_EMPTY_DATASTORES} = "1";
            $client_res->{VMFS_VMDK_DISK_TYPE} = "";
            $client_res->{NFS_VMDK_DISK_TYPE} = "";
            $client_res->{VVOL_VMDK_DISK_TYPE} = "";
            $client_res->{VMFS_VMDK_DIVISIBILITY_BYTE_SIZE} = "0";
            $client_res->{VVOL_VMDK_DIVISIBILITY_BYTE_SIZE} = "0";
            $client_res->{CONFIGURE_QFULL} = "1";
            $client_res->{CONFIGURE_SYSLOG} = "1";
            $client_res->{CLEANUP} = param_global()->get($type.".CLEANUP") // param_global()->get("ENGENX.CLEANUP") // 1;
            $client_res->{SETUP} = param_global()->get($type.".SETUP") // param_global()->get("ENGENX.SETUP") // 1;
            $client_res->{DECONFIG} = param_global()->get($type.".DECONFIG") // param_global()->get("ENGENX.DECONFIG") // 1;
            $client_res->{CONFIG} = param_global()->get($type.".CONFIG") // param_global()->get("ENGENX.CONFIG") // 1;
        }
        if (@vmware_list_arr) {
            $list_data->{HYPERVISORS}->{esx} = \@vmware_list_arr;
        }

        foreach my $client_res (@xen_list_arr) {
            foreach my $key (keys %$client_res) {
                $client_res->{uc($key)} = delete $client_res->{$key};
            }
            # TODO: process client res for xen
        }
        if (@xen_list_arr) {
            $list_data->{HYPERVISORS}->{cxs} = \@xen_list_arr;
        }
    }

    # convert virtual machine resources and store them in the clients resource list data
    my $vm_list = delete $xgen_resource_list->{'virtual_machine'};
    foreach my $client_res (@{$vm_list}) {
        foreach my $key (keys %$client_res) {
            $client_res->{uc($key)} = delete $client_res->{$key};
        }

        my $type = ref($client_res);
        $type =~ s/::/\./g;
        $client_res->{GROUP_ID} = 0;
        $client_res->{PROTOCOL} = join('|', @{delete $client_res->{'PROTOCOLS'}});
        $client_res->{HYPERVISOR} = delete($client_res->{HYPERVISOR_TYPE});
        $client_res->{HYPERVISOR_RESOURCE} = delete($client_res->{HYPERVISOR_NAME});
        $client_res->{KILL_IO_PROCESSES} = 1;
        $client_res->{CLEANUP} = param_global()->get($type.".CLEANUP") // param_global()->get("ENGENX.CLEANUP") // 1;
        $client_res->{SETUP} = param_global()->get($type.".SETUP") // param_global()->get("ENGENX.SETUP") // 1;
        $client_res->{DECONFIG} = param_global()->get($type.".DECONFIG") // param_global()->get("ENGENX.DECONFIG") // 1;
        $client_res->{CONFIG} = param_global()->get($type.".CONFIG") // param_global()->get("ENGENX.CONFIG") // 1;
        $client_res->{MP_TYPE} = 'none';
        $client_res->{LVM_TYPE} = 'none';
        $client_res->{IO_TYPE} = param_global()->get("CLIENT_IO_TYPE") // 'raw';
        $client_res->{LUN_PATH_REGEX} = 'qr/.*/';
    }
    push(@{$list_data->{HYPERVISORS}->{virtual_machines}}, @{$vm_list});
    push(@{$list_data->{CLIENTS}->{clients}}, @{$vm_list});

    # Client Cluster Group
    my $client_cluster_group_data = {
        'CONFIG' => '1',
        'SETUP' => '1',
        'CLEANUP' => '1',
        'DECONFIG' => '1',
        'REQUIRED_CLIENT_CLUSTERS' => '',
        'MAX_FAILURE_COUNT' => '0',
        'MAX_SETUP_FAILURE_COUNT' => '0',
        'GROUP_ID' => 0
    };
    bless($client_cluster_group_data, 'NACL::AppBuilder::EnGenX::Resource');
    $list_data->{CLIENT_CLUSTERS}->{client_cluster_groups}->[0] = $client_cluster_group_data;

    # Client Clusters
    my $client_cluster_list = delete $xgen_resource_list->{'client_cluster'};
    foreach my $client_res (@{$client_cluster_list}) {
        foreach my $key (keys %$client_res) {
            $client_res->{uc($key)} = delete $client_res->{$key};
        }

        my $type = ref($client_res);
        $type =~ s/::/\./g;
        $client_res->{'GROUP_ID'} = 0;
        $client_res->{CLEANUP} = param_global()->get($type.".CLEANUP") // param_global()->get("ENGENX.CLEANUP") // 1;
        $client_res->{SETUP} = param_global()->get($type.".SETUP") // param_global()->get("ENGENX.SETUP") // 1;
        $client_res->{CLUSTER_LUN_REGEX} = '.*';
    }
    push(@{$list_data->{CLIENT_CLUSTERS}->{client_clusters}}, @{$client_cluster_list});

    # Client Cluster Nodes
    my $node_list = delete $xgen_resource_list->{'client_cluster_node'};
    foreach my $client_res (@{$node_list}) {
        foreach my $key (keys %$client_res) {
            $client_res->{uc($key)} = delete $client_res->{$key};
        }

        my $type = ref($client_res);
        $type =~ s/::/\./g;
        $client_res->{'GROUP_ID'} = 0;
        $client_res->{'PROTOCOL'} = join('|', @{delete$client_res->{'PROTOCOLS'}});
        $client_res->{'HYPERVISOR'} = 'none';
        $client_res->{KILL_IO_PROCESSES} = 1;
        $client_res->{CLEANUP} = param_global()->get($type.".CLEANUP") // param_global()->get("ENGENX.CLEANUP") // 1;
        $client_res->{SETUP} = param_global()->get($type.".SETUP") // delete $client_res->{IS_PRIMARY};
        $client_res->{DECONFIG} = param_global()->get($type.".DECONFIG") // param_global()->get("ENGENX.DECONFIG") // 1;
        $client_res->{CONFIG} = param_global()->get($type.".CONFIG") // param_global()->get("ENGENX.CONFIG") // 1;
        $client_res->{'MP_TYPE'} = 'none' unless (defined $client_res->{'MP_TYPE'});
        $client_res->{'LVM_TYPE'} = 'none';
        $client_res->{'IO_TYPE'} = param_global()->get("CLIENT_IO_TYPE") // 'fs';
        $client_res->{'LUN_PATH_REGEX'} = 'qr/.*/';

        # if linux and fcp, default the driver type to be native
        if ($client_res->{OSTYPE} eq 'linux' and
            $client_res->{PROTOCOL} =~ /fcp/i) {
            $client_res->{'FCP_OPTIONS'} = {
                DRIVER_TYPE => 'native'
            };
        }
    }
    push(@{$list_data->{CLIENTS}->{clients}}, @{$node_list});

    # setup cluster attributes as a resource
    my $clusters_attributes_data = {};
    $clusters_attributes_data->{VALIDATE}      = '1';
    $clusters_attributes_data->{PARALLEL}      = '10';
    $clusters_attributes_data->{CLEANUP}       = '1';
    $clusters_attributes_data->{SETUP}         = '1';
    $clusters_attributes_data->{DECONFIG}      = '1';
    $clusters_attributes_data->{CONFIG}        = '1';
    $clusters_attributes_data->{PRECLEANUP}    = '0';
    $clusters_attributes_data->{CMD_INTERFACE} = 'zapi';
    bless($clusters_attributes_data, 'NACL::AppBuilder::EnGenX::Resource');
    $list_data->{CLUSTERS}->{attributes} = [$clusters_attributes_data];

    # fabrics resources
    my $fabric_list = delete $xgen_resource_list->{'fabric'};
    foreach my $fab_res (@{$fabric_list}) {
        foreach my $key (keys %$fab_res) {
            next if ($key =~ /^_/);
            $fab_res->{uc($key)} = delete $fab_res->{$key};
        }

        my $type = ref($fab_res);
        $type =~ s/::/\./g;
        $fab_res->{DECONFIG} = param_global()->get($type.".DECONFIG") // param_global()->get("ENGENX.DECONFIG") // 1;
        $fab_res->{CONFIG} = param_global()->get($type.".CONFIG") // param_global()->get("ENGENX.CONFIG") // 1;
    }
    $list_data->{FABRICS}->{fabrics} = $fabric_list;

    # fabric groups
    my $fabric_groups_data          = {};
    $fabric_groups_data->{CONFIG}   = '1';
    $fabric_groups_data->{DECONFIG} = '1';
    bless($fabric_groups_data, 'NACL::AppBuilder::EnGenX::Resource');
    $list_data->{FABRICS}->{fabric_groups} = [$fabric_groups_data];

    # switch resources
    my $switch_list = delete $xgen_resource_list->{'switch'};
    foreach my $sw_res (@{$switch_list}) {
        foreach my $key (keys %$sw_res) {
            next if ($key =~ /^_/);
            $sw_res->{uc($key)} = delete $sw_res->{$key};
        }

        my $type = ref($sw_res);
        $type =~ s/::/\./g;
        $sw_res->{PORT_RETRIES} = 5;
        $sw_res->{VERIFY_ON_ENABLE} = 1;
        $sw_res->{BACKEND_SWITCH} = 0;
        $sw_res->{CLEANUP} = param_global()->get($type.".CLEANUP") // param_global()->get("ENGENX.CLEANUP") // 0;
        $sw_res->{SETUP} = param_global()->get($type.".SETUP") // param_global()->get("ENGENX.SETUP") // 0;
        $sw_res->{DECONFIG} = param_global()->get($type.".DECONFIG") // param_global()->get("ENGENX.DECONFIG") // 0;
        $sw_res->{CONFIG} = param_global()->get($type.".CONFIG") // param_global()->get("ENGENX.CONFIG") // 0;
        $sw_res->{CONN_TYPE} = "THARNHOST";
    }
    $list_data->{SWITCHES}->{switches} = $switch_list;

    # install the actions resource
    my $actions_attributes_data         = {};
    $actions_attributes_data->{SETUP}   = '1';
    $actions_attributes_data->{CLEANUP} = '1';
    bless($actions_attributes_data, 'NACL::AppBuilder::EnGenX::Resource');
    $list_data->{ACTIONS}->{attributes} = [$actions_attributes_data];

    # Need to sort the resources by cluster.
    foreach my $resource_type (keys %$xgen_resource_list) {
        # Get all the resources of the current type
        my @resources = @{$xgen_resource_list->{$resource_type}};
        # Loop through and add the resource to the correct category.
        foreach my $resource_obj (@resources) {
            my $cluster_name_for_category = $resource_obj->{_cluster};
            if(not defined $cluster_name_for_category and $resource_type eq 'cluster') {
                $cluster_name_for_category = $resource_obj->{name}
            }
            next unless defined $cluster_name_for_category;
            my $category = 'CLUSTER_' . $cluster_name_for_category;
            # put transformed xgen_resource_list into list_data
            push(@{$list_data->{$category}->{$resource_type}},$resource_obj);
        }
    }
    # create the final resource list
    my $blessed_list_data = {};
    $blessed_list_data->{list_data} = $list_data;
    require NACL::AppBuilder::EnGenX::ResourceList;
    bless($blessed_list_data, 'NACL::AppBuilder::EnGenX::ResourceList');

    # put list_data into engenx_resource_list
    my $engenx_resource_list = {};
    $engenx_resource_list = $blessed_list_data;
    return $engenx_resource_list;
}

sub generate_resource_translation_hash {
    my ($self) = @_;

    my $translation_hash = {};

    # CLUSTER
    my $cluster_translation_hash = {};
    $cluster_translation_hash->{cluster} = 'name';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::Cluster'} = $cluster_translation_hash;

    # NODE
    my $node_translation_hash = {};
    $node_translation_hash->{node} = 'name';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::Node'} = $node_translation_hash;

    # AGGREGATE
    my $aggregate_translation_hash = {};
    $aggregate_translation_hash->{aggregate} = 'name';
    $aggregate_translation_hash->{diskcount} = 'num_disks';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::Aggregate'} = $aggregate_translation_hash;

    # CIFS SHARE
    my $cifs_share_translation_hash = {};
    $cifs_share_translation_hash->{cifs_share} = 'name';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::Cifs::Share'} = $cifs_share_translation_hash;

    # VSERVER
    my $vserver_translation_hash = {};
    $vserver_translation_hash->{vserver} = 'name';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::Vserver'} = $vserver_translation_hash;

    # LIF
    my $lif_translation_hash = {};
    $lif_translation_hash->{lif} = 'name';
    $lif_translation_hash->{data_protocol} = 'data-protocol';
    $lif_translation_hash->{vserver_name} = 'vserver';
    $lif_translation_hash->{home_node} = 'home-node';
    $lif_translation_hash->{home_port} = 'home-port';
    $lif_translation_hash->{address_range} = 'ips';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::Lif'} = $lif_translation_hash;

    # VOLUME
    my $volume_translation_hash = {};
    $volume_translation_hash->{volume} = 'name';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::Volume'} = $volume_translation_hash;

    # VOLUME
    my $volume_mount_translation_hash = {};
    $volume_mount_translation_hash->{volume} = 'name';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::Volume::Mount'} = $volume_mount_translation_hash;

    # FILE
    my $file_translation_hash = {};
    $file_translation_hash->{'file-size'} = 'size';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::File'} = $file_translation_hash;

    # IPspace
    my $ipspace_translation_hash = {};
    $ipspace_translation_hash->{ipspace} = 'name';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::IPspace'} = $ipspace_translation_hash;

    # BROADCAST DOMAIN
    my $broadcast_domain_translation_hash = {};
    $broadcast_domain_translation_hash->{broadcast_domain} = 'name';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::IPspace::BroadcastDomain'} = $broadcast_domain_translation_hash;

    # EXPORT POLICY
    my $export_policy_translation_hash = {};
    $export_policy_translation_hash->{export_policy} = 'name';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::ExportPolicy'} = $export_policy_translation_hash;

    # VSCAN ON ACCESS POLICY
    my $on_access_policy_translation_hash = {};
    $on_access_policy_translation_hash->{on_access_policy} = 'name';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::Vscan::OnAccessPolicy'} = $on_access_policy_translation_hash;

    # VSCAN SCANNER POOL
    my $scanner_pool_translation_hash = {};
    $scanner_pool_translation_hash->{scanner_pool} = 'name';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::Vscan::ScannerPool'} = $scanner_pool_translation_hash;

    # FPOLICY POLICY
    my $fpolicy_policy_translation_hash = {};
    $fpolicy_policy_translation_hash->{fpolicy} = 'name';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::Fpolicy::Policy'} = $fpolicy_policy_translation_hash;

    # FPOLICY EVENT
    my $fpolicy_event_translation_hash = {};
    $fpolicy_event_translation_hash->{fpolicy_event} = 'name';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::Fpolicy::Policy::Event'} = $fpolicy_event_translation_hash;

    # FPOLICY EXTERNAL ENGINE
    my $fpolicy_external_engine_translation_hash = {};
    $fpolicy_external_engine_translation_hash->{external_engine} = 'name';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::Fpolicy::Policy::ExternalEngine'} = $fpolicy_external_engine_translation_hash;

    # IFGRP
    my $ifgrp_translation_hash = {};
    $ifgrp_translation_hash->{ifgrp} = 'name';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::Ifgrp'} = $ifgrp_translation_hash;

    # LUN
    my $lun_translation_hash = {};
    $lun_translation_hash->{lun} = 'name';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::Lun'} = $lun_translation_hash;

    # VOLUME CLONE
    my $volume_clone_translation_hash = {};
    $volume_clone_translation_hash->{flexclone} = 'name';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::VolumeClone'} = $volume_clone_translation_hash;

    # LOCAL GROUP MEMBERSHIP
    my $membership_translation_hash = {};
    $membership_translation_hash->{member} = 'member-names';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::Cifs::LocalGroup::Membership'} = $membership_translation_hash;

    # IGROUP
    my $igroup_translation_hash = {};
    $igroup_translation_hash->{igroup} = 'name';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::Igroup'} = $igroup_translation_hash;

    #CLUSTER PEER
    my $cluster_peer_translation_hash = {};
    $cluster_peer_translation_hash->{rcluster} = 'remote_cluster';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::ClusterPeer'} = $cluster_peer_translation_hash;

    # INITIATOR
    my $initiator_translation_hash = {};
    $initiator_translation_hash->{client_name} = 'name';
    $initiator_translation_hash->{vserver_name} = 'vserver';
    $initiator_translation_hash->{igroup_name} = 'igroup';
    $initiator_translation_hash->{protocol} = '_protocol';
    $initiator_translation_hash->{initiator} = 'name';

    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::Igroup::Initiator'} = $initiator_translation_hash;

    # Wafliron
    my $wafliron_translation_hash = {};
    $wafliron_translation_hash->{aggregate} = 'name';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::Aggregate::WaflIron'} = $wafliron_translation_hash;

    # ClientCluster
    my $client_cluster_translation_hash = {};
    $client_cluster_translation_hash->{cluster_name} = 'name';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::ClientCluster'} = $client_cluster_translation_hash;

    # common
    my $common_translation_hash = {};
    $common_translation_hash->{cluster_name} = '_cluster';
    $translation_hash->{common} = $common_translation_hash;

    # store translation_hash on $self->{translation}
    $self->{resource_translation} = $translation_hash;

    return ;
}

sub generate_resource_list_translation_hash {
    my ($self) = @_;

    my $translation_hash = {};

    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::Cluster'} = 'cluster';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::Node'} = 'node';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::Aggregate'} = 'aggregate';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::Aggregate::ObjectStore'} = 'object_store';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::Nfs'} = 'nfs_service';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::Dns'} = 'dns_service';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::Cifs'} = 'cifs_service';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::Cifs::Security'} = 'cifs_security';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::Cifs::Options'} = 'cifs_vserver_options';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::Iscsi'} = 'iscsi_service';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::IscsiSecurity'} = 'iscsi_security';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::VserverIscsiIsns'} = 'isns_service';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::Fcp'} = 'fcp_service';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::Vserver'} = 'vserver';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::FailoverGroup'} = 'failover_group';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::Lif'} = 'lif';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::NetworkRoute'} = 'network_route';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::Security::Trace::Filter'} = 'security_trace_filter';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::ExportPolicy'} = 'export_policy';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::ExportPolicy::Rule'} = 'export_policy_rule';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::Vscan::OnDemandTask'} = 'on_demand_task';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::Vscan::OnAccessPolicy'} = 'on_access_policy';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::Vscan::ScannerPool'} = 'scanner_pool';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::Audit'} = 'audit';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::Cifs::Share'} = 'cifs_share';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::Cifs::Security'} = 'cifs_security';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::Cifs::Symlink'} = 'cifs_symlink';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::Cifs::Branchcache'} = 'cifs_branchcache';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::Cifs::ShareAccessControl'} = 'share_access_control';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::Cifs::HomeDirSearchPath'} = 'homedir_search_path';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::Cifs::LocalUser'} = 'cifs_local_user';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::Cifs::LocalGroup'} = 'cifs_local_group';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::Cifs::GroupPolicy'} = 'cifs_group_policy';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::Cifs::Privilege'} = 'cifs_privilege';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::Cifs::LocalGroup::Membership'} = 'local_group_membership';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::NameMapping'} = 'name_mapping';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::Volume'} = 'volume';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::Volume::Clone'} = 'volume_clone';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::Volume::Mount'} = 'volume_mount';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::File'} = 'file';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::Qtree'} = 'qtree';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::Volume::Efficiency'} = 'efficiency_options';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::Volume::EfficiencyPolicy'} = 'volume_efficiency_policy';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::Volume::QuotaPolicy'} = 'quota_policy';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::Volume::QuotaPolicyRule'} = 'quota_rule';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::Volume::Quota'} = 'volume_quota';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::Lun'} = 'lun';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::LunAttribute'} = 'lun_attribute';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::LunBind'} = 'lun_bind';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::LunMap'} = 'lun_map';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::Igroup'} = 'igroup';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::Igroup::Initiator'} = 'initiator';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::Client'} = 'client';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::ClientCluster'} = 'client_cluster';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::ClientCluster::Node'} = 'client_cluster_node';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::VirtualMachine'} = 'virtual_machine';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::Switch'} = 'switch';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::Fabric'} = 'fabric';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::Fpolicy::Policy'} = 'fpolicy';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::Fpolicy::Policy::Event'} = 'fpolicy_event';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::Fpolicy::Policy::Scope'} = 'fpolicy_scope';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::Fpolicy::Policy::ExternalEngine'} = 'external_engine';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::Fpolicy::Enable'} = 'fpolicy_enable';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::Fpolicy::ExternalEngineConnection'} = 'external_engine_connection';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::Security::FileDirectory::Ntfs'} = 'file_directory_ntfs';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::Security::FileDirectory::Policy'} = 'file_directory_policy';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::Security::FileDirectory::Apply'} = 'file_directory_policy_apply';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::Security::FileDirectory::Ntfs::Dacl'} = 'ntfs_dacl_setting';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::Security::FileDirectory::Ntfs::Sacl'} = 'ntfs_sacl_setting';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::Security::FileDirectory::Policy::PolicyTask'} = 'policy_task';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::DynamicDns'} = 'dns_dynamic_update';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::Snapmirror'} = 'snapmirror';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::SnapmirrorPolicy'} = 'snapmirror_policy';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::SnapmirrorPolicyRule'} = 'snapmirror_policy_rule';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::SnapshotPolicy'} = 'snapshot_policy';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::JobScheduleCron'} = 'job_schedule_cron';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::Nis'} = 'nis_service';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::VserverPeer'} = 'vserver_peer';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::SnapmirrorInitialize'} = 'snapmirror_initialize';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::QosPolicyGroup'} = 'qos_policy_group';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::IPspace'} = 'ipspace';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::IPspace::BroadcastDomain'} = 'broadcast_domain';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::IPspace::BroadcastDomain::Port'} = 'bd_port';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::ClusterPeer'} = 'cluster_peer';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::Aggregate::WaflIron'} = 'storage_aggregate_wafliron';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::Portset'} = 'portset';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::Ifgrp'} = 'ifgrp';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::SecurityLogin'} = 'security_login';
    $translation_hash->{'NACL::AppBuilder::Xgen::Resource::Vscan'} = 'vscan';

    $self->{resource_list_translation} = $translation_hash;

    return;
}

1;

__END__

=head1 NAME

NACL::AppBuilder::EnGenX::XgenFormatter

=head1 AUTHOR/MAINTAINER

=over

=item PTEFIT AUTOMATION (ng-ptefit-automation-core@netapp.com)

=back

   Last revised $Date: $ by $Author: $

=cut
