#
# Copyright (c) 2014-20* NetApp Inc.
# All rights reserved.
#
## @summary Driver class for various health checkers defined
##  inside lib/NACL/MTask/HealthChecker/ dir. It provides
##  functionality to register, unregister, enable, disable and
##  run health checkers.
##
## @author dl-nacl-dev@netapp.com
## @status Shared
## @pod here

package NACL::AppBuilder::HealthChecker::Driver;
use NACL::C::Node;
use Data::Dumper;
use Scalar::Util qw(blessed);
use Storable qw(freeze thaw dclone);

use feature 'state';
use Params::Validate qw(validate_with OBJECT SCALAR ARRAYREF);
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use NATE::Log qw(log_global);
use NATE::Messenger qw(messenger_global);
use NATE::Events qw(eventloop_drain);
use NATE::Process;

use NACL::APISet::Exceptions::ResponseException qw(:try);
use NATE::BaseException;
use NACL::AppBuilder::Exceptions::HealthChecker;

use Class::MethodMaker [
    array => 'nodes',
];

BEGIN {
    use Exporter qw(import);
    our @EXPORT_OK
        = qw($MESSENGER_ZEROMQ_PORT);
}

use vars qw($MESSENGER_ZEROMQ_PORT);
$MESSENGER_ZEROMQ_PORT = 54321;

# Global map to store all health checkers registered
# to health checker driver.
my %Registered_Health_Checkers;

# HealthCheckers can also be registered using name.
# This stores mapping b/w names and corresponding package.

# NOTE: currently empty. Add 'name to package' mapping
# to this hash whenever a new HealthChecker is added to MTask layer
# Some of the examples are:
#####################################################
# hc_name           =>       PackageName
#####################################################
# mgwd_ooq          =>       MgwdOoq
# filer_health      =>       FilerHealth
#####################################################
my %HC_Name_To_Package = ();

my $OneTrueSelf = undef;

=head1 METHODS

=head2 new

    my $HCD = NACL::AppBuilder::HealthChecker::Driver->new(
        nodes => \@nodes,
    );

(Singleton Class)
This method will create an object of type
NACL::AppBuilder::HealthChecker::Driver. Will return same object
if its already created

=over

=item Options

=over

=item C<< nodes => @nodes_list >>

(Required)
Each element of the list can either be L<NACL::C::Component::command_interface|lib-NACL-C-Component-pm/command_interface> OR node name.
If its node name then corresponding nate_hosts file should exist.

=back

=item Returns

Returns object of type NACL::AppBuilder::HealthChecker::Driver.

=back

=cut

sub new {
    $Log->enter() if $may_enter;
    if (!defined $OneTrueSelf) {
        my ($class, @args) = @_;
        state $spec = {
            nodes => { type => ARRAYREF },
        };

        my %opts = validate_with(
            params => \@args,
            spec   => $spec,
        );
        my $error = "Object passed should be of type 'NACL::C::CommandInterface::ONTAP' " .
               "But it was of type ";
        my @nodes = @{$opts{nodes}};
        my @node_objs;
        foreach my $node (@nodes) {
            if (blessed($node)) {
                if ($node->isa("NACL::C::CommandInterface::ONTAP")) {
                    push(@node_objs, $node);
                } else {
                    NATE::BaseException->throw($error . ref($node));
                }
            } else {
                push(@node_objs, NACL::C::Node->new(name => $node));
            }
        }

        my $self = {};
        $OneTrueSelf = bless $self, $class;
        $OneTrueSelf->nodes(@node_objs);

        $OneTrueSelf->{counter} = 1;
        # obtain the global messenger object using transport 'zeromq' (which is the default)
        $OneTrueSelf->_init_messenger();
    }

    $Log->exit() if $may_exit;
    return $OneTrueSelf;
}

=head2 hc_register

    my $hc_obj = NACL::MTask::HealthChecker::FilerHealth->new(
        nodes => \@nodes
    );
    my $HCD->hc_register(
        hc => $hc_obj,
    );

OR

    my $HCD->hc_register(
        hc => $hc_name,
    );

Method to register HealthChecker with the Driver.
It creates object for the health checker if name is passed instead of object.


=over

=item Options

=over

=item C<< hc => $hc >>

(Required)
It can either be a key in %HC_Name_To_Package OR
object of type NACL::MTask::HealthChecker::*

=back

=item Returns

Returns object of type NACL::MTask::HealthChecker::*.

=back

=cut

sub hc_register {
    $Log->enter() if $may_enter;
    my ($self, @args) = @_;
    state $spec = {
        hc => {type => SCALAR|OBJECT},
    };

    my %opts = validate_with(
        params => \@args,
        spec   => $spec,
    );

    my $hc = $opts{hc};
    my $hc_obj;

    if (blessed $hc) {
        my $ref = $hc->name();
        if (exists $Registered_Health_Checkers{$ref}) {
            $Log->comment("HealthChecker '$ref' is already registered with the driver");
        } else {
            $self->_register_hc_using_obj(hc => $hc);
        }
        $hc_obj = $hc;
    } else {
        if (exists $Registered_Health_Checkers{$hc}) {
            $Log->comment("HealthChecker '$hc' is already registered with the driver");
        } else {
            $self->_register_hc_using_name(hc => $hc);
        }
        $hc_obj = $Registered_Health_Checkers{$hc};
    }

    $Log->exit() if $may_exit;

    return $hc_obj;
}

sub _register_hc_using_obj {
    my ($self,%opts) = @_;
    my $hc = $opts{hc};
    my $ref = ref($hc);
    if (! $hc->isa("NACL::MTask::HealthChecker")) {
        my $error = "HealthChecker should be of type 'NACL::MTask::HealthChecker' " .
                    "but it is type of " . $ref;
        NACL::AppBuilder::Exceptions::HealthChecker->throw($error);
    }
    my ($name) = ($ref =~ /::(\w+)$/);
    $hc->name($name);
    $Registered_Health_Checkers{$name} = $hc;
    my @nodes = $self->nodes();
    my @hc_nodes = $hc->nodes();
    if (scalar(@hc_nodes) == 0) {
        $hc->nodes($self->nodes());
    }
}

sub _register_hc_using_name {
    my ($self,%opts) = @_;
    my $hc = $opts{hc};
    my $package = $HC_Name_To_Package{$hc};
    if (defined $package) {
        if ($package !~ /NACL::MTask::HealthChecker::/) {
            $package = "NACL::MTask::HealthChecker::" . $package;
        }
    } else {
        my $error = "Invalid HealthChecker name " . $hc . " passed." .
                    "Valid HealthChecker names are " . join(" ",keys(%HC_Name_To_Package));
        NACL::AppBuilder::Exceptions::HealthChecker->throw($error);
    }

    eval "require $package";
    if ($@) {
        NATE::BaseException->throw("Failed to load package $package. Error: " . $@ );
    }

    my @nodes = $self->nodes();
    my $hc_obj = $package->new(nodes => \@nodes);
    $hc_obj->name($hc);
    $Registered_Health_Checkers{$hc} = $hc_obj;
}

=head2 hc_unregister

    my $hc_obj = NACL::MTask::HealthChecker::FilerHealth->new(
        nodes => \@nodes
    );

    # register first
    $HCD->hc_register(
        hc => $hc_obj,
    );

    $HCD->hc_unregister(
        hc => $hc_obj,
    );

OR

    # register first
    $HCD->hc_register(
        hc => $hc_name,
    );

    my $HCD->hc_unregister(
        hc => $hc_name,
    );

Method to unregister HealthChecker from Driver.

=over

=item Options

=over

=item C<< hc => $hc >>

(Required)
It can either be a key in %HC_Name_To_Package OR
object of type NACL::MTask::HealthChecker::*

=back

=item Returns

HealthChecker to be unregistered .
Will return undef if HealthChecker is not already registered.

=back

=cut

sub hc_unregister {
    my ($self, @args) = @_;
    state $spec = {
        hc => {type => OBJECT|SCALAR},
    };

    my %opts = validate_with(
        params => \@args,
        spec   => $spec,
    );

    my $hc = $self->_get_hc_by_name($opts{hc});

    return (delete $Registered_Health_Checkers{$hc});
}

=head2 hc_disable

    my $hc_obj = NACL::MTask::HealthChecker::FilerHealth->new(
        nodes => \@nodes
    );

    # register first
    $HCD->hc_register(
        hc => $hc_obj,
    );

    $HCD->hc_disable(
        hc => $hc_obj,
    );

OR

    # register first
    $HCD->hc_register(
        hc => $hc_name,
    );

    my $HCD->hc_disable(
        hc => $hc_name,
    );

Method to disable HealthChecker. It will NOT delete hc from
the list of %Registered_Health_Checkers. hc can be enabled later
using hc_enable method

=over

=item Options

=over

=item C<< hc => $hc >>

(Required)
It can either be a key in %HC_Name_To_Package OR
object of type NACL::MTask::HealthChecker::*

=back

=back

=cut

sub hc_disable {
    my ($self, @args) = @_;
    state $spec = {
        hc => {type => OBJECT},
    };

    my %opts = validate_with(
        params => \@args,
        spec   => $spec,
    );

    my $hc = $self->_get_hc_by_name($opts{hc});
    my $hc_obj = $Registered_Health_Checkers{$hc};
    $hc_obj->is_enabled(0);
}

=head2 hc_enable

    my $hc_obj = NACL::MTask::HealthChecker::FilerHealth->new(
        nodes => \@nodes
    );

    # register first
    $HCD->hc_register(
        hc => $hc_obj,
    );

    $HCD->hc_enable(
        hc => $hc_obj,
    );

OR

    # register first
    $HCD->hc_register(
        hc => $hc_name,
    );

    my $HCD->hc_enable(
        hc => $hc_name,
    );

Method to re-enable HealthChecker. Will have no effect if hc is
already enabled.

=over

=item Options

=over

=item C<< hc => $hc >>

(Required)
It can either be a key in %HC_Name_To_Package OR
object of type NACL::MTask::HealthChecker::*

=back

=back

=cut

sub hc_enable {
    my ($self, @args) = @_;
    state $spec = {
        hc => {type => OBJECT},
    };

    my %opts = validate_with(
        params => \@args,
        spec   => $spec,
    );

    my $hc = $self->_get_hc_by_name($opts{hc});
    my $hc_obj = $Registered_Health_Checkers{$hc};
    $hc_obj->is_enabled(1);
}

=head2 hcs_start

    $HCD->hcs_start(
        health_checkers => \@hcs,
    );

Method to start all health checkers passed in @hcs.

=over

=item Options

=over

=item C<< health_checkers => @hcs >>

(Required)
List of health checkers to be start by Driver

=back

=back

=cut

sub hcs_start {
    my ($self, @args) = @_;
    state $spec = {
        health_checkers => {type => ARRAYREF},
    };

    my %opts = validate_with(
        params => \@args,
        spec   => $spec,
    );

    my @hcs = $self->enabled_hcs(%opts);
    foreach my $hc_obj (@hcs) {
        if ($hc_obj->is_enabled) {

            my $coderef = sub {
                my ($hc) = @_;
                try {
                    $hc->start();
                } finally {
                    # Listner should have been stopped by now
                    # but sometime in case of failures its not
                    # So removing the listner
                    $hc->_write_result_to_parent();
                };
            };

            my $runid = $self->{counter}++ . "_" . $hc_obj->name();
            my $proc = NATE::Process->new(
                codespec => $coderef,
                runid => $runid,
                background => 0,
                args  => [$hc_obj],
                onexit => \&NATE::Process::on_exit_propagate_worst_result,
            );

            $proc->start();
            $self->_read_result_from_hc(hc => $hc_obj);
        }
    }

}

=head2 is_any_hcs_failed

    $HCD->is_any_hcs_failed(
        health_checkers => \@hcs,
    );

checks whether any health_checker has failed or NOT.

=over

=item Returns

(bool)Returns true if any of the health_checker reported failure.
Else returns true.

=back

=cut

sub is_any_hcs_failed {
    my ($self) = @_;
    my @failed_hcs = $self->failed_hcs();

    return (scalar(@failed_hcs) > 0);

}

=head2 failed_hcs

    $HCD->failed_hcs(
        health_checkers => \@hcs,
    );

Returns the list of failed health_checkers.

=over

=item Returns

(Array)Returns the list of failed health_checkers.

=back

=cut

sub failed_hcs {
    my ($self) = @_;
    my @enabled_hcs = $self->enabled_hcs();

    my @failed_hcs = ();
    foreach my $hc(@enabled_hcs) {
        if ($hc->is_failed) {
            push(@failed_hcs, $hc);
        }
    }

    return @failed_hcs;
}

=head2 registered_hcs

    $HCD->registered_hcs();

Returns the list of registered health_checkers.

=over

=item Returns

(Array)Returns the list of registered health_checkers.

=back

=cut

sub registered_hcs {
    my ($self) = @_;
    my @registered_hcs = (values %Registered_Health_Checkers);

    return @registered_hcs;
}

=head2 enabled_hcs

    $HCD->enabled_hcs(
        health_checkers => \@hcs,
    );

Returns the list of health_checkers which are not disabled.

=over

=item Returns

(Array)Returns the list of health_checkers which are not disabled.

=back

=cut

sub enabled_hcs {
    my ($self, %opts) = @_;
    my @hcs;
    if (defined $opts{health_checkers}) {
        @hcs = @{$opts{health_checkers}};
    } else {
        @hcs = $self->registered_hcs();
    }

    my @enabled;
    foreach my $hc (@hcs) {
        if ($hc->is_enabled) {
            push(@enabled, $hc);
        }
    }

    return @enabled;
}

=head2 health_check_all

    $HCD->health_check_all();

Method to start all registered health checkers.

=cut


sub health_check_all {
    my ($self) = @_;

    my @hcs = $self->registered_hcs();
    $self->hcs_start(health_checkers => \@hcs);

    return $self->results();
}

=head2 health_check

    $HCD->health_check(
        health_checkers => \@hcs,
    );

Same as health_check_all if health_checkers are not passed.
else same as hcs_start.

=over

=item Options

=over

=item C<< health_checkers => @hcs >>

(Optional)
List of health checkers to be start by Driver

=back

=back

=cut

sub health_check {
    my ($self, %opts) = @_;

    if (!defined $opts{health_checkers}) {
        return $self->health_check_all();
    }

    $self->hcs_start(%opts);

    return $self->results();
}


=head2 results

    $HCD->results();

Returns consolidated result of all health checkers.

=over

=item Returns

=over

Returns result hashref in the following format.

Result => {
    HealthCheckerModuleName1 =>
              {
                   PASS =>  {
                            [checker_method1, checker_method1]
                            },
                   FAIL =>  { [
                              {checker_method1 => 'Error Message for checker_method1},
                              {checker_method2 => 'Error Message for checker_method2}
                              ]
                            }
               },
    HealthCheckerModuleName2 => {....},
    ........,
   }


=back

=back

=cut

sub results {
    my ($self) = @_;

    my $result = $self->{results} || {};
    return $result;
}

sub _read_result_from_hc {
    my ($self, @args) = @_;
    state $spec = {
        hc => {type => OBJECT},
    };

    my %opts = validate_with(
        params => \@args,
        spec   => $spec,
    );

    my $log = log_global();
$may_enter = $Log->may_enter();
$may_exit  = $Log->may_exit();
    my $hc = $opts{hc};
    my $msnger = messenger_global(transport => 'zeromq');

    # set up a listener for messages that have
    # type match 'qr/NACL_HealthChecker_<hc_name>/i'
    my $stop = 0;
    my $msg_type = "NACL_HealthChecker_" . $hc->name();

    my $listener = {
        filter   => sub { shift->type() =~ /$msg_type/ },
        callback => sub {
            my ($msg_obj) = shift;
            # reply to the sender acknowledging that we got their message
            my $msg = $msg_obj->message;
            foreach my $key(keys %$msg) {
                push(@{$self->{results}->{$hc->name}->{$key}},@{$msg->{$key}});
            }
            if(defined $msg_obj->message()->{FAIL}) {
                $hc->is_failed(1);
            }
            $stop = 1;
        },
        charges => -1 # never let this listener expire
    };
    $msnger->listener_add(listener => $listener);

    # sit in an eventloop until someone sets $stop = 1
    while ($stop == 0) {
        eventloop_drain();
    }

    # remove the listener since we set it to never expire
    $msnger->listener_remove(listener => $listener);
}

sub _init_messenger {
    my ($self) = @_;

    my $msnger = messenger_global(transport => 'zeromq');

    # prepare a listening socket so we can receive messages from others
    $msnger->init(
        socket_type => 'rep',
        protocol    => 'tcp',
        address     => '*', # basically means localhost,
        port        => $MESSENGER_ZEROMQ_PORT,
    );

}

sub _get_hc_by_name {
    my ($self, $hc) = @_;
    if (blessed $hc) {
        $hc->name();
    } else {
        return $hc;
    }
}

1;
