
#
# Copyright (c) 2001-2012 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Lun Controller Layer Module
## @author dl-nacl-dev
## @status shared
## @pod here

=head1 NAME

NACL::AppBuilder::SC::Controller::Lun

=head1 DESCRIPTION

NACL::AppBuilder::SC::Controller::Lun is a derived class of
NACL::AppBuilder::SC::Controller::Task.  It is used to Create and
Delete ONTAP Luns. All methods are static.

=cut

package NACL::AppBuilder::SC::Controller::Lun;

# Compiler directives
use strict;
use warnings;

# NATE and Standard Module import
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use NATE::Util::DataTypes qw(TRUE FALSE);
use Params::Validate
    qw(validate_with validate SCALAR ARRAYREF BOOLEAN UNDEF SCALARREF HASHREF);

# NACL and EnGenX Libs
use NACL::C::Lun;
use NACL::CS::Volume;
use NACL::STask::Lun;
use NACL::C::Job;
use NACL::AppBuilder::EnGenX::SubProcess;
use NACL::AppBuilder::EnGenX::EnGenXUtil;

# Exceptions
use NATE::BaseException qw(:try);
use NATE::Exceptions::Argument qw(:try);
use NACL::ComponentUtils qw(Dumper);

#Inherit base class
use base qw( NACL::AppBuilder::SC::Controller::Task );

use vars qw(%DTD_To_NACL_Translate
    %DTD_To_NACL_Ignore
    %Delete_Args);

# Arguments after converting from xml to a
# resource list that need their keys translated to
# keys that NACL methods will accept.
%DTD_To_NACL_Translate = (
    name   => 'lun',
    _reuse => 'if_exists'
);

# Arguments that are used by EnGenX but cannot be
# passed into the NACL methods
%DTD_To_NACL_Ignore = (
    parallel                       => '',
    map_list                       => '',
    require_same_lun_id_per_igroup => '',
    'lun-id'                       => '',
    lun_id_range                   => '',
);

# Arguments that can be passed into the NACL delete
# command
%Delete_Args = (
    command_interface => '',
    apiset_must       => '',
    apiset_should     => '',
    lun               => '',
    qtree             => '',
    volume            => '',
    vserver           => '',
    path              => '',
    force             => '',
);

=head1 METHODS

=head2 create

(Static method) Request creation of one or multiple luns serially
or in parallel and return whether the lun(s) was created successfully.
This method passes all user arguments through to EnGenX::Controller::Task
and adds specific arguments used by the EnGenX framework.

  NACL::AppBuilder::SC::Controller::Lun->create(
            command_interface => $Node,
            volume            => ( $Volume_name),
            lun                  => $Lun_name,
            vserver           => $Vserver_name,
            size              => "40M",
            ostype              => 'windows',
            'method-timeout'  => 3600,
            );

  NACL::AppBuilder::SC::Controller::Lun->create(
            command_interface => $Node,
            volume            => ( $Volume_name),
            lun                  => $Lun_name,
            vserver           => $Vserver_name,
            size              => "40M",
            ostype              => 'windows',
            'method-timeout'  => 3600,
            count             => 5
            );

  NACL::AppBuilder::SC::Controller::Lun->create(
            command_interfac  => $Node,
            resource_list      => @Luns,
            'method-timeout'  => 3600,
            );

  NACL::AppBuilder::SC::Controller::Lun->create(
            command_interface => $Node,
            resource_list      => @Luns,
            );

  NACL::AppBuilder::SC::Controller::Lun->create(
            resource_list      => @Luns,
            'method-timeout'  => 3600,
            );

  NACL::AppBuilder::SC::Controller::Lun->create(
            resource_list      => @Luns,
            );

=over

=item Options

Either the resource_list is required or the NACL Lun Creation Arguments
are required but both should NOT be used in the same call.

=over

=item C<< resource_list => @resource_list >>

(Optional)
If this is passed in the luns in the resource list will be created
based on the information that is supplied in the resource list. If
the resource list is passed in the NACL Lun Creation Arguments should
NOT be passed in.

=item C<< NACL Lun Creation Arguments >>

(Optional)
These are arguments that can be passed into NACL::C::Lun->Create.
If these arguments are passed in the resource_list argument should NOT
be passed in

=item C<< count => $lun_count >>

(Optional)
This arguments is used to create multiple luns from the same NACL Lun
Creation Arguments.  To create the names for the luns, the lun name
passed in is appended with the index that it was create at starting at 1.

=item C<< command_interface => $NACL_command_interface >>

(Optional)
This is the command_interface that will be used during NACL calls.  If a
command_interface is not passed in then one will be created.  Passing in a
command_interface helps performance because for every command_interface created
the version of the Filer must be determined by creating a connection to the filer
and then sending commands to get the version.

=item C<< method_timeout => $NACL_method_timeout >>

(Optional)
This is the method_timeout that will be used during NACL calls. If a
method_timeout is not passed in then the method_timeout will be defaulted.

=back

=item Return

This method returns a True or False based on whether the method succeeds or fails
in the creation of the Lun(s) based on the Options passed in.

=item Exceptions

This method produces and throws no exceptions of its own but pass through any
uncaught exceptions from other library calls.

=cut

sub create {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %args = @_;

    # Locals
    my $parent_answer;

    # Globals that will also be set as globals in the base class
    $args{'Name'} = "lun";

    # Super::_create is in Task.pm
    $parent_answer = $pkg->SUPER::_create(%args);

    return $parent_answer;
} ## end sub create

=head2 delete

(Static method) Request deletion of one or multiple luns serially
or in parallel and return whether the lun(s) was deleted successfully.
This method passes all user arguments through to EnGenX::Controller::Task
and adds specific arguments used by the EnGenX framework.

  NACL::AppBuilder::SC::Controller::Lun->delete(
            command_interface => $Node,
            volume            => ( $Volume_name),
            lun                  => $Lun_name,
            vserver           => $Vserver_name,
            'method-timeout'  => 3600,
            );

  NACL::AppBuilder::SC::Controller::Lun->delete(
            command_interface => $Node,
            volume            => ( $Volume_name),
            lun                  => $Lun_name,
            vserver           => $Vserver_name,
            count             => 5
            );

  NACL::AppBuilder::SC::Controller::Lun->delete(
            command_interface => $Node,
            resource_list      => @Luns,
            'method-timeout'  => 3600,
            );

  NACL::AppBuilder::SC::Controller::Lun->delete(
            command_interface => $Node,
            resource_list      => @Luns,
            );

  NACL::AppBuilder::SC::Controller::Lun->delete(
            resource_list      => @Luns,
            'method-timeout'  => 3600,
            );

  NACL::AppBuilder::SC::Controller::Lun->delete(
            resource_list      => @Luns,
            );

=over

=item Options

Either the resource_list is required or the NACL Lun Deletion Arguments
are required but both should NOT be used in the same call.

=over

=item C<< resource_list => @resource_list >>

(Required)
If this is passed in the luns in the resource list will be deleted
based on the information that is supplied in the resource list. If
the resource list is passed in the NACL Lun Deletion Arguments should
NOT be passed in.

=item C<< NACL Lun Deletion Arguments >>

(Required)
These are arguments that can be passed into NACL::C::Lun->Delete.
If these arguments are passed in the resource_list argument should NOT
be passed in.

=item C<< count => $lun_count >>

(Optional)
This arguments is used to delete multiple luns from the same NACL Lun
Deletion Arguments.  To delete the names for the luns, the lun name
passed in is appended with the index that it was deleted at starting at 1.

=item C<< command_interface => $NACL_command_interface >>

(Optional)
This is the command_interface that will be used during NACL calls.  If a
command_interface is not passed in then one will be created.  Passing in a
command_interface helps performance because for every command_interface created
the version of the Filer must be determined by creating a connection to the filer
and then sending commands to get the version.

=item C<< method_timeout => $NACL_method_timeout >>

(Optional)
This is the method_timeout that will be used during NACL calls. If a
method_timeout is not passed in then the method_timeout will be defaulted.

=back

=item Return

This method returns a True or False based on whether the method succeeds or fails
in the deletion of the Lun(s) based on the Options passed in.

=item Exceptions

This method produces and throws no exceptions of its own but pass through any
uncaught exceptions from other library calls.

=cut

sub delete {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %args = @_;

    # Locals
    my $parent_answer;

    # Globals that will also be set as globals in the base class
    $args{'Name'} = "lun";

    # Super::_delete is in Task.pm
    $parent_answer = $pkg->SUPER::_delete(%args);

    return $parent_answer;
} ## end sub delete

#==============================================================================
# Method:     _nacl_create_call
# Objective:  Makes the raw NACL calls for creation of a lun.
# Details:    Makes the NACL::C::Lun->create() call. Also works
#             around some common non-fatal errors that can be return from
#             lun creation and then has a retry mechanism to retry lun
#             creation.
# Inputs:     NACL::C::Lun->create arguments
#             retry => integer, is the current retry the method is on.
#                    This is used to count how many times the _nacl_create_call
#                    has been retried so that the retries can be limited.
# Outputs:    Returns whether the create was successful or a hash with success
#             and whether the lun was a duplicate if a hash is requested
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
#==============================================================================
sub _nacl_create_call {
    $Log->enter() if $may_enter;
    my ($pkg, %args) = @_;

    # Locals
    my $duplicated = FALSE;
    my $state = exists($args{'state'}) ? delete($args{'state'}) : 'online';
    my $allocation
        = exists($args{'space-allocation'})
        ? delete($args{'space-allocation'})
        : 'disabled';
    my $if_exists
        = exists($args{if_exists}) ? delete($args{if_exists}) : 'reuse';

    if (!exists($args{command_interface})) {
        my @Nodes = NACL::C::Node->find();
        $args{command_interface} = $Nodes[0];
    }

    $pkg->_set_ci_defaults(command_interface => $args{command_interface});

    # If the size argument has a percent then calculate the size.
    if ($args{size} =~ s/%//) {
        my $volume_cs_obj = NACL::CS::Volume->fetch(
            command_interface => $args{command_interface},
            requested_fields  => ['size'],
            filter            => {
                vserver => $args{vserver},
                volume  => $args{volume}
            },
        );
        my $percent = $args{size};
        $args{size} = ($args{size} / 100) * ($volume_cs_obj->size());
        $Log->comment("Setting Lun size to "
                . $args{size}
                . "b which is "
                . "$percent\% of volume "
                . $args{volume}
                . " in vserver "
                . $args{vserver}
                . " which is size "
                . $volume_cs_obj->size()
                . "b");
    }

    my $was_created;
    $args{_was_created}       = \$was_created;
    $args{nacltask_if_exists} = $if_exists;

    # Execute Create
    my $lun = NACL::STask::Lun->create(%args);

    # Add in space allocation.
    # NOTE the lun modify attirbutes were not in the
    # NACL docs at the time this was writen.
    if (   ($allocation =~ /enabled/i)
        or ($state eq 'offline')
        or (!$was_created)) {

        $lun->_construct_path(params => \%args);
        $lun->modify(
            command_interface  => $args{command_interface},
            path               => $args{path},
            'state'            => $state,
            'space-allocation' => $allocation
        );
    }
    $duplicated = TRUE if (!$was_created);
    $Log->exit() if $may_exit;

    # Returning either a list (hash) or scalar TRUE
    return (wantarray())
        ? ('success' => TRUE, 'duplicated' => $duplicated)
        : TRUE;

} ## end sub _nacl_create_call

#========================================================================================
# Method:     _nacl_delete_call
# Objective:  Makes the raw NACL calls for deletion of a Lun.
# Details:    Makes the NACL::C::Lun->delete() calls. Also works around
#             some common non-fatal errors that can be return from Lun
#             deletion and then has a retry mechanism to retry Lun deletion.
# Inputs:     NACL::C::Lun->delete arguments
#             retry => integer, is the current retry the method is on.
#                    This is used to count how many times the _nacl_delete_call
#                    has been retried so that the retries can be limited.
# Outputs:    Returns whether the create was successful or a hash with success
#             if a hash is requested
# Exceptions: This method produces and throws no exceptions of its own but
#             passes through any uncaught exceptions from other library calls.
#==============================================================================
sub _nacl_delete_call {
    $Log->enter() if $may_enter;
    my ($pkg, %args) = @_;

    my $if_exists;
    if (!exists($args{command_interface})) {
        my @Nodes = NACL::C::Node->find();
        $args{command_interface} = $Nodes[0];
    }
    $pkg->_set_ci_defaults(command_interface => $args{command_interface});
    $args{nacltask_if_purged} = 'pass';
    NACL::STask::Lun->purge(%args,);
    $Log->exit() if $may_exit;
    return TRUE;

} ## end sub _nacl_delete_call

#==============================================================================
# Method:     _get_dtd_nacl_hashes (translate, ignore and delete)
# Objective:  Getter methods for the variables of the same name.
# Details:    Retrieves the hash or a reference to one of the following:
#             - DTD_To_NACL_Translate.
#             - DTD_To_NACL_Ignore.
#             - Delete_Args.
# Inputs:     none.
# Outputs:    The reference or hash variable.
# Exceptions:
#==============================================================================
sub _get_dtd_nacl_translate {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = @_;

    $Log->exit() if $may_exit;
    return (wantarray()) ? %DTD_To_NACL_Translate : \%DTD_To_NACL_Translate;
} ## end sub _get_dtd_nacl_translate

#==============================================================================
sub _get_dtd_nacl_ignore {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = @_;

    $Log->exit() if $may_exit;
    return (wantarray()) ? %DTD_To_NACL_Ignore : \%DTD_To_NACL_Ignore;
} ## end sub _get_dtd_nacl_ignore

#==============================================================================
sub _get_dtd_nacl_delete {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = @_;

    $Log->exit() if $may_exit;
    return (wantarray()) ? %Delete_Args : \%Delete_Args;
} ## end sub _get_dtd_nacl_delete

1;
