# Copyright (c) 2012-2013 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Workflow Framework Components config file reading/manipulating
## @author rjray@netapp.com
## @status shared
## @pod here

package NACL::AppBuilder::WFC::Config;

use strict;
use warnings;

use File::Basename 'dirname';
use Config::General qw(ParseConfig);
use Params::Validate qw(validate validation_options SCALAR);
use Readonly;

use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use NACL::ComponentUtils qw(_verify_invocation);
use NATE::BaseException qw(:try);
use NATE::Exceptions::Argument;
use NATE::Database::Test qw(testdb_global);

# Set the on_fail behavior for Params::Validate to throw an exception (of the
# NATE::Exceptions::Argument class) on errors instead of a plain croak().
validation_options(
    on_fail => sub {
        NATE::Exceptions::Argument->throw(shift);
    }
);

# The default config file is this name, which is assumed to be located in the
# same directory as this module.
Readonly my $DEFAULT_CONFIG_FILE => 'wfc_variants_default.conf';

# These are the keys in each variant record, in the order they should appear
# when we create tables for MediaWiki. If the columns are changed/added to in
# the config file, they need to be changed/added to here as well.
Readonly my @CONFIG_COLUMNS => qw(
    pedigree_src
    aggr_type_src
    snapshots_src
    dedupe
    compression
    inline_compression
    space_guaranteed_src
    clone
    pedigree_dst
    aggr_type_dst
    snapshots_dst
    space_guaranteed_dst
    san
);

=head1 NAME

NACL::AppBuilder::WFC::Config - Configuration file support for WFC

=head1 SYNOPSIS

    use NACL::AppBuilder::WFC::Config;

    # Using the default location for the config file:
    my $config = NACL::AppBuilder::WFC::Config->new();
    my $variant = $config->get_variant(variant => 'rw_250ss');

    # Specifying an explicit location for the config file:
    my $config = NACL::AppBuilder::WFC::Config->new(file => $config_file);
    my $variant = $config->get_variant(variant => 'rw_250ss');

=head1 DESCRIPTION

The B<NACL::AppBuilder::WFC::Config> class is a utility module that
encapsulates the reading and parsing of variants configuration files for
Workflow Framework Components-based scripts. The file is in a simple format
that is parsed by the CPAN module B<Config::General> (which is installed
site-wide at NetApp). Through the object, the user can easily obtain the data
for any given variant provided by the configuration file.

For details of the configuration file format itself, see
L</CONFIGURATION FORMAT>.

=head1 METHODS

=head2 new

    # Using the default configuration file:
    $conf = NACL::AppBuilder::WFC::Config->new();

    # Using a specific configuration file:
    $conf = NACL::AppBuilder::WFC::Config->new(file => $config_file);

(Must be called as a class method)

This method creates a new object of this class, an object which encapsulates
the content of the configuration file read. If an error occurs at some point
in reading or parsing the configuration file, an exception is thrown.

=over

=item Options

Only one option is recognized by B<new>:

=over

=item C<< file =E<gt> $filename >>

Specify an alternative configuration file to read and parse. The value of the
option is a scalar containing the file name. If this option is not passed, the
default is to use the file C<wfc_variants_default.conf> in the same directory
as this library.

=back

=item Exceptions

There are two possible exceptions that may be thrown by B<new>:

=over

=item B<NATE::Exceptions::Argument>

If a value other than a simple scalar is passed for the I<file> argument,
this exception will be thrown.

=item B<NATE::BaseException>

If an error occurs in either the reading or the parsing of the configuration
file itself, this exception will be thrown.

=back

=back

=cut

###############################################################################
#
#   Sub Name:       new
#
#   Description:    Create a new NACL::AppBuilder::WFC::Config object, reading
#                   the specified config file in the process.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $class    in      scalar    Class we're blessing into
#                   @opts     in      list      Additional options, for now
#                                                 just "file"
#
#   Returns:        Success:    blessed hash ref
#                   Failure:    throws exception
#
###############################################################################
sub new {
    $Log->enter() if $may_enter;
    my ($class, @opts) = @_;
    # This must be a static method invocation:
    $class->_verify_invocation(style => 'static_only');
    my %opts = validate(
        @opts,
        {
            file => { type => SCALAR, default => $DEFAULT_CONFIG_FILE },
        }
    );
    my $self = {};
    my $file = $opts{file};
    my (@content, %hash);

    # If the file is not an absolute path, it is assumed to be relative to the
    # location of this library (which is also the location of the scripts).
    if ($file !~ m{^/}) {
        my $dir = dirname __FILE__;
        $file = "$dir/$file";
    }

    # Check that the file exists and is readable before passing it to
    # Config::General for parsing:
    if (open(my $fh, '<', $file)) {
        @content = <$fh>;
        if (! close($fh)) {
            $Log->exit() if $may_exit;
            NATE::BaseException->throw("Error closing $file: $!");
        }
        # Get the order of the <variant> block names in the file
        $self->{variant_names} = [];
        for (@content) {
            if (/^<variant\s+(.*?)>/) {
                push(@{$self->{variant_names}}, $1);
            }
        }
    } else {
        $Log->exit() if $may_exit;
        NATE::BaseException->throw(
            "Error opening file $file for reading: $!"
        );
    }

    # Register to the S2T DB so that testSuggester can suggest CITs if config
    # files are modified.
    testdb_global()->register_file_for_s2t(file_path => $file);
    $self->{file} = $file;
    my $retval = eval {
        %hash = ParseConfig(
            -String         => \@content,
            -LowerCaseNames => 1,
            -AutoTrue       => 1,
        );
        1;
    };
    if (! $retval) {
        NATE::BaseException->throw("Error parsing $file: $@");
    }

    # Copy what we want from %hash into $self
    $self->{labels} = $hash{labels};
    $self->{variants} = $hash{variant};

    bless $self, $class;

    $Log->exit() if $may_exit;
    return $self;
}

=head2 get_variant

    $hashref = $conf->get_variant(variant => $variant);

(Must be called as an instance method, not callable as a class method.)

Retrieve a hash reference of the data for a given variant. Takes the name of
the variant and returns the data from the configuration file specific to it.

The return value is a hash reference (not an object) with the following keys.
The keys correspond to the columns in the table of variants in the wikid page
L<http://wikid.netapp.com/w/Cmode_ADR_QA_Rapid_Regression_Suite>:

=over

=item pedigree_src

The type of the SRC volume for the variant

=item aggr_type_src

The type of aggregate (C<64-bit> or C<32-bit>) to create the source volume on

=item snapshots_src

The number of snapshots that should be on the source volume

=item dedupe

Boolean value, whether dedupe is to be enabled on the source volume

=item compression

Boolean value, whether background compression is to be enabled on the source
volume

=item inline_compression

Boolean value, whether inline compression is to be enabled on the source volume

=item space_guaranteed_src

Boolean value, whether the source volume is created with space guarantee
enabled

=item clone

Boolean value, whether a FlexClone is used

=item pedigree_dst

The type of the DST volume for the variant

=item aggr_type_dst

The type of aggregate (C<64-bit> or C<32-bit>) to create the destination volume
on

=item snapshots_dst

The number of snapshots that should be on the destination volume, usually
C<n/a> (meaning that it is not changed from the source)

=item space_guaranteed_dst

Boolean value, whether the destination volume is created with space guarantee
enabled

=item san

Boolean value, whether the I/O done for this variant is done through SAN.

=back

The return value will always be a copy of the internal data, so changes to it
will not affect the larger configuration.

=over

=item Options

Only one option is recognized, and it is required:

=over

=item C<< variant =E<gt> $string >>

The string containing the name of the variant to look up. If it is not
found, an exception is thrown.

=back

=item Exceptions

The only exceptions that may be thrown by this method are:

=over

=item B<NATE::Exceptions::Argument>

If an unknown option is passed in, or if the value for C<variant> is not a
scalar, this exception will be thrown.

=item B<NATE::BaseException>

If the variant specified is not found in the configuration, this exception
will be thrown.

=back

=back

=cut

###############################################################################
#
#   Sub Name:       get_variant
#
#   Description:    Return a hashref of the data for a given named variant.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#                   @opts     in      list      Options, in this case just
#                                                 "variant => name"
#
#   Returns:        Success:    hashref
#                   Failure:    throws exception
#
###############################################################################
sub get_variant {
    $Log->enter() if $may_enter;
    my ($self, @opts) = @_;
    # This must be an instance method invocation:
    $self->_verify_invocation(style => 'instance_only');
    my %opts = validate(
        @opts,
        {
            variant => { type => SCALAR },
        }
    );

    my $variant = $opts{variant};
    if (! $self->{variants}->{$variant}) {
        NATE::BaseException->throw("No variant '$variant' found in table");
    }

    # Make a copy of the hash, so that the user cannot accidentally alter it
    my $variant_copy = { %{$self->{variants}->{$variant}} };

    $Log->exit() if $may_exit;
    return $variant_copy;
}

=head2 make_wiki_table

    # Generate the table text for a wiki page
    $table = $conf->make_wiki_table();

(Instance method only)

Generate MediaWiki-style table mark-up for the configuration contents in the
calling object. The table will contain a header row, then one row for each
variant in the B<NACL::AppBuilder::WFC::Config> object. The table headings are
taken from the C<< E<lt>labelsE<gt> >> block in the file (see L</CONFIGURATION
FORMAT>).

The method does not write to a file or connect to the wiki. It simply returns
a string containing the content. The user may then use the content however they
want.

The rows in the table are in the same order that the variants appeared in the
configuration file itself. No re-ordering or sorting of the variants is done.

=over

=item Options

No options are recognized by this method.

=item Exceptions

No exceptions are thrown by this method.

=back

=cut

###############################################################################
#
#   Sub Name:       make_wiki_table
#
#   Description:    Create the markup for a MediaWiki table from the content
#                   of the config on this object.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#
#   Returns:        Success:    table text
#                   Failure:    throws exception
#
###############################################################################
sub make_wiki_table {
    $Log->enter() if $may_enter;
    my ($self) = @_;
    # This must be an instance method invocation:
    $self->_verify_invocation(style => 'instance_only');

    my @variants = @{$self->{variant_names}};
    # These are used in the definition of the booleans table:
    my $OFF_ON     = [ qw(off on) ];
    my $FALSE_TRUE = [ qw(false true) ];
    # This table is used to translate the boolean fields from 0/1 to the
    # labels used in the wiki table:
    my %BOOLEAN = (
        dedupe               => $OFF_ON,
        compression          => $OFF_ON,
        inline_compression   => $OFF_ON,
        space_guaranteed_src => $OFF_ON,
        clone                => $FALSE_TRUE,
        space_guaranteed_dst => $OFF_ON,
        san                  => $FALSE_TRUE,
    );

    # Start the table
    my $table = qq({|border="1" cellpadding="5" cellspacing="0"\n) .
        q(| align="center" style="background:#f0f0f0;"|) .
        qq('''Name Key'''\n);
    # Generate the column headings
    for my $column (@CONFIG_COLUMNS) {
        my $label = $self->{labels}->{$column} || $column;
        $table .= q(| align="center" style="background:#f0f0f0;"|) .
            qq('''$label'''\n);
    }
    # Generate a row for each variant
    for my $variant_name (@variants) {
        my $variant = $self->{variants}->{$variant_name};
        my @data = ($variant_name);
        for my $column (@CONFIG_COLUMNS) {
            if ($BOOLEAN{$column}) {
                push(@data, $BOOLEAN{$column}->[$variant->{$column}]);
            } else {
                push(@data, $variant->{$column});
            }
        }
        $table .= "|-\n| " . join(q{||}, @data) . "\n";
    }
    # Finish the table
    $table .= "|-\n|}\n";

    $Log->exit() if $may_exit;
    return $table;
}

=head1 CONFIGURATION FORMAT

The format of the configuration file is based on the B<Config::General> CPAN
module, and resembles the Apache web server configuration file format. Blank
lines are ignored, and comments are specified by a C<#> character, with
everything to the right of the C<#> being ignored.

The file has one block that defines the human-readable labels for each
configuration value. This block is called C<E<lt>labelsE<gt>>, and occurs first
in the file. Each value for a label is given as a key/value pair separated by
an equals sign and optional whitespace. The block is closed off as if it were
an XML tag:

    <labels>
        pedigree_src         = Vol Pedigree SRC
        aggr_type_src        = Aggr Type SRC
        snapshots_src        = \# snapshots SRC
        ...
    </labels>

(Note that the value for C<snapshots_src> includes a hash-sign, so it is
escaped to keep from being interpreted as a comment marker.)

Following the labels block are the C<E<lt>variantE<gt>> blocks. As with the
labels block, these are delimited by XML-style tags. However, unlike the label
block, each variant block includes the name of the variant:

    <variant rw_250ss>
        pedigree_src         = RW
        aggr_type_src        = 64-bit
        snapshots_src        = 250
        dedupe               = off
        compression          = off
        inline_compression   = off
        space_guaranteed_src = on
        clone                = false
        pedigree_dst         = DP
        aggr_type_dst        = 64-bit
        snapshots_dst        = n/a
        space_guaranteed_dst = on
        san                  = true
    </variant>

Everything following the C<variant> in the opening tag is taken as the name.
Within the block is the same sequence of field-names as were specified in the
labels block. In these blocks the values for each field are the actual values
for that variant.

There is no limit to the number of variant blocks, but the names must be
unique. If a name is re-used it will overwrite the previous definition for that
name.

=cut

1;
