# Copyright (c) 2012-2013 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.

## @summary Workflow Framework Components I/O operations
## @author rjray@netapp.com
## @status shared
## @pod here

package NACL::AppBuilder::WFC::IO;

use strict;
use warnings;

use Module::Load;

use Params::Validate qw(
    validate validate_with validation_options
    UNDEF SCALAR HASHREF OBJECT BOOLEAN
);
use Readonly;
use List::Util 'first';
use List::MoreUtils 'any';
use Scalar::Util 'blessed';

use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use NACL::STask::VolumeFile;
use NACL::C::Client;
use NACL::C::Client::Mkfs;
use NACL::CS::Client::Sanlun;
use NACL::CS::NetworkInterface;
use NACL::CS::VolumeFile;
use NACL::CS::VserverIscsi;
use NACL::STask::Client::Mount;
use NACL::STask::Lun;
use NACL::STask::LunIgroup;
use NACL::STask::VserverCifsShare;
use NACL::MTask::Traffic;
use NACL::ComponentUtils qw(_verify_invocation);
use NATE::BaseException qw(:try);
use NATE::Exceptions::Argument;

# The default size of populate data, used unless the user specifies an
# explicit value.
Readonly my $DEFAULT_POPULATE_SIZE => '1g';
# The default populate method/plugin to use, unless the user specifies one.
Readonly my $DEFAULT_POPULATE_METHOD => 'NaclData';
# The base class all population plugins should derive from:
Readonly my $POPULATE_BASE_CLASS => __PACKAGE__ . '::Populate';

Readonly my %FILE_SIZES => (
    'k' => 1024,
    'm' => 1024 * 1024,
    'g' => 1024 * 1024 * 1024,
    't' => 1024 * 1024 * 1024 * 1024,
);

# Params::Validate fragments to avoid re-typing the same specs over and over:
Readonly my $OPTIONAL_SCALAR => {
    type     => SCALAR,
    optional => 1,
};
Readonly my $OPTIONAL_HASHREF => {
    type     => HASHREF,
    optional => 1,
};
Readonly my $OPTIONAL_BOOLEAN => {
    type     => BOOLEAN,
    optional => 1,
};

# Note that all of these elements are all named as private attributes, thus
# ensuring that the methods C::MM generates are also named as private. We
# will define our own public accessors that are read-only for users to call.
# Also, I will be providing an explicit new() constructor, so that users can
# use the "visible" accessor names as options, while the data is stored under
# the private names.
use Class::MethodMaker [
    scalar => [
        {
            -type    => 'NACL::C::Client',
            -forward => [ 'is_windows' ],
        },
        '_client',
    ],
    scalar => [
        { -type => 'NACL::STask::Volume' },
        '_volume',
    ],
    scalar => [
        {
            -type    => 'NACL::STask::Client::Mount',
            -forward => [ 'mount_point' ],
        },
        '_mount',
    ],
    scalar => [
        {
            -type => 'NACL::STask::Lun'
        },
        '_lun',
    ],
    scalar => [
        {
            -type => 'NACL::STask::LunIgroup'
        },
        '_igroup',
    ],
    scalar => '_mount_access',
    scalar => '_traffic',
    scalar => '_san',
    scalar => '_lun_device',
    hash   => '_data_objects',
    scalar => '_data_current',
    scalar => '_data_method',
];

# These are the publically-visible accessors, all read-only:
for my $method (qw(client volume mount mount_access san lun igroup
                   data_method)) {
    ## no critic(ProhibitNoStrict)
    no strict 'refs';
    my $private = "_$method";
    *{$method} = sub {
        $Log->enter() if $may_enter;
        my $self = shift;

        $Log->exit() if $may_exit;
        return $self->$private();
    };
}

# Set the on_fail behavior for Params::Validate to throw an exception (of the
# NATE::Exceptions::Argument class) on errors instead of a plain croak().
validation_options(
    on_fail => sub {
        NATE::Exceptions::Argument->throw(shift);
    }
);

=head1 NAME

NACL::AppBuilder::WFC::IO - Data I/O generation for WFC

=head1 SYNOPSIS

    use NACL::AppBuilder::WFC::IO;

    # Create a handle on I/O generation:
    $io_handle = NACL::AppBuilder::WFC::IO->new(
        volume  => $volume,
        client  => $client
    );

    # Create the mount-point:
    $io_handle->create_mount(
        nfs_options => $opts,
        rw          => 1
    );

    # Populate some data:
    $io_handle->populate(size => '3g');

    # Unmount the volume:
    $io_handle->delete_mount();

=head1 DESCRIPTION

The B<NACL::AppBuilder::WFC::IO> class is used to do data population and/or I/O
traffic on volumes within the Workflow Framework Components lifecycle scripts.
It uses the client specification and examination of the volume to determine
whether to use NFS, CIFS or SAN protocols for the data. Ideally, the lifecycle
script should have little or no inherent knowledge of whether it is generating
traffic via NFS, CIFS or SAN.

An object of this class is capable of:

=over

=item *

B<Creating and deleting mount points>: Mount points may be created and deleted
in a way that removes some of the concern over Windows vs. UNIX (though some
additional information may be needed in some cases).

=item *

B<Populating data on a volume>: Data is populated using the
B<NACL::MTask::Data> library or other choices, depending on the user's options,
with the user not having to worry about the platform of the client (Windows vs.
UNIX) or whether SAN is involved.

=item *

B<Starting and stopping I/O traffic>: I/O traffic is done via the
B<NACL::MTask::Traffic> library, allowing the user to focus on just the
starting and stopping of traffic, not having to worry about other issues
related to starting sub-jobs or monitoring them for errors.

=item *

Initializing SAN elements on the volume, to enable population and traffic to
be done over SAN rather than NFS or CIFS.

=back

Note that data population and I/O traffic generation are very different tasks,
and that a user may do one without being compelled to do the other as well.
What I/O operations a script makes use of are up to the user writing the
script.

=head1 ACCESSORS

This class has the following property accessors. Each of these is a read-only
accessor, that cannot change the value of the underlying property. All of the
properties are read-only; to operate on a different volume or with a different
client, you will have to create a new B<NACL::AppBuilder::WFC::IO> object.

=over

=item client

The B<NACL::C::Client> object that represents the client this object was
instantiated with. If this object was only passed a client name, this property
will still return an object referent.

=item volume

The B<NACL::STask::Volume> object that represents the volume this object was
instantiated with.

=item mount

The B<NACL::STask::Client::Mount> object that represents the current
mount-point for the client/volume pair contained within this object. If there
is no current mount-point, this accessor returns C<undef>.

=item san

A boolean value that tells whether the volume being used is using SAN protocol
or not.

=item lun

If B<san> is set to true, then this accessor will return a B<NACL::STask::Lun>
object that represents the LUN on the volume that SAN I/O is being done
through. If B<san> is false, this accessor returns C<undef>.

=item igroup

If B<san> is set to true, then this accessor will return a
B<NACL::STask::LunIgroup> object that represents the initiator group that is
mapping the LUN to B<client>. If B<san> is false, this accessor returns
C<undef>.

=back

=head1 METHODS

This class provides the following methods. No subroutines are exported by this
package.

=head2 new

    $io_handle = NACL::AppBuilder::WFC::IO->new(
        volume  => $volume_obj,
        client  => $client_name_or_obj
    );

(Must be called as a class method)

Create a new I/O handle object and return it to the caller. Does not create any
mount-points or CIFS shares, SAN LUNs, etc. Derives from the arguments whether
the eventual data population will be via NFS, CIFS, etc.

All failures in B<new> throw an exception. See C<Exceptions>, below.

=over

=item Options

The following options are recognized by B<new>:

=over

=item C<< volume =E<gt> $volume_object >>

(Required) A volume object that represents the volume that will receive the
generated data. This must be either a B<NACL::C::Volume> or a
B<NACL::STask::Volume> instance. Additionally, it should already be mounted on
the filer itself. If the volume is a source volume created by the
B<NACL::AppBuilder::WFC::VolumeFactory> package, it will have been mounted for
you. If it was created outside of B<NACL::AppBuilder::WFC::VolumeFactory>, or
if it is a volume that was previously a DP volume that has been converted to RW
by means of a C<snapmirror break> operation, then you will have to mount it
explicitly prior to passing to this constructor.

=item C<< client =E<gt> $client_object_or_name >>

(Required) A client object or the name of a client. If a name is passed, a
client object will be created for you. If an object is passed, it must be
either a B<NACL::C::Client> instance or a B<NACL::STask::Client> instance.
This represents the client that will mount the volume and perform the actual
data population.

=item C<< lun =E<gt> $lun_object_or_name >>

(Optional) A LUN object or LUN name. If an object is passed, it must be an
instance of B<NACL::STask::Lun>. In both cases, the LUN must already exist
on the volume specified by the B<volume> option. If this option is given
(and is valid) then the newly-created object is marked to do I/O through
SAN.

If this option is not given, then the volume specified by B<volume> is
searched for a LUN. If one is found, then the found LUN is used and SAN
I/O is enabled. When searching, B<new> first looks for a LUN with the name
"I<volume-name>_lun". If that is not found, then the first LUN found by a
C<lun show> command is chosen. If no LUN is found at all, SAN I/O is not
enabled.

If the B<lun> option is passed with a false value (such as C<0> or C<undef>),
then the search for a LUN is disabled and SAN I/O is not enabled.

=back

=item Exceptions

The following exceptions may be thrown by B<new> in cases of failure:

=over

=item B<NATE::Exceptions::Argument>

If a required parameter is missing, or if the value of a paramter is invalid,
this exception will be thrown.

=item B<NATE::BaseException>

This exception may be thrown in cases where the variant specifies SAN I/O, but
the client is of a type other than Linux. At present, SAN I/O is only supported
with Linux clients.

=back

=back

=cut

###############################################################################
#
#   Sub Name:       new
#
#   Description:    Create a handle object to be used for the other method
#                   calls.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $class    in      scalar    Class we're blessing into
#                   @opts     in      list      Additional options
#
#   Globals:        $CONFIG
#
#   Returns:        Success:    object
#                   Failure:    throws exception
#
###############################################################################
sub new {
    $Log->enter() if $may_enter;
    my ($class, @opts) = @_;
    # This must be a static method invocation:
    $class->_verify_invocation(style => 'static_only');
    my %opts = validate(
        @opts,
        {
            volume => {
                type => OBJECT,
                isa  => 'NACL::C::Volume',
            },
            client => {
                type      => SCALAR | OBJECT,
                callbacks => {
                    'correct object type' => sub {
                        my $obj = shift;
                        # We know from type-validation that this is either a
                        # plain scalar or a blessed ref:
                        return (ref $obj) ? $obj->isa('NACL::C::Client') : 1;
                    },
                },
            },
            lun    => {
                type      => UNDEF | SCALAR | OBJECT,
                optional  => 1,
                callbacks => {
                    'correct object type' => sub {
                        my $obj = shift;
                        # We know from type-validation that this is either a
                        # plain scalar or a blessed ref:
                        return (ref $obj) ? $obj->isa('NACL::STask::Lun') : 1;
                    },
                },
            },
        }
    );
    my $self = bless {}, $class;
    # If "client" is a string, use it to instantiate an object:
    if (! ref($opts{client})) {
        $opts{client} = NACL::C::Client->new(name => $opts{client});
    }
    # If "volume" is not a NACL::STask::Volume, cast it:
    if (! $opts{volume}->isa('NACL::STask::Volume')) {
        $opts{volume} = $opts{volume}->cast_component_to_task;
    }
    # Set client and volume on the new object
    $self->_volume($opts{volume});
    $self->_client($opts{client});

    # Check for SAN. If they passed a value in for "lun", try to use that (as
    # either an object or a LUN name). If there is no value for "lun", check
    # the volume for any pre-existing LUN. However, if they passed "lun" as a
    # boolean false or undef, skip LUN checking entirely.
    my $skip_lun_check = exists($opts{lun}) && (! $opts{lun});
    if (! $skip_lun_check) {
        if ($opts{lun}) {
            # They either gave a LUN name or a LUN object. If it is the object,
            # confirm that it belongs to the volume. Otherwise, try to find it
            # on the volume. Due to the validate() call, above, we know that
            # this is either a scalar or specifically a NACL::STask::Lun
            # object.
            $self->_san(1);
            if (ref $opts{lun}) {
                if ($opts{lun}->get_one_state_attribute('volume') eq
                        $opts{volume}->volume) {
                    $self->_lun($opts{lun});
                } else {
                    $Log->exit() if $may_exit;
                    NATE::Exceptions::Argument->throw(
                        'The given LUN (' . $opts{lun}->path . ') does not ' .
                        'belong to the given volume (' .
                        $opts{volume}->volume . ')'
                    );
                }
            } else {
                my $lun = $self->_find_lun(lun_name => $opts{lun});
                if ($lun) {
                    $self->_lun($lun);
                } else {
                    $Log->exit() if $may_exit;
                    NATE::Exceptions::Argument->throw(
                        'The given LUN (' . $lun->path . ') was not found on ' .
                        'the given volume (' . $opts{volume}->volume . ')'
                    );
                }
            }
        } else {
            # No option passed. Look first for "volname_lun", then for any LUN
            # at all. If none are found, we just aren't doing SAN. (Yet.)
            my $lun;
            my $lunname = $opts{volume}->volume . '_lun';
            if (($lun = $self->_find_lun(lun_name => $lunname)) ||
                    ($lun = $self->_find_lun)) {
                $self->_lun($lun);
                $self->_san(1);
            } else {
                $self->_san(0);
            }
        }
    }

    # If the SAN flag is set, we detected an existing LUN. Make sure that it
    # is usable by ensuring that it is mapped and visible to the client.
    if ($self->_san) {
        # Sanity check: At present (10/11/2013), we cannot do SAN on Windows,
        # only on Linux.
        if ($opts{client}->type !~ /linux/) {
            $Log->exit() if $may_exit;
            NATE::BaseException->throw(
                'SAN I/O can currently only be done on Linux clients'
            );
        }

        $self->_setup_existing_lun;
    }

    # Init the _traffic listref.
    $self->_traffic([]);

    $Log->exit() if $may_exit;
    return $self;
}

=head2 create_mount

    $io_handle->create_mount(
        nfs_options  => $nfs_params_hash,
        cifs_options => $cifs_params_hash,
        access       => 'RW',
        is_ls        => 0
    );

(Must be called as an instance method)

Create the mount-point for the combination of volume and client that are
encapsulated in this object. Whether the mount point is for NFS or CIFS is
dependent on whether the client is a Windows host or not.

If there is an existing mount for this client/volume pair, it is removed and
a new one created. This is to prevent mounts that are read-write or read-only
from interfering in an attempt to create the opposite type.

If the operation is successful, the calling object is returned. Otherwise, an
exception is thrown.

=over

=item Options

The following options are recognized by this method. Note that certain options
will be silently ignored if they are not applicable; for example, the
C<nfs_options> value is ignored when creating a CIFS mount. This allows all
options to be passed all the time, without causing errors.

=over

=item C<< nfs_options =E<gt> $hashref >>

(Optional) Must be a hash-reference of the NFS-related options from the
B<NACL::C::Client::Mount::create()> method. These options are:

    proto ('tcp')
    vers (3)
    mountopts ('hard,intr')

Any of these three that are provided will be passed unchanged to the call to
create a NFS mount. Any keys other than these three will be ignored. Each of
these three has a default value (shown with the option, above). If the option
is not passed at all, the default value will be used. If you wish to suppress
the default value without providing a new value in the process, pass C<undef>
as a value:

    $io_handle->create_mount(
        # Other two NFS options will use their defaults:
        nfs_options => { mountopts => undef },
        ...
    );

Note that if SAN is in use, the NFS options are ignored.

=item C<< cifs_options =E<gt> $hashref >>

(Optional) Options to be used in the creation and mounting of a CIFS share
prior to mounting on a Windows client. This option is only used when the client
is a Windows machine. Must be a hash-reference. Unlike C<nfs_options>, above,
the different keys are used at different stages. The recognized keys are:

    user ('labclient')
    password ('sundance')
    properties (no default)

Each parameter has a default value shown above in parentheses (except for
C<properties>, which has no default). Any keys other than these three will be
ignored.

The value of the C<properties> option is passed directly as the
C<share-properties> option to B<NACL::C::VserverCifsShare::create()>, but only
if the share does not already exist (in which case the call to create is not
made). The C<user> and C<password> options are used when actually mounting the
share on the Windows client.

=item C<< access =E<gt> 'RW'|'RO' >>

(Optional) Determines whether the mount is created as read-only or read-write.
The default is C<RW> (read-write), and the value is treated as case-insensitive
so C<rw> is also acceptable.

=item C<< is_ls =E<gt> 0|1 >>

(Optional) A boolean value that tells whether the volume being mounted is
part of a LS mirror relationship. This affects the path used to mount the
volume when mounting it as read-write. The default is C<0> (false), meaning
that the volume is assumed to B<not> be part of a LS relationship.

=back

=item Exceptions

If an invalid value is passed for any of the options, an exception of the
class B<NATE::Exceptions::Argument> will be thrown.

Any of the operations for creating a CIFS share, SAN LUN, or mounting any of
these (including NFS) may throw exceptions. See the related manual pages for
the exceptions they can throw. If you need to catch an exception from this
method, it is probably best to catch B<NATE::BaseException> as an umbrella for
anything that could be thrown by the various libraries being used.

=back

=cut

###############################################################################
#
#   Sub Name:       create_mount
#
#   Description:    Create a mount-point for this object, based on the volume
#                   and client stored within. May also involve creating a
#                   CIFS share for the volume, if one does not already exist.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#                   @opts     in      array     Additional options, converted
#                                                 to a hash by validate()
#
#   Returns:        Success:    $self
#                   Failure:    throws exception
#
###############################################################################
sub create_mount {
    $Log->enter() if $may_enter;
    my ($self, @opts) = @_;
    # This must be an instance method invocation:
    $self->_verify_invocation(style => 'instance_only');
    my %opts = validate(
        @opts,
        {
            nfs_options  => $OPTIONAL_HASHREF,
            cifs_options => $OPTIONAL_HASHREF,
            is_ls        => $OPTIONAL_BOOLEAN,
            access       => {
                type      => SCALAR,
                default   => 'RW',
                callbacks => {
                    value => sub {
                        my $value = shift;
                        return ($value =~ /^r[wo]$/i);
                    },
                },
            },
        }
    );

    # These are the default options for NFS. Anything the user passes
    # overrides these, they do not augment these. SAN may add to these.
    my $nfs_options = {
        proto     => 'tcp',
        vers      => 3,
        mountopts => 'hard,intr',
    };
    # Copy over the NFS opts we recognize, discarding anything else.
    for my $opt (qw(proto vers mountopts)) {
        if (exists $opts{nfs_options}->{$opt}) {
            $nfs_options->{$opt} = $opts{nfs_options}->{$opt};
        }
    }

    # These are the default options for CIFS. Anything the user passes
    # overrides these, they do not augment these. SAN may (eventually) add
    # to these (when SAN on Windows is supported).
    my $cifs_options = {
        user       => 'labclient',
        password   => 'sundance',
        properties => q{},
    };
    # Copy over the CIFS opts we recognize, discarding anything else:
    for my $opt (qw(user password properties)) {
        if (exists $opts{cifs_options}->{$opt}) {
            $cifs_options->{$opt} = $opts{cifs_options}->{$opt};
        }
    }

    # This does nothing if the volume is already mounted. If it isn't, it
    # creates a junction-point based on the volume's name and mounts it:
    $self->_volume->mount;

    # For SAN, we have to set up some additional arguments and framework
    # before the actual call to mount or share is made.
    if ($self->_san) {
        my $lun     = $self->_lun;
        my $lunpath = $lun->path;
        my $vserver = $self->_volume->vserver;

        if ($self->is_windows) {
            $Log->exit() if $may_exit;
            NATE::BaseException->throw(
                'SAN I/O not yet supported for CIFS clients'
            );
        } else {
            # If we don't already have the LUN device then determine it so we
            # can mount as device instead of straight NFS. Do this by using
            # properties of the filer-side LUN object to locate the matching
            # client-side LUN and getting the device_filename from it.
            if (! $self->_lun_device) {
                my $client_lun_state = NACL::CS::Client::Sanlun->fetch(
                    command_interface => $self->_client,
                    filter            => {
                        controller   => $vserver,
                        lun_pathname => $lunpath,
                    }
                );
                $self->_lun_device($client_lun_state->device_filename);
            }
            delete @{$nfs_options}{qw(proto vers)};
            $nfs_options->{path}        = $self->_lun_device;
            $nfs_options->{type}        = 'device';
            $nfs_options->{mountopts}   = q{}; # Null for device mounts
            $nfs_options->{mount_point} =
                $self->_derive_lun_mountpoint(type => 'nfs');
        }
    }

    # Delegate to a private sub based on Windows or not
    if ($self->is_windows) {
        $self->_create_cifs_mount(
            options => $cifs_options,
            access  => lc $opts{access},
            is_ls   => $opts{is_ls}
        );
    } else {
        $self->_create_nfs_mount(
            options => $nfs_options,
            access  => lc $opts{access},
            is_ls   => $opts{is_ls}
        );
    }
    # If we didn't throw an exception by this point, then _mount is set. Also
    # set _mount_access, as well.
    $self->_mount_access(lc $opts{access});

    $Log->exit() if $may_exit;
    return $self;
}

=head2 delete_mount

    # Delete an existing mount
    $io_handle->delete_mount;

    # Delete with the "-f" (force) option passed:
    $io_handle->delete_mount(force => 1);

(Must be called as an instance method)

Delete a mount that was created by B<create_mount>. Removes the mount-point
on the client, but does not do an "unmount" of the volume itself at the filer
level.

If the operation is successful, the calling object is returned. Otherwise, an
exception is thrown.

=over

=item Options

The following option is the only recognized option to B<delete_mount>:

=over

=item C<< force =E<gt> 0|1 >>

(Optional) A boolean value that determines whether to "force" the unmount.
This maps to the C<force> option to B<NACL::C::Client::Mount::delete()>,
which is used for both NFS and CIFS mounts.

=back

=item Exceptions

If an invalid value is passed for C<force>, or if an unknown option is passed,
an exception of the type B<NATE::Exceptions::Argument> will be thrown. If
no mount-point exists currently (i.e., B<create_mount> has not been called) an
exception of the type B<NATE::BaseException> will be thrown.

Additionally, any exceptions that may be thrown by
B<NACL::STask::Client::Mount> may be thrown here, as they will be passed along
unchanged if they occur.

=back

=cut

###############################################################################
#
#   Sub Name:       delete_mount
#
#   Description:    Remove a mount-point that exists for the volume/client
#                   stored within this object. Remove even if we didn't create
#                   it ourselves.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#                   @opts     in      array     Additional options, converted
#                                                 to a hash by validate()
#
#   Returns:        Success:    $self
#                   Failure:    throws exception
#
###############################################################################
sub delete_mount {
    $Log->enter() if $may_enter;
    my ($self, @opts) = @_;
    # This must be an instance method invocation:
    $self->_verify_invocation(style => 'instance_only');
    my %opts = validate(
        @opts,
        {
            force        => $OPTIONAL_BOOLEAN,
            stop_traffic => $OPTIONAL_BOOLEAN,
        }
    );
    if (! $self->_mount_isset) {
        $Log->exit() if $may_exit;
        NATE::BaseException->throw(
            'Cannot delete_mount: no mount-point has been created yet'
        );
    }

    # Can't delete the mount-point if we're currently using it for traffic.
    if ($opts{stop_traffic}) {
        $self->traffic_stop;
    } else {
        if (@{$self->_traffic}) {
            $Log->exit() if $may_exit;
            NATE::BaseException->throw(
                'Cannot delete_mount: traffic is currently running'
            );
        }
    }

    $self->_mount->purge($opts{force} ? (force => 1) : ());
    $self->_mount_reset;
    $self->_mount_access_reset;

    $Log->exit() if $may_exit;
    return $self;
}

=head2 data_populate

    # Populate 2Gb of data:
    $io_handle->data_populate(
        size => '2g',
        path => 'data'
    );

    # Start a population but return immediately:
    my $data_handle = $io_handle->data_populate(
        size  => '2g',
        path  => 'data',
        async => 1,
    );
    # Do some stuff, like taking periodic snapshots...
    # Wait for background data_populate to finish (with
    # 15-min timeout):
    $io_handle->data_wait_for_completion(timeout => 900);

(Must be called as an instance method)

Populate data on the volume contained by this object, using the client also
contained by this object. The population may be done synchronously or
asynchronously, depending on the user's needs.

If called synchronously, the method returns when population has completed. If
called asynchronously, the method returns almost immediately. In both cases,
the return value is void. If an error occurs, an exception is thrown.

=over

=item Options

The following options are recognized by B<data_populate>:

=over

=item C<< method =E<gt> $string >>

(Optional) Specify the population method (plug-in) to use. See the modules
under B<NACL/AppBuilder/WFC/IO/Data> for the available plug-ins and any
options specific to them.

If not given, the default is C<NaclData>, which uses the B<NACL::MTask::Data>
module for population. See B<NACL::AppBuilder::WFC::IO::Data::NaclData>.

=item C<< size =E<gt> $string >>

(Optional) Specify the size of data to populate. The string should be of the
form, C<D+u>, where "D+" is one or more digits, and "u" is a single-letter unit
representing gigabytes, etc. At present, only C<g> (gigabytes) and C<m>
(megabytes) are supported for the units. The default is C<1g> (one gigabyte).

=item C<< path =E<gt> $string >>

(Optional) Specify the path from the root of the mounted volume where data
population should be rooted. The value is appended to the mount-point of the
volume. The default is no path (an empty string), which populates starting at
the root of the volume.

=item C<< async =E<gt> 0|1 >>

(Optional) Specify whether the population should be done asynchronously or
synchronously, that is, whether it should be backgrounded or not. If the
population is I<not> backgrounded (C<async> = 0), B<data_populate> will wait
until it is complete before returning. If it I<is> backgrounded (C<async> = 1),
B<data_populate> will return immediately. The default is C<0>, to do data
population synchronously.

=item C<< arguments =E<gt> $hashref >>

(Optional) Any additional arguments that should be passed to the underlying
population operation performed by the plugin. Must be a hash-reference of
key/value pairs. Each plugin is responsible for defining what (if any)
additional arguments are recognized. These additional arguments are generally
passed to the underlying calls without any processing or modification.

=back

=item Exceptions

The following exceptions may be thrown by B<data_populate>:

=over

=item B<NATE::Exceptions::Argument>

If an unknown argument is passed, or if a bad value for one of the known
arguments is passed, this exception will be thrown.

=item B<NATE::BaseException>

If B<data_populate> cannot start the population (for example, if you have not
yet called B<create_mount>) this exception will be thrown.

=back

In addition, any exceptions thrown by the chosen plug-in's B<data_populate>
method may be thrown here, as well.

=back

=cut

###############################################################################
#
#   Sub Name:       data_populate
#
#   Description:    Populate some amount of data on the volume encapsulated by
#                   this object, using the client encapsulated by this object.
#                   There must be an existing mount between the two or this
#                   will throw an exception. If this is done in an async manner
#                   then the method returns almost immediately, otherwise it
#                   waits for the operation to complete before returning.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#                   @opts     in      array     Additional options, converted
#                                                 to a hash by validate()
#
#   Returns:        Success:    void
#                   Failure:    throws exception
#
###############################################################################
sub data_populate {
    $Log->enter() if $may_enter;
    my ($self, @opts) = @_;
    # This must be an instance method invocation:
    $self->_verify_invocation(style => 'instance_only');

    # Population and modification are functionally the same except for the
    # name of the method called on the WFC::IO::Populate::* object. Any args
    # specific to either one will be in the "arguments" hashref that the user
    # passes.
    $self->_populate_or_modify(@opts, action => 'populate');

    $Log->exit;
    return;
}

=head2 data_modify

    # Modify 2Gb of data:
    $io_handle->data_modify(
        size => '2g',
        path => 'data'
    );

    # Start a modification but return immediately:
    $io_handle->data_modify(
        size  => '2g',
        path  => 'data',
        async => 1,
    );
    # Do some stuff, like taking periodic snapshots...
    # Wait for background data_modify to finish (with 15-min timeout):
    $io_handle->data_wait_for_completion(timeout => 900);

(Must be called as an instance method)

Modify data on the volume contained by this object, using the client also
contained by this object. The modification may be done synchronously or
asynchronously, depending on the user's needs.

If called synchronously, the method returns when modification has completed. If
called asynchronously, the method returns almost immediately. In both cases,
the return value is void. If an error occurs, an exception is thrown.

=over

=item Options

The following options are recognized by B<data_modify>:

=over

=item C<< method =E<gt> $string >>

(Optional) Specify the method (plug-in) to use. See the modules under
B<NACL/AppBuilder/WFC/IO/Populate> for the available plug-ins and any options
specific to them. (The same plug-in is used for B<data_populate> and
B<data_modify>.)

If not given, the default is C<NaclData>, which uses the B<NACL::MTask::Data>
module for modification. See B<NACL::AppBuilder::WFC::IO::Populate::NaclData>.

=item C<< size =E<gt> $string >>

(Optional) Specify the size of data to modify. The string should be of the
form, C<D+u>, where "D+" is one or more digits, and "u" is a single-letter unit
representing gigabytes, etc. At present, only C<g> (gigabytes) and C<m>
(megabytes) are supported for the units. The default is C<1g> (one gigabyte).

=item C<< path =E<gt> $string >>

(Optional) Specify the path from the root of the mounted volume where data
modification should be rooted. The value is appended to the mount-point of the
volume. The default is no path (an empty string), which modifies starting at
the root of the volume.

=item C<< async =E<gt> 0|1 >>

(Optional) Specify whether the modification should be done asynchronously or
synchronously, that is, whether it should be backgrounded or not. If the
modification is I<not> backgrounded (C<async> = 0), B<data_modify> will wait
until it is complete before returning. If it I<is> backgrounded (C<async> = 1),
B<data_modify> will return immediately. The default is C<0>, to do modification
synchronously.

=item C<< arguments =E<gt> $hashref >>

(Optional) Any additional arguments that should be passed to the underlying
modification operation performed by the plugin. Must be a hash-reference of
key/value pairs. Each plugin is responsible for defining what (if any)
additional arguments are recognized. These additional arguments are generally
passed to the underlying calls without any processing or modification.

=back

=item Exceptions

The following exceptions may be thrown by B<data_modify>:

=over

=item B<NATE::Exceptions::Argument>

If an unknown argument is passed, or if a bad value for one of the known
arguments is passed, this exception will be thrown.

=item B<NATE::BaseException>

If B<data_modify> cannot start the modification (for example, if you have not
yet called B<create_mount>) this exception will be thrown.

=back

In addition, any exceptions thrown by the chosen plug-in's B<data_modify>
method may be thrown here, as well.

=back

=cut

###############################################################################
#
#   Sub Name:       data_modify
#
#   Description:    Modify some amount of data on the volume encapsulated by
#                   this object, using the client encapsulated by this object.
#                   There must be an existing mount between the two or this
#                   will throw an exception. If this is done in an async manner
#                   then the method returns almost immediately, otherwise it
#                   waits for the operation to complete before returning.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#                   @opts     in      array     Additional options, converted
#                                                 to a hash by validate()
#
#   Returns:        Success:    void
#                   Failure:    throws exception
#
###############################################################################
sub data_modify {
    $Log->enter() if $may_enter;
    my ($self, @opts) = @_;
    # This must be an instance method invocation:
    $self->_verify_invocation(style => 'instance_only');

    # Population and modification are functionally the same except for the
    # name of the method called on the WFC::IO::Populate::* object. Any args
    # specific to either one will be in the "arguments" hashref that the user
    # passes.
    $self->_populate_or_modify(@opts, action => 'modify');

    $Log->exit;
    return;
}

=head2 data_is_running

    # Do something only if there is something running
    if ($io_handle->data_is_running) {
        ...
    }

(Must be called as an instance method)

Test whether a B<data_populate> or B<data_modify> operation is currently
underway. Note that (currently) this cannot tell if the remote job is
actually still running, but rather it tells whether the user has called either
of B<data_wait_for_completion> or B<data_kill> yet.

The return value is a boolean, C<true> if a job is running and C<false>
otherwise.

=over

=item Options

This method takes no options.

=item Exceptions

This method throws no exceptions.

=back

=cut

###############################################################################
#
#   Sub Name:       data_is_running
#
#   Description:    Boolean predicate telling whether we have a currently-
#                   running populate/modify job.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#
#   Returns:        boolean
#
###############################################################################
sub data_is_running {
    $Log->enter;
    my ($self) = @_;
    my $res = 0;

    # If we have a current populate/modify running, call down to its method.
    # Otherwise leave $res as 0.
    if ($self->_data_current_isset) {
        $res = $self->_data_current->data_is_running;
    }

    $Log->exit;
    return $res;
}

=head2 data_wait_for_completion

    # Wait for an async data operation to finish
    $io_handle->data_wait_for_completion;

(Must be called as an instance method)

Enters into a blocking wait for the current data operation (populate or modify,
that was initialized as asynchronous) to complete. If there is no running
operation, quietly does nothing.

If this method returns without throwing an exception, it clears out the slot
for the running data operation and as such B<data_is_running> (above) will then
return false.

=over

=item Options

All options passed to this method are passed along to the
B<data_wait_for_completion> method of the current data object.

The following is (currently) the only option recognized by the data classes:

=over

=item C<< timeout =E<gt> $number >>

Specify a timeout value for the wait. If the process takes longer than this
to finish, an exception will be thrown.

=back

=item Exceptions

This method may throw any exceptions that are thrown by the related methods of
the encapsulated data library (B<NACL::MTask::Data>, etc.).

=back

=cut

###############################################################################
#
#   Sub Name:       data_wait_for_completion
#
#   Description:    Wait for a current data operation to complete, if there
#                   is one.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#
#   Returns:        Success:    void
#                   Failure:    throws exception
#
###############################################################################
sub data_wait_for_completion {
    $Log->enter;
    my ($self, @opts) = @_;

    # If we have a current populate/modify running, call down to its method.
    if ($self->_data_current_isset) {
        $self->_data_current->data_wait_for_completion(@opts);
        $self->_data_current_reset;
    }

    $Log->exit;
    return;
}

=head2 data_kill

    # Kill any asynchronous operation currently running:
    $io_handle->data_kill;

(Must be called as an instance method)

Kills a running asynchronous data process without waiting for it to complete.
If there is not one running, or if the running process has already completed,
quietly does nothing.

If this method returns without throwing an exception, it clears out the slot
for the running data operation and as such B<data_is_running> (above) will then
return false.

=over

=item Options

This method takes no options.

=item Exceptions

This method may throw any exceptions that are thrown by the related methods of
the encapsulated data library (B<NACL::MTask::Data>, etc.).

=back

=cut

###############################################################################
#
#   Sub Name:       data_kill
#
#   Description:    Kill a running data operation, if there is one.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#
#   Returns:        Success:    void
#                   Failure:    throws exception
#
###############################################################################
sub data_kill {
    $Log->enter;
    my ($self) = @_;

    # If we have a current populate/modify running, call down to its method.
    if ($self->_data_current_isset) {
        $self->_data_current->data_kill;
        $self->_data_current_reset;
    }

    $Log->exit;
    return;
}

=head2 traffic_start

    # With options to control the traffic:
    my $id = $io_handle->traffic_start(
        traffic_file => $path,
        traffic_tool => 'sio',
        %additional_options
    );

    # Or with all defaults:
    my $id = $io_handle->traffic_start;

(Must be called as an instance method)

Start a traffic stream using the mount-point and client encapsulated by this
IO object. The return value is an integer ID for the stream that was created,
or an exception is thrown in case of error.

This is a wrapper around the B<NACL::MTask::Traffic> library. In theory it
supports all the tools that are supported by that library. However, at the
moment only C<sio> is supported, as there are not defaults defined for the
other tools. This should be augmented at a later date.

=over

=item Options

The following options are recognized:

=over

=item C<< traffic_file =E<gt> string >>

(Optional) The name of the file to use for traffic. If no name is passed, a
random name is generated.

=item C<< traffic_tool =E<gt> string >>

(Optional) The name of the traffic tool to use. Currently defaults to C<sio>,
which is currently the only supported tool.

=back

In addition, any options other than the above are passed directly to the
B<NACL::MTask::Traffic::start> method unchanged. There are some reasonable
defaults defined for each supported tool. Any options passed by the user
will override the default value of the same name.

=item Exceptions

The following exceptions may be thrown by this method:

=over

=item B<NATE::Exceptions::Argument>

If either of the options specified above has a non-scalar value, this
exception will be thrown. It will also be thrown if an unknown value
for C<traffic_tool> is passed in.

=item B<NATE::BaseException>

If the options to the traffic tool require read/write access, but the mount
is read-only, then this exception will be thrown.

=back

In addition, any exceptions thrown by B<NACL::MTask::Traffic::start> will be
allowed to pass through uncaught.

=back

=cut

###############################################################################
#
#   Sub Name:       traffic_start
#
#   Description:    Start a traffic stream, based on the options given. Return
#                   the ID of the new stream so that the caller can later
#                   stop the traffic stream.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#                   @opts     in      list      Additional arguments
#
#   Returns:        Success:    integer ID
#                   Failure:    throws exception
#
###############################################################################
sub traffic_start {
    $Log->enter() if $may_enter;
    my ($self, @opts) = @_;
    # This must be an instance method invocation:
    $self->_verify_invocation(style => 'instance_only');
    my %opts = validate_with(
        params      => \@opts,
        allow_extra => 1,
        spec        => {
            traffic_file => $OPTIONAL_SCALAR,
            traffic_tool => $OPTIONAL_SCALAR,
        }
    );
    my $traffic = $self->_traffic;
    my $id      = @{$traffic};
    my $tool    = delete $opts{traffic_tool} || 'sio';
    my $file    = delete $opts{traffic_file} ||
        sprintf('wfc_traffic_%d_%d', $$, time);

    # We must have a valid mount-point in order to run traffic:
    if (! $self->_mount_isset) {
        $Log->exit() if $may_exit;
        NATE::BaseException->throw(
            'Cannot call traffic_start() without a valid mount-point set'
        );
    }

    # Default options for the supported tools:
    my %default_options = (
        sio => {
            read           => 50,
            rand           => 100,
            block_size     => '4k',
            start_byte     => 0,
            file_size      => '40m',
            num_of_threads => 2,
        },
    );
    if (! $default_options{$tool}) {
        $Log->exit() if $may_exit;
        NATE::Exceptions::Argument->throw(
            "Traffic tool $tool not (yet) supported. Supported tools are: " .
            join(', ' => sort keys %default_options)
        );
    }

    # Make sure that the user doesn't try to sneakily override any of these
    # three options when we call NACL::MTask::Traffic::start() later:
    delete @opts{qw(command_interface paths tool)};
    my %traffic_opts = (%{$default_options{$tool}}, %opts);
    # Make sure the mount is OK with the type of traffic we're planning to
    # start:
    my $read_write_traffic =
        ($traffic_opts{read} eq 'random') || ($traffic_opts{read} < 100);
    if ($read_write_traffic && ($self->_mount_access ne 'rw')) {
        $Log->exit() if $may_exit;
        NATE::BaseException->throw(
            'Cannot start read/write traffic: mount is read-only'
        );
    }
    # Normalize the value of $traffic_opts{file_size}:
    $traffic_opts{file_size} = lc $traffic_opts{file_size};
    $traffic_opts{file_size} =~ s/b$//;

    if (my $size = $self->_validate_traffic_file(file => $file)) {
        # If the file exists and there's an explicit requested size, the
        # file needs to be at least that large. If there is no explicit
        # size, then use the file's size at the value to be passed to the
        # tool by overwriting $traffic_opts{file_size}.
        if ($opts{file_size}) {
            # If there was a particular size given in the original opts, then
            # convert it to a straight number and make sure that the current
            # file size is adequate. This is the tricky branch.
            if ($opts{file_size} =~ /^(\d+)([kmg])b?$/i) {
                my $req_size = $1 * $FILE_SIZES{lc $2};
                if ($req_size > $size) {
                    NATE::BaseException->throw(
                        "File '$file' exists on the volume already, but " .
                        'is too small for the traffic request. Cannot ' .
                        'use this file for traffic.'
                    );
                }
            } else {
                NATE::Exceptions::Argument->throw(
                    "Value for 'file_size' ($opts{file_size}) is not valid"
                );
            }
        } else {
            # If there is no explicit size, we just use the size that the
            # file is currently at. This is the easy branch.
            $traffic_opts{file_size} = $size;
        }
    } else {
        # This will use NACL::STask::VolumeFile for read-only, and will write
        # out a sparse file for read-write:
        $self->_create_traffic_file(
            file  => $file,
            size  => $traffic_opts{file_size},
            is_rw => ($self->_mount_access eq 'rw') ? 1 : 0
        );
    }

    # Make $file an absolute path:
    $file = NACL::MTask::Traffic->get_abs_path($self->_mount, $file);

    $traffic->[$id] = NACL::MTask::Traffic->start(
        command_interface => $self->_client,
        paths             => $file,
        tool              => $tool,
        %traffic_opts
    );

    $Log->exit() if $may_exit;
    return $id;
}

=head2 traffic_stop

    # Stop one traffic instance:
    $io_handle->traffic_stop(id => $id);

    # Stop all traffic:
    $io_handle->traffic_stop;

(Must be called as an instance method)

Stop a traffic process that was previously started by B<traffic_start>. If
called with no arguments, stops all traffic processes that are currently
active, that were created by this object.

=over

=item Options

This method recognizes only one parameter:

=over

=item C<< id =E<gt> integer >>

The ID returned by a previous call to B<traffic_start>. Optional, if no ID
is given then all active traffic streams are stopped.

=back

=item Exceptions

The following exceptions may be thrown by B<traffic_stop>:

=over

=item B<NATE::Exceptions::Argument>

If an unknown parameter is passed, or if the value of C<id> is not an
integer, this exception will be thrown.

=item B<NATE::BaseException>

If the C<id> given does not correspond to a traffic stream controlled by
this handle, this exception will be thrown.

=back

In addition, any exception that might be thrown by the B<stop> method of
B<NACL::Data::Traffic> may be thrown here, as well.

=back

=cut

###############################################################################
#
#   Sub Name:       traffic_stop
#
#   Description:    Stop one traffic stream, or all of them, depending on
#                   whether an argument is given.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#                   @opts     in      list      Additional arguments
#
#   Returns:        Success:    $self
#                   Failure:    throws exception
#
###############################################################################
sub traffic_stop {
    $Log->enter() if $may_enter;
    my ($self, @opts) = @_;
    # This must be an instance method invocation:
    $self->_verify_invocation(style => 'instance_only');
    my %opts = validate(
        @opts,
        {
            id => { type => SCALAR, regex => qr/^\d+$/, optional => 1 },
        }
    );

    my $traffic = $self->_traffic;
    if ($opts{id}) {
        if (my $traf_handle = $traffic->[$opts{id}]) {
            $traf_handle->stop(if_stopped => 'pass');
            undef $traffic->[$opts{id}];
        } else {
            $Log->exit() if $may_exit;
            NATE::BaseException->throw(
                'Invalid "id" argument passed, no such traffic running'
            );
        }
    } else {
        for my $traf_handle (@{$traffic}) {
            next if (! defined $traf_handle);
            $traf_handle->stop(if_stopped => 'pass');
            undef $traf_handle;
        }
    }

    # If there is now no traffic running at all, null the array:
    if (! any { defined($_) } @{$traffic}) {
        @{$traffic} = ();
    }

    $Log->exit() if $may_exit;
    return $self;
}

=head2 traffic_is_io_tool_running

    # Check a specific traffic handle:
    if ($io_handle->traffic_is_io_tool_running(id => $id)) {
        ...
    }

    # Check whatever handle is the first in the list:
    if ($io_handle->traffic_is_io_tool_running) {
        ...
    }

(Must be called as an instance method)

Check whether a traffic process created by a previous call to B<traffic_start>
is still running. Returns a boolean value (1/0) indicating whether the traffic
is running or not.

=over

=item Options

This method recognizes only one parameter:

=over

=item C<< id =E<gt> integer >>

The ID returned by a previous call to B<traffic_start>. Optional, if no ID
is given then the first traffic process found on the calling object is
queried. This is a short-hand for cases where the user is only ever creating
a single traffic stream at a time per I/O handle.

=back

=item Exceptions

The following exceptions may be thrown by B<traffic_stop>:

=over

=item B<NATE::Exceptions::Argument>

If an unknown parameter is passed, or if the value of C<id> is not an
integer, this exception will be thrown.

=item B<NATE::BaseException>

If the C<id> given does not correspond to a traffic stream controlled by
this handle, this exception will be thrown. This exception may also be thrown
if there are no current traffic streams on the I/O handle at all.

=back

In addition, any exception that might be thrown by the B<is_io_tool_running>
method of B<NACL::Data::Traffic> may be thrown here, as well.

=back

=cut

###############################################################################
#
#   Sub Name:       traffic_is_io_tool_running
#
#   Description:    Boolean predicate that indicates whether the specified
#                   I/O instance is still running.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#                   @opts     in      list      Additional arguments
#
#   Returns:        Success:    0/1
#                   Failure:    throws exception
#
###############################################################################
sub traffic_is_io_tool_running {
    $Log->enter;
    my ($self, @opts) = @_;
    # This must be an instance method invocation:
    $self->_verify_invocation(style => 'instance_only');
    my %opts = validate(
        @opts,
        {
            id => { type => SCALAR, regex => qr/^\d+$/, optional => 1 },
        }
    );
    my ($traffic, $is_running);

    my $all_traffic = $self->_traffic;
    if ($opts{id}) {
        if (! ($traffic = $all_traffic->[$opts{id}])) {
            $Log->exit() if $may_exit;
            NATE::BaseException->throw(
                'Invalid "id" argument passed, no such traffic running'
            );
        }
    } else {
        if (! ($traffic = first { defined($_) } @{$all_traffic})) {
            $Log->exit() if $may_exit;
            NATE::BaseException->throw(
                'No traffic running, nothing to check running status on.'
            );
        }
    }

    $is_running = $traffic->is_io_tool_running;

    $Log->exit;
    return $is_running;
}

=head2 initialize_san

    # Used as an instance method:
    my $io_handle = NACL::AppBuilder::WFC::IO->new(
        volume => $volume,
        client => $client
    );
    $io_handle->initialize_san(
        lun_name    => $name_of_lun,
        lun_size    => $size_of_lun,
        lun_fstype  => $fstype,
        igroup_name => $name_of_igroup
    );

    # Used as a static method:
    NACL::AppBuilder::WFC::IO->initialize_san(
        volume      => $volume,
        client      => $client,
        lun_name    => $name_of_lun,
        lun_size    => $size_of_lun,
        lun_fstype  => $fstype,
        igroup_name => $name_of_igroup
    );

    # Minimal usage, using all defaults:
    my $io_handle = NACL::AppBuilder::WFC::IO->initialize_san(
        volume => $volume,
        client => $client
    );

(May be called as a static method or an instance method.)

This method initializes SAN on the volume it encapsulates. It performs the
following steps:

=over

=item 1.

Creates a LUN on the volume through the command-interface of the volume.

=item 2.

Creates an initiator group on the vserver the volume belongs to, tied to the
encapsulated client.

=item 3.

Maps the LUN to the initiator group.

=item 4.

Performs the client-side discovery commands to cause the LUN to show up as a
mountable device.

=item 5.

Mounts the LUN on the client and creates a filesystem on the mounted device.

=back

Once this method completes successfully, the B<san> attribute on the object
is set to a true value, and the B<lun> and B<igroup> attributes are set to
return appropriate objects.

SAN initialization can be done with a minimum of data, or the user may provide
most of the elements directly.

The return value of this method is the calling object if the method was called
as an instance method. If it was called as a static method, the return value is
a new B<NACL::AppBuilder::WFC::IO> object that encapsulates the given volume
and client, as well as having SAN already set up. In both cases, the return
value is an object of this class that may be used to call other methods.

=over

=item Options

The following parameters are recognized by this method:

=over

=item C<< volume =E<gt> $volume_obj >>

The volume to use for creating the LUN. This is a required parameter if the
method is called statically. If it is called as an instance method, then the
contained volume is used by default. In that case, providing this parameter
overrides the internal volume (but does not replace it).

=item C<< client =E<gt> $client_obj >>

The client to use for creating the initiator group and mapping the LUN. This is
a required parameter if the method is called statically. If it is called as an
instance method, then the contained client is used by default. In that case,
providing this parameter overrides the internal client (but does not replace
it).

=item C<< lun_name =E<gt> $string >>

The name to use when creating the LUN. If not given, then the volume's name
will be used with the string C<_lun> appended to it.

=item C<< lun_size =E<gt> $string >>

The size to create the LUN as. The value may be either a typical size
expression (C<20gb>, C<100m>, etc.) or a percentage which is applied to the
volume size (C<50%>, C<90%>, etc.). If the size is larger than the volume, an
exception is thrown. If the size is too close to the size of the volume itself,
you may encounter errors later in the lifecycle. If not given, the default
value is C<50%>, or 50% of the size of the volume itself.

=item C<< lun_fstype =E<gt> $string >>

The type of filesystem to create on the LUN. If not given, the default value
is C<ext2> for Linux.

=item C<< igroup_name =E<gt> $string >>

The name to use when creating the initiator group to map the LUN to the client.
If not given, then the volume's name will be used with the string C<_igroup>
appended to it.

=item C<< lun =E<gt> $lun_obj >>

Optional: The user may pass an existing B<NACL::STask::Lun> object in to be
used by the SAN initialization. If this value is passed in, the B<lun_name>
and B<lun_size> options are ignored. The LUN is checked to make sure that it
belongs to the volume that SAN initialization is using, and an exception is
thrown if it does not belong to the volume. Note that mapping and creation of
the filesystem is still done, so if there is pre-existing data on the LUN it
will be lost.

=item C<< igroup =E<gt> $igroup_obj >>

Optional: The user may pass an existing B<NACL::STask::LunIgroup> object in to
be used by the SAN initialization. If this value is passed in, the
B<igroup_name> option is ignored. The object is not checked to ensure that it
is linked to the client, so the user must be sure this is the case or mounting
the device on the client will fail.

=back

=item Exceptions

The following exceptions may be thrown by B<initialize_san>:

=over

=item B<NATE::Exceptions::Argument>

If an unknown parameter is passed, or if a known parameter is passed with an
invalid value, this exception will be thrown.

=item B<NATE::BaseException>

If the initialization of SAN-related elements fails, this exception will be
thrown.

=back

In addition, any exceptions thrown by LUN creation or initiator group creation
may be thrown here, as well.

=back

=cut

###############################################################################
#
#   Sub Name:       initialize_san
#
#   Description:    Initialize SAN on this IO object by creating a LUN (or
#                   taking in an existing one), creating an Igroup (or taking
#                   in an existing one), mapping the LUN, discovering it on
#                   the client and creating a filesystem on it.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref/sc    Object instance or the class
#                                                 name
#                   @opts     in      list      Additional arguments
#
#   Returns:        Success:    $self
#                   Failure:    throws exception
#
###############################################################################
sub initialize_san {
    $Log->enter() if $may_enter;
    my ($self, @opts) = @_;
    my %opts = validate(
        @opts,
        {
            volume      => {
                type     => OBJECT,
                isa      => 'NACL::STask::Volume',
                optional => 1,
            },
            client       => {
                type     => OBJECT,
                isa      => 'NACL::C::Client',
                optional => 1,
            },
            lun         => {
                type     => OBJECT,
                isa      => 'NACL::STask::Lun',
                optional => 1,
            },
            igroup      => {
                type     => OBJECT,
                isa      => 'NACL::STask::LunIgroup',
                optional => 1,
            },
            lun_name    => $OPTIONAL_SCALAR,
            lun_size    => $OPTIONAL_SCALAR,
            lun_fstype  => $OPTIONAL_SCALAR,
            igroup_name => $OPTIONAL_SCALAR,
        }
    );
    my ($volume, $client, $lun, $lun_name, $lun_size, $igroup, $igroup_name,
        $fstype);

    if (blessed $self) {
        $volume = $opts{volume} || $self->_volume;
        $client = $opts{client} || $self->_client;
    } else {
        $volume = $opts{volume};
        $client = $opts{client};
        if (! ($volume && $client)) {
            $Log->exit() if $may_exit;
            NATE::Exceptions::Argument->throw(
                'This method must either be called as an instance method, or ' .
                'have both "volume" and "client" specified'
            );
        }
        # Create a temporary object for the sake of the private methods we'll
        # be calling:
        $self = $self->new(
            volume => $volume,
            client => $client,
            lun    => undef
        );
    }
    # If they are doing SAN on localhost, issue a warning
    if ($client->client eq 'localhost') {
        $Log->warn(
            'You are doing SAN set-up on "localhost". This may be ' .
            'a problem if you later try to connect to this LUN from ' .
            'a different (more specific) host.'
        );
    }

    # If they passed in a LUN, try to use it (though it has to belong to the
    # volume we're operating on). If they didn't pass in a LUN, get/set values
    # for $lun_name and $lun_size.
    if ($lun = $opts{lun}) {
        # First make sure that the LUN actually belongs to $volume.
        if ($lun->get_one_state_attribute('volume') ne $volume->volume) {
            $Log->exit() if $may_exit;
            NATE::Exceptions::Argument->throw(
                'The given LUN (' . $lun->path . ') does not belong to the ' .
                'given volume (' . $volume->volume . ')'
            );
        }
    } else {
        $lun_size = $opts{lun_size} || '50%';
        $lun_name = $opts{lun_name} || $volume->volume . '_lun';
    }
    # If they passed in an Igroup, we'll use it. If they didn't pass one in,
    # get/set $igroup_name for the creation of one.
    if (! ($igroup = $opts{igroup})) {
        $igroup_name = $opts{igroup_name} || $volume->volume . '_igroup';
    }

    # Create (if necessary) and map the LUN. Sets _lun and _igroup attrs.
    $self->_create_and_map_lun(
        lun         => $lun,
        igroup      => $igroup,
        lun_name    => $lun_name,
        lun_size    => $lun_size,
        igroup_name => $igroup_name
    );
    # Client-side discovery of the LUN. Sets the _lun_device attr.
    $self->_client_iscsi_discovery;
    # Create the filesystem on the LUN.
    $self->_create_lun_filesystem(lun_fstype => $opts{lun_fstype});

    # Set this in case it isn't already:
    $self->_san(1);

    $Log->exit() if $may_exit;
    return $self;
}

=head2 move_lun_mapping

    # Move a LUN mapping to an existing IO handle object:
    $io_handle->move_lun_mapping(
        lun    => $lun_obj,
        igroup => $igroup_obj
    );

    # Move a LUN mapping without having an existing handle:
    NACL::AppBuilder::WFC::IO->move_lun_mapping(
        volume => $volume_obj,
        client => $client_obj,
        lun    => $lun_obj,
        igroup => $igroup_obj
    );

(May be called as either an instance method or a static method.)

This method unmaps an existing LUN, finds a LUN on the given (or encapsulated)
volume by the same name, and maps the new LUN to the client using the given
initiator group. It then does the LUN discovery on the client so that the
client connects to the new volume for that particular LUN.

This method is primarily for cases where a LUN begins life on one volume, then
migrates to another volume (via snapmirror, volume cloning, etc.). Hence the
expectation that the new LUN is the same name as the current one.

The return value is the calling object in the case of an instance method call,
or a new B<NACL::AppBuilder::WFC::IO> object if called statically. The new
object will encapsulate the volume and client, as well has having SAN set up
with the LUN already configured.

=over

=item Options

The following parameters are recognized by this method:

=over

=item C<< volume =E<gt> $volume_obj >>

The volume to use for locating the new LUN. This is a required parameter if the
method is called statically. If it is called as an instance method, then the
contained volume is used by default. In that case, providing this parameter
overrides the internal volume (but does not replace it).

=item C<< client =E<gt> $client_obj >>

The client to map the LUN to. This is a required parameter if the method is
called statically. If it is called as an instance method, then the contained
client is used by default. In that case, providing this parameter overrides the
internal client (but does not replace it).

=item C<< lun =E<gt> $lun_obj >>

Required: The LUN as it I<currently> exists, on the old volume. This object is
used to both unmap the current LUN and to identify the new LUN on the "new"
volume.

=item C<< igroup =E<gt> $igroup_obj >>

Required: The initiator group being used to map the LUN to the given client.
It is assumed that the same initiator is being used on the new LUN as was
used to map the old LUN.

=back

=item Exceptions

The following exceptions may be thrown by B<move_lun_mapping>:

=over

=item B<NATE::Exceptions::Argument>

If an unknown parameter is passed, or if a known parameter is passed with an
invalid value, this exception will be thrown.

=back

Additionally, any exceptions possible from the mapping of a LUN or the client
discovery of the LUN may be thrown here.

=back

=cut

###############################################################################
#
#   Sub Name:       move_lun_mapping
#
#   Description:    Unmap the LUN as it currently is mapped, and map the LUN
#                   by the same path that exists on "volume".
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#                   @opts     in      List      Options for the operation
#
#   Returns:        Success:    $self
#                   Failure:    throws exception
#
###############################################################################
sub move_lun_mapping {
    $Log->enter() if $may_enter;
    my ($self, @opts) = @_;
    my %opts = validate(
        @opts,
        {
            volume      => {
                type     => OBJECT,
                isa      => 'NACL::STask::Volume',
                optional => 1,
            },
            client       => {
                type     => OBJECT,
                isa      => 'NACL::C::Client',
                optional => 1,
            },
            lun         => {
                type     => OBJECT,
                isa      => 'NACL::STask::Lun',
            },
            igroup      => {
                type     => OBJECT,
                isa      => 'NACL::STask::LunIgroup',
            },
        }
    );

    my ($volume, $client);
    if (blessed $self) {
        $volume = $opts{volume} || $self->_volume;
        $client = $opts{client} || $self->_client;
    } else {
        $volume = $opts{volume};
        $client = $opts{client};
        if (! ($volume && $client)) {
            $Log->exit() if $may_exit;
            NATE::Exceptions::Argument->throw(
                'This method must either be called as an instance method, or ' .
                'have both "volume" and "client" specified'
            );
        }
        # Create a temporary object for the sake of the private methods we'll
        # be calling:
        $self = $self->new(
            volume => $volume,
            client => $client,
            lun    => undef
        );
    }
    # Options from the caller, the existing LUN object and the igroup.
    my $lun    = $opts{lun};
    my $igroup = $opts{igroup};

    # Start by finding the LUN on $volume with the same path as $lun:
    my $new_lun = NACL::STask::Lun->find(
        command_interface => $volume->command_interface,
        filter            => {
            volume  => $volume,
            vserver => $volume->vserver,
            lun     => $lun->get_one_state_attribute('lun'),
        }
    );

    # $lun is already mapped, so just unmap it:
    $lun->unmap(igroup => $igroup);

    # Now map the new LUN, found from $volume.
    $new_lun->map(
        igroup                    => $igroup,
        nacltask_if_mapped_igroup => 'remap',
        nacltask_verify           => 1
    );

    # $self->_lun needs to point to the "new" LUN, now:
    $self->_lun($new_lun);
    # In case $self is a new object, we also need to save $igroup:
    $self->_igroup($igroup);

    # Client needs to re-discover the LUN on the new volume:
    $self->_client_iscsi_discovery;

    # Set this in case it isn't already:
    $self->_san(1);

    $Log->exit() if $may_exit;
    return $self;
}

###############################################################################
#
# Private methods
#
###############################################################################

###############################################################################
#
#   Sub Name:       _create_and_map_lun
#
#   Description:    Create a LUN for the volume on the filer. This just
#                   does the filer-side steps, it does not do the steps to
#                   discover the LUN on the client or create the initial
#                   filesystem. It does create the Igroup (if needed) and
#                   map the LUN.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#                   %opts     in      hash      Options for the operation
#                                                 ('volume', etc.)
#
#   Returns:        Success:    $self
#                   Failure:    throws exception
#
###############################################################################
sub _create_and_map_lun {
    $Log->enter() if $may_enter;
    my ($self, %opts) = @_;

    my $volume      = $self->_volume;
    my $client      = $self->_client;
    # Options from the caller, for SAN options.
    my $lun         = $opts{lun};
    my $lun_name    = $opts{lun_name};
    my $lun_size    = $opts{lun_size};
    my $igroup_name = $opts{igroup_name};
    my $igroup      = $opts{igroup};

    # Calculate the LUN size if it is in %, otherwise just validate it. Only
    # do this if $lun is not set, since they may not have specified a size if
    # they already have a LUN.
    if (! $lun) {
        if ($lun_size =~ /^(\d+)(%|[mgt]b?)?$/i) {
            my $num = $1;
            my $unit = $2 ? lc $2 : q{};
            my $vol_size = lc $volume->get_one_state_attribute('size');
            my ($vol_num, $vol_unit);

            if ($vol_size =~ /^(\d+(?:[.]\d+)?)(.*)$/) {
                $vol_num  = $1;
                $vol_unit = $2;
                if ($vol_unit) {
                    $vol_num *= $FILE_SIZES{substr($vol_unit, 0, 1)};
                }
                $vol_num = int $vol_num;
            } else {
                $Log->exit;
                NATE::BaseException->throw(
                    "Volume size ($vol_size) failed to parse"
                );
            }

            if ($unit eq q{%}) {
                if (($num > 100) || ($num < 1)) {
                    # Over 100%? Under 1%? Seriously?
                    $Log->exit;
                    NATE::Exceptions::Argument->throw(
                        "SAN option 'lun_size' ($lun_size) cannot be over " .
                        '100% or under 1%'
                    );
                }
                $lun_size = int(($num / 100) * $vol_num);
            } else {
                # Check that it isn't larger than the volume itself. I *think*
                # "lun create" would throw the error, but I'd rather avoid the
                # mess entirely since I can test it here.
                my $lsize = $unit ?
                    $num * $FILE_SIZES{substr($unit, 0, 1)} : $num;

                if ($lsize > $vol_num) {
                    $Log->exit;
                    NATE::Exceptions::Argument->throw(
                        "SAN option 'lun_size' ($lun_size) cannot exceed " .
                        "volume size ($vol_size)"
                    );
                }
            }
        } else {
            $Log->exit;
            NATE::Exceptions::Argument->throw(
                "Unparsable value for SAN option 'lun_size': $lun_size"
            );
        }
    }

    # Determine OS type of the client. We need this to be a type that
    # "lun create" understands, so for now we're limiting to linux and
    # windows.
    my $ostype = $client->type;
    if ($ostype =~ /linux/) {
        $ostype = 'linux';
    } elsif ($ostype =~ /windows/) {
        $ostype = 'windows';
    } else {
        $Log->exit;
        NATE::Exceptions::Argument->throw(
            "Client type '$ostype' is not one we can work with"
        );
    }

    # Now create the Igroup (if needed).
    if (! $igroup) {
        # Get the iSCSI initiator name value from the client, needed for igroup
        # creation:
        my $initiator = $client->get_initiator_name;

        $igroup = NACL::STask::LunIgroup->create(
            command_interface  => $volume->command_interface,
            vserver            => $volume->vserver,
            igroup             => $igroup_name,
            protocol           => 'iscsi',
            ostype             => $ostype,
            initiator          => [ $initiator ],
            nacltask_if_exists => 'purge',
            nacltask_verify    => 1
        );
    }

    # If $lun is not already given, create it:
    if (! $lun) {
        $lun = NACL::STask::Lun->create(
            command_interface => $volume->command_interface,
            lun               => $lun_name,
            volume            => $volume,
            vserver           => $volume->vserver,
            size              => $lun_size,
            ostype            => $ostype,
            if_exists         => 'purge',
            nacltask_verify   => 1
        );
    }

    # Now map the LUN.
    $lun->map(
        igroup                    => $igroup->igroup,
        nacltask_if_mapped_igroup => 'remap',
        nacltask_verify           => 1
    );

    # Save the LUN and Igroup objects for use by later steps in the SAN
    # lifecycle.
    $self->_lun($lun);
    $self->_igroup($igroup);

    $Log->exit() if $may_exit;
    return $self;
}

###############################################################################
#
#   Sub Name:       _client_iscsi_discovery
#
#   Description:    Send the client-side commands needed to do the
#                   initialization/discovery for iSCSI.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#
#   Returns:        Success:    Device node on client this LUN is mapped to
#                   Failure:    throws exception
#
###############################################################################
sub _client_iscsi_discovery {
    $Log->enter() if $may_enter;
    my ($self) = @_;

    # Values from the object, for volume, LUN, client and SAN options
    my $volume = $self->_volume;
    my $lun    = $self->_lun;
    my $client = $self->_client;

    # To keep from calling these methods multiple times:
    my $ci      = $volume->command_interface;
    my $vserver = $volume->vserver;

    # Get the IP address of the first iscsi-protocol-capable LIF on the
    # node. We're using a canned query here, to get the iscsi data interfaces
    # that are up.
    my $iface_addr;
    try {
        my $iface_state = NACL::CS::NetworkInterface->fetch(
            command_interface => $ci,
            canned_filters    => [ qw(up_data_iscsi_interfaces) ],
            filter            => { vserver => $vserver }
        );
        $iface_addr = $iface_state->address;
    } catch NACL::Exceptions::NoElementsFound with {
        my $exception = shift;
        $exception->set_text(
            'No data LIFs found for the iscsi protocol, cannot do SAN ' .
            'discovery for the client. Original error text: ' .
            $exception->text
        );
        $Log->exit() if $may_exit;
        $exception->throw;
    };
    my $filer_iqn;
    try {
        my $vs_iscsi_state = NACL::CS::VserverIscsi->fetch(
            command_interface => $ci,
            filter            => { vserver => $vserver }
        );
        $filer_iqn = $vs_iscsi_state->target_name;
    } catch NACL::Exceptions::NoElementsFound with {
        my $exception = shift;
        $exception->set_text(
            'No vserver iSCSI service found configured for vserver ' .
            "$vserver. Original error text: " .
            $exception->text
        );
        $Log->exit() if $may_exit;
        $exception->throw;
    };

    # Do "discovery" of the valid targets. This entails sending an iscsiadm
    # discovery command, then logging out and back in to the target.

    # Get the APISet object for $client. The commands we have to run are
    # only available at the APISet level.
    my $apiset = $client->apiset;
    # When scaled out to 20+ streams, these commands were sometimes hitting
    # the default time-out. So use connectrec-timeout to bump it to 5 min.
    $apiset->iscsiadm(
        mode                 => 1,
        discovery            => 1,
        type                 => 'sendtargets',
        'ip-port'            => $iface_addr,
        'privilege-level'    => 'root',
        'connectrec-timeout' => 300
    );
    $apiset->iscsiadm(
        mode                 => 1,
        node                 => 1,
        logout               => 1,
        'target-name'        => $filer_iqn,
        'ip-port'            => $iface_addr,
        'privilege-level'    => 'root',
        'connectrec-timeout' => 300
    );
    $apiset->iscsiadm(
        mode                 => 1,
        node                 => 1,
        login                => 1,
        'target-name'        => $filer_iqn,
        'ip-port'            => $iface_addr,
        'privilege-level'    => 'root',
        'connectrec-timeout' => 300
    );

    # It appears that there can be a delay between the login and the LUNs
    # actually showing up to the sanlun command. Give this about 3 tries
    # before we give up.
    my $lun_state;
    my $lunpath = $lun->path;
    for my $iter (1 .. 3) {
        $lun_state = NACL::CS::Client::Sanlun->fetch(
            command_interface => $client,
            allow_empty       => 1,
            filter            => {
                controller   => $vserver,
                lun_pathname => $lunpath,
            }
        );
        last if $lun_state;
        $Log->comment(
            "Failed to find the LUN after iscsiadm, iteration $iter."
        );
        Tharn::snooze(10);
    }
    if (! $lun_state) {
        NATE::BaseException->throw(
            'Commands for iSCSI discovery completed successfully, but the ' .
            'LUN did not appear in the sanlun table.'
        );
    }

    # Save the LUN device filename for later use in the SAN lifecycle
    $self->_lun_device($lun_state->device_filename);

    $Log->exit() if $may_exit;
    return $self;
}

###############################################################################
#
#   Sub Name:       _create_lun_filesystem
#
#   Description:    Create a filesystem on the LUN through the client
#                   command-interface. Currently only supports what can be
#                   done via NACL::C::Client::Mkfs.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#                   %opts     in      hash      Options for the operation
#
#   Returns:        Success:    void
#                   Failure:    throws exception
#
###############################################################################
sub _create_lun_filesystem {
    $Log->enter() if $may_enter;
    my ($self, %opts) = @_;

    # Options from the caller, for SAN options
    my $fstype = $opts{lun_fstype} || 'ext2';

    NACL::C::Client::Mkfs->mkfs(
        command_interface => $self->_client,
        device            => $self->_lun_device,
        fstype            => $fstype
    );

    $Log->exit() if $may_exit;
    return;
}

###############################################################################
#
#   Sub Name:       _find_lun
#
#   Description:    For our encapsulated Volume object, find the LUN that
#                   should be used for SAN-based I/O and save it to the _lun
#                   attribute.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#                   %opts     in      hash      Additional args. These have
#                                                 been pre-validated by the
#                                                 caller.
#
#   Returns:        Success:    The found LUN
#                   Failure:    throws exception
#
###############################################################################
sub _find_lun {
    $Log->enter() if $may_enter;
    my ($self, %opts) = @_;

    # Just find the first LUN on the volume, unless they specified the LUN
    # name. In that case search for it specifically.
    my $lun = NACL::STask::Lun->find(
        command_interface => $self->_volume->command_interface,
        allow_empty       => 1,
        filter            => {
            volume  => $self->_volume,
            vserver => $self->_volume->vserver,
            ($opts{lun_name}) ? (lun => $opts{lun_name}) : (),
        }
    );

    $Log->exit() if $may_exit;
    return $lun;
}

###############################################################################
#
#   Sub Name:       _derive_lun_mountpoint
#
#   Description:    From the given LUN, derive the local-filesystem mountpoint
#                   based on the type arg.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#                   %opts     in      hash      Additional args
#
#   Returns:        Success:    string
#                   Failure:    throws exception
#
###############################################################################
sub _derive_lun_mountpoint {
    $Log->enter() if $may_enter;
    my ($self, %opts) = @_;

    my $lun     = $self->_lun;
    my $lunname = $lun->lun;

    my $mountpoint;

    if ($opts{type} eq 'nfs') {
        # Get the data LIFs for the command-interface of this LUN object. We'll
        # use the first one to get the address to use in the mount-point.
        my $data_ip;
        try {
            my $iface_state = NACL::CS::NetworkInterface->fetch(
                command_interface => $lun->command_interface,
                canned_filters    => [ qw(up_data_interfaces) ],
                filter            => { vserver => $lun->vserver },
                requested_fields  => [ 'address' ]
            );
            $data_ip = $iface_state->address;
        } catch NACL::Exceptions::NoElementsFound with {
            my $exception = shift;
            $exception->set_text(
                'No data LIFs found for the LUN command-interface. Original ' .
                'error text: ' . $exception->text
            );
            $Log->exit() if $may_exit;
            $exception->throw;
        };

        $mountpoint = "/t/$data_ip/$lunname";
    } else {
        $Log->exit() if $may_exit;
        NATE::BaseException->throw(
            'Can only derive LUN mountpoint for Linux/UNIX and similar ' .
            'at this time.'
        );
    }

    $Log->exit() if $may_exit;
    return $mountpoint;
}

###############################################################################
#
#   Sub Name:       _create_nfs_mount
#
#   Description:    Create a mount-point for the client/volume contained in
#                   $self, using NFS as the protocol.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#                   %opts     in      hash      Additional args. These have
#                                                 been pre-validated by
#                                                 create_mount()
#
#   Returns:        Success:    $self
#                   Failure:    throws exception
#
###############################################################################
sub _create_nfs_mount {
    $Log->enter() if $may_enter;
    my ($self, %opts) = @_;

    # Extra opts to pass to the NACL::STask::Client::Mount->create() call. Used
    # to build the mountopts option, and to conditionally include the chmod
    # option.
    my @opts_for_mount;

    # No reason to call this accessor so many times:
    my $volume = $self->_volume;

    # We need to provide a path that includes .admin if we are mounting
    # the volume as RW from a LS mirror. But don't do this for SAN!
    if ($opts{is_ls} and ($opts{access} eq 'rw')) {
        if ($self->_san) {
            $Log->warn(
                'SAN and LS mirrors are not compatible. Not using the ' .
                'read-write "/.admin" NAS convention for the access of the ' .
                'mounted SAN LUN.'
            );
        } else {
            my $path = $volume->get_one_state_attribute('junction-path');
            push(@opts_for_mount, path => "/.admin$path");
        }
    }

    # For mountopts, we also have to factor in $opts{access} (which means
    # removing the opposite value if it is present). So handle this one
    # separately from the rest of $opts{options}.
    my @mountopts = split(/,/ => delete $opts{options}->{mountopts} || q{});
    if ($opts{access} eq 'rw') {
        @mountopts = grep { lc $_ ne 'ro' } @mountopts;
        push(@mountopts, 'rw');
        push(@opts_for_mount, chmod => '777');
    } else {
        # Don't need to explicitly test for 'ro', the param validation in
        # create_mount() assures that it is one of 'rw' or 'ro'.
        @mountopts = grep { lc $_ ne 'rw' } @mountopts;
        push(@mountopts, 'ro');
    }
    push(@opts_for_mount, mountopts => join(q{,} => @mountopts));
    # Now add the rest from $opts{options}:
    for my $key (keys %{$opts{options}}) {
        if ($opts{options}->{$key}) {
            push(@opts_for_mount, $key => $opts{options}->{$key});
        }
    }

    # OK, now we're ready to actually create the mount-point and
    # NACL::STask::Client::Mount object:
    my $mount = NACL::STask::Client::Mount->create(
        command_interface        => $self->_client,
        server_command_interface => $volume->command_interface,
        vserver                  => $volume->vserver,
        volume                   => $volume,
        exclusive                => 1,
        @opts_for_mount
    );
    # If that didn't throw an exception, we have a mount object in $mount:
    $self->_mount($mount);

    $Log->exit() if $may_exit;
    return $self;
}

###############################################################################
#
#   Sub Name:       _create_cifs_mount
#
#   Description:    Create a mount-point for the client/volume contained in
#                   $self, using CIFS as the protocol.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#                   %opts     in      hash      Additional args. These have
#                                                 been pre-validated by
#                                                 create_mount()
#
#   Returns:        Success:    $self
#                   Failure:    throws exception
#
###############################################################################
sub _create_cifs_mount {
    $Log->enter() if $may_enter;
    my ($self, %opts) = @_;

    # This will be either an existing share we find, or one we create:
    my $share;
    # The name of the share, for searching and creation:
    my $share_name;
    # These values are used in share creation and share mounting
    my ($properties, $user, $password) =
        @{$opts{options}}{qw(properties user password)};
    # This is just used so many times, it makes sense to only call the
    # accessor once:
    my $volume = $self->_volume;
    # Save this value to avoid calling this method twice:
    my $path = $volume->get_one_state_attribute('junction-path');

    # If this is for an LS mirror, the share will be slightly different based
    # on whether $opts{access} is 'ro' or 'rw'. Otherwise the share is always
    # the same.
    $share_name = $volume->volume;
    if ($opts{is_ls}) {
        $share_name .= "_ls_$opts{access}";
        if ($opts{access} eq 'rw') {
            $path = "/.admin$path";
        }
    }

    # Find or create the share for this volume:
    $share = NACL::STask::VserverCifsShare->create(
        command_interface  => $volume->command_interface,
        vserver            => $volume->vserver,
        'share-name'       => $share_name,
        path               => $path,
        nacltask_if_exists => 'reuse',
        ($properties) ? ('share-properties' => $properties) : ()
    );

    # OK, now we're ready to actually create the mount-point and
    # NACL::STask::Client::Mount object:
    my $mount = NACL::STask::Client::Mount->create(
        command_interface        => $self->_client,
        server_command_interface => $volume->command_interface,
        vserver                  => $volume->vserver,
        share                    => $share,
        user                     => $user,
        password                 => $password,
        nacltask_if_exists       => 'purge'
    );
    # If that didn't throw an exception, we have a mount object in $mount:
    $self->_mount($mount);

    $Log->exit() if $may_exit;
    return $self;
}

###############################################################################
#
#   Sub Name:       _setup_existing_lun
#
#   Description:    Given a volume with a pre-existing LUN, see if it is
#                   mapped and visible to the client. If not, map it (if
#                   necessary, creating an igroup if needed) and do the
#                   iscsi discovery on the client.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#
#   Returns:        Success:    void
#                   Failure:    throws exception
#
###############################################################################
sub _setup_existing_lun {
    $Log->enter;
    my ($self) = @_;

    # Things we'll need, that I don't want to be repeatedly calling accessors
    # for:
    my $lun               = $self->_lun;
    my $client            = $self->_client;
    my $volume            = $self->_volume;
    my $vserver           = $volume->vserver;
    my $command_interface = $volume->command_interface;

    my ($mapped_lun, $igroup, $initiator, $do_discovery);
    $initiator = $client->get_initiator_name;

    # 1. Check to see if the LUN is mapped on the filer.
    $mapped_lun = NACL::C::LunMapped->find(
        command_interface => $command_interface,
        allow_empty       => 1,
        filter            => {
            volume  => $volume->volume,
            lun     => $lun->lun,
            vserver => $vserver
        }
    );

    # 2. If it is not mapped, create an igroup and map it using the igroup.
    if (! $mapped_lun) {
        # Determine OS type of the client. We need this to be a type that "lun
        # igroup create" understands, so for now we're limiting to linux and
        # windows.
        my $ostype = $client->type;
        if ($ostype =~ /linux/) {
            $ostype = 'linux';
        } elsif ($ostype =~ /windows/) {
            $ostype = 'windows';
        } else {
            $Log->exit;
            NATE::Exceptions::Argument->throw(
                "Client type '$ostype' is not one we can work with"
            );
        }

        # Create an igroup, named based on the volume name.
        my $name = $volume->volume . '_igroup';
        $igroup = NACL::STask::LunIgroup->create(
            command_interface  => $command_interface,
            vserver            => $vserver,
            igroup             => $name,
            protocol           => 'iscsi',
            ostype             => $ostype,
            initiator          => [ $initiator ],
            nacltask_if_exists => 'purge',
            nacltask_verify    => 1
        );

        # Map the LUN to the igroup.
        $lun->map(
            igroup          => $name,
            nacltask_verify => 1
        );

        # We will have to do iscsi discovery for this.
        $do_discovery = 1;
    }

    # 3. If $igroup is non-null, then we just created it and thus we can
    #    safely assume that the client's initiator is in the igroup. But if
    #    the LUN was already mapped, $igroup is not set and we'll have to
    #    check and possibly add $initiator to the igroup.
    if (! $igroup) {
        $igroup = NACL::STask::LunIgroup->find(
            command_interface => $command_interface,
            filter            => {
                igroup  => $mapped_lun->igroup,
                vserver => $vserver,
            }
        );
        my $initiators = $igroup->get_one_state_attribute('initiator');

        # Check for this client's initiator in the list.
        if (any { lc $_ eq $initiator } @{$initiators}) {
            # It's in the list. Just make sure that the LUN is visible on
            # the client.
            my $client_lun_state = NACL::CS::Client::Sanlun->fetch(
                command_interface => $client,
                allow_empty       => 1,
                filter            => {
                    controller   => $vserver,
                    lun_pathname => $lun->path,
                }
            );

            # If it wasn't visible, we'll have to do iscsi discovery for the
            # client.
            if (! $client_lun_state) {
                $do_discovery = 1;
            }
        } else {
            # It isn't, so add it. And since we're adding it, we'll have to do
            # iscsi discovery on the client.
            $igroup->add(initiator => [ $initiator ]);
            $do_discovery = 1;
        }
    }

    # 4. If we determined a need for iSCSI discovery on the client, do it now.
    if ($do_discovery) {
        $self->_client_iscsi_discovery;
    }

    # 5. Save the igroup object for later use.
    $self->_igroup($igroup);

    $Log->exit;
    return;
}

###############################################################################
#
#   Sub Name:       _populate_or_modify
#
#   Description:    Run data_populate() or data_modify() on a population
#                   handle, based on the value of the "action" option from
#                   @opts/%opts.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#                   @opts     in      array     Additional options, converted
#                                                 to a hash by validate()
#
#   Returns:        Success:    void
#                   Failure:    throws exception
#
###############################################################################
sub _populate_or_modify {
    $Log->enter;
    my ($self, @opts) = @_;
    my %opts = validate(
        @opts,
        {
            action    => { type => SCALAR },
            method    => {
                type     => SCALAR | UNDEF,
                optional => 1,
            },
            size      => {
                type    => SCALAR,
                default => $DEFAULT_POPULATE_SIZE,
            },
            path      => $OPTIONAL_SCALAR,
            async     => $OPTIONAL_BOOLEAN,
            arguments => {
                type     => HASHREF | UNDEF,
                optional => 1,
            },
        }
    );

    my $action = $opts{action};
    if ($action eq 'populate' || $action eq 'modify') {
        $action = "data_$action";
    } else {
        $Log->exit;
        NATE::Exceptions::Argument->throw(
            "Parameter 'action' ($action) must be one of populate or modify"
        );
    }

    # I want to make the "method" option as flexible as possible, and even
    # allow them to just use $self->data_method() as a value without worrying
    # if it has been set yet. So this chain of defaults:
    my $method =
        $opts{method} || $self->_data_method || $DEFAULT_POPULATE_METHOD;

    # We must have a valid mount-point in order to populate:
    if (! $self->_mount_isset) {
        $Log->exit;
        NATE::BaseException->throw(
            "Cannot call $action() without a valid mount-point set"
        );
    }

    # Is there another populate/modify operation currently underway?
    if ($self->_data_current_isset &&
            $self->_data_current->data_is_running) {
        $Log->exit;
        NATE::BaseException->throw(
            "Cannot call $action() while a previous populate or modify " .
            'operation is still running. Call data_wait_for_completion() ' .
            'or data_kill(), first.'
        );
    }

    my $data_object = $self->_get_data_handle(type => $method);
    $self->_data_current($data_object);
    $self->_data_method($method);

    # Determine the path to populate/modify at. Start with the mount-point, add
    # "path" if given.
    my $path = $self->mount_point;
    if ($opts{path}) {
        if ($opts{path} !~ m{^/}) {
            $path .= q{/};
        }
        $path .= $opts{path};
    }

    # Start the data action:
    $data_object->$action(
        client    => $self->_client,
        path      => $path,
        size      => $opts{size},
        async     => $opts{async} || 0,
        arguments => $opts{arguments}
    );

    $Log->exit() if $may_exit;
    return;
}

###############################################################################
#
#   Sub Name:       _get_data_handle
#
#   Description:    Get a IO::Data::* handle based on the "type" parameter
#                   passed in. Try to use an existing object if there is one
#                   for the type.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#                   %opts     in      hash      Additional args
#
#   Returns:        Success:    Object that isa *::WFC::IO::Data
#                   Failure:    throws exception
#
###############################################################################
sub _get_data_handle {
    $Log->enter;
    my ($self, %opts) = @_;
    my ($type, $handle);

    if (! ($type = $opts{type})) {
        NATE::Exceptions::Argument->throw('Empty "type" argument passed');
    }

    if ($self->_data_objects_exists($type)) {
        # If we already have an object of this type, reuse it. If this happens
        # to be the current one, that's fine-- it's in this table, too.
        $handle = $self->_data_objects_index($type);
    } else {
        # We don't have an object of this type yet. Derive the class name and
        # load the plug-in. Then instantiate it and save the resulting object
        # in the table.
        my $class = $type;
        if ($class !~ /::/) {
            # If the type has no sequence of '::' anywhere, it is relative
            # to NACL::AppBuilder::WFC::IO::Data. If it has that string, it
            # is assumed to already be the full classname.
            $class = __PACKAGE__ . "::Data::$class";
        }

        # Attempt to load:
        if (! eval { load $class; 1; }) {
            NATE::BaseException->throw("Failed to load plugin '$class': $@");
        }

        # Create an object:
        $handle = $class->new;
        # Save it:
        $self->_data_objects_set($type => $handle);
    }

    $Log->exit;
    return $handle;
}

###############################################################################
#
#   Sub Name:       _validate_traffic_file
#
#   Description:    Determine if the given file already exists, and if it
#                   does return the file size as given by slist. Return 0
#                   if it doesn't exist. If any other exception occurs, let
#                   it through.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#                   %opts     in      hash      Additional args. These have
#                                                 been pre-validated.
#
#   Returns:        Success:    file size (0 if file doesn't exist)
#                   Failure:    throws exception
#
###############################################################################
sub _validate_traffic_file {
    $Log->enter() if $may_enter;
    my ($self, %opts) = @_;
    my $file = $opts{file};
    my $size;

    try {
        my $volume   = $self->_volume;
        my $pathname = $volume->construct_dblade_file_path(filename => $file);
        my $cs       = NACL::CS::VolumeFile->fetch(
            command_interface => $volume->command_interface,
            filter            => {
                path    => $pathname,
                vserver => $volume->vserver,
            }
        );
        $size = $cs->file_size;
    } catch NACL::Exceptions::NoElementsFound with {
        # This exception means that the path doesn't exist. Any other
        # exception should be allowed to bubble up.
        $size = 0;
    };

    $Log->exit() if $may_exit;
    return $size;
}

###############################################################################
#
#   Sub Name:       _create_traffic_file
#
#   Description:    Create a file of a given size for use by the traffic
#                   generation tools.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#                   %opts     in      hash      Additional args. These have
#                                                 been pre-validated.
#
#   Globals:        %FILE_SIZES
#
#   Returns:        Success:    void
#                   Failure:    throws an exception
#
###############################################################################
sub _create_traffic_file {
    $Log->enter() if $may_enter;
    my ($self, %opts) = @_;
    my $file = $opts{file};
    my $size = $opts{size};
    my $rw   = $opts{is_rw};

    if ($rw) {
        # For read-write traffic, use NACL::C::Client::Remote::File to remotely
        # create and write to the file. If we use NACL::STask::VolumeFile the
        # file ends up owned by root, and is thus useless to us.
        require NACL::C::Client::Remote::File;

        my $client   = $self->_client;
        my $fullname =
            NACL::MTask::Traffic->get_abs_path($self->_mount, $file);
        my ($total_units, $one_unit);

        if ($size =~ /^(\d+)([kmg])$/i) {
            my ($num, $unit) = ($1, lc $2);
            if ($unit eq 'g') {
                # I don't want to create any 1-gigabyte scalars, so just bump
                # up $total_units by a factor and use a megabyte scalar.
                $total_units = $num * 1024;
                $unit = 'm';
            } else {
                $total_units = $num;
            }
            $one_unit = chr(0) x $FILE_SIZES{$unit};
        } else {
            $Log->exit;
            NATE::BaseException->throw("Value for 'size' ($size) is not valid");
        }

        # Get the remote file handle object:
        my $fh_obj = NACL::C::Client::Remote::File->open(
            command_interface => $client,
            file_name         => $fullname,
            arguments         => {
                access_mode => 'write',
                open_flags  => 'create-new',
            }
        );
        # Write the "$one_unit" at the very end of the file. This will
        # create the rest of the file by extending the file pointer out to
        # the end.
        $fh_obj->write(
            content => $one_unit,
            offset  => ($total_units - 1) * length($one_unit)
        );
        $fh_obj->close;
    } else {
        # For read-only traffic, we need to create the complete file. But
        # the volume is mounted read-only on the client, so we have to go
        # through the filer to do it.

        # Get the volume object, and then get the dblade file name for our
        # requested file:
        my $volume   = $self->_volume;
        my $pathname = $volume->construct_dblade_file_path(filename => $file);

        # OK, we should be able to create the file:
        my $fileobj = NACL::STask::VolumeFile->create(
            command_interface => $volume->command_interface,
            path              => $pathname,
            vserver           => $volume->vserver,
            size              => $size,
            nacltask_wait     => 1
        );
    }

    $Log->exit;
    return;
}

###############################################################################
#
#   Sub Name:       DESTROY
#
#   Description:    Class destructor.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#
#   Returns:        Success:    void
#                   Failure:    throws exception
#
###############################################################################
sub DESTROY {
    $Log->enter;
    my ($self) = @_;

    # Make sure that all traffic streams controlled by this object are stopped,
    # if they haven't already been. This call will do that. If there are no
    # active traffic streams it will do nothing.
    if ($self->_traffic && @{$self->_traffic}) {
        $self->traffic_stop;
    }

    $Log->exit;
    return;
}

1;
