# Copyright (c) 2014 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.

## @summary Workflow Framework Components data pop. plugin for NACL::MTask::Data
## @author rjray@netapp.com
## @status shared
## @pod here

package NACL::AppBuilder::WFC::IO::Data::NaclData;

use strict;
use warnings;
use base qw(NACL::AppBuilder::WFC::IO::Data);

use Readonly;

use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use NATE::BaseException qw(:try);
use NATE::Exceptions::Argument;
use NACL::MTask::Data;

Readonly my %BASE_ARGUMENTS => (
    populate => {
        'g' => {
            create_directory => 1,
            file_size_avg    => 65536,
            file_size_dev    => 0,
            depth_avg        => 4,
            depth_dev        => 0,
            dir_avg          => 6,
            dir_dev          => 0,
            file_avg         => 9,
            file_dev         => 0,
            link_num         => 0,
            link_pct         => 0,
            slink_num        => 0,
            slink_pct        => 0,
        },
        'm' => {
            create_directory => 1,
            file_size_avg    => 1048576,
            file_size_dev    => 0,
            depth_avg        => 0,
            depth_dev        => 0,
            dir_avg          => 0,
            dir_dev          => 0,
            file_avg         => 1,
            file_dev         => 0,
            link_num         => 0,
            link_pct         => 0,
            slink_num        => 0,
            slink_pct        => 0,
        },
    },
    modify => {
        'g' => {
            create_directory => 1,
            file_size_avg    => 65536,
            file_size_dev    => 0,
            depth_avg        => 4,
            depth_dev        => 0,
            dir_avg          => 6,
            dir_dev          => 0,
            file_avg         => 9,
            file_dev         => 0,
            link_num         => 0,
            link_pct         => 0,
            slink_num        => 0,
            slink_pct        => 0,
        },
        'm' => {
            create_directory => 1,
            file_size_avg    => 1048576,
            file_size_dev    => 0,
            depth_avg        => 0,
            depth_dev        => 0,
            dir_avg          => 0,
            dir_dev          => 0,
            file_avg         => 1,
            file_dev         => 0,
            link_num         => 0,
            link_pct         => 0,
            slink_num        => 0,
            slink_pct        => 0,
        },
    },
);

# Note: None of the methods here are using Params::Validate for parameter
# validation. This is because all the code here is called from within the
# WFC::IO class, whose methods have already validated the parameters before
# calling in to here. There is no need to double-validate, since end-users will
# not be calling this code directly.

=head1 NAME

NACL::AppBuilder::WFC::IO::Data::NaclData - NACL::MTask::Data plugin

=head1 SYNOPSIS

This module is not used directly. See the B<NACL::AppBuilder::WFC::IO::Data>
class and the B<data_*> methods in B<NACL::AppBuilder::WFC::IO>.

=head1 DESCRIPTION

The B<NACL::AppBuilder::WFC::IO::Data::NaclData> class is a data plugin class
that is used with the B<NACL::AppBuilder::WFC::IO> B<data_populate> and
B<data_modify> methods. It provides data population and data modification by
way of the B<NACL::MTask::Data> module.

Users should not use this module directly. It will be loaded on demand by
B<NACL::AppBuilder::WFC::IO> as needed.

=head1 METHODS

This class provides the following methods. See also the methods in the
B<NACL::AppBuilder::WFC::IO::Data> base class.

=head2 data_populate

Perform data population by passing the arguments passed in to the B<populate>
method from B<NACL::MTask::Data>.

=over

=item Options

The following options should be passed from the calling IO object:

=over

=item C<< client =E<gt> $client >>

An object of the B<NACL::C::Client> class, that will be used to actually
perform the data population.

=item C<< path =E<gt> $string >>

The path relative to the mount point at which data should be populated.

=item C<< size =E<gt> $string >>

The size of data to be created.

=item C<< async =E<gt> $boolean >>

Whether or not to run the data population asynchronously.

=item C<< arguments =E<gt> $hashref >>

A hash reference of arguments to be passed directly to the B<populate>
method. See the documentation for B<NACL::MTask::Data> for details on the
available arguments. These arguments are passed in a way that guarantees they
will override the default parameters that are passed to B<populate>.

=back

=item Exceptions

Any exceptions thrown by B<NACL::MTask::Data::populate> will be passed along
unchanged.

=back

=cut

###############################################################################
#
#   Sub Name:       data_populate
#
#   Description:    Initiate data population on the client and path specified
#                   by the options to this method.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#                   %opts     in      hash      Additional arguments passed
#                                                 from the calling WFC::IO
#                                                 object
#
#   Returns:        Success:    void
#                   Failure:    throws exception
#
###############################################################################
sub data_populate {
    $Log->enter;
    my ($self, %opts) = @_;

    $self->_populate_or_modify(action => 'populate', %opts);

    $Log->exit;
    return;
}

=head2 data_modify

Perform data modification by passing the arguments passed in to the B<modify>
method from B<NACL::MTask::Data>.

=over

=item Options

The following options should be passed from the calling IO object:

=over

=item C<< client =E<gt> $client >>

An object of the B<NACL::C::Client> class, that will be used to actually
perform the data modification.

=item C<< path =E<gt> $string >>

The path relative to the mount point at which data should be modified.

=item C<< size =E<gt> $string >>

The size of data to be modified.

=item C<< async =E<gt> $boolean >>

Whether or not to run the data population asynchronously.

=item C<< arguments =E<gt> $hashref >>

A hash reference of arguments to be passed directly to the B<modify>
method. See the documentation for B<NACL::MTask::Data> for details on the
available arguments. These arguments are passed in a way that guarantees they
will override the default parameters that are passed to B<modify>.

=back

=item Exceptions

Any exceptions thrown by B<NACL::MTask::Data::populate> will be passed along
unchanged.

=back

=cut

###############################################################################
#
#   Sub Name:       data_modify
#
#   Description:    Initiate data modification on the client and path specified
#                   by the options to this method.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#                   %opts     in      hash      Additional arguments passed
#                                                 from the calling WFC::IO
#                                                 object
#
#   Returns:        Success:    void
#                   Failure:    throws exception
#
###############################################################################
sub data_modify {
    $Log->enter;
    my ($self, %opts) = @_;

    $self->_populate_or_modify(action => 'modify', %opts);

    $Log->exit;
    return;
}

=head2 data_wait_for_completion

Wait for the current populate or modify operation to complete. Returns
immediately if there is no current operation in progress, or if the operation
has already finished.

=over

=item Options

The following option is recognized, and should be passed from the calling IO
object:

=over

=item C<< timeout =E<gt> $number >>

Specify a timeout value for the wait. If the process takes longer than this
to finish, an exception will be thrown.

=back

=item Exceptions

Any exceptions thrown by B<NACL::MTask::Data::wait_for_completion> will be
passed along unchanged.

=back

=cut

###############################################################################
#
#   Sub Name:       data_wait_for_completion
#
#   Description:    Do a wait for a currently-backgrounded job to finish. If
#                   there is no job, quietly does nothing.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object
#                   %opts     in      hash      Extra options for the NACL
#                                                 call. Only "timeout" is
#                                                 looked for.
#
#   Returns:        Success:    void
#                   Failure:    throws exception
#
###############################################################################
sub data_wait_for_completion {
    $Log->enter;
    my ($self, %opts) = @_;
    my $timeout = $opts{timeout} || 0;

    # If we currently have a handle, do some stuff. If we don't have the handle
    # then we just quietly exit.
    if ($self->handle_isset) {
        # Use the NACL::MTask::Data call to wait for it to finish.
        $self->handle->wait_for_completion(timeout => $timeout);
        # Clear out the handle, now that we're done with it.
        $self->handle_reset;
    }

    $Log->exit;
    return;
}

=head2 data_kill

Immediately kill a running populate or modify operation. Does nothing if there
is no running operation.

=over

=item Options

No options are recognized by this method.

=item Exceptions

Any exceptions thrown by B<NACL::MTask::Data::kill> will be passed along
unchanged.

=back

=cut

###############################################################################
#
#   Sub Name:       data_kill
#
#   Description:    Kill a currently-backgrounded job. If there is no job,
#                   quietly does nothing.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object
#
#   Returns:        void
#
###############################################################################
sub data_kill {
    $Log->enter;
    my ($self) = @_;

    # If we currently have a handle, do some stuff. If we don't have the handle
    # then we just quietly exit.
    if ($self->handle_isset) {
        # Use the NACL::MTask::Data call to kill it.
        $self->handle->kill;
        # Clear out the handle, now that we're done with it.
        $self->handle_reset;
    }

    $Log->exit;
    return;
}

###############################################################################
#
#   Sub Name:       _populate_or_modify
#
#   Description:    Do the population or modification (based on the parameter
#                   "action").
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#                   %opts     in      hash      Additional arguments passed
#                                                 from the calling WFC::IO
#                                                 object
#
#   Globals:        %BASE_ARGUMENTS
#
#   Returns:        Success:    void
#                   Failure:    throws exception
#
###############################################################################
sub _populate_or_modify {
    $Log->enter;
    my ($self, %opts) = @_;

    my $action    = $opts{action};
    my $client    = $opts{client};
    my $path      = $opts{path};
    my $async     = $opts{async};
    my $size      = $opts{size};
    my $user_args = $opts{arguments} || {};

    # If we currently have a running handle, throw an exception. For now we're
    # limited to one operation per object at a time.
    if ($self->data_is_running) {
        NATE::BaseException->throw(
            "Cannot call data_$action() while a previous populate or modify " .
            'operation is still running. Call data_wait_for_completion() ' .
            'or data_kill(), first.'
        );
    }
    $self->handle_reset;

    # Start by setting up the base arguments for populate or modify. The hash
    # of options specifies a 1G and a 1M dataset. Based on the unit of the
    # value passed in, we copy the set we need, then multiply the desired size
    # out on the "file_avg" key, which is the number of files created.
    my %populate_args = ();
    if ($size =~ /^(\d+)([mg])b?$/i) {
        # First copy the set of parameters for the appropriate units given:
        %populate_args = %{$BASE_ARGUMENTS{$action}->{lc $2}};
        # Scale out the size of the dataset by increasing the number of files:
        $populate_args{file_avg} *= $1;
    } else {
        $Log->exit() if $may_exit;
        NATE::Exceptions::Argument->throw(
            "Value for 'size' ($size) is invalid or in units other than " .
            'megabytes (m) or gigabytes (g)'
        );
    }

    # Start the data operation:
    my $data_handle = NACL::MTask::Data->$action(
        command_interface => $client,
        path              => $path,
        async             => $async,
        template          => 'none',
        %populate_args,
        %{$user_args}
    );

    # If this was an async operation, then save the data handle on our local
    # accessor. If it wasn't async, then it's already completed and we don't
    # need it.
    if ($async) {
        $self->handle($data_handle);
    }

    $Log->exit;
    return;
}

=head1 SEE ALSO

B<NACL::AppBuilder::WFC::IO::Data>, B<NACL::AppBuilder::WFC::IO>,
B<NACL::MTask::Data>

=head1 AUTHOR

Randy J. Ray E<lt>rjray@netapp.comE<gt>

=cut

1;
