#
# Copyright (c) 2015 NetApp, Inc., All Rights Reserved.
#   Any other use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

package NACL::AppBuilder::Xgen::Config::Asup2Config::XMLParser;

use strict;
use warnings;

use NATE::Inc qw( find_file );
use Carp qw( confess );
use XML::LibXML;

# Creates the XMLParser object for the given file.  Internally, it
#   creates a xml doc object and stores it on self, and creates
#   the lookup table needed for qualifications.
# In:     (required) file => $filename
# Out:    $xml_parser_obj
# Fails:  Carp->confess("reason") for any failure
#           - No file provided as parameter
#           - Cannot open file
sub new {
    my ($class, %opts) = @_;

    # Verify file param passed in
    my $file = $opts{file} // confess("Parameter 'file' must be supplied");

    # Get path of file if relative
    $file = find_file($file);

    # Assign opts to self and bless itself
    my $self = \%opts;
    bless($self, $class);

    # Create doc object
    my $parser = XML::LibXML->new();
    my $doc;
    eval {
        $doc = $parser->parse_file($file);
    };
    if ($@) {
        confess("Could not parse XML file $file.");
    }

    # Store on self
    $self->{xml_doc} = $doc;

    # Create lookup table
    $self->_store_lookup_table();

    return $self;
}

# Returns all asup:ROWs in the XML file as an arrayref of hashrefs.
#   Field names will be changed to their smf names as defined in
#   in the xml.  The results are also filtered by the qualifications
#   and attribute filters passed in.
# In:     (optional) qualifications => ["create"|"key"|etc.]
#             Will filter the fields returned for each asup:ROW
#         (optional) attr_filters => { attr_name => $regex, ... }
#             Will filter the asup:ROWs returned based on matching
#             the attr_name's value with the $regex provided.
#         (optional) allow_empty_values => 0|1
#             Will allow empty value field to be include in data.
#             Default 0;
# Out:    $rows_arrayref
#             [ { attr_name => $value, ... } , { ... } ]
# Fails:  Carp->confess("reason") for any failure
#         However, attr_filters that match no asup:ROWs return []
#           - qualifications contains invalid value
sub get_rows {
    my ($self, %opts) = @_;

    #################
    # Initialization
    #################

    # Get parameters
    my $qualifications = delete($opts{qualifications});
    my $attr_filters = delete($opts{attr_filters});
    my $allow_empty_values = delete($opts{allow_empty_values}) // 0;
    # Get xml doc
    my $doc = $self->{xml_doc};

    # Initialize return value
    my @ret_rows = ();

    #################
    # Read ROWs into perl data structure
    #################

    foreach my $asup_row ($doc->findnodes('//asup:ROW')) {
        my %row;
        foreach my $field ($asup_row->getChildrenByTagName("*")) {
            # If the field has a asup:list of values
            if ($field->getChildrenByTagName("asup:list")) {
                my $list = $field->getChildrenByTagName("asup:list")->[0];
                my @elems = $list->getChildrenByTagName("asup:li");
                my @field_values;
                foreach my $elem (@elems) {
                    push(@field_values, $elem->textContent());
                }
                $row{$field->nodeName()} = \@field_values;
            }
            # If the field is a normal string value
            else {
                # Only allow empty value fields if the user wants them included.
                if ($field->textContent() or $allow_empty_values) {
                    # If a value has a space then encapsulate in double quotes.
                    if ($field->textContent() =~ /\s/) {
                        $row{$field->nodeName()} = '"' . $field->textContent() . '"';
                    } else {
                        $row{$field->nodeName()} = $field->textContent();
                    }
                }
            }
        }
        push(@ret_rows, \%row);
    }

    #################
    # Change field names to smf names
    #################
    my $ref = $self->{lookup_table};
    my @lookup_table = @$ref;

    foreach my $table_entry (@lookup_table) {
        my $orig_name = $table_entry->{'asup:tag'};
        my $smf_name = $table_entry->{'asup:smf_name'};

        foreach my $row (@ret_rows) {
            if (exists $row->{$orig_name}) {
                $row->{$smf_name} = delete $row->{$orig_name};
            }
        }
    }

    #################
    # Filter attr_filters
    #################
    if (defined $attr_filters) {
        my %filters = %$attr_filters;
        my @final_ret_rows;
        OUTER: foreach my $row (@ret_rows) {
            foreach my $filter_field (keys %filters) {
                if ((not defined $row->{$filter_field}) or
                    ($row->{$filter_field} !~ $filters{$filter_field})) {
                    next OUTER;
                }
            }
            push(@final_ret_rows, $row);
        }
        @ret_rows = @final_ret_rows;
    }

    #################
    # Filter qualifications
    #################
    if (defined $qualifications) {
        my @qualifications = @$qualifications;
        foreach my $row (@ret_rows) {
            foreach my $field (keys %$row) {
                foreach my $lookup (@lookup_table) {
                    if ($lookup->{'asup:smf_name'} eq $field) {
                        my $found = 0;
                        foreach my $qual (@qualifications) {
                            if ($lookup->{'asup:qualification'} eq $qual) {
                               $found = 1;
                               last;
                            }
                        }
                        delete $row->{$field} if ($found == 0) ;
                        last;
                    }
                }
            }
        }
    }

    return \@ret_rows;
}

sub _store_lookup_table {
    my ($self,%opts) = @_;

    my $doc = $self->{xml_doc};
    my @table;

    foreach my $param ($doc->findnodes('//asup:field')) {
        my %elem;
        foreach my $field ($param->getChildrenByTagName("*")) {
            $elem{$field->nodeName()} = $field->textContent();
        }
        push(@table, \%elem);
    }

    $self->{lookup_table} = \@table;
    return 1;
}
1;
