#
# Copyright (c) 2015 NetApp, Inc., All Rights Reserved.
#   Any other use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

package NACL::AppBuilder::Xgen::Config::DMG::Base;

use strict;
use warnings;

use parent qw( NACL::AppBuilder::Xgen::Config::ONTAP );
use NATE::Log qw( log_global );
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();

use NATE::BaseException qw(:try);
use NACL::AppBuilder::Xgen::Resource::ClusterPeer;
use NACL::AppBuilder::Xgen::Resource::VserverPeer;
use NACL::AppBuilder::Xgen::Resource::Vserver;
use NACL::AppBuilder::Xgen::Resource::Volume;
use NACL::AppBuilder::Xgen::Resource::Volume::Efficiency;
use NACL::AppBuilder::Xgen::Resource::Volume::EfficiencyPolicy;
use NACL::AppBuilder::Xgen::Resource::Lun;
use NACL::AppBuilder::Xgen::Resource::LunMap;
use NACL::AppBuilder::Xgen::Resource::File;
use NACL::AppBuilder::Xgen::Resource::Iscsi;
use NACL::AppBuilder::Xgen::Resource::Igroup;
use NACL::AppBuilder::Xgen::Resource::Igroup::Initiator;
use NACL::AppBuilder::Xgen::Resource::Lif;
use NACL::AppBuilder::Xgen::Resource::NetworkRoute;
use NACL::AppBuilder::Xgen::Resource::Snapmirror;
use NACL::AppBuilder::Xgen::Resource::SnapmirrorInitialize;
use NACL::AppBuilder::Xgen::Resource::SnapmirrorPolicy;
use NACL::AppBuilder::Xgen::Resource::SnapmirrorPolicyRule;
use NACL::AppBuilder::Xgen::Resource::SnapshotPolicy;
use NACL::AppBuilder::Xgen::Resource::JobScheduleCron;
use NACL::Service::Xgen::Util qw(convert_percent_layout_to_count_layout
                                 get_next_layout_element
                                 dumper
                                 positive_filter);
use NATE::BaseException;
use POSIX qw(ceil floor);
use Hostrec;

# These are the constant values that determine what happens in the config.
# This constant controls the number of aggregates per node for the config.
use constant AGGREGATES_PER_NODE => 1;
# This constant controls the total number of vservers in the config.
use constant TOTAL_VSERVER_COUNT => 4;
# This constant controls the total number of volumes in the config.
use constant TOTAL_VOLUME_COUNT => 88;
# This constant controls the total number of luns in the confg.
use constant TOTAL_LUN_COUNT => 176;
# This constant controls the total number of files in the config.
use constant TOTAL_FILE_COUNT => 176;
# This constant controls the total number of snapmirrors in the config.
use constant TOTAL_SNAPMIRROR_COUNT => 44;
# This constant controls the number of disks per aggregate.
use constant NUMBER_DISKS_PER_AGGREGATE => 10;
# This constant controls the size of all the volumes in the config.
use constant VOLUME_SIZE => "500m";
# This constant controls the size of all the luns in the config.
use constant LUN_SIZE => "35m";
# This constant controls the size of all the files in the config.
use constant FILE_SIZE => "20m";
# This is used to control the percentage of volumes with luns
use constant PERCENTAGE_OF_VOLUMES_WITH_LUNS => 100;
# This is used to control the percentage of volumes with files.
use constant PERCENTAGE_OF_VOLUMES_WITH_FILES => 100;
# This constant will control the number of cascade configs
use constant TOTAL_CASCADE_CONFIG_COUNT => 2;
# This constant will control the layouts of the cascade configs based on percents.
use constant CASCADE_TYPE_PERCENT_LAYOUT => {'ADR;VAULT'       => 50,
                                             'ADR;ADR;VAULT'   => 50};
# This constant will control the number of fanin configs
use constant TOTAL_FANIN_CONFIG_COUNT => 2;
# This constant will control the layouts of the fanin configs based on percents.
use constant FANIN_TYPE_PERCENT_LAYOUT => {'ADR;VAULT'       => 50,
                                           'ADR;ADR;VAULT'   => 50};
# This constant will control the number of fanin configs
use constant TOTAL_FANOUT_CONFIG_COUNT => 2;
# This constant will control the layouts of the fanin configs based on percents.
use constant FANOUT_TYPE_PERCENT_LAYOUT => {'ADR;VAULT'       => 50,
                                            'ADR;ADR;VAULT'   => 50};
# This constant will control the number of fanin configs
use constant TOTAL_BIDIRECTIONAL_CONFIG_COUNT => 2;
# This constant will control the layouts of the fanin configs based on percents.
use constant BIDIRECTIONAL_TYPE_PERCENT_LAYOUT => {'ADR;ADR'       => 50,
                                                   'VAULT;VAULT'   => 50};
# This constant controls the percentage of all Snapmirrors that will
# be each type of Snapmirror relationship.
use constant UNIDIRECTIONAL_SNAPMIRROR_TYPE_PERCENT_LAYOUT => {SMGR                            => 20,
                                                              ADR                             => 4,
                                                              ADR_NC                          => 4,
                                                              VAULT                           => 4,
                                                              VAULT_NC                        => 4,
                                                              VAULT_SRCSE                     => 4,
                                                              VAULT_SRCCOMPRESSION            => 4,
                                                              VAULT_ADDDEDUP                  => 4,
                                                              VAULT_LRE                       => 4,
                                                              LRSE_DR_MIRROR_LATEST           => 4,
                                                              LRSE_DR_MIRROR_ALL              => 4,
                                                              LRSE_DR_MIRROR_AND_VAULT        => 4,
                                                              LRSE_DR_MIRROR_LATEST_LRE       => 4,
                                                              LRSE_DR_MIRROR_ALL_LRE          => 4,
                                                              LRSE_DR_MIRROR_AND_VAULT_LRE    => 4,
                                                              LRSE_DR_MIRROR_LATEST_NC        => 4,
                                                              LRSE_DR_MIRROR_ALL_NC           => 4,
                                                              LRSE_DR_MIRROR_AND_VAULT_NC     => 4,
                                                              LRSE_DR_MIRROR_LATEST_NC_LRE    => 4,
                                                              LRSE_DR_MIRROR_ALL_NC_LRE       => 4,
                                                              LRSE_DR_MIRROR_AND_VAULT_NC_LRE => 4,};
# This constant controls the percentage of SMGR relationships that
# will be lun versus file.
use constant SMGR_CG_ITEM_PERCENT_LAYOUT => {'lun'   => 75,
                                             'file'  => 25};

# This constant controls the percentage of luns or files that are
# used in the SMGR relationships
use constant PERCENTAGE_OF_LUNS_FILES_USED_FOR_CG_ITEMS => 50;

# This constant controls the percentage of volumes that will have
# a vm-align-suffix and what the value will be.
# :NONE: means that the volumes will not have a vm align suffix.
use constant VOLUME_VM_ALIGN_SUFFIX_PERCENT_LAYOUT => {':NONE:' => 100,
                                                       '.vmdk'  => 0};
# This constant controls the percentage of volumes that will have
# a vm-align-sector and what the value will be.
# This percentage will only affect volumes that have a
# vm-align-suffix.
use constant VOLUME_VM_ALIGN_SECTOR_PERCENT_LAYOUT => {'1' => 50,
                                                       '2' => 50};
# The Following 2 constants control the minimum and maximum
# percent of files in a vm aligned volume that have the correct
# suffix. If the values are equal then the value is to percent
# that all volumes will recieve. If the values are not equal
# then the percent will be chosen randomly between the
# minimum and the maximum.  The random_seed if fixed by default.
# A user can pass in a random value for 'random_seed' in args to
# the config to set the random seed to a random value. This means
# that by default a config with the same constant values will always
# build the same thing unless random_seed is passed in.
use constant MINIMUM_PERCENT_VM_ALIGNED_FILES => 50;
use constant MAXIMUM_PERCENT_VM_ALIGNED_FILES => 50;
# This constant is used to turn on and off setup and cleanup for
# snapmirrors in EnGenX for this configuration. The snapmirror
# resources objects are still created but will not be created
# or deleted in ontap by EnGenX
use constant SETUP_CLEANUP_SNAPMIRRORS => 1;
# This constant is used to turn on and off snapmirror initialization.
use constant INITIALIZE_SNAPMIRRORS => 1;
# This constant is used to start or not start volume_efficiency
# Volume Efficiency will not start regardless for destinations of SMs.
use constant START_VOLUME_EFFICIENCY => 1;

#aggr name format: node_name, node_id, aggr_only
use constant AGGR_NAME_FORMAT => 'aggr_only' ;

#volume name format: vserver (append vserver name), volume_only
use constant VOLUME_NAME_FORMAT => 'volume_only' ;
###############################################################################
# Description: The parent generates aggregates based on the override method
#              generate_aggregate_resources.  Calls modular functions to
#              generate resources.
# Input: Arguments to pass to the module methods and other Xgen controls.
# Output: NONE
# Exceptions: None thrown directly by this method.
###############################################################################
sub build {
    my ($self, %args) = @_;

    #save xgen params
    $self->xgen_config_params($args{xgen_params} || {});

    $Log->comment("keys passed in to xgen dump:" . dumper $self->xgen_config_params()); #TODO set to trace or remove
    $Log->comment("resources passed in to xgen dump:" . dumper $args{resources});

    # We only want to do this if something did not pass in any Ethernet fabric definitions.
    # This fabric will include all the ips from all nodes and not sort specific to clusters.
    # This means that IPs will not be owned by a specific cluster and can be used by any cluster.
    unless (scalar keys %{$args{eth_fabric_data}}) {
        $Log->comment("Generating Ethernet Fabric data from Natehost files.");
        $args{eth_fabric_data} = $self->generate_eth_fabric_data_from_natehost_files(%args);
    }

    $Log->comment("Generating Parent Configuration.");
    $self->SUPER::build(%args);

    $Log->comment("Generating Cluster Peers.");
    # Call to generate cluster peer resources.
    $self->generate_cluster_peer_resources(%args);

    $Log->comment("Generating Vservers.");
    # Call to generate vserver_peer resources.
    $self->generate_vserver_resources(%args);

    $Log->comment("Generating ISCSI Lifs.");
    # Call to generate iscsi lif resources.
    $self->generate_iscsi_lif_resources(%args);

    $Log->comment("Generating NFS Lifs.");
    # Call to generate nfs lif resources.
    $self->generate_nfs_lif_resources(%args);

    $Log->comment("Generating Igroups.");
    # Call to genereate iscsi service resources.
    $self->generate_igroup_resources(%args);

    $Log->comment("Generating Initiators.");
    # Call to genereate iscsi service resources.
    $self->generate_igroup_initiator_resources(%args);

    $Log->comment("Generating ISCSI Services.");
    # Call to genereate iscsi service resources.
    $self->generate_iscsi_services_resources(%args);

    $Log->comment("Generating Volumes.");
    # Call to generate volume resources.
    $self->generate_volume_resources(%args);

    $Log->comment("Generating Snapmirror Resources.");
    # Call to generate snapmirror resources.
    $self->generate_snapmirror_resources(%args);

    $Log->comment("Generating Luns.");
    # Call to generate lun resources.
    $self->generate_lun_resources(%args);

    $Log->comment("Generating Lun Maps.");
    # Call to generate lun resources.
    $self->generate_lun_map_resources(%args);

    $Log->comment("Generating Files.");
    # Call to generate file resources.
    $self->generate_file_resources(%args);

    $Log->comment("Adding CG Items to SMGR Snapmirrors.");
    # Call to add_cg_items_to_SMGR_snapmirror_resources
    # This must be after luns and files are created
    $self->add_cg_items_to_SMGR_snapmirror_resources(%args);

    $Log->comment("Generating Volume Efficiency Options.");
    # Call to generate_volume_efficiency_resources
    # This must be after snapmirrors are created since
    # they control the volumes we want efficiency on.
    $self->generate_volume_efficiency_resources(%args);

    $Log->comment("Generating Vserver Peers.");
    # Call to generate vserver_peer resources.
    $self->generate_vserver_peer_resources(%args);
}

###############################################################################
# DESCRIPTION: Builds an ethernet fabric data hash from the nate_host files
#              for all the nodes found by NACL::C::Node->find();  Uses the
#              attributes DATA_IP, DATA_GATEWAY, DATA_MASK from the nate_host
#              files.
# INPUT: NONE
# OUTPUT: A hashref of the built ethernet fabric data hash.
# Exceptions: None thrown directly by this method.
###############################################################################
# get param, either passed in or the constant default
#
sub param {
    my $self  = shift;
    my $param = shift;

    my $value = $self->xgen_config_params()->{$param} // $self->get_constant($param);

    $Log->comment("xgen config param $param value: $value"); #TODO set to trace or remove
    #$Log->comment("xgen config params dump:" . dumper $self->xgen_config_params()); #TODO set to trace or remove

    return $value;
}

sub get_constant {
    my $self  = shift;
    my $param = shift;

    my $constant;

    my $method = $self->can($param);
    if (defined $method) {
        $constant = $self->$method;
        $Log->comment("constant param $param value: $constant\n"); #TODO set to trace or remove
    } else {
        $Log->comment("constant param $param NOT defined as a constant\n"); #TODO set to trace or remove
    }

    return $constant;
}

#setter/getter for xgen config params that are passed in
sub xgen_config_params {
    my $self = shift;

    my $param = shift;

    if (defined $param) {
        $self->{xgen_config_params} = $param;
    }

    return $self->{xgen_config_params};
}
#sub paramset {
#my $self = shift;
#$Log->comment("paramset dump:" . Dumper $Paramset); #TODO set to trace or remove
#return $Paramset;
#}

sub generate_eth_fabric_data_from_natehost_files {
    my ($self, %args) = @_;

    my @NODES = NACL::C::Node->find();
    my @ips;
    my $netmask = undef;
    my $gateway = undef;
    foreach my $node (@NODES) {
        my $hostrec = Hostrec->new(id=>$node->node());
        my %networks = $hostrec->param_fields();
        foreach my $key (keys %networks) {
            if($key =~ /DATA_IP/i) {
                push (@ips, $networks{$key});
            } elsif ($key =~ /DATA_GATEWAY/i) {
                $gateway = $networks{$key};
            } elsif ($key =~ /DATA_MASK/i) {
                $netmask = $networks{$key};
            }
        }
    }
    my %eth_fabric_data = ();
    $eth_fabric_data{'eth1'} = {
        'ipspace'           => 'Default',
        'broadcast_domain'  => 'Default',
        'address_range_v4'  => \@ips,
        'netmask_v4'        => $netmask,
        'gateway_v4'        => $gateway,
    };

    $Log->comment("The fabric data that was generated is " . dumper(\%eth_fabric_data));
    return \%eth_fabric_data
}

###############################################################################
# DESCRIPTION: Overrides the generate_aggregate_resources method from ONTAP.pm
#              so that it does not query for aggregates but instead creates
#              aggregates based on the AGGREGATES_PER_NODE and
#              NUMBER_DISKS_PER_AGGREGATE constants. The naming convention
#              use is n#_aggr#. Where both # start at 0 for each cluster.
#              TODO: Currently the N# does not start at 0 for each cluster.
#                    ONTAP.pm may need to be updated.
# INPUT: NONE
# OUTPUT: A hashref of the built ethernet fabric data hash.
# Exceptions: None thrown directly by this method.
###############################################################################
sub generate_aggregate_resources {
    my ($self, %args) = @_;

    my $nodes = $self->get_resources(type => "NACL::AppBuilder::Xgen::Resource::Node");

    #aggr name format: node_name, node_id, only_aggr
    my $aggr_name_format = $self->param('AGGR_NAME_FORMAT') || 'node_name';
    my %cluster_aggr_count_hash = ();
    for (my $i = 0; $i < $self->param('AGGREGATES_PER_NODE'); $i++) {
        foreach my $node (@{$nodes}) {
            my $node_name = $node->get_attribute('node');
            my $aggr_number = $cluster_aggr_count_hash{$node->get_attribute('cluster_name')} // 0;
            $cluster_aggr_count_hash{$node->get_attribute('cluster_name')}++;

            my $aggr_name;
            if ($aggr_name_format eq 'node_name') {
                $aggr_name = $node_name . "_aggr" . $aggr_number;
            } elsif ($aggr_name_format eq 'node_id') {
                $aggr_name = "node" . $self->get_node_id($node_name) . "_aggr" . $aggr_number;
            } elsif ($aggr_name_format eq 'aggr_only') {
                $aggr_name = "aggr_" . $aggr_number;
            } else {
                $aggr_name = "aggr_" . $aggr_number;
            }

            # make the aggregate resource
            my $aggregate = NACL::AppBuilder::Xgen::Resource::Aggregate->new('aggregate'    => $aggr_name,
                                                                             'diskcount'    => $self->param('NUMBER_DISKS_PER_AGGREGATE'),
                                                                             'nodes'        => [$node_name],
                                                                             'maxraidsize'  => $self->param('NUMBER_DISKS_PER_AGGREGATE'));
            $aggregate->add_parent_resources(resources => [$node]);
            $self->add_resources(resources => [$aggregate]);
        }
    }
}

###############################################################################
# DESCRIPTION: Generates NFS lif resources and adds them to the configuration.
#              N Lif resources are created in every vserver per fabric defined,
#              where N is the number of nodes in the cluster that the vserver
#              is in. The lifs use the IPs, Netmask, and Gateway described in
#              the fabrics in the eth_data_fabric hash. This method currently
#              supports IPV4.
# INPUT: eth_fabric_data - a hash ref of ethernet fabric data including
#                          ips, netmask, gateway(optional) for the different
#                          ethernet fabrics for the testbed.
# OUTPUT: NONE
# Exceptions: None thrown directly by this method.
###############################################################################
sub generate_nfs_lif_resources {
    my ($self, %args) = @_;

    my $vservers = $self->get_resources(type => "NACL::AppBuilder::Xgen::Resource::Vserver");

    foreach my $vs (@{$vservers}) {
        my $eth_fabric_data = $args{eth_fabric_data};

        # Each vserver should only have one ancestor that is a cluster.
        my ($cluster) = $vs->get_ancestors(type => "NACL::AppBuilder::Xgen::Resource::Cluster");
        # Get all the nodes that are children of the vservers cluster so that we get all the nodes.
        my @nodes = $cluster->get_descendants(type => "NACL::AppBuilder::Xgen::Resource::Node");
        my $already_create_routes;
        # make an iscsi lif on each node (on each fabric)
        foreach my $node (@nodes) {
            my $node_name = $node->get_attribute('node');
            my $node_id = $self->get_node_id($node_name);

            # obtain the ethenet fabric table for this node
            my $fabric_table = $self->get_fabric_lookup_table_by_node($node_name, 'eth');
            foreach my $fabric (keys %{$fabric_table}) {
                my $address_range_v4 = $eth_fabric_data->{$fabric}->{address_range_v4};
                my $netmask_v4       = $eth_fabric_data->{$fabric}->{netmask_v4};
                if (!$address_range_v4 or !$netmask_v4) {
                    NATE::Exceptions::Argument->throw("No address ranges and/or netmasks were provided for fabric '$fabric'");
                }
                # if this fabric has a gateway defined,
                # make a route resource on this vserver if we havn't built one yet for this fabric
                $already_create_routes->{$fabric} //= {};
                if (!$already_create_routes->{$fabric}->{route_resource} and $eth_fabric_data->{$fabric}->{gateway_v4}) {
                    my $route_resource = NACL::AppBuilder::Xgen::Resource::NetworkRoute->new(vserver => $vs->get_attribute('vserver'),
                                                                                             gateway => $eth_fabric_data->{$fabric}->{gateway_v4});
                    $route_resource->add_parent_resources(resources => [$vs]);
                    $self->add_resources(resources => [$route_resource]);
                    $already_create_routes->{$fabric}->{route_resource} = $route_resource;
                }

                # choose a port on this fabric for this node
                my $port_name = shift(@{$fabric_table->{$fabric}});
                push(@{$fabric_table->{$fabric}}, $port_name);
                my $port_id = $self->get_port_id($node_name, $port_name);

                # make the lif resource
                my $lif = NACL::AppBuilder::Xgen::Resource::Lif->new(vserver => $vs->get_attribute('vserver'),
                                                                     'data-protocol' => ['nfs'],
                                                                     address_range => $address_range_v4,
                                                                     netmask => $netmask_v4,
                                                                     lif => "nfs_lf_n" . $node_id . "_p" . $port_id,
                                                                     'home-node' => $node_name,
                                                                     'home-port' => $port_name);
                $lif->add_parent_resources(resources => [$vs]);
                $self->add_resources(resources => [$lif]);
            }
        }
    }
}

###############################################################################
# DESCRIPTION: Generates ISCSI lif resources and adds them to the configuration.
#              N Lif resources are created in every vserver per fabric defined,
#              where N is the number of nodes in the cluster that the vserver
#              is in. The lifs use the IPs, Netmask, and Gateway described in
#              the fabrics in the eth_data_fabric hash. This method currently
#              supports IPV4.
# INPUT: eth_fabric_data - a hash ref of ethernet fabric data including
#                          ips, netmask, gateway(optional) for the different
#                          ethernet fabrics for the testbed.
# OUTPUT: NONE
# Exceptions: None thrown directly by this method.
###############################################################################
sub generate_iscsi_lif_resources {
    my ($self, %args) = @_;

    my $vservers = $self->get_resources(type => "NACL::AppBuilder::Xgen::Resource::Vserver");

    foreach my $vs (@{$vservers}) {
        my $eth_fabric_data = $args{eth_fabric_data};

        # Each vserver should only have one ancestor that is a cluster.
        my ($cluster) = $vs->get_ancestors(type => "NACL::AppBuilder::Xgen::Resource::Cluster");
        # Get all the nodes that are children of the vservers cluster so that we get all the nodes.
        my @nodes = $cluster->get_descendants(type => "NACL::AppBuilder::Xgen::Resource::Node");
        my $already_create_routes;
        # make an iscsi lif on each node (on each fabric)
        foreach my $node (@nodes) {
            my $node_name = $node->get_attribute('node');
            my $node_id = $self->get_node_id($node_name);

            # obtain the ethenet fabric table for this node
            my $fabric_table = $self->get_fabric_lookup_table_by_node($node_name, 'eth');
            foreach my $fabric (keys %{$fabric_table}) {
                my $address_range_v4 = $eth_fabric_data->{$fabric}->{address_range_v4};
                my $netmask_v4       = $eth_fabric_data->{$fabric}->{netmask_v4};
                if (!$address_range_v4 or !$netmask_v4) {
                    NATE::Exceptions::Argument->throw("No address ranges and/or netmasks were provided for fabric '$fabric'");
                }
                # if this fabric has a gateway defined,
                # make a route resource on this vserver if we havn't built one yet for this fabric
                $already_create_routes->{$fabric} //= {};
                if (!$already_create_routes->{$fabric}->{route_resource} and $eth_fabric_data->{$fabric}->{gateway_v4}) {
                    my $route_resource = NACL::AppBuilder::Xgen::Resource::NetworkRoute->new(vserver => $vs->get_attribute('vserver'),
                                                                                             gateway => $eth_fabric_data->{$fabric}->{gateway_v4});
                    $route_resource->add_parent_resources(resources => [$vs]);
                    $self->add_resources(resources => [$route_resource]);
                    $already_create_routes->{$fabric}->{route_resource} = $route_resource;
                }

                # choose a port on this fabric for this node
                my $port_name = shift(@{$fabric_table->{$fabric}});
                push(@{$fabric_table->{$fabric}}, $port_name);
                my $port_id = $self->get_port_id($node_name, $port_name);

                # make the lif resource
                my $lif = NACL::AppBuilder::Xgen::Resource::Lif->new(vserver => $vs->get_attribute('vserver'),
                                                                     'data-protocol' => ['iscsi'],
                                                                     address_range => $address_range_v4,
                                                                     netmask => $netmask_v4,
                                                                     lif => "iscsi_lf_n" . $node_id . "_p" . $port_id,
                                                                     'home-node' => $node_name,
                                                                     'home-port' => $port_name);
                $lif->add_parent_resources(resources => [$vs]);
                $self->add_resources(resources => [$lif]);
            }
        }
    }
}

###############################################################################
# DESCRIPTION: Generates file resources and adds them to the configuration.
#              Generates a number of files equal to the TOTAL_FILE_COUNT
#              constant.  The files will only be generated on non DP type
#              volumes. The PERCENTAGE_OF_VOLUMES_WITH_FILES controls the
#              percent of the non DP type volumes that get files. If a
#              volume has a vm-align-suffix the MAXIMUM_PERCENT_VM_ALIGNED_FILES
#              and MINIMUM_PERCENT_VM_ALIGNED_FILES constants control the range
#              of random percentages of files that have the vm-align-suffix in
#              each volume. The file size is controlled by the FILE_SIZE
#              constant.
# INPUT: NONE
# OUTPUT: NONE
# Exceptions: None thrown directly by this method.
###############################################################################
sub generate_file_resources {
    my ($self, %args) = @_;

    # Get the clusters so that we can alternate vserver creation between multiple clusters.
    my $volume_resources = $self->get_resources(type => "NACL::AppBuilder::Xgen::Resource::Volume");

    # Filter the volumes resources by security style mixed.
    my $matches_hash = positive_filter(resources => $volume_resources,
                                       filter    => {'type' => 'DP'});

    my @volumes = @{$matches_hash->{'not_matches'}};

    my $volume_count = ceil((scalar(@volumes) * $self->param('PERCENTAGE_OF_VOLUMES_WITH_FILES')) / 100);
    my @files;

    return unless ($volume_count > 0);

    # Format is cluster->vserver->volume->desired_percentage
    #                                     aligned_files
    #                                     total_files
    my %volume_vm_aligned_hash = ();
    for (my $i = 0; $i < $self->param('TOTAL_FILE_COUNT'); $i++) {
        my $volume_index = $i%$volume_count;

        my $volume = $volumes[$volume_index];

        my $vm_align_suffix;

        try {
            $vm_align_suffix = $volume->get_attribute('vm-align-suffix');
        } otherwise {
            $vm_align_suffix = undef;
        };

        my $path;
        # If we have a suffix then we want to perform the suffix logic.
        if (defined $vm_align_suffix) {
            # Get the desired percent of aligned files for this volume.
            my $desired_percent_of_aligned = $volume_vm_aligned_hash{$volume->get_attribute('cluster_name')}
                                        {$volume->get_attribute('vserver')}
                                            {$volume->get_attribute('volume')}
                                                {'desired_percentage'};
            # If the desired percent of aligned files for the volume is undefined,
            # then this is the first file for the volume and we need to calculated the desired percent.
            unless (defined $desired_percent_of_aligned) {
                $desired_percent_of_aligned = $self->param('MINIMUM_PERCENT_VM_ALIGNED_FILES') +
                                                                     int(rand($self->param('MAXIMUM_PERCENT_VM_ALIGNED_FILES') -
                                                                              $self->param('MINIMUM_PERCENT_VM_ALIGNED_FILES')));
                $volume_vm_aligned_hash{$volume->get_attribute('cluster_name')}
                                        {$volume->get_attribute('vserver')}
                                            {$volume->get_attribute('volume')}
                                                {'desired_percentage'} = $desired_percent_of_aligned;
            }
            # Add to the total count for this volume.
            $volume_vm_aligned_hash{$volume->get_attribute('cluster_name')}
                                        {$volume->get_attribute('vserver')}
                                            {$volume->get_attribute('volume')}
                                                {'total_files'}++;
            my $total_count  = $volume_vm_aligned_hash{$volume->get_attribute('cluster_name')}
                                                            {$volume->get_attribute('vserver')}
                                                                {$volume->get_attribute('volume')}
                                                                    {'total_files'};
            my $aligned_count = $volume_vm_aligned_hash{$volume->get_attribute('cluster_name')}
                                                            {$volume->get_attribute('vserver')}
                                                                {$volume->get_attribute('volume')}
                                                                    {'aligned_files'} // 0;

            $path = '/vol/' . $volume->get_attribute("volume") . "/file$i";
            if ((($aligned_count / $total_count) * 100) < $desired_percent_of_aligned) {
                $volume_vm_aligned_hash{$volume->get_attribute('cluster_name')}{$volume->get_attribute('vserver')}{$volume->get_attribute('volume')}{'aligned_files'}++;
                $path .= $vm_align_suffix;
            }
        } else {
            $path = '/vol/' . $volume->get_attribute("volume") . "/file$i";
        }

        my $file_res = NACL::AppBuilder::Xgen::Resource::File->new('vserver' =>$volume->get_attribute('vserver'),
                                                                   'path'  => $path,
                                                                   'file-size'    =>$self->param('FILE_SIZE'),
                                                                   'xgen_metadata' => {retry_strings => ['Mounting unjunctioned Volume',
                                                                                                         'number of allowed concurrent mkfile threads exceeded']});

        # connect the vs resource to the aggr's cluster parent resource
        $file_res->add_parent_resources(resources => [$volume]);

        push(@files, $file_res);
    }

    $self->add_resources(resources => \@files);
}

###############################################################################
# DESCRIPTION: Generates lun resources and adds them to the configuration.
#              Generates a number of luns equal to the TOTAL_LUN_COUNT
#              constant.  The luns will only be generated on non DP type
#              volumes. The PERCENTAGE_OF_VOLUMES_WITH_LUNS controls the
#              percent of the non DP type volumes that get luns. The luns
#              ostype defaults to linux but will be changes if the lun is
#              mapped. The lun size is controlled by the LUN_SIZE constant.
# INPUT: NONE
# OUTPUT: NONE
# Exceptions: None thrown directly by this method.
###############################################################################
sub generate_lun_resources {
    my ($self, %args) = @_;

    # Get the clusters so that we can alternate vserver creation between multiple clusters.
    my $volume_resources = $self->get_resources(type => "NACL::AppBuilder::Xgen::Resource::Volume");

    # Filter the volumes resources by security style mixed.
    my $matches_hash = positive_filter(resources => $volume_resources,
                                       filter    => {'type' => 'DP'});

    my @volumes = @{$matches_hash->{'not_matches'}};

    my $volume_count = ceil((scalar(@volumes) * $self->param('PERCENTAGE_OF_VOLUMES_WITH_LUNS')) / 100);
    my @luns;

    return unless ($volume_count > 0);

    for (my $i = 0; $i < $self->param('TOTAL_LUN_COUNT'); $i++) {
        my $volume_index = $i%$volume_count;

        my $volume = $volumes[$volume_index];

        my $lun_res = NACL::AppBuilder::Xgen::Resource::Lun->new(vserver =>$volume->get_attribute('vserver'),
                                                                 volume  =>$volume->get_attribute("volume"),
                                                                 lun     => "lun$i",
                                                                 size    =>$self->param('LUN_SIZE'),
                                                                 ostype  => 'linux'); # Defaulting the ostype to linux. Will overwrite when maps created.

        # connect the vs resource to the aggr's cluster parent resource
        $lun_res->add_parent_resources(resources => [$volume]);

        push(@luns, $lun_res);
    }
    $self->add_resources(resources => \@luns);
}

###############################################################################
# DESCRIPTION: Generates lun map resources and adds them to the configuration.
#              Generates a lun map on every lun in the configuration.  Each lun
#              will be mapped to a single igroup.  In a vserver the lun
#              will be evenly distributed between igroups for maps. The luns
#              ostype will be changed to the ostype of the igroups client.
# INPUT: NONE
# OUTPUT: NONE
# Exceptions: None thrown directly by this method.
###############################################################################
sub generate_lun_map_resources {
    my ($self, %args) = @_;

    # Get the clusters so that we can alternate vserver creation between multiple clusters.
    my $vserver_resources = $self->get_resources(type => "NACL::AppBuilder::Xgen::Resource::Vserver");
    my @vservers = @{$vserver_resources};

    my @lun_maps;
    foreach my $vserver (@vservers) {
        my @luns = $vserver->get_descendants(type => "NACL::AppBuilder::Xgen::Resource::Lun");

        next unless (scalar @luns);

        my @igroups = $vserver->get_descendants(type => "NACL::AppBuilder::Xgen::Resource::Igroup");

        next unless (scalar @igroups);

        #(cl1,cl2)  vserver1:vol1:lun1  vserver2:vol1:lun1  might not work in this case so need better logic.
        foreach my $lun (@luns) {
            my $igroup = shift(@igroups);
            push(@igroups,$igroup);

            # Change the os_type of the lun to match the map.
            $lun->set_private_attributes(ostype => $igroup->get_attribute('client_os_type'));

            my $lunmap_res = NACL::AppBuilder::Xgen::Resource::LunMap->new(vserver => $lun->get_attribute('vserver'),
                                                                           lun => $lun->get_attribute('lun'),
                                                                           volume => $lun->get_attribute('volume'),
                                                                           igroup => $igroup->get_attribute('igroup'));

            # We want the vserver to be last.
            $lunmap_res->add_parent_resources(resources => [$igroup,$lun]);

            push(@lun_maps, $lunmap_res);
        }
    }
     $self->add_resources(resources => \@lun_maps);
}

###############################################################################
# DESCRIPTION: Generates igroup resources and adds them to the configuration.
#              Generates an igroup per ISCSI client on every vserver in the
#              configuration.  This method currently only supports ISCSI. The
#              igroups are named after the client they represent.
# INPUT: NONE
# OUTPUT: NONE
# Exceptions: None thrown directly by this method.
###############################################################################
sub generate_igroup_resources {
    my ($self, %args) = @_;

    # Get the clusters so that we can alternate vserver creation between multiple clusters.
    my $vserver_resources = $self->get_resources(type => "NACL::AppBuilder::Xgen::Resource::Vserver");
    my @vservers = @{$vserver_resources};

    # Get the clusters so that we can alternate vserver creation between multiple clusters.
    my $client_resources = $self->get_resources(type => "NACL::AppBuilder::Xgen::Resource::Client");
    my @clients = @{$client_resources};

    my @igroup;
    foreach my $vserver (@vservers) {
        foreach my $client (@clients) {
            # skip clients that are not doing iSCSI
            next if (!grep("iscsi", @{$client->get_attribute('protocols')}));

            my $igroup_res = NACL::AppBuilder::Xgen::Resource::Igroup->new(vserver => $vserver->get_attribute('vserver'),
                                                                           igroup => $client->get_attribute('name'),
                                                                           protocol => "iscsi",
                                                                           ostype => $client->get_attribute('ostype') =~ /windows/ ? "windows" : $client->get_attribute('ostype'),
                                                                           xgen_metadata => {client_os_type => $client->get_attribute('ostype')}); # used later for lun ostype.

            # We want the vserver to be last.
            $igroup_res->add_parent_resources(resources => [$client,$vserver]);

            push(@igroup, $igroup_res);
        }
    }
    $self->add_resources(resources => \@igroup);
}

###############################################################################
# DESCRIPTION: Generates igroup initiator resources and adds them to the
#              configuration. Generates an igroup initiator per igroup. Since
#              igroups are named after the client they represent we use the name
#              of the igroup for the client name.
# INPUT: NONE
# OUTPUT: NONE
# Exceptions: None thrown directly by this method.
###############################################################################
sub generate_igroup_initiator_resources {
    my ($self, %args) = @_;

    # Get the clusters so that we can alternate igroup creation between multiple clusters.
    my $igroup_resources = $self->get_resources(type => "NACL::AppBuilder::Xgen::Resource::Igroup");
    my @igroups = @{$igroup_resources};

    my @igroup_initiators;
    foreach my $igroup (@igroups) {
        my $initiator_res = NACL::AppBuilder::Xgen::Resource::Igroup::Initiator->new(vserver => $igroup->get_attribute('vserver'),
                                                                                     igroup => $igroup->get_attribute('igroup'),
                                                                                     protocol => "iscsi",
                                                                                     client_name => $igroup->get_attribute('igroup'));

        # We want the vserver to be last.
        $initiator_res->add_parent_resources(resources => [$igroup]);

        push(@igroup_initiators, $initiator_res);
    }

    $self->add_resources(resources => \@igroup_initiators);
}

###############################################################################
# DESCRIPTION: Generates ISCSI service resources and adds them to the
#              configuration. Generates an ISCSI service per vserver.
# INPUT: NONE
# OUTPUT: NONE
# Exceptions: None thrown directly by this method.
###############################################################################
sub generate_iscsi_services_resources {
    my ($self, %args) = @_;

    # Get the clusters so that we can alternate vserver creation between multiple clusters.
    my $vserver_resources = $self->get_resources(type => "NACL::AppBuilder::Xgen::Resource::Vserver");
    my @vservers = @{$vserver_resources};

    my @iscsi_services;

    foreach my $vserver (@vservers) {
        my $iscsi_res = NACL::AppBuilder::Xgen::Resource::Iscsi->new(vserver => $vserver->get_attribute('vserver'));

        # Add vserver as parent to iscsi service.
        $iscsi_res->add_parent_resources(resources => [$vserver]);

        push(@iscsi_services, $iscsi_res);
    }
    $self->add_resources(resources => \@iscsi_services);
}

###############################################################################
# DESCRIPTION: Generates cluster peer resources and adds them to the
#              configuration. Generates cluster peers in a cascade fasion.
#              C1->C2->C3. The order of the peers is based on the order off
#              clusters returned from the configuration and has no known
#              control for the order.
# INPUT: NONE
# OUTPUT: NONE
# Exceptions: None thrown directly by this method.
###############################################################################
sub generate_cluster_peer_resources {
    my ($self, %args) = @_;

    # Get the clusters so that we can alternate vserver creation between multiple clusters.
    my $cluster_resources = $self->get_resources(type => "NACL::AppBuilder::Xgen::Resource::Cluster");
    my @clusters = @{$cluster_resources};

    my @cluster_peers;

    return unless (scalar @clusters > 1);

    for (my $i = 1; $i < scalar @clusters; $i++) {
        my $source_cluster = $clusters[$i-1];
        my $destination_cluster = $clusters[$i];

        # May need to get a node from each cluster and use that as the name instead of the cluster_name
        # since nate_host files may not always exist for the clusters.
        my $cluster_peer_res = NACL::AppBuilder::Xgen::Resource::ClusterPeer->new(cluster => $source_cluster->get_attribute('cluster_name'),
                                                                                  rcluster => $destination_cluster->get_attribute('cluster_name'));

        # Add parents with source cluster last so that the command is run from the source cluster.
        $cluster_peer_res->add_parent_resources(resources => [$destination_cluster,$source_cluster]);

        push(@cluster_peers, $cluster_peer_res);
    }
    $self->add_resources(resources => \@cluster_peers);
}

###############################################################################
# DESCRIPTION: Generates vserver peer resources and adds them to the
#              configuration. Generates vserver peers based on snapmirrors in
#              the configuration. Gets all the snapmirrors from the
#              configurations and builds 1 vserver peer per pair of vservers
#              that need to be peered for snapmirrors to work.
# INPUT: NONE
# OUTPUT: NONE
# Exceptions: None thrown directly by this method.
###############################################################################
sub generate_vserver_peer_resources {
    my ($self, %args) = @_;

    # Get the clusters so that we can alternate vserver creation between multiple clusters.
    my $snapmirror_resources = $self->get_resources(type => "NACL::AppBuilder::Xgen::Resource::Snapmirror");
    my @snapmirrors = @{$snapmirror_resources};

    my @vserver_peers;

    return unless (scalar @snapmirrors > 0);

    my %vserver_peer_hash;
    for (my $i = 0; $i < scalar @snapmirrors; $i++) {
        my $source_vserver = $snapmirrors[$i]->get_attribute("source-vserver");
        my $destination_vserver = $snapmirrors[$i]->get_attribute("destination-vserver");
        my @vserver_resources = $snapmirrors[$i]->get_ancestors(type => "NACL::AppBuilder::Xgen::Resource::Vserver");
        if (defined $vserver_peer_hash{$source_vserver} and
            defined $vserver_peer_hash{$source_vserver}{$destination_vserver}) {
               # This will have already been added to the resource list.
               next;
        } elsif (defined $vserver_peer_hash{$destination_vserver} and
                 defined $vserver_peer_hash{$destination_vserver}{$source_vserver}) {
            # This will have already been added to the resource list.
            next;
        } else {
            # This combination was not found so we need to add this resource.
            $vserver_peer_hash{$source_vserver}{$destination_vserver} = 1;
            my %vserver_peer_args = (vserver => $source_vserver,
                                     'peer-vserver' => $destination_vserver);

            foreach my $vserver (@vserver_resources) {
                if ( $vserver->get_attribute('vserver') eq $source_vserver) {
                    $source_vserver = $vserver;
                } elsif ( $vserver->get_attribute('vserver') eq $destination_vserver) {
                    $destination_vserver = $vserver;
                }
            }

            if($source_vserver->get_attribute('cluster_name') ne $destination_vserver->get_attribute('cluster_name')) {
                $vserver_peer_args{'peer-cluster'} = $destination_vserver->get_attribute('cluster_name');
            }
            my $vserver_peer_res = NACL::AppBuilder::Xgen::Resource::VserverPeer->new(%vserver_peer_args);
            # Add parents with source cluster last so that the command is run from the source cluster.
            $vserver_peer_res->add_parent_resources(resources => [$destination_vserver,$source_vserver]);

            push(@vserver_peers, $vserver_peer_res);
        }
    }
    $self->add_resources(resources => \@vserver_peers);
}


###############################################################################
# DESCRIPTION: Generates vserver resources and adds them to the configuration.
#              Generates a number of vservers based on the TOTAL_VSERVER_COUNT
#              constant. The vservers will be named vs# where # starts at 0 and
#              counts up to TOTAL_VSERVER_COUNT-1. The vservers will alternate
#              between the clusters provided and inside each cluster will
#              alternate between the aggregates in that cluster that are part
#              of the configuration.
# INPUT: NONE
# OUTPUT: NONE
# Exceptions: None thrown directly by this method.
###############################################################################
sub generate_vserver_resources {
    my ($self, %args) = @_;

    # Get the clusters so that we can alternate vserver creation between multiple clusters.
    my $cluster_resources = $self->get_resources(type => "NACL::AppBuilder::Xgen::Resource::Cluster");
    my @clusters = @{$cluster_resources};
    my %cluster_aggregate_hash = ();
    foreach my $cluster_resource (@$cluster_resources) {
        @{$cluster_aggregate_hash{$cluster_resource->get_attribute('cluster_name')}} =
                $cluster_resource->get_descendants(type => "NACL::AppBuilder::Xgen::Resource::Aggregate");
        # {clus1 => [aggr1_obj,aggr2_obj], clus2 => [aggr1_obj,aggr2_obj]}
    }

    my @vservers;

    for (my $i = 0; $i < $self->param('TOTAL_VSERVER_COUNT'); $i++) {
        my $cluster = shift(@clusters);
        push (@clusters, $cluster);

        my $aggr = shift(@{$cluster_aggregate_hash{$cluster->get_attribute('cluster_name')}});
        push(@{$cluster_aggregate_hash{$cluster->get_attribute('cluster_name')}}, $aggr);

        my $vs_res = NACL::AppBuilder::Xgen::Resource::Vserver->new(vserver =>"vs$i",
                                                                    aggregate => $aggr->get_attribute("aggregate"));

        # connect the vs resource to the aggr's cluster parent resource
        $vs_res->add_parent_resources(resources => [$aggr]);

        push(@vservers, $vs_res);
    }
    $self->add_resources(resources => \@vservers);
}

###############################################################################
# DESCRIPTION: Generates volume resources and adds them to the configuration.
#              Generates a number of volumes based on the TOTAL_VOLUME_COUNT
#              constant. The volumes will be named Vol# where # starts at 0 and
#              counts up to TOTAL_VOLUME_COUNT-1. The volumes will alternate
#              between the clusters provided and inside each cluster will
#              alternate between the aggregates and vservers in that cluster
#              that are part of the configuration. The
#              VOLUME_VM_ALIGN_SUFFIX_PERCENT_LAYOUT will control the
#              percentage of the total volumes that have each of the
#              vm-align-suffix attributes defined in the layout. The
#              VOLUME_VM_ALIGN_SECTOR_PERCENT_LAYOUT will control the
#              percentage of the volumes with vm-align-sufix that have
#              each of the vm-align-sector attributes defined in the layout.
#              The volume size is controlled by the VOLUME_SIZE constant.
# INPUT: NONE
# OUTPUT: NONE
# Exceptions: None thrown directly by this method.
###############################################################################
sub generate_volume_resources {
    my ($self, %args) = @_;
  # Get the clusters so that we can alternate vserver creation between multiple clusters.
    my $cluster_resources = $self->get_resources(type => "NACL::AppBuilder::Xgen::Resource::Cluster");
    my @clusters = @{$cluster_resources};
    my %cluster_aggregate_hash = ();
    my %cluster_vserver_hash = ();
    foreach my $cluster_resource (@$cluster_resources) {
        @{$cluster_aggregate_hash{$cluster_resource->get_attribute('cluster_name')}} =
                $cluster_resource->get_descendants(type => "NACL::AppBuilder::Xgen::Resource::Aggregate");
        # {clus1 => [aggr1_obj,aggr2_obj], clus2 => [aggr1_obj,aggr2_obj]}
        @{$cluster_vserver_hash{$cluster_resource->get_attribute('cluster_name')}} =
                $cluster_resource->get_descendants(type => "NACL::AppBuilder::Xgen::Resource::Vserver");
        # {clus1 => [vs0_obj,vs2_obj], clus2 => [vs1_obj,vs3_obj]}
    }

    my @volumes;

    # Create a suffix layout based on the total number of volume and the percentage suffix layout constant.
    my $suffix_layout = convert_percent_layout_to_count_layout(layout        => $self->param('VOLUME_VM_ALIGN_SUFFIX_PERCENT_LAYOUT'),
                                                               total_objects => $self->param('TOTAL_VOLUME_COUNT'));
    # Get the number of none suffix volumes to use for the total for the sector layout.
    my $volumes_without_suffix_count = $suffix_layout->{':NONE:'} // 0;
    # Create a sector layout based on the total number of volumes that have a suffix and the percentage sector layout constant.
    my $sector_layout = convert_percent_layout_to_count_layout(layout        => $self->param('VOLUME_VM_ALIGN_SECTOR_PERCENT_LAYOUT'),
                                                               total_objects => $self->param('TOTAL_VOLUME_COUNT') - $volumes_without_suffix_count);

    for (my $i = 0; $i < $self->param('TOTAL_VOLUME_COUNT'); $i++) {
        my $cluster = shift(@clusters);
        push (@clusters, $cluster);

        my $aggr = shift(@{$cluster_aggregate_hash{$cluster->get_attribute('cluster_name')}});
        push(@{$cluster_aggregate_hash{$cluster->get_attribute('cluster_name')}}, $aggr);

        my $vserver = shift(@{$cluster_vserver_hash{$cluster->get_attribute('cluster_name')}});
        push(@{$cluster_vserver_hash{$cluster->get_attribute('cluster_name')}}, $vserver);

        my $vs_name = $vserver->get_attribute("vserver"),

        my $volume_name_format = $self->param('VOLUME_NAME_FORMAT') || 'vserver';
        my $volname;
        if ($volume_name_format eq 'vserver') {
            $volname = "vol_${i}_${vs_name}";
        } elsif ($volume_name_format eq 'volume_only') {
            $volname = "vol_$i";
        } elsif ($volume_name_format eq 'volumeonly') {
            $volname = "vol$i";
        } else {
            $volname = "vol_$i";
        }

        my %volume_arguments = (vserver   => $vs_name,
                                aggregate => $aggr->get_attribute("aggregate"),
                                volume    => $volname,
                                size      => $self->param('VOLUME_SIZE'));

        # Get the next suffix element to use from the suffix layout
        my $suffix = get_next_layout_element(iterator => $i,
                                             layout => $suffix_layout);

        # If there is a suffix then set the vm-align-suffix and get and set the vm-align-sector
        unless($suffix eq ':NONE:') {
            $volume_arguments{'vm-align-suffix'} = $suffix;
            $volume_arguments{'vm-align-sector'} = get_next_layout_element(iterator => $i,
                                                                           layout => $sector_layout);
        }

        my $volume_res = NACL::AppBuilder::Xgen::Resource::Volume->new(%volume_arguments);

        # connect the vs resource to the aggr's cluster parent resource
        $volume_res->add_parent_resources(resources => [$vserver,$aggr]);

        push(@volumes, $volume_res);
    }
    $self->add_resources(resources => \@volumes);
}

###############################################################################
# DESCRIPTION: Adds SMGR related data to SMGR type snapmirrors in the
#              configuration.  The SMGR_CG_ITEM_PERCENT_LAYOUT constant will
#              control the percentage of SMGR snapmirrors that get each type
#              defined in the layout. Current types known are file and lun.
#              Based on the type of cg items this method generates and add
#              source-path, destination-path, and cg-item-mappings to the
#              SMGR snapmirror resource and the SMGR snapmirror initialize
#              resource if it exists.
# INPUT: NONE
# OUTPUT: NONE
# Exceptions: None thrown directly by this method.
###############################################################################
sub add_cg_items_to_SMGR_snapmirror_resources {
    my ($self, %args) = @_;

    my $snapmirror_resources = $self->get_resources(type => "NACL::AppBuilder::Xgen::Resource::Snapmirror");

    # Filter the snapmirrors resources by snapmirror type of smgr.
    my $matches_hash = positive_filter(resources => $snapmirror_resources,
                                       filter    => {'sm_type' => 'smgr'});

    my @snapmirrors = @{$matches_hash->{'matches'}};

    # Create a SMGR layout based on the number of SMGR snapmirror required.
    my $smgr_layout = convert_percent_layout_to_count_layout(layout        => $self->param('SMGR_CG_ITEM_PERCENT_LAYOUT'),
                                                             total_objects => scalar(@snapmirrors));

    my $index = 0;
    foreach my $snapmirror (@snapmirrors) {
            # Get the next cg item type element to use from the smgr layout
        my $cg_item_type = get_next_layout_element(iterator => $index,
                                                   layout => $smgr_layout);
        $index++;
        my $cg_source_path = $snapmirror->get_attribute('source-vserver') . ':/cg/srccg_' . $snapmirror->get_attribute('source-volume');
        my $cg_destination_path = $snapmirror->get_attribute('destination-vserver') . ':/cg/dstcg_' . $snapmirror->get_attribute('destination-volume');

        if($cg_item_type =~ /file/i) {
            $cg_source_path .= '_file';
            $cg_destination_path .= '_file';
        }
        if($cg_item_type =~ /lun/i) {
            $cg_source_path .= '_lun';
            $cg_destination_path .= '_lun';
        }

        my @volumes = $snapmirror->get_ancestors(type => "NACL::AppBuilder::Xgen::Resource::Volume");
        my ($source_volume,$destination_volume);
        foreach my $volume (@volumes) {
            my $source = 0;
            my $destination = 0;
            if($volume->get_attribute('vserver') eq $snapmirror->get_attribute('source-vserver') and
               $volume->get_attribute('volume') eq $snapmirror->get_attribute('source-volume')) {
                $source_volume = $volume;
                next;
            } elsif($volume->get_attribute('vserver') eq $snapmirror->get_attribute('destination-vserver') and
               $volume->get_attribute('volume') eq $snapmirror->get_attribute('destination-volume')) {
               $destination_volume = $volume;
               next;
            } else {
                NATE::BaseException->throw("Found a child volume of an SMGR relationship when trying to add cg-item-mappings that is not a source or destination volume.")
            }
        }

        my %cg_item_mappings = ();

        if($cg_item_type =~ /file/i) {
            my @files = $source_volume->get_descendants(type => "NACL::AppBuilder::Xgen::Resource::File");
            # use only specified % of files for cg
            my $file_count = ceil((scalar(@files) * $self->param('PERCENTAGE_OF_LUNS_FILES_USED_FOR_CG_ITEMS')) / 100);
            my @files_for_cg = splice(@files, 0, $file_count);
            foreach my $file (@files_for_cg) {
                # Files were built using the path so we use the path.
                my $source_cg_item = $file->get_attribute('path');
                # The destination path will be the same as the source but we need to replace the source
                # volume name with the destination volume name.
                my $destination_cg_item = $source_cg_item;
                my $source_volume_name = $source_volume->get_attribute('volume');
                my $destination_volume_name = $destination_volume->get_attribute('volume');
                $destination_cg_item =~ s/$source_volume_name/$destination_volume_name/;
                $cg_item_mappings{$source_cg_item} = $destination_cg_item;
            }
        }
        if($cg_item_type =~ /lun/i) {
            my @luns = $source_volume->get_descendants(type => "NACL::AppBuilder::Xgen::Resource::Lun");
            # use only specified % of luns for cg
            my $lun_count = ceil((scalar(@luns) * $self->param('PERCENTAGE_OF_LUNS_FILES_USED_FOR_CG_ITEMS')) / 100);
            my @luns_for_cg = splice(@luns, 0, $lun_count);

            foreach my $lun (@luns_for_cg) {
                # Luns were created using the individual parameters so we need to buld the path.
                my $source_cg_item = '/vol/' . $lun->get_attribute('volume') . '/' . $lun->get_attribute('lun');
                my $destination_cg_item = '/vol/' . $destination_volume->get_attribute('volume') . '/' . $lun->get_attribute('lun');
                $cg_item_mappings{$source_cg_item} = $destination_cg_item;
            }
        }
        $snapmirror->set_private_attributes('cg-item-mappings' => \%cg_item_mappings,
                                            'source-path' => $cg_source_path,
                                            'destination-path' => $cg_destination_path);
        $snapmirror->set_metadata_attributes('cg_item_type' => $cg_item_type);
        my ($snapmirror_initialize) = $snapmirror->get_descendants(type => "NACL::AppBuilder::Xgen::Resource::SnapmirrorInitialize");
        if (defined $snapmirror_initialize) {
            $snapmirror_initialize->set_private_attributes('cg-item-mappings' => \%cg_item_mappings,
                                                'source-path' => $cg_source_path,
                                                'destination-path' => $cg_destination_path);
            $snapmirror_initialize->set_metadata_attributes('cg_item_type' => $cg_item_type);
        }
    }
}

###############################################################################
# DESCRIPTION: Generates Volume Efficiency Resources and Volume Efficiency
#              Policy resources in the configuration. This will generate
#              A volume efficiency resource on each volume that was
#              designated during snapmirror resource generation to need
#              volume efficiency. Uses the START_VOLUME_EFFICIENCY constant
#              to know whether to tell EnGenX to start volume efficiency.
#              Volume efficiency will never be started on destinations of
#              snapmirrors. The arguments for volume efficiency are based on a
#              template corresponding to the volumes volume_efficency_options
#              parameter.  If the Volume efficiency arguments contain a policy,
#              then a policy is created by calling
#              generate_volume_efficiency_policy_resource
# INPUT: NONE
# OUTPUT: NONE
# Exceptions: NATE::BaseException based on the volume_efficiency_options
#             template not existing.
###############################################################################
sub generate_volume_efficiency_resources {
    my ($self, %args) = @_;

    my $volume_resources = $self->get_resources(type => "NACL::AppBuilder::Xgen::Resource::Volume");

    # Filter the volumes resources by security style mixed.
    my $matches_hash = positive_filter(resources => $volume_resources,
                                       filter    => {'volume_efficiency_options' => undef});

    # We want the volumes that have volume efficiency options defined.
    my @volumes = @{$matches_hash->{'not_matches'}};

    return unless (scalar(@volumes) > 0);

    my $vserver_volume_efficiency_policy_hash = {};
    foreach my $volume (@volumes) {
        my $volume_efficiency_options_template = $volume->get_attribute('volume_efficiency_options');
        my $volume_efficiency_options_template_method = uc($volume_efficiency_options_template) . '_VOLUME_EFFICIENCY_OPTIONS';
        my $start;
        try {
            $volume->get_attribute('destination_for_sm_type');
            # If this is a destination then we do not want to start the service.
            $start = 0;
        } otherwise {
            $start = $self->param('START_VOLUME_EFFICIENCY');
        }; # Do Nothing.
        my $method = $self->can($volume_efficiency_options_template_method);
        if (defined $method) {
            #TODO - param check for these volume efficiency args
            my %volume_efficiency_args = %{$self->$method};
            my $volume_efficiency = NACL::AppBuilder::Xgen::Resource::Volume::Efficiency->new('vserver' => $volume->get_attribute('vserver'),
                                                                                              'volume' => $volume->get_attribute('volume'),
                                                                                              'start'  => $start,  # Put start here so it can be overridden by args
                                                                                              %volume_efficiency_args);
            $volume_efficiency->add_parent_resources(resources => [$volume]);
            $self->add_resources(resources => [$volume_efficiency]);

            my $volume_efficiency_policy = $volume_efficiency_args{'policy'};
            # Add the job_schedule to the destination cluster if it has not already been added.
            $self->generate_volume_efficiency_policy_resource('volume'                                => $volume,
                                                              'volume_efficiency_policy'              => $volume_efficiency_policy,
                                                              'vserver_volume_efficiency_policy_hash' => $vserver_volume_efficiency_policy_hash);
        } else {
            NATE::BaseException->throw("Tried to use a volume efficiency options template $volume_efficiency_options_template that this config does not understand.");
        }
    }
}

###############################################################################
# DESCRIPTION: Generates a single Volume Efficiency Policy resource
#              in the configuration. This will generate
#              a volume efficiency policy resource on the vserver of the
#              volume that is passed in. The arguments for volume efficiency
#              policy are based on a template corresponding to the
#              volume_efficiency_policy argument.
# INPUT: volume - the volume with volume efficiency options that require
#           this policy
#        volume_efficiency_policy - the name of the policy
#        vserver_volume_efficiency_policy_hash - a hash mapping policy name
#           to cluster and vserver. This is used to ensure no duplicate
#           policies.
# OUTPUT: NONE
# Exceptions: NATE::BaseException based on the volume_efficiency_policy
#             template not existing.
###############################################################################
sub generate_volume_efficiency_policy_resource {
    my ($self, %args) = @_;
    my $volume = $args{'volume'};
    my $policy_name = $args{'volume_efficiency_policy'};
    my $vserver_volume_efficiency_policy_hash = $args{'vserver_volume_efficiency_policy_hash'};

    if (not defined $policy_name or
        defined $vserver_volume_efficiency_policy_hash->{$volume->get_attribute('cluster_name')}
                                                      ->{$volume->get_attribute('vserver')}
                                                      ->{$policy_name}) {
        return; # Do nothing because the volume efficiency policy already exists.
    }

    $vserver_volume_efficiency_policy_hash->{$volume->get_attribute('cluster_name')}
                                          ->{$volume->get_attribute('vserver')}
                                          ->{$policy_name} = 1;
    my ($vserver) = $volume->get_ancestors(type => "NACL::AppBuilder::Xgen::Resource::Vserver");

    my $volume_efficiency_policy_method = uc($policy_name) . '_VOLUME_EFFICIENCY_POLICY';

    my $method = $self->can($volume_efficiency_policy_method);
    if (defined $method) {
        #TODO - param check for these volume efficiency args
        my %volume_efficiency_policy_args = %{$self->$method};
        my $volume_efficiency_policy = NACL::AppBuilder::Xgen::Resource::Volume::EfficiencyPolicy->new('vserver' => $vserver->get_attribute('vserver'),
                                                                                                       %volume_efficiency_policy_args);
        $volume_efficiency_policy->add_parent_resources(resources => [$vserver]);
        $self->add_resources(resources => [$volume_efficiency_policy]);
        return $volume_efficiency_policy;
    } else {
        NATE::BaseException->throw("Tried to use a volume efficiency policy $policy_name that this config does not understand.");
    }
}

###############################################################################
# DESCRIPTION: Generates a single Job Schedule Cron resource
#              in the configuration. This will generate
#              a Job Schedule Cron resource on the cluster of the
#              volume that is passed in. The arguments for job schedule
#              cron are based on a template corresponding to the
#              job_schedule_cron argument.
# INPUT: volume - the volume from a snapmirror that requires this schedule
#        job_schedule_cron - the name of the schedule
#        cluster_job_schedule_cron_hash - a hash mapping schedule name
#           to cluster. This is used to ensure no duplicate schedules.
# OUTPUT: NONE
# Exceptions: NATE::BaseException based on the job_schedule_cron
#             template not existing.
###############################################################################
sub generate_job_schedule_cron_resource {
    my ($self, %args) = @_;
    my $volume = $args{'volume'};
    my $schedule_name = $args{'job_schedule_cron'};
    my $cluster_job_schedule_cron_hash = $args{'cluster_job_schedule_cron_hash'};

    if (not defined $schedule_name or
        defined $cluster_job_schedule_cron_hash->{$volume->get_attribute('cluster_name')}
                                               ->{$schedule_name}) {
        return; # Do nothing because the schedule already exists.
    }

    $cluster_job_schedule_cron_hash->{$volume->get_attribute('cluster_name')}
                                   ->{$schedule_name} = 1;
    my ($cluster) = $volume->get_ancestors(type => "NACL::AppBuilder::Xgen::Resource::Cluster");

    my $schedule_method = uc($schedule_name) . '_JOB_SCHEDULE';

    my $method = $self->can($schedule_method);
    if (defined $method) {
        #TODO - param check for these JOB SCHEDULE args
        my %job_schedule_args = %{$self->$method};
        my $job_schedule = NACL::AppBuilder::Xgen::Resource::JobScheduleCron->new(%job_schedule_args);
        $job_schedule->add_parent_resources(resources => [$cluster]);
        $self->add_resources(resources => [$job_schedule]);
    } else {
        $Log->comment("schedule $schedule_name has no defined schedule method so we are assuming it is a default schedule and not making a schedule with that name.");
    }
}

###############################################################################
# DESCRIPTION: Generates a single Snapshot Policy resource
#              in the configuration. This will generate
#              a Snapshot Policy resource on the cluster of the
#              volume that is passed in. The arguments for job policy
#              cron are based on a template corresponding to the
#              snapshot_policy argument.
# INPUT: volume - the volume that requires this policy
#        snapshot_policy - the name of the policy
#        cluster_snapshot_policy_hash - a hash mapping policy name
#           to vserver and cluster. This is used to ensure no duplicate
#           policies.
# OUTPUT: NONE
# Exceptions: NATE::BaseException based on the snapshot_policy
#             template not existing.
###############################################################################
sub generate_snapshot_policy_resource {
    my ($self, %args) = @_;
    my $volume = $args{'volume'};
    my $policy_name = $args{'snapshot_policy'};
    my $vserver_snapshot_policy_hash = $args{'vserver_snapshot_policy_hash'};

    if (not defined $policy_name or
        defined $vserver_snapshot_policy_hash->{$volume->get_attribute('cluster_name')}
                                             ->{$volume->get_attribute('vserver')}
                                             ->{$policy_name}) {
        return; # Do nothing because the snapshot policy already exists.
    }

    $vserver_snapshot_policy_hash->{$volume->get_attribute('cluster_name')}
                                 ->{$volume->get_attribute('vserver')}
                                 ->{$policy_name} = 1;
    my ($vserver) = $volume->get_ancestors(type => "NACL::AppBuilder::Xgen::Resource::Vserver");

    my $policy_method = uc($policy_name) . '_SNAPSHOT_POLICY';

    my $method = $self->can($policy_method);
    if (defined $method) {
        #TODO - param check for these SNAPSHOT POLICY args
        my %snapshot_policy_args = %{$self->$method};
        my $snapshot_policy = NACL::AppBuilder::Xgen::Resource::SnapshotPolicy->new('vserver' => $vserver->get_attribute('vserver'),
                                                                                    'xgen_nacl_cs_validation' => 0,
                                                                                    %snapshot_policy_args);
        $snapshot_policy->add_parent_resources(resources => [$vserver]);
        $self->add_resources(resources => [$snapshot_policy]);
    } else {
        $Log->comment("Snapshot Policy $policy_name has no defined policy method so we are assuming it is a default policy and not making a policy with that name.");
    }
}

###############################################################################
# DESCRIPTION: Generates a single snapmirror Policy resource
#              in the configuration. This will generate
#              a snapmirror Policy resource on the cluster of the
#              volume that is passed in. The arguments for job policy
#              cron are based on a template corresponding to the
#              snapmirror_policy argument. Also generates Snapmirror Policy
#              Rule resources in the configuration based on rules defined in the
#              snapmirror policy arguments.
# INPUT: volume - the volume from a snapmirror that requires this policy
#        snapmirror_policy - the name of the policy
#        vserver_sm_policy_hash - a hash mapping policy name
#           to vserver and cluster. This is used to ensure no duplicate
#           policies.
# OUTPUT: NONE
# Exceptions: NATE::BaseException based on the snapmirror_policy
#             template not existing.
###############################################################################
sub generate_snapmirror_policy_resource {
    my ($self, %args) = @_;
    my $volume = $args{'volume'};
    my $policy_name = $args{'snapmirror_policy'};
    my $vserver_sm_policy_hash = $args{'vserver_sm_policy_hash'};
    if (not defined $policy_name or
        defined $vserver_sm_policy_hash->{$volume->get_attribute('cluster_name')}
                                             ->{$volume->get_attribute('vserver')}
                                             ->{$policy_name}) {
        return; # Do nothing because the snapmirror policy already exists.
    }

    $vserver_sm_policy_hash->{$volume->get_attribute('cluster_name')}
                                 ->{$volume->get_attribute('vserver')}
                                 ->{$policy_name} = 1;
    my ($vserver) = $volume->get_ancestors(type => "NACL::AppBuilder::Xgen::Resource::Vserver");

    my $policy_method = uc($policy_name) . '_SNAPMIRROR_POLICY';

    my $method = $self->can($policy_method);
    if (defined $method) {
        #TODO - param check for these SNAPSHOT POLICY args
        my %snapmirror_policy_args = %{$self->$method(%args)};
        my $snapmirror_policy_rules = delete $snapmirror_policy_args{'xgen_snapmirror_policy_rules'};
        my $snapmirror_policy = NACL::AppBuilder::Xgen::Resource::SnapmirrorPolicy->new('vserver' => $vserver->get_attribute('vserver'),
                                                                                        %snapmirror_policy_args);
        $snapmirror_policy->add_parent_resources(resources => [$vserver]);
        $self->add_resources(resources => [$snapmirror_policy]);

        foreach my $snapmirror_policy_rule_args (@{$snapmirror_policy_rules}) {
            my $snapmirror_policy_rule = NACL::AppBuilder::Xgen::Resource::SnapmirrorPolicyRule->new('vserver' => $vserver->get_attribute('vserver'),
                                                                                                     'policy'  => $policy_name,
                                                                                                     %{$snapmirror_policy_rule_args});
            $snapmirror_policy_rule->add_parent_resources(resources => [$snapmirror_policy]);
            $self->add_resources(resources => [$snapmirror_policy_rule]);
        }
    } else {
        $Log->comment("Snapmirror Policy $policy_name has no defined policy method so we are assuming it is a default policy and not making a policy with that name.");
    }
}

###############################################################################
# DESCRIPTION: Generates snapmirror resources
#              in the configuration. This method calls all the config
#              generation methods to generate snapmirrors in the configuration.
#              This also creates and holds the references to global data used
#              by some other methods to ensure no duplicates.
# INPUT: NONE
# OUTPUT: NONE
# Exceptions: None thrown directly by this method.
###############################################################################
sub generate_snapmirror_resources {
    my ($self, %args) = @_;

    my $volume_resources = $self->get_resources(type => "NACL::AppBuilder::Xgen::Resource::Volume");
    my @temp_volumes = @{$volume_resources};
    my $available_volumes = \@temp_volumes;
    # A mapping of vservers to clusters.
    my $cluster_vserver_hash = {};
    # A mapping of vservers to snapmirror policies.
    my $vserver_sm_policy_hash = {};
    # A mapping of vservers to snapshot policies.
    my $vserver_snapshot_policy_hash = {};
    # A mapping of clusters to job schedules.
    my $cluster_job_schedule_cron_hash = {};

    my $config_id = 0;
    my $created_snapmirrors = 0;

    $Log->comment("Generating FANIN Snapmirror configs.");
    # Create all the cascade snapmirror configs that were defined.
    $self->generate_fanin_snapmirror_configs(available_volumes                 => $available_volumes,
                                             created_snapmirrors               => \$created_snapmirrors,
                                             'vserver_sm_policy_hash'          => $vserver_sm_policy_hash,
                                             'vserver_snapshot_policy_hash'    => $vserver_snapshot_policy_hash,
                                             'cluster_job_schedule_cron_hash'  => $cluster_job_schedule_cron_hash,
                                             config_id                         => \$config_id);

    $Log->comment("Generating FANOUT Snapmirror configs.");
    # Create all the cascade snapmirror configs that were defined.
    $self->generate_fanout_snapmirror_configs(available_volumes                 => $available_volumes,
                                              created_snapmirrors               => \$created_snapmirrors,
                                              'vserver_sm_policy_hash'          => $vserver_sm_policy_hash,
                                              'vserver_snapshot_policy_hash'    => $vserver_snapshot_policy_hash,
                                              'cluster_job_schedule_cron_hash'  => $cluster_job_schedule_cron_hash,
                                              config_id                         => \$config_id);

    $Log->comment("Generating BIDIRECTIONAL Snapmirror configs.");
    # Create all the cascade snapmirror configs that were defined.
    $self->generate_bidirectional_snapmirror_configs(available_volumes                 => $available_volumes,
                                                     created_snapmirrors               => \$created_snapmirrors,
                                                     'vserver_sm_policy_hash'          => $vserver_sm_policy_hash,
                                                     'vserver_snapshot_policy_hash'    => $vserver_snapshot_policy_hash,
                                                     'cluster_job_schedule_cron_hash'  => $cluster_job_schedule_cron_hash,
                                                     config_id                         => \$config_id);

    $Log->comment("Generating CASCADE Snapmirror configs.");
    # Create all the cascade snapmirror configs that were defined.
    $self->generate_cascade_snapmirror_configs(available_volumes                 => $available_volumes,
                                               created_snapmirrors               => \$created_snapmirrors,
                                               'vserver_sm_policy_hash'          => $vserver_sm_policy_hash,
                                               'vserver_snapshot_policy_hash'    => $vserver_snapshot_policy_hash,
                                               'cluster_job_schedule_cron_hash'  => $cluster_job_schedule_cron_hash,
                                               config_id                         => \$config_id);

    $Log->comment("Generating UNIDIRECTIONAL Snapmirror config.");
    # Create all the unidirectional snapmirrors that were not taken by other configs.
    # Unidirectional does not mean all the snapmirrors will be c1-c2 as some can be c1-c1 or a mix.
    # It just means that we do not specifically choose the direction except that we favor cluster to cluster snapmirrors.
    $self->generate_unidirectional_snapmirror_config(snapmirror_count                 => $self->param('TOTAL_SNAPMIRROR_COUNT') - $created_snapmirrors,,
                                              available_volumes                 => $available_volumes,
                                              created_snapmirrors               => \$created_snapmirrors,
                                              'vserver_sm_policy_hash'          => $vserver_sm_policy_hash,
                                              'vserver_snapshot_policy_hash'    => $vserver_snapshot_policy_hash,
                                              'cluster_job_schedule_cron_hash'  => $cluster_job_schedule_cron_hash,
                                              config_id                         => \$config_id);



    # Some snapmirror src/dst volumes may have incorrect names if there are
    # multiple snapmirrors related with a particular volume (src/dst).
    # After snapmirrors are created, go back and make sure they have the correct volume names
    #
    $Log->comment("Updating src/dst Snapmirror volume names.");

    my @volumes = @{$volume_resources};

    foreach my $vol (@volumes) {
        #$Log->comment('volume:' . $vol->get_attribute('volume'));
        my $volname = $vol->get_attribute('volume');

        my @sm = $vol->get_descendants(type => "NACL::AppBuilder::Xgen::Resource::Snapmirror");
        foreach my $sm (@sm) {
            #$Log->comment(' yamo sm:'  . dumper($sm));
            my $source_vol = $sm->get_attribute('source-volume');
            my $dest_vol   = $sm->get_attribute('destination-volume');
            #$Log->comment(' srcvol:' . $source_vol);
            #$Log->comment(' dstvol:' . $dest_vol);

            foreach my $smvol ($source_vol, $dest_vol) {
                if ($volname =~ /$smvol/) {
                    if ($smvol ne $volname) {
                        $Log->comment("updating snapmirror volume name $smvol to $volname");
                        $smvol = $volname;
                    }
                }
            }

            #$sm->set_private_attributes('source-volume' => $source_vol);
            #$sm->set_private_attributes('destination-volume' => $dest_vol);
            $sm->set_metadata_attributes('source-volume' => $source_vol);
            $sm->set_metadata_attributes('destination-volume' => $dest_vol);
        }


    }

    #need to do it for snapmirror initialize also
    $Log->comment("Updating src/dst Snapmirror Initialize volume names.");

    my @volumes = @{$volume_resources};

    foreach my $vol (@volumes) {
        #$Log->comment('volume:' . $vol->get_attribute('volume'));
        my $volname = $vol->get_attribute('volume');

        my @sm = $vol->get_descendants(type => "NACL::AppBuilder::Xgen::Resource::SnapmirrorInitialize");
        foreach my $sm (@sm) {
            #$Log->comment(' yamo sm init:'  . dumper($sm));
            my $source_vol = $sm->get_attribute('source-volume');
            my $dest_vol   = $sm->get_attribute('destination-volume');
            #$Log->comment(' srcvol:' . $source_vol);
            #$Log->comment(' dstvol:' . $dest_vol);

            foreach my $smvol ($source_vol, $dest_vol) {
                if ($volname =~ /$smvol/) {
                    if ($smvol ne $volname) {
                        $Log->comment("updating snapmirror initialize volume name $smvol to $volname");
                        $smvol = $volname;
                    }
                }
            }

            $sm->set_metadata_attributes('source-volume' => $source_vol);
            $sm->set_metadata_attributes('destination-volume' => $dest_vol);
            #$sm->set_private_attributes('source-volume' => $source_vol);
            #$sm->set_private_attributes('destination-volume' => $dest_vol);
        }


    }


}

###############################################################################
# DESCRIPTION: Generates snapmirror resources in the configuration. This method
#              generates snapmirrors in the configuration based on the rules
#              for a bidirectional config.  Bidirectional in this method means
#              two snapmirrors that reverse the source and destination vservers
#              VS1->VS2 and VS2->VS1.
#              It also prefers vservers on different clusters.
#              Creates a number of configs based on the
#              TOTAL_BIDIRECTIONAL_CONFIG_COUNT constant and the percentage of
#              configs with each type of layout is based on the
#              BIDIRECTIONAL_TYPE_PERCENT_LAYOUT constant.
# INPUT: snapmirror_count - number of snapmirrors left to be created
#        available_volumes - reference to an array of available volumes.
#        created_snapmirrors - a scalar reference to the number of already
#           created snapmirrors.
#        config_id - The next config id to use for snapmirror configs.
#        vserver_sm_policy_hash - a hashref to a mapping of snapmirror
#           policy name to vserver and cluster.
#        vserver_snapshot_policy_hash - a hashref to a mapping of snapshot
#           policy name to vserver and cluster.
#        cluster_job_schedule_cron_hash - a hashref to a mapping of schedules
#           to cluster.
# OUTPUT: NONE
# Exceptions: None thrown directly by this method.
###############################################################################
sub generate_bidirectional_snapmirror_configs {
    my ($self, %args) = @_;
    my $snapmirror_count = $args{'snapmirror_count'};
    my $available_volumes = $args{'available_volumes'};
    my $created_snapmirrors = $args{'created_snapmirrors'};
    my $config_id           = $args{'config_id'};
    # A mapping of vservers to snapmirror policies.
    my $vserver_sm_policy_hash = $args{'vserver_sm_policy_hash'};
    # A mapping of vservers to snapshot policies.
    my $vserver_snapshot_policy_hash = $args{'vserver_snapshot_policy_hash'};
    # A mapping of clusters to job schedules.
    my $cluster_job_schedule_cron_hash = $args{'cluster_job_schedule_cron_hash'};

    # This will get the layout in terms of numbers instead of percents based on the total object count passed in.
    my $layout = convert_percent_layout_to_count_layout('layout'        => $self->param('BIDIRECTIONAL_TYPE_PERCENT_LAYOUT'),
                                                        'total_objects' => $self->param('TOTAL_BIDIRECTIONAL_CONFIG_COUNT'));
    # For the destination we start with no must use clusters.
    my @destination_must_use_clusters = ();
    # For the destination we start with no must use vservers.
    my @destination_must_use_vservers = ();
    # Loop through TOTAL_CASCADE_CONFIG_COUNT and create those configs.
    for (my $i = 0; $i < $self->param('TOTAL_BIDIRECTIONAL_CONFIG_COUNT'); $i++) {
        my $config_layout = get_next_layout_element('iterator' => $i,       # This is not required but allows the get_next_layout_element to
                                                                            # rotate through the layout instead of exhausting the first, then
                                                                            # the second, etc.
                                                    'layout'   => $layout); # layout is the numbers layout we got earlier.

        # Config the config layout into an array of the types of snapmirrors in this cascade config.
        my @config_type_array = split(';',$config_layout);

        my %last_resort_vservers_hash = ();
        my %last_resort_clusters_hash = ();


        # For the first destination volume it can come from anywhere.
        my @first_vservers_volumes = $self->get_available_snapmirror_volume('available_volumes'   => $available_volumes,
                                                                            'all_from_same_vserver' => 1,
                                                                            'volume_count'          => 2);

        my $first_vservers_cluster = $first_vservers_volumes[0]->get_attribute('cluster_name');
        my $first_vserver = $first_vservers_volumes[0]->get_attribute('vserver');

        # For the first destination volume it can come from anywhere.
        my @second_vservers_volumes = $self->get_available_snapmirror_volume('available_volumes'   => $available_volumes,
                                                                             'all_from_same_vserver' => 1,
                                                                             'volume_count'          => 2,
                                                                             'unuseable_vservers'    => [$first_vserver],
                                                                             'last_resort_clusters'  => [$first_vservers_cluster]);

        #Loop through and set the metadata on all the volumes.
        for (my $j = 0; $j < 2; $j++) {

            my $first_vservers_volume  = $first_vservers_volumes[$j];

            my $second_vservers_volume = $second_vservers_volumes[$j];

            # Update some metadata on the volume.
            $first_vservers_volume->set_metadata_attributes('sm_config'        => 'bidirectional',
                                                            'sm_config_layout' => $config_layout,
                                                            'sm_config_id'     => $$config_id,);

            # Update some metadata on the volume.
            $second_vservers_volume->set_metadata_attributes('sm_config'     => 'bidirectional',
                                                             'sm_config_layout' => $config_layout,
                                                             'sm_config_id'  => $$config_id,);
        }

        # Now we make our bidirectional config.
        # v1->v2 and v2->v1 based on the config type.
        # The SM Type is the config type arrays current index
        my $SM_TYPE = $config_type_array[0];
        $self->generate_snapmirror_resource('source_volume'                  => $first_vservers_volumes[0],
                                            'destination_volume'             => $second_vservers_volumes[0],
                                            'sm_type'                        => $SM_TYPE,
                                            'vserver_sm_policy_hash'         => $vserver_sm_policy_hash,
                                            'vserver_snapshot_policy_hash'   => $vserver_snapshot_policy_hash,
                                            'cluster_job_schedule_cron_hash' => $cluster_job_schedule_cron_hash,,
                                            'xgen_metadata_args'             => {'sm_config'      => 'bidirectional',
                                                                                 'sm_config_layout' => $config_layout,
                                                                                 'sm_config_id' => $$config_id});
        $$created_snapmirrors++;

        $SM_TYPE = $config_type_array[1];
        $self->generate_snapmirror_resource('source_volume'                  => $second_vservers_volumes[1],
                                            'destination_volume'             => $first_vservers_volumes[1],
                                            'sm_type'                        => $SM_TYPE,
                                            'vserver_sm_policy_hash'         => $vserver_sm_policy_hash,
                                            'vserver_snapshot_policy_hash'   => $vserver_snapshot_policy_hash,
                                            'cluster_job_schedule_cron_hash' => $cluster_job_schedule_cron_hash,,
                                            'xgen_metadata_args'             => {'sm_config'      => 'bidirectional',
                                                                                 'sm_config_layout' => $config_layout,
                                                                                 'sm_config_id' => $$config_id});
        $$created_snapmirrors++;
        # We finished the sm config so we need to update the config id
        $$config_id++;
    }
}

###############################################################################
# DESCRIPTION: Generates snapmirror resources in the configuration. This method
#              generates snapmirrors in the configuration based on the rules
#              for a fanin config.  Fanin in this method means
#              multiple snapmirrors that go to the same destination vserver
#              VS2->VS1 VS3->VS1.
#              It also prefers vservers on different clusters.
#              Creates a number of configs based on the
#              TOTAL_FANIN_CONFIG_COUNT constant and the percentage of
#              configs with each type of layout is based on the
#              FANIN_TYPE_PERCENT_LAYOUT constant.
# INPUT: snapmirror_count - number of snapmirrors left to be created
#        available_volumes - reference to an array of available volumes.
#        created_snapmirrors - a scalar reference to the number of already
#           created snapmirrors.
#        config_id - The next config id to use for snapmirror configs.
#        vserver_sm_policy_hash - a hashref to a mapping of snapmirror
#           policy name to vserver and cluster.
#        vserver_snapshot_policy_hash - a hashref to a mapping of snapshot
#           policy name to vserver and cluster.
#        cluster_job_schedule_cron_hash - a hashref to a mapping of schedules
#           to cluster.
# OUTPUT: NONE
# Exceptions: None thrown directly by this method.
###############################################################################
sub generate_fanin_snapmirror_configs {
    my ($self, %args) = @_;
    my $snapmirror_count = $args{'snapmirror_count'};
    my $available_volumes = $args{'available_volumes'};
    my $created_snapmirrors = $args{'created_snapmirrors'};
    my $config_id           = $args{'config_id'};
    # A mapping of vservers to snapmirror policies.
    my $vserver_sm_policy_hash = $args{'vserver_sm_policy_hash'};
    # A mapping of vservers to snapshot policies.
    my $vserver_snapshot_policy_hash = $args{'vserver_snapshot_policy_hash'};
    # A mapping of clusters to job schedules.
    my $cluster_job_schedule_cron_hash = $args{'cluster_job_schedule_cron_hash'};

    # This will get the layout in terms of numbers instead of percents based on the total object count passed in.
    my $layout = convert_percent_layout_to_count_layout('layout'        => $self->param('FANIN_TYPE_PERCENT_LAYOUT'),
                                                        'total_objects' => $self->param('TOTAL_FANIN_CONFIG_COUNT'));
    # For the destination we start with no must use clusters.
    my @destination_must_use_clusters = ();
    # For the destination we start with no must use vservers.
    my @destination_must_use_vservers = ();
    # Loop through TOTAL_CASCADE_CONFIG_COUNT and create those configs.
    for (my $i = 0; $i < $self->param('TOTAL_FANIN_CONFIG_COUNT'); $i++) {
        my $config_layout = get_next_layout_element('iterator' => $i,     # This is not required but allows the get_next_layout_element to
                                                                          # rotate through the layout instead of exhausting the first, then
                                                                          # the second, etc.
                                                    'layout'   => $layout); # layout is the numbers layout we got earlier.

        # Config the config layout into an array of the types of snapmirrors in this cascade config.
        my @config_type_array = split(';',$config_layout);

        my %last_resort_vservers_hash = ();
        my %last_resort_clusters_hash = ();


        # For the first destination volume it can come from anywhere.
        my @destination_volumes = $self->get_available_snapmirror_volume('available_volumes'   => $available_volumes,
                                                                         'all_from_same_vserver' => 1,
                                                                         'volume_count'          => scalar (@config_type_array));

        my $destination_cluster = $destination_volumes[0]->get_attribute('cluster_name');
        my $destination_vserver = $destination_volumes[0]->get_attribute('vserver');

        my @source_volumes = ();
        # Now we need to find our source_volumes.
        # We will make the destination cluster always a last_resort
        # and also try to rotate through any remaining clusters.
        for (my $j = 0; $j < scalar (@config_type_array); $j++) {
            my %clusters_hash = ();
            foreach my $volume (@{$available_volumes}) {
                $clusters_hash{$volume->get_attribute('cluster_name')}++;
            }
            my @clusters = keys (%clusters_hash);
            # If there are clusters in the
            my %temp_hash = ();
            foreach my $cluster (@clusters) {
                if(defined $last_resort_clusters_hash{$cluster}) {
                    $temp_hash{$cluster} = $last_resort_clusters_hash{$cluster};
                }
            }
            %last_resort_clusters_hash = %temp_hash;
            my @last_resort_clusters = ();
            # Build the last_resort_clusters based on the home many times each has been used.
            my $times_used = 0;
            foreach my $cluster (keys %last_resort_clusters_hash) {
                if ($last_resort_clusters_hash{$cluster} == $times_used) {
                    # If this is used the same number of times then add it to the list
                    push(@last_resort_clusters,$cluster);
                } elsif ($last_resort_clusters_hash{$cluster} > $times_used) {
                    # If this is used more number of times then make it the list and update
                    # the times used. This way we can have the last resorts be the highest used
                    # clusters.
                    @last_resort_clusters = ($cluster);
                    $times_used = $last_resort_clusters_hash{$cluster};
                } # Else we ignore since it is not the most used.
            }

            # make sure to always have the destination cluster in the last resort list.
            push(@last_resort_clusters,$destination_cluster);

            # if the last resort clusters is >= the available clusters that means that all clusters
            # are in the last resort so we can assume that we only want to have the destination cluster
            # as the last resort.
            if(scalar (@last_resort_clusters) >= scalar (@clusters)) {
                @last_resort_clusters = ($destination_cluster);
            }
            my @last_resort_vservers = ();
            # Build the last_resort_clusters based on the home many times each has been used.
            $times_used = 0;
            foreach my $vserver (keys %last_resort_vservers_hash) {
                if ($last_resort_vservers_hash{$vserver} == $times_used) {
                    # If this is used the same number of times then add it to the list
                    push(@last_resort_vservers,$vserver);
                } elsif ($last_resort_vservers_hash{$vserver} > $times_used) {
                    # If this is used more number of times then make it the list and update
                    # the times used. This way we can have the last resorts be the highest used
                    # vservers.
                    @last_resort_vservers = ($vserver);
                    $times_used = $last_resort_vservers_hash{$vserver};
                } # Else we ignore since it is not the most used.
            }

            # get a volume based on the criteria from the previous volumes.
            # If last_resort and unuseable have the same vservers unuseable takes precendence.
            my ($volume) = $self->get_available_snapmirror_volume('available_volumes'  => $available_volumes,
                                                                  'last_resort_clusters' => \@last_resort_clusters,
                                                                  'last_resort_vservers' => \@last_resort_vservers,
                                                                  'unuseable_vservers'   => [$destination_vserver]);
            # push the volume into our list of volumes for the chain.
            push(@source_volumes,$volume);

            # Now we need to update the criteria based on the volume found.
            $last_resort_vservers_hash{$volume->get_attribute('vserver')}++;
            $last_resort_clusters_hash{$volume->get_attribute('cluster_name')}++;
        }

        for (my $j = 0; $j < scalar (@config_type_array); $j++) {
            # The source volume is the volume at the current index.
            my $source_volume      = $source_volumes[$j];
            # The destination volume is the volume at the current index plus 1.
            my $destination_volume = $destination_volumes[$j];

            # Update some metadata on the volume.
            $source_volume->set_metadata_attributes('sm_config'        => 'fanin',
                                                    'sm_config_layout' => $config_layout,
                                                    'fanin_index'      => $j,
                                                    'sm_config_id'     => $$config_id,);

            # Update some metadata on the volume.
            $destination_volume->set_metadata_attributes('sm_config'     => 'fanin',
                                                         'sm_config_layout' => $config_layout,
                                                         'fanin_index' => $j,
                                                         'sm_config_id'  => $$config_id,);

            # The SM Type is the config type arrays current index
            my $SM_TYPE = $config_type_array[$j];
            $self->generate_snapmirror_resource('source_volume'                  => $source_volume,
                                                'destination_volume'             => $destination_volume,
                                                'sm_type'                        => $SM_TYPE,
                                                'vserver_sm_policy_hash'         => $vserver_sm_policy_hash,
                                                'vserver_snapshot_policy_hash'   => $vserver_snapshot_policy_hash,
                                                'cluster_job_schedule_cron_hash' => $cluster_job_schedule_cron_hash,,
                                                'xgen_metadata_args'             => {'sm_config'      => 'fanin',
                                                                                     'sm_config_layout' => $config_layout,
                                                                                     'fanin_index'  => $j,
                                                                                     'sm_config_id' => $$config_id});
            $$created_snapmirrors++;
        }
        # We finished the sm config so we need to update the config id
        $$config_id++;
    }
}

###############################################################################
# DESCRIPTION: Generates snapmirror resources in the configuration. This method
#              generates snapmirrors in the configuration based on the rules
#              for a fanout config.  Fanout in this method means
#              multiple snapmirrors that also have the same source volume.
#              VS1:Vol1->VS2:Vol2 and VS1:Vol1->VS3:Vol3
#              It also prefers volumes on different clusters, vserver, and
#              aggregates.
#              Creates a number of configs based on the
#              TOTAL_FANOUT_CONFIG_COUNT constant and the percentage of
#              configs with each type of layout is based on the
#              FANOUT_TYPE_PERCENT_LAYOUT constant.
# INPUT: snapmirror_count - number of snapmirrors left to be created
#        available_volumes - reference to an array of available volumes.
#        created_snapmirrors - a scalar reference to the number of already
#           created snapmirrors.
#        config_id - The next config id to use for snapmirror configs.
#        vserver_sm_policy_hash - a hashref to a mapping of snapmirror
#           policy name to vserver and cluster.
#        vserver_snapshot_policy_hash - a hashref to a mapping of snapshot
#           policy name to vserver and cluster.
#        cluster_job_schedule_cron_hash - a hashref to a mapping of schedules
#           to cluster.
# OUTPUT: NONE
# Exceptions: None thrown directly by this method.
###############################################################################
sub generate_fanout_snapmirror_configs {
    my ($self, %args) = @_;
    my $snapmirror_count = $args{'snapmirror_count'};
    my $available_volumes = $args{'available_volumes'};
    my $created_snapmirrors = $args{'created_snapmirrors'};
    my $config_id           = $args{'config_id'};
    # A mapping of vservers to snapmirror policies.
    my $vserver_sm_policy_hash = $args{'vserver_sm_policy_hash'};
    # A mapping of vservers to snapshot policies.
    my $vserver_snapshot_policy_hash = $args{'vserver_snapshot_policy_hash'};
    # A mapping of clusters to job schedules.
    my $cluster_job_schedule_cron_hash = $args{'cluster_job_schedule_cron_hash'};

    # This will get the layout in terms of numbers instead of percents based on the total object count passed in.
    my $layout = convert_percent_layout_to_count_layout('layout'        => $self->param('FANOUT_TYPE_PERCENT_LAYOUT'),
                                                        'total_objects' => $self->param('TOTAL_FANOUT_CONFIG_COUNT'));
    # For the destination we start with no must use clusters.
    my @destination_must_use_clusters = ();
    # For the destination we start with no must use vservers.
    my @destination_must_use_vservers = ();
    # Loop through TOTAL_CASCADE_CONFIG_COUNT and create those configs.
    for (my $i = 0; $i < $self->param('TOTAL_FANOUT_CONFIG_COUNT'); $i++) {
        my $config_layout = get_next_layout_element('iterator' => $i,     # This is not required but allows the get_next_layout_element to
                                                                          # rotate through the layout instead of exhausting the first, then
                                                                          # the second, etc.
                                                    'layout'   => $layout); # layout is the numbers layout we got earlier.

        # Config the config layout into an array of the types of snapmirrors in this cascade config.
        my @config_type_array = split(';',$config_layout);

        my %last_resort_aggregates_hash = ();
        my %last_resort_vservers_hash = ();
        my %last_resort_clusters_hash = ();


        # For the source volume it can come from anywhere.
        my ($source_volume) = $self->get_available_snapmirror_volume('available_volumes' => $available_volumes,);

        # Update some metadata on the volume.
        $source_volume->set_metadata_attributes('sm_config'        => 'fanout',
                                                'sm_config_layout' => $config_layout,
                                                'fanout_index'     => 0,
                                                'sm_config_id'     => $$config_id,);

        my $source_cluster = $source_volume->get_attribute('cluster_name');
        my $source_vserver = $source_volume->get_attribute('vserver');
        my $source_aggregate = $source_volume->get_attribute('aggregate');
        my @destination_volumes = ();
        # Now we need to find our destination_volumes.
        # We will make the source cluster always a last_resort
        # and also try to rotate through any remaining clusters.
        for (my $j = 0; $j < scalar (@config_type_array); $j++) {
            my %clusters_hash = ();
            foreach my $volume (@{$available_volumes}) {
                $clusters_hash{$volume->get_attribute('cluster_name')}++;
            }
            my @clusters = keys (%clusters_hash);
            # If there are clusters in the
            my %temp_hash = ();
            foreach my $cluster (@clusters) {
                if(defined $last_resort_clusters_hash{$cluster}) {
                    $temp_hash{$cluster} = $last_resort_clusters_hash{$cluster};
                }
            }
            %last_resort_clusters_hash = %temp_hash;
            my @last_resort_clusters = ();
            # Build the last_resort_clusters based on the home many times each has been used.
            my $times_used = 0;
            foreach my $cluster (keys %last_resort_clusters_hash) {
                if ($last_resort_clusters_hash{$cluster} == $times_used) {
                    # If this is used the same number of times then add it to the list
                    push(@last_resort_clusters,$cluster);
                } elsif ($last_resort_clusters_hash{$cluster} > $times_used) {
                    # If this is used more number of times then make it the list and update
                    # the times used. This way we can have the last resorts be the highest used
                    # clusters.
                    @last_resort_clusters = ($cluster);
                    $times_used = $last_resort_clusters_hash{$cluster};
                } # Else we ignore since it is not the most used.
            }

            # make sure to always have the source cluster in the last resort list.
            push(@last_resort_clusters,$source_cluster);

            # if the last resort clusters is >= the available clusters that means that all clusters
            # are in the last resort so we can assume that we only want to have the source cluster
            # as the last resort.
            if(scalar (@last_resort_clusters) >= scalar (@clusters)) {
                @last_resort_clusters = ($source_cluster);
            }

            my @last_resort_vservers = ();
            # Build the last_resort_clusters based on the home many times each has been used.
            $times_used = 0;
            foreach my $vserver (keys %last_resort_vservers_hash) {
                if ($last_resort_vservers_hash{$vserver} == $times_used) {
                    # If this is used the same number of times then add it to the list
                    push(@last_resort_vservers,$vserver);
                } elsif ($last_resort_vservers_hash{$vserver} > $times_used) {
                    # If this is used more number of times then make it the list and update
                    # the times used. This way we can have the last resorts be the highest used
                    # vservers.
                    @last_resort_vservers = ($vserver);
                    $times_used = $last_resort_vservers_hash{$vserver};
                } # Else we ignore since it is not the most used.
            }

            my @last_resort_aggregates = ();
            # Build the last_resort_clusters based on the home many times each has been used.
            $times_used = 0;
            foreach my $aggregate (keys %last_resort_aggregates_hash) {
                if ($last_resort_aggregates_hash{$aggregate} == $times_used) {
                    # If this is used the same number of times then add it to the list
                    push(@last_resort_aggregates,$aggregate);
                } elsif ($last_resort_aggregates_hash{$aggregate} > $times_used) {
                    # If this is used more number of times then make it the list and update
                    # the times used. This way we can have the last resorts be the highest used
                    # aggregates.
                    @last_resort_aggregates = ($aggregate);
                    $times_used = $last_resort_aggregates_hash{$aggregate};
                } # Else we ignore since it is not the most used.
            }

            # get a volume based on the criteria from the previous volumes.
            # If last_resort and unuseable have the same vservers unuseable takes precendence.
            my ($volume) = $self->get_available_snapmirror_volume('available_volumes'  => $available_volumes,
                                                                  'last_resort_clusters' => \@last_resort_clusters,
                                                                  'last_resort_vservers' => \@last_resort_vservers,
                                                                  'last_resort_aggregates' => \@last_resort_aggregates,
                                                                  'unuseable_vservers'   => [$source_vserver],
                                                                  'unuseable_aggregates' => [$source_aggregate],);
            # push the volume into our list of volumes for the chain.
            push(@destination_volumes,$volume);

            # Now we need to update the criteria based on the volume found.
            $last_resort_vservers_hash{$volume->get_attribute('vserver')}++;
            $last_resort_clusters_hash{$volume->get_attribute('cluster_name')}++;
        }

        for (my $j = 0; $j < scalar (@config_type_array); $j++) {
            # The destination volume is the volume at the current index plus 1.
            my $destination_volume = $destination_volumes[$j];

            # Update some metadata on the volume.
            $destination_volume->set_metadata_attributes('sm_config'     => 'fanout',
                                                         'sm_config_layout' => $config_layout,
                                                         'fanout_index' => $j,
                                                         'sm_config_id'  => $$config_id,);

            # The SM Type is the config type arrays current index
            my $SM_TYPE = $config_type_array[$j];
            $self->generate_snapmirror_resource('source_volume'                  => $source_volume,
                                                'destination_volume'             => $destination_volume,
                                                'sm_type'                        => $SM_TYPE,
                                                'vserver_sm_policy_hash'         => $vserver_sm_policy_hash,
                                                'vserver_snapshot_policy_hash'   => $vserver_snapshot_policy_hash,
                                                'cluster_job_schedule_cron_hash' => $cluster_job_schedule_cron_hash,,
                                                'xgen_metadata_args'             => {'sm_config'      => 'fanout',
                                                                                     'sm_config_layout' => $config_layout,
                                                                                     'fanout_index'  => $j+1,
                                                                                     'sm_config_id' => $$config_id});
            $$created_snapmirrors++;
        }
        # We finished the sm config so we need to update the config id
        $$config_id++;
    }
}

###############################################################################
# DESCRIPTION: Generates snapmirror resources in the configuration. This method
#              generates snapmirrors in the configuration based on the rules
#              for a cascade config.  Cascade in this method means
#              multiple snapmirrors in a change of volumes.
#              VOl1->Vol2->Vol3
#              It also prefers volumes on different clusters, and vservers
#              Creates a number of configs based on the
#              TOTAL_CASCADE_CONFIG_COUNT constant and the percentage of
#              configs with each type of layout is based on the
#              CASCADE_TYPE_PERCENT_LAYOUT constant.
# INPUT: snapmirror_count - number of snapmirrors left to be created
#        available_volumes - reference to an array of available volumes.
#        created_snapmirrors - a scalar reference to the number of already
#           created snapmirrors.
#        config_id - The next config id to use for snapmirror configs.
#        vserver_sm_policy_hash - a hashref to a mapping of snapmirror
#           policy name to vserver and cluster.
#        vserver_snapshot_policy_hash - a hashref to a mapping of snapshot
#           policy name to vserver and cluster.
#        cluster_job_schedule_cron_hash - a hashref to a mapping of schedules
#           to cluster.
# OUTPUT: NONE
# Exceptions: None thrown directly by this method.
###############################################################################
sub generate_cascade_snapmirror_configs {
    my ($self, %args) = @_;
    my $snapmirror_count = $args{'snapmirror_count'};
    my $available_volumes = $args{'available_volumes'};
    my $created_snapmirrors = $args{'created_snapmirrors'};
    my $config_id           = $args{'config_id'};
    # A mapping of vservers to snapmirror policies.
    my $vserver_sm_policy_hash = $args{'vserver_sm_policy_hash'};
    # A mapping of vservers to snapshot policies.
    my $vserver_snapshot_policy_hash = $args{'vserver_snapshot_policy_hash'};
    # A mapping of clusters to job schedules.
    my $cluster_job_schedule_cron_hash = $args{'cluster_job_schedule_cron_hash'};

    # This will get the layout in terms of numbers instead of percents based on the total object count passed in.
    my $layout = convert_percent_layout_to_count_layout('layout'        => $self->param('CASCADE_TYPE_PERCENT_LAYOUT'),
                                                        'total_objects' => $self->param('TOTAL_CASCADE_CONFIG_COUNT'));

    # Loop through TOTAL_CASCADE_CONFIG_COUNT and create those configs.
    for (my $i = 0; $i < $self->param('TOTAL_CASCADE_CONFIG_COUNT'); $i++) {
        my $config_layout = get_next_layout_element('iterator' => $i,     # This is not required but allows the get_next_layout_element to
                                                                          # rotate through the layout instead of exhausting the first, then
                                                                          # the second, etc.
                                                    'layout'   => $layout); # layout is the numbers layout we got earlier.

        # Config the config layout into an array of the types of snapmirrors in this cascade config.
        my @config_type_array = split(';',$config_layout);

        my @volumes = ();

        # For the first volume we set all the filters to nothing since the first
        # volume can go anywhere.
        my @unuseable_vservers = ();
        my %last_resort_vservers_hash = ();
        my %last_resort_clusters_hash = ();
        # We need to have snapmirrors+1 volumes so we use <=
        for (my $j = 0; $j <= scalar (@config_type_array); $j++) {
            my @last_resort_clusters = ();
            # Build the last_resort_clusters based on the home many times each has been used.
            my $times_used = 0;
            foreach my $cluster (keys %last_resort_clusters_hash) {
                if ($last_resort_clusters_hash{$cluster} == $times_used) {
                    # If this is used the same number of times then add it to the list
                    push(@last_resort_clusters,$cluster);
                } elsif ($last_resort_clusters_hash{$cluster} > $times_used) {
                    # If this is used more number of times then make it the list and update
                    # the times used. This way we can have the last resorts be the highest used
                    # clusters.
                    @last_resort_clusters = ($cluster);
                    $times_used = $last_resort_clusters_hash{$cluster};
                } # Else we ignore since it is not the most used.
            }

            my @last_resort_vservers = ();
            # Build the last_resort_clusters based on the home many times each has been used.
            $times_used = 0;
            foreach my $vserver (keys %last_resort_vservers_hash) {
                if ($last_resort_vservers_hash{$vserver} == $times_used) {
                    # If this is used the same number of times then add it to the list
                    push(@last_resort_vservers,$vserver);
                } elsif ($last_resort_vservers_hash{$vserver} > $times_used) {
                    # If this is used more number of times then make it the list and update
                    # the times used. This way we can have the last resorts be the highest used
                    # vservers.
                    @last_resort_vservers = ($vserver);
                    $times_used = $last_resort_vservers_hash{$vserver};
                } # Else we ignore since it is not the most used.
            }

            # get a volume based on the criteria from the previous volumes.
            # If last_resort and unuseable have the same vservers unuseable takes precendence.
            my ($volume) = $self->get_available_snapmirror_volume('available_volumes'  => $available_volumes,
                                                                  'last_resort_clusters' => \@last_resort_clusters,
                                                                  'last_resort_vservers' => \@last_resort_vservers,
                                                                  'unuseable_vservers'   => \@unuseable_vservers);
            # push the volume into our list of volumes for the chain.
            push(@volumes,$volume);

            # Now we need to update the criteria based on the volume found.
            # Can not use the same vserver for both source and destination of a single SM.
            @unuseable_vservers = ($volume->get_attribute('vserver'));
            $last_resort_vservers_hash{$volume->get_attribute('vserver')}++;
            $last_resort_clusters_hash{$volume->get_attribute('cluster_name')}++;

            # Update some metadata on the volume.
            $volume->set_metadata_attributes('sm_config'     => 'cascade',
                                             'sm_config_layout' => $config_layout,
                                             'cascade_index' => $j,
                                             'sm_config_id'  => $$config_id,)
        }

        # Now loop through and create all the Snapmirrors in the chain.
        for (my $j = 0; $j < scalar (@config_type_array); $j++) {
            # The source volume is the volume at the current index.
            my $source_volume      = $volumes[$j];
            # The destination volume is the volume at the current index plus 1.
            my $destination_volume = $volumes[$j+1];

            # The SM Type is the config type arrays current index
            my $SM_TYPE = $config_type_array[$j];
            $self->generate_snapmirror_resource('source_volume'                  => $source_volume,
                                                'destination_volume'             => $destination_volume,
                                                'sm_type'                        => $SM_TYPE,
                                                'vserver_sm_policy_hash'         => $vserver_sm_policy_hash,
                                                'vserver_snapshot_policy_hash'   => $vserver_snapshot_policy_hash,
                                                'cluster_job_schedule_cron_hash' => $cluster_job_schedule_cron_hash,,
                                                'xgen_metadata_args'             => {'sm_config'      => 'cascade',
                                                                                     'sm_config_layout' => $config_layout,
                                                                                     'cascade_index'  => $j,
                                                                                     'snapmirror_initialize_order' => $j,
                                                                                     'sm_config_id' => $$config_id,});
            $$created_snapmirrors++;
        }
        # We finished the sm config so we need to update the config id
        $$config_id++;
    }
}

###############################################################################
# DESCRIPTION: Generates snapmirror resources in the configuration. This method
#              generates snapmirrors in the configuration based on the rules
#              for a unidirectional config.  Unidirectional in this method means
#              each snapmirror creates is an atomic snapmirror with no
#              deliberate relationship to other snapmirrors.
#              It also prefers volumes on different clusters, and vservers.
#              Creates a number of snapmirrors based on the
#              snapmirror_count argument and the percentage of
#              snapmirrors with each type of layout is based on the
#              UNIDIRECTIONAL_TYPE_PERCENT_LAYOUT constant. This method
#              fills in snapmirrors up to the TOTAL_SNAPMIRROR_COUNT.
# INPUT: snapmirror_count - number of snapmirrors left to be created
#        available_volumes - reference to an array of available volumes.
#        created_snapmirrors - a scalar reference to the number of already
#           created snapmirrors.
#        config_id - The next config id to use for snapmirror configs.
#        vserver_sm_policy_hash - a hashref to a mapping of snapmirror
#           policy name to vserver and cluster.
#        vserver_snapshot_policy_hash - a hashref to a mapping of snapshot
#           policy name to vserver and cluster.
#        cluster_job_schedule_cron_hash - a hashref to a mapping of schedules
#           to cluster.
# OUTPUT: NONE
# Exceptions: None thrown directly by this method.
###############################################################################
sub generate_unidirectional_snapmirror_config {
    my ($self, %args) = @_;
    my $snapmirror_count    = $args{'snapmirror_count'};
    my $available_volumes   = $args{'available_volumes'};
    my $created_snapmirrors = $args{'created_snapmirrors'};
    my $config_id           = $args{'config_id'};
    # A mapping of vservers to snapmirror policies.
    my $vserver_sm_policy_hash = $args{'vserver_sm_policy_hash'};
    # A mapping of vservers to snapshot policies.
    my $vserver_snapshot_policy_hash = $args{'vserver_snapshot_policy_hash'};
    # A mapping of clusters to job schedules.
    my $cluster_job_schedule_cron_hash = $args{'cluster_job_schedule_cron_hash'};
    # Create a SMGR layout based on the number of SMGR snapmirror required.
    my $snapmirror_type_layout = convert_percent_layout_to_count_layout('layout'        => $self->param('UNIDIRECTIONAL_SNAPMIRROR_TYPE_PERCENT_LAYOUT'),
                                                                        'total_objects' => $snapmirror_count);

    for (my $i = 0; $i < $snapmirror_count; $i++) {
        # The source volume can be anywhere.
        my ($source_volume)      = $self->get_available_snapmirror_volume('available_volumes' => $available_volumes);
        # The destination volume cannot be on the same vserver and should try to be on a different cluster.
        my ($destination_volume) = $self->get_available_snapmirror_volume('available_volumes' => $available_volumes,
                                                                          'unuseable_vservers'   => [$source_volume->get_attribute('vserver')],
                                                                          'last_resort_clusters' => [$source_volume->get_attribute('cluster_name')]);
        $source_volume->set_metadata_attributes('sm_config'     => 'unidirectional',
                                                'sm_config_id'  => $$config_id,);
        $destination_volume->set_metadata_attributes('sm_config'     => 'unidirectional',
                                                     'sm_config_id'  => $$config_id,);
        # Get the next cg item type element to use from the smgr layout
        my $SM_TYPE = get_next_layout_element('iterator' => $i,
                                              'layout'   => $snapmirror_type_layout);
        $self->generate_snapmirror_resource('source_volume'                  => $source_volume,
                                            'destination_volume'             => $destination_volume,
                                            'sm_type'                        => $SM_TYPE,
                                            'vserver_sm_policy_hash'         => $vserver_sm_policy_hash,
                                            'vserver_snapshot_policy_hash'   => $vserver_snapshot_policy_hash,
                                            'cluster_job_schedule_cron_hash' => $cluster_job_schedule_cron_hash,
                                            'xgen_metadata_args'             => {'sm_config'     => 'unidirectional',
                                                                                 'sm_config_id' => $$config_id});
        $$created_snapmirrors++;
    }
    # Finished the unidirectional config so update the config_id.
    $$config_id++;
 }

###############################################################################
# DESCRIPTION: Generates a single snapmirror resource in the configuration.
#              This will generate a snapmirror resource using the volumes
#              passed in. The arguments for snapmirror are based on a
#              template corresponding to the snapmirror type.
#              snapmirror_policy argument.
#              Also calls method to generate snapmirror policies and
#              job schedule crons based on whether they are needed for the
#              snapmirror type.
#              Will Create Snapmirror Initialize Resources based on the
#              INITIALIZE_SNAPMIRRORS constant.
#              Will set metadata to turn off EnGenX setup and cleanup
#              for the snapmirror and snapmirror initialize based on the
#              SETUP_CLEANUP_SNAPMIRRORS constant.
# INPUT: source_volume - the source volume of the snapmirror
#        destination_volume - the destination volume of the snapmirror
#        vserver_sm_policy_hash - a hashref to a mapping of snapmirror
#           policy name to vserver and cluster.
#        vserver_snapshot_policy_hash - a hashref to a mapping of snapshot
#           policy name to vserver and cluster.
#        cluster_job_schedule_cron_hash - a hashref to a mapping of schedules
#           to cluster.
#        xgen_metadata_args - arguments to store on the snapmirror as metadata.
# OUTPUT: NONE
# Exceptions: None thrown directly by this method.
###############################################################################
sub generate_snapmirror_resource {
    my ($self, %args) = @_;

    my $source_volume = $args{'source_volume'};
    my $destination_volume = $args{'destination_volume'};
    my $SM_TYPE = $args{'sm_type'};

    my $source_volume_name = $source_volume->get_attribute('volume') . '_src_' . lc $SM_TYPE;
    my $dest_volume_name = $destination_volume->get_attribute('volume') . '_dst_' . lc $SM_TYPE;

    $source_volume->set_private_attributes(volume => $source_volume_name);
    $destination_volume->set_private_attributes(volume => $dest_volume_name);


    # A mapping of vservers to snapmirror policies.
    my $vserver_sm_policy_hash = $args{vserver_sm_policy_hash};
    # A mapping of vservers to snapshot policies.
    my $vserver_snapshot_policy_hash = $args{vserver_snapshot_policy_hash};
    # A mapping of clusters to job schedules.
    my $cluster_job_schedule_cron_hash = $args{cluster_job_schedule_cron_hash};
    my $xgen_metadata_args             = $args{xgen_metadata_args};
    my $method = $self->can($SM_TYPE);
    NATE::BaseException->throw("$SM_TYPE is not a supported Snapmirror Type in this configuration.\nPlease check your SM definition parameters.") unless(defined $method);
    my $snapmirror_args = $self->$method(
                   'source_volume' => $source_volume,
                   'destination_volume' => $destination_volume,
                   'vserver_snapshot_policy_hash' => $vserver_snapshot_policy_hash, # This is used for snapmirrors that require a snapshot policy.
                   );

    my ($source_aggr) = $source_volume->get_ancestors(type => 'NACL::AppBuilder::Xgen::Resource::Aggregate');
    my ($source_node) = $source_aggr->get_ancestors(type => 'NACL::AppBuilder::Xgen::Resource::Node');
    my ($destination_aggr) = $destination_volume->get_ancestors(type => 'NACL::AppBuilder::Xgen::Resource::Aggregate');
    my ($destination_node) = $destination_aggr->get_ancestors(type => 'NACL::AppBuilder::Xgen::Resource::Node');
    $snapmirror_args->{xgen_metadata}->{source_node}  = $source_node->get_attribute('node');
    $snapmirror_args->{xgen_metadata}->{destination_node}  = $destination_node->get_attribute('node');
    if(not $self->param('SETUP_CLEANUP_SNAPMIRRORS')) {
        $snapmirror_args->{xgen_metadata}->{setup}   = 0;
        $snapmirror_args->{xgen_metadata}->{cleanup} = 0;
    }

    if(defined $xgen_metadata_args) {
        foreach my $key (keys %{$xgen_metadata_args}) {
            $snapmirror_args->{xgen_metadata}->{$key} = $xgen_metadata_args->{$key};
        }
    }

    # If the volumes are on different clusters then we want to add the source-cluster as an argument.
    if($source_volume->get_attribute('cluster_name') ne $destination_volume->get_attribute('cluster_name')) {
        $snapmirror_args->{'source-cluster'} = $source_volume->get_attribute('cluster_name');
    }

    # Create the snapmirror resource.
    my $snapmirror = NACL::AppBuilder::Xgen::Resource::Snapmirror->new(%{$snapmirror_args});
    # The destination is required to be second since we need the destination cluster to be
    # the cluster used for these snapmirror commands.
    $snapmirror->add_parent_resources(resources => [$source_volume,$destination_volume]);
    $self->add_resources(resources => [$snapmirror]);

    my $snapmirror_policy = $snapmirror_args->{'policy'};
    # Add the snapmirror_policy to the destination vserver if it has not already been added.
    $self->generate_snapmirror_policy_resource('volume'                 => $destination_volume,
                                               'snapmirror_policy'      => $snapmirror_policy,
                                               'vserver_sm_policy_hash' => $vserver_sm_policy_hash);


    my $job_schedule_cron = $snapmirror_args->{'schedule'};
    # Add the job_schedule to the destination cluster if it has not already been added.
    $self->generate_job_schedule_cron_resource('volume'                         => $destination_volume,
                                               'job_schedule_cron'              => $job_schedule_cron,
                                               'cluster_job_schedule_cron_hash' => $cluster_job_schedule_cron_hash);


    # Add the job_schedule to the source cluster if it has not already been added.
    $self->generate_job_schedule_cron_resource('volume'                         => $source_volume,
                                               'job_schedule_cron'              => $job_schedule_cron,
                                               'cluster_job_schedule_cron_hash' => $cluster_job_schedule_cron_hash);

    if($self->param('INITIALIZE_SNAPMIRRORS')) {
        # Initialize does not have schedule so delete it.
        delete $snapmirror_args->{'schedule'};

        # Create the snapmirror initialize resource so the snapmirror gets initialized.
        my $snapmirror_initialize = NACL::AppBuilder::Xgen::Resource::SnapmirrorInitialize->new(%{$snapmirror_args});
        $snapmirror_initialize->add_parent_resources(resources => [$snapmirror]);
        $self->add_resources(resources => [$snapmirror_initialize]);
    }

}

###############################################################################
# VOLUME EFFICIENCY POLICY TEMPLATES BELOW
# Create new template by making the name of the method
# the policy name uppercased and suffixed with _VOLUME_EFFICIENCY_POLICY
# The method should return the arguments wanted for policy create.
###############################################################################

###############################################################################
# se_hourly policy that is just a default policy.
###############################################################################
sub SE_HOURLY_VOLUME_EFFICIENCY_POLICY {
    return {'policy' => 'se_hourly'};
}
###############################################################################
# VOLUME EFFICIENCY OPTION TEMPLATES BELOW
# Create new template by making the name of the method
# the template name uppercased and suffixed with _VOLUME_EFFICIENCY_OPTIONS
# The method should return the arguments wanted for volume efficiency modify.
###############################################################################

###############################################################################
# Default Storage Efficiency Turned on using SE_HOURLY policy and started
###############################################################################
sub STORAGE_EFFICIENCY_ON_VOLUME_EFFICIENCY_OPTIONS {
    return {'policy' => 'se_hourly'};
}

###############################################################################
# Storage efficiency on with compression
# Does not use a volume efficiency policy.
###############################################################################
sub COMPRESSION_ON_VOLUME_EFFICIENCY_OPTIONS {
    return {'compression'        => 'true',};
}

###############################################################################
# JOB SCHEDULE TEMPLATES BELOW
# Create new template by making the name of the method
# the Job Schedule name uppercased and suffixed with _JOB_SCHEDULE
# The method should return the arguments wanted for job schedule cron create.
###############################################################################

###############################################################################
# Job Schedule for 10 minutes after the hour every hour.
###############################################################################
sub HOURLY_10_JOB_SCHEDULE {
    return {'name'  =>'hourly_10',
            'minute'=>'10'};
}

###############################################################################
# Job Schedule for 15 minutes after the hour every hour.
###############################################################################
sub HOURLY_15_JOB_SCHEDULE {
    return {'name'  =>'hourly_15',
            'minute'=>'15'};
}

###############################################################################
# Job Schedule for 30 minutes after the hour every hour.
###############################################################################
sub SE_HOURLY_JOB_SCHEDULE {
    return {'name'  =>'se_hourly',
            'minute'=>'30'};
}

###############################################################################
# SNAPSHOT POLICY TEMPLATES BELOW
# Create new template by making the name of the method
# the Snapshot Policy name uppercased and suffixed with _SNAPSHOT_POLICY
# The method should return the arguments wanted for snapshot policy create.
###############################################################################

###############################################################################
# Snapshot policy to be used with some
###############################################################################
sub XDP_POLICY_SNAPSHOT_POLICY {
    return {'policy'=>'xdp_policy',
            'enabled'=>'true',
            'schedule1'=>'hourly', 'prefix1'=>'hourly', 'count1'=>'16', 'snapmirror-label1'=>'rpo_hourly',
            'schedule2'=>'daily',  'prefix2'=>'daily',  'count2'=>'2',  'snapmirror-label2'=>'rpo_daily',
            'schedule3'=>'weekly', 'prefix3'=>'weekly', 'count3'=>'2',  'snapmirror-label3'=>'rpo_weekly'};
}

###############################################################################
# SNAPMIRROR POLICY TEMPLATES BELOW
# Create new template by making the name of the method
# the Snapmirror Policy name uppercased and suffixed with _SNAPMIRROR_POLICY
# The method should return the arguments wanted for snapshot policy create.
# Also should return xgen_snapmirror_policy_rules as part of the return hash
# if snapmirror policy rules need to be created.
###############################################################################

###############################################################################
# Policy for SMGR snapmirrors.
###############################################################################
sub SMGR_SNAPMIRROR_POLICY {
    return {'tries'                   => 8,
            'policy'                  => 'smgr',
            'user-snapshot-timeout'   => 10,
            'type'                    => 'sync-mirror',
            'xgen_nacl_cs_validation' => 0};
}

###############################################################################
# Policy for ADR snapmirrors with net compression enabled.
###############################################################################
sub ADR_NETCOMP_SNAPMIRROR_POLICY {
    return {'policy'                          => 'ADR_netcomp',
            'type'                            => 'async-mirror',
            'is-network-compression-enabled'  => 'true'};
}

###############################################################################
# Policy for Vault Snapmirrors.
# Include snapmirror policy rules
###############################################################################
sub XDPRPO_SNAPMIRROR_POLICY {
    return {'policy' => 'XDPRPO',
            'xgen_snapmirror_policy_rules' =>
                [{'snapmirror-label'=>'rpo_hourly', 'keep'=>'24'},
                 {'snapmirror-label'=>'rpo_daily',  'keep'=>'7'},
                 {'snapmirror-label'=>'rpo_weekly', 'keep'=>'4'},
                 {'snapmirror-label'=>'weekly',     'keep'=>'4'}]};
}

###############################################################################
# Policy for Vault Snapmirrors with net compression enabled.
# Include snapmirror policy rules
###############################################################################
sub XDPRPO_NETCOMP_SNAPMIRROR_POLICY {
    return {'policy' => 'XDPRPO_netcomp',
            'is-network-compression-enabled'  => 'true',
            'xgen_snapmirror_policy_rules' =>
                [{'snapmirror-label'=>'rpo_hourly', 'keep'=>'24'},
                 {'snapmirror-label'=>'rpo_daily',  'keep'=>'7'},
                 {'snapmirror-label'=>'rpo_weekly', 'keep'=>'4'}]};
}

###############################################################################
# Policy for Mirror Latest Snapmirrors.
# Include snapmirror policy rules
###############################################################################
sub MIRRORLATEST_NC_SNAPMIRROR_POLICY {
    my ($self, %args) = @_;
    return {'policy'                          => 'MirrorLatest_NC',
            'type'                            => 'async-mirror',
            'is-network-compression-enabled'  => 'true'
            };
}

###############################################################################
# Policy for Mirror All Snapshots Snapmirrors.
# Include snapmirror policy rules
###############################################################################
sub MIRRORALLSNAPSHOTS_NC_SNAPMIRROR_POLICY {
    my ($self, %args) = @_;
    return {'policy' => 'MirrorAllSnapshots_NC',
            'type'                            => 'async-mirror',
            'is-network-compression-enabled'  => 'true',
            'xgen_snapmirror_policy_rules' =>
                [{'snapmirror-label'=>'all_source_snapshots', 'keep'=>'1',},]};
}

###############################################################################
# Policy for Vault Snapmirrors with net compression enabled.
# Include snapmirror policy rules
###############################################################################
sub MIRRORANDVAULT_NC_SNAPMIRROR_POLICY {
    my ($self, %args) = @_;
    return {'policy' => 'MirrorAndVault_NC',
            'xgen_snapmirror_policy_rules' =>
                [{'snapmirror-label'=>'rpo_hourly', 'keep'=>'24',},
                 {'snapmirror-label'=>'rpo_daily',  'keep'=>'7',},
                 {'snapmirror-label'=>'rpo_weekly', 'keep'=>'4',},
                 {'snapmirror-label'=>'daily',      'keep'=>'7',},
                 {'snapmirror-label'=>'weekly',     'keep'=>'52',},]};
}

###############################################################################
# Policy for Vault Snapmirrors with net compression enabled.
# Include snapmirror policy rules
###############################################################################
sub MIRRORANDVAULT_SNAPMIRROR_POLICY {
    my ($self, %args) = @_;
    return {'policy' => 'MirrorAndVault',
            'vserver' => $args{'volume'}->get_attribute('cluster_name'),
            'xgen_metadata' => {'setup' => 0,
                                'cleanup' => 0,},
            'xgen_snapmirror_policy_rules' =>
                [{'snapmirror-label'=>'rpo_hourly', 'keep'=>'24', 'vserver' => $args{'volume'}->get_attribute('cluster_name')},
                 {'snapmirror-label'=>'rpo_daily',  'keep'=>'7', 'vserver' => $args{'volume'}->get_attribute('cluster_name')},
                 {'snapmirror-label'=>'rpo_weekly', 'keep'=>'4', 'vserver' => $args{'volume'}->get_attribute('cluster_name')}]};
}

###############################################################################
# SNAPMIRROR TEMPLATES BELOW
# Create new template by making the name of the method
# the snapmirror template name uppercased
# The method should return the arguments wanted for snapmirror create, and
# set volume attributes if needed for the snapmirror.
###############################################################################

###############################################################################
# Creates the basic arguments for snapmirrors between volumes.
# Inputs:   source_volume - the source volume for the relationship.
#           destination_volume - the destination volume for the relationship.
#           sm_type - the type of snapmirror being created.
#           sm_policy - the name of the snapmirror policy to use
# Outputs:  A hash containing the snapmirror arguments.
#           Also updates the source and destination volumes with some metadata.
###############################################################################
sub generate_basic_snapmirror_args {
    my ($self, %args) = @_;
    my $source_volume = $args{'source_volume'};
    my $destination_volume = $args{'destination_volume'};
    my $sm_type = $args{'sm_type'};
    my $type    = $args{'type'};
    my $sm_policy = $args{'sm_policy'};

    my %optional_args = ();
    $optional_args{'policy'} = $sm_policy if (defined $sm_policy);
    $optional_args{'schedule'} = $args{'schedule'} if (defined $args{'schedule'});
    # Set metadata that can be queried for later.  When this gets to engenx all the fields will
    # be prepended with underscore.
    $source_volume->set_metadata_attributes('source_for_sm_type' => $sm_type);
    $destination_volume->set_metadata_attributes('destination_for_sm_type' => $sm_type);
    # Setup the source, destination, and type arguments.
    return {'source-vserver'    => $source_volume->get_attribute('vserver'),
            'source-volume'         => $source_volume->get_attribute('volume'),
            'destination-vserver'   => $destination_volume->get_attribute('vserver'),
            'destination-volume'    => $destination_volume->get_attribute('volume'),
            'type'                  => $type,
            %optional_args,
            xgen_metadata           => {'sm_type' => $sm_type} # set some metadata to be used later for filtering.
            };
}

###############################################################################
# Creates an smgr type Snapmirror Resource.
# The cg-item-mappings will be added later by the method
#     add_cg_items_to_SMGR_snapmirror_resources
# Defines each SMGR as a type file|lun
# Creates source_path/destination_path based on type.
# Uses policy smgr
###############################################################################
sub SMGR {
    my ($self, %args) = @_;
    my $source_volume = $args{'source_volume'};
    my $destination_volume = $args{'destination_volume'};

    # Set metadata that can be queried for later.  When this gets to engenx all the fields will
    # be prepended with underscore.
    $source_volume->set_metadata_attributes('source_for_sm_type' => 'smgr');
    $destination_volume->set_metadata_attributes('destination_for_sm_type' => 'smgr');
    # Setup the source, destination, and type arguments.  Many of the other required arguments will be added
    # by another method. This method just creates the shell that points to the correct volumes,type, and policy.
    return {'type'                => 'XDP',
            'policy'              => 'smgr',
            xgen_metadata         => {'sm_type' => 'smgr',
                                      'source-vserver'      => $source_volume->get_attribute('vserver'),
                                      'destination-vserver' => $destination_volume->get_attribute('vserver'),
                                      'source-volume'       => $source_volume->get_attribute('volume'),
                                      'destination-volume'  => $destination_volume->get_attribute('volume'),}
            };
}

###############################################################################
# Creates an ADR snapmirror
# type is DP
# Destination volume is type DP
# Uses the DPDefault policy
###############################################################################
sub ADR {
    my ($self, %args) = @_;
    $args{'destination_volume'}->set_private_attributes(type => 'DP',
                                                        'snapshot-policy' => 'none');
    my $snapmirror_args = $self->generate_basic_snapmirror_args('sm_type'   => 'adr',
                                                                'sm_policy' => 'DPDefault',
                                                                'type'      => 'DP',
                                                                'schedule'  => 'hourly_15',
                                                                %args);
    return $snapmirror_args;
}

###############################################################################
# Creates an ADR snapmirror with net compression enabled
# type is DP
# Destination volume is type DP
# Uses the ADR_netcomp policy
###############################################################################
sub ADR_NC {
    my ($self, %args) = @_;
    # Same as ADR but with a policy having net compression turned on
    my $snapmirror_args = $self->ADR('sm_type'   => 'adr_nc',
                                     'sm_policy' => 'ADR_netcomp',
                                     %args);
    return $snapmirror_args;
}

###############################################################################
# Creates an Vault snapmirror
# type is XDP
# Uses the XDPRPO policy
# Uses hourly_10 schedule
# Destination volume is type DP
# The source volume has a snapshot policy.
###############################################################################
sub VAULT {
    my ($self, %args) = @_;
    $args{'destination_volume'}->set_private_attributes(type => 'DP',
                                                        'snapshot-policy' => 'none');
    try {
        my $type = $args{source_volume}->get_attribute('type');
        # Only set the snapshot-policy if the type is not DP.
        if ($type ne 'DP') {
            $args{'source_volume'}->set_private_attributes('snapshot-policy' => 'xdp_policy');
        }
    } otherwise {
        # Only set the snapshot-policy if the type is not DP.
        $args{'source_volume'}->set_private_attributes('snapshot-policy' => 'xdp_policy');
    }
    $self->generate_snapshot_policy_resource('volume'          => $args{source_volume},
                                             'snapshot_policy' => 'xdp_policy',
                                             %args,);
    my $snapmirror_args = $self->generate_basic_snapmirror_args('sm_type'   => 'vault',
                                                                'sm_policy' => 'XDPRPO',
                                                                'type'      => 'XDP',
                                                                'schedule'  => 'hourly_10',
                                                                %args);
    return $snapmirror_args;
}

###############################################################################
# Creates an Vault where the source volume has storage efficiency on.
###############################################################################
sub VAULT_SRCSE {
    my ($self, %args) = @_;
    $args{'source_volume'}->set_metadata_attributes('volume_efficiency_options' => 'storage_efficiency_on');
    my $snapmirror_args = $self->VAULT('sm_type'   => 'vault_srcse',
                                       %args);
    return $snapmirror_args;
}

###############################################################################
# Creates an Vault where the source volume has compression on.
###############################################################################
sub VAULT_SRCCOMPRESSION {
    my ($self, %args) = @_;
    $args{'source_volume'}->set_metadata_attributes('volume_efficiency_options' => 'compression_on');
    my $snapmirror_args = $self->VAULT('sm_type'   => 'vault_srccompression',
                                       %args);
    return $snapmirror_args;
}

###############################################################################
# Creates an Vault where the destination volume has storage efficiency on.
###############################################################################
sub VAULT_ADDDEDUP {
    my ($self, %args) = @_;
    $args{'destination_volume'}->set_metadata_attributes('volume_efficiency_options' => 'storage_efficiency_on');
    my $snapmirror_args = $self->VAULT('sm_type'   => 'vault_adddedup',
                                       %args);
    return $snapmirror_args;
}

###############################################################################
# Creates an Vault where the destination volume has compression on.
###############################################################################
sub VAULT_LRE {
    my ($self, %args) = @_;
    $args{'destination_volume'}->set_metadata_attributes('volume_efficiency_options' => 'compression_on');
    my $snapmirror_args = $self->VAULT('sm_type'   => 'vault_srccompression',
                                       %args);
    return $snapmirror_args;
}

###############################################################################
# Creates an Vault with network compression on.
# Uses the XDPRPO_netcomp policy
###############################################################################
sub VAULT_NC {
    my ($self, %args) = @_;
    # Same as VAULT but with a policy having net compression turned on
    my $snapmirror_args = $self->VAULT('sm_type'   => 'vault_nc',
                                       'sm_policy' => 'XDPRPO_netcomp',
                                       %args);
    return $snapmirror_args;
}

###############################################################################
# Creates an Vault NC where destination volume has compression on.
###############################################################################
sub VAULT_NC_LRE {
    my ($self, %args) = @_;
    $args{'destination_volume'}->set_metadata_attributes('volume_efficiency_options' => 'compression_on');
    my $snapmirror_args = $self->VAULT_NC('sm_type'   => 'vault_srccompression',
                                          %args);
    return $snapmirror_args;
}

###############################################################################
# Creates an Vault with using a different snapmirror policy
# Uses the MirrorLatest
###############################################################################
sub LRSE_DR_MIRROR_LATEST {
    my ($self, %args) = @_;
    # Same as VAULT but with a policy having MirrorLatest policy
    my $snapmirror_args = $self->VAULT('sm_type'   => 'lrse_dr_mirror_latest',
                                       'sm_policy' => 'MirrorLatest',
                                       %args);
    return $snapmirror_args;
}

###############################################################################
# Creates an LRSE-DR-MIRROR-LATEST where the destination volume has
#   compression on.
###############################################################################
sub LRSE_DR_MIRROR_LATEST_LRE {
    my ($self, %args) = @_;
    $args{'destination_volume'}->set_metadata_attributes('volume_efficiency_options' => 'compression_on');
    # Same as MirrorLatest
    my $snapmirror_args = $self->LRSE_DR_MIRROR_LATEST('sm_type'   => 'lrse_dr_mirror_latest_lre',
                                                       %args);
    return $snapmirror_args;
}

###############################################################################
# Creates an Vault with using a different snapmirror policy
# Uses the MirrorAllSnapshots
###############################################################################
sub LRSE_DR_MIRROR_ALL {
    my ($self, %args) = @_;
    # Same as VAULT but with a policy having MirrorAllSnapshots policy
    my $snapmirror_args = $self->VAULT('sm_type'   => 'lrse_dr_mirror_all',
                                       'sm_policy' => 'MirrorAllSnapshots',
                                       %args);
    return $snapmirror_args;
}

###############################################################################
# Creates an LRSE_DR_MIRROR_ALL where the destination volume has
#   compression on
###############################################################################
sub LRSE_DR_MIRROR_ALL_LRE {
    my ($self, %args) = @_;
    $args{'destination_volume'}->set_metadata_attributes('volume_efficiency_options' => 'compression_on');
    # Same as LRSE_DR_MIRROR_ALL
    my $snapmirror_args = $self->LRSE_DR_MIRROR_ALL('sm_type'   => 'lrse_dr_mirror_all_lre',
                                                    %args);
    return $snapmirror_args;
}
###############################################################################
# Creates an Vault with using a different snapmirror policy
# Uses the MirrorAndVault
###############################################################################
sub LRSE_DR_MIRROR_AND_VAULT {
    my ($self, %args) = @_;
    # Same as VAULT but with a policy having MirrorAllSnapshots policy
    my $snapmirror_args = $self->VAULT('sm_type'   => 'lrse_dr_mirror_and_vault',
                                       'sm_policy' => 'MirrorAndVault',
                                       %args);
    return $snapmirror_args;
}

###############################################################################
# Creates an LRSE_DR_MIRROR_AND_VAULT where the destination volume has
#   compression on
###############################################################################
sub LRSE_DR_MIRROR_AND_VAULT_LRE {
    my ($self, %args) = @_;
    $args{'destination_volume'}->set_metadata_attributes('volume_efficiency_options' => 'compression_on');
    # Same as LRSE_DR_MIRROR_AND_VAULT
    my $snapmirror_args = $self->LRSE_DR_MIRROR_AND_VAULT('sm_type'   => 'lrse_dr_mirror_and_vault_lre',
                                                          %args);
    return $snapmirror_args;
}

###############################################################################
# Creates an LRSE-DR-MIRROR-LATEST with network compression on
# Uses the MirrorLatest_NC policy
###############################################################################
sub LRSE_DR_MIRROR_LATEST_NC {
    my ($self, %args) = @_;
    # Same as VAULT but with a policy having MirrorLatest policy
    my $snapmirror_args = $self->VAULT('sm_type'   => 'lrse_dr_mirror_latest_nc',
                                       'sm_policy' => 'MirrorLatest_NC',
                                       %args);
    return $snapmirror_args;
}

###############################################################################
# Creates an LRSE_DR_MIRROR_LATEST_NC where the destination volume has
#   compression on
###############################################################################
sub LRSE_DR_MIRROR_LATEST_NC_LRE {
    my ($self, %args) = @_;
    $args{'destination_volume'}->set_metadata_attributes('volume_efficiency_options' => 'compression_on');
    # Same as MirrorLatest
    my $snapmirror_args = $self->LRSE_DR_MIRROR_LATEST_NC('sm_type'   => 'lrse_dr_mirror_latest_nc_lre',
                                                       %args);
    return $snapmirror_args;
}

###############################################################################
# Creates an LRSE-DR-MIRROR-ALL with network compression on
# Uses the MirrorAllSnapshots_NC policy
###############################################################################
sub LRSE_DR_MIRROR_ALL_NC {
    my ($self, %args) = @_;
    # Same as VAULT but with a policy having MirrorAllSnapshots policy
    my $snapmirror_args = $self->VAULT('sm_type'   => 'lrse_dr_mirror_all_nc',
                                       'sm_policy' => 'MirrorAllSnapshots_NC',
                                       %args);
    return $snapmirror_args;
}

###############################################################################
# Creates an LRSE_DR_MIRROR_ALL_NC where the destination volume has
#   compression on
###############################################################################
sub LRSE_DR_MIRROR_ALL_NC_LRE {
    my ($self, %args) = @_;
    $args{'destination_volume'}->set_metadata_attributes('volume_efficiency_options' => 'compression_on');
    # Same as LRSE_DR_MIRROR_ALL
    my $snapmirror_args = $self->LRSE_DR_MIRROR_ALL_NC('sm_type'   => 'lrse_dr_mirror_all_nc_lre',
                                                    %args);
    return $snapmirror_args;
}
###############################################################################
# Creates an LRSE-DR-MIRROR-ALL with network compression on
# Uses the MirrorAndVault_NC policy
###############################################################################
sub LRSE_DR_MIRROR_AND_VAULT_NC {
    my ($self, %args) = @_;
    # Same as VAULT but with a policy having MirrorAllSnapshots policy
    my $snapmirror_args = $self->VAULT('sm_type'   => 'lrse_dr_mirror_and_vault_nc',
                                       'sm_policy' => 'MirrorAndVault_NC',
                                       %args);
    return $snapmirror_args;
}

###############################################################################
# Creates an LRSE_DR_MIRROR_AND_VAULT_NC where the destination volume has
#   compression on
###############################################################################
sub LRSE_DR_MIRROR_AND_VAULT_NC_LRE {
    my ($self, %args) = @_;
    $args{'destination_volume'}->set_metadata_attributes('volume_efficiency_options' => 'compression_on');
    # Same as LRSE_DR_MIRROR_AND_VAULT
    my $snapmirror_args = $self->LRSE_DR_MIRROR_AND_VAULT_NC('sm_type'   => 'lrse_dr_mirror_and_vault_nc_lre',
                                                          %args);
    return $snapmirror_args;
}

###############################################################################
# HELPER METHODS
###############################################################################

###############################################################################
# Gets Available Volumes for Snapmirrors based on filters.
# Inputs:   volume_count - The number of volumes to return matching the filters
#           available_volumes - a mapping of available volumes
#                               to cluster and vserver.
#           all_from_same_vserver - whether all the volumes must come from the
#                                   same vserver
#           must_use_clusters - array of cluster names required for the volumes
#           last_resort_clusters - array of cluster names that should only
#                                  be used if there are no volumes in other
#                                  clusters that match the criteria.
#           unuseable_clusters - array of cluster names that can not be used
#                                for the volumes
#           must_use_vservers - array of vserver names required for the volumes
#           last_resort_vservers - array of vserver names that should only
#                                  be used if there are no volumes in other
#                                  vservers that match the criteria.
#           unuseable_vservers - array of vserver names that can not be used
#                                for the volumes
#           must_use_aggregates - array of aggregate names required for the volumes
#           last_resort_aggregates - array of aggregate names that should only
#                                  be used if there are no volumes in other
#                                  aggregates that match the criteria.
#           unuseable_aggregates - array of aggregate names that can not be used
#                                for the volumes
# Outputs:  A hash containing the snapmirror arguments.
#           Also updates the source and destination volumes with some metadata.
###############################################################################
sub get_available_snapmirror_volume {
    my ($self, %args) = @_;

    # The number of volumes to return match the filters.
    # Defaults to 1.
    my $volume_count           = $args{volume_count} // 1;
    # A reference to a hash of the volumes mapped to clusters and vserver.
    my $available_volumes      = $args{available_volumes};

    # These are the filters for the method.
    # CLUSTER FILTERS
    my $must_use_clusters      = $args{must_use_clusters};
    my $unuseable_clusters     = $args{unuseable_clusters};
    my $last_resort_clusters   = $args{last_resort_clusters};
    # VSERVER FILTERS
    my $must_use_vservers      = $args{must_use_vservers};
    my $unuseable_vservers     = $args{unuseable_vservers};
    my $last_resort_vservers   = $args{last_resort_vservers};
    # AGGREGATE FILTERS
    my $must_use_aggregates    = $args{must_use_aggregates};
    my $unuseable_aggregates   = $args{unuseable_aggregates};
    my $last_resort_aggregates = $args{last_resort_aggregates};


    my $all_from_same_vserver  = $args{all_from_same_vserver} // 0;

    my @return_volumes = ();

    my $filtered_hash = {};
    ###############################################################################
    # The following section of code will build the filtered_hash.
    # The filtered hash is built in the following format based on passed in filters.
    # {
    # useable->cluster_name->{useable->vserver_name->useable->[array of volumes],
    #                                                unuseable->[array of volumes],
    #                                                last_resort->[array of volumes],
    #                         unuseable->vserver_name->useable->[array of volumes],
    #                                                  unuseable->[array of volumes],
    #                                                  last_resort->[array of volumes],
    #                         last_resort->vserver_name->useable->[array of volumes],
    #                                                    unuseable->[array of volumes],
    #                                                    last_resort->[array of volumes],
    # last_resort->cluster_name->{useable->vserver_name->useable->[array of volumes],
    #                                                    unuseable->[array of volumes],
    #                                                    last_resort->[array of volumes],
    #                             unuseable->vserver_name->useable->[array of volumes],
    #                                                      unuseable->[array of volumes],
    #                                                      last_resort->[array of volumes],
    #                             last_resort->vserver_name->useable->[array of volumes],
    #                                                        unuseable->[array of volumes],
    #                                                        last_resort->[array of volumes],
    # unuseable->cluster_name->{useable->vserver_name->useable->[array of volumes],
    #                                                  unuseable->[array of volumes],
    #                                                  last_resort->[array of volumes],
    #                           unuseable->vserver_name->useable->[array of volumes],
    #                                                    unuseable->[array of volumes],
    #                                                    last_resort->[array of volumes],
    #                           last_resort->vserver_name->useable->[array of volumes],
    #                                                      unuseable->[array of volumes],
    #                                                      last_resort->[array of volumes],
    #}
    ###############################################################################
    while (scalar @{$available_volumes}) {
        my $current_volume = shift @{$available_volumes};
        my $current_cluster = $current_volume->get_attribute('cluster_name');
        my $current_vserver = $current_volume->get_attribute('vserver');
        my $current_aggregate = $current_volume->get_attribute('aggregate');
        my $cluster_useability;
        my $vserver_useability;
        my $volume_useability;
        if (defined $must_use_clusters ) {
            if(grep(/^$current_cluster$/,@{$must_use_clusters})) {
                $cluster_useability = 'useable';
            } else {
                $cluster_useability = 'unuseable';
            }
        } elsif (defined $unuseable_clusters and grep(/^$current_cluster$/,@{$unuseable_clusters})) {
            $cluster_useability = 'unuseable';
        } elsif (defined $last_resort_clusters and grep(/^$current_cluster$/,@{$last_resort_clusters})) {
            $cluster_useability = 'last_resort';
        } else {
            $cluster_useability = 'useable';
        }

        if (defined $must_use_vservers ) {
            if(grep(/^$current_vserver$/,@{$must_use_vservers})) {
                $vserver_useability = 'useable';
            } else {
                $vserver_useability = 'unuseable';
            }
        } elsif (defined $unuseable_vservers and grep(/^$current_vserver$/,@{$unuseable_vservers})) {
            $vserver_useability = 'unuseable';
        } elsif (defined $last_resort_vservers and grep(/^$current_vserver$/,@{$last_resort_vservers})) {
            $vserver_useability = 'last_resort';
        } else {
            $vserver_useability = 'useable';
        }

        if (defined $must_use_aggregates ) {
            if(grep(/^$current_volume$/,@{$must_use_aggregates})) {
                $volume_useability = 'useable';
            } else {
                $volume_useability = 'unuseable';
            }
        } elsif (defined $unuseable_aggregates and grep(/^$current_volume$/,@{$unuseable_aggregates})) {
            $volume_useability = 'unuseable';
        } elsif (defined $last_resort_aggregates and grep(/^$current_volume$/,@{$last_resort_aggregates})) {
            $volume_useability = 'last_resort';
        } else {
            $volume_useability = 'useable';
        }

        push(@{$filtered_hash->{$cluster_useability}->{$current_cluster}->{$vserver_useability}->{$current_vserver}->{$volume_useability}}, $current_volume);
    }

    ###############################################################################
    # The following code will search through the filter hash for the best volumes
    # to use.
    # CU = Cluster Useable
    # VSU = Vserver Useable
    # VU  = Volume Useable
    # CL = Cluster Last Resort
    # VSL = Vserver Last Resort
    # VL  = Volume Resort
    # The search happens in the following order.
    # CU VSU VU
    # CU VSU VL
    # CU VSL VU
    # CU VSL VL
    # CL VSU VU
    # CL VSU VL
    # CL VSL VU
    # CL VSL VL
    ###############################################################################

    # Look for cluster useable, vserver useable, volume useable.
    if (scalar (@return_volumes) < $volume_count and
        defined $filtered_hash->{useable}) {
        my @clusters = keys (%{$filtered_hash->{useable}});
        # Look for vserver useable, volume useable
        foreach my $cluster (@clusters) {
            if (defined $filtered_hash->{useable}->{$cluster}->{useable}) {
                my @vservers = keys (%{$filtered_hash->{useable}->{$cluster}->{useable}});
                # Look for volume useable.
                foreach my $vserver (@vservers) {
                    if($all_from_same_vserver) {
                        # If we want all from same vserver and there are not enough useable then we skip this vserver for now.
                        next if (scalar @{$filtered_hash->{useable}->{$cluster}->{useable}->{$vserver}->{useable}} < $volume_count);
                        # If there are enough volumes then add the useable volumes
                        while ((scalar @{$filtered_hash->{useable}->{$cluster}->{useable}->{$vserver}->{useable}}) and
                               (scalar (@return_volumes) < $volume_count)) {
                            push (@return_volumes, (shift @{$filtered_hash->{useable}->{$cluster}->{useable}->{$vserver}->{useable}}));
                        }
                    # Otherwise we will just start adding volumes from the useable space of this vserver.
                    } else {
                        while ((scalar @{$filtered_hash->{useable}->{$cluster}->{useable}->{$vserver}->{useable}}) and
                               (scalar (@return_volumes) < $volume_count)) {
                            push (@return_volumes, (shift @{$filtered_hash->{useable}->{$cluster}->{useable}->{$vserver}->{useable}}));
                        }
                    }
                }
            }
        }
    }

    # Look for cluster useable, vserver useable, volume last_resort.
    if (scalar (@return_volumes) < $volume_count and
        defined $filtered_hash->{useable}) {
        my @clusters = keys (%{$filtered_hash->{useable}});
        # Look for vserver useable, volume last_resort
        foreach my $cluster (@clusters) {
            if (defined $filtered_hash->{useable}->{$cluster}->{useable}) {
                my @vservers = keys (%{$filtered_hash->{useable}->{$cluster}->{useable}});
                # Look for volume last_resort.
                foreach my $vserver (@vservers) {
                    if($all_from_same_vserver) {
                        # If we want all from same vserver and there are not enough useable and last_resort then we skip this vserver
                        next if (scalar @{$filtered_hash->{useable}->{$cluster}->{useable}->{$vserver}->{useable}} +
                                 scalar @{$filtered_hash->{useable}->{$cluster}->{useable}->{$vserver}->{last_resort}}  < $volume_count);
                        # Add all the useable volumes first, then the last resorts.
                        while ((scalar @{$filtered_hash->{useable}->{$cluster}->{useable}->{$vserver}->{useable}}) and
                               (scalar (@return_volumes) < $volume_count)) {
                            push (@return_volumes, (shift @{$filtered_hash->{useable}->{$cluster}->{useable}->{$vserver}->{useable}}));
                        }
                        while ((scalar @{$filtered_hash->{useable}->{$cluster}->{useable}->{$vserver}->{last_resort}}) and
                               (scalar (@return_volumes) < $volume_count)) {
                            push (@return_volumes, (shift @{$filtered_hash->{useable}->{$cluster}->{useable}->{$vserver}->{last_resort}}));
                        }
                    # Otherwise we will just start adding volumes from the useable space of this vserver.
                    } else {
                        while ((scalar @{$filtered_hash->{useable}->{$cluster}->{useable}->{$vserver}->{last_resort}}) and
                               (scalar (@return_volumes) < $volume_count)) {
                            push (@return_volumes, (shift @{$filtered_hash->{useable}->{$cluster}->{useable}->{$vserver}->{last_resort}}));
                        }
                    }
                }
            }
        }
    }

    # Look for cluster useable, vserver last_resort, volume useable.
    if (scalar (@return_volumes) < $volume_count and
        defined $filtered_hash->{useable}) {
        my @clusters = keys (%{$filtered_hash->{useable}});
        # Look for vserver useable, volume useable
        foreach my $cluster (@clusters) {
            if (defined $filtered_hash->{useable}->{$cluster}->{last_resort}) {
                my @vservers = keys (%{$filtered_hash->{useable}->{$cluster}->{last_resort}});
                # Look for volume useable.
                foreach my $vserver (@vservers) {
                    if($all_from_same_vserver) {
                        # If we want all from same vserver and there are not enough useable then we skip this vserver for now.
                        next if (scalar @{$filtered_hash->{useable}->{$cluster}->{last_resort}->{$vserver}->{useable}} < $volume_count);
                        # If there are enough volumes then add the useable volumes
                        while ((scalar @{$filtered_hash->{useable}->{$cluster}->{last_resort}->{$vserver}->{useable}}) and
                               (scalar (@return_volumes) < $volume_count)) {
                            push (@return_volumes, (shift @{$filtered_hash->{useable}->{$cluster}->{last_resort}->{$vserver}->{useable}}));
                        }
                    # Otherwise we will just start adding volumes from the useable space of this vserver.
                    } else {
                        while ((scalar @{$filtered_hash->{useable}->{$cluster}->{last_resort}->{$vserver}->{useable}}) and
                               (scalar (@return_volumes) < $volume_count)) {
                            push (@return_volumes, (shift @{$filtered_hash->{useable}->{$cluster}->{last_resort}->{$vserver}->{useable}}));
                        }
                    }
                }
            }
        }
    }

    # Look for cluster useable, vserver last_resort, volume last_resort.
    if (scalar (@return_volumes) < $volume_count and
        defined $filtered_hash->{useable}) {
        my @clusters = keys (%{$filtered_hash->{useable}});
        # Look for vserver useable, volume last_resort
        foreach my $cluster (@clusters) {
            if (defined $filtered_hash->{useable}->{$cluster}->{last_resort}) {
                my @vservers = keys (%{$filtered_hash->{useable}->{$cluster}->{last_resort}});
                # Look for volume last_resort.
                foreach my $vserver (@vservers) {
                    if($all_from_same_vserver) {
                        # If we want all from same vserver and there are not enough useable and last_resort then we skip this vserver
                        next if (scalar @{$filtered_hash->{useable}->{$cluster}->{last_resort}->{$vserver}->{useable}} +
                                 scalar @{$filtered_hash->{useable}->{$cluster}->{last_resort}->{$vserver}->{last_resort}}  < $volume_count);
                        # Add all the useable volumes first, then the last resorts.
                        while ((scalar @{$filtered_hash->{useable}->{$cluster}->{last_resort}->{$vserver}->{useable}}) and
                               (scalar (@return_volumes) < $volume_count)) {
                            push (@return_volumes, (shift @{$filtered_hash->{useable}->{$cluster}->{last_resort}->{$vserver}->{useable}}));
                        }
                        while ((scalar @{$filtered_hash->{useable}->{$cluster}->{last_resort}->{$vserver}->{last_resort}}) and
                               (scalar (@return_volumes) < $volume_count)) {
                            push (@return_volumes, (shift @{$filtered_hash->{useable}->{$cluster}->{last_resort}->{$vserver}->{last_resort}}));
                        }
                    # Otherwise we will just start adding volumes from the useable space of this vserver.
                    } else {
                        while ((scalar @{$filtered_hash->{useable}->{$cluster}->{last_resort}->{$vserver}->{last_resort}}) and
                               (scalar (@return_volumes) < $volume_count)) {
                            push (@return_volumes, (shift @{$filtered_hash->{useable}->{$cluster}->{last_resort}->{$vserver}->{last_resort}}));
                        }
                    }
                }
            }
        }
    }

    # Look for cluster last_resort, vserver useable, volume useable.
    if (scalar (@return_volumes) < $volume_count and
        defined $filtered_hash->{last_resort}) {
        my @clusters = keys (%{$filtered_hash->{last_resort}});
        # Look for vserver useable, volume useable
        foreach my $cluster (@clusters) {
            if (defined $filtered_hash->{last_resort}->{$cluster}->{useable}) {
                my @vservers = keys (%{$filtered_hash->{last_resort}->{$cluster}->{useable}});
                # Look for volume useable.
                foreach my $vserver (@vservers) {
                    if($all_from_same_vserver) {
                        # If we want all from same vserver and there are not enough useable then we skip this vserver for now.
                        next if (scalar @{$filtered_hash->{last_resort}->{$cluster}->{useable}->{$vserver}->{useable}} < $volume_count);
                        # If there are enough volumes then add the useable volumes
                        while ((scalar @{$filtered_hash->{last_resort}->{$cluster}->{useable}->{$vserver}->{useable}}) and
                               (scalar (@return_volumes) < $volume_count)) {
                            push (@return_volumes, (shift @{$filtered_hash->{last_resort}->{$cluster}->{useable}->{$vserver}->{useable}}));
                        }
                    # Otherwise we will just start adding volumes from the useable space of this vserver.
                    } else {
                        while ((scalar @{$filtered_hash->{last_resort}->{$cluster}->{useable}->{$vserver}->{useable}}) and
                               (scalar (@return_volumes) < $volume_count)) {
                            push (@return_volumes, (shift @{$filtered_hash->{last_resort}->{$cluster}->{useable}->{$vserver}->{useable}}));
                        }
                    }
                }
            }
        }
    }

    # Look for cluster last_resort, vserver useable, volume last_resort.
    if (scalar (@return_volumes) < $volume_count and
        defined $filtered_hash->{last_resort}) {
        my @clusters = keys (%{$filtered_hash->{last_resort}});
        # Look for vserver useable, volume last_resort
        foreach my $cluster (@clusters) {
            if (defined $filtered_hash->{last_resort}->{$cluster}->{useable}) {
                my @vservers = keys (%{$filtered_hash->{last_resort}->{$cluster}->{useable}});
                # Look for volume last_resort.
                foreach my $vserver (@vservers) {
                    if($all_from_same_vserver) {
                        # If we want all from same vserver and there are not enough useable and last_resort then we skip this vserver
                        next if (scalar @{$filtered_hash->{last_resort}->{$cluster}->{useable}->{$vserver}->{useable}} +
                                 scalar @{$filtered_hash->{last_resort}->{$cluster}->{useable}->{$vserver}->{last_resort}}  < $volume_count);
                        # Add all the useable volumes first, then the last resorts.
                        while ((scalar @{$filtered_hash->{last_resort}->{$cluster}->{useable}->{$vserver}->{useable}}) and
                               (scalar (@return_volumes) < $volume_count)) {
                            push (@return_volumes, (shift @{$filtered_hash->{last_resort}->{$cluster}->{useable}->{$vserver}->{useable}}));
                        }
                        while ((scalar @{$filtered_hash->{last_resort}->{$cluster}->{useable}->{$vserver}->{last_resort}}) and
                               (scalar (@return_volumes) < $volume_count)) {
                            push (@return_volumes, (shift @{$filtered_hash->{last_resort}->{$cluster}->{useable}->{$vserver}->{last_resort}}));
                        }
                    # Otherwise we will just start adding volumes from the useable space of this vserver.
                    } else {
                        while ((scalar @{$filtered_hash->{last_resort}->{$cluster}->{useable}->{$vserver}->{last_resort}}) and
                               (scalar (@return_volumes) < $volume_count)) {
                            push (@return_volumes, (shift @{$filtered_hash->{last_resort}->{$cluster}->{useable}->{$vserver}->{last_resort}}));
                        }
                    }
                }
            }
        }
    }

    # Look for cluster last_resort, vserver last_resort, volume useable.
    if (scalar (@return_volumes) < $volume_count and
        defined $filtered_hash->{last_resort}) {
        my @clusters = keys (%{$filtered_hash->{last_resort}});
        # Look for vserver useable, volume useable
        foreach my $cluster (@clusters) {
            if (defined $filtered_hash->{last_resort}->{$cluster}->{last_resort}) {
                my @vservers = keys (%{$filtered_hash->{last_resort}->{$cluster}->{last_resort}});
                # Look for volume useable.
                foreach my $vserver (@vservers) {
                    if($all_from_same_vserver) {
                        # If we want all from same vserver and there are not enough useable then we skip this vserver for now.
                        next if (scalar @{$filtered_hash->{last_resort}->{$cluster}->{last_resort}->{$vserver}->{useable}} < $volume_count);
                        # If there are enough volumes then add the useable volumes
                        while ((scalar @{$filtered_hash->{last_resort}->{$cluster}->{last_resort}->{$vserver}->{useable}}) and
                               (scalar (@return_volumes) < $volume_count)) {
                            push (@return_volumes, (shift @{$filtered_hash->{last_resort}->{$cluster}->{last_resort}->{$vserver}->{useable}}));
                        }
                    # Otherwise we will just start adding volumes from the useable space of this vserver.
                    } else {
                        while ((scalar @{$filtered_hash->{last_resort}->{$cluster}->{last_resort}->{$vserver}->{useable}}) and
                               (scalar (@return_volumes) < $volume_count)) {
                            push (@return_volumes, (shift @{$filtered_hash->{last_resort}->{$cluster}->{last_resort}->{$vserver}->{useable}}));
                        }
                    }
                }
            }
        }
    }

    # Look for cluster last_resort, vserver last_resort, volume last_resort.
    if (scalar (@return_volumes) < $volume_count and
        defined $filtered_hash->{last_resort}) {
        my @clusters = keys (%{$filtered_hash->{last_resort}});
        # Look for vserver useable, volume last_resort
        foreach my $cluster (@clusters) {
            if (defined $filtered_hash->{last_resort}->{$cluster}->{last_resort}) {
                my @vservers = keys (%{$filtered_hash->{last_resort}->{$cluster}->{last_resort}});
                # Look for volume last_resort.
                foreach my $vserver (@vservers) {
                    if($all_from_same_vserver) {
                        # If we want all from same vserver and there are not enough useable and last_resort then we skip this vserver
                        next if (scalar @{$filtered_hash->{last_resort}->{$cluster}->{last_resort}->{$vserver}->{useable}} +
                                 scalar @{$filtered_hash->{last_resort}->{$cluster}->{last_resort}->{$vserver}->{last_resort}}  < $volume_count);
                        # Add all the useable volumes first, then the last resorts.
                        while ((scalar @{$filtered_hash->{last_resort}->{$cluster}->{last_resort}->{$vserver}->{useable}}) and
                               (scalar (@return_volumes) < $volume_count)) {
                            push (@return_volumes, (shift @{$filtered_hash->{last_resort}->{$cluster}->{last_resort}->{$vserver}->{useable}}));
                        }
                        while ((scalar @{$filtered_hash->{last_resort}->{$cluster}->{last_resort}->{$vserver}->{last_resort}}) and
                               (scalar (@return_volumes) < $volume_count)) {
                            push (@return_volumes, (shift @{$filtered_hash->{last_resort}->{$cluster}->{last_resort}->{$vserver}->{last_resort}}));
                        }
                    # Otherwise we will just start adding volumes from the useable space of this vserver.
                    } else {
                        while ((scalar @{$filtered_hash->{last_resort}->{$cluster}->{last_resort}->{$vserver}->{last_resort}}) and
                               (scalar (@return_volumes) < $volume_count)) {
                            push (@return_volumes, (shift @{$filtered_hash->{last_resort}->{$cluster}->{last_resort}->{$vserver}->{last_resort}}));
                        }
                    }
                }
            }
        }
    }

    if (scalar (@return_volumes) < $volume_count){
        NATE::BaseException->throw("Could not find an available volume for snapmirroring that fit the following criteria.\n" .
                                   "volume_count = " . $volume_count . "\n" .
                                   "all_from_same_vserver = " . $all_from_same_vserver . "\n" .
                                   "must_use_clusters = \n" . dumper($must_use_clusters) . "\n" .
                                   "must_use_vservers = \n" . dumper($must_use_vservers) . "\n" .
                                   "unuseable_clusters = \n" . dumper($unuseable_clusters) . "\n" .
                                   "unuseable_vservers = \n" . dumper($unuseable_vservers) . "\n" .
                                   "last_resort_clusters = \n" . dumper($last_resort_clusters) . "\n" .
                                   "last_resort_vservers = \n" . dumper($last_resort_vservers) . "\n");
    }

    ###############################################################################
    # This code rebuilds the available volumes array from what is left in the filtered hash.
    # The volumes that we return from this method should no longer exist in the filtered_hash
    # or the available volumes array
    ###############################################################################
    foreach my $cluster_useability (keys (%{$filtered_hash})) {
        foreach my $cluster (keys (%{$filtered_hash->{$cluster_useability}})) {
            foreach my $vserver_useability (keys (%{$filtered_hash->{$cluster_useability}->{$cluster}})) {
                foreach my $vserver (keys (%{$filtered_hash->{$cluster_useability}->{$cluster}->{$vserver_useability}})) {
                    foreach my $volume_useability (keys (%{$filtered_hash->{$cluster_useability}->{$cluster}->{$vserver_useability}->{$vserver}})) {
                        push(@{$available_volumes}, @{$filtered_hash->{$cluster_useability}
                                                                                  ->{$cluster}
                                                                                  ->{$vserver_useability}
                                                                                  ->{$vserver}->{$volume_useability}});
                    }
                }
            }
        }
    }

    return (wantarray()) ? @return_volumes : \@return_volumes;
}

1;

__END__

=head1 NAME

PTEFIT::Xgen::Config::Example

=head1 AUTHOR/MAINTAINER

=over

=item PTEFIT AUTOMATION (ng-ptefit-automation-core@netapp.com)

=back

   Last revised $Date: $ by $Author: $

