#
# Copyright (c) 2015 NetApp, Inc., All Rights Reserved.
#   Any other use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

package NACL::AppBuilder::Xgen::Config::STE::EQ::Base_STE;

use parent qw(NACL::AppBuilder::Xgen::Config::STE::EQ::Base);

use List::Util 'shuffle';

use Data::Dumper;

# Default Base-line constants
use constant TOTAL_VS_COUNT => 30;

# The layout of VS_NODAL_LAYOUT.
# { node_count => vserver_count,
#   node_count => vserver_count,}
use constant VS_NODAL_LAYOUT => {
    6 => 5,
    4 => 5,
    2 => 5,
    1 => 15

};

use constant DISKS_PER_AGGR => 14;
use constant AGGRS_PER_NODE => 1;

use constant TOTAL_IPV4_LIFS_PER_VS_PER_NODE_PER_FABRIC => 1;
use constant TOTAL_IPV6_LIFS_PER_VS_PER_NODE_PER_FABRIC => 1;

use constant VSERVER_COUNT_CLUSTER_WIDE => 5;

use constant TOTAL_VOLUME_COUNT                  => 3000;
use constant VOL_SIZE                            => '400g';
use constant VOL_MAX_AUTOSIZE_MULTIPLIER         => 3;
use constant VOL_AUTOSIZE_MODE                   => "grow";
use constant VOL_AUTOSIZE_GROW_THRESHOLD_PERCENT => 60;

use constant TOTAL_QTREE_COUNT                => 30000;
use constant CIFS_SECURITY_TRACING_VS_PERCENT => 100;

use constant TOTAL_QTREE_HOMEDIR_COUNT => 3000;
use constant QTREE_HOMEDIR_LAYOUT      => {
    local  => 5,
    domain => 95
};

use constant TOTAL_QOS_POLICY_COUNT => 30;

use constant TOTAL_CIFS_SHARE_COUNT => 6000;
use constant CIFS_SHARE_LAYOUT      => {
    qtree  => 50,
    volume => 50
};

use constant TOTAL_SNAPMIRROR_COUNT => 1000;
use constant SNAPMIRROR_LAYOUT      => {
    "load_sharing"    => 0,
    "vault"           => 30,
    "data_protection" => 70
};

use constant CIFS_SECURITY_AES_ENCRYPTION_VS_PERCENT => 10;
use constant CIFS_SECURITY_SMB_SIGNING_VS_PERCENT    => 25;

use constant AUDIT_VS_PERCENT => 10;
use constant DDNS_VS_PERCENT  => 50;

use constant TOTAL_SYMLINK_COUNT                   => 9000;
use constant CIFS_OPTIONS_NODE_REFERRAL_VS_PERCENT => 10;
use constant CIFS_OPTIONS_DAC_VS_PERCENT           => 25;

use constant SYMLINK_LAYOUT => {
    qtree  => 90,
    volume => 10,
};

use constant TOTAL_LOCAL_USERS    => 300;
use constant TOTAL_DOMAIN_USERS   => 99;
use constant TOTAL_LOCAL_GROUPS   => 3000;
use constant MAX_USER_MEMBERSHIPS => 3000;

use constant TOTAL_DOMAIN_USER_NAME_MAPPINGS => 10;
use constant DOMAIN_USER_NAME_MAPPING_LAYOUT => {
    "local"   => 10,    # domain user will have a mapping to a local user in the vserver (direction: win-unix)
    "default" => 10,    # domain user will have a mapping to 'root' (direction: win-unix)
    "domain"  => 80     # domain user will have a mapping to themselves in both directions (direction: unix-win/win-unix)
};

use constant CIFS_NETBIOS_ALIASES_PER_VS => 10;

use constant DOMAIN_USER_PREFIX => "STaaS1u";

use constant LOCAL_GROUP_MEMBERSHIP_LAYOUT => {
    local  => 50,
    domain => 50
};

# 50% of vservers will have ntfs, 50% will have mixed
use constant VSERVER_ROOT_VOLUME_SECURITY_LAYOUT => {
    ntfs  => 50,
    mixed => 50
};

# all volumes will have vol efficiency enabled
use constant VOL_EFFICIENCY_PERCENT => 100;

use constant VSERVER_FSECURITY_ACCESS_CONTROL_LAYOUT => {
    slag             => 10,
    'file-directory' => 90
};

use constant TOTAL_LOCAL_USER_SACL_PER_NTFS  => 10;
use constant TOTAL_DOMAIN_USER_SACL_PER_NTFS => 10;
use constant TOTAL_LOCAL_USER_DACL_PER_NTFS  => 10;
use constant TOTAL_DOMAIN_USER_DACL_PER_NTFS => 10;

use constant SECURITY_FILTER_USERS_PER_VSERVER => 5;


use constant SAN_VSERVERS_COUNT => 1;
use constant PERCENTAGE_OF_VOLUMES_WITH_LUNS => 100;
use constant TOTAL_LUN_COUNT => 32;
use constant LUN_SIZE => "20m";

use strict;
use warnings;

use NATE::Log qw(log_global);
use Tharn qw($Log );
use NATE::BaseException;
use NATE::Exceptions::Argument qw(:try);

# resource libs
use POSIX qw(ceil floor);

sub build {
    my ($self, %args) = @_;
    #
    $Log->comment(Dumper(\%args));
    my $cifs_setup = delete($args{cifs_setup}) // 0;
    $self->SUPER::build(%args);

    log_global()->log("Generating aggregates");
    $self->generate_aggregate_resources(%args);

    # make vserver resources
    log_global()->log("Generating vservers");
    $self->generate_vserver_resources(%args);
    
    log_global()->log("Generating qos policy groups");
    $self->generate_qos_policy_resources(%args);
    
    # Call to genereate iscsi service resources.
    log_global()->log("Generating iscsi_services");
    $self->generate_iscsi_services_resources(%args);
    
    # Call to generate iscsi lif resources.
    log_global()->log("Generating iscsi_lif");
    $self->generate_iscsi_lif_resources(%args);
    
    # Call to genereate iscsi service resources.
    log_global()->log("Generating igroup");
    $self->generate_igroup_resources(%args);

    # Call to genereate iscsi service resources.
    log_global()->log("Generating igroup_initiator");
    $self->generate_igroup_initiator_resources(%args);

    log_global()->log("Generating vols");
    $self->generate_volume_resources(%args);

    log_global()->log("Generating lifs");
    $self->generate_lif_resources(%args);

    log_global()->log("Generating network routes");
    $self->generate_network_route_resources(%args);

    log_global()->log("Generating vol efficiency");
    $self->generate_volume_efficiency_resources(%args);

    log_global()->log("Generating export policies");
    $self->generate_export_policy_resources(%args);

    log_global()->log("Generating export policy rules");
    $self->generate_export_policy_rule_resources(%args);

    log_global()->log("Generating qtree");
    $self->generate_qtree_resources(%args);

    log_global()->log("Generating cifs_shares");
    $self->generate_cifs_share_resources(%args);

    log_global()->log("Generating root user mapping");
    $self->generate_root_user_name_mapping_resources(%args);

    log_global()->log("Generating snapmirrors");
    $self->generate_snapmirror_resources(%args);
    
    # Call to generate lun resources.
    $self->generate_lun_resources(%args);

    # Call to generate lun resources.
    $self->generate_lun_map_resources(%args);

    if ($cifs_setup) {
        log_global()->log("Generating cifs users/groups");
        $self->generate_cifs_user_and_group_resources(%args);

        log_global()->log("Generating cifs domain user name mappings");
        $self->generate_cifs_domain_user_name_mapping_resources(%args);

        log_global()->log("Generating cifs group policies");
        $self->generate_cifs_group_policy_resources(%args);

        log_global()->log("Generating cifs_options");
        $self->generate_cifs_options_resources(%args);

        log_global()->log("Generating Audits");
        $self->generate_audit_resources(%args);

        log_global()->log("Generating symlinks");
        $self->generate_symlink_resources(%args);

        log_global()->log("Generating cifs security");
        $self->generate_cifs_security_resources(%args);

        log_global()->log("Generating fsecurity");
        $self->generate_fsecurity_resources(%args);

        log_global()->log("Generating cifs branchcache");
        $self->generate_cifs_branchcache_resources(%args);

        log_global()->log("Generating cifs home directory search paths");
        $self->generate_cifs_homedirsearchpath_resources(%args);

        log_global()->log("Generating cifs security tracing");
        $self->generate_cifs_security_trace_filter_resources(%args);
    }

    #    log_global()->log("Generating DDNS");
    #    $self->generate_dynamic_dns_resources(%args);
} ## end sub build

1;

__END__

=head1 NAME

PTEFIT::Xgen::Config::ONTAP::CIFS::Base

=head1 AUTHOR/MAINTAINER

=over

=item PTEFIT AUTOMATION (ng-ptefit-automation-core@netapp.com)

=back

   Last revised $Date: $ by $Author: $

=cut

