use strict;
use warnings;

package NACL::AppBuilder::Xgen::Config::dt::Base;

use parent 'NACL::AppBuilder::Xgen::Config::ONTAP';

use constant TOTAL_VS_COUNT => 1;
use constant TOTAL_NAS_VOLUME_COUNT => 2;
use constant TOTAL_LUN_COUNT => 4;
use constant TOTAL_SAN_VOLUME_COUNT => 2;

use constant VOLUME_SIZE => "40%";  # percentage of aggregate
use constant LUN_SIZE => "40%";  # percentage of volume

use NATE::Log qw(log_global);
use NATE::BaseException;
use NATE::Exceptions::Argument qw(:try);

# resource libs
use NACL::AppBuilder::Xgen::Resource::Vserver;
use NACL::AppBuilder::Xgen::Resource::Volume;
use NACL::AppBuilder::Xgen::Resource::Dns;
use NACL::AppBuilder::Xgen::Resource::Lif;
use NACL::AppBuilder::Xgen::Resource::NetworkRoute;
use NACL::AppBuilder::Xgen::Resource::Cifs;
use NACL::AppBuilder::Xgen::Resource::Cifs::Share;
use NACL::AppBuilder::Xgen::Resource::Lun;
use NACL::AppBuilder::Xgen::Resource::Nfs;
use NACL::AppBuilder::Xgen::Resource::Iscsi;
use NACL::AppBuilder::Xgen::Resource::Igroup;
use NACL::AppBuilder::Xgen::Resource::Igroup::Initiator;
use NACL::AppBuilder::Xgen::Resource::LunMap;


use NACL::Service::Xgen::Util qw(convert_percent_layout_to_count_layout
                                 get_next_layout_element
                                 dumper
                                 positive_filter
                                 randomize_array);

sub build {
    my ($self, %args) = @_;

    # We only want to do this if something did not pass in any Ethernet fabric definitions.
    # This fabric will include all the ips from all nodes and not sort specific to clusters.
    # This means that IPs will not be owned by a specific cluster and can be used by any cluster.
    unless (scalar keys %{$args{eth_fabric_data}}) {
        log_global()->comment("Generating Ethernet Fabric data from Natehost files.");
        $args{eth_fabric_data} = $self->generate_eth_fabric_data_from_natehost_files(%args);
    }

    log_global()->comment("Generating Parent Configuration.");
    $self->SUPER::build(%args);

    log_global()->comment("Adding CIFS and NFS protocols to all clients.");
    $self->generate_client_protocols(%args);

    log_global()->comment("Generating Vservers.");
    $self->generate_vserver_resources(%args);

    log_global()->comment("Generating ISCSI Lifs.");
    $self->generate_iscsi_lif_resources(%args);

    log_global()->comment("Generating NFS Lifs.");
    $self->generate_nfs_lif_resources(%args);

    log_global()->comment("Generating Igroups.");
    $self->generate_igroup_resources(%args);

    log_global()->comment("Generating Initiators.");
    $self->generate_igroup_initiator_resources(%args);

    log_global()->comment("Generating ISCSI Services.");
    $self->generate_iscsi_services_resources(%args);

    log_global()->comment("Generating NFS Services.");
    $self->generate_nfs_services_resources(%args);

    log_global()->comment("Generating Volumes.");
    $self->generate_volume_resources(%args);

    log_global()->comment("Generating Luns.");
    $self->generate_lun_resources(%args);

    log_global()->comment("Generating Lun Maps.");
    $self->generate_lun_map_resources(%args);

}

sub generate_client_protocols {
    my ($self, %args) = @_;

    my $client_resources = $self->get_resources(type => "NACL::AppBuilder::Xgen::Resource::Client");
    my @clients = @{$client_resources};

    foreach my $client (@clients) {
        my $protocols = $client->get_attribute("protocols");
        my $protocol = ($client->get_attribute('ostype') =~ /windows/) ? "cifs" : "nfs";
        push($protocols, $protocol);
        $client->set_private_attributes('protocols' => $protocols);
    }
    return;
}

sub generate_eth_fabric_data_from_natehost_files {
    my ($self, %args) = @_;

    my @NODES = NACL::C::Node->find();
    my @ips;
    my $netmask = undef;
    my $gateway = undef;
    foreach my $node (@NODES) {
        my $hostrec = Hostrec->new(id=>$node->node());
        my %networks = $hostrec->param_fields();
        foreach my $key (keys %networks) {
            if($key =~ /DATA_IP/i) {
                push (@ips, $networks{$key});
            } elsif ($key =~ /DATA_GATEWAY/i) {
                $gateway = $networks{$key};
            } elsif ($key =~ /DATA_MASK/i) {
                $netmask = $networks{$key};
            }
        }
    }
    my %eth_fabric_data = ();
    $eth_fabric_data{'eth1'} = {
        'ipspace'           => 'Default',
        'broadcast_domain'  => 'Default',
        'address_range_v4'  => \@ips,
        'netmask_v4'        => $netmask,
        'gateway_v4'        => $gateway,
    };

    log_global()->comment("The fabric data that was generated is " . dumper(\%eth_fabric_data));
    return \%eth_fabric_data
}

###############################################################################
# DESCRIPTION: Generates vserver resources and adds them to the configuration.
#              Generates a number of vservers based on the TOTAL_VS_COUNT
#              constant. The vservers will be named vs# where # starts at 0 and
#              counts up to TOTAL_VS_COUNT-1. The vservers will alternate
#              between the clusters provided and inside each cluster will
#              alternate between the aggregates in that cluster that are part
#              of the configuration.
# INPUT: NONE
# OUTPUT: NONE
# Exceptions: None thrown directly by this method.
###############################################################################
sub generate_vserver_resources {
    my ($self, %args) = @_;

    # Get the clusters so that we can alternate vserver creation between multiple clusters.
    my $cluster_resources = $self->get_resources(type => "NACL::AppBuilder::Xgen::Resource::Cluster");
    my @clusters = @{$cluster_resources};
    my %cluster_aggregate_hash = ();
    foreach my $cluster_resource (@$cluster_resources) {
        @{$cluster_aggregate_hash{$cluster_resource->get_attribute('cluster_name')}} =
                $cluster_resource->get_descendants(type => "NACL::AppBuilder::Xgen::Resource::Aggregate");
        # {clus1 => [aggr1_obj,aggr2_obj], clus2 => [aggr1_obj,aggr2_obj]}
    }

    my @vservers;

    for (my $i = 0; $i < $self->TOTAL_VS_COUNT; $i++) {
        my $cluster = shift(@clusters);
        push (@clusters, $cluster);

        my $aggr = shift(@{$cluster_aggregate_hash{$cluster->get_attribute('cluster_name')}});
        push(@{$cluster_aggregate_hash{$cluster->get_attribute('cluster_name')}}, $aggr);

        my $vs_res = NACL::AppBuilder::Xgen::Resource::Vserver->new(vserver =>"vs$i",
                                                                    aggregate => $aggr->get_attribute("aggregate"));

        # connect the vs resource to the aggr's cluster parent resource
        $vs_res->add_parent_resources(resources => [$aggr]);

        push(@vservers, $vs_res);
    }
    $self->add_resources(resources => \@vservers);
}

###############################################################################
# DESCRIPTION: Generates volume resources and adds them to the configuration.
#              Generates a number of volumes based on the TOTAL_VOLUME_COUNT
#              constant. The volumes will be named Vol# where # starts at 0 and
#              counts up to TOTAL_VOLUME_COUNT-1. The volumes will alternate
#              between the clusters provided and inside each cluster will
#              alternate between the aggregates and vservers in that cluster
#              that are part of the configuration. The
#              VOLUME_VM_ALIGN_SUFFIX_PERCENT_LAYOUT will control the
#              percentage of the total volumes that have each of the
#              vm-align-suffix attributes defined in the layout. The
#              VOLUME_VM_ALIGN_SECTOR_PERCENT_LAYOUT will control the
#              percentage of the volumes with vm-align-sufix that have
#              each of the vm-align-sector attributes defined in the layout.
#              The volume size is controlled by the VOLUME_SIZE constant.
# INPUT: NONE
# OUTPUT: NONE
# Exceptions: None thrown directly by this method.
###############################################################################
sub generate_volume_resources {
    my ($self, %args) = @_;
    # Get the clusters so that we can alternate vserver creation between multiple clusters.
    my $cluster_resources = $self->get_resources(type => "NACL::AppBuilder::Xgen::Resource::Cluster");
    my @clusters = @{$cluster_resources};
    my %cluster_aggregate_hash = ();
    my %cluster_vserver_hash = ();
    foreach my $cluster_resource (@$cluster_resources) {
        @{$cluster_aggregate_hash{$cluster_resource->get_attribute('cluster_name')}} =
                $cluster_resource->get_descendants(type => "NACL::AppBuilder::Xgen::Resource::Aggregate");
        # {clus1 => [aggr1_obj,aggr2_obj], clus2 => [aggr1_obj,aggr2_obj]}
        @{$cluster_vserver_hash{$cluster_resource->get_attribute('cluster_name')}} =
                $cluster_resource->get_descendants(type => "NACL::AppBuilder::Xgen::Resource::Vserver");
        # {clus1 => [vs0_obj,vs2_obj], clus2 => [vs1_obj,vs3_obj]}
    }

    my @volumes;

    for (my $i = 0; $i < $self->TOTAL_NAS_VOLUME_COUNT; $i++) {
        my $cluster = shift(@clusters);
        push (@clusters, $cluster);

        my $aggr = shift(@{$cluster_aggregate_hash{$cluster->get_attribute('cluster_name')}});
        push(@{$cluster_aggregate_hash{$cluster->get_attribute('cluster_name')}}, $aggr);

        my $vserver = shift(@{$cluster_vserver_hash{$cluster->get_attribute('cluster_name')}});
        push(@{$cluster_vserver_hash{$cluster->get_attribute('cluster_name')}}, $vserver);

        my $vs_name = $vserver->get_attribute("vserver"),

        my $volname = "volume_nas_$i";

        my %volume_arguments = (vserver   => $vs_name,
                                aggregate => $aggr->get_attribute("aggregate"),
                                volume    => $volname,
                                size      => $self->VOLUME_SIZE,
                                'junction-path' => "/$volname",
                                xgen_metadata => { data_protocol => 'nas' });

        my $volume_res = NACL::AppBuilder::Xgen::Resource::Volume->new(%volume_arguments);

        # connect the vs resource to the aggr's cluster parent resource
        $volume_res->add_parent_resources(resources => [$vserver,$aggr]);

        push(@volumes, $volume_res);
    }

    for (my $i = 0; $i < $self->TOTAL_SAN_VOLUME_COUNT; $i++) {
        my $cluster = shift(@clusters);
        push (@clusters, $cluster);

        my $aggr = shift(@{$cluster_aggregate_hash{$cluster->get_attribute('cluster_name')}});
        push(@{$cluster_aggregate_hash{$cluster->get_attribute('cluster_name')}}, $aggr);

        my $vserver = shift(@{$cluster_vserver_hash{$cluster->get_attribute('cluster_name')}});
        push(@{$cluster_vserver_hash{$cluster->get_attribute('cluster_name')}}, $vserver);

        my $vs_name = $vserver->get_attribute("vserver"),

        my $volname = "volume_san_$i";

        my %volume_arguments = (vserver   => $vs_name,
                                aggregate => $aggr->get_attribute("aggregate"),
                                volume    => $volname,
                                size      => $self->VOLUME_SIZE,
                                xgen_metadata  => { data_protocol => 'san' });

        my $volume_res = NACL::AppBuilder::Xgen::Resource::Volume->new(%volume_arguments);

        # connect the vs resource to the aggr's cluster parent resource
        $volume_res->add_parent_resources(resources => [$vserver,$aggr]);

        push(@volumes, $volume_res);
    }
    $self->add_resources(resources => \@volumes);
}

###############################################################################
# DESCRIPTION: Generates NFS lif resources and adds them to the configuration.
#              N Lif resources are created in every vserver per fabric defined,
#              where N is the number of nodes in the cluster that the vserver
#              is in. The lifs use the IPs, Netmask, and Gateway described in
#              the fabrics in the eth_data_fabric hash. This method currently
#              supports IPV4.
# INPUT: eth_fabric_data - a hash ref of ethernet fabric data including
#                          ips, netmask, gateway(optional) for the different
#                          ethernet fabrics for the testbed.
# OUTPUT: NONE
# Exceptions: None thrown directly by this method.
###############################################################################
sub generate_nfs_lif_resources {
    my ($self, %args) = @_;

    my $vservers = $self->get_resources(type => "NACL::AppBuilder::Xgen::Resource::Vserver");

    foreach my $vs (@{$vservers}) {
        my $eth_fabric_data = $args{eth_fabric_data};

        # Each vserver should only have one ancestor that is a cluster.
        my ($cluster) = $vs->get_ancestors(type => "NACL::AppBuilder::Xgen::Resource::Cluster");
        # Get all the nodes that are children of the vservers cluster so that we get all the nodes.
        my @nodes = $cluster->get_descendants(type => "NACL::AppBuilder::Xgen::Resource::Node");
        my $already_create_routes;
        # make an iscsi lif on each node (on each fabric)
        foreach my $node (@nodes) {
            my $node_name = $node->get_attribute('node');
            my $node_id = $self->get_node_id($node_name);

            # obtain the ethenet fabric table for this node
            my $fabric_table = $self->get_fabric_lookup_table_by_node($node_name, 'eth');
            foreach my $fabric (keys %{$fabric_table}) {
                my $address_range_v4 = $eth_fabric_data->{$fabric}->{address_range_v4};
                my $netmask_v4       = $eth_fabric_data->{$fabric}->{netmask_v4};
                if (!$address_range_v4 or !$netmask_v4) {
                    NATE::Exceptions::Argument->throw("No address ranges and/or netmasks were provided for fabric '$fabric'");
                }
                # if this fabric has a gateway defined,
                # make a route resource on this vserver if we havn't built one yet for this fabric
                $already_create_routes->{$fabric} //= {};
                if (!$already_create_routes->{$fabric}->{route_resource} and $eth_fabric_data->{$fabric}->{gateway_v4}) {
                    my $route_resource = NACL::AppBuilder::Xgen::Resource::NetworkRoute->new(vserver => $vs->get_attribute('vserver'),
                                                                                             gateway => $eth_fabric_data->{$fabric}->{gateway_v4});
                    $route_resource->add_parent_resources(resources => [$vs]);
                    $self->add_resources(resources => [$route_resource]);
                    $already_create_routes->{$fabric}->{route_resource} = $route_resource;
                }

                # choose a port on this fabric for this node
                my $port_name = shift(@{$fabric_table->{$fabric}});
                push(@{$fabric_table->{$fabric}}, $port_name);
                my $port_id = $self->get_port_id($node_name, $port_name);

                # make the lif resource
                my $lif = NACL::AppBuilder::Xgen::Resource::Lif->new(vserver => $vs->get_attribute('vserver'),
                                                                     'data-protocol' => ['nfs','cifs'],
                                                                     address_range => $address_range_v4,
                                                                     netmask => $netmask_v4,
                                                                     lif => "nfs_lf_n" . $node_id . "_p" . $port_id,
                                                                     'home-node' => $node_name,
                                                                     'home-port' => $port_name);
                $lif->add_parent_resources(resources => [$vs]);
                $self->add_resources(resources => [$lif]);
            }
        }
    }
}

###############################################################################
# DESCRIPTION: Generates ISCSI lif resources and adds them to the configuration.
#              N Lif resources are created in every vserver per fabric defined,
#              where N is the number of nodes in the cluster that the vserver
#              is in. The lifs use the IPs, Netmask, and Gateway described in
#              the fabrics in the eth_data_fabric hash. This method currently
#              supports IPV4.
# INPUT: eth_fabric_data - a hash ref of ethernet fabric data including
#                          ips, netmask, gateway(optional) for the different
#                          ethernet fabrics for the testbed.
# OUTPUT: NONE
# Exceptions: None thrown directly by this method.
###############################################################################
sub generate_iscsi_lif_resources {
    my ($self, %args) = @_;

    my $vservers = $self->get_resources(type => "NACL::AppBuilder::Xgen::Resource::Vserver");

    foreach my $vs (@{$vservers}) {
        my $eth_fabric_data = $args{eth_fabric_data};

        # Each vserver should only have one ancestor that is a cluster.
        my ($cluster) = $vs->get_ancestors(type => "NACL::AppBuilder::Xgen::Resource::Cluster");
        # Get all the nodes that are children of the vservers cluster so that we get all the nodes.
        my @nodes = $cluster->get_descendants(type => "NACL::AppBuilder::Xgen::Resource::Node");
        my $already_create_routes;
        # make an iscsi lif on each node (on each fabric)
        foreach my $node (@nodes) {
            my $node_name = $node->get_attribute('node');
            my $node_id = $self->get_node_id($node_name);

            # obtain the ethenet fabric table for this node
            my $fabric_table = $self->get_fabric_lookup_table_by_node($node_name, 'eth');
            foreach my $fabric (keys %{$fabric_table}) {
                my $address_range_v4 = $eth_fabric_data->{$fabric}->{address_range_v4};
                my $netmask_v4       = $eth_fabric_data->{$fabric}->{netmask_v4};
                if (!$address_range_v4 or !$netmask_v4) {
                    NATE::Exceptions::Argument->throw("No address ranges and/or netmasks were provided for fabric '$fabric'");
                }
                # if this fabric has a gateway defined,
                # make a route resource on this vserver if we havn't built one yet for this fabric
                $already_create_routes->{$fabric} //= {};
                if (!$already_create_routes->{$fabric}->{route_resource} and $eth_fabric_data->{$fabric}->{gateway_v4}) {
                    my $route_resource = NACL::AppBuilder::Xgen::Resource::NetworkRoute->new(vserver => $vs->get_attribute('vserver'),
                                                                                             gateway => $eth_fabric_data->{$fabric}->{gateway_v4});
                    $route_resource->add_parent_resources(resources => [$vs]);
                    $self->add_resources(resources => [$route_resource]);
                    $already_create_routes->{$fabric}->{route_resource} = $route_resource;
                }

                # choose a port on this fabric for this node
                my $port_name = shift(@{$fabric_table->{$fabric}});
                push(@{$fabric_table->{$fabric}}, $port_name);
                my $port_id = $self->get_port_id($node_name, $port_name);

                # make the lif resource
                my $lif = NACL::AppBuilder::Xgen::Resource::Lif->new(vserver => $vs->get_attribute('vserver'),
                                                                     'data-protocol' => ['iscsi'],
                                                                     address_range => $address_range_v4,
                                                                     netmask => $netmask_v4,
                                                                     lif => "iscsi_lf_n" . $node_id . "_p" . $port_id,
                                                                     'home-node' => $node_name,
                                                                     'home-port' => $port_name);
                $lif->add_parent_resources(resources => [$vs]);
                $self->add_resources(resources => [$lif]);
            }
        }
    }
}

###############################################################################
# DESCRIPTION: Generates igroup resources and adds them to the configuration.
#              Generates an igroup per ISCSI client on every vserver in the
#              configuration.  This method currently only supports ISCSI. The
#              igroups are named after the client they represent.
# INPUT: NONE
# OUTPUT: NONE
# Exceptions: None thrown directly by this method.
###############################################################################
sub generate_igroup_resources {
    my ($self, %args) = @_;

    # Get the clusters so that we can alternate vserver creation between multiple clusters.
    my $vserver_resources = $self->get_resources(type => "NACL::AppBuilder::Xgen::Resource::Vserver");
    my @vservers = @{$vserver_resources};

    # Get the clusters so that we can alternate vserver creation between multiple clusters.
    my $client_resources = $self->get_resources(type => "NACL::AppBuilder::Xgen::Resource::Client");
    my @clients = @{$client_resources};

    my @igroup;
    foreach my $vserver (@vservers) {
        foreach my $client (@clients) {
            # skip clients that are not doing iSCSI
            next if (!grep("iscsi", @{$client->get_attribute('protocols')}));

            my $igroup_res = NACL::AppBuilder::Xgen::Resource::Igroup->new(vserver => $vserver->get_attribute('vserver'),
                                                                           igroup => $client->get_attribute('name'),
                                                                           protocol => "iscsi",
                                                                           ostype => $client->get_attribute('ostype') =~ /windows/ ? "windows" : $client->get_attribute('ostype'),
                                                                           xgen_metadata => {client_os_type => $client->get_attribute('ostype')}); # used later for lun ostype.

            # We want the vserver to be last.
            $igroup_res->add_parent_resources(resources => [$client,$vserver]);

            push(@igroup, $igroup_res);
        }
    }
    $self->add_resources(resources => \@igroup);
}

###############################################################################
# DESCRIPTION: Generates igroup initiator resources and adds them to the
#              configuration. Generates an igroup initiator per igroup. Since
#              igroups are named after the client they represent we use the name
#              of the igroup for the client name.
# INPUT: NONE
# OUTPUT: NONE
# Exceptions: None thrown directly by this method.
###############################################################################
sub generate_igroup_initiator_resources {
    my ($self, %args) = @_;

    # Get the clusters so that we can alternate igroup creation between multiple clusters.
    my $igroup_resources = $self->get_resources(type => "NACL::AppBuilder::Xgen::Resource::Igroup");
    my @igroups = @{$igroup_resources};

    my @igroup_initiators;
    foreach my $igroup (@igroups) {
        my $initiator_res = NACL::AppBuilder::Xgen::Resource::Igroup::Initiator->new(vserver => $igroup->get_attribute('vserver'),
                                                                                     igroup => $igroup->get_attribute('igroup'),
                                                                                     protocol => "iscsi",
                                                                                     client_name => $igroup->get_attribute('igroup'));

        # We want the vserver to be last.
        $initiator_res->add_parent_resources(resources => [$igroup]);

        push(@igroup_initiators, $initiator_res);
    }

    $self->add_resources(resources => \@igroup_initiators);
}

###############################################################################
# DESCRIPTION: Generates NFS service resources and adds them to the
#              configuration. Generates an NFS service per vserver.
# INPUT: NONE
# OUTPUT: NONE
# Exceptions: None thrown directly by this method.
###############################################################################
sub generate_nfs_services_resources {
    my ($self, %args) = @_;

    # Get the clusters so that we can alternate vserver creation between multiple clusters.
    my $vserver_resources = $self->get_resources(type => "NACL::AppBuilder::Xgen::Resource::Vserver");
    my @vservers = @{$vserver_resources};

    my @nfs_services;

    foreach my $vserver (@vservers) {
        my $nfs_res = NACL::AppBuilder::Xgen::Resource::Nfs->new(vserver => $vserver->get_attribute("vserver"),
                                                                 "v4.0" => "enabled",
                                                                 "v4.1-acl" => "enabled",
                                                                 "v3" => "enabled",
                                                                 "v4.0-acl" => "enabled",
                                                                 "rquota" => "enabled",
                                                                 "v4.1" => "enabled",
                                                                 "v4.1-pnfs" => "enabled",);


        # Add vserver as parent to nfs service.
        $nfs_res->add_parent_resources(resources => [$vserver]);

        push(@nfs_services, $nfs_res);
    }
    $self->add_resources(resources => \@nfs_services);
}

###############################################################################
# DESCRIPTION: Generates ISCSI service resources and adds them to the
#              configuration. Generates an ISCSI service per vserver.
# INPUT: NONE
# OUTPUT: NONE
# Exceptions: None thrown directly by this method.
###############################################################################
sub generate_iscsi_services_resources {
    my ($self, %args) = @_;

    # Get the clusters so that we can alternate vserver creation between multiple clusters.
    my $vserver_resources = $self->get_resources(type => "NACL::AppBuilder::Xgen::Resource::Vserver");
    my @vservers = @{$vserver_resources};

    my @iscsi_services;

    foreach my $vserver (@vservers) {
        my $iscsi_res = NACL::AppBuilder::Xgen::Resource::Iscsi->new(vserver => $vserver->get_attribute('vserver'));

        # Add vserver as parent to iscsi service.
        $iscsi_res->add_parent_resources(resources => [$vserver]);

        push(@iscsi_services, $iscsi_res);
    }
    $self->add_resources(resources => \@iscsi_services);
}

###############################################################################
# DESCRIPTION: Generates lun map resources and adds them to the configuration.
#              Generates a lun map on every lun in the configuration.  Each lun
#              will be mapped to a single igroup.  In a vserver the lun
#              will be evenly distributed between igroups for maps. The luns
#              ostype will be changed to the ostype of the igroups client.
# INPUT: NONE
# OUTPUT: NONE
# Exceptions: None thrown directly by this method.
###############################################################################
sub generate_lun_map_resources {
    my ($self, %args) = @_;

    # Get the clusters so that we can alternate vserver creation between multiple clusters.
    my $vserver_resources = $self->get_resources(type => "NACL::AppBuilder::Xgen::Resource::Vserver");
    my @vservers = @{$vserver_resources};

    my @lun_maps;
    foreach my $vserver (@vservers) {
        my @luns = $vserver->get_descendants(type => "NACL::AppBuilder::Xgen::Resource::Lun");

        next unless (scalar @luns);

        my @igroups = $vserver->get_descendants(type => "NACL::AppBuilder::Xgen::Resource::Igroup");

        next unless (scalar @igroups);

        #(cl1,cl2)  vserver1:vol1:lun1  vserver2:vol1:lun1  might not work in this case so need better logic.
        foreach my $lun (@luns) {
            my $igroup = shift(@igroups);
            push(@igroups,$igroup);

            # Change the os_type of the lun to match the map.
            $lun->set_private_attributes(ostype => $igroup->get_attribute('client_os_type'));

            my $lunmap_res = NACL::AppBuilder::Xgen::Resource::LunMap->new(vserver => $lun->get_attribute('vserver'),
                                                                           lun => $lun->get_attribute('lun'),
                                                                           volume => $lun->get_attribute('volume'),
                                                                           igroup => $igroup->get_attribute('igroup'));

            # We want the vserver to be last.
            $lunmap_res->add_parent_resources(resources => [$igroup,$lun]);

            push(@lun_maps, $lunmap_res);
        }
    }
     $self->add_resources(resources => \@lun_maps);
}


###############################################################################
# DESCRIPTION: Generates lun resources and adds them to the configuration.
#              Generates a number of luns equal to the TOTAL_LUN_COUNT
#              constant.  The luns will only be generated on non DP type
#              volumes. The PERCENTAGE_OF_VOLUMES_WITH_LUNS controls the
#              percent of the non DP type volumes that get luns. The luns
#              ostype defaults to linux but will be changes if the lun is
#              mapped. The lun size is controlled by the LUN_SIZE constant.
# INPUT: NONE
# OUTPUT: NONE
# Exceptions: None thrown directly by this method.
###############################################################################
sub generate_lun_resources {
    my ($self, %args) = @_;

    # Get the clusters so that we can alternate vserver creation between multiple clusters.
    my $volume_resources = $self->get_resources(type => "NACL::AppBuilder::Xgen::Resource::Volume");

    # Filter the volumes resources by security style mixed.
    my $matches_hash = positive_filter(resources => $volume_resources,
                                       filter    => {'data_protocol' => 'san'});

    my @volumes = @{$matches_hash->{'matches'}};

    my @luns;

    for (my $i = 0; $i < $self->TOTAL_LUN_COUNT; $i++) {
        my $volume = shift(@volumes);
        push(@volumes,$volume);

        my $lun_res = NACL::AppBuilder::Xgen::Resource::Lun->new(vserver =>$volume->get_attribute('vserver'),
                                                                 volume  =>$volume->get_attribute("volume"),
                                                                 lun     => "lun$i",
                                                                 size    =>$self->LUN_SIZE,
                                                                 ostype  => 'linux'); # Defaulting the ostype to linux. Will overwrite when maps created.

        # connect the vs resource to the aggr's cluster parent resource
        $lun_res->add_parent_resources(resources => [$volume]);

        push(@luns, $lun_res);
    }
    $self->add_resources(resources => \@luns);
}

1;
