#
# Copyright (c) 2014 NetApp, Inc., All Rights Reserved.
#   Any other use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

package NACL::AppBuilder::Xgen::Resource::Aggregate;

use strict;
use warnings;

use NACL::C::Node;
use NACL::CS::Aggregate;
use NACL::CS::StorageDisk;
use NATE::Log qw(log_global);
use POSIX;
use NACL::AppBuilder::Xgen::Exceptions::NoSpareDisks qw(:try);
use NATE::Exceptions::Argument;
use Params::Validate qw(validate_with BOOLEAN SCALAR);
use NACL::Service::Xgen::Exceptions::InvalidAttribute;

use parent qw( NACL::Service::Xgen::Resource );

use constant nacl_cs_class => 'NACL::CS::StorageAggregate';

my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();

sub _validate_private_attrs_with_nacl_cs_class {
    my ($self, %args) = @_;
    log_global()->enter() if $may_enter;

    # remove special args from the validation
    # or else the base class nacl cs class validation will complain
    delete($args{'nacltask_enable_composite'});
    delete($args{'nacltask_object_store_name'});
    delete($args{'nacltask_object_store_size_limit'});
    $self->SUPER::_validate_private_attrs_with_nacl_cs_class(%args);
    log_global()->exit() if $may_exit;
}

sub _validate_private_attrs {
    my ($self, $args) = @_;
    log_global()->enter() if $may_enter;

    eval {
        $args = validate_with(
            params => [%{$args}],
            spec => {
                nacltask_enable_composite        => { type => SCALAR, optional => 1},
                nacltask_object_store_name       => { type => SCALAR, optional => 1 },
                nacltask_object_store_size_limit => { type => SCALAR, optional => 1 },
            },
            allow_extra => 1);
    };
    if (my $err = $@) {
        NACL::Service::Xgen::Exceptions::InvalidAttribute->throw($err);
    }

    log_global()->exit() if $may_exit;
    return $args;
}

sub generate_aggr_info {
    my ($pkg, %args) = @_;
    $Log->enter() if $may_enter;

    my $nacl_node_objs = delete($args{nacl_node_objs})
                    // [NACL::C::Node->find()];

    # Variable to tell whether we want to create new aggregates or not.
    # Defaults to say we want to create new aggregates.
    my $no_new_aggrs = delete($args{no_new_aggr}) // 0;

    my %return = ();

    foreach my $nacl_node_obj (@{$nacl_node_objs}) {
        my $node = $nacl_node_obj->node();

        # We only want to perform the disk show and build the information
        # for building new aggrs if the user wanted to build new aggregates.
        # Otherwise this disk show can take a long time but provide nothing useful.
        if (! $no_new_aggrs) {
            # storage disk show
            my @disks = NACL::CS::StorageDisk->fetch(command_interface => $nacl_node_obj,
                                                     filter => {owner => $node},
                                                     allow_empty => 1);
            my $disk_count = scalar(@disks);
            my @spare_disks = grep {$_->container_type() eq "spare"} @disks;
            my $spare_disk_count = scalar(@spare_disks);

            try {
                $return{$node} = $pkg->_calculate_aggr_per_node(disks_per_node  => $disk_count,
                                                                spares_per_node => $spare_disk_count)
            } catch NACL::AppBuilder::Xgen::Exceptions::NoSpareDisks with {
                # we can not build new aggregates for this node
            };
        } # End if no_new_aggrs
        # discover aggregates and reuse
        $return{$node}->{current_aggr_layout} =
               [NACL::CS::StorageAggregate->fetch(command_interface => $nacl_node_obj,
                                                  filter => {'nodes'     => [$nacl_node_obj->node()],
                                                             'ha-policy' => 'sfo'},
                                                  allow_empty => 1)];
    }

    $Log->exit() if $may_exit;
    return \%return;
}

sub _calculate_aggr_per_node {
    my ($pkg, %args) = @_;
    $Log->enter() if $may_enter;

    my $disks_per_node = delete($args{disks_per_node})
                        // NATE::Exceptions::Argument->throw("disks_per_node is a required parameter");
    my $spares_per_node = delete($args{spares_per_node})
                        // NATE::Exceptions::Argument->throw("spares_per_node is a required parameter");
    my $spare_disk_percentage = delete($args{spare_disk_percentage})
                        // 0.15;

    # validate that the number of spares that we want to keep satisfies the requested
    # percentage
    my $spares_to_keep = POSIX::ceil($spare_disk_percentage * $disks_per_node);
    my $disks_to_use = $spares_per_node - $spares_to_keep;
    if ($disks_to_use <= 0) {
        NACL::AppBuilder::Xgen::Exceptions::NoSpareDisks->throw("There are not enough spare disks to satisfy " .
                                                                "the spare disk percentage of $spare_disk_percentage.");
    }

    # do not create new aggregates if the number of disks to use less than
    # the minimum sized raiddp aggregate
    if ($disks_to_use < 3) {
        NACL::AppBuilder::Xgen::Exceptions::NoSpareDisks->throw("There are not enough spare disks to " .
                                                                "create any aggregates.");
    }

    # calculate the max raid size to use for these aggregates
    my $aggr_count = $pkg->_calculate_aggr_count(disks_to_use => $disks_to_use);
    my $disks_per_aggr = POSIX::floor($disks_to_use / $aggr_count);
    my $max_raid_size = $pkg->_calculate_max_raid_size(disks_per_aggr => $disks_per_aggr);

    $Log->exit() if $may_exit;
    return {disks_per_aggr => $disks_per_aggr,
            aggr_count     => $aggr_count,
            max_raid_size  => $max_raid_size,
            spares_to_keep => $spares_to_keep + ($disks_to_use - $disks_per_aggr * $aggr_count)};
}

sub _calculate_aggr_count {
    my ($pkg, %args) = @_;
    $Log->enter() if $may_enter;

    my $disks_to_use = delete($args{disks_to_use})
                        // NATE::Exceptions::Argument->throw("disks_to_use is a required parameter");
    my $max_disks_per_aggr = delete($args{max_disks_per_aggr})
                        // 32;

    my $aggr_count = POSIX::ceil($disks_to_use / $max_disks_per_aggr);

    $Log->exit() if $may_exit;
    return $aggr_count;
}

sub _calculate_max_raid_size {
    my ($pkg, %args) = @_;
    $Log->enter() if $may_enter;

    my $disks_per_aggr = delete($args{disks_per_aggr})
                        // NATE::Exceptions::Argument->throw("disks_to_use is a required parameter");
    my $default_max_raid_size= delete($args{default_max_raid_size})
                        // 16;

    if ($disks_per_aggr > $default_max_raid_size) {
        $Log->exit() if $may_exit;
        return $disks_per_aggr;
    }
    $Log->exit() if $may_exit;
    return $default_max_raid_size;
}

sub _discover_current_aggr_layout_per_node {
    my ($pkg, %args) = @_;
    $Log->enter() if $may_enter;

    my $nacl_node_obj = delete($args{nacl_node});

    my $aggrs =
        NACL::CS::StorageDisk->fetch(node   => $nacl_node_obj,
                                     filter => {'node' => [$nacl_node_obj->node()]});

    $Log->exit() if $may_exit;
    return $aggrs;
}

1;

__END__

=head1 NAME

NACL::AppBuilder::Xgen::Resource::Aggregate

=head1 AUTHOR/MAINTAINER

=over

=item PTEFIT AUTOMATION (ng-ptefit-automation-core@netapp.com)

=back

   Last revised $Date: $ by $Author: $

=cut
