#
# Copyright (c) 2014 NetApp, Inc., All Rights Reserved.
#   Any other use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#
package NACL::AppBuilder::Xgen::Resource::Client;

use strict;
use warnings;

use Data::Dumper;

use Hostrec;

use parent qw( NACL::Service::Xgen::Resource );

use Params::Validate qw(validate_with SCALAR ARRAYREF);
use NACL::Service::Xgen::Exceptions::InvalidAttribute;

sub new {
    my ($class, %args) = @_;

    my $self = $class->SUPER::new(%args);

    $self->_initialize();

    return $self;
}

sub _validate_private_attrs {
    my ($self, $args) = @_;

    eval {
        $args = validate_with(
            params => [%{$args}],
            spec => {
                name             => {type => SCALAR},
                protocols        => {type => ARRAYREF},
                synthetic_client => {type => SCALAR, default => 0},
                ostype           => {type => SCALAR, optional => 1},
                virtual_machines => {type => ARRAYREF, default => []},
                vma_host => {type => SCALAR, optional => 1},
                is_hypervisor    => {type => SCALAR, optional => 1, default => 0},
                mp_type          => {type => SCALAR, optional => 1}
            });
    };
    if (my $err = $@) {
        NACL::Service::Xgen::Exceptions::InvalidAttribute->throw($err);
    }

    return $args;
}

sub _initialize {
    my ($self, %args) = @_;

    my $host_obj = Hostrec->new(id => $self->get_attribute('name'));
    my %hash = $host_obj->param_fields();

    my $hosttype = $host_obj->hosttype();
    my $os_type;
    my $synthetic_client = 0;
    my $mp_type = 'none';
    my $is_hypervisor = 0;
    my %extra = ();

    if ($hosttype =~ /windows/i) {
        if ($hash{OS_VER} and $hash{OS_VER} =~ /^5/) {
            $os_type = 'windows';
        } else {
            $os_type = 'windows_2008';
        }
        $mp_type = 'mpio';
    } elsif ($hosttype =~ /sanblaze/i) {
        $os_type = 'linux';
        $synthetic_client = 1;
    } elsif ($hosttype =~ /citrix/i) {
        $os_type = 'xen';
        $mp_type = 'dmmp';
        $is_hypervisor = 1;
    } elsif ($hosttype =~ /vmware/i) {
        $os_type = 'vmware';
        $mp_type = 'native';
        $is_hypervisor = 1;
        $extra{vma_host} = $hash{VMA_HOST};        
    } elsif ($hosttype =~ /linux/i) {
        $os_type = 'linux';
        $mp_type = 'dmmp';
    } elsif ($hosttype =~ /solaris/i) {
        $os_type = 'solaris';
    } elsif ($hosttype =~ /aix/i) {
        $os_type = 'aix';
    } elsif ($hosttype =~ /hpux/i) {
        $os_type = 'hpux';
    }

    if ($is_hypervisor) {
        $extra{virtual_machines} = $self->_get_virtual_machines();
    }

    $self->set_private_attributes(ostype => $os_type,
                                  mp_type => $mp_type,
                                  is_hypervisor => $is_hypervisor,
                                  synthetic_client => $synthetic_client,
                                  %extra);

    return;
}

sub _get_virtual_machines {
    my ($self) = @_;

    my $host_obj = Hostrec->new(id => $self->get_attribute('name'));
    my %hash = $host_obj->param_fields();
    if (my $vms = $hash{VIRTUAL_MACHINES}) {
        return [split(/\s*,\s*/, $vms)];
    }

    return [];
}

1;

__END__

=head1 NAME

NACL::AppBuilder::Xgen::Resource::Client

=head1 AUTHOR/MAINTAINER

=over

=item PTEFIT AUTOMATION (dl-ptefit-automation-core@netapp.com)

=back

   Last revised $Date: $ by $Author: $

=cut
