# Copyright (c) 2016 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.

## @summary AltaVault command interface wrapper, to resemble ONTAP nodes
## @author rjray@netapp.com,ng-shiner-qa@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::C::AVANode - A SystemNode-like class for AltaVault appliances

=head1 DESCRIPTION

B<NACL::C::AVANode> is a class that inherits from
B<NACL::C::CommandInterface::AVA>, extending it with attributes and methods that
are needed for full communication with AltaVault appliances.

As this is not an ONTAP-derived resource, there is no component-state class to
accompany this class.

=cut

package NACL::C::AVANode;

use 5.014;
use strict;
use warnings;
use base qw(NACL::C::CommandInterface);

use Data::Dumper;

use Const::Fast;
use List::MoreUtils qw(any);
use Net::Ping::External qw(ping);
use Params::Validate qw(validate validate_with validation_options :types);

use Tharn qw(host);
use NATE::Log qw(log_global);
use NATE::BaseException qw(:try);
use NATE::Exceptions::Argument;
use NACL::APISet::Exceptions::CommandNotFoundException;
use NACL::Exceptions::Timeout;
use NACL::ComponentUtils qw(_verify_invocation);

const my $Log       => log_global;
const my $may_enter => $Log->may_enter;
const my $may_exit  => $Log->may_exit;

# This class needs auto-loading:
## no critic(ClassHierarchies::ProhibitAutoloading)

=head1 ATTRIBUTES

=head2 node

The name of the node (the AltaVault appliance) that this object connects to.

=head1 METHODS

=head2 new

    my $ava = NACL::C::AVANode->new(name => $name);
    # or
    my $ava = NACL::C::AVANode->new(node => $nodename);
    # or
    my $ava = NACL::C::AVANode->new(hostrec => $hostrec_obj);

Can only be called as a static method.

Specifying any one of C<name>, C<node> or C<hostrec> will populate the other
two.

Returns a new B<AVANode> object, or throws an exception on failure.

=cut

use Class::MethodMaker [
    scalar => [
        {
            -default_ctor => sub { shift->_construct_primary_key; },
            -store_cb     => sub { shift->_verify_hostrec_callback(@_); },
        },
        'node',
    ],
];

# Set the on_fail behavior for Params::Validate to throw an exception (of the
# NATE::Exceptions::Argument class) on errors instead of a plain croak().
validation_options(
    on_fail => sub {
        NATE::Exceptions::Argument->throw(shift);
    }
);

our $AUTOLOAD;
sub AUTOLOAD {
    $Log->enter if $may_enter;

    my $method = $AUTOLOAD;
    $method =~ s/.*://;    # strip fully-qualified portion

    ## no critic(TestingAndDebugging::ProhibitNoStrict)
    no strict 'refs';
    *{$method} = sub {
        $Log->enter if $may_enter;
        my ($self, @args) = @_;
        my $result;
        my $apiset = $self->apiset;

        try {
            $result = $apiset->$method(@args);
        } otherwise {
            # TODO: When (if) we have an APISet-level exception for the EOF
            # condition that we're testing for below, refactor this to catch
            # the specific exception.
            my $exception = shift;

            if ($exception->text =~ /unexpected eof/i) {
                # This means that the connection is bad, probably because the
                # AVA rebooted. Try to reconnect and retry the command.
                # TODO: This might be better at the APISet level, but for now
                # (due to time constraints) it is here. Consider revisiting this
                # and trying to push it down to APISet.
                $apiset->api_reconnect_session;
                $result = $apiset->$method(@args);
            } else {
                # Anything else, we re-throw for now:
                $exception->throw;
            }
        } finally {
            # TODO: This is something else that should eventually be done at
            # the APISet level. For now, this is the only way of catching bad
            # commands since the execution layer doesn't know to throw any
            # exception.
            if ( !defined $result) {
                $apiset->api_reconnect_session;
                #NATE::BaseException->throw(
                #    "No output seen when running '$method':\n\n"
                #);
            } else {
                my @lines = split /\n/, $result->get_raw_output;
                if (my @errors = grep { /^%/ } @lines) {
                    if (any { /Unrecognized command/i } @errors) {
                        NACL::APISet::Exceptions::CommandNotFoundException->throw(
                            text            => "Command '$method' is unknown.",
                            response_object => $result
                        );
                    } else {
                        NATE::BaseException->throw(
                            "Error(s) detected when running '$method':\n\n" .
                            join "\n", @errors
                        );
                    }
                }
            }
        };

        $Log->exit if $may_exit;
        return $result;
    };
    use strict 'refs';

    $Log->exit if $may_exit;
    goto &{$method};
}

=head2 find

    $avanode = NACL::C::AVANode->find;
    # or
    @avanodes = NACL::C::AVANode->find;

Can only be called as a static method.

Find AVA nodes based on NATE parameters given to the run-time environment.
Works like any other B<find> on a NACL component, except that no parameters
are taken.

If called in a list context, a list of all discovered AVA nodes is returned. If
called in a scalar context, only the first element of the list is returned.

=cut

sub find {
    $Log->enter if $may_enter;
    my $pkg = shift;

    if (ref $pkg) {
        NATE::Exceptions::Argument->throw(
            'The NACL::C::AVANode::find method can only be called as a static' .
            ' method.'
        );
    }

    my @nodes = $pkg->SUPER::find();

    $Log->exit if $may_exit;
    return wantarray ? @nodes : $nodes[0];
}

=head2 reboot

    # Basic reboot:
    $avanode->reboot;

    # Reboot, but return immediately (without waiting):
    $avanode->reboot(wait_for_reboot => 0);

Reboot the AltaVault appliance.

By default, when the reboot command is sent the method will wait to detect that
the appliance has gone down. It will then wait for the appliance to come back
up. Each of these waits is configurable by the C<timeout> parameter. If either
of the wait-periods reaches the timeout, then an exception is thrown.

=over

=item Options

The following options are recognized:

=over

=item wait_for_reboot

A boolean value that defaults to C<1> (true). If true, then the method waits
until the appliance has successfully rebooted before returning, throwing an
exception if it does not come back up.

=item timeout

A numerical value that defaults to C<300>. This is the number of seconds that
the appliance has for each of the down/up cycles. If the appliance does not go
down within this timeframe, or if it goes down but does not come up within a
similar timeframe, then an exception will be thrown.

Note that the appliance is probed roughly every 5 seconds, for both checking
when it goes down and when it returns. So a C<timeout> value does not mean that
the full time will be required for each cycle. As soon as it is detected that
the appliance is down, the timeout loop for coming back up starts. And as soon
as it is detected that the appliance is back up, the method immediately
returns.

=back

=item Exceptions

The following exceptions may be thrown:

=over

=item B<NATE::Exceptions::Argument>

If an unknown parameter is passed, or if an invalid value is given for
C<timeout>, then this exception will be thrown.

=item B<NACL::Exceptions::Timeout>

If either of the down/up cycles of the reboot take longer than C<timeout>
seconds to complete, this exception will be thrown.

=back

=back

=cut

sub reboot {
    $Log->enter if $may_enter;
    my ($self, @opts) = @_;
    my %opts = validate(
        @opts,
        {
            wait_for_reboot => { type => BOOLEAN, default => 1 },
            timeout         => {
                type    => SCALAR,
                regex   => qr/^\d+$/,
                default => 300,
            },
        }
    );

    # Reboot AVA Node using admin command reboot
    my $result = $self->reload();

    my @lines = split /\n/, $result->get_raw_output;
    if (my @errors = grep { /^%/ } @lines) {
        NATE::BaseException->throw(
            "Error(s) detected when running 'reboot':\n\n" .
            join "\n", @errors
        );
    }

    if ($opts{wait_for_reboot}) {
        # While the host shows still alive, it isn't yet rebooted. Loop until
        # we get a false value from is_alive. Only give it $opts{timeout} sec
        # to go down, so that we don't loop forever if the reboot was stopped
        # for some reason.
        my $count = 0;
        my $flag;
        while ($flag = $self->is_alive) {
            last if (($count += 5) > $opts{timeout});
            Tharn::snooze(5);
        }

        # Did we reach here because it was finally down, or because of running
        # out the clock?
        if ($flag) {
            NACL::Exceptions::Timeout->throw(
                'AVA appliance ' . $self->node .
                " did not go down within $opts{timeout} seconds."
            );
        } else {
            $Log->comment("Possibly AVA has started to reboot");
        }

        if ( $self->uptime() =~ /(-1|0)/ ) {
            $Log->comment("Waiting for management back end to boot back...");
            $count = 0;
            while( $self->uptime() =~ /(-1)/ ) {
                last if (($count += 5) > $opts{timeout});
                Tharn::snooze(5);
            }

            if ( $self->uptime() =~ /(0)$/ ) {
                $Log->comment("Management back end is unavailable. "
                    . "Fixing using root access");
                $self->kill_process(process=>'/opt/rbt/bin/rfsd');
                $self->service_enable();
            }
        }

        if ( $self->uptime() =~ /(-1|0)$/ ) {
            NACL::Exceptions::Timeout->throw(
                'AVA appliance ' . $self->node .
                " did not come back up within $opts{timeout} seconds."
            );
        }
    }

    $Log->exit if $may_exit;
    return;
}

=head2 update_build

Update build on AVA 
    $avanode->update_build(imageurl=>"url...");

=over

=item Options

The following options are recognized:

=over

=item imageurl

A string that represents image url
Returns nothing

=back

=back

=cut

sub update_build {
    $Log->enter if $may_enter;
    my ($self, @opts) = @_;
    my %opts = validate(
        @opts,
        {
            imageurl => { type => SCALAR}
        }
    );

    $Log->comment('Image before update: ');
    $Log->comment(Dumper($self->show_version()->get_parsed_output()));
    my $output = $self->show_images()->get_raw_output();

    # Find boot image to install build
    $output =~ /Last boot partition:\s(\d)/;
    my $last_boot = $1;
    my $install_boot = $last_boot == 1 ? "2" : "1";


    # Disable optimization service
    try {
        $self->no_service_enable();
    } otherwise {
        my $exception = shift;

        # Ignore if optimization service is already stopped
        if ( $exception->text !~ /optimization service is not running/i ) {
            # Anything else, we re-throw for now:
            $exception->throw;
        }
    };

    # Fetch image
    my $imageurl = $opts{imageurl};
    my $apiset = $self->apiset;
    $apiset->set_timeout(timeout => 600);
    $apiset->image_fetch($imageurl => '');

    # Install image
    $apiset->image_install('image.img' => $install_boot, force => '');

    # Update boot image
    $self->image_boot($install_boot => '');

    # Write memory
    $self->write_memory();

    # Reboot
    $self->reboot();

    $Log->comment('Image after update: ');
    $Log->comment(Dumper($self->show_version()->get_parsed_output()));
    $Log->comment(Dumper($self->show_images()->get_raw_output()));

    $Log->exit if $may_exit;
}

=head2 get_destination_path

    # Returns destination path for a given volume name
    $avanode->get_destination_path(volume => 'vol1');

Return sAVA destination path for a given volume

=over

=item Options

The following options are recognized:

=over

=item volume

A string that represents volume.
Returns <ip>:/share/<volume>.

=back

=back

=cut

sub get_destination_path {
    $Log->enter if $may_enter;
    my ($self, @opts) = @_;
    my %opts = validate(
        @opts,
        {
            volume => { type => SCALAR}
        }
    );

    my $ip = $self->default_ip();
    my $volume = $opts{volume};
    my $destination_path = "$ip:/share/$volume";

    $Log->exit if $may_exit;
    return $destination_path;
}

=head2 uptime

Returns uptime of AVA
If AVA is not running, returns -1.
If AVA is running, but management back end is unavailable, returns 0.

=cut

sub uptime {
    $Log->enter if $may_enter;
    my $self = shift;

    my $uptime = -1;
    try {
        my $output = $self->show_version()->get_parsed_output();
        $Log->debug("uptime: " . Dumper($output));
        if ( defined $output->{Uptime} ) {
            # Obtain uptime when management back end is available.
            $uptime = $output->{Uptime};
        } else {
            # AVA is running but management back end is unavailable.
            $uptime = 0;
        }
    } otherwise {
        my $exception = shift;
        $Log->debug("Error: " . $exception->text);

        # Return -1 when AVA is not running,
        # 0 if management back end is unavailable.
        $Log->debug("Probably AVA is not running");
        if ($exception->text =~ /Management back end unavailable./i) {
            $uptime = 0;
        }
    };
    $Log->debug("Uptime: $uptime");

    $Log->exit if $may_exit;
    return $uptime;
}

=head2 add_whitelist_ips

    # Add snapmirror whitelist ips
    $avanode->add_whitelist_ips(ips=>\@ips_list);

Add snapmirror whitelist IPs.  If an IP is already added, print warning
and ignore adding it.  Propagate APISet errors if there are any.

=over

=item Options

The following options are recognized:

=over

=item ips

A reference to list of IPs to be added as whitelist
Returns nothing.

=item reboot

A BOOLEAN to say whether reboot is required after setting whitelist.
By default it is 0.
Returns nothing.

=back

=back

=cut

sub add_whitelist_ips {
    $Log->enter if $may_enter;
    my ($self, @args) = @_;
    my %opts = validate(
        @args,
        {
            ips    => { type => ARRAYREF },
            reboot => { type => BOOLEAN, default => 0 },
        }
    );

    my @ips = @{$opts{ips}};
    my $is_reboot = $opts{reboot};

    my $whitelist = $self->show_snapmirror_whitelist()->get_parsed_output();
    $Log->debug("IPs: ".Dumper($whitelist));

    foreach my $ip (@ips) {
        if ( $whitelist !~ qr/$ip/ ) {
            $self->snapmirror_whitelist_add(ip=>$ip);
        } else {
            $Log->comment("Skipping $ip as it is already part of whitelist");
        }
    }

    if ( $is_reboot ) {
        $Log->comment("Rebooting after adding whitelist ips");
        $self->reboot();
    }

    $Log->exit if $may_exit;
}

=head2 remove_whitelist_ips

    # Remove snapmirror whitelist ips
    $avanode->remove_whitelist_ips(ips=>\@ips_list);

Remove snapmirror whitelist IPs.  If an IP is not present, print warning
and ignore adding it.  Propagate APISet errors if there are any.

=over

=item Options

The following options are recognized:

=over

=item ips

A reference to list of IPs to be removed from whitelist
Returns nothing.

=item reboot

A BOOLEAN to say whether reboot is required after removing whitelist.
By default it is 0.
Returns nothing.

=back

=back

=cut

sub remove_whitelist_ips {
    $Log->enter if $may_enter;
    my ($self, @args) = @_;
    my %opts = validate(
        @args,
        {
            ips    => { type => ARRAYREF },
            reboot => { type => BOOLEAN, default => 0 },
        }
    );

    my @ips = @{$opts{ips}};
    my $is_reboot = $opts{reboot};

    my $whitelist = $self->show_snapmirror_whitelist()->get_parsed_output();
    $Log->debug("IPS: ".Dumper($whitelist));

    foreach my $ip (@ips) {
        if ( $whitelist =~ qr/$ip/ ) {
            $self->snapmirror_whitelist_delete(ip=>$ip);
        } else {
            $Log->comment("Skipping $ip as it is not part of whitelist");
        }
    }

    if ( $is_reboot ) {
        $Log->comment("Rebooting after removing whitelist ips");
        $self->reboot();
    }

    $Log->exit if $may_exit;
}

=head2 kill_process

    # Kills process using root access to AVA
    $avanode->kill_process(process=>'/opt/rbt/bin/rfsd');

Kills all named process using root access to AVA
Returns nothing.

=over

=item Options

The following options are recognized:

=over

=item process

Name of process to kill.

=back

=back

=cut

sub kill_process {
    $Log->enter if $may_enter;
    my ($self, @args) = @_;
    my %opts = validate(
        @args,
        {
            process => { type => SCALAR},
        }
    );

    my $process = $opts{process};

    my $response = $self->get_avaroot_apiset()->killall($process);
    $response = $self->get_avaroot_apiset()->execute_raw_command(
        command => "ps -ef | grep " . $process
    );

    $Log->comment("After killing process '$process' found: "
        . $response->get_raw_output());

    $Log->exit if $may_exit;
}


=head2 default_apiset_could

This is a static method that takes no arguments, and always returns the same
value. The value is a hash reference that defines the basic level of options
that AVA APISet objects would take.

=cut

sub default_apiset_could {
    return {
        category  => 'AVA',
        interface => 'CLI',
        set       => 'AVAAdmin',
    };
}

=head2 default_apiset_should

This is a static method that takes no arguments, and always returns the same
value. The value is a hash reference that defines what a default AVA APISet
object should do, in terms of options/features.

=cut

sub default_apiset_should {
    return {
        category  => 'AVA',
        interface => 'CLI',
        set       => 'AVAAdmin',
    };
}

=head2 default_apiset_must

This is a static method that takes no arguments, and always returns the same
value. The value is a hash reference that defines what APISet options must be
set (and their value) for instantiating an AVA APISet.

=cut

sub default_apiset_must {
    return {
        category        => 'AVA',
    };
}

=head2 apiset

    # Get the default APISet
    my $apiset = $avanode->apiset;

    # Get a specific set:
    my $root_apiset = $avanode->apiset(set => 'AVARoot');

(Instance method only)

This method fetches the APISet object that this node uses for making calls on
the underlying connection to the node that this object represents.

=over

=item Options

The following options are recognized by this method:

=over

=item category

A scalar (string) value. The category name for the desired APISet. This should
I<always> be C<AVA>.

=item interface

A scalar (string) value. The interface type for the desired APISet. At present,
only C<CLI> is supported.

=item set

A scalar (string) value. The type of APISet to return. Must be one of
C<AVAAdmin> or C<AVARoot>. If this is not passed, it will default based on the
values of the C<set> key in the hash tables returned by B<apiset_should> or
B<apiset_must> (as is the general case for APISets).

=item hostobj

An object of the B<Hostrec> class. The host object to use when establishing the
connection for the APISet object. Generally, this should not be needed as the
node object already has a B<Hostrec> object for the AltaVault appliance that
this object represents.

=item cache_apiset

A boolean value whose default is C<1> (true). When true, any newly-created
APISet object will be cached on this node object for future calls. The "key"
used to cache the object is the combination of the C<category>, C<interface>
and C<set> values.

=item return_new_object

A boolean value whose default is C<0> (false). When true, any existing cached
APISet object is ignored and a new object is always created. If C<cache_apiset>
is also true, then the new object will overwrite anything in the cache.

=item check_apiset_already_loaded

A boolean value whose default is C<0> (false). When true, rather than returning
the APISet object, the method returns a boolean value that indicates whether
an APISet object for this combination of (C<category>, C<interface>, C<set>)
currently exists on the node object's cache. If there is no object, a new
object is I<not> created.

=back

In addition to the above, any additional options passed in are passed directly
to the B<new> method of B<NACL::APISet>. If any of them are invalid, an
exception may be thrown by that method.

=item Exceptions

The following exception may be thrown by this method:

=over

=item B<NATE::Exceptions::Argument>

If an invalid parameter (or a parameter with an invalid value) is passed in,
this exception class will be thrown.

=back

In addition to the above, any exceptions that can be thrown by the B<new>
method of B<NACL::APISet> may be thrown here.

=back

The return value from this call is an object that derives from the
B<NACL::APISet> class.

=cut

sub apiset {
    $Log->enter if $may_enter;
    my ($self, @opts) = @_;
    my %opts = validate_with(
        params      => \@opts,
        allow_extra => 1,
        spec        => {
            cache_apiset                => { type => BOOLEAN, default => 1 },
            return_new_object           => { type => BOOLEAN, default => 0 },
            check_apiset_already_loaded => { type => BOOLEAN, default => 0 },
            category                    => { type => SCALAR, optional => 1 },
            interface                   => { type => SCALAR, optional => 1 },
            set                         => { type => SCALAR, optional => 1 },
            hostobj                     => {
                type     => OBJECT,
                isa      => 'Hostrec',
                optional => 1
            },
        }
    );
    my ($apiset, $apiset_cache_key);

    my $cache      = delete $opts{cache_apiset};
    my $return_new = delete $opts{return_new_object};
    my $check_only = delete $opts{check_apiset_already_loaded};
    if ($return_new) {
        # If they specifically want a new object, don't save it in the cache
        $cache = 0;
    }

    # Sanity-checking the actual APISet arguments.

    # The default value of hostobj is the hostrec value on this object:
    if (! $opts{hostobj}) {
        $opts{hostobj} = $self->hostrec;
    }

    # Fetch and merge the default and CI-specific rule-sets for APISets:
    my %apiset_opts = (
        %{$self->default_apiset_must},
        %{$self->default_apiset_could},
        $self->apiset_must,
        $self->apiset_should,
        # Any actual user-provided values take the highest priority
        %opts,
    );

    $apiset_cache_key =
        join q{|}, @apiset_opts{qw(category interface set)};
    $apiset = $self->_apisets_index($apiset_cache_key);

    # If we're just checking, then all we need to know is whether $apiset is
    # currently non-null:
    if ($check_only) {
        $apiset = (defined $apiset) ? 1 : 0;
    } else {
        if ($return_new || (! $apiset)) {
            # Either we're ignoring the cache, or there wasn't anything in it
            $apiset = NACL::APISet->new(%apiset_opts);
            if ($cache) {
                $self->_apisets_set($apiset_cache_key => $apiset);
            }
        }
    }

    $Log->exit if $may_exit;
    return $apiset;
}

=head2 get_avaadmin_apiset

    # Get an APISet specifically with the AVAAdmin set:
    my $apiset = $avanode->get_avaadmin_apiset;

This method gets an APISet object for the calling node, but restricts it to
an APISet with the C<AVAAdmin> privilege. Any arguments passed are passed
through to the B<apiset> method, but the C<set> argument is forced to
C<AVAAdmin>.

=over

=item Options

All options that may be passed to B<apiset> may be passed here. They are passed
unchanged to B<apiset>. Note that any value for the C<set> option will be
overridden here.

=item Exceptions

Any exceptions that may be thrown by B<apiset> may be thrown here.

=back

=cut

sub get_avaadmin_apiset {
    $Log->enter if $may_enter;
    my $self = shift;

    # Pass any additional args, but ensure that "set" is forced to AVAAdmin:
    my $apiset = $self->apiset(@_, set => 'AVAAdmin');

    $Log->exit if $may_exit;
    return $apiset;
}

=head2 get_avaroot_apiset

    # Get an APISet specifically with the AVARoot set:
    my $apiset = $avanode->get_avaadmin_apiset;

This method gets an APISet object for the calling node, but restricts it to
an APISet with the C<AVARoot> privilege. Any arguments passed are passed
through to the B<apiset> method, but the C<set> argument is forced to
C<AVARoot>.

=over

=item Options

All options that may be passed to B<apiset> may be passed here. They are passed
unchanged to B<apiset>. Note that any value for the C<set> option will be
overridden here.

=item Exceptions

Any exceptions that may be thrown by B<apiset> may be thrown here.

=back

=cut

sub get_avaroot_apiset {
    $Log->enter if $may_enter;
    my $self = shift;

    # Pass any additional args, but ensure that "set" is forced to AVARoot:
    my $apiset = $self->apiset(@_, set => 'AVARoot');

    $Log->exit if $may_exit;
    return $apiset;
}

=head2 is_alive

    if (! $avanode->is_alive) {
        $Log->error('AVA node is not reachable');
        # ...
    }

(Instance method only)

This method takes no arguments and returns only a boolean value that indicates
whether the AltaVault appliance that this object represents is reachable. It
tests this by using a network ping request.

=cut

sub is_alive {
    $Log->enter if $may_enter;
    my ($self) = @_;

    # This can only be an instance invocation:
    $self->_verify_invocation(style => 'instance_only');

    my $is_alive = ping(host => $self->default_ip());

    $Log->exit if $may_exit;
    return $is_alive;
}

=head1 ALTAVAULT COMMAND METHODS

At present, the AltaVault commands are not implemented as specific methods.
Any AVA command can be called on a node object by calling a method whose name
is constructed from the elements of the command with underscores (C<_>) in
place of spaces:

    $avanode->show_info; # Executes "show info"

If a command takes options, those are passed as key/value pairs:

    $avanode->snapmirror_share_unlock('share-name' => $share);

To ease in typing and avoid having to quote parameter names all the time, you
can also replace hyphens with doubled underscores:

    $avanode->snapmirror_share_unlock(share__name => $share);

If an AVA command takes multiple arguments, these are usually
position-sensitive, and should be passed in the order they would appear on the
command line:

    # Executes "snapmirror snapshot delete share-name X snapshot-name Y",
    # where the order of share-name and snapshot-name is important.
    $avanode->snapmirror_snapshot_delete(
        'share-name'    => $share,
        'snapshot-name' => $snapshot
    );

At present, there are no checks made to see if a command is valid before it is
executed. Likewise, those commands that take parameters do not currently have
any parameter validation. If an invalid command is given, an exception of the
B<NACL::APISet::Exceptions::CommandNotFoundException> class will be thrown. If
any other error occurs (such as invalid parameters), an exception of the class
B<NATE::BaseException> will be thrown, with the error message(s) from the
appliance.

=cut

###############################################################################
#
# Private methods
#
###############################################################################


# Specifies the method to use on a NACL::ResourceRecords object to fetch the
# resource records specific to AVA nodes:
sub resource_record_method_name {
    return 'avanodes';
}

# For this command interface, what is the common prefix for test params?
sub _ci_to_test_param_mapping {
    return 'AVA';
}

# Provide the validation spec entries that are specific to the primary key for
# this class:
sub _primary_keys_validate_spec {
    return (node => { type => SCALAR });
}

1;
