# $Id: $

#
# Copyright (c) 2001-2015 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary AggregateSnapshot Component Module
## @author sangavai@netapp.com dl-nacl-dev@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::C::AggregateSnapshot

=head1 DESCRIPTION

C<NACL::C::AggregateSnapshot> is a derived class of
L<NACL::C::Component::ONTAP|lib-NACL-C-Component-ONTAP-pm>.  It represents access to an
ONTAP AggregateSnapshot.  A related class is
L<NACL::CS::AggregateSnapshot|lib-NACL-CS-AggregateSnapshot-pm>, which represents
the past state of an ONTAP AggregateSnapshot.

=head1 ATTRIBUTES

=head2 command_interface

See L<NACL::C::Component::command_interface|lib-NACL-C-Component-pm/command_interface>

=head2 snapshot

The name of the snapshot, used to identify it uniquely.  (AggregateSnapshot
names are assumed to be unique within an aggregate)

=head2 aggregate

The name of the aggregate.

=cut

package NACL::C::AggregateSnapshot;
use strict;
use warnings;
use NACL::CS::AggregateSnapshot;
use NACL::CS::AggregateSnapshotSchedGet;
use NACL::ComponentUtils qw (_optional_scalars Dumper);
use base 'NACL::C::Component::ONTAP';
use Params::Validate qw(validate_with SCALAR SCALARREF ARRAYREF);
use NACL::C::Job;
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();

=head1 METHODS

=head2 new

  my $snapshot = NACL::C::AggregateSnapshot->new(command_interface=>$ci,..);

Return a new AggregateSnapshot component object with the given attributes.

=cut

use Class::MethodMaker [
    scalar => 'snapshot',
    scalar => 'aggregate',
];

=head2 create

  my $snapshot = NACL::C::AggregateSnapshot->create(command_interface=>$ci,
                                           aggregate => $aggregate_name,
                                           snapshot=>$snapshot_name,...);

(Class method) Request creation of a new snapshot element, and  return a new
snapshot component object that refers to it.  Does not wait for snapshot
creation to complete.  Uses a 7Mode-CLI APISet or Nodescope CLI APISet.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< snapshot=>$snapshot_name >>

(Required) Name of snapshot to create.

=item C<< aggregate=> $string >>

(Required) Name of aggregate to create snapshot in it.

=item C<< "method-timeout" => $scalar >>

(Optional)The new timeout value.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut


sub create {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    $pkg->_frontend_log_debug_opts(@_);
    my %opts = $pkg->_common_validate_with(
        params          => \@_,
    );
    my $ret = $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_create_7mode_cli',
                interface => 'CLI',
                set       => '7Mode|Nodescope',
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $ret;
} ## end sub create

sub _create_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};

    my %api_opts;
    $pkg->_hash_copy(
        source => \%opts,
        target => \%api_opts,
        copy   => [ qw(snapshot) ],
        map    => { aggregate => 'aggregate-name' },
    );
    $api_opts{aggregate} = 1;

    $pkg->_api_call_log_debug_opts(
        api      => 'snap_create',
        api_opts => \%api_opts,
    );

    $apiset->snap_create(%api_opts);
    $Log->exit() if $may_exit;
    return $pkg->new(
        command_interface => $command_interface,
        aggregate         => $opts{aggregate},
        snapshot          => $opts{snapshot}
    );
} ## end sub _create_7mode_cli


=head2 delete

  NACL::C::AggregateSnapshot->delete(command_interface=>$ci,
                             aggregate => $aggregate_name,
                             snapshot=>$snapshot_name,...);

                            or

  $snapshot_object->delete(...);

(Class or instance method) This method deletes an snapshot element.
Uses a 7Mode-CLI APISet or Nodescope CLI APISet.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< snapshot=>$snapshot_name >>

(Required for class method, Not Applicable for instance method)
Name of a snapshot.

=item C<< aggregate =>$aggregate_name >>

 (Required)Name of the aggregate

=item C<< "method-timeout" => $scalar >>

(Optional)The new timeout value.

=item C<< "quiet" => $boolean >>

(Optional).

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut


sub delete {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    $pkg_or_obj->_frontend_log_debug_opts(@_);
    my %opts       = $pkg_or_obj->_common_validate_with(
        params          => \@_,
        additional_spec => {
            _optional_scalars(
                qw(quiet )
            )
        }
    );

    my $ret =  $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_delete_7mode_cli',
                interface => 'CLI',
                set       => '7Mode|Nodescope',
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $ret;
} ## end sub delete

sub _delete_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my $allowed_7mode_options;

    my %snapshot_args;
    $snapshot_args{aggregate} = 1;

    if ($opts{snapshot} eq '*') {
        delete $opts{snapshot};
        $snapshot_args{all} = 1;
        $snapshot_args{force} = 1;
    }

    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%snapshot_args,
        copy   => [ qw(snapshot quiet) ],
        map    => { aggregate => 'aggregate-name' },
    );


    $pkg_or_obj->_api_call_log_debug_opts(
        api      => 'snap_delete',
        api_opts => \%snapshot_args,
    );

    $apiset->snap_delete(%snapshot_args);
    $Log->exit() if $may_exit;
} ## end sub _delete_7mode_cli

=head2 restore

   NACL::C::AggregateSnapshot->restore(
                                command_interface => $ci,
                                aggregate => $aggregate,
	                        snapshot => $snapshot,
                                  ...);
					 or

    $AggregateSnapshot_object->restore(...);

(Class or instance method)
This method is used to restore the aggregate to a snapshot..
Uses a 7Mode CLI APISet or Nodescope CLI APISet.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< snapshot=>$snapshot_name >>

(Required for class method, Not Applicable for instance method)
Name of a snapshot.

=item C<< aggregate =>$aggregate_name >>

(Required) Name of an aggregate.

=item C<< force => $boolean >>

=item C<< method-timeout => $string >>

(Optional)The new timeout value.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut


sub restore {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    $pkg_or_obj->_frontend_log_debug_opts(@_);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@_,
    );

    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => "_restore_7mode_cli",
                interface => "CLI",
                set       => "7Mode|Nodescope",
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $ret;
}

sub _restore_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset  = delete $opts{apiset};

    my %api_opts;

    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%api_opts,
        map    => { aggregate => 'aggregate-name',
                    snapshot  => 'snapshot-name',
        },
    );
    $api_opts{aggregate} = 1;
    $api_opts{force} = 1;

    $pkg_or_obj->_api_call_log_debug_opts(
        api      => 'snap_restore',
        api_opts => \%api_opts,
    );

    $apiset->snap_restore_aggr(%api_opts);
    $Log->exit() if $may_exit;
}## end sub _restore_7mode_cli

=head2 rename

   NACL::C::AggregateSnapshot->rename(
                                command_interface => $ci,
	                             aggregate => $aggregate,
	                             'snapshot' => $snapshot,
	                             'new-name' => $snapshot_new,
                                  ...);

								 or

    $AggregateSnapshot_object->rename(...);

(Class or instance method)
This method is used to delete a snapshot.
Uses a 7Mode CLI APISet or Nodescope CLI APISet.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< snapshot=>$snapshot_name >>

(Required for class method, Not Applicable for instance method)
Name of a snapshot.

=item C<< aggregate => $string >>

(Required) Name of aggregate of which the snapshot to be created.

=item C<< snapshot => $string >>

=item C<< new-name => $string >>

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut


sub rename {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    $pkg_or_obj->_frontend_log_debug_opts(@_);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@_,
        additional_spec => {
            _optional_scalars( qw( new-name ) ),
        },
    );

    my $ret =  $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_rename_7mode_cli',
                interface => 'CLI',
                set       => '7Mode|Nodescope',
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $ret;
}

sub _rename_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset    = delete $opts{apiset};


    my %api_opts;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%api_opts,
        copy   => [ qw( new-name) ],
        map    => { aggregate => 'aggregate-name' , snapshot => 'current-name'},
    );
    $api_opts{aggregate} = 1;

    $pkg_or_obj->_api_call_log_debug_opts(
        api      => 'snap_rename',
        api_opts => \%api_opts,
    );

    $apiset->snap_rename(%api_opts);
    $pkg_or_obj->snapshot($api_opts{'new-name'}) if (ref $pkg_or_obj);

    $Log->exit() if $may_exit;
} ## end sub _rename_7mode_cli

=head2 reserve

   NACL::C::AggregateSnapshot->reserve(
                                command_interface => $ci,
                                aggregate => $aggregate_name,
                                  ...);

								 or

    $AggregateSnapshot_object->reserve(...);

(Class or instance method)
This method is used to reserve a snapshot.
Uses a 7Mode CLI APISet or Nodescope CLI APISet.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< aggregate => $aggregate >>

 Name of the aggregate

=item C<< percentage => $string >>

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut


sub reserve {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    $pkg_or_obj->_frontend_log_debug_opts(@_);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@_,
        additional_spec => { _optional_scalars( qw(percentage snapshot) ) },
    );

    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => "_reserve_7mode_cli",
                interface => "CLI",
                set       => "7Mode|Nodescope",
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $ret;
}

sub _reserve_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset    = delete $opts{apiset};


    my %api_opts;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%api_opts,
        copy   => [ qw( percentage) ],
        map    => { aggregate => 'aggregate-name' },
    );
    $api_opts{aggregate} = 1;

    $pkg_or_obj->_api_call_log_debug_opts(
        api      => 'snap_reserve',
        api_opts => \%api_opts,
    );

    $apiset->snap_reserve(%api_opts);
    $Log->exit() if $may_exit;
} ## end sub _reserve_7mode_cli

=head2 reserve_get

    my $AggregateSnapshotReserveGet_object = NACL::C::AggregateSnapshot->reserve_get(
                                    command_interface => $ci,
                                    aggregate         => $aggregate,
                                    ...);

                     (or)

    my @AggregateSnapshotReserveGet_objects = NACL::C::AggregateSnapshot->reserve_get(command_interface => $ci);

                      (or)

   my $AggregateSnapshotReserveGet_object = NACL::C::AggregateSnapshot->reserve_get(command_interface => $ci);

(Class method)
This method is used to query the reserved percentage or size of one or all aggregate and
returns an object of type L<NACL::CS::AggregateSnapshotReserveGet|lib-NACL-CS-AggregateSnapshotReserveGet-pm>
Uses 7Mode CLI APISet

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< aggregate => $string >>

(Optional) Name of aggregate.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut


sub reserve_get {
    $Log->enter() if $may_enter;
    my $pkg = shift;
    $pkg->_frontend_log_debug_opts(@_);
    my %opts = $pkg->_common_validate_with(
        params              => \@_,
        ignore_primary_keys => 1,
        additional_spec     => { _optional_scalars( qw(aggregate) ) }
    );
    my @CS_objs = $pkg->call_on_apiset( %opts,
                  choices => [
                             {   method    => "_reserve_get_7mode_cli",
                                 interface => "CLI",
                                 set       => "7Mode|Nodescope",
                             },
                   ],
    );
    $Log->exit() if $may_exit;
    return wantarray ? @CS_objs : $CS_objs[0];
}

sub _reserve_get_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );

    my $command_interface = delete $opts{command_interface};
    my $apiset    = delete $opts{apiset};

    my %reserve_get_opts;
    $reserve_get_opts{aggregate} = 1;
    if (defined $opts{aggregate}) {
        $reserve_get_opts{'aggregate-name'} = $opts{aggregate};
    }

    $pkg->_api_call_log_debug_opts(
        api      => 'snap_reserve',
        api_opts => \%reserve_get_opts,
    );

    my $response = $apiset->snap_reserve(%reserve_get_opts);
    my $output = $response->get_parsed_output();
    require NACL::CS::AggregateSnapshotReserveGet;
    my @state_objs;
    foreach my $row (@$output) {
        my $obj = NACL::CS::AggregateSnapshotReserveGet->new(command_interface => $command_interface);
        my %aggr_row = %$row;

        # Translate "aggregate_name" to "aggregate", and in the value of
        # reserve, convert "k-bytes." to "KB" for consistency with other methods
        $aggr_row{aggregate} = delete $aggr_row{aggregate_name};
        $aggr_row{reserve} =~ s/k\-bytes./KB/;
        $obj->_set_fields(row => \%aggr_row);
        push @state_objs, $obj;
    }

    $Log->exit() if $may_exit;
    return wantarray ? @state_objs : $state_objs[0];
    #return @state_objs;
}

=head2 sched

   NACL::C::AggregateSnapshot->sched(
                                command_interface => $ci,
                                aggregate => $aggregate_name,
                                  ...);

								 or

    $AggregateSnapshot_object->sched(...);

(Class method)
This method is used to schedule a snapshot.
Uses a 7Mode CLI APISet or Nodescope CLI APISet.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< aggregate => $string >>

(Required) Name of the aggregate

=item C<< weeks => $string >>

=item C<< days => $string >>

=item C<< hours => $string >>

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut


sub sched {
    $Log->enter() if $may_enter;
    my $pkg = shift;
    $pkg->_frontend_log_debug_opts(@_);
    my %opts = $pkg->_common_validate_with(
        params          => \@_,
        additional_spec => {
             'hour-list' => { type => ARRAYREF, optional => 1 },
            _optional_scalars(
                qw(weeks days hours snapshot )
            ),
        },
    );

    my $ret = $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => "_sched_7mode_cli",
                interface => "CLI",
                set       => "7Mode|Nodescope",
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $ret;
}

sub _sched_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset    = delete $opts{apiset};

    my %api_opts;
    if ($opts{'hour-list'}) {
        $opts{'hour-list'} = join ',', @{$opts{'hour-list'}};
        $opts{'hours'} = $opts{'hours'}.'@'.$opts{'hour-list'};
    }

    if(($opts{hours} =~ /\@/) && (!defined($opts{weeks}))){
          $opts{weeks} = 0;
    }
    $pkg->_hash_copy(
        source => \%opts,
        map    => { aggregate => 'aggregate-name' },
        copy   => [ qw(weeks days hours) ],
        target => \%api_opts,
    );
    $api_opts{aggregate} = 1;

    $pkg->_api_call_log_debug_opts(
       api      => 'snap_sched',
       api_opts => \%api_opts,
    );


    $apiset->snap_sched(%api_opts);
    $Log->exit() if $may_exit;
} ## end sub _sched_7mode_cli

=head2 sched_get

    my $AggregateSnapshotSchedGet_object = NACL::C::AggregateSnapshotSchedGet->sched_get(
                                    command_interface => $ci,
                                    aggregate         => $aggregate,
                                    ...);

                     (or)

    my @AggregateSnapshotSchedGet_objects = NACL::C::AggregateSnapshot->sched_get(command_interface => $ci);

                      (or)

   my $AggregateSnapshotSchedGet_object = NACL::C::AggregateSnapshot->sched_get(command_interface => $ci);


(Class method)
This method is used to query the scheduled hourly,weekly, daily status of an aggregate and
returns an object of type L<NACL::CS::AggregateSnapshotSchedGet|lib-NACL-CS-AggregateSnapshotSchedGet-pm>
The hourly status can be either number_of_hourly_snap or number_of_hourly_snap@list_of_hourly snap
Uses 7Mode CLI APISet

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< aggregate => $string >>

(Optional) Name of aggregate.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut


sub sched_get {
    $Log->enter() if $may_enter;
    my $pkg = shift;
    $pkg->_frontend_log_debug_opts(@_);
    my %opts = $pkg->_common_validate_with(
        params              => \@_,
        ignore_primary_keys => 1,
        additional_spec     => { _optional_scalars( qw(aggregate snapshot) ) }
    );
    my @CS_objs = $pkg->call_on_apiset( %opts,
                  choices => [
                             {   method    => "_sched_get_7mode_cli",
                                 interface => "CLI",
                                 set       => "7Mode|Nodescope",
                             },
                   ],
    );
    $Log->exit() if $may_exit;
    return wantarray ? @CS_objs : $CS_objs[0];
}

sub _sched_get_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );

    my $command_interface = delete $opts{command_interface};
    my $apiset    = delete $opts{apiset};

    my %sched_get_opts;
    $sched_get_opts{aggregate} = 1;
    if (defined $opts{aggregate}) {
        $sched_get_opts{'aggregate-name'} = $opts{aggregate};
    }

    $pkg->_api_call_log_debug_opts(
        api      => 'snap_sched',
        api_opts => \%sched_get_opts,
    );

    my $response = $apiset->snap_sched(%sched_get_opts);
    my $output = $response->get_parsed_output();
    require NACL::CS::AggregateSnapshotSchedGet;
    my @state_objs;
    foreach my $row (@$output) {
        my $obj = NACL::CS::AggregateSnapshotSchedGet->new(command_interface => $command_interface);
        my %aggr_row = %$row;

        my $hours_from_output = $row->{hours};
        my $weeks_from_output = $row->{weeks};
        my $days_from_output = $row->{days};
        my $hour_list = [];
        if ($hours_from_output =~ /@/) {
            $hours_from_output =~ /(\d+)@(.*)/;
            $hours_from_output = $1;
            my $hour_list_str = $2;
            my @hour_list_arr = split ',', $hour_list_str;
            $hour_list = \@hour_list_arr;
        }
        $aggr_row{aggregate} = delete $aggr_row{aggregate_name};
        $aggr_row{hours} = $hours_from_output;
        $aggr_row{weeks} = $weeks_from_output;
        $aggr_row{days} = $days_from_output;
        $aggr_row{hour_list} = $hour_list;
        $obj->_set_fields(row => \%aggr_row);
        push @state_objs, $obj;

    }
    $Log->exit() if $may_exit;
    return @state_objs;

}

__PACKAGE__->_load_event_class();

sub _primary_keys_validate_spec {
    return (
        snapshot  => { type => SCALAR },
        aggregate => { type => SCALAR},
    );
} ## end sub _primary_keys_validate_spec
1;
