
#
# Copyright (c) 2001-2013 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Represents a client machine with mgmt interface and IP address
## @author madhav.shekar@netapp.com, dl-nacl-dev@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::C::Client

=head1 DESCRIPTION

C<NACL::C::Client> is a derived class of
L<NACL::C::CommandInterface|lib-NACL-C-CommandInterface-pm>.  It represents 
access to a Client (which is a Linux or Windows box). It can currently
only be used as a command interface for other components, though we do
plan to flesh it out into a component. That is why there is currently no
C<NACL::CS::Client>.

=head1 ATTRIBUTES

=head2 client

The name of the client, used to identify it uniquely.

=cut

package NACL::C::Client;

use strict;
use warnings;

use Params::Validate qw(validate_with SCALAR);
use NACL::ComponentUtils qw (_optional_scalars Dumper);
use NATE::BaseException qw(:try);
use NATE::Exceptions::Argument qw(:try);
use NACL::Exceptions::NoElementsFound;
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use NATE::Time qw(timeout2time parse_date_time);
use NATE::ParamSet qw(param_global);

use base qw(NACL::C::CommandInterface NACL::C::Component::Client);

=head1 METHODS

=head2 new

Returns a new Client component object with the given attributes.

    my $client = NACL::C::Client->new([name => $name],
                                      [client => $client],
                                      [hostrec => $hostrec]);

Specifying any one of the client, name or hostrec attributes will populate
the other two.

    my $client = NACL::C::Client->new(
        client                   => 'localhost',
        use_localhost_shell_conn => 1
    );

The default behaviour with Clients is to establish a connection through
rlogin/telnet. However, when the client is "localhost" (i.e. itself),
it is overkill to create a connection to itself -- there is overhead
(both in terms of CPU and memory) for establishing the connection and in
sending/receiving commands over a connection. Instead, communicating with
"localhost" simply requires firing up a new shell. Using a new shell eliminates
the overheads mentioned, so uses less CPU, less memory and commands are
executed quicker. Enabling this behaviour can be done by specifying
C<< use_localhost_shell_conn => 1 >> in the call to Client->new().

An alternative is to specify the command-line parameter
"NACL_CLIENT_USE_LOCALHOST_SHELL_CONN". This will have the effect of applying
it on all localhost Client objects, so it will not be necessary to modify
any scripts.

    ntest script.thpl NACL_CLIENT_USE_LOCALHOST_SHELL_CONN=1

=cut

use Class::MethodMaker [
    scalar => [
        {   '-default_ctor' => sub { $_[0]->_construct_primary_key(); },
            '-store_cb' =>
                sub { $_[0]->_verify_hostrec_callback( $_[1], $_[2] ); }
        },
        'client'
    ]
];

# Switch to 1 when we want to enable it by default.
my $default_localhost_shell = 0;

sub init {
    $Log->enter();
    my ($self, %opts) = @_;
    my $use_localhost_shell_conn = delete $opts{use_localhost_shell_conn};
    $self->SUPER::init(%opts);
    if ($self->name() eq 'localhost' && !$self->apiset_should_exists('connection')) {
        # By default, client connections on Linux use rlogin. However, when the
        # host is localhost (i.e. itself), it's unnecessary to have to "connect"
        # to itself. All that's needed to talk to itself is a new shell.
        # Eventually it would be nice if NATE handled this (see burt 603598).
        # Also, do this only if apiset_should->{connection} isn't already set.
        $use_localhost_shell_conn //= param_global()->get('NACL_CLIENT_USE_LOCALHOST_SHELL_CONN')
            // $default_localhost_shell;
        if ($use_localhost_shell_conn) {
            my $conn = $self->hostrec()->connect(
                template => "local",
                conntype => "rlogin",
                conncmd => ["/bin/bash","-i"]
            );
            $self->apiset_should(connection => $conn);
        }
    }
    $Log->exit();
}

=head1 METHODS

=head2 find

    my $Client  = NACL::C::Client->find(...);
    my @Clients = NACL::C::Client->find(...);

(Class method)

Finds all Clients and returns command interface objects for each (can also
be called in scalar context, in which case it returns the first client found).
Refer to L<find|lib-NACL-C-CommandInterface-pm/find> for more details.

You can also refer to user guide for more information on the options,
order of precedence for command line parameters (in case of find() is
invoked without specifying a command_interface) and examples.
L<http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/ComponentLayerUsersGuide/index.html#Obtaining_command_interface_objects>

Unlike other command interfaces, you cannot call 'find' of Client with
a command interface.

B<NOTE:> If a specific type of host is required (i.e. is UNIX or
is Windows), then it is better to use either L<find_unix_clients|lib-NACL-C-Client-pm/find_unix_clients>
or L<find_windows_clients|lib-NACL-C-Client-pm/find_windows_clients>, rather than assuming that Client->find()
returns the required type of host.

=head2 find_unix_clients

    my $unix_client  = NACL::C::Client->find_unix_clients();

    my @unix_clients = NACL::C::Client->find_unix_clients();

Returns objects for UNIX clients. For example, if CLIENTS was provided as
CLIENTS=winclient1,unixclient1,unixclient2,winclient2, then this method
would return objects only for unixclient1 and unixclient2.

This supports the standard facilities provided by C<find> such as
supporting scalar/list context and accepting C<allow_empty>.

=cut

sub find_unix_clients {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    my @unix_clients = $pkg->_find_clients_helper(@args, type => 'unix');
    $Log->exit() if $may_exit;
    return wantarray ? @unix_clients : $unix_clients[0];
}

=head2 find_windows_clients

    my $windows_client  = NACL::C::Client->find_windows_clients();

    my @windows_clients = NACL::C::Client->find_windows_clients();

Returns objects for windows clients. For example, if CLIENTS was provided as
CLIENTS=winclient1,unixclient1,unixclient2,winclient2, then this method
would return objects only for winclient1 and winclient2.

This supports the standard facilities provided by C<find> such as
supporting scalar/list context and accepting C<allow_empty>.

=cut

sub find_windows_clients {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    my @win_clients = $pkg->_find_clients_helper(@args, type => 'windows');
    $Log->exit() if $may_exit;
    return wantarray ? @win_clients : $win_clients[0];
}

=head2 find_ontap_deploy_clients

    my $ontap_deploy_client = NACL::C::Client->find_ontap_deploy_clients();

    my @ontap_deploy_client = NACL::C::Client->find_ontap_deploy_clients();

Returns objects for ONTAP Deploy clients. For example, if CLIENTS was provided as
CLIENTS=ONTAPDeployclient1,ONTAPDeployclient2,unixclient1,winclient2, then this method
would return objects only for ONTAPDeployclient1 and ONTAPDeployclient2.

This supports the standard facilities provided by C<find> such as
supporting scalar/list context and accepting C<allow_empty>.

=cut

sub find_ontap_deploy_clients {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    my @win_clients = $pkg->_find_clients_helper(@args, type => 'ontap-select-deploy');
    $Log->exit() if $may_exit;
    return wantarray ? @win_clients : $win_clients[0];
}

sub _find_clients_helper {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    my %opts = validate_with(
        params      => \@args,
        spec        => { type  => { type => SCALAR }, },
        allow_extra => 1
    );
    my $type = delete $opts{type};

    my @all_clients = $pkg->find(%opts);
    my @specific_clients;
    foreach my $client (@all_clients) {
        if ($client->hostrec()->hosttype() =~ /$type/ || ($type =~ /unix/ && $client->hostrec()->is_unix())) {
            push @specific_clients, $client;
        }
    }
    
    if (!@specific_clients && !$opts{allow_empty}) {
        $Log->exit() if $may_exit;
        NACL::Exceptions::NoElementsFound->throw("No $type clients found");
    }
    
    $Log->exit() if $may_exit;
    return wantarray ? @specific_clients : $specific_clients[0];
}

=head2 default_apiset_must

All clients have to be of category "Host", so this method returns
a hash with key 'category' whose value is "Host".

Refer to L<default_apiset_must|lib-NACL-C-CommandInterface-pm/default_apiset_must>.

=cut

sub default_apiset_must {
    my $self     = shift;
    my $hostrec  = $self->hostrec();
    my $hostid   = $hostrec->id();
    my $hosttype = $hostrec->hosttype();
    ## check if the hostrec object is of type Hostrec::Client
    NATE::BaseException->throw( "Host $hostid was supposed "
      . "to be a client but is actually of type '$hosttype'" ) if(!$hostrec->isa('Hostrec::Client') && !$hostrec->isa('Hostrec::Appliance::ONTAPSelectDeploy'));

    # We might need to update this later to be more specific in its
    # checks if we create APISets for different flavours of Linux and have
    # components use those different flavours
    my $set;
    if ( $hosttype =~ /^unix-solaris$/ ) {
        $set = 'Solaris';
    } elsif ( $hosttype =~ /^unix-aix$/ ) {
        $set = 'AIX';
    } elsif ( $hosttype =~ /^unix-hpux$/ ) {
        $set = 'HPUX';
    } elsif ( $hosttype =~ /^unix/ ) {
        $set = "Linux|SMTT";
    } elsif ( $hosttype =~ /^windows/ ) {
        $set = "Windows|SMTT";
    } elsif ( $hostrec->isa('Hostrec::LanForge::Fire') ) {
        $set = "Linux|SMTT";
    } elsif ($hosttype =~ /ontap-select-deploy/) {
	$set = "ONTAPSelectDeploy";
    }

    my $source = "command interface is a NACL::C::Client "
        . "and hosttype='$hosttype'";
    return {
        category        => "Host",
        category_source => $source,
        set             => $set,
        set_source      => $source,
    };
}

=head2 default_apiset_could

The ruleset to default to if values for any of the co-ordinates 'category', 
'interface' or 'set' cannot be determined.

This method returns category=Host;interface=CLI;set=Linux

Refer to L<default_apiset_could|lib-NACL-C-CommandInterface-pm/default_apiset_could>.

=cut

sub default_apiset_could {
    return {
        category  => "Host",
        interface => "CLI",
        set       => "Linux",
    };
} ## end sub default_apiset_could

=head2 _create_version_manager

   sub get_version_manager {
        ...
        if (keys %$apisets) {
            ...
        } else {
            $version_manager = $self->_create_version_manager();
        }
        return $version_manager;
    }

(Instance method)

This private method returns the version manager of the appropriate type.
This decision is based on the category ("Host" for Client command interfaces)
and the set.

=cut

sub _create_version_manager {
    $Log->enter() if $may_enter;
    my $self = shift;

    use NACL::APISet::Version;
    my $hostrec      = $self->hostrec();
    my $category     = "Host";
    my $default_must = $self->default_apiset_must();
    my $shell_type   = $default_must->{set};
    $Log->debug( "Version manager object with parameters "
            . "category = $category; shell_type = $shell_type; "
            . "hostobj type = "
            . $hostrec->hosttype()
            . " created" );
    $Log->exit() if $may_exit;
    return NACL::APISet::Version->new(
        category   => $category,
        hostobj    => $hostrec,
        shell_type => $shell_type
    );
}

=head2 find

    my @clients = NACL::C::Client->find();
    my $client  = NACL::C::Client->find();

Finds all Clients and returns command interface objects for each (can also
be called in scalar context, in which case it returns the first client found).
Refer to L<find|lib-NACL-C-CommandInterface-pm/find> for more details.

Unlike other command interfaces, we cannot call 'find' of Client with 
a command interface. The list of Clients is populated by using test parameters
(CLIENT or CLIENTS or CLIENT1,CLIENT2...). If 'find' is called and 
command_interface is provided as parameter, a C<NATE::Exceptions::Argument> 
exception is thrown.

=cut

=head2 apiset

Fetches the apiset object stored in command_interface of Client.
When developing client agnostic libraries/scripts using components, 
it is also desired to obtain the apiset object in client agnostic way and 
this could be because the future method calls made on the APISet object are also client agnostic
(Eg: Invocation of utilities whose input and output remains the same across the clients).

my $client = NACL::C::Client->find();
# Returns a Linux or Windows CLI APIset based on whether your client is a
# Linux or Windows host.
my $client_apiset = $client->apiset();

=cut

# XXX Need to investigate whether an API exists which can be run on a client
# which would give a list of other clients.
# At the moment, we do not allow command_interface to be passed.
sub find {
    my ( $pkg, %opts ) = @_;
    if ( defined $opts{command_interface} ) {
        NATE::Exceptions::Argument->throw( "Argument 'command_interface'"
                . "should not be passed in call to 'find' for the 'Client' command"
                . " interface" );
    } else {
        $pkg->SUPER::find(%opts);
    }
} ## end sub find

=head2 type

    my $type = $client_obj->type();

Returns the hosttype of the client. Nothing more.

=cut

sub type {
    my $obj = shift;
    if ( ref $obj ) {
        my $return_type ;
        if($obj->hostrec()->isa('Hostrec::LanForge::Fire')){
           $return_type = 'unix-linux';
        }else{
           $return_type = $obj->hostrec()->hosttype();
        }
        return $return_type;
    } else {
        NATE::BaseException->throw( "The method type() can only be "
                . "invoked on a NACL::C::Client object" );
    }
}

=head2 find_clients_by_os

    my $client  = NACL::C::Client->find_clients_by_os();

    my @clients = NACL::C::Client->find_clients_by_os();

This is a class method only.

Returns objects for specified operating system. For example, if CLIENTS was provided as
CLIENTS=windows2003client1,unixclient1,unixclient2,windows2003client2,and
operating system is provided Windows 2003  then this method
would return objects only for windows2003client1 and windows2003client2.

This supports the standard facilities provided by C<find> such as
supporting scalar/list context and accepting C<allow_empty>.

=cut

sub find_clients_by_os {
     $Log->enter() if $may_enter;
     my ($pkg, @args) = @_;
     my %opts = validate_with(
         params      => \@args,
         spec        => { os  => { type => SCALAR }, },
         allow_extra => 1
     );
     my $os = delete $opts{os};
     my @all_clients = $pkg->find(%opts);
     my @specific_clients;
     foreach my $client (@all_clients) {
           if ($client->hostrec()->OS() =~ /$os/) {
                push @specific_clients, $client;
           }
     }

     if (!@specific_clients && !$opts{allow_empty}) {
           $Log->exit() if $may_exit;
           NACL::Exceptions::NoElementsFound->throw("No $os clients found");
     }

     $Log->exit() if $may_exit;
     return wantarray ? @specific_clients : $specific_clients[0];
} ## end sub find_clients_by_os


=head2 execute

    my $output = $client_obj->execute(
        command => "arbitrary command line",
        get_status => \$execution_status
    );

This method is a pass-through to 
L<execute_raw_command|lib-NACL-APISet-Host-CLI-pm/execute_raw_command> 
and takes the same parameters as Connectrec->execute().  See 
L<http://web.netapp.com/engineering/nate/releases/latest/html/Connectrec.html#execute> 
for more information.

Returns scalar of raw output

=over

=item get_status

SCALARREF value that contains the execution status of the command.
It is send as one of the inputs to $apiset->execute_raw_command(%args)
Based on the execution status this method logs a warning in case of command
exiting with status other than zero.

=back

=cut

sub execute {
    my ( $self, %opts ) = @_;
    $Log->enter() if $may_enter;
    $self->_verify_invocation( style => 'instance_only' );
    my $execution_status ;

    my $ignore_status = NATE::ParamSet::param_global()->get('ALLOW_EXIT_STATUS_IGNORE', default => 0);

    my $get_status_undef = 0 ;

    ## check if get_status is provided as undef
    ## $obj->execute(get_status => undef); -> This should not warn
    if(!defined $opts{get_status} && exists $opts{get_status}){
           $get_status_undef = 1 ;
           delete $opts{get_status};
    }

    ## one can turn off the warning in case if the exit status other than 0 is expected
    ## this will be useful in negative test scenarios
    ## following are the calls that will turn off the warning:
    ## ALLOW_EXIT_STATUS_IGNORE = 1 is provided as CLI param(higher precedence)
    ## $obj->execute(get_status => undef); -> get_status is given the undef val

    my $warn_flag = 0;
     if((!$ignore_status && !$get_status_undef)){
       ## exclude those commands where prompts,match_table or echo _OK_ is provided
       if(!defined $opts{additional_prompts} && !defined $opts{match_table}
             && !defined $opts{prompts} && ($opts{command} !~ /echo _OK_/i)) {
         $warn_flag = 1;
         $opts{get_status} //= \$execution_status ;
         $opts{command} =~ s/;\s*$// ;
       }
    }
    my $output = $self->apiset(interface => 'CLI')->execute_raw_command(%opts);

    my $get_status_after = ${$opts{get_status}} if (exists $opts{get_status});

    if($warn_flag){
       if( defined $get_status_after && $get_status_after != 0 ){
            my $warn_msg = "The get_status argument was not passed to execute call " .
                  "hence logging the warning (burt#670359) \n" .
                  "The command: \"$opts{command}\" exited with " .
                  "unsuccessful status: \"$get_status_after\" " .
                  "with output: \"$output\" ";
           $Log->warn( $warn_msg );
       }
    }
    $Log->exit() if $may_exit;
    return $output;
}

=head2 is_ipv6_capable

    my $boolean = $client_obj->is_ipv6_capable();

Check whether the client is IPv6 capable or not. This is done by checking
whether any of the client's interfaces is an IPv6 address.

Supported for Linux, Solaris and Windows.

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of the users guide.

=back

=back

=cut


sub is_ipv6_capable {
    $Log->enter() if $may_enter;
    my $self = shift;
    $self->_verify_invocation( style => 'instance_only' );
    my %opts = $self->_common_validate_with( params => \@_ );
    my $boolean = $self->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_is_ipv6_capable_linux',
                interface => 'CLI',
                set       => 'Linux'
            },
            {   method    => '_is_ipv6_capable_solaris',
                interface => 'CLI',
                set       => 'Solaris'
            },
            {   method    => '_is_ipv6_capable_windows',
                interface => 'CLI',
                set       => 'Windows'
            },
        ],
    );

    $Log->exit() if $may_exit;
    return $boolean;
}

sub _is_ipv6_capable_linux {
    $Log->enter() if $may_enter;
    my $self = shift;
    my %opts = $self->_validate_backend_opts(@_);

    my $apiset   = $opts{apiset};
    my $response = $apiset->ifconfig( all => 1 );
    my $output   = $response->get_parsed_output();
    foreach my $row (@$output) {
        if ( exists $row->{inet6_info} ) {
            $Log->exit() if $may_exit;
            return 1;
        }
    }

    $Log->exit() if $may_exit;
    return 0;
}

sub _is_ipv6_capable_solaris {
    $Log->enter() if $may_enter;
    my $self = shift;
    my %opts = $self->_validate_backend_opts(@_);

    my $apiset   = $opts{apiset};
    my $response = $apiset->ifconfig( all => 1 );
    my $output   = $response->get_parsed_output();
    foreach my $row (@$output) {
        if ( exists $row->{configuration} ) {
            foreach my $config ( @{ $row->{configuration} } ) {
                if ( exists $config->{inet6} ) {
                    $Log->exit() if $may_exit;
                    return 1;
                }
            }
        }
    }

    $Log->exit() if $may_exit;
    return 0;
}

sub _is_ipv6_capable_windows {
    $Log->enter() if $may_enter;
    my $self = shift;
    my %opts = $self->_validate_backend_opts(@_);

    my $apiset   = $opts{apiset};
    my $response = $apiset->ipconfig( all => 1 );
    my $output   = $response->get_parsed_output();
    foreach my $row (@$output) {
        if ( exists $row->{ipv6} ) {
            $Log->exit() if $may_exit;
            return 1;
        }
    }

    $Log->exit() if $may_exit;
    return 0;
}

=head2 is_nfsv4_capable

    my $boolean = $client_obj->is_nfsv4_capable();

Check whether the client is nfsv4 capable or not. This is done by checking
nfs version 4 support in client.

Supported for Linux.

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method> of the users guide.

=back

=back

=cut


sub is_nfsv4_capable {
    $Log->enter() if $may_enter;
    my $self = shift;
    $self->_verify_invocation( style => 'instance_only' );
    my %opts = $self->_common_validate_with( params => \@_ );
    my $boolean = $self->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_is_nfsv4_capable_linux',
                interface => 'CLI',
                set       => 'Linux'
            },
        ],
    );

    $Log->exit() if $may_exit;
    return $boolean;
}

sub _is_nfsv4_capable_linux {
    $Log->enter() if $may_enter;
    my $self = shift;
    my %opts = $self->_validate_backend_opts(@_);
    my @paths = ("/etc/idmapd.conf", "/sbin/mount.nfs4");
    my $output;
    my $return = 0;
    my $apiset   = $opts{apiset};
    try {
        foreach (@paths) {
            $apiset->ls( paths => $_ );
        }
        $output
            = $apiset->execute_command( command => "ps -aef | grep idmapd" );
        if ($output =~ /rpc.idmapd/) {
            $return = 1;
        }
    } catch NACL::APISet::Exceptions::CommandFailedException with {
    };
    $Log->exit() if $may_exit;
    return $return;
}

# Clients will have some state information, but we don't as yet have a need
# for it. For the time being, we'll just define it to do nothing.
# Once we flesh out what this state information is, we'll need to define
# a CS file for Client.
sub state { }

sub _primary_keys_validate_spec {
    return ( client => { type => SCALAR } );
}

sub resource_record_method_name {
    return "clients";
}

# For this command interface, what is the common prefix for the test parameters?
sub _ci_to_test_param_mapping {
    $Log->enter() if $may_enter;
    $Log->exit() if $may_exit;
    return 'CLIENT';
}

=head2 is_linux

    my $boolean = $client_obj->is_linux();

Returns a boolean value describing whether the command interface is for a
linux client or not.

=cut

sub is_linux {
    $Log->enter() if $may_enter;
    my $self = shift;
    my $res = ( $self->type() =~ /^unix-linux/ ) ? 1 : 0;
    $Log->exit() if $may_exit;
    return $res;
}

=head2 is_windows

    my $boolean = $client_obj->is_windows();

Returns a boolean value describing whether the command interface is for a
Windows client or not.

=cut

sub is_windows {
    $Log->enter() if $may_enter;
    my $self = shift;
    my $res = ( $self->type() =~ /^windows/ ) ? 1 : 0;
    $Log->exit() if $may_exit;
    return $res;
}

=head2 is_solaris

    my $boolean = $client_obj->is_solaris();

Returns a boolean value describing whether the command interface is for a
solaris client or not.

=cut

sub is_solaris {
    $Log->enter() if $may_enter;
    my $self = shift;
    my $res = ( $self->type() =~ /^unix-solaris/ ) ? 1 : 0;
    $Log->exit() if $may_exit;
    return $res;
}

=head2 client_ip

    my $client_obj = NACL::C::Client->new();
    my $ip_address = $client_obj->client_ip();

(Instance method)
Returns the ip-address of the client.

=cut

sub client_ip {
    $Log->enter() if $may_enter;
    my $self = shift;
    my $res  = $self->hostrec()->default_ip();
    $Log->exit() if $may_exit;
    return $res;
}

=head2 get_initiator_name

    $Client        = NACL::C::Client->find();
    my $initiator_name = $Client->get_initiator_name();

(Instance call)
This subroutine is required to fetch the initiator name of the
client passed. Supported for Linux, Solaris and Windows clients.

=over

=item Exceptions

=over

=item C<NACL::APISet::Exceptions::InvalidParamValueException>

This exception is thrown when the initiator name is not present in
the /etc/iscsi path on the Linux machine.
This exception is also thrown when there is a failure in fetching the
initiator-name from the windows machine.

=item C<NACL::APISet::Exceptions::InvalidParamException>

This exception is thrown when this method is called for a client
which is not linux or winodws.

=back

=back

=cut

sub get_initiator_name {
    $Log->enter() if $may_enter;
    my $client_obj = shift;
    my ($initiator, $response, $actual_file);
    my $apiset = $client_obj->apiset(interface => "CLI");

    if($client_obj->command_interface->hostrec->is_linux()) {
        # checking the initiatorname.iscsi location in client
        $apiset->set_privilege('privilege-level' => "root");
        $actual_file = "initiatorname.iscsi";
        try {
            $response = $apiset->cat( 'files' => "/etc/iscsi/$actual_file" );
        } catch NACL::APISet::Exceptions::InvalidParamValueException with {
            my $exception_object = shift;
            my $resp             = $exception_object->text();
            if ( $resp =~ /No such file or directory/i ) {
                my $file = $actual_file . ".bak";
                try {
                    $response = $apiset->cat( 'files' => "/etc/iscsi/$file" );
                    $apiset->cp(
                        "source" => "/etc/iscsi/$file",
                        "dest"   => "/etc/iscsi/$actual_file"
                    );
                    $apiset->execute_command(
                        command => "service iscsi restart");
                } catch NACL::APISet::Exceptions::InvalidParamValueException with {
                    my $exception_object = $_[0];
                    my $resp             = $exception_object->text();
                    if ( $resp !~ /No such file or directory/i ) {
                        $exception_object->throw();
                    }
                }
            }
        };
        my $initiator_name = $response->get_processed_output();
        if ( $initiator_name =~ /.*Name=(\w.*)/ ) {
            $initiator = $1;
            chomp($initiator);
        }
    } elsif( $client_obj->type() =~ /windows/ ) {
        try {
            my $output = $apiset->mpc_iscsi_show_adapters();
            $initiator = $output->{parsed_output}->[0]->{initiator_name};
        } catch NACL::APISet::Exceptions::ResponseException with {
            $Log->exit() if $may_exit;
            NACL::APISet::Exceptions::InvalidParamValueException->throw(
                "Failed to fetch the initiator name on the windows machine");
        };
    } elsif ( $client_obj->type() =~ /solaris/i ) {
        try {
        my $apiset = $client_obj->apiset( interface => "CLI" );
        my $response
            = $apiset->iscsiadm_list( "direct-object" => "initiator-node" );
        $response  = $response->get_parsed_output();
        $initiator = $response->[0]->{'initiator_node'}->[0]->{'name'};
        } catch NACL::APISet::Exceptions::ResponseException with {
            $Log->exit() if $may_exit;
            NACL::APISet::Exceptions::InvalidParamValueException->throw(
                "Failed to fetch the initiator name on the solaris machine");
        };
    } else {
        $Log->exit() if $may_exit;
        NACL::APISet::Exceptions::InvalidParamException->throw($client_obj->type()
            ." is not supported");
    }

    $Log->exit() if $may_exit;
    return $initiator;
}

=head2 add_paths_for_transition

    $Client = NACL::C::Client->find();
    $Client->add_paths_for_transition();

OR

    $Client->add_paths_for_transition(
        path => 'C:\ProgramData\7-Mode Transition Tool\bin;C:\\Program Files\\Java\\jre6\\bin',
    );

(Instance call)
This subroutine is used to set the PATH environment variable for 7-Mode transition
tool.

There are two ways to invoke the method.

1) Without providing the parameters i.e "path"

2) With providing the parameters i.e "path"

In the first case default path for transition tool will be set i.e.
'C:\Program Files\NetApp\7-Mode Transition Tool\bin;'.

In the second case user defined path will be set. The path till the transition
binary has to be supplied. Multiple paths can be set by separating the paths with ';'.

Supported for Windows clients

=cut

sub add_paths_for_transition {
    $Log->enter() if $may_enter;
    my ($client_obj,%path_to_bin) = @_;
    $client_obj->apiset(interface => "CLI")->add_paths_for_transition(%path_to_bin);
    $Log->exit() if $may_exit;
}

sub delete_file_or_directory {
    $Log->enter() if $may_enter;
    my ($self, @args) = @_;
    $self->_verify_invocation( style => 'instance_only' );
    my %opts = $self->_common_validate_with(
        params => \@args,
        additional_spec => {
            filepath => {type => SCALAR},
        }
    );
    if ($self->is_directory(%opts)) {
        $self->delete_directory(%opts);
    } else {
        $self->delete_file(%opts);
    }

    $Log->exit() if $may_exit;
}

sub delete_directory {
    $Log->enter() if $may_enter;
    my ($self, @args) = @_;
    $self->_verify_invocation( style => 'instance_only' );
    my %opts = $self->_common_validate_with(
        params => \@args,
        additional_spec => {
            filepath => {type => SCALAR},
        }
    );
    my $boolean = $self->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_delete_directory_linux',
                interface => 'CLI',
                set       => 'Linux|Solaris'
            },
            {   method    => '_delete_directory_windows',
                interface => 'CLI',
                set       => 'Windows'
            },
        ],
    );

    $Log->exit() if $may_exit;
    return $boolean;
}

sub _delete_directory_linux {
    $Log->enter() if $may_enter;
    my ($self, @opts) = @_;
    my %args = $self->_validate_backend_opts(@opts);

    my $apiset = $args{apiset};
    my $resp = $apiset->ls(
        paths => $args{filepath},
        'comma-separated' => 1,
    );
    my $raw_output = $resp->get_raw_output;
    chomp($raw_output);
    my @files = split(/\,\s+/, $raw_output);
    foreach my $file (@files) {
        my $path = $args{filepath} . "/" . $file;
        $self->delete_file_or_directory(filepath => $path);
    }
    
    $apiset->rmdir(
        paths => $args{filepath},
        "ignore_fail_on_non_empty" => 1,
    );

    $Log->exit() if $may_exit;
}

sub _delete_directory_windows {
    $Log->enter() if $may_enter;
    my ($self, @opts) = @_;
    my %args = $self->_validate_backend_opts(@opts);

    my $apiset = $args{apiset};
    $apiset->rmdir(
        'recursive'  => 1,
        'quiet-mode' => 1,
        'path'       => $args{filepath}
    );

    $Log->exit() if $may_exit;
}

sub delete_file {
    $Log->enter() if $may_enter;
    my ($self, @args) = @_;
    $self->_verify_invocation( style => 'instance_only' );
    my %opts = $self->_common_validate_with(
        params => \@args,
        additional_spec => {
            filepath => {type => SCALAR},
        }
    );
    my $boolean = $self->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_delete_file_linux',
                interface => 'CLI',
                set       => 'Linux|Solaris'
            },
            {   method    => '_delete_file_windows',
                interface => 'CLI',
                set       => 'Windows'
            },
        ],
    );

    $Log->exit() if $may_exit;
    return $boolean;
}

sub _delete_file_linux {
    $Log->enter() if $may_enter;
    my ($self, @opts) = @_;
    my %args = $self->_validate_backend_opts(@opts);

    my $apiset = $args{apiset};
    $apiset->rm( paths => $args{filepath}, force => 1 );
    
    $Log->exit() if $may_exit;
}

sub _delete_file_windows {
    $Log->enter() if $may_enter;
    my ($self, @opts) = @_;
    my %args = $self->_validate_backend_opts(@opts);

    my $apiset = $args{apiset};
    $apiset->del(
        'force-delete' => 1,
        'quiet-mode'   => 1,
        files          => $args{filepath}
    );

    $Log->exit() if $may_exit;
}

sub is_directory {
    $Log->enter() if $may_enter;
    my ($self, @args) = @_;
    $self->_verify_invocation( style => 'instance_only' );
    my %opts = $self->_common_validate_with(
        params => \@args,
        additional_spec => {
            filepath => {type => SCALAR},
        }
    );
    my $boolean = $self->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_is_directory_linux',
                interface => 'CLI',
                set       => 'Linux|Solaris'
            },
            {   method    => '_is_directory_windows',
                interface => 'CLI',
                set       => 'Windows'
            },
        ],
    );

    $Log->exit() if $may_exit;
    return $boolean;
}

sub _is_directory_linux {
    $Log->enter() if $may_enter;
    my ($self, @opts) = @_;
    my %args = $self->_validate_backend_opts(@opts);

    my $apiset = $args{apiset};
    my $file_or_dir = $args{filepath};
    my $cmd = "file $file_or_dir";
    my $is_dir = 0;
    my $out = $apiset->execute_raw_command(command => $cmd);
    if ($out =~ /$file_or_dir:\s+directory/) {
        $is_dir = 1;
    }
    
    $Log->exit() if $may_exit;
    return $is_dir;
}

sub _is_directory_windows {
    $Log->enter() if $may_enter;
    my ($self, @opts) = @_;
    my %args = $self->_validate_backend_opts(@opts);

    my $apiset = $args{apiset};
    my $file_or_dir = $args{filepath};
    my $is_dir = 1;
    # Couldnt find a better way to determine whether
    # passed file is a file or Dir
    my $cmd = "cd $file_or_dir";
    my $out = $apiset->execute_raw_command(command => $cmd);
    if ($out =~ /The directory name is invalid/) {
        $is_dir = 0;
    }

    $Log->exit() if $may_exit;
    return $is_dir;
}

=head2 get_lun_ostype

    $Client        = NACL::C::Client->find();
    my $ostype = $Client->get_lun_ostype();

(Instance call)
This subroutine is required to fetch the hosttype of the
client passed and returns ostype value understood (burt:688898)
by the "lun create" command. Supported for Linux,Solaris and
Windows clients.

=over

=item Exceptions

=over

=item C<NACL::APISet::Exceptions::InvalidParamException>

This exception is thrown when this method is called for a client
which is not linux, solaris or winodws.

=back

=back

=cut

sub get_lun_ostype {
    $Log->enter() if $may_enter;
    my $client = shift;
    my $ostype = $client->type;
    if ($ostype =~ /linux/) {
        $ostype = 'linux';
    } elsif ($ostype =~ /solaris/) {
        $ostype = 'solaris';
    } elsif ($ostype =~ /windows/) {
        $ostype = 'windows';
    } elsif ($ostype =~ /hpux/) {
        $ostype = 'hpux';
    } elsif ($ostype =~ /vmware/) {
        $ostype = 'vmware';
    } elsif ($ostype =~ /aix/) {
        $ostype = 'aix'; 
    } elsif ($ostype =~ /unix-/) {
        $ostype = 'linux';
    }else {
        $Log->exit() if $may_exit;
        NATE::Exceptions::Argument->throw( $client->type() 
        ." is not supported" );
    }
    $Log->exit() if $may_exit;
    return $ostype;
}

=head2 uptime

    my $result = $client_obj->uptime();

Check the uptime of clients.

Supported for Linux/Unix.

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method> of the users guide.

=back

=back

=cut

sub uptime {
    $Log->enter() if $may_enter;
    my ($self, @args) = @_;
    my %opts = $self->_common_validate_with( params => \@args );
    my $result = $self->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_uptime_solaris',
                interface => 'CLI',
                set       => 'Solaris'
            },
            {   method    => '_uptime_linux',
                interface => 'CLI',
                set       => 'Linux'
            },
           ],
    );
    $Log->exit() if $may_exit;
    return $result;
} ## end sub uptime

#======================================================================
# we are using /proc/uptime file to get uptime because it returns
# uptime in seconds. While uptime doesnt return doesnt.
#
# Contents of /proc/uptime looks like. First field is the uptime in
# seconds and second is idle time in seconds
#######################################################################
## cat /proc/uptime
## 145956.15 135539.54
#######################################################################
sub _uptime_linux {
    my ($self, @opts) = @_;
    my %args = $self->_validate_backend_opts(@opts);
    my $apiset = $args{apiset};
    my $uptime = $apiset->execute_raw_command(
        command => 'cat /proc/uptime | awk \'{print $1}\'',
    );

    return $uptime;
} ## end sub _uptime_unix


## Few examples of output of uptime command on other Unix machines ####
# 05:45:37 up  1:27,  2 users,  load average: 0.00, 0.00, 0.00
# 04:05:31 up 20 min,  2 users,  load average: 1.71, 1.23, 0.57
# 03:43:10 up 8 days,  7:37,  2 users,  load average: 0.68, 0.26, 0.12
#======================================================================
sub _uptime_solaris {
    my ($self, @opts) = @_;
    my %args = $self->_validate_backend_opts(@opts);
    my $apiset = $args{apiset};
    my $uptime = $apiset->execute_raw_command(command => 'uptime');
    
    $uptime = parse_date_time(format => 'uptime', value => $uptime);
    
    return $uptime;
} ## end sub _uptime_unix

=head2 startup_time

    my $startup_time = NACL::C::Client->startup_time(
        command_interface => $command_interface,
    );

OR

    my $startup_time = $client_obj->startup_time();

Instance or Class method. Returns the start time for client machines. It will return an absolute 
Time::HiRes::time.

Supported for Windows.

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method> of the users guide.

=back

=back

=cut

sub startup_time {
    $Log->enter() if $may_enter;
    my ($self, @args) = @_;
    my %opts = $self->_common_validate_with( params => \@args );
    my $result = $self->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_startup_time_windows',
                interface => 'CLI',
                set       => 'Windows'
            }, 
            {   method    => '_startup_time_unix',
                interface => 'CLI',
                set       => 'Linux|Solaris'
            },
           ],
    );
    $Log->exit() if $may_exit;
    return $result;
} ## end sub uptime

#=======================================================================================
#This method is used to parse the uptime of windows based client
#Few uptime example output-
#C:\Documents and Settings\Administrator>net statistics server | find "Statistics since"
#Statistics since 7/23/2014 8:08 AM
#C:\>net statistics server | find "Statistics since"
#Statistics since 23-06-2014 09:50:59
#=======================================================================================
sub _startup_time_windows {
    my ($self, @opts) = @_;
    my %args = $self->_validate_backend_opts(@opts);
    
    my $apiset = $args{apiset};
    my $cmd = 'net statistics server | find "Statistics since" ';
    my $start_time = $apiset->execute_raw_command(command => $cmd);
    
    my $startup_time;
    if(defined $start_time and $start_time=~/Statistics\s+since/) {
        ($startup_time) = ($start_time=~/Statistics\s+since\s+(\S+\s+\S+\s+\S+)/);
        $startup_time = timeout2time($startup_time);
    } else {
        NATE::BaseException->throw("Failed to get windows client startup time");
    }

    return $startup_time;
} ## end sub _windows_uptime

sub _startup_time_unix {
    my ($self, @opts) = @_;
    my %args = $self->_validate_backend_opts(@opts);
    
    my $apiset = $args{apiset};
    my $uptime = $self->uptime();
    my $date = $apiset->execute_raw_command(command => 'date');
    my $now = parse_date_time(format => 'date', value => $date);
    my $startup_time = $now - $uptime;
    
    return $startup_time;
}

=head2 reboot

    NACL::C::Client->reboot(
        command_interface=>$ci,
    );

OR

    $client->reboot();

(Class OR Instance method)
This method is used to reboot a client. It will be using "shutdown"
command to reboot the machine

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< Options accepted by APISet >>

(Optional) It accepts all the options accepted by respective apisets
L<NACL::APISet|lib-NACL-APISet-pm>

=back

=back

=cut

sub reboot {
    my ($pkg, @args) = @_;
    $Log->enter() if $may_enter;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw (
                    no-sync dry-run no-wtmp-record force
                    shutdown-interfaces harddrive-standby poweroff
                    gui log-off shutdown restart-application
                    turn-off hibernate unexpected-shutdown timeout-shutdown
                    target reason-shutdown cancel-shutdown comment-shutdown
                    crash-dump fast-reboot suppress-messages no-logging quick
                )
            ),
        },
    );

    my $ci = $opts{command_interface};
    my $file_path = $opts{path};
    try {
        $pkg->call_on_apiset(
            %opts,
            choices => [
                {   method    => "_reboot_windows_cli",
                    interface => "CLI",
                    set       => "Windows"
                },
                {   method    => "_reboot_linux_cli",
                    interface => "CLI",
                    set       => "Linux"
                }, 
                {   method    => "_reboot_solaris_cli",
                    interface => "CLI",
                    set       => "Solaris"
                },
            ],
        );
    } catch NACL::APISet::Exceptions::TimeoutException with {
        # Its expected. Don't do anything
    } catch NATE::BaseException with {
        my $exception = shift;
        my $text      = $exception->text();
        if ($text =~ /Unexpected EOF while reading output from command|Timeout waiting for command to complete|Could not write to connection/i){
           # Its expected. Don't do anything
        } else {
            $Log->exit() if $may_exit;
            $exception->throw();
        }
    };

    $Log->exit() if $may_exit;
}

sub _reboot_windows_cli {
    my ($pkg, @args) = @_;
    $Log->enter() if $may_enter;

    my $map = {
        'force' => "force-shutdown",
    };
    my $api = 'shutdown';
    

    $pkg->_base_backend_client(
        @args,
        api => $api,
        map => $map,
        additional_args => {
            'connectrec-timeout' => 10,
            'reboot'             => 1, # should always reboot the system
        },
    );

    $Log->exit() if $may_exit;
}

sub _reboot_linux_cli {
    my ($pkg, @args) = @_;
    $Log->enter() if $may_enter;

    my $api = 'reboot';

    $pkg->_base_backend_client(
        @args,
        api => $api,
        additional_args => {
            'connectrec-timeout' => 10,
            'privilege-level'    => 'root',
        },
    );

    $Log->exit() if $may_exit;
}

sub _reboot_solaris_cli {
    my ($pkg, @args) = @_;
    $Log->enter() if $may_enter;

    my $api = 'reboot';

    $pkg->_base_backend_client(
        @args,
        api => $api,
        additional_args => {
            'connectrec-timeout' => 10,
            'privilege-level'    => 'root',
        },
    );

    $Log->exit() if $may_exit;
}

sub _get_up_or_start_time {
    my ($pkg_or_obj, @args) = @_;
    my %opts = $pkg_or_obj->_common_validate_with( params => \@args );
    if ($pkg_or_obj->is_windows) {
        return $pkg_or_obj->startup_time();
    } else {
        return $pkg_or_obj->uptime();
    }
}
1;
