#
# Copyright (c) 2012-13 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#
#
## @summary Fcp Component Module
## @author Vishwanath.Rawat@netapp.com, dl-nacl-dev@netapp.com
## @status shared
## @pod here
################################################################################

=head1 NAME

NACL::C::Client::Fcp

=head1 DESCRIPTION

C<NACL::C::Client::Fcp> Is a derived class of
L<NACL::C::Component::Client|lib-NACL-C-Component-pm>.
It represents access to FC on a Client Machine.

=head1 ATTRIBUTES

=head2 module

Name of the FC module installed on the client machine

=head2 card

Name of the FC card installed on the client machine

=head2 command_interface

See L<NACL::C::Component::command_interface|lib-NACL-C-Component-pm/command_interface>

=cut

###################
# Package
package NACL::C::Client::Fcp;

###################

use strict;
use warnings;
use Net::IP qw(ip_is_ipv6);

###################
# Module includes
use NACL::ComponentUtils qw (_optional_scalars Dumper);
use base 'NACL::C::Component::Client';
use Params::Validate qw(validate_with SCALAR BOOLEAN OBJECT);
use NACL::MTask::ClientUtils;
use NACL::CS::Client::Lsmod;

use NACL::APISet::Exceptions::ResponseException qw/:try/;
use NACL::Exceptions::NoElementsFound;
use NACL::C::Exceptions::DoesNotExist;

use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();

use Class::MethodMaker [
    scalar => [ qw(card module) ],
];

=head1 configure
    my $fcp = NACL::C::Client::Fcp->configure(
        command_interface => $ci,
    );

(Class method) Request to configure Fc on Client and return a new component object that refers to it.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< config_fc_driver_type => $string >>

(Optional) Fcp driver type used by client

Valid values are: native, vendor(default)

=back

=back

=cut

sub configure {
    $Log->enter();
    my ($pkg, @opts) = @_;
    $pkg->_frontend_log_debug_opts(@opts);
    my %opts = $pkg->_common_validate_with(
        params          => \@opts,
        additional_spec => {
            config_fc_driver_type => { type => SCALAR, default => "vendor" },
        },
    );

    my $ret = $pkg->call_on_apiset(
        %opts,
        choices => [
            {
                method    => "_configure_linux",
                interface => "CLI",
                set       => "Linux"
            },
        ],
    );

    $Log->exit();
    return $ret;
} ## end sub configure

sub _configure_linux {
    $Log->enter();
    my ($pkg, @opts)  = @_;
    my %opts = $pkg->_validate_backend_opts(@opts);

    my $ci     = $opts{command_interface};
    my $apiset = $opts{apiset};
    my $driver_type = delete $opts{config_fc_driver_type};

    my ($card, $module) = $pkg->_get_fc_card_and_module_type(
        apiset   => $apiset,
        config_fc_driver_type => $driver_type,
    );

    $pkg->_rescan_fc_linux(
        %opts,
        option => "configure_LUNs",
    );

    $pkg->load(module => $module, command_interface => $ci);

    my $obj = $pkg->new(
       command_interface => $ci,
       card              => $card,
       module            => $module,
    );

    $Log->exit();
    return $obj;
} ## end sub _configure_linux

=head1 rescan
    my $fcp = NACL::C::Client::Fcp->rescan(
        command_interface => $ci,
        option            => $option,
    );

    $fcp->rescan(
        option => $option,
    );

(Class or instance method) Will rescan scsi bus on Client to discover/remove LUNs.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< option => $string >>

(Optional) It specifies whether discover mapped LUNs or remove unmapped LUNs

Valid values are: remove_LUNs, configure_LUNs(default)

Applicable to Linux CLI.

=back

=back

=cut

sub rescan {
    $Log->enter();
    my ($pkg, @opts) = @_;
    $pkg->_frontend_log_debug_opts(@opts);
    my %opts = $pkg->_common_validate_with(
        params          => \@opts,
        additional_spec => {
            option   => { type => SCALAR },
        },
    );

    $pkg->call_on_apiset(
        %opts,
        choices => [
            {
                method    => "_rescan_fc_linux",
                interface => "CLI",
                set       => "Linux"
            },
        ],
    );

    $Log->exit();
} ## end sub rescan

sub _rescan_fc_linux {
    $Log->enter();
    my ($pkg, @opts) = @_;
    my %opts = $pkg->_validate_backend_opts(@opts);

    $pkg->_reboot_or_rescan_scsi_bus(%opts, protocol => "fcp");

    $Log->exit();
}

=head1 load
    my $fcp = NACL::C::Client::Fcp->load(
        command_interface => $ci,
        module            => $module,
    );

    $fcp->rescan(
        module => $module,
    );

(Class or instance method) Will load FC module on Client.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< module => $string >>

(Optional) Module to be loaded

Applicable to Linux CLI.

=back

=back

=cut

sub load {
    $Log->enter();
    my ($pkg, @opts) = @_;
    $pkg->_frontend_log_debug_opts(@opts);
    my %opts = $pkg->_common_validate_with(
        params          => \@opts,
        additional_spec => {
            module => { type => SCALAR },
        },
    );

    $pkg->call_on_apiset(
        %opts,
        choices => [
            {
                method    => "_load_module_linux",
                interface => "CLI",
                set       => "Linux"
            },
        ],
    );

    $Log->exit();
} ## end sub load

sub _load_module_linux {
    $Log->enter();
    my ($pkg, @opts) = @_;
    my %opts = $pkg->_validate_backend_opts(@opts);

    my $apiset = $opts{apiset};
    my $module = $opts{module};

    # Load fc module
    $pkg->_load_fc_module(%opts);

    # In case of lpfc we also need to load "lpfcdfc"
    if ($module =~ /lpfc/) {
        $pkg->_load_fc_module(module => "lpfcdfc");
    }

    $Log->exit();
}

=head1 unload
    my $fcp = NACL::C::Client::Fcp->unload(
        command_interface => $ci,
        module            => $module,
    );

    $fcp->unload(
        module => $module,
    );

(Class or instance method) Will unload FC module from Client.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< module => $string >>

(Optional) Module to be loaded

Applicable to Linux CLI.

=back

=back

=cut

sub unload {
    $Log->enter();
    my ($pkg, @opts) = @_;
    $pkg->_frontend_log_debug_opts(@opts);
    my %opts = $pkg->_common_validate_with(
        params          => \@opts,
        additional_spec => {
            module => { type => SCALAR },
        },
    );

    $pkg->call_on_apiset(
        %opts,
        choices => [
            {
                method    => "_unload_module_linux",
                interface => "CLI",
                set       => "Linux"
            },
        ],
    );

    $Log->exit();
} ## end sub unload

sub _unload_module_linux {
    $Log->enter();
    my ($pkg, @opts) = @_;
    my %opts = $pkg->_validate_backend_opts(@opts);

    my $apiset = $opts{apiset};
    my $module = delete $opts{module};

    if ($module =~ /qla2xxx/) {
        # Unload of qla2xxx will fail if "qlremote" is running
        # So killing qlremote before unloading qlogic driver
        try {
            my $resp = $apiset->pkill(pattern => "qlremote");
        } catch NACL::APISet::Exceptions::ResponseException with {
            # No need to fail here because even if
            # return status is NonZero, it will be killed
        };
    } elsif ($module =~ /lpfc/) {
        # hbaanyware needs to be stop in case of lpfc
        $apiset->execute_raw_command(
            command => "/usr/sbin/hbanyware/stop_hbanyware",
        );

        # unload this extra driver in case of emulex
        $pkg->_unload_fc_module(%opts, module => "lpfcdfc");
    }

    # unload fc module
    $pkg->_unload_fc_module(%opts, module => $module);

    $Log->exit();
}


=head1 reload
    my $fcp = NACL::C::Client::Fcp->reload(
        command_interface => $ci,
        module            => $module,
    );

    $fcp->reload(
        module => $module,
    );

(Class or instance method) Will reload FC module from Client. It is
basically wrapper around load and unload method

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< module => $string >>

(Optional) Module to be loaded

Applicable to Linux CLI.

=back

=back

=cut

sub reload {
    $Log->enter();
    my ($pkg, @opts) = @_;
    $pkg->_frontend_log_debug_opts(@opts);
    my %opts = $pkg->_common_validate_with(
        params          => \@opts,
        additional_spec => {
            module => { type => SCALAR },
        },
    );

    $pkg->call_on_apiset(
        %opts,
        choices => [
            {
                method    => "_reload_module_linux",
                interface => "CLI",
                set       => "Linux"
            },
        ],
    );

    $Log->exit();
} ## end sub load

sub _reload_module_linux {
    $Log->enter();
    my ($pkg, @opts) = @_;
    my %opts = $pkg->_validate_backend_opts(@opts);

    my $apiset = $opts{apiset};
    my $module = $opts{module};

    # Unoad fc module
    $pkg->_unload_module_linux(%opts);

    # Load fc module
    $pkg->_load_module_linux(%opts);

    $Log->exit();
}

sub _unload_fc_module {
    $Log->enter();
    my ($pkg, @opts)  = @_;
    my %opts = validate_with(
        params => \@opts,
        spec   => {
            module => { type => SCALAR },
            apiset => { type => OBJECT, isa => "NACL::APISet", },
            command_interface => { isa => 'NACL::C::CommandInterface', type => OBJECT },
        },
    );

    my $apiset = $opts{apiset};
    my $module = $opts{module};

    # Check if module is already loaded
    try {
        $pkg->_is_module_loaded(%opts);
        $apiset->modprobe(module => $module, remove => 1);
    } catch NACL::Exceptions::NoElementsFound with {
        # Module is not loaded. No need to unloade it.
    };

    $Log->exit();
}

sub _load_fc_module {
    $Log->enter();
    my ($pkg, @opts)  = @_;
    my %opts = validate_with(
        params => \@opts,
        spec   => {
            module => { type => SCALAR },
            apiset => { type => OBJECT, isa => "NACL::APISet", },
            command_interface => { isa => 'NACL::C::CommandInterface', type => OBJECT },
        },
    );

    my $apiset = $opts{apiset};
    my $module = $opts{module};

    # Check if module is already loaded
    try {
        $pkg->_is_module_loaded(%opts);
    } catch NACL::Exceptions::NoElementsFound with {
        my $resp = $apiset->modprobe(module => $module);
    };

    $Log->exit();
}

sub _is_module_loaded {
    $Log->enter();
    my ($pkg, @opts)  = @_;
    my %opts = validate_with(
        params => \@opts,
        spec   => {
            module            => { type => SCALAR },
            apiset            => { type => OBJECT, isa => "NACL::APISet", },
            command_interface => { isa => 'NACL::C::CommandInterface', type => OBJECT },
        },
    );

    my $module = $opts{module};
    my $ci     = $opts{command_interface};
    my $loaded_modules = NACL::CS::Client::Lsmod->fetch(
        command_interface => $ci,
        filter            => { module => $module },
    );

    $Log->exit();
}

sub _get_fc_card_and_module_type {
    $Log->enter();
    my ($pkg, @opts)  = @_;
    my %opts = validate_with(
        params => \@opts,
        spec   => {
            apiset                => { isa => "NACL::APISet", },
            config_fc_driver_type => { type => SCALAR, },
        },
    );

    my $apiset = $opts{apiset};
    my ($fc_card, $fc_module);
    my $driver_type = $opts{config_fc_driver_type};

    my @output_lspci = split(/\n/, $apiset->execute_raw_command(command => "lspci"));
    if (grep(/Cisco.*FCoE/i, @output_lspci)) {
        $fc_card   = "Cisco";
        $fc_module = "fnic";
    } elsif (grep(/Brocade/i, @output_lspci)) {
        $fc_card   = "Brocade";
        $fc_module = "bfa";
    } elsif (grep(/QLogic/i, @output_lspci)) {
        if ($driver_type =~ /vendor/) {
            my (@module_type) = (grep /QLogic/i,@output_lspci);
            if ($module_type[0] =~ /23\d\d/) {
                $fc_module = "qla2300";
            } elsif ($module_type[0] =~ /24\d\d/) {
                $fc_module = "qla2400";
            }
        } else {
            $fc_module = "qla2xxx";
        }
        $fc_card   = "QLogic";
    } elsif (grep(/Emulex|Fibre\s+Channel.*ServerEngines/i, @output_lspci)) {
        $fc_card   = "Emulex";
        $fc_module = "lpfc";
    } elsif (grep(/Intel\s+Corporation\s+82599/i, @output_lspci)) {
        $fc_card   = "Intel";
        $fc_module = "libfcoe";
    } elsif (grep(/Broadcom.*BCM57712/i, @output_lspci)) {
        $fc_card   = "Broadcom";
        $fc_module = "bnx2fc";
    } elsif (grep(/Chelsio Communications.*Unified Wire Storage Controller/i, @output_lspci)) {
        $fc_card   = "Chelsio";
        $fc_module = "csiostor";
    }

    if (! defined $fc_module) {
        NACL::C::Exceptions::DoesNotExist->throw(
            text => "No FC card is installed on this Client machine"
        );
    }

    return ($fc_card, $fc_module);
    $Log->exit();
}

sub _reboot_or_rescan_scsi_bus {
    $Log->enter();
    my ($pkg, @opts)  = @_;
    my %opts = validate_with(
        params => \@opts,
        spec   => {
            option            => { type => SCALAR, optional => 1, },
            protocol          => { type => SCALAR, },
            apiset            => { isa  => "NACL::APISet", },
            command_interface => { isa => 'NACL::C::CommandInterface', type => OBJECT },
        },
    );

    my $apiset = $opts{apiset};
    my $module = $opts{module};
    my $ci     = $opts{command_interface};

    my ($major, $minor) = $pkg->_get_major_minor(apiset => $apiset);
    if (($major =~ /sles9|rhel4/i) or
       (($major =~ /sles10/i) and $minor < 3) or
       (($major =~ /rhel5/i) and $minor < 4) or
       (!$pkg->_is_rpm_installed(rpm => "sg3_utils", apiset => $apiset))) {
        $Log->comment("System is rebooting to unload and load fcp driver");
        NACL::MTask::ClientUtils->reboot(command_interface => $ci);
    } else {
        $pkg->_rescan_scsi_bus(
            option   => $opts{option},
            apiset   => $apiset,
            protocol => $opts{protocol},
        );
    }

    $Log->exit();
}

sub _is_rpm_installed {
    $Log->enter();
    my ($pkg, @opts)  = @_;
    my %opts = validate_with(
        params => \@opts,
        spec   => {
            rpm    => { type => SCALAR, },
            apiset => { isa  => "NACL::APISet", },
        },
    );
    my $apiset = $opts{apiset};
    my $rpm    = $opts{rpm};
    my $install = 1;
    try {
        $apiset->rpm(query => 1, package => $rpm);
    } catch NACL::APISet::Exceptions::CommandFailedException with {
        $install = 0;
    };

    $Log->exit();
    return $install;
}

sub _primary_keys_validate_spec {
    return ();
}

1;
