#
# Copyright (c) 2014 NetApp, Inc., All Rights Reserved
# Any use, modification, or distribution is prohibited
# without prior written consent from NetApp, Inc.
#
## @summary File Operations Module
## @author dl-nacl-dev@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::C::Client::File

=head1 DESCRIPTION

This library is used to perform basic file system level operations
on a client such as open, close, write, read etc. and all these
operations would be performed on the local client (from which
the script is executed), so all of these mehtod does not require
command_interface as parameters.

=cut

package NACL::C::Client::File;
use strict;
use warnings;

use Params::Validate qw(validate SCALAR GLOBREF OBJECT);
use NACL::ComponentUtils qw(_optional_scalars Dumper);
use base 'NACL::C::Component::Client';
use NATE::BaseException qw(:try);
use NATE::Exceptions::Argument qw(:try);
use NACL::Exceptions::NoElementsFound;
use NACL::C::Exceptions::Client::File::FileException;
use threads 'exit'    => 'threads_only';
use constant KERNEL32 => 'C:\WINDOWS\system32\kernel32.dll';
use Tharn qw( host);
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use NATE::Process;
use NACL::Exceptions::Timeout;

my $Hosttype;
my $Type;

use Class::MethodMaker [
    scalar => 'file_name',
    scalar => 'file_handle',
    array  => 'lock_threads',
    scalar => 'process_obj',
    scalar => 'lock_pid',
];

my ($generic_read,    $generic_write,    $generic_execute,
    $generic_all,     $open_existing,    $open_always,
    $create_always,   $create_new,       $truncate_existing,
    $file_share_read, $file_share_write, $file_share_delete,
    $o_rdonly,        $o_wronly,         $o_rdwr,
    $o_creat,         $o_excl,           $o_append,
    $o_trunc,         $f_rdlck,          $f_wrlck,
    $f_unlck,         $f_setlk,          $f_setlkw,
    $seek_set,        $f_getlk,
);

BEGIN {

    my $linux_constant
        = '$o_rdonly = O_RDONLY; $o_wronly = O_WRONLY; $o_rdwr = O_RDWR; $o_creat = O_CREAT; $o_excl = O_EXCL; $o_append = O_APPEND; $o_trunc = O_TRUNC; $f_rdlck = F_RDLCK; $f_wrlck = F_WRLCK; $f_unlck = F_UNLCK; $f_setlk = F_SETLK; $f_setlkw = F_SETLKW, $seek_set = SEEK_SET, $f_getlk = F_GETLK; 1;';
    my $windows_constant
        = '$generic_read = GENERIC_READ; $generic_write = GENERIC_WRITE; $generic_execute = GENERIC_EXECUTE; $generic_all = GENERIC_ALL; $open_existing = OPEN_EXISTING; $open_always = OPEN_ALWAYS; $create_always = CREATE_ALWAYS; $create_new = CREATE_NEW; $truncate_existing = TRUNCATE_EXISTING; $file_share_read = FILE_SHARE_READ; $file_share_write = FILE_SHARE_WRITE; $file_share_delete = FILE_SHARE_DELETE; 1;';

    # constants to map to Windows' LockFileEx() values
    use constant LOCKFILE_FAIL_IMMEDIATELY => 1;
    use constant LOCKFILE_EXCLUSIVE_LOCK   => 2;

    # Get the local host
    my $host = host("localhost");
    $Hosttype = $host->hosttype();
    my $linux_packages = "use Fcntl; use File::FcntlLock; use File::Copy;1;";
    my $window_packages
        = "use Win32::API; use Win32API::File qw(:Func :Misc :FILE_ "
        . ":FILE_FLAG_ :FILE_SHARE_ :FILE_ATTRIBUTE_ :GENERIC_ );1;";

    # Load the libraries depending on the hosttype i.e "unix" or Windows"
    if ($host->is_unix() || $host->is_linux()) {
        $Type = 'linux';
        eval "$linux_packages";
        NATE::BaseException->throw(
            "Linux package could not be loaded due to error - $@")
            if ($@);
        eval "$linux_constant";
        NATE::BaseException->throw(
            "Linux constants could not be loaded due to error - $@")
            if ($@);

    } elsif ($host->is_windows()) {
        $Type = 'windows';
        eval "$window_packages";
        NATE::BaseException->throw(
            "Windows package could not be loaded due to error - $@")
            if ($@);
        eval "$windows_constant";
        NATE::BaseException->throw(
            "Windows constants could not be loaded due to error - $@")
            if ($@);
    } else {
        NATE::BaseException->throw(
            "Host $Hosttype is not supported client type'");
    }
} ## end BEGIN

sub init {
    $Log->enter() if $may_enter;
    my $self            = shift;
    my %validation_spec = $self->_primary_keys_validate_spec();
    validate(@_, \%validation_spec);
    $Log->exit() if $may_exit;
} ## end sub init

=head2 open()

Open (or create) a file(Class Method)

=over

=item Options

=item C<< file_name => $file_name >>

 (Required) The absolute or relative path of the file to be open

=item C<< access_mode => $access_mode >>

 (Required) The desired access mode for the file being opened
  Valid values are:  read, write, execute, read-write, all
The mapping for access modes of file for window and linux client is as
given below -
Parameter               Linux                   Windows
=========               ======                 =========
'read'                 O_RDONLY               GENERIC_READ
'write'                O_WRONLY               GENERIC_WRITE
'execute'                 ""                  GENERIC_EXECUTE
'all'                     ""                  GENERIC_ALL
'read-write'           O_RDWR                'GENERIC_READ|GENERIC_WRITE'

=item open_flags => $open_flags

 (Required) The desired open mode for the file being opened
 Valid values are:  open-existing, open-always, create-new,
 create-always, truncate-existing, open-append
The mapping for open flags of file for window and linux client is
as given below -
Parameter                     Linux                       Windows
=========                     ======                     =========
'open-existing'                ""                      OPEN_EXISTING
'open-always'                O_CREAT                   OPEN_ALWAYS
'create-always'              O_CREAT                   CREATE_ALWAYS
'create-new'             'O_CREAT|O_EXCL'              CREATE_NEW
'open-append'                O_APPEND                         ""
'truncate-existing'          O_TRUNC                   TRUNCATE_EXISTING

=item C<< permissions => $permissions >>

 (Optional) The permissions (4 bits) for the file being opened
 This is ignored for Windows.(Default - 0777)

=item C<< share_mode => $share_mode >>

 (Optional) The desired share mode for the file being opened
 Valid values are:  shared-read, shared-write, shared-read-write,
 shared-delete,shared-read-write-delete.
 This is ignored for Unix and defaults to shared-read-write for Windows
 parameters for share modes are only applicable for window client-
Parameter                                                Windows
=========                                               =========
'none'                                                        0
'shared-read'                                          FILE_SHARE_READ
'shared-write'                                         FILE_SHARE_WRITE
'shared-delete'                                       FILE_SHARE_DELETE
'shared-read-write'                   'FILE_SHARE_READ|FILE_SHARE_WRITE'
'shared-read-write-delete'  'SHARE_READ|FILE_SHARE_WRITE|FILE_SHARE_DELETE'

 Notes: For full details of supported flags and attributes,
        refer to MSDN's documentation for CreateFile() at:
        http://msdn.microsoft.com/en-us/library/aa363858%28v=vs.85%29.aspx

=item C<< attributes => $attributes >>

 (Optional) The desired flags and attributes for the file being opened
 This is ignored for Unix and can be any supported flags and attributes
 (or combinations combined with "|") specified in the MSDN documentation
 for CreateFile().  This is ignored for Unix.

=item Returns

Returns an object of type NACL::C::Client::File which has file_handle and
file_path stored as its attributes

=item Example

 my $fh_obj = NACL::C::Client::File->open(
     file_name => $file_name,
     access_mode => 'read',
     open_flags => 'create-new',
     share_mode  => 'shared-read',
     attributes => 'FILE_FLAG_DELETE_ON_CLOSE'
 );

=back

=cut

sub open {
    my $pkg = shift;
    $Log->enter() if $may_enter;

    my %opts = $pkg->_common_validate_with(
        params             => \@_,
        only_static_method => 1,
        additional_spec    => {
            access_mode => {type => SCALAR},
            open_flags  => {type => SCALAR},
            permissions => {
                type     => SCALAR,
                default  => '0777',
                optional => 1,
            },
            share_mode => {
                optional => 1,
                type     => SCALAR,
                default  => 'shared-read-write',
            },
            _optional_scalars(qw( attributes file_handle)),
        }
    );

    my $choices = {
        windows => "_open_windows",
        linux   => "_open_linux",
    };

    my $ret = $pkg->_choose_host(\%opts, $choices);

    $Log->exit() if $may_exit;
    return $ret;
} ## end sub open

sub _open_linux {
    my $pkg = shift;
    my ($file_handle, $file_name, $permissions, $flags, $orig_umask,);
    $Log->enter() if $may_enter;
    my %opts = @_;

    $file_name = $opts{file_name};
    my $access_mode = _translate_access_mode($opts{access_mode});

    my $open_flags = _translate_open_flags($opts{open_flags});

    # translate flags to integer equivalent and OR together
    $flags = eval "$access_mode";
    # append to open flags mask
    my $eval = eval $open_flags;
    if (defined $eval && $eval ne '') {
        if ($flags eq '') {
            $flags = $eval;
        } else {
            $flags |= $eval;
        }
    }
    $Log->comment("flag is $flags");

    # set the umask to '0' to allow user permissions
    $Log->comment("Set the umask to '0' to allow user permissions");
    $orig_umask = umask(0);

    # convert the permissions string to OCTAL (if required)
    if (defined $opts{permissions} and $opts{permissions} ne "") {
        if ($opts{permissions} =~ /^0/) {

            # convert perms to octal
            $permissions = oct $opts{permissions};
        } else {

            # assume perms are already octal
            $permissions = $opts{permissions};
        }
    } ## end if ( defined $opts{permissions...

    if ($Log->may_debug()) {
        $Log->debug("Options are file_handle => $file_handle, file_name => "
                . "$file_name, flags => $flags, permissions => $permissions");
    }

    # Open the file
    $Log->comment("Open the file $file_name with permission $permissions");
    sysopen($file_handle, $file_name, $flags, $permissions)
        or
        NATE::BaseException->throw("Failed to open file: $file_name: $!\n");

    # Reset the umask to original value
    $Log->comment("Reset the umask to original value");
    umask($orig_umask);

    $Log->exit() if $may_exit;

    return $pkg->new(file_handle => $file_handle, file_name => $file_name);
} ## end sub _open_linux

sub _open_windows {
    my $pkg  = shift;
    my %opts = @_;
    my ($file_handle, $file_name,  $permissions, $open_flags,
        $access_mode, $attributes, $share_mode
    );
    $Log->enter() if $may_enter;

    $file_name   = $opts{file_name};
    $access_mode = _translate_access_mode($opts{access_mode});
    $open_flags  = _translate_open_flags($opts{open_flags});
    $share_mode  = _translate_share_mode($opts{share_mode})
        if (defined $opts{share_mode});

    # translate access mode
    $access_mode |= eval "$access_mode";

    # translate open flags
    $open_flags |= eval "$open_flags";

    # translate shared mode flags
    $share_mode |= eval "$share_mode"
        if (defined $opts{share_mode});

    # translate flags & attributes
    $attributes |= eval "$opts{attributes}"
        if (defined $opts{attributes});

    if ($Log->may_debug()) {
        $Log->debug("Options are file_name => $file_name, access_mode => "
                . "$access_mode, share_mode => $share_mode, open_flags => "
                . "$open_flags, attributes => $attributes");
    }

    # Open the file
    $Log->comment("Open the file $file_name with access mode $access_mode");
    $file_handle
        = Win32API::File::CreateFile($file_name, $access_mode, $share_mode,
        [], $open_flags, $attributes, [])
        or
        NATE::BaseException->throw("Failed to open file: $file_name: $^E\n");
    $Log->exit() if $may_exit;

    return $pkg->new(file_handle => $file_handle, file_name => $file_name);

} ## end sub _open_windows

=head2 close()

Close a file handle (Class or Instance Method)

=over

=item Options

=item C<< file_handle => $fh >>

 (Required for Class Method) The file handle of the file to be closed
 (returned by open())

=item Returns

 None, Throws exception on failure

=item Example

 NACL::C::Client::File->close( file_handle => \*$Fh ); # $Fh returned by open()
or
 my $fh_obj = NACL::C::Client::File->open(
     file_name => $file_name,
     access_mode => 'read',
     open_flags => 'create-new'
 );
$fh_obj->close();

=back

=cut

sub close {
    my $pkg_or_obj = shift;
    $Log->enter() if $may_enter;
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@_,
        additional_spec => {
            file_handle => {type => GLOBREF | SCALAR},
            _optional_scalars(qw( file_name)),
        }
    );
    my $choices = {
        windows => "_close_windows",
        linux   => "_close_linux",
    };

    my $ret = $pkg_or_obj->_choose_host(\%opts, $choices);

    $Log->exit() if $may_exit;
} ## end sub close

sub _close_linux {
    my $pkg_or_obj = shift;
    my %opts       = @_;
    $Log->enter() if $may_enter;

    if ($Log->may_debug()) {
        $Log->debug("Options are file_handle => $opts{file_handle}");
    }

    $Log->comment("Close the file");
    CORE::close($opts{file_handle})
        or NATE::BaseException->throw("Failed to close file: $!\n");
    $Log->exit() if $may_exit;
} ## end sub _close_linux

sub _close_windows {
    my $pkg_or_obj = shift;
    my %opts       = @_;
    $Log->enter() if $may_enter;

    if ($Log->may_debug()) {
        $Log->debug("Options are file_handle => $opts{file_handle}");
    }

    $Log->comment("Close the file");
    Win32API::File::CloseHandle($opts{file_handle})
        or NATE::BaseException->throw("Failed to close file: $^E\n");
    $Log->exit() if $may_exit;
} ## end sub _close_windows

=head2 rename()

Rename a file to a new name (Instance or Class method)

=over

=item Options

=item C<< file_name => $file_name >>

 (Required for Class Method) The absolute or relative path of the file to be open

=item C<< new_file_name => $new_file >>

 (Required) The absolute or relative path of the new file

=item Returns

None, Throws exception on failure

=item Example

 NACL::C::Client::File->rename(
        file_name     => $file_name,
        new_file_name => $new_file_name
);
or
 my $fh_obj = NACL::C::Client::File->open(
     file_name => $file_name,
     access_mode => 'read',
     open_flags => 'create-new'
 );
 $fh_obj->rename(new_file_name => $new_file_name);

=back

=cut

sub rename {
    my $pkg_or_obj = shift;
    $Log->enter() if $may_enter;
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@_,
        additional_spec => {
            file_name     => {type => SCALAR},
            new_file_name => {type => SCALAR},
            file_handle   => {type => GLOBREF | SCALAR, optional => 1},
        },

    );

    my $choices = {
        windows => "_rename_windows",
        linux   => "_rename_linux",
    };
    my $ret = $pkg_or_obj->_choose_host(\%opts, $choices);

    $Log->exit() if $may_exit;

} ## end sub rename

sub _rename_linux {
    my $pkg_or_obj = shift;
    my %opts       = @_;
    my ($file_name, $new_file_name);
    $Log->enter() if $may_enter;

    $file_name     = $opts{file_name};
    $new_file_name = $opts{new_file_name};

    if ($Log->may_debug()) {
        $Log->debug(
            "Options are file_name => $file_name, new_file_name => $new_file_name"
        );
    }

    $Log->comment("Rename the file $file_name to $new_file_name");
    CORE::rename $file_name, $new_file_name
        or NATE::BaseException->throw(
        "failed to rename file $file_name to $new_file_name: $!\n");
    $pkg_or_obj->file_name($new_file_name) if ref($pkg_or_obj);
    $Log->exit() if $may_exit;
    return 0;
} ## end sub _rename_linux

sub _rename_windows {
    my $pkg_or_obj = shift;
    my %opts       = @_;
    my ($file_name, $new_file_name);
    $Log->enter() if $may_enter;

    $file_name     = $opts{file_name};
    $new_file_name = $opts{new_file_name};

    if ($Log->may_debug()) {
        $Log->debug(
            "Options are file_name => $file_name, new_file_name => $new_file_name"
        );
    }

    $Log->comment("Rename the file $file_name to $new_file_name");
    CORE::rename($file_name, $new_file_name)
        or NATE::BaseException->throw(
        "failed to rename file $file_name to $new_file_name: $^E\n");
    $pkg_or_obj->file_name($new_file_name) if ref($pkg_or_obj);
    $Log->exit() if $may_exit;
    return 0;
} ## end sub _rename_windows

=head2 move_file()

Move a file to a new location(Instance or Class method)

=over

=item Options

=item C << file_name >>

 (Required for Class Method) The absolute or relative path of the file to be open

=item C<< new_file_name >>

 (Required) The absolute or relative path of the new file

=item Returns

None, Throws exception on failure

=item Example

NACL::C::Client::File->move_file(
        file_name     => $File_name1,
        new_file_name => $File_name2
    );
or
 my $fh_obj = NACL::C::Client::File->open(
     file_name => $file_name,
     access_mode => 'read',
     open_flags => 'create-new'
 );
 $fh_obj->move_file(new_file_name => $File_name2);

=back

=cut

sub move_file {
    my $pkg_or_obj = shift;
    $Log->enter() if $may_enter;
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@_,
        additional_spec => {
            file_name     => {type => SCALAR},
            new_file_name => {type => SCALAR},
            file_handle   => {type => GLOBREF | SCALAR, optional => 1},
            }

    );

    my $choices = {
        windows => "_move_file_windows",
        linux   => "_move_file_linux",
    };
    my $ret = $pkg_or_obj->_choose_host(\%opts, $choices);

    $Log->exit() if $may_exit;

} ## end sub move_file

sub _move_file_linux {
    my $pkg_or_obj = shift;
    my %opts       = @_;
    my ($file_name, $new_file_name);
    $Log->enter() if $may_enter;

    $file_name     = $opts{file_name};
    $new_file_name = $opts{new_file_name};

    if ($Log->may_debug()) {
        $Log->debug(
            "Options are file_name => $file_name, new_file_name => $new_file_name"
        );
    }

    $Log->comment("Move the file $file_name to $new_file_name");
    CORE::rename($file_name, $new_file_name)
        or NATE::BaseException->throw(
        "Failed to move file $file_name to $new_file_name: $!\n");
    $pkg_or_obj->file_name($new_file_name) if ref($pkg_or_obj);
    $Log->exit() if $may_exit;
    return 0;
} ## end sub _move_file_linux

sub _move_file_windows {
    my $pkg_or_obj = shift;
    my %opts       = @_;
    my ($file_name, $new_file_name, $flags);
    $Log->enter() if $may_enter;

    $file_name     = $opts{file_name};
    $new_file_name = $opts{new_file_name};
    $flags |= (defined $opts{flags}) ? eval "$opts{flags}" : 0;

    if ($Log->may_debug()) {
        $Log->debug(
            "Options are file_name => $file_name, new_file_name => $new_file_name"
        );
    }

    $Log->comment("Move the file $file_name to $new_file_name");
    Win32API::File::MoveFileEx($file_name, $new_file_name, $flags)
        or NATE::BaseException->throw(
        "Failed to move file $file_name to $new_file_name: $^E\n");
    $pkg_or_obj->file_name($new_file_name) if ref($pkg_or_obj);
    $Log->exit() if $may_exit;
    return 0;
} ## end sub _move_file_windows

=head2 delete()

Delete a file(Instance or Class Method)

=over

=item Options

=item C<< file_name >>

(Required for Class Method) The absolute or relative path of the file to be open

=item Returns

None, Throws exception on failure

=item Example

 NACL::C::Client::File->delete( file_name => $File_name );
or
 my $fh_obj = NACL::C::Client::File->open(
     file_name => $file_name,
     access_mode => 'read',
     open_flags => 'create-new'
 );
 $fh_obj->delete();

=back

=cut

sub delete {
    my $pkg_or_obj = shift;
    $Log->enter() if $may_enter;
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@_,
        additional_spec => {
            file_name   => {type => SCALAR},
            file_handle => {type => GLOBREF | SCALAR, optional => 1},
        },
    );

    my $choices = {
        windows => "_delete_windows",
        linux   => "_delete_linux",
    };
    my $ret = $pkg_or_obj->_choose_host(\%opts, $choices);

    $Log->exit() if $may_exit;
} ## end sub delete

sub _delete_linux {
    my $pkg_or_obj = shift;
    my %opts       = @_;
    my $file_name;

    $Log->enter() if $may_enter;
    $file_name = $opts{file_name};

    if ($Log->may_debug()) {
        $Log->debug("Options are file_name => $file_name");
    }

    $Log->comment("Delete the file $file_name");
    my $success = unlink $file_name;
    if (!$success) {
        my $e = $!;
        if ($e =~ /No such file or directory/) {
            NACL::Exceptions::NoElementsFound->throw(
                "failed to delete file $file_name: $!\n");
        } else {
            NATE::BaseException->throw(
                "failed to delete file $file_name: $!\n");
        }
    }
    $Log->exit() if $may_exit;
    return 0;
} ## end sub _delete_linux

sub _delete_windows {
    my $pkg_or_obj = shift;
    my %opts       = @_;
    my $file_name;

    $Log->enter() if $may_enter;
    $file_name = $opts{file_name};

    if ($Log->may_debug()) {
        $Log->debug("Options are file_name => $file_name");
    }

    $Log->comment("Delete the file $file_name");
    my $success = Win32API::File::DeleteFile($file_name);
    if (!$success) {
        my $e = $^E;
        if ($e =~ /The system cannot find the file specified/i) {
            NACL::Exceptions::NoElementsFound->throw(
                "failed to delete file $file_name: $^E\n");
        } else {
            NATE::BaseException->throw(
                "failed to delete file $file_name: $^E\n");
        }
    }

    $Log->exit() if $may_exit;
    return 0;
} ## end sub _delete_windows

=head2 read()

Read from a file handle(Instance or class method)

=over

=item Options

=item C<< file_handle >>

 (Required for Class Method) The file handle to read from
  Type - Globref for Linux client and Scalar for Windows client

=item C<< offset >>

 (Optional) The offset into the file to begin reading from.
 If this option is not specified, then read would happen
 from current location of file pointer.

=item C<< num_bytes >>

 (Required) The number of bytes to read from the file

=item Returns

 SCALAR data if successfully read, or throws an exception

=item Example

$read_output = NACL::C::Client::File->read(
        file_handle => $Fh,
        offset      => 0,
        num_bytes   => 80
    );
or
$fh_obj->read(offset => 0,num_bytes   => 80);

=back

=cut

sub read {
    my $pkg_or_obj = shift;
    $Log->enter() if $may_enter;
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@_,
        additional_spec => {
            file_handle => {type => GLOBREF | SCALAR},
            offset      => {type => SCALAR, optional => 1},
            num_bytes   => {type => SCALAR},
            _optional_scalars(qw(file_name)),
        }
    );

    my $choices = {
        windows => "_read_windows",
        linux   => "_read_linux",
    };
    my $ret = $pkg_or_obj->_choose_host(\%opts, $choices);

    $Log->exit() if $may_exit;
    return $ret;
} ## end sub read

sub _read_linux {
    my $pkg_or_obj = shift;
    my %opts       = @_;
    my ($file_handle, $num_bytes, $buffer, $num_bytes_read);
    $Log->enter() if $may_enter;

    $file_handle = $opts{file_handle};
    $num_bytes   = $opts{num_bytes};

    if ($Log->may_debug()) {
        $Log->debug(
            "Options are file_handle => $file_handle, offset => $opts{offset}"
        );
    }

    # seek to the offset
    $Log->comment("Seek the offset to read the file");
    if (defined $opts{offset}) {
        sysseek($file_handle, $opts{offset}, $seek_set)
            or NATE::BaseException->throw(
            "failed to seek to offset $opts{offset}: $!\n");
    }

    if ($Log->may_debug()) {
        $Log->debug("Options are file_handle => $file_handle, buffer => "
                . "$buffer, num_bytes => $num_bytes");
    }

    # read
    $Log->comment("Read the file");
    $num_bytes_read = sysread($file_handle, $buffer, $num_bytes);
    if (!defined $num_bytes_read) {

        # sysread returns undef on failure
        NATE::BaseException->throw(
            "failed to read $num_bytes bytes from file: $!\n");
    }

    # print the actual number of bytes read
    $Log->comment("read: $num_bytes_read bytes read from file");
    $Log->debug("read: Data:\n$buffer\n");
    $Log->exit() if $may_exit;
    return $buffer;
} ## end sub _read_linux

sub _read_windows {
    my $pkg_or_obj = shift;
    my %opts       = @_;
    my ($file_handle, $num_bytes, $buffer, $num_bytes_read);
    $Log->enter() if $may_enter;

    $file_handle = $opts{file_handle};
    $num_bytes   = $opts{num_bytes};

    # set the current pointer to offset
    $Log->comment("Set the current pointer to offset");
    if (defined $opts{offset}) {
        Win32API::File::setFilePointer($file_handle, $opts{offset},
            "FILE_BEGIN")
            or NATE::BaseException->throw(
            "failed to set pointer to $opts{offset} before reading: $^E\n");
    } ## end if ( defined $opts{offset...

    if ($Log->may_debug()) {
        $Log->debug("Options are file_handle => $file_handle, buffer => "
                . "$buffer, num_bytes => $num_bytes, num_bytes_read => $num_bytes_read"
        );
    }

    # read
    $Log->comment("Read the file");
    Win32API::File::ReadFile($file_handle, $buffer, $num_bytes,
        $num_bytes_read, [])
        or NATE::BaseException->throw(
        "failed to read $num_bytes bytes from file: $^E\n");

    # print the actual number of bytes read
    $Log->comment("read: $num_bytes_read bytes read from file");
    $Log->debug("read: Data:\n$buffer\n");
    $Log->exit() if $may_exit;
    return $buffer;
} ## end sub _read_windows

=head2 write()

Write to a file handle(Instance or class method)

=over

=item Options

=item C<< file_handle >>

 (Required for Class Method) The file handle to write to
 Type - Globref for Linux client and Scalar for Windows client

=item C<< offset >>

 (Optional) The offset into the file to begin writing to.
  If this option is not specified, then write would happen
 from current location of file pointer.

=item C<< content >>

 (Required) The content to write to the file

=item Returns

 SCALAR number of bytes that were successfully written,
 or throws an exception

=item Example

 my $num_bytes_wrote = NACL::C::Client::File->write(
    file_handle => $file_handle,
    offset      => 10,
    content    => $content
    );
or
$fh_obj->write(offset => 10,content => $content);

=back

=cut

sub write {
    my $pkg_or_obj = shift;
    $Log->enter() if $may_enter;
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@_,
        additional_spec => {
            file_handle => {type => GLOBREF | SCALAR},
            offset      => {type => SCALAR, optional => 1, default => 0},
            content     => {type => SCALAR},
            _optional_scalars(qw(file_name)),
        },
    );

    my $choices = {
        windows => "_write_windows",
        linux   => "_write_linux",
    };
    my $ret = $pkg_or_obj->_choose_host(\%opts, $choices);

    $Log->exit() if $may_exit;
    return $ret;
} ## end sub write

sub _write_linux {
    my $pkg_or_obj = shift;
    my %opts       = @_;
    my ($file_handle, $content, $num_bytes_wrote);
    $Log->enter() if $may_enter;

    $file_handle = $opts{file_handle};
    $content     = $opts{content};

    # seek to the offset
    $Log->comment("Set the current pointer to offset");
    if (defined $opts{offset}) {
        sysseek($file_handle, $opts{offset}, $seek_set)
            or NATE::BaseException->throw(
            "failed to seek to offset $opts{offset}: $!\n");
    }

    if ($Log->may_debug()) {
        $Log->debug(
            "Options are file_handle => $file_handle, content => $content");
    }

    # write to the file
    $Log->comment("Write to the file");
    $num_bytes_wrote = syswrite($file_handle, $content)
        or NATE::BaseException->throw("failed to write to file: $!\n");
    $Log->exit() if $may_exit;
    return $num_bytes_wrote;
} ## end sub _write_linux

sub _write_windows {
    my $pkg_or_obj = shift;
    my %opts       = @_;
    my ($file_handle, $content, $num_bytes_wrote);
    $Log->enter() if $may_enter;

    $file_handle = $opts{file_handle};
    $content     = $opts{content};

    # set the current pointer to offset
    $Log->comment("Set the current pointer to offset");
    if (defined $opts{offset}) {
        Win32API::File::setFilePointer($file_handle, $opts{offset},
            "FILE_BEGIN")
            or NATE::BaseException->throw(
            "failed to set pointer to $opts{offset} before reading: $^E\n");
    } ## end if ( defined $opts{offset...

    if ($Log->may_debug()) {
        $Log->debug("Options are file_handle => $file_handle, content => "
                . "$content, num_bytes_wrote => $num_bytes_wrote");
    }

    # write
    $Log->comment("Write to the file");
    Win32API::File::WriteFile($file_handle, $content, 0, $num_bytes_wrote, [])
        or NATE::BaseException->throw("failed to write to file: $^E\n");
    $Log->exit() if $may_exit;
    return $num_bytes_wrote;
} ## end sub _write_windows

=head2 lock()

Lock a byte range in a file(Instance method)

=over

=item Options

=item C<< offset >>

 (Optional) The offset into the file to begin locking from

=item C<< num_bytes >>

 (Required) The number of bytes to lock

=item C<< lock_type >>

 (Required) The type of lock to acquire.The desired lock type values
 for the file being locked are:  read-lock, write-lock
The mapping for lock type of file for window and linux client is
as given below -
Parameter                    Linux                      Windows
=========                    ======                     =========
'read-lock'                  F_RDLCK                       ""
'write-lock'                 F_WRLCK             LOCKFILE_EXCLUSIVE_LOCK

=item C<< lock_wait >>

 (Default - non-waiting-lock )
 (Optional) The waiting symantics for the lock.The desired lock wait values
 for the file being locked are:  non-waiting-lock, waiting-lock
The mapping for lock wait of file for window and linux client is
as given below -
Parameter                    Linux                     Windows
=========                    ======                    =========
'non-waiting-lock'          F_SETLK           LOCKFILE_FAIL_IMMEDIATELY
'waiting-lock'              F_SETLKW                      ""

=item Returns

     SCALAR or throws an exception
     For non-waiting locks, returns lock id of lock.
     For waiting locks, a array of thread object is returned if user called its
     in array context or it returns last thread object

=item Example

 $lock_object = $file_obj->lock(
                   offset       => 10,
                   num_bytes    => 50,
                   lock_type    => 'write-lock',
                   lock_wait    => 'waiting-lock'
                  );

=back

=cut

sub lock {
    my $self = shift;
    $Log->enter() if $may_enter;
    $self->_verify_invocation(style => 'instance_only');
    my %opts = $self->_common_validate_with(
        params          => \@_,
        additional_spec => {
            file_handle => {type => GLOBREF | SCALAR},
            offset      => {type => SCALAR, optional => 1},
            num_bytes   => {type => SCALAR},
            lock_type   => {type => SCALAR},
            lock_wait => {
                type     => SCALAR,
                optional => 1,
                default  => 'non-waiting-lock'
            },
            _optional_scalars(qw(file_name)),
        }
    );

    my $choices = {
        windows => "_lock_windows",
        linux   => "_lock_linux",
    };
    my $ret = $self->_choose_host(\%opts, $choices);

    $Log->exit() if $may_exit;
    return $ret;
} ## end sub lock

sub _lock_linux {
    my $self = shift;
    my %opts = @_;
    my ($file_handle, $offset, $num_bytes, $lock_type, $lock_wait, $locker,
        $lock_op);
    $Log->enter() if $may_enter;

    $file_handle = $opts{file_handle};
    $offset      = $opts{offset};
    $num_bytes   = $opts{num_bytes};

    $lock_type = _translate_lock_type_params($opts{lock_type});
    $lock_wait = _translate_lock_wait_params($opts{lock_wait});
    $Log->comment("lock_type : $lock_type, lock_wait : $lock_wait");

    # translate flags to integer equivalent
    $lock_op = ($lock_type eq "F_UNLCK") ? "unlock" : "lock";
    $lock_type |= eval "$lock_type";
    $Log->comment("Flag lock_type : $lock_type, lock_wait : $lock_wait");

    if ($Log->may_debug()) {
        $Log->debug(
            "Options are file_handle => $file_handle, lock_wait => $lock_wait"
        );
    }

    # lock - check if this is a waiting lock request
    $Log->comment("lock - check if this is a waiting lock request");
    if ($lock_wait eq $f_setlkw) {

        # set up the lock structure
        $locker = File::FcntlLock->new();
        $locker->l_type($lock_type);    # type = F_WRLCK or F_RDLCK
        $locker->l_whence($seek_set);
        $locker->l_start($offset);
        $locker->l_len($num_bytes);
        $locker->l_pid($$);
        $lock_wait |= eval "$lock_wait";

        # spawn a thread for this waiting lock request
        $Log->comment("spawn a thread for this waiting lock request");
        my $thread_id
            = threads->new(\&_lock_handle_unix, $locker, $file_handle,
            $lock_wait);

        my $thread_obj
            = NACL::C::Client::Thread->new(thread_id => $thread_id);

        my @all_threads = $self->lock_threads();
        push @all_threads, $thread_obj;
        $self->lock_threads(@all_threads);
        return wantarray ? @all_threads : $thread_obj;

    } ## end if ( $lock_wait eq $f_setlkw)

    # else, this is a non-waiting lock request
    $Log->comment("This is non-waiting lock");
    my $get_lock =  '{
        my ($lock_type, $seek_set, $num_bytes, $lock_wait, $file_handle, $f_setlk, $lock_op, $offset) = @_;
        my $locker;
        $locker = File::FcntlLock->new();
        $locker->l_type($lock_type);    # type = F_WRLCK or F_RDLCK
        $locker->l_whence($seek_set);
        $locker->l_start($offset);
        $locker->l_len($num_bytes);
        $locker->l_pid($$);
        $lock_wait |= eval "$lock_wait";
        $locker->lock($file_handle, $lock_wait, $f_setlk);
        # check for lock error
        if ($locker->error()) {
            $Log->exit();
            NATE::BaseException->throw(
                  "failed to $lock_op $num_bytes bytes in the file: \n" . "("
                . $locker->lock_errno() . ") "
                . $locker->error()
                . "\n");
        } ## end if ( $locker->error() )
        while(1) { };
    }';
    $Log->comment("Lock the file");
    my $proc;
    $proc = NATE::Process->new(
                codespec => $get_lock,
                args   => [$lock_type, $seek_set, $num_bytes, $lock_wait, $file_handle, $f_setlk, $lock_op, $offset],
                runid    => "get-lock",
                onexit => \&NATE::Process::on_exit_propagate_worst_result);
    $proc->start();

    $Log->exit() if $may_exit;
    $self->process_obj($proc);
    $self->lock_pid($proc->pid());

} ## end sub _lock_linux

sub _lock_windows {
    my $self = shift;
    my %opts = @_;
    my ($file_handle, $offset, $num_bytes, $lock_type, $lock_wait, $locker,
        $lock_op,);
    $Log->enter() if $may_enter;
    $lock_wait   = $opts{lock_wait};
    $file_handle = $opts{file_handle};
    $offset      = $opts{offset};
    $num_bytes   = $opts{num_bytes};

    $lock_type = _translate_lock_type_params($opts{lock_type});

    # translate flags to integer equivalent
    $lock_type |= eval "$lock_type"
        if (defined $lock_type && $lock_type ne "");

    $Log->comment("lock_type is $lock_type");
    $Log->comment("lock_wait is $lock_wait");

    my $lock_object = new Win32::API(KERNEL32, 'LockFileEx', 'NNNNNP', 'N')
        or NATE::BaseException->throw(
        "failed to import a prototype object for LockFileEx: " . $^E);

    if ($Log->may_debug()) {
        $Log->debug(
            "Options are file_handle => $file_handle, lock_type => $lock_type, "
                . "num_bytes => $num_bytes, offset => $offset");
    }

    # lock - check if this is a waiting lock request
    $Log->comment("lock - check if this is a waiting lock request");
    if ($lock_wait eq "waiting-lock") {
        $lock_type = LOCKFILE_EXCLUSIVE_LOCK;

        # spawn a thread for this waiting lock request
        $Log->comment("spawn a thread for this waiting lock request");
        my $thread_id
            = threads->new(\&_lock_handle_windows, $file_handle, $lock_type,
            $num_bytes, $offset);

        my $thread_obj
            = NACL::C::Client::Thread->new(thread_id => $thread_id);

        my @all_threads = $self->lock_threads();
        push @all_threads, $thread_obj;
        $self->lock_threads(@all_threads);
        return wantarray ? @all_threads : $thread_obj;
    } ## end if ( $lock_wait eq "waiting-lock")

    # else, this is a non-waiting lock request
    $Log->comment("Lock the file");
    eval {
        my $status = $lock_object->Call(
            $file_handle,    # from open()
            $lock_type,
            0,               # RESERVED
            $num_bytes,      # offset (LOW)
            0,               # offset (HIGH)
            pack('PPLLL', 0, 0, $offset, 0, 0)    # LPOVERLAPPED
        );
        $Log->comment("lock command status = $status");
    };

    if ($@) {
        chomp $@;
        $Log->exit() if $may_exit;
        NATE::BaseException->throw(
            "" . (caller(0))[3] . ": " . $^E . " (" . $@ . ")");
    }
    $Log->exit() if $may_exit;

} ## end sub _lock_windows

=head2 unlock()

(Instance Method)Unlock a byte range lock on a file (uses F_UNLCK)

=over

=item Options

=item offset

 (Required*) The offset into the file to begin unlocking from
             * Not required if lock_id is provided

=item num_bytes

 (Required*) The number of bytes to unlock
             * Not required if lock_id is provided

=item unlock_wait

 (Optional) The waiting semantics for the lock. The desired unlock wait
 values for the file being unlocked are:  non-waiting-lock, waiting-lock
This is required for Linux client only.
Parameter                    Linux                    Windows
=========                    ======                 ===========
'non-waiting-lock'          F_SETLK           LOCKFILE_FAIL_IMMEDIATELY
'waiting-lock'              F_SETLKW                    ""

=item Returns

0 for SUCCESS else throw an exception.

=item Examples

 $file_obj->unlock(
        offset      => 10,
        num_bytes   => 50);

 $file_obj->unlock(unlock_wait => 'waiting-lock');

=back

=cut

sub unlock {
    my ($self, @args) = @_;
    $Log->enter() if $may_enter;
    $self->_verify_invocation(style => 'instance_only');
    my %opts = $self->_common_validate_with(
        params          => \@args,
        additional_spec => {
            file_handle => {type => GLOBREF | SCALAR},
            offset      => {type => SCALAR, optional => 1, default => 0},
            num_bytes   => {type => SCALAR, optional => 1},
            process_obj => {type => OBJECT, optional => 1},
            unlock_wait => {type => SCALAR, default  => 'non-waiting-lock'},
            _optional_scalars(qw(file_name lock_pid)),
        }
    );
    my $choices = {
        windows => "_unlock_windows",
        linux   => "_unlock_linux",
    };
    my $ret = $self->_choose_host(\%opts, $choices);;

    $Log->exit() if $may_exit;
    return $ret;
} ## end sub unlock

sub _unlock_linux {
    my $self = shift;
    my %opts = @_;
    $Log->enter() if $may_enter;
    if(defined $opts{process_obj} ) {
        my $process = delete $opts{process_obj};
        $process->stop() if $process->is_running();
    }

    if (defined $opts{unlock_wait}) {
        $opts{lock_wait} = $opts{unlock_wait};
        delete $opts{unlock_wait};
    }

    if ($Log->may_debug()) {
        $Log->debug("Options are offset => $opts{offset}, lock_wait => "
                . "$opts{lock_wait}, num_bytes => $opts{num_bytes}");
    }

    # since all code (except F_UNLCK) is common to
    # lock() and unlock(), call lock() with F_UNLCK
    $Log->comment("Unlock the file");
    my $return_val = $self->lock(
        lock_type => 'file-unlock',
        lock_wait => $opts{lock_wait},
        offset    => $opts{offset},
        num_bytes => $opts{num_bytes},
    );

    $Log->exit() if $may_exit;
    return $return_val;
} ## end sub _unlock_linux

sub _unlock_windows {
    my $self = shift;
    my %opts = @_;
    my ($file_handle, $offset, $num_bytes, $lock_type, $lock_wait, $locker,
        $lock_op);
    $Log->enter() if $may_enter;

    $file_handle = $opts{file_handle};
    $offset      = $opts{offset};
    $num_bytes   = $opts{num_bytes};

    my $unlock_object = new Win32::API(KERNEL32, 'UnlockFileEx', 'NNNNP', 'N')
        or NATE::BaseException->throw(
        "failed to import a prototype object for UnlockFileEx: " . $^E);

    if ($Log->may_debug()) {
        $Log->debug(
            "Options are file_handle => $file_handle, num_bytes => $num_bytes"
        );
    }

    # unlock
    $Log->comment("Unlock the file");
    eval {
        $unlock_object->Call(
            $file_handle,    # from open()
            0,               # RESERVED
            $num_bytes,      # offset (LOW)
            0,               # offset (HIGH)
            pack('PPLLL', 0, 0, $offset, 0, 0)    # LPOVERLAPPED
        );
    };
    if ($@) {
        chomp $@;
        $Log->exit() if $may_exit;
        NATE::BaseException->throw(
            "" . (caller(0))[3] . ": " . $^E . " (" . $@ . ")");
    }
    $Log->exit() if $may_exit;

} ## end sub _unlock_windows

# Private Functions

sub _translate_access_mode() {

    my ($access_mode) = @_;
    my %access_mode_values;
    no strict "vars";
    if ($Hosttype =~ /win/i) {
        %access_mode_values = (
            read         => $generic_read,
            write        => $generic_write,
            'execute'    => $generic_execute,
            'all'        => $generic_all,
            'read-write' => $generic_read | $generic_write,
        );
    } elsif ($Hosttype =~ /linux/i) {
        %access_mode_values = (
            read         => $o_rdonly,
            write        => $o_wronly,
            'read-write' => $o_rdwr,
            'execute'    => '',
            'all'        => ''
        );
    } else {
        $Log->exit() if $may_exit;
        NATE::BaseException->throw("Unsupported client type $Hosttype");
    }
    use strict "vars";

    if (exists $access_mode_values{$access_mode}) {
        return $access_mode_values{$access_mode};
    } else {
        $Log->exit() if $may_exit;
        NATE::Exceptions::Argument->throw(
            "Value of access_mode is not valid : $access_mode");
    }

} ## end sub _translate_access_mode()

sub _translate_open_flags() {

    my ($open_flags) = @_;
    my %open_flags_values;

    if ($Hosttype =~ /win/i) {
        %open_flags_values = (
            'open-existing'     => $open_existing,
            'open-always'       => $open_always,
            'create-always'     => $create_always,
            'create-new'        => $create_new,
            'open-append'       => '',
            'truncate-existing' => $truncate_existing
        );
    } elsif ($Hosttype =~ /linux/i) {
        %open_flags_values = (
            'open-existing'     => '',
            'open-always'       => $o_creat,
            'create-always'     => $o_creat,
            'create-new'        => $o_creat | $o_excl,
            'open-append'       => $o_append,
            'truncate-existing' => $o_trunc
        );
    } else {
        $Log->exit() if $may_exit;
        NATE::BaseException->throw("Unsupported client type $Hosttype");
    }

    if (exists $open_flags_values{$open_flags}) {
        return $open_flags_values{$open_flags};
    } else {
        $Log->exit() if $may_exit;
        NATE::Exceptions::Argument->throw("Value of open_flags is not valid");
    }
} ## end sub _translate_open_flags()

sub _translate_share_mode() {

    my ($share_mode) = @_;
    my %share_mode_values;

    if ($Hosttype =~ /win/i) {
        %share_mode_values = (
            'none'              => 0,
            'shared-read'       => $file_share_read,
            'shared-write'      => $file_share_write,
            'shared-delete'     => $file_share_delete,
            'shared-read-write' => $file_share_read | $file_share_write,
            'shared-read-write-delete' => $file_share_read | $file_share_write
                | $file_share_delete
        );
    } elsif ($Hosttype =~ /linux/i) {
        %share_mode_values = ();
    } else {
        $Log->exit() if $may_exit;
        NATE::BaseException->throw("Unsupported client type $Hosttype");
    }

    if (exists $share_mode_values{$share_mode}) {
        return $share_mode_values{$share_mode};
    } else {
        $Log->exit() if $may_exit;
        NATE::Exceptions::Argument->throw("Value of share_mode is not valid");
    }
} ## end sub _translate_share_mode()

sub _translate_lock_type_params() {
    my ($lock_type) = @_;
    my %lock_type_values;

    if ($Hosttype =~ /win/i) {
        %lock_type_values = (
            'read-lock'  => '',
            'write-lock' => LOCKFILE_EXCLUSIVE_LOCK
        );
    } elsif ($Hosttype =~ /linux/i) {
        %lock_type_values = (
            'read-lock'   => $f_rdlck,
            'write-lock'  => $f_wrlck,
            'file-unlock' => $f_unlck,
        );
    } else {
        $Log->exit() if $may_exit;
        NATE::BaseException->throw("Unsupported client type $Hosttype");
    }

    if (exists $lock_type_values{$lock_type}) {
        return $lock_type_values{$lock_type};
    } else {
        $Log->exit() if $may_exit;
        NATE::Exceptions::Argument->throw("Value of lock_type is not valid");
    }
} ## end sub _translate_lock_type_params()

sub _translate_lock_wait_params() {
    my ($lock_wait) = @_;
    my %lock_wait_values;

    if ($Hosttype =~ /win/i) {
        %lock_wait_values = (
            'non-waiting-lock' => LOCKFILE_FAIL_IMMEDIATELY,
            'waiting-lock'     => ''
        );
    } elsif ($Hosttype =~ /linux/i) {
        %lock_wait_values = (
            'non-waiting-lock' => $f_setlk,
            'waiting-lock'     => $f_setlkw,
        );
    } else {
        $Log->exit() if $may_exit;
        NATE::BaseException->throw("Unsupported client type $Hosttype");
    }

    if (exists $lock_wait_values{$lock_wait}) {
        return $lock_wait_values{$lock_wait};
    } else {
        $Log->exit() if $may_exit;
        NATE::Exceptions::Argument->throw("Value of lock_wait is not valid");
    }
} ## end sub _translate_lock_wait_params()

sub _choose_host {
    my $pkg_or_obj = shift;
    my $opts       = shift;
    my $method     = shift;

    my $call = $method->{$Type};
    my $ret;

    my $timeout = $opts->{'method-timeout'} || 60;

    # Adding timeout loop.
    my $new_exception_obj;

    local $SIG{ALRM} = sub {
        NACL::Exceptions::Timeout->throw
            ("Timed out ($timeout sec) in $method->{$Type} method\n");
    };

    alarm $timeout;
    try {
        $ret = $pkg_or_obj->$call(%$opts);
    } catch NACL::Exceptions::Timeout with {
        $_[0]->throw;
    } catch NATE::BaseException with {
        my $exception_obj = shift;
        $new_exception_obj =
            NACL::C::Exceptions::Client::File::FileException->convert
            (exception => $exception_obj);
        ## Deleting file_handle since this is GLOB item, While nfreeze it throws error. Please see burt:985904
        delete $opts->{file_handle};
        $new_exception_obj->file_args(%$opts);
        $new_exception_obj->throw();
    } finally {
        alarm 0;
    };

    return $ret;

} ## end sub _choose_host

#--------------------------------------------------------------------
# this method is for creating a thread for waiting lock requests.
# this method should NOT be called directly by any external caller.
#--------------------------------------------------------------------
sub _lock_handle_unix {

    # Thread 'cancellation' signal handler
    $SIG{'KILL'} = sub {
        return;
    };

    my ($locker, $file_handle, $lock_wait) = @_;

    $locker->lock($file_handle, $lock_wait); # lock_type = F_SETLK or F_SETLKW
                                             # check for lock error
    if ($locker->error()) {
        $Log->exit() if $may_exit;
        NATE::BaseException->throw("failed to lock the file: \n" . "("
                . $locker->lock_errno() . ") "
                . $locker->error()
                . "\n");
    } ## end if ( $locker->error() )

    return;
} ## end sub _lock_handle_unix

sub _lock_handle_windows {

    # Thread 'cancellation' signal handler
    $SIG{'KILL'} = sub {
        return;
    };

    my ($file_handle, $lock_type, $num_bytes, $offset) = @_;

    my $lock_object = new Win32::API(KERNEL32, 'LockFileEx', 'NNNNNP', 'N')
        or NATE::BaseException->throw(
        "failed to import a prototype object for LockFileEx: " . $^E);

    eval {
        my $status = $lock_object->Call(
            $file_handle, # from open()
            $lock_type,   # LOCKFILE_EXCLUSIVE_LOCK, LOCKFILE_FAIL_IMMEDIATELY
            0,            # RESERVED
            $num_bytes,   # offset (LOW)
            0,            # offset (HIGH)
            pack('PPLLL', 0, 0, $offset, 0, 0)    # LPOVERLAPPED
        );
        $Log->comment("Waiting lock command status = $status");
    };
    if ($@) {
        chomp $@;
        $Log->exit() if $may_exit;
        NATE::BaseException->throw(
            "" . (caller(0))[3] . ": " . $^E . " (" . $@ . ")");
    }

    return;
} ## end sub _lock_handle_windows

sub _primary_keys_validate_spec {
    return (
        file_name   => {type => SCALAR},
        file_handle => {type => GLOBREF | SCALAR},
        lock_pid    => {type => SCALAR, optional => 1},
        process_obj => {type => OBJECT, optional => 1},
    );
} ## end sub _primary_keys_validate_spec

# The same as the base class's version, but we need "no_command_interface"
# globally set for all methods.
sub _common_validate_with {
    $Log->exit() if $may_exit;
    my $pkg_or_obj = shift;
    my %opts       = $pkg_or_obj->SUPER::_common_validate_with(@_,
        no_command_interface => 1);
    $Log->exit() if $may_exit;
    return %opts;
}

=head2 reopen()

Reopen file handle object(Instance Method)

=over

=item open_flags => $open_flags

(Required) The desired open mode for the file being opened
Valid values are:  open-existing, open-always, create-new,
create-always, truncate-existing, open-append

=item C<< access_mode => $access_mode >>

(Required) The desired access mode for the file being opened
Valid values are:  read, write, execute, read-write

=item C<< permissions => $permissions >>

(Optional) The permissions (4 bits) for the file being opened
This is ignored for Windows.(Default - 0777)

=item C<< share_mode => $share_mode >>

(Optional) The desired share mode for the file being opened
Valid values are:  shared-read, shared-write, shared-read-write,
shared-delete,shared-read-write-delete.

=item C<< attributes => $attributes >>

(Optional) The desired flags and attributes for the file being opened
This is ignored for Unix and can be any supported flags and attributes
(or combinations combined with "|") specified in the MSDN documentation
for CreateFile().  This is ignored for Unix.

=item Returns

Returns an object of type NACL::C::Client::File, representing the same filename
on which the reopen() method was invoked.

=item Example

    $fh_obj->reopen(
        'open_flags' => 'open-existing',
        'access_mode' => 'read-write',
    );

=back

=cut


sub reopen {
    $Log->enter() if $may_enter;
    my ($self,@args)  = @_;

    my %opts = $self->_common_validate_with(
        params          => \@args,
        allow_extra     => 1,
    );

    #Close the file handle if it is open.
    try {
        $self->close();
    } catch NACL::C::Exceptions::Client::File::FileException with {};

    delete $opts{file_handle};

    my $file_obj = NACL::C::Client::File->open(%opts);
    $Log->exit() if $may_exit;
    return $file_obj;
}


=head1 NAME

NACL::C::Client::Thread

=head1 DESCRIPTION

This library is helper package for NACL::C::Client::File->lock()
to return an object representing the handle of the thread started by it.
It provides various methods to control the threads created like get_status,kill, etc.

=cut

package NACL::C::Client::Thread;

use NATE::Log qw(log_global);
$Log = log_global();
$may_enter = $Log->may_enter();
$may_exit  = $Log->may_exit();
use base 'NACL::C::Component';
use Params::Validate qw(SCALAR GLOBREF OBJECT);
use NACL::ComponentUtils qw(_optional_scalars Dumper);

use Class::MethodMaker [
    new    => ['-hash', '-init', 'new'],
    scalar => 'thread_id',
];

sub init {
    $Log->enter() if $may_enter;
    my $self = shift;
    my %opts = $self->_common_validate_with(
        params          => \@_,
        additional_spec => {thread_id => {type => SCALAR | GLOBREF}}
    );
    $Log->exit() if $may_exit;
} ## end sub init

=head2 get_status()

(Instance Method only)Return the status for a waiting lock operation request
Will be called on thread object

=over

=item Returns

 SCALAR ("waiting", "exited", "invalid")
     waiting = if the thread is still alive
     exited  = if the thread has exited
     invalid = if the thread object is undef (thread does not exist)

=item Example

 my $lock_status = $thread_obj->get_status();
 $Log->log("lock wait status is: $lock_status\n");

=back

=cut

sub get_status {

    my $self = shift;
    $Log->enter() if $may_enter;
    $self->_verify_invocation(style => 'instance_only');
    my %opts = $self->_common_validate_with(params => \@_);

    # get the waiting request's thread index
    my $thread_id = $self->thread_id();

    # check if the thread object exists
    return "invalid" if (!defined $thread_id);

    # check if the thread is still alive
    # if yes, return "waiting"
    return "waiting" if ($thread_id->is_running());

    # else... the thread seems to have exited
    # so, return "exited"
    return "exited";

} ## end sub get_status

=head2 kill_waiting_request()

(Instance Method only)Abort the waiting lock request and will be called
by thread object

=over

=item Returns

 0 for SUCCESS, or throws an exception

=item Example

 $thread_obj->kill_waiting_request();

=back

=cut

sub kill_waiting_request {

    my $self = shift;
    $Log->enter() if $may_enter;
    $self->_verify_invocation(style => 'instance_only');
    my %opts = $self->_common_validate_with(params => \@_);

    # get the waiting request's thread index
    my $thread_id = $self->thread_id();
    $Log->comment("Thread" . Dumper($thread_id));
    if (!defined $thread_id) {
        $Log->exit() if $may_exit;
        NATE::BaseException->throw("Thread id is not defined\n");
    }

    # Kill the thread
    $thread_id->kill('KILL')
        or NATE::BaseException->throw("failed to kill the thread - $!\n");

    if ($thread_id->is_running()) {
        $Log->exit() if $may_exit;
        NATE::BaseException->throw("\nwaiting lock thread did not die!\n");
    }
    return 1;
} ## end sub kill_waiting_request

# The same as the base class's version, but we need "no_command_interface"
# globally set for all methods.
sub _common_validate_with {
    $Log->exit() if $may_exit;
    my $pkg_or_obj = shift;
    my %opts       = $pkg_or_obj->SUPER::_common_validate_with(@_,
        no_command_interface => 1);
    $Log->exit() if $may_exit;
    return %opts;
}

1;
