#
# $Id: $
#
#
# Copyright (c) 2013 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#
#
## @summary Ifconfig Component Module
## @author sshaik@netapp.com, dl-nacl-dev@netapp.com
## @status shared
## @pod here
################################################################################

=head1 NAME

NACL::C::Client::Ifconfig

=head1 DESCRIPTION

C<NACL::C::Client::Ifconfig> Is a derived class of
L<NACL::C::Component|lib-NACL-C-Component-pm>. It represents access to
a paths.

=head1 ATTRIBUTES

=head2 interface

Name of the interface

=head2 command_interface

See L<NACL::C::Component::command_interface|lib-NACL-C-Component-pm/command_interface>

=cut

###################
# Package
package NACL::C::Client::Ifconfig;

###################
use strict;
use warnings;

###################
# Module includes
use NACL::CS::Client::Ifconfig;
use NACL::ComponentUtils qw (_optional_scalars Dumper);

use base 'NACL::C::Component::Client';
use Params::Validate qw(validate_with SCALAR BOOLEAN);
use NACL::APISet::Exceptions::ResponseException qw/:try/;
use Net::IP qw(:PROC);

use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();

=head1 METHODS

=head2 new

    my $ifconfig = NACL::C::Client::Ifconfig->new( command_interface => $ci ,
							interface => $interface);

Return a new ifconfig component object with the given attributes.

=cut

use Class::MethodMaker [
    scalar => 'interface',
    scalar => [ { -type => 'NACL::C::Client' }, 'command_interface', ],
];

=head2 set_address

    NACL::C::Client::Ifconfig->set_address(command_interface=>$ci,
                                           interface => $interface
                                           );

(Instance method or Class method) Request to assign only ipv4 address to an interface.

=over

=item Options

=over

=item C<< interface => $string >>

(Required) Name of the interface.

=item C<< address => $string >>

(Required) IPV4 address to be assign to the interface.

=item C<< netmask => $string >>

(optional) Netmask value to be assigned to the interface.

=item C<< max_transfer_unit => $string >>

(optional) MTU value needs to be assigned to the interface.

=item C<< command_interface=>$command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=back

=back

=cut

sub set_address {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    $pkg_or_obj->_frontend_log_debug_opts(@_);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@_,
        additional_spec => {
            'address'           => { type => SCALAR },
            'netmask'           => { type => SCALAR, optional => 1 },
            'prefixlen'         => { type => SCALAR, optional => 1 },
            'max-transfer-unit' => { type => SCALAR, optional => 1 },
        },
    );
    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => "_set_address_linux_cli",
                interface => "CLI",
                set       => "Linux"
            },
            {   method    => "_set_address_solaris_cli",
                interface => "CLI",
                set       => "Solaris"
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $ret;
}    ##end sub set_address

sub _set_address_linux_cli {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );

    delete $opts{command_interface};
    my $apiset = delete $opts{apiset};
    my %api_opts;
    my $allowed_opts = [qw (address netmask max-transfer-unit interface)];
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%api_opts,
        copy   => $allowed_opts,
    );
    $pkg_or_obj->_api_call_log_debug_opts(
        api_opts => \%api_opts,
        api      => 'ifconfig'
    );
    $api_opts{'up'} = 1;
    $apiset->ifconfig(%api_opts, 'privilege-level' => "root");
    $Log->exit() if $may_exit;
}    ##end sub _set_address_linux_cli

sub _set_address_solaris_cli {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );

    delete $opts{command_interface};
    my $apiset = delete $opts{apiset};
    my %api_opts;
    my $allowed_opts = [qw (address netmask max-transfer-unit interface
                            inet6)];

    my $is_ip6 = Net::IP::ip_is_ipv6( $opts{'address'} );
    if ( $is_ip6 == 1 ) {
        $opts{'inet6'} = 1;
        $api_opts{'up'} = 1;
        if ( $opts{'prefixlen'} ) {
            $opts{'address'}
                = $opts{'address'} . '/' . $opts{'prefixlen'};
        } else {
            $Log->exit() if $may_exit;
            NATE::Exceptions::Argument->throw('Prefixlen is required param for ipv6 address');

        }
    } else {
        $api_opts{'up'} = 1;
        if ( !$opts{'netmask'} ) {
            $Log->exit() if $may_exit;
            NATE::Exceptions::Argument->throw('Netmask is a required param for ipv4 address');
        }
    }

    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%api_opts,
        copy   => $allowed_opts,
    );
    $pkg_or_obj->_api_call_log_debug_opts(
        api_opts => \%api_opts,
        api      => 'ifconfig'
    );
    $api_opts{'up'} = 1;
    $apiset->ifconfig(%api_opts, 'privilege-level' => "root");
    $Log->exit() if $may_exit;
}    ##end sub _set_address_linux_cli

=head2 down

    NACL::C::Client::Ifconfig->down(command_interface=>$ci,
                                    interface => $interface
                                    );

(Instance method or Class method) Request to down an interface.

=over

=item Options

=over

=item C<< interface => $string >>

(Required) Name of the interface.

=item C<< command_interface=>$command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>


=back

=back

=cut

sub down {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    $pkg_or_obj->_frontend_log_debug_opts(@_);
    my %opts = $pkg_or_obj->_common_validate_with( params => \@_, );
    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => "_down_linux_cli",
                interface => "CLI",
                set       => "Linux"
            },
            {   method    => "_down_solaris_cli",
                interface => "CLI",
                set       => "Solaris"
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $ret;
}    ##end sub down

sub _down_linux_cli {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );

    delete $opts{command_interface};
    my $apiset = delete $opts{apiset};
    my %api_opts;
    my $allowed_opts = [qw (interface )];
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%api_opts,
        copy   => $allowed_opts,
    );
    $pkg_or_obj->_api_call_log_debug_opts(
        api_opts => \%api_opts,
        api      => 'ifconfig'
    );
    $api_opts{'down'} = 1;
    $apiset->ifconfig(%api_opts, 'privilege-level' => "root");
    $Log->exit() if $may_exit;
}    ##end sub _down_linux_cli

sub _down_solaris_cli {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );

    delete $opts{command_interface};
    my $apiset = delete $opts{apiset};
    my %api_opts;
    my $allowed_opts = [qw (interface)];
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%api_opts,
        copy   => $allowed_opts,
    );
    $pkg_or_obj->_api_call_log_debug_opts(
        api_opts => \%api_opts,
        api      => 'ifconfig'
    );
    $api_opts{'down'} = 1;
    $apiset->ifconfig(%api_opts, 'privilege-level' => "root");
    $Log->exit() if $may_exit;
}    ##end sub _down_linux_cli

=head2 up

    NACL::C::Client::Ifconfig->up(command_interface=>$ci,
                                  interface => $interface
                                  );

(Instance method or Class method ) Request to make state of an interface to up.

=over

=item Options

=over

=item C<< interface => $string >>

(Required) Name of the interface.

=item C<< command_interface=>$command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=back

=back

=cut

sub up {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    $pkg_or_obj->_frontend_log_debug_opts(@_);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@_,
        additional_spec => { "inet6" => { type => SCALAR, default => 0 }, },
    );

    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => "_up_linux_cli",
                interface => "CLI",
                set       => "Linux"
            },
            {   method    => "_up_solaris_cli",
                interface => "CLI",
                set       => "Solaris"
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $ret;
}    ##end sub up

sub _up_linux_cli {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );

    delete $opts{command_interface};
    my $apiset = delete $opts{apiset};
    my %api_opts;
    my $allowed_opts = [qw (interface inet6)];
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%api_opts,
        copy   => $allowed_opts,
    );
    $pkg_or_obj->_api_call_log_debug_opts(
        api_opts => \%api_opts,
        api      => 'ifconfig'
    );
    $api_opts{'up'} = 1;
    $apiset->ifconfig(%api_opts, 'privilege-level' => "root");
    $Log->exit() if $may_exit;
}    ##end sub _up_linux_cli

sub _up_solaris_cli {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );

    delete $opts{command_interface};
    my $apiset = delete $opts{apiset};
    my %api_opts;
    my $allowed_opts = [qw (interface inet6)];
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%api_opts,
        copy   => $allowed_opts,
    );
    $pkg_or_obj->_api_call_log_debug_opts(
        api_opts => \%api_opts,
        api      => 'ifconfig'
    );
    $api_opts{'up'} = 1;
    $apiset->ifconfig(%api_opts, 'privilege-level' => "root");
    $Log->exit() if $may_exit;
}    ##end sub _up_solaris_cli

=head2 set_mtu

    NACL::C::Client::Ifconfig->set_mtu(command_interface=>$ci,
                                       interface => $interface
                                       );

(Instance method or Class method) Request to assign mtu value to an interface.

=over

=item Options

=over

=item C<< interface => $string >>

(Required) Name of the interface.

=item C<< max-transfer-unit => $string >>

(optional) MTU value needs to be assigned to the interface.

=item C<< command_interface=>$command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=back

=back

=cut

sub set_mtu {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    $pkg_or_obj->_frontend_log_debug_opts(@_);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@_,
        additional_spec => { "max-transfer-unit" => { type => SCALAR }, },
    );
    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => "_set_mtu_linux_cli",
                interface => "CLI",
                set       => "Linux"
            },
            {   method    => "_set_mtu_solaris_cli",
                interface => "CLI",
                set       => "Solaris"
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $ret;
}    ##end sub set_mtu

sub _set_mtu_linux_cli {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );

    delete $opts{command_interface};
    my $apiset = delete $opts{apiset};
    my %api_opts;
    my $allowed_opts = [qw (max-transfer-unit interface)];
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%api_opts,
        copy   => $allowed_opts,
    );
    $pkg_or_obj->_api_call_log_debug_opts(
        api_opts => \%api_opts,
        api      => 'ifconfig'
    );
    $apiset->ifconfig(%api_opts, 'privilege-level' => "root");
    $Log->exit() if $may_exit;
}    ##end sub _set_mtu_linux_cli

sub _set_mtu_solaris_cli {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );

    delete $opts{command_interface};
    my $apiset = delete $opts{apiset};
    my %api_opts;
    my $allowed_opts = [qw (max-transfer-unit interface)];
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%api_opts,
        copy   => $allowed_opts,
    );
    $pkg_or_obj->_api_call_log_debug_opts(
        api_opts => \%api_opts,
        api      => 'ifconfig'
    );
    $apiset->ifconfig(%api_opts, 'privilege-level' => "root");
    $Log->exit() if $may_exit;
}    ##end sub _set_mtu_solaris_cli

=head2 add_address

   NACL::C::Client::Ifconfig->add_address(command_interface=>$ci,
                                          interface => $interface
                                          );

(Instance method or Class method) Request to add Ipv6/Ipv4 address to an interface.

=over

=item Options

=over

=item C<< interface => $string >>

(Required) Name of the interface to which ipv6 needs to be added.

=item C<< address  => $string >>

(Required) IPV4 or IPV6 address value needs to be added to the interface.

=item C<< netmask => $string >>

(optional) Netmask value to be assigned to the interface if it is IPV4 address.

=item C<< prefixlen => $string >>

(optional) prefixlen value to be assigned to the interface if it is IPV6 address.

=item C<< command_interface=>$command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=back

=item Exception

=item NACL::C::Exceptions::Client::Ifconfig::AlreadyExists

This type of exception is thrown when an attempt is made to add an already
added IPV6 address on an Interface

=back

=cut

sub add_address {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    $pkg_or_obj->_frontend_log_debug_opts(@_);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@_,
        additional_spec => {
            "address"   => { type => SCALAR },
            "netmask"   => { type => SCALAR, optional => 1 },
            "prefixlen" => { type => SCALAR, optional => 1 },
        },
    );

    my $ret;
    try {
            $ret = $pkg_or_obj->call_on_apiset(
            %opts,
            choices => [
                {   method    => "_add_address_linux_cli",
                    interface => "CLI",
                    set       => "Linux"
                },
                {   method    => "_add_address_solaris_cli",
                    interface => "CLI",
                    set       => "Solaris"
                },
            ],
        );
    }
    catch NACL::APISet::Exceptions::CommandFailedException with {
        my $exception = $_[0];
        NACL::C::Exceptions::Client::Ifconfig::AlreadyExists
            ->detect_convert_and_throw(exception => $exception);
        $Log->exit() if $may_exit;
        $exception->throw();
    };
    $Log->exit() if $may_exit;
    return $ret;
}    ##end sub add

sub _add_address_linux_cli {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );

    delete $opts{command_interface};
    my $apiset = delete $opts{apiset};
    my ( %api_opts, $allowed_opts );
    my $ip = Net::IP::ip_is_ipv6( $opts{'address'} );
    if ( $ip == 1 ) {
        $opts{'inet6'}  = 1;
        $api_opts{'up'} = 1;
        if ( $opts{'prefixlen'} ) {
            $opts{'add-address'}
                = $opts{'address'} . '/' . $opts{'prefixlen'};
        } else {
            $Log->exit() if $may_exit;
            NATE::Exceptions::Argument->throw('Prefixlen value is missing ');

        }
        $allowed_opts = [qw (interface inet6 add-address)];
        $pkg_or_obj->_hash_copy(
            source => \%opts,
            target => \%api_opts,
            copy   => $allowed_opts,
        );
    } else {
        $api_opts{'up'} = 1;
        if ( ! defined $opts{'netmask'} ) {
            $Log->exit() if $may_exit;
            NATE::Exceptions::Argument->throw('Netmask value is missing ');
        }
        $allowed_opts = [qw (interface address netmask)];
        $pkg_or_obj->_hash_copy(
            source => \%opts,
            target => \%api_opts,
            copy   => $allowed_opts,
        );
    }
    $pkg_or_obj->_api_call_log_debug_opts(
        api_opts => \%api_opts,
        api      => 'ifconfig'
    );
    $apiset->ifconfig(%api_opts, 'privilege-level' => "root");
    $Log->exit() if $may_exit;
}    ##end sub _add_address_linux_cli

sub _add_address_solaris_cli {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );

    delete $opts{command_interface};
    my $apiset = delete $opts{apiset};
    my ( %api_opts, $allowed_opts );
    my $ip = Net::IP::ip_is_ipv6( $opts{'address'} );
    if ( $ip == 1 ) {
        $opts{'inet6'} = 1;
        $api_opts{'up'} = 1;
        if ( $opts{'prefixlen'} ) {
            $opts{'add-address'}
                = $opts{'address'} . '/' . $opts{'prefixlen'};
        } else {
            $Log->exit() if $may_exit;
            NATE::Exceptions::Argument->throw('Prefixlen value is missing ');

        }
        $allowed_opts = [qw (interface inet6)];
        $pkg_or_obj->_hash_copy(
            source => \%opts,
            target => \%api_opts,
            copy   => $allowed_opts,
            map    => { "add-address" => "addif-address" },
        );
    } else {
        $api_opts{'up'} = 1;
        if ( !$opts{'netmask'} ) {
            $Log->exit() if $may_exit;
            NATE::Exceptions::Argument->throw('Netmask value is missing ');
        }
        $allowed_opts = [qw (interface netmask)];
        $pkg_or_obj->_hash_copy(
            source => \%opts,
            target => \%api_opts,
            copy   => $allowed_opts,
            map    => { "address" => "addif-address" },
        );
    }
    $pkg_or_obj->_api_call_log_debug_opts(
        api_opts => \%api_opts,
        api      => 'ifconfig'
    );
    $apiset->ifconfig(%api_opts, 'privilege-level' => "root");
    $Log->exit() if $may_exit;
}    ##end sub _add_address_solaris_cli

=head2 remove_address

    NACL::C::Client::Ifconfig->remove_address(command_interface=>$ci,
                                              interface => $interface
                                               );

(Instance method or Class method) Request to delete Ipv6/Ipv4 address to an interface.

=over

=item Options

=over

=item C<< interface => $string >>

(Required) Name of the interface to which ipv6 address need to be deleted

=item C<< address  => $string >>

(Required) IPV4 or IPV6 address value needs to be deleted to the interface.

=item C<< netmask => $string >>

(optional) Netmask value of the interface if it is IPV4 address.

=item C<< prefixlen => $string >>

(optional) prefixlen value of the interface if it is IPV6 address.

=item C<< command_interface=>$command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=back

=back

=cut

sub remove_address {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    $pkg_or_obj->_frontend_log_debug_opts(@_);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@_,
        additional_spec => {
            "address"   => { type => SCALAR },
            "netmask"   => { type => SCALAR, optional => 1 },
            "prefixlen" => { type => SCALAR, optional => 1 },
        },
    );
    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => "_remove_address_linux_cli",
                interface => "CLI",
                set       => "Linux"
            },
            {   method    => "_remove_address_solaris_cli",
                interface => "CLI",
                set       => "Solaris"
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $ret;
}    ##end sub remove

sub _remove_address_linux_cli {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );

    delete $opts{command_interface};
    my $apiset = delete $opts{apiset};
    my ( %api_opts, $allowed_opts );
    my $ip = Net::IP::ip_is_ipv6( $opts{'address'} );
    if ( $ip == 1 ) {
        $opts{'inet6'} = 1;
        if ( $opts{'prefixlen'} ) {
            $opts{'delete-address'}
                = $opts{'address'} . '/' . $opts{'prefixlen'};
        } else {
            $Log->exit() if $may_exit;
            NATE::Exceptions::Argument->throw('Prefixlen value is missing ');

        }
        $allowed_opts = [qw (interface inet6 delete-address)];
        $pkg_or_obj->_hash_copy(
            source => \%opts,
            target => \%api_opts,
            copy   => $allowed_opts,
        );
    } else {
        if ( !$opts{'netmask'} ) {
            $Log->exit() if $may_exit;
            NATE::Exceptions::Argument->throw('Netmask value is missing ');
        }
        $allowed_opts = [qw (interface netmask)];
        $pkg_or_obj->_hash_copy(
            source => \%opts,
            target => \%api_opts,
            copy   => $allowed_opts,
            map    => { "address" => "delete-address" },
        );
    }
    $pkg_or_obj->_api_call_log_debug_opts(
        api_opts => \%api_opts,
        api      => 'ifconfig'
    );
    $apiset->ifconfig(%api_opts, 'privilege-level' => "root");
    $Log->exit() if $may_exit;
}    ##end sub _remove_address_linux_cli

sub _remove_address_solaris_cli {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );

    delete $opts{command_interface};
    my $apiset = delete $opts{apiset};
    my %api_opts;
    my $ip = Net::IP::ip_is_ipv6( $opts{'address'} );
    if ( $ip == 1 ) {
        $opts{'inet6'}          = 1;
        $opts{'delete-address'} = $opts{'address'};
        my $allowed_opts = [qw (interface inet6 )];
        $pkg_or_obj->_hash_copy(
            source => \%opts,
            target => \%api_opts,
            copy   => $allowed_opts,
            map    => { "delete-address" => "removeif-address" },
        );
    } else {
        my $allowed_opts = [qw (interface )];
        $pkg_or_obj->_hash_copy(
            source => \%opts,
            target => \%api_opts,
            copy   => $allowed_opts,
            map    => { "address" => "removeif-address" },
        );
    }
    $pkg_or_obj->_api_call_log_debug_opts(
        api_opts => \%api_opts,
        api      => 'ifconfig'
    );
    $apiset->ifconfig(%api_opts, 'privilege-level' => "root");
    $Log->exit() if $may_exit;
}    ##end sub _remove_address_soalris_cli

sub _primary_keys_validate_spec {
    return ( 'interface' => { type => SCALAR }, );
}
1;

