#
#
# Copyright (c) 2013 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#
#
## @summary Mkfs Component Module
## @author skotagir@netapp.com, dl-nacl-dev@netapp.com
## @status shared
## @pod here
################################################################################

=head1 NAME

NACL::C::Client::Mkfs

=head1 DESCRIPTION

C<NACL::C::Client::Mkfs> Is a derived class of
L<NACL::C::Component|lib-NACL-C-Component-pm>. It represents access to
a filesystem.

=head1 ATTRIBUTES

=head2 device

Name of the device or hardisk partition on which you want to make filesystem

=head2 command_interface

See L<NACL::C::Component::command_interface|lib-NACL-C-Component-pm/command_interface>

=cut

###################
# Package
package NACL::C::Client::Mkfs;

###################

use strict;
use warnings;

###################
# Module includes
use NACL::ComponentUtils qw (_optional_scalars Dumper);

use base 'NACL::C::Component::Client';
use Params::Validate qw(validate_with ARRAYREF SCALAR BOOLEAN);
use NACL::APISet::Exceptions::ResponseException qw/:try/;

use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();

use Class::MethodMaker [
    scalar => 'device',
];

=head1 mkfs

    my $mkfs = NACL::C::Client::Mkfs->mkfs(command_interface=>$ci,
                                       device => <device_name>,
                                      );

(Class method) Request to make a filesystem on device and return a new component object that refers to it.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< device => $string >>

(Required) for make file system on the device or partition of a harddisk.
Applicable for Linux/Windows CLI.
Maps to:
    Windows CLI : volume

=item C<< verbose => $boolean_value >>

(Optional) If value is 1 produces verbose output the default is 0.

=item C<< fstype => $string >>

Specifies the type of file system to be built.  If not specified, the default file system type (currently ext2) is used

=item C<< volume-label => $string >>

(Optional) to set particular volume-label

=item C<< block-size => $string >>

(Optional) To mention the number of blocks to be used for mkfs

=item C<< file-name => $string >>

(Optional) Give the filename which are to be taken

=item C<< fragment-size => $string >>

(Optional) Mention the fragment size.

=item C<< bytes-per-inode => $string >>

(Optional) number of bytes to be set for per inode.

=item C<< journal => $boolean_value >>

(Optional) iif the value is set the journal(-j) flag is set, the default value is 0.

=item C<< journal-options => $string >>

(Optional) To set give Journal options.

=item C<< number-of-inodes => $string >>

(Optional) This represents the number of inodes to be created.

=item C<< reserved-blocks-percentage => $string >>

(Optional) this is used to set the percentage of reserved blocks

=item C<< blocks-per-group => $boolean_value >>

(Optional) To set the the size of each group.

=item C<< volume-label => $string >>

(Optional) To set the volume-label name.

=item C<< previous-mounted-directory => $string >>

(Optional) To give previous mounted directory name.

=item C<< feature => $string >>

(Optional) To give the feature name.

=item C<< fs-revision => $string >>

(Optional) To mention file system revision.

=item C<< library-version => $boolean_value >>

(Optional) If the value is 1 library vesion is set ,default value is 0.

=item C<< blocks-count => $string >>

(Optional) To give the count of blocks.

=item C<< options => $string >>

(Optional) To give options.

=item C<< uuid => $string >>

(Optional) To mention the uuid.

=item C<< size => $string >>

(Optional) To mention the size for mkfs

=item C<< validate => $boolean_value >>

(Optional) If the value is 1 validate switch is set ,default value is 0.

=item C<< "file-system" => $file_sys >>

(Required) The file-system which is to be created after format. This option is specific
to Windows host.

=item C<< label => $label >>

(Required) The new label which is to be assigned to the volume or disk. This option is specific
to Windows host.

=item C<< quick-format => <1|0> >>

(Required) This parameter takes as a boolean value whether to quick format a particular volume.
This option is specific to Windows host.

=item C<< "override-size" => $size >>

(Optional) This parameter takes size which is used to override the default allocation unit size.
NTFS supports 512, 1024, 2048, 4096, 8192, 16K, 32K, 64K.
FAT supports 512, 1024, 2048, 4096, 8192, 16K, 32K, 64K,
(128K, 256K for sector size > 512 bytes).
FAT32 supports 512, 1024, 2048, 4096, 8192, 16K, 32K, 64K,
(128K, 256K for sector size > 512 bytes).
exFAT supports 512, 1024, 2048, 4096, 8192, 16K, 32K, 64K,
128K, 256K, 512K, 1M, 2M, 4M, 8M, 16M, 32M.
This option is specific to Windows host.

=item C<< compressed => <1|0> >>

(Optional) This parameter takes boolean value,in order to compress the files created on the
new volume. Default behaviour is that it will compress the files.
This option is specific to Windows host.

=item C<< dismount => <1|0> >>

(Optional) This parameter takes boolean value, if specified as 1 it will force
the volume to dismount first if necessary.
This option is specific to Windows host.

=item C<< passes => <1|0> >>

(Optional) This parameter takes boolean value, if specified as 1 it will zero every
sector on the volume passes times. This parameter is not valid with "quick-format" option.
This option is specific to Windows host.

=item C<< duplicate => $boolean_value >>

(Optional) This parameter takes boolean value, if specified as 1 metadata will be
duplicated. This is applicable for UDF 2.50 only.
This option is specific to Windows host.

=item C<< revision => $revision >>

(Optional) If specified forces the format to a specific UDF version.
The possible values are 1.02, 1.50, 2.00, 2.01, 2.50. The default revision is 2.01.
This option is specific to Windows host.

=item C<< size => $boolean_value >>

(Optional) Specifies the size of the floppy disk to format.
This option is specific to Windows host.

=item C<< tracks => $tracks >>

(Optional) Specifies the number of tracks per disk side.
This option is specific to Windows host.

=item C<< sectors => $sectors >>

(Optional) Specifies the number of sectors per track.
This option is specific to Windows host.

=item C<< current_label => $value >>

(Optional) The current label of the volume or disk. This will be used when we
are trying to rename a label for a particular disk.
This option is specific to Windows host.

=back

=back

=cut

sub mkfs {
    $Log->enter() if $may_enter;
    my $pkg = shift;
    $pkg->_frontend_log_debug_opts(@_);
    my %opts = $pkg->_common_validate_with(
        params          => \@_,
        additional_spec => {
            _optional_scalars(
                qw (verbose fstype file-name block-size fragment-size bytes-per-inode
                    journal journal-options number-of-inodes reserved-blocks-percentage
                    creator-os blocks-per-group volume-label previous-mounted-directory
                    feature fs-revision options library-version blocks-count size validate
                    file-system label quick-format override-size compressed
                    dismount passes duplicate revision size tracks sectors
                    current_label
                    )
            ),
        },
    );
    my $ret = $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => "_mkfs_linux_cli",
                interface => "CLI",
                set       => "Linux"
            },
            {   method    => "_mkfs_solaris_cli",
                interface => "CLI",
                set       => "Solaris"
            },
            {   method    => "_mkfs_windows_cli",
                interface => "CLI",
                set       => "Windows"
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $ret;
} ## end sub mkfs

sub _mkfs_linux_cli {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );

    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my %api_opts;
    $pkg->_hash_copy(
        source => \%opts,
        target => \%api_opts,
        copy   => [
            qw(device verbose fstype file-name block-size fragment-size bytes-per-inode
                journal journal-options number-of-inodes reserved-blocks-percentage
                creator-os blocks-per-group volume-label previous-mounted-directory
                feature fs-revision options library-version blocks-count size
                )
        ],
    );
    my $prompts_answers = [ '.*Proceed\s+anyway?.*' => 'y' ];
    $api_opts{'connectrec-match_table'} = $prompts_answers;

    $apiset->mkfs( %api_opts, 'privilege-level' => "root" );

    my $obj = $pkg->new(
        device            => $api_opts{device},
        command_interface => $command_interface,
    );
    $Log->exit() if $may_exit;
    return $obj;
} ## end sub _mkfs_linux_cli

sub _mkfs_solaris_cli {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );

    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my %api_opts;
    $pkg->_hash_copy(
        source => \%opts,
        target => \%api_opts,
        copy   => [qw(fstype validate size device)],
    );

    my $prompts_answers = [ '.*Proceed\s+anyway?.*' => 'y' ];

    $apiset->mkfs( %api_opts, 'privilege-level' => "root" );

    my $obj = $pkg->new(
        device            => $api_opts{device},
        command_interface => $command_interface,
    );
    $Log->exit() if $may_exit;
    return $obj;
} ## end sub _mkfs_solaris_cli

sub _mkfs_windows_cli {
    $Log->enter() if $may_enter;
    my ( $pkg, @args)  = @_;
    my %opts = validate_with(
        params      => \@args,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );

    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my %api_opts;
    $pkg->_hash_copy(
        source => \%opts,
        target => \%api_opts,
        copy   => [ qw(
                file-system label quick-format override-size compressed
                dismount passes duplicate revision size tracks sectors
            ) ],
        map  => { "device" => "volume", },
    );

    my $prompts_answers = [
        '.*Proceed\s+with\s+Format.*' => 'y',
        '.*Enter current volume label for drive.*' => $opts{current_label},
        '.*Would you like to force a dismount on this volume.*'
            => 'y',
    ];

    $api_opts{'connectrec-match_table'} = $prompts_answers;
    $apiset->format(%api_opts);

    my $obj = $pkg->new(
        device            => $api_opts{volume},
        command_interface => $command_interface,
    );

    $Log->exit() if $may_exit;
    return $obj;
}

sub _primary_keys_validate_spec {
    return ( device => { type => SCALAR }, );
}

1;
