#
#
# Copyright (c) 2014 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#
#
## @summary Mount Component Module
## @author osgood@netapp.com, dl-nacl-dev@netapp.com
## @status shared
## @pod here
################################################################################

=head1 NAME

NACL::C::Client::Mount

=head1 DESCRIPTION

C<NACL::C::Client::Mount> is a derived class of
L<NACL::C::Component|lib-NACL-C-Component-pm>. It represents access to
a mount_point.

=head1 ATTRIBUTES

=head2 mount_point

What directory a remote filesystem was mounted onto

=head2 command_interface

See L<NACL::C::Component::command_interface|lib-NACL-C-Component-pm/command_interface>

=cut

###################
# Package
package NACL::C::Client::Mount;

###################
# Everytime
use strict;
use warnings;

###################
# Module includes
use NACL::CS::Client::Mount;
use NACL::ComponentUtils qw (_optional_scalars Dumper);
use NACL::TaskUtils qw (_get_client_path_to_file);
use NACL::Util::IPv6 qw(convert_ipv6_address_format handle_proto);
use base 'NACL::C::Component::Client';
use Params::Validate qw(validate_with SCALAR BOOLEAN);
use NACL::APISet::Exceptions::ResponseException qw/:try/;

use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use NACL::C::Exceptions::Client::Mount::FailedToMount;
use NACL::C::Exceptions::Client::Mount::DoesNotExist;

use Net::IP qw(:PROC);

=head1 METHODS

=head2 new

    my $mount = NACL::C::Client::Mount->new( command_interface => $ci );

Return a new mount component object with the given attributes.

=cut

use Class::MethodMaker [
    scalar => 'mount_point',
    scalar => '_auto_mount_point'
];

=head2 create

    my $mount = NACL::C::Client::Mount->create(command_interface=>$ci,
                                                server => $server,
                                                [ path => $path |
                                                  share_name => $share_name_on_server ]
                                               );

(Class method) Request creation of a new mount element and return a new component object that refers to it.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< server => $string >>

(Required) For type C<nfs>, the DNS name or IP address of a file server
to mount. For type C<device>, a more informal name of the server providing the
device (used in choosing the name of the mount_point, logging, etc.)

=item C<< path => $string >>

(Required for NFS) For type C<nfs>, the export point on that file server to mount.  For
type C<device>, the name of a block device (local to client) to mount.

=item C<< share_name => $string >>

(Required for CIFS) The share name on the file server that has to be mapped.

=item C<< type => $string >>

(Optional) Either C<nfs> or C<device>.  Default is C<nfs>.

=item C<< proto => $string >>

(Optional) For NFS mounts, what RPC transport protocol to use, such as "udp" or
"tcp".  Default is undef (which implies protocol left unspecified to
mount(8)).

=item C<< vers => $string >>

(Optional) For NFS mounts, what version of NFS to use, such as "2" or "3" or
"4").  Default is undef (which implies version left unspecified to
mount(8))

=item C<< mountopts => $string >>

(Optional) What mount options to use (what to pass to the C<-o> option of the
C<mount> executable).  For type C<device>, this value will be passed
to C<mount> unmodified, and the default is empty.  For type C<nfs>,
the default is 'hard,intr,noquota' and before the value is passed to
the C<mount> executable the C<proto> and C<vers> options given above
will be added and the options possibly translated.

The way the C<nfs>-mode option translation works is that options
should be given in Solaris style, and if the host performing the mount
is Linux or FreeBSD, known options will be translated into that
operating system style.  For Linux, options are left alone except
"noquota" is dropped if present and "vers=4" is dropped and expressed
by changing C<fstype> instead.  For FreeBSD/NetBSD/OpenBSD, the
options "hard", "soft", "intr", "bg", "proto", and "vers" are
translated into their equivalents.  No translation is yet performed
for HP/UX or AIX.

=item C<< fstype => $string >>

(Optional) The file system type to pass to the C<mount> executable (with "-t" or
"-F" option, depending on operating system).  The default is not to
pass any filesystem name, as the OS choice of defaults is generally
correct (except that on Linux when C<type> is C<nfs> and C<vers> is
C<4> then the fstype defaults to "nfs4", as that is what current
NFSv4-capable C<mount> executables on Linux require).

=item C<< mount_point => $string >>

(Optional) What directory to mount this filesystem onto (the second non-option
argument to the C<mount> executable).  The default is
"/t/server/extra" where "server" is the given C<$server> value, and
"extra" is a string generated from the C<$path> and the values given
for C<proto> and C<vers> options.

As described above, one should generally let C<mount> choose the value
of this option or else do at least as good a job of avoiding ever
having separate tests choose identical mount_point paths but
non-identical mount options.

=item C<< exclusive => $boolean_value >>

(Optional) Applicable only for Unix-like OSes; ignored (effectively
always true) for Windows.  If true, create an exclusive mount with a
unique mountpoint path, instead of the default behavior of trying to
share a mountpoint with any/all other callers who ask for a mount with
the same characteristics.

In order to provide an exclusive mount to many callers with the same
options, the mountpoint path will be appended with random hex string
suffix of the form "-xxxxxxxxx".  Note that if both the 'mount_point'
and 'exclusive' options are given, this means the mount_point that is
used and that $self->mount_point will refer to is not precisely the
mountpoint that was passed in (it will have this random suffix added)

=item C<< timeout => $string >>

(Optional) How many seconds to wait for a mount or umount attempt to finish.
The default is 60 seconds.  (C<mount> may take a total time which is
longer than this if it has to retry, since the timeout is for each
mount and umount)

=item C<< chmod => $string >>

(Optional) If given, what mode (permissions) to set on the mount_point after
mounting.  This can be any value accepted by the C<chmod(1)> Unix
command.

=item C<< chown => $string >>

(Optional) If '1', specifies that the mount_point should be 'chowned' to the
current euid. The default is '0'.

=item C<< retries => $string >>

(Optional) If given, the number of retries to attempt before failing on the
mount.  Defaults to 1.

=item C<< user => $string >>

(Optional) Username to be used to map the share

=item C<< password => $string >>

(Optional) Password for the user

=item C<< domain => $string >>

(Optional) Domain name of the user used to map the share

=item C<< savecred => $string >>

(Optional) Stores the provided credentials for reuse.

=item C<< smartcard => $boolean_value >>

(Optional) Specifies the network connection is to use the credentials on a
 smart card. If multiple smart cards are available, you are asked to specify
 the credential.

=item C<< persistent => $yes_or_no_string >>

(Optional) Either 'yes' or 'no.  Controls the use of persistent network
connections. The default is the setting used last. Deviceless connections are
not persistent. Yes saves all connections as they are made, and restores them
at next logon. No does not save the connection being made or subsequent
connections.

=item C<< home => $boolean_value >>

(Optional) A boolean option which connects a user to the home directory

=item C<< driveless_mapping >>

(Optional| default 0) A boolean option which deals with mount point
provided for net use command in case of windows client. If set to 1 mount point
can be skipped for net use command.For example when we map to UNC(Unified Naming
Convention), we dont specify drive(mount_point) like: "net use \\<ip> \<share>".
If * is specified then it will specify mount_point as  next available drive
otherwise it will specify devicename as mount_point.If set to 0 mount_point will
become a mandatory input.

=item C<< mount-prefix => $string >>

(Optional) If specified, mount-prefix string will be concatenated with mount command.
This allows user to run variety of troubleshooting tools, such as strace.

=back

=back

=cut

sub create {
    my ($pkg, @args) = @_;
    $Log->enter() if $may_enter;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            server            => { type => SCALAR },
            driveless_mapping => { type => BOOLEAN, default => 0 },
            server_command_interface =>  { isa => 'NACL::C::CommandInterface::ONTAP', optional => 1 },
            # path is required for Linux CLI
            # sharename is required for Windows CLI
            # hence both are optional
            _optional_scalars(
                qw (type proto vers fstype mount_point mountopts timeout
                    chmod chown verbose user password smartcard domain
                    savecred persistent home path exclusive share_name protocol
		    options file-system-type source-path mount-point )
            ),
            retries => { type => SCALAR, default => 1 },
            'mount-prefix' => {type => SCALAR, optional => 1 },
        },
        ignore_primary_keys => 1,
    );
    my $ret;

    try {
        $ret = $pkg->call_on_apiset(
            %opts,
            choices => [
                {   method    => "_create_windows_cli",
                    interface => "CLI",
                    set       => "Windows"
                },
                {   method    => "_create_linux_cli",
                    interface => "CLI",
                    set       => "Linux|Solaris"
                },
            ],
        );
    } except {
        my $coderef =   sub {
                            my $exp_obj = shift;
                            my $new_exception_obj =
                                NACL::C::Exceptions::Client::Mount::FailedToMount->convert
                                (exception => $exp_obj);

                            $new_exception_obj->mount_args(%opts);
                            $new_exception_obj->throw();
                        };
        return {
            "NACL::APISet::Exceptions::ResponseException" => $coderef,
            "NACL::APISet::Exceptions::TimeoutException" => $coderef,
        };
    };

    $Log->exit() if $may_exit;
    return $ret;
} ## end sub create

sub _create_windows_cli {
    my ( $pkg, @args ) = @_;
    $Log->enter() if $may_enter;
    my %opts = validate_with(
        params      => \@args,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );

    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my $driveless_mapping = delete $opts{driveless_mapping};
    my $retries           = delete $opts{retries};
    $retries              = $retries + 1;
    my %api_opts;
    $pkg->_hash_copy(
        source => \%opts,
        target => \%api_opts,
        copy   => [qw(user password smartcard savecred persistent home )],
        map    => {
            'share_name' => 'sharename',
            'timeout'    => 'connectrec-timeout'
        }
    );

    # Convert an IPv6 address to Microsofts .ip6-literal.net format
    if ( Net::IP::ip_is_ipv6( $opts{server} ) ) {
        $opts{server} = convert_ipv6_address_format(
            ipv6_address => $opts{server},
            client_type  => 'Windows'
        );
    } ## end if ( Net::IP::ip_is_ipv6...)

    # handle devicename (mount_point)
    # if ( !$api_opts{mount_point} && !$driveless_mapping ) {
    # NATE::Exceptions::Argument->throw(
    # "Option - mount_point is mandatory when driveless_mapping is false"
    # );
    # }

    $api_opts{devicename} = $opts{mount_point} || "*"
        if ( !$driveless_mapping );

    # make sharename also contain the IP as required by the API
    my $protocol = $opts{protocol} // '' ;

    if ( ($protocol ne 'nfsv3win')) {

        $api_opts{sharename}
            = "\\\\" . $opts{server} . '\\' . $api_opts{sharename};
    } else {
        $api_opts{sharename}  = "\\\\" . $opts{server};
    }

    $pkg->_api_call_log_debug_opts(
        api      => 'net_use',
        api_opts => \%api_opts
    );

    my $resp;
    my $exception;
    while ( $retries ) {
        try {
            $resp = $apiset->net_use(%api_opts);
            $retries = 0;
        }
        catch NACL::APISet::Exceptions::ResponseException with {
            my $e = $_[0];
            if ( --$retries ) {
                Tharn::snooze(10);
            } else {
               $Log->exit() if $may_exit;
               $e->throw();
            }
        };
    } ## end while ( $retries )


    my $mount_point;
    if ($driveless_mapping) {
        $mount_point = $api_opts{sharename};
    } else {
        if ( $api_opts{devicename} eq "*" ) {
            my $output = $resp->get_parsed_output();
            $mount_point = $output->[0]{local};
        } else {
            $mount_point = $api_opts{devicename};
        }
    } ## end else [ if ($driveless_mapping)]

    $Log->exit() if $may_exit;
    return $pkg->new(
        mount_point       => $mount_point,
        command_interface => $command_interface,
    );
} ## end sub _create_windows_cli

sub _create_linux_cli {
    my ( $pkg, @args ) = @_;
    $Log->enter() if $may_enter;

    my %opts = validate_with(
        params      => \@args,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );

    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my ($copy,%api_opts);
    my $is_ipv6 = "false";

    # convert server to IPv6 format and replace udp/tcp with udp6/tcp6
    # and vice versa, if the server value passed in is a literal IPv4
    # address or IPv6 address.
    if (!defined $opts{'type'} || ($opts{'type'} ne 'cifs')) {
        $opts{server} =~ s/^\[(.*)\]$/$1/;
        if ( Net::IP::ip_is_ipv6( $opts{server} ) ) {
            $opts{server} = convert_ipv6_address_format(
                ipv6_address => $opts{server},
                client_type  => 'Linux'
            );
        $is_ipv6 = "true";
        } ## end if ( Net::IP::ip_is_ipv6...)
    }
    if ( $opts{'proto'} ) {
        # To handle proto for supporting tcp6
        if ($is_ipv6 ne 'true' && $opts{'proto'} =~/tcp6/) {
            $is_ipv6 = 'true';
        }
        $opts{'proto'} = handle_proto(
            'is_ipv6' => $is_ipv6,
            'proto'   => $opts{'proto'}
        );
    } ## end if ( $opts{'proto'} )

    my %new_opts = (share_name   => 'share',
                    user         => 'username',
                    'timeout'    => 'connectrec-timeout',
    );

    if ( !(defined $opts{'file-system-type'} && $opts{'file-system-type'} =~ /cifs/) ) {
        $copy = [
            qw ( type proto vers fstype mountopts timeout domain password
                chmod chown verbose retries server path exclusive mount-prefix )
        ];
	$new_opts{'mount_point'} = 'mountpoint' if(defined($opts{'mount_point'}));

    } else {
        $copy = [
            qw ( all move bind list file-system-type optlist label uuid make-shared
		 make-slave make-private make-unbindable make-rshared make-rslave
		 make-rprivate make-runbindable options source-path mount-point ) 
	];

	$new_opts{'mount_point'} = 'mount-point' if(defined($opts{'mount_point'}));

    }

        $pkg->_hash_copy(
        source => \%opts,
        target => \%api_opts,
        map    => \%new_opts,
        copy   => $copy,
    );

    my $resp;
    if ( defined $api_opts{'file-system-type'} && $api_opts{'file-system-type'} =~ /cifs/ ) {
	$pkg->_api_call_log_debug_opts(
            api      => 'mount',
            api_opts => \%api_opts,
        );
        $resp = $apiset->mount(%api_opts,
                         'privilege-level' => 'root'
                  );
    } else {
        $pkg->_api_call_log_debug_opts(
            api      => 'do_mount',
            api_opts => \%api_opts,
        );
        $resp = $apiset->do_mount(%api_opts,
                     'privilege-level' => 'root'
                  );
    }

    my $output = $resp->get_parsed_output();
    $Log->exit() if $may_exit;

    my %hash;

    if ( !defined $opts{mount_point} ) {
        $hash{_auto_mount_point} = 1;
    }

    if ( defined $api_opts{'file-system-type'} && $api_opts{'file-system-type'} =~ /cifs/ ) {
        $hash{'mount-point'} = $api_opts{'mount-point'} ;
    } else {
        $hash{'mount_point'} = $output->[0]{mount_point};
    }

    return $pkg->new(
        command_interface => $command_interface,
        %hash
    );
} ## end sub _create_linux_cli

=head2 delete

  NACL::C::Client::Mount->delete(command_interface => $ci,
                             mount_point=>$mount_point);

                            or

  $mount_object->delete(...);

(Class or instance method) This method deletes a mount element.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< mount_point=> $string >>

(Required) The mount_point to be deleted

=item C<< remount-readonly=> $boolean >>

(Optional) Maps to -r option in umount

=item C<< free-loop-device=> $boolean >>

(Optional) Maps to -d option in umount

=item C<< all=> $boolean >>

(Optional) Maps to -a option in umount

=item C<< type=> $boolean >>

(Optional) Maps to -t option in umount

=item C<< Optlist=> $boolean >>

(Optional) Maps to -O option in umount

=item C<< force=> $boolean >>

(Optional) Maps to -f option in umount

=item C<< lazy=> $boolean >>

(Optional) Maps to -l option in umount

=back

=back

=cut

sub delete {
    my ($pkg_or_obj, @args) = @_;
    $Log->enter() if $may_enter;
    $Log->debug( "Opts to 'delete' front end are:\n" . Dumper( \@args ) );
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw (remount-readonly free-loop-device
                    all type optlist force lazy timeout
                    )
            ),
            retries => { type => SCALAR, default => 0 },
        }
    );

    if ( $opts{force} ) {
        if ( $opts{force} =~ /true/i ) {
            $opts{force} = 1;
        } elsif ( $opts{force} =~ /false/i ) {
            $opts{force} = 0;
        }
    }
    $opts{retries} = $opts{retries} + 1;
    try{
        $pkg_or_obj->call_on_apiset (
            %opts,
            choices => [
                {   method    => "_delete_linux_cli",
                    interface => "CLI",
                    set       => "Linux|Solaris"
                },
                {   method    => "_delete_windows_cli",
                    interface => "CLI",
                    set       => "Windows"
                }
            ],
        );
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = $_[0];
        NACL::C::Exceptions::Client::Mount::DoesNotExist->detect_convert_and_throw(
            exception => $exception);
        $Log->exit() if $may_exit;
        $exception->throw();
    };

    $Log->exit() if $may_exit;
} ## end sub delete

sub _delete_linux_cli {
    my ($pkg_or_obj, @args) = @_;
    $Log->enter() if $may_enter;
    my %opts = validate_with(
        params      => \@args,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1,
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my $retries           = delete $opts{retries};
    my $temp_retries      = $retries;
    my %api_opts;
    my $busy = 0;
    $pkg_or_obj->_hash_move(
        source => \%opts,
        target => \%api_opts,
        move   => [
            qw/remount-readonly free-loop-device all type optlist force
                lazy/
        ],
        map => { mount_point => 'paths', 'timeout' => 'connectrec-timeout' }
    );

    # map option name
    if ( $command_interface->type() =~ /solaris/ and $api_opts{force} ) {
        $api_opts{'forcily-umount'} = delete $api_opts{force};
    }

    # umount and rmdir the mount_point

    while ( $retries ) {

        try {
            $apiset->umount(%api_opts,
                            'privilege-level' => "root");
            $retries = 0;
        }
        catch NACL::APISet::Exceptions::ResponseException with {
            my $e = $_[0];
            my $text = $e->text();
            if ( --$retries ) {
               Tharn::snooze(10);
            } else {

               ## printing the trace of lsof -b
               $apiset->execute_raw_command( command => "lsof -b $api_opts{paths}" ) if($busy);

               $Log->exit() if $may_exit;
               $e->throw();
            }
            if ( $text =~ /busy/i ) {
               #
               #Handle the case when the client is busy
               #
               $apiset->execute_raw_command( command => "lsof $api_opts{paths}" );
               $apiset->execute_raw_command( command => "fuser -kc $api_opts{paths}" );

               $busy = 1;
            } elsif ( $text =~ /No such/i
                    && $busy  ) {
               #
               # Handle the case, where it gets the error message "no such
               # file or directory exist". In the previous interation of
               # unmounting system would have been busy,but it would have
               # unmounted it during the snooze time.
               # This can be treated as success
               #
               $retries = 0;
            } else {
               #
               #Report failure ,if it fails for any other reasons
               # like Invalid param etc..
               #

               ## printing the trace of lsof -b
               $apiset->execute_raw_command( command => "lsof -b $api_opts{paths}" ) if($busy);

               $Log->exit() if $may_exit;
               $e->throw();
            }
        };
    } ## end while ( $retries )

    # this seems to be a reasonably safe way to remove the directory
    # (mount point) that was created (if we created it)

         if ( ref $pkg_or_obj && $pkg_or_obj->_auto_mount_point() ) {
            $retries = $temp_retries;
            $busy = 0;
            my $non_empty = 0;
            while ( $retries ) {
             try {
                $apiset->rm( paths => $api_opts{paths},
                             recursive => 1,
                             force => 1,
                             'privilege-level' => "root" );
                $retries = 0;
             } ## end try
             catch NACL::APISet::Exceptions::ResponseException with {
                my $e = $_[0];
                my $text = $e->text();
                if ( --$retries ) {
                   Tharn::snooze(10);
                } else {
                   $Log->exit() if $may_exit;
                   $e->throw();
                }
                if ( $text =~ /busy/i ) {
                   #
                   #Handle the case when the client is busy
                   #
                   $apiset->execute_raw_command( command => "lsof $api_opts{paths}" );
                   $apiset->execute_raw_command( command => "fuser -c $api_opts{paths}" );

                   $busy = 1;
                } elsif ( $text =~ /not empty/i ) {
                   #
                   #Handle the case where umount has returned success ,but when it
                   #tries to delete the stale mount it would return "directory not empty"
                   #
                   $non_empty = 1;
                } elsif ( $text =~ /No such file/i
                        && ( $busy || $non_empty )  ) {
                   #
                   # In the previous interation when it tried to delete the stale mount
                   # system would have been busy , but it would have succeeded  during the
                   # snooze time. When it tries to delete in the next iteration ,it would
                   # throw "No such file or directorty" error. Which means that stale mount
                   # is cleaned up and it can be treated as success
                   #
                   $retries = 0;
                } else {
                   $Log->exit() if $may_exit;
                   $e->throw();
                }
             };
            }
         }
    $Log->exit() if $may_exit;
} ## end sub _delete_linux_cli

sub _delete_windows_cli {
    my ($pkg_or_obj, @args) = @_;
    $Log->enter() if $may_enter;
    my %opts = validate_with(
        params      => \@args,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1,
    );

    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};

    my %api_opts;
    my $force_prompt = [];

    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%api_opts,
        map    => {
            mount_point => 'devicename',
            timeout     => 'connectrec-timeout',
            force       => 'yes'
        }
    );

    if ( exists $opts{force} ) {
        if ( $opts{force} ) {
           $force_prompt = [
            "Is it OK to continue disconnecting and force them closed" => "Y" ];
        } else {
           $force_prompt = [
            "Is it OK to continue disconnecting and force them closed" => "N" ];
        }
    }

    while ( $opts{retries} ) {
        try {
            $apiset->net_use(
                %api_opts,
                delete                   => 1,
                'connectrec-match_table' => $force_prompt
            );
            $opts{retries} = 0;
        } ## end try
        catch NACL::APISet::Exceptions::ResponseException with {
	    my $e = $_[0];
            if ( --$opts{retries} ) {
               Tharn::snooze(10);
            } else {
               $Log->exit() if $may_exit;
               $e->throw();
           }
        };
    } ## end while ( $opts{retries} )

    $Log->exit() if $may_exit;
} ## end sub _delete_windows_cli

=head2 fetch_data_ip

  my $Mount_obj = NACL::C::Client::Mount->create(command_interface => $ci,
                        ......);

  $Mount_obj->fetch_data_ip(...);

(Instance method) This method is used to fetch the ip-address of the server
which has volumes / shares mounted on the client.

=cut

sub fetch_data_ip {
    my $self = shift;
    $Log->enter() if $may_enter;
    my $res  = $self->state()->server();
    $Log->exit() if $may_exit;
    return $res;
}

sub _primary_keys_validate_spec {
    return ( mount_point => { type => SCALAR }, );
}

=head2 get_client_path_to_file

    # Will return client mount path to file. Example:
    $client_path_to_file = $mount_obj->get_client_path_to_file(
        file_obj => $file_obj,
    );
    
=over

=item Options

=item C<< file_obj => $file_obj >>

An instance of type "NACL::C::VolumeFile".

=back

=cut    

sub get_client_path_to_file {
    $Log->enter() if $may_enter;
    my ($self, %args) = @_;
    
    my $client_path_to_file = _get_client_path_to_file(
        mount_obj => $self,
        file_obj  => $args{file_obj},
    );
    
    $Log->exit() if $may_exit;
    return $client_path_to_file;
}

1;
