#
#
# Copyright (c) 2017 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#
#
## @summary Cluster Component Module
## @author dl-nacl-dev@netapp.com
## @status shared
## @pod here
################################################################################

=pod

=head NAME

NACL::C::Client::ONTAPSelectDeploy::Cluster

=for html <hr>

=head DESCRIPTION

This library's purpose is to aid automation of workflows around the ONTAP Select Deploy Utility
by providing methods for cluster operations, such as cluster create, cluster offline,
cluster online, cluster delete and cluster create using json file.

=for html <hr>

=head SYNOPSIS

Create Cluster:
NACL::C::Client::ONTAPSelectDeploy::Cluster->create(
    command_interface => $ci,
    name => "cluster001",
    cluster_mgmt_ip => "10.227.64.53",
    netmask => "255.255.255.0",
    gateway => "10.168.0.1",
    node_mgmt_ips => \@node_mgmt_ips,
    node_hosts => \@node_hosts,
);

Create Cluster using JSON file:
NACL::C::ONTAPSelectDeploy::Cluster->create_using_json(
    command_interface => $ci,
    path_to_file => "/usr/bin",
    json_file => "cluster_create.json"
);

Delete Cluster:
NACL::C::ONTAPSelectDeploy::Cluster->delete(
    command_interface => $ci,
    name => "cluster001",
);

Offline Cluster:
NACL::C::ONTAPSelectDeploy::Cluster->offline(
    command_interface => $ci,
    name => "cluster001",
);

Online Cluster:
NACL::C::iONTAPSelectDeploy::Cluster->online(
    command_interface => $ci,
    name => "cluster001",
);

=head ATTRIBUTES

=head name

Name of the cluster to be created.

=cut

package NACL::C::Client::ONTAPSelectDeploy::Cluster;

use Exporter qw(import);
@EXPORT_OK = qw();

use strict;
use warnings;

use Options;
use Params::Validate qw(:all);
use NATE::BaseException;
use base 'NACL::C::Component::Client';
use NACL::ComponentUtils qw(_optional_scalars);

use Class::MethodMaker [
];

use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
###############################################################################
#           Cluster Operation methods
###############################################################################

=head create

Create clusters on the ONTAP Select Deploy Utility.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)

=item C<< name => SCALAR >>

 (Required) Name of the cluster to be created.

=item C<< cluster_mgmt_ip => SCALAR >>

 (Required) IP address to use for Cluster Mgmt LIF

=item C<< netmask => SCALAR >>

 (Required) Netmask for the Mgmt IP Addresses

=item C<< gateway => SCALAR >>

(Required) Default Gateway for the Mgmt IP Addresses

=item CC<< node_mgmt_ips => ARRAYREF >>

(Required) Array list of IPs to use for each Node Mgmt LIF

=item C<< node_hosts => ARRAYREF >>

(Required) Array list of Host IDs to use for this cluster.
The order of this list matches that of node-names.

=item C<< admin_password => SCALAR >>

(Optional) Password to set for the ONTAP admin user

=item C<< ontap_select_version => SCALAR >>

(Optional) The ONTAP Select version to use when deploying this cluster

=item C<< node_names => ARRAYREF >>

(Optional) Array list of names to use for each node. This is an optional parameter.
If the list of node names is not included, ONTAP Deploy will autogenerate node names

=item C<< node_mirrors => ARRAYREF >>

(Optional) Array list of the mirror of each node, the name of each mirror node is what is provided in this
command. This is an optional parameter. If this list is not included ONTAP Deploy will autogenerate the
mirror mapping.

=item C<< inhibit_rollback => SCALAR >>
 (Optional) Flag to not cleanup on a cluster create failure

=item C<< dns_ips => ARRAYREF >>

(Optional) Array list of DNS IP addresses to be provided to ONTAP.

=item C<< dns_domains => ARRAYREF >>

(Optional) Array list of DNS search domains to be provided to ONTAP.

=item C<< ntp_servers => ARRAYREF >>

(Optional) Space separated list of NTP server addresses to be provided to ONTAP.

=back

=item Examples

# create new ONTAP Select cluster
NACL::C::Client::ONTAPSelectDeploy::Cluster->create(
    command_interface => $ci,
    name => "cluster001",
    cluster_mgmt_ip => "10.227.64.53",
    netmask => "255.255.255.0",
    gateway => "10.168.0.1",
    node_mgmt_ips => \@node_mgmt_ips,
    node_hosts => \@node_hosts,
    admin_password => "admin123",
    ontap_select_version => "R8.4.1xN_161010_1155",
    node_names => \@node_names,
    node_mirrors => \@node_mirrors,
    inhibit_rollback => "false",
    dns_ips => \@dns_ips,
    dns_domains => \@dns_domains,
    ntp_servers => \@ntp_servers,
);

=back

=cut

sub create {
    my ($pkg, @args) = @_;
    $Log->enter() if $may_enter;
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
	additional_spec => {
	    name		=> { type => SCALAR },
	    'cluster_mgmt_ip'   => { type => SCALAR },
            netmask		=> { type => SCALAR },
            gateway		=> { type => SCALAR },
	    'node_mgmt_ips'	=> { type => ARRAYREF },
	    'node_hosts'	=> { type => ARRAYREF },
	    _optional_scalars(
		qw( admin_password ontap_select_version
		    inhibit_rollback )
	    ),
	    'node_names'	=> { type => ARRAYREF, optional => 1 },
	    'node_mirrors'	=> { type => ARRAYREF, optional => 1 },
            'dns_ips'		=> { type => ARRAYREF, optional => 1 },
            'dns_domains'	=> { type => ARRAYREF, optional => 1 },
            'ntp_servers'	=> { type => ARRAYREF, optional => 1 },
	    help		=> { type => BOOLEAN, optional => 1 },
	},
	allow_extra => 1,
    );
    my $ret = $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => "_create",
                interface => "CLI",
                set       => "ONTAPSelectDeploy"
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $ret;
} ## end sub create

sub _create {
    my ( $pkg, @args ) = @_;
    $Log->enter() if $may_enter;
    $pkg->_base_backend_ontapselectdeploy_cli( @args,
	api => 'cluster_create' );
    $Log->exit() if $may_exit;
} ## end sub _create

=head create_using_json

Create clusters on the ONTAP Select Deploy Utility using json file.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)

=item C<< json_file => SCALAR >>

 (Required) Name of the file containing cluster create JSON configuration.

=item C<< path_to_file => SCALAR >>

 (Optional) Path to cluster create JSON file location (Env: Home)

=back

=item Returns

None for success, throws an exception on failure.

=item Examples

# Create Cluster using JSON file:
NACL::C::Client::ONTAPSelectDeploy::Cluster->create_using_json(
    command_interface => $ci,
    path_to_file => "/usr/home",
    json_file => "cluster_create.json"
);

=back

=cut

sub create_using_json {
    my ($pkg, @args) = @_;
    $Log->enter() if $may_enter;
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
	additional_spec => {
	    'json_file'		=> { type => SCALAR },
            'path_to_file'	=> { type => SCALAR, optional => 1 },
	},
	allow_extra => 1,
    );
    my $ret = $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => "_create_using_json",
                interface => "CLI",
                set       => "ONTAPSelectDeploy"
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $ret;
} ## end sub create_using_json

sub _create_using_json {
    my ( $pkg, @args ) = @_;
    $Log->enter() if $may_enter;
    $pkg->_base_backend_ontapselectdeploy_cli( @args,
	api => 'cluster_create_using_json' );
    $Log->exit() if $may_exit;
} ## end sub _create_using_json

=head cluster_delete

Delete cluster using cluster name.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)

=item C<< cluster_name => SCALAR >>

 (Required) Name of the cluster to be deleted.

=back

=item Returns

None for success, throws an exception on failure.

=item Examples

# Delete Cluster :
NACL::C::Client::ONTAPSelectDeploy::Cluster->delete(
    command_interface => $ci,
    cluster_name => "cluster001",
);

=back

=cut

sub delete {
    my ($pkg, @args) = @_;
    $Log->enter() if $may_enter;
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
	additional_spec => {
	    cluster_name	=> { type => SCALAR },
	    force		=> { type => BOOLEAN, optional => 1 },
	    help		=> { type => BOOLEAN, optional => 1 },
	},
        allow_extra => 1,
    );
    my $ret = $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => "_delete",
                interface => "CLI",
                set       => "ONTAPSelectDeploy"
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $ret;
} ## end sub delete

sub _delete {
    my ( $pkg, @args ) = @_;
    $Log->enter() if $may_enter;
    $pkg->_base_backend_ontapselectdeploy_cli( @args,
	api => 'cluster_delete' );
    $Log->exit() if $may_exit;
} ## end sub _delete

=head offline

Offline cluster using cluster name.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)

=item C<< cluster_name => SCALAR >>

 (Required) Name of the cluster to be offlined.

=back

=item Returns

None for success, throws an exception on failure.

=item Examples

# Offline Cluster :
NACL::C::Client::ONTAPSelectDeploy::Cluster->offline(
    command_interface => $ci,
    cluster_name => "cluster001",
);

=back

=cut

sub offline {
    my ($pkg, @args) = @_;
    $Log->enter() if $may_enter;
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            cluster_name        => { type => SCALAR },
            force               => { type => BOOLEAN, optional => 1 },
            help                => { type => BOOLEAN, optional => 1 },
        },
        allow_extra => 1,
    );
    my $ret = $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => "_offline",
                interface => "CLI",
                set       => "ONTAPSelectDeploy"
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $ret;
} ## end sub offline

sub _offline {
    my ( $pkg, @args ) = @_;
    $Log->enter() if $may_enter;
    $pkg->_base_backend_ontapselectdeploy_cli( @args,
        api => 'cluster_offline' );
    $Log->exit() if $may_exit;
} ## end sub _offline

=head online

Online cluster using cluster name.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)

=item C<< cluster_name => SCALAR >>

 (Required) Name of the cluster to be brought online.

=back

=item Returns

None for success, throws an exception on failure.

=item Examples

# Online Cluster :
NACL::C::Client::ONTAPSelectDeploy::Cluster->online(
    command_interface => $ci,
    cluster_name => "cluster001",
);

=back

=cut

sub online {
    my ($pkg, @args) = @_;
    $Log->enter() if $may_enter;
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
	additional_spec => {
	    cluster_name	=> { type => SCALAR },
	    help		=> { type => BOOLEAN, optional => 1 },
	},
    );
    my $ret = $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => "_online",
                interface => "CLI",
                set       => "ONTAPSelectDeploy"
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $ret;
} ## end sub online

sub _online {
    my ( $pkg, @args ) = @_;
    $Log->enter() if $may_enter;
    $pkg->_base_backend_ontapselectdeploy_cli( @args,
	api => 'cluster_online' );
    $Log->exit() if $may_exit;
} ## end sub _online
1;
