#
#
# Copyright (c) 2017 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#
#
## @summary Configuration Component Module
## @author dl-nacl-dev@netapp.com
## @status shared
## @pod here
################################################################################

=pod

=head NAME

NACL::C::Client::ONTAPSelectDeploy::Configuration

=for html <hr>

=head DESCRIPTION

This library's purpose is to aid automation of workflows around the ONTAP Select Deploy Utility
by providing methods for configuration operations, such as configuration backup, configuration
restore.

=for html <hr>

=head SYNOPSIS

Backup the ONTAP Deploy VM:
NACL::C::Client::ONTAPSelectDeploy::Configuration->backup(
    command_interface => $ci,
    path => $path
);

Restore an ONTAP Deploy configuration backup:
NACL::C::Client::ONTAPSelectDeploy::Configuration->restore(
    command_interface => $ci,
    file_name => $file,
    path => $path
);
=head ATTRIBUTES

=head file_name

Name of the configuration file

=cut

package NACL::C::Client::ONTAPSelectDeploy::Configuration;

use Exporter qw(import);
@EXPORT_OK = qw();

use strict;
use warnings;

use Options;
use Params::Validate qw(:all);
use NATE::BaseException;
use base 'NACL::C::Component::Client';
use NACL::ComponentUtils qw(_optional_scalars);

use Class::MethodMaker [
];

use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
###############################################################################
#           Configuration Operation methods
###############################################################################

=head backup

Backup the ONTAP Deploy VM

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)

=item C<< path => SCALAR >>

 (Optional) Path to configuration restore file location. (Env: Home)

=back

=item Examples

# Backup the ONTAP Deploy VM:
NACL::C::Client::ONTAPSelectDeploy::Configuration->backup(
    command_interface => $ci,
    path => $path
);

=back

=cut

sub backup {
    my ($pkg, @args) = @_;
    $Log->enter() if $may_enter;
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
	additional_spec => {
	    path		=> { type => SCALAR, optional => 1 },
	    help		=> { type => BOOLEAN, optional => 1 },
	},
	allow_extra => 1,
    );
    my $ret = $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => "_backup",
                interface => "CLI",
                set       => "ONTAPSelectDeploy"
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $ret;
} ## end sub backup

sub _backup {
    my ( $pkg, @args ) = @_;
    $Log->enter() if $may_enter;
    $pkg->_base_backend_ontapselectdeploy_cli( @args,
	api => 'configuration_backup' );
    $Log->exit() if $may_exit;
} ## end sub _backup

=head restore

Restore an ONTAP Deploy configuration backup

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)

=item C<< file_name => SCALAR >>

 (Required) Name of the configuration file

=item C<< path => SCALAR >>

 (Optional) Path to configuration restore file location. (Env: Home)

=back

=item Examples

# Restore an ONTAP Deploy configuration backup:
NACL::C::Client::ONTAPSelectDeploy::Configuration->restore(
    command_interface => $ci,
    file_name => $file,
    path => $path
);

=back

=cut

sub restore {
    my ($pkg, @args) = @_;
    $Log->enter() if $may_enter;
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            file_name		=> { type => SCALAR },
            path		=> { type => SCALAR, optional => 1 },
            help                => { type => BOOLEAN, optional => 1 },
        },
    );
    my $ret = $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => "_restore",
                interface => "CLI",
                set       => "ONTAPSelectDeploy"
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $ret;
} ## end sub restore

sub _restore {
    my ( $pkg, @args ) = @_;
    $Log->enter() if $may_enter;
    $pkg->_base_backend_ontapselectdeploy_cli( @args,
        api => 'configuration_restore' );
    $Log->exit() if $may_exit;
} ## end sub _restore

1;
