#
#
# Copyright (c) 2017 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#
#
## @summary License Component Module
## @author dl-nacl-dev@netapp.com
## @status shared
## @pod here
################################################################################

=pod

=head NAME

NACL::C::Client::ONTAPSelectDeploy::License

=for html <hr>

=head DESCRIPTION

This library's purpose is to aid automation of workflows around the ONTAP Select Deploy Utility
by providing methods for License operations, such as license add and license delete.

=for html <hr>

=head SYNOPSIS

Add a new ONTAP Capacity License File:
NACL::C::Client::ONTAPSelectDeploy::License->add(
    command_interface => $ci,
    license_file => "deploy_license.txt",
    username => "admin",
    password => "netapp123",
    path_to_file => "/usr/env/home",
);

Delete a specific license from ONTAP Select Deploy:
NACL::C::Client::ONTAPSelectDeploy::License->delete(
    command_interface => $ci,
    serialnumber => "DNIVOG914",
);

=head ATTRIBUTES

=head license_file

License file name to add

=head serialnumber

Serial number associated with the license

=cut

package NACL::C::Client::ONTAPSelectDeploy::License;

use Exporter qw(import);
@EXPORT_OK = qw();

use strict;
use warnings;

use Options;
use Params::Validate qw(:all);
use NATE::BaseException;
use base 'NACL::C::Component::Client';
use NACL::ComponentUtils qw(_optional_scalars);

use Class::MethodMaker [
    scalar => 'name',
#    scalar => '_auto_mount_point'
];

use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
###############################################################################
#           License Operation methods
###############################################################################

=head add

Add a new ONTAP Capacity License File:

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)

=item C<< license_file => SCALAR >>

 (Required) License file name to add

=item C<< username => SCALAR >>

 (Optional) ONTAP administrative user

=item C<< password => SCALAR >>

 (Optional) Password for the ONTAP administrative user

=item C<< path_to_file => SCALAR >>

(Optional) Path to the location of the license file. env('HOME')

=back

=item Returns

None for success, throws an exception on failure.

=item Examples

#Add a new ONTAP Capacity License File:
NACL::C::Client::ONTAPSelectDeploy::License->add(
    command_interface => $ci,
    license_file => "deploy_license.txt",
    username => "admin",
    password => "netapp123",
    path_to_file => "/usr/env/home",
);

=back

=cut

sub add {
    my ($pkg, @args) = @_;
    $Log->enter() if $may_enter;
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
	additional_spec => {
	    license_file	=> { type => SCALAR,},
            username		=> { type => SCALAR, optional => 1,},
            password		=> { type => SCALAR, optional => 1,},
            path_to_file	=> { type => SCALAR, optional => 1,},
	    help		=> { type => BOOLEAN, optional => 1 },
        },
        allow_extra => 1,
    );
    my $ret = $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => "_add",
                interface => "CLI",
                set       => "ONTAPSelectDeploy"
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $ret;
} ## end sub add

sub _add {
    my ( $pkg, @args ) = @_;
    $Log->enter() if $may_enter;
    $pkg->_base_backend_ontapselectdeploy_cli( @args,
	api => 'license_add' );
    $Log->exit() if $may_exit;
} ## end sub _add

=head delete

Delete a specific license from ONTAP Select Deploy.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)

=item C<< serialnumber => SCALAR >>

 (Required) Serial number associated with the license

=back

=item Returns

None for success, throws an exception on failure.

=item Examples

# Delete a specific license from ONTAP Select Deploy.
NACL::C::Client::ONTAPSelectDeploy::License->delete(
    command_interface => $ci,
    serialnumber => "DNIVOG914",
);

=back

=cut

sub delete {
    my ($pkg, @args) = @_;
    $Log->enter() if $may_enter;
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
	additional_spec => {
	    serialnumber	=> {type => SCALAR,},
	    help		=> { type => BOOLEAN, optional => 1 },
	},
        allow_extra => 1,
    );
    my $ret = $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => "_delete",
                interface => "CLI",
                set       => "ONTAPSelectDeploy"
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $ret;
} ## end sub delete

sub _delete {
    my ( $pkg, @args ) = @_;
    $Log->enter() if $may_enter;
    $pkg->_base_backend_ontapselectdeploy_cli( @args,
	api => 'license_delete' );
    $Log->exit() if $may_exit;
} ## end sub _delete

1;
