#
#
# Copyright (c) 2017 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#
#
## @summary NetworkConnectivityCheck Component Module
## @author dl-nacl-dev@netapp.com
## @status shared
## @pod here
################################################################################

=pod

=head NAME

NACL::C::Client::ONTAPSelectDeploy::NetworkConnectivityCheck

=for html <hr>

=head DESCRIPTION

This library's purpose is to aid automation of workflows around the ONTAP Select Deploy Utility
by providing methods for network connectivity-check operations, such as network connectivity-check
start, network connectivity-check cancel, network connectivity-check delete.

=for html <hr>

=head SYNOPSIS

Initiate a connectivity check network setup between hosts:
NACL::C::Client::ONTAPSelectDeploy::NetworkConnectivityCheck->start(
    command_interface => $ci,
    host_ids => [1,2,3...],
    vswitch_type => 'StandardVSwitch',
    mode => 'quick',
);

Cancel a connectivity check run:
NACL::C::Client::ONTAPSelectDeploy::NetworkConnectivityCheck->cancel(
    command_interface => $ci,
    run_id => 123,
);

Delete a connectivity check run:
NACL::C::Client::ONTAPSelectDeploy::NetworkConnectivityCheck->delete(
    command_interface => $ci,
    run_id => 123,
);

=head ATTRIBUTES

=head host_ids

Host Ids(IP address or FQDN)

=cut

package NACL::C::Client::ONTAPSelectDeploy::NetworkConnectivityCheck;

use Exporter qw(import);
@EXPORT_OK = qw();

use strict;
use warnings;

use Options;
use Params::Validate qw(:all);
use NATE::BaseException;
use base 'NACL::C::Component::Client';
use NACL::ComponentUtils qw(_optional_scalars);

use Class::MethodMaker [
];

use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
###############################################################################
#           NetworkConnectivityCheck Operation methods
###############################################################################

=head start

Initiate a connectivity check network setup between hosts

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

 (Required for class method, Not Applicable for instance method)

=item C<< host_ids => ARRAYREF >>

 (Required) Comma separated Host Ids(IP address or FQDN) of the
 four hosts to run the test. The hosts must be authenticated
 and configured in ONTAP Deploy

=item C<< vswitch_type => SCALAR >>

 (Optional) {StandardVSwitch,DistributedVSwitch,OpenVSwitch}
 vSwitch type to be used for connectivity check run.
 Allowed values are StandardVSwitch, DistributedVSwitch, OpenVSwitch

=item C<< mode => SCALAR >>

 (Optional) {quick,extended,cleanup}
 Run mode for connectivity check. Allowed values are quick, extended,
 cleanup. Default: quick

=item C<< mtu => SCALAR >>

 (Optional) MTU size to be used for connectivity check run

=item C<< subnet => SCALAR >>

 (Optional) Subnet to be used to create virtual ports for testing

=item C<< sleep_time => SCALAR >>

 (Optional) Time in seconds to sleep before resuming connectivity
 check run after the NICs are brought up. Applies to extended mode only

=back

=item Examples

# Initiate a connectivity check network setup between hosts:
NACL::C::Client::ONTAPSelectDeploy::NetworkConnectivityCheck->start(
    command_interface => $ci,
    host_ids => [1,2,3...],
    vswitch_type => 'StandardVSwitch',
    mode => 'quick',
);

=back

=cut

sub start {
    my ($pkg, @args) = @_;
    $Log->enter() if $may_enter;
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
	additional_spec => {
	    host_ids		=> { type => ARRAYREF },
	    _optional_scalars(
		qw( vswitch_type mode mtu subnet sleep_time )
	    ),
	    help		=> { type => BOOLEAN, optional => 1 },
	},
	allow_extra => 1,
    );
    my $ret = $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => "_start",
                interface => "CLI",
                set       => "ONTAPSelectDeploy"
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $ret;
} ## end sub start

sub _start {
    my ( $pkg, @args ) = @_;
    $Log->enter() if $may_enter;
    $pkg->_base_backend_ontapselectdeploy_cli( @args,
	api => 'network_connectivity_check_start' );
    $Log->exit() if $may_exit;
} ## end sub _start

=head cancel

Cancel a connectivity check run

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

 (Required for class method, Not Applicable for instance method)

=item C<< run_id => SCALAR >>

 (Required) Id of the run to cancel

=back

=item Examples

# Cancel a connectivity check run:
NACL::C::Client::ONTAPSelectDeploy::NetworkConnectivityCheck->cancel(
    command_interface => $ci,
    run_id => 123,
);

=back

=cut

sub cancel{
    my ($pkg, @args) = @_;
    $Log->enter() if $may_enter;
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
	additional_spec => {
	    run_id         => { type => SCALAR,},
	    _optional_scalars(
		qw( )
	    ),
	    help		=> { type => BOOLEAN, optional => 1 },
	},
	allow_extra => 1,
    );
    my $ret = $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => "_cancel",
                interface => "CLI",
                set       => "ONTAPSelectDeploy"
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $ret;
} ## end sub cancel

sub _cancel{
    my ( $pkg, @args ) = @_;
    $Log->enter() if $may_enter;
    $pkg->_base_backend_ontapselectdeploy_cli( @args,
	api => 'network_connectivity_check_cancel' );
    $Log->exit() if $may_exit;
} ## end sub _cancel

=head delete

Delete a connectivity check run

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

 (Required for class method, Not Applicable for instance method)

=item C<< run_id => SCALAR >>

 (Required) Id of the run to cancel

=back

=item Examples

# Delete a connectivity check run:
NACL::C::Client::ONTAPSelectDeploy::NetworkConnectivityCheck->delete(
    command_interface => $ci,
    run_id => 123,
);

=back

=cut

sub delete{
    my ($pkg, @args) = @_;
    $Log->enter() if $may_enter;
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            run_id         => { type => SCALAR,},
            _optional_scalars(
                qw( )
            ),
            help                => { type => BOOLEAN, optional => 1 },
        },
        allow_extra => 1,
    );
    my $ret = $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => "_delete",
                interface => "CLI",
                set       => "ONTAPSelectDeploy"
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $ret;
} ## end sub delete

sub _delete{
    my ( $pkg, @args ) = @_;
    $Log->enter() if $may_enter;
    $pkg->_base_backend_ontapselectdeploy_cli( @args,
        api => 'network_connectivity_check_delete' );
    $Log->exit() if $may_exit;
} ## end sub _delete

1;
