#
#
# Copyright (c) 2017 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#
#
## @summary Node Component Module
## @author dl-nacl-dev@netapp.com
## @status shared
## @pod here
################################################################################

=pod

=head NAME

NACL::C::Client::ONTAPSelectDeploy::Node

=for html <hr>

=head DESCRIPTION

This library's purpose is to aid automation of workflows around the ONTAP Select Deploy Utility
by providing methods for node operations, such as node console connect, node console disconnect,
node coredump, node replace, node savecore, node start and node stop.

=for html <hr>

=head SYNOPSIS

Connect to the console port of the node:
NACL::C::Client::ONTAPSelectDeploy::Node->console_connect(
    command_interface => $ci,
    node_name => "node001",
    cluster_name => "cluster001",
);

Disconnect from the console port of the node:
NACL::C::Client::ONTAPSelectDeploy::Node->console_disconnect(
    command_interface => $ci,
    node_name => "node001",
    cluster_name => "cluster001",
);

Trigger Node to generate a coredump.
NACL::C::Client::ONTAPSelectDeploy::Node->coredump(
    command_interface => $ci,
    node_name => "node001",
    cluster_name => "cluster001",
    username => "admin",
    password => "admin123",
);

The node replace command is used to redeploy a failed node in an HA cluster:
NACL::C::Client::ONTAPSelectDeploy::Node->replace(
    command_interface => $ci,
    node_name => "node001",
    cluster_name => "cluster001",
    username => "admin",
    password => "admin123",
    hostid => "sti-ctl001.gdl.englab.netapp.com",
);

Save a coredump from the specified node:
NACL::C::Client::ONTAPSelectDeploy::Node->savecore(
    command_interface => $ci,
    node_name => "node001",
    cluster_name => "cluster001",
    case_number => "1001",
    message => "Savecore-from-test-induced-dump",
);

Start the Node - This will power on the VM:
NACL::C::Client::ONTAPSelectDeploy::Node->start(
    command_interface => $ci,
    node_name => "node001",
    cluster_name => "cluster001",
);

Shutdown the node:
NACL::C::Client::ONTAPSelectDeploy::Node->stop(
    command_interface => $ci,
    node_name => "node001",
    cluster_name => "cluster001",
    force => "true",
);

=head ATTRIBUTES

=head node_name

Name of the node in the cluster.

=head cluster_name

Name of the cluster.

=cut

package NACL::C::Client::ONTAPSelectDeploy::Node;

use Exporter qw(import);
@EXPORT_OK = qw();

use strict;
use warnings;

use Options;
use Params::Validate qw(:all);
use NATE::BaseException;
use base 'NACL::C::Component::Client';
use NACL::ComponentUtils qw(_optional_scalars);

use Class::MethodMaker [
];

use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
###############################################################################
#           Node Operation methods
###############################################################################

=head console_connect

Connect to the console port of the node:

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)

=item C<< node_name => SCALAR >>

 (Required) Name of the node to connect to.

=item C<< cluster_name => SCALAR >>

 (Required) Name of the cluster

=back

=item Returns

None for success, throws an exception on failure.

=item Examples

# Connect to the console port of the node
NACL::C::Client::ONTAPSelectDeploy::Node->console_connect(
    command_interface => $ci,
    node_name => "node001",
    cluster_name => "cluster001",
);

=back

=cut

sub console_connect {
    my ($pkg, @args) = @_;
    $Log->enter() if $may_enter;
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
	additional_spec => {
            node_name		=> { type => SCALAR,},
            cluster_name	=> { type => SCALAR,},
	    help		=> { type => BOOLEAN, optional => 1 },
	},
	allow_extra	=> 1,
    );
    my $ret = $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => "_console_connect",
                interface => "CLI",
                set       => "ONTAPSelectDeploy"
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $ret;
} ## sub console_connect

sub _console_connect {
    my ( $pkg, @args ) = @_;
    $Log->enter() if $may_enter;
    $pkg->_base_backend_ontapselectdeploy_cli( @args,
	api => 'node_console_connect' );
    $Log->exit() if $may_exit;
} ## sub _console_connect

=head console_disconnect

Disconnect from the console port of the node:

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)

=item C<< node_name => SCALAR >>

 (Required) Name of the node to disconnect from.

=item C<< cluster_name => SCALAR >>

 (Required) Name of the cluster

=back

=item Returns

None for success, throws an exception on failure.

=item Examples

# DisConnect from the console port of the node
NACL::C::Client::ONTAPSelectDeploy::Node->console_disconnect(
    command_interface => $ci,
    node_name => "node001",
    cluster_name => "cluster001",
);

=back

=cut

sub console_disconnect {
    my ($pkg, @args) = @_;
    $Log->enter() if $may_enter;
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            node_name		=> { type => SCALAR,},
            cluster_name	=> { type => SCALAR,},
            help                => { type => BOOLEAN, optional => 1 },
        },
        allow_extra	=> 1,
    );
    my $ret = $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => "_console_disconnect",
                interface => "CLI",
                set       => "ONTAPSelectDeploy"
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $ret;
} ## end sub console_disconnect

sub _console_disconnect {
    my ( $pkg, @args ) = @_;
    $Log->enter() if $may_enter;
    $pkg->_base_backend_ontapselectdeploy_cli( @args,
        api => 'node_console_disconnect' );
    $Log->exit() if $may_exit;
} ## end _console_disconnect

=head coredump

Trigger this Node to generate a coredump.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)

=item C<< cluster_name => SCALAR >>

 (Required) Name of the cluster

=item C<< node_name => SCALAR >>

 (Required) Name of the node

=item C<< username => SCALAR >>

 (Required) Valid user associated with the host

=item C<< password => SCALAR >>

 (Optional) Corresponding password associated with the user

=back

=item Returns

None for success, throws an exception on failure.

=item Examples

# Trigger this Node to generate a coredump:
NACL::C::Client::ONTAPSelectDeploy::Node::coredump(
    command_interface => $ci,
    node_name => "node001",
    cluster_name => "cluster001",
    username => "admin",
    password => "admin123",
);

=back

=cut

sub coredump {
    my ($pkg, @args) = @_;
    $Log->enter() if $may_enter;
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            node_name		=> { type => SCALAR,},
            cluster_name	=> { type => SCALAR,},
            username		=> { type => SCALAR,},
            password		=> { type => SCALAR, optional => 1,},
            help                => { type => BOOLEAN, optional => 1 },
        },
        allow_extra	=> 1,
    );
    my $ret = $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => "_coredump",
                interface => "CLI",
                set       => "ONTAPSelectDeploy"
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $ret;
} ## end sub coredump

sub _coredump {
    my ( $pkg, @args ) = @_;
    $Log->enter() if $may_enter;
    $pkg->_base_backend_ontapselectdeploy_cli( @args,
        api => 'node_coredump' );
    $Log->exit() if $may_exit;
} ## end sub _coredump

=head replace

The node replace command is used to redeploy a failed node in an HA cluster.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)

=item C<< cluster_name => SCALAR >>

 (Required) Name of the cluster

=item C<< node_name => SCALAR >>

 (Required) Name of the node

=item C<< username => SCALAR >>

 (Required) Valid user associated with the host

=item C<< password => SCALAR >>

 (Optional) Corresponding password associated with the user

=item C<< hostid => SCALAR >>

 (Optional) Name of the host to add replacing node on

=back

=item Returns

None for success, throws an exception on failure.

=item Examples

# To redeploy a failed node in an HA cluster:
NACL::C::Client::ONTAPSelectDeploy::Node::replace(
    command_interface => $ci,
    node_name => "node001",
    cluster_name => "cluster001",
    username => "admin",
    password => "admin123",
    hostid => "sti-ctl001.gdl.englab.netapp.com",
);

=back

=cut

sub replace {
    my ($pkg, @args) = @_;
    $Log->enter() if $may_enter;
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            node_name		=> { type => SCALAR,},
            cluster_name	=> { type => SCALAR,},
            username		=> { type => SCALAR,},
            password		=> { type => SCALAR, optional => 1,},
            hostid		=> { type => SCALAR, optional => 1,},
            help                => { type => BOOLEAN, optional => 1 },
        },
        allow_extra	=> 1,
    );
    my $ret = $pkg->call_on_apiset(
        %opts,
	choices => [
            {   method    => "_replace",
                interface => "CLI",
                set       => "ONTAPSelectDeploy"
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $ret;
} ## end sub replace

sub _replace {
    my ( $pkg, @args ) = @_;
    $Log->enter() if $may_enter;
    $pkg->_base_backend_ontapselectdeploy_cli( @args,
        api => 'node_replace' );
    $Log->exit() if $may_exit;
} ## end _replace

=head savecore

Save a coredump from the specified node.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)

=item C<< cluster_name => SCALAR >>

 (Required) Name of the cluster

=item C<< node_name => SCALAR >>

 (Required) Name of the node

=item C<< case_number => SCALAR >>

 (Required) Case-number for this core

=item C<< message => SCALAR >>

 (Optional) User message for sending the ASUP

=back

=item Returns

None for success, throws an exception on failure.

=item Examples

# Save a coredump from the specified node:
NACL::C::Client::ONTAPSelectDeploy::Node::savecore(
    command_interface => $ci,
    cluster_name => "cluster1",
    node_name => "node001",
);

=back

=cut

sub savecore {
    my ($pkg, @args) = @_;
    $Log->enter() if $may_enter;
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            node_name		=> { type => SCALAR,},
            cluster_name	=> { type => SCALAR,},
            case_number		=> { type => SCALAR, optional => 1,},
            message		=> { type => SCALAR, optional => 1,},
            help                => { type => BOOLEAN, optional => 1 },
        },
        allow_extra	=> 1,
    );
    my $ret = $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => "_savecore",
                interface => "CLI",
                set       => "ONTAPSelectDeploy"
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $ret;
} ## end sub savecore

sub _savecore {
    my ( $pkg, @args ) = @_;
    $Log->enter() if $may_enter;
    $pkg->_base_backend_ontapselectdeploy_cli( @args,
        api => 'node_savecore' );
    $Log->exit() if $may_exit;
} ## end _savecore

=head start

Start the Node - This will power on the VM.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)

=item C<< node_name => SCALAR >>

 (Required) Name of the node to be powered on.

=item C<< cluster_name => SCALAR >>

 (Required) Name of the cluster.

=back

=item Returns

None for success, throws an exception on failure.

=item Examples

# Start the Node - This will power on the VM:
NACL::C::Client::ONTAPSelectDeploy::Node::start(
    command_interface => $ci,
    node_name => "node001",
    cluster_name => "cluster001",
);

=back

=cut

sub start {
    my ($pkg, @args) = @_;
    $Log->enter() if $may_enter;
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            node_name		=> { type => SCALAR,},
            cluster_name	=> { type => SCALAR,},
            help                => { type => BOOLEAN, optional => 1 },
        },
        allow_extra	=> 1,
    );
    my $ret = $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => "_start",
                interface => "CLI",
                set       => "ONTAPSelectDeploy"
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $ret;
} ## end sub start

sub _start {
    my ( $pkg, @args ) = @_;
    $Log->enter() if $may_enter;
    $pkg->_base_backend_ontapselectdeploy_cli( @args,
        api => 'node_start' );
    $Log->exit() if $may_exit;
} ## end sub _start

=head stop

Shutdown the node.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)

=item C<< node_name => SCALAR >>

 (Required) Name of the node to be powered on.

=item C<< cluster_name => SCALAR >>

 (Required) Name of the cluster.

=back

=item Returns

None for success, throws an exception on failure.

=item Examples

# NACL::C::Client::ONTAPSelectDeploy::Node::stop(
    command_interface => $ci,
    node_name => "node001",
    cluster_name => "cluster001",
);

=back

=cut

sub stop {
    my ($pkg, @args) = @_;
    $Log->enter() if $may_enter;
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            node_name		=> { type => SCALAR,},
            cluster_name	=> { type => SCALAR,},
            help                => { type => BOOLEAN, optional => 1 },
        },
        allow_extra	=> 1,
    );
    my $ret = $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => "_stop",
                interface => "CLI",
                set       => "ONTAPSelectDeploy"
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $ret;
} ## end sub stop

sub _stop {
    my ( $pkg, @args ) = @_;
    $Log->enter() if $may_enter;
    $pkg->_base_backend_ontapselectdeploy_cli( @args,
        api => 'node_stop' );
    $Log->exit() if $may_exit;
} ## end sub _stop

1;
