#
#
# Copyright (c) 2017 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#
#
## @summary Storage Component Module
## @author dl-nacl-dev@netapp.com
## @status shared
## @pod here
################################################################################

=pod

=head NAME

NACL::C::Client::ONTAPSelectDeploy::NodeStorage

=for html <hr>

=head DESCRIPTION

This library's purpose is to aid automation of workflows around the ONTAP Select Deploy Utility
by providing methods for Node Storage operations, such as node storage add.

=for html <hr>

=head SYNOPSIS

Addition of storage to a node in Cluster:
NACL::C::Client::ONTAPSelectDeploy::NodeStorage->add(
    command_interface => $ci,
    cluster_name => "cluster_name",
    node_name => "node_name",
    local_pool => "local_pool",
    username => "admin",
    password => "admin123",
    mirror_pool => "mirror_pool",
    capacity => "capacity",
);

=head ATTRIBUTES

=head cluster_name

Cluster Name of the Storage.

=cut

package NACL::C::Client::ONTAPSelectDeploy::NodeStorage;

use Exporter qw(import);
@EXPORT_OK = qw();

use strict;
use warnings;

use Options;
use Params::Validate qw(:all);
use NATE::BaseException;
use base 'NACL::C::Component::Client';
use NACL::ComponentUtils qw(_optional_scalars);

use Class::MethodMaker [
];

use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
###############################################################################
#           Node Storage Operation methods
###############################################################################

=head add

Addition of storage to a node in Cluster

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)

=item C<< cluster_name => SCALAR >>

 (Required) Name of the cluster

=item C<< node_name => SCALAR >>

 (Required) Name of the node

=item C<< local_pool => SCALAR >>

 (Required) Local storage pool to use to add disks to this node in the cluster

=item C<< username => SCALAR >>

 (Required) Username with access to this Storage

=item C<< password => SCALAR >>

 (Required) Password for the provided Username

=item C<< mirror_pool => SCALAR >>

 (Optional) Mirror node's storage pool. If this node is part of an HA pair, the storage pool for the mirror node is required

=item C<< capacity => SCALAR >>

 (Optional)  Capacity to add to the node (GB).  For an HA pair, the capacity is consumed in both the local-pool and the mirror-pool.If a capacity is not specified, the entire pool space will be consumed.

=back

=item Returns

None for success, throws an exception on failure.

=item Examples

#Addition of storage to a node in Cluster:
NACL::C::Client::ONTAPSelectDeploy::NodeStorage->add(
    command_interface => $ci,
    cluster_name => "cluster_name",
    node_name => "node_name",
    local_pool => "local_pool",
    username => "admin",
    password => "admin123",
    mirror_pool => "mirror_pool",
    capacity => "capacity",
);

=back

=cut

sub add {
    my ($pkg, @args) = @_;
    $Log->enter() if $may_enter;
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
	additional_spec => {
	    cluster_name	=> { type => SCALAR,},
            node_name           => { type => SCALAR,},
            local_pool          => { type => SCALAR,},
            username		=> { type => SCALAR,},
            password		=> { type => SCALAR,},
            _optional_scalars(
		qw( mirror_pool capacity )
	    ),
	    help		=> { type => BOOLEAN, optional => 1 },
	},
	allow_extra => 1,
    );
    my $ret = $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => "_add",
                interface => "CLI",
                set       => "ONTAPSelectDeploy"
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $ret;
} ## end sub add

sub _add {
    my ( $pkg, @args ) = @_;
    $Log->enter() if $may_enter;
    $pkg->_base_backend_ontapselectdeploy_cli( @args,
	api => 'node_storage_add' );
    $Log->exit() if $may_exit;
} ## end sub _add

1;
