#
# Copyright (c) 2013 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#
## @summary PV Component Module
## @author dl-nacl-dev@netapp.com, praveent@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::C::Client::PV

=head1 DESCRIPTION

This library is used to perform operations of creating and deleting physical/
logical volumes on the hosts.

=head1 METHODS

=head2 new

  my $obj = NACL::C::Client::PV->new(command_interface  => $ci,
                                       'pv_name'        => $device,);

Return a new PV component object with the given attributes.

=cut

package NACL::C::Client::PV;

use strict;
use warnings;
use NACL::CS::Client::PV;
use NACL::ComponentUtils qw(_optional_scalars );
use base 'NACL::C::Component::Client';
use Params::Validate qw(validate_with SCALAR);
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();

use Class::MethodMaker [ scalar => 'pv_name', ];

=head2 create()

 my $obj = NACL::C::Client::PV->create(command_interface  => $ci,
                                           'pv_name'      => $device,);

(Class method) Request creation of a PV (physical volume)  element, and  return a new
component object that refers to it.

Uses Linux-CLI APISet.

Invokes "pvcreate" command for Linux CLI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Client|lib-NACL-C-Component-pm/command_interface>

=item C<< 'pv_name' => $volumes >>

Applicable for Linux-CLI.

(Required): Name of the device which needs to be converted to physical-volumes.

=item C<< 'debug' => $boolean >>

Applicable for Linux-CLI.

(Optional): debug field.

=item C<< 'force' => $boolean >>

Applicable for Linux-CLI.

(Optional) force option.

=item C<< 'yes' => $boolean >>

Applicable for Linux-CLI.

(Optional) yes option.

=item C<< 'test' => $boolean >>

Applicable for Linux-CLI.

(Optional) test option.

=item C<< 'verbose' => $boolean >>

Applicable for Linux-CLI.

(Optional) verbose option.

=item C<< 'labelsector' => $boolean >>

Applicable for Linux-CLI.

(Optional) labelsector option.

=item C<< 'version' => $boolean >>

Applicable for Linux-CLI.

(Optional) verbose option.

=item C<< 'metadatatype' => $datatype >>

Applicable for Linux-CLI.

(Optional) metadatatype option.

=item C<< 'metadatacopies' => $datacopies >>

Applicable for Linux-CLI.

(Optional) metadatacopies option.

=item C<< 'dataalignment' => $alignment >>

Applicable for Linux-CLI.

(Optional) dataalignment option.

=item C<< 'restorefile' => $file >>

Applicable for Linux-CLI.

(Optional) restorefile option.

=item C<< 'setphysicalvolumesize' => $size >>

Applicable for Linux-CLI.

(Optional) setphysicalvolumesize option.

=item C<< 'uuid' => $uuid >>

Applicable for Linux-CLI.

(Optional) uuid option.

=item C<< 'version' => $version >>

Applicable for Linux-CLI.

(Optional) version option.

=item C<< 'zero' => $zero >>

Applicable for Linux-CLI.

(Optional) zero option.

=item C<< 'yes' => $boolean >>

Applicable for Linux-CLI.

(Optional) yes option.

=back

=cut

sub create {
    $Log->enter() if $may_enter;
    my ( $pkg, @args ) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw(
                  debug force yes test verbose labelsector metadatatype
                  metadatacopies metadatasize dataalignment restorefile
                  setphysicalvolumesize uuid version zero
                  )
            ),
        },
    );
    my $obj = $pkg->call_on_apiset(
        %opts,
        choices => [
            {
                method    => '_create_linux_cli',
                interface => 'CLI',
                set       => 'Linux',
            },
        ],
    );

    $Log->exit() if $may_exit;
    return $obj;
}    ##end sub create

sub _create_linux_cli {
    $Log->enter() if $may_enter;
    my ( $pkg, @args ) = @_;
    my %opts = $pkg->_validate_backend_opts(@args);

    my $command_interface = $opts{command_interface};
    my $apiset            = $opts{apiset};
    my %linux_args;
    $pkg->_hash_copy(
        source => \%opts,
        target => \%linux_args,
        copy   => [
            qw (
              debug force yes test verbose labelsector
              metadatatype metadatacopies metadatasize dataalignment
              restorefile setphysicalvolumesize uuid version zero
              )
        ],
        map => { pv_name => 'physical-volumes' },
    );
    $apiset->pvcreate(%linux_args);

    my $obj = $pkg->new(
        command_interface => $command_interface,
        'pv_name'         => $opts{'pv_name'},
    );
    $Log->exit() if $may_exit;
    return $obj;
} ## end sub _create_linux_cli

=head2 remove()

 NACL::C::Client::PV->remove(command_interface  => $ci,
                               'pv_name'        => $device,);

                                        or
 $pv_obj->remove(...);

(Class method or Instance method) Request deletion of a PV (physical volume)  element, and  deletes
component object that refers to it.

Uses Linux-CLI APISet.

Invokes "pvremove" command for Linux CLI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Client|lib-NACL-C-Component-pm/command_interface>

=item C<< 'pv_name' => $volumes >>

Applicable for Linux-CLI.

(Required): Name of the physical volume which needs to be removed.

=item C<< 'debug' => $boolean >>

Applicable for Linux-CLI.

(Optional): debug field.

=item C<< 'force' => $boolean >>

Applicable for Linux-CLI.

(Optional) force option.

=item C<< 'yes' => $boolean >>

Applicable for Linux-CLI.

(Optional) yes option.

=item C<< 'test' => $boolean >>

Applicable for Linux-CLI.

(Optional) test option.

=item C<< 'verbose' => $boolean >>

Applicable for Linux-CLI.

(Optional) verbose option.

=back

=cut

sub remove {
    $Log->enter() if $may_enter;
    my ( $pkg, @args ) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw(
                  debug force yes test verbose
                  )
            ),
        },
    );
    my $obj = $pkg->call_on_apiset(
        %opts,
        choices => [
            {
                method    => '_remove_linux_cli',
                interface => 'CLI',
                set       => 'Linux',
            },
        ],
    );

    $Log->exit() if $may_exit;
    return $obj;
}    ##end sub remove

sub _remove_linux_cli {
    $Log->enter() if $may_enter;
    my ( $pkg, @args ) = @_;
    my %opts = $pkg->_validate_backend_opts(@args);

    my $command_interface = $opts{command_interface};
    my $apiset            = $opts{apiset};
    my %linux_args;
    $pkg->_hash_copy(
        source => \%opts,
        target => \%linux_args,
        copy   => [
            qw (
              debug force yes test verbose
              )
        ],
        map => { pv_name => 'physical-volumes' },
    );
    $apiset->pvremove(%linux_args);
    $Log->exit() if $may_exit;
} ## end sub _remove_linux_cli

sub _primary_keys_validate_spec {
    return ( pv_name => { type => SCALAR }, );
} ## end sub _primary_keys_validate_spec

1;
