
#
# Copyright (c) 2001-2012 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Cluster Component Module
## @author dl-nacl-dev@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::C::Cluster

=head1 DESCRIPTION

C<NACL::C::Cluster> is a derived class of
L<NACL::C::CommandInterface|lib-NACL-C-CommandInterface-pm>.  It represents
access to a CMode cluster. When used as a component object, it can be used
to perform cluster related operations but when used as a command interface in
find() method call ,the existing design will not return a command interface
object which represents a cserver a.k.a admin vserver. Rather it returns
objects which represent node vservers or nodes in the clusters.See
L<NACL::C::Component::command_interface|lib-NACL-C-Component-pm/command_interface>

=head1 ATTRIBUTES

=head2 name

The name of the first node in the cluster

=head2 node

The name of the epsilon node in the cluster

=head2 hostrec

The hostrec describing the host/cluster which is modeled as CommandInterface

=cut

# A component to represent a cluster of nodes running Data ONTAP.

# The main trait that test scripts can see in a "cluster" that is more
# than just a "list of nodes" is that the cluster has a floating
# logical mangement interface, and so this object is mainly about
# representing that interface.  Like Node.

# However, it could/should perhaps also represent the cluster concept
# that one can manipulate by sending commands.  As such it may should
# be like NACL::C::Vserver where it both is a command interface
# (because one can send commands to it) and has a command interface
# (because one can send commands to that other command interface to
# manipulate the cluster as a whole, such as with the "cluster" family
# of CMode CLI commands)

package NACL::C::Cluster;

use strict;
use warnings;
use NACL::CS::Cluster;
use NACL::CS::ClusterPingCluster;
use Params::Validate qw(validate_with SCALAR ARRAYREF BOOLEAN);
use NACL::ComponentUtils qw(_optional_scalars);
use base 'NACL::C::CommandInterface::ONTAP';
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use NATE::Exceptions::Argument qw(:try);

# Get methods common to Cluster and Cserver
use NACL::C::_Mixins::ClusterCserver qw(:all);
use NACL::C::_Mixins::Common qw(get_7m_or_nodescope_apiset);

=head1 METHODS 

=head2 new

my $cluster = NACL::C::Cluster->new([name => $name],
                              [node => $node],
                              [hostrec => $hostrec] );
Specifying any one of the node, name, hostrec attributes will populate
the other two.
Returns a new Cluster component object with the given attributes.

=cut

use Class::MethodMaker [
    scalar => [
        {   '-default_ctor' => sub { $_[0]->_construct_primary_key(); },
            '-store_cb' =>
                sub { $_[0]->_verify_hostrec_callback($_[1], $_[2]); }
        },
        'node'
    ]
];

=head2 find

    my $Cluster  = NACL::C::Cluster->find(command_interface => $ci, ...);
    my @Clusters = NACL::C::Cluster->find(command_interface => $ci, ...); 

(Class method)
A constructor of sorts, which discovers which elements are present (by
interacting with the given command interface).  In an array context it
creates a component object for each present element and returns them.
In a scalar context it creates a component object for just one
element.

This method is generally not implemented in individual components. The base
class (L<NACL::C::Component::ONTAP|lib-NACL-C-Component-ONTAP-pm> provides an implementation.
See <L<NACL::C::Component::find|lib-NACL-C-Component-pm/find> for a more
detailed description along with explanation of all the options accepted.
You can also refer to user guide for more information on the options,
order of precedence for command line parameters (in case of find() is
invoked without specifying a command_interface) and examples.
http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/ComponentLayerUsersGuide/index.html#Obtaining_command_interface_objects

=head2 state

    my $Cluster_state = $cluster_obj->state();

(Instance method)
Fetch the current state of the element associated
with this component, and return a Component State object corresponding
to this component (calling this on NACL::C::Cluster would return 
a NACL::CS::Cluster object)

This method is generally not implemented in individual components. The base
class (L<NACL::C::Component::ONTAP|lib-NACL-C-Component-ONTAP-pm> provides an implementation.
See <L<NACL::C::Component::state|lib-NACL-C-Component-pm/state> for a more
detailed description along with explanation of all the options accepted.

=head2 modify_ha

    NACL::C::Cluster->modify_ha(command_interface=>$ci,
                                configured=>$boolean,...);
                                       
                   or
                   
   $cluster_obj->modify_ha(...);

(Class or instance method) This method modifies an ha element.

Uses a CMode CLI APIset.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< configured=>$scalar >>

(Required)
This parameter specifies the configuration of HA in a cluster pair.

Applicable for CMode CLI.

=item C<< "extended_query" => $string >>

(Optional)

Applicable for CMode CLI

=item C<< other_options >>

This method takes other common options described in L<Common method options|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#Other_options_valid_on_component_methods>

=item C<< '*' >>

This method takes all other options supported by
"cluster ha modify" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub modify_ha {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {_optional_scalars(qw(configured node)),},
        api             => 'cluster_ha_modify',
    );
    $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => "_modify_ha_cmode_cli",
                interface => "CLI",
                set       => "CMode",
            },
        ],
        frontend => 'NACL::C::Cluster::modify_ha',
    );
    $Log->exit() if $may_exit;
} ## end sub modify_ha

sub _modify_ha_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@opts, api => 'cluster_ha_modify',);
    $Log->exit() if $may_exit;
} ## end sub _modify_ha_cmode_cli

=head2 create

    my $cluster_obj = NACL::C::Cluster->create(command_interface=>$ci,...);

(Class method) This method creates a cluster element and returns a
component object representing it.

Uses a CMode CLI/ZAPI APISet.

Invokes "cluster-create" API for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< license=> $string >>

Applicable for CMode CLI/ZAPI

Maps to:

CM ZAPI : license

=item C<< clustername=> $string >>

Applicable for CMode CLI/ZAPI

Maps to:

CM ZAPI : cluster-name

=item C<< 'mgmt-port'=> $string >>

Applicable for CMode CLI.

=item C<< 'mgmt-ip'=> $string >>

Applicable for CMode CLI.

=item C<< 'mgmt-netmask'=> $string >>

Applicable for CMode CLI.

=item C<< 'mgmt-gateway'=> $string >>

Applicable for CMode CLI.

=item C<< ipaddr1=> $string >>

Applicable for CMode CLI.

=item C<< netmask=> $string >>

Applicable for CMode CLI.

=item C<< mtu=> $string >>

Applicable for CMode CLI.

=item C<< ipaddr2=> $string >>

Applicable for CMode CLI.

=item C<< "method-timeout"=> $scalar >>

(Optional)The new timeout value.

=item C<< "extended_query" => $string >>

(Optional)

Applicable for CMode CLI

=item C<< other_options >>

This method takes other common options described in L<Common method options|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#Other_options_valid_on_component_methods>

=item C<< '*' >>

This method takes all other options supported by
"cluster create" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub create {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw(ipaddr1 netmask mtu ipaddr2 mgmt-port
                    license clustername mgmt-ip mgmt-netmask
                    mgmt-gateway node)
            ),
        },
        api => 'cluster_create',
    );

    $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => "_create_cmode_cli",
                interface => "CLI",
                set       => "CMode",
            },
            {   method    => "_create_cmode_zapi",
                interface => "ZAPI",
                set       => "CMode",
            },
        ],
        frontend => 'NACL::C::Cluster::create',
    );
    my $obj = $pkg->new(
        node              => $opts{command_interface}->node(),
        command_interface => $opts{command_interface}
    );
    $Log->exit() if $may_exit;
    return $obj;
} ## end sub create

sub _create_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, @opts) = @_;
    $pkg->_base_backend_cmode_cli(@opts, api => 'cluster_create',);
    $Log->exit() if $may_exit;
} ## end sub _create_cmode_cli

sub _create_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg, %opts) = @_;

    $pkg->_base_backend_cmode_zapi(
        %opts,
        api             => 'cluster-create',
        copy            => [qw(license )],
        map             => {"clustername" => "cluster-name",},
        frontend_method => 'create'
    );
    $Log->exit() if $may_exit;
}

=head2 join

    my $cluster_obj = NACL::C::Cluster->join(command_interface=>$ci,...);


(Class method) This method joins a Node (as specified by the command_interface)
to a cluster element and returns a component object representing it.

Uses a CMode CLI/ZAPI APISet.

Invokes "cluster-join" API for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< clusteripaddr=> $string >>

Applicable for CMode CLI/ZAPI

Maps to:

CM ZAPI : cluster-ip-address

=item C<< ipaddr1=> $string >>

Applicable for CMode CLI.

=item C<< netmask=> $string >>

Applicable for CMode CLI.

=item C<< mtu=> $string >>

Applicable for CMode CLI.

=item C<< ipaddr2=> $string >>

Applicable for CMode CLI.

=item C<< "extended_query" => $string >>

(Optional)

Applicable for CMode CLI

=item C<< other_options >>

This method takes other common options described in L<Common method options|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#Other_options_valid_on_component_methods>

=item C<< 'cluster-name' => $string >>

Cluster Name of the Cluster to Join
Applicable for CMode CLI

=item C<< '*' >>

This method takes all other options supported by
"cluster join" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub join {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw(ipaddr1 netmask mtu ipaddr2 clusteripaddr node
                    cluster-name
                    )
            ),
        },
        api => 'cluster_join',
    );
    $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => "_join_cmode_cli",
                interface => "CLI",
                set       => "CMode",
            },
            {   method    => "_join_cmode_zapi",
                interface => "ZAPI",
                set       => "CMode",
            },
        ],
        frontend => 'NACL::C::Cluster::join',
    );
    my $obj = $pkg->new(
        node              => $opts{command_interface}->name(),
        command_interface => $opts{command_interface}
    );
    $Log->exit() if $may_exit;
    return $obj;
} ## end sub join

sub _join_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, @opts) = @_;
    $pkg->_base_backend_cmode_cli(@opts, api => 'cluster_join',);
    $Log->exit() if $may_exit;
} ## end sub _join_cmode_cli

sub _join_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg, %opts) = @_;

    $pkg->_base_backend_cmode_zapi(
        %opts,
        api             => 'cluster-join',
        map             => {"clusteripaddr" => "cluster-ip-address",},
        frontend_method => 'join'
    );
    $Log->exit() if $may_exit;
}

=head2 unjoin

    NACL::C::Cluster->unjoin(command_interface=>$ci,...);
    
                   or
                   
    $cluster_obj->unjoin(...);


(Class or instance method) This method unjoins a Node from a cluster.

Uses a CMode CLI/ZAPI APIset.

Invokes API 'cluster-unjoin' for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< node=> $string >>

(Required)

Applicable for CMode CLI/ZAPI.

=item C<< clustername=> $string >>

Applicable for CMode CLI.

=item C<< "extended_query" => $string >>

(Optional)

Applicable for CMode CLI

=item C<< other_options >>

This method takes other common options described in L<Common method options|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#Other_options_valid_on_component_methods>

=item C<< 'force' => $string >>

Force Node to Unjoin
Applicable for CMode CLI

=item C<< '*' >>

This method takes all other options supported by
"cluster unjoin" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub unjoin {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw(clustername
                    force
                    )
            ),
        },
        api => 'cluster_unjoin',
    );
    $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => "_unjoin_cmode_cli",
                interface => "CLI",
                set       => "CMode",
            },
            {   method    => "_unjoin_cmode_zapi",
                interface => "ZAPI",
                set       => "CMode",
            },
        ],
        frontend => 'NACL::C::Cluster::unjoin',
    );
    $Log->exit() if $may_exit;
} ## end sub unjoin

sub _unjoin_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@opts, api => 'cluster_unjoin',);
    $Log->exit() if $may_exit;
} ## end sub _unjoin_cmode_cli

sub _unjoin_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg, %opts) = @_;
    $pkg->_base_backend_cmode_zapi(
        %opts,
        api  => 'cluster-unjoin',
        copy => [qw(node)],
    );
    $Log->exit() if $may_exit;
} ## end _unjoin_cmode_zapi

=head2 is_member

    my $boolean = NACL::C::Cluster->is_member(
                                command_interface => $ci,
                                ...);
                                
                    or
                    
    my $boolean = $cluster_obj->is_member();

(Class or instance method)
This method is used to print whether the node is part of a cluster or not.
This method returns 1 if it is a member, 0 if it is not.

Uses a CMode CLI APISet.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'method-timeout' => $time_in_seconds >>

(Optional)
The default NATE timeout is 60 seconds. If the command might take more
time to complete, this option can be used to specify a larger timeout value.
The value should be provided in seconds.

=item C<< "extended_query" => $string >>

(Optional)

Applicable for CMode CLI

=item C<< other_options >>

This method takes other common options described in L<Common method options|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#Other_options_valid_on_component_methods>

=item C<< '*' >>

This method takes all other options supported by
"cluster is-member" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub is_member {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {_optional_scalars(qw(node))},
        cli_cmd         => 'cluster is-member',
    );
    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_is_member_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::Cluster::is_member',
    );

    $Log->exit() if $may_exit;
    return $ret;
} ## end sub is_member

sub _is_member_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    my $response;
    $pkg_or_obj->_base_backend_cmode_cli(
        @opts,
        api                 => 'cluster_is_member',
        apiset_response_obj => \$response
    );

    my $raw_output = $response->get_raw_output();
    if ($raw_output =~ /This node is a member of a cluster/i) {
        $Log->exit() if $may_exit;
        return 1;
    } else {
        $Log->exit() if $may_exit;
        return 0;
    }
} ## end sub _is_member_cmode_cli

=head2 modify

    NACL::C::Cluster->modify(
                                command_interface => $ci,
                                'node' => $node,
                                ...);
                                
                     or
                     
    $cluster_obj->modify(...);

(Class or instance method)
This method is used to modify cluster node membership attributes.

Uses a CMode CLI/ZAPI APIset.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'node' => $node >>

(Required)

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

<cluster-node-modify>
    <node-name>$value</node-name>
</cluster-node-modify>

=item C<< 'epsilon' => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

<cluster-node-modify>
    <is-node-epsilon>$value</is-node-epsilon>
</cluster-node-modify>

=item C<< 'eligibility' => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

<cluster-node-modify>
    <is-node-eligibility>$value</is-node-eligibility>
</cluster-node-modify>

=item C<< "extended_query" => $string >>

(Optional)

Applicable for CMode CLI

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub modify {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    $pkg_or_obj->_frontend_log_debug_opts(@_);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@_,
        additional_spec => {
            _optional_scalars(
                qw( epsilon eligibility node
                    )
            ),
        },
        cli_cmd => 'cluster modify',
    );
    $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_modify_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_modify_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
        ],
    );
    frontend => 'NACL::C::Cluster::modify',

        $Log->exit() if $may_exit;
} ## end sub modify

sub _modify_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@opts, api => 'cluster_modify',);
    $Log->exit() if $may_exit;
} ## end sub _modify_cmode_cli

sub _modify_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    $pkg_or_obj->_base_backend_cmode_zapi(@opts,
            api    => 'cluster_node_modify',
            map    => {
                'node'        => 'node-name',
                'epsilon'     => 'is-node-epsilon',
                'eligibility' => 'is-node-eligible'
            },
    );
    $Log->exit() if $may_exit;
} ## end sub _modify_cmode_zapi

=head2 ping_cluster

    my @ping_cluster_cs_objs = NACL::C::Cluster->ping_cluster(
                                command_interface => $ci,
                                'node' => $node,
                                ...);
                                
                              or
                              
    my $ping_cluster_cs_obj = NACL::C::Cluster->ping_cluster(
                                command_interface => $ci,
                                'node' => $node,
                                ...);
                                
                              or
                              
    my @ping_cluster_cs_objs = $cluster_obj->ping_cluster(...);
    
                              or
                              
    my $ping_cluster_cs_obj = $cluster_obj->ping_cluster(...);


(Class or instance method)
This method is used to ping remote cluster interfaces and perform RPC server
check and it returns the output of cluster ping cluster.

This method returns objects of type
L<NACL::CS::ClusterPingCluster|lib-NACL-CS-ClusterPingCluster-pm>.

Uses a CMode CLI APISet.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'node' => $node >>

(Required)

Applicable for CMode CLI.

=item C<< 'use-sitelist' => $string >>

Applicable for CMode CLI.

=item C<< 'abort-signal' => $string >>

Applicable for CMode CLI.

=item C<< 'rpc-prognum' => $string >>

Applicable for CMode CLI.

=item C<< 'skip-rpccheck' => $string >>

Applicable for CMode CLI.

=item C<< 'cmd-pid' => $string >>

Applicable for CMode CLI.

=item C<< 'method-timeout' => $time_in_seconds >>

(Optional)
The default NATE timeout is 60 seconds. If the command might take more
time to complete, this option can be used to specify a larger timeout value.
The value should be provided in seconds.

=item C<< "extended_query" => $string >>

(Optional)

Applicable for CMode CLI

=item C<< other_options >>

This method takes other common options described in L<Common method options|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#Other_options_valid_on_component_methods>

=item C<< '*' >>

This method takes all other options supported by
"cluster ping-cluster" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub ping_cluster {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    $pkg_or_obj->_frontend_log_debug_opts(@_);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@_,
        additional_spec => {
            _optional_scalars(
                qw( use-sitelist abort-signal rpc-prognum skip-rpccheck cmd-pid
                    node )
            ),
        },
        cli_cmd => 'cluster ping-cluster',
    );

    my @ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_ping_cluster_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::Cluster::ping_cluster',
    );

    $Log->exit() if $may_exit;
    return wantarray ? @ret : $ret[0];
} ## end sub ping_cluster

sub _ping_cluster_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    my $ci = $opts{command_interface};

    my $response;
    $pkg_or_obj->_base_backend_cmode_cli(
        %opts,
        api                 => 'cluster_ping_cluster',
        apiset_response_obj => \$response
    );
    my @state_objs;
    my $output = $response->get_parsed_output();
    my $obj = NACL::CS::ClusterPingCluster->new(command_interface => $ci);
    foreach my $row (@$output) {
        $obj->_set_fields(row => $row, is_zapi => 0);
        push @state_objs, $obj;
    }

    $Log->exit() if $may_exit;
    return @state_objs;
} ## end sub _ping_cluster_cmode_cli

=head2 setup

    NACL::C::Cluster->setup(
                                command_interface => $ci,
                                ...);

(Class method)
This method is used to setup wizard.

It is suggested that this is invoked with the APISet or the connection
object since this component method does not have support for all of the
various prompts that could show up as a part of the wizard.

Uses a CMode CLI APISet.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'cluster_mgmt_gateway' => $string >>

Applicable for CMode CLI.

=item C<< 'gateway' => $string >>

Applicable for CMode CLI.

=item C<< 'ip' => $string >>

Applicable for CMode CLI.

=item C<< 'cluster_mgmt_netmask' => $string >>

Applicable for CMode CLI.

=item C<< 'node_mgmt_gateway' => $string >>

Applicable for CMode CLI.

=item C<< 'node_mgmt_netmask' => $string >>

Applicable for CMode CLI.

=item C<< 'advanced' => $string >>

Applicable for CMode CLI.

=item C<< 'yes_no' => $string >>

Applicable for CMode CLI.

=item C<< 'text' => $string >>

Applicable for CMode CLI.

=item C<< 'cluster_mgmt_ip' => $string >>

Applicable for CMode CLI.

=item C<< 'netmask' => $string >>

Applicable for CMode CLI.

=item C<< 'mtu' => $string >>

Applicable for CMode CLI.

=item C<< 'node_mgmt_ip' => $string >>

Applicable for CMode CLI.

=item C<< 'name-servers' => $string >>

Applicable for CMode CLI.

=item C<< 'join' => $string >>

Applicable for CMode CLI.

=item C<< 'reuse' => $string >>

Applicable for CMode CLI.

=item C<< 'create-join' => $string >>

Applicable for CMode CLI.

=item C<< 'cluster-ports-entered' => $string >>

Applicable for CMode CLI.

=item C<< 'method-timeout' => $time_in_seconds >>

(Optional)
The default NATE timeout is 60 seconds. If the command might take more
time to complete, this option can be used to specify a larger timeout value.
The value should be provided in seconds.

=item C<< 'single-node-cluster' => $string >>

Applicable for CMode CLI.

=item C<< 'sp_gateway' => $string >>

Applicable for CMode CLI.

=item C<< 'sp_ip' => $string >>

Applicable for CMode CLI.

=item C<< 'sp_netmask' => $string >>

Applicable for CMode CLI.

=item C<< "extended_query" => $string >>

(Optional)

Applicable for CMode CLI

=item C<< other_options >>

This method takes other common options described in L<Common method options|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#Other_options_valid_on_component_methods>

=item C<< '*' >>

This method takes all other options supported by
"cluster setup" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub setup {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    $Log->warn('It is suggested that this is invoked with the APISet or '
            . 'the connection object since this component method does not '
            . 'have support for all of the various prompts that could show '
            . 'up as a part of the wizard.');

    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( cluster_mgmt_gateway gateway ip cluster_mgmt_netmask node_mgmt_gateway
                    node_mgmt_netmask advanced yes_no text cluster_mgmt_ip netmask mtu
                    node_mgmt_ip name-servers join reuse create-join cluster-ports-entered

                    single-node-cluster

                    sp_gateway sp_ip sp_netmask
                    )
            ),
        },
        ignore_primary_keys => 1,
        cli_cmd             => 'cluster_setup',
    );
    $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_setup_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::Cluster::setup',
    );

    $Log->exit() if $may_exit;
} ## end sub setup

sub _setup_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, @opts) = @_;
    $pkg->_base_backend_cmode_cli(@opts, api => 'cluster_setup',);
    $Log->exit() if $may_exit;
} ## end sub _setup_cmode_cli


__PACKAGE__->_load_event_class();

sub _primary_keys_validate_spec {
    return (node => {type => SCALAR});
}

# ...

# To populate the 'primary key' - 'node' in static calls to
# Cluster component . The method verifies if the attribute 'node'
# is defined in the hash argument passed. If not, then goes and fetches the
# attribute from its corresponding command_interface and populates
# in the same argument hash passed. Returns the hash containing the final
# arguments .

sub _populate_node_name_in_static {

    my $pkg_or_obj = shift;
    my %opts       = @_;

    if (!defined($opts{'node'})) {
        my $command_interface = $opts{'command_interface'}
            || $pkg_or_obj->command_interface();
        my $node = $command_interface->node();
        $opts{node} = $node;
    } ## end if ( !defined( $opts{'node'...
    return %opts;
} ## end sub _populate_node_name_in_static

=head2 nodes

  NACL::C::Cluster->nodes(command_interface=>$ci,...);

  or

  $clusterobj->nodes();


(Class or instance method) This method returns an array ,
containing the name of the nodes contained in the cluster

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=back

=back

=cut

sub nodes {

    my $pkg_or_obj = shift;
    my %opts       = @_;
    %opts = $pkg_or_obj->_common_validate_with(
        params              => \@_,
        additional_spec     => {node => {type => SCALAR, optional => 1}},
        ignore_primary_keys => 1
    );

    my $command_interface = delete $opts{command_interface}
        || $pkg_or_obj->command_interface();
    delete $opts{node};
    my $pkg = ref($pkg_or_obj) || $pkg_or_obj;
    my @state = $pkg->find(command_interface => $command_interface, %opts);
    my @list_of_nodes = map $_->node(), @state;
    return @list_of_nodes;
} ## end sub nodes

=head2 node_objs

  NACL::C::Cluster->node_objs(command_interface=>$ci,...);

  or

  $clusterobj->node_objs();


(Class or instance method) This method returns an array ,
containing Node Objects , for every node present in the
cluster. When its called by NACL::STask::Cluster or its 
instance then it returns objects of type NACL::STask::Node ,
 else the returned type is NACL::C::Node 

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=back

=back

=cut

sub node_objs {

    my $pkg_or_obj = shift;
    my %opts       = @_;

    %opts = $pkg_or_obj->_common_validate_with(
        params          => \@_,
        additional_spec => {node => {type => SCALAR, optional => 1}},
    );

    my $pkg_name      = 'NACL::' . $pkg_or_obj->is_c_or_t() . '::Node';
    my @list_of_nodes = $pkg_or_obj->nodes(%opts);
    my @node_objects  = map $pkg_name->new(name => $_), @list_of_nodes;
    return @node_objects;
} ## end sub node_objs

=head2 cluster_interface

    NACL::C::Cluster->cluster_interface();
This method iterates over each nodes interface until the Cluster Management
LIF is found (thereby overcoming a node that is not currently running).
On successfully retrieving the LIF details of the Cluster Management Interface
this method connects via the interface to determine if the LIF is reachable
and functioning.
This method returns an object of type NACL::C::Cserver representing the 
cluster management LIF
=cut

sub cluster_interface {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    ## Find all the Node Objects
    my @node_objs = NACL::C::Node->find();
    my $cserver_interface;

    ## Iterate through the Node objects attempting to find the cluster management LIF details
    foreach my $node_interface (@node_objs) {
        try {
	    $cserver_interface = NACL::C::Cserver->find(
                command_interface => $node_interface,
            );
        } catch NATE::BaseException with {
            my $exc_obj = shift;
            if ($exc_obj->text() !~ /could not create connection for/i) {
                $exc_obj->throw();
            }
        };
        last if defined $cserver_interface;
    }
    NATE::BaseException->throw("Failed to create a connection to any of the " .
        "nodes in the cluster. Unable to retrieve Cluster Management LIF " .
        "information.")
        if !(defined $cserver_interface);

    ## Validate the cluster management. LIF (sometimes a bogus IP address used for the cluster management LIF).
    try {
        my @all_root_aggrs =  NACL::C::StorageAggregate->find_root(
            command_interface => $cserver_interface,
        );
    } catch NATE::BaseException with {
        my $exc_obj = shift;
        if ($exc_obj->text() =~ /could not create connection for/i) {
            $Log->error("This library is designed to use the cluster management " .
                "LIF. A connection to this LIF is failing. See the following " .
                "Failure Stack Trace for connection details.");
        }
        $exc_obj->throw();
    };
    return $cserver_interface;
    $Log->exit() if $may_exit;
}
1;
