
#
# Copyright (c) 2001-2017 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary ClusterPeer Component Module
## @author dl-nacl-dev@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::C::ClusterPeer

=head1 DESCRIPTION

C<NACL::C::ClusterPeer> is a derived class of
L<NACL::C::Component::ONTAP|lib-NACL-C-Component-ONTAP-pm>. It represents access
to an ONTAP ClusterPeer.  A related class is
L<NACL::CS::ClusterPeer|lib-NACL-CS-ClusterPeer-pm>, which represents
the past state of an ONTAP ClusterPeer.

=head1 ATTRIBUTES

=head2 command_interface

See L<NACL::C::Component::command_interface|lib-NACL-C-Component-pm/command_interface>

=head2 cluster

Name of the cluster peer.

=cut

package NACL::C::ClusterPeer;

use strict;
use warnings;

use NACL::CS::ClusterPeer;
use NACL::ComponentUtils qw (_optional_scalars Dumper);

use base 'NACL::C::Component::ONTAP';
use Params::Validate qw(validate validate_with SCALAR ARRAYREF);
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use NACL::C::_Mixins::ClusterPeer qw(:all);
use NACL::C::Exceptions::ClusterPeer::AlreadyExists qw(:try);
use NACL::C::Exceptions::ClusterPeer::DoesNotExist;

=head1 METHODS

=head2 new

  my $ClusterPeer = NACL::C::ClusterPeer->new(command_interface=>$ci,
                                   cluster => $cluster,
                                  );

Return a new ClusterPeer component object with the given attributes.

=cut

use Class::MethodMaker [scalar => 'cluster', scalar => 'passphrase'];

=head2 find

    my $ClusterPeer_obj  = NACL::C::ClusterPeer->find(command_interface => $ci, ...);
    my @ClusterPeer_objs = NACL::C::ClusterPeer->find(command_interface => $ci, ...);

(Class method)
A constructor of sorts, which discovers which elements are present (by
interacting with the given command interface).  In an array context it
creates a component object for each present element and returns them.
In a scalar context it creates a component object for just one
element.

This method is generally not implemented in individual components. The base
class (L<NACL::C::Component::ONTAP|lib-NACL-C-Component-ONTAP-pm> provides an implementation.
See <L<NACL::C::Component::find|lib-NACL-C-Component-pm/find> for a more
detailed description along with explanation of all the options accepted.

Uses a CMode CLI or CMode ZAPI APISet.

=head2 state

    my $ClusterPeer_state = $ClusterPeer_obj->state();

(Instance method)
Fetch the current state of the element associated
with this component, and return a Component State object corresponding
to this component (calling this on NACL::C::ClusterPeer would return
a NACL::CS::ClusterPeer object)

This method is generally not implemented in individual components. The base
class (L<NACL::C::Component::ONTAP|lib-NACL-C-Component-ONTAP-pm> provides an implementation.
See <L<NACL::C::Component::state|lib-NACL-C-Component-pm/state> for a more
detailed description along with explanation of all the options accepted.

Uses a CMode CLI or CMode ZAPI APISet.

=cut

=head2 create

    my $ClusterPeer_obj    = NACL::C::ClusterPeer->create(
                                command_interface => $ci,
                                'peer-addrs' => [ $peer_addr1, $peer_addr2, ...],
                                 ...);

(Class method)
This method is used to create a new cluster peer relationship.
Uses a CMode CLI or CMode ZAPI APISet.
Invokes 'cluster-peer-create' API for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "local-profile" => $string >>

Applicable only for CMode CLI

=item C<< "password" => $string >>

The password required to authenticate with the peer cluster.

Applicable for CMode CLI/ZAPI
Maps to:
    CM ZAPI : password

=item C<< "admin-role" => $string >>

Applicable only for CMode CLI

=item C<< "timeout" => $string >>

The timeout for operations in the cluster peer. (Defaults to being 60 seconds
if not provided, accepted range is between 25 to 180 seconds)

Applicable for CMode CLI/ZAPI
Maps to:
    CM ZAPI : timeout

=item C<< "stable-addrs" => $string >>

Applicable only for CMode CLI

=item C<< "username" => $string >>

The username required to authenticate with the peer cluster.

Applicable for CMode CLI/ZAPI
Maps to:
    CM ZAPI : user-name

=item C<< "peer-addrs" => [ $peer-addr1, $peer-addr2, ... ] >>

(Required)
(Arrayref, even if only a single value is to be sent it should specified
as an arrayref containing that single value)

Applicable for CMode CLI/ZAPI
Maps to:
    CM ZAPI : peer-addresses

=item C<< 'extended_query' => $string >>

(Optional)
Applicable only for CMode CLI

=item C<< other_options >>

This method takes other common options described in L<Common method options|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#Other_options_valid_on_component_methods>

=item C<< 'address-family' => $string >>

Address Family of Relationship possible value(s) are, ipv4,ipv6,ipv6z

Applicable for CMode CLI/ZAPI
Maps to:
    CM ZAPI :  address-family

=item C<< 'offer-expiration' => $string >>

Passphrase Match Deadline
Applicable for CMode CLI/ZAPI

=item C<< 'no-authentication' => $string >>

Do Not Use Authentication
Applicable for CMode CLI

=item C<< 'passphrase' => $string >>

Passphrase for cluster peer authentication
Applicable for CMode CLI/ZAPI

=item C<< 'update-ping-timeout' => $string >>
Timeout for Update Pings (seconds) ranges from 1 to 10
Applicable for CMode CLI/ZAPI.

=item C<< 'rpc-connect-timeout' => $string >>
Timeout for RPC Connect (seconds) ranges from 5 to 30
Applicable for CMode CLI/ZAPI.

=item C<< 'ipspace' => $string >>

IPspace for the Relationship
Applicable for CMode CLI/ZAPI.

Mapping for ZAPI: ipspace-name

=item C<< '*' >>

This method takes all other options supported by
"cluster peer create" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<NACL::C::Exceptions::ClusterPeer::AlreadyExists>

This type of exception gets thrown if an attempt is made to create a
cluster peer relationship that already exists.

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub create {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            'peer-addrs'   => {type => ARRAYREF,          optional => 1},
            'stable-addrs' => {type => ARRAYREF | SCALAR, optional => 1},
            _optional_scalars(
                qw( local-profile password admin-role timeout username
                    address-family offer-expiration no-authentication
                    passphrase rpc-connect-timeout update-ping-timeout
                    ipspace local-name )
            ),
        },
        ignore_primary_keys => 1,
        cli_cmd             => 'cluster peer create',
    );

    # Print a warning if "stable-addrs" was provided as a sclar. Also, split
    # it into an array.
    $pkg->_scalar_to_array_deprication_warning(
        params  => ['stable-addrs'],
        options => \%opts
    );

    my $ret;
    try {
        $ret = $pkg->call_on_apiset(
            %opts,
            choices => [
                {   method    => '_create_cmode_cli',
                    interface => 'CLI',
                    set       => 'CMode',
                },
                {   method    => '_create_cmode_zapi',
                    interface => 'ZAPI',
                    set       => 'CMode',
                },
            ],
            frontend => 'NACL::C::ClusterPeer::create',
        );
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = $_[0];
        NACL::C::Exceptions::ClusterPeer::AlreadyExists
            ->detect_convert_and_throw(exception => $exception);
        $Log->exit() if $may_exit;
        $exception->throw();
    };

    my $filter = {};

    $pkg->_hash_copy(
        source => \%opts,
        target => $filter,
        copy   => [qw(stable-addrs peer-addrs)]
    );

    my $cs = NACL::CS::ClusterPeer->fetch(
        command_interface => $opts{command_interface},
        requested_fields  => [qw(cluster)],
        filter            => $filter
    );

    my %extra_opts;
    $extra_opts{'passphrase'} = $ret->passphrase() if ( ( defined $ret ) && ( defined $ret->passphrase() ) );

    my $obj = $pkg->new(
        command_interface => $opts{command_interface},
        cluster           => $cs->cluster(),
	%extra_opts,
    );
    $Log->exit() if $may_exit;
    return $obj;
} ## end sub create

sub _create_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, %opts) = @_;
    my $password          = delete $opts{password};
    my $passphrase        = delete $opts{passphrase};
    my @array_match_table = ('Remote Password:' => $password,
        qr/[Enter|Confirm]\s+the\s+passphrase/i => $passphrase);
    my %additional_args   = ('connectrec-match_table' => \@array_match_table);
    $opts{additional_args} = \%additional_args;

    my ($response, @state_objs);
    $pkg->_base_backend_cmode_cli(%opts, api => 'cluster_peer_create',
		    apiset_response_obj => \$response,);

    my $output = $response->get_parsed_output();
    my $obj = NACL::CS::ClusterPeer->new(
		command_interface => $opts{command_interface});
    foreach my $row (@$output) {
	my %values = %$row;
	while ( my ( $key, $val ) = each %values ) {
	    delete $values{$key} if(!defined($val));
	}
        $obj->_set_fields(row => \%values);
	push @state_objs, $obj;
    }

    $Log->exit() if $may_exit;
    return @state_objs;
} ## end sub _create_cmode_cli

sub _create_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg, %opts) = @_;
    if (defined $opts{'offer-expiration'}) {
        $opts{'offer-expiration'} = $pkg->_convert_to_unix_time(
            command_interface  => $opts{command_interface},
            'offer-expiration' => $opts{'offer-expiration'}
        );
    }
    $opts{timeout} ||= 60;
    my $copy = [
        qw(password timeout address-family passphrase
           offer-expiration rpc-connect-timeout 
           update-ping-timeout )
    ];
    my $map = {
        "peer-addrs" => "peer-addresses",
        "username"   => "user-name",
        "ipspace"    => "ipspace-name",
    };

    my ($response, @state_objs);
    $pkg->_base_backend_cmode_zapi(
        %opts,
        api    => 'cluster-peer-create',
        copy   => $copy,
        map    => $map,
        method => 'create',
	apiset_response_obj => \$response,
   );

    my $passphrase = $response->get_parsed_output()->[0]->{'passphrase-info'}->[0]->{'cluster-peer-passphrase-info'}->[0]->{'passphrase'};
    my $obj = NACL::CS::ClusterPeer->new(
		command_interface => $opts{command_interface});
    $obj->_set_fields(row => {'passphrase' => $passphrase});
    push @state_objs, $obj;

    $Log->exit() if $may_exit;
    return @state_objs;
} ## end sub _create_cmode_zapi

# XXX When ZAPI support is added we need to take care of the "timeout" parameter
# CLI defaults it to 60 seconds, ZAPI defaults it to 15 seconds. So our
# call will need to send the value as 60 if not provided.

=head2 modify

   NACL::C::ClusterPeer->modify(
                                command_interface => $ci,
                                 cluster => $cluster,
                                  ...);

                                 or

    $ClusterPeer_obj->modify(...);

(Class or instance method)
This method is used to modify cluster peer relationships.
Uses a CMode CLI or CMode ZAPI APISet.
Invokes 'cluster-peer-modify' API for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "cluster" => $cluster >>

(Required for class method, Not Applicable for instance method)
Name of the cluster peer.

Applicable for CMode CLI/ZAPI
Maps to:
    CM ZAPI : cluster-name

=item C<< "timeout" => $string >>

The timeout for operations in the cluster peer. (Defaults to being 60 seconds
if not provided, accepted range is between 25 to 180 seconds)

Applicable for CMode CLI/ZAPI
Maps to:
    CM ZAPI : timeout

=item C<< "stable-addrs" => $string >>

Applicable only for CMode CLI

=item C<< "peer-addrs" => [ $peer-addr1, $peer-addr2, ... ] >>

(Arrayref, even if only a single value is to be sent it should specified
as an arrayref containing that single value)

Applicable for CMode CLI/ZAPI
Maps to:
    CM ZAPI : peer-addresses

=item C<< 'extended_query' => $string >>

(Optional)
Applicable only for CMode CLI

=item C<< other_options >>

This method takes other common options described in L<Common method options|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#Other_options_valid_on_component_methods>

=item C<< 'address-family' => $string >>

Address Family of Relationship possible value(s) are, ipv4,ipv6,ipv6z
Applicable for CMode CLI

=item C<< 'auth-status-admin' => $string >>

Authentication Status Administrative
Applicable for CMode CLI/ZAPI

=item C<< 'offer-expiration' => $string >>

Passphrase Match Deadline
Applicable for CMode CLI/ZAPI


=item C<< 'passphrase' => $string >>

Passphrase for cluster peer authentication
Applicable for CMode CLI/ZAPI

=item C<< 'update-ping-timeout' => $string >>
Timeout for Update Pings (seconds) ranges from 1 to 10
Applicable for CMode CLI/ZAPI.

=item C<< 'rpc-connect-timeout' => $string >>
Timeout for RPC Connect (seconds) ranges from 5 to 30
Applicable for CMode CLI/ZAPI.

=item C<< 'ipspace' => $string >>

IPspace for the Relationship
Applicable for CMode CLI/ZAPI.

Mapping for ZAPI: ipspace-name

=item C<< '*' >>

This method takes all other options supported by
"cluster peer modify" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub modify {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            'peer-addrs'   => {type => ARRAYREF,          optional => 1},
            'stable-addrs' => {type => ARRAYREF | SCALAR, optional => 1},
            _optional_scalars(
                qw( timeout address-family auth-status-admin passphrase
                    offer-expiration update-ping-timeout rpc-connect-timeout
                    ipspace )
            ),
        },
        cli_cmd  => 'cluster peer modify',
    );
    if (defined $opts{'stable-addrs'}) {
        if (ref $opts{'stable-addrs'} eq "ARRAY") {
            $pkg_or_obj->_print_deprecated_warning_array(
                params  => ['stable-addrs'],
                options => \%opts
            );
        } ## end if ( ref $opts{'stable-addrs'...
    } ## end if ( defined $opts{'stable-addrs'...
    $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_modify_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_modify_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::ClusterPeer::modify',
    );

    $Log->exit() if $may_exit;
} ## end sub modify

sub _modify_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@opts, api => 'cluster_peer_modify',
    );
    $Log->exit() if $may_exit;
} ## end sub _modify_cmode_cli

sub _modify_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;

    if (defined $opts{'offer-expiration'}) {
        $opts{'offer-expiration'} = $pkg_or_obj->_convert_to_unix_time(
            command_interface  => $opts{command_interface},
            'offer-expiration' => $opts{'offer-expiration'}
        );
    }

    my $copy = [qw(timeout passphrase auth-status-admin offer-expiration
                   update-ping-timeout rpc-connect-timeout address-family )];
    my $map  = {
        "peer-addrs" => "peer-addresses",
        "cluster"    => "cluster-name",
        "ipspace"    => "ipspace-name",
    };
    $pkg_or_obj->_base_backend_cmode_zapi(
        %opts,
        api    => 'cluster-peer-modify',
        copy   => $copy,
        map    => $map,
        method => 'modify'
    );
    $Log->exit() if $may_exit;
} ## end sub _modify_cmode_zapi

=head2 delete

   NACL::C::ClusterPeer->delete(
                                command_interface => $ci,
                                 cluster => $cluster,
                                  );

                                 or

    $ClusterPeer_obj->delete();

(Class or instance method)
This method is used to delete a cluster peer relationship.
Uses a CMode CLI or CMode ZAPI APISet.
Invokes 'cluster-peer-delete' API for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "cluster" => $cluster >>

(Required for class method, Not Applicable for instance method)
Name of the cluster peer.

Applicable for CMode CLI/ZAPI
Maps to:
    CM ZAPI : cluster-name

=item C<< 'extended_query' => $string >>

(Optional)
Applicable only for CMode CLI

=item C<< other_options >>

This method takes other common options described in L<Common method options|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#Other_options_valid_on_component_methods>

=item C<< '*' >>

This method takes all other options supported by
"cluster peer delete" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<NACL::C::Exceptions::ClusterPeer::DoesNotExist>

This type of exception gets thrown if an attempt is made to delete a
cluster peer relationship that does not exist.

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub delete {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params => \@args,
        cli_cmd    => 'cluster peer delete',
    );

    try {
        $pkg_or_obj->call_on_apiset(
            %opts,
            choices => [
                {   method    => '_delete_cmode_cli',
                    interface => 'CLI',
                    set       => 'CMode',
                },
                {   method    => '_delete_cmode_zapi',
                    interface => 'ZAPI',
                    set       => 'CMode',
                },
            ],
            frontend => 'NACL::C::ClusterPeer::delete',
        );
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = $_[0];
        NACL::C::Exceptions::ClusterPeer::DoesNotExist
            ->detect_convert_and_throw(exception => $exception);
        $Log->exit() if $may_exit;
        $exception->throw();
    };

    $Log->exit() if $may_exit;
} ## end sub delete

sub _delete_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@opts, api => 'cluster_peer_delete',
    );
    $Log->exit() if $may_exit;
} ## end sub _delete_cmode_cli

sub _delete_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    $pkg_or_obj->_base_backend_cmode_zapi(
        %opts,
        api    => 'cluster-peer-delete',
        map    => {"cluster" => "cluster-name",},
        method => 'delete'
    );

    $Log->exit() if $may_exit;
} ## end sub _delete_cmode_zapi

=head2 ping

    my @cluster_peer_ping_states = NACL::C::ClusterPeer->ping(%opts);

                                 or

    my $cluster_peer_ping_state  = NACL::C::ClusterPeer->ping(%opts);

(Class method only)
This method returns the details of intercluster connectivity test. It returns
objects of the type L<NACL::CS::ClusterPeerPing|lib-NACL-CS-ClusterPeerPing-pm>.

This method accepts all the parameters accepted by L<fetch|lib-NACL-CS-ComponentState-pm/fetch>
(requested_fields, filter, allow_empty) and has all of the other properties of
fetch methods (callable in scalar or list context, throws a NoElementsFound
exception if no matching elements are found)

=cut

sub ping {
    $Log->enter() if $may_enter;
    my $pkg = shift;
    require NACL::CS::ClusterPeerPing;
    my @state_objs = NACL::CS::ClusterPeerPing->fetch(@_);
    $Log->exit() if $may_exit;
    return wantarray ? @state_objs : $state_objs[0];
} ## end sub ping

=head2 diagnose

    NACL::C::ClusterPeer->diagnose(
                                command_interface => $ci,
                                ...);
   
(Class method)
This method is used to performs cluster peer diagnosis as well as collects the cluster peer infrastructure's internal state..
Uses CMode CLI .

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'collect-data' => $string >>

Should we collect diagnostic data possible value(s) are, true,false
Applicable for CMode CLI

=item C<< 'node' => $string >>

Node possible value(s) are, <nodename>,local
Applicable for CMode CLI

=item C<< 'peer-cluster' => $string >>

Name of the Peer Cluster
Applicable for CMode CLI

=item C<< '*' >>

This method takes all other options supported by "cluster peer diagnose" command in ONTAP

=item C<< other_options >>

This method takes other common options described in L<Common method options|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#Other_options_valid_on_component_methods>

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub diagnose {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( collect-data node peer-cluster
                    )
            ),
        },
        ignore_primary_keys => 1,
        cli_cmd             =>  'cluster peer diagnose',
    );
    $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_diagnose_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::ClusterPeer::diagnose',
    );

    $Log->exit() if $may_exit;
}

sub _diagnose_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, @opts) = @_;
    $pkg->_base_backend_cmode_cli(@opts, api => 'cluster_peer_diagnose',);
    $Log->exit() if $may_exit;
}


__PACKAGE__->_load_event_class();

sub _primary_keys_validate_spec {
    return ('cluster' => {type => SCALAR},);
}

1;
