
#
# Copyright (c) 2001-2017 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Represents a machine with mgmt interface and IP address
## @author dl-nacl-dev@netapp.com
## @status shared
## @pod here

# Design: http://wikid.netapp.com/w/QA/projects/Libraries_Initiative/Project/Common_Infrastructure/Design_Spec/Components_Design_Spec#Command_Interfaces

=head1 NAME

NACL::C::CommandInterface

=head1 DESCRIPTION

This is an abstract base class representing a machine (virtual or
physical) with a management interface and IP address.

It contains representation of the proposed or admin-intended
configuration of that machine (by containing a Hostrec) and the
cached sessions used to send commands to it (by containing a cache
of APISets, indexed by what category/interface/set [*] were used to
choose that APISet).

CommandInterfaces are really just Components which also have a management
interface and IP address, so CommandInterface is subclassed off
L<NACL::C::Component|lib-NACL-C-Component-pm>.

It is possible to specify parameters that specify for a particular host
which interface and set are required or preferred.
L<_determine_rules_from_hostrec()|lib-NACL-C-CommandInterface-pm/_determine_rules_from_hostrec> describes
the format of these parameters.

[*] While most APISets are indexed by category/interface/set, some are
indexed by some additional, optional co-ordinates. See
L<apiset|lib-NACL-C-CommandInterface-pm/apiset> for details.

=head1 ATTRIBUTES

In addition to the L<Attributes of NACL::C::Component|lib-NACL-C-Component-pm/ATTRIBUTES>,
CommandInterfaces have the following attributes.

=head2 hostrec

The hostrec describing the host which is modeled as a CommandInterface

=head2 apiset_must

A hash which represents the set of rules (values of category/interface/set)
that must be abided by while determining which APISet(s) can be chosen.

=head2 apiset_should

A hash which represents the set of rules (category/interface/set) that are
awarded the highest preference while determining which APISet(s) can be chosen.

=head2 version_manager

Returns the version manager object for this command interface. If this
particular type of command interface does not have versioning implemented,
then it throws a C<NATE::BaseException>.

=head2 exceptions_to_retry

An array containing the list of exception types,on catching anyone the exception
command would be retried


=head2 error_texts_to_retry

An array containing the list of error texts, on matching anyone the error text
command would be retried

=head2 retries

Applicable only when exceptions_to_retry or error_texts_to_retry is set,
a scalar defaulting to 3 retries

=head2 retry_interval

Applicable only when exceptions_to_retry or error_texts_to_retry is set,
a scalar defaulting to 10 seconds

=head2 error_texts_to_ignore

An array containing the list of error texts. Ignore command failure/Do not throw exceptions
when these errors are reported.
error_texts_to_ignore parameter takes precedence over retry parameters.

=head2 exceptions_to_ignore

An array containing the list of exceptions. Ignore command failure for any of the exceptions in the list.
exceptions_to_ignore parameter takes precedence over retry parameters.

=cut

package NACL::C::CommandInterface;

use strict;
use warnings;
use NACL::Exceptions::NoElementsFound qw(:try);
use NACL::ComponentUtils qw(_dump_one _optional_scalars Dumper);
use autouse 'NACL::GeneralUtils' => qw(_parse_hosts);
use NACL::Realm;
use base 'NACL::C::Component';
use Tharn qw( host param);
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use NACL::C::_Version;
use NACL::ResourceRecords;
use Hostrec;

use NACL::APISet ();
use Params::Validate qw(validate validate_with validate_pos :types);
use NATE::BaseException        ();
use NATE::Exceptions::Argument ();
use Scalar::Util qw(blessed);
use Tie::IxHash;

# Turn Params::Validate failures into NATE::Exceptions::Argument
BEGIN {
    Params::Validate::validation_options(
        on_fail => sub {
            NATE::Exceptions::Argument->throw(shift);
        }
    );
}

use Class::MethodMaker [
    new => ['-init', 'new'],

    # This is to delay the construction of the hostrec till it is required
    # Note that if a field has a -default_ctor or a -default value then
    # isset() will always return true. To check whether the attribute
    # currently has a value, we'd need to directly access the hash element.
    # So instead of "if ($self->hostrec_isset())", the condition will be
    # "if ($self->{hostrec})"
    scalar => [
        {   '-default_ctor' => sub {
                my $hostrec = $_[0]->construct_hostrec();
                $_[0]->_check_if_valid_hostrec($hostrec);
                return $hostrec;
            },
            '-store_cb' => sub {
                my ($self, $hostrec) = @_;
                $self->_check_if_valid_hostrec($hostrec);
                return $hostrec;
            },
            '-type' => 'Hostrec',
        },
        'hostrec'
    ],

    # When picking an apiset to use with this command interface, which
    # category/interface/set must be chosen.  ({interface=>'CLI'}
    # means command line interface must be used)
    hash => 'apiset_must',

    # When picking an apiset to use with this command interface, which
    # category/interface/set should preferable be chosen.  ({set=>'7Mode'}
    # means command set '7Mode' is most desired)
    hash => 'apiset_should',

    hash => 'cdef_path',

    # Private
    # It is not documented in POD because it is a private attribute which is
    # not meant to be accessed externally.
    # A cache of APISets used by the CommandInterface. These are cached by
    # category/interface/set, though this will probably be expanded to include
    # fields like connid (for console access).
    hash => [
        {   -store_cb => sub {

                # Validate that everything passed is an APISet object
                my ($self, $apiset_ref) = @_;

                # The value passed to this store callback could be either
                # an array-reference or a hash-reference.
                # 1) $ci->_apisets(key1 => value1, key2 => value2)
                # Value will be { key1 => value1, key2 => value2 }
                # 2) $ci->_apisets_set(key3 => value3, key4 => value4);
                # Value will be [ value3, value 4 ]
                my @apisets;
                if (ref $apiset_ref eq 'HASH') {
                    @apisets = values %$apiset_ref;
                } else {
                    @apisets = @$apiset_ref;
                }
                foreach my $apiset (@apisets) {
                    if (blessed($apiset)
                        && $apiset->isa('NACL::APISet')) {
                        if ($self->timeout_isset()) {
                            $apiset->set_timeout(timeout => $self->timeout());
                        }
                        if ($self->timeout_scaling_factor_isset()) {
                            $apiset->timeout_scaling_factor(
                                $self->timeout_scaling_factor());
                        }
                    } else {
                        my $ref = ref($apiset) || 'scalar';
                        NATE::Exceptions::Argument->throw('Each of the '
                                . 'values passed to _apisets should be '
                                . 'APISet objects but one of them was of type '
                                . "$ref. Dumper:\n"
                                . Dumper($apiset));
                    }
                }
                return $apiset_ref;
                }
        },
        '_apisets',
    ],

    scalar => [
        {   '-default_ctor' => sub {
                my $self = $_[0];
                if ($self->can('_create_version_manager')) {

                    # Don't want it bailing with an ugly script error,
                    # so check if can(), else throw a BaseException detailing
                    # what happened.
                    return $self->_create_version_manager();
                } else {
                    NATE::BaseException->throw('_create_version_manager has '
                            . 'not been implemented on '
                            . ref($self)
                            . '. Hence version manager could not be obtained'
                    );
                }
            },
            '-type' => 'NACL::APISet::Version',
        },
        'version_manager',
    ],

    # If the command interface object doesn't have its command interface set,
    # then it defaults to using itself as its command interface.
    # Do NOT use -default_ctor to default to $self. This causes a circular
    # reference, which prevents garbage collection (hence preventing
    # cleaning up of its APISets and connections). See burt 905644.
    scalar => [
        {   '-read_cb' => sub {
                my ($self, $val) = @_;
                $val //= $self;
                return $val;
            },
            '-type'         => 'NACL::C::CommandInterface'
        },
        'command_interface'
    ],

    scalar => [
        {   -store_cb => sub {
                my ($self, $timeout) = @_;
                my @apisets = $self->_apisets_values();
                foreach my $apiset (@apisets) {
                    $apiset->set_timeout(timeout => $timeout);
                }
                return $timeout;
            },
        },
        'timeout',
    ],

    scalar => [
        {   -store_cb => sub {
                my ($self, $timeout_scaling_factor) = @_;
                my @apisets = $self->_apisets_values();
                foreach my $apiset (@apisets) {
                    $apiset->timeout_scaling_factor($timeout_scaling_factor);
                }
                return $timeout_scaling_factor;
            },
        },
        'timeout_scaling_factor',
    ],
    array  => 'exceptions_to_retry',
    array  => 'error_texts_to_retry',
    array  => 'error_texts_to_ignore',
    array  => 'exceptions_to_ignore',
    scalar => [
        {-default => 3},
        'retries',
    ],
    scalar => [
        {-default => 10},
        'retry_interval',
    ],

    # Assign a UUID to each object so that when Component objects are
    # de-serialized, we can prevent the same command_interface from being
    # de-serialized multiple times. See burt 915491.
    scalar => [
        '-static',
        {   '-default_ctor' => sub {require Data::UUID; Data::UUID->new()},
            '-type'         => 'Data::UUID',
        },
        '_uuid_obj'
    ],
    scalar => '_uuid',
];

no warnings 'redefine';

# Even when the command_interface attribute is not set, invoking
# ->command_interface() will return a value (i.e. $self). Unfortunately,
# the _isset() method installed by MethodMaker reads only -default or
# -default_ctor, not -read_cb, so it will (incorrectly) return 0. We
# redefine it here to always return 1.
sub command_interface_isset {1}
use warnings 'redefine';

=head1 METHODS

=head2 default_apiset_could

I<Default APISet could> is the ruleset which (perhaps arbitarily) indicates
which category, interface and set are to be used when the user doesn't
really care (an APISet must always have these three values)
This method returns a hashref describing the I<default apiset could>. Derived
classed must override the implementation provided by NACL::C::CommandInterface
(which simply returns an empty hashref).

=head2 default_apiset_should

This method returns the default I<apiset_should> ruleset. It makes sense to
override this if we can assign particular rules based on the type of host.

=head2 default_apiset_must

This method returns the default I<apiset_must> ruleset. Again, it makes sense to
override this if we can assign particular rules based on the type of host. For
example, all ONTAP command interfaces must be of category "Node".

=cut

sub default_apiset_could  { return {}; }
sub default_apiset_should { return {}; }
sub default_apiset_must   { return {}; }

# The validate spec for options accepted by find()
sub _find_validate_spec {
    $Log->enter() if $may_enter;
    my $pkg = shift;
    my %find_validate_spec;

    # Get NACL::C::Component's _find_validate_spec
    %find_validate_spec = %{$pkg->SUPER::_find_validate_spec()};

    # And now add/override what we need

    # Command interfaces could take a ResourceRecords object
    $find_validate_spec{resource_records}
        = {isa => 'NACL::ResourceRecords', optional => 1};

    # Same validation as for the base class, but optional here
    $find_validate_spec{command_interface}{optional} = 1;

    $Log->exit() if $may_exit;
    return {%find_validate_spec};
} ## end sub _find_validate_spec

=head2 find

    my $widget  = NACL::C::Widget->find(...);
    my @widgets = NACL::C::Widget->find(...);

    my $widget  = NACL::C::Widget->find(command_interface => $ci, ...);
    my @widgets = NACL::C::Widget->find(command_interface => $ci, ...);

A constructor of sorts, which finds all of the command interfaces of that type.
It can be invoked either with or without passing a command interface object,
both types of invocation can be done in scalar or list context.

When invoked without providing a command interface, it uses Records present in
Resource Records to construct component object.

When a command interface is provided, we interact with to determine all the
command interfaces of that type present. This is done by calling the relevant
component state's "fetch" method. The returned component state objects are
then used to construct component objects.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Usually Required) The NACL::C::CommandInterface that will be used to
discover the list of elements.  This will also become the value of the
C<command_interface> attribute of the component objects that C<find>
creates and returns, so the methods on those components will continue
to operate by communicating with this command interface.

This option is generally optional for components which are themselves
command interfaces; when not specified, the command interface
components are discovered by examining data passed in by the user
(parameters, host record objects) and they will effectively be
constructed with themselves set to be their own command_interface.

=item C<< filter=>{ attribute1=>$value1, attribute2=>$value2, ... } >>

(Optional) A hashref that limits the components returned to be only
those whose attributes have the given values (If possible, the
commands/requests sent will be optimized to limit discovery
similarly).  The attribute names should match the fields used in the
ComponentState object corresponding to this Component. This option
should not be specified when you are not specifying command_interface in
which case, you can directly invoke the constructore new() to get a
component or command interface object for the specific element you are
interested at.

=item C<< apiset_must=>$ruleset >>

(Optional) A hashref constraining which API (mode, interface, etc.) on
the given command_interface will be used.  See L<lib::NACL::ChooseAPISet|lib-NACL-ChooseAPISet-pm>

=item C<< apiset_should=>$ruleset >>

(Optional) A hashref suggesting which API (mode, interface, etc.) on
the given command_interface is preferred.  See L<lib::NACL::ChooseAPISet|lib-NACL-ChooseAPISet-pm>

=item C<< allow_empty=>$boolean >>

(Optional) If false (the default) then throw an exception if no
elements are found.  If true, then return undef or an empty list if no
elements are found.

=item C<< hosttype => $hosttype >>

(Optional) Not all hosts (vservers and vfilers, for example) can be expected to
have a host record file. In that case, for a hostrec to be created we need
to supply the 'hosttype' value (creation of a hostrec requires hostid and
hosttype, at a minimum). This is, however, just a hint and will not override
a host record entry for hosttype, if present.

=back

=back

=cut

sub find {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;

    $pkg_or_obj->_extract_from_obj_for_find(arr_ref => \@args);

    my $pkg = $pkg_or_obj->get_package_name();
    my %opts = validate @args, $pkg->_find_validate_spec();

    my @components;
    my @primary_keys = $pkg->get_primary_keys();

    if (scalar(@primary_keys) != 1) {
        NATE::BaseException->throw("Internal error: command interface "
                . "components must only have one "
                . "primary key");
    }

    my $primary_key = $primary_keys[0];

    if (!(defined $opts{command_interface})) {

       # When command_interface is not specified, then specifying 'filter'
       # option is not supported at the moment and the filter option is simply
       # ignored. This behavior will affect users when we add support for
       # specifying filter option even when command_interface is not specified
       # So we should throw an exception now informing users that they are
       # using it wrongly and ask them to use new() directly.
        if (scalar keys(%{$opts{filter}})) {
            NATE::Exceptions::Argument->throw(
                      "For $pkg->find(), specifying 'filter' option is not "
                    . "supported if 'command_interface' option is not specified. "
                    . "you should resort to using $pkg->new() instead.");
        }

        # No command_interface, so these components are invented from
        # input given in our environment (host record files and test
        # parameters, or however this NACL::ResourceRecords was
        # formed) and use themselves as their own command interface.

        my $record_method = $pkg->resource_record_method_name();

        $opts{resource_records} ||= NACL::Realm->realm()->resource_records();

        my @records = @{$opts{resource_records}->$record_method()};

        # ... imagine using those filter options here to remove
        # undesired Hostrecs here ...

        if (!@records && !$opts{allow_empty}) {
            $record_method =~ s/(.*)s$/$1/;
            NACL::Exceptions::NoElementsFound->throw(
                      "No matching $record_method(s) found\n"
                    . "  Filter: "
                    . _dump_one($opts{filter})
                    . "\n");
        } ## end if ( !@records && !$opts...)

        my @components;
        foreach my $record (@records) {

            # We haven't yet supported storing the hostspec, so we'll
            # fail if a resource record does not have a hostrec set in it.
            if (!$record->hostrec_isset() || ( (defined $record->{'hostrec'}) && (! %{$record->{'hostrec'}}) ) ) {
                my $error = $record->_error();
                $Log->exit() if $may_exit;

                # $error is a NATE::Result::Config object. This
                # stringifies to the message(s) it stores
                NATE::BaseException->throw('When calling find() without a '
                        . 'command interface all the hosts specified through '
                        . 'test parameters must have nate-host files defined, '
                        . "however the host '"
                        . $record->_name()
                        . "' does not have one. Error:\n$error\n");
            }

            push @components,
                $pkg->new(
                hostrec         => $record->hostrec(),
                $primary_key    => $record->hostrec()->hostname(),
                resource_record => $record,
                );
        }
        return wantarray ? @components : $components[0];
    } else {

        # Another command_interface is passed as a parameter.  So, ask
        # that command interface what objects of our kind exist (using
        # our corresponding kind of state object's fetch method).
        # This merely fetches the names; most of the command interface
        # information takes the form of a Hostrec, which is still
        # filled out by a host record file. However, not all of these objects
        # can be used as command interfaces (and

        # Determine CS package for that component
        # For example, NACL::C::Node -> NACL::CS::Node
        my $CS_pkg = $pkg->get_CS_package_name();

        my %hash;
        $pkg->_hash_move(
            source => \%opts,
            target => \%hash,
            move   => ['resource_records']
        );

        my @state_objs
            = $CS_pkg->fetch(%opts, requested_fields => [$primary_key]);

        foreach my $state_object (@state_objs) {
            my $component = $pkg->new(
                "$primary_key"    => $state_object->$primary_key(),
                command_interface => $opts{command_interface},
                %hash
            );

            push @components, $component;
        } ## end foreach my $state_object (@state_objs)
    } ## end else [ if ( !( defined $opts{...}))]
    return wantarray ? @components : $components[0];
} ## end sub find

=head2 resource_record_method_name

    sub resource_record_method_name {
        return "nodes";
    }

(Abstract, all derived classes should provide this)
Returns the method of Resource Records to be called to get the records
of that type.

=head2 apiset

    Example:1
    # All the options can be explicitly specified.
    my $req_apiset = $cmd_int->apiset(category => $cat, interface => $int, set => $set, ...);

    Example:2
    my $node = NACL::C::Node->find();
    ...Do some operations
    $node->apiset_must(interface => 'ZAPI');
    # This would return a ZAPI APISet based on the preferences set in the
    # command interface object
    my $req_apiset = $node->apiset();

    Example:3
    my $vserver = NACL::C::Vserver->find();
    ...Do some operations
    # There are no apiset_must or apiset_should preferences set.Hence the
    # following call would return Vserver CLI APISet.
    my $req_apiset = $vserver->apiset();

    Example:4
    my $client = NACL::C::Client->find();
    # Returns a Linux or Windows CLI APIset based on whether your client is a
    # Linux or Windows host. Thus this call will help you to write client
    # agnostic scripts
    my $req_apiset = $client->apiset();

    Example:5
    my $vserver = NACL::C::Vserver->find();
    # You can specify any of 'category, 'interface' and 'set'. For other
    # parameters, defaults would be automatically assigned as explained in the
    # previous examples. The following call would return a Vserver ZAPI
    # APISet.
    my $req_apiset = $vserver->apiset(interface => 'ZAPI');

Each CommandInterface has a cache of apisets used. When requested for an
apiset, it checks the cache to see whether the appropriate apiset is
present and returns that. If not, it creates a new one and returns it.
APISets are indexed by category, interface and set and optionally
connid, shell_type and context_definition host. When any of the category,
interface and set parameters are not specified, then the preferences set in
apiset_must or apiset_should on command interface object would be considered.
If there are no preferences set, then we will try to assign reasonable
defaults.
NACL::C::Node object representing a CMode filer would return CM CLI apiset,
NACL::C::Node object representing a 7Mode filer would return 7M CLI apiset,
NACL::C::Cluster object would return CM CLI apiset,
NACL::C::Vfiler object would return Vfiler CLI apiset,
NACL::C::Vserver object would return Vserver CLI apiset,
NACL::C::Client object representing a Linux client would return Linux CLI
apiset,
NACL::C::Client object representing a Windows client would return Windows CLI
apiset,
NACL::C::DFM::DFMClient object would return DFM CLI apiset

=over

=item Options

=over

=item C<< category >>

(Optional)
The APISet 'category'.

=item C<< interface >>

(Optional)
The APISet 'interface'

=item C<< set >>

(Optional)
The APISet 'set'

=item C<< connid >>

(Optional)
The "connection id" specifies the type of connection to be established
to the host by the APISet. This would be used to specify the APISet object
should establish a non-default connection type. Though this could be used
to specify any connection ID, it would probably be most useful for console
access for situations where console access is mandatory.

=item C<< * >>

Any other options which the L<APISet constructor |lib-NACL-APISet-pm/new> allows
can be passed to this method.

=item C<< cache_apiset >>

(Optional)
This is a boolean option (accepts 1/0) which tells whether the APISet object
created newly should be cached or not. By default, any new APISet object
created will be cached (defaults to '1'). (If C<return_new_object> is set, then
this gets overridden to 0)

If an APISet object exists in the cache with all the specified preferences
then this method will return the same, in which case,  a new APISet
object is not created. This option will have no effect in such cases.

=item C<< return_new_object => 0 | 1 >>

(Optional)
If specified as 1, creates and returns a new APISet object (does not
cache it in the object and does not read from the command interface's
cache of APISet objects). If this is specified, then C<cache_apiset>
is reset to 0.

This is useful for cases where we should not use the APISet object cached
in the command_interface.

=back

=item Returns

An APISet object of the appropriate type.

=back

=cut

sub apiset {
    $Log->enter() if $may_enter;
    my ($self, @args) = @_;

    my %validation_spec = NACL::APISet->get_validation_spec();

    # Add the spec for 'cache_apiset' option.
    $validation_spec{'cache_apiset'} = {type => BOOLEAN, default => 1};

    # Add the spec for 'return_new_object' option.
    $validation_spec{'return_new_object'} = {type => BOOLEAN, default => 0};

    $validation_spec{'check_apiset_already_loaded'} = {type => BOOLEAN, default => 0};
    $validation_spec{'_was_new_apiset_created'} = {type => SCALARREF, optional => 1};

    # While category, interface and set are mandatory for APISet->new,
    # for apiset() they are not because apiset() is able to fill in values
    # for these based on type of command interface and settings on it.
    my %validate_spec_without_cat_int_set;

    # Tie this so that the order in which we read from it remains the same
    # for all APISets
    tie %validate_spec_without_cat_int_set, 'Tie::IxHash';
    %validate_spec_without_cat_int_set = %validation_spec;
    foreach my $val (qw(category interface set check_apiset_already_loaded
        _was_new_apiset_created)) {
        $validation_spec{$val}{optional} = 1;
        delete $validate_spec_without_cat_int_set{$val};
    }

    # Unfortunately many of the options accepted by apiset() are allowed
    # to be undef, even though APISet->new() doesn't allow it. Let's
    # add type UNDEF here.
    foreach my $value (keys %validate_spec_without_cat_int_set) {
        if (defined $validation_spec{$value}{type}) {
            $validation_spec{$value}{type}
                = $validation_spec{$value}{type} | UNDEF;
        }
    }

    my %opts              = validate(@args, \%validation_spec);
    my $cache_apiset      = delete $opts{cache_apiset};
    my $return_new_object = delete $opts{return_new_object};
    my $was_new_apiset_created = delete $opts{_was_new_apiset_created};

    # If we want a new object, then we implicitly don't want it cached
    $cache_apiset = 0 if $return_new_object;
    my $default_apiset_must  = $self->default_apiset_must();
    my $default_apiset_could = $self->default_apiset_could();
    my %apiset_must          = $self->apiset_must();
    my %apiset_should        = $self->apiset_should();

    # apiset() should not get affected by the command line params set for
    # 'interface' and 'set'. See burt 822337. Hence deleting the fields from the
    # two hashes.
    delete $apiset_should{interface};
    delete $apiset_must{interface};
    delete $apiset_should{set};
    delete $apiset_must{set};
    my %default_apiset       = (
        %$default_apiset_could, %$default_apiset_must,

        # Considering the must|should preferences set in the command interface
        # object as well
        %apiset_should, %apiset_must
    );
    my %req_params = (category => 1, interface => 1, set => 1);
    foreach my $apiset_opt ($self->_apiset_keys()) {
        if (defined $default_apiset{$apiset_opt}) {
            $default_apiset{$apiset_opt} =~ s/\s*(.+?)\s*\|.+/$1/;
            $opts{$apiset_opt} = $opts{$apiset_opt}
                || $default_apiset{$apiset_opt};
        }
        if ($req_params{$apiset_opt} && (!defined $opts{$apiset_opt})) {
            $Log->exit() if $may_exit;
            NATE::BaseException->throw("Internal error: "
                    . "User hasn't specified a value for "
                    . "'$apiset_opt' option and the command interface is also "
                    . "not able to identify a default");
        }
    }

    if ($opts{interface} eq 'RANDOM') {

       # If a setting like FILER_APISET_SHOULD_INTERFACE=RANDOM was made
       # on the command-line, then the setting in the object would
       # store "RANDOM" as the value. A call made without specifying the
       # interface value like:
       #     $node->apiset();
       # would result in the interface value being "RANDOM". It would later go
       #  down to APISet->new() and request for an APISet with interface
       # "RANDOM" and would fail. We're handling that case here.
       #
       # There's surely no actual use-case for getting back a random
       # APISet object. (calls on an APISet object are APISet-aware: i.e.
       # they require the APISet to be of a particular type) For this
       # reason, let's not actually return back a random APISet object, but
       # return one of type CLI.
        $opts{interface} = 'CLI';
    }

    ######################################################################
    # Node-scope ZAPI APISets require context_definition to be specified
    # For CLI, in the absence of context_definition it defaults to running
    # on the "local" node.
    # So we set the context definition only for Nodescope ZAPI APISet.
    # We want APISet to default to "local" for CLI and this is to handle the
    # case where the host id is different from the actual host name and users
    # don't have a 'hostname' entry in hostspec
    #######################################################################
    if (   $opts{set} eq 'Nodescope'
        && $opts{interface} eq 'ZAPI'
        && !defined $opts{context_definition}) {
        $opts{context_definition} = {schema_hostobj => $self->hostrec()};
    }

    my $my_dumper = sub {
        local $Data::Dumper::Terse  = 1;
        local $Data::Dumper::Indent = 0;
        return Data::Dumper::Dumper(@_);
    };

    # If connection is passed then no need to pass connection
    # overrides. See burt847870 for more details
    my @conn_overrides = NACL::APISet->get_conn_override_fields();
    my $has_conn_override;
    if (defined $opts{connection}) {
        $has_conn_override = 1;
        foreach my $override (@conn_overrides) {
            delete $opts{$override};
        }
    } else {
        foreach my $override (@conn_overrides) {
            if (defined $opts{$override}) {
                $has_conn_override = 1;
                last;
            }
        }
    }

    my $key = join ',', $opts{category}, $opts{interface}, $opts{set};
    foreach my $field (keys %validate_spec_without_cat_int_set) {
        my $value_of_field = $opts{$field};
        if (defined $value_of_field) {
            if (ref $value_of_field) {
                $key .= ',' . $my_dumper->($value_of_field);
            } else {
                $key .= ",$value_of_field";
            }
        }
    }

    my $apiset;
    my $check_apiset_already_loaded = delete $opts{check_apiset_already_loaded};
    unless ($return_new_object) {

        # Check the cache only if we don't want a new object
        $apiset = $self->_apisets_index($key);
    }
    if ( $check_apiset_already_loaded ) {
       if ($apiset) {
           $Log->exit() if $may_exit;
           return 1;
       } else {
           $Log->exit() if $may_exit;
           return 0;
       }
    }

    if (!$apiset) {
        # If "version_info" is provided in %opts, use it in %version_hash
        my $version_info_opts = delete $opts{version_info} // {};
        my %version_hash = (version_info => $version_info_opts);
        my %cdef_path_info = $self->cdef_path();
        $Log->debug(
            sub { "This is the cdef hash " . _dump_one(\%cdef_path_info) });

        if (exists $cdef_path_info{$opts{interface}}{uc $opts{set}}) {

            if ($cdef_path_info{$opts{interface}}{uc $opts{set}} eq
                'BUILD_ROOT') {
                $version_hash{version_info}{'cdefs_from_build'} //= 1;
            } else {

                $version_hash{cdef_path}
                    //= $cdef_path_info{$opts{interface}}{uc $opts{set}};
            }
        } ## end if ( exists $cdef_path_info...)

        try {

            # Get the version manager from the command interface
            my $vm = $self->version_manager();
            $version_hash{version_info}->{version_manager} //= $vm;

            # For CMode-CLI, we can reuse the connection present in the version
            # manager instead of creating a new one. Check if CMode-CLI,
            # there are no connection overrides and hostobj not provided
            # (hostobj and connection are mutually exclusive).
            if (!$has_conn_override && $opts{set} eq 'CMode'
                && $opts{interface} eq 'CLI' && !exists $opts{hostobj}) {
                my $vm_conn = $vm->cli_connection();
                $opts{connection} = $vm_conn if (defined $vm_conn);
            }
        }
        otherwise {

            # Could not get the version manager, proceed without
            # passing the version manager in the call to APISet->new
        };

        # "hostobj" and "connection" are mututally exclusive for the
        # APISet constructor
        unless (exists $opts{connection}) {
            $opts{hostobj} = $self->hostrec();
        }

        $apiset = NACL::APISet->new(%opts, %version_hash);
        $$was_new_apiset_created = 1;
        if ($cache_apiset) {
            $self->_apisets_set($key => $apiset);
        }
    } ## end if ( !$apiset )

    $Log->exit() if $may_exit;
    return $apiset;
} ## end sub apiset

sub _apiset_keys {
    return keys %{NACL::APISet->get_validation_spec()};
}

=head2 _determine_rules_from_hostrec

(Used internally by CommandInterface.pm)
This method looks at ruleset entries made in the tharnhost file and
updates the settings of this command interface object accordingly.
Entries in the nate_host file must be of the form
APISET_(MUST|SHOULD)_<KEY> = <value>

=cut

sub _determine_rules_from_hostrec {
    $Log->enter() if $may_enter;
    my $self    = shift;
    my @opts    = validate_pos(@_, {type => SCALAR});
    my $ruleset = $opts[0];

   # Set the rules specified by entries made by the user in the nate_host file
   # Do this only if this can indeed be used as a command interface
    if ($self->can_use_as_command_interface()) {
        my $hostrec        = $self->hostrec();
        my $name_attribute = $self->name_attribute();
        my $name           = $self->$name_attribute();

        foreach my $key (NACL::ChooseAPISet::ruleset_keys()) {
            my $rule = uc("${ruleset}_${key}");
            if ($hostrec->{$rule}) {
                my $exists = "${ruleset}_exists";
                unless ($self->$exists($key)) {
                    my $set = "${ruleset}_set";
                    $set->$set($key => $hostrec->{$rule});
                    $set->$set(
                        "${key}_source" => "'$rule' in Hostrec '$name'");
                }
            }
        }
    }

    $Log->exit() if $may_exit;
} ## end sub _determine_rules_from_hostrec

# For some reason, the primary key 'node' was originally called 'name'.
# This is to maintain backwards compatibility.
sub name {
    $Log->enter() if $may_enter;
    my $self           = shift;
    my $name_attribute = $self->name_attribute();
    if (scalar(@_)) {
        $self->$name_attribute(@_);

        # This should also clear the hostrec field so that if user tries to
        # fetch hostrec object, then the hostrec object corresponding to the
        # new name set will be returned.
        $self->hostrec_reset();
    } else {
        $Log->exit() if $may_exit;
        return $self->$name_attribute();
    }
    $Log->exit() if $may_exit;
} ## end sub name

sub name_attribute {
    $Log->enter() if $may_enter;
    my $self     = shift;
    my %pk_spec  = $self->_primary_keys_validate_spec();
    my @all_keys = keys %pk_spec;
    $Log->exit() if $may_exit;
    return $all_keys[0];
} ## end sub name_attribute

sub _verify_hostrec_callback {
    $Log->enter() if $may_enter;
    my $self = shift;
    my ($value, $attribute)
        = validate_pos(@_, {type => SCALAR}, {type => SCALAR});

    my @verify_name = _parse_hosts($value);
    $Log->debug(sub { "Parsed Hostspec value " . _dump_one({@verify_name}) });

    if (@verify_name > 2) {
        $Log->warn("The argument '$attribute' takes only the name of "
                . 'a single host. Instead, multiple hosts were specified. '
                . 'Only the first host will be used in this call. '
                . "(Value provided: $value)");
    }

    if ($verify_name[0] ne $verify_name[1]) {
        my $pkg_name = $self->get_package_name();
        $Log->warn("The argument '$attribute' (value: $value) should be the "
                . 'name of the host but instead a hostspec was provided as '
                . "the value. It is suggested that the 'hostrec' argument be "
                . "used for this. Example call:\n\t"
                . "$pkg_name->new(hostrec => Tharn::host(\$value));");

        my $host = host($value);
        my $id   = $host->id();
        $Log->exit() if $may_exit;
        return $id;
    }

    $Log->exit() if $may_exit;
    return $value;
} ## end sub _verify_hostrec_callback

=head2 init

An instance method, called internally by "new", to check that any
required attributes have been passed to "new". The base class,
NACL::C::CommandInterface, provides a generic implementation of this.

=cut

sub init {
    $Log->enter() if $may_enter;
    my $self        = shift;
    my $primary_key = $self->name_attribute();
    my %opts        = validate_with(
        params => \@_,
        spec   => {

            # These are the options which are accepted by new() but
            # do not directly get set as attributes of the object
            resource_record => {isa => 'NACL::ResourceRecord', optional => 1},
            resource_records => {type => ARRAYREF, default => []},
            _optional_scalars('name', $primary_key)
        },

        # Allow through the MethodMaker defined fields as options
        allow_extra => 1
    );

    my $resource_record  = delete $opts{resource_record};
    my $resource_records = delete $opts{resource_records};
    my $name             = delete $opts{name};

    # Set the values in the object (since "-hash" is not present in
    # the declaration of "new", the arguments passed into "new" do not
    # get automatically set)
    while (my ($key, $value) = each %opts) {
        $self->$key($value);
    }

    if ($opts{$primary_key} && $name) {
        $Log->exit() if $may_exit;
        NATE::Exceptions::Argument->throw("Either of '$primary_key' or "
                . "'name' should be provided in the call to new, but not both"
        );
    }

    unless ($opts{$primary_key} || $opts{hostrec} || $name) {
        $Log->exit() if $may_exit;
        NATE::Exceptions::Argument->throw("Any one of the three parameters "
                . "(name, hostrec, $primary_key) should be specified in the call"
                . " to new()");
    }

    # Set the value passed in 'name' to the specific primary key
    # (node, vfiler, vserver) for hostrec default constructor to work
    $self->$primary_key($name) if ($name);

    $self->_uuid($self->_uuid_obj()->create_str());

    # my $update_set_scoped_cdef_path = sub {
        # my ($apiset_must, $uc_prefix, $ci_to_test_param_name,
            # $check_cdef_path)
            # = @_;
        # my %all_cdef_paths;

        # $Log->debug(
            # sub {
                # "This is attribute cdef path set in Resource Record for $uc_prefix "
                    # . Dumper($check_cdef_path);
            # }
        # );
        # my @all_interfaces = NACL::ResourceRecords::_all_interfaces();
        # foreach my $ci_interface (@all_interfaces) {
            # foreach my $ci_set (@$apiset_must) {
                # my $uc_set = uc $ci_set;

                # First verify and set the PREFIX_INTERFACE_SET_CDEF_PATH
                # my $cdef_path
                    # = "${uc_prefix}_${ci_interface}_${uc_set}_CDEF_PATH";

                # my $cdef_param = param($cdef_path);

                # if ($cdef_param) {
                    # $all_cdef_paths{$ci_interface}{$uc_set} = $cdef_param;
                    # $all_cdef_paths{$ci_interface}{$uc_set . '_source'}
                        # = "Test parameter $cdef_path";
                # }

          # Verify and set the {ci_to_test_param_name}_INTERFACE_SET_CDEF_PATH
                # else {
                    # if ($uc_prefix ne $ci_to_test_param_name) {
                        # $cdef_path
                            # = "${ci_to_test_param_name}_${ci_interface}_${uc_set}_CDEF_PATH";

                        # $cdef_param = param($cdef_path);
                        # if ($cdef_param) {
                            # $all_cdef_paths{$ci_interface}{$uc_set}
                                # = $cdef_param;
                            # $all_cdef_paths{$ci_interface}
                                # {$uc_set . '_source'}
                                # = "Test parameter $cdef_path";
                        # }

            # Copy/expand the Interface only scoped cdef paths to
            # Interface and Set Scoped
            # Since this loop is hit at the end, if there are specific paths
            # provided for the given INTERFACE and SET they are stored already
                        # elsif (defined $check_cdef_path->{$ci_interface}) {
                            # my $value = $check_cdef_path->{$ci_interface};
                            # $all_cdef_paths{$ci_interface}{$uc_set} = $value;
                            # $all_cdef_paths{$ci_interface}
                                # {$uc_set . '_source'}
                                # = $check_cdef_path->{$ci_interface
                                    # . '_source'};
                        # }
                    # }    #($uc_prefix ne $ci_to_test_param_name )

                # }    #else
            # }    ## foreach $ci_set
        # }    ## foreach $ci_interface
        # $Log->debug(
            # sub {
                # "This is cdef path hash computed in init for $uc_prefix "
                    # . Dumper(\%all_cdef_paths);
            # }
        # );
        # return %all_cdef_paths;

    # };

    my $realm_resource_records = NACL::Realm->realm()->resource_records();

    # Provided a resource record, update the command interface object with
    # settings contained in the resource record.
    my $update_from_resource_record = sub {
        my ($resource_record)
            = validate_pos(@_, {isa => 'NACL::ResourceRecord'});
        foreach my $setting (
            NACL::ResourceRecord->_settings_read_from_parameters()) {
            my $isset = "${setting}_isset";
            if ($resource_record->$isset()) {
                if (!$self->$isset()) {
                    $self->$setting($resource_record->$setting());
                }
            }
        }

        # Code to handle cdef_paths set based on the SET values
        # my ($default_apiset_must_set_list, @apiset_must_set,
        # $ci_to_test_param_name,        $check_cdef_path,
        # $prefix,                       $uc_prefix,
        # %all_cdef_paths
        # );

      # $default_apiset_must_set_list = $self->default_apiset_must()->{'set'};
      # @apiset_must_set       = split( /\|/, $default_apiset_must_set_list );
      # $ci_to_test_param_name = $self->_ci_to_test_param_mapping();

        # The attribute _cdef_calculation_performed in ResourceRecords
        # is set to 1 when the 'SET' scoped CDEF_PATHs have been
        # processed and updated in cdef_path attribute of commandinterface
        # and ResourceRecords
        # if (  !$resource_record->_cdef_calculation_performed_isset()
        # || $resource_record->_cdef_calculation_performed() ne 1 )
        # {

        # $check_cdef_path = $self->cdef_path();
        # $prefix          = $resource_record->prefix();
        # $uc_prefix       = uc $prefix;
        # %all_cdef_paths  = $update_set_scoped_cdef_path->(
        # \@apiset_must_set,      $uc_prefix,
        # $ci_to_test_param_name, $check_cdef_path
        # );

        # $self->cdef_path(%all_cdef_paths);
        # $resource_record->cdef_path(%all_cdef_paths);
        # $resource_record->_cdef_calculation_performed('1');
        # }

        # my $generic_field
        # = 'generic_' . lc($ci_to_test_param_name) . '_resource_record';
        # my $generic_resource_record
        # = $realm_resource_records->$generic_field();
        # if (  !$generic_resource_record->_cdef_calculation_performed_isset()
        # || $generic_resource_record->_cdef_calculation_performed() ne 1 )
        # {

        # $prefix          = $generic_resource_record->prefix();
        # $uc_prefix       = uc $prefix;
        # $check_cdef_path = $generic_resource_record->cdef_path();
        # %all_cdef_paths  = $update_set_scoped_cdef_path->(
        # \@apiset_must_set,      $uc_prefix,
        # $ci_to_test_param_name, $check_cdef_path
        # );

        # $generic_resource_record->cdef_path(%all_cdef_paths);
        # $generic_resource_record->_cdef_calculation_performed(1);

        # }
    };

    my $self_name = $self->$primary_key();

    # Given an array of resource record objects, search for a resource
    # record for this particular host, and update the object with settings
    # of  the resource record
    my $search_from_list_and_update = sub {
        my @resource_records = @_;
        foreach my $record (@resource_records) {
            if (   $record->hostrec_isset()
                && $record->hostrec()->hostname() eq $self_name) {
                $update_from_resource_record->($record);
                last;
            }
        }
    };

    if ($resource_record) {
        $update_from_resource_record->($resource_record);
    } elsif (@$resource_records) {
        $search_from_list_and_update->(@$resource_records);
    } else {

        # Command interface objects constructed by calling find() without any
        # parameters will result in the test parameters being read. Command
        # interface objects constructed any other way (calling new() directly,
        # by calling create(), by calling find() with parameters) do not go
        # through the ResourceRecords code that reads the test parameters.
        # This code is to make it such that all command interface objects
        # (regardless of how they're obtained) read the test parameters.
        # The "generic" test parameters are looked at and those settings are
        # applied. To do so, we get the globally cached singleton resource
        # record and apply settings from it. We also check for a resource
        # record for this specific host.
        # (Example: for nodes looks in
        # $resource_records->nodes() for a resource record for this host)

        # Check if a resource record exists for this specific host
        my $record_method    = $self->resource_record_method_name();
        my @records          = @{$realm_resource_records->$record_method()};
        $search_from_list_and_update->(@records);

        # Use settings from the generic parameters for a particular type
        my $param_name    = $self->_ci_to_test_param_mapping();
        my $generic_field = 'generic_' . lc($param_name) . '_resource_record';
        my $resource_record = $realm_resource_records->$generic_field();
        $update_from_resource_record->($resource_record);

        # Use settings from the all_command_interface resource record
        my $all_res_rec
            = $realm_resource_records->all_command_interface_resource_record();
        $update_from_resource_record->($all_res_rec);
    }

    $Log->exit() if $may_exit;
} ## end sub init

=head2 refresh_command_interface

    $command_interface->refresh_command_interface([no_version_change => 0|1]);

An instance method, called to make the command_interface usable if the
underlying connection gets lost.

The connection to the filer could get lost as a result of:

=over

=item * reboot

=item * upgrade/revert

In this case not only would the connection have gotten lost but the OS version
being run on the host would have also changed.

=item * passing the command_interface to a subtest

Certain persistent connections might become unusable in a subtest (such as
c-login, since the socket connection established is inherently unsharable
across processes)

=item * disconnect being explicitly called on the connection object

=back

=over

=item Options

=back

=over

=item C<< no_version_change => 0|1 >>

(BOOLEAN, optional, defaults to 0 if not provided)
This is an optional parameter through which the user can specify whether the
OS version on the host has changed or not. The behavior of this method differs
based on the value of this parameter.

If the version has changed, then not only would the connection need to be
re-established but also the Schema needs to be updated to point to the
appropriate command definition files. This would require all of the APISet
objects currently cached by the command interface to be deleted. This would
mean that successive calls with the command interface would need to re-create
the appropriate APISet object(s).

However, if the version has not changed then only the persistent CLI
connection (if any) needs to be re-established. The cached APISet objects need
not be deleted.

Note that by default the version is assumed to have changed.

=item max_reconnect

(Optional) The number of retries on connection failure

=item max_reconnect_timewait

(Optional)
The number of seconds to sleep between each retry when max_reconnect is set

=back

=cut

sub refresh_command_interface {
    $Log->enter() if $may_enter;
    my ($command_interface, @args) = @_;
    my %opts = validate(
        @args,
        {   no_version_change      => {type => BOOLEAN, default  => 0},
            max_reconnect          => {type => SCALAR,  optional => 1},
            max_reconnect_timewait => {type => SCALAR,  optional => 1},
        }
    );
    my $no_version_change = delete $opts{no_version_change};
    if ($no_version_change) {

        # Version has not changed. Search for a cached CLI APISet and
        # reconnect its session.
        my $apisets = $command_interface->_apisets();
        while (my ($key, $apiset) = each %$apisets) {
            if ($apiset->{interface} eq 'CLI') {
                $Log->debug("Found '$key' APISet. Reconnecting its session");
                $apiset->api_reconnect_session(%opts);
            }
        }
    } else {

        # Version has changed, hence the currently cached APISet objects
        # should be deleted
        $Log->debug('Deleting the existing cached APISet objects and '
                . 'version manager');
        $command_interface->_apisets_reset();
        $command_interface->version_manager_reset();
    }

    $Log->exit() if $may_exit;
} ## end sub refresh_command_interface

=head2 get_version_manager

B<DEPRECATED>.

$command_interface->version_manager() should be used instead.

=cut

sub get_version_manager {
    $Log->enter() if $may_enter;
    my $self        = shift;
    my $version_mgr = $self->version_manager();
    $Log->exit() if $may_exit;
    return $version_mgr;
} ## end sub get_version_manager

# ...

sub apiset_allowed {
    my $self         = shift;
    my @ruleset_keys = NACL::ChooseAPISet::ruleset_keys();
    my (%opts)       = validate @_,
        {map { $_ => {type => SCALAR, optional => 1} } @ruleset_keys,};
    return NACL::ChooseAPISet::ruleset_allowed(rulesets =>
            [\%opts, $self->apiset_must(), $self->default_apiset_must(),],);
} ## end sub apiset_allowed

# Constructs a hostrec within the command_interface object when required
sub construct_hostrec {
    $Log->enter() if $may_enter;
    my $self           = shift;
    my $name_attribute = $self->name_attribute();
    my $name           = $self->$name_attribute();
    my $host;
    try {
        $host = host($name);
    }
    otherwise {
        my $config_error = shift;
        $Log->exit() if $may_exit;
        $self->_no_hostrec_error(
            name         => $name,
            config_error => $config_error
        );
    };
    $Log->exit() if $may_exit;
    return $host;
} ## end sub construct_hostrec

sub _no_hostrec_error {
    $Log->enter() if $may_enter;
    my $self = shift;
    my %opts = validate_with(
        params => \@_,
        spec   => {
            name         => {type => SCALAR},
            config_error => {isa  => 'NATE::Result::Config'},
            throw_exception => {type => SCALAR, default => 1}
        }
    );
    my $name            = $opts{name};
    my $config_error    = $opts{config_error};
    my $throw_exception = $opts{throw_exception};

    # If the call to Tharn::host() fails, a NATE::Result::Config object
    # gets thrown, not an exception.
    # The stringified form is the error message, which is what we use
    # in our exception text
    my $error_msg
        = 'A hostrec object could not be constructed for host '
        . "'$name'. This could be because '$name' does "
        . "not have a nate_host file defined. Error:\n"
        . $config_error;

    if ($throw_exception) {
        $Log->exit() if $may_exit;
        NATE::BaseException->throw($error_msg);
    } else {
        $Log->exit() if $may_exit;
        return $error_msg;
    }
} ## end sub _no_hostrec_error

sub _construct_primary_key {
    $Log->enter() if $may_enter;
    my $self = shift;

    # If node/vserver/vfiler is not already populated and if we are hitting
    # this default constructor, then it means that the object is created by
    # specifying a hostrec in the constructor. So we don't have to check
    # hostrec field is already populated or not
    my $name = $self->hostrec()->hostname();
    $Log->exit() if $may_exit;
    return $name;
} ## end sub _construct_primary_key

=head2 can_use_as_command_interface

    # Here CmdIntFoo is any command interface: Node, Vfiler, Vserver, Cluster
    my $cmd_int_foo = NACL::C::CmdIntFoo->find(command_interface => $ci);
    if ($foo->can_use_as_command_interface()) {
        NACL::C::Foo->some_method(command_interface => $cmd_int_foo, ...);
        ...
    }

(For component users)
Not all objects of type Node/Vfiler/Vserver/Cluster (all the command interfaces)
can be used as command interfaces for other components. This method checks
whether an object of any of these types can be used as command interface for
other components (it does this by checking if a hostrec can be created for the
host)

=cut

sub can_use_as_command_interface {
    $Log->enter() if $may_enter;
    my $self    = shift;
    my $can_use = 1;
    try {
        $self->hostrec();
    }
    catch NATE::BaseException with {
        $can_use = 0;
    };
    $Log->exit() if $may_exit;
    return $can_use;
} ## end sub can_use_as_command_interface

use Storable qw(freeze thaw);

sub STORABLE_freeze {
    my ($self, $cloning) = @_;
    my $data = {%$self};
    delete $data->{_apisets};
    return ('', $data);
} ## end sub STORABLE_freeze

sub STORABLE_thaw {
    my ($self, @opts) = @_;

    # Must reuse the thaw logic in Component.pm that ensures contained
    # command_interface objects are cached
    $self->SUPER::STORABLE_thaw(@opts);
}

=head2 is_feature_supported

    my $is_true = $self->is_feature_supported( feature => $feature_name );

(Instance method only)
This method is used to determine whether a particular feature is supported
or not. For the list of features which are currently supported, see
L<Supported Features|lib-NACL-C-_Version-pm/FEATURES>. Returns boolean value

    1 : If the specified feature is supported,
    0 : If the specified feature is not supported

=cut

sub is_feature_supported {
    $Log->enter() if $may_enter;
    my ($self, %opts) = @_;
    $self->_verify_invocation(style => 'instance_only');
    $Log->exit() if $may_exit;
    return NACL::C::_Version->is_feature_supported(
        command_interface => $self,
        %opts
    );
} ## end sub is_feature_supported

=head2 scale_timeout

    my $scaled_timeout = $ci_obj->scale_timeout('method-timeout' => $time);

Scales the C<method-timeout> value by the scaling factor value set in the
command interface object.

=over

=item Options

=over

=item C<< 'method-timeout' => $timeout >>

The timeout value to be scaled.

=back

=back

=cut

sub scale_timeout {
    $Log->enter() if $may_enter;
    my $self = shift;
    my %opts = validate_with(
        params => \@_,
        spec   => {'method-timeout' => {type => SCALAR, default => 60}},
    );
    my $scaled_timeout = $opts{'method-timeout'};

    if ($self->timeout_scaling_factor_isset()) {
        $scaled_timeout *= $self->timeout_scaling_factor();
    }

    $Log->exit() if $may_exit;
    return $scaled_timeout;
} ## end sub scale_timeout

=head2 clone_with_new_hostrec

    # Example here demonstrates how to run on a different lif for a given
    # vserver. Note, however, that this method is not limited to the Vserver
    # command interface or to IP addresses, but works for all command
    # interfaces and for any hostrec field.
    my $vserver_lif1 = NACL::C::Vserver->find();

    # Find a volume on the vserver
    my $vol_lif1 = NACL::C::Volume->find(command_interface => $vserver_lif1);
    ...

    my $vserver_lif2 = $vserver_lif1->clone_with_new_hostrec(default_ip => $new_ip);

    # Find a volume on the vserver, but this time the command/ZAPI is sent
    # through the new IP address provided
    my $vol_lif1 = NACL::C::Volume->find(command_interface => $vserver_lif2);

This method can be used to dynamically alter host-record details of a
command interface. Examples of instances where this method can be used are:

=over

=item *

For switching to using a different LIF for a particular host
(i.e. run on a different IP address)

=item *

To use a different username/password combination for the host

=back

This method can be used to alter any of the fields accepted by the Hostrec
constructor. (See
http://web.netapp.com/engineering/nate/releases/current/html/Hostrec.html#new)

Note that this method returns a B<new> command interface object with the
new host record. This is because merely modifying an existing hostrec object
does not result in the values stored in the Connection objects and APISet
objects being changed. Further, since hostrec objects are cached globally,
modifying the values of an existing hostrec can have unintended consequences
in a different part of the script. For this reason, rather than modifying
the existing hostrec object, a new one is created and returned in a new
command interface object.

=over

=item Options

The options accepted by C<< Hostrec->new >> are accepted by this method.
See http://web.netapp.com/engineering/nate/releases/current/html/Hostrec.html#new.

=back

=cut

sub clone_with_new_hostrec {
    $Log->enter() if $may_enter;
    my ($self, %opts) = @_;
    my $old_hostrec = $self->hostrec();
    my $new_hostrec = Hostrec->new(
        reuse => 0,
        save  => 0,
        id    => $old_hostrec->id(),

        # hosttype and default_ip could have been dynamically
        # determined, so NATE won't know about it. Hence we need to
        # copy these values over to the new hostrec. Of course, any
        # values supplied by the caller will override these
        hosttype   => $old_hostrec->hosttype(),
        default_ip => $old_hostrec->default_ip(),
        %opts
    );

    my $new_ci = thaw(freeze($self));
    $new_ci->hostrec($new_hostrec);
    $Log->exit() if $may_exit;
    return $new_ci;
} ## end sub clone_with_new_hostrec

=head2 default_ip

    if ($command_interface->default_ip()) {
        # Do something ...
    }

Must be called as an instance method.

This is a utility method that returns default IP of current command interface.

=cut

sub default_ip {
    $Log->enter() if $may_enter;
    my ($self) = @_;

    my $default_ip = $self->hostrec()->default_ip();

    $Log->exit() if $may_exit;
    return $default_ip;
}

# Sub-classes are expected to implement this. This method should check
# whether the hostrec supplied is valid for that type of command
# interface (for example, a hostrec without "filer" in the hosttype
# is invalid for all ONTAP command interfaces)
sub _check_if_valid_hostrec { }

# The "type" of an object is simply the last part of the package name.
# NACL::C::Node -> Node; NACL::STask::Node -> Node; NACL::C::Cluster -> Cluster,
# and so on.
sub _get_type {
    $Log->enter() if $may_enter;
    my ($self) = $_[0];
    my $type = $self->get_package_name();
    $type =~ s/.*:://;
    $Log->exit() if $may_exit;
    return $type;
}

=head2 type

    my $ci_type = $command_interface->type;

Must be called as an instance method.

Returns the "type" of the interface. Currently identifies the following as
distinct interface types:

    ava        AltaVault appliance
    eseries    ONTAP E-Series appliance
    ontap      Generic ONTAP filer (virtual or otherwise)

Takes no arguments. If the host-type of this interface is not known, a warning
will be issued, but no exception thrown. In this case, the return string will
be C<(unknown)>.

=cut

sub type {
    $Log->enter() if $may_enter;
    my ($self) = @_;
    my $hosttype = $self->hostrec()->hosttype();
    my $type;

    if ($hosttype =~ /^ava/) {
        $type = 'ava';
    } elsif ($hosttype =~ /^filer-eseries/) {
        $type = 'eseries';
    } elsif ($hosttype =~ /^filer/) {
        $type = 'ontap';
    } else {
        $Log->warn(
            "Could not determine interface type from hosttype of '$hosttype'." .
            ' Update NACL::C::CommandInterface::find() to properly identify ' .
            'this interface type.'
        );
        $type = '(unknown)';
    }

    $Log->exit() if $may_exit;
    return $type;
}

=head2 is_ava

    if ($command_interface->is_ava()) {
        # Do something AVA-specific...
    }

Must be called as an instance method.

This is a boolean predicate that tells whether the calling interface object is
representing an AltaVault appliance or not. If the B<type> method returns
C<ava> for this object, then the return value of this method will be true.

=cut

sub is_ava {
    $Log->enter() if $may_enter;
    my ($self) = @_;

    my $is_ava = $self->type() eq 'ava';

    $Log->exit() if $may_exit;
    return $is_ava;
}

=head2 is_ontap

    if ($command_interface->is_ontap()) {
        # Do something ONTAP-filer-specific...
    }

Must be called as an instance method.

This is a boolean predicate that tells whether the calling interface object is
representing an ONTAP filer or not. If the B<type> method returns C<ontap> for
this object, then the return value of this method will be true. (Note that this
will not return true for an E-Series appliance.)

=cut

sub is_ontap {
    $Log->enter() if $may_enter;
    my ($self) = @_;

    my $is_ontap = $self->type() eq 'ontap';

    $Log->exit() if $may_exit;
    return $is_ontap;
}

=head2 is_eseries

    if ($command_interface->is_eseries()) {
        # Do something ESeries-specific...
    }

Must be called as an instance method.

This is a boolean predicate that tells whether the calling interface object is
representing an ONTAP E-Series filer or not. If the B<type> method returns
C<eseries> for this object, then the return value of this method will be true.
(Note that this will not return true for a non-E-Series ONTAP interface.)

=cut

sub is_eseries {
    $Log->enter() if $may_enter;
    my ($self) = @_;

    my $is_eseries = $self->type() eq 'eseries';

    $Log->exit() if $may_exit;
    return $is_eseries;
}

1;
