#
# Copyright (c) 2001-2016 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Represents an ONTAP system
## @author dl-nacl-dev@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::C::Component::ONTAP

=head1 SYNOPSIS

    use base 'NACL::C::Component::ONTAP';

=head1 DESCRIPTION

NACL::C::Component::ONTAP is an abstract base class for all
ONTAP Component libraries (Node, Cluster, Vserver, Vfiler).
It is derived from L<NACL::C::Component|lib-NACL-C-Component-pm>
and it provides ONTAP specific component base class methods.
The implementation and documentation of provided functionality
is present in L<NACL::C::Component|lib-NACL-C-Component-pm>.

=head1 ATTRIBUTES

Same as those of L<NACL::C::Component|lib-NACL-C-Component-pm/ATTRIBUTES>.
(command interfaces are constrained to be sub-classes of
C<NACL::C::CommandInterface::ONTAP>.)

=cut

package NACL::C::Component::ONTAP;

use strict;
use warnings;
use feature 'state';
use Date::Parse;
use NACL::Realm;
use base 'NACL::C::Component';

BEGIN {
    Params::Validate::validation_options(
        on_fail => sub {
            NATE::Exceptions::Argument->throw(shift);
        }
    );
}

use NACL::CS::ComponentState::ONTAP;
use NACL::InstantComponent qw(_generate_method _generate_subcommand_ds);
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Params::Validate qw(UNDEF validate_with :types );
use NACL::ComponentUtils qw( _ontap_ci_validate_spec _optional_scalars
                             convert_to_unix_time enumerate_month Dumper);
use autouse 'NACL::GeneralUtils' => qw(coredump_on_timeout);
use NATE::Exceptions::Argument qw(:try);
use NACL::C::Exceptions::MultipleCommandFailed ();
use NACL::Exceptions::Timeout                  ();
use NACL::Exceptions::JobFailed                ();
use NACL::Exceptions::VerifyFailure;
use NACL::Exceptions::InvalidChoice;
use NACL::Exceptions::InvalidMethod;
use NACL::C::Job;
use NACL::C::JobPrivate;
use NACL::CS::JobHistory;
use autouse 'NACL::Tools::CGT::CGTUtils' => qw(parse_leaf_level_cmd);
use autouse 'NACL::Tools::CGT::CgtXml'   => qw(_get_help_xml_path parse_xmlfile);
use Scalar::Util qw(blessed);
use NATE::Time qw(timeout2time);
use feature 'state';
use List::MoreUtils qw( uniq );
use NACL::C::_Mixins::ComponentCommandInterfaceONTAP qw(:all);
use NACL::EventService qw(load_branch_events load_leaf_events);


use Class::MethodMaker [
    scalar => [
        { -type => 'NACL::C::CommandInterface::ONTAP' },
        'command_interface'
    ],
];

#
# Overriden can , which calls SUPER::can first, if it doesnt
# return a coderef ,it would walk the UI ,generates the
# method and returns a coderef
# Inorder to walk the UI  it needs command interface, if
# its not passed it would return undef
#
#
sub can {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my @opts       = $pkg_or_obj->_can_validation(@_);
    my @copy_opts  = @opts;
    my $pkg_name = $pkg_or_obj->get_package_name();
    my $method     = shift @copy_opts;
    my $error_ptr  = shift @copy_opts;
    my ( $code, $coderef, %opts_for_instant_c, $ci );

    # @opts is generally a hash of values, but it could just be a single
    # value (possible for accessor methods like $obj->field($val)). Assign
    # into a hash only if it's a hash of values (i.e. even number of values)
    %opts_for_instant_c = @opts if (scalar (@opts) % 2 == 0);

    if ( ref $pkg_or_obj && $pkg_or_obj->command_interface_isset() ) {
        $ci = $pkg_or_obj->command_interface();
    } else {
        $ci = $opts_for_instant_c{command_interface};
    }

    if ( $code = $pkg_or_obj->SUPER::can(@opts) ) {
        $Log->exit() if $may_exit;
        return $code;
    } elsif ( $method eq "show" ) {
        my $temp_pkg_name = $pkg_name;
        $temp_pkg_name =~ s/::C::/::CS::/g;
        $Log->warn(  $pkg_name."->show"." is invalid,"
                   . "call ".$temp_pkg_name."->fetch"." instead");
    } elsif ( $ci && $ci->is_cmode() ) {
        # If this is an alias component, then find out the real component's
        # name, and use that to construct the method.
        my $real_component = $pkg_name->_real_component();
        $pkg_name = $real_component if $real_component;
        my $component_ds;
        try{
            $component_ds = _generate_subcommand_ds(
                pkg_name          => $pkg_name,
                method_name       => $method,
                command_interface => $ci,
            );
        } catch NATE::BaseException with {
            $component_ds = _generate_subcommand_ds(
                pkg_name          => $pkg_name,
                method_name       => $method,
                command_interface => $ci,
                cdefs_from_local  => 1,
            );
        };
        if ( !$component_ds ) {
            $Log->exit() if $may_exit;
            $$error_ptr .= "\n Unable to find the command or cdef";
            return undef;
        }

        my $code = _generate_method(
            comp_method_hashref   => $component_ds,
            pkg_name              => $pkg_name,
            method                => $method,
            error_ptr             => $error_ptr,
            generate_method_alone => 1,
        );
        $code =~ s/require.*?;//g;
        $code =~ s/sub\s+.*?\{/sub \{/g;
        $coderef
            = eval "use NACL::ComponentUtils qw(_optional_scalars); $code";
        if ($@) {
            $Log->debug( "Could not eval the generated code: \n" . "$code"
                    . "Errors : \n"
                    . $@ );
            $$error_ptr
                .= "\n Could not eval the generated $method method Errors: $@";
            $coderef = undef;
        }
        $Log->exit() if $may_exit;
        return $coderef;
    } else {
        $Log->exit() if $may_exit;
        return undef;
    }
}

=head2 wait_on_job

    # Recommended syntax
    $job_obj->wait_on_job(
        'method-timeout'  => $time,
        [polling_interval => $interval, ]
        [apiset_must      => \%must_ruleset, ]
        [apiset_should    => \%should_ruleset ]
    );

    # $pkg_or_obj is a Component/STask package name/object.
    # The syntax below is not suggested, though it's still supported
    $pkg_or_obj->wait_on_job(
        job_component     => $job_component,
        'method-timeout'  => $time,
        [polling_interval => $interval, ]
        [apiset_must      => \%must_ruleset, ]
        [apiset_should    => \%should_ruleset ]
    );

(For component users)
This method, which can be called as a static or instance method, waits
for a specified amount of time for a job to complete. If the job is not found,
then it is assumed to have completed, so the job history table is searched
for a job of this ID. If it is found that the job has completed but has failed,
then a C<NACL::Exceptions::JobFailed> will be thrown. If a non-existent job id
(which isn't recognized by ONTAP as a valid id) is provided, then
C<NACL::Exceptions::NoElementsFound> would be thrown.

It is suggested that L<wait_for_completion|lib-NACL-C-Component-pm/wait_for_completion>
be used instead of this wherever possible since that provides a mechanism to
poll an attribute if no job was provided.

=over

=item Options

=over

=item C<< job_component => $job_component >>

(Required, isa C<NACL::C::Job>)
A job component modeling the job started for the operation.

Note that if this method is invoked with a Job Component object, then
this argument is not necessary.

=item C<< 'method-timeout' => $time_in_seconds >>

(Required)
The time, in seconds, for which to wait for the job to complete.

=item C<< polling_interval => $interval >>

(Optional, defaults to 10 seconds)
The interval at which the progress of the job should be shown. (This maps
to the "-interval" parameter of "job watch-progress")

=item C<< apiset_must => $must_ruleset >>

(Optional)

=item C<< apiset_should => $should_ruleset >>

(Optional)

=back

=back

=cut

sub wait_on_job {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            job_component => { type => OBJECT },
            $pkg_or_obj->_common_wait_for_validate_spec(),
            _optional_scalars( $pkg_or_obj->get_primary_keys() ),
            command_interface => { optional => 1 }
        }
    );
    my $job      = delete $opts{job_component};
    my $interval = delete $opts{polling_interval};

    if (!$job->isa('NACL::C::Job') && !$job->isa('NACL::C::JobPrivate')) {
        $Log->exit() if $may_exit;
        NATE::Exceptions::Argument->throw("'job_component' should be an "
            . "object of type 'NACL::C::Job' or 'NACL::C::JobPrivate' "
            . "but instead what was provided was of type '"
            . ref ($job)
            . "'"
        );
    }

    # For now, we don't expect users to pass nacl_retry_count and so we ignore
    # the nacl_retry_count added into %opts by _common_validate_with. The
    # individual calls made such as fetch, watch_progress will get its own
    # default later.
    delete $opts{nacl_retry_count};

    # At this point, %opts will have apiset_must, apiset_should along with
    # the primary keys. Move out the primary keys.
    $pkg_or_obj->_move_primary_keys( source => \%opts, target => {} );

    my $id = $job->id();
    my $ci = $job->command_interface();
    my $timeout = $ci->scale_timeout(
        'method-timeout' => $opts{'method-timeout'});

    try {
        my $end_time = timeout2time($timeout);
        my %state_opts = %opts;
        delete $state_opts{command_interface};
        my $success;

        while (time() < $end_time) {
            my $job_cs = $job->state(
                %state_opts,
                requested_fields => [ qw(state completion) ]
            );
            my $state = $job_cs->state();
            if ( $state =~ /Failure|Error|Quit|Dead/i ) {
                my $completion = $job_cs->completion();
                $Log->exit() if $may_exit;
                NACL::Exceptions::JobFailed->throw(
                    "Job $id failed:\n" . "Final state: $state\n"
                    . "Completion string: $completion\n",
                    state      => $state,
                    completion => $completion
                );
            } elsif ($state =~ /Success/i) {
                $success = 1;
                last;
            } else {
                Tharn::snooze($interval);
            }
        }
        if ((time() >= $end_time) && !$success) {
            $Log->exit() if $may_exit;

            my $dump_core = Tharn::param('NACL_COREDUMP_ON_TIMEOUT');
            if (defined $dump_core and $dump_core == 1) {
                coredump_on_timeout(command_interface => $ci);
            }

            NACL::Exceptions::Timeout->throw("After waiting for $timeout "
                 . "seconds the job did not complete");
        }
    }
    catch NACL::Exceptions::NoElementsFound with {

        # If a normal job, assume it has moved into the job history table.
        # If it's a private job, then there is no job history stored,
        # so we cannot look at the job history table.
        if ($job->isa('NACL::C::JobPrivate')) {
            $Log->warn("Final status of private job with id $id could "
                . 'not be determined. This is because the job has '
                . 'already completed and has been moved out of the '
                . 'private jobs table. We are assuming the job completed '
                . 'successfully and are proceeding.'
            );
            # Using return to break out of the catch block, so the
            # code following this in the block does not get executed.
            return;
        }

        my @states = NACL::CS::JobHistory->fetch(
            command_interface => $job->command_interface(),
            filter            => { id => $id },
            requested_fields  => [qw(completion event-type)],
            nacl_retry_count  => 0,
            %opts
        );
        foreach my $state (@states) {
            my $event_type = $state->event_type();
            if ( $event_type =~ /Failed|Error/i ) {
                $Log->exit() if $may_exit;
                NACL::Exceptions::JobFailed->throw(
                    "Historical job $id failed:\n"
                        . "  Final State: $event_type\n"
                        . "  Completion: "
                        . $state->completion() . "\n",
                    state      => $state->event_type,
                    completion => $state->completion()
                );
            }
        }
    };
    $Log->exit() if $may_exit;
}

=head2 _get_nodescope_node_name

    # Using NACL::C::VolumeFile->showfh as an example
	sub showfh {
		$Log->enter() if $may_enter;
		my $pkg  = shift;
		$pkg->_frontend_log_debug_opts(@_);
		my %opts = $pkg->_common_validate_with(
			params          => \@_,
			additional_spec => { _optional_scalars('nodescope-node-name', 'determine_nodescope_node_name') },
			only_static_method => 1
		);

		# get the name of the node based on the input options
		$opts{'nodescope-node-name'} = $pkg->_get_nodescope_node_name(%opts);

		my $obj = $pkg->call_on_apiset(
			%opts,
			choices => [
				{   method    => '_showfh_7mode_cli',
					interface => 'CLI',
					set       => 'Nodescope|7Mode',
				}
			],
		);

		$Log->exit() if $may_exit;
		return $obj;
	}

For component methods which support only Nodescope or 7Mode, this method
can be used to obtain the node name. If 'nodescope-node-name' is specified,
it just returns the same value as node name, else if
determine_appropriate_nodescope_node option is set to 1, it performs the
discovery by invoking the method specified in 'method' option.

=over

=item nodescope-node-name

The node-name on which to run the Nodescope command.

=item determine_appropriate_nodescope_node

If set to 1, the back-end implementation determines which node the
element is on and runs the command in the context of that Node.

=item method

The name of the method to be invoked for finding the node name in
which the element to be manipulated, is present. It defaults to
'get_local_node_name'.

=item %other_options

Other options to be passed to get_local_node_name() method

=back

=cut

sub _get_nodescope_node_name {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = validate_with(
        params => \@_,
        spec   => {
            command_interface =>
                { isa => 'NACL::C::CommandInterface::ONTAP' },
            'nodescope-node-name' => { type => SCALAR, optional => 1 },
            'determine_appropriate_nodescope_node' => {
                type     => BOOLEAN,
                optional => 1
            },
            method => {
                type    => SCALAR,
                default => 'get_local_node_name'
            }
        },
        allow_extra => 1
    );

    my $ci                  = $opts{command_interface};
    my $method              = delete $opts{method};
    my $nodescope_node_name = delete $opts{'nodescope-node-name'};
    my $determine_appropriate_nodescope_node
        = delete $opts{determine_appropriate_nodescope_node};
    my $name;

    if ( $ci->is_cmode ) {

        # Nodescope
        if ($nodescope_node_name) {
            $name = $nodescope_node_name;
        } elsif ($determine_appropriate_nodescope_node) {
            unless ( $pkg->can($method) ) {
                $Log->exit() if $may_exit;
                NATE::BaseException->throw( "Internal error: Method "
                        . "$method is not defined in '$pkg'" );
            }
            my %local_node_args;
            $pkg->_copy_primary_keys(
                source => \%opts,
                target => \%local_node_args
            );
            $pkg->_copy_common_component_params_with_ci(
                source => \%opts,
                target => \%local_node_args,
            );
            $name = $pkg->$method(%local_node_args);
        }
    }

    $Log->exit() if $may_exit;
    return $name;
}

=head2 _handle_cli_vserver_context, _handle_zapi_vserver_context

    # NACL/C/Foo.pm
    sub _foo_cmode_cli {
        ...
        my $command_interface = $opts{command_interface};
        my $apiset            = $opts{apiset};
        my $vserver           = $opts{vserver};
        ...
        $pkg_or_obj->_hash_copy(...);

        my $vserver_operated_on = $pkg_or_obj->_handle_cli_vserver_context(
            api_opts          => \%cmode_args,
            vserver           => $vserver,
            command_interface => $command_interface,
            frontend_method   => $method
        );

        $apiset->foo_api(%cmode_args);
    }

    # NACL/C/Foo.pm
    sub _foo_cmode_zapi {
        ...
        my $command_interface = $opts{command_interface};
        my $apiset            = $opts{apiset};
        my $vserver           = $opts{vserver};
        ...
        $pkg_or_obj->_hash_copy(...);

        my $vserver_operated_on = $pkg_or_obj->_handle_zapi_vserver_context(
            api_opts          => \%cmode_args,
            vserver           => $vserver,
            command_interface => $command_interface,
            frontend_method   => $method
        );

        $apiset->foo_api(%cmode_args);
    }

(For component developers)
Certain commands/ZAPIs can be run in Vserver context. These methods
are used to handle the differences between running these methods using
Vserver as a command interface or using Node/Cluster/Cserver as the
command interface. It returns the vserver being operated on (could
be the "vserver" argument passed in, or the name of the vserver which
the Vserver command interface is modeling)

For ZAPI, the method to be invoked is C<_handle_zapi_vserver_context>.
This ensures that if a vserver-scope ZAPI is being invoked with a Node
command interface, then it will set the API to run in the context of the
vserver passed as argument.

For CLI, the method to be invoked is C<_handle_cli_vserver_context>.
From SN.0 onwards, the "-vserver" option is no longer accepted in vserver
context. Hence, if a Vserver is used as command interface, then it
does not pass down the "vserver" option to the command.

This method throws a C<NATE::Exceptions::Argument> exception in the following
cases:

=over

=item 1)

If a vserver command interface is used and a vserver argument
is sent in the method call, but the argument does not match the
vserver of the command interface

=item 2)

The vserver argument is not passed when a non-Vserver
command interface is passed

=back

This method returns the name of the vserver on which we will operate. (This
could be the vserver attribute of the command interface or the vserver
option passed in)

=over

=item Options

=over

=item api_opts

(HASHREF, Mandatory)
A reference to the hash sent to the API call. This gets updated to contain
the vserver context to set to if a Node/Cluster command interface is used

=item vserver

(Mandatory)
The vserver option passed in

=item command_interface

(Isa 'NACL::C::CommandInterface::ONTAP', Mandatory)
The command interface being used

=back

=back

=cut

# Get implementations of _handle_zapi_vserver_context and
# _handle_vserver_context from the mix-in
use NACL::C::_Mixins::Component::ONTAP qw(:all);

sub _handle_cli_vserver_context {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    my $vserver = $pkg_or_obj->_handle_vserver_context( @args, is_zapi => 0 );
    $Log->exit() if $may_exit;
    return $vserver;
}

=head2 _vserver_context_init

    sub init {
        $Log->enter() if $may_enter;
        my ($self, @opts) = @_;
        $self->SUPER::init(@opts);
        $self->_vserver_context_init(@opts);
        # Perform any other checking that needs to be done
        $Log->exit() if $may_exit;
    }

(For component developers)
For components which could run in vserver context, we need to ensure that
the component object created has a vserver attribute if a Node/Cluster
command interface is used (since that attribute will be used to set the
vserver context).

This method is intended to be called in the validation "init" method. We perform
any checking the base class's init performs, call this method and then perform
any other checking we might want to do.

This method B<has> to be called for all components that could run in vserver
context.

=cut

sub _vserver_context_init {
    $Log->enter() if $may_enter;
    my ( $self, %opts ) = @_;
    my $command_interface = $opts{command_interface};
    my $vserver           = $opts{vserver};
    if ( $command_interface->isa('NACL::C::Vserver') ) {
        my $ci_vs = $command_interface->vserver();
        if ( defined $vserver ) {
            if ( $ci_vs ne $vserver ) {
                $Log->exit() if $may_exit;
                NATE::Exceptions::Argument->throw( 'The vserver attribute '
                        . "of the command interface ('$ci_vs') does "
                        . "not match the vserver parameter ('$vserver')"
                        . " passed in the call to new\n" );
            }
        } else {

            # Set the vserver attribute of the object to be the one
            # modeled by the ci
            $self->vserver($ci_vs);
        }
    } elsif ( $command_interface->is_cmode() ) {
        if ( !defined $vserver ) {
            my $ci_type = ref $command_interface;
            $ci_type =~ s/.*::(.*)/$1/;
            if ($ci_type ne 'Cserver') {
                my $pkg = ref $self;
                $Log->exit() if $may_exit;
                $Log->warn( "'vserver' parameter must be provided in the call "
                        . "to $pkg->new() when a $ci_type is used as "
                        . 'command interface. Please mail dl-nacl-dev@netapp.com '
                        . 'if this should not be the case for this component' );
            }
        }
    }
    $Log->exit() if $may_exit;
}

sub _handle_obj_cs_like_calls {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, %opts ) = @_;
    if ( ref $pkg_or_obj ) {
        if ( defined $opts{filter} && keys %{ $opts{filter} } ) {
            my @call = caller();
            $Log->exit() if $may_exit;
            NATE::Exceptions::Argument->throw( 'Filter should not be '
                    . "provided when $call[3] is invoked with an object" );
        }

        # Get the pri keys
        my %filter_hash = $pkg_or_obj->get_primary_keys_hash();

        # Construct the CS package name
        my @arr    = caller(1);
        my $method = $arr[3];
        if ( $method =~ /NACL::C::(.+)::(.+)/ ) {
            my $package = "NACL::CS::$1";
            $method = $2;

            $method =~ s/_(\S)/uc($1)/ge;
            $method =~ s/^(\S)/uc($1)/e;
            my $cs_package = $package . $method;

            # Delete the pri keys which are not valid for the CS like method
            # call
            eval "use $cs_package";
            foreach my $key ( keys %filter_hash ) {
                unless ( $cs_package->can($key) ) {
                    delete $filter_hash{$key};
                }
            }
        }
        $opts{filter}            = {%filter_hash};
        $opts{command_interface} = $pkg_or_obj->command_interface();
    }
    $Log->exit() if $may_exit;
    return %opts;
}

=head2 _cs_call_method_builder

    __PACKAGE__->_cs_call_method_builder(method => $method);

Provide a simple way to install methods at the Component-level that invoke
related component state methods and return back CS objects

For example, installing a method that runs the AggregateEcacheStats CS
and returns back the component state

    __PACKAGE__->_cs_call_method_builder(method => 'stats');

=over

=item Options

=over

=item C<< method => $method >>

(Mandatory)
The name of the method to install in the component. This would end up in
the CS name.

=item C<< extra_filters  => $arrayref >>

(Optional)

Arrayref of extra filter apart from that of primary key that CS would
accept. For example to accept filter 'keys' apart from the primary key
We need the below code

    __PACKAGE__->_cs_call_method_builder(method => 'find', extra_filters => ['key']);

=cut

sub _cs_call_method_builder {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    my %opts = validate_with(
        params => \@args,
        spec   => {
            method  => {type => SCALAR},
            extra_filters => {type => ARRAYREF, default => []},
        },
    );
    my $method  = $opts{method};

    my $cs_package = $pkg->get_CS_package_name($method);
    no strict 'refs';

    # Build cs call method
    *{"${pkg}::$method"} = sub {
        $Log->enter() if $may_enter;
        my ( $pkg_or_obj, %sub_args ) = @_;

        my %filter_hash = $pkg_or_obj->get_primary_keys_hash();

        require Module::Load;
        Module::Load::load($cs_package);

        map { $sub_args{filter}{$_} =
            delete $sub_args{$_} // $filter_hash{$_} // $sub_args{filter}{$_}
                if( $cs_package->can($_) );
        } ( keys %filter_hash );
        map { $sub_args{filter}{$_} =
            delete $sub_args{$_} // $filter_hash{$_} // $sub_args{filter}{$_}
        } ( @{ $opts{extra_filters} });

        $sub_args{command_interface} //= $pkg_or_obj->command_interface();
        $Log->exit() if $may_exit;
        return $cs_package->fetch(%sub_args);
    };
    use strict 'refs';
    $Log->exit() if $may_exit;
}

# Abstract base class implementation
sub _real_component { }

# Validation spec of options common to both ZAPI back-end methods
sub _common_zapi_backend_validate_spec {
    return {
        api            => { type => SCALAR },
        map            => { type => HASHREF, default => {} },
        copy           => { type => ARRAYREF, default => [] },
        cli_cmd        => { type => SCALAR, optional => 1 },
        vserver_operated_on => {type => SCALARREF, optional => 1},
        zapi_field_translations => {type => HASHREF, default => {}},
        inapplicable_for_help_xml => { type => ARRAYREF, default => [] },
        disable_help_xml_lookup => {type => SCALAR, default => 0},
    }
}

=head2 _iter_helper

    sub modify {
        my ($pkg_or_obj, @opts) = @_;
        $pkg_or_obj->_iter_helper(
            @opts,
            api  => $api,
            copy => [ @copy_list ],
            map  => { %mapping_hash },
            does_not_exist_error => $error,
            does_not_exist_exception => $exception_name
        );
    }

Certain modify and delete operations have an "-iter" equivalent. These
allow for modifying/deleting multiple elements at once by using the
C<query> facility (exposed through NACL code as C<extended_query>). This
method does the heavy lifting to make use of the "-iter" ZAPIs.

=over

=item Options

=over

=item C<< api => $api >>

This is the name of the ZAPI to be invoked.

item C<< copy => $copy_arrayref >>

(Optional)
The options which are named in ZAPI as they are in the CLI.

=item C<< map => $map_hashref >>

(Optional)
The mapping of options from CLI -> ZAPI.

=item C<< does_not_exist_error => $text | $hashref >>

(Optional)
When the element we're attempting to modify/delete does not exist,
this determines the text of the thrown exception. Note that if this is
not provided, then it defaults to "Foo does not exist". (where "Foo"
is the component name)

If provided as a scalar value, this is the text of the thrown exception.

It can also be provided as a hash-reference containing a code-reference
detailing how to build the error message. This hash-reference should
have two keys: "sub" and "args". The "sub" should be the code-reference,
while the "args" (an array-reference) specifies what arguments are to be
sent to it.

This is useful for cases where we'd like the error message to state the
name of the element that could not be found. For example, attempting to
modify a volume that does not exist results in an error message like
this on the CLI:

    Error: command failed: Volume "khsfg" not found. Reason: entry doesn't exist.

This is what we'd provide as the value for C<does_not_exist_error> to
make the exception text be CLI-like:

    does_not_exist_error => {
        sub => sub {
            my ($vol_name) = validate_pos(@_, { type => SCALAR });
            return 'Volume "' . $vol_name . '" not found. Reason: entry '
                . "doesn't exist";
        },
        args => [ $opts{volume} ]
    }

=item C<< does_not_exist_exception => $exception_name >>

The name of the exception to be thrown when the element being modified/deleted
does not exist. This would typically be NACL::C::Exceptions::Foo::DoesNotExist.

=item C<< jobid_field => $jobid_field >>

(Optional, defaults to "jobid")
If the operation is performed in a job, one of the fields of the ZAPI output
contains the ID of the job that was started. This field is generally named
"jobid", but in case this field is named differently, this argument
can be used to specify what the field is named.

=item C<< no_attributes => 0 | 1 >>

(Optional, defaults to 0)
Normally, this method would move the primary keys into the C<query> argument
for the ZAPI call while sending all the other options through the
C<attributes> argument.

However, certain ZAPIs (all -delete-iter ZAPIs, for example) do not accept
the C<attributes>. For these ZAPIs, all options need to be sent through the
C<query> argument. This can be achieved by setting this option to 1.

=item C<< fix_api_args => $coderef >>

(Optional)
This helper method uses a particular algorithm to determine the arguments
to the ZAPI call. However, if there are cases where we want to tweak the
arguments calculated, then we can "fix" the arguments using this option.

The value to be provided is an anonymous subroutine. This anonymous
subroutine is sent the API arguments as a hash. This hash could contain
either of both of "query" and "attributes". The subroutine is expected
to return the updated API arguments (as a hash).

Taking a purely hypothetical, let's assume the _iter_helper method
had constructed the input arguments as this:

( query => { a => $a, b => $b }, attributes => { c => $c } )

Let's say that this particular ZAPI is non-standard, because of which
the standard algorithm does not apply, and "b" should be in attributes
instead of "query". We would do so like this:

    fix_api_args => sub {
        my %args         = @_;
        my $query        = $args{query};
        my $attributes   = $args{attributes};
        $attributes->{b} = delete $query->{b};
        return %args;
    }

=item C<< zapi_field_translations => $hashref >>

See the description under L<_base_backend_cmode_zapi>.

=item C<< pass_in_query => $arrayref >>

(Optional) Array reference  containing the list of args to be passed only in query

=item C<< show_cmd => $show_cmd >>

(Optional) The ngsh show command for the component library

=item C<< inapplicable_for_help_xml => [] >>

(Optional) Arrayref containing the list of args for which ZAPI mapping need not be
looked up in help_xml

=back

=back

=cut

sub _iter_helper {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %iter_helper_spec = (
        %{$pkg_or_obj->_common_zapi_backend_validate_spec()},
        extended_query => { type => HASHREF, default => {} },
        does_not_exist_error =>
            { type => SCALAR | HASHREF, optional => 1 },
        does_not_exist_exception => { type => SCALAR },
        jobid_field => { type => SCALAR, default => 'jobid' },
        no_attributes => { type => BOOLEAN, default => 0 },
        fix_api_args  => { type => CODEREF, optional => 1 },
        pass_through  => { type => HASHREF, default => {} },
        pass_in_query => { type => ARRAYREF, default => [] },
        show_cmd      => { type => SCALAR, optional => 1 },
    );
    my %backend_validate_spec = %{$pkg_or_obj->_backend_validate_spec()};
    my %full_backend_validate_spec = (%backend_validate_spec, %iter_helper_spec);
    my %opts = validate_with(
        params => \@args,
        spec   => \%full_backend_validate_spec,
        # The other options are those which are for the API call
        allow_extra => 1
    );
    my $api                      = $opts{api};
    my $map                      = $opts{map};
    my $copy                     = $opts{copy};
    my $apiset                   = $opts{apiset};
    my $jobid_field              = $opts{jobid_field};
    my $command_interface        = $opts{command_interface};
    my $disable_help_xml_lookup  = delete $opts{disable_help_xml_lookup};

    if ( $opts{show_cmd}
         && $command_interface->can_get_version_details() ) {
        $pkg_or_obj->_update_mapping_translation(
            orig_args            => \%opts,
            known_backend_fields => [keys %full_backend_validate_spec],
            disable_help_xml_lookup => $disable_help_xml_lookup,
        );
    }

    # This call has to be before the options are translated
    $pkg_or_obj->_zapi_field_translation_code(\%opts);

    my $does_not_exist_exception = $opts{does_not_exist_exception};
    my $if_not_exists = $opts{does_not_exist_error};
    my $if_not_exists_msg;
    if ( defined $if_not_exists ) {

        # If passed as a reference, check that it's a hash-reference and
        # has the keys "sub" (type CODEREF) and "args" (type ARRAYREF)
        if ( ref $if_not_exists ) {

            # is a hash-ref, check that the "sub" and "args" keys are present
            # and are of the correct type
            my $method = ( caller(1) )[3];
            try {
                validate_with(
                    params => $if_not_exists,
                    spec   => {
                        sub  => { type => CODEREF },
                        args => { type => ARRAYREF },
                    }
                );
            }
            catch NATE::Exceptions::Argument with {
                my $exception = shift;
                $exception->set_text( 'Internal error: the '
                        . "'error_if_not_exists' argument was not passed "
                        . "correctly from $method. Error:\n"
                        . $exception->text() );
                $Log->exit() if $may_exit;
                $exception->throw();
            };
            my $sub = $if_not_exists->{sub};

            # Invoke the code-ref, sending the args specified
            $if_not_exists_msg = $sub->( @{ $if_not_exists->{args} } );
        } else {
            $if_not_exists_msg = $if_not_exists;
        }
    } else {

        # Defaults to "$pkg does not exist"
        my $pkg = $pkg_or_obj->get_package_name();
        $pkg =~ /.*::(.*)/;
        $pkg               = $1;
        $if_not_exists_msg = "$pkg does not exist";
    }

    my @cli_to_zapi = %{ $opts{map} };
    my @zapi_to_cli = reverse %{ $opts{map} };

    # copy => [ qw(a b) ] is equivalent to doing map =>{a =>['a'],b=>['b']}
    # For each element in copy, we make an entry of this form
    foreach ( @{ $opts{copy} } ) {
        push @cli_to_zapi, ( $_ => [$_] );
        push @zapi_to_cli, ( $_ => [$_] );
    }

    # The primary keys should not go into the attributes, they should go
    # into the query.
    my %primary_keys;
    $pkg_or_obj->_move_primary_keys(
        source => \%opts,
        target => \%primary_keys
    );

    my %api_args;

    $api_args{query} = {};

    if (@{ $opts{pass_in_query} }) {
        my %pass_as_query;
        $pkg_or_obj->_hash_move(
            source => \%opts,
            target => \%pass_as_query,
            move   => $opts{pass_in_query}
        );
        $pkg_or_obj->_zapi_hash_copy(
            source => \%pass_as_query,
            target => $api_args{query},
            map    => \@cli_to_zapi,
            target_skips => 1,
        );
    }

    # If "attributes" is not accepted by the ZAPI call (such as for
    # delete-iter's), then copy everything into "query", else copy it
    # into "attributes"
    if ($opts{no_attributes}) {
        $pkg_or_obj->_zapi_hash_copy(
            source => \%opts,
            target => $api_args{query},
            map    => \@cli_to_zapi,
            target_skips => 1,
        );
    } else {
        $api_args{attributes} = $pkg_or_obj->_zapi_hash_copy(
            source => \%opts,
            map    => \@cli_to_zapi,
            target_skips => 1,
        );
    }

    # Don't send in "attributes" if it's empty (this is necessary for
    # ZAPIs which don't accept the "attributes" input)
    delete $api_args{attributes} unless (keys %{$api_args{attributes}});

    if ( keys %{ $opts{extended_query} } ) {
        $pkg_or_obj->_zapi_hash_copy(
            source => $opts{extended_query},
            target => $api_args{query},
            map    => \@cli_to_zapi,
            target_skips => 1,
        );
    }

    $pkg_or_obj->_zapi_hash_copy(
        source => \%primary_keys,
        target => $api_args{query},
        map    => \@cli_to_zapi,
        target_skips => 1,
    );


    if ($opts{fix_api_args}) {
        %api_args = $opts{fix_api_args}->(%api_args);
    }

    my @errors;
    my $set_job_component;

    # The parsed output structure for success-list or failure-list is
    # (using success-list and volume-modify-iter as example):
    # {
    #   'success-list' => [
    #                      {
    #                        'volume-modify-iter-info' => [
    #                                                      {...}, {...}, ...
    #                                                     ]
    #
    # The only array-reference that matters is the inner-most one. This
    # code-ref is for extracting this inner-most arrayref.
    my $get_list_row = sub {
        my %opts = validate_with(
            params => \@_,
            spec   => {
                only_row  => { type => HASHREF },

                # This should be either "success-list" or "failure-list"
                list_type => { type => SCALAR },
            }
        );
        my $only_row_in_sub = $opts{only_row};
        my $list_type       = $opts{list_type};

        my $temp_hashref   = $only_row_in_sub->{$list_type}[0];
        my @keys           = keys %$temp_hashref;
        my $key            = $keys[0];
        my $inner_arrayref = $temp_hashref->{$key};
        return $inner_arrayref;
    };
    if(($api =~ /qos_policy_group_delete_iter/) && (defined($opts{force}))) {
        #field 'force' is an exceptional case for qos_policy_group_delete_iter
        #which needs special handling.
        $api_args{force} =  $opts{force};
    }
    while (1) {
        my $response = $apiset->$api(
            %api_args,

            %{$opts{pass_through}},

            # Default CLI behavior is to perform as many operations
            # as possible, even there are any failures in the middle
            # ZAPI defaults to failing at the first failure, so set
            # "continue-on-failure" to "true" to plough on
            'continue-on-failure' => 'true',

            # the modify/delete-iter's allow operating on a maximum of 100
            # records at at time
            'max-records' => 100
        );
        my $output   = $response->get_parsed_output();
        my $only_row = $output->[0];

        my $num_failed    = $only_row->{'num-failed'};
        my $num_succeeded = $only_row->{'num-succeeded'};
        if ($num_succeeded) {

            # If either job_component or job_components were passed, then
            # go through the success list to extract jobids (if any)
            if ( $opts{job_component} || $opts{job_components} ) {

                my $successes_arrayref = $get_list_row->(
                    only_row  => $only_row,
                    list_type => 'success-list',
                );
                foreach my $success_hashref (@$successes_arrayref) {
                    my $jobid = $success_hashref->{$jobid_field};
                    if ( defined $jobid ) {

                        my $job_obj = NACL::C::Job->new(
                            id                => $jobid,
                            command_interface => $opts{command_interface}
                        );

                        # Set job_component if it hasn't been set yet.
                        if ( exists $opts{job_component} && !$set_job_component )
                        {
                            ${ $opts{job_component} } = $job_obj;
                            $set_job_component = 1;
                        }

                        if ( exists $opts{job_components} ) {
                            push @{ $opts{job_components} }, $job_obj;
                        } else {
                            # User has requested only for one job: which we
                            # have already filled in. No need to go through
                            # the rest of the list looking for other job-IDs.
                            last;
                        }
                    }
                }
            }
        }

        if ($num_failed) {
            my $failures_arrayref = $get_list_row->(
                only_row  => $only_row,
                list_type => 'failure-list',
            );
            foreach my $failure_hashref (@$failures_arrayref) {
                my %failure = %$failure_hashref;

                # We will be deleting from %failure, so not iterating
                # over it
                while ( my ( $key, $value ) = each %$failure_hashref ) {
                    if ( $key =~ /-key/ ) {
                        $failure{text} = delete $failure{'error-message'};
                        $failure{error_code} = delete $failure{'error-code'};
                        delete $failure{$key};
                        my $info_outer = $value->[0];
                        my $info_key   = ( keys %$info_outer )[0];
                        my $info_hash  = $info_outer->{$info_key}[0];
                        $failure{primary_keys} = $pkg_or_obj->_zapi_hash_copy(
                            source => $info_hash,
                            map    => \@zapi_to_cli
                        );
                    }
                }
                push @errors, { %failure, response_object => $response };
            }
        } elsif ( !$num_succeeded ) {

            # If both num-failed and num-succeeded are 0 then it's possible
            # that no records got modified because the requested element
            # does not exist. (say we had provided vserver => 'abc', but
            # "abc" does not exist) We should throw an exception for this
            # case.
            # However, if the modify-iter ran multiple times, then the
            # last time it runs it will not operate on any element. This
            # should not be classified as an error, so we check whether
            # the call to modify-iter was the first one or not (done by
            # looking at the "tag" argument)
            if ( !$api_args{tag} ) {
                $Log->exit() if $may_exit;
                $does_not_exist_exception->throw(
                    text            => $if_not_exists_msg,
                    response_object => $response
                );
            }
        }

        $api_args{'tag'} = $only_row->{'next-tag'};
        last if ( !$api_args{'tag'} );
    }

    if (@errors) {
        if ( @errors == 1 ) {

            # Single failure: turn it into a CommandFailedException
            my $error_hashref = $errors[0];
            my $text          = $error_hashref->{text};
            my $response_obj  = $error_hashref->{response_object};
            $Log->exit() if $may_exit;
            NACL::APISet::Exceptions::CommandFailedException->throw(
                text            => $text,
                response_object => $response_obj
            );
        } else {

           # Multiple failures, turn it into a MultipleCommandFailed exception
            NACL::C::Exceptions::MultipleCommandFailed->throw(
                all_failures => \@errors );
        }
    }

    $Log->exit() if $may_exit;
}

sub _common_validate_with {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    my %opts = validate_with(
        params => \@args,
        spec   => {
            params          => { type => ARRAYREF },
            api             => { type => SCALAR, optional => 1 },
            cli_cmd         => { type => SCALAR, optional => 1 },
        },
        allow_extra => 1,
    );
    my $computed_spec   = {};
    my $computed_params = {};

    my $api         = delete $opts{api};
    my $cli_cmd     = delete $opts{cli_cmd};
    if (defined $cli_cmd && !defined $api) {
        $api = $cli_cmd;
        $api =~ s/\s/_/g;
        $api =~ s/-/_/g;
    }

    if ( defined $api ) {
        my $spec_method = $pkg_or_obj->_construct_spec_method(api => $api);
        my $pkg         = $pkg_or_obj->get_package_name();

        # The overridden can will use a CMode-CLI APISet to look for a
        # matching CDEF. This spec method is definitely not a CDEF, so we can
        # get it to skip the creation of this APISet by not passing in the
        # command interface.
        if ( $pkg->can($spec_method) ) {
            $opts{consolidated_spec} = $pkg_or_obj->$spec_method();
        }
    }

    my %options;
    try {
        %options = $pkg_or_obj->SUPER::_common_validate_with(
            %opts,
            get_computed_spec   => $computed_spec,
            get_computed_params => $computed_params
        );
    }
    catch NATE::Exceptions::Argument with {
        my $exception_obj = $_[0];
        my %params        = %{ $computed_params };
        my %spec          = %{ $computed_spec };
        if (    defined $params{command_interface}
             && blessed $params{command_interface}
             && $params{command_interface}->isa('NACL::C::CommandInterface::ONTAP')
             && $params{command_interface}->is_cmode()
             && defined $api )
        {
            my %spec_hash;
            map { $spec_hash{$_} = 1 } keys %spec;
            my @new_fields;
            map { push @new_fields, $_ if ( !exists $spec_hash{$_} ); }
                keys %params;
            if ( scalar @new_fields ) {
                my $updated_spec = $pkg_or_obj->_generate_spec(
                    spec   => $computed_spec,
                    params => $computed_params,
                    api    => $api,
                    cli_cmd => $cli_cmd
                );
                $Log->debug(
                    sub { "\n Updated spec: " .  Dumper($updated_spec) } );
                my %validate_with_opts = $pkg_or_obj->_get_validate_with_spec(
                    params => $computed_params,
                    opts   => \%opts
                );
                %options = validate_with(
                    params => $computed_params,
                    spec   => $updated_spec,
                    %validate_with_opts,
                );
            } else {
                $Log->exit() if $may_exit;
                $exception_obj->throw();
            }
        } else {
            $Log->exit() if $may_exit;
            $exception_obj->throw();
        }
    };

    $options{cli_cmd} = $cli_cmd if (defined $cli_cmd);
    $Log->exit() if $may_exit;
    return %options;
}

#
# Private function which returns the updated spec for new options
#
sub _generate_spec {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    my %opts = validate_with(
        params => \@args,
        spec   => {
            params  => { type => HASHREF },
            spec    => { type => HASHREF },
            api     => { type => SCALAR },
            cli_cmd => {type => SCALAR|UNDEF},
        }
    );
    my ( %new_fields, $ds, $cdef, @array_field );
    my $params              = $opts{params};
    my $original_spec       = $opts{spec};
    my %updated_spec        = %{$original_spec};
    my $instant_method_name = $pkg_or_obj->_construct_spec_method(
        api => $opts{api});

    my $ci = $opts{params}->{command_interface};


    my $help_xml_obj = NACL::Realm->realm()->get_help_xml_obj(command_interface => $ci);
    if ($help_xml_obj->help_xml_accessible()) {
        my $cmd = $opts{cli_cmd};
        if (!defined $cmd) {
            $cmd = $ci->translate_cm_cli_api_to_command(api => $opts{api});
        }
        my $alias_map;
        if ( $ci->apiset(check_apiset_already_loaded => 1) ) {
            my $cdef = $ci->apiset()->get_command_definition(command => $opts{api});
            $alias_map = $cdef->get_alias_map();
        }
        $ds = $help_xml_obj->get_attribute_map(command => $cmd,
                                               alias_map => $alias_map);
    }

    if (!$ds) {
        # Either help_xml files not accessible, or no help_xml file present
        # for this command
        $ds = $help_xml_obj->get_ui_details(command_interface => $ci, api => $opts{api});
    }

    if ($ds) {
        foreach my $field ( keys %{ $ds->{fields} } ) {
            if ( !$original_spec->{$field} ) {
                if ( $ds->{fields}->{$field}->{is_array} ) {
                    $updated_spec{$field}
                        = { type => ARRAYREF, optional => 1 };
                } else {
                    $updated_spec{$field} = { type => SCALAR, optional => 1 };
                }
            }
        }
    }
    my $package_name = $pkg_or_obj->get_package_name();
    my $coderef      = sub {
        $Log->enter() if $may_enter;
        $Log->exit() if $may_exit;
        return \%updated_spec;
    };
    no strict 'refs';
    *{"${package_name}::${instant_method_name}"} = $coderef;
    use strict 'refs';
    $Log->exit() if $may_exit;
    return \%updated_spec;
}

=head2 get_api_options

       my %target_hash = $pkg_or_obj->get_api_options(
                             api  => $api,
                             apiset => $apiset,
                             options => $options,
                            );

This method returns a hash containing the allowed options
which an api command can take. It deletes the invalid
options from $options hashref and returns the resulting
hash

=over

=item Options

=over

=item C<< api => $api >>

(Mandatory)  The name of the api command

=item C<< apiset => $apiset_obj >>

(Mandatory) HASHREF The apiset object referring to a node.

=back

=back

=cut

sub get_api_options {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    my %opts = validate_with(
        params => \@args,
        spec   => {
            api     => { type => SCALAR },
            options => { type => HASHREF },
            apiset  => {
                type => OBJECT,
                isa  => 'NACL::APISet'
            },
            cdef_ref => {type => SCALARREF, optional => 1},
        },
    );
    my $apiset        = $opts{apiset};
    my $api           = $opts{api};
    my $cdef_ref      = $opts{cdef_ref};
    my %input_options = %{ $opts{options} };
    my $cdef = $apiset->get_command_definition( command => $api );
    $$cdef_ref = $cdef if (defined $cdef_ref);

    my @options = $apiset->get_allowed_options_for_command(command => $api,
        cdef_obj => $cdef);
    push @options, 'extended_query';

    my $alias_map = $cdef->get_alias_map();

    if ( $alias_map ) {
        while (my ($old, $new) = each (%{ $alias_map })) {
            if ( $input_options{$old} ) {
                 $input_options{$new}=delete $input_options{$old};
            }
        }
    }
    my $target = $pkg_or_obj->_hash_copy(
        source => \%input_options,
        copy   => \@options
    );
    my %target_hash =  %{$target};
    foreach my $key ( keys %target_hash ) {
        if ( ref $target_hash{$key} eq "ARRAY" ) {
           $target_hash{$key} = join( ",", @{ $target_hash{$key} });
        }
    }
    $Log->exit() if $may_exit;
    return %target_hash;
}

=head2 _populate_job_components

    $pkg_or_obj->_populate_job_components(
        command_interface => $ci,
        response          => $response,
        job_component     => \$job,
        job_components    => \@jobs,
    );

This method creates the job objects and fill the
job_component/job_components based on input arguments.

=over

=item Options

=over

=item C<< command_interface => $ci >>

(Mandatory)

The command interface object.

=item C<< response => $response  >>

(Mandatory)  Response object

=item C<< job_component => \$job_component >>

(Optional)

This would be filled if the scalar ref passed to job_component

=item C<< job_components => \@job_components >>

(Optional)

This would be filled if the arrayref passed to job_components

=back

=back

=cut

sub _populate_job_components {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    my %opts = validate_with(
        params => \@args,
        spec   => {
            command_interface =>
                { isa => 'NACL::C::CommandInterface::ONTAP' },
            job_component  => { type => SCALARREF | UNDEF, optional => 1, },
            job_components => { type => ARRAYREF | UNDEF, optional => 1, },
            response => {
                type => OBJECT,
                isa  => 'NACL::APISet::Response'
            }
        },
        # Allow through the common options
        allow_extra => 1
    );

    my $job_component     = delete $opts{job_component};
    my $job_components    = delete $opts{job_components};
    my $command_interface = $opts{command_interface};
    my $response          = delete $opts{response};

    if (!defined $job_component && !defined $job_components) {
        # User has not requested for the job(s) to be returned, so nothing
        # more to do
        $Log->exit() if $may_exit;
        return;
    }

    my @jobids   = $response->get_job_ids();
    if (scalar @jobids) {
        # ZAPI does not appear to return private jobs, so the ZAPI Response
        # class does not have an is_private_job method implemented. Hence,
        # first check for presence of the is_private_job method.
        my $is_private;
        if ($response->can('is_private_job')) {
            $is_private = $response->is_private_job();
        } else {
            # Method not implemented; must be a public job
            $is_private = 0;
        }

        my @private_jobs;
        if ($is_private) {
            # JobPrivate requires the values of node+vserver to also
            # be filled into the object, so we need to make a call
            # to find() to get these values.
            my $ids_str = join '|', @jobids;
            @private_jobs = NACL::C::JobPrivate->find(
                %opts,
                filter => {id => $ids_str}
            );
        }

        if (defined $job_component) {
            if ($is_private) {
                $$job_component = $private_jobs[0];
            } else {
                $$job_component = NACL::C::Job->new(
                    id                => $jobids[0],
                    command_interface => $command_interface
                );
            }
        }
        if (defined $job_components) {
            if ($is_private) {
                push @{$job_components}, @private_jobs;
            } else {
                foreach my $id (@jobids) {
                    my $job_obj = NACL::C::Job->new(
                        id                => $id,
                        command_interface => $command_interface
                    );
                    push @{$job_components}, $job_obj;
                }
            }
        }
    } ## end if ($jobid)

    $Log->exit() if $may_exit;
}

sub _construct_spec_method {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts = validate_with(
        params => \@args,
        spec   => { api => { type => SCALAR } }
    );
    my $spec_method = '_' . $opts{api} . '_spec';
    $Log->exit($spec_method);
    return $spec_method;
}

=head2 _base_backend_cmode_cli

       $pkg_or_obj->_base_backend_cmode_cli(
            @_,
            api  => 'volume_create',
       );

This method is the generic base class method for
cmode cli backends.

=over

=item Options

=over

=item C<< api => $api  >>

(Mandatory)  The name of the api command

=item C<< additional_args => \%args >>

(Optional)

Hashref which contains additonal api arguments ,which
are not listed in the cdef file and connectrec match
table which would be used to answer the prompts

=item C<< vserver_operated_on  => \$scalar  >>

(Optional) Scalarref,when set it handles cli vserver
context and updates the scalarref with the
name of the vserver to be operated on

=item C<< apiset_response_obj  => \$scalar  >>

(Optional) Scalarref,updates the passed in
scalarref with the apiset response object

=back

=back

=cut

sub _base_backend_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my $spec = $pkg_or_obj->_backend_validate_spec();
    my %opts = validate_with(
        params => \@args,
        spec   => {
            %{$spec},
            api                 => {type => SCALAR},
            additional_args     => {type => HASHREF, default => {}},
            vserver_operated_on => {type => SCALARREF, optional => 1},
            apiset_response_obj => {type => SCALARREF, optional => 1}
        },
        allow_extra => 1
    );
    my $vs_context        = $opts{vserver_operated_on};
    my $api               = $opts{api};
    my $apiset            = $opts{apiset};
    my $execution_time    = $opts{cmd_execution_time};
    my $command_interface = $opts{command_interface};
    my $apiset_response_obj  = $opts{apiset_response_obj};
    my $cdef;
    my %cmode_args        = $pkg_or_obj->get_api_options(
        apiset  => $apiset,
        api     => $api,
        options => \%opts,
        cdef_ref => \$cdef,
    );

    if ( $vs_context ) {
        $api =~ /_(.*?)$/;
        $$vs_context = $pkg_or_obj->_handle_cli_vserver_context(
            api_opts          => \%cmode_args,
            command_interface => $command_interface,
            frontend_method   => $1,
        );
    } ## end if ($vs_context)

    %cmode_args = (%cmode_args, %{$opts{additional_args}});
    my $response = $apiset->$api(%cmode_args, cdef_obj => $cdef);

    if (defined $execution_time) {
        $$execution_time = $response->get_execution_time();
    }

    if (defined $apiset_response_obj) {
        $$apiset_response_obj = $response;
    }
    $pkg_or_obj->_populate_job_components(
        command_interface => $command_interface,
        response          => $response,
        job_component     => $opts{job_component},
        job_components    => $opts{job_components},
        apiset_should     => {interface => 'CLI'},
    );
    $Log->exit() if $may_exit;
} ## end sub _base_backend_cmode_cli

=head2 _base_backend_cmode_zapi

    sub _foo_cmode_zapi {
        $Log->enter() if $may_enter;
        my ($pkg_or_obj, @args) = @_;
        $pkg_or_obj->_base_backend_cmode_zapi(
            @args,
            api  => $zapi,
            map  => \%some_mapping,
            copy => \@some_copy,
            %other_opts
        );
        $Log->exit() if $may_exit;
    }

A base class helper for CMode-ZAPI back-end methods. This takes on
a lot of the heavy lifting (_hash_copy'ing elements, handling vserver
context, getting execution time, populating job objects) typically required
in CMode-ZAPI back-end methods.

=over

=item Options

=over

=item C<< backend options >>

All the arguments sent through to the back-end should be sent to this
method.

=item C<< api => $api >>

(Mandatory)

The ZAPI to be invoked.

=item C<< map => \%map >>

(Optional)

The mapping hash-reference.

=item C<< copy => \@copy >>

(Optional)

The copy array-reference.

=item C<< additional_args => \%additional_args >>

(Optional)

Any additional arguments (i.e. not covered by the copy/map) that need to
be sent through to the ZAPI call.

=item C<< vserver_operated_on => \$scalar >>

(Optional, applicable only if the ZAPI should be invoked in vserver context)

A reference to a scalar variable, which gets filled in with the vserver
in which the ZAPI gets invoked.

=item C<< frontend_method => $method >>

(Optional)

The name of the front-end method.

=item C<< show_cmd => $show_cmd >>

(Optional) The ngsh show command for the component library

=item C<< inapplicable_for_help_xml => [] >>

(Optional) Arrayref containing the list of args for which ZAPI mapping need not be
looked up in help_xml

=item C<< zapi_field_translations => $hash >>

A hash-reference that specifies the fields that are to be translated. Example:

    zapi_field_translations => {
        convert_to_bytes  => [qw(max-data-constituent-size
            max-namespace-constituent-size maxdir-size)
        ],
        remove_percentage => [qw(percent-snapshot-space)],
    },

The supported keys are:

=over

=item convert_to_bytes

The CLI accepts the value with size suffixes (KB,MB etc), but the ZAPI
accepts only an integer byte value.

=item remove_percentage

The CLI accepts the value with a "%" at the end, but the ZAPI accepts it as
only an integer.

=item on_off_boolean

The CLI accepts the value as on/off, but the ZAPI accepts it as true/false.

=item enable_disable

The CLI accepts the value as enable/disable, but the ZAPI accepts it as true/false.

=item octal_to_decimal

The CLI accepts the value as octals, but the ZAPI accepts it as decimal.

=item hex_to_decimal

The CLI accepts the value as hex, but the ZAPI accepts it as decimal.

=item string_to_timestamp

The CLI accepts date as a string("Thu Mar 27 01:47:52 2014"), but the ZAPI accepts it
as a unixtime.

=item mmddyyyy_to_timestamp

The CLI accepts date as a string("MM/DD/YYYY HH:MM:SS"), but the ZAPI accepts it
as a unixtime.

=item hrminsec_to_sec

The CLI accepts time intervals as a string("<integer>h<integer>m<integer>s"), but the ZAPI accepts it
as seconds.

=item yes_no_boolean

The CLI accepts the value as yes/no, but the ZAPI accepts it as true/false.

=item cron_time_integer

The CLI accepts the value as 'all', but the ZAPI accepts it as -1.

=back

=back

=back

=cut

sub _base_backend_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %backend_spec = %{$pkg_or_obj->_backend_validate_spec()};
    my %base_backend_spec = (
        %{$pkg_or_obj->_common_zapi_backend_validate_spec()},
        additional_args     => {type => HASHREF, default => {}},
        apiset_response_obj => {type => SCALARREF, optional => 1},
        frontend_method     => { type => SCALAR, optional => 1 },
    );
    my %full_backend_validate_spec = (%backend_spec, %base_backend_spec);
    my %opts = validate_with(
        params      => \@args,
        spec        => \%full_backend_validate_spec,
        allow_extra => 1
    );
    my $vs_context           = $opts{vserver_operated_on};
    my $api                  = $opts{api};
    my $apiset               = $opts{apiset};
    my $execution_time       = $opts{cmd_execution_time};
    my $command_interface    = $opts{command_interface};
    my $apiset_response_obj  = $opts{apiset_response_obj};
    my $disable_help_xml_lookup = delete $opts{disable_help_xml_lookup};

    if ( $opts{cli_cmd}
         && $command_interface->can_get_version_details()
       ) {
        $pkg_or_obj->_update_mapping_translation(
            orig_args => \%opts,
            known_backend_fields => [keys %full_backend_validate_spec],
            disable_help_xml_lookup => $disable_help_xml_lookup,
        );
    }
    # This call has to be before the options are translated
    $pkg_or_obj->_zapi_field_translation_code(\%opts);

    my %cmode_args;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%cmode_args,
        copy   => $opts{copy},
        map    => $opts{map},
    );

    if ( $vs_context ) {
        my %extra_args;
        $pkg_or_obj->_hash_copy(
            source => \%opts,
            target => \%extra_args,
            copy   => [ qw(frontend_method) ]
        );
        $$vs_context = $pkg_or_obj->_handle_zapi_vserver_context(
            api_opts          => \%cmode_args,
            vserver           => $opts{vserver},
            command_interface => $command_interface,
            %extra_args
        );
    } ## end if ($vs_context)

    %cmode_args = (%cmode_args, %{$opts{additional_args}});
    my $response = $apiset->$api(%cmode_args);

    if (defined $execution_time) {
        $$execution_time = $response->get_execution_time();
    }

    if (defined $apiset_response_obj) {
        $$apiset_response_obj = $response;
    }

    $pkg_or_obj->_populate_job_components(
        command_interface => $command_interface,
        response          => $response,
        job_component     => $opts{job_component},
        job_components    => $opts{job_components},
        apiset_should     => {interface => 'ZAPI'},
    );
    $Log->exit() if $may_exit;
} ## end sub _base_backend_cmode_zapi

sub _base_backend_cmode_gui {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my $spec = $pkg_or_obj->_backend_validate_spec();

    # many of the arguments listed below are kept for future support
    # currently only copy, map and api are supported.
    state $base_backend_spec = {
        api                 => {type => SCALAR},
        additional_args     => {type => HASHREF, default => {}},
        vserver_operated_on => {type => SCALARREF, optional => 1},
        apiset_response_obj => {type => SCALARREF, optional => 1},
        copy                => {type => ARRAYREF, default => []},
        map                 => {type => HASHREF, default => {} },
        cli_cmd             => { type => SCALAR, optional => 1 },
        inapplicable_for_help_xml => { type => ARRAYREF, default => [] },
    };
    my %opts = validate_with(
        params      => \@args,
        spec        => {%$spec, %$base_backend_spec},
        allow_extra => 1
    );

    my $vs_context           = $opts{vserver_operated_on};
    my $api                  = $opts{api};
    my $apiset               = $opts{apiset};

    my %cmode_args;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%cmode_args,
        copy   => $opts{copy},
        map    => $opts{map},
    );

    %cmode_args = (%cmode_args, %{$opts{additional_args}});
    $apiset->$api(%cmode_args);

    $Log->exit() if $may_exit;
} ## end sub _base_backend_cmode_gui

sub _zapi_field_translation_code {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, $opts_ptr) = @_;
    my $zapi_field_translations = $opts_ptr->{zapi_field_translations};
    if (keys %$zapi_field_translations) {
        my %known_translations = (
            convert_to_bytes    => '_zapi_translate_to_bytes',
            remove_percentage   => '_zapi_translate_percentage',
            on_off_boolean      => '_zapi_translate_on_off',
            enable_disable      => '_zapi_translate_enable_disable',
            octal_to_decimal    => '_zapi_translate_octal_decimal',
            hex_to_decimal      => '_zapi_translate_hex_decimal',
            string_to_timestamp => '_zapi_translate_string_to_timestamp',
            mmddyyyy_to_timestamp => '_zapi_translate_string_to_timestamp',
            hrminsec_to_sec   => '_zapi_translate_hrminsec_to_secs',
            yes_no_boolean      => '_zapi_translate_yes_no',
            cron_time_integer   => '_zapi_translate_cron_time_integer',
        );
        while (my ($key, $arrayref) = each %$zapi_field_translations) {
            my $translation_method = $known_translations{$key};
            if (!defined $translation_method) {
                $Log->exit() if $may_exit;
                NATE::BaseException->throw("$key not supported");
            }
            $pkg_or_obj->$translation_method(
                api_opts             => $opts_ptr,
                options_to_translate => $arrayref
            );
        }
    }
    $Log->exit() if $may_exit;
}

sub _nodescope_options_validate_spec {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = $_[0];
    my %nodename_spec = (_optional_scalars('determine_appropriate_nodescope_node'));
    if ($pkg_or_obj->isa('NACL::STask::STask')) {
        $nodename_spec{'determine_appropriate_nodescope_node'}{default} = 1;
    }
    my %hash = (
        %nodename_spec,
        _optional_scalars('nodescope-node-name')
    );
    $Log->exit() if $may_exit;
    return wantarray ? %hash : {%hash};
}

# I'm currently placing these two methods (_convert_multiple_to_bytes;
# _strip_percentage_from_end_for_multiple) in Component/ONTAP.pm since this
# is currently intended to handle CLI vs ZAPI discrepancies in ONTAP
# components. This can be moved up to Component.pm if deemed needed by
# non-ONTAP components.

# The convention is for these methods to be named "_zapi_translate_*".

# CLI accepts value with "%", but ZAPI accepts it only as an integer.
# Example: "50%" -> "50"
sub _zapi_translate_percentage {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_zapi_translate_val(@args,
        translation_method => '_zapi_percentage_translation_code');
    $Log->exit() if $may_exit;
}

sub _zapi_percentage_translation_code {
    my ($pkg_or_obj, $val) = @_;
    $val =~ s/%$//;
    return $val;
}

# CLI accepts value with size suffixes, but ZAPI accepts it only as an integer.
sub _zapi_translate_to_bytes {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_zapi_translate_val(@args,
        translation_method => '_zapi_bytes_translation_code');
    $Log->exit() if $may_exit;
}

sub _zapi_bytes_translation_code {
    my ($pkg_or_obj, $val) = @_;
    return $pkg_or_obj->convert_to_bytes($val);
}

# Convert on/off -> true/false
sub _zapi_translate_on_off {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_zapi_translate_val(@args,
        translation_method => '_zapi_on_off_translation_code');
    $Log->exit() if $may_exit;
}

sub _zapi_on_off_translation_code {
    my ($pkg_or_obj, $val) = @_;
    if ($val eq 'on') {
        return 'true';
    } elsif ($val eq 'off') {
        return 'false';
    } else {
        return $val;
    }
}

# Convert enable/disable -> true/false
sub _zapi_translate_enable_disable {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_zapi_translate_val(@args,
        translation_method => '_zapi_enable_disable_translation_code');
    $Log->exit() if $may_exit;
}

sub _zapi_enable_disable_translation_code {
    my ($pkg_or_obj, $val) = @_;
    if ($val =~ /enable/) {
        return 'true';
    } elsif ($val =~ /disable/) {
        return 'false';
    } else {
        return $val;
    }
}
# Convert yes/no -> true/false
sub _zapi_translate_yes_no {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_zapi_translate_val(@args,
        translation_method => '_zapi_yes_no_translation_code');
    $Log->exit() if $may_exit;
}

sub _zapi_yes_no_translation_code {
    my ($pkg_or_obj, $val) = @_;
    if ($val eq 'yes') {
        return 'true';
    } elsif ($val eq 'no') {
        return 'false';
    } else {
        return $val;
    }
}

sub _zapi_translate_cron_time_integer {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_zapi_translate_val(@args,
        translation_method => '_zapi_cron_time_integer_translation_code');
    $Log->exit() if $may_exit;
}

sub _zapi_cron_time_integer_translation_code {
    my ($pkg_or_obj, $val) = @_;
    my @new_val = @$val;
    map {s/^all$/-1/} @new_val;
    return \@new_val;
}

sub _zapi_translate_octal_decimal {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %args) = @_;
    $pkg_or_obj->_zapi_translate_val(%args,
        translation_method => '_zapi_octal_decimal_translation_code',);
    $Log->exit() if $may_exit;
}

sub _zapi_octal_decimal_translation_code {
    my ($pkg_or_obj, $val) = @_;
    return oct($val);
}

sub _zapi_translate_hex_decimal {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %args) = @_;
    $pkg_or_obj->_zapi_translate_val(%args,
        translation_method => '_zapi_hex_decimal_translation_code',);
    $Log->exit() if $may_exit;
}

sub _zapi_hex_decimal_translation_code {
    my ($pkg_or_obj, $val) = @_;
    return hex($val);
}

sub _zapi_translate_string_to_timestamp {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    state $spec = {
        api_opts             => {type => HASHREF},
        options_to_translate => {type => ARRAYREF},
    };
    my %opts = validate_with(
        params => \@args,
        spec   => $spec,
    );
    my $timezone;
    my $api_opts = $opts{api_opts};
    foreach my $translate_option (@{$opts{options_to_translate}}) {
        my $current_val = $api_opts->{$translate_option};
        if (defined $current_val) {
            if ( !defined $timezone ) {
                try {
                    $timezone = $api_opts->{command_interface}->get_timezone(
                    apiset_should => {interface => 'ZAPI'});
                } catch NATE::Exceptions::Argument with {
                    my $exception = $_[0];
                    $Log->warn('Attempt to translate CLI-like timestring to timestamp'
                      . "Reason:\n"
                      .$exception->text()
                     );
                };
                if ( !$timezone ) {
                    $Log->exit() if $may_exit;
                    return;
                }
            }
            $current_val =~ s/^\s+//g;
            $current_val =~ s/\s+$//g;
	#Skip Unix-timestamp conversion if value already in Unix-timestamp format
        if ( $current_val !~ /^[0-9]+$/ ){
            my ($ss,$mm,$hh,$day,$month,$year) = strptime($current_val);
            $month = $month + 1;
            $year = $year + 1900;
            $api_opts->{$translate_option} =
                $pkg_or_obj->convert_to_unix_time(
                    month  => $month,
                    day    => $day,
                    year   => $year,
                    hour   => $hh,
                    minute => $mm,
                    second => $ss,
                    timezone => $timezone,
                    interface => 'ZAPI'
            );
        }
        }
    }
    $Log->exit() if $may_exit;
}

sub _zapi_translate_hrminsec_to_secs {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %args) = @_;
    $pkg_or_obj->_zapi_translate_val(%args,
        translation_method => '_zapi_translate_hrminsec_secs',);
    $Log->exit() if $may_exit;
}

sub _zapi_translate_hrminsec_secs {
    my ($pkg_or_obj, $val) = @_;
    $Log->enter() if $may_enter;

    my $secs;
    if ( $val =~ /^\d+$/ ) {
        $secs = $val;
    } else {
        my %conversion = ('h' => 3600, 'm' => 60, 's' => 1);
        while ( $val =~ s/(\d+)(\w)// ) {
            $secs+= $1 * $conversion{$2};
        }
    }
    $Log->exit() if $may_exit;
    return $secs;
}

# All the "_zapi_translate_*" methods are wrappers around this
sub _zapi_translate_val {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    state $spec = {
        api_opts             => {type => HASHREF},
        options_to_translate => {type => ARRAYREF},
        translation_method   => {type => SCALAR},
    };
    my %opts = validate_with(
        params => \@args,
        spec   => $spec,
    );
    my $api_opts = $opts{api_opts};
    my $translation_method = $opts{translation_method};
    foreach my $translate_option (@{$opts{options_to_translate}}) {
        my $current_val = $api_opts->{$translate_option};
        if (defined $current_val) {

            $api_opts->{$translate_option} =
                $pkg_or_obj->$translation_method($current_val);
        }
    }
    $Log->exit() if $may_exit;
}

=head2 create_mcc_verification

    $command_interface->
    NACL::STask::VserverNfs->create_mcc_verification(
        command_interface => $ci,
        %primary_keys_of_Stask_vservernfs,
        remote_interface => $node, #not mandatory
    );

    (or)

    $pkg_or_obj->create_mcc_verification();

(Class or Instance call) This method will perform the verification
on the MCC cluster's partner node on whether the created object is replicated.
If the replication is not visible, the method throws <NATE::BaseException>
notifying that the replication has been failed.
Note that this method should be used when automating for MCC environment.
Before running this method please check whether the nodes are MCC clusters
using the following call,

    $command_interface->is_in_mcc_environment();

=over

=item Options

=over

=item (Required for class call) The primary keys of the STask or Component
needs to be passed for the class call. Please refer to the
_primary_keys_validate_spec() method of respective component.

=item C<< remote_interface => $node >>

(Optional) This a <NACL::C::CommandInterface::ONTAP> type of object of the remote
cluster or the partner cluster's node object in a MCC environment.

=item C<< "method-timeout" => $timeout >>

(Optional) This value contains the timeout in seconds in order to wait for the
replication to be done on the partner MCC node. Default : 120 secs.

=item C<< "nacltask_polling_interval" => $interval >>

(Optional) This value contains the interval between each polling for the MCC check.
Default: 10 secs.

=item command_interface, apiset_must, apiset_should, method-timeout, etc.

All of the other various options needs to verfied with the POD of the
respective Task or Component library.

=over

=item Exceptions

=over

=item C<NACL::Exceptions::VerifyFailure>

This type of exception is thrown when replication operation is not found on the
MCC partner cluster.

=back

=back

=cut

sub create_mcc_verification {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,%args) = @_;
    my $return_obj = $pkg_or_obj->_create_modify_mcc_verification(
        %args,
    );
    $Log->exit() if $may_exit;
    return $return_obj;
}

=head2 modify_mcc_verification

    $command_interface->
    NACL::STask::VserverNfs->modify_mcc_verification(
        command_interface => $ci,
        vserver           => $vserver,
        opts_to_find      => { 'access' => 'false' },
        remote_interface  => $node, #not mandatory
    );

    (or)

    $pkg_or_obj->modify_mcc_verification(%opts);

(Class or Instance call) This method will perform the verification
on the MCC cluster's partner node on whether the modified object is replicated.
If the replication is not visible, the method throws <NATE::BaseException>
notifying that the replication has been failed.
Note that this method should be used when automating for MCC environment.
Before running this method please check whether the nodes are MCC clusters
using the following call,

    $command_interface->is_in_mcc_environment();

=over

=item Options

=over

=item (Required for class call) The primary keys of the STask or Component
needs to be passed for the class call. Please refer to the
_primary_keys_validate_spec() method of respective component.

=item (Required) The option which needs to be verified on the MCC partner
cluster for that particular component or stask.

For eg., the 'access' parameter has been modified to false for the <NACL::STask::VserverNfs>
component, and if this parameter needs to verified on the partner cluster then
the call should be made as,

    NACL::STask::VsererNfs(
        command_interface => $node,
        vserver           => $vserver, #primary key
        access            => 'false',
    );

    (or)

    $VserverNfs_stask_obj->modify_mcc_verification(access => 'false');

For the using the applicable options, please check the respective Component
library.

=item C<< remote_interface => $node >>

(Optional) This a <NACL::C::CommandInterface::ONTAP> type of object of the remote
cluster or the partner cluster's node object in a MCC environment.

=item C<< "method-timeout" => $timeout >>

(Optional) This value contains the timeout in seconds in order to wait for the
replication to be done on the partner MCC node. Default : 120 secs.

=item C<< "nacltask_polling_interval" => $interval >>

(Optional) This value contains the interval between each polling for the MCC check.
Default: 10 secs.

=item command_interface, apiset_must, apiset_should, method-timeout, etc.

All of the other various options needs to verfied with the POD of the
respective Task or Component library.

=over

=item Exceptions

=over

=item C<NACL::Exceptions::VerifyFailure>

This type of exception is thrown when replication operation is not found on the
MCC partner cluster.

=back

=back

=cut

sub modify_mcc_verification {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,%args) = @_;
    my $return_obj = $pkg_or_obj->_create_modify_mcc_verification(
        %args,
    );
    $Log->exit() if $may_exit;
    return $return_obj;
}

sub _create_modify_mcc_verification {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    state $spec = {
            nacltask_polling_interval => { type => SCALAR, default  => 10},
            "method-timeout"          => { type => SCALAR, default  => 120},
            remote_interface          => { type => OBJECT, optional => 1},
    };
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => $spec,
        allow_extra     => 1,
    );

    my ($partner_ci,%common_params,%other_opts);
    my $ci        = $opts{command_interface};
    my $end_time  = timeout2time(delete $opts{'method-timeout'});
    my $poll_time = delete $opts{nacltask_polling_interval};

    $pkg_or_obj->_move_common_component_params(
        source => \%opts,
        target => \%common_params
    );

    if(defined $opts{remote_interface}) {
        $partner_ci = delete $opts{remote_interface};
    } else {
        $partner_ci = $ci->get_remote_interface_mcc( %common_params );
    }

    $pkg_or_obj->_update_opts_for_mcc(
        command_interface => $ci,
        options_hash_ref  => \%opts,
        %common_params,
    );

    my $primary_hash = $pkg_or_obj->get_primary_keys_options(%opts);

    my $CS_pkg_name = $pkg_or_obj->get_CS_package_name();

    while(1) {
        my $obj_found = $CS_pkg_name->fetch(
            command_interface => $partner_ci,
            filter            => $primary_hash,
            requested_fields  => [keys %opts],
            allow_empty       => 1
        );
        my ($replicated, $exception);
        if (defined $obj_found) {
            try {
                delete $opts{command_interface};
                $obj_found->verify_fields(%opts);
                $replicated = 1;
            } catch NACL::Exceptions::VerifyFailure with {
                $exception = $_[0];
            };
            last if $replicated;
        }
        if (time > $end_time) {
            if ($exception) {
                $Log->exit() if $may_exit;
                $exception->throw();
            } else {
                $Log->exit() if $may_exit;
                NACL::Exceptions::VerifyFailure->throw('Entry not showing up on '
                    . 'partner cluster');
            }
        } else {
            Tharn::snooze($poll_time);
        }
    }

    $Log->exit() if $may_exit;
}

=head2 delete_mcc_verification

    $command_interface->
    NACL::STask::VserverNfs->delete_mcc_verification(
        command_interface => $node,
        remote_interface  => $partner_node, #not mandatory
    );

    (or)

    $pkg_or_obj->delete_mcc_verification();

(Class or Instance call) This method will perform the verification
on the MCC cluster's partner node on whether the deleted object is replicated.
If the replication is not visible, the method throws <NATE::BaseException>
notifying that the replication has been failed.
Note that this method should be used when automating for MCC environment.
Before running this method please check whether the nodes are MCC clusters
using the following call,

    $command_interface->is_in_mcc_environment();

=over

=item Options

=over

=item (Required for class call) The primary keys of the STask or Component
needs to be passed for the class call. Please refer to the
_primary_keys_validate_spec() method of respective component.

=item C<< remote_interface => $node >>

(Optional) This a <NACL::C::CommandInterface::ONTAP> type of object of the remote
cluster or the partner cluster's node object in a MCC environment.

=item C<< "method-timeout" => $timeout >>

(Optional) This value contains the timeout in seconds in order to wait for the
replication to be done on the partner MCC node.

=item C<< "nacltask_polling_interval" => $interval >>

(Optional) This value contains the interval between each polling for the MCC check.
Default: 10 secs.

=item command_interface, apiset_must, apiset_should, method-timeout, etc.

All of the other various options needs to verfied with the POD of the
respective Task or Component library.

=over

=item Exceptions

=over

=item C<NACL::Exceptions::VerifyFailure>

This type of exception is thrown when replication operation is not found on the
MCC partner cluster.

=back

=back

=cut

sub delete_mcc_verification {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;

    state $spec = {
            nacltask_polling_interval => { type => SCALAR, default  => 10},
            "method-timeout"          => { type => SCALAR, default  => 120},
            remote_interface          => { type => OBJECT, optional => 1},
    };

    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => $spec,
    );

    my ($partner_ci,%common_params);
    my $ci        = $opts{command_interface};
    my $end_time  = timeout2time(delete $opts{'method-timeout'});
    my $poll_time = delete $opts{nacltask_polling_interval};

    $pkg_or_obj->_copy_common_component_params(
        source => \%opts,
        target => \%common_params
    );

    #fetch the primary keys for the STask/Component
    my $primary_hash   = $pkg_or_obj->get_primary_keys_options(%opts);
    my %primary_hash   = %$primary_hash;
    if(!defined $opts{remote_interface}) {
        $partner_ci = $ci->get_remote_interface_mcc( %common_params );
    } else {
        $partner_ci = delete $opts{remote_interface};
    }

    $pkg_or_obj->_update_opts_for_mcc(
        command_interface => $ci,
        options_hash_ref  => \%primary_hash,
        %common_params,
    );

    while(1) {
        my $obj_found = $pkg_or_obj->get_package_name()->find(
            command_interface => $partner_ci,
            filter            => { %primary_hash },
            %common_params,
            allow_empty => 1,
        );
        last if !$obj_found;
        if( time() > $end_time ) {
            $Log->exit() if $may_exit;
            NACL::Exceptions::VerifyFailure
                ->throw("Delete operation failed to replicate in".
                        " the metrocluster's partner");
        }
        Tharn::snooze($poll_time);
    }
    $Log->exit() if $may_exit;
}

sub _update_opts_for_mcc {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = validate_with(
        params          => \@args,
        spec => {
            command_interface => { type => OBJECT,
                isa => 'NACL::C::CommandInterface::ONTAP', },
            options_hash_ref  => { type => HASHREF },
        },
        allow_extra => 1,
    );

    my $opts_hash_ref = $opts{options_hash_ref};

    if( $opts_hash_ref->{vserver} ) {
        $opts_hash_ref->{vserver} = $opts_hash_ref->{vserver}."-mc";
    } elsif ( $opts_hash_ref->{aggregate} ) {
        my $ci = $opts{command_interface};
        my %common_params;
        $pkg_or_obj->_copy_common_component_params(
            source => \%opts,
            target => \%common_params
        );
        my $mcc_configured   = NACL::CS::ClusterIdentity->fetch(
            command_interface => $ci,
            requested_fields => [ qw( name ) ],
            %common_params,
        );
        $opts_hash_ref->{cluster} = $mcc_configured->name();
    }
    $Log->exit() if $may_exit;
}

sub _fallback_for_relational_regex {
     $Log->enter() if $may_enter;

     my ($pkg_or_obj, %opts) = @_;
     my @options = @{ delete $opts{fields_to_check} };
     foreach my $field (@options) {
         if (exists $opts{$field}) {
             if ($pkg_or_obj->_check_relational_regex_filter(filter_value => $opts{$field})) {
                 # field has special characters
                 $Log->exit() if $may_exit;
                 NACL::Exceptions::InvalidChoice->throw('Since the '
                    . "$field provided has special characters "
                    . "('$opts{$field}'), the ZAPI back-end cannot be used"
                 );
             }
         }
    }
    $Log->exit() if $may_exit;
}

# Running in the foreground is possible only through CLI, so if
# foreground => true is passed, then do not go through ZAPI
sub _foreground_zapi_check {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    if (defined $opts{foreground} && $opts{foreground} eq 'true') {
        $Log->exit() if $may_exit;
        NACL::Exceptions::InvalidChoice->throw('Waiting in the foreground '
                . "is possible only through CLI (foreground => 'true' was "
                . 'passed), therefore ZAPI cannot be used');
    }

    $Log->exit() if $may_exit;
}

sub _update_mapping_translation {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts = validate_with(
        params => \@args,
        spec   => {
            orig_args => {type => HASHREF, default => {}},
            known_backend_fields => {type => ARRAYREF, default => []},
            inapplicable_for_help_xml => {type => ARRAYREF, default => []},
            disable_help_xml_lookup => {type => SCALAR, default => 0}
        },
        allow_extra => 1,
    );

    if ($opts{disable_help_xml_lookup}) {
        $Log->exit() if $may_exit;
        return;
    }

    my $orig_args = $opts{orig_args};
    my %copy_opts = %$orig_args;
    my $known_backend_fields = $opts{known_backend_fields};

    my $copy = delete $copy_opts{copy};
    my $map  = delete $copy_opts{map};
    my $command_interface = delete $copy_opts{command_interface};
    my $realm = NACL::Realm->realm();
    my $xml_obj = $realm->get_help_xml_obj( command_interface => $command_interface );
    my $iter_helper = 0;
    # Check whether help_xml is accessible
    if ( $xml_obj->help_xml_accessible() ) {
        my (%cli_zapi, $cmd);
        if ( $copy_opts{show_cmd} ) {
           $cmd = delete $copy_opts{show_cmd};
           $iter_helper = 1;
        } elsif ( $copy_opts{cli_cmd} ) {
           $cmd = delete $copy_opts{cli_cmd};
        }

	#
        # The Zapi which runs on vserver context wont take vserver
        # as an input,but for certain zapis (snmp-comunity-add) has
        # vserver listed in the smf table . The following code handles
        # such issues.
        #
        if ( $copy_opts{vserver}
            && $copy_opts{vserver_operated_on}  ) {
            delete $copy_opts{vserver};
        }
        my $inapplicable_for_help_xml = delete $copy_opts{inapplicable_for_help_xml};
        $pkg_or_obj->_remove_opts_inapplicable_for_help_xml(
            inapplicable_for_help_xml => $inapplicable_for_help_xml,
            input                => \%copy_opts,
            known_backend_fields => $known_backend_fields,
        );

        map { $cli_zapi{$_} = 1 } @{ $copy };
        %cli_zapi = ( %cli_zapi, %{ $map } );
        map { delete $copy_opts{$_} } keys %cli_zapi;


        if ( keys %copy_opts ) {
           my @new = keys %copy_opts;
           my $alias_map;
           if ( $command_interface->apiset(check_apiset_already_loaded => 1) ) {
               my $api_cmd = $cmd;
               $api_cmd =~ s/\s+/_/g;
               $api_cmd =~ s/-/_/g;
               my $cdef = $command_interface->apiset()->get_command_definition(command => $api_cmd);
               $alias_map = $cdef->get_alias_map();
           }
           my $mapping_info  = $xml_obj->get_zapi_field_mapping(command => $cmd,
                      new_fields   => \@new,
                      mapping_as_ds => 1,
                      alias_map => $alias_map);
           if ( defined $mapping_info ) {
              %{ $map } = ( %{ $mapping_info }, %{ $map } );
              $orig_args->{map} = $map;
           }
           #Disabling the code to update value translations for iter_helper calls
           #because not all the value translations for CS are applicable for C methods
           if ( !$iter_helper ) {
               my $translation  = $xml_obj->get_zapi_translation(command => $cmd,
                                                                 alias_map => $alias_map);
               my $translation_from_comp =  $copy_opts{zapi_field_translations};
               #Check if there are any translations from help_xml, else nothing to do
               if ( $translation ) {
                   if ( $translation_from_comp ) {
                       foreach my $key ( keys %{ $translation } ) {
                           if ( !$translation_from_comp->{$key} ) {
                               $translation_from_comp->{$key} = $translation->{$key};
                           } else {
                               my @union = uniq( @{ $translation->{$key} }, @{ $translation_from_comp->{$key} } );
                               $translation_from_comp->{$key} = \@union;
                           }
                       }
                       $orig_args->{zapi_field_translations} = $translation_from_comp;
                   } else {
                       $orig_args->{zapi_field_translations} = $translation;
                   }
              }
           }
       }
    }
    $Log->exit() if $may_exit;
}



sub _remove_opts_inapplicable_for_help_xml {
    my ($pkg_or_obj, %opts) = @_;

    # Remove fields from %opts which are not passed down to the command.
    # This includes the common options (command_interface, apiset_should etc),
    # fields accepted by the respective base back-end method
    # ("known_backend_fields") and fields which the back-end specify
    # should not be considered for help_xml translation
    my @common_keys = (keys %{$pkg_or_obj->_common_validate_spec()},
        keys %{$pkg_or_obj->_additional_common_validate_spec()});
    my @fields_to_delete = (@{$opts{inapplicable_for_help_xml}},
        @{$opts{known_backend_fields}}, @common_keys);
    map { delete $opts{input}->{$_} } @fields_to_delete;
}

sub base_class { return __PACKAGE__; }

=head2 _nodescope_frontend_helper

    # Using NACL::C::Mkfile as an example
    sub mkfile {
        $Log->enter() if $may_enter;
        my $pkg = shift;
        $pkg->_frontend_log_debug_opts(@_);
        $pkg->_nodescope_frontend_helper(
            @_,
            command => 'mkfile',
            choices => [
                {   method    => '_mkfile_7mode_cli',
                    interface => 'CLI',
                    set       => '7Mode|Nodescope',
                },
            ],
        );
        $Log->exit() if $may_exit;
    }

For components which have only a single method for 7Mode/Nodescope
support for a command, this helper can be used. Rather than hard-code
the options accepted by the command, this helper method automatically
determines which options are accepted (by reading the CDEFs). In addition
to the options directly accepted by the API call, 7Mode/Nodescope component
methods support the following options:

=over

=item nodescope-node-name

The node-name on which to run the Nodescope command.

=item determine_appropriate_nodescope_node

If set to 1, the back-end implementation determines which node the
element is on and runs the command in the context of that Node.

=back

This method constructs the front-end validation spec as a combination
of those options accepted by the API call, along with the above two
special options. In addition, this method also checks if the command
interface supplied can be used for Nodescope operations. (see
L<_can_ci_be_used_for_nodescope>. It also finds the node in which
the element is present, if 'determine_appropriate_nodescope_node option' is
set to 1. This method expects the discovery logic be implemented in
the subroutine whose name is specified through 'node_name_discovery_method'
option.

=over

=item Options

=over

=item C<< command => $command >>

The name of the APISet method which will get invoked in the back-end.

=item C<< choices => $choices >>

The choices array-reference which would normally have been passed in the
call to call_on_apiset.

=item C<< common_validate_opts => $opts_hashref >>

(Optional)
The other customized options like 'ignore-primary-keys' to be passed
in the call to _common_validate_with

=item C<< node_name_discovery_method => $method_name >>

(Optional)
The name of the method to be invoked for finding the node name.

=back

=back

=cut

sub _nodescope_frontend_helper {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = validate_with(
        params => \@_,
        spec   => {
            command                    => { type => SCALAR },
            choices                    => { type => ARRAYREF },
            common_validate_opts       => { type => HASHREF, default => {} },
            node_name_discovery_method => { type => SCALAR, optional => 1 },
        },

        # The other options are those passed by the caller
        allow_extra => 1
    );
    my $command              = delete $opts{command};
    my $choices              = delete $opts{choices};
    my $common_validate_opts = delete $opts{common_validate_opts};
    my %node_args            = ();
    $pkg->_hash_move(
        source => \%opts,
        target => \%node_args,
        map    => { node_name_discovery_method => 'method' }
    );

    # Check that the command_interface is present in the options
    # passed by the caller.
    validate_with(
        params => \%opts,
        spec   => {
            command_interface =>
                { type => OBJECT, isa => 'NACL::C::CommandInterface::ONTAP' }
        },
        allow_extra => 1
    );

    # Determine the options accepted by the command
    my $apiset = $pkg->_can_ci_be_used_for_nodescope(
        command_interface => $opts{command_interface} );
    my @allowed_options
        = $apiset->get_allowed_options_for_command( command => $command );

    # Add the options accepted for Nodescope calls
    push @allowed_options,
        (
        'nodescope-node-name', 'determine_appropriate_nodescope_node',
        'vserver'
        );
    my %final_opts = $pkg->_common_validate_with(
        params          => [%opts],
        additional_spec => { _optional_scalars(@allowed_options) },
        %$common_validate_opts
    );

    $final_opts{'nodescope-node-name'}
        = $pkg->_get_nodescope_node_name( %opts, %node_args );

    my $ret = $pkg->call_on_apiset( %final_opts, choices => $choices );
    $Log->exit() if $may_exit;
    return $ret;
}

=head2 _nodescope_builder

    __PACKAGE__->_nodescope_builder(command => $cmd, method => $method);

Provide a simple way to install methods at the Component-level that invoke
Node-scoped commands. This method takes care of installing the front-end method
and back-end methods and defaults to determining the appropriate
nodescope-node-name.

For example, installing a method that runs the "zeroblock" command for
NACL::C::StorageDisk can be done like:

    __PACKAGE__->_nodescope_builder(command => 'zeroblock');

=over

=item Options

=over

=item C<< command => $cmd >>

(Mandatory)
The name of the APISet command being invoked.

=item C<< method => $method >>

(Optional)
The name of the method to install in the Component. By default, if this is
not provided, this will be the same as C<command>.

=item C<< map => $hashref >>

(Optional)
If the name of the primary key as accepted by the Component does not match
the name of the option defined in the APISet CDEF, then this option can be
used to specify the mapping of <component_name> -> <apiset_name>. Example:
for command "foo", the APISet accepts the option as "volume-name" but the
Component accepts it as "volume":

    __PACKAGE__->_nodescope_builder(command => 'foo', map => {volume => 'volume-name'});

=cut

sub _nodescope_builder {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    my %opts = validate_with(
        params => \@args,
        spec   => {
            method  => {type => SCALAR, optional => 1},
            command => {type => SCALAR},
            map     => {type => HASHREF, default => {}},
        },
    );
    my $command = $opts{command};
    my $method  = $opts{method} // $command;
    my $map     = $opts{map};
    my $backend_method = "_${method}_7m_cli";
    no strict 'refs';

    # Build front-end method
    *{"${pkg}::$method"} = sub {
        $Log->enter() if $may_enter;
        my ($pkg_or_obj_in_sub, %args_in_sub) = @_;
        $args_in_sub{command_interface} //= $pkg_or_obj_in_sub->command_interface();
        my %nodescope_args;
        if ($pkg_or_obj_in_sub->can('get_local_node_name')) {
            # Add determine_appropriate_nodescope_node only if
            # get_local_node_name() is defined for that package.
            $nodescope_args{determine_appropriate_nodescope_node} = 1;
        }
        my $parsed_output = $pkg_or_obj_in_sub->_nodescope_frontend_helper(
            %nodescope_args,
            command => $command,
            choices => [
                {
                    method    => $backend_method,
                    interface => 'CLI',
                    set       => '7Mode|Nodescope',
                }
            ],
            %args_in_sub,
        );
        $Log->exit() if $may_exit;
        return $parsed_output;
    };

    # Build back-end method
    *{"${pkg}::$backend_method"} = sub {
        $Log->enter() if $may_enter;
        my ($pkg_or_obj_backend, @args_backend) = @_;
        my %opts_backend = $pkg_or_obj_backend->_validate_backend_opts(@args_backend);
        my $apiset = $opts_backend{apiset};

        my %apiset_args;
        my @allowed_options
            = $apiset->get_allowed_options_for_command(command => $command);
        push @allowed_options, 'nodescope-node-name';

        $pkg_or_obj_backend->_hash_copy(
            source => \%opts_backend,
            target => \%apiset_args,
            copy   => \@allowed_options,
            map    => $map,
        );

        my $parsed_output = $apiset->$command(%apiset_args)->get_parsed_output();
        $Log->exit() if $may_exit;
        return $parsed_output;
    };
    use strict 'refs';
    $Log->exit() if $may_exit;
}

=head2 get_local_node_obj

    my $node_obj = $vol_obj->get_local_node_obj();

Returns a NACL::STask::Node object for the node hosting that particular
element.

=cut

sub get_local_node_obj {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    if ($pkg_or_obj->can('get_local_node_name')) {
        my $node_name = $pkg_or_obj->get_local_node_name(%opts);
        require NACL::STask::Node;
        my $node_obj = NACL::STask::Node->new(node => $node_name);
        $Log->exit() if $may_exit;
        return $node_obj;
    } else {
        $Log->exit() if $may_exit;
        NACL::Exceptions::InvalidMethod->throw("Method 'get_local_node_obj' "
            . "is not applicable on package '"
            . $pkg_or_obj->get_package_name()
            . "' because 'get_local_node_name' is not defined for it"
        );
    }
    $Log->exit() if $may_exit;
}

sub _additional_common_validate_spec {
    my ($pkg_or_obj, @opts) = @_;
    my $spec = $pkg_or_obj->SUPER::_additional_common_validate_spec(@opts);
    my $ontap_additional_spec = $pkg_or_obj->_get_additional_common_validate_spec();
    $spec = {%$ontap_additional_spec, %$spec};
    return $spec;
}

sub _load_event_class {
    my $pkg = shift;
    $pkg =~ /::(.*$)/;
    my $event_name = "NACL::Events::C_STask::".$1;
    load_branch_events(events => [ $event_name ]);
}

1;
