
#
# Copyright (c) 2001-2012 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Represents a client machine with mgmt interface and IP address and
## has the DFM service running.
## @author madhav.shekar@netapp.com, dl-nacl-dev@netapp.com,
## dl-mpo-nacl-dev@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::C::DFM::DFMClient

=head1 DESCRIPTION

C<NACL::C::DFM::DFMClient> is a derived class of 
L<NACL::C::Client|lib-NACL-C-Client-pm>. It represents clients which have
the DFM service running on it. It can currently be used only as a command
interface for DFM components.

Unlike C<NACL-C-Client-pm>, which defaults to the Linux CLI APISet, 
C<NACL::C::DFM::DFMClient> defaults to the DFM ZAPI APISet. All other attributes
and methods can be inherited from C<NACL::C::Client>.

=head1 ATTRIBUTES

Same as L<NACL::C::Client|lib-NACL-Client-pm/ATTRIBUTES>.

=cut

package NACL::C::DFM::DFMClient;

use strict;
use warnings;

use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Params::Validate
  qw( validate_with BOOLEAN HASHREF UNDEF OBJECT ARRAYREF SCALAR);
use base 'NACL::C::Client';
use NATE::Exceptions::Argument qw(:try);
use NACL::Exceptions::OperationFailed qw(:try);
use Data::Dumper;

=head1 METHODS

=head2 find

    my $Client  = NACL::C::DFM::DFMClient->find(...);
    my @Clients = NACL::C::DFM::DFMClient->find(...); 

(Class method)
Finds all DFM Hosts and returns command interface objects for each (can also
be called in scalar context, in which case it returns the first DFM host 
found).
Refer to L<find|lib-NACL-C-CommandInterface-pm/find> for more details.

You can also refer to user guide for more information on the options,
order of precedence for command line parameters (in case of find() is
invoked without specifying a command_interface) and examples.
http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/ComponentLayerUsersGuide/index.html#Obtaining_command_interface_objects

Unlike other command interfaces, you cannot call 'find' of DFMHost with
a command interface.

=head2 default_apiset_could

The ruleset to default to if values for the co-ordinates 'category',
'interface' or 'set' cannot be determined.

If DFM components used C<NACL::C::Client> as their command interface,
unless an apiset_must value for 'set' is defined as DFM (either as a test
parameter or in the tharnhost file for the hosts used), it will default to
use set as "Linux". By using C<NACL::C::DFM::DFMClient> instead, the set will
default to "DFM" without needing to make this setting for all tests/tharnhost
files.

Refer to L<default_apiset_could|lib-NACL-C-CommandInterface-pm/default_apiset_could>

=cut

sub default_apiset_could {
    return {
        category  => "Host",
        interface => "ZAPI",
        set       => "DFM"
    };
}

=head2 default_apiset_must

All DFM clients have to be of category "Host" and set "DFM", so this method 
returns a hash with keys 'category' whose value is "Host", 'set' whose
value is "DFM".

Refer to L<default_apiset_must|lib-NACL-C-CommandInterface-pm/default_apiset_must>.

=cut

sub default_apiset_must {
    my $self     = shift;
    my $hostrec  = $self->hostrec();
    my $hostid   = $hostrec->id();
    my $hosttype = $hostrec->hosttype();
    if ( $hosttype !~ /(?:unix|windows)/ ) {
        NATE::BaseException->throw( "Host $hostid was supposed "
              . "to be a DFM client but is actually of type '$hosttype'" );
    }
    return {
        category        => "Host",
        category_source => "command interface is a NACL::C::DFM::DFMClient",
        set             => "DFM",
        set_source      => "command interface is a NACL::C::DFM::DFMClient",
    };
}

=head2 _create_version_manager

   sub get_version_manager {
        ...
        if (keys %$apisets) {
            ...
        } else {
            $version_manager = $self->_create_version_manager();
        }
        return $version_manager;
    }
    
(Instance method)
This private method returns the version manager of the appropriate type.
This decision is based on the category ("Host" for DFM Client command interfaces)
and the set ("DFM").
    
=cut

sub _create_version_manager {
    $Log->enter() if $may_enter;
    my $self = shift;

    use NACL::APISet::Version;
    my $hostrec    = $self->hostrec();
    my $category   = "Host";
    my $shell_type = "DFM";
    $Log->debug( "Version manager object with parameters "
          . "category = $category; shell_type = $shell_type; "
          . "hostobj type = "
          . $hostrec->hosttype()
          . " created" );
    $Log->exit() if $may_exit;
    return NACL::APISet::Version->new(
        category   => $category,
        hostobj    => $hostrec,
        shell_type => $shell_type
    );
}

=head2 restart 

     $client->restart( ... ); 

(Instance method) 
Restarts the BARD server.

=over

=item Options

=item C<< wait_time => $scalar >>

(Optional, defaults to 300 if not provided )
This is the amount of time (in SECONDS)  for which the method waits for
the Bard server to get refreshed.

=item C<< polling_interval => $scalar >>

(Optional, defaults to 60 if not provided )
This is the interval (in SECONDS) , for which periodically a check is done
to see if the server is running.

=over

=back

=cut

#
# Define constants
#
use constant DEFAULT_WAIT_TIME        => 420;
use constant DEFAULT_POLLING_INTERVAL => 60;

sub restart {

    $Log->enter() if $may_enter;

    my $pkg = shift;

    my %opts = validate_with(
        params => \@_,
        spec   => {
            wait_time => { type => SCALAR, default => DEFAULT_WAIT_TIME, },
            polling_interval =>
              { type => SCALAR, default => DEFAULT_POLLING_INTERVAL, },
            username => { type => SCALAR | UNDEF, optional => 1 },
            password => { type => SCALAR | UNDEF, optional => 1 },
        },
    );

    my $wait_time        = $opts{'wait_time'};
    my $polling_interval = $opts{'polling_interval'};

    # Commands to restart the service
    my $ocie_cmd      = 'service ocie restart';
    my $ocieau_cmd    = 'service ocieau start';
    my $ocie_status   = 'service ocie status';
    my $ocieau_status = 'service ocieau status';

    # Global Variable
    my ($conn, $response, $sub_ref);
    # Assign the client
    my $Client = $pkg;

    # Create APISet object for CLI
    my %input_hash = ( interface => "CLI", );

    $input_hash{username} = $opts{'username'} if ( defined $opts{'username'} );
    $input_hash{password} = $opts{'password'} if ( defined $opts{'password'} );

    $Log->debug("Parameters for creating the APISet object: \%input_hash");

    my $apiset = $Client->apiset(%input_hash);

    $Log->debug("Created APISet object.");

  # Create Conn obj required for execute_start and execute_expect function calls
    $conn = $apiset->get_connection();

    # Create subrotuine ref which is passed as a input for on_timeout param
    $sub_ref = sub {

        $apiset->api_reconnect_session() if !$apiset->api_is_session_alive();

        $response = $apiset->execute_raw_command( command => $ocie_status );
        if ( $response !~ /ocie\sis\srunning/ ) {
            $Log->exit() if $may_exit;

            NACL::Exceptions::OperationFailed->throw(
                    "Failed to restart the server in the given time of "
                  . $wait_time
                  . "s. " );
        }
    };

    $conn->execute_start( command => $ocie_cmd );
    my $match_pattern = 'Restarting NetApp OnCommand Insight Essentials service: ocie';
    $conn->execute_expect(
        'timeout' => $wait_time,
        must_match => $match_pattern,
        on_timeout => $sub_ref
    );

    # Check the status
    my $exp_string = 'ocieau is running';
    $wait_time = $Client->scale_timeout( 'method-timeout' => $wait_time )
      ;    # Scale it accordingly
    my $start_time = time();
    my $end_time   = time() + $wait_time;

    do {
        Tharn::snooze $polling_interval;
        $response =
          $apiset->execute_raw_command( command => "$ocieau_cmd;$ocieau_status",
          );

      } while ( $response !~ /$exp_string/
        && time() < $end_time );

    # Check for time-out
    if ( $response !~ /$exp_string/ ) {

        $Log->exit() if $may_exit;

        NACL::Exceptions::OperationFailed->throw(
                "Failed to restart the server in the given time of "
              . $wait_time
              . "s. " );
    }

    $Log->exit() if $may_exit;

}

1;

