#
# Copyright (c) 2001-2010 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary DFM Role Component Module
## @author deepthis@netapp.com, dl-sm-nacl-dev@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::C::DFM::Role

=head1 DESCRIPTION

C<NACL::C::DFM::Role> is a derived class of
L<NACL::C::Component::DFM|lib-NACL-C-Component-DFM-pm>.  It represents moitoring
 a role on DFM.  A related class is
L<NACL::CS::DFM::Role|lib-NACL-CS-DFM-Role-pm>, which represents
the past state of an DFM Role.

=head1 ATTRIBUTES

=head2 command_interface

See L<NACL::C::Component::command_interface|lib-NACL-C-Component-pm/command_interface>

=head2 role

The name of the role, used to identify it uniquely. 

=cut

package NACL::C::DFM::Role;

use strict;
use warnings;
use Data::Dumper;

use NACL::CS::DFM::Role;
use NACL::ComponentUtils qw (_optional_scalars);
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Data::Dumper;

use base 'NACL::C::Component::DFM';
use Params::Validate qw(validate_with SCALAR);

=head1 METHODS

=head2 new

  my $role = NACL::C::DFM::Role->new(command_interface=>$ci, role => $role_name, 
                                      description => $description, owner-role => $owner-role );

Return a new Role component object with the given attributes.

=cut

use Class::MethodMaker [
    scalar => 'role',
];

=head2 

  NACL::C::DFM::Role->create( command_interface=>$ci, role => $role_name, 
                              description => $description, "owner-role" => $owner-role);

(Class method) Creates a new role on the DFM.
Uses a DFM-CLI APISet.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< role=>$role_name >>

(Required) Name of role added.

=item C<< description=>$description >>

(Optional) Description of the role.

= item C<< "owner-role"=>$owner_role >>

(Optional) The owner role of this role.

=back

=back

=cut

sub create {
    $Log->enter() if $may_enter;
	
    my $pkg  = shift;
	$Log->debug("Opts to create front end are:\n".Dumper(\@_));
    my %opts = $pkg->_common_validate_with(
        params => \@_,
        additional_spec =>
            { _optional_scalars(qw( description owner-role )), }
    );
	$Log->exit() if $may_exit;
    return $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => "_create_dfm_cli",
                interface => "CLI",
                set       => "DFM",
            }
        ],
    );
} ## end sub create

sub _create_dfm_cli {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );

    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};

    my $allowed_dfm_options = [qw( role description owner-role )];
    my $dfm_args            = $pkg->_hash_copy(
        source => \%opts,
        copy   => [qw( description owner-role )],
        map    => { "role" => "role-name" },
    );
    %opts = %$dfm_args;
    $Log->debug("Opts to API call 'dfm_role_create' are:\n".Dumper(\%opts));
	
    $apiset->dfm_role_create(%opts);
	
    $Log->exit() if $may_exit;
	
    # Return the package
    return $pkg->new(
        role              => $opts{'role-name'},
        command_interface => $command_interface
    );
} ## end sub _create_dfm_cli

=head2 add

  NACL::C::DFM::Role->add( command_interface=>$ci,
                           role => $role_name,
						   operation => $operation,
			               "resource-name" => $resource);
  or
  
  $role->add( ... );

(Class or instance method) Request addition of a new operation & resource pair to an existing role.
Uses a DFM-CLI APISet.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)See L<NACL::C::Component|lib::NACL::C::Component/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< role=>$role_name >>

(Required for class method, Not Applicable for instance method) Name of role added.

=item C<< operation=>$operation >>

Name of the operation to be performed on the resource.

=item C<< "resource-name"=>$resource >>

Name of the resource on which the operation is being performed.

=back

=back

=cut

sub add {

    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    $Log->debug("\nOpts to add front end are:\n".Dumper(\@_));
	
    my %opts = $pkg_or_obj->_common_validate_with(
        params => \@_,
        additional_spec =>
            { _optional_scalars(qw( operation resource-name )), }
    );
    $Log->exit() if $may_exit;
	
    return $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => "_add_dfm_cli",
                interface => "CLI",
                set       => "DFM",
            }
        ],
    );
} ## end sub add

sub _add_dfm_cli {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );

    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};

    my $dfm_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        map    => { "role" => "role-name" },
        copy   => [qw( operation resource-name)],
    );

    %opts = %$dfm_args;
    $Log->debug("Opts to API call 'dfm_role_add' are:\n".Dumper(\%opts));
	
    $apiset->dfm_role_add(%opts);
	$Log->exit() if $may_exit;

} ## end sub _add_dfm_cli

=head2 delete

  NACL::C::DFM::Role->delete( command_interface=>$ci,
                              role => $role_name,
                              operation => $operation,
			                  "resource-name" => $resource);
  or
  
  $role->delete( ... );

(Class or instance method) This method deletes a resource-operation pair associated with 
the role.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< operation=>$operation >>

Name of operation.

=item C<< "resource_name"=>$resource_name >>

Name of resource.

=back

=back

=cut

sub delete {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
	$Log->debug("Opts to delete front end are:\n".Dumper(\@_));
	
    my %opts       = $pkg_or_obj->_common_validate_with(
        params => \@_,
        additional_spec =>
            { _optional_scalars(qw( operation resource-name all )), }
    );
	$Log->exit() if $may_exit;
	
    return $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => "_delete_dfm_cli",
                interface => "CLI",
                set       => "DFM",
            }
        ],
    );
} ## end sub delete

sub _delete_dfm_cli {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );

    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};

    my $dfm_args;

    if ( defined( $opts{all} ) ) {

        $dfm_args = $pkg_or_obj->_hash_copy(
            source => \%opts,
            map    => { "role" => "role-name" },
            copy   => [qw( all)],
        );
    } else {

        $dfm_args = $pkg_or_obj->_hash_copy(
            source => \%opts,
            map    => { "role" => "role-name" },
            copy   => [qw( operation resource-name)],
        );
    }

    %opts = %$dfm_args;
    $Log->debug("Opts to API call 'dfm_role_delete' are:\n".Dumper(\%opts));
    $apiset->dfm_role_delete(%opts);
	
	$Log->exit() if $may_exit;

} ## end sub _delete_cmode_cli

=head2 modify

  NACL::C::DFM::Role->modify( command_interface=>$ci,
                              role => $role_name,
                              description => $description,
			                  "new-role-name" => $new_role);
  or
  
  $role->modify( ... );

(Class or instance method) Reset the name and/or description for a role.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< description=>$description >>

Description of the role.

=item C<< "new-role-name"=>$new_role >>

New name for the role.

=back

=back

=cut

sub modify {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
	$Log->debug("Opts to modify front end are:\n".Dumper(\@_));
	
    my %opts       = $pkg_or_obj->_common_validate_with(
        params => \@_,
        additional_spec =>
            { _optional_scalars(qw( description new-role-name )), }
    );
	$Log->exit() if $may_exit;
	
    return $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => "_modify_dfm_cli",
                interface => "CLI",
                set       => "DFM",
            }
        ],
    );
} ## end sub modify

sub _modify_dfm_cli {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );

    delete $opts{command_interface};
    my $apiset = delete $opts{apiset};

    my $dfm_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => [qw( description new-role-name )],
        map    => { "role" => "role-name" },
    );

    %opts = %$dfm_args;
    $Log->debug("Opts to API call 'dfm_role_modify' are:\n".Dumper(\%opts));
    $apiset->dfm_role_modify(%opts);

    # Important to change the value of the primary key
    $pkg_or_obj->role( $opts{'new-role-name'} ) if ( ref $pkg_or_obj );
	$Log->exit() if $may_exit;

} ## end sub _modify_dfm_cli

=head2 inherit

  NACL::C::DFM::Role->inherit( command_interface=>$ci,
                               role => $role_name,
                               "role-to-inherit-from" => $inherit_from);
  or
  
  $role->inherit( ... );

(Class or instance method) Allows a role to inherit capabilities
 from another role.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "role-to-inherit-from"=>$inherit_from >>

Name of role.

=back

=back

=cut

sub inherit {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
	$Log->debug("Opts to inherit front end are:\n".Dumper(\@_));
	
    my %opts       = $pkg_or_obj->_common_validate_with(
        params          => \@_,
        additional_spec => { _optional_scalars(qw( role-to-inherit-from )), }
    );
	$Log->exit() if $may_exit;
	
    return $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => "_inherit_dfm_cli",
                interface => "CLI",
                set       => "DFM",
            }
        ],
    );
} ## end sub inherit

sub _inherit_dfm_cli {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );

    delete $opts{command_interface};
    my $apiset = delete $opts{apiset};

    my $dfm_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => [qw( role-to-inherit-from )],
        map    => { "role" => "role-name" },
    );

    %opts = %$dfm_args;
    $Log->debug("Opts to API call 'dfm_role_inherit' are:\n".Dumper(\%opts));
    $apiset->dfm_role_inherit(%opts);
	
	$Log->exit() if $may_exit;

} ## end sub _inherit_dfm_cli

=head2 disinherit

   NACL::C::DFM::Role->disinherit( command_interface=>$ci,
                                   role => $role_name,
                                   "role-to-disinherit" => $disinherit_role);
  or
  
  $role->disinherit( ... );

(Class or instance method) Remove an inheritance from a role.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "role-to-disinherit"=>$disinherit_role >>

Name of role to disinherit.

=back

=back

=cut

sub disinherit {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
	$Log->debug("Opts to disinherit front end are:\n".Dumper(\@_));
	
    my %opts       = $pkg_or_obj->_common_validate_with(
        params          => \@_,
        additional_spec => { _optional_scalars(qw( role-to-disinherit )), }
    );
	$Log->exit() if $may_exit;
	
    return $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => "_disinherit_dfm_cli",
                interface => "CLI",
                set       => "DFM",
            }
        ],
    );
} ## end sub disinherit

sub _disinherit_dfm_cli {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );

    delete $opts{command_interface};
    my $apiset = delete $opts{apiset};

    my $dfm_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => [qw( role-to-disinherit )],
        map    => { "role" => "role-name" },
    );

    %opts = %$dfm_args;
    $Log->debug("Opts to API call 'dfm_role_disinherit' are:\n".Dumper(\%opts));
    $apiset->dfm_role_disinherit(%opts);
	
	$Log->exit() if $may_exit;

} ## end sub _disinherit_dfm_cli

=head2 destroy

  NACL::C::DFM::Role->destroy( command_interface=>$ci,
                               role => $role_name);
  or
  
  $role->destroy( );


(Class or instance method) Destroys a role.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< role=>$role_name >>

(Required for class method, Not Applicable for instance method)
Name of role.

=back

=back

=cut

sub destroy {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
	$Log->debug("Opts to destroy front end are:\n".Dumper(\@_));
	
    my %opts = $pkg_or_obj->_common_validate_with( params => \@_ );
	$Log->exit() if $may_exit;
	
    return $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => "_destroy_dfm_cli",
                interface => "CLI",
                set       => "DFM",
            }
        ],
    );
} ## end sub destroy

sub _destroy_dfm_cli {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );

    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};

    my $dfm_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        map    => { "role" => "role-name" },
    );

    %opts = %$dfm_args;
    $Log->debug("Opts to API call 'dfm_role_destroy' are:\n".Dumper(\%opts));
    $apiset->dfm_role_destroy(%opts);
	
	$Log->exit() if $may_exit;

} ## end sub _destroy_dfm_cli

# Defining this is a must
sub _primary_keys_validate_spec {
    return ( role => { type => SCALAR }, );
} ## end sub _primary_keys_validate_spec

1;

;
