# $Id$

#
# Copyright (c) 2001-2012 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Event Component Module
## @author dl-nacl-dev@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::C::Event

=head1 DESCRIPTION

C<NACL::C::Event> is a derived class of
L<NACL::C::Component::ONTAP|lib-NACL-C-Component-ONTAP-pm>. Unlike other components this
package does not model any element/service but is present to provide
component-level support for functionality. Since it does not model any
element/service there is no ComponentState module for this component.
For the same reason, 'find' and 'state' will not work for this component.

=cut

package NACL::C::Event;

use strict;
use warnings;

use NACL::ComponentUtils qw (_optional_scalars);
use NACL::CS::EventShowSuppression;
use NACL::CS::EventShowQueues;
use NACL::CS::EventShowStatistics;

use base 'NACL::C::Component::ONTAP';
use Params::Validate qw(validate validate_with SCALAR SCALARREF ARRAYREF);
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Data::Dumper;

=head2 show_queues

   NACL::C::Event->show_queues(
                                command_interface => $ci,
                                  ...);

(Class method)
This method is used to display EMS Engine Queue Sizes.
Returns objects of type NACL::CS::EventShowStatitics.
See L<NACL::CS::EventShowQueues|lib-NACL-CS-EventShowQueues-pm>
Uses a CMode CLI APISet.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>


=back

=back

=cut

sub show_queues {
    $Log->enter() if $may_enter;
    my $pkg    = shift;
    my @states = NACL::CS::EventShowQueues->fetch(@_);
    $Log->exit() if $may_exit;
    return wantarray ? @states : $states[0];
} ## end sub show_queues

=head2 generate

   NACL::C::Event->generate(
                                command_interface => $ci,
                                  ...);

(Class method)
This method is used to generate a test event.
Uses a CMode/7Mode CLI APISet.
Invokes 'ems-invoke' API for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "messagename" => $string >>

Applicable for CMode CLI, 7Mode CLI
Maps to:
    7M CLI  : event-name

=item C<< "values" => [ $value1, $value2, ... ] >>

Applicable for CMode CLI, 7Mode CLI
Maps to:
    7M CLI : values

=item C<< 'extended_query' => $string >>

(Optional)
Applicable only for CMode CLI

=item C<< other_options >>

This method takes other common options described in L<Common method options|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#Other_options_valid_on_component_methods>

=item C<< '*' >>

This method takes all other options supported by
"event generate" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub generate {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            "values" => {type => ARRAYREF, optional => 1},
            _optional_scalars(
                qw( messagename
                    )
            ),
        },
        ignore_primary_keys => 1,
        api                 => 'event_generate',
    );
    $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => "_generate_cmode_cli",
                interface => "CLI",
                set       => "CMode",
            },
            {   method    => "_generate_7mode_cli",
                interface => "CLI",
                set       => "7Mode",
            },
        ],
        frontend => 'NACL::C::Event::generate',
    );

    $Log->exit() if $may_exit;

} ## end sub generate

sub _generate_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, @opts) = @_;
    $pkg->_base_backend_cmode_cli(@opts, api => 'event_generate',);
    $Log->exit() if $may_exit;
} ## end sub _generate_cmode_cli

sub _generate_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );

    delete $opts{command_interface};
    my $apiset = delete $opts{apiset};

    my %cmode_args;
    $pkg->_hash_copy(
        source => \%opts,
        map    => {'messagename' => 'event-name'},
        copy   => [qw(values )],
        target => \%cmode_args,
    );
    %opts = %cmode_args;

    if (defined $opts{'values'}) {
        $opts{'values'} = join ' ', @{$opts{'values'}};
    }

    $apiset->ems_invoke(%opts);
    $Log->exit() if $may_exit;
} ## end sub _generate_7mode_cli

=head2 flush

   NACL::C::Event->flush(
                                command_interface => $ci,
                                  ...);

(Class method)
This method is used to flush the local EMS queues.
Uses a CMode CLI APISet.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< other_options >>

This method takes other common options described in L<Common method options|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#Other_options_valid_on_component_methods>

=item C<< '*' >>

This method takes all other options supported by
"event flush" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub flush {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params => \@args,
        api    => 'event_flush',
    );
    $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => "_flush_cmode_cli",
                interface => "CLI",
                set       => "CMode",
            },
        ],
        frontend => 'NACL::C::Event::flush',
    );

    $Log->exit() if $may_exit;

} ## end sub flush

sub _flush_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, @opts) = @_;
    $pkg->_base_backend_cmode_cli(@opts, api => 'event_flush',);
    $Log->exit() if $may_exit;
} ## end sub _flush_cmode_cli

=head2 sync_notifyd

   NACL::C::Event->sync_notifyd(
                                command_interface => $ci,
                                  ...);

(Class method)
This method is used to sync the route tables with notifyd.
Uses a CMode CLI APISet.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< other_options >>

This method takes other common options described in L<Common method options|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#Other_options_valid_on_component_methods>

=item C<< '*' >>

This method takes all other options supported by
"event sync notifyd" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub sync_notifyd {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params => \@args,
        api    => 'event_sync_notifyd',
    );
    $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => "_sync_notifyd_cmode_cli",
                interface => "CLI",
                set       => "CMode",
            },
        ],
        frontend => 'NACL::C::Event::sync_notifyd',
    );

    $Log->exit() if $may_exit;

} ## end sub sync_notifyd

sub _sync_notifyd_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, @opts) = @_;
    $pkg->_base_backend_cmode_cli(@opts, api => 'event_sync_notifyd',);
    $Log->exit() if $may_exit;
} ## end sub _sync_notifyd_cmode_cli

=head2 show_suppression

   my @state = NACL::C::Event->show_suppression(
                                command_interface => $ci,
                                  ...);

(Class method)
This method is used to display EMS Engine Suppression Statistics.
Returns objects of type NACL::CS::EventShowSuppression.
See L<NACL::CS::EventShowSuppression|lib-NACL-CS-EventShowSuppression-pm/command_interface>
Uses a CMode CLI APISet.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< filter => { 'is_on' => true ,...} >>

(Optional)

=item C<< requested_fields => ['is_on' ,...] >>

(Optional)

=back

=back

=cut

sub show_suppression {
    $Log->enter() if $may_enter;
    my $pkg    = shift;
    my @states = NACL::CS::EventShowSuppression->fetch(@_);
    $Log->exit() if $may_exit;
    return wantarray ? @states : $states[0];
} ## end sub show_suppression

=head2 show_statistics

    my @state =  NACL::C::Event->show_statistics(
                                command_interface => $ci,
                                  ...);

(Class method)
This method is used to display EMS Engine Statistics.
Returns objects of type NACL::CS::EventShowStatitics.
See L<NACL::CS::EventShowStatistics|lib-NACL-CS-EventShowStatistics-pm/command_interface>
Uses a CMode CLI APISet.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< filter => { 'receive-select-failures' => 0 ,...} >>

(Optional)

=item C<< requested_fields => ['receive-select-failures' ,...] >>

(Optional)

=back

=back

=cut

sub show_statistics {
    $Log->enter() if $may_enter;
    my $pkg    = shift;
    my @states = NACL::CS::EventShowStatistics->fetch(@_);
    $Log->exit() if $may_exit;
    return wantarray ? @states : $states[0];
} ## end sub show_statistics

=head2 rotate_files

   NACL::C::Event->rotate_files(
                                command_interface => $ci,
                                  ...);

(Class method)
This method is used to rotate the log files.
Uses a CMode/7Mode CLI APISet.
Invokes 'ems-rotate' API for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "node" => $string >>

Applicable only for CMode CLI

=item C<< 'extended_query' => $string >>

(Optional)
Applicable only for CMode CLI

=item C<< other_options >>

This method takes other common options described in L<Common method options|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#Other_options_valid_on_component_methods>

=item C<< '*' >>

This method takes all other options supported by
"event rotate-files" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub rotate_files {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( node
                    )
            ),
        },
        ignore_primary_keys => 1,
        cli_cmd             => 'event rotate-files',
    );
    $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => "_rotate_files_cmode_cli",
                interface => "CLI",
                set       => "CMode",
            },
            {   method    => "_rotate_files_7mode_cli",
                interface => "CLI",
                set       => "7Mode",
            },
        ],
        frontend => 'NACL::C::Event::rotate_files',
    );

    $Log->exit() if $may_exit;

} ## end sub rotate_files

sub _rotate_files_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, @opts) = @_;
    $pkg->_base_backend_cmode_cli(@opts, api => 'event_rotate_files',);
    $Log->exit() if $may_exit;
} ## end sub _rotate_files_cmode_cli

sub _rotate_files_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );

    delete $opts{command_interface};
    my $apiset = delete $opts{apiset};

    $apiset->ems_rotate();
    $Log->exit() if $may_exit;
} ## end sub _rotate_files_7mode_cli

=head2 generate_autosupport_log

    NACL::C::Event->generate_autosupport_log(
                                command_interface => $ci,
                                ...);
   
(Class method)
This method is used to generate an EMS event message, and an optional AutoSupport.
Uses a CMode CLI APISet.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'method-timeout' => $time_in_seconds >>

(Optional)
The default NATE timeout is 60 seconds. If the command might take more
time to complete, this option can be used to specify a larger timeout value.
The value should be provided in seconds.

=item C<< other_options >>

This method takes other common options described in L<Common method options|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#Other_options_valid_on_component_methods>

=item C<< '*' >>

This method takes all other options supported by
"event generate-autosupport-log" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub generate_autosupport_log {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params => \@args,
        additional_spec =>
            {'cmd_execution_time' => {type => SCALARREF, optional => 1},},
        ignore_primary_keys => 1,
        cli_cmd             => 'event generate-autosupport-log',
    );
    $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_generate_autosupport_log_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::Event::generate_autosupport_log',
    );

    $Log->exit() if $may_exit;
} ## end sub generate_autosupport_log

sub _generate_autosupport_log_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, @opts) = @_;
    $pkg->_base_backend_cmode_cli(@opts,
        api => 'event_generate_autosupport_log',);
    $Log->exit() if $may_exit;
} ## end sub _generate_autosupport_log_cmode_cli


__PACKAGE__->_load_event_class();

1;
