
#
# Copyright (c) 2001-2012 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary EventConfig Component Module
## @author dl-nacl-dev@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::C::EventConfig

=head1 DESCRIPTION

C<NACL::C::EventConfig> is a derived class of
L<NACL::C::Component::ONTAP|lib-NACL-C-Component-ONTAP-pm>. It represents access
to an ONTAP EventConfig.  A related class is
L<NACL::CS::EventConfig|lib-NACL-CS-EventConfig-pm>, which represents
the past state of an ONTAP EventConfig.
However, since it does not have any primary keys, 'find' and 'state' do not
work for this component. (This component is considered a 'singleton',
i.e. there is only one row shown by the relevant 'show' command)

=cut

=head1 ATTRIBUTES

=head2 command_interface

See L<NACL::C::Component::command_interface|lib-NACL-C-Component-pm/command_interface>

=cut

package NACL::C::EventConfig;

use strict;
use warnings;

use NACL::CS::EventConfig;
use NACL::ComponentUtils qw (_optional_scalars);

use base 'NACL::C::Component::ONTAP';
use Params::Validate qw(validate validate_with SCALAR ARRAYREF);
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Data::Dumper;

=head1 METHODS

=head2 new

  my $EventConfig = NACL::C::EventConfig->new(command_interface=>$ci,
                                  );

Return a new EventConfig component object with the given attributes.

=cut

=head2 modify

   NACL::C::EventConfig->modify(
                                command_interface => $ci,
                                  ...);

(Class method)
This method is used to modify log configuration parameters.
Uses a CMode CLI or CMode ZAPI APISet.
Invokes 'ems-config-modify' API for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "mailserver" => $string >>

Applicable for CMode CLI/ZAPI
Maps to:
    CM ZAPI : mail-server

=item C<< "maxtargeteventbytes" => $string >>

Applicable for CMode CLI/ZAPI
Maps to:
    CM ZAPI : max-target-event-bytes

=item C<< "console" => $string >>

Applicable for CMode CLI/ZAPI
Maps to:
    CM ZAPI : console

=item C<< "mailfrom" => $string >>

Applicable for CMode CLI/ZAPI
Maps to:
    CM ZAPI : mail-from

=item C<< "suppression" => $string >>

Applicable for CMode CLI/ZAPI
Maps to:
    CM ZAPI : suppression

=item C<< "maxlogshowsize" => $string >>

Applicable for CMode CLI/ZAPI
Maps to:
    CM ZAPI : max-log-show-size

=item C<< "extended_query" => $string >>

(Optional)
Applicable for CMode CLI

=item C<< "maxtargetlogsize" => $string >>

Applicable for CMode CLI/ZAPI
Maps to:
    CM ZAPI : max-target-log-size

=item C<< "consoleloglevel" => $string >>

Applicable for CMode CLI/ZAPI
Maps to:
    CM ZAPI : console-log-level

=item C<< "tracelogsize" => $string >>

Applicable for CMode CLI/ZAPI
Maps to:
    CM ZAPI : trace-log-size


=item C<< '*' >>

This method takes all other options supported by
"event config modify" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub modify {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( mailserver maxtargeteventbytes console mailfrom suppression maxlogshowsize
                    maxtargetlogsize consoleloglevel tracelogsize
                    )
            ),
        },
        cli_cmd => 'event config modify',
    );
    my $ret = $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_modify_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_modify_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::EventConfig::modify',
    );

    $Log->exit() if $may_exit;
    return $ret;
} ## end sub modify

sub _modify_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_base_backend_cmode_cli(@args,
        api     => 'event_config_modify',
    );
    $Log->exit() if $may_exit;
} ## end sub _modify_cmode_cli

sub _modify_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg, %opts) = @_;
    my %on_off_true_false_map = ('on' => 'true', 'off' => 'false');
    while (my ($key, $value) = each %opts) {
        if (exists $on_off_true_false_map{$value}) {
            $opts{$key} = $on_off_true_false_map{$value};
        }
    }
    my $allowed_cmode_options = [qw( console suppression )];
    $pkg->_base_backend_cmode_zapi(%opts,
        api    => 'ems_config_modify',
        copy   => $allowed_cmode_options,
        map    => {
            "mailfrom"            => "mail-from",
            "mailserver"          => "mail-server",
            "maxlogshowsize"      => "max-log-show-size",
            "tracelogsize"        => "trace-log-size",
            "consoleloglevel"     => "console-log-level",
            "maxtargetlogsize"    => "max-target-log-size",
            "maxtargeteventbytes" => "max-target-event-bytes",
        },
    );
    $Log->exit() if $may_exit;
} ## end sub _modify_cmode_zapi

=head2 delete

    NACL::C::EventConfig->delete(
                                command_interface => $ci,
                                ...);

(Class method)
Uses a CMode CLI APISet.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'method-timeout' => $time_in_seconds >>

(Optional)
The default NATE timeout is 60 seconds. If the command might take more
time to complete, this option can be used to specify a larger timeout value.
The value should be provided in seconds.

=item C<< '*' >>

This method takes all other options supported by
"event config delete" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub delete {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params  => \@args,
        cli_cmd => 'event config delete',
    );
    $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_delete_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::EventConfig::delete',
    );

    $Log->exit() if $may_exit;
} ## end sub delete

sub _delete_cmode_cli {
    $Log->enter() if $may_enter;
    my $pkg = shift;

    $pkg->_base_backend_cmode_cli(
        @_,
        api => 'event_config_delete'
    );

    $Log->exit() if $may_exit;
} ## end sub _delete_cmode_cli

=head2 create

    my $EventConfig_obj    = NACL::C::EventConfig->create(
                                command_interface => $ci,
                                ...);

(Class method)
Uses a CMode CLI APISet.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'mailserver' => $string >>

=item C<< 'maxtargeteventbytes' => $string >>

=item C<< 'console' => $string >>

=item C<< 'mailfrom' => $string >>

=item C<< 'suppression' => $string >>

=item C<< 'maxlogshowsize' => $string >>

=item C<< 'consoleloglevel' => $string >>

=item C<< 'maxtargetlogsize' => $string >>

=item C<< 'tracelogsize' => $string >>

=item C<< 'method-timeout' => $time_in_seconds >>

(Optional)
The default NATE timeout is 60 seconds. If the command might take more
time to complete, this option can be used to specify a larger timeout value.
The value should be provided in seconds.

=item C<< '*' >>

This method takes all other options supported by
"event config create" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub create {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( mailserver maxtargeteventbytes console mailfrom suppression maxlogshowsize
                    consoleloglevel maxtargetlogsize tracelogsize
                    )
            ),
        },
        cli_cmd => 'event config create',
    );
    my $obj = $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_create_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::EventConfig::create',
    );

    $Log->exit() if $may_exit;
    return $obj;
} ## end sub create

sub _create_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, %opts) = @_;

    $pkg->_base_backend_cmode_cli(
        %opts,
        api => 'event_config_create'
    );

    $Log->exit() if $may_exit;
    return $pkg->new(command_interface => $opts{command_interface},);
} ## end sub _create_cmode_cli


__PACKAGE__->_load_event_class();

sub _primary_keys_validate_spec {
    return ();
}

1;
