# $Id$

#
# Copyright (c) 2001-2012 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary EventDestination Component Module
## @author dl-nacl-dev@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::C::EventDestination

=head1 DESCRIPTION

C<NACL::C::EventDestination> is a derived class of
L<NACL::C::Component::ONTAP|lib-NACL-C-Component-ONTAP-pm>. It represents access
to an ONTAP EventDestination.  A related class is
L<NACL::CS::EventDestination|lib-NACL-CS-EventDestination-pm>, which represents
the past state of an ONTAP EventDestination.

=head1 ATTRIBUTES

=head2 command_interface

See L<NACL::C::Component::command_interface|lib-NACL-C-Component-pm/command_interface>

=head2 name

=cut

package NACL::C::EventDestination;

use strict;
use warnings;

use NACL::CS::EventDestination;
use NACL::ComponentUtils qw (_optional_scalars);

use base 'NACL::C::Component::ONTAP';
use Params::Validate qw(validate validate_with SCALAR ARRAYREF);
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use NACL::C::Exceptions::EventDestination::DoesNotExist ();
use NATE::BaseException qw(:try);
use NATE::Exceptions::Argument qw(:try);
use Data::Dumper;
use NACL::CS::ComponentState::ZapiSkip qw(make_zapi_skip);
use NACL::CS::ComponentState::ZapiArray qw(make_zapi_array);

=head1 METHODS

=head2 new

  my $EventDestination = NACL::C::EventDestination->new(command_interface=>$ci,
                                   name =>$name,
                                  );

Return a new EventDestination component object with the given attributes.

=cut

use Class::MethodMaker [scalar => 'name',];

=head2 find

    my $EventDestination  = NACL::C::EventDestination->find(command_interface => $ci, ...);
    my @EventDestinations = NACL::C::EventDestination->find(command_interface => $ci, ...);

(Class method)
A constructor of sorts, which discovers which elements are present (by
interacting with the given command interface).  In an array context it
creates a component object for each present element and returns them.
In a scalar context it creates a component object for just one
element.

This method is generally not implemented in individual components. The base
class (L<NACL::C::Component::ONTAP|lib-NACL-C-Component-ONTAP-pm> provides an implementation.
See <L<NACL::C::Component::find|lib-NACL-C-Component-pm/find> for a more
detailed description along with explanation of all the options accepted.

=head2 state

    my $EventDestination_state = $NACL::C::EventDestination->state();

(Instance method)
Fetch the current state of the element associated
with this component, and return a Component State object corresponding
to this component (calling this on NACL::C::EventDestination would return
a NACL::CS::EventDestination object)

This method is generally not implemented in individual components. The base
class (L<NACL::C::Component::ONTAP|lib-NACL-C-Component-ONTAP-pm> provides an implementation.
See <L<NACL::C::Component::state|lib-NACL-C-Component-pm/state> for a more
detailed description along with explanation of all the options accepted.

=cut

=head2 modify

   NACL::C::EventDestination->modify(
                                command_interface => $ci,
                                 name => $name,
                                  ...);

                                 or

    $EventDestination_object->modify(...);

(Class or instance method)
This method is used to modify an event destination.
Uses a CMode CLI/ZAPI APISet.
  Invokes "ems-destination-modify-iter" api for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "name" => $name >>

(Required for class method, Not Applicable for instance method)
Applicable for CMode CLI/ZAPI
Maps to:
    CM ZAPI : name

=item C<< "syslog-facility" => $string >>
Applicable for CMode CLI/ZAPI

Maps to:
  CM ZAPI: syslog-facility

=item C<< "hide-parameters" => $string >>
Applicable for CMode CLI/ZAPI

Maps to:
    CM ZAPI : hide-parameters

=item C<< "snmp-community" => $string >>
Applicable for CMode CLI/ZAPI

Maps to:
  CM ZAPI: snmp-community

=item C<< "snmp" => [ $snmp1, $snmp2, ... ] >>
Applicable for CMode CLI/ZAPI

Maps to:
  CM ZAPI: snmp

=item C<< "syslog" => [ $syslog1, $syslog2, ... ] >>
Applicable for CMode CLI/ZAPI

Maps to:
  CM ZAPI: syslog

=item C<< "mail" => [ $mail1, $mail2, ... ] >>
Applicable for CMode CLI/ZAPI

Maps to:
    CM ZAPI : mail

=item C<< "extended_query" => $string >>

(Optional)
Applicable for CMode CLI

=item C<< '*' >>

This method takes all other options supported by
"event destination modify" command in ONTAP.

Applicable for CMode CLI/ZAPI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=item C<NACL::C::Exceptions::EventDestination::DoesNotExist>

This type of exception gets thrown if an attempt is made to modify a
non-existing event destination.

=back

=back

=cut

sub modify {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            "snmp"   => {type => ARRAYREF, optional => 1},
            "syslog" => {type => ARRAYREF, optional => 1},
            "mail"   => {type => ARRAYREF, optional => 1},
            _optional_scalars(
                qw( syslog-facility hide-parameters snmp-community
                    )
            ),
        },
        cli_cmd => 'event destination modify',
    );
    my $ret;
    try {
        $ret = $pkg_or_obj->call_on_apiset(
            %opts,
            choices => [
                {   method    => '_modify_cmode_cli',
                    interface => 'CLI',
                    set       => 'CMode',
                },
                {   method    => '_modify_cmode_zapi',
                    interface => 'ZAPI',
                    set       => 'CMode',
                },
            ],
            frontend => 'NACL::C::EventDestination::modify',
        );
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = $_[0];
        NACL::C::Exceptions::EventDestination::DoesNotExist
            ->detect_convert_and_throw(exception => $exception);
        $Log->exit() if $may_exit;
        $exception->throw();
    };
    $Log->exit() if $may_exit;
    return $ret;
} ## end sub modify

sub _modify_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@opts,
        api => 'event_destination_modify',);
    $Log->exit() if $may_exit;
} ## end sub _modify_cmode_cli

sub _modify_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    require NACL::C::Exceptions::EventDestination::DoesNotExist;
    my $copy = [qw(syslog-facility hide-parameters snmp-community name)];
    my $map  = {
        'snmp'   => [
                      make_zapi_array('snmp'),
                      make_zapi_skip('remote-ip'),
                    ],
        'syslog' => [
                      make_zapi_array('syslog'),
                      make_zapi_skip('remote-ip'),
                    ],
        'mail' => [
                    make_zapi_array('mail'),
                    make_zapi_skip('mail-address'),
                  ],
    };
    $pkg_or_obj->_iter_helper(
        @opts,
        api      => 'ems-destination-modify-iter',
        show_cmd => 'event destination show',
        copy     => $copy,
        map      => $map,
        does_not_exist_exception =>
            'NACL::C::Exceptions::EventDestination::DoesNotExist',
    );
    $Log->exit() if $may_exit;
} ## end sub _modify_cmode_zapi

=head2 delete

   NACL::C::EventDestination->delete(
                                command_interface => $ci,
                                 name => $name,
                                  ...);

                                 or

    $EventDestination_object->delete(...);

(Class or instance method)
This method is used to delete an event destination.
Uses a CMode CLI/CMode ZAPI APISet.
  Invokes "ems-destination-destroy-iter" api for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "name" => $name >>

(Required for class method, Not Applicable for instance method)
Applicable for CMode CLI/ZAPI
Maps to:
    CM ZAPI : name

=item C<< "extended_query" => $string >>

(Optional)
Applicable for CMode CLI

=item C<< '*' >>

This method takes all other options supported by
"event destination delete" command in ONTAP.

Applicable for CMode CLI/ZAPI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=item C<NACL::C::Exceptions::EventDestination::DoesNotExist>

This type of exception gets thrown if an attempt is made to delete a
non-existing event destination.

=back

=back

=cut

sub delete {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params => \@args,
        cli_cmd => 'event destination delete',
    );
    my $ret;
    try {
        $ret = $pkg_or_obj->call_on_apiset(
            %opts,
            choices => [
                {   method    => '_delete_cmode_cli',
                    interface => 'CLI',
                    set       => 'CMode',
                },
                {   method    => '_delete_cmode_zapi',
                    interface => 'ZAPI',
                    set       => 'CMode',
                },
            ],
            frontend => 'NACL::C::EventDestination::delete',
        );
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = $_[0];
        NACL::C::Exceptions::EventDestination::DoesNotExist
            ->detect_convert_and_throw(exception => $exception);
        $Log->exit() if $may_exit;
        $exception->throw();
    };
    $Log->exit() if $may_exit;
    return $ret;
} ## end sub delete

sub _delete_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@opts,
        api => 'event_destination_delete',);
    $Log->exit() if $may_exit;
} ## end sub _delete_cmode_cli

sub _delete_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    require NACL::C::Exceptions::EventDestination::DoesNotExist;
    my $copy = [qw(syslog-facility hide-parameters snmp-community name)];
    my $map  = {
        'snmp'   => [
                      make_zapi_array('snmp'),
                      make_zapi_skip('remote-ip'),
                    ],
        'syslog' => [
                      make_zapi_array('syslog'),
                      make_zapi_skip('remote-ip'),
                    ],
        'mail' =>  [
                      make_zapi_array('mail'),
                      make_zapi_skip('mail-address'),
                   ],
    };
    $pkg_or_obj->_iter_helper(
        @opts,
        api      => 'ems-destination-destroy-iter',
        show_cmd => 'event destination show',
        copy     => $copy,
        map      => $map,
        does_not_exist_exception =>
            'NACL::C::Exceptions::EventDestination::DoesNotExist',
    );
    $Log->exit() if $may_exit;
} ## end sub _delete_cmode_zapi

=head2 create

    my $EventDestination_object    = NACL::C::EventDestination->create(
                                command_interface => $ci,
                                 name => $name,
                                  ...);

(Class method)
This method is used to create an event destination.
Uses a CMode CLI or CMode ZAPI APISet.
Invokes "ems-destination-create" API for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "name" => $name >>

(Required)
Applicable for CMode CLI/ZAPI
Maps to:
    CM ZAPI : name

=item C<< "syslog-facility" => $string >>

Applicable for CMode CLI/ZAPI
Maps to:
    CM ZAPI : syslog-facility

=item C<< "hide-parameters" => $string >>

Applicable for CMode CLI/ZAPI
Maps to:
    CM ZAPI : hide-parameters

=item C<< "snmp-community" => $string >>

Applicable for CMode CLI/ZAPI
Maps to:
    CM ZAPI : snmp-community

=item C<< "snmp" => [ $snmp1, $snmp2, ... ] >>

Applicable for CMode CLI/ZAPI
Maps to:
    CM ZAPI : snmp

=item C<< "syslog" => [ $syslog1, $syslog2, ... ] >>

Applicable for CMode CLI/ZAPI
Maps to:
    CM ZAPI : syslog

=item C<< "mail" => [ $mail1, $mail2, ... ] >>

Applicable for CMode CLI/ZAPI
Maps to:
    CM ZAPI : mail

=item C<< "extended_query" => $string >>

(Optional)
Applicable for CMode CLI

=item C<< '*' >>

This method takes all other options supported by
"event destination create" command in ONTAP.

Applicable for CMode CLI/ZAPI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub create {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            "snmp"   => {type => ARRAYREF, optional => 1},
            "syslog" => {type => ARRAYREF, optional => 1},
            "mail"   => {type => ARRAYREF, optional => 1},
            _optional_scalars(
                qw( syslog-facility hide-parameters snmp-community
                    )
            ),
        },
        cli_cmd => 'event destination create',
    );
    my $ret = $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_create_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_create_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::EventDestination::create',
    );

    $Log->exit() if $may_exit;
    return $pkg->new(
        command_interface => $opts{'command_interface'},
        name              => $opts{'name'},
    );
} ## end sub create

sub _create_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, @opts) = @_;
    $pkg->_base_backend_cmode_cli(@opts, api => 'event_destination_create',);
    $Log->exit() if $may_exit;
} ## end sub _create_cmode_cli

sub _create_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg, @opts) = @_;
    my $copy
        = [
        qw(syslog-facility snmp hide-parameters snmp-community syslog name mail)
        ];
    $pkg->_base_backend_cmode_zapi(
        @opts,
        api  => 'ems-destination-create',
        copy => $copy,
    );
    $Log->exit() if $may_exit;
} ## end sub _create_cmode_zapi


__PACKAGE__->_load_event_class();

sub _primary_keys_validate_spec {
    return ('name' => {type => SCALAR},);
}

1;
