#
# Copyright (c) 2012 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

=head1 NAME

NACL::C::Exceptions::Client::File::FileException


=head1 DESCRIPTION

This exception is thrown when a method call in NACL::C::Client::File failed.

Has gather_diagnostics() defined. If gather_diagnostics is called, then
diagnostic commands will be executed to provide more info for triage


=head1 EXAMPLE

    use NACL::C::Client::File;
    use NACL::C::Exceptions::Client::File::FileException qw(:try);

    my $fh_obj;
    try {
        $fh_obj = NACL::C::Client::File->open(
            file_name => "/dne/tmp.txt",
            access_mode => "write",
            open_flags => "create-new",
        );
    } catch NACL::C::Exceptions::Client::File::FileException with {

        # Do something special with the exception

    };


=head1 This implementation of gather_diagnostics()

In this implementation of gather_diagnostics(), the diagnostics it gathers are:
- "mount" -- what are the mounts currently open on the client?
- "ls" -- what are the files in the directory that the File->cmd() is called on?
- "stat $file_name" -- does $file_name exist?

=head1 SEE ALSO

=cut

use strict;
use warnings;

package NACL::C::Exceptions::Client::File::FileException;

use base qw(
    NACL::C::Exceptions::Component
);

use Class::MethodMaker [
    hash => 'file_args'
];


use Tharn qw( host);
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use NATE::BaseException qw(:try);

# The assumption of NACL::C::Client::File is that all the commands
# are run from the machine on which the script is being executed.
# Therefore, _exec_client_cmd is able to use the system() call
# in order to perform diagnostic commands
sub _exec_client_cmd {
    my ($self, %args) = @_;

    my $command = $args{command};

    $Log->comment("==========================================================");
    $Log->comment("Executing diagnostic command: $command" );
    $Log->comment("==========================================================");

    try {

        # Execute the given command on a NACL::C::Client obj
        system( $command );

    } catch NATE::BaseException with {

        # Print out the exception, then don't do anything else
        my $exception_obj = shift;
        $Log->comment( $exception_obj->stringify() );

    };

    $Log->comment("==========================================================");
    $Log->comment("== End of diagnostic cmd");
    $Log->comment("==========================================================");
}

sub _is_linux {
    my $host = host("localhost");
    if ($host->hosttype() =~ /unix/) {
        return 1;
    }
    else {
        return 0;
    }
}

sub gather_diagnostics {
    my $self = shift;
    my %file_args = $self->file_args();
    my $file_name = $file_args{file_name};

    # Perform diagnostic commands iff client is Linux
    if ($self->_is_linux()) {
        $self->_exec_client_cmd(command => "mount");
        $self->_exec_client_cmd(command => "ls");
        $self->_exec_client_cmd(command => "stat $file_name");
    }
    else {
        $Log->comment("======================================================");
        $Log->comment("== No diagnostic commands available for Windows");
        $Log->comment("======================================================");
    }
}

1;
