#
# Copyright (c) 2014-20* NetApp Inc.
# All rights reserved.
#
## @summary An exception for when User doesnt have enough
##  permission to operate on the file/dir
##
## @author dl-nacl-dev@netapp.com
## @status Shared
## @pod here

=head1 NAME

NACL::C::Exceptions::Client::PermissionDenied

=head1 DESCRIPTION

This exception is thrown when we attempt to execute a operation on aggregate which
is out of space.

=head1 EXAMPLE

   use NACL::C::Exceptions::Client::PermissionDenied qw(:try);

   try {
       NACL::C::Client::Ls->ls(paths => "/", %other_opts);
   } catch NACL::C::Exceptions::Client::PermissionDenied with {
        # User doesnt have permission to operate on the file
   };

=cut

package NACL::C::Exceptions::Client::PermissionDenied;
use NACL::APISet::Exceptions::ResponseException qw(:try);
use File::Spec;
use Data::Dumper;
use feature 'state';
use Params::Validate qw(validate_with OBJECT SCALAR);
use File::Spec::Win32;
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();

use base qw(
    NACL::C::Exceptions::Component
    NACL::APISet::Exceptions::ResponseException
);

use Class::MethodMaker [
    scalar => [ { '-type' => "NACL::C::Client" }, 'client_ci', ],
    scalar => [ { '-type' => "NACL::C::Node" }, 'node_ci', ],
    scalar => [ 'file_path', ],
];


sub detect {
    $Log->enter() if $may_enter;
    my ( $pkg, %args ) = @_;
    my $detected = $pkg->_detect_helper( %args,
        regexes => [qr/Permission denied/i] );
    $Log->exit() if $may_exit;
    return $detected;
}

sub detect_convert_and_throw {
    $Log->enter() if $may_enter;
    my ( $pkg, @args ) = @_;
    state $spec = {
        exception => { type => OBJECT },
        client_ci => { type => OBJECT, isa => "NACL::C::Client", optional => 1, },
        node_ci   => { type => OBJECT, isa => "NACL::C::Node", optional => 1, },
        file_path => { type => SCALAR, optional => 1 }
    };

    my %opts = validate_with(
        params => \@args,
        spec   => $spec,
    );

    my $ex = delete $opts{exception};
    my $new_ex;
    if($pkg->detect(exception => $ex)) {
        $new_ex = $pkg->convert(exception => $ex);
        foreach my $key (keys %opts) {
            my $value = $opts{$key};
            $new_ex->$key($value);
        }
        
        $Log->exit() if $may_exit;
        $new_ex->throw();
    }
}

sub gather_diagnostics {
    $Log->enter() if $may_enter;
    my ($self) = @_;

    $Log->comment("#"x40 . " Diag Info for Client " . "#"x40);
    $self->_gather_diag_info_for_client();
    $Log->comment("#"x40 . " Finished Gathering Diag Info " . "#"x36);
    $Log->comment("  \n\n  ");
    $Log->comment("#"x40 . " Diag Info for Node " . "#"x40);
    $self->_gather_diag_info_for_node();
    $Log->comment("#"x40 . " Finished Gathering Diag Info " . "#"x36);

    $Log->exit() if $may_exit;
}

sub _gather_diag_info_for_client {
    $Log->enter() if $may_enter;
    my ($self) = @_;
    my $ci = $self->client_ci();
    if(! defined $ci) {
        $Log->warn("Command Interface for Client is not passed. " .
                   "Skipping to collect diag info the Client.");
    } else {
        if ($ci->is_windows) {
            $self->_gather_diag_info_for_windows();
        } else {
            $self->_gather_diag_info_for_unix();
        }
    }

    $Log->exit() if $may_exit;
}

sub _gather_diag_info_for_unix {
    $Log->enter() if $may_enter;
    my ($self) = @_;
    my $ci = $self->client_ci();

    my $apiset = $ci->apiset();
    my $vm = $apiset->get_version_manager();

    $vm->get_version_attribute(attribute => "os_revision");
    my $major = $vm->{os_revision}->{major};
    my $minor = $vm->{os_revision}->{minor};
    my $uname = $apiset->uname(all => 1)->get_parsed_output();

    my $logged_user = $apiset->execute_command(command => "/usr/bin/whoami");

    $self->_print_file_path_permissions_unix();
    $Log->comment("Logged in User: " . $logged_user);
    $Log->comment("OS Kernel Version: " . Dumper($uname));
    $Log->comment("OS major Revision: $major");
    $Log->comment("OS minor Revision: $minor");

    $Log->exit() if $may_exit;
}

sub _gather_diag_info_for_windows {
    $Log->enter() if $may_enter;
    my ($self) = @_;
    my $ci = $self->client_ci();
    my $apiset = $ci->apiset();

    my $logged_user = $apiset->execute_command(command => "echo %username%");

    $Log->comment("Logged in User: " . $logged_user);
    $Log->comment("Permission of tree path: " . $self->_print_file_path_permissions_windows($dir_paths));
    $Log->comment("Windows Systeminfo: ");
    $apiset->execute_raw_command(command => "systeminfo");

    $Log->exit() if $may_exit;
}

sub _print_file_path_permissions_windows {
    $Log->enter() if $may_enter;
    my ($self) = @_;

    my $file = $self->file_path();
    my @files = File::Spec::Win32->splitdir('C:\Users\Administrator\Desktop');
    while (@files) {
        my $each_path = File::Spec::Win32->catfile(@files);
        pop(@files);
        push(@path_dirs, $each_path);
    }

    my $apiset = $self->client_ci()->apiset();

    foreach my $dir (@path_dirs) {
        $apiset->icacls(pattern => $dir);
    }

    $Log->exit() if $may_exit;
}

sub _print_file_path_permissions_unix {
    $Log->enter() if $may_enter;
    my ($self) = @_;

    my $file = $self->file_path();

    my @path_dirs;
    my @files = File::Spec->splitdir($file);
    while (@files) {
        my $each_path = File::Spec->catfile(@files);
        pop(@files);
        push(@path_dirs, $each_path);
    }

    my $apiset = $self->client_ci()->apiset();

    foreach my $path (@path_dirs) {
        try {
            $apiset->ls("paths" => $path, "long-list" => 1, "directory" => 1,);

        } otherwise {
            # Test has already failed so there is no point in
            # failing in gather_diagnostics call
        };
    }

    $Log->exit() if $may_exit;
}

sub _gather_diag_info_for_node {
    $Log->enter() if $may_enter;
    my ($self, %args) = @_;
    if(! defined $args{ci}) {
        $Log->warn("Command Interface for Node is not passed. " .
                   "Skipping collecting diag info for node.");
    } else {
        my $ci = $args{ci};
    }

    $Log->exit() if $may_exit;
}

1;
