#
# Copyright (c) 2009-2012 NetApp Inc.
# All rights reserved.
#
## @summary Base class for the exception that gets thrown when a feature
##          is not licensed.
##
## @author madhavs@netapp.com, dl-nacl-dev@netapp.com
## @status Shared
## @pod here

=head1 NAME

NACL::C::Exceptions::FeatureNotLicensed

=head1 DESCRIPTION

This is the base class for the exceptions that get thrown when a
command is run for a particular feature which has not been licensed.
All exceptions which are sub-classes of this exception have the attributes
and methods listed below.

=cut

package NACL::C::Exceptions::FeatureNotLicensed;

use strict;
use warnings;

use base 'NACL::C::Exceptions::Component';

=head1 ATTRIBUTES

Each of the attributes listed below have accessor methods, so (using the
C<feature> attribute as example) this can be done:

    my $feature = $exception_obj->feature();
    
=over

=item C<feature>

The feature which needs to be licensed to run the command.

=item C<command_interface>

The command_interface on which the feature needs to be licensed.

=back

=cut

use Class::MethodMaker [
    scalar => 'feature',
    scalar => [
        { -type => 'NACL::C::CommandInterface::ONTAP' },
        'command_interface'
    ],
];

use NACL::STask::SystemLicense;
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Params::Validate qw(validate_with SCALAR);
use NACL::ComponentUtils qw(_ontap_ci_validate_spec _hash_move);

=head1 EXAMPLES

(Using C<NACL::C::Exceptions::VserverNfs::FeatureNotLicensed> for the
examples below: that exception sub-classes this exception)

=head2 Throwing exception

    NACL::C::Exceptions::VserverNfs::FeatureNotLicensed->throw(
        # Like all other component-level exception, this has the
        # attributes of the APISet exceptions: "text" and "response_object"
        text              => $error_msg,
        response_object   => $resp,
        
        # These are specific to this exception and are mandatory
        feature           => $feature_name,
        command_interface => $ci
    );
    
=head2 Catching exception

    try {
        NACL::C::VserverNfs->create(%opts);
    } catch NACL::C::Exceptions::VserverNfs::FeatureNotLicensed with {
        my $exception = shift;
        
        # Getting the error text is the same as for other exceptions
        $Log->comment('Error is: ' . $exception->text());
        
        # Getting the response object, like any other component-level exception
        my $response_obj = $exception->get_response_object();
        
        # Getting the attributes special to this exception
        my $feature = $exception->feature();
        my $command_interface = $exception->command_interface();
    };

=cut

sub new {
    $Log->enter() if $may_enter;
    my ( $pkg, @args ) = @_;
    my %opts = validate_with(
        params => \@args,
        spec   => $pkg->_validate_spec(),

        # Options accepted by SUPER::new
        allow_extra => 1
    );
    my $self = $pkg->SUPER::new(%opts);
    $Log->exit() if $may_exit;
    return $self;
}

=head2 license_feature

    # An example showing the call failing because the feature is not licensed,
    # then licensing it and re-trying the call.
    LOOP: {
        use warnings;
        try {
            NACL::C::VserverNfs->create(command_interface => $ci, %other_opts);
        } catch NACL::C::Exceptions::VserverNfs::FeatureNotLicensed with {
            my $exception   = shift;
            my $license_obj = $exception->license_feature();
            no warnings;
            redo LOOP;
        };
    }

This method can be used to license the missing feature for which the
exception was thrown. The C<command_interface> and C<feature> are extracted
automatically from within the exception object and the
L<< NACL::STask::SystemLicense->add|lib-NACL-STask-SystemLicense-pm/add >>
method is invoked. Not that all of the other arguments accepted by the
L<< NACL::STask::SystemLicense->add|lib-NACL-STask-SystemLicense-pm/add >>
can be passed as arguments to this method.

This method returns a
L<NACL::STask::SystemLicense|lib-NACL-STask-SystemLicense-pm> object.

=cut

sub license_feature {
    $Log->enter() if $may_enter;
    my ( $self, @args ) = @_;
    my $license_obj = NACL::STask::SystemLicense->add(
        command_interface => $self->command_interface(),
        feature           => $self->feature(),
        @args
    );
    $Log->exit() if $may_exit;
    return $license_obj;
}

=head1 FOR DEVELOPERS

This exception introduces new mandatory arguments to the calls to
C<convert> and C<convert_and_throw>.

    my $licensed_exception = NACL::C::Exceptions::VserverNfs::FeatureNotLicensed->convert(
        exception         => $exception_obj,
        feature           => $feature_name,
        command_interface => $ci
    );
    
    NACL::C::Exceptions::VserverNfs::FeatureNotLicensed->convert_and_throw(
        exception         => $exception_obj,
        feature           => $feature_name,
        command_interface => $ci
    );

=cut

sub convert {
    $Log->enter() if $may_enter;
    my ( $pkg, @args ) = @_;
    my %validate_spec = $pkg->_validate_spec();
    my %opts          = validate_with(
        params => \@args,
        spec   => \%validate_spec,

        # These are the options for SUPER::convert
        allow_extra => 1
    );
    my %attributes;
    $pkg->_hash_move(
        source => \%opts,
        target => \%attributes,
        move   => [ keys %validate_spec ]
    );
    my $obj = $pkg->SUPER::convert(%opts);
    while ( my ( $key, $value ) = each %attributes ) {
        $obj->$key($value);
    }
    $Log->exit() if $may_exit;
    return $obj;
}

sub _validate_spec {
    my %hash = (
        feature           => { type => SCALAR },
        command_interface => _ontap_ci_validate_spec()
    );
    return wantarray ? %hash : {%hash};
}

1;
