# $Id: //depot/prod/test/nacldev/lib/NACL/C/Exceptions/TimeNotSynchronised.pm#$
#
# Copyright (c) 2009-2012 NetApp Inc.
# All rights reserved.
#
## @summary Exception to be thrown when the time accross the cluster is not synchronised
##
## @author chethanm@netapp.com, dl-nacl-dev@netapp.com
## @status Shared
## @pod here

=head1 NAME

NACL::C::Exceptions::TimeNotSynchronised

=head1 DESCRIPTION

This exception is thrown when the time accross cluster is not synchronised.
It is mainly used in NACL::CS::EventLog.

=head1 EXAMPLE

   use NACL::C::Exceptions::TimeNotSynchronised qw(:try);

   try {
       $event_obj->state(%opts);
   } catch NACL::C::Exceptions::TimeNotSynchronised with {
            my $ex = shift;
        my $event_objs = $ex->state_objects();
        $Test->nlog("Got the error " . $ex->text());
        $Test->nlog("Event Log Objects " . Dumper(\@event_objs));
   };

=cut

package NACL::C::Exceptions::TimeNotSynchronised;

use strict;
use warnings;

use base 'NACL::C::Exceptions::Component';

use Class::MethodMaker [ array => 'state_objects' ];
use NACL::ComponentUtils qw( _ontap_ci_validate_spec _hash_move );
use NACL::CS::ComponentState::ONTAP;
use Params::Validate qw(validate_with SCALAR OBJECT);

use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();

sub detect {
    $Log->enter() if $may_enter;
    my ( $pkg, @args ) = @_;
    my $detected = $pkg->_detect_helper( @args,
        regexes => [qr/time does not appear to be synchronised/] );
    $Log->exit() if $may_exit;
    return $detected;
}

=head1 FOR DEVELOPERS

This exception introduces new mandatory arguments to the calls to
C<convert> and C<convert_and_throw>.

    my $licensed_exception = NACL::C::Exceptions::TimeNotSynchronised->convert(
        exception         => $exception_obj,
        response_object   => $resp_obj,
        api               => $api,
        command_interface => $ci
    );

    my $licensed_exception = NACL::C::Exceptions::TimeNotSynchronised->convert_and_throw(
        exception         => $exception_obj,
        response_object   => $resp_obj,
        api               => $api,
        command_interface => $ci
    );

=over

=item C<< api => scalar >>

(Required) The api command used.

Ex: api => "event_log_show"

=back

=head2 command_interface

(Required) A component object that represents the host to which to send commands.

=cut

sub convert {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    my %validate_spec = $pkg->_validate_spec();
    my %opts          = validate_with(
        params => \@args,
        spec   => \%validate_spec,

        # These are the options for SUPER::convert
        allow_extra => 1
    );

    my %attributes;
    $pkg->_hash_move(
        source => \%opts,
        target => \%attributes,
        move   => [keys %validate_spec]
    );
    $attributes{'response_object'}
        = $opts{'exception'}->get_response_object();
    my $obj = $pkg->SUPER::convert(%opts);
    if (defined $attributes{'response_object'}->{'parsed_output'}) {
        my @cs_objs = NACL::CS::ComponentState::ONTAP->_construct_cs_objs_cm_cli(
            %attributes);
        $obj->state_objects(@cs_objs);
    }

    $Log->exit() if $may_exit;
    return $obj;
}

sub _validate_spec {
    my %hash = (
        api               => { type => SCALAR },
        command_interface => _ontap_ci_validate_spec()
    );
    return wantarray ? %hash : {%hash};
}

1;
