#
# Copyright (c) 2009-2012 NetApp Inc.
# All rights reserved.
#
## @summary An exception for when an attempt is made to mount a volume
##          which has already been mounted at a different junction path
##
## @author madhavs@netapp.com, dl-nacl-dev@netapp.com
## @status Shared
## @pod here

=head1 NAME

NACL::C::Exceptions::Volume::AlreadyMounted

=head1 DESCRIPTION

This exception is thrown when we attempt to mount a volume
which has already been mounted at a different junction path

=head1 EXAMPLE

   use NACL::C::Exceptions::Volume::AlreadyMounted qw(:try);

   try {
       $mounted_vol->mount('junction-path' => $existing_path);
   } catch NACL::C::Exceptions::Volume::AlreadyMounted with {
        # Failed because it is already mounted
   };

=cut

package NACL::C::Exceptions::Volume::AlreadyMounted;

use base qw(
    NACL::APISet::Exceptions::CommandFailedException
    NACL::C::Exceptions::Component
);

use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Params::Validate qw(:all);

=head1 ATTRIBUTES

Each of the attributes listed below have accessor methods, so (using the
C<feature> attribute as example) this can be done:

    my $feature = $exception_obj->feature();
    
=over

=item C<current_junction_path>

The junction-path at which the volume was mounted at the time of the
call to mount().

=back

=cut

use Class::MethodMaker [scalar => 'current_junction_path'];

=head1 EXAMPLES

=head2 Throwing exception

    NACL::C::Exceptions::Volume::AlreadyMounted->throw(
        # Like all other component-level exception, this has the
        # attributes of the APISet exceptions: "text" and "response_object"
        text              => $error_msg,
        response_object   => $resp,
        
        # These are specific to this exception and are mandatory
        current_junction_path => $path
    );
    
=head2 Catching exception

    try {
        NACL::C::Volume->mount(%opts);
    } catch NACL::C::Exceptions::Volume::AlreadyMounted with {
        my $exception = $_[0];
        
        # Getting the error text is the same as for other exceptions
        $Log->comment('Error is: ' . $exception->text());
        
        # Getting the response object, like any other component-level exception
        my $response_obj = $exception->get_response_object();
        
        # Getting the attributes special to this exception
        my $current_junction_path = $exception->current_junction_path();
    };

=cut

sub new {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    my $self = $pkg->_new_with_validate_spec(@args);
    $Log->exit() if $may_exit;
    return $self;
}

=head1 FOR DEVELOPERS

This exception introduces new arguments to the calls to
C<convert> and C<detect>.

=head1 detect

    my $current_junction_path;
    NACL::C::Exceptions::Volume::AlreadyMounted->detect(
        exception         => $exception_obj,
        get_junction_path => \$current_junction_path
    );
    $Log->comment("Current junction path is $current_junction_path");
    
The error message when mount fails because it's already mounted at a
different path is:

    Volume "myvol1" on Vserver "myvs" is mounted at "/newvol2".
    
As can be seen, the error message mentions where the volume is currently
mounted. The C<get_junction_path> argument can be used to extract this value.
Note that this argument is optional, and should be specified as a reference
to a scalar variable if provided.

=cut

sub detect {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    my %opts = validate_with(
        params => \@args,
        spec   => {
            exception => {
                type => OBJECT,
                isa  => 'NACL::APISet::Exceptions::ResponseException'
            },
            get_junction_path => {type => SCALARREF, optional => 1}
        }
    );
    my $text = $opts{exception}->text();

    my $dummy;
    $opts{get_junction_path} ||= \$dummy;

    if ($text =~ /is mounted at ["'](.*)['"]/i) {
        ${$opts{get_junction_path}} = $1;
        $Log->exit() if $may_exit;
        return 1;
    } else {
        $Log->exit() if $may_exit;
        return 0;
    }
}

=head2 convert

    NACL::C::Exceptions::Volume::AlreadyMounted->convert(
        exception             => $exception_obj,
        current_junction_path => $current_junction_path
    );
    
The C<convert> method requires an additional mandatory
C<current_junction_path> argument.

This implies that the call to C<convert_and_throw> also requires
this additional mandatory argument.

=cut

sub convert {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    my $obj = $pkg->_convert_with_validate_spec(@args);
    $Log->exit() if $may_exit;
    return $obj;
}

sub detect_convert_and_throw {
    $Log->enter() if $may_enter;
    my ($pkg, %opts) = @_;
    my $dummy;
    $opts{get_junction_path} ||= \$dummy;
    if ($pkg->detect(%opts)) {
        $opts{current_junction_path} = ${$opts{get_junction_path}};
        delete $opts{get_junction_path};
        $Log->exit() if $may_exit;
        $pkg->convert_and_throw(%opts);
    }
    $Log->exit() if $may_exit;
}

sub _validate_spec {
    my %spec = (current_junction_path => {type => SCALAR});
    return wantarray ? %spec : {%spec};
}

1;
