#
# Copyright (c) 2014 NetApp Inc.
# All rights reserved.
#
## @summary An exception is thrown when we attempt to write to a
##          file with offset larger than file length
##
## @author dl-nacl-dev@netapp.com
## @status Shared
## @pod here

=head1 NAME

NACL::C::Exceptions::VolumeFile::FileTooLarge

=head1 DESCRIPTION

This exception is thrown when we attempt to write to a file with
offset larger than file length

=head1 EXAMPLE

   use NACL::C::Exceptions::VolumeFile::FileTooLarge qw(:try);

   try {
       $file_obj->write(content => $content, overwrite => 'false');
   } catch NACL::C::Exceptions::VolumeFile::FileTooLarge with {
        # File size is too large.
   };

=cut

package NACL::C::Exceptions::VolumeFile::FileTooLarge;

use base qw(
    NACL::APISet::Exceptions::ResponseException
    NACL::C::Exceptions::Component
);
use NACL::C::VolumeFile;

use Class::MethodMaker [
    hash => 'file_args'
];

use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use NATE::BaseException qw(:try);
use Params::Validate qw(validate_with HASHREF);
use NACL::ComponentUtils qw( _hash_move );

sub detect {
    $Log->enter();
    my ( $pkg, @args ) = @_;

    my $detected = $pkg->_detect_helper(
        @args,
        regexes => [ qr/File too large/i ]
    );

    $Log->exit();
    return $detected;
}

=head1 FOR DEVELOPERS

This exception introduces new mandatory arguments to the calls to
C<convert> and C<convert_and_throw>.

    my $file_exception = NACL::C::Exceptions::VolumeFile::FileTooLarge->convert(
        exception         => $exception_obj,
        response_object   => $resp_obj,
        file_args         => \%args,
    );

    my $file_exception = NACL::C::Exceptions::VolumeFile::FileTooLarge->convert_and_throw(
        exception         => $exception_obj,
        response_object   => $resp_obj,
        file_args         => \%args,
    );

=over

=item C<< file_args => hashref >>

(Required) file arguments used.

=back

=cut

sub convert {
    $Log->enter();
    my ( $pkg, @args ) = @_;
    my %validate_spec = $pkg->_validate_spec();
    my %opts          = validate_with(
        params => \@args,
        spec   => \%validate_spec,

        # These are the options for SUPER::convert
        allow_extra => 1
    );
    my %attributes;
    $pkg->_hash_move(
        source => \%opts,
        target => \%attributes,
        move   => [ keys %validate_spec ]
    );
    my $obj = $pkg->SUPER::convert(%opts);
    while ( my ( $key, $value ) = each %attributes ) {
        $obj->$key($value);
    }
    $Log->exit();
    return $obj;
}

# Execute a diagnostic command.
# For ANY exception, print out the info, but do NOT
# let the exception propagate up the call stack.
# We print and ignore the exception because the exception
# was generated by diagnostics, NOT by the actual code
# that is being tested.
sub _exec_diagnostic_cmd {
    my ($self, %args) = @_;
    my $command = $args{command};

    $Log->comment("==============================================");
    $Log->comment("Executing diagnostic command: $command" );
    $Log->comment("==============================================");

    try {
        $command->();
    } catch NATE::BaseException with {

        # Print out the exception, then don't do anything else
        my $exception_obj = shift;
        $Log->comment( $exception_obj->stringify() );

    };

    $Log->comment("==============================================");
    $Log->comment("== End of diagnostic cmd");
    $Log->comment("==============================================");
}

sub gather_diagnostics {
    my $self = shift;
    my %file_args = $self->file_args();
    my $file_path = $file_args{path};
    # Perform diagnostic commands
    my $file_obj = NACL::C::VolumeFile->new(
        command_interface => $file_args{command_interface},
        vserver           => $file_args{vserver},
        path              => $file_args{path},
    );
    $self->_exec_diagnostic_cmd(
        command => sub {
            my $file_attributes = $file_obj->state();
        },
    );
}

sub _validate_spec {
    my %hash = (
        file_args   => { type => HASHREF },
    );
    return wantarray ? %hash : {%hash};
}
1;
