# $Id$

#
# Copyright (c) 2012-* NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Represents a HyperV machine with mgmt interface and IP address
## @author rawat@netapp.com, dl-nacl-dev@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::C::HyperV

=head1 DESCRIPTION

C<NACL::C::HyperV> is a derived class of
L<NACL::C::CommandInterface|lib-NACL-C-CommandInterface-pm>.  It represents
access to a HyperV Host.

=head1 ATTRIBUTES

=head2 name

The name of the HyperV, used to identify it uniquely.

=head2 hostrec

Hostrec object for the HyperV.

=head2 hypervhost

Command Interface for the HyperV. Its an alias to the name.

=cut

package NACL::C::HyperV;

use strict;
use warnings;

use Params::Validate qw(:all);
use NATE::BaseException qw(:try);
use Scalar::Util qw(blessed);
use NATE::Exceptions::Argument;
use NACL::APISet;
use NACL::C::HyperV::Task;
use NACL::CS::HyperV;
use NACL::ComponentUtils qw(_optional_scalars Dumper);

require VMware::VICommon;
require VMware::VIMRuntime;
require VMware::VIM25Stub;

use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();

use base 'NACL::C::CommandInterface';


use Class::MethodMaker [
    scalar => [
         { '-default_ctor' => sub { $_[0]->_construct_primary_key(); }
         },
         'hypervhost'
     ],
];

=head1 METHODS

=head2 new

my $hyperv = NACL::C::HyperV->new([name => $name],
                              [hypervhost => $hypervhost],
                              [hostrec => $hostrec] );
Specifying any one of the hyperv, name, hostrec attributes will populate
the other two.
Returns a new HyperV component object with the given attributes.

=cut

sub networksystem {
    my ($pkg_or_obj, @args) = @_;
    $Log->enter() if $may_enter;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
    );
    my $ret = $pkg_or_obj->call_on_apiset( %opts,
                        choices => [
                            {
                              method => '_networksystem_esx',
                              interface => 'VMAPI',
                              set   => 'ESX',
                            },
                        ],
                    );

    $Log->exit() if $may_exit;
    return $ret;
}

sub _networksystem_esx {
    my ($pkg_or_obj, @args) = @_;
    $Log->enter() if $may_enter;

    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = $opts{apiset};
    my $response_for_hostview = $apiset->find_entity_view(
        view_type  => 'HostSystem',
        properties => ['configManager.networkSystem']
    );
    my $hostview = $response_for_hostview->get_raw_output();

    my $response_for_networksystem = $apiset->get_view(
        mo_ref => $hostview->{'configManager.networkSystem'} );
    my $networksystem
        = $response_for_networksystem->get_raw_output();

    $Log->exit() if $may_exit;
    return $networksystem;
}

sub vdisksystem {
    my ($pkg_or_obj, @args) = @_;
    $Log->enter() if $may_enter;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
    );
    my $ret = $pkg_or_obj->call_on_apiset( %opts,
                        choices => [
                            {
                              method => '_vdisksystem_esx',
                              interface => 'VMAPI',
                              set   => 'ESX',
                            },
                        ],
                    );

    $Log->exit() if $may_exit;
    return $ret;
}

sub _vdisksystem_esx {
    my ($pkg_or_obj, @args) = @_;
    $Log->enter() if $may_enter;

    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = $opts{apiset};

    my $response_for_hostview = $apiset->find_entity_view(
                    view_type  => 'HostSystem',
                    properties => ['datastore']
                );
    my $hostview = $response_for_hostview->get_raw_output();

    my $response_for_service_content
        = $apiset->get_service_content();
    my $service_content
        = $response_for_service_content->get_raw_output();

    my $response_for_vdisk_manager = $apiset->get_view(
        mo_ref => $service_content->{virtualDiskManager} );
    my $vdisk_manager
        = $response_for_vdisk_manager->get_raw_output();

    $Log->exit() if $may_exit;
    return _VdiskSysem->new(
                hostview     => $hostview,
                vdiskmanager => $vdisk_manager
           );
}

sub datastoresystem {
    my ($pkg_or_obj, @args) = @_;
    $Log->enter() if $may_enter;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
    );
    my $ret = $pkg_or_obj->call_on_apiset( %opts,
                        choices => [
                            {
                              method => '_datastoresystem_esx',
                              interface => 'VMAPI',
                              set   => 'ESX',
                            },
                        ],
                    );

    $Log->exit() if $may_exit;
    return $ret;
}

sub _datastoresystem_esx {
    my ($pkg_or_obj, @args) = @_;
    $Log->enter() if $may_enter;

    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = $opts{apiset};

    my $response_for_hostview = $apiset->find_entity_view(
                    view_type  => 'HostSystem',
                    properties => ['configManager.datastoreSystem']
                );
    my $hostview = $response_for_hostview->get_raw_output();

    my $response_for_datastoresystem = $apiset->get_view(
        mo_ref => $hostview->{'configManager.datastoreSystem'} );
    my $datastoresystem
        = $response_for_datastoresystem->get_raw_output();

    $Log->exit() if $may_exit;
    return $datastoresystem;
}

sub virtualmachines {
    my ($pkg_or_obj, @args) = @_;
    $Log->enter() if $may_enter;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
    );
    my $ret = $pkg_or_obj->call_on_apiset( %opts,
                        choices => [
                            {
                              method => '_virtualmachines_xen',
                              interface => 'XAPI',
                              set   => 'Xen',
                            },
                            {
                              method => '_virtualmachines_esx',
                              interface => 'VMAPI',
                              set   => 'ESX',
                            },
                        ],
                    );

    $Log->exit() if $may_exit;
    return $ret;
}
sub _virtualmachines_xen {
    my ($pkg_or_obj, @args) = @_;
    $Log->enter() if $may_enter;

    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = $opts{apiset};

    my $xenhost = $apiset->xen_host();

    my $response = $apiset->simple_request(
                        "VM.get_all",
                   );

    my @vms;
    foreach my $vmref (@{$response->get_parsed_output()}) {
        my $resp = $apiset->simple_request("VM.get_is_a_template", $vmref);
        my $is_template = $resp->get_parsed_output()->[0];
        if (! $is_template){
            push(@vms, $vmref);
        }
    }
    
    $Log->exit();    
    return \@vms;
}

sub _virtualmachines_esx {
    my ($pkg_or_obj, @args) = @_;
    $Log->enter() if $may_enter;

    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = $opts{apiset};

    my $response_for_vmview = $apiset->find_entity_views(
                    view_type => 'VirtualMachine', );
    my $vmviews = $response_for_vmview->get_raw_output();

    $Log->exit() if $may_exit;
    return $vmviews;
}

sub tasks {
    my ($pkg_or_obj, @args) = @_;
    $Log->enter() if $may_enter;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
    );
    my $ret = $pkg_or_obj->call_on_apiset( %opts,
                        choices => [
                            {
                              method => '_tasks_esx',
                              interface => 'VMAPI',
                              set   => 'ESX',
                            },
                        ],
                    );

    $Log->exit() if $may_exit;
    return $ret;
}

sub _tasks_esx {
    my ($pkg_or_obj, @args) = @_;
    $Log->enter() if $may_enter;

    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = $opts{apiset};

    my $response_for_service_content
                    = $apiset->get_service_content();
    my $service_content
        = $response_for_service_content->get_raw_output();
    my $response_for_task = $apiset->get_view(
        mo_ref => $service_content->{taskManager} );
    my $taskviews = $response_for_task->get_raw_output();

    $Log->exit() if $may_exit;
    return $taskviews;
}

sub folder {
    my ($pkg_or_obj, @args) = @_;
    $Log->enter() if $may_enter;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
    );
    my $ret = $pkg_or_obj->call_on_apiset( %opts,
                        choices => [
                            {
                              method => '_folder_esx',
                              interface => 'VMAPI',
                              set   => 'ESX',
                            },
                        ],
                    );

    $Log->exit() if $may_exit;
    return $ret;
}

sub _folder_esx {
    my ($pkg_or_obj, @args) = @_;
    $Log->enter() if $may_enter;

    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = $opts{apiset};
    my $response_for_folder
                    = $apiset->find_entity_views( view_type => 'Folder', );
    my $folder = $response_for_folder->get_raw_output();

    $Log->exit() if $may_exit;
    return $folder;
}

sub hostsystem {
    my ($pkg_or_obj, @args) = @_;
    $Log->enter() if $may_enter;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
    );
    my $ret = $pkg_or_obj->call_on_apiset( %opts,
                        choices => [
                            {
                              method => '_hostsystem_xen',
                              interface => 'XAPI',
                              set   => 'Xen',
                            },
                            {
                              method => '_hostsystem_esx',
                              interface => 'VMAPI',
                              set   => 'ESX',
                            },
                        ],
                    );

    $Log->exit() if $may_exit;
    return $ret;
}

sub _hostsystem_esx {
    my ($pkg_or_obj, @args) = @_;
    $Log->enter() if $may_enter;

    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = $opts{apiset};
    my $response_for_hostview
                    = $apiset->find_entity_view( view_type => 'HostSystem', );
    my $host = $response_for_hostview->get_raw_output();

    $Log->exit() if $may_exit;
    return $host;
}

sub _hostsystem_xen {
    my ($pkg_or_obj, @args) = @_;
    $Log->enter() if $may_enter;

    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = $opts{apiset};
    my $host = $apiset->xen_host();

    $Log->exit() if $may_exit;
    return $host;
}

sub cluster_compute_resource {
    my ($pkg_or_obj, @args) = @_;
    $Log->enter() if $may_enter;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
    );
    my $ret = $pkg_or_obj->call_on_apiset( %opts,
                        choices => [
                            {
                              method => '_cluster_compute_resource_esx',
                              interface => 'VMAPI',
                              set   => 'ESX',
                            },
                        ],
                    );

    $Log->exit() if $may_exit;
    return $ret;
}

sub _cluster_compute_resource_esx {
    my ($pkg_or_obj, @args) = @_;
    $Log->enter() if $may_enter;

    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = $opts{apiset};
    my $response_for_cluster = $apiset->find_entity_view(
                    view_type => 'ClusterComputeResource', );
    my $cluster = $response_for_cluster->get_raw_output();

    $Log->exit() if $may_exit;
    return $cluster;
}

sub datacenter {
    my ($pkg_or_obj, @args) = @_;
    $Log->enter() if $may_enter;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
    );
    my $ret = $pkg_or_obj->call_on_apiset( %opts,
                        choices => [
                            {
                              method => '_datacenter_esx',
                              interface => 'VMAPI',
                              set   => 'ESX',
                            },
                        ],
                    );

    $Log->exit() if $may_exit;
    return $ret;
}

sub _datacenter_esx{
    my ($pkg_or_obj, @args) = @_;
    $Log->enter() if $may_enter;

    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = $opts{apiset};
    my $response_for_datacenter
                    = $apiset->find_entity_view( view_type => 'Datacenter', );
    my $datacenter = $response_for_datacenter->get_raw_output();

    $Log->exit() if $may_exit;
    return $datacenter;
}

sub resourcepools {
    my ($pkg_or_obj, @args) = @_;
    $Log->enter() if $may_enter;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
    );
    my $ret = $pkg_or_obj->call_on_apiset( %opts,
                        choices => [
                            {
                              method => '_resourcepools_esx',
                              interface => 'VMAPI',
                              set   => 'ESX',
                            },
                        ],
                    );

    $Log->exit() if $may_exit;
    return $ret;
}
sub _resourcepools_esx {
    my ($pkg_or_obj, @args) = @_;
    $Log->enter() if $may_enter;

    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = $opts{apiset};
    my $response_for_resourcepool
                    = $apiset->find_entity_views( view_type => 'ResourcePool', );
    my $resourcepool_views = $response_for_resourcepool->get_raw_output();

    $Log->exit() if $may_exit;
    return $resourcepool_views;
}

sub virtualnicmanager {
    my ($pkg_or_obj, @args) = @_;
    $Log->enter() if $may_enter;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
    );
    my $ret = $pkg_or_obj->call_on_apiset( %opts,
                        choices => [
                            {
                              method => '_virtualnicmanager_esx',
                              interface => 'VMAPI',
                              set   => 'ESX',
                            },
                        ],
                    );

    $Log->exit() if $may_exit;
    return $ret;
}

sub _virtualnicmanager_esx {
    my ($pkg_or_obj, @args) = @_;
    $Log->enter() if $may_enter;

    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = $opts{apiset};
    my $response_for_virtualnicmanager_view
                    = $apiset->find_entity_view(
                            view_type => 'HostSystem',
                            properties => ['configManager.virtualNicManager']    );
    my $virtualnicmanager_view = $response_for_virtualnicmanager_view->get_raw_output();

    my $response_for_virtualnicmanager = $apiset->get_view(
        mo_ref => $virtualnicmanager_view->{'configManager.virtualNicManager'} );
    my $virtualnicmanager
        = $response_for_virtualnicmanager->get_raw_output();

    $Log->exit() if $may_exit;
    return $virtualnicmanager;
}

=head2 add_standalone_host

    $Connect_Spec = {
        hostName => $ci->hostrec()->default_ip(),
        userName => $ci->hostrec()->default_username(),
        password => $ci->hostrec()->default_password(),
        force    => 1,
    };
    $result = NACL::C::HyperV->add_standalone_host(
                            command_interface => $ci,
                            addConnected      => 0,
                            spec              => $Connect_Spec
    );

    OR

    $result = $hyperv->add_standalone_host(
                            addConnected => 0,
                            spec         => $Connect_Spec
                       );

(Class or Instance method) This method adds a standalone host to the server.
Applicable for VMAPI ESX

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< license => $string >>

(optional) License to add.

=item C<< spec => host spec >>

(Required) Host spec to add.

=item C<< compResSpec => resource spec >>

(Optional) resource spec.

=item C<< addConnected=>$boolean >>

(Required) add connected flag

=back

=back

=cut

sub add_standalone_host {
    my ($pkg_or_obj, @args) = @_;
    $Log->enter() if $may_enter;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            addConnected => { type => BOOLEAN },
            task_object  => { type => SCALARREF, optional => 1 },
            task_object  => { type => SCALARREF, optional => 1 },
            license      => { type => SCALAR,    optional => 1 },
            spec         => { type => HASHREF,   optional => 1 },
            compResSpec  => { type => HASHREF,   optional => 1 },
        },
    );
    $pkg_or_obj->call_on_apiset( %opts,
        choices => [
            {
              method => '_add_standalone_host_esx',
              interface => 'VMAPI',
              set   => 'ESX',
            },
        ],
    );

    $Log->exit() if $may_exit;
} ## end sub add_standalone_host

sub _add_standalone_host_esx {
    my ($pkg_or_obj, @args) = @_;
    $Log->enter() if $may_enter;

    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my $task_object       = delete $opts{task_object};
    my $folder_obj;

    my $allowed_options = [ qw( spec compResSpec addConnected license ) ];
    my $folder          = $command_interface->folder();
    foreach my $folder_view (@$folder) {
        if ( $folder_view->name() eq "host" ) {
            my $childType = $folder_view->childType;
            foreach my $child (@$childType) {
                if ( $child eq "ComputeResource" ) {
                    $folder_obj = $folder_view;
                    last;
                }
            }
        }
    }
    my %args;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => $allowed_options,
        target => \%args,
    );
    my $host_connect_spec = HostConnectSpec->new(%{$args{'spec'}});
    $args{'spec'} = $host_connect_spec;
    my $result = $folder_obj->AddStandaloneHost_Task(%args);

    if ($task_object) {
        $$task_object = $pkg_or_obj->_get_task_object( apiset => $apiset, result => $result );
    }

    $Log->exit() if $may_exit;
} ## end sub _add_standalone_host

=head2 disconnect_host

  my $host  = NACL::C::HyperV->disconnect_host(
                                        command_interface => $ci );

  OR

  $hyperv->disconnect_host();

(Class or Instance method) This method disconnects host from the server.
Applicable for VMAPI ESX

=over

=item Options


=item C<< command_interface=>$command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=back

=cut

sub disconnect_host {
    my ($pkg_or_obj, @args) = @_;
    $Log->enter() if $may_enter;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
                task_object => { type => SCALARREF, optional => 1 },
        },
    );
    $pkg_or_obj->call_on_apiset( %opts,
        choices => [
            {
              method => '_disconnect_host_esx',
              interface => 'VMAPI',
              set   => 'ESX',
            },
        ],
    );

    $Log->exit() if $may_exit;
} ## end sub disconnect_host

sub _disconnect_host_esx {
    my ($pkg_or_obj, @args) = @_;
    $Log->enter() if $may_enter;

    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my $task_object     = delete $opts{task_object};

    my $host     = $command_interface->hostsystem();
    my $result   = $host->DisconnectHost_Task();
    if ($task_object) {
        $$task_object = $pkg_or_obj->_get_task_object( apiset => $apiset, result => $result );
    }

    $Log->exit() if $may_exit;
} ## end sub _disconnect_host

=head2 reconnect_host

  my $host  = NACL::C::HyperV->reconnect_host(
                                        command_interface => $ci );

  OR

  my $host  = $hyperv->reconnect_host();

(Class or Instance method) This method reconnects a host to the server.
Applicable for VMAPI ESX

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< cnxSpec => connection spec >>

(Optional) connection spec.

=back

=back

=cut

sub reconnect_host {
    my ($pkg_or_obj, @args) = @_;
    $Log->enter() if $may_enter;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            task_object => { type => SCALARREF, optional => 1 },
            cnxSpec     => { type => HASHREF, optional => 1 },
        },
    );
    my $ret = $pkg_or_obj->call_on_apiset( %opts,
                        choices => [
                            {
                              method => '_reconnect_host_esx',
                              interface => 'VMAPI',
                              set   => 'ESX',
                            },
                        ],
                    );

    $Log->exit() if $may_exit;
    return $ret;
} ## end sub reconnect_host

sub _reconnect_host_esx {
    my ($pkg_or_obj, @args) = @_;
    $Log->enter() if $may_enter;

    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my $task_object     = delete $opts{task_object};

    my $host            = $command_interface->hostsystem();
    my $allowed_options = [qw( cnxSpec )];
    my %args;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => $allowed_options,
        target => \%args,
    );
    my $result = $host->ReconnectHost_Task(%args);
    if ($task_object) {
        $$task_object = $pkg_or_obj->_get_task_object( apiset => $apiset, result => $result );
    }

    $Log->exit() if $may_exit;
    return $pkg_or_obj->new(
        command_interface => $command_interface,
    );
} ## end sub _reconnect_host

=head2 enter_maintenance_mode

    my $result = NACL::C::HyperV->enter_maintenance_mode(
        command_interface => $ci,
        timeout           => 240,
    );

    OR

    $result = $hyperv->enter_maintenance_mode(timeout => 240);

(Class or Instance method) This method enters host into maintenance mode.
Applicable for VMAPI ESX

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< evacuatePoweredOffVms => $boolean >>

(optional) This is a parameter only supported by VirtualCenter.

=item C<< task_object => \$task_object >>

(optional) If passed, a task object will be created
and will be assigned to it

=item C<< timeout => $timeout >>

(Required) timeout.

=back

=back

=cut

sub enter_maintenance_mode {
    my ($pkg_or_obj, @args) = @_;
    $Log->enter() if $may_enter;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            task_object           => { type => SCALARREF, optional => 1 },
            timeout               => { type => SCALAR },
            evacuatePoweredOffVms => { type => BOOLEAN, optional => 1 },
        },
    );

    $pkg_or_obj->call_on_apiset( %opts,
        choices => [
            {
              method => '_enter_maintenance_mode_esx',
              interface => 'VMAPI',
              set   => 'ESX',
            },
        ],
    );

    $Log->exit() if $may_exit;
} ## end sub enter_maintenance_mode

sub _enter_maintenance_mode_esx {
    my ($pkg_or_obj, @args) = @_;
    $Log->enter() if $may_enter;

    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my $task_object     = delete $opts{task_object};

    my $allowed_options = [qw( timeout evacuatePoweredOffVms )];
    my $host            = $command_interface->hostsystem();
    my %args;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => $allowed_options,
        target => \%args,
    );
    my $result = $host->EnterMaintenanceMode_Task(%args);

    if ($task_object) {
        $$task_object = $pkg_or_obj->_get_task_object( apiset => $apiset, result => $result );
    }

    $Log->exit() if $may_exit;
} ## end sub _enter_maintenance_mode

=head2 exit_maintenance_mode

    my $result = NACL::C::HyperV->exit_maintenance_mode(
        command_interface => $ci,
        timeout           => 240,
    );
    
    OR
    
    $result = $hyperv->exit_maintenance_mode(timeout => 240);

(Class or Instance method) This method exits the host from maintenance mode.
Applicable for VMAPI ESX

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< task_object => \$task_object >>

(optional) If passed, a task object will be created
and will be assigned to it

=item C<< timeout => $timeout >>

(Required) timeout.

=back

=back

=cut

sub exit_maintenance_mode {
    my ($pkg_or_obj, @args) = @_;
    $Log->enter() if $may_enter;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            task_object => { type => SCALARREF, optional => 1 },
            timeout => { type => SCALAR },
        },
    );
    $pkg_or_obj->call_on_apiset( %opts,
        choices => [
            {
              method => '_exit_maintenance_mode_esx',
              interface => 'VMAPI',
              set   => 'ESX',
            },
        ],
    );

    $Log->exit() if $may_exit;
} ## end sub exit_maintenance_mode

sub _exit_maintenance_mode_esx {
    my ($pkg_or_obj, @args) = @_;
    $Log->enter() if $may_enter;

    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my $task_object     = delete $opts{task_object};

    my $allowed_options = [qw( timeout )];
    my $host            = $command_interface->hostsystem();
    my %args;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => $allowed_options,
        target => \%args,
    );
    my $result = $host->ExitMaintenanceMode_Task(%args);
    if ($task_object) {
        $$task_object = $pkg_or_obj->_get_task_object( apiset => $apiset, result => $result );
    }

    $Log->exit() if $may_exit;
} ## end sub _exit_maintenance_mode

=head2 disable

    my $result = NACL::C::HyperV->disable(
        command_interface => $ci,
    );

    OR
    
    my $result = $hyperv->disable();

(Class or Instance method) This method disables the host.
Applicable for XAPI Xen

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=back

=back

=cut

sub disable {
    my ($pkg_or_obj, @args) = @_;
    $Log->enter() if $may_enter;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            task_object => { type => SCALARREF, optional => 1 },
        },
    );
    $pkg_or_obj->call_on_apiset( %opts,
        choices => [
            { method => '_disable_host_xen',
              interface => 'XAPI',
              set   => 'Xen',
            },
        ],
    );

    $Log->exit() if $may_exit;
} ## end sub disable

sub _disable_host_xen {
    my ($pkg_or_obj, @args) = @_;
    $Log->enter() if $may_enter;

    my %opts              = $pkg_or_obj->_validate_backend_opts(@args);
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};

    my $xenhost = $pkg_or_obj->hostsystem();
    my $req = "host.disable";

    # Execute the request
    $apiset->simple_request($req, $xenhost);

    $Log->exit() if $may_exit;

} ## end sub _disable_xen_host

=head2 enable_ha

    my $result = NACL::C::HyperV->enable_ha(
        command_interface      => $ci,
        storage_repository     => \@storage_repository,
        ha_config              => \%ha_config,
    );

    OR
    
    my $result = $hyperv->enable_ha(
        storage_repository     => \@storage_repository,
        ha_config              => \%ha_config,
    );

(Class or Instance method) This method enables HA on the pool.
Applicable for XAPI Xen

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< storage_repository=>\@storage_repository >>

Set of Storage Repositories(SRs) to use for storage heartbeating.

=item C<< ha_config=>\%ha_config, >>

Detailed HA configuration to apply

=back

=back

=cut

sub enable_ha {
    my ($pkg_or_obj, @args) = @_;
    $Log->enter() if $may_enter;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            storage_repository  => { type => ARRAYREF, default => [] },
            ha_config           => { type => HASHREF, default => {} },
        },
    );
    $pkg_or_obj->call_on_apiset( %opts,
        choices => [
            { method => '_enable_ha_xen',
              interface => 'XAPI',
              set   => 'Xen',
            },
        ],
    );

    $Log->exit() if $may_exit;
} ## end sub disable_ha

sub _enable_ha_xen {
    my ($pkg_or_obj, @args) = @_;
    $Log->enter() if $may_enter;

    my %opts              = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset            = delete $opts{apiset};

    my $req = "pool.enable_ha";

    # Execute the request
    $apiset->simple_request($req, $opts{storage_repository}, $opts{ha_config});

    $Log->exit() if $may_exit;

} ## end sub _enable_ha_xen

=head2 disable_ha

    my $result = NACL::C::HyperV->disable_ha(
        command_interface => $ci,
    );

    OR
    
    my $result = $hyperv->disable_ha();

(Class or Instance method) This method disables HA on the pool.
Applicable for XAPI Xen

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=back

=back

=cut

sub disable_ha {
    my ($pkg_or_obj, @args) = @_;
    $Log->enter() if $may_enter;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            task_object => { type => SCALARREF, optional => 1 },
        },
    );
    $pkg_or_obj->call_on_apiset( %opts,
        choices => [
            { method => '_disable_ha_xen',
              interface => 'XAPI',
              set   => 'Xen',
            },
        ],
    );

    $Log->exit() if $may_exit;
} ## end sub disable_ha

sub _disable_ha_xen {
    my ($pkg_or_obj, @args) = @_;
    $Log->enter() if $may_enter;

    my %opts              = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset            = delete $opts{apiset};

    my $req = "pool.disable_ha";

    # Execute the request
    $apiset->simple_request($req);

    $Log->exit() if $may_exit;

} ## end sub _disable_ha_xen

=head2 initialize_wlb

    my $result = NACL::C::HyperV->initialize_wlb(
        command_interface   => $ci,
        wlb_url             => $wlb_url,
        wlb_username        => $wlb_username,
        wlb_password        => $wlb_password,
        xenserver_username  => $xenserver_username,
        xenserver_password  => $xenserver_password,
    );

    OR
    
    my $result = $hyperv->initialize_wlb(
        wlb_url             => $wlb_url,
        wlb_username        => $wlb_username,
        wlb_password        => $wlb_password,
        xenserver_username  => $xenserver_username,
        xenserver_password  => $xenserver_password,
    );

(Class or Instance method) This method will initialize work load balancing(wlb) on the pool.
Applicable for XAPI Xen

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< wlb_url=>$wlb_url >>

The ip address and port to use when accessing
the wlb server

=item C<< wlb_username=>$wlb_username >>

The username used to authenticate with the
wlb server

=item C<< wlb_password=>$wlb_password >>

The password used to authenticate with the
wlb server

=item C<< xenserver_username=>$xenserver_username >>

The usernamed used by the wlb server to authenticate
with the xenserver

=item C<< xenserver_password=>$xenserver_password >>

The password used by the wlb server to authenticate
with the xenserver

=back

=back

=cut

sub initialize_wlb {
    my ($pkg_or_obj, @args) = @_;
    $Log->enter() if $may_enter;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            wlb_url             => { type => SCALAR, },
            wlb_username        => { type => SCALAR, },
            wlb_password        => { type => SCALAR, },
            xenserver_username  => { type => SCALAR, },
            xenserver_password  => { type => SCALAR, },
        },
    );
    $pkg_or_obj->call_on_apiset( %opts,
        choices => [
            { method => '_initialize_wlb_xen',
              interface => 'XAPI',
              set   => 'Xen',
            },
        ],
    );

    $Log->exit() if $may_exit;
} ## end sub initialize_wlb

sub _initialize_wlb_xen {
    my ($pkg_or_obj, @args) = @_;
    $Log->enter() if $may_enter;

    my %opts                = $pkg_or_obj->_validate_backend_opts(@args);
    my $command_interface   = delete $opts{command_interface};
    my $apiset              = delete $opts{apiset};
    my $wlb_url             = $opts{wlb_url};
    my $wlb_password        = $opts{wlb_password};
    my $wlb_username        = $opts{wlb_username};
    my $xenserver_username  = $opts{xenserver_username};
    my $xenserver_password  = $opts{xenserver_password};

    my $req = "pool.initialize_wlb";

    # Execute the request
    $apiset->simple_request(
        $req, $wlb_url, $wlb_username, $wlb_password,
        $xenserver_username, $xenserver_password,
    );

    $Log->exit() if $may_exit;

} ## end sub _initialize_wlb_xen


=head2 deconfigure_wlb

    my $result = NACL::C::HyperV->deconfigure_wlb(
        command_interface => $ci,
    );

    OR
    
    my $result = $hyperv->deconfigure_wlb();

(Class or Instance method) This method deconfigure work load balancing(wlb) on the pool.
Applicable for XAPI Xen

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=back

=back

=cut

sub deconfigure_wlb {
    my ($pkg_or_obj, @args) = @_;
    $Log->enter() if $may_enter;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(params => \@args);
    $pkg_or_obj->call_on_apiset( %opts,
        choices => [
            { method => '_deconfigure_wlb_xen',
              interface => 'XAPI',
              set   => 'Xen',
            },
        ],
    );

    $Log->exit() if $may_exit;
} ## end sub deconfigure_wlb

sub _deconfigure_wlb_xen {
    my ($pkg_or_obj, @args) = @_;
    $Log->enter() if $may_enter;

    my %opts              = $pkg_or_obj->_validate_backend_opts(@args);
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};

    my $req = "pool.deconfigure_wlb";

    # Execute the request
    $apiset->simple_request($req);

    $Log->exit() if $may_exit;

} ## end sub _deconfigure_wlb_xen

=head2 pool_join

    my $result = NACL::C::HyperV->pool_join(
        command_interface   => $ci,
        master_ip_address   => $master_ip_address,
        master_username     => $master_username,
        master_password     => $master_password,
    );

    OR
    
    my $result = $hyperv->pool_join(
        master_ip_address   => $master_ip_address,
        master_username     => $master_username,
        master_password     => $master_password,
    );

(Class or Instance method) This method join the host to the pool whose master's ip address is $master_ip_address.
Applicable for XAPI Xen

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< master_ip_address=>$master_ip_address >>

IP address of the master of the pool

=item C<< master_username=>$master_username >>

Username that can be used to login to the master of the pool

=item C<< master_password=>$master_password >>

Password for the user on the master of the pool

=item C<< force=>$boolean >>

(Optional) Force Xen host to join the pool if set

=back

=back

=cut

sub pool_join {
    my ($pkg_or_obj, @args) = @_;
    $Log->enter() if $may_enter;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            task_object       => { type => SCALARREF, optional => 1 },
            force             => { type => SCALAR, optional => 1 },
            master_ip_address => { type => SCALAR, },
            master_username   => { type => SCALAR, },
            master_password   => { type => SCALAR, },
        },
    );
    $pkg_or_obj->call_on_apiset( %opts,
        choices => [
            { method => '_pool_join_xen',
              interface => 'XAPI',
              set   => 'Xen',
            },
        ],
    );

    $Log->exit() if $may_exit;
} ## end sub pool_join

sub _pool_join_xen {
    my ($pkg_or_obj, @args) = @_;
    $Log->enter() if $may_enter;

    my %opts              = $pkg_or_obj->_validate_backend_opts(@args);
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my $master_ip         = $opts{master_ip_address};
    my $master_usr        = $opts{master_username};
    my $master_passwd     = $opts{master_password};
    my $force             = $opts{force};

    my $req = "pool.join";
    if ($force) {
        $req = "pool.join_force";
    }

    # Execute the request
    $apiset->simple_request($req, $master_ip, $master_usr, $master_passwd);

    $Log->exit() if $may_exit;

} ## end sub _pool_join_xen

=head2 pool_eject

    my $result = NACL::C::HyperV->pool_eject(
        command_interface   => $ci,
        xen_server          => $xen_server,
    );

    OR
    
    my $result = $hyperv->pool_eject(xen_server => $xen_server,);

(Class or Instance method) Instruct a pool master to eject a host from the pool.
Applicable for XAPI Xen

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< xen_server=>$xen_server >>

This can be any of the three types
1. Xen host ref
2. Xen Server name
3. Command Interface of Xen Server

=back

=back

=cut

sub pool_eject {
    my ($pkg_or_obj, @args) = @_;
    $Log->enter() if $may_enter;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            task_object => { type => SCALARREF, optional => 1 },
            xen_server  => { type => OBJECT|SCALAR,},
        },
    );
    $pkg_or_obj->call_on_apiset( %opts,
        choices => [
            {
              method    => '_pool_eject_xen',
              interface => 'XAPI',
              set       => 'Xen',
            },
        ],
    );

    $Log->exit() if $may_exit;
} ## end sub pool_eject

sub _pool_eject_xen {
    my ($pkg_or_obj, @args) = @_;
    $Log->enter() if $may_enter;

    my %opts              = $pkg_or_obj->_validate_backend_opts(@args);
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my $xenserver         = $opts{xen_server};

    my $xen_host = $pkg_or_obj->_get_xen_host(
                                    xenserver => $xenserver,
                                    apiset  => $apiset,
                                );

    my $req = "pool.eject";

    # Execute the request
    $apiset->simple_request($req, $xen_host);

    $Log->exit() if $may_exit;

} ## end sub _pool_eject_xen


=head2 designate_new_master

    my $result = NACL::C::HyperV->designate_new_master(
        command_interface   => $ci,
        xen_server          => $xen_server,
    );

    OR
    
    my $result = $hyperv->designate_new_master(xen_server => $xen_server,);

(Class or Instance method) Designate new master of the pool.
Applicable for XAPI Xen

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< xen_server=>$xen_server >>

This can be any of the three types
1. Xen host ref
2. Xen Server name
3. Command Interface of Xen Server

=back

=back

=cut

sub designate_new_master {
    my ($pkg_or_obj, @args) = @_;
    $Log->enter() if $may_enter;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            xen_server  => { type => OBJECT|SCALAR,},
        },
    );
    $pkg_or_obj->call_on_apiset( %opts,
        choices => [
            {
              method    => '_designate_new_master_xen',
              interface => 'XAPI',
              set       => 'Xen',
            },
        ],
    );

    $Log->exit() if $may_exit;
} ## end sub designate_new_master

sub _designate_new_master_xen {
    my ($pkg_or_obj, @args) = @_;
    $Log->enter() if $may_enter;

    my %opts              = $pkg_or_obj->_validate_backend_opts(@args);
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my $xenserver          = $opts{xen_server};

    my $xen_host = $pkg_or_obj->_get_xen_host(
                                    xenserver => $xenserver,
                                    apiset  => $apiset,
                                );

    my $req = "pool.designate_new_master";

    # Execute the request
    $apiset->simple_request($req, $xen_host);

    $Log->exit() if $may_exit;

} ## end sub _designate_new_master_xen

=head2 reboot_host

    my $result = NACL::C::HyperV->reboot_host(
        command_interface => $ci,
        force             => 1,
    );

    OR  

    my $result = $hyperv->reboot_host(force => 1);

(Class or Instance method) This method reboots the host.
Applicable for VMAPI ESX and XAPI Xen.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< force => $boolean >>

(Required) for ESX Host.
(Optional) for Xen Host.

=item C<< task_object => \$task_object >>

(optional) If passed, a task object will be created
and will be assigned to it

=back

=back

=cut

sub reboot_host {
    my ($pkg_or_obj, @args) = @_;
    $Log->enter() if $may_enter;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            task_object => { type => SCALARREF, optional => 1 },
            force => { type => BOOLEAN,  optional => 1},
        },
    );
    $pkg_or_obj->call_on_apiset( %opts,
        choices => [
            { method => '_reboot_host_esx',
              interface => 'VMAPI',
              set   => 'ESX',
            },
            { method => '_reboot_host_xen',
              interface => 'XAPI',
              set   => 'Xen'
            },
        ],
    );

    $Log->exit() if $may_exit;
} ## end sub reboot_host

sub _reboot_host_esx {
    my ($pkg_or_obj, @args) = @_;
    $Log->enter() if $may_enter;

    my %opts              = $pkg_or_obj->_validate_backend_opts(@args);
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my $task_object     = delete $opts{task_object};

    my $allowed_options = [qw( force )];
    my $host            = $command_interface->hostsystem();
    my %args;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => $allowed_options,
        target => \%args,
    );
    my $result = $host->RebootHost_Task(%args);
    if ($task_object) {
        $$task_object = $pkg_or_obj->_get_task_object( apiset => $apiset, result => $result );
    }

    $Log->exit() if $may_exit;
} ## end sub _reboot_esx_host

sub _reboot_host_xen {
    my ($pkg_or_obj, @args) = @_;
    $Log->enter() if $may_enter;

    my %opts              = $pkg_or_obj->_validate_backend_opts(@args);
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};

    my $xenhost = $pkg_or_obj->hostsystem();
    my $req = "host.reboot";

    # Execute the request
    $apiset->simple_request($req, $xenhost);

    $Log->exit() if $may_exit;
} ## end sub _reboot_xen_host

=head2 shutdown_host

    my $result = NACL::C::HyperV->shutdown_host(
        command_interface => $ci,
        force             => 1,
    );

    OR
    
    my $result = $hyperv->shutdown_host(force => 1);

(Class or Instance method) This method shutdown the host.
Applicable for VMAPI ESX and XAPI Xen.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< force => $boolean >>

(Required) For ESX host.
(Optional) For Xen host.

=item C<< task_object => \$task_object >>

(optional) If passed, a task object will be created
and will be assigned to it

=back

=back

=cut

sub shutdown_host {
    my ($pkg_or_obj, @args) = @_;
    $Log->enter() if $may_enter;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            task_object => { type => SCALARREF, optional => 1 },
            force => { type => BOOLEAN, optional => 1 },
        },
    );
    $pkg_or_obj->call_on_apiset( %opts,
        choices => [
            {
                method => '_shutdown_host_esx',
                interface => 'VMAPI',
                set   => 'ESX'
            },
            {
                method => '_shutdown_host_xen',
                interface => 'XAPI',
                set   => 'Xen'
            },
        ],
    );

    $Log->exit() if $may_exit;
} ## end sub shutdown_host

sub _shutdown_host_esx {
    my ($pkg_or_obj, @args) = @_;
    $Log->enter() if $may_enter;

    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my $task_object     = delete $opts{task_object};

    my $allowed_options = [qw( force )];
    my $host            = $command_interface->hostsystem();
    my %args;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => $allowed_options,
        target => \%args,
    );
    my $result = $host->ShutdownHost_Task(%args);
    if ($task_object) {
        $$task_object = $pkg_or_obj->_get_task_object( apiset => $apiset, result => $result );
    }

    $Log->exit() if $may_exit;
} ## end sub _shutdown_host

sub _shutdown_host_xen {
    my ($pkg_or_obj, @args) = @_;
    $Log->enter() if $may_enter;

    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};

    my $xenhost = $pkg_or_obj->hostsystem();
    my $req = "host.shutdown";

    # Execute the request
    $apiset->simple_request($req, $xenhost);

    $Log->exit() if $may_exit;
} ## end sub _shutdown_host

=head2 move_into_folder

    my $result = NACL::C::HyperV->move_into_folder(
        command_interface => $ci,
        folder_name       => 'datastore',
        list              => { datastore => [ $item ] },
    );

    OR
    
    my $result = $hyperv->move_into_folder(
        folder_name       => 'datastore',
        list              => { datastore => [ $item ] },
    );

(Class or Instance method) This method moves host into folder.
Applicable for VMAPI ESX

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< list=>hashref >>

(Required) The List is an hashref where key is the type of item to be added,
and value is the arrayref which is the name of the items to be moved.

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< list => $arrayref >>

(Required) The list of objects to be moved into the folder.

=item C<< task_object => \$task_object >>

(optional) If passed, a task object will be created
and will be assigned to it

=back

=back

=cut

sub move_into_folder {
    my ($pkg_or_obj, @args) = @_;
    $Log->enter() if $may_enter;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            task_objects => { type => ARRAYREF, optional => 1 },
            list         => { type => HASHREF },
        },
    );

    $pkg_or_obj->call_on_apiset( %opts,
        choices => [
                    {
                     method => '_move_into_folder_esx',
                     interface => 'VMAPI',
                     set   => 'ESX'
                    },
                   ],
     );

    $Log->exit() if $may_exit;
} ## end sub move_into_folder

sub _move_into_folder_esx {
    my ($pkg_or_obj, @args) = @_;
    $Log->enter() if $may_enter;

    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my $folder_name       = delete $opts{folder_name};
    my $task_objects      = delete $opts{task_objects};

    my $folder          = $command_interface->folder();
    $folder = _get_folder_entity( $folder, $folder_name );
    my %args;
    my (@list, @task_objs);
    foreach my $group (keys %{$opts{'list'}}) {
        foreach my $item (@{$opts{'list'}->{$group}}) {
            if ($group eq 'datastore') {
                my $mo_ref_ds = $command_interface->hostsystem->datastore;
                foreach my $temp (@{$mo_ref_ds}) {
                    my $response_for_task = $apiset->get_view( mo_ref => $temp );
                    my $task = $response_for_task->get_raw_output();
                    if($task->info->name =~ /$item/i) {
                        push @list, $temp;
                    }
                }
            } elsif($group eq 'virtualmachine') {
                my $mo_ref_ds = $command_interface->hostsystem->vm;
                foreach my $temp (@$mo_ref_ds) {
                    my $response_for_task = $apiset->get_view( mo_ref => $temp );
                    my $task = $response_for_task->get_raw_output();
                    if($task->config->name eq $item) {
                        push @list, $temp;
                    }
                }
            }
        }
        $args{'list'} = \@list;
        my $result;
        $result = $folder->MoveIntoFolder_Task(%args);
        if ($task_objects) {
            my $task_obj = $pkg_or_obj->_get_task_object( apiset => $apiset, result => $result );
            push @task_objs, $task_obj;
        }
    }

    @$task_objects = @task_objs;

    $Log->exit() if $may_exit;
} ## end sub _move_into_folder

=head2 add_host

    $Connect_Spec = {
        hostName => $ci->hostrec()->default_ip(),
        userName => $ci->hostrec()->default_username(),
        password => $ci->hostrec()->default_password(),
        force    => 1,
    };
    my $result = NACL::C::HyperV->add_host(
        command_interface => $ci,
        asConnected       => 0,
        spec              => $Connect_Spec,
    );

    OR
    
    my $result = $hyperv->add_host(
        asConnected       => 0,
        spec              => $Connect_Spec,
    );

(Class or Instance method) This method adds host to the cluster.
Applicable for VMAPI ESX

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

=item C<< license => $string >>

(optional) License to add.

=item C<< spec => host spec >>

(Required) Host spec to add.

=item C<< resourcePool => resource pool >>

(Optional) resource pool.

=item C<< asConnected=>$boolean >>

(Required) add connected flag

=item C<< task_object => \$task_object >>

(optional) If passed, a task object will be created
and will be assigned to it

=back

=back

=cut

sub add_host {
    my ($pkg_or_obj, @args) = @_;
    $Log->enter() if $may_enter;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            task_object => { type => SCALARREF, optional => 1 },
            asConnected => { type => BOOLEAN },
            license     => { type => SCALAR, optional => 1 },
        },
    );
    $pkg_or_obj->call_on_apiset( %opts,
        choices => [
                    {
                     method => '_add_host_esx',
                     interface => 'VMAPI',
                     set   => 'ESX'
                    },
                   ],
    );

    $Log->exit() if $may_exit;
} ## end sub add_host

sub _add_host_esx {
    my ($pkg_or_obj, @args) = @_;
    $Log->enter() if $may_enter;

    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my $task_object     = delete $opts{task_object};

    my $allowed_options = [ qw( spec resourcePool asConnected license ) ];
    my $folder          = $command_interface->cluster_compute_resource();
    my %args;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => $allowed_options,
        target => \%args,
    );
    my $host_connect_spec = HostConnectSpec->new(%{$args{'spec'}});
    $args{'spec'} = $host_connect_spec;
    my $result = $folder->AddHost_Task(%args);
    if ($task_object) {
        $$task_object = $pkg_or_obj->_get_task_object( apiset => $apiset, result => $result );
    }

    $Log->exit() if $may_exit;
} ## end sub _add_host

=head2 remove_host

  my $host  = NACL::C::HyperV->remove_host(
                                        command_interface => $ci );

  OR
  
  my $host = $hyperv->remove_host();

(Class or Instance method) This method removes host from the cluster.
Applicable for VMAPI ESX

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

=item C<< task_object => \$task_object >>

(optional) If passed, a task object will be created
and will be assigned to it

=back

=back

=cut

sub remove_host {
    my ($pkg_or_obj, @args) = @_;
    $Log->enter() if $may_enter;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
                task_object => { type => SCALARREF, optional => 1 },
        },
    );
    $pkg_or_obj->call_on_apiset( %opts,
        choices => [
                    {
                     method => '_remove_host_esx',
                     interface => 'VMAPI',
                     set   => 'ESX'
                    },
                   ],
    );

    $Log->exit() if $may_exit;
} ## end sub remove_host

sub _remove_host_esx {
    my ($pkg_or_obj, @args) = @_;
    $Log->enter() if $may_enter;

    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my $task_object     = delete $opts{task_object};

    my $host = $command_interface->hostsystem();
    my $target_host;
    if ( $host->parent->type eq 'ComputeResource' ) {
        $target_host = $host->parent;
        my $response_for_task = $apiset->get_view( mo_ref => $target_host );
        $target_host = $response_for_task->get_raw_output();
    } elsif ( $host->parent->type eq 'ClusterComputeResource' ) {
        $target_host = $host;
    }

    my $result = $target_host->Destroy_Task();
    if ($task_object) {
        $$task_object = $pkg_or_obj->_get_task_object( apiset => $apiset, result => $result );
    }

    $Log->exit() if $may_exit;
} ## end sub _remove_host

sub _get_xen_host {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts ) = @_;
    
    my %opts = validate_with(
        params => \@opts,
        spec   => {
            xenserver => { type => OBJECT|SCALAR },
            apiset  => { isa  => "NACL::APISet" },
        }
    );
    my $xenserver = $opts{xenserver};
    my $xen_host;
    if (blessed $xenserver) {
        if ($xenserver->isa('NACL::C::HyperV')) {
            $xen_host = $xenserver->hostsystem();
        } else {
            NATE::Exceptions::Argument->throw(
                "Command interface passed is not of type NACL::C::HyperV");
        }
    } elsif ($xenserver =~ /OpaqueRef/) {
        $xen_host = $xenserver;
    } else {
        my $resp = $opts{apiset}->simple_request("host.get_by_name_label", $xenserver);
        $xen_host = $resp->get_parsed_output()->[0];
    }

    $Log->exit() if $may_exit;
    return $xen_host;
}

sub _get_network_ref {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts ) = @_;

    my %args = validate_with(
        params => \@opts,
        spec   => {
            name => { type => SCALAR },
            apiset  => { isa  => "NACL::APISet" },
        }
    );

    # returned value of this is an array ref. Will return 1st value of array
    my $resp = $args{apiset}->simple_request("network.get_by_name_label", $args{name});
    
    my $vm_ref = $resp->get_parsed_output();
    $Log->exit() if $may_exit;
    
    return $vm_ref->[0];
} ## end sub _get_network_ref


sub _get_vm_ref {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts ) = @_;

    my %args = validate_with(
        params => \@opts,
        spec   => {
            vm_name => { type => SCALAR },
            apiset  => { isa  => "NACL::APISet" },
        }
    );

    # returned value of this is an array ref. Will return 1st value of array
    my $resp = $args{apiset}->simple_request("VM.get_by_name_label", $args{vm_name});
    
    my $vm_ref = $resp->get_parsed_output();
    $Log->exit() if $may_exit;
    
    return $vm_ref->[0];
} ## end sub _get_vm_ref

sub _find_target_vm {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts ) = @_;

    my %args = validate_with(
        params => \@opts,
        spec   => { vm_name => { type => HASHREF|SCALAR }, }
    );
    my $vm_name = $args{vm_name};
    my $vms = $pkg_or_obj->virtualmachines();
    foreach my $vm ( @{$vms} ) {
        if ( $vm->config->name =~ /$vm_name/ ) {
            $Log->exit() if $may_exit;
            return $vm;
        }
    }

    $Log->exit() if $may_exit;
    NATE::BaseException->throw("No such vm exists : $vm_name");
} ## end sub _find_target_vm

sub _get_task_object {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts ) = @_;

    my %args = validate_with(
        params => \@opts,
        spec   => {
            result => { type => HASHREF },
            apiset => { isa  => "NACL::APISet" },
        }
    );

    my $response_for_task = $args{apiset}->get_view( mo_ref => $args{result} );
    my $task              = $response_for_task->get_raw_output();

    my $task_obj          = NACL::C::HyperV::Task->new(
        'command_interface' => $pkg_or_obj,
        'key'               => $task->info()->key()
    );

    $Log->exit() if $may_exit;
    return $task_obj;
} ## end sub _get_task_object

sub _get_folder_entity {
    $Log->enter() if $may_enter;
    my ( $folder, $folder_name ) = @_;
    foreach (@$folder) {
        if ( $_->name eq $folder_name ) {
            $Log->exit() if $may_exit;
            return $_;
        }
    }
    $Log->exit() if $may_exit;
    return 0;
} ## end sub _get_folder_entity

=head2 default_apiset_could

The ruleset to default to if values for any of the co-ordinates 'category',
'interface' or 'set' cannot be determined.

This method returns category=Host;interface=VMAPI;set=ESX

Refer to L<default_apiset_could|lib-NACL-C-CommandInterface-pm/default_apiset_could>.

=cut

sub default_apiset_could {
    return {
        category  => "Host",
        interface => "VMAPI",
        set       => "ESX",
    };
} ## end sub default_apiset_could

=head2 default_apiset_must

All HyperV have to be of category "Host", so this method returns
a hash with key 'category' whose value is "Host". Return values
of Interface and SET will depend upon hosttype passed

Refer to L<default_apiset_must|lib-NACL-C-CommandInterface-pm/default_apiset_must>.

=cut

sub default_apiset_must {
    my ($self)     = @_;
    $Log->enter() if $may_enter;
    my $hostrec  = $self->hostrec();
    my $hosttype = $hostrec->hosttype();

    my ($interface, $set);
    if ($hosttype =~ /linux-redhat-xen/) {
        $interface = "XAPI";
        $set = "Xen";
    } else {
        $interface = "VMAPI";
        $set = "ESX";
    }

    $Log->exit() if $may_exit;
    return {
        category  => "Host",
        interface => $interface,
        set       => $set,
    };
} ## end sub default_apiset_must

sub _primary_keys_validate_spec {
    $Log->enter() if $may_enter;
    $Log->exit() if $may_exit;
    return ( 'hypervhost' => { type => SCALAR } );
}

sub resource_record_method_name {
    $Log->enter() if $may_enter;
    $Log->exit() if $may_exit;
    return "hypervhosts";
}

# For this command interface, what is the common prefix for the test parameters?
sub _ci_to_test_param_mapping {
    $Log->enter() if $may_enter;
    $Log->exit() if $may_exit;
    return 'HYPERV_HOST';
}

################################################################################
######################## Private Packages  #####################################
################################################################################

package _VdiskSysem;

use Class::MethodMaker [
    new  => [ '-hash', 'init', 'new' ],
    scalar => 'hostview',
    scalar => 'vdiskmanager',
];

1;
