# $Id$

#
# Copyright (c) 2001-2012 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary JobPrivate Component Module
## @author dl-nacl-dev@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::C::JobPrivate

=head1 DESCRIPTION

C<NACL::C::JobPrivate> is a derived class of
L<NACL::C::Component::ONTAP|lib-NACL-C-Component-ONTAP-pm>. It represents access
to an ONTAP JobPrivate.  A related class is
L<NACL::CS::JobPrivate|lib-NACL-CS-JobPrivate-pm>, which represents
the past state of an ONTAP JobPrivate.

=head1 ATTRIBUTES

=head2 command_interface

See L<NACL::C::Component::command_interface|lib-NACL-C-Component-pm/command_interface>

=head2 id

=head2 node

=cut

package NACL::C::JobPrivate;

use strict;
use warnings;

use NACL::CS::JobPrivate;
use NACL::CS::JobPrivateShowCompleted;
use NACL::ComponentUtils qw (_optional_scalars Dumper);
use NACL::C::Exceptions::JobPrivate::DoesNotExist ();
use NACL::APISet::Exceptions::ResponseException qw(:try);
use base 'NACL::C::Component::ONTAP';
use Params::Validate qw(validate validate_with SCALAR ARRAYREF );
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use NATE::BaseException qw(:try);
use NACL::C::_Mixins::Job qw(wait_on_job);

=head1 METHODS

=head2 new

  my $JobPrivate = NACL::C::JobPrivate->new(command_interface=>$ci,
                                   id =>$id,
                                   node =>$node,
                                  );

Return a new JobPrivate component object with the given attributes.

=cut

use Class::MethodMaker [
    scalar => 'id',
    scalar => 'node',
    scalar => 'vserver',
];

sub init {
    $Log->enter() if $may_enter;
    my ($self, @args) = @_;
    $self->SUPER::init(@args);
    $self->_vserver_context_init(@args);
    $Log->exit() if $may_exit;
} ## end sub init

=head2 find

    my $JobPrivate  = NACL::C::JobPrivate->find(command_interface => $ci, ...);
    my @JobPrivates = NACL::C::JobPrivate->find(command_interface => $ci, ...);

    Uses a CMode CLI or CMode ZAPI APISet.

(Class method)
A constructor of sorts, which discovers which elements are present (by
interacting with the given command interface).  In an array context it
creates a component object for each present element and returns them.
In a scalar context it creates a component object for just one
element.

This method is generally not implemented in individual components. The base
class (L<NACL::C::Component::ONTAP|lib-NACL-C-Component-ONTAP-pm> provides an implementation.
See <L<NACL::C::Component::find|lib-NACL-C-Component-pm/find> for a more
detailed description along with explanation of all the options accepted.

=head2 state

    my $JobPrivate_state = $NACL::C::JobPrivate->state();

    Uses a CMode CLI or CMode ZAPI APISet.

(Instance method)
Fetch the current state of the element associated
with this component, and return a Component State object corresponding
to this component (calling this on NACL::C::JobPrivate would return
a NACL::CS::JobPrivate object)

This method is generally not implemented in individual components. The base
class (L<NACL::C::Component::ONTAP|lib-NACL-C-Component-ONTAP-pm> provides an implementation.
See <L<NACL::C::Component::state|lib-NACL-C-Component-pm/state> for a more
detailed description along with explanation of all the options accepted.

=cut

=head2 pause

   NACL::C::JobPrivate->pause(
                                command_interface => $ci,
                                 id => $id,
                                 node => $node,
                                  ...);

                                 or

    $JobPrivate_object->pause(...);

(Class or instance method)
This method is used to pause a job.

Uses a CMode CLI or CMode ZAPI APISet.

Invokes job-private-pause-iter API for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "id" => $id >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI

Maps to:

CM ZAPI:

<job-private-pause-iter>
<query>
<job-info>
<job-id>$value</job-id>
</job-info>
</query>
</job-private-pause-iter>

=item C<< "node" => $node >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI

Maps to:

CM ZAPI :

<job-private-pause-iter>
<query>
<job-info>
<job-node>$value</job-node>
</job-info>
</query>
</job-private-pause-iter>

=item C<< "vserver" => $vserver >>

(Required)

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

<job-private-pause-iter>
<query>
<job-info>
<job-vserver>$value</job-vserver>
</job-info>
</query>
</job-private-pause-iter>

=item C<< "extended_query" => $string >>

(Optional)

Applicable for CMode CLI/ZAPI.

=item C<< '*' >>

This method takes all other options supported by
"job private pause" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=item C<NACL::C::Exceptions::JobPrivate::DoesNotExist>

This type of exception gets thrown if an attempt is made to pause a
non-existing job.

=back

=back

=cut

sub pause {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params => \@args,
        cli_cmd => 'job private pause',
    );
    my $ret;
    try {
        $pkg_or_obj->call_on_apiset(
            %opts,
            choices => [
                {   method    => '_pause_cmode_cli',
                    interface => 'CLI',
                    set       => 'CMode',
                },
                {   method    => '_pause_cmode_zapi',
                    interface => 'ZAPI',
                    set       => 'CMode',
                },
            ],
            frontend => 'NACL::C::JobPrivate::pause',
        );
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = $_[0];
        NACL::C::Exceptions::JobPrivate::DoesNotExist
            ->detect_convert_and_throw(exception => $exception);
        $Log->exit() if $may_exit;
        $exception->throw();
    };
    $Log->exit() if $may_exit;
    return $ret;
} ## end sub pause

sub _pause_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    my $vserver = $opts{vserver};

    $pkg_or_obj->_base_backend_cmode_cli(
        %opts,
        api => 'job_private_pause',
        vserver_operated_on => \$vserver,
    );

    $Log->exit() if $may_exit;
} ## end sub _pause_cmode_cli

sub _pause_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_iter_helper(
        @args,
        api      => 'job-private-pause-iter',
        show_cmd => 'job private show',
        map      => {
            "id"      => "job-id",
            "node"    => "job-node",
            "vserver" => "job-vserver",
        },
        does_not_exist_error => "entry doesn't exist",
        does_not_exist_exception =>
            'NACL::C::Exceptions::JobPrivate::DoesNotExist'
    );
    $Log->exit() if $may_exit;
} ## end sub _pause_cmode_zapi

=head2 watch_progress

   NACL::C::JobPrivate->watch_progress(
                                command_interface => $ci,
                                 id => $id,
                                 node => $node,
                                  ...);

                                 or

    $JobPrivate_object->watch_progress(...);

(Class or instance method)
This method is used to watch the progress of a job.

Uses a CMode CLI APISet.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "id" => $id >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI.

=item C<< "node" => $node >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI.

=item C<< "interval" => $string >>

Applicable for CMode CLI.

=item C<< "vserver" => $vserver >>

(Required)

Applicable for CMode CLI.

=item C<< "extended_query" => $string >>

(Optional)

Applicable for CMode CLI.

=item C<< '*' >>

This method takes all other options supported by
"job private watch-progress" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub watch_progress {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( interval
                    )
            ),
        },
        cli_cmd => 'job private watch-progress',
    );
    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_watch_progress_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::JobPrivate::watch_progress',
    );

    $Log->exit() if $may_exit;
    return $ret;
} ## end sub watch_progress

sub _watch_progress_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    my $vserver = $opts{vserver};

    $pkg_or_obj->_base_backend_cmode_cli(%opts,
            api => 'job_private_watch_progress',
            vserver_operated_on => \$vserver,);

    $Log->exit() if $may_exit;
} ## end sub _watch_progress_cmode_cli

=head2 show_completed

        my @jobprivateshowcompleted = NACL::C::JobPrivate->show_completed(...);
                                or
        my @jobprivateshowcompleted = $JobPrivate_obj->show_completed(...);

(Class or instance method)
This method is used to display a list of completed jobs.
It eturns objects of type L<NACL::CS::JobPrivateShowCompleted|lib-NACL-CS-JobPrivateShowCompleted-pm>.

The options accepted are the same as those accepted by
L<fetch|lib-NACL-CS-ComponentState-pm/fetch>. (requested_fields, filter, allow_empty)

=cut

sub show_completed {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    $pkg_or_obj->_handle_obj_cs_like_calls(@_);
    my @state_objs = NACL::CS::JobPrivateShowCompleted->fetch(@_);
    $Log->exit() if $may_exit;
    return wantarray ? @state_objs : $state_objs[0];
} ## end sub show_completed

=head2 stop

   NACL::C::JobPrivate->stop(
                                command_interface => $ci,
                                 id => $id,
                                 node => $node,
                                  ...);

                                 or

    $JobPrivate_object->stop(...);

(Class or instance method)
This method is used to stop a job.

Uses a CMode CLI/ZAPI APISet.

Invokes "job-private-stop-iter" API for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "id" => $id >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI

Maps to:

CM ZAPI :

<job-private-stop-iter>
<query>
<job-info>
<job-id>$value</job-id>
</job-info>
</query>
</job-private-stop-iter>

=item C<< "node" => $node >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI

Maps to:

CM ZAPI :

<job-private-stop-iter>
<query>
<job-info>
<job-node>$value</job-node>
</job-info>
</query>
</job-private-stop-iter>

=item C<< "vserver" => $vserver >>

(Required)

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

<job-private-stop-iter>
<query>
<job-info>
<job-vserver>$value</job-vserver>
</job-info>
</query>
</job-private-stop-iter>

=item C<< "extended_query" => $string >>

(Optional)

Applicable for CMode CLI/ZAPI.

=item C<< '*' >>

This method takes all other options supported by
"job private stop" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=item C<NACL::C::Exceptions::JobPrivate::DoesNotExist>

This type of exception gets thrown if an attempt is made to pause a
non-existing job.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=item C<NACL::C::Exceptions::JobPrivate::DoesNotExist>

This type of exception gets thrown if an attempt is made to pause a
non-existing job.

=back

=back

=cut

sub stop {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params => \@args,
        cli_cmd  => 'job private stop',
    );
    my $ret;
    try {
        $pkg_or_obj->call_on_apiset(
            %opts,
            choices => [
                {   method    => '_stop_cmode_cli',
                    interface => 'CLI',
                    set       => 'CMode',
                },
                {   method    => '_stop_cmode_zapi',
                    interface => 'ZAPI',
                    set       => 'CMode',
                },
            ],
            frontend => 'NACL::C::JobPrivate::stop',
        );
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = $_[0];
        NACL::C::Exceptions::JobPrivate::DoesNotExist
            ->detect_convert_and_throw(exception => $exception);
        $Log->exit() if $may_exit;
        $exception->throw();
    };
    $Log->exit() if $may_exit;
    return $ret;
} ## end sub stop

sub _stop_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    my $vserver = $opts{vserver};

    $pkg_or_obj->_base_backend_cmode_cli(%opts,
            api => 'job_private_stop',
            vserver_operated_on => \$vserver,
            );

    $Log->exit() if $may_exit;
} ## end sub _stop_cmode_cli

sub _stop_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;

    $pkg_or_obj->_iter_helper(
        @args,
        api      => 'job-private-stop-iter',
        show_cmd => 'job private show',
        map      => {
            "id"      => "job-id",
            "node"    => "job-node",
            "vserver" => "job-vserver",
        },
        does_not_exist_error => "entry doesn't exist",
        does_not_exist_exception =>
            'NACL::C::Exceptions::JobPrivate::DoesNotExist'
    );
    $Log->exit() if $may_exit;
} ## end sub _stop_cmode_zapi

=head2 delete

   NACL::C::JobPrivate->delete(
                                command_interface => $ci,
                                 id => $id,
                                 node => $node,
                                  ...);

                                 or

    $JobPrivate_object->delete(...);

(Class or instance method)
This method is used to delete a job.

Uses a CMode CLI/ZAPI APISet.

Invokes "job-private-delete-iter" API for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "id" => $id >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI

Maps to:

CM ZAPI :

<job-private-delete-iter>
<query>
<job-info>
<job-id>$value</job-id>
</job-info>
</query>
</job-private-delete-iter>

=item C<< "node" => $node >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI

Maps to:

CM ZAPI :

<job-private-delete-iter>
<query>
<job-info>
<job-node>$value</job-node>
</job-info>
</query>
</job-private-delete-iter>

=item C<< "vserver" => $vserver >>

(Optional)

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

<job-private-delete-iter>
<query>
<job-info>
<job-vserver>$value</job-vserver>
</job-info>
</query>
</job-private-delete-iter>

=item C<< "extended_query" => $string >>

(Optional)

Applicable for CMode CLI/ZAPI.

=item C<< '*' >>

This method takes all other options supported by
"job private delete" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=item C<NACL::C::Exceptions::JobPrivate::DoesNotExist>

This type of exception gets thrown if an attempt is made to pause a
non-existing job.

=back

=back

=cut

sub delete {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params => \@args,
        cli_cmd => 'job private delete',
        additional_spec => {
            'vserver' => {type => SCALAR, optional => 1},
        },
    );
    my $ret;
    try {
        $pkg_or_obj->call_on_apiset(
            %opts,
            choices => [
                {   method    => '_delete_cmode_cli',
                    interface => 'CLI',
                    set       => 'CMode',
                },
                {   method    => '_delete_cmode_zapi',
                    interface => 'ZAPI',
                    set       => 'CMode',
                },
            ],
            frontend => 'NACL::C::JobPrivate::delete',
        );
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = $_[0];
        NACL::C::Exceptions::JobPrivate::DoesNotExist
            ->detect_convert_and_throw(exception => $exception);
        $Log->exit() if $may_exit;
        $exception->throw();
    };

    $Log->exit() if $may_exit;
    return $ret;
} ## end sub delete

sub _delete_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    my $vserver = $opts{vserver};

    $pkg_or_obj->_base_backend_cmode_cli(%opts,
            api => 'job_private_delete',
            vserver_operated_on => \$vserver,);

    $Log->exit() if $may_exit;
} ## end sub _delete_cmode_cli

sub _delete_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;

    $pkg_or_obj->_iter_helper(
        @args,
        api      => 'job-private-delete-iter',
        show_cmd => 'job private show',
        map      => {
            "id"      => "job-id",
            "node"    => "job-node",
            "vserver" => "job-vserver",
        },
        does_not_exist_error => "entry doesn't exist",
        does_not_exist_exception =>
            'NACL::C::Exceptions::JobPrivate::DoesNotExist'
    );
    $Log->exit() if $may_exit;
} ## end sub _delete_cmode_zapi

=head2 softpause

   NACL::C::JobPrivate->softpause(
                                command_interface => $ci,
                                 id => $id,
                                 node => $node,
                                  ...);

                                 or

    $JobPrivate_object->softpause(...);

(Class or instance method)
This method is used to soft pause a job.

Uses a CMode CLI/ZAPI APISet.

Invokes "job-private-soft-pause-iter" API for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "id" => $id >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI

Maps to:

CM ZAPI :

<job-private-soft-pause-iter>
<query>
<job-info>
<job-id>&value</job-id>
</job-info>
</query>
</job-private-soft-pause-iter>

=item C<< "node" => $node >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI

Maps to:

CM ZAPI :

<job-private-soft-pause-iter>
<query>
<job-info>
<job-node>&value</job-node>
</job-info>
</query>
</job-private-soft-pause-iter>

=item C<< "vserver" => $vserver >>

(Required)

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

<job-private-soft-pause-iter>
<query>
<job-info>
<job-vserver>$value</job-vserver>
</job-info>
</query>
</job-private-soft-pause-iter>

=item C<< "extended_query" => $string >>

(Optional)

Applicable for CMode CLI/ZAPI.

=item C<< '*' >>

This method takes all other options supported by
"job private softpause" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=item C<NACL::C::Exceptions::JobPrivate::DoesNotExist>

This type of exception gets thrown if an attempt is made to pause a
non-existing job.

=back

=back

=cut

sub softpause {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params => \@args,
        cli_cmd  => 'job private softpause',
    );
    my $ret;
    try {
        $pkg_or_obj->call_on_apiset(
            %opts,
            choices => [
                {   method    => '_softpause_cmode_cli',
                    interface => 'CLI',
                    set       => 'CMode',
                },
                {   method    => '_softpause_cmode_zapi',
                    interface => 'ZAPI',
                    set       => 'CMode',
                },
            ],
            frontend => 'NACL::C::JobPrivate::softpause',
        );
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = $_[0];
        NACL::C::Exceptions::JobPrivate::DoesNotExist
            ->detect_convert_and_throw(exception => $exception);
        $Log->exit() if $may_exit;
        $exception->throw();
    };

    $Log->exit() if $may_exit;
    return $ret;
} ## end sub softpause

sub _softpause_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    my $vserver = $opts{vserver};

    $pkg_or_obj->_base_backend_cmode_cli(%opts,
            api => 'job_private_softpause',
            vserver_operated_on => \$vserver,);

    $Log->exit() if $may_exit;
} ## end sub _softpause_cmode_cli

sub _softpause_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;

    $pkg_or_obj->_iter_helper(
        @args,
        api      => 'job-private-soft-pause-iter',
        show_cmd => 'job private show',
        map      => {
            "id"      => "job-id",
            "node"    => "job-node",
            "vserver" => "job-vserver",
        },
        does_not_exist_error => "entry doesn't exist",
        does_not_exist_exception =>
            'NACL::C::Exceptions::JobPrivate::DoesNotExist'
    );

    $Log->exit() if $may_exit;
} ## end sub _softpause_cmode_zapi

=head2 resume

   NACL::C::JobPrivate->resume(
                                command_interface => $ci,
                                 id => $id,
                                 node => $node,
                                  ...);

                                 or

    $JobPrivate_object->resume(...);

(Class or instance method)
This method is used to resume a job.

Uses a CMode CLI/ZAPI APISet.

Invokes "job-private-resume-iter" API for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "id" => $id >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI

Maps to:

CM ZAPI :

<job-private-resume-iter>
<query>
<job-info>
<job-id>$value</job-id>
</job-info>
</query>
</job-private-resume-iter>

=item C<< "node" => $node >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI

Maps to:

CM ZAPI :

<job-private-resume-iter>
<query>
<job-info>
<job-node>$value</job-node>
</job-info>
</query>
</job-private-resume-iter>

=item C<< "vserver" => $vserver >>

(Required)

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

<job-private-resume-iter>
<query>
<job-info>
<job-vserver>$value</job-vserver>
</job-info>
</query>
</job-private-resume-iter>

=item C<< "extended_query" => $string >>

(Optional)

Applicable for CMode CLI/ZAPI.

=item C<< '*' >>

This method takes all other options supported by
"job private resume" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=item C<NACL::C::Exceptions::JobPrivate::DoesNotExist>

This type of exception gets thrown if an attempt is made to pause a
non-existing job.

=back

=back

=cut

sub resume {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params => \@args,
        cli_cmd   => 'job private resume',
    );
    my $ret;
    try {
        $pkg_or_obj->call_on_apiset(
            %opts,
            choices => [
                {   method    => '_resume_cmode_cli',
                    interface => 'CLI',
                    set       => 'CMode',
                },
                {   method    => '_resume_cmode_zapi',
                    interface => 'ZAPI',
                    set       => 'CMode',
                },
            ],
            frontend => 'NACL::C::JobPrivate::resume',
        );
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = $_[0];
        NACL::C::Exceptions::JobPrivate::DoesNotExist
            ->detect_convert_and_throw(exception => $exception);
        $Log->exit() if $may_exit;
        $exception->throw();
    };

    $Log->exit() if $may_exit;
    return $ret;
} ## end sub resume

sub _resume_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    my $vserver = $opts{vserver};

    $pkg_or_obj->_base_backend_cmode_cli(%opts,
            api => 'job_private_resume',
            vserver_operated_on => \$vserver,);

    $Log->exit() if $may_exit;
} ## end sub _resume_cmode_cli

sub _resume_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;

    $pkg_or_obj->_iter_helper(
        @args,
        api      => 'job-private-resume-iter',
        show_cmd => 'job private show',
        map      => {
            "id"      => "job-id",
            "node"    => "job-node",
            "vserver" => "job-vserver",
        },
        does_not_exist_error => "entry doesn't exist",
        does_not_exist_exception =>
            'NACL::C::Exceptions::JobPrivate::DoesNotExist'
    );

    $Log->exit() if $may_exit;
} ## end sub _resume_cmode_zapi

sub _primary_keys_validate_spec {
    return (
        'id'      => {type => SCALAR},
        'node'    => {type => SCALAR},
        'vserver' => {type => SCALAR },
    );
} ## end sub _primary_keys_validate_spec

1;
