#
# Copyright (c) 2013 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary LunIgroup Component Module
## @author dl-nacl-dev@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::C::LunIgroup

=head1 DESCRIPTION

C<NACL::C::LunIgroup> is a derived class of
L<NACL::C::Component::ONTAP|lib-NACL-C-Component-ONTAP-pm>. It represents access
to an ONTAP LunIgroup.  A related class is
L<NACL::CS::LunIgroup|lib-NACL-CS-LunIgroup-pm>, which represents
the past state of an ONTAP LunIgroup.

=head1 ATTRIBUTES

=head2 command_interface

See L<NACL::C::Component::command_interface|lib-NACL-C-Component-pm/command_interface>

=head2 igroup

=head2 vserver

=cut

package NACL::C::LunIgroup;

use strict;
use warnings;
use NACL::CS::LunIgroup;
use NACL::ComponentUtils qw (_optional_scalars Dumper);
use NACL::C::Exceptions::LunIgroup::AlreadyExists;
use NACL::C::Exceptions::LunIgroup::DoesNotExist;
use NACL::C::Exceptions::LunIgroup::MapExists;
use NACL::C::Exceptions::LunIgroup::AlreadyBound;
use NACL::C::Exceptions::LunIgroup::NotBound;
use NATE::BaseException qw(:try);
use NATE::Exceptions::Argument qw(:try);
use base 'NACL::C::Component::ONTAP';
use Params::Validate qw(validate validate_with SCALAR ARRAYREF UNDEF);
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();

=head1 METHODS

=head2 new

  my $LunIgroup = NACL::C::LunIgroup->new(command_interface=>$ci,
                                   igroup =>$igroup,
                                   vserver =>$vserver,
                    );

Return a new LunIgroup component object with the given attributes.

=cut

use Class::MethodMaker [
    scalar => 'igroup',
    scalar => 'vserver',
];

sub init {
    $Log->enter() if $may_enter;
    my ($self, @args) = @_;
    $self->SUPER::init(@args);
    $self->_vserver_context_init(@args);
    $Log->exit() if $may_exit;
} ## end sub init

=head2 find

    my $LunIgroup  = NACL::C::LunIgroup->find(command_interface => $ci, ...);
    my @LunIgroups = NACL::C::LunIgroup->find(command_interface => $ci, ...);

(Class method)
A constructor of sorts, which discovers which elements are present (by
interacting with the given command interface).  In an array context it
creates a component object for each present element and returns them.
In a scalar context it creates a component object for just one
element.

This method is generally not implemented in individual components. The base
class (L<NACL::C::Component::ONTAP|lib-NACL-C-Component-ONTAP-pm> provides an implementation.
See <L<NACL::C::Component::find|lib-NACL-C-Component-pm/find> for a more
detailed description along with explanation of all the options accepted.

=head2 state

    my $LunIgroup_state = $NACL::C::LunIgroup->state();

(Instance method)
Fetch the current state of the element associated
with this component, and return a Component State object corresponding
to this component (calling this on NACL::C::LunIgroup would return
a NACL::CS::LunIgroup object)

This method is generally not implemented in individual components. The base
class (L<NACL::C::Component::ONTAP|lib-NACL-C-Component-ONTAP-pm> provides an implementation.
See <L<NACL::C::Component::state|lib-NACL-C-Component-pm/state> for a more
detailed description along with explanation of all the options accepted.

=cut

=head2 rename

   NACL::C::LunIgroup->rename(
                                command_interface => $ci,
                                igroup => $igroup,
                                vserver => $vserver,
                                  ...);

                    or

    $LunIgroup_object->rename(...);

(Class or instance method)
This method is used to rename an existing initiator group.

Uses a CMode CLI/ZAPI, 7Mode CLI APISet.

Invokes "igroup-rename" API for CMode ZAPI.

Invokes "igroup rename" API for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "igroup" => $igroup >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI, 7Mode CLI

Maps to:

CM ZAPI: initiator-group-name

7M CLI : initiator group name (Programmatic name: "initiator-group")

=item C<< "vserver" => $vserver >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: ZAPI will be invoked in the context of the vserver specified to this
option.

=item C<< "new-name" => $string >>

Applicable for CMode CLI/ZAPI, 7Mode CLI

Maps to:

CM ZAPI: initiator-group-new-name

7Mode CLI: new igroup name (Programmatic name: "new-igroup-name")

=item C<< "extended_query" => $string >>

(Optional)

Applicable for CMode CLI.

=item C<< '*' >>

This method takes all other options supported by
"lun igroup rename" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub rename {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( new-name
                    )
            ),
        },
        cli_cmd => 'lun igroup rename',
    );
    my $vserver;
    $opts{vserver_operated_on} = \$vserver;
    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_rename_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_rename_7mode_cli',
                interface => 'CLI',
                set       => '7Mode',
            },
            {   method    => '_rename_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::LunIgroup::rename',
    );

    $pkg_or_obj->igroup($opts{'new-name'}) if (ref $pkg_or_obj);

    $Log->exit() if $may_exit;
    return $ret;
} ## end sub rename

sub _rename_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@args,  api => 'lun_igroup_rename');
    $Log->exit() if $may_exit;
} ## end sub _rename_cmode_cli

sub _rename_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );

    delete $opts{command_interface};
    my $apiset = delete $opts{apiset};

    my %valid_args;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        map    => {
            'igroup'   => 'initiator-group',
            'new-name' => 'new-igroup-name',
        },
        target => \%valid_args,
    );

    $apiset->igroup_rename(%valid_args);
    $Log->exit() if $may_exit;
} ## end sub _rename_7mode_cli

sub _rename_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_base_backend_cmode_zapi(@args,
            api    => 'igroup_rename',
            map    => {
                "igroup"   => "initiator-group-name",
                "new-name" => "initiator-group-new-name"
            },
    );
    $Log->exit() if $may_exit;
} ## end sub _rename_cmode_zapi

=head2 remove

   NACL::C::LunIgroup->remove(
                                command_interface => $ci,
                                igroup => $igroup,
                                vserver => $vserver,
                                  ...);

                            or

    $LunIgroup_object->remove(...);

(Class or instance method)
This method is used to remove initiator(s) from an initiator group.

Uses a CMode CLI/ZAPI, 7Mode CLI APISet.

Invokes "igroup-remove" API for CMode ZAPI.

Invokes "igroup remove" API for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "igroup" => $igroup >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI, 7Mode CLI

Maps to:

CM ZAPI: initiator-group-name

7M CLI : initiator group name (Programmatic name: "initiator-group")

=item C<< "vserver" => $vserver >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: ZAPI will be invoked in the context of the vserver specified to this
option.

=item C<< "force" => $string >>

Applicable for CMode CLI/ZAPI, 7Mode CLI

Maps to:

CM ZAPI : force

7M CLI  : -f (Programmatic name: "force")

=item C<< "initiator" => [ $initiator1, $initiator2, ... ] >>

Applicable for CMode CLI/ZAPI, 7Mode CLI.

Maps to:

CM ZAPI : initiator
7M CLI  : node name (Programmatic name: "node")

=item C<< "extended_query" => $string >>

(Optional)

Applicable for CMode CLI.

=item C<< '*' >>

This method takes all other options supported by
"lun igroup remove" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub remove {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            "initiator" => {type => ARRAYREF, optional => 1},
            _optional_scalars(
                qw( force
                    )
            ),
        },
        cli_cmd => 'lun igroup remove',
    );
    my $vserver;
    $opts{vserver_operated_on} = \$vserver;
    my $ret = $pkg_or_obj->call_on_apiset_with_exceptions(
        %opts,
        choices => [
            {   method    => '_remove_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_remove_7mode_cli',
                interface => 'CLI',
                set       => '7Mode',
            },
            {   method    => '_remove_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::LunIgroup::remove',
        nacl_exceptions => [qw(DoesNotExist)],
    );

    $Log->exit() if $may_exit;
    return $ret;
} ## end sub remove

sub _remove_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@args,  api => 'lun_igroup_remove');
    $Log->exit() if $may_exit;
} ## end sub _remove_cmode_cli

sub _remove_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );

    delete $opts{command_interface};
    my $apiset                = delete $opts{apiset};
    my $allowed_7mode_options = [qw( force )];
    my %valid_args;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        map    => {
            'igroup'    => 'initiator-group',
            'initiator' => 'node',
        },
        copy   => $allowed_7mode_options,
        target => \%valid_args,
    );
    $valid_args{'node'} = join ' ', @{$valid_args{'node'}};

    $apiset->igroup_remove(%valid_args);
    $Log->exit() if $may_exit;
} ## end sub _remove_7mode_cli

sub _remove_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    my $allowed_cmode_options = [
        qw( initiator force
            )
    ];
    my $arr_ref       = $opts{initiator};
    my @initiator_arr = @$arr_ref;
    my $exception;
    foreach my $element (@initiator_arr) {
        $opts{initiator} = $element;
        try{
            $pkg_or_obj->_base_backend_cmode_zapi(
            %opts,
            api  => 'igroup-remove',
            copy => $allowed_cmode_options,
            map  => {"igroup" => "initiator-group-name"},
            );
        } otherwise {
            # Burt930360 Don't fail immediately, but execute all ZAPIs first
            $exception = $_[0];
        };
    }
    if ($exception){
        $Log->exit() if $may_exit;
        $exception->throw();
    }
    $Log->exit() if $may_exit;
}

=head2 bind

   NACL::C::LunIgroup->bind(
                                command_interface => $ci,
                                igroup => $igroup,
                                vserver => $vserver,
                                  ...);

                                or

    $LunIgroup_object->bind(...);

(Class or instance method)
This method is used to bind an existing initiator group to a given portset.

Uses a CMode CLI/ZAPI, 7Mode CLI APISet.

Invokes "igroup-bind-portset" API for CMode ZAPI.

Invokes "igroup bind" command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "igroup" => $igroup >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI, 7Mode CLI

Maps to:

CM ZAPI: initiator-group-name

7Mode CLI : initiator group name (Programmatic name: "initiator-group")

=item C<< "vserver" => $vserver >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: ZAPI will be invoked in the context of the vserver specified to this
option.

=item C<< "portset-name" => $string >>

Applicable for CMode CLI.

=item C<< "portset" => $string >>

Applicable for CMode CLI/ZAPI, 7Mode CLI

Maps to:

CM ZAPI : portset-name

7M CLI  : portset name (Programmatic name: "portset")

=item C<< "extended_query" => $string >>

(Optional)

Applicable for CMode CLI.

=item C<< '*' >>

This method takes all other options supported by
"lun igroup bind" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub bind {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( portset-name portset
                    )
            ),
        },
        cli_cmd => 'lun igroup bind',
    );
    my $vserver;
    $opts{vserver_operated_on} = \$vserver;
    my $ret = $pkg_or_obj->call_on_apiset_with_exceptions(
        %opts,
        choices => [
            {   method    => '_bind_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_bind_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
            {   method    => '_bind_7mode_cli',
                interface => 'CLI',
                set       => '7Mode',
            },
        ],
        frontend => 'NACL::C::LunIgroup::bind',
        nacl_exceptions => [qw(AlreadyBound)],
    );

    $Log->exit() if $may_exit;
    return $ret;
} ## end sub bind

sub _bind_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@args,  api => 'lun_igroup_bind');
    $Log->exit() if $may_exit;
} ## end sub _bind_cmode_cli

sub _bind_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_base_backend_cmode_zapi(@args,
            api    => 'igroup_bind_portset',
            map    => {
                "igroup"  => "initiator-group-name",
                "portset" => "portset-name"
            },
    );
    $Log->exit() if $may_exit;
} ## end sub _bind_cmode_zapi

sub _bind_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface     = delete $opts{command_interface};
    my $apiset                = delete $opts{apiset};
    my $allowed_7mode_options = [qw( portset )];

    my %valid_args;
    $pkg->_hash_copy(
        source => \%opts,
        map    => {'igroup' => 'initiator-group',},
        copy   => $allowed_7mode_options,
        target => \%valid_args,
    );
    %opts = %valid_args;

    $apiset->igroup_bind(%opts);
    $Log->exit() if $may_exit;
} ## end sub _bind_7mode_cli

=head2 add

   NACL::C::LunIgroup->add(
                                command_interface => $ci,
                                igroup => $igroup,
                                vserver => $vserver,
                                  ...);

                            or

    $LunIgroup_object->add(...);

(Class or instance method)
This method is used to add initiator(s) to an initiator group.

Uses a CMode CLI/ZAPI, 7Mode CLI APISet.

Invokes "igroup-add" API for CMode ZAPI.

Invokes "igroup add" command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "igroup" => $igroup >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI, 7Mode CLI

Maps to:

CM ZAPI: initiator-group-name

7Mode CLI : initiator group name (Programmatic name: "initiator-group")

=item C<< "vserver" => $vserver >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: ZAPI will be invoked in the context of the vserver specified to this
option.

=item C<< "force" => $string >>

Applicable for CMode CLI/ZAPI, 7Mode CLI

Maps to:

CM ZAPI : force

7M CLI  : -f (Programmatic name: "force")

=item C<< "initiator" => [ $initiator1, $initiator2, ... ] >>

Applicable for CMode CLI/ZAPI

Maps to:

CM ZAPI: initiator

7M CLI : node name (Programmatic name: "node")

=item C<< "extended_query" => $string >>

(Optional)

Applicable for CMode CLI.

=item C<< '*' >>

This method takes all other options supported by
"lun igroup add" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub add {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            "initiator" => {type => ARRAYREF, optional => 1},
            _optional_scalars(
                qw( force
                    )
            ),
        },
        cli_cmd => 'lun igroup add',
    );
    my $vserver;
    $opts{vserver_operated_on} = \$vserver;
    my $ret = $pkg_or_obj->call_on_apiset_with_exceptions(
        %opts,
        choices => [
            {   method    => '_add_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_add_7mode_cli',
                interface => 'CLI',
                set       => '7Mode',
            },
            {   method    => '_add_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::LunIgroup::add',
        nacl_exceptions => [qw(AlreadyExists)],
    );

    $Log->exit() if $may_exit;
    return $ret;
} ## end sub add

sub _add_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@args,  api  => 'lun_igroup_add');
    $Log->exit() if $may_exit;
} ## end sub _add_cmode_cli

sub _add_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );

    delete $opts{command_interface};
    my $apiset                = delete $opts{apiset};
    my $allowed_7mode_options = [qw( force )];
    my %valid_args;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        map    => {
            'igroup'    => 'initiator-group',
            'initiator' => 'node',
        },
        copy   => $allowed_7mode_options,
        target => \%valid_args,
    );
    $valid_args{'node'} = join ' ', @{$valid_args{'node'}};

    $apiset->igroup_add(%valid_args);
    $Log->exit() if $may_exit;

} ## end sub _add_7mode_cli

sub _add_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    my $allowed_cmode_options = [
        qw( initiator force
            )
    ];
    my $arr_ref       = $opts{initiator};
    my @initiator_arr = @$arr_ref;
    my $exception;
    foreach my $element (@initiator_arr) {
        $opts{initiator} = $element;
        try{
            $pkg_or_obj->_base_backend_cmode_zapi(
            %opts,
            api  => 'igroup_add',
            copy => $allowed_cmode_options,
	    map  => {"igroup" => "initiator-group-name"},
            );
        }otherwise {
            # Burt930360 Don't fail immediately, but execute all ZAPIs first
            $exception = $_[0];
        };
    }
    if ($exception){
        $Log->exit() if $may_exit;
        $exception->throw();
    }
    $Log->exit() if $may_exit;
} ## end sub _add_cmode_zapi
=head2 set

   NACL::C::LunIgroup->set(
                                command_interface => $ci,
                                igroup => $igroup,
                                vserver => $vserver,
                                  ...);

                                or

    $LunIgroup_object->set(...);

(Class or instance method)
This method is used to sets an attribute for an initiator group.

Uses a CMode CLI/ZAPI APISet.

Invokes "igroup-set-attribute" API for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "igroup" => $igroup >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI

Maps to:

CM ZAPI: initiator-group-name

=item C<< "vserver" => $vserver >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: ZAPI will be invoked in the context of the vserver specified to this
option.

=item C<< "attribute" => $string >>

Applicable for CMode CLI/ZAPI

Maps to:

CM ZAPI : attribute

=item C<< "value" => $string >>

Applicable for CMode CLI/ZAPI

Maps to:

CM ZAPI : value
=item C<< "force" => $string >>

Applicable for CMode CLI/ZAPI

Maps to:

CM ZAPI : force

=item C<< "extended_query" => $string >>

(Optional)

Applicable for CMode CLI.

=item C<< '*' >>

This method takes all other options supported by
"lun igroup set" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub set {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( attribute value force
                    )
            ),
        },
        cli_cmd => 'lun igroup set',
    );
    my $vserver;
    $opts{vserver_operated_on} = \$vserver;
    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_set_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_set_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::LunIgroup::set',
    );

    $Log->exit() if $may_exit;
    return $ret;
} ## end sub set

sub _set_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@args,  api => 'lun_igroup_set');
    $Log->exit() if $may_exit;
} ## end sub _set_cmode_cli

sub _set_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
	my $allowed_cmode_options = [
        qw( attribute value force
            )
    ];
    $pkg_or_obj->_base_backend_cmode_zapi(@args,
            api    => 'igroup_set_attribute',
            copy   => $allowed_cmode_options,
            map    => {"igroup" => "initiator-group-name"},
    );
    $Log->exit() if $may_exit;
} ## end sub _set_cmode_zapi

=head2 create

    my $LunIgroup_object    = NACL::C::LunIgroup->create(
                                command_interface => $ci,
                                igroup => $igroup,
                                vserver => $vserver,
                                  ...);

(Class method)
This method is used to create a new initiator group.

Uses a CMode/ZAPI, 7Mode CLI APISet.

Invokes "igroup-create","igroup-add" API for CMode ZAPI.

Invokes "igroup create" command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "igroup" => $igroup >>

(Required)

Applicable for CMode CLI/ZAPI, 7Mode CLI

Maps to:

CM ZAPI: initiator-group-name

7M CLI : initiator group name (Programmatic name: "initiator-group")

=item C<< "vserver" => $vserver >>

(Required)

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: ZAPI will be invoked in the context of the vserver specified to this
option.

=item C<< "fcp" => $string >>

Applicable for CMode CLI, 7Mode CLI

Maps to:

7M CLI : -f (Programmatic name: "initiator-type-fcp")

=item C<< "portset" => $string >>

Applicable for CMode CLI/ZAPI, 7Mode CLI

Maps to:

CM ZAPI: bind-portset

7M CLI : -a <value> (Programmatic name: "port-set")

=item C<< "ostype" => $string >>

Applicable for CMode CLI/ZAPI, 7Mode CLI

Maps to:

CM ZAPI: os-type

7M CLI : -t <value> (Programmatic name: "os-type")

=item C<< "protocol" => $string >>

Applicable for CMode CLI/ZAPI

Maps to:

CM ZAPI: initiator-group-type

=item C<< "iscsi" => $string >>

Applicable for 7Mode CLI

Maps to:

7M CLI : -i (Programmatic name: "initiator-type-iscsi")

=item C<< "initiator" => [ $initiator1, $initiator2, ... ] >>

Applicable for 7Mode CLI/CMode CLI/ZAPI.

Maps to:

7M CLI : node name (Programmatic name: "node")

CM ZAPI : Invokes 'igroup-add' API for CMode ZAPI

=item C<< 'protocol-type' => $string >>

Applicable for CMode CLI.

=item C<< "extended_query" => $string >>

(Optional)

Applicable for CMode CLI.

=item C<< '*' >>

This method takes all other options supported by
"lun igroup create" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=item C<NACL::C::Exceptions::LunIgroup::AlreadyExists>

This type of exception gets thrown if an attempt is made to create a
lun igroup that already exists.

=back

=back

=cut

sub create {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            "initiator" => {type => ARRAYREF, optional => 1},
            _optional_scalars(
                qw( fcp portset protocol ostype iscsi
                    protocol-type
                    )
            ),
        },
    cli_cmd => 'lun igroup create',
    );
    my $vserver;
    $opts{vserver_operated_on} = \$vserver;
    try {
        $pkg->call_on_apiset(
            %opts,
            choices => [
                {   method    => '_create_cmode_cli',
                    interface => 'CLI',
                    set       => 'CMode',
                },
                {   method    => '_create_cmode_zapi',
                    interface => 'ZAPI',
                    set       => 'CMode',
                },
                {   method    => '_create_7mode_cli',
                    interface => 'CLI',
                    set       => '7Mode',
                },
            ],
            frontend => 'NACL::C::LunIgroup::create'
        );
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = $_[0];
        NACL::C::Exceptions::LunIgroup::AlreadyExists
            ->detect_convert_and_throw(exception => $exception);
        $Log->exit() if $may_exit;
        $exception->throw();
    };
    my $obj = $pkg->new(
        command_interface => $opts{command_interface},
        igroup            => $opts{'igroup'},
        vserver           => $opts{vserver},
    );
    $Log->exit() if $may_exit;
    return $obj;
} ## end sub create

sub _create_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_base_backend_cmode_cli(@args,  api => 'lun_igroup_create');
    $Log->exit() if $may_exit;
} ## end sub _create_cmode_cli

sub _create_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg,%opts) = @_;
    my $initiator;
    my %target;
    #Burt 797192, If not passed the argument for protocol,
    #the protocol default to 'mixed'
    $opts{protocol} ||= 'mixed';
    if (defined($opts{initiator})){
        $initiator  = delete $opts{initiator} ;
    }
    my $exception;
    try{
        $pkg->_base_backend_cmode_zapi(
            %opts,
            api  => 'igroup_create',
            map  => {
                "igroup"   => "initiator-group-name",
                "ostype"   => "os-type",
                "protocol" => "initiator-group-type",
                "portset"  => "bind-portset",
            },
        );
    } otherwise {
        #Burt952250 Don't fail immediately, but execute all ZAPIs first
        $exception = $_[0];
    };
    if (defined $initiator ) {
        $pkg->_hash_copy(
                source => \%opts,
                copy => [qw(command_interface vserver force igroup)],
                target => \%target,
                );
        $target{initiator} = $initiator;
        $pkg->add(
                %target,
                apiset_should => {interface => 'ZAPI'},
                );
    }
    if ($exception){
        $Log->exit() if $may_exit;
        $exception->throw();
    }
    $Log->exit() if $may_exit;
}

sub _create_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my %valid_args;
    if ($opts{'protocol'} =~ /fcp/i) {
        $opts{'fcp'} = 1;
    } elsif ($opts{'protocol'} =~ /iscsi/i) {
        $opts{'iscsi'} = 1;
    }
    $pkg->_hash_copy(
        source => \%opts,
        map    => {
            'igroup'    => 'initiator-group',
            'fcp'       => 'initiator-type-fcp',
            'iscsi'     => 'initiator-type-iscsi',
            'ostype'    => 'os-type',
            'portset'   => 'port-set',
            'initiator' => 'node',
        },
        target => \%valid_args,
    );
    %opts = %valid_args;

    if (defined $opts{node}) {
        $opts{node} = join ' ', @{$opts{node}};
    }

    $apiset->igroup_create(%opts);
    $Log->exit() if $may_exit;
    return $pkg->new(
        command_interface => $command_interface,
        igroup            => $opts{'initiator-group'},
    );
} ## end sub _create_7mode_cli

=head2 unbind

   NACL::C::LunIgroup->unbind(
                                command_interface => $ci,
                                igroup => $igroup,
                                vserver => $vserver,
                                  ...);

                            or

    $LunIgroup_object->unbind(...);

(Class or instance method)
This method is used to unbind an existing initiator group from a portset.

Uses a CMode CLI/ZAPI, 7Mode CLI APISet.

Invokes "igroup-unbind-portset" API for CMode ZAPI.

Invokes "igroup unbind" command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "igroup" => $igroup >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI, 7Mode CLI

Maps to:

CM ZAPI: initiator-group-name

7M CLI : initiator group name (Programmatic name: "initiator-group")

=item C<< "vserver" => $vserver >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: ZAPI will be invoked in the context of the vserver specified to this
option.

=item C<< "extended_query" => $string >>

(Optional)

Applicable for CMode CLI.

=item C<< '*' >>

This method takes all other options supported by
"lun igroup unbind" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub unbind {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params => \@args,
        cli_cmd    => 'lun igroup unbind',
    );
    my $vserver;
    $opts{vserver_operated_on} = \$vserver;
    my $ret = $pkg_or_obj->call_on_apiset_with_exceptions(
        %opts,
        choices => [
            {   method    => '_unbind_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_unbind_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
            {   method    => '_unbind_7mode_cli',
                interface => 'CLI',
                set       => '7Mode',
            },
        ],
        frontend => 'NACL::C::LunIgroup::unbind',
        nacl_exceptions => [qw(NotBound)],
    );

    $Log->exit() if $may_exit;
    return $ret;
} ## end sub unbind

sub _unbind_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@args,  api => 'lun_igroup_unbind');
    $Log->exit() if $may_exit;
} ## end sub _unbind_cmode_cli

sub _unbind_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_base_backend_cmode_zapi(@args,
            api    => 'igroup_unbind_portset',
            map    => {"igroup" => "initiator-group-name",},
    );
    $Log->exit() if $may_exit;
} ## end sub _unbind_cmode_zapi

sub _unbind_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my %valid_args;
    $pkg->_hash_copy(
        source => \%opts,
        map    => {'igroup' => 'initiator-group',},
        target => \%valid_args,
    );
    %opts = %valid_args;

    $apiset->igroup_unbind(%opts);
    $Log->exit() if $may_exit;

} ## end sub _unbind_7mode_cli

=head2 modify

   NACL::C::LunIgroup->modify(
                                command_interface => $ci,
                                igroup => $igroup,
                                vserver => $vserver,
                                  ...);

                            or

    $LunIgroup_object->modify(...);

(Class or instance method)
This method is used to modify an existing initiator group.

Uses a CMode CLI/ZAPI APISet.

Invokes "igroup-set-attribute" API for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "igroup" => $igroup >>

Applicable for CMode CLI/ZAPI

Maps to:

CM ZAPI: initiator-group-name

(Required for class method, Not Applicable for instance method)

=item C<< "vserver" => $vserver >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: ZAPI will be invoked in the context of the vserver specified to this
option.

=item C<< "ostype" => $string >>

Applicable for CMode CLI/ZAPIi.

Maps to:

CM ZAPI: attribute
(The value provided to ostype maps to 'value' in CM ZAPI)

=item C<< "initiator" => $string >>

Applicable for CMode CLI/ZAPI

Maps to:

CM ZAPI: initiators

=item C<< "alua" => $string >>

Applicable for CMode CLIi/ZAPI.

Maps to:

CM ZAPI: attribute
(The value provided to alua maps to 'value' in CM ZAPI)

=item C<< "force" => $string >>

Applicable for CMode CLI/ZAPI.

=item C<< "portset" => $string >>

Applicable for CMode CLI/ZAPI

Maps to:

CM ZAPI: initiator-group-portset-name

=item C<< "extended_query" => $string >>

(Optional)

Applicable for CMode CLI.

=item C<< "value" => $string >>

Applicable for CMode ZAPI

=item C<< '*' >>

This method takes all other options supported by
"lun igroup modify" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub modify {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( ostype initiator alua force portset value delete-on-unmap
                    )
            ),
        },
        cli_cmd => 'lun igroup modify',
    );
    my $vserver;
    $opts{vserver_operated_on} = \$vserver;
    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_modify_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_modify_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::LunIgroup::modify',
    );

    $Log->exit() if $may_exit;
    return $ret;
} ## end sub modify

sub _modify_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@args,  api => 'lun_igroup_modify');
    $Log->exit() if $may_exit;
} ## end sub _modify_cmode_cli

sub _modify_cmode_zapi {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );

    my %set_opts;
    $pkg_or_obj->_hash_move(
        source => \%opts,
        target => \%set_opts,
        move   => [qw(ostype alua)]
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my $vserver           = delete $opts{vserver};
    my %cmode_args;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        map    => {
            "igroup"    => "initiator-group-name",
            "portset"   => "initiator-group-portset-name",
            "initiator" => "initiators",
        },
        copy   => [qw (force value)],
        target => \%cmode_args,
    );

    $pkg_or_obj->_handle_zapi_vserver_context(
        api_opts          => \%cmode_args,
        vserver           => $vserver,
        command_interface => $command_interface
    );

    while (my ($key, $value) = each %set_opts) {
        $apiset->igroup_set_attribute(
            %cmode_args,
            attribute => $key,
            value     => $value
        );
    }
    $Log->exit() if $may_exit;
} ## end sub _modify_cmode_zapi

=head2 delete

   NACL::C::LunIgroup->delete(
                                command_interface => $ci,
                                igroup => $igroup,
                                vserver => $vserver,
                                  ...);

                            or

    $LunIgroup_object->delete(...);

(Class or instance method)
This method is used to delete an initiator group.

Uses a CMode CLI/ZAPI, 7Mode CLI APISet.

Invokes "igroup-destroy" API for CMode ZAPI.

Invokes "igroup destroy" command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "igroup" => $igroup >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI, 7Mode CLI

Maps to:

CM ZAPI: initiator-group-name

7M CLI : initiator group name (Programmatic name: "initiator-group")

=item C<< "vserver" => $vserver >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: ZAPI will be invoked in the context of the vserver specified to this
option.

=item C<< "force" => $string >>

Applicable for CMode CLI/ZAPI,7Mode CLI

Maps to:

CM ZAPI: force

7M CLI : -f (Programmatic name: "force")


=item C<< 'extended_query' => $string >>

(Optional)

Applicable for CMode CLI.

=item C<< '*' >>

This method takes all other options supported by
"lun igroup delete" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=item C<NACL::C::Exceptions::LunIgroup::DoesNotExist>

This type of exception gets thrown if an attempt is made to delete a
non-existing igroup.

=item C<NACL::C::Exceptions::LunIgroup::MapExists>

This type of exception gets thrown if an attempt is made to delete a
an igroup which is mapped.

=back

=back

=cut

sub delete {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( force
                    )
            ),
        },
        cli_cmd => 'lun igroup delete',
    );
    my $vserver;
    $opts{vserver_operated_on} = \$vserver;
    try {
        $pkg_or_obj->call_on_apiset(
            %opts,
            choices => [
                {   method    => '_delete_cmode_cli',
                    interface => 'CLI',
                    set       => 'CMode',
                },
                {   method    => '_delete_cmode_zapi',
                    interface => 'ZAPI',
                    set       => 'CMode',
                },
                {   method    => '_delete_7mode_cli',
                    interface => 'CLI',
                    set       => '7Mode',
                },
            ],
            frontend => 'NACL::C::LunIgroup::delete'
        );
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = $_[0];
        NACL::C::Exceptions::LunIgroup::DoesNotExist
            ->detect_convert_and_throw(exception => $exception);
        NACL::C::Exceptions::LunIgroup::MapExists->detect_convert_and_throw(
            exception => $exception);
        $Log->exit() if $may_exit;
        $exception->throw();
    };
    $Log->exit() if $may_exit;
    return;
} ## end sub delete

sub _delete_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@args,  api  => 'lun_igroup_delete');
    $Log->exit() if $may_exit;
} ## end sub _delete_cmode_cli

sub _delete_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
	my $allowed_cmode_options = [
        qw( force
            )
    ];
    $pkg_or_obj->_base_backend_cmode_zapi(@args,
            api    => 'igroup_destroy',
            copy   => $allowed_cmode_options,
            map    => {"igroup" => "initiator-group-name"},
	);
    $Log->exit() if $may_exit;
} ## end sub _delete_cmode_zapi

sub _delete_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface     = delete $opts{command_interface};
    my $apiset                = delete $opts{apiset};
    my $allowed_7mode_options = [qw( force )];
    my %valid_args;
    $pkg->_hash_copy(
        source => \%opts,
        map    => {'igroup' => 'initiator-group',},
        copy   => $allowed_7mode_options,
        target => \%valid_args,
    );
    %opts = %valid_args;

    $apiset->igroup_destroy(%opts);
    $Log->exit() if $may_exit;
} ## end sub _delete_7mode_cli

=head2 disable_aix_support

    NACL::C::LunIgroup->disable_aix_support(
                                command_interface => $ci,
                                ...);

(Class method)
This method is used to disabled SAN AIX Support on the cluster.

Uses a CMode CLI/ZAPI APISet.

Invokes "igroup-disable-aix-support" API for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'method-timeout' => $time_in_seconds >>

(Optional)
The default NATE timeout is 60 seconds. If the command might take more
time to complete, this option can be used to specify a larger timeout value.
The value should be provided in seconds.

=item C<< 'extended_query' => $string >>

(Optional)

Applicable for CMode CLI.

=item C<< '*' >>

This method takes all other options supported by
"lun igroup disable-aix-support" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub disable_aix_support {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params              => \@args,
        ignore_primary_keys => 1,
        cli_cmd             => 'lun igroup disable-aix-support',
    );
    $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_disable_aix_support_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_disable_aix_support_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::LunIgroup::disable_aix_support',
    );

    $Log->exit() if $may_exit;
} ## end sub disable_aix_support

sub _disable_aix_support_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;

    $pkg->_base_backend_cmode_cli(
        @args,
        api => 'lun_igroup_disable_aix_support',
    );

    $Log->exit() if $may_exit;
} ## end sub _disable_aix_support_cmode_cli

sub _disable_aix_support_cmode_zapi {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = $pkg->_validate_backend_opts(@_);

    my $apiset = $opts{apiset};

    $apiset->igroup_disable_aix_support();
    $Log->exit() if $may_exit;
} ## end sub _disable_aix_support_cmode_zapi


__PACKAGE__->_load_event_class();

sub _primary_keys_validate_spec {
    return (
        'igroup'  => {type => SCALAR},
        'vserver' => {type => SCALAR | UNDEF, optional => 1},
    );
} ## end sub _primary_keys_validate_spec

1;
