#
# Copyright (c) 2014 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary LunMove Component Module
## @author dl-nacl-dev@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::C::LunMove

=head1 DESCRIPTION

C<NACL::C::LunMove> is a derived class of
L<NACL::C::Component::ONTAP|lib-NACL-C-Component-ONTAP-pm>. It represents access
to an ONTAP LunMove.  A related class is
L<NACL::CS::LunMove|lib-NACL-CS-LunMove-pm>, which represents
the past state of an ONTAP LunMove.

=head1 ATTRIBUTES

=head2 command_interface

See L<NACL::C::Component::command_interface|lib-NACL-C-Component-pm/command_interface>

=head2 destination_path

=head2 vserver

=cut

package NACL::C::LunMove;

use strict;
use warnings;

use NACL::CS::LunMove;
use NACL::ComponentUtils qw (_optional_scalars Dumper);
use base 'NACL::C::Component::ONTAP';
use Params::Validate qw(validate_with SCALAR ARRAYREF SCALARREF);
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use NACL::C::_Mixins::LunCopyRelocation qw(:all);
use NACL::C::Exceptions::LunMove::MaximumSimultaneousOperation qw(:try);
use NACL::APISet::Exceptions::ResponseException;

=head1 METHODS

=head2 new

  my $LunMove = NACL::C::LunMove->new(
                                   command_interface => $ci,
                                   'destination-path' => $destination_path,
                                   'vserver' => $vserver,
								  );

Return a new LunMove component object with the given attributes.

=cut

use Class::MethodMaker [
    scalar => 'destination_path',
    scalar => 'vserver',
    scalar => 'job_uuid',
    scalar => 'lun_index',
];

=head2 find

    my $LunMove_obj  = NACL::C::LunMove->find(command_interface => $ci, ...);
	my @LunMove_objs = NACL::C::LunMove->find(command_interface => $ci, ...); 

(Class method)
A constructor of sorts, which discovers which elements are present (by
interacting with the given command interface).  In an array context it
creates a component object for each present element and returns them.
In a scalar context it creates a component object for just one
element.

This method is generally not implemented in individual components. The base
class (L<NACL::C::Component|lib-NACL-C-Component-pm> provides an implementation.
See <L<NACL::C::Component::find|lib-NACL-C-Component-pm/find> for a more
detailed description along with explanation of all the options accepted.

Uses a CMode CLI APISet.

=head2 state

    my LunMove_state = $LunMove_obj->state();

(Instance method)
Fetch the current state of the element associated
with this component, and return a Component State object corresponding
to this component (calling this on NACL::C::LunMove would return 
a NACL::CS::LunMove object)

This method is generally not implemented in individual components. The base
class (L<NACL::C::Component|lib-NACL-C-Component-pm> provides an implementation.
See <L<NACL::C::Component::state|lib-NACL-C-Component-pm/state> for a more
detailed description along with explanation of all the options accepted.

Uses a CMode CLI APISet.

=cut

sub init {
    $Log->enter() if $may_enter;
    my ($self, @args) = @_;
    $self->SUPER::init(@args);
    $self->_vserver_context_init(@args);
    $Log->exit() if $may_exit;
}

=head2 start

    my $LunMove_obj	= NACL::C::LunMove->start(
                                command_interface => $ci,
	                            'destination-path' => $destination_path,
	                            'vserver' => $vserver,
                                ...);
   
(Class method)
This method is used to start relocating a LUN or a set of LUNs from one volume to another within a Vserver.
Uses CMode CLI/ZAPI .

Invokes 'lun-move-start' API for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'destination-path' => $destination_path >>

(Required)
Destination Path
Applicable for CMode CLI/ZAPI

Maps to:

CM ZAPI:

<paths>
<lun-path-pair>
<destination-path>$scalar</destination-path>
</lun-path-pair>
</paths>

=item C<< 'vserver' => $vserver >>

(Required)
Vserver Name
Applicable for CMode CLI

=item C<< 'max-throughput' => $string >>

Maximum Transfer Rate (per sec)
Applicable for CMode CLI/ZAPI

=item C<< 'promote-late' => $string >>

Promote Late
Applicable for CMode CLI/ZAPI

=item C<< 'source-path' => $string >>

Source Path
Applicable for CMode CLI/ZAPI

Maps to:

CM ZAPI:

<paths>
<lun-path-pair>
<source-path>$scalar</source-path>
</lun-path-pair>
</paths>

=item C<< 'pause-transfer' => $string >>

Pause Transfer
Applicable for CMode CLI/ZAPI

=item C<< 'source-paths' => [ $source-path1, $source-path2, ... ] >>

(Arrayref, even if only a single value is to be sent it should specified
as an arrayref containing that single value)
Applicable for CMode CLI/ZAPI

Maps to:

CM ZAPI:

<paths>
<lun-path-pair>
<source-path>$scalar</source-path>
</lun-path-pair>
</paths>

=item C<< 'destination-paths' => [ $destination-path1, $destination-path2, ... ] >>

(Arrayref, even if only a single value is to be sent it should specified
as an arrayref containing that single value)
Applicable for CMode CLI/ZAPI

Maps to;

CM ZAPI:
<paths>
<lun-path-pair>
<destination-path>$scalar</destination-path>
</lun-path-pair>
</paths>

=item C<< 'reference-path' => $string >>

Source Reference Path
Applicable for CMode CLI

=item C<< '*' >>

This method takes all other options supported by "lun relocation start" command in ONTAP

=item C<< other_options >>

This method takes other common options described in L<Common method options|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#Other_options_valid_on_component_methods>

=back

=back

=over

=item Exceptions

=over

=item C<NACL::C::Exceptions::LunMove::MaximumSimultaneousOperation>

This exception gets thrown, when we attempt for
Lun move operation and maximum simultaneous file operations reached.


=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub start {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            'source-paths'      => {type => ARRAYREF, optional => 1},
            'destination-paths' => {type => ARRAYREF, optional => 1},
            _optional_scalars(
                qw( max-throughput promote-late source-path pause-transfer
                    destination-path
                    reference-path
                    )
            ),
        },
        api => 'lun_move_start',
    );
    my ($job_uuid, $vserver);
    $opts{_job_uuid}           = \$job_uuid;
    $opts{vserver_operated_on} = \$vserver;

    try {
        $pkg->call_on_apiset(
            %opts,
            choices => [
                {   method    => '_start_cmode_cli',
                    interface => 'CLI',
                    set       => 'CMode',
                },
                {   method    => '_start_cmode_zapi',
                    interface => 'ZAPI',
                    set       => 'CMode',
                    check     => '_check_zapi_option',
                },
            ],
            frontend => 'NACL::C::LunMove::start',
        );
    } catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = $_[0];
        NACL::C::Exceptions::LunMove::MaximumSimultaneousOperation
            ->detect_convert_and_throw(exception => $exception);
        $Log->exit() if $may_exit;
        $exception->throw();
    };
    my (@objs, $index);
    $opts{'destination-paths'} ||= [$opts{'destination-path'}];
    foreach my $dest_path (@{$opts{'destination-paths'}}) {
        my $obj = $pkg->new(
            command_interface  => $opts{command_interface},
            vserver            => $opts{vserver},
            'destination-path' => $dest_path,
        );
        if (defined $job_uuid) {
            $obj->job_uuid($job_uuid);
            $obj->lun_index($index++);
        }
        push @objs, $obj;
    }
    $Log->exit() if $may_exit;
    return wantarray ? @objs : $objs[0];
}

sub _start_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, @opts) = @_;
    $pkg->_base_backend_cmode_cli(@opts, api => 'lun_move_start',);
    $Log->exit() if $may_exit;
}

sub _start_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg, %opts) = @_;
    my $copy = $pkg->_start_common_copy();
    my %additional_args;
    $additional_args{'paths'} = $pkg->_construct_lun_path_pairs(%opts);
    my $response;
    $pkg->SUPER::_base_backend_cmode_zapi(
        %opts,
        copy                => $copy,
        apiset_response_obj => \$response,
        api                 => 'lun-move-start',
        frontend_method     => 'start',
        additional_args     => \%additional_args,
    );
    ${$opts{_job_uuid}} = $response->get_parsed_output()->[0]{'job-uuid'};
    $Log->exit() if $may_exit;
}

=head2 modify

    NACL::C::LunMove->modify(
                                command_interface => $ci,
	                            'destination-path' => $destination_path,
	                            'vserver' => $vserver,
                                ...);
	                             
								 or

    $LunMove_obj->modify(...);

(Class or instance method)
This method is used to modify an ongoing LUN relocation operation.
Uses CMode CLI/ZAPI .

Invokes API 'lun-move-modify' for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'destination-path' => $destination_path >>

(Required for class method, Not Applicable for instance method)
Destination Path
Applicable for CMode CLI

=item C<< 'vserver' => $vserver >>

(Required for class method, Not Applicable for instance method)
Vserver Name
Applicable for CMode CLI

=item C<< 'max-throughput' => $string >>

Maximum Transfer Rate (per sec)
Applicable for CMode CLI/ZAPI

=item C<< 'job-uuid' => $string >>

LUN Relocation Job UUID
Applicable for CMode CLI/ZAPI

=item C<< 'lun-index' => $string >>

LUN Index
Applicable for CMode CLI/ZAPI

=item C<< '*' >>

This method takes all other options supported by "lun relocation modify" command in ONTAP

=item C<< other_options >>

This method takes other common options described in L<Common method options|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#Other_options_valid_on_component_methods>

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub modify {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( max-throughput job-uuid lun-index
                    )
            ),
        },
        api => 'lun_move_modify',
    );
    my $vserver;
    $opts{vserver_operated_on} = \$vserver;
    $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_modify_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_modify_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::LunMove::modify',
    );

    $Log->exit() if $may_exit;
}

sub _modify_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@opts,
        api => 'lun_move_modify',);
    $Log->exit() if $may_exit;
}

sub get_api_options {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    my %hash = $pkg_or_obj->SUPER::get_api_options(%opts);
    if (exists $hash{vserver} || exists $hash{'destination-path'}) {
        delete $hash{'job-uuid'};
        delete $hash{'lun-index'};
    }
    $Log->exit() if $may_exit;
    return %hash;
}

sub _base_backend_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts = validate_with(
        params      => \@args,
        spec        => {copy => {type => ARRAYREF, default => []}},
        allow_extra => 1
    );
    $pkg_or_obj->_get_uuid_lun_index(\%opts);
    $opts{copy} = [@{$opts{copy}}, (qw(job-uuid lun-index))];
    $pkg_or_obj->SUPER::_base_backend_cmode_zapi(%opts);
    $Log->exit() if $may_exit;
}

sub _modify_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    $pkg_or_obj->_modify_cmode_zapi_helper(%opts,
        api => 'lun-move-modify',);
    $Log->exit() if $may_exit;
}

=head2 pause

    NACL::C::LunMove->pause(
                                command_interface => $ci,
	                            'destination-path' => $destination_path,
	                            'vserver' => $vserver,
                                ...);
	                             
								 or

    $LunMove_obj->pause(...);

(Class or instance method)
This method is used to pause an ongoing LUN relocation operation.
Uses CMode CLI/ZAPI .

Invokes API 'lun-move-pause' for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'destination-path' => $destination_path >>

(Required for class method, Not Applicable for instance method)
Destination Path
Applicable for CMode CLI

=item C<< 'vserver' => $vserver >>

(Required for class method, Not Applicable for instance method)
Vserver Name
Applicable for CMode CLI

=item C<< 'job-uuid' => $string >>

LUN Relocation Job UUID
Applicable for CMode CLI/ZAPI

=item C<< 'lun-index' => $string >>

LUN Index
Applicable for CMode CLI/ZAPI

=item C<< '*' >>

This method takes all other options supported by "lun relocation pause" command in ONTAP

=item C<< other_options >>

This method takes other common options described in L<Common method options|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#Other_options_valid_on_component_methods>

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub pause {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( job-uuid lun-index
                    )
            ),
        },
        api => 'lun_move_pause',
    );
    my $vserver;
    $opts{vserver_operated_on} = \$vserver;
    $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_pause_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_pause_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::LunMove::pause',
    );

    $Log->exit() if $may_exit;
}

sub _pause_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@opts,
        api => 'lun_move_pause',);
    $Log->exit() if $may_exit;
}

sub _pause_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    $pkg_or_obj->_base_backend_cmode_zapi(%opts,
        api => 'lun-move-pause',);
    $Log->exit() if $may_exit;
}

=head2 resume

    NACL::C::LunMove->resume(
                                command_interface => $ci,
	                            'destination-path' => $destination_path,
	                            'vserver' => $vserver,
                                ...);
	                             
								 or

    $LunMove_obj->resume(...);

(Class or instance method)
This method is used to resume a paused LUN relocation operation.
Uses CMode CLI/ZAPI .

Invokes API 'lun-move-resume' for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'destination-path' => $destination_path >>

(Required for class method, Not Applicable for instance method)
Destination Path
Applicable for CMode CLI

=item C<< 'vserver' => $vserver >>

(Required for class method, Not Applicable for instance method)
Vserver Name
Applicable for CMode CLI

=item C<< 'job-uuid' => $string >>

LUN Relocation Job UUID
Applicable for CMode CLI/ZAPI

=item C<< 'lun-index' => $string >>

LUN Index
Applicable for CMode CLI/ZAPI

=item C<< '*' >>

This method takes all other options supported by "lun relocation resume" command in ONTAP

=item C<< other_options >>

This method takes other common options described in L<Common method options|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#Other_options_valid_on_component_methods>

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub resume {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( job-uuid lun-index
                    )
            ),
        },
        api => 'lun_move_resume',
    );
    my $vserver;
    $opts{vserver_operated_on} = \$vserver;
    $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_resume_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_resume_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::LunMove::resume',
    );

    $Log->exit() if $may_exit;
}

sub _resume_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@opts,
        api => 'lun_move_resume',);
    $Log->exit() if $may_exit;
}

sub _resume_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    $pkg_or_obj->_base_backend_cmode_zapi(%opts,
        api => 'lun-move-resume',);
    $Log->exit() if $may_exit;
}

=head2 cancel

    NACL::C::LunMove->cancel(
                                command_interface => $ci,
	                            'destination-path' => $destination_path,
	                            'vserver' => $vserver,
                                ...);
	                             
								 or

    $LunMove_obj->cancel(...);

(Class or instance method)
This method is used to cancel a LUN relocation operation before the new LUN has been created.
Uses CMode CLI/ZAPI .

Invokes API 'lun-move-cancel' for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'destination-path' => $destination_path >>

(Required for class method, Not Applicable for instance method)
Destination Path
Applicable for CMode CLI

=item C<< 'vserver' => $vserver >>

(Required for class method, Not Applicable for instance method)
Vserver Name
Applicable for CMode CLI

=item C<< 'job-uuid' => $string >>

LUN Relocation Job UUID
Applicable for CMode CLI/ZAPI

=item C<< 'lun-index' => $string >>

LUN Index
Applicable for CMode CLI/ZAPI

=item C<< '*' >>

This method takes all other options supported by "lun relocation cancel" command in ONTAP

=item C<< other_options >>

This method takes other common options described in L<Common method options|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#Other_options_valid_on_component_methods>

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub cancel {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( job-uuid lun-index
                    )
            ),
        },
        api => 'lun_move_cancel',
    );
    my $vserver;
    $opts{vserver_operated_on} = \$vserver;
    $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_cancel_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_cancel_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::LunMove::cancel',
    );
    $Log->exit() if $may_exit;
}

sub _cancel_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@opts,
        api => 'lun_move_cancel',);
    $Log->exit() if $may_exit;
}

sub _cancel_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    $pkg_or_obj->_base_backend_cmode_zapi(%opts,
        api => 'lun-move-cancel',);
    $Log->exit() if $may_exit;
}

=head2 show_by_job_info

    my $show_by_job_info_obj = NACL::C::LunMove->show_by_job_info(command_interface => $ci,
                          [ filter => { %filter } ,]
			  [ requested_fields => [ @req_fields ]);
						  
                    or
					
    my @show_by_job_info_objs = NACL::C::LunMove->show_by_job_info(command_interface => $ci,
                          [ filter => { %filter } ,]
			  [ requested_fields => [ @req_fields ]);

                    or
  
   my $show_by_job_info_obj = $LunMove_obj->show_by_job_info(); 
						  
(Class or Instance method)

This method is a CS-like fetch method which accepts all the parameters of 
L<fetch|lib-NACL-CS-ComponentState-pm/fetch> (requested_fields, filter, allow_empty) 
and has all of the other properties of fetch methods (callable in scalar or list 
context, throws a NoElementsFound exception if no matching elements are found).
The requested fields and filter takes L<ATTRIBUTES|lib-NACL-CS-LunMoveShowByJobInfo-pm/ATTRIBUTES>
of NACL::CS::LunMoveShowByJobInfo as input. 
This method returns objects of the type L<NACL::CS::LunMoveShowByJobInfo|lib-NACL-CS-LunMoveShowByJobInfo-pm>.

=cut

sub show_by_job_info {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts = $pkg_or_obj->_handle_obj_cs_like_calls(@args);
    require NACL::CS::LunMoveShowByJobInfo;
    my @state_objs = NACL::CS::LunMoveShowByJobInfo->fetch(%opts);
    $Log->exit() if $may_exit;
    return wantarray ? @state_objs : $state_objs[0];
}

=head2 recover_source

    NACL::C::LunMove->recover_source(
                                command_interface => $ci,
	                            'vserver' => $vserver,
                                ...);
   
(Class method)
This method is used to recover the source LUN of a LUN relocation operation..
Uses CMode CLI .

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'vserver' => $vserver >>

(Required)
Vserver Name
Applicable for CMode CLI

=item C<< 'recover-to-path' => $string >>

New Path for the LUN
Applicable for CMode CLI

=item C<< 'source-path' => $string >>

Path to the source LUN
Applicable for CMode CLI

=item C<< '*' >>

This method takes all other options supported by "lun relocation recover source" command in ONTAP

=item C<< other_options >>

This method takes other common options described in L<Common method options|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#Other_options_valid_on_component_methods>

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub recover_source {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( recover-to-path vserver source-path
                    )
            ),
        },
        ignore_primary_keys => 1,
        cli_cmd             => 'lun move recover-source',
    );
    $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_recover_source_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::LunMove::recover_source',
    );

    $Log->exit() if $may_exit;
}

sub _recover_source_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, @opts) = @_;
    $pkg->_base_backend_cmode_cli(@opts,
        api => 'lun_move_recover_source',);
    $Log->exit() if $may_exit;
}


__PACKAGE__->_load_event_class();

sub _primary_keys_validate_spec {
    return (
        'destination-path' => {type => SCALAR},
        'vserver'          => {type => SCALAR},
    );
}

1;
