# $Id$

#
# Copyright (c) 2012 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary MagneticTape Component Module
## @author dl-dpg-nb-automation; dl-nacl-dev@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::C::MagneticTape

=head1 DESCRIPTION

C<NACL::C::MagneticTape> is a derived class of
L<NACL::C::Component::ONTAP|lib-NACL-C-Component-ONTAP-pm>.
It represents access to magnetic tape on a filer.

=cut

package NACL::C::MagneticTape;

use strict;
use warnings;

use NACL::CS::MagneticTape;
use NACL::ComponentUtils qw (_optional_scalars);

use base 'NACL::C::Component::ONTAP';
use Params::Validate qw(validate validate_with SCALAR ARRAYREF);
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Data::Dumper;

=head1 METHODS

=head2 new

    my $MagneticTape = NACL::C::MagneticTape->new(
                                   command_interface=>$ci,
                                   'tape-name'=>$tape_name,
                              );

Return a new MagneticTape component object with the given attributes.

=cut

use Class::MethodMaker [
    scalar => 'tape_name',
];

=head2 eof

   NACL::C::MagneticTape->eof(
                   command_interface => $ci,
                   'tape-name'=>$tape_name,
                  ...);
                              or

    $MagneticTape_object->eof(...);

(Class or instance method)
This method is used to write 'count' end-of-filemarks
beginning at the current position on tape.

Uses a 7Mode/Nodescope CLI APISet.

Invokes "mt" command for 7Mode/Nodescope CLI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "tape-name" => $tape_name >>

(Required for class method, Not Applicable for instance method)

Applicable for 7Mode CLI.

Maps to:

7M CLI : tape name (Programmatic name: "tape-name")

=item C<< "count" => $string >>

(Required) Number of filemarks.

Applicable for 7Mode CLI.

Maps to:

7M CLI : Number of filemarks. (Programmatic name: "count")

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut


sub eof {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            count => { type => SCALAR },
        },
    );
    my $ret = $pkg_or_obj->call_on_apiset(
          %opts,
          choices => [

              {
                  method    => '_eof_7mode_cli',
                  interface => 'CLI',
                  set       => '7Mode|Nodescope',
              },
          ],
    );

    $Log->exit() if $may_exit;
    return $ret;
}

sub _eof_7mode_cli {
      $Log->enter() if $may_enter;
      my $pkg_or_obj = shift;
      my %opts       = validate_with(
          params      => \@_,
          spec        => $pkg_or_obj->_backend_validate_spec(),
          allow_extra => 1
      );

      delete $opts{command_interface};
      my $apiset = delete $opts{apiset};

      $apiset->mt(%opts, 'end-of-file' => 1, 'tape-device-switch2' => 1);
      $Log->exit() if $may_exit;
}

=head2 weof

   NACL::C::MagneticTape->weof(
                   command_interface => $ci,
                   'tape-name'=>$tape_name,
                  ...);
                              or

    $MagneticTape_object->weof(...);

(Class or instance method)
This method is used to write 'count' end-of-filemarks
beginning at the current position on tape.

Uses a 7Mode/Nodescope CLI APISet.

Invokes "mt" command for 7Mode/Nodescope CLI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "tape-name" => $tape_name >>

(Required for class method, Not Applicable for instance method)

Applicable for 7Mode CLI.

Maps to:

7M CLI : tape name (Programmatic name: "tape-name")

=item C<< "count" => $string >>

(Required) Number of filemarks.

Applicable for 7Mode CLI.

Maps to:

7M CLI : Number of filemarks. (Programmatic name: "count")

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut


sub weof {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            count => { type => SCALAR },
        },
    );
    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [

            {
                method    => '_weof_7mode_cli',
                interface => 'CLI',
                set       => '7Mode|Nodescope',
            },
        ],
    );

    $Log->exit() if $may_exit;
    return $ret;
}

sub _weof_7mode_cli {
      $Log->enter() if $may_enter;
      my $pkg_or_obj = shift;
      my %opts       = validate_with(
          params      => \@_,
          spec        => $pkg_or_obj->_backend_validate_spec(),
          allow_extra => 1
      );

      delete $opts{command_interface};
      my $apiset = delete $opts{apiset};

      $apiset->mt(%opts, 'write-end-of-file' => 1, 'tape-device-switch2' => 1);
      $Log->exit() if $may_exit;
}

=head2 fsr

   NACL::C::MagneticTape->fsr(
                   command_interface => $ci,
                   'tape-name'=>$tape_name,
                  ...);
                              or

    $MagneticTape_object->fsr(...);

(Class or instance method)
This method is used to forward space 'count' records

Uses a 7Mode/Nodescope CLI APISet.

Invokes "mt" command for 7Mode/Nodescope CLI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "tape-name" => $tape_name >>

(Required for class method, Not Applicable for instance method)

Applicable for 7Mode CLI.

Maps to:

7M CLI : tape name (Programmatic name: "tape-name")

=item C<< "count" => $string >>

(Required) Count of records.

Applicable for 7Mode CLI.

Maps to:

7M CLI : Number of filemarks. (Programmatic name: "count")

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut


sub fsr {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            count => { type => SCALAR },
        },
    );
    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [

            {
                method    => '_fsr_7mode_cli',
                interface => 'CLI',
                set       => '7Mode|Nodescope',
            },
        ],
    );

    $Log->exit() if $may_exit;
    return $ret;
}

sub _fsr_7mode_cli {
      $Log->enter() if $may_enter;
      my $pkg_or_obj = shift;
      my %opts       = validate_with(
          params      => \@_,
          spec        => $pkg_or_obj->_backend_validate_spec(),
          allow_extra => 1
      );

      delete $opts{command_interface};
      my $apiset = delete $opts{apiset};

      $apiset->mt(%opts, 'forward-over-records' => 1, 'tape-device-switch2' => 1);
      $Log->exit() if $may_exit;
}

=head2 rewind

   NACL::C::MagneticTape->rewind(
                   command_interface => $ci,
                   'tape-name'=>$tape_name,
                  ...);
                              or

    $MagneticTape_object->rewind(...);

(Class or instance method)
This method is used to rewind the tape.

Uses a 7Mode/Nodescope CLI APIset.

Invokes "mt" command for 7Mode/Nodescope CLI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "tape-name" => $tape_name >>

(Required for class method, Not Applicable for instance method)

Applicable for 7Mode CLI.

Maps to:

7M CLI : tape name (Programmatic name: "tape-name")

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut


sub rewind {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@_,
    );
    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {
                method    => '_rewind_7mode_cli',
                interface => 'CLI',
                set       => '7Mode|Nodescope',
            },
        ],
    );

    $Log->exit() if $may_exit;
    return $ret;
}

sub _rewind_7mode_cli {
      $Log->enter() if $may_enter;
      my $pkg_or_obj = shift;
      my %opts       = validate_with(
          params      => \@_,
          spec        => $pkg_or_obj->_backend_validate_spec(),
          allow_extra => 1
      );

      delete $opts{command_interface};
      my $apiset = delete $opts{apiset};

      $apiset->mt(%opts, 'rewind' => 1, 'tape-device-switch2' => 1);
      $Log->exit() if $may_exit;
}

=head2 fsf

   NACL::C::MagneticTape->fsf(
                   command_interface => $ci,
                   'tape-name'=>$tape_name,
                  ...);
                              or

    $MagneticTape_object->fsf(...);

(Class or instance method)
This method is used to forward space over count filemarks.
Positions the tape on the end-of-tape side of the filemark.

Uses a 7Mode/Nodescope CLI APISet.

Invokes "mt" command for 7Mode/Nodescope CLI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "tape-name" => $tape_name >>

(Required for class method, Not Applicable for instance method)

Applicable for 7Mode CLI.

Maps to:

7M CLI : tape name (Programmatic name: "tape-name")

=item C<< "count" => $string >>

(Required) Number of filemarks.

Applicable for 7Mode CLI.

Maps to:

7M CLI : Number of filemarks. (Programmatic name: "count")

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut


sub fsf {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => { count => { type => SCALAR }, },
    );
    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [

            {
                method    => '_fsf_7mode_cli',
                interface => 'CLI',
                set       => '7Mode|Nodescope',
            },
        ],
    );

    $Log->exit() if $may_exit;
    return $ret;
}

sub _fsf_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );

    delete $opts{command_interface};
    my $apiset = delete $opts{apiset};

    $apiset->mt(
        %opts,
        'forward-over-filemarks' => 1,
        'tape-device-switch2'    => 1
    );
    $Log->exit() if $may_exit;
}

=head2 bsf

   NACL::C::MagneticTape->bsf(
                   command_interface => $ci,
                   'tape-name'=>$tape_name,
                  ...);
                              or

    $MagneticTape_object->bsf(...);

(Class or instance method)
This method is used to backward space over count filemarks.
Positions the tape on the beginning-of-tape side of the filemark.

Uses a 7Mode/Nodescope CLI APISet.

Invokes "mt" command for 7Mode/Nodescope CLI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "tape-name" => $tape_name >>

(Required for class method, Not Applicable for instance method)

Applicable for 7Mode CLI.

Maps to:

7M CLI : tape name (Programmatic name: "tape-name")

=item C<< "count" => $string >>

(Required) Number of filemarks.

Applicable for 7Mode CLI.

Maps to:

7M CLI : Number of filemarks. (Programmatic name: "count")

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut


sub bsf {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => { count => { type => SCALAR }, },
    );
    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [

            {
                method    => '_bsf_7mode_cli',
                interface => 'CLI',
                set       => '7Mode|Nodescope',
            },
        ],
    );

    $Log->exit() if $may_exit;
    return $ret;
}

sub _bsf_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );

    delete $opts{command_interface};
    my $apiset = delete $opts{apiset};

    $apiset->mt(
        %opts,
        'backward-over-filemarks' => 1,
        'tape-device-switch2'     => 1
    );
    $Log->exit() if $may_exit;
}

=head2 offline

   NACL::C::MagneticTape->offline(
                   command_interface => $ci,
                   'tape-name'=>$tape_name,
                  ...);
                              or

    $MagneticTape_object->offline(...);

(Class or instance method)
This method is used to rewind the tape and unload tape media.

Uses a 7Mode/Nodescope CLI APISet.

Invokes "mt" command for 7Mode/Nodescope CLI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "tape-name" => $tape_name >>

(Required for class method, Not Applicable for instance method)

Applicable for 7Mode CLI.

Maps to:

7M CLI : tape name (Programmatic name: "tape-name")

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut


sub offline {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with( params => \@args, );
    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [

            {
                method    => '_offline_7mode_cli',
                interface => 'CLI',
                set       => '7Mode|Nodescope',
            },
        ],
    );

    $Log->exit() if $may_exit;
    return $ret;
}

sub _offline_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );

    delete $opts{command_interface};
    my $apiset = delete $opts{apiset};

    $apiset->mt( %opts, 'offline' => 1, 'tape-device-switch2' => 1 );
    $Log->exit() if $may_exit;
}

=head2 erase

   NACL::C::MagneticTape->erase(
                   command_interface => $ci,
                   'tape-name'=>$tape_name,
                  ...);
                              or

    $MagneticTape_object->erase(...);

(Class or instance method)
This method is used to erase the tape beginning at the current
tape position. When the erase completes, the tape is positioned
to beginning-of-tape.

Uses a 7Mode/Nodescope CLI APISet.

Invokes "mt" command for 7Mode/Nodescope CLI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "tape-name" => $tape_name >>

(Required for class method, Not Applicable for instance method)

Applicable for 7Mode CLI.

Maps to:

7M CLI : tape name (Programmatic name: "tape-name")

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut


sub erase {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with( params => \@args, );
    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [

            {
                method    => '_erase_7mode_cli',
                interface => 'CLI',
                set       => '7Mode|Nodescope',
            },
        ],
    );

    $Log->exit() if $may_exit;
    return $ret;
}

sub _erase_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );

    delete $opts{command_interface};
    my $apiset = delete $opts{apiset};

    $apiset->mt( %opts, 'erase' => 1, 'tape-device-switch2' => 1 );
    $Log->exit() if $may_exit;
}

=head2 bsr

   NACL::C::MagneticTape->bsr(
                   command_interface => $ci,
                   'tape-name'=>$tape_name,
                  ...);
                              or

    $MagneticTape_object->bsr(...);

(Class or instance method)
This method is used to backward space [count] records
from the current position on tape.

Uses a 7Mode/Nodescope CLI APISet.

Invokes "mt" command for 7Mode/Nodescope CLI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "tape-name" => $tape_name >>

(Required for class method, Not Applicable for instance method)

Applicable for 7Mode CLI.

Maps to:

7M CLI : tape name (Programmatic name: "tape-name")

=item C<< "count" => $string >>

(Required) Number of records.

Applicable for 7Mode CLI.

Maps to:

7M CLI : Number of filemarks. (Programmatic name: "count")

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut


sub bsr {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            count => { type => SCALAR },
        },
    );
    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {
                method    => '_bsr_7mode_cli',
                interface => 'CLI',
                set       => '7Mode|Nodescope',
            },
        ],
    );

    $Log->exit() if $may_exit;
    return $ret;
}

sub _bsr_7mode_cli {
      $Log->enter() if $may_enter;
      my $pkg_or_obj = shift;
      my %opts       = validate_with(
          params      => \@_,
          spec        => $pkg_or_obj->_backend_validate_spec(),
          allow_extra => 1
      );

      delete $opts{command_interface};
      my $apiset = delete $opts{apiset};

      $apiset->mt(%opts, 'backward-over-records' => 1, 'tape-device-switch2' => 1);
      $Log->exit() if $may_exit;
}

=head2 eom

   NACL::C::MagneticTape->eom(
                   command_interface => $ci,
                   'tape-name'=>$tape_name,
                  ...);
                              or

    $MagneticTape_object->eom(...);

(Class or instance method)
This method is used to position the tape head at the end of tape media.

Uses a 7Mode/Nodescope CLI APISet.

Invokes "mt" command for 7Mode/Nodescope CLI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "tape-name" => $tape_name >>

(Required for class method, Not Applicable for instance method)

Applicable for 7Mode CLI.

Maps to:

7M CLI : tape name (Programmatic name: "tape-name")

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut


sub eom {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with( params => \@args );
    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [

            {
                method    => '_eom_7mode_cli',
                interface => 'CLI',
                set       => '7Mode|Nodescope',
            },
        ],
    );

    $Log->exit() if $may_exit;
    return $ret;
}

sub _eom_7mode_cli {
      $Log->enter() if $may_enter;
      my $pkg_or_obj = shift;
      my %opts       = validate_with(
          params      => \@_,
          spec        => $pkg_or_obj->_backend_validate_spec(),
          allow_extra => 1
      );

      delete $opts{command_interface};
      my $apiset = delete $opts{apiset};

      $apiset->mt(%opts, 'end-of-data' => 1, 'tape-device-switch2' => 1);
      $Log->exit() if $may_exit;
}


__PACKAGE__->_load_event_class();

sub _primary_keys_validate_spec {
      return ( 'tape-name' => { type => SCALAR } );
}

1;
