
#
# Copyright (c) 2001-2013 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary NetworkInterfaceFailoverGroups Component Module
## @author dl-nacl-dev@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::C::NetworkInterfaceFailoverGroups

=head1 DESCRIPTION

C<NACL::C::NetworkInterfaceFailoverGroups> is a derived class of
L<NACL::C::Component::ONTAP|lib-NACL-C-Component-ONTAP-pm>. It represents access
to an ONTAP NetworkInterfaceFailoverGroups.  A related class is
L<NACL::CS::NetworkInterfaceFailoverGroups|lib-NACL-CS-NetworkInterfaceFailoverGroups-pm>, which represents
the past state of an ONTAP NetworkInterfaceFailoverGroups.

=head1 ATTRIBUTES

=head2 command_interface

See L<NACL::C::Component::command_interface|lib-NACL-C-Component-pm/command_interface>

=head2 failover_group

Applicable for IPSpace and non-IPSPace builds

=head2 vserver

Applicable only for IPSpace builds

=head2 port

Applicable only for non-IPSPace builds

=head2 node

Applicable only for non-IPSPace builds

=cut

package NACL::C::NetworkInterfaceFailoverGroups;

use strict;
use warnings;

use NACL::CS::NetworkInterfaceFailoverGroups;
use NACL::ComponentUtils qw (_optional_scalars);

use base 'NACL::C::Component::ONTAP';
use Params::Validate qw(validate validate_with SCALAR ARRAYREF OBJECT);
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Data::Dumper;
use NACL::C::Exceptions::NetworkInterfaceFailoverGroups::AlreadyExists();
use NACL::C::Exceptions::NetworkInterfaceFailoverGroups::DoesNotExist();
use
    NACL::C::Exceptions::NetworkInterfaceFailoverGroups::FailoverGroupIsInUseByLif(
    );
use NATE::BaseException qw(:try);
use NACL::CS::ComponentState::ZapiSkip qw(make_zapi_skip);
use NACL::CS::ComponentState::ZapiArray qw(make_zapi_array);

=head1 METHODS

=head2 new

  my $NetworkInterfaceFailoverGroups = NACL::C::NetworkInterfaceFailoverGroups->new(command_interface=>$ci,
                                   'failover-group' =>$failover_group,
                                   port =>$port,
                                   node =>$node,
								  );

Return a new NetworkInterfaceFailoverGroups component object with the given attributes.

=cut

use Class::MethodMaker [
    scalar => 'failover_group',
    scalar => 'vserver',
    scalar => 'port',
    scalar => 'node',
];

=head2 find

    my $NetworkInterfaceFailoverGroups  = NACL::C::NetworkInterfaceFailoverGroups->find(command_interface => $ci, ...);
	my @NetworkInterfaceFailoverGroupss = NACL::C::NetworkInterfaceFailoverGroups->find(command_interface => $ci, ...); 

(Class method)
A constructor of sorts, which discovers which elements are present (by
interacting with the given command interface).  In an array context it
creates a component object for each present element and returns them.
In a scalar context it creates a component object for just one
element.

This method is generally not implemented in individual components. The base
class (L<NACL::C::Component::ONTAP|lib-NACL-C-Component-ONTAP-pm> provides an implementation.
See <L<NACL::C::Component::find|lib-NACL-C-Component-pm/find> for a more
detailed description along with explanation of all the options accepted.

=head2 state

    my $NetworkInterfaceFailoverGroups_state = $NACL::C::NetworkInterfaceFailoverGroups->state();

(Instance method)
Fetch the current state of the element associated
with this component, and return a Component State object corresponding
to this component (calling this on NACL::C::NetworkInterfaceFailoverGroups would return 
a NACL::CS::NetworkInterfaceFailoverGroups object)

This method is generally not implemented in individual components. The base
class (L<NACL::C::Component::ONTAP|lib-NACL-C-Component-ONTAP-pm> provides an implementation.
See <L<NACL::C::Component::state|lib-NACL-C-Component-pm/state> for a more
detailed description along with explanation of all the options accepted.

=cut

=head2 rename

    ## IPSpace build, class method call
    NACL::C::NetworkInterfaceFailoverGroups->rename(
        command_interface => $ci,
        'failover-group'  => $failover_group,
        vserver           => $vserver,
        ...
    );

    ## Non-IPSpace build, class method call
    NACL::C::NetworkInterfaceFailoverGroups->rename(
        command_interface => $ci,
	'failover-group'  => $failover_group,
	port              => $port,
	node              => $node,
        ...
    );

    # Instance method call
    failover_groups_obj->rename(...);

(Class method)
This method is used to rename a network interface failover group.

Uses a CMode CLI/ZAPI APIset.

Invokes "net-failover-group-rename" API for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "failover-group" => $failover_group >>

(Required)

Applicable for CMode CLI/ZAPI.

=item C<< "port" => $port >>

(Required)

Applicable for CMode CLI.

=item C<< "vserver" => $vserver>>

(Required)

Applicable for CMode CLI/ZAPI, IPSpace build only.

=item C<< "node" => $node >>

(Required)

Applicable for CMode CLI.

=item C<< "new-name" => $string >>

=item C<< 'new-failover-group-name' => $string >>

New name
Applicable for CMode CLI/ZAPI

=item C<< '*' >>

This method takes all other options supported by
"network interface failover-groups rename" command in ONTAP.

Applicable for CMode CLI.

=item C<< "extended_query" => $string >>

(Optional)

Applicable for CMode CLI

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub rename {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( new-name failover-group

                    new-failover-group-name
                    )
            ),
        },
        ignore_primary_keys => 1,
        cli_cmd             => 'network interface failover-groups rename',
    );

    my $ret = $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => "_rename_cmode_cli",
                interface => "CLI",
                set       => "CMode",
            },
            {   method    => "_rename_cmode_zapi",
                interface => "ZAPI",
                set       => "CMode",
            },
        ],
        frontend => 'NACL::C::NetworkInterfaceFailoverGroups::rename',
    );

    $Log->exit() if $may_exit;
    return $ret;
} ## end sub rename

sub _rename_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_base_backend_cmode_cli(@args, api => 'network_interface_failover_groups_rename');
    $Log->exit() if $may_exit;
} ## end sub _rename_cmode_cli

sub _rename_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;

    $pkg_or_obj->_base_backend_cmode_zapi(
        %opts,
        api  => 'net-failover-group-rename',
        copy => [qw( failover-group vserver )],
        map  => {'new-name' => 'new-failover-group-name'},
    );
    $Log->exit() if $may_exit;
} ## end sub _rename_cmode_zapi

=head2 create

    # IPSpace build, class method call
    NACL::C::NetworkInterfaceFailoverGroups->create(
        command_interface => $ci,
        'failover-group'  => $failover_group,
        vserver           => $vserver,
        ...
    );

    # Non-IPSpace build, class method call
    my $NetworkInterfaceFailoverGroups_object	= NACL::C::NetworkInterfaceFailoverGroups->create(
                                command_interface => $ci,
	                             'failover-group' => $failover_group,
	                             port => $port,
	                             node => $node,
                                  ...);

(Class method)
This method is used to create a new failover group or add an entry to an existing group.

Uses a CMode CLI/ZAPI APIset.

Invokes "net-failover-group-create" API for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "failover-group" => $failover_group >>

(Required)

Applicable for CMode CLI/ZAPI.

=item C<< "port" => $port >>

(Required)

Applicable for CMode CLI.

=item C<< "vserver" => $vserver>>

(Required)

Applicable for CMode CLI/ZAPI, IPSpace build only.

=item C<< "node" => $node >>

(Required)

Applicable for CMode CLI.

=item C<< 'targets' => [ $target1, $target2, ... ] >>

Failover Targets
(Arrayref, even if only a single value is to be sent it should specified
as an arrayref containing that single value)
Applicable for CMode CLI/ZAPI

=item C<< '*' >>

This method takes all other options supported by
"network interface failover-groups create" command in ONTAP.

Applicable for CMode CLI.

=item C<< "extended_query" => $string >>

(Optional)

Applicable for CMode CLI

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.


=item C<NACL::C::Exceptions::NetworkInterfaceFailoverGroups::AlreadyExists>

This type of exception gets thrown if an attempt is made to create a
failover-group that already exists.

=back

=back

=cut

sub create {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        cli_cmd         => 'network interface failover-groups create',
        additional_spec => {"targets" => {type => ARRAYREF, optional => 1},},
    );
    try {
        $pkg->call_on_apiset(
            %opts,
            choices => [
                {   method    => "_create_cmode_cli",
                    interface => "CLI",
                    set       => "CMode",
                },
                {   method    => "_create_cmode_zapi",
                    interface => "ZAPI",
                    set       => "CMode",
                },
            ],
            frontend => 'NACL::C::NetworkInterfaceFailoverGroups::create',
        );
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = $_[0];
        NACL::C::Exceptions::NetworkInterfaceFailoverGroups::AlreadyExists
            ->detect_convert_and_throw(exception => $exception);
        $Log->exit() if $may_exit;
        $exception->throw();
    };
    my $obj = $pkg->new(
        command_interface => $opts{command_interface},
        failover_group    => $opts{'failover-group'},
        vserver           => $opts{'vserver'},
    );
    $Log->exit() if $may_exit;
    return $obj;
} ## end sub create

sub _create_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    my %opts = validate_with(
        params      => \@args,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );

    my $command_interface = $opts{command_interface};
    my $apiset            = $opts{apiset};
    my $vserver           = $opts{vserver};

    my %keys = (
        command_interface => $command_interface,
        failover_group    => $opts{'failover-group'},
    );

    if ($opts{command_interface}->version_manager->has_uichange(
            uichange        => 'ipspaces-ms4',
            throw_exception => 0
        )
        )
    {
        $keys{vserver} = $vserver;
    } else {
        $keys{port} = $opts{'port'};
        $keys{node} = $opts{'node'};
    }

    $pkg->_base_backend_cmode_cli(%opts,
        api => 'network_interface_failover_groups_create',);

    $Log->exit() if $may_exit;
    return $pkg->new(%keys);

} ## end sub _create_cmode_cli

sub _create_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_base_backend_cmode_zapi(@args,
            api       => 'net_failover_group_create',
            copy      => [qw(failover-group vserver targets net-qualified-port-name)],
            target_skips  => 1,
    );
    $Log->exit() if $may_exit;
} ## end sub _create_cmode_zapi

=head2 delete

    # IPSpace build, class method call
    NACL::C::NetworkInterfaceFailoverGroups->delete(
        command_interface => $ci,
        'failover-group'  => $failover_group,
        vserver           => $vserver,
        ...
    );

    # Non-IPSpace build, class method call
    NACL::C::NetworkInterfaceFailoverGroups->delete(
                                command_interface => $ci,
	                             failover_group => $failover_group,
	                             port => $port,
	                             node => $node,
                                  ...);
	                             
								 or
    # Instance method call
    $NetworkInterfaceFailoverGroups_object->delete(...);

(Class or instance method)
This method is used to remove a port from a failover group.

Uses a CMode CLI/ZAPI APIset.

Invokes "net-failover-group-destroy" API for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "failover-group" => $failover_group >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI.

=item C<< "vserver" => $vserver>>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI, IPSpace build only.

=item C<< "port" => $port >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI.

=item C<< "node" => $node >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI.

=item C<< '*' >>

This method takes all other options supported by
"network interface failover-groups delete" command in ONTAP.

Applicable for CMode CLI.

=item C<< "extended_query" => $string >>

(Optional)

Applicable for CMode CLI

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=item C<NACL::C::Exceptions::NetworkInterfaceFailoverGroups::DoesNotExist>

This type of exception gets thrown if an attempt is made to delete a
non-existing failover-group.

=item C<NACL::C::Exceptions::NetworkInterfaceFailoverGroups::FailoverGroupIsInUseByLif>

This type of exception gets thrown if an attempt is made to delete a
failover-group which is in use by a lif.

=back

=back

=cut

sub delete {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params => \@args,
        cli_cmd  => 'network interface failover-groups delete',
    );
    my $ret;
    try {
        $ret = $pkg_or_obj->call_on_apiset(
            %opts,
            choices => [
                {   method    => "_delete_cmode_cli",
                    interface => "CLI",
                    set       => "CMode",
                },
                {   method    => "_delete_cmode_zapi",
                    interface => "ZAPI",
                    set       => "CMode",
                },
            ],
            frontend => 'NACL::C::NetworkInterfaceFailoverGroups::delete',
        );
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = $_[0];
        NACL::C::Exceptions::NetworkInterfaceFailoverGroups::DoesNotExist
            ->detect_convert_and_throw(exception => $exception);
        NACL::C::Exceptions::NetworkInterfaceFailoverGroups::FailoverGroupIsInUseByLif
            ->detect_convert_and_throw(exception => $exception);
        $Log->exit() if $may_exit;
        $exception->throw();
    };
    $Log->exit() if $may_exit;
    return $ret;
} ## end sub delete

sub _delete_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@args, api => 'network_interface_failover_groups_delete');
    $Log->exit() if $may_exit;
} ## end sub _delete_cmode_cli

sub _delete_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;

    $pkg_or_obj->_base_backend_cmode_zapi(
        %opts,
        api  => 'net-failover-group-destroy',
        copy => [qw( failover-group vserver )],
    );
    $Log->exit() if $may_exit;
} ## end sub _delete_cmode_zapi

=head2 add_port

    NACL::C::NetworkInterfaceFailoverGroups->add_port(
                                command_interface => $ci,
	                            'vserver' => $vserver,
	                            'failover-group' => $failover_group,
                                ...);
	                             
								 or

    $NetworkInterfaceFailoverGroups_obj->add_port(...);

(Class or instance method)
This method is used to add network ports to a failover group.
Uses CMode CLI/ZAPI .

Invokes "net-failover-group-add-ports" API for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'vserver' => $vserver >>

(Required for class method, Not Applicable for instance method)
Vserver Name
Applicable for CMode CLI/ZAPI

=item C<< 'failover-group' => $failover_group >>

(Required for class method, Not Applicable for instance method)
Failover Group Name
Applicable for CMode CLI/ZAPI

=item C<< 'ports' => [ $port1, $port2, ... ] >>

(Arrayref, even if only a single value is to be sent it should specified
as an arrayref containing that single value)
Applicable for CMode CLI/ZAPI

=item C<< '*' >>

This method takes all other options supported by "network interface failover groups add port" command in ONTAP

=item C<< other_options >>

This method takes other common options described in L<Common method options|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#Other_options_valid_on_component_methods>

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub add_port {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {'ports' => {type => ARRAYREF, optional => 1},},
        api             => 'network_interface_failover_groups_add_port',
    );
    if ($pkg_or_obj->_ports_renamed(%opts)) {
        $pkg_or_obj->_redirect_call(
            %opts,
            _old_method => 'add_port',
            _new_method => 'add_targets'
        );
        return;
    }
    $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_add_port_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_add_port_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::NetworkInterfaceFailoverGroups::add_port',
    );

    $Log->exit() if $may_exit;
} ## end sub add_port

sub _ports_renamed {
    my ($pkg_or_obj, %opts) = @_;
    my $ports_renamed = $opts{command_interface}->version_manager()->has_uichange(
        uichange        => 'vs-failover-ports-renamed-to-targets',
        throw_exception => 0
    );
    return $ports_renamed;
}

sub _redirect_call {
    my ($pkg_or_obj, %opts) = @_;
    my $pkg = $pkg_or_obj->get_package_name();
    my $old_method = delete $opts{_old_method};
    my $new_method = delete $opts{_new_method};
    $Log->warn('Ports has been renamed to targets, so the current call of:\n'
        . "    $pkg->$old_method(ports => \\\@ports);\n"
        . "should be changed to:\n"
        . "    $pkg->$new_method(targets => \\\@ports);\n"
    );
    $opts{targets} = delete $opts{ports};
    $pkg_or_obj->$new_method(%opts);
}

sub _add_port_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@opts,
        api => 'network_interface_failover_groups_add_port',);
    $Log->exit() if $may_exit;
} ## end sub _add_port_cmode_cli

sub _add_port_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_base_backend_cmode_zapi(@args,
            api     => 'net_failover_group_add_ports',
            copy    => [qw(failover-group vserver ports net-qualified-port-name)],
            target_skips => 1,
    );
    $Log->exit() if $may_exit;
} ## end sub _add_port_cmode_zapi

=head2 delete_port

    NACL::C::NetworkInterfaceFailoverGroups->delete_port(
                                command_interface => $ci,
	                            'vserver' => $vserver,
	                            'failover-group' => $failover_group,
                                ...);
	                             
								 or

    $NetworkInterfaceFailoverGroups_obj->delete_port(...);

(Class or instance method)
This method is used to delete network ports from a failover group.
Uses CMode CLI/ZAPI .

Invokes "net-failover-group-delete-ports" API for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'vserver' => $vserver >>

(Required for class method, Not Applicable for instance method)
Vserver Name
Applicable for CMode CLI/ZAPI

=item C<< 'failover-group' => $failover_group >>

(Required for class method, Not Applicable for instance method)
Failover Group Name
Applicable for CMode CLI/ZAPI

=item C<< 'ports' => [ $port1, $port2, ... ] >>

(Arrayref, even if only a single value is to be sent it should specified
as an arrayref containing that single value)
Applicable for CMode CLI/ZAPI

=item C<< '*' >>

This method takes all other options supported by "network interface failover groups delete port" command in ONTAP

=item C<< other_options >>

This method takes other common options described in L<Common method options|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#Other_options_valid_on_component_methods>

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub delete_port {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {'ports' => {type => ARRAYREF, optional => 1},},
        api             => 'network_interface_failover_groups_delete_port',
    );
    if ($pkg_or_obj->_ports_renamed(%opts)) {
        $pkg_or_obj->_redirect_call(
            %opts,
            _old_method => 'delete_port',
            _new_method => 'remove_targets'
        );
        return;
    }
    $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_delete_port_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_delete_port_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::NetworkInterfaceFailoverGroups::delete_port',
    );

    $Log->exit() if $may_exit;
} ## end sub delete_port

sub _delete_port_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@opts,
        api => 'network_interface_failover_groups_delete_port',);
    $Log->exit() if $may_exit;
} ## end sub _delete_port_cmode_cli

sub _delete_port_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_base_backend_cmode_zapi(@args,
            api      => 'net_failover_group_delete_ports',
            copy     => [qw(failover-group vserver ports net-qualified-port-name)],
            target_skips => 1,
    );
    $Log->exit() if $may_exit;
} ## end sub _delete_port_cmode_zapi

=head2 modify

    NACL::C::NetworkInterfaceFailoverGroups->modify(
                                command_interface => $ci,
	                            'vserver' => $vserver,
	                            'failover-group' => $failover_group,
                                ...);
	                             
								 or

    $NetworkInterfaceFailoverGroups_obj->modify(...);

(Class or instance method)
This method is used to modify a failover group.
Uses CMode CLI/ZAPI .

Invokes "net-failover-group-modify" API for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'vserver' => $vserver >>

(Required for class method, Not Applicable for instance method)
Vserver Name
Applicable for CMode CLI/ZAPI

=item C<< 'failover-group' => $failover_group >>

(Required for class method, Not Applicable for instance method)
Failover Group Name
Applicable for CMode CLI/ZAPI

=item C<< 'targets' => [ $target1, $target2, ... ] >>

(Arrayref, even if only a single value is to be sent it should specified
as an arrayref containing that single value)
Applicable for CMode CLI/ZAPI

=item C<< '*' >>

This method takes all other options supported by "network interface failover groups modify" command in ONTAP

=item C<< other_options >>

This method takes other common options described in L<Common method options|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#Other_options_valid_on_component_methods>

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub modify {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {'targets' => {type => ARRAYREF, optional => 1},},
        cli_cmd         => 'network interface failover-groups modify',
    );
    $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_modify_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_modify_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::NetworkInterfaceFailoverGroups::modify',
    );

    $Log->exit() if $may_exit;
} ## end sub modify

sub _modify_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@opts,
        api => 'network_interface_failover_groups_modify',);
    $Log->exit() if $may_exit;
} ## end sub _modify_cmode_cli

sub _modify_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_base_backend_cmode_zapi(@args,
            api         => 'net_failover_group_modify',
            copy        => [qw(failover-group vserver targets net-qualified-port-name)],
            target_skips => 1,
    );
    $Log->exit() if $may_exit;
} ## end sub _create_cmode_zapi

=head2 add_targets

    NACL::C::NetworkInterfaceFailoverGroups->add_targets(
                                command_interface => $ci,
	                            'vserver' => $vserver,
	                            'failover-group' => $failover_group,
                                ...);
		 or

    $NetworkInterfaceFailoverGroups_obj->add_targets(...);

(Class or instance method)
This method is used to add failover targets to a failover group.
Uses CMode CLI/ZAPI .
Invokes  net-failover-group-add-targets  for CMode Zapi non-iter

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>


=item C<< 'vserver' => $vserver >>

(Required for class method, Not Applicable for instance method)
Vserver Name
Applicable for CMode CLI/ZAPI, non-iter.

=item C<< 'failover-group' => $failover_group >>

(Required for class method, Not Applicable for instance method)
Failover Group Name
Applicable for CMode CLI/ZAPI, non-iter.

=item C<< 'targets' => [ $target1, $target2, ... ] >>

(Arrayref, even if only a single value is to be sent it should specified
as an arrayref containing that single value)
Applicable for CMode CLI/ZAPI, non-iter.

=item C<< '*' >>

This method takes all other options supported by "network interface failover groups add targets" command in ONTAP

=item C<< other_options >>

This method takes other common options described in L<Common method options|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#Other_options_valid_on_component_methods>

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub add_targets {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {'targets' => {type => ARRAYREF, optional => 1},},
        api             => 'network_interface_failover_groups_add_targets',
    );
    $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_add_targets_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
                zapi_type => 'none',
            },
            {   method    => '_add_targets_cmode_zapi_non_iter',
                interface => 'ZAPI',
                set       => 'CMode',
                zapi_type => 'non-iter',
            },
        ],
        frontend => 'NACL::C::NetworkInterfaceFailoverGroups::add_targets',
    );

    $Log->exit() if $may_exit;
}

sub _add_targets_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@opts,
        api => 'network_interface_failover_groups_add_targets',);
    $Log->exit() if $may_exit;
}

sub _add_targets_cmode_zapi_non_iter {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    my $copy = [qw(targets vserver failover-group)];
    $pkg_or_obj->_base_backend_cmode_zapi(
        @opts,
        api  => 'net-failover-group-add-targets',
        copy => $copy,
    );
    $Log->exit() if $may_exit;
}

=head2 remove_targets

    NACL::C::NetworkInterfaceFailoverGroups->remove_targets(
                                command_interface => $ci,
	                            'vserver' => $vserver,
	                            'failover-group' => $failover_group,
                                ...);
			 or

    $NetworkInterfaceFailoverGroups_obj->remove_targets(...);

(Class or instance method)
This method is used to remove failover targets from a failover group.
Uses CMode CLI/ZAPI .
Invokes  net-failover-group-remove-targets  for CMode Zapi non-iter

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>


=item C<< 'vserver' => $vserver >>

(Required for class method, Not Applicable for instance method)
Vserver Name
Applicable for CMode CLI/ZAPI, non-iter.

=item C<< 'failover-group' => $failover_group >>

(Required for class method, Not Applicable for instance method)
Failover Group Name
Applicable for CMode CLI/ZAPI, non-iter.

=item C<< 'targets' => [ $target1, $target2, ... ] >>

(Arrayref, even if only a single value is to be sent it should specified
as an arrayref containing that single value)
Applicable for CMode CLI/ZAPI, non-iter.

=item C<< '*' >>

This method takes all other options supported by "network interface failover groups remove targets" command in ONTAP

=item C<< other_options >>

This method takes other common options described in L<Common method options|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#Other_options_valid_on_component_methods>

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub remove_targets {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {'targets' => {type => ARRAYREF, optional => 1},},
        api             => 'network_interface_failover_groups_remove_targets',
    );
    $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_remove_targets_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
                zapi_type => 'none',
            },
            {   method    => '_remove_targets_cmode_zapi_non_iter',
                interface => 'ZAPI',
                set       => 'CMode',
                zapi_type => 'non-iter',
            },
        ],
        frontend => 'NACL::C::NetworkInterfaceFailoverGroups::remove_targets',
    );

    $Log->exit() if $may_exit;
}

sub _remove_targets_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@opts,
        api => 'network_interface_failover_groups_remove_targets',);
    $Log->exit() if $may_exit;
}

sub _remove_targets_cmode_zapi_non_iter {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    my $copy = [qw(targets vserver failover-group)];
    $pkg_or_obj->_base_backend_cmode_zapi(
        @opts,
        api  => 'net-failover-group-remove-targets',
        copy => $copy,
    );
    $Log->exit() if $may_exit;
}


__PACKAGE__->_load_event_class();

sub _primary_keys_validate_spec {
    my ($pkg_or_obj, @args) = @_;
    my %opts = validate(
        @args,
        {   command_interface => {
                type     => OBJECT,
                isa      => 'NACL::C::CommandInterface',
                optional => 1
            }
        }
    );
    my $command_interface = $opts{command_interface};

    my %default_spec = (
        'port'           => {type => SCALAR, optional => 1},
        'node'           => {type => SCALAR, optional => 1},
        'vserver'        => {type => SCALAR, optional => 1},
        'failover-group' => {type => SCALAR},
    );

    if (defined $command_interface) {
        if ($opts{command_interface}->version_manager->has_uichange(
                uichange        => 'ipspaces-ms4',
                throw_exception => 0
            )
            )
        {
            return (
                'vserver'        => {type => SCALAR},
                'failover-group' => {type => SCALAR},
            );
        } else {
            return (
                'port'           => {type => SCALAR},
                'node'           => {type => SCALAR},
                'failover-group' => {type => SCALAR},
            );
        }
    } else {
        return %default_spec;
    }

} ## end sub _primary_keys_validate_spec

1;
