# $Id$

#
# Copyright (c) 2001-2012 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary NetworkPort Component Module
## @author dl-nacl-dev@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::C::NetworkPort

=head1 DESCRIPTION

C<NACL::C::NetworkPort> is a derived class of
L<NACL::C::Component::ONTAP|lib-NACL-C-Component-ONTAP-pm>. It represents access
to an ONTAP NetworkPort.  A related class is
L<NACL::CS::NetworkPort|lib-NACL-CS-NetworkPort-pm>, which represents
the past state of an ONTAP NetworkPort.

=head1 ATTRIBUTES

=head2 command_interface

See L<NACL::C::Component::command_interface|lib-NACL-C-Component-pm/command_interface>

=head2 node

=head2 port

=cut

package NACL::C::NetworkPort;

use strict;
use warnings;

use NACL::CS::NetworkPort;
use NACL::C::NetworkPortBroadcastDomain;
use NACL::ComponentUtils qw (_optional_scalars Dumper);
use NACL::GeneralUtils qw(_move_common_component_params_with_ci);

use base 'NACL::C::Component::ONTAP';
use Params::Validate qw(validate validate_with SCALAR ARRAYREF UNDEF);
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use NACL::C::Exceptions::NetworkPort::DoesNotExist;
use NACL::APISet::Exceptions::ResponseException qw(:try);
use NATE::Exceptions::Argument;
use NACL::C::_Mixins::NetworkPort qw(:all);

=head1 METHODS

=head2 new

  my $NetworkPort = NACL::C::NetworkPort->new(command_interface=>$ci,
                                   node =>$node,
                                   port =>$port,
                    );

Return a new NetworkPort component object with the given attributes.

=cut

use Class::MethodMaker [
    scalar => 'node',
    scalar => 'port',
];

=head2 find

    my $NetworkPort  = NACL::C::NetworkPort->find(command_interface => $ci, ...);
    my @NetworkPorts = NACL::C::NetworkPort->find(command_interface => $ci, ...);

(Class method)
A constructor of sorts, which discovers which elements are present (by
interacting with the given command interface).  In an array context it
creates a component object for each present element and returns them.
In a scalar context it creates a component object for just one
element.

This method is generally not implemented in individual components. The base
class (L<NACL::C::Component::ONTAP|lib-NACL-C-Component-ONTAP-pm> provides an implementation.
See <L<NACL::C::Component::find|lib-NACL-C-Component-pm/find> for a more
detailed description along with explanation of all the options accepted.

=head2 state

    my $NetworkPort_state = $NACL::C::NetworkPort->state();

(Instance method)
Fetch the current state of the element associated
with this component, and return a Component State object corresponding
to this component (calling this on NACL::C::NetworkPort would return
a NACL::CS::NetworkPort object)

This method is generally not implemented in individual components. The base
class (L<NACL::C::Component::ONTAP|lib-NACL-C-Component-ONTAP-pm> provides an implementation.
See <L<NACL::C::Component::state|lib-NACL-C-Component-pm/state> for a more
detailed description along with explanation of all the options accepted.

=cut

=head2 modify

   NACL::C::NetworkPort->modify(
                                command_interface => $ci,
                                node => $node,
                                port => $port,
                                  ...);

                        or

    $NetworkPort_object->modify(...);

(Class or instance method)
This method is used to modify network port attributes.

Uses a  CMode CLI/ZAPI, 7Mode CLI APISet.

(Only Options mtu, port, flowcontrol-admin, up-admin are allowed for 7Mode)

Invokes "net-port-modify-iter" API for CMode ZAPI.

Invokes "ifconfig" command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "node" => $node >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI

Maps to:

CM ZAPI :

 <net-port-modify-iter>
 <attributes>
 <net-port-info>
 <node>$value</node>
 </net-port-info>
 </attributes>
 </net-port-modify-iter>

=item C<< "port" => $port >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI, 7Mode CLI

Maps to:

CM ZAPI :

 <net-port-modify-iter>
 <attributes>
 <net-port-info>
 <port>$value</port>
 </net-port-info>
 </attributes>
 </net-port-modify-iter>

7M CLI  : interface name (Programmatic name: "interface")

=item C<< "aggr-node" => $string >>

Applicable for CMode CLI.

=item C<< "aggr-distr-func" => $string >>

Applicable for CMode CLI.

=item C<< "ifgrp-mode" => $string >>

Applicable for CMode CLI/ZAPI

Maps to:

CM ZAPI :

 <net-port-modify-iter>
 <attributes>
 <net-port-info>
 <ifgrp-mode>$value</ifgrp-mode>
 </net-port-info>
 </attributes>
 </net-port-modify-iter>

=item C<< "speed-admin" => $string >>

Applicable for CMode CLI/ZAPI

Maps to:

CM ZAPI :

 <net-port-modify-iter>
 <attributes>
 <net-port-info>
 <administrative-speed>$value</administrative-speed>
 </net-port-info>
 </attributes>
 </net-port-modify-iter>

=item C<< "mtu" => $string >>

Applicable for CMode CLI/ZAPI, 7Mode CLI

Maps to:

CM ZAPI :

 <net-port-modify-iter>
 <attributes>
 <net-port-info>
 <mtu>$value</mtu>
 </net-port-info>
 </attributes>
 </net-port-modify-iter>

7M CLI  : mtusize <valule> (Programmatic name: "mtusize")

=item C<< "ifgrp-node" => $string >>

Applicable for CMode CLI/ZAPI

Maps to:

CM ZAPI :

 <net-port-modify-iter>
 <attributes>
 <net-port-info>
 <ifgrp-node>$value</ifgrp-node>
 </net-port-info>
 </attributes>
 </net-port-modify-iter>

=item C<< "ifgrp-port" => $string >>

Applicable for CMode CLI/ZAPI

Maps to:

CM ZAPI :

 <net-port-modify-iter>
 <attributes>
 <net-port-info>
 <ifgrp-port>$value</ifgrp-port>
 </net-port-info>
 </attributes>
 </net-port-modify-iter>

=item C<< "flowcontrol-admin" => $string >>

Applicable for CMode CLI/ZAPI, 7Mode CLI

Maps to:

CM ZAPI : administrative-flowcontrol

 <net-port-modify-iter>
 <attributes>
 <net-port-info>
 <administrative-flowcontrol>$value</administrative-flowcontrol>
 </net-port-info>
 </attributes>
 </net-port-modify-iter>

7M CLI  : flowcontrol <value> (Programmatic name: "flowcontrol")

=item C<< "ifgrp-distr-func" => $string >>

Applicable for CMode CLI/ZAPI

Maps to:

CM ZAPI : ifgrp-distribution-function

 <net-port-modify-iter>
 <attributes>
 <net-port-info>
 <ifgrp-distribution-function>$value</ifgrp-distribution-function>
 </net-port-info>
 </attributes>
 </net-port-modify-iter>

=item C<< "autonegotiate-admin" => $string >>

Applicable for CMode CLI/ZAPI

Maps to:

CM ZAPI :

 <net-port-modify-iter>
 <attributes>
 <net-port-info>
 <is-administrative-auto-negotiate>$value</is-administrative-auto-negotiate>
 </net-port-info>
 </attributes>
 </net-port-modify-iter>

=item C<< "up-admin" => $string >>

Applicable for CMode CLI/ZAPI

Maps to:

CM ZAPI :

 <net-port-modify-iter>
 <attributes>
 <net-port-info>
 <is-administrative-up>$value</is-administrative-up>
 </net-port-info>
 </attributes>
 </net-port-modify-iter>

=item C<< "aggr-lacp" => $string >>

Applicable for CMode CLI.

=item C<< "duplex-admin" => $string >>

Applicable for CMode CLI/ZAPI

Maps to:

CM ZAPI :

 <net-port-modify-iter>
 <attributes>
 <net-port-info>
 <administrative-duplex>$value</administrative-duplex>
 </net-port-info>
 </attributes>
 </net-port-modify-iter>

=item C<< "type" => $string >>

Applicable for CMode CLI/ZAPI

Maps to:

CM ZAPI : port-type

 <net-port-modify-iter>
 <attributes>
 <net-port-info>
 <port-type>$value</port-type>
 </net-port-info>
 </attributes>
 </net-port-modify-iter>

=item C<< "aggr-port" => $string >>

Applicable for CMode CLI.

=item C<< "role" => $string >>

Applicable for CMode CLI/ZAPI

Maps to:

CM ZAPI :

 <net-port-modify-iter>
 <attributes>
 <net-port-info>
 <role>$value</role>
 </net-port-info>
 </attributes>
 </net-port-modify-iter>

=item C<< 'autorevert-delay' => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CMode ZAPI: autorevert-delay

=item C<< 'ipspace' => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: ipspace

=item C<< 'extended_query' => $string >>

(Optional)

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI :

Any options specified under extended_query will translate to

 <net-port-modify-iter>
 <query>
 <net-port-info>
 <ZAPI-mapping-for-the-CLI-option-described-above>$value</...>
 </net-port-info>
 </query>
 </net-port-modify-iter>

=item C<< '*' >>

This method takes all other options supported by
"network port modify" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub modify {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( aggr-node aggr-distr-func ifgrp-mode speed-admin mtu ifgrp-node ifgrp-port
                    flowcontrol-admin ifgrp-distr-func autonegotiate-admin up-admin aggr-lacp
                    duplex-admin type aggr-port role autorevert-delay ipspace broadcast-domain
                    )
            ),
        },
        api => 'network_port_modify',
    );

   # If the uichange exists, there are three cases:
   # 1. Only "mtu" provided: should run only broadcast-domain modify
   # 2. Options other than "mtu" provided: should run only port modify
   # 3. Options includes mtu + other options: need to run both.
   # Let the $run_port_modify denote whether we need the port->modify
   # to run: it will be set to 0 only if uichange present + only mtu provided.
    my $run_port_modify = 1;
    if ($pkg_or_obj->_has_broadcast_domain_live_change(%opts)) {
        my %opts_copy = %opts;

        my %common_opts_with_ci;
        $pkg_or_obj->_move_common_component_params_with_ci(
            source => \%opts_copy,
            target => \%common_opts_with_ci
        );

        my %primary_keys;
        $pkg_or_obj->_move_primary_keys(
            source => \%opts_copy,
            target => \%primary_keys
        );

        my $mtu = delete $opts_copy{mtu};

        my $port_cs = NACL::CS::NetworkPort->fetch(
            %common_opts_with_ci,
            filter           => \%primary_keys,
            requested_fields => [qw(ipspace broadcast-domain)],
        );

        if ($port_cs->broadcast_domain() ne '-') {

            $Log->comment("Using broadcast-domain to modify mtu");

            delete $opts{mtu};
            delete $opts_copy{extended_query};
            if (!keys %opts_copy) {
                $run_port_modify = 0;
            }
            try {
                NACL::C::NetworkPortBroadcastDomain->modify(
                    %common_opts_with_ci,
                    ipspace            => $port_cs->ipspace(),
                    'broadcast-domain' => $port_cs->broadcast_domain(),
                    mtu                => $mtu,
                );
            }
            catch NACL::Exceptions::NoElementsFound with {
                my $exception = $_[0];
                if ($run_port_modify) {

                # Don't throw the exception: instead let the Port->modify run.
                # That way the DoesNotExist exception gets thrown.
                } else {
                    $Log->exit() if $may_exit;
                    $exception->throw();
                }
            };
        }
    }

    if (!$run_port_modify) {
        $Log->exit() if $may_exit;
        return;
    }

    my $ret;
    try {
        $ret = $pkg_or_obj->call_on_apiset(
            %opts,
            choices => [
                {   method    => '_modify_cmode_cli',
                    interface => 'CLI',
                    set       => 'CMode',
                },
                {   method    => '_modify_cmode_zapi',
                    interface => 'ZAPI',
                    set       => 'CMode',
                },
                {   method    => '_modify_7mode_cli',
                    interface => 'CLI',
                    set       => '7Mode',
                },
            ],
            frontend => 'NACL::C::NetworkPort::modify'
        );
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = $_[0];
        NACL::C::Exceptions::NetworkPort::DoesNotExist
            ->detect_convert_and_throw(exception => $exception);
        $Log->exit() if $may_exit;
        $exception->throw();
    };

    $Log->exit() if $may_exit;
    return $ret;
} ## end sub modify

sub _modify_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;

    if ($pkg_or_obj->_has_ipspaces_ms4(%opts)) {
        $pkg_or_obj->_map_role_to_ipspace(\%opts);
    }

    $pkg_or_obj->_base_backend_cmode_cli(%opts, api => 'network_port_modify');

    $Log->exit() if $may_exit;
} ## end sub _modify_cmode_cli

sub _modify_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );

    delete $opts{command_interface};
    my $apiset = delete $opts{apiset};

    my %args;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        map    => {
            "mtu"               => "mtusize",
            "flowcontrol-admin" => "flowcontrol",
            "port"              => "interface"
        },
        target => \%args,
    );
    if ($opts{'up-admin'} =~ /true/i) {
        $args{'up'} = 1;
    } else {
        $args{'down'} = 1;
    }

    $apiset->ifconfig(%args);
    $Log->exit() if $may_exit;
} ## end sub _modify_7mode_cli

sub _modify_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;

    if ($pkg_or_obj->_has_ipspaces_ms4(%opts)) {
        $pkg_or_obj->_map_role_to_ipspace(\%opts);
    }

    my $common_copy = $pkg_or_obj->_common_zapi_copy();
    my $common_map  = $pkg_or_obj->_common_zapi_map();
    $pkg_or_obj->_iter_helper(
        %opts,
        api                  => 'net-port-modify-iter',
        show_cmd             => 'network port show',
        copy                 => $common_copy,
        map                  => $common_map,
        does_not_exist_error => "entry doesn't exist",
        does_not_exist_exception =>
            'NACL::C::Exceptions::NetworkPort::DoesNotExist'
    );
    $Log->exit() if $may_exit;
}
=head2 delete

    NACL::C::NetworkPort->delete(
                                command_interface => $ci,
                                    'port' => $port,
                                    'node' => $node,
                                ...);

                                                                 or

    $NetworkPort_obj->delete(...);

(Class or instance method)
This method is used to delete a network port.
Uses CMode CLI/ZAPI .
Invokes  net-port-delete  for CMode Zapi non-iter

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>


=item C<< 'port' => $port >>

(Required for class method, Not Applicable for instance method)
Port possible value(s) are, <netport>,<ifgrp>
Applicable for CMode CLI/ZAPI, non-iter.

=item C<< 'node' => $node >>

(Required for class method, Not Applicable for instance method)
Node possible value(s) are, <nodename>,local
Applicable for CMode CLI/ZAPI, non-iter.

=item C<< '*' >>

This method takes all other options supported by "network port delete" command in ONTAP

=item C<< other_options >>

This method takes other common options described in L<Common method options|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#Other_options_valid_on_component_methods>

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub delete {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params  => \@args,
        cli_cmd => 'network port delete',
    );
    $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_delete_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
                zapi_type => 'none',
            },
            {   method    => '_delete_cmode_zapi_non_iter',
                interface => 'ZAPI',
                set       => 'CMode',
                zapi_type => 'non-iter',
            },
        ],
        frontend => 'NACL::C::NetworkPort::delete',
    );

    $Log->exit() if $may_exit;
}

sub _delete_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@opts,
        api => 'network_port_delete',);
    $Log->exit() if $may_exit;
}

sub _delete_cmode_zapi_non_iter {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    my $copy = [qw(port node)];
    $pkg_or_obj->_base_backend_cmode_zapi(
        @opts,
        api  => 'net-port-delete',
        copy => $copy,
    );
    $Log->exit() if $may_exit;
}

sub _check_non_iter {
    $Log->enter() if $may_enter;
    my ($pkg, @opts) = @_;
    $pkg->_base_check_non_iter(
        @opts,
        _primary_keys => [qw(port node)],
        _vs_context   => 1
    );
    $Log->exit() if $may_exit;
}


=head2 getdhcp

    NACL::C::NetworkPort->getdhcp(
                                command_interface => $ci,
	                            'port' => $port,
                                ...);
   
(Class method)
This method is used to get DHCP address for the port.
Uses CMode CLI .

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'port' => $port >>

(Required)
Management port
Applicable for CMode CLI

=item C<< '*' >>

This method takes all other options supported by "network port getdhcp" command in ONTAP

=item C<< other_options >>

This method takes other common options described in L<Common method options|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#Other_options_valid_on_component_methods>

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub getdhcp {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( port
                    )
            ),
        },
        ignore_primary_keys => 1,
        api                 => 'network_port_getdhcp',
    );
    $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_getdhcp_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::NetworkPort::getdhcp',
    );

}

sub _getdhcp_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, @opts) = @_;
    $pkg->_base_backend_cmode_cli(@opts, api => 'network_port_getdhcp',);
    $Log->exit() if $may_exit;
}

sub _has_broadcast_domain_live_change {
    my ($pkg, %opts) = @_;
    if ((   $opts{command_interface}->has_uichange(
                uichange        => 'broadcast-domain-live-change',
                throw_exception => 0
            )
        )
        && (defined $opts{mtu})
        )
    {
        return 1;
    }
}

sub _has_ipspaces_ms4 {
    my ($pkg, %opts) = @_;
    if (defined($opts{role})
        && ($opts{command_interface}->has_uichange(
                uichange        => 'ipspaces-ms4',
                throw_exception => 0
            )
        )
        && (!defined $opts{ipspace})
        )
    {
        return 1;
    }
}


__PACKAGE__->_load_event_class();

sub _primary_keys_validate_spec {
    return (
        'node' => {type => SCALAR, optional => 1},
        'port' => {type => SCALAR},
    );
} ## end sub _primary_keys_validate_spec

1;
