# $Id$
# A component to represent a machine running Data ONTAP

#
# Copyright (c) 2001-2010 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Node Component Module
## @author dl-nacl-dev@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::C::Node

=head1 DESCRIPTION

This component is an alias of L<NACL::C::SystemNode|lib-NACL-C-SystemNode-pm>.
The implementation and documentation of provided functionality is present in
L<NACL::C::SystemNode|lib-NACL-C-SystemNode-pm>.

(If a component 'aliascomp' is said to be an alias of component 'comp', then the
same functionality that is present in component 'comp' is also provided by
component 'aliascomp'.)

=cut

package NACL::C::Node;

use 5.014;
use strict;
use warnings;

use Module::Load;

use NACL::CS::Node;
use base 'NACL::C::SystemNode';
use NATE::Log qw(log_global);
use NATE::Exceptions::Argument;
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();

sub _real_component { 'NACL::C::SystemNode' }

=head1 METHODS

=head2 find

This is a small override of the B<NACL::C::SystemNode> method, that allows for
one special argument. If that argument is not present, then B<find> falls
through to the superclass version with all arguments intact.

=over

=item Options

Only the following option is relevant at this level:

=over

=item hosttype

A string value. If passed, must be one of:

    ava
    ontap

If the value is C<ava>, then this method will use the B<find> method of the
B<NACL::C::AVANode> class to find nodes. If the value is C<ontap>, then this
call threads through to the B<NACL::C::SystemNode> version of B<find>.

=back

=item Exceptions

In addition to any exceptions that could be thrown by the B<find> methods of
B<NACL::C::SystemNode> or B<NACL::C::AVANode>, the following may occur at this
level:

=over

=item B<NATE::Exceptions::Argument>

If the C<hosttype> parameter is passed, but is not one of the recognized
strings, this exception will be thrown.

=back

=back

The return value from this method is the same for any other B<find> method,
and is based on whether it is called in a list-context or scalar context.

=cut

sub find {
    $Log->enter() if $may_enter;
    my ($class, %opts) = @_;
    my @nodes;

    if (my $hosttype = delete $opts{hosttype}) {
        if ($hosttype eq 'ava') {
            @nodes = $class->find_ava_nodes(%opts);
        } elsif ($hosttype eq 'ontap') {
            @nodes = $class->SUPER::find(%opts);
        } else {
            NATE::Exceptions::Argument->throw("Unknown hosttype '$hosttype'");
        }
    } else {
        @nodes = $class->SUPER::find(%opts);
    }

    $Log->exit() if $may_exit;
    return wantarray ? @nodes : $nodes[0];
}

=head2 find_ava_nodes

    @avanodes = NACL::C::Node->find_ava_nodes;

This is generally only used internally by the B<find> method, above.

When called, this method will load the B<NACL::C::AVANode> class if needed,
and then call B<find> in that class to find all AltaVault nodes based on the
NATE parameters given to the current test-run.

This method takes no arguments.

=cut

sub find_ava_nodes {
    $Log->enter() if $may_enter;
    my ($class, @opts) = @_;

    state $classmap = {
        C     => 'NACL::C::AVANode',
        STask => 'NACL::STask::AVANode',
    };

    # Get the AVANode class to use based on whether this was called as a C or
    # STask:
    my $avaclass = $classmap->{$class->is_c_or_t()};
    load $avaclass;

    # Get the nodes for the proper class.
    my @nodes = $avaclass->find(@opts);

    $Log->exit() if $may_exit;
    return wantarray ? @nodes : $nodes[0];
}

1;
