# $Id$

#
# Copyright (c) 2017 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary QosAdaptivePolicyGroup Component Module
## @author dl-nacl-dev@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::C::QosAdaptivePolicyGroup

=head1 DESCRIPTION

C<NACL::C::QosAdaptivePolicyGroup> is a derived class of
L<NACL::C::Component::ONTAP|lib-NACL-C-Component-ONTAP-pm>. It represents access
to an ONTAP QosAdaptivePolicyGroup.  A related class is
L<NACL::CS::QosAdaptivePolicyGroup|lib-NACL-CS-QosAdaptivePolicyGroup-pm>, which represents
the past state of an ONTAP QosAdaptivePolicyGroup.

=head1 ATTRIBUTES

=head2 command_interface

See L<NACL::C::Component::command_interface|lib-NACL-C-Component-pm/command_interface>

=head2 policy_group

=cut

package NACL::C::QosAdaptivePolicyGroup;

use strict;
use warnings;

use NACL::CS::QosAdaptivePolicyGroup;
use NACL::ComponentUtils qw (_optional_scalars Dumper);

use base 'NACL::C::Component::ONTAP';
use Params::Validate qw(:all);

use NATE::Log qw(log_global);
use NACL::CS::ComponentState::ZapiSkip qw(make_zapi_skip);
use NACL::CS::ComponentState::ZapiArray qw(make_zapi_array);
use feature 'state';

=head1 METHODS

=head2 new

  my $QosAdaptivePolicyGroup = NACL::C::QosAdaptivePolicyGroup->new(
                                   command_interface => $ci,
                                   'policy-group' => $policy_group,
								  );

Return a new QosAdaptivePolicyGroup component object with the given attributes.

=cut

use Class::MethodMaker [scalar => 'policy_group',];
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();


=head2 create

    my $QosAdaptivePolicyGroup_obj	= NACL::C::QosAdaptivePolicyGroup->create(
                                command_interface => $ci,
	                            'policy-group' => $policy_group,
                                ...);
   
(Class method)
This method is used to create an adaptive policy group.
Uses CMode CLI/ZAPI .
Invokes  qos-adaptive-policy-group-create  for CMode Zapi non-iter

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>


=item C<< 'policy-group' => $policy_group >>

(Required)
Name
Applicable for CMode CLI/ZAPI, non-iter.

=item C<< 'peak-iops-allocation' => $string >>

Peak IOPS Allocation possible value(s) are, used-space,allocated-space
Applicable for CMode CLI/ZAPI, non-iter.

=item C<< 'vserver' => $string >>

Vserver
Applicable for CMode CLI/ZAPI, non-iter.

=item C<< 'is-expected-iops-enforced' => $string >>

Are Expected IOPS Enforced? possible value(s) are, true,false
Applicable for CMode CLI/ZAPI, non-iter.

=item C<< 'expected-iops' => $string >>

Expected IOPS
Applicable for CMode CLI/ZAPI, non-iter.

=item C<< 'peak-iops' => $string >>

Peak IOPS
Applicable for CMode CLI/ZAPI, non-iter.

=item C<< 'absolute-min-iops' => $string >>

Absolute Minimum IOPS
Applicable for CMode CLI/ZAPI, non-iter.

=item C<< '*' >>

This method takes all other options supported by "qos adaptive policy group create" command in ONTAP

=item C<< other_options >>

This method takes other common options described in L<Common method options|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#Other_options_valid_on_component_methods>

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub create {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params => \@args,
        additional_spec => {
            _optional_scalars(
                qw( peak-iops-allocation vserver is-expected-iops-enforced expected-iops
                    peak-iops absolute-min-iops)
            ),
        },
        cli_cmd => 'qos adaptive-policy-group create',
    );
    $pkg->call_on_apiset_with_exceptions(
        %opts,
        choices => [
            {   method    => '_create_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
                zapi_type => 'none',
            },
            {   method    => '_create_cmode_zapi_non_iter',
                interface => 'ZAPI',
                set       => 'CMode',
                zapi_type => 'non-iter',
            },
        ],
        nacl_exceptions => [qw(AlreadyExists)],
        frontend        => 'NACL::C::QosAdaptivePolicyGroup::create',
    );

    my %obj_opts = (
        command_interface => $opts{command_interface},
        policy_group      => $opts{'policy-group'},
    );
    my $obj = $pkg->new(%obj_opts);
    $Log->exit() if $may_exit;
    return $obj;
}

sub _create_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, @opts) = @_;
    $pkg->_base_backend_cmode_cli(@opts,
        api => 'qos_adaptive_policy_group_create',);
    $Log->exit() if $may_exit;
}

sub _create_cmode_zapi_non_iter {
    $Log->enter() if $may_enter;
    my ($pkg, @opts) = @_;
    my $copy = [
        qw(policy-group peak-iops-allocation is-expected-iops-enforced vserver
            expected-iops peak-iops absolute-min-iops)
    ];
    $pkg->_base_backend_cmode_zapi(
        @opts,
        api  => 'qos-adaptive-policy-group-create',
        copy => $copy,
    );
    $Log->exit() if $may_exit;
}

=head2 delete

NACL::C::QosAdaptivePolicyGroup->delete(
    command_interface => $ci,
    'policy-group'    => $policy_group,
    ...
);

or

$QosAdaptivePolicyGroup_obj->delete(...);

(Class or instance method)
This method is used to delete an adaptive policy group.
Uses CMode CLI / ZAPI.
Invokes qos-adaptive-policy-group-delete -iter
for CMode Zapi iter Invokes qos-adaptive-policy-group-delete
for CMode Zapi non-iter

=over

=item Options

=over

=item C << command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method) See L
< NACL::C::Component | lib-NACL -C -Component-pm / command_interface >

=item C << apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>


=item C<< 'policy-group' => $policy_group >>

(Required for class method, Not Applicable for instance method)
Name
Applicable for CMode CLI/ZAPI iter, non-iter.

Mapping for iter-zapi: <qos-adaptive-policy-group-info policy-group="$value" />

=item C << 'force' => $string >>

Force Delete Workloads for the QoS adaptive policy group
Applicable for CMode CLI/ZAPI iter, non-iter.

=item C << '*' >>

This method takes all other options supported by "qos adaptive policy group delete" command in ONTAP

=item C << other_options >>

This method takes other common options described in L<Common method options|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#Other_options_valid_on_component_methods>

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub delete {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params => \@args,
        additional_spec => {
            _optional_scalars(
                qw( force )
            ),
        },
        cli_cmd => 'qos adaptive-policy-group delete',
    );
    $pkg_or_obj->call_on_apiset_with_exceptions(
        %opts,
        choices => [
            {   method    => '_delete_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
                zapi_type => 'none',
            },
            {   method    => '_delete_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
                zapi_type => 'iter',
            },
            {   method    => '_delete_cmode_zapi_non_iter',
                interface => 'ZAPI',
                set       => 'CMode',
                zapi_type => 'non-iter',
            },
        ],
        nacl_exceptions => [qw(DoesNotExist)],
        frontend        => 'NACL::C::QosAdaptivePolicyGroup::delete',
    );

    $Log->exit() if $may_exit;
}

sub _delete_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@opts,
        api => 'qos_adaptive_policy_group_delete',);
    $Log->exit() if $may_exit;
}

sub _delete_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    require NACL::C::Exceptions::QosAdaptivePolicyGroup::DoesNotExist;
    my $copy = [
        qw(policy-group peak-iops-allocation is-expected-iops-enforced vserver
            num-workloads expected-iops peak-iops uuid absolute-min-iops pgid)
    ];
    $pkg_or_obj->_iter_helper(
        @opts,
        api      => 'qos-adaptive-policy-group-delete-iter',
        show_cmd => 'qos adaptive-policy-group show',
        copy     => $copy,
        does_not_exist_exception =>
            'NACL::C::Exceptions::QosAdaptivePolicyGroup::DoesNotExist',
    );
    $Log->exit() if $may_exit;
}

sub _delete_cmode_zapi_non_iter {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    my $copy = [qw(policy-group force)];
    $pkg_or_obj->_base_backend_cmode_zapi(
        @opts,
        api  => 'qos-adaptive-policy-group-delete',
        copy => $copy,
    );
    $Log->exit() if $may_exit;
}

=head2 modify

    NACL::C::QosAdaptivePolicyGroup->modify(
                                command_interface => $ci,
	                            'policy-group' => $policy_group,
                                ...);
	                             
								 or

    $QosAdaptivePolicyGroup_obj->modify(...);

(Class or instance method)
This method is used to modify an adaptive policy group.
Uses CMode CLI/ZAPI .
Invokes  qos-adaptive-policy-group-modify-iter for CMode Zapi iter
Invokes  qos-adaptive-policy-group-modify  for CMode Zapi non-iter

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>


=item C<< 'policy-group' => $policy_group >>

(Required for class method, Not Applicable for instance method)
Name
Applicable for CMode CLI/ZAPI iter, non-iter.

Mapping for iter-zapi: <qos-adaptive-policy-group-info policy-group="$value" />

=item C<< 'peak-iops-allocation' => $string >>

Peak IOPS Allocation possible value(s) are, used-space,allocated-space
Applicable for CMode CLI/ZAPI iter, non-iter.

Mapping for iter-zapi: <qos-adaptive-policy-group-info peak-iops-allocation="$value" />

=item C<< 'is-expected-iops-enforced' => $string >>

Are Expected IOPS Enforced? possible value(s) are, true,false
Applicable for CMode CLI/ZAPI iter, non-iter.

Mapping for iter-zapi: <qos-adaptive-policy-group-info is-expected-iops-enforced="$value" />

=item C<< 'expected-iops' => $string >>

Expected IOPS
Applicable for CMode CLI/ZAPI iter, non-iter.

Mapping for iter-zapi: <qos-adaptive-policy-group-info expected-iops="$value" />

=item C<< 'peak-iops' => $string >>

Peak IOPS
Applicable for CMode CLI/ZAPI iter, non-iter.

Mapping for iter-zapi: <qos-adaptive-policy-group-info peak-iops="$value" />

=item C<< 'absolute-min-iops' => $string >>

Absolute Minimum IOPS
Applicable for CMode CLI/ZAPI iter, non-iter.

Mapping for iter-zapi: <qos-adaptive-policy-group-info absolute-min-iops="$value" />

=item C<< '*' >>

This method takes all other options supported by "qos adaptive policy group modify" command in ONTAP

=item C<< other_options >>

This method takes other common options described in L<Common method options|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#Other_options_valid_on_component_methods>

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub modify {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params => \@args,
        additional_spec => {
            _optional_scalars(
                qw( peak-iops-allocation  is-expected-iops-enforced expected-iops
                    peak-iops absolute-min-iops)
            ),
        },
        cli_cmd => 'qos adaptive-policy-group modify',
    );
    $pkg_or_obj->call_on_apiset_with_exceptions(
        %opts,
        choices => [
            {   method    => '_modify_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
                zapi_type => 'none',
            },
            {   method    => '_modify_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
                zapi_type => 'iter',
            },
            {   method    => '_modify_cmode_zapi_non_iter',
                interface => 'ZAPI',
                set       => 'CMode',
                zapi_type => 'non-iter',
            },
        ],
        frontend => 'NACL::C::QosAdaptivePolicyGroup::modify',
    );

    $Log->exit() if $may_exit;
}

sub _modify_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@opts,
        api => 'qos_adaptive_policy_group_modify',);
    $Log->exit() if $may_exit;
}

sub _modify_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    require NACL::C::Exceptions::QosAdaptivePolicyGroup::DoesNotExist;
    my $copy = [
        qw(policy-group peak-iops-allocation is-expected-iops-enforced vserver
            num-workloads expected-iops peak-iops uuid absolute-min-iops pgid)
    ];
    $pkg_or_obj->_iter_helper(
        @opts,
        api      => 'qos-adaptive-policy-group-modify-iter',
        show_cmd => 'qos adaptive-policy-group show',
        copy     => $copy,
        does_not_exist_exception =>
            'NACL::C::Exceptions::QosAdaptivePolicyGroup::DoesNotExist',
    );
    $Log->exit() if $may_exit;
}

sub _modify_cmode_zapi_non_iter {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    my $copy = [
        qw(policy-group peak-iops-allocation is-expected-iops-enforced expected-iops
            peak-iops absolute-min-iops)
    ];
    $pkg_or_obj->_base_backend_cmode_zapi(
        @opts,
        api  => 'qos-adaptive-policy-group-modify',
        copy => $copy,
    );
    $Log->exit() if $may_exit;
}

=head2 rename

    NACL::C::QosAdaptivePolicyGroup->rename(
                                command_interface => $ci,
	                            'policy-group' => $policy_group,
                                ...);
	                             
								 or

    $QosAdaptivePolicyGroup_obj->rename(...);

(Class or instance method)
This method is used to rename an adaptive policy group.
Uses CMode CLI/ZAPI .
Invokes  qos-adaptive-policy-group-rename  for CMode Zapi non-iter

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>


=item C<< 'policy-group' => $policy_group >>

(Required for class method, Not Applicable for instance method)
Name
Applicable for CMode CLI/ZAPI, non-iter.

=item C<< 'new-name' => $string >>

New adaptive policy group name
Applicable for CMode CLI/ZAPI, non-iter.

=item C<< '*' >>

This method takes all other options supported by "qos adaptive policy group rename" command in ONTAP

=item C<< other_options >>

This method takes other common options described in L<Common method options|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#Other_options_valid_on_component_methods>

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub rename {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params => \@args,
        additional_spec => {
            _optional_scalars(
                qw( new-name)
            ),
        },
        cli_cmd => 'qos adaptive-policy-group rename',
    );
    $pkg_or_obj->call_on_apiset_with_exceptions(
        %opts,
        choices => [
            {   method    => '_rename_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
                zapi_type => 'none',
            },
            {   method    => '_rename_cmode_zapi_non_iter',
                interface => 'ZAPI',
                set       => 'CMode',
                zapi_type => 'non-iter',
            },
        ],
        frontend => 'NACL::C::QosAdaptivePolicyGroup::rename',
    );

    $Log->exit() if $may_exit;
    $pkg_or_obj->policy_group($opts{'new-name'}) if ref($pkg_or_obj);
}

sub _rename_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@opts,
        api => 'qos_adaptive_policy_group_rename',);
    $Log->exit() if $may_exit;
}

sub _rename_cmode_zapi_non_iter {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    my $copy = [qw(policy-group new-name)];
    $pkg_or_obj->_base_backend_cmode_zapi(
        @opts,
        api  => 'qos-adaptive-policy-group-rename',
        copy => $copy,
    );
    $Log->exit() if $may_exit;
}
__PACKAGE__->_load_event_class();

sub _primary_keys_validate_spec {
    return ('policy-group' => {type => SCALAR},);
}

1;
