#
# Copyright (c) 2001-2012 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary QosPolicy Component Module
## @author dl-nacl-dev@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::C::QosPolicy

=head1 DESCRIPTION

C<NACL::C::QosPolicy> is a derived class of
L<NACL::C::Component::ONTAP|lib-NACL-C-Component-ONTAP-pm>. It represents access
to an ONTAP QosPolicy.  A related class is
L<NACL::CS::QosPolicy|lib-NACL-CS-QosPolicy-pm>, which represents
the past state of an ONTAP QosPolicy.

=head1 ATTRIBUTES

=head2 command_interface

See L<NACL::C::Component::command_interface|lib-NACL-C-Component-pm/command_interface>

=head2 policy

=cut

package NACL::C::QosPolicy;

use strict;
use warnings;

use NACL::CS::QosPolicy;
use NACL::ComponentUtils qw (_optional_scalars);

use base 'NACL::C::Component::ONTAP';
use Params::Validate qw(validate validate_with SCALAR ARRAYREF);
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Data::Dumper;

=head1 METHODS

=head2 new

  my $QosPolicy = NACL::C::QosPolicy->new(command_interface=>$ci,
                                   policy =>$policy,
                                  );

Return a new QosPolicy component object with the given attributes.

=cut

use Class::MethodMaker [
    scalar => 'policy',
    scalar => 'uuid',
];

=head2 find

    my $QosPolicy  = NACL::C::QosPolicy->find(command_interface => $ci, ...);
    my @QosPolicys = NACL::C::QosPolicy->find(command_interface => $ci, ...);

(Class method)
A constructor of sorts, which discovers which elements are present (by
interacting with the given command interface).  In an array context it
creates a component object for each present element and returns them.
In a scalar context it creates a component object for just one
element.

This method is generally not implemented in individual components. The base
class (L<NACL::C::Component::ONTAP|lib-NACL-C-Component-ONTAP-pm> provides an implementation.
See <L<NACL::C::Component::find|lib-NACL-C-Component-pm/find> for a more
detailed description along with explanation of all the options accepted.

=head2 state

    my $QosPolicy_state = $NACL::C::QosPolicy->state();

(Instance method)
Fetch the current state of the element associated
with this component, and return a Component State object corresponding
to this component (calling this on NACL::C::QosPolicy would return
a NACL::CS::QosPolicy object)

This method is generally not implemented in individual components. The base
class (L<NACL::C::Component::ONTAP|lib-NACL-C-Component-ONTAP-pm> provides an implementation.
See <L<NACL::C::Component::state|lib-NACL-C-Component-pm/state> for a more
detailed description along with explanation of all the options accepted.

=cut

=head2 modify

   NACL::C::QosPolicy->modify(
                                command_interface => $ci,
                                 policy => $policy,
                                  ...);

                                 or

    $QosPolicy_object->modify(...);

(Class or instance method)
This method is used to modify policy.

Uses a CMode CLI/ZAPI APISet.

Invokes "qos-policy-modify" API for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "policy" => $policy >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI.

=item C<< "target-tput-burst" => $string >>

Applicable for CMode CLI.

=item C<< "max-tput" => $string >>

Applicable for CMode CLI.

=item C<< "max-latency" => $string >>

Applicable for CMode CLI.

=item C<< "target-tput" => $string >>

Applicable for CMode CLI.

=item C<< "max-tput-burst" => $string >>

Applicable for CMode CLI.

=item C<< "read-ahead" => $string >>

Applicable for CMode CLI/ZAPI

Maps to:

CM ZAPI : read-ahead-policy-name

=item C<< "cache" => $string >>

Applicable for CMode CLI/ZAPI

Maps to:

CM ZAPI : cache-policy-name

=item C<< 'cache-policy' => $string >>

Applicable for CMode CLI.

=item C<< 'read-ahead-policy' => $string >>

Applicable for CMode CLI.

=item C<< "extended_query" => $string >>

(Optional)

Applicable for CMode CLI.

=item C<< '*' >>

This method takes all other options supported by
"qos policy modify" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub modify {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( target-tput-burst max-tput max-latency target-tput max-tput-burst
                    read-ahead cache policy uuid
                    cache-policy read-ahead-policy
                    )
            ),
        },
        api => 'qos_policy_modify',
    );
    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_modify_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_modify_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::QosPolicy::modify',
    );

    $Log->exit() if $may_exit;
    return $ret;
} ## end sub modify

sub _modify_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts = validate_with(
        params      => \@args,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );

    delete $opts{command_interface};
    my $apiset = delete $opts{apiset};

    my %cmode_args = $pkg_or_obj->get_api_options(
        apiset  => $apiset,
        api     => 'qos_policy_modify',
        options => \%opts,
    );

    $apiset->qos_policy_modify(%cmode_args);
    $Log->exit() if $may_exit;
} ## end sub _modify_cmode_cli

sub _modify_cmode_zapi {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my $cs;
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );

    my $ci     = delete $opts{command_interface};
    my $apiset = delete $opts{apiset};
    my $policy = delete $opts{'policy'};

    my %cmode_args;

    $pkg_or_obj->_hash_copy(
        source => \%opts,
        map    => {
            "read-ahead" => "read-ahead-policy-name",
            "cache"      => "cache-policy-name"
        },
        target => \%cmode_args,
    );
    %opts = %cmode_args;

    $opts{'policy-uuid'} ||= $pkg_or_obj->_get_uuid(
        command_interface => $ci,
        policy            => $policy
    );

    $apiset->qos_policy_modify(%opts);
    $Log->exit() if $may_exit;
} ## end sub _modify_cmode_zapi

=head2 create

    my $QosPolicy_object    = NACL::C::QosPolicy->create(
                                command_interface => $ci,
                                 policy => $policy,
                                  ...);

(Class method)
This method is used to create policy.

Uses a CMode CLI/ZAPI APISet.

Invokes "qos-policy-create" API for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "policy" => $policy >>

(Required)

Applicable for CMode CLI/ZAPI

Maps to:

CM ZAPI : policy-name

=item C<< "target-tput-burst" => $string >>

Applicable for CMode CLI.

=item C<< "max-tput" => $string >>

Applicable for CMode CLI.

=item C<< "max-latency" => $string >>

Applicable for CMode CLI.

=item C<< "target-tput" => $string >>

Applicable for CMode CLI.

=item C<< "max-tput-burst" => $string >>

Applicable for CMode CLI.

=item C<< "read-ahead" => $string >>

Applicable for CMode CLI.

=item C<< "cache" => $string >>

Applicable for CMode CLI.

=item C<< 'cache-policy' => $string >>

Applicable for CMode CLI.

=item C<< 'read-ahead-policy' => $string >>

Applicable for CMode CLI.

=item C<< "extended_query" => $string >>

(Optional)

Applicable for CMode CLI.

=item C<< '*' >>

This method takes all other options supported by
"qos policy create" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub create {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( target-tput-burst max-tput max-latency target-tput max-tput-burst
                    read-ahead cache uuid
                    cache-policy read-ahead-policy
                    )
            ),
        },
        api => 'qos_policy_create',
    );
    my $ret = $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_create_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_create_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::QosPolicy::create',
    );

    $Log->exit() if $may_exit;
    return $ret;
} ## end sub create

sub _create_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    my %opts = validate_with(
        params      => \@args,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );

    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};

    my %cmode_args = $pkg->get_api_options(
        apiset  => $apiset,
        api     => 'qos_policy_create',
        options => \%opts,
    );

    $apiset->qos_policy_create(%cmode_args);
    $Log->exit() if $may_exit;
    return $pkg->new(
        command_interface => $command_interface,
        policy            => $cmode_args{'policy'},
    );
} ## end sub _create_cmode_cli

sub _create_cmode_zapi {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );

    my $ci     = delete $opts{command_interface};
    my $apiset = delete $opts{apiset};
    my %cmode_args;
    $pkg->_hash_copy(
        source => \%opts,
        map    => {"policy" => "policy-name",},
        target => \%cmode_args,
    );
    %opts = %cmode_args;

    my $response = $apiset->qos_policy_create(%opts);

    $opts{'policy-uuid'} ||= $pkg->_get_uuid(
        command_interface => $ci,
        policy            => $opts{'policy-name'}
    );
    $Log->exit() if $may_exit;

    return $pkg->new(
        command_interface => $ci,
        policy            => $opts{'policy-name'},
        uuid              => $opts{'policy-uuid'},
    );
} ## end sub _create_cmode_zapi

=head2 rename

   NACL::C::QosPolicy->rename(
                                command_interface => $ci,
                                 policy => $policy,
                                  ...);

                                 or

    $QosPolicy_object->rename(...);

(Class or instance method)
This method is used to rename policy.

Uses a CMode CLI APISet.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "policy" => $policy >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI.

=item C<< "newname" => $string >>

Applicable for CMode CLI.

=item C<< "extended_query" => $string >>

(Optional)

Applicable for CMode CLI.

=item C<< '*' >>

This method takes all other options supported by
"qos policy rename" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub rename {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( newname
                    )
            ),
        },
        api => 'qos_policy_rename',
    );
    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => "_rename_cmode_cli",
                interface => "CLI",
                set       => "CMode",
            },
        ],
        frontend => 'NACL::C::QosPolicy::rename',
    );

    $Log->exit() if $may_exit;
    return $ret;
} ## end sub rename

sub _rename_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts = validate_with(
        params      => \@args,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );

    delete $opts{command_interface};
    my $apiset = delete $opts{apiset};

    my %cmode_args = $pkg_or_obj->get_api_options(
        apiset  => $apiset,
        api     => 'qos_policy_rename',
        options => \%opts,
    );

    $apiset->qos_policy_rename(%cmode_args);
    $pkg_or_obj->policy($cmode_args{'newname'}) if ref($pkg_or_obj);
    $Log->exit() if $may_exit;
} ## end sub _rename_cmode_cli

=head2 delete

   NACL::C::QosPolicy->delete(
                                command_interface => $ci,
                                 policy => $policy,
                                  ...);

                                 or

    $QosPolicy_object->delete(...);

(Class or instance method)
This method is used to delete policy.

Uses a CMode CLI/ZAPI APISet.

Invokes "qos-policy-destroy" API for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "policy" => $policy >>

Applicable for CMode CLI.

=item C<< "uuid" => $uuid >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI.

=item C<< "extended_query" => $string >>

(Optional)

Applicable for CMode CLI.

=item C<< '*' >>

This method takes all other options supported by
"qos policy delete" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub delete {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {_optional_scalars(qw( uuid )),},
        api             => 'qos_policy_delete',
    );
    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_delete_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_delete_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::QosPolicy::delete',
    );

    $Log->exit() if $may_exit;
    return $ret;
} ## end sub delete

sub _delete_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts = validate_with(
        params      => \@args,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );

    delete $opts{command_interface};
    my $apiset = delete $opts{apiset};
    my $uuid   = delete $opts{uuid};

    my %cmode_args = $pkg_or_obj->get_api_options(
        apiset  => $apiset,
        api     => 'qos_policy_delete',
        options => \%opts,
    );

    $apiset->qos_policy_delete(%cmode_args);
    $Log->exit() if $may_exit;
} ## end sub _delete_cmode_cli

sub _delete_cmode_zapi {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );

    my $ci     = delete $opts{command_interface};
    my $apiset = delete $opts{apiset};
    my $policy = delete $opts{'policy'};

    my %cmode_args;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%cmode_args,
    );
    %opts = %cmode_args;

    $opts{'policy-uuid'} ||= $pkg_or_obj->_get_uuid(
        command_interface => $ci,
        policy            => $policy
    );

    $apiset->qos_policy_destroy(%opts);
    $Log->exit() if $may_exit;
} ## end sub _delete_cmode_zapi

sub _get_uuid {

    # Extract parameters
    my $pkg_or_obj = shift;
    my %opts       = @_;
    my $ci         = $opts{command_interface};
    my $policy     = $opts{policy};

    my $cs = NACL::CS::QosPolicy->fetch(
        command_interface => $ci,
        filter            => {policy => $policy},
        requested_fields  => [qw(uuid)],
        apiset_should     => {interface => 'ZAPI'}
    );
    my $uuid = $cs->uuid();

    # If it is an object and it does not have "uuid" set in it,
    # then set it, so that the next time we do not need to
    # re-query the filer
    if (ref $pkg_or_obj && !$pkg_or_obj->uuid_isset()) {
        $pkg_or_obj->uuid($uuid);
    }

    return $uuid;
} ## end sub _get_uuid


__PACKAGE__->_load_event_class();

sub _primary_keys_validate_spec {
    return (
        'policy' => {type => SCALAR},
        'uuid'   => {type => SCALAR, optional => 1}
    );
} ## end sub _primary_keys_validate_spec

1;
