
#
# Copyright (c) 2001-2012 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Workload Component Module
## @author dl-nacl=dev@netapp.com, dl-perf-qa-automation@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::C::QosWorkload

=head1 DESCRIPTION

C<NACL::C::QosWorkload> is a derived class of
L<NACL::C::Component::ONTAP|lib-NACL-C-Component-ONTAP-pm>. It represents access
to an ONTAP QosWorkload. A related class is
L<NACL::CS::QosWorkload|lib-NACL-CS-QosWorkload-pm>, which represents
the past state of an ONTAP QosWorkload.

=head1 ATTRIBUTES

=head2 command_interface

See L<NACL::C::Component::command_interface|lib-NACL-C-Component-pm/command_interface>

=head2 workload

=cut

package NACL::C::QosWorkload;

use strict;
use warnings;

use NACL::C::QosWorkloadStream;
use NACL::CS::QosWorkload;
use NACL::ComponentUtils qw(_optional_scalars);

use base 'NACL::C::Component::ONTAP';
use Params::Validate qw(validate validate_with SCALAR);
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Data::Dumper;
use NACL::C::Exceptions::QosWorkload::DoesNotExist;
use NATE::BaseException qw(:try);

=head1 METHODS

=head2 new

  my $QosWorkload = NACL::C::QosWorkload->new(command_interface=>$ci,
                                   workload =>$workload,
                                                                  );

Return a new QosWorkload component object with the given attributes.

=cut

use Class::MethodMaker [scalar => 'workload',];

# The "qos workload reset" command has been obsoleted by burt434645.
# Thus NACL::C::QosWorkload no longer supports the "reset" method.

=head2 modify

   NACL::C::QosWorkload->modify(
                                command_interface => $ci,
                                     workload => $workload,
                                  ...);

        or

    $QosWorkload_object->modify(...);

(Class or instance method)
This method is used to modify workload.

Uses a CMode CLI/ZAPI APISet.
  Invokes 'qos-workload-modify-iter' API for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface => $ci >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< workload => $workload >>

Name of workload.

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI

Maps to:

CM ZAPI:
        <attributes>
          <qos-workload-info>
                <workload-name>$workload_name</workload-name>
          </qos-workload-info>
        <attributes>

=item C<< "match-priority" => $string >>

Applicable only for CMode CLI/ZAPI

Maps to:

CM ZAPI:
        <attributes>
          <qos-workload-info>
                <match-priority>$match_priority</match-priority>
          </qos-workload-info>
        </attributes>

=item C<< "class" => $string >>

  Applicable only for CMode CLI/ZAPI

Maps to:

CM ZAPI:
        <attributes>
          <qos-workload-info>
                <workload-class>$read_ahead</workload-class>
          </qos-workload-info>
        </attributes>

=item C<< "policy" => $string >>

  Applicable only for CMode CLI

=item C<< 'policy-priority' => $string >>

  Applicable only for CMode CLI

=item C<< 'ext-cache' => $string >>

  Applicable only for CMode CLI/ZAPI

Maps to:

CM ZAPI:
        <attributes>
          <qos-workload-info>
                <ext-cache>$read_ahead</ext-cache>
          </qos-workload-info>
        </attributes>

=item C<< 'read-ahead' => $string >>

Applicable for CMode CLI/ZAPI

Maps to:

CM ZAPI:
        <attributes>
          <qos-workload-info>
                <read-ahead>$read_ahead</read-ahead>
          </qos-workload-info>
        </attributes>

=item C<< 'category' => $string >>

Applicable for CMode CLI/ZAPI

Maps to:

CM ZAPI:
        <attributes>
          <qos-workload-info>
                <category>$category</category>
          </qos-workload-info>
        </attributes>

=item C<< 'extended_query' => $string >>

(Optional)

Applicable for CMode CLI/ZAPI

=item C<< 'file' => $string >>

Applicable for CMode ZAPI

=item C<< 'lun' => $string >>

Applicable for CMode ZAPI

=item C<< 'policy-group' => $string >>

Applicable for CMode ZAPI

=item C<< 'qtree' => $string >>

Applicable for CMode ZAPI

=item C<< 'extended_query' => $string >>

(Optional)
  Applicable only for CMode CLI

=item C<< '*' >>

This method takes all other options supported by
"qos workload modify" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=item C<NACL::C::Exceptions::QosWorkload::DoesNotExist>

This exception is thrown when we attempt to modify a workload
when it doesn't exist.

=back

=back

=cut

sub modify {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {

            # Note: ZAPI calls will fail if passed extended syntax
            # CLI options. We'll let the APISet throw an exception
            # in that case.
            _optional_scalars qw(class match-priority policy streams uuid wid
                policy-priority
                ext-cache read-ahead
                ),
        },
        cli_cmd => 'qos workload modify',
    );

    my $ret;
    try {
        $ret = $pkg_or_obj->call_on_apiset(
            %opts,
            choices => [
                {   method    => '_modify_cmode_cli',
                    interface => 'CLI',
                    set       => 'CMode',
                },
                {   method    => '_modify_cmode_zapi',
                    interface => 'ZAPI',
                    set       => 'CMode',
                },
            ],
            frontend => 'NACL::C::QosWorkload::modify',
        );
    }
    catch NACL::C::Exceptions::Component with {
        my $exception = $_[0];
        $exception->throw();
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = $_[0];
        NACL::C::Exceptions::QosWorkload::DoesNotExist
            ->detect_convert_and_throw(exception => $exception);
        $Log->exit() if $may_exit;
        $exception->throw();
    };

    $Log->exit() if $may_exit;
    return $ret;
} ## end sub modify

sub _modify_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,%opts) = @_;

    $pkg_or_obj->_base_backend_cmode_cli(
        %opts,
        api => 'qos_workload_modify',
    );

    $Log->exit() if $may_exit;
} ## end sub _modify_cmode_cli

sub _modify_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts              = $pkg_or_obj->_validate_backend_opts(@args);
    my $command_interface = $opts{command_interface};

    my $policy = delete $opts{policy};
    if ($policy) {
        $opts{'policy-uuid'} = NACL::CS::QosPolicy->fetch(
            command_interface => $command_interface,
            apiset_should     => {interface => 'ZAPI'},
            filter            => {policy => $policy},
        )->uuid();
    } ## end if ($policy)
    $pkg_or_obj->_iter_helper(
        %opts,
        api                  => 'qos-workload-modify-iter',
        show_cmd             => 'qos workload show',
        copy                 => $pkg_or_obj->_sn_copy(),
        map                  => $pkg_or_obj->_sn_map(),
        does_not_exist_error => "entry doesn't exist",
        does_not_exist_exception =>
                'NACL::C::Exceptions::QosWorkload::DoesNotExist'
    );
    $Log->exit() if $may_exit;
} ## end sub _modify_cmode_zapi

=head2 delete

  NACL::C::QosWorkload->delete(
      command_interface => $ci,
      workload          => $workload,
      ...);

                            or

  $QosWorkload_object->delete(...);

(Class or instance method)
This method is used to delete workload.

Uses a CMode CLI/ZAPI APISet.

CMode ZAPI:
    Invokes 'qos-workload-delete-iter' API for SN build.
    Invokes 'qos-workload-destroy-iter' API for RR build.

=over

=item Options

=over

=item C<< command_interface => $ci >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< workload => $workload >>

(Required for class method, Not Applicable for instance method)
  Applicable for CMode CLI/ZAPI

Maps to:

CM ZAPI:
        <query>
          <qos-workload-info>
                <workload-name>$workload_name</workload-name>
          </qos-workload-info>
        </query>

=item C<< 'match-priority' => $string >>

  Applicable only for CMode CLI/ZAPI

Maps to:

CM ZAPI:
        <query>
          <qos-workload-info>
                <match-priority>$match_priority</match-priority>
          </qos-workload-info>
        </query>

=item C<< 'policy' => $string >>

  Applicable only for CMode CLI

=item C<< 'class' => $string >>

  Applicable only for CMode CLI/ZAPI

Maps to:

CM ZAPI:
        <query>
          <qos-workload-info>
                <workload-class>$workload_class</workload-class>
          </qos-workload-info>
        </query>

=item C<< 'streams' => $string >>

  Applicable only for CMode CLI

=item C<< 'uuid' => $string >>

  Applicable only for CMode CLI/ZAPI

Maps to:

CM ZAPI:
        <query>
          <qos-workload-info>
                <workload-uuid>$workload_uuid</workload-uuid>
          </qos-workload-info>
        </query>

=item C<< 'wid' => $string >>

 Applicable only for CMode CLI/ZAPI

Maps to:

CM ZAPI:
        <query>
          <qos-workload-info>
                <wid>$wid</wid>
          </qos-workload-info>
        </query>

=item C<< 'extended_query' => $string >>

(Optional)
  Applicable only for CMode CLI

=item C<< '*' >>

This method takes all other options supported by
"qos workload delete" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=item C<NACL::C::Exceptions::QosWorkload::DoesNotExist>

This exception is thrown when we attempt to delete a workload
when it doesn't exist.

=back

=back

=cut

sub delete {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {

            # Note: ZAPI calls will fail if passed extended syntax
            # CLI options. We'll let the APISet throw an exception
            # in that case.
            _optional_scalars
                qw(class match-priority policy streams uuid wid),
        },
        cli_cmd => 'qos workload delete',
    );

    my $ret;
    try {
        $ret = $pkg_or_obj->call_on_apiset(
            %opts,
            choices => [
                {   method    => '_delete_cmode_cli',
                    interface => 'CLI',
                    set       => 'CMode',
                },
                {   method    => '_delete_cmode_zapi',
                    interface => 'ZAPI',
                    set       => 'CMode',
                },
            ],
        );
    }
    catch NACL::C::Exceptions::Component with {
        my $exception = $_[0];
        $exception->throw();
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = $_[0];
        NACL::C::Exceptions::QosWorkload::DoesNotExist
            ->detect_convert_and_throw(exception => $exception);
        $Log->exit() if $may_exit;
        $exception->throw();
    };
    $Log->exit() if $may_exit;
    return $ret;
} ## end sub delete

sub _delete_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,%opts) = @_;

    $pkg_or_obj->_base_backend_cmode_cli(
        %opts,
        api => 'qos_workload_delete',
    );

    $Log->exit() if $may_exit;
} ## end sub _delete_cmode_cli

sub _delete_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts              = $pkg_or_obj->_validate_backend_opts(@args);
    my $command_interface = $opts{command_interface};
    my $policy            = delete $opts{policy};
    if ($policy) {
        $opts{'policy-uuid'} = NACL::CS::QosPolicy->fetch(
            command_interface => $command_interface,
            apiset_should     => {interface => 'ZAPI'},
            filter            => {policy => $policy},
        )->uuid();
    } ## end if ($policy)
    $pkg_or_obj->_iter_helper(
        %opts,
        api                  => 'qos-workload-delete-iter',
        show_cmd             => 'qos workload show',
        copy                 => $pkg_or_obj->_sn_copy(),
        map                  => $pkg_or_obj->_sn_map(),
        does_not_exist_error => "entry doesn't exist",
        does_not_exist_exception =>
            'NACL::C::Exceptions::QosWorkload::DoesNotExist'
    );
    $Log->exit() if $may_exit;
} ## end sub _delete_cmode_zapi

=head2 rename

  NACL::C::QosWorkload->rename(command_interface => $ci,
                               workload          => $workload_name,
                               newname           => $new_workload_name);

                            or

  $workload_object->rename(newname => $new_workload_name);

(Class or instance method)
This method is used to rename workload.

Uses a CMode CLI/ZAPI APISet.
  Invokes 'qos-workload-modify' API for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< workload => $workload >>

(Required for class method, Not Applicable for instance method)

=item C<< newname => $newname >>

(Required)
  Applicable only for CMode CLI

=item C<< 'vserver' => $string >>

  Applicable only for CMode CLI

=item C<< 'workload' => $string >>

  Applicable only for CMode CLI

=item C<< 'extended_query' => $string >>

(Optional)
  Applicable only for CMode CLI

=item C<< '*' >>

This method takes all other options supported by
"qos workload rename" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub rename {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {_optional_scalars qw(newname vserver),},
        api             => 'qos_workload_rename',
    );

    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_rename_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_rename_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::QosWorkload::rename',
    );
    $Log->exit() if $may_exit;
    return $ret;
} ## end sub rename

sub _rename_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,%opts) = @_;

    $pkg_or_obj->_base_backend_cmode_cli(
        %opts,
        api => 'qos_workload_rename',
    );

    $pkg_or_obj->workload($opts{'newname'}) if ref($pkg_or_obj);
    $Log->exit() if $may_exit;
} ## end sub _rename_cmode_cli

sub _rename_cmode_zapi {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1,
    );

    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my $oldname           = delete $opts{workload};
    my $newname           = delete $opts{newname};

    $opts{'workload-name'} = $newname;
    $opts{'workload-uuid'} = $pkg_or_obj->_fetch_and_cache_uuid(
        command_interface => $command_interface,
        workload          => $oldname,
    );

    $apiset->qos_workload_modify(%opts);
    $pkg_or_obj->workload($opts{'workload-name'}) if ref($pkg_or_obj);
    $Log->exit() if $may_exit;
} ## end sub _rename_cmode_zapi

=head2 create

  my $workload  = NACL::C::QosWorkload->create(
                      command_interface => $ci,
                      workload          => $workload, ...);

(Class method)
This method is used to create workload.

Uses a CMode CLI/ZAPI APISet.
  Invokes 'qos-workload-create' API for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface => $ci >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< workload => $workload >>

(Required)
  Applicable for CMode CLI/ZAPI
  Maps to:
    CM ZAPI: workload-name

=item C<< vserver => $string >>

  Applicable for CMode CLI/ZAPI
  Maps to:
    CM ZAPI: vserver

=item C<< volume => $string >>

  Applicable for CMode CLI/ZAPI
  Maps to:
    CM ZAPI: volume

=item C<< lun => $string >>

  Applicable for CMode CLI/ZAPI
  Maps to:
    CM ZAPI: lun

=item C<< qtree => $string >>

  Applicable for CMode CLI/ZAPI
  Maps to:
    CM ZAPI: qtree

=item C<< 'protocol' => $string >>

  Applicable only for CMode CLI

=item C<< 'ip' => $string >>

  Applicable only for CMode CLI

=item C<< 'policy-priority' => $string >>

  Applicable only for CMode CLI

=item C<< 'match-priority' => $string >>

  Applicable for CMode CLI/ZAPI
  Maps to:
    CM ZAPI: match-priority

=item C<< 'min-req-size' => $string >>

  Applicable only for CMode CLI

=item C<< 'wwn' => $string >>

  Applicable only for CMode CLI

=item C<< 'igroup' => $string >>

  Applicable only for CMode CLI

=item C<< 'policy' => $string >>

  Applicable for CMode CLI/ZAPI
  Maps to:
    CM ZAPI: policy-name

=item C<< 'max-req-size' => $string >>

  Applicable only for CMode CLI

=item C<< 'request-type' => $string >>

  Applicable only for CMode CLI

=item C<< 'file' => $string >>

  Applicable only for CMode CLI

=item C<< 'uuid' => $string >>

  Applicable for CMode CLI/ZAPI
  Maps to:
    CM ZAPI: workload-uuid

=item C<< 'class' => $string >>

  Applicable for CMode CLI/ZAPI
  Maps to:
    CM ZAPI: workload-class

=item C<< 'policy_uuid' => $string >>

  Applicable for CMode CLI/ZAPI
  Maps to:
    CM ZAPI: policy-uuid

=item C<< 'extended_query' => $string >>

(Optional)
  Applicable only for CMode CLI

=item C<< '*' >>

This method takes all other options supported by
"qos workload create" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub create {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars qw(vserver volume lun qtree
                protocol ip policy-priority match-priority min-req-size wwn igroup policy
                max-req-size request-type
                file uuid
                ),
        },
        api => 'qos_workload_create',
    );

    my $ret = $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_create_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_create_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::QosWorkload::create',
    );
    $Log->exit() if $may_exit;
    return $ret;
} ## end sub create

sub _create_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg,%opts) = @_;

    $pkg->_base_backend_cmode_cli(
        %opts,
        api => 'qos_workload_create',
    );

    $Log->exit() if $may_exit;
    return $pkg->new(
            workload          => $opts{workload},
            command_interface =>$opts{command_interface},
            );
} ## end sub _create_cmode_cli

sub _create_cmode_zapi {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1,
    );

    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my $workload          = $opts{workload};

    # Now deal with the stream arguments.

    my %stream_opts;
    $pkg->_hash_move(
        source => \%opts,
        move   => [qw(vserver volume qtree lun)],
        target => \%stream_opts,
    );

    my %workload_args;
    $pkg->_hash_copy(
        source => \%opts,
        target => \%workload_args,
        map    => {
            uuid           => 'workload-uuid',
            workload       => 'workload-name',
            class          => 'workload-class',
            match_priority => 'match-priority',
            policy         => 'policy-name',
            policy_uuid    => 'policy-uuid',
        },
    );

    my $response  = $apiset->qos_workload_create(%workload_args);
    my $output    = $response->get_parsed_output();
    my $wkld_uuid = $output->[0]->{'workload-uuid'};

    # Create a stream only if some stream characteristics
    # were specified.

    if (%stream_opts) {
        NACL::C::QosWorkloadStream->create(
            command_interface => $command_interface,
            apiset_must       => {interface => 'ZAPI'},
            workload          => $workload,
            %stream_opts,
        );
    } ## end if (%stream_opts)
   my $wkld_obj = $pkg->new(
           workload          => $opts{workload},
           command_interface => $command_interface,
           );

# Cache workload UUID for use by subsequent ZAPI instance
# method calls.
    $wkld_obj->{_uuid} = $wkld_uuid;
    $Log->exit() if $may_exit;
    return $wkld_obj;
} ## end sub _create_cmode_zapi


__PACKAGE__->_load_event_class();

sub _primary_keys_validate_spec {
    return ('workload' => {type => SCALAR},);
}

# ZAPI calls generally use the UUID, rather than the workload
# name, to identify a workload. The following routine will
# fetch the UUID when first called, and cache it in the
# object for subsequent use.

sub _fetch_and_cache_uuid {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate(
        @_,
        {   command_interface => {isa  => 'NACL::C::CommandInterface'},
            workload          => {type => SCALAR},
        },
    );

    # Instance method calls can cache the UUID, or use
    # its previously-cached value.

    if (ref $pkg_or_obj) {
        if (!$pkg_or_obj->{_uuid}) {
            $pkg_or_obj->{_uuid}
                = $pkg_or_obj->get_one_state_attribute('uuid',
                apiset_should => {interface => 'ZAPI'},);

        } ## end if ( !$pkg_or_obj->{_uuid...

        $Log->exit() if $may_exit;
        return $pkg_or_obj->{_uuid};
    } ## end if ( ref $pkg_or_obj )

    # However, static method calls cannot cache the UUID,
    # and must always fetch it.

    my $wkld_state = NACL::CS::QosWorkload->fetch(
        command_interface => $opts{command_interface},
        apiset_should     => {interface => 'ZAPI'},
        requested_fields  => [qw(uuid)],
        filter            => {workload => $opts{workload},},
    );

    $Log->exit() if $may_exit;
    return $wkld_state->uuid();
} ## end sub _fetch_and_cache_uuid

# Fields needs to be copied for delete and modify methods
# under SN build
sub _sn_copy {
    $Log->enter() if $may_enter;
    my $copy_fields = [
        qw(category ext-cache file lun match-priority
            policy-group qtree read-ahead volume vserver wid)
    ];
    $Log->exit() if $may_exit;
    return $copy_fields;
}

# Fields needs to be mapped for delete and modify methods
# under SN build
sub _sn_map {
    $Log->enter() if $may_enter;
    my $map_fields = {
        workload => 'workload-name',
        uuid     => 'workload-uuid',
        class    => 'workload-class',
    };
    $Log->exit() if $may_exit;
    return $map_fields;
}

# Fields needs to be copied for delete and modify methods
# under RR build
sub _rr_copy {
    $Log->enter() if $may_enter;
    my $copy_fields = [qw( match-priority policy-uuid)];
    $Log->exit() if $may_exit;
    return $copy_fields;
}

# Fields needs to be mapped for delete and modify methods
# under SN build
sub _rr_map {
    $Log->enter() if $may_enter;
    my $map_fields = {
        workload => 'workload-name',
        uuid     => 'workload-uuid',
        class    => 'workload-class',
    };
    $Log->exit() if $may_exit;
    return $map_fields;
}
1;
