# $Id$

#
# Copyright (c) 2012 NetApp, Inc., All Rights Reserved
# Any use, modification, or distribution is prohibited
# without prior written consent from NetApp, Inc.
#
## @summary restore Component Module for 7Mode and Nodescope.
## @author dl-dpg-nb-automation; dl-nacl-dev
## @status shared
## @pod here

=head1 NAME

NACL::C::Restore

=head1 DESCRIPTION

C<NACL::C::Restore> is a derived class of
L<NACL::C::Component::ONTAP|lib-NACL-C-Component-ONTAP-pm>.
It represents restore operations.

=cut

=head1 METHODS

=head2 Restore

   NACL::C::Restore->restore(
        command_interface => $node,
        "block-size"      =>  "rfDv",
        "tape-name"       =>  $TAPE,
        "target-location" =>  $VOL
        );

(Class method)
This method is used to restore from tape drive.

Uses a 7Mode/Nodescope CLI APISet.

Invokes "restore" command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

=item C<< 'block-size' => $BlockSize >>

(Required) block-size

Applicable for 7Mode CLI.

Maps to:

7M CLI : block size (Programmatic name: "block-size")

=item C<< 'tape-name' => 'nrst0a' >>

(Required) tape-name

Applicable for 7Mode CLI.

Maps to:

7M CLI : tape name (Programmatic name: "tape-name")

=item C<< 'target-location' =>  '/vol/vol0/home' >>

(Required) target-location

Applicable for 7Mode CLI.

Maps to:

7M CLI : target location (Programmatic name: "target-location")

=item C<< 'restore-all' => $boolean >>

Applicable for 7Mode CLI.

Maps to:

7M CLI : r (Programmatic name: "restore-all")

=item C<< 'restart-restore-all' => $boolean >>

Applicable for 7Mode CLI.

Maps to:

7M CLI : R (Programmatic name: "restore-all")

=item C<< 'list-backup-file' => $string >>

Applicable for 7Mode CLI.

Maps to:

7M CLI : t <value> (Programmatic name: "list-backup-file")

=item C<< 'list-qtree-root-files' => $string >>

Applicable for 7Mode CLI.

Maps to:

7M CLI : T <value> (Programmatic name: "list-qtree-root-files")

=item C<< 'extract-files' => $string >>

Applicable for 7Mode CLI.

Maps to:

7M CLI : x <value> (Programmatic name: "extract-files")

=item C<< 'ignore-acls' => $boolean >>

Applicable for 7Mode CLI.

Maps to:

7M CLI : A (Programmatic name: "ignore-acls")

=item C<< 'restore-target' => $string >>

Applicable for 7Mode CLI.

Maps to:

7M CLI : D <value> (Programmatic name: "restore-target")

=item C<< 'ignore-qtree-information' => $boolean >>

Applicable for 7Mode CLI.

Maps to:

7M CLI : Q (Programmatic name: "ignore-qtree-information")

=item C<< 'verify-only' => $boolean >>

Applicable for 7Mode CLI.

Maps to:

7M CLI : N (Programmatic name: "verify-only")

=item C<< 'force-continue' => $boolean >>

Applicable for 7Mode CLI.

Maps to:

7M CLI : F (Programmatic name: "force-continue")

=item C<< 'verbose' => $boolean >>

Applicable for 7Mode CLI.

Maps to:

7M CLI : v (Programmatic name: "verbose")

=item C<< 'confirm-prompt' => $boolean >>

Applicable for 7Mode CLI.

Maps to:

7M CLI : y (Programmatic name: "confirm-prompt")

=item C<< 'next-restore-file' => $string >>

Applicable for 7Mode CLI.

Maps to:

7M CLI : f (Programmatic name: "next-restore-file")

=item C<< 'restore-file-by-number' => $string >>

Applicable for 7Mode CLI.

Maps to:

7M CLI : s (Programmatic name: "restore-file-by-number")

=back

=back

=cut

package NACL::C::Restore;

use strict;
use warnings;
use NACL::ComponentUtils qw (_optional_scalars);
use base 'NACL::C::Component::ONTAP';
use Params::Validate qw(validate validate_with SCALAR ARRAYREF);
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Data::Dumper;
use NACL::APISet::Exceptions::InvalidParamValueException qw(:try);

use NACL::C::Options;

sub restore {
    my $pkg  = shift;
    my %opts = $pkg->_common_validate_with(
        params          => \@_,
        additional_spec => {
            _optional_scalars(
                qw (restore-all restart-restore-all list-backup-file list-qtree-root-files
                extract-files ignore-acls restore-target ignore-qtree-information method-timeout
                restore-block-size verify-only force-continue verbose confirm-prompt
                block-size target-location next-restore-file restore-file-by-number tape-name)
            )
        },
    );

    return $pkg->call_on_apiset(
        %opts,
        choices => [

            {   method    => '_restore_7mode_cli',
                interface => 'CLI',
                set       => '7Mode|Nodescope',
            }
        ],
    );
} ## end sub dump

sub _restore_7mode_cli {

    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );
    # enabling options at d blade for restore
    # Refer 505668 for more informaiton.
    try {
        NACL::C::Options->option(
                command_interface => $opts{command_interface},
                'option-name'     => 'nodescope.reenabledcmds',
                'option-value'    => 'restore',
            ) ;
    } catch NACL::APISet::Exceptions::InvalidParamValueException with {};

    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};

    $pkg->_hash_move(
        source => \%opts,
        target => \%opts,
        map => { 'method-timeout' => 'connectrec-timeout' }
    );

    $apiset->restore(%opts);
    $Log->exit() if $may_exit;
}


__PACKAGE__->_load_event_class();

sub _primary_keys_validate_spec {
    return ();
} ## end sub _primary_keys_validate_spec

1;
