#
# Copyright (c) 2001-2013 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary SecurityLogin Component Module
## @author dl-nacl-dev@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::C::SecurityLogin

=head1 DESCRIPTION

C<NACL::C::SecurityLogin> is a derived class of
L<NACL::C::Component::ONTAP|lib-NACL-C-Component-ONTAP-pm>. It represents access
to an ONTAP SecurityLogin.  A related class is
L<NACL::CS::SecurityLogin|lib-NACL-CS-SecurityLogin-pm>, which represents
the past state of an ONTAP SecurityLogin.

=head1 ATTRIBUTES

=head2 command_interface

See L<NACL::C::Component::command_interface|lib-NACL-C-Component-pm/command_interface>

=head2 authmethod

=head2 vserver

=head2 application

=head2 user-or-group-name

=cut

package NACL::C::SecurityLogin;

use strict;
use warnings;

use NACL::CS::SecurityLogin;
use NACL::ComponentUtils qw (_optional_scalars Dumper);
use NACL::C::Exceptions::SecurityLogin::AlreadyExists     ();
use NACL::C::Exceptions::SecurityLogin::DoesNotExist      ();
use NACL::C::Exceptions::SecurityLogin::DbFileBusy        ();
use NACL::C::Exceptions::SecurityLogin::InvalidOperationOrProtocol ();
use NACL::C::Exceptions::SecurityLogin::NewPasswordSameAsOldPassword ();
use NACL::C::Exceptions::SecurityLogin::InvalidPassword;
use NACL::Exceptions::InvalidChoice ();
use NATE::BaseException qw(:try);
use NATE::Exceptions::Argument qw(:try);
use NACL::CS::Vserver;
use NACL::Exceptions::InvalidChoice ();

use base 'NACL::C::Component::ONTAP';
use Params::Validate qw(validate validate_with SCALAR ARRAYREF UNDEF HASHREF);
use NACL::C::_Mixins::SecurityLogin qw(:all);
use NACL::C::Cserver;

use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use NACL::Exceptions::InvalidChoice;

=head1 METHODS

=head2 new

  my $SecurityLogin =
      NACL::C::SecurityLogin->new(
                                  command_interface=>$ci,
                                  authmethod =>$authmethod,
                                  vserver =>$vserver,
                                  application =>$application,
                                  user-or-group-name =>$user-or-group-name,
                                 );

Return a new SecurityLogin component object with the given attributes.

=cut

use Class::MethodMaker [
    scalar => 'authmethod',
    scalar => 'vserver',
    scalar => 'application',
    scalar => 'user_or_group_name',
    scalar => [
        {   '-default_ctor' => sub {
                    $_[0] ->user_or_group_name(),
            },
        },
        'username'
    ],
];

sub init {
    $Log->enter() if $may_enter;
    my ($self, @args) = @_;
    $self->SUPER::init(@args);
    $self->_vserver_context_init(@args);
    $Log->exit() if $may_exit;
}

=head2 create

    my $SecurityLogin_object =
        NACL::C::SecurityLogin->create(
                                       command_interface => $ci,
                                       authmethod => $authmethod,
                                       vserver => $vserver,
                                       application => $application,
                                       user-or-group-name => $user-or-group-name,
                                       role => $role,
                                       snmpv3-login-info => \%snmpv3_login_info,
                                       ...
                                      );

(Class method)
This method is used to add a login method.

Uses a CMode CLI/ZAPI APISet.
  Invokes 'security-login-create' API for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "authmethod" => $authmethod >>

(Required)
  Applicable for CMode CLI/ZAPI
  Maps to:
    CM ZAPI: authentication-method

=item C<< "vserver" => $vserver >>

(Required)
  Applicable for CMode CLI/ZAPI
  Maps to:
    CM ZAPI: vserver

=item C<< "application" => $application >>

(Required)
  Applicable for CMode CLI/ZAPI
  Maps to:
    CM ZAPI: application

=item C<< "user-or-group-name" => $user-or-group-name >>

(Required)
User Name or Active Directory Group Name
Applicable for CMode CLI/ZAPI, non-iter.
Maps to:
  CM ZAPI: user-name

=item C<< "password" => $password >>

(Required) For CLI, passed to "Please enter a password for user xyz"
prompt and "Please enter it again" prompts as they appear.
  Applicable for CMode CLI/ZAPI
  Maps to:
    CM ZAPI: password

=item C<< 'is-password-hashed' => $boolean >>

 Applicable for CMode ZAPI

(Optional) default=false ,if true password must be rehashed with
proper hash function before input otherwise password will be entered
as simple string

=item C<< 'comment' => $string >>

  Applicable for CMode CLI/ZAPI
  Maps to:
    CM ZAPI: comment

=item C<< 'role' => $string >>

  Applicable for CMode CLI/ZAPI
  Maps to:
    CM ZAPI: role-name

=item C<< 'require-passwd-update' => $string >>

  Applicable only for CMode CLI

=item C<< 'profile' => $string >>

  Applicable only for CMode CLI

=item C<< snmpv3-login-info => $snmpv3-login-info >>

(Optional) snmpv3-login-info field of snmp login usm info to create.^M
It is of type hashref. E.g.
following key - value pairs has to be passed in the snmpv3-login-info hash
'engine-id' => $value1,
'authentication-protocol' => $value2,
'authentication-password' => $value3,
'privacy-protocol'  => $value4,
'privacy-password' => $value5,

where $value1 to $value5 are simple string values

=item C<< 'extended_query' => $string >>

(Optional)
  Applicable only for CMode CLI

=item C<< '*' >>

This method takes all other options supported by
"security login create" command in ONTAP.

Applicable for CMode CLI.

=over

=back

=item Exceptions

=over

=item C<NACL::C::Exceptions::SecurityLogin::AlreadyExists>

This type of exception gets thrown if an attempt is made to create a
security login that already exists.

=item C<NACL::C::Exceptions::SecurityLogin::DbFileBusy>

This exception is thrown if the password db file is busy when a security login
operation is performed.

=item C<NACL::C::Exceptions::SecurityLogin::InvalidOperationOrProtocol>

This exception is thrown when Security Login performed with invalid protocol or operation.

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=back

=back

=cut

sub create {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( password
                    comment role
                    require-passwd-update
                    profile is-password-hashed
                    )
            ),
          'snmpv3-login-info' => { type => HASHREF, optional => 1 },
        },
        cli_cmd => 'security login create',
    );

    $pkg->call_on_apiset_with_exceptions(
        %opts,
        choices => [
            {   method    => '_create_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
                check     => '_check_cli_create',
            },
            {   method    => '_create_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::SecurityLogin::create',
        nacl_exceptions => [qw(AlreadyExists DbFileBusy
            InvalidOperationOrProtocol InvalidPassword)],
    );
    
    my $obj = $pkg->new(
        command_interface  => $opts{'command_interface'},
        authmethod         => $opts{'authmethod'},
        vserver            => $opts{'vserver'},
        application        => $opts{'application'},
        user_or_group_name => $opts{'user-or-group-name'} || $opts{'username'},
    );
    $Log->exit() if $may_exit;
    return $obj;
} ## end sub create

sub _create_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, %opts) = @_;
    my $snmpv3_info     = delete $opts{'snmpv3-login-info'} if ($opts{'snmpv3-login-info'});
    ## in case of no value , engine id defaults to empty string
    $snmpv3_info->{'engine-id'} = $snmpv3_info->{'engine-id'} || '' ;
    $snmpv3_info->{'authentication-protocol'} = $snmpv3_info->{'authentication-protocol'} || '' ;
    my $password        = delete $opts{password};

    # When an invalid password is supplied, it says the password is invalid and
    # then re-prompts for the password. The effect is that it keeps answering
    # with the invalid password, so goes into an infinite loop.
    # To prevent this, we keep track of how many times the password has been
    # sent, and on the second occasion send a Ctrl+C to interrupt the command.
    # See burt 955644.
    my $pwd_count = 0;
    my $prompts_answers = [
        "Please enter a password for user '\\S*': " => sub {
            if ($pwd_count++) {
                return "\cC";
            } else {
                return $password;
            }
        },
        "Please enter it again: "                                 => $password,
        '.*authoritative entity.*'                                => $snmpv3_info->{'engine-id'},
        '.*Which authentication protocol do you want to choose.*' => $snmpv3_info->{'authentication-protocol'},
        '.*authentication protocol password.*'                    => $snmpv3_info->{'authentication-password'},
        '.*authentication protocol password again.*'              => $snmpv3_info->{'authentication-password'},
        '.*Which privacy protocol do you want to choose.*'        => $snmpv3_info->{'privacy-protocol'},
        '.*privacy protocol password .*'                          => $snmpv3_info->{'privacy-password'},
        '.*privacy protocol password again.*'                     => $snmpv3_info->{'privacy-password'},
    ];

    my $response;
    try {
        $pkg->_base_backend_cmode_cli(
            %opts,
            api             => 'security_login_create',
            additional_args => {'connectrec-match_table' => $prompts_answers},
            frontend_method => 'create',
            apiset_response_obj => \$response,
        );
    } finally {
        if ($pwd_count > 1) {
            # This means that the password was prompted for more than once, so
            # the command was interrupted. To prevent off-by-one prompt
            # issues, the connection needs to be resynced.
            $opts{apiset}->get_connection()->resync();
            $Log->exit() if $may_exit;
            NACL::C::Exceptions::SecurityLogin::InvalidPassword->throw(
                text => "Password does not conform to filer conventions\n"
                    . $response->get_raw_output(),
                response_object => $response,
            );
        }
    };
   $Log->exit() if $may_exit;
} ## end sub _create_cmode_cli

sub _create_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg, %opts) = @_;
    my $snmpv3_info     = delete $opts{'snmpv3-login-info'};

    # provides default value for vserver and role for present cserver in command_interface
    unless (defined $opts{vserver}) {
        my $cserver =
            NACL::C::Cserver->find(command_interface => $opts{command_interface});
        if ($cserver) {
            $opts{vserver} = $cserver->cserver();
            $opts{role} ||= 'admin';
        }
    }
    $opts{'snmpv3-login-info'} = $snmpv3_info if($snmpv3_info);
    $pkg->_base_backend_cmode_zapi(
        %opts,
        api  => 'security-login-create',
        copy => [qw( snmpv3-login-info application vserver password comment is-password-hashed)],
        map  => {
            "user-or-group-name" => "user-name",
            "username"           => "user-name",
            "authmethod"         => "authentication-method",
            "role"               => "role-name",
        },
        frontend_method => 'create',
    );

    $Log->exit() if $may_exit;
} ## end sub _create_cmode_zapi

=head2 unlock

   NACL::C::SecurityLogin->unlock(
                                  command_interface => $ci,
                                  vserver => $vserver,
                                  user-or-group-name => $user-or-group-name,
                                  ...
                                 );

(Class method)
This method is used to unlock a user account with password auth method.

Uses a CMode CLI/ZAPI or 7Mode CLI APISet.
  Invokes 'security-login-unlock' API for CMode ZAPI.
  Invokes 'useradmin diaguser unlock' command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "vserver" => $vserver >>

(Required)
  Applicable for CMode CLI/ZAPI
  Maps to:
    CM ZAPI: vserver

=item C<< "user-or-group-name" => $user-or-group-name >>

(Required)

  Applicable for CMode CLI/ZAPI,7Mode CLI
  Maps to:
    CM ZAPI: user-or-group-name
    7M CLI : login name (Programmatic name: "login-name")

=item C<< 'extended_query' => $string >>

(Optional)
  Applicable only for CMode CLI

=item C<< '*' >>

This method takes all other options supported by
"security login unlock" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<NACL::C::Exceptions::SecurityLogin::DbFileBusy>

This exception is thrown if the password db file is busy when a security login
operation is performed

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub unlock {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    my $ret;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( vserver
                    )
            ),
        },
        ignore_primary_keys => 1,
        cli_cmd             => 'security login unlock',
    );
    try {
        $ret = $pkg->call_on_apiset(
            %opts,
            choices => [
                {   method    => '_unlock_cmode_cli',
                    interface => 'CLI',
                    set       => 'CMode',
                },
                {   method    => '_unlock_7mode_cli',
                    interface => 'CLI',
                    set       => '7Mode',
                },
                {   method    => '_unlock_cmode_zapi',
                    interface => 'ZAPI',
                    set       => 'CMode',
                },
            ],
            frontend => 'NACL::C::SecurityLogin::unlock',
        );
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = $_[0];
        NACL::C::Exceptions::SecurityLogin::DbFileBusy
            ->detect_convert_and_throw(exception => $exception);
        $Log->exit() if $may_exit;
        $exception->throw();
    };

    $Log->exit() if $may_exit;
    return $ret;
} ## end sub unlock

sub _unlock_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, %opts) = @_;
    $pkg->_base_backend_cmode_cli(%opts,
        api     => 'security_login_unlock',
    );
    $Log->exit() if $may_exit;
} ## end sub _unlock_cmode_cli

sub _unlock_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );
    my %mode_args;
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    $pkg->_hash_copy(
        source => \%opts,
        map    => {"username" => "login-name",
                   "user-or-group-name" => "login-name",},
        target => \%mode_args,
    );
    if ($mode_args{'login-name'} eq "diag") {
        $apiset->useradmin_diaguser_unlock();
        return $pkg->new(
            'vserver'         => undef,
            'username'        => $mode_args{'login-name'},
            command_interface => $command_interface
        );

    } ## end if ( $mode_args{'login-name'...
    $Log->exit() if $may_exit;
} ## end sub _unlock_7mode_cli

sub _unlock_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_base_backend_cmode_zapi(@args,
        api  => 'security_login_unlock',
        copy => [qw( vserver )],
        map  => {"user-or-group-name" => "user-name",
                 "username"           => "user-name",},
    );
    $Log->exit() if $may_exit;
} ## end sub _unlock_cmode_zapi

=head2 password

   NACL::C::SecurityLogin->password(
                                    command_interface => $ci,
                                    vserver => $vserver,
                                    user-or-group-name => $user-or-group-name,
                                    force => $boolean,
                                    oldpwd => $password,
                                    newpwd => $password,
                                    newpwd2 => $password,
                                    ...
                                   );

(Class method)
This method is used to modify a password for a user.

Uses a CMode CLI/ZAPI APISet.
  Invokes security-login-modify-password for CMode ZAPI.
  Invokes 'useradmin diaguser password' command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "authmethod" => $authmethod >>

(Required)

=item C<< "vserver" => $vserver >>

(Required)
  Applicable only for CMode CLI

=item C<< "application" => $application >>

(Required)

=item C<< "user-or-group-name" => $user-or-group-name >>

(Required)
  Applicable for CMode CLI/ZAPI,7Mode CLI
  Maps to:
    CM ZAPI: user-or-group-name
    7M CLI : login name (Programmatic name: "login-name")

=item C<< "oldpwd" => $string >>

(Optional)
  Applicable only for CMode CLI


=item C<< "newpwd" => $string >>

(Required) For CLI, passed to "Please enter a new password" prompt as
it appears.
  Applicable for CMode CLI/ZAPI
  Maps to:
    CM ZAPI: new-password

=item C<< "newpwd2" => $string >>

(Optional) For CLI, passed to "Please enter it again" prompt as it
appears.  This defaults to newpwd if not provided.  It will probably
be inapplicable for ZAPI (once ZAPI is supported).

  Applicable only for CMode CLI

=item C<< "force" => 0|1 >>

(Optional)
  Applicable only for CMode CLI

=item C<< 'profile' => $string >>

  Applicable only for CMode CLI

=item C<< 'extended_query' => $string >>

(Optional)
  Applicable only for CMode CLI

=item C<< 'new-password' => $string >>

New Password
Applicable for CMode CLI

=item C<< 'old-password' => $string >>

Old Password
Applicable for CMode CLI

=item C<< 'retype-password' => $string >>

Retype Password
Applicable for CMode CLI

=item C<< '*' >>

This method takes all other options supported by
"security login password" command in ONTAP.

Applicable for CMode CLI.

=item C<< 'is-password-hashed' => $boolean >>

Applicable for CMode ZAPI

(Optional) default=false ,if true password must be rehashed with
proper hash function before input otherwise password will be entered
as simple string

=back

=back

=over

=item Exceptions

=over

=item C<NACL::C::Exceptions::SecurityLogin::DbFileBusy>

This exception is thrown if the password db file is busy when a security login
operation is performed

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub password {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    my $ret;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( newpwd oldpwd newpwd2 force vserver
                    profile username

                    new-password old-password retype-password 
                    hash-function is-password-hashed
                    )
            ),
        },
        ignore_primary_keys => 1,
        cli_cmd             => 'security login password',
    );
    try {
        $ret = $pkg->call_on_apiset(
            %opts,
            choices => [
                {   method    => '_password_cmode_cli',
                    interface => 'CLI',
                    set       => 'CMode',
                    check     => '_check_cli_password',
                },
                {   method    => '_password_7mode_cli',
                    interface => 'CLI',
                    set       => '7Mode',
                },
                {   method    => '_password_cmode_zapi',
                    interface => 'ZAPI',
                    set       => 'CMode',
                    check     => '_cmode_zapi_check',
                },
            ],
            frontend => 'NACL::C::SecurityLogin::password',
        );
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = $_[0];
        NACL::C::Exceptions::SecurityLogin::DbFileBusy
            ->detect_convert_and_throw(exception => $exception);
        NACL::C::Exceptions::SecurityLogin::NewPasswordSameAsOldPassword
            ->detect_convert_and_throw(exception => $exception);
        $Log->exit() if $may_exit;
        $exception->throw();
    };
    $Log->exit() if $may_exit;
    return $ret;
} ## end sub password

sub _password_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, %opts) = @_;

    # Convert newpwd and newpwd2 into prompts instead of arguments
    my $oldpwd  = delete $opts{oldpwd};
    my $newpwd  = delete $opts{newpwd};
    my $newpwd2 = delete $opts{newpwd2};
    if (defined($newpwd) && !defined($newpwd2)) {
        $newpwd2 = $newpwd;
    }
    my $prompts_answers = [
        '.*your current password.*' => $oldpwd,
        '.*new password.*'          => $newpwd,
        '.*again.*'                 => $newpwd2,
    ];

    if (defined($opts{force})) {
        $opts{force}
            = ($opts{force} =~ /1|true/i) ? "true" : "false";
    }

    my $vserver;
    if (defined ($opts{vserver})) {
        $vserver = $opts{vserver};
        $opts{vserver_operated_on} = \$vserver;
    }

    $pkg->_base_backend_cmode_cli(%opts,
        api             => 'security_login_password',
        additional_args => {'connectrec-match_table' => $prompts_answers},
    );
    $Log->exit() if $may_exit;
} ## end sub _password_cmode_cli

sub _password_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );

    # Convert newpwd and newpwd2 into prompts instead of arguments

    my $newpwd  = delete $opts{newpwd};
    my $newpwd2 = delete $opts{newpwd2};
    if (defined($newpwd) && !defined($newpwd2)) {
        $newpwd2 = $newpwd;
    }
    $Log->comment("PASSWORD $newpwd\n");
    my $prompts_answers = [
        '.*new password.*' => $newpwd,
        '.*again.*'        => $newpwd2,
    ];
    my %mode_args;
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    $pkg->_hash_copy(
        source => \%opts,
        map    => {"username" => "login-name",
                   "user-or-group-name" => "login-name",},
        target => \%mode_args,
    );

    if ($mode_args{'login-name'} eq "diag") {
        my $name = delete $mode_args{'login-name'};
        $apiset->useradmin_diaguser_password(
            'connectrec-match_table' => $prompts_answers);
        return $pkg->new(
            'vserver'         => undef,
            'username'        => $name,
            command_interface => $command_interface
        );

    } ## end if ( $mode_args{'login-name'...
    $Log->exit() if $may_exit;

} ## end sub _password_7mode_cli

sub _password_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg, %opts) = @_;
    my $vserver;

    if (defined ($opts{vserver})) {
        $vserver = $opts{vserver};
        $opts{vserver_operated_on} = \$vserver;
    }

    $pkg->_base_backend_cmode_zapi(%opts,
        api    => 'security_login_modify_password',
        map    => {
            "username"           => 'user-name',
            "user-or-group-name" => 'user-name',
            'newpwd'             => 'new-password',
        },
        inapplicable_for_help_xml => [qw(oldpwd newpwd2)],
    );
    $Log->exit() if $may_exit;
} ## end sub _password_cmode_zapi

sub _cmode_zapi_check {
    $Log->enter() if $may_enter;
    my ($pkg, %opts) = @_;
    if (defined $opts{'force'}) {
          my $msg = 'Could not use ZAPI implementation of NACL::C::SecurityLogin->password().'
                    .' since the CMode ZAPI does not support "force" option';

          $Log->comment($msg);
          $Log->exit() if $may_exit;
          NACL::Exceptions::InvalidChoice->throw($msg);
     }


    $Log->exit() if $may_exit;
} ## end sub _cmode_zapi_check

sub _check_cli_password {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    if (defined $opts{'is-password-hashed'}) {
        $Log->exit() if $may_exit;
        NACL::Exceptions::InvalidChoice->throw('Could not use CLI '
                . 'since the option \'is-password-hashed\' is not supported '
                . 'by the CLI security_login_password');
    }
    $Log->exit() if $may_exit;
}

=head2 modify

   NACL::C::SecurityLogin->modify(
                                  command_interface => $ci,
                                  authmethod => $authmethod,
                                  vserver => $vserver,
                                  application => $application,
                                  user-or-group-name => $user-or-group-name,
                                  ...,
                                 );

         or

  $SecurityLogin_object->modify(...);

(Class or instance method)
This method is used to modify a login method.

Uses a CMode CLI/ZAPI APISet.
  Invokes 'security-login-modify-iter' API for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "authmethod" => $authmethod >>

(Required for class method, Not Applicable for instance method)
  Applicable for CMode CLI/ZAPI
  Maps to:
    CM ZAPI: authentication-method

=item C<< "vserver" => $vserver >>

(Required for class method, Not Applicable for instance method)
  Applicable for CMode CLI/ZAPI
  Maps to:
    CM ZAPI: vserver

=item C<< "application" => $application >>

(Required for class method, Not Applicable for instance method)
  Applicable for CMode CLI/ZAPI
  Maps to:
    CM ZAPI: application

=item C<< "user-or-group-name" => $user-or-group-name >>

(Required for class method, Not Applicable for instance method)
User Name or Active Directory Group Name
Maps to:
  CM ZAPI: user-name

=item C<< "acctlocked" => $string >>

(Optional)
  Applicable only for CMode CLI

=item C<< "role" => $string >>

  Applicable for CMode CLI/ZAPI
  Maps to:
    CM ZAPI: role-name

=item C<< 'comment' => $string >>

  Applicable only for CMode CLI

=item C<< 'require-passwd-update' => $string >>

  Applicable only for CMode CLI

=item C<< 'profile' => $string >>

  Applicable only for CMode CLI

=item C<< 'extended_query' => $string >>

(Optional)
  Applicable only for CMode CLI

=item C<< '*' >>

This method takes all other options supported by
"security login modify" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<NACL::C::Exceptions::SecurityLogin::DbFileBusy>

This exception is thrown if the password db file is busy when a security login
operation is performed

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub modify {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my $ret;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( acctlocked role

                    comment

                    require-passwd-update

                    profile
                    )
            ),
        },
        cli_cmd => 'security login modify',
    );
    $Log->exit() if $may_exit;

    try {
        $ret = $pkg_or_obj->call_on_apiset(
            %opts,
            choices => [
                {   method    => '_modify_cmode_cli',
                    interface => 'CLI',
                    set       => 'CMode',
                },
                {   method    => '_modify_cmode_zapi',
                    interface => 'ZAPI',
                    set       => 'CMode',
                },
            ],
            frontend => 'NACL::C::SecurityLogin::modify',
        );
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = $_[0];
        NACL::C::Exceptions::SecurityLogin::DbFileBusy
            ->detect_convert_and_throw(exception => $exception);
        $Log->exit() if $may_exit;
        $exception->throw();
    };
    return $ret;
} ## end sub modify

sub _modify_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    my $vserver = $opts{vserver};

    $pkg_or_obj->_base_backend_cmode_cli(
        %opts,
        api => 'security_login_modify',
        vserver_operated_on => \$vserver,
    );

    $Log->exit() if $may_exit;
} ## end sub _modify_cmode_cli

sub _modify_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my $common_copy = $pkg_or_obj->_common_zapi_copy();
    my $common_map  = $pkg_or_obj->_common_zapi_map();

    $pkg_or_obj->_iter_helper(
        @args,
        api                  => 'security-login-modify-iter',
        show_cmd             => 'security login show',
        copy                 => $common_copy,
        map                  => $common_map,
        does_not_exist_error => "entry doesn't exist",
        does_not_exist_exception =>
            'NACL::C::Exceptions::SecurityLogin::DoesNotExist'
    );
    $Log->exit() if $may_exit;
} ## end sub _modify_cmode_zapi

=head2 delete

   NACL::C::SecurityLogin->delete(
                                  command_interface => $ci,
                                  authmethod => $authmethod,
                                  vserver => $vserver,
                                  application => $application,
                                  user-or-group-name => $user-or-group-name,
                                 );

         or

   $SecurityLogin_object->delete(...);

(Class or instance method)
This method is used to delete a login method.

Uses a CMode CLI/ZAPI APISet.
  Invokes 'security-login-delete-iter' API for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "authmethod" => $authmethod >>

(Required for class method, Not Applicable for instance method)
  Applicable for CMode CLI/ZAPI
  Maps to:
    CM ZAPI: authentication-method

=item C<< "vserver" => $vserver >>

(Required for class method, Not Applicable for instance method)
  Applicable for CMode CLI/ZAPI
  Maps to:
    CM ZAPI: vserver

=item C<< "application" => $application >>

(Required for class method, Not Applicable for instance method)
  Applicable for CMode CLI/ZAPI
  Maps to:
    CM ZAPI: application

=item C<< "user-or-group-name" => $user-or-group-name >>

(Required for class method, Not Applicable for instance method)
User Name or Active Directory Group Name
Maps to:
  CM ZAPI: user-name

=item C<< 'extended_query' => $string >>

(Optional)
  Applicable only for CMode CLI

=item C<< '*' >>

This method takes all other options supported by
"security login delete" command in ONTAP.

Applicable for CMode CLI.

=over

=back

=item Exceptions

=over

=item C<NACL::C::Exceptions::SecurityLogin::DoesNotExist>

This type of exception gets thrown if an attempt is made to delete a
security login which does not exists.

=item C<NACL::C::Exceptions::SecurityLogin::DbFileBusy>

This exception is thrown if the password db file is busy when a security login
operation is performed

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=back

=back

=cut

sub delete {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my $ret;
    my %opts   = $pkg_or_obj->_common_validate_with(
        params => \@args,
        cli_cmd=> 'security login delete',
    );
    $Log->exit() if $may_exit;
    try {
        $ret = $pkg_or_obj->call_on_apiset(
            %opts,
            choices => [
                {   method    => '_delete_cmode_cli',
                    interface => 'CLI',
                    set       => 'CMode',
                },
                {   method    => '_delete_cmode_zapi',
                    interface => 'ZAPI',
                    set       => 'CMode',
                },
            ],
            frontend => 'NACL::C::SecurityLogin::delete',
        );
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = $_[0];
        NACL::C::Exceptions::SecurityLogin::DoesNotExist
            ->detect_convert_and_throw(exception => $exception);
        NACL::C::Exceptions::SecurityLogin::DbFileBusy
            ->detect_convert_and_throw(exception => $exception);
        $Log->exit() if $may_exit;
        $exception->throw();
    };
    return $ret;
} ## end sub delete

sub _delete_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    my $vserver = $opts{vserver};

    if (defined ($opts{'username'} )) {
        $opts{'user-or-group-name'} = delete $opts{'username'};
    }

    $pkg_or_obj->_base_backend_cmode_cli(
        %opts,
        api => 'security_login_delete',
        vserver_operated_on => \$vserver,
    );
        
    $Log->exit() if $may_exit;
} ## end sub _delete_cmode_cli

sub _delete_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;

    my $allowed_cmode_options = [qw( vserver application )];

    $pkg_or_obj->_iter_helper(
        @args,
        api      => 'security-login-delete-iter',
        show_cmd => 'security login show',
        copy     => $allowed_cmode_options,
        map      => {
            "user-or-group-name" => "user-name",
            "username"           => "user-name",
            "authmethod"         => "authentication-method",
            "role"               => "role-name",
            "acctlocked"         => "is-locked",
        },
        does_not_exist_error => "entry doesn't exist",
        does_not_exist_exception =>
            'NACL::C::Exceptions::SecurityLogin::DoesNotExist'
    );
    $Log->exit() if $may_exit;
} ## end sub _delete_cmode_zapi

=head2 lock

   NACL::C::SecurityLogin->lock(
                                command_interface => $ci,
                                vserver => $vserver,
                                user-or-group-name => $user-or-group-name,
                               );

(Class method)
This method is used to lock a user account with password auth method.

Uses a CMode CLI/ZAPI or 7Mode CLI APISet.
  Invokes 'security-login-lock' API for CMode ZAPI.
  Invokes 'useradmin diaguser lock' command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "vserver" => $vserver >>

(Required)
  Applicable for CMode CLI/ZAPI
  Maps to:
    CM ZAPI: vserver

=item C<< "user-or-group-name" => $user-or-group-name >>

(Required)
  Applicable for CMode CLI/ZAPI,7Mode CLI
  Maps to:
    CM ZAPI: user-or-group-name
    7M CLI : login-name

=item C<< 'locktime' => $string >>

  Applicable only for CMode CLI

=item C<< 'extended_query' => $string >>

(Optional)
  Applicable only for CMode CLI

=item C<< '*' >>

This method takes all other options supported by
"security login lock" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<NACL::C::Exceptions::SecurityLogin::DbFileBusy>

This exception is thrown if the password db file is busy when a security login
 operation is performed

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub lock {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    my $ret;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params              => \@args,
        ignore_primary_keys => 1,
        additional_spec     => {
            _optional_scalars(
                qw( vserver locktime
                    )
            ),
        },
        ignore_primary_keys => 1,
        cli_cmd             => 'security login lock',
    );
    $Log->exit() if $may_exit;
    try {
        $ret = $pkg->call_on_apiset(
            %opts,
            choices => [
                {   method    => '_lock_cmode_cli',
                    interface => 'CLI',
                    set       => 'CMode',
                },
                {   method    => '_lock_7mode_cli',
                    interface => 'CLI',
                    set       => '7Mode',
                },
                {   method    => '_lock_cmode_zapi',
                    interface => 'ZAPI',
                    set       => 'CMode',
                },
            ],
            frontend => 'NACL::C::SecurityLogin::lock',
        );
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = $_[0];
        NACL::C::Exceptions::SecurityLogin::DbFileBusy
            ->detect_convert_and_throw(exception => $exception);
        $Log->exit() if $may_exit;
        $exception->throw();
    };

    $Log->exit() if $may_exit;
    return $ret;
} ## end sub lock

sub _lock_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, %opts) = @_;
    $pkg->_base_backend_cmode_cli(%opts,
        api  => 'security_login_lock',
    );
    $Log->exit() if $may_exit;
} ## end sub _lock_cmode_cli

sub _lock_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );
    my %mode_args;
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    $pkg->_hash_copy(
        source => \%opts,
        map    => {"username" => "login-name",
                   "user-or-group-name" => "login-name",},
        target => \%mode_args,
    );
    if ($mode_args{'login-name'} eq "diag") {
        $apiset->useradmin_diaguser_lock();
        return $pkg->new(
            'vserver'         => undef,
            'username'        => $mode_args{'login-name'},
            command_interface => $command_interface
        );

    } ## end if ( $mode_args{'login-name'...
    $Log->exit() if $may_exit;
} ## end sub _lock_7mode_cli

sub _lock_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_base_backend_cmode_zapi(@args,
        api   => 'security_login_lock',
        copy  => [qw( vserver )],
        map    => {"user-or-group-name" => "user-name",
                   "username"           => "user-name",},
    );
    $Log->exit() if $may_exit;
} ## end sub _lock_cmode_zapi

=head2 whoami

    my $whoami_obj = NACL::C::SecurityLogin->whoami(command_interface => $ci,
                          [ filter => { %filter } ,]
			  [ requested_fields => [ @req_fields ]);

                    or

    my @whoami_objs = NACL::C::SecurityLogin->whoami(command_interface => $ci,
                          [ filter => { %filter } ,]
			  [ requested_fields => [ @req_fields ]);

                    or

   my $whoami_obj = $SecurityLogin_obj->whoami();

(Class or Instance method)

This method is a CS-like fetch method which accepts all the parameters of
L<fetch|lib-NACL-ComponentState-pm/fetch> (requested_fields, filter, allow_empty)
and has all of the other properties of fetch methods (callable in scalar or list
context, throws a NoElementsFound exception if no matching elements are found).
The requested fields and filter takes L<ATTRIBUTES|lib-NACL-CS-SecurityLoginWhoami-pm/ATTRIBUTES>
of NACL::CS::SecurityLoginWhoami as input.
This method returns objects of the type L<NACL::CS::SecurityLoginWhoami|lib-NACL-CS-SecurityLoginWhoami-pm>.

=cut
sub whoami {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts = $pkg_or_obj->_handle_obj_cs_like_calls(@args);
    require NACL::CS::SecurityLoginWhoami;
    my @state_objs = NACL::CS::SecurityLoginWhoami->fetch(%opts);
    $Log->exit() if $may_exit;
    return wantarray ? @state_objs : $state_objs[0];
}


=head2 password_prepare_to_downgrade
    NACL::C::SecurityLogin->password_prepare_to_downgrade(
                                command_interface => $ci,
                                ...);


(Class method)
This method is used to reset password features for compatibility with releases earlier than Data ONTAP 8.3.1 and 8.4.0.
Uses CMode CLI.


=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< '*' >>

This method takes all other options supported by "security login password prepare to downgrade" command in ONTAP

=item C<< 'password' => $string >>

It is the SHA2 hash function based passwords support available from LB.0 onwards.
Logged in console admin user password will be asked to be reset  during revert / downgrade operation.

Applicable for CMode CLI

=item C<< other_options >>

This method takes other common options described in L<Common method options|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#Other_options_valid_on_component_methods>

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub password_prepare_to_downgrade {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params              => \@args,
    additional_spec => {
            _optional_scalars(
                qw( password)
            )
        },
        ignore_primary_keys => 1,
        cli_cmd             => 'security login password-prepare-to-downgrade',
    );
    $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_password_prepare_to_downgrade_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
                zapi_type => 'none',
            },

        ],
        frontend => 'NACL::C::SecurityLogin::password_prepare_to_downgrade',
    );

    $Log->exit() if $may_exit;
}

sub _password_prepare_to_downgrade_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, %opts) = @_;
    my $password        = delete $opts{password};
    my $prompts_answers = [
        'Enter a new password:'    => $password,
        'Enter it again:'          => $password
    ];
    $pkg->_base_backend_cmode_cli(%opts,
        api => 'security_login_password_prepare_to_downgrade',
        additional_args => {'connectrec-match_table' => $prompts_answers},
    );
    $Log->exit() if $may_exit;
}


__PACKAGE__->_load_event_class();

sub _primary_keys_validate_spec {
    return (
        'authmethod'         => {type => SCALAR | UNDEF, optional => 1},
        'vserver'            => {type => SCALAR | UNDEF, optional => 1},
        'application'        => {type => SCALAR | UNDEF, optional => 1},
        'user-or-group-name' => {type => SCALAR | UNDEF, optional => 1},
        'username'           => {type => SCALAR | UNDEF, optional => 1},
    );
} ## end sub _primary_keys_validate_spec

sub _check_cli_create {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    if (defined $opts{'is-password-hashed'}) {
        $Log->exit() if $may_exit;
        NACL::Exceptions::InvalidChoice->throw('Could not use CLI '
                . 'since the option \'is-password-hashed\' is not supported '
                . 'by the CLI security_login_create');
    }
    $Log->exit() if $may_exit;
}

1;
