# $Id$
#
# Copyright (c) 2001-2012 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary SecurityLoginRole Component Module
## @author dl-nacl-dev@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::C::SecurityLoginRole

=head1 DESCRIPTION

C<NACL::C::SecurityLoginRole> is a derived class of
L<NACL::C::Component::ONTAP|lib-NACL-C-Component-ONTAP-pm>. It represents access
to an ONTAP SecurityLoginRole.  A related class is
L<NACL::CS::SecurityLoginRole|lib-NACL-CS-SecurityLoginRole-pm>, which represents
the past state of an ONTAP SecurityLoginRole.

=head1 ATTRIBUTES

=head2 command_interface

See L<NACL::C::Component::command_interface|lib-NACL-C-Component-pm/command_interface>

=head2 vserver

=head2 cmddirname

=head2 role

=cut

package NACL::C::SecurityLoginRole;

use strict;
use warnings;

use NACL::CS::SecurityLoginRole;
use NACL::ComponentUtils qw (_optional_scalars);

use base 'NACL::C::Component::ONTAP';
use Params::Validate qw(validate validate_with SCALAR ARRAYREF UNDEF);
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Data::Dumper;

=head1 METHODS

=head2 new

  my $SecurityLoginRole = NACL::C::SecurityLoginRole->new(command_interface=>$ci,
                                   vserver =>$vserver,
                                   cmddirname =>$cmddirname,
                                   role =>$role,
                                   );

Return a new SecurityLoginRole component object with the given attributes.

=cut

use Class::MethodMaker [
    scalar => 'vserver',
    scalar => 'cmddirname',
    scalar => 'role',
    scalar => 'capability',
    scalar => 'access',
];

sub init {
    $Log->enter() if $may_enter;
    my ($self, @args) = @_;
    $self->SUPER::init(@args);
    $self->_vserver_context_init(@args);
    $Log->exit() if $may_exit;
}

=head2 find

    my $SecurityLoginRole  = NACL::C::SecurityLoginRole->find(command_interface => $ci, ...);
    my @SecurityLoginRoles = NACL::C::SecurityLoginRole->find(command_interface => $ci, ...);

(Class method)
A constructor of sorts, which discovers which elements are present (by
interacting with the given command interface).  In an array context it
creates a component object for each present element and returns them.
In a scalar context it creates a component object for just one
element.

This method is generally not implemented in individual components. The base
class (L<NACL::C::Component::ONTAP|lib-NACL-C-Component-ONTAP-pm> provides an implementation.
See <L<NACL::C::Component::find|lib-NACL-C-Component-pm/find> for a more
detailed description along with explanation of all the options accepted.

=head2 state

    my $SecurityLoginRole_state = $NACL::C::SecurityLoginRole->state();

(Instance method)
Fetch the current state of the element associated
with this component, and return a Component State object corresponding
to this component (calling this on NACL::C::SecurityLoginRole would return
a NACL::CS::SecurityLoginRole object)

This method is generally not implemented in individual components. The base
class (L<NACL::C::Component::ONTAP|lib-NACL-C-Component-ONTAP-pm> provides an implementation.
See <L<NACL::C::Component::state|lib-NACL-C-Component-pm/state> for a more
detailed description along with explanation of all the options accepted.

=cut

=head2 modify

   NACL::C::SecurityLoginRole->modify(
                                command_interface => $ci,
                                vserver => $vserver,
                                cmddirname => $cmddirname,
                                role => $role,
                                ...);

                        or

    $SecurityLoginRole_object->modify(...);

(Class or instance method)
This method is used to modify an access control role.

Uses a CMode CLI/ZAPI or 7Mode CLI APISet.
  Invokes 'security-login-role-modify' API for CMode ZAPI.
  Invokes 'useradmin role modify' command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "vserver" => $vserver >>

(Required for class method, Not Applicable for instance method)
  Applicable for CMode CLI/ZAPI
  Maps to:
    CM ZAPI: vserver

=item C<< "cmddirname" => $cmddirname >>

(Required for class method, Not Applicable for instance method)
  Applicable for CMode CLI/ZAPI,7Mode CLI
  Maps to:
    CM ZAPI: command-directory-name
    7M CLI : -a (Programmatic name: 'capability')

=item C<< "role" => $role >>

(Required for class method, Not Applicable for instance method)
  Applicable for CMode CLI/ZAPI,7Mode CLI
  Maps to:
    CM ZAPI: role-name
    7M CLI : role name (Programmatic name: 'role-name')

=item C<< "query" => $string >>

  Applicable for CMode CLI/ZAPI
  Maps to:
    CM ZAPI: role-query

=item C<< "access" => $string >>

  Applicable for CMode CLI/ZAPI
  Maps to:
    CM ZAPI: access-level

=item C<< 'extended_query' => $string >>

(Optional)
  Applicable only for CMode CLI

=item C<< '*' >>

This method takes all other options supported by
"security login role modify" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub modify {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( query access
                    )
            ),
        },
        cli_cmd => 'security login role modify',
    );
    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_modify_7mode_cli',
                interface => 'CLI',
                set       => '7Mode',
            },
            {   method    => '_modify_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_modify_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::SecurityLoginRole::modify',
    );
    $Log->exit() if $may_exit;
    return $ret;
} ## end sub modify

sub _modify_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    my $vserver = $opts{vserver};
    $opts{query} = '"' . $opts{query} . '"'
        if ((defined($opts{query}))
        && ($opts{query} !~ /^".*"$/));
    $opts{cmddirname} = '"' . $opts{cmddirname} . '"'
        if ((defined($opts{cmddirname}))
        && ($opts{cmddirname} !~ /^".*"$/));
    $pkg_or_obj->_base_backend_cmode_cli(
        %opts,
        api => 'security_login_role_modify',
        vserver_opearted_on => \$vserver
    );
    $Log->exit() if $may_exit;
} ## end sub _modify_cmode_cli

sub _modify_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    delete $opts{command_interface};
    my $apiset = delete $opts{apiset};
    delete $opts{access};
    my %cmode_args;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        map    => {
            "role"       => "role-name",
            "cmddirname" => "capability",
        },
        target => \%cmode_args,
    );
    %opts = %cmode_args;

    if ($opts{'capability'} eq "DEFAULT") {
        $opts{'capability'} = "*";
    }
    $apiset->useradmin_role_modify(%opts);
    $Log->exit() if $may_exit;
} ## end sub _modify_7mode_cli

sub _modify_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my $allowed_cmode_options = [qw( vserver )];
    $pkg_or_obj->_base_backend_cmode_zapi(
        @args,
        api  => 'security_login_role_modify',
        copy   => $allowed_cmode_options,
	map    => {
            "role"       => "role-name",
            "cmddirname" => "command-directory-name",
            "access"     => "access-level",
            "query"      => "role-query",
        },
    );
    $Log->exit() if $may_exit;
} ## end sub _modify_cmode_zapi

=head2 delete
   NACL::C::SecurityLoginRole->delete(
                                command_interface => $ci,
                                vserver => $vserver,
                                cmddirname => $cmddirname,
                                role => $role,
                                ...);

(Class method)
This method is used to delete an access control role.

Uses a CMode CLI/ZAPI or 7Mode CLI APISet.
  Invokes 'security-login-role-delete' API for CMode ZAPI.
  Invokes 'useradmin role delete' command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "vserver" => $vserver >>

(Required)
  Applicable for CMode CLI/ZAPI
  Maps to:
    CM ZAPI: vserver

=item C<< "cmddirname" => $cmddirname >>

(Required)
  Applicable for CMode CLI/ZAPI
  Maps to:
    CM ZAPI: command-directory-name

=item C<< "role" => $role >>

(Required)
  Applicable for CMode CLI/ZAPI,7Mode CLI
  Maps to:
    CM ZAPI: role-name
    7M CLI : role name (Programmatic name: 'role-name')

=item C<< 'extended_query' => $string >>

(Optional)
  Applicable only for CMode CLI

=item C<< '*' >>

This method takes all other options supported by
"security login role delete" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub delete {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params  => \@args,
        cli_cmd => 'security login role delete',
    );
    my $ret = $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_delete_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_delete_7mode_cli',
                interface => 'CLI',
                set       => '7Mode',
            },
            {   method    => '_delete_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::SecurityLoginRole::delete',
    );
    $Log->exit() if $may_exit;
    return $ret;
} ## end sub delete

sub _delete_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, %opts) = @_;
    my $vserver = $opts{vserver};
    $opts{cmddirname} = '"' . $opts{cmddirname} . '"'
        if ((defined($opts{cmddirname}))
        && ($opts{cmddirname} !~ /^".*"$/));
            if($opts{cmddirname} eq '"*"')  {
                 $opts{cmddirname} = '*';
            }
    $pkg->_base_backend_cmode_cli(
        %opts,
        api => 'security_login_role_delete',
        vserver_operated_on => \$vserver,
    );
    $Log->exit() if $may_exit;
} ## end sub _delete_cmode_cli

sub _delete_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );
    delete $opts{command_interface};
    my $apiset = delete $opts{apiset};
    my %cmode_args;
    $pkg->_hash_copy(
        source => \%opts,
        map    => {'role' => 'role-name'},
        target => \%cmode_args,
    );
    $apiset->useradmin_role_delete(%cmode_args);
    $Log->exit() if $may_exit;
} ## end sub _delete_7mode_cli

sub _delete_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    my $allowed_cmode_options = [qw( vserver )];
    $pkg->_base_backend_cmode_zapi(
        @args,
        api  => 'security_login_role_delete',
        copy => $allowed_cmode_options,
        map => {
            "role"       => "role-name",
            "cmddirname" => "command-directory-name",
        },
    );
    $Log->exit() if $may_exit;
} ## end sub _delete_cmode_zapi

=head2 create

    my $SecurityLoginRole_object = NACL::C::SecurityLoginRole->create(
                                command_interface => $ci,
                                vserver => $vserver,
                                cmddirname => $cmddirname,
                                role => $role,
                                ...);

(Class method)
This method is used to add an access control role.

Uses a CMode CLI/ZAPI or 7Mode CLI APISet.
  Invokes 'security-login-role-create' API for CMode ZAPI.
  Invokes 'useradmin role add' command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "vserver" => $vserver >>

(Required)
  Applicable for CMode CLI/ZAPI
  Maps to:
    CM ZAPI: vserver

=item C<< "cmddirname" => $cmddirname >>

(Required)
  Applicable for CMode CLI/ZAPI,7Mode CLI
  Maps to:
    CM ZAPI: command-directory-name
    7M CLI : -a (Programmatic name: 'capability')

=item C<< "role" => $role >>

(Required)
  Applicable for CMode CLI/ZAPI,7Mode CLI
  Maps to:
    CM ZAPI: role-name
    7M CLI : role name (Programmatic name: 'role-name')

=item C<< "query" => $string >>

  Applicable for CMode CLI/ZAPI
  Maps to:
    CM ZAPI: role-query

=item C<< "access" => $string >>

  Applicable for CMode CLI/ZAPI
  Maps to:
    CM ZAPI: access-level

=item C<< 'extended_query' => $string >>

(Optional)
  Applicable only for CMode CLI

=item C<< '*' >>

This method takes all other options supported by
"security login role create" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub create {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( query access role
                    )
            ),
        },
        cli_cmd => 'security login role create',
    );
    my $ret = $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_create_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_create_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
            {   method    => '_create_7mode_cli',
                interface => 'CLI',
                set       => '7Mode',
            },
        ],
        frontend => 'NACL::C::SecurityLoginRole::create',
    );
    $Log->exit() if $may_exit;
    return $ret;
} ## end sub create

sub _create_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, %opts) = @_;
    my $vserver = $opts{vserver};
    $opts{query} = '"' . $opts{query} . '"'
        if ((defined($opts{query}))
        && ($opts{query} !~ /^".*"$/));
    $opts{cmddirname} = '"' . $opts{cmddirname} . '"'
        if ((defined($opts{cmddirname}))
        && ($opts{cmddirname} !~ /^".*"$/));

    $pkg->_base_backend_cmode_cli(
        %opts,
        api => 'security_login_role_create',
        vserver_operated_on => \$vserver,
    );

    $Log->exit() if $may_exit;
    return $pkg->new(
        command_interface => $opts{command_interface},
        vserver           => $opts{vserver},
        cmddirname        => $opts{'cmddirname'},
        role              => $opts{'role'},
    );
} ## end sub _create_cmode_cli

sub _create_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my %cmode_args;
    $pkg->_hash_copy(
        source => \%opts,
        map    => {
            "role"       => "role-name",
            "cmddirname" => "capability",
        },
        target => \%cmode_args,
    );
    %opts = %cmode_args;

    if ($cmode_args{'capability'} =~ /DEFAULT/i) {
        $cmode_args{'capability'} = "*";
    }
    $apiset->useradmin_role_add(%cmode_args);
    $Log->exit() if $may_exit;
    return $pkg->new(
        command_interface => $command_interface,
        vserver           => $opts{'vserver'},
        cmddirname        => $opts{'cmddirname'},
        role              => $opts{'role-name'},
    );
} ## end sub _create_7mode_cli

sub _create_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg, %opts) = @_;
    my $allowed_cmode_options = [qw(vserver )];

    $pkg->_base_backend_cmode_zapi(
        %opts,
        api  => 'security_login_role_create',
        copy   => $allowed_cmode_options,
	map    => {
            "role"       => "role-name",
            "cmddirname" => "command-directory-name",
            "access"     => "access-level",
            "query"      => "role-query",
        },
    );
    $Log->exit() if $may_exit;
    return $pkg->new(
        command_interface => $opts{'command_interface'},
        vserver           => $opts{'vserver'},
        cmddirname        => $opts{'cmddirname'},
        role              => $opts{'role'},
    );
} ## end sub _create_cmode_zapi

=head2 show_user_capability

    my $show_user_capability_obj = NACL::C::SecurityLoginRole->show_user_capability(command_interface => $ci,
                          [ filter => { %filter } ,]
              [ requested_fields => [ @req_fields ]);

                    or

    my @show_user_capability_objs = NACL::C::SecurityLoginRole->show_user_capability(command_interface => $ci,
                          [ filter => { %filter } ,]
              [ requested_fields => [ @req_fields ]);

                    or

   my $show_user_capability_obj = $SecurityLoginRole_obj->show_user_capability();

(Class or Instance method)

This method is a CS-like fetch method which accepts all the parameters of
L<fetch|lib-NACL-CS-ComponentState-pm/fetch> (requested_fields, filter, allow_empty)
and has all of the other properties of fetch methods (callable in scalar or list
context, throws a NoElementsFound exception if no matching elements are found).
The requested fields and filter takes L<ATTRIBUTES|lib-NACL-CS-SecurityLoginRoleShowUserCapability-pm/ATTRIBUTES>
of NACL::CS::SecurityLoginRoleShowUserCapability as input.
This method returns objects of the type L<NACL::CS::SecurityLoginRoleShowUserCapability|lib-NACL-CS-SecurityLoginRoleShowUserCapability-pm>.

=cut

sub show_user_capability {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = $pkg_or_obj->_handle_obj_cs_like_calls(@_);
    require NACL::CS::SecurityLoginRoleShowUserCapability;
    my @state_objs
        = NACL::CS::SecurityLoginRoleShowUserCapability->fetch(%opts);
    $Log->exit() if $may_exit;
    return wantarray ? @state_objs : $state_objs[0];
}

=head2 show_ontapi

    my $show_ontapi_obj = NACL::C::SecurityLoginRole->show_ontapi(command_interface => $ci,
                          [ filter => { %filter } ,]
			  [ requested_fields => [ @req_fields ]);
						  
                    or
					
    my @show_ontapi_objs = NACL::C::SecurityLoginRole->show_ontapi(command_interface => $ci,
                          [ filter => { %filter } ,]
			  [ requested_fields => [ @req_fields ]);

                    or
  
   my $show_ontapi_obj = $SecurityLoginRole_obj->show_ontapi(); 

(Class or Instance method)

This method is a CS-like fetch method which accepts all the parameters of 
L<fetch|lib-NACL-CS-ComponentState-pm/fetch> (requested_fields, filter, allow_empty) 
and has all of the other properties of fetch methods (callable in scalar or list 
context, throws a NoElementsFound exception if no matching elements are found).
The requested fields and filter takes L<ATTRIBUTES|lib-NACL-CS-SecurityLoginRoleShowOntapi-pm/ATTRIBUTES>
of NACL::CS::SecurityLoginRoleShowOntapi as input. 
This method returns objects of the type L<NACL::CS::SecurityLoginRoleShowOntapi|lib-NACL-CS-SecurityLoginRoleShowOntapi-pm>.

=cut

sub show_ontapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts = $pkg_or_obj->_handle_obj_cs_like_calls(@args);
    require NACL::CS::SecurityLoginRoleShowOntapi;
    my @state_objs = NACL::CS::SecurityLoginRoleShowOntapi->fetch(%opts);
    $Log->exit() if $may_exit;
    return wantarray ? @state_objs : $state_objs[0];
}


__PACKAGE__->_load_event_class();

sub _primary_keys_validate_spec {
    return (
        'vserver'    => {type => SCALAR | UNDEF, optional => 1},
        'cmddirname' => {type => SCALAR | UNDEF, optional => 1},
        'role'       => {type => SCALAR},
    );
} ## end sub _primary_keys_validate_spec
1;
