# $Id: //depot/prod/test/nacldev/lib/NACL/C/Smtape.pm#15 $

#
# Copyright (c) 2012 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Smtape Component Module
## @author dl-dpg-nb-automation; dl-nacl-dev@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::C::Smtape

=head1 DESCRIPTION

C<NACL::C::Smtape> is a derived class of
L<NACL::C::Component::ONTAP|lib-NACL-C-Component-ONTAP-pm>.
It represents Smtape service on a filer.

=cut

package NACL::C::Smtape;

use strict;
use warnings;

use NACL::CS::Smtape;
use NACL::CS::SmtapeRestoreGetImageHeader;
use NACL::CS::SmtapeRestoreGetVolumeGeometry;
use NACL::ComponentUtils qw (_optional_scalars);

use base 'NACL::C::Component::ONTAP';
use Params::Validate qw(validate validate_with SCALAR ARRAYREF);
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Data::Dumper;

=head1 METHODS

=head2 new

    my $smtape_object = NACL::C::Smtape->new(
                                   command_interface=>$ci,
                                   'job-id'=>$job_id,
                              );

Return a new Smtape component object with the given attributes.

=cut

use Class::MethodMaker [ scalar => 'job_id', ];

=head2 backup

   my $smtape_object = NACL::C::Smtape->backup(
                   command_interface => $ci,
                   'path'            => $path,
                   'tape-device'     => $tape_device,
                  ...);

(Class method)
This method is used to start a SnapMirror to Tape equivalent backup operation.
Does not wait for backup operation to complete.

Uses a 7Mode, Nodescope and CLI APISet.
  Invokes 'smtape backup' command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "path" => $path >>

(Required) Volume path to be backed up
  Applicable only for 7Mode CLI.

=item C<< "tape-device" => $tape_device >>

(Required) Tape device to which volume will be backed up
  Applicable only for 7Mode CLI.

=item C<< "volume-geometry" => $volume_geometry >>

(Optional) Geometry of the traditional volume of the destination
  Applicable only for 7Mode CLI.

=item C<< "block-size" => $block_size >>

(Optional) Tape record size in multiples of 4k-bytes (from 4KB to 256KB, default to 240KB)
  Applicable only for 7Mode CLI.

=item C<< "snapshot-name" => $snapshot_name >>

(Optional) New base snapshot name
  Applicable only for 7Mode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub backup {
    $Log->enter() if $may_enter;
    my ( $pkg, @args ) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            path          => { type => SCALAR },
            'tape-device' => { type => SCALAR },
            _optional_scalars(
                qw( node volume-geometry block-size snapshot-name
                    )
            ),
        },
        ignore_primary_keys => 1,
    );
    my $ret = $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_backup_7mode_cli',
                interface => 'CLI',
                set       => '7Mode|Nodescope',
            },
        ],
    );

    $Log->exit() if $may_exit;
    return $ret;
}

sub _backup_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );

    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};

    my $response      = $apiset->smtape_backup(%opts);
    my $parsed_output = $response->get_parsed_output();
    $Log->exit() if $may_exit;
    return $pkg->new(
        job_id            => $parsed_output->[0]{'job_id'},
        command_interface => $command_interface
    );
}

=head2 restore

   my $smtape_object = NACL::C::Smtape->restore(
                   command_interface => $ci,
                   'path'            => $path,
                   'tape-device'     => $tape_device,
                  ...);

(Class method)
This method is used to start a SnapMirror to Tape equivalent restore operation.
Does not wait for restore operation to complete.
Image Header and Volume Geometry cannot be obtained with this method. The user needs to
call NACL::C::Smtape->restore_get_image_header() and
NACL::C::Smtape->restore_get_volume_geometry() respectively, for the same.

Uses a 7Mode, Nodescope and CLI APISet.
  Invokes 'smtape restore' command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "path" => $path >>

(Required) Volume path to restore
  Applicable only for 7Mode CLI.

=item C<< "tape-device" => $tape_device >>

(Required) Tape device from which volume will be restored
  Applicable only for 7Mode CLI.

=item C<< "block-size" => $block_size >>

(Optional) Tape record size in multiples of 4k-bytes (from 4KB to 256KB, default to 240KB)
  Applicable only for 7Mode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub restore {
    $Log->enter() if $may_enter;
    my ( $pkg, @args ) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            path          => { type => SCALAR },
            'tape-device' => { type => SCALAR },
            _optional_scalars(qw( node block-size )),
        },
        ignore_primary_keys => 1,
    );
    my $ret = $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_restore_7mode_cli',
                interface => 'CLI',
                set       => '7Mode|Nodescope',
            },
        ],
    );

    $Log->exit() if $may_exit;
    return $ret;
}

sub _restore_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );

    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};

    my $response      = $apiset->smtape_restore(%opts);
    my $parsed_output = $response->get_parsed_output();
    $Log->exit() if $may_exit;
    return $pkg->new(
        job_id            => $parsed_output->[0]{'job_id'},
        command_interface => $command_interface
    );
}

=head2 restore_get_image_header

   my $smtape_restore_get_image_header_object =
            NACL::C::Smtape->restore_get_image_header(
                   command_interface => $ci,
                   'backup-image'    => $tape_name
            ...);

(Class method)
This method is used to list the contents of a backup image header.
For this purpose, use this method instead of NACL::C::Smtape->restore because
the latter is used only for performing an Smtape restore operation.

This method returns an object of type NACL::CS::SmtapeRestoreGetImageHeader.

Uses a 7Mode, Nodescope and CLI APISet.
  Invokes 'smtape restore' command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "backup-image" => $tape_name >>

(Required) Tape device for which image header will be retrieved
  Applicable only for 7Mode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub restore_get_image_header {
    $Log->enter() if $may_enter;
    my ( $pkg, @args ) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params              => \@args,
        additional_spec     => { 'backup-image' => { type => SCALAR }, },
        ignore_primary_keys => 1,
    );
    my $image_header_obj = $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_restore_get_image_header_7mode_cli',
                interface => 'CLI',
                set       => '7Mode|Nodescope',
            },
        ],
    );

    $Log->exit() if $may_exit;
    return $image_header_obj;
}

sub _restore_get_image_header_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );

    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};

    my $response      = $apiset->smtape_restore(%opts);
    my $parsed_output = $response->get_parsed_output();
    $Log->exit() if $may_exit;

    my $backup_sequence_no
        = delete $parsed_output->[0]{'backup_sequence_no.'};
    $parsed_output->[0]{'backup_sequence_no'} = $backup_sequence_no;
    my $image_header_obj = NACL::CS::SmtapeRestoreGetImageHeader->new(
        %{ $parsed_output->[0] } );
}

=head2 restore_get_volume_geometry

   my $smtape_restore_get_volume_geometry_object =
            NACL::C::Smtape->restore_get_volume_geometry(
                   command_interface => $ci,
                   'volume-path'     => $volume_path
            ...);

(Class method)
This method is used to list the volume geometry of the specified path.
For this purpose, use this method instead of NACL::C::Smtape->restore because
the latter is used only for performing an Smtape restore operation.

This method returns an object of type NACL::CS::SmtapeRestoreGetVolumeGeometry.

Uses a 7Mode, Nodescope and CLI APISet.
  Invokes 'smtape restore' command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "volume-path" => $volume_path >>

(Required) Path for which volume geometry will be retrieved
  Applicable only for 7Mode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub restore_get_volume_geometry {
    $Log->enter() if $may_enter;
    my ( $pkg, @args ) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params              => \@args,
        additional_spec     => { 'volume-path' => { type => SCALAR }, },
        ignore_primary_keys => 1,
    );
    my $volume_geometry_obj = $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_restore_get_volume_geometry_7mode_cli',
                interface => 'CLI',
                set       => '7Mode|Nodescope',
            },
        ],
    );

    $Log->exit() if $may_exit;
    return $volume_geometry_obj;
}

sub _restore_get_volume_geometry_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );

    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};

    my $response      = $apiset->smtape_restore(%opts);
    my $parsed_output = $response->get_parsed_output();
    $Log->exit() if $may_exit;

    return NACL::CS::SmtapeRestoreGetVolumeGeometry->new(
        %{ $parsed_output->[0] } );
}

=head2 abort

   NACL::C::smtape->abort(
                   command_interface => $ci,
                   'job-id'=>$job_id,
                  ...);
    or
   $smtape_object->abort(...);

(Class or instance method)
This method is used to abort the specified Job.

Uses a 7Mode, Nodescope and CLI APISet.
  Invokes 'smtape abort' command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "job-id" => $job_id >>

Required for class method, Not Applicable for instance method
  Applicable only for 7Mode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub abort {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with( params => \@args, );
    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_abort_7mode_cli',
                interface => 'CLI',
                set       => '7Mode|Nodescope',
            },
        ],
    );

    $Log->exit() if $may_exit;
    return $ret;
}

sub _abort_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );

    delete $opts{command_interface};
    my $apiset = delete $opts{apiset};

    $apiset->smtape_abort(%opts);
    $Log->exit() if $may_exit;
}

=head2 continue

   NACL::C::smtape->continue(
                   command_interface => $ci,
                   'job-id'=>$job_id,
                  ...);
    or
   $smtape_object->continue(...);

(Class or instance method)
This method is used to continue the specified Job after tape drive is full.

Uses a 7Mode, Nodescope and CLI APISet.
  Invokes 'smtape continue' command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "job-id" => $job_id >>

Required for class method, Not Applicable for instance method
  Applicable only for 7Mode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub continue {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( path
                    )
            ),
        },
    );
    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_continue_7mode_cli',
                interface => 'CLI',
                set       => '7Mode|Nodescope',
            },
        ],
    );

    $Log->exit() if $may_exit;
    return $ret;
}

sub _continue_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );

    delete $opts{command_interface};
    my $apiset = delete $opts{apiset};

    $apiset->smtape_continue(%opts);
    $Log->exit() if $may_exit;
}

=head2 status

    my $status_obj = NACL::C::Smtape->status(command_interface => $ci,
                          [ filter => { %filter } ,]
              [ requested_fields => [ @req_fields ]);

                    or

    my @status_objs = NACL::C::Smtape->status(command_interface => $ci,
                          [ filter => { %filter } ,]
              [ requested_fields => [ @req_fields ]);

                    or

   my $status_obj = $Smtape_obj->status();

(Class or Instance method)
This method returns objects of the type L<NACL::CS::SmtapeStatus|lib-NACL-CS-SmtapeStatus-pm>.

This method accepts all the parameters accepted by L<fetch|lib-NACL-CS-ComponentState-pm/fetch>
(requested_fields, filter, allow_empty) and has all of the other properties of
fetch methods (callable in scalar or list context, throws a NoElementsFound
exception if no matching elements are found)

=cut

sub status {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = $pkg_or_obj->_handle_obj_cs_like_calls(@_);
    require NACL::CS::SmtapeStatus;
    my @state_objs = NACL::CS::SmtapeStatus->fetch(%opts);
    $Log->exit() if $may_exit;
    return wantarray ? @state_objs : $state_objs[0];
}

=head2 break

    NACL::C::Smtape->break(
                                command_interface => $ci,
                                ...);

(Class method)
This method is used to break a restore relationship between a tape and a volume.

Uses a CMode CLI APISet.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'volume' => $string >>

  Applicable only for CMode CLI.

=item C<< 'vserver' => $string >>

  Applicable only for CMode CLI.

=item C<< 'method-timeout' => $time_in_seconds >>

(Optional)
The default NATE timeout is 60 seconds. If the command might take more
time to complete, this option can be used to specify a larger timeout value.
The value should be provided in seconds.

=item C<< "extended_query" => $string >>

(Optional)
  Applicable only for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub break {
    $Log->enter() if $may_enter;
    my $pkg = shift;
    $pkg->_frontend_log_debug_opts(@_);
    my %opts = $pkg->_common_validate_with(
        params          => \@_,
        additional_spec => {
            _optional_scalars(
                qw( volume vserver
                    )
            ),
        },
        ignore_primary_keys => 1,
    );
    $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_break_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
        ],
    );

    $Log->exit() if $may_exit;
}

sub _break_cmode_cli {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = $pkg->_validate_backend_opts(@_);

    delete $opts{command_interface};
    my $apiset = delete $opts{apiset};

    my $allowed_cmode_options = [
        qw( volume vserver
            extended_query )
    ];
    my %cmode_args;
    $pkg->_hash_copy(
        source => \%opts,
        copy   => $allowed_cmode_options,
        target => \%cmode_args,
    );

    $pkg->_api_call_log_debug_opts(
        api      => 'smtape_break',
        api_opts => \%cmode_args
    );

    $apiset->smtape_break(%cmode_args);
    $Log->exit() if $may_exit;
}

=head2 show_header

    NACL::C::Smtape->show_header(
                                command_interface => $ci,
                                ...);

(Class method)
This method is used to displays the SMTape header from a valid tape.

Uses a CMode CLI APISet.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'tape-block-size' => $string >>

  Applicable only for CMode CLI.

=item C<< 'tape' => $string >>

  Applicable only for CMode CLI.

=item C<< 'method-timeout' => $time_in_seconds >>

(Optional)
The default NATE timeout is 60 seconds. If the command might take more
time to complete, this option can be used to specify a larger timeout value.
The value should be provided in seconds.

=item C<< "extended_query" => $string >>

(Optional)
  Applicable only for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub show_header {
    $Log->enter() if $may_enter;
    my $pkg = shift;
    $pkg->_frontend_log_debug_opts(@_);
    my %opts = $pkg->_common_validate_with(
        params          => \@_,
        additional_spec => {
            _optional_scalars(
                qw( tape-block-size tape
                    )
            ),
        },
        ignore_primary_keys => 1,
    );
    $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_show_header_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
        ],
    );

    $Log->exit() if $may_exit;
}

sub _show_header_cmode_cli {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = $pkg->_validate_backend_opts(@_);

    delete $opts{command_interface};
    my $apiset = delete $opts{apiset};

    my $allowed_cmode_options = [
        qw( tape-block-size tape
            extended_query )
    ];
    my %cmode_args;
    $pkg->_hash_copy(
        source => \%opts,
        copy   => $allowed_cmode_options,
        target => \%cmode_args,
    );

    $pkg->_api_call_log_debug_opts(
        api      => 'smtape_show_header',
        api_opts => \%cmode_args
    );

    $apiset->smtape_show_header(%cmode_args);
    $Log->exit() if $may_exit;
}


__PACKAGE__->_load_event_class();

sub _primary_keys_validate_spec {
    return ( 'job-id' => { type => SCALAR } );
}

1;
