# $Id$
#
# Copyright (c) 2011 NetApp, Inc., All Rights Reserved
# Any use, modification, or distribution is prohibited
# without prior written consent from NetApp, Inc.
#

## @summary snapvault Component Module
## @author dl-dpg-nb-automation, dl-nacl-dev;
## @status shared
## @pod here

=head1 NAME

NACL::C::Snapvault

=head1 DESCRIPTION

C<NACL::C::Snapvault> is a derived class of
L<NACL::C::Component::ONTAP|lib-NACL-C-Component-ONTAP-pm>.
It represents snapvault command on a filer.

=cut

package NACL::C::Snapvault;

use strict;
use warnings;
use NACL::ComponentUtils qw (Dumper _optional_scalars);
use NACL::CS::Snapvault;
use base 'NACL::C::Component::ONTAP';
use Params::Validate qw(validate validate_with SCALAR ARRAYREF UNDEF);
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
=head1 METHODS

=head2 new

    my $session = NACL::C::Snapvault->new(
                                   command_interface=>$ci,
                                   'secondary-path'=>$secondary_path,
                              );

Return a new Snapvault component object with the given attributes.

=cut

use Class::MethodMaker [
    scalar => 'secondary_path',
];

=head2 start

   NACL::C::snapvault->start(
                   command_interface => $ci,
                   'primary-location' => $primary-location,
                   'secondary-path'   => $secondary-path
                  ...);

                OR

    $Destination_object->start('primary-location' => $primary-location,);
(Class method or Instance Method)
This method is used to start the baseline transfer.

Uses a 7Mode CLI APIset.

Invokes "snapvault start" command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "primary-location" => $primary_location >>

(Required for class method) primary location  path

Applicable for 7Mode CLI.

Maps to:

7M CLI : -S <value> (Programmatic name: "primary-location")

=item C<< "secondary-path" => $secondary_path >>

(Required for class method) secondary location path

Applicable for 7Mode CLI.

Maps to:

7M CLI : secondary path (Programmatic name: "secondary-path")

=item C<< "restart-backup" => $scalar >>

(Optional)restart updates from a different primary path.

Applicable for 7Mode CLI.

Maps to:

7M CLI : -r (Programmatic name: "restart-backup")

=item C<< "try-count" => $scalar >>

(Optional)sets the number of times that updates for the qtree should be tried before giving up.

Applicable for 7Mode CLI.

Maps to:

7M CLI : -t <value> (Programmatic name: "try-count")

=item C<< "max-transfer-rate" => $scalar >>

(Optional)maximum speed at which data is transferred in kilobytes per second.

Applicable for 7Mode CLI.

Maps to:

7M CLI : -k <value> (Programmatic name: "max-transfer-rate")

=item C<< "back-up-open-files" => $Boolean >>

(Optional) on OR  off

Applicable for 7Mode CLI.

=item C<< "compression" => $Boolean >>

(Optional) on OR  off

Applicable for 7Mode CLI.

=item C<< "gnore-atime" => $Boolean >>

(Optional) on OR  off

Applicable for 7Mode CLI.

=item C<< "wait" => $Boolean >>

(Optional)option causes the command not to return once the baseline transfer starts..

Applicable for 7Mode CLI.

Maps to:

7M CLI : -w (Programmatic name: "wait")

=item C<< "force" => $Boolean >>

(Optional)forceful start.

Applicable for 7Mode CLI.

Maps to:

7M CLI : -f (Programmatic name: "wait")

=item C<< "method-timeout" => $scalar >>

(Optional) timeout in seconds.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut


sub start {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars qw(
            restart-backup force try-count max-transfer-rate
            options wait primary-location method-timeout
            back-up-open-files ignore-atime compression
            ),
        },
    );
    # for ZAPI support, we need to split "options" paramter ( because in ZAPI, each option is taken as a separate parameter)
    # What i mean to say is, instead of options => "name=value,name2=value2" pass by user in the script,
    # user will now pass 'back-up-open-files' => 'on', compression => 'on'.
    # To access options using APILayer we have to pass "options => "name=value,name2=value2"".
    # To access component layer, we need to pass 'back-up-open-files' => 'on', compression => 'on'.
    my ($options_exist, $constructed_options_string);
    foreach my $option (qw(back-up-open-files ignore-atime compression)) {
        if (exists $opts{$option}) {
            $options_exist = 1;
            my $value = delete $opts{$option}; # Needs to be deleted so it's not sent to the API call
            my $underscored_form = $option;
            # Converting hyphenated  to underscore  for (back-up-open-files ignore-atime compression)
            $underscored_form =~ s/-/_/g;
            # constructing the string of the format -'back-up-open-files' => 'on', compression => 'on'
            $constructed_options_string .= "$underscored_form=$value,";
        }
    }
    # constructing -o part of the command
    # %opts is sent to the API call, the CDEF defines the mapping as "options" => "-o"
    if ($options_exist) {
        chop($constructed_options_string); # Remove the trailing ,
        $opts{options} = $constructed_options_string;
    }
    my $ret = $pkg->call_on_apiset(
          %opts,
          choices => [
              {
                  method    => '_start_7mode_cli',
                  interface => 'CLI',
                  set       => '7Mode',
              },
          ],
    );
    $Log->exit() if $may_exit;
    return $ret;
}
sub _start_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );
    delete $opts{command_interface};
    my $apiset = delete $opts{apiset};
    $pkg->_handle_boolean_options(
            api_options => \%opts,
            boolean_options => [ qw(force wait) ],
            is_7m_cli       => 1
    );
    $pkg->_hash_move(
          source => \%opts,
          target => \%opts,
          map => { 'method-timeout' => 'connectrec-timeout' }
    );
    $apiset->snapvault_start(%opts);
    $Log->exit() if $may_exit;
}
=head2 update

   NACL::C::snapvault->udpate(
                   command_interface => $ci,
                   'secondary-path'   => $secondary-path
                  ...);

                OR

    $Destination_object->udpate('primary-location' => $primary-location,);
(Class method or Instance Method)
This method is used start an update of the specified qtree on the secondary.

Uses a 7Mode CLI APIset.

Invokes "snapvault update" command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "secondary-path" => $secondary_path >>

(Required for class method) secondary location path

Applicable for 7Mode CLI.

=item C<< "max-transfer-rate" => $scalar >>

(Optional)maximum speed at which data is transferred in kilobytes per second.

Applicable for 7Mode CLI.

Maps to:

7M CLI: -k <value> (programmatic name: max-transfer-rate)

=item C<< "wait" => $Boolean >>

(Optional)option causes the command not to return once the incremental transfer starts..

Applicable for 7Mode CLI.

Maps to:

7M CLI: -w (programmatic name: wait)

=item C<< "method-timeout" => $scalar >>

(Optional) timeout in seconds.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub update {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars qw(
            max-transfer-rate wait method-timeout
            ),
        },
    );
    my $ret = $pkg->call_on_apiset(
          %opts,
            choices => [
                {
                  method    => '_update_7mode_cli',
                  interface => 'CLI',
                  set       => '7Mode',
                },
            ],
    );
    $Log->exit() if $may_exit;
    return $ret;
}

sub _update_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg = shift;
    my %opts        = validate_with(
        params      => \@_,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );
    delete $opts{command_interface};
    my $apiset = delete $opts{apiset};
    $pkg->_handle_boolean_options(
            api_options     => \%opts,
            boolean_options => [ qw(wait) ],
            is_7m_cli       => 1
    );
    $pkg->_hash_move(
          source    => \%opts,
          target    => \%opts,
          map       => { 'method-timeout' => 'connectrec-timeout' }
    );
    $apiset->snapvault_update(%opts);
    $Log->exit() if $may_exit;
}
=head2 modify

    NACL::C::snapvault->modify(
                   command_interface => $ci,
                   'secondary-path'   => $secondary-path
                  ...);

(Class method or Instance Method)
This method is used to modify specified attributes on the secondary.

Uses a 7Mode CLI APIset.

Invokes "snapvault modify" command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "secondary-path" => $secondary_path >>

(Required for class method) secondary location path

Applicable for 7Mode CLI.

Maps to:

7M CLI: secondary path (programmatic name: secondary-path)

=item C<< "primary-location" => $secondary_path >>

(Required for class method) primary location path

Applicable for 7Mode CLI.

Maps to:

7M CLI: -S <value> (programmatic name: primary-location)

=item C<< "max-transfer-rate" => $scalar >>

(Optional)maximum speed at which data is transferred in kilobytes per second.

Applicable for 7Mode CLI.

Maps to:

7M CLI: -k <value> (programmatic name: max-transfer-rate)

=item C<< "transfer" => $scalar >>

(Optional)option causes the command to try transfer $tries amount of time..

Applicable for 7Mode CLI.

Maps to:

7M CLI: -t <value> (programmatic name: transfer)

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub modify {
    $Log->enter() if $may_enter;
    my $pkg = shift;
    $pkg->_frontend_log_debug_opts(@_);
    my %opts = $pkg->_common_validate_with(
        params          => \@_,
        additional_spec => {
                            _optional_scalars qw(
                            try-count max-transfer-rate
                            options primary-location
                            back-up-open-files transfer
            ),
        },
    );
     my $ret = $pkg->call_on_apiset(
          %opts,
          choices => [
                {
                  method    => '_modify_7mode_cli',
                  interface => 'CLI',
                  set       => '7Mode',
                },
          ],
    );
    $Log->exit() if $may_exit;
    return $ret;
}
sub _modify_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );
    delete $opts{command_interface};
    my $apiset = delete $opts{apiset};

    $apiset->snapvault_modify(%opts);
    $Log->exit() if $may_exit;
}
=head2 stop

    NACL::C::snapvault->stop(
                   command_interface => $ci,
                   'secondary-path'   => $secondary-path
                  ...);

(Class method or Instance Method)
This method is used to stop snapvault transfer.

Uses a 7Mode CLI APIset.

Invokes "snapvault stop" command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "secondary-path" => $secondary_path >>

(Required for class method) secondary location path

Applicable for 7Mode CLI.

Maps to:

7M CLI: secondary path (programmatic name: secondary-path)

=item C<< "force" => $Boolean >>

(Optional)option causes the command to execute forcefully..

Applicable for 7Mode CLI.

Maps to:

7M CLI: -f (programmatic name: force)

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub stop {
    $Log->enter() if $may_enter;
    my $pkg = shift;
    $pkg->_frontend_log_debug_opts(@_);
    my %opts = $pkg->_common_validate_with(
        params          => \@_,
        additional_spec => {
            _optional_scalars qw(
            max-transfer-rate force
            ),
        },
    );
    my $ret = $pkg->call_on_apiset(
          %opts,
          choices => [
              {
                  method    => '_stop_7mode_cli',
                  interface => 'CLI',
                  set       => '7Mode',
              },
          ],
    );
    $Log->exit() if $may_exit;
    return $ret;
}
sub _stop_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );
    delete $opts{command_interface};
    my $apiset = delete $opts{apiset};
    my @prompts_answers = ('.*Are you sure you want to do this.*'=> 'y');
    $pkg->_handle_boolean_options(
            api_options => \%opts,
            boolean_options => [ qw(force) ],
            is_7m_cli       => 1
    );
    $opts{'connectrec-match_table'} = \@prompts_answers;
    $pkg->_api_call_log_debug_opts(api      => 'snapvault stop',
                                   api_opts => \%opts);

    $apiset->snapvault_stop(%opts);
    $Log->exit() if $may_exit;
}
=head2 prerevert

    NACL::C::snapvault->prerevert(
                   command_interface => $ci,
                   'secondary-path'   => $secondary-path
                  ...);

(Class method or Instance Method)
This method is used to prerevert the secondary.

Uses a 7Mode CLI APIset.

Invokes "snapvault prerevert" command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "secondary-path" => $secondary_path >>

(Required for class method) secondary location path

Applicable for 7Mode CLI.

Maps to:

7M CLI: secondary path (programmatic name: secondary-path)

=item C<< "force" => $Boolean >>

(Optional)option causes the command to execute forcefully..

Applicable for 7Mode CLI.

Maps to:

7M CLI: -f (programmatic name: force)

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub prerevert {
    $Log->enter() if $may_enter;
    my $pkg = shift;
    $pkg->_frontend_log_debug_opts(@_);
    my %opts = $pkg->_common_validate_with(
        params          => \@_,
        additional_spec => {
            _optional_scalars qw(
             force restart-backup
            ),
        },
    );
    my $ret = $pkg->call_on_apiset(
          %opts,
          choices => [
                {
                  method    => '_prerevert_7mode_cli',
                  interface => 'CLI',
                  set       => '7Mode',
                },
          ],
    );
    $Log->exit() if $may_exit;
    return $ret;
}
sub _prerevert_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );
    delete $opts{command_interface};
    my $apiset = delete $opts{apiset};
    my @prompts_answers = ('.*Are you REALLY sure you want to continue.*'=> 'y');
    delete $opts{'secondary-path'};
    $pkg->_handle_boolean_options(
            api_options     => \%opts,
            boolean_options => [ qw(force) ],
            is_7m_cli       => 1
    );
    $opts{'connectrec-match_table'} = \@prompts_answers;

    $apiset->snapvault_prerevert(%opts);
    $Log->exit() if $may_exit;
}
=head2 check_start

    NACL::C::snapvault->check_start(
                   command_interface => $ci,
                   'secondary-path'   => $secondary-path
                  ...);

(Class method or Instance Method)
This method is used to check start the snapvault transfer.

Uses a 7Mode CLI APIset.

Invokes "snapvault check start" command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "secondary-path" => $secondary_path >>

(Required for class method) secondary location path

Applicable for 7Mode CLI.

Maps to:

7M CLI: secondary path (programmatic name: secondary-path)

=item C<< "max-transfer-rate" => $scalar >>

(Optional)maximum speed at which data is transferred in kilobytes per second.

Applicable for 7Mode CLI.

Maps to:

7M CLI: -k <value> (programmatic name: max-transfer-rate)

=item C<< "wait" => $Boolean >>

(Optional)option causes the command not to return once the baseline transfer starts..

Applicable for 7Mode CLI.

Maps to:

7M CLI: -w (programmatic name: wait)

=item C<< "force" => $Boolean >>

(Optional)option causes the command to execute forcefully..

Applicable for 7Mode CLI.

Maps to:

7M CLI: -f (programmatic name: force)

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub check_start {
    $Log->enter() if $may_enter;
    my $pkg = shift;
    $pkg->_frontend_log_debug_opts(@_);
    my %opts = $pkg->_common_validate_with(
        params          => \@_,
        additional_spec => {
            _optional_scalars qw(
            max-transfer-rate force wait primary-location
            ),
        },
    );
    my $ret = $pkg->call_on_apiset(
          %opts,
          choices => [
                {
                  method    => '_check_start_7mode_cli',
                  interface => 'CLI',
                  set       => '7Mode',
                },
          ],
    );
    $Log->exit() if $may_exit;
    return $ret;
}
sub _check_start_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );
    delete $opts{command_interface};
    my $apiset = delete $opts{apiset};
    $pkg->_api_call_log_debug_opts(api      => 'snapvault check start',
                                   api_opts => \%opts);
    $apiset->snapvault_check_start(%opts);
    $Log->exit() if $may_exit;
}

=head2 abort

    NACL::C::snapvault->abort(
                   command_interface => $ci,
                   'secondary-path'   => $secondary-path
                  ...);

(Class method or Instance Method)
This method is used to abort the transfer.

Uses a 7Mode CLI APIset.

Invokes "snapvault abort" command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "secondary-path" => $secondary_path >>

(Required for class method) secondary location path

Applicable for 7Mode CLI.

Maps to:

7M CLI: secondary path (programmatic name: secondary-path)

=item C<< "force" => $Boolean >>

(Optional)option causes the command to execute forcefully..

Applicable for 7Mode CLI.

Maps to:

7M CLI: -f (programmatic name: force)

=item C<< "clear-checkpoint" => $Boolean >>

Applicable for 7Mode CLI.

Maps to:

7M CLI: -h (programmatic name: clear-checkpoint)

=item C<< "snapshot" => $string >>

Applicable for 7Mode CLI.

Maps to:

7M CLI: -s (programmatic name: snapshot)

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut


sub abort {
    $Log->enter() if $may_enter;
    my $pkg = shift;
    $pkg->_frontend_log_debug_opts(@_);
    my %opts = $pkg->_common_validate_with(
        params          => \@_,
        additional_spec => {
            _optional_scalars qw(
            force clear-checkpoint snapshot secondary-path
            ),
        },
    );
    my $ret = $pkg->call_on_apiset(
          %opts,
          choices => [
                {
                  method    => '_abort_7mode_cli',
                  interface => 'CLI',
                  set       => '7Mode',
                },
          ],
    );
    $Log->exit() if $may_exit;
    return $ret;
}

sub _abort_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );
    delete $opts{command_interface};
    my $apiset = delete $opts{apiset};
    $pkg->_handle_boolean_options(
            api_options     => \%opts,
            boolean_options => [ qw(force clear-checkpoint) ],
            is_7m_cli       => 1
    );

    $apiset->snapvault_abort(%opts);
    $Log->exit() if $may_exit;
}
=head2 check_abort

    NACL::C::snapvault->check_abort(
                   command_interface => $ci,
                   'secondary-path'   => $secondary-path
                  ...);

(Class method or Instance Method)
This method is used to check_abort the transfer.

Uses a 7Mode CLI APIset.

Invokes "snapvault check abort" command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "secondary-path" => $secondary_path >>

(Required for class method) secondary location path

Applicable for 7Mode CLI.

Maps to:

7M CLI: secondary path (programmatic name: secondary-path)

=item C<< "max-transfer-rate" => $scalar >>

Applicable for 7Mode CLI.

Maps to:

7M CLI: -k <value> (programmatic name: max-transfer-rate)

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub check_abort {
    $Log->enter() if $may_enter;
    my $pkg = shift;
    $pkg->_frontend_log_debug_opts(@_);
    my %opts = $pkg->_common_validate_with(
        params          => \@_,
        additional_spec => {
                _optional_scalars qw(
                max-transfer-rate force method-timeout
                ),
        },
    );
    my $ret = $pkg->call_on_apiset(
                %opts,
                choices => [
                {
                    method    => '_check_abort_7mode_cli',
                    interface => 'CLI',
                    set       => '7Mode',
                },
                ],
    );
    $Log->exit() if $may_exit;
    return $ret;
}
sub _check_abort_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg = shift;
    my %opts       = validate_with(
            params      => \@_,
            spec        => $pkg->_backend_validate_spec(),
            allow_extra => 1
    );
    delete $opts{command_interface};
    my $apiset = delete $opts{apiset};
    $pkg->_hash_move(
            source  => \%opts,
            target  => \%opts,
            map     => { 'method-timeout' => 'connectrec-timeout' }
    );

    $apiset->snapvault_check_abort(%opts);
    $Log->exit() if $may_exit;
}
=head2 convert

    NACL::C::snapvault->convert(
                   command_interface => $ci,
                   'secondary-path'   => $secondary-path
                  ...);

(Class method or Instance Method)
This method is used to convert the snapmirror qtree to snapvault.

Uses a 7Mode CLI APIset.

Invokes "snapvault convert, options" commands for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "secondary-path" => $secondary_path >>

(Required for class method) secondary location path

Applicable for 7Mode CLI.

Maps to:

7M CLI: secondary path (programmatic name: secondary-path)

=item C<< "qtree-path" => $secondary_path >>

(Required for class method) secondary location path

Applicable for 7Mode CLI.

Maps to:

7M CLI: qtree path (programmatic name: qtree-path)

=item C<< "method-timeout" => $scalar >>

(Optional) timeout in seconds.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut


sub convert {
    $Log->enter() if $may_enter;
    my $pkg = shift;
    $pkg->_frontend_log_debug_opts(@_);
    my %opts = $pkg->_common_validate_with(
        params          => \@_,
        additional_spec => {
            _optional_scalars qw(
            qtree-path secondary-path
            ),
        },
        ignore_primary_keys => 1,
    );
    my $ret = $pkg->call_on_apiset(
          %opts,
          choices => [
                {
                  method    => '_convert_7mode_cli',
                  interface => 'CLI',
                  set       => '7Mode',
                },
          ],
    );
    $Log->exit() if $may_exit;
    return $ret;
}

sub _convert_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );
    delete $opts{command_interface};
    my $apiset = delete $opts{apiset};
    #delete $opts{'secondary-path'};

    $apiset->options('snapvault.enable'=> 'off');
    $apiset->options('snapmirror.enable'=> 'off');
    $apiset->snapvault_convert(%opts);
    $apiset->options('snapvault.enable'=> 'on');
    $apiset->options('snapmirror.enable'=> 'on');
    $Log->exit() if $may_exit;
}

=head2 release

    NACL::C::snapvault->release(
                   command_interface => $ci,
                   'secondary-path'   => $secondary-path
                  ...);

(Class method or Instance Method)
This method is used to release the snapvaulted qtree.

Uses a 7Mode CLI APIset.

Invokes "snapvault release" command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "secondary-path" => $secondary_path >>

(Required for class method) secondary location path

Applicable for 7Mode CLI.

Maps to:

7M CLI: secondary path (programmatic name: secondary-path)

=item C<< "primary-path" => $primary_path >>

(Required for class method) primary location path

Applicable for 7Mode CLI.

Maps to:

7M CLI: primary path (programmatic name: primary-path)

=item C<< "is-primary" => $boolean >>

(Optional) To specify whether the snapvault release is run on the primary
or secondary. By default if nothing is specified '1' is considered, i.e primary.

Applicable for 7Mode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut


sub release {
    $Log->enter() if $may_enter;
    my $pkg = shift;
    $pkg->_frontend_log_debug_opts(@_);
    my %opts = $pkg->_common_validate_with(
        params          => \@_,
        additional_spec => {
            _optional_scalars qw(
            primary-path secondary-path is-primary
            ),
        },
    );
    my $ret = $pkg->call_on_apiset(
          %opts,
          choices => [
              {
                  method    => '_release_7mode_cli',
                  interface => 'CLI',
                  set       => '7Mode',
              },
          ],
    );
    $Log->exit() if $may_exit;
    return $ret;
}
sub _release_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );
    delete $opts{command_interface};
    my $apiset = delete $opts{apiset};

    $apiset->snapvault_release(%opts);
    $Log->exit() if $may_exit;
}


=head2 restore

    NACL::C::snapvault->restore(
                   command_interface => $ci,
                   'secondary-path'   => $secondary-path
                  ...);

(Class method or Instance Method)
This method is used to restore the qtree.

Uses a 7Mode CLI APIset.

Invokes "snapvault restore" command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "secondary-path" => $secondary_path >>

(Required for class method) secondary location path

Applicable for 7Mode CLI.

Maps to:

7M CLI: secondary path (programmatic name: secondary-path)

=item C<< "primary-location" => $secondary_path >>

(Required for class method) primary location path

Applicable for 7Mode CLI.

Maps to:

7M CLI: primary path (programmatic name: primary-path)

=item C<< "max-transfer-rate" => $scalar >>

(Optional)maximum speed at which data is transferred in kilobytes per second.

Applicable for 7Mode CLI.

Maps to:

7M CLI: -k <value> (programmatic name: max-transfer-rate)

=item C<< "incremental-restore" => $scalar >>

(Optional)incremental-restore option while restoring.

Applicable for 7Mode CLI.

Maps to:

7M CLI: -r (programmatic name: incremental-restore)

=item C<< "snapshot" => $scalar >>

(Optional)snapshot name on which operation has to be performed.

Applicable for 7Mode CLI.

Maps to:

7M CLI: -s <value> (programmatic name: snapshot)

=item C<< "force" => $Boolean >>

(Optional)option causes the command to execute forcefully..

Applicable for 7Mode CLI.

Maps to:

7M CLI: -f (programmatic name: force)

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut


sub restore {
    $Log->enter() if $may_enter;
    my $pkg = shift;
    $pkg->_frontend_log_debug_opts(@_);
    my %opts = $pkg->_common_validate_with(
        params          => \@_,
        additional_spec => {
            _optional_scalars qw(
            incremental-restore force snapshot max-transfer-rate
            primary-location secondary-location secondary-path
            ),
        },
    );
    my $ret = $pkg->call_on_apiset(
          %opts,
          choices => [
                {
                  method    => '_restore_7mode_cli',
                  interface => 'CLI',
                  set       => '7Mode',
                },
          ],
    );
    $Log->exit() if $may_exit;
    return $ret;
}

sub _restore_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );
    my $prompts_answers  = ['.*Are you sure you want to.*' => 'y'];
    delete $opts{command_interface};
    my $apiset = delete $opts{apiset};
    my $allowed_options = [
        qw( incremental-restore force snapshot max-transfer-rate
            primary-location secondary-location secondary-path
            )
    ];
    $pkg->_handle_boolean_options(
            api_options     => \%opts,
            boolean_options => [ qw(incremental-restore force) ],
            is_7m_cli       => 1
    );
    my %quota_args;
    $pkg->_hash_copy(
        source => \%opts,
        copy   => $allowed_options,
        target => \%quota_args
    );
    %opts = %quota_args;
    if ( $quota_args{'incremental-restore'} eq "" ) {
        if ( $quota_args{'max-transfer-rate'} eq "" ) {
            if ( $quota_args{'snapshot'} ne "" ) {
                delete $opts{'incremental-restore'};
                delete $opts{'max-transfer-rate'};
            } else {
                delete $opts{'incremental-restore'};
                delete $opts{'snapshot'};
                delete $opts{'max-transfer-rate'};
            }
        } else {
                delete $opts{'incremental-restore'};
                delete $opts{'snapshot'};
        }
        $apiset->snapvault_restore(%opts);
    } else {
        if ( $quota_args{'max-transfer-rate'} eq "" ) {
            if ( $quota_args{'snapshot'} ne "" ) {
                delete $opts{'max-transfer-rate'};
            } else {
                delete $opts{'snapshot'};
                delete $opts{'max-transfer-rate'};
            }
        } else {
            delete $opts{'snapshot'};
        }
        $apiset->snapvault_restore(%opts, "connectrec-match_table" => $prompts_answers );
    }
    $Log->exit() if $may_exit;
}

__PACKAGE__->_load_event_class();

sub _primary_keys_validate_spec {
       return ( 'secondary-path'    => { type => SCALAR }
              );
}
1;
