
#
# Copyright (c) 2001-2012 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Statistics Component Module
## @author Sumit.Mittal@netapp.com, dl-nacl-dev@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::C::Statistics

=head1 DESCRIPTION

C<NACL::C::Statistics> is a derived class of
L<NACL::C::Component::ONTAP|lib-NACL-C-Component-ONTAP-pm>.  It represents access
to an ONTAP Statistics.  A related class is
L<NACL::CS::Statistics|lib-NACL-CS-Statistics-pm>, which represents
the past state of an ONTAP Statistics.

=head1 ATTRIBUTES

=head2 command_interface

See L<NACL::C::Component::command_interface|lib-NACL-C-Component-pm/command_interface>

=head2 node

The name of the node, used to identify it uniquely.

=head2 category

The name of the Category.

=head2 object

The name of the object.

=head2 counter

The name of counter.

=head1 Handling differences between Statistics v1 and v2

As of SN.0, Statistics-v2 was introduced. This section will briefly
go over the changes and the modifications that would need to be made to
scripts to handle these changes. More definitive references of the changes
made are available here:

http://wikid.netapp.com/w/User:Neetug/NGSH_Statistics_CLI/SN_CLI_Changes

and the Functional Spec:

http://wikid.netapp.com/w/Performance_Counters/SierraNevada/Specs/Counter_Manager_Enhancement_FS

The primary change made was to allow for statistics samples to be taken over
multiple time periods simultaneously, as opposed to the earlier functionality
which only allowed taking statistics samples over a single time period. Before
going over the changes that would need to be made, how would one determine
whether a build is running statistics v1 or v2?

=head2 Determining if statistics v1 or v2 is running on the build

L<is_feature_supported|lib-NACL-C-CommandInterface-pm/is_feature_supported>
can be used for this.

    my $is_v2 = $ci->is_feature_supported(feature => 'statistics-v2');

=head2 Changes to be made to scripts to work for v2

The following sections list the changes made and the corresponding changes
to be made in scripts to handle them.

=head3 I have script that calls 'statistics show' command. How can update my script for SN.0 changes?

=head4 If "statistics show" was being invoked only once

If the script was invoking "statistics show" only once, then only the raw
counter values would be returned. In this case, the "raw" option needs to
be added to the filter to explicitly request for raw counter values.
Examples:

    # Old code
    NACL::C::Statistics->find(
        command_interface => $ci,
        filter            => { object => 'aggregate' }
    );

    # New Code
    NACL::C::Statistics->find(
        command_interface => $ci,
        filter            => {
            object => 'aggregate',
            raw    => 'true'
        }
    );

    # Old code
    NACL::CS::Statistics->fetch(
        command_interface => $ci,
        filter            => { object => 'aggregate' }
    );

    # New Code
    NACL::CS::Statistics->fetch(
        command_interface => $ci,
        filter            => {
            object => 'aggregate',
            raw    => 'true'
        }
    );

Here's an example way of handling the above to work for either v1 or v2:

    my %filter = ( object => 'aggregate' );
    my $is_v2 = $ci->is_feature_supported(feature => 'statistics-v2');
    if ($is_v2) {
        $filter{raw} = 'true';
    }
    NACL::CS::Statistics->fetch(
        command_interface => $ci,
        filter            => \%filter
    );

=head4 If "statistics show" was being invoked multiple times

If "statistics show" was being invoked multiple times then processed
counter values were being read. This would require first invoking
"statistics start", which will start a sample (identified by its id).
The subsequent calls to "statistics show" would then need to filter
by this sample-id to show processed statistics information for this sample.
Example (any lines with "..." is meant to indicate that some other code
is being run between the various calls to fetch):

    # Old Code
    NACL::CS::Statistics->fetch(
        command_interface => $node,
        filter            => { object => 'volume' }
    );
    ...
    NACL::CS::Statistics->fetch(
        command_interface => $node,
        filter            => { object => 'volume' }
    );
    ...
    NACL::CS::Statistics->fetch(
        command_interface => $node,
        filter            => { object => 'volume' }
    );

    # New Code
    my $stat_samp_obj = NACL::C::Statistics->start(
        command_interface => $ci,
        object            => 'volume',
        'sample-id'       => $id
    );
    ...
    NACL::CS::Statistics->fetch(
        command_interface => $ci,
        filter            => { 'sample-id' => $id }
    );
    ...
    NACL::CS::Statistics->fetch(
        command_interface => $ci,
        filter            => { 'sample-id' => $id }
    );
    # Once we want to stop collecting data for a sample, we stop it
    $stat_samp_obj->sample_stop();

=head3 Where is the 'delta' column from 'statistics show' command output?

The output of the "statistics show" in v1 had a "delta" column. In v2, this
field has been removed. This is because in v2, both raw and processed values
cannot show up in the output simultaneously.

If raw values are requested (i.e. the "-raw" option is passed), then the "value"
field in the output will be the raw counter value.

    my $some_counter = NACL::CS::Statistics->fetch(
        command_interface => $ci,
        filter            => {
            object   => $object,
            instance => $instance,
            counter  => $counter,
            raw      => 'true'
        }
    );
    $Log->comment('Raw value is: ' . $some_counter->value());

If processed values are requested for (i.e. without the "-raw" option), then
the "value" field in the output will be the processed counter value.

    my $some_counter = NACL::CS::Statistics->fetch(
        command_interface => $ci,
        filter            => {
            object      => $object,
            instance    => $instance,
            counter     => $counter,
            'sample-id' => $id
        }
    );
    $Log->comment('Processed value is: ' . $some_counter->value());

=head3 Fields no longer present in show output

The fields 'numerator', 'denominator', 'base-value', 'base-delta',
'base-rate', 'base-units', 'numerators', 'denominators' fields are no longer
applicable for v2.

=cut

package NACL::C::Statistics;

use strict;
use warnings;

use NACL::CS::Statistics;
use NACL::ComponentUtils
    qw (_optional_scalars Dumper _translate_to_obj_field_name);
use NATE::BaseException qw(:try);
use base 'NACL::C::Component::ONTAP';
use Params::Validate qw(validate SCALAR OBJECT);
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();

=head1 METHODS

=head2 new

  my $statistics = NACL::C::Statistics->new(command_interface=>$ci,
                                            node =>$node_name,
                                            ....);

Return a new statistics component object with the given attributes.

=cut

use Class::MethodMaker [
    scalar => 'node',
    scalar => 'instance',
    scalar => 'category',
    scalar => 'object',
    scalar => 'counter',
];

=head2 init

Derived class of init - checks for any optional keys and calls $self->SUPER::init(@_) first to get the rest of the argument checking

=cut

sub init {
    my ($self, %opts) = @_;
    $self->SUPER::init(%opts);
    if (!defined $opts{instance} && !defined $opts{category}) {
        NATE::Exceptions::Argument->throw(
                  "Either 'instance' or 'category' should"
                . 'be provided in the call to new');
    }
    if (defined $opts{instance} && $opts{category}) {
        NATE::Exceptions::Argument->throw(
                  "Both 'instance' and 'category' cannot be"
                . 'provided in the call to new (either one should be provided)'
        );
    } ## end if ( defined $opts{instance...

} ## end sub init

=head2 show_periodic

    my $StatisticsShowPeriodic_object = NACL::C::Statistics->show_periodic(
                                command_interface => $ci,
                                'counter' => $counter,
                                'node' => $node,
                                ...);

(Class method)
This method is used to continuously display performance statistics at regular intervals.
It returns a single object of type L<NACL::CS::StatisticsShowPeriodic|lib-NACL-CS-StatisticsShowPeriodic-pm>
which in turns holds objects of type L<NACL::CS::StatisticsShowPeriodicEntry|lib-NACL-CS-StatisticsShowPeriodic-pm>


Uses a CMode CLI and 7Mode CLI APISet.
  Invokes 'sysstat' command for 7Mode CLI.

=head3 Limitations

For CMode, this method supports all the input options accepted by
'statistics show-periodic' and also populates all the output fields in the
returned CS object. Each hyphen/space in the field name will be replaced by a
underscore to obtain the attribute name.
Eg: 'Number of Constituents' will have to be invoked as
$cs_obj->Number_of_Constituents()

For 7Mode, it will always run 'sysstat -x' and any of the CM specific
inputs like object, instance, counter would be ignored. Thus all the statistics
related fields supported by 7Mode are obtained and only the following
fields are supported in the output object as of now for 7Mode -
'cpu-busy' (CPU), 'nfs-ops' (NFS), 'cifs-ops' (CIFS), 'disk-read' (Disk-read),
'disk-write' (Disk-kB/s-write), 'data-sent' (Net-kB/s-out),
'data-recv' (Net-kB/s-in)

Since CMode provides very fine granularity in obtaining the statistics of the
specific object - instance - counter whereas the 7Mode command 'sysstat'
doesn't, the component support for this method is also limited for these
reasons.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'counter' => $counter >>

  Applicable only for CMode CLI

=item C<< 'node' => $node >>

  Applicable only for CMode CLI

=item C<< 'object' => $string >>

  Applicable only for CMode CLI

=item C<< 'vserver' => $string >>

  Applicable only for CMode CLI

=item C<< 'iterations' => $string >>

  Applicable for CMode CLI, 7Mode CLI.
  Maps to:
    7M CLI : -c (Programmatic name: 'count')

=item C<< 'object-name' => $string >>

  Applicable only for CMode CLI

=item C<< 'interval' => $string >>

  Applicable for CMode CLI, 7Mode CLI.
  Maps to:
    7M CLI : interval (Programmatic name: "interval")

=item C<< 'instance' => $string >>

  Applicable only for CMode CLI

=item C<< 'summary' => $string >>

  Applicable for CMode CLI, 7Mode CLI.
  Maps to:
    7M CLI : -s (Programmatic name: 'print-summary')

=item C<< 'method-timeout' => $time_in_seconds >>

(Optional)
Over here this parameter has been set as a product of
no of iterations and interval and also gave additional 10 sec to avoid race case.
If either of the parameters are not defined
then it will have a default NATE timeout of 60 seconds.
This option can be used to specify a larger timeout value.
The value should be provided in seconds.

=item C<< 'filter' => $string >>

Filter Data
  Applicable only for CMode CLI

=item C<< 'raw' => $string >>

Raw Values Request
  Applicable only for CMode CLI

=item C<< 'extended_query' => $string >>

(Optional)
  Applicable only for CMode CLI

=item C<< 'preset' => $string >>

Preset
Applicable for CMode CLI

=item C<< '*' >>

This method takes all other options supported by
"statistics show-periodic" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub show_periodic {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( object vserver iterations object-name interval instance summary
                    counter category
                    filter raw

                    preset
                    )
            ),
        },
        cli_cmd => 'statistics show-periodic',
    );
    if (!defined $opts{'method-timeout'}) {
        if (defined($opts{iterations} && $opts{interval})) {
            $opts{'method-timeout'} = $opts{iterations} * $opts{interval};
            $opts{'method-timeout'} = $opts{'method-timeout'} + 10;
        }
    }
    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_show_periodic_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_show_periodic_7mode_cli',
                interface => 'CLI',
                set       => '7Mode',
            },
        ],
        frontend => 'NACL::C::Statistics::show_periodic',
    );

    $Log->exit() if $may_exit;
    return $ret;
} ## end sub show_periodic

sub _show_periodic_cmode_cli {
    $Log->enter() if $may_enter;
    require NACL::CS::StatisticsShowPeriodic;
    require NACL::CS::StatisticsShowPeriodicEntry;
    my ($pkg_or_obj, %opts) = @_;
    my @state_objs;
    my $response;

    $pkg_or_obj->_base_backend_cmode_cli(
        %opts,
        api => 'statistics_show_periodic',
        apiset_response_obj => \$response,
    );

    my $type;
    my @arr;
    my $output   = $response->get_parsed_output();
    while (my ($key, $value) = each %{$output->[0]{'statistics'}->[0]}) {
        my $method = $pkg_or_obj->_translate_to_obj_field_name($key);
        $method =~ s/\s+/_/g;
        unless (NACL::CS::StatisticsShowPeriodicEntry->can($method)) {
            if (ref $value eq 'ARRAY') {
                $type = 'array';
            } elsif (!ref $value) {
                $type = 'scalar';
            } else {
                NATE::BaseException->throw(
                          "Some other typedef other than a scalar "
                        . " and array that is ref($value)\n"
                        . NACL::ComponentUtils::Dumper($value));
            }
            push @arr, $type, $method;
        }
    }
    if (@arr) {
        Class::MethodMaker->import(
            [   '-target_class' => 'NACL::CS::StatisticsShowPeriodicEntry',
                @arr
            ]
        );
    }
    foreach my $row (@{$output->[0]{'statistics'}}) {
        my $obj = NACL::CS::StatisticsShowPeriodicEntry->new();
        $obj->_set_fields(row => $row);
        push @state_objs, $obj;
    }

    my %statistics;
    if (defined($output->[0]{'maximums'})) {
        my $maximums = NACL::CS::StatisticsShowPeriodicEntry->new();
        $maximums->_set_fields(row => $output->[0]{'maximums'}[0]);
        $statistics{maximums} = $maximums;
    }

    if (defined($output->[0]{'minimums'})) {
        my $minimums = NACL::CS::StatisticsShowPeriodicEntry->new();
        $minimums->_set_fields(row => $output->[0]{'minimums'}[0]);
        $statistics{minimums} = $minimums;
    }

    if (defined($output->[0]{'averages'})) {
        my $averages = NACL::CS::StatisticsShowPeriodicEntry->new();
        $averages->_set_fields(row => $output->[0]{'averages'}[0]);
        $statistics{averages} = $averages;
    }

    if (defined($output->[0]{'timeinterval'})) {
        while (my ($key, $value) = each %{$output->[0]{'timeinterval'}->[0]})
        {
            $statistics{$key} = $value;
        }
    }

    $statistics{statistics} = \@state_objs;

    my $new = NACL::CS::StatisticsShowPeriodic->new();
    $new->_set_fields(row => \%statistics);

    $Log->exit() if $may_exit;
    return $new;
} ## end sub _show_periodic_cmode_cli

sub _show_periodic_7mode_cli {
    $Log->enter() if $may_enter;
    require NACL::CS::StatisticsShowPeriodic;
    require NACL::CS::StatisticsShowPeriodicEntry;
    my $pkg_or_obj = shift;
    my %opts       = $pkg_or_obj->_validate_backend_opts(@_);
    my @state_objs;
    delete $opts{command_interface};
    my $apiset = delete $opts{apiset};

    my $sevenmode_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => [qw( interval )],
        map    => {
            "summary"    => "print-summary",
            "iterations" => "count"
        }
    );

    $pkg_or_obj->_handle_boolean_options(
        api_options     => $sevenmode_args,
        boolean_options => [qw(print-summary)],
        is_7m_cli       => 1,
    );

    %opts = %$sevenmode_args;

    # Run 'sysstat -a' to get all the fields in the output and
    # this is to avoid input mapping
    $opts{'print-all'} = 1;
    my $response = $apiset->sysstat(%opts);
    my $output   = $response->get_parsed_output();
    my @arr;
    my $type;
    my %hash = (
        'CPU'             => 'cpu-busy',
        'NFS'             => 'nfs-ops',
        'CIFS'            => 'cifs-ops',
        'Disk-read'       => 'disk-read',
        'Disk-kB/s-write' => 'disk-write',
        'Net-kB/s-out'    => 'data-sent',
        'Net-kB/s-in'     => 'data-recv'
    );

    my $stat_hash                  = $output->[0]{'stat'}->[0];
    my @valid_sevenmode_attributes = keys %hash;
    foreach my $attributes (@valid_sevenmode_attributes) {
        my $value = $stat_hash->{$attributes};
        if (ref $value eq 'ARRAY') {
            $type = 'array';
        } elsif (!ref $value) {
            $type = 'scalar';
        } else {
            NATE::BaseException->throw(
                      "Some other typedef other than a scalar "
                    . " and array that is ref($value)\n"
                    . NACL::ComponentUtils::Dumper($value));
        }
        my $cmode_attributes = $hash{$attributes};
        my $method
            = $pkg_or_obj->_translate_to_obj_field_name($cmode_attributes);
        unless (NACL::CS::StatisticsShowPeriodicEntry->can($method)) {
            my $value = $stat_hash->{$attributes};
            if (ref $value eq 'ARRAY') {
                $type = 'array';
            } elsif (!ref $value) {
                $type = 'scalar';
            } else {
                NATE::BaseException->throw(
                          "Some other typedef other than a scalar "
                        . " and array that is ref($value)\n"
                        . NACL::ComponentUtils::Dumper($value));
            }
            push @arr, $type, $method;
        }
    }
    Class::MethodMaker->import(
        ['-target_class' => 'NACL::CS::StatisticsShowPeriodicEntry', @arr]);
    foreach my $row (@{$output->[0]{'stat'}}) {
        my $obj          = NACL::CS::StatisticsShowPeriodicEntry->new();
        my $row_modified = $pkg_or_obj->_hash_copy(
            source => $row,
            map    => \%hash
        );
        $obj->_set_fields(row => $row_modified);
        push @state_objs, $obj;
    }
    my %statistics;

    if (defined($output->[0]{'Max'})) {
        my $max = $pkg_or_obj->_hash_copy(
            source => $output->[0]{'Max'},
            map    => \%hash
        );
        my $maximums = NACL::CS::StatisticsShowPeriodicEntry->new();
        $maximums->_set_fields(row => $max);
        $statistics{maximums} = $maximums;
    }

    if (defined($output->[0]{'Min'})) {
        my $min = $pkg_or_obj->_hash_copy(
            source => $output->[0]{'Min'},
            map    => \%hash
        );
        my $minimums = NACL::CS::StatisticsShowPeriodicEntry->new();
        $minimums->_set_fields(row => $min);
        $statistics{minimums} = $minimums;
    }

    if (defined($output->[0]{'Avg'})) {
        my $avg = $pkg_or_obj->_hash_copy(
            source => $output->[0]{'Avg'},
            map    => \%hash
        );
        my $averages = NACL::CS::StatisticsShowPeriodicEntry->new();
        $averages->_set_fields(row => $avg);
        $statistics{averages} = $averages;
    }

    $statistics{statistics} = \@state_objs;

    my $new = NACL::CS::StatisticsShowPeriodic->new();
    $new->_set_fields(row => \%statistics);

    $Log->exit() if $may_exit;
    return $new;
} ## end sub _show_periodic_7mode_cli

=head2 stop

    NACL::C::Statistics->stop(
                                command_interface => $ci,
                                ...);

(Class method)
This method is used to stop Sample.

Uses a CMode CLI APISet.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

  Applicable only for CMode CLI

=item C<< 'sample-id' => $string >>
Sample Identifier

  Applicable only for CMode CLI

=item C<< 'method-timeout' => $time_in_seconds >>

(Optional)
The default NATE timeout is 60 seconds. If the command might take more
time to complete, this option can be used to specify a larger timeout value.
The value should be provided in seconds.

=item C<< "extended_query" => $string >>

(Optional)
  Applicable only for CMode CLI.

=item C<< 'wait' => $string >>

Wait on the Stop to Finish
Applicable for CMode CLI

=item C<< '*' >>

This method takes all other options supported by
"statistics stop" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub stop {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( sample-id

                    wait
                    )
            ),
        },
        ignore_primary_keys => 1,
        cli_cmd             => 'statistics stop',
    );
    $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_stop_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_stop_cmode_zapi_non_iter',
                interface => 'ZAPI',
                set       => 'CMode',
                zapi_type => 'non-iter',
            },
        ],
        frontend => 'NACL::C::Statistics::stop',
    );

    $Log->exit() if $may_exit;
} ## end sub stop

sub _stop_cmode_cli {
    $Log->enter() if $may_enter;
    my($pkg, @args) = @_;

    $pkg->_base_backend_cmode_cli(
          @args,
          api => 'statistics_stop',
    );

    $Log->exit() if $may_exit;
} ## end sub _stop_cmode_cli

sub _stop_cmode_zapi_non_iter {
    $Log->enter() if $may_enter;
    my ($pkg, @opts) = @_;
    my $copy = [qw(sample-id)];
    $pkg->_base_backend_cmode_zapi(
        @opts,
        api  => 'perf-object-stop',
        copy => $copy,
    );
    $Log->exit() if $may_exit;
}

=head2 start

    my $StatisticsSamples_obj    = NACL::C::Statistics->start(
                                command_interface => $ci,
                                'object' => $object
                                ...);

(Class method)
This method is used to start Sample. It returns an object of type
L<NACL::C::StatisticsSamples|lib-NACL-C-StatisticsSamples-pm>.

Uses a CMode CLI APISet.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'object' => $object >>

(Required)
Object
  Applicable only for CMode CLI

=item C<< 'instance-uuid' => $instance_uuid >>

(Optional)
  Applicable only for CMode CLI

=item C<< 'counter' => $counter >>

(Optional)
Counter
  Applicable only for CMode CLI

=item C<< 'instance' => $instance >>

(Optional)
Instance
  Applicable only for CMode CLI

=item C<< 'filter' => $string >>

Filter
  Applicable only for CMode CLI

=item C<< 'vserver' => $string >>

Vserver
  Applicable only for CMode CLI

=item C<< 'sample-id' => $string >>

Sample Identifier
  Applicable for CMode CLI/ZAPI

=item C<< 'node' => $string >>

Node
  Applicable only for CMode CLI

=item C<< 'method-timeout' => $time_in_seconds >>

(Optional)
The default NATE timeout is 60 seconds. If the command might take more
time to complete, this option can be used to specify a larger timeout value.
The value should be provided in seconds.

=item C<< 'duration' => $string >>

Sample Duration in Minutes
Applicable for CMode CLI

=item C<< 'preset' => $string >>

Preset
Applicable for CMode CLI

=item C<< 'sample-type' => $string >>

Sample Type possible value(s) are, User,System
Applicable for CMode CLI

=item C<< 'wait' => $string >>

Wait on the Start to Finish
Applicable for CMode CLI

=item C<< '*' >>

This method takes all other options supported by
"statistics start" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub start {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( filter vserver sample-id node instance counter

                    duration

                    preset

                    sample-type

                    wait
                    )
            ),
        },
        cli_cmd => 'statistics start',
    );
    my $obj = $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_start_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_start_cmode_zapi_non_iter',
                interface => 'ZAPI',
                set       => 'CMode',
                zapi_type => 'non-iter',
                check     => '_check_zapi',
            },
        ],
        frontend => 'NACL::C::Statistics::start',
    );

    $Log->exit() if $may_exit;
    return $obj;
} ## end sub start

sub _start_cmode_cli {
    $Log->enter() if $may_enter;
    my($pkg, %opts) = @_;
    my $response;

    $pkg->_base_backend_cmode_cli(
          %opts,
          api => 'statistics_start',
          apiset_response_obj => \$response,
    );

    my $sample_id = $response->get_parsed_output->[0]->{'sample_id'};
    require NACL::C::StatisticsSamples;
    my $obj       = NACL::C::StatisticsSamples->new(
        command_interface => $opts{command_interface},
        'sample_id'       => $sample_id
    );

    $Log->exit() if $may_exit;
    return $obj;
} ## end _start_cmode_cli


__PACKAGE__->_load_event_class();

sub _start_cmode_zapi_non_iter {
    $Log->enter() if $may_enter;
    my ($pkg, %opts) = @_;
    my $copy
        = [
        qw(object filter sort-key sample-id max instance sort-order sample-type counter)
        ];

    $pkg->_base_backend_cmode_zapi(
        %opts,
        api  => 'perf-object-start',
        copy => $copy,
    );
    my $sample_id = $opts{'sample-id'};
    require NACL::C::StatisticsSamples;
    my $obj       = NACL::C::StatisticsSamples->new(
        command_interface => $opts{command_interface},
        'sample_id'       => $sample_id
    );
    $Log->exit() if $may_exit;
    return $obj;
}
sub _check_zapi {
     $Log->enter() if $may_enter;
     my ($pkg, %opts) = @_;
     if ($pkg->_check_relational_regex_filter(filter_value => $opts{object})){
         my $msg = "Since the value of object contains * hence falling back to CLI";
         $Log->comment($msg);
         $Log->exit() if $may_exit;
         NACL::Exceptions::InvalidChoice->throw($msg);
    }
    $Log->exit() if $may_exit;
}


sub _check_non_iter {
    $Log->enter() if $may_enter;
    my ($pkg, @opts) = @_;
    $pkg->_base_check_non_iter(
        @opts,
        _primary_keys => [qw(object instance counter)],
        _vs_context   => 1
    );
    $Log->exit() if $may_exit;
}

sub _primary_keys_validate_spec {
    my $pkg_or_obj = shift;
    my %opts       = validate(
        @_,
        {   command_interface => {
                type     => OBJECT,
                isa      => 'NACL::C::CommandInterface',
                optional => 1
            }
        }
    );
    my $command_interface = $opts{command_interface};

    my %default_spec = (
        node     => {type => SCALAR, optional => 1},
        category => {type => SCALAR, optional => 1},
        instance => {type => SCALAR, optional => 1},
        object   => {type => SCALAR},
        counter  => {type => SCALAR}
    );

    if (defined $command_interface) {
        return (
            node     => {type => SCALAR, optional => 1},
            instance => {type => SCALAR},
            object   => {type => SCALAR},
            counter  => {type => SCALAR}
        );
    } else {
        return %default_spec;
    }
} ## end sub _primary_keys_validate_spec

1;
