#
# Copyright (c) 2001-2012 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary StorageAggregate64bitUpgrade Component Module
## @author dl-nacl-dev@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::C::StorageAggregate64bitUpgrade

=head1 DESCRIPTION

C<NACL::C::StorageAggregate64bitUpgrade> is a derived class of
L<NACL::C::Component::ONTAP|lib-NACL-C-Component-ONTAP-pm>.  It represents access
to an ONTAP Aggregate.  A related class is
L<NACL::CS::StorageAggregate64bitUpgrade|lib-NACL-CS-StorageAggregate64bitUpgrade-pm>, which represents
the past state of an ONTAP Storage Aggregate 64-bit Upgrade.

=head1 ATTRIBUTES

=head2 command_interface

See L<NACL::C::Component::command_interface|lib-NACL-C-Component-pm/command_interface>

=head2 aggregate

=cut

package NACL::C::StorageAggregate64bitUpgrade;

use strict;
use warnings;

use NACL::ComponentUtils qw (_optional_scalars Dumper);

use base 'NACL::C::Component::ONTAP';
use NACL::CS::StorageAggregate64bitUpgrade;
use NACL::CS::StorageAggregate64bitUpgradeStartCheck;

use Params::Validate qw(validate validate_with SCALAR ARRAYREF);
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();

use NATE::BaseException qw(:try);

# Component exceptions
use NACL::C::Exceptions::StorageAggregate64bitUpgrade::AlreadyDisabled ();

=head1 METHODS

=head2 new

  my $StorageAggregate64bit = NACL::C::StorageAggregate64bitUpgrade->new(command_interface=>$ci,
            aggregate => $aggregate );

Return a new StorageAggregate64bitUpgrade component object with the given attributes.

=cut

use Class::MethodMaker [scalar => 'aggregate',];

=head2 scan_start

   NACL::C::StorageAggregate64bitUpgrade->scan_start(
                                command_interface => $ci,
                                aggregate         => $aggregate,
                                  ...);

                        or
    $aggr64bitupgrade_obj->scan_start(...);

(Class or instance method)
This method is used to start the 64-bit upgrade Scanner on an aggregate.

Uses a CMode CLI or 7Mode CLI APISet.
  Invokes 'aggr 64bit upgrade scan start' command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< aggregate => $string >>

(Required for class method, Not Applicable for instance method)
Name of the aggregate.
  Applicable for CMode CLI, 7Mode CLI.
  Maps to:
    7M CLI : aggregate name (Programmatic name: "aggregate")

=item C<< 'extended_query' => $string >>

(Optional)
  Applicable only for CMode CLI

=item C<< '*' >>

This method takes all other options supported by
"storage aggregate 64bit-upgrade scan-start" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub scan_start {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( aggregate
                    )
            )
        },
        api => 'storage_aggregate_64bit_upgrade_scan_start',
    );
    $Log->exit() if $may_exit;

    return $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => "_scan_start_cmode_cli",
                interface => "CLI",
                set       => "CMode",
            },
            {   method    => "_scan_start_7mode_cli",
                interface => "CLI",
                set       => "7Mode",
            },

        ],
        frontend => 'NACL::C::StorageAggregate64bitUpgrade::scan_start',
    );
} ## end sub scan_start

sub _scan_start_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;

    $pkg_or_obj->_base_backend_cmode_cli(
        %opts,
        api => 'storage_aggregate_64bit_upgrade_scan_start',
    );

    $Log->exit() if $may_exit;
} ## end sub _scan_start_cmode_cli

sub _scan_start_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );

    delete $opts{command_interface};
    my $apiset = delete $opts{apiset};

    my $allowed_7mode_options = [qw( aggregate )];

    my %args;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => $allowed_7mode_options,
        target => \%args,
    );
    %opts = %args;

    $apiset->aggr_64bit_upgrade_scan_start(%opts);
    $Log->exit() if $may_exit;
} ## end sub _scan_start_7mode_cli

=head2 start

    my $StorageAggregate64bitUpgrade_object = NACL::C::StorageAggregate64bitUpgrade->start(
                                command_interface => $ci,
                                  ...);

(Class method)
This method is used to start 64-bit upgrade on an aggregate.

Uses a CMode CLI (or) 7Mode CLI APISet.
  Invokes "aggr 64bit upgrade start" command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< aggregate => $string >>

(Required)
Name of aggregate
  Applicable for CMode CLI, 7Mode CLI.
  Maps to:
    7M CLI : aggregate name (Programmatic name: "aggregate")

=item C<< mode => $string >>

  Applicable for CMode CLI, 7Mode CLI.
  Maps to:
    7M CLI : -mode <value> (Programmatic name: "mode")

=item C<< 'extended_query' => $string >>

(Optional)
  Applicable only for CMode CLI

=item C<< '*' >>

This method takes all other options supported by
"storage aggregate 64bit-upgrade start" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub start {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw(mode
                    aggregate
                    )
            )
        },
        api => 'storage_aggregate_64bit_upgrade_start',
    );
    $Log->exit() if $may_exit;

    return $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => "_start_cmode_cli",
                interface => "CLI",
                set       => "CMode",
            },
            {   method    => "_start_7mode_cli",
                interface => "CLI",
                set       => "7Mode",
            }
        ],
        frontend => 'NACL::C::StorageAggregate64bitUpgrade::start',
    );
} ## end sub start

sub _start_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, %opts) = @_;

    $pkg->_base_backend_cmode_cli(
        %opts,
        api => 'storage_aggregate_64bit_upgrade_start',
    );

    $Log->exit() if $may_exit;
    return $pkg->new(
        command_interface => $opts{command_interface},
        aggregate         => $opts{aggregate},
    );
} ## end sub _start_cmode_cli

sub _start_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );

    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};

    my $allowed_7mode_options = [qw(mode aggregate)];
    my %args;
    $pkg->_hash_copy(
        source => \%opts,
        copy   => $allowed_7mode_options,
        target => \%args,
    );
    %opts = %args;

    $apiset->aggr_64bit_upgrade_start(%opts);
    $Log->exit() if $may_exit;

    return $pkg->new(
        command_interface => $command_interface,
        aggregate         => $opts{aggregate},
    );
} ## end sub _start_7mode_cli

=head2 start_check

my $object = NACL::C::StorageAggregate64bitUpgrade->start_check(
                                                    command_interface => $ci,
                                                    aggregate => $aggregate);
(or)
my @object = NACL::C::StorageAggregate64bitUpgrade->start_check(
                                                    command_interface => $ci,
                                                    aggregate => $aggregate);

(Class method)
This method is used to returns the space required for  64-bit upgrade of an
aggregate and it returns an object of type
L<NACL::CS::StorageAggregate64bitUpgradeStartCheck|lib-NACL-CS-StorageAggregate64bitUpgradeStartCheck-pm>
(or)
It can return an arrayref ,which contains objects of type
L<NACL::CS::StorageAggregate64bitUpgradeStartCheck|lib-NACL-CS-StorageAggregate64bitUpgradeStartCheck-pm>

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< aggregate => $string >>

(Required) Name of Aggregate.
  Applicable for CMode CLI, 7Mode CLI.
  Maps to:
    7M CLI : aggregate name (Programmatic name: "aggregate")

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub start_check {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(params => \@args,);

    my @objs = $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => "_start_check_cmode_cli",
                interface => "CLI",
                set       => "CMode",
            },
            {   method    => "_start_check_7mode_cli",
                interface => "CLI",
                set       => "7Mode",
            },
        ],
        frontend => 'NACL::C::StorageAggregate64bitUpgrade::start_check',
    );
    $Log->exit() if $may_exit;
    return wantarray ? @objs : $objs[0];
} ## end sub start_check

sub _start_check_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    my $obj;
    my @state_objs;
    my %opts = validate_with(
        params      => \@args,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );

    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};

    my $response = $apiset->storage_aggregate_64bit_upgrade_start(
        mode      => 'check',
        aggregate => $opts{'aggregate'}
    );
    my $output = $response->get_parsed_output();
    foreach my $row (@$output) {
        $obj = NACL::CS::StorageAggregate64bitUpgradeStartCheck->new();
        $obj->_set_fields(row => $row);
        push(@state_objs, $obj);
    }
    $Log->exit() if $may_exit;
    return @state_objs;
} ## end sub _start_check_cmode_cli

sub _start_check_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg = shift;
    my $obj;
    my @state_objs;
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );

    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};

    my $response = $apiset->aggr_64bit_upgrade_start(
        mode      => 'check',
        aggregate => $opts{'aggregate'}
    );
    my $output = $response->get_parsed_output();
    foreach my $row (@$output) {
        $obj = NACL::CS::StorageAggregate64bitUpgradeStartCheck->new();
        $obj->_set_fields(row => $row);
        push(@state_objs, $obj);
    }
    $Log->exit() if $may_exit;
    return @state_objs;
} ## end sub _start_check_7mode_cli

=head2 scan_stop

   NACL::C::StorageAggregate64bitUpgrade->scan_stop(
                                command_interface => $ci,
                                aggregate         => $aggregate,
                                  ...);
                    or

    $aggr64bitupgrade_obj->scan_stop(...);

(Class or instance method)
This method is used to stop the 64-bit upgrade Scanner on an aggregate.

Uses a CMode CLI or 7Mode CLI APISet.
  Invokes 'aggr 64bit upgrade scan stop' command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< aggregate => $string >>

(Required for class method, Not Applicable for instance method)
Name of the aggregate
  Applicable for CMode CLI, 7Mode CLI.
  Maps to:
    7M CLI : aggregate name (Programmatic name: "aggregate")

=item C<< 'extended_query' => $string >>

(Optional)
  Applicable only for CMode CLI

=item C<< '*' >>

This method takes all other options supported by
"storage aggregate 64bit-upgrade scan-stop" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub scan_stop {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( aggregate
                    )
            )
        },
        api => 'storage_aggregate_64bit_upgrade_scan_stop',
    );
    $Log->exit() if $may_exit;

    return $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => "_scan_stop_cmode_cli",
                interface => "CLI",
                set       => "CMode",
            },
            {   method    => "_scan_stop_7mode_cli",
                interface => "CLI",
                set       => "7Mode",
            },

        ],
        frontend => 'NACL::C::StorageAggregate64bitUpgrade::scan_stop',
    );
} ## end sub scan_stop

sub _scan_stop_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;

    $pkg_or_obj->_base_backend_cmode_cli(
        %opts,
        api => 'storage_aggregate_64bit_upgrade_scan_stop',
    );

    $Log->exit() if $may_exit;
} ## end sub _scan_stop_cmode_cli

sub _scan_stop_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );

    delete $opts{command_interface};
    my $apiset = delete $opts{apiset};

    my $allowed_7mode_options = [qw( aggregate )];

    my %args;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => $allowed_7mode_options,
        target => \%args,
    );
    %opts = %args;

    $apiset->aggr_64bit_upgrade_scan_stop(%opts);
    $Log->exit() if $may_exit;
} ## end sub _scan_stop_7mode_cli

=head2 status

    my $status_obj = NACL::C::StorageAggregate64bitUpgrade->status(command_interface => $ci,
                          [ filter => { %filter } ,]
              [ requested_fields => [ @req_fields ]);

                    or

    my @status_objs = NACL::C::StorageAggregate64bitUpgrade->status(command_interface => $ci,
                          [ filter => { %filter } ,]
              [ requested_fields => [ @req_fields ]);

                    or

   my $status_obj = $StorageAggregate64bitUpgrade_obj->status();

(Class or Instance method)
This method returns objects of the type L<NACL::CS::StorageAggregate64bitUpgradeStatus|lib-NACL-CS-StorageAggregate64bitUpgradeStatus-pm>.

This method accepts all the parameters accepted by L<fetch|lib-NACL-CS-ComponentState-pm/fetch>
(requested_fields, filter, allow_empty) and has all of the other properties of
fetch methods (callable in scalar or list context, throws a NoElementsFound
exception if no matching elements are found)

=cut

sub status {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = $pkg_or_obj->_handle_obj_cs_like_calls(@_);
    require NACL::CS::StorageAggregate64bitUpgradeStatus;
    my @state_objs
        = NACL::CS::StorageAggregate64bitUpgradeStatus->fetch(%opts);
    $Log->exit() if $may_exit;
    return wantarray ? @state_objs : $state_objs[0];
} ## end sub status

=head2 prepare_to_downgrade

    NACL::C::StorageAggregate64bitUpgrade->prepare_to_downgrade(
                                command_interface => $ci,
                                ...);

(Class method)
This method is used to restore 64-bit upgrade settings for releases earlier than Data ONTAP 8.2.1.
Uses CMode CLI .

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< '*' >>

This method takes all other options supported by "storage aggregate 64bit upgrade prepare to downgrade" command in ONTAP

=item C<< other_options >>

This method takes other common options described in L<Common method options|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#Other_options_valid_on_component_methods>

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub prepare_to_downgrade {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params              => \@args,
        ignore_primary_keys => 1,
        api => 'storage_aggregate_64bit_upgrade_prepare_to_downgrade',
    );
    $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_prepare_to_downgrade_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
        ],
        frontend =>
            'NACL::C::StorageAggregate64bitUpgrade::prepare_to_downgrade',
    );

    $Log->exit() if $may_exit;
}

sub _prepare_to_downgrade_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, @opts) = @_;
    $pkg->_base_backend_cmode_cli(@opts,
        api => 'storage_aggregate_64bit_upgrade_prepare_to_downgrade',);
    $Log->exit() if $may_exit;
}

=head2 storage_aggregate_64bit_upgrade_32bit_disable

    NACL::C::StorageAggregate64bitUpgrade->storage_aggregate_64bit_upgrade_32bit_disable(
                                command_interface => $ci,
                                ...);

(Class method)
This method is used to disable 32-bit block format.
Uses CMode CLI .

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'check' => $string >>

Check for 32bit storage objects
Applicable for CMode CLI

=item C<< 'undo' => $string >>

Enable format_32bit capability
Applicable for CMode CLI

=item C<< 'force' => $string >>

Force 32bit Disable
Applicable for CMode CLI

=item C<< '*' >>

This method takes all other options supported by "storage aggregate 64bit upgrade 32bit disable" command in ONTAP

=item C<< other_options >>

This method takes other common options described in L<Common method options|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#Other_options_valid_on_component_methods>

=back

=back

=over

=item Exceptions

=over

=item C<NACL::C::Exceptions::StorageAggregate64bitUpgrade::AlreadyDisabled>

This type of exception gets thrown if an attempt is made to disable 
32-bit aggregate support when its already disabled.

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub storage_aggregate_64bit_upgrade_32bit_disable {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( check undo force
                    )
            ),
        },
        ignore_primary_keys => 1,
        api => 'storage_aggregate_64bit_upgrade_32bit_disable',
    );
    try {
        $pkg->call_on_apiset(
            %opts,
            choices => [
                {   method =>
                        '_storage_aggregate_64bit_upgrade_32bit_disable_cmode_cli',
                    interface => 'CLI',
                    set       => 'CMode',
                },
            ],
            frontend =>
                'NACL::C::StorageAggregate64bitUpgrade::storage_aggregate_64bit_upgrade_32bit_disable',
        );
    }
    catch NACL::APISet::Exceptions::CommandFailedException with {
        my $exception = $_[0];
        NACL::C::Exceptions::StorageAggregate64bitUpgrade::AlreadyDisabled
            ->detect_convert_and_throw(exception => $exception);
        $Log->exit() if $may_exit;
        $exception->throw();
    };

    $Log->exit() if $may_exit;
}

sub _storage_aggregate_64bit_upgrade_32bit_disable_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, @opts) = @_;
    $pkg->_base_backend_cmode_cli(@opts,
        api => 'storage_aggregate_64bit_upgrade_32bit_disable',);
    $Log->exit() if $may_exit;
}


__PACKAGE__->_load_event_class();

sub _primary_keys_validate_spec {
    return (aggregate => {type => SCALAR},);
}
1;
