#
# Copyright (c) 2011 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Represents a switch machine with mgmt interface and IP address
## @author rawat@netapp.com, dl-nacl-dev@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::C::Switch

=head1 DESCRIPTION

C<NACL::C::Switch> is a derived class of
L<NACL::C::CommandInterface|lib-NACL-C-CommandInterface-pm>.  It represents
access to a Switch (Can be Network or FC switch). It can currently
only be used as a command interface for other components, though we do
plan to flesh it out into a component. That is why there is currently no
C<NACL::CS::Switch>.

=head1 ATTRIBUTES

=head2 switch

The name of the switch, used to identify it uniquely.

=head2 hostrec

Hostrec object for the switch.

=head2 Switch

Command Interface for the switch. Its an alias to the name.

=cut

package NACL::C::Switch;

use strict;
use warnings;

use Params::Validate qw(validate_pos validate_with SCALAR OBJECT ARRAYREF);
use NACL::CS::Switch;
use NATE::BaseException qw(:try);
use NATE::Exceptions::Argument;
use NACL::APISet::Exceptions::TimeoutException;
use NACL::ComponentUtils qw (_optional_scalars Dumper);
use NACL::GeneralUtils qw(_poll _powercycle);
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();

use base 'NACL::C::CommandInterface';

=head1 METHODS

=head2 new

my $Switch = NACL::C::Switch->new([name => $name],
                              [Switch => $Switch],
                              [hostrec => $hostrec] );
Specifying any one of the switch, name, hostrec attributes will populate
the other two.
Returns a new Switch component object with the given attributes.

=cut

use Class::MethodMaker [
    scalar => [
        {   '-default_ctor' => sub { $_[0]->_construct_primary_key(); },
            '-store_cb' => sub { $_[0]->_verify_hostrec_callback($_[1],$_[2]); }
        },
        'switch'
    ]
];

=head1 METHODS

=head2 enable

NACL::C::Switch->enable(
    command_interface   => $ci,
);

                            or

  $switch_object->enable();

(Class or instance method) This method enables the switch.
list of supported sets are BrocadeFC, Mellanox

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=back

=back

=cut

sub enable {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@opts);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@opts,
    );

    $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => "_enable_brocadefc_cli",
                interface => "CLI",
                set       => "BrocadeFC",
            },
            {
                method    => "_enable_mellanox_cli",
                interface => "CLI",
                set       => "Mellanox",
            },
        ],
    );
    $Log->exit() if $may_exit;
}

sub _enable_brocadefc_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    my %opts = validate_with(
        params      => \@opts,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1,
    );

    my $apiset = $opts{apiset};

    $apiset->switchenable();

    $Log->exit() if $may_exit;
}

sub _enable_mellanox_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    my %opts = validate_with(
        params      => \@opts,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1,
    );

    my $apiset = $opts{apiset};

    $apiset->enable();

    $Log->exit() if $may_exit;
}

=head2 disable

NACL::C::Switch->disable(
    command_interface   => $ci,
);

                            or

  $switch_object->disable();

(Class or instance method) This method disables the switch.
list of supported sets are BrocadeFC

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=back

=back

=cut

sub disable {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@opts);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@opts,
    );

    $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => "_disable_brocadefc_cli",
                interface => "CLI",
                set       => "BrocadeFC",
            },
        ],
    );
    $Log->exit() if $may_exit;
}

sub _disable_brocadefc_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    my %opts = validate_with(
        params      => \@opts,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1,
    );

    my $apiset = $opts{apiset};

    $apiset->switchdisable();

    $Log->exit() if $may_exit;
}

=head2 shutdown

NACL::C::Switch->shutdown(
    command_interface   => $ci,
);

                            or

  $switch_object->shutdown();

(Class or instance method) This method shutdown the switch.
list of supported sets are CiscoFC, DellEthernet

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< 'port' => $string >>

(Required for Dell Ethernet CLI)
Applicable for DellEthernet CLI.

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=back

=back

=cut

sub shutdown {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@opts);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@opts,
        additional_spec => {
            "port"      => { type => SCALAR, optional => 1 },
        },
    );

    $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => "_shutdown_ciscofc_cli",
                interface => "CLI",
                set       => "CiscoFC",
            },
            {
                method    => "_shutdown_dellethernet_cli",
                interface => "CLI",
                set       => "DellEthernet",
            },
        ],
    );
    $Log->exit() if $may_exit;
}

sub _shutdown_ciscofc_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    my %opts = validate_with(
        params      => \@opts,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1,
    );

    my $apiset = $opts{apiset};

    try {
        $apiset->shutdown();
    }
    catch NACL::APISet::Exceptions::TimeoutException with {}
    catch NATE::BaseException with {
        my $exception = shift;
        my $text      = $exception->text();

        if ( $text !~ /Unexpected EOF while reading output from command|Timeout waiting for command to complete/i ) {
            $Log->exit() if $may_exit;
            $exception->throw();
        }
    };

    $Log->exit() if $may_exit;
}

sub _shutdown_dellethernet_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);

    my $apiset = $opts{apiset};
    $apiset->configure( "subsystem" => "terminal", );
    try {
        $apiset->interface_tengigabitethernet(
                'unit-slot-port-name' => $opts{port},
        );
        $apiset->shutdown();
    } finally {
        $apiset->end();
    };
    $Log->exit() if $may_exit;
}

=head2 reboot

NACL::C::Switch->reboot(
    command_interface   => $ci,
);

                            or

  $switch_object->reboot();

(Class or instance method) This method reboot the switch.
list of supported sets are CiscoFC, BrocadeFC

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=back

=back

=cut

sub reboot {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@opts);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@opts,
        additional_spec => {
            retries => { type => SCALAR, default => 5 },
        },
    );

    $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => "_reboot_ciscofc_cli",
                interface => "CLI",
                set       => "CiscoFC",
            },
            {   method    => "_reboot_brocadefc_cli",
                interface => "CLI",
                set       => "BrocadeFC",
            },
            {   method    => "_reboot_cisconetwork_cli",
                interface => "CLI",
                set       => "CiscoNetwork",
            },
            {   method    => "_reboot_netappnetwork_cli",
                interface => "CLI",
                set       => "NetAppNetwork",
            },
        ],
    );
    $Log->exit() if $may_exit;
}

sub _reboot_ciscofc_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    my %opts = validate_with(
        params      => \@opts,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1,
    );

    my $apiset = $opts{apiset};

    my @prompt_answer = ('This command will reboot the system. (y/n)?' => "y");

    $pkg_or_obj->_reboot(
        prompts           => \@prompt_answer,
        command           => "reload",
        apiset            => $apiset,
        retries           => $opts{retries},
        command_interface => $opts{command_interface},
    );

    $Log->exit() if $may_exit;
}

sub _reboot_brocadefc_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    my %opts = validate_with(
        params      => \@opts,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1,
    );

    my $apiset = $opts{apiset};

    my @prompt_answer = ('Are you sure you want to reboot the switch \[y\/n\]?' => "y");

    $pkg_or_obj->_reboot(
        prompts           => \@prompt_answer,
        command           => "reboot",
        apiset            => $apiset,
        retries           => $opts{retries},
        command_interface => $opts{command_interface},
    );

    $Log->exit() if $may_exit;
}


sub _reboot_cisconetwork_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    my %opts = validate_with(
        params      => \@opts,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1,
    );

    my $apiset = $opts{apiset};

    my @prompt_answer = (
        'This command will reboot the system. (y/n)?' => "y",
        'Proceed with reload\?\s*\[confirm\]' => "y",
        'System configuration has been modified. Save\?\s+\[yes\/no\]:' => "y",
        'Do you want to continue\?\s*\(y\/n\)\s*\[n\]' => "y",
    );

    $pkg_or_obj->_reboot(
        prompts           => \@prompt_answer,
        command           => "reload",
        apiset            => $apiset,
        retries           => $opts{retries},
        command_interface => $opts{command_interface},
    );

    $Log->exit() if $may_exit;
}

sub _reboot_netappnetwork_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    my %opts = validate_with(
        params      => \@opts,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1,
    );

    my $apiset = $opts{apiset};

    my @prompt_answer = (
        'Would you like to save them now\?\s*\(y\/n\)' => "y",
        'Are you sure you would like to reset the system\?\s*\(y\/n\)' => "y",
    );

    $pkg_or_obj->_reboot(
        command_interface => $opts{command_interface},
        prompts           => \@prompt_answer,
        command           => "reload",
        apiset            => $apiset,

        # I know this is not a good command to run to
        # verify that switch is rebooted. But NetApp ethernet switches
        # doesnt support "echo" command.
        verify_cmd        => "show hosts",
        retries           => $opts{retries},
    );

    $Log->exit() if $may_exit;
}

sub _reboot {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    my %opts = validate_with(
        params      => \@opts,
        spec        => {
            command           => { type => SCALAR },
            retries           => { type => SCALAR, default => 10 },
            prompts           => { type => ARRAYREF, optional => 1 },
            verify_cmd        => { type => SCALAR, default => "echo hello" },
            apiset            => { type => OBJECT, isa => "NACL::APISet" },
            command_interface => { type => OBJECT, isa => "NACL::C::Switch" },
        }
    );

    my $apiset  = $opts{apiset};
    my $prompts = $opts{prompts};
    my $ci      = $opts{command_interface};
    my $command = $opts{command};
    my $retries = $opts{retries};
    my $client  = $ci->hostrec->hostname();
    my $verify_cmd = $opts{verify_cmd};


    try {
        $apiset->execute_raw_command(
            'connectrec-match_table' => $prompts,
            command                  => $command,
            timeout                  => 5,
        );
    }
    catch NATE::BaseException with {
        my $exception = shift;
        my $text      = $exception->text();

        if ( $text !~ /Unexpected EOF while reading output from command|Timeout waiting for command to complete|Timeout has occured/i ) {
            $Log->exit() if $may_exit;
            $exception->throw();
        } else {
            my $result = _poll(
                command_interface => $ci,
                retries           => $retries,
                verify_cmd        => $verify_cmd,
            );

            if ( $result == 1 ) {
                $Log->debug("Client \"$client\" rebooted successfully.");
            } else {
                _powercycle(
                    client_to_powercycle => $ci,
                    retries              => $retries,
                    verify_cmd           => $verify_cmd,
                );
            }
        }
    };

    $Log->exit() if $may_exit;
}

=head2 no_shutdown

    NACL::C::Switch->no_shutdown(
        command_interface   => $ci,
        port                => $port,
    );

              or

    $switch_object->no_shutdown();

(Class or instance method) This method disable a port in the switch.
list of supported sets are DellEthernet

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< 'port' => $string >>

(Required)
Applicable for DellEthernet CLI.

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=back

=back

=cut

sub no_shutdown {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@_);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            "port"      => {type => SCALAR,},
        },
    );

    $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {
                method    => "_no_shutdown_dellethernet_cli",
                interface => "CLI",
                set       => "DellEthernet",
            },
        ],
    );
    $Log->exit() if $may_exit;
}

sub _no_shutdown_dellethernet_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts =  $pkg_or_obj->_validate_backend_opts(@args);

    my $apiset = $opts{apiset};
    $apiset->configure( "subsystem" => "terminal", );
    try {
        $apiset->interface_tengigabitethernet(
            'unit-slot-port-name' => $opts{port},
        );
        $apiset->no_shutdown();
    } finally {
        $apiset->end();
    };
    $Log->exit() if $may_exit;
}

=head2 default_apiset_must

All switchs have to be of category "Switch", so this method returns
a hash with key 'category' whose value is "Switch".

Refer to L<default_apiset_must|lib-NACL-C-CommandInterface-pm/default_apiset_must>.

=cut

sub default_apiset_must {
    my $self     = shift;
    my $hostrec  = $self->hostrec();
    my $hostid   = $hostrec->id();
    my $hosttype = $hostrec->hosttype();

    my $set;
    if ( $hosttype =~ /^ethernet-switch-cisco/ ) {
        $set="CiscoNetwork";
    } elsif ( $hosttype =~ /^switch-brocade/ ) {
        $set="BrocadeFC";
    } elsif ( $hosttype =~ /^switch-cisco/ ) {
        $set="CiscoFC";
    } elsif ( $hosttype =~ /^armadillo-bridge/ ) {
        $set="Armadillo";
    } elsif ( $hosttype =~ /^switch-corduroyl/ ) {
        $set="NetAppNetwork";
    } elsif ( $hosttype =~ /^switch-dell(ethernet)?/ ) {
        $set="DellEthernet";
    } elsif ( $hosttype =~ /^switch-qlogic/ ) {
         $set="QlogicFC";
    } elsif ( $hosttype =~ /^(ethernet-)?switch-mellanox/ ) {
         $set="Mellanox";
    } else {
        NATE::BaseException->throw( "Switch $hostid was supposed "
          . "to be a switch but is actually of type '$hosttype'" );
    }
    my $source="command interface is a NACL::C::Switch " .
               "and hosttype='$hosttype'";
    return {
        category => "Switch",
        category_source => $source,
        set => $set,
        set_source => $source,
       };
}

=head2 default_apiset_could

The ruleset to default to if values for any of the co-ordinates 'category',
'interface' or 'set' cannot be determined.

This method returns category=Switch;interface=CLI;set=CiscoNetwork

Refer to L<default_apiset_could|lib-NACL-C-CommandInterface-pm/default_apiset_could>.

=cut

sub default_apiset_could {
    return {
        category  => "Switch",
        interface => "CLI",
        set       => "CiscoNetwork",
    };
} ## end sub default_apiset_could

=head2 _create_version_manager

   sub get_version_manager {
        ...
        if (keys %$apisets) {
            ...
        } else {
            $version_manager = $self->_create_version_manager();
        }
        return $version_manager;
    }

(Instance method)
This private method returns the version manager of the appropriate type.
This decision is based on the category ("Switch" for Switch command interfaces)
and the set.

=cut

sub _create_version_manager {
    $Log->enter() if $may_enter;
    my $self = shift;

    use NACL::APISet::Version;
    my $hostrec      = $self->hostrec();
    my $category     = "Switch";
    my $default_must = $self->default_apiset_must();
    my $shell_type   = $default_must->{set};
    $Log->debug( "Version manager object with parameters "
            . "category = $category; shell_type = $shell_type; "
            . "hostobj type = "
            . $hostrec->hosttype()
            . " created" );
    $Log->exit() if $may_exit;
    return NACL::APISet::Version->new(
        category   => $category,
        hostobj    => $hostrec,
        shell_type => $shell_type
    );
}

=head2 find

    my @switches = NACL::C::Switch->find();
    my $switch  = NACL::C::Switch->find();

Finds all Switches and returns command interface objects for each (can also
be called in scalar context, in which case it returns the first switch found).
Refer to L<find|lib-NACL-C-CommandInterface-pm/find> for more details.

You can also refer to user guide for more information on the options,
order of precedence for command line parameters (in case of find() is
invoked without specifying a command_interface) and examples.
http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/ComponentLayerUsersGuide/index.html#Obtaining_command_interface_objects

Unlike other command interfaces, you cannot call 'find' of Switch with
a command interface. The list of Switches is populated by using test parameters
(SWITCH or SWITCHES or SWITCH1,SWITCH2...). If 'find' is called and
command_interface is provided as parameter, a C<NATE::Exceptions::Argument>
exception is thrown.

=cut

# XXX Need to investigate whether an API exists which can be run on a switch
# which would give a list of other switches.
# At the moment, we do not allow command_interface to be passed.
sub find {
    my ( $pkg, %opts ) = @_;
    if ( defined $opts{command_interface} ) {
        NATE::Exceptions::Argument->throw( "Argument 'command_interface'"
                . "should not be passed in call to 'find' for the 'Switch' command"
                . " interface" );
    } else {
        $pkg->SUPER::find(%opts);
    }
} ## end sub find

=head2 type
    my $type = $switch_obj->type();

Returns the hosttype of the switch. Nothing more.

=cut

sub type {
    my $obj = shift;
    if (ref $obj) {
        return $obj->hostrec()->hosttype();
    } else {
        NATE::BaseException->throw( "The method type() can only be ".
                            "invoked on a NACL::C::Switch object" );
    }
}

# Switches will have some state information, but we don't as yet have a need
# for it. For the time being, we'll just define it to do nothing.
# Once we flesh out what this state information is, we'll need to define
# a CS file for Switch.
sub state { }

sub _primary_keys_validate_spec {
    return ( switch => { type => SCALAR } );
}

sub resource_record_method_name {
    return "switches";
}

# For this command interface, what is the common prefix for the test parameters?
sub _ci_to_test_param_mapping {
    $Log->enter() if $may_enter;
    $Log->exit() if $may_exit;
    return 'SWITCH';
}

1;
