#
# Copyright (c) 2012-2013 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#
## @summary Port Component Module
## @author gupendra@netapp.com, dl-nacl-dev@netapp.com
## @status shared
## @pod here
################################################################################

=head1 NAME

NACL::C::Switch::Port

=head1 DESCRIPTION

C<NACL::C::Switch::Port> Is a derived class of
L<NACL::C::Component::Switch|lib-NACL-C-Component-Switch-pm>. It represents access to
a Port.

=head1 ATTRIBUTES

=head2 port_num

Port on switch

=head2 command_interface

See L<NACL::C::Component::command_interface|lib-NACL-C-Component-pm/command_interface>

=cut

###################
# Package
package NACL::C::Switch::Port;
use NACL::CS::Switch::Port;
###################
# Everytime
use strict;
use warnings;

###################

use base 'NACL::C::Component::Switch';

# NATE Module
use NATE::Exceptions::Argument qw(:try);
use Params::Validate qw(validate_pos validate_with SCALAR);
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();

use NACL::ComponentUtils qw(Dumper _optional_scalars);

=head1 METHODS

=head2 new

  my $port = NACL::C::Switch::Port->new(command_interface=>$ci,port => $port_num,
                                  );

Return a new Port component object with the given attributes.

=cut

=head2 enable

NACL::C::Switch::Port->enable(
    command_interface   => $ci,
    port => $port_name,
);

                            or

  $port->enable();

(Class or instance method) This method enables port on the switch.
list of supported sets are BrocadeFC,CiscoFC and Mellanox

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< port => $number >>

=item C<< index => $number >>

=item C<< slot-number => $number >>

=item C<< ignore-nonexisting-port => $boolean >>

=item C<< port-channel => $number >>

=item C<< port-type => $string >>
Based on its value different apiset methods will be called in case of CiscoFC switches for example if port-type = 'eth' then interface_eth will be called

=back

=back

=cut

use Class::MethodMaker [
    new    => [ '-hash', 'init', 'new' ],
    scalar => 'port',
    scalar => [ { -type => 'NACL::C::Switch' }, 'command_interface', ],
];

sub enable {

    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    $pkg_or_obj->_frontend_log_debug_opts(@_);
    my %opts = $pkg_or_obj->_common_validate_with(
        allow_extra     => 1,
        params          => \@_,
        additional_spec => {
            _optional_scalars(
                qw ( slot-number index port ignore-nonexisting-port deconfigure )
            ),
            'port-type' => { type => SCALAR, optional => 1, default => 'fc' },
        },
    );
    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {
                method    => "_enable_brocadefc_cli",
                interface => "CLI",
                set       => "BrocadeFC"
            },
            {
                method    => "_enable_ciscofc_cli",
                interface => "CLI",
                set       => "CiscoFC"
            },
            {
                method    => "_enable_mellanox_cli",
                interface => "CLI",
                set       => "Mellanox",
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $ret;

}    #end of sub enable

sub _enable_brocadefc_cli {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $apiset            = delete $opts{apiset};
    my $command_interface = delete $opts{command_interface};
    my %api_opts;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%api_opts,
        copy   => [qw( slot-number index port ignore-nonexisting-port )],
    );
    $apiset->portenable(%api_opts);
    my $ret = $pkg_or_obj->new(
        'port'            => $opts{'port'},
        command_interface => $command_interface,
    );
    $Log->exit() if $may_exit;
    return $ret;

}    #end sub _enable_brocadefc_cli

sub _enable_ciscofc_cli {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );

    my $port_type         = delete $opts{'port-type'};
    my $apiset            = delete $opts{apiset};
    my $command_interface = delete $opts{command_interface};
    my %api_opts;

    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%api_opts,
        copy   => [qw( port port-channel)],
    );

    my ($ret,$cmd);
    if ( $port_type =~ /eth/) {
         $cmd = "interface_eth";
         $apiset->configure();
    } elsif( $port_type =~ /vfc/ ) {
          $cmd = "interface_vfc";
          $apiset->configure();
    } else{
          $cmd = "interface_fc";
          $apiset->configure( subsystem => "terminal" );
    }

    $apiset->$cmd(%api_opts);
    $api_opts{'deconfigure'} = 1;
    delete $api_opts{port};
    $apiset->shutdown(%api_opts);
    $apiset->end();
    $ret = $pkg_or_obj->new(
       'port'            => $opts{'port'},
       command_interface => $command_interface,
    );

    $Log->exit() if $may_exit;
    return $ret;

}    #end sub _enable_ciscofc_cli

sub _enable_mellanox_cli {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );

    my $port_type         = delete $opts{'port-type'};
    my $apiset            = delete $opts{apiset};
    my $command_interface = delete $opts{command_interface};
    my (%api_opts,$ret,$cmd);

    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%api_opts,
        copy   => [qw( port port-channel)],
    );

    $apiset->enable();
    $apiset->configure_terminal();

    if ( $port_type =~ /eth/) {
        $cmd = "interface_eth";
    } else {
        $cmd = "interface_ib";
    }

    $apiset->$cmd(%api_opts);
    $api_opts{'deconfigure'} = 1;
    delete $api_opts{port};
    $apiset->shutdown(%api_opts);
    $apiset->exit();
    $apiset->exit();
    $apiset->disable();
    $ret = $pkg_or_obj->new(
       'port'            => $opts{'port'},
       command_interface => $command_interface,
    );

    $Log->exit() if $may_exit;
    return $ret;
}   #end sub _enable_mellanox_cli

=head2 disable

NACL::C::Switch::Port->disable(
    command_interface   => $ci,
    port => $port_name,
);

                            or

  $port->disable();

(Class or instance method) This method enables port on the switch.
list of supported sets are BrocadeFC,CiscoFC and Mellanox

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< slot-number => $number >>

=item C<< index => $number >>

=item C<< port => $number >>

=item C<< port-type => $string >>
Based on its value different apiset methods will be called in case of CiscoFC switches for example if port-type = 'eth' then interface_eth will be called

=item C<< ignore-nonexisting-port => $number >>

=back

=back

=cut

sub disable {

    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    $pkg_or_obj->_frontend_log_debug_opts(@_);
    my %opts = $pkg_or_obj->_common_validate_with(
        allow_extra     => 1,
        params          => \@_,
        additional_spec => {
            _optional_scalars(
                qw ( slot-number index port ignore-nonexisting-port)
            ),
            'port-type' => { type => SCALAR, optional => 1, default => 'fc' },
        },
    );
    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {
                method    => "_disable_brocadefc_cli",
                interface => "CLI",
                set       => "BrocadeFC"
            },
            {
                method    => "_disable_ciscofc_cli",
                interface => "CLI",
                set       => "CiscoFC"
            },
            {
                method    => "_disable_mellanox_cli",
                interface => "CLI",
                set       => "Mellanox",
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $ret;

}    #end of sub disable

sub _disable_ciscofc_cli {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );

    my $port_type         = delete $opts{'port-type'};
    my $apiset            = delete $opts{apiset};
    my $command_interface = delete $opts{command_interface};
    my (%api_opts,$cmd);

    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%api_opts,
        copy   => [qw( port)],
    );

    if( $port_type =~ /eth/ ) {
        $cmd  = "interface_eth";
        $apiset->configure();
    } elsif ( $port_type =~ /vfc/) {
          $cmd = "interface_vfc";
          $apiset->configure();
    } else {
          $cmd = "interface_fc";
          $apiset->configure( subsystem => "terminal" );
    }

    $apiset->$cmd(%api_opts);
    delete $api_opts{port};
    $apiset->shutdown(%api_opts);
    $apiset->end();

    $Log->exit() if $may_exit;

}    #end sub _disable_ciscofc_cli

sub _disable_brocadefc_cli {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $apiset            = delete $opts{apiset};
    my $command_interface = delete $opts{command_interface};
    my %api_opts;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%api_opts,
        copy   => [qw( slot-number index port ignore-nonexisting-port )],
    );
    $apiset->portdisable(%api_opts);
    $Log->exit() if $may_exit;

}    #end of sub _disable_brocadefc_cli

sub _disable_mellanox_cli {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $port_type         = delete $opts{'port-type'};
    my $apiset            = delete $opts{apiset};
    my $command_interface = delete $opts{command_interface};
    my (%api_opts,$cmd);

    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%api_opts,
        copy   => [qw( port)],
    );

    $apiset->enable();
    $apiset->configure_terminal();

    if ( $port_type =~ /eth/ ) {
        $cmd = "interface_eth";
    } else {
        $cmd = "interface_ib";
    }

    $apiset->$cmd(%api_opts);
    delete $api_opts{port};
    $apiset->shutdown(%api_opts);
    $apiset->exit();
    $apiset->exit();
    $apiset->disable();

    $Log->exit() if $may_exit;

}    # end of sub _disable_mellanox_cli

=head2 portstats_clear

NACL::C::Switch::Port->portstats_clear(
    command_interface   => $ci,
    port => $port_name,
);

                            or

  $port->portstats_clear();

(Class or instance method) This method enables port on the switch.
list of supported sets are BrocadeFC

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< port => $number >>

=item C<< port-index => $number >>

=back

=back

=cut

sub portstats_clear {

    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    $pkg_or_obj->_frontend_log_debug_opts(@_);
    my %opts = $pkg_or_obj->_common_validate_with(
        allow_extra     => 1,
        params          => \@_,
        additional_spec => { _optional_scalars( qw ( port port-index ) ), },
    );
    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {
                method    => "_portstats_clear_brocadefc_cli",
                interface => "CLI",
                set       => "BrocadeFC"
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $ret;

}    #end of sub portstats_clear

sub _portstats_clear_brocadefc_cli {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $apiset            = delete $opts{apiset};
    my $command_interface = delete $opts{command_interface};
    my %api_opts;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%api_opts,
        copy   => [qw( port port-index )],
    );
    $apiset->portstatsclear(%api_opts);
    $Log->exit() if $may_exit;

}    #end of sub _portstats_clear_brocadefc_cli

=head2 portlicense_enable

NACL::C::Switch::Port->portlicense_enable(
    command_interface   => $ci,
    port => $port_name,
);

                            or

  $port->portlicense_enable();

(Class or instance method) This method enables portlicense.
list of supported sets are CiscoFC

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< port => $number >>

=item C<< acquire-port-license  >>

=back

=back

=cut

sub portlicense_enable {

    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    $pkg_or_obj->_frontend_log_debug_opts(@_);
    my %opts = $pkg_or_obj->_common_validate_with(
        allow_extra => 1,
        params      => \@_,
        additional_spec =>
          { _optional_scalars( qw ( acquire-port-license ) ), },
    );
    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {
                method    => "_portlicense_enable_ciscofc_cli",
                interface => "CLI",
                set       => "CiscoFC"
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $ret;

}    #end of sub portlicense_enable

sub _portlicense_enable_ciscofc_cli {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $apiset            = delete $opts{apiset};
    my $command_interface = delete $opts{command_interface};
    my %api_opts;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%api_opts,
        copy   => [qw( port acquire-port-license )],
    );
    $apiset->configure( subsystem => "terminal" );
    $apiset->interface_fc(%api_opts);
    delete $api_opts{port};
    $apiset->shutdown();
    $apiset->port_license("acquire-port-license" => "1");
    $apiset->end();
    $Log->exit() if $may_exit;

}    #end of sub _portlicense_enable_ciscofc_cli

=head2 portlicense_disable

NACL::C::Switch::Port->portlicense_disable(
    command_interface   => $ci,
    port => $port_name,
);

                            or

  $port->portlicense_disable();

(Class or instance method) This method disables the portlicense.
list of supported sets are CiscoFC

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< port => $number >>

=item C<< acquire-port-license  >>

=back

=back

=cut

sub portlicense_disable {

    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    $pkg_or_obj->_frontend_log_debug_opts(@_);
    my %opts = $pkg_or_obj->_common_validate_with(
        allow_extra     => 1,
        params          => \@_,
        additional_spec => {
            _optional_scalars( qw ( deconfigure acquire-port-license ) ),
        },
    );
    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {
                method    => "_portlicense_disable_ciscofc_cli",
                interface => "CLI",
                set       => "CiscoFC"
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $ret;

}    #end of sub portlicense_disable

sub _portlicense_disable_ciscofc_cli {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $apiset            = delete $opts{apiset};
    my $command_interface = delete $opts{command_interface};
    my %api_opts;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%api_opts,
        copy   => [qw( port acquire-port-license )],
    );
    $apiset->configure( subsystem => "terminal" );
    $apiset->interface_fc(%api_opts);
    $api_opts{'deconfigure'} = 1;
    delete $api_opts{port};
    $apiset->shutdown();
    $apiset->port_license(%api_opts);
    $apiset->end();
    $Log->exit() if $may_exit;

}    #end of sub _portlicense_disable_ciscofc_cli

=head2 online

    NACL::C::Switch::Port->online(
        command_interface  => $ci,
        port               => $port_name,
    );

                     or

  $port->online();

(Class or instance method) This method sets port online on the switch.
List of supported sets are QLogicFC.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< port => $number >>

=back

=back

=cut

sub online {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@_);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
    );
    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {
                method    => "_online_qlogicfc_cli",
                interface => "CLI",
                set       => "QlogicFC"
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $ret;
}    #end of sub online

sub _online_qlogicfc_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset            = $opts{apiset};
    my $command_interface = $opts{command_interface};
    my %api_opts;
    $pkg_or_obj->_hash_move(
        source => \%opts,
        target => \%api_opts,
        map    => { port => 'port-num' },
    );
    $api_opts{'state'} = "online";
    $apiset->admin_start();
    try {
        $apiset->set_port(%api_opts);
    } finally {
        $apiset->admin_stop();
    };

    $Log->exit() if $may_exit;
}    #end sub _enable_brocadefc_cli

=head2 offline

    NACL::C::Switch::Port->offline(
        command_interface  => $ci,
        port               => $port_name,
    );

                     or

    $port->offline();

(Class or instance method) This method sets port offline on the switch.
List of supported sets are QLogicFC.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< port => $number >>

=back

=back

=cut

sub offline {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@_);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
    );
    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {
                method    => "_offline_qlogicfc_cli",
                interface => "CLI",
                set       => "QlogicFC"
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $ret;
}    #end of sub offline

sub _offline_qlogicfc_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset            = $opts{apiset};
    my $command_interface = $opts{command_interface};
    my %api_opts;
    $pkg_or_obj->_hash_move(
        source => \%opts,
        target => \%api_opts,
        map    => { port => 'port-num' },
    );
    $api_opts{'state'} = "offline";
    $apiset->admin_start();
    try {
        $apiset->set_port(%api_opts);
    } finally {
        $apiset->admin_stop();
    };

    $Log->exit() if $may_exit;
}

sub _primary_keys_validate_spec {
    return ( 'port' => { type => SCALAR }, );
}

1;
