#
# Copyright (c) 2012 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Portcfgspeed Component Module
## @author gupendra@netapp.com, dl-nacl-dev@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::C::Switch::Portcfgspeed

=head1 DESCRIPTION

C<NACL::C::Switch::Portcfgspeed> Is a derived class of
L<NACL::C::Component::Switch|lib-NACL-C-Component-Switch-pm>. It represents access to
a Port.

=head1 ATTRIBUTES

=head2 port_num

Port number  on switch

=head2 speed_level

Speed level/number to set on switch

=head2 command_interface

See L<NACL::C::Component::command_interface|lib-NACL-C-Component-pm/command_interface>

=cut

package NACL::C::Switch::Portcfgspeed;
###################
# Everytime
use strict;
use warnings;

###################

use base 'NACL::C::Component::Switch';

# NATE Module
use NATE::Exceptions::Argument qw(:try);
use Params::Validate qw(validate_pos validate_with SCALAR ARRAYREF HASHREF);
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();

use NACL::ComponentUtils qw(Dumper _optional_scalars);

=head1 METHODS

=head2 new

    my $port = NACL::C::Switch::Portcfgspeed->new(
        command_interface   => $ci,
        'port-num'     => $port_num,
    );

Return a new Portcfgspeed component object with the given attributes.

=cut

use Class::MethodMaker [
    new    => [ '-hash', 'init', 'new' ],
    scalar => 'port_num',
    scalar => 'speed_level',
    scalar => [ { -type => 'NACL::C::Switch' }, 'command_interface', ],
];

=head2 set_speed
    my $port = NACL::C::Switch::Portcfgspeed->set_speed(
        command_interface   =>  $ci,
        'port-num'     => $port_num,
        'speed-level'  =>$port_speed,
    );

(Class method) This method  sets value of the C<< speed-level >> on specified port.
list of supported sets are CiscoFC and BrocadeFC

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'port-num'=>$number >>

=item C<< 'speed-level'=>$string >>
Represents a level (1 or 2)in Brocade or actual speed(4000) in Cisco switch.


=back

=back

=cut

sub set_speed {

    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    $pkg_or_obj->_frontend_log_debug_opts(@_);
    my %opts = $pkg_or_obj->_common_validate_with(
        allow_extra     => 1,
        params          => \@_,
        additional_spec => {
            _optional_scalars( qw ( port-num speed-level port-speed port deconfigure ) ),
        },
    );
    $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {
                method    => "_set_speed_brocadefc_cli",
                interface => "CLI",
                set       => "BrocadeFC"
            },
            {
                method    => "_set_speed_ciscofc_cli",
                interface => "CLI",
                set       => "CiscoFC"
            },
        ],
    );
    $Log->exit() if $may_exit;

}    #end of sub set_speed

sub _set_speed_brocadefc_cli {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $apiset            = delete $opts{apiset};
    my $command_interface = delete $opts{command_interface};
    my %api_opts;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%api_opts,
        copy   => [qw( port-num speed-level  )],
    );
    $apiset->portcfgspeed(%api_opts);
    $Log->exit() if $may_exit;

}    #end sub _set_speed_brocadefc_cli

sub _set_speed_ciscofc_cli {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $apiset            = delete $opts{apiset};
    my $command_interface = delete $opts{command_interface};
    my %api_opts;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%api_opts,
        copy   => [qw( port-speed port )],
        map    => {
            'speed-level' => 'port-speed',
            'port-num'    => 'port',
        },
    );
    $apiset->configure( subsystem => "terminal" );
    $apiset->interface_fc( port => "$api_opts{port}" );
    $apiset->switchport( 'port-speed' => $api_opts{'port-speed'} );
    $apiset->end();
    $Log->exit() if $may_exit;

}    #end sub _set_speed_ciscofc_cli

sub _primary_keys_validate_spec {
    return (
        'port-num'    => { type => SCALAR },
        'speed-level' => { type => SCALAR, optional => 1 },
    );

}

1;
