#
#
# Copyright (c) 2011-2012 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#
#
## @summary Switch Component Module
## @author rawat@netapp.com, dl-nacl-dev@netapp.com
## @status shared
## @pod here
################################################################################

=head1 NAME

NACL::C::Switch::Switch

=head1 DESCRIPTION

C<NACL::C::Switch::Switch> Is a derived class of
L<NACL::C::Component::Switch|lib-NACL-C-Component-Switch-pm>. It represents access to
a Switch.

=head1 ATTRIBUTES

=head2 command_interface

See L<NACL::C::Component::command_interface|lib-NACL-C-Component-pm/command_interface>

=cut

###################
# Package
package NACL::C::Switch::Switch;

###################
# Everytime
use strict;
use warnings;

###################
use base 'NACL::C::Component::Switch';

# NATE Module includes
use Params::Validate qw(validate_pos validate_with SCALAR);
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();

# NACL Modules
use NACL::CS::Switch::Switch;
use NACL::ComponentUtils qw (_optional_scalars Dumper);
use NATE::Exceptions::Argument qw(:try);
use NATE::BaseException;
use NACL::APISet::Exceptions::TimeoutException;

=head1 METHODS
 
=head2 new

    my $switch = NACL::C::Switch::Switch->new(
        command_interface   => $ci,
    );

Return a new switch component object with the given attributes.

=cut

=head2 enable

NACL::C::Switch::Switch->enable(
    command_interface   => $ci,
);

                            or

  $switch_object->enable();

(Class or instance method) This method enables the switch.
list of supported sets are BrocadeFC

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=back

=back

=cut

sub enable {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    $pkg_or_obj->_frontend_log_debug_opts(@_);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@_,
    );

    $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => "_enable_brocadefc_cli",
                interface => "CLI",
                set       => "BrocadeFC",
            },
        ],
    );
    $Log->exit() if $may_exit;
}

sub _enable_brocadefc_cli {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1,
    );

    my $apiset = $opts{apiset};
    
    $apiset->switchenable();
    
    $Log->exit() if $may_exit;
}

=head2 disable

NACL::C::Switch::Switch->disable(
    command_interface   => $ci,
);

                            or

  $switch_object->disable();

(Class or instance method) This method disables the switch.
list of supported sets are BrocadeFC

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=back

=back

=cut

sub disable {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    $pkg_or_obj->_frontend_log_debug_opts(@_);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@_,
    );

    $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => "_disable_brocadefc_cli",
                interface => "CLI",
                set       => "BrocadeFC",
            },
        ],
    );
    $Log->exit() if $may_exit;
}

sub _disable_brocadefc_cli {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1,
    );

    my $apiset = $opts{apiset};
    
    $apiset->switchdisable();
    
    $Log->exit() if $may_exit;
}

=head2 shutdown

NACL::C::Switch::Switch->shutdown(
    command_interface   => $ci,
);

                            or

  $switch_object->shutdown();

(Class or instance method) This method shutdown the switch.
list of supported sets are CiscoFC

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=back

=back

=cut

sub shutdown {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    $pkg_or_obj->_frontend_log_debug_opts(@_);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@_,
    );

    $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => "_shutdown_ciscofc_cli",
                interface => "CLI",
                set       => "CiscoFC",
            },
        ],
    );
    $Log->exit() if $may_exit;
}

sub _shutdown_ciscofc_cli {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1,
    );

    my $apiset = $opts{apiset};
    
    try {
        $apiset->shutdown();
    }
    catch NACL::APISet::Exceptions::TimeoutException with {}
    catch NATE::BaseException with {
        my $exception = shift;
        my $text      = $exception->text();

        if ( $text !~ /Unexpected EOF while reading output from command|Timeout waiting for command to complete/i ) {
            $Log->exit() if $may_exit;
            $exception->throw();
        }
    };
    
    $Log->exit() if $may_exit;
}

=head2 reboot

NACL::C::Switch::Switch->reboot(
    command_interface   => $ci,
);

                            or

  $switch_object->reboot();

(Class or instance method) This method reboot the switch.
list of supported sets are CiscoFC, BrocadeFC

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=back

=back

=cut

sub reboot {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@opts);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@opts,
    );

    $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => "_reboot_ciscofc_cli",
                interface => "CLI",
                set       => "CiscoFC",
            },
            {   method    => "_reboot_brocade_cli",
                interface => "CLI",
                set       => "BrocadeFC",
            },
        ],
    );
    $Log->exit() if $may_exit;
}

sub _reboot_ciscofc_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    my %opts = validate_with(
        params      => \@opts,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1,
    );

    my $apiset = $opts{apiset};
    
    my @prompt_answer = ('This command will reboot the system. (y/n)?' => "y");
    try {
        $apiset->execute_raw_command(command => "reload",
                                     'connectrec-match_table' => \@prompt_answer,
                                    );
    }
    catch NACL::APISet::Exceptions::TimeoutException with {}
    catch NATE::BaseException with {
        my $exception = shift;
        my $text      = $exception->text();

        if ( $text !~ /Unexpected EOF while reading output from command|Timeout waiting for command to complete/i ) {
            $Log->exit() if $may_exit;
            $exception->throw();
        }
    };
    
    $Log->exit() if $may_exit;
}

sub _reboot_brocade_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    my %opts = validate_with(
        params      => \@opts,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1,
    );

    my $apiset = $opts{apiset};
    
    my @prompt_answer = ('Are you sure you want to reboot the switch \[y\/n\]?' => "y");
    try {
        $apiset->execute_raw_command(command => "reboot",
                                     'connectrec-match_table' => \@prompt_answer,
                                    );
    }
    catch NACL::APISet::Exceptions::TimeoutException with {}
    catch NATE::BaseException with {
        my $exception = shift;
        my $text      = $exception->text();

        if ( $text !~ /Unexpected EOF while reading output from command|Timeout waiting for command to complete/i ) {
            $Log->exit() if $may_exit;
            $exception->throw();
        }
    };
    
    $Log->exit() if $may_exit;
}

sub _common_validate_with {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    
    $Log->warn( "This Library(NACL::C::Switch::Switch) is deprecated. " .
                "No new functionality will be added to this library. " .
                "Use NACL::C::Switch instead."
          );
    
    
    $Log->exit() if $may_exit;
    return $pkg_or_obj->SUPER::_common_validate_with(@opts);
}

1;
