# Copyright (c) 2011 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#
#
## @summary Vlan Component Module
## @author rawat@netapp.com, dl-nacl-dev@netapp.com
## @status shared
## @pod here
################################################################################

=head1 NAME

NACL::C::Switch::Vlan

=head1 DESCRIPTION

C<NACL::C::Switch::Vlan> Is a derived class of
L<NACL::C::Component::Switch|lib-NACL-C-Component-Switch-pm>. It represents access to
a mount_point.

=head1 ATTRIBUTES

=head2 vlan_id

vlan id on switch

=head2 command_interface

See L<NACL::C::Component::command_interface|lib-NACL-C-Component-pm/command_interface>

=cut

###################
# Package
package NACL::C::Switch::Vlan;

###################
# Everytime
use strict;
use warnings;

###################
# Module includes
use NACL::CS::Switch::Vlan;
use NACL::ComponentUtils qw (_optional_scalars Dumper);

use base 'NACL::C::Component::Switch';
use Params::Validate qw(validate_with SCALAR);
use NACL::APISet::Exceptions::ResponseException qw/:try/;

use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();

=head1 METHODS
 
=head2 new

    my $vlan_obj = NACL::C::Switch::Vlan->new(
        command_interface   => $ci,
        vlan_id             => 777
    );

Return a new vlan component object with the given attributes.

=cut

use Class::MethodMaker [
    scalar => 'vlan_id',
];

=head2 create

my $vlan_obj = NACL::C::Switch::Vlan->create(
    command_interface   => $ci,
    'vlan-id'           => 777
);

(Class method) Request creation of a new vlan element and return a new component object that refers to it.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'vlan-id'=>$string >>

(Required) c<< id >> of the vlan on which action will be performed

=back

=back

=cut

sub create {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
			'vlan-id' => { type => SCALAR },
        },
    );
    $Log->exit() if $may_exit;
    return $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => "_create_cisconetwork_cli",
                interface => "CLI",
                set       => "CiscoNetwork"
            },
        ],
    );
}

sub _create_cisconetwork_cli {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    my %opts = validate_with(
        params      => \@args,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );
    
    $pkg->_set_vlan_cisconetwork_cli(%opts);

	$Log->exit() if $may_exit;
    return $pkg->new(
		vlan_id		   	  => $opts{'vlan-id'},
        command_interface => $opts{'command_interface'},
    );	
}

=head2 set_mtu

NACL::C::Switch::Vlan->set_mtu(
    command_interface   =>  $ci,
    'vlan-id'           =>  777
    mtu                 =>  $mtu,
);

                            or

  $vlan_object->set_mtu( mtu => 9216 );

(Class or instance method) This method sets vale of the C<< mtu >> for vlan.
list of supported sets are CiscoNetwork

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'vlan-id'=>$string >>

(Required) c<< id >> of the vlan on which action will be performed

=item C<< mtu   =>  $int >>

(Required) mtu size to be set for the vlan

=item C<< deconfigure   =>  $boolean >>

(Optional) set this value to remove mtu setting from vlan. default to 0

=back

=back

=cut

sub set_mtu {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
		additional_spec => {
            mtu => { type => SCALAR },
            deconfigure => { type => SCALAR, default => 0 },
        },
    );

    $pkg_or_obj->call_on_apiset(
        %opts,
		choices => [
            {   method    => "_set_mtu_cisconetwork_cli",
                interface => "CLI",
                set       => "CiscoNetwork"
            },
        ],
    );
    $Log->exit() if $may_exit;
}

sub _set_mtu_cisconetwork_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts = validate_with(
        params      => \@args,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1,
    );

    my $pkg = ref($pkg_or_obj) || $pkg_or_obj;
    $pkg->_set_vlan_cisconetwork_cli(%opts);

    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    
    my %api_opts;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%api_opts,
        copy   => [ qw(mtu deconfigure)],
    );
	
	if ($command_interface->hostrec()->hosttype =~ /nxos/) {
		$Log->warn("Setting mtu size is not supported on Nexus OS");
	} else {
		$apiset->mtu(%api_opts);
        $pkg_or_obj->_exit_config_modes_cisconetwork(apiset => $apiset);
	}

	$Log->exit() if $may_exit;
}

=head2 set_name

NACL::C::Switch::Vlan->set_name(
    command_interface   =>  $ci,
    'vlan-id'           =>  777
    name                =>  $name,
);

                            or

  $vlan_object->set_name( name => nacl_vlan1 );

(Class or instance method) This method sets value of the C<< name >> for vlan.
list of supported sets are CiscoNetwork

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'vlan-id'=>$string >>

(Required) c<< id >> of the vlan on which action will be performed

=item C<< name   =>  $string >>

(Required) name to be set for the vlan

=item C<< deconfigure   =>  $boolean >>

(Optional) set this value to remove mtu setting from vlan. default to 0

=back

=back

=cut

sub set_name {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
		additional_spec => {
            name => { type => SCALAR },
            deconfigure => { type => SCALAR, default => 0 },
        },
    );

    $pkg_or_obj->call_on_apiset(
        %opts,
		choices => [
            {   method    => "_set_name_cisconetwork_cli",
                interface => "CLI",
                set       => "CiscoNetwork"
            },
        ],
    );
    $Log->exit() if $may_exit;
}

sub _set_name_cisconetwork_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts = validate_with(
        params      => \@args,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1,
    );

    my $pkg = ref($pkg_or_obj) || $pkg_or_obj;
    $pkg->_set_vlan_cisconetwork_cli(%opts);

    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    
    my %api_opts;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%api_opts,
        copy   => [ qw(name deconfigure)],
    );

    $apiset->name(%api_opts);
    $pkg_or_obj->_exit_config_modes_cisconetwork(apiset => $apiset);

	$Log->exit() if $may_exit;
}


=head2 set_state

NACL::C::Switch::Vlan->set_state(
    command_interface   => $ci,
    vlan-id             => 777,
    state               => "active"
);

                            or

$vlan_object->set_state( state => "active" );

(Class or instance method) This method sets value of the C<< state >> on vlan.
list of supported sets are CiscoNetwork

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'vlan-id'=>$string >>

(Required) c<< id >> of the vlan on which action will be performed

=item C<< deconfigure   =>  $boolean >>

(Optional) set this value to remove state setting from vlan. default to 0

=item C<< state=>$string >>

(Required) state to be set for this vlan

=back

=back

=cut

sub set_state {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
		additional_spec => {
            state => { type => SCALAR },
            deconfigure => { type => SCALAR, default => 0 },
        },
    );

    $pkg_or_obj->call_on_apiset(
        %opts,
		choices => [
            {   method    => "_set_state_cisconetwork_cli",
                interface => "CLI",
                set       => "CiscoNetwork"
            },
        ],
    );
    $Log->exit() if $may_exit;
}

sub _set_state_cisconetwork_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts = validate_with(
        params      => \@args,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1,
    );

    my $pkg = ref($pkg_or_obj) || $pkg_or_obj;
    $pkg->_set_vlan_cisconetwork_cli(%opts);
   
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    
    my %api_opts;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%api_opts,
        copy   => [ qw(state deconfigure)],
    );
    
    $apiset->state(%api_opts);
    $pkg_or_obj->_exit_config_modes_cisconetwork(apiset => $apiset);
    
	$Log->exit() if $may_exit;
}

=head2 up

  NACL::C::Switch::Vlan->up( command_interface => $ci, vlan-id => 777 );

                            or

  $vlan_object->up();

(Class or instance method) This method will bring vlan status to UP.
list of supported sets are CiscoNetwork

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

=item C<< 'vlan-id'=>$string >>

(Required) c<< id >> of the vlan on which action will be performed

=back

=back

=cut

sub up {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %args) = @_;
    $pkg_or_obj->down(%args, deconfigure => 1);
    $Log->exit() if $may_exit;
}

=head2 down

  NACL::C::Switch::Vlan->down(command_interface => $ci, vlan-id => 777);

                            or

  $vlan_object->down();

(Class or instance method) This method will bring vlan status to DOWN.
list of supported sets are CiscoNetwork

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

=item C<< 'vlan-id'=>$string >>

(Required) c<< id >> of the vlan on which action will be performed

=back

=back

=cut

sub down {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            deconfigure => { type => SCALAR, default => 0 },
        },
    );

    $pkg_or_obj->call_on_apiset(
        %opts,
		choices => [
            {   method    => "_down_cisconetwork_cli",
                interface => "CLI",
                set       => "CiscoNetwork"
            },
        ],
    );
    $Log->exit() if $may_exit;
}

sub _down_cisconetwork_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts = validate_with(
        params      => \@args,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1,
    );
    
    my $pkg = ref($pkg_or_obj) || $pkg_or_obj;
    $pkg->_set_vlan_cisconetwork_cli(%opts);
   
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    
    my %api_opts;
    $pkg->_hash_copy(
        source => \%opts,
        target => \%api_opts,
        copy   => [ qw(deconfigure)],
    );
    
    $apiset->shutdown(%api_opts);
    $pkg->_exit_config_modes_cisconetwork(apiset => $apiset);
    $Log->exit() if $may_exit;
}

=head2 purge

$vlan_obj->purge();

NACL::C::Switch::Vlan->purge(
    command_interface   => $command_interface,
    'vlan-id'           => '666',
);

(Class or Instance Method) This Component method will delete vlan from switch.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'vlan-id'=>$string >>

(Required) c<< id >> of the vlan on which action will be performed

=back

=back

=cut

sub purge {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts = $pkg_or_obj->_common_validate_with( params => \@args );

    my $pkg = ref($pkg_or_obj) || $pkg_or_obj;
    
    $Log->exit() if $may_exit;
    return $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => "_purge_cisconetwork_cli",
                interface => "CLI",
                set       => "CiscoNetwork"
            },
        ],
    );
} #end of sub purge

sub _purge_cisconetwork_cli {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    my %opts = validate_with(
        params      => \@args,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1,
    );

    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    
    ## delete vlan
    $pkg->_exit_config_modes_cisconetwork(apiset => $apiset);
    $apiset->configure(subsystem => "terminal");
    $apiset->vlan(%opts, deconfigure => 1);

    $Log->exit() if $may_exit;
}

#Private Method to set cli for vlan
sub _set_vlan_cisconetwork_cli {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    my %opts = validate_with(
        params      => \@args,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );

    my $command_interface = $opts{command_interface};
    my $apiset            = $opts{apiset};
    my %api_opts;
    $pkg->_hash_copy(
        source => \%opts,
        target => \%api_opts,
        copy   => [ qw(vlan-id)],
    );
    
    $pkg->_exit_config_modes_cisconetwork(apiset => $apiset);
	$apiset->configure(subsystem => "terminal");
    $apiset->vlan(%api_opts);
    
	$Log->exit() if $may_exit;
}

sub _primary_keys_validate_spec {
    return ( 
			'vlan-id' => { type => SCALAR },
		   );
}

1;
