# $Id: //depot/prod/test/nacldev/lib/NACL/C/Switch/Zone.pm# $

#
# Copyright (c) 2012 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Zone Component Module
## @author skotagir@netapp.com , dl-nacl-dev@netapp.com
## @status shared
## @pod here
################################################################################

=head1 NAME

NACL::C::Switch::Zone

=head1 DESCRIPTION

C<NACL::C::Switch::Zone> Is a derived class of
L<NACL::C::Component::Switch|lib-NACL-C-Component-Switch-pm>. It represents access to
a Zone.

=head1 ATTRIBUTES

=head2 zone-name

Name of the Zone on switch

=head2 command_interface

See L<NACL::C::Component::command_interface|lib-NACL-C-Component-pm/command_interface>

=cut

###################
# Package
package NACL::C::Switch::Zone;
use  NACL::CS::Switch::Zone;

###################
# Everytime
use strict;
use warnings;

###################

use base 'NACL::C::Component::Switch';

# NATE Module
use NATE::Exceptions::Argument qw(:try);
use NACL::C::Exceptions::Switch::ElementAlreadyExists;
use NACL::C::Exceptions::Switch::DoesNotExists;

use Params::Validate qw(validate_pos validate_with SCALAR ARRAYREF HASHREF);
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();

# NACL Module
use NACL::ComponentUtils qw(Dumper _optional_scalars);

=head1 METHODS

=head2 new

    my $zone = NACL::C::Switch::Zone->new(
        command_interface   => $ci,
        zone_name     => $zone_name,
    );

Return a new interface component object with the given attributes.

=cut

use Class::MethodMaker [
    scalar => 'zone_name',
    scalar => 'vsan_id',
];

=head2 create

    my $zone = NACL::C::Switch::Zone->create(
        command_interface   =>  $ci,
        'zone-name'    =>  $Zone_name,
        'vsan-id'     =>  $Vsan_id,
        'member'      =>  $Member,
    );

(Class method) Request creation of a new zone and return a new component object that refers to it.
Uses a CiscoFC-CLI or BrocadeFC-CLI APISet.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'zone-name'=>$string >>

(Required) c<< name >> of the zone is required on which action will be performed

=item C<< 'vsan-id'=>$string >>

Id of the vsan is required on which zone action will be performed
(Required) for CiscoFC Switch.
(Optional) for others

=item C<< 'member'=>$string >>

c<< member >> that are to be added to zone while creating zone
(Required) for BrocadeFC Switch.
(Optional) for others

=back

=item Exceptions

=over

=item C<NACL::C::Exceptions::Switch::ElementAlreadyExists>

This type of exception is thrown when an attempt is made to create
a element (e.g. zone) which already exists.

=item C<NACL::C::Exceptions::Switch::DoesNotExists.pm>

This type of exception is thrown when an attempt is access
a element (e.g. zone) which does not exists.

=back 

=back

=cut

sub create {

    $Log->enter() if $may_enter;
    my $pkg = shift;
    $pkg->_frontend_log_debug_opts(@_);
    my %opts = $pkg->_common_validate_with(
        allow_extra     => 1,
        params          => \@_,
        additional_spec => { _optional_scalars( qw ( member ) ), },
    );
    my $ret = $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => "_create_ciscofc_cli",
                interface => "CLI",
                set       => "CiscoFC"
            },
            {   method    => "_create_brocadefc_cli",
                interface => "CLI",
                set       => "BrocadeFC"
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $ret;

}    #end of sub create()

sub _create_ciscofc_cli {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );

    my $apiset            = delete $opts{apiset};
    my $command_interface = delete $opts{command_interface};
    
    # Cisco FC switch doesnt throw any error if Zone already exists.
    # Check if zone exists using "show zone". Throw exception if exists 
    
    my $zone = $opts{'zone-name'};
    my $zones = $pkg->find(command_interface => $command_interface, 
                           'filter'          => { 'zone-name' => "$zone" },
                           'allow_empty'     => 1,
                          );
    $Log->trace('Exception is not thrown it means zone already exists. ' .
                'Error out.');
    if ($zones) {
        NACL::C::Exceptions::Switch::ElementAlreadyExists->throw("Zone already exists");
    }
    
    my %api_opts;
    $pkg->_hash_copy(
        source => \%opts,
        target => \%api_opts,
        copy   => [qw( vsan-id zone-name)],
    );

    $apiset->configure( subsystem => "terminal" );
    $apiset->zone_name(%api_opts);
    $apiset->end();

    my $ret = $pkg->new(
        'zone-name'         => $api_opts{'zone-name'},
        'command_interface' => $command_interface,
        'vsan_id'           => $api_opts{'vsan-id'},
    );
    $Log->exit() if $may_exit;
    return $ret;

}    #end of sub _create_ciscofc_cli()

sub _create_brocadefc_cli {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );
    my $apiset            = delete $opts{apiset};
    my $command_interface = delete $opts{command_interface};
    $opts{member} = "\"" . "$opts{member}" . "\"";
    $opts{'zone-name'} = "\"" . "$opts{'zone-name'}" . "\"";
    my %api_opts;
    $pkg->_hash_copy(
        source => \%opts,
        target => \%api_opts,
        copy   => [qw( member zone-name)],
    );
    
    try {
        $apiset->zonecreate(%api_opts);
    } catch NACL::APISet::Exceptions::ResponseException with{
        my $exception = shift;
        $Log->exit() if $may_exit;
        NACL::C::Exceptions::Switch::ElementAlreadyExists
                ->detect_convert_and_throw('exception' => $exception, );
        
        $exception->throw;
    };

    my $ret = $pkg->new(
        'zone-name'         => $api_opts{'zone-name'},
        'command_interface' => $command_interface,
    );
    $Log->exit() if $may_exit;
    return $ret;
}    #end of sub_create_brocadefc_cli()

=head2 delete

    NACL::C::Switch::Zone->delete(
        command_interface   =>  $ci,
        'zone-name'    =>  $Zone_name,
        'vsan-id'     =>  $Vsan_id
    );

         or

   $zone_object->delete();

(Class or Instance method) Request deletion of an existing zone.
Uses a CiscoFC-CLI or BrocadeFC-CLI APISet.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'zone-name'=>$string >>

(Required) c<< name >> of the zone is required on which action will be performed

=item C<< 'vsan-id'=>$string >>

Id of the vsan is required on which the zone is present
(Required) for CiscoFC Switch.
(Optional) for others

=back

=back

=cut

sub delete {

    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    $pkg_or_obj->_frontend_log_debug_opts(@_);
    my %opts = $pkg_or_obj->_common_validate_with(
        allow_extra => 1,
        params      => \@_,
    );
    my $ret;
    try {
        $ret = $pkg_or_obj->call_on_apiset(
            %opts,
            choices => [
                {   method    => "_delete_ciscofc_cli",
                    interface => "CLI",
                    set       => "CiscoFC"
                },
                {   method    => "_delete_brocadefc_cli",
                    interface => "CLI",
                    set       => "BrocadeFC"
                },
            ],
        );
    } catch NACL::APISet::Exceptions::ResponseException with{
        my $exception = shift;
        NACL::C::Exceptions::Switch::DoesNotExists
                ->detect_convert_and_throw('exception' => $exception, );
        
        $Log->exit() if $may_exit;
        $exception->throw;
    };
    $Log->exit() if $may_exit;
    return $ret;

} ## end sub delete

sub _delete_ciscofc_cli {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );

    my $apiset            = delete $opts{apiset};
    my $command_interface = delete $opts{command_interface};
    my %api_opts;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%api_opts,
        copy   => [qw( vsan-id zone-name )],
    );

    $api_opts{'deconfigure'} = 1;
    $apiset->configure( subsystem => "terminal" );
    $apiset->zone_name(%api_opts);
    $apiset->end();

    $Log->exit() if $may_exit;
} ## end sub _delete_ciscofc_cli

sub _delete_brocadefc_cli {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $apiset            = delete $opts{apiset};
    my $command_interface = delete $opts{command_interface};
    my %api_opts;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%api_opts,
        copy   => [qw( zone-name)],
    );

    $apiset->zonedelete(%api_opts);

    $Log->exit() if $may_exit;
} ## end sub _delete_brocadefc_cli

=head2 alias_create

    my $zonealias = NACL::C::Switch::Zone->alias_create(
        'command_interface'   =>  $ci,
        'zone-name'    =>  $Zone_name,
        'vsan-id'     =>  $Vsan_id,
        'member'      =>  $Member,
    );

(Class method) Request creation of a new zone alias and return a new component object that refers to it.
Uses a CiscoFC-CLI or BrocadeFC-CLI APISet.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'zone-name'=>$string >>

(Required) c<< name >> of the zone is required on which action will be performed

=item C<< 'vsan-id'=>$string >>

Id of the vsan is required on which zone action will be performed
(Required) for CiscoFC Switch.
(Optional) for others`

=item C<< 'member'=>$string >>

 c<< member >> that are to be added to zone while creating zone
(Required) for BrocadeFC Switch.
(Optional) for others

=back

=back

=cut

sub alias_create {

    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    $pkg_or_obj->_frontend_log_debug_opts(@_);
    my %opts = $pkg_or_obj->_common_validate_with(
        allow_extra     => 1,
        params          => \@_,
        additional_spec => { _optional_scalars( qw ( member ) ), },
    );
    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => "_alias_create_ciscofc_cli",
                interface => "CLI",
                set       => "CiscoFC"
            },
            {   method    => "_alias_create_brocadefc_cli",
                interface => "CLI",
                set       => "BrocadeFC"
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $ret;

} ## end sub alias_create

sub _alias_create_brocadefc_cli {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );
    my $apiset            = delete $opts{apiset};
    my $command_interface = delete $opts{command_interface};
    $opts{'member'}    = "\"" . "$opts{member}" . "\"";
    $opts{'zone-name'} = "\"" . "$opts{'zone-name'}" . "\"";

    my %api_opts;
    $pkg->_hash_copy(
        source => \%opts,
        target => \%api_opts,
        copy   => [qw( member)],
        map    => { 'zone-name' => 'alias-name', }
    );

    $apiset->ali_create(%api_opts);

    my $ret = $pkg->new(
        'zone-name'         => $api_opts{'alias-name'},
        'command_interface' => $command_interface,
    );
    $Log->exit() if $may_exit;
    return $ret;

} ## end sub _alias_create_brocadefc_cli

sub _alias_create_ciscofc_cli {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );
    my $apiset            = delete $opts{apiset};
    my $command_interface = delete $opts{command_interface};

    my %api_opts;
    $pkg->_hash_copy(
        source => \%opts,
        target => \%api_opts,
        map    => {
            'zone-name' => 'fcalias-name',
            'vsan-id'   => 'vsan',
        }
    );

    $apiset->configure( subsystem => "terminal" );
    $apiset->fcalias(%api_opts);
    $apiset->end();

    my $ret = $pkg->new(
        'zone-name'         => $api_opts{'fcalias-name'},
        'command_interface' => $command_interface,
        'vsan_id'           => $api_opts{'vsan'},
    );
    $Log->exit() if $may_exit;
    return $ret;

} ## end sub _alias_create_ciscofc_cli

=head2 alias_delete

    NACL::C::Switch::Zone->alias_delete(
        command_interface   =>  $ci,
        'zone-name'    =>  $Zone_name,
        'vsan-id'     =>  $Vsan_id
    );
            or
    $zonealias_object->alias_delete();

(Class or Instance method) Request deletion of an existing zone alias.
Uses a CiscoFC-CLI or BrocadeFC-CLI APISet.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'zone-name'=>$string >>

(Required) c<< name >> of the zone is required on which action will be performed

=item C<< 'vsan-id'=>$string >>

Id of the vsan is required on which the zone is present
(Required) for CiscoFC Switch.
(Optional) for others

=back

=back

=cut

sub alias_delete {

    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    $pkg_or_obj->_frontend_log_debug_opts(@_);
    my %opts = $pkg_or_obj->_common_validate_with(
        allow_extra => 1,
        params      => \@_,
    );
    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => "_alias_delete_ciscofc_cli",
                interface => "CLI",
                set       => "CiscoFC"
            },
            {   method    => "_alias_delete_brocadefc_cli",
                interface => "CLI",
                set       => "BrocadeFC"
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $ret;

} ## end sub alias_delete

sub _alias_delete_brocadefc_cli {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $apiset            = delete $opts{apiset};
    my $command_interface = delete $opts{command_interface};
    $opts{'zone-name'} = "\"" . "$opts{'zone-name'}" . "\"";

    my %api_opts;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%api_opts,
        map    => { 'zone-name' => 'alias-name', },
    );

    $apiset->ali_delete(%api_opts);

    $Log->exit() if $may_exit;

} ## end sub _alias_delete_brocadefc_cli

sub _alias_delete_ciscofc_cli {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );
    my $apiset            = delete $opts{apiset};
    my $command_interface = delete $opts{command_interface};

    my %api_opts;
    $pkg->_hash_copy(
        source => \%opts,
        target => \%api_opts,
        map    => {
            'zone-name' => 'fcalias-name',
            'vsan-id'   => 'vsan',
        }
    );

    $apiset->configure( subsystem => "terminal" );
    $apiset->fcalias( %api_opts, 'deconfigure' => "1" );
    $apiset->end();

    $Log->exit() if $may_exit;

} ## end sub _alias_delete_ciscofc_cli

=head2 add_port

    NACL::C::Switch::Zone->add_port(
        'command_interface'   =>  $ci,
        'zone-name'    =>  $Zone_name,
        'vsan-id'     =>  $Vsan_id,
        'port-number'      =>  $Port_id,
        'domain-id' => $Domain_id,
    );

(Class method) adds a port to zone.
Uses a CiscoFC-CLI APISet.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'zone-name'=>$string >>

(Required) c<< name >> of the zone is required on which action will be performed

=item C<< 'vsan-id'=>$string >>

Id of the vsan is required on which zone action will be performed
(Required) for CiscoFC Switch.
(Optional) for others

=item C<< 'domain-id'=>$string >>

domain id of port that is to be added to zone

=item C<< 'port-number'=>$string >>

port number which we are goin to add to zone

=back

=back

=cut

sub add_port {

    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    $pkg_or_obj->_frontend_log_debug_opts(@_);
    my %opts = $pkg_or_obj->_common_validate_with(
        allow_extra => 1,
        params      => \@_,
        additional_spec =>
            { _optional_scalars( qw ( port-number domain-id ) ), },
    );
    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => "_add_port_ciscofc_cli",
                interface => "CLI",
                set       => "CiscoFC"
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $ret;

} ## end sub add_port

sub _add_port_ciscofc_cli {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );
    my $apiset            = delete $opts{apiset};
    my $command_interface = delete $opts{command_interface};

    my %zone_opts;
    my %member_opts;
    $pkg->_hash_copy(
        source => \%opts,
        target => \%zone_opts,
        copy   => [qw( vsan-id zone-name )],
    );
    $pkg->_hash_copy(
        source => \%opts,
        target => \%member_opts,
        copy   => [qw( port-number domain-id)],
    );

    $apiset->configure( subsystem => "terminal" );
    $apiset->zone_name(%zone_opts);
    $apiset->member(%member_opts);
    $apiset->end();

    $Log->exit() if $may_exit;

} ## end sub _add_port_ciscofc_cli

=head2 delete_port

    NACL::C::Switch::Zone->delete_port(
        command_interface   =>  $ci,
        'zone-name'    =>  $Zone_name,
        'vsan-id'     =>  $Vsan_id
        'domain-id' => "$Domain_id",
        'port-number' => "$Port_num",
    );

(Class method) Request deletion of a port from zone.
Uses a CiscoFC-CLI APISet.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'zone-name'=>$string >>

(Required) c<< name >> of the zone is required on which action will be performed

=item C<< 'vsan-id'=>$string >>

Id of the vsan is required on which the zone is present
(Required) for CiscoFC Switch.
(Optional) for others

=item C<< 'domain-id'=>$string >>

Domain id of the port we are deleting from zone

=item C<< 'port-number'=>$string >>

port number which we are goin to delete from zone

=back

=back

=cut

sub delete_port {

    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    $pkg_or_obj->_frontend_log_debug_opts(@_);
    my %opts = $pkg_or_obj->_common_validate_with(
        allow_extra => 1,
        params      => \@_,
        additional_spec =>
            { _optional_scalars( qw ( port-number domain-id vsan-id ) ), },
    );
    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => "_delete_port_ciscofc_cli",
                interface => "CLI",
                set       => "CiscoFC"
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $ret;

} ## end sub delete_port

sub _delete_port_ciscofc_cli {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );
    my $apiset            = delete $opts{apiset};
    my $command_interface = delete $opts{command_interface};

    my %zone_opts;
    my %member_opts;
    $pkg->_hash_copy(
        source => \%opts,
        target => \%zone_opts,
        copy   => [qw( vsan-id zone-name )],
    );
    $pkg->_hash_copy(
        source => \%opts,
        target => \%member_opts,
        copy   => [qw( port-number domain-id)],
    );

    $apiset->configure( subsystem => "terminal" );
    $apiset->zone_name(%zone_opts);
    $apiset->member( %member_opts, 'deconfigure' => "1" );
    $apiset->end();

    $Log->exit() if $may_exit;

} ## end sub _delete_port_ciscofc_cli

=head2 add_wwpn

    NACL::C::Switch::Zone->add_wwpn(
        'command_interface'   =>  $ci,
        'zone-name'    =>  $Zone_name,
        'vsan-id'     =>  $Vsan_id,
        'wwpn'     => $pwwn_id,
    );

(Class method) adds a wwpn to zone.
Uses a CiscoFC-CLI APISet.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'zone-name'=>$string >>

(Required) c<< name >> of the zone is required on which action will be performed

=item C<< 'vsan-id'=>$string >>

Id of the vsan is required on which zone action will be performed
(Required) for CiscoFC Switch.
(Optional) for others

=item C<< 'wwpn'=>$string >>

id of wwpn that is to be added to zone

=back

=back

=cut

sub add_wwpn {

    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    $pkg_or_obj->_frontend_log_debug_opts(@_);
    my %opts = $pkg_or_obj->_common_validate_with(
        allow_extra     => 1,
        params          => \@_,
        additional_spec => { _optional_scalars( qw ( wwpn ) ), },
    );
    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => "_add_wwpn_ciscofc_cli",
                interface => "CLI",
                set       => "CiscoFC"
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $ret;

} ## end sub add_wwpn

sub _add_wwpn_ciscofc_cli {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );
    my $apiset            = delete $opts{apiset};
    my $command_interface = delete $opts{command_interface};

    my %zone_opts;
    my %member_opts;
    $pkg->_hash_copy(
        source => \%opts,
        target => \%zone_opts,
        copy   => [qw( vsan-id zone-name )],
    );
    $pkg->_hash_copy(
        source => \%opts,
        target => \%member_opts,
        map    => { "wwpn" => "pwwn-id", },
    );

    $apiset->configure( subsystem => "terminal" );
    $apiset->zone_name(%zone_opts);
    $apiset->member(%member_opts);
    $apiset->end();

    $Log->exit() if $may_exit;

} ## end sub _add_wwpn_ciscofc_cli

=head2 delete_wwpn

    NACL::C::Switch::Zone->delete_wwpn(
        'command_interface'   =>  $ci,
        'zone-name'    =>  $Zone_name,
        'vsan-id'     =>  $Vsan_id,
        'wwpn'     => $pwwn_id,
    );

(Class method) deletes a wwpn from zone.
Uses a CiscoFC-CLI APISet.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'zone-name'=>$string >>

(Required) c<< name >> of the zone is required on which action will be performed

=item C<< 'vsan-id'=>$string >>

Id of the vsan is required on which zone action will be performed
(Required) for CiscoFC Switch.
(Optional) for others

=item C<< 'wwpn'=>$string >>

id of wwpn that is to be deleted from zone

=back

=back

=cut

sub delete_wwpn {

    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    $pkg_or_obj->_frontend_log_debug_opts(@_);
    my %opts = $pkg_or_obj->_common_validate_with(
        allow_extra     => 1,
        params          => \@_,
        additional_spec => { _optional_scalars( qw ( wwpn ) ), },
    );
    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => "_delete_wwpn_ciscofc_cli",
                interface => "CLI",
                set       => "CiscoFC"
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $ret;

} ## end sub delete_wwpn

sub _delete_wwpn_ciscofc_cli {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );
    my $apiset            = delete $opts{apiset};
    my $command_interface = delete $opts{command_interface};

    my %zone_opts;
    my %member_opts;
    $pkg->_hash_copy(
        source => \%opts,
        target => \%zone_opts,
        copy   => [qw( vsan-id zone-name )],
    );
    $pkg->_hash_copy(
        source => \%opts,
        target => \%member_opts,
        map    => { "wwpn" => "pwwn-id", },
    );

    $apiset->configure( subsystem => "terminal" );
    $apiset->zone_name(%zone_opts);
    $apiset->member( %member_opts, 'deconfigure' => "1" );
    $apiset->end();

    $Log->exit() if $may_exit;

} ## end sub _delete_wwpn_ciscofc_cli

=head2 save

    NACL::C::Switch::Zone->save();

(Class method) Request to save the configurtion of zones.
Uses a BrocadeFC-CLI APISet.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=back

=back

=cut

sub save {
    $Log->enter() if $may_enter;
    my ($pkg , @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        allow_extra => 1,
        params => \@args,
        ignore_primary_keys => 1,
    );

    my $ret = $pkg->call_on_apiset(
         %opts,
        choices => [
            {   method    => "_save_zonecfg_brocade_cli",
                interface => "CLI",
                set       => "BrocadeFC"
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $ret;

} ## end save

sub _save_zonecfg_brocade_cli {
    $Log->enter() if $may_enter;
    my ($pkg , @args) = @_;
    my %api_opts;
    my %opts = validate_with(
        params      => \@args,
        spec        => $pkg->_backend_validate_spec(),
        #allow_extra =>1
    );

    my $apiset = delete $opts{apiset};
    my $prompts_answers = ['.*Do you want to\s+.*(the)?Defined\s+\n*zoning configuration only?.*' => 'yes'];
    $api_opts{'connectrec-match_table'} = $prompts_answers;
    $apiset->cfgsave(%api_opts);

    $Log->exit() if $may_exit;
} ## end sub _save_zonecg_brocade_cli


sub _primary_keys_validate_spec {
    return (
        'zone-name' => { type => SCALAR },
        'vsan-id'   => { type => SCALAR, optional => 1 },
    );
} ## end sub _primary_keys_validate_spec

1;
