# $Id$

#
# Copyright (c) 2012 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Zoneset Component Module
## @author skotagir@netapp, dl-nacl-dev@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::C::Switch::Zoneset

=head1 DESCRIPTION

C<NACL::C::Switch::Zoneset> Is a derived class of
L<NACL::C::Component::Switch|lib-NACL-C-Component-Switch-pm>. It represents access to
a Zoneset.

=head1 ATTRIBUTES

=head2 command_interface

See L<NACL::C::Component::command_interface|lib-NACL-C-Component-pm/command_interface>

=head2 Zoneset

Name of the Zoneset.

=head2 vsan-id

vsan on which the Zoneset is present
This is required for ciscoFC switch

=head2 zone-name

Zone-name which are to be added to zoneset while creating
This is required for BrocadeFC while creating zoneset

=cut

package NACL::C::Switch::Zoneset;
use NACL::CS::Switch::Zoneset;

###################
# Everytime
use strict;
use warnings;

###################

use base 'NACL::C::Component::Switch';

# NATE Module
use NATE::Exceptions::Argument qw(:try);
use Params::Validate qw(validate_pos validate_with SCALAR ARRAYREF HASHREF);
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();

# NACL Module
use NACL::ComponentUtils qw(Dumper _optional_scalars);

use Class::MethodMaker [
    scalar => 'zoneset',
    scalar => 'vsan',
];

=head1 METHODS

=head2 create

    my $zoneset = NACL::C::Switch::Zoneset->create(
        command_interface   =>  $ci,
        'zoneset'    =>  $Zoneset_name,
        'vsan-id'     =>  $Vsan_id,
        'zone-name'      =>  $zone_name,
    );

(Class method) Request creation of a new zoneset and return a new component object that refers to it.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'zone-name'=>$string >>

c<< name >> of the zone is required while creating a zoneset.
(Required) for BrocadeFC Switch.
(Optional) for others

=item C<< 'vsan-id'=>$string >>

Id of the vsan is required on which zoneset action will be performed
(Required) for  CiscoFC Switch.
(Optional) for others.

=item C<< 'zoneset'=>$string >>

(Required) c<< zoneset >> name to be created on switch

=back

=back

=cut

sub create {
    $Log->enter() if $may_enter;
    my $pkg = shift;
    $pkg->_frontend_log_debug_opts(@_);
    my %opts = $pkg->_common_validate_with(
        allow_extra => 1,
        params      => \@_,
        additional_spec =>
            { _optional_scalars( qw ('vsan-id' 'zone-name') ), },
    );
    my $ret = $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => "_create_ciscofc_cli",
                interface => "CLI",
                set       => "CiscoFC"
            },
            {   method    => "_create_brocadefc_cli",
                interface => "CLI",
                set       => "BrocadeFC"
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $ret;

}    #end of sub create

sub _create_brocadefc_cli {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );

    my $apiset            = delete $opts{apiset};
    my $command_interface = delete $opts{command_interface};
    $opts{'zone-name'} = "\"" . "$opts{'zone-name'}" . "\"";
    $opts{'zoneset'}   = "\"" . "$opts{'zoneset'}" . "\"";
    my %api_opts;
    $pkg->_hash_copy(
        source => \%opts,
        target => \%api_opts,
        map    => {
            "zoneset"   => "cfgname",
            "zone-name" => "member",
        },
    );

    $pkg->_api_call_log_debug_opts(
        api      => 'cfgcreate',
        api_opts => \%api_opts
    );

    $apiset->cfgcreate(%api_opts);

    my $ret = $pkg->new(
        'zoneset'           => $api_opts{'cfgname'},
        'command_interface' => $command_interface,
    );
    $Log->exit() if $may_exit;
    return $ret;

} ## end sub _create_brocadefc_cli

sub _create_ciscofc_cli {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );

    my $apiset            = delete $opts{apiset};
    my $command_interface = delete $opts{command_interface};
    my %api_opts;
    $pkg->_hash_copy(
        source => \%opts,
        target => \%api_opts,
        copy   => [qw( vsan-id zoneset )],
    );

    $pkg->_api_call_log_debug_opts(
        api      => 'zoneset_name',
        api_opts => \%api_opts
    );

    $apiset->configure( subsystem => "terminal" );
    $apiset->zoneset_name(%api_opts);
    $apiset->end();

    my $ret = $pkg->new(
        'zoneset'           => $api_opts{'zoneset'},
        'vsan'              => $api_opts{'vsan-id'},
        'command_interface' => $command_interface,
    );
    $Log->exit() if $may_exit;
    return $ret;

} ## end sub _create_ciscofc_cli

=head2 delete

    NACL::C::Switch::Zone->delete(
        command_interface   =>  $ci,
        'zoneset'    =>  $Zoneset_name,
        'vsan-id'     =>  $Vsan_id
    );

(Class and Instance method)Request deletion of an existing zoneset.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'zoneset'=>$string >>

(Required) c<< name >> of the zoneset is required on which action will be performed

=item C<< 'vsan-id'=>$string >>

Id of the vsan is required on which the zone is present
This option is required for CiscoFC Switch only

=back

=back

=cut

sub delete {

    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    $pkg_or_obj->_frontend_log_debug_opts(@_);
    my %opts = $pkg_or_obj->_common_validate_with(
        allow_extra     => 1,
        params          => \@_,
        additional_spec => { _optional_scalars( qw ( 'vsan-id' ) ), },
    );
    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => "_delete_ciscofc_cli",
                interface => "CLI",
                set       => "CiscoFC"
            },
            {   method    => "_delete_brocadefc_cli",
                interface => "CLI",
                set       => "BrocadeFC"
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $ret;

} ## end sub delete

sub _delete_brocadefc_cli {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );

    my $apiset            = delete $opts{apiset};
    my $command_interface = delete $opts{command_interface};

    my %api_opts;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%api_opts,
        map    => { "zoneset" => "cfgname", },
    );

    $pkg_or_obj->_api_call_log_debug_opts(
        api      => 'cfgdelete',
        api_opts => \%api_opts
    );

    $apiset->cfgdelete(%api_opts);

    $Log->exit() if $may_exit;

} ## end sub _delete_brocadefc_cli

sub _delete_ciscofc_cli {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );

    my $apiset            = delete $opts{apiset};
    my $command_interface = delete $opts{command_interface};
    my %api_opts;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%api_opts,
        copy   => [qw( vsan-id zoneset )],
    );

    $pkg_or_obj->_api_call_log_debug_opts(
        api      => 'zoneset_name',
        api_opts => \%api_opts
    );

    $api_opts{'deconfigure'} = 1;
    $apiset->configure( subsystem => "terminal" );
    $apiset->zoneset_name(%api_opts);
    $apiset->end();

    $Log->exit() if $may_exit;
} ## end sub _delete_ciscofc_cli

=head2 add_zone

    NACL::C::Switch::Zone->add_zone(
        command_interface   =>  $ci,
        'zoneset'    =>  $Zoneset_name,
        'vsan-id'     =>  $Vsan_id
        'zone-name'   => $Zone_name
    );

(Class and Instance method)This will add a zone to zoneset.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'zoneset'=>$string >>

(Required) c<< name >> of the zoneset is required on which action will be performed

=item C<< 'zone-name'=>$string >>

Name of the zone which is to be added to zoneset

=back

=back

=cut

sub add_zone {

    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    $pkg_or_obj->_frontend_log_debug_opts(@_);
    my %opts = $pkg_or_obj->_common_validate_with(
        allow_extra     => 1,
        params          => \@_,
        additional_spec => { _optional_scalars( qw ('zone-name') ), },
    );
    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => "_add_zone_ciscofc_cli",
                interface => "CLI",
                set       => "CiscoFC"
            },
            {   method    => "_add_zone_brocadefc_cli",
                interface => "CLI",
                set       => "BrocadeFC"
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $ret;

}    #end of sub add_zone

sub _add_zone_brocadefc_cli {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );

    my $apiset            = delete $opts{apiset};
    my $command_interface = delete $opts{command_interface};
    $opts{'zone-name'} = "\"" . "$opts{'zone-name'}" . "\"";
    $opts{'zoneset'}   = "\"" . "$opts{'zoneset'}" . "\"";

    my %api_opts;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%api_opts,
        map    => {
            'zoneset'   => 'cfgname',
            'zone-name' => 'member',
        }
    );

    $pkg_or_obj->_api_call_log_debug_opts(
        api      => 'cfgadd',
        api_opts => \%api_opts
    );

    $apiset->cfgadd(%api_opts);

    $Log->exit() if $may_exit;

} ## end sub _add_zone_brocadefc_cli

sub _add_zone_ciscofc_cli {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );

    my $apiset            = delete $opts{apiset};
    my $command_interface = delete $opts{command_interface};

    my %api_opts;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%api_opts,
        copy   => [qw( 'zone-name' 'zoneset' 'vsan-id')],
    );

    $apiset->configure( subsystem => "terminal" );
    $apiset->zoneset_name(
        zoneset   => $opts{zoneset},
        'vsan-id' => $opts{'vsan-id'}
    );
    $apiset->member( 'zone-name' => $opts{'zone-name'} );
    $apiset->end();

    $Log->exit() if $may_exit;
} ## end sub _add_zone_ciscofc_cli

=head2 delete_zone

    NACL::C::Switch::Zone->delete_zone(
        command_interface   =>  $ci,
        'zoneset'    =>  $Zoneset_name,
        'vsan-id'     =>  $Vsan_id
        'zone-name'   => $Zone_name
    );

(Class and Instance method)This will delete a zone from zoneset.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'zoneset'=>$string >>

(Required) c<< name >> of the zoneset is required on which action will be performed

=item C<< 'zone-name'=>$string >>

Name of the zone which is to be deleted from the zoneset

=back

=back

=cut

sub delete_zone {

    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    $pkg_or_obj->_frontend_log_debug_opts(@_);
    my %opts = $pkg_or_obj->_common_validate_with(
        allow_extra     => 1,
        params          => \@_,
        additional_spec => { _optional_scalars( qw ('zone-name') ), },
    );
    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => "_delete_zone_ciscofc_cli",
                interface => "CLI",
                set       => "CiscoFC"
            },
            {   method    => "_remove_zone_brocadefc_cli",
                interface => "CLI",
                set       => "BrocadeFC"
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $ret;

}    #end of sub delete_zone

sub _remove_zone_brocadefc_cli {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );

    my $apiset            = delete $opts{apiset};
    my $command_interface = delete $opts{command_interface};
    $opts{'zone-name'} = "\"" . "$opts{'zone-name'}" . "\"";
    $opts{'zoneset'}   = "\"" . "$opts{'zoneset'}" . "\"";

    my %api_opts;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%api_opts,
        map    => {
            zoneset     => "cfgname",
            'zone-name' => "member",
        }
    );

    $pkg_or_obj->_api_call_log_debug_opts(
        api      => 'cfgremove',
        api_opts => \%api_opts
    );

    $apiset->cfgremove(%api_opts);

    $Log->exit() if $may_exit;

} ## end sub _remove_zone_brocadefc_cli

sub _delete_zone_ciscofc_cli {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );

    my $apiset            = delete $opts{apiset};
    my $command_interface = delete $opts{command_interface};
    my %api_opts;

    $apiset->configure( subsystem => "terminal" );
    $apiset->zoneset_name(
        'zoneset' => $opts{zoneset},
        'vsan-id' => $opts{'vsan-id'}
    );
    $apiset->member(
        'zone-name'   => $opts{'zone-name'},
        'deconfigure' => "1"
    );
    $apiset->end();

    $Log->exit() if $may_exit;
} ## end sub _delete_zone_ciscofc_cli

=head2 enable

    NACL::C::Switch::Zone->enable(
        command_interface   =>  $ci,
        'zoneset'    =>  $Zoneset_name,
    );

          or

   $zoneset_object->enable();

(Class or Instance method) request for enabling a zoneset

Uses a BrocadeFC-CLI, CiscoFC-CLI APISet

Invokes "zoneset activate name" command for CiscoFC CLI.

Invokes "cfgenable" command for BrocadeFC CLI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'zoneset'=>$string >>

(Required) c<< name >> of the zoneset is required on which action will be performed

Applicable for CiscoFC CLI, BrocadeFC CLI.

Maps to:

BrocadeFC CLI : cfg name( Programmatic option: 'cfg_name' )

CiscoFC CLI : zoneset name (Programmatic option: 'zoneset' )

=item C<< 'vsan-id'=>$string >>

(Required) c<< vsan-id >> on which the zoneset is present

Applicable for CiscoFC CLI.

Maps to:

CiscoFC CLI : vsan id ( Programmatic option: 'vsan-id' )

=back

=back

=cut

sub enable {

    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            zoneset => { type => SCALAR },
            _optional_scalars( qw (vsan-id) ),
        },
    );
    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => "_enable_ciscofc_cli",
                interface => "CLI",
                set       => "CiscoFC"
            },
            {   method    => "_enable_brocadefc_cli",
                interface => "CLI",
                set       => "BrocadeFC"
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $ret;

}    #end of sub enable

sub _enable_ciscofc_cli {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    my %opts = validate_with(
        params      => \@args,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );

    my $apiset          = $opts{apiset};
    my $prompts_answers = ['Do you want to continue?.*' => 'y' ];
    my %api_opts;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%api_opts,
        copy   => [qw( vsan-id zoneset )],
    );
    $api_opts{'connectrec-match_table'} = $prompts_answers;
    $apiset->configure( subsystem => "terminal" );
    $apiset->zoneset_activate_name(%api_opts);
    $apiset->end();

    $Log->exit() if $may_exit;

} ## end sub _enable_ciscofc_cli

sub _enable_brocadefc_cli {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    my %opts = validate_with(
        params      => \@args,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );

    my $apiset            = delete $opts{apiset};
    my $command_interface = delete $opts{command_interface};

    my %api_opts;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%api_opts,
        map    => { "zoneset" => "cfg_name", },
    );

    my $prompts_answers
        = [ '.*Do you want to enable.*configuration.*' => 'yes' ];
    $api_opts{'connectrec-match_table'} = $prompts_answers;

    $apiset->cfgenable(%api_opts);

    $Log->exit() if $may_exit;

} ## end sub _enable_brocadefc_cli

=head2 disable

    NACL::C::Switch::Zone->disable(
        command_interface   =>  $ci,
    );

          or

   $zoneset_object->disable();

(Class or Instance method) request for disable of a zoneset
For BrocadeFC current zoning configuration will be disabled.

Uses a BrocadeFC-CLI, CiscoFC-CLI APISet

Invokes "no zoneset activate name" command for CiscoFC CLI.

Invokes "cfgdisable" command for BrocadeFC CLI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'zoneset'=>$string >>

(Required) c<< name >> of the zoneset is required on which action will be performed

Applicable for CiscoFC CLI.

Maps to:

CiscoFC CLI : zoneset name (Programmatic option: 'zoneset' )

=item C<< 'vsan-id'=>$string >>

(Required) c<< vsan-id >> on which the zoneset is present

Applicable for CiscoFC CLI.

Maps to:

CiscoFC CLI : vsan id ( Programmatic option: 'vsan-id' )

=back

=back

=cut

sub disable {

    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            zoneset => { type => SCALAR , optional => 1},
            _optional_scalars( qw (vsan-id) ),
        },
        ignore_primary_keys => 1,
    );
    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => "_disable_ciscofc_cli",
                interface => "CLI",
                set       => "CiscoFC"
            },
            {   method    => "_disable_brocadefc_cli",
                interface => "CLI",
                set       => "BrocadeFC"
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $ret;

}    #end of sub disable

sub _disable_ciscofc_cli {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    my %opts = validate_with(
        params      => \@args,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );

    my $apiset = $opts{apiset};

    my %api_opts;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%api_opts,
        copy   => [qw( vsan-id zoneset )],
    );

    $apiset->configure( subsystem => "terminal" );
    $apiset->zoneset_activate_name( %api_opts, 'deconfigure' => "1" );
    $apiset->end();

    $Log->exit() if $may_exit;
} ## end sub _disable_ciscofc_cli

sub _disable_brocadefc_cli {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    my %opts = validate_with(
        params => \@args,
        spec   => $pkg_or_obj->_backend_validate_spec(),
    );

    my $apiset            = delete $opts{apiset};
    my $command_interface = delete $opts{command_interface};

    my %api_opts;
    my $prompts_answers
        = [ '.*Do you want to disable zoning configuration.*' => 'yes' ];
    $api_opts{'connectrec-match_table'} = $prompts_answers;

    $apiset->cfgdisable(%api_opts);

    $Log->exit() if $may_exit;

} ## end sub _disable_brocadefc_cli

sub _primary_keys_validate_spec {
    return ( 'zoneset' => { type => SCALAR }, );
}

1;
