# $Id$

#
# Copyright (c) 2001-2011 NetApp, Inc., All Rights Reserved
#  Any use, modification, or distribution is prohibited
#  without prior written consent from NetApp, Inc.
#

## @summary Cifs Component Module
## @author vikas.p@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::C::SysManager::Cifs

=head1 DESCRIPTION

C<NACL::C::SysManager::Cifs> is a derived class of
L<NACL::C::Component::SysManager|lib-NACL-C-Component-SysManager-pm>.  It represents access
to an ONTAP Cifs present on the System Manger GUI Screen.  A related class is
L<NACL::CS::SysManager::Cifs|lib-NACL-CS-SysManager-Cifs-pm>, which represents
the State Info of an ONTAP Cifs present on the System Manger GUI Screen.

=head1 ATTRIBUTES

=head2 command_interface

See L<NACL::C::Component::command_interface|lib-NACL-C-Component-pm/command_interface>

=head2 vserver

(Required for CMode, Not Applicable for 7Mode) The name of the vserver
in which the cifs is present.

=cut

package NACL::C::SysManager::Cifs;

use strict;
use warnings;
use Params::Validate
    qw(validate_with validate SCALAR ARRAYREF BOOLEAN UNDEF SCALARREF HASHREF);

use NACL::CS::SysManager::Cifs;
use NACL::ComponentUtils qw(_optional_scalars Dumper);
use NATE::BaseException qw(:try);
use NATE::Exceptions::Argument qw(:try);
use base 'NACL::C::Component::SysManager';
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();

=head1 METHODS

=head2 new

	my $cifs = NACL::C::SysManager::Cifs->new(command_interface => $ci,
											  vserver           => $vserver_name);

Return a new Cifs component object with the given attributes.

=cut

####################### Method Maker ###############################
use Class::MethodMaker [ scalar => 'vserver', ];

=head2 setup

	NACL::C::SysManager::Cifs->setup (command_interface => $ci,
									  vserver           => $vserver_name,...);

(Class method) Setup Cifs.  
Uses a  7Mode GUI or CMode GUI APISet.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>


=item C<< vserver => $string >>

(Required for CMode , Not Applicable for 7Mode)The name of the vserver
in which the cifs has to be setup.

=item C<< "vfiler" => $string >>

(Optional) Name of the VFiler

=item C<< "security-style" => $string >>

(Optional) Security Style for Cifs

=item C<< "windows-authentication" => $string >>

(Optional) Windows Authentication for Cifs

=item C<< "domain-name" => $string >>

(Optional) Domain Name

=item C<< "admin-name" => $string >>

(Optional) Admin Name

=item C<< "admin-password" => $string >>

(Optional) Admin Password

=item C<< "work-group" => $string >>

(Optional) Work Group

=item C<< "system-name" => $string >>

(Optional) System Name

=item C<< "bios-name" => $string >>

(Optional) BIOS Name

=item C<< "server-address" => [$server1, $server2, $server3, ... ] >>

(Optional) List of Server Address

=back

=back

=cut

sub setup {
    $Log->enter() if $may_enter;
    my ( $pkg, @args ) = @_;

    $pkg->_frontend_log_debug_opts(@args);

    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            "server-address" => { type => ARRAYREF, optional => 1 },
            "domain-name"    => { type => SCALAR },
            "admin-name"     => { type => SCALAR },
            "admin-password" => { type => SCALAR },
            _optional_scalars(
                qw (vfiler security-style windows-authentication
                    work-group system-name bios-name)
            ),
        },
    );

    my $ret = $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method => '_setup_7mode_gui',
                set    => 'SysManager_7Mode',
            },
            {   method => '_setup_cmode_gui',
                set    => 'SysManager_CMode',
            },
        ],
    );

    $Log->exit() if $may_exit;
    return $ret;
} ## end sub setup

####################### GUI 7Mode Setup ###############################
sub _setup_7mode_gui {
    $Log->enter() if $may_enter;
    my ( $pkg, @args ) = @_;
    my %opts = $pkg->_validate_backend_opts(@args);

    my $command_interface = $opts{command_interface};
    my $apiset            = $opts{apiset};

    my $cifs_args = $pkg->_hash_copy(
        source => \%opts,
        map    => {
            "vfiler"                 => "-vfiler",
            "security-style"         => "-secstyle",
            "windows-authentication" => "-windowsauth",
            "domain-name"            => "-domainname",
            "admin-name"             => "-adminname",
            "admin-password"         => "-adminpassword",
            "work-group"             => "-workgroup",
            "system-name"            => "-systemname",
            "server-address"         => "-serveraddress",
        }
    );

    $apiset->setup_cifs(%$cifs_args);
    $Log->exit() if $may_exit;

    return $pkg->new( command_interface => $command_interface );
} ## end sub _setup_7mode_gui

####################### GUI CMode Setup ###############################
sub _setup_cmode_gui {
    $Log->enter() if $may_enter;
    my ( $pkg, @args ) = @_;
    my %opts = $pkg->_validate_backend_opts(@args);

    my $command_interface = $opts{command_interface};
    my $apiset            = $opts{apiset};

    my $cifs_args = $pkg->_hash_copy(
        source => \%opts,
        map    => {
            "vserver"        => "-vservername",
            "bios-name"      => "-biosname",
            "domain-name"    => "-domainname",
            "admin-name"     => "-adminname",
            "admin-password" => "-adminpassword",
        }
    );

    $apiset->setup_cifs(%$cifs_args);
    $Log->exit() if $may_exit;

    return $pkg->new(
        vserver           => $cifs_args->{-vservername},
        command_interface => $command_interface
    );
} ## end sub _setup_cmode_gui

=head2 start

	NACL::C::SysManager::Cifs->start(command_interface => $ci,...);

                            or

	$cifs_object->start(...);

(Class or instance method) This method is used to Start Cifs.  
Uses a 7Mode GUI APISet.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=back

=back

=cut

sub start {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);

    my %opts = $pkg_or_obj->_common_validate_with( params => \@args, );

    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method => '_start_7mode_gui',
                set    => 'SysManager_7Mode',
            },
        ],
    );

    $Log->exit() if $may_exit;
    return $ret;
} ## end sub start

####################### GUI 7Mode Start ###############################
sub _start_7mode_gui {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    my %opts   = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = $opts{apiset};

    my $cifs_args = $pkg_or_obj->_hash_copy( source => \%opts, );

    $apiset->start_cifs_service(%$cifs_args);
    $Log->exit() if $may_exit;
} ## end sub _start_7mode_gui

=head2 stop

  NACL::C::SysManager::Cifs->stop (command_interface => $ci,...);

                            or

  $cifs_object->stop(...);

(Class or instance method)This method is used to Stop Cifs.  
Uses a 7Mode GUI APISet.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=back

=back

=cut

sub stop {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);

    my %opts = $pkg_or_obj->_common_validate_with( params => \@args, );

    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method => '_stop_7mode_gui',
                set    => 'SysManager_7Mode',
            },
        ],
    );

    $Log->exit() if $may_exit;
    return $ret;
} ## end sub stop

####################### GUI 7Mode Stop ###############################
sub _stop_7mode_gui {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    my %opts   = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = $opts{apiset};

    my $cifs_args = $pkg_or_obj->_hash_copy( source => \%opts, );

    $apiset->stop_cifs_service(%$cifs_args);
    $Log->exit() if $may_exit;
} ## end sub _stop_7mode_gui

=head2 lookup

	NACL::C::SysManager::Cifs->lookup(command_interface => $ci,...);

                            or

   $cifs_object->lookup(...);

(Class or instance method) This method is used to lookup Cifs.  
Uses a 7Mode GUI APISet.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "sid" => $string >>

(Optional) The ID for which the settings has to be lookup

=back

=back

=cut

sub lookup {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);

    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => { _optional_scalars(qw (sid)), },
    );

    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method => '_lookup_7mode_gui',
                set    => 'SysManager_7Mode',
            },
        ],
    );

    $Log->exit() if $may_exit;
    return $ret;
} ## end sub lookup

####################### GUI 7Mode Lookup ###############################
sub _lookup_7mode_gui {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    my %opts   = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = $opts{apiset};

    my $cifs_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        map    => { "sid" => "-sid", }
    );

    $apiset->cifs_lookup(%$cifs_args);
    $Log->exit() if $may_exit;
} ## end sub _lookup_7mode_gui

=head2 auditing_edit

	NACL::C::SysManager::Cifs->auditing_edit(command_interface => $ci,...);

                            or

	$cifs_object->auditing_edit(...);

(Class or instance method) This method is used to edit Cifs auditing.  
Uses a 7Mode GUI APISet.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "cifs-auditing" => $string >>

(Optional) Cifs Auditing

=item C<< "nfs-auditing" => $string >>

(Optional) NFS Auditing

=item C<< "filter-file" => $string >>

(Optional) Filter File

=item C<< "log-file-size" => $string >>

(Optional) Log File Size

=item C<< "log-file-size-unit" => $string >>

(Optional) Log File Size Unit

=item C<< "file-access" => $string >>

(Optional) File Access

=item C<< "logon-logoff" => $string >>

(Optional) Logon and Logoff option

=item C<< "account-management" => $string >>

(Optional) Account Management

=back

=back

=cut

sub auditing_edit {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);

    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw (cifs-auditing nfs-auditing filter-file
                    log-file-size log-file-size-unit file-access
                    logon-logoff account-management)
            ),
        },
    );

    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method => '_auditing_edit_7mode_gui',
                set    => 'SysManager_7Mode',
            },
        ],
    );

    $Log->exit() if $may_exit;
    return $ret;
} ## end sub auditing_edit

####################### GUI 7Mode Auditing Edit ###############################
sub _auditing_edit_7mode_gui {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    my %opts   = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = $opts{apiset};

    my $cifs_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        map    => {
            "cifs-auditing"      => "-cifsaudit",
            "nfs-auditing"       => "-nfsaudit",
            "filter-file"        => "-auditfile",
            "log-file-size"      => "-filesize",
            "log-file-size-unit" => "-filesizemode",
            "file-access"        => "-fileaccess",
            "logon-logoff"       => "-logonoff",
            "account-management" => "-accountmgmt",
        }
    );

    $apiset->edit_cifs_auditing_options(%$cifs_args);
    $Log->exit() if $may_exit;
} ## end sub _auditing_edit_7mode_gui

=head2 auditing_clear_log

	NACL::C::SysManager::Cifs->auditing_clear_log(command_interface => $ci,...);

                            or

	$cifs_object->auditing_clear_log(...);

(Class or instance method) This methos is used to clear Cifs auditing Log.  
Uses a 7Mode GUI APISet.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=back

=back

=cut

sub auditing_clear_log {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);

    my %opts = $pkg_or_obj->_common_validate_with( params => \@args, );

    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method => '_auditing_clear_log_7mode_gui',
                set    => 'SysManager_7Mode',
            },
        ],
    );

    $Log->exit() if $may_exit;
    return $ret;
} ## end sub auditing_clear_log

####################### GUI 7Mode Auditing Clear Log ###############################
sub _auditing_clear_log_7mode_gui {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    my %opts   = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = $opts{apiset};

    my $cifs_args = $pkg_or_obj->_hash_copy( source => \%opts, );

    $apiset->cifs_clear_log(%$cifs_args);
    $Log->exit() if $may_exit;
} ## end sub _auditing_clear_log_7mode_gui

=head2 edit_home_directory

	NACL::C::SysManager::Cifs->edit_home_directory(command_interface => $ci,...);

                            or

	$cifs_object->edit_home_directory(...);

(Class or instance method) This method is used to edit Cifs Home Directory.  
Uses a 7Mode GUI APISet.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< style => $string >>

=item C<< "add-home-directory" => [$hd1, $hd2, $hd3, ... ] >>

(Optional) List of Home Directory to be added

=item C<< "delete-home-directory" => [$hd1, $hd2, ... ] >>

(Optional) List of Home Directory to be deleted

=item C<< "style" => $string >>

(Optional) Style of the Home directory

=back

=back

=cut

sub edit_home_directory {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);

    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                "add-home-directory" => { type => ARRAYREF, optional => 1 },
                "delete-home-directory" =>
                    { type => ARRAYREF, optional => 1 },
                qw (style)
            ),
        },
    );

    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method => '_edit_home_directory_7mode_gui',
                set    => 'SysManager_7Mode',
            },
        ],
    );

    $Log->exit() if $may_exit;
    return $ret;
} ## end sub edit_home_directory

####################### GUI 7Mode Edit Home Directory ###############################
sub _edit_home_directory_7mode_gui {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    my %opts   = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = $opts{apiset};

    my $cifs_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        map    => {
            "style"                 => "-style",
            "add-home-directory"    => "-addhomepath",
            "delete-home-directory" => "-delhomepath",
        }
    );

    $apiset->edit_cifs_home_directory(%$cifs_args);
    $Log->exit() if $may_exit;
} ## end sub _edit_home_directory_7mode_gui

=head2 edit

	NACL::C::SysManager::Cifs->edit(command_interface => $ci,
								    vserver           => $vserver_name,...);

                            or

	$cifs_object->edit(...);

(Class or instance method) This method is used to edit Cifs Options.  
Uses a  7Mode GUI or CMode GUI APISet.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< vserver => $string >>

(Required for CMode , Not Applicable for 7Mode)The name of the vserver
in which the cifs has to be setup.

=item C<< "system-description" => $string >>

(Optional) System Description

=item C<< "session-timeout" => $string >>

(Optional) Session Timeout

=item C<< "concurrent-operations" => $string >>

(Optional) Concurrent Operations

=item C<< "snapshot-access" => $string >>

(Optional) Snapshot Access

=item C<< "support-older-client" => $string >>

(Optional) Support Older Client Option

=item C<< "wins-server" => $string >>

(Optional) Wins Server

=item C<< "netbios-aliases" => $string >>

(Optional) Net BIOS Aliases

=item C<< "netbios-over-tcp" => $string >>

(Optional) Net BIOS Over TCP Option

=item C<< "restrict-access" => $string >>

(Optional) Restrict Access Option

=item C<< "smb-signing" => $string >>

(Optional) SMB Signing

=item C<< "default-unix-user" => $string >>

(Optional) Default Unis User Field

=back

=back

=cut

sub edit {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);

    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw (system-description session-timeout concurrent-operations
                    snapshot-access support-older-client wins-server
                    netbios-aliases netbios-over-tcp restrict-access smb-signing
                    default-unix-user wins-delete-server)
            ),
        },
    );

    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method => '_edit_7mode_gui',
                set    => 'SysManager_7Mode',
            },
            {   method => '_edit_cmode_gui',
                set    => 'SysManager_CMode',
            },
        ],
    );

    $Log->exit() if $may_exit;
    return $ret;
} ## end sub edit

####################### GUI 7Mode Edit ###############################
sub _edit_7mode_gui {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    my %opts   = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = $opts{apiset};

    my $cifs_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        map    => {
            "system-description"    => "-systemdesc",
            "session-timeout"       => "-timeout",
            "concurrent-operations" => "-concurrentoper",
            "snapshot-access"       => "-enableaccess",
            "support-older-client"  => "-supportolder",
            "wins-server"           => "-winsip",
            "wins-delete-server"    => "-delwinsip",
            "netbios-aliases"       => "-alias",
            "netbios-over-tcp"      => "-enabletcp",
            "restrict-access"       => "-restrict",
            "smb-signing"           => "-enablesmb",
        }
    );

    $apiset->edit_cifs_options(%$cifs_args);
    $Log->exit() if $may_exit;
} ## end sub _edit_7mode_gui

####################### GUI CMode Edit ###############################
sub _edit_cmode_gui {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    my %opts   = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = $opts{apiset};

    my $cifs_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        map    => {
            "vserver"            => "-vservername",
            "default-unix-user"  => "-defaultunixuser",
            "wins-server"        => "-win_ip_address",
            "wins-delete-server" => "-delwinsip",
        }
    );

    $apiset->edit_cifs_options(%$cifs_args);
    $Log->exit() if $may_exit;
} ## end sub _edit_cmode_gui

=head2 manage_add

	NACL::C::SysManager::Cifs->manage_add(command_interface => $ci,
										  vserver           => $vserver_name,...);

                            or

	$cifs_object->manage_add(...);

(Class or instance method)This method is used to Add Manage Cifs.  
Uses a CMode GUI APISet.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< vserver => $string >>

The name of the vserver in which the cifs home directory add has to be done.

=item C<< "home-directory-path" => $string >>

(Optional) Home Directory path to be added

=back

=back

=cut

sub manage_add {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);

    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => { _optional_scalars(qw (home-directory-path)), },
    );

    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method => '_manage_add_cmode_gui',
                set    => 'SysManager_CMode',
            },
        ],
    );

    $Log->exit() if $may_exit;
    return $ret;
} ## end sub manage_add

####################### GUI CMode Manage Add ###############################
sub _manage_add_cmode_gui {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    my %opts   = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = $opts{apiset};

    my $cifs_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        map    => {
            "vserver"             => "-vservername",
            "home-directory-path" => "-homedirpath",
        }
    );

    $apiset->cifs_configuration_manage_add(%$cifs_args);
    $Log->exit() if $may_exit;
} ## end sub _manage_add_cmode_gui

=head2 manage_delete

	NACL::C::SysManager::Cifs->manage_delete(command_interface => $ci,
										     vserver           => $vserver_name,...);

                            or

	$cifs_object->manage_delete(...);

(Class or instance method) This method is used to delete Cifs Manage.  
Uses a CMode GUI APISet.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< vserver => $string >>

The name of the vserver in which the cifs home directory deletion has to be done.

=item C<< "home-directory-path" => $string >>

(Optional) Home Directory Path to be deleted

=back

=back

=cut

sub manage_delete {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);

    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => { _optional_scalars(qw (home-directory-path)), },
    );

    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method => '_manage_delete_cmode_gui',
                set    => 'SysManager_CMode',
            },
        ],
    );

    $Log->exit() if $may_exit;
    return $ret;
} ## end sub manage_delete

####################### GUI CMode Manage Delete ###############################
sub _manage_delete_cmode_gui {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    my %opts   = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = $opts{apiset};

    my $cifs_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        map    => {
            "vserver"             => "-vservername",
            "home-directory-path" => "-homedirpath",
        }
    );

    $apiset->cifs_configuration_manage_delete(%$cifs_args);
    $Log->exit() if $may_exit;
} ## end sub _manage_delete_cmode_gui

=head2 refresh

	NACL::C::SysManager::Cifs->refresh(command_interface => $ci,
									   vserver           => $vserver_name,...);

                            or

	$cifs_object->refresh(...);

(Class or instance method) refreshes Cifs page.  
Uses a CMode GUI or 7Mode GUI APISet.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< vserver => $string >>

(Required for class method, Not Applicable for instance method)
(Required for CMode , Not Applicable for 7Mode)
The name of the vserver.

=back

=back

=cut

sub refresh {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);

    my %opts = $pkg_or_obj->_common_validate_with( params => \@args, );

    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method => '_refresh_7mode_gui',
                set    => 'SysManager_7Mode',
            },
            {   method => '_refresh_cmode_gui',
                set    => 'SysManager_CMode',
            },
        ],
    );

    $Log->exit() if $may_exit;
    return $ret;
} ## end sub refresh

####################### GUI 7Mode Refresh ###############################
sub _refresh_7mode_gui {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    my %opts   = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = $opts{apiset};

    my $cifs_args = $pkg_or_obj->_hash_copy( source => \%opts, );

    $apiset->cifs_refresh(%$cifs_args);
    $Log->exit() if $may_exit;
} ## end sub _refresh_7mode_gui

####################### GUI CMode Refresh ###############################
sub _refresh_cmode_gui {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    my %opts   = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = $opts{apiset};

    my $cifs_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        map    => { "vserver" => "-vservername", }
    );

    $apiset->cifs_refresh(%$cifs_args);
    $Log->exit() if $may_exit;
} ## end sub _refresh_cmode_gui

####################### Primary Keys Validate ###############################
sub _primary_keys_validate_spec {
    return ( vserver => { type => SCALAR | UNDEF, optional => 1 }, );
}

1;
