# $Id$

#
# Copyright (c) 2001-2011 NetApp, Inc., All Rights Reserved
#  Any use, modification, or distribution is prohibited
#  without prior written consent from NetApp, Inc.
#

## @summary CifsSymlinks Component Module
## @author vikas.p@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::C::SysManager::CifsSymlinks

=head1 DESCRIPTION

C<NACL::C::SysManager::CifsSymlinks> is a derived class of
L<NACL::C::Component::SysManager|lib-NACL-C-Component-SysManager-pm>.  It represents access
to an ONTAP CifsSymlinks present on the System Manger GUI Screen.  A related class is
L<NACL::CS::SysManager::CifsSymlinks|lib-NACL-CS-SysManager-CifsSymlinks-pm>, which represents
the State Info of an ONTAP CifsSymlinks present on the System Manger GUI Screen.

=head1 ATTRIBUTES

=head2 command_interface

See L<NACL::C::Component::command_interface|lib-NACL-C-Component-pm/command_interface>

=head2 path

The unix path, used to identify it uniquely. (Unix path
names are assumed to be unique in a vserver in Cmode)

=head2 vserver

The name of the vserver in which the cifs symlinks is present.

=cut

package NACL::C::SysManager::CifsSymlinks;

use strict;
use warnings;
use Params::Validate
    qw(validate_with validate SCALAR ARRAYREF BOOLEAN UNDEF SCALARREF HASHREF);

use NACL::CS::SysManager::CifsSymlinks;
use NACL::ComponentUtils qw(_optional_scalars Dumper);
use NATE::BaseException qw(:try);
use NATE::Exceptions::Argument qw(:try);
use base 'NACL::C::Component::SysManager';
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();

=head1 METHODS

=head2 new

	my $cifs = NACL::C::SysManager::CifsSymlinks->new(command_interface => $ci,
											          vserver           => $vserver_name);

Return a new CifsSymlinks component object with the given attributes.

=cut

####################### Method Maker ###############################
use Class::MethodMaker [
    scalar => 'vserver',
    scalar => 'path',
];

=head2 create

	NACL::C::SysManager::CifsSymlinks->create(command_interface => $ci,
											           path              => "/",
                                                       vserver           => $vserver_name,...);

                            or

	$cifs_object->create(...);

(Class or instance method) This method is used to create Cifs Symlinks.  
Uses a CMode GUI APISet.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< vserver => $string >>

The name of the vserver in which the cifs Symlinks has to be created.

=item C<< "path" => $string >>

(Required) Unix path

=item C<< "cifs-share" => $string 

(Required) CifsSymlinks Share

=item C<< "cifs-path" => $string >>

(Required) CifsSymlinks Path

=item C<< "remote-server" => $string >>

(Optional) Remote Server

=back

=back

=cut

sub create {
    $Log->enter() if $may_enter;
    my ( $pkg, @args ) = @_;
    $pkg->_frontend_log_debug_opts(@args);

    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            "cifs-share" => { type => SCALAR },
            "cifs-path"  => { type => SCALAR },
            _optional_scalars(qw (remote-server)),
        },
    );

    my $ret = $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method => '_create_cmode_gui',
                set    => 'SysManager_CMode',
            },
        ],
    );

    $Log->exit() if $may_exit;
    return $ret;
} ## end sub create

####################### GUI CMode Symlink Create ###############################
sub _create_cmode_gui {
    $Log->enter() if $may_enter;
    my ( $pkg, @args ) = @_;
    my %opts = $pkg->_validate_backend_opts(@args);

    my $apiset            = $opts{apiset};
    my $command_interface = $opts{command_interface};

    my $cifs_args = $pkg->_hash_copy(
        source => \%opts,
        map    => {
            "vserver"       => "-vservername",
            "path"          => "-unix_path",
            "cifs-share"    => "-cifs_share",
            "cifs-path"     => "-cifs_path",
            "remote-server" => "-remote_server",
        }
    );

    $apiset->cifs_symlinks_create(%$cifs_args);
    $Log->exit() if $may_exit;

    return $pkg->new(
        path              => $cifs_args->{-unix_path},
        vserver           => $cifs_args->{-vservername},
        command_interface => $command_interface
    );
} ## end sub _create_cmode_gui

=head2 edit

	NACL::C::SysManager::CifsSymlinks->edit(command_interface => $ci,
										 vserver           => $vserver_name,...);

                            or

	$cifs_object->edit(...);

(Class or instance method) This method is used to edit Cifs Symlinks.  
Uses a CMode GUI APISet.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< vserver => $string >>

The name of the vserver in which the cifs Symlinks has to be edited.

=item C<< "path" => $string >>

(Required) Unix path

=item C<< "cifs-share" => $string 

(Required) CifsSymlinks Share

=item C<< "cifs-path" => $string >>

(Required) CifsSymlinks Path

=item C<< "remote-server" => $string >>

(Optional) Remote Server

=back

=back

=cut

sub edit {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);

    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            "cifs-share" => { type => SCALAR },
            "cifs-path"  => { type => SCALAR },
            _optional_scalars(qw (remote-server)),
        },
    );

    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method => '_edit_cmode_gui',
                set    => 'SysManager_CMode',
            },
        ],
    );

    $Log->exit() if $may_exit;
    return $ret;
} ## end sub edit

####################### GUI CMode Symlink Edit ###############################
sub _edit_cmode_gui {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    my %opts   = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = $opts{apiset};

    my $cifs_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        map    => {
            "vserver"       => "-vservername",
            "path"          => "-unix_path",
            "cifs-share"    => "-cifs_share",
            "cifs-path"     => "-cifs_path",
            "remote-server" => "-remote_server",
        }
    );

    $apiset->cifs_symlinks_edit(%$cifs_args);
    $Log->exit() if $may_exit;
} ## end sub _edit_cmode_gui

=head2 delete

	NACL::C::SysManager::CifsSymlinks->delete(command_interface => $ci,
											  vserver           => $vserver_name,...);

                            or

	$cifs_object->delete(...);

(Class or instance method) This method is used to delete Cifs Symlinks.  
Uses a CMode GUI APISet.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< vserver => $string >>

The name of the vserver in which the cifs Symlinks has to be deleted.

=item C<< "path" => $string >>

(Required) Unix path

=back

=back

=cut

sub delete {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);

    my %opts = $pkg_or_obj->_common_validate_with( params => \@args, );

    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method => '_delete_cmode_gui',
                set    => 'SysManager_CMode',
            },
        ],
    );

    $Log->exit() if $may_exit;
    return $ret;
} ## end sub delete

####################### GUI CMode Symlink Delete ###############################
sub _delete_cmode_gui {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    my %opts   = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = $opts{apiset};

    my $cifs_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        map    => {
            "vserver" => "-vservername",
            "path"    => "-unix_path",
        }
    );

    $apiset->cifs_symlinks_delete(%$cifs_args);
    $Log->exit() if $may_exit;
} ## end sub _delete_cmode_gui

=head2 refresh

	NACL::C::SysManager::CifsSymlinks->refresh(command_interface => $ci,
											vserver           => $vserver_name,...);

                            or

	$cifs_object->refresh(...);

(Class or instance method) This method is used to refresh Cifs Symlinks page.  
Uses a CMode GUI APISet.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< vserver => $string >>

The name of the vserver in which the cifs Symlinks page has to be refreshed.

=back

=back

=cut

sub refresh {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);

    my %opts = $pkg_or_obj->_common_validate_with( params => \@args, );

    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method => '_refresh_cmode_gui',
                set    => 'SysManager_CMode',
            },
        ],
    );

    $Log->exit() if $may_exit;
    return $ret;
} ## end sub refresh

####################### GUI CMode Symlink Refresh ###############################
sub _refresh_cmode_gui {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    my %opts   = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = $opts{apiset};

    my $cifs_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        map    => { "vserver" => "-vservername", }
    );

    $apiset->cifs_symlink_refresh(%$cifs_args);
    $Log->exit() if $may_exit;
} ## end sub _refresh_cmode_gui

####################### Primary Keys Validate ###############################
sub _primary_keys_validate_spec {
    return (
        path    => { type => SCALAR },
        vserver => { type => SCALAR },
    );
} ## end sub _primary_keys_validate_spec

1;
