# $Id$

#
# Copyright (c) 2001-2011 NetApp, Inc., All Rights Reserved
# Any use, modification, or distribution is prohibited
#  without prior written consent from NetApp, Inc.
#

## @summary Export Policy Component Module
## @author vikas.p@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::C::SysManager::ExportPolicy

=head1 DESCRIPTION

C<NACL::C::SysManager::ExportPolicy> is a derived class of
L<NACL::C::Component::SysManager|lib-NACL-C-Component-SysManager-pm>.  It represents access
to an ONTAP ExportPolicy present on the System Manger GUI Screen.  A related class is
L<NACL::CS::SysManager::ExportPolicy|lib-NACL-CS-SysManager-ExportPolicy-pm>, which represents
the State Info of an ONTAP ExportPolicy present on the System Manger GUI Screen.

=head1 ATTRIBUTES

=head2 command_interface

See L<NACL::C::Component::command_interface|lib-NACL-C-Component-pm/command_interface>

=head2 policy

The name of the policy, used to identify it uniquely.  (ExportPolicy
names are assumed to be unique in a vserver in Cmode)

=head2 vserver

The name of the vserver in which the export policy is present.

=cut

package NACL::C::SysManager::ExportPolicy;

use strict;
use warnings;
use Params::Validate
    qw(validate_with validate SCALAR ARRAYREF BOOLEAN UNDEF SCALARREF HASHREF);

use NACL::CS::SysManager::ExportPolicy;
use NACL::ComponentUtils qw(_optional_scalars Dumper);
use NATE::BaseException qw(:try);
use NATE::Exceptions::Argument qw(:try);
use base 'NACL::C::Component::SysManager';
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();

=head1 METHODS

=head2 new

    my $policy = NACL::C::SysManager::ExportPolicy->new(command_interface => $ci,
												        policy            => $pol_name,
												        vserver           => $vserver_name);

Return a new ExportPolicy component object with the given attributes.

=cut

use Class::MethodMaker [
    scalar => 'policy',
    scalar => 'vserver',
];

=head2 create

	my $exportPolicy  = 
		NACL::C::SysManager::ExportPolicy->create(command_interface => $ci,
												  policy            => $policy_name,
												  vserver           => $vserver_name);

(Class method) Request creation of a new exportPolicy element, and  return a new
exportPolicy component object that refers to it.
Uses CMode GUI APISet.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< policy=> $string >>

(Required) Name of exportPolicy.

=item C<< vserver=>$string >>

The name of the vserver in which the exportPolicy is present.

=item C<< "copy-rules"=> \$scalar >>

(Optional) Enable Copy rules 

=item C<< "source-policy"=> \$scalar >>

(Optional) Source of the policy

=item C<< "add-rules"=> \$scalar >>

(Optional) Rules to be added

=item C<< "client-specification"=> \$scalar >>

(Optional) Client specification option

=item C<< "protocol"=> \$scalar >>

(Optional) Protocol option

=item C<< "access-read-only"=> \$scalar >>

(Optional) Enable access read only

=item C<< "add-access-read-only-option"=> \$scalar >>

(Optional) add access read only options

=item C<< "access-read-write"=> \$scalar >>

(Optional) Enable access read write

=item C<< "add-access-read-write-option"=> \$scalar >>

(Optional) add access read write option only

=item C<< "access-superuser"=> \$scalar >>

(Optional) Enable access superuser

=item C<< "add-access-superuser-option"=> \$scalar >>

(Optional) add access superuser options 

=item C<< "edit-rules"=> \$scalar >>

(Optional) rules to be edited

=item C<< "edit-protocol"=> \$scalar >>

(Optional) protocols to be edited

=item C<< "edit-access-read-only-option"=> \$scalar >>

(Optional) access read only options to be edited

=item C<< "edit-access-read-write-option"=> \$scalar >>

(Optional) access read write options to be edited

=item C<< "edit-access-superuser-option"=> \$scalar >>

(Optional) access superuser options to be edited

=item C<< "delete-rules"=> \$scalar >>

(Optional) rules to be deleted

=back

=back

=cut

sub create {
    $Log->enter() if $may_enter;
    my $pkg = shift;
    $pkg->_frontend_log_debug_opts(@_);

    my %opts = $pkg->_common_validate_with(
        params          => \@_,
        additional_spec => {
            _optional_scalars qw(copy-rules source-policy add-rules
                client-specification protocol access-read-only
                add-access-read-only-option access-read-write
                add-access-read-write-option access-superuser
                add-access-superuser-option edit-rules edit-protocol
                edit-access-read-only-option edit-access-read-write-option
                edit-access-superuser-option delete-rules),
        },
    );

    my $ret = $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method => '_create_cmode_gui',
                set    => 'SysManager_CMode',
            },
        ],
    );

    $Log->exit() if $may_exit;
    return $ret;
} ## end sub create

####################### GUI CMode Create ###############################
sub _create_cmode_gui {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = $pkg->_validate_backend_opts(@_);

    my $command_interface = $opts{command_interface};
    my $apiset            = $opts{apiset};

    my $exp_pol_args = $pkg->_hash_copy(
        source => \%opts,
        map    => {
            "policy"                        => "-policyName",
            "vserver"                       => "-vservername",
            "copy-rules"                    => "-copyrules",
            "source-policy"                 => "-sourcepolicy",
            "add-rules"                     => "-addrules",
            "client-specification"          => "-clientspec",
            "protocol"                      => "-protocol",
            "access-read-only"              => "-readonly",
            "add-access-read-only-option"   => "-addreadonlyopt",
            "access-read-write"             => "-readwrite",
            "add-access-read-write-option"  => "-addreadwriteopt",
            "access-superuser"              => "-superuser",
            "add-access-superuser-option"   => "-addsuperuseropt",
            "edit-rules"                    => "-editrules",
            "edit-protocol"                 => "-editprotocol",
            "edit-access-read-only-option"  => "-editreadonlyopt",
            "edit-access-read-write-option" => "-editreadwriteopt",
            "edit-access-superuser-option"  => "-editsuperuseropt",
            "delete-rules"                  => "-deleterules",
        }
    );

    $apiset->export_policy_create(%$exp_pol_args);
    $Log->exit() if $may_exit;

    return $pkg->new(
        policy            => $exp_pol_args->{-policyName},
        vserver           => $exp_pol_args->{-vservername},
        command_interface => $command_interface
    );

} ## end sub _create_cmode_gui

=head2 delete

	NACL::C::SysManager::ExportPolicy->delete(command_interface => $ci,
											  exportPolicy      => $policy_name,
											  vserver           => $vserver_name);

                            or

	$export_policy_object->delete(...);

(Class or instance method)This method deletes exportPolicy.
Uses CMode GUI APISet.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< exportPolicy=> $string >>

(Required) Name of exportPolicy.

=item C<< vserver=>$string >>

The name of the vserver in which the exportPolicy is present.

=item C<< "confirm-delete"=> \$scalar >>

(Optional) Enable Delete Confirmation

=back

=back

=cut

sub delete {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    $pkg_or_obj->_frontend_log_debug_opts(@_);

    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@_,
        additional_spec => { _optional_scalars qw(confirm-delete), },
    );

    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method => '_delete_cmode_gui',
                set    => 'SysManager_CMode',
            },
        ],
    );

    $Log->exit() if $may_exit;
    return $ret;
} ## end sub delete

####################### GUI CMode Delete ###############################
sub _delete_cmode_gui {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = $pkg_or_obj->_validate_backend_opts(@_);
    my $apiset     = $opts{apiset};

    my $exp_pol_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        map    => {
            "policy"         => "-policyName",
            "vserver"        => "-vservername",
            "confirm-delete" => "-checkbox",
        }
    );

    $apiset->export_policy_delete(%$exp_pol_args);
    $Log->exit() if $may_exit;
} ## end sub _delete_cmode_gui

=head2 rename

  NACL::C::SysManager::ExportPolicy->rename(command_interface => $ci,
									        exportPolicy      =>$policy_name,
										    vserver           => $vserver_name);

                            or

  $export_policy_object->rename(...);

(Class or instance method)This method rename a exportPolicy element.
Uses CMode GUI APISet.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< exportPolicy=> $string >>

(Required) Name of exportPolicy.

=item C<< vserver=>$string >>

The name of the vserver in which the exportPolicy is present.

=item C<< "rename"=> \$scalar >>

(Optional) The Export policy new name.

=back

=back

=cut

sub rename {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    $pkg_or_obj->_frontend_log_debug_opts(@_);

    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@_,
        additional_spec => { _optional_scalars qw(rename), },
    );

    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method => '_rename_cmode_gui',
                set    => 'SysManager_CMode',
            },
        ],
    );

    $Log->exit() if $may_exit;
    return $ret;
} ## end sub rename

####################### GUI CMode Rename ###############################
sub _rename_cmode_gui {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = $pkg_or_obj->_validate_backend_opts(@_);
    my $apiset     = $opts{apiset};

    my $exp_pol_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        map    => {
            "policy"  => "-policyName",
            "vserver" => "-vservername",
            "rename"  => "-rename",
        }
    );

    $apiset->export_policy_rename(%$exp_pol_args);
    $Log->exit() if $may_exit;
} ## end sub _rename_cmode_gui

=head2 refresh

  NACL::C::SysManager::ExportPolicy->refresh(command_interface => $ci,
									         vserver           => $vserver_name,...);

                            or

  $export_policy_object->refresh();

(Class or instance method) This method refreshs a exportPolicy page.
Uses CMode GUI APIset.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< vserver=>$string >>

The name of the vserver in which the exportPolicy is present.

=back

=back

=cut

sub refresh {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    $pkg_or_obj->_frontend_log_debug_opts(@_);

    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@_,
        additional_spec => { _optional_scalars qw(vserver), },
    );

    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method => '_refresh_cmode_gui',
                set    => 'SysManager_CMode',
            },
        ],
    );

    $Log->exit() if $may_exit;
    return $ret;
} ## end sub refresh

####################### GUI CMode Refresh ###############################
sub _refresh_cmode_gui {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = $pkg_or_obj->_validate_backend_opts(@_);
    my $apiset     = $opts{apiset};

    my $exp_pol_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        map    => { "vserver" => "-vservername", }
    );

    $apiset->export_policy_refresh(%$exp_pol_args);
    $Log->exit() if $may_exit;
} ## end sub _refresh_cmode_gui

####################### Primary Keys Validate ###############################
sub _primary_keys_validate_spec {
    return (
        policy  => { type => SCALAR },
        vserver => { type => SCALAR },
    );
} ## end sub _primary_keys_validate_spec

1
