# $Id$

#
# Copyright (c) 2001-2010 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Fcp Module
## @author Neethu.George@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::C::SysManager::Fcp

=head1 DESCRIPTION


C<NACL::C::SysManager::Fcp> is a derived class of
L<NACL::C::Component::SysManager|lib-NACL-C-Component-SysManager-pm>.  It represents access
to ONTAP FCP present on the System Manger GUI Screen.  A related class is
L<NACL::CS::SysManager::Fcp|lib-NACL-CS-SysManager-Fcp-pm>, which represents
the State Info of ONTAP Fcp present on the System Manger GUI Screen.

=head1 ATTRIBUTES

=head2 command_interface

See L<NACL::C::Component::command_interface|lib-NACL-C-Component-pm/command_interface>

=head2  fcpname

Name of the fc protocol.

=head2 vserver

(Required for CMode, Not Applicable for 7Mode)The name of the vserver
in which the FCP is present.

=cut

package NACL::C::SysManager::Fcp;

use strict;
use warnings;
use Params::Validate
    qw(validate_with validate SCALAR ARRAYREF BOOLEAN UNDEF SCALARREF HASHREF);

use NACL::CS::SysManager::Fcp;
use NACL::ComponentUtils qw(_optional_scalars Dumper);
use NATE::BaseException qw(:try);
use NATE::Exceptions::Argument qw(:try);
use base 'NACL::C::Component::SysManager';
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();

=head1 METHODS

=head2 new

  my $fcp_obj  = NACL::C::SysManager::Fcp->new(command_interface => $ci,
											   fcpname           => $fcpname,
                                               vserver           => $vserver_name );

Return a new FCP component object with the given attributes.

=cut

####################### Method Maker ###############################

use Class::MethodMaker [
    scalar => 'fcpname',
    scalar => 'vserver',
];

=head2 edit

  my $fcp_obj  = NACL::C::SysManager::Fcp->edit(command_interface => $ci,
												fcpname           => $fcpname,
                                                vserver           => $vserver_name);

(Class method) Edit fcp element, and  return a new
fcp component object that refers to it.
Uses a CMode GUI or 7mode GUI APISet.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< vserver => $string >>

(Required for CMode, Not Applicable for 7Mode)The name of the vserver.

=item C<< fcpname  => $string >>

(Required) Name of new fc protocol.

=back

=back

=cut

sub edit {
    $Log->enter() if $may_enter;
    my $pkg = shift;

    $pkg->_frontend_log_debug_opts(@_);

    my %opts = $pkg->_common_validate_with( params => \@_, );

    my $ret = $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method => '_edit_7mode_gui',
                set    => 'SysManager_7Mode',
            },
            {   method => '_edit_cmode_gui',
                set    => 'SysManager_CMode',
            },
        ],
    );

    $Log->exit() if $may_exit;
    return $ret;

} ## end sub edit
####################### GUI 7Mode edit ###############################
sub _edit_7mode_gui {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = $pkg->_validate_backend_opts(@_);

    my $command_interface = $opts{command_interface};
    my $apiset            = $opts{apiset};

    my $fcp_args = $pkg->_hash_copy(
        source => \%opts,
        map    => { "fcpname" => "-fcpname", }
    );

    $apiset->fcp_edit(%$fcp_args);
    $Log->exit() if $may_exit;

    return $pkg->new(
        fcpname           => $fcp_args->{-fcpname},
        command_interface => $command_interface
    );
} ## end sub _edit_7mode_gui

####################### GUI CMode create ###############################
sub _edit_cmode_gui {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = $pkg->_validate_backend_opts(@_);

    my $command_interface = $opts{command_interface};
    my $apiset            = $opts{apiset};

    my $fcp_args = $pkg->_hash_copy(
        source => \%opts,
        map    => {
            "vserver" => "-vservername",
            "fcpname" => "-fcpname"
        }
    );

    $apiset->fcp_edit(%$fcp_args);
    $Log->exit() if $may_exit;

    return $pkg->new(
        fcpname           => $fcp_args->{-fcpname},
        vserver           => $fcp_args->{-vservername},
        command_interface => $command_interface
    );

} ## end sub _edit_cmode_gui

=head2 start

  NACL::C::SysManager::Fcp->start(command_interface => $ci,);

                            or

  $fcp_obj->start();

(Class or instance method) This starts fcp service.
Uses a 7Mode GUI or CMode GUI APIset.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< vserver => $string >>

(Required for class method, Not Applicable for instance method)
(Required for CMode, Not Applicable for 7Mode)The name of the vserver

=back

=back

=cut

sub start {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    $pkg_or_obj->_frontend_log_debug_opts(@_);

    my %opts = $pkg_or_obj->_common_validate_with( params => \@_, );

    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method => '_start_7mode_gui',
                set    => 'SysManager_7Mode',
            },
            {   method => '_start_cmode_gui',
                set    => 'SysManager_CMode',
            },
        ],
    );

    $Log->exit() if $may_exit;
    return $ret;

} ## end sub start

####################### GUI 7Mode start ###############################
sub _start_7mode_gui {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = $pkg_or_obj->_validate_backend_opts(@_);
    my $apiset     = $opts{apiset};

    my $fcp_args = $pkg_or_obj->_hash_copy( source => \%opts, );

    $apiset->fcp_status( -status => "start", %$fcp_args );
    $Log->exit() if $may_exit;
} ## end sub _start_7mode_gui

####################### GUI CMode start ###############################
sub _start_cmode_gui {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = $pkg_or_obj->_validate_backend_opts(@_);
    my $apiset     = $opts{apiset};

    my $fcp_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        map    => { "vserver" => "-vservername" }
    );

    $apiset->fcp_status( -status => "start", %$fcp_args );
    $Log->exit() if $may_exit;
} ## end sub _start_cmode_gui

=head2 stop

  NACL::C::SysManager::Fcp->stop(command_interface => $ci,);

                            or

  $fcp_obj->stopt();

(Class or instance method) This stop fcp service .
Uses a 7Mode GUI or CMode GUI APIset.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< vserver => $string >>

(Required for class method, Not Applicable for instance method)
(Required for CMode, Not Applicable for 7Mode)The name of the vserver

=back

=back

=cut

sub stop {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    $pkg_or_obj->_frontend_log_debug_opts(@_);

    my %opts = $pkg_or_obj->_common_validate_with( params => \@_, );

    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method => '_stop_7mode_gui',
                set    => 'SysManager_7Mode',
            },
            {   method => '_stop_cmode_gui',
                set    => 'SysManager_CMode',
            },
        ],
    );

    $Log->exit() if $may_exit;
    return $ret;

} ## end sub stop

####################### GUI 7Mode stop ###############################
sub _stop_7mode_gui {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = $pkg_or_obj->_validate_backend_opts(@_);
    my $apiset     = $opts{apiset};

    my $fcp_args = $pkg_or_obj->_hash_copy( source => \%opts, );

    $apiset->fcp_status( -status => "stop", %$fcp_args );
    $Log->exit() if $may_exit;
} ## end sub _stop_7mode_gui

####################### GUI CMode stop ###############################
sub _stop_cmode_gui {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = $pkg_or_obj->_validate_backend_opts(@_);
    my $apiset     = $opts{apiset};

    my $fcp_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        map    => { "vserver" => "-vservername" }
    );

    $apiset->fcp_status( -status => "stop", %$fcp_args );
    $Log->exit() if $may_exit;
} ## end sub _stop_cmode_gui

=head2 refresh

  NACL::C::SysManager::Fcp->refresh(command_interface => $ci,);

                            or


  $fcp_obj->refresh();

(Class or instance method) This method refreshs fcp page.
Uses a 7Mode GUI or CMode GUI APIset.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< vserver => $string >>

(Required for class method, Not Applicable for instance method)
(Required for CMode, Not Applicable for 7Mode)The name of the vserver

=back

=back

=cut

sub refresh {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    $pkg_or_obj->_frontend_log_debug_opts(@_);

    my %opts = $pkg_or_obj->_common_validate_with( params => \@_, );

    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method => '_refresh_7mode_gui',
                set    => 'SysManager_7Mode',
            },
            {   method => '_refresh_cmode_gui',
                set    => 'SysManager_CMode',
            },
        ],
    );

    $Log->exit() if $may_exit;
    return $ret;
} ## end sub refresh

####################### GUI 7Mode Refresh ###############################
sub _refresh_7mode_gui {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = $pkg_or_obj->_validate_backend_opts(@_);
    my $apiset     = $opts{apiset};

    my $fcp_args = $pkg_or_obj->_hash_copy( source => \%opts, );

    $apiset->fcp_refresh();
    $Log->exit() if $may_exit;
} ## end sub _refresh_7mode_gui

####################### GUI CMode Refresh ###############################
sub _refresh_cmode_gui {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = $pkg_or_obj->_validate_backend_opts(@_);
    my $apiset     = $opts{apiset};

    my $fcp_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        map    => { "vserver" => "-vservername" }
    );

    $apiset->refresh_fcp(%$fcp_args);
    $Log->exit() if $may_exit;
} ## end sub _refresh_cmode_gui

####################### Primary Keys Validate ###############################
sub _primary_keys_validate_spec {
    return (
        fcpname => { type => SCALAR },
        vserver => { type => SCALAR | UNDEF, optional => 1 },
    );
} ## end sub _primary_keys_validate_spec

1;
