# $Id$

#
# Copyright (c) 2001-2010 NetApp, Inc., All Rights Reserved
#  Any use, modification, or distribution is prohibited
#  without prior written consent from NetApp, Inc.
#

## @summary IscsiTargetPortalGroup Module
## @author Neethu.George@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::C::SysManager::IscsiTargetPortalGroup

=head1 DESCRIPTION

C<NACL::C::SysManager::IscsiTargetPortalGroup> is a derived class of
L<NACL::C::Component::SysManager|lib-NACL-C-Component-SysManager-pm>.  It represents access
to an ONTAP Target Portal Group present on the System Manger GUI Screen. A related class is
L<NACL::CS::SysManager::IscsiTargetPortalGroup|lib-NACL-CS-SysManager-IscsiTargetPortalGroup-pm>, which represents
the State Info of an ONTAP IscsiTargetPortalGroup present on the System Manger GUI Screen.

=head1 ATTRIBUTES

=head2 command_interface

See L<NACL::C::Component::command_interface|lib-NACL-C-Component-pm/command_interface>

=head2  tpgname

Name of the target portal group, used to identify it uniquely. (TPG
names are assumed to be unique within a node or within a cluster in 7mode
and in a vserver in Cmode)

=head2 vserver

(Required for CMode, Not Applicable for 7Mode) The name of the vserver
in which the TPG is present. 

=cut

package NACL::C::SysManager::IscsiTargetPortalGroup;

use strict;
use warnings;
use Params::Validate
    qw(validate_with validate SCALAR ARRAYREF BOOLEAN UNDEF SCALARREF HASHREF);

use NACL::CS::SysManager::IscsiTargetPortalGroup;
use NACL::ComponentUtils qw(_optional_scalars Dumper);
use NATE::BaseException qw(:try);
use NATE::Exceptions::Argument qw(:try);
use base 'NACL::C::Component::SysManager';
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();

=head1 METHODS

=head2 new

	my $tpg_obj  = NACL::C::SysManager::IscsiTargetPortalGroup->new(command_interface => $ci,
																	vserver           => $vserver_name,
																	tpgname           => $tpgname );
								 
Return a new target portal group component object with the given attributes.

=cut

####################### Method Maker ###############################

use Class::MethodMaker [
    scalar => 'tpgname',
    scalar => 'vserver',
];

=head2 create

	my $tpg_obj  = 
		NACL::C::SysManager::IscsiTargetPortalGroup->create(command_interface => $ci,
															vserver           => $vserver_name,
															tpgname           => $tpgname);

(Class method) Request creation of a new target portal group element, and return a new
tpg component object that refers to it.
Uses a CMode GUI or 7mode GUI APISet.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< vserver => $string >>

(Required for CMode, Not Applicable for 7Mode)The name of the vserver
in which the tpg to be created.

=item C<< tpgname  => $string >>

(Required) Name of target portal group to be created.

=item C<< tag  => $string >>

(Optional) Tag value for tpg group.

=item C<< interface  => $string >>

(Optional) Name of interface to be added in the tpg.

=back

=back

=cut

sub create {
    $Log->enter() if $may_enter;
    my ( $pkg, @args ) = @_;
    $pkg->_frontend_log_debug_opts(@args);

    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => { _optional_scalars(qw ( tag interface)) }
    );

    my $ret = $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method => '_create_7mode_gui',
                set    => 'SysManager_7Mode',
            },
            {   method => '_create_cmode_gui',
                set    => 'SysManager_CMode',
            },
        ],
    );

    $Log->exit() if $may_exit;
    return $ret;

} ## end sub create
####################### GUI 7Mode create ###############################
sub _create_7mode_gui {
    $Log->enter() if $may_enter;
    my ( $pkg, @args ) = @_;
    my %opts = $pkg->_validate_backend_opts(@args);

    my $command_interface = $opts{command_interface};
    my $apiset            = $opts{apiset};

    my $tpg_args = $pkg->_hash_copy(
        source => \%opts,
        map    => {
            "tpgname"   => "-tpgname",
            "tag"       => "-tag",
            "interface" => "-interface"
        }
    );

    $apiset->tpg_create(%$tpg_args);

    $Log->exit() if $may_exit;

    return $pkg->new(
        tpgname           => $tpg_args->{-tpgname},
        command_interface => $command_interface
    );
} ## end sub _create_7mode_gui

####################### GUI CMode create ###############################
sub _create_cmode_gui {
    $Log->enter() if $may_enter;
    my ( $pkg, @args ) = @_;
    my %opts = $pkg->_validate_backend_opts(@args);

    my $command_interface = $opts{command_interface};
    my $apiset            = $opts{apiset};

    my $tpg_args = $pkg->_hash_copy(
        source => \%opts,
        map    => {
            "vserver"   => "-vservername",
            "tpgname"   => "-tpgname",
            "tag"       => "-tag",
            "interface" => "-interface"
        }
    );

    $apiset->tpg_create(%$tpg_args);

    $Log->exit() if $may_exit;

    return $pkg->new(
        tpgname           => $tpg_args->{-tpgname},
        vserver           => $tpg_args->{-vservername},
        command_interface => $command_interface
    );

} ## end sub _create_cmode_gui

=head2 edit

	NACL::C::SysManager::IscsiTargetPortalGroup->edit(command_interface => $ci,
													  vserver           => $vserver_name,
												      tpgname           => $tpgname);

                            or

	$tpg_obj->edit(...);

(Class or instance method) Modify a target portal group.  
Uses a CMode GUI or a 7Mode GUI APISet.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< vserver => $string >>

(Required for class method, Not Applicable for instance method.)
(Required for CMode, Not Applicable for 7Mode).
Name of vserver that contains the tpg.

=item C<< tpgname => $string >>

(Required for class method, Not Applicable for instance method)
The tpg  to be edited.

=item C<< interface-add => $string >>

(Optional) Interface to be added to the tpg.

=item C<< interface-deleted => $string >>

(Optional) Interface to be deleted from the tpg.

=back

=back

=cut

sub edit {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);

    my %opts = $pkg_or_obj->_common_validate_with(
        params => \@args,
        additional_spec =>
            { _optional_scalars(qw ( interface-add interface-delete )) }
    );

    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method => '_edit_7mode_gui',
                set    => 'SysManager_7Mode',
            },
            {   method => '_edit_cmode_gui',
                set    => 'SysManager_CMode',
            },
        ],
    );

    $Log->exit() if $may_exit;
    return $ret;

} ## end sub edit
####################### GUI 7Mode edit ###############################
sub _edit_7mode_gui {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    my %opts       = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset     = $opts{apiset};

    my $tpg_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        map    => {
            "tpgname"          => "-tpgname",
            "interface-add"    => "-interfaceadd",
            "interface-delete" => "-interfaceremove"
        }
    );

    $apiset->tpg_edit(%$tpg_args);
    $Log->exit() if $may_exit;
} ## end sub _edit_7mode_gui

####################### GUI CMode edit ###############################
sub _edit_cmode_gui {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    my %opts       = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset     = $opts{apiset};

    my $tpg_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        map    => {
            "vserver"          => "-vservername",
            "tpgname"          => "-tpgname",
            "interface-add"    => "-interfaceadd",
            "interface-delete" => "-interfaceremove"
        }
    );

    $apiset->tpg_edit(%$tpg_args);
    $Log->exit() if $may_exit;
} ## end sub _edit_cmode_gui

=head2 delete

	NACL::C::SysManager::IscsiTargetPortalGroup->delete(command_interface => $ci,
														vserver           => $vserver_name,
												        tpgname           => $tpgname);

                            or

  $tpg_obj->delete(...);

(Class or instance method) Delete a target portal group.  
Uses a CMode GUI or a 7Mode GUI APISet.


=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< vserver => $string >>

(Required for class method, Not Applicable for instance method.)
(Required for CMode, Not Applicable for 7Mode).
Name of vserver that contains the tpg.

=item C<< tpgname => $string >>

(Required for class method, Not Applicable for instance method)
The tpg  to be deleted.

=back

=back

=cut

sub delete {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);

    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => { tpgarray => { type => ARRAYREF }, }
    );

    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method => '_delete_7mode_gui',
                set    => 'SysManager_7Mode',
            },
            {   method => '_delete_cmode_gui',
                set    => 'SysManager_CMode',
            },
        ],
    );

    $Log->exit() if $may_exit;
    return $ret;

} ## end sub delete

####################### GUI 7Mode delete ###############################
sub _delete_7mode_gui {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    my %opts       = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset     = $opts{apiset};

    my $tpg_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        map    => { "tpgarray" => "-tpgarray", }
    );

    $apiset->tpg_delete(%$tpg_args);
    $Log->exit() if $may_exit;
} ## end sub _delete_7mode_gui

####################### GUI CMode delete ###############################
sub _delete_cmode_gui {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    my %opts       = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset     = $opts{apiset};

    my $tpg_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        map    => {
            "vserver"  => "-vservername",
            "tpgarray" => "-tpgarray",
        }
    );

    $apiset->tpg_delete(%$tpg_args);
    $Log->exit() if $may_exit;
} ## end sub _delete_cmode_gui

=head2 refresh

	NACL::C::SysManager::IscsiTargetPortalGroup->refresh(command_interface => $ci,
														 vserver           => $vserver_name,);

                            or

	$tpg_obj->refresh(...);

(Class or instance method) This method refreshs tpg page.
Uses a 7Mode GUI or CMode GUI APIset.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< vserver => $string >>

(Required for class method, Not Applicable for instance method)
(Required for CMode, Not Applicable for 7Mode)The name of the vserver
in which the tpg is present.

=back

=back

=cut

sub refresh {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);

    my %opts = $pkg_or_obj->_common_validate_with( params => \@args, );

    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method => '_refresh_7mode_gui',
                set    => 'SysManager_7Mode',
            },
            {   method => '_refresh_cmode_gui',
                set    => 'SysManager_CMode',
            },
        ],
    );

    $Log->exit() if $may_exit;
    return $ret;

} ## end sub refresh

####################### GUI 7Mode Refresh ###############################
sub _refresh_7mode_gui {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    my %opts       = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset     = $opts{apiset};

    my $tpg_args = $pkg_or_obj->_hash_copy( source => \%opts, );

    $apiset->tpg_refresh(%$tpg_args);
    $Log->exit() if $may_exit;
} ## end sub _refresh_7mode_gui

####################### GUI CMode Refresh ###############################
sub _refresh_cmode_gui {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    my %opts       = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset     = $opts{apiset};

    my $tpg_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        map    => { "vserver" => "-vservername" }
    );

    $apiset->tpg_refresh(%$tpg_args);
    $Log->exit() if $may_exit;
} ## end sub _refresh_cmode_gui

####################### Primary Keys Validate ###############################
sub _primary_keys_validate_spec {
    return (
        tpgname => { type => SCALAR },
        vserver => { type => SCALAR | UNDEF, optional => 1 },
    );
} ## end sub _primary_keys_validate_spec

1;
