# $Id$

#
# Copyright (c) 2001-2011 NetApp, Inc., All Rights Reserved
# Any use, modification, or distribution is prohibited
#  without prior written consent from NetApp, Inc.
#

## @summary License Component Module
## @author vikas.p@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::C::SysManager::License

=head1 DESCRIPTION

C<NACL::C::SysManager::License> is a derived class of
L<NACL::C::Component::SysManager|lib-NACL-C-Component-SysManager-pm>.  It represents access
to an ONTAP License present on the System Manger GUI Screen.  A related class is
L<NACL::CS::SysManager::License|lib-NACL-CS-SysManager-License-pm>, which represents
the State Info of an ONTAP License present on the System Manger GUI Screen.

=head1 ATTRIBUTES

=head2 command_interface

See L<NACL::C::Component::command_interface|lib-NACL-C-Component-pm/command_interface>

=head2 licenses

The List of license keys, used to identify it uniquely. (License
keys are assumed to be unique in 7mode and Cmode)

=cut

package NACL::C::SysManager::License;

use strict;
use warnings;
use Params::Validate
    qw(validate_with validate SCALAR ARRAYREF BOOLEAN UNDEF SCALARREF HASHREF);

use NACL::CS::SysManager::License;
use NACL::ComponentUtils qw(_optional_scalars Dumper);
use NATE::BaseException qw(:try);
use NATE::Exceptions::Argument qw(:try);
use base 'NACL::C::Component::SysManager';
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();

=head1 METHODS

=head2 new

	my $License = NACL::C::SysManager::License->new(command_interface => $ci,
													license           => [$license_keys]);

Return a new License component object with the given attributes.

=cut

####################### Method Maker ###############################

use Class::MethodMaker [ 
    scalar => 'license', ];

=head2 add

	my $License_object = NACL::C::SysManager::License->add(command_interface => $ci,
														   License           => [$license_keys]);

(Class method) Request addition of a new License elements, and return a new
License component object that refers to it.
Uses a 7Mode GUI or CMode GUI APISet.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< license => [$License1, $License2, $License3, ... ] >>

(Required) List of License Keys to be added.

=back

=back

=cut

####################### GUI Add ###############################
sub add {
    $Log->enter() if $may_enter;
    my ( $pkg, @args ) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    
    $Log->debug("Opts to 'add' front end are: @args");

    my %opts = $pkg->_common_validate_with( params => \@args, );

    my $ret = $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method => '_add_7mode_gui',
                set    => 'SysManager_7Mode',
            },
            {   method => '_add_cmode_gui',
                set    => 'SysManager_CMode',
            },
        ],
    );

    $Log->exit() if $may_exit;
    return $ret;

} ## end sub add

####################### GUI 7Mode Add ###############################
sub _add_7mode_gui {
    $Log->enter() if $may_enter;
    my ( $pkg, @args ) = @_;
    my %opts = $pkg->_validate_backend_opts(@args);

    my $command_interface = $opts{command_interface};
    my $apiset            = $opts{apiset};

    my $license_args = $pkg->_hash_copy(
        source => \%opts,
        map    => { "license" => "-licenses", }
    );

    $apiset->license_add(%$license_args);
    $Log->exit() if $may_exit;

    return $pkg->new(
        license           => $license_args->{-licenses},
        command_interface => $command_interface
    );
} ## end sub _add_7mode_gui

####################### GUI CMode Add ###############################
sub _add_cmode_gui {
    $Log->enter() if $may_enter;
    my ( $pkg, @args ) = @_;
    my %opts = $pkg->_validate_backend_opts(@args);

    my $command_interface = $opts{command_interface};
    my $apiset            = $opts{apiset};

    my $license_args = $pkg->_hash_copy(
        source => \%opts,
        map    => { "license" => "-licenses", }
    );

    $apiset->license_add(%$license_args);
    $Log->exit() if $may_exit;

    return $pkg->new(
        license           => $license_args->{-licenses},
        command_interface => $command_interface
    );

} ## end sub _add_cmode_gui

=head2 delete

	NACL::C::SysManager::License->delete(command_interface => $ci,
									     license           => [$license_keys]);

                            or

	$License_object->delete(...);

(Class or instance method)This method deletes License elements.
Uses a 7Mode GUI or CMode GUI APISet.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< license => [$License1, $License2, $License3, ... ] >>

(Required for class method, Not Applicable for instance method)
License Names to be deleted.


=back

=back

=cut

####################### GUI Delete ###############################
sub delete {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);

    my %opts = $pkg_or_obj->_common_validate_with( params => \@args, );

    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method => '_delete_7mode_gui',
                set    => 'SysManager_7Mode',
            },
            {   method => '_delete_cmode_gui',
                set    => 'SysManager_CMode',
            },
        ],
    );

    $Log->exit() if $may_exit;
    return $ret;
} ## end sub delete

####################### GUI 7Mode Delete ###############################
sub _delete_7mode_gui {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    my %opts   = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = $opts{apiset};

    my $license_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        map    => { "license" => "-licenses", }
    );

    $apiset->license_delete(%$license_args);
    $Log->exit() if $may_exit;
} ## end sub _delete_7mode_gui

####################### GUI CMode Delete ###############################
sub _delete_cmode_gui {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    my %opts   = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = $opts{apiset};

    my $license_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        map    => { "license" => "-licenses", }
    );

    $apiset->license_delete(%$license_args);
    $Log->exit() if $may_exit;
} ## end sub _delete_cmode_gui

=head2 enable

	NACL::C::SysManager::License->enable(command_interface => $ci,
									     license           => [$license_keys]);

                            or

	$License_object->enable(...);

(Class or instance method) This method enables all License elements.
Uses a 7Mode GUI APIset.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< license => [$License1, $License2, $License3, ... ] >>

(Required for class method, Not Applicable for instance method)
License Names to be enabled.


=back

=back

=cut

####################### GUI Enable ###############################
sub enable {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);

    my %opts = $pkg_or_obj->_common_validate_with( params => \@args, );

    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method => '_enable_7mode_gui',
                set    => 'SysManager_7Mode',
            },
        ],
    );

    $Log->exit() if $may_exit;
    return $ret;
} ## end sub enable

####################### GUI 7Mode Enable ###############################
sub _enable_7mode_gui {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    my %opts   = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = $opts{apiset};

    my $license_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        map    => { "license" => "-licenses", }
    );

    $apiset->license_enable(%$license_args);
    $Log->exit() if $may_exit;
} ## end sub _enable_7mode_gui

=head2 disable

	NACL::C::SysManager::License->disable(command_interface => $ci,
									      license           => [$license_keys]);

                            or

	$License_object->disable(...);

(Class or instance method) This method disables all License elements.
Uses 7Mode GUI APIset.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< license => [$License1, $License2, $License3, ... ] >>

(Required for class method, Not Applicable for instance method)
License Names to be disabled.

=back

=back

=cut

####################### GUI Disable ###############################
sub disable {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);

    my %opts = $pkg_or_obj->_common_validate_with( params => \@args, );

    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method => '_disable_7mode_gui',
                set    => 'SysManager_7Mode',
            },
        ],
    );

    $Log->exit() if $may_exit;
    return $ret;
} ## end sub disable

####################### GUI 7Mode Disable ###############################
sub _disable_7mode_gui {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    my %opts   = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = $opts{apiset};

    my $license_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        map    => { "license" => "-licenses", }
    );

    $apiset->license_disable(%$license_args);
    $Log->exit() if $may_exit;
} ## end sub _disable_7mode_gui

=head2 refresh

  NACL::C::SysManager::License->refresh(command_interface => $ci);

                            or

  $License_object->refresh();

(Class or instance method) This method refreshs a License page.
Uses a 7Mode GUI or CMode GUI APIset.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=back

=back

=cut

####################### GUI Refresh ###############################
sub refresh {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);

    my %opts = $pkg_or_obj->_common_validate_with( params => \@args, );

    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method => '_refresh_7mode_gui',
                set    => 'SysManager_7Mode',
            },
            {   method => '_refresh_cmode_gui',
                set    => 'SysManager_CMode',
            },
        ],
    );

    $Log->exit() if $may_exit;
    return $ret;
} ## end sub refresh

####################### GUI 7Mode Refresh ###############################
sub _refresh_7mode_gui {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    my %opts   = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = $opts{apiset};

    my $license_args = $pkg_or_obj->_hash_copy( source => \%opts, );

    $apiset->license_refresh(%$license_args);
    $Log->exit() if $may_exit;
} ## end sub _refresh_7mode_gui

####################### GUI CMode Refresh ###############################
sub _refresh_cmode_gui {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    my %opts   = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = $opts{apiset};

    my $license_args = $pkg_or_obj->_hash_copy( source => \%opts, );

    $apiset->license_refresh(%$license_args);
    $Log->exit() if $may_exit;
} ## end sub _refresh_cmode_gui

####################### Primary Keys Validate ###############################
sub _primary_keys_validate_spec {
    return ( license => { type => SCALAR }, );
}

1;
