# $Id$

#
# Copyright (c) 2001-2011 NetApp, Inc., All Rights Reserved
# Any use, modification, or distribution is prohibited
#  without prior written consent from NetApp, Inc.
#

## @summary NFS Component Module
## @author vikas.p@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::C::SysManager::NFS

=head1 DESCRIPTION

C<NACL::C::SysManager::NFS> is a derived class of
L<NACL::C::Component::SysManager|lib-NACL-C-Component-SysManager-pm>.  It represents access
to ONTAP NFS present on the System Manger GUI Screen. A related class is
L<NACL::CS::SysManager::NFS|lib-NACL-CS-SysManager-NFS-pm>, which represents
the State Info of ONTAP NFS configured on the system manager GUI screen.

=head1 ATTRIBUTES

=head2 command_interface

See L<NACL::C::Component::command_interface|lib-NACL-C-Component-pm/command_interface>

=head2 vserver

(Required for CMode, Not Applicable for 7Mode) The name of the vserver
in which the NFS has to be modified.

=cut

package NACL::C::SysManager::NFS;

use strict;
use warnings;
use Params::Validate
    qw(validate_with validate SCALAR ARRAYREF BOOLEAN UNDEF SCALARREF HASHREF);

use NACL::CS::SysManager::NFS;
use NACL::ComponentUtils qw(_optional_scalars Dumper);
use NATE::BaseException qw(:try);
use NATE::Exceptions::Argument qw(:try);
use base 'NACL::C::Component::SysManager';
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();

=head1 METHODS

=head2 new

	my $nfs = NACL::C::SysManager::NFS->new(command_interface => $ci,
											vserver           => $vserver_name);

Return a new NFS component object with the given attributes.

=cut

####################### Method Maker ###############################

use Class::MethodMaker [ scalar => 'vserver', ];

=head2 edit

	NACL::C::SysManager::NFS->edit(command_interface => $ci,
								 vserver           => $vserver_name,...);

(Class method or instance method) Modify a NFS element. 
Uses a 7Mode GUI or CMode GUI APISet.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< vserver => $string >>

(Required for CMode , Not Applicable for 7Mode)The name of the vserver
in which the NFS has to be edited.

=item C<< "support-version3" => $string >>

(Optional) Enable Support Version 3

=item C<< "support-version4.0" => $string >>

(Optional) Enable Support Version 4.0

=item C<< "acls" => $string >>

(Optional) Enable ACLs

=item C<< "read-delegation" => $string >>

(Optional) Enable Read Delegation

=item C<< "write-delegation" => $string >>

(Optional) Enable Write Delegation

=item C<< "support-version2" => $string >>

(Optional) Enable Support Version 2

=item C<< "support-version4.1" => $string >>

(Optional) Enable Support Version 4.1

=item C<< "default-windows-user" => $string >>

(Optional) Default Windows User option

=back

=back

=cut

####################### GUI Edit ###############################
sub edit {
    $Log->enter() if $may_enter;
    my ( $pkg, @args ) = @_;

    $pkg->_frontend_log_debug_opts(@args);

    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw (support-version3 support-version4.0 acls
                    read-delegation write-delegation
                    support-version2 support-version4.1
                    default-windows-user)
            ),
        },
    );

    my $ret = $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method => '_edit_7mode_gui',
                set    => 'SysManager_7Mode',
            },
            {   method => '_edit_cmode_gui',
                set    => 'SysManager_CMode',
            },
        ],
    );

    $Log->exit() if $may_exit;
    return $ret;
} ## end sub edit

####################### GUI 7Mode Edit ###############################
sub _edit_7mode_gui {
    $Log->enter() if $may_enter;
    my ( $pkg, @args ) = @_;
    my %opts = $pkg->_validate_backend_opts(@args);

    my $command_interface = $opts{command_interface};
    my $apiset            = $opts{apiset};

    my $nfs_args = $pkg->_hash_copy(
        source => \%opts,
        map    => {
            "support-version3"   => "-supportnfs3",
            "support-version4.0" => "-supportnfs4",
            "acls"               => "-supportacl",
            "read-delegation"    => "-supportreaddelegate",
            "write-delegation"   => "-supportwritedelegate",
        }
    );

    $apiset->edit_nfs_settings(%$nfs_args);
    $Log->exit() if $may_exit;

    return $pkg->new(
        vserver           => $nfs_args->{-vservername},
        command_interface => $command_interface
    );
} ## end sub _edit_7mode_gui

####################### GUI CMode Edit ###############################
sub _edit_cmode_gui {
    $Log->enter() if $may_enter;
    my ( $pkg, @args ) = @_;
    my %opts = $pkg->_validate_backend_opts(@args);

    my $command_interface = $opts{command_interface};
    my $apiset            = $opts{apiset};

    my $nfs_args = $pkg->_hash_copy(
        source => \%opts,
        map    => {
            "vserver"              => "-vservername",
            "support-version2"     => "-support_version2",
            "support-version3"     => "-support_version3",
            "support-version4.0"   => "-support_version4_0",
            "acls"                 => "-nfs_acl",
            "read-delegation"      => "-nfs_read_delagation",
            "write-delegation"     => "-nfs_write_delegation",
            "support-version4.1"   => "-support_version4_1",
            "default-windows-user" => "-windows_user",
        }
    );

    $apiset->nfs_edit(%$nfs_args);
    $Log->exit() if $may_exit;

    return $pkg->new( command_interface => $command_interface );
} ## end sub _edit_cmode_gui

=head2 disable

	NACL::C::SysManager::NFS->disable(command_interface => $ci,
									  vserver           => $vserver_name,...);

                            or

	$nfs_object->disable(...);

(Class or instance method) This method disables nfs.
Uses a CMode GUI or 7Mode GUI APIset.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< vserver =>$string >>

(Required for class method, Not Applicable for instance method)
(Required for CMode , Not Applicable for 7Mode)The name of the vserver
in which the NFS has to be disabled.

=back

=back

=cut

####################### GUI Disable ###############################
sub disable {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);

    my %opts = $pkg_or_obj->_common_validate_with( params => \@args, );

    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method => '_disable_7mode_gui',
                set    => 'SysManager_7Mode',
            },
            {   method => '_disable_cmode_gui',
                set    => 'SysManager_CMode',
            },
        ],
    );

    $Log->exit() if $may_exit;
    return $ret;
} ## end sub disable

####################### GUI 7Mode Disable ###############################
sub _disable_7mode_gui {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    my %opts   = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = $opts{apiset};

    my $nfs_args = $pkg_or_obj->_hash_copy( source => \%opts, );

    $apiset->nfs_disable(%$nfs_args);
    $Log->exit() if $may_exit;
} ## end sub _disable_7mode_gui

####################### GUI CMode Disable ###############################
sub _disable_cmode_gui {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    my %opts   = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = $opts{apiset};

    my $nfs_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        map    => { "vserver" => "-vservername", }
    );

    $apiset->nfs_disable(%$nfs_args);
    $Log->exit() if $may_exit;
} ## end sub _disable_cmode_gui

=head2 enable

	NACL::C::SysManager::NFS->enable(command_interface => $ci,
									 vserver           => $vserver_name,...);

                            or

	$nfs_object->enable(...);

(Class or instance method) This method enables nfs.
Uses a CMode GUI or 7Mode GUI APIset.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< vserver =>$string >>

(Required for class method, Not Applicable for instance method)
(Required for CMode , Not Applicable for 7Mode)The name of the vserver
in which the NFS has to be enabled.

=back

=back

=cut

####################### GUI Enable ###############################
sub enable {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);

    my %opts = $pkg_or_obj->_common_validate_with( params => \@args, );

    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method => '_enable_7mode_gui',
                set    => 'SysManager_7Mode',
            },
            {   method => '_enable_cmode_gui',
                set    => 'SysManager_CMode',
            },
        ],
    );

    $Log->exit() if $may_exit;
    return $ret;
} ## end sub enable

####################### GUI 7Mode Disable ###############################
sub _enable_7mode_gui {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    my %opts   = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = $opts{apiset};

    my $nfs_args = $pkg_or_obj->_hash_copy( source => \%opts, );

    $apiset->nfs_enable(%$nfs_args);
    $Log->exit() if $may_exit;
} ## end sub _enable_7mode_gui

####################### GUI CMode Enable ###############################
sub _enable_cmode_gui {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    my %opts   = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = $opts{apiset};

    my $nfs_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        map    => { "vserver" => "-vservername", }
    );

    $apiset->nfs_enable(%$nfs_args);
    $Log->exit() if $may_exit;
} ## end sub _enable_cmode_gui

=head2 refresh

	NACL::C::SysManager::NFS->refresh(command_interface => $ci,
									  vserver           => $vserver_name,...);

                            or

  $nfs_object->refresh(...);

(Class or instance method) This method refresh nfs page.
Uses a CMode GUI or 7Mode GUI APIset.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< vserver =>$string >>

(Required for class method, Not Applicable for instance method)
(Required for CMode , Not Applicable for 7Mode)The name of the vserver
in which the NFS page has to be refreshed.

=back

=back

=cut

####################### GUI Refresh ###############################
sub refresh {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);

    my %opts = $pkg_or_obj->_common_validate_with( params => \@args, );

    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method => '_refresh_7mode_gui',
                set    => 'SysManager_7Mode',
            },
            {   method => '_refresh_cmode_gui',
                set    => 'SysManager_CMode',
            },
        ],
    );

    $Log->exit() if $may_exit;
    return $ret;
} ## end sub refresh

####################### GUI 7Mode Refresh ###############################
sub _refresh_7mode_gui {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    my %opts   = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = $opts{apiset};

    my $nfs_args = $pkg_or_obj->_hash_copy( source => \%opts, );

    $apiset->nfs_refresh(%$nfs_args);
    $Log->exit() if $may_exit;
} ## end sub _refresh_7mode_gui

####################### GUI CMode Refresh ###############################
sub _refresh_cmode_gui {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    my %opts   = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = $opts{apiset};

    my $nfs_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        map    => { "vserver" => "-vservername", }
    );

    $apiset->nfs_refresh(%$nfs_args);
    $Log->exit() if $may_exit;
} ## end sub _refresh_cmode_gui

####################### Primary Keys Validate ###############################
sub _primary_keys_validate_spec {
    return ( vserver => { type => SCALAR | UNDEF, optional => 1 }, );
}

1;
