# $Id$

#
# Copyright (c) 2001-2011 NetApp, Inc., All Rights Reserved
# Any use, modification, or distribution is prohibited
#  without prior written consent from NetApp, Inc.
#

## @summary NetworkInterface Component Module
## @author Neethu.George@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::C::SysManager::NetworkInterface

=head1 DESCRIPTION

C<NACL::C::SysManager::NetworkInterface> is a derived class of
L<NACL::C::Component::SysManager|lib-NACL-C-Component-SysManager-pm>.  It represents access
to an ONTAP NetworkInterface present on the System Manger GUI Screen.  A related class is
L<NACL::CS::SysManager::NetworkInterface|lib-NACL-CS-SysManager-NetworkInterface-pm>, which represents
the State Info of an ONTAP NetworkInterface present on the System Manger GUI Screen.


=head1 ATTRIBUTES

=head2 command_interface

See L<NACL::C::Component::command_interface|lib-NACL-C-Component-pm/command_interface

=head2 interfacename

The name of the Interface, used to identify it uniquely. (Interface
names are assumed to be unique within a node or within a cluster in 7mode
and in a vserver in Cmode)

=head2 vserver

(Required for CMode, Not Applicable for 7Mode) The name of the vserver
in which the Interface is present. 

=cut

package NACL::C::SysManager::NetworkInterface;

use strict;
use warnings;
use Params::Validate
    qw(validate_with validate SCALAR ARRAYREF BOOLEAN UNDEF SCALARREF HASHREF);

use NACL::CS::SysManager::NetworkInterface;
use NACL::ComponentUtils qw(_optional_scalars Dumper);
use NATE::BaseException qw(:try);
use NATE::Exceptions::Argument qw(:try);
use base 'NACL::C::Component::SysManager';
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();

=head1 METHODS

=head2 new

  my $Network_Obj = NACL::C::SysManager::NetworkInterface->new(command_interface => $ci,
															   interfacename     => $interface,
															   vserver           => $vserver_name,);

Return a new NetworkInterface component object with the given attributes.

=cut

####################### Method Maker ###############################

use Class::MethodMaker [
    scalar => 'interfacename',
    scalar => 'vserver',
];

=head2 create

	my $Network_Obj = 
		NACL::C::SysManager::NetworkInterface->create(command_interface => $ci,
													  vserver           => $vserver_name,
													  interfacename     => $interface);

(Class method) Request creation of a new network interface element, 
and return a new network interface component object that refers to it.
Uses a CMode GUI APISet.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< vserver => $string >>

(Required for CMode, Not Applicable for 7Mode)

=item C<< "interfacename" => $string >>

(Required) Name of the network interface to be created.

=item C<< "netmask" => $string >>

(Required) Netmask name.

=item C<< "homeport" => $string >>

(Required) Home Port for the interface

=item C<< "ipaddress" => $string >>

(Required)IP Address for the Interface

=item C<< "subprotocol" => $string >>

(Optional) Sub protocols to be enabled

=item C<< "role" => $string >>

(Optional) Role to be added

=item C<< "protocol" => $string >>

(Optional) Protocols to be enabled

=item C<< "gateway" => $string >>

(Optional) Gateway for the Interface

=back

=back

=cut

sub create {
    $Log->enter() if $may_enter;
    my ( $pkg, @args ) = @_;

    $pkg->_frontend_log_debug_opts(@args);

    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            homeport  => { type => SCALAR },
            ipaddress => { type => SCALAR },
            netmask   => { type => SCALAR },
            _optional_scalars(qw (subprotocol role protocol gateway))
        }
    );

    my $ret = $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method => '_create_cmode_gui',
                set    => 'SysManager_CMode',
            },
        ],
    );

    $Log->exit() if $may_exit;
    return $ret;

} ## end sub create

####################### GUI CMode Create ###############################
sub _create_cmode_gui {
    $Log->enter() if $may_enter;
    my ( $pkg, @args ) = @_;
    my %opts = $pkg->_validate_backend_opts(@args);

    my $command_interface = $opts{command_interface};
    my $apiset            = $opts{apiset};

    my $network_args = $pkg->_hash_copy(
        source => \%opts,
        map    => {
            "vserver"       => "-vservername",
            "interfacename" => "-interfacename",
            "netmask"       => "-netmask",
            "ipaddress"     => "-ipaddress",
            "homeport"      => "-homeport",
            "subprotocol"   => "-subprotocol",
            "role"          => "-role",
            "protocol"      => "-protocol",
            "gateway"       => "-gateway"
        }
    );

    $apiset->create_network_interface(%$network_args);
    $Log->exit() if $may_exit;

    return $pkg->new(
        interfacename     => $network_args->{-interfacename},
        vserver           => $network_args->{-vservername},
        command_interface => $command_interface
    );

} ## end sub _create_cmode_gui

=head2 edit

	NACL::C::SysManager::NetworkInterface->edit(command_interface => $ci,
												vserver           => $vserver_name,
												interfacename     => $interface);
				or 		
				
	$Network_Obj->edit(...);					 

(Class or instance method)This method edits network interface.
Uses a CMode GUI APISet.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< vserver => $string >>

(Required for instance method, Optional for class method)
The Name of the Vserver

=item C<< "interfacename" => $string >>

(Required for instance method, Optional for class method)
Name of the network interface to be edited.

=item C<< "netmask" => $string >>

(Required) Netmask name.

=item C<< "homeport" => $string >>

(Required) Home port of the Interface

=item C<< "ipaddress" => $string >>

(Required) IP Adress for the Interface

=item C<< "gateway" => $string >>

(Optional) Default gateway.

=back

=back

=cut

sub edit {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);

    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            homeport  => { type => SCALAR },
            ipaddress => { type => SCALAR },
            netmask   => { type => SCALAR },
            _optional_scalars(qw (gateway))
        }
    );

    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method => '_edit_cmode_gui',
                set    => 'SysManager_CMode',
            },
        ],
    );

    $Log->exit() if $may_exit;
    return $ret;

} ## end sub edit

####################### GUI CMode Edit ###############################
sub _edit_cmode_gui {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    my %opts   = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = $opts{apiset};

    my $network_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        map    => {
            "vserver"       => "-vservername",
            "interfacename" => "-interfacename",
            "netmask"       => "-netmask",
            "ipaddress"     => "-ipaddress",
            "homeport"      => "-homeport",
            "gateway"       => "-gateway"
        }
    );

    $apiset->edit_network_interface(%$network_args);
    $Log->exit() if $may_exit;

} ## end sub _edit_cmode_gui

=head2 delete

	NACL::C::SysManager::NetworkInterface->delete(command_interface => $ci,
												  vserver           => $vserver_name,
												  interfacename     => $interface,);
				or 		
	
	$Network_Obj->delete();					 

(Class or instance method)This method deletes a network interface element.
Uses a CMode GUI APISet.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< vserver => $string >>

(Required for class method, Not Applicable for instance method)
The name of the vserver

=item C<< "interfacename" => $string >>

(Required for instance method, Optional for class method) 
Name of the network interface to be deleted.

=back

=back

=cut

sub delete {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);

    my %opts = $pkg_or_obj->_common_validate_with( params => \@args, );

    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method => '_delete_cmode_gui',
                set    => 'SysManager_CMode',
            },
        ],
    );

    $Log->exit() if $may_exit;
    return $ret;

} ## end sub delete

####################### GUI CMode Delete ###############################
sub _delete_cmode_gui {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    my %opts   = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = $opts{apiset};

    my $network_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        map    => {
            "vserver"       => "-vservername",
            "interfacename" => "-interfacename"
        }
    );

    $apiset->delete_network_interface(%$network_args);
    $Log->exit() if $may_exit;

} ## end sub _delete_cmode_gui

=head2 enable

	NACL::C::SysManager::NetworkInterface->enable(command_interface => $ci,
												  vserver           => $vserver_name,
												  interfacename     => $interface,);
												  
				or 		
	
	$Network_Obj->enable();					 

(Class or instance method) This method enables a network interface.
Uses a CMode GUI APISet.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< vserver => $string >>

(Required for class method, Not Applicable for instance method)
The name of the vserver

=item C<< "interfacename" => $string >>

(Required for instance method, Optional for class method) 
Name of the network interface to be enabled.

=back

=back

=cut

sub enable {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);

    my %opts = $pkg_or_obj->_common_validate_with( params => \@args, );

    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method => '_enable_cmode_gui',
                set    => 'SysManager_CMode',
            },
        ],
    );

    $Log->exit() if $may_exit;
    return $ret;

} ## end sub enable

####################### GUI CMode Enable ###############################
sub _enable_cmode_gui {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    my %opts   = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = $opts{apiset};

    my $network_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        map    => {
            "vserver"       => "-vservername",
            "interfacename" => "-interfacename"
        }
    );

    $apiset->enable_network_interface(%$network_args);
    $Log->exit() if $may_exit;

} ## end sub _enable_cmode_gui

=head2 disable

	NACL::C::SysManager::NetworkInterface->disable(command_interface => $ci,
												   vserver           => $vserver_name,
												   interfacename     => $interface,);
	
				or 		
	
	$Network_Obj->disable();					 

(Class or instance method)This method disables a network interface.
Uses a CMode GUI APISet.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< vserver => $string >>

(Required for class method, Not Applicable for instance method)
The name of the vserver

=item C<< "interfacename" => $string >>

(Required for instance method, Optional for class method) 
Name of the network interface to be disabled.

=back

=back

=cut

sub disable {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);

    my %opts = $pkg_or_obj->_common_validate_with( params => \@args, );

    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method => '_disable_cmode_gui',
                set    => 'SysManager_CMode',
            },
        ],
    );

    $Log->exit() if $may_exit;
    return $ret;

} ## end sub disable

####################### GUI CMode Disable ###############################
sub _disable_cmode_gui {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    my %opts   = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = $opts{apiset};

    my $network_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        map    => {
            "vserver"       => "-vservername",
            "interfacename" => "-interfacename"
        }
    );

    $apiset->disable_network_interface(%$network_args);
    $Log->exit() if $may_exit;

} ## end sub _disable_cmode_gui

=head2 refresh

	NACL::C::SysManager::NetworkInterface->refresh(command_interface => $ci,
									   );

                            or

	$Network_Object->refresh();

(Class or instance method)This method refreshs network interface page

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=back

=back

=cut

sub refresh {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);

    my %opts = $pkg_or_obj->_common_validate_with( params => \@args, );

    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method => '_refresh_cmode_gui',
                set    => 'SysManager_CMode',
            },
        ],
    );

    $Log->exit() if $may_exit;
    return $ret;

} ## end sub refresh

####################### GUI CMode Refresh ###############################
sub _refresh_cmode_gui {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    my %opts   = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = $opts{apiset};

    my $network_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        map    => {
            "vserver"       => "-vservername",
        }
    );

    $apiset->refresh_network_interface(%$network_args);
    $Log->exit() if $may_exit;
} ## end sub _refresh_cmode_gui

=head2 sendhome
	
    NACL::C::SysManager::NetworkInterface->sendhome(command_interface => $ci,
													vserver           => $vserver_name,
													interfacename     => $interface,);

                            or

	$Network_Object->sendhome();

(Class or instance method)Clicks send_to_home buttong from network interface page

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< vserver => $string >>

(Required for class method, Not Applicable for instance method)
The name of the vserver

=item C<< "interfacename" => $string >>

(Required for instance method, Optional for class method) 
Name of the network interface to be sent Home.

=back

=back

=cut

sub sendhome {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);

    my %opts = $pkg_or_obj->_common_validate_with( params => \@args, );

    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method => '_sendhome_cmode_gui',
                set    => 'SysManager_CMode',
            },
        ],
    );

    $Log->exit() if $may_exit;
    return $ret;

} ## end sub sendhome

####################### GUI CMode sendhome ###############################
sub _sendhome_cmode_gui {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    my %opts   = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = $opts{apiset};

    my $network_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        map    => {
            "vserver"       => "-vservername",
            "interfacename" => "-interfacename"
        }
    );

    $apiset->send_to_home(%$network_args);
    $Log->exit() if $may_exit;

} ## end sub _sendhome_cmode_gui

##################################################################################
sub _primary_keys_validate_spec {
    return (
        interfacename => { type => SCALAR },
        vserver       => { type => SCALAR },
    );
} ## end sub _primary_keys_validate_spec

1;
