# $Id$

#
# Copyright (c) 2001-2011 NetApp, Inc., All Rights Reserved
#  Any use, modification, or distribution is prohibited
#  without prior written consent from NetApp, Inc.
#

## @summary Portset Component Module
## @author vikas.p@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::C::SysManager::Portset

=head1 DESCRIPTION

C<NACL::C::SysManager::Portset> is a derived class of
L<NACL::C::Component::SysManager|lib-NACL-C-Component-SysManager-pm>.  It represents access
to an ONTAP Portset present on the System Manger GUI Screen.  A related class is
L<NACL::CS::SysManager::Portset|lib-NACL-CS-SysManager-Portset-pm>, which represents
the State Info of an ONTAP Portset present on the System Manger GUI Screen.

=head1 ATTRIBUTES

=head2 command_interface

See L<NACL::C::Component::command_interface|lib-NACL-C-Component-pm/command_interface>

=head2 portset

The name of the portset, used to identify it uniquely.  (Portset
names are assumed to be unique)

=head2 vserver

The name of the vserver in which the portset is present.

=cut

package NACL::C::SysManager::Portset;

use strict;
use warnings;
use Params::Validate
    qw(validate_with validate SCALAR ARRAYREF BOOLEAN UNDEF SCALARREF HASHREF);

use NACL::CS::SysManager::Portset;
use NACL::ComponentUtils qw(_optional_scalars Dumper);
use NATE::BaseException qw(:try);
use NATE::Exceptions::Argument qw(:try);
use base 'NACL::C::Component::SysManager';
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();

=head1 METHODS

=head2 new

  my $portset = NACL::C::SysManager::Portset->new(command_interface => $ci,
												  portset           => $portset_name,
												  vserver           => $vserver_name);

Return a new Portset component object with the given attributes.

=cut

####################### Method Maker ###############################

use Class::MethodMaker [
    scalar => 'portset',
    scalar => 'vserver',
];

=head2 create

  my $portset_object  = NACL::C::SysManager::Portset->create( command_interface => $ci,
															  portset           => $portset_name,
															  vserver           => $vserver_name);

(Class method) Request creation of a new portset element, and  return a new
portset component object that refers to it.
Uses a CMode GUI APISet.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< portset => $string >>

(Required) Name of portset.

=item C<< vserver =>$string >>

(Required)The name of the vserver
in which the portset is present.

=item C<< protocol => $string >>

(Optional) Protocols type of the portset to be created.

=item C<< interface => $arrayref >>

(Optional) List of Interfaces which has to be associated with portset.


=back

=back

=cut

####################### GUI Create ###############################
sub create {
    $Log->enter() if $may_enter;
    my ( $pkg, @args ) = @_;

    $pkg->_frontend_log_debug_opts(@args);

    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            interface => { type => ARRAYREF, optional => 1 },
            _optional_scalars qw(protocol),
        },
    );

    my $ret = $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method => '_create_cmode_gui',
                set    => 'SysManager_CMode',
            },
        ],
    );

    $Log->exit() if $may_exit;
    return $ret;

} ## end sub create

####################### GUI CMode Create ###############################
sub _create_cmode_gui {
    $Log->enter() if $may_enter;
    my ( $pkg, @args ) = @_;
    my %opts = $pkg->_validate_backend_opts(@args);

    my $command_interface = $opts{command_interface};
    my $apiset            = $opts{apiset};

    my $portset_args = $pkg->_hash_copy(
        source => \%opts,
        map    => {
            "vserver"   => "-vservername",
            "portset"   => "-portsetName",
            "protocol"  => "-protocol",
            "interface" => "-interface",
        }
    );

    $apiset->portset_create(%$portset_args);
    $Log->exit() if $may_exit;

    return $pkg->new(
        portset           => $portset_args->{-portsetName},
        vserver           => $portset_args->{-vservername},
        command_interface => $command_interface
    );

} ## end sub _create_cmode_gui

=head2 delete

  NACL::C::SysManager::Portset->delete( command_interface => $ci,
									    portset           => $portset_name,
									    vserver           => $vserver_name);

                            or

  $portset_object->delete(...);

(Class or instance method)This method deletes a portset element.
Uses a CMode GUI APISet.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< portset=> $string >>

(Required for class method, Not Applicable for instance method)
Name of portset.

=item C<< vserver => $string >>

(Required for class method, Not Applicable for instance method)
The name of the vserver in which the portset is present.


=back

=back

=cut

####################### GUI Delete ###############################
sub delete {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);

    my %opts = $pkg_or_obj->_common_validate_with( params => \@args, );

    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method => '_delete_cmode_gui',
                set    => 'SysManager_CMode',
            },
        ],
    );

    $Log->exit() if $may_exit;
    return $ret;
} ## end sub delete

####################### GUI CMode Delete ###############################
sub _delete_cmode_gui {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    my %opts   = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = $opts{apiset};

    my $portset_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        map    => {
            "portset" => "-portsetName",
            "vserver" => "-vservername"
        }
    );

    $apiset->portset_delete(%$portset_args);
    $Log->exit() if $may_exit;
} ## end sub _delete_cmode_gui

=head2 modify

  NACL::C::SysManager::Portset->modify(command_interface => $ci,
									   portset           => $portset_name,
									   vserver           => $vserver_name,...);

                            or

  $portset_object->modify(...);

(Class or instance method) Modify a portset element.  
Uses a  CMode GUI APISet.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< portset => $string >>

(Required for class method, Not Applicable for instance method)
Name of portset.

=item C<< vserver => $string >>

(Required for class method, Not Applicable for instance method)
The name of the vserver in which the portset is present.

=item C<< interface => $arrayref >>

(Optional) List of Interfaces which has to be associated with portset.

=back

=back

=cut

####################### GUI Modify ###############################
sub modify {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);

    my %opts = $pkg_or_obj->_common_validate_with(
        params => \@args,
        additional_spec =>
            { interface => { type => ARRAYREF, optional => 1 }, },
    );

    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method => '_modify_cmode_gui',
                set    => 'SysManager_CMode',
            },
        ],
    );

    $Log->exit() if $may_exit;
    return $ret;
} ## end sub modify

####################### GUI CMode Modify ###############################
sub _modify_cmode_gui {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    my %opts   = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = $opts{apiset};

    my $portset_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        map    => {
            "portset"   => "-portsetName",
            "vserver"   => "-vservername",
            "interface" => "-interface",
        }
    );

    $apiset->portset_edit(%$portset_args);
    $Log->exit() if $may_exit;
} ## end sub _modify_cmode_gui

=head2 refresh

  NACL::C::SysManager::Portset->refresh(command_interface => $ci,
									    vserver           => $vserver_name,...);

                            or

  $portset_object->refresh();

(Class or instance method) This method refreshs portset page.
Uses a CMode GUI APIset.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< vserver => $string >>

(Required for class method, Not Applicable for instance method)
The name of the vserver in which the portset is present.


=back

=back

=cut

####################### GUI Refresh ###############################
sub refresh {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);

    my %opts = $pkg_or_obj->_common_validate_with( params => \@args, );

    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method => '_refresh_cmode_gui',
                set    => 'SysManager_CMode',
            },
        ],
    );

    $Log->exit() if $may_exit;
    return $ret;
} ## end sub refresh

####################### GUI CMode Refresh ###############################
sub _refresh_cmode_gui {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    my %opts   = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = $opts{apiset};

    my $portset_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        map    => { "vserver" => "-vservername" }
    );

    $apiset->portset_refresh(%$portset_args);
    $Log->exit() if $may_exit;
} ## end sub _refresh_cmode_gui

####################### Primary Keys Validate ###############################
sub _primary_keys_validate_spec {
    return (
        portset => { type => SCALAR },
        vserver => { type => SCALAR },
    );
} ## end sub _primary_keys_validate_spec

1;
