# $Id$

#
# Copyright (c) 2001-2011 NetApp, Inc., All Rights Reserved
#  Any use, modification, or distribution is prohibited
#  without prior written consent from NetApp, Inc.
#

## @summary Quota Component Module
## @author raja.sindhe@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::C::SysManager::Quota

=head1 DESCRIPTION

C<NACL::C::SysManager::Quota> is a derived class of
L<NACL::C::Component::SysManager|lib-NACL-C-Component-SysManager-pm>.  It represents access
to an ONTAP Quota present on the System Manger GUI Screen.  A related class is
L<NACL::CS::SysManager::Quota|lib-NACL-CS-SysManager-Quota-pm>, which represents
the State Info of an ONTAP Quota present on the System Manger GUI Screen.

=head1 ATTRIBUTES

=head2 command_interface

See L<NACL::C::Component::command_interface|lib-NACL-C-Component-pm/command_interface>

=head2 container

Container having the Quota path, used to identify it uniquely.

=head2 vserver

(Required for CMode, Not Applicable for 7Mode) 

=cut

package NACL::C::SysManager::Quota;

use strict;
use warnings;
use Params::Validate
    qw(validate_with validate SCALAR ARRAYREF BOOLEAN UNDEF SCALARREF HASHREF);

use NACL::CS::SysManager::Quota;
use NACL::ComponentUtils qw(_optional_scalars Dumper);
use NATE::BaseException qw(:try);
use NATE::Exceptions::Argument qw(:try);
use base 'NACL::C::Component::SysManager';
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();

=head1 METHODS

=head2 new

	my $quota = NACL::C::SysManager::Quota->new(command_interface => $ci,
												vserver           => $vserver_name,
												container         => $container);

Return a new Quota component object with the given attributes.

=cut

####################### Method Maker ###############################

use Class::MethodMaker [
    scalar => 'vserver',
    scalar => 'container',
];

=head2 create

	my $quota  = NACL::C::SysManager::Quota->create(command_interface => $ci,
													vserver           => $vserver_name,
													container         => $container,
													qtype             => $qtype,...);

(Class method) Request creation of a new quota object, and  return a new
quota component object that refers to it.
Uses a CMode GUI or 7mode GUI APISet.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< vserver => $string >>

(Required for CMode, Not Applicable for 7Mode)The name of the vserver
in which the quota is present.

=item C<< container => $string >>

(Required) Name of container.

=item C<< qtype => $string >>

(Required) Name of qtype (qtree/user/group).

=item C<< "quser" => $string >>

(Optional) quota for all users or specific user

=item C<< "qgroup" => $string >>

(Optional) quota for all groups or specific group

=item C<< "specific" => $string >>

(Optional) specific user or group

=item C<< "spacehardlt" => $string >>

(Optional) Space hard limit

=item C<< "spacehardltut" => $string >>

(Optional) Space hard limit unit (KB/MB/GB/TB)

=item C<< "spacesoftlt" => $string >>

(Optional) Space soft limit

=item C<< "spacesoftltut" => $string >>

(Optional) Space soft limit (KB/MB/GB/TB)

=item C<< "spacethlt" => $string >>

(Optional) Space threshhold limit

=item C<< "spacethltut" => $string >>

(Optional) Space threshhold limit unit

=item C<< "filehardlt" => $string >>

(Optional) file hard limit

=item C<< "filehardltut" => $string >>

(Optional) file hard limit unit

=item C<< "filesoftlt" => $string >>

(Optional) file soft limit

=item C<< "filesoftltut" => $string >>

(Optional) file soft limit unit

=back

=back

=cut

####################### GUI Create ###############################
sub create {
    $Log->enter() if $may_enter;
    my ( $pkg, @args ) = @_;

    $pkg->_frontend_log_debug_opts(@args);

    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw (qtype quser qgroup specific spacehardlt spacehardltut spacesoftlt
                    spacesoftltut spacethlt spacethltut filehardlt filehardltut filesoftlt filesoftltut)
            )
        }
    );

    my $ret = $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method => '_create_7mode_gui',
                set    => 'SysManager_7Mode',
            },
            {   method => '_create_cmode_gui',
                set    => 'SysManager_CMode',
            },
        ],
    );

    $Log->exit() if $may_exit;
    return $ret;

} ## end sub create

####################### GUI 7Mode Create ###############################
sub _create_7mode_gui {
    $Log->enter() if $may_enter;
    my ( $pkg, @args ) = @_;
    my %opts = $pkg->_validate_backend_opts(@args);

    my $command_interface = $opts{command_interface};
    my $apiset            = $opts{apiset};

    my $quota_args = $pkg->_hash_copy(
        source => \%opts,
        map    => {
            "container"     => "-container",
            "qtype"         => "-qtype",
            "quser"         => "-quser",
            "qgroup"        => "-qgroup",
            "specific"      => "-specific",
            "spacehardlt"   => "-spacehardlt",
            "spacehardltut" => "-spacehardltut",
            "spacesoftlt"   => "-spacesoftlt",
            "spacesoftltut" => "-spacesoftltut",
            "spacethlt"     => "-spacethlt",
            "spacethltut"   => "-spacethltut",
            "filehardlt"    => "-filehardlt",
            "filehardltut"  => "-filehardltut",
            "filesoftlt"    => "-filesoftlt",
            "filesoftltut"  => "-filesoftltut",
        }
    );

    $apiset->quota_create(%$quota_args);
    $Log->exit() if $may_exit;

    return $pkg->new(
        container         => $quota_args->{-container},
        command_interface => $command_interface
    );
} ## end sub _create_7mode_gui

####################### GUI CMode Create ###############################
sub _create_cmode_gui {
    $Log->enter() if $may_enter;
    my ( $pkg, @args ) = @_;
    my %opts = $pkg->_validate_backend_opts(@args);

    my $command_interface = $opts{command_interface};
    my $apiset            = $opts{apiset};

    my $quota_args = $pkg->_hash_copy(
        source => \%opts,
        map    => {
            "vserver"       => "-vservername",
            "container"     => "-container",
            "qtype"         => "-qtype",
            "quser"         => "-quser",
            "qgroup"        => "-qgroup",
            "specific"      => "-specific",
            "spacehardlt"   => "-spacehardlt",
            "spacehardltut" => "-spacehardltut",
            "spacesoftlt"   => "-spacesoftlt",
            "spacesoftltut" => "-spacesoftltut",
            "spacethlt"     => "-spacethlt",
            "spacethltut"   => "-spacethltut",
            "filehardlt"    => "-filehardlt",
            "filehardltut"  => "-filehardltut",
            "filesoftlt"    => "-filesoftlt",
            "filesoftltut"  => "-filesoftltut",
        }
    );

    $apiset->quota_create(%$quota_args);
    $Log->exit() if $may_exit;

    return $pkg->new(
        container         => $quota_args->{-container},
        vserver           => $quota_args->{-vservername},
        command_interface => $command_interface
    );

} ## end sub _create_cmode_gui

=head2 edit

	NACL::C::SysManager::Quota->edit(command_interface => $ci,
									 vserver           => $vserver_name,
									 container         => $container,);
	
	                        or

	$quota_object->edit(...);

(Class or instance method)This method edits the Quota
Uses a CMode GUI or 7mode GUI APISet.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< vserver => $string >>

(Required for class method, Not Applicable for instance method)
(Required for CMode, Not Applicable for 7Mode)The name of the vserver
in which the quota is present.

=item C<< container => $string >>

(Required for class method, Not Applicable for instance method)
Name of container.

=item C<< qtype => $string >>

(Required) Name of qtype (qtree/user/group).

=item C<< "spacehardlt" => $string >>

(Optional) Space hard limit

=item C<< "spacehardltut" => $string >>

(Optional) Space hard limit unit (KB/MB/GB/TB)

=item C<< "spacesoftlt" => $string >>

(Optional) Space soft limit

=item C<< "spacesoftltut" => $string >>

(Optional) Space soft limit (KB/MB/GB/TB)

=item C<< "spacethlt" => $string >>

(Optional) Space threshhold limit

=item C<< "spacethltut" => $string >>

(Optional) Space threshhold limit unit

=item C<< "filehardlt" => $string >>

(Optional) file hard limit

=item C<< "filehardltut" => $string >>

(Optional) file hard limit unit

=item C<< "filesoftlt" => $string >>

(Optional) file soft limit

=item C<< "filesoftltut" => $string >>

(Optional) file soft limit unit


=back

=back

=cut

sub edit {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);

    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw (qtype spacehardlt spacehardltut spacesoftlt spacesoftltut
                    spacethlt spacethltut filehardlt filehardltut filesoftlt filesoftltut)
            )
        }
    );

    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method => '_edit_7mode_gui',
                set    => 'SysManager_7Mode',
            },
            {   method => '_edit_cmode_gui',
                set    => 'SysManager_CMode',
            },
        ],
    );

    $Log->exit() if $may_exit;
    return $ret;

} ## end sub edit
####################### GUI 7Mode Create ###############################
sub _edit_7mode_gui {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    my %opts   = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = $opts{apiset};

    my $quota_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        map    => {
            "container"     => "-quotaname",
            "qtype"         => "-qtype",
            "spacehardlt"   => "-spacehardlt",
            "spacehardltut" => "-spacehardltut",
            "spacesoftlt"   => "-spacesoftlt",
            "spacesoftltut" => "-spacesoftltut",
            "spacethlt"     => "-spacethlt",
            "spacethltut"   => "-spacethltut",
            "filehardlt"    => "-filehardlt",
            "filehardltut"  => "-filehardltut",
            "filesoftlt"    => "-filesoftlt",
            "filesoftltut"  => "-filesoftltut",
        }
    );

    $apiset->quota_edit(%$quota_args);
    $Log->exit() if $may_exit;

} ## end sub _edit_7mode_gui

####################### GUI CMode Create ###############################
sub _edit_cmode_gui {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    my %opts   = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = $opts{apiset};

    my $quota_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        map    => {
            "vserver"       => "-vservername",
            "container"     => "-quotaname",
            "qtype"         => "-qtype",
            "spacehardlt"   => "-spacehardlt",
            "spacehardltut" => "-spacehardltut",
            "spacesoftlt"   => "-spacesoftlt",
            "spacesoftltut" => "-spacesoftltut",
            "spacethlt"     => "-spacethlt",
            "spacethltut"   => "-spacethltut",
            "filehardlt"    => "-filehardlt",
            "filehardltut"  => "-filehardltut",
            "filesoftlt"    => "-filesoftlt",
            "filesoftltut"  => "-filesoftltut",
        }
    );

    $apiset->quota_edit(%$quota_args);
    $Log->exit() if $may_exit;

} ## end sub _edit_cmode_gui

=head2 delete

	NACL::C::SysManager::Quota->delete(command_interface => $ci,
									   vserver           => $vserver_name,
									   container         => $container,);
	
	                        or

	$quota_object->delete(...);

(Class or instance method)This method deletes a quota element.
Uses a CMode GUI or 7mode GUI APISet.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< vserver => $string >>

(Required for class method, Not Applicable for instance method)
(Required for CMode, Not Applicable for 7Mode)The name of the vserver
in which the quota is present.

=item C<< container => $string >>

(Required for class method, Not Applicable for instance method)
Name of container.

=item C<< qtype => $string >>

(Optional) Name of qtype (qtree/user/group).

=item C<< multiplequota => $arrayref >>

(Optional) List of Quotas to be deleted.

=back

=back

=cut

sub delete {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);

    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            'multiplequota' => { type => ARRAYREF, optional => 1 },
            _optional_scalars(qw (container qtype))
        }
    );

    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method => '_delete_7mode_gui',
                set    => 'SysManager_7Mode',
            },
            {   method => '_delete_cmode_gui',
                set    => 'SysManager_CMode',
            },
        ],
    );

    $Log->exit() if $may_exit;
    return $ret;

} ## end sub delete
####################### GUI 7Mode Create ###############################
sub _delete_7mode_gui {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    my %opts   = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = $opts{apiset};

    my $quota_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        map    => {
            "container"     => "-quotaname",
            "qtype"         => "-qtype",
            "multiplequota" => "-multipledel",
        }
    );

    $apiset->quota_delete(%$quota_args);
    $Log->exit() if $may_exit;

} ## end sub _delete_7mode_gui

####################### GUI CMode Create ###############################
sub _delete_cmode_gui {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    my %opts   = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = $opts{apiset};

    my $quota_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        map    => {
            "vserver"       => "-vservername",
            "container"     => "-quotaname",
            "qtype"         => "-qtype",
            "multiplequota" => "-multipledel",
        }
    );

    $apiset->quota_delete(%$quota_args);
    $Log->exit() if $may_exit;

} ## end sub _delete_cmode_gui

=head2 quotastatus

	NACL::C::SysManager::Quota->quotastatus(command_interface => $ci,
											vserver           => $vserver_name,
											container         => $container,);
	
	                        or

	$quota_object->quotastatus(...);

(Class or instance method)This method sets the quota status
Uses a CMode GUI or 7mode GUI APISet.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< vserver => $string >>

(Required for class method, Not Applicable for instance method)
(Required for CMode, Not Applicable for 7Mode)The name of the vserver
in which the quota is present.

=item C<< container => $string >>

(Required for class method, Not Applicable for instance method)
Name of container.

=item C<< state => $string >>

(Optional) new state values (on/off).

=item C<< resize => $string >>

(Optional) resize state (yes/no).

=back

=back

=cut

sub quotastatus {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);

    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => { _optional_scalars(qw (state resize)) }
    );

    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method => '_quotastatus_7mode_gui',
                set    => 'SysManager_7Mode',
            },
            {   method => '_quotastatus_cmode_gui',
                set    => 'SysManager_CMode',
            },
        ],
    );

    $Log->exit() if $may_exit;
    return $ret;

} ## end sub quotastatus

####################### GUI 7Mode Create ###############################
sub _quotastatus_7mode_gui {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    my %opts   = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = $opts{apiset};

    my $quota_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        map    => {
            "container" => "-container",
            "state"     => "-state",
            "resize"    => "-resize",
        }
    );

    $apiset->quota_manage(%$quota_args);
    $Log->exit() if $may_exit;

} ## end sub _quotastatus_7mode_gui

####################### GUI CMode Create ###############################
sub _quotastatus_cmode_gui {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    my %opts   = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = $opts{apiset};

    my $quota_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        map    => {
            "vserver"   => "-vservername",
            "container" => "-container",
            "state"     => "-state",
            "resize"    => "-resize",
        }
    );

    $apiset->quota_manage(%$quota_args);
    $Log->exit() if $may_exit;

} ## end sub _quotastatus_cmode_gui

=head2 refresh

	NACL::C::SysManager::Quota->refresh(command_interface => $ci,
									    vserver           => $vserver_name,...);

                            or

	$quota_object->refresh();

(Class or instance method) This method refreshs quota page.
Uses a 7Mode GUI or CMode GUI APIset.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< vserver => $string >>

(Required for class method, Not Applicable for instance method)
(Required for CMode, Not Applicable for 7Mode)The name of the vserver
in which the quota is present.

=back

=back

=cut

####################### GUI Refresh ###############################
sub refresh {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);

    my %opts = $pkg_or_obj->_common_validate_with( params => \@args, );

    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method => '_refresh_7mode_gui',
                set    => 'SysManager_7Mode',
            },
            {   method => '_refresh_cmode_gui',
                set    => 'SysManager_CMode',
            },
        ],
    );

    $Log->exit() if $may_exit;
    return $ret;
} ## end sub refresh

####################### GUI 7Mode Refresh ###############################
sub _refresh_7mode_gui {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    my %opts   = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = $opts{apiset};

    my $quota_args = $pkg_or_obj->_hash_copy( source => \%opts, );

    $apiset->quota_refresh(%$quota_args);
    $Log->exit() if $may_exit;
} ## end sub _refresh_7mode_gui

####################### GUI CMode Refresh ###############################
sub _refresh_cmode_gui {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    my %opts   = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = $opts{apiset};

    my $quota_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        map    => { "vserver" => "-vservername" }
    );

    $apiset->quota_refresh(%$quota_args);
    $Log->exit() if $may_exit;
} ## end sub _refresh_cmode_gui

=head2 manage_refresh

	NACL::C::SysManager::Quota->manage_refresh(command_interface => $ci,
											   vserver           => $vserver_name,...);

                            or

	$quota_object->manage_refresh();

(Class or instance method) This method refreshs manage quota page.
Uses a 7Mode GUI or CMode GUI APIset.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< vserver => $string >>

(Required for class method, Not Applicable for instance method)
(Required for CMode, Not Applicable for 7Mode)The name of the vserver
in which the quota is present.

=back

=back

=cut

####################### GUI Manage Refresh ###############################
sub manage_refresh {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);

    my %opts = $pkg_or_obj->_common_validate_with( params => \@args, );

    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method => '_manage_refresh_7mode_gui',
                set    => 'SysManager_7Mode',
            },
            {   method => '_manage_refresh_cmode_gui',
                set    => 'SysManager_CMode',
            },
        ],
    );

    $Log->exit() if $may_exit;
    return $ret;
} ## end sub manage_refresh

####################### GUI 7Mode Manage Refresh ###############################
sub _manage_refresh_7mode_gui {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    my %opts   = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = $opts{apiset};

    my $quota_args = $pkg_or_obj->_hash_copy( source => \%opts, );

    $apiset->quota_manage_refresh(%$quota_args);
    $Log->exit() if $may_exit;
} ## end sub _manage_refresh_7mode_gui

####################### GUI CMode Manage Refresh ###############################
sub _manage_refresh_cmode_gui {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    my %opts   = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = $opts{apiset};

    my $quota_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        map    => { "vserver" => "-vservername" }
    );

    $apiset->quota_manage_refresh(%$quota_args);
    $Log->exit() if $may_exit;
} ## end sub _manage_refresh_cmode_gui

####################### Primary Keys Validate ###############################
sub _primary_keys_validate_spec {
    return (
        container => { type => SCALAR },
        vserver   => { type => SCALAR | UNDEF, optional => 1 },
    );
} ## end sub _primary_keys_validate_spec

1;
