# $Id$

#
# Copyright (c) 2001-2011 NetApp, Inc., All Rights Reserved
#  Any use, modification, or distribution is prohibited
#  without prior written consent from NetApp, Inc.
#

## @summary Roles Component Module
## @author Vikas.P@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::C::SysManager::Roles

=head1 DESCRIPTION

C<NACL::C::SysManager::Roles> is a derived class of
L<NACL::C::Component::SysManager|lib-NACL-C-Component-SysManager-pm>.  It represents access
to an ONTAP Roles present on the System Manger GUI Screen.  A related class is
L<NACL::CS::SysManager::Roles|lib-NACL-CS-SysManager-Roles-pm>, which represents
the State Info of an ONTAP Roles present on the System Manger GUI Screen.

=head1 ATTRIBUTES

=head2 command_interface

See L<NACL::C::Component::command_interface|lib-NACL-C-Component-pm/command_interface>

=head2 role

The name of the role, used to identify it uniquely. (role
names are assumed to be unique in Cmode)

=cut

package NACL::C::SysManager::Roles;

use strict;
use warnings;
use Params::Validate
    qw(validate_with validate SCALAR ARRAYREF BOOLEAN UNDEF SCALARREF HASHREF);

use NACL::CS::SysManager::Roles;
use NACL::ComponentUtils qw(_optional_scalars Dumper);
use NATE::BaseException qw(:try);
use NATE::Exceptions::Argument qw(:try);
use base 'NACL::C::Component::SysManager';
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();

=head1 METHODS

=head2 new

	my $user = 
		NACL::C::SysManager::Roles->new(command_interface => $ci,
										role              => $role,);

Return a new role component object with the given attributes.

=cut

use Class::MethodMaker [
    scalar => 'role',
    scalar => 'route',
    scalar => 'vserver',
    scalar => 'command_name',
];

=head2 add

	NACL::C::SysManager::Roles->add(command_interface => $ci,
								    role              => $role, 
								    route             => $route,...);
                                 
(Class method)This method adds a role
Uses a CMode GUI APISet.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< route => $route >>

(Required) Either vserver or cluster.

=item C<< role => $role >>

(Required) Role to be created.

=item C<< command => $password >>

(Required) The directory command.

=item C<< query => $query >>

(Optional) Query for the role

=item C<< access-level = $access_level >>

(Optional) Access level for the role

=item C<< vserver => $vserver >>

(Optional) Vserver name.

=back

=back

=cut

sub add {
    $Log->enter() if $may_enter;
    my ( $pkg, @args ) = @_;

    $pkg->_frontend_log_debug_opts(@args);

    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            command          => { type => ARRAYREF },
            'edit-command'   => { type => ARRAYREF, optional => 1 },
            'delete-command' => { type => ARRAYREF, optional => 1 },
            _optional_scalars(qw(query access-level vserver)),
        },
    );

    my $ret = $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method => '_add_cmode_gui',
                set    => 'SysManager_CMode',
            },
        ],
    );

    $Log->exit() if $may_exit;
    return $ret;

} ## end sub add

####################### GUI CMode add ###############################
sub _add_cmode_gui {
    $Log->enter() if $may_enter;
    my ( $pkg, @args ) = @_;
    my %opts = $pkg->_validate_backend_opts(@args);

    my $command_interface = $opts{command_interface};
    my $apiset            = $opts{apiset};

    my $security_args = $pkg->_hash_copy(
        source => \%opts,
        map    => {
            "role"           => "-rolename",
            "route"          => "-workflow",
            "access-level"   => "-access_level",
            "command"        => "-addcommand",
            "edit-command"   => "-editcommand",
            "delete-command" => "-deletecommand",
            "query"          => "-query",
            "vserver"        => "-vservername"
        }
    );

    $apiset->security_role_add(%$security_args);
    $Log->exit() if $may_exit;

    return $pkg->new(
        role              => $security_args->{-rolename},
        route             => $security_args->{-workflow},
        command_interface => $command_interface
    );

} ## end sub _add_cmode_gui

=head2 modify

   NACL::C::SysManager::Roles->modify(command_interface => $ci,
									  role              => $role, 
									  route             => $route,...);
                                 
								or

    $Role_object->modify(...);

(Class or instance method)This method modifies role
Uses a CMode GUI APISet.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< route => $route >>

(Required) Either vserver or cluster

=item C<< options => $options >>

(Required) Operations to be performed - add, edit, delete.

=item C<< role => $role >>

(Required) Role name to be modified.

=item C<< command => $command >>

(Optional) Command for the role

=item C<< query => $query >>

(Optional) Query for the role

=item C<< edit-command => $editcmd >>

(Optional)  Command to be edited.

=item C<< delete-command => $delcmd >>

(Optional) Command to be deleted.

=item C<< access-level => $access_level >>

(Optional) Access level for the role

=item C<< vserver => $vserver >>

(Optional) Vserver name.

=back

=back

=cut

sub modify {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);

    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            command          => { type => ARRAYREF, optional => 1 },
            'edit-command'   => { type => ARRAYREF, optional => 1 },
            'delete-command' => { type => ARRAYREF, optional => 1 },
            _optional_scalars(qw(access-level vserver)),
        },
    );

    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method => '_modify_cmode_gui',
                set    => 'SysManager_CMode',
            },
        ],
    );

    $Log->exit() if $may_exit;
    return $ret;

} ## end sub modify

####################### GUI CMode modify ###############################
sub _modify_cmode_gui {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    my %opts   = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = $opts{apiset};

    my $security_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        map    => {
            "route"          => "-workflow",
            "role"           => "-rolename",
            "command"        => "-addcommand",
            "access-level"   => "-access_level",
            "edit-command"   => "-editcommand",
            "delete-command" => "-deletecommand",
            "vserver"        => "-vservername"
        }
    );

    $apiset->security_role_modify(%$security_args);
    $Log->exit() if $may_exit;

} ## end sub _modify_cmode_gui

=head2 refresh

	NACL::C::SysManager::Roles->refresh(command_interface => $ci,
									    route             => $route);

                            or

	$Role_object->refresh();

(Class or instance method)This method refreshs Roles page
Uses a CMode GUI APISet.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< route => $route >>

(Required) Either vserver or cluster.

=back

=back

=cut

sub refresh {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);

    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            route => { type => SCALAR },
            _optional_scalars(qw(vserver)),
        }
    );

    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method => '_refresh_cmode_gui',
                set    => 'SysManager_CMode',
            },
        ],
    );

    $Log->exit() if $may_exit;
    return $ret;

} ## end sub refresh

####################### GUI CMode refresh ###############################
sub _refresh_cmode_gui {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    my %opts   = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = $opts{apiset};

    my $role_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        map    => {
            "route"   => "-workflow",
            "vserver" => "-vservername"
        }
    );

    $apiset->role_refresh(%$role_args);
    $Log->exit() if $may_exit;

} ## end sub _refresh_cmode_gui

####################### Primary Keys Validate ###############################
sub _primary_keys_validate_spec {
    return (
        role         => { type => SCALAR },
        route        => { type => SCALAR },
        vserver      => { type => SCALAR | UNDEF, optional => 1 },
        command_name => { type => SCALAR | UNDEF, optional => 1 },
    );
} ## end sub _primary_keys_validate_spec

1;
