# $Id$

#
# Copyright (c) 2001-2011 NetApp, Inc., All Rights Reserved
#  Any use, modification, or distribution is prohibited
#  without prior written consent from NetApp, Inc.
#

## @summary VolumeClone Component Module
## @author vikasp.p@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::C::SysManager::VolumeClone

=head1 DESCRIPTION

C<NACL::C::SysManager::VolumeClone> is a derived class of
L<NACL::C::SysManager::Volume|lib-NACL-C-SysManager-Volume-pm>.  It represents access
to an ONTAP VolumeClone present on the System Manger GUI Screen. Since this is a subclass
of Volume, all of the functionality of Volume is available here.  A related class is
L<NACL::CS::SysManager::VolumeClone|lib-NACL-CS-SysManager-VolumeClone-pm>, which represents
the State Info of an ONTAP VolumeClone present on the System Manger GUI Screen.

=head1 ATTRIBUTES

=head2 command_interface

See L<NACL::C::Component::command_interface|lib-NACL-C-Component-pm/command_interface>

=head2 volume

The name of the Clone, used to identify it uniquely. (Volume Clone
names are assumed to be unique within a node or within a cluster in 7mode
and in a vserver in Cmode)

=head2 vserver

(Required for CMode, Not Applicable for 7Mode) The name of the vserver
in which the volume is present.

=cut

package NACL::C::SysManager::VolumeClone;

use strict;
use warnings;
use Params::Validate
    qw(validate_with validate SCALAR ARRAYREF UNDEF SCALARREF );

use NACL::CS::SysManager::VolumeClone;
use NACL::ComponentUtils qw(_optional_scalars Dumper);
use NATE::BaseException qw(:try);
use NATE::Exceptions::Argument qw(:try);
use base 'NACL::C::SysManager::Volume';
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();

=head1 METHODS 

=head2 new

	my $Volclone = NACL::C::SysManager::VolumeClone->new(command_interface => $ci,
													     volume            => $clone,
													     vserver           => $vserver,...);

Return a new VolumeClone component object with the given attributes.

=cut

use Class::MethodMaker [
    scalar => 'volume',
    scalar => 'vserver',
];

=head2 create

	my $volumeclone  =
		NACL::C::SysManager::VolumeClone->create(command_interface => $ci,
												 clone             => $clone,
												 vserver           => $vserver...);

(Class method) Request creation of a new volumeclone element, and return a new 
volumeclone component object that refers to it.
Uses a CMode GUI or 7Mode GUI APISet.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< clone => $string >>

(Required) Name of clone to create.

=item C<< vserver => $vserver >>

(Required for CMode, Not Applicable for 7Mode)
Name of vserver.

=item C<< "volume" => $string >>

(Required) Name of the Volume.

=item C<< "type" => $string >>

(Required) type of the VolumeClone to be created(Values can be volume|file)

=item C<< "clone-parser" => $string >>

(Optional) Clone Parser Option

=item C<< "parent-copy" => $string >>

(Optional) Enable Parent copy 

=item C<< "parent-name" => $string >>

(Optional) Parent name for the clone

=item C<< "file-path" => $string >>

(Optional) file path for the clone creation

=back

=back

=cut

sub create {
    $Log->enter() if $may_enter;
    my ( $pkg, @args ) = @_;

    $pkg->_frontend_log_debug_opts(@args);

    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            volume => { type => SCALAR },
            clone  => { type => SCALAR },
            type   => { type => SCALAR },
            _optional_scalars(
                qw(clone-parser parent-copy parent-name file-path))
        }
    );

    my $ret = $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method => '_create_7mode_gui',
                set    => 'SysManager_7Mode',
            },
            {   method => '_create_cmode_gui',
                set    => 'SysManager_CMode',
            },
        ],
    );

    $Log->exit() if $may_exit;
    return $ret;
} ## end sub create

####################### GUI CMode Create ###############################
sub _create_cmode_gui {
    $Log->enter() if $may_enter;
    my ( $pkg, @args ) = @_;
    my %opts = $pkg->_validate_backend_opts(@args);

    my $command_interface = $opts{command_interface};
    my $apiset            = $opts{apiset};

    my $vol_clone_args = $pkg->_hash_copy(
        source => \%opts,
        map    => {
            "volume"       => "-volname",
            "vserver"      => "-vservername",
            "clone"        => "-clonename",
            "type"         => "-type",
            "clone-parser" => "-colnespaceres",
            "parent-copy"  => "-parentcopy",
            "parent-name"  => "-parentname",
            "file-path"    => "-filepath",
        }
    );

    $apiset->clone_volume(%$vol_clone_args);
    $Log->exit() if $may_exit;

    return $pkg->new(
        volume            => $vol_clone_args->{-clonename},
        vserver           => $vol_clone_args->{-vservername},
        command_interface => $command_interface
    );
} ## end sub _create_cmode_gui

####################### GUI 7Mode Create ###############################
sub _create_7mode_gui {
    $Log->enter() if $may_enter;
    my ( $pkg, @args ) = @_;
    my %opts = $pkg->_validate_backend_opts(@args);

    my $command_interface = $opts{command_interface};
    my $apiset            = $opts{apiset};

    my $vol_clone_args = $pkg->_hash_copy(
        source => \%opts,
        map    => {
            "volume"       => "-volname",
            "clone"        => "-clonename",
            "type"         => "-type",
            "clone-parser" => "-colnespaceres",
            "parent-copy"  => "-parentcopy",
            "parent-name"  => "-parentname",
            "file-path"    => "-filepath",
        }
    );

    $apiset->clone_volume(%$vol_clone_args);
    $Log->exit() if $may_exit;

    return $pkg->new(
        volume            => $vol_clone_args->{-clonename},
        command_interface => $command_interface
    );
} ## end sub _create_7mode_gui

=head2 split

	NACL::C::SysManager::VolumeClone->split(command_interface => $ci,
											volume            => $clone,
											vserver           => $vserver...);
                             
                            or
                            
	$volclone_object->split(...);

(Class or instance method) This method splits a volumeclone element. 
Uses a CMode GUI or 7Mode GUI APISet

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< clone => $string >>

(Required for class method, Not Applicable for instance method)
Name of clone.
 
=item C<< vserver => $vserver_name >>

(Required for class method, Not Applicable for instance method)
(Required for CMode, Not Applicable for 7Mode)
Name of vserver.
 
=item C<< "split" => $string >>

(Optional) Split the Clone option

=back

=back

=cut

sub split {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);

    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => { _optional_scalars(qw(split)) },
    );

    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method => '_split_7mode_gui',
                set    => 'SysManager_7Mode',
            },
            {   method => '_split_cmode_gui',
                set    => 'SysManager_CMode',
            },
        ],
    );

    $Log->exit() if $may_exit;
    return $ret;
} ## end sub split

####################### GUI CMode Split ###############################
sub _split_cmode_gui {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    my %opts   = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = $opts{apiset};

    my $vol_clone_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        map    => {
            "volume"  => "-clonename",
            "vserver" => "-vservername",
            "split"   => "-clonesplit",
        }
    );

    $apiset->clone_split(%$vol_clone_args);
    $Log->exit() if $may_exit;
} ## end sub _split_cmode_gui

####################### GUI 7Mode Split ###############################
sub _split_7mode_gui {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    my %opts   = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = $opts{apiset};

    my $vol_clone_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        map    => {
            "volume" => "-clonename",
            "split"  => "-clonesplit",
        }
    );

    $apiset->clone_split(%$vol_clone_args);
    $Log->exit() if $may_exit;
} ## end sub _split_7mode_gui

####################### Primary Keys Validate ###############################
sub _primary_keys_validate_spec {
    return (
        volume  => { type => SCALAR },
        vserver => { type => SCALAR | UNDEF, optional => 1 },
    );
} ## end sub _primary_keys_validate_spec

1;
