# $Id$

#
# Copyright (c) 2001-2011 NetApp, Inc., All Rights Reserved
#  Any use, modification, or distribution is prohibited
#  without prior written consent from NetApp, Inc.
#

## @summary VolumeSnapshot Component Module
## @author Neethu.George@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::C::SysManager::VolumeSnapshot

=head1 DESCRIPTION

C<NACL::C::SysManager::VolumeSnapshot> is a derived class of
L<NACL::C::Component::SysManager|lib-NACL-C-Component-SysManager-pm>.  It represents access
to an ONTAP Volume Snapshot present on the System Manger GUI Screen.  A related class is
L<NACL::CS::SysManager::VolumeSnapshot|lib-NACL-CS-SysManager-VolumeSnapshot-pm>, which represents
the State Info of an ONTAP Volume Snapshot present on the System Manger GUI Screen.

=head1 ATTRIBUTES

=head2 command_interface

See L<NACL::C::Component::command_interface|lib-NACL-C-Component-pm/command_interface>

=head2 snapshot

The name of the snapshot, used to identify it uniquely. (VolumeSnapshot
names are assumed to be unique within a node or within a cluster in 7mode
and in a vserver in Cmode)

=head2 volume

The name of the volume on which snapshot is created, used to identify it uniquely. 
(Volume names are assumed to be unique within a node or within a cluster in 7mode
and in a vserver in Cmode)

=head2 vserver

(Required for CMode, Not Applicable for 7Mode) The name of the vserver
in which the volume is present.

=cut

package NACL::C::SysManager::VolumeSnapshot;

use strict;
use warnings;
use Params::Validate
    qw(validate_with validate SCALAR ARRAYREF BOOLEAN UNDEF SCALARREF HASHREF);

use NACL::CS::SysManager::VolumeSnapshot;
use NACL::ComponentUtils qw(_optional_scalars Dumper);
use NATE::BaseException qw(:try);
use NATE::Exceptions::Argument qw(:try);
use base 'NACL::C::Component::SysManager';
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();

=head1 METHODS


=head2 new

	my $snapshot = 
		NACL::C::SysManager::VolumeSnapshot->new(command_interface => $ci,
												 vserver           => $vserver_name,
												 volume            => $volume_name,
												 snapshot          => $snapshot,...);

Return a new VolumeSnapshot component object with the given attributes.

=cut

use Class::MethodMaker [
    scalar => 'vserver',
    scalar => 'volume',
    scalar => 'snapshot',
];

=head2 create

	my $snapshot = 
		NACL::C::SysManager::VolumeSnapshot->create(command_interface => $ci,
													vserver           => $vserver_name,
													volume            => $volume_name,
													snapshot          => $snapshot,...);

(Class method) Request creation of a new snapshot element, and  return a new
snapshot component object that refers to it. 
Uses a CMode GUI or 7Mode GUI APISet.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< snapshot => $string >>

(Required) Name of snapshot to create.

=item C<< volume => $string >>

(Required) Name of volume of which the snapshot to be created.

=item C<< vserver => $string >>

(Required for CMode, Not Applicable for 7Mode)
Name of vserver.

=back

=back

=cut

sub create {
    $Log->enter() if $may_enter;
    my ( $pkg, @args ) = @_;

    $pkg->_frontend_log_debug_opts(@args);

    my %opts = $pkg->_common_validate_with( params => \@args, );

    my $ret = $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method => '_create_7mode_gui',
                set    => 'SysManager_7Mode',
            },
            {   method => '_create_cmode_gui',
                set    => 'SysManager_CMode',
            },
        ],
    );

    $Log->exit() if $may_exit;
    return $ret;

} ## end sub create

####################### GUI 7Mode Create ###############################
sub _create_7mode_gui {
    $Log->enter() if $may_enter;
    my ( $pkg, @args ) = @_;
    my %opts = $pkg->_validate_backend_opts(@args);

    my $command_interface = $opts{command_interface};
    my $apiset            = $opts{apiset};

    my $snapshot_args = $pkg->_hash_copy(
        source => \%opts,
        map    => {
            "volume"   => "-volname",
            "snapshot" => "-snapname"
        }
    );

    $apiset->create_volsnap(%$snapshot_args);
    $Log->exit() if $may_exit;

    return $pkg->new(
        volume            => $snapshot_args->{-volname},
        snapshot          => $snapshot_args->{-snapname},
        command_interface => $command_interface
    );
} ## end sub _create_7mode_gui

####################### GUI CMode Create ###############################
sub _create_cmode_gui {
    $Log->enter() if $may_enter;
    my ( $pkg, @args ) = @_;
    my %opts = $pkg->_validate_backend_opts(@args);

    my $command_interface = $opts{command_interface};
    my $apiset            = $opts{apiset};

    my $snapshot_args = $pkg->_hash_copy(
        source => \%opts,
        map    => {
            "volume"   => "-volname",
            "snapshot" => "-snapname",
            "vserver"  => "-vservername"
        }
    );

    $apiset->create_volsnap(%$snapshot_args);
    $Log->exit() if $may_exit;

    return $pkg->new(
        volume            => $snapshot_args->{-volname},
        snapshot          => $snapshot_args->{-snapname},
        command_interface => $command_interface
    );
} ## end sub _create_cmode_gui

=head2 delete

	NACL::C::SysManager::VolumeSnapshot->delete(command_interface => $ci,
											    snapshot          => $snapshot_name,...);

                            or

	$snapshot_object->delete(...);

(Class or instance method) This method deletes an snapshot element.
Uses a CMode GUI or 7Mode GUI APISet.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< snapshot=>$snapshot_name >>

(Required for class method, Not Applicable for instance method)
Name of a snapshot.

=item C<< volume => $string >>

(Required for class method, Not Applicable for instance method)
(Required) Name of volume of which the snapshot to be created.

=item C<< vserver => $string >>

(Required for class method, Not Applicable for instance method)
(Required for CMode, Not Applicable for 7Mode)
Name of vserver.

=back

=back

=cut

sub delete {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);

    my %opts = $pkg_or_obj->_common_validate_with( params => \@args, );

    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method => '_delete_7mode_gui',
                set    => 'SysManager_7Mode',
            },
            {   method => '_delete_cmode_gui',
                set    => 'SysManager_CMode',
            },
        ],
    );

    $Log->exit() if $may_exit;
    return $ret;

} ## end sub delete

####################### GUI 7Mode Delete ###############################
sub _delete_7mode_gui {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    my %opts   = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = $opts{apiset};

    my $snapshot_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        map    => {
            "volume"   => "-volname",
            "snapshot" => "-snapname"
        }
    );

    $apiset->manage_volsnap( -option => 'delete', %$snapshot_args );
    $Log->exit() if $may_exit;
} ## end sub _delete_7mode_gui

####################### GUI CMode Delete ###############################
sub _delete_cmode_gui {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    my %opts   = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = $opts{apiset};

    my $snapshot_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        map    => {
            "volume"   => "-volname",
            "vserver"  => "-vservername",
            "snapshot" => "-snapname"
        }
    );

    $apiset->manage_volsnap( -option => 'delete', %$snapshot_args, );
    $Log->exit() if $may_exit;
} ## end sub _delete_cmode_gui

=head2 restore

	NACL::C::SysManager::VolumeSnapshot->restore(command_interface => $ci,
												 volume            => $volume,
												 vserver           => $vserver,
												 snapshot          => $snapshot,...);
	                             
								 or

    $VolumeSnapshot_object->restore(...);

(Class or instance method)
This method is used to restore a volume from a snapshot.
Uses a CMode GUI or 7Mode GUI APISet.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< snapshot => $string >>

(Required for class method, Not Applicable for instance method)
Name of a snapshot.

=item C<< volume => $string >>

(Required for class method, Not Applicable for instance method)
(Required) Name of volume of which the snapshot to be created.

=item C<< vserver => $string >>

(Required for class method, Not Applicable for instance method)
(Required for CMode, Not Applicable for 7Mode)
Name of vserver.

=back

=back

=cut

sub restore {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);

    my %opts = $pkg_or_obj->_common_validate_with( params => \@args );

    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method => '_restore_7mode_gui',
                set    => 'SysManager_7Mode',
            },
            {   method => '_restore_cmode_gui',
                set    => 'SysManager_CMode',
            },
        ],
    );

    $Log->exit() if $may_exit;
    return $ret;
} ## end sub restore

####################### GUI 7Mode Restore ###############################
sub _restore_7mode_gui {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    my %opts   = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = $opts{apiset};

    my $snapshot_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        map    => {
            "volume"   => "-volname",
            "snapshot" => "-snapname"
        }
    );

    $apiset->manage_volsnap( -option => 'restore', %$snapshot_args );
    $Log->exit() if $may_exit;
} ## end sub _restore_7mode_gui

####################### GUI CMode restore ###############################
sub _restore_cmode_gui {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    my %opts   = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = $opts{apiset};

    my $snapshot_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        map    => {
            "volume"   => "-volname",
            "snapshot" => "-snapname",
            "vserver"  => "-vservername"
        }
    );

    $apiset->manage_volsnap( -option => 'restore', %$snapshot_args );
    $Log->exit() if $may_exit;
} ## end sub _restore_cmode_gui

=head2 rename

	NACL::C::SysManager::VolumeSnapshot->rename(command_interface => $ci,
											    volume            => $volume,
											    vserver           => $vserver,
											    snapshot          => $snapshot,...);
	                             
								 or

    $VolumeSnapshot_object->rename(...);

(Class or instance method)
This method is used to renames snapshot.
Uses a CMode GUI or 7Mode-GUI APISet.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< snapshot=>$string >>

(Required for class method, Not Applicable for instance method)
Name of a snapshot.

=item C<< volume => $string >>

(Required for class method, Not Applicable for instance method)
(Required) Name of volume of which the snapshot to be created.

=item C<< vserver => $string >>

(Required for class method, Not Applicable for instance method)
(Required for CMode, Not Applicable for 7Mode)
Name of vserver.

=item C<< new-name => $string >>

(Required) New Name for the snapshot

=back

=back

=cut

sub rename {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);

    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => { _optional_scalars(qw (new-name)) }
    );

    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method => '_rename_7mode_gui',
                set    => 'SysManager_7Mode',
            },
            {   method => '_rename_cmode_gui',
                set    => 'SysManager_CMode',
            },
        ],
    );

    $Log->exit() if $may_exit;
    return $ret;

} ## end sub rename

####################### GUI 7Mode Rename ###############################
sub _rename_7mode_gui {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    my %opts   = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = $opts{apiset};

    my $snapshot_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        map    => {
            "volume"   => "-volname",
            "snapshot" => "-snapname",
            "new-name" => "-snaprename"
        }
    );

    $apiset->manage_volsnap( -option => 'rename', %$snapshot_args );

    $pkg_or_obj->snapshot( $opts{'new-name'} ) if ( ref $pkg_or_obj );

    $Log->exit() if $may_exit;
} ## end sub _rename_7mode_gui

####################### GUI CMode rename ###############################
sub _rename_cmode_gui {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    my %opts   = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = $opts{apiset};

    my $snapshot_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        map    => {
            "volume"   => "-volname",
            "snapshot" => "-snapname",
            "vserver"  => "-vservername",
            "new-name" => "-snaprename"
        }
    );

    $apiset->manage_volsnap( -option => 'rename', %$snapshot_args );

    $pkg_or_obj->snapshot( $opts{'new-name'} ) if ( ref $pkg_or_obj );

    $Log->exit() if $may_exit;
} ## end sub _rename_cmode_gui

=head2 sched

    NACL::C::SysManager::VolumeSnapshot->sched(command_interface => $ci,
											   volume            => $volume,
											   vserver           => $vserver,...);
	                             
								 or

    $VolumeSnapshot_object->sched(...);

(Class method)This method is used to schedule a snapshot.
Uses a 7Mode GUI or CMode GUI APISet.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< volume => $string >>

(Required for class method, Not Applicable for instance method)
Name of the Volume

=item C<< vserver => $string >>

(Required for class method, Not Applicable for instance method)
(Required for CMode, Not Applicable for 7Mode)
Name of vserver.

=item C<< policy => $string >>

(Optional) Policy name

=item C<< weeks => $string >>

(Optional) Week Value

=item C<< days => $string >>

(Optional) Day Value

=item C<< hours => $string >>

(Optional) Hours value


=back

=back

=cut

sub sched {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);

    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => { _optional_scalars(qw (policy weeks days hours snapshot)) }
    );

    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method => '_sched_7mode_gui',
                set    => 'SysManager_7Mode',
            },
            {   method => '_sched_cmode_gui',
                set    => 'SysManager_CMode',
            },
        ],
    );

    $Log->exit() if $may_exit;
    return $ret;

} ## end sub sched

####################### GUI 7Mode Sched ###############################
sub _sched_7mode_gui {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    my %opts   = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = $opts{apiset};

    my $snapshot_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        map    => {
            "volume"     => "-volname",
            "percentage" => "-snapres",
            "weeks"      => "-weeklyvalue",
            "days"       => "-dailyvalue",
            "hours"      => "-hourlyvalue"
        }
    );

    $apiset->manage_volsnap(
        -option       => 'config',
        -snapschedule => 'enable',
        %$snapshot_args,
    );
    $Log->exit() if $may_exit;
} ## end sub _sched_7mode_gui

####################### GUI CMode sched ###############################
sub _sched_cmode_gui {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    my %opts   = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = $opts{apiset};

    my $snapshot_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        map    => {
            "volume"  => "-volname",
            "vserver" => "-vservername",
            "policy"  => "-policy"
        }
    );

    $apiset->manage_volsnap(
        -option       => 'config',
        -snapschedule => 'enable',
        %$snapshot_args,
    );
    $Log->exit() if $may_exit;
} ## end sub _sched_cmode_gui

=head2 snapdirshow

    NACL::C::SysManager::VolumeSnapshot->snapdirshow(command_interface => $ci,
													 volume            => $volume,
													 vserver           => $vserver,...);
	                             
								 or

    $VolumeSnapshot_object->snapdirshow(...);

(Class method)
This method is used for the visibility of .snapshot directory on 
the  volume at the client mount points.
Uses a 7mode GUI or Cmode GUI APISet.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< volume => $string >>

(Required for class method, Not Applicable for instance method)
Name of the Volume

=item C<< vserver => $string >>

(Required for class method, Not Applicable for instance method)
(Required for CMode, Not Applicable for 7Mode)
Name of vserver.


=back

=back

=cut

sub snapdirshow {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);

    my %opts = $pkg_or_obj->_common_validate_with( 
        params => \@args,
        additional_spec => { _optional_scalars(qw (snapshot))} );

    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method => '_snapdirshow_7mode_gui',
                set    => 'SysManager_7Mode',
            },
            {   method => '_snapdirshow_cmode_gui',
                set    => 'SysManager_CMode',
            },
        ],
    );

    $Log->exit() if $may_exit;
    return $ret;

} ## end sub snapdirshow

####################### GUI 7Mode Snapdirshow ###############################
sub _snapdirshow_7mode_gui {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    my %opts   = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = $opts{apiset};

    my $snapshot_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        map    => { "volume" => "-volname" }
    );

    $apiset->manage_volsnap(
        -option  => 'config',
        -dirshow => 'enable',
        %$snapshot_args,
    );
    $Log->exit() if $may_exit;
} ## end sub _snapdirshow_7mode_gui

####################### GUI CMode Snapdirshow ###############################
sub _snapdirshow_cmode_gui {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    my %opts   = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = $opts{apiset};

    my $snapshot_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        map    => {
            "volume"  => "-volname",
            "vserver" => "-vservername"
        }
    );

    $apiset->manage_volsnap(
        -option  => 'config',
        -dirshow => 'enable',
        %$snapshot_args,
    );
    $Log->exit() if $may_exit;
} ## end sub _snapdirshow_cmode_gui

####################### Primary Keys Validate ###############################
sub _primary_keys_validate_spec {
    return (
        volume   => { type => SCALAR },
        vserver  => { type => SCALAR | UNDEF, optional => 1 },
        snapshot => { type => SCALAR },
    );
} ## end sub _primary_keys_validate_spec

1;
