#
# Copyright (c) 2001-2012 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary SystemConfigurationBackup Component Module
## @author dl-nacl-dev@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::C::SystemConfigurationBackup

=head1 DESCRIPTION

C<NACL::C::SystemConfigurationBackup> is a derived class of
L<NACL::C::Component::ONTAP|lib-NACL-C-Component-ONTAP-pm>. It represents access
to an ONTAP SystemConfigurationBackup.  A related class is
L<NACL::CS::SystemConfigurationBackup|lib-NACL-CS-SystemConfigurationBackup-pm>, which represents
the past state of an ONTAP SystemConfigurationBackup.

=head1 ATTRIBUTES

=head2 command_interface

See L<NACL::C::Component::command_interface|lib-NACL-C-Component-pm/command_interface>

=head2 node

=head2 backup

=cut

package NACL::C::SystemConfigurationBackup;

use strict;
use warnings;

use NACL::CS::SystemConfigurationBackup;
use NACL::ComponentUtils qw (_optional_scalars Dumper);

use base 'NACL::C::Component::ONTAP';
use Params::Validate qw(validate validate_with SCALAR ARRAYREF UNDEF);
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();

=head1 METHODS

=head2 new

  my $SystemConfigurationBackup = NACL::C::SystemConfigurationBackup->new(command_interface=>$ci,
                                   node =>$node,
                                   backup =>$backup,
                                  );

Return a new SystemConfigurationBackup component object with the given attributes.

=cut

use Class::MethodMaker [
    scalar => 'node',
    scalar => 'backup',
];

=head2 find

    my $SystemConfigurationBackup  = NACL::C::SystemConfigurationBackup->find(command_interface => $ci, ...);
    my @SystemConfigurationBackups = NACL::C::SystemConfigurationBackup->find(command_interface => $ci, ...);

(Class method)
A constructor of sorts, which discovers which elements are present (by
interacting with the given command interface).  In an array context it
creates a component object for each present element and returns them.
In a scalar context it creates a component object for just one
element.

This method is generally not implemented in individual components. The base
class (L<NACL::C::Component::ONTAP|lib-NACL-C-Component-ONTAP-pm> provides an implementation.
See <L<NACL::C::Component::find|lib-NACL-C-Component-pm/find> for a more
detailed description along with explanation of all the options accepted.

=head2 state

    my $SystemConfigurationBackup_state = $NACL::C::SystemConfigurationBackup->state();

(Instance method)
Fetch the current state of the element associated
with this component, and return a Component State object corresponding
to this component (calling this on NACL::C::SystemConfigurationBackup would return
a NACL::CS::SystemConfigurationBackup object)

This method is generally not implemented in individual components. The base
class (L<NACL::C::Component::ONTAP|lib-NACL-C-Component-ONTAP-pm> provides an implementation.
See <L<NACL::C::Component::state|lib-NACL-C-Component-pm/state> for a more
detailed description along with explanation of all the options accepted.

=cut

=head2 rename

   NACL::C::SystemConfigurationBackup->rename(
                                command_interface => $ci,
                                 node => $node,
                                 backup => $backup,
                                  ...);

                                 or

    $SystemConfigurationBackup_object->rename(...);

(Class or instance method)
This method is used to rename a configuration backup.

Uses a CMode CLI/ZAPI APISet.
  Invokes 'config-backup-rename' API for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "node" => $node >>

(Required for class method, Not Applicable for instance method)
  Applicable for CMode CLI/ZAPI.
  Maps to:
    CM ZAPI: node

=item C<< "backup" => $backup >>

  Applicable for CMode CLI/ZAPI.
  Maps to:
    CMode ZAPI: "backup-name" of "config_backup_rename"

(Required for class method, Not Applicable for instance method)

=item C<< "new-name" => $string >>

  Applicable for CMode CLI/ZAPI.
  Maps to:
    CMode ZAPI: "new-backup-name" of "config_backup_rename"

=item C<< "local" => $string >>

  Applicable only for CMode CLI.

=item C<< 'extended_query' => $string >>

(Optional)
  Applicable only for CMode CLI.

=item C<< 'cmd_execution_time' => \$scalar >>

(Optional) Scalar reference , it will be updated with the command execution
time returned by the apiset response object

=item C<< "job_component" => \$scalar >>

(Optional)
If "system configuration backup rename" command returns job_id
then \$scalar gets updated with the corresponding job component

Applicable for CMode CLI

=item C<< "job_components" => \@array >>

(Optional)
If "system configuration backup rename" command returns multiple job_ids
then \@array gets updated with an array of corresponding job components

Applicable for CMode CLI

=item C<< '*' >>

This method takes all other options supported by
"system configuration backup rename" command in ONTAP.

Applicable for CMode CLI/ZAPI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub rename {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( new-name local
                    )
            ),
        },
        cli_cmd => 'system configuration backup rename',
    );
    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_rename_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_rename_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::SystemConfigurationBackup::rename',
    );
    $Log->exit() if $may_exit;
    $pkg_or_obj->backup($opts{'new-name'}) if ref($pkg_or_obj);
} ## end sub rename

sub _rename_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@args,
        api => 'system_configuration_backup_rename');
    $Log->exit() if $may_exit;
} ## end sub _rename_cmode_cli

sub _rename_cmode_zapi {  
    $Log->enter() if $may_enter;
    my ($pkg, @opts) = @_;
    $pkg->_base_backend_cmode_zapi(
        @opts,
        copy => [qw(node)],
	map  => {
            "backup"   => "backup-name",
            "new-name" => "new-backup-name",
        },
        api  => 'config_backup_rename',
    );
    $Log->exit() if $may_exit;
} ## end sub _rename_cmode_zapi

=head2 create

    my $SystemConfigurationBackup_object    = NACL::C::SystemConfigurationBackup->create(
                                command_interface => $ci,
                                 node => $node,
                                 backup => $backup,
                                  ...);

(Class method)
This method is used to create a configuration backup.

Uses a CMode CLI/ZAPI APISet.
  Invokes 'config-backup-create' API for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "node" => $node >>

(Required)
  Applicable for CMode CLI/ZAPI.
  Maps to:
    CM ZAPI: node

=item C<< "backup" => $backup >>

(Required)

=item C<< "backup-type" => $string >>

  Applicable for CMode CLI/ZAPI.
  Maps to:
    CM ZAPI: backup-type

=item C<< "local" => $string >>

  Applicable only for CMode CLI.

=item C<< "backup-name" => $string >>

  Applicable for CMode CLI/ZAPI.
  Maps to:
    CM ZAPI: backup-name

=item C<< 'extended_query' => $string >>

(Optional)
  Applicable only for CMode CLI.

=item C<< "job_component" => \$scalar >>

(Optional)
If "system configuration backup create" command returns job_id
then \$scalar gets updated with the corresponding job component

Applicable for CMode CLI/ZAPI

=item C<< "job_components" => \@array >>

(Optional)
If "system configuration backup create" command returns multiple job_ids
then \@array gets updated with an array of corresponding job components

Applicable for CMode CLI/ZAPI

=item C<< '*' >>

This method takes all other options supported by
"system configuration backup create" command in ONTAP.

Applicable for CMode CLI/ZAPI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub create {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( backup-type local backup-name
                    )
            ),
        },
        cli_cmd => 'system configuration backup create',
    );
    $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_create_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_create_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::SystemConfigurationBackup::create',
    );
    my $obj = $pkg->new(
        command_interface => $opts{command_interface},
        node              => $opts{'node'},
        'backup'          => $opts{'backup-name'} . '.7z',
    );
    $Log->exit() if $may_exit;
    return $obj;
} ## end sub create

sub _create_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_base_backend_cmode_cli(@args,
        api => 'system_configuration_backup_create');
    $Log->exit() if $may_exit;
} ## end sub _create_cmode_cli

sub _create_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg, %opts) = @_;
    $pkg->_base_backend_cmode_zapi(
        %opts,
        copy => [qw(backup-type node backup-name)],
        api  => 'config_backup_create',
    );
    $Log->exit() if $may_exit;
} ## end sub _create_cmode_zapi

=head2 download

   NACL::C::SystemConfigurationBackup->download(
                                command_interface => $ci,
                                 node => $node,
                                 backup => $backup,
                                  ...);

(Class method)
This method is used to download a configuration backup.

Uses a CMode CLI/ZAPI APISet.
  Invokes 'config-backup-download' API for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "node" => $node >>

(Required)
  Applicable for CMode CLI/ZAPI.
  Maps to:
    CM ZAPI: node

=item C<< "backup" => $backup >>

(Required)
  Applicable only for CMode CLI.

=item C<< "source" => $string >>

  Applicable for CMode CLI/ZAPI.
  Maps to:
    CMode ZAPI: "source-url" of "config_backup_download"

=item C<< "sequence" => $string >>

  Applicable only for CMode CLI.

=item C<< "user-name" => $user_name >>

(Required) user name for ftp login.
  Applicable only for CMode CLI.

=item C<< "password" => $password >>

(Required) password for ftp login.
  Applicable only for CMode CLI.

=item C<< "local" => $string >>

  Applicable only for CMode CLI.

=item C<< "backup-name" => $string >>

  Applicable for CMode CLI/ZAPI.
  Maps to:
    CM ZAPI: backup-name

=item C<< 'extended_query' => $string >>

(Optional)
  Applicable only for CMode CLI.

=item C<< 'cmd_execution_time' => \$scalar >>

(Optional) Scalar reference , it will be updated with the command execution
time returned by the apiset response object

=item C<< "job_component" => \$scalar >>

(Optional)
If "system configuration backup download" command returns job_id
then \$scalar gets updated with the corresponding job component

Applicable for CMode CLI

=item C<< "job_components" => \@array >>

(Optional)
If "system configuration backup download" command returns multiple job_ids
then \@array gets updated with an array of corresponding job components

Applicable for CMode CLI

=item C<< '*' >>

This method takes all other options supported by
"system configuration backup download" command in ONTAP.

Applicable for CMode CLI/ZAPI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub download {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( source sequence local backup-name
                    user-name password
                    )
            ),
        },
        cli_cmd => 'system configuration backup download',
    );
    $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_download_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_download_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::SystemConfigurationBackup::download',
    );

    $Log->exit() if $may_exit;
} ## end sub download

sub _download_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, %args) = @_;
    $args{additional_args} = {
        'connectrec-match_table' => [
            qr/username:/i => delete $args{'user-name'},
            qr/password:/i => delete $args{password}
        ]
    };
    $pkg->_base_backend_cmode_cli(%args,
        api => 'system_configuration_backup_download');
    $Log->exit() if $may_exit;
} ## end sub _download_cmode_cli

sub _download_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg, %opts) = @_;
    $pkg->_base_backend_cmode_zapi(
        %opts,
        copy => [qw(node backup-name)],
        map    => {"source" => "source-url"},
        api  => 'config_backup_download',
    );
    $Log->exit() if $may_exit;
} ## end sub _download_cmode_zapi

=head2 copy

    my $SystemConfigurationBackup_object    = NACL::C::SystemConfigurationBackup->copy(
                                command_interface => $ci,
                                 node => $node,
                                 backup => $backup,
                                  ...);

(Class method)
This method is used to copy a configuration backup.

Uses a CMode CLI/ZAPI APISet.
  Invokes 'config-backup-copy' API for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "node" => $node >>

(Required)
  Applicable only for CMode CLI.

=item C<< "backup" => $backup >>

  Applicable for CMode CLI/ZAPI.
  Maps to:
    CMode ZAPI: "backup-name" of "config-backup-copy"

(Required)

=item C<< "from-node" => $string >>

  Applicable for CMode CLI/ZAPI.
  Maps to:
    CM ZAPI: from-node

=item C<< "to-node" => $string >>

  Applicable for CMode CLI/ZAPI.
  Maps to:
    CM ZAPI: to-node

=item C<< "local" => $string >>

  Applicable only for CMode CLI.

=item C<< 'extended_query' => $string >>

(Optional)
  Applicable only for CMode CLI.

=item C<< 'cmd_execution_time' => \$scalar >>

(Optional) Scalar reference , it will be updated with the command execution
time returned by the apiset response object

=item C<< "job_component" => \$scalar >>

(Optional)
If "system configuration backup copy" command returns job_id
then \$scalar gets updated with the corresponding job component

Applicable for CMode CLI

=item C<< "job_components" => \@array >>

(Optional)
If "system configuration backup copy" command returns multiple job_ids
then \@array gets updated with an array of corresponding job components

Applicable for CMode CLI

=item C<< '*' >>

This method takes all other options supported by
"system configuration backup copy" command in ONTAP.

Applicable for CMode CLI/ZAPI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub copy {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( from-node to-node local backup node
                    )
            ),
        },
        ignore_primary_keys => 1,
        cli_cmd             => 'system configuration backup copy',
    );
    $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_copy_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_copy_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::SystemConfigurationBackup::copy',
    );

    $Log->exit() if $may_exit;
    return $pkg->new(
        command_interface => $opts{command_interface},
        node              => $opts{'node'},
        backup            => $opts{'backup'},
    );
} ## end sub copy

sub _copy_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_base_backend_cmode_cli(@args,
        api => 'system_configuration_backup_copy');
} ## end sub _copy_cmode_cli

sub _copy_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg, %opts) = @_;
    $pkg->_base_backend_cmode_zapi(
        %opts,
        copy => [qw(from-node to-node)],
        map  => {"backup" => "backup-name"},
        api  => 'config_backup_copy',
    );
    $Log->exit() if $may_exit;
} ## end sub _copy_cmode_zapi

=head2 upload

   NACL::C::SystemConfigurationBackup->upload(
                                command_interface => $ci,
                                 node => $node,
                                 backup => $backup,
                                  ...);

                                 or

    $SystemConfigurationBackup_object->upload(...);

(Class or instance method)
This method is used to upload a configuration backup.

Uses a CMode CLI/ZAPI APISet.
  Invokes 'config-backup-upload' API for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "node" => $node >>

(Required for class method, Not Applicable for instance method)
  Applicable for CMode CLI/ZAPI.
  Maps to:
    CM ZAPI: node

=item C<< "backup" => $backup >>

  Applicable for CMode CLI/ZAPI.
  Maps to:
    CMode ZAPI: "backup-name" of "config_backup_upload"

(Required for class method, Not Applicable for instance method)

=item C<< "sequence" => $string >>

  Applicable only for CMode CLI.

=item C<< "destination" => $string >>

  Applicable for CMode CLI/ZAPI.
  Maps to:
    CMode ZAPI: "destination-url"  of "config_backup_upload"

=item C<< "user-name" => $string >>

(Required) User name for ftp login.
  Applicable for CMode CLI.

=item C<< "password" => $string >>

(Required) password for ftp login.
  Applicable for CMode CLI.

=item C<< "local" => $string >>

  Applicable only for CMode CLI.

=item C<< 'extended_query' => $string >>

(Optional)
  Applicable only for CMode CLI.

=item C<< 'cmd_execution_time' => \$scalar >>

(Optional) Scalar reference , it will be updated with the command execution
time returned by the apiset response object

=item C<< "job_component" => \$scalar >>

(Optional)
If "system configuration backup upload" command returns job_id
the \$scalar gets updated with the job component

Applicable for CMode CLI

=item C<< "job_components" => \@array >>

(Optional)
If "system configuration backup upload" command returns multiple job_ids
the \@array gets updated with the array of job components

Applicable for CMode CLI

=item C<< '*' >>

This method takes all other options supported by
"system configuration backup upload" command in ONTAP.

Applicable for CMode CLI/ZAPI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub upload {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( sequence destination local
                    user-name password
                    )
            ),
        },
        cli_cmd => 'system configuration backup upload',
    );
    $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_upload_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_upload_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::SystemConfigurationBackup::upload',
    );

    $Log->exit() if $may_exit;
} ## end sub upload

sub _upload_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %args) = @_;
    $args{additional_args} = {
        'connectrec-match_table' => [
            qr/username:/i => delete $args{'user-name'},
            qr/password:/i => delete $args{password}
        ]
    };
    $pkg_or_obj->_base_backend_cmode_cli(%args,
        api => 'system_configuration_backup_upload');
    $Log->exit() if $may_exit;
} ## end sub _upload_cmode_cli

sub _upload_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg, %opts) = @_;
    $pkg->_base_backend_cmode_zapi(
        %opts,
        copy => [qw(node)],
        map    => {
            "backup"      => "backup-name",
            "destination" => "destination-url",
        },
        api  => 'config_backup_upload',
    );
    $Log->exit() if $may_exit;
} ## end sub _upload_cmode_zapi

=head2 delete

   NACL::C::SystemConfigurationBackup->delete(
                                command_interface => $ci,
                                 node => $node,
                                 backup => $backup,
                                  ...);

                                 or

    $SystemConfigurationBackup_object->delete(...);

(Class or instance method)
This method is used to delete a configuration backup.

Uses a CMode CLI/ZAPI APISet.
  Invokes 'config-backup-delete' API for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "node" => $node >>

(Required for class method, Not Applicable for instance method)
  Applicable for CMode CLI/ZAPI.
  Maps to:
    CM ZAPI: node

=item C<< "backup" => $backup >>

  Applicable for CMode CLI/ZAPI.
  Maps to:
    CMode ZAPI: "backup-name" of "config_backup_delete"

(Required for class method, Not Applicable for instance method)

=item C<< "local" => $string >>

  Applicable only for CMode CLI.

=item C<< 'extended_query' => $string >>

(Optional)
  Applicable only for CMode CLI.

=item C<< 'cmd_execution_time' => \$scalar >>

(Optional) Scalar reference , it will be updated with the command execution
time returned by the apiset response object

=item C<< "job_component" => \$scalar >>

(Optional)
If "system configuration backup delete" command returns job_id
then \$scalar gets updated with the corresponding job component

Applicable for CMode CLI

=item C<< "job_components" => \@array >>

(Optional)
If "system configuration backup delete" command returns multiple job_ids
then \@array gets updated with the array of corresponding job components

Applicable for CMode CLI

=item C<< '*' >>

This method takes all other options supported by
"system configuration backup delete" command in ONTAP.

Applicable for CMode CLI/ZAPI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub delete {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( local
                    )
            ),
        },
        cli_cmd => 'system configuration backup delete',
    );
    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_delete_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_delete_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::SystemConfigurationBackup::delete',
    );

    $Log->exit() if $may_exit;
    return $ret;
} ## end sub delete

sub _delete_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@args,
        api => 'system_configuration_backup_delete');
    $Log->exit() if $may_exit;
} ## end sub _delete_cmode_cli

sub _delete_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg, %opts) = @_;
    $pkg->_base_backend_cmode_zapi(
        %opts,
        copy => [qw(node)],
        map  => {"backup" => "backup-name"},
        api  => 'config_backup_delete',
    );
    $Log->exit() if $may_exit;
} ## end sub _delete_cmode_zapi


__PACKAGE__->_load_event_class();

sub _primary_keys_validate_spec {
    return (
        'node'   => {type => SCALAR | UNDEF},
        'backup' => {type => SCALAR | UNDEF, optional => 1},
    );
} ## end sub _primary_keys_validate_spec

1;
