# Copyright (c) 2001-2013 NetApp, Inc. All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary SystemServicesFirewallPolicy Component Module
## @author dl-nacl-dev@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::C::SystemServicesFirewallPolicy

=head1 DESCRIPTION

C<NACL::C::SystemServicesFirewallPolicy> is a derived class of
L<NACL::C::Component::ONTAP|lib-NACL-C-Component-ONTAP-pm>. It represents access
to an ONTAP SystemServicesFirewallPolicy.  A related class is
L<NACL::CS::SystemServicesFirewallPolicy|lib-NACL-CS-SystemServicesFirewallPolicy-pm>, which represents
the past state of an ONTAP SystemServicesFirewallPolicy.

=head1 ATTRIBUTES

=head2 command_interface

See L<NACL::C::Component::command_interface|lib-NACL-C-Component-pm/command_interface>

=head2 policy

=head2 action

Applicable for pre-FS

=head2 service

=head2 vserver

Applicable for FS onwards

=cut

package NACL::C::SystemServicesFirewallPolicy;

use strict;
use warnings;

use NACL::CS::SystemServicesFirewallPolicy;
use NACL::ComponentUtils qw (_optional_scalars);

use NACL::C::Exceptions::SystemServicesFirewallPolicy::AlreadyExists ();
use NACL::C::Exceptions::SystemServicesFirewallPolicy::DoesNotExist  ();
use NATE::BaseException qw(:try);
use NATE::Exceptions::Argument qw(:try);

use base 'NACL::C::Component::ONTAP';
use Params::Validate qw(validate validate_with SCALAR ARRAYREF OBJECT);
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Data::Dumper;

=head1 METHODS

=head2 new

  ## Call for pre-FS builds
  my $SystemServicesFirewallPolicy = NACL::C::SystemServicesFirewallPolicy->new(command_interface=>$ci,
                                   policy =>$policy,
                                   action =>$action,
                                   service =>$service,
                                );

  ## Call for FS onwards
  my $SystemServicesFirewallPolicy = NACL::C::SystemServicesFirewallPolicy->new(command_interface=>$ci,
                                       policy =>$policy,
                                       service =>$service,
                                       vserver =>$vserver
                                ); 



Return a new SystemServicesFirewallPolicy component object with the given attributes.

=cut

use Class::MethodMaker [
    scalar => 'policy',
    scalar => 'action',
    scalar => 'service',
    scalar => 'vserver',
];

=head2 find

    my $SystemServicesFirewallPolicy  = NACL::C::SystemServicesFirewallPolicy->find(command_interface => $ci, ...);
	my @SystemServicesFirewallPolicys = NACL::C::SystemServicesFirewallPolicy->find(command_interface => $ci, ...);

(Class method)
A constructor of sorts, which discovers which elements are present (by
interacting with the given command interface).  In an array context it
creates a component object for each present element and returns them.
In a scalar context it creates a component object for just one
element.

This method is generally not implemented in individual components. The base
class (L<NACL::C::Component::ONTAP|lib-NACL-C-Component-ONTAP-pm> provides an implementation.
See <L<NACL::C::Component::find|lib-NACL-C-Component-pm/find> for a more
detailed description along with explanation of all the options accepted.

=head2 state

    my $SystemServicesFirewallPolicy_state = $NACL::C::SystemServicesFirewallPolicy->state();

(Instance method)
Fetch the current state of the element associated
with this component, and return a Component State object corresponding
to this component (calling this on NACL::C::SystemServicesFirewallPolicy would return
a NACL::CS::SystemServicesFirewallPolicy object)

This method is generally not implemented in individual components. The base
class (L<NACL::C::Component::ONTAP|lib-NACL-C-Component-ONTAP-pm> provides an implementation.
See <L<NACL::C::Component::state|lib-NACL-C-Component-pm/state> for a more
detailed description along with explanation of all the options accepted.

=cut

=head2 create

    ## Call for pre-FS builds
    my $SystemServicesFirewallPolicy_object = NACL::C::SystemServicesFirewallPolicy->create(
                                command_interface => $ci,
                                policy => $policy,
                                action => $action,
                                service => $service,
                                  ...);

    ## Call for FS onwards
    my $SystemServicesFirewallPolicy_object = NACL::C::SystemServicesFirewallPolicy->create(
                                 command_interface => $ci,
                                 policy => $policy,
                                 service => $service,
                                 vserver =>$vserver
                                  ...);


(Class method)

This method is used to add a service to an existing firewall policy.

Uses a CMode CLI APIset/ZAPI.

Invokes API 'net-firewall-policy-create' for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "policy" => $policy >>

(Required)

Applicable for CMode CLI/ZAPI

=item C<< "action" => $action >>

Required for pre-FS builds, Not Applicable for FS onwards

Applicable for CMode CLI.

=item C<< 'vserver' => $string >>

Required for FS onwards, Not Applicable pre-FS

Applicable for CMode CLI/ZAPI

=item C<< "service" => $service >>

(Required)

Applicable for CMode CLI/ZAPI

=item C<< "ip-list" => [ $ip-list1, $ip-list2, ... ] >>

Applicable for CMode CLI/ZAPI

Maps to: ip-and-mask

=item C<< "extended_query" => $string >>

(Optional)

Applicable for CMode CLI

=item C<< 'allow-list' => [ $allow-1, $allow-2, ... ] >>

Allowed IPs
(Arrayref, even if only a single value is to be sent it should specified
as an arrayref containing that single value)

Applicable for CMode CLI/ZAPI

=item C<< 'deny-list' => [ $deny-1, $deny-2, ... ] >>

Denied IPs
(Arrayref, even if only a single value is to be sent it should specified
as an arrayref containing that single value)
Applicable for CMode CLI

=item C<< '*' >>

This method takes all other options supported by
"system services firewall policy create" command in ONTAP.

Applicable for CMode CLI/ZAPI.

=back

=back

=over

=item Exceptions

=over

=item C<NACL::C::Exceptions::SystemServicesFirewallPolicy::AlreadyExists>

This type of exception gets thrown if an attempt is made to create a
firewall policy that already exists.

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub create {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            "allow-list" => {type => ARRAYREF, optional => 1},
            "deny-list"  => {type => ARRAYREF, optional => 1},
            _optional_scalars(
                qw( vserver
                    )
            ),
            "ip-list" => {type => SCALAR | ARRAYREF, optional => 1},
        },
        cli_cmd => 'system services firewall policy create',
    );
    $pkg->_scalar_to_array_deprication_warning(
        params  => ['ip-list'],
        options => \%opts
    );
    my $ret;
    try {
        $ret = $pkg->call_on_apiset(
            %opts,
            choices => [
                {   method    => '_create_cmode_cli',
                    interface => 'CLI',
                    set       => 'CMode',
                },
                {   method    => '_create_cmode_zapi',
                    interface => 'ZAPI',
                    set       => 'CMode',
                },
            ],
            frontend => 'NACL::C::SystemServicesFirewallPolicy::create',
        );
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = $_[0];
        NACL::C::Exceptions::SystemServicesFirewallPolicy::AlreadyExists
            ->detect_convert_and_throw(exception => $exception);
        $Log->exit() if $may_exit;
        $exception->throw();
    };
    $Log->exit() if $may_exit;
    return $ret;
} ## end sub create

sub _create_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, %opts) = @_;
    my $action            = $opts{action};
    my %keys              = (
        command_interface => $opts{'command_interface'},
        policy            => $opts{'policy'},
        service           => $opts{'service'},
    );

    if ($opts{command_interface}->version_manager->has_uichange(
            uichange        => 'ipspaces-ms4',
            throw_exception => 0
        )
        )
    {
        $keys{vserver} = $opts{'vserver'};
    } else {
        $keys{action} = $action;
    }

    $pkg->_base_backend_cmode_cli(%opts,
        api => 'system_services_firewall_policy_create',);

    $Log->exit() if $may_exit;
    return $pkg->new(%keys);

} ## end sub _create_cmode_cli

sub _create_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg, %opts) = @_;

    $pkg->_base_backend_cmode_zapi(
        %opts,
        api             => 'net-firewall-policy-create',
        copy            => [qw( allow-list policy service vserver )],
        map             => {"ip-list" => "ip-and-mask",},
        frontend_method => 'create'
    );
    $Log->exit() if $may_exit;
    return $pkg->new(
        command_interface => $opts{'command_interface'},
        policy            => $opts{'policy'},
        service           => $opts{'service'},
        vserver           => $opts{'vserver'},
    );
}

=head2 modify

   ## Call for pre-FS builds
   NACL::C::SystemServicesFirewallPolicy->modify(
                                command_interface => $ci,
                                policy => $policy,
                                action => $action,
                                service => $service,
                                  ...);

   ## Call for FS onwards
   NACL::C::SystemServicesFirewallPolicy->modify(
                                 command_interface => $ci,
                                 policy => $policy,
                                 service => $service,
                                 vserver =>$vserver
                                   ...);
                                or

   $SystemServicesFirewallPolicy_object->modify(...);

(Class or instance method)
This method is used to modify a policy's IP-list.

Uses a CMode CLI APIset/ZAPI.

Invokes API 'net-firewall-policy-modify' for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "policy" => $policy >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI

=item C<< "action" => $action >>

(Required for class method for pre-FS, Not Applicable for instance method or FS onwards)

(Applicable for all other class/instance methods)

=item C<< "service" => $service >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI.

=item C<< "ip-list" => [ $ip-list1, $ip-list2, ... ] >>

Applicable for CMode CLI/ZAPI

Maps to: ip-and-mask

=item C<< 'vserver' => $string >>

(Required for class method for FS, Not Applicable for instance method or pre-FS)

Applicable for CMode CLI/ZAPI.

=item C<< "extended_query" => $string >>

(Optional)

Applicable for CMode CLI

=item C<< 'allow-list' => [ $allow-1, $allow-2, ... ] >>

Allowed IPs
(Arrayref, even if only a single value is to be sent it should specified
as an arrayref containing that single value)
Applicable for CMode CLI

Applicable for CMode CLI/ZAPI.

=item C<< 'deny-list' => [ $deny-1, $deny-2, ... ] >>

Denied IPs
(Arrayref, even if only a single value is to be sent it should specified
as an arrayref containing that single value)
Applicable for CMode CLI

=item C<< '*' >>

This method takes all other options supported by
"system services firewall policy modify" command in ONTAP.

Applicable for CMode CLI/ZAPI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub modify {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            "allow-list" => {type => ARRAYREF, optional => 1},
            "deny-list"  => {type => ARRAYREF, optional => 1},
            _optional_scalars(
                qw( vserver
                    )
            ),
            "ip-list" => {type => SCALAR | ARRAYREF, optional => 1},
        },
        cli_cmd => 'system services firewall policy modify',
    );
    $pkg_or_obj->_scalar_to_array_deprication_warning(
        params  => ['ip-list'],
        options => \%opts
    );
    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_modify_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_modify_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::SystemServicesFirewallPolicy::modify',
    );

    $Log->exit() if $may_exit;
    return $ret;
} ## end sub modify

sub _modify_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@opts,
        api => 'system_services_firewall_policy_modify');
    $Log->exit() if $may_exit;
} ## end sub _modify_cmode_cli

sub _modify_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg, %opts) = @_;

    $pkg->_base_backend_cmode_zapi(
        %opts,
        api             => 'net-firewall-policy-modify',
        copy            => [qw( allow-list policy service vserver )],
        map             => {"ip-list" => "ip-and-mask",},
        frontend_method => 'modify'
    );
    $Log->exit() if $may_exit;
}    ##end sub _modify_cmode_zapi

=head2 clone

   ## Call for pre-FS builds
   NACL::C::SystemServicesFirewallPolicy->clone(
                                command_interface => $ci,
                                policy => $policy,
                                  ...);

   ## Call for FS onwards
   NACL::C::SystemServicesFirewallPolicy->clone(
                                    command_interface => $ci,
                                    policy => $policy,
                                    vserver => $vserver,
                                    ...);


(Class method)
This method is used to clone an existing firewall policy.

Uses a CMode CLI APIset.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "policy" => $policy >>

(Required)

Applicable for CMode CLI.

=item C<< 'vserver' => $string >>

(Required for FS, Not Applicable pre-FS)

=item C<< "new-policy-name" => $string >>

Applicable for CMode CLI.

=item C<< "extended_query" => $string >>

(Optional)

Applicable for CMode CLI.

=item C<< 'destination-vserver' => $string >>

Vserver owning the New Firewall Policy
Applicable for CMode CLI

=item C<< 'destination-policy' => $string >>

Name of New Firewall Policy
Applicable for CMode CLI

=item C<< '*' >>

This method takes all other options supported by
"system services firewall policy clone" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub clone {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( new-policy-name
                    policy vserver

                    destination-vserver destination-policy
                    service
                    )
            ),
        },
        cli_cmd => 'system services firewall policy clone',
    );
    my $ret;
    try {
        $ret = $pkg->call_on_apiset(
            %opts,
            choices => [
                {   method    => '_clone_cmode_cli',
                    interface => 'CLI',
                    set       => 'CMode',
                },
            ],
            frontend => 'NACL::C::SystemServicesFirewallPolicy::clone',
        );
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = shift;
        NACL::C::Exceptions::SystemServicesFirewallPolicy::AlreadyExists
            ->detect_convert_and_throw(exception => $exception);
        $Log->exit() if $may_exit;
        $exception->throw();
    };
    $Log->exit() if $may_exit;
    return $ret;
} ## end sub clone

sub _clone_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_base_backend_cmode_cli(@args,
        api     => 'system_services_firewall_policy_clone',
    );    
    $Log->exit() if $may_exit;
} ## end sub _clone_cmode_cli

=head2 delete

   ##Call for pre-FS builds
   NACL::C::SystemServicesFirewallPolicy->delete(
                                command_interface => $ci,
                                policy => $policy,
                                service => $service,
                                action  => $action
                                  ...);

   ##Call for FS onwards
   NACL::C::SystemServicesFirewallPolicy->delete(
                                      command_interface => $ci,
                                      policy => $policy,
                                      vserver => $vserver,
                                      service => $service,
                                or

    $SystemServicesFirewallPolicy_object->delete(...);

(Class or instance method)

This method is used to remove a service from a firewall policy.

Uses a CMode CLI APIset/ZAPI.

Invokes API 'net-firewall-policy-destroy' for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "policy" => $policy >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI.

=item C<< "action" => $action >>

(Required for class method for pre-FS, Not Applicable for instance method or FS onwards)

(Applicable for all other class/instance methods)

=item C<< 'vserver' => $string >>

(Required for class method for FS, Not Applicable for instance method or pre-FS)

(Applicable for all other class/instance methods)

Applicable for CMode CLI/ZAPI.

=item C<< "service" => $service >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI

=item C<< "extended_query" => $string >>

(Optional)

Applicable for CMode CLI

=item C<< '*' >>

This method takes all other options supported by
"system services firewall policy delete" command in ONTAP.

Applicable for CMode CLI/ZAPI.

=back

=back

=over

=item Exceptions

=over

=item C<NACL::C::Exceptions::SystemServicesFirewallPolicy::DoesNotExist>

This type of exception gets thrown if an attempt is made to delete a
security login which does not exists.

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub delete {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( vserver
                    )
            )
        },
        cli_cmd => 'system services firewall policy delete',
    );
    my $ret;
    try {
        my $ret = $pkg_or_obj->call_on_apiset(
            %opts,
            choices => [
                {   method    => '_delete_cmode_cli',
                    interface => 'CLI',
                    set       => 'CMode',
                },
                {   method    => '_delete_cmode_zapi',
                    interface => 'ZAPI',
                    set       => 'CMode',
                    check     => '_check_service_value',
                },
            ],
            frontend => 'NACL::C::SystemServicesFirewallPolicy::delete',
        );
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = $_[0];
        NACL::C::Exceptions::SystemServicesFirewallPolicy::DoesNotExist
            ->detect_convert_and_throw(exception => $exception);
        $Log->exit() if $may_exit;
        $exception->throw();
    };
    $Log->exit() if $may_exit;
    return $ret;
} ## end sub delete

sub _delete_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@args,
        api => 'system_services_firewall_policy_delete');

    $Log->exit() if $may_exit;
} ## end sub _delete_cmode_cli

sub _delete_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg, %opts) = @_;
    $pkg->_base_backend_cmode_zapi(
        %opts,
        api             => 'net-firewall-policy-destroy',
        copy            => [qw( policy service vserver )],
        frontend_method => 'delete'
    );
    $Log->exit() if $may_exit;
}    ##end sub _delete_cmode_zapi

#The Param service cannot understand value '*' in ZAPI and
#considers it to be a literal rather than 'delete all'
#if it is passed ,It should be re-directed to CLI

sub _check_service_value {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    if ($opts{service} eq '*') {
        $Log->exit() if $may_exit;
        NACL::Exceptions::InvalidChoice->throw(
            "Param service with '*' value is not supported in ZAPI");
    }
    $Log->exit() if $may_exit;
}


__PACKAGE__->_load_event_class();

sub _primary_keys_validate_spec {
    my ($pkg_or_obj, @args) = @_;
    my %opts = validate(
        @args,
        {   command_interface => {
                type     => OBJECT,
                isa      => 'NACL::C::CommandInterface',
                optional => 1
            }
        }
    );
    my $command_interface = $opts{command_interface};
    my %default_spec      = (
        'policy'  => {type => SCALAR},
        'service' => {type => SCALAR},
        'action'  => {type => SCALAR, optional => 1},
        'vserver' => {type => SCALAR, optional => 1},
    );
    if (defined $command_interface) {
        if ($opts{command_interface}->version_manager->has_uichange(
                uichange        => 'ipspaces-ms4',
                throw_exception => 0
            )
            )
        {
            return (
                'policy'  => {type => SCALAR},
                'service' => {type => SCALAR},
                'vserver' => {type => SCALAR},
            );
        } else {

            return (
                'policy'  => {type => SCALAR},
                'service' => {type => SCALAR},
                'action'  => {type => SCALAR},

            );
        }
    } else {
        return %default_spec;
    }

} ## end sub _primary_keys_validate_spec

1;
