# $Id$

#
# Copyright (c) 2012 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary SystemServicesNdmp Component Module
## @author dl-dpg-nb-automation; dl-nacl-dev@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::C::SystemServicesNdmp

=head1 DESCRIPTION

C<NACL::C::SystemServicesNdmp> is a derived class of
L<NACL::C::Component::ONTAP|lib-NACL-C-Component-ONTAP-pm>.
It represents NDMP service on a filer.
Note that there is no command to create a new NDMP session and we need to
rely on an external agent to establish NDMP sessions. We can only use the
component to affect settings of the NDMP service and of existing NDMP
sessions.

=cut

package NACL::C::SystemServicesNdmp;

use strict;
use warnings;

use NACL::CS::SystemServicesNdmp;

use NACL::CS::SystemServicesNdmpGetVersion;
use NACL::CS::SystemServicesNdmpGetPassword;
use NACL::CS::SystemServicesNdmpShow;

use NATE::Exceptions::Argument qw(:try);

use NACL::ComponentUtils qw (_optional_scalars);

use base 'NACL::C::Component::ONTAP';

use Params::Validate qw(validate validate_with SCALAR SCALARREF ARRAYREF);
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Data::Dumper;

=head1 METHODS

=head2 new

    my $session = NACL::C::SystemServicesNdmp->new(
                                   command_interface => $ci,
                                   'session-id'      => $session_id,
                                   'node'            => $node,
                              );

Return a new SystemServicesNdmp component object with the given attributes.

=cut

use Class::MethodMaker [
    scalar => 'session_id',
    scalar => 'node',
];

=head2 find

    my $ndmp  = NACL::C::SystemServicesNdmp->find(command_interface => $ci, ...);
    my @ndmps = NACL::C::SystemServicesNdmp->find(command_interface => $ci, ...);

(Class method)
A constructor of sorts, which discovers which elements are present (by
interacting with the given command interface).  In an array context it
creates a component object for each present element and returns them.
In a scalar context it creates a component object for just one
element.

This method is generally not implemented in individual components. The base
class (L<NACL::C::Component::ONTAP|lib-NACL-C-Component-ONTAP-pm> provides an implementation.
See <L<NACL::C::Component::find|lib-NACL-C-Component-pm/find> for a more
detailed description along with explanation of all the options accepted.

=head2 state

    my $ndmp_state = $ndmp->state();

(Instance method)
Fetch the current state of the element associated
with this component, and return a Component State object corresponding
to this component (calling this on NACL::C::VserverLocks would return
a NACL::CS::VserverLocks object)

This method is generally not implemented in individual components. The base
class (L<NACL::C::Component::ONTAP|lib-NACL-C-Component-ONTAP-pm> provides an implementation.
See <L<NACL::C::Component::state|lib-NACL-C-Component-pm/state> for a more
detailed description along with explanation of all the options accepted.

=head2 on

    NACL::C::ndmpd->on(command_interface => $ci [, node => $node]);

(Class method)
This method is used to switch on the NDMP service.

Uses a 7Mode, Nodescope and CMode CLI/ZAPI APISet.

Invokes 'ndmpd on' command for 7Mode/Nodescope CLI.

Invokes  ndmp-node-service-enable  for CMode Zapi non-iter

Note that this can also be achieved by doing
    NACL::C::SystemServicesNdmp->modify(enable => 'true');

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< node => $string >>

Node on which to switch on NDMP.

Applicable for CMode CLI.

=item C<< '*' >>

This method takes all other options supported by
"system services ndmp modify" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub on {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params              => \@args,
        additional_spec     => {_optional_scalars(qw(node))},
        ignore_primary_keys => 1,
        api                 => 'system_services_ndmp_modify',
    );
    $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_on_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_on_7mode_cli',
                interface => 'CLI',
                set       => '7Mode|Nodescope',
            },
            {   method    => '_on_cmode_zapi_non_iter',
                interface => 'ZAPI',
                set       => 'CMode',
                zapi_type => 'non-iter',
            },
        ],
        frontend => 'NACL::C::SystemServicesNdmp::on',
    );

    $Log->exit() if $may_exit;
} ## end sub on

sub _on_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    my $password          = delete $opts{password};
    if (!defined $password) {
        $password = "admin";
    }
    my $prompts_answers = [
        'Please enter password:' => sub {$password},
        'Confirm password:'      => sub {$password}
    ];
    my %additional_args;
    $additional_args{'connectrec-match_table'} = $prompts_answers;

    # Switching on the NDMP service can be done in CMode either through
    # "system services ndmp modify -enable true" or through
    # "system services ndmp on". However, "on" is an alias of
    # "modify -enable true", hence there is no CDEF for "on".
    # Hence, here we need to use "system services ndmp modify"
    $opts{enable} = 'true';
    $pkg_or_obj->_base_backend_cmode_cli(
        %opts,
        api             => 'system_services_ndmp_modify',
        additional_args => \%additional_args
    );
    $Log->exit() if $may_exit;
} ##end sub _on_cmode_cli

sub _on_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );

    delete $opts{command_interface};
    my $apiset = delete $opts{apiset};

    $apiset->ndmpd_on();
    $Log->exit() if $may_exit;
} ## end sub _on_7mode_cli

sub _on_cmode_zapi_non_iter {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    my $map = {'node' => 'node-name',};
    $pkg_or_obj->_base_backend_cmode_zapi(
        @opts,
        api => 'ndmp-node-service-enable',
        map => $map,
    );
    $Log->exit() if $may_exit;
}

=head2 killall

    NACL::C::SystemServicesNdmp->killall(command_interface => $ci [, node => $node] );

(Class method)
This method is used to kill all NDMP sessions.

Uses a 7Mode, Nodescope and CMode CLI APISet.

Invokes 'ndmpd killall' command for 7Mode/Nodescope CLI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< node => $string >>

(Optional - only applicable for CMode)
The node on which to kill all NDMP sessions.

Applicable for CMode CLI.

=item C<< '*' >>

This method takes all other options supported by
"system services ndmp killall" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub killall {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params              => \@args,
        additional_spec     => {_optional_scalars(qw(node))},
        ignore_primary_keys => 1,
        cli_cmd             => 'system services ndmp kill-all',
    );
    $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_killall_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_killall_7mode_cli',
                interface => 'CLI',
                set       => '7Mode|Nodescope',
            },
        ],
        frontend => 'NACL::C::SystemServicesNdmp::killall',
    );

    $Log->exit() if $may_exit;
} ## end sub killall

sub _killall_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, %opts) = @_;

    $pkg->_base_backend_cmode_cli(
        %opts,
        api => 'system_services_ndmp_killall',
    );

    $Log->exit() if $may_exit;
} ## end sub _killall_cmode_cli

sub _killall_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );

    delete $opts{command_interface};
    my $apiset = delete $opts{apiset};

    $apiset->ndmpd_killall();
    $Log->exit() if $may_exit;
} ## end sub _killall_7mode_cli

=head2 kill

    NACL::C::SystemServicesNdmp->kill(
                                command_interface => $ci,
                                'session-id' => $session_id,
                                  ...);

                                 or

    $Session_object->kill(...);

(Class or instance method)
This method is used to kill the specified NDMP session.

Uses a 7Mode, Nodescope and CMode CLI APISet.

Invokes 'ndmpd kill' command for 7Mode/Nodescope CLI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "session-id" => $string >>

(Required for class method, Not Applicable for instance method)
The ID of the NDMP session to be killed

=item C<< node => $string >>

(Optional - only applicable for CMode)
The node on which to kill the specified NDMP session.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub kill {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    $Log->debug("Opts to 'kill' front end are:\n" . Dumper(\@_));
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@_,
        additional_spec => {_optional_scalars(qw(node))},
    );
    $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_kill_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_kill_7mode_cli',
                interface => 'CLI',
                set       => '7Mode|Nodescope',
            },
        ],
    );

    $Log->exit() if $may_exit;
} ## end sub kill

sub _kill_cmode_cli {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );

    delete $opts{command_interface};
    my $apiset = delete $opts{apiset};

    $Log->debug("Opts to API call 'system_services_ndmp_kill' are: \n"
            . Dumper(\%opts));

    # "system services ndmp kill" takes one (un-named) parameter which is the
    # session id. "system services ndmp killsession", however, takes node
    # and session-id as named parameters, so we call that here.
    $apiset->system_services_ndmp_killsession(%opts);
    $Log->exit() if $may_exit;
} ## end sub _kill_cmode_cli

sub _kill_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );

    delete $opts{command_interface};
    my $apiset = delete $opts{apiset};

    $apiset->ndmpd_kill(%opts);
    $Log->exit() if $may_exit;
} ## end sub _kill_7mode_cli

=head2 get_version

    my @objs = NACL::C::SystemServicesNdmp->get_version(
                                command_interface => $ci
                                [, node => $node], ...);

                    or

    my $obj  = NACL::C::SystemServicesNdmp->get_version(
                                command_interface => $ci
                                [, node => $node], ...);

(Class method)

Uses a 7Mode, Nodescope and CMode CLI APISet.

Invokes 'ndmpd version' command for 7Mode/Nodescope CLI.

This method is used to obtain default NDMP version. The object(s) returned are
of the type
L<NACL::CS::SystemServicesNdmpGetVersion|lib-NACL-CS-SystemServicesNdmpGetVersion-pm>,
which contains two fields: node and version.

For CMode, the "system services ndmp version" command returns the NDMP version
for each node in the cluster unless a specific node is provided. Each row of
output is modeled as an object.

For 7Mode, the "ndmpd version" command returns the NDMP version. Hence, for
7Mode there is only one object returned, which only has a "version" field
(no "node" field).

This method can be invoked in either list context (in which case it returns all
objects) or scalar context (in which case it returns the first object).

(Note that this method does not support the filter or requested_fields
parameters. It also does not throw a "NoElementsFound" exception. It returns a
CS object since any method that returns useful output also returns
some sort of CS object)

Both the CMode command "system services ndmp version" and the 7Mode/Nodescope
command "ndmpd version" can be used either to query the current version
or set a new value. For consistency of the return value, we've split this
functionality into two methods: one is this method (to get the version). To set
the version L<set_version|lib-NACL-C-SystemServicesNdmp-pm/set_version> should
be used.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< node => $string >>

(Optional - only applicable for CMode)
The node on which to query for the NDMP version. If this is specified, the NDMP
version on only that node is returned, else it will return the NDMP version
for each node in the cluster.

Applicable for CMode CLI.

=item C<< '*' >>

This method takes all other options supported by
"system services ndmp version" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub get_version {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params              => \@args,
        additional_spec     => {_optional_scalars(qw(node))},
        ignore_primary_keys => 1,
        cli_cmd             => 'system services ndmp version',
    );
    my @objs = $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_get_version_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_get_version_7mode_cli',
                interface => 'CLI',
                set       => '7Mode|Nodescope',
            },
        ],
        frontend => 'NACL::C::SystemServicesNdmp::get_version',
    );

    $Log->exit() if $may_exit;
    return wantarray ? @objs : $objs[0];
} ## end sub get_version

sub _get_version_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    my $response;

    $pkg->_base_backend_cmode_cli(
        @args,
        api => 'system_services_ndmp_version',
        apiset_response_obj => \$response,
    );

    my $output = $response->get_parsed_output();
    my @objs;

    foreach my $row (@{$output}) {
        my $obj = NACL::CS::SystemServicesNdmpGetVersion->new();
        $row->{version} = delete $row->{maxversion};
        $obj->_set_fields(row => $row);
        push @objs, $obj;
    } ## end foreach my $row ( @{$output...
        
    $Log->exit() if $may_exit;
    return @objs;
} ## end sub _get_version_cmode_cli

sub _get_version_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );

    delete $opts{command_interface};
    my $apiset = delete $opts{apiset};

    my $response = $apiset->ndmpd_version();
    my $output   = $response->get_parsed_output();
    my $version  = $output->[0]{version};
    my $obj
        = NACL::CS::SystemServicesNdmpGetVersion->new(version => $version);
    my @objs;
    push @objs, $obj;
    $Log->exit() if $may_exit;
    return @objs;
} ## end sub _get_version_7mode_cli

sub set_version {
    $Log->enter() if $may_enter;
    my $pkg = shift;
    $Log->debug("Opts to 'set_version' front end are:\n" . Dumper(\@_));
    my %opts = $pkg->_common_validate_with(
        params              => \@_,
        additional_spec     => {_optional_scalars(qw(node maxversion))},
        ignore_primary_keys => 1,
    );
    $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_set_version_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_set_version_7mode_cli',
                interface => 'CLI',
                set       => '7Mode|Nodescope',
            },
        ],
    );

    $Log->exit() if $may_exit;
} ## end sub set_version

sub _set_version_cmode_cli {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );

    delete $opts{command_interface};
    my $apiset = delete $opts{apiset};

    $Log->debug("Opts to API call 'system_services_ndmp_version' are: \n"
            . Dumper(\%opts));

    $apiset->system_services_ndmp_version(%opts);

    $Log->exit() if $may_exit;
} ## end sub _set_version_cmode_cli

sub _set_version_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );

    delete $opts{command_interface};
    my $apiset = delete $opts{apiset};

    my %api_opts;
    $pkg->_hash_copy(
        source => \%opts,
        target => \%api_opts,
        map    => {'maxversion' => 'ndmpd-max-version'},
    );

    $Log->debug(
        "Opts to API call 'ndmpd_version' are:\n" . Dumper(\%api_opts));

    $apiset->ndmpd_version(%api_opts);

    $Log->exit() if $may_exit;
} ## end sub _set_version_7mode_cli

=head2 off

   NACL::C::SystemServicesNdmp->off(command_interface => $ci [, node => $node]);

(Class method)
This method is used to switch off NDMP.

Uses 7Mode, Nodescope and CMode CLI/ZAPI APISet.

Invokes 'ndmpd off' command for 7Mode/Nodescope CLI.

Invokes  ndmp-node-service-disable  for CMode Zapi non-iter

Note that this can also be achieved by doing
    NACL::C::SystemServicesNdmp->modify(enable => 'false');

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< node => $string >>

Node on which to switch off NDMP.

Applicable for CMode CLI.

=item C<< '*' >>

This method takes all other options supported by
"system services ndmp modify" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub off {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params              => \@args,
        additional_spec     => {_optional_scalars(qw(node))},
        ignore_primary_keys => 1,
        cli_cmd             => 'system services ndmp modify',
    );
    $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_off_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_off_7mode_cli',
                interface => 'CLI',
                set       => '7Mode|Nodescope',
            },
            {   method    => '_off_cmode_zapi_non_iter',
                interface => 'ZAPI',
                set       => 'CMode',
                zapi_type => 'non-iter',
            },
        ],
        frontend => 'NACL::C::SystemServicesNdmp::off',
    );

    $Log->exit() if $may_exit;
} ## end sub off

sub _off_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, %opts) = @_;

    $opts{additional_args}->{'enable'} = 'false';

    # Switching off the NDMP service can be done in CMode either through
    # "system services ndmp modify -enable false" or through
    # "system services ndmp off". However, "off" is an alias of
    # "modify -enable false", hence there is no CDEF for "off".
    # Hence, here we need to use "system services ndmp modify"

    $pkg->_base_backend_cmode_cli(
        %opts,
        api => 'system_services_ndmp_modify',
    );

    $Log->exit() if $may_exit;
} ## end sub _off_cmode_cli

sub _off_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );

    delete $opts{command_interface};
    my $apiset = delete $opts{apiset};

    $apiset->ndmpd_off();
    $Log->exit() if $may_exit;
} ## end sub _off_7mode_cli

sub _off_cmode_zapi_non_iter {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    my $map = {'node' => 'node-name',};
    $pkg_or_obj->_base_backend_cmode_zapi(
        @opts,
        api => 'ndmp-node-service-disable',
        map => $map,
    );
    $Log->exit() if $may_exit;
}

=head2 get_password

    my $password_obj = NACL::C::SystemServicesNdmp->get_password(
                                command_interface => $node,
                                'password-for-non-root-user-account' => $username);

    $Log->comment('Password is: ' . $password_obj->password());


(Class method)
This method returns the password for non root user account. (7Mode/Nodescope only)
An object of type
L<CS::SystemServicesNdmpGetPassword|lib-NACL-CS-SystemServicesNdmpGetPassword-pm>
is returned which contains one field: the password.

(Note that while it might have been possible to return the password as a scalar,
we opted to return it as an object to keep consistency with other methods
which have useful output returned as CS objects)

Uses 7Mode and Nodescope CLI APISet.

This maps to the 7Mode/Nodescope CLI command "ndmpd password" which returns
the password for the specified user account. Though CMode CLI has a
"system services ndmp password" command, its purpose is completely
different. (It is used to change the NDMP password for the node) Since the
CMode and 7Mode/Nodescope commands are for different purposes, they've been
split into separate component methods. For changing the NDMP password for
the node (CMode support only) use
L<set_password|lib-NACL-C-SystemServicesNdmp-pm/set_password>.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'password-for-non-root-user-account' => $username >>

The user account for which we want the password

Applicable for 7Mode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub get_password {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = $pkg->_common_validate_with(
        params          => \@_,
        additional_spec => {
            _optional_scalars(qw (password-for-non-root-user-account))

        },
        ignore_primary_keys => 1,
    );
    my $password_obj = $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_get_password_7mode_cli',
                interface => 'CLI',
                set       => '7Mode|Nodescope',
            }
        ],
    );

    $Log->exit() if $may_exit;
    return $password_obj;
} ## end sub get_password

sub _get_password_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );

    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};

    $Log->debug("Opts to API call 'ndmp_password' are: \n" . Dumper(\%opts));
    my $response = $apiset->ndmpd_password(%opts);
    my $output   = $response->get_parsed_output();
    my $password = $output->[0]{password};
    my $obj
        = NACL::CS::SystemServicesNdmpGetPassword->new(password => $password);
    $Log->exit() if $may_exit;
    return $obj;
} ## end sub _get_password_7mode_cli

=head2 set_password

    NACL::C::SystemServicesNdmp->set_password(command_interface => $node,
                                              'new-password'    => $password
                                              [, node           => $node]);

(Class method)
This method is used to change the NDMP password for the node. (CMode only)

Uses CMode CLI/ZAPI APISet.

This maps to the CMode CLI command "system services ndmp password".

Invokes  ndmp-node-modify-password  for CMode Zapi non-iter

The 7Mode CLI command "ndmpd password" is used for a different purpose
(returns the password of the specified user account). To retrieve the
current password, the method
L<get_password|lib-NACL-C-SystemServicesNdmp-pm/get_password> should be used.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'new-password' => $password >>

The new password.

Applicable for CMode CLI.

=item C<< node => $string >>

(Optional)
The node on which to change the password.

Applicable for CMode CLI.

=item C<< '*' >>

This method takes all other options supported by
"system services ndmp password" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub set_password {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    my %opts = $pkg->_common_validate_with(
        params              => \@args,
        additional_spec     => {_optional_scalars(qw (node new-password))},
        ignore_primary_keys => 1,
        cli_cmd             => 'system services ndmp password',
    );
    $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_set_password_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_password_cmode_zapi_non_iter',
                interface => 'ZAPI',
                set       => 'CMode',
                zapi_type => 'non-iter',
            },
        ],
        frontend => 'NACL::C::SystemServicesNdmp::set_password',
    );

    $Log->exit() if $may_exit;
} ## end sub set_password

sub _set_password_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, %opts) = @_;
    my $password = delete $opts{'new-password'};

    my $prompts_answers = [
        'Please enter password:' => sub {$password},
        'Confirm password:'      => sub {$password}
    ];

    $opts{additional_args}->{'connectrec-match_table'} = $prompts_answers;

    $pkg->_base_backend_cmode_cli(
        %opts,
        api => 'system_services_ndmp_password',
    );
    
    $Log->exit() if $may_exit;
} ## end sub _setpassword_cmode_cli

sub _password_cmode_zapi_non_iter {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    my $copy = [qw(password)];
    my $map = {'node' => 'node-name',
               'new-password' => 'password',};
    $pkg_or_obj->_base_backend_cmode_zapi(
        @opts,
        api  => 'ndmp-node-modify-password',
        copy => $copy,
        map  => $map,
    );
    $Log->exit() if $may_exit;
}

=head2 modify

   NACL::C::SystemServicesNdmp->modify(command_interface => $ci, ...);

(Class method)
This method is used to modify NDMP service configuration.

Uses a CMode CLI, 7Mode/Nodescope CLIi/ZAPI APISet.

Invokes 'ndmpd on' command for 7Mode CLI.

Invokes  ndmp-node-attributes-modify  for CMode Zapi non-iter

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< node => $string >>

The node on which to modify the NDMP configurations.

Applicable for CMode CLI.

=item C<< enable => $boolean >>

NDMP service enabled?
Possible values: true; false

Note that enabling the service can also be done by doing
    NACL::C::SystemServicesNdmp->on(...);

And disabling the service can also be done by doing
    NACL::C::SystemServicesNdmp->off(...);

Applicable for CMode CLI.

=item C<< 'clear-text' => $boolean >>

Allow clear text password?
Possible values: true; false

Applicable for CMode CLI.

=item C<< 'user-id' => $string >>

NDMP User Id

Applicable for CMode CLI.

=item C<< 'password' => $string >>

NDMP password. This password is used in CMode CLI to answer
the prompts "Please enter password" and "Confirm password".

=item C<< 'common-sessions' => $string >>

NDMP Common Sessions
Applicable for CMode CLI

=item C<< '*' >>

This method takes all other options supported by
"system services ndmp modify" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub modify {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw (node enable clear-text user-id password
                    common-sessions
                   )
            )
        },
        ignore_primary_keys => 1,
        api                 => 'system_services_ndmp_modify',
    );
    $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_modify_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_modify_7mode_cli',
                interface => 'CLI',
                set       => '7Mode|Nodescope',
            },
            {   method    => '_modify_cmode_zapi_non_iter',
                interface => 'ZAPI',
                set       => 'CMode',
                zapi_type => 'non-iter',
            },
        ],
        frontend => 'NACL::C::SystemServicesNdmp::modify',
    );

    $Log->exit() if $may_exit;
} ## end sub modify

sub _modify_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    my $password          = delete $opts{password};
    if (!defined $password) {
        $password = "admin";
    }
    my $prompts_answers = [
        'Please enter password:' => sub {$password},
        'Confirm password:'      => sub {$password}
    ];
    my %additional_args;
    $additional_args{'connectrec-match_table'} = $prompts_answers;

    $pkg_or_obj->_base_backend_cmode_cli(
        %opts,
        api             => 'system_services_ndmp_modify',
        additional_args => \%additional_args
    );
    $Log->exit() if $may_exit;
} ## end sub _modify_cmode_cli

sub _modify_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg = shift;

    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );

    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};

    my $enable = delete $opts{enable};

    # The options 'node', 'enable', 'clear-text' and 'user-id' can be
    # modified in CMode. The only thing that can be modified in 7Mode is
    # 'enable', which can be done through ndmpd_on() or ndmpd_off()
    if (keys %opts) {
        $Log->debug('The options ' . join ',',
            (keys %opts) . 'were provided but cannot be modified in 7Mode');
    } ## end if ( keys %opts )

    if ($enable eq 'true') {
        $apiset->ndmpd_on();
    } elsif ($enable eq 'false') {
        $apiset->ndmpd_off();
    } else {
        $Log->exit() if $may_exit;
        NATE::Exceptions::Argument->throw("Option 'enable' should be either "
                . "'true' or 'false' but it was passed as '$enable'");
    }

    $Log->exit() if $may_exit;
} ## end sub _modify_7mode_cli

sub _modify_cmode_zapi_non_iter {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    my $copy = [qw(password enable user-id clear-text)];
    my $map = {'node' => 'node-name',};
    $pkg_or_obj->_base_backend_cmode_zapi(
        @opts,
        api  => 'ndmp-node-attributes-modify',
        copy => $copy,
        map  => $map,
    );
    $Log->exit() if $may_exit;
}

=head2 show

    my @objs = NACL::C::SystemServicesNdmp->show(command_interface => $ci, ...);

                           or

    my $obj  = NACL::C::SystemServicesNdmp->show(command_interface => $ci, ...);

(Class method)
This method is used to display NDMP service configuration.

Uses a CMode, 7Mode and Nodescope CLI APISet.

This method returns objects of the type
L<NACL::CS::SystemServicesNdmpShow|lib-NACL-CS-SystemServicesNdmpShow-pm>. Each
object represents the NDMP service configurations of a node in the cluster.

Note that in 7Mode/Nodescope the only field of information we can fill in is
whether the NDMP service is enabled or not.

This method provides the same interface as fetch methods on CS objects do
(i.e. the input parameters filter, requested_fields and allow_empty are
supported; a NoElementsFound exception is thrown if no elements are found;
the method can be invoked in either scalar or list context).

=over

=item Options

The same options accepted by CS::fetch are accepted by this method. A detailed
explanation of all the parameters supported can be found in
L<ComponentState::fetch|lib-NACL-CS-ComponentState-pm/fetch>.

=back

=cut

sub show {
    $Log->enter() if $may_enter;
    my $pkg = shift;
    if (wantarray) {
        my @state_objs = NACL::CS::SystemServicesNdmpShow->fetch(@_);
        $Log->exit() if $may_exit;
        return @state_objs;
    } else {
        my $state_obj = NACL::CS::SystemServicesNdmpShow->fetch(@_);
        $Log->exit() if $may_exit;
        return $state_obj;
    }
} ## end sub show

=head2 start

    my $SystemServicesNdmp_obj    = NACL::C::SystemServicesNdmp->start(
                                command_interface => $ci,
                                'node' => $node,
                                ...);

(Class method)
This method is used to start the NDMP service.

Uses a CMode CLI APISet.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'node' => $node >>

(Required)

Applicable for CMode CLI

=item C<< 'method-timeout' => $time_in_seconds >>

(Optional)
The default NATE timeout is 60 seconds. If the command might take more
time to complete, this option can be used to specify a larger timeout value.
The value should be provided in seconds.

=item C<< 'extended_query' => $string >>

(Optional)

Applicable for CMode CLI

=item C<< '*' >>

This method takes all other options supported by
"system services ndmp start" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub start {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params     => \@args,
        cli_cmd    => 'system services ndmp start',
    );
    my $obj = $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_start_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::SystemServicesNdmp::start',
    );

    $Log->exit() if $may_exit;
} ## end sub start

sub _start_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;

    $pkg->_base_backend_cmode_cli(
        @args,
        api => 'system_services_ndmp_start',
    );

    $Log->exit() if $may_exit;
} ## end sub _start_cmode_cli

=head2 terminate

    NACL::C::SystemServicesNdmp->terminate(
                                command_interface => $ci,
                                'node' => $node,
                                ...);

(Class only method)
This method is used to terminate all NDMP sessions.

Uses a CMode CLI APISet.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'node' => $node >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI

=item C<< 'method-timeout' => $time_in_seconds >>

(Optional)
The default NATE timeout is 60 seconds. If the command might take more
time to complete, this option can be used to specify a larger timeout value.
The value should be provided in seconds.

=item C<< 'extended_query' => $string >>

(Optional)

Applicable for CMode CLI

=item C<< '*' >>

This method takes all other options supported by
"system services ndmp terminate" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub terminate {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params     => \@args,
        cli_cmd    => 'system services ndmp terminate',
    );
    $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_terminate_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::SystemServicesNdmp::terminate',
    );

    $Log->exit() if $may_exit;
} ## end sub terminate

sub _terminate_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;

    $pkg_or_obj->_base_backend_cmode_cli(
        @args,
        api => 'system_services_ndmp_terminate',
    );

    $Log->exit() if $may_exit;
}  ## end sub _terminate_cmode_cli

=head2 stop

    NACL::C::SystemServicesNdmp->stop(
                                command_interface => $ci,
                                'node' => $node,
                                ...);

(Class only method)
This method is used to stop the NDMP service.

Uses a CMode CLI APISet.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'node' => $node >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI.

=item C<< 'method-timeout' => $time_in_seconds >>

(Optional)
The default NATE timeout is 60 seconds. If the command might take more
time to complete, this option can be used to specify a larger timeout value.
The value should be provided in seconds.

=item C<< 'extended_query' => $string >>

(Optional)

Applicable for CMode CLI.

=item C<< '*' >>

This method takes all other options supported by
"system services ndmp stop" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub stop {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params     => \@args,
        cli_cmd    => 'system services ndmp stop',
    );
    $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_stop_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::SystemServicesNdmp::stop',
    );

    $Log->exit() if $may_exit;
} ## end sub stop

sub _stop_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;

    $pkg_or_obj->_base_backend_cmode_cli(
        @args,
        api => 'system_servcies_ndmp_stop',
    );

    $Log->exit() if $may_exit;
} ## end sub _stop_cmode_cli

=head2 password

    NACL::C::SystemServicesNdmp->password(
                                command_interface => $ci,
                                'node' => $node,
                                ...);

(Class only method)

This method is used to change the NDMP password for the node.

Uses a CMode CLI APISet.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'node' => $node >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI.

=item C<< 'method-timeout' => $time_in_seconds >>

(Optional)
The default NATE timeout is 60 seconds. If the command might take more
time to complete, this option can be used to specify a larger timeout value.
The value should be provided in seconds.

=item C<< 'extended_query' => $string >>

(Optional)

Applicable for CMode CLI.

=item C<< '*' >>

This method takes all other options supported by
"system services ndmp password" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub password {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params     => \@args,
        cli_cmd    => 'system services ndmp password',
    );
    $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_password_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::SystemServicesNdmp::password',
    );

    $Log->exit() if $may_exit;
} ## end sub password

sub _password_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;

    $pkg_or_obj->_base_backend_cmode_cli(
        @args,
        api => 'system_services_ndmp_password',
    );

    $Log->exit() if $may_exit;
} ## end sub _password_cmode_cli

=head2 status

    my $status_obj = NACL::C::SystemServicesNdmp->status(command_interface => $ci,
                          [ filter => { %filter } ,]
              [ requested_fields => [ @req_fields ]);

                    or

    my @status_objs = NACL::C::SystemServicesNdmp->status(command_interface => $ci,
                          [ filter => { %filter } ,]
              [ requested_fields => [ @req_fields ]);

                    or

   my $status_obj = $SystemServicesNdmp_obj->status();

(Class or Instance method)
This method returns objects of the type L<NACL::CS::SystemServicesNdmpStatus|lib-NACL-CS-SystemServicesNdmpStatus-pm>.

This method accepts all the parameters accepted by L<fetch|lib-NACL-CS-ComponentState-pm/fetch>
(requested_fields, filter, allow_empty) and has all of the other properties of
fetch methods (callable in scalar or list context, throws a NoElementsFound
exception if no matching elements are found)

=cut

sub status {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = $pkg_or_obj->_handle_obj_cs_like_calls(@_);
    require NACL::CS::SystemServicesNdmpStatus;
    my @state_objs = NACL::CS::SystemServicesNdmpStatus->fetch(%opts);
    $Log->exit() if $may_exit;
    return wantarray ? @state_objs : $state_objs[0];
} ## end sub status

=head2 version

    NACL::C::SystemServicesNdmp->version(
                                command_interface => $ci,
                                'node' => $node,
                                ...);

                                 or

    $SystemServicesNdmp_obj->version(...);

(Class or instance method)
This method is used to display default NDMP version.

Uses a CMode CLI APISet.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'node' => $node >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI.

=item C<< 'maxversion' => $string >>

Applicable for CMode CLI.

=item C<< 'method-timeout' => $time_in_seconds >>

(Optional)
The default NATE timeout is 60 seconds. If the command might take more
time to complete, this option can be used to specify a larger timeout value.
The value should be provided in seconds.

=item C<< 'extended_query' => $string >>

(Optional)

Applicable for CMode CLI.

=item C<< '*' >>

This method takes all other options supported by
"system services ndmp version" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub version {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( maxversion
                    )
            ),
        },
        cli_cmd => 'system services ndmp version',
    );
    $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_version_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::SystemServicesNdmp::version',
    );

    $Log->exit() if $may_exit;
} ## end sub version

sub _version_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;

    $pkg_or_obj->_base_backend_cmode_cli(
        @args,
        api => 'system_services_ndmp_version',
    );

    $Log->exit(); 
} ## end sub _version_cmode_cli

=head2 kill_all

    NACL::C::SystemServicesNdmp->kill_all(
                                command_interface => $ci,
                                'node' => $node,
                                ...);

                                 or

    $SystemServicesNdmp_obj->kill_all(...);

(Class or instance method)
This method is used to kill all NDMP sessions.

Uses a CMode CLI/ZAPI APISet.

Invokes  ndmp-node-session-kill-all  for CMode Zapi non-iter

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'node' => $node >>

(Required for class method, Not Applicable for instance method)
Node

Applicable for CMode CLI.

=item C<< 'method-timeout' => $time_in_seconds >>

(Optional)
The default NATE timeout is 60 seconds. If the command might take more
time to complete, this option can be used to specify a larger timeout value.
The value should be provided in seconds.

=item C<< 'extended_query' => $string >>

(Optional)

Applicable for CMode CLI.

=item C<< '*' >>

This method takes all other options supported by
"system services ndmp kill-all" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub kill_all {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params              => \@args,
        additional_spec     => {_optional_scalars(qw(session-id))},
        cli_cmd             => 'system services ndmp kill-all',
    );
    $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_kill_all_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_kill_all_cmode_zapi_non_iter',
                interface => 'ZAPI',
                set       => 'CMode',
                zapi_type => 'non-iter',
            },
        ],
        frontend => 'NACL::C::SystemServicesNdmp::kill_all',
    );

    $Log->exit() if $may_exit;
} ## end sub kill_all

sub _kill_all_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;

    $pkg_or_obj->_base_backend_cmode_cli (
        %opts,
        api => 'system_services_ndmp_kill_all',
    );

    $Log->exit() if $may_exit;
} ## end sub _kill_all_cmode_cli

sub _kill_all_cmode_zapi_non_iter {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    my $map = {'node' => 'node-name',};
    $pkg_or_obj->_base_backend_cmode_zapi(
        @opts,
        api => 'ndmp-node-session-kill-all',
        map => $map,
    );
    $Log->exit() if $may_exit;
}

sub _check_non_iter {
    $Log->enter() if $may_enter;
    my ($pkg, @opts) = @_;
    $pkg->_base_check_non_iter(@opts, _primary_keys => [qw(node)],);
    $Log->exit() if $may_exit;
}

=head2 killsession

    NACL::C::SystemServicesNdmp->killsession(
                                command_interface => $ci,
	                            'node' => $node,
                                ...);
	                             
								 or

    $SystemServicesNdmp_obj->killsession(...);

(Class or instance method)
This method is used to kill the specified NDMP session.

Uses a CMode CLI/ZAPI APISet.

Invokes  ndmp-node-session-kill  for CMode Zapi non-iter

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'node' => $node >>

(Required for class method, Not Applicable for instance method)
Node possible value(s) are, <nodename>,local

Applicable for CMode CLI

=item C<< 'session-id' => $string >>

Session Identifier possible value(s) are, integer

Applicable for CMode CLI

=item C<< 'cmd_execution_time' => \$scalar >>

(Optional) Scalar reference , it will be updated with the command execution
time returned by the apiset response object

=item C<< 'method-timeout' => $time_in_seconds >>

(Optional)
The default NATE timeout is 60 seconds. If the command might take more
time to complete, this option can be used to specify a larger timeout value.
The value should be provided in seconds.

=item C<< '*' >>

This method takes all other options supported by
"system services ndmp killsession" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub killsession {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            'cmd_execution_time' => {type => SCALARREF, optional => 1},
            _optional_scalars(
                qw( session-id
                    )
            ),
        },
        cli_cmd => 'system services ndmp killsession',
    );
    $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_killsession_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_killsession_cmode_zapi_non_iter',
                interface => 'ZAPI',
                set       => 'CMode',
                zapi_type => 'non-iter',
            },
        ],
        frontend => 'NACL::C::SystemServicesNdmp::killsession',
    );

    $Log->exit() if $may_exit;
} ## end sub killsession

sub _killsession_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;

    $pkg_or_obj->_base_backend_cmode_cli(
        @opts,
        api => 'system_services_ndmp_killsession',
    );

    $Log->exit() if $may_exit;
} ## end sub _killsession_cmode_cli

sub _killsession_cmode_zapi_non_iter {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    my $copy = [qw(session-id)];
    my $map = {'node' => 'node-name',};
    $pkg_or_obj->_base_backend_cmode_zapi(
        @opts,
        api  => 'ndmp-node-session-kill',
        copy => $copy,
        map  => $map,
    );
    $Log->exit() if $may_exit;
}

=head2 maxversion

    NACL::C::SystemServicesNdmp->maxversion(
                                command_interface => $ci,
	                            'node' => $node,
                                ...);
	                             
								 or

    $SystemServicesNdmp_obj->maxversion(...);

(Class or instance method)
This method is used to display default NDMP version.

Uses a CMode CLI APISet.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'node' => $node >>

(Required for class method, Not Applicable for instance method)
Node possible value(s) are, <nodename>,local

Applicable for CMode CLI

=item C<< 'maxversion' => $string >>

NDMP Version possible value(s) are, integer

Applicable for CMode CLI

=item C<< 'cmd_execution_time' => \$scalar >>

(Optional) Scalar reference , it will be updated with the command execution
time returned by the apiset response object

=item C<< 'method-timeout' => $time_in_seconds >>

(Optional)
The default NATE timeout is 60 seconds. If the command might take more
time to complete, this option can be used to specify a larger timeout value.
The value should be provided in seconds.

=item C<< '*' >>

This method takes all other options supported by
"system services ndmp maxversion" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub maxversion {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            'cmd_execution_time' => {type => SCALARREF, optional => 1},
            _optional_scalars(
                qw( maxversion
                    )
            ),
        },
        cli_cmd => 'system services ndmp maxversion',
    );
    $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_maxversion_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::SystemServicesNdmp::maxversion',
    );

    $Log->exit() if $may_exit;
} ## end sub maxversion

sub _maxversion_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;

    $pkg_or_obj->_base_backend_cmode_cli(
        @opts,
        api => 'system_services_ndmp_maxversion',
    );

    $Log->exit() if $may_exit;
} ## end sub _maxversion_cmode_cli


__PACKAGE__->_load_event_class();

sub _primary_keys_validate_spec {
    return (
        'session-id' => {type => SCALAR},
        'node'       => {type => SCALAR, optional => 1},
    );
} ## end sub _primary_keys_validate_spec

1;
