# $Id: $

#
# Copyright (c) 2001-2012 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Storage Unit Conversion Routines
## @author  steve.latif@netapp.com, dl-nacl-dev@netapp.com
## @status  Shared
## @pod here

=head1 NAME

NACL::C::UnitNormalization

=cut

package NACL::C::UnitNormalization;

use strict;
use warnings;

use NATE::BaseException qw(:try);
use NATE::Exceptions::Argument;
use Params::Validate qw(validate SCALAR validate_pos);
use Tharn qw(script_result );
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Module::Common;
use Data::Dumper;
use Math::BigFloat;
my ( %unit2bytes, %abbrev2unit, %unit2abbrev );
our $ALLOWED_UNITS_RE;
my $SIZE_WO_UNIT_RE = qr/[+-]?\d+(?:\.\d+)?/;

BEGIN {

    #               0 1  2  3  4
    my @bytes = qw( B KB MB GB TB );
    my @constants;

    # generate the constants and calculate the multiplier
    # for each unit, which should be 1024 ** $idx
    # (B => 1, KB => 1024, MB => 1024 ** 2... etc)
    #
    # so we're basically doing UNIT_B => 'B', UNIT_KB => 'KB',
    # and $unit2bytes{B} = 1, $unit2bytes{KB} => 1024 ..
    require constant;
    for my $idx ( 0 .. $#bytes ) {
        my $unit   = $bytes[$idx];
        my $symbol = "UNIT_$unit";

        # create constant...
        constant->import( $symbol, $unit );
        push @constants, $symbol;

        # calculate appropriate factor..
        $unit2bytes{$unit} = 1024**$idx;

        # now for convenience, map abbreviation
        # (first letter) to actual constant
        my $abbrev1 = lc( substr( $unit, 0, 1 ) );
        my $abbrev2 = lc($unit);
        $abbrev2unit{$abbrev1} = $unit;
        $abbrev2unit{$abbrev2} = $unit;
        $unit2abbrev{$unit}    = $abbrev1;
    } ## end for my $idx ( 0 .. $#bytes)

    $ALLOWED_UNITS_RE = qr!(?:(?i)(b|[kmgt]b?))!;

    # XXX Modify this to explicitly export the right things.
    Module::Common::setup_export_tags(
        convert => [
            qw(
                guess_unit
                convert_units
                to_unit
                convert_to_bytes
                convert_to_bytes_int
                )
        ],
        constants => \@constants
    );
} ## end BEGIN

BEGIN {
    Params::Validate::validation_options(
        on_fail => sub {
            my @msgs = @_;
            chomp(@msgs);
            NATE::Exceptions::Argument->throw( script_result(@msgs) );
        }
    );
} ## end BEGIN

=head1 METHODS 

=head2 to_unit($size,$convertSizeToUnit)
	
	$converted_size= NACL::C::UnitNormalization::to_unit($size,$convertSizeToUnit);

Converts $size into $ConvertSizeToUnit and returns the $convertedSize.
Possible values for $ConvertSizeToUnit can be any of "b,k|kb,m|mb,g|gb,t|tb".
*If there is no suffix like b,k,m,g,t,kb,mb,gb,tb in $size then by default,
$size will be considered in Bytes. 

=head2 to_bytes($size)

	$size_in_bytes =NACL::C::UnitNormalization::to_bytes($size);

Returns $size in Bytes.possible values that $size can take is ,
"b,k|kb,m|mb,g|gb,t|tb".
*If there is no suffix like b,k,m,g,t,kb,mb,gb,tb in,
$size then by default $size will be considered in Bytes.

=head2 to_kilobytes($size)

	$size_inkilo_bytes =NACL::C::UnitNormalization::to_kilobytes($size);	

Returns $size in Kilobytes.Possible values that $size can take is,
"b,k|kb,m|mb,g|gb,t|tb".
*If there is no suffix like b,k,m,g,t,kb,mb,gb,tb in,
$size then by default $size,will be considered in Bytes.


=head2 to_megabytes($size)

        $size_in_megabytes =NACL::C::UnitNormalization::to_megabytes($size);

Returns  $size in Megabytes.Possible values that $size can take is,
"b,k|kb,m|mb,g|gb,t|tb".
*If there is no suffix like b,k,m,g,t,kb,mb,gb,tb in,
$size then by default $size will be considered in Bytes.


=head2 to_gigabytes($size)

        $size_in_gigabytes =NACL::C::UnitNormalization::to_gigabytes($size);

Returns  $size in Gigabytes.Possible values that $size can take is "b,k|kb,m|mb,g|gb,t|tb".
*If there is no suffix like b,k,m,g,t,kb,mb,gb,tb in  $size then by default $size
will be considered in Bytes.

=head2 to_terabytes($size)

        $size_in_terabytes =NACL::C::UnitNormalization::to_terabytes($size);

Returns $size in Terabytes.Possible values that $size can ,
take is "b,k|kb,m|mb,g|gb,t|tb".
*If there is no suffix like b,k,m,g,t,kb,mb,gb,tb in $size,
then by default $size will be considered in Bytes.

=head2 normalize($size,$magnitude)
	
	NACL::C::UnitNormalization::normalize($size,$magnitude);
		
Returns $size without suffix if $magnitude is pased as 'i'.
Returns $size,If suffix passed in $size matches with $magnitude 
Returns $size in bytes without any suffix ,If suffix Passed in $size,
does matches with $magnitude. 
* Possibe values $magnitude can have are 'i,k,b,m,g,t,kb,mb,gb,tb'. 

=cut

sub to_unit {
    my $size = $_[0];
    my $unit = $abbrev2unit{ lc( $_[1] ) };
    my $convert;

    my $guess = guess_unit($size);

    unless ($guess) {
        NATE::Exceptions::Argument->throw(
            "Could not guess the number $size: $!");
    }
    my $c = convert_to_bytes($size);
    $convert = convert_units( strip_unit_spec($c), UNIT_B, $unit );
    $convert .= $unit2abbrev{$unit};
    return $convert;
} ## end sub to_unit

sub to_bytes {
    my $size  = $_[0];
    my $bytes = convert_to_bytes($size);
    $bytes .= 'B';
    return $bytes;
} ## end sub to_bytes

sub to_kilobytes {
    my $size  = $_[0];
    my $kb    = "";
    my $guess = guess_unit($size);
    if ( guess_unit($size) eq UNIT_KB ) {
        return $size;
    } else {

        my $bytes = convert_to_bytes($size);
        $kb = convert_units( strip_unit_spec($bytes), UNIT_B, UNIT_KB );
        $kb .= 'KB';
    }
    return $kb;
} ## end sub to_kilobytes

sub to_megabytes {
    my $size  = $_[0];
    my $mb    = "";
    my $guess = guess_unit($size);

    if ( guess_unit($size) eq UNIT_MB ) {
        return $size;
    } else {
        my $bytes = convert_to_bytes($size);
        $mb = convert_units( strip_unit_spec($bytes), UNIT_B, UNIT_MB );
        $mb .= 'MB';
    }
    return $mb;
} ## end sub to_megabytes

sub to_gigabytes {
    my $size = $_[0];
    my $gb;
    if ( guess_unit($size) eq UNIT_GB ) {
        return $size;
    } else {
        my $bytes = convert_to_bytes($size);
        $gb = convert_units( strip_unit_spec($bytes), UNIT_B, UNIT_GB );
        $gb .= 'GB';
    }
    return $gb;
} ## end sub to_gigabytes

sub to_terabytes {
    my $size = $_[0];
    my $tb;
    if ( guess_unit($size) eq UNIT_TB ) {
        return $size;
    } else {
        my $bytes = convert_to_bytes($size);
        $tb = convert_units( strip_unit_spec($bytes), UNIT_B, UNIT_TB );
        $tb .= 'TB';
    }
    return $tb;
} ## end sub to_terabytes

##
## Stolen from Utils::Units
##

sub convert_units {
    my ( $value, $from, $to ) = validate_pos(
        @_,
        { regex => qr/^$SIZE_WO_UNIT_RE$/ },
        {   callbacks => {
                'valid unit' => sub { exists $unit2bytes{ $_[0] } }
            }
        },
        {   callbacks => {
                'valid unit' => sub {
                    exists $unit2bytes{ $_[0] };
                    }
            }
        }
    );

    if ( $from eq $to ) {
        return $value;
    }
    my $factor_from = $unit2bytes{$from};
    my $factor_to   = $unit2bytes{$to};

    # first convert to bytes, and then convert to whatever
    # the user is asking for...
    my $bytes = $value * $factor_from;
    my $ret   = $bytes / $factor_to;

    #Create a  BigFloat object ,using the computed value
    #This will convert exponential form to floating point number
    $ret = Math::BigFloat->new($ret);

    #Convert BigFloat object to string and return
    return $ret->bstr();
} ## end sub convert_units

=head2 guess_unit($string)

Given a string that contains something like "100k", guesses the units
that the string is represented in and returns it. On failure returns
undef, setting $@.

If $string does not end with a known unit character, assumes that 
$string is in bytes.

=cut

sub guess_unit {
    my ($size) = validate_pos( @_, { type => SCALAR } );

    my $matched;
    if ( $size !~ /^$SIZE_WO_UNIT_RE\s*($ALLOWED_UNITS_RE)?$/ ) {
        NATE::Exceptions::Argument->throw(
            "Unable to guess the unit for $size");
    }
    $matched = $1;

    # XXX - $matched can be from 2 characters to 0 characters.
    # if 2 characters, we need to get the first character only
    if ( $matched && length($matched) > 1 ) {
        $matched = lc( substr( $matched, 0, 1 ) );
    }

    my $unit;
    if ( !$matched ) {
        $unit = UNIT_B;
    } else {
        $unit = $abbrev2unit{ lc($matched) };
    }

    return $unit;
} ## end sub guess_unit

=head2 convert_to_bytes

    use NACL::C::UnitNormalization qw(convert_to_bytes);
    ...
    my $val_in_bytes = convert_to_bytes($size);

Returns the value of $size in bytes. The value returned with not have the
unit suffix "B", it will simply be an integer value.

=cut

sub convert_to_bytes {
    $Log->enter() if $may_enter;
    my ($size) = validate_pos(@_, { type => SCALAR });
    my $unit   = guess_unit($size);

    # strip out non-numbers from $size's end
    my $value_in_bytes = convert_units( strip_unit_spec($size), $unit, UNIT_B );
    $Log->exit() if $may_exit;
    return $value_in_bytes;
} ## end sub convert_to_bytes($)

=head2 convert_to_bytes_int

    use NACL::C::UnitNormalization qw(convert_to_bytes_int);
    ...
    my $val_in_bytes = convert_to_bytes_int($size);

Returns the integer value of $size in bytes. The value returned with not have the
unit suffix "B", it will simply be an integer value.

=cut

sub convert_to_bytes_int {
    return int (convert_to_bytes($_[0]));
} ## end convert_to_bytes_int


sub strip_unit_spec {
    my ($v) = validate_pos( @_, { type => SCALAR } );
    $v =~ s/\s*$ALLOWED_UNITS_RE$//;
    return $v;
}

=head2 _get_suffix($string)

Given a string that contains something like "100k", return the suffix. If it is an
integer return false.

=cut

sub _get_suffix {
    my ($size) = validate_pos( @_, { type => SCALAR } );
    if ( $size eq "" || $size !~ /^$SIZE_WO_UNIT_RE\s*($ALLOWED_UNITS_RE)?$/ ) {
        NATE::Exceptions::Argument->throw("Unable to parse string: $size");
    }
    my $matched = $1;
    return $matched;
}

=head2 _remove_suffix($string)

Given a string that contains something like "100k", return the string without the suffix. 
If it has no suffix then just return the string.

=cut

sub _remove_suffix {
    my ($size) = validate_pos( @_, { type => SCALAR } );
    if ( $size eq '' || $size !~ /^[+-]?(\d+(?:\.\d+)?)$ALLOWED_UNITS_RE?$/ )
    {
        NATE::Exceptions::Argument->throw("Unable to parse string: $size");
    }
    my $matched = $1;
    return $matched;
}

sub normalize {

    my ( $size, $magnitude )
        = validate_pos( @_, { type => SCALAR }, { type => SCALAR } );

# Nasty little hack to deal with those situations where we just
# need to send a value with no suffix. BR/RR are currently inconsistent
# about how they handle units. Volume create expects suffixes, aggr create
# does not and assumes everything is GB, coverting to bytes would not be good.
    my $suffix;
    if ( $magnitude eq 'i' ) {
        if ( $suffix = _get_suffix($size) ) {
            return _remove_suffix($size);
        } else {
            return $size;
        }
    } else {
        if ( $magnitude eq "" || $magnitude !~ /^$ALLOWED_UNITS_RE?$/ ) {
            NATE::Exceptions::Argument->throw(
                "Unable to parse string: $magnitude");
        }
        my $short_mag = substr( $magnitude, 0, 1 );
        if ( $suffix = _get_suffix($size) ) {
            my $val;
            my $short_suffix = substr( $suffix, 0, 1 );
            if ( substr( $suffix, 0, 1 ) !~ /$short_mag/i ) {
                my $bytes = to_unit( $size, 'b' );
                return _remove_suffix($bytes);
            } else {
                return $size;
            }
        }

     # The $size paramter does not have a suffix, so we just return the value.
        else {
            return $size;
        }
    }
}
1;
