# $Id$

#
# Copyright (c) 2014 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary VPserver Component Module
## @author dl-nacl-dev@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::C::VPserver

=head1 DESCRIPTION

C<NACL::C::VPserver> is a derived class of
L<NACL::C::Component::VPserver|lib-NACL-C-Component-VPserver-pm>. It represents access
to a VPserver.

=head1 ATTRIBUTES

=head2 command_interface

See L<NACL::C::VPserver|lib-NACL-C-VPserver-pm>

=head2 credential

A VPserver credential object of type L<NATE::SOAP::Data|nate-lib-NATE-SOAP-Data-pm>

=cut

package NACL::C::VPserver;

use strict;
use warnings;

use base 'NACL::C::SOAPCommandInterface';
use Params::Validate qw(validate_with SCALAR ARRAYREF SCALARREF);
use NATE::Log qw(log_global);

use NATE::BaseException qw(:try);
use NACL::APISet::Exceptions::InvalidParamValueException;
use NACL::APISet;
use NATE::SOAP::Data;
use Hostrec;

=head1 METHODS

=head2 new

    use NACL::C::VPserver;

    # implicitly specify the credential
    my $vpserver_interface_obj = NACL::C::VPserver->new(vpserver => $vpserver_name);

    # explicitly define the credential if you need to override the username & password
    use NATE::SOAP::Data;
    my $credential = NATE::SOAP::Data->new(credential => {VPUsername => "VPserver",
                                                          VPPassword => "secretpassword",
                                                          applicationName => "my_application_name"});
    my $vpserver_interface_obj = NACL::C::VPserver->new(vpserver => $vpserver_name,
                                                        credential => $credential);

(Class method)
This method is used to create a new VPserver object. It is the command interface used for
all NACL VPserver Components.

=over

=item Options

=over

=item C<< vpserver => $vpserver >>

(Required) The name of the vpserver to create the VPserver component for.

=item C<< credential => $credential_object >>

(Optional) The credential object parameter to be used for this call against the vpserver. See L<NATE::SOAP::Data|nate-lib-NATE-SOAP-Data-pm>.
If not specified, a credential object will be generated for this vpserver.

=item C<< timeout => $timeout >>

(Optional) The timeout value to use for all calls against the vpserver.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

use Class::MethodMaker [
        new => ['-hash','-init', 'new'],
        scalar => 'vpserver',
        scalar => 'timeout',
        scalar => [
            {-type => 'NATE::SOAP::Data'},
            'credential'
        ],
        scalar => [
            {-type => 'NACL::APISet::Host::SOAP::VPserver'},
            'command_interface'
        ],
];
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();

sub init {
    $Log->enter() if $may_enter;
    my ($pkg, %opts) = @_;

    if (!$opts{vpserver}) {
        NACL::APISet::Exceptions::MissingRequiredParamException->throw(text => "Mandatory parameter 'vpserver' missing in call to NACL::C::VPserver::new.");
    }

    my %create_opts;
    my $host = Hostrec->new(id => $opts{vpserver},
                            hosttype => 'svp');

    my $connection = $host->connect(template => 'soap');
    my $credential = $opts{credential};
    if (!$credential) {
        my $username = $opts{username} // $connection->username;
        my $password = $opts{password} // $connection->password;

        $credential = NATE::SOAP::Data->new(credential => {VPUsername => $username,
                                                           VPPassword => $password,
                                                           applicationName => 'nacl_component_vpserver_interface'},
                                           );
    }
    $create_opts{credential} = $credential;
    $pkg->credential($credential);

    my $timeout = $opts{timeout} // 180;
    $pkg->timeout($timeout);

    $create_opts{hostobj} = $host;
    $create_opts{category} = 'Host';
    $create_opts{interface} = 'SOAP';
    $create_opts{set} = 'VPserver';

    my $api_set_obj;
    try {
        $api_set_obj = NACL::APISet->new(%create_opts);
        $pkg->command_interface($api_set_obj);
    } otherwise {
        my $exception = shift;
        $exception->throw();
    };

    $Log->exit() if $may_exit;
    return $pkg;
}

=head2 call

    (Class Call)
    my $param1 = NATE::SOAP::Data->new(fullDetail => "true");
    ### or ###
    my $param1 = {fullDetail => "true"};

    my $ci = NACL::APISet->new(
        hostobj   => $host,
        category  => "Host",
        interface => "SOAP",
        set       => "VPserver",
    );
    my $result = NACL::C::VPserver->call(command_interface => $ci,
                                         command => 'clusterList',
                                         params => [$param1]);

    (Instance Call)
    use NACL::C::VPserver;
    my $vpserver_command_interface = NACL::C::VPserver->new(vpserver => $vpserver_name);

    use NATE::SOAP::Data;
    my $param1 = NATE::SOAP::Data->new(fullDetail => "true");
    ### or ###
    my $param1 = {fullDetail => "true"};

    my $result = $vpserver_command_interface->call(command => 'clusterList',
                                                   params  => [$param1]);

This method is used to manually send a SOAP API call to the vpserver.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method) See L<NACL::APISet::Host::SOAP::VPserver|lib-NACL-APISet-Host-SOAP-VPserver-pm>.

=item C<< command => $command >>

(Required) The command we wish to execute against the vpserver.

=item C<< params => \@params >>

(Optional) The arrayref of parameters for the command to be run against the vpserver.

=item C<< credential => $credential >>

(Optional) The credential object parameter to be used for this call against the vpserver. See L<NATE::SOAP::Data|nate-lib-NATE-SOAP-Data-pm>.
This defaults to the credential object stored on the vpserver command interface.

=item C<< timeout => $timeout >>

(Optional) The timeout value to use for this call against the vpserver.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub call {
    $Log->enter() if $may_enter;
    my ($pkg, %opts) = @_;

    $pkg->command_interface($opts{command_interface}) if (!$pkg->command_interface);
    if (!$pkg->command_interface) {
        NACL::APISet::Exceptions::MissingRequiredParamException->throw(text => "Mandatory parameter 'command_interface' missing in call to NACL::C::VPserver::call.");
    }

    if (!$opts{command}) {
        NACL::APISet::Exceptions::MissingRequiredParamException->throw(text => "Mandatory parameter 'command' missing in call to NACL::C::VPserver::call.");
    }

    if ($opts{credential_overrides}) {
        $pkg->_override_credential(credential_overrides => $opts{credential_overrides});
    }

    if (!$opts{credential}) {
        $opts{credential} = $pkg->credential;
    }

    $opts{timeout} //= $pkg->timeout;

    my $parsed_output = $pkg->SUPER::call(%opts);

    $Log->exit() if $may_exit;
    return $parsed_output;
}

=head2 register_with_vcenter_vsc

    (Class Call)

    my $ci = NACL::APISet->new(
        hostobj   => $host,
        category  => "Host",
        interface => "SOAP",
        set       => "VPserver",
    );
    my $result = NACL::C::VPserver->register_with_vcenter_vsc(command_interface => $ci,
                                                              vcenter_ip => $vcenter_ip,
                                                              vcenter_username => $username,
                                                              vcenter_password => $password,
                                                              vsc_ip => $vsc_ip);

    (Instance Call)
    use NACL::C::VPserver;
    my $vpserver_command_interface = NACL::C::VPserver->new(vpserver => "myvpserver");
    my $register_with_vcenter_vsc_result  = $vpserver_command_interface->register_with_vcenter_vsc(vcenter_ip => $vcenter_ip,
                                                                                                   vcenter_username => $username,
                                                                                                   vcenter_password => $password,
                                                                                                   vsc_ip => $vsc_ip);

The method registers the VPserver with the specified Vcenter and VSC servers.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method) See L<NACL::APISet::Host::SOAP::VPserver>.

=item C<< vcenter_ip => $vcenter_ip >>

(Required) The ip address or hostname of the vcenter to register with.

=item C<< vcenter_username => $username >>

(Required) The username for the vcenter to register with.

=item C<< vcenter_password => $password >>

(Required) The password for the vcenter to register with.

=item C<< vsc_ip => $vsc_ip >>

(Required) The ip address or hostname of the vsc to register with.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub register_with_vcenter_vsc {
    $Log->enter() if $may_enter;
    my ($pkg, %opts) = @_;

    $pkg->command_interface($opts{command_interface}) if (!$pkg->command_interface);
    if (!$pkg->command_interface) {
        NACL::APISet::Exceptions::MissingRequiredParamException->throw(text => "Mandatory parameter 'command_interface' missing in call to NACL::C::VPserver::call.");
    }

    if (!$opts{vcenter_ip}) {
        NACL::APISet::Exceptions::MissingRequiredParamException->throw(text => "Mandatory parameter 'vcenter_ip' missing in call to NACL::C::VPserver::call.");
    }

    if (!$opts{vcenter_username}) {
        NACL::APISet::Exceptions::MissingRequiredParamException->throw(text => "Mandatory parameter 'vcenter_username' missing in call to NACL::C::VPserver::call.");
    }

    if (!$opts{vcenter_password}) {
        NACL::APISet::Exceptions::MissingRequiredParamException->throw(text => "Mandatory parameter 'vcenter_password' missing in call to NACL::C::VPserver::call.");
    }

    if (!$opts{vsc_ip}) {
        NACL::APISet::Exceptions::MissingRequiredParamException->throw(text => "Mandatory parameter 'vsc_ip' missing in call to NACL::C::VPserver::call.");
    }

    my @params;
    my $credential_overrides;

    # Perform Validation
    if ($opts{vcenter_ip}) {
        push(@params, {vCenterIP => $opts{vcenter_ip}});
        if ($opts{vcenter_username} and $opts{vcenter_password}) {
            push(@params, {vCenterUsername => $opts{vcenter_username}});
            push(@params, {vCenterPassword => $opts{vcenter_password}});

            $credential_overrides->{vCenterIP} = $opts{vcenter_ip};
            $credential_overrides->{vCenterUsername} = $opts{vcenter_username};
            $credential_overrides->{vCenterPassword} = $opts{vcenter_password};
        }
    }

    push(@params, {vscIp => $opts{vsc_ip}}) if ($opts{vsc_ip});
    push(@params, {vasa10 => 'false'});

    $pkg->call(command => 'registerVasaProviderWithVsc',
               params => \@params,
               credential_overrides => $credential_overrides);

    $Log->exit() if $may_exit;
}

=head2 unregister_with_vcenter

    (Class Call)
    my $ci = NACL::APISet->new(
        hostobj   => $host,
        category  => "Host",
        interface => "SOAP",
        set       => "VPserver",
    );
    my $result = NACL::C::VPserver->unregister_with_vcenter(command_interface => $ci,
                                                            vcenter_ip => $vcenter_ip,
                                                            vcenter_username => $username,
                                                            vcenter_password => $password);

    (Instance Call)
    my $vpserver_command_interface = NACL::C::VPserver->new(vpserver => "myvpserver");
    my $unregister_with_vcenter_result = $vpserver_command_interface->unregister_with_vcenter(vcenter_ip => $vcenter_ip,
                                                                                              vcenter_username => $username,
                                                                                              vcenter_password => $password);

The method unregisters the VPserver with the specified Vcenter server.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method) See L<NACL::APISet::Host::SOAP::VPserver>.

=item C<< vcenter_ip => $vcenter_ip >>

(Required) The ip address or hostname of the vcenter to unregister.

=item C<< vcenter_username => $username >>

(Required) The username for the vcenter to unregister.

=item C<< vcenter_password => $password >>

(Required) The password for the vcenter to unregister.

=item C<< vc_guid => $vc_guid >>

(Optional) The VcGuid for the vcenter to unregister. If only one vcenter is
registered with the VPserver Command interface, this does not need to be specified.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub unregister_with_vcenter {
    $Log->enter() if $may_enter;
    my ($pkg, %opts) = @_;

    $pkg->command_interface($opts{command_interface}) if (!$pkg->command_interface);
    if (!$pkg->command_interface) {
        NACL::APISet::Exceptions::MissingRequiredParamException->throw(text => "Mandatory parameter 'command_interface' missing in call to NACL::C::VPserver::unregister_with_vcenter.");
    }

    if (!$opts{vcenter_ip}) {
        NACL::APISet::Exceptions::MissingRequiredParamException->throw(text => "Mandatory parameter 'vcenter_ip' missing in call to NACL::C::VPserver::unregister_with_vcenter.");
    }

    if (!$opts{vcenter_username}) {
        NACL::APISet::Exceptions::MissingRequiredParamException->throw(text => "Mandatory parameter 'vcenter_username' missing in call to NACL::C::VPserver::unregister_with_vcenter.");
    }

    if (!$opts{vcenter_password}) {
        NACL::APISet::Exceptions::MissingRequiredParamException->throw(text => "Mandatory parameter 'vcenter_password' missing in call to NACL::C::VPserver::unregister_with_vcenter.");
    }

    my @params;
    my $credential_overrides;
    # Perform Validation
    if ($opts{vcenter_ip}) {
        push(@params, {vCenterIP => $opts{vcenter_ip}});
        if ($opts{vcenter_username} and $opts{vcenter_password}) {
            push(@params, {vCenterUsername => $opts{vcenter_username}});
            push(@params, {vCenterPassword => $opts{vcenter_password}});

            $credential_overrides->{vCenterIP} = $opts{vcenter_ip};
            $credential_overrides->{vCenterUsername} = $opts{vcenter_username};
            $credential_overrides->{vCenterPassword} = $opts{vcenter_password};
        }
    }
    push(@params, {vasa10 => 'false'});
    push(@params, {vcIdentifier => {vcGuid => $opts{vc_guid}}});

    $pkg->call(command => 'unregisterVasaProvider',
               params => \@params,
               credential_overrides => $credential_overrides);

    $Log->exit() if $may_exit;
    return $pkg;
}

sub _override_credential {
    my ($pkg, %opts) = @_;
    my $credential_overrides = $opts{credential_overrides};
    my $credential_attrs;

    if (!$opts{credential_overrides}) {
        NACL::APISet::Exceptions::MissingRequiredParamException->throw(text => "Mandatory parameter 'credential_overrides' missing in call to NACL::C::VPserver::_override_credential.");
    }
    if (ref($opts{credential_overrides}) ne 'HASH') {
        NACL::APISet::Exceptions::InvalidParamValueException->(text => "The parameter 'credential_overrides' parameter must be a hash, but it was a "
                                                               . ref($opts{credential_overrides}) . " refernce instead.");
    }

    # copy all existing values
    my @soap_data = ${pop($pkg->credential->get_soap_data_objects())->value}->value;
    foreach my $existing_value (@soap_data) {
        $credential_attrs->{$existing_value->name} = $existing_value->value;
    }

    # Override
    foreach my $override (keys %{$credential_overrides}) {
        $credential_attrs->{$override} = $credential_overrides->{$override};
    }

    my $credential = NATE::SOAP::Data->new(credential => $credential_attrs);
    $pkg->credential($credential);
}

1;
