#
# Copyright (c) 2014 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary NACL::C::VPserver::VCenter Component Module
## @author dl-nacl-dev@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::C::VPserver::VCenter

=head1 DESCRIPTION

C<NACL::C::VPserver::VCenter> is a derived class of
L<NACL::C::Component::VPserver|lib-NACL-C-Component-VPserver-pm>. It represents access
to VPserver VCenter Objects.

=head1 ATTRIBUTES

=head2 command_interface

See L<NACL::C::Component::VPserver|lib-NACL-C-Component-VPserver-pm>

=cut

package NACL::C::VPserver::VCenter;

use strict;
use warnings;

use base 'NACL::C::Component::VPserver';

use NACL::ComponentUtils qw (_optional_scalars Dumper);
use NACL::C::VPserver;

use Params::Validate qw(validate_with SCALAR ARRAYREF SCALARREF);
use NATE::Log qw(log_global);

use NACL::APISet::Exceptions::MissingRequiredParamException qw(:try);

=head1 METHODS

=head2 new

  my $vp = NACL::C::VPserver->new(vpserver => "myvpserver");
  my $vcenter_component = NACL::C::VPserver::VCenter->new(command_interface => $vp);

Return a new VPserver VCenter component object.

=cut

use Class::MethodMaker [
        scalar => [
            {-type => 'NACL::C::VPserver'},
            'command_interface'
        ],
];
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();

sub init {
    $Log->enter() if $may_enter;
    my ($pkg, %opts) = @_;

    if (!$opts{command_interface}) {
        NACL::APISet::Exceptions::MissingRequiredParamException->throw(
            text => "Mandatory parameter 'command_interface' missing in call to NACL::C::VPserver::VCenter::new.");
    }

    $Log->exit() if $may_exit;
    return $pkg;
}

=head2 get_esx_hosts

    my $host_moref = 'HostSystem:host-12345';
    my $protocol = 'fcp';

    my $vp = NACL::C::VPserver->new(vpserver => "myvpserver");
    my $esx_hosts  = NACL::C::VPserver::VCenter->get_esx_hosts(command_interface => $vp,
                                                              contianer_overrides => \%overrides,
                                                              host_moref => $host_moref,
                                                              protocol => $protocol);

A method which returns a list of ESX hosts that are visible to the VCenter registered with the VPserver (by
interacting with the given vpserver command interface).

=over

=item Options

=over

=item C<< command_interface => $vp_ci >>

(Optional) If executed as an instance method.

(Required) If executed as a class method.

See L<NACL::C::VPserver|lib-NACL-C-VPserver-pm>

=item C<< host_moref => $host_moref >>

(Required) The VMware host MoRef (Managed Object Reference) for the host to retrieve.

e.g. "HostSystem:host-12"

=item C<< protocol => $protocol >>

(Required) The protocol that the esx host must match.

e.g. "fcp", "iscsi", "nfs"

=item C<< vc_guid => $vc_guid >>

(Optional) The vcenter GUID to target the API against. By default the vpserver will use the default vcenter it knows of.

=back

=item Returns

This method returns back an array reference of hash references. Each hash reference represents
an ESX host that matched the specified filters.

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub get_esx_hosts {
    $Log->enter() if $may_enter;
    my ($pkg, %opts) = @_;

    # check required params

    $pkg->command_interface($opts{command_interface}) if (!$pkg->command_interface);
    if (!$pkg->command_interface) {
        NACL::APISet::Exceptions::MissingRequiredParamException->throw(
            text => "Mandatory parameter 'command_interface' missing in call to NACL::C::VPserver::VCenter->get_esx_hosts()");
    }

    $opts{host_moref} //
        NACL::APISet::Exceptions::MissingRequiredParamException->throw(
            text => "Mandatory parameter 'host_moref' missing in call to NACL::C::VPserver::VCenter->get_esx_hosts()");
    $opts{protocol} //
        NACL::APISet::Exceptions::MissingRequiredParamException->throw(
            text => "Mandatory parameter 'protocol' missing in call to NACL::C::VPserver::VCenter->get_esx_hosts()");

    my @params;
    push(@params, {moref => $opts{host_moref}}) if ($opts{host_moref});
    push(@params, {protocol => $opts{protocol}});
    push(@params, {vcIdentifier => {vcGuid => $opts{vc_guid} // ''}});

    my $response = $pkg->command_interface->call(command => "vcenterGetEsxHosts",
                                                 params => \@params);
    $Log->exit() if $may_exit;
    return $response;
}

=head2 sync

    my $vp = NACL::C::VPserver->new(vpserver => "myvpserver");
    NACL::C::VPserver::VCenter->sync(command_interface => $vp);

A method that forces the VPserver to sync its databases with the VCenter that it is registered with.

=over

=item Options

=over

=item C<< command_interface => $vp_ci >>

(Optional) If executed as an instance method.

(Required) If executed as a class method.

See L<NACL::C::VPserver|lib-NACL-C-VPserver-pm>

=item C<< vc_guid => $vc_guid >>

(Optional) The vcenter GUID to target the API against. By default the vpserver will use the default vcenter it knows of.

=back

=item Returns

Nothing

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub sync {
    $Log->enter() if $may_enter;
    my ($pkg, %opts) = @_;

    $pkg->command_interface($opts{command_interface}) if (!$pkg->command_interface);
    if (!$pkg->command_interface) {
        NACL::APISet::Exceptions::MissingRequiredParamException->throw(
            text => "Mandatory parameter 'command_interface' missing in call to NACL::C::VPserver::VCenter->sync()");
    }

    my @params;
    push(@params, {vcIdentifier => {vcGuid => $opts{vc_guid} // ''}});

    my $response = $pkg->command_interface->call(command => "vcenterSync",
                                                 params  => \@params);

    $Log->exit() if $may_exit;
    return $response;
}

1;
