# $Id$

#
# Copyright (c) 2014 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary VVolDatastore Component Module
## @author dl-nacl-dev@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::C::VPserver::VVolDatastore

=head1 DESCRIPTION

C<NACL::C::VPserver::VVolDatastore> is a derived class of
L<NACL::C::Component::ONTAP|lib-NACL-C-Component-ONTAP-pm>. It represents access
to VPserver VVolDatastore Objects.

=head1 ATTRIBUTES

=head2 command_interface

See L<NACL::C::VPserver|lib-NACL-C-VPserver-pm>

=cut

package NACL::C::VPserver::VVolDatastore;

use strict;
use warnings;

use base 'NACL::C::Component::VPserver';

use NACL::ComponentUtils qw (_optional_scalars Dumper);
use NACL::C::VPserver;

use Params::Validate qw(validate_with SCALAR ARRAYREF SCALARREF);
use NATE::Log qw(log_global);

use NACL::Exceptions::InvalidParam qw(:try);
use NACL::C::Exceptions::VPserver::VVolDatastore::AlreadyExists;

=head1 METHODS

=head2 new

  my $VVolDatastore = NACL::C::VPserver::VVolDatastore->new(command_interface => $ci);

Return a new VPserver VVolDatastore component object with the given attributes.

=cut

use Class::MethodMaker [
        scalar => [
            {-type => 'NACL::C::VPserver'},
            'command_interface'
        ],
];
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();

sub init {
    $Log->enter() if $may_enter;
    my ($pkg, %opts) = @_;

    if (!$opts{command_interface}) {
        NACL::Exceptions::InvalidParam->throw("Mandatory parameter 'command_interface' missing in call to NACL::C::Vpserver::VVolDatastore::new.");
    }

    $Log->exit() if $may_exit;
    return $pkg;
}

=head2 get_unmounted_hosts

    my $unmounted_hosts  = NACL::C::VPserver::VVolDatastore->get_unmounted_hosts(command_interface => $vp_ci,
                                                                                 override_credentials => \%overrides,
                                                                                 datastore => $datastore);

A method which returns a list of all unmounted hosts that are visible to the VPserver (by
interacting with the given vpserver command interface).

Uses a SOAP VPserver APISet.

=over

=item Options

=over

=item C<< command_interface => $vp_ci >>

(Optional) See L<NACL::C::VPserver|lib-NACL-C-VPserver-pm>

=item C<< datastore => $datastore >>

(Required) The data store for which to return the list of unmounted hosts.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub get_unmounted_hosts {
    $Log->enter() if $may_enter;
    my ($pkg, %opts) = @_;

    $pkg->command_interface($opts{command_interface}) if (!$pkg->command_interface);
    if (!$pkg->command_interface) {
        NACL::Exceptions::InvalidParam->throw("Mandatory parameter 'command_interface' missing in call to NACL::C::Vpserver::VVolDatastore::get_unmounted_hosts.");
    }
    if (!$opts{datastore}) {
        NACL::Exceptions::InvalidParam->throw("Mandatory parameter 'datastore' missing in call to NACL::C::Vpserver::VVolDatastore::get_unmounted_hosts.");
    }

    my $response = $pkg->command_interface->call(command => "getVvolDSUnmountedHosts",
                                                 params => [{datastore => $opts{datastore}}]);
    $Log->exit() if $may_exit;
    return $response;
}

=head2 mount_vvol_datastore_on_host

A method which mounts a VVol datastore to a host using the VPserver (by interacting with the
given vpserver command interface).

    my $cluster_add_result = NACL::C::VPserver::VVolDatastore->mount_vvol_datastore_on_host(command_interface => $vp_ci,
                                                                                            host_moref => $host_moref,
                                                                                            container_id => $container_id,
                                                                                            datastore => $datastore);

(Class method)
A method which mounts a VVol datastore to a host using the VPserver (by interacting with the
given vpserver command interface).

Uses a SOAP VPserver APISet.

=over

=item Options

=over

=item C<< command_interface => $vp_ci >>

(Required) See L<NACL::C::VPserver|lib-NACL-C-VPserver-pm>

=item C<< host_moref => \@host_moref >>

(Required) An arrayref of VMware MoRef (Managed Object Reference) of the host to mount the datastore to.

=item C<< container_id => $container_id >>

(Required) The VMware identifer for the container which holds the data store to mount.

=item C<< datastore => $datastore >>

(Required) The name of the datastore to be mounted.

=item C<< vc_guid => $vc_guid >>

(Optional) The VcGuid for the vcenter to use for mounting the vvol datastore. If only one vcenter is
registered with the VPserver Command interface, this does not need to be specified.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub mount_vvol_datastore_on_host {
    $Log->enter() if $may_enter;
    my ($pkg, %opts) = @_;

    $pkg->command_interface($opts{command_interface}) if (!$pkg->command_interface);
    if (!$pkg->command_interface) {
        NACL::Exceptions::InvalidParam->throw("Mandatory parameter 'command_interface' missing in call to NACL::C::VPserver::VVolDatastore::mount_vvol_datastore_on_host.");
    }
    if (!$opts{datastore}) {
        NACL::Exceptions::InvalidParam->throw("Mandatory parameter 'datastore' missing in call to NACL::C::VPserver::VVolDatastore::mount_vvol_datastore_on_host.");
    }
    if (!$opts{host_moref} and (ref($opts{host_moref} ne 'ARRAY'))) {
        NACL::Exceptions::InvalidParam->throw("Mandatory parameter 'host_moref' missing or was not an ARRAYREF in call to NACL::C::VPserver::VVolDatastore::mount_vvol_datastore_on_host.");
    }
    if (!$opts{container_id}) {
        NACL::Exceptions::InvalidParam->throw("Mandatory parameter 'container_id' missing in call to NACL::C::VPserver::VVolDatastore::mount_vvol_datastore_on_host.");
    }

    $opts{vc_guid} //= "";
    my $response;
    try {
        $response = $pkg->command_interface->call(command => "mountVvolDatastoreOnHost",
                                                  params => [{hostMoref => $opts{host_moref}},
                                                             {vvolDatastoreName => $opts{datastore}},
                                                             {containerId => $opts{container_id}},
                                                             {vcIdentifier => {vcGuid => $opts{vc_guid}}}
                                                            ]);
    } catch NATE::BaseException with {
        my $exception = shift;
        NACL::C::Exceptions::VPserver::VVolDatastore::AlreadyExists->detect_convert_and_throw(exception => $exception);

        $Log->exit() if $may_exit;
        $exception->throw();
    };

    $Log->exit() if $may_exit;
    return $response;
}

=head2 unmount_vvol_datastore_on_all_hosts

A method which unmounts a VVol datastore for all hosts using the VPserver (by interacting with the
given vpserver command interface).

    my $cluster_add_result = NACL::C::VPserver::VVolDatastore->unmount_vvol_datastore_on_all_hosts(command_interface => $vp_ci,
                                                                                                   datastore => $datastore);

Uses a SOAP VPserver APISet.

=over

=item Options

=over

=item C<< command_interface => $vp_ci >>

(Required) See L<NACL::C::VPserver|lib-NACL-C-VPserver-pm>

=item C<< datastore => $datastore >>

(Required) The name of the datastore to unmount from all hosts.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub unmount_vvol_datastore_on_all_hosts {
    $Log->enter() if $may_enter;
    my ($pkg, %opts) = @_;

    $pkg->command_interface($opts{command_interface}) if (!$pkg->command_interface);
    if (!$pkg->command_interface) {
        NACL::Exceptions::InvalidParam->throw("Mandatory parameter 'command_interface' missing in call to NACL::C::VPserver::VVolDatastore::unmount_vvol_datastore_on_all_hosts.");
    }
    if (!$opts{datastore}) {
        NACL::Exceptions::InvalidParam->throw("Mandatory parameter 'datastore' missing in call to NACL::C::VPserver::VVolDatastore::unmount_vvol_datastore_on_all_hosts.");
    }

    my $response = $pkg->command_interface->call(command => "unmountVvolDatastoreOnAllHosts",
                                                 params => [{vvolDatastoreName => $opts{datastore}}]);
    $Log->exit() if $may_exit;
    return $response;
}

=head2 unmount_vvol_datastore_on_host

A method which unmounts a VVol datastore for a specified host using the VPserver (by interacting with the
given vpserver command interface).

    my $cluster_add_result = NACL::C::VPserver::VVolDatastore->unmount_vvol_datastore_on_host(command_interface => $vp_ci,
                                                                                              host => $host_moref,
                                                                                              datastore => $datastore);

Uses a SOAP VPserver APISet.

=over

=item Options

=over

=item C<< command_interface => $vp_ci >>

(Required) See L<NACL::C::VPserver|lib-NACL-C-VPserver-pm>

=item C<< host => $host_moref >>

(Required) The VMware host MoREF (Managed Object Reference).

=item C<< datastore => $datastore >>

(Required) The name of the datastore to unmount from all hosts.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub unmount_vvol_datastore_on_host {
    $Log->enter() if $may_enter;
    my ($pkg, %opts) = @_;

    $pkg->command_interface($opts{command_interface}) if (!$pkg->command_interface);
    if (!$pkg->command_interface) {
        NACL::Exceptions::InvalidParam->throw("Mandatory parameter 'command_interface' missing in call to NACL::C::VPserver::VVolDatastore::unmount_vvol_datastore_on_host.");
    }
    if (!$opts{host}) {
        NACL::Exceptions::InvalidParam->throw("Mandatory parameter 'host' missing in call to NACL::C::VPserver::VVolDatastore::unmount_vvol_datastore_on_host.");
    }
    if (!$opts{datastore}) {
        NACL::Exceptions::InvalidParam->throw("Mandatory parameter 'datastore' missing in call to NACL::C::VPserver::VVolDatastore::unmount_vvol_datastore_on_host.");
    }

    my $response = $pkg->command_interface->call(command => "unmountVvolDatastoreOnHost",
                                                 params => [{hostMoref => $opts{host}},
                                                            {vvolDatastoreName => $opts{datastore}}]);
    $Log->exit() if $may_exit;
    return $response;
}

=head2 convert_datastore_moref_to_container_id

    my $convert_ds_moref_result = NACL::C::VPserver::VVolDatastore->convert_datastore_moref_to_container_id(command_interface => $vp_ci,
                                                                                                            datastore_moref => $datastore_moref);

A method which converts the VVol datastore MoRef (Managed Object Reference) to the corresponding container on which it resides (by interacting with the given vpserver command interface).

Uses a SOAP VPserver APISet.

=over

=item Options

=over

=item C<< command_interface => $vp_ci >>

(Optional) See L<NACL::C::VPserver|lib-NACL-C-VPserver-pm>

=item C<< datastore_moref => $datastore_moref >>

(Required) The VMware MoRef (Managed Object Reference) of the datastore to convert to it's correspoinding container id.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub convert_datastore_moref_to_container_id {
    $Log->enter() if $may_enter;
    my ($pkg, %opts) = @_;

    $pkg->command_interface($opts{command_interface}) if (!$pkg->command_interface);
    if (!$pkg->command_interface) {
        NACL::Exceptions::InvalidParam->throw("Mandatory parameter 'command_interface' missing in call to NACL::C::Vpserver::convert_datastore_moref_to_container_id.");
    }
    if (!$opts{datastore_moref}) {
        NACL::Exceptions::InvalidParam->throw("Mandatory parameter 'datastore_moref' missing in call to NACL::C::Vpserver::VVolDatastore::convert_datastore_moref_to_container_id.");
    }

    my $response = $pkg->command_interface->call(command => "convertVvolDatastoreMorefToContainerId",
                                                 params => [{vvolDatastoreMoref => $opts{datastore_moref}}]);
    $Log->exit() if $may_exit;
    return $response;
}

1;
