# $Id$

#
# Copyright (c) 2012 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary VfilerDr Component Module
## @author dl-nacl-dev@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::C::VfilerDr

=head1 DESCRIPTION

C<NACL::C::VfilerDr> is a derived class of
L<NACL::C::Component::ONTAP|lib-NACL-C-Component-ONTAP-pm>.  It represents access
to an ONTAP Vfiler Disastor Recovery. A related class is
L<NACL::CS::VfilerDr|lib-NACL-CS-VfilerDr-pm>, which represents
the past state of an ONTAP Vfiler disastor recovery.

=cut

package NACL::C::VfilerDr;

use strict;
use warnings;
use NACL::CS::VfilerDr;
use base 'NACL::C::Component::ONTAP';
use NACL::ComponentUtils qw (_optional_scalars Dumper);
use Params::Validate qw(SCALAR ARRAYREF BOOLEAN validate validate_with);

use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();

=head1 METHODS (for component users)


=head2 configure

  NACL::C::VfilerDr->configure(command_interface => $ci,
                                      'remote_login' => "root:netapp1!", ...);

(Class method) This method cofigures a specified remote vfiler
from the specified remote filer for disaster recovery
This can be used as a class method only because it is a setting applicable
for the remote filer hosting the vfiler (not for individual vfilers).

Uses a 7Mode-CLI APISet.

Invokes 'vfiler dr configure' command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< remote-login-name=>$login_name >>

(Mandatory)

Applicable for 7Mode CLI.

=item C<< remote-login-password=>$login_password >>

(Mandatory)

Applicable for 7Mode CLI.

=item C<< interface-details=>[$interface1,$interface2] >>

(Optional) syntax to be specified 'interfacename:ipaddr:netmask'.

Applicable for 7Mode CLI.

=item C<< remote-vfiler=>$remote_vfiler_name >>

(Mandatory)

Applicable for 7Mode CLI.

=item C<< remote-filer=>$remote_filer >>

(Mandatory)

Applicable for 7Mode CLI.

=item C<< dns-server=>[ $dns_address1,$dns_address1 ] >>

(Optional)

Applicable for 7Mode CLI.

=item C<< nis-server=>[ $nis_server1 , $nis_server2] >>

(Optional)

Applicable for 7Mode CLI.

=item C<< sync-snapmirror=>$boolean >>

(Optional)

Applicable for 7Mode CLI.

=item C<< alternate-source=>$alternative_source >>

(Optional)

Applicable for 7Mode CLI.

=item C<< alternate-destination=>$alternative_destination >>

(Optional)

Applicable for 7Mode CLI.

=item C<< skip-snapmirror-init=>$boolean >>

(Optional)

Applicable for 7Mode CLI.

=item C<< secure=>$string >>

(Optional)

Applicable for 7Mode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut


sub configure {
    $Log->enter() if $may_enter;
    my $pkg = shift;

    $Log->debug( "Opts to 'configure' front end are:\n" . Dumper( \@_ ) );
    my %opts = $pkg->_common_validate_with(
        params          => \@_,
        additional_spec => {
            "dns-server"            => { type => ARRAYREF, optional => 1 },
            "nis-server"            => { type => ARRAYREF, optional => 1 },
            "remote-login-name"     => { type => SCALAR },
            "remote-login-password" => { type => SCALAR },
            "interface-details"     => { type => ARRAYREF, optional => 1 },
            "remote-vfiler"         => { type => SCALAR },
            "remote-filer"          => { type => SCALAR },
            _optional_scalars(
                qw(alternate-source alternate-destination skip-snapmirror-init secure
                    sync-snapmirror )
            ),
        },
    );
    my $ret = $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => "_configure_7mode_cli",
                interface => "CLI",
                set       => "7Mode",
            },
        ],
    );

    $Log->exit() if $may_exit;
    return $ret;
} ## end sub configure

sub _configure_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );
    delete $opts{command_interface};
    my $apiset = delete $opts{apiset};

    my $allowed_options = [
        qw(dns-server nis-server remote-login-name remote-login-password interface-details sync-snapmirror
            remote-filer remote-vfiler alternate-destination alternate-source
            skip-snapmirror-init secure)
    ];
    my %cli_args;
    $pkg->_hash_copy(
        source => \%opts,
        copy   => $allowed_options,
        target => \%cli_args,
    );
    if ( exists( $cli_args{'interface-details'} ) ) {
        $cli_args{'interface-details'} = join ",",
            @{ $cli_args{'interface-details'} };
    }
    if ( exists( $cli_args{'dns-server'} ) ) {
        $cli_args{'dns-server'} = join ":", @{ $cli_args{'dns-server'} };
    }

    if ( exists( $cli_args{'nis-server'} ) ) {
        $cli_args{'nis-server'} = join ":", @{ $cli_args{'nis-server'} };
    }

    $apiset->vfiler_dr_configure(%cli_args);
    $Log->exit() if $may_exit;
} ## end sub _configure_7mode_cli

=head2 delete

  NACL::C::VfilerDr->delete(command_interface => $ci,
                                 'remote-vfiler-details' => "vf3@filer");

(Class method) This method delete specified remote vfiler
from the specified remote filer for disaster recovery

Uses a 7Mode CLI APISet.

This can be used as a class method only because it is a setting applicable
for the remote filer hosting the vfiler (not for individual vfilers).

Invokes 'vfiler dr delete' command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< remote-vfiler=>$remote_vfiler_name >>

(Mandatory)

Applicable for 7Mode CLI.

=item C<< remote-filer=>$remote_filer >>

(Mandatory)

Applicable for 7Mode CLI.

=item C<< secure=>$string >>

(Optional)

Applicable for 7Mode CLI.

=item C<< force=>$boolean >>

(Optional)

Applicable for 7Mode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut


sub delete {
    $Log->enter() if $may_enter;
    my $pkg = shift;

    $Log->debug( "Opts to 'delete' front end are:\n" . Dumper( \@_ ) );

    my %opts = $pkg->_common_validate_with(
        params          => \@_,
        additional_spec => {
            "remote-vfiler" => { type => SCALAR },
            "remote-filer"  => { type => SCALAR },
            _optional_scalars(qw(force secure)),
        },
    );

    my $ret = $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => "_delete_7mode_cli",
                interface => "CLI",
                set       => "7Mode",
            },
        ],
    );

    $Log->exit() if $may_exit;
    return $ret;
} ## end sub delete

sub _delete_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );
    delete $opts{command_interface};
    my $apiset = delete $opts{apiset};

    my $allowed_options = [qw(remote-vfiler remote-filer secure force)];
    my %cli_args;
    $pkg->_hash_copy(
        source => \%opts,
        copy   => $allowed_options,
        target => \%cli_args,
    );

    $apiset->vfiler_dr_delete(%cli_args);
    $Log->exit() if $may_exit;
} ## end sub _delete_7mode_cli

=head2 activate

  NACL::C::VfilerDr->activate(command_interface => $ci,
                                 'remote-vfiler-details' => "vf3@filer");

(Class method) This method activates specified remote vfiler
from the specified remote filer for disaster recovery

Uses a 7mode CLI APISet.

This can be used as a class method only because it is a setting applicable
for the remote filer hosting the vfiler (not for individual vfilers).

Invokes 'vfiler dr activate' command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< remote-vfiler-details=>$vfiler_details >>

(Mandatory) syntax to be specified 'remote_vfiler@remote_filer'.

Applicable for 7Mode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut


sub activate {
    $Log->enter() if $may_enter;
    my $pkg = shift;

    $Log->debug( "Opts to 'activate' front end are:\n" . Dumper( \@_ ) );
    my %opts = $pkg->_common_validate_with(
        params          => \@_,
        additional_spec => {
            "remote-vfiler" => { type => SCALAR },
            "remote-filer"  => { type => SCALAR },
        },
    );

    my $ret = $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => "_activate_7mode_cli",
                interface => "CLI",
                set       => "7Mode",
            },
        ],
    );

    $Log->exit() if $may_exit;
    return $ret;
} ## end sub activate

sub _activate_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );
    delete $opts{command_interface};
    my $apiset = delete $opts{apiset};

    my $allowed_options = [qw(remote-vfiler remote-filer)];
    my %cli_args;
    $pkg->_hash_copy(
        source => \%opts,
        copy   => $allowed_options,
        target => \%cli_args,
    );

    $apiset->vfiler_dr_activate(%cli_args);
    $Log->exit() if $may_exit;
} ## end sub _activate_7mode_cli

=head2 resync

  NACL::C::VfilerDr->resync(command_interface => $ci,
                                      'remote_login' => "root:netapp1!", ...);

(Class method) This method resync a specified remote vfiler
from the specified remote filer for disaster recovery
This can be used as a class method only because it is a setting applicable
for the remote filer hosting the vfiler (not for individual vfilers).

Uses a 7Mode CLI APISet.

Invokes 'vfiler dr resync' command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< remote-login-name=>$login_name >>

(Mandatory)

Applicable for 7Mode CLI.

=item C<< remote-login-password=>$login_password >>

(Mandatory)

Applicable for 7Mode CLI.

=item C<< remote-vfiler=>$remote_vfiler_name >>

(Mandatory)

Applicable for 7Mode CLI.

=item C<< remote-filer=>$remote_filer >>

(Mandatory)

Applicable for 7Mode CLI.

=item C<< sync-snapmirror=>$boolean >>

(Optional)

Applicable for 7Mode CLI.

=item C<< alternate-host-details=>$host_details >>

(Optional) syntax to be specified 'alt-src,alt-dst'.

Applicable for 7Mode CLI.

=item C<< secure=>$string >>

(Optional)

Applicable for 7Mode CLI.
=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut


sub resync {
    $Log->enter() if $may_enter;
    my $pkg = shift;

    $Log->debug( "Opts to 'resync' front end are:\n" . Dumper( \@_ ) );

    my %opts = $pkg->_common_validate_with(
        params          => \@_,
        additional_spec => {
            "remote-vfiler"         => { type => SCALAR },
            "remote-filer"          => { type => SCALAR },
            "remote-login-name"     => { type => SCALAR },
            "remote-login-password" => { type => SCALAR },
            _optional_scalars(
                qw(alternate-source alternate-destination secure sync-snapmirror)
            ),
        },
    );

    my $ret = $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => "_resync_7mode_cli",
                interface => "CLI",
                set       => "7Mode",
            },
        ],
    );

    $Log->exit() if $may_exit;
    return $ret;
} ## end sub resync

sub _resync_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );
    delete $opts{command_interface};
    my $apiset = delete $opts{apiset};

    my $allowed_options = [
        qw(remote-login-name remote-login-password remote-filer alternate-destination
            alternate-source sync-snapmirror secure remote-vfiler)
    ];
    my %cli_args;
    $pkg->_hash_copy(
        source => \%opts,
        copy   => $allowed_options,
        target => \%cli_args,
    );

    $apiset->vfiler_dr_resync(%cli_args);
    $Log->exit() if $may_exit;
} ## end sub _resync_7mode_cli

=head2 status

  NACL::C::VfilerDr->status(command_interface => $ci,
                 'remote-vfiler-details' => "vf@filer", ...);

(Class method) This method status returns NACL::CS::VfilerDr state objects of
a specified remote vfiler from the specified remote filer for disaster recovery

Uses a 7Mode CLI APISet.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< remote-vfiler-details=>$vfiler_details >>

(Mandatory) syntax to be specified 'remote_vfiler@remote_filer'.

Applicable for 7Mode CLI.

=item C<< filter=>{status => "mirrored"} >>

(Optional) HASHREF.

Applicable for 7Mode CLI.

=item C<< requested_fileds=>["status"] >>

(Optional) ARRAYREF.

Applicable for 7Mode CLI.

=back

=back

=cut

sub status {
    $Log->enter() if $may_enter;
    my $pkg    = shift;
    my @states = NACL::CS::VfilerDr->fetch(@_);
    $Log->exit() if $may_exit;
    return wantarray ? @states : $states[0];
} ## end sub status


__PACKAGE__->_load_event_class();

1;
