# Copyright (c) 2001-2012 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Volume64bitUpgrade Component Module
## @author dl-nacl-dev@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::C::Volume64bitUpgrade

=head1 DESCRIPTION

C<NACL::C::Volume64bitUpgrade> is a derived class of
L<NACL::C::Component::ONTAP|lib-NACL-C-Component-ONTAP-pm>.  It represents the operation
of upgrading a 32-bit volume to 64-bits.  A related class is
L<NACL::CS::Volume64bitUpgrade|lib-NACL-CS-Volume64bitUpgrade-pm>, which represents
the operation of upgrading a 32-bit volume to 64-bits.

=head1 ATTRIBUTES

=head2 volume

=head2 vserver

=cut

package NACL::C::Volume64bitUpgrade;

use strict;
use warnings;

use NACL::ComponentUtils qw (_optional_scalars Dumper);

use base 'NACL::C::Component::ONTAP';
use NACL::CS::Volume64bitUpgrade;
use NACL::CS::Volume64bitUpgradeStartCheck;

use Params::Validate qw(validate validate_with SCALAR ARRAYREF UNDEF);
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();

use Class::MethodMaker [
    scalar => 'volume',
    scalar => 'vserver',
];

=head2 start

    my $Volume64bitUpgrade_object = NACL::C::Volume64bitUpgrade->start(
                                    command_interface => $ci,
                                    volume            => $volume,
                                    vserver           => $vserver,
                                    ...);

(Class method)
This method is used to start 64-bit upgrade on a volume.

Uses CMode CLI, 7Mode CLI APISets.

Invokes 'vol 64bit upgrade start' command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< volume => $string >>

(Required) Name of volume.

Applicable for CMode CLI, 7Mode CLI.

Maps to:

7M CLI : volume

=item C<< mode => $string >>

(Optional) The mode for upgrading the volume to 64bit .

Applicable for CMode CLI, 7Mode CLI.

Maps to:

7M CLI : -mode <value> (Programmatic name: mode)

=item C<< vserver => $string >>

Required for CMode not applicable for 7Mode

Applicable for CMode CLI.

=item C<< 'extended_query' => $string >>

(Optional)

Applicable for CMode CLI.

=item C<< '*' >>

This method takes all other options supported by
"volume 64bit-upgrade start" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub start {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw(mode
                    volume vserver
                    )
            )
        },
        api => 'volume_64bit_upgrade_start',
    );
    $Log->exit() if $may_exit;

    return $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => "_start_cmode_cli",
                interface => "CLI",
                set       => "CMode",
            },
            {   method    => "_start_7mode_cli",
                interface => "CLI",
                set       => "7Mode",
            },
        ],
        frontend => 'NACL::C::Volume64bitUpgrade::start',
    );
} ## end sub start

sub _start_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, %opts) = @_;

    $pkg->_base_backend_cmode_cli(
        %opts,
        api => 'volume_64bit_upgrade_start',
    );

    $Log->exit() if $may_exit;
    return $pkg->new(
        command_interface => $opts{command_interface},
        volume            => $opts{volume},
        vserver           => $opts{vserver},
    );
} ## end sub _start_cmode_cli

sub _start_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );

    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};

    my $allowed_7mode_options = [qw(volume mode)];
    my %args;
    $pkg->_hash_copy(
        source => \%opts,
        copy   => $allowed_7mode_options,
        target => \%args,
    );
    %opts = %args;

    $apiset->vol_64bit_upgrade_start(%opts);
    $Log->exit() if $may_exit;

    return $pkg->new(
        command_interface => $command_interface,
        volume            => $opts{volume},
    );
} ## end sub _start_7mode_cli

=head2 start_check

    my $Volume64bitUpgradeStartCheck_object = NACL::C::Volume64bitUpgrade->start_check(
                                    command_interface => $ci,
                                    volume            => $volume,
                                    vserver           => $vserver,
                                    ...);

(Class method)
This method is used to query the potential 64-bit upgrade of a volume and
returns an object of type L<NACL::CS::Volume64bitUpgradeStartCheck|lib-NACL-CS-Volume64bitUpgradeStartCheck-pm>

Uses CMode CLI, 7Mode CLI APISet

Invokes 'vol 64bit upgrade start' command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< volume => $string >>

(Required) Name of volume.

Applicable for CMode CLI.

=item C<< vserver => $string >>

Required for CMode, not applicable for 7Mode

Applicable for CMode CLI.

=item C<< mode => $string >>

Applicable for CMode CLI.

=item C<< 'extended_query' => $string >>

(Optional)

Applicable for CMode CLI.

=item C<< '*' >>

This method takes all other options supported by
"volume 64bit-upgrade start" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub start_check {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params => \@args,
        api    => 'volume_64bit_upgrade_start',
    );
    $Log->exit() if $may_exit;

    return $pkg->call_on_apiset(
        %opts,
        'mode'  => "check",
        choices => [
            {   method    => "_start_check_cmode_cli",
                interface => "CLI",
                set       => "CMode",
            },
            {   method    => "_start_check_7mode_cli",
                interface => "CLI",
                set       => "7Mode",
            },
        ],
        frontend => 'NACL::C::Volume64bitUpgrade::start',
    );
} ## end sub start_check

sub _start_check_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    my $response;

    $pkg->_base_backend_cmode_cli(
        @args,
        api => 'volume_64bit_upgrade_start',
        apiset_response_obj => \$response,
    );

    my $output = $response->get_parsed_output();
    my $only_row = $output->[0];
    my $obj      = NACL::CS::Volume64bitUpgradeStartCheck->new();
    $obj->_set_fields(row => $only_row);
    $Log->exit() if $may_exit;
    return $obj;
} ## end sub _start_check_cmode_cli

sub _start_check_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );

    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};

    my $response = $apiset->vol_64bit_upgrade_start(
        mode   => $opts{mode},
        volume => $opts{volume}
    );
    my $output   = $response->get_parsed_output();
    my $only_row = $output->[0];
    my $obj      = NACL::CS::Volume64bitUpgradeStartCheck->new();
    $obj->_set_fields(row => $only_row);
    $Log->exit() if $may_exit;
    return $obj;
} ## end sub _start_check_7mode_cli

=head2 start_normal

    my $Volume64bitUpgrade_object = NACL::C::Volume64bitUpgrade->start_normal(
                                    command_interface => $ci,
                                    volume            => $volume,
                                    vserver           => $vserver,
                                    ...);

(Class method)
This method is used to start 64-bit upgrade on a volume in normal mode.
This method has been added for backward compatibility,so that it doesnt
break any existing scripts
Ideally users should be calling start method by passing mode => 'normal'

=cut

sub start_normal {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(params => \@args,);
    $Log->exit() if $may_exit;
    return $pkg->start(%opts, 'mode' => 'normal');
} ## end sub start_normal

=head2 status

    my $status_obj = NACL::C::Volume64bitUpgrade->status(command_interface => $ci,
                          [ filter => { %filter } ,]
              [ requested_fields => [ @req_fields ]);

                    or

    my @status_objs = NACL::C::Volume64bitUpgrade->status(command_interface => $ci,
                          [ filter => { %filter } ,]
              [ requested_fields => [ @req_fields ]);

                    or

   my $status_obj = $Volume64bitUpgrade_obj->status();

(Class or Instance method)
This method returns objects of the type L<NACL::CS::Volume64bitUpgradeStatus|lib-NACL-CS-Volume64bitUpgradeStatus-pm>.

This method accepts all the parameters accepted by L<fetch|lib-NACL-CS-ComponentState-pm/fetch>
(requested_fields, filter, allow_empty) and has all of the other properties of
fetch methods (callable in scalar or list context, throws a NoElementsFound
exception if no matching elements are found)

=cut

sub status {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = $pkg_or_obj->_handle_obj_cs_like_calls(@_);
    require NACL::CS::Volume64bitUpgradeStatus;
    my @state_objs = NACL::CS::Volume64bitUpgradeStatus->fetch(%opts);
    $Log->exit() if $may_exit;
    return wantarray ? @state_objs : $state_objs[0];
} ## end sub status


__PACKAGE__->_load_event_class();

sub _primary_keys_validate_spec {
    return (
        volume  => {type => SCALAR},
        vserver => {type => SCALAR | UNDEF, optional => 1},
    );
} ## end sub _primary_keys_validate_spec

1;
