# $Id$

#
# Copyright (c) 2001-2012 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary VolumeClone Component Module
## @author Dheeraj.K@netapp.com, dl-nacl-dev@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::C::VolumeClone

=head1 DESCRIPTION

C<NACL::C::VolumeClone> is a derived class of
L<NACL::C::Component::ONTAP|lib-NACL-C-Component-ONTAP-pm>.  It represents access
to an ONTAP VolumeClone.  A related class is
L<NACL::CS::VolumeClone|lib-NACL-CS-VolumeClone-pm>, which represents
the past state of an ONTAP VolumeClone.

=head1 ATTRIBUTES

=head2 command_interface

See L<NACL::C::Component::command_interface|lib-NACL-C-Component-pm/command_interface>

=head2 flexclone

The name of the flexcone, used to identify it uniquely.

=head3 vserver

The name of the vserver.

=cut

package NACL::C::VolumeClone;

use strict;
use warnings;
use NACL::CS::VolumeClone;
use NACL::ComponentUtils qw (_optional_scalars);
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use NACL::ComponentUtils qw(Dumper);
use base 'NACL::C::Component::ONTAP';
use Params::Validate qw(validate_with SCALAR SCALARREF UNDEF);
use NACL::APISet::Exceptions::MethodNotFoundException qw(:try);
use NACL::C::Job;
use NACL::Exceptions::InvalidChoice qw(:try);
use NACL::STask::Volume;
use NACL::C::Exceptions::VolumeClone::FeatureNotLicensed;
use NACL::C::Exceptions::VolumeClone::AlreadyExists;
use NATE::BaseException qw(:try);

=head1 Linked components

(See
http://wikid.netapp.com/w/QA/projects/Libraries_Initiative/Project/Common_Infrastructure/Users_Guide/Component_Layer_Users_Guide#Linked_components
for a more detailed explanation on linked components)

This component is linked to the following components:

=begin html

<table border="1">
<tr> <th>Prefix</th> <th>Linked Component</th> </tr>
<tr> <td>split</td> <td>VolumeCloneSplit</td> </tr>
</table>

=end html

Examples:

    # Invoking the "estimate" method of the VolumeCloneSplit component
    my $estimate = $vol_clone_obj->split_estimate(%opts);

    # Invoking the "stop" method of the VolumeCloneSplit component
    $vol_clone_obj->split_stop(%opts);

=cut

sub can {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my @opts       = $pkg_or_obj->_can_validation(@_);

    my $code;
    my $method    = $opts[0];
    my $error_ptr = $opts[1];
    if ($code = $pkg_or_obj->NACL::C::Component::can(@opts)) {
        $Log->exit() if $may_exit;
        return $code;
    } else {

        # See if it can be found in the linked components
        $code = $pkg_or_obj->_handle_component_aggregation(
            method    => $method,
            error_ptr => $error_ptr,
            table     => [
                {   component => 'NACL::C::VolumeCloneSplit',
                    prefix    => 'split',
                },
            ]
        );
        if ($code) {
            $Log->exit() if $may_exit;
            return $code;
        } else {

            # See if it can be found in Volume
            my $c_or_t = $pkg_or_obj->is_c_or_t();
            my $vol_c_or_t_pkg
                = $pkg_or_obj->_get_package_at_same_layer(file => 'Volume');

            # If the method is "volume_blah", look for method "blah" in Volume
            if ($method =~ /^volume_/) {
                $method =~ s/volume_(.*)/$1/;
            }
            if ($vol_c_or_t_pkg->can($method)) {
                my $sub = sub {
                    $Log->enter() if $may_enter;
                    my ($pkg_or_obj, %args) = @_;
                    my %opts = $pkg_or_obj->_common_validate_with(
                        params      => [%args],
                        allow_extra => 1
                    );

                    $pkg_or_obj->_hash_move(
                        source => \%opts,
                        target => \%opts,
                        map    => {flexclone => 'volume'}
                    );

                    # Do NOT use the $vol_c_or_t_pkg defined outside.
                    # The closure will get tied to that Volume package,
                    # regardless of the invoking VolumeClone package
                    my $vol_c_or_t_pkg_in_sub
                        = $pkg_or_obj->_get_package_at_same_layer(
                        file => 'Volume');

                    my $ret = $vol_c_or_t_pkg_in_sub->$method(%opts);
                    $Log->exit() if $may_exit;
                    return $ret;
                };
                $Log->exit() if $may_exit;
                return $sub;
            } else {
                $error_ptr .= "Could not find '$method' in package "
                    . "'$vol_c_or_t_pkg\n";
                if ($code = $pkg_or_obj->SUPER::can(@_)) {
                    $Log->exit() if $may_exit;
                    return $code;
                } else {
                    $Log->exit() if $may_exit;
                    return undef;
                }
            }
        }
    }
} ## end sub can

=head2 clone_create
    
    my $clone_c_obj     = $clone_c_obj->clone_create();
    
    my $clone_stask_obj = $clone_stask_obj->clone_create();
    
Creates a clone of the volume on which the method was invoked. The
C<command_interface>, C<vserver> and C<parent-volume> arguments do not
need to be passed since these are extracted from the flexclone object.

If invoked with a component object, then the arguments that can be passed
to this method are those accepted by
L<the component|lib-NACL-C-VolumeClone-pm/create> and the object returned is
of type C<NACL::C::VolumeClone>.

If invoked with a component object, then the arguments that can be passed
to this method are those accepted by
L<the task|lib-NACL-STask-VolumeClone-pm/create> and the object returned is
of type C<NACL::STask::VolumeClone>.

=cut

# Tested in the Volume unit test
# This needs to be explicitly declared and cannot be handled by AUTOLOAD
# because then flexclone could be provided as an options and will be an
# attribute in the object, so the mismatch in value will cause
# _common_validate_with to complain.
sub clone_create {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %orig_opts) = @_;

    # Some juggling of options needed. Assume a call like:
    #    $flexclone_obj->clone_create(flexclone => $val);
    # At this point, delete "flexclone" from %opts, so it does not
    # clash with the value in the object.
    my $flexclone = delete $orig_opts{flexclone};

    # Now invoke _common_validate_with. The value of the attribute in
    # the object now gets stored in $opts{flexclone}
    my %opts = $pkg_or_obj->_common_validate_with(
        params      => [%orig_opts],
        allow_extra => 1
    );

    # For our purposes, this is actually the "parent-volume", so move this
    # into $opts{'parent-volume'}
    $opts{'parent-volume'} = delete $opts{flexclone};

    # And if a value of flexclone was passed as argument, add it back
    $opts{flexclone} = $flexclone if (defined $flexclone);

    my $clone_pkg = 'NACL::' . $pkg_or_obj->is_c_or_t() . '::VolumeClone';
    my $clone_obj = $clone_pkg->create(%opts);
    $Log->exit() if $may_exit;
    return $clone_obj;
}

=head1 METHODS

=head2 new

  my $Volclone = NACL::C::VolumeClone->new(command_interface=>$ci,
                                           flexclone=>$flexclone,
                                           vserver =>$vserver);

Return a new VolumeClone component object with the given attributes.

=cut

use Class::MethodMaker [
    scalar => 'flexclone',
    scalar => 'vserver',
];

sub init {
    $Log->enter() if $may_enter;
    my ($self, @args) = @_;
    $self->SUPER::init(@args);
    $self->_vserver_context_init(@args);
    $Log->exit() if $may_exit;
} ## end sub init

=head2 create

  my $volumeclone  = NACL::C::VolumeClone->create(command_interface=>$ci,
                                                  flexclone=>$flexclone,
                                                  vserver =>$vserver...);

(Class method) Request creation of a new volumeclone element, and  return a new
volumeclone component object that refers to it.  Does not wait for volumeclone
creation to complete.

Uses a CMode CLI/ZAPI, 7Mode CLI APISet.

Invokes "volume-clone-create" API for CMode ZAPI.

Invokes "vol clone create" command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< flexclone=>$flexclone_name >>

(Required) Name of flexclone to create.

Applicable for CMode CLI/ZAPI, 7Mode CLI.

Maps to:

CM ZAPI: volume

7M CLI : clone volume

=item C<< vserver=>$vserver >>

(Required) Name of vserver.

Applicable for CMode CLI.

=item C<< "parent-volume" => $string >>

Applicable for CMode CLI/ZAPI, 7Mode CLI.

Maps to:

CM ZAPI: parent-volume

7M CLI : -b <value> (Programmatic name: parent flex volume)

=item C<< "parent-snapshot" => $string >>

Applicable for CMode CLI/ZAPI, 7Mode CLI.

Maps to:

CM ZAPI: parent-snapshot

7M CLI : parent snapshot

=item C<< "junction-path" => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: junction-path

=item C<< "junction-active" => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: junction-active

=item C<< "space-guarantee" => $string >>

Applicable for CMode CLI/ZAPI, 7Mode CLI.

Maps to:

CM ZAPI: space-reserve

7M CLI : -s <value> (Programmatic name: space reserve)

=item C<< "comment" => $string >>

Applicable for CMode CLI.

=item C<< "foreground" => $string >>

Applicable for CMode CLI.

=item C<< "method-timeout"=> $scalar >>

(Optional)The new timeout value.

=item C<< "job_component"=> \$scalar >>

=item C<< 'use-snaprestore-license' => $string >>

Applicable for CMode CLI.

=item C<< 'type' => $string >>

Applicable for CMode CLI/ZAPI.

=item C<< 'qos-policy-group' => $string >>

Applicable for CMode CLI/ZAPI.

=item C<< 'external-cache' => $string >>

External Cache Policy Name

Applicable for CMode CLI

=item C<< "force" => $string >>

Applicable for 7Mode CLI.

Maps to:

7M CLI : force

=item C<< 'extended_query' => $string >>

(Optional)

Applicable for CMode CLI.

=item C<< 'caching-policy' => $string >>

Caching Policy Name
Applicable for CMode CLI/ZAPI

=item C<< '*' >>

This method takes all other options supported by
"volume clone create" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=item C<NACL::C::Exceptions::VolumeClone::FeatureNotLicensed>

This type of exception gets thrown if an attempt is made to create a
volume clone when flexclone is not licensed.

=item C<NACL::C::Exceptions::VolumeClone::AlreadyExists>

This exception is thrown when we attempt to create a volume clone which already cloned.

=back

=back

=cut

sub create {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            job_component => {type => SCALARREF, optional => 1},
            _optional_scalars(
                qw(parent-volume parent-snapshot junction-path
                    junction-active space-guarantee comment foreground
                    force
                    use-snaprestore-license

                    type

                    qos-policy-group

                    external-cache

                    caching-policy
                    )
            )
        },
        cli_cmd => 'volume clone create',
    );
    my $ret;
    try {
        $ret = $pkg->call_on_apiset(
            %opts,
            choices => [
                {   method    => "_create_cmode_cli",
                    interface => "CLI",
                    set       => "CMode",
                },
                {   method    => "_create_cmode_zapi",
                    interface => "ZAPI",
                    set       => "CMode",
                },
                {   method    => "_create_7mode_cli",
                    interface => "CLI",
                    set       => "7Mode",
                }
            ],
            frontend => 'NACL::C::VolumeClone::create',
        );
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = $_[0];
        if (NACL::C::Exceptions::VolumeClone::FeatureNotLicensed->detect(
                exception => $exception
            )
            )
        {
            $Log->exit() if $may_exit;
            NACL::C::Exceptions::VolumeClone::FeatureNotLicensed
                ->convert_and_throw(
                exception         => $exception,
                feature           => 'flexclone',
                command_interface => $opts{command_interface}
                );
        }
        NACL::C::Exceptions::VolumeClone::AlreadyExists
            ->detect_convert_and_throw(exception => $exception);
        $Log->exit() if $may_exit;
        $exception->throw();
    };

    $Log->exit() if $may_exit;
    return $ret;
} ## end sub create

sub _create_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, %opts) = @_;
    my $vserver = $opts{vserver};

    $pkg->_base_backend_cmode_cli (
        %opts,
        api => 'volume_clone_create',
        vserver_operated_on => \$vserver,
    );

    $Log->exit() if $may_exit;
    return $pkg->new(
        flexclone         => $opts{flexclone},
        vserver           => $opts{vserver},
        command_interface => $opts{command_interface},
    );
} ## end sub _create_cmode_cli

sub _create_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    my %opts = $pkg->_validate_backend_opts(@args);
    
    # ZAPI accepts this value in all lower-case.
    if (defined $opts{type}) {
        $opts{type} = lc $opts{type};
    }
    
    delete $opts{'parent-vserver'} if ( defined $opts{'parent-vserver'} );
    delete $opts{'force'} if ( defined $opts{'force'} );
    my $vserver;
    $pkg->_base_backend_cmode_zapi(
        %opts,
        api  => 'volume-clone-create',
        copy => [qw(parent-volume parent-snapshot junction-path
            junction-active caching-policy)
        ],
        map  => {
            'space-guarantee'  => 'space-reserve',
            'flexclone'        => 'volume',
            'qos-policy-group' => 'qos-policy-group-name',
            type               => 'volume-type',
        },
        vserver_operated_on => \$vserver,
    );

    my $obj = $pkg->new(
        vserver           => $vserver,
        flexclone         => $opts{flexclone},
        command_interface => $opts{command_interface}
    );

    $Log->exit() if $may_exit;
    return $obj;
} ## end sub _create_cmode_zapi

sub _create_7mode_cli {
    my $pkg  = shift;
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};

    my $allowed_7mode_options = [qw(force parent-snapshot)];
    my %clone_args;
    $pkg->_hash_copy(
        source => \%opts,
        copy   => $allowed_7mode_options,
        map    => {
            flexclone         => "clone-volume",
            "space-guarantee" => "space-reserve",
            "parent-volume"   => "parent-flex-volume"
        },
        target => \%clone_args,
    );
    my %args     = %clone_args;
    my $response = $apiset->vol_clone_create(%args);

    return $pkg->new(
        flexclone         => $opts{flexclone},
        command_interface => $command_interface
    );
} ## end sub _create_7mode_cli

=head2 split

  NACL::C::VolumeClone->split(command_interface=>$ci,
                               flexclone=>$flexclone,
                               vserver =>$vserver...);

                            or

  $volclone_object->split(...);

(Class or instance method) This method splits a volumeclone element.

(Uses a CMode CLI APISet for BR, for RR releases onwards see
L<NACL::C::VolumeCloneSplit|lib-NACL-C-VolumeCloneSplit-pm/start>
for the modes/interfaces suppported)

For BR releases, this runs "volume clone split". For RR releases onwards
this automatically invokes C<NACL::C::VolumeCloneSplit->start()> (since from
RR "split" got changed from a command into a directory of commands)

The options listed below are for the BR release. For documentation
of this method for RR onwards see
L<NACL::C::VolumeCloneSplit|lib-NACL-C-VolumeCloneSplit-pm/start>.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< flexclone=>$flexclone_name >>

(Required for class method, Not Applicable for instance method)
Name of flexclone.

Applicable for CMode CLI.

=item C<< vserver=>$vserver_name >>

(Required for class method, Not Applicable for instance method)
Name of vserver.

Applicable for CMode CLI.

=item C<< "override-space-check" => $string >>

Applicable for CMode CLI.

=item C<< foreground => $string >>

Applicable for CMode CLI.

=item C<< "method-timeout"=> $scalar >>

(Optional)The new timeout value.

Applicable for CMode CLI.

=item C<< "job_component"=> \$scalar >>

=item C<< 'extended_query' => $string >>

(Optional)

Applicable for CMode CLI.

=item C<< '*' >>

This method takes all other options supported by
"volume clone split" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub split {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts = $pkg_or_obj->_common_validate_with(
        params      => \@args,
        allow_extra => 1,
    );
    require NACL::C::VolumeCloneSplit;
    return NACL::C::VolumeCloneSplit->start(%opts);
    $Log->exit() if $may_exit;
} ## end sub split

=head2 modify

    NACL::C::VolumeClone->modify(
                                command_interface => $ci,
                                'flexclone' => $flexclone,
                                'vserver' => $vserver,
                                ...);

                                 or

    $VolumeClone_obj->modify(...);

(Class or instance method)

Uses a CMode CLI APISet.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'flexclone' => $flexclone >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI.

=item C<< 'vserver' => $vserver >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI.

=item C<< 'qos-policy-group' => $string >>

Applicable for CMode CLI.

=item C<< 'method-timeout' => $time_in_seconds >>

(Optional)
The default NATE timeout is 60 seconds. If the command might take more
time to complete, this option can be used to specify a larger timeout value.
The value should be provided in seconds.

=item C<< 'external-cache' => $string >>

External Cache Policy Name

Applicable for CMode CLI

=item C<< 'extended_query' => $string >>

(Optional)

Applicable for CMode CLI.

=item C<< '*' >>

This method takes all other options supported by
"volume clone modify" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub modify {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( qos-policy-group

                    external-cache
                    )
            ),
        },
        api => 'volume_clone_modify',
    );
    $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_modify_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::VolumeClone::modify',
    );

    $Log->exit() if $may_exit;
} ## end sub modify

sub _modify_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    my $vserver = $opts{vserver};

    $pkg_or_obj->_base_backend_cmode_cli(
        %opts,
        api => 'volume_clone_modify',
        vserver_operated_on => \$vserver,
    );

    $Log->exit() if $may_exit;
} ## end sub _modify_cmode_cli


__PACKAGE__->_load_event_class();

sub _primary_keys_validate_spec {
    return (
        flexclone => {type => SCALAR},
        vserver   => {type => SCALAR | UNDEF, optional => 1},
    );
} ## end sub _primary_keys_validate_spec

=head2 Other Volume methods

All functionality available for volumes are applicable for volume clones
as well. This module allows for direct reuse of the methods defined under
C<NACL::C::Volume>, by providing "volume_" as prefix. Taking the example
of C<delete>, it is possible to do:

    NACL::C::VolumeClone->volume_delete(
        command_interface => $ci,
        flexclone         => $flexclone_name,
        vserver           => $vserver_name,
    );

                    or

    $vol_clone_obj->volume_delete();

See L<NACL::C::Volume|lib-NACL-C-Volume-pm> for a description of the methods
available.

(Note that the key is named "volume" in NACL::C::Volume, but is "flexclone"
here. However, this module takes care of the translation of this field, so
calls with object instances work.

Also note that it is possible to invoke the methods under C<NACL::C::Volume>
directly, like:

    $vol_clone_obj->delete();

This is not suggested, since if at some point commands of the same name
are added under "volume clone" but with different functionality, then
the scripts using it will break. For this reason, it is suggested that
the "volume_" prefix is used while attempting to access functionality
provided by C<NACL::C::Volume>.
)

=head2 volume

    # Get the value
    my $flexclone_name = $clone_obj->volume();

    # Set the value
    $clone_obj->volume('b');

Get or set the name of the flexclone. Functionally equivalent to the
C<flexclone> accessor method. This is provided since volume clones are
also volumes.

=cut

# This is necessary for cases where we provide a volume clone object as the
# value for "flexclone", and that method passes through to the Volume
# component.
sub volume {
    $Log->enter() if $may_enter;
    $Log->exit() if $may_exit;
    goto &flexclone;
}

=head2 cast_to_volume

    my $vol_obj = $vol_clone_obj->cast_to_volume();

As described above, volume clones are also volumes, so operations which
can be performed on volumes can be performed on volume clones as well.
Certain Component and Task methods accept Volume objects but not
VolumeClone objects. This method can be used to cast the VolumeClone
object into a Volume object - which can then be passed into the
method. An example:

    my $vol_obj = $vol_clone_obj->cast_to_volume();
    NACL::STask::Client::Mount->create(
        volume => $vol_obj,
        %other_opts
    );

Note that if invoked on a NACL::C::VolumeClone object, a NACL::C::Volume
object is returned, while if is invoked on a NACL::STask::VolumeClone
object, a NACL::STask::Volume object is returned.

=cut

sub cast_to_volume {
    $Log->enter() if $may_enter;
    my $self    = $_[0];
    my $c_or_t  = $self->is_c_or_t();
    my $vol_pkg = "NACL::${c_or_t}::Volume";
    my %pk_hash = $self->get_pk_hash_with_ci();
    $pk_hash{volume} = delete $pk_hash{flexclone};
    my $vol_obj = $vol_pkg->new(%pk_hash);
    $Log->exit() if $may_exit;
    return $vol_obj;
} ## end sub cast_to_volume

1;
