# Copyright (c) 2001-2012 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary VolumeCloneSplit Component Module
## @author dl-nacl-dev@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::C::VolumeCloneSplit

=head1 DESCRIPTION

C<NACL::C::VolumeCloneSplit> is a derived class of
L<NACL::C::Component::ONTAP|lib-NACL-C-Component-ONTAP-pm>. It represents access
to an ONTAP VolumeCloneSplit.  A related class is
L<NACL::CS::VolumeCloneSplit|lib-NACL-CS-VolumeCloneSplit-pm>, which represents
the past state of an ONTAP VolumeCloneSplit.

=head1 ATTRIBUTES

=head2 command_interface

See L<NACL::C::Component::command_interface|lib-NACL-C-Component-pm/command_interface>

=head2 flexclone

=head2 vserver

=cut

package NACL::C::VolumeCloneSplit;

use strict;
use warnings;

use NACL::CS::VolumeCloneSplit;
use NACL::ComponentUtils qw (_optional_scalars Dumper);

use base 'NACL::C::Component::ONTAP';
use NACL::C::Job;
use Params::Validate
    qw(validate validate_with SCALAR ARRAYREF SCALARREF UNDEF);
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use NATE::BaseException qw(:try);
use NATE::Exceptions::Argument qw(:try);
use NACL::C::Exceptions::VolumeCloneSplit::SplitInProgress;

=head1 METHODS

=head2 new

  my $VolumeCloneSplit = NACL::C::VolumeCloneSplit->new(command_interface=>$ci,
                                   flexclone =>$flexclone,
                                   vserver =>$vserver,
                                  );

Return a new VolumeCloneSplit component object with the given attributes.

=cut

use Class::MethodMaker [
    scalar => 'flexclone',
    scalar => 'vserver',
];

sub init {
    $Log->enter() if $may_enter;
    my ($self, @args) = @_;
    $self->SUPER::init(@args);
    $self->_vserver_context_init(@args);
    $Log->exit() if $may_exit;
} ## end sub init

=head2 find

    my $VolumeCloneSplit  = NACL::C::VolumeCloneSplit->find(command_interface => $ci, ...);
    my @VolumeCloneSplits = NACL::C::VolumeCloneSplit->find(command_interface => $ci, ...); 

(Class method)
A constructor of sorts, which discovers which elements are present (by
interacting with the given command interface).  In an array context it
creates a component object for each present element and returns them.
In a scalar context it creates a component object for just one
element.

This method is generally not implemented in individual components. The base
class (L<NACL::C::Component::ONTAP|lib-NACL-C-Component-ONTAP-pm> provides an implementation.
See <L<NACL::C::Component::find|lib-NACL-C-Component-pm/find> for a more
detailed description along with explanation of all the options accepted.

=head2 state

    my VolumeCloneSplit_state = $NACL::C::VolumeCloneSplit->state();

(Instance method)
Fetch the current state of the element associated
with this component, and return a Component State object corresponding
to this component (calling this on NACL::C::VolumeCloneSplit would return
a NACL::CS::VolumeCloneSplit object)

This method is generally not implemented in individual components. The base
class (L<NACL::C::Component::ONTAP|lib-NACL-C-Component-ONTAP-pm> provides an implementation.
See <L<NACL::C::Component::state|lib-NACL-C-Component-pm/state> for a more
detailed description along with explanation of all the options accepted.

=cut

=head2 start

    my $VolumeCloneSplit_object    = NACL::C::VolumeCloneSplit->start(
                                command_interface => $ci,
                                 flexclone => $flexclone,
                                 vserver => $vserver,
                                  ...);

(Class method)
This method is used to split a FlexClone from the parent volume.

Uses a CMode CLI/ZAPI, 7Mode CLI APISet.

Invokes 'volume-clone-split-start' API for CMode ZAPI.

Invokes 'vol clone split start' command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "flexclone" => $flexclone >>

(Required)

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: volume

=item C<< "vserver" => $vserver >>

(Required)

Applicable for CMode CLI.

=item C<< "foreground" => $string >>

Applicable for CMode CLI.

=item C<< "job_component"=> \$scalar >>

This is a reference to a variable, which gets filled in with a job component
if the operation was performed through a background job. We can then wait
on the completion of the job by doing
  $job->watch_progress('method-timeout' => $time_it_might_take_in_seconds);

=item C<< 'extended_query' => $string >>

(Optional)

Applicable for CMode CLI.

=item C<< '*' >>

This method takes all other options supported by
"volume clone split start" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=item C<NACL::C::Exceptions::VolumeCloneSplit::SplitInProgress>
This exception is thrown when split operation is in progress for a volume

=back

=back

=cut

sub start {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            job_component => {type => SCALARREF, optional => 1},
            _optional_scalars(qw( foreground )),
        },
        cli_cmd => 'volume clone split start',
    );
    my $vserver;
    $opts{vserver_operated_on} = \$vserver;
    try {
        $pkg->call_on_apiset(
            %opts,
            choices => [
                {   method    => '_start_cmode_cli',
                    interface => 'CLI',
                    set       => 'CMode',
                },
                {   method    => '_start_cmode_zapi',
                    interface => 'ZAPI',
                    set       => 'CMode',
                },
                {   method    => '_start_7mode_cli',
                    interface => 'CLI',
                    set       => '7Mode',
                }
            ]
        );
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = $_[0];
        NACL::C::Exceptions::VolumeCloneSplit::SplitInProgress
            ->detect_convert_and_throw(exception => $exception);
        $Log->exit() if $may_exit;
        $exception->throw();
    };
    my $obj = $pkg->new(
        command_interface => $opts{command_interface},
        flexclone         => $opts{'flexclone'},
        vserver           => $vserver,
    );
    $Log->exit() if $may_exit;
    return $obj;
} ## end sub start

sub _start_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, %opts) = @_;
    my $response;
    $pkg->_base_backend_cmode_cli(%opts,
        api => 'volume_clone_split_start',
        apiset_response_obj => \$response
    );
    $Log->exit() if $may_exit;
} ## end sub _start_cmode_cli

sub _start_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    my $response;
    $pkg_or_obj->_base_backend_cmode_zapi(%opts,
            api    => 'volume_clone_split_start',
            map    => {'flexclone' => 'volume'},
            apiset_response_obj => \$response
    );
    $Log->exit() if $may_exit;
} ## end sub _start_cmode_zapi

sub _start_7mode_cli {
    $Log->enter() if $may_enter;
    my %args;
    my $pkg  = shift;
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );

    my $apiset            = delete $opts{apiset};
    my $command_interface = delete $opts{command_interface};
    $args{volume} = $opts{flexclone};
    $apiset->vol_clone_split_start(%args);
    $Log->exit() if $may_exit;
    return $pkg->new(
        command_interface => $command_interface,
        flexclone         => $args{volume}
    );
} ## end sub _start_7mode_cli

=head2 stop

    NACL::C::VolumeCloneSplit->stop(
                                command_interface => $ci,
                                 flexclone => $flexclone,
                                 vserver => $vserver,
                                  ...);

                                 or

    $VolumeCloneSplit_object->stop(...);

(Class or instance method)
This method is used to stop an ongoing FlexClone split job.

Uses a CMode CLI/ZAPI, 7Mode CLI APISet.

Invokes 'volume-clone-split-stop' API for CMode ZAPI.

Invokes 'vol clone split stop' command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "flexclone" => $flexclone >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: volume

=item C<< "vserver" => $vserver >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI.

=item C<< 'extended_query' => $string >>

(Optional)

Applicable for CMode CLI.

=item C<< '*' >>

This method takes all other options supported by
"volume clone split stop" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub stop {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params  => \@args,
        cli_cmd => 'volume clone split stop',
    );
    my $vserver;
    $opts{vserver_operated_on} = \$vserver;
    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_stop_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_stop_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
            {   method    => '_stop_7mode_cli',
                interface => 'CLI',
                set       => '7Mode',
            },
        ],
        frontend => 'NACL::C::VolumeCloneSplit::stop',
    );

    $Log->exit() if $may_exit;
    return $ret;
} ## end sub stop

sub _stop_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@args,  api => 'volume_clone_split_stop');
    $Log->exit() if $may_exit;
} ## end sub _stop_cmode_cli

sub _stop_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_base_backend_cmode_zapi(@args,
            api    => 'volume_clone_split_stop',
            map    => {"flexclone" => "volume"},
    );
    $Log->exit() if $may_exit;
} ## end sub _stop_cmode_zapi

sub _stop_7mode_cli {
    $Log->enter() if $may_enter;
    my %args;
    my $pkg  = shift;
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );
    my $apiset            = delete $opts{apiset};
    my $command_interface = delete $opts{command_interface};
    $args{volume} = $opts{flexclone};
    $apiset->vol_clone_split_stop(%args);
    $Log->exit() if $may_exit;
} ## end sub _stop_7mode_cli

=head2 estimate

    NACL::C::VolumeCloneSplit->estimate(
                                     command_interface => $ci,
                                     flexclone => $flexclone,
                                     vserver => $vserver,
                                  ...);

                                                                 or

    $estimate = $VolumeCloneSplit_object->estimate(...);

(Class or instance method)
This method is used to get the split estimate of a FlexClone
returns a scalar

Uses a CMode CLI, 7Mode CLI APISet.

Invokes 'vol clone split estimate' command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "flexclone" => $flexclone >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI.

=item C<< "vserver" => $vserver >>

(Required for cmode )

Applicable for CMode CLI.

=item C<< '*' >>

This method takes all other options supported by
"volume clone show" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub estimate {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params  => \@args,
        cli_cmd => 'volume clone show',
    );
    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_estimate_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_estimate_7mode_cli',
                interface => 'CLI',
                set       => '7Mode',
            },
        ],
        frontend => 'NACL::C::VolumeCloneSplit::estimate',
    );

    $Log->exit() if $may_exit;
    return $ret;
} ## end sub estimate

sub _estimate_7mode_cli {
    $Log->enter() if $may_enter;
    my %args;
    my $pkg  = shift;
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );

    my $apiset = delete $opts{apiset};
    $args{'volume'} = $opts{'flexclone'};
    my $response = $apiset->vol_clone_split_estimate(%args);
    my $results  = $response->get_parsed_output();
    $Log->exit() if $may_exit;
    return @$results[0]->{split_estimate};
} ## end sub _estimate_7mode_cli

#
#The cdef for "volume clone split estimate" is not generated for cmode.
#Alternative way of find the split estimate is to do a
#volume clone show -flexclone component_test_clone -vserver component_vserver
#-fields split-estimate
#
#
sub _estimate_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, %opts) = @_;
    my $vserver = $opts{vserver};
    my $response;

    $opts{'additional_args'}->{'fields'} = 'split-estimate';

    $pkg->_base_backend_cmode_cli(
        %opts,
        api => 'volume_clone_show',
        vserver_opearted_on => \$vserver,
        apiset_response_obj => \$response,
    );
    my $output = $response->get_parsed_output();
    $Log->exit() if $may_exit;
    return @$output[0]->{'split-estimate'};
} ## end sub _estimate_cmode_cli


__PACKAGE__->_load_event_class();

sub _primary_keys_validate_spec {
    return (
        'flexclone' => {type => SCALAR},
        'vserver'   => {type => SCALAR | UNDEF, optional => 1},
    );
} ## end sub _primary_keys_validate_spec

1;
