#
# Copyright (c) 2001-2017 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Volume VolumeEfficiency Component Module
## @author Peter.Rutenbar@netapp.com, dl-nacl-dev@netapp.com, benjaram@netapp.com
## @status public
## @pod here

=head1 NAME

NACL::C::VolumeEfficiency

=head1 DESCRIPTION

C<NACL::C::VolumeEfficiency> is a derived class of
L<NACL::C::Component::ONTAP|lib-NACL-C-Component-ONTAP-pm>.  It represents access to
ONTAP volume efficiency.  A related class is
L<NACL::CS::VolumeEfficiency|lib-NACL-CS-VolumeEfficiency-pm>, which represents
the state of efficiency on an ONTAP volume.

=head1 ATTRIBUTES

=head2 command_interface

See L<NACL::C::Component::command_interface|lib-NACL-C-Component-pm/command_interface>

=head2 volume

The name of the volume on which to manage efficiency.

=head2 vserver

(Required for CMode, Not Applicable for 7Mode) The name of the vserver in which the volume is present.

=cut

package NACL::C::VolumeEfficiency;

use strict;
use warnings;
use NACL::CS::VolumeEfficiency;
use Params::Validate
    qw(validate_with SCALAR ARRAYREF HASHREF UNDEF SCALARREF OBJECT);
use NACL::ComponentUtils qw (_optional_scalars);
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use NACL::C::Vserver;
use Scalar::Util qw(blessed);

# Exceptions
use NACL::C::Exceptions::VolumeEfficiency::OperationAlreadyEnabled  ();
use NACL::C::Exceptions::VolumeEfficiency::OperationAlreadyDisabled ();
use NACL::C::Exceptions::VolumeEfficiency::OperationCurrentlyIdle   ();
use NACL::C::Exceptions::VolumeEfficiency::VolumeDoesNotExist       ();
use NACL::C::Exceptions::VolumeEfficiency::VserverDoesNotExist      ();
use NACL::C::Exceptions::VolumeEfficiency::RestrictedVolume         ();
use NACL::C::Exceptions::VolumeEfficiency::OperationCurrentlyActive ();
use NACL::C::Exceptions::VolumeEfficiency::OfflineVolume            ();
use NACL::C::Exceptions::VolumeEfficiency::OperationNotEnabled      ();
use NACL::C::Exceptions::VolumeEfficiency::OperationNotDisabled     ();
use NACL::C::Exceptions::VolumeEfficiency::DisableInlineCompression ();
use NACL::C::Exceptions::VolumeEfficiency::CompressionNotEnabled    ();
use NACL::C::Exceptions::VolumeEfficiency::RevertToFailedSnapshotCopied    ();
use NACL::C::Exceptions::VolumeEfficiency::RevertToFailedUndoCompression   ();
use NACL::C::Exceptions::VolumeEfficiency::PromoteDeprioritizedVolume       ();
use NACL::APISet::Exceptions::CommandFailedException qw(:try);
use NACL::C::Exceptions::VolumeEfficiency::NotEnoughDiskSpace ();
use
    NACL::C::Exceptions::VolumeEfficiency::VolumeIsDestinationForFileTransfers (
    );

use NATE::BaseException qw(:try);

use NACL::C::_Mixins::Common qw(:iv_methods);

use base 'NACL::C::Component::ONTAP';
use NATE::ParamSet qw( param_global);

=head1 METHODS

=head2 new

 my $ef = NACL::C::VolumeEfficiency->new(command_interface=>$ci,
                                     vserver => "my_vserver",
                                     volume => "my_volume");

                            or

 my $ef = NACL::C::VolumeEfficiency->new(command_interface=>$ci,
                                     vserver => "my_vserver",
                                     path    => "my_path");

Return a new VolumeEfficiency component object with the given attributes.

=cut

use Class::MethodMaker [
    scalar => 'vserver',
    scalar => [
        {   '-default_ctor' => sub {
                my $self = $_[0];
                my $path = '/vol/' . $self->volume();
                return $path;
            },
        },
        'path'
    ],
    scalar => [
        {   '-default_ctor' => sub {
                if (ref $_[0]) {
                    return $_[0]
                        ->_extract_volume_from_path(path => $_[0]->path());
                }
            },
        },
        'volume'
    ],

    # Not a primary key, but used to determine which ZAPI to invoke.
    # Hence, is a private attribute and is undocumented.
    scalar => '_is_infinivol',
];

sub isa {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, $kind) = @_;
    my $isa = $pkg_or_obj->_build_isa(
        kind  => $kind,
        alias => 'NACL::C::Efficiency'
    );
    $Log->exit() if $may_exit;
    return $isa;
} ## end sub isa

=head2 on

 $ef->on()

(Class or instance method) Enable efficiency on a volume.

Uses CMode CLI/ZAPI, 7Mode CLI.

Invokes 'sis-enable-async' or 'sis-enable' API for CMode ZAPI.

Invokes 'sis on' command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< volume=> $string >>

(Required if path is not provided) Name of volume.

Applicable for CMode CLI/ZAPI, 7Mode CLI.

Maps to:

CM ZAPI:

Since this option is not passed to the apilayer call ,there is no mapping
information provided ,path will be constructed using this
argument and passed to the apilayer call.

7M CLI:

Since this option is not passed to the apilayer call ,there is no mapping
information provided ,path will be constructed using
this argument and passed to the apilayer call.

=item C<< vserver=>$string >>

(Required for CMode, Not Applicable for 7Mode) The name of the vserver
in which the volume is present.

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: ZAPI will be invoked in the context of the vserver specified to this
option.

=item C<< path=>$string >>

(Required if volume is not provided)
Path to the volume

Applicable for CMode CLI/ZAPI, 7Mode CLI.

7M CLI:

path (Programmatic name: path)

If this option is not provided then path will be constructed from the volume
and passed to API layer.

CM ZAPI: path

If the volume is on repository vserver then 'volume-name' param will be
extracted from the path and invokes sis-enable-async API.

For flexible volume this option is mapped to 'path', if its not provided
then path will be constructed from the volume and passed to sis-enable API.

=item C<< 'start-lrse-xfer' => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: lrse

=item C<< "job_component"=> \$scalar >>

Filled in for CMode CLI/ZAPI.

Applicable for infinite volume.

=item C<< 'is-repository' => $boolean >>

B<DEPRECATED>. Use C<is_infinivol> instead.

=item C<< is_infinivol => 0 | 1 >>

(Optional, Applicable for CMode ZAPI)

This boolean option describes whether the volume is an infinite volume or not.
For CMode ZAPI, the "async" form of the ZAPI must be used for the infinite
volume. If this option is not specified, the library will attempt to
determine if an infinite volume is being operated on (this could involve
querying the filer).

=item C<< is_infinivol_ref => \$scalar >>

(Optional, SCALARREF)

If the method was invoked through CMode ZAPI, the variable gets updated
describing whether it operated on an infinivol (used the "async" ZAPI) or not
(used the normal ZAPI).

=item C<< "extended_query" => $string >>

(Optional)

Applicable for CMode CLI.

=item C<< 'needs-upgrade' => $string >>


Enable efficiency on the volume only if the efficiency policy of the volume is needs-upgrade
Applicable for CMode CLI

=item C<< '*' >>

This method takes all other options supported by
"volume efficiency on" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=item C<NACL::C::Exceptions::VolumeEfficiency::VolumeDoesNotExist>

This type of exception gets thrown if an attempt is made to perform
volume efficiency on a volume which does not exists.

=item C<NACL::C::Exceptions::VolumeEfficiency::VserverDoesNotExist>

This type of exception gets thrown if an attempt is made to perform
volume efficiency on a vserver which does not exists.

=item C<NACL::C::Exceptions::VolumeEfficiency::RestrictedVolume>

This type of exception gets thrown if an attempt is made to perform
volume efficiency on a volume which is in restricted state.

=item C<NACL::C::Exceptions::VolumeEfficiency::OperationAlreadyEnabled>

This type of exception gets thrown if an attempt is made to on
volume efficiency when it is already enabled.

=item C<NACL::C::Exceptions::VolumeEfficiency::OfflineVolume>

This type of exception gets thrown if an attempt is made to perform
volume efficiency on a volume which is offline.

=back

=back

=cut

sub on {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            job_component => {type => SCALARREF, optional => 1},
            _optional_scalars qw(
                start-lrse-xfer
                volume path is-repository

                needs-upgrade
                ),
        },
        cli_cmd => 'volume efficiency on',
    );

    my $is_infinivol_ref;
    $opts{is_infinivol_ref} ||= \$is_infinivol_ref;

    try {
        $pkg_or_obj->call_on_apiset(
            %opts,
            choices => [
                {   method    => "_on_cmode_cli",
                    interface => "CLI",
                    set       => "CMode",
                },
                {   method    => "_on_cmode_zapi",
                    interface => "ZAPI",
                    set       => "CMode",
                    check     => '_infinivol_vol_vs_check',
                },
                {   method    => "_on_7mode_cli",
                    interface => "CLI",
                    set       => "7Mode",
                }
            ],
            frontend => 'NACL::C::VolumeEfficiency::on',
        );

    } ## end try
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception_object = $_[0];
        foreach my $name (
            qw(NotEnoughDiskSpace OfflineVolume RestrictedVolume OperationAlreadyEnabled VolumeDoesNotExist VserverDoesNotExist)
            )
        {
            my $full_name = "NACL::C::Exceptions::VolumeEfficiency::$name";
            $full_name->detect_convert_and_throw(
                exception => $exception_object);
        }
        $Log->exit() if $may_exit;
        $exception_object->throw();
    };

    my %new_args = (path => $opts{path});
    if ($opts{command_interface}->is_cmode()) {
        $new_args{vserver} = $opts{vserver};
    }

    $pkg_or_obj->_process_options(options => \%new_args);
    my $self = $pkg_or_obj->get_package_name()
        ->new(%new_args, command_interface => $opts{command_interface});
    $self->_update_infinivol_val_in_self(%opts);

    $Log->exit() if $may_exit;
    return $self;
} ## end sub on

sub _on_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    my $vserver = $opts{vserver};

    $pkg_or_obj->_base_backend_cmode_cli(
        %opts,
        api                 => 'volume_efficiency_on',
        vserver_operated_on => \$vserver,
    );

    $Log->exit() if $may_exit;
} ## end sub _on_cmode_cli

sub _on_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    $pkg_or_obj->_backend_cmode_zapi_infinivol_builder(%opts,
        api => 'sis-enable',);
    $Log->exit() if $may_exit;
} ## end sub _on_cmode_zapi

sub _on_7mode_cli {
    my $pkg_or_obj = shift;
    $Log->enter() if $may_enter;
    my %args;
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};

    $args{path} = $opts{path};

    $apiset->sis_on(%args);
    $Log->exit() if $may_exit;
} ## end sub _on_7mode_cli

=head2 off

 NACL::C::VolumeEfficiency->off(command_interface=>$ci,volume =>$vol_name,
                           vserver => $vserver_name);

                            or

 $ef->off();
                            or
    NACL::C::VolumeEfficiency->off(
    command_interface => $ci,
        vserver           => $vserver,
        path              => $path
        ...
    )


(Class or instance method) Disable efficiency on a volume.

Uses CMode CLI/ZAPI, 7Mode CLI APISet.

Invokes 'sis-disable-async' or 'sis-disable' API for CMode ZAPI.

Invokes 'sis off' command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< volume=> $string >>

(Required) Name of volume.

Applicable for CMode CLI/ZAPI, 7Mode CLI.

Maps to:

CM ZAPI:

Since this option is not passed to the apilayer call ,there is no mapping
information provided ,path will be constructed using this
argument and passed to the apilayer call.

7M CLI:

Since this option is not passed to the apilayer call ,there is no mapping
information provided ,path will be constructed using
this argument and passed to the apilayer call.

=item C<< vserver=>$string >>

(Required for CMode, Not Applicable for 7Mode) The name of the vserver
in which the volume is present.

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: ZAPI will be invoked in the context of the vserver specified to this
option.

=item C<< path=>$string >>

(Required if volume is not provided)
Path to the volume

Applicable for CMode CLI/ZAPI, 7Mode CLI.

7M CLI: path (Programmatic name: path)

If this option is not provided then path will be constructed from the volume
and passed to API layer.

CM ZAPI: path

If the volume is on repository vserver then 'volume-name' param will be
extracted from the path and invokes sis-disable-async API.

For flexible volume this option is mapped to 'path', if its not provided
then path will be constructed from the volume and passed to sis-disable API.

=item C<< 'end-lrse-xfer' => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: lrse

=item C<< "job_component"=> \$scalar >>

Filled in for CMode CLI/ZAPI.

Applicable for infinite volume.

=item C<< 'is-repository' => $boolean >>

B<DEPRECATED>. Use C<is_infinivol> instead.

=item C<< is_infinivol => 0 | 1 >>

(Optional, Applicable for CMode ZAPI)

This boolean option describes whether the volume is an infinite volume or not.
For CMode ZAPI, the "async" form of the ZAPI must be used for the infinite
volume. If this option is not specified, the library will attempt to
determine if an infinite volume is being operated on (this could involve
querying the filer).

=item C<< is_infinivol_ref => \$scalar >>

(Optional, SCALARREF)

If the method was invoked through CMode ZAPI, the variable gets updated
describing whether it operated on an infinivol (used the "async" ZAPI) or not
(used the normal ZAPI).

=item C<< 'extended_query' => $string >>

(Optional)

Applicable for CMode CLI.

=item C<< '*' >>

This method takes all other options supported by
"volume efficiency off" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=item C<NACL::C::Exceptions::VolumeEfficiency::VolumeDoesNotExist>

This type of exception gets thrown if an attempt is made to perform
volume efficiency on a volume which does not exists.

=item C<NACL::C::Exceptions::VolumeEfficiency::VserverDoesNotExist>

This type of exception gets thrown if an attempt is made to perform
volume efficiency on a vserver which does not exists.

=item C<NACL::C::Exceptions::VolumeEfficiency::RestrictedVolume>

This type of exception gets thrown if an attempt is made to perform
volume efficiency on a volume which is in restricted state.

=item C<NACL::C::Exceptions::VolumeEfficiency::OperationAlreadyDisabled>

This type of exception gets thrown if an attempt is made to off
volume efficiency when it is already disabled.

=item C<NACL::C::Exceptions::VolumeEfficiency::OfflineVolume>

This type of exception gets thrown if an attempt is made to perform
volume efficiency off on a volume which is offline.

=item C<NACL::C::Exceptions::VolumeEfficiency::OperationCurrentlyActive>

This type of exception gets thrown if an attempt is made to off volume efficiency
on a volume when the operation is currently active.

=item C<NACL::C::Exceptions::VolumeEfficiency::OperationNotEnabled>

This type of exception is thrown when an attempt is made to disable the volume efficiency when
efficiency was not enabled.

=item C<NACL::C::Exceptions::VolumeEfficiency::OperationAlreadyDisabled>

This type of exception is thrown when an attempt is made to disable volume efficiency when it is already disabled.

=back

=back

=cut

sub off {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            job_component => {type => SCALARREF, optional => 1},
            _optional_scalars qw(
                end-lrse-xfer
                volume path is-repository
                ),
        },
        cli_cmd => 'volume efficiency off',
    );

    my $ret;

    try {
        $ret = $pkg_or_obj->call_on_apiset(
            %opts,
            choices => [
                {   method    => "_off_cmode_cli",
                    interface => "CLI",
                    set       => "CMode",
                },
                {   method    => "_off_cmode_zapi",
                    interface => "ZAPI",
                    set       => "CMode",
                    check     => '_infinivol_vol_vs_check',
                },
                {   method    => "_off_7mode_cli",
                    interface => "CLI",
                    set       => "7Mode",
                }
            ],
            frontend => 'NACL::C::VolumeEfficiency::off',
        );
    } ## end try
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception_object = $_[0];
        foreach my $name (
            qw(RestrictedVolume OperationAlreadyDisabled OfflineVolume
            OperationCurrentlyActive VolumeDoesNotExist VserverDoesNotExist
            OperationNotEnabled )
            )
        {
            my $full_name = "NACL::C::Exceptions::VolumeEfficiency::$name";
            $full_name->detect_convert_and_throw(
                exception => $exception_object);
        }
        $Log->exit() if $may_exit;
        $exception_object->throw();
    };
    $Log->exit() if $may_exit;
    return $ret;

} ## end sub off

sub _off_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    my $vserver = $opts{vserver};

    $pkg_or_obj->_base_backend_cmode_cli(
        %opts,
        api                 => 'volume_efficiency_off',
        vserver_operated_on => \$vserver,
    );

    $Log->exit() if $may_exit;
} ## end sub _off_cmode_cli

sub _off_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_backend_cmode_zapi_infinivol_builder(@args,
        api => 'sis-disable',);
    $Log->exit() if $may_exit;
}

sub _off_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %args;
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};

    $args{path} = $opts{path};

    $apiset->sis_off(%args);
    $Log->exit() if $may_exit;

} ## end sub _off_7mode_cli

=head2 config

 NACL::C::VolumeEfficiency->config(command_interface=>$ci,volume =>$vol_name,
                             vserver => $vserver_name,
                             ...);

                            or

 $ef->config(...);
                            or

 NACL::C::VolumeEfficiency->config(command_interface=>$ci,path => $path,
                             vserver => $vserver_name,
                             ...);

(Class or instance method) Set the efficiency schedule for a volume.

Uses a  CMode CLI, 7Mode CLI APISet.

Invokes "sis config" command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< volume=> $string >>

(Required) Name of volume.

Applicable for CMode CLI, 7Mode CLI.

Maps to:

7M CLI:

Since this option is not passed to the apilayer call ,there is no mapping
information provided ,path will be constructed using
this argument and passed to the apilayer call.

=item C<< vserver=>$string >>

(Required for CMode, Not Applicable for 7Mode) The name of the vserver
in which the volume is present.

Applicable for CMode CLI.

=item C<< path=>$string >>

(Required if volume is not provided)
Path to the volume

Applicable for CMode CLI, 7Mode CLI.

Maps to:

7M CLI: path (Programmatic name: path)

If this option is not provided then path will be constructed from the volume
and passed to API layer.

=item C<< background-compression=>$boolean >>

Applicable for CMode CLI, 7Mode CLI.

Maps to:

7M CLI : -C <value> (Programmatic name: "enable-compression")

=item C<< compression=>$boolean >>

Applicable for CMode CLI, 7Mode CLI.

Maps to:

7M CLI : -C <value> (Programmatic name: "enable-compression")

=item C<< inline-compression=>$boolean >>

Applicable for CMode CLI, 7Mode CLI.

Maps to:

7M CLI : -I <value> (Programmatic name: "inline-compression")

=item C<< min-blocks-shared=>$scalar >>

Applicable for CMode CLI, 7Mode CLI.

Maps to:

7M CLI : -m <value> (Programmatic name: "minimum-blocks-shared")

=item C<< policy =>$string >>

Applicable for CMode CLI.

=item C<< schedule =>$string >>

Applicable for CMode CLI, 7Mode CLI.

Maps to:

7M CLI : -s <value> (Programmatic name: "schedule")

=item C<< idd =>$string >>

Applicable for CMode CLI, 7Mode CLI.

Maps to:

7M CLI : -Q <value> (Programmatic name: "idd")

=item C<< quick-check-fsize =>$string >>

Applicable for CMode CLI, 7Mode CLI.

Maps to:

7M CLI : -z <value> (Programmatic name: "quick-check-fsize")

=item C<< "extended_query" => $string >>

(Optional)

=item C<< 'verify-trigger-rate' => $string >>

Verify Trigger Rate possible value(s) are, integer
Applicable for CMode CLI

=item C<< 'optimize' => $string >>

Optimization possible value(s) are, performance,space-saving
Applicable for CMode CLI

=item C<< 'volume' => $string >>

Volume Name possible value(s) are, volume name
Applicable for CMode CLI

=item C<< 'path' => $string >>

Path to volume possible value(s) are, /vol/volume
Applicable for CMode CLI

=item C<< is_infinivol => 0 | 1 >>

(Optional, Applicable for CMode ZAPI)

This boolean option describes whether the volume is an infinite volume or not.
For CMode ZAPI, the "async" form of the ZAPI must be used for the infinite
volume. If this option is not specified, the library will attempt to
determine if an infinite volume is being operated on (this could involve
querying the filer).

=item C<< is_infinivol_ref => \$scalar >>

(Optional, SCALARREF)

If the method was invoked through CMode ZAPI, the variable gets updated
describing whether it operated on an infinivol (used the "async" ZAPI) or not
(used the normal ZAPI).

=item C<< '*' >>

This method takes all other options supported by
"volume efficiency config" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=item C<NACL::C::Exceptions::VolumeEfficiency::VolumeDoesNotExist>

This type of exception gets thrown if an attempt is made to perform
volume efficiency on a volume which does not exists.

=item C<NACL::C::Exceptions::VolumeEfficiency::VserverDoesNotExist>

This type of exception gets thrown if an attempt is made to perform
volume efficiency on a vserver which does not exists.

=item C<NACL::C::Exceptions::VolumeEfficiency::RestrictedVolume>

This type of exception gets thrown if an attempt is made to perform
volume efficiency on a volume which is in restricted state.

=item C<NACL::C::Exceptions::VolumeEfficiency::DisableInlineCompression>

This type of exception gets thrown if an attempt is made to perform volume efficiency
config when inline compression is enabled.

=item C<NACL::C::Exceptions::VolumeEfficiency::OfflineVolume>

This type of exception gets thrown if an attempt is made to perform
volume efficiency off on a volume which is offline.

=item C<NACL::C::Exceptions::VolumeEfficiency::CompressionNotEnabled>

This type of exception gets thrown if an attempt is made to perform volume efficiency
config when compression is not enabled.

=item C<< "inline-dedupe" => $string >>

( Boolean Optional )

Inline dedup defines the state of the volume Attributes : non-creatable, non-modifiable

Applicable for CMode CLI/ZAPI

Maps to:

CM ZAPI : is-inline-dedupe-enabled

=item C<< "data-compaction" => $string >>

( Boolean Optional )

Enable/Disable data compaction on the sis volume.

Applicable for CMode CLI/ZAPI

Maps to:

CM ZAPI : enable-data-compaction

=back

=back

=cut

sub config {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars qw(
                background-compression compression inline-compression
                min-blocks-shared policy schedule idd quick-check-fsize

                verify-trigger-rate optimize

                volume path inline-dedupe data-compaction
                )
        },
        cli_cmd => 'volume efficiency config',
    );
    try {
        $pkg_or_obj->call_on_apiset(
            %opts,
            choices => [
                {   method    => "_config_cmode_cli",
                    interface => "CLI",
                    set       => "CMode",
                },

                # "config" is really a hidden alias to "modify", so reuse the
                # back-end methods
                {   method    => "_modify_7mode_cli",
                    interface => "CLI",
                    set       => "7Mode",
                },
                {   method    => "_modify_cmode_zapi",
                    interface => "ZAPI",
                    set       => 'CMode',
                },
                {   method    => '_modify_7mode_zapi',
                    interface => 'ZAPI',
                    set       => '7Mode',
                },
            ],
            frontend => 'NACL::C::VolumeEfficiency::config',
        );
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception_object = $_[0];

        foreach my $name (
            qw(RestrictedVolume DisableInlineCompression OfflineVolume CompressionNotEnabled  VolumeDoesNotExist VserverDoesNotExist )
            )
        {
            my $full_name = "NACL::C::Exceptions::VolumeEfficiency::$name";
            $full_name->detect_convert_and_throw(
                exception => $exception_object);
        }
        $Log->exit() if $may_exit;
        $exception_object->throw();
    };
    $Log->exit() if $may_exit;
} ## end sub config

sub _config_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    my $vserver = $opts{vserver};

    $pkg_or_obj->_base_backend_cmode_cli(
        %opts,
        api                 => 'volume_efficiency_config',
        vserver_operated_on => \$vserver,
    );

    $Log->exit() if $may_exit;
} ## end sub _config_cmode_cli

=head2 stop

 NACL::C::VolumeEfficiency->stop(command_interface=>$ci,volume =>$vol_name,
                             vserver => $vserver_name, ...);

                            or

 $ef->stop(...);
                            or

 NACL::C::VolumeEfficiency->stop(command_interface=>$ci,path=> $path,
                             vserver => $vserver_name, ...);

(Class or instance method) Aborts currently active SIS operations on a volume.

Uses a  CMode CLI/ZAPI, 7Mode CLI APISet.

Invokes "sis stop" command for 7Mode CLI.

Invokes "sis-stop/sis-stop-async" API's for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< volume=> $string >>

(Required) Name of volume.

Applicable for CMode CLI/ZAPI, 7Mode CLI.

Maps to:

CM ZAPI:

Since this option is not passed to the apilayer call ,there is no mapping
information provided ,path will be constructed using this
argument and passed to the apilayer call.

7M CLI:

Since this option is not passed to the apilayer call ,there is no mapping
information provided ,path will be constructed using
this argument and passed to the apilayer call.

=item C<< vserver=>$string >>

(Required for CMode, Not Applicable for 7Mode) The name of the vserver
in which the volume is present.

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: ZAPI will be invoked in the context of the vserver specified to this
option.

=item C<< all=>$boolean >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: all-operations

=item C<< path=>$string >>

(Required if volume is not provided)
Path to the volume

Applicable for CMode CLI/ZAPI, 7Mode CLI.

7M CLI: path (Programmatic name: path)

If this option is not provided then path will be constructed from the volume
and passed to API layer.

CM ZAPI: path

If the volume is on repository vserver then 'volume-name' param will be
extracted from the path and invokes sis-stop-async API.

For flexible volume this option is mapped to 'path', if its not provided
then path will be constructed from the volume and passed to sis-stop API.

=item C<< 'extended_query' => $string >>

(Optional)

Applicable for CMode CLI.

=item C<< "job_component"=> \$scalar >>

Filled in for CMode CLI/ZAPI.

Applicable for infinite volume.

=item C<< 'is-repository' => $boolean >>

B<DEPRECATED>. Use C<is_infinivol> instead.

=item C<< is_infinivol => 0 | 1 >>

(Optional, Applicable for CMode ZAPI)

This boolean option describes whether the volume is an infinite volume or not.
For CMode ZAPI, the "async" form of the ZAPI must be used for the infinite
volume. If this option is not specified, the library will attempt to
determine if an infinite volume is being operated on (this could involve
querying the filer).

=item C<< is_infinivol_ref => \$scalar >>

(Optional, SCALARREF)

If the method was invoked through CMode ZAPI, the variable gets updated
describing whether it operated on an infinivol (used the "async" ZAPI) or not
(used the normal ZAPI).

=item C<< '*' >>

This method takes all other options supported by
"volume efficiency stop" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

This type of exception gets thrown if an attempt is made to perform
volume efficiency on a volume which does not exists.

=item C<NACL::C::Exceptions::VolumeEfficiency::VserverDoesNotExist>

This type of exception gets thrown if an attempt is made to perform
volume efficiency on a vserver which does not exists.

=item C<NACL::C::Exceptions::VolumeEfficiency::RestrictedVolume>

This type of exception gets thrown if an attempt is made to perform
volume efficiency on a volume which is in restricted state.

=item C<NACL::C::Exceptions::VolumeEfficiency::OfflineVolume>

This type of exception gets thrown if an attempt is made to perform
volume efficiency off on a volume which is offline.

=item C<NACL::C::Exceptions::VolumeEfficiency::OperationCurrentlyIdle>

This type of exception gets thrown if an attempt is made to stop volume efficiency
 when operation is currently idle.

=item C<NACL::C::Exceptions::VolumeEfficiency::OperationNotEnabled>

This type of exception is thrown when an attempt is made to stop the volume efficiency when
efficiency was not enabled.

=item C<NACL::C::Exceptions::VolumeEfficiency::OperationAlreadyDisabled>

This type of exception is thrown when an attempt is made to stop volume efficiency when it is already disabled.

=back

=back

=cut

sub stop {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            job_component => {type => SCALARREF, optional => 1},
            _optional_scalars qw(volume all
                path is-repository
                ),
        },
        cli_cmd => 'volume efficiency stop',
    );
    try {
        $pkg_or_obj->call_on_apiset(
            %opts,
            choices => [
                {   method    => "_stop_cmode_cli",
                    interface => "CLI",
                    set       => "CMode",
                },
                {   method    => "_stop_7mode_cli",
                    interface => "CLI",
                    set       => "7Mode",
                },
                {   method    => "_stop_cmode_zapi",
                    interface => "ZAPI",
                    set       => "CMode",
                    check     => '_infinivol_vol_vs_check',
                },
            ],
            frontend => 'NACL::C::VolumeEfficiency::stop',
        );
    } ## end try
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = $_[0];
        foreach my $name (
            qw(RestrictedVolume OperationCurrentlyIdle OfflineVolume
            VolumeDoesNotExist VserverDoesNotExist OperationNotEnabled
            OperationAlreadyDisabled)
            )
        {
            my $full_name = "NACL::C::Exceptions::VolumeEfficiency::$name";
            $full_name->detect_convert_and_throw(exception => $exception);
        }
        $Log->exit() if $may_exit;
        $exception->throw();
    };

    $Log->exit() if $may_exit;
} ## end sub stop

sub _stop_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    my $vserver = $opts{vserver};

    $pkg_or_obj->_base_backend_cmode_cli(
        %opts,
        api                 => 'volume_efficiency_stop',
        vserver_operated_on => \$vserver,
    );

    $Log->exit() if $may_exit;
} ## end sub _stop_cmode_cli

sub _stop_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %args;
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    delete $opts{command_interface};
    my $apiset = delete $opts{apiset};
    $args{path} = $opts{path};
    $apiset->sis_stop(%args);
    $Log->exit() if $may_exit;

} ## end sub _stop_7mode_cli

sub _stop_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_backend_cmode_zapi_infinivol_builder(
        @args,
        api => 'sis-stop',
        map => {all => 'all-operations'},
    );
    $Log->exit() if $may_exit;
}

=head2 start

 NACL::C::VolumeEfficiency->start(command_interface=>$ci,volume =>$vol_name,
                             vserver => $vserver_name, ...);

                            or

 $ef->start(...);
                            or

 NACL::C::VolumeEfficiency->start(command_interface=>$ci,path=> $path,
                             vserver => $vserver_name, ...);

(Class or instance method) Start a SIS operation on a volume.

Uses a  CMode CLI/ZAPI, 7Mode CLI APISet.

Invokes "sis-start-async or sis-start" API for CMode ZAPI.

Invokes "sis start" command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< volume=> $string >>

(Required) Name of volume.

Applicable for CMode CLI/ZAPI, 7Mode CLI.

Maps to:

CM ZAPI:

Since this option is not passed to the apilayer call ,there is no mapping
information provided ,path will be constructed using this
argument and passed to the apilayer call.

7M CLI:

Since this option is not passed to the apilayer call ,there is no mapping
information provided ,path will be constructed using
this argument and passed to the apilayer call.

=item C<< vserver=>$string >>

(Required for CMode, Not Applicable for 7Mode) The name of the vserver
in which the volume is present.

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: ZAPI will be invoked in the context of the vserver specified to this
option.

=item C<< scan=>$boolean >>

Applicable for CMode CLI/ZAPI,7Mode CLI.

Maps to:

CM ZAPI: scan

7M CLI : -s (Programmatic name: "scan")

=item C<< use-checkpoint=>$boolean >>

Applicable for CMode CLI, 7Mode CLI.

Maps to:

7M CLI : -p (Programmatic name: "use-checkpoint")

=item C<< del-checkpoint=>$boolean >>

Applicable for CMode CLI, 7Mode CLI.

Maps to:

7M CLI : -d (Programmatic name: "delete-checkpoint")

=item C<< compression=>$boolean >>

Applicable for CMode CLI, 7Mode CLI.

Maps to:

7M CLI : -C (Programmatic name: "compression")

=item C<< dedupe=>$boolean >>

Applicable for CMode CLI, 7Mode CLI.
Maps to:

7M CLI : dedupe

=item C<< shared-blocks=>$boolean >>

Applicable for CMode CLI, 7Mode CLI.

Maps to:

7M CLI : shared-blocks

=item C<< snapshot-blocks=>$boolean >>

Applicable for CMode CLI, 7Mode CLI.
Maps to:

7M CLI : snapshot-blocks

=item C<< queue=>$boolean >>

Applicable for CMode CLI, 7Mode CLI.

Maps to:

CM ZAPI: queue-operation

7M CLI : -q (Programmatic name: "enqueue")

=item C<< force=>$boolean >>

Applicable for CMode CLI, 7Mode CLI.

Maps to:

7M CLI : -f (Programmatic name: "force")

=item C<< 'delete-checkpoint' => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: delete-checkpoint

=item C<< 'scan-old-data' => $string >>

Applicable for CMode CLI, 7Mode CLI.

Maps to:

7M CLI : -s (Programmatic name: "scan")

=item C<< path=>$string >>

(Required if volume is not provided)
Path to the volume

Applicable for CMode CLI/ZAPI, 7Mode CLI.

7M CLI: path (Programmatic name: path)

If this option is not provided then path will be constructed from the volume
and passed to API layer.

CM ZAPI: path

If the volume is on repository vserver then 'volume-name' param will be
extracted from the path and invokes sis-start-async API.

For flexible volume this option is mapped to 'path', if its not provided
then path will be constructed from the volume and passed to sis-start API.

=item C<< 'vault-update-done' => $string >>

Applicable for CMode CLI.

=item C<< 'qos-policy' => $string >>

QoS Policy

Applicable for CMode CLI

=item C<< 'scan-all' => $string >>

Scan all the data without shread block optimization(if scanning old data)

Applicable for CMode CLI.

=item C<< 'build-metadata' => $string >>

Build metadata wihtout sharing(if scanning old data)

Applicable for CMode CLI.

=item C<< "job_component"=> \$scalar >>

Filled in for CMode CLI/ZAPI.

Applicable for infinite volume.

=item C<< 'extended_query' => $string >>

(Optional)

Applicable for CMode CLI.

=item C<< 'is-repository' => $boolean >>

B<DEPRECATED>. Use C<is_infinivol> instead.

=item C<< is_infinivol => 0 | 1 >>

(Optional, Applicable for CMode ZAPI)

This boolean option describes whether the volume is an infinite volume or not.
For CMode ZAPI, the "async" form of the ZAPI must be used for the infinite
volume. If this option is not specified, the library will attempt to
determine if an infinite volume is being operated on (this could involve
querying the filer).

=item C<< is_infinivol_ref => \$scalar >>

(Optional, SCALARREF)

If the method was invoked through CMode ZAPI, the variable gets updated
describing whether it operated on an infinivol (used the "async" ZAPI) or not
(used the normal ZAPI).

=item C<< 'qos-level' => $string >>

QoS Level
Applicable for CMode CLI

=item C<< '*' >>

This method takes all other options supported by
"volume efficiency start" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=item C<NACL::C::Exceptions::VolumeEfficiency::VserverDoesNotExist>

This type of exception gets thrown if an attempt is made to perform
volume efficiency on a vserver which does not exists.

=item C<NACL::C::Exceptions::VolumeEfficiency::RestrictedVolume>

This type of exception gets thrown if an attempt is made to perform
volume efficiency on a volume which is in restricted state.

=item C<NACL::C::Exceptions::VolumeEfficiency::OfflineVolume>

This type of exception gets thrown if an attempt is made to perform
volume efficiency off on a volume which is offline.

=item C<NACL::C::Exceptions::VolumeEfficiency::OperationNotEnabled>

This type of exception gets thrown if an attempt is made to start volume efficiency
when it is not enabled.

=back

=back

=cut

sub start {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            job_component => {type => SCALARREF, optional => 1},
            _optional_scalars qw(
                scan use-checkpoint del-checkpoint compression
                dedupe shared-blocks snapshot-blocks queue force

                delete-checkpoint scan-old-data

                vault-update-done
                volume path

                qos-policy

                scan-all build-metadata is-repository

                qos-level
                ),
        },
        api => 'volume_efficiency_start',
    );
    try {
        $pkg_or_obj->call_on_apiset(
            %opts,
            choices => [
                {   method    => "_start_cmode_cli",
                    interface => "CLI",
                    set       => "CMode",
                },
                {   method    => "_start_7mode_cli",
                    interface => "CLI",
                    set       => "7Mode",
                },
                {   method    => "_start_cmode_zapi",
                    interface => "ZAPI",
                    set       => "CMode",
                    check     => '_infinivol_vol_vs_check',
                },
            ],
            frontend => 'NACL::C::VolumeEfficiency::start',
        );
    } ## end try
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception_object = $_[0];
        foreach my $name (
            qw(RestrictedVolume OperationNotEnabled OfflineVolume  VolumeDoesNotExist VserverDoesNotExist )
            )
        {
            my $full_name = "NACL::C::Exceptions::VolumeEfficiency::$name";
            $full_name->detect_convert_and_throw(
                exception => $exception_object);
        }
        $Log->exit() if $may_exit;
        $exception_object->throw();
    };

    $Log->exit() if $may_exit;
} ## end sub start

sub _start_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    my $vserver = $opts{vserver};

    $pkg_or_obj->_base_backend_cmode_cli(
        %opts,
        api                 => 'volume_efficiency_start',
        vserver_operated_on => \$vserver,
    );

    $Log->exit() if $may_exit;
} ## end sub _start_cmode_cli

sub _start_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %args;
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};

    my $allowed_7mode_options
        = [
        qw( scan use-checkpoint compression shared-blocks snapshot-blocks dedupe force path )
        ];
    my $args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => $allowed_7mode_options,
        map    => {
            'del-checkpoint' => 'delete-checkpoint',
            'queue'          => 'enqueue',
            'scan-old-data'  => 'scan'
        }
    );
    my %prompt_hash;
    my $prompts_answers = ['.*Are you sure you want to proceed.*' => 'y'];
    $prompt_hash{'connectrec-match_table'} = $prompts_answers;

    $apiset->sis_start(%$args, %prompt_hash);
    $Log->exit() if $may_exit;

} ## end sub _start_7mode_cli

sub _start_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_backend_cmode_zapi_infinivol_builder(
        @args,
        api  => 'sis-start',
        copy => [
            qw(build-metadata delete-checkpoint qos-policy
                restart-checkpoint scan-all)
        ],
        map => {
            'scan-old-data' => 'scan',
            'queue'         => 'queue-operation'
        },
    );
    $Log->exit() if $may_exit;
}

=head2 stat

    # Class method, list context
    my @eff_stats = NACL::C::VolumeEfficiency->stat(%opts);

                           or

    # Class method, scalar context
    my $eff_stat  = NACL::C::VolumeEfficiency->stat(%opts);

                           or

    # Instance method
    my $eff_stat = $eff_obj->stat(%opts);

(Class or instance method)
This method returns the statistics for volume efficiency operations. It returns
objects of the type L<NACL::CS::VolumeEfficiencyStat|lib-NACL-CS-VolumeEfficiencyStat-pm>.

When invoked with an object instance, the object returned will be only
for the efficiency operation modeled by the C<VolumeEfficiency> object
with which the method was invoked.

This method accepts all the parameters accepted by L<fetch|lib-NACL-CS-ComponentState-pm/fetch>
(requested_fields, filter, allow_empty) and has all of the other properties of
fetch methods (callable in scalar or list context, throws a NoElementsFound
exception if no matching elements are found)

=cut

sub stat {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts = $pkg_or_obj->_handle_obj_cs_like_calls(@args);
    require NACL::CS::VolumeEfficiencyStat;
    my @state_objs = NACL::CS::VolumeEfficiencyStat->fetch(%opts);
    $Log->exit() if $may_exit;
    return wantarray ? @state_objs : $state_objs[0];
} ## end sub stat

=head2 revert_to

  NACL::C::VolumeEfficiency->revert_to(command_interface=>$ci
                                 version => 7.3);

                              or
(Class or instance method) Request to downgrade ONTAP.

Uses a CMode CLI/ZAPI, 7Mode CLI APISet.

Invokes "sis-revert-to" API of CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< version=>$version >>

(Required) Version of ONTAP to downgrade

Applicable for 7Mode CLI,CMode CLI/ZAPI.

=item C<< path=>$path >>

(Required) volume path like /vol/$volname

Applicable for 7Mode CLI,CMode CLI/ZAPI.

Since its filled by backend method, mapping information not provided.

=item C<< volume=>$volume >>

Applicable for CMode CLI/ZAPI, 7Mode CLI.

CM ZAPI:

Since this option is not passed to the apilayer call ,there is no mapping
information provided ,path will be constructed using this
argument and passed to the apilayer call.

7M CLI : volume name (Programmatic name: "volume")

=item C<< clean-up=>$cleanup >>

Applicable for CMode CLI/ZAPI.

=item C<< delete=>$delete >>

Applicable for CMode CLI/ZAPI.

=item C<< 'volume' => $string >>

Volume Name possible value(s) are, volume name
Applicable for CMode CLI

=item C<< 'path' => $string >>

Volume Path possible value(s) are, /vol/volume
Applicable for CMode CLI

=item C<< '*' >>

This method takes all other options supported by
"volume efficiency revert-to" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=item C<NACL::C::Exceptions::VolumeEfficiency::VserverDoesNotExist>

This type of exception gets thrown if an attempt is made to perform
volume efficiency on a vserver which does not exists.

=item C<NACL::C::Exceptions::VolumeEfficiency::RestrictedVolume>

This type of exception gets thrown if an attempt is made to perform
volume efficiency on a volume which is in restricted state.

=item C<NACL::C::Exceptions::VolumeEfficiency::OfflineVolume>

This type of exception gets thrown if an attempt is made to perform
volume efficiency off on a volume which is offline.

=item C<NACL::C::Exceptions::VolumeEfficiency::OperationCurrentlyActive>

This type of exception gets thrown if an attempt is made to revert volume efficiency
on a volume when the operation is currently active.

=back

=back

=cut

sub revert_to {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars qw(version delete clean-up

                volume path
                )
        },
        cli_cmd => 'volume efficiency revert-to',
    );
    my $vserver;
    $opts{vserver_operated_on} = \$vserver;
    try {
        $pkg_or_obj->call_on_apiset(
            %opts,
            choices => [
                {   method    => "_revert_to_cmode_cli",
                    interface => "CLI",
                    set       => "CMode",
                },
                {   method    => "_revert_to_7mode_cli",
                    interface => "CLI",
                    set       => "7Mode",
                },
                {   method    => "_revert_to_cmode_zapi",
                    interface => "ZAPI",
                    set       => "CMode",
                },
            ],
            frontend => 'NACL::C::VolumeEfficiency::revert_to',
        );
    } ## end try
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception_object = $_[0];
        foreach my $name (
            qw(RestrictedVolume OfflineVolume  VolumeDoesNotExist VserverDoesNotExist OperationCurrentlyActive )
            )
        {
            my $full_name = "NACL::C::Exceptions::VolumeEfficiency::$name";
            $full_name->detect_convert_and_throw(
                exception => $exception_object);
        }
        $Log->exit() if $may_exit;
        $exception_object->throw();
    };
    $Log->exit() if $may_exit;
} ## end sub revert_to

sub _revert_to_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );

    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};

    my $allowed_7mode_options = [qw(version path)];
    my $args                  = $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => $allowed_7mode_options,
    );

    %opts = %$args;
    $apiset->sis_revert_to(%opts);
    $Log->exit() if $may_exit;
} ## end sub _revert_to_7mode_cli

sub _revert_to_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my $allowed_cmode_options = [qw( version path volume clean-up delete )];
    $pkg_or_obj->_base_backend_cmode_zapi(
        @args,
        api  => 'sis_revert_to',
        copy => $allowed_cmode_options,
    );
    $Log->exit() if $may_exit;
} ## end sub _revert_to_cmode_zapi

sub _revert_to_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@args,
        api => 'volume_efficiency_revert_to');
    $Log->exit() if $may_exit;
} ## end sub _revert_to_cmode_cli

=head2 dump

  NACL::C::VolumeEfficiency->dump(command_interface=>$ci
                            path => $path);

(Class or instance method) Dump SIS metafile records present in aggregate to the 'sisdump'
   directory of a specified volume.

Uses a 7Mode CLI APISet.

Invokes "sis dump" command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< path=>$path >>

(Required if volume is not provided)
Path to the volume

Applicable for 7Mode CLI.

=item C<< volume=>$volume >>

Applicable for 7Mode CLI.

=item C<< to => $volumename >>

Applicable for 7Mode CLI.

=item C<< file  => $scalar >>

Applicable for 7Mode CLI.

=item C<< offset => $scalar >>

Applicable for 7Mode CLI.

=item C<< dir => $scalar >>

Applicable for 7Mode CLI.

=item C<< verbose-mode => $boolean >>

Applicable for 7Mode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub dump {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params => \@args,
        additional_spec =>
            {_optional_scalars qw( to file offset dir verbose-mode )}   ## end

    );

    $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => "_dump_7mode_cli",
                interface => "CLI",
                set       => "7Mode|Nodescope",
            }
        ],
    );
    $Log->exit() if $may_exit;
} ## end sub dump

sub _dump_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );

    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};

    my $allowed_7mode_options = [qw(to file offset dir verbose-mode path)];
    my $args                  = $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => $allowed_7mode_options,
    );

    %opts = %$args;

    $pkg_or_obj->_handle_boolean_options(
        api_options     => \%opts,
        boolean_options => [qw(verbose-mode)],
        is_7m_cli       => 1
    );

    $apiset->sis_dump(%opts);
    $Log->exit() if $may_exit;
} ## end sub _dump_7mode_cli

=head2 ls

    my @ls = NACL::C::VolumeEfficiency->ls(%opts);

                               or

    my $ls  = NACL::C::VolumeEfficiency->ls(%opts);

(Class method only)
This method returns the metafiles for a sis enabled volumes. It returns
objects of the type L<NACL::CS::SisLs|lib-NACL-CS-SisLs-pm>.

This method accepts all the parameters accepted by L<fetch|lib-NACL-CS-ComponentState-pm/fetch>
(requested_fields, filter, allow_empty) and has all of the other properties of
fetch methods (callable in scalar or list context, throws a NoElementsFound
exception if no matching elements are found)

=cut

sub ls {
    $Log->enter() if $may_enter;
    my $pkg = shift;
    require NACL::CS::SisLs;
    my @state_objs = NACL::CS::SisLs->fetch(@_);
    $Log->exit() if $may_exit;
    return wantarray ? @state_objs : $state_objs[0];
} ## end sub ls

=head2 reset

  NACL::C::VolumeEfficiency->reset(command_interface=>$ci,
                             path =>$path);

(Class or instance method) resets sis settings

Uses a 7Mode-CLI APISet.

Invokes "sis reset" command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< path=>$path >>

(Required if volume is not provided)
Path to the volume

Applicable for 7Mode CLI.

=item C<< volume=>$volume >>

Applicable for 7Mode CLI.

=item C<< relative-path=>$scalar >>

Applicable for 7Mode CLI.

=item C<< location=>$scalar >>

Applicable for 7Mode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub reset {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {_optional_scalars qw(relative-path location)}
    );

    $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => "_reset_7mode_cli",
                interface => "CLI",
                set       => "7Mode|Nodescope",
            }
        ],
    );
    $Log->exit() if $may_exit;
} ## end sub reset

sub _reset_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );

    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};

    my $allowed_7mode_options = [qw(path relative-path location)];
    my $args                  = $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => $allowed_7mode_options,
    );

    %opts = %$args;
    my $prompts_answers = ['.*Are you sure you want to proceed.*' => 'y'];
    $opts{'connectrec-match_table'} = $prompts_answers;

    $apiset->sis_reset(%opts);
    $Log->exit() if $may_exit;
} ## end sub _reset_7mode_cli

=head2 modify

    NACL::C::VolumeEfficiency->modify(command_interface => $ci,
                                 volume => $volume,
                                 vserver => $vserver,
                                  ...);

                                 or

    $VolumeEfficiency_object->modify(...);

                                                                 or
    NACL::C::VolumeEfficiency->modify(command_interface => $ci,
                                     path    => $path,
                                     vserver => $vserver,
                                  ...);

(Class or instance method)
This method is used to modify efficiency configuration of a volume.

Uses a CMode CLI/ZAPI or 7Mode CLI APISet.

Invokes 'sis-set-config-async' or 'sis-set-config' API for CMode ZAPI.

Invokes 'sis config' command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "volume" => $volume >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI, 7Mode CLI.

CM ZAPI:

Since this option is not passed to the apilayer call ,there is no mapping
information provided ,path will be constructed using this
argument and passed to the apilayer call.

7M CLI : volume name (Programmatic name: "volume")

=item C<< "vserver" => $vserver >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: ZAPI will be invoked in the context of the vserver specified to this
option.

=item C<< "path" => $path >>

(Required for class method, Not Applicable for instance method)
(Required if volume is not provided)
Path to the volume

Applicable for CMode CLI/ZAPI, 7Mode CLI

Maps to:

CM ZAPI : path

7M CLI  : path name (Programmatic name: "path")

=item C<< "inline" => $string >>

Applicable for CMode CLI

=item C<< "background-compression" => $string >>

Applicable for CMode/7Mode CLI

Maps to:

7M CLI : -C <value> (Programmatic name : 'enable-compression')

=item C<< "schedule" => $string >>

Applicable for CMode CLI/ZAPI,7Mode CLI.

Maps to:

CM ZAPI: schedule

7M CLI : -s <value>(Programmatic name: "schedule")

=item C<< "compression" => $string >>

Applicable for CMode CLI/ZAPI, 7Mode CLI

Maps to:

7M CLI  : -C  <value> (Programmatic name : 'enable-compression')

CM ZAPI : enable-compression

=item C<< "verify-trigger-rate" => $string >>

Applicable for CMode CLI

=item C<< "inline-compression" => $string >>

Applicable for CMode CLI/ZAPI, 7Mode CLI

Maps to:

CM ZAPI : enable-inline-compression

7M CLI  : -I <value> (Programmatic name : 'inline-compression')

=item C<< "policy" => $string >>

Applicable for CMode CLI/ZAPI

Maps to:

CM ZAPI : policy-name

=item C<< "min-blocks-shared" => $string >>

Applicable for CMode CLI,7Mode CLI

Maps to:

7M CLI : -m <value> (Programmatic name : 'minimum-blocks-shared')

=item C<< 'idd' => $string >>

Applicable for CMode CLI/ZAPI, 7Mode CLI

Maps to:

CM ZAPI : enable-idd

7M CLI  : -Q <value> (Programmatic name : 'idd')

=item C<< 'quick-check-fsize' => $string >>

Applicable for CMode/7Mode CLI

Maps to:

7M CLI : -z  <value>(Programmatic name : 'quick-check-fsize')

=item C<< 'optimize' => $string >>

Applicable for CMode CLI

*Optimization, possible value(s) are performance,space-saving

Applicable for CMode CLI

=item C<< "compression-type" => $string >>

Applicable for CMode CLI/ZAPI.

It is a global parameter which takes preference  when it is passed in the parameter NACL_SIS_COMPRESSION_TYPE

CM ZAPI: compression-type

=item C<< "job_component"=> \$scalar >>

Filled in for CMode CLI/ZAPI.

Applicable for infinite volume.

=item C<< 'is-repository' => $boolean >>

B<DEPRECATED>. Use C<is_infinivol> instead.

=item C<< is_infinivol => 0 | 1 >>

(Optional, Applicable for CMode ZAPI)

This boolean option describes whether the volume is an infinite volume or not.
For CMode ZAPI, the "async" form of the ZAPI must be used for the infinite
volume. If this option is not specified, the library will attempt to
determine if an infinite volume is being operated on (this could involve
querying the filer).

=item C<< is_infinivol_ref => \$scalar >>

(Optional, SCALARREF)

If the method was invoked through CMode ZAPI, the variable gets updated
describing whether it operated on an infinivol (used the "async" ZAPI) or not
(used the normal ZAPI).

=item C<< 'extended_query' => $string >>

(Optional)

Applicable for CMode CLI.

=item C<< '*' >>

This method takes all other options supported by
"volume efficiency modify" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=item C<NACL::C::Exceptions::VolumeEfficiency::VserverDoesNotExist>

This type of exception gets thrown if an attempt is made to perform
volume efficiency on a vserver which does not exists.

=item C<NACL::C::Exceptions::VolumeEfficiency::RestrictedVolume>

This type of exception gets thrown if an attempt is made to perform
volume efficiency on a volume which is in restricted state.

=item C<NACL::C::Exceptions::VolumeEfficiency::OfflineVolume>

This type of exception gets thrown if an attempt is made to perform
volume efficiency off on a volume which is offline.

=back

=back

=cut

sub modify {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            job_component => {type => SCALARREF, optional => 1},
            _optional_scalars(
                qw( inline background-compression schedule compression verify-trigger-rate
                    inline-compression policy min-blocks-shared

                    idd

                    quick-check-fsize

                    volume path

                    optimize is-repository
                    )
            ),
        },
        cli_cmd => 'volume efficiency modify',
    );
    my $param_global     = param_global();
    my $compression_type = $param_global->get('NACL_SIS_COMPRESSION_TYPE');
    if ($compression_type && defined $opts{compression}) {

 # If compression is being passed as true , then only setting compression-type
        if ($opts{compression} =~ /true/) {
            $Log->comment("Setting 'compression-type' to '$compression_type' "
                    . "because parameter 'NACL_SIS_COMPRESSION_TYPE' was provided"
            );
            $opts{'compression-type'} = $compression_type;
        }
    }

    my $ret;
    try {
        $ret = $pkg_or_obj->call_on_apiset(
            %opts,
            choices => [
                {   method    => '_modify_cmode_cli',
                    interface => 'CLI',
                    set       => 'CMode',
                },
                {   method    => '_modify_cmode_zapi',
                    interface => 'ZAPI',
                    set       => 'CMode',
                    check     => '_check_modify_cmode_zapi',
                },
                {   method    => '_modify_7mode_cli',
                    interface => 'CLI',
                    set       => '7Mode',
                },
                {   method    => '_modify_7mode_zapi',
                    interface => 'ZAPI',
                    set       => '7Mode',
                },
            ],
            frontend => 'NACL::C::VolumeEfficiency::modify',
        );
    } ## end try
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception_object = $_[0];
        foreach my $name (
            qw(RestrictedVolume OfflineVolume  VolumeDoesNotExist VserverDoesNotExist )
            )
        {
            my $full_name = "NACL::C::Exceptions::VolumeEfficiency::$name";
            $full_name->detect_convert_and_throw(
                exception => $exception_object);
        }
        $Log->exit() if $may_exit;
        $exception_object->throw();
    };

    $Log->exit() if $may_exit;
    return $ret;
} ## end sub modify

sub _modify_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    my $vserver = $opts{vserver};

    $pkg_or_obj->_base_backend_cmode_cli(
        %opts,
        api                 => 'volume_efficiency_modify',
        vserver_operated_on => \$vserver,
    );

    $Log->exit() if $may_exit;
} ## end sub _modify_cmode_cli

sub _modify_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_backend_cmode_zapi_infinivol_builder(
        @args,
        api  => 'sis-set-config',
        copy => $pkg_or_obj->_modify_zapi_copy(),
        map  => $pkg_or_obj->_modify_zapi_map(),
    );

    $Log->exit() if $may_exit;
}

sub _modify_zapi_copy {
    return [qw(schedule quick-check-fsize compression-type)];
}

sub _modify_zapi_map {
    return {
        'policy'             => 'policy-name',
        'compression'        => 'enable-compression',
        'inline-compression' => 'enable-inline-compression',
        'idd'                => 'enable-idd',
        'inline-dedupe'      => 'enable-inline-dedupe',
        'data-compaction'    => 'enable-data-compaction'

    };
}

sub _check_modify_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;

    if (exists $opts{'min-blocks-shared'}) {
      $Log->exit() if $may_exit;
      NACL::Exceptions::InvalidChoice->throw('Could not use the ZAPI '
        . "implementation of vol efficiency modify because the CLI-specific "
        . "argument 'min-blocks-shared' was provided"
      );
    }

    ## check for infinivol
    $pkg_or_obj->_infinivol_vol_vs_check(%opts);
}

sub _modify_7mode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $copy = $pkg_or_obj->_modify_zapi_copy();
    push @$copy, 'path';
    my $map = $pkg_or_obj->_modify_zapi_map();

    my %api_opts;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%api_opts,
        copy   => $copy,
        map    => $map
    );
    $opts{apiset}->sis_set_config(%api_opts);

    $Log->exit() if $may_exit;
}

sub _modify_7mode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @arguments) = @_;
    my %opts = validate_with(
        params      => \@arguments,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );

    my $apiset = $opts{apiset};

    my $allowed_7mode_options
        = [qw( schedule path inline-compression idd quick-check-fsize )];
    my $args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => $allowed_7mode_options,
        map    => {
            'min-blocks-shared'      => 'minimum-blocks-shared',
            'background-compression' => 'enable-compression',
            'compression'            => 'enable-compression',
        }
    );
    $apiset->sis_config(%{$args});
    $Log->exit() if $may_exit;
} ## end sub _modify_7mode_cli

=head2 undo

    NACL::C::VolumeEfficiency->undo(
                                command_interface => $ci,
                                'volume' => $volume,
                                'vserver' => $vserver,
                                ...);

                                 or

    $VolumeEfficiency_obj->undo(...);
                                                                 or
     NACL::C::VolumeEfficiency->undo(
                                command_interface => $ci,
                                    path      => $path,
                                    'vserver' => $vserver,
                                ...);

(Class or instance method)
This method is used to undo efficiency on a volume.

Uses a CMode CLI/ZAPI or 7Mode CLI APISet.

Invokes 'sis undo' command for 7Mode CLI.

Invokes sis-undo in CMode ZAPI

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'volume' => $volume >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI.

=item C<< 'vserver' => $vserver >>

(Required for class method, Not Applicable for instance method)

Applicable only for CMode CLI.

=item C<< 'path' => $path >>

(Required if volume is not provided)
Path to the volume

Applicable for CMode CLI/ZAPI.

=item C<< 'dedupe' => $string >>

Applicable for CMode CLI, 7Mode CLI.

Maps to:

7M CLI : -D (Programmatic name: "compression")

=item C<< 'inode' => $string >>

Applicable for CMode CLI, 7Mode CLI.

Maps to:

7M CLI : -i <value> (Programmatic name: "inode")

=item C<< 'force' => $string >>

Applicable for CMode CLI.

=item C<< 'compression' => $string >>

Applicable for CMode CLI, 7Mode CLI.

Maps to:

7M CLI : -C (Programmatic name: "compression")

=item C<< 'log' => $string >>

Applicable for CMode CLI, 7Mode CLI.

Maps to:

7M CLI : -d <value> (Programmatic name: "log")

=item C<< 'undo-type' => $string >>

Applicable for CMode CLI, 7Mode CLI.

Maps to:

7M CLI : -t <value> (Programmatic name: "undo-type")

=item C<< 'method-timeout' => $time_in_seconds >>

(Optional)
The default NATE timeout is 60 seconds. If the command might take more
time to complete, this option can be used to specify a larger timeout value.
The value should be provided in seconds.

=item C<< 'volume' => $string >>

Applicable for CMode CLI.

=item C<< 'vserver' => $string >>

Applicable for CMode CLI.

=item C<< 'path' => $string >>

Applicable for CMode CLI, 7Mode CLI.
Maps to:

7M CLI : path

=item C<< 'extended_query' => $string >>

(Optional)

Applicable for CMode CLI.

=item C<< '*' >>

This method takes all other options supported by
"volume efficiency undo" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=item C<NACL::C::Exceptions::VolumeEfficiency::VserverDoesNotExist>

This type of exception gets thrown if an attempt is made to perform
volume efficiency on a vserver which does not exists.

=item C<NACL::C::Exceptions::VolumeEfficiency::RestrictedVolume>

This type of exception gets thrown if an attempt is made to perform
volume efficiency on a volume which is in restricted state.

=item C<NACL::C::Exceptions::VolumeEfficiency::OfflineVolume>

This type of exception gets thrown if an attempt is made to perform
volume efficiency off on a volume which is offline.

=item C<NACL::C::Exceptions::VolumeEfficiency::OperationNotDisabled>

This exception is thrown when volume efficiency undo is performed before
switching off volume efficiency.

=item C<NACL::C::Exceptions::VolumeEfficiency::VolumeIsDestinationForFileTransfers>


This exception is thrown when we attempt to undo volume efficiency
on a volume while the volume is a destination for file transfer operations

=back

=back

=cut

sub undo {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( dedupe inode force compression log undo-type
                    volume path
                    )
            ),
        },
        cli_cmd => 'volume efficiency undo',
    );
    my $vserver;
    $opts{vserver_operated_on} = \$vserver;
    try {
        $pkg_or_obj->call_on_apiset(
            %opts,
            choices => [
                {   method    => '_undo_cmode_cli',
                    interface => 'CLI',
                    set       => 'CMode',
                },
                {   method    => '_undo_7mode_cli',
                    interface => 'CLI',
                    set       => '7Mode',
                },
                {   method    => '_undo_cmode_zapi',
                    interface => 'ZAPI',
                    set       => 'CMode',
                }
            ],
            frontend => 'NACL::C::VolumeEfficiency::undo',
        );
    } ## end try
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception_object = $_[0];
        foreach my $name (
            qw(RestrictedVolume OfflineVolume VolumeDoesNotExist
            VserverDoesNotExist OperationNotDisabled VolumeIsDestinationForFileTransfers )
            )
        {
            my $full_name = "NACL::C::Exceptions::VolumeEfficiency::$name";
            $full_name->detect_convert_and_throw(
                exception => $exception_object);
        }
        $Log->exit() if $may_exit;
        $exception_object->throw();
    };

    $Log->exit() if $may_exit;
} ## end sub undo

sub _undo_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@args,
        api => 'volume_efficiency_undo');
    $Log->exit() if $may_exit;
} ## end sub _undo_cmode_cli

sub _undo_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my $allowed_cmode_options = [
        qw( dedupe  path inode  compression log undo-type
            volume
            )
    ];
    $pkg_or_obj->_base_backend_cmode_zapi(
        @args,
        api  => 'sis_undo',
        copy => $allowed_cmode_options,
    );
    $Log->exit() if $may_exit;
} ## end sub _undo_cmode_zapi

sub _undo_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = $pkg_or_obj->_validate_backend_opts(@_);

    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my $allowed_7mode_options
        = [qw(compression dedupe inode undo-type log path )];
    my $args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => $allowed_7mode_options,
    );

    %opts = %$args;

    $pkg_or_obj->_handle_boolean_options(
        api_options     => \%opts,
        boolean_options => [qw(log)],
        is_7m_cli       => 1
    );

    $apiset->sis_undo(%opts);
    $Log->exit() if $may_exit;
} ## end sub _undo_7mode_cli

=head2 check

    NACL::C::VolumeEfficiency->check(
                                command_interface => $ci,
                                'volume' => $volume,
                                'vserver' => $vserver,
                                'path' => $path,
                                ...);

                                 or

    $VolumeEfficiency_obj->check(...);

(Class or instance method)
This method is used to scrub efficiency metadata of a volume.

Uses a CMode CLI/ZAPI, 7Mode CLI APISet.

Invokes 'sis check' command for 7Mode CLI.

Invokes sis-check API in CMode ZAPI

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'volume' => $volume >>

(Required for class method, Not Applicable for instance method)
Volume Name

Applicable for CMode CLI/ZAPI.

CM ZAPI:

Since this option is not passed to the apilayer call ,there is no mapping
information provided ,path will be constructed using this
argument and passed to the apilayer call.

7M CLI:

Since this option is not passed to the apilayer call ,there is no mapping
information provided ,path will be constructed using
this argument and passed to the apilayer call.

=item C<< 'vserver' => $vserver >>

(Required for class method, Not Applicable for instance method)
Vserver Name

Applicable for CMode CLI.

=item C<< 'path' => $path >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI, 7Mode CLI

Maps to:

7M CLI : path

=item C<< 'delete-checkpoint' => $string >>

*Delete Checkpoint, possible value(s) are true,false

Applicable for CMode CLI/ZAPI, 7Mode CLI.

Maps to:

7M CLI : delete-checkpoint

=item C<< 'method-timeout' => $time_in_seconds >>

(Optional)
The default NATE timeout is 60 seconds. If the command might take more
time to complete, this option can be used to specify a larger timeout value.
The value should be provided in seconds.

=item C<< 'changelog' => $string >>

Applicable for 7Mode CLI.

=item C<< 'verbose-mode' => $string >>

Applicable for 7Mode CLI.

=item C<< 'extended_query' => $string >>

(Optional)
Applicable for CMode CLI.

=item C<< 'volume' => $string >>

Volume Name possible value(s) are, volume name
Applicable for CMode CLI

=item C<< 'path' => $string >>

Volume Path possible value(s) are, /vol/volume
Applicable for CMode CLI

=item C<< '*' >>

This method takes all other options supported by
"volume efficiency check" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=item C<NACL::C::Exceptions::VolumeEfficiency::VserverDoesNotExist>

This type of exception gets thrown if an attempt is made to perform
volume efficiency on a vserver which does not exists.

=item C<NACL::C::Exceptions::VolumeEfficiency::RestrictedVolume>

This type of exception gets thrown if an attempt is made to perform
volume efficiency on a volume which is in restricted state.

=item C<NACL::C::Exceptions::VolumeEfficiency::OfflineVolume>

This type of exception gets thrown if an attempt is made to perform
volume efficiency off on a volume which is offline.

=back

=back

=cut

sub check {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( delete-checkpoint

                    volume path
                    )
            ),
        },
        cli_cmd => 'volume efficiency check',
    );
    my $vserver;
    $opts{vserver_operated_on} = \$vserver;
    try {
        $pkg_or_obj->call_on_apiset(
            %opts,
            choices => [
                {   method    => '_check_cmode_cli',
                    interface => 'CLI',
                    set       => 'CMode',
                },
                {   method    => "_check_7mode_cli",
                    interface => "CLI",
                    set       => "7Mode",
                },
                {   method    => '_check_cmode_zapi',
                    interface => 'ZAPI',
                    set       => 'CMode',
                },
            ],
            frontend => 'NACL::C::VolumeEfficiency::check',
        );
    } ## end try
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception_object = $_[0];
        foreach my $name (
            qw(RestrictedVolume OfflineVolume  VolumeDoesNotExist VserverDoesNotExist )
            )
        {
            my $full_name = "NACL::C::Exceptions::VolumeEfficiency::$name";
            $full_name->detect_convert_and_throw(
                exception => $exception_object);
        }
        $Log->exit() if $may_exit;
        $exception_object->throw();
    };

    $Log->exit() if $may_exit;
} ## end sub check

sub _check_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@args,
        api => 'volume_efficiency_check');
    $Log->exit() if $may_exit;
} ## end sub _check_cmode_cli

sub _check_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my $allowed_cmode_options = [qw( volume delete-checkpoint path)];
    $pkg_or_obj->_base_backend_cmode_zapi(
        @args,
        api  => 'sis_check',
        copy => $allowed_cmode_options,
    );
    $Log->exit() if $may_exit;
} ## end sub _check_cmode_zapi

sub _check_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );

    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};

    my $allowed_7mode_options
        = [qw(changelog verbose-mode delete-checkpoint path )];
    my $args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => $allowed_7mode_options,
    );
    %opts = %$args;
    $apiset->sis_check(%opts);
    $Log->exit() if $may_exit;
} ## end sub _check_7mode_cli

=head2 node_statistics

    my $node_statistics_obj = NACL::C::VolumeEfficiency->node_statistics(command_interface => $ci,
                          [ filter => { %filter } ,]
              [ requested_fields => [ @req_fields ]);

                    or

    my @node_statistics_objs = NACL::C::VolumeEfficiency->node_statistics(command_interface => $ci,
                          [ filter => { %filter } ,]
              [ requested_fields => [ @req_fields ]);

                    or

   my $node_statistics_obj = $VolumeEfficiency_obj->node_statistics();

(Class or Instance method)

This method is a CS-like fetch method which accepts all the parameters of
L<fetch|lib-NACL-CS-ComponentState-pm/fetch> (requested_fields, filter, allow_empty)
and has all of the other properties of fetch methods (callable in scalar or list
context, throws a NoElementsFound exception if no matching elements are found).
The requested fields and filter takes L<ATTRIBUTES|lib-NACL-CS-VolumeEfficiencyNodeStatistics-pm/ATTRIBUTES>
of NACL::CS::VolumeEfficiencyNodeStatistics as input.
This method returns objects of the type L<NACL::CS::VolumeEfficiencyNodeStatistics|lib-NACL-CS-VolumeEfficiencyNodeStatistics-pm>.

=cut

sub node_statistics {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts = $pkg_or_obj->_handle_obj_cs_like_calls(@args);
    require NACL::CS::VolumeEfficiencyNodeStatistics;
    my @state_objs = NACL::CS::VolumeEfficiencyNodeStatistics->fetch(%opts);
    $Log->exit() if $may_exit;
    return wantarray ? @state_objs : $state_objs[0];
}

=head2 clear_statistics

    NACL::C::VolumeEfficiency->clear_statistics(
                                command_interface => $ci,
                                'vserver' => $vserver,
                                'path' => $path,
                                'volume' => $volume,
                                ...);

                                 or

    $VolumeEfficiency_obj->clear_statistics(...);

(Class or instance method)
This method is used to clear the volume efficiency statistics.
Uses a CMode CLI APISet.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'vserver' => $vserver >>

(Required for class method, Not Applicable for instance method)
Vserver Name
Applicable for CMode CLI

=item C<< 'path' => $path >>

(Required for class method, Not Applicable for instance method)
Volume Path
Applicable for CMode CLI

=item C<< 'volume' => $volume >>

(Required for class method, Not Applicable for instance method)
Volume Name
Applicable for CMode CLI

=item C<< 'z' => $string >>

Clear Efficiency Statistics
Applicable for CMode CLI

=item C<< 'cmd_execution_time' => \$scalar >>

(Optional) Scalar reference , it will be updated with the command execution
time returned by the apiset response object

=item C<< 'method-timeout' => $time_in_seconds >>

(Optional)
The default NATE timeout is 60 seconds. If the command might take more
time to complete, this option can be used to specify a larger timeout value.
The value should be provided in seconds.

=item C<< '*' >>

This method takes all other options supported by "volume efficiency clear statistics" command in ONTAP

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub clear_statistics {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            'cmd_execution_time' => {type => SCALARREF, optional => 1},
            _optional_scalars(
                qw( z
                    )
            ),
        },
        cli_cmd => 'volume efficiency clear-statistics',
    );
    $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_clear_statistics_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::VolumeEfficiency::clear_statistics',
    );

    $Log->exit() if $may_exit;
}

sub _clear_statistics_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;

    $pkg_or_obj->_base_backend_cmode_cli(@args,
        api => 'volume_efficiency_clear_statistics',);

    $Log->exit() if $may_exit;
} ## end sub _clear_statistics_cmode_cli

=head2 promote

    NACL::C::VolumeEfficiency->promote(
                                command_interface => $ci,
                                'vserver' => $vserver,
                                'path' => $path,
                                'volume' => $volume,
                                ...);

                                 or

    $VolumeEfficiency_obj->promote(...);

(Class or instance method)
This method is used to adds a volume to the preferred set of volumes                   for efficiency processing.
Uses CMode CLI .

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'vserver' => $vserver >>

(Required for class method, Not Applicable for instance method)
Vserver Name
Applicable for CMode CLI

=item C<< 'path' => $path >>

(Required for class method, Not Applicable for instance method)
Volume Path
Applicable for CMode CLI

=item C<< 'volume' => $volume >>

(Required for class method, Not Applicable for instance method)
Volume Name
Applicable for CMode CLI

=item C<< '*' >>

This method takes all other options supported by "volume efficiency promote" command in ONTAP

=item C<< other_options >>

This method takes other common options described in L<Common method options|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#Other_options_valid_on_component_methods>

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub promote {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params => \@args,
        api    => 'volume_efficiency_promote',
    );
    $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_promote_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::VolumeEfficiency::promote',
    );

    $Log->exit() if $may_exit;
}

sub _promote_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@opts,
        api => 'volume_efficiency_promote',);
    $Log->exit() if $may_exit;
}

=head2 demote

    NACL::C::VolumeEfficiency->demote(
                                command_interface => $ci,
                                'vserver' => $vserver,
                                'path' => $path,
                                'volume' => $volume,
                                ...);

                                 or

    $VolumeEfficiency_obj->demote(...);

(Class or instance method)
This method is used to removes a volume from the preferred set of                   volumes for efficiency processing.
Uses CMode CLI .

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'vserver' => $vserver >>

(Required for class method, Not Applicable for instance method)
Vserver Name
Applicable for CMode CLI

=item C<< 'path' => $path >>

(Required for class method, Not Applicable for instance method)
Volume Path
Applicable for CMode CLI

=item C<< 'volume' => $volume >>

(Required for class method, Not Applicable for instance method)
Volume Name
Applicable for CMode CLI

=item C<< '*' >>

This method takes all other options supported by "volume efficiency demote" command in ONTAP

=item C<< other_options >>

This method takes other common options described in L<Common method options|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#Other_options_valid_on_component_methods>

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub demote {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params => \@args,
        api    => 'volume_efficiency_demote',
    );
    $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_demote_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::VolumeEfficiency::demote',
    );

    $Log->exit() if $may_exit;
}

sub _demote_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@opts,
        api => 'volume_efficiency_demote',);
    $Log->exit() if $may_exit;
}

=head2 prepare_to_downgrade

    NACL::C::VolumeEfficiency->prepare_to_downgrade(
                                command_interface => $ci,
                                ...);
   
(Class method)
This method is used to identify any incompatable volumes or Snapshot copies before downgrade.
Uses CMode CLI/ZAPI .
Invokes  sis-prepare-to-downgrade  for CMode Zapi non-iter

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>


=item C<< 'disable-feature-set' => $string >>

Data ONTAP Version
Applicable for CMode CLI/ZAPI, non-iter.

=item C<< 'skip-offline-volumes' => $string >>

Skip downgrade on offline volumes possible value(s) are, true,false
Applicable for CMode CLI/ZAPI, non-iter.

=item C<< '*' >>

This method takes all other options supported by "volume efficiency prepare to downgrade" command in ONTAP

=item C<< other_options >>

This method takes other common options described in L<Common method options|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#Other_options_valid_on_component_methods>

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub prepare_to_downgrade {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( disable-feature-set skip-offline-volumes
                    )
            ),
        },
        ignore_primary_keys => 1,
        cli_cmd             => 'volume efficiency prepare-to-downgrade',
    );
    $pkg->call_on_apiset_with_exceptions(
        %opts,
        choices => [
            {   method    => '_prepare_to_downgrade_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
                zapi_type => 'none',
            },
            {   method    => '_prepare_to_downgrade_cmode_zapi_non_iter',
                interface => 'ZAPI',
                set       => 'CMode',
                zapi_type => 'non-iter',
            },
        ],
	nacl_exceptions => [qw(RevertToFailedSnapshotCopied RevertToFailedUndoCompression PromoteDeprioritizedVolume)],
        frontend => 'NACL::C::VolumeEfficiency::prepare_to_downgrade',
    );

    $Log->exit() if $may_exit;
}

sub _prepare_to_downgrade_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, @opts) = @_;
    $pkg->_base_backend_cmode_cli(@opts,
        api => 'volume_efficiency_prepare_to_downgrade',);
    $Log->exit() if $may_exit;
}

sub _prepare_to_downgrade_cmode_zapi_non_iter {
    $Log->enter() if $may_enter;
    my ($pkg, @opts) = @_;
    $pkg->_base_backend_cmode_zapi(@opts, api => 'sis-prepare-to-downgrade',);
    $Log->exit() if $may_exit;
}

__PACKAGE__->_load_event_class();

sub _primary_keys_validate_spec {
    return (
        vserver => {type => SCALAR | OBJECT | UNDEF, optional => 1},
        path    => {type => SCALAR},
    );
} ## end sub _primary_keys_validate_spec

#
#(Overriding the _common_validate_with definition of the
# base)
#VolumeEfficiency subcommand takes path or a volume,
#not both .If a user does not pass path ,the volume
#name is being extracted and added to the input hash
#

sub _common_validate_with {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my $flag       = 0;
    my %opts       = validate_with(
        params => \@_,
        spec   => {
            params              => {type => ARRAYREF},
            additional_spec     => {type => HASHREF, default => {}},
            ignore_primary_keys => {type => SCALAR, default => 0},
        },
        allow_extra => 1
    );
    my %orig_params = @{$opts{params}};
    my %spec        = %{$opts{additional_spec}};
    $spec{volume} ||= {type => SCALAR, optional => 1};

    # Make "is_infinivol" a valid option for all calls
    $spec{is_infinivol} ||= {type => SCALAR, optional => 1};

    # Make "is_infinivol_ref" a valid option for all calls
    $spec{is_infinivol_ref} ||= {type => SCALARREF, optional => 1};

 #
 # This is a hack, since path is the primary key ,if path has not been passed
 # we insert a dummy path ,so that SUPER::_common_validate_with  wont throw an
 # error
 #
    if (   !ref $pkg_or_obj
        && !$orig_params{path}
        && !$opts{ignore_primary_keys})
    {
        $orig_params{path} = "dummy";
        $flag = 1;
    }
    my %params = $pkg_or_obj->SUPER::_common_validate_with(
        %opts,
        params          => [%orig_params],
        additional_spec => \%spec,
    );
    if ($flag) {
        delete $params{path};
    }

    my %options_to_process;
    $pkg_or_obj->_hash_move(
        source => \%params,
        target => \%options_to_process,
        move   => [qw(volume)]
    );

    if (!$opts{ignore_primary_keys}) {
        $pkg_or_obj->_process_options(options => \%options_to_process);
        if ($params{path}
            && ($options_to_process{volume}))
        {
            NATE::Exceptions::Argument->throw(
                "Either of 'volume' or 'path' should have been provided but both were"
            );
        } ## end if ( $params{path} && ...)

        if (!$params{path} && ($options_to_process{volume})) {
            $params{path} = '/vol/' . delete $options_to_process{volume};
        }
    } ## end if ( !$opts{ignore_primary_keys...})
    $Log->exit() if $may_exit;
    return %params;
} ## end sub _common_validate_with

# These fields cannot be added to _common_validate_spec since they end up
# causing validation failures on the CS side, hence this extra spec for
# options that are common to all methods on this Component on the C side.
sub _additional_common_validate_spec {
    my ($pkg_or_obj, @opts) = @_;
    my $spec    = $pkg_or_obj->SUPER::_additional_common_validate_spec(@opts);
    my $iv_spec = $pkg_or_obj
        ->_add_infinivol_params_to_additional_common_validate_spec(@opts);
    $spec = {%$iv_spec, %$spec};
    return $spec;
}

sub init {
    $Log->enter() if $may_enter;
    my $self = shift;
    my %opts = validate_with(
        params      => \@_,
        spec        => {command_interface => {required => 1}},
        allow_extra => 1
    );
    $Log->exit() if $may_exit;
    if (defined($opts{path}) && defined($opts{volume})) {
        NATE::Exceptions::Argument->throw("Specify either path or volume");
    }
    unless ($opts{path} || $opts{volume}) {
        NATE::Exceptions::Argument->throw(
            "Either of 'volume' or 'path' should have been provided but neither were"
        );
    }
    if ( defined $opts{vserver} ) {
         if ( blessed( $opts{vserver} )
            && $opts{vserver}->isa('NACL::C::Vserver') )
         {
          $opts{vserver} = $opts{vserver}->vserver();
         }
    }
    $self->_vserver_context_init(%opts);
    $Log->exit() if $may_exit;
} ## end sub init

sub _extract_volume_from_path {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts = validate_with(
        params => \@args,
        spec   => {path => {type => SCALAR}},
    );
    my $path = $opts{path};
    $path =~ m!/vol/(\w+)!;
    my $volume = $1;
    $Log->exit() if $may_exit;
    return $volume;
}

sub _backend_cmode_zapi_infinivol_builder {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts = validate_with(
        params => \@args,
        spec   => {
            api              => {type => SCALAR},
            copy             => {type => ARRAYREF, default => []},
            is_infinivol_ref => {type => SCALARREF, optional => 1},
        },
        allow_extra => 1
    );
    my $is_infinivol_ref = delete $opts{is_infinivol_ref};

    my $volume = $pkg_or_obj->_extract_volume_from_path(path => $opts{path});
    my $is_infinivol
        = $pkg_or_obj->_determine_is_infinivol(%opts, volume => $volume);
    if ($is_infinivol) {
        $opts{api} .= '-async';
        $opts{'volume-name'} = $volume;
        push @{$opts{copy}}, 'volume-name';
        @{$opts{copy}} = grep { $_ ne 'path' } @{$opts{copy}};
        delete $opts{path} if (defined $opts{path});
    } else {
        push @{$opts{copy}}, 'path';
    }

    $$is_infinivol_ref = $is_infinivol if ($is_infinivol_ref);

    my $vserver;
    $opts{vserver_operated_on} ||= \$vserver;
    $pkg_or_obj->_base_backend_cmode_zapi(%opts);

    $Log->exit() if $may_exit;
}

1;
