#
# Copyright (c) 2014-2016 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary VolumeFile Component Module
## @author dl-nacl-dev@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::C::VolumeFile

=head1 DESCRIPTION

C<NACL::C::VolumeFile> is a derived class of
L<NACL::C::Component::ONTAP|lib-NACL-C-Component-ONTAP-pm>. It represents access
to an ONTAP VolumeFile. A related class is
L<NACL::CS::VolumeFile|lib-NACL-CS-VolumeFile-pm>, which represents
the past state of an ONTAP volume file.
This class makes heavy use of the mixin class C<NACL::C::_Mixins::VolumeFileDirectory>. 

=cut

package NACL::C::VolumeFile;

use strict;
use warnings;

# The original implementation had methods named with "_file". For example
# there were methods "read_file", "delete_file" etc. Considering the component
# is named "VolumeFile", this trailing "_file" is redundant and should
# be removed. Here we're defining alias methods to maintain backwards
# compatibility.
BEGIN {
    no strict 'refs';
    foreach my $method (qw(read delete)) {
        *{"${method}_file"} = *{$method};
    }
    use strict 'refs';
}

use NACL::CS::VolumeFileShowFhmap;
use Params::Validate qw(:all);
use NACL::ComponentUtils qw (_optional_scalars Dumper);
use NACL::TaskUtils qw (_get_client_path_to_file);
use base 'NACL::C::Component::ONTAP';
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();

use NATE::Exceptions::Argument qw(:try);
use NACL::C::Exceptions::VolumeFile::AlreadyExists;
use NACL::C::Exceptions::VolumeFile::DoesNotExist;
use NACL::C::Exceptions::VolumeFile::BlockIsAHole;
use NACL::C::Exceptions::VolumeFile::FileTooLarge;
use NACL::Exceptions::InvalidChoice;

use NACL::CS::VolumeFile;
use NACL::C::VolumeDirectory;
use NACL::C::Volume;
use NACL::C::VolumeFileClone;
use NACL::C::VolumeFileCopy;
use NACL::C::VolumeFileMove;
use NACL::CS::VolumeQtree;

use NACL::C::_Mixins::VolumeFileDirectory qw(
    :all :file_create :file_delete :file_operation
);
use constant ONEMB => 1048576;

=head1 ATTRIBUTES

=head2 command_interface

See L<NACL::C::Component::command_interface|lib-NACL-C-Component-pm/command_interface>

=head2 path

The path to the file

=head2 vserver

(Required for CMode, Not Applicable for 7Mode) The name of the vserver
in which the volume hosting the file is present.

=head1 The concept of "name-ordinal"

(This is applicable only for CMode)

When there are multiple volumes of the same name on the same node, the
dblade cannot list the same name for all of them. The C<name-ordinal>
(this is accessible through NACL::CS::Volume) is used to distinguish between
these volumes.

The first volume created will have a "name-ordinal" value of
"base", while successive volumes of that name will be assigned consecutive
integer "name-ordinal" values. For example, let's say we create 3 volumes on a
node, all named "myvol", each of which have a file called "myfile". In this
case, the respective "name-ordinal", dblade names, and dblade paths are:

First creation:
    name-ordinal: base
    dblade name:  myvol
    dblade path:  /vol/myvol/myfile

Second creation:
    name-ordinal: 1
    dblade name:  myvol(1)
    dblade path:  /vol/myvol(1)/myfile

Third creation:
    name-ordinal: 2
    dblade name:  myvol(2)
    dblade path:  /vol/myvol(2)/myfile

=head2 Handling the name-ordinal

All of the component methods accept an argument called C<handle_name_ordinal>,
whose default value is 1. Its functionality is described below.

=over

=item For Nodescope-CLI

=over

=item 1.

The underlying Nodescope commands require the path to contain the
name-ordinal. The component handles this by auto-determining the
name-ordinal and internally updating the path accordingly. For example,
let's say we want to create a file on a volume "myvol" on vserver "myvs"
whose name-ordinal is 1. If a call was made like:

    NACL::C::VolumeFile->create(
        %other_opts,
        vserver => 'myvs',
        path    => '/vol/myvol/filename'
    );

The component automatically determines the name-ordinal and updates the
path to be "/vol/myvol(1)/filename".

=item 2.

Note that if we don't want the component to determine the name-ordinal,
then C<handle_name_ordinal> should be specified as 0.

    NACL::C::VolumeFile->create(
        %other_opts,
        handle_name_ordinal => 0,
        vserver             => 'myvs',
        path                => '/vol/myvol/filename'
    );

For our example, this call would fail since the component would not
automatically update the path to incorporate the name-ordinal.

=item 3.

If the call is made with the name-ordinal explicitly provided, then the
assumption is that this name-ordinal is accurate and the component does
B<not> attempt to determine the name-ordinal. Example:

    NACL::C::VolumeFile->create(
        %other_opts,
        vserver             => 'myvs',
        path                => '/vol/myvol(1000)/filename'
    );

Here an invalid name-ordinal was provided, so the call would end up
failing. As mentioned above, the component does B<not> attempt to
determine the name-ordinal in this case.

=back

=item For ZAPI

=over

=item 1.

The name-ordinal value is inapplicable for ZAPI. The component ensures that
even if the value of the path provided has the name-ordinal value, it strips
it out before making the ZAPI call. Example:

    NACL::C::VolumeFile->delete(
        vserver             => 'myvs',
        path                => '/vol/myvol(1000)/filename',
        apiset_should       => { interface => 'ZAPI' }
    );

The component strips out the name-ordinal and sends the path as
"/vol/myvol/filename" to the ZAPI call.

=item 2.

If we do not want the name-ordinal to be stripped from the path (say for
negative testing), then C<handle_name_ordinal> should be specified as 0.

    NACL::C::VolumeFile->delete(
        vserver             => 'myvs',
        path                => '/vol/myvol(1000)/filename',
        handle_name_ordinal => 0,
        apiset_should       => { interface => 'ZAPI' }
    );

The above call would fail since it would send the path as is to the
ZAPI call, which would error out.

=back

=back

=head2 So, if the ZAPI path and the dblade path are different, then which value is stored in the object?

The value of the C<path> attribute is always stored B<without> the name-ordinal.

=head2 So, how would we get the dblade path?

The L<dblade_path> method can be used for this.

=cut

# Implementation of dblade_path, get_containing_volume(obj|name) from
# the mix-in

=head1 HELPER METHODS

=head2 dblade_path

(See the concept of "name-ordinal" first)

    my $dblade_file_path = $file_obj->dblade_path();

(Instance method)
Returns the dblade path of the file.

(For 7Mode since there isn't a dblade, so this method simply returns the
value of the C<path> attribute)

The only options accepted by this method are the common options (apiset_must,
apiset_should, method-timeout).

=head2 get_containing_volume_obj

    my $volume_c_obj = $file_c_obj->get_containing_volume_obj();

    my $volume_stask_obj = $file_stask_obj->get_containing_volume_obj();

(Instance method only)
Returns a Volume object for the volume containing this file. If this method
is invoked with a C<NACL::C::VolumeFile> object, then a C<NACL::C::Volume>
object is returned. If this method is invoked with a C<NACL::STask::VolumeFile>
object, then a C<NACL::STask::Volume> object is returned.

=head2 get_containing_volume_name

    my $vol_name = $file_obj->get_containing_volume_name();

Returns the name of the volume containing this file.

=head2 copy_to_logdir

See documentation in L<< NACL::C::_Mixins::VolumeFileDirectory->copy_to_logdir|lib-NACL-C-_Mixins-VolumeFileDirectory-pm/copy_to_logdir >>

=head2 systemshell_path

See documentation in L<< NACL::C::_Mixins::VolumeFileDirectory->systemshell_path|lib-NACL-C-_Mixins-VolumeFileDirectory-pm/systemshell_path >>

=head2 get_containing_directory_path

    my $dir_path = $file_obj->get_containing_directory_path();

Returns the path of the directory containing this file. For example, if
the file path is "/vol/vol0/dir/myfile", then the returned directory path
would be "/vol/vol0/dir".

=head2 get_containing_directory_obj

    my $dir_obj = $file_obj->get_containing_directory_obj();

Returns a L<NACL::C::VolumeDirectory|lib-NACL-C-VolumeDirecotry-pm> object
representing the directory containing this file. For example, if
the file path is "/vol/vol0/dir/myfile", then the returned directory object
would model the directory "/vol/vol0/dir".

=head2 get_local_node_name

    my $node_name = $file_obj->get_local_node_name();

Returns the name of the node hosting the volume file.

=head2 get_local_node_obj

    my $node_obj = $file_obj->get_local_node_obj();

Similar to C<get_local_node_name>, but returns a C<NACL::STask::Node> object
for the node hosting the volume file.

=cut

sub get_containing_directory_obj {
    $Log->enter() if $may_enter;
    my $self           = $_[0];
    my $directory_path = $self->get_containing_directory_path();
    my $dir_obj = NACL::C::VolumeDirectory->new($self->get_pk_hash_with_ci(),
        path => $directory_path);
    $Log->exit() if $may_exit;
    return $dir_obj;
}

=head2 clone_create

    my $clone = $file_obj->clone_create('destination-path' => $clone_path, %other_opts);

Clones the file and returns a new C<NACL::C::VolumeFile> object representing
the clone file. The arguments accepted by
L<< NACL::C::VolumeFileClone->create|lib-NACL-C-VolumeFileClone-pm/create >>
are accepted by this method. The C<source-path> argument is automatically
supplied by extracting the value of the C<path> attribute of the file object.

=cut

sub clone_create {
    $Log->enter() if $may_enter;
    my ($self, %opts) = @_;
    my %obj_attributes = $self->get_pk_hash_with_ci();
    my $path           = delete $obj_attributes{path};
    my $elements_obj   = $self->_extract_volume_file_from_path(path => $path);
    my $vol            = $elements_obj->volume();
    $opts{'source-path'} = $elements_obj->file();
    $opts{volume} = $vol;
    NACL::C::VolumeFileClone->create(%opts, %obj_attributes);
    delete $obj_attributes{'source-path'};
    my $clone_path = $self->_construct_file_path(
        volume => $vol,
        file   => $opts{'destination-path'}
    );
    my $clone = $self->get_package_name()
        ->new(%obj_attributes, path => $clone_path);
    $Log->exit() if $may_exit;
    return $clone;
}

=head2 move_start, copy_start

Documented together since they support similar features and options.

    my $copy_obj = $file_obj->copy_start(
        'destination-paths' => $vs . ':' . $vol_name . '/' . $dest_file,
        %other_opts
    );

Starts a volume file copy operation and returns a
L<NACL::C::VolumeFileCopy|lib-NACL-C-VolumeFileCopy-pm> object.

    my $move_obj = $file_obj->move_start(
        'destination-paths' => $vs . ':' . $vol_name . '/' . $dest_file,
        %other_opts
    );

Starts a volume file copy operation and returns a
L<NACL::C::VolumeFileMove|lib-NACL-C-VolumeFileMove-pm> object.

For both methods, only a single file can be copied/moved (the one represented
by the file object), so only a single destination path should be provided.

The command_interface is extracted from the file object and the
C<source-paths> argument is automatically constructed using attributes
of the file object. All other options accepted by the respective C<start>
method are accepted by this method.

Note that neither method waits for the copy/move to complete. To do so, we
should use the C<wait_for_finish> method implemented on the STask. Example:

    my $move_c_obj = $file->move_start('destination-paths' => $dest);
    my $move_task_obj->cast_component_to_task();
    $move_task_obj->wait_for_finish();

=cut

sub copy_start {
    $Log->enter() if $may_enter;
    my ($self, @args) = @_;
    my $copy_obj = $self->_move_copy_start_helper(@args,
        _package => 'NACL::C::VolumeFileCopy');
    $Log->exit() if $may_exit;
    return $copy_obj;
}

sub move_start {
    $Log->enter() if $may_enter;
    my ($self, @args) = @_;
    my $move_obj = $self->_move_copy_start_helper(@args,
        _package => 'NACL::C::VolumeFileMove');
    $Log->exit() if $may_exit;
    return $move_obj;
}

sub _move_copy_start_helper {
    $Log->enter() if $may_enter;
    my ($self, @args) = @_;
    my %opts = validate_with(
        params      => \@args,
        spec        => {_package => {type => SCALAR}},
        allow_extra => 1
    );
    my $package = delete $opts{_package};

    my %obj_attributes = $self->get_pk_hash_with_ci();
    my $vserver        = delete $obj_attributes{vserver};
    my $path           = delete $obj_attributes{path};
    $path =~ s/^\/vol\///;
    $opts{'source-paths'} = "$vserver:$path";
    my $obj = $package->start(%opts, %obj_attributes);

    $Log->exit() if $may_exit;
    return $obj;
}

sub _form_dest_obj_from_path {
    $Log->enter() if $may_enter;
    my ($self, @args) = @_;
    my %opts = validate_with(
        params => \@args,
        spec   => {'destination-paths' => {type => SCALAR | ARRAYREF}}
    );

    my $dest_path;
    if (ref $opts{'destination-paths'}) {
        $dest_path = $opts{'destination-paths'}->[0];
    } else {
        $dest_path = $opts{'destination-paths'};
    }

    $dest_path =~ /(.*):(.*)\/(.*)/;
    my ($vserver, $volume, $file) = ($1, $2, $3);
    my $path = $self->_construct_file_path(volume => $volume, file => $file);
    my $new_file_obj = $self->get_package_name()->new(
        command_interface => $self->command_interface(),
        vserver           => $self->vserver(),
        path              => $path
    );

    $Log->exit() if $may_exit;
    return $new_file_obj;
}

=head1 METHODS

=head2 new

    my $file_obj = NACL::C::VolumeFile->new(
        command_interface => $ci,
        path              => $file_path,
        vserver           => $vserver_name  # Required for CMode
    );

Creates a new file object. Note that this does not create a new file
on the filer, it only creates a Perl object.

=cut

use Class::MethodMaker [
    scalar => 'path',
    scalar => 'vserver',
];

# Remove the name-ordinal and quotes from the path value and store in the
# object.
sub init {
    $Log->enter() if $may_enter;
    my ($self, %opts) = @_;
    $self->SUPER::init(%opts);
    $self->_vserver_context_init(%opts);
    my $cmode_path
        = $self->_convert_path_to_cmode_format(path => $opts{path});
    $self->path($cmode_path);
    $Log->exit() if $may_exit;
}

=head2 create

    my $file_obj = NACL::C::VolumeFile->create(
        command_interface => $ci,
        path              => $file_path,
        vserver           => $vserver_name,
        ...
    );

    my @file_objs = NACL::C::VolumeFile->create(
        command_interface => $ci,
        path              => $file_path,
        vserver           => $vserver_name,
        filecount         => $count,
        ...
    );

(Class method only)

Creates a file at the specified path and returns a Component object
modeling it.

When the "filecount" argument is passed (let's say its value is N), then N
number of files are created and N number of objects are returned. The files
created will be "$filepath.0", "$filepath.1" .. "$filepath.N-1".

Uses a 7Mode CLI or Nodescope-CLI APISet.

Invokes "mkfile".

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< path => $path >>

(Required)
The path to the file.

=item C<< vserver => $vserver >>

(Required in CMode, ignored in 7Mode)
The vserver through which the volume hosting the file is accessible.

=item C<< filecount => $count >>

When provided, $count number of files are created and $count number of objects
are returned.

=item C<< size => $size >>

The size of the file to create. If provided as an integer without any units,
then the command assumes to it to be in terms of bytes. If a size suffix is
provided, no space should be present between the size and the suffix. The
allowed suffixes are "b/B", "k/K", "m/M", "g/G". Examples of valid values are:
'10', '100k', '2T'.

=item C<< 'deduplicate-zero-data' => 0 | 1 >>

A value of 1 would detect and deduplicate zeroed data (only valid on vols
with dedup enabled.)i

=item C<< 'step-size-in-blocks' => $num_blocks >>

Number of blocks to skip in between each write.

=item C<< 'stream-name' => $stream >>

A stream is an on disk structure containing a sequence of bytes, analogous to
a file. A stream associated with a file is identified by a name, and data may
be read or written to a file or stream.

=item C<< handle_name_ordinal => 0 | 1 >>

(Optional, defaults to 1)
Described at the beginning of this file.

=item C<< fragment-file => 0 | 1 >>

Fragment the new file(s) on disk

=item C<< contiguous_fbns => $num_blocks >>

Number of File Blocks(FBNs). It is only valid with fragment-file (-f) option.

=back

=back

=cut

sub create {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    my $objs = $pkg->_create_helper(@args, _is_vol_file => 1);
    $Log->exit() if $may_exit;
    return wantarray ? @$objs : $objs->[0];
}

=head2 create_stream

    $file_obj->create_stream('stream-name' => $stream_name, size => $size);

Create a stream on the file.

=cut

sub create_stream {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    $pkg_or_obj->create(@opts);
    $Log->exit() if $may_exit;
}

=head2 read

The method C<read_file> is an alias of this method. C<read_file> was the
older name for this method, so that has been aliased to this method to
retain backwards compatibility.

    # Class method
    my $contents = NACL::C::VolumeFile->read(
        command_interface => $ci,
        path              => $path,
        %other_opts
    );

               or

    # Instance method
    my $contents = $vol_file_obj->read(%other_opts);

(Class or instance method)

Read and return the contents of the file.

While this method is supported for both CLI and ZAPI, if any of the
ZAPI-specific options (described below) are provided, then this method
will run through ZAPI.

Uses a 7Mode CLI, Nodescope-CLI, CMode-ZAPI or 7Mode-ZAPI APIset.

Defaults to using ZAPI.

Invokes "rdfile" command for 7Mode/Nodescope CLI.

Invokes "file-read-file" for CMode/7Mode ZAPI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< path => $path >>

(Required for Class Method, Not Applicable for Instance Method)
The path to the file.

=item C<< vserver => $vserver >>

(Required for Class Method in CMode, Not Applicable for Instance Method, ignored in 7Mode)
The vserver through which the volume hosting the file is accessible.

=item C<< length => $length >>

(ZAPI specific option)
The number of bytes to read from the file. Value is allowed to be in the range
0..1048576.

=item C<< offset => $offset >>

(ZAPI specific option)
Offset into file to start reading from. Value is allowed to be in the range
0 .. (2^63)-1

=item C<< 'stream-name' => $stream >>

(Specific to CMode ZAPI)
A stream is an on disk structure which contains a sequence of bytes, analogous
to a file. A stream is always associated with exactly one file. Characters
that are valid for a file name are also valid for the stream-name. e.g.,
"meta", "resource1". Use this option if the data should be read from the
named stream associated with file specified by <path> Default is empty stream
name, in which case the data is read from file specified by <path>.

=item C<< num_bytes_read => \$num >>

(Reference to a scalar variable)
If provided, the scalar variable gets updated with the number of bytes that
were read.

    my $num;
    my $contents = NACL::C::VolumeFile->read_file(..., num_bytes_read => \$num);
    $Log->comment("Number of bytes read: $num");

=item C<< 'nodescope-node-name' => $node_name >>

(Applicable only for Nodescope-CLI)
Explicitly specify the node on which to run the Nodescope command (the APISet
defaults to running on node "local")

=item C<< determine_appropriate_nodescope_node => 0 | 1 >>

(Optional, defaults to 0, applicable only for Nodescope-CLI)
If set to 1, the method determines which node the volume hosting the file
is present on and runs the Nodescope CLI command on that particular node.

=item C<< handle_name_ordinal => 0 | 1 >>

(Optional, defaults to 1)
Described at the beginning of this file.

=item C<< file => $file >>

B<DEPRECATED>. The argument C<path> should be used instead, support for
this argument has been maintained purely for backwards compatibility. Scripts
using this argument should be updated to use C<path> instead.

=back

=back

=over

=item Exceptions

=over

=item C<NACL::C::Exceptions::VolumeFile::DoesNotExist>

This type of exception gets thrown if an attempt read the file
which does not exists.

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub read {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(qw(length offset stream-name)),
            $pkg_or_obj->_nodescope_options_validate_spec(),
            num_bytes_read => {type => SCALARREF, optional => 1},
        },
        _method => 'read'
    );

    my $contents;
    try {
         $contents = $pkg_or_obj->call_on_apiset(
            %opts,
            choices => [
                {   method    => '_read_zapi',
                    interface => 'ZAPI',
                    set       => 'CMode|7Mode',
                    check     => '_check_for_vserver'
                },
                {   method    => '_read_7mode_cli',
                    interface => 'CLI',
                    set       => '7Mode|Nodescope',
                    check     => '_check_read',
                },
            ],
            frontend => 'NACL::C::VolumeFile::read'
        );
    } catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = $_[0];

        NACL::C::Exceptions::VolumeFile::DoesNotExist
            ->detect_convert_and_throw(exception => $exception);

        $Log->exit() if $may_exit;
        $exception->throw();
    };

    $Log->exit() if $may_exit;
    return $contents;

} ## end sub read_file

sub _read_7mode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts              = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset            = $opts{apiset};
    my $command_interface = $opts{command_interface};

    my @allowed_options
        = $apiset->get_allowed_options_for_command(command => 'rdfile');
    push @allowed_options, 'nodescope-node-name';

    my %api_args;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%api_args,
        copy   => \@allowed_options,
        map    => {path => 'file'}
    );

    my $response = $apiset->rdfile(%api_args);
    my $output   = $response->get_raw_output();

    if ($opts{num_bytes_read}) {
        ${$opts{num_bytes_read}} = length $output;
    }

    $Log->exit() if $may_exit;
    return $output;
}

sub _read_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts              = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset            = $opts{apiset};
    my $command_interface = $opts{command_interface};

    my $is_cmode        = $command_interface->is_cmode();
    my @allowed_options = (qw(path length offset));
    if ($is_cmode) {
        push @allowed_options, 'stream-name';
    }

    unless (defined $opts{length}) {

        # Maximum allowed value is 1 MB
        $opts{length} = ONEMB;
    }

    unless (defined $opts{offset}) {

        # offset = 0 means "start reading from the beginning of the file"
        $opts{offset} = 0;
    }

    my %api_args;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%api_args,
        copy   => \@allowed_options,
    );

    if ($is_cmode) {
        $pkg_or_obj->_handle_zapi_vserver_context(
            api_opts          => \%api_args,
            vserver           => $opts{vserver},
            command_interface => $command_interface,
            frontend_method   => 'read',
        );
    }

    my ($total_length, $full_output);

    # The ZAPI only allows reading 1 MB at a time, so let's keep reading in
    # chunks of 1 MB until we reach the end of the file
    # Note: CMode currently allows reading more than 1MB at a time, but let's
    # not rely it, it's a bug. See burt 630271.
    while (1) {
        my $response    = $apiset->file_read_file(%api_args);
        my $row         = $response->get_parsed_output()->[0];
        my $zapi_output = $row->{data};
        my $length      = $row->{length};

        # Doesn't complain even if offset is beyond EOF. See burt 632388.
        last unless $length;

        # The data returned is in terms of hex characters.
        # We need to convert this to a normal string.
        # Every two characters in the returned data is one hex value.
        # Hence, split the data into pairs of characters.
        # Then convert the hex value into a character. Invoking "hex"
        # on a hexadecimal value returns its value in decimal, invoking
        # "chr" on it returns the characters equivalent.
        #
        # Here's an example:
        #   data => '64617461'
        # Step 1: Get each hexadecimal value
        #   @hexadecimal_array = ( '64', '61', '74', '61' )
        # Step 2: Convert each of character (in turn done in two steps:
        # Step 2.1: Hex -> Decimal using hex()
        # Step 2.2: Decimal -> Character using chr()
        #   @chars = ( 'd', 'a', 't', 'a' )
        # Step 3: Join together to form an entire string
        #   $string = 'data'
        #
        # Surely there's a better way to do this?
        # TODO: Figure out what that "better way" is :)
        my @single_hex_chars = split '|', $zapi_output;
        my @hexadecimal_array;
        for (my $i = 0; $i < @single_hex_chars; $i += 2) {
            my $hex_char = $single_hex_chars[$i] . $single_hex_chars[$i + 1];
            push @hexadecimal_array, $hex_char;
        }
        my @char = map { chr hex $_ } @hexadecimal_array;
        my $output = join '', @char;

        $total_length += $length;
        $full_output .= $output;

        if ($length < ONEMB) {

           # Read less than 1MB, so we must have read till the end of the file
            last;
        } else {

          # Read 1MB, there could be more, let's move ahead 1MB and read again
            $api_args{offset} += ONEMB;
        }
    }

    if ($opts{num_bytes_read}) {
        ${$opts{num_bytes_read}} = $total_length;
    }

    $Log->exit() if $may_exit;
    return $full_output;
}

sub _check_read {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    my @zapi_options;
    foreach my $opt (qw(length offset)) {
        if (exists $opts{$opt}) {
            push @zapi_options, "'$opt'";
        }
    }

    if ($opts{command_interface}->is_cmode()) {

        # "stream-name" is applicable on for CMode
        if (exists $opts{'stream-name'}) {
            push @zapi_options, "'stream-name'";
        }
    }

    if (@zapi_options) {
        $Log->exit() if $may_exit;
        NACL::Exceptions::InvalidChoice->throw('Cannot use the CLI '
                . 'implementation because the following ZAPI-specific options '
                . "were passed:\n"
                . join(' ', @zapi_options)
                . "\n");
    }

    $Log->exit() if $may_exit;
}

=head2 delete

The method C<delete_file> is an alias of this method. C<delete_file> was the
older name for this method, so that has been aliased to this method to
retain backwards compatibility.

    NACL::C::VolumeFile->delete(
        command_interface => $ci,
        path              => $path,
        vserver           => $vserver
    );

                  or

    $vol_file_obj->delete();

                  or

    # Deleting multiple files at once
    NACL::C::VolumeFile->delete(
        command_interface => $ci,
        paths             => \@paths,
        vserver           => $vserver
    );

(Class or instance method)

Delete a file.

Uses a CMode-ZAPI, 7Mode-ZAPI, 7Mode CLI or Nodescope-CLI APIset.

Defaults to using ZAPI.

Invokes "rm" in 7Mode and Nodescope CLI.

Invokes "file-delete-file" in CMode and 7Mode ZAPI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for Class Method, Not Applicable for Instance Method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< path => $file_path >>

Path to the file to be deleted.

For class method calls, either this or C<paths> is necessary, for instance
method calls, this argument is not applicable.

=item C<< paths => \@paths >>

(Optional, Applicable only for Class Method)
An array of file paths to be deleted.

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'nodescope-node-name' => $node_name >>

(Applicable only for Nodescope-CLI)
Explicitly specify the node on which to run the Nodescope command (the APISet
defaults to running on node "local")

=item C<< determine_appropriate_nodescope_node => 0 | 1 >>

(Optional, defaults to 0, applicable only for Nodescope-CLI)
If set to 1, the method determines which node the volume hosting the file
is present on and runs the Nodescope CLI command on that particular node.

=item C<< handle_name_ordinal => 0 | 1 >>

(Optional, defaults to 1)
Described at the beginning of this file.

=item C<< file => $file >>

B<DEPRECATED>. The argument C<path> should be used instead, support for
this argument has been maintained purely for backwards compatibility. Scripts
using this argument should be updated to use C<path> instead.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub delete {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    $pkg_or_obj->_delete_front_end_helper( @args, _is_vol_file => 1 );
    $Log->exit() if $may_exit;
}

sub _delete_7mode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts   = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = $opts{apiset};
    my $ci     = $opts{command_interface};
    my @paths  = @{$opts{paths}};

    my %api_args;
    if ($ci->is_cmode()) {

        # Nodescope
        $pkg_or_obj->_hash_copy_defined(
            source => \%opts,
            target => \%api_args,
            copy   => [qw(nodescope-node-name)]
        );

        if (@paths == 1) {

            # The common case: Only a single path provided
            my $hash = $pkg_or_obj->_get_node_and_name_ordinal(%opts,
                path => $paths[0]);
            $api_args{paths} = $hash->{path} || $paths[0];
            if (  !defined $api_args{'nodescope-node-name'}
                && defined $hash->{'nodescope-node-name'})
            {
                $api_args{'nodescope-node-name'}
                    = $hash->{'nodescope-node-name'};
            }
        } else {

            # More than a single path provided
            # First let's determine if either "name-ordinal" or "node"
            # need to be queried for
            my @requested_fields;
            if ($opts{handle_name_ordinal}) {
                push @requested_fields, 'name-ordinal';
            }

            if ($opts{determine_appropriate_nodescope_node}
                && !$opts{'nodescope-node-name'})
            {
                push @requested_fields, 'node';
            }

            if (@requested_fields) {

                # Either (or both of) "node" or "name-ordinal" need to be
                # queried for.
                # Iterate through the paths and get the details (volume +
                # file + contains_name_ordinal) of each. While iterating,
                # keep a track of the volumes seen, so that later when
                # we call Volume->fetch, we filter for only these volumes.
                my %vol_hash;
                foreach my $path (@paths) {
                    my $obj = $pkg_or_obj->_extract_volume_file_from_path(
                        path => $path);
                    my $volume = $obj->volume();
                    if (exists $vol_hash{$volume}) {
                        push @{$vol_hash{$volume}{objs}}, $obj;
                    } else {
                        $vol_hash{$volume} = {objs => [$obj]};
                    }
                }

                my %filter = (vserver => $opts{vserver});
                $filter{volume} = join '|', keys %vol_hash;

                my @vol_cs = NACL::CS::Volume->fetch(
                    command_interface => $ci,
                    filter            => \%filter,
                    requested_fields  => \@requested_fields
                );

                foreach my $cs (@vol_cs) {

                    # Invoke rm for all of the paths belonging to a volume
                    # These are all guaranteed to be on the same node
                    my $objs = $vol_hash{$cs->volume()}{objs};
                    my @paths_for_volume;
                    my %local_api_args;
                    foreach my $obj (@$objs) {
                        if ($obj->contains_name_ordinal()) {
                            push @paths_for_volume, $obj->path();
                        } else {
                            my $this_path
                                = $pkg_or_obj->_construct_file_path(
                                volume         => $obj->volume(),
                                file           => $obj->file(),
                                'name-ordinal' => $cs->name_ordinal()
                                );
                            push @paths_for_volume, $this_path;
                        }
                    }
                    my $local_paths = join ' ', @paths_for_volume;
                    $local_api_args{paths} = join ' ', @paths_for_volume;
                    if (exists $api_args{'nodescope-node-name'}) {
                        $local_api_args{'nodescope-node-name'}
                            = $api_args{'nodescope-node-name'};
                    } elsif ($cs->node_isset()) {
                        $local_api_args{'nodescope-node-name'} = $cs->node();
                    }
                    $apiset->rm(%local_api_args);
                }

                $Log->exit() if $may_exit;
                return;
            } else {

                # Neither "node" nor "name-ordinal" need to be queried for,
                # so we can just invoke rm for all paths together
                $api_args{paths} = join ' ', @paths;
            }
        }
    } else {

        # 7Mode
        $api_args{paths} = join ' ', @paths;
    }

    $apiset->rm(%api_args);
    $Log->exit() if $may_exit;
}

sub _delete_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts              = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset            = $opts{apiset};
    my $command_interface = $opts{command_interface};

    my %api_args;
    my $is_cmode = $command_interface->is_cmode();
    if ($is_cmode) {
        $pkg_or_obj->_handle_zapi_vserver_context(
            api_opts          => \%api_args,
            vserver           => $opts{vserver},
            command_interface => $command_interface,
            frontend_method   => 'delete',
        );
    }

    foreach my $path (@{$opts{paths}}) {

        # Do not modify $path since that would modify the $opts{paths}
        # array-ref; instead make a copy and operate on that
        my $new_path = $path;
        if ($is_cmode) {
            $new_path = $pkg_or_obj->_remove_name_ordinal(path => $path);
        }

        # Strip out quotes (ZAPIs do not accept values with quotes)
        $new_path =~ s/"//g;

        $apiset->file_delete_file(%api_args, path => $new_path);
    }

    $Log->exit() if $may_exit;
}

=head2 write_file

    my $file_obj = NACL::C::VolumeFile->write_file(
        command_interface => $ci,
        path              => $path,
        vserver           => $vserver,
        ...
    );

                 or

    $file_obj->write_file(...);

(Class or instance method)

Writes the provided content to the file. If the file did not exist, it
creates the file and writes to it. If invoked as a class method it returns
a new component object.

If the file existed prior to the call, this overwrites the B<entire> file.
(This is in contrast to the C<write> method which does not overwrite
the entire file, but only the individual bytes)

Uses a 7Mode CLI or Nodescope-CLI APIset. See L<write> for the ZAPI
implementation.

Invokes "wrfile" in 7Mode and Nodescope CLI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for Class Method, Not Applicable for Instance Method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< path => $file_path >>

(Required for Class Method, Not Applicable for Instance Method)
Path to the file to be written to.

=item C<< vserver => $vserver >>

(Required for Class Method in CMode, Not Applicable for Instance Method, ignored in 7Mode)
The vserver through which the volume hosting the file is accessible.

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< content => $scalar >>

(Required) Content to be written to the file.

This can be provided in one of two forms:

* As a scalar variable

The scalar variable should hold the content.

    NACL::C::VolumeFile->write_file(..., content => $content);

* As a reference to a scalar variable

This is the suggested mechanism for passing the value. The reason is that
if the content is large, passing it by value would result in this large
value being passed along the stack. Passing it by reference eliminates this
overhead.

    NACL::C::VolumeFile->write_file(..., content => \$content);

Maps to:

CMode/7Mode ZAPI: "data"

=item C<< append => 0 | 1 >>

If set as 1, appends contents to the end of the file. Note that this
argument is mutually exclusive with C<offset>.

Maps to:

7Mode/Nodescope CLI: The -a option of "wrfile"

=item C<< private => 0 | 1 >>

Flag for creating a file in private namespace.

=item C<< inum => $scalar >>

Inode number.

=item C<< 'nodescope-node-name' => $node_name >>

(Applicable only for Nodescope-CLI)
Explicitly specify the node on which to run the Nodescope command (the APISet
defaults to running on node "local")

=item C<< determine_appropriate_nodescope_node => 0 | 1 >>

(Optional, defaults to 0, applicable only for Nodescope-CLI)
If set to 1, the method determines which node the volume hosting the file
is present on and runs the Nodescope CLI command on that particular node.

=item C<< handle_name_ordinal => 0 | 1 >>

(Optional, defaults to 1)
Described at the beginning of this file.

=item C<< file => $file >>

B<DEPRECATED>. The argument C<path> should be used instead, support for
this argument has been maintained purely for backwards compatibility. Scripts
using this argument should be updated to use C<path> instead.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub write_file {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);

    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(qw(append private inum)),
            content => {type => SCALAR | SCALARREF, optional => 1},
            $pkg_or_obj->_nodescope_options_validate_spec(),
        },
        _method => 'write_file'
    );

    if ($opts{command_interface}->is_cmode() && !$opts{append}) {
        $Log->exit() if $may_exit;
        NATE::BaseException->throw('Call to wrfile does not work in '
                . 'Nodescope without setting append to 1. Either set append '
                . 'to 1 in the call or switch to use the "write" method');
    }

   # Always turn $opts{content} into a scalar-ref. This way it's always
   # passed by reference rather than needing to be copied into the stack
   # (there are a couple more calls before it makes it to the back-end method)
    unless (ref $opts{content}) {
        my $content = $opts{content};
        $opts{content} = \$content;
    }

    my $obj = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_write_file_7mode_cli',
                interface => 'CLI',
                set       => '7Mode|Nodescope',
            },
        ],
        frontend => 'NACL::C::VolumeFile::write_file'
    );

    $Log->exit() if $may_exit;
    return $obj;
}

sub _write_file_7mode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts   = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = $opts{apiset};
    my $ci     = $opts{command_interface};

    my @allowed_options
        = (qw(append private inum content nodescope-node-name));

    my %api_args;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%api_args,
        copy   => \@allowed_options,
        map    => {path => 'file'}
    );

    $pkg_or_obj->_handle_boolean_opts(
        all_opts     => \%api_args,
        boolean_opts => [qw(append private)]
    );

    # "content" is guaranteed to be a scalar-reference, so de-reference it
    # to get its value
    $api_args{content} = ${$api_args{content}};

    $apiset->wrfile(%api_args);

    my $obj = $pkg_or_obj->_create_obj_for_write(%opts);

    $Log->exit() if $may_exit;
    return $obj;
}

=head2 write

    my $file_obj = NACL::C::VolumeFile->write(
        command_interface => $ci,
        path              => $path,
        vserver           => $vserver,
        ...
    );

                 or

    $file_obj->write(...);

(Class or instance method)

Writes the provided content to the file. If the file did not exist, it
creates the file and writes to it. If invoked as a class method it returns
a new component object.

If the file existed prior to the call, this overwrites only the individual
bytes, not the entire file.
(This is in contrast to the C<write_file> method which overwrites the
entire file)

Uses a CMode-ZAPI or a 7Mode-ZAPI APISet.

Invokes "file-write-file" in CMode and 7Mode ZAPI. See L<write_file>
for the CLI implementation.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for Class Method, Not Applicable for Instance Method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< path => $file_path >>

(Required for Class Method, Not Applicable for Instance Method)
Path to the file to be written to.

=item C<< vserver => $vserver >>

(Required for Class Method in CMode, Not Applicable for Instance Method, ignored in 7Mode)
The vserver through which the volume hosting the file is accessible.

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< content => $scalar | \$scalar >>

(Required) Content to be written to the file.

This can be provided in one of two forms:

* As a scalar variable

The scalar variable should hold the content.

    NACL::C::VolumeFile->write(..., content => $content);

* As a reference to a scalar variable

This is the suggested mechanism for passing the value. The reason is that
if the content is large, passing it by value would result in this large
value being passed along the stack. Passing it by reference eliminates this
overhead.

    NACL::C::VolumeFile->write(..., content => \$content);

=item C<< allow_empty_content > 0|1 >>

(Optional)

By default, if empty content is provided, then this method logs a warning
(with the assumption being that the user wanted to write some content but
accidentally provided an undefined value in the call). However, if writing
empty content is indeed being done explicitly and is desired, then set this to
1 to disable the warning that would normally have been generated.

=item C<< append => 0 | 1 >>

If set as 1, appends contents to the end of the file. Note that this
argument is mutually exclusive with C<offset>.

=item C<< offset => $offset >>

The offset from which to start writing within the file. This argument
and C<append> are mutually exclusive.

The value is allowed to be in the range -1 .. (2^63)-1

A value of -1 means the content should be appended to the end of the
file. (this has the same effect as sending "append => 1")

=item C<< overwrite => 'true' | 'false' >>

If set to "false", then throws a
L<NACL::C::Exceptions::VolumeFile::AlreadyExists|lib-NACL-C-Exceptions-VolumeFile-AlreadyExists-pm>
exception if the file already existed.

=item C<< 'stream-name' => $stream >>

(CMode specific option)
A stream is an on disk structure which contains a sequence of bytes, analogous
to a file. A stream is always associated with exactly one file. Characters that
are valid for a file name are also valid for the stream-name. e.g., "meta",
"resource1". Use this option if the data should be written to the named stream
associated with file specified by <path>.

=item C<< handle_name_ordinal => 0 | 1 >>

(Optional, defaults to 1)
Described at the beginning of this file.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub write {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);

    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(qw(append offset overwrite stream-name
                allow_empty_content)),
            content => {type => SCALAR | SCALARREF, optional => 1},
        },
        _method => 'write'
    );

    if (exists $opts{append} && exists $opts{offset}) {
        $Log->exit() if $may_exit;
        NATE::Exceptions::Argument->throw("The argument 'append' and "
                . "'offset' are mutually exclusive but both were provided in "
                . "the call to 'write'");
    }

   # Always turn $opts{content} into a scalar-ref. This way it's always
   # passed by reference rather than needing to be copied into the stack
   # (there are a couple more calls before it makes it to the back-end method)
    unless (ref $opts{content}) {
        my $content = $opts{content};
        $opts{content} = \$content;
    }

    my $obj;
    try {
        $obj = $pkg_or_obj->call_on_apiset(
            %opts,
            choices => [
                {   method    => '_write_zapi',
                    interface => 'ZAPI',
                    set       => 'CMode|7Mode',
                }
            ],
            frontend => 'NACL::C::VolumeFile::write'
        );
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = $_[0];

        NACL::C::Exceptions::VolumeFile::AlreadyExists
            ->detect_convert_and_throw(exception => $exception);

        NACL::C::Exceptions::VolumeFile::FileTooLarge
            ->detect_convert_and_throw(exception => $exception, file_args => \%opts);

        $Log->exit() if $may_exit;
        $exception->throw();
    };

    $Log->exit() if $may_exit;
    return $obj;
}

sub _write_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts              = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset            = $opts{apiset};
    my $command_interface = $opts{command_interface};
    my $is_cmode          = $command_interface->is_cmode();

    my %api_args;
    if ($is_cmode) {
        $pkg_or_obj->_handle_zapi_vserver_context(
            api_opts          => \%api_args,
            vserver           => $opts{vserver},
            command_interface => $command_interface,
            frontend_method   => 'write',
        );
    }

    unless (exists $opts{offset}) {
        if (exists $opts{append}) {
            $opts{offset} = -1;
        } else {

            # Overwrite the file
            $opts{offset} = 0;
        }
    }

   # $opts{content} is guaranteed to be a scalar-reference, so de-reference it
   # to get its value
    my $content = ${$opts{content}};

    # We need to convert each decimal character into a hexadecimal character
    # for the ZAPI to work. First step is to get each individual decimal
    # character. The simplest way would be to just use split to do this.
    # Unforunately, I've found that not only does split slow down
    # considerably as the size of the string increases, but for a 40MB
    # file content it was consistently resulting in segmentation faults
    # (regardless of which client ntest was being invoked from). Hence,
    # instead, we'll split into substrings and then split those into
    # individual characters.

    # Allows writing 1MB of hex characters at a time, so that's
    # half an MB of decimal characters. This will be the size of each
    # substring.
    my $onemb  = ONEMB / 2;
    my $offset = 0;
    my $length = length $content;
    my @substrs;
    while ($offset < $length) {
        push @substrs, (substr $content, $offset, $onemb);
        $offset += $onemb;
    }

    my @copy_list = (qw(offset path overwrite));
    push(@copy_list, 'stream-name') if ($is_cmode);
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%api_args,
        copy   => \@copy_list,
    );

    unless (@substrs) {

        if (!$opts{allow_empty_content}) {
            # The ZAPI allows the data to be undef, but it's unlikely
            # the user wanted to write an empty file, so at least warn about it
            $Log->warn('The content value was not passed in the call to '
                . 'NACL::C::VolumeFile->write(), so an empty file will be '
                . 'written. Please check the script to ensure this is '
                . 'desired. If this is desired, then this warning can be '
                . 'disabled by passing allow_empty_content => 1 in the '
                . 'call to ->write().'
            );
        }
        $apiset->file_write_file(%api_args, data => undef);
    }

    # Can only write 1MB at a time, so let's split the content
    # into 1MB chunks and write one chunk per iteration.
    # Note: CMode currently allows writing more than 1MB at a time, but
    # let's not rely on it because it's a bug. See burt 630271.
    #
    # File contents are provided in human-readable characters
    # Need to convert to hex characters
    # Steps:
    # 1. Split into individual characters (using split)
    # For each character:
    # 2.1 Get numeric equivalent of character (using ord())
    # 2.2 Convert to hex by using sprintf
    # 2.3. If the returned value is less than 2 characters, prepend a "0"
    # to make it 2 characters long. (two hex characters represent one byte)
    #
    # 3. Join all these hex characters together into a string
    #
    # For content = "data\n", the returned value is 646174610a
    foreach my $substr (@substrs) {
        my $str = unpack('H*', $substr);
        $apiset->file_write_file(%api_args, data => $str);

        # If we're not appending to the file, then the successive write
        # should start from an offset 1MB further than the current offset.
        $api_args{offset} += $onemb if ($api_args{offset} != -1);
    }

    my $obj = $pkg_or_obj->_create_obj_for_write(%opts);

    $Log->exit() if $may_exit;
    return $obj;
}

sub _create_obj_for_write {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;

    my $obj;
    if (ref $pkg_or_obj) {
        $obj = $pkg_or_obj;
    } else {
        my %new_args;
        my $ci = $opts{command_interface};
        if ($ci->is_cmode()) {
            if (defined $opts{vserver}) {
                $new_args{vserver} = $opts{vserver};
            } else {

                # Older support of VolumeFile was 7M/Nodescope only, so
                # vserver was not accepted then. For these older callers,
                # we should ensure the code still works, so we shouldn't
                # call new(). (new would fail since vserver would be undef)
                # Instead, let's not create an object and let's just return.
                $Log->exit() if $may_exit;
                return;
            }
        }
        $obj = $pkg_or_obj->new(
            command_interface => $ci,
            path              => $opts{path},
            %new_args
        );
    }
    $Log->exit() if $may_exit;
    return $obj;
}

=head2 show_disk_usage

    my @vol_show_disk_usage = NACL::C::VolumeFile->show_disk_usage(...);

                                          or

    my $vol_show_disk_usage = NACL::C::VolumeFile->show_disk_usage(...);

(Class method only)
This method returns the disk usage of the files present in volumes. It returns objects of the
type L<NACL::CS::VolumeFileShowDiskUsage|lib-NACL-CS-VolumeFileShowDiskUsage-pm>.

This method accepts all the parameters accepted by L<fetch|lib-NACL-CS-ComponentState-pm/fetch>
(requested_fields, filter, allow_empty) and has all of the other properties of
fetch methods (callable in scalar or list context, throws a NoElementsFound
exception if no matching elements are found)

=cut

sub show_disk_usage {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my @state_objs = NACL::CS::VolumeFileShowDiskUsage->fetch(@_);
    $Log->exit() if $may_exit;
    return wantarray ? @state_objs : $state_objs[0];
} ## end sub show_disk_usage

=head2 show_fhmap

   my $show_fhamp_obj = NACL::C::VolumeFile->show_fhmap(command_interface => $ci,...);

(Class or Instance method)
This method returns objects of the type L<NACL::CS::VolumeFileShowFhmap|lib-NACL-CS-VolumeFileShowFhmap-pm>.

This method accepts all the parameters accepted by L<fetch|lib-NACL-CS-ComponentState-pm/fetch>
(requested_fields, filter, allow_empty) and has all of the other properties of
fetch methods (callable in scalar or list context, throws a NoElementsFound
exception if no matching elements are found)

=cut

sub show_fhmap {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    my %opts = $pkg->_handle_obj_cs_like_calls(@args);
    require NACL::CS::VolumeFileShowFhmap;
    my @state_objs = NACL::CS::VolumeFileShowFhmap->fetch(%opts);
    $Log->exit() if $may_exit;
    return wantarray ? @state_objs : $state_objs[0];
}

=head2 reservation

    # Class method, setting a value
    NACL::C::VolumeFile->reservation(
        command_interface => $ci,
        vserver           => $vs,
        path              => $file_path,
        'is-enabled'      => 'enable'
    );

    # Instance method, setting a value
    $file_obj->reservation('is-enabled' => 'enable');

    # Class method, getting the value of the reservation status
    my $status = NACL::C::VolumeFile->reservation(
        command_interface => $ci,
        vserver           => $vs,
        path              => $file_path,
    );

    # Instance method, getting the value of the reservation status
    my $status = $file_obj->reservation();

(Class method)
This method is used to get/set the space reservation info for the named file.

When invoked in get context (see above), the status is returned. The value will
be either "on" or "off".

Uses a CMode CLI/ZAPI, 7Mode CLI APISet.

Note that CMode ZAPI can only set the values of the "is-fill-enabled"
and "is-overwrite-enabled" fields. If an attempt is made to set the value
of the "is-enabled" field or to get the value, then ZAPI is inapplicable
(the implementation will fall back to using CLI).

Invokes "file-set-space-reservation-info" API for CMode ZAPI.

Invokes "file reservation" command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'is-fill-enabled' => $string >>

Applicable for CMode CLI/ZAPI. Valid values are "true" and "false".

Maps to:

CM ZAPI: is-fill-enabled

=item C<< 'vserver' => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: ZAPI will be invoked in the context of the vserver specified to this
         option.

=item C<< 'path' => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: path

=item C<< 'is-enabled' => $string >>

Applicable for CMode CLI. Valid values are "enable" and "disable".

=item C<< 'is-overwrite-enabled' => $string >>

Applicable for CMode CLI/ZAPI. Valid values are "true" and "false".

Maps to:

CM ZAPI: is-overwrite-enabled

=item C<< 'method-timeout' => $time_in_seconds >>

(Optional)
The default NATE timeout is 60 seconds. If the command might take more
time to complete, this option can be used to specify a larger timeout value.
The value should be provided in seconds.

=item C<< "extended_query" => $string >>

(Optional)

Applicable for CMode CLI.

=item C<< '*' >>

This method takes all other options supported by
"volume file reservation" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub reservation {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( is-fill-enabled is-enabled is-overwrite-enabled
                    )
            ),
        },
        api => 'volume_file_reservation',
    );
    my $status = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_reservation_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_reservation_7mode_cli',
                interface => 'CLI',
                set       => '7Mode',
            },
            {   method    => '_reservation_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
                check     => '_reservation_check'
            },
        ],
        frontend => 'NACL::C::VolumeFile::reservation',
    );

    $Log->exit() if $may_exit;
    return $status;
} ## end sub reservation

sub _reservation_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my $response;
    $pkg_or_obj->_base_backend_cmode_cli(
        @args,
        api                 => 'volume_file_reservation',
        apiset_response_obj => \$response
    );
    my $status = $response->get_parsed_output()->[0]{status};
    $Log->exit() if $may_exit;
    return $status;
} ## end sub _reservation_cmode_cli

sub _reservation_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);

    my $command_interface = $opts{command_interface};
    my $apiset            = $opts{apiset};
    my $vserver           = $opts{vserver};

    my $allowed_cmode_options = [
        qw( is-fill-enabled  path is-overwrite-enabled
            )
    ];

    my %cmode_args;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => $allowed_cmode_options,
        target => \%cmode_args,
    );
    $pkg_or_obj->_handle_zapi_vserver_context(
        api_opts          => \%cmode_args,
        vserver           => $vserver,
        command_interface => $command_interface,
    );
    $apiset->file_set_space_reservation_info(%cmode_args);
    $Log->exit() if $may_exit;
} ## end sub _reservation_cmode_zapi

sub _reservation_7mode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts   = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = $opts{apiset};

    my %final_7mode_args;
    if (defined $opts{'is-enabled'}) {
        my $is_enabled = $opts{'is-enabled'};
        if ($is_enabled eq 'enable') {
            $final_7mode_args{enable} = 1;
        } elsif ($is_enabled eq 'disable') {
            $final_7mode_args{disable} = 1;
        } else {
            $Log->exit() if $may_exit;
            NATE::Exceptions::Argument->throw("The value for 'is-enabled' "
                    . "should be either 'enable' or 'disable' but was provided "
                    . "as '$is_enabled'");
        }
    }

    $final_7mode_args{path} = $opts{'path'};

    my $response = $apiset->file_reservation(%final_7mode_args);
    my $status   = $response->get_parsed_output()->[0]{status};

    $Log->exit() if $may_exit;
    return $status;
} ## end sub _reservation_7mode_cli

# ZAPI does not work for:
# 1) get context
# 2) For setting 'is-enabled'
# Check if either of these is the case and if so, disable ZAPI
sub _reservation_check {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;

    my $str1 = "The ZAPI 'file-set-space-reservation-info' cannot be used";
    my $str2 = 'so the Component ZAPI implementation cannot be be used';

    if (exists $opts{'is-enabled'}) {
        $Log->exit() if $may_exit;
        NACL::Exceptions::InvalidChoice->throw(
            "$str1 to set the value of 'is-enabled', $str2\n");
    } else {

        # Figure out the options that would go through to the API call.
        # This is done by removing common back-end options (as described
        # by "_backend_validate_spec"), "_method" (sent to the "check"
        # method by call_on_apiset) and "handle_name_ordinal" (this is
        # specific to this package)
        # If there are no options (other than the primary keys) to be sent to
        # the API, then this is get context, in which case ZAPI cannot be used
        my $backend_validate_spec = $pkg_or_obj->_backend_validate_spec();
        my @backend_keys          = keys %$backend_validate_spec;

        push @backend_keys, $pkg_or_obj->get_primary_keys();

      # "_method" represents the front-end method invoked. This is passed only
      # to the check method, not to back-end methods, so will not be available
      # in backend_validate_spec.
        push @backend_keys, '_method';

        # Specific to this package
        push @backend_keys, 'handle_name_ordinal';

        $pkg_or_obj->_hash_move(
            source => \%opts,
            target => {},
            move   => \@backend_keys
        );

        unless (keys %opts) {
            $Log->exit() if $may_exit;
            NACL::Exceptions::InvalidChoice->throw(
                "$str1 in get context, $str2\n");
        }
    }

    $Log->exit() if $may_exit;
}

=head2 modify

    NACL::C::VolumeFile->modify(
                                command_interface => $ci,
                                ...);
                      or

    $volume_file_object->modify(...);

(Class or instance  method)
This method is used to manage the association of a QoS policy group with a file.

Uses a CMode CLI or CMode ZAPI APISet.
Invokes 'file-assign-qos' API for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'volume' => $string >>

Volume name

Applicable for CMode CLI/ZAPI
Maps to:
   CM ZAPI:
       <file-assign-qos>
           <volume>$value</volume>
       </file-assign-qos>

=item C<< 'vserver' => $string >>

Vserver managing volume

Applicable for CMode CLI/ZAPI
Maps to:
   CM ZAPI:
       <file-assign-qos>
           <vserver>$value</vserver>
       </file-assign-qos>

=item C<< 'qos-policy-group' => $string >>

QoS Policy Group Name

Applicable for CMode CLI/ZAPI
Maps to:
   CM ZAPI:
       <file-assign-qos>
           <qos-policy-group-name>$value</qos-policy-group-name>
       </file-assign-qos>

=item C<< 'file' => $string >>

File path

Applicable for CMode CLI/ZAPI
Maps to:
   CM ZAPI:
       <file-assign-qos>
           <qos-policy-group-name>$value</qos-policy-group-name>
       </file-assign-qos>

=item C<< 'method-timeout' => $time_in_seconds >>

(Optional)
The default NATE timeout is 60 seconds. If the command might take more
time to complete, this option can be used to specify a larger timeout value.
The value should be provided in seconds.

=item C<< 'external-cache' => $string >>

External Cache Policy Name

Applicable for CMode CLI.

=item C<< "extended_query" => $string >>

 (Optional)
Applicable only for CMode CLI

 =item C<< 'caching-policy' => $string >>

Caching Policy Name
Applicable for CMode CLI

=item C<< 'qos-policy-group-uuid' => $string >>

QoS Policy Group UUID
Applicable for CMode CLI

=item C<< '*' >>

This method takes all other options supported by
"volume file modify" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub modify {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    # Don't use NACL::C::VolumeFile's _common_validate_with since
    # that converts file -> path, but this is not applicable here.
    # the UI of only this command is different
    if (ref $pkg_or_obj) {
        my $elements_obj =   $pkg_or_obj->_extract_volume_file_from_path( path => $pkg_or_obj->path() );
        push @args,('volume', $elements_obj->volume(),'file',$elements_obj->file());
    }
    my %opts = $pkg_or_obj->SUPER::_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( volume vserver qos-policy-group file
                    external-cache

                    caching-policy

                    qos-policy-group-uuid

                    path
                    )
            ),
        },
        ignore_primary_keys => 1,
        cli_cmd             => 'volume file modify',
    );

    # Similar reasoning to above: don't use this package's call_on_apiset, but
    # just use the base's one instead.
    $pkg_or_obj->SUPER::call_on_apiset(
        %opts,
        choices => [
            {   method    => '_modify_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_modify_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::VolumeFile::modify',
    );

    $Log->exit() if $may_exit;
} ## end sub modify

sub _modify_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@args,  api => 'volume_file_modify');
    $Log->exit() if $may_exit;
} ## end sub _modify_cmode_cli

sub _modify_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my $allowed_cmode_options = [
            qw( volume file vserver
            )
    ];
    $pkg_or_obj->_base_backend_cmode_zapi(@args,
            api    => 'file_assign_qos',
            copy   => $allowed_cmode_options,
            map    => {'qos-policy-group' => 'qos-policy-group-name'},
    );
    $Log->exit() if $may_exit;
}    ##end sub _modify_cmode_zapi

=head2 show_filehandle

    my $show_filehandle_obj = NACL::C::VolumeFile->show_filehandle(command_interface => $ci,
                          [ filter => { %filter } ,]
              [ requested_fields => [ @req_fields ]);

                    or

    my @show_filehandle_objs = NACL::C::VolumeFile->show_filehandle(command_interface => $ci,
                          [ filter => { %filter } ,]
              [ requested_fields => [ @req_fields ]);

                    or

   my $show_filehandle_obj = $VolumeFile_obj->show_filehandle();

(Class or Instance method)

This method is a CS-like fetch method which accepts all the parameters of
L<fetch|lib-NACL-CS-ComponentState-pm/fetch> (requested_fields, filter, allow_empty)
and has all of the other properties of fetch methods (callable in scalar or list
context, throws a NoElementsFound exception if no matching elements are found).
The requested fields and filter takes L<ATTRIBUTES|lib-NACL-CS-VolumeFileShowFilehandle-pm/ATTRIBUTES>
of NACL::CS::VolumeFileShowFilehandle as input.
This method returns objects of the type L<NACL::CS::VolumeFileShowFilehandle|lib-NACL-CS-VolumeFileShowFilehandle-pm>.

=cut

sub show_filehandle {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts = $pkg_or_obj->_handle_obj_cs_like_calls(@args);
    require NACL::CS::VolumeFileShowFilehandle;
    my @state_objs = NACL::CS::VolumeFileShowFilehandle->fetch(%opts);
    $Log->exit() if $may_exit;
    return wantarray ? @state_objs : $state_objs[0];
}

=head2 repository_fhmap_show

    my $repository_fhmap_show_obj = NACL::C::VolumeFile->repository_fhmap_show(command_interface => $ci,
                          [ filter => { %filter } ,]
			  [ requested_fields => [ @req_fields ]);
						
                    or
					
    my @repository_fhmap_show_objs = NACL::C::VolumeFile->repository_fhmap_show(command_interface => $ci,
                          [ filter => { %filter } ,]
			  [ requested_fields => [ @req_fields ]);

                    or

   my $repository_fhmap_show_obj = $VolumeFile_obj->repository_fhmap_show();
						
(Class or Instance method)

This method is a CS-like fetch method which accepts all the parameters of
L<fetch|lib-NACL-CS-ComponentState-pm/fetch> (requested_fields, filter, allow_empty)
and has all of the other properties of fetch methods (callable in scalar or list
context, throws a NoElementsFound exception if no matching elements are found).
The requested fields and filter takes L<ATTRIBUTES|lib-NACL-CS-VolumeFileRepositoryFhmapShow-pm/ATTRIBUTES>
of NACL::CS::VolumeFileRepositoryFhmapShow as input.
This method returns objects of the type L<NACL::CS::VolumeFileRepositoryFhmapShow|lib-NACL-CS-VolumeFileRepositoryFhmapShow-pm>.

=cut

sub repository_fhmap_show {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts = $pkg_or_obj->_handle_obj_cs_like_calls(@args);
    require NACL::CS::VolumeFileRepositoryFhmapShow;
    my @state_objs = NACL::CS::VolumeFileRepositoryFhmapShow->fetch(%opts);
    $Log->exit() if $may_exit;
    return wantarray ? @state_objs : $state_objs[0];
}

=head2 file_reservation

B<DEPRECATED>. Use the method L<reservation> instead.

=cut

sub file_reservation {
    my $warn_message = 'This method is deprecated. Please use the method '
            . "'reservation' instead";
    NACL::C::Component->_print_deprecation_warning($warn_message);
    goto &reservation;
}

=head2 showfh

    my $showfh_cs_obj = NACL::C::VolumeFile->showfh(
        command_interface => $ci,
        vserver           => $vserver_name,
        path              => $file_path,
        ...
    );

                          or

    my $showfh_cs_obj = $file_obj->showfh(...);

(Class or instance method) This method executes 'showfh' DBlade command.

The implementation is present in
L<NACL::C::_Mixins::VolumeFileDirectory::showfh|lib-NACL-C-_Mixins-VolumeFileDirectory-pm>.

Supports 7Mode CLI and Nodescope CLI (for CMode).

Uses a 7Mode/Nodescope CLI APIset.

Invokes "showfh" command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< path => $file_path >>

(Required) File Path

Applicable for 7Mode CLI/Nodescope .

=item C<< 'nodescope-node-name' => $node_name >>

(Optional)
When run in nodescope, the component method defaults to running on the "local"
node. This argument can be used to run the method on node other than "local".

Applicable for 7Mode CLI/Nodescope .

=item C<< determine_appropriate_nodescope_node => 0|1 >>

(Optional, defaults to 0)
When set to 1, the method first determines which node the file is present
and makes the call in the context of that node. Note that when
C<nodescope-node-name> overrides this argument, meaning that if both
C<nodescope-node-name> is provided and C<determine_appropriate_nodescope_node> is
set to 1 then the node name provided in C<nodescope-node-name> is used.

This argument defaults to 0.

Applicable for 7Mode CLI/Nodescope.

=item C<< handle_name_ordinal => 0 | 1 >>

(Optional, defaults to 1)
Described at the beginning of this file.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

The implementation is present in
L<NACL::C::_Mixins::VolumeFileDirectory::showfh|lib-NACL-C-_Mixins-VolumeFileDirectory-pm>.

=cut


=head2 dumpblock_info

    my $dumpblock_cs_obj =   NACL::C::VolumeFile->dumpblock_info(command_interface => $ci,...);

(Class method only) This method executes 'dumpblock info' DBlade command and returns the
information returned by the command in
L<NACL::CS::VolumeFileDumpblockInfo|lib-NACL-CS-VolumeFileDumpblockInfo-pm> object.

The implementation is present in
L<NACL::C::_Mixins::VolumeFileDirectory::dumpblock_info|lib-NACL-C-_Mixins-VolumeFileDirectory-pm>.

Supports 7Mode CLI and Nodescope CLI (for CMode).

Uses a 7Mode/Nodescope CLI APIset.

Invokes "dumpblock info" command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< file => $file_path >>

(Required) File Path

Applicable for 7Mode CLI/Nodescope .

=item C<< 'nodescope-node-name' => $node_name >>

(Optional)
When run in nodescope, the component method defaults to running on the "local"
node. This argument can be used to run the method on node other than "local".

Applicable for 7Mode CLI/Nodescope .

=item C<< determine_appropriate_nodescope_node => 0|1 >>

(Optional, defaults to 0)
When set to 1, the method first determines which node the file is present
and makes the call in the context of that node. Note that when
C<nodescope-node-name> overrides this argument, meaning that if both
C<nodescope-node-name> is provided and C<determine_appropriate_nodescope_node> is
set to 1 then the node name provided in C<nodescope-node-name> is used.

This argument defaults to 0.

Applicable for 7Mode CLI/Nodescope.

=item Command options

All of the options listed under "PROGRAMMATIC NAME" on the
L<dumpblock|lib-NACL-APISet-Node-CLI-7Mode-default-dumpblock-pm--dumpblock_info>
page are accepted by this method.

Applicable for 7Mode CLI/Nodescope.

=item C<< handle_name_ordinal => 0 | 1 >>

(Optional, defaults to 1)
Described at the beginning of this file.

=back

=back

=over

=item Exceptions

=over

=item C<NACL::C::Exceptions::VolumeFile::BlockIsAHole>

This exception is thrown when an attempt is made to get the block information
for a block that is a hole.

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

The implementation is present in
L<NACL::C::_Mixins::VolumeFileDirectory::dumpblock_info|lib-NACL-C-_Mixins-VolumeFileDirectory-pm>.

=cut

=head2 show_inode

    my $show_inode_obj = NACL::C::VolumeFile->show_inode(command_interface => $ci,
                          [ filter => { %filter } ,]
			  [ requested_fields => [ @req_fields ]);
						
                    or
					
    my @show_inode_objs = NACL::C::VolumeFile->show_inode(command_interface => $ci,
                          [ filter => { %filter } ,]
			  [ requested_fields => [ @req_fields ]);

                    or

   my $show_inode_obj = $VolumeFile_obj->show_inode();
						
(Class or Instance method)

This method is a CS-like fetch method which accepts all the parameters of
L<fetch|lib-NACL-ComponentState-pm/fetch> (requested_fields, filter, allow_empty)
and has all of the other properties of fetch methods (callable in scalar or list
context, throws a NoElementsFound exception if no matching elements are found).
The requested fields and filter takes L<ATTRIBUTES|lib-NACL-CS-VolumeFileShowInode-pm/ATTRIBUTES>
of NACL::CS::VolumeFileShowInode as input.
This method returns objects of the type L<NACL::CS::VolumeFileShowInode|lib-NACL-CS-VolumeFileShowInode-pm>.

=cut

sub show_inode {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts = $pkg_or_obj->_handle_obj_cs_like_calls(@args);
    require NACL::CS::VolumeFileShowInode;
    my @state_objs = NACL::CS::VolumeFileShowInode->fetch(%opts);
    $Log->exit() if $may_exit;
    return wantarray ? @state_objs : $state_objs[0];
}

=head2 truncate

    NACL::C::VolumeFile->truncate(
        command_interface => $ci,
        vserver           => $vs_name, # Required for CMode
        path              => $file_path,
        size              => $size
    );

                          OR

    $file_obj->truncate(size => $size);

(Class or Instance method)

Truncate a file. Any data past the truncation point will be lost.

Supports CMode ZAPI and 7Mode ZAPI. Invokes "file-truncate-file" for both
CMode and 7Mode ZAPI.

=over

=item Options

All options are applicable for both CMode and 7Mode.

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for Instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< path => $file_path >>

(Required for class method, Not Applicable for Instance method)
File path

=item C<< size => $size >>

(Optional, defaults to 0)
The file size in bytes. Value is allowed to be between 0 .. (2^63)-1. Note
that the provided value should be only an integer and should B<not> have
any unit suffixes provided.

=back

=back

=cut

sub truncate {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);

    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {_optional_scalars(qw(size))},
    );

    $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_truncate_zapi',
                interface => 'ZAPI',
                set       => 'CMode|7Mode',
            },
        ],
    );
}

sub _truncate_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts   = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = $opts{apiset};

    my %api_args;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%api_args,
        copy   => [qw(path size)]
    );

    if ($opts{command_interface}->is_cmode()) {
        $pkg_or_obj->_handle_zapi_vserver_context(
            api_opts          => \%api_args,
            vserver           => $opts{vserver},
            command_interface => $opts{command_interface},
            frontend_method   => 'truncate_file',
        );
    }

    $apiset->file_truncate_file(%api_args);
    $Log->exit() if $may_exit;
}

=head2 punch_hole

    NACL::C::VolumeFile->punch_hole(
        command_interface  => $ci,
        path               => $file_path,
        vserver            => $vserver_name, # Required for CMode
        'hole-range-info'  => [
            { 'hole-start' => $start1, 'hole-size' => $size1 },
            { 'hole-start' => $start2, 'hole-size' => $size2 },
            ...
        ],
    );

    $file_obj->punch_hole(
        'hole-range-info' => [
            { 'hole-start' => $start1, 'hole-size' => $size1 },
            { 'hole-start' => $start2, 'hole-size' => $size2 },
            ...
        ],
    );

(Class or Instance Method)
Punch hole in the file. Hole punching involves reclaiming of blocks in a
file by unallocating them and then direct or indirect blocks can be made
to point to 0.

Supports CMode ZAPI and 7Mode/Nodescope CLI.
Invokes "file_punch_hole_iter" for CMode ZAPI.
Invokes "file punchhole" for 7Mode/Nodescope CLI.

For CMode, this method defaults to using the ZAPI implementation (unless
'punch-bad' => 1 is provided, in which case it uses the CLI implementation
since punching VBN_BAD holes is possible only through the CLI).

=over

=item Options

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for Instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< path => $file_path >>

(Required for class method, Not Applicable for Instance method)
File path

=item C<< vserver => $vserver_name >>

(Required for class method in CMode, Not Applicable for Instance method or for 7Mode)
The vserver on which the file is present.

=item C<< 'hole-range-info' => [ { 'hole-start' => $start1, 'hole-size' => $size1 }, {...}, ... ] >>

(Arrayref of hashrefs)
Information of the holes to be punched. Each entry in the array-reference
is a hash-reference and each such hash-reference describes one hole to be
punched. Each hash-reference should have the following keys:

* 'hole-start': Starting offset of the hole. Range: [0..2^63-1]

* 'hole-size': Size of the hole in terms of bytes. If punching a hole WAFL
presently limits this value to 4MB. Range: [0..2^63-1]

=item C<< 'punch-bad' => 0 | 1 >>

(Optional, applicable only from FS.0 onwards)
If set to 1, punches VBN_BAD blocks. (Else a VBN_HOLE block is punched)
Note that since this can only be done through the CLI, if this option is
provided, then the CLI implementation is used.

More information is available on burt 683120.

=item C<< 'waffi' => 1 >>

(Optional, applicable only from metawafl/HN.0 onwards)
To mark as VBN_BAD, need to set 'waffi'  to 1 in conjunction with 'punch-bad'.
Note that since this can only be done through the CLI, if this option is
provided, then the CLI implementation is used.

More information is available on burt 979522.

=item C<< 'nodescope-node-name' => $node_name >>

(Optional)
When run in nodescope, the component method defaults to running on the "local"
node. This argument can be used to run the method on node other than "local".

Applicable for 7Mode CLI/Nodescope .

=item C<< determine_appropriate_nodescope_node => 0|1 >>

(Optional, defaults to 0)
When set to 1, the method first determines which node the file is present
and makes the call in the context of that node. Note that when
C<nodescope-node-name> overrides this argument, meaning that if both
C<nodescope-node-name> is provided and C<determine_appropriate_nodescope_node> is
set to 1 then the node name provided in C<nodescope-node-name> is used.

This argument defaults to 0.

Applicable for 7Mode CLI/Nodescope.

=item C<< handle_name_ordinal => 0 | 1 >>

(Optional, defaults to 1)
Described at the beginning of this file.

=back

=cut

sub punch_hole {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);

    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            'hole-range-info' => {type => ARRAYREF},
            'punch-bad'       => {type => SCALAR, optional => 1},
            _optional_scalars( qw( waffi ) ),
            $pkg_or_obj->_nodescope_options_validate_spec(),
        },
    );

    $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_punch_hole_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
                check     => '_check_punch_hole_cmode_zapi',
            },
            {   method    => '_punch_hole_7mode_cli',
                interface => 'CLI',
                set       => '7Mode|Nodescope',
            },
        ],
    );
}

sub _punch_hole_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts   = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = $opts{apiset};

    my %api_args = (
        path              => $opts{path},
        'attributes-list' => $opts{'hole-range-info'}
    );

    $pkg_or_obj->_handle_zapi_vserver_context(
        api_opts          => \%api_args,
        vserver           => $opts{vserver},
        command_interface => $opts{command_interface},
        frontend_method   => 'punch_hole',
    );

    while (1) {
        my $response = $apiset->file_punch_hole_iter(%api_args);
        my $only_row = $response->get_parsed_output()->[0];

        # "next-tag" indicates that the iter needs to be run again
        my $tag = $only_row->{'next-tag'};
        if ($tag) {
            $api_args{tag} = $tag;
        } else {
            last;
        }
    }

    $Log->exit() if $may_exit;
}

sub _punch_hole_7mode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts   = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = $opts{apiset};
    my %api_args;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%api_args,
        copy   => [qw(nodescope-node-name path punch-bad waffi)]
    );

    foreach my $range_hash (@{$opts{'hole-range-info'}}) {
        my $ref = ref $range_hash || '';
        if ($ref ne 'HASH') {
            $Log->exit() if $may_exit;
            NATE::Exceptions::Argument->throw('Each entry in the '
                    . "'hole-range-info' array-ref should be a hash-reference "
                    . "but one of the entries was of type '$ref'. Dumper:\n"
                    . NACL::ComponentUtils::Dumper($range_hash));
        }
        foreach my $key (qw(hole-start hole-size)) {
            if (!exists $range_hash->{$key}) {
                $Log->exit() if $may_exit;
                NATE::Exceptions::Argument->throw('Each entry in the '
                        . "'hole-range-info' array-ref should be a hash-reference "
                        . "containing the keys 'hole-start' and 'hole-size' but "
                        . "one of the entries did not contain the key '$key'. "
                        . "Dumper:\n"
                        . NACL::ComponentUtils::Dumper($range_hash));
            }
        }
        $apiset->file_punchhole(
            %api_args,
            offset => $range_hash->{'hole-start'},
            length => $range_hash->{'hole-size'}
        );
    }

    $Log->exit() if $may_exit;
}

sub _check_punch_hole_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;

    if ($opts{'punch-bad'}) {
        $Log->exit() if $may_exit;
        NACL::Exceptions::InvalidChoice->throw('Punching VBN_BAD blocks is '
                . 'possible only through the CLI, so the ZAPI implementation '
                . 'cannot be used');
    }

    $Log->exit() if $may_exit;
}

=head2 get_space_reservation_info

    my $file_space_obj =
        NACL::C::VolumeFile->get_space_reservation_info(command_interface => $ci,
            [ filter => { %filter } ,]

                    or

    my $file_space_obj = $File_obj->get_space_reservation_info();

(Class or Instance method)

This method is a CS-like fetch method which accepts filter parameter of
L<fetch|lib-NACL-ComponentState-pm/fetch> and has all of the other properties
of fetch methods (callable in scalar or list context, throws a NoElementsFound exception
if no matching elements are found). The filter takes
L<ATTRIBUTES|lib-NACL-CS-VolumeFileGetSpaceReservationInfo-pm/ATTRIBUTES> of
NACL::CS::VolumeFileGetSpaceReservationInfo as input.This method returns objects of the type L<NACL::CS::VolumeFileGetSpaceReservationInfo|lib-NACL-CS-VolumeFileGetSpaceReservationInfo-pm>.

=cut

sub get_space_reservation_info {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args)  = @_;
    my %opts       = $pkg_or_obj->_handle_obj_cs_like_calls(@args);
    require NACL::CS::VolumeFileGetSpaceReservationInfo;
    my @state_objs = NACL::CS::VolumeFileGetSpaceReservationInfo->fetch(%opts);
    $Log->exit() if $may_exit;
    return wantarray ? @state_objs : $state_objs[0];
} ## end sub show_space


__PACKAGE__->_load_event_class();

sub _primary_keys_validate_spec {
    return (
        path    => {type => SCALAR | OBJECT},
        vserver => {type => SCALAR | OBJECT | UNDEF, optional => 1}
    );
}

# 1. Translate file -> path (while logging a warning if file is provided)
# 2. Log a warning if "vserver" is not provided for a CMode system
# 3. Make handle_name_ordinal a valid argument for all calls (defaulting to 1)
sub _common_validate_with {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts = validate_with(
        params => \@args,
        spec   => {
            params          => {type => ARRAYREF},
            _method         => {type => SCALAR, optional => 1},
            additional_spec => {type => HASHREF, default => {}},
        },
        allow_extra => 1
    );
    my $method = delete $opts{_method};
    my %params = @{$opts{params}};
    my $cli_cmd = $opts{cli_cmd} || '';

    if ( $cli_cmd !~ /privileged-delete/ ) {
	if (exists $params{file}) {
            _file_warning($method);
	}

	$pkg_or_obj->_hash_move(
	    source => \%params,
            target => \%params,
            map    => {file => 'path'}
	);
    }

    my %additional_spec = (
        handle_name_ordinal => {type => SCALAR, default => 1},
        %{$opts{additional_spec}}
    );

    my %validated_opts = $pkg_or_obj->SUPER::_common_validate_with(
        %opts,
        params          => [%params],
        additional_spec => \%additional_spec
    );

    if (   exists $validated_opts{command_interface}
        && $validated_opts{command_interface}->is_cmode()
        && !defined $validated_opts{vserver})
    {
        my $to_method = '';
        if (defined $method) {
            $to_method = "to $method";
        }
        $Log->warn("Call $to_method was made on a CMode system but "
                . "the 'vserver' argument was not passed. Please update "
                . "the call to send the 'vserver' argument. In the "
                . 'absence of this argument, certain features such as '
                . 'automatic-determination of node-name and ZAPI cannot '
                . 'be used');
    }

    $Log->exit() if $may_exit;
    return wantarray ? %validated_opts : {%validated_opts};
}

sub _check_for_vserver {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    my $ci = $opts{command_interface};

    if ($ci->is_cmode()) {
        if (!defined $opts{vserver}) {
            my $method = $opts{_method};
            $Log->exit() if $may_exit;
            NACL::Exceptions::InvalidChoice->throw('Cannot use ZAPI '
                    . 'implementation because vserver was not passed for a '
                    . 'CMode command_interface');
        }
    }

    $Log->exit() if $may_exit;
}

sub _check_for_vserver_sksim {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    $pkg_or_obj->_check_sksim(%opts);
    $pkg_or_obj->_check_for_vserver(%opts);
    $Log->exit() if $may_exit;
}

sub _handle_boolean_opts {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts = validate_with(
        params => \@args,
        spec   => {
            all_opts     => {type => HASHREF},
            boolean_opts => {type => ARRAYREF},
        }
    );
    my $all_opts = $opts{all_opts};

    foreach my $boolean (@{$opts{boolean_opts}}) {
        if (defined $all_opts->{$boolean}) {
            if ($all_opts->{$boolean} =~ /false/i) {
                $all_opts->{$boolean} = 0;
            }
        }
    }

    $Log->exit() if $may_exit;
}

my %File_Warning;

sub _file_warning {
    $Log->enter() if $may_enter;
    my $method = $_[0] || '';
    unless (exists $File_Warning{$method}) {
        my $to_method = '';
        if ($method) {
            $to_method = "to $method";
        }
        my $warn_message = "The argument 'file' was passed in the call $to_method "
                . "but is deprecated. The argument 'path' should be used instead";
        NACL::C::Component->_print_deprecation_warning($warn_message);
        $File_Warning{$method} = 1;
    }
    $Log->exit() if $may_exit;
}

my %Vserver_Warning;

sub _vserver_warning {
    $Log->enter() if $may_enter;
    my $method = $_[0] || '';
    unless (exists $Vserver_Warning{$method}) {
        my $to_method;
        if (defined $method) {
            $to_method = "to $method";
        }
        $Log->warn("Call $to_method was made on a CMode system but "
                . "the 'vserver' argument was not passed. Please update "
                . "the call to send the 'vserver' argument. In the "
                . 'absence of this argument, certain features such as '
                . 'automatic-determination of node-name and ZAPI cannot '
                . 'be used');
        $Vserver_Warning{$method} = 1;
    }
    $Log->exit() if $may_exit;
}

sub call_on_apiset {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my @ret = $pkg_or_obj->_call_on_apiset(@args);
    $Log->exit() if $may_exit;
    return wantarray ? @ret : $ret[0];
}

=head2 get_sfod_path

    my $sfod_path = $file_obj->get_sfod_path();

Returns a path to the file suitable for passing to NACL::C::VolumeFileCopy or
NACL::C::VolumeFileMove. The STask for each of these will call this method if
directly passed a NACL::C::VolumeFile instance.

=cut

sub get_sfod_path {
    my $self = $_[0];

    # vs0:volume/dir/file
    # /vol/volume/dir/path => "", "vol", "volume", "dir", "path"
    my $obj = $self->_extract_volume_file_from_path(path => $self->path(),);

    my @components = split /\//, $self->path();
    shift @components;
    shift @components;
    return $self->vserver() . ":" . join "/", @components;
}

=head2 new_destination_file

    $dest_file = $file->new_destination_file();

    $dest_file = $file->new_destination_file(
        volume => $other_volume_instance,
    );

    $dest_file = $file->new_destination_file(
        volume => $other_volume_instance,
        dest_dir => $destination_directory,
    );

Instantiate a NACL::C::VolumeFile instance with a path suitable for use as a
destination file. This is useful for file level replication, mobility, or
cloning operations to automatically create a destination file path. The file
will not exist until some other action is taken to create the element on the
filer, this call will just instantiate a component for an element that does not
exist yet.

=over

=item Options

=item C<< volume=>$volume >>

(optional)
Replace the volume name with the volume from the passed NACL::C::Volume instance

=item C<< dest_dir=>$destination_directory >>

(optional)
Add the destination directory to the path. For example, for the destination
directory /subdir1/subdir2/ the final destination path will look like
/vol/volumename/subdir1/subdir2/file.txt. This function assumes the
destination directory exists and does not attempt to create one for user.

=back

=cut

sub new_destination_file {
    my ($self, %args) = @_;

    my $current_volume        = $self->get_containing_volume_obj();
    my $destination_volume    = $args{volume} || $current_volume;
    my $destination_directory = $args{dest_dir};
    my $path                  = $self->path();

    if ($path =~ /_dest_(\d+)$/) {
        my $new_num = $1 + 1;
        $path =~ s/_dest_(\d+)$/_dest_$new_num/;
    } else {
        $path .= '_' . int(rand(1000000)) . "_dest_1";
    }

    my @components = split /\//, $path;
    $components[2] = $destination_volume->volume();

    if ($destination_directory) {
        my @dest_dir_components = split /\//, $destination_directory;
        @dest_dir_components = grep length($_), @dest_dir_components;
        splice @components, 3, 0, @dest_dir_components;
    }

    my $type = blessed $self;

    return $type->new(
        command_interface => $destination_volume->command_interface(),
        vserver           => $destination_volume->vserver(),
        path              => (join "/", @components),
    );
}

=head2 alternate_file_location

    # The snapmirror object can be for a DP or XDP relationship
    $file_on_vault = $file->alternate_file_location(
        snapmirror => $snapmirror_obj,
    );

    # The object passed in must be a NACL::C::Volume, not a VolumeClone. This
    # same format can be used to get a file on a LS mirror volume or an
    # unrelated volume.
    $file_on_clone = $file->alternate_file_location(
        volume => $volume_clone_volume,
    );

    # Operations can be chained
    $file_in_mirror_snapshot = $file->alternate_file_location(
        snapmirror => $snapmirror_obj,
    )->alternate_file_location(
        snapshot => $snapshot_in_dp_volume,
    );

    # Alternatively, the caller can just directly pass the snapshot object.
    # The vserver and volume from the snapshot object will be used for the
    $file_in_snapshot = $file->alternate_file_location(
        snapshot => $snapshot_obj,
    );

    # This method can be used to get the AFS path of a file in a snapshot
    $file_in_afs = $file_in_snapshot->alternate_file_location();


(Instance Method)

Files can have alternate paths to different version of the file. Files can live
in snapshots, snapmirrors, volume clones, and snapvaults. This method allows the
caller to easily get a NACL::C::VolumeFile instance for alternate versions of
the file in other locations.

The file instance that this method is called on may be a file that is or is not
in a snapshot. Any .snapshot component will be stripped from the path before the
if specifying a volume argument or no arguments. If a snapmirror argument is
specified, we will preserve any snapshot path from the original file. This is
done because most callers will want the file in the snapshot on the
destination. If a snapshot is passed, then the vserver, volume, and snapshot
for the alternate file will be taken from that snapshot object.

=over

=item C<< snapshot >>

A NACL::C::VolumeSnapshot instance for where the file
has an alternate location. This snapshot may or may not be on the same volume as
the original file. This argument must be specified by itself.

=item C<< snapmirror >>

A NACL::C::Snapmirror instance for where the file has
an alternate location. This will get the file on the destination volume of the
SnapMirror/SnapVault. This argument must be specified by itself.

=item C<< volume >>

A NACL::C::Volume instance for where the file has an
alternate location. This argument must be specified by itself.

=back

=cut

sub alternate_file_location {
    my ($file, @args) = @_;
    $Log->enter() if $may_enter;

    my %opts = validate(
        @args,
        {   snapmirror => {type => OBJECT, optional => 1},
            snapshot   => {type => OBJECT, optional => 1,},
            volume     => {type => OBJECT, optional => 1},
        }
    );

    my $snapmirror = $opts{snapmirror};
    my $snapshot   = $opts{snapshot};
    my $volume     = $opts{volume};

    my $arg_count = 0;
    $arg_count++ if $snapmirror;
    $arg_count++ if $snapshot;
    $arg_count++ if $volume;
    if ($arg_count > 1) {
        NATE::Exceptions::Argument->throw("Must specify no more than one "
                . "argument to alternate_file_location");
    }

    my $destination_vserver_name;
    my $destination_volume_name;
    my $destination_command_interface;

   # If someone passes in a snapshot object, get the volume, vserver, and
   # command interface for the destination file from that. If they pass in a
   # snapmirror, we can all three from that. A volume can be passed to get the
   # file on a second volume, so get them from that if passed. If none were
   # passed in, get them from the file object.
    if ($snapshot) {
        $destination_volume_name       = $snapshot->volume();
        $destination_vserver_name      = $snapshot->vserver();
        $destination_command_interface = $snapshot->command_interface();
    } elsif ($snapmirror) {
        my $destination_volume
            = $snapmirror->cast_component_to_task()->get_destination_volume();
        $destination_volume_name  = $destination_volume->volume();
        $destination_vserver_name = $destination_volume->vserver();
        $destination_command_interface
            = $destination_volume->command_interface();
    } elsif ($volume) {
        $destination_volume_name       = $volume->volume();
        $destination_vserver_name      = $volume->vserver();
        $destination_command_interface = $volume->command_interface();
    } else {

        # This case is when the caller specifies no snapshot or snapmirror. It
        # is used to take a file that is in the snapshot and get the same file
        # in the AFS.
        my $containing_volume = $file->get_containing_volume_obj();
        $destination_volume_name  = $containing_volume->volume();
        $destination_vserver_name = $containing_volume->vserver();
        $destination_command_interface
            = $containing_volume->command_interface();
    }

    # Remove the /vol/volume from the original file
    my @file_components = split /\//,
        $file->get_file_or_dir_name_from_path(path => $file->path(),);

    # If the original file was in a snapshot, remove that
    my $original_snapshot_name;
    if ($file_components[0] eq '.snapshot') {
        shift @file_components;
        $original_snapshot_name = shift @file_components;
    }

   # If the destination is supposed to be a snapshot, add that back in. If
   # someone passed in a snapmirror, assume they want the same snapshot on the
   # destination volume, so preserve the original snapshot if it had one.
    if ($snapshot) {
        @file_components
            = ('.snapshot', $snapshot->snapshot(), @file_components);
    } elsif ($snapmirror && $original_snapshot_name) {
        @file_components
            = ('.snapshot', $original_snapshot_name, @file_components);
    }

    my $file_path = join '/', @file_components;
    my $destination_file = NACL::STask::VolumeFile->new(
        command_interface => $destination_command_interface,
        vserver           => $destination_vserver_name,
        path =>
            join('/', '', 'vol', $destination_volume_name, @file_components,)
    );
    $Log->exit() if $may_exit;
    return $destination_file;
}

sub _remove_invalid_verify_state_opts {
    $Log->enter() if $may_enter;
    my ($pkg, %opts) = @_;

    #handle_name_ordinal is not the fetched field.
    delete $opts{handle_name_ordinal};
    $Log->exit() if $may_exit;
    return %opts;
}

=head2 get_client_path_to_file

    # Will return client mount path to file. Example:
    $client_path_to_file = $file->get_client_path_to_file(
        mount_obj => $mount_obj,
    );
    
=over

=item Options

=item C<< mount_obj => $mount_obj >>

An instance of type "NACL::C::Client::Mount".

=back

=cut    

sub get_client_path_to_file {
    $Log->enter() if $may_enter;
    my ($self, %args) = @_;

    my $client_path_to_file = _get_client_path_to_file(
        file_obj  => $self,
        mount_obj => $args{mount_obj},
    );

    $Log->exit() if $may_exit;
    return $client_path_to_file;
}

=head2 get_file_name

Will return the file name from the path. Example:
my $file_name = $file_obj->get_file_name();
(Instance method)
Returns the file name.

=cut

sub get_file_name {
    $Log->enter() if $may_enter;
    my $self = $_[0];

    my $file_name = $self->get_file_or_dir_name_from_path(path => $self->path());

    $Log->exit() if $may_exit;
    return $file_name;
}

=head2 list_streams

    @streams = NACL::C::VolumeFile->list_streams(
        command_interface  => $ci,
        path               => $file_path,
    );

    @streams = $file->list_streams();

(Class or instance method)

Returns the streams associated with the file.

Applicable only for CMode CLI.

Invokes NACL::C::Volume->explore.

=over

=item Options

=item C<< command_interface => $command_interface >>

(Required for Class Method, Not Applicable for Instance Method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< path => $file_path >>

(Required for Class Method, Not Applicable for Instance Method)
Path to the file to get list of streams from.

=back

=cut

sub list_streams {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    my ( $command_interface, $path , $volume, $file_name, $vol_inode_info);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@opts,
    );
    $command_interface = $opts{command_interface};
    $path = $opts{path};
    #extract volume name & file
    my $obj = $pkg_or_obj->_extract_volume_file_from_path(path => $path);
    $volume = $obj->volume();
    $file_name = $obj->file();
    my %common_opts;
    $pkg_or_obj->_copy_common_component_params(
        source => \%opts,
        target => \%common_opts
    );
    $vol_inode_info = NACL::C::Volume->explore(
        %common_opts,
        command_interface => $command_interface,
        format => "inode",
        scope  => "$volume."."/".$file_name,
    );
    my @vol_exp = NACL::C::Volume->explore(
        %common_opts,
        command_interface => $command_interface,
        format => "dir",
        scope  => "$volume.". $vol_inode_info->sinode(),
    );
    my @streams_list;
    #push streams to list
    foreach my $explore_obj (@vol_exp) {
        my @names = $explore_obj->name();
        #remove "." and ".." being current & parent directory
        if ($names[0] !~ /^\.\.?$/) {
            push (@streams_list, $names[0]);
        }
    }
    $Log->exit() if $may_exit;
    return @streams_list;
}

=head2 clear_error

    $response_obj = NACL::C::VolumeFile->clear_error(
        command_interface  => $ci,
        path               => $file_path,
        vserver            => $vserver_name, # Required for CMode
        inaccessible       => 1
        %other_option
    );

            or

    $response_obj = $file_obj->clear_error(
        %other_option
    );

(Class or Instance Method)
Clear errors on blocks of the specified file and returns the response object,
which is returned by the APISet.

Supports 7Mode/Nodescope CLI.
Invokes "file_clear_error" for 7Mode/Nodescope CLI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for Instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< path => $file_path >>

(Required for class method, Not Applicable for Instance method)
File path

=item C<< vserver => $vserver_name >>

(Required for class method in CMode, Not Applicable for Instance method or for 7Mode)
The vserver on which the file is present.


=item C<< unverified => 0|1 >>

(Optional)
unverified option clears E_UNVERIFIED_BLOCK error.

=item C<< inaccessible => 0|1 >>

(Optional)
inaccessible option clears iron marked inaccessible block ranges

=item C<< nodescope-node-name => $node_name >>

(Optional)
When run in nodescope, the component method defaults to running on the "local"
node. This argument can be used to run the method on node other than "local".

Applicable for 7Mode CLI/Nodescope .

=item C<< determine_appropriate_nodescope_node => 0|1 >>

(Optional, defaults to 0)
When set to 1, the method first determines which node the file is present
and makes the call in the context of that node. Note that when
C<nodescope-node-name> overrides this argument, meaning that if both
C<nodescope-node-name> is provided and C<determine_appropriate_nodescope_node> is
set to 1 then the node name provided in C<nodescope-node-name> is used.

This argument defaults to 0.

Applicable for 7Mode CLI/Nodescope.

=item C<< handle_name_ordinal => 0 | 1 >>

(Optional, defaults to 1)
Described at the beginning of this file.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=item C<NACL::C::Exceptions::VolumeFile::DoesNotExist>

This type of exception gets thrown if an attempt is made to clear error
on a file which does not exists.

=back

=back

=cut

sub clear_error {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;

    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            $pkg_or_obj->_nodescope_options_validate_spec(),
            _optional_scalars(
                qw(unverified fbns inaccessible)
            ),
        }
    );

    my $output;

    try {
        $output = $pkg_or_obj->call_on_apiset(
            %opts,
            choices => [
                {   method    => '_clear_error_7mode_cli',
                    interface => 'CLI',
                    set       => '7Mode|Nodescope',
                },
            ],
        );
    } catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = $_[0];
        NACL::C::Exceptions::VolumeFile::DoesNotExist
            ->detect_convert_and_throw(exception => $exception);
        $Log->exit() if $may_exit;
        $exception->throw();
    };

    $Log->exit() if $may_exit;
    return $output;
}

sub _clear_error_7mode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts              = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset            = $opts{apiset};

    my @allowed_options =
        $apiset->get_allowed_options_for_command(command => 'file_clear_error');
    push @allowed_options, 'nodescope-node-name';

    my %api_args;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%api_args,
        copy   => \@allowed_options,
        map    => {path => 'file', volume => 'vol'}
    );

    my $response = $apiset->file_clear_error(%api_args);

    $Log->exit() if $may_exit;
    return $response->get_parsed_output();
}

=head2 is_file_in_qtree

    # Instance call
    my $boolean = $file_obj->is_file_in_qtree(%common_option);

(instance method)
Returns 1 (true) if the file resides in VolumeQtree, or else 0 (false).

=cut

sub is_file_in_qtree {
    $Log->enter() if $may_enter;
    my ( $self, @args ) = @_;

    my %opts = $self->_common_validate_with( params => \@args, );

    my %common_opts_with_ci;

    $self->_copy_common_component_params_with_ci(
        source => \%opts,
        target => \%common_opts_with_ci,
    );

    my $obj = $self->_extract_volume_file_from_path( path => $self->path() );

    my @file_components = split /\/+/, $obj->file();

    my $qtree = NACL::CS::VolumeQtree->fetch(
        %common_opts_with_ci,
        filter =>
          { 'qtree-path' => "/vol/" . $obj->volume() . "/$file_components[0]" },
        allow_empty => 1
    );

    $Log->exit() if $may_exit;

    return ( defined $qtree ) ? 1 : 0;
}

=head2 is_file_in_snapshot

    # Instance call
    my $boolean = $file_obj->is_file_in_snapshot(%common_option);

(instance method)
Returns 1 (true) if the file resides in snapshot, or else 0 (false).

=cut

sub is_file_in_snapshot {
    $Log->enter() if $may_enter;
    my ( $self, @args ) = @_;

    my %opts = $self->_common_validate_with( params => \@args, );

    my @file_components = split /\/+/, $self->path();

    $Log->exit() if $may_exit;
    return ( $file_components[3] eq '.snapshot' ) ? 1 : 0;
}

=head2 testsetattr

Execute testsetattr on the file, without having to depend on apiset and path names
Only works for debug-build.

    my $file = NACL::STask::VolumeFile->create(...);
    my $file_attributes = $file->testsetattr('get-attributes' => 1);
    $file_offset_length = $file->testsetattr('read-offset-and-length' => '0 100');
    $file->testsetattr('write-offset-and-length' => '0 100');

Use it as a pass through to apiset testsetattr.

=cut

 __PACKAGE__->_nodescope_builder(command => 'testsetattr');

=head2 wreckcxt

 Metwafl nodescope command 'wreckcxt'

=cut

__PACKAGE__->_nodescope_builder(command => 'wreckcxt');

=head2 filersio

 filersio support for VolumeFile

=cut

__PACKAGE__->_nodescope_builder(
    command     => 'filersio_asyncio_active',
    map         => { path => 'filename' },
);

__PACKAGE__->_nodescope_builder(
    command     => 'filersio_asyncio_pause',
    map         => { path => 'filename' },
);

__PACKAGE__->_nodescope_builder(command => 'filersio_status');
__PACKAGE__->_nodescope_builder(command => 'filersio_stop');
__PACKAGE__->_nodescope_builder(command => 'filersio_spc1');

=head2 wafl_showvbns

Executes the wafl showvbns command on the file without needing to pass path
and command interface

    my $file = $Volume->file_create(...);
    $file->wafl_showvbns(...);

This accepts all input accepts by wafl_showvbns apiset

=cut

__PACKAGE__->_cs_call_method_builder(method => 'wafl_showvbns', extra_filters => ['path']);

=head2 hammer

Executes the hammer command on the file without needing to pass filename
and command interface

    my $file = $Volume->file_create(...);
    $file->hammer(...);

This accepts all input accepts by hammer apiset

=cut

__PACKAGE__->_nodescope_builder(
    command => 'hammer',
    map     => { path => 'file-name' },
);

=head2 wafl_get_child_and_next_fbn

Executes the wafl_get_child_and_next_fbn command on the file without
needing to pass path and command interface

    my $file = $Volume->file_create(...);
    $file->wafl_get_child_and_next_fbn(...);

=cut

__PACKAGE__->_cs_call_method_builder(method => 'wafl_get_child_and_next_fbn', extra_filters => ['path','fbn','level']);

=head2 wafl_get_subtree_level

Executes the wafl_get_subtree_level command on the file without
needing to pass path and command interface

    my $file = $Volume->file_create(...);
    $file->wafl_get_subtree_level(...);

=cut

__PACKAGE__->_cs_call_method_builder(method => 'wafl_get_subtree_level', extra_filters => ['path','fbn']);

1;
