#
# Copyright (c) 2001-2013 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary VolumeFlexcache Component Module
## @author dl-nacl-dev@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::C::VolumeFlexcache

=head1 DESCRIPTION

C<NACL::C::VolumeFlexcache> is a derived class of
L<NACL::C::Component::ONTAP|lib-NACL-C-Component-ONTAP-pm>. It represents access
to an ONTAP VolumeFlexcache.  A related class is
L<NACL::CS::VolumeFlexcache|lib-NACL-CS-VolumeFlexcache-pm>, which represents
the past state of an ONTAP VolumeFlexcache.

=head1 ATTRIBUTES

=head2 command_interface

See L<NACL::C::Component::command_interface|lib-NACL-C-Component-pm/command_interface>

=head2 origin_volume

=head2 vserver

=head2 cache_volume

=cut

package NACL::C::VolumeFlexcache;

use strict;
use warnings;

use NACL::CS::VolumeFlexcache;
use NACL::ComponentUtils qw (_optional_scalars);

use base 'NACL::C::Component::ONTAP';
use Params::Validate qw(validate validate_with SCALAR ARRAYREF UNDEF);
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Data::Dumper;
use NATE::BaseException;

=head1 METHODS

=head2 new

  my $VolumeFlexcache = NACL::C::VolumeFlexcache->new(command_interface => $ci,
                                   origin_volume => $origin_volume,
                                   vserver => $vserver,
                                  );

Return a new VolumeFlexcache component object with the given attributes.

=cut

use Class::MethodMaker [
    scalar => 'origin_volume',
    scalar => 'vserver',
    scalar => 'cache_volume',
];

=head2 find

    my $VolumeFlexcache  = NACL::C::VolumeFlexcache->find(command_interface => $ci, ...);
    my @VolumeFlexcaches = NACL::C::VolumeFlexcache->find(command_interface => $ci, ...);

(Class method)
A constructor of sorts, which discovers which elements are present (by
interacting with the given command interface).  In an array context it
creates a component object for each present element and returns them.
In a scalar context it creates a component object for just one
element.

This method is generally not implemented in individual components. The base
class (L<NACL::C::Component::ONTAP|lib-NACL-C-Component-ONTAP-pm> provides an implementation.
See <L<NACL::C::Component::find|lib-NACL-C-Component-pm/find> for a more
detailed description along with explanation of all the options accepted.

=head2 state

    my $VolumeFlexcache_state = $volumeflexcache_obj->state();

(Instance method)
Fetch the current state of the element associated
with this component, and return a Component State object corresponding
to this component (calling this on NACL::C::VolumeFlexcache would return
a NACL::CS::VolumeFlexcache object)

This method is generally not implemented in individual components. The base
class (L<NACL::C::Component::ONTAP|lib-NACL-C-Component-ONTAP-pm> provides an implementation.
See <L<NACL::C::Component::state|lib-NACL-C-Component-pm/state> for a more
detailed description along with explanation of all the options accepted.

=cut

=head2 delete

    NACL::C::VolumeFlexcache->delete(command_interface => $ci,
                                 'origin-volume' => $origin_volume,
                                  vserver => $vserver,
                                  ...);

                    or

    $volume_flexcache_obj->delete();

(Class or instance method)
This method is used to delete caching for a given volume throughout the cluster.

Uses a CMode CLI/ZAPI APISet.

Invokes "flexcache-delete" API for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "vserver" => $vserver >>

(Required for class method, Not Applicable for instance method)
Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: vserver

=item C<< "origin-volume" => $origin_volume >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: origin-volume

=item C<< "extended_query" => $string >>

(Optional)

Applicable for CMode CLI.

=item C<< '*' >>

This method takes all other options supported by
"volume flexcache delete" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub delete {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( origin-volume
                    )
            )
        },
        api => 'volume_flexcache_delete',
    );

    $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_delete_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_delete_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::VolumeFlexcache::delete',
    );

    $Log->exit() if $may_exit;
} ## end sub delete

sub _delete_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@args,
        api => 'volume_flexcache_delete');
    $Log->exit() if $may_exit;
} ## end sub _delete_cmode_cli

sub _delete_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_base_backend_cmode_zapi(
        @args,
        api  => 'flexcache-delete',
        copy => [qw(origin-volume vserver)]
    );
    $Log->exit() if $may_exit;
} ## end sub _delete_cmode_zapi

=head2 create

    my $VolumeFlexcache_object    = NACL::C::VolumeFlexcache->create(
                                command_interface => $ci,
                                 origin_volume => $origin_volume,
                                 vserver => $vserver,
                                  ...);

(Class method)
This method is used to cache a volume throughout the cluster and returns a
VolumeFlexcache component object.

Note that Flexcache is not supported for low memory platforms. (See
http://wikid.netapp.com/w/WAFL/SN.0/Change_Request/WAFL_LowMemPlat_Limitations
for more details.) For this reason, test-cases which use flexcache and
that might run on any of the listed low memory platforms should be disabled
if run against a low memory platform. Code like this can be used at the
beginning of the TCD-based test case:

    sub testcase {
        my $reason;
        if (!$node->is_feature_supported(
            feature    => 'flexcache',
            reason_ref => \$reason
        )) {
            $Test->nlog($reason);
            return $TCD::NA;
        }
        ... Rest of the test case here ...
    }

Uses a CMode CLI/ZAPI APISet.

Invokes "flexcache-create" API for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "vserver" => $vserver >>

(Required)

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: vserver

=item C<< "origin-volume" => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: origin-volume

=item C<< "extended_query" => $string >>

(Optional)

Applicable for CMode CLI.

=item C<< '*' >>

This method takes all other options supported by
"volume flexcache create" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub create {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw(origin-volume
                    )
            )
        },
        api => 'volume_flexcache_create',
    );

    $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_create_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_create_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::VolumeFlexcache::create',
    );

    my %primary_keys;
    $pkg->_copy_primary_keys(
        source => \%opts,
        target => \%primary_keys
    );

    # It is possible for the command to not have created any cache volumes
    # if we're running on un-supported hardware or if the aggregates are
    # too full to accomodate the cache volume. So let's verify that
    # at least one cache volume was created.
    my %common_opts;
    $pkg->_copy_common_component_params_with_ci(
        source => \%opts,
        target => \%common_opts
    );
    unless (
        $pkg->find(%common_opts, filter => {%primary_keys}, allow_empty => 1))
    {

        # Did not create any caches, was it because we ran on unsupported
        # hardware?
        my $reason;
        my $supported = $opts{command_interface}->is_feature_supported(
            feature    => 'flexcache',
            reason_ref => \$reason
        );

        my $error_text;
        if ($supported) {

            # Ran on supported hardware, yet did not create cache volumes.
            # Probably because the aggregates are full.
            $error_text
                = 'VolumeFlexcache->create did not create any '
                . 'cache volumes. This could be because the candidate '
                . 'aggregates were too full to accomodate the cache volume';
        } else {
            my $text = <<'HEREDOC';
A suggested mechanism for disabling testcases that do use flexcache on any of
these low memory platforms is to add a code snippet like this to the
beginning of the (TCD-based) test-case:
    sub testcase {
        my $reason;
        if (!$node->is_feature_supported(
            feature    => 'flexcache',
            reason_ref => \$reason
        )) {
            $Test->nlog($reason);
            return $TCD::NA;
        }
        ... Rest of the test case here ...
    }
HEREDOC

            $error_text = "$reason\n$text";
        }

        $Log->exit() if $may_exit;
        NATE::BaseException->throw($error_text);
    }

    $pkg->_process_options(options => \%primary_keys);
    my $self = $pkg->new(%primary_keys,
        command_interface => $opts{command_interface});

    $Log->exit() if $may_exit;
    return $self;
} ## end sub create

sub _create_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_base_backend_cmode_cli(@args, api => 'volume_flexcache_create');
    $Log->exit() if $may_exit;
} ## end sub _create_cmode_cli

sub _create_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_base_backend_cmode_zapi(
        @args,
        api  => 'flexcache-create',
        copy => [qw(origin-volume vserver)]
    );
    $Log->exit() if $may_exit;
} ## end sub _create_cmode_zapi

=head2 reset

   NACL::C::VolumeFlexcache->reset(command_interface => $ci);

(Class method)
This method is used to reset all flexcache tables.

Uses a CMode CLI APISet.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< '*' >>

This method takes all other options supported by
"volume flexcache reset" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub reset {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params              => \@args,
        ignore_primary_keys => 1,
        cli_cmd             => 'volume flexcache reset',
    );
    $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_reset_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::VolumeFlexcache::reset',
    );

    $Log->exit() if $may_exit;
} ## end sub reset

sub _reset_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_base_backend_cmode_cli(@args, api => 'volume_flexcache_reset');
    $Log->exit() if $may_exit;
} ## end sub _reset_cmode_cli


__PACKAGE__->_load_event_class();

sub _primary_keys_validate_spec {
    return (
        'origin-volume' => {type => SCALAR},
        'vserver'       => {type => SCALAR},
        'cache-volume'  => {type => SCALAR | UNDEF, optional => 1},
    );
} ## end sub _primary_keys_validate_spec

1;
