# $Id$

#
# Copyright (c) 2001-2012 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary VolumeMove Component Module
## @author Akshay Mohit(makshay@netapp.com), dl-nacl-dev@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::C::VolumeMove

=head1 DESCRIPTION

C<NACL::C::VolumeMove> is a derived class of
L<NACL::C::Component::ONTAP|lib-NACL-C-Component-ONTAP-pm>. It models the volume move
operation. There is currently no support for find and state, because CMode
does not currently provide any mechanism of seeing the state of a volume
move operation.

=head1 ATTRIBUTES

=head2 command_interface

See L<NACL::C::Component::command_interface|lib-NACL-C-Component-pm/command_interface>

=head2 vserver

The vserver on which the source volume is present.

=head2 volume

The source volume to be moved.

=cut

package NACL::C::VolumeMove;

use strict;
use warnings;

use NACL::ComponentUtils qw (_optional_scalars Dumper);

use base 'NACL::C::Component::ONTAP';
use NACL::C::Job;
use NACL::CS::VolumeMove;
use Params::Validate qw(validate_with SCALAR SCALARREF UNDEF CODEREF);
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use NACL::C::Exceptions::VolumeMove::IsAFlexClone;
use NACL::C::Exceptions::VolumeMove::VolumeNotOnline;
use NACL::C::Exceptions::VolumeMove::VolumeMoveAlreadyStarted;
use NATE::BaseException qw(:try);
use NATE::Exceptions::Argument;
use NACL::Exceptions::InvalidChoice;

=head1 Linkage from Volume Component

All of the methods in this package are callable with a Volume Component object,
by using the "move_" prefix. For example, if we want to invoke the C<start>
method of C<NACL::C::VolumeMove>, we can do:

    my $vol_move_c_obj = $vol_c_obj->move_start(%opts);

The volume name, vserver name (for CMode) and command_interface are extracted
from the volume object, so they do not need to be provided as options
in the method call. So for example if we'd like to start a volume
move in the foreground, we could simply do:

    my $vol_move_c_obj = $vol_c_obj->move_start(
        'destination-aggregate' => $aggr_name,
        foreground              => 'true'
    );

=head1 METHODS

=head2 new

  my $VolumeMove= NACL::C::VolumeMove->new(
                                   command_interface  => $ci,
                                   volume             => $vol,
                                   vserver            => $vserver
                                 );

Return a new VolumeMove component object with the given attributes.

=cut

use Class::MethodMaker [
    scalar => 'volume',
    scalar => 'vserver',
];

=head2 trigger_cutover

    NACL::C::VolumeMove->trigger_cutover(
                                command_interface => $ci,
                                volume            => $vol,
                                vserver           => $vserver
                                ...);

                    or

    $VolumeMove_obj->trigger_cutover(...);

(Class method or instance method)

Uses a CMode CLI/ZAPI, 7Mode CLI APISet.

Invokes "volume-move-trigger-cutover" API for CMode ZAPI.

Invokes "vol move cutover" command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< volume => $string >>

(Required for class method, Not Applicable for instance method)
Source volume to be moved.

Applicable for CMode CLI/ZAPI,7Mode CLI.

Maps to:

CM ZAPI: source-volume

7M CLI : volume name (Programmatic name: "volume")

=item C<< vserver => $string >>

(Required for class method, Not Applicable for instance method)
Vserver on which the source volume resides.

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: vserver

=item C<< force => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: force

=item C<< cutover-window => $string >>

Applicable for 7Mode CLI.

Maps to:

7M CLI : cutover window (Programmatic name: "cutover-window")

 =item C<< "extended_query" => $string >>

(Optional)

Applicable for CMode CLI.

=item C<< '*' >>

This method takes all other options supported by
"volume move trigger-cutover" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub trigger_cutover {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw(force cutover-window

                    volume vserver
                    )
            )
        },
        cli_cmd => 'volume move trigger-cutover',
    );
    $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_trigger_cutover_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_trigger_cutover_7mode_cli',
                interface => 'CLI',
                set       => '7Mode',
            },
            {   method    => '_trigger_cutover_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::VolumeMove::trigger_cutover',
    );

    $Log->exit() if $may_exit;
} ## end sub trigger_cutover

sub _trigger_cutover_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;

    $pkg_or_obj->_base_backend_cmode_cli(
        @args,
        api => 'volume_move_trigger_cutover',
    );

    $Log->exit() if $may_exit;
} ## end sub _trigger_cutover_cmode_cli

sub _trigger_cutover_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_base_backend_cmode_zapi(
        @args,
        copy => [qw(vserver force)],
        map  => {volume => 'source-volume'},
        api  => 'volume-move-trigger-cutover',
    );
    $Log->exit() if $may_exit;
} ## end sub _trigger_cutover_cmode_zapi

sub _trigger_cutover_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );

    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};

    my $allowed_7mode_options = [qw( volume cutover-window )];

    my %mode7_args;
    $pkg->_hash_copy(
        source => \%opts,
        copy   => $allowed_7mode_options,
        target => \%mode7_args,
    );

    my $response = $apiset->vol_move_cutover(%mode7_args);
    $Log->exit() if $may_exit;
} ## end sub _trigger_cutover_7mode_cli

=head2 start

    my $VolumeMove_object= NACL::C::VolumeMove->start(
                                command_interface => $ci,
                                volume            => $vol,
                                vserver           => $vs,
                                  ...);

(Class method)
This method is used to start moving a volume from one aggregate to another aggregate.
It returns a new VolumeMove component object.

Uses a CMode CLI/ZAPI, 7Mode CLI APISet.

Invokes "volume-move-start" API for CMode ZAPI.

Invokes "vol move start" command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< volume => $volume >>

(Required)
The volume to be moved.

Applicable for CMode CLI/ZAPI,7Mode CLI.

Maps to:

CM ZAPI: source-volume

7M CLI : volume name (Programmatic name: "volume")

=item C<< vserver => $vserver >>

(Required)
The vserver on which the source volume is present.

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: vserver

=item C<< "destination-aggregate" => $string >>

(CMode) Destination aggreagate to which the volume is moved.

Applicable for CMode CLI/ZAPI,7Mode CLI.

Maps to:

CM ZAPI: dest-aggr

7M CLI : destination path (Programmatic name: "destination")

=item C<< "destination-volume" => $string >>

Applicable for CMode CLI.

=item C<< "cutover-action" => $string >>

Action for Cutover
Possible values: abort_on_failure; defer_on_failure; force; wait

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: cutover-action

=item C<< "cutover-window" => $string >>

Cutover time window in seconds
Valid range: 30 to 300

Applicable for CMode CLI/ZAPI,7Mode CLI.

Maps to:

CM ZAPI: cutover-window

7M CLI :  -w <value> (Programmatic name: "cutover-window")

=item C<< "tiering-policy" => $string >>

Tiering Policy
Possible Values: snapshot-only; none; backup

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: tiering-policy

=item C<< "perform-validation-only" => $string >>

Performs validation checks only
Possible values: true; false

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: perform-validation-only

=item C<< "destination-stripe-format" => $string >>

Destination Stripe Format
Possible values: data; full

Applicable for CMode CLI.

=item C<< "cutover-attempts" => $string >>

Number of cutover attempts
Valid range: 1 to 25

Applicable for CMode CLI/ZAPI,7Mode CLI.

Maps to:

CM ZAPI: cutover-attempts

7M CLI : -r <value> (Programmatic name: "cutover-attempts")

=item C<< "force-cutover" => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: force-cutover

=item C<< "keep-source" => $string >>

Applicable for CMode CLI/ZAPI,7Mode CLI.

Maps to:

CM ZAPI: is-keep-source

7M CLI :  -k <value> (Programmatic name: "keep-source")

=item C<< "manual-cut-over" => $string >>

Applicable for CMode CLI/ZAPI,7Mode CLI.

Maps to:

CM ZAPI: is-manual-cutover

7M CLI : -M (Programmatic name: "manual-cut-over")

=item C<< "ignore-redundancy" => $string >>

Applicable for CMode CLI.

=item C<< foreground => $string >>

Foreground process
Possible values: true; false

Applicable for CMode CLI.

=item C<< "method-timeout"=> $scalar >>

(Optional) The new timeout value.

=item C<< job_component => \$scalar >>

This is a reference to a variable, which gets filled in with a job component
if the operation was performed through a background job. We can then wait
on the completion of the job by doing

  my $job;
  NACL::C::VolumeMove->start(..., job_component => \$job);
  # $job will remain as undef if no job was started.
  # Check if $job was filled in and then wait for it.
  if ($job) {
      $job->watch_progress('method-timeout' => $time_it_might_take_in_seconds);
  }

=item C<< "destination" => $string >>

Applicable for 7Mode CLI

=item C<< "extended_query" => $string >>

(Optional)

Applicable for CMode CLI.

=item C<< 'repository-operation' => $string >>

Repository Operation possible value(s) are, true,false
Applicable for CMode CLI

=item C<< 'bypass-throttling' => $string >>

Bypass Replication Engine Throttling possible value(s) are, true,false
Applicable for CMode CLI

=item C<< 'skip-delta-calculation' => $string >>

Skip the Delta Calculation possible value(s) are, true,false
Applicable for CMode CLI

=item C<< '*' >>

This method takes all other options supported by
"volume move start" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=item C<NACL::C::Exceptions::VolumeMove::IsAFlexClone>

This exception is thrown when we attempt to perform volume move start operation
on a flexclone volume.

=item C<NACL::C::Exceptions::VolumeMove::VolumeNotOnline>

This exception is thrown when we attempt to perform volume move start operation
on an offlined volume.


=back

=back

=cut

sub start {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            job_component => {type => SCALARREF, optional => 1},
            _optional_scalars(
                qw( destination-volume cutover-action cutover-window
                    perform-validation-only destination-stripe-format cutover-attempts
                    force-cutover foreground destination-aggregate keep-source
                    manual-cut-over ignore-redundancy

                    repository-operation

                    bypass-throttling skip-delta-calculation
                    )
            ),

            # Needed for the STask. External callers should NOT use these
            # directly in the call to ->start since these are subject to
            # be changed/removed
            'connectrec-max_idle' => {type => SCALAR,  optional => 1},
            'connectrec-on_idle'  => {type => CODEREF, optional => 1},
        },
        api => 'volume_move_start',
    );

    try {
        $pkg->call_on_apiset(
            %opts,
            choices => [
                {   method    => '_start_cmode_cli',
                    interface => 'CLI',
                    set       => 'CMode',
                },
                {   method    => '_start_cmode_zapi',
                    interface => 'ZAPI',
                    set       => 'CMode',
                    check     => '_foreground_zapi_check',
                },
                {   method    => '_start_7mode_cli',
                    interface => 'CLI',
                    set       => '7Mode',
                },
            ],
            frontend => 'NACL::C::VolumeMove::start',
        );
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = $_[0];
        foreach my $type (qw( IsAFlexClone VolumeNotOnline VolumeMoveAlreadyStarted )) {
            my $full_type = "NACL::C::Exceptions::VolumeMove::$type";
            $full_type->detect_convert_and_throw(exception => $exception);
        }

        $Log->exit() if $may_exit;
        $exception->throw();
    };

    my %new_args = ();
    $pkg->_hash_copy(
        source => \%opts,
        target => \%new_args,
        copy   => [qw(command_interface volume)]
    );
    if ($opts{command_interface}->is_cmode()) {
        $new_args{vserver} = $opts{vserver};
    }

    $pkg->_process_options(options => \%new_args);

    my $self = $pkg->new(%new_args);
    $Log->exit() if $may_exit;
    return $self;
} ## end sub start

sub _start_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, %opts) = @_;

    my %additional_args;
    $pkg->_hash_move(
        source => \%opts,
        target => \%additional_args,
        move   => [qw(connectrec-max_idle connectrec-on_idle)]
    );
    $opts{additional_args} = {%additional_args} if (keys %additional_args);

    $pkg->_base_backend_cmode_cli(%opts, api => 'volume_move_start');

    $Log->exit() if $may_exit;
} ## end sub _start_cmode_cli

sub _start_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;

    $pkg->_base_backend_cmode_zapi(
        @args,
        api  => 'volume-move-start',
        copy => [
            qw(cutover-action vserver tiering-policy cutover-window perform-validation-only
                cutover-attempts force-cutover encrypt-destination)
        ],
        map => {
            'destination-aggregate' => 'dest-aggr',
            'volume'                => 'source-volume',
            'keep-source'           => 'is-keep-source',
            'manual-cut-over'       => 'is-manual-cutover',
        },
    );

    $Log->exit() if $may_exit;
} ## end sub _start_cmode_zapi

sub _start_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );

    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};

    my $allowed_7mode_options = [
        qw( volume cutover-window keep-source destination
            cutover-attempts manual-cut-over )
    ];
    my %mode7_args;
    $pkg->_hash_copy(
        source => \%opts,
        copy   => $allowed_7mode_options,
        map    => {'destination-aggregate' => "destination"},
        target => \%mode7_args,
    );

    my $response = $apiset->vol_move_start(%mode7_args);
    $Log->exit() if $may_exit;
} ## end sub _start_7mode_cli

=head2 pause

   NACL::C::VolumeMove->pause(
                         command_interface => $ci,
                         volume            => $vol,
                         ...);

   or

   $obj->pause();

(Class or Instance method )
This method is used to pause moving a volume from one aggregate to another aggregate.

Uses a CMode CLI/ZAPI, 7Mode CLI APISet.

Invokes "volume-move-pause" API for CMode ZAPI.

Invokes "vol move pause" command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< volume => $volume >>

(Required)
The volume to be paused.

Applicable for CMode CLI/ZAPI,7Mode CLI.

=item C<< vserver => $vserver >>

(Required)
The vserver on which the volume to be paused is present.

Applicable for CMode CLI/ZAPI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub pause {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(params => \@args);
    my $obj = $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_pause_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_pause_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
            {   method    => '_pause_7mode_cli',
                interface => 'CLI',
                set       => '7Mode',
            },
        ],
    );

    $Log->exit() if $may_exit;
    return $obj;
} ## end sub pause

sub _pause_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@args,  api => 'volume_move_pause');
    $Log->exit() if $may_exit;
} ## end sub _pause_cmode_cli

sub _pause_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_base_backend_cmode_zapi(@args,
            api    => 'volume_move_pause',
            copy   => [qw (vserver volume)],
    );
    $Log->exit() if $may_exit;
} ## end sub _pause_cmode_zapi

sub _pause_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );

    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};

    my $allowed_7mode_options = [qw( volume )];
    my %mode7_args;
    $pkg->_hash_copy(
        source => \%opts,
        copy   => $allowed_7mode_options,
        target => \%mode7_args,
    );

    my $response = $apiset->vol_move_pause(%mode7_args);
    $Log->exit() if $may_exit;
} ## end sub _pause_7mode_cli

=head2 resume

   NACL::C::VolumeMove->resume(
                         command_interface => $ci,
                         volume            => $vol,
                         ...);

   or

   $obj->resume( cutover-window => 30 );

(Class or Instance method )
This method is used to resume moving a volume from one aggregate to another aggregate.

Uses a CMode CLI/ZAPI, 7Mode CLI APISet.

Invokes "volume-move-resume" API for CMode ZAPI.

Invokes "vol move resume" command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< volume => $volume >>

(Required)
The volume to be resume.

Applicable for CMode CLI/ZAPI,7Mode CLI.

=item C<< keep-source => $boolean >>

Applicable for 7Mode CLI.

=item C<< manual-cut-over => $boolean >>

Applicable for 7Mode CLI.

=item C<< cutover-attempts => $integer >>

Applicable for 7Mode CLI.

=item C<< cutover-window => $integer >>

Applicable for 7Mode CLI.

=item C<< ignore-redundancy => $boolean >>

Applicable for 7Mode CLI.

=item C<< vserver => $vserver >>

(Required)
The vserver on which the volume to be resume is present.

Applicable for CMode CLI/ZAPI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub resume {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params => \@args,
	additional_spec => {
            _optional_scalars(
                qw( keep-source manual-cut-over cutover-attempts
                    cutover-window ignore-redundancy )
            ),
        },
	cli_cmd => 'volume move resume',
    );
    my $obj = $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_resume_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_resume_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
            {   method    => '_resume_7mode_cli',
                interface => 'CLI',
                set       => '7Mode',
            },
        ],
    );

    $Log->exit() if $may_exit;
    return $obj;
} ## end sub resume

sub _resume_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@args,  api => 'volume_move_resume');
    $Log->exit() if $may_exit;
} ## end sub _resume_cmode_cli

sub _resume_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_base_backend_cmode_zapi(@args,
            api    => 'volume_move_resume',
            copy   => [qw (vserver volume)],
    );
    $Log->exit() if $may_exit;
} ## end sub _resume_cmode_zapi

sub _resume_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );

    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};

    my $allowed_7mode_options = [
        qw( volume keep-source manual-cut-over cutover-attempts
            cutover-window ignore-redundancy )
    ];
    my %mode7_args;
    $pkg->_hash_copy(
        source => \%opts,
        copy   => $allowed_7mode_options,
        target => \%mode7_args,
    );

    $Log->debug(
        "Opts to API call 'vol_move_resume' are: \n" . Dumper(\%mode7_args));

    my $response = $apiset->vol_move_resume(%mode7_args);
    $Log->exit() if $may_exit;
} ## end sub _resume_7mode_cli

=head2 abort

   NACL::C::VolumeMove->abort(
                         command_interface => $ci,
                         volume            => $vol,
                         ...);

   or

   $obj->abort();

(Class or Instance method )
This method is used to abort moving a volume from one aggregate to another aggregate.

Uses a CMode/7Mode CLI, CMode ZAPI APISet.

Invokes "vol move abort" command for 7Mode CLI.

Invokes "volume-move-trigger-abort" API for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< volume => $volume >>

(Required)
The volume to be aborted.

Applicable for CMode/7Mode CLI, CMode ZAPI

Maps to:

CM ZAPI: source-volume

7M CLI : volume name (Programmatic name: "volume")

=item C<< vserver => $vserver >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: vserver

=item C<< '*' >>

This method takes all other options supported by
"volume move abort" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub abort {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params  => \@args,
        cli_cmd => 'volume move abort',
    );
    my $obj = $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_abort_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_abort_7mode_cli',
                interface => 'CLI',
                set       => '7Mode',
            },
            {   method    => '_abort_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
                check     => '_check_zapi_option',
            },
        ],
        frontend => 'NACL::C::VolumeMove::abort',
    );

    $Log->exit() if $may_exit;
    return $obj;
} ## end sub abort

sub _abort_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@args,  api => 'volume_move_abort');
    $Log->exit() if $may_exit;
} ## end sub _abort_cmode_cli

sub _abort_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );

    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};

    my $allowed_7mode_options = [qw( volume )];
    my %mode7_args;
    $pkg->_hash_copy(
        source => \%opts,
        copy   => $allowed_7mode_options,
        target => \%mode7_args,
    );

    my $response = $apiset->vol_move_abort(%mode7_args);
    $Log->exit() if $may_exit;
} ## end sub _abort_7mode_cli

sub _abort_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_base_backend_cmode_zapi(@args,
            api    => 'volume_move_trigger_abort',
            copy   => [qw (vserver)],
            map    => {'volume' => 'source-volume',},
    );
    $Log->exit() if $may_exit;
} ## end sub _abort_cmode_zapi

=head2 modify

    NACL::C::VolumeMove->modify(
                                command_interface => $ci,
                                'volume' => $volume,
                                'vserver' => $vserver,
                                ...);

                                 or
    $VolumeMove_obj->modify(...);

(Class or instance method)
This method is used to modify parameters for a running volume move operation.

Uses a CMode CLI/ZAPI APISet.

Ivokes "volume-move-modify" API for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'volume' => $volume >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI.

Maps to : 

CM ZAPI : 'source-volume'

=item C<< 'vserver' => $vserver >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI.

Maps to :

CM ZAPI : 'vserver'

=item C<< 'cutover-action' => $string >>

Applicable for CMode CLI/ZAPI.

Maps to :

CM ZAPI : 'cutover-action'

=item C<< 'cutover-attempts' => $string >>

Applicable for CMode CLI/ZAPI.

Maps to :

CM ZAPI : 'cutover-attempts'

=item C<< 'cutover-window' => $string >>

Applicable for CMode CLI.

=item C<< 'method-timeout' => $time_in_seconds >>

(Optional)
The default NATE timeout is 60 seconds. If the command might take more
time to complete, this option can be used to specify a larger timeout value.
The value should be provided in seconds.

=item C<< "extended_query" => $string >>

(Optional)

Applicable for CMode CLI.

=item C<< '*' >>

This method takes all other options supported by
"volume move modify" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub modify {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( cutover-action cutover-attempts cutover-window
                    )
            ),
        },
        cli_cmd => 'volume move modify',
    );
    $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_modify_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_modify_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::VolumeMove::modify',
    );

    $Log->exit() if $may_exit;
} ## end sub modify

sub _modify_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@args,  api => 'volume_move_modify');
    $Log->exit() if $may_exit;
} ## end sub _modify_cmode_cli

sub _modify_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my $allowed_cmode_options
        = [qw( cutover-action  cutover-attempts vserver cutover-window )];
    $pkg_or_obj->_base_backend_cmode_zapi(@args,
            api    => 'volume_move_modify',
            copy   => $allowed_cmode_options,
            map    => {'volume' => 'source-volume'},
    );
    $Log->exit() if $may_exit;
}

sub _check_zapi_option {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    if (exists $opts{'volume'} && $opts{'volume'} =~/\| || \,/i ) {
        my $msg = 'Could not use ZAPI implementation because field '
              . "'source-volume' cannot accept fields separated by comma(,)".
              "or pipe(|) for the ZAPI command 'volume-move-trigger-abort'";
       $Log->comment($msg);
       $Log->exit() if $may_exit;
      NACL::Exceptions::InvalidChoice->throw($msg);
   }
  $Log->exit() if $may_exit;
}

__PACKAGE__->_load_event_class();

sub _primary_keys_validate_spec {
    return (
        vserver => {type => SCALAR | UNDEF, optional => 1},
        volume  => {type => SCALAR}
    );
} ## end sub _primary_keys_validate_spec
1;
