# $Id: //depot/prod/test/nacldev/lib/NACL/C/VolumeQtree.pm#30 $

#
# Copyright (c) 2001-2012 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary VolumeQtree Component Module
## @author Dheeraj.K@netapp.com, pragyan@netapp.com, dl-nacl-dev@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::C::VolumeQtree

=head1 DESCRIPTION

C<NACL::C::VolumeQtree> is a derived class of
L<NACL::C::Component::ONTAP|lib-NACL-C-Component-ONTAP-pm>.  It represents access
to an ONTAP VolumeQtree.  A related class is
L<NACL::CS::VolumeQtree|lib-NACL-CS-VolumeQtree-pm>, which represents
the past state of an ONTAP VolumeQtree.

=head1 ATTRIBUTES

=head2 command_interface

See L<NACL::C::Component::command_interface|lib-NACL-C-Component-pm/command_interface>

=head2 vserver

(Required for CMode-CLI, Not Applicable for 7Mode-CLI)
The name of the vserver, in which the qtree is created

=head2 volume

=head2 qtree

=head2 qtree_path

=cut

package NACL::C::VolumeQtree;

use strict;
use warnings;

use NACL::CS::VolumeQtree;
use NACL::ComponentUtils qw (_optional_scalars Dumper);

use base 'NACL::C::Component::ONTAP';
use NACL::C::Job;
use Params::Validate
    qw(validate_with SCALAR ARRAYREF HASHREF UNDEF SCALARREF OBJECT);
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use NATE::Exceptions::Argument qw(:try);
use NACL::C::_Mixins::VolumeQtree qw(:all);
use NACL::C::Exceptions::VolumeQtree::AlreadyExists ();
use NACL::C::Exceptions::VolumeQtree::DoesNotExist  ();
use NACL::C::Exceptions::VolumeQtree::QtreeContainsLuns ();
use NATE::BaseException qw(:try);

=head1 METHODS

=head2 new

  my $qtree = NACL::C::VolumeQtree->new(command_interface=>$ci,
                                  vserver => $vserver_name,
                                  volume=> $volume_name,
                                  qtree => $qtree);

Return a new VolumeQtree component object with the given attributes.

=cut

use Class::MethodMaker [
    scalar => 'volume',
    scalar => 'vserver',
    scalar => 'qtree',
    scalar => [
        {   '-default_ctor' => sub {
                if (defined $_[0]->volume() && defined $_[0]->qtree()) {
                    my $path
                        = "/vol/" . $_[0]->volume() . "/" . $_[0]->qtree();
                    return $path;
                }
            },
        },
        'qtree_path'
    ],
];

sub isa {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, $kind) = @_;
    my $isa = $pkg_or_obj->_build_isa(
        kind  => $kind,
        alias => 'NACL::C::Qtree'
    );
    $Log->exit() if $may_exit;
    return $isa;
} ## end sub isa

=head2 cast_to_directory

    my $dir_obj = $qtree_obj->cast_to_directory();

This method is used to cast the Qtree Component object to Volume Directory
Component object as qtree and directory are similar for some operations.
So, in some cases qtree can be treated as directory.

Note that if invoked on a NACL::C::Qtree object, a NACL::C::VolumeDirectory
object is returned.

=cut

sub cast_to_directory {
    $Log->enter() if $may_enter;
    my $self    = $_[0];
    my $dir_pkg = "NACL::C::VolumeDirectory";
    my %pk_hash = $self->get_pk_hash_with_ci();
    my $volume = delete $pk_hash{volume};
    my $qtree  = delete $pk_hash{qtree};
    $pk_hash{path} = "/vol/" . $volume . "/" . $qtree;
    my $dir_obj = $dir_pkg->new(%pk_hash);
    $Log->exit() if $may_exit;
    return $dir_obj;
} ## end sub cast_to_file

=head2 create

  my $qtree  = NACL::C::VolumeQtree->create(command_interface=>$ci,
                                      vserver => $vserver_name,
                                      volume=> $volume_name,
                                      qtree => $qtree);

(Class method) Request creation of a new qtree element, and  return a new
qtree component object that refers to it.

Uses a CMode CLI/ZAPI, 7Mode CLI APISet.

Invokes "qtree-create" API for CMode ZAPI.

Invokes "qtree create" command for 7Mode CLI.

Invokes "qtree oplocks" command for 7Mode CLI if $opts{'oplock-mode'} exists.

Invokes "qtree security" command for 7Mode CLI if $opts{'security-style'}
  exists.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< vserver=>$vserver_name >>

(Required for CMode-CLI, Not Applicable for 7Mode-CLI)
Name of vserver in which the qtree is created.

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: ZAPI will be invoked in the context of the vserver specified to this
             option.

=item C<< volume=>$volume_name >>

(Required) Name of volume in which the qtree is created.

Applicable for CMode CLI.

=item C<< qtree=>$qtree_name >>

(Required) Name of qtree created.

Applicable for CMode CLI.

=item C<< "security-style" => $string >>

Applicable for CMode CLI/ZAPI, 7Mode CLI.

Maps to:

CM ZAPI: security-style

7M CLI : securiyt style (Programmatic name: "style") of
         "qtree security" command

=item C<< "oplock-mode" => $string >>

Applicable for CMode CLI, 7Mode CLI.

Maps to:

7M CLI : oplock state (Programmatic name: "state") of
         "qtree oplock" command

=item C<< "unix-permissions" => $string >>

Applicable for CMode CLI,7Mode CLI.

Maps to:

7M CLI : -m <value> (Programmatic name: "unix-permissions")

=item C<< 'oplock-mode-zapi' => $string >>

Applicable for CMode CLI.

=item C<< 'qtree-path' => $string >>

Applicable for CMode CLI/ZAPI,7Mode CLI.

Maps to:

CM ZAPI: qtree-path

7M CLI : path name (Programmatic name: "path")

=item C<< 'oplocks' => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: oplock-mode

=item C<< 'mode' => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: unix-permissions

=item C<< "extended_query" => $string >>

(Optional)

Applicable for CMode CLI.

=item C<< 'export-policy' => $string >>

Export Policy
Applicable for CMode CLI

=item C<< '*' >>

This method takes all other options supported by
"volume qtree create" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=item C<NACL::C::Exceptions::VolumeQtree::AlreadyExists>

This type of exception gets thrown if an attempt is made to create a
qtree that already exists.

=back

=back

=cut

sub create {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw (security-style oplock-mode unix-permissions
                    oplock-mode-zapi qtree-path
                    export-policy
                    )
            )
        },
        api => 'volume_qtree_create',
    );

    my $vserver;
    $opts{vserver_operated_on} = \$vserver;

    try {
        $pkg->call_on_apiset(
            %opts,
            choices => [
                {   method    => '_create_cmode_cli',
                    interface => 'CLI',
                    set       => 'CMode',
                },
                {   method    => '_create_7mode_cli',
                    interface => 'CLI',
                    set       => '7Mode',
                },
                {   method    => '_create_cmode_zapi',
                    interface => 'ZAPI',
                    set       => 'CMode',
                },
            ],
            frontend => 'NACL::C::VolumeQtree::create',
        );
    } ## end try
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = $_[0];
        NACL::C::Exceptions::VolumeQtree::AlreadyExists
            ->detect_convert_and_throw(exception => $exception);
        $Log->exit() if $may_exit;
        $exception->throw();
    };
    my $target = $pkg->_hash_copy(
        source => \%opts,
        copy   => [$pkg->get_primary_keys()]
    );

    $pkg->_process_options(options => $target);
    $Log->exit() if $may_exit;
    return $pkg->new(%$target, command_interface => $opts{command_interface});
} ## end sub create

sub _create_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, %opts) = @_;
    $pkg->_base_backend_cmode_cli(
        %opts,
        api             => 'volume_qtree_create',
        frontend_method => 'create',
    );
    $Log->exit() if $may_exit;
} ## end sub _create_cmode_cli

sub _create_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg               = shift;
    my %opts              = $pkg->_validate_backend_opts(@_);
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};

    $pkg->_construct_qtree_path(args_ptr => \%opts);

    my $qtree_args = $pkg->_hash_copy(
        source => \%opts,
        map    => {
            "unix-permissions" => "mode",
            "qtree-path"       => "path"
        },
    );

    $apiset->qtree_create(%$qtree_args);

    if (exists($opts{'oplock-mode'})) {
        $apiset->qtree_oplocks(
            'path'  => $qtree_args->{path},
            'state' => $opts{'oplock-mode'}
        );
    } ## end if ( exists( $opts{'oplock-mode'...}))

    if (exists($opts{'security-style'})) {
        $apiset->qtree_security(
            'path'  => $qtree_args->{path},
            'style' => $opts{'security-style'}
        );
    } ## end if ( exists( $opts{'security-style'...}))

    $Log->exit() if $may_exit;
} ## end sub _create_7mode_cli

sub _create_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg, %opts) = @_;
    if (defined $opts{'oplock-mode'}) {
        if ($opts{'oplock-mode'} eq "enable") {
            $opts{'oplock-mode'} = "enabled";
        } elsif ($opts{'oplock-mode'} eq "disable") {
            $opts{'oplock-mode'} = "disabled";
        }
    }
    $pkg->_base_backend_cmode_zapi(
        %opts,
        api  => 'qtree-create',
        copy => [qw( security-style volume qtree export-policy)],
        map  => {
            'oplock-mode'      => 'oplocks',
            'unix-permissions' => 'mode'
        },
        frontend_method => 'create'
    );
    $Log->exit() if $may_exit;

} ## end sub _create_cmode_zapi

=head2 delete

  NACL::C::VolumeQtree->delete(command_interface=>$ci,
                         vserver => $vserver_name,
                         volume => $volume_name,
                         qtree => $qtree_name);

                            or

  $qtree_object->delete(...);

(Class or instance method) This method deletes an qtree element.
Note that for CMode, this method defaults to deleting asynchronously (i.e. it
deletes the qtree in a background job). To make the deletion happen
synchronously, then it is necessary to provide foreground => 'true'.

    NACL::C::VolumeQtree->delete(%Other_opts, foreground => 'true');

    $qtree_obj->delete(foreground => 'true');

Uses a CMode CLI/ZAPI, 7Mode CLI APISet.

Invokes "qtree-delete-async" API for CMode ZAPI (asynchronous -- this is the
default); invokes "qtree-delete" for CMode ZAPI if foreground => 'true' is
provided.

Invokes "qtree delete" command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< vserver=>$vserver_name >>

(Required for class method, Not Applicable for instance method)
(Required for CMode-CLI, Not Applicable for 7Mode-CLI)
Name of vserver that contains the qtree.

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: ZAPI will be invoked in the context of the vserver specified to this
         option.

=item C<< force => $boolean >>

(Optional) A boolean when true will do a forceful delete of the qtree
component.

Applicable for CMode CLI/ZAPI,7Mode CLI.

Maps to:

CM ZAPI: force

7M CLI : -f (Programmatic name: "force")

=item C<< foreground => $boolean >>

Applicable for CMode CLI.

=item C<< "method-timeout"=> $scalar >>

(Optional)The new timeout value.

=item C<< "job_component"=> \$scalar >>

=item C<< 'qtree-path' => $string >>

Applicable for CMode CLI/ZAPI,7Mode CLI.

Maps to:

CM ZAPI: qtree-path

7M CLI : path name (Programmatic name: "path")

=item C<< "extended_query" => $string >>

(Optional)

Applicable for CMode CLI.

=item C<< '*' >>

This method takes all other options supported by
"volume qtree delete" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=item C<NACL::C::Exceptions::VolumeQtree::DoesNotExist>

This type of exception gets thrown if an attempt is made to delete a
non-existing qtree.

=back

=back

=cut

sub delete {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            job_component => {type => SCALARREF, optional => 1},
            _optional_scalars(
                qw (force foreground
                    qtree-path
                    )
            )
        },
        api => 'volume_qtree_delete',
    );

    my $ret;
    try {
        $ret = $pkg_or_obj->call_on_apiset(
            %opts,
            choices => [
                {   method    => '_delete_cmode_cli',
                    interface => 'CLI',
                    set       => 'CMode',
                },
                {   method    => '_delete_7mode_cli',
                    interface => 'CLI',
                    set       => '7Mode',
                },
                {   method    => '_delete_cmode_zapi',
                    interface => 'ZAPI',
                    set       => 'CMode',
                },
            ],
            frontend => 'NACL::C::VolumeQtree::delete',
        );
    } ## end try
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = $_[0];
        foreach my $type (
            qw(DoesNotExist QtreeContainsLuns)
            )
        {
            my $full_type = "NACL::C::Exceptions::VolumeQtree::$type";
            $full_type->detect_convert_and_throw(
                exception => $exception);
        }
        $Log->exit() if $may_exit;
        $exception->throw();
    };
    $Log->exit() if $may_exit;
    return $ret;
} ## end sub delete

sub _delete_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    my $vserver = $opts{vserver};

    $pkg_or_obj->_base_backend_cmode_cli(
        %opts,
        api => 'volume_qtree_delete',
        vserver_operated_on => \$vserver,
    );

    $Log->exit() if $may_exit;
} ## end sub _delete_cmode_cli

sub _delete_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg_or_obj        = shift;
    my %opts              = $pkg_or_obj->_validate_backend_opts(@_);
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};

    $pkg_or_obj->_construct_qtree_path(args_ptr => \%opts);

    my $qtree_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => [qw(force)],
        map    => {"qtree-path" => "path"},
    );

    %opts = %$qtree_args;
    $apiset->qtree_delete(%opts);
    $Log->exit() if $may_exit;
} ## end sub _delete_7mode_cli

sub _delete_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts                = $pkg_or_obj->_validate_backend_opts(@args);
    $pkg_or_obj->_construct_qtree_path(args_ptr => \%opts);
    my $api;
    if (defined $opts{foreground} && $opts{foreground} eq 'true') {
        $api = 'qtree-delete';
    } else {
        $api = 'qtree-delete-async';
    }
    my $vserver;
    $pkg_or_obj->_base_backend_cmode_zapi(
        %opts,
        api  => $api,
        copy => [qw(force)],
        map  => {'qtree-path' => 'qtree'},
        vserver_operated_on => \$vserver,
    );
    $Log->exit() if $may_exit;
} ## end sub _delete_cmode_zapi

=head2 modify

  NACL::C::VolumeQtree->modify(command_interface=>$ci,
                         vserver => $vserver_name,
                         volume  => $volume_name,
                         qtree   => $volume_name );

                            or

  $qtree_object->modify(...);

(Class or instance method) Modify a qtree element.

Uses a CMode CLI/ZAPI, 7Mode CLI APISet.

Invokes "qtree oplocks" command for 7Mode CLI if $opts{'state'} exists.

Invokes "qtree security" command for 7Mode CLI if $opts{'style'} exists.

Invokes 'qtree-modify' API for CMode ZAPI.

For CMode, the capabilities of the "volume qtree modify" command are available.
For 7Mode, "modify" only knows so far how to modify
the "security" attribute of the qtree element, setting it to be unix|mixed|ntfs,
and the "oplocks" attribute of the qtree element, setting it to enable|disable.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< vserver=>$vserver_name >>

(Required for class method, Not Applicable for instance method.
Required for CMode, Not Applicable for 7Mode).
Name of vserver that contains the qtree.

Applicable for CMode CLI/ZAPI.

=item C<< "security-style" => $string >>

Applicable for CMode CLI/ZAPI, 7Mode CLI.

Maps to:

7M CLI : security style (Programmatic name: "style")

CM ZAPI:

<qtree-modify>
<security-style>$value</security-style>
</qtree-modify>

=item C<< "oplock-mode" => $string >>

Applicable for CMode CLI, 7Mode CLI.

Maps to:

7M CLI : oplock state (Programmatic name: "state")

=item C<< "unix-permissions" => $string >>

Applicable for CMode CLI.

=item C<< 'qtree-path' => $string >>

Applicable for CMode CLI/ZAPI, 7Mode CLI.

Maps to:

7M CLI : path name (Programmatic name: "path")

CM ZAPI:

<qtree-modify>
<qtree-path>$value</qtree-path>
</qtree-modify>

=item C<< "extended_query" => $string >>

(Optional)

Applicable for CMode CLI.

=item C<< 'export-policy' => $string >>

Export Policy
Applicable for CMode CLI

=item C<< '*' >>

This method takes all other options supported by
"volume qtree modify" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub modify {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw (security-style oplock-mode
                    unix-permissions qtree-path
                    export-policy
                    )
            )
        },
        api => 'volume_qtree_modify',
    );

    my $vserver;
    $opts{vserver_operated_on} = \$vserver;

    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_modify_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_modify_7mode_cli',
                interface => 'CLI',
                set       => '7Mode',
            },
            {   method    => '_modify_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            }
        ],
        frontend => 'NACL::C::VolumeQtree::modify',
    );
    $Log->exit() if $may_exit;
    return $ret;
} ## end sub modify

sub _modify_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(
        %opts,
        api             => 'volume_qtree_modify',
        frontend_method => 'modify'
    );
    $Log->exit() if $may_exit;
} ## end sub _modify_cmode_cli

sub _modify_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg_or_obj        = shift;
    my %opts              = $pkg_or_obj->_validate_backend_opts(@_);
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};

    $pkg_or_obj->_construct_qtree_path(args_ptr => \%opts);

    my $qtree_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        map    => {
            "oplock-mode"    => "state",
            "security-style" => "style",
            "qtree-path"     => "path",
        },
    );

    %opts = %$qtree_args;

    # Modifying the oplocks
    if (exists $opts{'state'}) {
        $apiset->qtree_oplocks(
            'path'  => $opts{'path'},
            'state' => $opts{'state'}
        );
    } ## end if ( exists $opts{'state'...})

    # Modifying the security style
    if (exists $opts{'style'}) {
        $apiset->qtree_security(
            'path'  => $opts{'path'},
            'style' => $opts{'style'}
        );
    } ## end if ( exists $opts{'style'...})
    $Log->exit() if $may_exit;
} ## end sub _modify_7mode_cli

sub _modify_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;

    if (defined $opts{'oplock-mode'}) {
        if ($opts{'oplock-mode'} eq "enable") {
            $opts{'oplock-mode'} = "enabled";
        } elsif ($opts{'oplock-mode'} eq "disable") {
            $opts{'oplock-mode'} = "disabled";
        }
    }

    $pkg_or_obj->_base_backend_cmode_zapi(
        %opts,
        api  => 'qtree-modify',
        copy => [qw( security-style volume qtree export-policy)],
        map  => {
            'oplock-mode'      => 'oplocks',
            'unix-permissions' => 'mode'
        },
        frontend_method => 'modify',
    );
    $Log->exit() if $may_exit;
} ## end sub _modify_cmode_zapi

=head2 oplocks

   NACL::C::VolumeQtree->oplocks(
                                command_interface => $ci,
                                 volume => $volume,
                                 qtree => $qtree,
                                 vserver => $vserver,
                                  ...);

                                 or

    $VolumeQtree_object->oplocks(...);

(Class or instance method)
This method is used to modify qtree oplock mode.

Uses a CMode CLI, 7Mode CLI APISet.

Invokes "qtree oplocks" command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< vserver=>$vserver_name >>

(Required for class method, Not Applicable for instance method)
(Required for CMode-CLI, Not Applicable for 7Mode-CLI)
Name of vserver that contains the qtree.

Applicable for CMode CLI.

=item C<< oplock-mode => $string >>

Applicable for CMode CLI, 7Mode CLI.

Maps to:

7M CLI : oplock state (Programmatic name: "state")

=item C<< qtree-path=>$string >>

Applicable for CMode CLI, 7Mode CLI.

Maps to:

7M CLI : path name (Programmatic name: "path")

=item C<< "extended_query" => $string >>

(Optional)

Applicable for CMode CLI.

=item C<< '*' >>

This method takes all other options supported by
"volume qtree oplocks" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub oplocks {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( oplock-mode qtree-path
                    )
            ),
        },
        cli_cmd => 'volume qtree oplocks',
    );
    my $vserver;
    $opts{vserver_operated_on} = \$vserver;
    $Log->exit() if $may_exit;

    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_oplocks_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_oplocks_7mode_cli',
                interface => 'CLI',
                set       => '7Mode',
            },
        ],
        frontend => 'NACL::C::VolumeQtree::oplocks',
    );
    return $ret;
} ## end sub oplocks

sub _oplocks_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@args,  api => 'volume_qtree_oplocks');
    $Log->exit() if $may_exit;
} ## end sub _oplocks_cmode_cli

sub _oplocks_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg_or_obj        = shift;
    my %opts              = $pkg_or_obj->_validate_backend_opts(@_);
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};

    $pkg_or_obj->_construct_qtree_path(args_ptr => \%opts);

    my $qtree_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        map    => {
            "oplock-mode" => "state",
            "qtree-path"  => "path"
        },
    );

    %opts = %$qtree_args;
    $apiset->qtree_oplocks(%opts);
    $Log->exit() if $may_exit;
} ## end sub _oplocks_7mode_cli

=head2 rename

   NACL::C::VolumeQtree->rename(
                                command_interface => $ci,
                                 volume => $volume,
                                 qtree => $qtree,
                                 vserver => $vserver,
                                  ...);

                                 or

    $VolumeQtree_object->rename(...);

(Class or instance method)
This method is used to rename an existing qtree.

Uses a CMode CLI/ZAPI, 7Mode CLI APISet.

Invokes "qtree-rename" API for CMode ZAPI.

Invokes "qtree rename" command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< vserver=>$vserver_name >>

(Required for class method, Not Applicable for instance method)
(Required for CMode-CLI, Not Applicable for 7Mode-CLI)
Name of vserver that contains the qtree.

Applicable for CMode CLI/ZAPI.

Maps to:

  CM ZAPI: ZAPI will be invoked in the context of the vserver specified to this
           option.

=item C<< newname=>$string >>

Applicable for CMode CLI/ZAPI,7Mode CLI.

Maps to:

CM ZAPI: newname

=item C<< 'qtree-path' => $string >>

=item C<< 'qtree' => $string >>

Qtree Name possible value(s) are, qtree name
Applicable for CMode CLI

=item C<< 'volume' => $string >>

Volume Name possible value(s) are, volume name
Applicable for CMode CLI

=item C<< '*' >>

This method takes all other options supported by
"volume qtree rename" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

Applicable for CMode CLI/ZAPI,7Mode CLI.

Maps to:

CM ZAPI: qtree-path

7M CLI : path name (Programmatic name: "path")

=item C<< "extended_query" => $string >>

(Optional)

Applicable for CMode CLI.

=back

=back

=cut

sub rename {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( newname qtree-path
                    )
            ),
        },
        cli_cmd => 'volume qtree rename',
    );
    my $vserver;
    $opts{vserver_operated_on} = \$vserver;
    $Log->exit() if $may_exit;

    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_rename_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_rename_7mode_cli',
                interface => 'CLI',
                set       => '7Mode',
            },
            {   method    => '_rename_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::VolumeQtree::rename',
    );

    if (ref($pkg_or_obj)) {
        $pkg_or_obj->volume($opts{'volume'});
        $pkg_or_obj->qtree($opts{'newname'});
    }
    return $ret;
} ## end sub rename

sub _rename_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@args,  api => 'volume_qtree_rename');
    $Log->exit() if $may_exit;
} ## end sub _rename_cmode_cli

sub _rename_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg_or_obj        = shift;
    my %opts              = $pkg_or_obj->_validate_backend_opts(@_);
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};

    $pkg_or_obj->_construct_qtree_path(args_ptr => \%opts);

    my $qtree_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => [qw(newname )],
        map    => {"qtree-path" => "path"},
    );

    my $newname = delete $qtree_args->{newname};
    $qtree_args->{srcpath} = delete $qtree_args->{path};

    $qtree_args->{dstpath}
        = _construct_new_qtree_path($qtree_args->{srcpath}, $newname);
    $apiset->qtree_rename(%$qtree_args);
    $Log->exit() if $may_exit;
} ## end sub _rename_7mode_cli

sub _rename_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj)      = shift ;
    my %opts              = $pkg_or_obj->_validate_backend_opts(@_);
    $pkg_or_obj->_construct_qtree_path(args_ptr => \%opts);

    $opts{'newname'} =  _construct_new_qtree_path($opts{'qtree-path'},$opts{'newname'});
    $pkg_or_obj->_base_backend_cmode_zapi(
        %opts,
        api  => 'qtree-rename',
        map  => {
            'newname'    => 'new-qtree-name',
            'qtree-path' => 'qtree',
        },
        
        # Input contains volume + qtree, but only qtree-path is valid,
        # so skip looking up help_xml for these two fields.
        inapplicable_for_help_xml => [qw(volume qtree)],
    );
    $Log->exit() if $may_exit;
} ## end sub _rename_cmode_zapi

=head2 security

   NACL::C::VolumeQtree->security(
                                command_interface => $ci,
                                 volume => $volume,
                                 qtree => $qtree,
                                 vserver => $vserver,
                                  ...);

                                 or

    $VolumeQtree_object->security(...);

(Class or instance method)
This method is used to modify qtree security style.

Uses a CMode CLI, 7Mode CLI APISet.

Invokes "qtree security" command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< vserver=>$vserver_name >>

(Required for class method, Not Applicable for instance method)
(Required for CMode-CLI, Not Applicable for 7Mode-CLI)
Name of vserver that contains the qtree.

Applicable for CMode CLI.

=item C<< security-style => $string >>

Applicable for CMode CLI, 7Mode CLI.

Maps to:

7M CLI : security style (Programmatic name: "style")

=item C<< 'qtree-path' => $string >>

=item C<< 'qtree' => $string >>

Qtree Name possible value(s) are, qtree name
Applicable for CMode CLI

=item C<< 'volume' => $string >>

Volume Name possible value(s) are, volume name
Applicable for CMode CLI

Applicable for CMode CLI, 7Mode CLI.

Maps to:

7M CLI : path name (Programmatic name: "path")

=item C<< "extended_query" => $string >>

(Optional)

Applicable for CMode CLI.

=item C<< '*' >>

This method takes all other options supported by
"volume qtree security" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub security {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( security-style qtree-path
                    )
            ),
        },
        cli_cmd => 'volume qtree security',
    );
    $Log->exit() if $may_exit;

    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_security_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_security_7mode_cli',
                interface => 'CLI',
                set       => '7Mode',
            },
        ],
        frontend => 'NACL::C::VolumeQtree::security',
    );
    return $ret;
} ## end sub security

sub _security_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    my $vserver = $opts{vserver};

    $pkg_or_obj->_base_backend_cmode_cli(
        %opts,
        api => 'volume_qtree_security',
        vserver_operated_on => \$vserver,
    );

    $Log->exit() if $may_exit;
} ## end sub _security_cmode_cli

sub _security_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg_or_obj        = shift;
    my %opts              = $pkg_or_obj->_validate_backend_opts(@_);
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};

    $pkg_or_obj->_construct_qtree_path(args_ptr => \%opts);

    my $qtree_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        map    => {
            "security-style" => "style",
            "qtree-path"     => "path"
        },
    );

    %opts = %$qtree_args;
    $apiset->qtree_security(%opts);
    $Log->exit() if $may_exit;
} ## end sub _security_7mode_cli

=head2 statistics_reset

    NACL::C::VolumeQtree->statistics_reset(
                                command_interface => $ci,
                                'volume' => $volume,
                                'vserver' => $vserver,
                                ...);

(Class method)
This method is used to reset qtree statistics in a volume.

Uses a CMode CLI APISet.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'volume' => $volume >>

(Required)

Applicable for CMode CLI.

=item C<< 'vserver' => $vserver >>

(Required)

Applicable for CMode CLI.

=item C<< 'method-timeout' => $time_in_seconds >>

(Optional)
The default NATE timeout is 60 seconds. If the command might take more
time to complete, this option can be used to specify a larger timeout value.
The value should be provided in seconds.

=item C<< "extended_query" => $string >>

(Optional)

Applicable for CMode CLI.

=item C<< '*' >>

This method takes all other options supported by
"volume qtree statistics-reset" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub statistics_reset {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( qtree-path
                    )
            ),
        },
        cli_cmd => 'volume qtree statistics-reset',
    );
    $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_statistics_reset_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::VolumeQtree::statistics_reset',
    );

    $Log->exit() if $may_exit;
} ## end sub statistics_reset

sub _statistics_reset_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, %opts) = @_;
    my $vserver = $opts{vserver};

    $pkg->_base_backend_cmode_cli(
        %opts,
        api => 'volume_qtree_statistics_reset',
        vserver_operated_on => \$vserver,
    );

    $Log->exit() if $may_exit;
} ## end sub _statistics_reset_cmode_cli

=head2 statistics

    my $statistics_obj = NACL::C::VolumeQtree->statistics(command_interface => $ci,
                          [ filter => { %filter } ,]
              [ requested_fields => [ @req_fields ]);

                    or

    my @statistics_objs = NACL::C::VolumeQtree->statistics(command_interface => $ci,
                          [ filter => { %filter } ,]
              [ requested_fields => [ @req_fields ]);

                    or

   my $statistics_obj = $VolumeQtree_obj->statistics();

(Class or Instance method)
This method returns objects of the type L<NACL::CS::VolumeQtreeStatistics|lib-NACL-CS-VolumeQtreeStatistics-pm>.

This method accepts all the parameters accepted by L<fetch|lib-NACL-CS-ComponentState-pm/fetch>
(requested_fields, filter, allow_empty) and has all of the other properties of
fetch methods (callable in scalar or list context, throws a NoElementsFound
exception if no matching elements are found)

=cut

sub statistics {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = $pkg_or_obj->_handle_obj_cs_like_calls(@_);
    require NACL::CS::VolumeQtreeStatistics;
    my @state_objs = NACL::CS::VolumeQtreeStatistics->fetch(%opts);
    $Log->exit() if $may_exit;
    return wantarray ? @state_objs : $state_objs[0];
} ## end sub statistics

=head2 stats

    my $stats_obj = NACL::C::VolumeQtree->stats(command_interface => $ci,
                          [ filter => { %filter } ,]
			  [ requested_fields => [ @req_fields ]);

                    or

    my @stats_objs = NACL::C::VolumeQtree->stats(command_interface => $ci,
                          [ filter => { %filter } ,]
			  [ requested_fields => [ @req_fields ]);

                    or

   my $stats_obj = $VolumeQtree_obj->stats();

(Class or Instance method)

This method is a CS-like fetch method which accepts all the parameters of
L<fetch|lib-NACL-CS-ComponentState-pm/fetch> (requested_fields, filter, allow_empty)
and has all of the other properties of fetch methods (callable in scalar or list
context, throws a NoElementsFound exception if no matching elements are found).
The requested fields and filter takes L<ATTRIBUTES|lib-NACL-CS-VolumeQtreeStats-pm/ATTRIBUTES>
of NACL::CS::VolumeQtreeStats as input.
This method returns objects of the type L<NACL::CS::VolumeQtreeStats|lib-NACL-CS-VolumeQtreeStats-pm>.

=cut

sub stats {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts = $pkg_or_obj->_handle_obj_cs_like_calls(@args);
    require NACL::CS::VolumeQtreeStats;
    my @state_objs = NACL::CS::VolumeQtreeStats->fetch(%opts);
    $Log->exit() if $may_exit;
    return wantarray ? @state_objs : $state_objs[0];
} ## end sub stats

=head2 status

    my $status_obj = NACL::C::VolumeQtree->status(command_interface => $ci,
                          [ filter => { %filter } ,]
			  [ requested_fields => [ @req_fields ]);

                    or

    my @status_objs = NACL::C::VolumeQtree->status(command_interface => $ci,
                          [ filter => { %filter } ,]
			  [ requested_fields => [ @req_fields ]);

                    or

   my $status_obj = $VolumeQtree_obj->status();

(Class or Instance method)

This method is a CS-like fetch method which accepts all the parameters of
L<fetch|lib-NACL-CS-ComponentState-pm/fetch> (requested_fields, filter, allow_empty)
and has all of the other properties of fetch methods (callable in scalar or list
context, throws a NoElementsFound exception if no matching elements are found).
The requested fields and filter takes L<ATTRIBUTES|lib-NACL-CS-VolumeQtreeStatus-pm/ATTRIBUTES>
of NACL::CS::VolumeQtreeStatus as input.
This method returns objects of the type L<NACL::CS::VolumeQtreeStatus|lib-NACL-CS-VolumeQtreeStatus-pm>.

=cut

sub status {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts = $pkg_or_obj->_handle_obj_cs_like_calls(@args);
    require NACL::CS::VolumeQtreeStatus;
    my @state_objs = NACL::CS::VolumeQtreeStatus->fetch(%opts);
    $Log->exit() if $may_exit;
    return wantarray ? @state_objs : $state_objs[0];
} ## end sub status

=head2 prepare_for_downgrade

    NACL::C::VolumeQtree->prepare_for_downgrade(
                                command_interface => $ci,
                                ...);
   
(Class method)
This method is used to disable qtree exports.
Uses CMode CLI .

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< '*' >>

This method takes all other options supported by "volume qtree prepare for downgrade" command in ONTAP

=item C<< other_options >>

This method takes other common options described in L<Common method options|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#Other_options_valid_on_component_methods>

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub prepare_for_downgrade {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params              => \@args,
        ignore_primary_keys => 1,
        api                 => 'volume_qtree_prepare_for_downgrade',
    );
    $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_prepare_for_downgrade_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::VolumeQtree::prepare_for_downgrade',
    );

    $Log->exit() if $may_exit;
}

sub _prepare_for_downgrade_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, @opts) = @_;
    $pkg->_base_backend_cmode_cli(@opts,
        api => 'volume_qtree_prepare_for_downgrade',);
    $Log->exit() if $may_exit;
}


__PACKAGE__->_load_event_class();

sub _primary_keys_validate_spec {
    return (
        vserver  => {type => SCALAR | OBJECT | UNDEF, optional => 1},
        'volume' => {type => SCALAR | OBJECT},
        'qtree'  => {type => SCALAR | OBJECT},
    );
} ## end sub _primary_keys_validate_spec

sub _common_validate_with {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;

    my %opts = validate_with(
        params => \@args,
        spec   => {
            params              => {type => ARRAYREF},
            additional_spec     => {type => HASHREF, default => {}},
            ignore_primary_keys => {type => SCALAR, default => 0},
            allow_extra         => {type => SCALAR, default => 0},
        },
        allow_extra => 1
    );

    my %params = @{$opts{params}};

    $params{'qtree-path'} = $params{'qtree-path'}->qtree_path()
        if ref $params{'qtree-path'};

    if (exists $params{'qtree-path'}) {
        $pkg_or_obj->_convert_path_to_volume_qtree(
            command_interface => $params{command_interface},
            args_ptr          => \%params
        );
    } ## end if ( exists $params{'qtree-path'...})

    @{$opts{params}} = %params;

    %params = $pkg_or_obj->SUPER::_common_validate_with(%opts);

    $Log->exit() if $may_exit;
    return %params;
} ## end sub _common_validate_with

sub init {
    $Log->enter() if $may_enter;
    my ($self, %args) = @_;

    $self->_convert_path_to_volume_qtree(
        command_interface => $args{command_interface},
        args_ptr          => \%args
    );

    my @args = %args;
    my %opts = validate_with(
        params => \@args,
        spec   => {
            command_interface => {required => 1},
            $self->_primary_keys_validate_spec()
        },
        allow_extra => 1
    );

    $self->volume($opts{volume});
    $self->qtree($opts{qtree});
    $self->SUPER::init(%opts);
    $self->_vserver_context_init(%opts);
    $Log->exit() if $may_exit;
} ## end sub init

=head1 Linked components

(See
http://wikid.netapp.com/w/QA/projects/Libraries_Initiative/Project/Common_Infrastructure/Users_Guide/Component_Layer_Users_Guide#Linked_components
for a more detailed explanation on linked components)

This component is linked to the following components:

=begin html

<table border="1">
<tr> <th>Prefix</th> <th>Linked Component</th> </tr>
<tr> <td>lun</td> <td>Lun</td> </tr>
</table>

=end html

Examples:

    # Invoking the "delete" method of the Lun component
    $qtree_obj->lun_delete();

=cut

sub can {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my @opts = $pkg_or_obj->_can_validation(@args);
    my $code;
    if ($code = $pkg_or_obj->NACL::C::Component::can(@opts)) {
        $Log->exit() if $may_exit;
        return $code;
    } else {
        $code = $pkg_or_obj->_handle_component_aggregation(
            method    => $opts[0],
            error_ptr => $opts[1],
            table     => [
                {   component => 'NACL::C::Lun',
                    prefix    => 'lun'
                },
            ]
        );
        if ($code) {
            $Log->exit() if $may_exit;
            return $code;
        } elsif ($code = $pkg_or_obj->SUPER::can(@opts)) {
            $Log->exit() if $may_exit;
            return $code;
        }
    } ## end else [ if ( $code = $pkg_or_obj...)]

    $Log->exit() if $may_exit;
    return undef;
} ## end sub can

1;
