# $Id: $

#
# Copyright (c) 2001-2012 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary VolumeQuotaPolicy Component Module
## @author Dheeraj.K@netapp.com, dl-nacl-dev@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::C::VolumeQuotaPolicy

=head1 DESCRIPTION

C<NACL::C::VolumeQuotaPolicy> is a derived class of
L<NACL::C::Component::ONTAP|lib-NACL-C-Component-ONTAP-pm>.  It represents access to an
ONTAP Volume Quota Policy.  A related class is
L<NACL::CS::VolumeQuotaPolicy|lib-NACL-CS-VolumeQuotaPolicy-pm>, which represents
the past state of an ONTAP Volume Quota Policy.

=head1 ATTRIBUTES

=head2 command_interface

See L<NACL::C::Component::command_interface|lib-NACL-C-Component-pm/command_interface>

=head2 vserver

The name of the vserver, used to identify it uniquely.  (vserver names are
assumed to be unique within a node or within a cluster)

=head2 policy_name

The name of the quota policy you are creating.

=cut

package NACL::C::VolumeQuotaPolicy;

use strict;
use warnings;
use NACL::CS::VolumeQuotaPolicy ();
use NACL::ComponentUtils qw (_optional_scalars);

use base 'NACL::C::Component::ONTAP';
use Params::Validate qw(validate_with SCALAR);
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();

=head2 create

  my $quota_policy = NACL::C::VolumeQuotaPolicy->create(command_interface=>$ci,
                                                  vserver=>$vserver_name,
                          'policy-name'=>$policy_name,
                          ...);

(Class method) Request creation of a new Quota Policy element, and  return a
new Volume Quota Policy component object that refers to it.  Does not wait for Quota
Policy creation to complete.

Uses a CMode-CLI APIset.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< vserver=>$string >>

(Required) Name of vserver

Applicable for CMode CLI.

=item C<< policy-name => $string >>

(Required) This parameter specifies the name of the quota policy you are
creating.

Applicable for CMode CLI.

=item C<< "extended_query" => $string >>

(Optional)

Applicable for CMode CLI.

=item C<< '*' >>

This method takes all other options supported by
"volume quota policy create" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=cut

use Class::MethodMaker [
    scalar => 'vserver',
    scalar => 'policy_name',
];

sub init {
    $Log->enter() if $may_enter;
    my ($self, @args) = @_;
    $self->SUPER::init(@args);
    $self->_vserver_context_init(@args);
    $Log->exit() if $may_exit;
} ## end sub init

sub isa {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, $kind) = @_;
    my $isa = $pkg_or_obj->_build_isa(
        kind  => $kind,
        alias => 'NACL::C::QuotaPolicy'
    );
    $Log->exit() if $may_exit;
    return $isa;
} ## end sub isa

sub create {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    my %opts = $pkg->_common_validate_with(
        params   => \@args,
        cli_cmd  => 'volume quota policy create'
    );
    my $obj = $pkg->call_on_apiset_with_exceptions(
        %opts,
        choices => [
            {   method    => "_create_cmode_cli",
                interface => "CLI",
                set       => "CMode",
            },
        ],
        frontend        => 'NACL::C::VolumeQuotaPolicy::create',
        nacl_exceptions => [qw(AlreadyExists)]
    );
    $Log->exit() if $may_exit;
    return $obj;
} ## end sub create

sub _create_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, %opts) = @_;
    my $vserver = $opts{vserver};

    $pkg->_base_backend_cmode_cli(
        %opts,
        api => 'volume_quota_policy_create',
        vserver_operated_on => \$vserver,
    );

    $Log->exit() if $may_exit;
    return $pkg->new(
        vserver           => $opts{vserver},
        policy_name       => $opts{'policy-name'},
        command_interface => $opts{command_interface}
    );
} ## end sub _create_cmode_cli

=head2 delete

  my $quota_policy = NACL::C::VolumeQuotaPolicy->delete(command_interface=>$ci,
                                                  vserver=>$vserver_name,
                          'policy-name'=>$policy_name,
                              ...);

                            or

  $quota_policy_object->delete(...);

(Class or instance method) This method deletes a Quota Policy element.

Uses a CMode-CLI APIset.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< vserver=>$string >>

(Required for class method, Not Applicable for instance method)
(Required) Name of vserver

Applicable for CMode CLI.

=item C<< policy-name => $string >>

(Required for class method, Not Applicable for instance method)
(Required) This parameter specifies the name of the quota policy you are
creating.

Applicable for CMode CLI.

=item C<< "extended_query" => $string >>

(Optional)

Applicable for CMode CLI.

=item C<< '*' >>

This method takes all other options supported by
"volume quota policy delete" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub delete {
    my ($pkg_or_obj, @args) = @_;
    my %opts = $pkg_or_obj->_common_validate_with(
        params     => \@args,
        cli_cmd    => 'volume quota policy delete'
    );

    $pkg_or_obj->call_on_apiset_with_exceptions(
        %opts,
        choices => [
            {   method    => "_delete_cmode_cli",
                interface => "CLI",
                set       => "CMode",
            },
        ],
        frontend        => 'NACL::C::VolumeQuotaPolicy::delete',
        nacl_exceptions => [qw(DoesNotExist)],
    );
} ## end sub delete

sub _delete_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    my $vserver = $opts{vserver};

    $pkg_or_obj->_base_backend_cmode_cli(
        %opts,
        api => 'volume_quota_policy_delete',
        vserver_operated_on => \$vserver,
    );

    $Log->exit() if $may_exit;
} ## end sub _delete_cmode_cli

=head2 rename

  my $quota_policy = NACL::C::VolumeQuotaPolicy->rename(command_interface=>$ci,
                                                  vserver=>$vserver_name,
                                   'policy-name'=>$policy_name,
                          ...);

                            or

  $quota_policy_object->rename(...);

(Class or instance method) This method renames a Quota Policy element.

Uses a CMode-CLI APIset.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< vserver=>$string >>

(Required for class method, Not Applicable for instance method)
(Required) Name of vserver

Applicable for CMode CLI.

=item C<< policy-name => $string >>

(Required for class method, Not Applicable for instance method)
(Required) This parameter specifies the name of the quota policy.

Applicable for CMode CLI.

=item C<< new-policy-name => $string >>

(Required for class method, Not Applicable for instance method)
(Required) This parameter specifies the new name of the quota policy.

Applicable for CMode CLI.

=item C<< "extended_query" => $string >>

(Optional)

Applicable for CMode CLI.

=item C<< '*' >>

This method takes all other options supported by
"volume quota policy rename" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub rename {
    my ($pkg_or_obj, @args) = @_;
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {_optional_scalars(qw(new-policy-name))},
        cli_cmd         => 'volume quota policy rename',
    );

    return $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => "_rename_cmode_cli",
                interface => "CLI",
                set       => "CMode",
            },
        ],
        frontend => 'NACL::C::VolumeQuotaPolicy::rename',
    );
} ## end sub rename

sub _rename_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    my $vserver = $opts{vserver};

    $pkg_or_obj->_base_backend_cmode_cli(
        %opts,
        api => 'volume_quota_policy_rename',
        vserver_operated_on => \$vserver,
    );

    $pkg_or_obj->policy_name($opts{'new-policy-name'})
        if ref($pkg_or_obj);

    $Log->exit() if $may_exit;
} ## end sub _rename_cmode_cli

=head2 copy

  my $quota_policy = NACL::C::VolumeQuotaPolicy->copy(command_interface=>$ci,
                                                  vserver=>$vserver_name,
                          'policy-name'=>$policy_name,
                          ...);

                            or

  $quota_policy_object->copy(...);

(Class or instance method) This method copies a Quota Policy element.

Uses a CMode-CLI APIset.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< vserver=>$string >>

(Required for class method, Not Applicable for instance method)
(Required) Name of vserver

Applicable for CMode CLI.

=item C<< policy-name => $string >>

(Required for class method, Not Applicable for instance method)
(Required) This parameter specifies the name of the quota policy.

Applicable for CMode CLI.

=item C<< new-policy-name => $string >>

(Required) This parameter specifies the name of the new name of the quota
policy you are copying.

Applicable for CMode CLI.

=item C<< "extended_query" => $string >>

(Optional)

Applicable for CMode CLI.

=item C<< '*' >>

This method takes all other options supported by
"volume quota policy copy" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub copy {
    my ($pkg_or_obj, @args) = @_;
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {_optional_scalars(qw(new-policy-name))},
        cli_cmd         => 'volume quota policy copy',
    );

    return $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => "_copy_cmode_cli",
                interface => "CLI",
                set       => "CMode",
            },
        ],
        frontend => 'NACL::C::VolumeQuotaPolicy::copy',
    );
} ## end sub copy

sub _copy_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    my $vserver = $opts{vserver};

    $pkg_or_obj->_base_backend_cmode_cli(
        %opts,
        api => 'volume_quota_policy_copy',
        vserver_operated_on => \$vserver,
    );

    $Log->exit() if $may_exit;
    return NACL::C::VolumeQuotaPolicy->new(
        vserver           => $opts{'vserver'},
        policy_name       => $opts{'new-policy-name'},
        command_interface => $opts{command_interface},
    );
} ## end sub _copy_cmode_cli


__PACKAGE__->_load_event_class();

sub _primary_keys_validate_spec {
    return (
        vserver       => {type => SCALAR},
        'policy-name' => {type => SCALAR},
    );
} ## end sub _primary_keys_validate_spec

1;
