#
# Copyright (c) 2001-2017 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary VolumeSnapshot Component Module
## @author Akshay Mohit(makshay@netapp.com), chethan.m@netapp.com, pragyan@netapp.com,
##         dl-nacl-dev@netapp.com, benjaram@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::C::VolumeSnapshot

=head1 DESCRIPTION

C<NACL::C::VolumeSnapshot> is a derived class of
L<NACL::C::Component::ONTAP|lib-NACL-C-Component-ONTAP-pm>.  It represents access to an
ONTAP VolumeSnapshot.  A related class is
L<NACL::CS::VolumeSnapshot|lib-NACL-CS-VolumeSnapshot-pm>, which represents
the past state of an ONTAP VolumeSnapshot.

=head1 ATTRIBUTES

=head2 command_interface

See L<NACL::C::Component::command_interface|lib-NACL-C-Component-pm/command_interface>

=head2 snapshot

The name of the snapshot, used to identify it uniquely.  (VolumeSnapshot
names are assumed to be unique within a volume)

=cut

package NACL::C::VolumeSnapshot;

use strict;
use warnings;
use NACL::CS::VolumeSnapshot;
use NACL::ComponentUtils qw (_optional_scalars Dumper _dump_one);
use NACL::C::Vserver;

use base 'NACL::C::Component::ONTAP';
use Params::Validate qw(:all);
use NACL::C::Job;
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use NACL::C::Exceptions::VolumeSnapshot::AlreadyExists           ();
use NACL::C::Exceptions::VolumeSnapshot::DoesNotExist            ();
use NACL::C::Exceptions::VolumeSnapshot::InUse                   ();
use NACL::C::Exceptions::VolumeSnapshot::SnapRestoreNotLicensed  ();
use NACL::C::Exceptions::VolumeSnapshot::ClonesBackedBySnapshots ();
use NACL::C::Exceptions::VolumeSnapshot::SnapshotHasOwner        ();
use NACL::C::Exceptions::VolumeSnapshot::RetryCmd                ();
use NACL::C::Exceptions::VolumeSnapshot::SnapshotFenced          ();
use NACL::C::Exceptions::VolumeSnapshot::CurrentlyBusy           ();
use NATE::BaseException qw(:try);
use NATE::Exceptions::Argument ();
use NACL::Exceptions::InvalidChoice;

use NACL::C::_Mixins::Common qw(:iv_methods);

=head1 METHODS

=head2 new

  my $snapshot = NACL::C::VolumeSnapshot->new(command_interface=>$ci,
                                         snapshot =>$snapshot_name);

Return a new VolumeSnapshot component object with the given attributes.

=cut

use Class::MethodMaker [
    scalar => 'snapshot',
    scalar => 'volume',
    scalar => 'vserver',

    # Not a primary key, but used to determine which ZAPI to invoke.
    # Hence, is a private attribute and is undocumented.
    scalar => '_is_infinivol',

    # Snapshot operations not allowed in ZAPI for constituent volumes,
    # so set this to 1 if a constituent so subsequence calls can
    # skip ZAPI altogether.
    scalar => '_is_const_vol',
];

sub init {
    $Log->enter() if $may_enter;
    my ($self, @args) = @_;
    $self->SUPER::init(@args);
    $self->_vserver_context_init(@args);
    $Log->exit() if $may_exit;
} ## end sub init

sub isa {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, $kind) = @_;
    my $isa = $pkg_or_obj->_build_isa(
        kind  => $kind,
        alias => 'NACL::C::Snapshot'
    );
    $Log->exit() if $may_exit;
    return $isa;
} ## end sub isa

=head2 create

  my $snapshot = NACL::C::VolumeSnapshot->create(command_interface=>$ci,
                                           snapshot=>$snapshot_name,...);

(Class method) Request creation of a new snapshot element, and  return a new
snapshot component object that refers to it.  Does not wait for snapshot
creation to complete.

Uses a CMode CLI/ZAPI, 7Mode CLI APISet.

Invokes "snap create" command for 7Mode CLI.

For CM ZAPI:

Invokes "snapshot-create-async" API for infinite volumes

Invokes "snapshot-create" API for flexible volumes.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< snapshot=>$snapshot_name >>

(Required) Name of snapshot to create.

Applicable for CMode CLI/ZAPI,7Mode CLI.

Maps to:

CM ZAPI: snapshot

7M CLI : snapshot name (Programmatic name: "snapshot")

=item C<< volume => $string >>

(Required) Name of volume of which the snapshot to be created.

Applicable for CMode CLI/ZAPI,7Mode CLI.

Maps to:

CM ZAPI: volume

7M CLI : volume name (Programmatic name: "vol-name")

=item C<< vserver => $string >>

(Required) Name of vserver.

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: ZAPI will be invoked in the context of the vserver specified to this
option.

=item C<< comment => $string >>

(Optional) Comment.

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: comment

=item C<< foreground => $boolean >>

(Optional) Foreground Process (default: true).

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

async of 'snapshot-create' ZAPI.

Not applicable for 'snapshot-create-async' ZAPI.

=item C<< "method-timeout" => $scalar >>

(Optional)The new timeout value.

=item C<< "job_component"=> \$scalar >>

Filled in for CMode CLI/ZAPI.

CM ZAPI:

Applicable only for infinite volumes.

=item C<< 'test-version-uuid' => $string >>

Applicable for CMode CLI.

=item C<< 'cpcount-sort' => $string >>

Applicable for CMode CLI.

=item C<< 'snapmirror-label' => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: snapmirror-label

=item C<< 'is-repository' => $string >>

B<DEPRECATED>. Use C<is_infinivol> instead.

=item C<< is_infinivol => 0 | 1 >>

(Optional, Applicable for CMode ZAPI)

This boolean option describes whether the volume is an infinite volume or not.
For CMode ZAPI, the "async" form of the ZAPI must be used for the infinite
volume. If this option is not specified, the library will attempt to
determine if an infinite volume is being operated on (this could involve
querying the filer).

=item C<< is_infinivol_ref => \$scalar >>

(Optional, SCALARREF)

If the method was invoked through CMode ZAPI, the variable gets updated
describing whether it operated on an infinivol (used the "async" ZAPI) or not
(used the normal ZAPI).

=item C<< "extended_query" => $string >>

(Optional)

Applicable for CMode CLI.

=item C<< '*' >>

This method takes all other options supported by
"volume snapshot create" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<NACL::C::Exceptions::VolumeSnapshot::AlreadyExists>

This type of exception gets thrown if an attempt is made to create a
snapshot that already exists.

=item C<NACL::C::Exceptions::VolumeSnapshot::ClonesBackedBySnapshots>

This type of exception gets thrown if an attempt is made to create a
snapshot when a clone operation is running background for the volume.

=item C<NACL::C::Exceptions::VolumeSnapshot::SnapshotFenced>

This type of exception gets thrown if an attempt is made to create a
volume snapshot where volume is fenced temporarily.

=item C<NACL::C::Exceptions::VolumeSnapshot::CurrentlyBusy>

This type of exception gets thrown if an attempt is made to create
volume snapshot where volume is currently busy.

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub create {
    my ($pkg, @args) = @_;
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            job_component => {type => SCALARREF, optional => 1},
            _optional_scalars(
                qw (comment foreground
                    aggregate test-version-uuid cpcount-sort
                    snapmirror-label is-repository

                    )
            ),
        },
        api => 'volume_snapshot_create',
    );

    my $is_infinivol_ref;
    $opts{is_infinivol_ref} ||= \$is_infinivol_ref;

    my $const_vol_ref;

    try {
        $pkg->call_on_apiset(
            %opts,
            _is_const_vol_ref => \$const_vol_ref,
            choices           => [
                {   method    => '_create_cmode_cli',
                    interface => 'CLI',
                    set       => 'CMode',
                },
                {   method    => '_create_7mode_cli',
                    interface => 'CLI',
                    set       => '7Mode',
                },
                {   method    => '_create_cmode_zapi',
                    interface => 'ZAPI',
                    set       => 'CMode',
                },
            ],
            frontend => 'NACL::C::VolumeSnapshot::create',
        );
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        $pkg->_create_exception_handler(@_);
    };

    my %new_hash;
    $pkg->_copy_primary_keys(
        source => \%opts,
        target => \%new_hash
    );
    $pkg->_process_options(options => \%new_hash);
    my $self
        = $pkg->new(%new_hash, command_interface => $opts{command_interface});
    $self->_update_infinivol_val_in_self(%opts);
    $self->_is_const_vol($const_vol_ref) if $const_vol_ref;

    $Log->exit() if $may_exit;
    return $self;
} ## end sub create

sub _create_exception_handler {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, $exception) = @_;
    NACL::C::Exceptions::VolumeSnapshot::AlreadyExists
        ->detect_convert_and_throw(exception => $exception);
    NACL::C::Exceptions::VolumeSnapshot::ClonesBackedBySnapshots
        ->detect_convert_and_throw(exception => $exception);
    NACL::C::Exceptions::VolumeSnapshot::SnapshotFenced
        ->detect_convert_and_throw(exception => $exception);
    NACL::C::Exceptions::VolumeSnapshot::CurrentlyBusy
        ->detect_convert_and_throw(exception => $exception);
    $Log->exit() if $may_exit;
    $exception->throw();
}

sub _create_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_base_backend_cmode_cli(@args,
                  api => 'volume_snapshot_create',
    );
    $Log->exit() if $may_exit;
} ## end sub _create_cmode_cli

sub _create_7mode_cli {
    my ($pkg, @args) = @_;
    my %opts = validate_with(
        params      => \@args,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};

    my $snapshot_args = $pkg->_hash_copy(
        source => \%opts,
        copy   => [qw(snapshot aggregate)],
        map    => {"volume" => "vol-name",},
    );

    %opts = %$snapshot_args;
    $apiset->snap_create(%opts);

    return $pkg->new(
        snapshot          => $opts{snapshot},
        volume            => $opts{"vol-name"},
        command_interface => $command_interface
    );
} ## end sub _create_7mode_cli

sub _create_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg, %opts) = @_;

    # The CLI's "foreground" and the ZAPI's "async" are opposites, i.e
    # for foreground=true => async=false; foreground=false => async=true
    # Note that foreground='' is equivalent to foreground=true, so handle
    # that as well.
    if (defined $opts{foreground}) {
        if ($opts{foreground} eq 'true' || !length $opts{foreground}) {
            $opts{async} = 'false';
        } elsif ($opts{foreground} eq 'false') {
            $opts{async} = 'true';
        } else {
            # Invalid value provided for "foreground", let it go down to the
            # ZAPI and fail
            $opts{async} = $opts{foreground};
        }
    }
    $pkg->_backend_cmode_zapi_infinivol_builder(
        %opts,
        api       => 'snapshot-create',
        copy      => [qw(comment snapshot volume snapmirror-label)],
        non_async => {copy => [qw(async)]},
    );
    $Log->exit() if $may_exit;
} ## end sub _create_cmode_zapi

=head2 multicreate

   my @snapshot_objs =  NACL::C::VolumeSnapshot->multicreate(
                                command_interface => $ci,
                                cleanup => 'true',
                                snapshot => $old_snapshot_name,
                                vserver => $vserver,
                                'volume-names' => [ vol1 vol2 ],
                                  ...);

(Class  method)
This method is used to create snapshots for all specified volumes in volume-names.
Returns the array of SnapShot objects created.

Uses a CMode ZAPI APISet.

Invokes "snapshot-multicreate" API for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< 'snapshot' =>$snapshot_name >>

(Required for class method)
Name of a snapshot.

Applicable for CMode ZAPI.

Maps to:

CM ZAPI: snapshot

=item C<< cleanup => $boolean >>

(Optional default true  values: true|false)

(Required for class method)
This is a boolean option. If it is specified as true then cleanup of snapshots on volumes incase of any error will be handled by script.For time critical window manual removal of snapshot from volumes will be done with cleanup value equal to false.

Applicable for CMode ZAPI.

Maps to:

CM ZAPI: cleanup

=item C<< 'volume-names' => \$array >>

(Required for class method, Not Applicable for instance method)
(Required) Name of volumes for which the snapshot to be created.

Applicable for CMode ZAPI.

Maps to:

CM ZAPI: volume-names

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=item C<NACL::C::Exceptions::VolumeSnapshot::MultiOperationFailed>

This exception is thrown when we attempt to invoke any multi-api(create|delete|rename|validate) on  snapshot for multiple volumes and it got failed for one or more volumes.

=back

=back

=cut


sub multicreate {
    my ($pkg, @args) = @_;
    my %opts = $pkg->_common_validate_with(
    params          => \@args,
       additional_spec => {
           cleanup       => {type => BOOLEAN, optional => 1},
           snapshot      => {type => SCALAR},
           'volume-names'  => {type =>ARRAYREF},
       },
       ignore_primary_keys => 1,
       allow_extra         => 1,
    );

    my $vserver;
    $opts{vserver_operated_on} = \$vserver;

    $pkg->call_on_apiset(
           %opts,
           choices           => [
               {   method    => '_multicreate_cmode_zapi',
                   interface => 'ZAPI',
                   set       => 'CMode',
               },
            ],
            frontend => 'NACL::C::VolumeSnapshot::multicreate',
    );

    my @objs;
    foreach my $vol_name (@{$opts{'volume-names'}}) {
        push @objs, $pkg->new(
            command_interface => $opts{command_interface},
            vserver           => $opts{vserver},
            snapshot          => $opts{snapshot},
            volume            => $vol_name
        );
    }
    return @objs;

    $Log->exit() if $may_exit;

} ## end sub muticreate

sub _multicreate_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_multi_base_backend_cmode_zapi(
        @args,
        api       => 'snapshot-multicreate',
        copy      => [qw(cleanup snapshot volume-names)],
    );
    $Log->exit() if $may_exit;
} ## end sub _multicreate_cmode_zapi

=head2 multicreate_validate

   NACL::C::VolumeSnapshot->multicreate_validate(
                                command_interface => $ci,
                                snapshot => $snapshot_name,
                                vserver => $vserver,
                                'volume-names' => [ vol1 vol2 ],
                                  ...);

(Class  method)
This method is used to validate whether snapshots for all specified volumes in volume-names can be created or not.It does not give the guarantee that any error will not occur.

Uses a CMode ZAPI APISet.

Invokes "snapshot-multicreate-validate" API for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< 'snapshot' =>$snapshot_name >>

(Required for class method)
Name of a snapshot.

Applicable for CMode ZAPI.

Maps to:

CM ZAPI: snapshot

=item C<< 'volume-names' => [ vol1,vol2 ] >>

(Arrayref)
(Required for class method, Not Applicable for instance method)
(Required) Name of volumes for which the snapshot to be created.

Applicable for CMode ZAPI.

Maps to:

CM ZAPI: volume-names

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=item C<NACL::C::Exceptions::VolumeSnapshot::MultiOperationFailed>

This exception is thrown when we attempt to invoke any multi-api(create|delete|rename|validate) on  snapshot for multiple volumes and it got failed for one or more volumes.

=back

=back

=cut


sub multicreate_validate {
    my ($pkg, @args) = @_;
    my %opts = $pkg->_common_validate_with(
       params          => \@args,
       additional_spec => {
          snapshot     => { type => SCALAR },
          'volume-names' => { type => ARRAYREF },
       },
       ignore_primary_keys => 1,
       allow_extra =>1,
    );

    my $vserver;
    $opts{vserver_operated_on} = \$vserver;

    $pkg->call_on_apiset(
        %opts,
        choices           => [
            {   method    => '_multicreate_validate_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::VolumeSnapshot::multicreate_validate',
     );

     $Log->exit() if $may_exit;
     return;
}

sub _multicreate_validate_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_multi_base_backend_cmode_zapi(
            @args,
            api    => 'snapshot_multicreate_validate',
            copy   => [qw(snapshot volume-names )],
    );
    $Log->exit() if $may_exit;
} ## end sub _multicreate_validate_cmode_zapi

=head2 delete

  NACL::C::VolumeSnapshot->delete(command_interface=>$ci,
                             snapshot=>$snapshot_name,...);

                            or

  $snapshot_object->delete(...);

(Class or instance method) This method deletes an snapshot element.

Uses a CMode-CLI/ZAPI, 7Mode-CLI APISet.

Invokes "snap delete" command for 7Mode CLI.

CM ZAPI:

Invokes "snapshot-delete-async" API for infinite volumes

Invokes "snapshot-delete" API for flexible volumes

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< snapshot=>$snapshot_name >>

(Required for class method, Not Applicable for instance method)
Name of a snapshot.

Applicable for CMode CLI/ZAPI,7Mode CLI.

Maps to:

CM ZAPI: snapshot

7M CLI : snapshot (Programmatic name: "snapshot")

=item C<< volume => $string >>

(Required for class method, Not Applicable for instance method)
(Required) Name of volume of which the snapshot to be created.

Applicable for CMode CLI/ZAPI,7Mode CLI.

Maps to:

CM ZAPI: volume

7M CLI : volume name (Programmatic name: "vol-name")

=item C<< vserver => $string >>

(Required for class method, Not Applicable for instance method)
(Required for CMode-CLI, Not
Applicable for 7Mode-ZAPI or 7Mode-CLI)
Name of vserver.

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: ZAPI will be invoked in the context of the vserver specified to this
option.

=item C<< force => $boolean >>

(Optional) A boolean when true will do a forceful delete of the snapshot
component (default: false).

Applicable for CMode CLI, 7Mode CLI.

Maps to:

7M CLI : -f (Programmatic name: "force")

=item C<< foreground => $boolean >>

(Optional) Foreground Process (default: true).

Applicable for CMode CLI.

=item C<< ignore-owners => $boolean >>

(Optional) Ignore VolumeSnapshot Owners (default: false).

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: ignore-owners

=item C<< instance-uuid => $UUID >>

(Optional) UUID.

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: snapshot-instance-uuid

=item C<< cpcount-sort => $cpcount_sort >>

Applicable for CMode CLI.

=item C<< "method-timeout" => $scalar >>

(Optional)The new timeout value.

=item C<< "all" => $boolean >>

(Optional)Applicable to all.

Applicable for CMode CLI, 7Mode CLI.

Maps to:

7M CLI : -a (Programmatic name: "all")

=item C<< "quiet" => $boolean >>

(Optional).

Applicable for CMode CLI, 7Mode CLI.

Maps to:

7M CLI : -q (Programmatic name: "quiet")

=item C<< "job_component"=> \$scalar >>

Filled in for CMode CLI/ZAPI.

CM ZAPI:

Applicable only for infinite volumes.

=item C<< "job_components"=> \@array >>

Filled in for CMode CLI/ZAPI.

If multiple job ids are obtained, because of deleting multiple snapshots
by passing snapshot => '*', then this option would be populated
with the reference to the array of job component objects, if specified.

CM ZAPI:

Applicable only for infinite volumes.

=item C<< 'is-repository' => $string >>

B<DEPRECATED>. Use C<is_infinivol> instead.

=item C<< is_infinivol => 0 | 1 >>

(Optional, Applicable for CMode ZAPI)

This boolean option describes whether the volume is an infinite volume or not.
For CMode ZAPI, the "async" form of the ZAPI must be used for the infinite
volume. If this option is not specified, the library will attempt to
determine if an infinite volume is being operated on (this could involve
querying the filer).

=item C<< is_infinivol_ref => \$scalar >>

(Optional, SCALARREF)

If the method was invoked through CMode ZAPI, the variable gets updated
describing whether it operated on an infinivol (used the "async" ZAPI) or not
(used the normal ZAPI).

=item C<< "extended_query" => $string >>

(Optional)

Applicable for CMode CLI.

=item C<< '*' >>

This method takes all other options supported by
"volume snapshot delete" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<NACL::C::Exceptions::VolumeSnapshot::DoesNotExist>

This type of exception gets thrown if an attempt is made to delete a
non-existing snapshot.

=item C<NACL::C::Exceptions::VolumeSnapshot::InUse>

This type of exception gets thrown if an attempt is made to delete a
snapshot which is currently in use.

=item C<NACL::C::Exceptions::VolumeSnapshot::SnapshotFenced>

This type of exception gets thrown if an attempt is made to delete a
volume snapshot which is fenced temporarily.

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub delete {
    my ($pkg_or_obj, @args) = @_;
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            job_component  => {type => SCALARREF, optional => 1},
            job_components => {type => ARRAYREF,  optional => 1},
            _optional_scalars(
                qw(foreground force instance-uuid aggregate cpcount-sort
                    ignore-owners all force quiet snapshot is-repository)
            )
        },
        cli_cmd => 'volume snapshot delete',
    );

    my $ret;
    try {
        $ret = $pkg_or_obj->call_on_apiset(
            %opts,
            choices => [
                {   method    => '_delete_cmode_cli',
                    interface => 'CLI',
                    set       => 'CMode',
                },
                {   method    => '_delete_7mode_cli',
                    interface => 'CLI',
                    set       => '7Mode',
                },
                {   method    => '_delete_cmode_zapi',
                    interface => 'ZAPI',
                    set       => 'CMode',
                    check     => '_infinivol_vol_vs_check',
                },
            ],
            frontend => 'NACL::C::VolumeSnapshot::delete',
        );
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        $pkg_or_obj->_delete_exception_handler(@_);
    };

    $Log->exit() if $may_exit;
    return $ret;
} ## end sub delete

sub _delete_exception_handler {
    my ($pkg, $exception) = @_;
    foreach my $type (qw(DoesNotExist InUse SnapshotFenced)) {
        my $full_type = "NACL::C::Exceptions::VolumeSnapshot::$type";
        $full_type->detect_convert_and_throw(exception => $exception);
    }
    $Log->exit() if $may_exit;
    $exception->throw();
}

sub _delete_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    my $vserver = $opts{vserver};
    my $response;

    $pkg_or_obj->_base_backend_cmode_cli(
        %opts,
        api => 'volume_snapshot_delete',
        vserver_operated_on => \$vserver,
        apiset_response_obj => \$response,
    );

    $Log->exit() if $may_exit;
} ## end sub _delete_cmode_cli

sub _delete_7mode_cli {
    my ($pkg_or_obj, @args) = @_;
    my %opts = validate_with(
        params      => \@args,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my $allowed_7mode_options;

# To have the mode agnostisim, when '*' is specified as the snapshot name, it is considered
# as the user wants to delete all the snapshots created for the given volume
# in 7Mode CLI , when the option 'all' is specified, the option 'snapshot' should be knocked off
# ref: burt507172
    if (exists $opts{'snapshot'} && $opts{'snapshot'} eq '*') {
        $allowed_7mode_options = [qw(aggregate force all quiet)];
        $opts{'all'} = 1;
        $opts{force} = 1 if (!defined $opts{'force'});
    } else {
        if (!exists $opts{'snapshot'}) {
            $opts{'all'}           = 1;
            $opts{force}           = 1;
            $allowed_7mode_options = [qw(aggregate all force quiet)];
        } else {
            if (exists $opts{'force'} and exists $opts{'all'}) {

                # "snap delete -a -f vol-name snap-name" results in error.
                $allowed_7mode_options
                    = [qw(snapshot aggregate force all quiet)];
            } else {
                $allowed_7mode_options = [qw(snapshot aggregate quiet)];
            }
        } ## end else [ if ( !exists $opts{'snapshot'...})]
    } ## end else [ if ( exists $opts{'snapshot'...})]

    my $snapshot_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => $allowed_7mode_options,
        map    => {"volume" => "vol-name",},
    );
    %opts = %$snapshot_args;

    $apiset->snap_delete(%opts);
} ## end sub _delete_7mode_cli

sub _delete_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;

    delete $opts{force} if( defined $opts{force} );
    my $delete = sub {
        $pkg_or_obj->_backend_cmode_zapi_infinivol_builder(
            %opts,
            api  => 'snapshot-delete',
            copy => [qw(ignore-owners snapshot volume)],
            map  => {'instance-uuid' => 'snapshot-instance-uuid'},
            @_
        );
    };

    if ($opts{snapshot} eq '*') {
        my $pkg               = $pkg_or_obj->get_package_name();
        my @vol_snapshot_list = $pkg->find(
            command_interface => $opts{command_interface},
            filter            => {volume => $opts{volume}},
            apiset_should     => {interface => 'ZAPI'},
            allow_empty       => 1
        );

        my @job_ids;
        if (@vol_snapshot_list) {

            # Possibly multiple invocations of the ZAPI. Once the first call
            # is done, we'll know whether we're running on an infinivol or
            # not, so we can pass that in to successive calls.
            my $infinivol_ref;
            $opts{is_infinivol_ref} ||= \$infinivol_ref;

            foreach my $snap_c (@vol_snapshot_list) {
                if (  !defined $opts{is_infinivol}
                    && defined ${$opts{is_infinivol_ref}})
                {
                    $opts{is_infinivol} = ${$opts{is_infinivol_ref}};
                }

                my $response;
                $delete->(
                    snapshot            => $snap_c->snapshot(),
                    apiset_response_obj => \$response
                );
                push @job_ids, $response->get_job_ids();
            }
            if (@job_ids) {
                my @jobs;
                foreach my $id (@job_ids) {
                    push @jobs,
                        NACL::C::Job->new(
                        id                => $id,
                        command_interface => $opts{command_interface}
                        );
                }
                if ($opts{job_component}) {
                    ${$opts{job_component}} = $jobs[0];
                }
                if ($opts{job_components}) {
                    @{$opts{job_components}} = @jobs;
                }
            }
        } else {
            $delete->();
        }
    } else {
        $delete->();
    }

    $Log->exit() if $may_exit;

} ## end sub _delete_cmode_zapi

=head2 multidelete

   NACL::C::VolumeSnapshot->multidelete(
                                command_interface => $ci,
                                snapshot => $snapshot_name,
                                vserver => $vserver,
                                'volume-names' => [ vol1 vol2 ],
                                  ...);

(Class  method)
This method is used to delete snapshots for all specified volumes in volume-names.

Uses a CMode ZAPI APISet.

Invokes "snapshot-multidelete" API for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< 'snapshot' =>$snapshot_name >>

(Optional)
Name of a snapshot.

Applicable for CMode ZAPI.

Maps to:

CM ZAPI: snapshot

=item C<< 'volume-names' => \$array >>

(Required for class method, Not Applicable for instance method)
(Required) Name of volumes for which the snapshot to be created.

Applicable for CMode ZAPI.

Maps to:

CM ZAPI: volume-names

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=item C<NACL::C::Exceptions::VolumeSnapshot::MultiOperationFailed>

This exception is thrown when we attempt to invoke any multi-api(create|delete|rename|validate) on  snapshot for multiple volumes and it got failed for one or more volumes.

=back

=back

=cut

sub multidelete {
    my ($pkg, @args) = @_;
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            snapshot      => {type => SCALAR},
            'volume-names'  => {type =>ARRAYREF},
        },
        ignore_primary_keys => 1,
        allow_extra         => 1,
    ),
    my $vserver;
    $opts{vserver_operated_on} = \$vserver;

    my $ret = $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_multidelete_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
         ],
         frontend => 'NACL::C::VolumeSnapshot::delete',
     );

    $Log->exit() if $may_exit;
    return $ret;
} ## end sub multidelete

sub _multidelete_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg, %opts) = @_;
    $pkg->_multi_base_backend_cmode_zapi(
            %opts,
            api  => 'snapshot-multidelete',
            copy => [qw(snapshot volume-names)],
    );
    $Log->exit() if $may_exit;
} ## end sub _multidelete_cmode_zapi


=head2 restore_file

   NACL::C::VolumeSnapshot->restore_file(
                             command_interface => $ci,
                             volume => $volume,
                             vserver => $vserver,
                             snapshot => $snapshot,
                                  ...);

 or

    $VolumeSnapshot_object->restore_file(...);

(Class or instance method)
This method is used to restore a file from a snapshot.

Uses a CMode CLI/ZAPI, 7Mode-CLI APISet.

Invokes "snapshot-restore-file" API for CMode ZAPI.

Invokes "snap restore" command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< snapshot=>$snapshot_name >>

(Required for class method, Not Applicable for instance method)
Name of a snapshot.

Applicable for CMode CLI/ZAPI,7Mode CLI.

Maps to:

CM ZAPI: snapshot

7M CLI : -s <value> (Programmatic name: "snapshot-name")

=item C<< volume => $string >>

(Required for class method, Not Applicable for instance method)
(Required) Name of volume of which the snapshot to be created.

Applicable for CMode CLI/ZAPI,7Mode CLI.

Maps to:

CM ZAPI: volume

7M CLI : volume name (Programmatic name: "vol-name")

=item C<< vserver => $string >>

(Required for class method, Not Applicable for instance method)
(Required for CMode-CLI, Not
Applicable for 7Mode-CLI)
Name of vserver.

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: ZAPI will be invoked in the context of the vserver specified to this
             option.

=item C<< volume-dsid => $string >>

Applicable for CMode CLI.

=item C<< instance-uuid => $string >>

Applicable for CMode CLI/ZAPI,7Mode CLI.

Maps to:

CM ZAPI: snapshot-instance-uuid

=item C<< split-disabled => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: space-efficient-split-disabled

=item C<< restore-path => $string >>

Applicable for CMode CLI/ZAPI,7Mode CLI.

Maps to:

CM ZAPI: restore-path

7M CLI : restore path (Programmatic name: "restore-as-path")

=item C<< path => $string >>

Applicable for CMode CLI/ZAPI,7Mode CLI.

Maps to:

CM ZAPI: path

7M CLI : restore from path (Programmatic name: "restore-from-path")

=item C<< force => $boolean >>

Applicable for CMode CLI, 7Mode CLI.

Maps to:

7M CLI : -f (Programmatic name: "force")

=item C<< method-timeout => $string >>

(Optional)The new timeout value.

=item C<< "job_component"=> \$scalar >>

Filled in for CMode CLI.

=item C<< 'foreground' => $string >>

Applicable for CMode CLI.

=item C<< "extended_query" => $string >>

(Optional)

Applicable for CMode CLI.

=item C<< '*' >>

This method takes all other options supported by
"volume snapshot restore-file" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<NACL::C::Exceptions::VolumeSnapshot::CurrentlyBusy>

This type of exception gets thrown if an attempt is made to restore
volume snapshot where volume is currently busy

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub restore_file {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            job_component => {type => SCALARREF, optional => 1},
            _optional_scalars(
                qw( volume volume-dsid instance-uuid aggregate force
                    restore-path path split-disabled
                    foreground
                    )
            ),
        },
        cli_cmd => 'volume snapshot restore-file',
    );
    my $vserver;
    $opts{vserver_operated_on} = \$vserver;
    try {
        $pkg_or_obj->call_on_apiset(
            %opts,
            choices => [
                {   method    => '_restore_file_cmode_cli',
                    interface => 'CLI',
                    set       => 'CMode',
                },
                {   method    => '_restore_file_7mode_cli',
                    interface => 'CLI',
                    set       => '7Mode',
                },
                {   method    => '_restore_file_cmode_zapi',
                    interface => 'ZAPI',
                    set       => 'CMode',
                },
            ],
        frontend => 'NACL::C::VolumeSnapshot::restore_file',
        );
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = $_[0];
        NACL::C::Exceptions::VolumeSnapshot::CurrentlyBusy
            ->detect_convert_and_throw(exception => $exception);

        $Log->exit() if $may_exit;
        $exception->throw();
    };
    $Log->exit() if $may_exit;
} ## end sub restore_file

sub _restore_file_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    my $response;
    $pkg_or_obj->_base_backend_cmode_cli(%opts,
            api => 'volume_snapshot_restore_file',
            apiset_response_obj => \$response
    );
    $Log->exit() if $may_exit;
} ## end sub _restore_file_cmode_cli

sub _restore_file_7mode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts   = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = $opts{apiset};

    my $snapshot_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => [qw(force instance-uuid)],
        map    => {
            'snapshot'     => 'snapshot-name',
            'restore-path' => 'restore-as-path',
            'path'         => 'restore-from-path',
            'volume'       => 'vol-name',
        },
    );
    $snapshot_args->{type} = 'file';

    # "volume" and "restore-from-path" are mutually exclusive
    # If both are provided, we'll use "restore-from-path"
    if (   defined $snapshot_args->{'vol-name'}
        && defined $snapshot_args->{'restore-from-path'})
    {
        delete $snapshot_args->{'vol-name'};
    }

    my $prompt_answers = $pkg_or_obj->_restore_7mode_prompts();

    $apiset->snap_restore(%$snapshot_args,
        'connectrec-match_table' => $prompt_answers);
    $Log->exit() if $may_exit;
} ## end sub _restore_file_7mode_cli

sub _restore_file_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_base_backend_cmode_zapi(@args,
            api    => 'snapshot_restore_file',
            copy   => [qw( path restore-path snapshot volume )],
            map    => {
                'instance-uuid'  => 'snapshot-instance-uuid',
                'split-disabled' => 'space-efficient-split-disabled',
            },
    );
    $Log->exit() if $may_exit;
} ## end sub _restore_file_cmode_zapi

=head2 restore

   NACL::C::VolumeSnapshot->restore(
                                command_interface => $ci,
                             volume => $volume,
                             vserver => $vserver,
                             snapshot => $snapshot,
                                  ...);

 or

    $VolumeSnapshot_object->restore(...);

(Class or instance method)
This method is used to restore the volume to a snapshot..

Uses a CMode CLI/ZAPI, 7Mode CLI APISet.

Invokes "snap restore" command for 7Mode CLI.

CM ZAPI:

Invokes "snapshot-restore-volume-async" API for infinite volumes

Invokes "snapshot-restore-volume" API for flexible volumes

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< snapshot=>$snapshot_name >>

(Required for class method, Not Applicable for instance method)
Name of a snapshot.

Applicable for CMode CLI/ZAPI,7Mode CLI.

Maps to:

CM ZAPI: snapshot

7M CLI : snapshot (Programmatic name: "snapshot-name")

=item C<< volume => $string >>

(Required for class method, Not Applicable for instance method)
(Required) Name of volume of which the snapshot to be created.

Applicable for CMode CLI/ZAPI,7Mode CLI.

Maps to:

CM ZAPI: volume

7M CLI : volume name (Programmatic name: "vol-name")

=item C<< vserver => $string >>

(Required for class method, Not Applicable for instance method)
(Required for CMode-CLI, Not
Applicable for 7Mode-CLI)
Name of vserver.

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: ZAPI will be invoked in the context of the vserver specified to this
             option.

=item C<< cpcount-sort => $string >>

Applicable for CMode CLI.

=item C<< force => $boolean >>

Applicable for CMode CLI.

=item C<< instance-uuid => $string >>

Applicable for CMode CLI/ZAPI,7Mode CLI.

Maps to:

CM ZAPI: snapshot-instance-uuid

=item C<< preserve-lun-ids => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

preserve-lun-ids option of snapshot-restore-volume ZAPI API.

Not applicable for snapshot-restore-volume-async

=item C<< method-timeout => $string >>

(Optional)The new timeout value.

=item C<< "job_component"=> \$scalar >>

Filled in for CMode CLI/ZAPI.

CM ZAPI:

Applicable for only infinite volumes.

=item C<< 'is-repository' => $string >>

B<DEPRECATED>. Use C<is_infinivol> instead.

=item C<< is_infinivol => 0 | 1 >>

(Optional, Applicable for CMode ZAPI)

This boolean option describes whether the volume is an infinite volume or not.
For CMode ZAPI, the "async" form of the ZAPI must be used for the infinite
volume. If this option is not specified, the library will attempt to
determine if an infinite volume is being operated on (this could involve
querying the filer).

=item C<< is_infinivol_ref => \$scalar >>

(Optional, SCALARREF)

If the method was invoked through CMode ZAPI, the variable gets updated
describing whether it operated on an infinivol (used the "async" ZAPI) or not
(used the normal ZAPI).

=item C<< "extended_query" => $string >>

(Optional)

Applicable for CMode CLI.

=item C<< '*' >>

This method takes all other options supported by
"volume snapshot restore" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<NACL::C::Exceptions::VolumeSnapshot::SnapshotFenced>

This type of exception gets thrown if an attempt is made to restore a
volume snapshot which is fenced temporarily.

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub restore {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            job_component => {type => SCALARREF, optional => 1},
            _optional_scalars(
                qw( force instance-uuid preserve-lun-ids
                    cpcount-sort is-repository)
            ),
        },
        cli_cmd => 'volume snapshot restore',
    );

    try {
        $pkg_or_obj->call_on_apiset(
            %opts,
            choices => [
                {   method    => "_restore_cmode_cli",
                    interface => "CLI",
                    set       => "CMode",
                },
                {   method    => '_restore_cmode_zapi',
                    interface => 'ZAPI',
                    set       => 'CMode',
                    check     => '_restore_zapi_check',
                },
                {   method    => "_restore_7mode_cli",
                    interface => "CLI",
                    set       => "7Mode",
                },
            ],
            frontend => 'NACL::C::VolumeSnapshot::restore',
        );
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = $_[0];

        NACL::C::Exceptions::VolumeSnapshot::SnapRestoreNotLicensed
            ->detect_convert_and_throw(
            exception         => $exception,
            feature           => 'SnapRestore',
            command_interface => $opts{command_interface}
            );
        NACL::C::Exceptions::VolumeSnapshot::SnapshotFenced
            ->detect_convert_and_throw(exception => $exception);

        $Log->exit() if $may_exit;
        $exception->throw();
    };

    $Log->exit() if $may_exit;
} ## end sub restore

sub _restore_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    my $vserver = $opts{vserver};

    $pkg_or_obj->_base_backend_cmode_cli(
        %opts,
        api => 'volume_snapshot_restore',
        vserver_operated_on => \$vserver,
    );

    $Log->exit() if $may_exit;
} ## end sub _restore_cmode_cli

sub _restore_7mode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts   = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = $opts{apiset};

    my $snapshot_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => [qw(force instance-uuid)],
        map    => {
            'snapshot' => 'snapshot-name',
            'volume'   => 'vol-name',
        },
    );
    $snapshot_args->{type} = 'vol';

    my $prompt_answers = $pkg_or_obj->_restore_7mode_prompts();

    $apiset->snap_restore(%$snapshot_args,
        'connectrec-match_table' => $prompt_answers);
    $Log->exit() if $may_exit;
} ## end sub _restore_7mode_cli

sub _restore_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_backend_cmode_zapi_infinivol_builder(
        @args,
        api       => 'snapshot-restore-volume',
        copy      => [qw(snapshot volume)],
        map       => {'instance-uuid' => 'snapshot-instance-uuid'},
        non_async => {copy => ['preserve-lun-ids']},
    );
    $Log->exit() if $may_exit;
} ## end sub _restore_cmode_zapi

sub _restore_zapi_check {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    if (defined $opts{force} && $opts{force} eq 'true') {
        my $msg = 'Could not use the ZAPI implementation because '
            . 'the force option was provided but snapshot-restore does '
            . 'not support this option';
        $Log->comment($msg);
        $Log->exit() if $may_exit;
        NACL::Exceptions::InvalidChoice->throw($msg);
    }
    $Log->exit() if $may_exit;
}

=head2 rename

   NACL::C::VolumeSnapshot->rename(
                                command_interface => $ci,
                             volume => $volume,
                             vserver => $vserver,
                             snapshot => $snapshot,
                                  ...);

 or

    $VolumeSnapshot_object->rename(...);

(Class or instance method)
This method is used to rename a snapshot.

Uses a CMode CLI/ZAPI, 7Mode CLI APISet.

Invokes "snapshot-rename" API for CMode ZAPI.

Invokes "snap rename" command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< snapshot=>$snapshot_name >>

(Required for class method, Not Applicable for instance method)
Name of a snapshot.

Applicable for CMode CLI/ZAPI,7Mode CLI.

Maps to:

CM ZAPI: current-name

7M CLI : current snapshot name (Programmatic name: "current-name")

=item C<< volume => $string >>

(Required for class method, Not Applicable for instance method)
(Required) Name of volume of which the snapshot to be created.

Applicable for CMode CLI/ZAPI,7Mode CLI.

Maps to:

CM ZAPI: volume

7M CLI : volume name (Programmatic name: "vol-name")

=item C<< vserver => $string >>

(Required for class method, Not Applicable for instance method)
Name of vserver.

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: ZAPI will be invoked in the context of the vserver specified to this
option.

=item C<< new-name => $string >>

Applicable for CMode CLI/ZAPI,7Mode CLI.

Maps to:

CM ZAPI: new-name

7M CLI : new name (Programmatic name: "new-name")

=item C<< instance-uuid => $string >>

Applicable for CMode CLI.

=item C<< "job_component"=> \$scalar >>

Filled in for CMode CLI.

=item C<< 'cpcount-sort' => $string >>

Applicable for CMode CLI.

=item C<< "extended_query" => $string >>

(Optional)

Applicable for CMode CLI.

=item C<< '*' >>

This method takes all other options supported by
"volume snapshot rename" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<NACL::C::Exceptions::VolumeSnapshot::CurrentlyBusy>

This type of exception gets thrown if an attempt is made to rename
volume snapshot where volume is currently busy

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub rename {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            job_component => {type => SCALARREF, optional => 1},
            _optional_scalars(
                qw( instance-uuid new-name

                    cpcount-sort
                    )
            ),
        },
        cli_cmd => 'volume snapshot rename',
    );
    my $vserver;
    $opts{vserver_operated_on} = \$vserver;
    try {
        $pkg_or_obj->call_on_apiset(
            %opts,
            choices => [
                {   method    => '_rename_cmode_cli',
                    interface => 'CLI',
                    set       => 'CMode',
                },
                {   method    => '_rename_7mode_cli',
                    interface => 'CLI',
                    set       => '7Mode',
                },
                {   method    => '_rename_cmode_zapi',
                    interface => 'ZAPI',
                    set       => 'CMode',
                },
            ],
            frontend => 'NACL::C::VolumeSnapshot::rename',
        );
    }
    catch NACL::APISet::Exceptions::ResponseException with {
       my $exception = $_[0];
       NACL::C::Exceptions::VolumeSnapshot::CurrentlyBusy
            ->detect_convert_and_throw(exception => $exception);
       $Log->exit() if $may_exit;
       $exception->throw();
    };
    $pkg_or_obj->snapshot($opts{'new-name'}) if (ref $pkg_or_obj);
    $Log->exit() if $may_exit;
} ## end sub rename

sub _rename_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    my $response;
    $pkg_or_obj->_base_backend_cmode_cli(%opts,
            api => 'volume_snapshot_rename',
            apiset_response_obj => \$response
    );
    $Log->exit() if $may_exit;
} ## end sub _rename_cmode_cli

sub _rename_7mode_cli {
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    delete $opts{command_interface};
    my $apiset        = delete $opts{apiset};
    my $snapshot_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => [qw(aggregate new-name)],
        map    => {
            "snapshot" => "current-name",
            "newname"  => "new-name",
            "volume"   => "vol-name",
        }
    );
    %opts = %$snapshot_args;
    $apiset->snap_rename(%opts);
} ## end sub _rename_7mode_cli

sub _rename_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my $allowed_cmode_options = [qw( volume new-name)];
    $pkg_or_obj->_base_backend_cmode_zapi(@args,
            api    => 'snapshot_rename',
            copy   => $allowed_cmode_options,
            map    => {"snapshot" => "current-name",},
    );
    $Log->exit() if $may_exit;
} ## end sub _rename_cmode_zapi

=head2 multirename

   NACL::C::VolumeSnapshot->multirename(
                                command_interface => $ci,
                                'new-snapshot-name' => $new_snapshot_name,
                                'snapshot-name' => $old_snapshot_name,
                                vserver => $vserver,
                                'volume-names' => [ vol1 vol2 ],
                                  ...);

(Class  method)
This method is used to rename all snapshots.

Uses a CMode ZAPI APISet.

Invokes "snapshot-multirename" API for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< 'snapshot-name' =>$snapshot_name >>

(Required for class method)
Name of a snapshot.

Applicable for CMode ZAPI.

Maps to:

CM ZAPI: snapshot-name

=item C<< 'new-snapshot-name' =>$new_snapshot_name >>

(Required for class method)
New name of a snapshot.

Applicable for CMode ZAPI.

Maps to:

CM ZAPI: new-snapshot-name

=item C<< 'volume-names' => [vol1 vol2] >>

(Required for class method, Not Applicable for instance method)
(Required) Name of volumes for which the snapshot to be created.

Applicable for CMode ZAPI.

Maps to:

CM ZAPI: volume-names

=item C<< '_error_typedef' => $scalar >>

defines error type in multirename operation

Applicable for CMode ZAPI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=item C<NACL::C::Exceptions::VolumeSnapshot::MultiOperationFailed>

This exception is thrown when we attempt to invoke any multi-api(create|delete|rename|validate) on  snapshot for multiple volumes and it got failed for one or more volumes.

=back

=back

=cut

sub multirename {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
       params          => \@args,
       additional_spec => {
           'new-snapshot-name' => {type => SCALAR},
           'snapshot-name'     => {type => SCALAR},
           'volume-names'      => {type => ARRAYREF},
       },
       ignore_primary_keys => 1,
       allow_extra         =>1,
    );
    my $vserver;
    $opts{vserver_operated_on} = \$vserver;
    $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_multirename_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
         ],
         frontend => 'NACL::C::VolumeSnapshot::multirename',
    );

    $Log->exit() if $may_exit;
}

sub _multirename_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg, %opts) = @_;

    $opts{'_error_typedef'}      = 'snapshot-multirename-errors';

    $pkg->_multi_base_backend_cmode_zapi(%opts,
            api    => 'snapshot-multirename',
            copy   => [qw(new-snapshot-name snapshot-name volume-names)],
    );
    $Log->exit() if $may_exit;
} ## end sub _multirename_cmode_zapi

=head2 reserve

   NACL::C::VolumeSnapshot->reserve(
                                command_interface => $ci,
                             volume => $volume,
                             vserver => $vserver,
                             snapshot => $snapshot,
                                  ...);

 or

    $VolumeSnapshot_object->reserve(...);

(Class or instance method)
This method is used to reserve a snapshot.

Uses a 7Mode CLI APISet.

Invokes "snap reserve" command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< volume => $volume >>

(Required for class method, Not Applicable for instance method)

Applicable for 7Mode CLI.

=item C<< snapshot => $snapshot >>

(Required for class method, Not Applicable for instance method)

Applicable for 7Mode CLI.

=item C<< percentage => $string >>

Applicable for 7Mode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub reserve {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params              => \@args,
        additional_spec     => {_optional_scalars(qw(volume percentage))},
        ignore_primary_keys => 1,
    );
    $Log->exit() if $may_exit;

    return $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => "_reserve_7mode_cli",
                interface => "CLI",
                set       => "7Mode",
            },
        ],
    );
} ## end sub reserve

sub _reserve_7mode_cli {
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    delete $opts{command_interface};
    my $apiset        = delete $opts{apiset};
    my $snapshot_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => [qw(aggregate percentage)],
        map    => {"volume" => "vol-name"},
    );
    %opts = %$snapshot_args;
    $apiset->snap_reserve(%opts);
} ## end sub _reserve_7mode_cli

=head2 sched

   NACL::C::VolumeSnapshot->sched(
                                command_interface => $ci,
                             volume => $volume,
                             vserver => $vserver,
                             snapshot => $snapshot,
                                  ...);

 or

    $VolumeSnapshot_object->sched(...);

(Class method)
This method is used to sched a snapshot.

Uses a 7Mode CLI APISet.

Invokes "snap sched" command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< volume => $volume >>

(Required)

Applicable for 7Mode CLI.

=item C<< weeks => $string >>

Applicable for 7Mode CLI.

=item C<< days => $string >>

Applicable for 7Mode CLI.

=item C<< hours => $string >>

Applicable for 7Mode CLI.

=item C<< hour-list => [ $hour1, $hour2, ... ] >>

Applicable for 7Mode CLI.

=item C<< minutes => $string >>

Applicable for 7Mode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub sched {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            "hour-list" => {type => ARRAYREF, optional => 1},
            _optional_scalars(qw( weeks days hours minutes volume )),
        },
        ignore_primary_keys => 1,
    );
    $Log->exit() if $may_exit;

    return $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => "_sched_7mode_cli",
                interface => "CLI",
                set       => "7Mode",
            },
        ],
    );
} ## end sub sched

sub _sched_7mode_cli {
    my $pkg  = shift;
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );
    delete $opts{command_interface};
    my $apiset        = delete $opts{apiset};
    my $snapshot_args = $pkg->_hash_copy(
        source => \%opts,
        copy   => [qw( aggregate weeks days hours hour-list minutes)],
        map    => {"volume" => "vol-name"},
    );
    %opts = %$snapshot_args;
    if (defined($opts{'hour-list'})) {

        my $hour_list = join ',', @{$opts{'hour-list'}};
        delete $opts{'hour-list'};
        $opts{'hours'} = $opts{'hours'} . '@' . $hour_list;
    }

    $apiset->snap_sched(%opts);
} ## end sub _sched_7mode_cli

=head2 partial_restore_file

   NACL::C::VolumeSnapshot->partial_restore_file(
                                command_interface => $ci,
                                     volume => $volume,
                                     vserver => $vserver,
                                     snapshot => $snapshot,
                                  ...);

                                                                 or

    $VolumeSnapshot_object->partial_restore_file(...);

(Class or instance method)
This method is used to restore part of a file from a snapshot.

Uses a CMode CLI/ZAPI APISet.

Invokes "snapshot-partial-restore-file" API for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< volume => $volume >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: volume

=item C<< vserver => $vserver >>
(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: ZAPI will be invoked in the context of the vserver specified to this
option.

=item C<< blocks => $string >>

Applicable for CMode CLI.

=item C<< cumulative-percentage-of-used-blocks => $string >>

Applicable for CMode CLI.

=item C<< preserve-lun-ids => $string >>

Applicable for CMode CLI.

=item C<< async => $string >>

Applicable for CMode CLI.

=item C<< version-uuid => $string >>

Applicable for CMode CLI.

=item C<< dsid => $string >>

Applicable for CMode CLI.

=item C<< byte-count => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: byte-count

=item C<< max-bytes => $string >>

Applicable for CMode CLI.

=item C<< volume-dsid => $string >>

Applicable for CMode CLI.

=item C<< new-name => $string >>

Applicable for CMode CLI.

=item C<< physical-snap-id => $string >>

Applicable for CMode CLI.

=item C<< blocks-reserved => $string >>

Applicable for CMode CLI.

=item C<< logical-snap-id => $string >>

Applicable for CMode CLI.

=item C<< cumulative-total => $string >>

Applicable for CMode CLI.

=item C<< cpcount-sort => $string >>

Applicable for CMode CLI.

=item C<< ignore-owners => $string >>

Applicable for CMode CLI.

=item C<< instance-uuid => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: snapshot-instance-uuid

=item C<< percentage => $string >>

Applicable for CMode CLI.

=item C<< test-version-uuid => $string >>

Applicable for CMode CLI.

=item C<< foreground => $string >>

Applicable for CMode CLI.

=item C<< busy => $string >>

Applicable for CMode CLI.

=item C<< size-kb => $string >>

Applicable for CMode CLI.

=item C<< fs-version => $string >>

Applicable for CMode CLI.

=item C<< record-owner => $string >>

Applicable for CMode CLI.

=item C<< msid => $string >>

Applicable for CMode CLI.

=item C<< split-disabled => $string >>

Applicable for CMode CLI.

=item C<< size => $string >>

Applicable for CMode CLI.

=item C<< dependency => $string >>

Applicable for CMode CLI.

=item C<< is-7-mode => $string >>

Applicable for CMode CLI.

=item C<< owners => $string >>

Applicable for CMode CLI.

=item C<< percent-reserved => $string >>

Applicable for CMode CLI.

=item C<< create-time => $string >>

Applicable for CMode CLI.

=item C<< start-byte => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: start-byte

=item C<< limit => $string >>

Applicable for CMode CLI.

=item C<< owner => $string >>

Applicable for CMode CLI.

=item C<< size-blocks => $string >>

Applicable for CMode CLI.

=item C<< cumulative-percentage-of-total-blocks => $string >>

Applicable for CMode CLI.

=item C<< progress => $string >>

Applicable for CMode CLI.

=item C<< cpcount => $string >>

Applicable for CMode CLI.

=item C<< path => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: path

=item C<< internal-status => $string >>

Applicable for CMode CLI.

=item C<< force => $string >>

Applicable for CMode CLI.

=item C<< tags => $string >>

Applicable for CMode CLI.

=item C<< comment => $string >>

Applicable for CMode CLI.

=item C<< restore-path => $string >>

Applicable for CMode CLI.

=item C<< usedblocks => $string >>

Applicable for CMode CLI.

=item C<< method-timeout => $string >>

(Optional)The new timeout value.

=item C<< snapshot => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: snapshot

=item C<< "extended_query" => $string >>

(Optional)

Applicable for CMode CLI.

=item C<< '*' >>

This method takes all other options supported by
"volume snapshot partial-restore-file" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub partial_restore_file {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( blocks cumulative-percentage-of-used-blocks preserve-lun-ids async
                    version-uuid dsid byte-count max-bytes volume-dsid new-name physical-snap-id
                    blocks-reserved logical-snap-id cumulative-total cpcount-sort
                    ignore-owners instance-uuid percentage test-version-uuid foreground busy
                    size-kb fs-version record-owner msid split-disabled size dependency is-7-mode
                    owners percent-reserved create-time start-byte limit owner size-blocks
                    cumulative-percentage-of-total-blocks progress cpcount path
                    internal-status force tags comment restore-path usedblocks
                    )
            ),
        },
        cli_cmd => 'volume snapshot partial-restore-file',
    );
    my $vserver;
    $opts{vserver_operated_on} = \$vserver;
    my $ret_obj = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_partial_restore_file_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_partial_restore_file_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::VolumeSnapshot::partial_restore_file',
    );
    $Log->exit() if $may_exit;
    return $ret_obj;
} ## end sub partial_restore_file

sub _partial_restore_file_cmode_cli {
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@args,  api => 'volume_snapshot_partial_restore_file');
} ## end sub _partial_restore_file_cmode_cli

sub _partial_restore_file_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my $allowed_cmode_options
        = [qw( byte-count path snapshot start-byte volume )];
    $pkg_or_obj->_base_backend_cmode_zapi(@args,
            api    => 'snapshot_partial_restore_file',
            copy   => $allowed_cmode_options,
            map    => {"instance-uuid" => "snapshot-instance-uuid",},
    );
} ## end sub _partial_restore_file_cmode_zapi

=head2 restore_volume

Note that this method is B<DEPRECATED>. The method L<restore|lib-NACL-C-VolumeSnapshot-pm/restore> should
be used instead to restore a volume from a snapshot.

=cut

sub restore_volume {
    $Log->enter() if $may_enter;
    my $warn_message = 'This method (restore_volume) is deprecated. Use the '
            . "method 'restore' instead";
    NACL::C::Component->_print_deprecation_warning($warn_message);
    my ($pkg_or_obj, @args) = @_;
    my $ret = $pkg_or_obj->restore(@args);
    $Log->exit() if $may_exit;
    return $ret;
} ## end sub restore_volume

=head2 modify

    NACL::C::VolumeSnapshot->modify(
                                command_interface => $ci,
                            'volume' => $volume,
                            'vserver' => $vserver,
                            'snapshot' => $snapshot,
                            'cpcount-sort' => $cpcount_sort,
                                ...);

 or

    $VolumeSnapshot_obj->modify(...);

(Class or instance method)
This method is used to modify snapshot attributes.

Uses a CMode CLI APISet.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'volume' => $volume >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI.

=item C<< 'vserver' => $vserver >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI.

=item C<< 'snapshot' => $snapshot >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI.

=item C<< 'cpcount-sort' => $cpcount_sort >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI.

=item C<< 'comment' => $string >>

Applicable for CMode CLI.

=item C<< 'method-timeout' => $time_in_seconds >>

(Optional)
The default NATE timeout is 60 seconds. If the command might take more
time to complete, this option can be used to specify a larger timeout value.
The value should be provided in seconds.

=item C<< 'cpcount-sort' => $string >>

Applicable for CMode CLI.

=item C<< 'snapmirror-label' => $string >>

Applicable for CMode CLI.

=item C<< '*' >>

This method takes all other options supported by
"volume snapshot modify" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub modify {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( comment

                    cpcount-sort

                    snapmirror-label
                    )
            ),
        },
        cli_cmd => 'volume snapshot modify',
    );
    my $vserver;
    $opts{vserver_operated_on} = \$vserver;
    $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_modify_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::VolumeSnapshot::modify',
    );

    $Log->exit() if $may_exit;
} ## end sub modify

sub _modify_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@args,  api => 'volume_snapshot_modify');
    $Log->exit() if $may_exit;
} ## end sub _modify_cmode_cli

=head2 hammer_ownership

    NACL::C::VolumeSnapshot->hammer_ownership(
                                command_interface => $ci,
                            'volume' => $volume,
                            'vserver' => $vserver,
                            'snapshot' => $snapshot,
                            'cpcount-sort' => $cpcount_sort,
                                ...);

 or

    $VolumeSnapshot_obj->hammer_ownership(...);

(Class or instance method)
This method is used to add and remove ownership to a given snapshot 100 times.

Uses a CMode CLI/ZAPI APISet.

Invokes "snapshot-hammer-ownership" API for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'volume' => $volume >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: volume

(Required for class method, Not Applicable for instance method)

=item C<< 'vserver' => $vserver >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: ZAPI will be invoked in the context of the vserver specified to this
             option.

=item C<< 'snapshot' => $snapshot >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: name

=item C<< 'cpcount-sort' => $cpcount_sort >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI.

=item C<< 'owner' => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: owner

=item C<< 'method-timeout' => $time_in_seconds >>

(Optional)
The default NATE timeout is 60 seconds. If the command might take more
time to complete, this option can be used to specify a larger timeout value.
The value should be provided in seconds.

=item C<< "extended_query" => $string >>

(Optional)

Applicable for CMode CLI.

=item C<< '*' >>

This method takes all other options supported by
"volume snapshot hammer-ownership" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub hammer_ownership {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( owner
                    cpcount-sort
                    )
            ),
        },
        api => 'volume_snapshot_hammer_ownership',
    );
    $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_hammer_ownership_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_hammer_ownership_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::VolumeSnapshot::hammer_ownership',
    );

    $Log->exit() if $may_exit;
} ## end sub hammer_ownership

sub _hammer_ownership_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    my $vserver = $opts{vserver};

    $pkg_or_obj->_base_backend_cmode_cli (
        %opts,
        api => 'volume_snapshot_hammer_ownership',
        vserver_operated_on => \$vserver,
    );

    $Log->exit() if $may_exit;
} ## end sub _hammer_ownership_cmode_cli

sub _hammer_ownership_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj ,@args)  = @_ ;

    $pkg_or_obj->_base_backend_cmode_zapi(@args,
                  api => 'snapshot_hammer_ownership' ,
                  copy => [qw( owner volume)],
                  map  => {"snapshot" => "name"},
    );

    $Log->exit() if $may_exit;

} ## end sub _hammer_ownership_cmode_zapi

=head2 compute_reclaimable

    NACL::C::VolumeSnapshot->compute_reclaimable(
                                'vserver' => $vserver,
                                ...);

(Class method)
This method is used to calculate the reclaimable space if specified snapshots are deleted.

Uses a CMode CLI/ZAPI, 7Mode CLI APISet.

Invokes "snapshot-reclaimable-info" API for CMode ZAPI.

Invokes "snap reclaimable" command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'volume' => $volume >>

(Required)

Applicable for CMode CLI/ZAPI,7Mode CLI.

Maps to:

CM ZAPI: volume

7M CLI : volume name (Programmatic name: "vol-name")

=item C<< 'vserver' => $vserver >>

(Required)

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: ZAPI will be invoked in the context of the vserver specified to this
             option.

=item C<< 'snapshots' => [ $snapshot1, $snapshot2, ... ] >>

(Arrayref, even if only a single value is to be sent it should specified
as an arrayref containing that single value)

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: snapshots

=item C<< 'method-timeout' => $time_in_seconds >>

(Optional)
The default NATE timeout is 60 seconds. If the command might take more
time to complete, this option can be used to specify a larger timeout value.
The value should be provided in seconds.

=item C<< "extended_query" => $string >>

(Optional)

Applicable for CMode CLI.

=item C<< 'reclaimable-size' => $string >>

Size Reclaimable
Applicable for CMode CLI

=item C<< '*' >>

This method takes all other options supported by
"volume snapshot compute-reclaimable" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub compute_reclaimable {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            'snapshots' => {type => ARRAYREF, optional => 1},
            _optional_scalars(
                qw( volume vserver

                    reclaimable-size
                    )
            ),
        },
        ignore_primary_keys => 1,
        cli_cmd             => 'volume snapshot compute-reclaimable',
    );
    my $vserver;
    $opts{vserver_operated_on} = \$vserver;
    my $ret = $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_compute_reclaimable_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_compute_reclaimable_7mode_cli',
                interface => 'CLI',
                set       => '7Mode',
            },
            {   method    => '_compute_reclaimable_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::VolumeSnapshot::compute_reclaimable',
    );
    $Log->exit() if $may_exit;
    return $ret;
} ## end sub compute_reclaimable

sub _compute_reclaimable_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    my $response;
    $pkg->_base_backend_cmode_cli(@args,
            api => 'volume_snapshot_compute_reclaimable',
            apiset_response_obj => \$response
    );
    my $bytes_freed = $response->get_parsed_output()->[0]{bytes_freed};
    $Log->exit() if $may_exit;
    return $bytes_freed;
} ## end sub _compute_reclaimable_cmode_cli

sub _compute_reclaimable_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    my $response;
    my $allowed_cmode_options = [qw( volume snapshots )];
    $pkg->_base_backend_cmode_zapi(@args,
            api    => 'snapshot_reclaimable_info',
            copy   => $allowed_cmode_options,
            apiset_response_obj => \$response,
    );
    my $bytes_freed = $response->get_parsed_output()->[0]{'reclaimable-size'};
    $Log->exit() if $may_exit;
    return $bytes_freed;
} ## end sub _compute_reclaimable_cmode_zapi

sub _compute_reclaimable_7mode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    my %opts              = $pkg->_validate_backend_opts(@args);
    my $command_interface = $opts{command_interface};
    my $apiset            = $opts{apiset};

    my $snapshot_args = $pkg->_hash_copy(
        source => \%opts,
        map    => {'volume' => 'vol-name',},
    );
    my $temp = join ' ', @{$opts{'snapshots'}};
    $snapshot_args->{snapshot} = $temp;

    my $response    = $apiset->snap_reclaimable(%$snapshot_args);
    my $bytes_freed = $response->get_parsed_output()->[0]{bytes_freed};
    $Log->exit() if $may_exit;
    return $bytes_freed;
} ## end sub _compute_reclaimable_7mode_cli

=head2 remove_owner

    NACL::C::VolumeSnapshot->remove_owner(
                                command_interface => $ci,
                            'volume' => $volume,
                            'vserver' => $vserver,
                            'snapshot' => $snapshot,
                            'cpcount-sort' => $cpcount_sort,
                                ...);

 or

    $VolumeSnapshot_obj->remove_owner(...);

(Class or instance method)
This method is used to remove owner from the ownership list.

Uses a CMode CLI/ZAPI APISet.

Invokes "snapshot-remove-owner" API for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'volume' => $volume >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: volume

=item C<< 'vserver' => $vserver >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: ZAPI will be invoked in the context of the vserver specified to this
             option.

=item C<< 'snapshot' => $snapshot >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: name

=item C<< 'cpcount-sort' => $cpcount_sort >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI.

=item C<< 'owner' => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: owner

=item C<< 'method-timeout' => $time_in_seconds >>

(Optional)
The default NATE timeout is 60 seconds. If the command might take more
time to complete, this option can be used to specify a larger timeout value.
The value should be provided in seconds.

=item C<< "extended_query" => $string >>

(Optional)

Applicable for CMode CLI.

=item C<< '*' >>

This method takes all other options supported by
"volume snapshot remove-owner" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub remove_owner {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( owner
                    cpcount-sort
                    )
            ),
        },
        cli_cmd => 'volume snapshot remove-owner',
    );
    my $vserver;
    $opts{vserver_operated_on} = \$vserver;
    $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_remove_owner_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_remove_owner_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::VolumeSnapshot::remove_owner',
    );

    $Log->exit() if $may_exit;
} ## end sub remove_owner

sub _remove_owner_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@args,  api => 'volume_snapshot_remove_owner');
    $Log->exit() if $may_exit;
} ## end sub _remove_owner_cmode_cli

sub _remove_owner_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my $allowed_cmode_options = [qw( owner volume )];
    $pkg_or_obj->_base_backend_cmode_zapi(@args,
            api    => 'snapshot_remove_owner',
            copy   => $allowed_cmode_options,
            map    => {'snapshot' => 'name'},
    );
    $Log->exit() if $may_exit;
} ## end sub _remove_owner_cmode_zapi

=head2 exercise_query

    NACL::C::VolumeSnapshot->exercise_query(
                                command_interface => $ci,
                            'volume' => $volume,
                            'vserver' => $vserver,
                            'snapshot' => $snapshot,
                            'cpcount-sort' => $cpcount_sort,
                                ...);

 or

    $VolumeSnapshot_obj->exercise_query(...);

(Class or instance method)
This method is used to exercise the snapshot query (for performance).

Uses a CMode CLI/ZAPI APISet.

Invokes "snapshot-exercise-query" API for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'volume' => $volume >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: volume

=item C<< 'vserver' => $vserver >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: ZAPI will be invoked in the context of the vserver specified to this
             option.

=item C<< 'snapshot' => $snapshot >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: snapshot

=item C<< 'cpcount-sort' => $cpcount_sort >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI.

=item C<< 'method-timeout' => $time_in_seconds >>

(Optional)
The default NATE timeout is 60 seconds. If the command might take more
time to complete, this option can be used to specify a larger timeout value.
The value should be provided in seconds.

=item C<< "extended_query" => $string >>

(Optional)

Applicable for CMode CLI.

=item C<< '*' >>

This method takes all other options supported by
"volume snapshot exercise-query" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub exercise_query {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( cpcount-sort
                    )
            )
        },
        api => 'volume_snapshot_exercise_query',
    );
    $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_exercise_query_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_exercise_query_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::VolumeSnapshot::exercise_query',
    );

    $Log->exit() if $may_exit;
} ## end sub exercise_query

sub _exercise_query_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    my $vserver = $opts{vserver};

    $pkg_or_obj->_base_backend_cmode_cli(
        %opts,
        api => 'volume_snapshot_exercise_query',
        vserver_operated_on => \$vserver,
    );

    $Log->exit() if $may_exit;
} ## end sub _exercise_query_cmode_cli

sub _exercise_query_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj ,@args)  = @_ ;
    $pkg_or_obj->_base_backend_cmode_zapi(@args,
                  api => 'snapshot_exercise_query' ,
                  copy => [qw(volume)],
                  map  => {"snapshot" => "name"},
    );
    $Log->exit() if $may_exit;
} ## end sub _exercise_query_cmode_zapi

=head2 add_test_owners

    NACL::C::VolumeSnapshot->add_test_owners(
                                command_interface => $ci,
                            'volume' => $volume,
                            'vserver' => $vserver,
                            'snapshot' => $snapshot,
                            'cpcount-sort' => $cpcount_sort,
                                ...);

 or

    $VolumeSnapshot_obj->add_test_owners(...);

(Class or instance method)
This method is used to add test owners to a snapshot.

Uses a CMode CLI/ZAPI APISet.

Invokes "snapshot-add-test-owners" API for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'volume' => $volume >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: volume

=item C<< 'vserver' => $vserver >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: ZAPI will be invoked in the context of the vserver specified to this
             option.

=item C<< 'snapshot' => $snapshot >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: name

=item C<< 'cpcount-sort' => $cpcount_sort >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI.

=item C<< 'method-timeout' => $time_in_seconds >>

(Optional)
The default NATE timeout is 60 seconds. If the command might take more
time to complete, this option can be used to specify a larger timeout value.
The value should be provided in seconds.

=item C<< "extended_query" => $string >>

(Optional)

Applicable for CMode CLI.

=item C<< '*' >>

This method takes all other options supported by
"volume snapshot add-test-owners" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub add_test_owners {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( cpcount-sort
                    )
            )
        },
        cli_cmd => 'volume snapshot add-test-owners',
    );
    my $vserver;
    $opts{vserver_operated_on} = \$vserver;
    $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_add_test_owners_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_add_test_owners_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::VolumeSnapshot::add_test_owners',
    );

    $Log->exit() if $may_exit;
} ## end sub add_test_owners

sub _add_test_owners_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@args,  api => 'volume_snapshot_add_test_owners');
    $Log->exit() if $may_exit;
} ## end sub _add_test_owners_cmode_cli

sub _add_test_owners_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my $allowed_cmode_options = [
            qw( volume
            )
    ];
    $pkg_or_obj->_base_backend_cmode_zapi(@args,
            api    => 'snapshot_add_test_owners',
            copy   => $allowed_cmode_options,
            map    => {"snapshot" => "name"},
    );
    $Log->exit() if $may_exit;
} ## end sub _add_test_owners_cmode_zapi

=head2 name_attribute

Returns 'snapshot'.

=cut

sub name_attribute {
    return 'snapshot';
}

=head2 promote

    NACL::C::VolumeSnapshot->promote(
                                command_interface => $ci,
	                            'vserver' => $vserver,
	                            'snapshot' => $snapshot,
	                            'volume' => $volume,
                                ...);

								 or

    $VolumeSnapshot_obj->promote(...);

(Class or instance method)
This method is used to promote a snapshot to its volume.
Uses a CMode CLI APISet.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'vserver' => $vserver >>

(Required for class method, Not Applicable for instance method)
Vserver possible value(s) are, vserver name

Applicable for CMode CLI

=item C<< 'snapshot' => $snapshot >>

(Required for class method, Not Applicable for instance method)
Snapshot possible value(s) are, snapshot name

Applicable for CMode CLI

=item C<< 'volume' => $volume >>

(Required for class method, Not Applicable for instance method)
Volume possible value(s) are, volume name

Applicable for CMode CLI

=item C<< 'preserve-lun-ids' => $string >>

Preserve LUN Identifiers possible value(s) are, true,false

Applicable for CMode CLI

=item C<< 'instance-uuid' => $string >>

Instance UUID possible value(s) are, UUID

Applicable for CMode CLI

=item C<< 'force' => $string >>

Force Restore possible value(s) are, true

Applicable for CMode CLI

=item C<< 'cmd_execution_time' => \$scalar >>

(Optional) Scalar reference , it will be updated with the command execution
time returned by the apiset response object

=item C<< 'method-timeout' => $time_in_seconds >>

(Optional)
The default NATE timeout is 60 seconds. If the command might take more
time to complete, this option can be used to specify a larger timeout value.
The value should be provided in seconds.

=item C<< "extended_query" => $string >>

(Optional)

Applicable for CMode CLI.

=item C<< '*' >>

This method takes all other options supported by
"volume snapshot promote" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub promote {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            'cmd_execution_time' => {type => SCALARREF, optional => 1},
            _optional_scalars(
                qw( preserve-lun-ids instance-uuid force
                    )
            ),
        },
        cli_cmd => 'volume snapshot promote',
    );
    $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_promote_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::VolumeSnapshot::promote',
    );

    $Log->exit() if $may_exit;
} ## end sub promote

sub _promote_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;

    $pkg_or_obj->_base_backend_cmode_cli(
        @args,
        api => 'volume_snapshot_promote',
    );

    $Log->exit() if $may_exit;
} ## end sub _promote_cmode_cli

=head2 prepare_for_revert

    NACL::C::VolumeSnapshot->prepare_for_revert(
                                command_interface => $ci,
                                ...);

(Class method)
This method is used to deletes multiple Snapshot copies of the current File System version..
Uses CMode CLI .

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'node' => $string >>

Node
Applicable for CMode CLI

=item C<< 'snapshots_deleted' => $string >>

Snapshots Deleted ranges from 0 to 2147483647
Applicable for CMode CLI

=item C<< 'zapi_count' => $string >>

Zapi Count ranges from 0 to 256
Applicable for CMode CLI

=item C<< '*' >>

This method takes all other options supported by "volume snapshot prepare for revert" command in ONTAP

=item C<< other_options >>

This method takes other common options described in L<Common method options|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#Other_options_valid_on_component_methods>

=back

=back

=over

=item Exceptions

=over

=item C<NACL::C::Exceptions::VolumeSnapshot::SnapshotHasOwner>

This type of exception gets thrown if an attempt to prepare_for_revert
and snapshot is in use and owned by another OnTap subsystem.

=item C<NACL::C::Exceptions::VolumeSnapshot::RetryCmd>

This type of exception gets thrown if an attempt to prepare_for_revert
fails with same command suggestion.

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub prepare_for_revert {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( node

                    snapshots_deleted zapi_count
                    )
            ),
        },
        ignore_primary_keys => 1,
        cli_cmd             => 'volume snapshot prepare-for-revert',
    );
    try {
        $pkg->call_on_apiset(
            %opts,
            choices => [
                {   method    => '_prepare_for_revert_cmode_cli',
                    interface => 'CLI',
                    set       => 'CMode',
                },
            ],
            frontend => 'NACL::C::VolumeSnapshot::prepare_for_revert',
        );
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = $_[0];

        foreach my $type (qw(SnapshotHasOwner RetryCmd DisableSnapshotPolicies)) {
            my $full_type = "NACL::C::Exceptions::VolumeSnapshot::$type";
            $full_type->detect_convert_and_throw(exception => $exception);
        }
        $Log->exit() if $may_exit;
        $exception->throw();
    };

    $Log->exit() if $may_exit;
}

sub _prepare_for_revert_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, @opts) = @_;
    $pkg->_base_backend_cmode_cli(@opts,
        api => 'volume_snapshot_prepare_for_revert',);
    $Log->exit() if $may_exit;
}

=head2 delta

    NACL::C::VolumeSnapshot->delta(
                                command_interface => $ci,
                                ...);

(Class method)
Uses CMode CLI .

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< '*' >>

This method takes all other options supported by "volume snapshot delta" command in ONTAP

=item C<< other_options >>

This method takes other common options described in L<Common method options|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#Other_options_valid_on_component_methods>

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub delta {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params              => \@args,
        ignore_primary_keys => 1,
        cli_cmd             => 'volume snapshot delta',
    );
    $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_delta_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::VolumeSnapshot::delta',
    );

    $Log->exit() if $may_exit;
}

sub _delta_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, @opts) = @_;
    $pkg->_base_backend_cmode_cli(@opts, api => 'volume_snapshot_delta',);
    $Log->exit() if $may_exit;
}

=head2 get_snapshot_UUID

    NACL::C::VolumeSnapshot->get_snapshot_UUID(
                                command_interface => $ci,
	                            'vserver' => $vserver,
	                            'snapshot' => $snapshot,
	                            'volume' => $volume,
                                ...);

								 or

    $VolumeSnapshot_obj->get_snapshot_UUID(...);

(Class or instance method)
This method is used to get UUID of given snapshot.
Uses CMode CLI .

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'vserver' => $vserver >>

(Required for class method, Not Applicable for instance method)
Vserver
Applicable for CMode CLI

=item C<< 'snapshot' => $snapshot >>

(Required for class method, Not Applicable for instance method)
Snapshot
Applicable for CMode CLI

=item C<< 'volume' => $volume >>

(Required for class method, Not Applicable for instance method)
Volume
Applicable for CMode CLI

=item C<< 'reposId' => $string >>

Repository Id
Applicable for CMode CLI

=item C<< '*' >>

This method takes all other options supported by "volume snapshot get snapshot UUID" command in ONTAP

=item C<< other_options >>

This method takes other common options described in L<Common method options|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#Other_options_valid_on_component_methods>

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub get_snapshot_UUID {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( reposId
                    )
            ),
        },
        cli_cmd => 'volume snapshot get_snapshot_UUID',
    );
    $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_get_snapshot_UUID_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::VolumeSnapshot::get_snapshot_UUID',
    );

    $Log->exit() if $may_exit;
}

sub _get_snapshot_UUID_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@opts,
        api => 'volume_snapshot_get_snapshot_UUID',);
    $Log->exit() if $may_exit;
}

=head2 get_NS_Constituent

    NACL::C::VolumeSnapshot->get_NS_Constituent(
                                command_interface => $ci,
	                            'vserver' => $vserver,
	                            'snapshot' => $snapshot,
	                            'volume' => $volume,
                                ...);

								 or

    $VolumeSnapshot_obj->get_NS_Constituent(...);

(Class or instance method)
This method is used to get snapMsid of Infinite Volume NS Constituent.
Uses CMode CLI .

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'vserver' => $vserver >>

(Required for class method, Not Applicable for instance method)
Vserver
Applicable for CMode CLI

=item C<< 'snapshot' => $snapshot >>

(Required for class method, Not Applicable for instance method)
Snapshot
Applicable for CMode CLI

=item C<< 'volume' => $volume >>

(Required for class method, Not Applicable for instance method)
Volume
Applicable for CMode CLI

=item C<< 'reposId' => $string >>

Repository Id
Applicable for CMode CLI

=item C<< 'needMirror' => $string >>

Needs Mirror possible value(s) are, true,false
Applicable for CMode CLI

=item C<< 'instance-uuid' => $string >>

Instance UUID
Applicable for CMode CLI

=item C<< '*' >>

This method takes all other options supported by "volume snapshot get NS Constituent" command in ONTAP

=item C<< other_options >>

This method takes other common options described in L<Common method options|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#Other_options_valid_on_component_methods>

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub get_NS_Constituent {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( reposId needMirror instance-uuid
                    )
            ),
        },
        cli_cmd => 'volume snapshot get_NS_Constituent',
    );
    $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_get_NS_Constituent_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::VolumeSnapshot::get_NS_Constituent',
    );

    $Log->exit() if $may_exit;
}

sub _get_NS_Constituent_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@opts,
        api => 'volume_snapshot_get_NS_Constituent',);
    $Log->exit() if $may_exit;
}

=head2 get_Data_Constituent

    NACL::C::VolumeSnapshot->get_Data_Constituent(
                                command_interface => $ci,
	                            'vserver' => $vserver,
	                            'snapshot' => $snapshot,
	                            'volume' => $volume,
                                ...);

								 or

    $VolumeSnapshot_obj->get_Data_Constituent(...);

(Class or instance method)
This method is used to get snapMsid of Infinite Volume Data Constituent.
Uses CMode CLI .

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'vserver' => $vserver >>

(Required for class method, Not Applicable for instance method)
Vserver
Applicable for CMode CLI

=item C<< 'snapshot' => $snapshot >>

(Required for class method, Not Applicable for instance method)
Snapshot
Applicable for CMode CLI

=item C<< 'volume' => $volume >>

(Required for class method, Not Applicable for instance method)
Volume
Applicable for CMode CLI

=item C<< 'reposId' => $string >>

Repository Id
Applicable for CMode CLI

=item C<< 'instance-uuid' => $string >>

Instance UUID
Applicable for CMode CLI

=item C<< '*' >>

This method takes all other options supported by "volume snapshot get Data Constituent" command in ONTAP

=item C<< other_options >>

This method takes other common options described in L<Common method options|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#Other_options_valid_on_component_methods>

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub get_Data_Constituent {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( reposId instance-uuid
                    )
            ),
        },
        cli_cmd => 'volume snapshot get_Data_Constituent',
    );
    $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_get_Data_Constituent_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::VolumeSnapshot::get_Data_Constituent',
    );

    $Log->exit() if $may_exit;
}

sub _get_Data_Constituent_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@opts,
        api => 'volume_snapshot_get_Data_Constituent',);
    $Log->exit() if $may_exit;
}

=head2 get_Constituent_SnapShot_ID

    NACL::C::VolumeSnapshot->get_Constituent_SnapShot_ID(
                                command_interface => $ci,
	                            'vserver' => $vserver,
	                            'snapshot' => $snapshot,
	                            'volume' => $volume,
                                ...);

								 or

    $VolumeSnapshot_obj->get_Constituent_SnapShot_ID(...);

(Class or instance method)
This method is used to get snapID of Infinite Volume Data Constituent.
Uses CMode CLI .

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'vserver' => $vserver >>

(Required for class method, Not Applicable for instance method)
Vserver
Applicable for CMode CLI

=item C<< 'snapshot' => $snapshot >>

(Required for class method, Not Applicable for instance method)
Snapshot
Applicable for CMode CLI

=item C<< 'volume' => $volume >>

(Required for class method, Not Applicable for instance method)
Volume
Applicable for CMode CLI

=item C<< 'constituentMsid' => $string >>

Constituent Msid
Applicable for CMode CLI

=item C<< 'reposId' => $string >>

Repository Id
Applicable for CMode CLI

=item C<< 'instance-uuid' => $string >>

Instance UUID
Applicable for CMode CLI

=item C<< '*' >>

This method takes all other options supported by "volume snapshot get Constituent SnapShot ID" command in ONTAP

=item C<< other_options >>

This method takes other common options described in L<Common method options|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#Other_options_valid_on_component_methods>

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub get_Constituent_SnapShot_ID {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( constituentMsid reposId instance-uuid
                    )
            ),
        },
        cli_cmd => 'volume snapshot get_Constituent_SnapShot_ID',
    );
    $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_get_Constituent_SnapShot_ID_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::VolumeSnapshot::get_Constituent_SnapShot_ID',
    );

    $Log->exit() if $may_exit;
}

sub _get_Constituent_SnapShot_ID_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@opts,
        api => 'volume_snapshot_get_Constituent_SnapShot_ID',);
    $Log->exit() if $may_exit;
}

=head2 sched_create

    my $VolumeSnapshot_obj	= NACL::C::VolumeSnapshot->sched_create(
                                command_interface => $ci,
	                            'vserver' => $vserver,
	                            'snapshot' => $snapshot,
	                            'volume' => $volume,
                                ...);

(Class method)
This method is used to kick off one run of a scheduled Infinite Volume snapshot job.
Uses CMode CLI .

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'vserver' => $vserver >>

(Required)
Vserver
Applicable for CMode CLI

=item C<< 'snapshot' => $snapshot >>

(Required)
Snapshot
Applicable for CMode CLI

=item C<< 'volume' => $volume >>

(Required)
Volume
Applicable for CMode CLI

=item C<< 'foreground' => $string >>

Foreground Process possible value(s) are, true,false
Applicable for CMode CLI

=item C<< 'schedule' => $string >>

Schedule
Applicable for CMode CLI

=item C<< '*' >>

This method takes all other options supported by "volume snapshot sched create" command in ONTAP

=item C<< other_options >>

This method takes other common options described in L<Common method options|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#Other_options_valid_on_component_methods>

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub sched_create {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( foreground schedule
                    )
            ),
        },
        cli_cmd => 'volume snapshot sched-create',
    );
    $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_sched_create_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::VolumeSnapshot::sched_create',
    );

    my $obj = $pkg->new(
        command_interface => $opts{command_interface},
        vserver           => $opts{'vserver'},
        snapshot          => $opts{'snapshot'},
        volume            => $opts{'volume'},
    );
    $Log->exit() if $may_exit;
    return $obj;
}

sub _sched_create_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, @opts) = @_;
    $pkg->_base_backend_cmode_cli(@opts,
        api => 'volume_snapshot_sched_create',);
    $Log->exit() if $may_exit;
}

=head2 fix_snaplock_state

    NACL::C::VolumeSnapshot->fix_snaplock_state(
                                command_interface => $ci,
	                            'volume' => $volume,
	                            'vserver' => $vserver,
                                ...);

(Class method)
This method is used to reconstruct SnapLock Snapshot state in the SnapLock Volume.
Uses CMode CLI.


=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>


=item C<< 'volume' => $volume >>

(Required)
Volume
Applicable for CMode CLI

=item C<< 'vserver' => $vserver >>

(Required)
Vserver
Applicable for CMode CLI

=item C<< '*' >>

This method takes all other options supported by "volume snapshot fix snaplock state" command in ONTAP

=item C<< other_options >>

This method takes other common options described in L<Common method options|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#Other_options_valid_on_component_methods>

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub fix_snaplock_state {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( volume vserver
                    )
            ),
        },
        ignore_primary_keys => 1,
        cli_cmd             => 'volume snapshot fix-snaplock-state',
    );
#    my $vserver;
#    $opts{vserver_operated_on} = \$vserver;
    $pkg->call_on_apiset_with_exceptions(
        %opts,
        choices => [
            {   method    => '_fix_snaplock_state_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
                zapi_type => 'none',
            },

        ],
        frontend => 'NACL::C::VolumeSnapshot::fix_snaplock_state',
    );

    $Log->exit() if $may_exit;
}

sub _fix_snaplock_state_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, @opts) = @_;
    $pkg->_base_backend_cmode_cli(@opts,
        api => 'volume_snapshot_fix_snaplock_state',);
    $Log->exit() if $may_exit;
}

=head2 modify_snaplock_expiry_time

    NACL::C::VolumeSnapshot->modify_snaplock_expiry_time(
                                command_interface => $ci,
	                            'snapshot' => $snapshot,
	                            'volume' => $volume,
	                            'vserver' => $vserver,
                                ...);

								 or

    $VolumeSnapshot_obj->modify_snaplock_expiry_time(...);

(Class or instance method)
This method is used to modify expiry time of a SnapLock Snapshot copy.
Uses CMode CLI/ZAPI .
Invokes  snapshot-modify-snaplock-expiry-time  for CMode Zapi non-iter

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>


=item C<< 'snapshot' => $snapshot >>

(Required for class method, Not Applicable for instance method)
Snapshot
Applicable for CMode CLI/ZAPI, non-iter.

=item C<< 'volume' => $volume >>

(Required for class method, Not Applicable for instance method)
Volume
Applicable for CMode CLI/ZAPI, non-iter.

=item C<< 'vserver' => $vserver >>

(Required for class method, Not Applicable for instance method)
Vserver
Applicable for CMode CLI/ZAPI, non-iter.

=item C<< 'infinite-expiry-time' => $string >>

Infinite SnapLock Expiry Time possible value(s) are, true,false
Applicable for CMode CLI/ZAPI, non-iter.

=item C<< 'expiry-time' => $string >>

SnapLock Expiry Time
Applicable for CMode CLI/ZAPI, non-iter.

=item C<< '*' >>

This method takes all other options supported by "volume snapshot modify snaplock expiry time" command in ONTAP

=item C<< other_options >>

This method takes other common options described in L<Common method options|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#Other_options_valid_on_component_methods>

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub modify_snaplock_expiry_time {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( infinite-expiry-time expiry-time
                    )
            ),
        },
        cli_cmd => 'volume snapshot modify-snaplock-expiry-time',
    );
    my $vserver;
    $opts{vserver_operated_on} = \$vserver;    #ZAPI is vserver family
    $pkg_or_obj->call_on_apiset_with_exceptions(
        %opts,
        choices => [
            {   method    => '_modify_snaplock_expiry_time_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
                zapi_type => 'none',
            },
            {   method => '_modify_snaplock_expiry_time_cmode_zapi_non_iter',
                interface => 'ZAPI',
                set       => 'CMode',
                zapi_type => 'non-iter',
            },
        ],
        frontend => 'NACL::C::VolumeSnapshot::modify_snaplock_expiry_time',
    );

    $Log->exit() if $may_exit;
}

sub _modify_snaplock_expiry_time_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@opts,
        api => 'volume_snapshot_modify_snaplock_expiry_time',);
    $Log->exit() if $may_exit;
}

sub _modify_snaplock_expiry_time_cmode_zapi_non_iter {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    my ($index) = grep { $opts[$_] eq 'expiry-time'} 0..$#opts;
    ## Convert expiry-time in epoch format if the input is in format MM/DD/YYYY HH:MM:SS or MM/DD/YYYY HH:MM:SS [{+|-}hh:mm]
    if ( $opts[$index+1] =~ /\d+\/\d+\/\d+\s+\d+\:\d+\:\d+$/ || $opts[$index+1] =~ /\d+\/\d+\/\d+\s+\d+\:\d+\:\d+\s+(\+|\-)\d+\:\d+/ ){
        use Date::Parse;
        $opts[$index+1] = str2time("$opts[$index+1]");
    }
    my $copy = [qw(infinite-expiry-time volume expiry-time snapshot)];
    $pkg_or_obj->_base_backend_cmode_zapi(
        @opts,
        api  => 'snapshot-modify-snaplock-expiry-time',
        copy => $copy,
    );
    $Log->exit() if $may_exit;
}


__PACKAGE__->_load_event_class();

sub _primary_keys_validate_spec {
    return (
        snapshot => {type => SCALAR},
        volume   => {type => SCALAR},
        vserver  => {type => SCALAR | UNDEF, optional => 1},
    );
} ## end sub _primary_keys_validate_spec

sub _restore_7mode_prompts {
    return [
        'Are you sure you want to do this' => 'y',
        'Proceed with revert'              => 'y',

        # Introduced in SN.0
        'Proceed with restore' => 'y',
    ];
} ## end sub _restore_7mode_prompts

sub _multi_base_backend_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg, @args ) = @_;
    my %opts = validate_with(
        params      => \@args,
        spec        => {
            api            => {type => SCALAR},
            map            => {type => HASHREF, default => {}},
            copy           => {type => ARRAYREF , default => []},
            _error_typedef => {type => SCALAR, default => 'volume-errors'},
        },
        allow_extra => 1,
    );
    my $error_typedef = delete $opts{_error_typedef};
    my $response;
    $pkg->_base_backend_cmode_zapi(%opts, apiset_response_obj =>\$response);
    my $output = $response->get_parsed_output()->[0];
    if ( exists $output->{$error_typedef}) {

         # If field-name is "volume-errors" ,inner typedef name is
         # "volume-error". In other words we can get inner typedef
         # name by just  removing the last character
         my $inner_typedef = substr $error_typedef,0,-1;
         my $errors        = $output->{$error_typedef}[0]{$inner_typedef};
         my $err_str       = 'The following error was encountered for the'
                             ."snapshot multi operation:\n";
         foreach my $error(@$errors) {
             $err_str .= _dump_one($error) . "\n";
         }
         NACL::C::Exceptions::VolumeSnapshot::MultiOperationFailed->throw(
             text             => $err_str,
             response_object  => $response,
             errors           => $errors,
         );
     }

     $Log->exit() if $may_exit;

} ## end sub _multi_base_backend_cmode_zapi

sub _backend_cmode_zapi_infinivol_builder {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts = validate_with(
        params => \@args,
        spec   => {
            api              => {type => SCALAR},
            map              => {type => HASHREF, default => {}},
            copy             => {type => ARRAYREF, default => []},
            is_infinivol_ref => {type => SCALARREF, optional => 1},
            non_async        => {type => HASHREF, default => {}},
        },
        allow_extra => 1
    );
    my $is_infinivol_ref = delete $opts{is_infinivol_ref};
    my $non_async_orig   = delete $opts{non_async};
    my %non_async        = validate_with(
        params => [%$non_async_orig],
        spec   => {
            map  => {type => HASHREF,  default => {}},
            copy => {type => ARRAYREF, default => []},
        }
    );

    my $is_infinivol = $pkg_or_obj->_determine_is_infinivol(%opts,apiset_must => { interface => "ZAPI" });

    if ($is_infinivol) {
        $opts{api} .= '-async';
    } else {
        $opts{copy} = [@{$opts{copy}}, @{$non_async{copy}}];
        $opts{map}  = {%{$opts{map}},  %{$non_async{map}}};
    }

    $$is_infinivol_ref = $is_infinivol if ($is_infinivol_ref);

    my $vserver;
    $opts{vserver_operated_on} ||= \$vserver;
    $pkg_or_obj->_base_backend_cmode_zapi(%opts);

    $Log->exit() if $may_exit;
}

sub call_on_apiset {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;

    # Reference which should be updated if is a constituent volume
    my $const_vol_ref = delete $opts{_is_const_vol_ref};
    my @ret;

    # ZAPI does not allow operations on constituent volumes. This code-ref
    # walks through the list of choices, removes any CMode-ZAPI choices,
    # and calls SUPER::call_on_apiset.
    my $exclude_zapi_and_run = sub {

        # Putting this at comment level, so it's clear what's going on
        $Log->comment('Snapshot operations on constituent volumes '
                . 'are not allowed through ZAPI. Attempt to drop back to '
                . 'the CLI to perform this operation');

        # Weed out any CMode-ZAPI choices
        my @choices;
        foreach my $choice (@{$opts{choices}}) {
            if ($choice->{interface} eq 'ZAPI' && $choice->{set} eq 'CMode') {
                next;
            } else {
                push @choices, $choice;
            }
        }
        @ret
            = $pkg_or_obj->SUPER::call_on_apiset(%opts, choices => \@choices);
    };

    if ($pkg_or_obj->_is_const_vol()) {
        $exclude_zapi_and_run->();
    } else {
        try {
            @ret = $pkg_or_obj->SUPER::call_on_apiset(%opts);
        }
        catch NACL::APISet::Exceptions::ResponseException with {
            my $exception = $_[0];
            if ($exception->text()
                =~ /Operation not supported on Infinite Volume constituents/i)
            {
                # This is the error message generated when attempting to use
                # ZAPI on a constituent volume
                $pkg_or_obj->_is_const_vol(1) if ref $pkg_or_obj;
                if ($const_vol_ref) {
                    $$const_vol_ref = 1;
                }
                $exclude_zapi_and_run->();
            } else {
                $Log->exit() if $may_exit;
                $exception->throw();
            }
        };
    }

    $Log->exit() if $may_exit;
    return wantarray ? @ret : $ret[0];
}

# These fields cannot be added to _common_validate_spec since they end up
# causing validation failures on the CS side, hence this extra spec for
# options that are common to all methods on this Component on the C side.
sub _additional_common_validate_spec {
    my ($pkg_or_obj, @opts) = @_;
    my $spec = $pkg_or_obj->SUPER::_additional_common_validate_spec(@opts);
    my $iv_spec = $pkg_or_obj->_add_infinivol_params_to_additional_common_validate_spec(@opts);
    $spec = {%$iv_spec, %$spec};
    return $spec;
}

1;
