# Copyright (c) 2001-2014 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary SnaphotPolicy Component Module
## @author Dheeraj.K@netapp.com, dl-nacl-dev@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::C::VolumeSnapshotPolicy

=head1 DESCRIPTION

C<NACL::C::VolumeSnapshotPolicy> is a derived class of
L<NACL::C::Component::ONTAP|lib-NACL-C-Component-ONTAP-pm>.  It represents access 
to an ONTAP VolumeSnapshotPolicy.  A related class is
L<NACL::CS::VolumeSnapshotPolicy|lib-NACL-CS-VolumeSnapshotPolicy-pm>, which represents
the past state of an ONTAP Aggregate.

=head1 ATTRIBUTES

=head2 command_interface

See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=head2 policy

The name of the policy, used to identify it uniquely.  (Policy
names are assumed to be unique within a node or within a Vserver)

=cut

package NACL::C::VolumeSnapshotPolicy;

use strict;
use warnings;

use Params::Validate
    qw(validate validate_with SCALAR SCALARREF ARRAYREF HASHREF BOOLEAN UNDEF OBJECT);
use NACL::CS::VolumeSnapshotPolicy;
use NACL::ComponentUtils
    qw (_optional_scalars Dumper _ontap_ci_validate_spec);
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use NACL::C::Cserver;
use NACL::Exceptions::InvalidChoice;
use NACL::CS::VolumeSnapshotPolicy;
use NACL::C::Exceptions::Component qw(:try);
use NACL::C::Exceptions::VolumeSnapshotPolicy::DoesNotExist qw(:try);
use NACL::C::Exceptions::VolumeSnapshotPolicy::AlreadyExists;
use NACL::APISet::Exceptions::ResponseException qw(:try);

use base 'NACL::C::Component::ONTAP';

=head1 METHODS 

=head2 new

  my $snap_policy = NACL::C::VolumeSnapshotPolicy->new(command_interface=>$ci,
                                         policy =>$policy_name);

Return a new VolumeSnapshotPolicy component object with the given attributes.

=cut

use Class::MethodMaker [
    scalar => 'policy',
    scalar => [
        {   -default_ctor => sub { $_[0]->_construct_vserver() }
        },
        'vserver',
    ],
];

sub init {
    $Log->enter() if $may_enter;
    my ($self, %opts) = @_;
    $self->SUPER::init(%opts);
    if ($opts{command_interface}->isa('NACL::C::Vserver')) {
        $self->_vserver_context_init(%opts);
    }
    $Log->exit() if $may_exit;
}

sub isa {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, $kind) = @_;
    my $isa = $pkg_or_obj->_build_isa(
        kind  => $kind,
        alias => 'NACL::C::SnapshotPolicy'
    );
    $Log->exit() if $may_exit;
    return $isa;
}

=head2 create

  my $snap_policy  = NACL::C::VolumeSnapshotPolicy->create(command_interface=>$ci,
                                              policy=>$policy_name,...);

(Class method) Request creation of a new snapshot policy element, and  return
a new snapshot policy component object that refers to it.  Does not wait for
snapshot policy creation to complete.

Uses a CMode CLI/ZAPI APISet.

Invokes "snapshot-policy-create" API for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< policy =>$policy_name >>

(Required) Name of snapshot policy to create.

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: policy

=item C<< 'vserver' => $string >>

Applicable for CMode CLI/ZAPI

(Optional for class method, Not Applicable for Instance method)
This is the vserver in which the policy is present. If not provided,
then the method operates on the admin vserver (i.e. the cserver).

For ZAPI, this specifies to run the context of the specified vserver

=item C<< enabled => [true|false] >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: enabled

=item C<< schedule1 => $schedule_name >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: schedule1

=item C<< count1 => $count >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: count1

=item C<< comment => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: comment

=item C<< schedule2 => $schedule_name >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: schedule2

=item C<< count2 => $count >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: count2

=item C<< schedule3 => $schedule_name >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: schedule3

=item C<< count3 => $count >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: count3

=item C<< schedule4 => $schedule_name >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: schedule4

=item C<< count4 => $count >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: count4

=item C<< schedule5 => $schedule_name >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: schedule5

=item C<< count5 => $count >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: count5

=item C<< 'prefix1' => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: prefix1

=item C<< 'prefix4' => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: prefix4

=item C<< 'prefix3' => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: prefix3

=item C<< 'prefix2' => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: prefix2

=item C<< 'prefix5' => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: prefix5

=item C<< 'snapmirror-label2' => $string >>

Applicable for CMode CLI

=item C<< 'snapmirror-label3' => $string >>

Applicable for CMode CLI

=item C<< 'snapmirror-label5' => $string >>

Applicable for CMode CLI

=item C<< 'snapmirror-label4' => $string >>

Applicable for CMode CLI

=item C<< 'snapmirror-label1' => $string >>

Applicable for CMode CLI

=item C<< "extended_query" => $string  >>

(Optional)

Applicable for CMode CLI

=item C<< other_options >>

This method takes other common options described in L<Common method options|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#Other_options_valid_on_component_methods>

=item C<< '*' >>

This method takes all other options supported by
"volume snapshot policy create" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub create {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw (enabled comment schedule1 count1
                    schedule2 count2 schedule3 count3
                    schedule4 count4 schedule5 count5
                    prefix1 prefix4 prefix3 prefix2 prefix5

                    snapmirror-label2 snapmirror-label3 snapmirror-label5 snapmirror-label4
                    snapmirror-label1

                    )
            )
        },
        cli_cmd => 'volume snapshot policy create',
    );

    my $vserver;
    $opts{vserver_operated_on} = \$vserver;
    try {
        $pkg->call_on_apiset(
            %opts,
            choices => [
                {   method    => '_create_cmode_cli',
                    interface => 'CLI',
                    set       => 'CMode',
                },
                {   method    => '_create_cmode_zapi',
                    interface => 'ZAPI',
                    set       => 'CMode',
                }    ## end

            ],
            frontend => 'NACL::C::VolumeSnapshotPolicy::create',
        );
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = $_[0];
        NACL::C::Exceptions::VolumeSnapshotPolicy::AlreadyExists
            ->detect_convert_and_throw(exception => $exception);
        $Log->exit() if $may_exit;
        $exception->throw();
    };

    my $obj = $pkg->new(
        policy            => $opts{policy},
        vserver           => $opts{vserver},
        command_interface => $opts{command_interface},
    );
    $Log->exit() if $may_exit;
    return $obj;
} ## end sub create

sub _create_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, @opts) = @_;
    $pkg->_base_backend_cmode_cli(@opts,
        api => 'volume_snapshot_policy_create',);
    $Log->exit() if $may_exit;
} ## end sub _create_cmode_cli

sub _create_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg, @opts) = @_;
    my $copy = [
        qw( policy enabled comment schedule1 count1
            schedule2 count2 schedule3 count3
            schedule4 count4 schedule5 count5 prefix1
            prefix2 prefix3 prefix4 prefix5
            snapmirror-label2 snapmirror-label3 snapmirror-label5
            snapmirror-label4 snapmirror-label1)
    ];
    $pkg->_base_backend_cmode_zapi(
        @opts,
        api  => 'snapshot-policy-create',
        copy => $copy
    );

    $Log->exit() if $may_exit;
} ## end sub _create_cmode_zapi

=head2 delete

  my $snap_policy = NACL::C::VolumeSnapshotPolicy->delete(
                                          command_interface=>$ci,
                                          policy =>$policy_name,
                                          ...);
                           or

  $snap_policy->delete(...);

(Class or instance method) This method deletes a snapshot policy element.

Uses a CMode CLI/ZAPI APISet.

Invokes "snapshot-policy-delete" API for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< policy=>$policy_name >>

(Required for class method, Not Applicable for instance method)
(Required) Name of policy to delete an snapshot policy.

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: policy

=item C<< 'vserver' => $string >>

Applicable for CMode CLI

(Optional for class method, Not Applicable for Instance method)
This is the vserver in which the policy is present. If not provided,
then the method operates on the admin vserver (i.e. the cserver).

=item C<< "extended_query" => $string  >>

(Optional)

Applicable for CMode CLI

=item C<< other_options >>

This method takes other common options described in L<Common method options|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#Other_options_valid_on_component_methods>

=item C<< '*' >>

This method takes all other options supported by
"volume snapshot policy delete" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub delete {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {},
        cli_cmd         => 'volume snapshot policy delete',
    );

    my $vserver;
    $opts{vserver_operated_on} = \$vserver;

    my $ret_obj;
    try {
        $ret_obj = $pkg_or_obj->call_on_apiset(
            %opts,
            choices => [
                {   method    => '_delete_cmode_cli',
                    interface => 'CLI',
                    set       => 'CMode',
                },
                {   method    => '_delete_cmode_zapi',
                    interface => 'ZAPI',
                    set       => 'CMode',
                }    ## end

            ],
            frontend => 'NACL::C::VolumeSnapshotPolicy::delete',
        );
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = $_[0];
        NACL::C::Exceptions::VolumeSnapshotPolicy::DoesNotExist
            ->detect_convert_and_throw(exception => $exception);
        $Log->exit() if $may_exit;
        $exception->throw();
    };
    $Log->exit() if $may_exit;
    return $ret_obj;
} ## end sub delete

sub _delete_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@opts,
        api => 'volume_snapshot_policy_delete',);
    $Log->exit() if $may_exit;
} ## end sub _delete_cmode_cli

sub _delete_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    $pkg_or_obj->_base_backend_cmode_zapi(
        @opts,
        api  => 'snapshot-policy-delete',
        copy => [qw (policy)],
    );
    $Log->exit() if $may_exit;
} ## end sub _delete_cmode_zapi

=head2 modify

  NACL::C::VolumeSnapshotPolicy->modify(command_interface=>$ci,
                             policy=>$snap_policy_name,...);

                            or

  $snap_policy_object->modify(...);

(Class or instance method) Modify an snapshot policy element.

Uses a CMode CLI/ZAPI APISet.

Invokes "snapshot-policy-modify" API for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< policy=>$snap_policy_name >>

(Required for class method, Not Applicable for instance method)
Name of snapshot policy.

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: policy

=item C<< enabled => $boolean >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: policy

=item C<< comment => $string >>

Applicable for CMode CLI/ZAPI

=item C<< 'vserver' => $string >>

Applicable for CMode CLI

(Optional for class method, Not Applicable for Instance method)
This is the vserver in which the policy is present. If not provided,
then the method operates on the admin vserver (i.e. the cserver).

=item C<< 'snapmirror-labels' => [ $snapmirror-label1, $snapmirror-label2, ... ] >>

Label for SnapMirror Operations
(Arrayref, even if only a single value is to be sent it should specified
as an arrayref containing that single value)

Applicable for CMode CLI

=item C<< "extended_query" => $string  >>

(Optional)

Applicable for CMode CLI

=item C<< other_options >>

This method takes other common options described in L<Common method options|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#Other_options_valid_on_component_methods>

=item C<< '*' >>

This method takes all other options supported by
"volume snapshot policy modify" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=item C<NACL::C::Exceptions::VolumeSnapshotPolicy::DoesNotExist>

Throws a C<NACL::C::Exceptions::VolumeSnapshotPolicy::DoesNotExist>,  if not matching snapshot
are found. 

=back

=back

=cut

sub modify {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            "snapmirror-labels" => {type => ARRAYREF, optional => 1},
            _optional_scalars(
                qw (enabled comment
                    )
            )
        },
        cli_cmd => 'volume snapshot policy modify',
    );
    my $vserver;
    $opts{vserver_operated_on} = \$vserver;
    my $ret_obj;
    try {
        $ret_obj = $pkg_or_obj->call_on_apiset(
            %opts,
            choices => [
                {   method    => '_modify_cmode_cli',
                    interface => 'CLI',
                    set       => 'CMode',
                },
                {   method    => '_modify_cmode_zapi',
                    interface => 'ZAPI',
                    set       => 'CMode',
                },

            ],
            frontend => 'NACL::C::VolumeSnapshotPolicy::modify',
        );
    }
    catch NACL::APISet::Exceptions::NoMatchingEntriesException with {
        my $exception = $_[0];
        NACL::C::Exceptions::VolumeSnapshotPolicy::DoesNotExist
            ->detect_convert_and_throw(exception => $exception);
        $Log->exit() if $may_exit;
        $exception->throw();
    };

    $Log->exit() if $may_exit;
    return $ret_obj;
} ## end sub modify

sub _modify_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@opts,
        api => 'volume_snapshot_policy_modify',);
    $Log->exit() if $may_exit;
} ## end sub _modify_cmode_cli

sub _modify_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    if (defined $opts{'snapmirror-labels'}) {
        my $sm_labels   = delete $opts{'snapmirror-labels'};
        my $label_count = scalar @$sm_labels;
        my $policy_obj  = NACL::CS::VolumeSnapshotPolicy->fetch(
            command_interface => $opts{command_interface},
            filter            => {
                policy  => $opts{policy},
                vserver => $opts{vserver},
            },
            requested_fields => [qw(schedules)],
            apiset_should    => {interface => 'ZAPI'}
        );
        my $schedules       = $policy_obj->schedules();
        my $sched_count     = scalar @$schedules;
        my $modify_schedule = sub {
            NACL::C::VolumeSnapshotPolicy->modify_schedule(
                command_interface => $opts{command_interface},
                policy            => $opts{policy},
                vserver           => $opts{vserver},
                apiset_should     => {interface => 'ZAPI'},
                @_
            );
        };
        if ($label_count == 1 && $sm_labels->[0] eq "") {
            foreach (@$schedules) {
                $modify_schedule->(
                    schedule              => $_,
                    'newsnapmirror-label' => "",
                );
            }
        } else {
            for (0 .. $sched_count - 1) {
                $modify_schedule->(
                    schedule              => $schedules->[$_],
                    'newsnapmirror-label' => $sm_labels->[$_]
                );
            }
        }
    }
    $pkg_or_obj->_base_backend_cmode_zapi(
        %opts,
        api  => 'snapshot-policy-modify',
        copy => [qw (enabled policy comment)],
    );
    $Log->exit() if $may_exit;
} ## end sub _modify_cmode_zapi

=head2 add_schedule

   NACL::C::VolumeSnapshotPolicy->add_schedule(
                                command_interface => $ci,
                                     policy => $policy,
                                  ...);

                                                                 or

    $VolumeSnapshotPolicy_object->add_schedule(...);

(Class or instance method)
This method is used to add a schedule to snapshot policy.

Uses a CMode CLI/ZAPI APISet.

Invokes "snapshot-policy-add-schedule" API for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "policy" => $policy >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: policy

=item C<< 'vserver' => $string >>

Applicable for CMode CLI

(Optional for class method, Not Applicable for Instance method)
This is the vserver in which the policy is present. If not provided,
then the method operates on the admin vserver (i.e. the cserver).

=item C<< "count" => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: count

=item C<< "schedule" => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: schedule

=item C<< "prefix" => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: prefix

=item C<< 'snapmirror-label' => $string >>

Applicable for CMode CLI

=item C<< "extended_query" => $string  >>

(Optional)

Applicable for CMode CLI

=item C<< other_options >>

This method takes other common options described in L<Common method options|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#Other_options_valid_on_component_methods>

=item C<< '*' >>

This method takes all other options supported by
"volume snapshot policy add-schedule" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub add_schedule {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( count schedule prefix

                    snapmirror-label

                    )
            ),
        },
        cli_cmd => 'volume snapshot policy add-schedule',
    );

    my $vserver;
    $opts{vserver_operated_on} = \$vserver;

    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_add_schedule_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_add_schedule_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            }    ## end

        ],
        frontend => 'NACL::C::VolumeSnapshotPolicy::add_schedule',
    );

    $Log->exit() if $may_exit;
    return $ret;
} ## end sub add_schedule

sub _add_schedule_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@opts,
        api => 'volume_snapshot_policy_add_schedule',);
    $Log->exit() if $may_exit;
} ## end sub _add_schedule_cmode_cli

sub _add_schedule_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    $pkg_or_obj->_base_backend_cmode_zapi(
        @opts,
        api  => 'snapshot-policy-add-schedule',
        copy => [qw(count policy prefix schedule snapmirror-label)]
    );
    $Log->exit() if $may_exit;
} ## end sub _add_schedule_cmode_zapi

=head2 remove_schedule

   NACL::C::VolumeSnapshotPolicy->remove_schedule(
                                command_interface => $ci,
                                     policy => $policy,
                                  ...);

                                                                 or

    $VolumeSnapshotPolicy_object->remove_schedule(...);

(Class or instance method)
This method is used to remove a schedule from snapshot policy.

Uses a CMode CLI/ZAPI APISet.

Invokes "snapshot-policy-remove-schedule" API for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "policy" => $policy >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: policy

=item C<< 'vserver' => $string >>

Applicable for CMode CLI

(Optional for class method, Not Applicable for Instance method)
This is the vserver in which the policy is present. If not provided,
then the method operates on the admin vserver (i.e. the cserver).

=item C<< "schedule" => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: schedule

=item C<< "extended_query" => $string  >>

(Optional)

Applicable for CMode CLI

=item C<< other_options >>

This method takes other common options described in L<Common method options|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#Other_options_valid_on_component_methods>

=item C<< '*' >>

This method takes all other options supported by
"volume snapshot policy remove-schedule" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub remove_schedule {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {_optional_scalars(qw( schedule)),},
        cli_cmd         => 'volume snapshot policy remove-schedule',
    );

    my $vserver;
    $opts{vserver_operated_on} = \$vserver;

    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_remove_schedule_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_remove_schedule_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            }    ## end

        ],
        frontend => 'NACL::C::VolumeSnapshotPolicy::remove_schedule',
    );

    $Log->exit() if $may_exit;
    return $ret;
} ## end sub remove_schedule

sub _remove_schedule_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@opts,
        api => 'volume_snapshot_policy_remove_schedule',);
    $Log->exit() if $may_exit;
} ## end sub _remove_schedule_cmode_cli

sub _remove_schedule_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    $pkg_or_obj->_base_backend_cmode_zapi(
        @opts,
        api  => 'snapshot-policy-remove-schedule',
        copy => [qw(policy schedule)],
    );
    $Log->exit() if $may_exit;
} ## end sub _remove_schedule_cmode_cli

=head2 modify_schedule

   NACL::C::VolumeSnapshotPolicy->modify_schedule(
                                command_interface => $ci,
                                     policy => $policy,
                                  ...);

                                                                 or

    $VolumeSnapshotPolicy_object->modify_schedule(...);

(Class or instance method)
This method is used to modify a schedule within snapshot policy.

Uses a CMode CLI/ZAPI APISet.

Invokes "snapshot-policy-modify-schedule" API for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "policy" => $policy >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: policy

=item C<< "newcount" => $string >>

Applicable for CMode CLI

=item C<< 'vserver' => $string >>

Applicable for CMode CLI

(Optional for class method, Not Applicable for Instance method)
This is the vserver in which the policy is present. If not provided,
then the method operates on the admin vserver (i.e. the cserver).

=item C<< "schedule" => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: schedule

=item C<< 'newsnapmirror-label' => $string >>

Applicable for CMode CLI

=item C<< "extended_query" => $string  >>

(Optional)

Applicable for CMode CLI

=item C<< other_options >>

This method takes other common options described in L<Common method options|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#Other_options_valid_on_component_methods>

=item C<< '*' >>

This method takes all other options supported by
"volume snapshot policy modify-schedule" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub modify_schedule {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( newcount schedule

                    newsnapmirror-label

                    )
            ),
        },
        cli_cmd => 'volume snapshot policy modify-schedule',
    );

    my $vserver;
    $opts{vserver_operated_on} = \$vserver;

    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_modify_schedule_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_modify_schedule_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            }
        ],
        frontend => 'NACL::C::VolumeSnapshotPolicy::modify_schedule',
    );

    $Log->exit() if $may_exit;
    return $ret;
} ## end sub modify_schedule

sub _modify_schedule_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@opts,
        api => 'volume_snapshot_policy_modify_schedule',);
    $Log->exit() if $may_exit;
} ## end sub _modify_schedule_cmode_cli

sub _modify_schedule_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    $pkg_or_obj->_base_backend_cmode_zapi(
        @opts,
        api  => 'snapshot-policy-modify-schedule',
        copy => [qw (policy schedule)],
        map  => {
            'newcount'            => 'new-count',
            'newsnapmirror-label' => 'new-snapmirror-label',
        }
    );

    $Log->exit() if $may_exit;
} ## end sub _modify_schedule_cmode_zapi

=head2 reset_prefix

    NACL::C::VolumeSnapshotPolicy->reset_prefix(
                                command_interface => $ci,
                                ...);

(Class method)
This method is used to set prefix to its associated schedule name..

Uses a CMode CLI APISet.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'method-timeout' => $time_in_seconds >>

(Optional)
The default NATE timeout is 60 seconds. If the command might take more
time to complete, this option can be used to specify a larger timeout value.
The value should be provided in seconds.

=item C<< "extended_query" => $string  >>

(Optional)

Applicable for CMode CLI

=item C<< other_options >>

This method takes other common options described in L<Common method options|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#Other_options_valid_on_component_methods>

=item C<< '*' >>

This method takes all other options supported by
"volume snapshot policy reset-prefix" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub reset_prefix {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params              => \@args,
        ignore_primary_keys => 1,
        api                 => 'volume_snapshot_policy_reset_prefix',
    );
    my $vserver;
    $opts{vserver_operated_on} = \$vserver;

    $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_reset_prefix_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::VolumeSnapshotPolicy::reset_prefix',
    );

    $Log->exit() if $may_exit;
} ## end sub reset_prefix

sub _reset_prefix_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, @opts) = @_;
    $pkg->_base_backend_cmode_cli(@opts,
        api => 'volume_snapshot_policy_reset_prefix',);
    $Log->exit() if $may_exit;
} ## end sub _reset_prefix_cmode_cli

=head2 update_snapmirror_labels

    NACL::C::VolumeSnapshotPolicy->update_snapmirror_labels(
                                command_interface => $ci,
                                'policy' => $policy,
                                ...);

                                 or

    $VolumeSnapshotPolicy_obj->update_snapmirror_labels(...);

(Class or instance method)
This method is used to update all of the SnapMirror labels in a Snapshot policy.

Uses a CMode CLI APISet.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'policy' => $policy >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI

=item C<< 'snapmirror-label' => $string >>

Applicable for CMode CLI

=item C<< 'method-timeout' => $time_in_seconds >>

(Optional)
The default NATE timeout is 60 seconds. If the command might take more
time to complete, this option can be used to specify a larger timeout value.
The value should be provided in seconds.

=item C<< 'vserver' => $string >>

Applicable for CMode CLI

=item C<< "extended_query" => $string  >>

(Optional)

Applicable for CMode CLI

=item C<< other_options >>

This method takes other common options described in L<Common method options|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#Other_options_valid_on_component_methods>

=item C<< '*' >>

This method takes all other options supported by
"volume snapshot policy update-snapmirror-labels" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub update_snapmirror_labels {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( snapmirror-label

                    )
            ),
        },
        api => 'volume_snapshot_policy_update_snapmirror_labels',
    );
    my $vserver;
    $opts{vserver_operated_on} = \$vserver;

    $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_update_snapmirror_labels_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::VolumeSnapshotPolicy::update_snapmirror_labels',
    );

    $Log->exit() if $may_exit;
} ## end sub update_snapmirror_labels

sub _update_snapmirror_labels_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@opts,
        api => 'volume_snapshot_policy_update_snapmirror_labels',);
    $Log->exit() if $may_exit;
} ## end sub _update_snapmirror_labels_cmode_cli


__PACKAGE__->_load_event_class();

sub _primary_keys_validate_spec {

    my %default_spec = (
        policy  => {type => SCALAR},
        vserver => {type => SCALAR | UNDEF, optional => 1}
    );

    return %default_spec;

} ## end sub _primary_keys_validate_spec

=head2 name_attribute

Returns 'policy'.

=cut

sub name_attribute {
    return 'policy';
}

sub _handle_zapi_vserver_context {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    my $ret = $pkg_or_obj->_handle_vscontext_helper(%opts,
        _method => '_handle_zapi_vserver_context');
    $Log->exit() if $may_exit;
    return $ret;
}

sub _handle_cli_vserver_context {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    my $ret = $pkg_or_obj->_handle_vscontext_helper(%opts,
        _method => '_handle_cli_vserver_context');
    $Log->exit() if $may_exit;
    return $ret;
}

sub _handle_vscontext_helper {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts = validate_with(
        params      => \@args,
        spec        => {_method => {type => SCALAR}},
        allow_extra => 1
    );
    my $method = delete $opts{_method};
    my $vserver = $opts{api_opts}->{vserver} || $opts{vserver};
    if (   !$opts{command_interface}->isa('NACL::C::Vserver')
        && !defined $vserver)
    {

        # The base class implementation will either throw a warning (for
        # CLI) for error out (for ZAPI) if the command interface is not
        # a vserver and vserver was not passed in %opts. However, for this
        # component ONTAP will take this to mean that the admin vserver
        # is the vserver, so we should not fail or warn for this case.
        $Log->exit() if $may_exit;
        return;
    } else {
        my $super_method = "SUPER::$method";
        $vserver = $pkg_or_obj->$super_method(%opts);
        $Log->exit() if $may_exit;
        return $vserver;
    }
}

# Even after these commands were vserverized, it is possible to run these
# commands on Node/Cluster shell without the vserver argument - in this
# case ONTAP uses the admin vserver. This -default_ctor is to handle
# cases where the object was constructed with a vserver
sub _construct_vserver {
    $Log->enter() if $may_enter;
    my $self = $_[0];
    unless (ref $self) {
        $Log->exit() if $may_exit;
        return;
    }

    my $ci      = $self->command_interface();
    my $boolean = $ci->is_feature_supported(
        feature => 'snapshotpolicy_vserverization');
    if ($boolean) {
        my $name = NACL::C::Cserver->find(command_interface => $ci)->name();
        $Log->exit() if $may_exit;
        return $name;
    } else {
        $Log->exit() if $may_exit;
        return undef;
    }
}

1;
