#
# Copyright (c) 2001-2016 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Vserver Component Module
## @author dl-nacl-dev@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::C::Vserver

=head1 DESCRIPTION

C<NACL::C::Vserver> is a derived class of
L<NACL::C::CommandInterface::ONTAP|lib-NACL-C-CommandInterface-ONTAP-pm>.
It represents access to an ONTAP Vserver.  A related class is
L<NACL::CS::Vserver|lib-NACL-CS-Vserver-pm>, which represents
the past state of an ONTAP Vserver. A Vserver can act as command interface
to other components.See
L<NACL::C::Component::command_interface|lib-NACL-C-Component-pm/command_interface>

=head1 ATTRIBUTES

=head2 vserver

The name of the vserver, used to identify it uniquely.

=cut

package NACL::C::Vserver;

use strict;
use warnings;
use NACL::CS::Vserver;
use NACL::ComponentUtils qw(_optional_scalars);
use Params::Validate qw(:all);
use base 'NACL::C::CommandInterface::ONTAP';
use NACL::C::Job;
use NACL::CS::Volume;
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use NATE::Exceptions::Argument qw(:try);
use NACL::Exceptions::NoElementsFound                   ();
use NACL::C::Exceptions::Vserver::AlreadyExists         ();
use NACL::C::Exceptions::Vserver::DoesNotExist          ();
use NACL::C::Exceptions::Vserver::VolumesExist          ();
use NACL::C::Exceptions::Vserver::InfiniteVolumesExist  ();
use NACL::C::Exceptions::Vserver::IgroupsExist          ();
use NACL::C::Exceptions::Vserver::CifsShareExists       ();
use NACL::C::Exceptions::Vserver::TpgroupsExist         ();
use NACL::C::Exceptions::Vserver::PortsetsExist         ();
use NACL::C::Exceptions::Vserver::IscsiIsnsExist        ();
use NACL::C::Exceptions::Vserver::IscsiSecurityExist    ();
use NACL::C::Exceptions::Vserver::PeerRelationshipExist ();
use NACL::C::Exceptions::Vserver::AlreadyStarted        ();
use NACL::C::Exceptions::Vserver::AlreadyStopped        ();
use NACL::C::Exceptions::Vserver::LifEnabled	        ();
use NACL::Exceptions::InvalidChoice                     ();
use NACL::C::_Mixins::Vserver qw(:all);
use NATE::BaseException qw(:try);
use NACL::CS::VserverOptions;

use NACL::STask::VserverPeer;

=head1 Linked components

(See
http://wikid.netapp.com/w/QA/projects/Libraries_Initiative/Project/Common_Infrastructure/Users_Guide/Component_Layer_Users_Guide#Linked_components
for a more detailed explanation on linked components)

This component is linked to the following components:

=begin html

<table border="1">
<tr> <th>Prefix</th> <th>Linked Component</th> </tr>
<tr> <td>cifs</td> <td>VserverCifs</td> </tr>
<tr> <td>nfs</td> <td>VserverNfs</td> </tr>
<tr> <td>peer</td> <td>VserverPeer</td> </tr>
<tr> <td>snapmirror_policy</td> <td>SnapmirrorPolicy</td> </tr>
<tr> <td>security_file_directory</td> <td>VserverSecurityFileDirectory</td> </tr>
<tr> <td>security_file_directory_job</td> <td>VserverSecurityFileDirectoryJob</td> </tr>
<tr> <td>security_file_directory_ntfs</td> <td>VserverSecurityFileDirectoryNtfs</td> </tr>
<tr> <td>security_file_directory_ntfs_dacl</td> <td>VserverSecurityFileDirectoryNtfsDacl</td> </tr>
<tr> <td>security_file_directory_ntfs_sacl</td> <td>VserverSecurityFileDirectoryNtfsSacl</td> </tr>
<tr> <td>security_file_directory_policy</td> <td>securityfiledirectory_policy</td> </tr>
<tr> <td>security_file_directory_policy_task</td> <td>VserverSecurityFileDirectoryPolicyTask</td> </tr>
</table>

=end html

Examples:

    # Invoking the "delete" method of the VserverCifs component
    $vs_obj->cifs_delete();

    # Invoking the "create" method of the VserverNfs component
    my $nfs_obj = $vs_obj->nfs_create();

=cut

sub can {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my @opts = $pkg_or_obj->_can_validation(@args);
    my $code;
    if ($code = $pkg_or_obj->NACL::C::Component::can(@opts)) {
        $Log->exit() if $may_exit;
        return $code;
    } else {
        $code = $pkg_or_obj->_find_wrapper_generator(@opts);
        if ($code) {
            $Log->exit() if $may_exit;
            return $code;
        } else {
            $code = $pkg_or_obj->_handle_component_aggregation(
                method    => $opts[0],
                error_ptr => $opts[1],
                table     => [
                    {   component => 'NACL::C::VserverCifs',
                        prefix    => 'cifs'
                    },
                    {   component => 'NACL::C::VserverNfs',
                        prefix    => 'nfs'
                    },
                    {   component => 'NACL::C::VserverPeer',
                        prefix    => 'peer'
                    },
                    {   component => 'NACL::C::SnapmirrorPolicy',
                        prefix    => 'snapmirror_policy'
                    },
                    {   component => 'NACL::C::VserverSecurityFileDirectoryJob',
                        prefix    => 'security_file_directory_job'
                    },
                    {   component => 'NACL::C::VserverSecurityFileDirectoryNtfsDacl',
                        prefix    => 'security_file_directory_ntfs_dacl'
                    },
                    {   component => 'NACL::C::VserverSecurityFileDirectoryNtfsSacl',
                        prefix    => 'security_file_directory_ntfs_sacl'
                    },
                    {   component => 'NACL::C::VserverSecurityFileDirectoryNtfs',
                        prefix    => 'security_file_directory_ntfs'
                    },
                    {   component => 'NACL::C::VserverSecurityFileDirectoryPolicyTask',
                        prefix    => 'security_file_directory_policy_task'
                    },
                    {   component => 'NACL::C::VserverSecurityFileDirectoryPolicy',
                        prefix    => 'security_file_directory_policy'
                    },
                    {   component => 'NACL::C::VserverSecurityFileDirectory',
                        prefix    => 'security_file_directory'
                    },
                ]
            );
            if ($code) {
                $Log->exit() if $may_exit;
                return $code;
            } elsif ($code = $pkg_or_obj->SUPER::can(@opts)) {
                $Log->exit() if $may_exit;
                return $code;
            }
        } ## end else [ if ($code) ]
    } ## end else [ if ( $code = $pkg_or_obj...)]

    $Log->exit() if $may_exit;
    return undef;
} ## end sub can

=head1 METHODS

=head2 new

my $vserver = NACL::C::Vserver->new([name => $name],
                              [vserver => $vserver],
                              [hostrec => $hostrec] );
Specifying any one of the vserver, name, hostrec attributes will populate
the other two.
Returns a new Vserver component object with the given attributes.

=cut

use Class::MethodMaker [
    scalar => [
        {   '-default_ctor' => sub { $_[0]->_construct_primary_key(); },
            '-store_cb' =>
                sub { $_[0]->_verify_hostrec_callback($_[1], $_[2]); }
        },
        'vserver',
    ],
    scalar => [
        {   '-default_ctor' => sub {
                $_[0]->determine_is_repository();
                }
        },
        '_is_repository',
    ]
];

# Backwards compatibility for the "is_repository" method
no strict 'refs';
*is_repository = *_is_repository;
use strict 'refs';

# ...

sub default_apiset_must {
    my ($self) = @_;
    return {
        %{$self->SUPER::default_apiset_must()},
        set        => "Vserver|CMode",
        set_source => "command interface is a NACL::C::Vserver",
    };
} ## end sub default_apiset_must

# First preference is to connect to the Vserver and run commands directly on it.
# If the user wants to run it through the node (i.e: "vserver context"), they'd
# need to specify the shell_type to be CMode.
sub default_apiset_could {
    return {
        shell_type => "Vserver",
        category   => "Node",
        interface  => "CLI",
        set        => "Vserver"
    };
} ## end sub default_apiset_could

=head2 find

    my $Vserver  = NACL::C::Vserver->find(command_interface => $ci, ...);
    my @Vservers = NACL::C::Vserver->find(command_interface => $ci, ...);

(Class method)
A constructor of sorts, which discovers which elements are present (by
interacting with the given command interface).  In an array context it
creates a component object for each present element and returns them.
In a scalar context it creates a component object for just one
element.

This method is generally not implemented in individual components. The base
class (L<NACL::C::Component::ONTAP|lib-NACL-C-Component-ONTAP-pm> provides an implementation.
See <L<NACL::C::Component::find|lib-NACL-C-Component-pm/find> for a more
detailed description along with explanation of all the options accepted.

You can also refer to user guide for more information on the options,
order of precedence for command line parameters (in case of find() is
invoked without specifying a command_interface) and examples.
http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/ComponentLayerUsersGuide/index.html#Obtaining_command_interface_objects

=head2 CANNED FILTERS

See
http://wikid.netapp.com/w/QA/projects/Libraries_Initiative/Project/Common_Infrastructure/Users_Guide/Component_Layer_Users_Guide#Canned_filters
for a description of canned filters.

=head3 data_vservers

Returns data vservers, i.e. those that have type as "cluster" or "data".

=head2 repository_vservers

Returns repository vservers, i.e. those that have their "is-repository" field
being "true".

=head2 non_repository_vservers

Returns non-repository vservers, i.e. those that have their "is-repository"
fields being "false".

This will include node and admin vservers.

=head2 data_non_repository_vservers

Returns vservers with are both data vservers as well as non-repository
vservers.

=head2 FIND WRAPPERS

Each of the canned filters have a wrapper method (implemented through
AUTOLOAD) of the form:

    find_<canned-filter-name>

For example, there is a C<find_data_vservers> method for the C<data_vservers>
canned filter.

    NACL::C::Vserver->find_data_vservers(...);

These methods accept all of the arguments normally accepted by find()
but automatically apply the canned filter contained in its name.

=head2 state

    my $Vserver_state = $vserver_obj->state();

(Instance method)
Fetch the current state of the element associated
with this component, and return a Component State object corresponding
to this component (calling this on NACL::C::Vserver would return
a NACL::CS::Vserver object)

This method is generally not implemented in individual components. The base
class (L<NACL::C::Component::ONTAP|lib-NACL-C-Component-ONTAP-pm> provides an implementation.
See <L<NACL::C::Component::state|lib-NACL-C-Component-pm/state> for a more
detailed description along with explanation of all the options accepted.

=head2 create

  my $vserver  = NACL::C::Vserver->create(command_interface=>$ci,
                                          vserver =>$vserver_name,...);

 (Class method) Request creation of a new vserver element, and  return a new
 vserver component object that refers to it.  Does not wait for vserver
 creation to complete.

Uses a CMode CLI/ZAPI/GUI APISet.

Invokes "vserver-create" API for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< vserver=>$vserver_name >>

(Required) Name of vserver to create.

Applicable for CMode CLI/ZAPI/GUI.

Maps to:

CM ZAPI: vserver-name

=item C<< rootvolume => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: root-volume

=item C<< aggregate => $string >>

Applicable for CMode CLI/ZAPI/GUI.

Maps to:

CM ZAPI: root-volume-aggregate

=item C<< "ns-switch" => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: name-server-switch

For ui-change ns-switch-ui this option will be hidden for 'vserver create'
command,Use NACL::C::VserverServicesNameServiceNsSwitch->create()
to create the vserver ns-switch.

=item C<< "rootvolume-security-style" => $string >>

Applicable for CMode CLI/ZAPI/GUI.

Maps to:

CM ZAPI: root-volume-security-style

=item C<< language => $string >>

Applicable for CMode CLI/ZAPI/GUI.

Maps to:

CM ZAPI: language

=item C<< "snapshot-policy" => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: snapshot-policy

=item C<< "antivirus-on-access-policy" => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: antivirus-on-access-policy

=item C<< comment => $string >>

Comment (a string up to a length of 255 characters)

Applicable to CMode CLI/ZAPI.

Maps to:

CMode ZAPI: comment

=item C<< "quota-policy => $string >>

Quota Policy.

Applicable to CMode CLI/ZAPI.

Maps to:

CMode ZAPI: quota-policy

=item C<< "job_component"=> \$scalar >>

Applicable to CMode CLI

=item C<< "cmd_execution_time"=> \$scalar >>

(Optional) Populate the param with execution time of command

=item C<< 'admin-state' => $string >>

Applicable to CMode CLI

=item C<< 'is-repository' => $string >>

Applicable to CMode CLI/ZAPI/GUI.

Maps to:

CMode ZAPI: is-repository-vserver

=item C<< 'ipspace' => $string >>

Applicable to CMode CLI/ZAPI/GUI.

=item C<< 'qos-policy-group' => $string >>

Applicable to CMode CLI

=item C<< 'external-cache' => $string >>

External Cache Policy

Applicable for CMode CLI

=item C<< 'subtype' => $string >>

Vserver subtype
Applicable for CMode CLI/ZAPI

Map to:

CM ZAPI: vserver-subtype

=item C<< 'snapshot-policy-uuid' => $string >>

Snapshot Policy UUID
Applicable for CMode CLI

=item C<< 'caching-policy' => $string >>

Caching Policy Name
Applicable for CMode CLI/ZAPI

Maps to:

CM ZAPI: caching-policy

=item C<< other_options >>

This method takes other common options described in L<Common method options|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#Other_options_valid_on_component_methods>

=item C<< 'next-rtpid' => $string >>

Next Available RTPID
Applicable for CMode CLI

=item C<< 'san-tgt-dev-id' => $string >>

SAN Target Device ID
Applicable for CMode CLI

=item C<< 'name' => $string >>

Vserver name for config replication
Applicable for CMode CLI

=item C<< 'uuidforEngineID' => $string >>

UUID for generating EngineID
Applicable for CMode CLI

=item C<< '*' >>

This method takes all other options supported by
"vserver create" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item GUI-specific options 

=item C<< 'dnsDomain' = $string >>

=item C<< 'dnsServer' = $string >>

=item C<< 'protocol' = [$protocol1, $protocol2..] >>

=item C<< 'ddns' = $string >>

=item C<< 'foreground' = $string >>

run process in foreground if foreground => 'true'

=back

=over

=item Exceptions

=over

=item C<NACL::C::Exceptions::Vserver::AlreadyExists>

This type of exception gets thrown if an attempt is made to create a
vserver that already exists.

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub create {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            job_component      => {type => SCALARREF,         optional => 1},
            cmd_execution_time => {type => SCALARREF,         optional => 1},
            'nm-switch'        => {type => ARRAYREF | SCALAR, optional => 1},
            'ns-switch'        => {type => ARRAYREF | SCALAR, optional => 1},
            _optional_scalars(
                qw(rootvolume aggregate
                    rootvolume-security-style language snapshot-policy
                    antivirus-on-access-policy quota-policy comment
                    admin-state is-repository
                    ipspace qos-policy-group external-cache
                    subtype snapshot-policy-uuid caching-policy
                    next-rtpid san-tgt-dev-id name
                    uuidforEngineID
                    dnsDomain dnsServer protocol ddns foreground
                    )
            ),
        },
        cli_cmd => 'vserver create',
    );

    if ($pkg->_has_nsswitch_ui(%opts)) {
        my $warn_message = "The fields nm-switch and ns-switch are deprecated. "
                      . "Use 'vserver services name-service ns-switch create' command "
                      . "to configure name-service and name-mapping\n";
        NACL::C::Component->_print_deprecation_warning($warn_message);
    }

    try {
        $pkg->call_on_apiset(
            %opts,
            choices => [
                {   method    => "_create_cmode_cli",
                    interface => "CLI",
                    set       => "CMode",
                },

                {   method    => "_create_cmode_zapi",
                    interface => "ZAPI",
                    set       => "CMode",
                },

                {   method    => "_create_cmode_gui",
                    interface => "GUI",
                    set       => "CMode",
                },

            ],
            frontend => 'NACL::C::Vserver::create'
        );
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = $_[0];
        NACL::C::Exceptions::Vserver::AlreadyExists
            ->detect_convert_and_throw(exception => $exception);
        $Log->exit() if $may_exit;
        $exception->throw();
    };

    my $obj = $pkg->new(
        vserver           => $opts{vserver},
        command_interface => $opts{command_interface},
    );

    $Log->exit() if $may_exit;
    return $obj;
} ## end sub create

sub _create_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, @opts) = @_;

    my %opts = validate_with(
        params      => \@opts,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );

    $pkg->_array_to_scalar_deprication_warning(
        params  => ['nm-switch', 'ns-switch'],
        options => \%opts
    );

    $pkg->_base_backend_cmode_cli(%opts, api => 'vserver_create',);
    $Log->exit() if $may_exit;
} ## end sub _create_cmode_cli

sub _create_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg, %opts) = @_;

    # Explanation present in the header for this private method
    $pkg->_handle_subtype_zapi_val(options => \%opts);

    my $copy = [
        qw(antivirus-on-access-policy caching-policy comment ipspace
           language quota-policy qos-policy-group snapshot-policy)
    ];

    my $map = {
        vserver                     => 'vserver-name',
        aggregate                   => 'root-volume-aggregate',
        rootvolume                  => 'root-volume',
        'rootvolume-security-style' => 'root-volume-security-style',
        'is-repository'             => 'is-repository-vserver',
        subtype                     => 'vserver-subtype',
        'nm-switch'                 => 'name-mapping-switch',
        'ns-switch'                 => 'name-server-switch',
    };

    my $ci = $opts{command_interface};
    my $api = 'vserver_create';
    if(defined $opts{foreground}){
        if($opts{foreground} =~ /false/i){
            $api .= '_async';
        }
    }
    $pkg->_base_backend_cmode_zapi(
        %opts,
        api  => $api,
        copy => $copy,
        map  => $map,
    );

    $Log->exit() if $may_exit;
} ## end sub _create_cmode_zapi

sub _create_cmode_gui {
    $Log->enter() if $may_enter;
    my ( $pkg, @args ) = @_;
    my %opts = $pkg->_validate_backend_opts(@args);

    my $copy = [
        qw( vserver ipspace language is-repository protocol
            aggregate dnsDomain dnsServer ddns rootvolume-security-style )
    ];

    $pkg->_base_backend_cmode_gui(
        %opts,
        api  => 'createVserver',
        copy => $copy,
    );

    $Log->exit() if $may_exit;

} ## end sub _create_cmode_gui

=head2 delete

 NACL::C::Vserver->delete(command_interface=>$ci,
                             vserver=>$vserver_name,...);

                            or

  $vserver_object->delete(...);

(Class or instance method) This method deletes a vserver element.

Uses a CMode CLI/ZAPI/GUI APISet.

Invokes "vserver-destroy" API for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
 See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< vserver =>$vserver_name >>

(Required for class method, Not Applicable for instance method)
 Name of vserver.

Applicable for CMode CLI/ZAPI/GUI.

Maps to:

CM ZAPI: vserver-name

=item C<< "cmd_execution_time"=> \$scalar >>

(Optional) Populate the param with execution time of command

=item C<< "extended_query" => $string  >>

(Optional)

Applicable for CMode CLI

=item C<< other_options >>

This method takes other common options described in L<Common method options|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#Other_options_valid_on_component_methods>

=item C<< 'foreground => $string >>

run process in foreground if foreground => 'true'

Applicable for CMode CLI

=item C<< '*' >>

This method takes all other options supported by
"vserver delete" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=item C<NACL::C::Exceptions::Vserver::DoesNotExist>

This type of exception gets thrown if an attempt is made to delete a
non-existing vserver.

=item C<NACL::C::Exceptions::Vserver::InfiniteVolumesExist>

This type of exception gets thrown if infinite volumes exist on vserver.

=item C<NACL::C::Exceptions::Vserver::PortsetsExist>

This type of exception gets thrown if portsets are associated with vserver.

=item C<NACL::C::Exceptions::Vserver::CifsShareExists>

This type of exception gets thrown if cifs share exists on vserver.

=item C<NACL::C::Exceptions::Vserver::IgroupsExist>

This type of exception gets thrown if igroups are associated with vserver.

=item C<NACL::C::Exceptions::Vserver::IscsiIsnsExist>

This type of exception gets thrown if iSCSI iSNS are associated with vserver.

=item C<NACL::C::Exceptions::Vserver::IscsiSecurityExist>

This type of exception gets thrown if iSCSI Security are associated with vserver.

=item C<NACL::C::Exceptions::Vserver::TpgroupsExist>

This type of exception gets thrown if tpgroups associated with vserver.

=item C<NACL::C::Exceptions::Vserver::VolumesExist>

This type of exception gets thrown if volumes exist in vserver.

=item C<NACL::C::Exceptions::Vserver::PeerRelationshipExist>

This type of exception gets thrown if vserver peer relationship exists

=item C<NACL::C::Exceptions::Vserver::MirroredVolumesExist>

This is exception is raised when an attempt is made to delete a vserver which has
mirrors configured on it. Probably the vserver will be part of a destination in a
Vserver DR configuration.

=item C<NACL::C::Exceptions::Vserver::CannotDeleteDPVserver>

This exception is raised when the vserver that is about to be deleted is a destination
vserver in a Vserver DR configuration.

=back

=back

=cut

sub delete {
    my ($pkg_or_obj, @args) = @_;
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( vserver foreground 
                    )
            ),
            cmd_execution_time => {type => SCALARREF, optional => 1},
        },
        cli_cmd => 'vserver delete',
    );

    my $ret;
    try {
        $ret = $pkg_or_obj->call_on_apiset(
            %opts,
            choices => [
                {   method    => '_delete_cmode_cli',
                    interface => 'CLI',
                    set       => 'CMode',
                },
                {   method    => '_delete_cmode_zapi',
                    interface => 'ZAPI',
                    set       => 'CMode',
                },
                {   method    => '_delete_cmode_gui',
                    interface => 'GUI',
                    set       => 'CMode',
                },

            ],
            frontend => 'NACL::C::Vserver::delete'
        );
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = $_[0];
        NACL::C::Exceptions::Vserver::DoesNotExist->detect_convert_and_throw(
            exception => $exception);
        NACL::C::Exceptions::Vserver::PeerRelationshipExist
            ->detect_convert_and_throw(exception => $exception);
        NACL::C::Exceptions::Vserver::VolumesExist->detect_convert_and_throw(
            exception => $exception);
        NACL::C::Exceptions::Vserver::TpgroupsExist
            ->detect_convert_and_throw(exception => $exception);
        NACL::C::Exceptions::Vserver::IgroupsExist->detect_convert_and_throw(
            exception => $exception);
        NACL::C::Exceptions::Vserver::IscsiIsnsExist
            ->detect_convert_and_throw(exception => $exception);
        NACL::C::Exceptions::Vserver::IscsiSecurityExist
            ->detect_convert_and_throw(exception => $exception);
        NACL::C::Exceptions::Vserver::CifsShareExists
            ->detect_convert_and_throw(exception => $exception);
        NACL::C::Exceptions::Vserver::PortsetsExist
            ->detect_convert_and_throw(exception => $exception);
        NACL::C::Exceptions::Vserver::InfiniteVolumesExist
            ->detect_convert_and_throw(exception => $exception);
        NACL::C::Exceptions::Vserver::MirroredVolumesExist
            ->detect_convert_and_throw(exception => $exception);
        NACL::C::Exceptions::Vserver::CannotDeleteDPVserver
            ->detect_convert_and_throw(exception => $exception);
        NACL::C::Exceptions::Vserver::LifEnabled
            ->detect_convert_and_throw(exception => $exception);
        $Log->exit() if $may_exit;
        $exception->throw();
    };
    return $ret;
} ## end sub delete

sub _delete_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@opts, api => 'vserver_delete',);
    $Log->exit() if $may_exit;
} ## end sub _delete_cmode_cli

sub _delete_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg, %args) = @_;
    my $api = 'vserver_destroy';
    if(defined $args{foreground}){
        if($args{foreground} =~ /false/i){
            $api .= '_async';
        }
    }
    $pkg->_base_backend_cmode_zapi(%args,
        api    => $api,
        map    => {"vserver" => "vserver-name",},
    );
    $Log->exit() if $may_exit;
} ## end sub _delete_cmode_zapi

sub _delete_cmode_gui {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    $pkg_or_obj->_base_backend_cmode_gui( @args,
        api  => 'deleteVserver',
        copy => [ 'vserver' ],
    );
    $Log->exit() if $may_exit;
} ## end sub _delete_cmode_gui

=head2 modify

    NACL::C::Vserver->modify(
        command_interface => $ci,
        vserver           => $vserver_name,
        ...
    );

                            or

  $vserver_object->modify(...);

(Class or instance method) This method modifies a vserver element.

Uses a CMode CLI/ZAPI/GUI APISet.

Supports "vserver-modify-iter" or "vserver-modify" API for CMode ZAPI.
"vserver-modify" is selected when
1. selected interface is "ZAPI" and
2. zapi_type is selected as "non-iter"

"vserver-modify-iter" is default "ZAPI" API for modify.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
 See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< vserver =>$vserver_name >>

(Required for class method, Not Applicable for instance method)
 Name of vserver.

Applicable for CMode CLI/ZAPI/GUI.

Maps to:

CM ZAPI:

<vserver-modify-iter>
<query>
<vserver-info>
<vserver-name>$value</vserver-name>
</vserver-info>
</query>
</vserver-modify-iter>

=item C<< "ns-switch" => [ $ns_sw1, $ns_sw2, ... ] >>

(Arrayref, even if only a single value is to be sent it should be specified as
an arrayref containing that single value)

Name Service Switch. Valid values: nis | file | ldap

Applicable for CMode CLI/ZAPI/GUI.

Maps to:

CM ZAPI:

<vserver-modify-iter>
<query>
<vserver-info>
<name-server-switch>@values</name-server-switch>
</vserver-info>
</query>
</vserver-modify-iter>

For ui-change ns-switch-ui this option will be hidden for 'vserver modify'
command,Use NACL::C::VserverServicesNameServiceNsSwitch->modify()
to modify the vserver ns-switch value.

=item C<< "nm-switch" => [ $nm_sw1, $nm_sw2, ... ] >>

(Arrayref, even if only a single value is to be sent it should be specified as
an arrayref containing that single value)

Name Mapping Switch. Valid values: file | ldap

Applicable for CMode CLI/ZAPI/GUI.

Maps to:

CM ZAPI:

<vserver-modify-iter>
<query>
<vserver-info>
<name-mapping-switch>@values</name-mapping-switch>
</vserver-info>
</query>
</vserver-modify-iter>

For ui-change ns-switch-ui this option will be hidden for 'vserver modify'
command,Use NACL::C::VserverServicesNameServiceNsSwitch->modify()
to modify the vserver nm-switch value.

=item C<< comment => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

<vserver-modify-iter>
<query>
<vserver-info>
<comment>$value</comment>
</vserver-info>
</query>
</vserver-modify-iter>

=item C<< "snapshot-policy" => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

<vserver-modify-iter>
<query>
<vserver-info>
<snapshot-policy>$value</snapshot-policy>
</vserver-info>
</query>
</vserver-modify-iter>

=item C<< "antivirus-on-access-policy" => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

<vserver-modify-iter>
<query>
<vserver-info>
<antivirus-on-access-policy>$value</antivirus-on-access-policy>
</vserver-info>
</query>
</vserver-modify-iter>

=item C<< "quota-policy" => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

<vserver-modify-iter>
<query>
<vserver-info>
<quota-policy>$value</quota-policy>
</vserver-info>
</query>
</vserver-modify-iter>

=item C<< "maxsshsessions" => $integer >>

Applicable for CMode CLI

=item C<< "maxvolumes" => $integer >>

Applicable for CMode CLI

=item C<< maxStorage => $string >>

Applicable for CMode CLI

=item C<< "admin-state" => $string >>

Applicable for CMode CLI

=item C<< aggrlist => [$aggr1, $aggr2, $aggr3, ... ] >>

Applicable for CMode CLI/GUI

=item C<< 'aggr-list' => [ $aggr1, $aggr2, ... ] >>

(Arrayref, even if only a single value is to be sent it should specified
as an arrayref containing that single value)

List of aggregates assigned to the Vserver.

Applicable for CMode CLI/ZAPI/GUI.

Maps to:

CMode ZAPI: aggr-list

<vserver-modify-iter>
<query>
<vserver-info>
<aggr-list>
<aggr-name>@values</aggr-name>
</aggr-list>
</vserver-info>
</query>
</vserver-modify-iter>

=item C<< 'max-volumes' => $integer >>

Maximum number of volumes.

Applicable for CMode CLI/ZAPI.

Maps to:

CMode ZAPI:

<vserver-modify-iter>
<query>
<vserver-info>
<max-volumes>$value</max-volumes>
</vserver-info>
</query>
</vserver-modify-iter>

=item C<< "cmd_execution_time"=> \$scalar >>

(Optional) Populate the param with execution time of command

=item C<< 'allowed-protocols' => [ $allowed_proto1, $allowed_proto2, ... ] >>

(Arrayref, even if only a single value is to be sent it should be specified as
an arrayref containing that single value)

Allowed Protocols. Valid values are: nfs | cifs | fcp | iscsi

Applicable for CMode CLI/ZAPI/GUI.

Maps to:

CMode ZAPI: allowed-protocols

<vserver-modify-iter>
<query>
<vserver-info>
<allowed-protocols>
<protocol>@values</protocol>
</allowed-protocols>
</vserver-info>
</query>
</vserver-modify-iter>

=item C<< 'disallowed-protocols' => [ $disallowed1, $disallowed2, ... ] >>

(Arrayref, even if only a single value is to be sent it should be specified as
an arrayref containing that single value)

Disallowed Protocols. Valid values are: nfs | cifs | fcp | iscsi

Applicable for CMode CLI/ZAPI/GUI.

Maps to:

CMode ZAPI: disallowed-protocols

<vserver-modify-iter>
<query>
<vserver-info>
<disallowed-protocols>
<protocol>@values</max-volumes>
</disallowed-protocols>
</vserver-info>
</query>
</vserver-modify-iter>

=item C<< 'protocol-services-use-data-lifs' => $string >>

Applicable for CMode CLI

=item C<< 'qos-policy-group' => $string >>

Applicable for CMode CLI

=item C<< 'external-cache' => $string >>

External Cache Policy

Applicable for CMode CLI

=item C<< "extended_query" => $string  >>

(Optional)

Applicable for CMode CLI/ZAPI

=item C<< 'snapshot-policy-uuid' => $string >>

Snapshot Policy UUID
Applicable for CMode CLI

=item C<< 'caching-policy' => $string >>

Caching Policy Name
Applicable for CMode CLI

=item C<< 'volume-delete-retention-hours' => $string >>

Volume Delete Retention Period
Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

<vserver-modify-iter>
<query>
<vserver-info>
<volume-delete-retention-hours>$value</volume-delete-retention-hours>
</vserver-info>
</query>
</vserver-modify-iter>

=item C<< 'qos-policy-group-uuid' => $string >>

QoS Policy Group UUID
Applicable for CMode CLI

=item C<< other_options >>

This method takes other common options described in L<Common method options|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#Other_options_valid_on_component_methods>

=item C<< 'language' => $string >>

Default Volume Language Code
Applicable for CMode CLI/ZAPI for builds that support the Vserver Multiple
Languages feature (i.e. from SN.0 onwards).

Maps to:

CMode ZAPI: language

=item C<< 'aggr-uuid-list' => [ $aggr-uuid-1, $aggr-uuid-2, ... ] >>

List of Aggregates Assigned
(Arrayref, even if only a single value is to be sent it should specified
as an arrayref containing that single value)
Applicable for CMode CLI

=item C<< 'next-rtpid' => $string >>

Next Available RTPID
Applicable for CMode CLI

=item C<< 'san-tgt-dev-id' => $string >>

SAN Target Device ID
Applicable for CMode CLI

=item C<< 'name' => $string >>

Vserver name for config replication
Applicable for CMode CLI

=item C<< 'protocol' => [ $protocol1, $protocol2, ... ] >>

Protocols possible value(s) are, nfs,cifs,fcp,iscsi,ndmp
(Arrayref, even if only a single value is to be sent it should specified
as an arrayref containing that single value)
Applicable for CMode CLI/GUI

=item C<< 'uuidforEngineID' => $string >>

UUID for generating EngineID
Applicable for CMode CLI

=item C<< 'newname' => $string >>

newname of vserver for renaming.
Applicable for CMode GUI.

=item C<< '*' >>

This method takes all other options supported by
"vserver modify" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=item C<NACL::C::Exceptions::Vserver::DoesNotExist>

This type of exception gets thrown if an attempt is made to modify a
non-existing vserver.

=back

=back

=cut


sub modify {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            "protocol"             => {type => ARRAYREF,          optional => 1},
            "aggr-uuid-list"       => {type => ARRAYREF,          optional => 1},
            'aggrlist'             => {type => ARRAYREF | SCALAR, optional => 1},
            'aggr-list'            => {type => ARRAYREF,          optional => 1},
            'nm-switch'            => {type => ARRAYREF | SCALAR, optional => 1},
            'ns-switch'            => {type => ARRAYREF | SCALAR, optional => 1},
            'allowed-protocols'    => {type => ARRAYREF,          optional => 1},
            'disallowed-protocols' => {type => ARRAYREF,  optional => 1},
            cmd_execution_time     => {type => SCALARREF, optional => 1},

            _optional_scalars(
                qw( snapshot-policy comment
                    antivirus-on-access-policy quota-policy maxsshsessions
                    maxvolumes maxStorage admin-state max-volumes
                    protocol-services-use-data-lifs
                    qos-policy-group
                    external-cache
                    vserver newname
                    snapshot-policy-uuid caching-policy
                    volume-delete-retention-hours qos-policy-group-uuid
                    language
                    next-rtpid san-tgt-dev-id name
                    uuidforEngineID
                    )
            ),
        },
        api => 'vserver_modify',
    );

    if ($pkg_or_obj->_has_nsswitch_ui(%opts)) {
        my $warn_message = "The fields nm-switch and ns-switch are deprecated. "
                . "Use 'vserver services name-service ns-switch modify' command "
                . "to modify name-service and name-mapping\n";
        NACL::C::Component->_print_deprecation_warning($warn_message);
    }

    try {
        $pkg_or_obj->call_on_apiset(
            %opts,
            choices => [
                {   method    => "_modify_cmode_cli",
                    interface => "CLI",
                    set       => "CMode",
                    zapi_type => 'none',
                },
                {   method    => "_modify_cmode_zapi",
                    interface => "ZAPI",
                    zapi_type => "iter",
                    set       => "CMode",
                },
                {   method    => "_modify_cmode_zapi_non_iter",
                    interface => "ZAPI",
                    zapi_type => "non-iter",
                    set       => "CMode",
                    check     => "_base_check_non_iter",
                },
                {   method    => "_modify_cmode_gui",
                    interface => "GUI",
                    set       => "CMode",
                    zapi_type => "none",
                },

            ],
            frontend => 'NACL::C::Vserver::modify'
        );
    }
    catch NACL::APISet::Exceptions::MultipleCommandFailedException with {
        my $exception = shift;
        my $sub       = sub {
            my ($error) = validate_pos(@_, {type => SCALAR});
            $error =~ /failed on vserver "(.*?)"/;
            my $pk_hash = {vserver => $1};
            $Log->exit() if $may_exit;
            return $pk_hash;
        };
        my $new_exception = $exception->convert_to_base_component_exception(
            extract_pk_sub => $sub);
        $Log->exit() if $may_exit;
        $new_exception->throw();
    };
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = $_[0];
        NACL::C::Exceptions::Vserver::DoesNotExist->detect_convert_and_throw(
            exception => $exception);
        $Log->exit() if $may_exit;
        $exception->throw();
    };

    $Log->exit() if $may_exit;
} ## end sub modify

sub _modify_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;

    my %opts = validate_with(
        params      => \@opts,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );

    $pkg_or_obj->_array_to_scalar_deprication_warning(
        params  => ['aggrlist', 'nm-switch', 'ns-switch'],
        options => \%opts
    );

    $pkg_or_obj->_base_backend_cmode_cli(%opts, api => 'vserver_modify',);
    $Log->exit() if $may_exit;
} ## end sub _modify_cmode_cli

sub _modify_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;

    my $common_copy = $pkg_or_obj->_common_zapi_copy();
    my $common_map  = $pkg_or_obj->_common_zapi_map();

    $pkg_or_obj->_iter_helper(
        @opts,
        api                  => 'vserver-modify-iter',
        copy                 => $common_copy,
        map                  => $common_map,
        does_not_exist_error => "entry doesn't exist",
        does_not_exist_exception =>
            'NACL::C::Exceptions::Vserver::DoesNotExist'
    );

    $Log->exit() if $may_exit;
} ## end sub _modify_cmode_zapi

sub _modify_cmode_zapi_non_iter {
    $Log->enter() if $may_enter;
    my ($pkg, %opts) = @_;

    my $zapi_map = {
        'vserver'     => 'vserver-name',
        'admin-state' => 'state',
        'nm-switch'   => 'name-mapping-switch',
        'ns-switch'   => 'name-server-switch',
    };

    my $zapi_copy = [
        qw(antivirus-on-access-policy caching-policy comment language
            max-volumes quota-policy qos-policy-group snapshot-policy
            volume-delete-retention-hours aggr-list disallowed-protocols
            allowed-protocols)
    ];

    $pkg->_base_backend_cmode_zapi(
        %opts,
        api  => 'vserver-modify',
        map  => $zapi_map,
        copy => $zapi_copy
    );

    $Log->exit() if $may_exit;
    return;
} ## end sub _modify_cmode_zapi

sub _modify_cmode_gui {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, %opts ) = @_;

    my $copy = [
        qw( vserver newname nm-switch ns-switch aggr-list
            protocol disallowed-protocols allowed-protocols )
    ];

    $pkg_or_obj->_base_backend_cmode_gui(
        %opts,
        api  => 'editVserver',
        copy => $copy,
    );

    if ( ( $opts{'newname'} ) && ( ref $pkg_or_obj ) ) {
        $pkg_or_obj->vserver( $opts{'newname'} );
    }

    $Log->exit() if $may_exit;
} ## end sub _modify_cmode_gui

=head2 rename

   NACL::C::Vserver->rename(command_interface => $ci,
                            vserver => $vserver,
                            ...);
                or
    $Vserver_object->rename(...);

(Class or instance method)
This method is used to rename a Vserver.

Uses a CMode CLI/ZAPI/GUI APISet.

Invokes "vserver-rename" API for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "vserver" => $vserver >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI/GUI.

Maps to:

CM ZAPI: vserver-name

=item C<< "newname" => $string >>

Applicable for CMode CLI/ZAPI/GUI.

Maps to:

CM ZAPI: new-name

=item C<< "cmd_execution_time"=> \$scalar >>

(Optional) Populate the param with execution time of command

item C<< "extended_query" => $string  >>

(Optional)

Applicable for CMode CLI

=item C<< other_options >>

This method takes other common options described in L<Common method options|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#Other_options_valid_on_component_methods>

=item C<< '*' >>

This method takes all other options supported by
"vserver rename" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub rename {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            cmd_execution_time => {type => SCALARREF, optional => 1},
            _optional_scalars(
                qw( foreground newname vserver )
            ),
        },
        cli_cmd => 'vserver rename',
    );
    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_rename_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_rename_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
            {   method    => '_rename_cmode_gui',
                interface => 'GUI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::Vserver::rename',
    );

    $pkg_or_obj->vserver($opts{'newname'}) if ref($pkg_or_obj);
    $Log->exit() if $may_exit;
    return $ret;
} ## end sub rename

sub _rename_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@opts, api => 'vserver_rename',);
    $Log->exit() if $may_exit;
} ## end sub _rename_cmode_cli

sub _rename_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg, %opts) = @_;
    my $response;
    my $execution_time = $opts{cmd_execution_time};
    my $foreground = $opts{foreground};
    my $api = 'vserver_rename';
    if(defined $opts{foreground}){
        if($foreground =~ /false/i){
            $api .= '_async';
        }
    }

    $pkg->_base_backend_cmode_zapi(%opts,
        apiset_response_obj => \$response,
        api                 => $api,
        map                 => {
            "vserver" => "vserver-name",
            "newname" => "new-name",
        },
    );

    if (defined $execution_time) {
        $$execution_time = $response->get_execution_time();
    }
    $Log->exit() if $may_exit;

} ## end sub _rename_cmode_zapi

sub _rename_cmode_gui {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, %opts ) = @_;
    $pkg_or_obj->_base_backend_cmode_gui( %opts,
        api  => 'editVserver',
        copy => [ 'vserver', 'newname' ]
    );
    $Log->exit() if $may_exit;
} ## end sub _rename_cmode_gui


=head2 start

   NACL::C::Vserver->start(command_interface => $ci,
                           vserver => $vserver,
                           ...);
                or
    $Vserver_object->start(...);

(Class or instance method)
This method is used to start a Vserver.

Uses a CMode CLI/ZAPI/GUI APISet.

Invokes "vserver-start" API for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "vserver" => $vserver >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI/GUI.

Maps to:

CM ZAPI: vserver-name

=item C<< "foreground" => $string >>

Applicable for CMode CLI

=item C<< "method-timeout"=> $scalar >>

(Optional) The new timeout value.

=item C<< "cmd_execution_time"=> \$scalar >>

(Optional) Populate the param with execution time of command

=item C<< "extended_query" => $string  >>

(Optional)

Applicable for CMode CLI

=item C<< other_options >>

This method takes other common options described in L<Common method options|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#Other_options_valid_on_component_methods>

=item C<< 'force' => $string >>

Force start
Applicable for CMode CLI

=item C<< '*' >>

This method takes all other options supported by
"vserver start" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<NACL::C::Exceptions::Vserver::AlreadyStarted>

This type of exception gets thrown if an attempt is made to start a
vserver that is already started.

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub start {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( foreground

                    vserver

                    force
                    )
            ),
        },
        cli_cmd => 'vserver start',
    );
    my $ret = $pkg_or_obj->call_on_apiset_with_exceptions(
        %opts,
        choices => [
            {   method    => '_start_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_start_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
                check     => '_foreground_zapi_check',
            },
            {   method    => '_start_cmode_gui',
                interface => 'GUI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::Vserver::start',
        nacl_exceptions => [qw(AlreadyStarted)],
    );
    $Log->exit() if $may_exit;
    return $ret;
} ## end sub start

sub _start_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    my $apiset = $opts{apiset};
    my $enabled_confirmations = 0;
    my $prompts_answers = [
        qr/.*Do you want to continue.*/i => 'y',
    ];

    START: {
        use warnings qw(exiting);
        try {
            $pkg_or_obj->_base_backend_cmode_cli(%opts, api => 'vserver_start',
                additional_args => {'connectrec-match_table' => $prompts_answers});
        }
        catch NACL::APISet::Exceptions::CommandFailedException with {
            my $e = shift;
            if ($e->text() =~
                /cannot\sstart\sthe\sVserver\swhen\scommand\sconfirmations\sare\sdisabled/i)
            {
                $apiset->set(confirmations => 'on');
                $enabled_confirmations = 1;
                no warnings qw(exiting);
                redo START;
            } else {
                $Log->exit() if $may_exit;
                $e->throw();
            }
        } finally {
            $apiset->set(confirmations => 'off') if ($enabled_confirmations);
        };
    }
    $Log->exit() if $may_exit;
} ## end sub _start_cmode_cli

sub _start_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    $pkg_or_obj->_base_backend_cmode_zapi(%opts,
        api  => 'vserver_start',
        map  => {"vserver" => "vserver-name",},
        inapplicable_for_help_xml => ['foreground'],
    );
    $Log->exit() if $may_exit;
} ## end sub _start_cmode_zapi

sub _start_cmode_gui {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, %opts ) = @_;
    $pkg_or_obj->_base_backend_cmode_gui(
        %opts,
        api => 'startVserver',
        copy => [ 'vserver' ],
    );
    $Log->exit() if $may_exit;
} ## end sub _start_cmode_gui

sub _foreground_zapi_check {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    if (defined $opts{foreground} && $opts{foreground} eq 'false') {
        my $msg = 'Could not use the ZAPI implementaion for '
            . "$opts{_method} because foreground => 'false' was provided "
            . 'in the call, but the ZAPI does not support asynchronous '
            . 'operations';
        $Log->comment($msg);
        $Log->exit() if $may_exit;
        NACL::Exceptions::InvalidChoice->throw($msg);
    }
    $Log->exit() if $may_exit;
}

=head2 stop

   NACL::C::Vserver->stop(command_interface => $ci,
                          vserver => $vserver,
                          ...);
                or
    $Vserver_object->stop(...);

(Class or instance method)
This method is used to stop a Vserver.

Uses a CMode CLI/ZAPI/GUI APISet.

Invokes "vserver-stop" API for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "vserver" => $vserver >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI/GUI.

Maps to:

CM ZAPI: vserver-name

=item C<< "foreground" => $string >>

Applicable for CMode CLI

=item C<< "method-timeout"=> $scalar >>

(Optional) The new timeout value.

=item C<< "cmd_execution_time"=> \$scalar >>

(Optional) Populate the param with execution time of command

=item C<< "extended_query" => $string  >>

(Optional)

Applicable for CMode CLI

=item C<< other_options >>

This method takes other common options described in L<Common method options|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#Other_options_valid_on_component_methods>

=item C<< '*' >>

This method takes all other options supported by
"vserver stop" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub stop {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( foreground

                    vserver
                    )
            ),
        },
        cli_cmd => 'vserver stop',
    );

    my $ret = $pkg_or_obj->call_on_apiset_with_exceptions(
        %opts,
        choices => [
            {   method    => '_stop_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_stop_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
            {   method    => '_stop_cmode_gui',
                interface => 'GUI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::Vserver::stop',
        nacl_exceptions => [qw(AlreadyStopped)],
    );
    $Log->exit() if $may_exit;
    return $ret;
} ## end sub stop

sub _stop_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@opts, api => 'vserver_stop',);
    $Log->exit() if $may_exit;
} ## end sub _stop_cmode_cli

sub _stop_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    $pkg_or_obj->_base_backend_cmode_zapi(%opts,
        api  => 'vserver_stop',
        map  => {"vserver" => "vserver-name"},
        inapplicable_for_help_xml => ['foreground'],
    );
    $Log->exit() if $may_exit;
} ## end sub _stop_cmode_zapi

sub _stop_cmode_gui {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, %opts ) = @_;
    $pkg_or_obj->_base_backend_cmode_gui(
        %opts,
        api => 'stopVserver',
        copy => [ 'vserver' ],
    );
    $Log->exit() if $may_exit;
} ## end sub _stop_cmode_gui

# All components should now implement this method.
# This method should return the validation spec for the primary keys
# Note that here, we do not include hostrec, command_interface in this
# validation spec

=head2 context

    NACL::C::Vserver->context(
                                command_interface => $ci,
                                'vserver' => $vserver,
                                ...);

                                 or

    $Vserver_obj->context(...);

(Class or instance method)
This method is used to set Vserver context.
To run the commands in vserver context, we recommend you to refer the user
guide section L<"Setting Vserver
context"|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#Vserver_and_Vfiler_context>
rather than setting the context by using this method.

The only usecase for using this method would be, if the command which you
would want to execute in vserver context, cannot be invoked from ngshell by
passing '-vserver' option and the interface of your vserver is not set up to
make a direct connection to it. This said, we recommend you to prefer this
method as the last choice, since this method changes the execution context
without the knowledge of command interface and this may result in undesirable
behavior, if the context is not reset back properly.

Uses a CMode CLI APISet.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'vserver' => $vserver >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI

=item C<< 'username' => $string >>

Applicable for CMode CLI

=item C<< "cmd_execution_time"=> \$scalar >>

(Optional) Populate the param with execution time of command

=item C<< 'method-timeout' => $time_in_seconds >>

(Optional)
The default NATE timeout is 60 seconds. If the command might take more
time to complete, this option can be used to specify a larger timeout value.
The value should be provided in seconds.

=item C<< "extended_query" => $string  >>

(Optional)

Applicable for CMode CLI

=item C<< other_options >>

This method takes other common options described in L<Common method options|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#Other_options_valid_on_component_methods>

=item C<< '*' >>

This method takes all other options supported by
"vserver context" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub context {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            cmd_execution_time => {type => SCALARREF, optional => 1},
            _optional_scalars(
                qw( username

                    vserver
                    )
            ),
        },
        api => 'vserver_context',
    );
    $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_context_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::Vserver::context',
    );

    $Log->exit() if $may_exit;
} ## end sub context

sub _context_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@opts, api => 'vserver_context',);
    $Log->exit() if $may_exit;
} ## end sub _context_cmode_cli

=head2 test_context

    NACL::C::Vserver->test_context(
                                command_interface => $ci,
                                ...);

(Class method)
This method is used to test VserverContext Class.

Uses a CMode CLI APISet.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'vserverid' => $string >>

Applicable for CMode CLI

=item C<< 'vserverip' => $string >>

Applicable for CMode CLI

=item C<< 'vserveruuid' => $string >>

Applicable for CMode CLI

=item C<< 'vservername' => $string >>

Applicable for CMode CLI

=item C<< "cmd_execution_time"=> \$scalar >>

(Optional) Populate the param with execution time of command

=item C<< 'method-timeout' => $time_in_seconds >>

(Optional)
The default NATE timeout is 60 seconds. If the command might take more
time to complete, this option can be used to specify a larger timeout value.
The value should be provided in seconds.

=item C<< "extended_query" => $string  >>

(Optional)

Applicable for CMode CLI

=item C<< other_options >>

This method takes other common options described in L<Common method options|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#Other_options_valid_on_component_methods>

=item C<< '*' >>

This method takes all other options supported by
"vserver test-context" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub test_context {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            cmd_execution_time => {type => SCALARREF, optional => 1},
            _optional_scalars(
                qw( vserverid vserverip vserveruuid vservername
                    )
            ),
        },
        ignore_primary_keys => 1,
        api                 => 'vserver_test_context',
    );
    $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_test_context_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::Vserver::test_context',
    );

    $Log->exit() if $may_exit;
} ## end sub test_context

sub _test_context_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, @opts) = @_;
    $pkg->_base_backend_cmode_cli(@opts, api => 'vserver_test_context',);
    $Log->exit() if $may_exit;
} ## end sub _test_context_cmode_cli

=head2 setup

    NACL::C::Vserver->setup(
                                command_interface => $ci,
                                'vserver' => $vserver,
                                ...);

                                 or

    $Vserver_obj->setup(...);

(Class or instance method)
This method is used to vserver setup wizard.

Uses a CMode CLI APISet.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'vserver' => $vserver >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI

=item C<< 'protocol-configure' => $string >>

Applicable for CMode CLI

=item C<< 'network' => $string >>

Applicable for CMode CLI

=item C<< 'lif_type' => $string >>

Applicable for CMode CLI

=item C<< 'yes_no' => $string >>

Applicable for CMode CLI

=item C<< 'text' => $string >>

Applicable for CMode CLI

=item C<< 'service-configure' => $string >>

Applicable for CMode CLI

=item C<< 'volSize' => $string >>

Applicable for CMode CLI

=item C<< 'storage' => $string >>

Applicable for CMode CLI

=item C<< 'services' => [ $service1, $service2, ... ] >>

(Arrayref, even if only a single value is to be sent it should specified
as an arrayref containing that single value)

Applicable for CMode CLI

=item C<< 'protocols' => [ $protocol1, $protocol2, ... ] >>

(Arrayref, even if only a single value is to be sent it should specified
as an arrayref containing that single value)

Applicable for CMode CLI

=item C<< 'method-timeout' => $time_in_seconds >>

(Optional)
The default NATE timeout is 60 seconds. If the command might take more
time to complete, this option can be used to specify a larger timeout value.
The value should be provided in seconds.

=item C<< 'ip-protocols' => $string >>

Applicable for CMode CLI

=item C<< "cmd_execution_time"=> \$scalar >>

(Optional) Populate the param with execution time of command

=item C<< "extended_query" => $string  >>

(Optional)

Applicable for CMode CLI

=item C<< other_options >>

This method takes other common options described in L<Common method options|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#Other_options_valid_on_component_methods>

=item C<< '*' >>

This method takes all other options supported by
"vserver setup" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub setup {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            cmd_execution_time => {type => SCALARREF, optional => 1},
            'services'         => {type => ARRAYREF,  optional => 1},
            'protocols'        => {type => ARRAYREF,  optional => 1},
            _optional_scalars(
                qw( protocol-configure network lif_type yes_no text service-configure volSize
                    storage ip-protocols

                    vserver
                    )
            ),
        },
        api => 'vserver_setup',
    );
    $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_setup_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::Vserver::setup',
    );

    $Log->exit() if $may_exit;
} ## end sub setup

sub _setup_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@opts, api => 'vserver_setup',);
    $Log->exit() if $may_exit;
} ## end sub _setup_cmode_cli

=head2 options

    # Set a new value, package call
    NACL::C::Vserver->options(
        command_interface => $ci,
        vserver           => $vserver,
        'option-name'     => $option,
        'option-value'    => $val,
    );

    # Set a new value, instance call
    $vs_obj->options(
        'option-name'     => $option,
        'option-value'    => $val,
    );

    # Get a value. Returns an object of type NACL::CS::VserverOptions
    my $vs_options = NACL::C::Vserver->options(
        command_interface => $ci,
        vserver           => $vserver,
    );

(Class or instance method)
This method is used to display/modify cluster wide options.

Uses a CMode CLI/ZAPI APISet.

Invokes 'options-modify-iter' API for CMode ZAPI if 'option-name' is specified.

Invokes 'options-get-iter' API for CMode ZAPI if 'option-name' is not specified.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'vserver' => $vserver >>

(Required for class method, Not Applicable for instance method)
Vserver possible value(s) are, vserver name
Applicable for CMode CLI

=item C<< 'option-value' => $string >>

Option Value possible value(s) are, text
Applicable for CMode CLI

=item C<< 'option-name' => $string >>

Option Name possible value(s) are, text
Applicable for CMode CLI/ZAPI

Maps to:

CMode ZAPI: name

=item C<< 'cmd_execution_time' => \$scalar >>

(Optional) Scalar reference , it will be updated with the command execution
time returned by the apiset response object

=item C<< 'method-timeout' => $time_in_seconds >>

(Optional)
The default NATE timeout is 60 seconds. If the command might take more
time to complete, this option can be used to specify a larger timeout value.
The value should be provided in seconds.

=item C<< 'vserver' => $string >>

Vserver possible value(s) are, vserver name

Applicable for CMode CLI/ZAPI.

Maps to:

CMode ZAPI : vserver

=item C<< other_options >>

This method takes other common options described in L<Common method options|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#Other_options_valid_on_component_methods>

=item C<< '*' >>

This method takes all other options supported by
"vserver options" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub options {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            'cmd_execution_time' => {type => SCALARREF, optional => 1},
            _optional_scalars(
                qw( option-value option-name

                    vserver
                    )
            ),
        },
        api => 'vserver_options',
    );

    my @ret;
    try {
        @ret = $pkg_or_obj->call_on_apiset(
            %opts,
            choices => [
                {   method    => '_options_cmode_cli',
                    interface => 'CLI',
                    set       => 'CMode',
                },
                {   method    => '_options_cmode_zapi',
                    interface => 'ZAPI',
                    set       => 'CMode',
                },
            ],
            frontend => 'NACL::C::Vserver::options',
        );
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = $_[0];
        NACL::C::Exceptions::Vserver::DoesNotExist->detect_convert_and_throw(
            exception => $exception);
        $Log->exit() if $may_exit;
        $exception->throw();
    };
    $Log->exit() if $may_exit;
    return wantarray ? @ret : $ret[0];
} ## end sub options

sub _options_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    if (defined ($opts{'option-value'})) {
        $pkg_or_obj->_base_backend_cmode_cli(%opts, api => 'vserver_options');
    } else {
        my @state_objs = $pkg_or_obj->_get_options_cs(%opts, interface => 'CLI');
        $Log->exit() if $may_exit;
        return @state_objs;
    }
} ## end sub _options_cmode_cli

sub _options_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    if (defined ($opts{'option-value'})) {
        $pkg_or_obj->_iter_helper(
            @args,
            api  => 'options-modify-iter',
            copy => [qw( vserver )],
            map  => {
                "option-name"  => "name",
                "option-value" => "value",
            },
            fix_api_args => sub {
                my %arg        = @_;
                my $query      = $arg{query};
                my $attributes = $arg{attributes};
                $query->{name} = delete $attributes->{name};
                return %arg;
            },
            does_not_exist_error => "entry doesn't exist",
            does_not_exist_exception =>
                'NACL::C::Exceptions::Vserver::DoesNotExist'
        );
    } else {
        my @state_objs = $pkg_or_obj->_get_options_cs(%opts, interface => 'ZAPI');
        $Log->exit() if $may_exit;
        return @state_objs;
    }
} ## end sub _options_cmode_zapi

sub _get_options_cs {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    my %filter;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%filter,
        copy   => [qw(vserver option-name)]
    );
    my @state_objs = NACL::CS::VserverOptions->fetch(
        command_interface => $opts{command_interface},
        filter            => \%filter,
        apiset_should     => {interface => $opts{interface}}
    );
    $Log->exit() if $may_exit;
    return @state_objs;
}

=head2 lock

    NACL::C::Vserver->lock(
                                command_interface => $ci,
                                'vserver' => $vserver,
                                ...);

                                 or

    $Vserver_obj->lock(...);

(Class or instance method)
This method is used to lock Vserver configuration.
Uses CMode CLI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'vserver' => $vserver >>

(Required for class method, Not Applicable for instance method)
Vserver
Applicable for CMode CLI

Maps to:

CM ZAPI: vserver-name

=item C<< '*' >>

This method takes all other options supported by "vserver lock" command in ONTAP

=item C<< other_options >>

This method takes other common options described in L<Common method options|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#Other_options_valid_on_component_methods>

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub lock {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params => \@args,
        api    => 'vserver_lock',
    );
    $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_lock_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::Vserver::lock',
    );

}

sub _lock_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@opts, api => 'vserver_lock',);
    $Log->exit() if $may_exit;
}

=head2 restrict

    NACL::C::Vserver->restrict(
                                command_interface => $ci,
                                'vserver' => $vserver,
                                ...);

                                 or

    $Vserver_obj->restrict(...);

(Class or instance method)
This method is used to restrict a Vserver.
Uses CMode CLI .

Note that this command was removed in FS.0, so this method should not be used
for builds running FS.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'vserver' => $vserver >>

(Required for class method, Not Applicable for instance method)
Vserver
Applicable for CMode CLI

=item C<< '*' >>

This method takes all other options supported by "vserver restrict" command in ONTAP

=item C<< other_options >>

This method takes other common options described in L<Common method options|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#Other_options_valid_on_component_methods>

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

# Command was removed in FS.0. When we get to LB, this should be removed.
sub restrict {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params => \@args,
        api    => 'vserver_restrict',
    );
    $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_restrict_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::Vserver::restrict',
    );

}

sub _restrict_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@opts, api => 'vserver_restrict',);
    $Log->exit() if $may_exit;
}

=head2 unlock

    NACL::C::Vserver->unlock(
                                command_interface => $ci,
                                'vserver' => $vserver,
                                ...);

                                 or

    $Vserver_obj->unlock(...);

(Class or instance method)
This method is used to unlock Vserver configuration.
Uses CMode CLI .

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'vserver' => $vserver >>

(Required for class method, Not Applicable for instance method)
Vserver
Applicable for CMode CLI/ZAPI

=item C<< 'force' => $string >>

Force Unlock possible value(s) are, true,false
Applicable for CMode CLI

=item C<< '*' >>

This method takes all other options supported by "vserver unlock" command in ONTAP

=item C<< other_options >>

This method takes other common options described in L<Common method options|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#Other_options_valid_on_component_methods>

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub unlock {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( force
                    )
            ),
        },
        api => 'vserver_unlock',
    );
    $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_unlock_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_unlock_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::Vserver::unlock',
    );

}

sub _unlock_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@opts, api => 'vserver_unlock',);
    $Log->exit() if $may_exit;
}

sub _unlock_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    try {
        $pkg_or_obj->_base_backend_cmode_zapi(
            %opts,
            api  => 'vserver-unlock',
            map  => {vserver => 'vserver-name'}
        );
    } catch NACL::APISet::Exceptions::MissingRequiredParamException with {
        # User is running an older build, so has the old mapping.
        # Once the fix for burt 799418 is available in DOT:main, we can
        # remove these try-catch blocks and leave only the new mapping.
        $pkg_or_obj->_base_backend_cmode_zapi(
            %opts,
            api  => 'vserver-unlock',
            copy => ['vserver'],
        );
    };
    $Log->exit() if $may_exit;
} ## end sub _unlock_cmode_zapi

=head2 prepare_for_revert

    NACL::C::Vserver->prepare_for_revert(
                                command_interface => $ci,
                                ...);

(Class method)
This method is used to prepares Vservers to be reverted.
Uses CMode CLI .

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< '*' >>

This method takes all other options supported by "vserver prepare for revert" command in ONTAP

=item C<< other_options >>

This method takes other common options described in L<Common method options|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#Other_options_valid_on_component_methods>

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub prepare_for_revert {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params              => \@args,
        ignore_primary_keys => 1,
        api                 => 'vserver_prepare_for_revert',
    );
    $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_prepare_for_revert_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::Vserver::prepare_for_revert',
    );

    $Log->exit() if $may_exit;
}

sub _prepare_for_revert_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, @opts) = @_;
    $pkg->_base_backend_cmode_cli(@opts, api => 'vserver_prepare_for_revert',
    );
    $Log->exit() if $may_exit;
}

=head2 add_aggregates

    NACL::C::Vserver->add_aggregates(
                                command_interface => $ci,
                                'vserver' => $vserver,
                                ...);

                                 or

    $Vserver_obj->add_aggregates(...);

(Class or instance method)
This method is used to add aggregates to the Vserver.
Uses CMode CLI/ZAPI/GUI .
Invokes  vserver-add-aggregates for CMode Zapi


=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'vserver' => $vserver >>

(Required for class method, Not Applicable for instance method)
Vserver

Applicable for CMode CLI/ZAPI/GUI.


Maps to:

CMode ZAPI : vserver-name

=item C<< 'aggregates' => [ $aggregate1, $aggregate2, ... ] >>

(Arrayref, even if only a single value is to be sent it should specified
as an arrayref containing that single value)


Applicable for CMode CLI/ZAPI/GUI.

Maps to:

CMode GUI : aggr-list


=item C<< '*' >>

This method takes all other options supported by "vserver add aggregates" command in ONTAP

=item C<< other_options >>

This method takes other common options described in L<Common method options|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#Other_options_valid_on_component_methods>

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub add_aggregates {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params => \@args,
        additional_spec =>
            {'aggregates' => {type => ARRAYREF, optional => 1},},
        api => 'vserver_add_aggregates',
    );
    $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_add_aggregates_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_add_aggregates_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
            {   method    => '_add_aggregates_cmode_gui',
                interface => 'GUI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::Vserver::add_aggregates',
    );

    $Log->exit() if $may_exit;
}

sub _add_aggregates_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@opts,
        api => 'vserver_add_aggregates',);
    $Log->exit() if $may_exit;
}

sub _add_aggregates_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    my $map = {
        'vserver'  => 'vserver-name',
        aggregates => 'aggr-list',
    };
    $pkg_or_obj->_base_backend_cmode_zapi(
        @opts,
        api  => 'vserver-add-aggregates',
        map  => $map,
    );
    $Log->exit() if $may_exit;
}

sub _add_aggregates_cmode_gui {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    $pkg_or_obj->_base_backend_cmode_gui(
        @opts,
        api  => 'editVserver',
        copy => [ 'vserver' ],
        map  => {aggregates => 'aggr-list'}
    );
    $Log->exit() if $may_exit;
}

=head2 add_protocols

    NACL::C::Vserver->add_protocols(
                                command_interface => $ci,
                                'vserver' => $vserver,
                                ...);

                                 or

    $Vserver_obj->add_protocols(...);

(Class or instance method)
This method is used to add protocols to the Vserver.
Uses CMode CLI/ZAPI/GUI .
Invokes  vserver-add-protocols for CMode Zapi


=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'vserver' => $vserver >>

(Required for class method, Not Applicable for instance method)
Vserver

Applicable for CMode CLI/ZAPI/GUI


Maps to:

CMode ZAPI : vserver-name

=item C<< 'protocols' => [ $protocol1, $protocol2, ... ] >>

(Arrayref, even if only a single value is to be sent it should specified
as an arrayref containing that single value)


Applicable for CMode CLI/ZAPI/GUI

Maps to:

CMode GUI : allowed-protocols

=item C<< '*' >>

This method takes all other options supported by "vserver add protocols" command in ONTAP

=item C<< other_options >>

This method takes other common options described in L<Common method options|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#Other_options_valid_on_component_methods>

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub add_protocols {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params => \@args,
        additional_spec =>
            {'protocols' => {type => ARRAYREF, optional => 1},},
        api => 'vserver_add_protocols',
    );
    $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_add_protocols_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_add_protocols_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
            {   method    => '_add_protocols_cmode_gui',
                interface => 'GUI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::Vserver::add_protocols',
    );

    $Log->exit() if $may_exit;
}

sub _add_protocols_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@opts,
        api => 'vserver_add_protocols',);
    $Log->exit() if $may_exit;
}

sub _add_protocols_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    my $copy = [qw(protocols)];
    my $map = {'vserver' => 'vserver-name',};
    $pkg_or_obj->_base_backend_cmode_zapi(
        @opts,
        api  => 'vserver-add-protocols',
        copy => $copy,
        map  => $map,
    );
    $Log->exit() if $may_exit;
}

sub _add_protocols_cmode_gui {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    $pkg_or_obj->_base_backend_cmode_gui(
        @opts,
        api  => 'editVserver',
        copy => [ 'vserver' ],
        map  => { 'protocols' => 'allowed-protocols' }
    );
    $Log->exit() if $may_exit;
}

=head2 remove_aggregates

    NACL::C::Vserver->remove_aggregates(
                                command_interface => $ci,
                                'vserver' => $vserver,
                                ...);

                                 or

    $Vserver_obj->remove_aggregates(...);

(Class or instance method)
This method is used to remove aggregates from the Vserver.
Uses CMode CLI/ZAPI .
Invokes  vserver-remove-aggregates for CMode Zapi


=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'vserver' => $vserver >>

(Required for class method, Not Applicable for instance method)
Vserver

Applicable for CMode CLI/ZAPI


Maps to:

CMode ZAPI : vserver-name

=item C<< 'aggregates' => [ $aggregate1, $aggregate2, ... ] >>

(Arrayref, even if only a single value is to be sent it should specified
as an arrayref containing that single value)


Applicable for CMode CLI/ZAPI

=item C<< '*' >>

This method takes all other options supported by "vserver remove aggregates" command in ONTAP

=item C<< other_options >>

This method takes other common options described in L<Common method options|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#Other_options_valid_on_component_methods>

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub remove_aggregates {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params => \@args,
        additional_spec =>
            {'aggregates' => {type => ARRAYREF, optional => 1},},
        api => 'vserver_remove_aggregates',
    );
    $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_remove_aggregates_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_remove_aggregates_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::Vserver::remove_aggregates',
    );

    $Log->exit() if $may_exit;
}

sub _remove_aggregates_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@opts,
        api => 'vserver_remove_aggregates',);
    $Log->exit() if $may_exit;
}

sub _remove_aggregates_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    my $map = {
        'vserver'  => 'vserver-name',
        aggregates => 'aggr-list'
    };
    $pkg_or_obj->_base_backend_cmode_zapi(
        @opts,
        api  => 'vserver-remove-aggregates',
        map  => $map,
    );
    $Log->exit() if $may_exit;
}

=head2 remove_protocols

    NACL::C::Vserver->remove_protocols(
                                command_interface => $ci,
                                'vserver' => $vserver,
                                ...);

                                 or

    $Vserver_obj->remove_protocols(...);

(Class or instance method)
This method is used to remove protocols from the Vserver.
Uses CMode CLI/ZAPI/GUI .
Invokes  vserver-remove-protocols for CMode Zapi


=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'vserver' => $vserver >>

(Required for class method, Not Applicable for instance method)
Vserver

Applicable for CMode CLI/ZAPI/GUI


Maps to:

CMode ZAPI : vserver-name

=item C<< 'protocols' => [ $protocol1, $protocol2, ... ] >>

(Arrayref, even if only a single value is to be sent it should specified
as an arrayref containing that single value)


Applicable for CMode CLI/ZAPI/GUI

Maps to:

CMode GUI : disallowed-protocols

=item C<< '*' >>

This method takes all other options supported by "vserver remove protocols" command in ONTAP

=item C<< other_options >>

This method takes other common options described in L<Common method options|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#Other_options_valid_on_component_methods>

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub remove_protocols {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params => \@args,
        additional_spec =>
            {'protocols' => {type => ARRAYREF, optional => 1},},
        api => 'vserver_remove_protocols',
    );
    $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_remove_protocols_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_remove_protocols_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
            {   method    => '_remove_protocols_cmode_gui',
                interface => 'GUI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::Vserver::remove_protocols',
    );

    $Log->exit() if $may_exit;
}

sub _remove_protocols_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@opts,
        api => 'vserver_remove_protocols',);
    $Log->exit() if $may_exit;
}

sub _remove_protocols_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    my $copy = [qw(protocols)];
    my $map = {'vserver' => 'vserver-name',};
    $pkg_or_obj->_base_backend_cmode_zapi(
        @opts,
        api  => 'vserver-remove-protocols',
        copy => $copy,
        map  => $map,
    );
    $Log->exit() if $may_exit;
}

sub _remove_protocols_cmode_gui {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;

    $pkg_or_obj->_base_backend_cmode_gui(
        @opts,
        api  => 'editVserver',
        copy => [ 'vserver' ],
        map  => {'protocols' => 'disallowed-protocols' },
    );
    $Log->exit() if $may_exit;
}

=head2 show_protocols

    my $show_protocols_obj = NACL::C::Vserver->show_protocols(command_interface => $ci,
                          [ filter => { %filter } ,]
              [ requested_fields => [ @req_fields ]);

                    or

    my @show_protocols_objs = NACL::C::Vserver->show_protocols(command_interface => $ci,
                          [ filter => { %filter } ,]
              [ requested_fields => [ @req_fields ]);

                    or

   my $show_protocols_obj = $Vserver_obj->show_protocols();

(Class or Instance method)

This method is a CS-like fetch method which accepts all the parameters of
L<fetch|lib-NACL-CS-ComponentState-pm/fetch> (requested_fields, filter, allow_empty)
and has all of the other properties of fetch methods (callable in scalar or list
context, throws a NoElementsFound exception if no matching elements are found).
The requested fields and filter takes L<ATTRIBUTES|lib-NACL-CS-VserverShowProtocols-pm/ATTRIBUTES>
of NACL::CS::VserverShowProtocols as input.
This method returns objects of the type L<NACL::CS::VserverShowProtocols|lib-NACL-CS-VserverShowProtocols-pm>.

=cut

sub show_protocols {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts = $pkg_or_obj->_handle_obj_cs_like_calls(@args);
    require NACL::CS::VserverShowProtocols;
    my @state_objs = NACL::CS::VserverShowProtocols->fetch(%opts);
    $Log->exit() if $may_exit;
    return wantarray ? @state_objs : $state_objs[0];
}


__PACKAGE__->_load_event_class();

sub _primary_keys_validate_spec {
    return (vserver => {type => SCALAR});
}

# This returns the appropriate method in ResourceRecords to be called
# to return the records for that command interface
sub resource_record_method_name {
    return "vservers";
}

=head1 METHODS (for component developers)

=head2 apiset

See L<NACL::C::CommandInterface::apiset|lib-NACL-C-CommandInterface-pm/apiset>
for a description of what the generic apiset method does.

We need to override apiset here to provide suitable values for shell_type
and context_definition, before calling the generic apiset defined in
NACL::C::CommandInterface.

=cut

sub apiset {
    my ($self, %opts) = @_;

    # Note that the specification in the 'choices' hash of call_on_apiset
    # would specify set to be "CMode". But if we are using Vserver as a
    # command interface, we are in "vserver context". Meaning, the APISet
    # we'd need to instantiate it to be of set 'Vserver'
    if ((defined $opts{set}) && ($opts{set} !~ /(Vserver|CMode)/)) {
        $Log->exit() if $may_exit;
        NATE::Exceptions::Argument->throw("Invalid set value '$opts{set}' "
                . "specified. "
                . "With a command_interface of type NACL::C::Vserver, you can "
                . "only create Vserver APISet object");
    } ## end if ( ( defined $opts{set...}))

    # APISet expects the 'set' to be 'Vserver' if the shell_type is Vserver.
    # The component frontend methods don't list 'Vserver' in the choices hash
    # Hence call_on_apiset cannot pick 'Vserver' as the set value and so the
    # value selected by it would be 'CMode'. Hence we are
    # working around here by reassigning the set value to be 'Vserver'.
    $opts{set} = "Vserver";

    # We need to set $self->hostrec() back to being the hostrec of the
    # vserver. Here's a use case for why we need to explicitly set
    # it back: we have a component $foo whose cmd_int is Vserver.
    # $foo->method(apiset_must => {shell_type => "CMode"});
    # $foo->method2(apiset_must => {shell_type => "Vserver"});
    # After the call to 'method', $self->hostrec() is of the parent filer,
    # but for call to 'method2' $self->hostrec() needs to be be of the
    # Vserver.

    # If shell_type is not specified, then we will set the default
    # shell_type to be "Vserver". Thus the default is to make a direct
    # connection to Vserver to execute the commands

    $opts{shell_type} = $opts{shell_type} || "Vserver";

    # Have a copy before modifying it in the foll 'if' statement.
    my $hostrec = $self->hostrec();

    if ($opts{shell_type} ne "Vserver") {

        # Note that we don't explicitly bother checking whether shell_type is
        # CMode. If shell_type=CMode is passed, the APISet code would
        # take care of throwing an exception, saving us from needing to
        # make that check here.
        my $vserver_host = $self->hostrec();
        my $pfiler       = $vserver_host->{PFILER};
        unless (defined $pfiler) {
            $Log->exit() if $may_exit;
            NATE::Exceptions::Argument->throw("You have specified the "
                    . "shell_type to be '$opts{shell_type}' for which a connection "
                    . "needs to be established to node or cluster and the commands"
                    . " should be executed in the context of vserver. But "
                    . "information about cluster or node is missing in the PFILER "
                    . "attribute of the vserver hostrec object");
        } ## end unless ( defined $pfiler )
        my $pfiler_host = Tharn::host($pfiler);
        $self->hostrec($pfiler_host);
        my %context_definition;
        $context_definition{schema_hostobj} = $vserver_host;
        $opts{context_definition}           = \%context_definition;
    } ## end if ( $opts{shell_type}...)

    # Call CommandInterface's apiset()
    my $apiset;
    try {
        $apiset = $self->SUPER::apiset(%opts);
    }
    finally {

        # Regardless of whether the SUPER call passed or failed set the
        # hostrec attribute back to what it originally was
        $self->hostrec($hostrec);
    };

    return $apiset;
} ## end sub apiset

=head2 find_vservers_with_online_root_volume

    # Scalar context
    my $first_vs = NACL::C::Vserver->find_vservers_with_online_root_volume(
        command_interface => $ci,
        ...
    );

    # List context
    my @all_vs = NACL::C::Vserver->find_vservers_with_online_root_volume(
        command_interface => $ci,
        ...
    );

(Class method only)
This method returns Vserver Component objects for those vservers whose
root-volume is online. (note that the normal find() will return all vservers
present, regardless of whether their root-volume is online or not).

When invoked in scalar context, it returns the first such vserver, in list
context it returns an array of all such vservers.

This method throws a C<NACL::Exceptions::NoElementsFound> exception if
there are no vservers with online root-volume.

This method is applicable for CMode-CLI and CMode-ZAPI.

=over

=item Options

=over

=item C<< command_interface => $ci >>

(Required)
The command_interface on which to search for online vservers.

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< randomize_order => 0|1 >>

(Optional) Specifies whether the order in which the vservers are returned
should be randomized.

=back

=back

=cut

sub find_vservers_with_online_root_volume {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = $pkg->_common_validate_with(
        params              => \@_,
        additional_spec     => {_optional_scalars(qw(randomize_order))},
        ignore_primary_keys => 1
    );

    my @vs_cs_list = NACL::CS::Vserver->fetch(
        %opts,
        canned_filters   => ['data_vservers'],
        requested_fields => ['rootvolume']
    );

    my @vols;
    map { push @vols, $_->rootvolume() } @vs_cs_list;
    my $vol_list = join '|', @vols;

    my @vol_cs_list = NACL::CS::Volume->fetch(
        %opts,
        filter => {
            volume => $vol_list,
            state  => 'online'
        },
        requested_fields => ['state'],
    );

    # This hash is keyed by a string of this form:
    # <volume>X;X<vserver>
    my %vol_cs_hash;
    foreach my $vol_cs (@vol_cs_list) {
        my $key = $vol_cs->volume() . 'X;X' . $vol_cs->vserver();
        $vol_cs_hash{$key} = 1;
    }

    my $get_instance;
    if ($pkg->isa('NACL::STask::STask')) {
        $get_instance = 'get_task_instance';
    } else {
        $get_instance = 'get_component_instance';
    }
    my @online_list;
    foreach my $vs_cs (@vs_cs_list) {

        # Form the key and check if its available in the vol list hash
        # If so, then this vserver's root-volume is online
        my $key = $vs_cs->rootvolume() . 'X;X' . $vs_cs->vserver();
        if (exists $vol_cs_hash{$key}) {
            push @online_list, $vs_cs->$get_instance();
        }
    } ## end foreach my $vs_cs (@vs_cs_list)

    unless (@online_list) {
        $Log->exit() if $may_exit;
        NACL::Exceptions::NoElementsFound->throw(
            'No vservers found with online root-volume');
    }

    $Log->exit() if $may_exit;
    return wantarray ? @online_list : $online_list[0];
} ## end sub find_vservers_with_online_root_volume

=head2 find_vservers_in_peer_relationship

    # Callable in list context
    my @vservers_in_peer = NACL::C::Vserver->find_vservers_in_peer_relationship(
        command_interface => $ci);

    # Callable in scalar context
    my $vserver_in_peer  = NACL::C::Vserver->find_vservers_in_peer_relationship(
        command_interface => $ci);

    # Accepts all of the other options accepted by find() methods
    my @vservers_in_peer = NACL::C::Vserver->find_vservers_in_peer_relationship(
        command_interface => $ci,
        filter            => { type => 'data' },
        allow_empty       => 1
    );

Finds and returns objects for vservers which are in a vserver peer relationship.
This provides the same facilities (callable in scalar/list context, throws
NoElementsFound exception) and accepts the same options (such as filter,
allow_empty etc) as all other find methods.

=cut

sub find_vservers_in_peer_relationship {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    my @vservers = $pkg->find(@args, is_in_peer => 1);
    $Log->exit() if $may_exit;
    return wantarray ? @vservers : $vservers[0];
}

=head2 find_vservers_not_in_peer_relationship

    # Callable in list context
    my @vservers_not_in_peer = NACL::C::Vserver->find_vservers_not_in_peer_relationship(
        command_interface => $ci);

    # Callable in scalar context
    my $vserver_not_in_peer  = NACL::C::Vserver->find_vservers_not_in_peer_relationship(
        command_interface => $ci);

    # Accepts all of the other options accepted by find() methods
    my @vservers_not_in_peer = NACL::C::Vserver->find_vservers_not_in_peer_relationship(
        command_interface => $ci,
        filter            => { type => 'data' },
        allow_empty       => 1
    );

Finds and returns objects for vservers which are not in a vserver peer relationship.
This provides the same facilities (callable in scalar/list context, throws
NoElementsFound exception) and accepts the same options (such as filter,
allow_empty etc) as all other find methods.

=cut

sub find_vservers_not_in_peer_relationship {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    my @vservers = $pkg->find(@args, is_in_peer => 0);
    $Log->exit() if $may_exit;
    return wantarray ? @vservers : $vservers[0];
}

=head2 is_in_peer_relationship

    my $is_in_peer = $vs_obj->is_in_peer_relationship();

    # More advanced form: check if it matches certain filter criteria
    my $is_in_peer = $vs_obj->is_in_peer_relationship(
        filter => { 'peer-vserver' => $other_vs });

(Instance method only)

Checks if the given Vserver is in a vserver peer relationship and returns 1
if so (returns 0 otherwise).

It is also possible to provide filter criteria. Example 2 checks whether
$vs_obj is in a vserver peer relationship with $other_vs. The options
accepted in the filter argument are the attributes listed in
L<NACL::CS::VserverPeer|lib-NACL-CS-VserverPeer-pm>.

=cut

sub is_in_peer_relationship {
    $Log->enter() if $may_enter;
    my ($self, @args) = @_;
    my %opts = validate_with(
        params      => \@args,
        spec        => {filter => {type => HASHREF, default => {}}},
        allow_extra => 1
    );

    my %filter = %{$opts{filter}};
    $filter{vserver} = $self->vserver();
    $opts{filter}    = \%filter;

    my $is_in_peer = 0;
    my $peer       = NACL::C::VserverPeer->find(
        command_interface => $self->command_interface(),
        allow_empty       => 1,
        %opts
    );
    $is_in_peer = 1 if (defined $peer);
    $Log->exit() if $may_exit;
    return $is_in_peer;
}

=head2 name_attribute

Returns 'vserver'.

=cut

sub name_attribute {
    return 'vserver';
}

=head2 determine_is_repository

    my $bool = NACL::C::Vserver->determine_is_repository(
        command_interface => $ci,
        vserver           => $vserver
    );

    $vserver_obj->determine_is_repository();

(Class or Instance method)
This method returns boolean value 1 is the specified vserver is a repository
vserver otherwise it will return 0.

=over

=item Options

=over

=item C<< command_interface => $ci >>

(Required for Class Methods, Not Applicable for Instance Methods)

See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< "vserver" => $vserver >>

(Required for Class Methods, Not Applicable for Instance Methods)

The vserver name.

=item Other options

The common component options such as C<method-timeout>, C<apiset_must>,
C<apiset_should> are also accepted.

=back

=back

=cut

sub determine_is_repository {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts = $pkg_or_obj->_common_validate_with(
        params => \@args,
        additional_spec =>
            {'is-repository' => {type => SCALAR | UNDEF, optional => 1}}
    );
    my $is_repository = delete $opts{'is-repository'};

    if (defined $is_repository) {
        $Log->exit() if $may_exit;
        return $is_repository;
    } elsif (ref($pkg_or_obj) && defined $pkg_or_obj->{_is_repository}) {

        # Do NOT use isset above since this would lead to an infinite
        # recursion when this method is invoked by the -default_ctor.
        # (isset() would return true, which would then invoke is_repository
        # which leads to an infinite recursion)
        $Log->exit() if $may_exit;
        return $pkg_or_obj->_is_repository();
    } else {
        my $vserver = delete $opts{vserver};

        my $cs = NACL::CS::Vserver->fetch(
            filter           => {vserver => $vserver},
            requested_fields => [qw(is-repository)],
            %opts,
        );

        if (   $cs->is_repository_isset()
            && $cs->is_repository() eq 'true')
        {
            $is_repository = 1;
        } else {
            $is_repository = 0;
        }
        $pkg_or_obj->_is_repository($is_repository) if (ref $pkg_or_obj);
        $Log->exit() if $may_exit;
        return $is_repository;
    }
} ## end sub determine_is_repository

=head2 get_vsid

    my $vsid = NACL::C::Vserver->get_vsid(
        command_interface => $ci,
        vserver           => $vserver
    );

    $vserver_obj->get_vsid();

(Class or Instance method)
This method returns the vserver id.

=over

=item Options

=over

=item C<< command_interface => $ci >>

(Required for Class Methods, Not Applicable for Instance Methods)

See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< "vserver" => $vserver >>

(Required for Class Methods, Not Applicable for Instance Methods)

The vserver name.

=item Other options

The common component options such as C<method-timeout>, C<apiset_must>,
C<apiset_should> are also accepted.

=back

=back

=cut

sub get_vsid {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts = $pkg_or_obj->_common_validate_with(params => \@args,);

    my $apiset = $opts{command_interface}->apiset(
        interface => 'CLI',
        set       => 'CMode'
    );

    my $response = $apiset->debug_smdb_table_show(
        table => 'vserver_id_to_name',
        name  => $opts{vserver},
    );
    my $vsid = $response->get_parsed_output()->[0]{id};

    $Log->exit() if $may_exit;
    return $vsid;
}

=head2 _create_version_manager

    scalar => [
        {   '-default_ctor' => sub {
                my $self = $_[0];
                if ( $self->can('_create_version_manager') ) {
                   ...
                   return $self->_create_version_manager();
                } else {
                    ...
                }
             }
        },
        'version_manager',
    ]

(Instance method)
This private method returns the version manager of type
NACL::APISet::Version::Node::CMode or NACL::APISet::Version::Node::Vserver
depending on whether the vserver component instance is playing the role of
pure command_interface or it is treated as a component (in which case, it
is a command_interface as well). In simple words, if vserver instance is
used as a component, then the command_interface associated with it cannot be of
type NACL::C::Vserver, basically the command_interface should be of type
'NACL::C::Node' or 'NACL::C::Cluster' so that it can manipulate a vserver.

This implementation is specific to vservers and we would like to return
a version manager of type NACL::APISet::Version::Node::CMode whenever possible
so that a user gets accessibility to discover various version related
attributes which could be discovered by connecting to the node (which
has cluster context)than limiting himself to the attributes which could be
directly discovered from vserver shell.

Returned version manager would be of type NACL::APISet::Version::Node::Vserver
if the vserver instance is a pure command_interface, else it
would be a version manager object of type NACL::APISet::Version::Node::CMode.

=cut

sub _create_version_manager {
    $Log->enter() if $may_enter;
    my $self = shift;

    require NACL::APISet::Version;
    my $vm;
    if (!$self->command_interface()->isa('NACL::C::Vserver')) {
        $Log->debug("vserver instance is not associated with vserver "
                . "command_interface");
        $vm = $self->command_interface()->get_version_manager();
    } else {
        $Log->debug("vserver instance is associated with vserver"
                . " command_interface");
        $vm = NACL::APISet::Version->new(
            category   => "Node",
            hostobj    => $self->hostrec(),
            shell_type => "Vserver"
        );
    } ## end else [ if ( !$self->command_interface...)]
    $Log->exit() if $may_exit;
    return $vm;
} ## end sub _create_version_manager

# For this command interface, what is the common prefix for the test parameters?
sub _ci_to_test_param_mapping {
    $Log->enter() if $may_enter;
    $Log->exit() if $may_exit;
    return 'VSERVER';
}

# When we dynamically create hostrecs, this is the hosttype which we should
# assign
sub _hosttype {
    $Log->enter() if $may_enter;
    $Log->exit() if $may_exit;
    return 'filer-ng-vserver';
}

sub _check_if_valid_hostrec {
    $Log->enter() if $may_enter;
    my ($self, @opts) = @_;
    $self->SUPER::_check_if_valid_hostrec(@opts);
    $self->_check_if_hostrec_cmode(@opts);
    my ($hostrec) = @opts;
    if ($hostrec->hosttype() !~ /filer\-ng\-vserver/) {

        # Don't invoke ->vserver() here. Since we could have reached here
        # from the -default_ctor of hostrec, and the -default_ctor of
        # vserver invokes ->hostrec(), we could end up with infinite
        # recursion
        my $vserver = $self->{vserver} || $hostrec->id();
        $Log->warn("The hosttype for Vserver '$vserver' " . "is '"
                . $hostrec->hosttype()
                . "' but for vservers it should ideally be 'filer-ng-vserver' "
                . '(this is because NATE makes vserver-specific settings '
                . 'with this hosttype)');
    } ## end if ( $hostrec->hosttype...)
    $Log->exit() if $may_exit;
} ## end sub _check_if_valid_hostrec

sub _find_validate_spec {
    $Log->enter() if $may_enter;
    my $pkg  = $_[0];
    my $spec = $pkg->SUPER::_find_validate_spec();
    $spec->{is_in_peer} = {type => SCALAR, optional => 1};
    $Log->exit() if $may_exit;
    return $spec;
}

sub _has_nsswitch_ui {
    my ($pkg, %opts) = @_;

    if (defined $opts{'ns-switch'} || defined $opts{'nm-switch'}) {
        return 1;
    }
    return 0;
}

# Values accepted by the CLI for "subtype": default; dp-destination;
# sync-source; sync-destination; cache
# Values accepted by the ZAPI: default; dp_destination; sync_source
# sync_destination; cache
# So for ZAPI, we translate all hyphens to underscores for the value of subtype
sub _handle_subtype_zapi_val {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts = validate_with(
        params => \@args,
        spec   => {options => {type => HASHREF}},
    );
    my $options = $opts{options};

    if (defined $options->{subtype}) {
        $options->{subtype} =~ s/-/_/g;
    }

    $Log->exit() if $may_exit;
}

1;
