#
# Copyright (c) 2001-2012 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary VserverCifsSecurity Component Module
## @author dl-nacl-dev@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::C::VserverCifsSecurity

=head1 DESCRIPTION

C<NACL::C::VserverCifsSecurity> is a derived class of
L<NACL::C::Component::ONTAP|lib-NACL-C-Component-ONTAP-pm>.  It represents access
to an ONTAP Vserver CIFS Security.  A related class is
L<NACL::CS::VserverCifsSecurity|lib-NACL-CS-VserverCifsSecurity-pm>, which
represents the past state of an ONTAP Vserver CIFS Security

=head1 ATTRIBUTES

=head2 command_interface

See L<NACL::C::Component::command_interface|lib-NACL-C-Component-pm/command_interface>

=head2 vserver

The name of the vserver, used to identify it uniquely.

=cut

package NACL::C::VserverCifsSecurity;

use strict;
use warnings;
use Params::Validate qw(validate_with SCALAR UNDEF);
use NACL::CS::VserverCifsSecurity;
use NACL::ComponentUtils qw (_optional_scalars);
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Data::Dumper;
use base 'NACL::C::Component::ONTAP';

=head1 METHODS

=head2 new

  my $cifs_security = NACL::C::VserverCifsSecurity->new(command_interface=>$ci,
                                                     vserver =>$vserver_name);

Return a new Vserver CIFS Security component object with the given attributes.

=cut

use Class::MethodMaker [scalar => 'vserver',];

sub init {
    $Log->enter() if $may_enter;
    my ($self, @args) = @_;
    $self->SUPER::init(@args);
    $self->_vserver_context_init(@args);
    $Log->exit() if $may_exit;
}

=head2 modify

  NACL::C::VserverCifsSecurity->modify(command_interface=>$ci,
                                       vserver=>$vserver_name,...);

                            or

  $cifs_security_object->modify(...);

(Class or instance method) This method modifies an vserver cifs security
element.

Uses a CMode CLI/CMode ZAPI APISet.

Invokes "cifs-security-modify" API for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< vserver=>$vserver_name >>

(Required for class method, Not Applicable for instance method)
This parameter specifies the virtual server on which the CIFS server was
created.

Applicable for CMode CLI.

=item C<< 'is-signing-required' => $string >>

Applicable for CMode CLI/ZAPI.

=item C<< 'password-complexity-required' => $string >>

Require Password Complexity for Local User Accounts, possible value(s) are
true,false

Applicable for CMode CLI.

=item C<< 'is-password-complexity-required' => $string >>

Applicable for CMode CLI/ZAPI.

=item C<< 'kerberos-clock-skew' => $kerberos-clock-skew >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: kerberos-clock-skew

=item C<< 'kerberos-ticket-age' => $kerberos-ticket-age >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: kerberos-ticket-age

=item C<< 'kerberos-renew-age' => $kerberos-renew-age >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: kerberos-renew-age

=item C<< "extended_query" => $string >>

(Optional)

Applicable for CMode CLI.

=item C<< 'use-start-tls-for-ad-ldap' => $string >>


Use start_tls for CIFS LDAP connections possible value(s) are, true,false
Applicable for CMode CLI/ZAPI

=item C<< 'is-aes-encryption-enabled' => $string >>


Is AES-128 and AES-256 Encryption Enabled possible value(s) are, true,false
Applicable for CMode CLI/ZAPI

=item C<< 'admin-username' => $string >>

Password Reset Username
Applicable for CMode CLI

=item C<< 'admin-password' => $string >>

Password Reset Password
Applicable for CMode CLI

=item C<< 'lm-compatibility-level' => $string >>


LM Compatibility Level possible value(s) are, lm-ntlm-ntlmv2-krb,ntlm-ntlmv2-krb,ntlmv2-krb,krb
Applicable for CMode CLI

=item C<< is_smb_encryption_required >>

Require SMB Encryption for Incoming CIFS Traffic possible value(s) are, true,false

Filled in for CMode CLI/ZAPI.

=item C<< '*' >>

This method takes all other options supported by
"vserver cifs security modify" command in ONTAP.

Applicable for CMode CLI/ZAPI.

=item C<< username=>$username>>

(Optional)

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: admin-username

=item C<< password=>$password>>

(Optional)

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: admin-password

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub modify {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( kerberos-clock-skew kerberos-ticket-age
                    kerberos-renew-age username password
                    is-signing-required

                    password-complexity-required

                    is-password-complexity-required

                    use-start-tls-for-ad-ldap

                    is-aes-encryption-enabled

                    admin-username admin-password

                    lm-compatibility-level

                    is-smb-encryption-required
                    )
            )
        },
        cli_cmd => 'vserver cifs security modify',
    );
    my $vserver;
    $opts{vserver_operated_on} = \$vserver;
    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_modify_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_modify_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            }
        ],
        frontend => 'NACL::C::VserverCifsSecurity::modify',
    );

    $Log->exit() if $may_exit;
    return $ret;

} ## end sub modify

sub _modify_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, %opts) = @_;
    my $password        = delete $opts{password};
    my $username        = delete $opts{username};
    my $prompts_answers = [
        "Enter your user ID: "  => $username,
        "Enter your password: " => $password,
    ];
    $pkg->_base_backend_cmode_cli(
        %opts,
        api             => 'vserver_cifs_security_modify',
        additional_args => {'connectrec-match_table' => $prompts_answers},
        frontend_method => 'modify',
    );
    $Log->exit() if $may_exit;

} ## end sub _modify_cmode_cli

sub _modify_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg, @opts) = @_;
    $pkg->_base_backend_cmode_zapi(
        @opts,
        copy => [
            qw(use-start-tls-for-ad-ldap kerberos-clock-skew
                kerberos-ticket-age kerberos-renew-age
                is-signing-required is-password-complexity-required
                is-aes-encryption-enabled is-smb-encryption-required)
        ],
        map =>
            {"username" => "admin-username", "password" => "admin-password",},
        api             => 'cifs-security-modify',
        frontend_method => 'modify',
    );
    $Log->exit() if $may_exit;
} ## end sub _modify_cmode_zapi


__PACKAGE__->_load_event_class();

sub _primary_keys_validate_spec {
    return (vserver => {type => SCALAR | UNDEF, optional => 1},);
}

1;
