#
# Copyright (c) 2001-2013 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary VserverCifsShare Component Module
## @author Rupamjyoti.Baruah@netapp.com, dl-nacl-dev@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::C::VserverCifsShare

=head1 DESCRIPTION

C<NACL::C::VserverCifsShare> is a derived class of
L<NACL::C::Component::ONTAP|lib-NACL-C-Component-ONTAP-pm>.  It represents access
to an ONTAP Vserver CIFS Share.  A related class is
L<NACL::CS::VserverCifsShare|lib-NACL-CS-VserverCifsShare-pm>, which
represents the past state of an ONTAP Vserver CIFS Share

=head1 ATTRIBUTES

=head2 command_interface

See L<NACL::C::Component::command_interface|lib-NACL-C-Component-pm/command_interface>

=head2 vserver

The name of the vserver, used to identify it uniquely.

=head2 share-name

This parameter specifies the name of the CIFS share that is to be created. A
share name can be up to 256 characters long. If this is a dynamic share as
specified by the values of dynamicshare and dynamic-path on the
-share-properties parameter, you can make the share name dynamic by specifying
 the %u (user name) or %d (domain name) variables as the value of this
 parameter.

=cut

package NACL::C::VserverCifsShare;

use strict;
use warnings;

use NACL::CS::VserverCifsShare;
use NACL::ComponentUtils qw (_optional_scalars Dumper);
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use base 'NACL::C::Component::ONTAP';
use Params::Validate qw(validate_with SCALAR ARRAYREF UNDEF);
use NACL::Exceptions::InvalidChoice qw(:try);
use NACL::C::Exceptions::VserverCifsShare::AlreadyExists ();
use NACL::C::Exceptions::VserverCifsShare::DoesNotExist  ();

require Data::Dumper;

=head1 METHODS

=head2 new

  my $cifs_share = NACL::C::VserverCifsShare->new(command_interface=>$ci,
                                                  vserver =>$vserver_name,
                                                  ...);

Return a new Vserver CIFS Share component object with the given attributes.

=cut

use Class::MethodMaker [
    scalar => 'vserver',
    scalar => 'share_name',
];

sub init {
    $Log->enter() if $may_enter;
    my ($self, @args) = @_;
    $self->SUPER::init(@args);
    $self->_vserver_context_init(@args);
    $Log->exit() if $may_exit;
}

=head2 create

  NACL::C::VserverCifsShare->create(command_interface=>$ci,
                                    vserver=>$vserver_name,...);

(Class method) Request creation of a new vserver cifs share element, and
return a new  vserver cifs share component object that refers to it.  Does not
wait for vserver cifs share creation to complete.

Uses a CMode CLI/ZAPI, 7Mode CLI/ZAPI APISet.

Invokes "cifs-share-add" API for 7Mode ZAPI.

Invokes "cifs shares" command for 7Mode CLI.

Invokes "cifs-share-create" API for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< vserver=>$vserver_name >>

(Required) This parameter specifies the CIFS-enabled virtual server on which a
 CIFS share is to be created.

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: ZAPI will be invoked in the context of the vserver specified to this
option.

=item C<< share-name=>$share_name >>

(Required) This parameter specifies the name of the CIFS share that is to be
created. A share name can be up to 256 characters long. If this is a dynamic
share as specified by the values of dynamicshare and dynamicpath on the
-share-properties parameter, you can make the share name dynamic by specifying
 the %u (user name) or %d (domain name) variables as the value of this
 parameter..

Applicable for CMode CLI/ZAPI, 7Mode CLI/ZAPI

Maps to:

CM ZAPI: share-name

7M CLI : share name (Programmatic name: "share-name")

7M ZAPI: share-name

=item C<< path=>$path >>

(Required) This parameter specifies the path to the CIFS share. This path must
 exist in a volume. A directory path name can be up to 255
 characters long. If there is a space in the path name, the entire string must
 be quoted (for example, "/new volume/mount here"). If this is a dynamic share
 as specified by values of dynamicshare and dynamicpath on the
 -share-properties parameter, you can make the path name dynamic by specifying
 the %u (user name) or %d (domain name) variables as the value of this
 parameter.

Applicable for CMode CLI/ZAPI, 7Mode CLI/ZAPI

Maps to:

CM ZAPI: path

7M CLI : path name (Programmatic name: "path")

7M ZAPI: path

=item C<< share-properties => $share_properties >>

(Optional) This parameter specifies  a list of properties for the share.
(default: oplocks)
{dynamicshare, dynamicpath, oplocks, browsable, showsnapshot, changenotify}

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: share-properties

=item C<< file-umask => $file_umask >>

(Optional) (File Mode Creation Mask) This optional parameter specifies the
default UNIX umask for new files created on the share. If this parameter is
not specified, the mask defaults to 022..

Applicable for CMode CLI/ZAPI, 7Mode CLI/ZAPI

Maps to:

CM ZAPI: file-umask

7M CLI : -file_umask <value> (Programmatic name: "file-umask")

7M ZAPI: file-umask

=item C<< dir-umask => $dir_umask >>

(Optional) (Directory Mode Creation Mask)
This optional parameter specifies the default UNIX umask for new directories
created on the share. If this parameter is not specified, the mask defaults to
 022.

Applicable for CMode CLI/ZAPI, 7Mode CLI/ZAPI

Maps to:

CM ZAPI: dir-umask

7M CLI : -dir_umask <value> (Programmatic name: "dir-umask")

7M ZAPI: dir-umask

=item C<< comment => $comment >>

(Optional) This optional parameter specifies a text comment for the share that
is made available to Windows clients. The comment can be up to 255 characters
long. If there is a space in the descriptive remark or the path, the entire
string must be quoted (for example, "This is engineering's share.").

Applicable for CMode CLI/ZAPI, 7Mode CLI/ZAPI

Maps to:

CM ZAPI: comment

7M CLI : -comment <value> (Programmatic name: "comment")

7M ZAPI: comment

=item C<< caching => $caching >>

(Optional) If specified, the following is done based on the value of the string:
 "no_caching": disallow Windows clients from caching any files on this share.
 This is the default value. "auto_document_caching": allow Windows clients to
 cache user documents on this share. "auto_program_caching": allow Windows
 clients to cache programs on this share. The actual caching behavior depends on
 the Windows client.

Applicable for 7Mode ZAPI.

=item C<< branchcache => $branchcache >>

Applicable for 7Mode CLI.

=item C<< forcegroup => $group >>

(Optional) If specified, provides name of the group to which files to be created
 in the share belong. The groupname is the name of a group in the UNIX group
 database. If it is an empty string or else not specified, then files to be
 created in the share do not belong to a particular UNIX group.
 That is, each file belongs to the same group as the owner of the file.

Applicable for 7Mode CLI/ ZAPI

Maps to:

7M CLI : -forcegroup <value> (Programmatic name: "forcegroup")

7M ZAPI: forcegroup

=item C<< is-access-based-enum => $value >>

(Optional) If true Access Based Enumeration (ABE) is enabled, else it is
disabled. ABE filtered shared folders are visible to a user based on that
individual user's access rights, preventing the display of folders or other
shared resources that the user does not have rights to access.

Applicable for 7Mode CLI/ ZAPI

Maps to:

7M CLI : -accessbasedenum (Programmatic name: "'accessbasedenum")

7M ZAPI: is-access-based-enum

=item C<< is-browse => $is_browse >>

(Optional) If true or not specified, shares are visible to browsing clients and
 can be enumerated (e.g. "net view", SrvMgr, AD).

Applicable for 7Mode CLI/ ZAPI

Maps to:

7M CLI : -nobrowse (Programmatic name: "nobrowse")

7M ZAPI: is-browse

=item C<< is-namespace-caching-allowed => $is_namespace_caching_allowed >>

(Optional) If true, namespace caching is enabled on the share. If namespace
caching is enabled on a share, clients are allowed to cache the directory
enumeration results for better performance.

Applicable for 7Mode CLI/ ZAPI

Maps to:

7M CLI : -namespace_caching (Programmatic name: "namespace-caching")

7M ZAPI: is-namespace-caching-allowed

=item C<< is-symlink-strict-security => $value >>

(Optional) If true or not specified, strict symlink security is enabled. If
false, allows clients to follow symbolic links to destinations on this filer
but outside of the current share. It is not checked if the client is
authenticated to the symbolic link's    destination.

Applicable for 7Mode CLI/ ZAPI

Maps to:

7M CLI : -nosymlink_strict_security (Programmatic name: "is-symlink-strict-security")

7M ZAPI: is-symlink-strict-security

=item C<< is-vscan => $is_vscan >>

(Optional) If true or not specified, virus scan is done when clients open files
on this share, else no virus scan is done.

Applicable for 7Mode CLI/ ZAPI

Maps to:

7M CLI : -novscan (Programmatic name: "novscan")

7M ZAPI: is-vscan

=item C<< is-vscanread => $is_vscanread >>

(Optional) If true or not specified, virus scan is done when clients open files
on this share for read access, else no virus scan is done for read access on
this share.

Applicable for 7Mode CLI/ ZAPI

Maps to:

7M CLI : -noscanread (Programmatic name: "novscanread")

7M ZAPI: is-vscanread

=item C<< is-widelink => $is_widelink >>

(Optional) If true, allows clients to follow absolute symbolic links outside of
this share, subject to NT security. This feature requires an entry in the
/etc/symlink.translations file and it requires that the client supports
Microsoft's Distributed File System (DFS). If false or not specified, widelinks
in the share are disabled.

Applicable for 7Mode CLI/ ZAPI

Maps to:

7M CLI : -widelink (Programmatic name: "widelink")

7M ZAPI: is-widelink

=item C<< maxusers => $comment >>

(Optional) If specified, gives the maximum number of simultaneous connections to
 the new share. It must be a positive number. If not specified, the filer does
 not impose a limit on the number of connections to the share.

Applicable for 7Mode CLI/ ZAPI

Maps to:

7M CLI : -maxusers (Programmatic name: "maxusers")

7M ZAPI: maxusers

=item C<< umask => $umask >>

(Optional) If specified, sets file mode creation mask for a share in qtrees with
 Unix or mixed security styles. The mask restricts the initial permissions
 setting of a newly created file or directory. If not specified, the file mode
 creation mask of the share is 0. This field is ignored when both dir-umask and
 file-umask are present.

Applicable for 7Mode CLI/ ZAPI

Maps to:

7M CLI : -umask <value> (Programmatic name: "umask")

7M ZAPI: umask

=item C<< 'offline-files' => $string >>

Offline Files (default: manual)

Applicable for CMode CLI/ZAPI.

=item C<< "symlink-properties" => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: symlink-properties

=item C<< "extended_query" => $string >>

(Optional)

Applicable for CMode CLI.

=item C<< 'vscan-fileop-profile' => $string >>


Vscan File-Operations Profile possible value(s) are, no_scan,standard,strict,writes_only
Applicable for CMode CLI/ZAPI

=item C<< 'max-connections-per-share' => $string >>

Maximum tree-connect on share
Applicable for CMode CLI/ZAPI

=item C<< 'force-group-for-create' => $string >>

UNIX Group for File Create
Applicable for CMode CLI/ZAPI

=item C<< 'attribute-cache-ttl' => $string >>

Applicable for CMode CLI/ZAPI

=item C<< 'is-path-validation-enabled' => $string >>


Enable/Disable the CIFS share path validation possible value(s) are, true,false
Applicable for CMode CLI

=item C<< '*' >>

This method takes all other options supported by
"vserver cifs share create" command in ONTAP.

Applicable for CMode CLI

=back

=back

=over

=item Exceptions

=over

=item C<NACL::C::Exceptions::VserverCifsShare::AlreadyExists>

This type of exception gets thrown if an attempt is made to create a
cifs share that already exists.

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub create {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my $ret;
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            "share-properties"   => {type => ARRAYREF, optional => 1},
            "symlink-properties" => {type => ARRAYREF, optional => 1},
            _optional_scalars(
                qw( path file-umask dir-umask comment attribute-cache-ttl
                    maxusers forcegroup is-widelink is-symlink-strict-security
                    is-vscan is-vscanread umask caching is-access-based-enum
                    is-browse is-namespace-caching-allowed branchcache
                    offline-files

                    vscan-fileop-profile

                    max-connections-per-share

                    force-group-for-create

                    is-path-validation-enabled
                    )
            )
        },
        api => 'vserver_cifs_share_create',
    );
    my $vserver;
    $opts{vserver_operated_on} = \$vserver;
    try {
        $ret = $pkg->call_on_apiset(
            %opts,
            choices => [
                {   method    => '_create_cmode_cli',
                    interface => 'CLI',
                    set       => 'CMode',
                },
                {   method    => '_create_cmode_zapi',
                    interface => 'ZAPI',
                    set       => 'CMode',
                    check     => '_generic_zapi_check'
                },
                {   method    => '_create_7mode_cli',
                    interface => 'CLI',
                    set       => '7Mode',
                },
                {   method    => '_create_7mode_zapi',
                    interface => 'ZAPI',
                    set       => '7Mode',
                },
            ],
            frontend => 'NACL::C::VserverCifsShare::create',
        );
    } ## end try
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = shift;
        my $text      = $exception->text();

        if (   $text =~ /.*Duplicate\s+entry.*/i
            || $text =~ /.* already a share named.*/i)
        {
            $Log->exit() if $may_exit;
            NACL::C::Exceptions::VserverCifsShare::AlreadyExists
                ->convert_and_throw(exception => $exception);
        } else {
            $Log->exit() if $may_exit;
            $exception->throw();
        }
    };
    $Log->exit() if $may_exit;
    return $ret;
} ## end sub create

sub _create_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );

    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};

    # for options that have the same name in 7Mode and CMode
    my $common_options_7mode = [qw(path comment)];

    # options that are exclusive to 7Mode (common between CLI and ZAPI)
    my $exclusive_7mode_options
        = [qw( maxusers forcegroup umask branchcache )];

    # for options that have different names in 7Mode CLI and ZAPI
    my $key_map_7mode_options = {
        'share-name'                   => "sharename",
        'dir-umask'                    => 'dir_umask',
        'file-umask'                   => 'file_umask',
        'is-widelink'                  => 'widelink',
        'is-access-based-enum'         => 'accessbasedenum',
        'is-namespace-caching-allowed' => 'namespace_caching',
        'is-symlink-strict-security'   => 'nosymlink_strict_security',
        'is-browse'                    => 'nobrowse',
        'is-vscan'                     => 'novscan',
        'is-vscanread'                 => 'novscanread',
    };

    my %final_7mode_args;

    $pkg->_hash_copy(
        source => \%opts,
        copy   => [@$exclusive_7mode_options, @$common_options_7mode],
        map    => $key_map_7mode_options,
        target => \%final_7mode_args,
    );

    $pkg->_handle_boolean_options(
        api_options     => \%final_7mode_args,
        boolean_options => [qw /widelink accessbasedenum namespace_caching/],
        is_7m_cli       => 1
    );

    # for switches of type '-no_foo', the switch must be sent if the value
    # for the 'is-foo' switch in the front end call is false

    foreach (qw /nosymlink_strict_security novscan novscanread nobrowse/) {
        if (defined $final_7mode_args{$_}
            && (   $final_7mode_args{$_} eq 'false'
                || $final_7mode_args{$_} eq '0'
                || $final_7mode_args{$_} eq '')
            )
        {
            $final_7mode_args{$_} = 1;
        } else {
            delete $final_7mode_args{$_};
        }
    } ## end foreach (...)

   # for the -nobrowse option, we will have to look at share-properties to see
   # if it does not contain "browsable" as one of the values in the array and
   # only then send in that switch
   # This has to happen only if the 'share-properties' parameter is passed in

    if ($opts{'share-properties'}) {
        $final_7mode_args{nobrowse} = 1
            if !grep { $_ =~ /browsable/ } @{$opts{'share-properties'}};
    }

    # for 7mode cli, the keys to the method call are
    # no_caching | auto_document_caching | -auto_program_caching

    if ($opts{'caching'}) {
        $final_7mode_args{$opts{caching}} = 1;
    }

    $apiset->cifs_shares(%final_7mode_args, add => 1);
    $Log->exit() if $may_exit;

    return $pkg->new(
        share_name        => $opts{'share-name'},
        command_interface => $command_interface
    );
} ## end sub _create_7mode_cli

sub _create_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, %opts) = @_;
    $pkg->_array_to_scalar_deprication_warning(
        params  => ['share-properties', 'symlink-properties'],
        options => \%opts
    );
    $pkg->_base_backend_cmode_cli(%opts, api => 'vserver_cifs_share_create',);
    $Log->exit() if $may_exit;

    return $pkg->new(
        vserver           => $opts{vserver},
        share_name        => $opts{'share-name'},
        command_interface => $opts{'command_interface'},
    );
} ## end sub _create_cmode_cli

sub _create_7mode_zapi {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};

    # for options that have the same name in 7Mode and CMode
    my $common_options_7mode = [qw(path comment dir-umask file-umask)];

    # options that are exclusive to 7Mode (common between CLI and ZAPI)
    my $exclusive_7mode_options = [
        qw( caching forcegroup
            is-symlink-strict-security is-vscan
            is-vscanread is-widelink maxusers path
            share-name umask is-browse
            is-access-based-enum
            is-namespace-caching-allowed)
    ];

    my %final_7mode_args;

    $pkg->_hash_copy(
        source => \%opts,
        copy   => [@$exclusive_7mode_options, @$common_options_7mode],
        target => \%final_7mode_args,
    );

   # for the -nobrowse option, we will have to look at share-properties to see
   # if it does not contain "browsable" as one of the values in the array and
   # only then send in that switch
   # This has to happen only if the 'share-properties' parameter is passed in

    if ($opts{'share-properties'}) {
        $final_7mode_args{'is-browse'} = 'false'
            if !grep { $_ =~ /browsable/ } @{$opts{'share-properties'}};
    }

    $apiset->cifs_share_add(%final_7mode_args);
    $Log->exit() if $may_exit;

    return $pkg->new(
        share_name        => $opts{'share-name'},
        command_interface => $command_interface,
    );
} ## end sub _create_7mode_zapi

sub _create_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg, %opts) = @_;
    my $copy = [
        qw(share-name path share-properties
            symlink-properties file-umask dir-umask comment
            attribute-cache-ttl force-group-for-create
            max-connections-per-share vscan-fileop-profile)
    ];
    $pkg->_base_backend_cmode_zapi(
        %opts,
        copy => $copy,
        map  => {'offline-files' => 'offline-files-mode'},
        api  => 'cifs-share-create',
        zapi_field_translations => {
            octal_to_decimal => [qw( dir-umask file-umask)],
            hrminsec_to_sec  => [qw(attribute-cache-ttl)],
        },
    );
    $Log->exit() if $may_exit;

    return $pkg->new(
        share_name        => $opts{'share-name'},
        command_interface => $opts{'command_interface'},
        vserver           => $opts{vserver},
    );
} ## end sub _create_cmode_zapi

=head2 delete

  NACL::C::VserverCifsShare->delete(command_interface=>$ci,
                                    vserver=>$vserver_name,...);

                            or

  $vserver_cifs_share_object->delete(...);

(Class or instance method) This method deletes an vserver cifs share element.

Uses a CMode CLI/ZAPI, 7Mode CLI/ZAPI APISet.

Invokes "cifs-share-delete" API for CMode ZAPI.

Invokes "cifs shares" command for 7Mode CLI.

Invokes "cifs-share-delete" API for 7Mode ZAPI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< vserver=>$vserver_name >>

(Required for class method, Not Applicable for instance method)
(Required) This parameter specifies the CIFS-enabled virtual server on which a
 CIFS share is to be created.

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: ZAPI will be invoked in the context of the vserver specified to this
option.

=item C<< share-name=>$share_name >>

(Required for class method, Not Applicable for instance method)
(Required) This parameter specifies the name of the CIFS share.

Applicable for CMode CLI/ZAPI,7Mode CLI.

Maps to:

CM ZAPI: share-name

7M CLI : share name (Programmatic name: "share-name")

=item C<< "extended_query" => $string >>

(Optional)

Applicable for CMode CLI.

=item C<< '*' >>

This method takes all other options supported by
"vserver cifs share delete" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<NACL::C::Exceptions::VserverCifsShare::DoesNotExist>

This type of exception gets thrown if an attempt is made to delete a
non-existing cifs share.

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub delete {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {force => {type => SCALAR, optional => 1}},
        cli_cmd         => 'vserver cifs share delete'
    );
    my $vserver;
    $opts{vserver_operated_on} = \$vserver;
    my $ret;
    try {
        $ret = $pkg_or_obj->call_on_apiset(
            %opts,
            choices => [
                {   method    => '_delete_cmode_cli',
                    interface => 'CLI',
                    set       => 'CMode',
                },
                {   method    => '_delete_cmode_zapi',
                    interface => 'ZAPI',
                    set       => 'CMode',
                    check     => '_generic_zapi_check'
                },
                {   method    => '_delete_7mode_cli',
                    interface => 'CLI',
                    set       => '7Mode',
                },
                {   method    => '_delete_7mode_zapi',
                    interface => 'ZAPI',
                    set       => '7Mode',
                }
            ],
            frontend => 'NACL::C::VserverCifsShare::delete',
        );
    } ## end try
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = shift;
        my $text      = $exception->text();
        if (   $text =~ /.*entry doesn't exist/i
            || $text =~ /.*no share named.*/i)
        {
            $Log->exit() if $may_exit;
            NACL::C::Exceptions::VserverCifsShare::DoesNotExist
                ->convert_and_throw(exception => $exception);
        } else {
            $Log->exit() if $may_exit;
            $exception->throw();
        }
    };
    $Log->exit() if $may_exit;
    return $ret;
} ## end sub delete

sub _delete_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};

    my %final_7mode_args;

    $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => ['force'],
        map    => {'share-name' => "sharename"},
        target => \%final_7mode_args,
    );

    $apiset->cifs_shares(%final_7mode_args, delete => 1);
    $Log->exit() if $may_exit;

} ## end sub _delete_7mode_cli

sub _delete_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@args,
        api     => 'vserver_cifs_share_delete',
    );
    $Log->exit() if $may_exit;
} ## end sub _delete_cmode_cli

sub _delete_7mode_zapi {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};

    # remove and hence ignore the other arguments
    %opts = ('share-name' => $opts{'share-name'});

    $apiset->cifs_share_delete(%opts);
    $Log->exit() if $may_exit;

} ## end sub _delete_7mode_zapi

sub _delete_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;

    $pkg_or_obj->_base_backend_cmode_zapi(@args,
        api    => 'cifs_share_delete',
        copy   => [qw(share-name)],
    );
    $Log->exit() if $may_exit;

} ## end sub _delete_cmode_zapi

=head2 modify

   NACL::C::VserverCifsShare->modify(
                                command_interface => $ci,
                                     share_name => $share_name,
                                     vserver => $vserver,
                                  ...);

                                                                 or

    $VserverCifsShare_object->modify(...);

(Class or instance method)
This method is used to modify a CIFS share.

Uses a CMode CLI/ZAPI, 7Mode CLI APISet.

Invokes "cifs-share-modify" API for CMode ZAPI.

Invokes "cifs shares" command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "share-name" => $share_name >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI, 7Mode CLI.

Maps to:

CM ZAPI: share-name

7M CLI : share name (Programmatic name: "sharename")

=item C<< "vserver" => $vserver >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI.

=item C<< "file-umask" => $string >>

Applicable for CMode CLI/ZAPI, 7Mode CLI.

Maps to:

CM ZAPI: file-umask

7M CLI : -file_umask (Programmatic name: "file-umask")

=item C<< "attribute-cache-ttl" => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: attribute-cache-ttl

=item C<< "comment" => $string >>

Applicable for CMode CLI/ZAPI, 7Mode CLI.

Maps to:

CM ZAPI: comment

7M CLI : -comment <value> (Programmatic name: "comment")

=item C<< "path" => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: path

=item C<< "symlink-properties" => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: symlink-properties

=item C<< "dir-umask" => $string >>

Applicable for CMode CLI/ZAPI, 7Mode CLI.

Maps to:

CM ZAPI: dir-umask

7M CLI : -dir_umask <value> (Programmatic name: "dir-umask")

=item C<< "share-properties" => [ $share-propertie1, $share-propertie2, ... ] >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: share-properties

=item C<< 'offline-files' => $string >>

Offline Files

Applicable for CMode CLI/ZAPI.

=item C<< is-symlink-strict-security => $value >>

Applicable for 7Mode CLI.

Maps to:

7M CLI : -nosymlink_strict_security (Programmatic name: "is-symlink-strict-security")

=item C<< is-widelink => $is_widelink >>

Applicable for 7Mode CLI.

Maps to:

7M CLI : -widelink (Programmatic name: "widelink")

=item C<< is-vscan => $is_vscan >>

Applicable for 7Mode CLI

Maps to:

7M CLI : -vscan (Programmatic name: "vscan")

=item C<< is-vscanread => $is_vscanread >>

Applicable for 7Mode CLI

Maps to:

7M CLI : -noscanread (Programmatic name: "vscanread")

=item C<< is-access-based-enum => $value >>

Applicable for 7Mode CLI/ZAPI

Maps to:

7M CLI : -accessbasedenum (Programmatic name: "'accessbasedenum")

=item C<< is-namespace-caching-allowed => $is_namespace_caching_allowed >>

Applicable for 7Mode CLI/ZAPI

Maps to:

7M CLI : -namespace_caching (Programmatic name: "namespace-caching")

=item C<< "extended_query" => $string >>

(Optional)

Applicable for CMode CLI.

=item C<< 'vscan-fileop-profile' => $string >>


Vscan File-Operations Profile possible value(s) are, no_scan,standard,strict,writes_only
Applicable for CMode CLI/ZAPI

=item C<< 'max-connections-per-share' => $string >>

Maximum tree-connect on share
Applicable for CMode CLI/ZAPI

=item C<< 'force-group-for-create' => $string >>

UNIX Group for File Create
Applicable for CMode CLI/ZAPI

=item C<< 'is-path-validation-enabled' => $string >>


Enable/Disable the CIFS share path validation possible value(s) are, true,false
Applicable for CMode CLI

=item C<< '*' >>

This method takes all other options supported by
"vserver cifs share modify" command in ONTAP.

Applicable for CMode CLI

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub modify {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            "share-properties"   => {type => ARRAYREF, optional => 1},
            "symlink-properties" => {type => ARRAYREF, optional => 1},
            _optional_scalars(
                qw( file-umask attribute-cache-ttl comment path dir-umask
                    symlink_strict_security nosymlink_strict_security branchcache
                    offline-files

                    vscan-fileop-profile

                    max-connections-per-share

                    force-group-for-create

                    is-path-validation-enabled
                    )
            ),
        },
        cli_cmd => 'vserver cifs share modify',
    );
    my $vserver;
    $opts{vserver_operated_on} = \$vserver;
    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_modify_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_modify_7mode_cli',
                interface => 'CLI',
                set       => '7Mode',
            },
            {   method    => '_modify_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
                check     => '_generic_zapi_check'
            }
        ],
        frontend => 'NACL::C::VserverCifsShare::modify',
    );

    $Log->exit() if $may_exit;
    return $ret;
} ## end sub modify

sub _modify_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );

    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};

    # for options that have the same name in 7Mode and CMode
    my $common_options_7mode = [qw(comment)];

    # options that are exclusive to 7Mode CLI & ZAPI
    my $exclusive_7mode_options
        = [qw( maxusers forcegroup umask branchcache )];

    # for options that have different names in 7Mode CLI
    my $key_map_7mode_options = {
        'share-name'                   => "sharename",
        'dir-umask'                    => 'dir_umask',
        'file-umask'                   => 'file_umask',
        'is-symlink-strict-security'   => 'symlink_strict_security',
        'is-widelink'                  => "widelink",
        'is-vscan'                     => "vscan",
        'is-vscanread'                 => "vscanread",
        'is-access-based-enum'         => "accessbasedenum",
        'is-namespace-caching-allowed' => "namespace_caching"
    };

    my %final_7mode_args;

    $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => [@$exclusive_7mode_options, @$common_options_7mode],
        map    => $key_map_7mode_options,
        target => \%final_7mode_args,
    );

    my %hash = {
        'is-symlink-strict-security'   => 'symlink_strict_security',
        'is-widelink'                  => "widelink",
        'is-vscan'                     => "vscan",
        'is-vscanread'                 => "vscanread",
        'is-access-based-enum'         => "accessbasedenum",
        'is-namespace-caching-allowed' => "namespace_caching",
    };

    foreach my $keys (keys(%hash)) {
        if ($opts{$keys} =~ /true/) {
            $final_7mode_args{$hash{$keys}} = 1;
        } elsif ($opts{$keys} =~ /false/) {
            if ($keys =~ /is-namespace-caching-allowed/) {
                $final_7mode_args{'no_' . $hash{$keys}} = 1;
            } else {
                $final_7mode_args{'no' . $hash{$keys}} = 1;
            }
        } ## end elsif ( $opts{$keys} =~ /false/)
    } ## end foreach my $keys ( keys(%hash...))

    if ($opts{'share-properties'}) {
        if (!grep { $_ =~ /browsable/ } @{$opts{'share-properties'}}) {
            $final_7mode_args{nobrowse} = 1;
        } else {
            $final_7mode_args{browse} = 1;
        }
    } ## end if ( $opts{'share-properties'...})

    if ($opts{'caching'}) {
        $final_7mode_args{$opts{caching}} = 1;
    }

    $apiset->cifs_shares(%final_7mode_args, change => 1);
    $Log->exit() if $may_exit;

    return $pkg_or_obj->new(
        share_name        => $opts{'share-name'},
        command_interface => $command_interface
    );

} ## end sub _modify_7mode_cli

sub _modify_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $command_interface = $opts{command_interface};
    
    # Need this to determine if _base_backend_cmode_cli should be invoked
    # later on.
    my %cmode_args = $pkg_or_obj->get_api_options(
        apiset  => $opts{apiset},
        api     => 'vserver_cifs_share_modify',
        options => \%opts,
    );
    my %primary_keys;
    $pkg_or_obj->_move_primary_keys(
        source => \%cmode_args,
        target => \%primary_keys
    );
    
    # Used to be possible to modify share-properties using
    # "vserver cifs share modify" till SN, but was then moved to
    # "vserver cifs share properties add/remove"
    my $share_properties = delete $opts{'share-properties'};
    if (defined $share_properties) {
        # Get the current share-properties
        require NACL::C::VserverCifsShareProperties;
        my $existing_share_properties = NACL::CS::VserverCifsShareProperties->fetch(
            command_interface => $command_interface,
            filter            => {%primary_keys},
            requested_fields  => [qw(share-properties)],
        )->share_properties();

        # Compare the existing share-properties against the desired list and
        # determine which properties (if any) need to be added or removed.
        require List::Compare;
        my $list_obj = List::Compare->new($existing_share_properties, $share_properties);
        my @shares_to_remove = $list_obj->get_Lonly();
        my @shares_to_add = $list_obj->get_Ronly();
        
        if (@shares_to_remove) {
            NACL::C::VserverCifsShareProperties->remove(
                %primary_keys,
                command_interface  => $command_interface,
                'share-properties' => \@shares_to_remove
            );
        }
        
        if (@shares_to_add) {
            NACL::C::VserverCifsShareProperties->add(
                %primary_keys,
                command_interface  => $command_interface,
                'share-properties' => \@shares_to_add
            );
        }
    }

    # Only if options are still left, then invoke vserver_cifs_share_modify. For
    # example, if only share-properties is provided, then we should not call
    # modify.
    if (keys %cmode_args) {
        $pkg_or_obj->_base_backend_cmode_cli(
            %opts,
            api => 'vserver_cifs_share_modify'
        );
    }

    $Log->exit() if $may_exit;
} ## end sub _modify_cmode_cli

sub _modify_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my $allowed_cmode_options = [
        qw( file-umask share-name attribute-cache-ttl comment path
            symlink-properties share-properties dir-umask
            force-group-for-create vscan-fileop-profile
            max-connections-per-share
            )
    ];

    $pkg_or_obj->_base_backend_cmode_zapi(@args,
        api    => 'cifs_share_modify',
        copy   => $allowed_cmode_options,
        map    => {'offline-files' => 'offline-files-mode'},
        zapi_field_translations => {hrminsec_to_sec => [qw(attribute-cache-ttl)]},
    );

    $Log->exit() if $may_exit;
} ## end sub _modify_cmode_zapi


__PACKAGE__->_load_event_class();

sub _primary_keys_validate_spec {
    return (
        vserver      => {type => SCALAR | UNDEF, optional => 1},
        'share-name' => {type => SCALAR},
    );
} ## end sub _primary_keys_validate_spec

# All of the "cifs-share-*" ZAPIs are available from RR onwards
# Check the build of the host to check if the ZAPI back-end is applicable or
# not. All the individual _check_* methods in this component should invoke
# this
sub _generic_zapi_check {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params => \@_,
        spec   => {
            command_interface => {isa => 'NACL::C::CommandInterface::ONTAP'}
        },
        allow_extra => 1
    );

    my $command_interface = $opts{command_interface};

    if ($command_interface->isa('NACL::C::Vserver')) {

        # No versioning possible for vservers, just return if a vserver
        # is the command interface
        $Log->debug('Versioning is currently not possible with vservers. '
                . 'Returning from the "check" subroutine');
        $Log->exit() if $may_exit;
        return;
    } ## end if ( $command_interface...)

    my $version_mgr = $command_interface->get_version_manager();
    my $release = $version_mgr->get_version_attribute(attribute => 'release');
    my $codename = $version_mgr->map_release_to_codename(release => $release);
    if ($codename =~ /(boilermaker)|(tricky)/i) {
        my $msg
            = "The 'cifs-share-*' ZAPIs are only available from "
            . "Rollingrock onwards but this host '"
            . $command_interface->name()
            . "' is running '$codename' (release string is '$release')\n";
        $Log->exit() if $may_exit;
        NACL::Exceptions::InvalidChoice->throw($msg);
    } ## end if ( $codename =~ /(boilermaker)|(tricky)/i)
    $Log->exit() if $may_exit;
} ## end sub _generic_zapi_check

1;
