# $Id$
# Copyright (c) 2001-2012 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
## @summary VserverExportPolicy Component Module
## @author Dheeraj.K@netapp.com@netapp.com, dl-nacl-dev@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::C::VserverExportPolicy

=head1 DESCRIPTION

C<NACL::C::VserverExportPolicy> is a derived class of
L<NACL::C::Component::ONTAP|lib-NACL-C-Component-ONTAP-pm>.  It represents access
to an ONTAP vserver export policy.  A related class is
L<NACL::CS::VserverExportPolicy|lib-NACL-CS-VserverExportPolicy-pm>, which represents
the past state of an ONTAP vserver export policy.

=head1 ATTRIBUTES

=head2 command_interface

See L<NACL::C::Component::command_interface|lib-NACL-C-Component-pm/command_interface>

=head2 vserver

The name of the vserver, used to identify it uniquely.

=head2 policyname

The name of the vserver policy, used to identify it uniquely.

=cut

package NACL::C::VserverExportPolicy;

use strict;
use warnings;

use NACL::CS::VserverExportPolicy;
use NATE::BaseException qw(:try);
use NATE::Exceptions::Argument qw(:try);
use NACL::ComponentUtils qw (_optional_scalars Dumper);
use NACL::C::Exceptions::VserverExportPolicy::AlreadyExists ();
use NACL::C::Exceptions::VserverExportPolicy::DoesNotExist  ();
use NACL::C::Exceptions::VserverExportPolicy::CannotDeleteDefault ();
use base 'NACL::C::Component::ONTAP';
use Params::Validate qw( validate_with SCALAR UNDEF);
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();

=head1 METHODS

=head2 new

my $vserver_export_policy = NACL::C::VserverExportPolicy->new(
                                            vserver => $vserver_name,
                                            policyname => $policy_name,
                                            ... );

Return a new Vserver Export Policy component object with the given attributes.

=cut

use Class::MethodMaker [
    scalar => 'policyname',
    scalar => 'vserver',
];

sub init {
    $Log->enter() if $may_enter;
    my ($self, @args) = @_;
    $self->SUPER::init(@args);
    $self->_vserver_context_init(@args);
    $Log->exit() if $may_exit;
}

sub isa {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, $kind) = @_;
    my $isa = $pkg_or_obj->_build_isa(
        kind  => $kind,
        alias => 'NACL::C::ExportPolicy'
    );
    $Log->exit() if $may_exit;
    return $isa;
}

=head2 create

  my $vserver_export_policy = NACL::C::VserverExportPolicy->create(
                                          command_interface=>$ci,
                                          vserver =>$vserver_name,
                                          policyname =>$policy_name,
                                          ...);

 (Class method) Request creation of a new export policy element, and  return
 a new vserver export policy component object that refers to it.  Does not wait
 for export policy creation to complete.

Uses a CMode CLI/ZAPI APISet.

Invokes "export-policy-create" API for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< vserver=>$vserver_name >>

(Required) Name of vserver to create an export policy.

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: ZAPI will be invoked in the context of the vserver specified to this
option.

=item C<< policyname=>$policy_name >>

(Required) Name of Export Policy.

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: policy-name

=item C<< "extended_query" => $string >>

(Optional)

Applicable for CMode CLI.

=item C<< other_options >>

This method takes other common options described in L<Common method options|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#Other_options_valid_on_component_methods>

=item C<< '*' >>

This method takes all other options supported by
"vserver export-policy create" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<NACL::C::Exceptions::VserverExportPolicy::AlreadyExists>

This type of exception gets thrown if an attempt is made to create a
export policy that already exists.

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub create {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    my %opts = $pkg->_common_validate_with(
        params  => \@args,
        cli_cmd => 'vserver export-policy create',
    );

    my $vserver;
    $opts{vserver_operated_on} = \$vserver;
    try {
        $pkg->call_on_apiset(
            %opts,
            choices => [
                {   method    => '_create_cmode_cli',
                    interface => 'CLI',
                    set       => 'CMode',
                },
                {   method    => '_create_cmode_zapi',
                    interface => 'ZAPI',
                    set       => 'CMode',
                },
            ],
            frontend => 'NACL::C::VserverExportPolicy::create',
        );
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = $_[0];
        NACL::C::Exceptions::VserverExportPolicy::AlreadyExists
            ->detect_convert_and_throw(exception => $exception);
        $Log->exit() if $may_exit;
        $exception->throw();
    };
    my $obj = $pkg->new(
        vserver           => $vserver,
        policyname        => $opts{policyname},
        command_interface => $opts{command_interface},
    );
    $Log->exit() if $may_exit;
    return $obj;
} ## end sub create

sub _create_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, @opts) = @_;
    $pkg->_base_backend_cmode_cli(@opts,
        api => 'vserver_export_policy_create',);
    $Log->exit() if $may_exit;
} ## end sub _create_cmode_cli

sub _create_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg, @opts) = @_;
    my $map = {'policyname' => 'policy-name',};
    $pkg->_base_backend_cmode_zapi(
        @opts,
        api => 'export-policy-create',
        map => $map,
    );
    $Log->exit() if $may_exit;
} ## end sub _create_cmode_zapi

=head2 delete

  my $vserver_export_policy = NACL::C::VserverExportPolicy->delete(
                                          command_interface=>$ci,
                                          vserver =>$vserver_name,
                                          policyname =>$policy_name,
                                          ...);
                           or

  $vserver_export_policy->delete(...);

(Class or instance method) This method deletes a vserver export policy element.

Uses a CMode CLI/ZAPI APISet.

Invokes "export-policy-destroy" API for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< vserver=>$vserver_name >>

(Required for class method, Not Applicable for instance method)
(Required) Name of vserver to delete an export policy.

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: ZAPI will be invoked in the context of the vserver specified to this
option.

=item C<< policiname=>$policy_name >>

(Required for class method, Not Applicable for instance method)
(Required) Name of Export Policy.

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: policy-name

=item C<< "extended_query" => $string >>

(Optional)

Applicable for CMode CLI.

=item C<< other_options >>

This method takes other common options described in L<Common method options|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#Other_options_valid_on_component_methods>

=item C<< '*' >>

This method takes all other options supported by
"vserver export-policy delete" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<NACL::C::Exceptions::VserverExportPolicy::DoesNotExist>

This type of exception gets thrown if an attempt is made to delete a
non-existing export policy.

=item C<NACL::C::Exceptions::VserverExportPolicy::CannotDeleteDefault>

This type of exception gets thrown if an attempt is made to delete the default export policy.

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub delete {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts = $pkg_or_obj->_common_validate_with(
        params  => \@args,
        cli_cmd => 'vserver export-policy delete',
    );
    my $vserver;
    $opts{vserver_operated_on} = \$vserver;
    try {
        $pkg_or_obj->call_on_apiset(
            %opts,
            choices => [
                {   method    => '_delete_cmode_cli',
                    interface => 'CLI',
                    set       => 'CMode',
                },
                {   method    => '_delete_cmode_zapi',
                    interface => 'ZAPI',
                    set       => 'CMode',
                },
            ],
            frontend => 'NACL::C::VserverExportPolicy::delete'
        );
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = $_[0];
        NACL::C::Exceptions::VserverExportPolicy::DoesNotExist
            ->detect_convert_and_throw(exception => $exception);
        NACL::C::Exceptions::VserverExportPolicy::CannotDeleteDefault
            ->detect_convert_and_throw(exception => $exception);
        $Log->exit() if $may_exit;
        $exception->throw();
    };
    $Log->exit() if $may_exit;
    return;
} ## end sub delete

sub _delete_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@opts,
        api => 'vserver_export_policy_delete',);
    $Log->exit() if $may_exit;
} ## end sub _delete_cmode_cli

sub _delete_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    my $map = {'policyname' => 'policy-name',};
    $pkg_or_obj->_base_backend_cmode_zapi(
        @opts,
        api => 'export-policy-destroy',
        map => $map,
    );
    $Log->exit() if $may_exit;
} ## end sub _delete_cmode_zapi

=head2 rename

  my $vserver_export_policy = NACL::C::VserverExportPolicy->rename(
                                          command_interface=>$ci,
                                          vserver =>$vserver_name,
                                          policyname =>$policy_name,
                                          newpolicyname =>$new_policy_name,
                                          ...);
                           or

  $vserver_export_policy->rename(...);

(Class or instance method) This method renames a vserver export policy element
to a new name.

Uses a CMode CLI/ZAPI APISet.

Invokes "export-policy-rename" API for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< vserver=>$vserver_name >>

(Required for class method, Not Applicable for instance method)
(Required) Name of vserver to create an export policy.

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: ZAPI will be invoked in the context of the vserver specified to this
option.

=item C<< policyname=>$policy_name >>

(Required for class method, Not Applicable for instance method)
(Required) Name of Export Policy.

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: policy-name

=item C<< newpolicyname=>$new_policy_name >>

(Required) New name of Export Policy.

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: new-policy-name

=item C<< "extended_query" => $string >>

(Optional)

Applicable for CMode CLI.

=item C<< other_options >>

This method takes other common options described in L<Common method options|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#Other_options_valid_on_component_methods>

=item C<< '*' >>

This method takes all other options supported by
"vserver export-policy rename" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub rename {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {_optional_scalars(qw(newpolicyname))},
        cli_cmd         => 'vserver export-policy rename',
    );
    my $vserver;
    $opts{vserver_operated_on} = \$vserver;
    $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_rename_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_rename_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::VserverExportPolicy::rename',
    );

    $pkg_or_obj->policyname($opts{newpolicyname}) if ref($pkg_or_obj);
    $Log->exit() if $may_exit;

} ## end sub rename

sub _rename_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@opts,
        api => 'vserver_export_policy_rename',);
    $Log->exit() if $may_exit;
} ## end sub _rename_cmode_cli

sub _rename_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    my $map = {
        'policyname'    => 'policy-name',
        'newpolicyname' => 'new-policy-name'};
    $pkg_or_obj->_base_backend_cmode_zapi(
        @opts,
        api  => 'export-policy-rename',
        map  => $map,
    );
    $Log->exit() if $may_exit;
} ## end sub _rename_cmode_zapi

=head2 copy

  my $vserver_export_policy = NACL::C::VserverExportPolicy->copy(
                                          command_interface=>$ci,
                                          vserver =>$vserver_name,
                                          newvserver =>$new_vserver_name,
                                          policyname =>$policy_name,
                                          newpolicyname =>$new_policy_name,
                                          ...);
                           or

  $vserver_export_policy->rename(...);

(Class or instance method) This method renames a vserver export policy element
to a new name.

Uses a CMode CLI APISet.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< vserver=>$vserver_name >>

(Required for class method, Not Applicable for instance method)
(Required) Name of vserver to create an export policy.

Applicable for CMode CLI.

=item C<< newvserver=>$new_vserver_name >>

(Required) Name of new vserver where policy will be copied.

Applicable for CMode CLI.

=item C<< policiname=>$policy_name >>

(Required for class method, Not Applicable for instance method)
(Required) Name of Export Policy.

Applicable for CMode CLI.

=item C<< newpoliciname=>$new_policy_name >>

(Required) New name of Export Policy.

Applicable for CMode CLI.

=item C<< "extended_query" => $string >>

(Optional)

Applicable for CMode CLI.

=item C<< other_options >>

This method takes other common options described in L<Common method options|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#Other_options_valid_on_component_methods>

=item C<< '*' >>

This method takes all other options supported by
"vserver export-policy copy" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub copy {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {_optional_scalars(qw(newvserver newpolicyname))},
        cli_cmd         => 'vserver export-policy copy',
    );
    my $vserver;
    $opts{vserver_operated_on} = \$vserver;
    $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => "_copy_cmode_cli",
                interface => "CLI",
                set       => "CMode",
            },
        ],
        frontend => 'NACL::C::VserverExportPolicy::copy',
    );
    my $obj = NACL::C::VserverExportPolicy->new(
        vserver           => $opts{'newvserver'},
        policyname        => $opts{'newpolicyname'},
        command_interface => $opts{command_interface},
    );
    $Log->exit() if $may_exit;
    return $obj;
} ## end sub copy

sub _copy_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@opts,
        api => 'vserver_export_policy_copy',);
    $Log->exit() if $may_exit;
} ## end sub _copy_cmode_cli

=head2 check_access

    my $check_access_obj = NACL::C::VserverExportPolicy->check_access(command_interface => $ci,
                          [ filter => { %filter } ,]
			  [ requested_fields => [ @req_fields ]);

                    or

    my @check_access_objs = NACL::C::VserverExportPolicy->check_access(command_interface => $ci,
                          [ filter => { %filter } ,]
			  [ requested_fields => [ @req_fields ]);

                    or

   my $check_access_obj = $VserverExportPolicy_obj->check_access();

(Class or Instance method)

This method is a CS-like fetch method which accepts all the parameters of
L<fetch|lib-NACL-CS-ComponentState-pm/fetch> (requested_fields, filter, allow_empty)
and has all of the other properties of fetch methods (callable in scalar or list
context, throws a NoElementsFound exception if no matching elements are found).
The requested fields and filter takes L<ATTRIBUTES|lib-NACL-CS-VserverExportPolicyCheckAccess-pm/ATTRIBUTES>
of NACL::CS::VserverExportPolicyCheckAccess as input.
This method returns objects of the type L<NACL::CS::VserverExportPolicyCheckAccess|lib-NACL-CS-VserverExportPolicyCheckAccess-pm>.

=cut

sub check_access {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts = $pkg_or_obj->_handle_obj_cs_like_calls(@args);
    require NACL::CS::VserverExportPolicyCheckAccess;
    my @state_objs = NACL::CS::VserverExportPolicyCheckAccess->fetch(%opts);
    $Log->exit() if $may_exit;
    return wantarray ? @state_objs : $state_objs[0];
}


__PACKAGE__->_load_event_class();

sub _primary_keys_validate_spec {
    return (
        policyname => {type => SCALAR},
        vserver    => {type => SCALAR | UNDEF, optional => 1}
    );
} ## end sub _primary_keys_validate_spec

1;
