# Copyright (c) 2001-2015 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary VserverExportPolicyRule Component Module
## @author Dheeraj.K@netapp.com@netapp.com, dl-nacl-dev@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::C::VserverExportPolicyRule

=head1 DESCRIPTION

C<NACL::C::VserverExportPolicyRule> is a derived class of
L<NACL::C::Component::ONTAP|lib-NACL-C-Component-ONTAP-pm>.  It represents access
to an ONTAP Vserver Export Policy Rule.  A related class is
L<NACL::CS::VserverExportPolicyRule|lib-NACL-CS-VserverExportPolicyRule-pm>, which represents
the past state of an ONTAP Vserver Export Policy Rule. A Vserver can act as
command interface to other components.See
L<NACL::C::Component::command_interface|lib-NACL-C-Component-pm/command_interface>

=head1 ATTRIBUTES

=head2 vserver

The name of the vserver, used to identify it uniquely.

=head2 policyname

The name of the vserver policy, used to identify it uniquely.

=head2 ruleindex

The export rule index number.

=cut

package NACL::C::VserverExportPolicyRule;

use strict;
use warnings;

use NACL::CS::VserverExportPolicyRule;
use NACL::C::Exceptions::VserverExportPolicyRule::DoesNotExist ();
use NACL::APISet::Exceptions::ResponseException qw(:try);
use NACL::ComponentUtils qw (_optional_scalars);
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use base 'NACL::C::Component::ONTAP';
use Params::Validate qw(validate_with SCALAR ARRAYREF UNDEF HASHREF);

=head1 METHODS

=head2 new

my $export_policy_rule = NACL::C::VserverExportPolicyRule->new(
                                            vserver => $vserver_name,
                                            policyname => $policy_name,
                                            ruleindex  => $rule_index,
                                            ... );

Return a new Vserver Export Policy Rule component object with the given
attributes.

=cut

use Class::MethodMaker [
    scalar => 'policyname',
    scalar => 'vserver',
    scalar => 'ruleindex',
];

sub init {
    $Log->enter() if $may_enter;
    my ($self, @args) = @_;
    $self->SUPER::init(@args);
    $self->_vserver_context_init(@args);
    $Log->exit() if $may_exit;
}

sub isa {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, $kind) = @_;
    my $isa = $pkg_or_obj->_build_isa(
        kind  => $kind,
        alias => 'NACL::C::ExportPolicyRule'
    );
    $Log->exit() if $may_exit;
    return $isa;
}

=head2 create

  my $export_policy_rule = NACL::C::VserverExportPolicyRule->create(
                                          command_interface=>$ci,
                                          vserver =>$vserver_name,
                                          policyname =>$policy_name,
                                          ruleindex  =>$rule_index,
                                          ...);

(Class method) Request creation of a new export policy rule element, and
return a new vserver export policy component object that refers to it.  Does not wait
for export policy creation to complete.

Uses a CMode CLI/ZAPI APISet.

Invokes "export-rule-create" API for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< vserver=>$vserver_name >>

(Required) Name of vserver to create an export policy.

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: ZAPI will be invoked in the context of the vserver specified to this
option.

=item C<< policyname=>$policy_name >>

(Required) Name of Export Policy.

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: policy-name


=item C<< ruleindex=>$rule_index >>

(Optional) Rule Index.

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: rule-index

=item C<< protocol=> [$protocol1, $protocol2, $protocol3, ...] >>

(Optional) Access Protocol.

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: protocol

=item C<< clientmatch=>$clientmatch >>

(Required) Client Match Spec.

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: client-match

=item C<< rorule=> [$rorule1, $rorule2, ...] >>

(Required) RO Access Rule.

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: ro-rule

=item C<< rwrule=> [$rwrule1, $rwrule2, $rwrule3, ...] >>

(Required) RW Access Rule.

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: rw-rule

=item C<< anon=>$anon >>

(Optional) User ID to which anonymous users are mapped.

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: anonymous-user-id

=item C<< superuser=> [$superuser1, $superuser2, ...] >>

(Optional) Superuser Security Flavors.

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: super-user-security

=item C<< 'allow-suid'=>$allow_suid >>

(Optional) Honor SetUID Bits In SETATTR (default: true).

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: is-allow-set-uid-enabled

=item C<< 'allow-dev'=>$allow_dev >>

(Optional) Allow Creation of Devices (default: true).

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: is-allow-dev-is-enabled

=item C<< 'ntfs-unix-security-ops'=>$ntfs_unix_security_ops >>

(Optional) NTFS Unix Security Options (default: fail).

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: export-ntfs-unix-security-ops

=item C<< 'chown-mode'=>$chown_mode >>

(Optional) Change Ownership Mode (default: restricted).

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: export-chown-mode

=item C<< "extended_query" => $string >>
Optional)

Applicable for CMode CLI.

=item C<< '*' >>

This method takes all other options supported by
"vserver export-policy rule create" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub create {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_arrayrefs(qw(rorule rwrule protocol superuser)),
            _optional_scalars(
                qw (clientmatch anon allow-suid vserver
                    allow-dev ntfs-unix-security-ops chown-mode
                    policyname ruleindex)
            )
        },
        cli_cmd => 'vserver export-policy rule create',
    );

    my $vserver;
    $opts{vserver_operated_on} = \$vserver;

    $pkg->call_on_apiset_with_exceptions(
        %opts,
        choices => [
            {   method    => '_create_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_create_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::VserverExportPolicyRule::create',
        nacl_exceptions => [qw(AlreadyExists)]
    );

    # If ruleindex is not supplied, ONTAP sets it to be one more than the last
    # one already present. In other words, if ruleindex is not supplied, we
    # need to query the filer for the last ruleindex, so we know what value to
    # set in the object.
    # However, if the user does not want an object back, then calling find()
    # is unnecessary and we can save time by not calling it (this is
    # particularly important if there are a large number of export rules; see
    # burt 808857). We can know if this method is invoked in void context
    # by checking if wantarray() returns undef.
    if (defined wantarray) {
        # User wants the object returned
        my (%filter_opt, $rule_index);

        if (defined $opts{ruleindex}) {
            $rule_index = $opts{ruleindex};
        } else {
            my %common_params = ();
            $pkg->_copy_common_component_params(
                source => \%opts,
                target => \%common_params
            );

            $filter_opt{policyname} = $opts{policyname};
            $filter_opt{vserver} = $opts{vserver} if defined($opts{vserver});

            my @ruleindexs = NACL::C::VserverExportPolicyRule->find(
                command_interface => $opts{command_interface},
                filter            => \%filter_opt,
                %common_params
            );
            $rule_index = $ruleindexs[$#ruleindexs]->ruleindex();
        }

        my %new_opts;
        $pkg->_copy_primary_keys(
            source => \%opts,
            target => \%new_opts
        );
        $pkg->_process_options(options => \%new_opts);

        my $obj = $pkg->new(
            %new_opts,
            command_interface => $opts{command_interface},
            ruleindex         => $rule_index,
        );
        $Log->exit() if $may_exit;
        return $obj;
    } else {
        # Calling in void context, return nothing
        $Log->exit() if $may_exit;
        return;
    }
} ## end sub create

sub _create_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, %opts) = @_;
    $pkg->_array_to_scalar_deprication_warning(
        params  => ['protocol', 'rorule', 'rwrule', 'superuser'],
        options => \%opts,
    );
    $pkg->_base_backend_cmode_cli(%opts,
        api => 'vserver_export_policy_rule_create',);
    $Log->exit() if $may_exit;
} ## end sub _create_cmode_cli

sub _create_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg, %opts) = @_;
    $pkg->_base_backend_cmode_zapi(
        %opts,
        api  => 'export-rule-create',
        copy => [qw(protocol)],
        map  => {
            "policyname"             => "policy-name",
            "ruleindex"              => "rule-index",
            "clientmatch"            => "client-match",
            "rorule"                 => "ro-rule",
            "rwrule"                 => "rw-rule",
            "anon"                   => "anonymous-user-id",
            "superuser"              => "super-user-security",
            "allow-suid"             => "is-allow-set-uid-enabled",
            "allow-dev"              => "is-allow-dev-is-enabled",
            "ntfs-unix-security-ops" => "export-ntfs-unix-security-ops",
            "chown-mode"             => "export-chown-mode"
        },
        frontend_method => 'create'
    );
    $Log->exit() if $may_exit;
} ## end sub _create_cmode_zapi

=head2 delete

  my $export_policy_rule = NACL::C::VserverExportPolicyRule->delete(
                                          command_interface=>$ci,
                                          vserver =>$vserver_name,
                                          policyname =>$policy_name,
                                          ruleindex  =>$rule_index,
                                          ...);
                           or

  $export_policy_rule->delete(...);

(Class or instance method) This method deletes a vserver export policy rule
element.

Uses a CMode CLI/ZAPI APISet.

Invokes "export-rule-destroy" API for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
 See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< vserver =>$vserver_name >>

(Required for class method, Not Applicable for instance method)
 Name of vserver.

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: ZAPI will be invoked in the context of the vserver specified to this
option.

=item C<< policyname=>$policy_name >>

(Required) Name of Export Policy.
(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: policy-name

=item C<< ruleindex=>$rule_index >>

(Optional) Rule Index.

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: rule-index

=item C<< "extended_query" => $string >>

(Optional)

Applicable for CMode CLI.

=item C<< '*' >>

This method takes all other options supported by
"vserver export-policy rule delete" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub delete {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params  => \@args,
        cli_cmd => 'vserver export-policy rule delete',
    );
    my $vserver;
    $opts{vserver_operated_on} = \$vserver;
    my $ret;
    $ret = $pkg_or_obj->call_on_apiset_with_exceptions(
                %opts,
                choices => [
                    {   method    => '_delete_cmode_cli',
                        interface => 'CLI',
                        set       => 'CMode',
                    },
                    {   method    => '_delete_cmode_zapi',
                        interface => 'ZAPI',
                        set       => 'CMode',
                        check     => '_delete_zapi_check'
                    },
                ],
                frontend => 'NACL::C::VserverExportPolicyRule::delete',
                nacl_exceptions => [qw(DoesNotExist)]
            );
    
    $Log->exit() if $may_exit;
    return $ret;

} ## end sub delete

sub _delete_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@opts,
        api => 'vserver_export_policy_rule_delete',);
    $Log->exit() if $may_exit;
} ## end sub _delete_cmode_cli

sub _delete_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    my $map = {
        'ruleindex'  => 'rule-index',
        'policyname' => 'policy-name',
    };
    $pkg_or_obj->_base_backend_cmode_zapi(
        @opts,
        api => 'export-rule-destroy',
        map => $map,
    );
    $Log->exit() if $may_exit;
} ## end sub _delete_cmode_zapi

sub _delete_zapi_check {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    my @fields_to_check = ('ruleindex');
    $pkg_or_obj->_fallback_for_relational_regex(%opts, fields_to_check => \@fields_to_check);

    $Log->exit() if $may_exit;
} ## end sub delete_zapi_check

=head2 modify

  my $export_policy_rule = NACL::C::VserverExportPolicyRule->modify(
                                          command_interface=>$ci,
                                          vserver =>$vserver_name,
                                          policyname =>$policy_name,
                                          ruleindex  =>$rule_index,
                                          ...);
                           or

  $export_policy_rule->modify(...);

(Class or instance method) This method modifys a vserver export policy rule
element.

Uses a CMode CLI/ZAPI APISet.

Invokes "export-rule-modify" API for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
 See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< vserver =>$vserver_name >>

(Required for class method, Not Applicable for instance method)
 Name of vserver.

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: ZAPI will be invoked in the context of the vserver specified to this
option.

=item C<< policyname=>$policy_name >>

(Required) Name of Export Policy.
(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: policy-name

=item C<< ruleindex=>$rule_index >>

(Required) Rule Index.
(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: rule-index

=item C<< protocol=> [$protocol1, $protocol2, $protocol3, ...] >>

(Optional) Access Protocol. {any|nfs2|nfs3|nfs|cifs|nfs4|flexcache}

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: protocol

=item C<< clientmatch=>$clientmatch >>

(Optional) Client Match Spec.

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: client-match

=item C<< rorule=> [$rorule1, $rorule2, $rorule3, ...] >>

(Optional) RO Access Rule. {any|none|never|krb5|ntlm|sys}

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: ro-rule

=item C<< rwrule=> [$rwrule1, $rwrule2, $rwrule3, ...] >>

(Optional) RW Access Rule. {any|none|never|krb5|ntlm|sys}

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: rw-rule

=item C<< anon=>$anon >>

(Optional) User ID to which anonymous users are mapped.

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: anonymous-user-id

=item C<< superuser=> [$superuser1, $superuser2, $superuser3, ...] >>

(Optional) Superuser Security Flavors.

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: super-user-security

=item C<< 'allow-suid'=>$allow_suid >>

(Optional) Honor SetUID Bits In SETATTR (default: true).

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: is-allow-set-uid-enabled

=item C<< 'allow-dev'=>$allow_dev >>

(Optional) Allow Creation of Devices (default: true).

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: is-allow-dev-is-enabled

=item C<< 'ntfs-unix-security-ops'=>$ntfs_unix_security_ops >>

(Optional) NTFS Unix Security Options (default: fail).

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: export-ntfs-unix-security-ops

=item C<< 'chown-mode'=>$chown_mode >>

(Optional) Change Ownership Mode (default: restricted).

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: export-chown-mode

=item C<< "extended_query" => $string >>

(Optional)

Applicable for CMode CLI.

=item C<< '*' >>

This method takes all other options supported by
"vserver export-policy rule modify" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub modify {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_arrayrefs(qw(rorule rwrule protocol superuser)),
            _optional_scalars(
                qw (clientmatch anon allow-suid
                    allow-dev ntfs-unix-security-ops chown-mode)
            )
        },
        cli_cmd => 'vserver export-policy rule modify',
    );

    my $vserver;
    $opts{vserver_operated_on} = \$vserver;

    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_modify_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_modify_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
                check     => "_modify_zapi_check",
            },
        ],
        frontend => 'NACL::C::VserverExportPolicyRule::modify',
    );

    $Log->exit() if $may_exit;
    return $ret;

} ## end sub modify

sub _modify_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    $pkg_or_obj->_array_to_scalar_deprication_warning(
        params  => ['protocol', 'rorule', 'rwrule', 'superuser'],
        options => \%opts,
    );
    $pkg_or_obj->_base_backend_cmode_cli(%opts,
        api => 'vserver_export_policy_rule_modify',);
    $Log->exit() if $may_exit;
} ## end sub _modify_cmode_cli

sub _modify_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    $pkg_or_obj->_base_backend_cmode_zapi(
        %opts,
        api  => 'export-rule-modify',
        copy => [qw(protocol)],
        map  => {
            "policyname"             => "policy-name",
            "ruleindex"              => "rule-index",
            "clientmatch"            => "client-match",
            "rorule"                 => "ro-rule",
            "rwrule"                 => "rw-rule",
            "anon"                   => "anonymous-user-id",
            "superuser"              => "super-user-security",
            "allow-suid"             => "is-allow-set-uid-enabled",
            "allow-dev"              => "is-allow-dev-is-enabled",
            "ntfs-unix-security-ops" => "export-ntfs-unix-security-ops",
            "chown-mode"             => "export-chown-mode"
        },
        frontend_method => 'modify',
    );
    $Log->exit() if $may_exit;
} ## end sub _modify_cmode_zapi

sub _modify_zapi_check {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    my @fields_to_check = ('ruleindex');
    $pkg_or_obj->_fallback_for_relational_regex(%opts, fields_to_check => \@fields_to_check);

    $Log->exit() if $may_exit;
} ## end sub _modify_zapi_check

=head2 setindex

  my $export_policy_rule = NACL::C::VserverExportPolicyRule->setindex(
                                          command_interface=>$ci,
                                          vserver =>$vserver_name,
                                          policyname =>$policy_name,
                                          ruleindex  =>$rule_index,
                                          ...);
                           or

  $export_policy_rule->setindex(...);

(Class or instance method) This method modifys a vserver export policy rule
element.

Uses a CMode CLI/ZAPI APISet.

Invokes "export-rule-set-index" API for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
 See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< vserver =>$vserver_name >>

(Required for class method, Not Applicable for instance method)
 Name of vserver.

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: ZAPI will be invoked in the context of the vserver specified to this
option.

=item C<< policyname=>$policy_name >>

(Required) Name of Export Policy.
(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: policy-name

=item C<< ruleindex=>$rule_index >>

(Required) Rule Index.
(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: rule-index

=item C<< newruleindex=>$new_rule_index >>

(Required) New rule index.

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: new-rule-index

=item C<< "extended_query" => $string >>

(Optional)

Applicable for CMode CLI.

=item C<< '*' >>

This method takes all other options supported by
"vserver export-policy rule setindex" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub setindex {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {_optional_scalars(qw (newruleindex))},
        cli_cmd         => 'vserver export-policy rule setindex',
    );
    my $vserver;
    $opts{vserver_operated_on} = \$vserver;
    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_setindex_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_setindex_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::VserverExportPolicyRule::setindex',
    );

    $Log->exit() if $may_exit;
    return $ret;

} ## end sub setindex

sub _setindex_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(%opts,
        api => 'vserver_export_policy_rule_setindex',);
    $pkg_or_obj->ruleindex($opts{'newruleindex'}) if ref($pkg_or_obj);
    $Log->exit() if $may_exit;
} ## end sub _setindex_cmode_cli

sub _setindex_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    my $map  = {
        'policyname'   => 'policy-name',
        'ruleindex'    => 'rule-index',
        'newruleindex' => 'new-rule-index',
    };
    $pkg_or_obj->_base_backend_cmode_zapi(
        %opts,
        api  => 'export-rule-set-index',
        map  => $map,
    );

    $pkg_or_obj->ruleindex($opts{'new-rule-index'})
        if ref($pkg_or_obj);

    $Log->exit() if $may_exit;
} ## end sub _setindex_cmode_zapi


__PACKAGE__->_load_event_class();

sub _primary_keys_validate_spec {
    return (
        policyname => {type => SCALAR},
        vserver    => {type => SCALAR | UNDEF, optional => 1},
        ruleindex  => {type => SCALAR | UNDEF, optional => 1},
    );
} ## end sub _primary_keys_validate_spec

sub _optional_arrayrefs {
    my (@names) = @_;
    return (map { $_ => {type => ARRAYREF, optional => 1} } @names);
}

1;
