#
# Copyright (c) 2014 NetApp Inc.
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary VserverIscsiSecurity Component Module
## @author dl-nacl-dev@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::C::VserverIscsiSecurity

=head1 DESCRIPTION

C<NACL::C::VserverIscsiSecurity> is a derived class of
L<NACL::C::Component::ONTAP|lib-NACL-C-Component-ONTAP-pm>. It represents access
to an ONTAP VserverIscsiSecurity.  A related class is
L<NACL::CS::VserverIscsiSecurity|lib-NACL-CS-VserverIscsiSecurity-pm>, which represents
the past state of an ONTAP VserverIscsiSecurity.

=head1 ATTRIBUTES

=head2 command_interface

See L<NACL::C::Component::command_interface|lib-NACL-C-Component-pm/command_interface>

=head2 initiator_name

=head2 vserver

=cut

package NACL::C::VserverIscsiSecurity;

use strict;
use warnings;

use NACL::CS::VserverIscsiSecurity;
use NACL::ComponentUtils qw (_optional_scalars Dumper);

use base 'NACL::C::Component::ONTAP';
use Params::Validate qw(validate validate_with SCALAR UNDEF);
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use NACL::Exceptions::InvalidChoice;
use NACL::C::Exceptions::VserverIscsiSecurity::AlreadyExists;
use NACL::C::Exceptions::VserverIscsiSecurity::DoesNotExist;
use NACL::C::Exceptions::VserverIscsiSecurity::NoNodesAvailable;


=head1 METHODS

=head2 new

  my $VserverIscsiSecurity = NACL::C::VserverIscsiSecurity->new(command_interface=>$ci,
                                   initiator_name =>$initiator_name,
                                   vserver =>$vserver,
                                    );

Return a new VserverIscsiSecurity component object with the given attributes.

=cut

use Class::MethodMaker [
    scalar => 'initiator',
    scalar => 'initiator_name',
    scalar => 'vserver',
];

sub init {
    $Log->enter() if $may_enter;
    my ($self, @args) = @_;
    $self->SUPER::init(@args);
    $self->_vserver_context_init(@args);
    $Log->exit() if $may_exit;
}

=head2 find

    my $VserverIscsiSecurity  = NACL::C::VserverIscsiSecurity->find(command_interface => $ci, ...);
    my @VserverIscsiSecuritys = NACL::C::VserverIscsiSecurity->find(command_interface => $ci, ...);

(Class method)
A constructor of sorts, which discovers which elements are present (by
interacting with the given command interface).  In an array context it
creates a component object for each present element and returns them.
In a scalar context it creates a component object for just one
element.

This method is generally not implemented in individual components. The base
class (L<NACL::C::Component::ONTAP|lib-NACL-C-Component-ONTAP-pm> provides an implementation.
See <L<NACL::C::Component::find|lib-NACL-C-Component-pm/find> for a more
detailed description along with explanation of all the options accepted.

=head2 state

    my $VserverIscsiSecurity_state = $NACL::C::VserverIscsiSecurity->state();

(Instance method)
Fetch the current state of the element associated
with this component, and return a Component State object corresponding
to this component (calling this on NACL::C::VserverIscsiSecurity would return
a NACL::CS::VserverIscsiSecurity object)

This method is generally not implemented in individual components. The base
class (L<NACL::C::Component::ONTAP|lib-NACL-C-Component-ONTAP-pm> provides an implementation.
See <L<NACL::C::Component::state|lib-NACL-C-Component-pm/state> for a more
detailed description along with explanation of all the options accepted.

=cut

=head2 modify

   NACL::C::VserverIscsiSecurity->modify(
                                command_interface => $ci,
                                initiator_name => $initiator_name,
                                vserver => $vserver,
                                ...);

(Class method)
This method is used to modify the iSCSI authentication configuration for an initiator.

Uses a CMode CLI/ZAPI, 7Mode CLI APISet.

Invokes "iscsi-initiator-modify-chap-params" API for CMode ZAPI.

Invokes "iscsi security add" command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "initiator-name" => $initiator_name >>

(Required)

Applicable for CMode CLI/ZAPI,7Mode CLI.

Maps to:

CM ZAPI: initiator

7M CLI : -i <value> (Programmatic name: "initiator")

=item C<< "vserver" => $vserver >>

(Required)

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: ZAPI will be invoked in the context of the vserver specified to this
option.

=item C<< "remove-outbound" => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: remove-outbound

=item C<< "auth-type" => $string >>

Applicable for CMode CLI, 7Mode CLI.

Maps to:

7M CLI : -s <value> (Programmatic name: "auth-type")

=item C<< "password" => $string >>

Applicable for CMode CLI/ZAPI,7Mode CLI.

Maps to:

CM ZAPI: password

7M CLI : -p <value> (Programmatic name: "force")

=item C<< "outbound-user-name" => $string >>

Applicable for CMode CLI/ZAPI,7Mode CLI.

Maps to:

CM ZAPI: password

7M CLI : -m <value> (Programmatic name: "outbound-user-name")

=item C<< "radius" => $string >>

Applicable for CMode CLI/ZAPI,7Mode CLI.

Maps to:

CM ZAPI: radius

7M CLI : -f <value> (Programmatic name: "radius")

=item C<< "outbound-password" => $string >>

Applicable for CMode CLI/ZAPI,7Mode CLI.

Maps to:

CM ZAPI: outbound-password

7M CLI : -o <value> (Programmatic name: "outbound-password")

=item C<< "user-name" => $string >>

Applicable for CMode CLI/ZAPI,7Mode CLI.

Maps to:

CM ZAPI: user-name

7M CLI : -n <value> (Programmatic name: "user-name")

=item C<< 'clear-outbound' => $string >>

Applicable for CMode CLI.

=item C<< "extended_query" => $string >>

(Optional)

Applicable for CMode CLI.

=item C<< '*' >>

This method takes all other options supported by
"vserver iscsi security modify" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=over

=item Returns

This method returns an object of type NACL::C::VserverIscsiSecurity.

=back

=back

=cut

sub modify {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( remove-outbound auth-type password outbound-user-name radius
                    outbound-password user-name initiator initiator-name

                    clear-outbound
                    )
            ),
        },
        cli_cmd => 'vserver iscsi security modify',
    );
    my $vserver;
    $opts{vserver_operated_on} = \$vserver;
    $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_modify_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_modify_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
            {   method    => '_modify_7mode_cli',
                interface => 'CLI',
                set       => '7Mode',
            },
        ],
        frontend => 'NACL::C::VserverIscsiSecurity::modify',
    );

    my $initiator = $opts{'initiator'} || $opts{'initiator-name'};
    my $obj = $pkg->new(
        command_interface => $opts{command_interface},
        initiator_name    => $initiator,
        vserver           => $opts{vserver},
    );

    $Log->exit() if $may_exit;
    return $obj;
} ## end sub modify

sub _modify_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_base_backend_cmode_cli(@args,
        api => 'vserver_iscsi_security_modify',
    );
    $Log->exit() if $may_exit;
} ## end sub _modify_cmode_cli


sub _modify_cmode_zapi{
    $Log->enter() if $may_enter;
    my ($pkg, %opts) = @_;
    my $apiset = $opts{apiset};

    my $allowed_cmode_options = [
        qw( remove-outbound outbound-user-name radius
            user-name )
    ];

    my $api;
    if ($pkg->_check_for_remove_outbound(%opts)) {
        delete $opts{'auth-type'} if (defined $opts{'auth-type'});
    }
    if (defined $opts{'auth-type'}){
        delete $opts{'clear-outbound'} if (defined $opts{'clear-outbound'});
        delete $opts{'remove-outbound'} if (defined $opts{'remove-outbound'});
        if((defined $opts{'initiator-name'}) && ($opts{'initiator-name'} eq 'default') ){
            delete $opts{'initiator-name'};
            $api = 'iscsi_initiator_set_default_auth';
        }
        else{
            $api = 'iscsi_initiator_add_auth';
        }
    }else{
        $api = 'iscsi_initiator_modify_chap_params';
    }

    $pkg->_base_backend_cmode_zapi(%opts,
        api => $api,
        copy   => $allowed_cmode_options,
        map    => {
                     'initiator-name' => 'initiator',
                     'password' => 'passphrase',
                     'outbound-password' => 'outbound-passphrase'
         },
    );
    $Log->exit() if $may_exit;
} ## end sub _modify_cmode_zapi

sub _modify_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );

    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};

    my $allowed_7mode_options = [
        qw( auth-type password outbound-user-name radius
            outbound-password user-name
            )
    ];
    my %quota_args;
    $pkg->_hash_copy(
        source => \%opts,
        copy   => $allowed_7mode_options,
        map    => {"initiator-name" => "initiator"},
        target => \%quota_args,
    );
    %opts = %quota_args;
    if ($quota_args{'auth-type'} eq "CHAP") {
        if ($quota_args{'radius'} eq "RADIUS") {
            delete $opts{'user-name'};
            delete $opts{'password'};
            delete $opts{'outbound-password'};
            delete $opts{'outbound-user-name'};
        } else {
            delete $opts{'radius'};
        }
    } else {
        delete $opts{'radius'};
        delete $opts{'user-name'};
        delete $opts{'password'};
        delete $opts{'outbound-password'};
        delete $opts{'outbound-user-name'};
    } ## end else [ if ( $quota_args{'auth-type'...
    $apiset->iscsi_security_add(%opts);
    $Log->exit() if $may_exit;
} ## end sub _modify_7mode_cli

=head2 create

    my $VserverIscsiSecurity_object = NACL::C::VserverIscsiSecurity->create(
                                     command_interface => $ci,
                                     initiator_name => $initiator_name,
                                     vserver => $vserver,
                                  ...);

                             or

    NACL::C::VserverIscsiSecurity->create(
                                     command_interface => $ci,
                                     initiator_name => $initiator_name,
                                     vserver => $vserver,
                                  ...);


(Class method)
This method is used to create an iSCSI authentication configuration for an initiator.

Uses a CMode CLI/ZAPI, 7Mode CLI APISet.

Invokes "iscsi-initiator-add-auth" API for CMode ZAPI.

Invokes "iscsi security add" command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "initiator-name" => $initiator_name >>

(Required)

Applicable for CMode CLI/ZAPI,7Mode CLI.

Maps to:

CM ZAPI: initiator

7M CLI : -i <value> (Programmatic name: "initiator")

=item C<< "vserver" => $vserver >>

(Required)

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: ZAPI will be invoked in the context of the vserver specified to this
option.

=item C<< "auth-type" => $string >>

Applicable for CMode CLI/ZAPI,7Mode CLI.

Maps to:

CM ZAPI: auth-type

7M CLI : -s <value> (Programmatic name: "auth-type")

=item C<< "password" => $string >>

Applicable for CMode CLI/ZAPI,7Mode CLI.

Maps to:

CM ZAPI: password

7M CLI : -p <value> (Programmatic name: "password")

=item C<< "outbound-user-name" => $string >>

Applicable for CMode CLI/ZAPI,7Mode CLI.

Maps to:

CM ZAPI: outbound-user-name

7M CLI : -m <value> (Programmatic name: "outbound-user-name")

=item C<< "radius" => $string >>

Applicable for CMode CLI/ZAPI,7Mode CLI.

Maps to:

CM ZAPI: radius

7M CLI : -f <value> (Programmatic name: "radius")

=item C<< "outbound-password" => $string >>

Applicable for CMode CLI/ZAPI,7Mode CLI.

Maps to:

CM ZAPI: outbound-password

7M CLI : -o <value> (Programmatic name: "outbound-password")

=item C<< "user-name" => $string >>

Applicable for CMode CLI/ZAPI,7Mode CLI.

Maps to:

CM ZAPI: user-name

7M CLI : -n <value> (Programmatic name: "user-name")

=item C<< "extended_query" => $string >>

(Optional)

Applicable for CMode CLI.

=item C<< '*' >>

This method takes all other options supported by
"vserver iscsi security create" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub create {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( auth-type password outbound-user-name radius outbound-password user-name
                    initiator-name initiator )
            ),
        },
        cli_cmd => 'vserver iscsi security create',
    );
    my $vserver;
    $opts{vserver_operated_on} = \$vserver;
    $pkg->call_on_apiset_with_exceptions(
        %opts,
        choices => [
            {   method    => '_create_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_create_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
            {   method    => '_create_7mode_cli',
                interface => 'CLI',
                set       => '7Mode',
            },
        ],
        frontend => 'NACL::C::VserverIscsiSecurity::create',
        nacl_exceptions => [qw(AlreadyExists NoNodesAvailable)]             
    );

    my $initiator = $opts{'initiator'} || $opts{'initiator-name'};
    my $obj = $pkg->new(
        command_interface => $opts{command_interface},
        initiator_name    => $initiator,
        vserver           => $opts{vserver},
    );

    $Log->exit() if $may_exit;
    return $obj;
} ## end sub create

sub _create_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_base_backend_cmode_cli(@args,
        api     => 'vserver_iscsi_security_create',
    );
    $Log->exit() if $may_exit;
} ## end sub _create_cmode_cli

sub _create_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg, %opts) = @_;
    my $apiset = $opts{apiset};

    my $allowed_cmode_options = [
        qw( auth-type outbound-user-name radius
            user-name )
    ];

    $pkg->_base_backend_cmode_zapi(%opts,
        api  => 'iscsi_initiator_add_auth',
        copy => $allowed_cmode_options,
        map  => {
                    'initiator-name' => 'initiator',
                    'password' => 'passphrase',
                    'outbound-password' => 'outbound-passphrase'
                },
    );

    $Log->exit() if $may_exit;
} ## end sub _create_cmode_zapi

sub _create_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );

    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};

    my $allowed_7mode_options = [
        qw( auth-type password outbound-user-name radius
            outbound-password user-name
            )
    ];
    my %quota_args;
    $pkg->_hash_copy(
        source => \%opts,
        copy   => $allowed_7mode_options,
        map    => {"initiator-name" => "initiator"},
        target => \%quota_args,
    );
    %opts = %quota_args;
    if ($quota_args{'auth-type'} eq "CHAP") {
        if ($quota_args{'radius'} eq "RADIUS") {
            delete $opts{'user-name'};
            delete $opts{'password'};
            delete $opts{'outbound-password'};
            delete $opts{'outbound-user-name'};
        } else {
            delete $opts{'radius'};
        }
    } else {
        delete $opts{'radius'};
        delete $opts{'user-name'};
        delete $opts{'password'};
        delete $opts{'outbound-password'};
        delete $opts{'outbound-user-name'};
    } ## end else [ if ( $quota_args{'auth-type'...
    $apiset->iscsi_security_add(%opts);
    $Log->exit() if $may_exit;
    return $pkg->new(
        command_interface => $command_interface,
        initiator_name    => $opts{'initiator'},
    );
} ## end sub _create_7mode_cli

=head2 generate

   NACL::C::VserverIscsiSecurity->generate(
                                     command_interface => $ci,
                                  ...);

(Class method)
This method is used to generate a 128-bit random CHAP secret.

Uses a CMode CLI/ZAPI, 7Mode CLI APISet.

Invokes "iscsi-auth-generate-chap-password" API for CMode ZAPI.

Invokes "iscsi security generate" command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "secret" => $string >>

Applicable for CMode CLI.

=item C<< "vserver" => $vserver >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: ZAPI will be invoked in the context of the vserver specified to this
option.

=item C<< '*' >>

This method takes all other options supported by
"vserver iscsi security generate" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub generate {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( secret initiator-name
                    )
            ),
        },
        cli_cmd => 'vserver iscsi security generate',
    );
    my $vserver;
    $opts{vserver_operated_on} = \$vserver;
    my $ret = $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_generate_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_generate_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
            {   method    => '_generate_7mode_cli',
                interface => 'CLI',
                set       => '7Mode',
            },
        ],
        frontend => 'NACL::C::VserverIscsiSecurity::generate',
    );

    $Log->exit() if $may_exit;
    return $ret;
} ## end sub generate

sub _generate_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, %opts) = @_;

    if (exists $opts{vserver}) {
        delete $opts{vserver};
    }

    $pkg->_base_backend_cmode_cli(%opts,
        api => 'vserver_iscsi_security_generate',
    );
    $Log->exit() if $may_exit;
} ## end sub _generate_cmode_cli

sub _generate_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg, %opts) = @_;
    delete $opts{'initiator-name'};
    $pkg->_base_backend_cmode_zapi(%opts,
        api => 'iscsi_auth_generate_chap_password',
    );
    $Log->exit() if $may_exit;
} ## end sub _generate_cmode_zapi

sub _generate_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );
    delete $opts{command_interface};
    my $apiset = delete $opts{apiset};

    my %quota_args;
    $pkg->_hash_copy(
        source => \%opts,
        target => \%quota_args,
    );
    %opts = %quota_args;

    $apiset->iscsi_security_generate(%opts);
    $Log->exit() if $may_exit;
} ## end sub _generate_7mode_cli

=head2 delete

   NACL::C::VserverIscsiSecurity->delete(
                                command_interface => $ci,
                                initiator_name => $initiator_name,
                                vserver => $vserver,
                                ...);

(Class method)
This method is used to delete the iSCSI authentication configuration for an initiator.

Uses a CMode CLI/ZAPI, 7Mode CLI APISet.

Invokes "iscsi-initiator-delete-auth" API for CMode ZAPI.

Invokes "iscsi security delete" command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "initiator-name" => $initiator_name >>

(Required)

Applicable for CMode CLI/ZAPI,7Mode CLI.

Maps to:

CM ZAPI: initiator

7M CLI : -i <value> (Programmatic name: "initiator")

=item C<< "vserver" => $vserver >>

(Required)

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: ZAPI will be invoked in the context of the vserver specified to this
option.

=item C<< 'initiator' => $string >>

Applicable for CMode CLI.

=item C<< "extended_query" => $string >>

(Optional)

Applicable for CMode CLI.

=item C<< '*' >>

This method takes all other options supported by
"vserver iscsi security delete" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub delete {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( initiator initiator-name
                    )
            )
        },
        cli_cmd => 'vserver iscsi security delete',
    );
    my $vserver;
    $opts{vserver_operated_on} = \$vserver;
    my $ret = $pkg->call_on_apiset_with_exceptions(
        %opts,
        choices => [
            {   method    => '_delete_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_delete_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
            {   method    => '_delete_7mode_cli',
                interface => 'CLI',
                set       => '7Mode',
            },
        ],
        frontend => 'NACL::C::VserverIscsiSecurity::delete',
        nacl_exceptions => [qw(DoesNotExist NoNodesAvailable)]
    );

    $Log->exit() if $may_exit;
    return $ret;
} ## end sub delete

sub _delete_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_base_backend_cmode_cli(@args,
        api     => 'vserver_iscsi_security_delete',
    );
    $Log->exit() if $may_exit;
} ## end sub _delete_cmode_cli

sub _delete_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_base_backend_cmode_zapi(@args,
        api    => 'iscsi_initiator_delete_auth',
        map    => {"initiator-name" => "initiator"},
    );
    $Log->exit() if $may_exit;
} ## end sub _delete_cmode_zapi

sub _delete_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );

    delete $opts{command_interface};
    my $apiset = delete $opts{apiset};

    my %quota_args;
    $pkg->_hash_copy(
        source => \%opts,
        map    => {"initiator-name" => "initiator"},
        target => \%quota_args
    );
    %opts = %quota_args;

    $apiset->iscsi_security_delete(%opts);
    $Log->exit() if $may_exit;
} ## end sub _delete_7mode_cli

=head2 default

   NACL::C::VserverIscsiSecurity->default(
                                command_interface => $ci,
                                vserver => $vserver,
                                ...);

(Class method)
This method is used to configure the default authentication settings.

Uses a CMode CLI/ZAPI, 7Mode CLI APISet.

Invokes "iscsi-initiator-set-default-auth" API for CMode ZAPI.

Invokes "iscsi security default" command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "vserver" => $vserver >>

(Required)

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: ZAPI will be invoked in the context of the vserver specified to this
option.

=item C<< "remove-outbound" => $string >>

Applicable for CMode CLI.

=item C<< "auth-type" => $string >>

Applicable for CMode CLI/ZAPI,7Mode CLI.

Maps to:

CM ZAPI: auth-type

7M CLI : -s <value> (Programmatic name: "auth-type")

=item C<< "radius" => $string >>

Applicable for CMode CLI/ZAPI,7Mode CLI.

Maps to:

CM ZAPI: radius

7M CLI : -f <value> (Programmatic name: "radius")

=item C<< "user-name" => $string >>

Applicable for CMode CLI/ZAPI,7Mode CLI.

Maps to:

CM ZAPI: user-name

7M CLI : -n <value> (Programmatic name: "user-name")

=item C<< "password" => $string >>

Applicable for CMode CLI/ZAPI,7Mode CLI.

Maps to:

CM ZAPI: password

7M CLI : -p <value> (Programmatic name: "password")

=item C<< "outbound-user-name" => $string >>

Applicable for CMode CLI/ZAPI,7Mode CLI.

Maps to:

CM ZAPI: outbound-user-name

7M CLI : -m <value> (Programmatic name: "outbound-user-name")

=item C<< "outpassword" => $string >>

Applicable for CMode CLI.

=item C<< "security" => $string >>

Applicable for CMode CLI.

=item C<< "outbound-password" => $string >>

Applicable for CMode CLI/ZAPI,7Mode CLI.

Maps to:

CM ZAPI: outbound-password

7M CLI : -o <value> (Programmatic name: "outbound-password")

=item C<< "outname" => $string >>

Applicable for CMode CLI.

=item C<< "fetch" => $string >>

Applicable for CMode CLI.

=item C<< "inname" => $string >>

Applicable for CMode CLI.

=item C<< "inpassword" => $string >>

Applicable for CMode CLI.

=item C<< 'clear-outbound' => $string >>

Applicable for CMode CLI.

=item C<< 'initiator' => $string >>

Applicable for CMode CLI.

=item C<< "extended_query" => $string >>

(Optional)

Applicable for CMode CLI.

=item C<< '*' >>

This method takes all other options supported by
"vserver iscsi security default" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub default {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( remove-outbound auth-type radius user-name password outbound-user-name
                    outpassword security outbound-password outname fetch inname inpassword

                    clear-outbound

                    initiator
                    )
            ),
        },
        cli_cmd => 'vserver iscsi security default',
    );
    my $vserver;
    $opts{vserver_operated_on} = \$vserver;
    my $ret = $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_default_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_default_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
            {   method    => '_default_7mode_cli',
                interface => 'CLI',
                set       => '7Mode',
            },
        ],
        frontend => 'NACL::C::VserverIscsiSecurity::default',
    );

    $Log->exit() if $may_exit;
    return $ret;
} ## end sub default

sub _default_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_base_backend_cmode_cli(@args,
        api     => 'vserver_iscsi_security_default',
    );
    $Log->exit() if $may_exit;
} ## end sub _default_cmode_cli

sub _default_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg, %opts) = @_;
    my $apiset = $opts{apiset};

    my $allowed_cmode_options = [
        qw( auth-type radius user-name
            outbound-user-name )
    ];
    my $api;
    if ($pkg->_check_for_remove_outbound(%opts)){
        delete $opts{'auth-type'} if (defined $opts{'auth-type'});
        $opts{'initiator'} = 'default';
        $allowed_cmode_options = [qw(initiator)];
        $api = 'iscsi_initiator_modify_chap_params';
    }
    else {
        delete $opts{'clear-outbound'} if (defined $opts{'clear-outbound'});
        delete $opts{'remove-outbound'} if (defined $opts{'remove-outbound'});
        $api = 'iscsi_initiator_set_default_auth';
    }

    $pkg->_base_backend_cmode_zapi(%opts,
            api  => $api,
            copy => $allowed_cmode_options,
            map  => {
                        'password' => 'passphrase',
                        'outbound-password' => 'outbound-passphrase'
                    },
    );

    $Log->exit() if $may_exit;
} ## end sub _default_cmode_zapi

sub _default_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );

    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};

    my $allowed_7mode_options = [
        qw( auth-type password outbound-user-name radius
            outbound-password user-name
            )
    ];
    my %quota_args;
    $pkg->_hash_copy(
        source => \%opts,
        copy   => $allowed_7mode_options,
        target => \%quota_args,
    );
    %opts = %quota_args;
    if ($quota_args{'auth-type'} eq "CHAP") {
        if ($quota_args{'radius'} eq "RADIUS") {
            delete $opts{'user-name'};
            delete $opts{'password'};
            delete $opts{'outbound-password'};
            delete $opts{'outbound-user-name'};
        } else {
            delete $opts{'radius'};
        }
    } else {
        delete $opts{'radius'};
        delete $opts{'user-name'};
        delete $opts{'password'};
        delete $opts{'outbound-password'};
        delete $opts{'outbound-user-name'};
    } ## end else [ if ( $quota_args{'auth-type'...
    $apiset->iscsi_security_default(%opts);
    $Log->exit() if $may_exit;
} ## end sub _default_7mode_cli


__PACKAGE__->_load_event_class();

sub _primary_keys_validate_spec {
    return (
        'initiator'      => {type => SCALAR | UNDEF, optional => 1},
        'initiator-name' => {type => SCALAR | UNDEF, optional => 1},
        'vserver'        => {type => SCALAR | UNDEF, optional => 1},
    );
} ## end sub _primary_keys_validate_spec

## As per burt956401
## If user has provided auth-type AND clear/remove-outbound
## and auth-type is CHAP OR not defined, just drop the chap parm and call iscsi-initiator-modify-chap-parms
## (use initiator=default)
sub _check_for_remove_outbound {
    my ($pkg, %opts) = @_;
    my $result = 0;
    my $remove_outbound = 'false';

    $remove_outbound = 'true' if ((defined $opts{'clear-outbound'} && $opts{'clear-outbound'} =~ /^true$/i) ||(   defined $opts{'remove-outbound'} && $opts{'remove-outbound'} =~ /^true$/i));
    if (defined $opts{'auth-type'} ){
        if ($opts{'auth-type'} =~ /^CHAP$/i && $remove_outbound eq 'true'){
            $result = 1;
        }
    }elsif($remove_outbound  eq 'true'){
         $result = 1;
    }
    return $result;
} ## end sub _check_for_remove_outbound
1;
