# $Id$

#
# Copyright (c) 2001-2012 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary VserverServicesDns Component Module
## @author dl-nacl-dev@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::C::VserverServicesDns

=head1 DESCRIPTION

C<NACL::C::VserverServicesDns> is a derived class of
L<NACL::C::Component::ONTAP|lib-NACL-C-Component-ONTAP-pm>. It represents access
to an ONTAP VserverServicesDns.  A related class is
L<NACL::CS::VserverServicesDns|lib-NACL-CS-VserverServicesDns-pm>, which represents
the past state of an ONTAP VserverServicesDns.

=head1 ATTRIBUTES

=head2 command_interface

See L<NACL::C::Component::command_interface|lib-NACL-C-Component-pm/command_interface>

=head2 vserver

=cut

package NACL::C::VserverServicesDns;

use strict;
use warnings;

use NACL::CS::VserverServicesDns;
use NACL::ComponentUtils qw (_optional_scalars);

use base 'NACL::C::Component::ONTAP';
use Params::Validate qw(validate validate_with SCALAR ARRAYREF UNDEF);
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Data::Dumper;

use NACL::C::Exceptions::VserverServicesDns::AlreadyExists ();
use NACL::C::Exceptions::VserverServicesDns::DoesNotExist  ();
use NACL::C::Exceptions::VserverServicesDns::DnsServerException ();
use NATE::BaseException qw(:try);
use NATE::Exceptions::Argument qw(:try);

=head1 METHODS

=head2 new

  my $VserverServicesDns = NACL::C::VserverServicesDns->new(command_interface=>$ci,
                                   vserver =>$vserver,
                                  );

Return a new VserverServicesDns component object with the given attributes.

=cut

use Class::MethodMaker [scalar => 'vserver',];

sub init {
    $Log->enter() if $may_enter;
    my ($self, @args) = @_;
    $self->SUPER::init(@args);
    $self->_vserver_context_init(@args);
    $Log->exit() if $may_exit;
}

=head2 find

    my $VserverServicesDns  = NACL::C::VserverServicesDns->find(command_interface => $ci, ...);
    my @VserverServicesDnss = NACL::C::VserverServicesDns->find(command_interface => $ci, ...);

(Class method)
A constructor of sorts, which discovers which elements are present (by
interacting with the given command interface).  In an array context it
creates a component object for each present element and returns them.
In a scalar context it creates a component object for just one
element.

This method is generally not implemented in individual components. The base
class (L<NACL::C::Component::ONTAP|lib-NACL-C-Component-ONTAP-pm> provides an implementation.
See <L<NACL::C::Component::find|lib-NACL-C-Component-pm/find> for a more
detailed description along with explanation of all the options accepted.

=head2 state

    my $VserverServicesDns_state = $NACL::C::VserverServicesDns->state();

(Instance method)
Fetch the current state of the element associated
with this component, and return a Component State object corresponding
to this component (calling this on NACL::C::VserverServicesDns would return
a NACL::CS::VserverServicesDns object)

This method is generally not implemented in individual components. The base
class (L<NACL::C::Component::ONTAP|lib-NACL-C-Component-ONTAP-pm> provides an implementation.
See <L<NACL::C::Component::state|lib-NACL-C-Component-pm/state> for a more
detailed description along with explanation of all the options accepted.

=cut

=head2 create

    my $VserverServicesDns_object    = NACL::C::VserverServicesDns->create(
                                command_interface => $ci,
                                 vserver => $vserver,
                                  ...);

(Class method)
This method is used to create a new DNS table entry.
Uses a CMode CLI or CMode ZAPI APISet.
Invokes "dns-add" API for 7Mode CLI.
Invokes "net-dns-create" API for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "vserver" => $vserver >>

(Required)
Applicable for CMode CLI/ZAPI
Maps to the same option name for CMode ZAPI

=item C<< "timeout" => $string >>

Applicable for CMode CLI/ZAPI
Maps to the same option name for CMode ZAPI

=item C<< "attempts" => $string >>

Applicable for CMode CLI/ZAPI
Maps to the same option name for CMode ZAPI

=item C<< "state" => $string >>

Applicable for CMode CLI/ZAPI
This maps to dns-state in CMode ZAPI

=item C<< "domains" => [ $domain1, $domain2, ... ] >>

Applicable for CMode CLI/ZAPI
Maps to the same option name for CMode ZAPI

Applicable only for 7Mode CLI
Maps to:
    7Mode CLI: domainname

=item C<< "name-servers" => [ $name-server1, $name-server2, ... ] >>

Applicable for CMode CLI/ZAPI
Maps to the same option name for CMode ZAPI

Maps to:
    7Mode CLI: ip-address

=item C<< "extended_query" => $string >>

Applicable for CMode CLI

=item Exceptions

=over

=item C<NACL::C::Exceptions::VserverServicesDns::AlreadyExists>

This type of exception gets thrown if an attempt is made to create a
vserver services dns that already exists.

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=item C<< '*' >>

This method takes all other options supported by
"vserver services dns create" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=back

=cut

sub create {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my $ret;
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            "domains"      => {type => ARRAYREF},
            "name-servers" => {type => ARRAYREF},
            _optional_scalars(
                qw( timeout attempts state domainname ip-address
                    skip-config-validation
                    )
            ),
        },
        cli_cmd => 'vserver services dns create',
    );
    my $vserver;
    $opts{vserver_operated_on} = \$vserver;
    try {
        $pkg->call_on_apiset(
            %opts,
            choices => [
                {   method    => '_create_cmode_cli',
                    interface => 'CLI',
                    set       => 'CMode',
                },
                {   method    => '_create_cmode_zapi',
                    interface => 'ZAPI',
                    set       => 'CMode',
                },
                {   method    => '_create_7mode_cli',
                    interface => 'CLI',
                    set       => '7Mode',
                },
            ],
            frontend => 'NACL::C::VserverServicesDns::create',
        );
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = $_[0];
	foreach my $type ( qw(AlreadyExists DnsServerException) ) {
            my $fulltype = "NACL::C::Exceptions::VserverServicesDns::$type";
            $fulltype->detect_convert_and_throw(exception => $exception);
	}
        $Log->exit() if $may_exit;
        $exception->throw();
    };

    my $obj = $pkg->new(
        command_interface => $opts{command_interface},
        vserver           => $opts{vserver},
    );
    $Log->exit() if $may_exit;
    return $obj;
} ## end sub create

sub _create_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_base_backend_cmode_cli(@args,
        api     => 'vserver_services_dns_create',
    );
    $Log->exit() if $may_exit;
} ## end sub _create_cmode_cli

sub _create_7mode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    my %opts = validate_with(
        params      => \@args,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );

    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my %final_7mode_args;

    $pkg->_hash_copy(
        source => \%opts,
        map    => {
            'domains'      => 'domainname',
            'name-servers' => 'ip-address',
        },
        target => \%final_7mode_args,
    );

    $pkg->_api_call_log_debug_opts(
        api_opts => \%opts,
        api      => 'dns_add'
    );

    foreach my $array_element (qw(domainname ip-address)) {
        if (defined $final_7mode_args{$array_element}) {
            $final_7mode_args{$array_element} = join ',',
                @{$final_7mode_args{$array_element}};
        }
    } ## end foreach my $array_element qw(domainname ip-address)
    $apiset->dns_add(%final_7mode_args);
    $Log->exit() if $may_exit;
    return $pkg->new(
        command_interface => $command_interface,
        vserver           => undef,
    );
} ## end sub _create_7mode_cli

sub _create_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;

    my $allowed_cmode_options = [
        qw( domains timeout name-servers attempts
            )
    ];

    $pkg->_base_backend_cmode_zapi(@args,
        api    => 'net_dns_create',
        copy   => $allowed_cmode_options,
        map    => {"state" => "dns-state"},
    );
    $Log->exit() if $may_exit;
} ## end sub _create_cmode_zapi

=head2 modify

   NACL::C::VserverServicesDns->modify(
                                command_interface => $ci,
                                 vserver => $vserver,
                                  ...);

                                 or

    $VserverServicesDns_object->modify(...);

(Class or instance method)
This method is used to change a DNS table entry.
Uses a CMode CLI or CMode ZAPI APISet.
Invokes "net-dns-modify" API for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "vserver" => $vserver >>

(Required for class method, Not Applicable for instance method)
Applicable for CMode CLI/ZAPI
Maps to the same option name for CMode ZAPI

=item C<< "timeout" => $string >>

Applicable for CMode CLI/ZAPI
Maps to the same option name for CMode ZAPI

=item C<< "attempts" => $string >>

Applicable for CMode CLI/ZAPI
Maps to the same option name for CMode ZAPI

=item C<< "state" => $string >>

Applicable for CMode CLI/ZAPI
This maps to dns-state in CMode ZAPI

=item C<< "domains" => [ $domain1, $domain2, ... ] >>

Applicable for CMode CLI/ZAPI
Maps to the same option name for CMode ZAPI

=item C<< "name-servers" => [ $name-server1, $name-server2, ... ] >>

Applicable for CMode CLI/ZAPI
Maps to the same option name for CMode ZAPI

=item C<< "extended_query" => $string >>

Applicable for CMode CLI

=item C<< '*' >>

This method takes all other options supported by
"vserver services dns modify" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub modify {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            "domains"      => {type => ARRAYREF, optional => 1},
            "name-servers" => {type => ARRAYREF, optional => 1},
            _optional_scalars(
                qw( timeout attempts state
                    )
            ),
        },
        cli_cmd => 'vserver services dns modify',
    );
    my $vserver;
    $opts{vserver_operated_on} = \$vserver;
    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_modify_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_modify_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::VserverServicesDns::modify',
    );

    $Log->exit() if $may_exit;
    return $ret;
} ## end sub modify

sub _modify_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@args,
        api     => 'vserver_services_dns_modify',
    );
    $Log->exit() if $may_exit;
} ## end sub _modify_cmode_cli

sub _modify_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my $allowed_cmode_options = [
        qw( domains timeout name-servers attempts
            extended_query )
    ];

    $pkg_or_obj->_base_backend_cmode_zapi(@args,
        api    => 'net_dns_modify',
        copy   => $allowed_cmode_options,
        map    => {"state" => "dns-state"},
    );
    $Log->exit() if $may_exit;
} ## end sub _modify_cmode_zapi

=head2 delete

   NACL::C::VserverServicesDns->delete(
                                command_interface => $ci,
                                 vserver => $vserver,
                                  ...);

                                 or

    $VserverServicesDns_object->delete(...);

(Class or instance method)
This method is used to remove a DNS table entry.
Uses a CMode/7Mode CLI or CMode ZAPI APISet.

Invokes "dns-delete" API for 7Mode CLI.
Invokes "net-dns-destroy" API for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "vserver" => $vserver >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI
Maps to the same option name for CMode ZAPI

=item C<< "domains" => [ $domain1, $domain2, ... ] >>

Applicable only for 7Mode CLI
Maps to:
    7Mode CLI: domainname

=item C<< "name-servers" => [ $name-server1, $name-server2, ... ] >>

Applicable only for 7Mode CLI
Maps to:
    7Mode CLI: ip-address

=item C<< "extended_query" => $string >>

Applicable for CMode CLI

=over

=item Exceptions

=over

=item C<NACL::C::Exceptions::VserverServicesDns::DoesNotExist>

This type of exception gets thrown if an attempt is made to delete
vserver services dns which does not exists.

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=item C<< '*' >>

This method takes all other options supported by
"vserver services dns delete" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=back

=back

=cut

sub delete {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my $ret;
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            "domains"      => {type => ARRAYREF, optional => 1},
            "name-servers" => {type => ARRAYREF, optional => 1},
            _optional_scalars(qw(domainname ip-address)),
        },
        cli_cmd => 'vserver services dns delete',
    );
    my $vserver;
    $opts{vserver_operated_on} = \$vserver;
    try {
        $ret = $pkg_or_obj->call_on_apiset(
            %opts,
            choices => [
                {   method    => '_delete_cmode_cli',
                    interface => 'CLI',
                    set       => 'CMode',
                },
                {   method    => '_delete_cmode_zapi',
                    interface => 'ZAPI',
                    set       => 'CMode',
                },
                {   method    => '_delete_7mode_cli',
                    interface => 'CLI',
                    set       => '7Mode',
                },
            ],
            frontend => 'NACL::C::VserverServicesDns::delete'
        );
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = $_[0];
        NACL::C::Exceptions::VserverServicesDns::DoesNotExist
            ->detect_convert_and_throw(exception => $exception);
        $Log->exit() if $may_exit;
        $exception->throw();
    };
    $Log->exit() if $may_exit;
    return $ret;
} ## end sub delete

sub _delete_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@args,
        api     => 'vserver_services_dns_delete',
    );
    $Log->exit() if $may_exit;
} ## end sub _delete_cmode_cli

sub _delete_7mode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    my %opts = validate_with(
        params      => \@args,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );

    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my %final_7mode_args;

    $pkg->_hash_copy(
        source => \%opts,
        map    => {
            'domains'      => 'domainname',
            'name-servers' => 'ip-address',
        },
        target => \%final_7mode_args,
    );

    $pkg->_api_call_log_debug_opts(
        api_opts => \%opts,
        api      => 'dns_delete'
    );

    foreach my $array_element (qw(domainname ip-address)) {
        if (defined $final_7mode_args{$array_element}) {
            $final_7mode_args{$array_element} = join ',',
                @{$final_7mode_args{$array_element}};
        }
    } ## end foreach my $array_element qw(domainname ip-address)

    $apiset->dns_delete(%final_7mode_args);
    $Log->exit() if $may_exit;
    return $pkg->new(
        command_interface => $command_interface,
        vserver           => undef,
    );
} ## end sub _delete_7mode_cli

sub _delete_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_base_backend_cmode_zapi(@args,
        api => 'net_dns_destroy',
    );
    $Log->exit() if $may_exit;
} ## end sub _delete_cmode_zapi

=head2 show_statistics

    my $show_statistics_obj = NACL::C::VserverServicesDns->show_statistics(command_interface => $ci,
                          [ filter => { %filter } ,]
			  [ requested_fields => [ @req_fields ]);
						  
                    or
					
    my @show_statistics_objs = NACL::C::VserverServicesDns->show_statistics(command_interface => $ci,
                          [ filter => { %filter } ,]
			  [ requested_fields => [ @req_fields ]);

                    or
  
   my $show_statistics_obj = $VserverServicesDns_obj->show_statistics(); 

(Class or Instance method)

This method is a CS-like fetch method which accepts all the parameters of 
L<fetch|lib-NACL-CS-ComponentState-pm/fetch> (requested_fields, filter, allow_empty) 
and has all of the other properties of fetch methods (callable in scalar or list 
context, throws a NoElementsFound exception if no matching elements are found).
The requested fields and filter takes L<ATTRIBUTES|lib-NACL-CS-VserverServicesDnsShowStatistics-pm/ATTRIBUTES>
of NACL::CS::VserverServicesDnsShowStatistics as input. 
This method returns objects of the type L<NACL::CS::VserverServicesDnsShowStatistics|lib-NACL-CS-VserverServicesDnsShowStatistics-pm>.

=cut

sub show_statistics {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts = $pkg_or_obj->_handle_obj_cs_like_calls(@args);
    require NACL::CS::VserverServicesDnsShowStatistics;
    my @state_objs = NACL::CS::VserverServicesDnsShowStatistics->fetch(%opts);
    $Log->exit() if $may_exit;
    return wantarray ? @state_objs : $state_objs[0];
}


__PACKAGE__->_load_event_class();

sub _primary_keys_validate_spec {
    return ('vserver' => {type => SCALAR | UNDEF, optional => 1},);
}

1;
