#
# Copyright (c) 2001-2013 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary VserverServicesLdapClient Component Module
## @author Rupamjyoti.Baruah@netapp.com, dl-nacl-dev@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::C::VserverServicesLdapClient

=head1 DESCRIPTION

C<NACL::C::VserverServicesLdapClient> is a derived class of
L<NACL::C::Component::ONTAP|lib-NACL-C-Component-ONTAP-pm>.  It represents access
to an ONTAP Vserver Services Ldap Client.  A related class is
L<NACL::CS::VserverServicesLdapClient|lib-NACL-CS-VserverServicesLdapClient-pm>,
 which represents the past state of an ONTAP Vserver Services Ldap Client

=head1 ATTRIBUTES

=head2 command_interface

See L<NACL::C::Component::command_interface|lib-NACL-C-Component-pm/command_interface>

=head2 client-config

This parameter specifies the name of the LDAP client configuration, used to
identify it uniquely.

=cut

package NACL::C::VserverServicesLdapClient;

use strict;
use warnings;
use NACL::CS::VserverServicesLdapClient;
use NACL::ComponentUtils qw (_optional_scalars Dumper);
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use base 'NACL::C::Component::ONTAP';
use Params::Validate qw(validate validate_with SCALAR ARRAYREF UNDEF OBJECT);
use NACL::C::Options;
use NACL::C::_Version;
use NACL::C::Exceptions::VserverServicesLdapClient::AlreadyExists ();
use NACL::C::Exceptions::VserverServicesLdapClient::DoesNotExist  ();
use NATE::BaseException qw(:try);
use NATE::Exceptions::Argument qw(:try);

=head1 METHODS

=head2 new

  my $ldap_client = NACL::C::VserverServicesLdapClient->new(
                                                    command_interface=>$ci,
                                                    client-config =>$string,
                                                    );

Return a new Vserver Services LDAP Client component object with the given
attributes.

=cut

use Class::MethodMaker [
    scalar => 'client_config',
    scalar => 'vserver',
];

sub init {
    $Log->enter() if $may_enter;
    my ($self, @args) = @_;
    $self->SUPER::init(@args);
    $self->_vserver_context_init(@args);
    $Log->exit() if $may_exit;
}

=head2 create

  NACL::C::VserverServicesLdapClient->create( command_interface=>$ci,
                                              'client-config'=>$string,
                                              ...);

(Class method) Request creation of a new Vserver LDAP Clientname mapping
element, and  return a new  Vserver LDAP Client component object that refers
to it.  Does not wait for Vserver LDAP Client creation to complete.

Uses a CMode-CLI/ZAPI APISet.
  Invokes 'ldap-client-create' API for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< client-config=>$string >>

(Required) This parameter specifies the name of the LDAP client configuration.
  Applicable for CMode CLI/ZAPI.
  Maps to:
    CM ZAPI: ldap-client-config

=item C<< servers=>$string >>

(Required) This specifies the LDAP Server List.
  Applicable for CMode CLI/ZAPI.
  Maps to:
    CM ZAPI: servers

=item C<< ad-domain=>$string >>

(Required) This specifies the Active Directory Domain.
  Applicable for CMode CLI/ZAPI.
  Maps to:
    CM ZAPI: ad-domain

=item C<< preferred-ad-servers=> [$preferred-ad-server1, $preferred-ad-server2, ....]  >>
  Applicable for CMode CLI/ZAPI.
  Maps to:
    CM ZAPI: preferred-ad-servers

(Optional) This specifies the Preferred Active Directory Servers.

=item C<< bind-as-cifs-server=>$boolean >>

(Optional) This specifies whether to bind using the Virtual Server's CIFS
Credentials or not.
  Applicable for CMode CLI/ZAPI.
  Maps to:
    CM ZAPI: bind-as-cifs-server

=item C<< schema=>$string >>

(Required) This parameter specifies the schema template.
  Applicable for CMode CLI/ZAPI.
  Maps to:
    CM ZAPI: schema

=item C<< port=>$number >>

(Optional) This optional parameter specifies the LDAP server port
(default: 389).
  Applicable for CMode CLI/ZAPI.
  Maps to:
    CM ZAPI: tcp-port

=item C<< query-timeout=>$number >>

(Optional) This optional parameter specifies the query timeout in seconds
(default: 3).
  Applicable only for CMode CLI/ZAPI.
  Maps to:
  CM ZAPI: query-timeout

=item C<< min-bind-level=>$string >>

(Optional) This optional parameter specifies Minimum Bind Authentication Level.
  Applicable for CMode CLI/ZAPI.
  Maps to:
    CM ZAPI: min-bind-level

=item C<< bind-dn=>$string >>

(Optional) This optional parameter specifies the bind distinguished name (DN).
  Applicable for CMode CLI/ZAPI.
  Maps to:
    CM ZAPI: bind-dn

=item C<< bind-password=>$string >>

(Optional) This optional parameter specifies the bind password.
  Applicable for CMode CLI/ZAPI.
  Maps to:
    CM ZAPI: bind-password

=item C<< base-dn=>$string >>

(Optional) This optional parameter specifies the default base DN for all
searches, including user, group, and netgroup searches.
  Applicable for CMode CLI/ZAPI.
  Maps to:
    CM ZAPI: base-dn

=item C<< base-scope=>$string >>

(Optional) This optional parameter specifies the Base Search Scope.
  Applicable for CMode CLI/ZAPI.
  Maps to:
    CM ZAPI: base-scope

=item C<< "vserver" => $string >>

  Applicable for CMode CLI/ZAPI.
  Maps to:
    CM ZAPI: ZAPI will be invoked in the context of the vserver specified to this
             option.

=item C<< "user-dn" => $string >>

  Applicable for CMode CLI/ZAPI.
  Maps to:
    CM ZAPI: user-dn

=item C<< "user-scope" => $string >>

  Applicable for CMode CLI/ZAPI.
  Maps to:
    CM ZAPI: user-scope

=item C<< "group-dn" => $string >>

  Applicable for CMode CLI/ZAPI.
  Maps to:
    CM ZAPI: group-dn

=item C<< "group-scope" => $string >>

  Applicable for CMode CLI/ZAPI.
  Maps to:
    CM ZAPI: group-scope

=item C<< "netgroup-dn" => $string >>

  Applicable for CMode CLI/ZAPI.
  Maps to:
    CM ZAPI: netgroup-dn

=item C<< "netgroup-scope" => $string >>

  Applicable for CMode CLI/ZAPI.
  Maps to:
    CM ZAPI: netgroup-scope

=item C<< 'extended_query' => $string >>

(Optional)
  Applicable only for CMode CLI.

=over

=item Exceptions

=over

=item C<NACL::C::Exceptions::VserverServicesLdapClient::AlreadyExists>

This type of exception gets thrown if an attempt is made to create a
vserver services ldap client that already exists.

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=item C<< 'use-start-tls' => $string >>

Use start-tls Over LDAP Connections possible value(s) are, true,false
Applicable for CMode CLI/ZAPI

Maps to:

CM ZAPI: use-start-tls

=item C<< '*' >>

This method takes all other options supported by
"vserver services ldap client create" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=back

=back

=cut

sub create {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            servers => {type => ARRAYREF, optional => 1},
            'preferred-ad-servers' =>
                {type => SCALAR | ARRAYREF, optional => 1},
            _optional_scalars(
                qw(client-config ad-domain schema
                    bind-as-cifs-server port query-timeout min-bind-level bind-dn
                    bind-password base-dn base-scope user-dn user-scope group-dn
                    group-scope netgroup-dn netgroup-scope asymmetric-name-mapping
                    return-record is-netgroup-byhost-enabled use-start-tls
                    netgroup-byhost-dn
                    )
            )
        },
        cli_cmd => 'vserver services ldap client create',
    );
    $pkg->_scalar_to_array_deprication_warning(
        params  => ['preferred-ad-servers'],
        options => \%opts
    );
    my $ret;
    try {
      $ret =  $pkg->call_on_apiset(
              %opts,
              choices => [
                {   method    => '_create_cmode_cli',
                    interface => 'CLI',
                    set       => 'CMode',
                },
                {   method    => '_create_cmode_zapi',
                    interface => 'ZAPI',
                    set       => 'CMode',
                },
            ],
            frontend => 'NACL::C::VserverServicesLdapClient::create',
        );
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = $_[0];
        NACL::C::Exceptions::VserverServicesLdapClient::AlreadyExists
            ->detect_convert_and_throw(exception => $exception);
        $Log->exit() if $may_exit;
        $exception->throw();
    };

    $Log->exit() if $may_exit;
    return $ret;
} ## end sub create

sub _create_cmode_cli {
    my ($pkg, %opts) = @_;
    $pkg->_array_to_scalar_deprication_warning(
        params  => ['servers'],
        options => \%opts,
    );

    my $vserver;
    $opts{vserver_operated_on} = \$vserver;

    $pkg->_base_backend_cmode_cli(%opts,
        api => 'vserver_services_ldap_client_create');

    my %vserver_opt;
    $vserver_opt{vserver} = $vserver if (defined $vserver);

    my $obj = $pkg->new(
        command_interface => $opts{command_interface},
        client_config     => $opts{'client-config'},
        %vserver_opt
    );

    $Log->exit() if $may_exit;
    return $obj;
} ## end sub _create_cmode_cli

sub _create_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg, %opts) = @_;

    my $vserver;
    $opts{vserver_operated_on} = \$vserver;

    # Defaulting to mimic CLI behaviour (NACL burt 815665; ONTAP burt 815662)
    $opts{schema} ||= 'RFC-2307';

    $pkg->_base_backend_cmode_zapi(
        %opts,
        api  => 'ldap-client-create',
        copy => [
            qw( servers ad-domain schema
                preferred-ad-servers bind-as-cifs-server min-bind-level
                bind-dn bind-password base-dn base-scope user-dn user-scope group-dn
                group-scope netgroup-dn netgroup-scope query-timeout return-record use-start-tls
                is-netgroup-byhost-enabled netgroup-byhost-dn )
        ],
        map => {
            "client-config" => "ldap-client-config",
            "port"          => "tcp-port"
        },
        frontend_method => 'create'
    );

    my %vserver_opt;
    $vserver_opt{vserver} = $vserver if (defined $vserver);

    my $obj = $pkg->new(
        command_interface => $opts{command_interface},
        client_config     => $opts{'client-config'},
        %vserver_opt,
    );

    $Log->exit() if $may_exit;
    return $obj;
} ## end sub _create_cmode_zapi

=head2 delete

  NACL::C::VserverServicesLdapClient->delete(command_interface=>$ci,
                                             client-config=>$string);

                            or

  $ldap_client_object->delete(...);

(Class or instance method) This method deletes an vserver name mapping element.

Uses a CMode-CLI/ZAPI APISet.
  Invokes 'ldap-client-delete' API for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< client-config=>$string >>

(Required for class method, Not Applicable for instance method)
(Required) This parameter specifies the name of the LDAP client configuration.
  Applicable for CMode CLI/ZAPI.
  Maps to:
    CM ZAPI: ldap-client-config

=item C<< "vserver" => $string >>

  Applicable for CMode CLI/ZAPI.
  Maps to:
    CM ZAPI: ZAPI will be invoked in the context of the vserver specified to this
             option.

=item C<< 'extended_query' => $string >>

(Optional)
  Applicable only for CMode CLI.

=over

=item Exceptions

=over

=item C<NACL::C::Exceptions::VserverServicesLdapClient::DoesNotExist>

This type of exception gets thrown if an attempt is made to delete
vserver services ldap client which does not exists.

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=item C<< '*' >>

This method takes all other options supported by
"vserver services ldap client delete" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=back

=back

=cut

sub delete {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts = $pkg_or_obj->_common_validate_with(
        params => \@args,
        cli_cmd => 'vserver services ldap client delete',
    );
    my $vserver;
    $opts{vserver_operated_on} = \$vserver;
    my $ret;
    try {
        $ret = $pkg_or_obj->call_on_apiset(
                %opts,
                choices => [
                {   method    => '_delete_cmode_cli',
                    interface => 'CLI',
                    set       => 'CMode',
                },
                {   method    => '_delete_cmode_zapi',
                    interface => 'ZAPI',
                    set       => 'CMode',
                },
            ],
            frontend => 'NACL::C::VserverServicesLdapClient::delete',
        );
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = $_[0];
        NACL::C::Exceptions::VserverServicesLdapClient::DoesNotExist
            ->detect_convert_and_throw(exception => $exception);
        $Log->exit() if $may_exit;
        $exception->throw();
    };

    $Log->exit() if $may_exit;
    return $ret;
} ## end sub delete

sub _delete_cmode_cli {
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@args,
        api => 'vserver_services_ldap_client_delete',
    );
    $Log->exit() if $may_exit;
} ## end sub _delete_cmode_cli

sub _delete_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_base_backend_cmode_zapi(@args,
        api => 'ldap_client_delete',
        map => {"client-config" => "ldap-client-config"},
    );
    $Log->exit() if $may_exit;
} ## end sub _delete_cmode_zapi

=head2 modify
   NACL::C::VserverServicesLdapClient->modify(
                                command_interface => $ci,
                                  ...);
(Class method)
This method is used to modify an LDAP client configuration.

Uses a CMode CLI/ZAPI or 7Mode CLI APISet.
  Invokes 'ldap-client-modify' API for CMode ZAPI.
  Invokes 'options' command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>
(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>
(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>
(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "user-scope" => $string >>

  Applicable for CMode CLI/ZAPI.
  Maps to:
    CM ZAPI: user-scope

=item C<< "base-dn" => $string >>

  Applicable for CMode CLI/ZAPI.
  Maps to:
    CM ZAPI: base-dn

=item C<< "schema" => $string >>
  Applicable for CMode CLI/ZAPI.
  Maps to:
    CM ZAPI: schema

=item C<< "user-dn" => $string >>

  Applicable for CMode CLI/ZAPI.
  Maps to:
    CM ZAPI: user-dn

=item C<< "asymmetric-name-mapping" => $string >>

  Applicable only for CMode CLI.

=item C<< "bind-as-cifs-server" => $string >>

  Applicable for CMode CLI/ZAPI.
  Maps to:
    CM ZAPI: bind-as-cifs-server

=item C<< "base-scope" => $string >>

  Applicable for CMode CLI/ZAPI.
  Maps to:
    CM ZAPI: base-scope

=item C<< "bind-dn" => $string >>

  Applicable for CMode CLI/ZAPI,7Mode CLI.
  Maps to:
    CM ZAPI: bind-dn
    7M CLI : ldap.usermap.base

=item C<< "group-dn" => $string >>

  Applicable for CMode CLI/ZAPI.
  Maps to:
    CM ZAPI: group-dn

=item C<< "bind-password" => $string >>

  Applicable for CMode CLI/ZAPI,7Mode CLI.
  Maps to:
    CM ZAPI: bind-password
    7M CLI : ldap.base.passwd

=item C<< "min-bind-level" => $string >>

  Applicable for CMode CLI/ZAPI,7Mode CLI.
  Maps to:
    CM ZAPI: min-bind-level
    7M CLI : ldap.minimum bind level

=item C<< "netgroup-dn" => $string >>

  Applicable for CMode CLI/ZAPI.
  Maps to:
    CM ZAPI: netgroup-dn

=item C<< "ad-domain" => $string >>

  Applicable for CMode CLI/ZAPI,7Mode CLI.
  Maps to:
    CM ZAPI: ad-domain
    7M CLI : ldap.ADdomain

=item C<< "port" => $string >>

  Applicable for CMode CLI/ZAPI,7Mode CLI.
  Maps to:
    CM ZAPI: tcp-port
    7M CLI : ldap.port

=item C<< "group-scope" => $string >>

  Applicable for CMode CLI/ZAPI.
  Maps to:
    CM ZAPI: group-scope

=item C<< "netgroup-scope" => $string >>

  Applicable for CMode CLI/ZAPI.
  Maps to:
    CM ZAPI: netgroup-scope

=item C<< "client-config" => $string >>

  Applicable for CMode CLI/ZAPI,7Mode CLI.
  Maps to:
    CM ZAPI: ldap-client-config
    7M CLI : ldap.name

=item C<< "query-timeout" => $string >>

Applicable for CMode CLI/ZAPI, 7Mode CLI.
Maps to:
    CM ZAPI : query-timeout
    7M CLI : ldap.timeout

=item C<< "preferred-ad-servers" => [ $preferred-ad-server1, $preferred-ad-server2, ... ] >>

  Applicable for CMode CLI/ZAPI,7Mode CLI.
  Maps to:
    CM ZAPI: preferred-ad-servers
    7M CLI : ldap.servers.preferred

=item C<< "servers" => [ $server1, $server2, ... ] >>

Applicable for CMode CLI, 7Mode CLI.
Maps to:
    7M CLI : ldap.servers

=item C<< "vserver" => $string >>

  Applicable for CMode CLI/ZAPI.
  Maps to:
    CM ZAPI: ZAPI will be invoked in the context of the vserver specified to this
             option.

=item C<< 'extended_query' => $string >>

(Optional)
  Applicable only for CMode CLI.

=item C<< 'use-start-tls' => $string >>

Use start-tls Over LDAP Connections possible value(s) are, true,false
Applicable for CMode CLI/ZAPI

Maps to:

CM ZAPI : use-start-tls

=item C<< '*' >>

This method takes all other options supported by
"vserver services ldap client modify" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub modify {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            "preferred-ad-servers" => {type => ARRAYREF, optional => 1},
            "servers"              => {type => ARRAYREF, optional => 1},
            _optional_scalars(
                qw( user-scope base-dn schema user-dn asymmetric-name-mapping
                    bind-as-cifs-server base-scope bind-dn group-dn bind-password
                    min-bind-level netgroup-dn ad-domain port group-scope netgroup-scope
                    client-config query-timeout is-netgroup-byhost-enabled use-start-tls 
                    netgroup-byhost-dn
                    )
            ),
        },
        cli_cmd => 'vserver services ldap client modify',
    );

   my $ret = $pkg->call_on_apiset(
             %opts,
             choices => [
             {   method    => '_modify_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
             },
             {   method    => '_modify_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
             },
             {   method    => '_modify_7mode_cli',
                interface => 'CLI',
                set       => '7Mode',
             },
          ],
        frontend => 'NACL::C::VserverServicesLdapClient::modify',
    );
    $Log->exit() if $may_exit;
    return $ret;
} ## end sub modify

sub _modify_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, %opts) = @_;
    my $vserver;
    $opts{vserver_operated_on} = \$vserver;

    $pkg->_base_backend_cmode_cli(%opts,
        api => 'vserver_services_ldap_client_modify');
    $Log->exit() if $may_exit;
} ## end sub _modify_cmode_cli

sub _modify_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg, %opts) = @_;
    my $vserver;
    $opts{vserver_operated_on} = \$vserver;

    $pkg->_base_backend_cmode_zapi(
        %opts,
        api  => 'ldap-client-modify',
        copy => [
            qw( user-scope base-dn schema user-dn preferred-ad-servers
                bind-as-cifs-server base-scope bind-dn group-dn
                bind-password min-bind-level netgroup-dn ad-domain
                group-scope servers netgroup-scope query-timeout use-start-tls
                is-netgroup-byhost-enabled netgroup-byhost-dn )
        ],
        map => {
            "client-config" => "ldap-client-config",
            "port"          => "tcp-port"
        },
        frontend_method => 'modify'
    );
    $Log->exit() if $may_exit;
} ## end sub _modify_cmode_zapi

sub _modify_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );

    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};

    my %args_7mode;
    $pkg->_hash_copy(
        source => \%opts,
        map    => {
            'client-config'        => 'ldap.name',
            'servers'              => 'ldap.servers',
            'ad-domain'            => 'ldap.ADdomain',
            'preferred-ad-servers' => 'ldap.servers.preferred',
            'port'                 => 'ldap.port',
            'query-timeout'        => 'ldap.timeout',
            'min-bind-level'       => 'ldap.minimum_bind_level',
            'bind-dn'              => 'ldap.usermap.base',
            'bind-password'        => 'ldap.base.passwd',
        },
        target => \%args_7mode,
    );
    if ($opts{'base-dn'}) {
        my $base_scope = "";
        $base_scope = uc($opts{'base-scope'});
        $args_7mode{'ldap.base'} = $opts{'base-dn'} . ":" . $base_scope;
    }
    if ($opts{'group-dn'}) {
        my $group_scope = "";
        $group_scope = uc($opts{'group-scope'});
        $args_7mode{'ldap.base.group'}
            = $opts{'group-dn'} . ":" . $group_scope;
    } ## end if ( $opts{'group-dn'})
    if ($opts{'netgroup-dn'}) {
        my $netgroup_scope = "";
        $netgroup_scope = uc($opts{'netgroup-scope'});
        $args_7mode{'ldap.base.netgroup'}
            = $opts{'netgroup-dn'} . ":" . $netgroup_scope;
    } ## end if ( $opts{'netgroup-dn'...

    $pkg->_array_to_scalar_deprication_warning(
        params  => ['ldap.servers'],
        options => \%args_7mode
    );

    while (my ($key, $value) = each(%args_7mode)) {
        $apiset->options($key => $value);
    }
    $Log->exit() if $may_exit;
} ## end sub _modify_7mode_cli

=head2 modify_bind_password

    NACL::C::VserverServicesLdapClient->modify_bind_password(
                                command_interface => $ci,
                                'vserver' => $vserver,
                                'client-config' => $client_config,
                                ...);

                                 or

    $VserverServicesLdapClient_obj->modify_bind_password(...);

(Class or instance method)
This method is used to modify Bind Password of an LDAP client configuration.

Uses a CMode CLI APISet.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'vserver' => $vserver >>

(Required for class method, Not Applicable for instance method)
  Applicable only for CMode CLI.

=item C<< 'client-config' => $client_config >>

(Required for class method, Not Applicable for instance method)
  Applicable only for CMode CLI.

=item C<< 'method-timeout' => $time_in_seconds >>

(Optional)
The default NATE timeout is 60 seconds. If the command might take more
time to complete, this option can be used to specify a larger timeout value.
The value should be provided in seconds.

=item C<< 'extended_query' => $string >>

(Optional)
  Applicable only for CMode CLI.

=item C<< '*' >>

This method takes all other options supported by
"vserver services ldap client modify-bind-password" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from 
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub modify_bind_password {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params => \@args,
        cli_cmd    => 'vserver services ldap client modify bind password',
    );
    $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_modify_bind_password_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
        ],
        frontend =>
            'NACL::C::VserverServicesLdapClient::modify_bind_password',
    );

    $Log->exit() if $may_exit;
} ## end sub modify_bind_password

sub _modify_bind_password_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    my $vserver = $opts{vserver};

    $pkg_or_obj->_base_backend_cmode_cli(
        %opts,
        api => 'vserver_services_ldap_client_modify_bind_password',
        vserver_operated_on => \$vserver,
    );

    $Log->exit() if $may_exit;
} ## end sub _modify_bind_password_cmode_cli


__PACKAGE__->_load_event_class();

sub _primary_keys_validate_spec {
    my $pkg_or_obj = shift;
    require NACL::C::Node;
    my %opts = validate(
        @_,
        {   command_interface => {
                type     => OBJECT,
                isa      => 'NACL::C::CommandInterface',
                optional => 1
            }
        }
    );
    my $command_interface = $opts{command_interface};

    my %default_spec = (
        'client-config' => {type => SCALAR},
        vserver         => {type => SCALAR | UNDEF, optional => 1}
    );

    if (defined $command_interface
        && $command_interface->mode() =~ /7Mode/i)
    {
        return (%default_spec);
    }

    if(( defined $command_interface )
        && ($command_interface->mode() =~ /CMode/i))
    {
        return (
            'client-config' => {type => SCALAR},
            vserver         => {type => SCALAR, optional => 1}
        );
    } else {
        return %default_spec;
    }
} ## end sub _primary_keys_validate_spec

1;
