#
# Copyright (c) 2001-2012 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary VserverServicesUnixGroup Component Module
## @author Dheeraj.K@netapp.com, dl-nacl-dev@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::C::VserverServicesUnixGroup

=head1 DESCRIPTION

C<NACL::C::VserverServicesUnixGroup> is a derived class of
L<NACL::C::Component::ONTAP|lib-NACL-C-Component-ONTAP-pm>.  It represents access
to an ONTAP Unix Group.  A related class is
L<NACL::CS::VserverServicesUnixGroup|lib-NACL-CS-UnixGroup-pm>, which represents
the past state of an ONTAP Unix Group.

=head1 ATTRIBUTES

=head2 command_interface

See L<NACL::C::Component::command_interface|lib-NACL-C-Component-pm/command_interface>

=head2 name

The name of the unix group, used to identify it uniquely.  (unix group
names are assumed to be unique within a vserver)

=head2 vserver

The name of the vserver or Node.

=cut

package NACL::C::VserverServicesUnixGroup;

use strict;
use warnings;
use NACL::CS::VserverServicesUnixGroup;
use NACL::ComponentUtils qw (_optional_scalars);
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use NACL::ComponentUtils qw(Dumper);

use base 'NACL::C::Component::ONTAP';
use Params::Validate qw(validate_with SCALAR UNDEF);
use NACL::C::Exceptions::VserverServicesUnixGroup::AlreadyExists qw(:try);
use NACL::C::Exceptions::VserverServicesUnixGroup::DoesNotExist;

=head1 METHODS

=head2 new

  my $unix_group = NACL::C::VserverServicesUnixGroup->new(command_interface=>$ci,
                                           name =>$unix_group_name);

Return a new Unix Group component object with the given attributes.

=cut

use Class::MethodMaker [
    scalar => 'vserver',
    scalar => 'name',
];

sub init {
    $Log->enter() if $may_enter;
    my ($self, @args) = @_;
    $self->SUPER::init(@args);
    $self->_vserver_context_init(@args);
    $Log->exit() if $may_exit;
}

sub isa {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, $kind) = @_;
    my $isa = $pkg_or_obj->_build_isa(
        kind  => $kind,
        alias => 'NACL::C::UnixGroup'
    );
    $Log->exit() if $may_exit;
    return $isa;
}

=head2 create

  my $unix_group = NACL::C::VserverServicesUnixGroup->create(command_interface=>$ci,
                                              name=>$unix_group_name,...);

(Class method) Request creation of a new unix group element, and  return a new
unix group component object that refers to it.  Does not wait for unix group
creation to complete.

Uses a CMode-CLI/ZAPI,7Mode CLI APISet.
  Invokes 'name-mapping-unix-group-create' API for CMode ZAPI.
  Invokes 'useradmin group add' command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< vserver=>$vserver_name >>

(Required) Name of vserver or Node.
  Applicable for CMode CLI/ZAPI.
  Maps to:
    CM ZAPI: ZAPI will be invoked in the context of the vserver specified to this
             option.



=item C<< name=>$unix_group_name >>

(Required) Name of unix group to create.
  Applicable for CMode CLI/ZAPI,7Mode CLI.
  Maps to:
    CMode ZAPI: group-name
    7Mode CLI:  group-name

=item C<< id=>$integer >>

(Required) Integer number to represent a unix group.
  Applicable to CMode CLI/ZAPI
  Maps to:
    CMode ZAPI: group-id

=item C<< extended_query => $string >>

(Optional)
Applicable only for CMode CLI

=item C<< '*' >>

This method takes all other options supported by
"vserver services unix-group create" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=item C<NACL::C::Exceptions::VserverServicesUnixGroup::AlreadyExists>

This type of exception is thrown when an attempt is made to create a
unix group when a group by that name or id already exist.

=back

=back

=cut

sub create {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {_optional_scalars(qw(id))},
        api             => 'vserver_services_unix_group_create',
    );
    my $ret_obj;
    try {
        $ret_obj = $pkg->call_on_apiset(
            %opts,
            choices => [
                {   method    => '_create_7mode_cli',
                    interface => 'CLI',
                    set       => '7Mode',
                },
                {   method    => '_create_cmode_cli',
                    interface => 'CLI',
                    set       => 'CMode',
                },
                {   method    => '_create_cmode_zapi',
                    interface => 'ZAPI',
                    set       => 'CMode',
                },

            ],
            frontend => 'NACL::C::VserverServicesUnixGroup::create',
        );
    } catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = shift;
        NACL::C::Exceptions::VserverServicesUnixGroup::AlreadyExists
            ->detect_convert_and_throw('exception' => $exception,);
        $Log->exit() if $may_exit;
        $exception->throw();
    };

    $Log->exit() if $may_exit;
    return $ret_obj;
} ## end sub create

sub _create_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, %opts) = @_;
    
    my $vserver = $opts{vserver};
    $pkg->_base_backend_cmode_cli(
        %opts, 
        api => 'vserver_services_unix_group_create',
        vserver_operated_on => \$vserver,
    );

    $Log->exit() if $may_exit;
    return $pkg->new(
        'vserver'         => $vserver,
        'name'            => $opts{name},
        command_interface => $opts{command_interface},
    );
} ## end sub _create_cmode_cli

sub _create_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg, %opts) = @_;
    
    my $vserver = $opts{vserver};
    $pkg->_base_backend_cmode_zapi(
        %opts,
        frontend_method => 'create',
        vserver_operated_on => \$vserver,
        api => 'name-mapping-unix-group-create',
        map => {
            "id"   => "group-id",
            "name" => "group-name",
        },
    );
    
    $Log->exit() if $may_exit;
    return $pkg->new(
        'vserver'         => $vserver,
        'name'            => $opts{'name'},
        command_interface => $opts{command_interface},
    );
} ## end sub _create_cmode_zapi

sub _create_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};

    my %seven_mode_args;
    $pkg->_hash_copy(
        source => \%opts,
        map    => {"name" => "group-name",},
        target => \%seven_mode_args,
    );
    %opts = %seven_mode_args;

    $apiset->useradmin_group_add(%opts);
    $Log->exit() if $may_exit;
    return $pkg->new(
        'name'            => $opts{'group-name'},
        command_interface => $command_interface
    );
} ## end sub _create_7mode_cli

=head2 delete

  NACL::C::VserverServicesUnixGroup->delete(command_interface=>$ci,
                             name=>$unix_group_name,...);

                            or

  $unix_group_object->delete(...);

(Class or instance method) This method deletes an Unix Group element.

Uses a CMode-CLI/ZAPI,7Mode CLI APISet.
  Invokes 'name-mapping-unix-group-destroy' API for CMode ZAPI.
  Invokes 'useradmin group delete' command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< name=>$unix_group_name >>

(Required for class method, Not Applicable for instance method)
Name of Unix Group.
  Applicable for CMode CLI/ZAPI, 7Mode CLI
  Maps to:
    CMode ZAPI: group-name
    7Mode CLI:  group-name

=item C<< vserver=>$vserver_name >>

(Required for class method, Not Applicable for instance method)
(Required) Name of vserver or Node.
  Applicable for CMode CLI/ZAPI.
  Maps to:
    CM ZAPI: ZAPI will be invoked in the context of the vserver specified to this
             option.

=item C<< extended_query => $string >>

(Optional)
Applicable only for CMode CLI

=item C<< '*' >>

This method takes all other options supported by
"vserver services unix-group delete" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=item C<NACL::C::Exceptions::VserverServicesUnixGroup::DoesNotExist>

This type of exception is thrown when an attempt is made to delete a
a unix user group which doesn't exist.

=back

=back

=cut

sub delete {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {},
        api             => 'vserver_services_unix_group_delete',
    );
    
    my $ret_obj;
    try {
        $ret_obj = $pkg_or_obj->call_on_apiset(
            %opts,
            choices => [
                {   method    => '_delete_cmode_cli',
                    interface => 'CLI',
                    set       => 'CMode',
                },
                {   method    => '_delete_cmode_zapi',
                    interface => 'ZAPI',
                    set       => 'CMode',
                },
                {   method    => '_delete_7mode_cli',
                    interface => 'CLI',
                    set       => '7Mode',
                },
            ],
            frontend => 'NACL::C::VserverServicesUnixGroup::delete',
        );
    } catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = shift;
        NACL::C::Exceptions::VserverServicesUnixGroup::DoesNotExist
            ->detect_convert_and_throw('exception' => $exception,);
        $Log->exit() if $may_exit;
        $exception->throw();
    };

    $Log->exit() if $may_exit;
    return $ret_obj;
} ## end sub delete

sub _delete_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, %opts) = @_;
    
    my $vserver = $opts{vserver};
    $pkg->_base_backend_cmode_cli(
        %opts, 
        api => 'vserver_services_unix_group_delete',
        vserver_operated_on => \$vserver,
    );

    $Log->exit() if $may_exit;
} ## end sub _delete_cmode_cli

sub _delete_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg,%opts) = @_;
     
    my $vserver = $opts{vserver};
    $pkg->_base_backend_cmode_zapi(
        %opts,
        frontend_method => 'delete',
        vserver_operated_on => \$vserver,
        api => 'name-mapping-unix-group-destroy',
        map => {
            "name" => "group-name",
        },
    );

    $Log->exit() if $may_exit;
} ## end sub _delete_cmode_zapi

sub _delete_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};

    my %seven_mode_args;
    $pkg->_hash_copy(
        source => \%opts,
        map    => {"name" => "group-name",},
        target => \%seven_mode_args,
    );
    %opts = %seven_mode_args;

    $apiset->useradmin_group_delete(%opts);
    $Log->exit() if $may_exit;
} ## end sub _delete_7mode_cli

=head2 adduser

  NACL::C::VserverServicesUnixGroup->adduser(command_interface=>$ci,
                             name=>$unix_group_name,...);

                            or

  $unix_group_object->adduser(...);

(Class or instance method) This method adds users to an existing Unix Group
element.

Uses a CMode-CLI/ZAPI APISet.
  Invokes 'name-mapping-unix-group-add-user' for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< name=>$unix_group_name >>

(Required for class method, Not Applicable for instance method)
Name of Unix Group.
  Applicable for CMode CLI/ZAPI
  Maps to:
    CMode ZAPI:group-name

=item C<< vserver=>$vserver_name >>

(Required for class method, Not Applicable for instance method)
(Required) Name of vserver or Node.
  Applicable for CMode CLI/ZAPI.
  Maps to:
    CM ZAPI: ZAPI will be invoked in the context of the vserver specified to this
             option.

=item C<< username=>$string >>

(Required) Name of a user to add to the existing unix group.
  Applicable for CMode CLI/ZAPI
  Maps to:
    CMode ZAPI:user-name

=item C<< extended_query => $string >>

(Optional)
Applicable only for CMode CLI

=item C<< '*' >>

This method takes all other options supported by
"vserver services unix-group adduser" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub adduser {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {_optional_scalars(qw(username))},
        cli_cmd         => 'vserver services unix-group adduser',
    );

    my $vserver;
    $opts{vserver_operated_on} = \$vserver;

    my $ret_obj = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_adduser_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_adduser_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            }
        ],
        frontend => 'NACL::C::VserverServicesUnixGroup::adduser',
    );
    $Log->exit() if $may_exit;
    return $ret_obj;
} ## end sub adduser

sub _adduser_cmode_cli {
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@args,
        api     => 'vserver_services_unix_group_adduser',
    );
} ## end sub _adduser_cmode_cli

sub _adduser_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_base_backend_cmode_zapi(@args,
        api    => 'name_mapping_unix_group_add_user',
        map    => {
            "name"     => "group-name",
            "username" => "user-name",
        },
    );
    $Log->exit() if $may_exit;
} ## end sub _adduser_cmode_zapi

=head2 modify

  NACL::C::VserverServicesUnixGroup->modify(command_interface=>$ci,
                             name=>$unix_group_name,...);

                            or

  $unix_group_object->modify(...);

(Class or instance method) This method modifies an existing Unix Group element.

Uses a CMode-CLI/ZAPI APISet.
  Invokes 'name-mapping-unix-group-modify' for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< name=>$unix_group_name >>

(Required for class method, Not Applicable for instance method)
Name of Unix Group.
  Applicable for CMode CLI/ZAPI
  Maps to:
    CMode ZAPI:group-name

=item C<< vserver=>$vserver_name >>

(Required for class method, Not Applicable for instance method)
(Required) Name of vserver or Node.
  Applicable for CMode CLI/ZAPI.
  Maps to:
    CM ZAPI: ZAPI will be invoked in the context of the vserver specified to this
             option.

=item C<< id=>$integer >>

(Required for class method, Not Applicable for instance method)
(Required) Integer number to represent unix group.
  Applicable for CMode CLI/ZAPI
  Maps to:
    CMode ZAPI:group-id

=item C<< extended_query => $string >>

(Optional)
Applicable only for CMode CLI

=item C<< '*' >>

This method takes all other options supported by
"vserver services unix-group modify" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub modify {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {_optional_scalars(qw(id))},
        cli_cmd         => 'vserver services unix-group modify',
    );
    my $vserver;
    $opts{vserver_operated_on} = \$vserver;
    my $ret_obj = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_modify_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_modify_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            }    ## end

        ],
        frontend => 'NACL::C::VserverServicesUnixGroup::modify',
    );
    $Log->exit() if $may_exit;
    return $ret_obj;
} ## end sub modify

sub _modify_cmode_cli {
    my ($pkg_or_obj, %opts) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(%opts,
        api => 'vserver_services_unix_group_modify',
    );
} ## end sub _modify_cmode_cli

sub _modify_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    $pkg_or_obj->_base_backend_cmode_zapi(%opts,
        api    => 'name_mapping_unix_group_modify',
        map    => {
            "id"   => "group-id",
            "name" => "group-name",
        },
    );
    $Log->exit() if $may_exit;
} ## end sub _modify_cmode_zapi

=head2 deluser

  NACL::C::VserverServicesUnixGroup->deluser(command_interface=>$ci,
                             name=>$unix_group_name,...);

                            or

  $unix_group_object->deluser(...);

(Class or instance method) This method delete a user from a local UNIX group
element.

Uses a CMode-CLI/ZAPI APISet.
  Invokes 'name-mapping-unix-group-delete-user' for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< name=>$unix_group_name >>

(Required for class method, Not Applicable for instance method)
This specifies the local UNIX group from which a user is to be deleted.
  Applicable for CMode CLI/ZAPI
  Maps to:
    CMode ZAPI:group-name

=item C<< vserver=>$vserver_name >>

(Required for class method, Not Applicable for instance method)
(Required) Name of vserver or Node.
  Applicable for CMode CLI/ZAPI.
  Maps to:
    CM ZAPI: ZAPI will be invoked in the context of the vserver specified to this
             option.

=item C<< username=>$string >>

(Required) This specifies the user name that is to be deleted from the local
UNIX group.
  Applicable for CMode CLI/ZAPI
  Maps to:
    CMode ZAPI:user-name

=item C<< extended_query => $string >>

(Optional)
Applicable only for CMode CLI

=item C<< '*' >>

This method takes all other options supported by
"vserver services unix-group deluser" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub deluser {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {_optional_scalars(qw(username))},
        cli_cmd         => 'vserver services unix-group deluser',
    );
    my $vserver;
    $opts{vserver_operated_on} = \$vserver;
    my $ret_obj = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_deluser_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_deluser_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            }
        ],
        frontend => 'NACL::C::VserverServicesUnixGroup::deluser',
    );
    $Log->exit() if $may_exit;
    return $ret_obj;
} ## end sub deluser

sub _deluser_cmode_cli {
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@args,
        api => 'vserver_services_unix_group_deluser',
    );
} ## end sub _deluser_cmode_cli

sub _deluser_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_base_backend_cmode_zapi(@args,
        api    => 'name_mapping_unix_group_delete_user',
        map    => {
            "name"     => "group-name",
            "username" => "user-name",
        },
    );
    $Log->exit() if $may_exit;
} ## end sub _deluser_cmode_zapi

=head2 load_from_uri

    NACL::C::VserverServicesUnixGroup->load_from_uri(
                                command_interface => $ci,
                                'vserver' => $vserver,
                                ...);

(Class method)
This method is used to load one or more local UNIX groups from a URI.

Uses a CMode CLI APISet.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'vserver' => $vserver >>

(Required)
Applicable only for CMode CLI.

=item C<< 'overwrite' => $string >>

Applicable only for CMode CLI.

=item C<< 'uri' => $string >>

Applicable only for CMode CLI.

=item C<< 'method-timeout' => $time_in_seconds >>

(Optional)
The default NATE timeout is 60 seconds. If the command might take more
time to complete, this option can be used to specify a larger timeout value.
The value should be provided in seconds.

=item C<< 'extended_query' => $string >>

(Optional)
Applicable only for CMode CLI.

=item C<< '*' >>

This method takes all other options supported by
"vserver services unix-group load-from-uri" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub load_from_uri {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( vserver overwrite uri
                    )
            ),
        },
        ignore_primary_keys => 1,
        cli_cmd             => 'vserver services unix-group load-from-uri',
    );
    $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_load_from_uri_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::VserverServicesUnixGroup::load_from_uri',
    );

    $Log->exit() if $may_exit;
} ## end sub load_from_uri

sub _load_from_uri_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, %opts) = @_;
    my $vserver = $opts{vserver};

    $pkg->_base_backend_cmode_cli(
        %opts,
        api                 => 'vserver_services_unix_group_load_from_uri',
        vserver_operated_on => \$vserver,
    );

    $Log->exit() if $may_exit;
} ## end sub _load_from_uri_cmode_cli


__PACKAGE__->_load_event_class();

sub _primary_keys_validate_spec {
    return (
        vserver => {type => SCALAR | UNDEF, optional => 1},
        name    => {type => SCALAR},
    );
} ## end sub _primary_keys_validate_spec

1;
