# $Id: $

#
# Copyright (c) 2001-2012 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary VserverServicesUnixUser Component Module
## @author Dheeraj.K@netapp.com, dl-nacl-dev@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::C::VserverServicesUnixUser

=head1 DESCRIPTION

C<NACL::C::VserverServicesUnixUser> is a derived class of
L<NACL::C::Component::ONTAP|lib-NACL-C-Component-ONTAP-pm>.  It represents access
to an ONTAP Unix Group.  A related class is
L<NACL::CS::VserverServicesUnixUser|lib-NACL-CS-UnixUser-pm>, which represents
the past state of an ONTAP Unix User.

=head1 ATTRIBUTES

=head2 command_interface

See L<NACL::C::Component::command_interface|lib-NACL-C-Component-pm/command_interface>

=head2 user

The name of the unix user, used to identify it uniquely.  (unix user
names are assumed to be unique within a unix user group)

=head2 vserver

This specifies the virtual server on which the local UNIX user is to be
located.

=cut

package NACL::C::VserverServicesUnixUser;

use strict;
use warnings;
use NACL::CS::VserverServicesUnixUser;
use NACL::ComponentUtils qw (_optional_scalars);
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();

use base 'NACL::C::Component::ONTAP';
use Params::Validate qw(validate_with SCALAR UNDEF);
use NACL::C::Exceptions::VserverServicesUnixUser::AlreadyExists qw(:try);
use NACL::C::Exceptions::VserverServicesUnixUser::DoesNotExist;

=head1 METHODS

=head2 new

  my $unix_user = NACL::C::VserverServicesUnixUser->new(command_interface=>$ci,
                                           user =>$unix_user_name);

Return a new Unix User component object with the given attributes.

=cut

use Class::MethodMaker [
    scalar => 'vserver',
    scalar => 'user',
];

sub init {
    $Log->enter() if $may_enter;
    my ($self, @args) = @_;
    $self->SUPER::init(@args);
    $self->_vserver_context_init(@args);
    $Log->exit() if $may_exit;
}

sub isa {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, $kind) = @_;
    my $isa = $pkg_or_obj->_build_isa(
        kind  => $kind,
        alias => 'NACL::C::UnixUser'
    );
    $Log->exit() if $may_exit;
    return $isa;
}

=head2 create

  my $unix_user = NACL::C::VserverServicesUnixUser->create(command_interface=>$ci,
                                            user=>$unix_user_name,...);

(Class method) Request creation of a new unix user element, and  return a new
unix user component object that refers to it.  Does not wait for unix user
creation to complete.

Uses a CMode-CLI APISet.
  Invokes 'useradmin user add' command for 7Mode CLI.
  Invokes 'name-mapping-unix-user-create' API for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< vserver=>$vserver_name >>

(Required) This specifies the virtual server on which the local UNIX user is
to be located.
  Applicable for CMode CLI/ZAPI
  Maps to:
    CM ZAPI: ZAPI will be invoked in the context of the vserver specified to this
             option.



=item C<< user=>$unix_user_name >>

(Required) This specifies the user account that is to be created.
  Applicable for CMode CLI/ZAPI,7Mode CLI
  Maps to:
    7Mode CLI: login name
    CMode ZAPI: user-name

=item C<< id=>$integer >>

(Required) This specifies an ID number for the user.
  Applicable for CMode CLI/ZAPI.
  Maps to:
    CMode ZAPI: user-id

=item C<< primary-gid=>$interger >>

(Required) This specifies the ID number of the user's primary group.
  Applicable for CMode CLI/ZAPI.
  Maps to:
    CMode ZAPI: group-id

=item C<< full-name=>$user_full_name >>

(Optional) This optionally specifies the user's full name.
  Applicable for CMode CLI/ZAPI,7Mode CLI
  Maps to:
    7Mode CLI : full name
    CMode ZAPI: full-name

=item C<< group => $group >>

  Applicable only for 7Mode CLI
  Maps to:
    7Mode CLI: group

=item C<< password => $password >>

  Applicable only for 7Mode CLI
  Maps to:
    7Mode CLI: password

=item C<< extended_query => $string >>

(Optional)
Applicable for CMode CLI

=item C<< '*' >>

This method takes all other options supported by
"vserver services unix-user create" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=item C<NACL::C::Exceptions::VserverServicesUnixUser::AlreadyExists>

This type of exception is thrown when an attempt is made to create a
unix user which already exist.

=back

=back

=cut

sub create {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params => \@args,
        additional_spec =>
            {_optional_scalars(qw(id primary-gid full-name password group))},
        cli_cmd => 'vserver services unix-user create',
    );
    my $ret_obj;
    try {
        $ret_obj = $pkg->call_on_apiset(
            %opts,
            choices => [
                {   method    => '_create_cmode_cli',
                    interface => 'CLI',
                    set       => 'CMode',
                },
                {   method    => '_create_cmode_zapi',
                    interface => 'ZAPI',
                    set       => 'CMode',
                },
                {   method    => '_create_7mode_cli',
                    interface => 'CLI',
                    set       => '7Mode',
                }
            ],
            frontend => 'NACL::C::VserverServicesUnixUser::create',
        );
    } catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = shift;
        NACL::C::Exceptions::VserverServicesUnixUser::AlreadyExists
            ->detect_convert_and_throw('exception' => $exception,);
        $Log->exit() if $may_exit;
        $exception->throw();
    };

    $Log->exit() if $may_exit;
    return $ret_obj;
} ## end sub create

sub _create_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, %opts) = @_;
    my $vserver = $opts{vserver};
    $pkg->_base_backend_cmode_cli(
        %opts,
        api => 'vserver_services_unix_user_create',
        vserver_operated_on => \$vserver,
    );

    $Log->exit() if $may_exit;
    return $pkg->new(
        'vserver'         => $vserver,
        'user'            => $opts{'user'},
        command_interface => $opts{command_interface},
    );
} ## end sub _create_cmode_cli

sub _create_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg, %opts) = @_;

    my $vserver = $opts{vserver};
    my $allowed_cmode_options = [qw(full-name)];
    $pkg->_base_backend_cmode_zapi(
        %opts,
        frontend_method => 'create',
        vserver_operated_on => \$vserver,
        api    => 'name-mapping-unix-user-create',
        copy   => $allowed_cmode_options,
        map    => {
            "primary-gid" => "group-id",
            "id"          => "user-id",
            "user"        => "user-name",
        },
    );

    $Log->exit() if $may_exit;
    return $pkg->new(
        vserver           => $vserver,
        user              => $opts{user},
        command_interface => $opts{command_interface},
    );
} ## end sub _create_cmode_zapi

sub _create_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};

    my $allowed_7mode_options = [qw(group password full-name)];
    my %mode_args;
    $pkg->_hash_copy(
        source => \%opts,
        copy   => $allowed_7mode_options,
        target => \%mode_args,
        map    => {"user" => "login-name",},
    );
    %opts = %mode_args;

    $apiset->useradmin_user_add(%opts);

    $Log->exit() if $may_exit;
    return $pkg->new(
        'vserver'         => undef,
        'user'            => $opts{'login-name'},
        command_interface => $command_interface
    );
} ## end sub _create_7mode_cli

=head2 delete

  NACL::C::VserverServicesUnixUser->delete(command_interface=>$ci,
                            user=>$unix_user_name,...);

                            or

  $unix_user_object->delete(...);

(Class or instance method) This method deletes an Unix User element.

Uses a CMode-CLI APISet.
  Invokes 'useradmin user delete' command for 7Mode CLI.
  Invokes 'name-mapping-unix-user-destroy' API for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< vserver=>$vserver_name >>

(Required for class method, Not Applicable for instance method)
(Required) This specifies the virtual server on which the local UNIX user is
located.
  Applicable for CMode CLI/ZAPI.
  Maps to:
    CM ZAPI: ZAPI will be invoked in the context of the vserver specified to this
             option.



=item C<< user=>$unix_user_name >>

(Required for class method, Not Applicable for instance method)
(Required) This specifies the user account that is to be deleted.
  Applicable for CMode CLI/ZAPI,7Mode CLI.
  Maps to:
    7Mode CLI: login-name
    CMode ZAPI: user-name

=item C<< extended_query => $string >>

(Optional)
Applicable only for CMode CLI

=item C<< '*' >>

This method takes all other options supported by
"vserver services unix-user delete" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=item C<NACL::C::Exceptions::VserverServicesUnixUser::DoesNotExist>

This type of exception is thrown when an attempt is made to delete 
a unix user which doesn't exist.

=back

=back

=cut

sub delete {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params => \@args,
        cli_cmd    => 'vserver services unix-user delete',
    );

    my $ret_obj;
    try {
        $ret_obj = $pkg_or_obj->call_on_apiset(
            %opts,
            choices => [
                {   method    => '_delete_cmode_cli',
                    interface => 'CLI',
                    set       => 'CMode',
                },
                {   method    => '_delete_cmode_zapi',
                    interface => 'ZAPI',
                    set       => 'CMode',
                },
                {   method    => '_delete_7mode_cli',
                    interface => 'CLI',
                    set       => '7Mode',
                }
            ],
            frontend => 'NACL::C::VserverServicesUnixUser::delete',
        );
    } catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = shift;
        NACL::C::Exceptions::VserverServicesUnixUser::DoesNotExist
            ->detect_convert_and_throw('exception' => $exception,);
        $Log->exit() if $may_exit;
        $exception->throw();
    };

    $Log->exit() if $may_exit;
    return $ret_obj;
} ## end sub delete

sub _delete_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, %opts) = @_;

    my $vserver = $opts{vserver};
    $pkg->_base_backend_cmode_cli(
        %opts,
        api => 'vserver_services_unix_user_delete',
        vserver_operated_on => \$vserver,
    );

    $Log->exit() if $may_exit;
} ## end sub _delete_cmode_cli

sub _delete_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg,%opts) = @_;

    my $vserver = $opts{vserver};
    $pkg->_base_backend_cmode_zapi(
        %opts,
        frontend_method => 'delete',
        vserver_operated_on => \$vserver,
        api => 'name-mapping-unix-user-destroy',
        map => {
            "user" => "user-name",
        },
    );

    $Log->exit() if $may_exit;
} ## end sub _delete_cmode_zapi

sub _delete_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};

    my %mode_args;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%mode_args,
        map    => {"user" => "login-name",},
    );
    %opts = %mode_args;

    $apiset->useradmin_user_delete(%opts);
    $Log->exit() if $may_exit;
    return $pkg_or_obj->new(
        'vserver'         => undef,
        'user'            => $opts{'login-name'},
        command_interface => $command_interface
    );

} ## end sub _delete_7mode_cli

=head2 modify

  NACL::C::VserverServicesUnixUser->modify(command_interface=>$ci,
                            user=>$unix_user_name,...);

                            or

  $unix_user_object->modify(...);

(Class or instance method) This method modifies an Unix User element.

Uses a CMode-CLI APISet.
  Invokes 'useradmin user modify' command for 7Mode CLI.
  Invokes 'name-mapping-unix-user-modify' API for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< vserver=>$vserver_name >>

(Required for class method, Not Applicable for instance method)
(Required) This specifies the virtual server on which the local UNIX user is
located.
  Applicable for CMode CLI/ZAPI.
  Maps to:
    CM ZAPI: ZAPI will be invoked in the context of the vserver specified to this
             option.



=item C<< user=>$unix_user_name >>

(Required for class method, Not Applicable for instance method)
(Required) This specifies the user account that is to be modified.
  Applicable for CMode CLI/ZAPI,7Mode CLI
  Maps to:
    7Mode CLI: login-name
    CMode ZAPI: user-name

=item C<< id=>$integer >>

(Optional) This specifies an ID number for the user.
  Applicable for CMode CLI/ZAPI.
  Maps to:
    CMode ZAPI: user-id

=item C<< primary-gid=>$interger >>

(Optional) This specifies the ID number of the user's primary group.
  Applicable for CMode CLI/ZAPI.
  Maps to:
    CMode ZAPI: group-id

=item C<< full-name=>$user_full_name >>

(Optional) This optionally specifies the user's full name.
  Applicable for CMode CLI/ZAPI,7Mode CLI
  Maps to:
    7Mode CLI : full-name
    CMode ZAPI: full-name

=item C<<extended_query=>$string >>

(Optional)
Applicable only for CMode CLI

=item C<< '*' >>

This method takes all other options supported by
"vserver services unix-user modify" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub modify {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {_optional_scalars(qw (id primary-gid full-name))},
        cli_cmd         => 'vserver services unix-user modify',
    );
    my $vserver;
    $opts{vserver_operated_on} = \$vserver;
    my $ret_obj = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_modify_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_modify_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
            {   method    => '_modify_7mode_cli',
                interface => 'CLI',
                set       => '7Mode',
            }
        ],
        frontend => 'NACL::C::VserverServicesUnixUser::modify',
    );
    $Log->exit() if $may_exit;
    return $ret_obj;
} ## end sub modify

sub _modify_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@args,
        api => 'vserver_services_unix_user_modify',
    );
    $Log->exit() if $may_exit;
} ## end sub _modify_cmode_cli

sub _modify_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};

    my $allowed_cmode_options = [qw( full-name)];
    my %cmode_args;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => $allowed_cmode_options,
        target => \%cmode_args,
        map    => {"user" => "login-name",},
    );
    %opts = %cmode_args;
    $apiset->useradmin_user_modify(%opts);
    $Log->exit() if $may_exit;
} ## end sub _modify_7mode_cli

sub _modify_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_base_backend_cmode_zapi(@args,
        api    => 'name_mapping_unix_user_modify',
        copy   => [qw(full-name)],
        map    => {
            "primary-gid" => "group-id",
            "id"          => "user-id",
            "user"        => "user-name",
        },
    );
    $Log->exit() if $may_exit;
} ## end sub _modify_cmode_zapi

=head2 load_from_uri

    NACL::C::VserverServicesUnixUser->load_from_uri(
                                command_interface => $ci,
                                'vserver' => $vserver,
                                ...);

(Class method)
This method is used to load one or more local UNIX users from a URI.

Uses a CMode CLI APISet.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'vserver' => $vserver >>

(Required)
Applicable only for CMode CLI.

=item C<< 'overwrite' => $string >>

Applicable only for CMode CLI.

=item C<< 'uri' => $string >>

Applicable only for CMode CLI.

=item C<< 'method-timeout' => $time_in_seconds >>

(Optional)
The default NATE timeout is 60 seconds. If the command might take more
time to complete, this option can be used to specify a larger timeout value.
The value should be provided in seconds.

=item C<< 'extended_query' => $string >>

(Optional)
Applicable only for CMode CLI.

=item C<< '*' >>

This method takes all other options supported by
"vserver services unix-user load-from-uri" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub load_from_uri {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( vserver overwrite uri
                    )
            ),
        },
        ignore_primary_keys => 1,
        cli_cmd             => 'vserver services unix-user load-from-uri',
    );
    $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_load_from_uri_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::VserverServicesUnixUser::load_from_uri',
    );

    $Log->exit() if $may_exit;
} ## end sub load_from_uri

sub _load_from_uri_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, %opts) = @_;
    my $vserver = $opts{vserver};
    $pkg->_base_backend_cmode_cli(
            %opts,
            api     => 'vserver_services_unix_user_load_from_uri',
            vserver_operated_on => \$vserver,
        );

    $Log->exit() if $may_exit;
} ## end sub _load_from_uri_cmode_cli


__PACKAGE__->_load_event_class();

sub _primary_keys_validate_spec {
    return (
        vserver => {type => SCALAR | UNDEF, optional => 1},
        user    => {type => SCALAR},
    );
} ## end sub _primary_keys_validate_spec

1;
