# $Id $

#
# Copyright (c) 2001-2012 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Common methods for the StorageAggregate.pm and Volume.pm Components
## @author dl-nacl-dev@netapp.com
## @status shared
## @pod here

package NACL::C::_Mixins::AggregateVolume;

use strict;
use warnings;

use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Params::Validate qw(validate_with validate :types);
use NATE::Exceptions::Argument;

BEGIN {
    use Exporter qw(import);
    our @EXPORT_OK = qw(_translate_options_values _handle_snaplock_type);
    our %EXPORT_TAGS = ( 'all' => \@EXPORT_OK );
}

################################################################################
# Method:    _translate_options_values
# Objective: Many of the aggregate and volume options that show up in 7Mode 
#            have values on/off while their CMode equivalents have values 
#            true/false. This subroutine is to translate accordingly.
# Inputs:
#         1. hashref
#            This is the hash reference within which we want to translate the
#            values of the options
#         2. direction
#            This can be either 'SEVENTOCLUSTER' or 'CLUSTERTOSEVEN' and it
#            specifies the sort of translation we desire.
#            (does an on -> off and off -> on if
#            it is 'SEVEN', and does a true -> false and false -> true if it
#            is 'CLUSTER'
#         3. values_to_translate
#            These are the option names whose values are to be translated
#            (This can be populated by using _option_values_to_translate()
#            defined in the respective packages
#
# Returns:   Nothing. The input hashref has the required option values translated
###################################################################################

sub _translate_options_values {

    my %opts = validate(
        @_,
        {   hashref             => { type => HASHREF },
            direction           => { type => SCALAR },
            values_to_translate => { type => ARRAYREF },
        },
    );

    my $hashref             = $opts{hashref};
    my $direction           = $opts{direction};
    my @values_to_translate = @{ $opts{values_to_translate} };

    my %true_false_hash;
    if ( $direction eq 'SEVENTOCLUSTER' ) {
        %true_false_hash = ( 'on' => 'true', 'off' => 'false' );
    } elsif ( $direction eq 'CLUSTERTOSEVEN' ) {
        %true_false_hash = ( 'true' => 'on', 'false' => 'off' );
    } else {
        NATE::Exceptions::Argument->throw( 'Internal error: The value of the '
                . "'direction' parameter should "
                . "be either of 'SEVENTOCLUSTER' or 'CLUSTERTOSEVEN' but "
                . "the value provided was $direction" );
    } ## end else [ if ( $direction eq 'SEVENTOCLUSTER')

    foreach my $val_to_translate (@values_to_translate) {
        if ( defined $hashref->{$val_to_translate} ) {
            if ( exists $true_false_hash{ $hashref->{$val_to_translate} } ) {
                $hashref->{$val_to_translate}
                    = $true_false_hash{ $hashref->{$val_to_translate} };
            }
        } ## end if ( defined $hashref->...
    } ## end foreach my $val_to_translate...
} ## end sub _translate_options_values

sub _handle_snaplock_type {
   $Log->enter();
   my ($pkg, @args) = @_;
   my %opts = validate_with(
       params => \@args,
       spec   => {
           options => {type => HASHREF},
           final_hashref => {type => HASHREF},
       },
    );
    my $options = $opts{options};
    my $snaplock_type;
    if (exists $options->{snaplock_compliance}) {
         $snaplock_type = 'compliance';
    } elsif (exists $options->{snaplock_enterprise}) {
        $snaplock_type = 'enterprise';
    }
    my $final_hashref = $opts{final_hashref};
    if (defined $snaplock_type) {
         $final_hashref->{'snaplock-type'} = $snaplock_type;
         $final_hashref->{'is-snaplock'} = 'true';
    } else {
         $final_hashref->{'is-snaplock'} = 'false';
         $final_hashref->{'snaplock-type'} = 'non-snaplock';
    }
}

1;
