#
# Copyright (c) 2012-13 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Common methods for various components
## @author dl-nacl-dev@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::C::_Mixins::Common

=head1 DESCRIPTION

This mix-in contains implementations for methods common to various components.

=cut

package NACL::C::_Mixins::Common;

use strict;
use warnings;

BEGIN {
    use Exporter qw(import);
    our @iv_methods = qw(
        _determine_is_repository
        _determine_is_infinivol
        _read_infinivol_val
        _update_infinivol_val_in_self
        _infinivol_vol_vs_check
        _add_infinivol_params_to_additional_common_validate_spec
    );
    our @EXPORT_OK = (
        @iv_methods,
        qw(
            _check_relational_regex_filter
            get_7m_or_nodescope_apiset
        )
    );
    our %EXPORT_TAGS = (
        'all'      => \@EXPORT_OK,
        iv_methods => \@iv_methods,
    );
}

use Params::Validate qw(:all);
use NACL::ComponentUtils qw(_optional_scalars);

use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use NACL::STask::Vserver;
use NACL::STask::Volume;
use NATE::BaseException;
use Data::Dumper;

=head1 METHODS

=head2 _determine_is_repository

    # Can also be imported using the ":all" tag
    use NACL::C::_Mixins::Common qw(_determine_is_repository);
    ...
    
    sub _foo_cmode_zapi {
        ...
        my $boolean = $pkg_or_obj->_determine_is_repository(
            command_interface => $ci,
            vserver           => $vserver,
            'is-repository'   => $opts{'is-repository'}
        );
        
        if ($boolean) {
            # invoke async ZAPI
        } else {
            # invoke normal ZAPI
        }
    }
    
Numerous ZAPIs do not work for repository vserver. For these ZAPIs, there is
an equivalent *-async ZAPI that should be used. This method can be invoked
by CMode ZAPI back-end methods to determine which ZAPI to use.

This method requires that the component into which it's being mixed-into
has a scalar "_is_repository" attribute defined. Also, for each CMode ZAPI
back-end method that invokes this, the front-end method should list
"is-repository" as an optional scalar (and should have documentation for
the same)

This method does the following:

* If 'is-repository' is passed in the method call, then it returns that value
  (since the user has explicitly stated what to use)
  
* If the component object has a value cached, return it.

* If a vserver command interface is being used and it has a value cached,
  return it.
 
* Else, query the filer to check whether this vserver is a repository
  vserver or not.
  
=over

=item Options

=over

=item C<< command_interface => $ci >>

(Mandatory)
The command interface

=item C<< vserver => $vserver >>

(Mandatory)
The vserver

=item C<< 'is-repository' => 0 | 1 >>

(Optional)


=back

=back

=cut

sub _determine_is_repository {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my @primary_keys = $pkg_or_obj->get_primary_keys();
    my %opts         = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(@primary_keys),
            'is-repository' => {type => SCALAR | UNDEF, optional => 1}
        },
        ignore_primary_keys => 1,
        allow_extra         => 1
    );
    unless ($pkg_or_obj->can('_is_repository')) {
        $Log->exit() if $may_exit;
        NATE::BaseException->throw("Internal error: The package '"
                . $pkg_or_obj->get_package_name()
                . "' should have an '_is_repository' attribute defined "
                . 'but it does not (this attribute is required for the method '
                . "'_determine_is_repository' to work)");
    }

    my $ci             = $opts{command_interface};
    my $ci_isa_vserver = $ci->isa('NACL::C::Vserver');
    if (!defined $opts{vserver} && $ci_isa_vserver) {
        $opts{vserver} = $ci->vserver();
    }

    my %is_repo_opts;
    $pkg_or_obj->_copy_common_component_params_with_ci(
        source => \%opts,
        target => \%is_repo_opts
    );
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%is_repo_opts,
        copy   => [qw(vserver is-repository)]
    );

    if (defined $opts{'is-repository'}) {

        # If a value of "is-repository" was explicitly passed in the method
        # call, use that value
        $Log->exit() if $may_exit;
        return $opts{'is-repository'};
    } elsif (ref($pkg_or_obj) && $pkg_or_obj->_is_repository_isset()) {

        # If a value is cached in the component object, return it
        $Log->exit() if $may_exit;
        return $pkg_or_obj->_is_repository();
    } elsif ($ci_isa_vserver && defined $ci->{_is_repository}) {

        # If a vserver command interface is being used and it has cached
        # a value, return it.
        # Not using _isset, since it has a default constructor and this
        # would cause an infinite recursion
        $Log->exit() if $may_exit;
        return $ci->_is_repository();
    } else {
        my $is_repository
            = NACL::C::Vserver->determine_is_repository(%is_repo_opts,
            apiset_should => {interface => 'ZAPI'});

        # Update the object (if invoked with an object instance)
        $pkg_or_obj->_is_repository($is_repository) if (ref $pkg_or_obj);

        # If the ci is a vserver, update it
        $ci->_is_repository($is_repository) if ($ci_isa_vserver);

        $Log->exit() if $may_exit;
        return $is_repository;
    }
}

=head2 _determine_is_infinivol

    # Can also be imported using the ":all" tag
    use NACL::C::_Mixins::Common qw(_determine_is_infinivol);
    
    # Class method
    my $is_infinivol = NACL::C::Volume->_determine_is_infinivol(
        volume  => $vol_name,
        vserver => $vs_name,
        %other_opts
    );
    
    # Instance method
    my $is_infinivol = $vol_obj->_determine_is_infinivol(%other_opts);

Determine if the provided volume is an infinivol or not. If invoked with
an object instance, it updates the "_is_infinivol" attribute with this value.

=over

=item Options

=over

=item For class method calls only

=over

=item C<< volume => $vol_name | $vol_obj >>

=item C<< command_interface => $ci >>

=item C<< vserver => $vs_name | $vs_obj >>

Required if the C<command_interface> is not a Vserver object.

=back

=item C<common options>

Such as apiset_must, apiset_should etc

=item Any other option

All extra options are accepted (does not result in validation error) but
are ignored.

=back

=back

=cut

sub _determine_is_infinivol {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts = validate_with(
        params => \@args,
        spec   => { volume => { type => OBJECT | SCALAR } },
        allow_extra => 1
    );
    
    unless ($pkg_or_obj->can('_is_infinivol')) {
        $Log->exit() if $may_exit;
        NATE::BaseException->throw("Internal error: The package '"
                . $pkg_or_obj->get_package_name()
                . "' should have an '_is_infinivol' attribute defined "
                . 'but it does not (this attribute is required for the method '
                . "'_determine_is_infinivol' to work)");
    }
    
    my $is_infinivol = $opts{nacltask_is_infinivol} || $opts{is_infinivol};
    
    unless (defined $is_infinivol) {
        if (ref ($pkg_or_obj) && $pkg_or_obj->_is_infinivol_isset()) {
            # Use the value cached in the component object
            $is_infinivol = $pkg_or_obj->_is_infinivol();
        } else {
            my %common_opts_ci;
            $pkg_or_obj->_copy_common_component_params_with_ci(
                source => \%opts,
                target => \%common_opts_ci
            );
            my @vols_cs = NACL::CS::Volume->fetch(
                %common_opts_ci,
                filter           => {
                    volume  => $opts{volume},
                    vserver => $opts{vserver},
                },
                requested_fields => ['volume-style'],
                allow_empty      => 1
            );
            if (@vols_cs) {
                if (@vols_cs == 1) {
                    if ($vols_cs[0]->volume_style() eq 'infinitevol') {
                        $is_infinivol = 1;
                    } else {
                        $is_infinivol = 0;
                    }
                } else {
                    $Log->exit() if $may_exit;
                    NATE::Exceptions::Argument->throw('Internal error: '
                        . 'The volume and vserver should have been '
                        . 'provided in the call to _determine_is_infinivol '
                        . 'but were not');
                }
            } else {
                # Could not find the volume, the underlying API call will
                # fail, so it doesn't really matter what value we set
                # is_infinivol to. Let's just set it to 0.
                $is_infinivol = 0;
            }
        }
    }
    
    if (ref ($pkg_or_obj) && !$pkg_or_obj->_is_infinivol_isset()) {
        # Set the component's attribute to this value
        $pkg_or_obj->_is_infinivol($is_infinivol);
    }
    
    $Log->exit($is_infinivol);
    return $is_infinivol;
}

sub _read_infinivol_val {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts = validate_with(
        params => \@args,
        spec   => {
            is_infinivol     => {type => SCALAR,    optional => 1},
            is_infinivol_ref => {type => SCALARREF, optional => 1},
        },
        allow_extra => 1
    );
    my $is_infinivol;
    if (defined $opts{is_infinivol}) {
        $is_infinivol = $opts{is_infinivol};
    } elsif (defined ${$opts{is_infinivol_ref}}) {
        $is_infinivol = ${$opts{is_infinivol_ref}};
    }

    $Log->exit() if $may_exit;
    return $is_infinivol;
}

sub _update_infinivol_val_in_self {
    $Log->enter() if $may_enter;
    my ($self, @args) = @_;
    my $is_infinivol = $self->_read_infinivol_val(@args);
    if (defined $is_infinivol) {
        $self->_is_infinivol($is_infinivol);
    }
    $Log->exit() if $may_exit;
}

sub _add_infinivol_params_to_additional_common_validate_spec {
    return {
        is_infinivol     => {type => SCALAR, optional => 1},
        is_infinivol_ref => {type => SCALARREF, optional => 1},
    };
}

=head2 _check_relational_regex_filter

Method which checks whether the passed in, value of a filter field  contains relational or
regex expressions . Returns 0 or 1.

=cut

sub _check_relational_regex_filter {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts = validate (@args, {filter_value => {type => SCALAR|OBJECT|ARRAYREF}});
    my $ret = 0;
    # Check if the filter value is anyone of the type  value*, !value, >=value,
    # <=value, >value, <value,
    # value1..value2 and value1|value2  AND is not
    # within quotes (values within quotes are to be considered as literals)
    my $filter_value_ref = (ref $opts{filter_value} eq 'ARRAY') ? $opts{filter_value} : [$opts{filter_value}];
    foreach my $filter_value (@$filter_value_ref) {
        if ( (   $filter_value =~ /^[!><]/
            || $filter_value =~ /\S+(\.\.|\|)\S+/
            || $filter_value =~ /\*/) 
            && $filter_value !~ /^".*"$/ && !(ref($opts{filter_value}) 
            && !ref $opts{filter_value} eq 'ARRAY')) {
            $ret = 1;
            last;
        }
    }   
    $Log->exit() if $may_exit;
    return $ret;
}

sub get_7m_or_nodescope_apiset {
    $Log->enter() if $may_enter;
    my ($self, @args) = @_;
    $Log->debug(sub { "Opts are:\n" . Dumper({@_}) });
    my %opts = validate_with(
       params => \@args,
       spec   => {interface => {type => SCALAR, default => 'CLI'}},

       # All other options are validated by the apiset() method
       allow_extra => 1
    );
    foreach my $not_allowed (qw(category set)) {
       if (exists $opts{$not_allowed}) {
           $Log->exit() if $may_exit;
            NATE::Exceptions::Argument->throw("The argument '$not_allowed' "
                      . 'is not applicable in a call to '
                      . "'get_7m_or_nodescope_apiset'");
       }
    }

    $opts{category} = 'Node';

    if ($self->is_7mode()) {
            $opts{set} = '7Mode';
    } else {
            $opts{set} = 'Nodescope';
    }

    local $Data::Dumper::Maxdepth = 2;
    local $Data::Dumper::Terse    = 1;
    $Log->debug(sub { 'Returning $self->apiset(' . Dumper(\%opts) . ')' });
    my $apiset = $self->apiset(%opts);
    $Log->exit() if $may_exit;
    return $apiset;
} ## end sub get_7m_or_nodescope_apiset

# For cases where there is an async ZAPI, the back-end chooses whether to run
# the normal ZAPI or the async form depending on whether the volume is an
# infinivol or not. If the call is not limited to a single volume (by
# not providing volume/vserver, or if they have special characters), then
# we won't know whether to use the normal or the async version, so we
# fall-back to the CLI.
sub _infinivol_vol_vs_check {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    $pkg_or_obj->_fallback_for_relational_regex(%opts,
        fields_to_check => [qw(volume vserver)]);
    $Log->exit() if $may_exit;
}

1;
