#
# Copyright (c) 2001-2012 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Common methods for the ONTAP Component and ComponentState modules
## @author dl-nacl-dev@netapp.com
## @status shared
## @pod here

package NACL::C::_Mixins::Component::ONTAP;

use strict;
use warnings;
use feature 'state';

use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Params::Validate qw(validate_with HASHREF SCALAR UNDEF ARRAYREF);
use NACL::ComponentUtils qw( _ontap_ci_validate_spec);
use NATE::Exceptions::Argument qw(:try);
use NATE::BaseException qw(:try);

BEGIN {
    use Exporter qw(import);
    our @EXPORT_OK = qw(
        _handle_zapi_vserver_context
        _handle_vserver_context
        _build_isa
        _base_check_non_iter
    );
    our %EXPORT_TAGS = ( 'all' => \@EXPORT_OK );
    # Turn Params::Validate failures into NATE::Exceptions::Argument
    Params::Validate::validation_options(
        on_fail => sub {
            NATE::Exceptions::Argument->throw(shift);
        }
    );
}

=head2 _handle_zapi_vserver_context

    # NACL/C/Foo.pm
    sub _foo_cmode_zapi {
        ...
        my $command_interface = $opts{command_interface};
        my $apiset            = $opts{apiset};
        my $vserver           = $opts{vserver};
        ...
        $pkg_or_obj->_hash_copy(...);

        my $vserver_operated_on = $pkg_or_obj->_handle_zapi_vserver_context(
            api_opts          => \%cmode_args,
            vserver           => $vserver,
            command_interface => $command_interface,
            frontend_method   => $method
        );

        $apiset->foo_api(%cmode_args);
    }

(For component developers)
Certain commands/ZAPIs can be run in Vserver context. These methods
are used to handle the differences between running these methods using
Vserver as a command interface or using Node/Cluster/Cserver as the
command interface. It returns the vserver being operated on (could
be the "vserver" argument passed in, or the name of the vserver which
the Vserver command interface is modeling)

For ZAPI, the method to be invoked is C<_handle_zapi_vserver_context>.
This ensures that if a vserver-scope ZAPI is being invoked with a Node
command interface, then it will set the API to run in the context of the
vserver passed as argument.

This method throws a C<NATE::Exceptions::Argument> exception in the following
cases:

=over

=item 1)

If a vserver command interface is used and a vserver argument
is sent in the method call, but the argument does not match the
vserver of the command interface

=item 2)

The vserver argument is not passed when a non-Vserver
command interface is passed

=back

This method returns the name of the vserver on which we will operate. (This
could be the vserver attribute of the command interface or the vserver
option passed in)

=over

=item Options

=over

=item api_opts

(HASHREF, Mandatory)
A reference to the hash sent to the API call. This gets updated to contain
the vserver context to set to if a Node/Cluster command interface is used

=item vserver

(Mandatory)
The vserver option passed in

=item command_interface

(Isa 'NACL::C::CommandInterface::ONTAP', Mandatory)
The command interface being used

=back

=back

=cut

sub _handle_zapi_vserver_context {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    my $vserver = $pkg_or_obj->_handle_vserver_context(@args, is_zapi => 1);
    $Log->exit() if $may_exit;
    return $vserver;
}

#
# _base_check_non_iter is used to check of non-iter choice is valid.
# non-iter can work only for 2 case.
# case1. When only id used to make the ontap call.
# case2. method call is made using the class call but filter contains
#        all primary keys.
# If these conditions are not met, InvalidChoice Exception is thrown.
#
sub _base_check_non_iter {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    
    if (ref $pkg_or_obj) {
        # Invoked with an object, so is limited to a single element,
        # hence non-iter is applicable
        $Log->exit() if $may_exit;
        return;
    }
    
    # If we've reached here, then $pkg_or_obj is package name.

    state $spec = {
        _primary_keys => {type => ARRAYREF, optional => 1},
        _vs_context   => {type => SCALAR, optional => 1},
    };
    my %opts = validate_with(
        params      => \@args,
        spec        => $spec,
        allow_extra => 1,
    );
    my $primary_keys_input = delete $opts{_primary_keys};
    my $vs_context = delete $opts{_vs_context};

    my %opts_to_look_at;
    require Module::Load;
    my $C_pkg;
    if ($pkg_or_obj->isa("NACL::CS::ComponentState")) {
        %opts_to_look_at = %{$opts{filter}};
        $C_pkg           = $pkg_or_obj->get_C_package_name();
    } else {
        %opts_to_look_at = %opts;
        $C_pkg = $pkg_or_obj;
    }

    my @primary_keys;
    if (defined $primary_keys_input) {
        @primary_keys = @$primary_keys_input;
    } else {
        # Attempt to get the primary keys
        eval { Module::Load::load($C_pkg) };
        if ($@) {
            # Could not load the C package (likely because the original
            # caller was a CS-like method, so there's no corresponding C
            # package)
            $Log->exit() if $may_exit;
            NATE::BaseException->throw('Internal NACL error: For non-iter '
                . 'back-end method, the primary keys were not provided and '
                . 'could not be determined.'
            );
        } else {
            @primary_keys = $C_pkg->get_primary_keys(command_interface =>
                $opts{command_interface});
        }
    }
    
    # If this is a vserver-context ZAPI, then either a vserver
    # command_interface should be used or "vserver" should be provided
    # and should not contain special characters. Let's remove "vserver"
    # from the primary keys list so that we can handle it separately.
    if ($vs_context) {
        @primary_keys = grep {$_ ne 'vserver'} @primary_keys;
        my $opts_vserver = $opts_to_look_at{vserver};
        if (defined $opts_vserver) {
            my $has_sp_chr = $pkg_or_obj->_check_relational_regex_filter(
                filter_value => $opts_vserver);
            delete $opts_to_look_at{vserver} if $has_sp_chr;
        }
        
        try {
            $pkg_or_obj->_handle_zapi_vserver_context(
                api_opts          => \%opts_to_look_at,
                command_interface => $opts{command_interface},
            );
        } otherwise {
            # _handle_zapi_vserver_context failed because it could not
            # determine the vserver context. Mark this back-end as
            # inapplicable by throwing an InvalidChoice exception.
            my $vs_err_str;
            if (defined $opts_vserver) {
                $vs_err_str = "and the vserver value provided ('$opts_vserver') "
                    . 'has special characters.';
            } else {
                $vs_err_str = 'and the vserver value was not provided';
            }
            $Log->exit() if $may_exit;
            NACL::Exceptions::InvalidChoice->throw('Cannot use non-iter '
                . 'because the ZAPI is vserver-context, but command_interface '
                . 'was not a Vserver object (isa '
                . ref ($opts{command_interface})
                . " object) $vs_err_str"
            );
        };
    }

    foreach my $pk (@primary_keys) {
        if (exists $opts_to_look_at{$pk}) {
            # make sure value of $pk does not contain special character.
            my $value = $opts_to_look_at{$pk};
            if (defined $value) {
                my $has_sp_chr = $pkg_or_obj->_check_relational_regex_filter(
                    filter_value => $value);
                if ($has_sp_chr) {
                    $Log->exit() if $may_exit;
                    NACL::Exceptions::InvalidChoice->throw(
                        'non-iter cannot be used as Primary key '
                        . "$pk contains special characters");
                }
            } else {
                $Log->exit() if $may_exit;
                NACL::Exceptions::InvalidChoice->throw('non-iter cannot '
                    . "be used as Primary key $pk not specified for $pkg_or_obj");
            }
        } else {
            $Log->exit() if $may_exit;
            NACL::Exceptions::InvalidChoice->throw("non-iter cannot be used".
            " as Primary key $pk not specified for $pkg_or_obj");
        }
    }
    $Log->exit() if $may_exit;
    return;
}

sub _handle_vserver_context {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts = validate_with(
        params => \@args,
        spec   => {
            api_opts => { type => HASHREF },

            # Should look to get rid of this -- the value of this can be
            # obtained from api_opts->{vserver}
            vserver  => { type => SCALAR | UNDEF, optional => 1 },
            command_interface => _ontap_ci_validate_spec(),
            is_zapi           => { type => SCALAR },
            frontend_method   => { type => SCALAR, optional => 1 },
        }
    );
    my $api_opts          = $opts{api_opts};
    my $vserver           = $api_opts->{vserver} || $opts{vserver};
    my $command_interface = $opts{command_interface};
    my $is_zapi           = $opts{is_zapi};

    if ( $command_interface->isa('NACL::C::Vserver') ) {
        my $vs_ci = $command_interface->vserver();
        if (defined $vserver) {
            # For ZAPI, check that the vserver as modeled by the
            # command_interface is the same as the vserver option passed in.
            # (For CLI we don't perform this check because we want it
            # to go down to the command and the command should fail)
            if ($vs_ci eq $vserver) {
                # "vserver" option removed for commands
                # in vserver context in FS.0
                if (!$is_zapi &&
                    $command_interface->version_manager()->is_version_ge(release
                        => '8.3')) {
                    delete $api_opts->{vserver};
                }
            } elsif ($is_zapi) {
                # The vserver modeled by the CI and the vserver provided in
                # the input arguments do not match, and this is ZAPI
                my $first_line;
                if ( defined $opts{frontend_method} ) {
                    my $pkg = $pkg_or_obj->get_package_name();
                    $first_line = "${pkg}::$opts{frontend_method}: The ";
                } else {
                    $first_line = 'The ';
                }
                $Log->exit() if $may_exit;
                NATE::Exceptions::Argument->throw( $first_line
                        . "command interface's vserver ('$vs_ci')"
                        . 'does not match the vserver option passed in the '
                        . "method call ('$vserver')\n" );
            }
        } else {

            # The vserver being operated on is the one modeled by the CI
            $vserver = $vs_ci;
        }
    } else {
        if ( (!defined $vserver) && (ref($command_interface) ne 'NACL::C::Cserver') ) {

            # "vserver" not provided and command_interface was not
            # a vserver. Error out.
            my $ci_type = ref $command_interface;
            $ci_type =~ s/.*:://;

            my $first_line;
            if ( defined $opts{frontend_method} ) {
                my $pkg = $pkg_or_obj->get_package_name();
                $first_line
                    = "${pkg}::$opts{frontend_method}: The 'vserver' ";
            } else {
                $first_line = "The 'vserver' ";
            }
            my $error
                = $first_line
                . 'parameter must be provided when call '
                . "is made with a $ci_type command interface\n";
            if ($is_zapi) {
                $Log->exit() if $may_exit;
                NATE::Exceptions::Argument->throw($error);
            } else {
                $Log->warn( $error
                        . 'Please mail dl-nacl-dev@netapp.com if '
                        . 'this is not the case for this method' );
            }
        } elsif ($is_zapi) {

            # Make the ZAPI run in the context of the vserver
            $api_opts->{'vserver-override-name'} = $vserver;
        }
    }

    $Log->exit() if $may_exit;
    return $vserver;
}

=head2 _build_isa

    # Using NACL::C::SystemNode as example
    sub isa {
        $Log->enter() if $may_enter;
        my ($pkg_or_obj, $kind) = @_;
        my $isa = $pkg_or_obj->_build_isa(
            kind  => $kind,
            alias => 'NACL::C::Node'
        );
        $Log->exit() if $may_exit;
        return $isa;
    }

For components that have a related alias component, we need to write
a custom "isa" to make the real component "isa" the alias component.

=over

=item Options

=over

=item C<< kind => $kind >>

The package name passed as argument to "isa".

=item C<< alias => $alias >>

The name of the alias component.

=back

=back

=cut

sub _build_isa {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts = validate_with(
        params => \@args,
        spec   => {
            kind  => { type => SCALAR },
            alias => { type => SCALAR }
        }
    );
    my $kind = $opts{kind};
    if ($kind eq $opts{alias}) {
        $Log->exit() if $may_exit;
        return 1;
    } else {
        # The CPAN module UNIVERSAL::isa is supposed to catch and fix
        # incorrect usages of isa. Ironically enough, it flags the below
        # call (which is perfectly valid) as invalid and logs a warning.
        # We disable warnings only for this one call so that the logs don't
        # get flooded with warnings for what is perfectly valid code.
        no warnings;
        my $isa = $pkg_or_obj->SUPER::isa($kind);
        use warnings;
        $Log->exit() if $may_exit;
        return $isa;
    }
}

1;
