#
# Copyright (c) 2001-2012 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Common methods for the StorageAggregate Component and ComponentState
## @author dl-nacl-dev@netapp.com
## @status shared
## @pod here

package NACL::C::_Mixins::StorageAggregate;

use strict;
use warnings;

use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use NATE::Exceptions::Argument qw(:try);
use Params::Validate qw(validate SCALAR);

BEGIN {
    use Exporter qw(import);
    our @EXPORT_OK
        = qw(_aggr_options_copy _aggr_options_map _option_values_to_translate);
    our %EXPORT_TAGS = ( 'all' => \@EXPORT_OK );
}

# This is the list of options that shows up in the 7Mode command "aggr options"
# that has no CMode equivalent and hence can be copied untranslated
sub _aggr_options_copy {
    return [
        qw( nosnap lost_write_protect snapmirrored fs_size_fixed no_delete_log
            metadata_checking root resyncsnaptime )
    ];
}

# This is the map of the volume options from the CMode name to the
# respective 7Mode name
sub _aggr_options_map {
    return {
        'maxraidsize'            => 'raidsize',
        'raid-block-checksum'    => 'raid_cv',
        'raid-zoned-checksum'    => 'raid_zoned',
        'raid-lost-write'        => 'raid_lost_write',
        'ignore-inconsistent'    => 'ignore_inconsistent',
        'snapshot-autodelete'    => 'snapshot_autodelete',
        'ha-policy'              => 'ha_policy',
        'hybrid-enabled'         => 'hybrid_enabled',
        'percent-snapshot-space' => 'percent_snapshot_space',
        'free-space-realloc'     => 'free_space_realloc',
        'raid-block-checksum'    => 'raid_cv',
        'thorough-scrub'         => 'thorough_scrub',
        'max-write-alloc-blocks' => 'max_write_alloc_blocks',
        'raid-zoned-checksum'    => 'raid_zoned',
        'resync-snap-time'       => 'resyncsnaptime',
        'cache-raid-group-size'  => 'cache_raid_group_size'
    };
}

##############################################################################
# Method:    _option_values_to_translate
# Objective: This method returns the options whose values are to be translated
#            (true/false to on/off or vice-versa)
# Inputs:
#         1. mode_of_options
#            This should be either 'SEVEN' or 'CLUSTER'. Options of this mode
#            whose values are to be translated are returned.
# Returns:   An arrayref to the names of the options whose values are to be
#            translated.
###############################################################################

sub _option_values_to_translate {
    $Log->enter() if $may_enter;
    my (%opts) = validate( @_, { mode_of_options => { type => SCALAR } } );
    my $mode = $opts{mode_of_options};
    if ( ( $mode ne 'SEVEN' ) && ( $mode ne 'CLUSTER' ) ) {
        $Log->exit() if $may_exit;
        NATE::Exceptions::Argument->throw( 'Internal error: Mode should be '
                . "either of 'SEVEN' or 'CLUSTER' but it was provided as "
                . "'$mode'" );
    }
    my $sevenmode_values_to_translate = [qw( hybrid_enabled )];

    if ( $mode eq 'SEVEN' ) {
        $Log->exit() if $may_exit;
        return $sevenmode_values_to_translate;
    } else {
        my @cmode_values_to_translate;
        my %aggr_options = reverse %{ _aggr_options_map() };
        foreach my $sevenmode_value (@$sevenmode_values_to_translate) {
            my $cmode_value = $aggr_options{$sevenmode_value}
                || $sevenmode_value;
            push @cmode_values_to_translate, $cmode_value;
        }
        $Log->exit() if $may_exit;
        return [@cmode_values_to_translate];
    } ## end else [ if ( $mode eq 'SEVEN' )
} ## end sub _option_values_to_translate

1;
