# Copyright (c) 2001-2012 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Common methods for the SystemNodeAutosupport Component and ComponentState
## @author dl-nacl-dev@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::C::_Mixins::SystemNodeAutosupport

=head1 DESCRIPTION

This package is a mix-in consisting of methods common to the
SystemNodeAutosupport Component and ComponentState modules. All of the
methods defined in this package are for developers only and should not
be used by external users.

=head1 WHAT TO DO TO ADD SUPPORT FOR NEW FIELDS

When support needs to be added for new fields, the following should
be done:

=over

=item 1) For 7Mode CLI, add the CMode -> 7Mode mapping in L<_c_7_fields_map>.
For CMode ZAPI, add the copy/mapping information in L<_cli_zapi_copy>
and L<_cli_zapi_map> respectively.

=item 2) Check if any value translation is necessary.

=over

=item 2.1

If the CMode values for the fields are enable/disable, then add the
CMode name to L<_enable_disable_fields>.

=item 2.2

If the CMode values for the fields are true/false, then add the
CMode name to L<_true_false_fields>.

=item 2.3

If the field is an array, add the CMode name to L<_array_fields_cmode>.

=item 2.4

If the field accepts units in CMode, add the CMode name to
L<_unit_normalize_fields>.

=item 2.5

If the field is a time field, add the CMode name to L<_time_fields>.

=back

=back

=cut

package NACL::C::_Mixins::SystemNodeAutosupport;

use strict;
use warnings;

use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Params::Validate qw(:all);

use NACL::CS::ComponentState::ZapiSkip qw(make_zapi_skip);
use NACL::CS::ComponentState::ZapiArray qw(make_zapi_array);

BEGIN {
    use Exporter qw(import);
    our @EXPORT_OK = qw(
        _c_7_fields_map
        _enable_disable_fields
        _translate_enable_disable
        _true_false_fields
        _translate_true_false
        _array_fields_cmode
        _array_fields_7mode
        _unit_normalize_fields
        _translate
        _get_7mode_list
        _convert_time
        _time_fields
        _cli_zapi_copy
        _cli_zapi_map
    );
    our %EXPORT_TAGS = ( 'all' => \@EXPORT_OK );
}

=head1 METHODS

=head2 _c_7_fields_map

Returns the CMode -> 7Mode mapping of fields. (returns a hash in list
context and a hash-reference in scalar context)

=cut

sub _c_7_fields_map {
    $Log->enter() if $may_enter;
    my %hash = (
        state        => 'autosupport.enable',
        'mail-hosts' => 'autosupport.mailhost',
        from         => 'autosupport.from',
        to           => 'autosupport.to',
        noteto       => 'autosupport.noteto',
        'partner-address' => 'autosupport.partner.to',
        'max-http-size'   => 'autosupport.max_http_size',
        'transport'       => 'autosupport.support.transport',
        'url'             => 'autosupport.support.url',
        'put-url'         => 'autosupport.support.put_url',
        'proxy-url'       => 'autosupport.support.proxy',
        'nht'             => 'autosupport.nht_data.enable',
        'perf'            => 'autosupport.performance_data.enable',
        'retry-interval'  => 'autosupport.retry.interval',
        'retry-count'     => 'autosupport.retry.count',
        'payload-format'  => 'autosupport.payload_format',
        'periodic-tx-window' => 'autosupport.periodic.tx_window',
        'max-smtp-size'   => 'autosupport.max_smtp_size',
        'validate-digital-certificate' => 'autosupport.validate_digital_certificate',
        'reminder'        => 'autosupport.support.reminder',
        'throttle'        => 'autosupport.throttle',
        'local-collection' => 'autosupport.local_collection',
        'ondemand-state'  => 'autosupport.ondemand.state',
        'ondemand-remotediag-state' => 'autosupport.ondemand.remotediag.state',
        'ondemand-server-url' => 'autosupport.ondemand.server_url',
        'ondemand-polling-interval' => 'autosupport.ondemand.polling_interval',
    );
    $Log->exit() if $may_exit;
    return wantarray ? %hash : {%hash};
}

=head2 _enable_disable_fields

The values of certain fields in CMode are enable/disable, while in 7Mode
they're on/off. This method returns this list of CMode fields. (returns
an array in list context, an array-reference in scalar context)

=cut

sub _enable_disable_fields {
    $Log->enter() if $may_enter;
    my @arr = ( qw(state ondemand-state ondemand-remotediag-state support) );
    $Log->exit() if $may_exit;
    return wantarray ? @arr : [@arr];
}

=head2 _translate_enable_disable

    # Input side
    $pkg_or_obj->_translate_enable_disable(
        options         => \%opts,
        mode_of_options => 'CMode'
    );
    
    # Output side
    $pkg_or_obj->_translate_enable_disable(
        options         => \%opts,
        mode_of_options => '7Mode'
    );
    
If C<mode_of_options> is "CMode" translates the values enable/disable to on/off.
If C<mode_of_options> is "7Mode" translates the values on/off to enable/disable.

=cut

sub _translate_enable_disable {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_translate(
        @args,
        _translation_map => { enable => 'on', disable => 'off' },
        _fields_method   => '_enable_disable_fields'
    );
    $Log->exit() if $may_exit;
}

=head2 _true_false_fields

The values of certain fields in CMode are true/false, while in 7Mode
they're on/off. This method returns this list of CMode fields. (returns
an array in list context, an array-reference in scalar context)

=cut

sub _true_false_fields {
    $Log->enter() if $may_enter;
    my @arr = ( qw(nht perf validate-digital-certificate reminder throttle
        local-collection) );
    $Log->exit() if $may_exit;
    return wantarray ? @arr : [@arr];
}

=head2 _translate_true_false

    # Input side
    $pkg_or_obj->_translate_true_false(
        options         => \%opts,
        mode_of_options => 'CMode'
    );
    
    # Output side
    $pkg_or_obj->_translate_true_false(
        options         => \%opts,
        mode_of_options => '7Mode'
    );
    
If C<mode_of_options> is "CMode" translates the values true/false to on/off.
If C<mode_of_options> is "7Mode" translates the values on/off to true/false.

=cut

sub _translate_true_false {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_translate(
        @args,
        _translation_map => { true => 'on', false => 'off' },
        _fields_method   => '_true_false_fields'
    );
    $Log->exit() if $may_exit;
}

=head2 _array_fields_cmode

Returns the list of CMode fields which are array fields (returns an array
in list context, an array-reference in scalar context)

=cut

sub _array_fields_cmode {
    $Log->enter() if $may_enter;
    my @arr = ( qw(mail-hosts to noteto partner-address) );
    $Log->exit() if $may_exit;
    return wantarray ? @arr : [@arr];
}

=head2 _array_fields_7mode

Returns the list of 7Mode fields which are array fields (returns an array
in list context, an array-reference in scalar context)

=cut

sub _array_fields_7mode {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = $_[0];
    my $arr_cmode  = $pkg_or_obj->_array_fields_cmode();
    my @arr        = $pkg_or_obj->_get_7mode_list(cmode_list => $arr_cmode);
    $Log->exit() if $may_exit;
    return wantarray ? @arr : [@arr];
}

=head2 _unit_normalize_fields

Returns the list of fields (CMode names) which need to be unit normalized
in 7Mode. (returns an array in list context, returns an array-reference in
scalar context)

=cut

sub _unit_normalize_fields {
    $Log->enter() if $may_enter;
    my @arr = ( qw(max-http-size max-smtp-size) );
    $Log->exit() if $may_exit;
    return wantarray ? @arr : [@arr];
}

# Private helper for the two translate routines
sub _translate {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts = validate_with(
        params => \@args,
        spec   => {
            options          => { type => HASHREF },
            mode_of_options  => { type => SCALAR },
            _translation_map => { type => HASHREF },
            _fields_method   => { type => SCALAR },
        }
    );
    my $mode = $opts{mode_of_options};
    if ($mode ne 'CMode' && $mode ne '7Mode') {
        my $me = (caller(1))[3];
        $Log->exit() if $may_exit;
        NATE::Exceptions::Argument->throw('Internal error: Value of '
            . "'mode_of_options' to '$me' should have been CMode or "
            . "7Mode but was provided as '$mode'");
    }
    
    my $fields_method = $opts{_fields_method};
    my @fields = $pkg_or_obj->$fields_method();
    my %translation_map = %{$opts{_translation_map}};
    
    if ($mode eq '7Mode') {
        @fields = $pkg_or_obj->_get_7mode_list(cmode_list => \@fields);
        %translation_map = reverse %translation_map;
    }
    
    my $options = $opts{options};
    foreach my $field (@fields) {
        my $value = $options->{$field};
        if (defined $value) {
            if (exists $translation_map{$value}) {
                $options->{$field} = $translation_map{$value};
            }
        }
    }
    
    $Log->exit() if $may_exit;
}

# Helper that gets the 7Mode list corresponding to a CMode list (looks
# up the mapping and returns the corresponding fields)
sub _get_7mode_list {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts = validate_with(
        params => \@args,
        spec   => { cmode_list => { type => ARRAYREF } }
    );
    my $map = $pkg_or_obj->_c_7_fields_map();
    my $map_subset = {};
    $pkg_or_obj->_hash_copy(
        source => $map,
        target => $map_subset,
        copy   => $opts{cmode_list}
    );
    my @list_7m = values %$map_subset;
    $Log->exit() if $may_exit;
    return wantarray ? @list_7m : [@list_7m];
}

=head2 _convert_time

CMode accepts time in the following format:
<int>h<int>m<int>s

Therefore, values such as:
1h2m
2m3s
300

Are all accepted. However, 7Mode is less flexible. This method converts
the value passed in into the number of seconds.

=cut

sub _convert_time {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts = validate_with(
        params => \@args,
        spec   => { time => { type => SCALAR } }
    );
    my $time = $opts{time};
    
    if ($time =~ /^(?:(\d+)h)?(?:(\d+)m)?(?:(\d+)s?)?$/) {
        my $hours   = $1;
        my $minutes = $2;
        my $seconds = $3;
        my $new_val = 0;
        if ($hours) {
            $new_val += $hours * 60 * 60;
        }
        if ($minutes) {
            $new_val += $minutes * 60;
        }
        if ($seconds) {
            $new_val += $seconds;
        }
        
        $Log->trace("Newval: $new_val");
        $Log->exit() if $may_exit;
        return $new_val;
    } else {
        $Log->exit() if $may_exit;
        NATE::Exceptions::Argument->throw("The value of 'retry-interval' "
            . 'is allowed to be in the form <int>h<int>m<int>s but was '
            . "provided as '$time'");
    }
}

=head2 _time_fields

Returns the time-related fields which are of the format
<int>h<int>m<int>s in CMode.

=cut

sub _time_fields {
    $Log->enter() if $may_enter;
    my @arr = ( qw(retry-interval periodic-tx-window) );
    $Log->exit() if $may_exit;
    return wantarray ? @arr : [@arr];
}

=head2 _cli_zapi_copy

Returns the 

=cut

sub _cli_zapi_copy {
    $Log->enter() if $may_enter;
    my @arr = (
        qw(log-file-size transport retry-interval from periodic-tx-window
            max-smtp-size payload-format put-url
            retry-count max-http-size proxy-url last-subject noteto
            retry-interval to transport ondemand-server-url ondemand-polling-interval
            minimum-private-data-length)
    );
    $Log->exit() if $may_exit;
    return wantarray ? @arr : [@arr]; 
}

sub _cli_zapi_map {
    $Log->enter() if $may_enter;
    $Log->exit() if $may_exit;
    return {
        'hostname-subj'   => 'is-node-in-subject',
        'node'            => 'node-name',
        'nht'             => 'is-nht-data-enabled',
        'perf'            => 'is-perf-data-enabled',
        'throttle'        => 'is-throttle-enabled',
        'support-address' => [
            make_zapi_array('support-address'),
            make_zapi_skip('mail-address')
        ],
        'state'                 => 'is-enabled',
        'remove-private-data' => 'is-private-data-removed',
        'support'               => 'is-support-enabled',
        'last-time'           => 'last-timestamp',
        'trigger-now'         => 'last-trigger',
        'url'                 => 'post-url',
        'reminder'            => 'is-reminder-enabled',
        'mail-hosts'       => [
            make_zapi_array('mail-hosts'),
            make_zapi_skip('string')
        ],
        'local-collection' => 'is-local-collection-enabled',
        'partner-address'  => [
            make_zapi_array('partner-address'),
            make_zapi_skip('mail-address')
        ],
    };
}

1;
