# Copyright (c) 2001-2013 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Common methods for the VolumeFileMove and VolumeFileCopy components
## @author dl-nacl-dev@netapp.com
## @status shared
## @pod here

package NACL::C::_Mixins::VolumeFileCopyMove;

use strict;
use warnings;

BEGIN {
    use Exporter qw(import);
    our @EXPORT_OK = qw(
        _create_obj
        _start_cmode_zapi_helper
        _destroy_cmode_zapi_helper
        _normalize_paths
        _start_zapi_check
        _start_cli_check
        _create_common_code
    );
    our %EXPORT_TAGS = ('all' => \@EXPORT_OK);
}

use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Params::Validate qw(:all);
use NATE::BaseException qw(:try);
use NATE::Exceptions::Argument ();
use NACL::Exceptions::InvalidChoice;
use NACL::ComponentUtils qw(_ontap_ci_validate_spec);
use NACL::APISet::Exceptions::ResponseException;

sub _create_obj {
    $Log->enter();
    my ($pkg, @args) = @_;
    my %opts = validate_with(
        params => \@args,
        spec   => {
            'job-uuid'        => { type => SCALAR },
            'source-paths'    => { type => ARRAYREF },
            command_interface => _ontap_ci_validate_spec(),
        },
        allow_extra => 1
    );

    my (@instances, $index);
    foreach my $path (@{$opts{'source-paths'}}) {
        push @instances, $pkg->new(
            command_interface => $opts{command_interface},
            job_uuid          => $opts{'job-uuid'},
            file_index        => $index++
        );
    }

    $Log->exit();
    return @instances;
}

sub _start_cmode_zapi_helper {
    $Log->enter();
    my ($pkg, @args) = @_;
    my %opts         = $pkg->_validate_backend_opts(@args);

    # CLI accepts size suffixes, while ZAPI allows only an integer
    # value (i.e. number of bytes)
    if (defined $opts{'max-throughput'}) {
        $opts{'max-throughput'} = $pkg->convert_to_bytes($opts{'max-throughput'});
    }
    my $response;
    $pkg->_base_backend_cmode_zapi(
        %opts,
        copy                => [qw(max-throughput source-paths
            destination-paths reference-path hold-quiescence)],
        map                 => {
            'max-cutover-time'           => 'cutover-time',
            'reference-max-cutover-time' => 'reference-cutover-time',
        },
        apiset_response_obj => \$response,
        inapplicable_for_help_xml => [qw(_job_uuid report-completed)],
    );

    ${$opts{_job_uuid}} = $response->get_parsed_output()->[0]{'job-uuid'};
    $Log->exit();
}

sub _destroy_cmode_zapi_helper{
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_base_backend_cmode_zapi(
        @args,
        copy => [qw(file-index job-uuid allow-destroy-file)]
    );
    $Log->exit();
}

sub _normalize_paths {
    $Log->enter();
    my ($pkg, @args) = @_;
    my ($opts) = validate_pos(@args, { type => HASHREF });
    foreach my $normalize_arrayref (qw(source-paths destination-paths)) {
        if (exists $opts->{$normalize_arrayref} && !ref $opts->{$normalize_arrayref}) {
            $opts->{$normalize_arrayref} = [$opts->{$normalize_arrayref}];
        }
    }

    $Log->exit();
}

sub _start_zapi_check {
    $Log->enter();
    my ($pkg, %opts) = @_;    

    if (exists $opts{'scanner-paused'}) {
        $Log->exit();
        NACL::Exceptions::InvalidChoice->throw('Could not use the ZAPI '
            . "implementation of ${pkg}::start because the CLI-specific "
            . "argument 'scanner-paused' was provided"
        );
    }
    
    # If report-completed => true is provided, we can still use the ZAPI even
    # though the option is not supported in ZAPI because that is the default
    # behaviour. However, ZAPI cannot support report-completed => false.
    my $report_completed = $opts{'report-completed'};
    if (defined $report_completed && $report_completed eq 'false') {
        my $msg = 'Could not use the ZAPI '
            . "implementation of ${pkg}::start because 'report-completed' "
            . "=> 'false' was specified in the call and the ZAPI does "
            . "not accept the 'report-completed' parameter";
        $Log->comment($msg);
        $Log->exit();
        NACL::Exceptions::InvalidChoice->throw($msg);
    }
    if ($pkg->isa('NACL::C::VolumeFileMove')) {
        my $msg = 'Could not use the ZAPI '
            . "implementation of ${pkg}::start because the 'hold-quiescence' "
            . "was specified in the call and the ZAPI does not accept"
            . "the 'hold-quiescence' parameter";
        if (exists $opts{'hold-quiescence'}) {
            $Log->comment($msg);
            $Log->exit() if $may_exit;
            NACL::Exceptions::InvalidChoice->throw($msg);
        }
    }

    $Log->exit();
}

# CLI not available for promoted builds
sub _start_cli_check {
    $Log->enter();
    my ($pkg, %opts) = @_;
    if ($opts{command_interface}->version_manager()->is_promoted()) {
        $Log->exit();
        NACL::Exceptions::InvalidChoice->throw('Could not use the CLI '
            . "implementation of ${pkg}::start because the CLI start command "
            . 'is not available on promoted builds'
        );
    }
    $Log->exit();
}

# Code common to the create front-end methods of both components.
sub _create_common_code {
    $Log->enter();
    my ($pkg, @args) = @_;
    my %opts = validate_with(
        params      => \@args,
        spec        => {frontend => {type => SCALAR}},
        allow_extra => 1
    );
    my $frontend = delete $opts{frontend};

    $pkg->_normalize_paths(\%opts);

    my $job_uuid;
    try {
        $pkg->call_on_apiset(
            %opts,
            _job_uuid => \$job_uuid,
            choices   => [
                # Keep the ZAPI choice first here! Reason is that the CLI
                # command is not available on promoted builds, while the ZAPI is,
                # so the CLI command is not available on customer builds. For
                # this reason, we'd like this method to default to the ZAPI
                # implementation.
                # See burt 740994.
                {   method    => '_start_cmode_zapi',
                    interface => 'ZAPI',
                    set       => 'CMode',
                    check     => '_start_zapi_check',
                },
                {   method    => '_start_cmode_cli',
                    interface => 'CLI',
                    set       => 'CMode',
                    check     => '_start_cli_check',
                },
            ],
            frontend => $frontend,
        );
    } catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = $_[0];        
        my $exception_pkg;
        if ($pkg->isa('NACL::C::VolumeFileCopy')) {
            $exception_pkg =
                "NACL::C::Exceptions::VolumeFileCopy::MaximumSimultaneousOperation";
        } else {
            $exception_pkg =
                "NACL::C::Exceptions::VolumeFileMove::MaximumSimultaneousOperation";
        }
        eval "require $exception_pkg";
        unless ($@) {
            $exception_pkg->detect_convert_and_throw(exception => $exception);
        }
        $Log->exit();
        $exception->throw();
    };

    my @instances = $pkg->_create_obj(%opts, 'job-uuid' => $job_uuid);

    $Log->exit();
    return @instances;
}

1;
