#
# Copyright (c) 2001-2010 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary AntivirusRemedy ComponentState Module
## @author dl-nacl-dev@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::CS::AntivirusRemedy

=head1 DESCRIPTION

C<NACL::CS::AntivirusRemedy> is a derived class of
L<NACL::CS::ComponentState::ONTAP|lib-NACL-CS-ComponentState-ONTAP-pm>. It represents the
state of an ONTAP AntivirusRemedy.  A related class is
L<NACL::C::AntivirusRemedy|lib-NACL-C-AntivirusRemedy-pm>, which represents access
to an ONTAP AntivirusRemedy.

=head1 ATTRIBUTES

The individual pieces of data that are part of the  state of the AntivirusRemedy
element are the attributes of the AntivirusRemedy ComponentState.

=over

=item C<< "quarantine_action" >>

Filled in for CMode CLI/ZAPI.

Maps to:

CM ZAPI: option

=item C<< "quarantine_ext" >>

Filled in for CMode CLI/ZAPI.

Maps to:

CM ZAPI: extension

=item C<< "failed_repair_action" >>

Filled in for CMode CLI/ZAPI.

Maps to:

CM ZAPI: repair

=item C<< "quarantine_dir" >>

Filled in for CMode CLI/ZAPI.

Maps to:

CM ZAPI: directory

=item C<< "action" >>

Filled in for CMode CLI/ZAPI.

Maps to:

CM ZAPI: action

=back

=cut

package NACL::CS::AntivirusRemedy;

use strict;
use warnings;
use Params::Validate qw(validate);
use NACL::ComponentUtils qw(_dump_one);
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Data::Dumper;

use NACL::Exceptions::NoElementsFound qw(:try);

use base 'NACL::CS::ComponentState::ONTAP';

use Class::MethodMaker [
    scalar => 'quarantine_action',
    scalar => 'quarantine_ext',
    scalar => 'failed_repair_action',
    scalar => 'quarantine_dir',
    scalar => 'action',
];

=head1 METHODS

=head2 fetch

    my $AntivirusRemedy_state   = NACL::CS::AntivirusRemedy->fetch(command_interface => $ci, ...);
	my @AntivirusRemedy_states = NACL::CS::AntivirusRemedy->fetch(command_interface => $ci, ...);

(Class method)
Discovers which elements are present and returns their state in ComponentState
objects. Called in scalar context it returns only one state object, in list
context it returns all state objects.
See L<NACL::CS::ComponentState|lib-NACL-CS-ComponentState-pm/fetch> for a
more detailed description along with a complete explanation of the options
it accepts.

Supports CMode CLI/ZAPI.

Invokes "av-get-remedy-info" API for CMode ZAPI.

=over

=item Exceptions

=over

=item C<NACL::Exceptions::NoElementsFound>

When there are no elements matching the query specified or elements of that type
doesn't exist, then this exception will be thrown.

=back

=back

=cut

sub fetch {
    $Log->enter() if $may_enter;
    my $pkg        = shift;
    my @state_objs = $pkg->SUPER::fetch(
        @_,
        choices => [
            {   method    => '_fetch_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_fetch_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
        ],
        exception_text => 'No matching antivirus remedy(s) found',
        is_singleton   => 1,
        show_cmd       => 'antivirus remedy show',
    );

    $Log->exit() if $may_exit;
    return wantarray ? @state_objs : $state_objs[0];
} ## end sub fetch

sub _fetch_cmode_cli {
    $Log->enter() if $may_enter;
    my $pkg        = shift;
    my @state_objs = $pkg->SUPER::_fetch_cmode_cli(
        @_,
        api          => 'antivirus_remedy_show',
        is_singleton => 1
    );
    $Log->exit() if $may_exit;
    return @state_objs;
} ## end sub _fetch_cmode_cli

sub _fetch_cmode_zapi {
    $Log->enter() if $may_enter;
    my $pkg    = shift;
    my %opts   = validate @_, $pkg->_fetch_backend_validate_spec();
    my $apiset = $opts{apiset};
    my $ci_response;

    # find the antivirus present.
    $ci_response = $apiset->av_get_remedy_info();

    my $ci_output = $ci_response->get_parsed_output();
    $Log->debug(Dumper($ci_output));
    my @state_objs;
    my $state_fields = $pkg->_zapi_hash_copy(
        source => $ci_output->[0],
        copy   => [qw(action)],
        map    => [
            "extension" => "quarantine-ext",
            "option"    => "quarantine-action",
            "repair"    => "failed-repair-action",
            "directory" => "quarantine-dir",
        ],
    );
    my $obj = $pkg->new();
    $obj->_set_fields(row => $state_fields);
    push @state_objs, $obj;
    $Log->exit() if $may_exit;
    return @state_objs;
} ## end sub _fetch_cmode_zapi

1;
