# Copyright (c) 2001-2012 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary ClusterPeerHealth ComponentState Module
## @author dl-nacl-dev@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::CS::ClusterPeerHealth

=head1 DESCRIPTION

C<NACL::CS::ClusterPeerHealth> is a derived class of
L<NACL::CS::ComponentState::ONTAP|lib-NACL-CS-ComponentState-ONTAP-pm>. It represents the
state of an ONTAP ClusterPeerHealth.  A related class is
L<NACL::C::ClusterPeerHealth|lib-NACL-C-ClusterPeerHealth-pm>, which represents access
to an ONTAP ClusterPeerHealth.

=head1 ATTRIBUTES

The individual pieces of data that are part of the  state of the ClusterPeerHealth
element are the attributes of the ClusterPeerHealth ComponentState.

=over

=item C<< "originating_node" >>

Filled in for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

For 'requested_fields', 'filter' and Output mapping:

 <cluster-peer-health-info>
    <originating-node>$value</originating-node>
 </cluster-peer-health-info>

=item C<< "bypass_cache" >>

Filled in for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

For 'requested_fields', 'filter' and Output mapping:

 <cluster-peer-health-info>
    <bypass-cache>$value</bypass-cache>
 </cluster-peer-health-info>

=item C<< "data_ping" >>

Filled in for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

For 'requested_fields', 'filter' and Output mapping:

 <cluster-peer-health-info>
    <data-ping>$value</data-ping>
 </cluster-peer-health-info>

=item C<< "availability" >>

Filled in for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

For 'requested_fields', 'filter' and Output mapping:

 <cluster-peer-health-info>
    <is-destination-node-available>$value</is-destination-node-available>
 </cluster-peer-health-info>

=item C<< "node_health" >>

Filled in for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

For 'requested_fields', 'filter' and Output mapping:

 <cluster-peer-health-info>
    <is-node-healthy>$value</is-node-healthy>
 </cluster-peer-health-info>

=item C<< "destination_node" >>

Filled in for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

For 'requested_fields', 'filter' and Output mapping:

 <cluster-peer-health-info>
     <destination-node>$value</destination-node>
 </cluster-peer-health-info>

=item C<< "destination_cluster" >>

Filled in for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

For 'requested_fields', 'filter' and Output mapping:

 <cluster-peer-health-info>
    <destination-cluster>$value</destination-cluster>
 </cluster-peer-health-info>

=item C<< "cluster_health" >>

Filled in for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

For 'requested_fields', 'filter' and Output mapping:

 <cluster-peer-health-info>
    <is-cluster-healthy>$value</is-cluster-healthy>
 </cluster-peer-health-info>

=item C<< icmp_ping >>

Filled in for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

For 'requested_fields', 'filter' and Output mapping:

 <cluster-peer-health-info>
    <icmp-ping>$value</icmp-ping>
 </cluster-peer-health-info>

=item C<< destination_cluster_uuid >>

Filled in for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

For 'requested_fields', 'filter' and Output mapping:

 <cluster-peer-health-info>
    <destination-cluster-uuid>$value</destination-cluster-uuid>
 </cluster-peer-health-info>

=item C<< last_updated >>

Last Update Time
Filled in for CMode CLI/ZAPI.

Mapping for CMode ZAPI: <cluster-peer-health-info last-updated="$value" />

=back

=cut

package NACL::CS::ClusterPeerHealth;

use strict;
use warnings;
use Params::Validate qw(validate);
use NACL::ComponentUtils qw(_dump_one Dumper);
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();

use NACL::Exceptions::NoElementsFound qw(:try);
use NACL::ComponentUtils qw (enumerate_month);
use base 'NACL::CS::ComponentState::ONTAP';

use Class::MethodMaker [
    scalar => 'originating_node',
    scalar => 'bypass_cache',
    scalar => 'destination_cluster_UUID',
    scalar => 'data_ping',
    scalar => 'availability',
    scalar => 'ICMP_ping',
    scalar => 'node_health',
    scalar => 'destination_node',
    scalar => 'destination_cluster',
    scalar => 'cluster_health',
    scalar => 'icmp_ping',
    scalar => 'destination_cluster_uuid',
    scalar => 'last_updated',
];

=head1 METHODS

=head2 fetch

    my $ClusterPeerHealth_state   = NACL::CS::ClusterPeerHealth->fetch(command_interface => $ci, ...);
	my @ClusterPeerHealth_states = NACL::CS::ClusterPeerHealth->fetch(command_interface => $ci, ...);

(Class method)
Discovers which elements are present and returns their state in ComponentState
objects. Called in scalar context it returns only one state object, in list
context it returns all state objects.
See L<NACL::CS::ComponentState|lib-NACL-CS-ComponentState-pm/fetch> for a
more detailed description along with a complete explanation of the options
it accepts.

Supports CMode CLI/ZAPI.

Invokes "cluster-peer-health-info-get-iter" API for CMode ZAPI.

=over

=item Exceptions

=over

=item C<NACL::Exceptions::NoElementsFound>

When there are no elements matching the query specified or elements of that type
doesn't exist, then this exception will be thrown.

=back

=back

=cut

sub fetch {
    $Log->enter() if $may_enter;
    my $pkg        = shift;
    my @state_objs = $pkg->SUPER::fetch(
        @_,
        choices => [
            {   method    => '_fetch_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_fetch_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
        ],
        exception_text => 'No matching cluster peer health(s) found',
        show_cmd       => 'cluster peer health show',
    );

    $Log->exit() if $may_exit;
    return wantarray ? @state_objs : $state_objs[0];
} ## end sub fetch

sub _fetch_cmode_cli {
    $Log->enter() if $may_enter;
    my $pkg        = shift;
    my @state_objs = $pkg->SUPER::_fetch_cmode_cli(@_,
        api => 'cluster_peer_health_show');
    $Log->exit() if $may_exit;
    return @state_objs;
} ## end sub _fetch_cmode_cli

sub _fetch_cmode_zapi {
    $Log->enter() if $may_enter;
    my $pkg        = shift;
    my @state_objs = $pkg->SUPER::_fetch_cmode_zapi(
        @_,
        api  => 'cluster_peer_health_info_get_iter',
        copy => [
            qw(originating-node destination-cluster destination-node
               destination-cluster-uuid data-ping icmp-ping bypass-cache
               last-updated )
        ],
        map => {
            "node-health"    => ['is-node-healthy'],
            "cluster-health" => ['is-cluster-healthy'],
            "availability"   => ['is-destination-node-available'],
        }
    );
    return @state_objs;
} ## end sub _fetch_cmode_zapi

sub _update_state_objs_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->SUPER::_update_state_objs_cmode_zapi(
        @args,
        zapi_field_translations => {
            timestamp_to_mmddyyyy => [
                qw(last-updated)
            ],
        },
    );
    $Log->exit() if $may_exit;
}
1;
