#
# Copyright (c) 2001-2012 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary EventDestination ComponentState Module
## @author dl-nacl-dev@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::CS::EventDestination

=head1 DESCRIPTION

C<NACL::CS::EventDestination> is a derived class of
L<NACL::CS::ComponentState::ONTAP|lib-NACL-CS-ComponentState-ONTAP-pm>. It represents the
state of an ONTAP EventDestination.  A related class is
L<NACL::C::EventDestination|lib-NACL-C-EventDestination-pm>, which represents access
to an ONTAP EventDestination.

=head1 ATTRIBUTES

The individual pieces of data that are part of the  state of the EventDestination
element are the attributes of the EventDestination ComponentState.

=over

=item C<< "snmp" >>

(Array)

Filled in for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

For "requested_fields", "filter" and Output mapping:

 <ems-destination-info>
    <snmp>
         <remote-ip>@values</remote-ip>
    </snmp>
 </ems-destination-info>

=item C<< "syslog_facility" >>

Filled in for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

For "requested_fields", "filter" and Output mapping:

 <ems-destination-info>
    <syslog-facility>$value</syslog-facility>
 </ems-destination-info>

=item C<< "syslog" >>

(Array)

Filled in for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

For "requested_fields", "filter" and Output mapping:

 <ems-destination-info>
    <syslog>
         <remote-ip>@values</remote-ip>
    </syslog>
 </ems-destination-info>

=item C<< "hide_parameters" >>

Filled in for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

For "requested_fields", "filter" and Output mapping:

 <ems-destination-info>
    <hide-parameters>$value</hide-parameters>
 </ems-destination-info>

=item C<< "snmp_community" >>

Filled in for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

For "requested_fields", "filter" and Output mapping:

 <ems-destination-info>
    <snmp-community>$value</snmp-community>
 </ems-destination-info>

=item C<< "name" >>

Filled in for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

For "requested_fields", "filter" and Output mapping:

 <ems-destination-info>
    <name>$value</name>
 </ems-destination-info>

=item C<< "mail" >>

(Array)

Filled in for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

For "requested_fields", "filter" and Output mapping:

 <ems-destination-info>
    <mail>
         <mail-address>@values</mail-address>
    </mail>
 </ems-destination-info>


=back

=cut

package NACL::CS::EventDestination;

use strict;
use warnings;
use Params::Validate qw(validate);
use NACL::ComponentUtils qw(_dump_one);
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Data::Dumper;

use NACL::Exceptions::NoElementsFound qw(:try);

use base 'NACL::CS::ComponentState::ONTAP';
use NACL::CS::ComponentState::ZapiSkip qw(make_zapi_skip);
use NACL::CS::ComponentState::ZapiArray qw(make_zapi_array);

use Class::MethodMaker [
    array  => 'snmp',
    scalar => 'syslog_facility',
    array  => 'syslog',
    scalar => 'hide_parameters',
    scalar => 'snmp_community',
    scalar => 'name',
    array  => 'mail',
];

=head1 METHODS

=head2 fetch

    my $EventDestination_state   = NACL::CS::EventDestination->fetch(command_interface => $ci, ...);
	my @EventDestination_states = NACL::CS::EventDestination->fetch(command_interface => $ci, ...);

(Class method)
Discovers which elements are present and returns their state in ComponentState
objects. Called in scalar context it returns only one state object, in list
context it returns all state objects.
See L<NACL::CS::ComponentState|lib-NACL-CS-ComponentState-pm/fetch> for a
more detailed description along with a complete explanation of the options
it accepts.

Supports CMode CLI/ZAPI.

Invokes "ems-destination-get-iter" API for CMode ZAPI

=over

=item Exceptions

=over

=item C<NACL::Exceptions::NoElementsFound>

When there are no elements matching the query specified or elements of that type
doesn't exist, then this exception will be thrown.

=back

=back

=cut

sub fetch {
    $Log->enter() if $may_enter;
    my $pkg        = shift;
    my @state_objs = $pkg->SUPER::fetch(
        @_,
        choices => [
            {   method    => '_fetch_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_fetch_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
        ],
	show_cmd       => 'event destination show',
        exception_text => 'No matching event destination(s) found'
    );

    $Log->exit() if $may_exit;
    return wantarray ? @state_objs : $state_objs[0];
} ## end sub fetch

sub _fetch_cmode_cli {
    $Log->enter() if $may_enter;
    my $pkg = shift;
    my @state_objs
        = $pkg->SUPER::_fetch_cmode_cli(@_, api => 'event_destination_show');
    $Log->exit() if $may_exit;
    return @state_objs;
} ## end sub _fetch_cmode_cli

sub _fetch_cmode_zapi {
    my $pkg    = shift;
    my @states = $pkg->SUPER::_fetch_cmode_zapi(
        @_,
        api  => "ems_destination_get_iter",
        copy => [qw( name syslog-facility hide-parameters snmp-community )],
        map  => {
            "snmp" => [make_zapi_array('snmp'), make_zapi_skip('remote-ip')],
            "syslog" =>
                [make_zapi_array('syslog'), make_zapi_skip('remote-ip')],
            "mail" =>
                [make_zapi_array('mail'), make_zapi_skip('mail-address')],
        },
    );

    return @states;
} ## end sub _fetch_cmode_zapi

1;
