#
# Copyright (c) 2001-2012 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary JobQueue componentState Module
## @author valsala@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::CS::JobQueue

=head1 DESCRIPTION

C<NACL::CS::JobQueue> is a derived class of
L<NACL::CS::ComponentState::ONTAP|lib-NACL-CS-ComponentState-ONTAP-pm>. It represents the
state of an ONTAP JobQueue. 

=head1 ATTRIBUTES

The individual pieces of data that are part of the  state of the JobQueue
element are the attributes of the JobQueue ComponentState.

=over

=item C<< "queue" >>

Filled in for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

For "filter", "requested_fields" and Output mapping:

 <job-queue-info>
     <job-queue>$value</job-queue>
 </job-queue-info>

=item C<< "id" >>

Filled in for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

For "filter", "requested_fields" and Output mapping:

 <job-queue-info>
     <job-uuid>$value</job-uuid>
 </job-queue-info>

=item C<< "runtime" >>

Filled in for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

For "filter", "requested_fields" and Output mapping:

 <job-queue-info>
     <job-scheduled-time>$value</job-scheduled-time>
 </job-queue-info>

item C<< "priority" >>

Filled in for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

For "filter", "requested_fields" and Output mapping:

 <job-queue-info>
     <job-priority>$value</job-priority>
 </job-queue-info>

=back

=cut

package NACL::CS::JobQueue;

use strict;
use warnings;
use Params::Validate qw(validate);
use NACL::ComponentUtils qw(_dump_one);
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use NACL::ComponentUtils qw(_optional_scalars Dumper);

use NACL::Exceptions::NoElementsFound qw(:try);

use base 'NACL::CS::ComponentState::ONTAP';

use Class::MethodMaker [
    scalar => 'queue',
    scalar => 'id',
    scalar => 'runtime',
    scalar => 'priority',

];

=head1 METHODS

=head2 fetch

    my $JobQueue_state   = NACL::CS::JobQueue->fetch(command_interface => $ci, ...);
	my @JobQueue_states = NACL::CS::JobQueue->fetch(command_interface => $ci, ...);

(Class method)
Discovers which elements are present and returns their state in ComponentState
objects. Called in scalar context it returns only one state object, in list
context it returns all state objects.
See L<NACL::CS::ComponentState|lib-NACL-CS-ComponentState-pm/fetch> for a
more detailed description along with a complete explanation of the options
it accepts.

Supports CMode CLI/ZAPI.

Invokes "job-queue-get-iter" API for CMode ZAPI.

=over

=item Exceptions

=over

=item C<NACL::Exceptions::NoElementsFound>

When there are no elements matching the query specified or elements of that type
doesn't exist, then this exception will be thrown.

=back

=back

=cut

sub fetch {
    $Log->enter() if $may_enter;
    my $pkg        = shift;
    my @state_objs = $pkg->SUPER::fetch(
        @_,
        choices => [
            {   method    => '_fetch_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_fetch_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
        ],
	show_cmd       => 'job queue show',
        exception_text => 'No matching job queue(s) found'
    );

    $Log->exit() if $may_exit;
    return wantarray ? @state_objs : $state_objs[0];
}

sub _fetch_cmode_cli {
    $Log->enter() if $may_enter;
    my $pkg = shift;
    my @state_objs
        = $pkg->SUPER::_fetch_cmode_cli(@_, api => 'job_queue_show');
    $Log->exit() if $may_exit;
    return @state_objs;
} ## end sub _fetch_cmode_cli

sub _fetch_cmode_zapi {
    $Log->enter() if $may_enter;
    my $pkg        = shift;
    my @state_objs = $pkg->SUPER::_fetch_cmode_zapi(
        @_,
        api => 'job-queue-get-iter',
        map => {
            "queue"    => ['job-queue'],
            "id"       => ['job-uuid'],
            "runtime"  => ['job-scheduled-time'],
            "priority" => ['job-priority'],
        },

    );
    return @state_objs;
} ## end sub _fetch_cmode_zapi
1;
