#
# Copyright (c) 2012-2013 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#
#
## @summary Port ComponentState Module
## @author gupendra@netapp.com, dl-nacl-dev@netapp.com
## @status shared
## @pod here
################################################################################

=head1 NAME

NACL::CS::Switch::Port

=head1 DESCRIPTION

C<NACL::CS::Switch::Port> is a derived class of
L<NACL::CS::ComponentState::Switch|lib-NACL-CS-ComponentState-Switch-pm>.  It represents the
state of Port element.  A related class is
L<NACL::C::Switch::Port|lib-NACL-C-Switch-Port-pm>, which represents access
to a Port element.

=head1 ATTRIBUTES

The individual pieces of data that are part of the state of the Port element
are the individual attributes of the Port ComponentState

=over

=item C<< 'port'=>$string >>

(Represents port number in 3 or 1/1 form)

=item C<< portindex >>

=item C<< portname >>

=item C<< protocol >>

=item C<< portifid >>

=item C<< portstate >>

=item C<< portspeed >>

=item C<< porttype >>

#Filled in for Cisco

=item C<< port_capability_mode >>

(Array)List of port capability modes with min speed,max speed

#Filled in for Cisco

=item C<< operational_information >>

(Array)List of operational information of  port

#Filled in for Cisco

=item C<< min_speed >>

#Filled in for Cisco

=item C<< max_speed >>

#Filled in for Cisco

=item C<< fcid >>

=back

=cut

###################
# Package
package NACL::CS::Switch::Port;

###################
# Everytime
use strict;
use warnings;

###################
# Module includes
use Params::Validate qw (validate);
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use NACL::ComponentUtils qw (_optional_scalars  _dump_one);
use NATE::BaseException;
use NACL::Exceptions::NoElementsFound qw(:try);
use NACL::APISet::Exceptions::InvalidParamValueException qw(:try );

use base 'NACL::CS::ComponentState::Switch';

use Class::MethodMaker [
    new    => [ '-hash', 'new' ],
    scalar => "port",
    scalar => "portindex",
    scalar => "portname",
    scalar => "protocol",
    scalar => "portifid",
    scalar => "portstate",
    scalar => "portspeed",
    scalar => "porttype",

    #filled in for cisco
    array  => "port_capability_mode",
    array  => "operational_information",
    scalar => "min_speed",
    scalar => "max_speed",
    scalar => "fcid",

    #filled in for qlogic
    scalar => "adminstate",
    scalar => "operationalstate",
    scalar =>  "loginstatus",
    scalar => "configtype",
    scalar => "runningtype",
    scalar => "linkstate",
    scalar => "linkspeed",

];

=head1 METHODS

=head2 fetch

    my $port = NACL::CS::Switch::Port->fetch(command_interface=>$ci,...);
    my @port = NACL::CS::Switch::Port->fetch(command_interface=>$ci,...);

see L<NACL::CS::ComponentState|lib-NACL-CS-ComponentState-pm>

=cut

sub fetch {
    my $pkg        = shift;
    my @state_objs = $pkg->SUPER::fetch(
        @_,
        choices => [
            {
                method    => "_fetch_brocadefc_cli",
                interface => "CLI",
                set       => "BrocadeFC"
            },
            {
                method    => "_fetch_ciscofc_cli",
                interface => "CLI",
                set       => "CiscoFC"
            },
            {
                method    => "_fetch_qlogicfc_cli",
                interface => "CLI",
                set       => "QlogicFC"
            },

        ],
        exception_text => 'No matching elements found'
    );
    return wantarray ? @state_objs : $state_objs[0];
}

sub _fetch_brocadefc_cli {
    my $pkg = shift;
    my %opts = validate @_, $pkg->_fetch_backend_validate_spec();
    my %api_opts;
    my @state_objs;
    my $apiset                = delete $opts{apiset};
    my $command_interface     = delete $opts{command_interface};
    my %copy_filter           = %{ $opts{filter} };
    my @copy_requested_fields = @{ $opts{requested_fields} };
    my $deleted_filter        = $pkg->_remove_relational_regex_filters(
        filter           => \%copy_filter,
        requested_fields => \@copy_requested_fields
    );
    my $filter           = \%copy_filter;
    my $requested_fields = \@copy_requested_fields;
    my $port;

    if ( defined( $filter->{port} ) ) {
        $port = $filter->{port};
    }
    else {
        NATE::BaseException->throw("No port number was passed in filter");
    }
    my $response = $apiset->portshow( 'port' => $port );
    my $output = $response->get_parsed_output();

    foreach my $list_output ( @{$output} ) {
        my $target = $pkg->_hash_copy(
            source => $list_output,
            copy =>
              [qw(portstate portspeed porttype protocol portname portifid)],
            map => { 'portindex' => 'port' },
        );
        my $obj = $pkg->new( command_interface => $command_interface, );
        $obj->_set_fields( row => $target );
        push @state_objs, $obj;
    } ## end foreach my $list_output (@$output)

    return @state_objs;

}    #end sub _fetch_brocadefc_cli

sub _fetch_ciscofc_cli {
    my $pkg = shift;
    my %opts = validate @_, $pkg->_fetch_backend_validate_spec();
    my %api_opts;
    my @state_objs;
    my $apiset                = delete $opts{apiset};
    my $command_interface     = delete $opts{command_interface};
    my %copy_filter           = %{ $opts{filter} };
    my @copy_requested_fields = @{ $opts{requested_fields} };
    my $deleted_filter        = $pkg->_remove_relational_regex_filters(
        filter           => \%copy_filter,
        requested_fields => \@copy_requested_fields
    );
    my $filter           = \%copy_filter;
    my $requested_fields = \@copy_requested_fields;
    my $port;

    if ( defined( $filter->{port} ) ) {
        $port = $filter->{port};
    }
    else {
        NATE::BaseException->throw(
            "No port number was passed in filter");
    }
    my $response = $apiset->show_port(
        "internal-info"            => 1,
        "show-data-structure-info" => 1,
        "interface"                => 1,
        "fiber-channel"            => "$port"
    );
    my $output = $response->get_parsed_output();
    foreach my $row (@$output) {
        my $target = $pkg->_hash_copy(
            source => $row,
            copy =>
              [qw(operational_information port port_capability_mode sfp fcid)],
            map => {
                'port_type' => 'porttype',

            },
        );
        my $state_value      = $row->{'operational_information'}->[0]{'state'};
        my $speed_value      = $row->{'operational_information'}->[0]{'speed'};
        my %final_attributes = (
            %$target,
            'portstate' => $state_value,
            'portspeed' => $speed_value,
        );
        my $obj = $pkg->new( command_interface => $command_interface );
        $obj->_set_fields( row => \%final_attributes );
        push @state_objs, $obj;
    }
    $Log->exit() if $may_exit;
    return @state_objs;
}    #end sub _fetch_ciscofc_cli

sub _fetch_qlogicfc_cli {
    $Log->enter() if $may_enter;
    my ($pkg, @opts) = @_;
    my %opts   = validate @opts, $pkg->_fetch_backend_validate_spec();
    my $apiset = $opts{apiset};
    my @state_objs;
    my %show_args ;
    my ($output, $response, $caught_exception);
    my %filter = %{$opts{filter}};
    my @requested_fields = @{$opts{requested_fields}};
    $pkg->_remove_relational_regex_filters(
        filter           => \%filter,
        requested_fields => \@requested_fields
    );
    my $requested_fields = $opts{requested_fields};
    my $filter           = $opts{filter};
    my %req_field_filter
        = ( requested_fields => $requested_fields, filter => $filter );

    if ( defined( $filter->{'port'})) {
        $show_args{'port-num'} = $filter->{'port'};
    }
    try {
        $response = $apiset->show_port(%show_args);
    }
    catch NACL::APISet::Exceptions::InvalidParamValueException with {
            $caught_exception = 1;
    };
    if ($caught_exception) {
        $Log->exit() if $may_exit;
        return;
    }

    $output = $response->get_parsed_output();
    foreach my $list_output (@$output) {
        my $target = $pkg->_hash_copy(
            source => $list_output,
            copy   => [qw(port adminstate operationalstate loginstatus
                configtype runningtype linkstate linkspeed)],
            map => { 'port_number' => 'port' },
        );
        my $obj = $pkg->new( command_interface => $opts{command_interface} );
        $obj->_set_fields( row => $target );
        push @state_objs, $obj;
    }

    $Log->exit() if $may_exit;
    return @state_objs;
}  #end sub _fetch_qlogicfc_cli
1;
