#
#
# Copyright (c) 2011 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#
#
## @summary PortChannel ComponentState Module
## @author rawat@netapp.com, dl-nacl-dev@netapp.com
## @status shared
## @pod here
################################################################################

=head1 NAME

NACL::CS::Switch::PortChannel

=head1 DESCRIPTION

C<NACL::CS::Switch::PortChannel> is a derived class of
L<NACL::CS::ComponentState::Switch|lib-NACL-CS-ComponentState-Switch-pm>.  It represents the
state of PortChannel element.  A related class is
L<NACL::C::Switch::PortChannel|lib-NACL-C-Switch-PortChannel-pm>, which represents access
to a PortChannel element.

=head1 ATTRIBUTES

The individual pieces of data that are part of the state of the PortChannel 
element are the individual attributes of the PortChannel ComponentState

=over

=item C<< port_channel_group >>

(Required) number of the port-channel on which action will be performed

=item C<< ports >>

information about the physical ports in this port-channel.

=item C<< ports_up >>

number of ports up in this port-channel. only available in Nexus.

=item C<< last_membership_update >>

status of last membership update. only available in Nexus.

=item C<< last_bundled_member >>

last port to join this group. only available in Nexus.

=item C<< age_of_the_port_channel >>

only available in Nexus.

=item C<< first_operational_port >>

only available in Nexus.

=item C<< time_since_last_bundle >>

only available in Nexus.

=item C<< ports_in_total >>

Total number of ports in this port-channel. only available in Nexus.

=item C<< running_conf >>

running configuration of this port-channel. only available in IOS.

=item C<< flags >>

=item C<< no_of_channel_group_in_use >>

only available in IOS. number of channel-groups in use.

=item C<< no_of_aggregators >>

only available in IOS.

=item C<< protocol >>

only available in IOS.

=item C<< port_channel >>

only available in IOS.

=item C<< interface_num >>

=item C<< interface_type >>

=back

=cut

###################
# Package
package NACL::CS::Switch::PortChannel;

###################
# Everytime
use strict;
use warnings;

###################
# Module includes
use Params::Validate qw (validate);
use NACL::ComponentUtils qw (_dump_one);
use NACL::Exceptions::NoElementsFound qw(:try);
use NACL::APISet::Exceptions::InvalidParamValueException;

use base 'NACL::CS::ComponentState::Switch';

# define constant
use constant INTERFACE_TYPE => "port-channel";

use Class::MethodMaker [    

    # Fields common to both IOS and Nexus OS
    scalar => "interface_num",
    scalar => "interface_type",
    array  => "ports",
    scalar => "port_channel_group",

    # Fields for Nexus OS
    scalar => "ports_up",
    scalar => "last_membership_update",
    scalar => "last_bundled_member",
    scalar => "age_of_the_port_channel",
    scalar => "first_operational_port",
    scalar => "time_since_last_bundle",
    scalar => "ports_in_total",

    # Fields for IOS
    array  => "running_conf",
    hash   => "flags",
    scalar => "no_of_channel_group_in_use",
    scalar => "no_of_aggregators",
    scalar => "protocol",
    scalar => "port_channel",
];

=head1 METHODS

=head2 fetch

    my $port_channel_state = NACL::CS::Switch::PortChannel->fetch(command_interface=>$ci,...);
    my @port_channel_states = NACL::CS::Switch::PortChannel->fetch(command_interface=>$ci,...);

see L<NACL::CS::ComponentState|lib-NACL-CS-ComponentState-pm>

=cut

sub fetch {
    my $pkg        = shift;
    my @state_objs = $pkg->SUPER::fetch(
        @_,
        choices => [
            {   method    => "_fetch_cisconetwork_cli",
                interface => "CLI",
                set       => "CiscoNetwork"
            },
        ],
        exception_text => 'No matching elements found'
    );
    return wantarray ? @state_objs : $state_objs[0];
}

sub _fetch_cisconetwork_cli {
    my $pkg    = shift;
    my %opts   = validate @_, $pkg->_fetch_backend_validate_spec();
    my $apiset = $opts{apiset};
    my $filter = $opts{filter};

    my @state_objs;

    my $hosttype = $opts{command_interface}->hostrec()->hosttype();

    my ($output, $response, $obj, $caught_exception);

    $obj = $pkg->new( command_interface => $opts{command_interface} );

    if($hosttype =~ /nxos/i) {
        try {
            $response = $apiset->show_port_channel(
                'port-channel' => $filter->{'port-channel-group'},
                'field' => "database",
                'entity'  => "interface",
            );
        } catch NACL::APISet::Exceptions::InvalidParamValueException with {
            $caught_exception = 1;
        };
        return if $caught_exception;
        
        $output = $response->get_parsed_output();
        foreach my $row (@$output) {
            $obj->_set_fields( row => $row );
        }
    } elsif($hosttype =~ /ios/i) {
        $pkg->_exit_config_modes_cisconetwork(apiset => $apiset);
        if ( $pkg->_want_any_field_of(
                                      requested_fields     => $opts{requested_fields},
                                      filter               => $filter,
                                      fields_filled_by_api => [ qw(port_channel protocol group ports no_of_channel_group_in_use no_of_aggregators) ],
                                    )
        ) {
            try {
                $response = $apiset->show_etherchannel(
                    'etherchannel-num' => $filter->{'port-channel-group'},
                    summary => 1,
                );
            } catch NACL::APISet::Exceptions::InvalidParamValueException with {
                $caught_exception = 1;
            }; 
            
            if ($caught_exception) {
                $apiset->configure(subsystem => "terminal");
                $apiset->interface(
                    'interface-num'  => $opts{filter}{'port-channel-group'},
                    'interface-type' => INTERFACE_TYPE,
                );
                return ;
            }
            
            $output = $response->get_parsed_output();
            my @ports = @{$output->[0]->{ports}};
            my $i = 0;
            foreach my $port (@ports) {
                $output->[0]->{ports}->[$i++]->{interface_type} = $pkg->_expand_interface_type($port->{interface_type});
            }
            
            foreach my $row (@$output) {
                $row->{'port_channel_group'} = delete $row->{group};
                $obj->_set_fields( row => $row );
            }
        }

        # Get the details of running config for this interface
        if ($pkg->_want_any_field_of(
                                requested_fields     => $opts{requested_fields},
                                filter               => $filter,
                                fields_filled_by_api => [ qw(running_conf) ],
            )
        ) {
	        try {
	            $response = $apiset->show_running_config( 
	                'entity-name' => $opts{filter}{'port-channel-group'},
	                'entity-type' => INTERFACE_TYPE,
	                entity => "interface",
	            );
	        } catch NACL::APISet::Exceptions::InvalidParamValueException with {
	            $caught_exception = 1;
	        }; 
	        
	        if ($caught_exception) {
	            $apiset->configure(subsystem => "terminal");
	            $apiset->interface(
	                'interface-num'  => $opts{filter}{'port-channel-group'},
	                'interface-type' => INTERFACE_TYPE,
	            );
	            return ;
	        }
	        
	        $output = $response->get_parsed_output();
            $output->[0]->{interface_type} = $pkg->_expand_interface_type($output->[0]->{interface_type});
	        foreach my $row (@$output) {
                $row->{'port_channel_group'} = $row->{'interface_num'};
	            $obj->_set_fields( row => $row );
	        }
        }

        $apiset->configure(subsystem => "terminal");
        $apiset->interface(
            'interface-num'  => $opts{filter}{'port-channel-group'},
            'interface-type' => INTERFACE_TYPE,
        );

    }

    push @state_objs, $obj;

    return @state_objs;
}

1;
