#
# Copyright (c) 2012 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#
#
## @summary PortStats ComponentState Module
## @author pbala@netapp.com, dl-nacl-dev@netapp.com
## @status shared
## @pod here
################################################################################

=head1 NAME

NACL::CS::Switch::PortStats

=head1 DESCRIPTION

C<NACL::CS::Switch::PortStats> is a derived class of
L<NACL::CS::ComponentState::Switch|lib-NACL-CS-ComponentState-Switch-pm>.  It represents the
statistics details of a port.  A related class is
L<NACL::C::Switch::Port|lib-NACL-C-Switch-Port-pm>, which represents access
to a Port element.

=head1 ATTRIBUTES

The individual pieces of data that are part of the state of the Port element
are the individual attributes of the Port ComponentState

=over

=item C<< 'port_number'=>$string >>

(Represents port number in 3 or 1/1 format)

=item C<< time >>

(Array)Lists time specific statistics details of port

=item C<< error >>

(Array)Lists error specific statistics details of port

=item C<< stat >>

(Array)Lists statistics details of port

=back

=cut

###################
# Package
package NACL::CS::Switch::PortStats;

###################
# Everytime
use strict;
use warnings;

###################
# Module includes
use Params::Validate qw (validate);
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use NATE::BaseException qw(:try);
use base 'NACL::CS::ComponentState::Switch';

use Class::MethodMaker [

    scalar => "port_number",
    array  => "time",
    array  => "error",
    array  => "stat",
];

=head1 METHODS

=head2 fetch

    my $portstats = NACL::CS::Switch::PortStats->fetch(command_interface=>$ci,...);
    my @portstats = NACL::CS::Switch::PortStats->fetch(command_interface=>$ci,...);

see L<NACL::CS::ComponentState|lib-NACL-CS-ComponentState-pm>

=cut

sub fetch {
    my ($pkg, @args) = @_;
    $Log->enter() if $may_enter;
    my @state_objs = $pkg->SUPER::fetch(
        @args,
        choices => [
            {
                method    => "_fetch_brocadefc_cli",
                interface => "CLI",
                set       => "BrocadeFC"
            },
        ],
        exception_text => 'No matching elements found'
    );
    $Log->exit() if $may_exit;
    return wantarray ? @state_objs : $state_objs[0];
}

sub _fetch_brocadefc_cli {
    my ($pkg, @args) = @_;
    $Log->enter() if $may_enter;
    my %opts = validate @args, $pkg->_fetch_backend_validate_spec();
    my ( $output, $response, @state_objs, $caught_exception );
    my $apiset            = delete $opts{apiset};
    my $port_number       = $opts{filter}->{'port-number'};
    my $command_interface = delete $opts{command_interface};
    if ( !(defined $port_number) ) {
        $Log->exit() if $may_exit;
        NATE::Exceptions::Argument->throw("No port number was passed in filter");
    }
    try{
        $response = $apiset->portstatsshow( 'port-number' => $port_number );
    }
    catch NACL::APISet::Exceptions::InvalidParamException with {
       $caught_exception = 1;
    };
    if ($caught_exception) {
        $Log->exit() if $may_exit;
	return;
    }
    $output = $response->get_parsed_output();
    foreach my $row (@$output) {
        my $target = $pkg->_hash_copy(
            source => $row,
            copy   => [qw(stat)],
            map => {
                'tim' => 'time',
                'er'  => 'error',
            },
        );
        my $obj = $pkg->new(
            command_interface => $command_interface,
            'port-number'     => $port_number
        );
        $obj->_set_fields( row => $target );
        push @state_objs, $obj;
    }
    $Log->exit() if $may_exit;
    return @state_objs;

}    #end sub _fetch_brocadefc_cli

1;
