# $Id$

#
# Copyright (c) 2001-2011 NetApp, Inc., All Rights Reserved
# Any use, modification, or distribution is prohibited
#  without prior written consent from NetApp, Inc.
#

## @summary CIFS ComponentState Module
## @author Vikas.p@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::CS::SysManager::CIFS

=head1 DESCRIPTION

C<NACL::CS::SysManager::CIFS> is a derived class of
L<NACL::CS::ComponentState::SysManager|lib-NACL-CS-ComponentState-SysManager-pm>.  It represents the  
state of an CIFS server configuration present on the System Manger GUI Screen.  A related class is
L<NACL::C::SysManager::CIFS|lib-NACL-C-SysManager-CIFS-pm>, which represents access
to an cifs setup

=head1 ATTRIBUTES

The individual pieces of data that are part of the  state of the cifs setup 
element are the attributes of the cifs ComponentState.

=over

=item C<< vserver >>

Filled in for CMode.

=item C<< addomainname >>

Filled in for CMode.

=item C<< securityStyle >>

Filled in for 7Mode.

=item C<< cifsAuditing >>

Filled in for 7Mode.

=item C<< nfsAuditing >>

Filled in for 7Mode.

=item C<< auditLogFile >>

Filled in for 7Mode.

=item C<< homeDirectoryStyle >>

Filled in for 7Mode.

=item C<< homeDirectoryPaths >>

Filled in for 7Mode.

=item C<< serviceStatus >>

Filled in for 7Mode.

=item C<< netbiosName >>

Filled in for 7Mode.

=item C<< dnsDomain >>

Filled in for 7Mode.

=item C<< domainType >>

Filled in for 7Mode.

=back

=cut

package NACL::CS::SysManager::Cifs;

use strict;
use warnings;
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Data::Dumper;
use Params::Validate qw(validate);

use base 'NACL::CS::ComponentState::SysManager';

use NACL::C::SysManager::Cifs;
use NACL::Exceptions::NoElementsFound qw(:try);

# Import Exception class
use NACL::APISet::Exceptions::ResponseException qw(:try);
use NACL::APISet::Exceptions::CommandFailedException;

use Class::MethodMaker [
    scalar => 'systemName',
    scalar => 'authenticationStyle',

    #7mode Only
    scalar => 'securityStyle',
    scalar => 'cifsAuditing',
    scalar => 'nfsAuditing',
    scalar => 'auditLogFile',
    scalar => 'homeDirectoryStyle',
    scalar => 'homeDirectoryPaths',
    scalar => 'serviceStatus',
    scalar => 'netbiosName',
    scalar => 'dnsDomain',
    scalar => 'domainType',

    #Cmode Only
    scalar => 'vserver',
    scalar => 'addomainname',
];

=head1 METHODS 

=head2 fetch

    my $cifs_state  = 
		NACL::CS::SysManager::Cifs->fetch(command_interface=>$ci,...);
    my @cifs_states = 
		NACL::CS::SysManager::Cifs->fetch(command_interface=>$ci,...);

see L<NACL::CS::ComponentState|lib-NACL-CS-ComponentState-pm>.

Uses a CMode GUI or a 7Mode GUI APISet.

=cut

sub fetch {
    $Log->enter() if $may_enter;
    my ( $pkg, @args ) = @_;

    my @state_objs = $pkg->SUPER::fetch(
        @args,
        choices => [
            {   method => "_fetch_cmode_gui",
                set    => "SysManager_CMode",
            },
            {   method => '_fetch_7mode_gui',
                set    => 'SysManager_7Mode',
            },
        ],
        exception_text => 'CIFS not licensed/Configured',
    );

    $Log->exit() if $may_exit;
    return wantarray ? @state_objs : $state_objs[0];
} ## end sub fetch

sub _fetch_7mode_gui {
    $Log->enter() if $may_enter;
    my ( $pkg, @args ) = @_;
    my %opts = validate @args, $pkg->_fetch_backend_validate_spec();

    my $apiset            = $opts{apiset};
    my $requested_fields  = $opts{requested_fields};
    my $filter            = $opts{filter};
    my $command_interface = $opts{command_interface};

    my $cifs_args = $pkg->_hash_copy( source => $filter, );

    my ( $response, $caught_exception, $parsed_output );
    try {

        # caling the gui api to get cifs status
        $response      = $apiset->get_cifs_configuration_info(%$cifs_args);
        $parsed_output = $response->get_parsed_output();
    } ## end try
    catch NACL::APISet::Exceptions::CommandFailedException with {

        # A caught exception indicates that the CIFS is either not licensed
        # or not configured. We catch the exception and return immediately.
        # The 'fetch' frontend decides whether to throw a NoElementsFound
        # exception based on the value of 'allow_empty'
        $Log->debug(
            'CommandFailedException caught. CIFS not licensed/configured');
        $caught_exception = 1;
    };
    if ($caught_exception) {
        $Log->exit() if $may_exit;
        return;
    }

    my %final_attributes = %$parsed_output;

    my $obj = $pkg->new( command_interface => $command_interface );
    $obj->_set_fields( row => \%final_attributes );
    $Log->exit() if $may_exit;

    return $obj;
} ## end sub _fetch_7mode_gui

sub _fetch_cmode_gui {
    $Log->enter() if $may_enter;
    my ( $pkg, @args ) = @_;
    my %opts = validate @args, $pkg->_fetch_backend_validate_spec();

    my $apiset            = $opts{apiset};
    my $requested_fields  = $opts{requested_fields};
    my $filter            = $opts{filter};
    my $command_interface = $opts{command_interface};

    my $cifs_args = $pkg->_hash_copy( source => $filter, );

    my ( $response, $caught_exception, $parsed_output );
    try {

        # caling the gui api to get CIFS status
        $response      = $apiset->get_cifs_configuration_info(%$cifs_args);
        $parsed_output = $response->get_parsed_output();
    } ## end try
    catch NACL::APISet::Exceptions::CommandFailedException with {

        # A caught exception indicates that the CIFS is either not licensed
        # or not configured. We catch the exception and return immediately.
        # The 'fetch' frontend decides whether to throw a NoElementsFound
        # exception based on the value of 'allow_empty'
        $Log->debug(
            'CommandFailedException caught. CIFS not licensed/configured');
        $caught_exception = 1;
    };
    if ($caught_exception) {
        $Log->exit() if $may_exit;
        return;
    }

    my %final_attributes = %$parsed_output;

    my $obj = $pkg->new( command_interface => $command_interface );
    $obj->_set_fields( row => \%final_attributes );
    $Log->exit() if $may_exit;

    return $obj;
} ## end sub _fetch_cmode_gui

1;
