# $Id$

#
# Copyright (c) 2001-2011 NetApp, Inc., All Rights Reserved
# Any use, modification, or distribution is prohibited
#  without prior written consent from NetApp, Inc.
#

## @summary IscsiTargetPortalGroup ComponentState Module
## @author vikas.p@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::CS::SysManager::IscsiTargetPortalGroup

=head1 DESCRIPTION

C<NACL::CS::SysManager::IscsiTargetPortalGroup> is a derived class of
L<NACL::CS::ComponentState::SysManager|lib-NACL-CS-ComponentState-SysManager-pm>.  It represents the  
state of an ONTAP IscsiTargetPortalGroup.  A related class is
L<NACL::C::SysManager::IscsiTargetPortalGroup|lib-NACL-C-SysManager-IscsiTargetPortalGroup-pm>, 
which represents access to an ONTAP IscsiTargetPortalGroup.

=head1 ATTRIBUTES

The individual pieces of data that are part of the state of the IscsiTargetPortalGroup 
element are the attributes of the IscsiTargetPortalGroup ComponentState.

=over

=item C<< tpgname >>

The name of the target portal group name element whose state is being represented.

=item C<< vserver >>

=item C<< tpg >>

=item C<< type >>

=item C<< tag >>

=back

=cut

package NACL::CS::SysManager::IscsiTargetPortalGroup;

use strict;
use warnings;
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Data::Dumper;
use Params::Validate qw(validate);

use base 'NACL::CS::ComponentState::SysManager';

use NACL::C::SysManager::IscsiTargetPortalGroup;
use NACL::Exceptions::NoElementsFound qw(:try);

# Import Exception class
use NACL::APISet::Exceptions::ResponseException qw(:try);
use NACL::APISet::Exceptions::CommandFailedException;

use Class::MethodMaker [
    scalar => 'vserver',
    scalar => 'tpgname',
    scalar => 'tpg',
    scalar => 'type',
    scalar => 'tag',
];

=head1 METHODS 

=head2 fetch

    my $tpg_state  = 
		NACL::CS::SysManager::IscsiTargetPortalGroup->fetch(command_interface => $ci,...);
    my @tpg_states = 
		NACL::CS::SysManager::IscsiTargetPortalGroup->fetch(command_interface => $ci,...);

see L<NACL::CS::ComponentState|lib-NACL-CS-ComponentState-pm>.

Uses a CMode GUI or a 7Mode GUI APISet.
(Works only when tpgname and vserver parameters are passed under filter option).

=cut

####################### GUI Fetch ###############################
sub fetch {
    $Log->enter() if $may_enter;
    my ( $pkg, @args ) = @_;

    my @state_objs = $pkg->SUPER::fetch(
        @args,
        choices => [
            {   method => "_fetch_cmode_gui",
                set    => "SysManager_CMode",
            },
            {   method => '_fetch_7mode_gui',
                set    => 'SysManager_7Mode',
            },
        ],
        exception_text => 'No matching target portal group found',
    );

    $Log->exit() if $may_exit;
    return wantarray ? @state_objs : $state_objs[0];
} ## end sub fetch

####################### GUI CMode Fetch ###############################
sub _fetch_cmode_gui {
    $Log->enter() if $may_enter;
    my ( $pkg, @args ) = @_;
    my %opts = validate @args, $pkg->_fetch_backend_validate_spec();

    my $apiset            = $opts{apiset};
    my $requested_fields  = $opts{requested_fields};
    my $filter            = $opts{filter};
    my $command_interface = $opts{command_interface};

    my $tpg_args = $pkg->_hash_copy(
        source => $filter,
        map    => {
            "tpgname" => "-tpgname",
            "vserver" => "-vservername",
        }
    );
	
	$pkg->_check_for_required_filter_fields(
        filter          => $filter,
        required_fields => [qw(tpgname vserver)],
    );
	
	my ( $response, $caught_exception, $parsed_output );
    try {

        # caling the gui api to get tpg status
        $response      = $apiset->get_tpg_info(%$tpg_args);
        $parsed_output = $response->get_parsed_output();
    } ## end try
    catch NACL::APISet::Exceptions::CommandFailedException with {

        # A caught exception indicates that the tpg being looked for
        # does not exist. We catch the exception and return immediately.
        # The 'fetch' frontend decides whether to throw a NoElementsFound
        # exception based on the value of 'allow_empty'
        $Log->debug(
            'CommandFailedException caught. Requested Target portal Group does '
                . 'not exist' );
        $caught_exception = 1;
    };
    if ($caught_exception) {
        $Log->exit() if $may_exit;
        return;
    }

    my %state_base_field_settings;

    $state_base_field_settings{'tpgname'} = $tpg_args->{'-tpgname'};
    $state_base_field_settings{'vserver'} = $tpg_args->{'-vservername'};

    my %final_attributes = ( %state_base_field_settings, %$parsed_output );

    my $obj = $pkg->new( command_interface => $command_interface );
    $obj->_set_fields( row => \%final_attributes );
    $Log->exit() if $may_exit;

    return $obj;
} ## end sub _fetch_cmode_gui

####################### GUI 7Mode Fetch ###############################
sub _fetch_7mode_gui {
    $Log->enter() if $may_enter;
    my ( $pkg, @args ) = @_;
    my %opts = validate @args, $pkg->_fetch_backend_validate_spec();

    my $apiset            = $opts{apiset};
    my $requested_fields  = $opts{requested_fields};
    my $filter            = $opts{filter};
    my $command_interface = $opts{command_interface};

    my $tpg_args = $pkg->_hash_copy(
        source => $filter,
        map    => { "tpgname" => "-tpgname", }
    );
	
	$pkg->_check_for_required_filter_fields(
        filter          => $filter,
        required_fields => [qw(tpgname)],
    );
	
	my ( $response, $caught_exception, $parsed_output );
    try {

        # caling the gui api to get tpg status
        $response      = $apiset->get_tpg_info(%$tpg_args);
        $parsed_output = $response->get_parsed_output();
    } ## end try
    catch NACL::APISet::Exceptions::CommandFailedException with {

        # A caught exception indicates that the tpg being looked for
        # does not exist. We catch the exception and return immediately.
        # The 'fetch' frontend decides whether to throw a NoElementsFound
        # exception based on the value of 'allow_empty'
        $Log->debug(
            'CommandFailedException caught. Requested Target Portal Group does '
                . 'not exist' );
        $caught_exception = 1;
    };
    if ($caught_exception) {
        $Log->exit() if $may_exit;
        return;
    }

    my %state_base_field_settings;

    $state_base_field_settings{'tpgname'} = $tpg_args->{'-tpgname'};

    my %final_attributes = ( %state_base_field_settings, %$parsed_output );
    my $obj = $pkg->new( command_interface => $command_interface );
    $obj->_set_fields( row => \%final_attributes );
    $Log->exit() if $may_exit;

    return $obj;
} ## end sub _fetch_7mode_gui

1;
