# $Id$

#
# Copyright (c) 2001-2011 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Lun ComponentState Module
## @author vikas.p@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::CS::SysManager::Lun

=head1 DESCRIPTION

C<NACL::CS::SysManager::Lun> is a derived class of
L<NACL::CS::ComponentState::SysManager|lib-NACL-CS-ComponentState-SysManager-pm>.  It represents the  
state of an ONTAP Lun.  A related class is
L<NACL::C::SysManager::Lun|lib-NACL-C-SysManager-Lun-pm>, which represents access
to an ONTAP Lun.

=head1 ATTRIBUTES

The individual pieces of data that are part of the  state of the Lun 
element are the attributes of the lun ComponentState.

=over

=item C<< lun >>

The name of the lun element whose state is being represented.

=item C<< vserver >>

=item C<< volume >>

=item C<< name >>

=item C<< path >>

=item C<< size >>

=item C<< status >>

=item C<< type >>

=item C<< serial_no >>

=item C<< description >>

=item C<< igroupdetails >>

=item C<< initiatordetais >>

=back

=cut

package NACL::CS::SysManager::Lun;

use strict;
use warnings;
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Data::Dumper;
use Params::Validate qw(validate);

use base 'NACL::CS::ComponentState::SysManager';

use NACL::C::SysManager::Lun;
use NACL::Exceptions::NoElementsFound qw(:try);

# Import Exception class
use NACL::APISet::Exceptions::ResponseException qw(:try);
use NACL::APISet::Exceptions::CommandFailedException;

use Class::MethodMaker [
    scalar => 'vserver',
    scalar => 'lun',
    scalar => 'volume',
    scalar => 'name',
    scalar => 'path',
    scalar => 'size',
    scalar => 'status',
    scalar => 'type',
    scalar => 'serial_no',
    scalar => 'description',
    scalar => 'igroupdetails',
    scalar => 'initiatordetais',
    scalar => 'thinprovision',
];

=head1 METHODS 

=head2 fetch

    my $lun_state  = 
		NACL::CS::SysManager::Lun->fetch(command_interface=>$ci,...);
    my @lun_states = 
		NACL::CS::SysManager::Lun->fetch(command_interface=>$ci,...);

see L<NACL::CS::ComponentState|lib-NACL-CS-ComponentState-pm>.

Uses a CMode GUI or a 7Mode GUI APISet.
(Works only when lun and vserver parameters are passed under filter option).

=cut

sub fetch {
    $Log->enter() if $may_enter;
    my ( $pkg, @args ) = @_;

    my @state_objs = $pkg->SUPER::fetch(
        @args,
        choices => [
            {   method => "_fetch_cmode_gui",
                set    => "SysManager_CMode",
            },
            {   method => '_fetch_7mode_gui',
                set    => 'SysManager_7Mode',
            },
        ],
        exception_text => 'No matching lun found',
    );

    $Log->exit() if $may_exit;
    return wantarray ? @state_objs : $state_objs[0];
} ## end sub fetch

sub _fetch_cmode_gui {
    $Log->enter() if $may_enter;
    my ( $pkg, @args ) = @_;
    my %opts = validate @args, $pkg->_fetch_backend_validate_spec();

    my $apiset            = $opts{apiset};
    my $requested_fields  = $opts{requested_fields};
    my $filter            = $opts{filter};
    my $command_interface = $opts{command_interface};

    my $lun_args = $pkg->_hash_copy(
        source => $filter,
        map    => {
            "lun"     => "-lunname",
            "volume"  => "-new_volname",
            "vserver" => "-vservername",
        }
    );

    $pkg->_check_for_required_filter_fields(
        filter          => $filter,
        required_fields => [qw(lun vserver)],
    );

    my ( $response, $caught_exception, $parsed_output );
    try {

        # caling the gui api to get lun status
        $response      = $apiset->lun_info(%$lun_args);
        $parsed_output = $response->get_parsed_output();
    } ## end try
    catch NACL::APISet::Exceptions::CommandFailedException with {

        # A caught exception indicates that the lun being looked for
        # does not exist. We catch the exception and return immediately.
        # The 'fetch' frontend decides whether to throw a NoElementsFound
        # exception based on the value of 'allow_empty'
        $Log->debug( 'CommandFailedException caught. Requested lun does '
                . 'not exist' );
        $caught_exception = 1;
    };
    if ($caught_exception) {
        $Log->exit() if $may_exit;
        return;
    }

    my %state_base_field_settings;

    $state_base_field_settings{'lun'}     = $lun_args->{'-lunname'};
    $state_base_field_settings{'volume'}  = $lun_args->{'-new_volname'};
    $state_base_field_settings{'vserver'} = $lun_args->{'-vservername'};

    #$state_base_field_settings{'state'}   = $parsed_output->{'status'};

    my %final_attributes = ( %state_base_field_settings, %$parsed_output );

    my $obj = $pkg->new( command_interface => $command_interface );
    $obj->_set_fields( row => \%final_attributes );
    $Log->exit() if $may_exit;

    return $obj;
} ## end sub _fetch_cmode_gui

sub _fetch_7mode_gui {
    $Log->enter() if $may_enter;
    my ( $pkg, @args ) = @_;
    my %opts = validate @args, $pkg->_fetch_backend_validate_spec();

    my $apiset            = $opts{apiset};
    my $requested_fields  = $opts{requested_fields};
    my $filter            = $opts{filter};
    my $command_interface = $opts{command_interface};

    my $lun_args = $pkg->_hash_copy(
        source => $filter,
        map    => {
            "lun"    => "-lunname",
            "volume" => "-new_volname",
        }
    );

    $pkg->_check_for_required_filter_fields(
        filter          => $filter,
        required_fields => [qw(lun)],
    );

    my ( $response, $caught_exception, $parsed_output );
    try {

        # caling the gui api to get lun status
        $response      = $apiset->lun_info(%$lun_args);
        $parsed_output = $response->get_parsed_output();
    } ## end try
    catch NACL::APISet::Exceptions::CommandFailedException with {

        # A caught exception indicates that the lun being looked for
        # does not exist. We catch the exception and return immediately.
        # The 'fetch' frontend decides whether to throw a NoElementsFound
        # exception based on the value of 'allow_empty'
        $Log->debug( 'CommandFailedException caught. Requested lun does '
                . 'not exist' );
        $caught_exception = 1;
    };
    if ($caught_exception) {
        $Log->exit() if $may_exit;
        return;
    }

    my %state_base_field_settings;

    $state_base_field_settings{'lun'}    = $lun_args->{'-lunname'};
    $state_base_field_settings{'volume'} = $lun_args->{'-new_volname'};

    #$state_base_field_settings{'state'}  = $parsed_output->{'status'};

    my %final_attributes = ( %state_base_field_settings, %$parsed_output );

    my $obj = $pkg->new( command_interface => $command_interface );
    $obj->_set_fields( row => \%final_attributes );
    $Log->exit() if $may_exit;

    return $obj;
} ## end sub _fetch_7mode_gui

1;
