# $Id$

#
# Copyright (c) 2001-2011 NetApp, Inc., All Rights Reserved
# Any use, modification, or distribution is prohibited
#  without prior written consent from NetApp, Inc.
#

## @summary SnapshotPolicy ComponentState Module
## @author vikas.p@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::CS::SysManager::SnapshotPolicy

=head1 DESCRIPTION

C<NACL::CS::SysManager::SnapshotPolicy> is a derived class of
L<NACL::CS::ComponentState::SysManager|lib-NACL-CS-ComponentState-SysManager-pm>.  It represents the  
state of an ONTAP SnapshotPolicy.  A related class is
L<NACL::C::SysManager::SnapshotPolicy|lib-NACL-C-SysManager-SnapshotPolicy-pm>, which represents access
to an ONTAP SnapshotPolicy.

=head1 ATTRIBUTES

The individual pieces of data that are part of the state of the SnapshotPolicy 
element are the attributes of the SnapshotPolicy ComponentState.

=over

=item C<< policy >>

The name of the Snapshot Policy element whose state is being represented.

=item C<< status >>

=item C<< max_snapshot_copies >>

=back

=cut

package NACL::CS::SysManager::SnapshotPolicy;

use strict;
use warnings;
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Data::Dumper;
use Params::Validate qw(validate);

use base 'NACL::CS::ComponentState::SysManager';

use NACL::C::SysManager::SnapshotPolicy;
use NACL::Exceptions::NoElementsFound qw(:try);

# Import Exception class
use NACL::APISet::Exceptions::ResponseException qw(:try);
use NACL::APISet::Exceptions::CommandFailedException;

use Class::MethodMaker [
    scalar => 'policy',
    scalar => 'status',
    scalar => 'max_snapshot_copies',
];

=head1 METHODS 

=head2 fetch

    my $snapshot_policy_state  = 
		NACL::CS::SysManager::SnapshotPolicy->fetch(command_interface => $ci,...);
    my @snapshot_policy_states = 
		NACL::CS::SysManager::SnapshotPolicy->fetch(command_interface => $ci,...);

see L<NACL::CS::ComponentState|lib-NACL-CS-ComponentState-pm>.

Uses a CMode GUI APISet.
(Works only when policy and vserver parameters are passed under filter option).

=cut

sub fetch {
    $Log->enter() if $may_enter;
    my ( $pkg, @args ) = @_;

    my @state_objs = $pkg->SUPER::fetch(
        @args,
        choices => [
            {   method => "_fetch_cmode_gui",
                set    => "SysManager_CMode",
            },
        ],
        exception_text => 'No matching snapshot policy found',
    );

    $Log->exit() if $may_exit;
    return wantarray ? @state_objs : $state_objs[0];
} ## end sub fetch

sub _fetch_cmode_gui {
    $Log->enter() if $may_enter;
    my ( $pkg, @args ) = @_;
    my %opts = validate @args, $pkg->_fetch_backend_validate_spec();

    my $apiset            = $opts{apiset};
    my $requested_fields  = $opts{requested_fields};
    my $filter            = $opts{filter};
    my $command_interface = $opts{command_interface};

    my $snapshot_args = $pkg->_hash_copy(
        source => $filter,
        map    => { "policy" => "-policyName", }
    );

    $pkg->_check_for_required_filter_fields(
        filter          => $filter,
        required_fields => [qw(policy)],
    );

    my ( $response, $caught_exception, $parsed_output );
    try {

        # caling the gui api to get snapshot policy status
        $response      = $apiset->get_snapshotpolicy_info(%$snapshot_args);
        $parsed_output = $response->get_parsed_output();
    } ## end try
    catch NACL::APISet::Exceptions::CommandFailedException with {

      # A caught exception indicates that the snapshot policy being looked for
      # does not exist. We catch the exception and return immediately.
      # The 'fetch' frontend decides whether to throw a NoElementsFound
      # exception based on the value of 'allow_empty'
        $Log->debug(
            'CommandFailedException caught. Requested snapshot policy does '
                . 'not exist' );
        $caught_exception = 1;
    };
    if ($caught_exception) {
        $Log->exit() if $may_exit;
        return;
    }

    my %state_base_field_settings;

    $state_base_field_settings{'policy'} = $snapshot_args->{'-policyName'};

    my %final_attributes = ( %state_base_field_settings, %$parsed_output );
    my $obj = $pkg->new( command_interface => $command_interface );
    $obj->_set_fields( row => \%final_attributes );
    $Log->exit() if $may_exit;

    return $obj;
} ## end sub _fetch_cmode_gui

1;
