# $Id$

#
# Copyright (c) 2001-2011 NetApp, Inc., All Rights Reserved
# Any use, modification, or distribution is prohibited
#  without prior written consent from NetApp, Inc.
#

## @summary VolumeSnapshot ComponentState Module
## @author vikas.p@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::CS::SysManager::VolumeSnapshot

=head1 DESCRIPTION

C<NACL::CS::SysManager::VolumeSnapshot> is a derived class of
L<NACL::CS::ComponentState::SysManager|lib-NACL-CS-ComponentState-SysManager-pm>.  It represents the  
state of an ONTAP VolumeSnapshot.  A related class is
L<NACL::C::SysManager::VolumeSnapshot|lib-NACL-C-SysManager-VolumeSnapshot-pm>, 
which represents access to an ONTAP VolumeSnapshot.

=head1 ATTRIBUTES

The individual pieces of data that are part of the state of the VolumeSnapshot 
element are the attributes of the VolumeSnapshot ComponentState.

=over

=item C<< snapshot >>

The name of the Snapshot element whose state is being represented.

=item C<< volume >>

The name of the volume where snapshot exists.

=item C<< vserver >>

=item C<< name >>

=item C<< date_time >>

=item C<< total_size >>

=item C<< cumulative_total_size >>

=item C<< status >>

=item C<< application_dependency >>

=back

=cut

package NACL::CS::SysManager::VolumeSnapshot;

use strict;
use warnings;
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Data::Dumper;
use Params::Validate qw(validate);

use base 'NACL::CS::ComponentState::SysManager';

use NACL::C::SysManager::VolumeSnapshot;
use NACL::Exceptions::NoElementsFound qw(:try);

# Import Exception class
use NACL::APISet::Exceptions::ResponseException qw(:try);
use NACL::APISet::Exceptions::CommandFailedException;

use Class::MethodMaker [
    scalar => 'vserver',
    scalar => 'snapshot',
    scalar => 'volume',
    scalar => 'name',
    scalar => 'date_time',
    scalar => 'total_size',
    scalar => 'cumulative_total_size',
    scalar => 'status',
    scalar => 'application_dependency',
];

=head1 METHODS 

=head2 fetch

    my $volume_snapshot_state  = 
		NACL::CS::SysManager::VolumeSnapshot->fetch(command_interface => $ci,...);
    my @volume_snapshot_states = 
		NACL::CS::SysManager::VolumeSnapshot->fetch(command_interface => $ci,...);

see L<NACL::CS::ComponentState|lib-NACL-CS-ComponentState-pm>.

Uses a CMode GUI or a 7Mode GUI APISet.
(Works only when volume, snapshot and vserver parameters are passed under filter option).

=cut

sub fetch {
    $Log->enter() if $may_enter;
    my ( $pkg, @args ) = @_;

    my @state_objs = $pkg->SUPER::fetch(
        @args,
        choices => [
            {   method => "_fetch_cmode_gui",
                set    => "SysManager_CMode",
            },
            {   method => '_fetch_7mode_gui',
                set    => 'SysManager_7Mode',
            },
        ],
        exception_text => 'No matching snapshot found',
    );

    $Log->exit() if $may_exit;
    return wantarray ? @state_objs : $state_objs[0];
} ## end sub fetch

sub _fetch_cmode_gui {
    $Log->enter() if $may_enter;
    my ( $pkg, @args ) = @_;
    my %opts = validate @args, $pkg->_fetch_backend_validate_spec();

    my $apiset            = $opts{apiset};
    my $requested_fields  = $opts{requested_fields};
    my $filter            = $opts{filter};
    my $command_interface = $opts{command_interface};

    my $snapshot_args = $pkg->_hash_copy(
        source => $filter,
        map    => {
            "volume"   => "-volname",
            "vserver"  => "-vservername",
            "snapshot" => "-snapname",
        }
    );

    $pkg->_check_for_required_filter_fields(
        filter          => $filter,
        required_fields => [qw(volume snapshot vserver)],
    );

    my ( $response, $caught_exception, $parsed_output );
    try {

        # caling the gui api to get snapshot status
        $response      = $apiset->get_snapshot_info(%$snapshot_args);
        $parsed_output = $response->get_parsed_output();
    } ## end try
    catch NACL::APISet::Exceptions::CommandFailedException with {

        # A caught exception indicates that the snapshot being looked for
        # does not exist. We catch the exception and return immediately.
        # The 'fetch' frontend decides whether to throw a NoElementsFound
        # exception based on the value of 'allow_empty'
        $Log->debug( 'CommandFailedException caught. Requested snapshot does '
                . 'not exist' );
        $caught_exception = 1;
    };
    if ($caught_exception) {
        $Log->exit() if $may_exit;
        return;
    }

    my %state_base_field_settings;

    $state_base_field_settings{'volume'}   = $snapshot_args->{'-volname'};
    $state_base_field_settings{'vserver'}  = $snapshot_args->{'-vservername'};
    $state_base_field_settings{'snapshot'} = $snapshot_args->{'-snapname'};

    my %final_attributes = ( %state_base_field_settings, %$parsed_output );
    my $obj = $pkg->new( command_interface => $command_interface );
    $obj->_set_fields( row => \%final_attributes );
    $Log->exit() if $may_exit;

    return $obj;
} ## end sub _fetch_cmode_gui

sub _fetch_7mode_gui {
    $Log->enter() if $may_enter;
    my ( $pkg, @args ) = @_;
    my %opts = validate @args, $pkg->_fetch_backend_validate_spec();

    my $apiset            = $opts{apiset};
    my $requested_fields  = $opts{requested_fields};
    my $filter            = $opts{filter};
    my $command_interface = $opts{command_interface};

    my $snapshot_args = $pkg->_hash_copy(
        source => $filter,
        map    => {
            "volume"   => "-volname",
            "snapshot" => "-snapname",
        }
    );

    $pkg->_check_for_required_filter_fields(
        filter          => $filter,
        required_fields => [qw(volume snapshot)],
    );

    my ( $response, $caught_exception, $parsed_output );
    try {

        # caling the gui api to get snapshot status
        $response      = $apiset->get_snapshot_info(%$snapshot_args);
        $parsed_output = $response->get_parsed_output();
    } ## end try
    catch NACL::APISet::Exceptions::CommandFailedException with {

        # A caught exception indicates that the snapshot being looked for
        # does not exist. We catch the exception and return immediately.
        # The 'fetch' frontend decides whether to throw a NoElementsFound
        # exception based on the value of 'allow_empty'
        $Log->debug( 'CommandFailedException caught. Requested snapshot does '
                . 'not exist' );
        $caught_exception = 1;
    };
    if ($caught_exception) {
        $Log->exit() if $may_exit;
        return;
    }

    my %state_base_field_settings;

    $state_base_field_settings{'volume'}   = $snapshot_args->{'-volname'};
    $state_base_field_settings{'snapshot'} = $snapshot_args->{'-snapname'};

    my %final_attributes = ( %state_base_field_settings, %$parsed_output );
    my $obj = $pkg->new( command_interface => $command_interface );
    $obj->_set_fields( row => \%final_attributes );
    $Log->exit() if $may_exit;

    return $obj;
} ## end sub _fetch_7mode_gui

1;
