# $Id$
#
# Copyright (c) 2001-2012 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#
## @summary SystemNodeCoredumpConfig ComponentState Module
## @author dl-nacl-dev@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::CS::SystemNodeCoredumpConfig

=head1 DESCRIPTION

C<NACL::CS::SystemNodeCoredumpConfig> is a derived class of
L<NACL::CS::ComponentState::ONTAP|lib-NACL-CS-ComponentState-ONTAP-pm>. It represents the
state of an ONTAP SystemNodeCoredumpConfig.  A related class is
L<NACL::C::SystemNodeCoredumpConfig|lib-NACL-C-SystemNodeCoredumpConfig-pm>, which represents access
to an ONTAP SystemNodeCoredumpConfig.

=head1 ATTRIBUTES

The individual pieces of data that are part of the  state of the SystemNodeCoredumpConfig
element are the attributes of the SystemNodeCoredumpConfig ComponentState.

=over

=item C<< "save_attempts" >>

Filled in for CMode CLI/ZAPI

Maps to:

CMode ZAPI:

For "requested_fields", "filter" and Output mapping:

 <coredump-config-info>
    <save-attempts>$value</save-attempts>
 </coredump-config-info>

=item C<< "coredump_attempts" >>

Filled in for CMode CLI/ZAPI, 7Mode CLI.

Maps to:

CMode ZAPI:

For "requested_fields", "filter" and Output mapping:

 <coredump-config-info>
    <coredump-attempts>$value</coredump-attempts>
 </coredump-config-info>

7Mode CLI:

For "filter":
Applicable, Filtering will be done by Components.

For "requested_fields",
Not applicable, but the field will be populated in the CS object.

=item C<< "upload_location" >>

Filled in for CMode CLI/ZAPI

Maps to:

CMode ZAPI:

For "requested_fields", "filter" and Output mapping:

 <coredump-config-info>
    <upload-location>$value</upload-location>
 </coredump-config-info>.

=item C<< "sparsecore_enabled" >>

Filled in for CMode CLI/ZAPI

Maps to:

CMode ZAPI:

For "requested_fields", "filter" and Output mapping:

 <coredump-config-info>
    <is-sparsecore-enabled>$value</is-sparsecore-enabled>
 </coredump-config-info>.

=item C<< "save_onstartup" >>

Filled in for CMode CLI/ZAPI

Maps to:

CMode ZAPI:

For "requested_fields", "filter" and Output mapping:

 <coredump-config-info>
    <save-onstartup>$value</save-onstartup>
 </coredump-config-info>

=item C<< "node" >>

Filled in for CMode CLI/ZAPI

Maps to:

CMode ZAPI:

For "requested_fields", "filter" and Output mapping:

 <coredump-config-info>
    <node-name>$value</node-name>
 </coredump-config-info>.

=item C<< "min_free" >>

Filled in for CMode CLI/ZAPI

Maps to:

CMode ZAPI:

For "requested_fields", "filter" and Output mapping:

 <coredump-config-info>
    <min-free>$value</min-free>
 </coredump-config-info>

=back

=cut

package NACL::CS::SystemNodeCoredumpConfig;

use strict;
use warnings;
use Params::Validate qw(validate);
use NACL::ComponentUtils qw(_dump_one);
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();

use NACL::Exceptions::NoElementsFound qw(:try);

use base 'NACL::CS::ComponentState::ONTAP';

use Class::MethodMaker [
    scalar => 'save_attempts',
    scalar => 'coredump_attempts',
    scalar => 'upload_location',
    scalar => 'sparsecore_enabled',
    scalar => 'save_onstartup',
    scalar => 'node',
    scalar => 'min_free',
];

=head1 METHODS

=head2 fetch

    my $SystemNodeCoredumpConfig_state   = NACL::CS::SystemNodeCoredumpConfig->fetch(command_interface => $ci, ...);
	my @SystemNodeCoredumpConfig_states = NACL::CS::SystemNodeCoredumpConfig->fetch(command_interface => $ci, ...);

(Class method)
Discovers which elements are present and returns their state in ComponentState
objects. Called in scalar context it returns only one state object, in list
context it returns all state objects.
See L<NACL::CS::ComponentState|lib-NACL-CS-ComponentState-pm/fetch> for a
more detailed description along with a complete explanation of the options
it accepts.

Supports CMode CLI/ZAPI, 7Mode CLI.

Invokes "coredump-config-get-iter" API for CMode ZAPI.

Invokes "options" command for 7Mode CLI.

=over

=item Exceptions

=over

=item C<NACL::Exceptions::NoElementsFound>

When there are no elements matching the query specified or elements of that type
doesn't exist, then this exception will be thrown.

=back

=back

=cut

sub fetch {
    $Log->enter() if $may_enter;
    my $pkg        = shift;
    my @state_objs = $pkg->SUPER::fetch(
        @_,
        choices => [
            {   method    => '_fetch_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_fetch_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
            {   method    => '_fetch_7mode_cli',
                interface => 'CLI',
                set       => '7Mode',
            },
        ],
	show_cmd       => 'system node coredump config show',
        exception_text => 'No matching system node coredump config(s) found'
    );

    $Log->exit() if $may_exit;
    return wantarray ? @state_objs : $state_objs[0];
} ## end sub fetch

sub _fetch_cmode_cli {
    $Log->enter() if $may_enter;
    my $pkg        = shift;
    my @state_objs = $pkg->SUPER::_fetch_cmode_cli(@_,
        api => 'system_node_coredump_config_show');
    $Log->exit() if $may_exit;
    return @state_objs;
} ## end sub _fetch_cmode_cli

sub _fetch_cmode_zapi {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = @_;
    my $copy = [
        qw(min-free coredump-attempts
            save-attempts save-onstartup upload-location)
    ];
    my $map = {
        "node"               => ['node-name'],
        "sparsecore_enabled" => ['is-sparsecore-enabled'],
    };

    $Log->exit() if $may_exit;
    return $pkg->SUPER::_fetch_cmode_zapi(
        %opts,
        api  => "coredump-config-get-iter",
        copy => $copy,
        map  => $map,
    );
} ## end sub _fetch_cmode_zapi

sub _fetch_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg = shift;
    my %opts = validate @_, $pkg->_fetch_backend_validate_spec();
    my @state_objs;
    my $apiset            = delete $opts{apiset};
    my $command_interface = delete $opts{command_interface};
    my $response          = $apiset->options("coredump");
    my $output            = $response->get_parsed_output();
    my $node_name         = $command_interface->node();

    foreach my $list_output (@{$output}) {
        my $target = $pkg->_hash_copy(
            source => $list_output,
            map    => {'coredump.dump.attempts' => 'coredump-attempts',},
        );
        my $obj = $pkg->new(command_interface => $command_interface,);
        $obj->node($node_name);
        $obj->_set_fields(row => $target);
        push @state_objs, $obj;
    } ## end foreach my $list_output (@$output)
    $Log->exit() if $may_exit;
    return @state_objs;
} ## end sub _fetch_7mode_cli

1;
