#
# Copyright (c) 2001-2012 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary SystemNodeImage ComponentState Module
## @author dl-nacl-dev@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::CS::SystemNodeImage

=head1 DESCRIPTION

C<NACL::CS::SystemNodeImage> is a derived class of
L<NACL::CS::ComponentState::ONTAP|lib-NACL-CS-ComponentState-ONTAP-pm>. It represents the
state of an ONTAP SystemNodeImage.  A related class is
L<NACL::C::SystemNodeImage|lib-NACL-C-SystemNodeImage-pm>, which represents access
to an ONTAP SystemNodeImage.

=head1 ATTRIBUTES

The individual pieces of data that are part of the state of the SystemNodeImage
element are the attributes of the SystemNodeImage ComponentState.

=over

=item C<< build_type >>

Filled in for CMode CLI.

=item C<< installdate >>

Filled in for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

For "requested_fields", "filter" and Output mapping:

 <system-image-attributes>
    <install-time>$value</install-time>
 </system-image-attributes>

=item C<< force_setdefault >>

Filled in for CMode CLI.

=item C<< build_info >>

Filled in for CMode CLI.

=item C<< rootfs_path >>

Filled in for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

For "requested_fields", "filter" and Output mapping:

 <system-image-attributes>
    <rootfs-path>$value</rootfs-path>
 </system-image-attributes>

=item C<< node >>

Filled in for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

For "requested_fields", "filter" and Output mapping:

 <system-image-attributes>
    <node>$value</node>
 </system-image-attributes>

=item C<< build_time >>

Filled in for CMode CLI.

=item C<< isdefault >>

Filled in for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

For "requested_fields", "filter" and Output mapping:

 <system-image-attributes>
    <is-default>$value</is-default>
 </system-image-attributes>

=item C<< iscurrent >>

Filled in for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

For "requested_fields", "filter" and Output mapping:

 <system-image-attributes>
    <is-current>$value</is-current>
 </system-image-attributes>

=item C<< deltas >>

Filled in for CMode CLI.

=item C<< change >>

Filled in for CMode CLI.

=item C<< version >>

Filled in for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

For "requested_fields", "filter" and Output mapping:

 <system-image-attributes>
    <version>$value</version>
 </system-image-attributes>

=item C<< newest_image_as_default >>

Filled in for CMode CLI.

=item C<< image >>

Filled in for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

For "requested_fields", "filter" and Output mapping:

 <system-image-attributes>
    <image>$value</image>
 </system-image-attributes>

=item C<< build_path >>

Filled in for CMode CLI.

=item C<< kernel_path >>

Filled in for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

For "requested_fields", "filter" and Output mapping:

 <system-image-attributes>
    <kernel-path>$value</kernel-path>
 </system-image-attributes>

=item C<< client >>

Filled in for CMode CLI.

=item C<< package >>

Package File Name

Filled in for CMode CLI.

=back

=cut

package NACL::CS::SystemNodeImage;

use strict;
use warnings;
use Params::Validate qw(validate);
use NACL::ComponentUtils qw(_dump_one Dumper);
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();

use NACL::Exceptions::NoElementsFound qw(:try);

use base 'NACL::CS::ComponentState::ONTAP';

use Class::MethodMaker [
    scalar => 'build_type',
    scalar => 'installdate',
    scalar => 'force_setdefault',
    scalar => 'build_info',
    scalar => 'rootfs_path',
    scalar => 'node',
    scalar => 'build_time',
    scalar => 'isdefault',
    scalar => 'iscurrent',
    scalar => 'deltas',
    scalar => 'change',
    scalar => 'version',
    scalar => 'newest_image_as_default',
    scalar => 'image',
    scalar => 'build_path',
    scalar => 'kernel_path',
    scalar => 'client',
    scalar => 'package',
];

=head1 METHODS

=head2 fetch

    my $SystemNodeImage_state   = NACL::CS::SystemNodeImage->fetch(command_interface => $ci, ...);
	my @SystemNodeImage_states = NACL::CS::SystemNodeImage->fetch(command_interface => $ci, ...);

(Class method)
Discovers which elements are present and returns their state in ComponentState
objects. Called in scalar context it returns only one state object, in list
context it returns all state objects.
See L<NACL::CS::ComponentState|lib-NACL-CS-ComponentState-pm/fetch> for a
more detailed description along with a complete explanation of the options
it accepts.

Supports CMode CLI/ZAPI.

Invokes "systemr-image-get-iter" API for CMode ZAPI

=over

=item Exceptions

=over

=item C<NACL::Exceptions::NoElementsFound>

When there are no elements matching the query specified or elements of that type
doesn't exist, then this exception will be thrown.

=back

=back

=cut

sub fetch {
    $Log->enter() if $may_enter;
    my $pkg        = shift;
    my @state_objs = $pkg->SUPER::fetch(
        @_,
        choices => [
            {   method    => '_fetch_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_fetch_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
        ],
	show_cmd       => 'system node image show',
        exception_text => 'No matching system node image(s) found'
    );

    $Log->exit() if $may_exit;
    return wantarray ? @state_objs : $state_objs[0];
} ## end sub fetch

sub _fetch_cmode_cli {
    $Log->enter() if $may_enter;
    my $pkg = shift;
    my @state_objs
        = $pkg->SUPER::_fetch_cmode_cli(@_, api => 'system_node_image_show');
    $Log->exit() if $may_exit;
    return @state_objs;
} ## end sub _fetch_cmode_cli

sub _fetch_cmode_zapi {
    $Log->enter() if $may_enter;
    my $pkg        = shift;
    my @state_objs = $pkg->SUPER::_fetch_cmode_zapi(
        @_,
        api  => 'system_image_get_iter',
        copy => [qw (node rootfs-path image kernel-path version)],
        map  => {
            'installdate' => ['install-time'],
            'iscurrent'   => ['is-current'],
            'isdefault'   => ['is-default'],
            ,
        },
    );

    $Log->exit() if $may_exit;
    return @state_objs;
} ## end sub _fetch_cmode_zapi
1;
