#
# Copyright (c) 2001-2012 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary SystemNodeImagePackage ComponentState Module
## @author dl-nacl-dev@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::CS::SystemNodeImagePackage

=head1 DESCRIPTION

C<NACL::CS::SystemNodeImagePackage> is a derived class of
L<NACL::CS::ComponentState::ONTAP|lib-NACL-CS-ComponentState-ONTAP-pm>. It represents the
state of an ONTAP SystemNodeImagePackage.  A related class is
L<NACL::C::SystemNodeImagePackage|lib-NACL-C-SystemNodeImagePackage-pm>, which represents access
to an ONTAP SystemNodeImagePackage.

=head1 ATTRIBUTES

The individual pieces of data that are part of the  state of the SystemNodeImagePackage
element are the attributes of the SystemNodeImagePackage ComponentState.

=over

=item C<< "repository" >>

=item C<< "node" >>

Filled in for CMode CLI/ZAPI, 7Mode CLI.

Maps to:

CMode ZAPI:

For "requested_fields", "filter" and Output mapping:

 <system-image-package-attributes>
	<node>$value</node>
 </<system-image-package-attributes

7Mode CLI:

For "filter":
Applicable, Filtering will be done by Components.

For "requested_fields",
Not applicable, but the field will be populated in the CS object.

=item C<< "package" >>

Filled in for CMode CLI/ZAPI, 7Mode CLI.

Maps to:

CMode ZAPI:

For "requested_fields", "filter" and Output mapping:

 <system-image-package-attributes>
	<package>$value</package>
 </<system-image-package-attributes

7Mode CLI:

For "filter":
Applicable, Filtering will be done by Components.

For "requested_fields",
Not applicable, but the field will be populated in the CS object.

=back

=cut

package NACL::CS::SystemNodeImagePackage;

use strict;
use warnings;
use Params::Validate qw(validate);
use NACL::ComponentUtils qw(_dump_one Dumper);
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();

use NACL::Exceptions::NoElementsFound qw(:try);

use base 'NACL::CS::ComponentState::ONTAP';

use Class::MethodMaker [
    scalar => 'repository',
    scalar => 'node',
    scalar => 'package',
];

=head1 METHODS

=head2 fetch

    my $SystemNodeImagePackage_state   = NACL::CS::SystemNodeImagePackage->fetch(command_interface => $ci, ...);
	my @SystemNodeImagePackage_states = NACL::CS::SystemNodeImagePackage->fetch(command_interface => $ci, ...);

(Class method)
Discovers which elements are present and returns their state in ComponentState
objects. Called in scalar context it returns only one state object, in list
context it returns all state objects.
See L<NACL::CS::ComponentState|lib-NACL-CS-ComponentState-pm/fetch> for a
more detailed description along with a complete explanation of the options
it accepts.

Supports CMode CLI/ZAPI, 7Mode CLI.

Invokes "system-image-package-get-iter" API for CMode ZAPI.

Invokes "software list" command for 7Mode CLI.

=over

=item Exceptions

=over

=item C<NACL::Exceptions::NoElementsFound>

When there are no elements matching the query specified or elements of that type
doesn't exist, then this exception will be thrown.

=back

=back

=cut

sub fetch {
    $Log->enter() if $may_enter;
    my $pkg        = shift;
    my @state_objs = $pkg->SUPER::fetch(
        @_,
        choices => [
            {   method    => "_fetch_cmode_cli",
                interface => "CLI",
                set       => "CMode",
            },
            {   method    => "_fetch_7mode_cli",
                interface => "CLI",
                set       => "7Mode",
            },
            {   method    => '_fetch_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
        ],
        exception_text => 'No matching system node image package(s) found'
    );

    $Log->exit() if $may_exit;
    return wantarray ? @state_objs : $state_objs[0];
} ## end sub fetch

sub _fetch_cmode_cli {
    $Log->enter() if $may_enter;
    my $pkg        = shift;
    my @state_objs = $pkg->SUPER::_fetch_cmode_cli(@_,
        api => 'system_node_image_package_show');
    $Log->exit() if $may_exit;
    return @state_objs;
} ## end sub _fetch_cmode_cli

sub _fetch_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg               = shift;
    my %opts              = validate @_, $pkg->_fetch_backend_validate_spec();
    my $apiset            = delete $opts{apiset};
    my $command_interface = delete $opts{'command_interface'};
    my @state_objs;

    my $response  = $apiset->software_list();
    my $output    = $response->get_parsed_output();
    my $node_name = $command_interface->node();
    $Log->debug(Data::Dumper::Dumper($output));
    foreach my $package (@{$output->[0]->{'software_images'}}) {
        my $obj = $pkg->new();
        $obj->node($node_name);
        $obj->package($package);
        push @state_objs, $obj;
    } ## end foreach my $package ( @{ $output...
    $Log->exit() if $may_exit;
    return @state_objs;
} ## end sub _fetch_7mode_cli

sub _fetch_cmode_zapi {
    $Log->enter() if $may_enter;
    my $pkg        = shift;
    my @state_objs = $pkg->SUPER::_fetch_cmode_zapi(
        @_,
        api  => 'system_image_package_get_iter',
        copy => [qw (node package)],
    );

    $Log->exit() if $may_exit;
    return @state_objs;
} ## end sub _fetch_cmode_zapi

1;
