#
# Copyright (c) 2001-2013 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary VolumeEfficiencyPolicy ComponentState Module
## @author dl-nacl-dev@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::CS::VolumeEfficiencyPolicy

=head1 DESCRIPTION

C<NACL::CS::VolumeEfficiencyPolicy> is a derived class of
L<NACL::CS::ComponentState::ONTAP|lib-NACL-CS-ComponentState-ONTAP-pm>. It represents the
state of an ONTAP VolumeEfficiencyPolicy.  A related class is
L<NACL::C::VolumeEfficiencyPolicy|lib-NACL-C-VolumeEfficiencyPolicy-pm>, which represents access
to an ONTAP VolumeEfficiencyPolicy.

=head1 ATTRIBUTES

The individual pieces of data that are part of the  state of the VolumeEfficiencyPolicy
element are the attributes of the VolumeEfficiencyPolicy ComponentState.

=over

=item C<< "duration_mins" >>

Filled in for CMode CLI.

=item C<< "vserver" >>

Filled in for CMode CLI/ZAPI.

Maps to:

CMode ZAPI:

For "requested_fields", "filter" and Output mapping:

 <sis-policy-info>
     <vserver>$value</vserver>
 </sis-policy-info>

=item C<< "duration" >>

Filled in for CMode CLI/ZAPI.

Maps to:

CMode ZAPI:

For "requested_fields", "filter" and Output mapping:

 <sis-policy-info>
     <duration>$value</duration>
 </sis-policy-info>

=item C<< "schedule" >>

Filled in for CMode CLI/ZAPI.

Maps to:

CMode ZAPI:

For "requested_fields", "filter" and Output mapping:

 <sis-policy-info>
     <schedule>$value</schedule>
 </sis-policy-info>

=item C<< "policyid" >>

Filled in for CMode CLI.

item C<< "comment" >>

Filled in for CMode CLI/ZAPI.

Maps to:

CMode ZAPI:

For "requested_fields", "filter" and Output mapping:

 <sis-policy-info>
     <comment>$value</comment>
 </sis-policy-info>

=item C<< "policy" >>

Filled in for CMode CLI/ZAPI.

Maps to:

CMode ZAPI:

For "requested_fields", "filter" and Output mapping:

 <sis-policy-info>
     <policy-name>$value</policy-name>
 </sis-policy-info>

=item C<< "enabled" >>

Filled in for CMode CLI/ZAPI.

Maps to:

CMode ZAPI:

For "requested_fields", "filter" and Output mapping:

 <sis-policy-info>
     <enabled>$value</enabled>
 </sis-policy-info>

=item C<< threshold_percent >>

Filled in for CMode CLI.

=item C<< qos_policy >>

Filled in for CMode CLI.

=item C<< type >>

Filled in for CMode CLI.

=item C<< policy_owner >>

Owner of the Policy

Filled in for CMode CLI/ZAPI.

Maps to:

CMode ZAPI:

For "requested_fields", "filter" and Output mapping:

 <sis-policy-info>
    <policy-owner>$value</policy-owner>
 </sis-policy-info>

=item C<< second_window_qos_level >>

Second Window QoS Level

Filled in for CMode CLI.

=item C<< window_only >>

Window Only possible value(s) are, true,false

Filled in for CMode CLI.

=item C<< second_window_start >>

Second Window Start

Filled in for CMode CLI.

=item C<< first_window_start >>

First Window Start

Filled in for CMode CLI.

=item C<< qos_level >>

QoS Level

Filled in for CMode CL/ZAPI.

Maps to:

CMode ZAPI:

For "requested_fields", "filter" and Output mapping:

 <sis-policy-info>
    <qos-level>$value</qos-level>
 </sis-policy-info>

=item C<< first_window_qos_level >>

First Window QoS Level

Filled in for CMode CLI.

=item C<< first_window_duration >>

First Window Duration

Filled in for CMode CLI.

=item C<< second_window_duration >>

Second Window Duration

Filled in for CMode CLI.

=item C<< system_managed >>

System Managed possible value(s) are, true,false

Filled in for CMode CLI.


=item C<< start_threshold_percent >>

Threshold Percentage

Filled in for CMode CLI.


=back

=cut

package NACL::CS::VolumeEfficiencyPolicy;

use strict;
use warnings;
use Params::Validate qw(validate);
use NACL::ComponentUtils qw(_dump_one Dumper);
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();

use NACL::Exceptions::NoElementsFound qw(:try);

use base 'NACL::CS::ComponentState::ONTAP';

use Class::MethodMaker [
    scalar => 'duration_mins',
    scalar => 'vserver',
    scalar => 'duration',
    scalar => 'schedule',
    scalar => 'policyid',
    scalar => 'comment',
    scalar => 'policy',
    scalar => 'enabled',
    scalar => 'threshold_percent',
    scalar => 'qos_policy',
    scalar => 'type',
    scalar => 'policy_owner',
    scalar => 'second_window_qos_level',
    scalar => 'window_only',
    scalar => 'second_window_start',
    scalar => 'first_window_start',
    scalar => 'qos_level',
    scalar => 'first_window_qos_level',
    scalar => 'first_window_duration',
    scalar => 'second_window_duration',
    scalar => 'system_managed',
    scalar => 'start_threshold_percent',
];

=head1 METHODS

=head2 fetch

    my $VolumeEfficiencyPolicy_state   = NACL::CS::VolumeEfficiencyPolicy->fetch(command_interface => $ci, ...);
	my @VolumeEfficiencyPolicy_states = NACL::CS::VolumeEfficiencyPolicy->fetch(command_interface => $ci, ...);

(Class method)
Discovers which elements are present and returns their state in ComponentState
objects. Called in scalar context it returns only one state object, in list
context it returns all state objects.
See L<NACL::CS::ComponentState|lib-NACL-CS-ComponentState-pm/fetch> for a
more detailed description along with a complete explanation of the options
it accepts.

Supports CMode CLI/ZAPI.

Invokes "sis-policy-get-iter" API for CMode ZAPI.

=over

=item Exceptions

=over

=item C<NACL::Exceptions::NoElementsFound>

When there are no elements matching the query specified or elements of that type
doesn't exist, then this exception will be thrown.

=back

=back

=cut

sub fetch {
    $Log->enter() if $may_enter;
    my $pkg        = shift;
    my @state_objs = $pkg->SUPER::fetch(
        @_,
        choices => [
            {   method    => '_fetch_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_fetch_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
        ],
	show_cmd       => 'volume efficiency policy show',
        exception_text => 'No matching volume efficiency policy(s) found'
    );

    $Log->exit() if $may_exit;
    return wantarray ? @state_objs : $state_objs[0];
} ## end sub fetch

sub _fetch_cmode_cli {
    $Log->enter() if $may_enter;
    my $pkg        = shift;
    my @state_objs = $pkg->SUPER::_fetch_cmode_cli(@_,
        api => 'volume_efficiency_policy_show');
    $Log->exit() if $may_exit;
    return @state_objs;
} ## end sub _fetch_cmode_cli

sub _fetch_cmode_zapi {
    my $pkg    = shift;
    my @states = $pkg->SUPER::_fetch_cmode_zapi(
        @_,
        api  => "sis_policy_get_iter",
        copy => [
            qw(vserver schedule duration enabled comment policy-owner qos-level)
        ],
        map => {"policy" => "policy-name",},
    );

    return @states;
} ## end sub _fetch_cmode_zapi

1;
