#
# Copyright (c) 2001-2012 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary VolumeFlexcache ComponentState Module
## @author dl-nacl-dev@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::CS::VolumeFlexcache

=head1 DESCRIPTION

C<NACL::CS::VolumeFlexcache> is a derived class of
L<NACL::CS::ComponentState::ONTAP|lib-NACL-CS-ComponentState-ONTAP-pm>. It represents the
state of an ONTAP VolumeFlexcache.  A related class is
L<NACL::C::VolumeFlexcache|lib-NACL-C-VolumeFlexcache-pm>, which represents access
to an ONTAP VolumeFlexcache.

=head1 ATTRIBUTES

The individual pieces of data that are part of the  state of the VolumeFlexcache
element are the attributes of the VolumeFlexcache ComponentState.

=over

=item C<< "cache_state" >>

Filled in for CMode CLI/ZAPI.

Maps to:

CMode ZAPI:

For "requested_fields", "filter" and Output mapping:

 <flexcache-info>
     <cache-state>$value</cache-state>
 </flexcache-info>

=item C<< "cache_volume" >>

Filled in for CMode CLI/ZAPI.

Maps to:

CMode ZAPI:

For "requested_fields", "filter" and Output mapping:

 <flexcache-info>
     <cache-volume>$value</cache-volume>
 </flexcache-info>

=item C<< "vserver" >>

Filled in for CMode CLI/ZAPI.

Maps to:

CMode ZAPI:

For "requested_fields", "filter" and Output mapping:

 <flexcache-info>
     <vserver>$value</vserver>
 </flexcache-info>

=item C<< "cache_aggregate" >>

Filled in for CMode CLI/ZAPI.

Maps to:

CMode ZAPI:

For "requested_fields", "filter" and Output mapping:

 <flexcache-info>
     <cache-aggregate>$value</cache-aggregate>
 </flexcache-info>

=item C<< "origin_volume" >>

Filled in for CMode CLI/ZAPI.

Maps to:

CMode ZAPI:

For "requested_fields", "filter" and Output mapping:

 <flexcache-info>
     <origin-volume>$value</origin-volume>
 </flexcache-info>

=item C<< "origin_aggregate" >>

Filled in for CMode CLI/ZAPI.

Maps to:

CMode ZAPI:

For "requested_fields", "filter" and Output mapping:

 <flexcache-info>
     <origin-aggregate>$value</origin-aggregate>
 </flexcache-info>

=item C<< "connection_status" >>

Filled in for CMode CLI/ZAPI.

Maps to:

CMode ZAPI:

For "requested_fields", "filter" and Output mapping:

 <flexcache-info>
     <connection-status>$value</connection-status>
 </flexcache-info>

=item C<< "cache_available" >>

Filled in for CMode CLI/ZAPI.

Maps to:

CMode ZAPI:

For "requested_fields", "filter" and Output mapping:

 <flexcache-info>
     <cache-available>$value</cache-available>
 </flexcache-info>

=item C<< "cache_size" >>

Filled in for CMode CLI/ZAPI.

Maps to:

CMode ZAPI:

For "requested_fields", "filter" and Output mapping:

 <flexcache-info>
     <cache-size>$value</cache-size>
 </flexcache-info>

=item C<< "cache_percent_used" >>

Filled in for CMode CLI/ZAPI.

Maps to:

CMode ZAPI:

For "requested_fields", "filter" and Output mapping:

 <flexcache-info>
     <cache-percent-used>$value</cache-percent-used>
 </flexcache-info>

=item C<< "origin_state" >>

Filled in for CMode CLI/ZAPI.

Maps to:

CMode ZAPI:

For "requested_fields", "filter" and Output mapping:

 <flexcache-info>
     <origin-state>$value</origin-state>
 </flexcache-info>

=item C<< origin_node >>

Origin Volume Node Name

Filled in for CMode CLI.

=item C<< cache_node >>

Cache Node Name

Filled in for CMode CLI.

=back

=cut

package NACL::CS::VolumeFlexcache;

use strict;
use warnings;
use Params::Validate qw(validate);
use NACL::ComponentUtils qw(_dump_one);
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Data::Dumper;

use NACL::Exceptions::NoElementsFound qw(:try);

use base 'NACL::CS::ComponentState::ONTAP';

use Class::MethodMaker [
    scalar => 'cache_state',
    scalar => 'cache_volume',
    scalar => 'vserver',
    scalar => 'cache_aggregate',
    scalar => 'origin_volume',
    scalar => 'origin_aggregate',
    scalar => 'connection_status',
    scalar => 'cache_available',
    scalar => 'cache_size',
    scalar => 'cache_percent_used',
    scalar => 'origin_state',
    scalar => 'origin_node',
    scalar => 'cache_node',
];

=head1 METHODS

=head2 fetch

    my $VolumeFlexcache_state   = NACL::CS::VolumeFlexcache->fetch(command_interface => $ci, ...);
	my @VolumeFlexcache_states = NACL::CS::VolumeFlexcache->fetch(command_interface => $ci, ...);

(Class method)
Discovers which elements are present and returns their state in ComponentState
objects. Called in scalar context it returns only one state object, in list
context it returns all state objects.
See L<NACL::CS::ComponentState|lib-NACL-CS-ComponentState-pm/fetch> for a
more detailed description along with a complete explanation of the options
it accepts.

Supports CMode CLI/ZAPI.

Invokes "flexcache-get-iter" API for CMode ZAPI.

=over

=item Exceptions

=over

=item C<NACL::Exceptions::NoElementsFound>

When there are no elements matching the query specified or elements of that type
doesn't exist, then this exception will be thrown.

=back

=back

=cut

sub fetch {
    $Log->enter() if $may_enter;
    my $pkg        = shift;
    my @state_objs = $pkg->SUPER::fetch(
        @_,
        show_cmd => 'volume flexcache show',
        choices  => [
            {   method    => '_fetch_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_fetch_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
        ],
        exception_text => 'No matching volume flexcache(s) found'
    );

    $Log->exit() if $may_exit;
    return wantarray ? @state_objs : $state_objs[0];
} ## end sub fetch

sub _fetch_cmode_cli {
    $Log->enter() if $may_enter;
    my $pkg = shift;
    my @state_objs
        = $pkg->SUPER::_fetch_cmode_cli(@_, api => 'volume_flexcache_show');
    $Log->exit() if $may_exit;
    return @state_objs;
} ## end sub _fetch_cmode_cli

sub _fetch_cmode_zapi {
    $Log->enter() if $may_enter;
    my $pkg    = shift;
    my @states = $pkg->SUPER::_fetch_cmode_zapi(
        @_,
        api  => "flexcache_get_iter",
        copy => [
            qw(vserver cache-volume origin-volume cache-aggregate
                cache-size cache-state cache-available cache-percent-used
                connection-status origin-aggregate origin-state)
        ],
    );
    $Log->exit() if $may_exit;
    return @states;
} ## end sub _fetch_cmode_zapi
1;
