#
# Copyright (c) 2001-2013 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary VolumeFlexcacheCachePolicy ComponentState Module
## @author dl-nacl-dev@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::CS::VolumeFlexcacheCachePolicy

=head1 DESCRIPTION

C<NACL::CS::VolumeFlexcacheCachePolicy> is a derived class of
L<NACL::CS::ComponentState::ONTAP|lib-NACL-CS-ComponentState-ONTAP-pm>. It represents the
state of an ONTAP VolumeFlexcacheCachePolicy.  A related class is
L<NACL::C::VolumeFlexcacheCachePolicy|lib-NACL-C-VolumeFlexcacheCachePolicy-pm>, which represents access
to an ONTAP VolumeFlexcacheCachePolicy.

=head1 ATTRIBUTES

The individual pieces of data that are part of the  state of the VolumeFlexcacheCachePolicy
element are the attributes of the VolumeFlexcacheCachePolicy ComponentState.

=over

=item C<< "reg_ttl" >>

Filled in for CMode CLI/ZAPI.

Maps to:

CMode ZAPI:

For "requested_fields", "filter" and Output mapping:

 <flexcache-cache-policy-info>
     <time-to-live-files>$value</time-to-live-files>
 </flexcache-cache-policy-info>

=item C<< "dir_ttl" >>

Filled in for CMode CLI/ZAPI.

Maps to:

CMode ZAPI:

For "requested_fields", "filter" and Output mapping:

 <flexcache-cache-policy-info>
     <time-to-live-directories>$value</time-to-live-directories>
 </flexcache-cache-policy-info>

=item C<< "vserver" >>

Filled in for CMode CLI/ZAPI.

Maps to:

CMode ZAPI:

For "requested_fields", "filter" and Output mapping:

 <flexcache-cache-policy-info>
     <vserver>$value</vserver>
 </flexcache-cache-policy-info>

=item C<< "policy" >>

Filled in for CMode CLI/ZAPI.

Maps to:

CMode ZAPI:

For "requested_fields", "filter" and Output mapping:

 <flexcache-cache-policy-info>
     <cache-policy-name>$value</cache-policy-name>
 </flexcache-cache-policy-info>

=item C<< "id" >>

Filled in for CMode CLI.

=item C<< meta_ttl >>

*ONTAP Metafile TTL in Seconds

Filled in for CMode CLI/ZAPI.

Maps to:

CMode ZAPI:

For "requested_fields", "filter" and Output mapping:

 <flexcache-cache-policy-info>
     <time-to-live-metafiles>$value</time-to-live-metafiles>
 </flexcache-cache-policy-info>

=item C<< deleg_lru_timeout >>

*Delegation LRU Timeout in Seconds

Filled in for CMode CLI/ZAPI.

Maps to:

CMode ZAPI:

For "requested_fields", "filter" and Output mapping:

 <flexcache-cache-policy-info>
     <deleg-lru-timeout>$value</deleg-lru-timeout>
 </flexcache-cache-policy-info>

=item C<< other_ttl >>

*Other File TTL in Seconds

Filled in for CMode CLI/ZAPI.

Maps to:

CMode ZAPI:

For "requested_fields", "filter" and Output mapping:

 <flexcache-cache-policy-info>
     <time-to-live-other>$value</time-to-live-other>
 </flexcache-cache-policy-info>

=item C<< sym_ttl >>

*Symbolic Link TTL in Seconds

Filled in for CMode CLI/ZAPI.

Maps to:

CMode ZAPI:

For "requested_fields", "filter" and Output mapping:

 <flexcache-cache-policy-info>
     <time-to-live-symbolic>$value</time-to-live-symbolic>
 </flexcache-cache-policy-info>

=item C<< prefer_local_cache >>

*Local Cache Preferred over Local Origin, possible value(s) are true,false

Filled in for CMode CLI/ZAPI.

Maps to:

CMode ZAPI:

For "requested_fields", "filter" and Output mapping:

 <flexcache-cache-policy-info>
     <prefer-local-cache>$value</prefer-local-cache>
 </flexcache-cache-policy-info>

=back

=cut

package NACL::CS::VolumeFlexcacheCachePolicy;

use strict;
use warnings;
use Params::Validate qw(validate);
use NACL::ComponentUtils qw(_dump_one Dumper);
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();

use NACL::Exceptions::NoElementsFound qw(:try);
use NACL::C::_Mixins::VolumeFlexcacheCachePolicy qw(:all);

use base 'NACL::CS::ComponentState::ONTAP';

use Class::MethodMaker [
    scalar => 'reg_ttl',
    scalar => 'dir_ttl',
    scalar => 'vserver',
    scalar => 'policy',
    scalar => 'id',
    scalar => 'meta_ttl',
    scalar => 'deleg_lru_timeout',
    scalar => 'other_ttl',
    scalar => 'sym_ttl',
    scalar => 'prefer_local_cache',
];

=head1 METHODS

=head2 fetch

    my $VolumeFlexcacheCachePolicy_state   = NACL::CS::VolumeFlexcacheCachePolicy->fetch(command_interface => $ci, ...);
	my @VolumeFlexcacheCachePolicy_states = NACL::CS::VolumeFlexcacheCachePolicy->fetch(command_interface => $ci, ...);

(Class method)
Discovers which elements are present and returns their state in ComponentState
objects. Called in scalar context it returns only one state object, in list
context it returns all state objects.
See L<NACL::CS::ComponentState|lib-NACL-CS-ComponentState-pm/fetch> for a
more detailed description along with a complete explanation of the options
it accepts.

Supports CMode CLI/ZAPI.

Invokes "flexcache-cache-policy-get-iter" API for CMode ZAPI.

=over

=item Exceptions

=over

=item C<NACL::Exceptions::NoElementsFound>

When there are no elements matching the query specified or elements of that type
doesn't exist, then this exception will be thrown.

=back

=back

=cut

sub fetch {
    $Log->enter() if $may_enter;
    my $pkg        = shift;
    my @state_objs = $pkg->SUPER::fetch(
        @_,
        choices => [
            {   method    => '_fetch_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_fetch_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
        ],
	show_cmd       => 'volume flexcache cache-policy show',
        exception_text => 'No matching volume flexcache cache policy(s) found'
    );

    $Log->exit() if $may_exit;
    return wantarray ? @state_objs : $state_objs[0];
} ## end sub fetch

sub _fetch_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, %opts) = @_;
    my @state_objs = $pkg->SUPER::_fetch_cmode_cli(%opts,
        api => 'volume_flexcache_cache_policy_show');
    $Log->exit() if $may_exit;
    return @state_objs;
} ## end sub _fetch_cmode_cli

sub _fetch_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg, %opts) = @_;
    my $common_map  = $pkg->_common_zapi_map();
    my $common_copy = $pkg->_common_zapi_copy();
    my @state_objs  = $pkg->SUPER::_fetch_cmode_zapi(
        %opts,
        api  => "flexcache-cache-policy-get-iter",
        copy => $common_copy,
        map  => $common_map,
    );
    $Log->exit() if $may_exit;
    return @state_objs;

} ## end sub _fetch_cmode_zapi

1;
