#
# Copyright (c) 2001-2012 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary VolumeFlexcacheOptions ComponentState Module
## @author dl-nacl-dev@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::CS::VolumeFlexcacheOptions

=head1 DESCRIPTION

C<NACL::CS::VolumeFlexcacheOptions> is a derived class of
L<NACL::CS::ComponentState::ONTAP|lib-NACL-CS-ComponentState-ONTAP-pm>. It represents the
state of an ONTAP VolumeFlexcacheOptions.  A related class is
L<NACL::C::VolumeFlexcacheOptions|lib-NACL-C-VolumeFlexcacheOptions-pm>, which represents access
to an ONTAP VolumeFlexcacheOptions.

=head1 ATTRIBUTES

The individual pieces of data that are part of the  state of the VolumeFlexcacheOptions
element are the attributes of the VolumeFlexcacheOptions ComponentState.

=over

=item C<< "per_cache_statistics" >>

For "filter":
Applicable, Filtering will be done by Components.

For "requested_fields", Output mapping:

 <flexcache-options-info>
     <per-cache-statistics>$value<per-cache-statistics>
 </flexcache-options-info>

=item C<< "delegations_high_watermark" >>

For "filter":
Applicable, Filtering will be done by Components.

For "requested_fields", Output mapping:

 <flexcache-options-info>
     <delegations-high-watermark>$value<delegations-high-watermark>
 </flexcache-options-info>

=item C<< "delegations_low_watermark" >>

For "filter":
Applicable, Filtering will be done by Components.

For "requested_fields", Output mapping:

 <flexcache-options-info>
     <delegations-low-watermark>$value<delegations-low-watermark>
 </flexcache-options-info>

=item C<< "capability" >>

Filled in for CMode CLI.

=back

=cut

package NACL::CS::VolumeFlexcacheOptions;

use strict;
use warnings;
use Params::Validate qw(validate);
use NACL::ComponentUtils qw(_dump_one);
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Data::Dumper;

use NACL::Exceptions::NoElementsFound qw(:try);

use base 'NACL::CS::ComponentState::ONTAP';

use Class::MethodMaker [
    scalar => 'per_cache_statistics',
    scalar => 'delegations_high_watermark',
    scalar => 'delegations_low_watermark',
    scalar => 'capability',
];

=head1 METHODS

=head2 fetch

    my $VolumeFlexcacheOptions_state   = NACL::CS::VolumeFlexcacheOptions->fetch(command_interface => $ci, ...);
	my @VolumeFlexcacheOptions_states = NACL::CS::VolumeFlexcacheOptions->fetch(command_interface => $ci, ...);

(Class method)
Discovers which elements are present and returns their state in ComponentState
objects. Called in scalar context it returns only one state object, in list
context it returns all state objects.
See L<NACL::CS::ComponentState|lib-NACL-CS-ComponentState-pm/fetch> for a
more detailed description along with a complete explanation of the options
it accepts.

Supports CMode CLI/ZAPI.

Invokes "flexcache-options-get" API for CMode ZAPI.

=over

=item Exceptions

=over

=item C<NACL::Exceptions::NoElementsFound>

When there are no elements matching the query specified or elements of that type
doesn't exist, then this exception will be thrown.

=back

=back

=cut

sub fetch {
    $Log->enter() if $may_enter;
    my $pkg        = shift;
    my @state_objs = $pkg->SUPER::fetch(
        @_,
        choices => [
            {   method    => '_fetch_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_fetch_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
        ],
        exception_text => 'No matching volume flexcache options(s) found',
        is_singleton   => 1
    );

    $Log->exit() if $may_exit;
    return wantarray ? @state_objs : $state_objs[0];
} ## end sub fetch

sub _fetch_cmode_cli {
    $Log->enter() if $may_enter;
    my $pkg        = shift;
    my @state_objs = $pkg->SUPER::_fetch_cmode_cli(
        @_,
        api          => 'volume_flexcache_options_show',
        is_singleton => 1,

    );
    $Log->exit() if $may_exit;
    return @state_objs;
} ## end sub _fetch_cmode_cli

sub _fetch_cmode_zapi {
    $Log->enter() if $may_enter;
    my $pkg              = shift;
    my %opts             = validate @_, $pkg->_fetch_backend_validate_spec();
    my $apiset           = $opts{apiset};
    my $requested_fields = $opts{requested_fields};
    my $filter           = $opts{filter};
    my $caught_exception;
    my $flexcache_response;
    my %flexcache_options_get_args;

    # find the volume flexcache options.
    try {
        my %desired_attributes_hash;
        foreach (@{$requested_fields}) {
            $desired_attributes_hash{$_} = 1;
        }
        if (keys %desired_attributes_hash) {
            $flexcache_options_get_args{'desired-attributes'}
                = {%desired_attributes_hash};
        }
        $flexcache_response
            = $apiset->flexcache_options_get(%flexcache_options_get_args);
    } ## end try
    catch NACL::APISet::Exceptions::ResponseException with {

# A caught exception indicates that the volume flexcache options being looked for
# does not exist. We catch the exception and return immediately. The
# 'fetch' frontend decides whether to throw a NoElementsFound
# exception based on the value of 'allow_empty'
        $caught_exception = 1;
    };
    return if ($caught_exception);

    my $flexcache_output = $flexcache_response->get_parsed_output();
    $Log->debug(Dumper($flexcache_output));
    my @state_objs;
    foreach my $row (
        @{  $flexcache_output->[0]->{'attributes'}->[0]
                ->{'flexcache-options-info'}
        }
        )
    {
        my $state_fields = $pkg->_zapi_hash_copy(
            source                  => $row,
            source_has_extra_arrays => 1,
            copy                    => [
                qw(per-cache-statistics
                    delegations-high-watermark
                    delegations-low-watermark)
            ],
        );
        my $obj = $pkg->new();
        $obj->_set_fields(row => $state_fields);

        push @state_objs, $obj;
    } ## end foreach my $row ( @{ $flexcache_output...

    $Log->exit() if $may_exit;
    return @state_objs;
} ## end sub _fetch_cmode_zapi

1;
