#
# Copyright (c) 2001-2012 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary VserverIscsiSecurity ComponentState Module
## @author dl-nacl-dev@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::CS::VserverIscsiSecurity

=head1 DESCRIPTION

C<NACL::CS::VserverIscsiSecurity> is a derived class of
L<NACL::CS::ComponentState::ONTAP|lib-NACL-CS-ComponentState-ONTAP-pm>. It represents the
state of an ONTAP VserverIscsiSecurity.  A related class is
L<NACL::C::VserverIscsiSecurity|lib-NACL-C-VserverIscsiSecurity-pm>, which represents access
to an ONTAP VserverIscsiSecurity.

=head1 ATTRIBUTES

The individual pieces of data that are part of the  state of the VserverIscsiSecurity
element are the attributes of the VserverIscsiSecurity ComponentState.

=over

=item C<< "remove_outbound" >>

Filled in for CMode CLI.

=item C<< "auth_type" >>

Filled in for CMode CLI/ZAPI.

Maps to:

CMode ZAPI:

For "requested_fields", "filter" and Output mapping:

 <iscsi-security-entry-info>
     <auth-type>$value</auth-type>
 </iscsi-security-entry-info>

=item C<< "radius" >>

Filled in for CMode CLI.

=item C<< "vserver" >>

Filled in for CMode CLI/ZAPI.

Maps to:

CMode ZAPI:

For "requested_fields", "filter" and Output mapping:

 <iscsi-security-entry-info>
     <vserver>$value</vserver>
 </iscsi-security-entry-info>

=item C<< "auth_chap_policy" >>

Filled in for CMode CLI/ZAPI.

Maps to:

CMode ZAPI:

For "requested_fields", "filter" and Output mapping:

 <iscsi-security-entry-info>
     <auth-chap-policy>$value</auth-chap-policy>
 </iscsi-security-entry-info>

=item C<< "user_name" >>

Filled in for CMode CLI/ZAPI.

Maps to:

CMode ZAPI:

For "requested_fields", "filter" and Output mapping:

 <iscsi-security-entry-info>
     <user-name>$value</user-name>
 </iscsi-security-entry-info>

=item C<< "initiator" >>

Filled in for CMode CLI.

=item C<< "initiator_name" >>

Filled in for CMode CLI/ZAPI.

Maps to:

CMode ZAPI:

For "requested_fields", "filter" and Output mapping:

 <iscsi-security-entry-info>
     <initiator>$value</initiator>
 </iscsi-security-entry-info>

=item C<< "password" >>

Filled in for CMode CLI.

=item C<< "outbound_user_name" >>

Filled in for CMode CLI/ZAPI, 7Mode CLI.

Maps to:

CMode ZAPI:

For "requested_fields", "filter" and Output mapping:

 <iscsi-security-entry-info>
     <outbound-user-name>$value</outbound-user-name>
 </iscsi-security-entry-info>

7Mode CLI:

For "filter":
Applicable, Filtering will be done by Components.

For "requested_fields",
Not applicable, but the field will be populated in the CS object.

=item C<< "outpassword" >>

Filled in for CMode CLI.

=item C<< "security" >>

Filled in for CMode CLI, 7Mode CLI.

Maps to:

7Mode CLI:

For "filter":
Applicable, Filtering will be done by Components.

For "requested_fields",
Not applicable, but the field will be populated in the CS object.

=item C<< "outbound_password" >>

Filled in for CMode CLI, 7Mode CLI.

Maps to:

7Mode CLI:

For "filter":
Applicable, Filtering will be done by Components.

For "requested_fields",
Not applicable, but the field will be populated in the CS object.

=item C<< "outname" >>

Filled in for CMode CLI.

=item C<< "c_fetch" >>

This attribute stores the value of the "fetch" field that's present
in the show output. Note that there is no "c-fetch" field present
in the output, this attribute has been named this so that there is
no clash with the C<fetch> method.

Filled in for CMode CLI.

=item C<< "inname" >>

Filled in for CMode CLI, 7Mode CLI.

Maps to:

7Mode CLI:

For "filter":
Applicable, Filtering will be done by Components.

For "requested_fields",
Not applicable, but the field will be populated in the CS object.

=item C<< "inpassword" >>

Filled in for CMode CLI, 7Mode CLI.

Maps to:

7Mode CLI:

For "filter":
Applicable, Filtering will be done by Components.

For "requested_fields",
Not applicable, but the field will be populated in the CS object.

=item C<< clear_outbound >>

Filled in for CMode CLI.

=item C<< vserver_uuid >>

Vserver UUID

Filled in for CMode CLI.


=back

=cut

package NACL::CS::VserverIscsiSecurity;

use strict;
use warnings;
use Params::Validate qw(validate);
use NACL::ComponentUtils qw(_dump_one Dumper);
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();

use NACL::Exceptions::NoElementsFound qw(:try);

use base 'NACL::CS::ComponentState::ONTAP';

use Class::MethodMaker [
    scalar => 'remove_outbound',
    scalar => 'auth_type',
    scalar => 'radius',
    scalar => 'vserver',
    scalar => 'auth_chap_policy',
    scalar => 'user_name',
    scalar => 'initiator',
    scalar => 'initiator_name',
    scalar => 'password',
    scalar => 'outbound_user_name',
    scalar => 'outpassword',
    scalar => 'security',
    scalar => 'outbound_password',
    scalar => 'outname',
    scalar => 'c_fetch',
    scalar => 'inname',
    scalar => 'inpassword',
    scalar => 'clear_outbound',
    scalar => 'vserver_uuid',
];

=head1 METHODS

=head2 fetch

    my $VserverIscsiSecurity_state   = NACL::CS::VserverIscsiSecurity->fetch(command_interface => $ci, ...);
	my @VserverIscsiSecurity_states = NACL::CS::VserverIscsiSecurity->fetch(command_interface => $ci, ...);

(Class method)
Discovers which elements are present and returns their state in ComponentState
objects. Called in scalar context it returns only one state object, in list
context it returns all state objects.
See L<NACL::CS::ComponentState|lib-NACL-CS-ComponentState-pm/fetch> for a
more detailed description along with a complete explanation of the options
it accepts.

Supports CMode CLI/ZAPI, 7Mode CLI.

Invokes "iscsi-initiator-auth-get-iter" API for CMode ZAPI.

Invokes "iscsi security show" command for 7Mode CLI.

=over

=item Exceptions

=over

=item C<NACL::Exceptions::NoElementsFound>

When there are no elements matching the query specified or elements of that type
doesn't exist, then this exception will be thrown.

=back

=back

=cut

sub fetch {
    $Log->enter() if $may_enter;
    my $pkg        = shift;
    my @state_objs = $pkg->SUPER::fetch(
        @_,
        show_cmd => 'vserver iscsi security show',
        choices  => [
            {   method    => '_fetch_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_fetch_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
            {   method    => '_fetch_7mode_cli',
                interface => 'CLI',
                set       => '7Mode',
            },
        ],
        exception_text => 'No matching vserver iscsi security(s) found'
    );

    $Log->exit() if $may_exit;
    return wantarray ? @state_objs : $state_objs[0];
} ## end sub fetch

sub _fetch_cmode_cli {
    $Log->enter() if $may_enter;
    my $pkg        = shift;
    my @state_objs = $pkg->SUPER::_fetch_cmode_cli(@_,
        api => 'vserver_iscsi_security_show');
    $Log->exit() if $may_exit;
    return @state_objs;
} ## end sub _fetch_cmode_cli

sub _fetch_cmode_zapi {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my $copy = [
        qw(auth-type auth-chap-policy outbound-user-name user-name vserver )];
    my $map = {"initiator-name" => ['initiator']};

    my @state_objs = $pkg->SUPER::_fetch_cmode_zapi(
        @_,
        api  => "iscsi-initiator-auth-get-iter",
        copy => $copy,
        map  => $map,
    );
    $Log->exit() if $may_exit;
    return @state_objs;

} ## end sub _fetch_cmode_zapi

sub _fetch_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg         = shift;
    my %opts        = validate @_, $pkg->_fetch_backend_validate_spec();
    my $apiset      = $opts{apiset};
    my %copy_filter = %{$opts{filter}};
    my @copy_requested_fields = @{$opts{requested_fields}};
    my $deleted_filter        = $pkg->_remove_relational_regex_filters(
        filter           => \%copy_filter,
        requested_fields => \@copy_requested_fields
    );
    my $filter           = \%copy_filter;
    my $requested_fields = \@copy_requested_fields;
    my $req_field_filter;
    $req_field_filter->{'requested_fields'} = $requested_fields;
    $req_field_filter->{'filter'}           = \%copy_filter;
    my %cmd_opts;
    my @state_objs;

    my $response = $apiset->iscsi_security_show();
    my $output   = $response->get_parsed_output();
    foreach my $row (@$output) {
        my $obj;
        $obj = $pkg->new(command_interface => $opts{command_interface});
        my $row_modified = $pkg->_hash_copy(
            source => $row,
            map    => {
                'default_sec'            => 'security',
                'outbound_username'      => 'outbound_user_name',
                'outbound_password'      => 'outbound_password',
                'inbound_username'       => 'inname',
                'local_inbound_password' => 'inpassword',
            },
        );
        $obj->_set_fields(row => $row_modified);
        push @state_objs, $obj;
    } ## end foreach my $row (@$output)
    $Log->exit() if $may_exit;
    $Log->debug("value of the parsed output is :" . Dumper(\@state_objs));

    return @state_objs;
} ## end sub _fetch_7mode_cli
1;
