#
# Copyright (c) 2001-2013 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary VserverPeer ComponentState Module
## @author dl-nacl-dev@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::CS::VserverPeer

=head1 DESCRIPTION

C<NACL::CS::VserverPeer> is a derived class of
L<NACL::CS::ComponentState::ONTAP|lib-NACL-CS-ComponentState-ONTAP-pm>. It represents the
state of an ONTAP VserverPeer.  A related class is
L<NACL::C::VserverPeer|lib-NACL-C-VserverPeer-pm>, which represents access
to an ONTAP VserverPeer.

=head1 ATTRIBUTES

The individual pieces of data that are part of the state of the VserverPeer
element are the attributes of the VserverPeer ComponentState. 

Additionally, the command_interface used to obtain the ComponentState object is 
also an attribute of the object. This makes it easier to obtain the component
object corresponding to the CS object, using 
L<get_component_instance|lib-NACL-CS-ComponentState-pm/get_component_instance>.

=over

=item command_interface

The command_interface with which the CS object was obtained. (Available in
all CS objects regardless of requested_fields and the mode/interface)

=item C<< peer_state >>

Filled in for CMode CLI; CMode ZAPI (both iter and non-iter).

Maps to:

CMode ZAPI:

For "requested_fields", "filter" and Output mapping:

<vserver-peer-info>
    <peer-state>$value</peer-state>	
</vserver-peer-info>
 
=item C<< vserver >>

Filled in for CMode CLI; CMode ZAPI (both iter and non-iter).

Maps to:

CMode ZAPI:

For "requested_fields", "filter" and Output mapping:

<vserver-peer-info>
    <vserver>$value</vserver>	
</vserver-peer-info>
 
=item C<< peer_vserver >>

Filled in for CMode CLI; CMode ZAPI (both iter and non-iter).

Maps to:

CMode ZAPI:

For "requested_fields", "filter" and Output mapping:

<vserver-peer-info>
    <peer-vserver>$value</peer-vserver>	
</vserver-peer-info>
 
=item C<< applications >>

Filled in for CMode CLI; CMode ZAPI (both iter and non-iter).
(Array)

Maps to:

CMode ZAPI:

For "requested_fields", "filter" and Output mapping:

<vserver-peer-info>
    <applications>
        <vserver-peer-application>$value</vserver-peer-application>	
    </applications>
</vserver-peer-info>
 
Note that for array fields, the accessor method can be invoked in either
scalar or list context.
 my $applications = $obj->applications();
# $applications contains a reference to the array of values
 my @applications = $obj->applications();
# @applications contains the array of values

=item C<< peer_cluster >>

Filled in for CMode CLI; CMode ZAPI (both iter and non-iter).

Maps to:

CMode ZAPI:

For "requested_fields", "filter" and Output mapping:

<vserver-peer-info>
    <peer-cluster>$value</peer-cluster>	
</vserver-peer-info>

=item C<< peer_vserver_uuid >>

Peer Vserver UUID

Filled in for CMode CLI; CMode ZAPI (both iter and non-iter).

=item C<< vserver_uuid >>

Local Vserver UUID

Filled in for CMode CLI; CMode ZAPI (both iter and non-iter).

 
=back

=cut

package NACL::CS::VserverPeer;

use strict;
use warnings;
use Params::Validate qw(validate);
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();

use NACL::Exceptions::NoElementsFound qw(:try);

use base 'NACL::CS::ComponentState::ONTAP';

use NACL::CS::_Mixins::VserverPeer qw(:all);

use Class::MethodMaker [
    scalar => 'peer_state',
    scalar => 'vserver',
    scalar => 'peer_vserver',
    array  => 'applications',
    scalar => 'peer_cluster',
    scalar => 'peer_vserver_uuid',
    scalar => 'vserver_uuid',
];

=head1 METHODS

=head2 fetch

    my $VserverPeer_state  = NACL::CS::VserverPeer->fetch(command_interface => $ci, ...);
	my @VserverPeer_states = NACL::CS::VserverPeer->fetch(command_interface => $ci, ...);

(Class method)
Discovers which elements are present and returns their state in ComponentState
objects. Called in scalar context it returns only one state object, in list
context it returns all state objects.
See L<NACL::CS::ComponentState|lib-NACL-CS-ComponentState-pm/fetch> for a
more detailed description along with a complete explanation of the options
it accepts.

Uses a CMode CLI or CMode ZAPI APISet (supports both iter and non-iter).

Invokes vserver-peer-get-iter for iter CMode ZAPI.

Invokes vserver-peer-get for non-iter CMode ZAPI.

=over

=item Exceptions

=over

=item C<NACL::Exceptions::NoElementsFound>

When there are no elements matching the query specified or elements of that type
doesn't exist, then this exception will be thrown.

=back

=back

=cut

sub fetch {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    my @state_objs = $pkg->SUPER::fetch(@args, $pkg->_fetch_opts(), show_cmd => 'vserver peer show',);
    $Log->exit() if $may_exit;
    return wantarray ? @state_objs : $state_objs[0];
}

sub _fetch_cmode_cli {
    $Log->enter() if $may_enter;
    my $pkg = shift;
    my @state_objs
        = $pkg->SUPER::_fetch_cmode_cli(@_, api => 'vserver_peer_show',);
    $Log->exit() if $may_exit;
    return @state_objs;
}

sub _fetch_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    my @state_objs = $pkg->SUPER::_fetch_cmode_zapi(@args,
        $pkg->_fetch_cmode_zapi_opts(),);
    $Log->exit() if $may_exit;
    return @state_objs;
}

sub _super_fetch_cmode_zapi_non_iter {
    $Log->enter() if $may_enter;
    my ($pkg, @opts) = @_;
    my @state_objs = $pkg->SUPER::_fetch_cmode_zapi_non_iter(@opts);
    $Log->exit() if $may_exit;
    return @state_objs;
}

1;
