# $Id$

#
# Copyright (c) 2001-2012 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary VserverServicesDns ComponentState Module
## @author dl-nacl-dev@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::CS::VserverServicesDns

=head1 DESCRIPTION

C<NACL::CS::VserverServicesDns> is a derived class of
L<NACL::CS::ComponentState::ONTAP|lib-NACL-CS-ComponentState-ONTAP-pm>. It represents the
state of an ONTAP VserverServicesDns.  A related class is
L<NACL::C::VserverServicesDns|lib-NACL-C-VserverServicesDns-pm>, which represents access
to an ONTAP VserverServicesDns.

=head1 ATTRIBUTES

The individual pieces of data that are part of the  state of the VserverServicesDns
element are the attributes of the VserverServicesDns ComponentState.

=over

=item C<< "domains" >>

(Array)

Filled in for CMode CLI/ZAPI, 7Mode CLI.

Maps to:

CMode ZAPI:

For "requested_fields", "filter" and Output mapping:

 <net-dns-info>
     <domains>
	     <string>@values</string>
	 </domains>
 </net-dns-info>

7Mode CLI:

For "filter":
Applicable, Filtering will be done by Components.

For "requested_fields",
Not applicable, but the field will be populated in the CS object.

=item C<< "timeout" >>

Filled in for CMode CLI/ZAPI.

Maps to:

CMode ZAPI:

For "requested_fields", "filter" and Output mapping:

 <net-dns-info>
     <timeout>$value</timeout>
 </net-dns-info>

=item C<< "vserver" >>

Filled in for CMode CLI/ZAPI.

Maps to:

CMode ZAPI:

For "requested_fields", "filter" and Output mapping:

 <net-dns-info>
     <vserver-name>$value</vserver-name>
 </net-dns-info>

=item C<< "name_servers" >>

(Array)

Filled in for CMode CLI/ZAPI, 7Mode CLI.

Maps to:

CMode ZAPI:

For "requested_fields", "filter" and Output mapping:

 <net-dns-info>
     <name-servers>
	     <ip-address>@values</ip-address>
	 </name-servers>
 </net-dns-info>

7Mode CLI:

For "filter":
Applicable, Filtering will be done by Components.

For "requested_fields",
Not applicable, but the field will be populated in the CS object.

=item C<< "attempts" >>

Filled in for CMode CLI/ZAPI.

Maps to:

CMode ZAPI:

For "requested_fields", "filter" and Output mapping:

 <net-dns-info>
     <attempts>$value</attempts>
 </net-dns-info>

=item C<< "state" >>

Filled in for CMode CLI/ZAPI, 7Mode CLI.

Maps to:

CMode ZAPI:

For "requested_fields", "filter" and Output mapping:

 <net-dns-info>
     <dns-state>$value</dns-state>
 </net-dns-info>

7Mode CLI:

For "filter":
Applicable, Filtering will be done by Components.

For "requested_fields",
Not applicable, but the field will be populated in the CS object.

=item C<< user_modified >>

User Modified This Row possible value(s) are, true,false

Filled in for CMode CLI.


=back

=cut

package NACL::CS::VserverServicesDns;

use strict;
use warnings;
use Params::Validate qw(validate);
use NACL::ComponentUtils qw(_dump_one);
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Data::Dumper;
use NACL::CS::ComponentState::ZapiSkip qw(make_zapi_skip);
use NACL::CS::ComponentState::ZapiArray qw(make_zapi_array);

use NACL::Exceptions::NoElementsFound qw(:try);

use base 'NACL::CS::ComponentState::ONTAP';

use Class::MethodMaker [
    array  => 'domains',
    scalar => 'timeout',
    scalar => 'vserver',
    array  => 'name_servers',
    scalar => 'attempts',
    scalar => 'state',
    scalar => 'user_modified',
];

=head1 METHODS

=head2 fetch

    my $VserverServicesDns_state   = NACL::CS::VserverServicesDns->fetch(command_interface => $ci, ...);
    my @VserverServicesDns_states = NACL::CS::VserverServicesDns->fetch(command_interface => $ci, ...);

(Class method)
Discovers which elements are present and returns their state in ComponentState
objects. Called in scalar context it returns only one state object, in list
context it returns all state objects.
See L<NACL::CS::ComponentState|lib-NACL-CS-ComponentState-pm/fetch> for a
more detailed description along with a complete explanation of the options
it accepts.

Supports CMode CLI/ZAPI, 7Mode CLI.

Invokes "net-dns-get-iter" API for CMode ZAPI.

Invokes "dns info" command for 7Mode CLI.

=over

=item Exceptions

=over

=item C<NACL::Exceptions::NoElementsFound>

When there are no elements matching the query specified or elements of that type
doesn't exist, then this exception will be thrown.

=back

=back

=cut

sub fetch {
    $Log->enter() if $may_enter;
    my $pkg        = shift;
    my @state_objs = $pkg->SUPER::fetch(
        @_,
        show_cmd => 'vserver services dns show',
        choices  => [
            {   method    => '_fetch_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_fetch_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
            {   method    => '_fetch_7mode_cli',
                interface => 'CLI',
                set       => '7Mode',
            },
        ],
        exception_text => 'No matching vserver services dns(s) found'
    );

    $Log->exit() if $may_exit;
    return wantarray ? @state_objs : $state_objs[0];
} ## end sub fetch

sub _fetch_cmode_cli {
    $Log->enter() if $may_enter;
    my $pkg        = shift;
    my @state_objs = $pkg->SUPER::_fetch_cmode_cli(@_,
        api => 'vserver_services_dns_show');
    $Log->exit() if $may_exit;
    return @state_objs;
} ## end sub _fetch_cmode_cli

sub _fetch_cmode_zapi {
    $Log->enter() if $may_enter;
    my $pkg        = shift;
    my @state_objs = $pkg->SUPER::_fetch_cmode_zapi(
        @_,
        api  => 'net_dns_get_iter',
        copy => [qw(attempts timeout)],
        map  => {
            "vserver" => ['vserver-name'],
            "state"   => ['dns-state'],
            "domains" =>
                [make_zapi_array('domains'), make_zapi_skip('string')],
            "name-servers" => [
                make_zapi_array('name-servers'),
                make_zapi_skip('ip-address')
            ],
        },
    );
    $Log->exit() if $may_exit;
    return @state_objs;
} ## end sub _fetch_cmode_zapi

sub _fetch_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg = shift;
    my %opts = validate @_, $pkg->_fetch_backend_validate_spec();
    my @state_objs;
    my (%temp_attributes, %final_attributes);
    my $apiset = delete $opts{apiset};
    delete $opts{filter};
    my $response = $apiset->dns_info();
    my $output   = $response->get_parsed_output();
    $temp_attributes{'search_domains'} = $output->[0]->{'search_domains'};
    $temp_attributes{'status'}         = $output->[0]->{'status'};
    $temp_attributes{'ip_address'}
        = $output->[0]->{'table'}->[0]->{'ip_address'};
    $pkg->_hash_copy(
        source => \%temp_attributes,
        copy   => [qw(domain)],
        map    => {
            'search_domains' => 'domains',
            'status'         => 'state',
            'ip_address'     => 'name-servers'
        },
        target => \%final_attributes
    );
    my $obj = $pkg->new(command_interface => $opts{command_interface},);
    $obj->_set_fields(row => \%final_attributes);
    push @state_objs, $obj;
    $Log->exit() if $may_exit;
    return @state_objs;
} ## end sub _fetch_7mode_cli

1;
