#
# Copyright (c) 2001-2014 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary NACL::Cleanup::Database::MySQL
## @author dl-nacl-dev@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::Cleanup::Database::MySQL

=cut

package NACL::Cleanup::Database::MySQL;
use strict;
use warnings;
use Params::Validate qw(validate_with :types);
use Tharn qw( param);
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use NATE::ParamSet;
use File::Spec;
use constant MAX_DB_NAME_LEN => 64;
use NACL::GeneralUtils qw(random_name_generator);
use base 'NACL::Cleanup::Database';
use Class::MethodMaker [
    new => [ '-init', 'new' ],
    scalar => [ { -type => 'DBI::db' }, 'dbh' ],
    scalar => 'host',
    scalar => 'port',
];

sub init {
    $Log->enter() if $may_enter;
    my $self = shift;
    my %opts = validate_with(
        params      => \@_,
        spec        => $self->_connection_validate_spec(),
        allow_extra => 1
    );
    my $params = $self->_verify_get_connection_params(%opts);

    while ( my ( $key, $value ) = each %{$params} ) {
        $self->$key($value);
    }
    $self->deploy_schema( $self->should_schema_be_deployed(%opts, %{$params}) );
    $Log->exit() if $may_exit;
}

sub _verify_get_connection_params {
    $Log->enter() if $may_enter;

    my ( $self, %opts ) = @_;
    my %params;
    my $nate_global_params = NATE::ParamSet::param_global;
    $params{port} = $opts{'port'} || $nate_global_params->get('NATE_DB_PORT');
    $params{host} = $opts{'host'} || $nate_global_params->get('NATE_DB_HOST');
    $params{user} = $opts{'user'} || $nate_global_params->get('NATE_DB_USER');
    $params{password} = $opts{'password'}
        || $nate_global_params->get('NATE_DB_PASSWORD');
    $params{dsn}
        = "DBI:mysql:database=$opts{db_name};host=$params{host};port=$params{port}";
    $self->verify_set_dbh(%params);

    $params{db_name} = $self->validate_dbname(
        'db_name'  => $opts{'db_name'},
        'random_db' => $opts{'random_db'}
    );
    $params{dsn}
        = "DBI:mysql:database=$params{db_name};host=$params{host};port=$params{port}";

    return \%params;
    $Log->exit() if $may_exit;

}

sub verify_set_dbh {
    $Log->enter() if $may_enter;
    my ( $self, %opts ) = @_;
    if ( !$self->_is_dbh_valid() ) {
        my $dbh = $self->_get_dbh(%opts);
        $self->dbh($dbh);
        my $xx=$self->_is_dbh_valid();
    }
    $Log->exit() if $may_exit;
}

sub _get_dbh {
    my ($self, %opts) = @_;
    my $dsn_to_metadb = $opts{'dsn'};
    $dsn_to_metadb =~ s/(.*database=)(.*?)(\;.*)/$1information_schema$3/;
    my $dbh
        = DBI->connect( $dsn_to_metadb, $opts{user}, $opts{password},
        { RaiseError => 1, AutoCommit => 0 } )
        or die $DBI::errstr;
    return $dbh;
}

sub _is_dbh_valid {
    $Log->enter() if $may_enter;
    my ($self) = @_;
    $Log->exit() if $may_exit;
    return ( $self->dbh_isset && $self->dbh->ping ) ? 1 : 0;
}

sub _connection_validate_spec {
    return {
        host     => { type => SCALAR, optional => 1 },
        port     => { type => SCALAR, optional => 1 },
        user     => { type => SCALAR, optional => 1 },
        password => { type => SCALAR, optional => 1 },
        db_name  => { type => SCALAR }
    };
}

sub _create_database {
    $Log->enter() if $may_enter;
    my ( $self, %opts ) = @_;
    my $dbh      = $self->dbh();
    my $database = $self->db_name();
    $dbh->do("create database if not exists $database")
        or die "Cannot create the database $database\n" . $dbh->errstr();
    $Log->exit() if $may_exit;
}

sub _drop_database {
    $Log->enter() if $may_enter;
    my ( $self ) = @_;
    my $dbh       = $self->dbh();
    my $database  = $self->db_name();
    my $statement = qq/drop database if exists $database/;
    $dbh->do($statement) or die $dbh->errstr;
    $dbh->disconnect()   or warn $dbh->errstr;
    $Log->exit() if $may_exit;
}

sub should_schema_be_deployed {
    $Log->enter() if $may_enter;
    my ( $self, %opts ) = @_;
    my $deploy_schema = $opts{'deploy_schema'} || $self->deploy_schema();
    my $database      = $opts{'db_name'} || $self->db_name();
    $database = $self->extract_filename_from_path($database);
    $database =~ s/\.db$//;
    my $database_exists
        = $self->_does_database_exists( 'db_name' => $database );
    $deploy_schema = ( $deploy_schema && !$database_exists ) || 0;

    return $deploy_schema;
}
sub _does_database_exists {
    $Log->enter() if $may_enter;
    my ( $self, %opts ) = @_;

    my $database = $opts{'db_name'};
    my $dbh      = $opts{'dbh'} || $self->dbh();

    my $db_query = qq/SELECT IF('$database' IN(SELECT SCHEMA_NAME FROM INFORMATION_SCHEMA.SCHEMATA), 1, 0) AS found/;
    my $exists_or_not = $dbh->selectall_arrayref($db_query)->[0][0];
    if($exists_or_not){
        my $schema_query = qq/SELECT table_name FROM information_schema.tables WHERE table_schema = '$database' AND table_name = 'Primary_Keys'/;
        my $schema_exists_or_not = $dbh->selectall_arrayref($schema_query)->[0][0];
        if(! $schema_exists_or_not ){
            $exists_or_not = 0;
        }
    }

    $Log->exit() if $may_exit;
    return $exists_or_not;
}

sub validate_dbname {
    $Log->enter() if $may_enter;
    my ( $self, %opts ) = @_;
    my $database;
    my $retry = 100;
    my $uuid = $self->_get_nate_uuid();
    my $rn_length = ( MAX_DB_NAME_LEN - length($uuid) - 1 );
    if ( $opts{'random_db'} ) {
        while ($retry) {
            --$retry;
            my $rand_name = random_name_generator(
                prefix => "resourceList",
                size   => '1',
                length => $rn_length,
            );
            $rand_name = $self->_db_name_with_unique_id('name' => $rand_name);
            my $db_name_length = length($rand_name);
            # with length being specified to random_name_generator, it is always
            # ensured that the length would not exceed 31 chars, but below check
            # is in place to recheck in case nate uuid has increased its digits.
            if($db_name_length > 64) {
                $Log->comment("db name - $rand_name whose length -
                $db_name_length characters exceeds the maximum database name
                length supported by MySQL. Hence, trying to regenerate the db
                name.");
                next;
            }
            unless ( $self->_does_database_exists( 'db_name' => $rand_name ) )
            {
                $database = $rand_name;
                last;
            }
        }
    } else {
        $database = $self->extract_filename_from_path( $opts{'db_name'} );
        $database = $self->_db_name_with_unique_id('name' => $database);
    }
    $Log->exit() if $may_exit;
    return $database;
}

sub get_connection_params {

    $Log->enter() if $may_enter;
    my ( $self, %opts ) = @_;
    my $params = $self->_connection_validate_spec();
    my %connect_params;
    foreach my $key ( keys %{$params} ) {
        $connect_params{$key} = $self->{$key} if(exists $self->{$key});
    }

    $Log->exit() if $may_exit;
    return \%connect_params;
}

sub create_schema {
    $Log->enter() if $may_enter;
    my ($self, %opts) = @_;
    $self->_create_database();
    $self->SUPER::create_schema(%opts);
    $Log->exit() if $may_exit;
}


sub DESTROY {
    my $self = shift;
    $self->dbh()->disconnect or die $DBI::errstr if $self->dbh_isset();
}

1;
