#   Copyright (c) 2017 NetApp, Inc., All Rights Reserved.
#   Any use, modification, or distribution is prohibited
#    without prior written consent from NetApp, Inc.

###
# Documentation:
# https://confluence.eng.netapp.com/x/r13S
###

package NACL::Init::Config::ONTAPSelect::Deploy::Appliance::ESX;

use strict;
use warnings;
use Carp qw(croak);
use List::Util qw(first);

use NATE::ParamSet qw(param_global);

use NACL::Init::Utils qw(process_init_args);

use Moo;
extends qw(NATE::Init::Config);

sub setup_args
{
    my $self = shift;

    my @definitions = (
        ##########################
        ## One of These Required
        ##########################
        {
            'name'        => 'DEPLOY_OVA',
            'description' => 'Path to the ONTAP Deploy ova file',
            'required'    => 0,
        },
        {
            'name'        => 'BUILDROOT',
            'description' => 'Path to the ONTAP buildroot - Sets DEPLOY_OVA to '
                           . 'that release nightly',
            'required'    => 0,
        },
        ##########################
        ## Required
        ##########################
        {
            'name'        => 'DNS_SERVER',
            'description' => 'Comma-separated list of DNS server addresses (IPs)',
            'spec'        => 's@',
            'required'    => 1,
        },
        ##########################
        ## Optional
        ##########################
        {
            'name'        => 'DNS_SEARCH',
            'description' => 'FQDNs to search for short hostnames',
            'spec'        => 's@',
            'default'     => [
                'ctl.gdl.englab.netapp.com',
                'gdl.englab.netapp.com',
                'gdl.eng.netapp.com',
                'rtp.netapp.com',
            ],
        },
        {
            'name'        => 'DEPLOY_NETWORKNAME',
            'description' => 'Management network name for deploy',
            'default'     => 'ONTAP-Management',
        },
        {
            'name'        => 'OVFTOOL_INSTALLER',
            'description' => 'Path to the ovftool installer',
            'default'     => '/x/eng/ctl/share/VMware-ovftool-4.2.0-4586971-lin.x86_64.bundle',
        },
    );

    return process_init_args(\@definitions);
}

sub generate_resources {
    my $self = shift;

    # Get the value of the Resources.
    my $resources_param = param_global->get('DEPLOY_APPLIANCE') //
        NATE::BaseException->throw('Missing Mandatory DEPLOY_APPLIANCE');
    my @resources = split(/,|:/, $resources_param);
    $self->resources(\@resources);
}

sub generate_setup_tasks
{
    my $self = shift;

    my $setup_args = $self->setup_args;

    if (! $setup_args->{'DEPLOY_OVA'} && ! $setup_args->{'BUILDROOT'})
    {
        croak("You must provide either DEPLOY_OVA or BUILDROOT");
    }

    $self->add_task(
        'task_name' => 'NACL::Init::Tasks::ONTAPSelect::ESX::InstallOvfTool',
        'alias'     => 'install_ovftool',
        'run_scheme' => 'all_together',
        'task_args'  => $setup_args,
    );

    $self->add_task(
        'task_name'  => 'NACL::Init::Tasks::ONTAPSelect::Server::Power',
        'alias'      => 'power_on_host',
        'task_args'  => {
            'HOST_OS'      => 'esx',
            'POWER_ACTION' => 'on',
        },
    );

    $self->add_task(
        'task_name'  => 'NACL::Init::Tasks::ONTAPSelect::ESX::Host::Reconnect',
        'alias'      => 'reconnect_host',
    );

    $self->add_task(
        'task_name' => 'NACL::Init::Tasks::ONTAPSelect::Deploy::Appliance::ESX::Add',
        'alias'     => 'create',
        'task_args'  => $setup_args,
        'timeout'    => 7200,
    );
}

sub generate_cleanup_tasks
{
    my $self = shift;

    $self->add_task(
        'task_name' => 'NACL::Init::Tasks::ONTAPSelect::Deploy::Appliance::ESX::Delete',
        'alias'     => 'delete',
        'timeout'   => 1200,
    );
}

1;
