#   Copyright (c) 2017 NetApp, Inc., All Rights Reserved.
#   Any use, modification, or distribution is prohibited
#    without prior written consent from NetApp, Inc.

###
# Documentation:
# https://wikid.netapp.com/w/AERO/ONTAP_Select_Projects/init.thpl#NACL::Init::Config::ONTAPSelect::KVM::Node::Ubuntu
###

package NACL::Init::Config::ONTAPSelect::KVM::Node::Ubuntu;

use strict;
use warnings;
use List::Util   qw(first);

use Params::Validate qw(validate :types);

use NATE::ParamSet qw(param_global);

use Moo;
extends qw(NATE::Init::Config);

sub global_args
{
    my $self = shift;

    my @definitions = (
        {
            'name'        => 'KVMADMIN_PATH',
            'description' => 'Path to kvmadmin_cli.py',
            'default'     => '/x/eng/rlse/ontapdeploy/devN/offtap/lib/deploy/src/kvmadmin/kvmadmin_cli.py',
        },
        {
            'name'        => 'MEDIATOR_IP',
            'description' => 'IP address of an ONTAPSelect mediator',
        },
        {
            'name'        => 'CLUSTER_NAME',
            'description' => 'Name of the ONTAPSelect cluster',
            'default'     => eval(
                 "my \$suffix = qx(cat /dev/urandom | tr -cd 'A-F0-9' | head -c 8);
                 chomp(\$suffix); # mutatable
                 return \"C\$suffix\";"
            ),
        },
    );

    param_global->parameters('definitions' => \@definitions);
    param_global->process;

    my @params = map {$_->{'name'}} @definitions;
    my %set = param_global->walk(
        'filter' => sub {
            my ($element) = @_;
            return first {$_ eq $element->longname} @params;
        },
        'fill_callback' => sub {
            my ($element) = @_;
            return $element->longname => $element->value;
        },
    );

    # Must always be this for this config
    $set{'NODE_TYPE'} = 'kvm';

    return \%set;
}

sub generate_resources {
    my ($self) = shift;

    # Get the value of the Resources.
    my $resources_param = param_global->get('NODES') // NATE::BaseException->throw('Missing Mandatory NODES');
    my @resources = split(',', $resources_param);
    $self->resources(\@resources);
}

sub generate_setup_tasks
{
    my $self = shift;

    my $global_args = $self->global_args;

    $self->add_task(
        'task_name'  => 'NACL::Init::Tasks::ONTAPSelect::Node::MediatorSetup',
        'alias'      => 'mediator_setup',
        'run_scheme' => 'all_together',
        'task_args'  => $global_args,
    );
    $self->add_task(
        'task_name'  => 'NACL::Init::Tasks::ONTAPSelect::Node::GenerateJSON',
        'alias'      => 'generate_json',
        'run_scheme' => 'all_together',
        'task_args'  => $global_args,
    );
    $self->add_task(
        'task_name' => 'NACL::Init::Tasks::ONTAPSelect::KVM::Node::DiskPoolCreate',
        'alias'     => 'disk_pool_create',
        'task_args'  => $global_args,
    );
    $self->add_task(
        'task_name' => 'NACL::Init::Tasks::ONTAPSelect::KVM::Host::Ubuntu::NetworkingSetup',
        'alias'     => 'networking_setup',
        'task_args'  => $global_args,
    );
    $self->add_task(
        'task_name' => 'NACL::Init::Tasks::ONTAPSelect::KVM::Node::Create',
        'alias'     => 'create',
        'task_args'  => $global_args,
    );
    $self->add_task(
        'task_name' => 'NACL::Init::Tasks::ONTAPSelect::KVM::Node::PowerOn',
        'alias'     => 'power_on',
        'task_args'  => $global_args,
    );
    $self->add_task(
        'task_name'  => 'NACL::Init::Tasks::ONTAPSelect::Node::ClusterValidate',
        'alias'      => 'cluster_validate',
        'run_scheme' => 'all_together',
        'task_args'  => $global_args,
    );
    $self->add_task(
        'task_name' => 'NACL::Init::Tasks::ONTAPSelect::Node::DiskAssign',
        'alias'     => 'disk_assign',
        'task_args'  => $global_args,
    );
    $self->add_task(
        'task_name' => 'NACL::Init::Tasks::ONTAPSelect::Node::LicenseInstall',
        'alias'     => 'licenses_install',
        'task_args'  => $global_args,
    );
}

sub generate_cleanup_tasks
{
    my $self = shift;

    my $global_args = $self->global_args;

    $self->add_task(
        'task_name' => 'NACL::Init::Tasks::ONTAPSelect::KVM::Node::DomainDestroy',
        'alias'     => 'domain_destroy',
        'task_args'  => $global_args,
    );
    $self->add_task(
        'task_name' => 'NACL::Init::Tasks::ONTAPSelect::KVM::Node::DomainUndefine',
        'alias'     => 'domain_undefine',
        'task_args'  => $global_args,
    );
    $self->add_task(
        'task_name' => 'NACL::Init::Tasks::ONTAPSelect::KVM::Node::PoolDestroy',
        'alias'     => 'pool_destroy',
        'task_args'  => $global_args,
    );
    $self->add_task(
        'task_name' => 'NACL::Init::Tasks::ONTAPSelect::KVM::Node::PoolUndefine',
        'alias'     => 'pool_undefine',
        'task_args'  => $global_args,
    );
    $self->add_task(
        'task_name' => 'NACL::Init::Tasks::ONTAPSelect::KVM::Node::VgRemove',
        'alias'     => 'vg_remove',
        'task_args'  => $global_args,
    );
    $self->add_task(
        'task_name' => 'NACL::Init::Tasks::ONTAPSelect::KVM::Node::BrRemove',
        'alias'     => 'br_remove',
        'task_args'  => $global_args,
    );
}

1;
