#
#   Copyright (c) 2017 NetApp, Inc., All Rights Reserved.
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

package NACL::Init::Tasks::ONTAP::ConfigureFlashOptimized;

use strict;
use warnings;

use NATE::Log qw(log_global);
use NACL::Transit;
use NACL::C::Node;
use Carp qw(confess);
use Moo;

extends qw(NATE::Init::Task);

# need to do this to prevent Moo from clashing with NATE::BaseException
BEGIN {undef *with;}

use NATE::BaseException qw(:try);
use NACL::APISet::Exceptions::InvalidParamValueException;

sub execute {
    my ($self, %args) = @_;

    my $node = $args{resource};

    log_global()->log("Transit to CLI");
    my $t = NACL::Transit->new(name => $node);
    $t->change_state(to => "CLI");

    my $command_interface = NACL::C::Node->new(name => $node);
    my $apiset = $command_interface->apiset(
        category => "Node",
        interface => "CLI",
        set => "Systemshell"
    );

    my $new_value = $command_interface->hostrec()->{FLASH_OPTIMIZED} ? "true" : "false";
    log_global()->log("Requested to set bootarg.init.flash_optimized to $new_value");

    # we need to run all the following commands with 'sudo', so switch to root
    $apiset->set_privilege("privilege-level" => "root");

    # check if the bootarg is already set to the value we want
    my $response;
    my $cant_be_set = 0;
    try {
        $response = $apiset->kenv(variable => "bootarg.init.flash_optimized");
    } catch NACL::APISet::Exceptions::InvalidParamValueException with {
        my $e = shift;
        if ($e->text() =~ /unable to get bootarg/i) {
            # the boot arg does not exist, assume that it is implicitly not set
            # spoof the response to indicate that it is set to false
            $response->{output} = "false";
        } else {
            # some other bad thing happened, rethrow the exception
            $e->throw();
        }
    };

    # See if the desired value is already set, if so, go to the next node
    if ($response->{output} =~ /$new_value/i) {
        log_global()->log("bootarg.init.flash_optimized is already set to " .
                          "$new_value (or it is not set at all) on $node");
        return;
    }

    # set the bootarg and reboot the node
    $response = $apiset->kenv(
        persist => 1,
        variable => "bootarg.init.flash_optimized",
        value => $new_value
    );

    log_global()->log("Rebooting the node for it to take effect");
    $t->change_state(to => "FIRMWARE");
    $t->change_state(to => "CLI");

    return;
}

1;
