#   Copyright (c) 2016 NetApp, Inc., All Rights Reserved.
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.

package NACL::Init::Tasks::ONTAP::NetworkConfig;

use strict;
use warnings;

use NATE::Log qw(log_global);
use NATE::ParamSet qw(param_global);
use Hostrec;

use NACL::Transit;
use NACL::C::Node;
use NACL::C::SecurityLogin;
use NACL::CS::NetworkPort;
use NACL::CS::NetworkInterface;
use NACL::CS::ClusterDate;
use NACL::CS::SystemNodeDate;
use NACL::CS::SecurityLogin;
use NACL::APISet::Exceptions::ResponseException;

use Tie::IxHash;

use Moo;
extends qw(NATE::Init::Task);

# need to do this to prevent Moo from clashing with NATE::BaseException
BEGIN {undef *with;}

use NATE::BaseException qw(:try);

use constant defaults => {
    TIMEZONE => "GMT",
};

sub execute {
    my ($self, %args) = @_;

    my $node = $args{resource};

    my $task_args = $self->task_args;

    my $host = Hostrec->new( id => $node );
    my $t = NACL::Transit->new(name => $host);
    $t->change_state(to => "CLI");

    my $conn = $t->get_connection();
    my $api_set_obj = NACL::APISet->new(
        connection => $conn,
        interface  => 'CLI',
        category   => 'Node',
        set        => 'CMode',
    );

    my $node_obj = NACL::C::Node->new(
        node => $node,
        apiset_must => {connid => "console"}
    );

    NACL::C::SecurityLogin->password(
        command_interface  => $node_obj,
        username           => "admin",
        oldpwd             => "\r",
        newpwd             => "netapp1!",
        newpwd2            => "netapp1!",
    );

    NACL::C::SystemNode->rename(
        command_interface => $node_obj,
        node              => "local",
        newname           => $node
    );

    $node_obj->apiset()->execute_raw_command(
        command => "debug smdb table node show",
    );

    $node_obj->apiset()->security_login_new(
        username    => 'admin',
        application => 'ssh',
        authmethod  => 'password',
    );

    # cluster mgmt port mtu change
    # network port modify -port e0d -node smoke10-vsim-ucs10b6p -mtu 1500
    my @nets = $host->net_sort(include=>{nettype=>'cluster_mgmt'});
    foreach my $net (@nets) {
        NACL::C::NetworkPort->modify(
            command_interface => $node_obj,
            port              => $net->{interface}, # cluster mgmt port from hostrec
            node              => $node,
            mtu               => $net->{mtu}, # mtu from inter cluster mtu
        );
    }

    # we are not changing mtu to 9000. Just using the value from tharnhost files.
    my $cluster_ports_obj = NACL::CS::NetworkIpspace->fetch(
        command_interface => $node_obj,
        filter            => { ipspace => "Cluster" },
    );

    my $cluster_ports = $cluster_ports_obj->ports();
    @nets = $host->net_sort(include=>{nettype=>'cluster'});
    foreach my $port (@{$cluster_ports}) {
        if ($port =~ /\S+\:(\S+$)/) {
            NACL::C::NetworkPort->modify(
                command_interface => $node_obj,
                port              => $1,
                node              => $node,
                mtu               => "9000",
            );
        }
    }

    # network int create -lif mgmt
    @nets = $host->net_sort(include=>{nettype=>'management', family => "inet4"});
    foreach my $net (@nets) {
        my $interface = $net->{interface};
        my $address   = $net->{address};
        my $netmask   = $net->{netmask};
        my $gateway   = $net->{gateway};

        # create the mgmt life for this node
        if ($interface and $address and $netmask) {
            try {
                NACL::C::NetworkInterface->create(
                    command_interface => $node_obj,
                    vserver           => "Default",
                    lif               => "mgmt1",
                    role              => "node-mgmt",
                    address           => $address,
                    netmask           => $netmask,
                    'home-node'       => $node,
                    'home-port'       => $interface);
            } catch NACL::C::Exceptions::NetworkInterface::AlreadyExists with {
                my $ex = shift;
                log_global()->comment("Management lif already exists, reusing.");
            };

            # attempt to create a network route for the gateway of this interface
            if ($gateway) {
                try {
                    NACL::C::NetworkRoute->create(
                        command_interface => $node_obj,
                        vserver           => "Default",
                        gateway           => $gateway,
                        destination       => "0.0.0.0/0", # defaulting the destination
                    );
                } catch NACL::C::Exceptions::NetworkRoute::AlreadyExists with {
                    my $ex = shift;
                    log_global()->comment("A network route already exists for gateway: $gateway");
                };
            }
            # if we succeeded in creating a mgmt lif for this node, then we are done searching
            # through the interfaces.
            last;
        }
    }


    # system node modify -node smoke10-vsim-ucs10b6o -owner smoke -location smoke_bubble
    my $user = $ENV{USER} || $ENV{LOGNAME} || getlogin() || ( getpwuid($<) )[0];
    my $site = param_global()->get("SITE");
    NACL::C::SystemNode->modify(
        command_interface => $node_obj,
        node              => $node,
        owner             => $user,
        location          => $site,
    );

    # cluster date modify -timezone $timezone
    NACL::C::ClusterDate->modify(
        command_interface => $node_obj,
        timezone          => $task_args->{TIMEZONE},
    );
}

1;
