#   Copyright (c) 2017 NetApp, Inc., All Rights Reserved.
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.

package NACL::Init::Tasks::ONTAPSelect::Deploy::Appliance::ESX::Add;

use strict;
use warnings;
use Carp qw(croak);
use Cwd  qw(abs_path);
use Data::Dumper qw(Dumper);
use IO::File;
use Net::Ping;

use Mojo::Util qw(url_escape);

# vmware
use VMware::VICommon;
use VMware::VIMRuntime;
use VMware::VIM25Stub;
use VMware::VILib;
use VMware::VIExt;

use Hostrec;
use NATE::Events   qw(eventloop);
use NATE::Log      qw(log_global);
use NATE::Process;
use NATE::Util::Appliance qw(buildroot_to_version);

use Moo;
extends qw(NATE::Init::Task);

use NACL::Init::Utils::ONTAPSelect::Deploy qw(get_deploy_apiclient);
use NACL::Init::Utils::VMware qw(rot13);

sub execute
{
    my ($self, %args) = @_;

    my $deploy_appliance = $args{'resource'};

    my $log       = log_global;
    my %task_args = %{$self->task_args};

    my $ova;
    if ($task_args{'DEPLOY_OVA'})
    {
        $ova = $task_args{'DEPLOY_OVA'};
    }
    elsif ($task_args{'BUILDROOT'})
    {
        my $p4info_file = "$task_args{'BUILDROOT'}/final/bedrock/export/common/p4info/p4info";
        if (-f $p4info_file)
        {
            my $fh = IO::File->new($p4info_file, 'r');
            while (my $line = <$fh>)
            {
                if ($line =~ m/^P4BRANCH='DOT\/(.*)'/)
                {
                    my $branch = $1;
                    $ova = "/x/eng/rlse/ontapdeploy/${branch}N/offtap/lib/deploy/bedrock/"
                        . "export/linux64.debug/final/installer/ONTAPdeploy.ova";
                    last;
                }
            }
            $fh->close;
        }
        else
        {
            croak("Unable to find $p4info_file to determine what nightly to "
                . "use for the OVA.");
        }
    }

    if (! -e $ova)
    {
        croak("Unable to find OVA file: $ova");
    }

    $log->info("OVA File: $ova");

    my $hostrec_da = Hostrec->new('id' => $deploy_appliance);

    my $host = $hostrec_da->simhost;
    $host .= '.' . $hostrec_da->{'DOMAIN'};

    my $vcenter = $hostrec_da->{'VCENTER'};
    my $hostrec_vc = Hostrec->new('id' => $vcenter);

    my $vc_password = rot13('string' => $hostrec_vc->default_password);

    $log->info("Logging in to $vcenter vcenter...");
    my $vc = Vim::login(
        'service_url' => 'https://' . $vcenter . '/sdk/vimService',
        'user_name'   => $hostrec_vc->default_username,
        'password'    => $vc_password,
    );
    $log->info("Logged in to $vcenter.");

    my $hostSystem = $vc->find_entity_view(
        'view_type' => 'HostSystem',
        'filter'    => {'name' => $host},
    );

    my $clusterComputeResource = Vim::get_view('mo_ref' => $hostSystem->parent);
    my $folder                 = Vim::get_view('mo_ref' => $clusterComputeResource->parent);
    my $datacenter             = Vim::get_view('mo_ref' => $folder->parent);

    $log->info("Host: $host");
    $log->info('ClusterComputeResource: ' . $clusterComputeResource->name);
    $log->info('Folder: ' . $folder->name);
    $log->info('Datacenter: ' . $datacenter->name);

    my $netrec = $hostrec_da->net_best;

    my $vc_addr = 'vi://' . $hostrec_vc->default_username . ':' . $vc_password . '@';
    $vc_addr   .= $hostrec_vc->default_ip . '/';
    $vc_addr   .= join('/', (url_escape($datacenter->name), url_escape($folder->name), url_escape($clusterComputeResource->name)));

    my $proc = NATE::Process->new(
        'codespec' => '/usr/bin/ovftool',
        'onexit'   => \&NATE::Process::on_exit_die_on_non_zero_exit_status,
        'args'     => [
            '--datastore=' . $hostrec_da->{'DATASTORE'},
            '--diskMode=thin',
            '--net:"ONTAP Select Deploy VM Network=' . $task_args{'DEPLOY_NETWORKNAME'} .'"',
            '--name=' . $deploy_appliance,
            '--prop:product_company=NetApp',
            '--prop:password=' . $hostrec_da->default_password,
            '--prop:ipAddress=' . $netrec->{'address'},
            '--prop:netMask=' . $netrec->{'netmask'},
            '--prop:gateway=' . $netrec->{'gateway'},
            '--prop:hostName=' . $deploy_appliance,
            '--prop:primaryDNS=' . $task_args{'DNS_SERVER'}->[0],
            '--prop:secondaryDNS=' . $task_args{'DNS_SERVER'}->[1],
            '--prop:searchDomains="' . join(' ', @{$task_args{'DNS_SEARCH'}}) . '"',
            '--noSSLVerify',
            '--powerOn',
            '--overwrite',
            $ova,
            $vc_addr,
        ],
    );

    $log->comment(Dumper($proc));
    $proc->start;
    $proc->wait;
    $proc->destroy;

    _wait_for_on($hostrec_da);
}

sub _wait_for_on
{
    my $hostrec = shift;

    my $log = log_global;

    while (1)
    {
        my $exit_status = system('ping', '-c', 5, $hostrec->default_ip);
        if ($exit_status != 0)
        {
            $log->info("Still unresponsive, waiting for it to "
                . 'be responsive...');
            eventloop('seconds' => 5);
        }
        else
        {
            $log->info("Reponsive, assuming it is powered on.");
            last;
        }
    }

    $log->info("Waiting until we can login to the host...");
    while (1)
    {
        while (1)
        {
            Connectrec->new(
                'id'       => $hostrec->id,
                'host'     => $hostrec->id,
                'conntype' => 'ssh-persist_service',
                'connect'  => 1,
            );
            if (my $err = $@)
            {
                $log->info("Unable to login still...");
                eventloop('seconds' => 30);
            }
            else
            {
                $log->info("Successfully logged in.");
                return 1;
            }
        }
    }
}

1;
