#   Copyright (c) 2017 NetApp, Inc., All Rights Reserved.
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.

package NACL::Init::Tasks::ONTAPSelect::Deploy::Cluster::KVM::GenerateJSON;

# core
use strict;
use warnings;
use Carp         qw(croak);
use Cwd          qw(abs_path getcwd);
use File::Copy;
use File::Spec::Functions qw(catfile);
use List::Util   qw(first);
use IO::File;

# cpan
use Moo;
extends qw(NATE::Init::Task);
use JSON::XS qw(encode_json);

#nate
use NATE::Log qw(log_global);
use NATE::ParamSet qw(param_global);

# nacl
use NACL::Init::Utils::JSON qw(json_true_false);
use NACL::Init::Utils::ONTAPSelect::Deploy qw(get_deploy_apiclient);
use NACL::Init::Utils::STB qw(in_a_stb copy_file_to_stb);

sub execute
{
    my ($self, %args) = @_;

    my $log       = log_global;
    my @nodes     = @{$args{'resource'}};
    my %task_args = %{$self->task_args};

    my $n = 0;

    my $nodecount = scalar(@nodes);

    my $client = get_deploy_apiclient($task_args{'DEPLOY_APPLIANCE'});

       # Determine if Deploy is promoted or not
    my $version = $client->GET('path' => 'version');
    my $promoted = $version->{'production'};

    # Pick the right image based on these parameters:
    #   ONTAP_IMAGE
    #   ONTAP_IMAGE_CONSOLE
    $log->comment('ONTAP_IMAGE='.$task_args{'ONTAP_IMAGE'}) if ($task_args{'ONTAP_IMAGE'});
    $log->comment('ONTAP_IMAGE_CONSOLE='.$task_args{'ONTAP_IMAGE_CONSOLE'}) if ($task_args{'ONTAP_IMAGE_CONSOLE'});

    my $ontap_image;
    if ($task_args{'ONTAP_IMAGE'})
    {
        $ontap_image = $task_args{'ONTAP_IMAGE'};
    }
    else
    {
        my $branch;
        my $p4info_file = param_global->get('BUILDROOT') . '/final/bedrock/export/common/p4info/p4info';
        if (-f $p4info_file)
        {
            my $fh = IO::File->new($p4info_file, 'r');
            while (my $line = <$fh>)
            {
                if ($line =~ m/^P4BRANCH='DOT\/(.*)'/)
                {
                    $branch = $1;
                    last;
                }
            }
            $fh->close;
        }
        else
        {
            croak("Unable to find $p4info_file to determine what nightly to "
                . "use for the image.");
        }

        my $images = $client->GET('path' => 'images');
        if ($promoted)
        {
            $ontap_image = $images->{'default_image'};
        }
        elsif ($task_args{'ONTAP_IMAGE_CONSOLE'})
        {
            my $non_promo_image = first {$_ !~ m/Px/ && $_ =~ m/$task_args{ONTAP_IMAGE_CONSOLE}/ && $_ =~ m/$branch/i} @{$images->{'images'}};
            if ($non_promo_image)
            {
                $ontap_image = $non_promo_image;
            }
            else
            {
                $log->warn("Unable to find an image with a console type of $task_args{ONTAP_IMAGE_CONSOLE}.  Will use the serialconsole version of the default.");

                # Load the serial version of the default_image then
                $ontap_image = $images->{'default_image'};
                $ontap_image =~ s/-.*/-serialconsole/;
            }
        }
    }

    my $template = {
        'instance_type'       => $task_args{'INSTANCE_TYPE'},
        '__internal_console_port' => 7000,
        'name'                => $task_args{'CLUSTER_NAME'},
        'ntp_servers'         => $task_args{'NTP_SERVER'},
        'inhibit_rollback'    => json_true_false('true'),
        'ontap_image_version' => $ontap_image,
        'nodes'               => [],
        'dns_info'            => {
            'dns_ips' => $task_args{'DNS_SERVER'},
            'domains' => $task_args{'DNS_SEARCH'},
        },
    };

    for my $node (@nodes)
    {
        my $hostrec_partner;
        my $hostrec_node = Hostrec->new('id' => $node);

        my $host = $hostrec_node->simhost;
        my $hostrec_host = Hostrec->new('id' => $host);
        $log->log("Host: $host");
        $log->log("Node: $node");

        my $domain = $hostrec_host->{'DOMAIN'};

        my @c_mgmts = $hostrec_node->net_filter('include' => {'family' => 'inet4', 'nettype' => 'cluster_management'});
        my @n_mgmts = $hostrec_node->net_filter('include' => {'family' => 'inet4', 'nettype' => 'management'});

        my $c_mgmt  = $c_mgmts[0];
        my $n_mgmt = $n_mgmts[0];

        if ($n == 0)
        {
            $template->{'cluster_mgmt_ip'} = $c_mgmt->{'address'};
            $template->{'netmask'} = $c_mgmt->{'netmask'};
            $template->{'gateway'} = $c_mgmt->{'gateway'};
            $template->{'admin_password'} = $hostrec_node->default_password;
        }

        if ($nodecount > 1)
        {
            if ($n % 2)
            {
                $log->log("Partner: " . $nodes[$n-1]);
                $hostrec_partner = Hostrec->new('id' => $nodes[$n-1]);
            }
            else
            {
                $log->log("Partner: " . $nodes[$n+1]);
                $hostrec_partner = Hostrec->new('id' => $nodes[$n+1]);
            }
        }

        my $network = {
            'mgmt_network' => {
                'name' => $hostrec_host->{'KVM_MGMT_NETWORK_NAME'},
            },
            'data_network' => {
                'name' => $hostrec_host->{'KVM_DATA_NETWORK_NAME'},
            },
            'internal_network' => {
                'name' => $hostrec_host->{'KVM_INTERNAL_NETWORK_NAME'},
            },
        };

        # if single bridge, tagging required on all 3
        # otherwise tagging only required on data
        if ($hostrec_host->{'KVM_MGMT_NETWORK_NAME'} eq $hostrec_host->{'KVM_DATA_NETWORK_NAME'} &&
            $hostrec_host->{'KVM_MGMT_NETWORK_NAME'} eq $hostrec_host->{'KVM_INTERNAL_NETWORK_NAME'})
        {
            $log->info("Single bridge, tagging vlans on all networks");
            if (defined($hostrec_host->{'MANAGEMENT_VLANID'}))
            {
                $network->{'mgmt_network'}->{'vlan_id'} = int($hostrec_host->{'MANAGEMENT_VLANID'}) 
            }
            if (defined($hostrec_host->{'EXTERNAL_VLANID'}))
            {
                $network->{'data_network'}->{'vlan_id'} = int($hostrec_host->{'EXTERNAL_VLANID'});
            }
            if (defined($hostrec_host->{'INTERNAL_VLANID'}))
            {
                $network->{'internal_network'}->{'vlan_id'} = int($hostrec_host->{'INTERNAL_VLANID'});
            }
        }
        else
        {
            $log->info("Multi-bridge, tagging only data network");
            if (defined($hostrec_host->{'EXTERNAL_VLANID'}))
            {
                $network->{'data_network'}->{'vlan_id'} = int($hostrec_host->{'EXTERNAL_VLANID'});
            }
        }

        # KVM need vlans tagged
        my $storage = {
            'name'     => 'ontap-select-storage-pool',
            'is_ssd'   => 0,
        };

        my $conn = Connectrec->new(
            'id'       => $host,
            'host'     => $host,
            'template' => 'kvm_ssh',
        );
        my $output = $conn->execute(
            'command'      => ['virsh', 'pool-info', 'ontap-select-storage-pool'],
            'check_status' => 1,
        );
        for my $line (split(/\n/, $output))
        {
            if ($line =~ m/Capacity:\s+([0-9\.]+)\s([A-Za-z]+)/)
            {
                my $number = $1;
                my $unit   = $2;

                if ($unit eq 'TiB')
                {
                    $storage->{'capacity'} = int($number * 1024 * 0.80);
                }
                elsif ($unit eq 'GiB')
                {
                    $storage->{'capacity'} = int($number * 0.80);
                }
                else
                {
                    croak("Unit was $unit which was unexpected (Expecting TiB or GiB)!");
                }

                last;
            }
        }

        my $node_tpl = {
            'name'               => $node,
            'host'               => $hostrec_node->{'simhost'} . ".$domain",
            'node_mgmt_ip'       => $n_mgmt->{'address'},
            'host_configuration' => {
                'serial_number'    => ('32000000' . $n + 22) . '', # has to be a string, the only way to "type-cast" this
                'instance_type'    => $task_args{'INSTANCE_TYPE'},
                'eval'             => json_true_false('false'),
                'location'         => $task_args{'HOST_LOCATION'},
                'storage_pool'     => [$storage],
                %{$network},
            },
        };

        if ($nodecount > 1)
        {
            $node_tpl->{'mirror'} = $hostrec_partner->id;
        }

        push(@{$template->{'nodes'}}, $node_tpl);

        $n++;
    }

    my $filename = 'cluster.json';
    my $filepath = catfile(param_global->get('TOP_LOGDIR'), $filename);
    $filepath = abs_path($filepath);
    my $json = encode_json($template);
    my $fh = IO::File->new($filepath, 'w');
    my $failure = $!;
    if ($fh)
    {
        print({$fh} $json);
        $fh->close;
        $log->log($json);
        $log->log("Wrote file to: $filepath");

        # Move json file to testbed directory if within an STB testbed
        if (in_a_stb())
        {
            copy_file_to_stb('source' => $filepath, 'dest' => 'cluster.json');
        }
    }
    else
    {
        if ($failure)
        {
            croak("Failed to write json file with error: $failure");
        }
        else
        {
            croak("Failed to write json file for an unknown reason.");
        }
    }
}

1;
