#   Copyright (c) 2017 NetApp, Inc., All Rights Reserved.
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.

package NACL::Init::Tasks::ONTAPSelect::ESX::Host::Add;

use strict;
use warnings;
use Carp         qw(croak);
use Digest::SHA  qw(sha1_hex);
use File::Temp   qw(tempfile);
use MIME::Base64 qw(decode_base64);

use Params::Validate qw(validate :types);

# vmware
use VMware::VICommon;
use VMware::VIMRuntime;
use VMware::VIM25Stub;
use VMware::VILib;
use VMware::VIExt;

use Moo;
extends qw(NATE::Init::Task);
# need to do this to prevent Moo from clashing with NATE::BaseException
BEGIN {undef *with;}

use Connectrec;
use Hostrec;
use NATE::BaseException qw(:try);
use NATE::Log           qw(log_global);

use NACL::Init::Utils::VMware qw(wait_for_task rot13);

sub execute
{
    my ($self, %args) = @_;

    my $resource = $args{'resource'};
    my $log = log_global;

    my $hostrec_resource = Hostrec->new('id' => $resource);
    my $hostrec;

    if ($hostrec_resource->simhost)
    {
         $hostrec = Hostrec->new('id' => $hostrec_resource->simhost);
    }
    else
    {
        $hostrec = $hostrec_resource;
    }

    my $host = $hostrec->id . '.' . $hostrec->{'DOMAIN'};

    $log->info("Logging in to $hostrec->{'VCENTER'} vcenter...");
    my $hostrec_vc = Hostrec->new('id' => $hostrec->{'VCENTER'});
    my $vc = Vim::login(
        'service_url' => 'https://' . $hostrec->{'VCENTER'} . '/sdk/vimService',
        'user_name'   => $hostrec_vc->default_username,
        'password'    => rot13('string' => $hostrec_vc->default_password),
    );
    $log->info("Logged in to $hostrec->{'VCENTER'}.");

    eval {
        my $hostSystem = $vc->find_entity_view(
            'view_type' => 'HostSystem',
            'filter'    => {'name' => $host},
        );
    };
    if (! $@)
    {
        $log->info("No error while trying to find $host HostSystem. It must already exist.");
        return;
    }

    # This is just to get connection info, do not actually connect
    my $conn = Connectrec->new(
        'host'     => $hostrec,
        'template' => 'esx_ssh',
        'connect'  => 0,
    );

    $log->info("Retrieving $hostrec->{'VCENTER_DATACENTER'} Datacenter view...");
    my $dc_view = $vc->find_entity_view(
        'view_type' => 'Datacenter',
        'filter'    => {'name' => $hostrec->{'VCENTER_DATACENTER'}},
    );
    $log->info("Received $hostrec->{'VCENTER_DATACENTER'} Datacenter view.");

    $log->info("Retrieving $hostrec->{'VCENTER_CLUSTER'} Cluster view...");
    my $cluster_view = $vc->find_entity_view(
        'view_type' => 'ClusterComputeResource',
        'filter'    => {'name' => $hostrec->{'VCENTER_CLUSTER'}},
    );
    $log->info("Received $hostrec->{'VCENTER_CLUSTER'} Cluster.");

    $log->info("Retrieving SSL thumbprint for $host HostSystem...");
    my $cert = get_esx_certificate($host, $conn->username, $conn->password);
    my $ssl_thumbprint = generate_ssl_thumbprint('pem' => $cert);
    $log->info("Received SSL thumbprint for $host HostSystem.");

    my $hostconnect_spec = HostConnectSpec->new(
        'hostName'      => $host,
        'userName'      => $conn->username,
        'password'      => $conn->password,
        'force'         => 'true',
        'sslThumbprint' => $ssl_thumbprint,
    );

    $log->info("Registering $host HostSystem...");
    my $addhost_task = $cluster_view->AddHost_Task(
        'spec'        => $hostconnect_spec,
        'asConnected' => 'true',
    );
    wait_for_task('task' => $addhost_task, 'message' => "Registered $host HostSystem.");

    $log->info("Retrieving $host HostSystem view...");
    my $host_view = $vc->find_entity_view(
        'view_type' => 'HostSystem',
        'filter'    => {'name' => $host},
    );
    $log->info("Retrieved $host HostSystem view.");
}

sub get_esx_certificate
{
    my ($host, $user, $pass) = @_;

    my $ua = LWP::UserAgent->new;
    my $realm_name = "VMware HTTP server";
    $ua->credentials("$host:443", $realm_name, $user => $pass);
    my $response = $ua->get("https://$host:443/host/ssl_cert");

    croak $response->status_line unless $response->is_success;

    return $response->decoded_content;
}

sub generate_ssl_thumbprint {
    my %args = @_;

    my ($pem, $der, $digest, $sslthumbprint);
    $pem = delete($args{'pem'});

    # Strip PEM tags to get Base64 encoded certificate data
    $pem =~ s/-{1,}(BEGIN|END) CERTIFICATE-{1,}//g;

    # Convert PEM to DER (decode Base64)
    $der = decode_base64($pem);

    # Generate SHA1 hex digest
    $digest = sha1_hex($der);

    # Format thumbprint
    $sslthumbprint = "";
    for (my $i=0; $i < length($digest); $i+=2)
    {
        my $substring = substr($digest, $i, 2);
        $sslthumbprint .= uc($substring);
        unless ($i >= 38)
        {
            $sslthumbprint .= ":";
        }
    }

    return $sslthumbprint;
}

1;
