#   Copyright (c) 2017 NetApp, Inc., All Rights Reserved.
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.

package NACL::Init::Tasks::ONTAPSelect::ESX::Host::NetworkCreate;

use strict;
use warnings;
use Carp qw(croak);

# vmware
use VMware::VICommon;
use VMware::VIMRuntime;
use VMware::VIM25Stub;
use VMware::VILib;
use VMware::VIExt;

use Moo;
extends qw(NATE::Init::Task);
# need to do this to prevent Moo from clashing with NATE::BaseException
BEGIN {undef *with;}

use Hostrec;
use NATE::BaseException qw(:try);
use NATE::Log           qw(log_global);

use NACL::Init::Utils::VMware qw(rot13);

use constant 'defaults' => {
    'ESX_VSWITCH_NAME'  'vSwitchSdot',
};

sub execute
{
    my ($self, %args) = @_;

    my $resource = $args{'resource'};
    my $log = log_global;
    my %task_args = %{$self->task_args};

    my $hostrec_resource = Hostrec->new('id' => $resource);
    my $hostrec;

    if ($hostrec_resource->simhost)
    {
         $hostrec = Hostrec->new('id' => $hostrec_resource->simhost);
    }
    else
    {
        $hostrec = $hostrec_resource;
    }

    my $host = $hostrec->id . '.' . $hostrec->{'DOMAIN'};

    $log->info("Logging in to $hostrec->{'VCENTER'} vcenter...");
    my $hostrec_vc = Hostrec->new('id' => $hostrec->{'VCENTER'});
    my $vc = Vim::login(
        'service_url' => 'https://' . $hostrec->{'VCENTER'} . '/sdk/vimService',
        'user_name'   => $hostrec_vc->default_username,
        'password'    => rot13('string' => $hostrec_vc->default_password),
    );
    $log->info("Logged in to $hostrec->{'VCENTER'}.");

    my $hostSystem = $vc->find_entity_view(
        'view_type' => 'HostSystem',
        'filter'    => {'name' => $host},
    );

    my $networkSystem = $vc->get_view('mo_ref' => $hostSystem->configManager->networkSystem);

    my @inics = split(',', $hostrec->{'ESX_INTERNAL_PNICS'} // '');
    my @enics = split(',', $hostrec->{'ESX_EXTERNAL_PNICS'} // '');
    my @nics  = (@inics, @enics);
    my $vswitchSpec = HostVirtualSwitchSpec->new(
        'numPorts' => 128,
        'mtu'      => 9000,
        'bridge'   => HostVirtualSwitchBondBridge->new(
            'nicDevice' => \@nics,
        ),
    );

    eval {
        $networkSystem->AddVirtualSwitch(
            'vswitchName' => $task_args{'ESX_VSWITCH_NAME'},
            'spec'        => $vswitchSpec,
        );
    };
    if ($@ && $@ !~ m/already exists/)
    {
        croak("Something bad happened while adding the vSwitch: $@");
    }

    my $internalNetworkPolicy;
    my $externalNetworkPolicy;
    if (scalar(@enics) && ! scalar(@inics))
    {
        #####################################################################
        # 2 NIC Configuration
        #####################################################################
        $externalNetworkPolicy = new HostNetworkPolicy;
        $internalNetworkPolicy = new HostNetworkPolicy;
    }
    elsif (scalar(@enics) && scalar(@inics))
    {
        #####################################################################
        # 4 NIC Configuration
        #####################################################################
        $externalNetworkPolicy = HostNetworkPolicy->new(
            'nicTeaming' => HostNicTeamingPolicy->new(
                'nicOrder' => HostNicOrderPolicy->new(
                    'activeNic' => \@enics,
                ),
            ),
        );
        $internalNetworkPolicy = HostNetworkPolicy->new(
            'nicTeaming' => HostNicTeamingPolicy->new(
                'nicOrder' => HostNicOrderPolicy->new(
                    'activeNic' => \@inics,
                ),
            ),
        );
    }

    $networkSystem->AddPortGroup(
        'portgrp' => HostPortGroupSpec->new(
            'name'        => $hostrec->{'ESX_DATA_NETWORK_NAME'},
            'vswitchName' => $task_args{'ESX_VSWITCH_NAME'},
            'vlanId'      => $hostrec->{'EXTERNAL_VLANID'},
            'policy'      => $externalNetworkPolicy,
        ),
    );

    $networkSystem->AddPortGroup(
        'portgrp' => HostPortGroupSpec->new(
            'name'        => $hostrec->{'ESX_INTERNAL_NETWORK_NAME'},
            'vswitchName' => $task_args{'ESX_VSWITCH_NAME'},
            'vlanId'      => $hostrec->{'INTERNAL_VLANID'},
            'policy'      => $internalNetworkPolicy,
        ),
    );

    $networkSystem->AddPortGroup(
        'portgrp' => HostPortGroupSpec->new(
            'name'        => $hostrec->{'ESX_MGMT_NETWORK_NAME'},
            'vswitchName' => $task_args{'ESX_VSWITCH_NAME'},
            'vlanId'      => $hostrec->{'MANAGEMENT_VLANID'},
            'policy'      => $externalNetworkPolicy,
        ),
    );
}

1;
