#   Copyright (c) 2017 NetApp, Inc., All Rights Reserved.
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.

package NACL::Init::Tasks::ONTAPSelect::Node::ClusterValidate;

use strict;
use warnings;
use Carp qw(croak);

use NATE::Log    qw(log_global);
use NATE::Events qw(eventloop);

use NACL::Init::Utils::ONTAPSelect::KVMadmin qw(get_host_sub_shell);

use Moo;
extends qw(NATE::Init::Task);

use constant 'defaults' => {
    'KVM_PASS'                 => 'netapp1!',
    'KVMADMIN_PATH'            => '/x/eng/rlse/vsa/devsDOTN/kvmadmin/kvmadmin_cli.py',
    'CREATE_CLUSTER'           => 'TRUE',
    'VALIDATE_CLUSTER_TIMEOUT' => 3600,
    'CREATE_NODE'              => undef,
};

sub execute
{
    my ($self, %args) = @_;

    my @nodes = @{$args{'resource'}};

    my $log = log_global;

    my %task_args = %{$self->task_args};

    if ($task_args{'CREATE_CLUSTER'} eq 'FALSE')
    {
        $log->info('Was asked to not create the cluster.  No need to validate cluster.');
        return 0;
    }

    my $timeout = $task_args{'VALIDATE_CLUSTER_TIMEOUT'};

    my $create_node = $task_args{'CREATE_NODE'} // $nodes[0];

    my $elapsed_time = 0;
    my $healthy = 0;
    my $expected_nodes_found = scalar(@nodes);

    while ($elapsed_time < $timeout && !$healthy)
    {
        my $start_time = time;
        $healthy = 1;
        eval
        {
            require NACL::C::Node;
            require NACL::CS::ClusterIdentity;
            my $n = NACL::C::Node->find;
            my $c = NACL::CS::ClusterIdentity->fetch('command_interface' => $n);
            $log->info('ClusterName: ' . $c->name);

            my $apiset = NACL::APISet->new(
                'name' => $create_node,
                'interface' => 'CLI',
                'set' => 'CMode',
                'category' => 'Node',
            );
            my $response = $apiset->cluster_show();
            my $parsed_outputs = $response->get_parsed_output;

            for my $unparsed (@{$response->{'not_parsed'}})
            {
                if ($unparsed =~ m/warning|fail|failure|warn/i)
                {
                    $healthy = 0;
                    $log->info('Found a warning or failure in "cluster show" output: ');
                    $log->info($unparsed);
                }
            }

            my $found_nodes = 0;
            for my $parsed_output (@{$parsed_outputs})
            {
                # Make sure the right node is part of the cluster
                if (grep($parsed_output->{'node'}, @nodes))
                {
                    $found_nodes++;
                }

                if ($parsed_output->{'health'} ne 'true')
                {
                    $log->info('Health for ' . $parsed_output->{'node'} . ' is false');
                    $healthy = 0;
                }

                if ($parsed_output->{'eligibility'} ne 'true')
                {
                    $log->info('Eligibility for ' . $parsed_output->{'node'} . ' is false');
                    $healthy = 0;
                }
            }

            if ($found_nodes ne $expected_nodes_found)
            {
                $log->info("Found an unexpected number of nodes: Expected: $expected_nodes_found Found: $found_nodes");
                $healthy = 0;
            }
        };
        if (my $err = $@)
        {
            $log->info($err);
            $healthy = 0;
            $log->info('Cluster is not healthy yet');
        }

        $log->info('Elapsed: ' . $elapsed_time . ', sleeping 30 seconds.');
        $elapsed_time = (time - $start_time) + $elapsed_time;
        eventloop('seconds' => 30);
    }

    if ($elapsed_time >= $timeout)
    {
        croak('Failed waiting for cluster to be healthy');
    }
}

1;
