#   Copyright (c) 2017 NetApp, Inc., All Rights Reserved.
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.

package NACL::Init::Tasks::ONTAPSelect::Node::GenerateJSON;

use strict;
use warnings;
use Carp qw(croak);
use Cwd qw(abs_path);
use Data::Dumper qw(Dumper);
use File::Spec::Functions qw(catfile);
use IO::File;
use List::Util qw(first);

use JSON::XS qw(encode_json decode_json);

use NATE::BaseException      qw(try);
use NATE::Events             qw(eventloop);
use NATE::Log                qw(log_global);
use NATE::ParamSet           qw(param_global);
use NATE::Process;
use NATE::ServiceAPI::KeyVal qw(keyval_global);

use NACL::Init::Utils::ONTAPSelect::KVMadmin qw(get_host_sub_shell);

use Moo;
extends qw(NATE::Init::Task);

use constant 'defaults' => {
    'ESX_EXTERNAL_NETWORKNAME'   => 'ONTAP-External',
    'ESX_INTERNAL_NETWORKNAME'   => 'ONTAP-Internal',

    'KVM_EXTERNAL_NETWORKNAME'   => 'ontap-br',
    'KVM_INTERNAL_NETWORKNAME'   => 'ontap-br',

    'CPU_COUNT'                  => 4,
    'MEM_SIZE_MB'                => 16384,
    'SERIAL_PORT_NAMES'          => ['console', 'gdb'],

    'BUILDROOT'                  => '/x/eng/rlse/DOT/devN',
    'VARIANT'                    => 'x86_64.ovf.sdot',
    'MEDIATOR_ROOT'              => '/x/eng/rlse/vsa/devsDOTN',
    'IMAGE'                      => '',

    'MEDIATOR_IP'                => '', # required for 2 node

    'CREATE_CLUSTER'             => 'TRUE',

    'CORE_DISK_SIZE'             => 122880,
    'ROOT_DISK_SIZE'             => 69632,
    'DATA_DISK_SIZE'             => 102400,
    'DATA_DISK_ALLOWED_CAPACITY' => 1048576,
    'DATA_DISKS'                 => 2,

    'EXTERNAL_NICS'              => 2,
    'INTERNAL_NICS'              => 4,

    'NODE_LOCATION'              => 'rtp',

    'CLUSTER_NAME'               => undef,

    'TOP_LOGDIR'                 => undef,
};

sub execute
{
    my ($self, %args) = @_;

    my $log   = log_global;

    my @nodes = @{$args{'resource'}};

    my %task_args = %{$self->task_args};

    if (scalar(@nodes) == 2 && ! $task_args{'MEDIATOR_IP'})
    {
        croak("A 2 node configuration requires a MEDIATOR_IP to be specified!");
    }
    ###

    my $n = 0;

    my %json_files;

    my $cluster_uuid = qx(/usr/software/bin/uuidgen);
    chomp($cluster_uuid);

    # Determine release to pick proper esxadmin script
    my $version_string = catfile($task_args{'BUILDROOT'}, 'version', 'cdot', 'version_string');
    my $vsfh = IO::File->new($version_string, 'r');
    my $version = <$vsfh>;
    chomp($version);
    $version =~ s/^.*_//; # git rid of <ReleaseName>__
    $vsfh->close;

    my($major, $minor, $patch) = split(/\./, $version);

    if (int($major) >= 9 &&
        int($minor) >= 2)
    {
        my $esxadmin_path = '/x/eng/rlse/DOT/devN/offtap/lib/deploy/src/esxadmin/src/esxadmin_cli';
        $log->log("Release is >= HN, setting ESXADMIN_PATH to $esxadmin_path");
        $task_args{'ESXADMIN_PATH'} = $esxadmin_path;
    }
    else
    {
        my $esxadmin_path = '/x/eng/rlse/vsa/devsDOTN/vsautil/src/esxadmin';
        $log->log("Release is <= LB.1, setting ESXADMIN_PATH to $esxadmin_path");
        $task_args{'ESXADMIN_PATH'} = $esxadmin_path;
    }

    if (int($major) >= 9 &&
        int($minor) >= 3)
    {
        $task_args{'EXTERNAL_NICS'} = 3;
    }

    my $nodecount = scalar(@nodes);

    #  create node is first node in the list
    my $create_node = $nodes[0];

    for my $node (@nodes)
    {
        my $hostrec_node = Hostrec->new('id' => $node);
        my $host = $hostrec_node->simhost;
        my $hostrec_host = Hostrec->new('id' => $host);
        my $hostrec_partner;

        $log->log("Host: $host");
        $log->log("Node: $node");

        if ($nodecount > 1)
        {
            if ($n % 2)
            {
                $log->log("Partner: " . $nodes[$n-1]);
                $hostrec_partner = Hostrec->new('id' => $nodes[$n-1]);
            }
            else
            {
                $log->log("Partner: " . $nodes[$n+1]);
                $hostrec_partner = Hostrec->new('id' => $nodes[$n+1]);
            }
        }

        $log->log("Starting template generation...");

        my $template = {};
        $template->{'GuestConfig'} = {};

        $template->{'Name'} = $node;

        # KVM needs NODE_UUID
        if ($task_args{'NODE_TYPE'} eq 'kvm')
        {
            my $node_uuid = qx(/usr/software/bin/uuidgen);
            chomp($node_uuid);
            $template->{'NODE_UUID'} = $node_uuid;
        }

        $template->{'CpuCount'} = int($task_args{'CPU_COUNT'});
        $template->{'MemsizeMB'} = int($task_args{'MEM_SIZE_MB'});
        $template->{'SerialPorts'} = [];
        for my $serial_port (@{$task_args{'SERIAL_PORT_NAMES'}})
        {
            if ($serial_port eq 'console')
            {
                push(@{$template->{'SerialPorts'}}, {'PortName' => $serial_port, 'PortNumber' => $hostrec_node->{'CONSOLE_PORT'}});
            }
            elsif ($serial_port eq 'gdb')
            {
                push(@{$template->{'SerialPorts'}}, {'PortName' => $serial_port});
            }
        }

        if ($task_args{'NODE_TYPE'} eq 'esx')
        {
            $template->{'ONTAPImage'} = $task_args{'BUILDROOT'}
                . '/final/bedrock/export/x86_64.debug.ovf.sdot/'
                . '/sdot/DataONTAPv-esx.ova';

            if ($nodecount >=4) # needs an internal mediator
            {
                $template->{'MediatorImage'} = $task_args{'MEDIATOR_ROOT'}
                    . '/final/mediator/DataONTAPv-mediator.vmdk';
            }

            # Single node cluster doesn't need a mediator
            # Two node cluster doesn't need an internal mediator, needs a remote one
        }
        else
        {
            $template->{'ONTAPImage'} = $task_args{'BUILDROOT'}
                . '/final/bedrock/export/x86_64.debug.ovf.sdot/'
                . '/sdot/DataONTAPv-kvm.raw.tar';

            if ($nodecount >=4) # needs an internal mediator
            {
                $template->{'MediatorImage'} = $task_args{'MEDIATOR_ROOT'}
                    . '/final/mediator/DataONTAPv-mediator.raw';
            }

            # Single node cluster doesn't need a mediator
            # Two node cluster doesn't need an internal mediator, needs a remote one
        }

        my $datastore;
        if ($hostrec_node->{'STORAGE_POOL'})
        {
            $datastore = $hostrec_node->{'STORAGE_POOL'};
        }
        else
        {
            if ($task_args{'NODE_TYPE'} eq 'esx')
            {
                $datastore = 'sDOT-' . $host;
            }
            elsif ($task_args{'NODE_TYPE'} eq 'kvm')
            {
                $datastore = 'ontap-select-storage-pool';
            }
        }

        $template->{'Storage'} = {};
        $template->{'Storage'}->{'BootDisk'} = {'Pool' => $datastore};

        if ($nodecount > 2)
        {
            $template->{'Storage'}->{'MediatorDisk'} = {
                'Pool'   => $datastore,
                'SizeMB' => 556,
            };
        }

        $template->{'Storage'}->{'PropertyISO'} = {'Pool' => $datastore};

        $template->{'Storage'}->{'CoreDisk'} = {'Pool' => $datastore, 'SizeMB' => int($task_args{'CORE_DISK_SIZE'}), 'EagerZero' => 0};

        if ($task_args{'RAID_TYPE'} eq 'hardware')
        {
            $template->{'Storage'}->{'RootDisks'} = [];
            push(@{$template->{'Storage'}->{'RootDisks'}}, {'Pool' => $datastore, 'SizeMB' => int($task_args{'ROOT_DISK_SIZE'}), 'EagerZero' => 0});
            # single node cluster only gets one
            if ($nodecount > 1)
            {
                push(@{$template->{'Storage'}->{'RootDisks'}}, {'Pool' => $datastore, 'SizeMB' => int($task_args{'ROOT_DISK_SIZE'}), 'EagerZero' => 0});
            }

            $template->{'Storage'}->{'DataDisks'} = [];
            # single node cluster only gets 1
            my $data_disks = $nodecount > 1 ? $task_args{'DATA_DISKS'} : 1;
            my $data_disk_size = $task_args{'DATA_DISK_SIZE'};

            my $requested_capacity = $data_disks * $data_disk_size;

            my $allowed_capacity = $task_args{'DATA_DISK_ALLOWED_CAPACITY'};

            if ($requested_capacity > $allowed_capacity)
            {
                croak("Requested disk capacity: $requested_capacity MB, exceeds allowed $allowed_capacity MB capacity per VM");
            }
            else
            {
                $log->log("Requested disk capacity: $requested_capacity MB, within bounds of $allowed_capacity MB capacity per VM");
            }

            for (my $j = 0; $j < $data_disks; $j++)
            {
                push(@{$template->{'Storage'}->{'DataDisks'}}, {'Pool' => $datastore, 'SizeMB' => int($data_disk_size), 'EagerZero' => 0});
            }
        }
        elsif ($task_args{'RAID_TYPE'} eq 'software')
        {

            $template->{'GuestConfig'}->{'BootArgs'} = [
               "bootarg.vm.raidtype=RAID4",
               "bootarg.vm.passthru_enabled=true",
               "wafl-rootvol-size-in-gib=70"
            ];

            $template->{'Storage'}->{'PhysicalDisks'} = [];

            # Get physical disks via kvmadmin_cli
            my $host = $node;
            $host =~ s/[a-z]$//;

            my $conn = Connectrec->new(
                'id'       => $host,
                'host'     => $host,
                'template' => 'kvm_ssh',
                'connect'  => 1,
            );

            my $exp = get_host_sub_shell(
                'script'   => $task_args{'KVMADMIN_PATH'},
                'host'     => $host,
                'sub'      => 'storage',
                'password' => $conn->password,
            );

            my $proc = NATE::Process->new(
                'args' => [$exp, $host],
                'codespec' => sub {
                    my($exp, $host) = @_;
                    $exp->send("pdisk show_ssd\n");
                    $exp->expect(30, "KVMadmin host $host storage >");
                    my $output = $exp->before;
                    $output =~ s/pdisk show_ssd//;

                    my $disks = decode_json($output)->[0];

                    $log->log("Got some Physical Disks ... : " . Dumper($disks));
                    $log->log("Template ... : " . Dumper($template));

                    my @physical_disks;
                    for my $disk (@{$disks})
                    {
                        if (! $disk->{'used_by'})
                        {
                            push(@physical_disks, {'Disk' => $disk->{'name'}});
                        }
                        elsif ($disk->{'used_by'} eq '[unknown]')
                        {
							# Get device name
                            my $device = $conn->execute(
                                'command' => ['readlink', '-f', $disk->{'name'}],
                                'check_status' => 1,
                            );
                            chomp($device);

							$conn->execute(
								'command' => ['pvremove', '-ff', '-y', $device],
								'check_status' => 1,
							);

                            push(@physical_disks, {'Disk' => $disk->{'name'}});
                        }
                    }

                    keyval_global->hset('hash_name' => 'sdot', 'key' => 'PHYSICAL_DISKS', 'val' => \@physical_disks);

                    $exp->soft_close;
                },
            );
            $proc->start;
            $proc->wait;
            $proc->destroy;

            $template->{'Storage'}->{'PhysicalDisks'} = keyval_global->hget('hash_name' => 'sdot', 'key' => 'PHYSICAL_DISKS', 'request_timeout' => 5);

            # RAID4 requires 5 disks
            if (scalar(@{$template->{'Storage'}->{'PhysicalDisks'}}) < 5)
            {
                croak("Not enough PhysicalDisks for RAID4 (Need >= 5): " . Dumper($template->{'Storage'}->{'PhysicalDisks'}));
            }
        }

        $template->{'NetworkInterfaces'} = [];
        # single node cluster always gets 2
        my $external_nics = $nodecount > 1 ? $task_args{'EXTERNAL_NICS'} : 3;

        my %mgmt_vlan_arg = $hostrec_host->{'MANAGEMENT_VLANID'} ?
                                ('VLANid' => int($hostrec_host->{'MANAGEMENT_VLANID'})) :
                                ();

        my %external_vlan_arg = $hostrec_host->{'EXTERNAL_VLANID'} ?
                                ('VLANid' => int($hostrec_host->{'EXTERNAL_VLANID'})) :
                                ();

        my %internal_vlan_arg = $hostrec_host->{'INTERNAL_VLANID'} ?
                                ('VLANid' => int($hostrec_host->{'INTERNAL_VLANID'})) :
                                ();

        for (my $i = 1; $i <= $external_nics; $i++)
        {
            if ($task_args{'NODE_TYPE'} eq 'esx')
            {
                if ($i == 1)
                {
                    push(@{$template->{'NetworkInterfaces'}}, {'Network' => 'ONTAP-Management'});
                }
                else
                {
                    push(@{$template->{'NetworkInterfaces'}}, {'Network' => $task_args{'ESX_EXTERNAL_NETWORKNAME'}});
                }
            }
            elsif ($task_args{'NODE_TYPE'} eq 'kvm' &&
                   ! $hostrec_host->{'MANAGEMENT_MAC1'} &&
                   ! $hostrec_host->{'MANAGEMENT_MAC2'})
            {
                if ($i == 1)
                {
                    push(@{$template->{'NetworkInterfaces'}},
                        {
                            'Network' => lc($task_args{'KVM_EXTERNAL_NETWORKNAME'}),
                            %mgmt_vlan_arg,
                        });
                }
                else
                {
                    push(@{$template->{'NetworkInterfaces'}},
                        {
                            'Network' => lc($task_args{'KVM_EXTERNAL_NETWORKNAME'}),
                            %external_vlan_arg,
                        });
                }
            }
            elsif ($task_args{'NODE_TYPE'} eq 'kvm' &&
                   $hostrec_host->{'MANAGEMENT_MAC1'} &&
                   $hostrec_host->{'MANAGEMENT_MAC2'})
            {
                if ($i == 1)
                {
                    push(@{$template->{'NetworkInterfaces'}},
                        {
                            'Network' => 'ontap-external',
                            %mgmt_vlan_arg,
                        });
                }
                else
                {
                    push(@{$template->{'NetworkInterfaces'}},
                        {
                            'Network' => 'ontap-external',
                            %external_vlan_arg,
                        });
                }
            }

            last if ($i == 2);
        }

        # single node cluster does not need internal nics
        my $internal_nics = $nodecount > 1 ? $task_args{'INTERNAL_NICS'} : 0;
        for (my $i = 1; $i <= $internal_nics; $i++)
        {
            if ($i == $internal_nics)
            {
                if ($task_args{'NODE_TYPE'} eq 'esx')
                {
                    push(@{$template->{'NetworkInterfaces'}}, {
                        'Network' => $task_args{'ESX_INTERNAL_NETWORKNAME'},
                        'MACAddress' => $hostrec_node->{'LOCALMAC'},
                    });
                }
                elsif ($task_args{'NODE_TYPE'} eq 'kvm' &&
                       ! $hostrec_host->{'MANAGEMENT_MAC1'} &&
                       ! $hostrec_host->{'MANAGEMENT_MAC2'})
                {
                    push(@{$template->{'NetworkInterfaces'}}, {
                        'Network' => $task_args{'KVM_INTERNAL_NETWORKNAME'},
                        'MACAddress' => $hostrec_node->{'LOCALMAC'},
                        %internal_vlan_arg,
                    });
                }
                elsif ($task_args{'NODE_TYPE'} eq 'kvm' &&
                       $hostrec_host->{'MANAGEMENT_MAC1'} &&
                       $hostrec_host->{'MANAGEMENT_MAC2'})
                {
                    push(@{$template->{'NetworkInterfaces'}}, {
                        'Network' => 'ontap-internal',
                        'MACAddress' => $hostrec_node->{'LOCALMAC'},
                        %internal_vlan_arg,
                    });
                }
            }
            else
            {
                if ($task_args{'NODE_TYPE'} eq 'esx')
                {
                    push(@{$template->{'NetworkInterfaces'}}, {'Network' => $task_args{'ESX_INTERNAL_NETWORKNAME'}});

                }
                elsif ($task_args{'NODE_TYPE'} eq 'kvm')
                {
                    push(@{$template->{'NetworkInterfaces'}},
                        {
                            'Network' => $task_args{'KVM_INTERNAL_NETWORKNAME'},
                            %internal_vlan_arg,
                        });
                }
            }
        }

        # Only 2 can exist first, last set of them have to be last in NetworkInterfaces list
        my $external_nics_left = $external_nics - 2;
        $log->log("External NICs: $external_nics");
        $log->log("External NICs Left: $external_nics_left");

        # This is only supported in 9.3+
        while ($external_nics_left > 0)
        {
            if ($task_args{'NODE_TYPE'} eq 'esx')
            {
                push(@{$template->{'NetworkInterfaces'}}, {'Network' => lc($task_args{'ESX_EXTERNAL_NETWORKNAME'})});
            }
            elsif ($task_args{'NODE_TYPE'} eq 'kvm' && $node =~ m/c6220|rx2540/)
            {
                push(@{$template->{'NetworkInterfaces'}}, {'Network' => lc($task_args{'KVM_EXTERNAL_NETWORKNAME'}), 'VLANid' => 3089});
            }
            elsif ($task_args{'NODE_TYPE'} eq 'kvm' && ($node =~ m/c240-004[1-9]/ || $node =~ m/c240-005[0-9]/ || $node =~ m/c240-0060/))
            {
                push(@{$template->{'NetworkInterfaces'}}, {'Network' => lc($task_args{'ESX_EXTERNAL_NETWORKNAME'})});
            }
            elsif ($task_args{'NODE_TYPE'} eq 'kvm' && ($node =~ m/c240/))
            {
                push(@{$template->{'NetworkInterfaces'}}, {'Network' => lc($task_args{'ESX_EXTERNAL_NETWORKNAME'})});
            }

            $external_nics_left--;
        }

        $template->{'GuestConfig'}->{'CLUSTER_NAME'}  = $task_args{'CLUSTER_NAME'};
        $template->{'GuestConfig'}->{'CLUSTER_UUID'}  = $cluster_uuid;
        $template->{'GuestConfig'}->{'NODE_LOCATION'} = $task_args{'NODE_LOCATION'};

        # single node cluster node is the create node
        if ($node eq $create_node)
        {
            $template->{'GuestConfig'}->{'CLUSTER_CREATE'} = $task_args{'CREATE_CLUSTER'};
        }
        else
        {
            $template->{'GuestConfig'}->{'CLUSTER_CREATE'} = 'FALSE';
        }

        $template->{'GuestConfig'}->{'NODE_COUNT'} = $nodecount;
        $template->{'GuestConfig'}->{'ADMIN_PASSWORD'} = $hostrec_node->default_password;

        $log->log("Getting cluster management interface");
        my @cluster_management_v4s = $hostrec_node->net_filter('include' => {'family' => 'inet4', 'nettype' => 'cluster_management'});
        my $cluster_management_v4 = $cluster_management_v4s[0];
        $template->{'GuestConfig'}->{'CLUSTER_MGMT_IP'} = $cluster_management_v4->{'address'};
        $template->{'GuestConfig'}->{'CLUSTER_MGMT_NETMASK'} = $cluster_management_v4->{'netmask'};
        $template->{'GuestConfig'}->{'CLUSTER_MGMT_GATEWAY'} = $cluster_management_v4->{'gateway'};

        $log->log("Getting management interface");
        my @management_v4s = $hostrec_node->net_filter('include' => {'family' => 'inet4', 'nettype' => 'management', 'mtu' => 1500});
        my $management_v4 = $management_v4s[0];
        $template->{'GuestConfig'}->{'NODE_MGMT_IP'} = $management_v4->{'address'};
        $template->{'GuestConfig'}->{'NODE_MGMT_NETMASK'} = $management_v4->{'netmask'};
        $template->{'GuestConfig'}->{'NODE_MGMT_GATEWAY'} = $management_v4->{'gateway'};

        # External Mediator
        if ($nodecount == 2)
        {
            my $tgt = keyval_global->hget('hash_name' => 'sdot', 'key' => 'ISCSI_XMEDIATOR_TGT', 'request_timeout' => 5) // croak('Must provide an ISCS_XMEDIATOR_TGT!');
            $log->log("Getting RLM interface to use as mediator communication interface");
            my $rlm_ip = qx(host $node-rlm | sed 's/^.*address //');
            chomp($rlm_ip);
            $template->{'GuestConfig'}->{'ISCSI_XMEDIATOR_IP'} = $task_args{'MEDIATOR_IP'};
            $template->{'GuestConfig'}->{'ISCSI_XMEDIATOR_TGT'} = $tgt;
        }

        # single node cluster doesn't have a partner
        if ($nodecount > 1)
        {
            $template->{'GuestConfig'}->{'HA_PARTNER_IC_MAC'} = $hostrec_partner->{'LOCALMAC'};
            $template->{'GuestConfig'}->{'HA_PARTNER'} = $hostrec_partner->hostname;
        }

        $template->{'GuestConfig'}->{'SERIAL_NUMBER'} = '32000000' . $n + 32;
        $template->{'GuestConfig'}->{'SERIAL_NUMBER'} .= ""; # force string, lolersauce....
        $template->{'GuestConfig'}->{'BUILTIN_LICENSES'} = 'YJSJBKLHSOVOGDAAAAAAAAAAAAAA,SCFVEKLHSOVOGDAAAAAAAAAAAAAA,KEQYBKLHSOVOGDAAAAAAAAAAAAAA,MPUUAKLHSOVOGDAAAAAAAAAAAAAA,UNJRDKLHSOVOGDAAAAAAAAAAAAAA,EXCKFKLHSOVOGDAAAAAAAAAAAAAA';
        $log->log("Done.");

        my $file = File::Spec->catfile($task_args{'TOP_LOGDIR'}, "$node.json");
        my $json = JSON::XS->new->pretty(1)->encode($template);
        my $fh = IO::File->new($file, 'w');
        my $failure = $!;
        if ($fh)
        {
            print({$fh} $json);
            $log->log($json);
            $log->log("Wrote file to: " . abs_path($file));
            $json_files{$node} = abs_path($file);
            $log->log("Stored in $node of json_files hash");
        }
        else
        {
            if ($failure)
            {
                croak("Failed to write json file with error: $failure");
            }
            else
            {
                croak("Failed to write json file for an unknown reason.");
            }
        }

        $n++; # for partner info accessing @nodes +- n
    }

    # Store json files
    $task_args{'JSON_FILES'} = \%json_files;
}

1;
