#   Copyright (c) 2017 NetApp, Inc., All Rights Reserved.
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.

package NACL::Init::Tasks::ONTAPSelect::Server::Power;

use strict;
use warnings;
use Carp       qw(croak);
use File::Temp qw(tempfile);
use Net::Ping;

use Params::Validate qw(validate :types);

use Moo;
extends qw(NATE::Init::Task);
# need to do this to prevent Moo from clashing with NATE::BaseException
BEGIN {undef *with;}

use Hostrec;
use NATE::BaseException qw(:try);
use NATE::Events        qw(eventloop);
use NATE::Log           qw(log_global);

use NACL::Init::Utils::VMware qw(get_esx_certificate);

use constant is_retryable => 1;

use constant 'defaults' => {
    'POWER_ACTION' => '',
    'HOST_OS'      => '',
};

sub execute
{
    my ($self, %args) = @_;

    my $resource      = $args{'resource'};
    my $log       = log_global;
    my %task_args = %{$self->task_args};

    if (!$task_args{'POWER_ACTION'} ||
        ($task_args{'POWER_ACTION'} && $task_args{'POWER_ACTION'} !~ m/^on|off|reboot|$/))
    {
        croak("You must specify a POWER_ACTION of either on, off, or reboot.");
    }

    my $hostrec_resource = Hostrec->new('id' => $resource);
    my $hostrec_host;
    if ($hostrec_resource->simhost)
    {
         $hostrec_host = Hostrec->new('id' => $hostrec_resource->simhost);
    }
    else
    {
        $hostrec_host = $hostrec_resource;
    }

    # Check if it is ipmi
    if ($hostrec_host->powercmd =~ m/ipmitool/)
    {
        $log->info('Installing ipmitool...');
        my $proc = NATE::Process->new(
            'codespec' => '/usr/bin/yum',
            'runid'    => 'install_ipmitool',
            'onexit'   => \&NATE::Process::on_exit_die_on_non_zero_exit_status,
            'username' => 'root',
            'args'    => [
                'install',
                '-y',
                'ipmitool',
            ],
        );
        $proc->start;
        $proc->wait;
        $proc->destroy;
        $log->info('Done.');
    }

    if ($task_args{'POWER_ACTION'} eq 'on')
    {
        $log->info('Powering on ' . $hostrec_host->id);

        $hostrec_host->powercycle('on');
        _wait_for_on($hostrec_host, $task_args{'HOST_OS'});
        $log->info($hostrec_host->id . ' has been powered on.')
    }
    elsif ($task_args{'POWER_ACTION'} eq 'off')
    {
        $log->info('Powering off ' . $hostrec_host->id);
        $hostrec_host->powercycle('off');
        _wait_for_off($hostrec_host);
        $log->info($hostrec_host->id . ' has been powered off.')
    }
    elsif ($task_args{'POWER_ACTION'} eq 'reboot')
    {
        # Do not use reboot diretly, we want to make sure
        #   it really did power off before powering on.
        $log->info('Powering off ' . $hostrec_host->id);
        $DB::Single=2;
        $hostrec_host->powercycle('off');
        _wait_for_off($hostrec_host);

        $log->info('Powering on ' . $hostrec_host->id);
        $hostrec_host->powercycle('on');
        _wait_for_on($hostrec_host, $task_args{'HOST_OS'});

        $log->info($hostrec_host->id . ' has been rebooted.');
    }

}

sub _wait_for_off
{
    my $hostrec = shift;
    my $log = log_global;

    while (1)
    {
        my $exit_status = system('ping', '-c', 1, '-t', 60, $hostrec->default_ip);
        if ($exit_status == 0)
        {
            $log->info("Still responsive, waiting for it to "
                . 'be unresponsive...');
            eventloop('seconds' => 10);
        }
        else
        {
            $log->info("Not responsive, assuming it is powered "
                . 'completely off.');
            return 1;
        }
    }
}

sub _wait_for_on
{
    my $hostrec = shift;
    my $os   = shift;

    my $log = log_global;

    my $conn_tpl = ($os =~ m/esx/i ) ? 'esx_ssh' : 'kvm_ssh';

    while (1)
    {
        my $exit_status = system('ping', '-c', 1, '-t', 60, $hostrec->default_ip);
        if ($exit_status != 0)
        {
            $log->info("Still unresponsive, waiting for it to "
                . 'be responsive...');
            eventloop('seconds' => 5);
        }
        else
        {
            $log->info("Reponsive, assuming it is powered on.");
            last;
        }
    }


    $log->info("Waiting until we can login to the host...");
    while (1)
    {
        eval {
            Connectrec->new(
                'host'     => $hostrec,
                'template' => $conn_tpl,
                'connect'  => 1,
            );
        };
        if (my $err = $@)
        {
            $log->info("Unable to login still...");
            eventloop('seconds' => 30);
        }
        else
        {
            $log->info("Successfully logged in.");
            last;
        }
    }

    # Wait until HTTP GET (via get_esx_certificate) is successful
    if ($os =~ m/esx/i)
    {
        $log->info("Waiting until we can get_esx_certificate...");
        my $conn = Connectrec->new(
            'host'     => $hostrec,
            'template' => $conn_tpl,
            'connect'  => 0,
        );

        while (1)
        {
            eval {
                get_esx_certificate($hostrec->id, $conn->username, $conn->password);
            };
            if ($@)
            {
                $log->info("Unable to get ESX certificate still: $@");
                eventloop('seconds' => 30);
            }
            else
            {
                $log->info("Success!  All done....");
                last;
            }
        }
    }

    return 1;
}

1;
