#   Copyright (c) 2016 NetApp, Inc., All Rights Reserved.
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.

package NACL::Init::Tasks::Vsim::Make;

use strict;
use warnings;
use NATE::Log qw(log_global);
use Data::Dumper;
use NACL::Init::Tasks::VsimUtils qw(get_vsim_details_from_node_name prepare_client_apiset);
use NACL::C::Node;
use Carp qw(confess);
use NATE::ParamSet;
use Options;
use NATE::BaseException qw(try);
use Moo;
extends qw(NATE::Init::Task);

use constant defaults => {
    VSIM_SCRIPT         => "/usr/software/test/bin/vsim",
    BUILDROOT           => undef,
    VARIANT             => "x86_64.sim",
    IMAGE               => undef,
    ONTAP_BUILDROOT     => undef,
    ONTAP_VARIANT       => undef,
    ONTAP_INSTALL_IMAGE => undef,
    MEM                 => undef,
    NICS                => "5",
    DISK_PREZERO        => 0,
    VSPC_SCRIPT         => "/usr/software/test/bin/vSPC.py",
    VSTORE              => "vscsi:8g:3:0,vscsi:1500m:11:0,vscsi:1500m:14:1",
    CPUS                => undef,
    PNVR                => undef,
};

sub execute {

    my ($self, %args) = @_;

    my @nodes = $args{resource};

    my ($task, %task_args);
    $task = $self->task_args();
    %task_args = %$task;

    my $vsim_tool           = $task->{"VSIM_SCRIPT"};
    my $mem                 = $task->{"MEM"};
    my $ontap_buildroot     = $task->{"ONTAP_BUILDROOT"} // $task->{"BUILDROOT"};
    my $ontap_variant       = $task->{"ONTAP_VARIANT"} // $task->{"VARIANT"};
    my $vspc_script         = $task->{"VSPC_SCRIPT"};
    my $vstore              = $task->{"VSTORE"};
    my $nics                = $task->{"NICS"};
    my $cpus                = $task->{"CPUS"};
    my $ontap_install_image = $task->{"ONTAP_INSTALL_IMAGE"} // $task->{"IMAGE"};
    my $pnvr_option         = $task->{"PNVR"};

    my $command_string;

    if($task->{DISK_PREZERO} == 1) {
        $command_string .= " -dzero";
    }
    if (defined($cpus)) {
        $command_string .= " -cpus $cpus";
    }
    if (defined($mem)) {
        $command_string .= " -m $mem";
    }
    if (defined($pnvr_option)) {
        $command_string .= " -pnvr";
    }

    if(!defined $ontap_buildroot || !defined $ontap_variant) {
        if(defined $ontap_install_image){
            # parse through image to get the buildroot and variant
            if ($ontap_install_image =~ m#^(.+?)/final/bedrock/export/(.+?)/#) {
                # assuming the path for non-promotional ontap image to be of the format:
                # /x/eng/bbrtp/daemon/DOT/R9.3x/daemon_base/.snapshot/4498145/final/bedrock/export/x86_64.debug.sim/tarball/image.tgz
                # ontap_buildroot: /x/eng/bbrtp/daemon/DOT/R9.3x/daemon_base/.snapshot/4498145
                $ontap_buildroot = $1;

                # ontap_variant: x86_64.debug.sim
                $ontap_variant = $2;

            } elsif ($ontap_install_image =~ m#^(.+?)/promo/.+?/(.+?)/#) {
                # assuming the path for promotional image to be of the format:
                # /x/eng/rlse49/DOT/released/9.3X6/promo/9.3X6/x86_64.sim/image.tgz
                # ontap_buildroot: /x/eng/rlse49/DOT/released/9.3X6
                $ontap_buildroot = $1;

                # ontap_variant: x86_64.sim
                $ontap_variant = $2;
            }
        } else {
            confess "One or more parameter is not passed in: ONTAP_VARIANT,ONTAP_BUILDROOT,ONTAP_INSTALL_IMAGE";
        }
    }

    log_global()->comment("Make all vsims");

    foreach my $node (@nodes) {
        # identify the vsim user to operate on for this
        my $vsim_details = get_vsim_details_from_node_name($node);
        my $vsim_user = $vsim_details->{user};
        my $vsim_dir = "/u/".$vsim_details->{user}."/vsims";
        my $sudo_user = "/usr/software/bin/sudo -i -u ".$vsim_details->{user};

        log_global()->comment("Preparing client APISet");
        my $c_api = prepare_client_apiset();

        my $status = 0;

        my $command = "$sudo_user $vsim_tool make -ir $ontap_buildroot -px $ontap_variant -bubble_net -pcons ".
            "-vstore $vstore -nics $nics -simdir $vsim_dir -smoke $vsim_details->{id}";
        if ($command_string) {
            $command .= $command_string;
        }

        $c_api->execute_raw_command(command    => $command,
                                    get_status => \$status);
        if ($status ne 0) {
            NATE::BaseException->throw("Encountered a bad status while making vsim $node");
        }
    }
}

1;
