#   Copyright (c) 2016 NetApp, Inc., All Rights Reserved.
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.

package NACL::Init::Tasks::Vsim::Setenv;

use strict;
use warnings;
use NATE::Log qw(log_global);
use NACL::Init::Tasks::VsimUtils qw(get_vsim_details_from_node_name 
                                    prepare_client_apiset);
use Carp qw(confess);
use NACL::Transit;
use Connectrec;
use NACL::C::Node;
use Tie::IxHash;

use Moo;

extends qw(NATE::Init::Task);

# need to do this to prevent Moo from clashing with NATE::BaseException
BEGIN {undef *with;}

use NATE::BaseException qw(:try);

use constant defaults => {
    DEFAULT_SETENV  => {
        "AUTOBOOT"                  => "false",
        "bootarg.vm.rapidsavecore"  => "false",
        "qos.classify.panic_how"    => 0,
    },
    CONFIG_SETENV    => {},
    SETENV           => undef,
    VSIM_SETENV      => undef,
    UNSETENV         => [],
    VSIM_SCRIPT      => "/usr/software/test/bin/vsim",
};

sub execute {

    my ($self, %args) = @_;

    my @nodes = $args{resource};

    my ($task, %task_args, %default_setenv, %config_setenv, %setenv);

    # This code is used to split the SETENV/VSIM_SETENV passed on command line and make a hash.
    # Now merge the hash with the default values, the command line values taking
    # precedence over the default values. (The default values hash has already been
    # merged with the config_setenv hash, the config values taking precedence over
    # default values).

    $task = $self->task_args;
    %task_args = %$task;
    tie %default_setenv, 'Tie::IxHash';
    tie %config_setenv, 'Tie::IxHash';
    tie %setenv, 'Tie::IxHash';
    my $defaultsetenv = $task->{DEFAULT_SETENV};
    my $configsetenv = $task->{CONFIG_SETENV};
    # setenv from command line will look like this: SETENV/VSIM_SETENV="p1=>v1,p2=>v2". parse this string and make a hash and them merge with defaults hash
    my $task_setenv = "";
    if($task->{"SETENV"}){
        $task_setenv = $task->{"SETENV"};
    }
    if($task->{"VSIM_SETENV"}){
        $task_setenv .= "," if($task_setenv); # append ',' if task_setenv already has values assigned
        $task_setenv .= $task->{"VSIM_SETENV"};
    }
    my @setenv_array = split /,/,$task_setenv;
    foreach my $setenv_element (@setenv_array) {
        my @temp = split /=>/,$setenv_element;
        $setenv{$temp[0]} = $temp[1];
    }
    %default_setenv = %$defaultsetenv;
    %config_setenv = %$configsetenv;
    # merge default and config, replace duplicates with values of config
    @default_setenv{ keys %config_setenv } = values %config_setenv;
    # now merge the above result with setnev, giving setenv values priority
    foreach my $key2 ( keys %setenv ) {
        if( exists $default_setenv{$key2} ) {
            $default_setenv{$key2} = $setenv{$key2};
            next;
        } else {
            $default_setenv{$key2} = $setenv{$key2};
        }
    }
    my $Vsim_tool = $task->{VSIM_SCRIPT};

    log_global()->comment("setenv on vsims");
    foreach my $node (@nodes) {
        my $vsim_details = get_vsim_details_from_node_name($node);
        my $vsim_user = $vsim_details->{user};
        my $vsim_dir = "/u/".$vsim_user."/vsims/";
        my $sudo_user = "/usr/software/bin/sudo -i -u ".$vsim_details->{user};

        log_global()->comment("Preparing client APISet");
        my $c_api = prepare_client_apiset();

        foreach my $key (keys %default_setenv) {
            my $status = 0;
            $c_api->execute_raw_command(command => "$sudo_user $Vsim_tool setenv -smoke $vsim_details->{id} -simdir $vsim_dir $key $default_setenv{$key}",
                                          get_status => \$status);
            if ($status ne 0) {
                NATE::BaseException->throw("Encountered a bad status while setting env variable ");
            }
        }
    }
}

1;
