#   Copyright (c) 2017 NetApp, Inc., All Rights Reserved.
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.

# REST::Client package available in CPAN does not support authentication
package NACL::Init::Utils::ONTAPSelect::Deploy::APIClient;

###
# NOTE: This is a temporary module used in NACL::Init ONTAPSelectDeploy until
#        APISet for ONTAPSelectDeploy is available.
###
use strict;
use warnings;
use Carp         qw(croak);
use Data::Dumper qw(Dumper);
use MIME::Base64 qw(encode_base64);
use Scalar::Util qw(blessed);

use base qw(LWP::UserAgent);

use HTTP::Request;
use HTTP::Response;
use HTTP::Status;
use JSON::XS         qw(encode_json decode_json);
use Params::Validate qw(validate validate_with :types);

use NATE::Log qw(log_global);

$ENV{'PERL_LWP_SSL_VERIFY_HOSTNAME'} = 0;

sub new
{
    my $class = shift;
    my %opts = validate_with(
    	'params'      => \@_,
    	'allow_extra' => 1,
    	'spec'        => {
    		'baseuri'         => {'type' => SCALAR},
    		'username'        => {'type' => SCALAR,  'optional' => 1},
    		'password'        => {'type' => SCALAR,  'optional' => 1},
    		'default_headers' => {'type' => ARRAYREF, 'default'  => [
					{'Content-type' => 'application/json'},
				],
			},
		},
	);

    if (($opts{'username'} && !$opts{'password'}) ||
    	($opts{'password'} && !$opts{'username'}))
    {
    	croak("If you supply a username or password, you must supply both.");
    }

    my $self = $class->SUPER::new(@_);

    $self->{'_na'} = \%opts;

    return $self;
}

sub _shared_request
{
	my $self = shift;
	my %opts = validate(@_, {
		'method' => {'type' => SCALAR},
		'path'   => {'type' => SCALAR},
		'data'   => {'type' => HASHREF, 'optional' => 1},
	});

	my $request = HTTP::Request->new(
		$opts{'method'} => join('/', ($self->{'_na'}->{'baseuri'}, $opts{'path'})),
	);

	if ($self->{'_na'}->{'username'} && $self->{'_na'}->{'username'})
	{
		$request->header('Authorization', 'Basic ' . encode_base64($self->{'_na'}->{'username'} . ':' . $self->{'_na'}->{'password'}));
	}

	if ($self->{'_na'}->{'default_headers'})
    {
    	for my $header (@{$self->{'_na'}->{'default_headers'}})
    	{
    		$request->header(%{$header});
    	}
    }

	if ($opts{'data'})
	{
		$request->content(encode_json($opts{'data'}));
	}

	print Dumper $request;

	my $response = $self->request($request);

	if ($response->is_success)
	{
		my $data;
		eval {
			$data = decode_json($response->decoded_content);
		};
		if ($@)
		{
			$data = $response->decoded_content;
		}

		return $data;
	}
	else
	{
		croak($response->status_line . " : " . $response->decoded_content);
	}
}

sub GET
{
	my $self = shift;
	my %opts = validate(@_, {
		'path' => {'type' => SCALAR},
		'data' => {'type' => HASHREF, 'optional' => 1},
	});
	return $self->_shared_request('method' => 'GET', %opts);
}

sub PUT
{
	my $self = shift;
	my %opts = validate(@_, {
		'path' => {'type' => SCALAR},
		'data' => {'type' => HASHREF, 'optional' => 1},
	});
	return $self->_shared_request('method' => 'PUT', %opts);
}

sub POST
{
	my $self = shift;
	my %opts = validate(@_, {
		'path' => {'type' => SCALAR},
		'data' => {'type' => HASHREF, 'optional' => 1},
	});
	return $self->_shared_request('method' => 'POST', %opts);
}

sub DELETE
{
	my $self = shift;
	my %opts = validate(@_, {
		'path' => {'type' => SCALAR},
		'data' => {'type' => HASHREF, 'optional' => 1},
	});
	return $self->_shared_request('method' => 'DELETE', %opts);
}

sub PATCH
{
	my $self = shift;
	my %opts = validate(@_, {
		'path' => {'type' => SCALAR},
		'data' => {'type' => HASHREF, 'optional' => 1},
	});
	return $self->_shared_request('method' => 'PATCH', %opts);
}

1;