#   Copyright (c) 2017 NetApp, Inc., All Rights Reserved.
#   Any use, modification, or distribution is prohibited
#    without prior written consent from NetApp, Inc.

package NACL::Init::Utils::VMware;

use strict;
use warnings;
use Carp qw(croak);
use Digest::SHA  qw(sha1_hex);
use MIME::Base64 qw(decode_base64);

use Params::Validate qw(validate :types);

use VMware::VICommon;
use VMware::VIMRuntime;
use VMware::VIM25Stub;
use VMware::VILib;
use VMware::VIExt;

use NATE::Events qw(eventloop);
use NATE::Log qw(log_global);

BEGIN {
    require Exporter;
    our @ISA = qw(Exporter);
    our @EXPORT_OK = qw(
        wait_for_task
        get_esx_certificate
        generate_ssl_thumbprint
        rot13
    );
}

sub wait_for_task
{
    my %opts = validate(@_, {
        'task'    => {'type' => OBJECT, 'isa' => 'ManagedObjectReference'},
        'message' => {'type' => SCALAR},
    });

    my $task_view = Vim::get_view('mo_ref' => $opts{'task'});

    while (1)
    {
        my $info = $task_view->info;

        if ($info->state->val eq 'success')
        {
            log_global->info($opts{'message'});
            return 1;
        }
        elsif ($info->state->val eq 'error')
        {
            croak(join("\n", ($info->error->fault, $info->error->localizedMessage)));
        }

        eventloop('seconds' => 5);

        $task_view->ViewBase::update_view_data();
    }
}

sub get_esx_certificate
{
    my ($host, $user, $pass) = @_;

    my $ua = LWP::UserAgent->new;
    my $realm_name = "VMware HTTP server";
    $ua->credentials("$host:443", $realm_name, $user => $pass);
    my $response = $ua->get("https://$host:443/host/ssl_cert");

    croak $response->status_line unless $response->is_success;

    return $response->decoded_content;
}

sub generate_ssl_thumbprint
{
    my %args = @_;

    my ($pem, $der, $digest, $sslthumbprint);
    $pem = delete($args{'pem'});

    # Strip PEM tags to get Base64 encoded certificate data
    $pem =~ s/-{1,}(BEGIN|END) CERTIFICATE-{1,}//g;

    # Convert PEM to DER (decode Base64)
    $der = decode_base64($pem);

    # Generate SHA1 hex digest
    $digest = sha1_hex($der);

    # Format thumbprint
    $sslthumbprint = "";
    for (my $i=0; $i < length($digest); $i+=2)
    {
        my $substring = substr($digest, $i, 2);
        $sslthumbprint .= uc($substring);
        unless ($i >= 38)
        {
            $sslthumbprint .= ":";
        }
    }

    return $sslthumbprint;
}

sub rot13
{
    my %opts = validate(@_, {
        'string' => {'type' => SCALAR},
    });

    $opts{'string'} =~ y/A-Za-z/n-za-mN-ZA-M/;

    return $opts{'string'};
}

1;
