# Copyright (c) 2016 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Composite Aggregate Object Store Task Module
## @author ng-compaggr-smg-qa@netapp.com
## @status shared
## @pod here

package NACL::MTask::AggregateObjectStore;

use strict;
use warnings;

use base qw(NACL::STask::Aggregate);
use NATE::Log qw(log_global);
my $Log       = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Params::Validate qw(validate :types);
use NACL::STask::Aggregate;
use NACL::CS::Aggregate;
use NATE::BaseException qw(:try);
use NACL::STask::Volume;
use NACL::STask::Vserver;
use NACL::ComponentUtils qw(Dumper);
use Class::MethodMaker [
    scalar => 'vserver',
    scalar => 'percent_to_fill',
    scalar => 'aggregate',
];

use constant DEFAULT_VOLUME_TIERING_POLICY => 'snapshot-only';

sub VERIFY {
    my ($flag, $msg) = @_;
    if (!$flag) {
        NATE::BaseException->throw($msg);
    }
} ## end sub VERIFY

sub config_and_attach {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;

    #$Log->comment("Args:". Dumper(\@args));
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            node         => {type => OBJECT, isa      => 'NACL::C::Node'},
            'size-limit' => {type => SCALAR},
            aggregate    => {type => SCALAR, optional => 1},
            'object-store-name' => {type => SCALAR, optional => 1},
            'provider-type'   => {type => SCALAR, reqex => /AWS_S3|SGWS|S3EMU/},
            'ip-address'      => {type => SCALAR, optional => 1},
            's3-name'         => {type => SCALAR, optional => 1},
            'server'          => {type => SCALAR, optional => 1},
            'container-name'  => {type => SCALAR, optional => 1},
            'access-key'      => {type => SCALAR, optional => 1},
            'secret-password' => {type => SCALAR, optional => 1},
            port    => {type => SCALAR, optional => 1, default => 80},
            encrypt => {type => SCALAR, optional => 1},
            'encryption-key-id'   => {type => SCALAR, optional => 1},
            'object-checker-node' => {type => SCALAR, optional => 1},
            'ssl-enabled'         => {type => SCALAR, optional => 1},
            ipspace               => {type => SCALAR, optional => 1},
        },
        allow_extra         => 1,
        ignore_primary_keys => 1,
    );

	my $ip_address = delete $opts{'ip-address'};
	my $server = $opts{'server'};
	my $s3_name = delete $opts{'s3-name'};
	my $container_name = $opts{'container-name'};

	if($ip_address ne ""){
		$Log->warn("This field 'ip-address' is now changed to 'server', support for ip-address is only available on debug builds for short time, update your script");
	}
	if($server eq "" and $ip_address ne ""){
		$opts{server} = $ip_address;
		$Log->warn("This field 'server' is now changed to accept only FQDN (like aws.netapp.com), support for ip-address=172.xx.yy.zz format is only available on debug builds for short time, update your script");
	}
	if($server eq "" and $ip_address eq ""){
		NATE::BaseException->throw("The fields -server and/or -ip-address are not defined, you must provide -server=<SGWS/AWS-S3 FDQN>");
	}
	if($s3_name ne ""){
		$Log->warn("This field 's3-name' is now changed to 'container-name', support for container-name is only available on debug builds for short time, update your script");
	}
	if($container_name eq "" and $s3_name ne ""){
		$opts{'container-name'} = $s3_name;
		$Log->warn("This field 's3-name' is now changed to 'container-name', support for this auto-conversion is only available on debug builds for short time, update your script");
	}
	if(!$opts{'container-name'} or !$opts{server}){
		NATE::BaseException->throw("The fields -container-name and/or -s3-name are not defined, you must provide -conainter-name=<bucket-name>");
	}
    #$Log->comment("Opts:". Dumper(\%opts));
    $Log->comment("Aggregate: " . $opts{aggregate});
    my ($percent, $command_interface, $aggr_obj, $vs_obj, $volume_size,
        $vol_obj, $self);
    my $aggr = delete $opts{aggregate};
    my $node = $opts{node};
    my ($e_aggr, %topts);
    if ($aggr) {
        $Log->comment("Aggr: " . $aggr);
        $topts{aggregate} = $aggr;
    }

    # find an aggregate that is already composite
    $e_aggr = NACL::STask::Aggregate->find(
        command_interface => $node,
        filter            => {
            composite => 'true',
            %topts
        },
    );

    # find an existing data aggregate that can be converted to composite
    if (!$e_aggr) {
        $e_aggr = NACL::STask::Aggregate->find(
            command_interface => $node,
            filter            => {
                hybrid           => 'false',
                'hybrid-enabled' => 'false',
                %topts
            },
        );
    }

    # create an aggregate as last resort
    if (!$e_aggr) {
        $e_aggr = NACL::STask::Aggregate->create(
            command_interface => $node,
            %topts,
        );
    }

    delete $opts{node};
    delete $opts{client};
    require NACL::STask::StorageAggregateObjectStore;
    my %c = NACL::STask::StorageAggregateObjectStore->config_show(%opts);
    if (!$c{'object-store-name'} eq $opts{'object-store-name'}) {
        NACL::STask::StorageAggregateObjectStore->config_create(%opts);
        NACL::STask::StorageAggregateObjectStore->config_show(%opts);
    }
    $opts{aggregate}         = $e_aggr->aggregate;
    $opts{command_interface} = $node;
    my %attach_opts = (
        aggregate           => $e_aggr->aggregate,
        command_interface   => $node,
        'object-store-name' => $opts{'object-store-name'},
        'size-limit'        => $opts{'size-limit'},
    );
    NACL::STask::StorageAggregateObjectStore->attach(%attach_opts);

    my %args;
    $pkg->_hash_copy(
        source => \%opts,
        copy   => [qw(aggregate command_interface)],
        target => \%args,
    );

    $self = $pkg->new(%args);
    $Log->exit() if $may_exit;
    return $self;
} ## end sub config_and_attach

sub volume_create {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            command_interface => {type => OBJECT, isa => 'NACL::C::Node'},
            size      => {type => SCALAR, optional => 1, default => '20m'},
            aggregate => {type => SCALAR, optional => 1},
            volume    => {type => SCALAR, optional => 1},
            type      => {type => SCALAR, optional => 1},
            vserver => {type => SCALAR, optional => 1, default => 'vs0'},
            'tiering-policy' =>
                {type => SCALAR, optional => 1, default => 'snapshot-only'},
            'space-guarantee' =>
                {type => SCALAR, optional => 1, default => 'none'},
            'unix-permissions' =>
                {type => SCALAR, optional => 1, default => '---rwxrwxrwx'},
            'security-style' =>
                {type => SCALAR, optional => 1, default => 'mixed'},
        },
        allow_extra         => 1,
        ignore_primary_keys => 1,
    );
    if ($opts{type} eq 'DP') {
        delete $opts{'unix-permissions'};
    }
    my ($vserver, $command_interface, $percent, $vs_obj, $volume_size);
    $command_interface = $opts{command_interface};
    unless ($opts{vserver}) {
        my $vserver = NACL::STask::Vserver->find_data_vservers(
            command_interface => $opts{command_interface},
            filter            => {subtype => 'default'},
        );
        $opts{vserver} = $vserver->vserver;
    }
    my %taggr_opts = (
        aggregate => $opts{aggregate},
        composite => 'true'
    );
    my $aggr_obj = NACL::CS::Aggregate->fetch(
        command_interface => $command_interface,
        filter            => {%taggr_opts},
    );
    $opts{aggregate} = $aggr_obj->aggregate;
    require Data::Dumper;
    $Data::Dumper::MaxDepth = 1;
    $Log->comment("Vol opts: " . Dumper(\%opts));
    my $vol_obj = NACL::STask::Volume->create(%opts,);
    my $vol_tiering_policy
        = $vol_obj->get_one_state_attribute('tiering-policy');
    VERIFY(
        (          $vol_tiering_policy eq $opts{'tiering-policy'}
                or $vol_tiering_policy eq DEFAULT_VOLUME_TIERING_POLICY
        ),
        'volume tiering policy is not set to default'
    );
    $Log->exit() if $may_exit;
    return $vol_obj;
} ## end sub volume_create

sub new {
    my ($pkg, %opts) = @_;
    my $self = bless {}, $pkg;
    while (my ($key, $value) = each %opts) {
        $self->$key($value);
    }
    return $self;
} ## end sub new

1;
