
# Copyright (c) 2011 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Collect Autosupport messages as they occur
## @author dl-nacl-dev@netapp.com
## @status shared
## @pod here

package NACL::MTask::AutosupportReceiver;
use base qw(NACL::MTask::MTask);

use strict;
use warnings;

use Params::Validate qw(validate validate_with SCALAR BOOLEAN);

use Tharn ( '$Log', 'subtest' );
use NATE::BaseException                       (':try');
use NACL::C::SystemNodeAutosupport            ();
use NACL::Tools::httpd::control               ();
use NACL::MTask::AutosupportReceiver::Message ();

=head1 NAME

NACL::MTask::AutosupportReceiver

=head1 SYNOPSIS

    # Set up a temporary server to receive autosupports:

        my $ar = NACL::MTask::AutosupportReceiver->start(
            command_interface => $node,
            redirect => 1,
        );

    # Trigger a user-initiated autosupport message.  (As
    # SystemNodeAutosupport->invoke, but fills in command_interface
    # and url)

        $ar->invoke(node=>"local");

    # Return a list of autosupport messages which have been received

        my @messages = $ar->messages();

    # Return the next new message which has been received (once
    # new_message or wait_for_message returns a message, it becomes no
    # longer new)

        my $message  = $ar->new_message();

    # As C<new_message>, but first wait for there to be at least one
    # message.

        my $message  = $ar->wait_for_message("method-timeout"=>$timeout);

    # The messages themselves are
    # NACL::MTask::AutosupportReceiver::Message objects (see that
    # library)

        my $contents = $message->get_file(name=>$foo);

    # Kill the temporary server.  Set the autosupport vendor URL to
    # its previous value.

        $ar->stop();

=head1 DESCRIPTION

This library allows tests to set themselves up to receive autosupport
messages from an ONTAP cluster, to wait for an autosupport message to
be received, and to access the files/sections contained within the
autosupport message.  This is done by starting a listening process on
the local host.

This library implementation supports only C-mode ONTAP for now (but is
mapped out so that it could support both).  It also is structured to
support multiple transports (such as C<SMTP> or C<HTTP> or C<HTTPS>),
but only a non-HTTPS C<HTTP_Proxy> is implemented for now.
C<HTTP_Proxy> support is implemented by spawning an HTTP proxy server
(on the host executing this library) listening on an ephemeral port,
which forwards the autosupport message to its usual "vendor"/"support"
destination, but also captures the messages as they go by.

If the contents of the archives sent in autosupport will be examined,
the local host will need to have the appropriate unpacker available:
7za for ".7z" (7-zip) files, tar and gzip for ".tgz" files.

=head1 ATTRIBUTES

=head2 repository

The path to the directory into which autosupport messages are being
received.  (See C<start> for how this is set initially)

=cut

use Class::MethodMaker [

    # Private.  The public method is 'start', instead. create an
    # instance of this object and assign values to these attributes.
    new => [ '-hash', 'new' ],

    # Public, documented above.
    scalar => 'repository',

    # Private.  The command_interface passed to start(), used by other
    # instance methods to talk to the node/cluster
    scalar =>
      [ { -type => 'NACL::C::CommandInterface::ONTAP' }, 'command_interface' ],

    # Private.  A NACL::C::SystemNodeAutosupport
    scalar =>
      [ { -type => 'NACL::C::SystemNodeAutosupport' }, '_asup_component' ],

    # Private.  A NACL::Tools::httpd::control.  An object controlling
    # the subprocess that is acting as an HTTP proxy
    scalar =>
      [ { -type => 'NACL::Tools::httpd::control' }, '_proxy' ],

    # Private.  If defined, the URL of the HTTP Proxy
    scalar => '_new_proxy_url',

    # Private.  If defined, then start did 'redirect' to reassign
    # 'http_proxy', and this is the value that 'stop' should reset it
    # back to.
    scalar => '_old_proxy_url',

    # Private.  If defined, then start did 'redirect' to reassign
    # 'transport', and this is the value that 'stop' should reset it
    # back to.
    scalar => '_old_transport',

    # Private.  An arrayref of
    # NACL::MTask::AutosupportReceiver::Message objects.  Have been
    # seen by new_message or wait_for_message.
    scalar => '_old_messages',

    # Private.  An arrayref of
    # NACL::MTask::AutosupportReceiver::Message objects.  Have NOT
    # been seen by new_message or wait_for_message.
    scalar => '_new_messages',
];

=head1 METHODS

=head2 start

        my $ar = NACL::MTask::AutosupportReceiver->start(
            command_interface => $node,
            redirect => 1,
        );

Start a receiver process in the background to receive autosupport
messages.  This returns a new C<NACL::MTask::AutosupportReceiver> that
can be used to interact with this background receiver.

=over

=item Options

=over

=item command_interface => $command_interface

(Required) A NACL::C::CommandInterface object (such as a NACL::C::Node
or NACL::C::Cserver) representing the management interface that will
have its autosupport settings changed or that will be told to send an
autosupport message.

=item repository => $path

(Optional in most tests) The directory into which autosupport messages
will be received and unpacked.  If the TOP_LOGDIR and RUNID test
parameters are defined, this defaults to
"$TOP_LOGDIR/$RUNID.asup.$count/" where $count is the first whole
number such that the given repository directory does not already
exist.  (In other words, it will be a directory alongside the usual
logfile for this subtest).  Otherwise, there is no default and this is
a required option.

=item transport => "HTTP_Proxy"

(Optional) What transport to use for autosupport.  Defaults to
"HTTP_Proxy", which is also the only currently allowed value.

=item redirect => 0|1

(Optional) If true, then modify the ONTAP autosupport settings to send
to our server by default.  For "HTTP_Proxy", this means changing the
"vendor"/"support" proxy server to point to our temporary proxy server
(but leaving the final destination the same as before).  It will be
set back to the previous value when C<stop> is called.

=back

=back

=cut

sub start {
    $Log->enter();
    my $pkg  = shift;
    my %opts = validate @_,
      {
        command_interface => { isa  => 'NACL::C::CommandInterface::ONTAP' },
        repository        => { type => SCALAR, optional => 1 },
        transport => {
            type    => SCALAR,
            regex   => qr/^HTTP_Proxy$/,
            default => "HTTP_Proxy",
        },
        redirect => { type => BOOLEAN, default => 1 },
      };
    my $redirect          = $opts{redirect};
    my $command_interface = $opts{command_interface};

    my $repository = $opts{repository} || $pkg->default_repository();
    if ( !-d $repository ) {
        mkdir($repository)
          or NATE::BaseException->throw(
            "Could not create repository directory '$repository': $^E");
    }

    my $asup_component = NACL::C::SystemNodeAutosupport->new(
        command_interface => $command_interface,
        node              => 'local'
    );

    my $proxy         = NACL::Tools::httpd::control->new();
    my $new_transport = "http";
    my $new_proxy_url = $proxy->start_proxy(
        peer       => $command_interface->hostrec->default_ip,
        repository => $repository
    );

    # ONTAP wants proxy-url attribute not to have the leading
    # "http://" or "https://" part.
    $new_proxy_url =~ s@http://@@;

    my $old_proxy_url;
    my $old_transport;
    if ($redirect) {
        my $state = $asup_component->state(
            requested_fields => [ "proxy-url", "transport" ] );
        my @modify_opts;
        if ( $new_proxy_url ne $state->proxy_url ) {
            $old_proxy_url = $state->proxy_url() ? $state->proxy_url() : "''";
            push @modify_opts, ( "proxy-url" => $new_proxy_url );
        }
        if ( $new_transport ne $state->transport ) {
            $old_transport = $state->transport;
            push @modify_opts, ( transport => $new_transport );
        }
        if (@modify_opts) {
            $asup_component->modify(@modify_opts);
        }
    }

    $Log->exit();
    return $pkg->new(
        command_interface => $command_interface,
        repository        => $repository,
        _asup_component   => $asup_component,
        _proxy            => $proxy,
        _new_proxy_url    => $new_proxy_url,
        _old_proxy_url    => $old_proxy_url,
        _old_transport    => $old_transport,
        _old_messages     => [],
        _new_messages     => [],
    );
}

=head2 invoke

   $ar->invoke(node=>'local',type=>'all',...);

This is a wrapper around C<< NACL::C::SystemNodeAutosupport->invoke
>>, and has the same behavior except the following option defaults:

=over

=item Options

=over

=item command_interface => $command_interface

This defaults to the C<command_interface> value passed to C<start>

=item url => $url

The C<url> option is supplied to talk to this library's transport
service, as determined by the C<transport> option supplied to
C<start>.  (Actually, this isn't true for C<< transport=>HTTP_Proxy
>>.  Instead, there is no override for C<url> because the default
"vendor"/"support" address is already appropriate)

=item %other_options

All the other key=>$value pairs passed to this C<invoke> are dealt
with by L<< NACL::C::SystemNodeAutosupport->invoke|lib-NACL-C-SystemNodeAutosupport-pm/invoke >>.

=back

=back

=cut

sub invoke {
    $Log->enter();
    my $self = shift;
    my %opts = validate_with(
        params => \@_,
        spec =>
          { command_interface => { default => $self->command_interface }, },
        allow_extra => 1,
    );
    $Log->exit();
    return NACL::C::SystemNodeAutosupport->invoke(%opts);
}

=head2 messages

    my @messages = $ar->messages();

This returns a C<NACL::MTask::AutosupportReceiver::Message> object for
each autosupport message that this instance of
C<NACL::MTask::AutosupportReceiver> has received.  If there are no
messages, an empty list will be returned.

=cut

sub messages {
    $Log->enter();
    my $self = shift;
    $self->_scan_messages();
    $Log->exit();
    return ( @{ $self->_old_messages }, @{ $self->_new_messages } );
}

=head2 new_message

    my $message  = $ar->new_message();

This returns the first NACL::MTask::AutosupportReceiver::Message
object which has not yet been returned by a call to a C<new_message>
or C<wait_for_message>.  If there are no new messages, undef will be
returned.

=cut

sub new_message {
    $Log->enter();
    my $self = shift;
    if ( !scalar( @{ $self->_new_messages } ) ) {
        $self->_scan_messages();
    }
    my $message = pop @{ $self->_new_messages };
    if ( defined($message) ) {
        push @{ $self->_old_messages }, $message;
    }
    $Log->exit();
    return $message;
}

=head2 wait_for_message

    my $message  = $ar->wait_for_message("method-timeout"=>$timeout);

Like C<new_message>, but it first waits for there to be a new message
before returning that message.

=over

=item Options

=over

=item "method-timeout" => $timeout

(Required) Throw an exception if the given timeout is exceeded.  This
is a usual NATE timeout (-1 infinite, small values are seconds delta,
large values are absolute time() values).

=back

=back

=cut

sub wait_for_message {
    $Log->enter();
    my $self = shift;
    my %opts = validate @_, { "method-timeout" => { type => SCALAR }, };
    if ( !scalar( @{ $self->_new_messages } ) ) {
        $self->_scan_messages( timeout => $opts{"method-timeout"} );
    }
    $Log->exit();
    return $self->new_message();
}

=head2 stop

    $ar->stop();

This does two things:

=over

=item 1

If the C<redirect> option was used, undoes the effects of C<redirect>
(resets the proxy server, mail server, or vendor URL, whichever were
changed, to their previous values)

=item 2

Kills/stops whatever process this instance had been using to recieve
autosupport messages.

=back

=cut

sub stop {
    $Log->enter();
    my $self = shift;
    my @modify_opts;
    if ( defined $self->_old_proxy_url ) {
        push @modify_opts, ( "proxy-url" => $self->_old_proxy_url );
    }
    if ( defined $self->_old_transport ) {
        push @modify_opts, ( transport => $self->_old_transport );
    }
    if (@modify_opts) {
        $self->_asup_component->modify(@modify_opts);
    }
    $self->_proxy->stop();
    $Log->exit();
}

# Private Routines

# Check for any/all new files that have been received, and fill in
# $self->_new_messages with a message object for each.

# If timeout option is defined, then wait up to that amount of time
# for at least one new message to have been received (and if none do,
# raise an exception)

sub _scan_messages {
    $Log->enter();
    my $self = shift;
    my %opts = validate @_, { timeout => { type => SCALAR, optional => 1 }, };
    my @new_filenames;

    # Gather as many files as we can without spending any time blocked
    # on reading.
  SCAN: while (1) {
        my $filename = $self->_proxy->wait_for_file(
            timeout    => 0,
            on_timeout => sub { no warnings qw(exiting); last SCAN; }
        );
        push @new_filenames, $filename;
    }

    # Still no files?  Did our caller want us to wait for some
    # (timeout option given)?  If so, wait for more, and it's an error
    # if none appear in time.

    if ( !@new_filenames && defined( $opts{timeout} ) ) {
        my $filename = $self->_proxy->wait_for_file(
            timeout    => $opts{timeout},
            on_timeout => sub {
                NATE::BaseException->throw(
"Timeout exceeded ($opts{timeout}) waiting for new autosupport messages"
                );
            },
        );
        push @new_filenames, $filename;
    }

    # Insert messages for each filename seen
    foreach my $filename (@new_filenames) {
        push @{ $self->_new_messages },
          NACL::MTask::AutosupportReceiver::Message->new(
            packed_path => $filename );
    }
    $Log->exit();
}

sub default_repository {
    $Log->enter();
    my $top_logdir = Tharn::param('TOP_LOGDIR');
    my $runid      = Tharn::param('RUNID');
    if (   defined($top_logdir)
        && length($top_logdir)
        && defined($runid)
        && length($runid) )
    {
        my $suffix = "";
        my $count  = 0;
        while (1) {
            my $name = "$top_logdir/$runid.asup$suffix";
            if ( !-e $name ) {
                $Log->exit();
                return $name;
            }
            $count++;
            $suffix = ".$count";
        }
    }
    NATE::BaseException->throw(
"'repository' option required (since TOP_LOGDIR and RUNID not defined; in other words, because this isn't a NATE test or it's a test without a log directory)"
    );
}

1;
