# Copyright (c) 2015 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#
## @summary NACL::MTask::BSDFileOps Module
## @author  nimishp@netapp.com
## @status public
## @pod here

=head1 NAME

NACL::MTask::BSDFileOps

=head1 DESCRIPTION

C<NACL::MTask::BSDFileOps> provides a series of well-defined methods to execute native
filesystem calls such as write, read, unlink, etc. inside Data ONTAP's systemshell
(FreeBSD shell) to test internal filesystems such as SpinVFS (/mroot) and ClusFS (/clus).
Each instance of this library can be used for a single node.
Use multiple instances of this library for multiple nodes.

The library installs file_test and invokes its CLI to execute file operations.
"file_test" is available in the standard tools location at:
    $Test_Root/freebsd_vfs/bedrock/export/<variant>/uexe/file_test

Help and usge for file_test can be found here:
L<https://wikid.netapp.com/w/User:Nimishp/file_test/help#Usage_.26_Examples>

Exceptions for file operations return 'NACL::MTask::Exceptions::FileOperationFailed'.
Other exceptions by underlying NACL libraries will be different.

=head1 Author/Maintainer

dl-kernel-qa

=head1 ATTRIBUTES

=head2 command_interface

command interface object for the node whose systemshell will execute file operations.

=head2 _uniq_id

A scalar, this unique ID represents an instance of this library to handle multiple instances.

=head2 _installed_binary_path

A scalar, the path of the tool on the node, to use for executing file operations.

=head2 file_interface

NACL::STask::Node, a copy of the 'command_interface', to serve as a separate interface for file_test.

=head2 _freebsd_apiset

NACL::APISet, the systemshell APISet from the above 'file_interface', dedicated for running file_test's commands.

=head2 _unsupported_version_file_test

A scalar that if defined, holds the current "unsupported" version of file_test, thus blocking
any subsequent file_test operations from running. A mimimum supported version (2.08) is required.

=cut

package NACL::MTask::BSDFileOps;

=head1 SYNOPSIS

    # Create a new object instance
    my $node_fileops = NACL::MTask::BSDFileOps->new();

    # Setup the library (e.g. - install the test binary on the filer)
    $node_fileops->setup(command_interface => $node_object);

    # Open an existing file (or create a new one if nonexistent)
    my $fh1 = $node_fileops->open(
                path  => "/clus/vs0/vol-1/foo1.txt",
                flags => O_CREAT|O_RDWR,
                mode  => 755,       # permissions
            );

    # Open-create a new file for write-only access
    my $fh2 = $node_fileops->open(
                path  => "/clus/vs1/vol-2/foo2.txt",
                flags => O_WRONLY,
                mode  => 600,       # permissions
            );

    # Write to file-1
    my $written_data = $node_fileops->write(
        filehandle => $fh1,
        offset     => 0,
        num_bytes  => 10,
    );

    # [NEGATIVE]: Attempt to read from WRONLY file-2
    try {
        my $data = $node_fileops->read(
                filehandle => $fh2,
                offset     => 0,
                num_bytes  => 50,
            );
    } catch NACL::MTask::Exceptions::FileOperationFailed with {
        my $exception = shift;
        # handle exception here
        ...
        ...
        ...
    };

    # Read from file-1
    my $data = $node_fileops->read(
            filehandle => $fh1,
            offset     => 0,
            num_bytes  => 10,
    );
    $Test->nlog("Data read back from the file is: [$data]");

    # Close file-1's filehandle
    $node_fileops->close(filehandle => $fh1);

    # Close file-1's filehandle
    $node_fileops->close(filehandle => $fh2);

    # Cleanup: close all open file handles, uninstall test tool
    $node_fileops->cleanup();

=cut

use strict;
use warnings;

use Params::Validate qw(:all);
use Switch;
use Data::Dumper qw(Dumper);

use NACL::MTask::Exceptions::FileOperationFailed;
use NACL::MTask::Exceptions::PathValidationFailed;
use NACL::MTask::Exceptions::IncorrectParameterUsage;

use Os::BSDExecutor;
use NATE::Log qw(log_global);
use NATE::BaseException qw(:try);
use NACL::STask::Node;
use NACL::MTask::BSDInstall;
use NACL::GeneralUtils qw(random_name_generator);

# constants
use constant BINARY_NAME          => 'file_test';
use constant MIN_FILETEST_VERSION => '2.1';
use constant ERR_EIO              => 'Input/output error';

# method generators for new/set/get
use Class::MethodMaker [
    scalar => [ {-type => 'NACL::STask::Node'}, 'command_interface' ],
    scalar => [ {-type => 'NACL::STask::Node'}, 'file_interface' ],
    scalar => '_uniq_id',
    scalar => '_alt_binary_source_path',
    scalar => '_installed_binary_path',
    scalar => '_freebsd_apiset',
    scalar => '_unsupported_version_file_test',
    new    => 'new',
];

my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();

=head1 METHODS

=head2 new

Instantiates a new NACL::MTask::BSDFileOps object.

=over

=item Options

=over

=item NONE

=back

=back

=over

=item Returns

Returns a new NACL::MTask::BSDFileOps object instance.

=item Example

 my $node_fileops = NACL::MTask::BSDFileOps->new();

=back

=cut

### new() - auto-generated by Class::MethodMaker ###

=head2 setup

Sets up the object instance to run file operations.

=over

=item Options

=over

=item C<< "command_interface => $ci" >>

(Required) The command interface (NACL::STask::Node object) of the filer
on which file operations are to be executed.

=item C<< "binary_source_path => $alt_path_to_binary" >>

(Optional) An alternate path to the 'file_test' binary.
Default: pick the binary from the build:
  $Test_Root/freebsd_vfs/bedrock/export/<variant>/uexe/file_test

=back

=back

=over

=item Returns

Nothing if successful, exceptions for failures.

=item Example

 my @Nodes = NACL::STask::Node->find();
 $node_fileops->setup(command_interface => $Nodes[0]);

=back

=cut

sub setup {

    my $self = shift;
    my %args = Params::Validate::validate_with(
        params => \@_,
        spec   => {
            command_interface   => { isa => 'NACL::STask::Node' },
            binary_source_path  => { type => SCALAR, optional => 1 },
        },
    );

    # save the node as the command interface
    $self->command_interface( $args{command_interface} );

    # if provided, remember the alternate source path for the binary
    if (defined $args{binary_source_path}) {
        $self->_alt_binary_source_path( $args{binary_source_path} );
    }

    # install file_test if needed, and start it up
    $self->_load_bsd_file_tool();

    return;
} ## end sub setup

=head2 cleanup

Cleans up up the object instance, including uninstalling the test binary.

=over

=item Options

=over

=item NONE

=back

=back

=over

=item Returns

Nothing if successful, exceptions for failures.

=item Example

 $node_fileops->cleanup();

=back

=cut

sub cleanup {

    my $self = shift;
    my %args = Params::Validate::validate_with(
        params => \@_,
        spec   => {},
    );

    if (defined $self->{_installed_binary_path}) {
        # quit the tool (if running)
        $self->_get_freebsd_apiset()
            ->execute_raw_command(command => "quit");

        # uninstall the tool
        $Log->comment("Uninstalling the "
                . BINARY_NAME
                . " binary from node "
                . $self->command_interface()->node());
        $self->{_bsd_install}->uninstall_binaries(binaries => [ $self->{_uniq_id} ]);

        # clear the saved path
        undef $self->{_installed_binary_path};
    }

    return;
} ## end sub cleanup

###############################################################################
#           Filepath operation methods
###############################################################################

=head2 open

Use native open() to open/create a file and return its filehandle.

=over

=item Options

=over

=item C<< path => $file_path >>

(Required, SCALAR) The absolute path or relative path of the file on the filer.
    Example: "/clus/vs0/vol1/file1", "/mroot/etc/dummyfile1"

=item C<< flags => $flags >>

(Required, SCALAR) A string containing flags for open().
    Currently supported flags are:
        O_RDONLY, O_WRONLY, O_RDWR
        O_CREAT, O_EXCL, O_APPEND, O_TRUNC, O_DIRECTORY, O_SYNC
    These flags can be combined as follows:
        my $flags = "O_CREAT|O_RDWR";

=item C<< mode => $mode >>

(Required, SCALAR) mode, or permissions for the file.
    Example: 777, 644, etc.

=item C<< nacltask_verify => 1|0 >>

(Optional, BOOLEAN) if nacltask_verify = 1, verify if the file path exists.
    Default = 1 (verify path existence)

=item C<< nacltask_fail_for_eio => 1|0 >>

(Optional, BOOLEAN) if set to 1, fail for the generic EIO ("Input/Output") error.
    Default = 0  (false: retry a max of "nacltask_retry_limit" times)
    Note: if set to 1, then EIO cannot be part of the retryable list of errors (see below).

=item C<< nacltask_error_retry_list => [ "error-string 1" "error string 2" ] >>

(Optional, ARRAYREF) if a list is provided, retry a max of "nacltask_retry_limit" times.
    Default = throw exception on first occurrence of failure
    Note: will retry by default for EIO ("Input/Output") error unless suppressed.
        if the list contains EIO, cannot set "nacltask_fail_for_eio" to 1 above.

=item C<< nacltask_retry_limit => SCALAR >>

(Optional, SCALAR) maximum allowable retries on error.
    Default = 1 (one more attempt after hitting an error)

=back

=back

=over

=item Returns

Returns a SCALAR containing the 'filehandle' for success,
else throws an exception of type NACL::MTask::Exceptions::FileOperationFailed for failure.
If nacltask_verify = 1, and path validation fails, throws an exception of type NACL::MTask::Exceptions::PathValidationFailed.
The exception will contain the native error code and error string,
available to catch for negative testing.

=item Examples

 # Open an existing file (or create a new one if nonexistent)
    my $fh1 = $node_fileops->open(
        path  => "/clus/vs0/vol-1/foo1.txt",
        flags => O_CREAT|O_RDWR,
        mode  => 755,       # permissions
    );

 # Open an existing file for write-only access
    my $fh2 = $node_fileops->open(
                path  => "/clus/vs1/vol-2/foo2.txt",
                flags => O_WRONLY,
                mode  => 600,       # permissions
            );

=back

=cut

sub open {

    my $self = shift;
    my %args = Params::Validate::validate_with(
        params => \@_,
        spec   => {
            path  => {type => SCALAR,},
            flags => {type => SCALAR,},
            mode  => {     # aka permissions
                type => SCALAR,
            },
            nacltask_verify => {
                type    => SCALAR,
                default => 1,
            },
            nacltask_fail_for_eio => {
                type    => SCALAR,
                default => 0,
            },
            nacltask_error_retry_list => {
                type    => ARRAYREF,
                optional=> 1,
            },
            nacltask_retry_limit => {
                type    => SCALAR,
                default => 1,       # retry once after first failure, bail if error persists
            },
            timeout => {
                type    => SCALAR,
                default => 60,
            },
        },
    );

    # Steps:
    # ------
    # 1. build the command for file_test
    # 2. execute the command on the filer
    #    e.g. -    $node->get_systemshell_apiset()->execute_raw_command(
    #                   command => "...cmd...",
    #               );
    # 3. parse the output, return FH or throw an exception of type
    #    NACL::MTask::Exceptions::FileOperationFailed

    my $path   = delete $args{path};
    my $flags  = $self->_translate_flags_for_open(flags => $args{flags});
    my $mode   = (defined $args{mode}) ? delete $args{mode} : "";
    my $cmd    = "open $path $flags $mode";
    my $verify = delete $args{nacltask_verify};
    delete $args{flags};

    $Log->comment("\nOpening file: [$path]...\n");
    my $output = $self->_run_file_command(command => $cmd, %args);

    # check the filehandle
    my $filehandle = $self->_parse_file_handle(output => $output);
    if (!defined $filehandle) {
        throw NACL::MTask::Exceptions::FileOperationFailed(
                  "open() returned an undefined filehandle, "
                . "maybe the file operation failed: \n$output\n");
    }

    # validate file path ?
    $self->verify_file_path(path => $path)      if ($verify);

    return $filehandle;
} ## end sub open

=head2 close

Use native close() to close an open filehandle.

=over

=item Options

=over

=item C<< filehandle => $filehandle >>

(Required, SCALAR) The filehandle returned by the open() call above.

=item C<< nacltask_fail_for_eio => 1|0 >>

(Optional, BOOLEAN) if set to 1, fail for the generic EIO ("Input/Output") error.
    Default = 0  (false: retry a max of "nacltask_retry_limit" times)
    Note: if set to 1, then EIO cannot be part of the retryable list of errors (see below).

=item C<< nacltask_error_retry_list => [ "error-string 1" "error string 2" ] >>

(Optional, ARRAYREF) if a list is provided, retry a max of "nacltask_retry_limit" times.
    Default = throw exception on first occurrence of failure
    Note: will retry by default for EIO ("Input/Output") error unless suppressed.
        if the list contains EIO, cannot set "nacltask_fail_for_eio" to 1 above.

=item C<< nacltask_retry_limit => SCALAR >>

(Optional, SCALAR) maximum allowable retries on error.
    Default = 1 (one more attempt after hitting an error)

=back

=back

=over

=item Returns

Returns nothing for success,
else throws an exception of type NACL::MTask::Exceptions::FileOperationFailed for failure.
The exception will contain the native error code and error string,
available to catch for negative testing.

=item Examples

    # Close a filehandle
    $node_fileops->close(filehandle => $filehandle);

=back

=cut

sub close {

    my $self = shift;
    my %args = Params::Validate::validate_with(
        params => \@_,
        spec   => {
            filehandle => { type => SCALAR },
            nacltask_fail_for_eio => {
                type    => SCALAR,
                default => 0,
            },
            nacltask_error_retry_list => {
                type    => ARRAYREF,
                optional=> 1,
            },
            nacltask_retry_limit => {
                type    => SCALAR,
                default => 1,       # retry once after first failure, bail if error persists
            },
            timeout => {
                type    => SCALAR,
                default => 60,
            },
        },
    );

    my $fh = delete $args{filehandle};
    return $self->_run_file_command(command => "close $fh", %args);
} ## end sub close

=head2 link

Use native link() to create a hardlink to a file.

=over

=item Options

=over

=item C<< path => $file_path >>

(Required, SCALAR) The absolute path or relative path of the file on the filer.
Example: "/clus/vs0/vol1/file1", "/mroot/etc/dummyfile1"

=item C<< link => $link_path >>

(Required, SCALAR) The absolute path or relative path of the link on the filer.
Example: "/clus/vs0/vol1/link1", "/mroot/etc/link1"

=item C<< nacltask_verify => 1|0 >>

(Optional, BOOLEAN) if nacltask_verify = 1, verify if the link file path exists.
    Default = 1 (verify path existence)

=item C<< nacltask_fail_for_eio => 1|0 >>

(Optional, BOOLEAN) if set to 1, fail for the generic EIO ("Input/Output") error.
    Default = 0  (false: retry a max of "nacltask_retry_limit" times)
    Note: if set to 1, then EIO cannot be part of the retryable list of errors (see below).

=item C<< nacltask_error_retry_list => [ "error-string 1" "error string 2" ] >>

(Optional, ARRAYREF) if a list is provided, retry a max of "nacltask_retry_limit" times.
    Default = throw exception on first occurrence of failure
    Note: will retry by default for EIO ("Input/Output") error unless suppressed.
        if the list contains EIO, cannot set "nacltask_fail_for_eio" to 1 above.

=item C<< nacltask_retry_limit => SCALAR >>

(Optional, SCALAR) maximum allowable retries on error.
    Default = 1 (one more attempt after hitting an error)

=back

=back

=over

=item Returns

Returns nothing for success,
else throws an exception of type NACL::MTask::Exceptions::FileOperationFailed for failure.
If nacltask_verify = 1, and path validation fails, throws an exception of type NACL::MTask::Exceptions::PathValidationFailed.
The exception will contain the native error code and error string,
available to catch for negative testing.

=item Examples

    # Link a file
    $node_fileops->link(
        path => "/clus/vs0/vol-1/foo1.txt",
        link => "/clus/vs0/vol-1/link1",
    );

=back

=cut

sub link {

    my $self = shift;
    my %args = Params::Validate::validate_with(
        params => \@_,
        spec   => {
            path => {type => SCALAR,},
            link => {type => SCALAR,},
            nacltask_verify => {
                type    => SCALAR,
                default => 1,
            },
            nacltask_fail_for_eio => {
                type    => SCALAR,
                default => 0,
            },
            nacltask_error_retry_list => {
                type    => ARRAYREF,
                optional=> 1,
            },
            nacltask_retry_limit => {
                type    => SCALAR,
                default => 1,       # retry once after first failure, bail if error persists
            },
            timeout => {
                type    => SCALAR,
                default => 60,
            },
        },
    );

    my $path   = delete $args{path};
    my $link   = delete $args{link};
    my $verify = delete $args{nacltask_verify};

    my $ret = $self->_run_file_command(command => "hlink $path $link", %args);

    # validate file path ?
    $self->verify_file_path(path => $link)      if ($verify);

    return $ret;
} ## end sub link

=head2 symlink

Use native symlink() to create a symlink to a file.

=over

=item Options

=over

=item C<< path => $file_path >>

(Required, SCALAR) The absolute path or relative path of the file on the filer.
Example: "/clus/vs0/vol1/file1", "/mroot/etc/dummyfile1"

=item C<< link => $link_path >>

(Required, SCALAR) The absolute path or relative path of the link on the filer.
Example: "/clus/vs0/vol1/link1", "/mroot/etc/link1"

=item C<< nacltask_verify => 1|0 >>

(Optional, BOOLEAN) if nacltask_verify = 1, verify if the link file path exists.
    Default = 1 (verify path existence)

=item C<< nacltask_fail_for_eio => 1|0 >>

(Optional, BOOLEAN) if set to 1, fail for the generic EIO ("Input/Output") error.
    Default = 0  (false: retry a max of "nacltask_retry_limit" times)
    Note: if set to 1, then EIO cannot be part of the retryable list of errors (see below).

=item C<< nacltask_error_retry_list => [ "error-string 1" "error string 2" ] >>

(Optional, ARRAYREF) if a list is provided, retry a max of "nacltask_retry_limit" times.
    Default = throw exception on first occurrence of failure
    Note: will retry by default for EIO ("Input/Output") error unless suppressed.
        if the list contains EIO, cannot set "nacltask_fail_for_eio" to 1 above.

=item C<< nacltask_retry_limit => SCALAR >>

(Optional, SCALAR) maximum allowable retries on error.
    Default = 1 (one more attempt after hitting an error)

=back

=back

=over

=item Returns

Returns nothing for success,
else throws an exception of type NACL::MTask::Exceptions::FileOperationFailed for failure.
If nacltask_verify = 1, and path validation fails, throws an exception of type NACL::MTask::Exceptions::PathValidationFailed.
The exception will contain the native error code and error string,
available to catch for negative testing.

=item Examples

    # Symlink a file
    $node_fileops->symlink(
        path => "/clus/vs0/vol-1/foo1.txt",
        link => "/clus/vs0/vol-1/link1",
    );

=back

=cut

sub symlink {

    my $self = shift;
    my %args = Params::Validate::validate_with(
        params => \@_,
        spec   => {
            path => {type => SCALAR,},
            link => {type => SCALAR,},
            nacltask_verify => {
                type    => SCALAR,
                default => 1,
            },
            nacltask_fail_for_eio => {
                type    => SCALAR,
                default => 0,
            },
            nacltask_error_retry_list => {
                type    => ARRAYREF,
                optional=> 1,
            },
            nacltask_retry_limit => {
                type    => SCALAR,
                default => 1,       # retry once after first failure, bail if error persists
            },
            timeout => {
                type    => SCALAR,
                default => 60,
            },
        },
    );

    my $path   = delete $args{path};
    my $link   = delete $args{link};
    my $verify = delete $args{nacltask_verify};

    my $ret = $self->_run_file_command(command => "slink $path $link", %args);

    # validate file path ?
    $self->verify_file_path(path => $link)      if ($verify);

    return $ret;
} ## end sub symlink

=head2 unlink

Use native unlink() to unlink to a file.

=over

=item Options

=over

=item C<< path => $file_path >>

(Required, SCALAR) The absolute path or relative path of the file on the filer.
Example: "/clus/vs0/vol1/file1", "/mroot/etc/dummyfile1"

=item C<< nacltask_verify => 1|0 >>

(Optional, BOOLEAN) if nacltask_verify = 1, verify that the link file path does not exist.
    Default = 1 (verify path's non-existence)

=item C<< nacltask_fail_for_eio => 1|0 >>

(Optional, BOOLEAN) if set to 1, fail for the generic EIO ("Input/Output") error.
    Default = 0  (false: retry a max of "nacltask_retry_limit" times)
    Note: if set to 1, then EIO cannot be part of the retryable list of errors (see below).

=item C<< nacltask_error_retry_list => [ "error-string 1" "error string 2" ] >>

(Optional, ARRAYREF) if a list is provided, retry a max of "nacltask_retry_limit" times.
    Default = throw exception on first occurrence of failure
    Note: will retry by default for EIO ("Input/Output") error unless suppressed.
        if the list contains EIO, cannot set "nacltask_fail_for_eio" to 1 above.

=item C<< nacltask_retry_limit => SCALAR >>

(Optional, SCALAR) maximum allowable retries on error.
    Default = 1 (one more attempt after hitting an error)

=back

=back

=over

=item Returns

Returns nothing for success,
else throws an exception of type NACL::MTask::Exceptions::FileOperationFailed for failure.
If nacltask_verify = 1, and path validation fails, throws an exception of type NACL::MTask::Exceptions::PathValidationFailed.
The exception will contain the native error code and error string,
available to catch for negative testing.

=item Examples

    # Unink a file
    $node_fileops->unlink(
        path => "/clus/vs0/vol-1/foo1.txt",
    );

=back

=cut

sub unlink {

    my $self = shift;
    my %args = Params::Validate::validate_with(
        params => \@_,
        spec   => {
            path => {type => SCALAR,},
            nacltask_verify => {
                type    => SCALAR,
                default => 1,
            },
            nacltask_fail_for_eio => {
                type    => SCALAR,
                default => 0,
            },
            nacltask_error_retry_list => {
                type    => ARRAYREF,
                optional=> 1,
            },
            nacltask_retry_limit => {
                type    => SCALAR,
                default => 1,       # retry once after first failure, bail if error persists
            },
            timeout => {
                type    => SCALAR,
                default => 60,
            },
        },
    );

    my $path   = delete $args{path};
    my $verify = delete $args{nacltask_verify};

    my $ret = $self->_run_file_command(command => "ulink $path", %args);

    # validate file path ?
    $self->verify_file_path(path => $path, nacltask_ignore_failure => 1)
        if ($verify);

    return $ret;
} ## end sub unlink

=head2 remove

Use native remove() to unlink/delete to a file.

=over

=item Options

=over

=item C<< path => $file_path >>

(Required, SCALAR) The absolute path or relative path of the file on the filer.
Example: "/clus/vs0/vol1/file1", "/mroot/etc/dummyfile1"

=item C<< nacltask_verify => 1|0 >>

(Optional, BOOLEAN) if nacltask_verify = 1, verify that the file path does not exist.
    Default = 1 (verify path's non-existence)

=item C<< nacltask_fail_for_eio => 1|0 >>

(Optional, BOOLEAN) if set to 1, fail for the generic EIO ("Input/Output") error.
    Default = 0  (false: retry a max of "nacltask_retry_limit" times)
    Note: if set to 1, then EIO cannot be part of the retryable list of errors (see below).

=item C<< nacltask_error_retry_list => [ "error-string 1" "error string 2" ] >>

(Optional, ARRAYREF) if a list is provided, retry a max of "nacltask_retry_limit" times.
    Default = throw exception on first occurrence of failure
    Note: will retry by default for EIO ("Input/Output") error unless suppressed.
        if the list contains EIO, cannot set "nacltask_fail_for_eio" to 1 above.

=item C<< nacltask_retry_limit => SCALAR >>

(Optional, SCALAR) maximum allowable retries on error.
    Default = 1 (one more attempt after hitting an error)

=back

=back

=over

=item Returns

Returns nothing for success,
else throws an exception of type NACL::MTask::Exceptions::FileOperationFailed for failure.
If nacltask_verify = 1, and path validation fails, throws an exception of type NACL::MTask::Exceptions::PathValidationFailed.
The exception will contain the native error code and error string,
available to catch for negative testing.

=item Examples

    # Remove a file
    $node_fileops->remove(
        path => "/clus/vs0/vol-1/foo1.txt",
    );

=back

=cut

sub remove {

    my $self = shift;
    my %args = Params::Validate::validate_with(
        params => \@_,
        spec   => {
            path => {type => SCALAR,},
            nacltask_verify => {
                type    => SCALAR,
                default => 1,
            },
            nacltask_fail_for_eio => {
                type    => SCALAR,
                default => 0,
            },
            nacltask_error_retry_list => {
                type    => ARRAYREF,
                optional=> 1,
            },
            nacltask_retry_limit => {
                type    => SCALAR,
                default => 1,       # retry once after first failure, bail if error persists
            },
            timeout => {
                type    => SCALAR,
                default => 60,
            },
        },
    );

    my $path   = delete $args{path};
    my $verify = delete $args{nacltask_verify};

    my $ret = $self->_run_file_command(command => "delete $path", %args);

    # validate file path ?
    $self->verify_file_path(path => $path, nacltask_ignore_failure => 1)
        if ($verify);

    return $ret;
} ## end sub remove

=head2 truncate

Use native truncate() to truncate a file.

=over

=item Options

=over

=item C<< path => $file_path >>

(Required, SCALAR) The absolute path or relative path of the source file on the filer.
Example: "/clus/vs0/vol1/file1", "/mroot/etc/dummyfile1"


=item C<< length => $length >>

(Required, SCALAR) The length in bytes to which file will be truncated .

                this->util->print("truncate a file to a specified length in bytes\n");


=item C<< nacltask_fail_for_eio => 1|0 >>

(Optional, BOOLEAN) if set to 1, fail for the generic EIO ("Input/Output") error.
    Default = 0  (false: retry a max of "nacltask_retry_limit" times)
    Note: if set to 1, then EIO cannot be part of the retryable list of errors (see below).

=item C<< nacltask_error_retry_list => [ "error-string 1" "error string 2" ] >>

(Optional, ARRAYREF) if a list is provided, retry a max of "nacltask_retry_limit" times.
    Default = throw exception on first occurrence of failure
    Note: will retry by default for EIO ("Input/Output") error unless suppressed.
        if the list contains EIO, cannot set "nacltask_fail_for_eio" to 1 above.

=item C<< nacltask_retry_limit => SCALAR >>

(Optional, SCALAR) maximum allowable retries on error.
    Default = 1 (one more attempt after hitting an error)

=back

=back

=over

=item Returns

Returns nothing for success,
else throws an exception of type NACL::MTask::Exceptions::FileOperationFailed for failure.
The exception will contain the native error code and error string,
available to catch for negative testing.

=item Examples

    # Rename a file
    $node_fileops->truncate(
        path => "/clus/vs0/vol-1/foo1.txt",
        length => "100",
    );

=back

=cut

sub truncate {

    my $self = shift;
    my %args = Params::Validate::validate_with(
        params => \@_,
        spec   => {
            path     => {type => SCALAR,},
            length => {type => SCALAR,},
            nacltask_fail_for_eio => {
                type    => SCALAR,
                default => 0,
            },
            nacltask_error_retry_list => {
                type    => ARRAYREF,
                optional=> 1,
            },
            nacltask_retry_limit => {
                type    => SCALAR,
                default => 1,       # retry once after first failure, bail if error persists
            },
            timeout => {
                type    => SCALAR,
                default => 60,
            },
        },
    );

    my $path     = delete $args{path};
    my $length   = delete $args{length};

    my $ret = $self->_run_file_command(command => "truncate $path $length",%args);
    return $ret;
} ## end sub truncate

=head2 rename

Use native rename() to rename a file.

=over

=item Options

=over

=item C<< path => $file_path >>

(Required, SCALAR) The absolute path or relative path of the source file on the filer.
Example: "/clus/vs0/vol1/file1", "/mroot/etc/dummyfile1"

=item C<< new_path => $new_path >>

(Required, SCALAR) The new absolute or relative path or filename on the filer.
Example: "/clus/vs0/vol1/file2", "/mroot/etc/foofile1234"

=item C<< nacltask_verify => 1|0 >>

(Optional, BOOLEAN) if nacltask_verify = 1, verify if the new file path exists.
    Default = 1 (verify new path's existence)

=item C<< nacltask_fail_for_eio => 1|0 >>

(Optional, BOOLEAN) if set to 1, fail for the generic EIO ("Input/Output") error.
    Default = 0  (false: retry a max of "nacltask_retry_limit" times)
    Note: if set to 1, then EIO cannot be part of the retryable list of errors (see below).

=item C<< nacltask_error_retry_list => [ "error-string 1" "error string 2" ] >>

(Optional, ARRAYREF) if a list is provided, retry a max of "nacltask_retry_limit" times.
    Default = throw exception on first occurrence of failure
    Note: will retry by default for EIO ("Input/Output") error unless suppressed.
        if the list contains EIO, cannot set "nacltask_fail_for_eio" to 1 above.

=item C<< nacltask_retry_limit => SCALAR >>

(Optional, SCALAR) maximum allowable retries on error.
    Default = 1 (one more attempt after hitting an error)

=back

=back

=over

=item Returns

Returns nothing for success,
else throws an exception of type NACL::MTask::Exceptions::FileOperationFailed for failure.
If nacltask_verify = 1, and path validation fails, throws an exception of type NACL::MTask::Exceptions::PathValidationFailed.
The exception will contain the native error code and error string,
available to catch for negative testing.

=item Examples

    # Rename a file
    $node_fileops->rename(
        path => "/clus/vs0/vol-1/foo1.txt",
        new_path => "/clus/vs0/vol-1/dir2/foo2.txt",
    );

=back

=cut

sub rename {

    my $self = shift;
    my %args = Params::Validate::validate_with(
        params => \@_,
        spec   => {
            path     => {type => SCALAR,},
            new_path => {type => SCALAR,},
            nacltask_verify => {
                type    => SCALAR,
                default => 1,
            },
            nacltask_fail_for_eio => {
                type    => SCALAR,
                default => 0,
            },
            nacltask_error_retry_list => {
                type    => ARRAYREF,
                optional=> 1,
            },
            nacltask_retry_limit => {
                type    => SCALAR,
                default => 1,       # retry once after first failure, bail if error persists
            },
            timeout => {
                type    => SCALAR,
                default => 60,
            },
        },
    );

    my $path     = delete $args{path};
    my $new_path = delete $args{new_path};
    my $verify   = delete $args{nacltask_verify};

    my $ret = $self->_run_file_command(command => "rename $path $new_path",%args);

    if ($verify) {
        # validate file path ?
        $self->verify_file_path(path => $new_path, nacltask_ignore_failure => 1);

        # - validate old path does not exist
        $self->verify_file_path(path => $new_path, nacltask_ignore_failure => 1);
    }

    return $ret;
} ## end sub rename

=head2 move

Use native rename() to move a file.
There is no special move() API, just a wrapper over rename().

=over

=item Options

=over

=item C<< path => $file_path >>

(Required, SCALAR) The absolute path or relative path of the source file on the filer.
Example: "/clus/vs0/vol1/file1", "/mroot/etc/dummyfile1"

=item C<< new_path => $new_path >>

(Required, SCALAR) The absolute path or relative path of the destination on the filer.
Example: "/clus/vs0/vol1/file2", "/mroot/etc/foofile1234"

=item C<< nacltask_verify => 1|0 >>

(Optional, BOOLEAN) if nacltask_verify = 1, verify if the new file path exists.
    Default = 1 (verify new path's existence)

=item C<< nacltask_fail_for_eio => 1|0 >>

(Optional, BOOLEAN) if set to 1, fail for the generic EIO ("Input/Output") error.
    Default = 0  (false: retry a max of "nacltask_retry_limit" times)
    Note: if set to 1, then EIO cannot be part of the retryable list of errors (see below).

=item C<< nacltask_error_retry_list => [ "error-string 1" "error string 2" ] >>

(Optional, ARRAYREF) if a list is provided, retry a max of "nacltask_retry_limit" times.
    Default = throw exception on first occurrence of failure
    Note: will retry by default for EIO ("Input/Output") error unless suppressed.
        if the list contains EIO, cannot set "nacltask_fail_for_eio" to 1 above.

=item C<< nacltask_retry_limit => SCALAR >>

(Optional, SCALAR) maximum allowable retries on error.
    Default = 1 (one more attempt after hitting an error)

=back

=back

=over

=item Returns

Returns nothing for success,
else throws an exception of type NACL::MTask::Exceptions::FileOperationFailed for failure.
If nacltask_verify = 1, and path validation fails, throws an exception of type NACL::MTask::Exceptions::PathValidationFailed.
The exception will contain the native error code and error string,
available to catch for negative testing.

=item Examples

    # Move a file
    $node_fileops->move(
        path => "/clus/vs0/vol-1/foo1.txt",
        new_path => "/clus/vs0/vol-1/dir2/foo2.txt",
    );

=back

=cut

sub move {

    my $self = shift;

    return $self->rename(@_);
} ## end sub move

=head2 stat

Use native stat() to stat to a file.

=over

=item Options

=over

=item C<< path => $file_path >>

(Required, SCALAR) The absolute path or relative path of the file on the filer.
Example: "/clus/vs0/vol1/file1", "/mroot/etc/dummyfile1"

=item C<< return_data => \%stat_data >>

(Optional, HASHREF) A reference to an empty hash to return stat data.
If not provided, stat data will simply be printed to the logfile.

=item C<< nacltask_fail_for_eio => 1|0 >>

(Optional, BOOLEAN) if set to 1, fail for the generic EIO ("Input/Output") error.
    Default = 0  (false: retry a max of "nacltask_retry_limit" times)
    Note: if set to 1, then EIO cannot be part of the retryable list of errors (see below).

=item C<< nacltask_error_retry_list => [ "error-string 1" "error string 2" ] >>

(Optional, ARRAYREF) if a list is provided, retry a max of "nacltask_retry_limit" times.
    Default = throw exception on first occurrence of failure
    Note: will retry by default for EIO ("Input/Output") error unless suppressed.
        if the list contains EIO, cannot set "nacltask_fail_for_eio" to 1 above.

=item C<< nacltask_retry_limit => SCALAR >>

(Optional, SCALAR) maximum allowable retries on error.
    Default = 1 (one more attempt after hitting an error)

=back

=back

=over

=item Returns

Returns nothing for success,
else throws an exception of type NACL::MTask::Exceptions::FileOperationFailed for failure.
The exception will contain the native error code and error string,
available to catch for negative testing.

=item Examples

    # Stat a file
    my %stat_data;
    $node_fileops->stat(
        path        => "/clus/vs0/vol-1/foo1.txt",
        return_data => \%stat_data,
    );

=back

=cut

sub stat {

    my $self = shift;
    my %args = Params::Validate::validate_with(
        params => \@_,
        spec   => {
            path => {
                type => SCALAR,
            },
            nacltask_fail_for_eio => {
                type    => SCALAR,
                default => 0,
            },
            nacltask_error_retry_list => {
                type    => ARRAYREF,
                optional=> 1,
            },
            nacltask_retry_limit => {
                type    => SCALAR,
                default => 1,       # retry once after first failure, bail if error persists
            },
            return_data => {
                type     => HASHREF,
                optional => 0,
            },
            timeout => {
                type    => SCALAR,
                default => 60,
            },
            __unit_test_arg_count_attempts => {
                type     => SCALARREF,
                optional => 1,
            },
        },
    );

    my $path     = delete $args{path};
    my $ref_data = delete $args{return_data};

    my $output = $self->_run_file_command(command => "stat $path", %args);

    # parse stat data?
    if (defined $ref_data) {
        my @lines = split(/\n/, $output);
        foreach my $line (@lines) {
            if ($line =~ /(\S+)\s+\=\s+(\S+)/) {
                $ref_data->{$1} = $2;
            }
        }
    }

    return;
} ## end sub stat

=head2 getcwd

Use native getcwd() to list the current working directory.

=over

=item Options

=over

=item C<< max_path_len => $max_len >>

(Optional, SCALAR) The max path name length to allow, default = (UNIX - PATH_MAX).

=item C<< nacltask_fail_for_eio => 1|0 >>

(Optional, BOOLEAN) if set to 1, fail for the generic EIO ("Input/Output") error.
    Default = 0  (false: retry a max of "nacltask_retry_limit" times)
    Note: if set to 1, then EIO cannot be part of the retryable list of errors (see below).

=item C<< nacltask_error_retry_list => [ "error-string 1" "error string 2" ] >>

(Optional, ARRAYREF) if a list is provided, retry a max of "nacltask_retry_limit" times.
    Default = throw exception on first occurrence of failure
    Note: will retry by default for EIO ("Input/Output") error unless suppressed.
        if the list contains EIO, cannot set "nacltask_fail_for_eio" to 1 above.

=item C<< nacltask_retry_limit => SCALAR >>

(Optional, SCALAR) maximum allowable retries on error.
    Default = 1 (one more attempt after hitting an error)

=back

=back

=over

=item Returns

Returns a SCALAR containing the current path for success,
else throws an exception of type NACL::MTask::Exceptions::FileOperationFailed for failure.
The exception will contain the native error code and error string,
available to catch for negative testing.

=item Examples

    # Get current working directory
    my $cwd = $node_fileops->getcwd();

    # Get current working directory with max_path_len
    my $cwd = $node_fileops->getcwd(max_path_len => 4096);

=back

=cut

sub getcwd {

    my $self = shift;
    my %args = Params::Validate::validate_with(
        params => \@_,
        spec   => {
            max_path_len => {
                type     => SCALAR,
                optional => 1
            },
            nacltask_fail_for_eio => {
                type    => SCALAR,
                default => 0,
            },
            nacltask_error_retry_list => {
                type    => ARRAYREF,
                optional=> 1,
            },
            nacltask_retry_limit => {
                type    => SCALAR,
                default => 1,       # retry once after first failure, bail if error persists
            },
            timeout => {
                type    => SCALAR,
                default => 60,
            },
        },
    );

    my $max_path_len = (defined $args{max_path_len}) ?
                        " $args{max_path_len}" : "";
    delete $args{max_path_len};

    my $output = $self->_run_file_command(
        command      => "getcwd" . $max_path_len,
        match_output => 'Current working directory',
        %args
    );

    $_ = ($output =~ /Current working directory\:\s+(\S+)/);

    return $1;
} ## end sub getcwd

=head2 chdir

Use native chdir() to change the current working directory.

=over

=item Options

=over

=item C<< path => $new_dir_path >>

(Required, SCALAR) The absolute path or relative path to change to on the filer.
Example: "/clus/vs0/vol1", "/mroot/"

=item C<< nacltask_fail_for_eio => 1|0 >>

(Optional, BOOLEAN) if set to 1, fail for the generic EIO ("Input/Output") error.
    Default = 0  (false: retry a max of "nacltask_retry_limit" times)
    Note: if set to 1, then EIO cannot be part of the retryable list of errors (see below).

=item C<< nacltask_error_retry_list => [ "error-string 1" "error string 2" ] >>

(Optional, ARRAYREF) if a list is provided, retry a max of "nacltask_retry_limit" times.
    Default = throw exception on first occurrence of failure
    Note: will retry by default for EIO ("Input/Output") error unless suppressed.
        if the list contains EIO, cannot set "nacltask_fail_for_eio" to 1 above.

=item C<< nacltask_retry_limit => SCALAR >>

(Optional, SCALAR) maximum allowable retries on error.
    Default = 1 (one more attempt after hitting an error)

=back

=back

=over

=item Returns

Returns nothing for success,
else throws an exception of type NACL::MTask::Exceptions::FileOperationFailed for failure.
The exception will contain the native error code and error string,
available to catch for negative testing.

=item Examples

    # Remove a file
    $node_fileops->chdir(path => "/clus/vs0/vol-1/");

=back

=cut

sub chdir {

    my $self = shift;
    my %args = Params::Validate::validate_with(
        params => \@_,
        spec   => {
            path    => {type => SCALAR,},
            nacltask_fail_for_eio => {
                type    => SCALAR,
                default => 0,
            },
            nacltask_error_retry_list => {
                type    => ARRAYREF,
                optional=> 1,
            },
            nacltask_retry_limit => {
                type    => SCALAR,
                default => 1,       # retry once after first failure, bail if error persists
            },
            timeout => {
                type    => SCALAR,
                default => 60,
            },
        },
    );

    my $path = delete $args{path};
    return $self->_run_file_command(command => "chdir $path", %args);

} ## end sub chdir

=head2 opendir

Use native opendir() to read a working directory.

=over

=item Options

=over

=item C<< path => $new_dir_path >>

(Required, SCALAR) The absolute path or relative path to change to on the filer.
Example: "/clus/vs0/vol1", "/mroot/"

=item C<< nacltask_fail_for_eio => 1|0 >>

(Optional, BOOLEAN) if set to 1, fail for the generic EIO ("Input/Output") error.
    Default = 0  (false: retry a max of "nacltask_retry_limit" times)
    Note: if set to 1, then EIO cannot be part of the retryable list of errors (see below).

=item C<< nacltask_error_retry_list => [ "error-string 1" "error string 2" ] >>

(Optional, ARRAYREF) if a list is provided, retry a max of "nacltask_retry_limit" times.
    Default = throw exception on first occurrence of failure
    Note: will retry by default for EIO ("Input/Output") error unless suppressed.
        if the list contains EIO, cannot set "nacltask_fail_for_eio" to 1 above.

=item C<< nacltask_retry_limit => SCALAR >>

(Optional, SCALAR) maximum allowable retries on error.
    Default = 1 (one more attempt after hitting an error)

=back

=back

=over

=item Returns

Returns a SCALAR containing the 'dir handle' for success
else throws an exception of type NACL::MTask::Exceptions::FileOperationFailed for failure.
The exception will contain the native error code and error string,
available to catch for negative testing.

=item Examples

    # open a directory handle
    my $dirp = $node_fileops->opendir(path => "/clus/vs0/vol-1/");

=back

=cut

sub opendir {

    my $self = shift;
    my %args = Params::Validate::validate_with(
        params => \@_,
        spec   => {
            path    => {type => SCALAR,},
            nacltask_fail_for_eio => {
                type    => SCALAR,
                default => 0,
            },
            nacltask_error_retry_list => {
                type    => ARRAYREF,
                optional=> 1,
            },
            nacltask_retry_limit => {
                type    => SCALAR,
                default => 1,       # retry once after first failure, bail if error persists
            },
            timeout => {
                type    => SCALAR,
                default => 60,
            },
        },
    );

    my $path = delete $args{path};

    $Log->comment("\nOpening dir: [$path]...\n");
    my $output = $self->_run_file_command(command => "opendir $path", %args);

    # check the filehandle
    my $dirhandle = $self->_parse_dir_handle(output => $output);
    if (!defined $dirhandle) {
        throw NACL::MTask::Exceptions::FileOperationFailed(
                  "opendir() returned an undefined directory handle, "
                . "maybe the opendir operation failed: \n$output\n");
    }

    return $dirhandle;

} ## end sub opendir

=head2 readdir

Use native readdir() to read a working directory.

=over

=item Options

=over

=item C<< dirhandle => $dir_handle >>

(Required, SCALAR) The dirhandle returned by the opendir() call above.

=item C<< count => $count >>

(optional, SCALAR) The count of directory entries to read from the directory handle.
defaulted to 1 entries

=item C<< return_data => \%stat_data >>

(optional, HASHREF) A reference to an empty hash to return READDIR output.
If not provided, readdir data will simply be printed to the logfile.

=item C<< nacltask_fail_for_eio => 1|0 >>

(Optional, BOOLEAN) if set to 1, fail for the generic EIO ("Input/Output") error.
    Default = 0  (false: retry a max of "nacltask_retry_limit" times)
    Note: if set to 1, then EIO cannot be part of the retryable list of errors (see below).

=item C<< nacltask_error_retry_list => [ "error-string 1" "error string 2" ] >>

(Optional, ARRAYREF) if a list is provided, retry a max of "nacltask_retry_limit" times.
    Default = throw exception on first occurrence of failure
    Note: will retry by default for EIO ("Input/Output") error unless suppressed.
        if the list contains EIO, cannot set "nacltask_fail_for_eio" to 1 above.

=item C<< nacltask_retry_limit => SCALAR >>

(Optional, SCALAR) maximum allowable retries on error.
    Default = 1 (one more attempt after hitting an error)

=back

=back

=over

=item Returns

Returns nothing for success,
else throws an exception of type NACL::MTask::Exceptions::FileOperationFailed for failure.
The exception will contain the native error code and error string,
available to catch for negative testing.

=item Examples

    # Read $count directory entries
    $node_fileops->readdir(dirhandle => $dir_handle,count = "5");

=back

=cut

sub readdir {

    my $self = shift;
    my %args = Params::Validate::validate_with(
        params => \@_,
        spec   => {
            dirhandle    => {type => SCALAR,},
            count    => {type => SCALAR,
                         default => 1,
            },
            nacltask_fail_for_eio => {
                type    => SCALAR,
                default => 0,
            },
            nacltask_error_retry_list => {
                type    => ARRAYREF,
                optional=> 1,
            },
            nacltask_retry_limit => {
                type    => SCALAR,
                default => 1,       # retry once after first failure, bail if error persists
            },
            return_data => {
                type     => HASHREF,
                optional => 1,
            },
            timeout => {
                type    => SCALAR,
                default => 60,
            },
        },
    );

    my $dirhandle = delete $args{dirhandle};
    my $count = delete $args{count};
    my $ref_data = delete $args{return_data};

    my $output = $self->_run_file_command(command => "readdir $dirhandle $count", %args);

    # parse readdir data?
    if (defined $ref_data) {
        my @lines = split(/\n/, $output);
        my $dir_name ;
        foreach my $line (@lines) {
            if ($line =~ /Dirent\s+details\s+of\s+->\s+(\S+)/) {
                $dir_name = $1;
            }
            if ($line =~ /(\S+)\s+\=\s+(\S+)/) {
                $ref_data->{$dir_name}{$1} = $2;
            }
        }
    }

    return ;
} ## end sub readdir


=head2 closedir

Use native closedir() to close an open directory handle.

=over

=item Options

=over

=item C<< dirhandle => $dir_handle >>

(Required, SCALAR) The directory handle returned by the opendir() call above.

=item C<< nacltask_fail_for_eio => 1|0 >>

(Optional, BOOLEAN) if set to 1, fail for the generic EIO ("Input/Output") error.
    Default = 0  (false: retry a max of "nacltask_retry_limit" times)
    Note: if set to 1, then EIO cannot be part of the retryable list of errors (see below).

=item C<< nacltask_error_retry_list => [ "error-string 1" "error string 2" ] >>

(Optional, ARRAYREF) if a list is provided, retry a max of "nacltask_retry_limit" times.
    Default = throw exception on first occurrence of failure
    Note: will retry by default for EIO ("Input/Output") error unless suppressed.
        if the list contains EIO, cannot set "nacltask_fail_for_eio" to 1 above.

=item C<< nacltask_retry_limit => SCALAR >>

(Optional, SCALAR) maximum allowable retries on error.
    Default = 1 (one more attempt after hitting an error)

=back

=back

=over

=item Returns

Returns nothing for success,
else throws an exception of type NACL::MTask::Exceptions::FileOperationFailed for failure.
The exception will contain the native error code and error string,
available to catch for negative testing.

=item Examples

    # Close a dirhandle
    $node_fileops->close(dirhandle => $dir_handle);

=back

=cut

sub closedir {

    my $self = shift;
    my %args = Params::Validate::validate_with(
        params => \@_,
        spec   => {
            dirhandle => { type => SCALAR },
            nacltask_fail_for_eio => {
                type    => SCALAR,
                default => 0,
            },
            nacltask_error_retry_list => {
                type    => ARRAYREF,
                optional=> 1,
            },
            nacltask_retry_limit => {
                type    => SCALAR,
                default => 1,       # retry once after first failure, bail if error persists
            },
            timeout => {
                type    => SCALAR,
                default => 60,
            },
        },
    );

    my $dh = delete $args{dirhandle};
    return $self->_run_file_command(command => "closedir $dh", %args);
} ## end sub closedir

###############################################################################
#           Filehandle operation methods
###############################################################################

=head2 seek

Use native seek() to seek an open filehandle's file pointer to a specific offset.

=over

=item Options

=over

=item C<< filehandle => $filehandle >>

(Required, SCALAR) The filehandle returned by the open() call above.

=item C<< offset => $offset >>

(Required, SCALAR) The relative offset to seek, depending on the method below.

=item C<< seek_method => $seek_method >>

(Required, SCALAR) The method for how to seek the file pointer:
    Legal values are one of the following:
        SEEK_SET
        SEEK_CURRENT
        SEEK_END

=item C<< nacltask_fail_for_eio => 1|0 >>

(Optional, BOOLEAN) if set to 1, fail for the generic EIO ("Input/Output") error.
    Default = 0  (false: retry a max of "nacltask_retry_limit" times)
    Note: if set to 1, then EIO cannot be part of the retryable list of errors (see below).

=item C<< nacltask_error_retry_list => [ "error-string 1" "error string 2" ] >>

(Optional, ARRAYREF) if a list is provided, retry a max of "nacltask_retry_limit" times.
    Default = throw exception on first occurrence of failure
    Note: will retry by default for EIO ("Input/Output") error unless suppressed.
        if the list contains EIO, cannot set "nacltask_fail_for_eio" to 1 above.

=item C<< nacltask_retry_limit => SCALAR >>

(Optional, SCALAR) maximum allowable retries on error.
    Default = 1 (one more attempt after hitting an error)

=back

=back

=over

=item Returns

Returns nothing for success,
else throws an exception of type NACL::MTask::Exceptions::FileOperationFailed for failure.
The exception will contain the native error code and error string,
available to catch for negative testing.

=item Examples

    # Seek 5 bytes from the start of the file
    $node_fileops->seek(
        filehandle  => $filehandle,
        offset      => 5,
        seek_method => "SEEK_SET",
    );

    # Seek the file pointer's position to
    # 10 bytes from its current position
    $node_fileops->seek(
        filehandle  => $filehandle,
        offset      => 10,
        seek_method => "SEEK_CURRENT",
    );

=back

=cut

sub seek {

    my $self = shift;
    my %args = Params::Validate::validate_with(
        params => \@_,
        spec   => {
            filehandle  => {type => SCALAR,},
            offset      => {type => SCALAR,},
            seek_method => {
                type  => SCALAR,
                regex => qr/SEEK_SET|SEEK_CURRENT|SEEK_END/,
            },
            nacltask_fail_for_eio => {
                type    => SCALAR,
                default => 0,
            },
            nacltask_error_retry_list => {
                type    => ARRAYREF,
                optional=> 1,
            },
            nacltask_retry_limit => {
                type    => SCALAR,
                default => 1,       # retry once after first failure, bail if error persists
            },
            timeout => {
                type    => SCALAR,
                default => 60,
            },
        },
    );

    my $filehandle = delete $args{filehandle};
    my $offset     = delete $args{offset};

    # translate the seek method to a file_test acceptable value
    my $seek_method
        = $self->_translate_flags_for_seek(flag => $args{seek_method});
    delete $args{seek_method};

    return $self->_run_file_command(command => "seek $filehandle $offset $seek_method v", %args);
} ## end sub seek

=head2 write

Use native write() to write to an open filehandle.

=over

=item Options

=over

=item C<< filehandle => $filehandle >>

(Required, SCALAR) The filehandle returned by the open() call above.

=item C<< offset => $offset >>

(Optional, SCALAR) The offset to begin writing from.
If no offset is provided, the write will begin from the current position
of the file pointer. The file pointer can be set using seek().

=item C<< num_bytes => $num_bytes >>

(Required, SCALAR) The number of bytes to write, random data will be
automatically generated.

=item C<< nacltask_fail_for_eio => 1|0 >>

(Optional, BOOLEAN) if set to 1, fail for the generic EIO ("Input/Output") error.
    Default = 0  (false: retry a max of "nacltask_retry_limit" times)
    Note: if set to 1, then EIO cannot be part of the retryable list of errors (see below).

=item C<< nacltask_error_retry_list => [ "error-string 1" "error string 2" ] >>

(Optional, ARRAYREF) if a list is provided, retry a max of "nacltask_retry_limit" times.
    Default = throw exception on first occurrence of failure
    Note: will retry by default for EIO ("Input/Output") error unless suppressed.
        if the list contains EIO, cannot set "nacltask_fail_for_eio" to 1 above.

=item C<< nacltask_retry_limit => SCALAR >>

(Optional, SCALAR) maximum allowable retries on error.
    Default = 1 (one more attempt after hitting an error)

=back

=back

=over

=item Returns

Returns the written data if successful,
else throws an exception of type NACL::MTask::Exceptions::FileOperationFailed for failure.
The exception will contain the native error code and error string,
available to catch for negative testing.

=item Examples

    # Write 5 bytes starting at offset 10
    my $data = $node_fileops->write(
        filehandle => $fh,
        offset     => 10,
        num_bytes  => 15,
    );

    # Write from the current file pointer's position
    my $data = $node_fileops->write(
        filehandle => $fh,
        num_bytes  => 20,
    );

=back

=cut

sub write {

    my $self = shift;
    my %args = Params::Validate::validate_with(
        params => \@_,
        spec   => {
            filehandle => {type => SCALAR,},
            offset     => {
                type    => SCALAR,
                default => 'c',      # 'c' = current position of file pointer
            },
            num_bytes  => {type => SCALAR,},
            nacltask_fail_for_eio => {
                type    => SCALAR,
                default => 0,
            },
            nacltask_error_retry_list => {
                type    => ARRAYREF,
                optional=> 1,
            },
            nacltask_retry_limit => {
                type    => SCALAR,
                default => 1,       # retry once after first failure, bail if error persists
            },
            timeout    => {
                type    => SCALAR,
                default => 60,
            },
        },
    );

    # NOTE:
    # 'v' (verbose) is a hidden option to file_test to print the written data

    my $filehandle = delete $args{filehandle};
    my $offset     = delete $args{offset};
    my $num_bytes  = delete $args{num_bytes};

    my $output
        = $self->_run_file_command(command => "write $filehandle $offset $num_bytes v", %args);

    # extract the actual file data
    my $written_data;
    if ($output =~ /data\:\s+\{\s+(.*)\s+___\(EOF\)___/s) {
        $written_data = $1;
    }

    return $written_data;
} ## end sub write

=head2 fsync

Use native fsync() to sync to an open filehandle.

=over

=item Options

=over

=item C<< filehandle => $filehandle >>

(Required, SCALAR) The filehandle returned by the open() call above.

=item C<< nacltask_fail_for_eio => 1|0 >>

(Optional, BOOLEAN) if set to 1, fail for the generic EIO ("Input/Output") error.
    Default = 0  (false: retry a max of "nacltask_retry_limit" times)
    Note: if set to 1, then EIO cannot be part of the retryable list of errors (see below).

=item C<< nacltask_error_retry_list => [ "error-string 1" "error string 2" ] >>

(Optional, ARRAYREF) if a list is provided, retry a max of "nacltask_retry_limit" times.
    Default = throw exception on first occurrence of failure
    Note: will retry by default for EIO ("Input/Output") error unless suppressed.
        if the list contains EIO, cannot set "nacltask_fail_for_eio" to 1 above.

=item C<< nacltask_retry_limit => SCALAR >>

(Optional, SCALAR) maximum allowable retries on error.
    Default = 1 (one more attempt after hitting an error)

=back

=back

=over

=item Returns

Returns nothing for success,
else throws an exception of type NACL::MTask::Exceptions::FileOperationFailed for failure.
The exception will contain the native error code and error string,
available to catch for negative testing.

=item Examples

    # fsync an open filehandle
    $node_fileops->fsync(filehandle => $filehandle);

=back

=cut

sub fsync {

    my $self = shift;
    my %args = Params::Validate::validate_with(
        params => \@_,
        spec   => {
            filehandle => {type => SCALAR,},
            nacltask_fail_for_eio => {
                type    => SCALAR,
                default => 0,
            },
            nacltask_error_retry_list => {
                type    => ARRAYREF,
                optional=> 1,
            },
            nacltask_retry_limit => {
                type    => SCALAR,
                default => 1,       # retry once after first failure, bail if error persists
            },
            timeout => {
                type    => SCALAR,
                default => 60,
            },
        },
    );

    my $filehandle = delete $args{filehandle};

    return $self->_run_file_command(command => "fsync $filehandle", %args);
} ## end sub fsync

=head2 read

Use native read() to write to an open filehandle.

=over

=item Options

=over

=item C<< filehandle => $filehandle >>

(Required, SCALAR) The filehandle returned by the open() call above.

=item C<< offset => $offset >>

(Optional, SCALAR) The offset to begin reading from.
If no offset is provided, the read will begin from the current position
of the file pointer. The file pointer can be set using seek().

=item C<< num_bytes => $bytes_to_read >>

(Required, SCALAR) The number of bytes to read from the file.

=item C<< validate_data => $file_data >>

(Optional, SCALAR) A string containing the data to valiadte against the read data.

=item C<< nacltask_fail_for_eio => 1|0 >>

(Optional, BOOLEAN) if set to 1, fail for the generic EIO ("Input/Output") error.
    Default = 0  (false: retry a max of "nacltask_retry_limit" times)
    Note: if set to 1, then EIO cannot be part of the retryable list of errors (see below).

=item C<< nacltask_error_retry_list => [ "error-string 1" "error string 2" ] >>

(Optional, ARRAYREF) if a list is provided, retry a max of "nacltask_retry_limit" times.
    Default = throw exception on first occurrence of failure
    Note: will retry by default for EIO ("Input/Output") error unless suppressed.
        if the list contains EIO, cannot set "nacltask_fail_for_eio" to 1 above.

=item C<< nacltask_retry_limit => SCALAR >>

(Optional, SCALAR) maximum allowable retries on error.
    Default = 1 (one more attempt after hitting an error)

=back

=back

=over

=item Returns

Returns read data for success,
else throws an exception of type NACL::MTask::Exceptions::FileOperationFailed for failures.
The exception will contain the native error code and error string,
available to catch for negative testing.

=item Examples

    # Read 5 bytes starting at offset 10
    my $read_data = $node_fileops->read(
        filehandle => $filehandle,
        offset     => 10,
        num_bytes  => 5,
    );

    # Read 5 bytes from the current file pointer's position
    # and validate the data
    $node_fileops->read(
        filehandle    => $filehandle,
        offset        => 10,
        num_bytes     => 5,
        validate_data => "ab_12#",
    );

=back

=cut

sub read {

    my $self = shift;
    my %args = Params::Validate::validate_with(
        params => \@_,
        spec   => {
            filehandle => {type => SCALAR,},
            offset     => {
                type    => SCALAR,
                default => 'c',      # 'c' = current position of file pointer
            },
            num_bytes     => {type => SCALAR,},
            validate_data => {
                type     => SCALAR,
                optional => 1,
            },
            nacltask_fail_for_eio => {
                type    => SCALAR,
                default => 0,
            },
            nacltask_error_retry_list => {
                type    => ARRAYREF,
                optional=> 1,
            },
            nacltask_retry_limit => {
                type    => SCALAR,
                default => 1,       # retry once after first failure, bail if error persists
            },
            timeout => {
                type    => SCALAR,
                default => 60,
            },
        },
    );

    my $filehandle      = delete $args{filehandle};
    my $offset          = delete $args{offset};
    my $num_bytes       = delete $args{num_bytes};
    my $validate_data   = delete $args{validate_data};

    # run the 'read' operation
    my $output
        = $self->_run_file_command(command => "read $filehandle $offset $num_bytes", %args);

    # extract the actual file data
    my $read_data;
    if ($output =~ /data\:\s+\{\s+(.*)\s+___\(EOF\)___/s) {
        $read_data = $1;
    }

    # do we need to validate data
    if (defined $validate_data) {
        if ($validate_data eq $read_data) {
            $Log->comment(
                "Read data matches the provided validation data!\n");

        } else {
            throw NACL::MTask::Exceptions::FileOperationFailed(
                "Read data does not match the provided validation data!\n"
                    . "- Expected data: {$validate_data}\n"
                    . "- Read     data: {$read_data}\n");
        }
    }

    return $read_data;
} ## end sub read

=head2 fallocate

Use ONTAP's implementation of fallocate() to punch holes in an open filehandle.

=over

=item Options

=over

=item C<< filehandle => $filehandle >>

(Required, SCALAR) The filehandle returned by the open() call above.

=item C<< offset => $offset >>

(Required, SCALAR) The offset to begin punching holes from.

=item C<< num_bytes => $num_bytes >>

(Required, SCALAR) The number of bytes to punch holes on.

=item C<< FALLOC_FL_KEEP_SIZE|FALLOC_FL_PUNCH_HOLE >>

[Automatically passed to fallocate()]
NOTE: this is NOT an accepted param, but is listed here to tell the
reader that these flags will be passed to fallocate() implicitly.

=item C<< nacltask_fail_for_eio => 1|0 >>

(Optional, BOOLEAN) if set to 1, fail for the generic EIO ("Input/Output") error.
    Default = 0  (false: retry a max of "nacltask_retry_limit" times)
    Note: if set to 1, then EIO cannot be part of the retryable list of errors (see below).

=item C<< nacltask_error_retry_list => [ "error-string 1" "error string 2" ] >>

(Optional, ARRAYREF) if a list is provided, retry a max of "nacltask_retry_limit" times.
    Default = throw exception on first occurrence of failure
    Note: will retry by default for EIO ("Input/Output") error unless suppressed.
        if the list contains EIO, cannot set "nacltask_fail_for_eio" to 1 above.

=item C<< nacltask_retry_limit => SCALAR >>

(Optional, SCALAR) maximum allowable retries on error.
    Default = 1 (one more attempt after hitting an error)

=back

=back

=over

=item Returns

Returns nothing for success,
else throws an exception of type NACL::MTask::Exceptions::FileOperationFailed for failure.
The exception will contain the native error code and error string,
available to catch for negative testing.

=item Examples

    # Punch holes over 15 bytes starting at offset 10
    my $data = $node_fileops->fallocate(
        filehandle => $fh,
        offset     => 10,
        num_bytes  => "15",
    );

=back

=cut

sub fallocate {

    my $self = shift;
    my %args = Params::Validate::validate_with(
        params => \@_,
        spec   => {
            filehandle => {type => SCALAR,},
            offset     => {type => SCALAR,},
            num_bytes  => {type => SCALAR,},
            nacltask_fail_for_eio => {
                type    => SCALAR,
                default => 0,
            },
            nacltask_error_retry_list => {
                type    => ARRAYREF,
                optional=> 1,
            },
            nacltask_retry_limit => {
                type    => SCALAR,
                default => 1,       # retry once after first failure, bail if error persists
            },
            timeout    => {
                type    => SCALAR,
                default => 60,
            },
        },
    );

    my $filehandle  = delete $args{filehandle};
    my $offset      = delete $args{offset};
    my $num_bytes   = delete $args{num_bytes};

    return $self->_run_file_command(
        command => "falloc $filehandle $offset $num_bytes ks ph",
        %args
    );
} ## end sub fallocate

=head2 lock

Use native fcntl() to acquire byte range locks on an open filehandle.

=over

=item Options

=over

=item C<< filehandle => $filehandle >>

(Required, SCALAR) The filehandle returned by the open() call above.

=item C<< offset => $offset >>

(Required, SCALAR) The offset to begin locking from.

=item C<< num_bytes => $num_bytes_to_lock >>

(Required, SCALAR) The number of bytes to lock.

=item C<< type => $lock_method >>

(Required, SCALAR) The lock type to acquire.
    Legal values are one of the following:
        F_WRLCK     (write/exclusive lock)
        F_RDLCK     (read/shared lock)

=item C<< command => $lock_command >>

(Required, SCALAR) The lock command to use.
    Legal values are one of the following:
        F_SETLK     (non-waiting lock - fail immediately for conflicts)
        F_SETLKW    (waiting lock - wait till conflicting lock is releaseed)

=item C<< seek_method => $seek_method >>

(Optional, SCALAR) The method for how to seek the file pointer:
    Legal values are one of the following:
        SEEK_SET
        SEEK_CURRENT
        SEEK_END
    Default = undef

=item C<< nacltask_fail_for_eio => 1|0 >>

(Optional, BOOLEAN) if set to 1, fail for the generic EIO ("Input/Output") error.
    Default = 0  (false: retry a max of "nacltask_retry_limit" times)
    Note: if set to 1, then EIO cannot be part of the retryable list of errors (see below).

=item C<< nacltask_error_retry_list => [ "error-string 1" "error string 2" ] >>

(Optional, ARRAYREF) if a list is provided, retry a max of "nacltask_retry_limit" times.
    Default = throw exception on first occurrence of failure
    Note: will retry by default for EIO ("Input/Output") error unless suppressed.
        if the list contains EIO, cannot set "nacltask_fail_for_eio" to 1 above.

=item C<< nacltask_retry_limit => SCALAR >>

(Optional, SCALAR) maximum allowable retries on error.
    Default = 1 (one more attempt after hitting an error)

=back

=back

=over

=item Returns

Returns nothing for success,
else throws an exception of type NACL::MTask::Exceptions::FileOperationFailed for failure.
The exception will contain the native error code and error string,
available to catch for negative testing.

=item Examples

 # Lock 5 bytes starting at offset 10 for shared access
    $node_fileops->lock(
        filehandle  => $filehandle,
        offset      => 10,
        num_bytes   => 5,
        type        => "F_RDLCK",
        command     => "F_SETLK",
    );

 # Wait to lock 5 bytes at offset 10 for exclusive access,
 #
    $node_fileops->lock(
        filehandle  => $filehandle,
        offset      => 10,
        num_bytes   => 5,
        type        => "F_WRLCK",
        command     => "F_SETLKW",
    );

=back

=cut

sub lock {

    my $self = shift;
    my %args = Params::Validate::validate_with(
        params => \@_,
        spec   => {
            filehandle => {type => SCALAR,},
            offset     => {type => SCALAR,},
            num_bytes  => {type => SCALAR,},
            type       => {
                type  => SCALAR,
                regex => qr/F_WRLCK|F_RDLCK/,
            },
            command => {
                type  => SCALAR,
                regex => qr/F_SETLK|F_SETLKW/,
            },
            seek_method => {
                type    => SCALAR,
                default => undef,
                regex   => qr/SEEK_SET|SEEK_CURRENT|SEEK_END/,
            },
            nacltask_fail_for_eio => {
                type    => SCALAR,
                default => 0,
            },
            nacltask_error_retry_list => {
                type    => ARRAYREF,
                optional=> 1,
            },
            nacltask_retry_limit => {
                type    => SCALAR,
                default => 1,       # retry once after first failure, bail if error persists
            },
            timeout => {
                type    => SCALAR,
                default => 60,
            },
        },
    );

    my $filehandle  = delete $args{filehandle};
    my $offset      = delete $args{offset};
    my $num_bytes   = delete $args{num_bytes};
    my $lk_type     = delete $args{type};
    my $lk_cmd      = delete $args{command};
    my $seek_method = delete $args{seek_method};

    # build the basic 'lock' command
    my $cmd
        = "lock $filehandle $offset $num_bytes "
        . $self->_translate_flags_for_lock(flag => $lk_type) . " "
        . $self->_translate_flags_for_lock(flag => $lk_cmd);

    # append the 'seek_method' flag if requested
    if (defined $seek_method) {
        $cmd .= " "
            . $self->_translate_flags_for_seek(flag => $seek_method);
    }

    return $self->_run_file_command(command => $cmd, %args);
} ## end sub lock

=head2 unlock

Use native fcntl() to unlock byte range locks on an open filehandle.

=over

=item Options

=over

=item C<< filehandle => $filehandle >>

(Required, SCALAR) The filehandle returned by the open() call above.

=item C<< offset => $offset >>

(Required, SCALAR) The offset to begin unlocking from.

=item C<< bytes => $num_bytes_to_lock >>

(Required, SCALAR) The number of bytes to unlock.

=item C<< command => $lock_command >>

(Required, SCALAR) The unlock command to use.
    Legal values are one of the following:
        F_SETLK     (non-waiting unlock - fail immediately for conflicts)
        F_SETLKW    (waiting unlock - wait till conflicting lock is releaseed)

=item C<< seek_method => $seek_method >>

(Optional, SCALAR) The method for how to seek the file pointer:
    Legal values are one of the following:
        SEEK_SET
        SEEK_CURRENT
        SEEK_END
    Default = SEEK_SET

=item C<< nacltask_fail_for_eio => 1|0 >>

(Optional, BOOLEAN) if set to 1, fail for the generic EIO ("Input/Output") error.
    Default = 0  (false: retry a max of "nacltask_retry_limit" times)
    Note: if set to 1, then EIO cannot be part of the retryable list of errors (see below).

=item C<< nacltask_error_retry_list => [ "error-string 1" "error string 2" ] >>

(Optional, ARRAYREF) if a list is provided, retry a max of "nacltask_retry_limit" times.
    Default = throw exception on first occurrence of failure
    Note: will retry by default for EIO ("Input/Output") error unless suppressed.
        if the list contains EIO, cannot set "nacltask_fail_for_eio" to 1 above.

=item C<< nacltask_retry_limit => SCALAR >>

(Optional, SCALAR) maximum allowable retries on error.
    Default = 1 (one more attempt after hitting an error)

=back

=back

=over

=item Returns

Returns nothing for success,
else throws an exception of type NACL::MTask::Exceptions::FileOperationFailed for failure.
The exception will contain the native error code and error string,
available to catch for negative testing.

=item Examples

    # Unlock 5 bytes starting at offset 10
    $node_fileops->unlock(
        filehandle  => $filehandle,
        offset      => 10,
        bytes       => 5,
        command     => "F_SETLK",
    );

    # Wait to unlock 5 bytes at offset 10
    $node_fileops->unlock(
        filehandle  => $filehandle,
        offset      => 10,
        bytes       => 5,
        command     => "F_SETLKW",
    );

=back

=cut

sub unlock {

    my $self = shift;
    my %args = Params::Validate::validate_with(
        params => \@_,
        spec   => {
            filehandle => {type => SCALAR,},
            offset     => {type => SCALAR,},
            num_bytes  => {type => SCALAR,},
            command    => {
                type  => SCALAR,
                regex => qr/F_SETLK|F_SETLKW/,
            },
            seek_method => {
                type    => SCALAR,
                default => undef,
                regex   => qr/SEEK_SET|SEEK_CURRENT|SEEK_END/,
            },
            nacltask_fail_for_eio => {
                type    => SCALAR,
                default => 0,
            },
            nacltask_error_retry_list => {
                type    => ARRAYREF,
                optional=> 1,
            },
            nacltask_retry_limit => {
                type    => SCALAR,
                default => 1,       # retry once after first failure, bail if error persists
            },
            timeout     => {
                type    => SCALAR,
                default => 60,
            },
        },
    );

    my $filehandle  = delete $args{filehandle};
    my $offset      = delete $args{offset};
    my $num_bytes   = delete $args{num_bytes};
    my $lk_cmd      = delete $args{command};
    my $seek_method = delete $args{seek_method};

    # build the basic 'unlock' command
    my $cmd
        = "unlock $filehandle $offset $num_bytes "
        . $self->_translate_flags_for_lock(flag => $lk_cmd);

    # append the 'seek_method' flag if requested
    if (defined $seek_method) {
        $cmd .= " "
            . $self->_translate_flags_for_seek(flag => $seek_method);
    }

    return $self->_run_file_command(command => $cmd, %args);
} ## end sub unlock

###############################################################################
#           SpinVFS operation methods
###############################################################################

=head2 join_av_group

Use setgroups() to join the 'antivirus' group.

=over

=item Options

=over

=item C<< timeout => $timeout_in_seconds >>

(Optional, SCALAR) Override default timeout for command execution.
                    [Default: 60 seconds]

=back

=back

=over

=item Returns

NONE for success,
else throws an exception of type NACL::MTask::Exceptions::FileOperationFailed for failure.
The exception will contain the native error code and error string,
available to catch for negative testing.

=item Examples

    # Join the 'antivirus' group
    my $cwd = $node_fileops->join_av_group();

=back

=cut

sub join_av_group {

    my $self = shift;
    my %args = Params::Validate::validate_with(
        params => \@_,
        spec   => {
            timeout => {
                type    => SCALAR,
                default => 60,
            },
        },
    );

    return $self->_run_file_command(
        command => "setgrp",
        timeout => $args{timeout},
    );
} ## end sub join_av_group

=head2 fh_to_path

Use extattr_get_file() to translate  filehandle to its full path.

=over

=item Options

=over

=item C<< dsid_major => $dsid_major >>

(Required, SCALAR) The DECIMAL or HEX value of the DSID (major) of the volume hosting the file.
NOTE: HEX values MUST start with '0x'

=item C<< fileid => $fileid >>

(Required, SCALAR) The DECIMAL or HEX value of the fileid of the file to look up.
NOTE: HEX values MUST start with '0x'

=item C<< uniq => $uniq >>

(Required, SCALAR) The DECIMAL or HEX value of the uniq generation number in the filehandle.
NOTE: HEX values MUST start with '0x'

=item C<< timeout => $timeout_in_seconds >>

(Optional, SCALAR) Override default timeout for command execution.
                    [Default: 60 seconds]

=item C<< nacltask_fail_for_eio => 1|0 >>

(Optional, BOOLEAN) if set to 1, fail for the generic EIO ("Input/Output") error.
    Default = 0  (false: retry a max of "nacltask_retry_limit" times)
    Note: if set to 1, then EIO cannot be part of the retryable list of errors (see below).

=item C<< nacltask_error_retry_list => [ "error-string 1" "error string 2" ] >>

(Optional, ARRAYREF) if a list is provided, retry a max of "nacltask_retry_limit" times.
    Default = throw exception on first occurrence of failure
    Note: will retry by default for EIO ("Input/Output") error unless suppressed.
        if the list contains EIO, cannot set "nacltask_fail_for_eio" to 1 above.

=item C<< nacltask_retry_limit => SCALAR >>

(Optional, SCALAR) maximum allowable retries on error.
    Default = 1 (one more attempt after hitting an error)

=back

=back

=over

=item Returns

Returns a SCALAR containing the current path for success,
else throws an exception of type NACL::MTask::Exceptions::FileOperationFailed for failure.
The exception will contain the native error code and error string,
available to catch for negative testing.

=item Examples

    # Get current working directory
    my $path = $node_fileops->fh_to_path(dsid_major => 1026, fileid => 96, uniq => 71469);

=back

=cut

sub fh_to_path {

    my $self = shift;
    my %args = Params::Validate::validate_with(
        params => \@_,
        spec   => {
            dsid_major => {
                type => SCALAR,
            },
            fileid => {
                type => SCALAR,
            },
            uniq => {
                type => SCALAR,
            },
            nacltask_fail_for_eio => {
                type    => SCALAR,
                default => 0,
            },
            nacltask_error_retry_list => {
                type    => ARRAYREF,
                optional=> 1,
            },
            nacltask_retry_limit => {
                type    => SCALAR,
                default => 1,       # retry once after first failure, bail if error persists
            },
            timeout => {
                type    => SCALAR,
                default => 60,
            },
        },
    );

    my $dsid_major = delete $args{dsid_major};
    my $fileid     = delete $args{fileid};
    my $uniq       = delete $args{uniq};

    # file_test accepts decimal values only, since the underlying ext_getattr_file()
    # API also uses decimal values... so let's convert HEX to DEC if HEX values
    # were provided; if DEC values were provided, use them as-is
    if ($dsid_major =~ /^0x/) {
        $dsid_major = hex($dsid_major);
        $Log->debug("convert HEX(dsid_major) to DEC: [$dsid_major]");
    }
    if ($fileid =~ /^0x/) {
        $fileid = hex($fileid);
        $Log->debug("convert HEX(fileid) to DEC: [$fileid]");
    }
    if ($uniq =~ /^0x/) {
        $uniq = hex($uniq);
        $Log->debug("convert HEX(dsid_major) to DEC: [$uniq]");
    }

    # run the file_test command and get its output
    my $output = $self->_run_file_command(
        command => "f2p $dsid_major $fileid $uniq",
        %args,
    );
    if (!defined $output or $output eq '') {
        throw NATE::BaseException(
            "INTERNAL_ERROR: empty output or unable to parse the output returned!\n"
        );
    }

    # parse the output and get the returned path
    $_ = ($output =~ /Received\s+path\:\s+\[(\S+)\]/);

    return $1;
} ## end sub getcwd

=head2 reserve_space

Use /usr/sbin/setextattr to enable or disable space reservation.

=over

=item Options

=over

=item C<< path => $file_name >>

(Required, SCALAR) SpinVfs or Clusfs Pathname.

=item C<< reservation => ON|OFF >>

(Optional, SCALAR) Either the string "ON" to enable space reservation or the string "OFF" to disable.
NOTE: Defaults to ON if not specified.

=item C<< reservation_size => $reserve_size >>

(Optional, SCALAR) File Reservation Size.
NOTE: Defaults to 0 which will just enable file reservations.

=back

=back

=over

=item Returns

Returns nothing for success,
else throws an exception of type NACL::MTask::Exceptions::FileOperationFailed for failure.
The exception will contain the error string,

=item Examples

    # Enable File Reservation
    $node_fileops->reserve_space(path => "/clus/vs2/junc/foo", reservation => "ON");

    # Enable File Reservation with filesize of 1024
    $node_fileops->reserve_space(path => "/clus/vs2/junc/foo", reservation_size => 1024);

=back

=cut

sub reserve_space {

    my $self = shift;
    my %args = Params::Validate::validate_with(
        params => \@_,
        spec   => {
            path => {
                type     => SCALAR,
                optional => 0,
            },
            reservation => {
                type     => SCALAR,
                optional => 1,
                default => "on",
            },
            reservation_size => {
                type     => SCALAR,
                optional => 1,
                default => 0,
                regex => qr/^\d+$/
            },
        },
    );

    my $file_name = $args{path};
    my $reserv_size = $args{reservation_size};
    my $reserv = $args{reservation};

    $Log->debug("file name:". $file_name .
                  " Reservation:". $reserv .
		  " Rsize: ".  $reserv_size);

    my $cmd;

    if ( $reserv_size == 0 ) {
        $cmd = "sudo /usr/sbin/setextattr user space_reserve " .  $reserv . " " . $file_name;
    } else {
        $cmd = "sudo /usr/sbin/setextattr user space_reserve " .  $reserv_size . " " . $file_name;
    }

    # run the file_test command and get its output

    my $output = $self->command_interface()->get_systemshell_apiset()->execute_raw_command(command => $cmd);

    my $return = $self->command_interface()->get_systemshell_apiset()->execute_raw_command(command =>'echo $?');

    $Log->debug("return:" . $return);
    throw NACL::MTask::Exceptions::FileOperationFailed("File Space Reservation failed: $output\n")  if ($output =~ /failed:/i or $return !~ /0/);

}

=head2 get_reserve_space_state

Use ngsh command to determine if file space is state.

=over

=item Options

=over

=item C<< ontappath => $ontap_path >>

(Required, SCALAR) Ontap's Pathname /vol/<volname>/filepath.

=item C<< vserver => $vserver_name >>

(Required, SCALAR) Vserver node. For the filer's root volume the vserver name is node name.

=back

=back

=over

=item Returns

Returns "on" for enabled, "off" for disable.

=item Examples

    # Check to see if space reservation is enabled for /clus/vs2/junc/foo
    my $reserv = $node_fileops->get_reserve_space_state(ontappath => "/vol/junc/foo", vserver => "vs2");

    # Check to see if space reservation is enabled for /mroot/nfg/foo
    my $reserv = $node_fileops->get_reserve_space_state(ontappath => "/vol/vol0/nfg/foo", vserver => "node_name");

=back

=cut

sub get_reserve_space_state {

    my $self = shift;
    my %args = Params::Validate::validate_with(
        params => \@_,
        spec   => {
            vserver => {
                type     => SCALAR,
            },
            ontappath => {
                type     => SCALAR,
            },
        },
    );

    my $vserver = $args{vserver};
    my $ontap_file = $args{ontappath};

    my $enabled = NACL::C::VolumeFile->reservation (
        command_interface => $self->command_interface(),
        vserver           => $vserver,
        path              => $ontap_file,
    );

    return $enabled;
}
=head2 get_filesize

Use stat() to obtain the size of a file.

=over

=item Options

=over

=item C<< path => $file_path >>

(Required, SCALAR) The absolute path or relative path of the file on the filer.
Example: "/clus/vs0/vol1/file1", "/mroot/etc/dummyfile1"

=item C<< nacltask_fail_for_eio => 1|0 >>

(Optional, BOOLEAN) if set to 1, fail for the generic EIO ("Input/Output") error.
    Default = 0  (false: retry a max of "nacltask_retry_limit" times)
    Note: if set to 1, then EIO cannot be part of the retryable list of errors (see below).

=item C<< nacltask_error_retry_list => [ "error-string 1" "error string 2" ] >>

(Optional, ARRAYREF) if a list is provided, retry a max of "nacltask_retry_limit" times.
    Default = throw exception on first occurrence of failure
    Note: will retry by default for EIO ("Input/Output") error unless suppressed.
        if the list contains EIO, cannot set "nacltask_fail_for_eio" to 1 above.

=item C<< nacltask_retry_limit => SCALAR >>

(Optional, SCALAR) maximum allowable retries on error.
    Default = 1 (one more attempt after hitting an error)

=back

=back

=over

=item Returns

file size
else throws an exception of type NACL::MTask::Exceptions::FileOperationFailed for failure.
The exception will contain the native error code and error string,
available to catch for negative testing.

=item Examples

    # Get the file size
    my $size;
    $size = $node_fileops->get_filesize(
        path        => "/clus/vs0/vol-1/foo1.txt",
    );

=back

=cut

sub get_filesize {

    my $self = shift;
    my %args = @_;      # let stat() do the params validation
    my %stat_data;

    $self->stat(return_data => \%stat_data, %args);

    return $stat_data{st_size};

} ## end sub stat

=head2 reload_file_interface

In case a node was rebooted, refresh connections, relaunch file ops tool.

=over

=item Options

=over

=item NONE

=back

=back

=over

=item Returns

Nothing if successful, exceptions for failures.

=item Example

 $node_fileops->reload_file_interface();

=back

=cut

sub reload_file_interface {

    my $self = shift;
    my %args = Params::Validate::validate_with(
        params => \@_,
        spec   => {},
    );

    # refresh the node's APISet (primary cmd interface)
    $Log->comment("Refreshing Node connection APISet (command_interface), "
            . "will retry every 2 minutes, 30 times...\n");
    $self->command_interface()->refresh_command_interface(
        'no_version_change'      => 1,
        'max_reconnect'          => 30,
        'max_reconnect_timewait' => 120
    );

    # refresh the node's APISet (file interface)
    $Log->comment("Refreshing Node connection APISet (file interface), "
            . "will retry every 2 minutes, 30 times...\n");
    $self->{file_interface}->refresh_command_interface(
        'no_version_change'      => 1,
        'max_reconnect'          => 30,
        'max_reconnect_timewait' => 120
    );

    # relaunch file_tet
    $Log->comment("\nRelaunching the fileops tool interface...\n");
    $self->_load_bsd_file_tool();

    return;
} ## end sub reload_file_interface


#-----------------------------------------------------------------------------#
# METHOD  : _validate_output
#
# SUMMARY : Validate a file path's existence
#
# INPUTS  : SCALAR (path)
#           SCALAR (nacltask_ignore_failure = 1|0)
#
# RETURNS : NONE on success
#           NACL::MTask::Exceptions::PathValidationFailed for failure
#
#-----------------------------------------------------------------------------#

=head2 verify_file_path

In case a node was rebooted, refresh connections, relaunch file ops tool.

=over

=item Options

=over

=item C<< path => $path_to_file >>

(Required, SCALAR) The path of the file to verify.

=item C<< nacltask_ignore_failure => 0|1 >>

(Optional, SCALAR) (0) = 'fail' if validation fails, (1) = 'ignore' if expected failure.
Default: 0 (fail for validation failures)

=back

=back

=over

=item Returns

Nothing if successful, exceptions for failures.

=item Example

 $node_fileops->verify_file_path(
    path => '/mroot/file1'
 );

=back

=cut

sub verify_file_path {

    my $self = shift;
    my %args = Params::Validate::validate_with(
        params => \@_,
        spec   => {
            path => {type => SCALAR,},
            nacltask_ignore_failure => {
                type    => SCALAR,
                default => 0,        # 0 = no, 1 = yes
            },
        },
    );

    # run 'ls' to list the file
    my $output = $self->command_interface()
        ->get_systemshell_apiset()
        ->execute_raw_command(command => "ls -l $args{path}");

    # validate output
    if ($output =~ /No such file/ and $args{nacltask_ignore_failure} == 0) {
        throw NACL::MTask::Exceptions::PathValidationFailed(
            "File path [$args{path}] does not exist! \n$output\n");
    }

    return;
} ## end sub verify_file_path


###############################################################################
#                 Private helper methods for internal purposes
###############################################################################

#-----------------------------------------------------------------------------#
# METHOD  : _get_freebsd_apiset
#
# SUMMARY : Get a different systemshell APISet for file operations
#
# INPUTS  : NONE
#
# RETURNS : NACL::APISet on success, throws an exception on failure.
#
#-----------------------------------------------------------------------------#

sub _get_freebsd_apiset {

    my $self = shift;

    if (!defined $self->{_freebsd_apiset}) {

        # get a NEW apiset for the node, because we
        # cannot share it with others who wish to run
        # FreeBSD commands in the systemshell
        if (!defined $self->command_interface()) {
            throw NATE::BaseException(
                "The {command_interface} for this instance is undefined!\n"
                    . "Did you forget to run setup() on this library's instance?\n"
            );
        }

        # rediscover this node to create a new SSH connection
        $self->{file_interface} = NACL::STask::Node->find(
            command_interface => $self->command_interface(),
            filter            => {node => $self->command_interface()->node(),}
        );

        # save the node APISet
        $self->{_freebsd_apiset}
            = $self->{file_interface}->get_systemshell_apiset();
    }

    return $self->{_freebsd_apiset};
} ## end sub _get_freebsd_apiset

#-----------------------------------------------------------------------------#
# METHOD  : _get_test_root_from_build
#
# SUMMARY : Get the test root from the current build.
#
# INPUTS  : NONE
#
# RETURNS : SCALAR (path) on success, throws an exception on failure.
#
#-----------------------------------------------------------------------------#

sub _get_test_root_from_build {

    my $self = shift;

    my $test_root =
        $self->command_interface()
            ->get_version_manager()
            ->get_version_attribute(attribute => 'build_path');

    $Log->debug(
        "\nGot test_root [$test_root] from the build.\n");

    if (!defined $test_root) {
        throw NATE::BaseException(
            "Unable to get the test root path for the binary!\n"
        );
    }

    # get the image variant
    my $image_variant =
        $self->command_interface()
            ->get_version_manager()
            ->get_version_attribute(attribute => 'build_options');

    my $image_path = "";
    $image_path = '.debug' if ($image_variant =~ /DEBUG/i);
    if ($image_variant =~ /LGCOV/i) {
       $image_path = $image_path . '.lgcov';
    } elsif ($image_variant =~ /GCOV/i) {
       $image_path = $image_path . '.gcov';
    }

    # complete the path to the binary
    my $src_path = $test_root . "freebsd_vfs/bedrock/export/x86_64" . $image_path . "/uexe/";

    $Log->debug(
        "\nWill get binary using path [$src_path].\n");

    return $src_path;
}

#-----------------------------------------------------------------------------#
# METHOD  : _load_bsd_file_tool
#
# SUMMARY : Start up (if not running) the BSD fileops tool.
#           Generate a uniq ID if not done yet.
#
# INPUTS  : NONE
#
# RETURNS : NONE on success, throws an exception on failure.
#
#-----------------------------------------------------------------------------#

sub _load_bsd_file_tool {

    my $self = shift;

    if (defined $self->{_installed_binary_path}) {

# check if the tool is running on the filer using "ps -wwwaux | grep ... $uniq_id".
# the _uniq_id should be part of the tool's command to distinguish
# between multiple instances.
        my $pid = Os::BSDExecutor::find_process(
            node => $self->{file_interface},
            name => $self->{_installed_binary_path},
        );

        if ($pid == 0) {
            $Log->comment("Binary was not running, starting it up again...");

        } else {

            # nothing to do...
            return;
        }

    } else {

        my $binary_source_path = undef;

        # do we need to get the binary from an alternate path?
        if (defined $self->_alt_binary_source_path()) {
            # yes, override the default source path
            $binary_source_path = $self->_alt_binary_source_path();

        } else {
            # no, get the test_root (default)
            $binary_source_path = $self->_get_test_root_from_build();
        }

        # generate a uniq_id for this installed instance
        $self->{_uniq_id} = NACL::GeneralUtils::random_name_generator(
            prefix => "file_test_uid_",
            size   => 2,
        );

        # install the file_test binary
        $Log->comment("Installing the "
                . BINARY_NAME
                . " binary on node "
                . $self->command_interface()->node());
        $self->{_bsd_install} = NACL::MTask::BSDInstall->install_binaries(
            command_interface         => $self->command_interface(),
            source_path               => $binary_source_path,
            binaries                  => [ BINARY_NAME ],
            nacltask_if_exists        => 'reuse',
            nacltask_verify           => 1,
            timeout                   => 150,   # for Smokes
        );

        # rename the file_test binary to the _uniq_id
        $Log->comment("Renaming the "
                . BINARY_NAME
                . " binary to "
                . $self->{_uniq_id});
        NACL::MTask::BSDInstall->rename_binaries(
            command_interface   => $self->command_interface(),
            rename_mapping      => { BINARY_NAME() => $self->{_uniq_id} },
            nacltask_verify     => 1,
        );

        # update the saved path
        $self->{_installed_binary_path}
            = $self->{_bsd_install}->filer_binary_install_path()
            . "/" . $self->{_uniq_id};

    }    # end of 'if (defined $self->{_installed_binary_path})'

    # start up the tool using the _uniq_id
    my $output = $self->_get_freebsd_apiset()
        ->execute_raw_command(command => "sudo " . $self->{_installed_binary_path});

    # check for its prompt
    if ($output =~ /Version:\s+(\S+)/) {
        my $version = $1;
        if ($version < MIN_FILETEST_VERSION) {
            throw NATE::BaseException(
                "The file_test binary's version [$version] is lower than the "
                    . "minimum required version ["
                    . MIN_FILETEST_VERSION
                    . "]\n");
            $self->{_unsupported_version_file_test_} = $version;
        }

    } else {
        throw NATE::BaseException("Failed to start binary "
                . $self->{_installed_binary_path}
                . "\nDid not match expected prompt [Version: ] in output [$output]\n"
        );
    }

    return;

} ## end sub _load_bsd_file_tool

#-----------------------------------------------------------------------------#
# METHOD  : _translate_flags_for_open
#
# SUMMARY : Translate UNIX-style flags to file_test accepted flags for open()
#
# INPUTS  : SCALAR
#
# RETURNS : SCALAR on success, throws an exception on failure.
#
#-----------------------------------------------------------------------------#

sub _translate_flags_for_open {

    my $self = shift;
    my %args = Params::Validate::validate_with(
        params => \@_,
        spec   => {flags => {type => SCALAR,},},
    );

    my $flags       = $args{flags};
    my $access_mode = "";
    my $open_flags  = undef;
    my @extra_flags = ();

    # split possibly combined flags
    my @flags = split(/\|/, $flags);

    foreach my $flag (@flags) {
        switch ($flag) {

            # <open_flags>

            case 'O_CREAT' {
                $open_flags = 'cn';
            }

            case 'O_APPEND' {
                $open_flags = 'ap';
            }

            case 'O_TRUNC' {
                $open_flags = 'tr';
            }

            case 'O_EXCL' {
                push(@extra_flags, 'excl');
            }

            case 'O_SYNC' {
                push(@extra_flags, 'sync');
            }

            case 'O_DIRECTORY' {
                push(@extra_flags, 'dir');
            }

            # </open_flags>

            # <access_modes>

            case 'O_RDONLY' {
                $access_mode = 'r';
            }

            case 'O_WRONLY' {
                $access_mode = 'w';
            }

            case 'O_RDWR' {
                $access_mode = 'rw';
            }

            # </access_modes>

        default: {
                throw NATE::BaseException(
                    "Unknown open flag '$flag' provided!\n");
            }

        }    # end of 'switch'

    }    # end of 'foreach @flags'

    # default open_flags if none were specified
    if (!defined $open_flags) {
        $open_flags = 'oe';
    }

    # combine all flags into a single string for file_test
    my $flags_str = "$open_flags $access_mode";
    if (scalar @extra_flags) {
        $flags_str .= ' ' . join(" ", @extra_flags);
    }

    return $flags_str;
} ## end sub _translate_flags_for_open

#-----------------------------------------------------------------------------#
# METHOD  : _translate_flags_for_seek
#
# SUMMARY : Translate UNIX-style flags to file_test accepted flags for seek ops
#
# INPUTS  : SCALAR
#
# RETURNS : SCALAR on success, throws an exception on failure.
#
#-----------------------------------------------------------------------------#

sub _translate_flags_for_seek {

    my $self = shift;
    my %args = Params::Validate::validate_with(
        params => \@_,
        spec   => {flag => {type => SCALAR,},},
    );

    my $flag_str = undef;

    switch ($args{flag}) {

        case 'SEEK_SET' {
            $flag_str = 's';
        }

        case 'SEEK_CURRENT' {
            $flag_str = 'c';
        }

        case 'SEEK_END' {
            $flag_str = 'e';
        }

    default: {
            throw NATE::BaseException(
                "Unknown seek flag '$args{flag}' provided!\n");
        }

    }    # end of 'switch'

    return $flag_str;
} ## end sub _translate_flags_for_seek

#-----------------------------------------------------------------------------#
# METHOD  : _translate_flags_for_lock
#
# SUMMARY : Translate UNIX-style flags to file_test accepted flags for lock ops
#
# INPUTS  : SCALAR
#
# RETURNS : SCALAR on success, throws an exception on failure.
#
#-----------------------------------------------------------------------------#

sub _translate_flags_for_lock {

    my $self = shift;
    my %args = Params::Validate::validate_with(
        params => \@_,
        spec   => {flag => {type => SCALAR,},},
    );

    my $flag_str = undef;

    switch ($args{flag}) {

        case 'F_WRLCK' {
            $flag_str = 'w';
        }

        case 'F_RDLCK' {
            $flag_str = 'r';
        }

        case 'F_SETLK' {
            $flag_str = 'n';
        }

        case 'F_SETLKW' {
            $flag_str = 'w';
        }

    default: {
            throw NATE::BaseException(
                "Unknown lock flag '$args{flag}' provided!\n");
        }

    }    # end of 'switch'

    return $flag_str;
} ## end sub _translate_flags_for_lock

#-----------------------------------------------------------------------------#
# METHOD  : _parse_file_handle
#
# SUMMARY : Parse file_test's output to extract the filehandle
#
# INPUTS  : SCALAR (output)
#
# RETURNS : SCALAR on success, undef for failure.
#
#-----------------------------------------------------------------------------#

sub _parse_file_handle {

    my $self = shift;
    my %args = Params::Validate::validate_with(
        params => \@_,
        spec   => {output => {type => SCALAR,},},
    );

    my $output = $args{output};
    my @lines  = split(/\n/, $output);
    my $fh     = undef;

    foreach my $line (@lines) {
        if ($line =~ /Successfully\s+opened\s+file\s+handle\s+(\d+)/i) {
            $fh = $1;
            last;
        } else {
            next;
        }
    }    # end of 'foreach' loop

    return $fh;
} ## end sub _parse_file_handle

#-----------------------------------------------------------------------------#
# METHOD  : _parse_dir_handle
#
# SUMMARY : Parse file_test's output to extract the directory handle
#
# INPUTS  : SCALAR (output)
#
# RETURNS : SCALAR on success, undef for failure.
#
#-----------------------------------------------------------------------------#

sub _parse_dir_handle {

    my $self = shift;
    my %args = Params::Validate::validate_with(
        params => \@_,
        spec   => {output => {type => SCALAR,},},
    );

    my $output = $args{output};
    my @lines  = split(/\n/, $output);
    my $dh     = undef;

    foreach my $line (@lines) {
        if ($line =~ /Successfully\s+opened\s+diretory\s+handle:\s+(\d+)/i) {
            $dh = $1;
            last;
        } else {
            next;
        }
    }    # end of 'foreach' loop

    return $dh;
} ## end sub _parse_dir_handle

#-----------------------------------------------------------------------------#
# METHOD  : _run_file_command
#
# SUMMARY : Run file_test command, check for errors
#
# INPUTS  :
#           SCALAR (command) (Required)
#           SCALAR (timeout) (Optional - default = '30' seconds)
#
# RETURNS : SCALAR (output) on success,
#           NACL::MTask::Exceptions::FileOperationFailed for failure.
#
#-----------------------------------------------------------------------------#

sub _run_file_command {

    my $self = shift;
    my %args = Params::Validate::validate_with(
        params => \@_,
        spec   => {
            command       => {type => SCALAR,},
            match_output  => {
                type    => SCALAR,
                default => 'Successfully',
            },
            timeout       => {
                type    => SCALAR,
                default => 60,
            },
            nacltask_fail_for_eio => {
                type    => SCALAR,
                default => 0,
            },
            nacltask_error_retry_list => {
                type    => ARRAYREF,
                optional=> 1,
            },
            nacltask_retry_limit => {
                type    => SCALAR,
                default => 1,       # retry once after first failure, bail if error persists
            },
            __unit_test_arg_count_attempts => {
                type     => SCALARREF,
                optional => 1,
            },
        },
    );

    # grab unit test args (internal use only)
    my $_ut_arg_count_attempt_ref = (defined $args{__unit_test_arg_count_attempts}) ?
            delete $args{__unit_test_arg_count_attempts} : undef;

    # check for nacltask_* usage errors first
    my %retry_args;
    $self->_validate_retry_args(ref_retry_args => \%retry_args, %args);

    # clear all nacltask_* args from %args
    delete @{\%args} { grep /nacltask_/, keys %args };

    # do we have the minimum required version?
    if (defined $self->{_unsupported_version_file_test_}) {
        throw NATE::BaseException("The file_test binary's version ["
                . $self->{_unsupported_version_file_test_}
                . "] is lower than the "
                . "minimum required version ["
                . MIN_FILETEST_VERSION
                . "]\n");
    }

    # extract additional args not recognized by NACL
    my $match_output = delete $args{match_output};

    my $output;
    my $exception;
    my $attempt = 1;

    # master do-while loop to handle command execution and retries
    do {
        # if this is a retry after the first failure...
        if ($attempt > 1) {
            # this is a retry...
            $Log->comment("\n...retrying operation (retry attempt # " .
                          ($attempt - 1) . ")\n");

            # reset the exception and output
            $output    = undef;
            $exception = undef;
        }

        # controlled execution of the fileop
        try {
            # run the command
            $output
                = $self->_get_freebsd_apiset()->execute_raw_command(%args);

            # validate the output
            $self->_validate_output(
                output       => $output,
                match_output => $match_output);

        } catch NACL::MTask::Exceptions::FileOperationFailed with {
            # -----
            # NOTE:
            # -----
            # we will NOT catch NACL::MTask::Exceptions::PathValidationFailed
            # will retry for failed file operations, NOT for post-op validations
            $exception = shift;

            $Log->debug("\ncaught exception:\n" . $exception->text());
            $Log->debug("\nretryable errors:\n" . Dumper($retry_args{nacltask_error_retry_list}));

            if (grep { $exception->text() =~ $_ } @{$retry_args{nacltask_error_retry_list}}) {
                # acceptable error, retry (if attempts are available)
                $Log->comment("Caught a retryable exception:\n" . $exception->text());

            } else {
                $exception->throw();
            }

        };  # end of catch block

        # on to the next attempt...
        $attempt++;

    } while (($attempt <= $retry_args{nacltask_retry_limit} + 1) and defined $exception);

    # UNIT TEST ONLY
    # populate the SCALAR ref provided by the caller, with the value of
    # how many attempts were made for this operation
    if (defined $_ut_arg_count_attempt_ref) {
        $$_ut_arg_count_attempt_ref = ($attempt - 1);   # -1 to take off the loop breaking increment
    }

    # did the command succeed?
    if (defined $exception) {
        # if we reached here, we can assume
        # - we have exhausted all retry attempts
        # - the command is still failing
        # - we have not hit any other exception that would have been thrown already
        $Log->comment("\nExhausted all retry attempts, operation is still failing... " .
                      "rethrowing exception...\n\n");
        $exception->throw();
    }
    

    return $output;
} ## end sub _run_file_command

#-----------------------------------------------------------------------------#
# METHOD  : _validate_retry_args
#
# SUMMARY : Validate retry related args set by the caller.
#
# INPUTS  : nacltask_* args for retry logic
#           HASHREF to return just the retry related args
#
# RETURNS : NONE on success
#           NACL::MTask::Exceptions::IncorrectParameterUsage for failure
#
#-----------------------------------------------------------------------------#

sub _validate_retry_args {

    my $self = shift;
    my %args = Params::Validate::validate_with(
        params => \@_,
        spec   => {
            nacltask_fail_for_eio => {
                type    => SCALAR,
                default => 0,
            },
            nacltask_error_retry_list => {
                type    => ARRAYREF,
                default => [ ERR_EIO ],
            },
            nacltask_retry_limit => {
                type    => SCALAR,
                default => 1,       # retry once after first failure, bail if error persists
            },
            ref_retry_args => {
                type    => HASHREF,
            },
        },
        allow_extra => 1,   # we'll ignore the non-NACL args
    );

    # check if caller asked to fail for EIO,
    # but also requested retries for it
    if ($args{nacltask_fail_for_eio} == 1 and
        grep { $_ =~ ERR_EIO } @{$args{nacltask_error_retry_list}}) {

        throw NACL::MTask::Exceptions::IncorrectParameterUsage(
            "Usage error: conflicting values to arguments:\n" .
            "Requested failure for EIO via (nacltask_fail_for_eio => 1), " .
            "but nacltask_error_retry_list contains EIO in retryable errors' list.\n"
        );
    }

    # do we have a good value for max retry limit?
    if ($args{nacltask_retry_limit} < 0) {
        throw NACL::MTask::Exceptions::IncorrectParameterUsage(
            "Usage error: invalid value [$args{nacltask_retry_limit}] provided " .
            "for optional argument 'nacltask_retry_limit'. Minimum = 0 (no retries).\n"
        );
    }

    # all checks out, copy off all retry related nacltask args to the args
    my $ref_retry_args = delete $args{ref_retry_args};
    foreach my $key (keys %args) {
        $ref_retry_args->{$key} = $args{$key}
                if ($key =~ /nacltask_/);
    }

    return;
} ## end sub _validate_output


#-----------------------------------------------------------------------------#
# METHOD  : _validate_output
#
# SUMMARY : Parse file_test's output to check if the command succeeded or not
#
# INPUTS  : SCALAR (output)
#
# RETURNS : NONE on success, NACL::MTask::Exceptions::FileOperationFailed for failure.
#
#-----------------------------------------------------------------------------#

sub _validate_output {

    my $self = shift;
    my %args = Params::Validate::validate_with(
        params => \@_,
        spec   => {
            output => {
                type => SCALAR,
            },
            match_output  => {
                type => SCALAR,
            },
        },
    );

    if ($args{output} !~ /$args{match_output}/) {
        throw NACL::MTask::Exceptions::FileOperationFailed(
            "File operation failed with: \n$args{output}\n");
    }

    return;
} ## end sub _validate_output


1;
