# $Id$

## @summary CIFS Symlink
## @author dl-nacl-dev@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::MTask::CIFSSymlink

=head1 DESCRIPTION

C<NACL::MTask::CIFSSymlink> provides a number of well-defined but potentially
complex or multi-step methods related to CIFSSymlink in ONTAP. The
functionality provided mainly involves the components
C<NACL::C::VserverCifsSymlink> and C<NACL::C::VserverCifsShare> (since the
current implementation covers only CMode). See L<NACL::C::VserverCifsSymlink>
and L<NACL::C::VserverCifsShare> for details. The MTask will mainly be
concerned with creation of symlink in the share by mounting the corresponding
junction path in a linux box and also creates the symlink entry for the local
symlink or widelink on a given file/directory on  the filer.

=head1 ATTRIBUTES

=head2 link

The symlink created on the share from a linux mount.

=head2 linkdir_created

The directory created by the task inside the share and the link file will be present inside this directory.

=head2 symlinkentry

The symlink entry created in the Filer side.

=head2 mnt_path

The mount path in the linux box.

=head2 mountdirectory_created

The directory created by the task when the complete/part of the mount path does not exist.

=head1 METHODS

=head2 setup

    my $symlink_obj = NACL::MTask::CIFSSymlink->setup(
        cifsshare  => $cifsshare_obj,
        link       => $link,
        mount_path => $mount_path,
        locality   => $locality,
        target     => $target,
        unix-path  => $unix_path,
        cifs-path  => $cifs_path,
        client => $client,
        junction-path => $junction_path,
        destination-sharename => $dest_sharename,
        destination-server => $dest_server,
        nacltask_if_exists  => $action, # default 'die'
        %other_options
    );

(Class method) Given a unix path, this task creates local symlink or widelink
to it. This includes the creation of the link file on the Filer mount and the
Symlink Table Entry on the Filer.

=over 

=item Options

=over

=item C<< nacltask_if_exists=>$action >>

(Optional) Specify what needs to be done if the link and/or the symlink table
entry (in the filer) already exists.
    If $action is "die" (the default), then an execption is raised.
    If $action is "purge", then purge the existing link/symlink table entry 
    and continue.
Default: "die"

=item C<< cifsshare  => $cifsshare_obj >>

(Required) A C<NACL::C::VserverCifsShare> instance.

=item C<< locality   => $locality >>

(Optional) A string indicating the type of link to be created.
Default: local

=item C<< target     => $target >>

(Required) The path of the file/directory to which the link has to be created.
This is a "path" string that the symlink should point to( target in the ln command). No check is done to
verify if this target file/directory exists or not.

=item C<< link       => $link >>

(Required) The link file name or the path of the link to be created relative to mount point. If the relative path
does not exist the task will create it and then create the link inside it.

=item C<< unix-path  => $unix_path >>

(Optional) The unix path prefix to be matched for the mapping. This should be the unix-path for the symlink entry 
in the filer
Default: $mount_path/

=item C<< cifs-path  => $cifs_path >>

(Optional) The cifs path on the destination to which the link is mapped(this will be the cifs-path entry in the
symlink table on the filer). This is mandatory if $unix_path is provided.
Default: /

=item C<< client => $client >>

(Required) The unix client which is used to mount the junction path of the share and create the link(The tharnhost
file name of the client is given as input).

=item C<< mount_path => $mount_path >>

(Required) The temporary mount path in the linux box used for mounting the junction path and creation of the link

=item C<< junction-path => $junction_path >>

(Optional) The junction path used for mounting on the linux box, This parameter is required to be passed only when 
the share is of Homedirectory type. 

=item C<< destination-sharename => $dest_sharename >>

(Optional) The destination share name to be passed in the symlink entry when the locality is of type widelink,
This parameter is required when the locality is widelink.

=item C<< destination-server => $dest_server >>

(Optional) The destination server to be passed in the symlink entry when the locality is of type widelink,
This parameter is required when the locality is widelink.

=back

=item command_interface, apiset_must, apiset_should, method-timeout, all the
other options supported by NACL::C::CIFSSymlink->create is supported by
NACL::MTask::CIFSSymlink->setup as well.

=back

=head2 cleanup

    $symlink_obj->cleanup();

Delete the link,the entry in the symlink table and the directories created if any for the
given symlink object.

=over 

=item Options

    None

=back

=cut

# this package is
package NACL::MTask::CIFSSymlink;

# this package is a
use base qw/NACL::MTask::MTask/;

# we need these libs
use NACL::C::NetworkInterface;
use NACL::C::VserverCifsShare;                      #for create
use NACL::C::VserverCifsSymlink;
use NACL::C::Client;
use NACL::C::Volume;
use NACL::STask::Client::Mount;
use NACL::APISet;                                   # for systemshell apiset
use NACL::APISet::Exceptions::CommandFailedException qw ( :try);
use Params::Validate qw/validate_with SCALAR ARRAYREF/;
use Tharn;
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
#utils
use strict;
use warnings;
use Data::Dumper;


###########################################################################################################
# Name        : setup
# Description : Actual task function to be invoked by test scripts. POD above has complete details
#               Creates links, Adds entry in the Symlink table as per options
###########################################################################################################


sub setup {
    # obtain the package name
    my $pkg = shift;
    NATE::BaseException->throw(
        "Setup can only be invoked with a proper package call")
        if ref $pkg;
    my $obj;
	my %optwide;
	
	#get the mount_path value and assign as the default value for the unix path
	my %para = @_;
	my $default_unixpath = $para{mount_path}."/";
	
	#if the Unix path is not mentioned then the mount path is taken as the default unix path.
	unless(defined($default_unixpath)){
		$default_unixpath = "/tmp/cifssymlink/"; 
	}	
		
	my %opts = validate_with(
        params => \@_,
        spec   => {
			 cifsshare => { isa  => 'NACL::C::VserverCifsShare' },
			 link => {type => SCALAR},
			 target => {type => SCALAR},
			 client  => {type => SCALAR},
			 locality => {
					type      => SCALAR,
					optional => 1,
					default	  => 'local'
					},
			'unix-path' => {
					type      => SCALAR,
					optional => 1,
					default => $default_unixpath,
					depends  => ['cifs-path'],
					},
			'cifs-path' => {
					type      => SCALAR,
					optional => 1,
					default	  => '/'
					},
			'junction-path' => {
					type      => SCALAR,
					optional => 1,
					},
			'destination-sharename' => {
					type      => SCALAR,
					optional => 1,
					depends  => ['destination-server'],
					},		
			'destination-server' => {
					type      => SCALAR,
					optional => 1,
					},		
			mount_path => {type      => SCALAR},		
			nacltask_if_exists => {
                type      => SCALAR,
                default   => 'die',
                callbacks => {
                    'Any of "die" or "purge"' => sub {
                        if (   $_[0] eq 'die'
                             || $_[0] eq 'purge' )
                        {
                            return 1;
                        }
                        return 0;
                        }
					}
                }
            }
			
	);

	# extract the parameters	
	my $cifs_shareobj = $opts{cifsshare};
	my $link 		 = $opts{link};
	my $mount_path  = $opts{mount_path};
	my $locality    = $opts{locality};
	my $target      = $opts{target};
	my $unix_path   = $opts{'unix-path'};
	my $cifs_path   = $opts{'cifs-path'};
	my $client	 = $opts{client};
	my $if_exists = $opts{nacltask_if_exists};
	my $destination_sharename = $opts{'destination-sharename'};
	my $destination_server = $opts{'destination-server'};
	
	#newly added
	my $volobj;

	# get the command interface for the other operations
	my $command_interface = $cifs_shareobj->command_interface();
		NATE::BaseException->throw("Only CMode is supported as of now")
			if $command_interface->mode() ne 'CMode';

	# Get the Host name and the Host type and create the API set object

     my $client_obj = NACL::C::Client->new( name => $client );
     my $APISet_obj = $client_obj->apiset();

    $obj->{apiset} = $APISet_obj;
    #get the vserver name from the share
    my $vserver_name = $cifs_shareobj->get_one_state_attribute("vserver");

    #get the share name and junction path from the share object
    my $share_name = $cifs_shareobj->state()->share_name();
    my $junction   = $cifs_shareobj->state()->path();
    if ( $junction =~ /%w|%u|%d/ ) {
        $junction = $opts{'junction-path'};
    }

    #newly added for IPV6
    $volobj = NACL::C::Volume->find(
        command_interface => $command_interface,
        filter            => {
            'junction-path' => $junction,
            vserver         => $vserver_name
        }
    );
	$optwide{'share-name'} = $share_name;

	#get the share properties to check whether its a home directory
	my @share_properties = @{$cifs_shareobj->state()->share_properties()};

	my $home_directory = "false";
	foreach(@share_properties){
		if($_ =~ /homedirectory/){
			$home_directory = "true";
			$junction = $opts{'junction-path'};
			NACL::APISet::Exceptions::MissingRequiredParamException->throw(
				"Must specify the Junction path as parameter for mounting when the Share is of type Homedirectory")
					if(!defined $junction);
			last;		
		}
	}
	
	if($locality =~ /widelink/i){
		NACL::APISet::Exceptions::MissingRequiredParamException->throw(
				"Must specify the destination-sharename and destination-server as parameter when the locality is widelink") 
					if((!defined $destination_sharename) || (!defined $destination_server));
			$optwide{'share-name'} = $destination_sharename;
			$optwide{'cifs-server'} = $destination_server; 
			
	}
	
	# create the entry in the symlink table.
	# make the parameter list for the symlink entry creation
	my %component_call_opts = (
			%opts,
			%optwide,
			command_interface => $command_interface,
			   vserver           => $vserver_name,
			   'unix-path'       => $unix_path,
			   'cifs-path'       => $cifs_path,
			   'locality'		 => $locality,
			   'home-directory'  => $home_directory
		);

	# create a symlink entry.	
	$obj->{symlinkentry} = _handle_creation(
			%component_call_opts,
			component     => 'NACL::C::VserverCifsSymlink',
			method        => 'create',
			delete_method => 'delete',

		);

	# on the client side mount and create the environment
	#get the Ip address
	my $lif = NACL::C::NetworkInterface->find(command_interface => $command_interface, filter => { vserver => $vserver_name});
	my $IP_Address = $lif->state()->address();

	#extract the link name from the relative path
	my @linkarr = split("/",$link);
	my $linkname = pop @linkarr;
	my $linkpath = $mount_path."/".join("/",@linkarr);

	#check if mount path exist, if not create it and add to the class attribute
	my $response ;
	try{
		$obj->{mountdirectory_created}	= _handle_dircreation($APISet_obj,$mount_path);
				
		#cd to the / path in the unix box and mount the junction path 
		#my $response = $APISet_obj->execute_raw_command('command'=>'cd /');
		$response = $APISet_obj->cd('paths' => '/');
		my $source_path = $IP_Address.":".$junction;

		# # # # # # # # # # # # # # $response = $APISet_obj->mount('source-path'=>$source_path, 
												   # # # # # # # # # # # # # # 'mount-point'=>$mount_path);

		# added new to support IPV6

		my $lnxmnt3obj = NACL::STask::Client::Mount->create(
											command_interface => $client_obj,
											volume => $volobj,
											'mount_point' => $mount_path ,
											nacltask_if_exists => "purge",
											nacltask_verify    => 1
											);
											
		$obj->{mnt_obj} = $lnxmnt3obj;
		#end if new	
												   
		$obj->{mnt_path} = $mount_path;
		$obj->{linkdir_created} = _handle_dircreation($APISet_obj,$linkpath);
		
		#move to the subdir where the link has to be created
		$response = $APISet_obj->cd('paths' => $linkpath);

	}otherwise {
        my $exception_obj = shift;
		cleanup($obj);
		$exception_obj->throw();
    };
	
	try{	
	#need to check if the link exist then needs to delete or die
	#create the link using the ln command 
		$response = $APISet_obj->ln('symlink' => 1,
									'link-name' => $linkname,
									'files' => $target
									);
		$obj->{link} = 	$linkpath."/".$linkname;						
	}otherwise {
        my $exception_obj = shift;
	
		#if the file exist then check whether it needs to be deleted and continued.
		if(($exception_obj =~ /File exists/i) && ($if_exists eq 'purge')){
			try{	
				$APISet_obj->rm( 'force' => 1, 'recursive' => 1, paths => $linkname);
				$response = $APISet_obj->ln('symlink' => 1,
									'link-name' => $linkname,
									'files' => $target
									);
			}otherwise{
				my $e = shift;
				cleanup($obj);
				$e->throw();
			}
		}else{
			$exception_obj->throw();
		}
       
    };
	
	bless $obj, $pkg;
	return $obj;
}

###########################################################################################################
# Name        : _handle_creation
# Description : Invoked from setup. If there are duplicate entries already existing on the filer, based on
#               if_exists option which might be to die or purge, appropriate action is taken. If e.g.
#               if_exists => purge, It will call purge action which deletes the entries and recreates
#               the entries. If that configuration is not there, it simply creates the entry. Returns an
#				object of the requested component type only if this method creates one so that this can be
#				stored as an attribute for deletion during cleanup.
###########################################################################################################

sub _handle_creation {
    my %opts = @_;

    # get the name of the component and the name of the method required for
    # create/add
    my $pkg    = $opts{component};
    my $method = $opts{method};

    # extract the command interface.
    my $command_interface = $opts{command_interface};
    my ( $object, $is_created );

    my %actions = (
        "die" => {
            check  => 0,
            exists => undef,
        },
        "purge" => {
            check  => 1,
            exists => sub {

                _purge(%opts);

                # Forget that the component had already
                # been found, to force create step
                undef $object;
            },
        },
    );

    # get the hashref that is the action for the nacltask_if_exists value
    my $action = $actions{ $opts{nacltask_if_exists} };

    # if a check has to be done for the selected action
    if ( $action->{check} ) {

        # build the primary key hash for this element from all the options
        # that we have
        my $pk_hash = $pkg->get_primary_keys_options(%opts);

        # check if an element with the same primary key values as what we want
        # already exists
        $object = $pkg->find(
            command_interface => $command_interface,
            filter            => $pk_hash,
            allow_empty       => 1
        );

        # if $object is not defined, then we didn't find a matching element
    }

    # if an object exists, then we either use the same object or delete it
    # depending on what the 'exists' coderef does for us

    if ($object) {
        &{ $action->{exists} };
    }

    # at this point, $object may not be defined (if it was either deleted or
    # not found. so create it if its not defined

    if ( !$object ) {

       # call the approriate method (create / add ) on the appropriate package
       # with all the opts sent to this function. allow_extra is for making
       # the component method ignore the irrelevant opts

        $object = $pkg->$method(
            %opts,
            command_interface => $command_interface,
            allow_extra       => 1
        );
        $is_created = 1;
    }
    return $object if $is_created;
    return undef;
}

###########################################################################################################
# Name        : _purge
# Description : Invoked from _handle_creation. It takes the appropriate delete method to be used and the
#               component that has requested for purge, frames the method with appropriate parameters
#               to be called and invokes the same.
###########################################################################################################

sub _purge {
    my %opts   = @_;
    my $pkg    = $opts{component};
    my $method = $opts{delete_method};

    my $command_interface = $opts{command_interface};

    # build the primary key hash for this element from all the options
    # that we have
    my $pk_hash = $pkg->get_primary_keys_options(%opts);

    # call either delete/remove on the package provided
    $pkg->$method( command_interface => $command_interface, %$pk_hash );
}


###########################################################################################################
# Name        : _handle_dircreation
# Description : Invoked from setup. make the directories based on the mount path or the link if it does not 
#				exist. If this method creates the directories this can be stored as an attribute for 
#				deletion during cleanup.
###########################################################################################################

sub _handle_dircreation {
    my ($APISet_obj,$dirpath) = @_; 
	my @arr = split("/",$dirpath);
	my $path ="/";
	my $flag = 0;
	
	#remove the space from the array which will be added from the split
	shift @arr if ($arr[0] eq "");
		
	foreach (@arr)
	{
		next if($_ eq "/");
		$path .= $_;

# check whether the directory exist or not 
		#my $response = $APISet_obj->ls(paths =>$path); 
		my $response = $APISet_obj->execute_raw_command(command =>"ls $path"); 
		print Dumper $response;	
		if ($response =~ /No such file or directory/)
		{
			$flag = 1;	
			last;
		}
			$path .= "/";
	}
	
	if ($flag == 1){
		#mkdir recursively for $dirpath and return the path from where its created
		my $response = $APISet_obj->mkdir(	'make-parents' => 1,	
											'paths' => $dirpath );
		return $path;
	}else{
		return undef;
	}
}

###########################################################################################################
# Name        : cleanup
# Description : Invoked with an object that was returned from setup
#               destroys everything that was created during setup
###########################################################################################################

sub cleanup {
    my $obj = shift;
	
	try{
		#remove the link file created
		$obj->{apiset}->cd('paths' => " /");
		
		$obj->{apiset}
            ->rm( 'force' => 1, 'recursive' => 1, paths => $obj->{link} )
            if $obj->{link};	
		
		#remove the link directories created	
		$obj->{apiset}
            ->rm( 'force' => 1, 'recursive' => 1, paths => $obj->{linkdir_created} )
			if $obj->{linkdir_created};
		
		#delete the symlink entry from the filer
		$obj->{symlinkentry}->delete()  if $obj->{symlinkentry};
		
		#unmount the path
		# # if($obj->{mnt_path}){
					# # $obj->{apiset}->cd('paths' => '/');
					# # $obj->{apiset}->umount('paths'=> $obj->{mnt_path});
					
		$obj->{mnt_obj}->purge() if($obj->{mnt_obj});			
		
		$obj->{apiset}
            ->rm( 'force' => 1, 'recursive' => 1, paths => $obj->{mountdirectory_created} )
            if $obj->{mountdirectory_created};	
		
	}
	otherwise {
       my $exception_obj = shift;
	   $exception_obj->throw();
    };  	
	
}

1;
