# Copyright (c) 2015 NetApp, Inc., All Rights Reserved.
# Any use, modification, or distribution is prohibited
# without prior written consent from NetApp, Inc.
# @summary Create partitions by setting net access mask on the nodes MTask module
# @author bramya@netapp.com, dl-ptefit-automation-core@netapp.com
# @status shared
# @pod here

=head1 NAME

NATE::MTask::CtranPartitioning

=head1 DESCRIPTION

C<NATE::MTask::CtranPartitioning> provides methods to calculate net access mask for the partition with any number of
nodes, set the net access mask on each node, create and destroy partitions.

=head1 METHODS

=head2 new

    # Example for creating CtranPartitioning object
    my $CtranPartitioning_obj = NACL::MTask::CtranPartitioning->new(
                                    command_interface => $cmd_int,
                                );

This method will return a CtranPartitioning object to allow method calls for defined methods. This
method also discovers the cluster size (by interacting with the given command interface) and also
creates a hash of the node names and their node ids.

=over

=item Arguments

=over

=item C<< command_interface => $cmd_int >>

(Required) An instance of a subclass of NACL::C::CommandInterface, which represents what element
should receive the commands to influence this element.

=back

=item Return values

A C<NACL::MTask::CtranPartitioning> object.

=back

=head2 calculate_mask_for_partition

This method will calculate the mask that needs to be set on the nodes in order to form a partition.

=over

=item Arguments

=over

=item C<< nodes => \@node_names >>

(Required) An ARRAYREF of the node names which would be in the partition.

=back

=item Return values

A net access mask that needs to be set on the nodes.

=back

=head2 create_partitions

This method will accept an ARRAYREF of ARRAYREFs of partitions where the group of nodes in a partition will only respond to
CLAM heartbeats from each other. This method will create partitions in parallel i.e., all nodes will have their masks set simultaneously by default.

=over

=item Arguments

=over

=item C<< partitions => \@partitions >>

(Required) An ARRAYREF of ARRAYREFs of partitions. Each partition contains the node names which would be in the partition.
Example: [["node1","node2"],["node3","node4"]]

=item C<< runid_prefix => $runid_prefix >>

(Optional) A scalar value that helps in identifying logs in case of multiple calls of create_partitions().

=item C<< auto_partition_remaining_nodes => 0|1 >>

(Optional) This parameter specifies whether to create a partition of all the nodes in the cluster that aren't part of any partition. Default value is 0.

=back

=item Return values

A HASHREF of the nodenames, the respective masks and the time at which the masks were set on the nodes.
Example:
    {
        $node1 => {time => ..., mask => ...},
        $node2 => {time => ..., mask => ...},
    }

=back

=head2 reset_partitions

This method will reset the partitions i.e., the net access mask on each node provided in the cluster is reset.

=over

=item Arguments

=over

=item C<< nodes => \@node_names >>

(Optional) An ARRAYREF of the node names in the cluster. The default value will be all nodes in the cluster.

=item C<< runid_prefix => $runid_prefix >>

(Optional) A scalar value that helps in identifying logs in case of multiple calls of reset_partitions().

=back

=item Return values

None

=back

=head2 set_mask_on_node

(Static Method) This method will set the net access mask on the node.

=over

=item Arguments

=over

=item C<< node => $node_name >>

(Required) The node name on which the net access mask needs to be set.

=item C<< mask => $mask >>

(Required) The net access mask to be set on the node.

=back

=item Return values

None

=back

=head2 unset_mask_on_node

(Static Method) This method will reset the net access mask on the node.

=over

=item Arguments

=over

=item C<< node => $node_name >>

(Required) The node name on which the net access mask needs to be set.

=back

=item Return values

None

=back

=head2 get_majority_partition_nodes

This method will return a ARRAYREF of the majority partition nodes.

=item C<< node => $node_name >>

(Required) A node name to establish a connection to the cluster..
=over

=item Return values

ARRAYREF of the majority partition nodes.

=back

=cut

package NACL::MTask::CtranPartitioning;

use strict;
use warnings;

use base qw/NACL::MTask::MTask/;
use Array::Utils qw(:all);
use Data::Dumper;
use NACL::C::Cluster;
use NACL::CS::DebugSitelist;
use NACL::C::Node;
use NATE::Log qw(log_global);
use NATE::BaseException;
use Params::Validate qw(validate_with SCALAR SCALARREF ARRAYREF HASHREF BOOLEAN OBJECT);
use NATE::ProcManager;
use NATE::Process;
use Time::HiRes qw(time);
use NATE::Events qw(eventloop_drain);
use NATE::ServiceAPI::Event qw(event_global);
use NATE::Messenger qw(messenger_global);
use Guard qw(scope_guard);
use NACL::CS::ClusterKernelService;

#######################################################################################
## Name        : new
## Description : Returns a CtranPartitioning object and also discovers the cluster size
# and creates a hash of node names and their node ids.
#######################################################################################

sub new {
    log_global()->enter();
    my ($pkg, @args) = @_;
    my %opts = validate_with(
        params          => \@args,
        spec => {
            command_interface  => { type => ARRAYREF|HASHREF, optional => 0},
        },
        allow_extra => 1,
    );

    my (%node_name_to_id, %node_name_to_order);
    my @nodes = NACL::C::Cluster->nodes(command_interface => $opts{command_interface});
    my $cluster_size = scalar @nodes;

    my @node_sitelist_state = NACL::CS::DebugSitelist->fetch(command_interface => $opts{command_interface});

    for(my $i=0; $i < scalar @node_sitelist_state; $i++) {
        my $node = $node_sitelist_state[$i]->node();
        my $node_id = $node_sitelist_state[$i]->sid();
        $node_name_to_id{$node} = $node_id;
    }

    my @sorted_ids = sort values %node_name_to_id;
    my $index = 0; 
    foreach my $id (@sorted_ids) {
        foreach my $key (keys %node_name_to_id) {
            if ($node_name_to_id{$key} eq $id) {
                $node_name_to_order{$key} = $index;
                $index++;
                last;
            }
        }
    }
    
    $opts{cluster_size} = $cluster_size;
    $opts{node_name_to_id} = \%node_name_to_id;
    $opts{node_name_to_order} = \%node_name_to_order;

    log_global()->exit();
    return bless (\%opts , $pkg);
}

#######################################################################################
## Name        : create_partitions
## Description : This method will create partitions in parallel.
#######################################################################################

sub create_partitions {
    log_global()->enter();
    my ($self, @args) = @_;
    my %opts = validate_with(
        params => \@args,
        spec   => {
            partitions => { type => ARRAYREF, optional => 0} ,
            runid_prefix => { type => SCALAR, optional => 1},
            auto_partition_remaining_nodes => { type => BOOLEAN, default => 0}
        },
    );

    #check no duplicates in partitions; throw error if duplicates exist.
    my @partitions = @{$opts{partitions}};
    my $num_of_partitions = scalar @partitions;
    my @union_of_partitions = ();
    my ($i, $j);
    my @array_of_partitions = ();

    for($i=0; $i < $num_of_partitions; $i++) {
        for($j=$i+1; $j < $num_of_partitions; $j++) {
            my @intersect = intersect(@{$partitions[$i]},@{$partitions[$j]});
            if (scalar @intersect) {
                NATE::BaseException->throw("The node partitions cannot overlap");
            }
        }
        @union_of_partitions = unique(@{$partitions[$i]},@union_of_partitions);
        push(@array_of_partitions, $partitions[$i]);
    }

    my @node_obj = NACL::C::Node->find();
    my @nodes;
    foreach my $node (@node_obj) {
        push( @nodes, $node->name() );
    }

    if($opts{auto_partition_remaining_nodes}) {
        my @individual_partitions = array_minus(@nodes, @union_of_partitions);
        push(@array_of_partitions, \@individual_partitions);
    }

    my %summary;
    my %node_name_to_mask;
    foreach my $partition (@array_of_partitions) {
        my $mask = $self->calculate_mask_for_partition(nodes => $partition);
        foreach my $node (@{$partition}) {
            $node_name_to_mask{$node} = $mask;
            $summary{$node}{"mask"} = $mask;
        }
    }

    my $sync_domain = __PACKAGE__ . "_${self}";
    my @proc_obj;
    my $listener_obj = event_global()->subscribe(
                            topic    => "sync",
                            domain   => $sync_domain,
                            listener => {
                                callback  => \&_handlesync,
                                synccount => 0,
                            }
                        );
    $listener_obj->{domain} = $sync_domain;

    my $e = messenger_global()->init(socket_type => 'rep');

    my $runid_prefix = "";
    if ($opts{runid_prefix}) {
        $runid_prefix = $opts{runid_prefix};
    }

    foreach my $node (keys %node_name_to_mask) {
        my $proc = NATE::Process->new(
            codespec => \&_set_mask_on_node,
            runid    => $runid_prefix."_setting_mask_on_node_".$node,
            args     => [ $e, $node, $node_name_to_mask{$node}, $sync_domain ],
            onexit => \&NATE::Process::on_exit_propagate_worst_result,
        );
        push( @proc_obj, $proc );
    }

    my $procMan = NATE::ProcManager->new(
        proc_info => \@proc_obj,
        runid     => "",
    );
    $listener_obj->{procobj} = \@proc_obj;
    $procMan->start();
    $procMan->wait();

    event_global()->unsubscribe(topic    => "sync", domain   => $sync_domain, listener => $listener_obj);
    my $listener = {
        count    => scalar @proc_obj,
        stop_count => 1,
        filter   => sub { shift->type() eq 'partition_summary' },
        callback => sub {
            my ($msg_obj) = shift;
            my ($list_obj) = shift;
            my $msg = $msg_obj->message();
            my @mask_set_time = @{$msg};
            $summary{$mask_set_time[0]}{"time"} = $mask_set_time[1];
            if ($list_obj->{stop_count} != $list_obj->{count}){
                $list_obj->{stop_count}++;
            }
        },
        charges => scalar @proc_obj,
    };
    messenger_global()->listener_add(listener => $listener);

    while ($listener->{stop_count} != (scalar @proc_obj)) {
        eventloop_drain();
    }

    messenger_global()->listener_remove(listener => $listener);

    log_global()->exit();
    return \%summary;
}

#######################################################################################
## Name        : _set_mask_on_node
## Description : This method will call set_mask_on_node() method and returns the time
# at which the mask is set..
#######################################################################################

sub _set_mask_on_node {
    log_global()->enter();
    my ( $e, $node, $mask, $domain ) = @_;

    _sync($domain);

    my $time = time();

    set_mask_on_node(node => $node, mask => $mask);
    messenger_global()->put(endpoint => $e, type => "partition_summary", message => [$node, $time]);

    log_global()->exit();
    return $time;
}

#######################################################################################
## Name        : reset_partitions
## Description : This method will reset the mask on the nodes in parallel.
#######################################################################################
sub reset_partitions {
    log_global()->enter();
    my ($self, @args) = @_;
    my %opts = validate_with(
        params => \@args,
        spec   => { nodes => { type => ARRAYREF , optional => 1},
                    runid_prefix => { type => SCALAR , optional => 1},
                  },
    );

    my $nodes;
    if (defined $opts{nodes}) {
        $nodes = $opts{nodes};
    } else {
        my $node_hash = $self->{node_name_to_id};
        my %node_hash = %$node_hash;
        my @node_names = keys %node_hash;
        $nodes = \@node_names;
    }

    my $sync_domain = __PACKAGE__ . "_${self}";
    my @proc_obj;
    my $listener_obj = event_global()->subscribe(
                            topic    => "sync",
                            domain   => $sync_domain,
                            listener => {
                                callback  => \&_handlesync,
                                synccount => 0,
                            }
                        );
    $listener_obj->{domain} = $sync_domain;
    my $runid_prefix = "";
    if ($opts{runid_prefix}) {
        $runid_prefix = $opts{runid_prefix};
    }

    foreach my $node (@{$nodes}) {
        my $proc = NATE::Process->new(
            codespec => \&_unset_mask_on_node,
            runid    => $runid_prefix."_unsetting_mask_on_node_".$node,
            args     => [ $node, $sync_domain ],
            onexit => \&NATE::Process::on_exit_propagate_worst_result,
        );
        push( @proc_obj, $proc );
    }

    my $procMan = NATE::ProcManager->new(
        proc_info => \@proc_obj,
        runid     => "",
    );
    $listener_obj->{procobj} = \@proc_obj;
    $procMan->start();
    $procMan->wait();

    event_global()->unsubscribe(topic    => "sync", domain   => $sync_domain, listener => $listener_obj);
    log_global()->exit();
}

#######################################################################################
## Name        : _unset_mask_on_node
## Description : This method calls unset_mask_on_node() method.
#######################################################################################

sub _unset_mask_on_node {
    log_global()->enter();
    my ( $node, $domain ) = @_;

    _sync($domain);

    unset_mask_on_node(node => $node);
    log_global()->exit();
}

#######################################################################################
## Name        : calculate_mask_for_partition
## Description : This method will calculate mask for the partition
#######################################################################################

sub calculate_mask_for_partition {
    log_global()->enter();
    my ($self, @args) = @_;
    my %opts = validate_with(
        params => \@args,
        spec   => { nodes => { type => ARRAYREF , optional => 0} },
    );

    my $nodes = $opts{nodes};
    my $mask;
    foreach my $node (@{$nodes}) {
        my $mask_for_node = $self->_calculate_mask_for_node(node => $node);
        $mask |= $mask_for_node
    }

    log_global()->exit();
    return $mask;
}

#######################################################################################
## Name        : get_majority_partition_nodes
## Description : This method will return an ARRAYREF of majority partition nodes.
#######################################################################################

sub get_majority_partition_nodes {
    log_global()->enter();

    my (%args) = @_;
    my $node = $args{'node'};
    my $node_obj = NACL::C::Node->new(node => $node);

    # getting majority partition nodes
    my @majority_partition_nodes = ();
    my $master_node = "";
    my @status_quorum = NACL::CS::ClusterKernelService->fetch(
                                                    command_interface => $node_obj,
                                                );
    foreach my $state (@status_quorum) {
        if (($state->master_node() eq $state->cluster_node()) && ($state->status_quorum() eq "in-quorum")) {
            $master_node = $state->master_node();
            last;
        }
    }

    foreach my $state (@status_quorum) {
        if (($state->master_node() eq $master_node) && ($state->status_quorum() eq "in-quorum")) {
            push (@majority_partition_nodes, $state->cluster_node());
        }
    }

    log_global()->exit();
    return \@majority_partition_nodes;
}

#######################################################################################
## Name        : _calculate_mask_for_node
## Description : This method will calculate mask for a node
#######################################################################################

sub _calculate_mask_for_node {
    log_global()->enter();
    my ($self, @args) = @_;
    print Dumper($self);
    my %opts = validate_with(
        params => \@args,
        spec   => { node => { type => SCALAR , optional => 0} },
    );
    my $node = $opts{node};
    my $index = $self->{node_name_to_order}{$node};

    my $cluster_size = $self->{cluster_size};
    my $mask = 0x1 << ($cluster_size - 1 - $index);
    print Dumper($mask);
    log_global()->exit();
    return $mask;
}

#######################################################################################
## Name        : set_mask_on_node
## Description : This static method will set the mask on the node.
#######################################################################################

sub set_mask_on_node {
    log_global()->enter();
    my (%args) = @_;
    my $node = $args{'node'};
    my $mask = $args{'mask'};

    my $command = {name => 'sysvar.kcs.ctran.debug.net_access_mask', value => $mask};

    my $nacl_node_obj = NACL::C::Node->new( name => $node);
    my $apiset = $nacl_node_obj->get_systemshell_apiset();
    $apiset->sysctl('privilege-level' => 'root', %$command);
    # display the value set for mask
    my $cmd = {name => 'sysvar.kcs.ctran.debug.net_access_mask'};
    $apiset->sysctl('privilege-level' => 'root', %$cmd);

    log_global()->exit();
}

#######################################################################################
## Name        : unset_mask_on_node
## Description : This static method will unset the mask on the node.
#######################################################################################

sub unset_mask_on_node {
    log_global()->enter();
    my (%args) = @_;
    my $node = $args{'node'};

    my $command = {name => 'sysvar.kcs.ctran.debug.net_access_mask', value => 0};

    my $nacl_node_obj = NACL::C::Node->new( name => $node);
    my $apiset = $nacl_node_obj->get_systemshell_apiset();
    $apiset->sysctl('privilege-level' => 'root', %$command);
    # display the value after unset of mask
    my $cmd = {name => 'sysvar.kcs.ctran.debug.net_access_mask'};
    $apiset->sysctl('privilege-level' => 'root', %$cmd);

    log_global()->exit();
}

#######################################################################################
## Name        : _handlesync
## Description : The callback method for the listener object to allow child processes to
# continue after sync.
#######################################################################################

sub _handlesync {
    my $msg_obj      = shift;
    my $listener_obj = shift;
    my $proccount    = 0;
    foreach my $proc ( @{ $listener_obj->{procobj} } ) {
        if ( $proc->is_running() ) {
            $proccount++;
        }
    }
    $listener_obj->{synccount}++;
    log_global()->debug( "Received sync request : " . $listener_obj->{synccount} );

    # publish a continue when all the processes have reached a sync point
    if ( $listener_obj->{synccount} eq $proccount ) {
        log_global()->debug("sending continue");

        $listener_obj->{synccount} = 0;

        event_global()->publish(
            topic  => "continue",
            domain => $listener_obj->{domain},
        );
    }
}

#######################################################################################
## Name        : _sync
## Description : This method is for achieving synchronization between the child processes.
# continue after sync.
#######################################################################################

sub _sync {
    my ( $domain ) = @_;
    my $continue     = 0;
    my $listener_obj = event_global()->subscribe(
                topic    => "continue",
                domain   => $domain,
                listener => {
                    "callback" => sub { $continue = 1; }
                },
                charges => 1,
    );

    log_global()->debug("sending sync request");
    event_global()->publish(
                topic  => "sync",
                domain => $domain,
    );
    while ( !$continue ) {
        eventloop_drain();
    }

    log_global()->debug("finished sync request");
}

1;

