# $Id: //depot/prod/test/nacldev/lib/NACL/MTask/DFM/Monitor.pm#1 $
#
## Copyright (c) 2012 NetApp, Inc., All Rights Reserved
##   Any use, modification, or distribution is prohibited
##   without prior written consent from NetApp, Inc.
##
#

## @summary Monitor Task Module
## @author Deepthi Shetty, dl-mpo-nacl-dev@netapp.com
## @change 2012-01-06 deepthis Adding contents
##
## @status shared
## @pod here

=head1 NAME

NACL::MTask::DFM::Monitor

=head1 DESCRIPTION

This task is to be used for triggering monitors and related operations on the OnCommand Server.

Many testcases require triggering certain monitors and then verifying if they have run successfully
or not. This package provides methods for selecting various monitors and selecting the wait time for
the monitor to be refreshed. 

    # Begining of test which calls trigger() by default when $monitor->run() is called. 
    use NACL::MTask::DFM::Monitor;
    ...
    my $monitor = NACL::MTask::DFM::Monitor->new(
        command_interface  => $Client,
	object_name_or_ids => $object_ids_ref, 
	... 
    );
    ...

    $monitor->run();

    # End test

=cut

package NACL::MTask::DFM::Monitor;

use strict;
use warnings;

use base qw(NACL::MTask::MTask );

use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Data::Dumper;
use Params::Validate qw(validate validate_with BOOLEAN HASHREF OBJECT ARRAYREF SCALAR);
use NATE::Exceptions::Argument qw(:try);
use NACL::Exceptions::OperationFailed;

use NACL::C::DFM::Job;

#
# Default Values
#
use constant DEFAULT_WAIT_TIME        => 900;
use constant DEFAULT_POLLING_INTERVAL => 60;

=head1 METHODS

=head2 new

    my $monitor = NACL::MTask::DFM::Monitor->new(
        command_interface  => $Client,
        object_name_or_ids => $object_list_ref,
        [ monitor_list => $monitor_names_array_ref, ]  
        [ complete_on_states => $state_array, ] 
        [ wait_time => $time_in_seconds, ]
        [ polling_interval => $time_in_seconds, ]
    );


Create a Monitor object that can be used to trigger monitors on
the OnCommand machine. 

=over

=back

=cut

use Class::MethodMaker [
    new    => [ '-hash', '-init', 'new' ],
    scalar => 'object_name_or_ids',
    scalar => 'monitor_list',
    scalar => 'complete_on_states',
    scalar => [ { -default => DEFAULT_WAIT_TIME },         'wait_time' ],
    scalar => [ { -default => DEFAULT_POLLING_INTERVAL },  'polling_interval' ],
    scalar => [ { -type    => 'NACL::C::DFM::DFMClient' }, 'command_interface', ],
];

# Validate the parameters

sub init {

    $Log->enter() if $may_enter;

    my $self = shift;

    my %opts = validate_with(
        params => \@_,
        spec   => {
            command_interface  => { type => OBJECT },
            object_name_or_ids => { type => ARRAYREF },
            monitor_list       => { type => ARRAYREF, optional => 1 },
            complete_on_states => { type => ARRAYREF, optional => 1 },
            wait_time          => { type => SCALAR, optional => 1 },
            polling_interval   => { type => SCALAR, optional => 1 },
        },
    );
    $Log->exit() if $may_exit;
}

=head2 run

    $monitor->run();
(Instance method)

    NACL::MTask::DFM::Monitor->run( 
	'command_interface'    => $ci,
	'object_name_or_ids'   => $object_list_ref,
	[ 'monitor_list'       => $monitor_list_ref, ] 
	[ 'complete_on_states' => $states_list_ref, ] 
	[ 'wait_time',	       => $wait_time, ] 
	[ 'polling_interval    => $polling_interval, ] 
    );
(Class method)
    
Run the monitor operation.

=item Options

=over

=item C<< command_interface => $ci >>

(Required, isa NACL::C::DFM::DFMClient)
The command interface through which the monitors for object will be refreshed.
See L<NACL::C::Component::command_interface|lib-NACL-C-Component-pm/command_interface>

=item C<< object_name_or_ids => $array_ref >>

(Required)
Reference to a list of object-name-or-ids containing the Id or Name of one or more 
objects whose monitors need to be refreshed. If no value is given, the throws a 
NACL::Exceptions::MissingArg exception with appropriate error message.   

=item C<< monitor_list => $array_ref >>

( Optional )
This is a reference to an array containing the list of monitors to be checked against for
completion of object refresh. If not specified, all monitors are checked for job completion.

=item C<< complete_on_states => $array_ref  >> 

( Optional )
This a reference to an array containing the states, on reaching any of which the monitor refresh 
will be cosidered as complete (or not). The default value is 'normal' i.e. when the output 
of the refresh has a job-status "normal" within the given wait period, the task will be considered successful. 
Valid values are: normal(default), error, partial_failures, warning.

=item C<< wait_time => $scalar >>

( Optional, defaults to 900 if not provided )
This is the amount of time (in SECONDS)  for which the method waits for all the monitors 
to get refreshed i.e the total time the method waits for the job completion.

=item C<< polling_interval => $scalar >>

( Optional, defaults to 60 if not provided )
This is the interval (in SECONDS) , for which periodically a check is done on the job status 
to check if it has completed or not. The default value is 60s, meaning: Every one minute 
the job status is checked to see if it has 'SUCCEEDED'.

=back

=over

=cut

sub run {

    $Log->enter() if $may_enter;
    my $self = shift;

    my %opts = $self->_common_validate_with( params => \@_ );

    my $cmd_interface    = $opts{'command_interface'};
    my $object_ref       = $opts{'object_name_or_ids'};
    my $monitor_ref      = $opts{'monitor_list'};
    my $completion_state = $opts{'complete_on_states'};
    my $wait_time        = $opts{'wait_time'};
    my $polling_interval = $opts{'polling_interval'};

    # Assign the client
    my $Client = $cmd_interface;

    # Create APISet object for ZAPI
    my $apiset_zapi = $Client->apiset( interface => "ZAPI" );

    # Dereference and store the array values
    my @object_ids = @{$object_ref};
    my @monitor_list = @{$monitor_ref} if defined $monitor_ref;

    # Dereference the array to get completion states
    my @success_statuses = ();

    ## Map te OLD States to new job statuses
    foreach my $end_state ( @{$completion_state} ) {

        my $warn_flag = 0;
        my $old_state = $end_state;

        if ( $end_state eq 'SUCCEEDED' ) {

            $warn_flag = 1;
            push( @success_statuses, "normal" );

        } elsif ( $end_state eq 'PARTIALLY_SUCCEEDED' ) {

            $warn_flag = 1;
            push( @success_statuses, "partial_failures" );

        } elsif ( $end_state =~ /(FAILED|SKIPPED)/ ) {

            $warn_flag = 1;
            $old_state = $1;
            push( @success_statuses, "error" );
        } else {
            push( @success_statuses, $end_state );
        }

        # Log a warning message for states
        if ($warn_flag) {
            $Log->warn(
                "The state $old_state is deprecated in NACL::MTask::DFM::Monitor. Valid states can be one of these: "
                  . "http://bard-quality.gdl.englab.netapp.com/job/bard_main_documentation/zapidoc/job/index.html#job-status"
            );
        }
    }    # end foreach @{$completion_state}

    # Define failure states and then remove elements of @success_statuses from it
    my @failure_statuses = qw ( error warning partial_failures );

    # Make two hashes containing as keys the states

    # %failed_status_hash contains states of @failure_statuses - @success_statuses
    my %failed_status_hash = map { $_ => "1" } @failure_statuses;
    delete @failed_status_hash{@success_statuses};

    # %success_status_hash contains states of @success_statuses
    my %success_status_hash = map { $_ => "1" } @success_statuses;

    # Array of 'Job' components
    my @job_components = ();

    # Array of job states for the Job components
    my @job_states = ();

    # Call object refresh for each object
    foreach my $obj (@object_ids) {

        # Call object-refresh
        my $response = $apiset_zapi->dfm_object_refresh( "object-name-or-id" => $obj );

        my $parsed_output = $response->get_parsed_output();
        my $job_id        = $parsed_output->[0]->{'job-id'};

        $Log->debug("\nJob id is $job_id for object $obj.");

        my $job_c = NACL::C::DFM::Job->new(
            command_interface => $cmd_interface,
            'job_id'          => $job_id,
        );
        push( @job_components, $job_c );

    }

    # Call job-wait-for completion for each job id
    foreach my $job (@job_components) {

        $job->wait_for_state( 'wait-for-state-timeout' => $wait_time, );

    }

    #### Check the status of the jobs ####

    # Status flags
    my $refresh_failed = 0;    # to check for any state in %failed_status_hash

    @job_states = ();

    # Check for each job triggered the status
    foreach my $job (@job_components) {

        # Job's state
        my $job_state_info = $job->state();

        push( @job_states, $job_state_info );

        # If no individual monitor list is given
        my $job_status = $job_state_info->status();

        if ( exists $success_status_hash{$job_status} ) {
            $Log->debug( "\nRefresh of monitors completed successfully for job-id: "
                  . $job_state_info->job_id()
                  . " with status: "
                  . $job_status );

        } elsif ( exists $failed_status_hash{$job_status} ) {

            $Log->debug( "\nRefresh of monitors with job-id "
                  . $job_state_info->job_id()
                  . " FAILED! with status: "
                  . $job_status );
            $refresh_failed = 1;

        }

    }    # end of foreach $job

    my $completion_info = "\nFinal state of all the jobs:\n";

    # Store info of final status of the job(s)
    foreach my $state_obj (@job_states) {

        $completion_info .= "\nJob-id: " . $state_obj->job_id() . "  Status: " . $state_obj->status();
    }

    if ($refresh_failed) {

        # Prepare the exception string
        my $exception_message = "\nRefresh failed in $wait_time seconds.";

        $exception_message .= $completion_info;

        $Log->exit() if $may_exit;

        # Throw exception
        NACL::Exceptions::OperationFailed->throw($exception_message);

    }

    $Log->log("Refresh of monitors completed successfully.");
    $Log->log($completion_info);

    $Log->exit() if $may_exit;

} ## end sub run

sub _object_attributes_validate_spec {

    return => {
        object_name_or_ids => { type => ARRAYREF },
        monitor_list       => { type => ARRAYREF, optional => 1 },
        complete_on_states => {
            type     => ARRAYREF,
            optional => 1,
            default  => [qw( normal )],
        },
        wait_time => {
            type     => SCALAR,
            optional => 1,
            default  => DEFAULT_WAIT_TIME
        },
        polling_interval => {
            type     => SCALAR,
            optional => 1,
            default  => DEFAULT_POLLING_INTERVAL
        },
    };

} ## end sub _object_attributes_validate_spec

1;
