# $Id: $
#
## Copyright (c) 2012 NetApp, Inc., All Rights Reserved
##   Any use, modification, or distribution is prohibited
##   without prior written consent from NetApp, Inc.
##

#
## @summary StorageServiceSetup MTask Module
## @author kvishal
## @change
##
## @status shared
## @pod here

=head1 NAME

NACL::MTask::DFM::StorageServiceSetup

=head1 DESCRIPTION

This task does the following

=item * Choose source and destination aggr ( default & optional )

=item * Create source and destination vserver

=item * Create volumes on source and destination server ( non default & optional )

=item ** Creates volume in parllel, if count > 1

=item * Create resource resourcepool and adding destination aggr and vserver as members

=head1 ATTRIBUTES

=head2 source_aggregate

Aggr object which is assigned as the source

=head2 destination_aggregate

Aggr object which is assigned as the destination

=head2 other_destination_aggregates

Aggr objects apart from destination aggr

=head2 source_vserver

vserver object assigned as source

=head2 destination_vserver

vserver object assigned as destination

=head2 source_aggregate_resource_key

Aggr resource key of source aggregate

=head2 destination_aggregate_resource_key

Aggr resource key of destination aggregate

=head2 source_vserver_resource_key

vserver resource key of source vserver

=head2 destination_vserver_resource_key

vserver resource key of destination vserver

=head2 source_volume_names

volume names source vserver

=head2 source_volume_names_dfm

volume names in DFM format created on source vserver

=head2 source_volume_resource_keys

resource keys of volume created on source vserver

=head2 destination_volume_names

volume names destination vserver

=head2 destination_volume_names_dfm

volume names in DFM format created on destination vserver

=head2 destination_volume_resource_keys

redestination keys of volume created on destination vserver

=cut

package NACL::MTask::DFM::StorageServiceSetup;

use strict;
use warnings;
use base qw(NACL::MTask::MTask);

use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Data::Dumper;
use Params::Validate
  qw(validate validate_with BOOLEAN HASHREF OBJECT ARRAYREF SCALAR SCALARREF);
use NATE::Exceptions::Argument qw(:try);
use NACL::ComponentUtils qw( _optional_scalars get_package_name );

use NACL::STask::Vserver;
use NACL::STask::Volume;
use NACL::STask::Aggregate;
use NACL::STask::DFM::Datasource;
use NACL::STask::DFM::ResourcePool;
use NACL::STask::Cluster;
use NACL::C::DFM::Resource;
use NACL::MTask::VolumeMulti;
use NACL::MTask::ClusterPeer;
use NACL::GeneralUtils qw(random_name_generator);

use Class::MethodMaker [
    new => [ '-hash', '-init', 'new' ],
    scalar => [ { -type => 'NACL::C::DFM::DFMClient' }, 'command_interface', ],
    scalar => [ { -type => 'NACL::STask::DFM::Datasource' }, 'source_cluster_obj' ],
    scalar => [ { -type => 'NACL::STask::DFM::Datasource' }, 'destination_cluster_obj' ],
    scalar => [ { -type => 'NACL::C::Aggregate' }, 'source_aggregate' ],
    scalar => [ { -type => 'NACL::C::Aggregate' }, 'destination_aggregate' ],
    array  => [ { -type => 'NACL::C::Aggregate' }, 'other_destination_aggregates' ],
    scalar => [ { -type => 'NACL::STask::Vserver' }, 'source_vserver' ],
    scalar => [ { -type => 'NACL::STask::Vserver' }, 'destination_vserver' ],
    scalar => [ { -type => 'NACL::STask::DFM::ResourcePool' }, 'resource_pool' ],
    scalar => [ { -type => 'NACL::MTask::ClusterPeer' }, 'cluster_peer_obj' ],
    scalar => [ 'source_aggregate_resource_key' ],
    scalar => [ 'destination_aggregate_resource_key' ],
    scalar => [ 'source_vserver_resource_key' ],
    scalar => [ 'destination_vserver_resource_key' ],
    array  => [ 'source_volume_names' ],
    array  => [ 'source_volume_names_dfm' ],
    array  => [ 'source_volume_resource_keys' ],
    array  => [ 'destination_volume_names' ],
    array  => [ 'destination_volume_names_dfm' ],
    array  => [ 'destination_volume_resource_keys' ],
];

use constant PREFIX => 'GMTASK';

my (@datasource_added_flag);

# Validate the parameters

sub init {

    $Log->enter() if $may_enter;

    my $pkg = shift;

    my %opts = validate_with(
        params => \@_,
        spec   => {
            command_interface  => { type => OBJECT },
        },
    );
    $Log->exit() if $may_exit;
}

=head1 METHODS


=head2 create_setup

    $setup->create_setup(
        source_cluster_interface => $src_cluster,
        destination_cluster_interface => $dst_cluster,
        nacltask_source_volume_count => 2,
    );
(Instance method)

=item Options

=over

=item C<< command_interface => $ci >>

(Required, isa NACL::C::DFM::DFMClient)
The command interface through which the monitors for object will be refreshed.
See L<NACL::C::Component::command_interface|lib-NACL-C-Component-pm/command_interface>

=item C<< source_cluster_interface => $src_cluster >>

(Required, isa NACL::C::Cluster)
The cluster interface through which the Vserver and Volume will be created in source cluster.
See L<NACL::C::Component::command_interface|lib-NACL-C-Component-pm/command_interface>

=item C<< destination_cluster_interface => $dst_cluster >>

(Optional, isa NACL::C::Cluster)
The destination cluster interface through which the Vserver and Volume will be subscribed.
default value for destination_cluster_interface is source_cluster_interface
See L<NACL::C::Component::command_interface|lib-NACL-C-Component-pm/command_interface>

=item C<< source_vserver_name => $src_vs >>

(Optional, Scalar)
Name of the pre-created source vserver. If not given MTask will creates its own src vserver

=item C<< nacltask_source_volume_count => 2 >>

(Optional, Scalar)
The number of volumes to be created on the source vserver. Default is 0

=item C<< nacltask_destination_volume_count => 2 >>

(Optional, Scalar)
The number of volumes to be created on the destination vserver. Default is 0

=item C<< nacltask_aggregate_add => 1 >>

(Optional, Scalar)
This flag decides if destination aggregate and vserver are added to resourcepool. Default is 1

=item C<< nacltask_vserver_properties => $vserver_properties >>

(Optional, Hashref)
A hash containing the inputs for src Vserver creation. It takes all the inputs given for
the create function of L<NACL::STask::Vserver|lib-NACL-STask-Vserver-pm>.
Refer L<NACL::STask::Vserver|lib-NACL-STask-Vserver-pm> for default values.

=item C<< nacltask_src_volume_properties => $volume_properties >>

(Optional, Hashref)
A hashref contianing the inputs for Volume creation. It takes all the inputs given for
the create function of L<NACL::STask::Volume|lib-NACL-STask-Volume-pm>.
Refer L<NACL::STask::Volume|lib-NACL-STask-Volume-pm> for default values.

=item C<< nacltask_dst_volume_properties => $volume_properties >>

(Optional, Hashref)
A hashref contianing the inputs for dest Volume creation. It takes all the inputs given for
the create function of L<NACL::STask::Volume|lib-NACL-STask-Volume-pm>.
Refer L<NACL::STask::Volume|lib-NACL-STask-Volume-pm> for default values.

=back

=over

=cut

sub create_setup {

    $Log->enter() if $may_enter;

    my $pkg = shift;

    $pkg->_verify_invocation( style => 'instance_only' );

    my %opts = $pkg->_common_validate_with(
        params          => \@_,
        additional_spec => {
            source_cluster_interface     => { type => 'NACL::C::Cluster' },
            destination_cluster_interface => { type => 'NACL::C::Cluster', optional => 1 },
            source_vserver_name          => { type => SCALAR, optional => 1, default =>undef },
            nacltask_source_volume_count => { type => SCALAR, optional => 1, default => 0 },
            nacltask_destination_volume_count => { type => SCALAR, optional => 1, default => 0 },
            nacltask_aggregate_add       => { type => SCALAR, optional => 1, default => 1 },
            nacltask_vserver_properties  => { type => HASHREF,   optional => 1 },
            nacltask_src_volume_properties   => { type => HASHREF,   optional => 1 },
            nacltask_dst_volume_properties   => { type => HASHREF,   optional => 1 },
        },
        ignore_primary_keys => 1,
    );
    my $command_interface       = $opts{command_interface};
    my $src_cluster_interface   = $opts{source_cluster_interface};
    my $dst_cluster_interface   = $opts{destination_cluster_interface};
    my $src_vserver_name        = $opts{source_vserver_name};
    my $src_volume_count        = $opts{nacltask_source_volume_count};
    my $dst_volume_count        = $opts{nacltask_destination_volume_count};
    my $aggregate_add           = $opts{nacltask_aggregate_add};
    my $vserver_properties      = delete $opts{nacltask_vserver_properties};
    my $src_volume_properties   = delete $opts{nacltask_src_volume_properties};
    my $dst_volume_properties   = delete $opts{nacltask_dst_volume_properties};
    my ( %vserver_properties, %src_volume_properties, %dst_volume_properties, @clusters );

    %vserver_properties = %{$vserver_properties}
      if ( defined $vserver_properties );
    %src_volume_properties = %{$src_volume_properties}
      if ( defined $src_volume_properties );
    %dst_volume_properties = %{$dst_volume_properties}
      if ( defined $dst_volume_properties );
    $dst_cluster_interface ||= $src_cluster_interface
      if ( !defined $dst_cluster_interface );

    ## Finding aggregates in source cluster & Destination cluster
    push( @clusters, $src_cluster_interface );

    my ( $src_aggregate, $dst_aggregate, @other_dest_aggrs );
    if ( $src_cluster_interface != $dst_cluster_interface ) {
        $Log->comment('Src and Dst clusters are different');
        $src_aggregate =
            NACL::STask::Aggregate->find_online_non_root_64_bit_arranged_by_most_availsize (
                command_interface => $src_cluster_interface );
        ( $dst_aggregate, @other_dest_aggrs ) =
            NACL::STask::Aggregate->find_online_non_root_64_bit_arranged_by_most_availsize (
                command_interface => $dst_cluster_interface );
        push( @clusters, $dst_cluster_interface );
        $pkg->_cluster_peer_clusters( $src_cluster_interface, $dst_cluster_interface );
    } else {
        $Log->comment('Src and Dst clusters are same');
        ( $src_aggregate, $dst_aggregate, @other_dest_aggrs ) =
            NACL::STask::Aggregate->find_online_non_root_64_bit_arranged_by_most_availsize (
                command_interface => $src_cluster_interface );
    }

    $pkg->source_aggregate( $src_aggregate );
    $pkg->destination_aggregate( $dst_aggregate );
    $pkg->other_destination_aggregates( @other_dest_aggrs );

    ## Creating Vserver on source cluster
    if( ! $src_vserver_name ) {
        $src_vserver_name = PREFIX . "src_vs" . random_name_generator;
        $Log->comment("Creating vserver: $src_vserver_name");
        my $src_vserver = NACL::STask::Vserver->create(
            command_interface  => $src_cluster_interface,
            aggregate          => $src_aggregate,
            vserver            => $src_vserver_name,
            nacltask_if_exists => 'purge',
            %vserver_properties
        );
        $pkg->source_vserver( $src_vserver );
    }


    ## Creating Volumes on source cluster
    my( $src_vol_names_ref, $src_vol_dfm_names_ref ) =  &_create_volumes(
        'src', $src_volume_count, $src_cluster_interface, $src_aggregate,
        $src_vserver_name, %src_volume_properties );
    $pkg->source_volume_names( @$src_vol_names_ref );
    $pkg->source_volume_names_dfm( @$src_vol_dfm_names_ref );

    ## Creating Vserver in destination aggregates
    my $dst_vserver_name = PREFIX . "dst_vs" . random_name_generator;
    $Log->comment('Creating destination vserver: $dst_vserver_name');
    my $dst_vserver = NACL::STask::Vserver->create(
        command_interface  => $dst_cluster_interface,
        aggregate          => $dst_aggregate,
        vserver            => $dst_vserver_name,
        nacltask_if_exists => 'purge',
        %vserver_properties
    );
    $pkg->destination_vserver( $dst_vserver );

    ## Creating Volumes on destination cluster
    my( $dst_vol_names_ref, $dst_vol_dfm_names_ref ) =  &_create_volumes(
        'dst', $dst_volume_count, $dst_cluster_interface, $dst_aggregate,
        $dst_vserver_name, %dst_volume_properties );
    $pkg->destination_volume_names( @$dst_vol_names_ref );
    $pkg->destination_volume_names_dfm( @$dst_vol_dfm_names_ref );

    ## Adding Datasource to the DFM
    my ( @datasource_added_flag, @datasource_objs, $datasource_added_flag );
    $Log->comment('Adding datasources');
    foreach my $cluster (@clusters) {
        my $cluster_obj = NACL::STask::DFM::Datasource->add(
            command_interface       => $command_interface,
            hostobj                 => $cluster->hostrec(),
            nacltask_if_exists      => 'reuse',
            _was_added              => \$datasource_added_flag,
        );

        push( @datasource_objs,       $cluster_obj );
        push( @datasource_added_flag, $datasource_added_flag );
    }
    $pkg->source_cluster_obj( $datasource_objs[0] );
    $pkg->destination_cluster_obj( $datasource_objs[1] );
    my $src_aggr_reskey = _get_aggr_reskey( $command_interface,  $src_aggregate );
    my $src_vs_reskey = _get_vserver_reskey( $command_interface, $src_vserver_name);
    $pkg->source_vserver_resource_key( $src_vs_reskey );
    $pkg->source_aggregate_resource_key( $src_aggr_reskey );
    $Log->comment('Source Vserver resource key: ' . $src_vs_reskey);
    $Log->comment('Source Aggregate resource key: ' . $src_aggr_reskey);

    ## Storing vol objects
    my @src_volume_reskeys =
        map { _get_volume_reskey($command_interface, $_) } @$src_vol_dfm_names_ref
        if( $src_volume_count );
    $pkg->source_volume_resource_keys( @src_volume_reskeys );
    $Log->comment( 'Source vol reskeys: ' . Dumper \@src_volume_reskeys );

    my @dst_volume_reskeys =
        map { _get_volume_reskey($command_interface, $_) } @$dst_vol_dfm_names_ref
        if( $dst_volume_count );
    $pkg->destination_volume_resource_keys( @dst_volume_reskeys );
    my $dst_vs_reskey = _get_vserver_reskey($command_interface, $dst_vserver_name);
    my $dst_aggr_reskey = _get_aggr_reskey($command_interface, $dst_aggregate);
    $pkg->destination_aggregate_resource_key( $dst_aggr_reskey );
    $pkg->destination_vserver_resource_key( $dst_vs_reskey );
    $Log->comment( 'Dest vol reskeys: ' . Dumper \@dst_volume_reskeys );
    $Log->comment('Destination Vserver resource key: ' . $dst_vs_reskey);
    $Log->comment('Destination Aggregate resource key: ' . $dst_aggr_reskey);

    ##  Creating the resource resourcepool
    $Log->comment('Creating resourcepool');
    my $resourcepool_obj = NACL::STask::DFM::ResourcePool->create(
        command_interface   => $command_interface,
        description         => 'Created by MTask',
    );

    $pkg->resource_pool( $resourcepool_obj );
    $Log->comment( ' ResourcePool resource key: ' . $resourcepool_obj->resource_key );

    ## Adding member to the resourcepool
    $Log->comment('Adding members to resourcepool');
    if( $aggregate_add ) {
        $resourcepool_obj->aggregate_add( 'aggregate' => $dst_aggr_reskey );
    }

    #
    # Set explicit associations for this new dst vserver,
    # for both mirror and vault connection types.
    #
    $pkg->_set_vserver_destination_association();

    $Log->exit;
}

=head2 purge_setup

$resourcepool_obj->purge_setup();

(Instance method)

=cut

sub purge_setup {
    $Log->enter() if $may_enter;
    my $pkg = shift;
    $pkg->_verify_invocation( style => 'instance_only' );

    my $command_interface      = $pkg->command_interface();
    my $src_vserver_obj        = $pkg->source_vserver();
    my $dst_vserver_obj        = $pkg->destination_vserver();
    my $src_cluster            = $pkg->source_cluster_obj();
    my $dst_cluster            = $pkg->destination_cluster_obj();
    my $resourcepool_obj       = $pkg->resource_pool();
    my $cluster_peer_obj       = $pkg->cluster_peer_obj();

    if ( defined $src_vserver_obj ) {
        $Log->comment('Destroying source vserver');
        try {
            $src_vserver_obj->purge( nacltask_if_purged => 'pass' );
        }
        catch NATE::BaseException with {
            my $ex = shift;
            $Log->error( "Error in src vserver cleanup: " . $ex->stacktrace );
        };

    }

    if ( defined $dst_vserver_obj ) {
        $Log->comment("Clear vserver destination associaton");
        $pkg->_set_vserver_destination_association( 'true' );

        $Log->comment('Destroying destination vserver');
        try {
            $dst_vserver_obj->purge( nacltask_if_purged => 'pass' );
        }
        catch NATE::BaseException with {
            my $ex = shift;
            $Log->error( "Error in dst vserver cleanup: " . $ex->stacktrace );
        };

    }

    if ( defined $resourcepool_obj ) {
        $Log->comment('Destroying resourcepool');
        try {
            $resourcepool_obj->destroy( nacltask_if_purged => 'pass' );
        }
        catch NATE::BaseException with {
            my $ex = shift;
            $Log->error( "Error in resourcepool cleanup: " . $ex->stacktrace );
        };

    }

    if ( $cluster_peer_obj ) {
        $Log->comment("Delete Cluster peer relationship.");
        try {
            $cluster_peer_obj->purge( nacltask_delete_lifs => 1 );
        }
        catch NATE::BaseException with {
            my $ex = shift;
            $Log->comment( "Error in Cluster peer relationship cleanup: "
                  . $ex->stacktrace );
        };
    }
    
    $Log->comment('Purging datasources');
    if ( defined $src_cluster && defined $datasource_added_flag[0] ) {
        $src_cluster->purge( nacltask_if_purged => 'pass' );
    }
    if ( defined $dst_cluster && defined $datasource_added_flag[1] ) {
        $dst_cluster->purge( nacltask_if_purged => 'pass' );
    }

    $Log->exit;
}

## private subroutines
sub _create_volumes {
    my( $type, $cnt, $interface, $aggr, $vserver, %volume_properties ) = @_;
    my( @volume_names, @dfm_vol_names );
    my $vol = PREFIX . "$type\_vol" . random_name_generator;
    if( $cnt ) {
        $Log->comment("Creating $type volumes: $cnt");
        if( $cnt > 1 ) {
            @volume_names = NACL::MTask::VolumeMulti->create_multi(
                command_interface           => $interface,
                aggregate                   => $aggr,
                vserver                     => $vserver,
                volume                      => $vol,
                nacltask_num_vols           => $cnt,
                nacltask_start_index        => 1,
                %volume_properties
            );
            @dfm_vol_names  = map { "$vserver:/$_"} @volume_names;
        } else {
            NACL::STask::Volume->create(
                command_interface  => $interface,
                aggregate          => $aggr,
                vserver            => $vserver,
                volume             => $vol,
                %volume_properties
            );
            push( @volume_names, $vol );
            push( @dfm_vol_names, "$vserver:/$vol" );
        }
    }
    return( \@volume_names, \@dfm_vol_names );
}

sub _get_aggr_reskey {
    my $command_interface = shift;
    my $aggregate = shift;
    my $node_name = $aggregate->get_local_node_name();
    my $aggr_name = $node_name . ":" . $aggregate->aggregate();

    return NACL::C::DFM::Resource->lookup(
        command_interface   => $command_interface,
        type                => 'aggregate',
        'qualified-name'    => $aggr_name,
    );
}

sub _get_vserver_reskey {
    return NACL::C::DFM::Resource->lookup(
        command_interface   => shift,
        type                => 'vserver',
        'qualified-name'    => shift,
    );
}

sub _get_volume_reskey {
    return NACL::C::DFM::Resource->lookup(
        command_interface   => shift,
        type                => 'volume',
        'qualified-name'    => shift,
    );
}

################################################################################
#
# Function: _set_vserver_destination_association
#
# Summary:
#     Create a cluster wide new vserver destination  mirror and vault associations
#     or clear an existing one.
#
#  Input:
#        client: The DFM command interface
#        destination-vserver-name: Name of the destination vserver
#        clear-association: If true, clear the specificed association.
#                If false, add the new association.
#                (Optional. Default is false.)
#
#  Output: NONE.
#
################################################################################

sub _set_vserver_destination_association {

        my $pkg = shift;
        my $clear_association = shift;
        $clear_association ||= 'false';
        my $apiset = $pkg->command_interface->apiset;

        $Log->comment("Set/clear vserver destination...");
        # define associations for both mirror & vault
        foreach ( 'mirror', 'vault' ) {
            $apiset->storage_service_vserver_set_protection_destination(
                'destination-vserver-resource-key' =>  $pkg->destination_vserver_resource_key,
                'source-vserver-resource-key' =>  $pkg->source_vserver_resource_key,
                'connection-type'       => $_,
                'clear-association' => $clear_association
            );
        }


} ## end sub _set_vserver_destination_association

sub _cluster_peer_clusters() {
    my ( $pkg, $cluster1, $cluster2 ) = @_;
    my $cluster_peer_was_established;

    my @cluster1_nodes = $cluster1->node_objs();
    my @cluster2_nodes = $cluster2->node_objs();

    my $cluster_peer_obj = NACL::MTask::ClusterPeer->create(
        home_cluster_interface    => $cluster1_nodes[0],
        remote_cluster_interface  => $cluster2_nodes[0],
        "nacltask_if_exists_peer" => 'reuse',
        nacltask_verify           => 1,
        _was_created              => \$cluster_peer_was_established,
    );

    if ($cluster_peer_was_established) {
        $Log->comment('Cluster peer established between');
        $pkg->cluster_peer_obj( $cluster_peer_obj );
    }
    else {
        $Log->comment('Cluster peer already established');
    }
}

1;
