#
# Copyright (c) 2012 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Data population Task Module
## @author dl-nacl-dev@netapp.com
## @status shared
## @pod here


=head1 NAME

NACL::MTask::Data - Data population on Unix and Windows.

=head1 DESCRIPTION

C<NACL::MTask::Data> provides a number of methods related to data population
on a client command interface.

=head1 METHODS

=cut

package NACL::MTask::Data;


use Params::Validate qw(validate SCALAR HASHREF OBJECT);
use Tharn qw/$Log/;

use NACL::STask::Client::RPC::Worker;
use strict;
use warnings;
use NACL::Exceptions::OperationFailed qw/:try/;
use NACL::APISet::Exceptions::TimeoutException qw/:try/;
use NATE::BaseException qw/:try/;

my @cont_opts = qw(
    dirs
    files
    streams
    sparse
    compress
    encrypt
    archive
    hidden
    acl
);

my %args_map = (
    file_size_avg   => 'size_avg',
    file_size_dev   => 'size_dev',
    stream_size_avg => 'size_avg',
    stream_size_dev => 'size_dev',
);

my $_seed;
# A hash that provides the default parameters for population
my %pop_template;
# A hash that provides the default parameters for modify
my %mod_template;

sub _get_seed();
sub _get_cont_opts(\%$;);
sub _get_args(\%$$;);
sub _create_none(\%;);

=head2 populate

Populate a directory on a client using the options specified. This method will 
start a process on the remote client where the data population has to happen.
By default, this call is blocking.

This method and its options have been ported from lib/RichData.pm

=over

=item Arguments

=over

=item C<< path => $directory_path >>

(Required) 

The absolute path of the directory on the client which will be populated.

=item C<< command_interface => $client_obj >>

(Required) 

C<NACL::C::Client> instance which represents the 
client on which the operation should be performed.

=item C<< template => $string >>

(Optional)

Template for population which sets the default values of all the following
options. Default is the 'default' template. Values corresponding to the 
templates can be found in the %pop_template hash.

=item C<< async => $boolean >>

(Optional) A true value indicates that the call will be non blocking
A false value (default) indicates that the call will be blocking.

=item C<< name_avg=>$name_avg >>

(Optional) Average number of characters for the name of objects.

=item C<< name_dev=>$name_dev >>

(Optional) Standard deviation of the number of characters for the name of objects.

=item C<< dir_avg=>$dir_avg >>

(Optional) Average number of directories to create under each directory.

=item C<< dir_dev=>$dir_dev >>

(Optional) Standard deviation of the number of directories to create under each directory.

=item C<< file_avg=>$file_avg >>

(Optional) Average number of files to create under each directory.

=item C<< file_dev=>$file_dev >>

(Optional) Standard deviation of the number of files to create under each directory.

=item C<< file_size_avg=>$file_size_avg >>

(Optional) Average size of files.

=item C<< file_size_dev=>$file_size_dev >>

(Optional) Standard deviation of the size of files.

=item C<< depth_avg=>$depth_avg >>

(Optional) Average directory depth.

=item C<< depth_dev=>$depth_dev >>

(Optional) Standard deviation of directory depth.

=item C<< stream_avg=>$stream_avg >>

(Optional) Average number of streams to create on file/directory.

=item C<< stream_dev=>$stream_dev >>

(Optional) Standard deviation of the number of streams to create on file/directory.

=item C<< stream_size_avg=>$stream_size_avg >>

(Optional) Average size of streams.

=item C<< stream_size_dev=>$stream_size_dev >>

(Optional) Standard deviation of the size of streams.

=item C<< dir_acl_num=>$dir_acl_num >>

(Optional) Number of directories to add an ACL.

=item C<< dir_acl_pct=>$dir_acl_pct >>

(Optional)Percentage of directories to add an ACL.

=item C<< file_acl_num=>$file_acl_num >>

(Optional) Number of files to add an ACL.

=item C<< file_acl_pct=>$file_acl_pct >>

(Optional)Percentage of files to add an ACL.

=item C<< dir_stream_num=>$dir_stream_num >>

(Optional) Number of directories to add streams.

=item C<< dir_stream_pct=>$dir_stream_pct >>

(Optional) Percentage of directories to add streams.

=item C<< file_stream_num=>$file_stream_num >>

(Optional) Number of files to add streams.

=item C<< file_stream_pct=>$file_stream_pct >>

(Optional) Percentage of files to add streams.

=item C<< sparse_num=>$sparse_num >>

(Optional)Number of files to be sparse files.

=item C<< sparse_pct=>$sparse_pct >>

(Optional) Percentage of files to be sparse files.

=item C<< link_num=>$link_num >>

(Optional) Number of files to be links.

=item C<< link_pct=>$link_pct >>

(Optional) Percentage of files to be links.

=item C<< Slink_num=>$Slink_num >>

(Optional) Number of files to be symbolic links.

=item C<< slink_pct=>$slink_pct >>

(Optional) Percentage of files to be symbolic links.

=item C<< compress_num=>$compress_num >>

(Optional)Number of objects to set the compress attribute.

=item C<< compress_pct=>$compress_pct >>

(Optional) Percentage of objects to set the compress attribute.

=item C<< encrypt_num=>$encrypt_num >>

(Optional) Number of objects to set the encrypt attribute.

=item C<< encrypt_pct=>$encrypt_pct >>

(Optional) Percentage of objects to set the encrypt attribute.

=item C<< archive_num=>$archive_num >>

(Optional) Number of objects to set the archive attribute.

=item C<< archive_pct=>$archive_pct >>

(Optional)Percentage of objects to set the archive attribute.

=item C<< hidden_num=>$hidden_num >>

(Optional) Number of objects to set the hidden attribute.

=item C<< hidden_pct=>$hidden_pct >>

(Optional) Percentage of objects to set the hidden attribute.

=item C<< acl_opts=>$acl_opts >>

(Optional) Hash reference of options for how to create each ACL.

      acl_masks     => Use a maximum of this number of access mask flags (0-18)
                       (default is 8; 0 means use maximum available).
      acl_flags     => Use a maximum of this number of flags (0-7)
                       (default is 4; 0 means use maximum available).
      acl_types     => Use a maximum of this number of types (0-3)
                       (default is 1; 0 means use maximum available).
      acl_users     => Use a maximum of this number of users (0-7)
                       (default is 2; 0 means use maximum available).
      users         => List of users/groups to use instead of the defaults.
      acl           => Create file with an ACL (1 or 2).
                       1: the number of ACEs varies between 1 and the product
                          of the acl_* options
                       2: the number of ACEs is the product of the
                          acl_* options (fixed).

=item C<< sparse_opts=>$sparse_opts >>

(Optional) Hash reference of options for how to create sparse files.

      ndata      => Number of data regions to create.
      ndata_dev  => Standard deviation of the number of data regions.
      offset     => Offset between data regions.
      offset_dev => Standard deviation of the offset between data regions.
      size       => Size of each data region.
      size_dev   => Standard deviation of the size of each data region.

=item C<< timeout => $number >>

(Optional)
Timeout value for function execution.
Default is 1800.      

=back

=back

=over

=item Returns

An object which represents the data population operation.
When the method call is non blocking, this object can be used later to wait for 
operation to finish.
The data population  in progress can also be killed using this object.

=back

=cut

sub populate {
    my $self = shift;
    my (%opts) = validate (
        @_,
        {
            path            => { type => SCALAR },
            command_interface => { type => OBJECT, isa => 'NACL::C::Client' },
            template        => { type => SCALAR,   default => 'default' },
            name_avg        => { type => SCALAR,  optional => 1 },
            name_dev        => { type => SCALAR,  optional => 1 },
            dir_avg         => { type => SCALAR,  optional => 1 },
            dir_dev         => { type => SCALAR,  optional => 1 },
            file_avg        => { type => SCALAR,  optional => 1 },
            file_dev        => { type => SCALAR,  optional => 1 },
            file_size_avg   => { type => SCALAR,  optional => 1 },
            file_size_dev   => { type => SCALAR,  optional => 1 },
            depth_avg       => { type => SCALAR,  optional => 1 },
            depth_dev       => { type => SCALAR,  optional => 1 },
            stream_avg      => { type => SCALAR,  optional => 1 },
            stream_dev      => { type => SCALAR,  optional => 1 },
            stream_size_avg => { type => SCALAR,  optional => 1 },
            stream_size_dev => { type => SCALAR,  optional => 1 },
            dir_acl_num     => { type => SCALAR,  optional => 1 },
            dir_acl_pct     => { type => SCALAR,  optional => 1 },
            file_acl_num    => { type => SCALAR,  optional => 1 },
            file_acl_pct    => { type => SCALAR,  optional => 1 },
            dir_stream_num  => { type => SCALAR,  optional => 1 },
            dir_stream_pct  => { type => SCALAR,  optional => 1 },
            file_stream_num => { type => SCALAR,  optional => 1 },
            file_stream_pct => { type => SCALAR,  optional => 1 },
            sparse_num      => { type => SCALAR,  optional => 1 },
            sparse_pct      => { type => SCALAR,  optional => 1 },
            link_num        => { type => SCALAR,  optional => 1 },
            link_pct        => { type => SCALAR,  optional => 1 },
            slink_num       => { type => SCALAR,  optional => 1 },
            slink_pct       => { type => SCALAR,  optional => 1 },
            compress_num    => { type => SCALAR,  optional => 1 },
            compress_pct    => { type => SCALAR,  optional => 1 },
            encrypt_num     => { type => SCALAR,  optional => 1 },
            encrypt_pct     => { type => SCALAR,  optional => 1 },
            archive_num     => { type => SCALAR,  optional => 1 },
            archive_pct     => { type => SCALAR,  optional => 1 },
            hidden_num      => { type => SCALAR,  optional => 1 },
            hidden_pct      => { type => SCALAR,  optional => 1 },
            acl_opts        => { type => HASHREF, optional => 1 },
            sparse_opts     => { type => HASHREF, optional => 1 },
            fill_vol        => { type => SCALAR,   default => 100 },
            timeout         => { type => SCALAR,   default => 1800 },
            dbfile          => { type => SCALAR,  default => "" },
            create_directory => { type => SCALAR, default => 0 },
            async => {type => SCALAR, default => 0}
        }
    );
    my ($sparse_opts, $acl_opts);
    my $client = $opts{command_interface};
    my ($obj, $dataset);
    my $async = delete $opts{async};
    
    $acl_opts    = $pop_template{$opts{template}}{acl_opts};
    $sparse_opts = $pop_template{$opts{template}}{sparse_opts};
    %opts = ( %{$pop_template{$opts{template}}}, %opts, _mode => 'populate' );

    # Sparse file options
    $opts{sparse_opts} = {
        # Values for sparse file options from template
        %$sparse_opts,
        # Overwrite defaults with user values
        %{$opts{sparse_opts}},
    };

    # ACL options
    $opts{acl_opts} = {
        # Values for ACL options from template
        %$acl_opts,
        # Overwrite defaults with user values
        %{$opts{acl_opts}},
    };

    
    my $worker = NACL::STask::Client::RPC::Worker->new( client => $client );
    $dataset->{worker} = $worker;
    
    $worker->start();
    $worker->call( eval => 'require NACL::MTask::DataUtils::LocalOps;');
    my $obj_wrapper = $worker->call( eval => 'NACL::MTask::DataUtils::LocalOps->new(@_)',
                     args => [
                                path => $opts{path} ,
                                fillpct => $opts{fill_vol},
                                save    => 1,
                                seed    => _get_seed,
                                timeout => $opts{timeout},
                                dbfile  => $opts{dbfile},
                             ],
                      wrap_return => 1
                    );
    $dataset->{localops} = $obj_wrapper;
    
    bless $dataset, $self;
    
    $dataset->_check_options(\%opts);
    _init_create_opts(\%opts);

    my $future = $obj_wrapper->_rpc_submit(method => 'create_files', 
                                          args => [ %{$opts{_file_args}} ] );
    push( @{$dataset->{jobs}}, $future );
    
    if ( $opts{depth_avg} > 0 ) {
            $future = $obj_wrapper->_rpc_submit( method => 'create_dirs', 
                                       args => [ %{$opts{_dir_args}} ]
                                     );
            push( @{$dataset->{jobs}}, $future );
    }
    
    if (!$async) {
        $dataset->wait_for_completion( timeout => $opts{timeout} );
    }
    
    

    return $dataset;
}

=head2 modify

Modify a data set.
The *_opt options is used to overwrite some of the main options, especially
the conditional options (dirs, files, streams, sparse, compress, encrypt,
archive, hidden and acl).

This method and its options have been ported from lib/RichData.pm

=over

=item Arguments

=over

=item C<< path => $string >>

(Required)
Top level directory of tree to modify.

=item C<< command_interface => $client_obj >>

(Required)
C<NACL::C::Client> instance which represents the 
client on which the operation should be performed.

=item C<< template => $string >>

(Optional)
Template for modify which sets the default values of all the following
options. Default is the 'default' template.
Values corresponding to the templates can be found in the %mod_template hash

=item C<< name_avg => $number >>

(Optional)
Average number of characters for the name of objects.

=item C<< name_dev => $number >>

(Optional)
Standard deviation of the number of characters for the name of objects.

=item C<< dir_avg => $number >>

(Optional)
Average number of directories to create under each directory.

=item C<< dir_dev => $number >>

(Optional)
Standard deviation of the number of directories to create under each directory.

=item C<< file_avg => $number >>

(Optional)
Average number of files to create under each directory.

=item C<< file_dev => $number >>

(Optional)
Standard deviation of the number of files to create under each directory.

=item C<< file_size_avg => $number >>

(Optional)
Average size of files.

=item C<< file_size_dev => $number >>

(Optional)
Standard deviation of the size of files.

=item C<< stream_avg => $number >>

(Optional)
Average number of streams to create on file/directory.

=item C<< stream_dev => $number >>

(Optional)
Standard deviation of the number of streams to create on file/directory.

=item C<< stream_size_avg => $number >>

(Optional)
Average size of streams.

=item C<< stream_size_dev => $number >>

(Optional)
Standard deviation of the size of streams.

=item C<< dir_acl_num => $number >>

(Optional)
Number of directories to add an ACL.

=item C<< dir_acl_pct => $number >>

(Optional)
Percentage of directories to add an ACL.

=item C<< file_acl_num => $number >>

(Optional)
Number of files to add an ACL.

=item C<< file_acl_pct => $number >>

(Optional)
Percentage of files to add an ACL.

=item C<< dir_stream_num => $number >>

(Optional)
Number of directories to add streams.

=item C<< dir_stream_pct => $number >>

(Optional)
Percentage of directories to add streams.

=item C<< file_stream_num => $number >>

(Optional)
Number of files to add streams.

=item C<< file_stream_pct => $number >>

(Optional)
Percentage of files to add streams.

B<sparse_num => I<number>

(Optional)
Number of files to be sparse files.

=item C<< sparse_pct => $number >>

(Optional)
Percentage of files to be sparse files.

=item C<< link_num => $number >>

(Optional)
Number of files to be links.

=item C<< link_pct => $number >>

(Optional)
Percentage of files to be links.

=item C<< slink_num => $number >>

(Optional)
Number of files to be symbolic links.

=item C<< slink_pct => $number >>

(Optional)
Percentage of files to be symbolic links.

=item C<< compress_num => $number >>

(Optional)
Number of objects to set the compress attribute.

=item C<< compress_pct => $number >>

(Optional)
Percentage of objects to set the compress attribute.

B<encrypt_num=> I<number>

(Optional)
Number of objects to set the encrypt attribute.

=item C<< encrypt_pct => $number >>

(Optional)
Percentage of objects to set the encrypt attribute.

=item C<< archive_num => $number >>

(Optional)
Number of objects to set the archive attribute.

=item C<< archive_pct => $number >>

(Optional)
Percentage of objects to set the archive attribute.

=item C<< hidden_num => $number >>

(Optional)
Number of objects to set the hidden attribute.

=item C<< hidden_pct => $number >>

(Optional)
Percentage of objects to set the hidden attribute.

=item C<< create_on_dir_num => $number >>

(Optional)
Number of directories to add more objects.

=item C<< create_on_dir_pct => $number >>

(Optional)
Percentage of directories to add more objects.

=item C<< create_on_file_num => $number >>

(Optional)
Number of files to add more objects.

=item C<< create_on_file_pct => $number >>

(Optional)
Percentage of files to add more objects.

=item C<< modify_file_num => $number >>

(Optional)
Number of files to modify.

=item C<< modify_file_pct => $number >>

(Optional)
Percentage of files to modify.

=item C<< modify_stream_num => $number >>

(Optional)
Number of streams to modify.

=item C<< modify_stream_pct => $number >>

(Optional)
Percentage of streams to modify.

=item C<< modify_sparse_num => $number >>

(Optional)
Number of sparse files to modify.

=item C<< modify_sparse_pct => $number >>

(Optional)
Percentage of sparse files to modify.

=item C<< remove_dir_num => $number >>

(Optional)
Number of directories to remove. It only removes the leaf directories
(directories which do not contain other directories)

=item C<< remove_dir_pct => $number >>

(Optional)
Percentage of directories to remove. It only removes the leaf directories
(directories which do not contain other directories)

=item C<< remove_file_num => $number >>

(Optional)
Number of files to remove.

=item C<< remove_file_pct => $number >>

(Optional)
Percentage of files to remove.

=item C<< remove_stream_num => $number >>

(Optional)
Number of streams to remove.

=item C<< remove_stream_pct => $number >>

(Optional)
Percentage of streams to remove.

=item C<< rename_dir_num => $number >>

(Optional)
Number of directories to rename.

=item C<< rename_dir_pct => $number >>

(Optional)
Percentage of directories to rename.

=item C<< rename_file_num => $number >>

(Optional)
Number of files to rename.

=item C<< rename_file_pct => $number >>

(Optional)
Percentage of files to rename.

=item C<< truncate_file_num => $number >>

(Optional)
Number of files to truncate.

=item C<< truncate_file_pct => $number >>

(Optional)
Percentage of files to truncate.

=item C<< truncate_stream_num => $number >>

(Optional)
Number of streams to truncate.

=item C<< truncate_stream_pct => $number >>

(Optional)
Percentage of streams to truncate.

=item C<< acl_mod_dir_num => $number >>

(Optional)
Maximum number of directories to modify ACL's.

=item C<< acl_mod_dir_pct => $number >>
Percentage of directories to modify ACL's.

(Optional)

=item C<< acl_mod_file_num => $number >>

(Optional)
Maximum number of files to modify ACL's.

=item C<< acl_mod_file_pct => $number >>

(Optional)
Percentage of files to modify ACL's.

=item C<< acl_del_dir_num => $number >>

(Optional)
Maximum number of directories to delete ACL's.

=item C<< acl_del_dir_pct => $number >>

(Optional)
Percentage of directories to delete ACL's.

=item C<< acl_del_file_num => $number >>

(Optional)
Maximum number of files to delete ACL's.

=item C<< acl_del_file_pct => $number >>

(Optional)
Percentage of files to delete ACL's.

=item C<< sattrs_dir_num => $number >>

(Optional)
Maximum number of directories to set attributes.

=item C<< sattrs_dir_pct => $number >>

(Optional)
Percentage of directories to set attributes.

=item C<< sattrs_file_num => $number >>

(Optional)
Maximum number of files to set attributes.

=item C<< sattrs_file_pct => $number >>

(Optional)
Percentage of files to set attributes.

B<dirs> => C<yes|no|any>

(Optional)
Modify directories if they contain directories.

B<files> => C<yes|no|any>

(Optional)
Modify directories if they contain files.

B<streams> => C<yes|no|any>

(Optional)
Modify directories/files if they contain streams.

B<sparse> => C<yes|no|any>

(Optional)
Modify files if they are sparse files.

B<compress> => C<yes|no|any>

(Optional)
Modify files if they are compressed.

B<encrypt> => C<yes|no|any>

(Optional)
Modify files if they are encrypted.

B<archive> => C<yes|no|any>

(Optional)
Modify files if they are archived.

B<hidden> => C<yes|no|any>

(Optional)
Modify files if they are hidden.

B<acl> => C<yes|no|any>

(Optional)
Modify directories/files with ACL.

=item C<< modify_file_opt => $hash_options >>

(Optional)
Modify options for files. Half of the files to be modified use a full
modify which replaces the entire contents of the file. The other half,
use a block modify which only changes the contents of the given number
of blocks.

      file_size_avg => New average size of file (full modify).
                       Default is the value of main file_size_avg option.
      file_size_dev => Standard deviation of new size of file (full modify).
                       Default is the value of main file_size_dev option.
      blocks        => Number of blocks to modify. Default is 4.
      block_size    => Size of each block to modify. Default is 1024.

=item C<< modify_stream_opt => $hash_options >>

(Optional)
Modify options for streams. Same as modify_file_opt.

=item C<< modify_sparse_opt => $hash_options >>

(Optional)
Modify options for sparse files. Overwrite the sparse options given by
the main option sparse_opts.

=item C<< remove_dir_opt => $hash_options >>

(Optional)
Remove options for directories.

=item C<< remove_file_opt => $hash_options >>

(Optional)
Remove options for files.

=item C<< rename_dir_opt => $hash_options >>

(Optional)
Rename options for directories.

=item C<< rename_file_opt => $hash_options >>

(Optional)
Rename options for files.

=item C<< truncate_file_opt => $hash_options >>

(Optional)
Truncate options for files.

      file_size_avg => New average size of file.
                       Default is the value of main file_size_avg option.
      file_size_dev => Standard deviation of new size of file.
                       Default is the value of main file_size_dev option.

=item C<< truncate_stream_opt => $hash_options >>

(Optional)
Truncate options for files. Same as truncate_file_opt.

=item C<< acl_mod_dir_opt => $hash_options >>

(Optional)
Modify ACL options for directories.

=item C<< acl_mod_file_opt => $hash_options >>

(Optional)
Modify ACL options for files.

=item C<< acl_del_dir_opt => $hash_options >>

(Optional)
Remove ACL options for directories.

=item C<< acl_del_file_opt => $hash_options >>

(Optional)
Remove ACL options for files.

=item C<< sattrs_dir_opt => $hash_options >>

(Optional)
Set attributes options for directories.

=item C<< sattrs_file_opt => $hash_options >>

(Optional)
Set attributes options for files.

=item C<< acl_opts => $hash_options >>

(Optional)
Hash reference of options for how to create each ACL.

      acl_masks => Use a maximum of this number of access mask flags (0-18)
                   (default is 8; 0 means use maximum available).
      acl_flags => Use a maximum of this number of flags (0-7)
                   (default is 4; 0 means use maximum available).
      acl_types => Use a maximum of this number of types (0-3)
                   (default is 1; 0 means use maximum available).
      acl_users => Use a maximum of this number of users (0-7)
                   (default is 2; 0 means use maximum available).
      users     => List of users/groups to use instead of the defaults.
      acl       => Create file with an ACL (1 or 2).
                   1: the number of ACEs varies between 1 and the product
                      of the acl_* options
                   2: the number of ACEs is the product of the
                      acl_* options (fixed).

=item C<< sparse_opts => $hash_options >>

(Optional)
Hash reference of options for how to create sparse files.

      ndata      => Number of data regions to create.
      ndata_dev  => Standard deviation of the number of data regions.
      offset     => Offset between data regions.
      offset_dev => Standard deviation of the offset between data regions.
      size       => Size of each data region.
      size_dev   => Standard deviation of the size of each data region.

=item C<< sattrs_opts => $hash_options >>

(Optional)
Hash reference of options for attributes.

      compress => Set or clear the compress attribute.
      encrypt  => Set or clear the encrypt attribute.
      archive  => Set or clear the archive attribute.
      hidden   => Set or clear the hidden attribute.

=item C<< fill_vol => $number >>

(Optional)
The percentage of space usage that will not be exceeded.
Only supported on Windows. Default is 100.

=item C<< timeout => $number >>

(Optional)
Timeout value for function execution.
Default is 1800.

=item C<< dbfile => $string >>

(Optional)
Unique database file name to allow multiple clients to work in the same
directory.

=item C<< async => $boolean >>

(Optional) A true value indicates that the call will be non blocking
A false value (default) indicates that the call will be blocking.

=back

=back

=over

=item Returns

An object which represents the data modify operation.
When the method call is non blocking, this object can be used later to wait for 
operation to finish.
The data population  in progress can also be killed using this object.

=back

=cut

#---------------------------------------------------------------------
sub modify {
    my $self = shift;
    my (%opts) = validate (
        @_,
        {
            path                => { type => SCALAR },
            command_interface   => { type => OBJECT,   isa => 'NACL::C::Client' },
            template            => { type => SCALAR,   default => 'default' },
            name_avg            => { type => SCALAR,  optional => 1 },
            name_dev            => { type => SCALAR,  optional => 1 },
            dir_avg             => { type => SCALAR,  optional => 1 },
            dir_dev             => { type => SCALAR,  optional => 1 },
            file_avg            => { type => SCALAR,  optional => 1 },
            file_dev            => { type => SCALAR,  optional => 1 },
            file_size_avg       => { type => SCALAR,  optional => 1 },
            file_size_dev       => { type => SCALAR,  optional => 1 },
            stream_avg          => { type => SCALAR,  optional => 1 },
            stream_dev          => { type => SCALAR,  optional => 1 },
            stream_size_avg     => { type => SCALAR,  optional => 1 },
            stream_size_dev     => { type => SCALAR,  optional => 1 },
            dir_acl_num         => { type => SCALAR,  optional => 1 },
            dir_acl_pct         => { type => SCALAR,  optional => 1 },
            file_acl_num        => { type => SCALAR,  optional => 1 },
            file_acl_pct        => { type => SCALAR,  optional => 1 },
            dir_stream_num      => { type => SCALAR,  optional => 1 },
            dir_stream_pct      => { type => SCALAR,  optional => 1 },
            file_stream_num     => { type => SCALAR,  optional => 1 },
            file_stream_pct     => { type => SCALAR,  optional => 1 },
            sparse_num          => { type => SCALAR,  optional => 1 },
            sparse_pct          => { type => SCALAR,  optional => 1 },
            link_num            => { type => SCALAR,  optional => 1 },
            link_pct            => { type => SCALAR,  optional => 1 },
            slink_num           => { type => SCALAR,  optional => 1 },
            slink_pct           => { type => SCALAR,  optional => 1 },
            compress_num        => { type => SCALAR,  optional => 1 },
            compress_pct        => { type => SCALAR,  optional => 1 },
            encrypt_num         => { type => SCALAR,  optional => 1 },
            encrypt_pct         => { type => SCALAR,  optional => 1 },
            archive_num         => { type => SCALAR,  optional => 1 },
            archive_pct         => { type => SCALAR,  optional => 1 },
            hidden_num          => { type => SCALAR,  optional => 1 },
            hidden_pct          => { type => SCALAR,  optional => 1 },
            create_on_dir_num   => { type => SCALAR,  optional => 1 },
            create_on_dir_pct   => { type => SCALAR,  optional => 1 },
            create_on_file_num  => { type => SCALAR,  optional => 1 },
            create_on_file_pct  => { type => SCALAR,  optional => 1 },
            modify_file_num     => { type => SCALAR,  optional => 1 },
            modify_file_pct     => { type => SCALAR,  optional => 1 },
            modify_stream_num   => { type => SCALAR,  optional => 1 },
            modify_stream_pct   => { type => SCALAR,  optional => 1 },
            modify_sparse_num   => { type => SCALAR,  optional => 1 },
            modify_sparse_pct   => { type => SCALAR,  optional => 1 },
            remove_dir_num      => { type => SCALAR,  optional => 1 },
            remove_dir_pct      => { type => SCALAR,  optional => 1 },
            remove_file_num     => { type => SCALAR,  optional => 1 },
            remove_file_pct     => { type => SCALAR,  optional => 1 },
            remove_stream_num   => { type => SCALAR,  optional => 1 },
            remove_stream_pct   => { type => SCALAR,  optional => 1 },
            rename_dir_num      => { type => SCALAR,  optional => 1 },
            rename_dir_pct      => { type => SCALAR,  optional => 1 },
            rename_file_num     => { type => SCALAR,  optional => 1 },
            rename_file_pct     => { type => SCALAR,  optional => 1 },
            truncate_file_num   => { type => SCALAR,  optional => 1 },
            truncate_file_pct   => { type => SCALAR,  optional => 1 },
            truncate_stream_num => { type => SCALAR,  optional => 1 },
            truncate_stream_pct => { type => SCALAR,  optional => 1 },
            acl_mod_dir_num     => { type => SCALAR,  optional => 1 },
            acl_mod_dir_pct     => { type => SCALAR,  optional => 1 },
            acl_mod_file_num    => { type => SCALAR,  optional => 1 },
            acl_mod_file_pct    => { type => SCALAR,  optional => 1 },
            acl_del_dir_num     => { type => SCALAR,  optional => 1 },
            acl_del_dir_pct     => { type => SCALAR,  optional => 1 },
            acl_del_file_num    => { type => SCALAR,  optional => 1 },
            acl_del_file_pct    => { type => SCALAR,  optional => 1 },
            sattrs_dir_num      => { type => SCALAR,  optional => 1 },
            sattrs_dir_pct      => { type => SCALAR,  optional => 1 },
            sattrs_file_num     => { type => SCALAR,  optional => 1 },
            sattrs_file_pct     => { type => SCALAR,  optional => 1 },
            dirs                => { type => SCALAR,  optional => 1 },
            files               => { type => SCALAR,  optional => 1 },
            streams             => { type => SCALAR,  optional => 1 },
            sparse              => { type => SCALAR,  optional => 1 },
            compress            => { type => SCALAR,  optional => 1 },
            encrypt             => { type => SCALAR,  optional => 1 },
            archive             => { type => SCALAR,  optional => 1 },
            hidden              => { type => SCALAR,  optional => 1 },
            acl                 => { type => SCALAR,  optional => 1 },
            modify_file_opt     => { type => HASHREF, optional => 1 },
            modify_stream_opt   => { type => HASHREF, optional => 1 },
            modify_sparse_opt   => { type => HASHREF, optional => 1 },
            remove_dir_opt      => { type => HASHREF, optional => 1 },
            remove_file_opt     => { type => HASHREF, optional => 1 },
            rename_dir_opt      => { type => HASHREF, optional => 1 },
            rename_file_opt     => { type => HASHREF, optional => 1 },
            truncate_file_opt   => { type => HASHREF, optional => 1 },
            truncate_stream_opt => { type => HASHREF, optional => 1 },
            acl_mod_dir_opt     => { type => HASHREF, optional => 1 },
            acl_mod_file_opt    => { type => HASHREF, optional => 1 },
            acl_del_dir_opt     => { type => HASHREF, optional => 1 },
            acl_del_file_opt    => { type => HASHREF, optional => 1 },
            sattrs_dir_opt      => { type => HASHREF, optional => 1 },
            sattrs_file_opt     => { type => HASHREF, optional => 1 },
            acl_opts            => { type => HASHREF, optional => 1 },
            sparse_opts         => { type => HASHREF, optional => 1 },
            sattrs_opts         => { type => HASHREF, optional => 1 },
            fill_vol            => { type => SCALAR,   default => 100 },
            timeout             => { type => SCALAR,   default => 1800 },
            dbfile              => { type => SCALAR,  default => "" },
            async               => { type => SCALAR, default => 0 }
        }
    );
    
    my (%dir_opts, %file_opts, %stream_opts, %args);
    my (%fcont_opts, %dcont_opts, $sparse_opts, $acl_opts, $sattrs_opts);
    my (%cont_opts, $size, $dev);
    my $dir_pct    = 100;
    my $file_pct   = 100;
    my $stream_pct = 100;
    
    my $client = $opts{command_interface};
    my ($obj, $dataset);
    my $async = delete $opts{async};

    
        # Set options according to template
    # Options in template are overwritten by explicitly setting the options
    $acl_opts    = $mod_template{$opts{template}}{acl_opts};
    $sparse_opts = $mod_template{$opts{template}}{sparse_opts};
    $sattrs_opts = $mod_template{$opts{template}}{sattrs_opts};
    %opts = ( %{$mod_template{$opts{template}}}, %opts, _mode => 'modify' );

    # ACL options
    $opts{acl_opts} = {
        # Values for ACL options from template
        %$acl_opts,
        # Overwrite defaults with user values
        %{$opts{acl_opts}},
    };

    # Sparse file options
    $opts{sparse_opts} = {
        # Values for sparse file options from template
        %$sparse_opts,
        # Overwrite defaults with user values
        %{$opts{sparse_opts}},
    };

    # Set attribute options
    $opts{sattrs_opts} = {
        # Values for set attribute options from template
        %$sattrs_opts,
        # Overwrite defaults with user values
        %{$opts{sattrs_opts}},
    };
    
    my $worker = NACL::STask::Client::RPC::Worker->new( client => $client );
    $dataset->{worker} = $worker;
    
    $worker->start();
    $worker->call( eval => 'require NACL::MTask::DataUtils::LocalOps;');
    my $obj_wrapper = $worker->call( eval => 'NACL::MTask::DataUtils::LocalOps->new(@_)',
                     args => [
                                path => $opts{path} ,
                                fillpct => $opts{fill_vol},
                                save    => 1,
                                create => 0,
                                seed    => _get_seed,
                                timeout => $opts{timeout},
                                dbfile  => $opts{dbfile},
                             ],
                      wrap_return => 1
                    );
    $dataset->{localops} = $obj_wrapper;
    
    bless $dataset, $self;
    
    my $future = $obj_wrapper->_rpc_submit(method => 'set_original');
    push @{$dataset->{jobs}}, $future;
    
    $dataset->_check_options(\%opts);
    _init_create_opts(\%opts);
    
    # set signal handler
    
    # Get main 'contain' attributes from options
    %cont_opts = map { $_ => $opts{$_} } grep { exists $opts{$_} } @cont_opts;
    
    # Setup both files and dirs 'contain' options
    %dcont_opts = %cont_opts;
    %fcont_opts = %cont_opts;
    delete $fcont_opts{dirs};
    delete $fcont_opts{files};
    delete $dcont_opts{sparse};
    
    # Options for removing directories which have no directories (dir leaves)
    %dir_opts = (
        _get_cont_opts(%dcont_opts, $opts{remove_dir_opt}),
        dirs     => 'no',
        number   => $opts{remove_dir_num},
        percent  => 100 * $opts{remove_dir_pct} / $dir_pct,
        original => 1,
        mode     => 'remove',
    );
    
    # Options for removing files
    %file_opts = (
        _get_cont_opts(%fcont_opts, $opts{remove_file_opt}),
        number   => $opts{remove_file_num},
        percent  => 100 * $opts{remove_file_pct} / $file_pct,
        original => 1,
        mode     => 'remove',
    );
    
    # Options for removing streams
    %stream_opts = (
        number   => $opts{remove_stream_num},
        percent  => 100 * $opts{remove_stream_pct} / $stream_pct,
        original => 1,
        mode     => 'remove',
    );
    
    $future = $obj_wrapper->_rpc_submit(method => 'modify', args => 
        [  dirs    => \%dir_opts,
           files   => \%file_opts,
           streams => \%stream_opts
        ]
    );
    push @{$dataset->{jobs}}, $future;
    
    # Calculate remaining percentage for next modify
    $dir_pct    -= $opts{remove_dir_pct};
    $file_pct   -= $opts{remove_file_pct};
    $stream_pct -= $opts{remove_stream_pct};
    $dir_pct     = 1 if $dir_pct    < 1;
    $file_pct    = 1 if $file_pct   < 1;
    $stream_pct  = 1 if $stream_pct < 1;
    
    # Options for modifying files (full file modify)
    %file_opts = (
        _get_cont_opts(%fcont_opts, $opts{modify_file_opt}),
        sparse   => 'no',
        number   => $opts{modify_file_num} / 2,
        percent  => 50 * $opts{modify_file_pct} / $file_pct,
        original => 1,
        mode     => 'modify',
        args     => {
            size     => _get_args(%opts, 'file_size_avg', 'modify_file_opt'),
            size_dev => _get_args(%opts, 'file_size_dev', 'modify_file_opt'),
        },
    );
    
    if ($client->hostrec->hosttype =~ /solaris/) {
        %args = (
            size     => _get_args(%opts, 'stream_size_avg','modify_stream_opt'),
            size_dev => _get_args(%opts, 'stream_size_dev','modify_stream_opt'),
        );
    } else {
        $size = _get_args(%opts, 'stream_size_avg', 'modify_stream_opt');
        %args = (
            blocks   => exists $opts{modify_stream_opt}{blocks} ?
                               $opts{modify_stream_opt}{blocks} : 4,
            size     => exists $opts{modify_stream_opt}{block_size} ?
                               $opts{modify_stream_opt}{block_size} : 1024,
            max_size => $size,
        );
    }
    
    %stream_opts = (
        number   => $opts{modify_stream_num} / 2,
        percent  => 50 * $opts{modify_stream_pct} / $stream_pct,
        original => 1,
        mode     => 'modify',
        args     => \%args,
    );
    
    $future = $obj_wrapper->_rpc_submit(method => 'modify', args => 
        [  files   => \%file_opts,
           streams => \%stream_opts
        ]
    );
    push @{$dataset->{jobs}}, $future;

    
    # Swap the modification args so
    # full stream modify and some blocks modify for files
    $file_opts{args} = {
        blocks   => exists $opts{modify_file_opt}{blocks} ?
                           $opts{modify_file_opt}{blocks} : 4,
        size     => exists $opts{modify_file_opt}{block_size} ?
                           $opts{modify_file_opt}{block_size} : 1024,
        max_size => _get_args(%opts, 'file_size_avg', 'modify_file_opt'),
    };
    
    $stream_opts{args} = {
        size     => _get_args(%opts, 'stream_size_avg','modify_stream_opt'),
        size_dev => _get_args(%opts, 'stream_size_dev','modify_stream_opt'),
    };
    
    # Make sure it modifies the correct number
    $file_opts{number}   += $opts{modify_file_num} % 2;
    $stream_opts{number} += $opts{modify_stream_num} % 2;
    
    # Modify files/streams
    $future = $obj_wrapper->_rpc_submit(method => 'modify', args => 
        [  files   => \%file_opts,
           streams => \%stream_opts
        ]
    );
    push @{$dataset->{jobs}}, $future;
    
        # Calculate remaining percentage for next modify
    $file_pct   -= $opts{modify_file_pct};
    $stream_pct -= $opts{modify_stream_pct};
    $file_pct    = 1 if $file_pct   < 1;
    $stream_pct  = 1 if $stream_pct < 1;

    # Options for modifying sparse files
    %file_opts = (
        _get_cont_opts(%fcont_opts, $opts{modify_sparse_opt}),
        sparse   => 'yes',
        number   => $opts{modify_sparse_num},
        percent  => $opts{modify_sparse_pct},
        original => 1,
        mode     => 'modify',
        args     => {
            sparse_opts => { %{$opts{sparse_opts}},
                             (exists $opts{modify_sparse_opt}{sparse_opts} ?
                              %{$opts{modify_sparse_opt}{sparse_opts}} : ()) }
        },
    );

    $future = $obj_wrapper->_rpc_submit(method => 'modify', args => 
        [  files   => \%file_opts,
        ]
    );
    push @{$dataset->{jobs}}, $future;

    # Options for renaming directories
    %dir_opts = (
        _get_cont_opts(%dcont_opts, $opts{rename_dir_opt}),
        number   => $opts{rename_dir_num},
        percent  => 100 * $opts{rename_dir_pct} / $dir_pct,
        original => 1,
        mode     => 'rename',
    );

    # Options for renaming files
    %file_opts = (
        _get_cont_opts(%fcont_opts, $opts{rename_file_opt}),
        number   => $opts{rename_file_num},
        percent  => 100 * $opts{rename_file_pct} / $file_pct,
        original => 1,
        mode     => 'rename',
    );

    # Modify files/streams
    $future = $obj_wrapper->_rpc_submit(method => 'modify', args => 
        [  files   => \%file_opts,
           dirs    => \%dir_opts
        ]
    );
    push @{$dataset->{jobs}}, $future;
    
    # Calculate remaining percentage for next modify
    $dir_pct  -= $opts{rename_dir_pct};
    $file_pct -= $opts{rename_file_pct};
    $dir_pct   = 1 if $dir_pct  < 1;
    $file_pct  = 1 if $file_pct < 1;

    # Options for truncating files
    %file_opts = (
        _get_cont_opts(%fcont_opts, $opts{truncate_file_opt}),
        number   => $opts{truncate_file_num},
        percent  => 100 * $opts{truncate_file_pct} / $file_pct,
        original => 1,
        mode     => 'trunc',
        args     => {
            size     => _get_args(%opts, 'file_size_avg', 'truncate_file_opt'),
            size_dev => _get_args(%opts, 'file_size_dev', 'truncate_file_opt'),
        },
    );

    # Options for truncating streams
    $size = _get_args(%opts, 'stream_size_avg', 'truncate_stream_opt');
    $dev  = _get_args(%opts, 'stream_size_dev', 'truncate_stream_opt');
    %stream_opts = (
        number   => $opts{truncate_stream_num},
        percent  => 100 * $opts{truncate_stream_pct} / $stream_pct,
        original => 1,
        mode     => 'trunc',
        args     => {
            size     => $size,
            size_dev => $dev,
        },
    );

    $future = $obj_wrapper->_rpc_submit(method => 'modify', args => 
        [  files   => \%file_opts,
           streams => \%stream_opts
        ]
    );
    push @{$dataset->{jobs}}, $future;
    
    # Calculate remaining percentage for next modify
    $file_pct   -= $opts{truncate_file_pct};
    $stream_pct -= $opts{truncate_stream_pct};
    $file_pct    = 1 if $file_pct   < 1;
    $stream_pct  = 1 if $stream_pct < 1;
    
        # Options for modifying ACL's from directories
    %dir_opts = (
        _get_cont_opts(%dcont_opts, $opts{acl_mod_dir_opt}),
        number   => $opts{acl_mod_dir_num},
        percent  => 100 * $opts{acl_mod_dir_pct} / $dir_pct,
        original => 1,
        mode     => 'acl_mod',
    );

    # Options for modifying ACL's from files
    %file_opts = (
        _get_cont_opts(%fcont_opts, $opts{acl_mod_file_opt}),
        number   => $opts{acl_mod_file_num},
        percent  => 100 * $opts{acl_mod_file_pct} / $file_pct,
        original => 1,
        mode     => 'acl_mod',
    );
    
    $future = $obj_wrapper->_rpc_submit(method => 'modify', args => 
        [  files   => \%file_opts,
           dirs => \%dir_opts
        ]
    );
    push @{$dataset->{jobs}}, $future;
    
        # Calculate remaining percentage for next modify
    $dir_pct  -= $opts{acl_mod_dir_pct};
    $file_pct -= $opts{acl_mod_file_pct};
    $dir_pct   = 1 if $dir_pct  < 1;
    $file_pct  = 1 if $file_pct < 1;

    # Options for removing ACL's from directories
    %dir_opts = (
        _get_cont_opts(%dcont_opts, $opts{acl_del_dir_opt}),
        number   => $opts{acl_del_dir_num},
        percent  => 100 * $opts{acl_del_dir_pct} / $dir_pct,
        original => 1,
        mode     => 'acl_del',
    );

    # Options for removing ACL's from files
    %file_opts = (
        _get_cont_opts(%fcont_opts, $opts{acl_del_file_opt}),
        number   => $opts{acl_del_file_num},
        percent  => 100 * $opts{acl_del_file_pct} / $file_pct,
        original => 1,
        mode     => 'acl_del',
    );

    $future = $obj_wrapper->_rpc_submit(method => 'modify', args => 
        [  files   => \%file_opts,
           dirs => \%dir_opts
        ]
    );
    push @{$dataset->{jobs}}, $future;

    # Calculate remaining percentage for next modify
    $dir_pct  -= $opts{acl_del_dir_pct};
    $file_pct -= $opts{acl_del_file_pct};
    $dir_pct   = 1 if $dir_pct  < 1;
    $file_pct  = 1 if $file_pct < 1;

    # Options for setting attributes on directories
    %dir_opts = (
        _get_cont_opts(%dcont_opts, $opts{sattrs_dir_opt}),
        number   => $opts{sattrs_dir_num},
        percent  => 100 * $opts{sattrs_dir_pct} / $dir_pct,
        original => 1,
        mode     => 'sattrs',
        args     => $opts{sattrs_opts},
    );

    # Options for setting attributes on files
    %file_opts = (
        _get_cont_opts(%fcont_opts, $opts{sattrs_file_opt}),
        number   => $opts{sattrs_file_num},
        percent  => 100 * $opts{sattrs_file_pct} / $file_pct,
        original => 1,
        mode     => 'sattrs',
        args     => $opts{sattrs_opts},
    );
    
    $future = $obj_wrapper->_rpc_submit(method => 'modify', args => 
        [  files   => \%file_opts,
           dirs => \%dir_opts
        ]
    );
    push @{$dataset->{jobs}}, $future;
    
        # Calculate remaining percentage for next modify
    $dir_pct  -= $opts{acl_del_dir_pct};
    $file_pct -= $opts{acl_del_file_pct};
    $dir_pct   = 1 if $dir_pct  < 1;
    $file_pct  = 1 if $file_pct < 1;

    # Options for adding new objects to directories
    %dir_opts = (
        %dcont_opts,
        number   => $opts{create_on_dir_num},
        percent  => 100 * $opts{create_on_dir_pct} / $dir_pct,
        original => 1,
        depth    => 0, # Maintain same depth
        mode     => 'create',
        args     => $opts{_dir_args},
    );

    # Options for adding new objects to files
    %file_opts = (
        %fcont_opts,
        number   => $opts{create_on_file_num},
        percent  => 100 * $opts{create_on_file_pct} / $file_pct,
        original => 1,
        mode     => 'create',
        args     => $opts{_file_args},
    );

    $future = $obj_wrapper->_rpc_submit(method => 'modify', args => 
        [  files   => \%file_opts,
           dirs => \%dir_opts
        ]
    );
    push @{$dataset->{jobs}}, $future;
    
    if (!$async) {
        $dataset->wait_for_completion( timeout => $opts{timeout} );
    }

    return $dataset;
}

=head2 wait_for_completion

Wait for the data population to finish on the remaining jobs

=over

=item Arguments

=over

=item C<< timeout => $timeout_value_in_seconds >>

(Required)

How long to wait for the remaining jobs to complete

=back

=back

=over

=item Returns

Nothing.

=back

=cut

sub wait_for_completion {
    my $self = shift;
    my %opts = @_;
    my $timeout = $opts{timeout} || 0;
    
    # get the status of the job
    # if done, then simply return
    # else, wait for the a maximum duration (specified by timeout)
    # and see if the jobs are done
    # if not, throw an exception
    
    my ($start_time, $end_time, $time_spent, $exception, $error_message, $e);
    $error_message = "";
    
    my @jobs = @{$self->_get_jobs()};

    foreach my $job (  @jobs ) {
        $start_time = time();
        $timeout = 0 if $timeout < 0;
        try {
        $job->result( timeout => $timeout,
                      on_timeout => sub { 
                        if(!$exception) {
                            $exception = "NACL::APISet::Exceptions::TimeoutException";
                            $error_message = "One or more operations timed".
                                             " out.\nLater operations would".
                                             " have timed out due to earlier ".
                                             "ones timing out. The List of".
                                             " jobs is:\n". join (",", @jobs).
                                             "\nThe jobs that have timed out".
                                             " are:\n"."$job";
                        } else {
                            $error_message .= ", $job";
                        }
                      }
                    );
        } otherwise {
            $e = shift;
            $error_message .= "\n$job failed because of: ".$e->text();
            $exception = "NACL::Exceptions::OperationFailed" if !$exception;
        };
        $end_time = time();
        $time_spent = $end_time - $start_time;
        
        # calculate the remaining timeout
        $timeout = $timeout - $time_spent;
    }
    
    $self->{jobs} = [];
    $self->kill();
    $exception->throw($error_message) if $exception;
}

=head2 kill

Kill the data population that is in progress
Nothing will happen if the job is already complete

=over

=item Arguments

None

=back

=over

=item Returns

Nothing

=back

=cut

sub kill {
    my $self = shift;
    $self->{worker}->kill();
}

# Internal methods and functions

# Internal method to get the list of jobs that were submitted

sub _get_jobs {
    my $self = shift;
    return $self->{jobs};
}

# Common starter options
# Initialize all hashes for the create options for both
# the populate and modify

sub _init_create_opts {
    my $opts = shift;

    # Name options
    $opts->{_name_args} = {
        name_size     => $opts->{name_avg},
        name_size_dev => $opts->{name_dev},
    };

    # Creating streams option hash
    $opts->{_stream_args} = {
        number => $opts->{stream_avg},
        dev    => $opts->{stream_dev},
        args   => {
            %{$opts->{_name_args}},
            size     => $opts->{stream_size_avg},
            size_dev => $opts->{stream_size_dev},
        }
    };

    # Creating files option hash
    $opts->{_file_args} = {
        number  => $opts->{file_avg},
        dev     => $opts->{file_dev},
        streams => $opts->{_stream_args},
        args    => {
            %{$opts->{_name_args}},
            %{$opts->{_attr_args}},
            %{$opts->{acl_opts}},
            size        => $opts->{file_size_avg},
            size_dev    => $opts->{file_size_dev},
            acl_num     => $opts->{file_acl_num},
            acl_pct     => $opts->{file_acl_pct},
            stream_num  => $opts->{file_stream_num},
            stream_pct  => $opts->{file_stream_pct},
            sparse_num  => $opts->{sparse_num},
            sparse_pct  => $opts->{sparse_pct},
            sparse_opts => $opts->{sparse_opts},
            link_num    => $opts->{link_num},
            link_pct    => $opts->{link_pct},
            slink_num   => $opts->{slink_num},
            slink_pct   => $opts->{slink_pct},
        }
    };

    # Creating directories option hash
    $opts->{_dir_args} = {
        number  => $opts->{dir_avg},
        dev     => $opts->{dir_dev},
        files   => $opts->{_file_args},
        streams => $opts->{_stream_args},
        args    => {
            %{$opts->{_name_args}},
            %{$opts->{_attr_args}},
            %{$opts->{acl_opts}},
            acl_num    => $opts->{dir_acl_num},
            acl_pct    => $opts->{dir_acl_pct},
            stream_num => $opts->{dir_stream_num},
            stream_pct => $opts->{dir_stream_pct},
        }
    };

    if ( $opts->{_mode} eq 'populate' ) {
        # Add the depth options for populate
        $opts->{_dir_args}{depth_avg} = $opts->{depth_avg},
        $opts->{_dir_args}{depth_dev} = $opts->{depth_dev},
    }

    return;
}

# Internal function to obtain a seed to send to the constructor 
# of the LocalOps library

sub _get_seed() {
    my $seed;

    # Get seed to use
    if ( defined($_seed) ) {
        # One from the package already been seeded
        $seed = $_seed + 1;
    } elsif ( defined($main::DATA_SEED) ) {
        # From the environment
        $seed = $main::DATA_SEED;
    } else {
        # Create a new seed
        do {
            $seed = time() ^ ($$ + ($$ << 15));
        } while (!$seed);
    }
    $_seed = $seed;
    return $seed;
}

# Internal function to ignore unsupported options and 
# generate a warning if they are used

sub _check_options {
    my $self = shift;
    my $opts = shift;
    my $data = $self->{localops};
    my ($vol_info, $vol);
    my $streams = 0;
    my %flags = (
        acls     => 0,
        sparse   => 1,
        encrypt  => 0,
        compress => 0,
    );
    
    my $hosttype = $data->_rpc_call(method => 'hosttype');

    if ( $hosttype =~ /windows/ ) {
        # Streams are supported
        $streams = 1;
        # Get volume information
        $vol = $opts->{path};
        $vol = getcwd() if $vol !~ /^.:/;
        $vol =~ s/:.*/:/;
        my $worker=$self->{worker};
        #my $worker = NACL::STask::Client::RPC::Worker->new( client => $opts->{command_interface}->hostrec() );
        $worker->call( eval => 'require NACL::MTask::DataUtils::FileManip;');
        $vol_info = $worker->call( eval => 'NACL::MTask::DataUtils::FileManip::vol_info(@_)',
                     args => [
                                vol => $vol
                             ],
                    );
        %flags = %{$vol_info->{sysflags}};
    } else {
        # Check if path supports streams
        $streams = $self->_streams_supported(opts => $opts, hosttype => $hosttype);
    }

    if(!$opts->{command_interface}->hostrec()->is_unix()){
        # Ignore symbolic link option if not unix
        $opts->{slink_pct} = 0;
        $Log->debug("Symbolic link option is ignored: Symbolic links are not ".
                    "supported on " . $hosttype);
    }

    if ( ! $streams ) {
        # Ignore stream options if OS does not support streams
        ( $opts->{$_} = 0 ) foreach grep { /stream_pct$/ } keys %$opts;
        $Log->debug("Stream options are ignored: streams are not supported on ".
                     "path");
    }

    if ( ! $flags{acls} ) {
        # Ignore ACL options, it is only supported on windows
        ( $opts->{$_} = 0 ) foreach grep { /acl_pct$/ } keys %$opts;
        ( $opts->{$_} = 0 ) foreach grep { /^acl_.*_pct$/ } keys %$opts;
        $Log->debug("ACL options are ignored: ACLs are not supported on ".
                   $hosttype !~ /windows/ ? $hosttype : 'volume');
    }

    if ( ! $flags{sparse} ) {
        $opts->{sparse_pct} = 0;
        $Log->debug('Sparse options are ignored: '.
                    'Sparse files are not supported on volume');
    }

    if ( $hosttype !~ /windows/ ) {
       #Ignore Windows attribute options, it is only supported on windows
        $opts->{_attr_args} = {};
        ( $opts->{$_} = 0 ) foreach grep { /^sattrs_.*_pct$/ } keys %$opts;
        ( $opts->{$_} = 0 ) foreach grep { /^.*_attr_pct$/ } keys %$opts;
        $Log->debug('Windows attribute options are ignored: Windows '.
                    'attributes are not supported on ' . $hosttype);
    } else {
        if ( ! $flags{encrypt} ) {
            $opts->{encrypt_pct} = 0;
            delete $opts->{sattrs_opts}{encrypt};
            $Log->debug('Encrypt options are ignored: '.
                        'Encrypt is not supported on volume');
        }

        if ( ! $flags{compress} ) {
            $opts->{compress_pct} = 0;
            delete $opts->{sattrs_opts}{compress};
            $Log->debug('Compress options are ignored: ' .
                         'Compress is not supported on volume');
        }

        # Windows attribute options
        $opts->{_attr_args} = {
            archive_num  => $opts->{archive_num},
            archive_pct  => $opts->{archive_pct},
            hidden_num   => $opts->{hidden_num},
            hidden_pct   => $opts->{hidden_pct},
            encrypt_num  => $opts->{encrypt_num},
            encrypt_pct  => $opts->{encrypt_pct},
            compress_num => $opts->{compress_num},
            compress_pct => $opts->{compress_pct},
        };
    }

    return;
}

# Internal function to check if streams are supported for 
# the given client and the given path

sub _streams_supported {
    my $self = shift;
    my %args = @_;
    my $data = $self->{localops};
    my ($opts, $hosttype) = ($args{opts}, $args{hosttype});
    my ($opt_pct, $opt_num, $ver, $rel, $path, $dobj, $err);
    my $streams = 0;

    if ( $hosttype =~ /solaris/ && ($opts->{_mode} eq 'read'
      || $opts->{_mode} eq 'populate' || $opts->{_mode} eq 'modify') ) {
        # Check if streams need to be created/modified
        foreach $opt_pct (grep { /stream_pct$/ } keys %$opts) {
            ($opt_num = $opt_pct) =~ s/pct$/num/;
            if ( $opts->{$opt_pct} > 0 && $opts->{$opt_num} > 0 ) {
                # Options are set to create streams
                $streams = 1;
                last;
            }
        }
        if ( $streams ) {
            # Check if streams are supported
            ($ver, $rel) = ((POSIX::uname)[2] =~ /(\d+)\.(\d+)/);
            $ver .= sprintf("%03d", $rel);
            $streams = 0;
            if ($ver >= 5010) {
                # This is Solaris 5.8 or later which supports streams,
                # now check if path actually supports streams
                $path = "$opts->{path}/.test_streams.$$";
                Log->debug('Trying to create a stream to check whether path ' .
                    'supports streams');

                # Create the data manipulation object
                my $client = $opts->{client};
                my $worker = NACL::STask::Client::RPC::Worker->new( client => $client );
                $worker->call(eval => 'require NACL::MTask::DataUtils::LocalOps;');
                my $obj_wrapper = $worker->call( eval => 'NACL::MTask::DataUtils::LocalOps->new(@_)',
                     args => [
                                path => $opts->{path} ,
                                fillpct => $opts->{fill_vol},
                                save    => 1,
                                seed    => _get_seed,
                                timeout => $opts->{timeout},
                                dbfile  => $opts->{dbfile},
                             ],
                      wrap_return => 1
                    );
                my $dataset;
                $dataset->{localops} = $obj_wrapper;
                bless $dataset, ref $self;
                
                $obj_wrapper->_rpc_call(method => 'create_files', args =>[ 
                                                                        number => 1,
                                                                         dev    => 0,
                                                                         streams => {
                                                                            number => 1,
                                                                            dev    => 0,
                                                                         },
                                                                        ] 
                );
                $Log->debug('Now removing the stream');
                $worker->call(eval => 'NACL::MTask::DataUtils::FileManip::remove_file(@_)', args => 
                                                                [
                                                                 file => $path,
                                                                 recursive => 1
                                                                ] 
                              );
                
                $worker->kill();
                $streams = 1;
            }
        }
    }
    return $streams;
}

sub _get_cont_opts(\%$;) {
    my ($cont_opts, $opts) = @_;

    return (
        %$cont_opts,
        map { $_ => $opts->{$_} } grep { exists $opts->{$_} } @cont_opts,
    );
}

sub _get_args(\%$$;) {
    my ($opts, $key, $ind_opt) = @_;

    return exists $opts->{$ind_opt}{$args_map{$key}} ?
                  $opts->{$ind_opt}{$args_map{$key}} : $opts->{$key};
}

sub _create_none(\%;) {
    my $template = shift;
    my $item;
    my @list = qw(dir file depth);

    # Copy the default options to the 'none' template
    $template->{none} = { %{$template->{default}} };

    foreach $item ( keys %{$template->{none}} ) {
        $template->{none}{$item} = 0 if $item =~ /_(num|pct)$/;
    }

    foreach $item ( @list ) {
        if ( exists $template->{none}{"${item}_avg"} ) {
            $template->{none}{"${item}_avg"} = 0;
            $template->{none}{"${item}_dev"} = 0;
        }
    }
}

#---------------------------------------------------------------------
# Common options for the default templates for populate and modify
# Refer to tools/dp/cifs_tools_design.txt for more detailed information
# on the parameters in the templates below.
#---------------------------------------------------------------------
my %_pop_mod_opts = (
    dir_avg         => 3.2,
    dir_dev         => 0.2,
    file_size_avg   => 32768,
    file_size_dev   => 16384,
    name_avg        => 8,
    name_dev        => 3,
    stream_size_avg => 10000,
    stream_size_dev => 5000,

    dir_stream_num  => 99999999999,
    dir_stream_pct  => 3,
    file_stream_num => 99999999999,
    file_stream_pct => 3,

    link_num        => 99999999999,
    link_pct        => 1,
    slink_num       => 99999999999,
    slink_pct       => 1,
    compress_num    => 99999999999,
    compress_pct    => 5,
    encrypt_num     => 99999999999,
    encrypt_pct     => 5,
    archive_num     => 0,
    archive_pct     => 0,
    hidden_num      => 99999999999,
    hidden_pct      => 5,
    sparse_num      => 99999999999,
    sparse_pct      => 1,
    dir_acl_num     => 99999999999,
    dir_acl_pct     => 4,
    file_acl_num    => 99999999999,
    file_acl_pct    => 4,

    acl_opts        => { acl=>1 },
    sparse_opts => {
        ndata      => 3,
        ndata_dev  => 2,
        offset     => 1000000,
        offset_dev => 200000,
        size       => 4096,
        size_dev   => 1024,
    },
);
my %_common_cont_opts = (
    dirs     => 'any',
    files    => 'any',
    sparse   => 'any',
    streams  => 'any',
    compress => 'any',
    encrypt  => 'any',
    archive  => 'any',
    hidden   => 'any',
    acl      => 'any',
);


#---------------------------------------------------------------------
# Populate templates
#---------------------------------------------------------------------
%pop_template = (
    default => {
        %_pop_mod_opts,
        file_avg   => 40,
        file_dev   => 2,
        stream_avg => 5,
        stream_dev => 0.5,
        depth_avg  => 4.3,
        depth_dev  => 0.2,
    },
);
_create_none(%pop_template);
$pop_template{small} = {
    %{$pop_template{default}},
    depth_avg     => 3,
    file_avg      => 20,
    file_size_avg => 8192,
    file_size_dev => 2048,
};

#---------------------------------------------------------------------
# Modify templates
#---------------------------------------------------------------------
%mod_template = (
    default => {
        %_pop_mod_opts,
        %_common_cont_opts,
        file_avg   => 8,
        file_dev   => 2,
        stream_avg => 0.5,
        stream_dev => 0.5,

        create_on_dir_num   => 99999999999,
        create_on_dir_pct   => 20,
        create_on_file_num  => 99999999999,
        create_on_file_pct  => 20,
        modify_file_num     => 99999999999,
        modify_file_pct     => 20,
        modify_stream_num   => 99999999999,
        modify_stream_pct   => 20,
        modify_sparse_num   => 99999999999,
        modify_sparse_pct   => 20,
        remove_dir_num      => 99999999999,
        remove_dir_pct      => 4,
        remove_file_num     => 99999999999,
        remove_file_pct     => 4,
        remove_stream_num   => 99999999999,
        remove_stream_pct   => 4,
        rename_dir_num      => 99999999999,
        rename_dir_pct      => 5,
        rename_file_num     => 99999999999,
        rename_file_pct     => 5,
        truncate_file_num   => 99999999999,
        truncate_file_pct   => 4,
        truncate_stream_num => 99999999999,
        truncate_stream_pct => 4,
        acl_mod_dir_num     => 99999999999,
        acl_mod_dir_pct     => 5,
        acl_mod_file_num    => 99999999999,
        acl_mod_file_pct    => 5,
        acl_del_dir_num     => 99999999999,
        acl_del_dir_pct     => 5,
        acl_del_file_num    => 99999999999,
        acl_del_file_pct    => 5,
        sattrs_dir_num      => 99999999999,
        sattrs_dir_pct      => 5,
        sattrs_file_num     => 99999999999,
        sattrs_file_pct     => 5,

        sattrs_opts => {
            compress => 1,
            encrypt  => 1,
            hidden   => 1,
        },
    },
);
_create_none(%mod_template);
$mod_template{small} = {
    %{$mod_template{default}},
    file_avg           => 4,
    file_dev           => 1,
    create_on_dir_pct  => 10,
    create_on_file_pct => 10,
    remove_dir_pct     => 15,
    remove_file_pct    => 15,
};


1;
