#
# Copyright (c) 2014 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

package NACL::MTask::DataSet::FileBoundaries;

use Moose;
use MTF::MooseCompatibleException qw(:all);

extends 'NACL::MTask::DataSet';

use Data::Dumper;

use NACL::Exceptions::UnexpectedState;
use NACL::C::Exceptions::VolumeFile::FileTooLarge;
use NACL::C::VolumeFile;

use constant {
    ZERO_SIZE => 0,
    MAX_INLINE_SIZE => 64,
};

=head1 NAME

NACL::MTask::DataSet::FileBoundaries

=head1 SYNOPSIS

    my $fileboundaries_dataset = NACL::MTask::DataSet::FileBoundaries->create(
        volume => $volume,
        clients => [ @clients ],
    );

    $fileboundaries_dataset = NACL::MTask::DataSet::FileBoundaries->find(
        volume => $volume,
        clients => [ @clients ],
    );

    $fileboundaries_dataset->validate();
    $fileboundaries_dataset->purge();

=head1 DESCRIPTION

Please read L<NACL::MTask::DataSet|lib-NACL-MTask-DataSet-pm> to understand the
general behavior and capabilities of DataSet libraries.

The FileBoundaries DataSet contains files that line up on interesting WAFL
boundaries.

=head1 ATTRIBUTES

All attributes available from NACL::MTask::DataSet are available on this derived
class.

=cut

=head1 METHODS

Please read L<NACL::MTask::DataSet|lib-NACL-MTask-DataSet-pm> to understand the
general behavior and capabilities of DataSet libraries. Any methods defined on
C<NACL::MTask::DataSet> are applicable to this DataSet.

=head2 file_max_size

Returns a L<NACL::C::VolumeFile|lib-NACL-C-VolumeFile-pm> instance of a file 
that has the maximum file size.

=head2 file_zero_size

Returns a L<NACL::C::VolumeFile|lib-NACL-C-VolumeFile-pm> instance of a file 
that has zero size.

=head2 file_max_inline_size

Returns a L<NACL::C::VolumeFile|lib-NACL-C-VolumeFile-pm> instance of a file 
that is the maximum size for an inline file.

=head2 validate

Validate will ensure the files have the proper size. It will also verify the
maximum file still has the last data written to it.


Throws C<NACL::Exceptions::UnexpectedState> or 
C<NACL::Exceptions::VerifyFailure> if the validation fails.

=cut

has '_max_file_content' => (
    is => 'ro',
    isa => 'Str',
    default => '7E57CA5E',
);

sub create {
    my ($self_or_pkg, %args) = @_;
    my $self = $self_or_pkg->_new_if_required(%args);

    my $dataset_directory = $self->_create_dataset_directory();

    # This gets a NACL component for a file that doesn't yet exist. It will be
    # created soon.
    my $file_max_size = $self->file_max_size();

    $self->_probe_maximum_size($file_max_size);

    my $file_zero_size = $self->file_zero_size();
    $file_zero_size->write(
        content => '',
        offset => 0,
        allow_empty_content => 1
    );

    my $file_max_inline_size = $self->file_max_inline_size();
    $file_max_inline_size->write(
        content => 'Z'x MAX_INLINE_SIZE,
        offset => 0,
    );

    $self->_write_manifest_file();

    $self->_fixup_permissions();

    return $self;
}

sub _probe_maximum_size {
    my ($self, $file) = @_;

    my $state = {
        high => 2**63,
        low => 2,
    };
 
    my $spread;
    my $mid;
    # Do a binary search to find the maximum size of a file
    do {
        $spread = ($state->{high}-$state->{low});
        $mid = int($spread/2)+$state->{low};
        $Log->trace("spread: $spread, mid: $mid");
        try {
            $file->write(
                content => 'This is some data to be written',
                offset => $mid,
            );
            $state->{low} = $mid;
        } catch NACL::C::Exceptions::VolumeFile::FileTooLarge with_handler {
            my $exception = shift;
            $state->{high} = $mid;
            $exception->gather_diagnostics();
        };
    } while ($spread > 1);

    # Write a special identifier and the file_size into the last section of the
    # file so we can read that out later. This will let us validate the file
    # size has not changed and the contents of the block were preserved
    my $file_size = $file->state()->file_size();
    $file->write(
        path => $file->path(),
        content => $self->_max_file_content().$file_size,
        offset => $state->{low},
    );

    return $state->{low};
}

sub _check_file_size {
    my ($self, $file, $expected_size) = @_;

    my $file_state = $file->state();

    # Change to the following when burt795758 is fixed. 
    # $file->verify_state('file-size' => $expected_size);
    if ($file_state->file_size() != $expected_size) {
        NACL::Exceptions::UnexpectedState->throw(
            "Expected file ".$file->path()." to be $expected_size bytes, but ".
            "it was ".$file_state->file_size()." bytes "
        );
    }
}

sub _file_instance {
    my ($self, $name) = @_;
    my $dataset_directory = $self->_get_dataset_directories();

    return NACL::C::VolumeFile->new(
        command_interface => $dataset_directory->command_interface(),
        vserver => $dataset_directory->vserver(),
        path => $dataset_directory->path()."/$name",
    );
   
}

sub file_max_size {
    my ($self) = @_;
    return $self->_file_instance("file_max_size");
}

sub file_zero_size {
    my ($self) = @_;
    return $self->_file_instance("file_zero_size");
}

sub file_max_inline_size {
    my ($self) = @_;
    return $self->_file_instance("file_max_inline_size");
}

sub _validate_file_max_size {
    my ($self, $file) = @_;

    my $file_state = $file->state();
    my $file_size = $file_state->file_size();
    my $expected_content = $self->_max_file_content().$file_size;

    my $read_offset = $file_size - 4096;

    my $actual_content = $file->read(
        offset => $read_offset,
        length => 4096,
    );

    if ($actual_content !~ /$expected_content/) {
        NACL::Exceptions::UnexpectedState->throw(
            "Expected to find $expected_content in the last 4096 bytes of the ".
            "file, but $actual_content did not contain this"
        );
    }
}

sub validate {
    my ($self) = @_;

    $self->_check_file_size(
        $self->file_zero_size(),
        ZERO_SIZE,
    );

    $self->_check_file_size(
        $self->file_max_inline_size(),
        MAX_INLINE_SIZE,
    );

    $self->_validate_file_max_size(
        $self->file_max_size(),
    );

    # There is not much we can do to validate the maximum file size. Hardcoding
    # the value isn't a good idea. It is also possible that the DataSet will be
    # created on one release and then the system upgraded to a version that has
    # a larger limit.

    return $self;
}

=head1 REGRESSION TESTS

Regression tests for this library are located at
lib/NACL/UnitTest/MTask/DataSetFileBoundaries.thpl

=head1 SEE ALSO

Please read L<NACL::MTask::DataSet|lib-NACL-MTask-DataSet-pm> to understand the
general behavior and capabilities of DataSet libraries.

Additional background on the DataSet framework is available at
https://wikid.netapp.com/w/NACL/Tasks/DataSet

=cut

1;
