#
# Copyright (c) 2014 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

package NACL::MTask::DataSet::HardlinkBasic;

use Moose;
use MTF::MooseCompatibleException qw(:all);

extends 'NACL::MTask::DataSet';

use Data::Dumper;

use NACL::C::Volume;
use NACL::C::VolumeFile;
use NACL::C::VolumeDirectory;
use NACL::Exceptions::UnexpectedState;
use NACL::GeneralUtils qw(random_name_generator);

=head1 NAME

NACL::MTask::DataSet::HardlinkBasic

=head1 SYNOPSIS

    my $hardlink_basic_dataset = NACL::MTask::DataSet::HardlinkBasic->create(
        volume => $volume,
        clients => [ @clients ],
    );

    $hardlink_basic_dataset = NACL::MTask::DataSet::HardlinkBasic->find(
        volume => $volume,
        clients => [ @clients ],
    );

    $hardlink_basic_dataset->validate();
    $hardlink_basic_dataset->purge();

=head1 DESCRIPTION

Please read L<NACL::MTask::DataSet|lib-NACL-MTask-DataSet-pm> to understand the
general behavior and capabilities of DataSet libraries.

The HardlinkBasic DataSet contains files that have hardlinks. A number of
interesting files are present in the DataSet, such as inline files and cross
directory hardlinks. 

=head1 ATTRIBUTES

All attributes available from NACL::MTask::DataSet are available on this derived
class.

=head2 regular_file

A L<NACL::C::VolumeFile|lib-NACL-C-VolumeFile-pm> instance of a regular (not
inline) file that has hardlinks to it.

=head2 regular_file_xdir

A L<NACL::C::VolumeFile|lib-NACL-C-VolumeFile-pm> instance of a regular (not
inline) file that has cross directory hardlinks to it.

=head2 inline_file

A L<NACL::C::VolumeFile|lib-NACL-C-VolumeFile-pm> instance of an inline file
that fits inside a single inode that has hardlinks to it.

=head2 inline_file_xdir

A L<NACL::C::VolumeFile|lib-NACL-C-VolumeFile-pm> instance of an inline file
that fits inside a single inode that has cross directory hardlinks to it.

=cut

my @File_Methods = qw(regular_file regular_file_xdir inline_file
    inline_file_xdir);

has 'regular_file' => (
    is => 'rw',
    isa => 'Object',
);

has 'regular_file_xdir' => (
    is => 'rw',
    isa => 'Object',
);

has 'inline_file' => (
    is => 'rw',
    isa => 'Object',
);

has 'inline_file_xdir' => (
    is => 'rw',
    isa => 'Object',
);

# Private attributes
has 'num_links' => (
    is => 'rw',
    isa => 'Int',
);

has 'junction_path_cache' => (
    is => 'rw',
    isa => 'HashRef',
    default => sub { return {}; },
);

=head1 METHODS

Please read L<NACL::MTask::DataSet|lib-NACL-MTask-DataSet-pm> to understand the
general behavior and capabilities of DataSet libraries. Any methods defined on
C<NACL::MTask::DataSet> are applicable to this DataSet.

=head2 validate

The validation provided is that the names have at least the number of hardlinks
they originally had. An exact match is not done to allow other users to make
additional hardlinks without breaking the validation. An I2P lookup is done and
the number of names is checked to see that it is equal or greater than the
minimum number of links.

Throws C<NACL::Exceptions::UnexpectedState> if the validation fails.

=cut

sub create {
    my ($self_or_pkg, %args) = @_;
    my $self = $self_or_pkg->_new_if_required(%args);

    my $dataset_directory = $self->_create_dataset_directory();

    my $links_to_make = $self->fast_mode() ? 4 : 128;
    # The number of links on the file will be 1 more than the number we make
    # this is because a file starts out with nlink==1
    $self->num_links($links_to_make+1);

    # Make an inline file w/ links 
    my $inline_file = $self->_file_for_target($dataset_directory);
    $inline_file->write(
        content => '8BADF00D',
    );
    $self->inline_file($inline_file);
    
    $self->_make_links(
        $inline_file,
        $links_to_make
    );
    # Make a regular file w/ links

    my $regular_file = $self->_file_for_target($dataset_directory);
    $regular_file->write(
        content => '8BADF00D'x1024,
    );
    $self->regular_file($regular_file);
    
    $self->_make_links(
        $regular_file,
        $links_to_make
    );

    # Make an inline file w/ xdir links
    my $inline_file_xdir = $self->_file_for_target($dataset_directory);
    $inline_file_xdir->write(
        content => '8BADF00D',
    );
    $self->inline_file_xdir($inline_file_xdir);
    
    $self->_make_xdir_links(
        $inline_file_xdir,
        $links_to_make
    );

    # Make a regular file w/ xdir links
    my $regular_file_xdir = $self->_file_for_target($dataset_directory);
    $regular_file_xdir->write(
        content => '8BADF00D'x1024,
    );
    $self->regular_file_xdir($regular_file_xdir);

    $self->_make_xdir_links(
        $regular_file_xdir,
        $links_to_make,
    );

    $self->_write_manifest_file();

    $self->_fixup_permissions();

    return $self;
}


# Freeze and thaw have been implemented to comply with the interface required
# for freeze/thaw. The concern is that we must not capture the volume name in
# the frozen information because the DataSet may be replicated to another volume
# via SnapMirror/SnapVault. The path to the file relative to the root of the
# volume is stored in the bundle. 
sub freeze {
    my ($self) = @_;

    my $bundle;
    foreach my $method (@File_Methods) {
        my $file = $self->$method();

        $bundle->{$method} = $self->_file_to_volume_relative_path($file);
    }

    $bundle->{num_links} = $self->num_links();

    return $bundle;
}

sub thaw {
    my ($self, $thaw_bundle) = @_;

    my $volume = $self->volume();

    foreach my $method (@File_Methods) {
        my $relative_path = $thaw_bundle->{$method};

        $self->$method(
            $self->_volume_relative_path_to_file($volume,$relative_path)
        );
    }

    $self->num_links($thaw_bundle->{num_links});

}

# The validation performed is described in the POD. All of these validations are
# read-only. When read-write validations are supported, we can also create
# additional links and do rename with victim to shake the I2P B+ tree.
sub validate {
    my ($self) = @_;

    foreach my $file (
        $self->regular_file(),      $self->inline_file(),
        $self->regular_file_xdir(), $self->inline_file_xdir()
        ) {
        my $file_state = $file->state();
        
        if ($file_state->hard_links_count < $self->num_links()) {
            NACL::Exceptions::UnexpectedState->throw("Expected file ".
                $file->path()." to have at least ".
                $self->num_links()." links but it had ".
                $file_state->hard_links_count()." links "
            );
        }

        $self->_validate_i2p($file);
    }

    return $self;
}

# Function: _file_to_volume_relative_path
# Arguments: 
#   [0] - NACL::C::VolumeFile instance of the file to get a volume relative
#   path for.
# Returns: path to the file relative to the root of the volume
sub _file_to_volume_relative_path {
    my ($self, $file) = @_;

    # Avoid duplicate / characters in the path
    my $trim_path = $file->path();
    $trim_path =~ s/(\/){2,}/\//g;

    my @components = split '/', $trim_path;
    # Remove '', 'vol', and 'volume_name'
    shift @components;
    shift @components;
    shift @components;

    return join '/', @components;
}

# Function: _volume_relative_path_to_file
#   This function is basically the inverse of _file_to_volume_relative_path
# Arguments:
#   [0] - NACL::C::Volume instance of the volume the file is on.
#   [1] - Relative path to the file on the volume
# Returns:
#   NACL::C::VolumeFile instance for the file
sub _volume_relative_path_to_file {
    my ($self, $volume, $relative_path) = @_;

    return NACL::C::VolumeFile->new(
        command_interface => $volume->command_interface(),
        vserver => $volume->vserver(),
        path => "/vol/".$volume->volume()."/$relative_path",
    );
}

# Function: _hardlink_for_file
# Arguments:
#   [0] - NACL::C::VolumeFile instance for the file to hardlink to
#   [1] - NACL::C::VolumeDirectory instance to put the link in
# Returns:
#   NACL::C::VolumeFile instance of the newly created hardlink
sub _hardlink_for_file {
    my ($self, $target_file, $directory) = @_;

    my $systemshell = $target_file->command_interface->apiset(
        category => 'Node',
        interface => 'CLI',
        set => 'Systemshell',
    );

    my $link_file = $self->_file_for_link($directory);

    $systemshell->link(
        files => $self->_file_to_systemshell_path($target_file),
        'link-name' => $self->_file_to_systemshell_path($link_file),
        'privilege-level' => 'root',
    );

    return $link_file;
}

# Function: _junction_path_for_volume
#   Memoized return of junction-path to prevent repeated lookups. Reduces the
#   time spent to create many hardlinks by 2x.
# Arguments:
#   [0] - NACL::C::Volume instance for the volume to get a junction-path for
# Returns:
#   Junction path to the volume
sub _junction_path_for_volume {
    my ($self, $volume) = @_;

    my $key = $volume->vserver().":".$volume->volume();

    if (!$self->junction_path_cache()->{$key}) {
        $self->junction_path_cache()->{$key} =
            $volume->state()->junction_path();
    }

    return $self->junction_path_cache()->{$key};

}

sub _file_to_systemshell_path {
    my ($self, $file) = @_;

    my $volume = $file->get_containing_volume_obj();
    my $junction_path = $self->_junction_path_for_volume($volume);
    my $file_path = $self->_file_to_volume_relative_path($file);

    return "/clus/".$volume->vserver()."/".$junction_path."/".$file_path;
}

# Function: _file_for_target
#   Return a VolumeFile instance for a yet to be created file
# Arguments:
#   [0] - NACL::C::VolumeDirectory instance to instantiate file in
# Returns:
#   A NACL::C::VolumeFile instance for a file yet to be created and to be used
#   as a link target.
sub _file_for_target {
    my ($self,$directory) = @_;

    return NACL::C::VolumeFile->new(
        command_interface => $directory->command_interface(),
        vserver => $directory->vserver(),
        path => $directory->path()."/file_with_links".
            random_name_generator(size => 3),
    );
}

# Function: _file_for_link
#   Return a VolumeFile instance for a yet to be created file
# Arguments:
#   [0] - NACL::C::VolumeDirectory instance to instantiate file in
# Returns:
#   A NACL::C::VolumeFile instance for a file yet to be created and to be used
#   as a link
sub _file_for_link {
    my ($self, $directory) = @_;

    my $file = $self->_file_for_target($directory);
    $file->path($file->path()."_link");
    return $file;
}

# Function:: _make_links
#   Create a number of hardlinks to a file
# Arguments:
#   [0] - NACL::C::VolumeFile instance of the file to link to.
#   [1] - The number of links to make to the file;
# Returns:
#   An array of NACL::C::VolumeFile instances for the hardlinks created.

sub _make_links {
    my ($self, $file, $links_to_make) = @_;

    my $dataset_directory = $self->_get_dataset_directories();
    my @links;
    foreach my $i (1..$links_to_make) {
        push @links, $self->_hardlink_for_file($file,$dataset_directory);
    }
    return @links;
}

# Function:: _make_xdir_links
#   Create a number of cross directory hardlinks to a file
# Arguments:
#   [0] - NACL::C::VolumeFile instance of the file to link to.
#   [1] - The number of links to make to the file;
# Returns:
#   An array of NACL::C::VolumeFile instances for the hardlinks created.
sub _make_xdir_links {
    my ($self, $file, $links_to_make) = @_;

    my $dataset_directory = $self->_get_dataset_directories();
    my @links;

    foreach my $i (1..$links_to_make) {
        my $subdir = NACL::C::VolumeDirectory->create(
            command_interface => $dataset_directory->command_interface(),
            vserver => $dataset_directory->vserver(),
            path => $dataset_directory->path()."/".
                random_name_generator(size => 3),
            perm => '0777',
            determine_appropriate_nodescope_node => 1,
        );
        push @links, $self->_hardlink_for_file($file,$subdir);
    }

    return @links;
}


# Function: _validate_i2p
#   Make sure I2P lookup returns names and the minimum number of names are
#   returned.
# Arguments: 
#   [0] - NACL::C::VolumeFile instance of the file to validate I2P information
#   for.
# Returns:
#   instance of object
# Throws:
#   NACL::Exceptions::UnexpectedState if validation fails.
sub _validate_i2p {
    my ($self, $file) = @_;

    my $file_path = $self->_file_to_volume_relative_path($file);
    my $inode_info = NACL::C::Volume->explore(
        command_interface => $file->command_interface(),
        format => 'inode',
        scope => $file->get_containing_volume_obj()->volume()."./".$file_path,
    );
    if (scalar($inode_info->name()) < $self->num_links()) {
            NACL::Exceptions::UnexpectedState->throw("Expected file ".
            $file->path()." to have at least ".
            $self->num_links()." names but it had ".
            scalar($inode_info->name())." names"
        );
    }
    return $self;
}

=head1 REGRESSION TESTS

Regression tests for this library are located at
lib/NACL/UnitTest/MTask/DataSetHardlinkBasic.thpl

=head1 SEE ALSO

Please read L<NACL::MTask::DataSet|lib-NACL-MTask-DataSet-pm> to understand the
general behavior and capabilities of DataSet libraries.

Additional background on the DataSet framework is available at
https://wikid.netapp.com/w/NACL/Tasks/DataSet

=cut

1;
