#
# Copyright (c) 2014 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

package NACL::MTask::DataSet::LUNsInQtrees;

use Moose;
use MTF::MooseCompatibleException qw(:all);

extends 'NACL::MTask::DataSet::QtreeBasic';

use NACL::C::Lun;

=head1 NAME

NACL::MTask::DataSet::LUNsInQtrees

=head1 SYNOPSIS

    my $luns_in_qtrees_dataset = NACL::MTask::DataSet::LUNsInQtrees->create(
        volume => $volume,
        clients => [ @clients ],
    );

    $luns_in_qtrees_dataset = NACL::MTask::DataSet::LUNsInQtrees->find(
        volume => $volume,
        clients => [ @clients ],
    );

    $luns_in_qtrees_dataset->validate();
    $luns_in_qtrees_dataset->purge();

=head1 DESCRIPTION

Please read L<NACL::MTask::DataSet|lib-NACL-MTask-DataSet-pm> to understand the
general behavior and capabilities of DataSet libraries.

This DataSet extends
L<NACL::MTask::DataSet::QtreeBasic|lib-NACL-MTask-DataSet-QtreeBasic-pm>.
Please read the documentation for that library to fully understand how this
library works.

The LUNsInQtrees DataSet creates a few qtrees with LUNs contained inside of them.
At least one qtree of each security-style will be created. The LUNs will have a
minimal amount of data populated in them so they are not empty.

=head1 ATTRIBUTES

All attributes available from NACL::MTask::DataSet are available on this derived
class.

=head2 mixed_qtree_lun

A L<NACL::C::Lun|lib-NACL-C-Lun-pm> instance of LUN in a mixed security style
LUN

=head2 unix_qtree_lun

A L<NACL::C::Lun|lib-NACL-C-Lun-pm> instance of LUN in a unix security style
LUN

=head2 ntfs_qtree_lun

A L<NACL::C::Lun|lib-NACL-C-Lun-pm> instance of LUN in a ntfs security style
LUN

=cut

has 'mixed_qtree_lun' => (
    is => 'rw',
    isa => 'Object',
);

has 'unix_qtree_lun' => (
    is => 'rw',
    isa => 'Object',
);

has 'ntfs_qtree_lun' => (
    is => 'rw',
    isa => 'Object',
);

=head1 METHODS

Please read L<NACL::MTask::DataSet|lib-NACL-MTask-DataSet-pm> to understand the
general behavior and capabilities of DataSet libraries. Any methods defined on
C<NACL::MTask::DataSet> are applicable to this DataSet.

=head2 validate

The validation for LUNs makes sure that state can be called on the component
instance. 

Throws C<NACL::Exceptions::UnexpectedState> if the validation fails. (thrown by
base class)

=cut

sub create {
    my ($self_or_pkg, %args) = @_;

    my $self = $self_or_pkg->_new_if_required(%args);

    $self->SUPER::create(%args);

    my $lun_count = 3;

    foreach my $sec_style (@{$self->security_styles()}) {
        my $qtree_method_name = $sec_style."_qtree";
        my $qtree = $self->$qtree_method_name();

        my $lun_method_name = $sec_style."_qtree_lun";
        foreach my $lun_number (1..$lun_count) {
            $self->$lun_method_name(
                NACL::C::Lun->create(
                    command_interface => $qtree->command_interface(),
                    vserver => $qtree->vserver(),
                    volume => $qtree->volume(),
                    qtree => $qtree->qtree(),
                    lun => $self->_lun_name($sec_style,$lun_number),
                    size => '10MB',
                    ostype => 'linux',
                )
            );

            if (!$self->fast_mode()) {
                my $node_name = $self->volume()->get_local_node_name();

                my $apiset = $self->volume()->command_interface()->get_7m_or_nodescope_apiset();
                $apiset->filersio_asyncio_active(
                    'read-percentage' => 0,
                    'rand-percentage' => 0,
                    'blk-size' => '64k',
                    'start-byte' => 0,
                    'end-byte' => '1m',
                    'run-time' => 10,
                    'num-active' => 1,
                    # When burt795637 is implemented, replace this with
                    # filename => 
                    # $self->$lun_method_name()->cast_to_file()->construct_dblade_path()
                    filename => $self->volume()->construct_dblade_path()."/".
                        $self->$qtree_method_name()->qtree()."/".
                        $self->_lun_name($sec_style,$lun_number),
                    synchronous => 1,
                    'nodescope-node-name' => $node_name,
                );
            }
        }
    }

    $self->_write_manifest_file();
    return $self;
}

# This thaw method doesn't use anything from the bundle, but it does use the
# instance via _lun_name. It will instantiate component instances for the
# qtrees that are made accessible via the public interface. No freeze method is
# necessary, all required information can be obtained from the base class'
# thawed attributes.
sub thaw {
    my ($self, $thaw_manifest) = @_;

    $self->SUPER::thaw($thaw_manifest);

    foreach my $sec_style (@{$self->security_styles()}) {
        my $qtree_method_name = $sec_style."_qtree";
        my $lun_method_name = $sec_style."_qtree_lun";
        $self->$lun_method_name(
            NACL::C::Lun->new(
                command_interface => $self->volume()->command_interface(),
                vserver => $self->volume()->vserver(),
                path => "/vol/".$self->volume()->volume()."/".
                    $self->$qtree_method_name()->qtree()."/".
                    $self->_lun_name($sec_style,1),
            )
        );
    }
    return $self;
}

sub purge {
    my ($self) = @_;

    foreach my $sec_style (@{$self->security_styles()}) {
        my $qtree_method_name = $sec_style."_qtree";
        my $qtree = $self->$qtree_method_name();

        # Delete all LUNs in the qtree
        my @qtree_luns = NACL::STask::Lun->find(
            command_interface => $qtree->command_interface(),
            filter => {
                vserver => $qtree->vserver(),
                volume => $qtree->volume(),
                qtree => $qtree->qtree(),
            },
            allow_empty => 1,
        );
        foreach my $qtree (@qtree_luns) {
            $qtree->delete();
        }
    }

    $self->SUPER::purge();

    return $self;
}

sub validate {
    my ($self) = @_;

    $self->SUPER::validate();

    foreach my $sec_style (@{$self->security_styles()}) {
        my $lun_method_name = $sec_style."_qtree_lun";
        my $lun_state = $self->$lun_method_name()->state();
    }
    return $self;
}

sub _lun_name {
    my ($self, $sec_style, $lun_number) = @_;

    return $sec_style.'_lun_'.$self->instance()."_$lun_number";
}


=head1 REGRESSION TESTS

Regression tests for this library are located at
lib/NACL/UnitTest/MTask/DataSetLUNsInQtrees.thpl

=head1 SEE ALSO

Please read L<NACL::MTask::DataSet|lib-NACL-MTask-DataSet-pm> to understand the
general behavior and capabilities of DataSet libraries.

Additional background on the DataSet framework is available at
https://wikid.netapp.com/w/NACL/Tasks/DataSet

=cut

1;
