#
# Copyright (c) 2014 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

package NACL::MTask::DataSet::NTFSSecurityBasic;

use Moose;
use MTF::MooseCompatibleException qw(:all);

extends 'NACL::MTask::DataSet';

use Data::Dumper;

use NACL::CS::VserverSecurityFileDirectory;
use NACL::C::VolumeFile;
use NACL::Exceptions::UnexpectedState;
use NACL::MTask::VserverSecurityFileDirectory;

=head1 NAME

NACL::MTask::DataSet::NTFSSecurityBasic

=head1 SYNOPSIS

    my $ntfs_security_dataset = NACL::MTask::DataSet::NTFSSecurityBasic->create(
        volume => $volume,
        clients => [ @clients ],
    );

    $ntfs_security_dataset = NACL::MTask::DataSet::NTFSSecurityBasic->find(
        volume => $volume,
        clients => [ @clients ],
    );

    $ntfs_security_dataset->validate();
    $ntfs_security_dataset->purge();

=head1 DESCRIPTION

Please read L<NACL::MTask::DataSet|lib-NACL-MTask-DataSet-pm> to understand the
general behavior and capabilities of DataSet libraries.

The NTFSSecurityBasic DataSet contains files and directories that have NTFS
security descriptors. The ACLs will be configured to allow the test to access
the files.

=head1 ATTRIBUTES

All attributes available from NACL::MTask::DataSet are available on this derived
class.

=cut

=head1 METHODS

Please read L<NACL::MTask::DataSet|lib-NACL-MTask-DataSet-pm> to understand the
general behavior and capabilities of DataSet libraries. Any methods defined on
C<NACL::MTask::DataSet> are applicable to this DataSet.

=head2 file_with_acl

Returns a L<NACL::C::VolumeFile|lib-NACL-C-VolumeFile-pm> instance of a file 
that has an ACL.

=head2 file_in_dir_with_acl

Returns a L<NACL::C::VolumeFile|lib-NACL-C-VolumeFile-pm> instance of a file 
that is in a directory with an ACL but does not have an ACL.


Throws C<NACL::Exceptions::UnexpectedState> if the validation fails.

=cut

sub _apply_security_descriptor {
    my ($self, $file_or_dir) = @_;
    # Save and restore the preferences to work around burt791624, burt791630, 
    # and burt791637
    my %restore_pref = $file_or_dir->command_interface()->apiset_should();
    $file_or_dir->command_interface()->apiset_should(interface => 'CLI');
    NACL::MTask::VserverSecurityFileDirectory->set_ntfs_acl(
        command_interface => $file_or_dir->command_interface(),
        vserver => $file_or_dir->vserver(),
        path => [ $self->_file_to_vserver_relative_path($file_or_dir) ],
        ntfs_mode => 'propagate',
        dacl => [
            {
                account => 'everyone',
                'access-type' => 'allow',
                rights => 'full-control',
                'apply-to' => [ 'this-folder', 'sub-folders', 'files' ],
            }
        ],
        cleanup => 0,
    );
    $file_or_dir->command_interface()->apiset_should(%restore_pref);
}

sub create {
    my ($self_or_pkg, %args) = @_;
    my $self = $self_or_pkg->_new_if_required(%args);

    my $dataset_directory = $self->_create_dataset_directory();

    # This gets a NACL component for a file that doesn't yet exist. It will be
    # created soon.
    my $file_with_acl = $self->file_with_acl();
    $file_with_acl->write(
        content => 'small_file_content',
        offset => 0,
    );
    $self->_apply_security_descriptor($file_with_acl);

    my $dir_with_acl = $self->dir_with_acl();
    NACL::C::VolumeDirectory->create(
        command_interface => $dir_with_acl->command_interface(),
        vserver => $dir_with_acl->vserver(),
        path => $dir_with_acl->path(),
        perm => '777',
    );
    $self->_apply_security_descriptor($dir_with_acl);

    my $file_in_dir_with_acl = $self->file_in_dir_with_acl();
    $file_in_dir_with_acl->write(
        content => 'small_file_content',
        offset => 0,
    );
    
    $self->_write_manifest_file();

    $self->_fixup_permissions();

    return $self;
}

# burt689038 requests this to be added to NACL::C::VolumeFile
sub _file_to_vserver_relative_path {
    my ($self, $file) = @_;
    my $path = $file->path();
    my @components = split /\//, $path;
    shift @components; # ''
    shift @components; # 'vol'
    shift @components; # 'volumename'
    
    my $volume_relative_path = join '/', @components;
    
    return $file->get_containing_volume_obj()->state()->junction_path()."/".
        $volume_relative_path;
}

sub _file_instance {
    my ($self, $name) = @_;
    my $dataset_directory = $self->_get_dataset_directories();

    return NACL::C::VolumeFile->new(
        command_interface => $dataset_directory->command_interface(),
        vserver => $dataset_directory->vserver(),
        path => $dataset_directory->path()."/$name",
    );
   
}

sub file_with_acl {
    my ($self) = @_;
    return $self->_file_instance("file_with_acl");
}

sub dir_with_acl {
    my ($self) = @_;
    my $dataset_directory = $self->_get_dataset_directories();

    return NACL::C::VolumeDirectory->new(
        command_interface => $dataset_directory->command_interface(),
        vserver => $dataset_directory->vserver(),
        path => $dataset_directory->path()."/dir_with_acl",
    );
}

sub file_in_dir_with_acl {
    my ($self) = @_;
    my $dir_with_acl = $self->dir_with_acl();

    return NACL::C::VolumeFile->new(
        command_interface => $dir_with_acl->command_interface(),
        vserver => $dir_with_acl->vserver(),
        path => $dir_with_acl->path()."/file_in_dir_with_acl",
    );
}

sub _validate_file_has_applied_ace {
    my ($self, $file, $expected_ace_count) = @_;

    my $acl_state = NACL::CS::VserverSecurityFileDirectory->fetch(
        command_interface => $file->command_interface(),
        filter => {
            vserver => $file->vserver(),
            path => $self->_file_to_vserver_relative_path($file),
        },
        allow_empty => 1,
        # ZAPI parsing is completely broken for the ACLS field
        # burt795728
        apiset_must => { interface => 'CLI' },
    );
    my @acls = ();
    if ($acl_state) {
        @acls = $acl_state->acls();
    }
    my @not_inherited = grep(!/Inherited/i, @acls);
    my @allow_everyone = grep(/ALLOW-Everyone/i, @not_inherited);
    if (scalar(@allow_everyone) != $expected_ace_count) {
        NACL::Exceptions::UnexpectedState->throw(
            "ACL did not have a single ACE for ALLOW-Everyone, was ".
            Dumper(\@acls)
        );
    }
}

sub validate {
    my ($self) = @_;

    my $file_with_acl = $self->file_with_acl();
    $self->_validate_file_has_applied_ace($file_with_acl,1);

    my $file_in_dir_with_acl = $self->file_in_dir_with_acl();
    $self->_validate_file_has_applied_ace($file_in_dir_with_acl,0);

    my $dir_with_acl = $self->dir_with_acl();
    $self->_validate_file_has_applied_ace($dir_with_acl,1);

    return $self;
}

=head1 REGRESSION TESTS

Regression tests for this library are located at
lib/NACL/UnitTest/MTask/DataSetNTFSSecurityBasic.thpl

=head1 SEE ALSO

Please read L<NACL::MTask::DataSet|lib-NACL-MTask-DataSet-pm> to understand the
general behavior and capabilities of DataSet libraries.

Additional background on the DataSet framework is available at
https://wikid.netapp.com/w/NACL/Tasks/DataSet

=cut

1;
