#
# Copyright (c) 2014 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

package NACL::MTask::DataSet::SisCloneBasic;

use Moose;
use MTF::MooseCompatibleException qw(:all);

extends 'NACL::MTask::DataSet';

use NACL::Exceptions::UnexpectedState;
use data_mobility::ondemand::lib::MountManagerSingleton;
use NACL::STask::VolumeFile;
use NACL::GeneralUtils qw(random_name_generator);

has clone_cascading_depth => (
    is => 'rw',
    isa => 'Int',
    default => 6,
);

has change_buftree_shape => (
    is => 'rw',
    isa => 'Str',
    required => 0,
);

has sis_clone_fixup => (
    is => 'rw',
    isa => 'Str',
    default => 'enable',
);

=head1 NAME

NACL::MTask::DataSet::SisCloneBasic

=head1 SYNOPSIS

    my $sis_clone_dataset = NACL::MTask::DataSet::SisCloneBasic->create(
        volume => $volume,
        clients => [ @clients ],
        ...
    );

    $sis_clone_dataset = NACL::MTask::DataSet::SisCloneBasic->find(
        volume => $volume,
        clients => [ @clients ],
    );

    $sis_clone_dataset->validate();
    $sis_clone_dataset->purge();

=head1 DESCRIPTION

Please read L<NACL::MTask::DataSet|lib-NACL-MTask-DataSet-pm> to understand the
general behavior and capabilities of DataSet libraries.

The SisCloneBasic DataSet contains sis clone files. User can make the sis clone
data set rich by specifying cascading depth, buftree shape or disabling sis
clone fixup.

=head1 ATTRIBUTES

=head2 clone_cascading_depth

(OPTIONAL, Defaults to 6) This represents depth of cascading while creating
sis clones. (clone of clone of clone..)

=head2 change_buftree_shape

(OPTIONAL, Defaults to notDefined) User can specify this option to create sis
clones with different buftree shapes.

=head2 sis_clone_fixup

(OPTIONAL, Defaults to enable) This parameter can be used to enable or disable
clone split.

All attributes available from NACL::MTask::DataSet are available on this derived
class.

=cut

=head1 METHODS

Please read L<NACL::MTask::DataSet|lib-NACL-MTask-DataSet-pm> to understand the
general behavior and capabilities of DataSet libraries. Any methods defined on
C<NACL::MTask::DataSet> are applicable to this DataSet.

=head2 get_sisclone_files

Returns list of sis clone files which are instance of
L<NACL::STask::VolumeFile|lib-NACL-STask-VolumeFile-pm>

=head2 get_sisclone_parent_file

Returns the sis clone parent file which is an instance of
L<NACL::STask::VolumeFile|lib-NACL-STask-VolumeFile-pm>

=head2 disable_sis_clone_fixup

Disables the sis clone fixup by setting flag

=head2 enable_sis_clone_fixup

Enables the sis clone fixup by re-setting flag

Throws C<NACL::Exceptions::UnexpectedState> if the validation fails.

=cut

sub create {
    my ($self_or_pkg, %args) = @_;
    my $self = $self_or_pkg->_new_if_required(%args);

    # Create the dataset directory
    my $dataset_directory = $self->_create_dataset_directory();

    # Create the parent file
    my $parent_file_obj = $self->_create_parent_file();
    $Log->comment("Parent File Path : " . $parent_file_obj->state()->path());

    # set bpass-throttle to true, to prevent failure due to clone back pressure.
    # This only applies if the fixup is enabled.
    my $bypass_throttle = 'true';
    if ('disable' eq $self->sis_clone_fixup()) {
        $bypass_throttle = 'false';
        $self->disable_sis_clone_fixup(); 
    }

    # Create clones of the parent file
    foreach my $i (1..$self->clone_cascading_depth) {

        my $clone_file_name = "sisclone_dataset_clone_file_" .
            random_name_generator() . "_" . $i;
        my $clone_file_obj = $parent_file_obj->clone_create(
            'destination-path' =>
                $dataset_directory->state()->name() . "/" .$clone_file_name,
            'bypass-throttle' => $bypass_throttle,
        );

        $Log->comment("Created new sis lone file : "
            . $clone_file_obj->state()->path());

        # Determine if we need to poke the clone file before creating new clone
        # of the clone
        if($self->change_buftree_shape()) {
            $self->_change_level_of_file($clone_file_obj);
        }

        # The clone file will become new parent since we are cascading clones.
        # Don't get confuse here. We are NOT deleting the parent file. This is
        # just reusing the same variable.
        $parent_file_obj = $clone_file_obj;
    }

    # Time to create metafile for our dataset
    $self->_write_manifest_file();

    # We should be okay with permission since the files are created with
    # dd utility. But let's make sure everybody can access this file.
    $self->_fixup_permissions();

    return $self;
}

sub _create_parent_file {
    my $self = shift;

    my $volume_obj = $self->volume();
    my $dataset_directory = $self->_get_dataset_directories();

    # Get the unix client information
    my $unix_client = $self->_get_unix_client();
    my $client_apiset = $unix_client->apiset();
    $client_apiset->set_privilege( 'privilege-level' => 'root' );

    # Mount the volume
    my $mm = data_mobility::ondemand::lib::MountManagerSingleton->new();
    my $mount = $mm->get_mount(
        volume => $volume_obj,
        client => $unix_client,
    );

    my $directory_client_path =
        $mount->mount_point(). '/' .$dataset_directory->state()->name();

    my $parent_file_name = "sisclone_dataset_parent_file_" .
        random_name_generator();

    # Create the Parent File
    $client_apiset->dd(
        'input-file' => '/dev/urandom',
        'output-file' => $directory_client_path . '/' . $parent_file_name,
        'bytes-rate' => 4096,
        'block-count' => (255 * 5),
        #Increase the timeout value to 120 seconds due to burt885491.
        #The default is 60 seconds.
        #Sometimes, dd may be very slow on some clients.
        #See the burt for more info.
        'connectrec-timeout' => 120,
    );

    my $dblade_path_to_parent_file =
        $dataset_directory->dblade_path() . '/' . $parent_file_name;
    my $parent_file = NACL::STask::VolumeFile->new(
        command_interface => $volume_obj->command_interface(),
        vserver  => $volume_obj->vserver(),
        path => $dblade_path_to_parent_file,
    );

    return $parent_file;
}

sub _change_level_of_file {
    my ($self, $file) = @_;

    my $volume_obj = $self->volume();

    # Get the unix client information
    my $unix_client = $self->_get_unix_client();
    my $client_apiset = $unix_client->apiset();
    $client_apiset->set_privilege( 'privilege-level' => 'root' );

    # Mount the volume
    my $mm = data_mobility::ondemand::lib::MountManagerSingleton->new();
    my $mount = $mm->get_mount(
        volume => $volume_obj,
        client => $unix_client,
    );

    my $client_path_to_file = $mount->mount_point().'/'.$file->state()->name();

    # Append a block to random level
    my $random_level = int (rand(3)) + 1;

    my $seek = ( ( 255**($random_level) ) * 5 );
    $client_apiset->dd(
        'input-file' => '/dev/urandom',
        'output-file' => $client_path_to_file,
        'bytes-rate' => 4096,
        'skipped-output-blocks-count' => $seek,
        'convert-options' => 'notrunc',
        'block-count' => 1,
    );
}

sub disable_sis_clone_fixup {
    my ($self) = @_;
    my $volume_obj = $self->volume();
    $volume_obj->command_interface()->apiset(
        category => 'Node',
        interface => 'CLI',
        set => 'CMode',
    )->system_node_run(
        node => $volume_obj->get_local_node_name(),
        command => "\"priv set test; setflag wafl_disable_sis_clone_fixup 1\"",
    );
    $self->sis_clone_fixup('disable');
}

sub enable_sis_clone_fixup {
    my ($self) = @_;
    my $volume_obj = $self->volume();
    $volume_obj->command_interface()->apiset(
        category => 'Node',
        interface => 'CLI',
        set => 'CMode',
    )->system_node_run(
        node => $volume_obj->get_local_node_name(),
        command => "\"priv set test; setflag wafl_disable_sis_clone_fixup 0\"",
    );
    $self->sis_clone_fixup('enable');
}

sub get_sisclone_files {
    my ($self) = @_;
    my $dataset_directory = $self->_get_dataset_directories();

    my @sis_clone_files = $dataset_directory->find_files(
        filter => { 'name' => 'sisclone_dataset_*' }
    );

    return @sis_clone_files;
}

sub get_sisclone_parent_file {
    my ($self) = @_;
    my $dataset_directory = $self->_get_dataset_directories();

    my $sis_clone_parent_file = $dataset_directory->find_files(
        filter => { 'name' => 'sisclone_dataset_parent_*' }
    );

    return $sis_clone_parent_file;
}

sub validate {
    my ($self) = @_;

    # Log state of each file
    map { $_->state() } ($self->get_sisclone_files());

    return $self;
}

sub DEMOLISH {
    my $self = shift;

    if($self->sis_clone_fixup =~ /disable/i) {
        $self->enable_sis_clone_fixup();
    }
}

=head1 REGRESSION TESTS

Regression tests for this library are located at
lib/NACL/UnitTest/MTask/DataSetSisCloneBasic.thpl

=head1 SEE ALSO

Please read L<NACL::MTask::DataSet|lib-NACL-MTask-DataSet-pm> to understand the
general behavior and capabilities of DataSet libraries.

Additional background on the DataSet framework is available at
https://wikid.netapp.com/w/NACL/Tasks/DataSet

=cut

1;
