#
# Copyright (c) 2013 NetApp, Inc., All Rights Reserved
# Any use, modification, or distribution is prohibited
# without prior written consent from NetApp, Inc.

## @summary DebugCrsStreams MTask Module
## @author dl-cluster-qa@netapp.com, dl-nacl-dev@netapp.com
## @status shared
## @pod here

package NACL::MTask::DebugCrsStreams;

use strict;
use warnings;

use base qw(NACL::MTask::MTask);
use NACL::STask::DebugCrsStream;
use NACL::C::DebugCrsStreamOperational;
use NACL::CS::ClusterIdentity;
use NACL::CS::Vserver;
use NACL::CS::DebugCrsStreamMappings;

use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Params::Validate qw(validate_with OBJECT SCALAR);

=head1 NAME

NACL::MTask::DebugCrsStreams

=head1 DESCRIPTION

C<NACL::MTask::DebugCrsStreams> is a wrapper around
L<NACL::STask::DebugCrsStream|lib-NACL-STask-DebugCrsStream-pm> that
simplifies handling both sender and receiver streams.

(CRS stands for Configuration Replication Service which is an internally
used for Metro-cluster and Vserver DR scenarios of backing up configurations.)

=head1 ATTRIBUTES

=head2 sender_stream

A C<NACL::STask::DebugCrsStream> object representing the sender stream.

=head2 receiver_stream

A C<NACL::STask::DebugCrsStream> object representing the receiver stream.

=cut

use Class::MethodMaker [
    new    => [ '-hash', 'new' ],
    scalar => [
        { -type => 'NACL::STask::DebugCrsStream' },
        'sender_stream'
    ],
    scalar => [
        { -type => 'NACL::STask::DebugCrsStream' },
        'receiver_stream'
    ],
];

=head1 METHODS

=head2 create_streams

    my $streams_mtask_obj = NACL::MTask::DebugCrsStreams->create_streams(
        command_interface => $command_interface,
        peer_interface    => $command_interface2,
        'stream-uuid'     => $uuid,
        
        # While other options are also accepted, they are not required
        # because they are defaulted by the MTask and the underlying STask
    );
    
    my $sender_stream_stask = $streams_mtask->sender_stream();
    my $receiver_stream_stask = $streams_mtask->receiver_stream();

(Class Method) Method that will create streams on both a sender and 
receiver cluster. The same options will be used for both.

It returns a C<NACL::MTask::DebugCrsStreams> object containing the
attributes C<sender_stream> and C<receiver_stream>.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) A command_interface that represents the sender.

=item C<< peer_interface    => $command_interface2 >>

(Required) A command_interface that represents the receiver.

=item Other options

All the other options accepted by
L<< NACL::STask::DebugCrsStream->create_stream|lib-NACL-STask-DebugCrsStream-pm/create_stream >>
are accepted by this method. The arguments defaulted by that library are
used here as well. In addition to what the STask defaults, the MTasks
defaults the values for C<remote-cluster-uuid>, C<peer-domain-instance-uuid>
(for the source); and C<domain-instance-uuid> and C<remote-cluster-uuid>
for the receiver, so these need not be provided in the call.

=over

=item For the sender stream

=over

=item 1)

The C<remote-cluster-uuid> is auto-determined if not provided
by quering the cluster identity of the command interface.

=back

=item For the receiver stream

=over

=item 1)

The C<domain-instance-uuid> is auto-determined if not provided
by querying for the uuid on a non-admin Vserver on the peer interface.

=item 2)

The C<remote-cluster-uuid> is auto-determined by querying the cluster
identity of the peer receiver command interface.

=back

=back

=back

=back

=cut

sub create_streams {
    $Log->enter();
    my ($pkg, @args) = @_;
    my %opts = validate_with(
        params => \@args,
        spec   => {
            command_interface => {
                type => OBJECT,
                isa => 'NACL::C::CommandInterface::ONTAP'
            },
            peer_interface    => {
                type => OBJECT,
                isa => 'NACL::C::CommandInterface::ONTAP'
            },
        },
        # All other options are sent through to the calls to create_stream
        # where they'll get validated
        allow_extra => 1,
    );

    my $command_interface  = delete $opts{command_interface};
    my $command_interface2 = delete $opts{peer_interface};
    
    my %common_opts;
    $pkg->_copy_common_component_params(
        source => \%opts,
        target => \%common_opts
    );
    
    my $sender_cluster_uuid = NACL::CS::ClusterIdentity->fetch(
        command_interface => $command_interface,
        %common_opts
    )->uuid();
    
    my $receiver_cluster_uuid;
    if (defined $opts{'remote-cluster-uuid'}) {
        $receiver_cluster_uuid = delete $opts{'remote-cluster-uuid'};
    } else {
        $receiver_cluster_uuid = NACL::CS::ClusterIdentity->fetch(
            command_interface => $command_interface2,
            %common_opts
        )->uuid();
    }

    my $receiver_domain_uuid;
    if (defined $opts{'peer-domain-instance-uuid'}) {
        $receiver_domain_uuid = delete $opts{'peer-domain-instance-uuid'};
    } else {
        $receiver_domain_uuid = NACL::CS::Vserver->fetch(
            %common_opts,
            command_interface => $command_interface2,
            filter            => { 'type' => '!admin' },
            requested_fields  => ['uuid'],
        )->uuid();
    }

    $opts{'league-uuid'} ||= $receiver_cluster_uuid;

    my $sender_stream = NACL::STask::DebugCrsStream->create_stream(
        %opts,
        command_interface      => $command_interface,
        direction              => 'sender',
        'remote-cluster-uuid'  => $receiver_cluster_uuid,
    );

    my $receiver_stream = NACL::STask::DebugCrsStream->create_stream(
        %opts,
        command_interface      => $command_interface2,
        direction              => 'receiver',
        'domain-instance-uuid' => $receiver_domain_uuid,
        'remote-cluster-uuid'  => $sender_cluster_uuid,
    );

    my $self = $pkg->new();
    $self->sender_stream($sender_stream);
    $self->receiver_stream($receiver_stream);
    
    $Log->exit();
    return $self;
}

=head2 purge

    $streams_mtask_obj->purge();

(Instance Method only) Deletes both the sender and receiver streams held
by the object.

=cut

sub purge {
    $Log->enter();
    my ($self, %opts) = @_;

    # Iterate through the streams and delete them
    $self->sender_stream()->purge(%opts);
    $self->receiver_stream()->purge(%opts);

    $Log->exit();
}

=head2 reset_held_streams

    $streams_mtask->reset_held_streams();

(Instance Method only) Function that will reset both the sender and
receiver streams held by the object and checks for the notification of 
'in-error' and remove it as well.

=cut

sub reset_held_streams {
    $Log->enter();
    my ($self, %opts) = @_;
    my @streams = ( $self->sender_stream(), $self->receiver_stream() );
    foreach my $stream (@streams) {
        my $command_interface = $stream->command_interface();
        my $stream_uuid = $stream->stream_uuid();
        my @streams = ($stream_uuid);

        my $domain_uuid = NACL::CS::DebugCrsStreamMappings->fetch(
            %opts,
            command_interface => $command_interface,
            filter            => { 'streams' => \@streams },
            requested_fields  => [ 'domain-instance-uuid' ],
        )->domain_instance_uuid();


        NACL::STask::DebugCrsStream->reset_streams(
            %opts,
            command_interface      => $command_interface,
            'domain-instance-uuid' => $domain_uuid,
            streams                => \@streams,
        );

         my $DebugCrsStreamOperational_obj = NACL::C::DebugCrsStreamOperational->find(
            %opts,
            command_interface => $command_interface,
            filter => {
                    'stream-uuid' => $stream_uuid,
                    'oper-state'  => 'in-error',
            },
            allow_empty => 1,
        );

        if ($DebugCrsStreamOperational_obj) {
            my $apiset = $command_interface->apiset(
                interface => 'CLI', 
                set       => 'CMode'
            );

            #Ignore errors from this command because it may not be in error
            $apiset->execute_raw_command( 
                command => "debug smdb table crs_pending_dm_notifications delete -stream-uuid $stream_uuid"
            );
        }
    }

    $Log->exit();
}

1;
