#
# Copyright (c) 2013 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#
#
## @summary MTask FileIterator Base class
## @author asada@netapp.com, dl-nacl-dev@netapp.com
## @status shared
## @pod here
##############################################################################
package NACL::MTask::FileIterator::OperationalFileIterator;

use Moose;
use Moose::Util::TypeConstraints;

use Tharn;
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use NACL::STask::Volume;
use NACL::STask::VolumeFile;
use NACL::Exceptions::NoElementsFound;
use NACL::APISet::Exceptions::InvalidParamValueException;
use NACL::C::Exceptions::VolumeDirectory::AlreadyExists qw(:try);

use NACL::GeneralUtils qw(random_name_generator);
use File::Spec::Functions qw(catdir);

#
# file_iterator is required object that is iterated and each element is used to
# do further iteration with the applied operation.
#
has file_iterator => (
    is => 'rw',
    isa => 'Object',
    required => 1,
);

# Create separate directory for files at the destination is any
has create_destination_directory => (
    is => 'ro',
    isa => 'Bool',
    default => 1,
);

# name of the destination directory
has destination_directory => (
    is => 'rw',
    isa => 'Str',
    default => '',
);

#
# Group operations flag - allow iterator to do next group operation on files
#
has group_operations_flag => (
    is => 'ro',
    isa => 'Bool',
    required => 0,
    default => 0,
);

#
# Number of files to perform an operation on in a go
#
has num_of_bulk_files => (
    is => 'ro',
    isa => 'Int',
    required => 0,
    default => 40,
);

has _next_file => (
    is => 'rw',
    isa => 'NACL::C::VolumeFile',
    clearer => '_clear_next_file',
    predicate => '_has_next_file',
);

#
# Array to store group operationed files will only be filled if group operations flag set
# and the iterator performed a group op.
#
has _operationed_files => (
    is => 'rw',
    isa => 'ArrayRef[NACL::C::VolumeFile]',
    clearer => '_clear_next_operationed_files',
    predicate => '_has_next_operationed_files',
    default => sub { return []; },
);

#
# Checks if the iterator has a next operationable file
#
sub has_next {
    my $self = $_[0];
    $Log->enter() if $may_enter;
    $self->_refill_next_file();
    if ($self->group_operations_flag) {
        if ($self->_has_next_operationed_files()) {
            $Log->exit() if $may_exit;
            return 1;
        }
    } else {
        if ($self->_has_next_file()) {
        $Log->exit() if $may_exit;
        return 1;
        }
    }

    $Log->exit() if $may_exit;
    return 0;
}

# It is a core refill function refills the iterator
sub _refill_next_file {
    my $self = $_[0];
    $Log->enter() if $may_enter;

    # If we have a file that is ready for next to return, stop now
    # same goes for group
    if (($self->group_operations_flag == 0) && ($self->_has_next_file())) {
        $Log->exit() if $may_exit;
        return;
    }

    if (($self->group_operations_flag == 1) && ($self->_has_next_operationed_files())) {
        $Log->exit() if $may_exit;
        return;
    }

    if ($self->group_operations_flag){
        # For group operations we will collect the files for source upto
        # num_of_bulk_files and pass that on to child class for operation
        my $count = 0;
        my @source_files = ();
        while ($self->file_iterator()->has_next() && $count < $self->num_of_bulk_files) {
            my $file = $self->file_iterator()->next();
            # If operation can be performed on the file push it to the list
            if ($self->_is_operationable($file)) {
                $Log->comment ("Selected source file: ". $file->path());
                push @source_files, $file;
            }

            $count++;
        }
        if (scalar @source_files > 0) {
            # We have collected upto num_of_bulk_files so lets call operation on them
            my @processed_files = $self->_do_file_operation(@source_files);
            $self->_operationed_files(\@processed_files);
            if (scalar @processed_files == 0){
                $Log->exit() if $may_exit;
                NATE::BaseException->throw("Iterator Operation not performed");
            }
        }
    } else {
        while ($self->file_iterator()->has_next()) {
            my $file = $self->file_iterator()->next();
            my @source_files = ();
            # If operation can be performed on the file, just do it.
            if ($self->_is_operationable($file)) {
                $Log->comment ("Selected source file: ". $file->path());
                push @source_files, $file;
                my $processed_file =$self->_do_file_operation(@source_files);
                if (!defined $processed_file){
                    $Log->exit() if $may_exit;
                    NATE::BaseException->throw("Iterator Operation not performed");
                }

                $self->_next_file($processed_file);
                $Log->exit() if $may_exit;
                return;
            }
        }
    }
    $Log->exit() if $may_exit;
    return;
}

# Base method 'only filters for files', need to override in child classes
sub _is_operationable {
    my $self = shift;
    my $file = shift;

    $Log->enter() if $may_enter;

    my $type = "";
    try{
        $type = $file->get_one_state_attribute('file-type');
    } catch NACL::Exceptions::NoElementsFound with {
        # File might have gone away btw the time
        # we made it here from the original iterator
        # if the file isnt there its obviously un operationable :)
    };

    $Log->exit() if $may_exit;
    return 0 if ($type ne 'file');
    return 1;
}

# Base method, need to override in child classes
sub _do_file_operation {
    my $self = shift;
    my $file = shift;
    return $file;
}

#
# does the iterator next
#
sub next {
    my $self = $_[0];
    $Log->enter() if $may_enter;
    if ($self->group_operations_flag) {
        $Log->exit() if $may_exit;
        NACL::Exceptions::InvalidMethod->throw ("next() cannot be called with group_operations_flag, use next_n_operations() instead.");
    }

    $self->_refill_next_file();
    if (!$self->_has_next_file()) {
        $Log->exit() if $may_exit;
        NACL::Exceptions::NoElementsFound->throw(
            "Called next and no elements were remaining"
        );
    }

    my $file = $self->_next_file();
    $self->_clear_next_file();
    $Log->exit() if $may_exit;
    return $file;
}

#
# does the next N operations, returns the N operationed files
#
sub next_n_operations {
    my $self = $_[0];
    $Log->enter() if $may_enter;
    if (!$self->group_operations_flag) {
        $Log->exit() if $may_exit;
        NACL::Exceptions::InvalidMethod->throw ("next_n_operations() cannot be called without group_operations_flag, use next() instead.");
    }

    $self->_refill_next_file();
    if (!$self->_has_next_operationed_files()) {
        $Log->exit() if $may_exit;
        NACL::Exceptions::NoElementsFound->throw(
            "Called next and no elements were remaining"
        );
    }

    my @op_files = @{$self->_operationed_files()};
    $self->_clear_next_operationed_files();
    $Log->exit() if $may_exit;
    return @op_files;
}

#
# rewind the iterator
#
sub rewind {
    my $self = $_[0];
    $Log->enter() if $may_enter;
    $self->file_iterator()->rewind();
    $Log->exit() if $may_exit;
}

# Private Method: _zapi_volume_create_dir
# Description: Create a directory on the specified volume.
sub _zapi_volume_create_dir {
    my ($self, $volume, $dir_path) = @_;
    $Log->enter() if $may_enter;

    try {
        $volume->directory_create(dirname => $dir_path, perm => '777');
    } catch NACL::C::Exceptions::VolumeDirectory::AlreadyExists with {
        $Log->debug ("Directory exists so re-use. ");
    };

    $Log->exit() if $may_exit;
}

1;

=head1 NAME

NACL::MTask::FileIterator::OperationalFileIterator - Provide an iterator
interface base class to be used by various child iterator classes.

=head1 SYNOPSIS

This is a base iterator class and if used will just iterate over the iterator provided

extends NACL::MTask::FileIterator::OperationalFileIterator;

This class is not intented to be used directly buts its childern iterator will be.

=head1 DESCRIPTION

This is a base iterator class for various file iterator api.

=head1 FUNCTIONS

=head2 new

Create a new iterator object instance and return it.

=over

=item C<< file_iterator >> An instance of a file iterator, it must support next
and has_next. The next method must return NACL::C::VolumeFile instances.

=back

=head2 has_next

Returns 1 if there is another "operationed" file available to get via a call to next or
0 if there are no more files available.

=head2 next

Return the next "operationed" file as a NACL::C::VolumeFile instance.
Note: this cannot be used if doing group operations if you pass group_operations_flag then
use next_n_operations()

=head2 next_n_operations

Does the file "operation" on N files and returns all the "operationed files" as a Array Reference of
"NACL::C::VolumeFile" objects.
Note: This function cannot be used without group operations flag

=head2 rewind()

Rewinds the iterator so the user can re-use this iterator again.
Note: if there is a destination directory to be specfied, a separate directory is created
for it.
Caution: Keep in mind previous on demand operations might still be happening. Make sure the
source is ready before iterating again.

=head1 SEE ALSO

=cut
