#
# Copyright (c) 2013 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#
#
## @summary MTask SFmoD File Iterator Class
## @author asada@netapp.com, dl-nacl-dev@netapp.com
## @status shared
## @pod here
##############################################################################

package NACL::MTask::FileIterator::SfmodFileIterator;
use Moose;

use Tharn;
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use NACL::STask::Volume;
use NACL::STask::VolumeFile;
use NACL::STask::VolumeFileMove;
use NATE::BaseException;
use NACL::APISet::Exceptions::ResponseException;
use NACL::GeneralUtils qw(random_name_generator);

use constant WAIT_TIME => '10';

# iterator functions and attributes are provided by the base class
extends qw (NACL::MTask::FileIterator::OperationalFileIterator);

#
# destination volume is a required field for file move on demand iterator
#
has destination_volume => (
    is => 'ro',
    isa => 'NACL::C::Volume',
    required => 1,
);

# Overide parent function to do SFMoD operation on a file
override '_do_file_operation' => sub {
    my ($self, @files) = @_;
    $Log->enter();
    my $volume = $self->destination_volume();
    my $create_destination_directory = $self->create_destination_directory();

    my $sfmod_dir_path;
    # create a destination directory
    if ($create_destination_directory) {
        my $new_dir_path = "";
        # If we have previously created a directory use it for all the sfmod ops with this
        # iterator
        if ( $self->destination_directory() eq  '') {
            $new_dir_path = random_name_generator(
                prefix => 'sfmod_dir',
                length => 128
            );

            $self->destination_directory($new_dir_path);
        }
        else {
            $new_dir_path = $self->destination_directory();
        }

        # call parents dir create method -reuse if exists
        $self->_zapi_volume_create_dir($volume, $new_dir_path);
        $sfmod_dir_path = "/vol/".$volume->volume()."/$new_dir_path";
    }
    else {
        $sfmod_dir_path = "/vol/".$volume->volume();
    }

    # For all the files passed lets make destination file objects and pass it to
    # move operation
    my @dest_files = ();
    foreach my $file (@files) {
        # Create a random name for moved file
        my $newfilename = random_name_generator(
            prefix => 'sfmod_file',
            length => 128
        );

        my $dest_file_obj = NACL::STask::VolumeFile->new(
            command_interface => $file->command_interface(),
            vserver => $volume->vserver(),
            path => $sfmod_dir_path."/$newfilename",
        );

        push @dest_files, $dest_file_obj;
    }

    my $command_interface = $files[0]->command_interface();
    # make a copy of destination files objects as file move 
    # changes them from objects to a concatenated string
    my @new_dest_files =  @dest_files;

    # Call file move
    my $VolumeFileMove_objs;
    # Keep re-trying until a slot opens up for move job to be slotted
    while (!$VolumeFileMove_objs) {
        try {
            $VolumeFileMove_objs = NACL::STask::VolumeFileMove->start(
                command_interface => $command_interface,
                'source-paths' => \@files,
                'destination-paths' => \@new_dest_files,
                nacltask_wait => 0,
            );
        }
        catch NACL::APISet::Exceptions::ResponseException with {
            my $ex = shift;
            if ($ex->text() =~ /Maximum simultaneous file operations reached/i) {
                $Log->trace("Maximum number of operations running, not removing ".
                "the operation from the staging area");
                 Tharn::sleep(WAIT_TIME);
            } else {
                $Log->exit();
                @dest_files = ();
                $ex->throw();
            }
        };
    }

    $VolumeFileMove_objs->wait_for_destination_ready();

    $Log->exit();
    # create VolumeFile object for the moved file and return
    if ($self->group_operations_flag){
        return @dest_files;
    } else {
        return $dest_files[0];
    }
};

# Overide parent function to to overide for sfmod (rewind to move directory on destination)
override 'rewind' => sub {
    my $self = $_[0];
    $Log->enter();
    require NACL::MTask::VolumeFileIterator;
    require NATE::BaseException;

    my $volume = $self->destination_volume();

    if ($self->destination_directory() eq '') {
        NATE::BaseException->throw(
            "Cannot call rewind on sfmod iterator before iterating once."
        );
    }

    # create new iterator for the moved files :)
    my $moved_file_iter_obj = NACL::MTask::VolumeFileIterator->new(
                volume   => $volume,
                path     => $self->destination_directory(),
                method   => 'zapi',
                criteria => {
                    exclude_types => [
                        'symlink', 'lun',  'blockdev', 'chardev',
                        'socket',  'fifo', 'stream'
                    ]
                },
            );

    $self->file_iterator($moved_file_iter_obj);
    $Log->exit();
};

1;
=head1 NAME

NACL::MTask::FileIterator::SfmodFileIterator - Provide an iterator object from iterating and doing a
file ondemand move operation.

=head1 SYNOPSIS

This is a File on demand move iterator class and if used iterate and do a move operation the files in
provided by input iterator.

extends NACL::MTask::FileIterator::SfmodFileIterator;

The class overides base functions to check if the Volume File object input is able to be moved and also overides
the operation function in this case a file move operation is performed.

=head1 DESCRIPTION

This Class provides an iterator that for each next() function will do the file move operation before return
the NACL::C::VolumeFile object for the moved file.

=head1 FUNCTIONS

=head2 new

Create a new iterator object instance and return it.

=over

=item C<< file_iterator >> An instance of a file iterator, it must support next
and has_next. The next method must return NACL::C::VolumeFile instances.

=back

=head2 has_next

Returns 1 if there is another "ondemand move" file available to get via a call to next or
0 if there are no more files available.

=head2 next

Return the next "ondemand moved" file as a NACL::C::VolumeFile instance.

=head1 SEE ALSO

=cut
