#
# Copyright (c) 2013 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#
#
## @summary MTask SFR Vault File Iterator Class
## @author asada@netapp.com, dl-nacl-dev@netapp.com
## @status shared
## @pod here
##############################################################################

package NACL::MTask::FileIterator::SfrFileIterator;
use Moose;

use Tharn;
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use NACL::STask::Volume;
use NACL::STask::Snapmirror;
use NATE::BaseException;
use NACL::APISet::Exceptions::ResponseException;
use NACL::Exceptions::InvalidParam;
use NACL::GeneralUtils qw(random_name_generator);

# iterator functions and attributes are provided by the base class
extends qw (NACL::MTask::FileIterator::OperationalFileIterator);

# Retries
use constant RETRIES => 5;
# timeout
use constant METHOD_TIMEOUT => 180;
# Polling Interval
use constant POLLING_INTERVAL => 30;
#
# Destination volume object, required to restore file too,
# Has to be a type RW volume will only be used as ro object here.
#
has destination_volume => (
    is => 'ro',
    isa => 'NACL::C::Volume',
    required => 1,
);

#
# Snapshot name from which the file will be restored from
#
has snapshot_name => (
    is => 'ro',
    isa => 'Str',
    required => 1,
);

override '_is_operationable' => sub {
    my ($self, $file) = @_;
    $Log->enter() if $may_enter;
    # Also check the destination volume and see if its a RW volume
    # Note: If its type RW imples it wont be a destination of a snapmirror relationship
    my $volume = $self->destination_volume();
    my $vol_type = $volume->get_one_state_attribute('type');
    if ($vol_type !~ /RW/i ) {

        NACL::Exceptions::InvalidParam->throw("Destination volume for single file restore from vault has to be type rw or RW and not: $vol_type");
        $Log->exit() if $may_exit;
        return 0;
    }
    # Get the file type
    my $type = "";
    try{
        $type = $file->get_one_state_attribute('file-type');
    } catch NACL::Exceptions::NoElementsFound with {
        # File might have gone away btw the time
        # we made it here from the original iterator
        # if the file isnt there its obviously un operationable :)
    };

    $Log->exit() if $may_exit;
    # Only file for now
    return 1 if ($type eq 'file' );
    return 0;
};

# Overide parent function to do SFR operation on a file
override '_do_file_operation' => sub {
    my ($self, $file) = @_;
    $Log->enter() if $may_enter;
    my $dest_file_obj =undef;
    my $volume = $self->destination_volume();
    my $create_destination_directory = $self->create_destination_directory();

    # Lets get the volume object for the file
    my $source_vol_object = $file->get_containing_volume_obj();
    $Log->debug ("Printing file path:" . $file->path());
    my $src_vol_name = $source_vol_object->volume();
    my $dst_vol_name = $volume->volume();
    my $snapshot_name = $self->snapshot_name();

    my $error =0;
    # Lets get the directory path from the source file, replace the the source
    # volume with destination volume and check if the directory exists
    my $dir_path = $file->get_containing_directory_path();

    # Lets remove the volume portion /vol/volume_name
    $dir_path =~ s/^\/vol\/${src_vol_name}//;
    # lets remove the snapshot name as well
    $dir_path =~ s/^\/.snapshot\/$snapshot_name//;
    # now lets create the directories at destination if they are not there
    my @dirs = split ('/', $dir_path);
    # Lets see if the dirs exists else create them on the destination volume
    my $dir_chain = "";
    my $skip_format = 1;
    $Log->debug ("Number of subdirs: ". scalar @dirs);
    foreach my $dir (@dirs) {
        $Log->debug ("DIR name: $dir");
        if (length $dir > 0) {
            if ($skip_format) {
                $dir_chain = $dir;
                $skip_format = 0;
            }else {
                # create the dir/sub_dir as we go
                $dir_chain = $dir_chain."/".$dir;
            }

            my $dest_file_dir_obj;
            try {
                $dest_file_dir_obj = $volume->directory_create(dirname => $dir_chain, perm => '777');
            }
            catch NACL::C::Exceptions::VolumeDirectory::AlreadyExists with {
                # Dont do anything if it exists its even better
            }
            catch NATE::BaseException with {
                $error = 1;
                my $e = shift;
                my $d_path = "/vol/$dst_vol_name$dir_chain";
                NATE::BaseException->throw("Could not create directory ($d_path) at destination volume: $dst_vol_name with error ". $e->text());
            };

            if ($error == 1) {
                last;
            }

        }
    }

    # We could not replicate directory structure for restore on destination so cant
    # proceed
    if ($error == 1) {
        $Log->exit() if $may_exit;
        return $dest_file_obj; # return undefined object as we can progress
    }

    # extract the partial file path from the file path /dir/file1 for api call
    my $partial_file_path = $file->path();
    $partial_file_path =~ s/^\/vol\/${src_vol_name}//;
    $partial_file_path =~ s/^\/.snapshot\/$snapshot_name//;

    # Call file restore note: all files will be copied to the destination directory
    my $restorefile_obj;
    my $job;
    my $retries = 0;
    my %options = ();
    $options{command_interface} = $volume->command_interface();
    $options{'destination-path'} = $volume->{vserver}.":".$dst_vol_name;
    $options{'source-path'} = $source_vol_object->{vserver}.":".$src_vol_name;
    $options{'source_ci'}   = $volume->command_interface();
    while (!$restorefile_obj) {
        try {
            $restorefile_obj =
                NACL::STask::Snapmirror->restore(
                    job_component => \$job,
                    'file-list' => [ $partial_file_path, ],
                    'source-snapshot' => $snapshot_name,
                    %options,
            );

            # Assuming all went well destination path should be the same as source path volume down
            # Create a volume file object for the destination file at destination dir
            $dest_file_obj = NACL::STask::VolumeFile->new(
                command_interface => $file->command_interface(),
                vserver => $volume->vserver(),
                path => "/vol/$dst_vol_name$partial_file_path",
            );
        }
        catch NATE::BaseException with{
            my $e = shift;
            if ($retries < RETRIES) {
                if ($e =~ /Another restore operation is partially completed using a different source volume/i) {
                    # Previous partial relation ship might be left hanging lets use clean-up-failure on re-try
                    # example:  Cannot restore to destination volume "CSIQA624010VS1:ksink_src1" from volume "
                    #CSIQA624013VS4:ksink_src6_XDP". Another restore operation is
                    #partially completed using a different source volume.
                    my $cleanup_restore = NACL::STask::Snapmirror->restore(
                        %options,
                        'clean-up-failure' => 'true',
                        job_component => \$job,
                        );

                    if (defined $job) {
                        $job->wait_on_job(
                        'method-timeout' => METHOD_TIMEOUT,
                        polling_interval => POLLING_INTERVAL,
                        );
                    }
                } else {
                    $e->set_text("Retry: $retries Could not do a file restore from snapmirror with Source Path: ".$file->path()
                    . " Original failure: ". $e->text());
                    $e->throw();
                }
                Tharn::sleep(5);
                $retries++;
            } else {
                $e->throw();
            }
        };
    }

    if (defined $job) {
        $job->wait_on_job(
            'method-timeout' => METHOD_TIMEOUT,
            polling_interval => POLLING_INTERVAL,
        );
    }

    $Log->exit() if $may_exit;
    # Return the restored file as a VolumeFileObject or undef in error case
    return $dest_file_obj;
};

1;
=head1 NAME

NACL::MTask::FileIterator::SfrFileIterator - Provide an iterator object from iterating and doing a
snapmirror restore on files

=head1 SYNOPSIS

This is a File restore from snapmirror iterator class and if used iterate and do a restore operation the files in
provided by input iterator of a volume.

extends NACL::MTask::FileIterator::SfrFileIterator;

The class overides base functions to check if the Volume File object input is able to be restored and also overides
the operation function in this case a file restore from snapmirror operation is performed.

=head1 DESCRIPTION

This Class provides an iterator that for each next() function will do the file restore operation before return
the NACL::C::VolumeFile object for the restored file.

=head1 FUNCTIONS

=head2 new

Create a new iterator object instance and return it.
my $test_iterator = NACL::MTask::FileIterator::SfrFileIterator->new(
    destination_volume => $Dest_Vol,
    file_iterator => $vol_iter_obj,
    snapshot_name => $snap_obj->{snapshot},
);

=over

=item C<< file_iterator >> An instance of a file iterator, it must support next
and has_next. The next method must return NACL::C::VolumeFile instances.

=back

=head2 has_next

Returns 1 if there is another "snapmirror file restore" file available to get via a call to next or
0 if there are no more files available.

=head2 next

Return the next "restored" file as a NACL::C::VolumeFile instance.

=head1 SEE ALSO

=cut
