#
# Copyright (c) 2013 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary NodeRouter MTask module
## @author dl-nacl-dev@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::MTask::Group::NodeRouter

=head1 DESCRIPTION

This module provides an object which is to be used as a replacement
for NACL::C::Node objects when SFO testing is performed.

When a node is taken over, the command_interface modeling it also
becomes unusable (since that node has gone down). This makes all Component
objects tied to that command_interface unusable.

A C<NACL::MTask::Group::NodeRouter> object helps handle these situations
by issuing the commands to a node in the cluster which is up and usable.
This is how it is to be used:

* Instead of using NACL::C::Node objects as command_interface, use a
C<NACL::MTask::Group::Router> object as command_interface.

* Associate all the nodes in the cluster with this NodeRouter.

* If we know a node is going down for some reason, specify to the router
that it is to be disabled by invoking the L<disable|lib-NACL-MTask-Group-NodeRouter-pm/disable> API.

* Once a node is disabled, the NodeRouter will ensure commands are issued
to some other node in the cluster.

* Once the node comes back up, specify to the NodeRouter that the node is
ready for use again by invoking the L<enable|lib-NACL-MTask-Group-NodeRouter-pm/enable> API.

=over

=item Basic Code example

    # Get any node in the cluster
    my $node = NACL::C::Node->find();
    
    # Get all nodes in the cluster
    my @nodes_in_cluster = NACL::C::Node->find(command_interface => $node);
    
    # Create a NodeRouter objects associated with all nodes
    my $router = NACL::MTask::Group::NodeRouter->new(nodes => \@nodes_in_cluster);
    
    # This object can be used in place of a command_interface
    my $aggr = NACL::STask::Aggregate->create(command_interface => $router);
    
    # ... Perform other operations ...
    
    # Issue command that causes $nodes_in_cluster[0] to go down and a
    # takeover to occur
    # All that needs to be done is to mark that node as disabled
    $router->disable($nodes_in_cluster[0]);
    
    # There is no need to change the command_interface, the NodeRouter will
    # take care of routing commands to a different node
    my $vol = NACL::C::Volume->find(command_interface => $router);
    
    # Component/Task objects created prior to the takeover continue to work
    # without any changes
    $aggr->modify(%some_opts);
    
    # $nodes_in_cluster[0] is brought back up and giveback is issued
    # Specify to the router that this node is back up again and reconnect it.
    $router->enable($nodes_in_cluster[0]);
    
    # And continue as usual ...
    my $vol_cs = $vol->state();

=back

=head1 ATTRIBUTES

Note that it is not advised to directly modify the attributes of the
object.

=head2 enabled_set

A C<NACL::MTask::Set::SystemNode> object containing all of the nodes
which are usable. For example, if we'd like to get all of the enabled
nodes from this set, we'd do:

    my $enabled_set     = $router->enabled_set();
    my @enabled_members = $enabled_set->get_all_members();

The documentation of L<NACL::MTask::Set|lib-NACL-MTask-Set-pm> has more
details of the functionality available on Set objects.

=head2 disabled_set

A C<NACL::MTask::Set::SystemNode> object containing all of the nodes
which are currently unusable. For example, if we'd like to get all of the disabled
nodes from this set, we'd do:

    my $disabled_set     = $router->disabled_set();
    my @disabled_members = $disabled_set->get_all_members();

The documentation of L<NACL::MTask::Set|lib-NACL-MTask-Set-pm> has more
details of the functionality available on Set objects.

=head2 algorithm

The algorithm that's used to choose which node to issue commands to. Valid
values are C<first> and C<round_robin>.

A value of C<first> means that the commands are issued to the first node in
the enabled set.

A value of C<round_robin> means that each command is issued to a different
node with the nodes being chosen in a round-robin fashion. For example,
if the enabled_set includes nodes N1, N2 and N3 a value of C<round_robin>
would make the first command be issued against N1, the next one against
N2, the next one against N3, the next one against N1 and so on.

=cut

package NACL::MTask::Group::NodeRouter;

use strict;
use warnings;

use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use NACL::MTask::Set::SystemNode;
use NACL::STask::Node;
use Params::Validate qw(:all);
use NACL::Exceptions::InvalidMethod;

# Have to use this to fix Class::MethodMaker wrongly invoking
# UNIVERSAL's isa as a function instead of a method call.
# Class::MethodMaker does
#    UNIVERSAL::isa($obj, $method)
# instead of:
#    $obj->isa($method)
# The second form is the correct one and would read overloaded isa (such as
# the one defined here).
use UNIVERSAL::isa;

use Class::MethodMaker [
    scalar => [{-type => 'NACL::MTask::Set::SystemNode'}, 'enabled_set',],
    scalar => [
        {   -type         => 'NACL::MTask::Set::SystemNode',
            -default_ctor => sub { NACL::MTask::Set::SystemNode->new() },
        },
        'disabled_set',
    ],
    scalar => 'algorithm',

    # Our own little form of an iterator. Note that the values of the
    # enabled_set is dynamic so we cannot use any of the freely available
    # iterator CPAN modules.
    scalar => [
        {   -read_cb => sub {
                my ($self)  = $_[0];
                my $members = $self->enabled_set()->get_all_members();
                my $index   = $self->{_index};
                $index++;

                # Gone past the end of the set, reset to the beginning
                if ($index >= @$members) {
                    $index = 0;
                }
                $self->{_index} = $index;
                return $index;
            },
            -default_ctor => sub {
                -1
            },
        },
        '_index',
    ],
];

=head1 METHODS

=head2 new

    # Basic example
    # Find all nodes in the cluster
    my @nodes  = NACL::C::Node->find(command_interface => $node);
    my $router = NACL::MTask::Group::NodeRouter->new(nodes => \@nodes);

    # Example with algorithm being passed.
    my @nodes  = NACL::C::Node->find(command_interface => $node);
    my $router = NACL::MTask::Group::NodeRouter->new(
        nodes     => \@nodes,
        algorithm => 'round_robin'
    );

    # A small variation: here's an example showing how we'd use routers to
    # invoke commands on only the highest version nodes of the cluster.
    my $version_grp_obj = $node->group_by_version();
    my @highest_version_nodes = $version->node_objs_in_highest_version();
    my $router = NACL::MTask::Group::NodeRouter->new(nodes => \@highest_version_nodes);

Create a new NodeRouter object. C<nodes> should be provided an array-reference
of C<NACL::C::Node> or C<NACL::STask::Node> objects. The C<algorithm> can be
either 'first' or 'round_robin' (described above in the Attributes section)

=cut

sub new {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    my %opts = validate_with(
        params => \@args,
        spec   => {
            nodes     => {type => ARRAYREF},
            algorithm => {type => SCALAR, default => 'first'},
        },
    );

    my $all_nodes
        = NACL::MTask::Set::SystemNode->new(members => $opts{nodes});
    my $self = bless {}, $pkg;

    $self->enabled_set($all_nodes);
    $self->algorithm($opts{algorithm});

    $Log->exit() if $may_exit;
    return $self;
}

=head2 disable

    # Can specify a single node
    $router->disable($node);
    
    # Or can specify multiple nodes
    $router->disable(@many_nodes);

Specifies to the NodeRouter that the node(s) specified are unusable so should
be disabled. (net effect is that commands are not issued to these nodes) This
should be done when we know a particular node is going down.

=cut

sub disable {
    $Log->enter() if $may_enter;
    my ($self, @nodes) = @_;

    $self->enabled_set()->remove(@nodes);
    $self->disabled_set()->add(@nodes);

    $Log->exit() if $may_exit;
}

=head2 enable

    # Can specify a single node
    $router->enable($node);
    
    # Or can specify multiple nodes
    $router->enable(@many_nodes);

Specifies to the NodeRouter that the node(s) specified are usable so should
be re-enabled. This should be done when we know a particular node is being
brought back up. This takes care of re-establishing the connection as well.

=cut

sub enable {
    $Log->enter() if $may_enter;
    my ($self, @nodes) = @_;
    $self->disabled_set()->remove(@nodes);
    $self->enabled_set()->add(@nodes);

    # Re-establish the connections
    foreach my $node (@nodes) {
        $node->refresh_command_interface(no_version_change => 1);
    }

    $Log->exit() if $may_exit;
}

=head2 get_any_enabled_node

    my $enabled_node = $router->get_any_enabled_node();

Returns a NACL::C::Node object for any node that is enabled (i.e. is usable)

=cut

sub get_any_enabled_node {
    $Log->enter() if $may_enter;
    my ($self, @args) = @_;
    my %opts = validate_with(
        params => \@args,
        spec   => {algorithm => {type => SCALAR, optional => 1}},
    );
    my $members = $self->enabled_set()->get_all_members();
    if (!@$members) {
        $Log->exit() if $may_exit;
        NATE::BaseException->throw(
            'No nodes present in the set which are ' . 'enabled');
    }

    my $algorithm = $opts{algorithm} || $self->algorithm();
    if ($algorithm eq 'first') {
        my $first_member = $members->[0];
        $Log->exit() if $may_exit;
        return $first_member;
    } else {
        my $index = $self->_index();
        $Log->exit() if $may_exit;
        return $members->[$index];
    }
}

sub _get_first_enabled_node {
    $Log->enter() if $may_enter;
    my $self = $_[0];
    my $node = $self->get_any_enabled_node(algorithm => 'first');
    $Log->exit() if $may_exit;
    return $node;
}

# Look away, some fairly hacky code below.

# Code that would have been expecting a command_interface object would
# be greeted by a NACL::MTask::Group::NodeRouter object. We need to trick
# other code into treating this just like it would a command_interface.
# We need to override can, AUTOLOAD and isa for this.

# Methods on the Node object should be available on a NodeRouter object
# as well.
sub can {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    $Log->exit() if $may_exit;
    return NACL::STask::Node->can(@opts);
}

# We cannot simply make this a sub-class of NACL::C::SystemNode.
# This is because we need calls to the Node methods (hostrec, apiset, etc)
# to be redirected. A call like:
#    $router->hostrec()
# should instead get redirected to the current node, i.e. should be
# implemented like:
#    $router->current_node()->hostrec();

sub AUTOLOAD {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    our $AUTOLOAD;
    my $method = $AUTOLOAD;
    $method =~ s/.*:://;

    if (my $code = $pkg_or_obj->can($method)) {

        # Do NOT install this into the symbol table because the closure
        # is tied to a node, so all calls to that methods will end up
        # getting invoked on that node.
        my $updated_code = sub {

            # Methods would be invoked with a NodeRouter object, so
            # $self = NACL::MTask::Group::NodeRouter
            # Redirect the method call to an enabled node
            my ($self, @opts_in_sub) = @_;

            # Cannot use get_any_enabled_node here. This is because
            # if round-robin is enabled, then each method that hits AUTOLOAD
            # will update the index, so it's not guaranteed that we'll
            # really be issuing commands in a round-robin fashion.
            # We should ensure that the only thing getting nodes in a
            # round-robin fashion is apiset(), because every call that
            # issues commands goes through the apiset() method.
            # For all other methods, getting the first node should suffice.
            my $current_node = $self->_get_first_enabled_node();
            return $code->($current_node, @opts_in_sub);
        };
        $Log->exit() if $may_exit;
        return $updated_code->($pkg_or_obj, @args);
    } else {
        my $ref = ref ($pkg_or_obj) || $pkg_or_obj;
        $Log->exit() if $may_exit;
        NACL::Exceptions::InvalidMethod->throw("Method '$method' not defined "
            . "on package '$ref'");
    }
}

sub apiset {
    $Log->enter() if $may_enter;
    my ($self, @opts) = @_;
    my $current_node = $self->get_any_enabled_node();
    my $apiset       = $current_node->apiset(@opts);
    $Log->exit() if $may_exit;
    return $apiset;
}

sub DESTROY { }

# We need to ensure that code that validated expecting a command_interface
# object does not fail.
# The following should be true:
#     $router->isa('NACL::C::SystemNode')
#     $router->isa('NACL::C::Node')
#     $router->isa('NACL::STask::Node')
#     $router->isa('NACL::C::CommandInterface')
#     $router->isa('NACL::C::CommandInterface::ONTAP')
# However, it should be false for the following:
#     $router->isa('NACL::C::Vserver')
#     $router->isa('NACL::C::Cluster')
# To satisfy the above, we have the conditions listed in the code below.
# Note that we cannot just have
#    $kind->isa('NACL::C::CommandInterface')
# because that would then return true for the last two cases as well.
sub isa {
    $Log->enter() if $may_enter;

    my ($pkg_or_obj, $kind) = @_;
    my $isa;
    if ($kind->isa('NACL::C::SystemNode')) {
        $isa = 1;
    } elsif ($kind eq 'NACL::C::CommandInterface') {
        $isa = 1;
    } elsif ($kind eq 'NACL::C::CommandInterface::ONTAP') {
        $isa = 1;
    } else {
        $isa = $pkg_or_obj->SUPER::isa($kind);
    }

    $Log->exit() if $may_exit;
    return $isa;
}

1;
