#
# Copyright (c) 2014-20* NetApp Inc.
# All rights reserved.
#
## @summary An exception for when User doesnt have enough
##  permission to operate on the file/dir
##
## @author dl-nacl-dev@netapp.com
## @status Shared
## @pod here

package NACL::MTask::HealthChecker::ClusterHealth;

use base qw(NACL::MTask::HealthChecker);

use NACL::C::Node;
use Data::Dumper;
use NACL::C::Cserver;
use NACL::CS::Cluster;
use NACL::CS::NetworkInterface;
use NACL::CS::DebugVreport;
use NACL::CS::StorageAggregate;
use List::MoreUtils qw(uniq);

use feature 'state';
use Params::Validate qw(validate_with OBJECT SCALAR ARRAYREF);
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();

use NACL::APISet::Exceptions::ResponseException qw(:try);
use NACL::AppBuilder::Exceptions::HealthChecker;
use NATE::BaseException;

use Class::MethodMaker [
    array => 'dead_nodes',
];

=head1 METHODS

=head2 new

See the documentation for method 'new' in parent class
L<NACL::MTask::HealthChecker|lib-NACL-MTask-HealthChecker-pm/new>.

=cut

=head2 start

It is wrapper around base class' 'start' method. Since command are
clusterwide commands, it finds cserver for each the cluster and passes
them to SUPER::start.
Also See the documentation of base class method 'start'
L<NACL::MTask::HealthChecker|lib-NACL-MTask-HealthChecker-pm/start>.

=cut

sub start {
    my ($self) = @_;
    my @nodes = $self->nodes();

    my @cservers;
    my @dead_nodes;
    foreach my $node (@nodes) {
        if ($node->is_alive) {
            my $cserver = NACL::C::Cserver->find(command_interface => $node);
            push(@cservers, $cserver);
        } else {
            push(@dead_nodes, $node);
        }
    }

    @cservers = List::MoreUtils::uniq(@cservers);
    
    $self->dead_nodes(@dead_nodes);

    if(scalar(@cservers) == 0) {
        NACL::AppBuilder::Exceptions::HealthChecker->throw(
            "None of the nodes in the cluster are alive. " .
            "We need at least one node alive to be able run cluster commands"
        );
    }

    $self->SUPER::start(execute_in_parallel => 0, command_interfaces => \@cservers);
}

sub check_health_cluster_show {
    my ($self, %opts) = @_;
    my $cserver = $opts{command_interface};

    my @states = NACL::CS::Cluster->fetch(
        command_interface => $cserver,
        filter => { health => 'false' },
        allow_empty => 1,
    );
    if (@states > 0) {
        my $nodes = join(", ", map {$_->node} @states);
        my $msg = $nodes . " nodes are not healthy on cluster " . $cserver->name;
        $self->_report_error($msg);
    } else {
        $Log->comment( "All aggregates are online");
    }
}

sub check_health_cluster_ring_show {
    my ($self, %opts) = @_;
    my $cserver = $opts{command_interface};

    my @states = NACL::CS::ClusterRing->fetch(
        command_interface => $cserver,
        filter => { online => 'offline' },
        allow_empty => 1,
    );
    if (@states > 0) {
        my $nodes = join(", ", map {$_->unitname . " on node " . $_->node} @states);
        my $msg = "Units " . $nodes . " are offline for cluster " . $cserver->name;
        $self->_report_error($msg);
    } else {
        $Log->comment( "Node of cluster ring units are offline");
    }
}

sub check_health_debug_vreport_show {
    my ($self, %opts) = @_;
    my $cserver = $opts{command_interface};
    my @states = NACL::CS::DebugVreport->fetch(
        command_interface => $cserver,
        allow_empty => 1,
    );
}

sub check_health_storage_failover_show {
    my ($self, %opts) = @_;
    my $cserver = $opts{command_interface};
    my @states = NACL::CS::StorageFailover->fetch(
        command_interface => $cserver,
        allow_empty => 1,
    );
}

sub check_health_network_interface_show {
    my ($self, %opts) = @_;
    my $cserver = $opts{command_interface};

    my @states = NACL::CS::NetworkInterface->fetch(
        command_interface => $cserver,
        filter => { 'is-home' => 'false' },
        allow_empty => 1,
    );
    if (@states > 0) {
        my $lifs = join(", ", map {$_->lif . " on vserver " . $_->vserver} @states);
        my $msg = "lifs " . $lifs . " are not home for cluster " . $cserver->name;
        $self->_report_error($msg);
    } else {
        $Log->comment( "All interfaces are home");
    }
}

sub check_health_network_lif_status {
    my ($self, %opts) = @_;
    my $cserver = $opts{command_interface};

    my @states = NACL::CS::NetworkInterface->fetch(
        command_interface => $cserver,
        filter => { 'status-oper' => '!up' },
        allow_empty => 1,
    );
    if (@states > 0) {
        my $lifs = join(", ", map {$_->lif} @states);
        my $msg = "$lifs lifs are not up.";
        $self->_report_error($msg);
    } else {
        $Log->comment( "All lifs are operational");
    }
}

sub start_msg {
    my ($self, $method) = @_;

    state $start_msgs = {
        check_health_cluster_show =>
            "Checking for cluster health",
        check_health_cluster_ring_show =>
            "Check cluster ring show to detect rings that are currently out-of-quorum",
        check_health_debug_vreport_show =>
            "Checking for vldb inconsistencies",
        check_health_storage_failover_show =>
            "Storage failover show to make sure nodes aren't takenover",
        check_health_network_interface_show =>
            "Check for interfaces that aren't home",
        check_health_network_lif_status =>
            "Check for unhealthy lifs",
    };

    return $start_msgs->{$method};
}

1;
