#
# Copyright (c) 2014-20* NetApp Inc.
# All rights reserved.
#
## @summary An exception for when User doesnt have enough
##  permission to operate on the file/dir
##
## @author dl-nacl-dev@netapp.com
## @status Shared
## @pod here

package NACL::MTask::HealthChecker::FilerHealth;

use base qw(NACL::MTask::HealthChecker);

# use base qw(NACL::AppBuilder::HealthChecker::Driver);
use NACL::C::Node;
use Data::Dumper;
use NACL::CS::Volume;
use NACL::CS::StorageAggregate;

use feature 'state';
use Params::Validate qw(validate_with OBJECT SCALAR ARRAYREF);
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();

use NACL::APISet::Exceptions::ResponseException qw(:try);
use NATE::BaseException;

use Class::MethodMaker [
    scalar => 'mroot_min_available_size',
];

=head1 METHODS

=head2 new

See the documentation for method 'new' in parent class
L<NACL::MTask::HealthChecker|lib-NACL-MTask-HealthChecker-pm/new>.

=cut

sub new {
    my ($class, @args) = @_;

    state $spec = {
        nacltask_mroot_min_available_size => {type => SCALAR, default => '100m'},
    };

    my %opts = validate_with(
        params => \@args,
        spec   => $spec,
        allow_extra => 1,
    );

    my $mroot_min_available_size = delete $opts{nacltask_mroot_min_available_size};

    my $self = $class->SUPER::new(%opts);

    $self->mroot_min_available_size($mroot_min_available_size);

    bless($self, $class);

    return $self;
}

=head2 start

See the documentation for method 'start' in parent class
L<NACL::MTask::HealthChecker|lib-NACL-MTask-HealthChecker-pm/start>.

=cut

sub start {
    my ($self) = @_;
    $self->SUPER::start(execute_in_parallel => 1);
}

sub check_health_aggregates {
    my ($self, %opts) = @_;
    my $ci = $opts{command_interface};

    my @states = NACL::CS::StorageAggregate->fetch(
        command_interface => $ci,
        filter => { state => '!online' },
        allow_empty => 1,
    );
    if (@states > 0) {
        my $aggrs = join(", ", map {$_->aggregate} @states);
        my $msg = $aggrs . " aggregates are not online on node " . $ci->node;
        $self->_report_error($msg);
    } else {
        $Log->comment( "All aggregates are online");
    }
}

sub check_health_volume {
    my ($self, %opts) = @_;
    my $ci = $opts{command_interface};

    my @states = NACL::CS::Volume->fetch(
        command_interface => $ci,
        filter => { state => '!online' },
        allow_empty => 1,
    );
    my $node = $ci->node;
    if (@states > 0) {
        my $volumes = join(", ", map {$_->volume} @states);
        my $msg = $volumes . " volumes are not online on node " . $node;
        $self->_report_error($msg);
    } else {
        $Log->comment( "All volumes are online on node " . $node);
    }
}

sub check_health_mroot {
    my ($self, %opts) = @_;
    my $ci = $opts{command_interface};

    my $available_space = $self->mroot_min_available_size();
    my @states = NACL::CS::Volume->fetch(
        command_interface => $ci,
        filter => { 'is-mroot' => 'true', available => "<$available_space" },
        allow_empty => 1,
    );
    my $node = $ci->node;
    if (@states > 0) {
        my $volumes = join(", ", map {$_->volume . " on aggregate " . $_->aggregate} @states);

        my $msg = "mroot volumes: " . $volumes . " has less than $available_space space available.";
        $self->_report_error($msg);
    } else {
        $Log->comment("All rmoot volumes meet minimum requirement");
    }
}

sub check_health_zapi_server_check {
    my ($self, %opts) = @_;
    my $ci = $opts{command_interface};
    my $zapi_apiset = $ci->apiset(interface => 'ZAPI');
    $zapi_apiset->clock_get_clock();
    $Log->comment("ZAPI check passed");
}

sub start_msg {
    my ($self, $method) = @_;

    state $start_msgs = {
        check_health_aggregates =>
            "Checking for any unhealthy aggregates",
        check_health_volume =>
            "Checking for unhealthy volumes",
        check_health_zapi_server_check =>
            "Checking if ZAPI server is working",
        check_health_mroot =>
            "Checking if mroot is online and has sufficient space",
    };

    return $start_msgs->{$method};
}

1;
