# $Id$
# Copyright (c) 2011-2012 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary BlockingLoad Task Module
## @author Isshwarya.M@netapp.com, dl-nacl-dev@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::MTask::Load::BlockingLoad

=head1 SYNOPSIS

  use NACL::MTask::Load::BlockingLoad;

  # Initiate the load object
  my $load_obj = NACL::MTask::Load::BlockingLoad->new(
                              command_interface => $ci,
                              clients_info      => \@clients_info_AoH,
                    );
  # Start the load
  $load_obj->load_cpu(min => $min, max => $max, %other_optional_args);

  # Run the testcases which needs to be tested under I/O workload
  ...

  # Ensure that the load is still in the expected range
  $load_obj->load_cpu();

  # Continue testing
  ...

  # Stop the load
  $load_obj->stop();


=head1 DESCRIPTION

To ramp up the CPU load of filer running Data ONTAP to the specified range
and continue running the I/O tool with the tuned input values, till the caller
desires to stop the I/O load.

Once the load reaches the desired range, the first time tuning is deemed
to be completed and the control will be returned back to the caller.
The caller can test ONTAP under I/O load which continues to run in
the background with the tuned parameters. Note that,
further maintenance tuning which ensures that the load
continues to stay in the desired range will not be handled by this task and
the onus of verifying whether the CPU load is in the expected range or not and
doing tuning when needed is on the direct callers of this task.

If you need more control on tuning or if you desire non-periodic tuning, only
then this task would be a best fit, otherwise we recommend you to use
L<NACL::MTask::Load|lib-NACL-MTask-Load-pm>

If you want to keep loading the filer with the same amount, irrespective of
the CPU load overshoot caused by the test script (say, by invoking some ONTAP
operation which raises the CPU load), then you would not desire the
automatic maintenance tuning of L<NACL::MTask::Load|lib-NACL-MTask-Load-pm>
in which case, this task would be the best fit.

=head1 CONFIGURATION REQUIREMENTS

The caller is expected to establish mount points (with whatever read/write
sizes desired in the case of NAS mount). Handling of the setup steps by
the task requires more information about the test clients
and the cluster configuration and this is not covered in the initial version
of this task.

More details on this section can be referred
L<here|http://wikid.netapp.com/w/User:Isshwar/CPULoadTuning#Configuration_requirements>

=head1 NOTE

Please note that the tuning logic is in experimental stage and based on the user
feedback, the logic will be enhanced and fine tuned further.

=head1 ATTRIBUTES

=head2 command_interface

(Required, isa 'NACL::C::CommandInterface::ONTAP')
A command interface object which represents the cluster or node to be loaded.
See L<NACL::C::Component::command_interface|lib-NACL-C-Component-pm/command_interface>.

=head2 clients_info

(Required)
This array input takes the information about the clients and the absolute
file paths or paths to raw devices on which I/O can be started to load the
CPU of the filer. You can either provide an array of hash or an array of
L<NACL::MTask::Load::ClientInfo|lib-NACL-MTask-Load-ClientInfo-pm> objects.
Each array element corresponds to a traffic client that can be used to
for tuning purpose.

If you are providing array of hash as the input, then the hash keys can
be any of the attributes of
L<NACL::MTask::Load::ClientInfo|lib-NACL-MTask-Load-ClientInfo-pm>
'client_ci' holds the command_interface to the client and it is an object
of type L<NACL::C::Client|lib-NACL-C-Client-pm>. 'paths' is an array
of simple scalars holding the list of absolute file paths or paths to raw
devices. For other interesting options, that can be passed, please
refer to L<NACL::MTask::Load::ClientInfo|lib-NACL-MTask-Load-ClientInfo-pm>.
If you provide array of hashes as the input, the task would
turn them into objects of type NACL::MTask::Load::ClientInfo before storing
them.

Eg:

 my @clients = NACL::C::Client->find();
 my $load_obj = NACL::MTask::Load::BlockingLoad->new(
        command_interface => $ci,
        clients_info => [
            {client_ci =>$clients[0], paths => ['/mount/dir1/file1', '/mount/dir1/file2']},
            {client_ci =>$clients[1], paths => ['/dev/sda1', '/dev/sda2']},
            {client_ci =>$clients[2], paths => ['Y:\dir1', 'J:\dir2']},
        ]
 );

If you are providing array of NACL::MTask::Load::ClientInfo objects,
then each object should have 'client_ci' and 'paths' attribute filled in.

Eg:

 my $client_info_1 = NACL::MTask::Load::ClientInfo->new(
        client_ci => $linux_ci, paths => ['/mount/dir1/file1', '/mount/dir1/file2']);
 my $client_info_2 = NACL::MTask::Load::ClientInfo->new(
        client_ci => $windows_ci, paths => ['Y:\dir1', 'J:\dir2']);
 my $load_obj = NACL::MTask::Load::BlockingLoad->new(
        command_interface => $ci,
        clients_info => [$client_info_1, $client_info_2]
 );

Additional note is that,

$load_obj->client_info() will return an array of objects of type
NACL::MTask::Load::ClientInfo when invoked in array context and it would
return reference to the array when invoked in the scalar context.

=cut

package NACL::MTask::Load::BlockingLoad;

use strict;
use warnings;

# Base class
use base qw(NACL::MTask::MTask);

# Load Perl and NATE modules
use Tharn qw( sleep);
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Params::Validate qw(:all);

# Load Exceptions
use NATE::BaseException qw(:try);
use NATE::Exceptions::Argument;
use NACL::Exceptions::LoadCPUFailed;

# Load other needed NACL libraries
use NACL::C::Statistics;
use NACL::ComponentUtils qw(Dumper get_abs_path);
use NACL::MTask::Traffic;
use NACL::MTask::Load::_Mixin qw(clients_info_write_cb get_cpu_load);
use NACL::GeneralUtils qw(random_name_generator);

=head1 METHODS

=head2 new

  my $load_obj = NACL::MTask::Load::BlockingLoad->new(
                              command_interface => $ci,
                              clients_info           => \@clients_info_AoH,
                    );
Returns a new NACL::MTask::Load::BlockingLoad object. Refer to
L<ATTRIBUTES|lib-NACL-MTask-Load-BlockingLoad-pm/ATTRIBUTES> section
for the description of the input options.

=cut

use Class::MethodMaker [
    new   => [ '-hash', 'new' ],
    array => [
        {   '-store_cb' => sub {
                my $self = shift;
                return $self->clients_info_write_cb(@_);
            },

        },
        'clients_info'
    ],
    scalar => [
        { '-type' => 'NACL::C::CommandInterface::ONTAP' },
        'command_interface'
    ],
    hash => '_opts_cache',
];

=head2 load_cpu_once

  # By default, tunes the CPU load of the overall cluster.
  $load_obj->load_cpu_once (
        min => $min, max => $max,
        %other_optional_options
  );

  # Tune the CPU of a single node.
  $load_obj->load_cpu_once (
        min => $min, max => $max,
        load_entity => $node_name_hosting_the_volumes
        %other_optional_options
  );

(Instance method)

This method loads the CPU to the specified range by performing I/O
load tuning in the foreground. On reaching the specified range, this
method returns back the control to the caller. The I/O tool invocation
will not be stopped and it will continue to run in the background with
the tuned inputs.

Note that this method will not perform maintenance tuning in the
background, unlike
L<load_cpu method of NACL::MTask::Load|lib-NACL-MTask-Load-pm/load_cpu>
which takes care of maintaining the load in the desired level by regularly
checking the CPU load and trigger tuning, if CPU load goes out of range.

To explain briefly about the tuning logic, it starts the specified I/O
tool on all the clients and then it measures the CPU load of the cluster
or node, after the specified wait time. If the load is not in the
expected range, it kills the instances of the I/O tool and starts
a new instance of the tool with the tuned inputs. It repeats these steps,
untill the load reaches the desired range or the max_attempts is hit.

For interested readers, the tuning logic is described in detail at
L<CPU Load Tuning section|http://wikid.netapp.com/w/User:Isshwar/CPULoadTuning>

=head3 Options

=over

=item min

(Required)

The lower limit of the expected CPU load range. Valid range: 0-100

=item max

(Required)

The upper limit of the expected CPU load range. Valid range: 0-100

=item load_entity

(Optional)

The valid values are node name and 'cluster'. This specifies the
entity whose CPU load has to be tuned in the given range.
The value 'cluster' indicates that the CPU of the overall cluster has to be
tuned and the name of the cluster itself should not be passed.

NOTE: The support for loading the entire cluster is not supported
in the current revision and this will be added in the next revision.

Default: 'cluster'

=item max_attempts

(Optional)

The maximum number of times to iterations to tune the I/O load to be
between the values specified in the options C<min> and C<max>.

Default: 30

=item wait_time_in_tuning_phase

(Optional)

The wait time in seconds between the iterations of a tuning cycle. This is the
time to wait before measuring the CPU load and after starting the
I/O tool with the tuned inputs.

Default: 10

=item num_load_samples

(Optional)

The no.of CPU load samples to be taken to obtain the average CPU load.

Default: 3

=item sampling_interval

(Optional)

The time interval in seconds between the measurement of any two CPU samples.

Default: 30

=item tool

(Optional)

The name of the I/O tool. Currently only 'sio' is supported.

Default: sio

=item tune_io_step

(Optional)

Delay time (to be applied for I/O tool) in milliseconds per overshoot
in CPU load %.
If the measured CPU load is higher, then the tuning logic will
scale the delay time by the value specified in this option. This option
is still in experimental. If this option is not specified, the task
will increase the delay time by 1ms.

=item cache_options

(Optional)

This boolean option accepts 0/1 and specifies whether the task should cache the
input options or not.

Default: 0

Example usecases for enabling this option:

 my $load_obj = NACL::MTask::Load::BlockingLoad->new(%input_args);
 $load_obj->load_cpu_once(min => 40, max => 50, cache_options => 1);
 ....Load is in the expected range. Do the testing....
 ...
 $load_obj->load_cpu_once(); # Inputs need not be passed as they are cached
 ...
 # 'min' alone is overriden and 'max' will be picked up from cache
 $load_obj->load_cpu_once( min=> 35);
 ...
 # No inputs are specified again, 'min' is taken to be '40' now as the
 # cache is not updated in the last call.
 $load_obj->load_cpu_once();
 ...
 # Override 'max' limit, as well as update the cache
 $load_obj->load_cpu_once(max => '60', cache_options => 1);

=item traffic_options

(Optional)

This option takes a hash reference as the input and can be used to
override the I/O tool related input options.
All the input options accepted by L<NACL::MTask::Traffic|lib-NACL-MTask-Traffic-pm/start>.
It's recommended not to override C<on_exception>, C<on_success>, C<on_done>
handlers accepted by NACL::MTask::Traffic->start() through this option.

Note that overrides specified through this option will be applicable
for all the clients involved in the tuning.

Example:
   # considering 'sio' as the I/O tool, the following overrides instruct
   # the task to start with threadcount as 4 and 'random I/O' to be 40%
   # for the first iteration of the tuning cycle.
   traffic_options => {num_of_threads => 4, rand => 40}

=item per_client_traffic_options

(Optional)

This option can be used to override the I/O tool related input options on
a per client basis. It takes a hash reference as the input whose keys are
name (or id) of the clients and the corresponding values are traffic
related overrides for the I/O tool to be started in that client.

Example:
  # considering 'sio' as the I/O tool, the following overrides instruct
  # the task to start with threadcount as 4 and 'random I/O' to be 40%
  # on client1 and threadcount as 5 on client2.
  per_client_traffic_options => {
        client1 => {num_of_threads => 4, rand => 40},
        client2 => {num_of_threads => 5},
  }

=back

=cut

=head3 Exceptions

=over

=item NATE::Exceptions::Argument

If the the value specified to 'min' option is greater than 'max', then
this exception will be thrown.

Also if 'per_client_traffic_options'
and/or 'traffic_options' is specified and if 'tool' is not
specified, then this exception will be thrown.

=item NACL::Exceptions::LoadCPUFailed

When the CPU cannot be loaded to the specified range within the
specified 'max_attempts', then this exception will be thrown.

If the clients are overloaded and starts thrashing, then this
exception will be thrown.

=item NACL::Exceptions::TrafficFailed

NACL::Exceptions::TrafficFailed thrown by NACL::MTask::Traffic
which conveys the failures occured while running traffic will
simply be propagated.

=back

=cut


sub load_cpu_once {
    $Log->enter() if $may_enter;
    my ( $self, %orig_opts ) = @_;
    $self->_verify_invocation( style => 'instance_only' );
    $Log->debug( "Original input options are:\n" . Dumper( \%orig_opts ) )
        if ( $Log->may_debug() );
    my %opts;
    my $cache_options = delete $orig_opts{cache_options};
    if ( $self->_opts_cache_isset ) {

        # No inputs are specified.
        my %cache_opts = $self->_opts_cache();
        $Log->trace('Merging %cache_opts and %orig_opts');
        %opts = ( %cache_opts, %orig_opts );
        $Log->debug("Merged options are:\n". Dumper(\%opts))
            if ( $Log->may_debug() );
    } else {
        %opts = %orig_opts;
    }
    %opts = validate_with(
        params => [%opts],
        spec   => {
            min                        => { type => SCALAR },
            max                        => { type => SCALAR },
            max_attempts               => { type => SCALAR, default => 30 },
            wait_time_in_tuning_phase  => { type => SCALAR, default => 10 },
            num_load_samples           => { type => SCALAR, default => 3 },
            sampling_interval          => { type => SCALAR, default => 30 },
            traffic_options            => { type => HASHREF, default => {} },
            per_client_traffic_options => { type => HASHREF, default => {} },
            tool => {
                type      => SCALAR,
                optional  => 1,
                callbacks => {
                    "Value of 'tool' must be 'sio'" => sub {
                        if ( defined $_[0] ) { $_[0] =~ /^sio$/i }
                        }
                }
            },
            tune_io_step  => { type => SCALAR,  optional => 1 },
            cache_options => { type => BOOLEAN, default  => 0 },
            load_entity   => { type => SCALAR,  default  => 'cluster' },
        }
    );

    $opts{min} =~ s/\s*\%$//;
    $opts{max} =~ s/\s*\%$//;

    my $cpu_min = $opts{min};
    my $cpu_max = $opts{max};
    if ( $cpu_min > $cpu_max ) {
        $Log->exit() if $may_exit;
        NATE::Exceptions::Argument( "cpu_min cannot be greater than cpu_max: "
                . "cpu_min:$cpu_min, cpu_max:$cpu_max" );
    }

    if ($cache_options) {

        # %orig_opts is cached and %opts is intentionally avoided
        # so that we need not cache the options for which the task has
        # provided the defaults.

        # This is the convention of hash attributes declared using
        # Class::MethodMaker to update the specified keys alone without
        # overriding the complete hash structure.

        # %n = $x->h; # (a=>1,b=>2,c=>3) (in some order)
        # $x->h_set(b=>4,d=>7);
        # %n = $x->h; # (a=>1,b=>4,c=>3,d=>7) (in some order)
        #
        # per_client_traffic_options merged with traffic_options will already
        # be set by the tuning methods at this point of time. This will have
        # processed and tuned values which we wouldn't want to override here.
        delete $orig_opts{traffic_options};
        delete $orig_opts{per_client_traffic_options};
        $self->_opts_cache_set(%orig_opts);
    }

    $self->_load_cpu_logic(%opts);
    $Log->exit() if $may_exit;
}

=head2 stop

    $load_obj->stop();

(Instance method)

This method stops the I/O load started earlier to load the CPU.

=cut

sub stop {
    $Log->enter() if $may_enter;
    my $self         = shift;
    my @clients_info = $self->clients_info;
    foreach my $client (@clients_info) {
        if ( $client->traffic_obj_isset ) {
            $client->traffic_obj->stop( if_stopped => 'pass' );
            $client->traffic_obj_reset;
        }
    }
    $Log->exit() if $may_exit;
}

=head2 get_cpu_load

The description can be referred at,
L<get_cpu_load|lib-NACL-MTask-Load-_Mixin-pm/get_cpu_load>

=cut

####################################
#       Private Methods            #
####################################

sub _load_cpu_logic {
    $Log->enter() if $may_enter;
    my ( $self, %opts ) = @_;
    $Log->debug( "Opts to '_load_cpu_logic':\n" . Dumper( \@_ ) )
        if ( $Log->may_debug() );

    my $per_client_traffic_options = $opts{per_client_traffic_options};
    my %traffic_options            = %{ delete $opts{traffic_options} };
    my $tool                       = delete $opts{tool};

    # When only one of the hash is specified, merge has to happen
    if (   ( scalar %traffic_options )
        || ( scalar %$per_client_traffic_options ) )
    {
        foreach my $client_info_obj ( $self->clients_info ) {

            # Merge traffic_options (which is applicable for all clients)
            # and per client overrides.
            my $client_name = $client_info_obj->get_id;
            my %overrides_per_client
                = ( defined $per_client_traffic_options->{$client_name} )
                ? %{ $per_client_traffic_options->{$client_name} }
                : ();
            $per_client_traffic_options->{$client_name}
                = { %traffic_options, %overrides_per_client };
        }
    }

    # When traffic overrides are specified, tool name should be compulsorily
    # specified
    if ( scalar(%$per_client_traffic_options) && !( defined $tool ) ) {
        $Log->exit() if $may_exit;
        NATE::Exceptions::Argument->throw(
                  'If either of \'traffic_options\' or '
                . '\'per_client_traffic_options\' is specified, then '
                . 'tool name must also be specified in the option \'tool\''
                . 'The merged traffic_options and per_client_traffic_options is'
                . "\n"
                . Dumper($per_client_traffic_options) );
    }

    # Set the default
    unless ( defined $tool ) {
        $tool = 'sio';
    }

    # Add support for other tools - This section will be expanded
    if ( $tool eq 'sio' ) {
        $self->_load_cpu_using_sio(%opts);
    }

    $Log->exit;
}

# Create files in the mount dir (or workdir) for each client
sub _create_files {
    $Log->enter() if $may_enter;
    my $self = $_[0];
    foreach my $client_info_obj ( $self->clients_info ) {
        unless ( $client_info_obj->paths_isset ) {
            my $workdir = $self->get_abs_path($client_info_obj->workdir);
            $workdir =~ s/[\\\/]$//;
            my $filecount = $client_info_obj->file_count;
            my @paths     = ();
            my $apiset
                = $client_info_obj->client_ci->apiset( cache_apiset => 0 );
            for ( my $i = 1; $i <= $filecount; $i++ ) {
                my $singlepath
                    = random_name_generator( prefix => "${i}_file" );
                # By default, we use the Linux dir separator.
                my $char = '/';
                if ( $client_info_obj->client_ci->is_windows ) {
                    # Update if it is windows client
                    $char = '\\';
                }
                $singlepath = $workdir . $char . $singlepath;
                $Log->comment(
                    "Info: Creating the file '$singlepath' in the client "
                        . $client_info_obj->get_id );

                # A very light weight approach of file creation is needed.
                # NACL::C::Client::Remote::File starts a remote process which
                # needs NATE_LIB, LOGDIR set appropriately and this is a
                # hard requirement on the Traffic users.
                # Hence used OS based commands intentionally.
                if ( $client_info_obj->client_ci->is_windows ) {

                    $apiset->copy(
                        'suppress-prompt' => 1,
                        source           => 'NUL',
                        destination      => $singlepath
                    );
                } else {
                    $apiset->touch( path => $singlepath );
                }
                push( @paths, $singlepath );
            }
            $client_info_obj->paths(@paths);
        }
    }
    $Log->exit() if $may_exit;
}

# Load CPU to the desired range using sio tool
sub _load_cpu_using_sio {
    $Log->enter() if $may_enter;
    my ( $self, %opts ) = @_;

    my %traffic_opts = %{ $opts{per_client_traffic_options} };
    my $cpu_min      = $opts{min};
    my $cpu_max      = $opts{max};

    # Check if files need to be created for any of the client.
    # files will be created only for the first time
    $self->_create_files();

    # Options to get the CPU load.
    my %stats_inputs;
    $self->_hash_move(
        target => \%stats_inputs,
        source => \%opts,
        move   => [qw(num_load_samples sampling_interval load_entity)]
    );

    # subroutine to get cpu load
    # TODO #############################
    my $get_cpu_load = sub {
        my $current_cpu_load;
        try {
            $current_cpu_load = $self->get_cpu_load(%stats_inputs);
        }
        catch NACL::Exceptions::TrafficFailed with {

            # We have received TrafficFailed Exception from the worker
            # which is performing IO and so just rethrow it.
            my $ex = shift;
            $self->stop;
            $ex->throw;
        }
        catch NATE::BaseException with {
            my $ex = shift;
            $Log->debug( "Error is of type "
                    . ref($ex)
                    . " and the reason is:\n"
                    . $ex->text() );
            $self->stop;
            NACL::Exceptions::LoadCPUFailed->throw(
                      "Failed to load the CPU to the specified range "
                    . "because the following error is hit while trying "
                    . "to measure the CPU load:\n"
                    . $ex->text() );
        };
        return $current_cpu_load;
    };

    $self->_opts_cache_set( tool => 'sio' );

    # Start sio for the first time
    my $load_range = $opts{min} . '% - ' . $opts{max} . '%';
    unless ( $self->clients_info->[0]->traffic_obj_isset ) {

        # Traffic object is not set in the first client which should be
        # the same case for other clients as well because we start
        # traffic from all the clients specified and then control
        # the cpu load with the help of tuning the parameters but still
        # running some load from all the clients specified.
        $Log->comment(
            'Info: Setting up initial load to achieve the range ' . $load_range );

        # Traffic object will not be set only for the first tuning
        # and so, the load started at this point would be initial load.
        foreach my $client_info_obj ( $self->clients_info ) {

            # Build sio options
            my $client_name = $client_info_obj->get_id;
            my %sio_options
                = ( defined $traffic_opts{$client_name} )
                ? %{ $traffic_opts{$client_name} }
                : ();
            $self->_populate_defaults_for_sio(
                traffic_options => \%sio_options,
                command_interface => $client_info_obj->client_ci
            );

            # Store it back for further tuning
            $traffic_opts{$client_name} = \%sio_options;
            $Log->comment( 'Info: Starting traffic in the client \''
                    . $client_name
                    . "\'.\nThreadcount: "
                    . $sio_options{num_of_threads}
                    . ', think_time: '
                    . $sio_options{think} );
            my $obj = NACL::MTask::Traffic->start(
                %sio_options,
                paths             => [ $client_info_obj->paths ],
                command_interface => $client_info_obj->client_ci
            );
            $client_info_obj->traffic_obj($obj);
        }
        $Log->debug( "wait time after first load is "
                . $opts{wait_time_in_tuning_phase} );
        sleep( $opts{wait_time_in_tuning_phase} );
    }

    my $tuning_completed = 0;

    # To start with, all the clients would be involved in tuning
    my @clients_in_tuning = $self->clients_info;

    # Instead of setting undef, -1 is set here which will be considered as
    # an unseen load
    my $previous_cpu_load = -1;

    # Reference: SANInterop/io/host_io.thpl
    # If the current cpu load continues to be less than the previous
    # measured load continuously for three times, then it's an indication of
    # thrashing.
    my $badness = 0;

    # -think option of sio cannot go lesser than 0ms
    my $sio_delay_check = sub {
        my $value = shift;
        $value = ( $value >= 0 ) ? $value : 0;
        $value = int($value);
        return $value;
    };
    my $current_cpu_load;
    my $fine_tuning = 0;
    my ($is_threads_last_tuned_for_deficient,
        $is_threads_last_tuned_for_overshoot
    ) = ( 0, 0 );
    my $current_tuned_clients_count  = scalar(@clients_in_tuning);
    my $previous_tuned_clients_count = $current_tuned_clients_count;
    for ( my $i = 1; $i <= $opts{max_attempts}; $i++ ) {
        $current_cpu_load = $get_cpu_load->();
        $Log->comment( "Info: Tuning attempt: $i out of $opts{max_attempts}. "
                . "current cpu load is $current_cpu_load. Expected range is $load_range"
        );
        if (   ( $current_cpu_load >= $cpu_min )
            && ( $current_cpu_load <= $cpu_max ) )
        {
            $Log->comment(
                "Info: Load is within the range. Current load is ${current_cpu_load}\%"
            );
            $tuning_completed = 1;

            # Store the tuned values so that future tunings can start from the
            # tuned values of previous tuning cycle.
            $self->_opts_cache_reset('traffic_options');
            $self->_opts_cache_set(
                per_client_traffic_options => \%traffic_opts );
            last;
        } else {
            # Load is out of given range.
            # Stop the load in all the clients, let's tune the parameters
            # and restart the I/O tool.
            $Log->trace('Load is not in the given range');
            $self->stop();
            # $is_threads_last_tuned_for_deficient is set, then it indicates that the
            # thread count is increased in the last iteration. If the CPU load
            # still remains low, then it's a bad indication.
            # This assumes that the thread count is increased across all clients
            # involved in the tuning (the individual values may be different, but
            # the values are increased from the previous value) if $is_threads_last_tuned_for_deficient
            # is set and that's how the current logic works.
            if ( ($current_cpu_load < $previous_cpu_load) && $is_threads_last_tuned_for_deficient ) {

                # May be an indication of thrashing
                $badness++;
            } else {

                # Reset
                $badness = 0;
            }
            if ( $badness == 3 ) {
                $Log->exit() if $may_exit;
                NACL::Exceptions::LoadCPUFailed->throw(
                      "Increasing thread count has not helped after 3 "
                    . "consecutive tuning attempts that did not yield "
                    . "any progress.\nGiving up, please check to make sure "
                    . "your hosts are not overloaded and are not thrashing"
                );
            }

           # Load has just slipped through the specified range, so fine tuning
           # is needed. Start tuning 3/4 th no.of clients at this point.
           # Loads are comparable this way, only if the no.of clients used in the
           # previous and current iterations are the same.
            if (($current_tuned_clients_count == $previous_tuned_clients_count) &&
                # This is a new tuning cycle, hence previous_cpu_load is -1
                ($previous_cpu_load != -1) &&
                ((      ( $previous_cpu_load < $cpu_min )
                    && ( $current_cpu_load > $cpu_max )
                )
                || (   ( $previous_cpu_load > $cpu_max )
                    && ( $current_cpu_load < $cpu_min ) )
                ))
            {
                $Log->trace( "Load has just slipped the specified range: "
                        . "$previous_cpu_load to $current_cpu_load" );
                $fine_tuning                  = 1;
                $previous_tuned_clients_count = scalar(@clients_in_tuning);
                @clients_in_tuning            = @clients_in_tuning[ 0 ..
                    int( $#clients_in_tuning * 0.75 ) ];
                $current_tuned_clients_count = scalar(@clients_in_tuning);
            } else {
                $previous_tuned_clients_count = $current_tuned_clients_count;
                # This below check is mostly useful when load_cpu_once is invoked
                # as part of maintenance tuning cycles and the overshoot/deficient
                # is very less.
                if( abs($current_cpu_load-$cpu_min) <10 || abs($current_cpu_load-$cpu_max) < 10) {
                    # Load is out of range by just 10%. This calls for fine tuning.
                    $fine_tuning                  = 1;
                }
            }

            $Log->trace( "No.of clients involved in tuning: "
                    . scalar(@clients_in_tuning) );
            $Log->trace( sub { "Clients are,\n" . Dumper \@clients_in_tuning }
            );
            if ( $current_cpu_load < $cpu_min ) {
                if ( $current_tuned_clients_count
                        == $previous_tuned_clients_count ) {
                    if ($fine_tuning) {
                        $Log->comment( "Info: Load is less. Either one of ".
                            "threadcount or delay will be tuned. ".
                            "Current load is ${current_cpu_load}\%"
                        );
                    } else {
                        $Log->comment( "Info: Load is less. Increasing the thread count, "
                                . "decreasing the think time. ".
                                "Current load is ${current_cpu_load}\%"
                        );
                    }
                }
                foreach my $client_info_obj (@clients_in_tuning) {
                    my $client_name = $client_info_obj->get_id;
                    my %sio_options = %{ $traffic_opts{$client_name} };

                    # When previous and current clients count is not the same,
                    # this means that no.of clients is altered and so use the
                    # same thread count and delay time as set earlier.
                    # Otherwise, tune the threadcount/delay params.
                    if ( $current_tuned_clients_count
                        == $previous_tuned_clients_count )
                    {
                        unless ($fine_tuning) {
                            $Log->trace('Fine tuning zone is not hit');
                            $sio_options{num_of_threads}++;
                            $sio_options{think}
                                = ( $sio_options{think} == 0 )
                                ? 0
                                : $sio_options{think} / 2;
                        } else {

                  # Tune either threads or delay time as we are in the zone of
                  # fine tuning.
                            my $threads = $sio_options{num_of_threads};
                            my $think   = $sio_options{think};
                            my ( $new_threads, $new_think );
                            if (   $think
                                && $is_threads_last_tuned_for_deficient )
                            {

                                # Tuning think time
                                $new_think = $think / 2 ;
                                $new_threads = $threads;
                                $is_threads_last_tuned_for_deficient = 0;
                            } else {

                                # Tuning thread count
                                $new_threads = $threads + 1;
                                $new_think   = $think;
                                $is_threads_last_tuned_for_deficient = 1;
                            }
                            $sio_options{num_of_threads} = $new_threads;
                            $sio_options{think}          = $new_think;
                        }
                    } else {
                        $Log->trace('Client count has changed from '.
                            $previous_tuned_clients_count . ' to '.
                            $current_tuned_clients_count
                        );
                    }
                    $sio_options{think}
                        = $sio_delay_check->( $sio_options{think} );
                    $Log->comment( 'Info: Starting traffic in the client '
                            . $client_name
                            . ". Threadcount: "
                            . $sio_options{num_of_threads}
                            . ', think_time: '
                            . $sio_options{think} );
                    %{ $traffic_opts{$client_name} } = %sio_options;
                    my $obj = NACL::MTask::Traffic->start(
                        %sio_options,
                        paths             => [ $client_info_obj->paths ],
                        command_interface => $client_info_obj->client_ci
                    );
                    $client_info_obj->traffic_obj($obj);
                }
            } else {
                if ( $current_tuned_clients_count
                        == $previous_tuned_clients_count ) {
                    if ($fine_tuning) {
                        $Log->comment( "Info: Load is high. Either one of ".
                            "threadcount or delay will be tuned. ".
                            "Current load is ${current_cpu_load}\%"
                        );
                    } else {
                        $Log->comment( "Info: Load is high. Decreasing the thread count, "
                                . "increasing the think time. ".
                                "Current load is ${current_cpu_load}\%"
                        );
                    }
                }
                foreach my $client_info_obj (@clients_in_tuning) {
                    my $client_name = $client_info_obj->get_id;
                    my %sio_options = %{ $traffic_opts{$client_name} };

                    # When previous and current clients count is not the same,
                    # this means that no.of clients is altered and so use the
                    # same thread count and delay time as set earlier.
                    if ( $current_tuned_clients_count
                        == $previous_tuned_clients_count )
                    {
                        unless ($fine_tuning) {
                            $Log->trace('Fine tuning zone is not hit');

                            $sio_options{num_of_threads}
                                = ( $sio_options{num_of_threads} == 1 )
                                ? $sio_options{num_of_threads}
                                : --$sio_options{num_of_threads};
                            $sio_options{think}++;
                        } else {

                  # Tune either threads or delay time as we are in the zone of
                  # fine tuning.
                            my $threads = $sio_options{num_of_threads};
                            my $think   = $sio_options{think};
                            my ( $new_threads, $new_think );

                            if ( ( $threads > 1 )
                                && !$is_threads_last_tuned_for_overshoot )
                            {

                                # Tuning thread count
                                $new_threads = $threads - 1;
                                $new_think   = $think;
                                $is_threads_last_tuned_for_overshoot = 1;
                            } else {

                                # Tuning think time
                                $new_think   = $think + 1;
                                $new_threads = $threads;
                                $is_threads_last_tuned_for_overshoot = 0;
                            }
                            $sio_options{num_of_threads} = $new_threads;
                            $sio_options{think}          = $new_think;
                        }

                    } else {
                        $Log->trace('Client count has changed from '.
                            $previous_tuned_clients_count . ' to '.
                            $current_tuned_clients_count
                        );
                    }
                    $sio_options{think}
                        = $sio_delay_check->( $sio_options{think} );
                    $Log->comment( 'Info: Starting traffic in the client \''
                            . $client_name
                            . "\'. Threadcount: "
                            . $sio_options{num_of_threads}
                            . ', think_time: '
                            . $sio_options{think} );
                    %{ $traffic_opts{$client_name} } = %sio_options;
                    my $obj = NACL::MTask::Traffic->start(
                        %sio_options,
                        paths             => [ $client_info_obj->paths ],
                        command_interface => $client_info_obj->client_ci
                    );
                    $client_info_obj->traffic_obj($obj);
                }
            }

            # For next cycle, current load would be the previous load
            $previous_cpu_load = $current_cpu_load;

           # Load is started, wait for the specified time before measuring the
           # current cpu load
            $Log->debug( "wait time in tuning phase is "
                    . $opts{wait_time_in_tuning_phase} );
            sleep( $opts{wait_time_in_tuning_phase} );
        }
    }
    unless ($tuning_completed) {
        $Log->debug('Stop any running load, if any');
        $self->stop();
        $Log->exit() if $may_exit;
        NACL::Exceptions::LoadCPUFailed->throw(
                  "Unable to tune the CPU load to the desired range "
                . "$load_range even after the max_attempts of "
                . $opts{max_attempts} );
    }

    $Log->exit() if $may_exit;
}

=head2 Task selected defaults for I/O tools

=head3 sio

=over

=item 'num_of_threads'- '3'

=item 'think' - '0'

=item 'read' - '0'

=item 'rand' - '0'

=item 'block_size' - '64K'

=item 'start_byte' - '0'

=item 'file_size' - '50m'

=item 'direct' - '1'

=item 'noflock' - '1'

=item 'no_dsync'- '1'

=item 'disable_host_monitoring' - '1'

=back

=cut

sub _populate_defaults_for_sio {
    $Log->enter() if $may_enter;
    my ( $self, %args ) = @_;
    my $opts = $args{traffic_options};
    my $ci = $args{command_interface};

    my $set_default = sub {
        my ( $option, $default ) = @_;
        unless ( defined $opts->{$option} ) {
            $opts->{$option} = $default;
        }
    };

    $set_default->( 'num_of_threads',          '3' );
    $set_default->( 'think',                   '0' );
    $set_default->( 'read',                    '0' );
    $set_default->( 'rand',                    '0' );
    $set_default->( 'block_size',              '64K' );
    $set_default->( 'start_byte',              '0' );
    $set_default->( 'file_size',               '50m' );
    $set_default->( 'direct',                  '1' );
    # The below two options are linux specific options
    unless ($ci->is_windows) {
        $set_default->( 'noflock',                 '1' );
        $set_default->( 'no_dsync',                '1' );
    }
    $set_default->( 'disable_host_monitoring', '1' );

    # No choice in the tool name at this point
    $opts->{tool} = 'sio';

    $Log->exit() if $may_exit;
}

1;
