# $Id: //depot/prod/test/nacldev/lib/NACL/MTask/Load/_Mixin.pm#1 $

#
# Copyright (c) 2011-2012 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Common methods for NACL::MTask::Load and NACL::MTask::Load::BlockingLoad
## @author Isshwarya.M@netapp.com, dl-nacl-dev@netapp.com
## @status shared

package NACL::MTask::Load::_Mixin;

use strict;
use warnings;

use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use NATE::Exceptions::Argument qw(:try);
use NACL::MTask::Load::ClientInfo;
use NACL::C::Statistics;
use NATE::BaseException;
use NACL::ComponentUtils qw(Dumper);
use Params::Validate qw(:all);


BEGIN {
    use Exporter qw(import);
    our @EXPORT_OK = qw(
        clients_info_write_cb
        get_cpu_load
    );
    our %EXPORT_TAGS = ( 'all' => \@EXPORT_OK );
}

# Callback to validate 'clients_info' input
sub clients_info_write_cb {
    $Log->enter() if $may_enter;
    my ($self, $value) = @_;

    # Fix how the array attribute is stored in the object
    my @clients_info = @$value;
    if (ref($clients_info[0]) eq 'ARRAY' && scalar(@clients_info) == 1) {
        @clients_info = @{$clients_info[0]};
    }

    # validate clients_info
    my @clients_info_obj;
    foreach my $client (@clients_info) {
        if (ref($client) eq 'HASH') {
            push(@clients_info_obj, NACL::MTask::Load::ClientInfo->new(%$client));
        } elsif ($client->isa('NACL::MTask::Load::ClientInfo')) {
            push(@clients_info_obj, $client);
        } else {
            $Log->exit() if $may_exit;
            NATE::Exceptions::Argument->throw('Each array element passed to '.
                '\'clients_info\' should either be a hash reference or an object'.
                ' of type NACL::MTask::Load::ClientInfo, but you have '.
                'passed, '. "\n" . Dumper($client));
        }
    }
    $Log->exit() if $may_exit;
    return \@clients_info_obj;
}


=head2 get_cpu_load

  my $cpu_load = $load_obj->get_cpu_load ();

(Instance method)

=head3 Options

=over

=item num_load_samples

(Optional)

The no.of CPU load samples to be taken to obtain the average CPU load.

Default: 3

=item sampling_interval

(Optional)

The time interval in seconds between the measurement of any two CPU samples.

Default: 30

=item load_entity

(Optional)

The valid values are node name and 'cluster'. This specifies the
entity whose CPU load has to be tuned in the given range.
The value 'cluster' indicates that the CPU of the overall cluster has to be
tuned and the name of the cluster itself should not be passed.

This field is not applicable for 7Mode and if passed, it would be ignored.

Default: 'cluster'

=item handle_fluctuations

(Optional)

This is a boolean option.
If the delta between minimum - average - maximum samples of CPU load is
greater than 10%, then it's considered that the load is highly fluctuating
and enabling this option will result in taking more samples as indicated
by C<max_tries> option.

Default: 1

=item max_retries

(Optional)

This option takes an integer input which gives the max no.of times, the
CPU load% can be measured, when the load is in fluctuation state.

Default: 3

Eg:
   $self->get_cpu_load(handle_fluctuations => 1, max_retries => 2, num_load_samples => 3, sampling_interval => 10);

When the load is in fluctuation state, the above call will invoke the
command 'statistics show-periodic' twice as,
   "statistics show-periodic -iteration 3 -interval 10"
   "statistics show-periodic -iteration 3 -interval 10"

Note that if the load is not fluctuating, then it will invoke the command
only once, irrespective of the value specified to C<max_retries>

=back

=cut

sub get_cpu_load {
    $Log->enter() if $may_enter;
    my ($self, @args) = @_;
    $self->_verify_invocation( style => 'instance_only' );
    my %opts = validate_with(
        params => \@args,
        spec   => {
            num_load_samples => { type => SCALAR, default => 3},
            sampling_interval => { type => SCALAR, default => 30},
            load_entity       => { type => SCALAR, default => 'cluster'},
            handle_fluctuations => { type => BOOLEAN, default => 1},
            max_tries         => { type => SCALAR, default => 3},
        }
    );
    my $current_cpu_load;
    my $load_entity = delete $opts{load_entity};
    my $handle_fluctuations = delete $opts{handle_fluctuations};
    my $max_tries = delete $opts{max_tries};
    my $repeat = 0;
    my %additional_opts = ();

    my ($object, $instance, $field);
    if ($load_entity eq 'cluster') {
        $object = 'cluster';
        $instance = 'cluster';
        $field  = 'cpu_busy';
    } else {
        if ($self->command_interface->version_manager->is_version_ge(release => "8.2")) {
            $object = 'processor:node',
            $instance = $load_entity;
            $field = 'processor_busy';
        } else {
            $object = 'node';
            $instance = 'node';
            $field = 'cpu_busy';
            $additional_opts{node} =  $load_entity;
        }
    }

    if ($self->command_interface->is_7mode) {
        $field = 'cpu_busy';
    }
    my $statistics;
    my $remove_percentage = sub {
        my $val = shift;
        $val =~ s/\%\s*$//;
        return $val;
    };
 READ: {
    $statistics = NACL::C::Statistics->show_periodic(
        command_interface => $self->command_interface,
        summary => 'true',
        iterations => $opts{num_load_samples},
        interval => $opts{sampling_interval},
        object => $object,
        instance => $instance,
        %additional_opts
    );
    my ($avg, $min, $max);
    if ($statistics->averages_isset()) {
        $avg = $statistics->averages()->$field();
        $min = $statistics->minimums()->$field();
        $max = $statistics->maximums()->$field();
        if ($handle_fluctuations) {
            # If the delta between min-avg-max is > 10%, then the traffic
            # has not reached stable state. We need to take some more
            # samples to get more accurate current state.
            $avg = $remove_percentage->($avg);
            $min = $remove_percentage->($min);
            $max = $remove_percentage->($max);
            if ((($avg-$min)>=10) || (($max-$avg) >= 10)) {
                $repeat++;
                unless ($repeat > $max_tries) {
                    redo READ;
                }
            } else {
                $current_cpu_load = $avg;
            }
        } else {
            $current_cpu_load = $statistics->averages()->$field();
        }
    } else {
        # In case of single sample, summary will not be displayed and so
        # we will have to get the cpu_load from the single sample directly.
        $current_cpu_load = $statistics->statistics()->[0]->$field();
    }
}
    if (defined $current_cpu_load) {
        $current_cpu_load = $remove_percentage->($current_cpu_load);
    } else {
        $Log->exit() if $may_exit;
        NATE::BaseException->throw(
            "Unable to obtain the cpu load information.".
            "The obtained statistics object had,\n". Dumper($statistics));
    }
    $Log->exit() if $may_exit;
    return $current_cpu_load;
}
1;

