#
# Copyright (c) 2013 NetApp, Inc., All Rights Reserved
# Any use, modification, or distribution is prohibited
# without prior written consent from NetApp, Inc.
#
#

## @summary SPEventLogDetector Task Module
## @author rahula@netapp.com, dl-nacl-dev@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::MTask::LogDetector::SPEventLogDetector

=head1 DESCRIPTION

This task is to be used for validating the sp events logs for presence or
absence of given message/regex.

Many testcases require verification that certain sp event log messages have or
have not been triggered. This package provides methods for selecting the time
ranges between which the event log messages will be verified
(for presence/absence) of the given sp events message or regex.

=head2 EXAMPLE

    #Tests which calls check() by default when $SPEventLogDetector->stop()
     is called.
    use NACL::MTask::LogDetector::SPEventLogDetector;
    ...
    my $detector = NACL::MTask::LogDetector::SPEventLogDetector->new(
        sp_apiset_object => $Sp_Apiset_Obj,
        events           => \@events,
    );
    $detector->start();
    ...
    #Perform any operations here
    ...
    $detector->stop();

    #End test

    #Tests which explicitly calls check().
    use NACL::MTask::LogDetector::SPEventLogDetector;
    ...
    my $detector = NACL::MTask::LogDetector::SPEventLogDetector->new(
        sp_apiset_object  => $Sp_Apiset_Obj,
        events            => \@events,
    );
    $detector->start();
    ...
    # Perform any operations here
    ...
    $detector->stop('check' => 0);
    #do wait/or some other operations
    $detector->check();
    # End test

    Also note that if the SP login is rebooted or if the SP connection is lost
    during operation, the SP login session needs to be reconnected.
    Using the below command,
    $APIset_Object->api_reconnect_session();

=head1 ATTRIBUTES

=head2 sp_apiset_object

(Required)
The SP login APISet object of the filer on which we are checking for the sp events.
See L<NACL::APISet::Node::CLI::SP|lib-NACL-APISet-Node-CLI-SP-pm>.

=head2 event_begin_time

(Required) The begin time from which we need to check the particular sp_event.

=head2 event_end_time

(Required) The end time before which we need to check the particular sp_event.

=head2 events

(Required) This attribute will contain all the events present in the sp events log.

=head2 count

(Required) The count which contains the number of times the particular message
    has been pattern matched.

=head2 matched_records

(Required) The pattern matched records which are found in the sp events log for
    the timeframe.

=head1 EXCEPTIONS

    1. NACL::Exceptions::SPEventCheckFailure
       (If none of the events passed to check_for_any_presence parameter are not
        found in the sp events log)
    2. NACL::Exceptions::SPEventCheckFailure
       (If any one of the events passed to check_for_all_presence parameter are
        not found in the sp events log)
    3. NACL::Exceptions::SPEventCheckFailure
       (If any one of the events passed to check_for_absence parameter are found
        in the sp events log)
    4. NACL::Exceptions::InvalidParam
       (If the event_start_time is greater than event_end_time)
    5. NACL::Exceptions::InvalidParam
       (This exception occurs when none of the options check_for_any_presence/
        check_for_all_presence/check_for_absence are passed)

=cut

package NACL::MTask::LogDetector::SPEventLogDetector;

use strict;
use warnings;

use base qw(NACL::MTask::MTask);

use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Params::Validate
    qw(validate validate_with BOOLEAN HASHREF OBJECT ARRAYREF SCALAR);
use NATE::Exceptions::Argument qw(:try);
use NACL::Exceptions::SPEventCheckFailure qw(:try);
use NACL::Exceptions::InvalidParam qw(:try);
use NACL::ComponentUtils qw( Dumper enumerate_month);
use Date::Parse;

use constant DEFAULT_TIMEOUT  => 120;

BEGIN {
    use Exporter qw(import);
    our %EXPORT_TAGS
        = (
        all => [qw( start stop check )]
        );
    our @EXPORT_OK = qw( get_sysdate );
} ## end BEGIN

=head1 METHODS

=head2 new

my $detector = NACL::MTask::LogDetector::SPEventLogDetector->new(
    sp_apiset_object => $Node,
);

Create a MTask::LogDetector::SPEventLogDetector object that can be used to check for the
presence or absence of events during a time range.

=over

=item Options

=over

=item C<< sp_apiset_object => $apiset_obj >>

(Required)
The command interface on which to read and verify the event logs.
See L<NACL::APISet::Node::CLI::SP|lib-NACL-APISet-Node-CLI-SP-pm>.

=back

=back

=cut

use Class::MethodMaker [
    new    => ['-hash', 'new'],
    array  => 'events',
    scalar => 'event_begin_time',
    scalar => 'event_end_time',
    scalar => 'count',
    scalar => 'pattern',
    hash   => 'matched_records',
    scalar =>
        [{-type => 'NACL::APISet::Node::CLI::SP'}, 'sp_apiset_object'],
];

# Validate the parameters

sub init {
    $Log->enter() if $may_enter;
    my ($self, @opts) = @_;
    my %opts = validate_with(
        params => \@opts,
        spec   => {
            sp_apiset_object       => {type => OBJECT},
            events                 => {type => ARRAYREF, optional => 1},
            check_for_all_presence => {type => ARRAYREF, optional => 1},
            check_for_absence      => {type => ARRAYREF, optional => 1},
            check_for_any_presence => {type => ARRAYREF, optional => 1},
            event_begin_time       => {type => SCALAR, optional => 1},
            event_end_time         => {type => SCALAR, optional => 1},
            matched_records        => {type => HASHREF, optional => 1},
            count                  => {type => SCALAR, optional => 1},
            %{NACL::C::Component->_common_validate_spec_without_ci()}
        },
    );
    $Log->exit() if $may_exit;
} ## end sub init

=head2 start

    $SPEventLogDetector->start();

(Instance method)
Marks the time stamp which will be used as the starting point for sp events log
to be validated.

This does not have any return value. Instead, it updates the time stamp on
the instance with the key "event_begin_time".

=cut

sub start {
    $Log->enter() if $may_enter;
    my ($self, @opts) = @_;
    my %opts = validate_with(
        params => \@opts,
        spec   => NACL::C::Component->_common_validate_spec_without_ci()
    );

    my $apiset_obj = $self->sp_apiset_object();
    my $sys_date
        = $self->get_sysdate('sp_apiset_object' => $apiset_obj, %opts);
    $self->event_begin_time($sys_date);
    $Log->exit() if $may_exit;
} ## end sub start

=head2 stop

    $SPEventLogDetector->stop();

(Instance method)
Marks the time stamp which will be used as the end point for sp events log to be
validated. By default, this method internally calls check() to validate the logs
unless the param 'check' is passed with value 0.

=over

=item Options

=over

=item C<< check_for_all_presence => $array_ref >>

( Optional, defaults to reference to a null array )
Reference to a list of events names to check. If any of the given event names
are not found, throws a NACL::Exceptions::SPEventCheckFailure exception with
appropriate error message.

=item C<< check_for_absence => $array_ref >>

( Optional, defaults to reference to a null array )
Reference to a list of events names to check. If any of the given event names
are found, throws a NACL::Exceptions::SPEventCheckFailure exception with
appropriate error message.

=item C<< check_for_any_presence => $array_ref >>

( Optional, defaults to reference to a null array )
Reference to a list of events names to check. If none of the given event names
are found, throws a NACL::Exceptions::SPEventCheckFailure exception with
appropriate error message.

=item C<< check => 0|1 >>

( Optional, defaults to 1 if not provided )
This param is checked to see if the log verification has to be triggered when
stop() is called. Unless it is 'check' => 0, this method calls check().

=back

=back

=cut

sub stop {
    $Log->enter() if $may_enter;
    my ($self, @opts) = @_;
    my %opts = validate_with(
        params => \@opts,
        spec   => {
            check => {type => SCALAR, default => 1},
            %{NACL::C::Component->_common_validate_spec_without_ci()},
        },
        allow_extra => 1
    );
    my $apiset_obj = $self->sp_apiset_object();
    my $check      = delete $opts{'check'};
    my $sys_date = $self->get_sysdate('sp_apiset_object' => $apiset_obj, %opts);
    $self->event_end_time($sys_date);

    $Log->exit() if $may_exit;
    return $self->check(%opts) if ($check);
} ## end sub stop

=head2 check

    $detector->check(check_for_all_presence=>['wafl.iron.mount.times']);
(Instance method)

or

   NACL::MTask::LogDetector::SPEventLogDetector->check(
    'sp_apiset_object'       => $apiset_obj,
    'check_for_all_presence' => ['message.must.occur'],
    'event_begin_time'       => $begin_time,
    'event_end_time '        => $end_time
    );
(Class method)

Verifies the events occurred during the time frame between the calls to start() and
stop() methods been called based on the options provided. In the case of static
call the time frames are passed to the event_begin_time & event_end_time
parameters. NACL::Exceptions::SPEventCheckFailure with the appropriate message
will be thrown if the verification fails.

=over

=item Options

=over

=item C<< event_begin_time => $begin_time >>

( mandatory for class method call )
Begin time to check for event logs. Date should be either the in the
format "DAY MONTH DATE HH:MM:SS time-line YEAR" / "DAY MONTH DATE HH:MM:SS YEAR".
eg : Thu Jan  1 00:00:35 GMT 1970 / Thu Jan  1 00:00:35 1970.

=item C<< event_end_time => $end_time >>

( mandatory for class method call )
Begin time to check for event logs. Date should be either the in the
format "DAY MONTH DATE HH:MM:SS time-line YEAR"/"DAY MONTH DATE HH:MM:SS YEAR".
eg : Thu Jan  1 00:00:35 GMT 1970 / Thu Jan  1 00:00:35 1970.

=item C<< check_for_all_presence => $array_ref >>

( Optional, defaults to reference to a null array )
Reference to a list of events names/Regexp to check. If any of the given event
names are not found, throws a NACL::Exceptions::SPEventCheckFailure exception
with appropriate error message.

=item C<< events => $array_ref >>

( Mandatory for class method, Not Applicable for instance method )
Reference to a list of sp events log objects from which the events will be
validated. These events will be the parsed_output which we get from the
command "events <all|newest|oldest>".
eg:
[
  {
    'record' => '542',
    'time' => '11:58:06',
    'date' => 'Fri Apr 12',
    'message' => '[SP CLI.notice]: admin "log in from 10.72.8.245"',
    'year' => '2013'
  },
  {
    'record' => '543',
    'time' => '11:58:43',
    'date' => 'Fri Apr 12',
    'message' => '[SP CLI.notice]: admin "log in from 10.72.8.245"',
    'year' => '2013'
  },
];

=item C<< check_for_absence => $array_ref >>

( Optional, defaults to reference to a null array )
Reference to a list of events names/Regexp to check. If any of the given event
names are found, throws a NACL::Exceptions::SPEventCheckFailure exception with
appropriate error message.

=item C<< check_for_any_presence => $array_ref >>

( Optional, defaults to reference to a null array )
Reference to a list of events names/Regexp to check. If none of the given event
names are found, throws a NACL::Exceptions::SPEventCheckFailure exception with
appropriate error message.


=back

=back

=cut

sub check {
    $Log->enter() if $may_enter;
    my ($self, @opts) = @_;
    my %opts = validate_with(
        params => \@opts,
        spec   => {
            %{$self->_check_validate_spec()},
            %{NACL::C::Component->_common_validate_spec_without_ci()},
        },
        allow_extra => 1,
    );

    my ($all_sp_events,          $total_records,
        $apiset_obj,             $check_for_absence,
        $check_for_any_presence, $check_for_all_presence,
        $cnt,                    $apiset,
        $begin_time,             $end_time,
        $scalar_begin_date,      $matched_hash,
        $scalar_end_date,
    );

    my $timeout       = delete $opts{timeout}|| DEFAULT_TIMEOUT;
    $apiset_obj = delete $opts{sp_apiset_object} || $self->sp_apiset_object();
    $check_for_absence = $opts{'check_for_absence'}
        || $self->check_for_absence();
    $check_for_any_presence = $opts{'check_for_any_presence'}
        || $self->check_for_any_presence();
    $check_for_all_presence = $opts{'check_for_all_presence'}
        || $self->check_for_all_presence();

    $cnt        = 0;
    $begin_time = $opts{'event_begin_time'} || $self->event_begin_time();
    $end_time   = $opts{'event_end_time'} || $self->event_end_time();

    $scalar_begin_date
        = $self->_convert_time_for_compare(time => $begin_time);
    $scalar_end_date = $self->_convert_time_for_compare(time => $end_time);

    if ($scalar_begin_date > $scalar_end_date) {
        $Log->exit() if $may_exit;
        NACL::Exceptions::InvalidParam->throw(
            "Time seems to be unsynchronized since start date-time is greater than end date-time"
        );
    }

    if (!(     $check_for_any_presence
            || $check_for_all_presence
            || $check_for_absence
        )
        ) {
        $Log->exit() if $may_exit;
        NACL::Exceptions::InvalidParam->throw(
            "Any one of the option check check_for_any_presence/check_for_all_presence/check_for_absence should have been passed"
        );
    }

    if ($opts{events}) {
        $all_sp_events = $opts{events};
    } else {
        {
             local $NATE::Log::CommEvent::CommLogLevel = "debug";
             $Log->comment(
             " Inhibiting the logging of the command \"sp events all\" execution".
             " into the log file as this might increase the file size if there".
             " too many events produced. This can be removed by providing the".
             " LOGSPEC as \"debug\"");
             $all_sp_events = $apiset_obj->sp_events( all => 1, 'connectrec-timeout' => $timeout )->get_parsed_output();
        }
    }
    if (defined NATE::ParamSet::param_global()->get('NACL_BURT_916668')) {
        $Log->comment("begin time is: " . $begin_time);
        $Log->comment("end time is: " . $end_time);
        $Log->comment("all sp events are " . Data::Dumper::Dumper(\$all_sp_events));
    }
    if(!defined($all_sp_events)) {
        $Log->exit() if $may_exit;
        NACL::Exceptions::SPEventCheckFailure->throw(
            "The events are either not passed or its not generated from".
            " SP login");
    }

    if (@$check_for_all_presence) {
        my %match_all_presence;
        foreach my $msg (@$check_for_all_presence) {
            foreach my $records (@$all_sp_events) {
                my $record_date
                    = $records->{date} . " "
                    . $records->{time} . " "
                    . $records->{year};
                my $scalar_record_date
                    = $self->_convert_time_for_compare(time => $record_date);
                if (   ($records->{message} =~ /$msg/)
                    && ($scalar_record_date >= $scalar_begin_date)
                    && ($scalar_record_date <= $scalar_end_date)) {
                    $matched_hash->{records}->[$cnt] = $records;
                    $cnt++;
                    $match_all_presence{$msg} = $cnt;
                }
            }
        }
        my @events_not_found;
        foreach (@$check_for_all_presence) {
            if (defined($match_all_presence{$_})) {
                next;
            } else {
                push @events_not_found, $_;
            }
        }
        if (@events_not_found) {
            $Log->exit() if $may_exit;
            NACL::Exceptions::SPEventCheckFailure->throw(
                "Events not found when they were passed in check_for_all_presence parameter",
                'unmatched_events' => \@events_not_found,
            );
        }
        $matched_hash->{count} = $cnt;
    }
    if (@$check_for_any_presence) {
        foreach my $msg (@$check_for_any_presence) {
            foreach my $records (@$all_sp_events) {
                my $record_date
                    = $records->{date} . " "
                    . $records->{time} . " "
                    . $records->{year};
                my $scalar_record_date
                    = $self->_convert_time_for_compare(time => $record_date);
                if (   ($records->{message} =~ /$msg/)
                    && ($scalar_record_date >= $scalar_begin_date)
                    && ($scalar_record_date <= $scalar_end_date)) {
                    $matched_hash->{records}->[$cnt] = $records;
                    $cnt++;
                }
            }
        }
        if ($cnt > 0) {
            $matched_hash->{count} = $cnt;
        } else {
            $Log->exit() if $may_exit;
            NACL::Exceptions::SPEventCheckFailure->throw(
                "No events found when check_for_any_presence parameter was passed",
                'unmatched_events' => $check_for_any_presence,
            );
        }
    }
    if (@$check_for_absence) {
        my %match_absence;
        foreach my $msg (@$check_for_absence) {
            my $unexpected_records;
            foreach my $records (@$all_sp_events) {
                my $record_date
                    = $records->{date} . " "
                    . $records->{time} . " "
                    . $records->{year};
                my $scalar_record_date
                    = $self->_convert_time_for_compare(time => $record_date);
                if (   ($scalar_record_date >= $scalar_begin_date)
                    && ($scalar_record_date <= $scalar_end_date)) {
                    if ($records->{message} !~ /$msg/) {
                        next;
                    } elsif ($records->{message} =~ /$msg/) {
                        $match_absence{$msg} = $unexpected_records++;
                    }
                }
            }
        }
        if (keys %match_absence) {
            push my @events_matched, keys %match_absence;
            $Log->exit() if $may_exit;
            NACL::Exceptions::SPEventCheckFailure->throw(
                "Unexpected events are matched, which should not have occured",
                'matched_events' => \@events_matched,
            );
        } else {
            $Log->comment(
                "None of the events mentioned in check_for_absence option are matched"
            );
            $matched_hash = {};
            $matched_hash->{count} = 0;
        }
    }

    if (!ref $self) {
        my $obj = $self->new(
            matched_records   => $matched_hash,
            sp_apiset_object  => $apiset_obj,
            event_begin_time  => $begin_time,
            event_end_time    => $end_time,
            events            => $all_sp_events,
            count             => $matched_hash->{count},
        );
        $Log->exit() if $may_exit;
        return $obj;
    } else {
        $self->{matched_records} = $matched_hash;
        $self->{events}          = $all_sp_events;
        $self->{count}           = $matched_hash->{count};
        $Log->exit() if $may_exit;
        return $self;
    }
} ## end sub check

=head2 get_sysdate

    $SPEventLogDetector->get_sysdate();

    (or)

    NACL::MTask::LogDetector::SPEventLogDetector->get_sysdate(
        sp_apiset_object => $Node);

(Class or Instance method)
This method is used to fetch the system date at that point of time.

Also note that if the SP login is rebooted or if the SP connection is lost
during operation, the SP login session needs to be reconnected.
Using the below command,
    $APIset_Object->api_reconnect_session();

=over

=item Options

=over

=item C<< sp_apiset_object => $node >>

( Mandatory for class call )
The SP login APISet object of the filer on which we are checking for the sp events.
See L<NACL::APISet::Node::CLI::SP|lib-NACL-APISet-Node-CLI-SP-pm>.


=back

=back

=cut

sub get_sysdate {
    $Log->enter() if $may_enter;
    my ($self, @opts) = @_;
    my %opts = validate_with(
        params => \@opts,
        spec   => {'sp_apiset_object' => {type => OBJECT, optional => 1}},
        allow_extra => 1
    );
    my $apiset_obj = $opts{sp_apiset_object} || $self->sp_apiset_object();
    my $parsed_output = $apiset_obj->date()->get_parsed_output();
    my $date          = $parsed_output->[0]{date};
    $Log->exit() if $may_exit;
    return $date;
} ## end sub get_sysdate

# Returns validate spec for check() method based on if the check() is invoked as instance method or class method
sub _check_validate_spec {
    my $self = shift;
    if (ref($self)) {
        return {
            'check_for_all_presence' =>
                {type => ARRAYREF, optional => 1, default => []},
            'check_for_absence' =>
                {type => ARRAYREF, optional => 1, default => []},
            'check_for_any_presence' =>
                {type => ARRAYREF, optional => 1, default => []},
        };
    } else {
        return {
            'event_begin_time' => {type => SCALAR},
            'event_end_time'   => {type => SCALAR},
            'events'           => {type => ARRAYREF, optional => 1},
            'check_for_all_presence' =>
                {type => ARRAYREF, optional => 1, default => []},
            'check_for_absence' =>
                {type => ARRAYREF, optional => 1, default => []},
            'check_for_any_presence' =>
                {type => ARRAYREF, optional => 1, default => []},
            'sp_apiset_object' => {type => OBJECT},
        };
    }
} ## end sub _check_validate_spec

sub _convert_time_for_compare {
    $Log->enter() if $may_enter;
    my ($self, @opts) = @_;
    my %opts = validate_with(
        params => \@opts,
        spec   => {time => {type => HASHREF | SCALAR, optional => 1},},
    );

    my ($month, $dd, $time, $yr, $mm, $str_to_cmp);
    my $date = $opts{time};

    #Thu Jan  1 00:01:12 GMT 1970  (or) Thu Jan  1 00:01:12 1970
    if ($date =~ /\w+\s+(\w+)\s+(\d+)\s+(\d+:\d+:\d+)(\.\d+)?\s*\w*\s+(\d+)/) {
        $month = $1;
        $dd    = $2;
        $time  = $3;
        $time  = $3.$4 if ($4);
        $yr    = $5;
        $mm    = $self->enumerate_month(month => $month);
    }

    #format YYYY-MM-DD HH:MM:SS
    #1970-1-1 00:01:12
    my $time_to_return = str2time($yr . "-" . $mm . "-" . $dd . " " . $time);
    $Log->exit() if $may_exit;
    return $time_to_return;
} ## end sub _convert_time_for_compare

1;
