#
# Copyright (c) 2013 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#
#
## @summary MTask Lun Copy Iterator Class-
## @author asada@netapp.com, dl-nacl-dev@netapp.com
## @status shared
## @pod here
##############################################################################

package NACL::MTask::LunIterators::LunCopyIterator;
use Moose;

use Tharn;
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Data::Dumper;
use NACL::STask::Volume;
use NACL::STask::Lun;
use NACL::C::LunCopy;
use NATE::BaseException;
use NACL::APISet::Exceptions::ResponseException;
use NACL::GeneralUtils qw(random_name_generator);

# iterator functions and attributes are provided by the base class
extends qw (NACL::MTask::LunIterators::OperationalLunIterator);

# Constants
use constant WAIT_TIME => '10';
use constant RETRIES => '5';
use constant METHOD_TIMEOUT => '180';
use constant POLLING_INTERVAL => '20';

#
# destination volume is a required field for lun copy iterator
#
has destination_volume => (
    is => 'rw',
    isa => 'NACL::C::Volume',
    required => 1,
);

# Overide parent function to do Lun copy operation on a lun
override '_do_lun_operation' => sub {
    my ($self, @luns) = @_;
    $Log->enter() if $may_enter;
    my $volume = $self->destination_volume();

    # create array of Lun objects for the copied Luns
    my @dest_lun_objs = ();
    my $dest_lun_obj = undef;
    my @dest_lun_paths = ();
    my @source_lun_paths = ();
    foreach my $lun (@luns) {
        # extract the lun name out of lun path
        my @split_path = split ('/', $lun->{path} );
        my $lun_name = $split_path[scalar @split_path - 1];
        if (length $lun_name > 10) {
            $lun_name = substr $lun_name, 0, 10;
        }
        # Create a random name for copied lun
        my $new_lun_name = random_name_generator(
            prefix => "copied_".$lun_name,
            length => 20,
        );

        # /vol/volname/lunname for destination
        my $copy_destination_path = "/vol/".$volume->volume()."/$new_lun_name";
        push @dest_lun_paths, $copy_destination_path;
        push @source_lun_paths, $lun->{path};

        $Log->debug ("LUN: ". $lun->{path});

        # Create objects for the destination luns
        $dest_lun_obj = NACL::STask::Lun->new(
            command_interface => $lun->command_interface(),
            vserver => $volume->vserver(),
            path => $copy_destination_path,
        );

        push @dest_lun_objs, $dest_lun_obj;
    }

    # Note: all luns belong to the same vserver and are copied to the same
    # destination vserver volume
    my $command_interface = $luns[0]->command_interface();
    my $source_vserver = $luns[0]->vserver();

    # Call lun copy
    my $lun_copy_obj = undef;
    my $job = undef;
    my $retries = 0;
    while (!$lun_copy_obj  && $retries < RETRIES) {
        try {
            $lun_copy_obj = NACL::C::LunCopy->start(
                command_interface => $command_interface,
                'source-vserver' => $source_vserver,
                'source-paths' => \@source_lun_paths,
                'vserver' => $volume->vserver(),
                'destination-paths' => \@dest_lun_paths,
                'job_component'  => \$job,
            );

            if (defined $job){
                $job->wait_for_completion(
                    'method-timeout'   => METHOD_TIMEOUT,
                    'polling_interval' => POLLING_INTERVAL,
                );
            }
        }
        catch NACL::APISet::Exceptions::ResponseException with {
            $retries++;
            my $ex = shift;
            if ($ex->text() !~ /Maximum simultaneous file operations reached/i) {
                # Error: command failed: Cannot start file operation. Maximum simultaneous file operations reached.
                $Log->trace("Maximum number of operations running, not removing ".
                "the operation from the staging area");
                Tharn::sleep(WAIT_TIME);
            }  else {
                @dest_lun_objs = (); # reset
                $ex->throw();
            }
        };
    }

    $Log->exit() if $may_exit;
    if ($self->group_operations_flag){
        return @dest_lun_objs;
    } else {
        return $dest_lun_objs[0];
    }
};

1;
=head1 NAME

NACL::MTask::LunIterators::LunCopyIterator - Provide an iterator object for iterating and doing a
lun copy operation.

=head1 SYNOPSIS

This is a lun on copy iterator class and it allows to iterate over luns and do a copy operation on the luns
provided by input lun iterator.

extends NACL::MTask::LunIterators::LunCopyIterator;

The class overides base functions to check if the lun object input is able to be copied and also overides
the operation function in this case a lun copy operation is performed.

=head1 DESCRIPTION

This Class provides an iterator that for each next() function will do the lun copy operation before return
the NACL::STask::Lun object for the copied lun.

=head1 FUNCTIONS

=head2 new

Create a new iterator object instance and return it.

=over

=item C<< lun_iterator >> An instance of a lun iterator, it must support next
and has_next. The next method must return NACL::STask::Lun instances.

=back

=head2 has_next

Returns 1 if there is another "copy" lun available to get via a call to next or
0 if there are no more luns available.

=head2 next

Return the next "copied" lun as a NACL::STask::Lun instance.

=head2 next_n_operations

Does the lun "operation" on n luns and returns all the "operationed luns" as a Array Reference of
"NACL::STask::Lun" objects.
Note: This function cannot be used without group operations flag


=head1 SEE ALSO

MTask::LunIterators::OperationalLunIterator for documentation on other functions

=cut
