#
# Copyright (c) 2013 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#
#
## @summary MTask LunIterator Base class
## @author asada@netapp.com, dl-nacl-dev@netapp.com
## @status shared
## @pod here
##############################################################################
package NACL::MTask::LunIterators::OperationalLunIterator;

use Moose;
use Moose::Util::TypeConstraints;

use Tharn;
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use NACL::STask::Volume;
use NACL::STask::Lun;
use NACL::Exceptions::NoElementsFound;
use NACL::Exceptions::InvalidMethod;
use NACL::APISet::Exceptions::InvalidParamValueException;

use NACL::GeneralUtils qw(random_name_generator);

#
# lun_iterator is required object that is iterated and each element is used to
# do further iteration with the applied operation.
#
has lun_iterator => (
    is => 'rw',
    isa => 'Object',
    required => 1,
);

#
# The next lun to be returned
#
has _next_lun => (
    is => 'rw',
    isa => 'NACL::C::Lun',
    clearer => '_clear_next_lun',
    predicate => '_has_next_lun',
);

#
# Group operations flag - allow iterator to do next group operation on luns
#
has group_operations_flag => (
    is => 'ro',
    isa => 'Bool',
    required => 0,
    default => 0,
);

#
# Number of luns to perform an operation on in a go
#
has num_of_bulk_luns => (
    is => 'ro',
    isa => 'Int',
    required => 0,
    default => 400,
);

#
# Array to store group operationed luns will only be filled if group operations flag set
# and the iterator performed a group op.
#
has _operationed_luns => (
    is => 'rw',
    isa => 'ArrayRef[NACL::C::Lun]',
    clearer => '_clear_next_operationed_luns',
    predicate => '_has_next_operationed_luns',
    default => sub { return []; },
);

#
# Checks if the iterator has a next operationable lun
#
sub has_next {
    my $self = $_[0];
    $Log->enter() if $may_enter;
    $self->_refill_next_lun();

    if ($self->group_operations_flag) {
        if ($self->_has_next_operationed_luns()) {
            $Log->exit() if $may_exit;
            return 1;
        }
    } else {
        if ($self->_has_next_lun()) {
            $Log->exit() if $may_exit;
            return 1;
        }
    }

    $Log->exit() if $may_exit;
    return 0;
}

# It is a core refill function refills the iterator
sub _refill_next_lun {
    my $self = $_[0];
    $Log->enter() if $may_enter;

    # If we have a lun that is ready for next to return, stop now
    # same goes for group
    if (($self->group_operations_flag == 0) && ($self->_has_next_lun())) {
        $Log->exit() if $may_exit;
        return;
    }

    if (($self->group_operations_flag == 1) && ($self->_has_next_operationed_luns())) {
        $Log->exit() if $may_exit;
        return;
    }

    if ($self->group_operations_flag){
        # For group operations we will collect the luns for source upto
        # num_of_bulk_luns and pass that on to child class for operation
        my $count = 0;
        my @source_luns = ();
        while ($self->lun_iterator()->has_next() && $count < $self->num_of_bulk_luns) {
            my $lun = $self->lun_iterator()->next();
            # If operation can be performed on the lun push it to the list
            if ($self->_is_operationable($lun)) {
                push @source_luns, $lun;
            }
            
            $Log->comment ("Selected source file: ". $lun->path());
            $count++;
        }
        if (scalar @source_luns > 0) {
            # We have collected upto num_of_bulk_luns luns so lets call operation on them
            my @processed_luns = $self->_do_lun_operation(@source_luns);
            if (scalar @processed_luns == 0){
                $Log->exit() if $may_exit;
                NATE::BaseException->throw("Iterator Operation not performed");
            }

            $self->_operationed_luns(\@processed_luns);
        }
    } else {
        while ($self->lun_iterator()->has_next()) {
            my $lun = $self->lun_iterator()->next();
            my @source_luns = ();
            # If operation can be performed on the lun;
            if ($self->_is_operationable($lun)) {
                push @source_luns, $lun;
                $Log->comment ("Selected source file: ". $lun->path());
                my $processed_lun = $self->_do_lun_operation(@source_luns);
                if (!defined $processed_lun){
                    $Log->exit() if $may_exit;
                    NATE::BaseException->throw("Iterator Operation not performed");
                }

                $self->_next_lun($processed_lun);
                $Log->exit() if $may_exit;
                return;
            }
        }
    }

    $Log->exit() if $may_exit;
    return;
}

# Base method 'only filters for luns', need to override in child classes if required
sub _is_operationable {
    my $self = shift;
    my $lun = shift;

    $Log->enter() if $may_enter;

    # Get the lun ostype make sure it is defined
    my $ostype = "";
    try {
        $ostype = $lun->get_one_state_attribute('ostype');
    } catch NACL::Exceptions::NoElementsFound with {
        # Lun might have gone away btw the time
        # we made it here from the original iterator
        # if the lun isnt there its obviously un operationable :)
    };

    $Log->exit() if $may_exit;
    if (defined $ostype && $ostype ne "") {
        return 1;
    } else {
        return 0;
    }
}

# Base method, need to override in child classes
sub _do_lun_operation {
    my $self = shift;
    my $lun = shift;
    return $lun;
}

#
# does the iterator next. Note: Can't be used with group operations
#
sub next {
    my $self = $_[0];
    $Log->enter() if $may_enter;
    if ($self->group_operations_flag) {
        $Log->exit() if $may_exit;
        NATE::BaseException->throw ("next() cannot be called with group_operations_flag, use next_n_operations() instead.");
    }

    $self->_refill_next_lun();
    if (!$self->_has_next_lun()) {
        $Log->exit() if $may_exit;
        NACL::Exceptions::NoElementsFound->throw(
            "Called next and no elements were remaining"
        );
    }

    my $lun = $self->_next_lun();
    $self->_clear_next_lun();
    $Log->exit() if $may_exit;
    return $lun;
}

#
# does the next N operations, returns the N operationed luns
#
sub next_n_operations {
    my $self = $_[0];
    $Log->enter() if $may_enter;
    if (!$self->group_operations_flag) {
        $Log->exit() if $may_exit;
        NATE::BaseException->throw ("next_n_operations() cannot be called without group_operations_flag, use next() instead.");
    }

    $self->_refill_next_lun();
    if (!$self->_has_next_operationed_luns()) {
        $Log->exit() if $may_exit;
        NACL::Exceptions::NoElementsFound->throw(
            "Called next and no elements were remaining"
        );
    }

    my @op_luns = @{$self->_operationed_luns()};
    $self->_clear_next_operationed_luns();
    $Log->exit() if $may_exit;
    return @op_luns;
}

#
# rewind the iterator
#
sub rewind {
    my $self = $_[0];
    $Log->enter() if $may_enter;
    $self->lun_iterator()->rewind();
    $Log->exit() if $may_exit;
}

1;
=head1 NAME

NACL::MTask::LunIterators::OperationalLunIterator - Provide an iterator
interface base class to be used by various child iterator classes.

=head1 SYNOPSIS

This is a base iterator class and if used will just iterate over the iterator provided

extends NACL::MTask::LunIterators::OperationalLunIterator;

This class is not intented to be used directly buts its childern iterator will be.

=head1 DESCRIPTION

This is a base iterator class for various lun iterator api.

=head1 FUNCTIONS

=head2 new

Create a new iterator object instance and return it.

=over

=item C<< lun_iterator >> An instance of a lun iterator, it must support next
and has_next. The next method must return NACL::STask::Lun instances.

=back

=head2 has_next

Returns 1 if there is another "operationed" lun available to get via a call to next or
0 if there are no more luns available.

=head2 next

Return the next "operationed" lun as a NACL::STask::Lun instance.
Note: this cannot be used if doing group operations if you pass group_operations_flag then
use next_n_operations()

=head2 next_n_operations

Does the lun "operation" on n luns and returns all the "operationed luns" as a Array Reference of
"NACL::STask::Lun" objects.
Note: This function cannot be used without group operations flag

=head2 rewind()

Rewinds the iterator so the user can re-use this iterator again.
Caution: Keep in mind previous operations might still be happening. Make sure the
source is ready before iterating again.

=head1 SEE ALSO

=cut
