# $Id: //depot/prod/test/main/lib/NACL/MTask/MBUSTest/Receiver.pm#1 $
# Copyright (c) 2011 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary MBUSTest Task Module
## @author clausner@netapp.com
## @status shared
## @pod here

package NACL::MTask::MBUSTest::Receiver;

use strict;
use warnings;

use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Params::Validate qw(validate validate_with BOOLEAN SCALAR HASHREF OBJECT SCALARREF);
use NATE::Exceptions::Argument qw(:try);
use NATE::BaseException;
use NACL::APISet;

use NACL::ComponentUtils qw(Dumper);

use Time::ParseDate qw(parsedate);

=head1 NAME

NACL::MTask::MBUSTest::Receiver

=head1 DESCRIPTION

C<NACL::MTask::MBUSTest::Receiver> is derived class of
L<NACL::MTask::MBUSTest|lib-NACL-MTask-MBUSTest-pm>. It provides a 
well-defined method to return an mbustest object that can be
used to listen to MRT messages.

MRT stands for Message Router. This process controls where 
messages can go.

MBUSTest is the test client to transmit messages and receive
messages through the MRT process.

=head1 METHODS

=cut

use base qw (NACL::MTask::MBUSTest);
use Class::MethodMaker [
    scalar => 'node',
    scalar => 'class_id',
    scalar => 'duration',
    scalar => 'message_type',
    scalar => 'initial_delay',
];


=head2 new

    (Class method)
      $mbustest = NACL::MTask::MBUSTest::Receiver->new(
                node            => $node,           # required
                class_id        => 6                # optional, defaults to 2
                duration        => 10,              # optional, defaults to 10
                message_type    => 1,               # optional
                initial_delay   => 10,              # optional, defaults to 3
            );

    Creates a task that will have the ability to listen to messages.
    Constructor method that stores all values to be used.

=over

=item Options

=over

=item C<< node  => $node >>
    (Required) C<NACL::C::Node> that will be used to run the command from.

=item C<< class_id  => 6 >>

    (Optional) Integer value that allows the class of the listener to be
    set. The default value for the mbustest client is 2.

    This is meant to simulate clients that may want to be grouped together.
    For instance mgwd may have class of 5 and another new client may have the
    type of 5. Another client could send a broadcast message to type 5 and only
    those two clients will receive messages.

=item C<< duration    => 10 >>

    (Optional) Integer value that will be used to determine the task duration
    in seconds. Default is 10 seconds.

=item C<< message_type => 1 >>

    (Optional) As an integer. This option will register to listen to the specified
    message id. It will also, always listen to 5.

=item C<< initial_delay   => 10 >>

    (Optional) Integer that determines how long to wait(in seconds) before attempting
    to receive messages. Default is 3 seconds to allow the task to receive updates
    from the mrt task.

=back

=back

=cut

sub new {
    $Log->enter() if $may_enter;
    my $self  = shift;

    my %opts = validate_with(
        params => \@_,
        spec   => {
            node              => { isa  => 'NACL::C::Node' },
            duration          => { type => SCALAR, default => 10 },
            class_id          => { type => SCALAR, optional => 1 },
            message_type      => { type => SCALAR, optional => 1 },
            initial_delay     => { type => SCALAR, default => 3 },
        },
    );

    if($opts{node}->mode() ne 'CMode') {
       NACL::APISet::Exceptions::InvalidParamValueException->throw(
                text=>'Filer isn\'t CMode',
            );
    }

    $Log->debug(
        'opts to NACL::MTask::MBUSTest::Receiver->new ' . Dumper(%opts) );

    ## Bless the package instance
    $self = bless {}, $self;

    $self->initial_delay($opts{initial_delay});
    $self->node($opts{node});
    $self->duration($opts{duration});

    if(defined($opts{class_id})) {
        $self->class_id($opts{class_id});
    }

    if(defined($opts{message_type})) {
        $self->message_type($opts{message_type});
    }

    $Log->exit() if $may_exit;
    return $self;
} ## end sub new

=head2 run

    (Instance method)
      $mbustest->run(
                class_id        => 6                # optional
                duration        => 10,              # optional
                message_type    => 1,               # optional
                initial_delay   => 10,              # optional
          );


    Creates a short lived task that runs on the Freebsd that will listen  
    for messages. The parameters will determine what type of messages
    it will listen to.

=over

=item Options

=over

=item C<< class_id  => 2 >>

    (Optional) Integer value that allows the class of the listener to be
    set. This can override the constructors value.

    This is meant to simulate clients that may want to be grouped together.
    For instance mgwd may have class of 5 and another new client may have the
    type of 5. Another client could send a broadcast message to type 5 and only
    those two clients will receive messages.


=item C<< duration    => 10 >>

    (Optional) Integer value that will be used to determine the task duration
    in seconds.
    Value that can override the constructor values.

=item C<< message_type    => 1 >>

    (Optional) As an integer. This will register to listen to this message id.
    It will also, always listen to 5. This can override the value that was
    passed into the constructor.

=item C<< initial_delay   => 10 >>

    (Optional) Integer that determines how long to wait(in seconds) before attempting
    to receive messages. Value that can override the constructor values. This value
    allows the task to delay starting to listen to receive updates from the mrt
    task as to what all clients are available within the cluster network.

=back

=back

=cut

sub run {
    $Log->enter() if $may_enter;
    my $self  = shift;

    $Log->debug(
        'Package information of NACL::MTask::MBUSTest::Receiver ' . Dumper($self) );

    my %opts = validate_with(
        params => \@_,
        spec   => {
            duration          => { type => SCALAR, default => $self->duration() },
            class_id          => { type => SCALAR, default => $self->class_id()  },
            message_type      => { type => SCALAR, default => $self->message_type() },
            initial_delay     => { type => SCALAR, default => $self->initial_delay() },
        },
    );

    $Log->debug(
        'opts to NACL::MTask::MBUSTest::Receiver::run ' . Dumper(%opts) );

    my %params_to_use = (
            log   =>   'verbose',
        );

    $params_to_use{delay} = $opts{initial_delay};
    $params_to_use{time} = $opts{duration};

    if(defined($opts{message_type})) {
        $params_to_use{register} = $opts{message_type};
    }

    if(defined($opts{class_id})) {
        $params_to_use{class} = $opts{class_id};
    }

    my $apiset = $self->node()->apiset(
                category => 'Node',
                interface => 'CLI',
                set => 'Systemshell',
                shell_type => 'CMode' );    

    # Declare local variable
    my $output;
    my @output_array;
    my $output_ref;

    # Run the command
    $output = $apiset->mbustest( %params_to_use );
    $output_ref = $output->{parsed_output};

    @output_array = @$output_ref;
    my %actual_hash;
    foreach my $hash_ref ( @output_array ) {
        %actual_hash = %$hash_ref;
    }

    $self->output(\%actual_hash);

    $Log->exit() if $may_exit;
    return;
} ## end sub run


=head2 get_messages

    (Instance method)
      @messages_received = $mbus_object->get_messages(
            message_type    => 5, #optional
        );


    Method that will retrieve all the messages that were received
    by the base object or can be filtered down to a specific message
    type.

=over

=item Options

=over

=item C<< message_type    => 5 >>

    (Optional) Integer filter for the desired messages.

=back

=back

=cut

sub get_messages {
    my $self  = shift;
    my %opts = validate_with(
        params => \@_,
        spec   => {
            message_type          => { type => SCALAR, optional => 1 },
        }
        );
    
    my %message_hash = ();
    my $output_hash = $self->output();
    my %output_hash = %$output_hash;

    my $total_received = $output_hash{received_max};
    my $total_for_return = 0;

    if( defined($total_received) ) {
        for( my $i = 0; $i < $total_received; $i++) {
            my $current_message = "message$i";
            if(defined($opts{message_type}) ) {
                if( ('0x' . $opts{message_type}) eq $output_hash{$current_message}->{message_type}) {
                    $message_hash{$current_message} = $output_hash{$current_message};
                    $total_for_return++;
                }
            } else {
                $message_hash{$current_message} = $output_hash{$current_message};
                $total_for_return++;
            }
        }

        if($total_for_return != 0 ) {
            $message_hash{received_max} = $total_for_return;
        }
    }
    return (\%message_hash);

} ## end sub get_messages


=head2 cleanup_messages

    (Parent method)
      $mbus_object->cleanup_messages();

    Method that will cleanup the message received container.

=over

=over

=back

=back

=cut

sub cleanup_messages {

    my $self  = shift;
    my %empty_hash = ();
    $self->output(\%empty_hash);

} ## end sub cleanup_messages

1;
