# $Id$

# this is here due to our standard usage of @_ in Params::Validate methods.
##no critic (RequireArgUnpacking)
##no critic (ControlStructures::ProhibitUnlessBlocks)
##no critic (ValuesAndExpressions::ProhibitInterpolationOfLiterals)

#
# Copyright (c) 2011-2021 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary MCC Cluster Module
## @author dl-mcc-dpg@netapp.com
## @status shared
## @pod here

package NACL::MTask::MCC::Cluster;

use strict;
use warnings;

use base qw (NACL::MTask::MTask);

use Carp;
use NACL::MTask::MCC::DR_Coordinator ();

use NACL::C::Node();
use NACL::C::Cluster          ();
use NACL::C::Metrocluster     ();
use NACL::C::ClusterPeer      ();
use NACL::C::ClusterIdentity  ();
use NACL::C::NetworkInterface ();

use NACL::CS::EventLog;
use NACL::CS::JobHistory;
use NACL::CS::MetroclusterOperationHistory;
use NACL::CS::MetroclusterInterconnectMirror;

use NATE::BaseException qw(:try);

use NACL::MTask::MCC::Asup;
use NACL::MTask::MCC::Exceptions::Validation;
use NACL::MTask::MCC::Utils;

use NACL::Transit ();
use NACL::Transit::Exceptions::TransitException;

use Tharn qw( subtest );
use NATE::Log qw(log_global);
my $Log       = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();

use Params::Validate
  qw(validate validate_with BOOLEAN HASHREF OBJECT ARRAYREF SCALAR UNDEF);
use List::MoreUtils qw(any);
use Data::Dumper;
use Subtest;

=head1 NAME

NACL::MTask::MCC::Cluster

=head1 DESCRIPTION

This class is to be used for C-Mode Metrocluster testing.  It has only the
functionality required for MCC testing and should not be confused with
more general implementations of ONTAP cluster functionality such as
NACL::C::Cluster and NACL::STask::Cluster.  Think of it as a filter which in
some sense helps define the functionality needed to test clusters in the MCC
realm.

    # Start test
    use NACL::MTask::MCC::Metrocluster;
    ...
    my $cluster_A = $mcc->get_cluster(site => 'A');
    ...
    # Perform test here
    ...
    # End test

=cut

=head1 Public Methods

There is no public constructor for this class.  A NACL::MTask::MCC::Cluster
object is obtained via calls such as

    my $cluster_A = $mcc->get_cluster(site => 'A');

=cut

# Automatically generate the (non-public) constructor (includes a call to
# 'init') and also creates non-public members and non-public accesors for
# them.
use Class::MethodMaker [
    new => [ '-init', 'new' ],
    scalar =>
      [ { -type => 'NACL::MTask::MCC::Metrocluster' }, '_metrocluster', ],
    scalar =>
      [ { -type => 'NACL::MTask::MCC::DR_Coordinator' }, '_DR_coordinator', ],
    scalar => [
        qw/ _site
          _name
          _pre_switchover_halted_node_transit_objs
          _pre_switchover_intercluster_network_interfaces_downed
          _pre_switchover_disabled_isl_ports
          _prepare_for_switchover_execution_nodes_powered_off
          _remote_storage_reachable
          /,
    ],
];

# validate the parameters passed in and what else?
sub init {
    my $trace = NACL::MTask::MCC::Log::TRACELOG->start_timer();    #metrics

    my $self = shift;
    my %opts = validate_with(
        params => \@_,
        spec   => {
            metrocluster => { isa   => 'NACL::MTask::MCC::Metrocluster' },
            site         => { regex => qr/^(A|B)$/ },
        },
    );

    # set members per the passed constructor arguments.
    $self->_metrocluster( $opts{'metrocluster'} );
    $self->_site( $opts{'site'} );
    return;
}

=head2 get_Metrocluster_component()

    my $site_specific_mcc_component = $cluster_X->get_Metrocluster_component();

Returns the NACL::C::Metrocluster object associated with the execution of
metrocluster commands at $cluster_X's site.

=cut

sub get_Metrocluster_component {
    my $trace = NACL::MTask::MCC::Log::TRACELOG->start_timer();    #metrics

    my $self = shift;

    return $self->_metrocluster()
      ->get_Metrocluster_component( site => $self->_site() );
}

=head2 get_Cluster_component()

    my $cluster_component = $cluster_X->get_Cluster_component();

Returns the NACL::C::Cluster object associated with this cluster.

=cut

sub get_Cluster_component {
    my $trace = NACL::MTask::MCC::Log::TRACELOG->start_timer();    #metrics

    my $self = shift;

    my $cluster_component =
      $self->_metrocluster()
      ->get_Metrocluster_component( site => $self->_site() )
      ->command_interface();

    # this method (and a lot else) expects that the Metrocluster components'
    # command interfaces are NACL::C::Cluster objects.
    unless ( $cluster_component->isa('NACL::C::Cluster') ) {
        confess('Metrocluster CI found to not be a NACL::C::Cluster!!');
    }

    return $cluster_component;
}

=head2 get_site()

    my $cluster_site = $cluster->get_site();

Returns this cluster's site ('A'|'B').

=cut

sub get_site {
    my $trace = NACL::MTask::MCC::Log::TRACELOG->start_timer();    #metrics

    my $self = shift;

    return $self->_site();
}

=head2 get_name()

    my $cluster_name = $cluster_X->get_name();

Returns the name of this cluster.

=cut

# TBD: do we want to use metrocluster 'cluster' field for this?
sub get_name {
    my $trace = NACL::MTask::MCC::Log::TRACELOG->start_timer();    #metrics

    my $self = shift;

    my $metro_comp = $self->get_Metrocluster_component();

    if ( !defined $metro_comp ) {
        NACL::BaseException->throw(
            "Unable to get Metrocluster component in get_name()!");
    }

    my $ClusterIdentity_state = NACL::CS::ClusterIdentity->fetch(
        command_interface => $metro_comp->command_interface(),
        requested_fields  => ['name']
    );
    my $name = $ClusterIdentity_state->name();

    return $name;
}

sub get_site_peer {
    my $trace = NACL::MTask::MCC::Log::TRACELOG->start_timer();    #metrics

    my $self = shift;

    return $self->get_site() eq 'A' ? 'B' : 'A';
}

=head2 get_auso_config_state()

Gets the latest AUSO status on local MCC cluster

    my $auso_config_state = $cluster_X->get_auso_config_state();

Returns

auso-on-cluster-disaster : AUSO is configured for cluster wide disaster

auso-on-dr-group-disaster : AUSO is configure for cluster disaster within the DR group

auso-disabled : AUSO is disabled

undef : Unable to grab AUSO configurations

=cut

sub get_auso_config_state {
    my $trace = NACL::MTask::MCC::Log::TRACELOG->start_timer();    #metrics

    my $self = shift;
    my $local_cluster_name;

    #here try to get the cluster name. If the cluster is down this call will
    #fail. If the IGNORE_BAD_MCC_STATE is set then it is ok to ignore the
    #failure and return after the try/catch block. But if the param
    #is not set set then rethrow the exception.
    try {
        $local_cluster_name = $self->get_name();
    }
    catch NATE::BaseException with {
        if ( $ENV{'IGNORE_BAD_MCC_STATE'} ) {
            $Log->comment('IGNORE BAD STATE: get auso config state');

            #Perl will not let me return here hence do it after the catch.
        }
        else {
            my $exception = shift;
            $exception->throw();
        }
    };

    #We return from here with the auso value set to undef if we did not
    #get a value for local_cluster_name and if IGNORE_BAD_MCC_STATE = 1
    if ( !defined $local_cluster_name && $ENV{'IGNORE_BAD_MCC_STATE'} ) {
        $Log->comment( 'IGNORE BAD STATE: local cluster name not available.'
              . 'Testbed MAY be in a bad state! Returning from here.' );
        return undef;
    }

    # In LB the auso setting is verified using "metrocluster show"
    # the local node settings are not used for set/get of auso.
    my $mcc_cstate = NACL::CS::Metrocluster->fetch(
        command_interface => $self->get_Cluster_component(),
        apiset_must       => { interface => 'CLI' },
    );

    my $auso_state = $mcc_cstate->local_auso_failure_domain();
    $Log->comment("AUSO config state for $local_cluster_name is $auso_state");

    return $auso_state;

} ## end sub get_auso_config_state

=head2 modify_auso_config_state()

Modifies the AUSO config on a given MCC cluster

    $cluster_A->modify_auso_config_state(
            auso_failure_domain => 'auso-on-cluster-disaster',
    );

=over

=item Options

=item C<< auso_failure_domain => 'auso-on-cluster-disaster'|
'auso-on-dr-group-disaster'|'auso-disabled' (default = 0) >>

(Required) Specify the auso type to set on this cluster.

=back

=cut

sub modify_auso_config_state {
    my $self     = shift;
    my %raw_opts = @_;

    # validate opts
    my %opts = validate_with(
        params => \%raw_opts,
        spec   => {
            auso_failure_domain => {
                regex => qr/^(
                                auso-on-cluster-disaster  |
                                auso-on-dr-group-disaster |
                                auso-disabled             |
                               )$/x,
                optional => 0,
            },
        },
    );

    my $auso_failure_domain = $opts{'auso_failure_domain'};
    my $local_auso          = undef;
    my $local_cluster;
    my $local_cluster_name;

    #here try to get the cluster name. If the cluster is down this call will
    #fail. If the IGNORE_BAD_MCC_STATE is set then it is ok to ignore the
    #failure and return after the try/catch block. But if the param
    #is not set set then rethrow the exception.
    $Log->comment('Get the name of the cluster');
    try {
        $local_cluster_name = $self->get_name();
    }
    catch NATE::BaseException with {
        if ( $ENV{'IGNORE_BAD_MCC_STATE'} ) {
            $Log->comment('IGNORE BAD STATE: get auso config state');

            #Perl will not let me return here hence do it after the catch.
        }
        else {
            my $exception = shift;
            $exception->throw();
        }
    };

    #We return from here with the auso value set to undef if we did not
    #get a value for local_cluster_name and if IGNORE_BAD_MCC_STATE = 1
    if ( !defined $local_cluster_name && $ENV{'IGNORE_BAD_MCC_STATE'} ) {
        $Log->comment( 'IGNORE BAD STATE: local cluster name not available.'
              . 'Testbed MAY be in a bad state! Returning from here.' );
        return $local_auso;
    }

    $Log->comment(
        'Modifying AUSO config state for cluster: ' . $local_cluster_name );

    $local_cluster =
      $self->_metrocluster()
      ->get_Metrocluster_component( site => $self->_site() );
    if ( defined $local_cluster ) {

        # In LB the auso setting is verified using "metrocluster show"
        # and the local node settings using "metrocluster node show "
        # are not used for verifying auso.

        my $mcc_cstate = NACL::CS::Metrocluster->fetch(
            command_interface => $local_cluster->command_interface(),
            apiset_must       => { interface => 'CLI' },
        );
        if (
            lc $mcc_cstate->local_auso_failure_domain() ne
            $auso_failure_domain )
        {
            $Log->comment(
                "Modifying AUSO config to '$auso_failure_domain' on cluster "
                  . $mcc_cstate->local_cluster_name() );
            NACL::C::Metrocluster->modify(
                command_interface => $local_cluster->command_interface(),
                'auto-switchover-failure-domain' => $auso_failure_domain
            );

            # Check the settings after modifications
            $mcc_cstate = NACL::CS::Metrocluster->fetch(
                command_interface => $local_cluster->command_interface(),
                apiset_must       => { interface => 'CLI' },
            );

            if (
                lc $mcc_cstate->local_auso_failure_domain() ne
                $auso_failure_domain )
            {
                NATE::BaseException->throw(
"Failed to set '$auso_failure_domain' for AUSO config on cluster "
                      . $mcc_cstate->local_cluster_name() );
            }
        }
        else {
            $Log->comment('AUSO configuration is already set on this cluster');
        }

    }
    else {
        NATE::BaseException->throw(
'Unable to get AUSO config state for local cluster before modifying it'
        );
    }

} ## end sub modify_auso_config_state

# expects a single argument remote_cluster_name
sub _get_ClusterPeer_component {
    my $trace = NACL::MTask::MCC::Log::TRACELOG->start_timer();    #metrics

    my $self = shift;
    my %opts = @_;

    my $remote_cluster_name = $opts{'remote_cluster_name'};

    my $mcc_component = $self->get_Metrocluster_component();

    my $ClusterPeer_component = NACL::C::ClusterPeer->find(
        command_interface => $mcc_component->command_interface(),
        filter            => { 'rcluster' => $remote_cluster_name },
    );

    return $ClusterPeer_component;
}

=head2 get_peer()

    my $cluster_B = $cluster_A->get_peer();

Returns the other cluster (NACL::MTask::MCC::Cluster) in the MCC relationship.

=cut

sub get_peer {
    my $trace = NACL::MTask::MCC::Log::TRACELOG->start_timer();    #metrics

    my $self = shift;

    my $peer_site = ( $self->_site() =~ /A/ ) ? 'B' : 'A';

    return $self->_metrocluster()->get_cluster( site => "$peer_site" );
}

=head1 is_negotiated_switchover()

   $cluster_A->is_negotiated_switchover()

If we are (or have been) switched over, returns true if the last switchover
was negotiated (as opposed to an unplanned switchover).

=cut

sub is_negotiated_switchover {
    my $trace = NACL::MTask::MCC::Log::TRACELOG->start_timer();    #metrics
    my $self  = shift;

    # TODO Check for down nodes
    my ($mcc_node) = @{ $self->L_get_cluster_mcc_nodes() };
    my $node = $mcc_node->get_Node_component();

    my @events;

    try {
        my $cs = NACL::CS::MetroclusterOperation->fetch(
            command_interface => $node,
            'method-timeout'  => 30,
        );

        my $start_time = $cs->start_time();

        $Log->comment("Start time: $start_time");

        my $end_time_cs = NACL::CS::ClusterDate->fetch(
            command_interface => $node,
            requested_fields  => ['date']
        );

        my $end_time = $end_time_cs->date();

        $end_time =~ s/ -\d{4}$//;

        $Log->comment("End time: $end_time");

        my $time_str = "${start_time}..${end_time}";

        @events = NACL::CS::EventLog->fetch(
            command_interface => $node,
            filter            => {
                messagename => 'mcc.drsom.fsmStateEntry',
                time        => $time_str
            },
        );

    }
    catch NACL::Exceptions::NoElementsFound with {
        $Log->warn("Unable to find fsmStateEntries");

        return !$self->get_DR_Coordinator()->L_was_forced_on_disaster();

        return;
    };

    foreach my $event (@events) {
        $event->event() =~ m{mcc\.drsom\.fsmStateEntry:
                               \s
                             state="([^"]+)"
                             }x
          or next;

        my $state = $1;

        $state eq 'normal' and return;
        $state =~ m{^negotiated_so_} and return 1;
    }

    return;

}

sub _is_negotiated_switchover_fallback {
    return;
}

=head2 get_DR_Coordinator()

Each cluster in the MCC relationship has a DR Coordinator.

    my $dr_coordinator = $cluster_B->get_DR_Coordinator();

Returns the NACL::MTask::MCC::DR_Coordinator associated with this cluster.

=cut

sub get_DR_Coordinator {
    my $trace = NACL::MTask::MCC::Log::TRACELOG->start_timer();    #metrics

    my $self = shift;

    if ( !$self->_DR_coordinator_isset() ) {
        my $mcc_component = $self->get_Metrocluster_component();
        $self->_DR_coordinator(
            NACL::MTask::MCC::DR_Coordinator->new(
                metrocluster => $self->_metrocluster(),
                site         => $self->_site()
            )
        );
    }

    return $self->_DR_coordinator();
}

=head2 get_interconnect_state()

Does a metrocluster interconnect show for the cluster.

    my $interconnect_cs = $cluster_A->get_interconnect_state();

By default, requests the fields:

=over 4

=item * type

=item * node

=item * partner-type

=item * mirror-oper-status

=item * adapter

=item * status

=back

Returns the (arrayref) NACL::CS::MetroclusterInterconnectMirror
associated with this cluster.

=cut

sub get_interconnect_state {
    my $self = shift;

    my %opts = validate_with(
        params => \@_,
        spec   => {
            requested_fields => {
                type    => ARRAYREF,
                default => [
                    'type',               'node',
                    'partner-name',       'partner-type',
                    'mirror-oper-status', 'adapter',
                    'status'
                ],
            },
        },
    );

    my $requested_fields = $opts{requested_fields};
    my $ci        = $self->get_Metrocluster_component()->command_interface();
    my @ic_states = NACL::CS::MetroclusterInterconnectMirror->fetch(
        command_interface => $ci,
        requested_fields  => $requested_fields,
    );

    $Log->comment( '   Interconnect States: '
          . NACL::MTask::MCC::Utils->dump( \@ic_states ) );
    return \@ic_states;
}

=head2 prepare_for_switchover_execution()

    $cluster_A->prepare_for_switchover_execution(
                # default values shown here
            do_intercluster_network_interface_disable   => 1,
            do_isl_link_down                            => 0,
            do_peer_metrocluster_site_nodes_halt        => 1,
            do_wait_minutes_for_cluster_appears_down    => 3,);

This method performs various actions often/usually performed before this site's
cluster executes a DR switchover command.

=over

=item Options

=item C<< do_intercluster_network_interface_disable => 0|1 (default = 0) >>

(Optional) This option will disable the intercluster network interface on this
cluster which allows this cluster to see its Metrocluster peer cluster.

=item C<< do_ic_lif_disable_after_disaster => 0|1 (default = 1) >>

(Optional) determines whether to disable intercluster network interface
after disaster site is brought down or before

If set to 1 do_ic_lif_disable_after_disaster => 1 ic lifs will be brought
down after

=item C<< do_isl_link_down => 0|1 (default = 0) >>

(Optional) set to disable ISL ports of switch to bring down ISL link

If set to 1 do_isl_link_down => 1 all FC ports will be brought down and will
be brought back up by prepare_for_switchback_execution

=item C<< do_isl_link_down_after_disaster => 0|1 (default = 1) >>

(Optional) determines whether to disable ISLs after disaster site is
brought down or before

If set to 1 do_isl_link_down_after_disaster => 1 ISLs will be brought
down after

=item C<< do_isl_link_down_delay =>  (optional) >>

(Optional) value in seconds to wait between disabling each switch's isl
ports to simulate rolling disaster

=item C<< do_wait_minutes_for_cluster_appears_down => $minutes (default = 3) >>

(Optional) wait for specified number of minutes

=item C<< do_ignore_quorum_warnings => 0|1 (default = 1) >>

Determines whether to pass the 'do_ignore_quorum_warnings' argument to
halt and reboot commands. Defaults to 1.

=item C<< do_peer_metrocluster_site_nodes_halt => 0|1 (optional) >>
DEPRECATED

DO NOT USE THIS OPTION.  If you are using it, please convert the functionality
to the newer do_peer_metrocluster_site_nodes_disaster option instead!

This option is equivalent to the 'halt_inhibit_takeover' option for
do_peer_metrocluster_site_nodes_disaster.  An explicit '0' setting for
this option is equivalent to the 'none' option for
do_peer_metrocluster_site_nodes_disaster.

(Optional) This option will attempt to halt all the nodes in this cluster's
Metrocluster peer cluster.  (I.e., if this called on cluster A, it will try
to halt all nodes in cluster B.)

=item C<< do_peer_metrocluster_site_nodes_disaster => <disaster_type> (default = halt_inhibit_takeover) >>

This is the primary way to specify what disaster is being performed now that
do_peer_metrocluster_site_nodes_halt is deprecated.

The operation is performed on both disaster site nodes.

disaster_type is one of:

=over

=item  custom

Allows a custom combination of different disaster options.
Set these options in desired combinations (not all work!):
disaster_type, panic_core_dump, disable_core_file, inhibit_takeover,
skip_lif_migration


=item  halt

plain halt with no options

=item  halt_inhibit_takeover

halt with inhibit_takeover true

=item  none

no disaster performed, usually for NSO (negotiated switchover)

=item  panic

halt with dump true

=item  panic_nocore

halt with dump true but core dumps disabled

=item  panic_skip_lif_migration

halt with dump true and skip_lif_migration true

=item  power

dirty power off

Disaster-side nodes will be powered down and
will be powered up again during 'prepare_for_switchback_execution'.

=item  reboot

plain reboot

=item  reboot_inhibit_takeover

reboot with inhibit_takeover true

=back

=item C<< disaster_type => (halt|none|power|reboot) (optional) >>

Only used when do_peer_metrocluster_site_nodes_disaster == 'custom'
This option is REQUIRED in that case however.
If 'halt' or 'reboot' that option is used for the 'system node' command.
If 'power' then a poweroff is performed.
If 'none' then no disaster operation is performed.

=item C<< panic_core_dump => 0|1 (optional) >>

Only used when do_peer_metrocluster_site_nodes_disaster == 'custom'

If '1' then the '-dump true' option is enabled for halt|reboot.

=item C<< disable_core_file => 0|1 (optional) >>

Only used when do_peer_metrocluster_site_nodes_disaster == 'custom'

If '1' then core files are disabled.  (Presumably used in conjuction with
disable_core_file.)

=item C<< inhibit_takeover => 0|1 (optional) >>

Only used when do_peer_metrocluster_site_nodes_disaster == 'custom'

If '1' then the '-inhibit_takeover true' option is enabled for halt|reboot.

=item C<< skip_lif_migration => 0|1 (optional) >>

Only used when do_peer_metrocluster_site_nodes_disaster == 'custom'

If '1' then the '-skip_lif_migration true' option is enabled for halt|reboot.

=item C<< transit_timeout => timeout in seconds (default = 2000) >>

Time to pass NACL::Transit change_state call that halts/reboots disaster
nodes.  The passed value will override any default set internally.
(Some disaster types have different internal default timeouts.)

=item C<< release_disk_reservations_if_takenover => 0|1 (default = 0) >>

Control whether NACL::Transit will attempt to release disk reservations and
halt the nodes if they are in a taken over state. Only applicable if
'do_peer_metrocluster_site_nodes_disaster' is set to the halt or reboot
variants.

=back

=cut

sub prepare_for_switchover_execution {
    my $trace = NACL::MTask::MCC::Log::TRACELOG->start_timer();    #metrics

    my $self     = shift;
    my %raw_opts = @_;

    # validate opts
    my %opts = validate_with(
        params => \%raw_opts,
        spec   => {
            do_intercluster_network_interface_disable => {
                regex   => qr/^[01]$/,
                default => 0,
            },
            do_ic_lif_disable_after_disaster => {
                regex   => qr/^[01]$/,
                default => 1,
            },
            do_isl_link_down_delay => {
                regex    => qr/^\d+[.]?\d*$/,
                optional => 1,
            },
            do_isl_link_down => {
                regex   => qr/^[01]$/,
                default => 0,
            },
            do_isl_link_down_after_disaster => {
                regex   => qr/^[01]$/,
                default => 1,
            },
            do_storage_link_down => {
                regex   => qr/^[01]$/,
                default => 0,
            },
            do_peer_metrocluster_site_nodes_halt => {
                regex    => qr/^[01]$/,
                optional => 1,
            },
            do_wait_minutes_for_cluster_appears_down => {
                regex    => qr/^\d+[.]?\d*$/,
                optional => 1,
            },
            do_ignore_quorum_warnings => {
                regex   => qr/^[01]$/,
                default => 1,
            },
            do_peer_metrocluster_site_nodes_disaster => {
                regex => qr/^(
                                custom                   |
                                halt                     |
                                halt_inhibit_takeover    |
                                none                     |
                                panic                    |
                                panic_nocore             |
                                panic_skip_lif_migration |
                                power                    |
                                reboot                   |
                                reboot_inhibit_takeover
                               )$/x,

                #to support the old halt option we can't default here
                #default => 'halt_inhibit_takeover',
                optional => 1,
            },
            transit_timeout => {
                regex    => qr/^\d+$/,
                optional => 1,
            },

            # these options are only used with 'custom' disaster
            disaster_type => {
                regex => qr/^(
                                halt                    |
                                none                    |
                                power                   |
                                reboot
                               )$/x,
                optional => 1,
            },
            panic_core_dump => {
                regex    => qr/^[01]$/,
                optional => 1,
            },
            disable_core_file => {
                regex    => qr/^[01]$/,
                optional => 1,
            },
            inhibit_takeover => {
                regex    => qr/^[01]$/,
                optional => 1,
                default  => 1,
            },
            skip_lif_migration => {
                regex    => qr/^[01]$/,
                optional => 1,
                default  => 1,
            },
            release_disk_reservations_if_takenover => {
                regex   => qr/^[01]$/,
                default => 0,
            },
            suspend_bdfu => {
                regex   => qr/^[01]$/,
                default => 1,
            },
            skip_asup_generation => {
                regex   => qr/^[01]$/,
                default => 0,
            },
        },
    );

    my $disaster_type        = $opts{'disaster_type'};
    my $panic_core_dump      = $opts{'panic_core_dump'};
    my $inhibit_takeover     = $opts{'inhibit_takeover'};
    my $skip_lif_migration   = $opts{'skip_lif_migration'};
    my $skip_asup_generation = $opts{'skip_asup_generation'};
    my $disable_core_file    = $opts{'disable_core_file'};
    my $transit_timeout      = $opts{'transit_timeout'};
    my $release_disk_reservations_if_takenover =
      $opts{'release_disk_reservations_if_takenover'};

    my %disaster_map = (
        'custom' => {
            'disaster_type' => $disaster_type,
            defined $inhibit_takeover
            ? ( 'inhibit_takeover' => $inhibit_takeover )
            : (),
            defined $panic_core_dump ? ( 'panic_core_dump' => $panic_core_dump )
            : (),
            defined $skip_lif_migration
            ? ( 'skip_lif_migration' => $skip_lif_migration )
            : (),
            defined $disable_core_file
            ? ( 'disable_core_file' => $disable_core_file )
            : (),
            defined $transit_timeout ? ( 'transit_timeout' => $transit_timeout )
            : (),
        },

        #wafl 4
        'halt' => { 'disaster_type' => 'halt', },

        #walf 1
        #'skip_lif_migration' param added as a fix for BURT #1005300. STO
        #with 'inhibit-takeover=true' doesn't work without it.
        'halt_inhibit_takeover' => {
            'disaster_type'      => 'halt',
            'inhibit_takeover'   => 1,
            'skip_lif_migration' => 1,
        },

        #wafl 6
        'none'  => { 'disaster_type' => 'none', },
        'panic' => {
            'disaster_type'   => 'halt',
            'panic_core_dump' => 1,
        },

        #wafl 2
        'panic_skip_lif_migration' => {
            'disaster_type'      => 'halt',
            'panic_core_dump'    => 1,
            'skip_lif_migration' => 1,
        },
        'panic_nocore' => {
            'disaster_type'     => 'halt',
            'panic_core_dump'   => 1,
            'disable_core_file' => 1,
            'transit_timeout'   => 600,
        },

        #wafl 0
        'power' => { 'disaster_type' => 'power', },

        #wafl 3
        'reboot_inhibit_takeover' => {
            'disaster_type'    => 'reboot',
            'inhibit_takeover' => 1,
        },

        #wafl 5
        'reboot' => { 'disaster_type' => 'reboot', },

    );

    my $disaster = $opts{'do_peer_metrocluster_site_nodes_disaster'};

    if ( defined $opts{do_peer_metrocluster_site_nodes_halt} ) {

        #try to provide support for this deprecated option ...
        #map these values to the new format, but if the new
        #format is also defined, used that instead and ignore the
        #older ones
        my $halt = $opts{do_peer_metrocluster_site_nodes_halt};
        my $new_disaster;
        if ($halt) {

            #1 == halt inhibit takeover
            $new_disaster = 'halt_inhibit_takeover';
        }
        else {
            #don't halt means do nothing
            $new_disaster = 'none';
        }
        if ( !defined $disaster ) {

            #disaster isn't defined so go ahead and do the mapping
            $disaster = $new_disaster;
        }
        else {
            #they've passed both new and old halt values so the new
            #ones override the older one...
            $Log->warn( 'passed do_peer_metrocluster_site_nodes_disaster '
                  . "value '$disaster' as well as older "
                  . 'do_peer_metrocluster_site_nodes_halt option.  Using '
                  . "$disaster but the call should be changed to pass only one"
            );
        }

    }
    else {
        #set the default value here if no value is passed in
        $disaster = 'power' unless defined $disaster;
    }

    $disaster_type = $opts{'disaster_type'};

    my @interfaces_to_disable;
    if ( $opts{do_intercluster_network_interface_disable} ) {

        #get interfaces here as after disaster cluster peer query will fail
        $Log->comment('getting intercluster interfaces to disaster site...');
        @interfaces_to_disable =
          $self->_find_DR_cluster_to_THIS_cluster_intercluster_interfaces();
    }
    if ( $opts{do_intercluster_network_interface_disable}
        && !$opts{do_ic_lif_disable_after_disaster} )
    {
        $Log->comment('Disabling intercluster interfaces to disaster site...');

        foreach my $network_interface (@interfaces_to_disable) {
            $network_interface->modify( 'status-admin' => 'down' );
        }

        # keep track of these for re-enable after switchback
        $self->_pre_switchover_intercluster_network_interfaces_downed(
            \@interfaces_to_disable );
    }
    else {
        $self->_pre_switchover_intercluster_network_interfaces_downed(undef);
    }

    if ( $opts{do_isl_link_down} && !$opts{do_isl_link_down_after_disaster} ) {
        $self->_do_isl_links_modify(
            modify => 'down',
            (
                defined $opts{do_isl_link_down_delay}
                ? ( delay => $opts{do_isl_link_down_delay} )
                : ()
            ),
        );
    }

    $Log->comment("Disaster type: $disaster") if $disaster;

    if ( $disaster_map{$disaster}->{disaster_type} =~ /halt|reboot/ ) {
        my $map = $disaster_map{$disaster};

        my $halt_cmd_base =
            'set -confirmations off;'
          . "system node $map->{disaster_type} "
          . ( $map->{inhibit_takeover}   ? '-inhibit-takeover true '   : '' )
          . ( $map->{skip_lif_migration} ? '-skip-lif-migration true ' : '' )
          . ( $map->{panic_core_dump}    ? '-dump true '               : '' );

        if ( $opts{'do_ignore_quorum_warnings'} ) {
            $halt_cmd_base .= '-ignore-quorum-warnings true ';
        }

        $halt_cmd_base .= '-reason prepare_for_switchover_execution -node ';

        my $type =
          $map->{panic_core_dump}
          ? 'Panicking'
          : ucfirst $map->{disaster_type} . 'ing';
        $Log->comment( $type
              . ( $map->{inhibit_takeover}   ? ' with inhibit takeover'   : '' )
              . ( $map->{skip_lif_migration} ? ' with skip lif migration' : '' )
              . ' the MCC associated '
              . 'nodes at disaster site...' );
        my @subtest_handles     = ();
        my @halted_node_objects = ();
        my @peer_cluster_mcc_nodes =
          @{ $self->get_peer()->L_get_cluster_mcc_nodes() };

        foreach my $peer_mcc_node (@peer_cluster_mcc_nodes) {

            # Set options to cause panic without core dumps
            # FYI: This sysctl option stays into effect only for 1 iteration and
            # will automatically default to "panic with core" after a reboot
            if ( $map->{disable_core_file} ) {
                $Log->comment('Configuring kernel to panic without core dumps');
                $peer_mcc_node->get_Node_component->get_systemshell_apiset
                  ->sysctl(
                    name              => 'debug.dont_dump_core',
                    value             => 1,
                    'privilege-level' => 'root'
                  );
            }
            my $node_name = $peer_mcc_node->get_name();
            my $halt_cmd  = $halt_cmd_base . $node_name;
            my $timeout   = $transit_timeout || $map->{transit_timeout} || 2000;
            my $trans_obj = NACL::Transit->new( name => $node_name );
            push @subtest_handles, subtest(
                sub {
                    my %add_args = ();
                    if ($release_disk_reservations_if_takenover) {

                        # release disk reservations if we are in a takeover
                        # state
                        $add_args{iftakenover} = "takeback";

                        # answer the "halt the node question" if we had to
                        # release disk reservations
                        $add_args{user_defined_states} = [
"Do you wish to halt this node rather than wait \\[y/n\\]\\? "
                              => "yes",
                            "Please answer yes or no" => "yes"
                        ];
                    }
                    $trans_obj->change_state(
                        to      => 'FIRMWARE',
                        haltcmd => $halt_cmd,
                        timeout => $timeout,
                        %add_args
                    );
                },
                '-bg',
                '-runid',
                $node_name . '_prepare_for_switchover_execution_down'
            );
            push( @halted_node_objects, $trans_obj );
        }
        Subtest::wait_finish( subtests => \@subtest_handles );

        # save this information for rebooting halted nodes
        $self->_pre_switchover_halted_node_transit_objs(
            \@halted_node_objects );
    }
    elsif ( $disaster_map{$disaster}->{disaster_type} =~ /power/ ) {

        my @powered_off_nodes =
          @{ $self->get_peer()->L_get_cluster_mcc_nodes() };

        my @sys_nodes = map { $_->get_Node_component() } @powered_off_nodes;

        # Skip ASUPs generation if user sets skip_asup_generation flag
        if ( !$skip_asup_generation ) {
            $Log->comment('Triggering ASUPs');
            my $asup = NACL::MTask::MCC::Asup->new( nodes => \@sys_nodes );
            $asup->invoke_all();
            $asup->validate_sent();
        }
        else {
            $Log->comment('Skipping ASUPs generation...');
        }

        my @subtests = ();
        foreach my $node (@powered_off_nodes) {
            $Log->comment( "powering off node " . $node->get_name() );
            my $subtest = $node->subtest_system_power( 'power' => 'off' );
            push( @subtests, $subtest );
        }
        $self->_prepare_for_switchover_execution_nodes_powered_off(
            \@powered_off_nodes );
        Subtest::wait_finish( subtests => \@subtests );
    }
    elsif ( $disaster_map{$disaster}->{disaster_type} =~ /none/ ) {
        $Log->comment("Leaving 'disaster' site nodes alone (likely NSO)");

        # save this information for rebooting halted nodes
        #$self->_pre_switchover_halted_node_transit_objs(
        #\@halted_node_objects);
    }
    else {
        $self->_pre_switchover_halted_node_transit_objs(undef);
    }

    if (   $opts{do_intercluster_network_interface_disable}
        && $opts{do_ic_lif_disable_after_disaster} )
    {
        $Log->comment('Disabling intercluster interfaces to disaster site...');

        foreach my $network_interface (@interfaces_to_disable) {
            $network_interface->modify( 'status-admin' => 'down' );
        }

        # keep track of these for re-enable after switchback
        $self->_pre_switchover_intercluster_network_interfaces_downed(
            \@interfaces_to_disable );
    }
    else {
        $self->_pre_switchover_intercluster_network_interfaces_downed(undef);
    }

    if ( $opts{do_isl_link_down} && $opts{do_isl_link_down_after_disaster} ) {
        if ( $self->has_switches() ) {
            $self->_do_isl_links_modify(
                modify => 'down',
                (
                    defined $opts{do_isl_link_down_delay}
                    ? ( delay => $opts{do_isl_link_down_delay} )
                    : ()
                ),
            );
        }
        elsif ( $self->is_glowworm() ) {
            $self->_do_remote_sas_adapters_down();
        }
        elsif ( $self->is_atto() ) {
            ${ $self->L_get_cluster_mcc_nodes() }[0]
              ->_do_remote_atto_adapters_down();
        }
    }

    if ( $opts{do_wait_minutes_for_cluster_appears_down} ) {
        $Log->warn( 'Sleeping for '
              . $opts{do_wait_minutes_for_cluster_appears_down}
              . ' minutes...' );
        sleep( 60 * $opts{do_wait_minutes_for_cluster_appears_down} );
    }

    if ( $opts{suspend_bdfu} ) {
        $self->L_set_bdfu('off');
    }
}

sub _do_remote_sas_adapters_down {
    my $trace = NACL::MTask::MCC::Log::TRACELOG->start_timer();    #metrics

    my $self = shift;

    $Log->comment("Taking remote SAS adapters down");

    return $self->_do_remote_sas_adapters_modify(
        command => 'adapter_offline' );

}

sub _do_remote_sas_adapters_up {
    my $trace = NACL::MTask::MCC::Log::TRACELOG->start_timer();    #metrics

    my $self = shift;

    return $self->_do_remote_sas_adapters_modify( command => 'adapter_online' );

}

sub _do_remote_sas_adapters_modify {
    my $trace = NACL::MTask::MCC::Log::TRACELOG->start_timer();    #metrics

    my $self = shift;

    # validate opts
    my %opts = validate_with(
        params => \@_,
        spec   => { command => { type => SCALAR, }, },
    );

    my ($node) = @{ $self->L_get_cluster_mcc_nodes() };

    my $nnode = $node->L_get_network_node_functionality();

    my ( $local_adapters, $remote_adapters ) =
      $nnode->get_sas_host_adapters_local_remote();

    $Log->comment( 'Local adapters: '
          . Dumper($local_adapters)
          . 'Remote adapters: '
          . Dumper($remote_adapters) );

    my $apiset =
      $node->get_Node_component->get_7m_or_nodescope_apiset(
        interface => 'CLI' );

    foreach my $adapter ( @{$remote_adapters} ) {
        $apiset->execute_command(
            command => "sasadmin adapter_state $adapter" );
        $apiset->execute_command(
            command => "sasadmin $opts{'command'} $adapter" );
        $apiset->execute_command(
            command => "sasadmin adapter_state $adapter" );
    }

    return 1;

}

sub has_switches {
    my $self = shift;

    my ($node) = @{ $self->L_get_cluster_mcc_nodes() };

    return $node->L_has_switches();

}

sub has_bridges {
    my $self = shift;

    my ($node) = @{ $self->L_get_cluster_mcc_nodes() };

    return $node->L_has_bridges();
}

#helper method to disable isls
sub _do_isl_links_modify {
    my $trace = NACL::MTask::MCC::Log::TRACELOG->start_timer();    #metrics

    my $self = shift;

    # validate opts
    my %opts = validate_with(
        params => \@_,
        spec   => {
            delay  => { type  => SCALAR, default => 0, },
            modify => { regex => qr/^(up|down)$/, },
        },
    );

    my $down = $opts{modify} eq 'down';

    my @isl_ports = ();
    my $dr_groups = $self->_metrocluster()->get_DR_groups();

    my %isl_ports;

    if ($down) {
        foreach my $dr_group ( values( %{$dr_groups} ) ) {
            foreach my $fabric (
                @{ $dr_group->get_network_topology()->get_intersite_fabrics() }
              )
            {
                my $switch =
                  $fabric->get_switch( site => $self->get_peer()->_site() );
                my @ports = @{ $switch->get_isl_ports() };
                push( @isl_ports, @ports );
                push @{ $isl_ports{ $switch->get_switch_name() } }, @ports;
            }
        }
        $self->_pre_switchover_disabled_isl_ports( \@isl_ports );
    }
    else {
        #up
        my $isl_ports = $self->_pre_switchover_disabled_isl_ports();
        if ( !$isl_ports ) {
            $Log->comment(
                'cannot bring ISLs up, ports not brought down this run');
            return;
        }
        @isl_ports = @{$isl_ports};

        #make hash
        foreach my $port (@isl_ports) {
            push @{ $isl_ports{ $port->get_switch()->get_switch_name() } },
              $port;
        }
    }

    my @subtest_handles = ();

    #do each switch in parallel, but due to connection limits on
    #switches do the switch ports serially
    my $notfirst;
    foreach my $switch ( keys %isl_ports ) {
        if ( $notfirst++ && $opts{delay} ) {
            $Log->comment( "rolling disaster, pause $opts{delay} sec "
                  . 'until bringing down next switch ISL' );
            Tharn::snooze( $opts{delay} );
        }
        push @subtest_handles, subtest(
            sub {
                foreach my $isl_port ( @{ $isl_ports{$switch} } ) {
                    if ($down) {
                        $isl_port->disable();
                    }
                    else {
                        $isl_port->enable();
                    }
                }
            },
            '-bg',
            '-runid',
            $switch . "_switch_isl_$opts{modify}"
        );

    }
    Subtest::wait_finish( subtests => \@subtest_handles );

}

#private helper method to create transit objects for this cluster's peer node
sub _create_peer_transit_objects {
    my $trace = NACL::MTask::MCC::Log::TRACELOG->start_timer();    #metrics

    my $self = shift;

    my @peer_cluster_mcc_nodes =
      @{ $self->get_peer()->L_get_cluster_mcc_nodes() };
    my @transit_objects;

    foreach my $peer_mcc_node (@peer_cluster_mcc_nodes) {
        my $node_name = $peer_mcc_node->get_name();
        my $trans_obj = NACL::Transit->new( name => $node_name );
        push @transit_objects, $trans_obj;
    }

    return \@transit_objects;

}

#private helper method to determine if this cluster's peer nodes are powered
#off and if so return node objects for nodes that are powered off
sub _get_powered_off_peer_nodes {
    my $trace = NACL::MTask::MCC::Log::TRACELOG->start_timer();    #metrics

    my $self = shift;

    my @peer_cluster_mcc_nodes =
      @{ $self->get_peer()->L_get_cluster_mcc_nodes() };
    my @powered_off_nodes;

    foreach my $peer_mcc_node (@peer_cluster_mcc_nodes) {
        if ( $peer_mcc_node->system_power_status() eq 'off' ) {
            push @powered_off_nodes, $peer_mcc_node;
        }
    }

    return \@powered_off_nodes;

}

=head2 prepare_for_heal_aggrs_execution()

    #usually just call without params
    $cluster_A->prepare_for_heal_aggrs_execution();

    $cluster_A->prepare_for_heal_aggrs_execution(
            # default values shown here
            do_isl_link_up => 1,
    );


=over

=item Options

=item C<< do_isl_link_up => 0|1 (default = 1) >>

(Optional) This option will attempt to bring up all isl ports that
were taken down during prepare_for_switchover_execution.  Since we
would normally want to do this it defaults to 1.

=item C<< suspend_bdfu => 0|1 (default = 1) >>

If true, suspends background disk firmware updates on all cluster
nodes.

=back

=cut

sub prepare_for_heal_aggrs_execution {
    my $trace = NACL::MTask::MCC::Log::TRACELOG->start_timer();    #metrics

    my $self     = shift;
    my %raw_opts = @_;

    # validate opts
    my %opts = validate_with(
        params => \%raw_opts,
        spec   => {
            do_isl_link_up => { regex => qr/^[01]$/, default => 1, },
            suspend_bdfu   => { regex => qr/^[01]$/, default => 1, }
        },
    );

    if ( $opts{do_isl_link_up} ) {
        $Log->comment("prepare_for_heal_aggrs_execution: bring up ISL link");
        if ( $self->has_switches() ) {
            $self->_do_isl_links_modify( modify => 'up' );
        }
        elsif ( $self->is_glowworm() ) {
            $self->_do_remote_sas_adapters_modify(
                command => 'adapter_online' );
        }
        elsif ( $self->is_atto() ) {
            ${ $self->L_get_cluster_mcc_nodes() }[0]
              ->_do_remote_atto_adapters_up();
        }
        else {
            $Log->warn("Should not get here");

            # Should not get here
        }
    }

    if ( $opts{suspend_bdfu} ) {
        $self->L_set_bdfu('off');
    }

}

sub enable_isl_ports {
    my $trace = NACL::MTask::MCC::Log::TRACELOG->start_timer();    #metrics

    my $self = shift;

    #my %raw_opts = @_;

    # validate opts
    #my %opts = validate_with(
    #params => \%raw_opts,
    #spec   => {
    #},
    #);

    # enable ISL ports if they are disabled earlier
    my $isl_ports = $self->_pre_switchover_disabled_isl_ports();
    if ( defined($isl_ports) && scalar @{$isl_ports} ) {
        $Log->comment('Enabling the ISL ports');
        foreach my $isl_port ( @{$isl_ports} ) {
            $isl_port->enable();
        }
    }

}

=head2 prepare_for_switchback_execution()

    $cluster_A->prepare_for_switchback_execution(
                # default values shown here
            do_peer_metrocluster_site_nodes_boot           => 1,
            do_intercluster_network_interface_enable       => 0,
            do_peer_metrocluster_command_interface_refresh => 1,
            do_wait_seconds_for_cluster_peering_up         => 90,

                # vsims are slow re disk migration
                # so is reserver-id currently in real systems
            timeout_for_slow_disk_SANOWN_updates      => 1800,

            # this is obsolete
            do_disaster_site_aggrs_offline                 => 0,

This method performs various actions often/usually performed before this site's
cluster executes a DR switchback command.

=over

=item Options

=item C<< do_peer_metrocluster_site_nodes_boot => 0|1 (default = 1) >>

(Optional) This option will attempt to boot all the nodes in this cluster's
Metrocluster peer cluster.  (I.e., if this called on cluster A, it wiil try
to boot all nodes in cluster B.)

=item C<< do_intercluster_network_interface_enable => 0|1 (default = 1) >>

(Optional) This option will enable the intercluster network interface on this
cluster which allows this cluster to see its Metrocluster peer cluster.  (See
also the do_intercluster_network_interface_disable option in
prepare_for_switchover_execution().)

=item C<< do_peer_metrocluster_command_interface_refresh => 0|1 (default = 1) >>

(Optional) This option will re-enable the command interface to allow
automation to access the disaster site CLI again.

=item C<< do_wait_seconds_for_cluster_peering_up => 360 (default = 360) >>

(Optional) This option manipulates the amount of time to wait for cluster
peering to be reinstated beteen the metrocluster sites after the disaster site
nodes have been brought back up.

=item C<< timeout_for_slow_disk_SANOWN_updates => 1800 (default = 1800) >>

(Optional) This is the same phenomenon as is addressed in
perform_switchback_execution_recovery().

=item C<< do_disaster_site_aggrs_offline => 0|1 (default = 0) >>

(Optional) This option will offline all the DR aggrs before switchback
operation to avoid panic.

NOTE: sn_warm chunk 5 renders this option obsolete.  The default is now set to
0 (false), and a warning will be logged if it is set to 1 by the user.  Plan to
remove this option at the start of chunk 6.

=item C<< wait_for_auto_giveback => 0|1 (default = 1) >>

(Optional) This option will control whether we should wait for an auto giveback
to process if one of the nodes boots up into the 'waiting for giveback' prompt.

=item C<< wait_for_aggregates => 0|1 (default = 1) >>

(Optional) This option will control whether we wait for all aggregates
to resync before returning.

=back

=cut

sub prepare_for_switchback_execution {
    my $trace = NACL::MTask::MCC::Log::TRACELOG->start_timer();    #metrics

    my $self     = shift;
    my %raw_opts = @_;

    # validate opts
    my %opts = validate_with(
        params => \%raw_opts,
        spec   => {
            do_peer_metrocluster_site_nodes_boot =>
              { regex => qr/^[01]$/, default => 1, },
            do_intercluster_network_interface_enable =>
              { regex => qr/^[01]$/, default => 1, },
            do_peer_metrocluster_command_interface_refresh =>
              { regex => qr/^[01]$/, default => 1, },
            do_wait_seconds_for_cluster_peering_up =>
              { regex => qr/^\d+$/, default => 360, },
            timeout_for_slow_disk_SANOWN_updates =>
              { regex => qr/^\d+$/, default => 7200, },
            do_disaster_site_aggrs_offline =>
              { regex => qr/^[01]$/, default => 0, },
            wait_for_auto_giveback => { regex => qr/^[01]$/, default => 1, },
            wait_for_aggregates    => { regex => qr/^[01]$/, default => 1, },
            suspend_bdfu           => { regex => qr/^[01]$/, default => 1, },

        },
    );

    ##rvalsan MCCIP: set param do_peer_metrocluster_site_nodes_boot to 0
    ##incase of MCCIP
    if ( $self->_metrocluster()->is_mccip() ) {
        $opts{do_peer_metrocluster_site_nodes_boot} = 0;
    }

    my @subtests = ();

    #if we didn't power off any nodes, check to see if they are powered
    #off anyway.  Potentially this adds a little time here so maybe we should
    #only do this during bad mcc startup...but we removed that restriction
    #from halted transit object creation so maybe we should do the same here
    if ( !defined $self->_prepare_for_switchover_execution_nodes_powered_off() )
    {
        $self->_prepare_for_switchover_execution_nodes_powered_off(
            $self->_get_powered_off_peer_nodes() );
    }

    if ( defined $self->_prepare_for_switchover_execution_nodes_powered_off() )
    {

        foreach my $node (
            @{ $self->_prepare_for_switchover_execution_nodes_powered_off() } )
        {
            $Log->comment( "powering on node " . $node->get_name() );

   # power on the nodes but don't attempt to transit to the CLI,
   # let the "do_peer_metrocluster_site_nodes_boot" option control that behavior
            my $subtest =
              $node->subtest_system_power( 'power' => 'on', 'goto_cli' => 0 );
            push( @subtests, $subtest );
        }
        Subtest::wait_finish( subtests => \@subtests );
        $self->_prepare_for_switchover_execution_nodes_powered_off(undef);

    }

    if ( $opts{do_peer_metrocluster_site_nodes_boot} ) {
        $Log->comment('Booting MCC associated nodes at disaster site...');
        my @subtest_handles = ();
        if ( !defined $self->_pre_switchover_halted_node_transit_objs() ) {

  # if there are no halted node transit objects cached from the switchover phase
  # recreate them for all the nodes of the cluster peer
            $Log->comment(
'Transit objects for halted nodes do not exist, recreating them...'
            );
            $self->_pre_switchover_halted_node_transit_objs(
                $self->_create_peer_transit_objects() );
        }
        my $halted_node_obj = $self->_pre_switchover_halted_node_transit_objs();
        foreach my $transit_obj ( @{$halted_node_obj} ) {
            push @subtest_handles, subtest(
                sub {
                    my $iftakenover;
                    if ( $opts{wait_for_auto_giveback} ) {
                        $iftakenover = "wait";
                    }
                    else {
                        $iftakenover = "error";
                    }
                    try {
                        $transit_obj->change_state(
                            to          => 'CLI',
                            timeout     => 1500,
                            iftakenover => $iftakenover
                        );
                    }
                    catch NACL::Transit::Exceptions::TransitException with {
                        my $e = shift;
                        if (    $e->text() eq 'partner has taken over'
                            and $iftakenover eq 'error' )
                        {
                 # in this case, we are not waiting for an auto giveback
                 # and this node is in a waiting for giveback state, continue on
                            $Log->trace(
'Node is currently in a taken over state. Not waiting '
                                  . 'for an auto giveback to occur.' );
                        }
                        else {
                            # unknown error, rethrow the exception
                            $e->throw();
                        }
                    };
                },
                '-bg',
                '-runid',
                $transit_obj->{conn}->{connhost}
                  . '_prepare_for_switchback_execution_up'
            );
        }
        Subtest::wait_finish( subtests => \@subtest_handles );
        $self->_pre_switchover_halted_node_transit_objs(undef);
    }

    if ( $opts{do_peer_metrocluster_command_interface_refresh} ) {

        # Refresh both site metrocluster component connections
        $Log->comment('Refreshing all inactive SSH connections');
        $self->_metrocluster()
          ->refresh_command_interfaces( max_reconnect => 20 );
    }

    # additional workaround stuff for do_peer_metrocluster_site_nodes_boot
    if ( $opts{do_peer_metrocluster_site_nodes_boot} ) {

        # big workaround after nodes reboot...
        unless ( $self->_metrocluster()
            ->has_uichange( uichange => 'cluster-stream-down-on-switchover' ) )
        {
            $Log->comment(
                    'BURT 554679 + 662064 workaround.  BEGIN to bring up '
                  . 'admin-state of crs stream' );

            #BURT 662064 mods:
            #survivor SENDER stream up before switcback
            #we are skipping disaster side stuff as dev _thinks_ it's OK.
            #put in if needed
            ##disaster RECEIVER stream up before switchback
            #
            #survivor RECEIVER stream up AFTER switcback
            #we are skipping disaster side stuff as dev _thinks_ it's OK.
            #put in if needed
            ##disaster SENDER stream up AFTER switchback

            my $survivor_site_ci =
              $self->get_Metrocluster_component()->command_interface();

            my $parsed_output =
              $survivor_site_ci->apiset( interface => 'CLI' )
              ->cluster_identity_show->get_parsed_output();

            my $cluster_uuid = $parsed_output->[0]->{'uuid'};

            $parsed_output =
              $survivor_site_ci->apiset( interface => 'CLI' )
              ->debug_crs_stream_show(
                'direction' => 'sender',
                'synopsis'  => '"MetroCluster Cluster Domain Sender Stream"',
              )->get_parsed_output();

            my $cmd =
                'debug smdb table mccStreamDBv2 modify -cluster-uuid '
              . $cluster_uuid
              . ' -stream-target-admin-state up ';

            foreach my $stream ( @{$parsed_output} ) {
                my $c_cmd = $cmd . '-stream-uuid ' . $stream->{'stream-uuid'};

                $survivor_site_ci->apiset( interface => 'CLI' )
                  ->execute_command( command => $c_cmd );

                $survivor_site_ci->apiset( interface => 'CLI' )
                  ->debug_crs_stream_modify_stream(
                    'stream-uuid' => $stream->{'stream-uuid'},
                    'admin-state' => 'up'
                  );

            }

            # the stream may not come up immediately, retry for a while...
            my $end_time = time() + ( 2 * 60 );    # time allowed
            my $stream_done = 0;
            while ( !$stream_done && ( time() < $end_time ) ) {
                $parsed_output =
                  $survivor_site_ci->apiset( interface => 'CLI' )
                  ->debug_crs_stream_show(
                    'direction' => 'sender',
                    'synopsis' => '"MetroCluster Cluster Domain Sender Stream"',
                  )->get_parsed_output();

                my $check = 1;
                foreach my $stream ( @{$parsed_output} ) {
                    if ( $stream->{'admin-state'} ne 'up' ) {
                        $check = 0;

                        # iff retry time ran out, throw an exception here
                        if ( time() > $end_time ) {
                            NACL::MTask::MCC::Exceptions::BaseException->throw(
                                'Failed to modify stream admin-state to up');
                        }
                    }
                }
                $stream_done = $check;
                if ( !$stream_done ) {
                    $Log->comment("sleeping 10 to retry stream up check...");
                    Tharn::snooze(10);
                }
            }

            # shouldn't really need this except for corner cases maybe
            if ( !$stream_done ) {
                NACL::MTask::MCC::Exceptions::BaseException->throw(
                    'Failed to modify stream admin-state to up');
            }
        }

        unless (
            $self->_metrocluster()->workaround_expired(
                uichange => 'mcc_expire_workaround_490829',
                burt     => '490829'
            )
          )
        {
            $Log->comment( 'Part 2/3 of the BURT 645504 (+490829) '
                  . 'workaround.  BEGIN...' );
            $Log->comment( 'Stopping all data vservers (not the special '
                  . 'vservers such as vs0, MCC_Vserver and '
                  . 'CRS_Vserver) at the disaster site...' );
            my $survivor_site_ci =
              $self->get_Metrocluster_component()->command_interface();

            # re "disaster_site_ci": makes sense only after heal roots...
            my $disaster_site_ci =
              $self->get_peer()->get_Metrocluster_component()
              ->command_interface();

            # subtype seems to be missing from component spec.  must use apiset.
            #my @disaster_site_src_data_vservers = NACL::C::Vserver->find(
            #           command_interface => $disaster_site_ci,
            #           filter            => { type    => 'data|cluster',
            #                                  subtype => 'sync-source', },
            #           allow_empty       => 1 );
            try {
                #large # of vservers can take a long time to show
                my $old_timeout = $disaster_site_ci->apiset->get_timeout();
                $disaster_site_ci->apiset->set_timeout(
                    'connectrec-timeout' => 10 * 60 );
                my $vserver_specs;
                try {
                    $vserver_specs =
                      $disaster_site_ci->apiset( interface => 'CLI' )
                      ->vserver_show(
                        type    => 'data',
                        subtype => 'sync-source'
                      )->get_parsed_output();
                }
                finally {
                    $disaster_site_ci->apiset->set_timeout(
                        'connectrec-timeout' => $old_timeout );
                };
                my @disaster_site_src_data_vservers = map {
                    NACL::C::Vserver->new(
                        command_interface => $disaster_site_ci,
                        vserver           => $_->{vserver}
                      )
                } @{$vserver_specs};

                # remove any vservers on node's root aggr (is redundant check?)
                # TBD If really required

                # stop each such vserver found
                foreach my $vserver (@disaster_site_src_data_vservers) {
                    try {
                        $vserver->stop();
                    }
                    catch NACL::APISet::Exceptions::CommandFailedException
                      with {
                        # ignore already stopped vservers
                        my $exception = shift;
                        my $acceptable_error_msg =
                            'Vserver stop requires admin-state of the '
                          . 'Vserver to be running or restricted';
                        unless ( $exception->text() =~ /$acceptable_error_msg/ )
                        {
                            # rethrow unexpected exception
                            $exception->throw();
                        }
                      };
                }
            }
            catch NACL::APISet::Exceptions::NoMatchingEntriesException with {

                # why this is an exception is beyond me...  :p    ignore it.
            };

            $Log->comment('Sleeping 60 sec for vserver stops...');
            Tharn::snooze(60);

            $Log->comment( 'On survivor site: re-start the switched over '
                  . 'data vservers...' );

            # subtype seems to be missing from component spec.  must use apiset.
            #my @SOed_data_vservers = NACL::C::Vserver->find(
            #           command_interface => $survivor_site_ci,
            #           filter            => { type    => 'data|cluster',
            #                                  subtype => 'sync-destination', },
            #           allow_empty       => 1 );
            my @SOed_data_vservers = ();
            try {
                #large # of vservers can take a long time to show
                my $old_timeout = $survivor_site_ci->apiset->get_timeout();
                $survivor_site_ci->apiset->set_timeout(
                    'connectrec-timeout' => 10 * 60 );
                my $vserver_specs;
                try {
                    $vserver_specs =
                      $survivor_site_ci->apiset( interface => 'CLI' )
                      ->vserver_show(
                        type    => 'data',
                        subtype => 'sync-destination'
                      )->get_parsed_output();
                }
                finally {
                    $survivor_site_ci->apiset->set_timeout(
                        'connectrec-timeout' => $old_timeout );
                };
                @SOed_data_vservers = map {
                    NACL::C::Vserver->new(
                        command_interface => $survivor_site_ci,
                        vserver           => $_->{vserver}
                      )
                } @{$vserver_specs};

                # remove any vservers on node's root aggr (is redundant check?)
                # TBD If really required

                # start each such vserver found
                map { $_->start() } @SOed_data_vservers;
            }
            catch NACL::APISet::Exceptions::NoMatchingEntriesException with {

                # why this is an exception is beyond me...  :p    ignore it.
                @SOed_data_vservers = ();
            };

            $Log->comment('Sleeping 120 sec for table to update...');
            Tharn::snooze(120);
        }

        unless (
            $self->_metrocluster()->workaround_expired(
                uichange => 'mcc_expire_workaround_652320',
                burt     => '645504'
            )
          )
        {

            $Log->comment('Part 2/3 of the BURT 645504 workaround.  BEGIN...');
            my $survivor_site_ci =
              $self->get_Metrocluster_component()->command_interface();

            # re "disaster_site_ci": makes sense only after heal roots...
            my $disaster_site_ci =
              $self->get_peer()->get_Metrocluster_component()
              ->command_interface();

            $Log->comment('Showing dm vserver streams @ survivor site:');
            $survivor_site_ci->apiset( interface => 'CLI' )
              ->execute_command( command => 'debug dm vserver stream show' );
            $Log->comment('Showing dm vserver streams @ disaster site:');
            $disaster_site_ci->apiset( interface => 'CLI' )
              ->execute_command( command => 'debug dm vserver stream show' );

         #don't delete vserver streams until crs agent
         #(and cluster/vserver dm) has acquired master duties:
         # 5 minutes timeout here OK
         #
         #siteB::*> debug crs service-level modules show -module-name CrsAgent
         #
         #               Module Name: CrsAgent
         #                     Service Admin State: up
         #                     Service Operational State: acquired_master_duties
         #debug crs service-level modules show
         #               Admin
         #               Module Name    State    Operational State       Remarks
         #               -------------  -------  ----------------------  -------
         #               CrsAgent       up       acquired_master_duties  Ready
         #

            my %modules = (
                'CrsAgent'  => 1,
                'ClusterDM' => 1,
                'VserverDM' => 1,
            );
            $Log->comment( 'Waiting for CrsAgent, ClusterDM and VserverDM to'
                  . ' acquire master '
                  . 'duties before deleting streams at the '
                  . 'disaster site...' );
            my $end_time = time() + ( 5 * 60 );    # time allowed
            my $crs_agent_done = 0;
            while ( !$crs_agent_done && ( time() < $end_time ) ) {
                my @crsagent;

               #We actually need to check both nodes of the cluster because this
               #is a node scoped command (see burt 701848)
                my $mcc_disaster_site_nodes =
                  $self->get_peer()->L_get_cluster_mcc_nodes();
                foreach my $mcc_node ( @{$mcc_disaster_site_nodes} ) {
                    my $crsagent =
                      $mcc_node->get_Node_component()->command_interface()
                      ->apiset( interface => 'CLI' )
                      ->debug_crs_service_level_modules_show(
                        'module-name' => 'CrsAgent|ClusterDM|VserverDM', )
                      ->get_parsed_output();

                    #print Dumper $crsagent;
                    push @crsagent, @{$crsagent};
                }

                foreach my $module ( keys %modules ) {
                    if (
                        grep {
                            ( $_->{'module-name'} eq $module )
                              && (
                                $_->{'oper-state'} eq 'acquired_master_duties' )
                        } @crsagent
                      )
                    {
                        delete $modules{$module};
                        if ( !scalar keys %modules ) {
                            $crs_agent_done = 1;
                        }
                    }
                    else {
                        $Log->comment('Sleeping 15 sec till next check...');
                        Tharn::snooze(15);
                    }
                }
            }

            unless ($crs_agent_done) {
                NACL::MTask::MCC::Exceptions::BaseException->throw(
                    join( ',', keys %modules )
                      . ' Failed to acquire master duties after 5 minutes.' );
            }

            $Log->comment( 'Deleting all sender side streams at the '
                  . 'disaster site...' );
            my $cmd = 'debug dm vserver stream delete -stream-uuid * '
              . '-abort-pending-requests true';
            $disaster_site_ci->apiset( interface => 'CLI' )
              ->execute_command( command => $cmd );

            $Log->comment('Sleeping 120 sec for table to update...');
            Tharn::snooze(120);

            # verification
            $cmd = 'debug dm vserver stream show';
            my $output =
              $disaster_site_ci->apiset( interface => 'CLI' )
              ->execute_command( command => $cmd );

            unless ( $output =~ /This table is currently empty/i ) {
                NACL::MTask::MCC::Exceptions::BaseException->throw(
                    'Failed to delete vserver steams.');
            }

  #debug crs stream delete here - ui change string mcc_stream_delete_698006
  #debug crs stream delete-all-streams
  #warning, could fail, may have to retry: Error: command failed: RPC: Timed out
  #
  #do we only need this on the disaster site?? manually, both didn't hurt but
  #may not be necessary
            unless (
                $self->_metrocluster()->workaround_expired(
                    uichange => 'mcc_stream_delete_698006',
                    burt     => '698006'
                )
              )
            {
                $end_time = time() + ( 1 * 60 );    # time allowed
                my $crs_stream_delete = 0;
                while ( !$crs_stream_delete && ( time() < $end_time ) ) {
                    $Log->comment( 'Deleting all crs streams at the '
                          . 'disaster site...(workaround)' );
                    try {
                        my $out =
                          $disaster_site_ci->apiset( interface => 'CLI' )
                          ->debug_crs_stream_delete_all_streams()
                          ->get_parsed_output();
                        $crs_stream_delete = 1;
                    }
                    catch NATE::BaseException with {
                        my $out =
                          $disaster_site_ci->apiset( interface => 'CLI' )
                          ->debug_crs_stream_show()->get_parsed_output();
                        if ( scalar @{$out} ) {
                            $Log->comment("crs stream delete failed, retry...");
                        }
                        else {
                            #no crs streams, delete successful
                            $crs_stream_delete = 1;
                        }
                    };
                }
            }

            # convenience variables only
            my $REBASELINE_COMMAND = 'debug smdb table '
              . 'metrocluster_vserver_rebaseline_all_vservers run -drc-job-id 1';
            my $TRACKER_TABLE = 'metrocluster_vserver_rebaseline_tracker';

            # deal with the possibility of old entries in
            # metrocluster_vserver_rebaseline_tracker table
            {
                my $entries =
                  $survivor_site_ci->apiset( interface => 'CLI' )
                  ->debug_smdb_table_show( table => $TRACKER_TABLE )
                  ->get_parsed_output();
                if ( ref($entries) eq 'ARRAY' ) {
                    if ( scalar( @{$entries} ) > 0 ) {
                        $Log->comment('Cleaning up old tracker entries.');
                        $survivor_site_ci->apiset( interface => 'CLI' )
                          ->execute_command( command => $REBASELINE_COMMAND );
                        $Log->comment('Sleeping 30 sec for table to update...');
                        Tharn::snooze(30);
                    }
                }

                #TODO failing with multiple vservers at this time, retry
                my $end_time = time() + ( 5 * 60 );    # time allowed
                while ( time() < $end_time ) {

                    # retry better give what we expect now...
                    $entries =
                      $survivor_site_ci->apiset( interface => 'CLI' )
                      ->debug_smdb_table_show( table => $TRACKER_TABLE )
                      ->get_parsed_output();
                    unless ( ( ref($entries) eq 'ARRAY' )
                        && ( scalar( @{$entries} ) == 0 ) )
                    {
                        $Log->comment( "Failed to clean the $TRACKER_TABLE "
                              . "table! (1), retrying for "
                              . ( $end_time - time )
                              . " more second..." );
                        Tharn::snooze(30);
                    }
                    else {
                        last;
                    }
                }

                $entries =
                  $survivor_site_ci->apiset( interface => 'CLI' )
                  ->debug_smdb_table_show( table => $TRACKER_TABLE )
                  ->get_parsed_output();
                unless ( ( ref($entries) eq 'ARRAY' )
                    && ( scalar( @{$entries} ) == 0 ) )
                {
                    NACL::MTask::MCC::Exceptions::BaseException->throw(
                        "Failed to clean the $TRACKER_TABLE table! (1)");
                }
            }

            $Log->comment( 'On survivor site: rebaseline vservers and wait '
                  . 'for its completion...' );
            $survivor_site_ci->apiset( interface => 'CLI' )
              ->execute_command( command => $REBASELINE_COMMAND );

            $Log->comment('Sleeping 120 sec for table to update...');
            Tharn::snooze(120);

            # subtype seems to be missing from component spec.  must use apiset.
            #my @SOed_data_admin_vservers = NACL::C::Vserver->find(
            #           command_interface => $survivor_site_ci,
            #           filter            => { type    => 'data|cluster',
            #                                  subtype => 'sync-destination', },
            #           allow_empty       => 1 );
            my @SOed_data_admin_vservers = ();
            try {
                #large # of vservers can take a long time to show
                my $old_timeout = $survivor_site_ci->apiset->get_timeout();
                $survivor_site_ci->apiset->set_timeout(
                    'connectrec-timeout' => 10 * 60 );
                my $vserver_specs;
                my $admin_vserver_specs;
                try {
                    $vserver_specs =
                      $survivor_site_ci->apiset( interface => 'CLI' )
                      ->vserver_show(
                        type    => 'data',
                        subtype => 'sync-destination'
                      )->get_parsed_output();
                    $admin_vserver_specs =
                      $survivor_site_ci->apiset( interface => 'CLI' )
                      ->vserver_show( type => 'admin' )->get_parsed_output();
                }
                finally {
                    $survivor_site_ci->apiset->set_timeout(
                        'connectrec-timeout' => $old_timeout );
                };
                @SOed_data_admin_vservers = map {
                    NACL::C::Vserver->new(
                        command_interface => $survivor_site_ci,
                        vserver           => $_->{vserver}
                      )
                } ( @{$vserver_specs}, @{$admin_vserver_specs} );

                # remove any vservers on node's root aggr (is redundant check?)
                # TBD If really required

            }
            catch NACL::APISet::Exceptions::NoMatchingEntriesException with {

                # why this is an exception is beyond me...  :p    ignore it.
                @SOed_data_admin_vservers = ();
            };

            #need this sleep?? probably only needed for starting ...TODO ...

            #$Log->comment('Sleeping 120 sec for table to update...');
            #Tharn::snooze(120);

            # this is essentially a list of vservers we are waiting for
            # rebaseline success on
            my %vserver_rebaseline_waits =
              map { $_->get_one_state_attribute('uuid') => 1; }
              @SOed_data_admin_vservers;
            $Log->comment( 'vserver_rebaseline_waits (1) = '
                  . NACL::MTask::MCC::Utils->dump(%vserver_rebaseline_waits) );

            my $all_vserver_rebaselines_were_successful = 0;
            $end_time = time() + ( 5 * 60 );    # time allowed
            while ( time() < $end_time ) {
                my $parsed_output =
                  $survivor_site_ci->apiset( interface => 'CLI' )
                  ->debug_smdb_table_show( table => $TRACKER_TABLE )
                  ->get_parsed_output();
                foreach my $element ( @{$parsed_output} ) {
                    my $success_vserver_uuids = $element->{'success-vservers'};
                    if ( ref($success_vserver_uuids) ) {
                        my @uuids = @{$success_vserver_uuids};
                        foreach my $uuid (@uuids) {
                            $Log->comment(
                                "Deleting success vserver uuid = $uuid\n");
                            delete( $vserver_rebaseline_waits{$uuid} );
                        }
                    }
                    elsif ( $success_vserver_uuids ne q{-} ) {

                        # i don't think this case should ever happen.
                        $Log->comment( 'Deleting success vserver uuids = '
                              . $success_vserver_uuids );
                        delete(
                            $vserver_rebaseline_waits{$success_vserver_uuids} );
                    }
                    else {
                        $Log->comment("Ignoring vserver uuid = '-'.");
                    }
                }

                # not waiting on any more successes?
                $Log->comment( 'vserver_rebaseline_waits (2) = '
                      . NACL::MTask::MCC::Utils->dump(%vserver_rebaseline_waits)
                );
                if ( scalar( keys(%vserver_rebaseline_waits) ) == 0 ) {
                    $all_vserver_rebaselines_were_successful = 1;
                    last;
                }
                else {
                    Tharn::snooze(15);
                }
            }

            unless ($all_vserver_rebaselines_were_successful) {

                #dump debugging info before throwing exception...
                $Log->comment('rebaseline failed, dumping debug info...');
                $Log->comment('disaster site:');
                foreach my $cmd (
                    'debug dm vserver statistics latency queue show',
                    'debug dm vserver statistics latency baseline show',
                    'debug crs queue show',
                    'debug crs stream operational show',
                  )
                {
                    $disaster_site_ci->apiset( interface => 'CLI' )
                      ->execute_command( command => $cmd );
                }

                $Log->comment('survivor site:');
                foreach my $cmd (
                    'debug dm vserver statistics latency queue show',
                    'debug crs queue show',
                    'debug crs stream operational show',
                  )
                {
                    $survivor_site_ci->apiset( interface => 'CLI' )
                      ->execute_command( command => $cmd );
                }

                NACL::MTask::MCC::Exceptions::BaseException->throw(
                        'Failed to rebaseline vservers: '
                      . join( q{, }, keys(%vserver_rebaseline_waits) )
                      . q{!} );
            }

            $Log->comment( 'Cleaning up all the tracker entries and '
                  . 're-establishing the streams in the forward direction...' );
            $survivor_site_ci->apiset( interface => 'CLI' )
              ->execute_command( command => $REBASELINE_COMMAND );

            my $entries =
              $survivor_site_ci->apiset( interface => 'CLI' )
              ->debug_smdb_table_show( table => $TRACKER_TABLE )
              ->get_parsed_output();

            unless ( ( ref($entries) eq 'ARRAY' )
                && ( scalar( @{$entries} ) == 0 ) )
            {
                NACL::MTask::MCC::Exceptions::BaseException->throw(
                    "Failed to clean the $TRACKER_TABLE table! (2)");
            }

            $Log->comment('Part 2/3 of the BURT 645504 workaround.  ...END.');
        }
    }

    #used to enable ISL ports here, moved to prepare_for_heal_aggrs_execution

    if ( $opts{do_intercluster_network_interface_enable} ) {
        $Log->comment('Enabling intercluster interfaces to partner site...');
        my $interfaces_to_enable =
          $self->_pre_switchover_intercluster_network_interfaces_downed();

        if ( defined $interfaces_to_enable ) {
            foreach my $network_interface ( @{$interfaces_to_enable} ) {
                $network_interface->modify( 'status-admin' => 'up' );
            }
            $self->_pre_switchover_intercluster_network_interfaces_downed(
                undef);
        }
    }

    # TBD: reimplement as a more exacting way of doing this.
    if ( $opts{do_wait_seconds_for_cluster_peering_up} ) {
        my $wait_secs = $opts{do_wait_seconds_for_cluster_peering_up};
        $Log->comment( "Waiting $wait_secs seconds for reinstatement of "
              . "cluster peering..." );

        #sleep($wait_secs);
        my @mcc_nodes = @{ $self->L_get_cluster_mcc_nodes() };
        $self->L_check_cluster_peer_connectivity( timeout => $wait_secs );
    }

    $self->L_workaround_force_reserver_id_update();

    if ( $opts{timeout_for_slow_disk_SANOWN_updates} ) {
        my $wait_secs =
          $opts{timeout_for_slow_disk_SANOWN_updates} -
          $opts{do_wait_seconds_for_cluster_peering_up};
        if ( $wait_secs > 0 ) {
            my $dr_groups          = $self->_metrocluster()->get_DR_groups();
            my $sanown_timeout_end = time + $wait_secs;

            # we don't want to wait more than timeout (not timeout per call)
            $Log->comment("Verifying SANOWN disk state migrations...");

            foreach my $dr ( sort keys %{$dr_groups} ) {
                my $dr_group = $dr_groups->{$dr};
                $Log->comment(
                    "Waiting for SANOWN disk state migrations: $dr ...");

                # timeout is total time, not each
                # L_wait_for_slow_SANOWN_state_migration logs warnings if needed
                $dr_group->get_shared_storage()
                  ->L_wait_for_slow_SANOWN_state_migration(
                    timeout => $sanown_timeout_end - time );
            }
        }
    }

    if ( $opts{do_disaster_site_aggrs_offline} ) {
        $Log->warn( 'prepare_for_switchback_execution('
              . 'do_disaster_site_aggrs_offline => 1) is probably an '
              . 'obsolete usage!' );
        my @mcc_nodes = @{ $self->L_get_cluster_mcc_nodes() };
        foreach my $mcc_node (@mcc_nodes) {
            my $node = $mcc_node->get_Node_component();
            my $response =
              $node->get_7m_or_nodescope_apiset( interface => 'CLI' )
              ->execute_command( command => 'aggr status' );

            my @aggrs = split( /\n/, $response );

            foreach my $aggr_names (@aggrs) {
                if ( $aggr_names =~ /offline|online/ ) {
                    my @aggr_name = split( /offline|online/, $aggr_names );
                    my $cmd = 'raid_config info listaggr ' . $aggr_name[0];
                    $response =
                      $node->get_7m_or_nodescope_apiset( interface => 'CLI' )
                      ->execute_command(
                        command => $cmd,

                        #      'privilege-level' => 'test'
                      );

                    if ( $response !~ /dr_home_owner_id=0/i ) {
                        $cmd      = 'aggr offline ' . $aggr_name[0] . ' -a -f';
                        $response = $node->run( command => $cmd );
                        $response = $node->get_7m_or_nodescope_apiset(
                            interface => 'CLI' )->execute_command(
                            command   => $cmd,

                            #      'privilege-level' => 'test'
                            );
                        $Log->comment($response);
                    }
                }
            }
        }
    }

    # setting the disaster_only option because for SB all we care
    # about is really just the disaster site aggrs to have completed
    # resyncing
    if ( $opts{wait_for_aggregates} ) {
        $self->_metrocluster()->wait_for_aggr_resync(
            disaster_only => 1,
            reason        => 'Preparing for Metrocluster switchback'
        );
    }

    if ( $opts{suspend_bdfu} ) {
        $self->L_set_bdfu('off');
    }
}

# runs workaround to try to force reserver ids to update quickly
sub L_workaround_force_reserver_id_update {
    my $trace = NACL::MTask::MCC::Log::TRACELOG->start_timer();    #metrics
    my $self  = shift;

    if ( $ENV{DISABLE_MCC_DISK_SHOW_Z} ) {
        $Log->comment("NOT running workaround 649633 force reserver id update");
        return;

        #we are manually disabling this as the UI change string below was never
        #checked in with the fix
    }

    #we are now re-enabling this functionality per burt 821566
    #on the claim that it won't affect mcc directly

    # validate opts
    my %opts = validate_with(
        params => \@_,
        spec   => {
            known_live_sites => {
                type     => ARRAYREF,
                optional => 1,
            },
        },
    );

    my $live_sites_ref =
      ( defined( $opts{known_live_sites} ) )
      ? $opts{known_live_sites}
      : $self->_metrocluster()->L_determine_live_sites();

    #unless ($self->_metrocluster()->workaround_expired(
    #uichange => 'mcc_expire_workaround_649633',
    #burt => '649633')) {
    my @live_sites = @{$live_sites_ref};
    foreach my $site (@live_sites) {
        my $cluster = $self->_metrocluster()->get_cluster( site => $site );
        my $sites_nodes = $cluster->L_get_cluster_mcc_nodes();

        foreach my $mcc_node ( @{$sites_nodes} ) {
            try {
                my $ci = $mcc_node->get_Node_component()->command_interface();
                my $out =
                  $ci->apiset( interface => 'CLI' )
                  ->execute_command(
                    command => 'node run local disk show -v -z' );

                #print $mcc_node->get_name() . ' node run local disk show:' .
                #Dumper $out;
            }
            catch NATE::BaseException with {
                $Log->comment("Ignoring error in workaround attempt ");
            };
        }
    }

    #}

}

=head2 cycle_to_normal_DR_state()

    $site_X_cluster->cycle_to_normal_DR_state( );
    $site_X_cluster->cycle_to_normal_DR_state( verify => {disks => 0} );

Perform necessary operations to bring DR state back to normal.

=over

=item Options

=item C<< verify => { disks }  >>

Verify is a hashref containing the various things you can verify.
Currently disks is the only value.

=over

=item C<< disks => 0|1 (default = 1) >>

(Optional) This option will verify the disks after each state transition
if set to 1. Internally it just calls library methods to verify disks.

=back

=back

=cut

sub cycle_to_normal_DR_state {
    my $trace = NACL::MTask::MCC::Log::TRACELOG->start_timer();    #metrics

    my ( $self, %opts ) = @_;

    $opts{'state'} = 'normal';

    my $state = $self->cycle_to_DR_state(
        %opts,    #validate in cycle_to_DR_state
    );

    return $state;
}

=head2 cycle_to_DR_state()

    $site_X_cluster->cycle_to_DR_state( state => 'normal' );
    $site_X_cluster->cycle_to_DR_state( state => 'normal', verify=> {disks => 0} );

Perform necessary operations to bring DR state to the specified state.

=over

=item Options

=item C<< state => normal | switchover_completed | heal_aggrs_completed |
heal_roots_completed >>

=item C<< disable_cycle_short_circuit => 0 | 1 (default = 0) >>

(Optional)  By default (disable_cycle_short_circuit=0) the DR state will not
change if it is asked to cycle to a state it is already in, it will simply log
that it is already in that state and exit.

If disable_cycle_short_circuit=1, it will force a full cycle back to the
specified state even if that is the start state.

=item C<< verify => { disks } >>

Verify is a hashref containing the various things you can verify.
Currently disks is the only value.

=over

=item C<< run_extended >>

Passed down to run_all_validations.  Will only run any specified
extended validations if this is 1

=over

=item C<< disks => 0|1 (default = 1) >>

(Optional) This option will verify the disks after each state transition
if set to 1. Internally it just calls library methods to verify disks.

=back

=over

=item C<< prepare_for_switchover_execution passthru params >>

Most params that can be passed directly to prepare_for_switchover_execution
can also be passed to this method and are then passed directly
down to prepare_for_switchover_execution

=back

=item C<< force_nvfail_all => (0|1) (default = 0) >>

(Optional) Boolean. Defaults to false.

=back

=item C<< expect-auso => (0|1|undef) (default = undef) >>

(Optional) Flag to indicate if MCC Lib should expect AUSO on 2-node MCC. Defaults to undef which would auto-select the value based on current setting.

=back

=item C<< handle_unmirrored_aggr_presence => (0|1) (default = undef) >>

(Optional) Flag to indicate if DR SO/SB operations should be handled for unmmirrored aggr presence. Lib would automatically scan for unmirrored aggrs if set to undef. Defaults to undef.

=back

=cut

sub cycle_to_DR_state {
    my $trace = NACL::MTask::MCC::Log::TRACELOG->start_timer();    #metrics

    my ( $self, %opts ) = @_;

    # get list of all validations from mcc - create validations for param
    # and default to 1
    # validate verify opts (because Params::Validate can't handle
    # recursive hashrefs)
    $opts{verify} = validate_with(
        params => $opts{verify} || {},
        spec => $self->_metrocluster()->get_all_validations_params_spec(),
    );

    #these should only be terminal states (keys)
    my $next = {
        'normal'               => 'switchover_completed',
        'switchover_completed' => 'heal_aggrs_completed',
        'heal_aggrs_completed' => 'heal_roots_completed',
        'heal_roots_completed' => 'normal',

        #partial states:
        'switchover_partially_successful' => 'switchover_completed',
        'heal_aggrs_partially_successful' => 'heal_aggrs_completed',
        'heal_roots_partially_successful' => 'heal_roots_completed',
        'switchback_partially_successful' => 'normal',
    };

    my $next_regex = join q{|}, reverse sort { length($a) <=> length($b) }
      sort keys %{$next};

    # validate rest of opts
    %opts = validate_with(
        params => \%opts,
        spec   => {
            verify                    => HASHREF,
            skip_validations          => { regex => qr/^[01]$/, default => 0 },
            run_extended              => { type => BOOLEAN, default => 0 },
            die_on_validation_failure => { type => BOOLEAN, default => 1 },
            disable_cycle_short_circuit => { type => BOOLEAN, default => 0 },
            state => { regex => qr/^(?:$next_regex)$/ },

            #these options only for switchover
            do_intercluster_network_interface_disable =>
              { type => BOOLEAN, optional => 1, },
            'do_ic_lif_disable_after_disaster' =>
              { type => BOOLEAN, optional => 1 },
            do_peer_metrocluster_site_nodes_disaster =>
              { type => SCALAR, optional => 1 },
            'forced-on-disaster'     => { type => BOOLEAN, optional => 1 },
            'do_isl_link_down'       => { type => BOOLEAN, optional => 1 },
            'do_isl_link_down_delay' => { type => SCALAR,  optional => 1 },
            'do_isl_link_down_after_disaster' =>
              { type => BOOLEAN, optional => 1 },

            #these options only for heal aggrs
            'do_isl_link_up' => { type => BOOLEAN, optional => 1 },

            #these options only for switchover/switchback
            'override-vetoes' => { type => BOOLEAN, optional => 1 },

            #these are to support 'custom' disaster types
            'disaster_type'      => { type => SCALAR | UNDEF, optional => 1 },
            'transit_timeout'    => { type => SCALAR,         optional => 1 },
            'panic_core_dump'    => { type => BOOLEAN,        optional => 1 },
            'disable_core_file'  => { type => BOOLEAN,        optional => 1 },
            'inhibit_takeover'   => { type => BOOLEAN,        optional => 1 },
            'skip_lif_migration' => { type => BOOLEAN,        optional => 1 },
            'force_nvfail_all'   => { type => BOOLEAN,        optional => 1 },

            #param to expect AUSO on 2-node MCC
            'expect-auso'  => { type  => BOOLEAN | UNDEF, default => undef },
            'suspend_bdfu' => { regex => qr/^[01]$/,      default => 1 },

            #param for supporting unmirrored aggrs
            'handle_unmirrored_aggr_presence' => {
                type     => BOOLEAN | UNDEF,
                optional => 1,
                default  => undef
            },
        },
    );

    my $target_state = delete $opts{state};

    my $disable_cycle_short_circuit = delete $opts{disable_cycle_short_circuit};

    $Log->comment("Target metrocluster state $target_state");

    my $dr_coordinator = $self->get_DR_Coordinator();
    my $dr_groups      = $self->_metrocluster()->get_DR_groups();

    my $state = $dr_coordinator->get_DR_state();
    $Log->comment("Currently in metrocluster state $state");

    my $transitions = 0;    # number of transitions

    #if this is the start and disable_cycle_short_circuit=1,
    #transition anyway even if in target state
    while ( ( !$transitions && $disable_cycle_short_circuit )
        || $state ne $target_state )
    {
        if ( !$opts{'skip_validations'} ) {
            $Log->comment('run all (pre) validations - from cycle to...');
            $self->_metrocluster->run_all_validations(
                validation_filter => $opts{verify},
                disaster_status   => $opts{disaster_status},
                run_extended      => $opts{run_extended},
                pre_dr_operation  => 1,
                die_on_failure    => $opts{die_on_validation_failure},
            );
        }

        my $next_state = $next->{$state};
        $Log->comment("Init transition to metrocluster state $next_state");

        #we could shove these calls into a hash but this allows more flexibility
        my $disaster = $opts{do_peer_metrocluster_site_nodes_disaster};
        my $forced_on_disaster = $opts{'forced-on-disaster'};
        if ( $next_state eq 'switchover_completed' ) {

       #skip non normal state runs here (ie partial states) or get disaster
       #status will get stuck in a loop waiting for ending in a good state
       #and we really only want to run the diffs in a good starting state anyway
            if (
                $state eq 'normal'
                && (   $opts{verify}->{check_config_cluster_diff}
                    || $opts{verify}->{check_config_vserver_diff} )
                && !$ENV{MCC_SKIP_CONFIG_DIFF_PRE_VALIDATION}
              )
            {
                my $disaster_status = $opts{disaster_status}
                  || $self->_metrocluster()
                  ->get_disaster_status( include_HA => 1 );

                if ( $opts{verify}->{check_config_cluster_diff} ) {
                    $Log->comment(
                        ' validate metrocluster check config cluster diff...');
                    $self->_metrocluster()->validate_check_config_cluster_diff(
                        disaster_status => $disaster_status,
                        die_on_failure  => $opts{die_on_validation_failure}
                    );
                }
                if ( $opts{verify}->{check_config_vserver_diff} ) {
                    $Log->comment(
                        ' validate metrocluster check config vserver diff...');
                    $self->_metrocluster()->validate_check_config_vserver_diff(
                        disaster_status => $disaster_status,
                        die_on_failure  => $opts{die_on_validation_failure}
                    );
                }
            }
            else {
                $Log->comment( "dr state: $state - "
                      . 'skipping metrocluster check config diffs validation...'
                );
            }

            #It is best to calculate this value on runtime rather than saving
            #it when the object is created. The reason being the object would
            #have been created in a bad state in which case auso will be
            #disabled and the value will be 0. But when the same object is used
            #to run the second iteration of DR lifecycle, when AUSO will be
            #enabled the library should know that it is suppose to expect it
            #which it wont if the value was pre-stored in a bad mcc state.
            #This will cause the library to attempt an USO and AUSO would
            #have also kicked in causing test failure.

            #leaving it in there for debugging
            $Log->comment( 'Passed in value for expect-auso is : '
                  . NACL::MTask::MCC::Utils->dump( $opts{'expect-auso'} ) );
            if ( !defined $opts{'expect-auso'} ) {
                $Log->comment(
'User has not passed in a value for expect-auso; querying it...'
                );
                $Log->comment('Get auso config state for the local cluster');
                my $local_auso = $self->get_auso_config_state();
                $Log->comment( 'site A auso config state: '
                      . NACL::MTask::MCC::Utils->dump($local_auso) );

                $Log->comment('Get remote cluster object');
                my $remote_cluster = $self->get_peer();

                $Log->comment('Get auso config state for the remote cluster');
                my $remote_auso = $remote_cluster->get_auso_config_state();
                $Log->comment( 'site B auso config state: '
                      . NACL::MTask::MCC::Utils->dump($remote_auso) );

                $Log->comment(
                    'Determining AUSO eligibility for this MCC system');
                if (   ( defined $local_auso && lc $local_auso !~ /disabled/ )
                    && ( defined $remote_auso && lc $remote_auso !~ /disabled/ )
                  )
                {
                    $opts{'expect-auso'} = 1;
                }
                else {
                    $opts{'expect-auso'} = 0;
                }
            }

            # leaving it here for debugging
            $Log->comment( 'expect-auso value being passed to SO : '
                  . NACL::MTask::MCC::Utils->dump( $opts{'expect-auso'} ) );

            if ( $self->_metrocluster()->has_unmirrored_aggrs() ) {
                $Log->comment( 'Found unmirrored aggrs on this system. SO will '
                      . 'complete with manual recovery required state!' );
                $opts{'handle_unmirrored_aggr_presence'} = 1;
            }

            $self->prepare_for_switchover_execution(

                #do_peer_metrocluster_site_nodes_halt      => 1,
                #TO DO for burt 490829:
                #do_wait_minutes_for_cluster_appears_down  => 0,
                (
                    $disaster
                    ? ( do_peer_metrocluster_site_nodes_disaster => $disaster )
                    : ()
                ),
                (
                    defined $opts{do_intercluster_network_interface_disable}
                    ? ( do_intercluster_network_interface_disable =>
                          $opts{do_intercluster_network_interface_disable} )
                    : ()
                ),
                (
                    defined $opts{do_ic_lif_disable_after_disaster}
                    ? ( do_ic_lif_disable_after_disaster =>
                          $opts{do_ic_lif_disable_after_disaster} )
                    : ()
                ),
                (
                    defined $opts{do_isl_link_down}
                    ? ( do_isl_link_down => $opts{do_isl_link_down} )
                    : ()
                ),
                (
                    defined $opts{do_isl_link_down_delay}
                    ? ( do_isl_link_down_delay => $opts{do_isl_link_down_delay}
                      )
                    : ()
                ),
                (
                    defined $opts{do_isl_link_down_after_disaster}
                    ? ( do_isl_link_down_after_disaster =>
                          $opts{do_isl_link_down_after_disaster} )
                    : ()
                ),

                #only for custom disaster types
                (
                    defined $opts{disaster_type}
                    ? ( disaster_type => $opts{disaster_type} )
                    : ()
                ),
                (
                    defined $opts{panic_core_dump}
                    ? ( panic_core_dump => $opts{panic_core_dump} )
                    : ()
                ),
                (
                    defined $opts{disable_core_file}
                    ? ( disable_core_file => $opts{disable_core_file} )
                    : ()
                ),
                (
                    defined $opts{inhibit_takeover}
                    ? ( inhibit_takeover => $opts{inhibit_takeover} )
                    : ()
                ),
                (
                    defined $opts{skip_lif_migration}
                    ? ( skip_lif_migration => $opts{skip_lif_migration} )
                    : ()
                ),
                (
                    defined $opts{transit_timeout}
                    ? ( transit_timeout => $opts{transit_timeout} )
                    : ()
                ),
                (
                    defined $opts{suspend_bdfu}
                    ? ( suspend_bdfu => $opts{suspend_bdfu} )
                    : ()
                ),

            );

            $dr_coordinator->DR_switchover(

                #we want to support passing 0 for forced on disaster
                (
                    defined $forced_on_disaster
                    ? ( 'forced-on-disaster' => $forced_on_disaster )
                    : ()
                ),
                (
                    defined $opts{'override-vetoes'}
                    ? ( 'override-vetoes' => $opts{'override-vetoes'} )
                    : ()
                ),
                (
                    defined $opts{'force_nvfail_all'}
                    ? ( 'force_nvfail_all' => $opts{'force_nvfail_all'} )
                    : ()
                ),
                (
                    defined $opts{'expect-auso'}
                    ? ( 'expect-auso' => $opts{'expect-auso'} )
                    : ()
                ),
                (
                    defined $opts{'handle_unmirrored_aggr_presence'}
                    ? ( 'handle_SO_warnings_for_unm_aggr' =>
                          $opts{'handle_unmirrored_aggr_presence'} )
                    : ()
                ),
            );
            if ( defined $forced_on_disaster && !$forced_on_disaster ) {
                $Log->comment( 'Performed negotiated switchover, verify that '
                      . 'disaster site was brought down automatically...' );
                $self->_ensure_disaster_site_halted();
            }
        }
        elsif ( $next_state eq 'heal_aggrs_completed' ) {
            ##rvalsan MCCIP: bootup the disaster nodes incase of MCCIP
            if ( $self->_metrocluster()->is_mccip() ) {
                $self->bootup_disaster_mccip_nodes();
            }
            $self->prepare_for_heal_aggrs_execution(
                (
                    defined $opts{do_isl_link_up}
                    ? ( do_isl_link_up => $opts{do_isl_link_up} )
                    : ()
                ),

                # Pass down BDFU to turn off
                (
                    defined $opts{suspend_bdfu}
                    ? ( suspend_bdfu => $opts{suspend_bdfu} )
                    : ()
                ),

            );
            $dr_coordinator->DR_heal_aggrs();
        }
        elsif ( $next_state eq 'heal_roots_completed' ) {
            $dr_coordinator->DR_heal_root_aggrs();

            # this is placed here because it really is part of the recovery
            # after the heal roots phase, and before disk validations can be
            # done in this new state.
            $self->prepare_for_switchback_execution(
                do_peer_metrocluster_site_nodes_boot           => 1,
                do_peer_metrocluster_command_interface_refresh => 1,
            );
        }
        elsif ( $next_state eq 'normal' ) {
            $dr_coordinator->DR_switchback(
                (
                    defined $opts{'override-vetoes'}
                    ? ( 'override-vetoes' => $opts{'override-vetoes'} )
                    : ()
                ),
            );

            $self->perform_switchback_execution_recovery(
                do_peer_metrocluster_site_nodes_boot => 0, );

            if ( $self->_metrocluster()->is_mccip() ) {
                my @nacl_nodes = @{ $self->L_get_cluster_mcc_nodes() };
                foreach my $nacl_node (@nacl_nodes) {
                    my $nacl_node_obj = $nacl_node->get_Node_component();
                    NACL::MTask::MCC::Shared_Storage->check_remote_disks(
                        node => $nacl_node_obj );
                }
            }
        }
        else {
            # how did we get here?!
            $Log->comment("ERROR: bad next state $next_state (state=$state)");
            return;
        }

        # Turn BDFU back on
        if ( $opts{'suspend_bdfu'} ) {
            $self->L_set_bdfu('on');
        }

        my $disaster_status = $opts{disaster_status}
          || $self->_metrocluster()->get_disaster_status( include_HA => 1 );

        if ( !$opts{'skip_validations'} ) {
            my %validation_ops;
            $Log->comment('run all validations - from cycle to...');
            if ( $opts{'handle_unmirrored_aggr_presence'} ) {
                $validation_ops{die_on_failure} = 0;
            }
            $self->_metrocluster->run_all_validations(
                validation_filter => $opts{verify},
                disaster_status   => $disaster_status,
                run_extended      => $opts{run_extended},
                %validation_ops,
            );
        }

        $transitions++;
        $state = $dr_coordinator->get_DR_state();
    }

    return $state;
}

#rvalsan MCCIP Bootup diaster nodes before heal phase
sub bootup_disaster_mccip_nodes {
    my $trace = NACL::MTask::MCC::Log::TRACELOG->start_timer();    #metrics

    my $self     = shift;
    my %raw_opts = @_;

    # validate opts
    my %opts = validate_with(
        params => \%raw_opts,
        spec   => {
            do_peer_metrocluster_site_nodes_boot =>
              { regex => qr/^[01]$/, default => 1, },
            do_intercluster_network_interface_enable =>
              { regex => qr/^[01]$/, default => 1, },
            do_peer_metrocluster_command_interface_refresh =>
              { regex => qr/^[01]$/, default => 1, },
            wait_for_auto_giveback => { regex => qr/^[01]$/, default => 1, },
        },
    );
    my @subtests = ();

    #if we didn't power off any nodes, check to see if they are powered
    #off anyway.  Potentially this adds a little time here so maybe we should
    #only do this during bad mcc startup...but we removed that restriction
    #from halted transit object creation so maybe we should do the same here
    if ( !defined $self->_prepare_for_switchover_execution_nodes_powered_off() )
    {
        $self->_prepare_for_switchover_execution_nodes_powered_off(
            $self->_get_powered_off_peer_nodes() );
    }

    if ( defined $self->_prepare_for_switchover_execution_nodes_powered_off() )
    {

        foreach my $node (
            @{ $self->_prepare_for_switchover_execution_nodes_powered_off() } )
        {
            $Log->comment( "powering on node " . $node->get_name() );

   # power on the nodes but don't attempt to transit to the CLI,
   # let the "do_peer_metrocluster_site_nodes_boot" option control that behavior
            my $subtest =
              $node->subtest_system_power( 'power' => 'on', 'goto_cli' => 0 );
            push( @subtests, $subtest );
        }
        Subtest::wait_finish( subtests => \@subtests );
        $self->_prepare_for_switchover_execution_nodes_powered_off(undef);

    }

    if ( $opts{do_peer_metrocluster_site_nodes_boot} ) {
        $Log->comment('Booting MCC associated nodes at disaster site...');
        my @subtest_handles = ();
        if ( !defined $self->_pre_switchover_halted_node_transit_objs() ) {

  # if there are no halted node transit objects cached from the switchover phase
  # recreate them for all the nodes of the cluster peer
            $Log->comment(
'Transit objects for halted nodes do not exist, recreating them...'
            );
            $self->_pre_switchover_halted_node_transit_objs(
                $self->_create_peer_transit_objects() );
        }
        my $halted_node_obj = $self->_pre_switchover_halted_node_transit_objs();
        foreach my $transit_obj ( @{$halted_node_obj} ) {
            push @subtest_handles, subtest(
                sub {
                    my $iftakenover;
                    if ( $opts{wait_for_auto_giveback} ) {
                        $iftakenover = "wait";
                    }
                    else {
                        $iftakenover = "error";
                    }
                    try {
                        $transit_obj->change_state(
                            to          => 'CLI',
                            timeout     => 1500,
                            iftakenover => $iftakenover
                        );
                    }
                    catch NACL::Transit::Exceptions::TransitException with {
                        my $e = shift;
                        if (    $e->text() eq 'partner has taken over'
                            and $iftakenover eq 'error' )
                        {
                 # in this case, we are not waiting for an auto giveback
                 # and this node is in a waiting for giveback state, continue on
                            $Log->trace(
'Node is currently in a taken over state. Not waiting '
                                  . 'for an auto giveback to occur.' );
                        }
                        else {
                            # unknown error, rethrow the exception
                            $e->throw();
                        }
                    };
                },
                '-bg',
                '-runid',
                $transit_obj->{conn}->{connhost}
                  . '_prepare_for_healaggr_execution_up'
            );
        }
        Subtest::wait_finish( subtests => \@subtest_handles );
        $self->_pre_switchover_halted_node_transit_objs(undef);
    }

    if ( $opts{do_peer_metrocluster_command_interface_refresh} ) {

        # Refresh both site metrocluster component connections
        $Log->comment('Refreshing all inactive SSH connections');
        $self->_metrocluster()
          ->refresh_command_interfaces( max_reconnect => 20 );
    }

    if ( $opts{do_intercluster_network_interface_enable} ) {
        $Log->comment('Enabling intercluster interfaces to partner site...');
        my $interfaces_to_enable =
          $self->_pre_switchover_intercluster_network_interfaces_downed();

        if ( defined $interfaces_to_enable ) {
            foreach my $network_interface ( @{$interfaces_to_enable} ) {
                $network_interface->modify( 'status-admin' => 'up' );
            }
            $self->_pre_switchover_intercluster_network_interfaces_downed(
                undef);
        }
    }
    Tharn::snooze(90);
}

sub L_set_bdfu {
    my $self = shift;
    my ($setting) = @_;

    # Implementation note: trying to call get_disaster_status() while
    # in HA switchover times out. So we're just doing this, with a short
    # timeout to keep from unnecessarily delaying things.

    foreach my $mcc_node ( @{ $self->L_get_cluster_mcc_nodes } ) {
        try {
            $mcc_node->get_Node_component->get_7m_or_nodescope_apiset(
                interface            => 'CLI' )->execute_command(
                command              => "options raid.bdfu.suspend $setting",
                'connectrec-timeout' => 5
                );
        }
        catch NATE::BaseException with {
            my $exception = shift;
            $Log->comment( "Caught exception " . $exception->text() );
        };
    }
}

sub _ensure_disaster_site_halted {
    my $self = shift;
    my %opts =
      validate( @_,
        { transit_objs_ref => { type => ARRAYREF, optional => 1, }, } );

    my $objs =
         $opts{transit_objs_ref}
      || $self->_pre_switchover_halted_node_transit_objs()
      || $self->_pre_switchover_halted_node_transit_objs(
        $self->_create_peer_transit_objects() );
    foreach my $obj (@$objs) {
        my $timeout = time + 300;
        my $state = $obj->get_state( timeout => 300, transient => 1 );
        $Log->comment(
            $obj->{conn}->{hostp}->{hostname} . " state is: '$state'" );
        while (( $state ne 'HALTED' )
            && ( $state ne 'FIRMWARE' ) )
        {
            Tharn::snooze(15);
            $state = $obj->get_state( timeout => 300, transient => 1 );
            $Log->comment(
                $obj->{conn}->{hostp}->{hostname} . " state is: '$state'" );
            if ( time >= $timeout ) {
                if ( $state ne 'HALTED' && $state ne 'FIRMWARE' ) {
                    NACL::MTask::MCC::Exceptions::Validation->throw(
                        'The disaster node did not halt in 5 min');
                }
                else {
                    last;
                }
            }

        }
    }
    $Log->comment("disaster site nodes halted as expected");
    return;
}

=head2 unconfigure()

This method runs the unconfigure command which destroys the metrocluster
configuration.  Without any arguments it destroys all metroclusters related
to this site.  A node argument will limit this to just the nodes associated
with the passed node.

    my $result = $mcc->unconfigure();

    (Currently valid - burt #659676)
    my $result = $mcc->unconfigure(dr_group_id => 'dr_group_to_remove');

    This will become implicit shortly (it will be ignored if passed)
    (Perform a local unconfigure)
    my $result = $mcc->unconfigure(force => 'true');


Static usage is deprecated and unsupported

=over

=item Options

=over


=item C<< dr_group_id =E<gt> E<lt> dr group idE<gt> (optional) >>

If specified, limits unconfigure to just the DR partners related to that dr
group.

=item C<< force =E<gt> 'true'|'false' >>

This option if specified will run unconfigure on any one of the site.
This opton, if passed ONLY does a local unconfigure. You must
run it on the other site to complete the unconfigure operation. If the
dr_group_id is specified it will do a local unconfigure only for that dr group
on the site that has been passed.

$mcc->unconfigure(
            site  =E<gt> 'A',
            dr_group_id =E<gt> 1,
            force =E<gt> 'true',
);

The same command needs to be re-run on site 'B' as well to complete the
operation.


=back

=back

=cut

sub unconfigure {
    my $trace = NACL::MTask::MCC::Log::TRACELOG->start_timer();    #metrics
    my ( $self, @opts ) = @_;

    #validate opts
    my %opts = validate_with(
        params => \@opts,
        spec   => {
            dr_group_id => { type => SCALAR | UNDEF, optional => 1 },
            force       => { type => SCALAR | UNDEF, optional => 1 },
        },
        allow_extra => 1,
    );
    my $component      = $self->get_Metrocluster_component();
    my $peer_component = $self->get_peer()->get_Metrocluster_component();

    my $force       = $opts{force};
    my $dr_group_id = $opts{'dr_group_id'};

    my %args;    # for passing down to component unconfig call

    if ($dr_group_id) {
        $Log->comment("Dr group id is: $dr_group_id");
        $args{dr_group_id} = $dr_group_id;
    }

    my $site = $self->get_site();

    my $local_unconfig = $self->_metrocluster()->workaround_expired(
        uichange   => 'mcc-unconfigure-local-only',
        burt       => 794422,
        workaround => 0
    );

    #convert to boolean friendly
    if ( defined $force && $force eq 'true' ) {
        $args{force} = $opts{force};
        $force = 1;
    }
    elsif ( defined $force && $force eq 'false' ) {
        $args{force} = $opts{force};
        $force = 0;
    }
    else {
        $force = 0;
    }

    if ($local_unconfig) {

        #force is implicit now, so turn it on explicitly
        #but the option is no longer there so we don't
        #pass it down via args
        #TODO REMOVE THIS NEXT LINE, for test on old builds
        #$args{force} = 'true';
        $force = 1;
    }

    my $mcc_nodes =
      $self->L_get_cluster_mcc_nodes(
        $dr_group_id ? ( dr_group_id => $dr_group_id ) : () );

    my $peer_mcc_nodes =
      $self->get_peer()
      ->L_get_cluster_mcc_nodes(
        $dr_group_id ? ( dr_group_id => $dr_group_id ) : () );

    my %jobs_site;
    my %run_site;
    $Log->comment("unconfigure site: $site");

    my $run;
    my @nodes_to_be_unconfigured;
    my %hash;

    if ($force) {

        #get the nodes that will be unconfigured
        #only local site if force
        @nodes_to_be_unconfigured = @{$mcc_nodes};
        %hash = map { $_ => 1 } @nodes_to_be_unconfigured;
    }
    else {
        #get the nodes that will be unconfigured
        @nodes_to_be_unconfigured = ( @{$mcc_nodes}, @{$peer_mcc_nodes} );
        %hash = map { $_ => 1 } @nodes_to_be_unconfigured;
    }

    $Log->comment("Running unconfigure");
    my $timeout = 120 * 10;
    my %common_opts;
    $component->_copy_common_component_params(
        source => \%args,
        target => \%common_opts
    );

    my $job;
    my $result;

  RETRY_UNCONFIGURE:

    # Handle for both CommandFailedException and JobFailedException
    try {
        $result = $component->unconfigure(
            %args,
            'method-timeout' => $timeout,
            job_component    => \$job,
        );
    }
    catch NACL::APISet::Exceptions::CommandFailedException with {
        my $ex = shift;

        # Handle a known exception
        if ( $ex->text =~ /A metrocluster check operation is in progress/ ) {
            if ( $ex->text() =~
                /metrocluster operation history show -job-id (\d+)/ )
            {
                $Log->comment("Waiting for check job (id=$1) to finish ...");
                my $check_job = NACL::C::Job->find(
                    command_interface => $component->command_interface(),
                    filter            => { id => $1, }
                );

                # Wait for the metrocluster check job to go back to queued state
                my $check_job_timeout = time + 120;    # set 2 min timeout
                while ( lc $check_job->state()->state() ne 'queued' ) {
                    $Log->comment( 'Metrocluster check job is in progress. '
                          . 'Recheck in 5 secs...' );
                    Tharn::snooze(5);
                    if ( time > $check_job_timeout ) {
                        NACL::MTask::MCC::Exceptions::BaseException->throw(
                                'Timeout waiting for metrocluster check '
                              . 'job to finish!' );
                    }
                }
                goto RETRY_UNCONFIGURE;
            }
            else {
                NACL::MTask::MCC::Exceptions::BaseException->throw(
                    'Unable to get metrocluster check job ID');
            }
        }
        else {
            # Throw the exception if it is not known
            $ex->throw();
        }
    }
    catch NACL::APISet::Exceptions::JobFailedException with {
        my $ex = shift;

        # Handle a known exception
        if ( $ex->text =~ /A metrocluster check operation is in progress/ ) {
            if ( $ex->text() =~
                /metrocluster operation history show -job-id (\d+)/ )
            {
                $Log->comment("Waiting for check job (id=$1) to finish ...");
                my $check_job = NACL::C::Job->find(
                    command_interface => $component->command_interface(),
                    filter            => { id => $1, }
                );

                # Wait for the metrocluster check job to go back to queued state
                my $check_job_timeout = time + 120;    # set 2 min timeout
                while ( lc $check_job->state()->state() ne 'queued' ) {
                    $Log->comment( 'Metrocluster check job is in progress. '
                          . 'Recheck in 5 secs...' );
                    Tharn::snooze(5);
                    if ( time > $check_job_timeout ) {
                        NACL::MTask::MCC::Exceptions::BaseException->throw(
                                'Timeout waiting for metrocluster check '
                              . 'job to finish!' );
                    }
                }
                goto RETRY_UNCONFIGURE;
            }
            else {
                NACL::MTask::MCC::Exceptions::BaseException->throw(
                    'Unable to get metrocluster check job ID');
            }
        }
        else {
            # Throw the exception if it is not known
            $ex->throw();
        }
    };

    if ( defined $job ) {
        $job->wait_on_job( %common_opts, 'method-timeout' => $timeout );
    }

    my %job_key;
    $job_key{name} = '"MetroCluster Unconfigure"';
    $job_key{'event-type'} = 'Succeeded';
    $Log->comment( "job key\n" . NACL::MTask::MCC::Utils->dump(%job_key) );

    #then optionally check for completion (check metrocluster show)
    $Log->comment("validate unconfigure completion for site: $site");
    my $job_show = NACL::CS::JobHistory->fetch(
        command_interface => $component->command_interface(),
        filter            => {%job_key},
    );
    $Log->comment( NACL::MTask::MCC::Utils->dump( target => $job_show ) );

    if ( $job_show->{'completion'} !~ /successful/ ) {
        NACL::MTask::MCC::Exceptions::Validation->throw(
            "unconfigure failed with error: $job_show->{'completion'}");
    }

    $Log->comment('Nodes to be unconfigured');
    $Log->comment( join ',', map { $_->get_name() } @nodes_to_be_unconfigured );

    #TODO finish cleaning up unconfig.... simplify and remove old stuff...
    foreach my $mccnode (@nodes_to_be_unconfigured) {
        my $node = $mccnode->get_Node_component();
        $Log->comment(
            'Fetching the configuration state for node ' . $node->node );
        my $op = NACL::MTask::MCC::Utils->metrocluster_node_show(
            command_interface => $node,
            requested_fields  => [ 'configuration-state', ],
        );
        $Log->comment("State is: $op->{'configuration_state'}");
        if ( $op->{'configuration_state'} !~ /Ready To Configure/i ) {
            if ( $force && ( scalar @nodes_to_be_unconfigured > 0 ) ) {
                $Log->comment('force option set');
                my $name = $node->node;
                $Log->comment("verifying node: $name");
                if ( defined $hash{$name} ) {
                    $Log->comment( 'WARN!! This node should '
                          . 'have been unconfigured  '
                          . 'which is not the case.' );

                    #last;
                }
            }
            else {
                $Log->comment( 'WARN!! All nodes need to be '
                      . ' unconfigured which is not the case' );
            }
        }
    }
    $Log->comment('Checking output for metrocluster node show');

    # we are using components now and background operation is not possible
    # anymore either so we will assume we have finished.
    my $finished = 1;

    #additional unconfigure checks to make sure unconfigure is operating
    #as expected
    #implicitly, this is only (potentially) run if we are waiting for completion

    if ($finished) {
        $Log->comment( 'unconfigure finished (job finished and metrocluster '
              . 'show empty.)  Running additional internal checks...' );
        my $extra_checks_fail = 0;

        my $cmd = 'set test;debug smdb table mccDRCOperation show';

        #should be: This table is currently empty.\r\n
        my $output =
          $component->command_interface()->apiset( interface => 'CLI' )
          ->execute_command( command => $cmd );

        #The table will be empty for the below cases:
        ## 1. unconfigure was done for both the local and remote cluster
        ## 2. unconfigure was done locally for the cluster we are checking the
        ##    command on.
        #The same applies for all the below commands.

        if ($output) {
            if ( $output !~ /table is currently empty/ ) {
                $extra_checks_fail++;
                $Log->warn("$cmd failure - table NOT empty as expected");
            }
            $Log->comment( NACL::MTask::MCC::Utils->dump( target => $output ) );
        }

        if ( !$force ) {

            #check to make sure peer unconfigured if not force/local only
            $output =
              $peer_component->command_interface()->apiset( interface => 'CLI' )
              ->execute_command( command => $cmd );
            if ( $output !~ /table is currently empty/ ) {
                $extra_checks_fail++;
                $Log->warn("$cmd failure - table empty which is not  expected");
            }
            $Log->comment( NACL::MTask::MCC::Utils->dump( target => $output ) );
        }

        $cmd = 'set test;vserver peer show';

        #should be: There are no Vserver peering relationships
        $output =
          $component->command_interface()->apiset( interface => 'CLI' )
          ->execute_command( command => $cmd );
        if ( $output !~ /There are no Vserver peer relationships/i ) {
            $extra_checks_fail++;
            $Log->warn("$cmd failure - table NOT empty as expected");
        }
        $Log->comment( NACL::MTask::MCC::Utils->dump( target => $output ) );

        if ( !$force ) {

            #check to make sure peer unconfigured if not force/local only
            $output =
              $peer_component->command_interface()->apiset( interface => 'CLI' )
              ->execute_command( command => $cmd );
            if ( $output !~ /There are no Vserver peer relationships/i ) {
                $extra_checks_fail++;
                $Log->warn("$cmd failure - table NOT empty as expected");
            }
            $Log->comment( NACL::MTask::MCC::Utils->dump( target => $output ) );
        }

        $cmd = 'set test;debug crs stream show';

        #should be: This table is currently empty.\r\n
        $output =
          $component->command_interface()->apiset( interface => 'CLI' )
          ->execute_command( command => $cmd );
        if ( $output !~ /table is currently empty/ ) {
            $extra_checks_fail++;
            $Log->warn("$cmd failure - table NOT empty as expected");
        }
        $Log->comment( NACL::MTask::MCC::Utils->dump( target => $output ) );

        if ( !$force ) {

            #check to make sure peer unconfigured if not force/local only
            $output =
              $peer_component->command_interface()->apiset( interface => 'CLI' )
              ->execute_command( command => $cmd );
            if ( $output !~ /table is currently empty/ ) {
                $extra_checks_fail++;
                $Log->warn("$cmd failure - table NOT empty as expected");
            }
            $Log->comment( NACL::MTask::MCC::Utils->dump( target => $output ) );
        }

        #If you created servers, 'vserver show' should have
        #deleted sync-destination servers (not sync-source)
        #
        $cmd = 'set test;vserver show -subtype sync-destination';

        #should be: There are no entries matching your query
        $output =
          $component->command_interface()->apiset( interface => 'CLI' )
          ->execute_command( command => $cmd );
        if ( $output !~ /are no entries matching your query/ ) {
            $extra_checks_fail++;
            $Log->warn("$cmd failure - entries NOT empty as expected");
        }
        $Log->comment( NACL::MTask::MCC::Utils->dump( target => $output ) );

        if ( !$force ) {

            #check to make sure peer unconfigured if not force/local only
            $output =
              $peer_component->command_interface()->apiset( interface => 'CLI' )
              ->execute_command( command => $cmd );
            if ( $output !~ /are no entries matching your query/ ) {
                $extra_checks_fail++;
                $Log->warn("$cmd failure - entries NOT empty as expected");
            }
            $Log->comment( NACL::MTask::MCC::Utils->dump( target => $output ) );
        }

        if ($extra_checks_fail) {
            if ($force) {
                $Log->warn( "WARN!! These failures could be incorrect since "
                      . "the force option is on. Please manually "
                      . "inspect the failure!!." );
            }
            else {
                NACL::MTask::MCC::Exceptions::Validation->throw(
"unconfigure appeared to complete BUT failed $extra_checks_fail"
                      . "  additional check(s)" );
            }
        }
    }

    #If we do a unconfigure then a configure, and then we run a metrocluster
    #operation show it could return an empty table (on the cluster that
    #configure wasn't run on)...  So in order to prevent it throwing an
    #exception, we enable the flag saying it's OK to ignore this until
    #it first finds a good operation state...it shouldn't ever be empty
    #unless it's just been configured
    $Log->comment("resetting dr op OK to be empty - unconfig");
    foreach my $site (qw(A B)) {
        my $cluster = $self->_metrocluster()->get_cluster( site => $site );
        $cluster->get_DR_Coordinator()->_dr_operation_table_ok_to_be_empty(1);
    }

    return;
}

=head2 validate_operation()

Validate metrocluster operation (show) for this cluster.  Validates operation
as follows:

  Anything in the Errors field is considered a failure.

  This method does NOT validate the operation and the state (That is
  done by DR_Coordinator::get_DR_state() method )

=over

=item Options

=item C<< disaster_status =E<gt> $mcc-E<gt>get_disaster_status() (optional) >>

(Optional) Pass in a previously generated disaster status structure if
you have one or a new one will be generated

=item C<< die_on_failure =E<gt> 0 | 1 (default = 1) >>

Whether to throw a NACL::MTask::MCC::Exceptions::Validation exception
on failure or to just warn

=back

=cut

sub validate_operation {
    my $trace = NACL::MTask::MCC::Log::TRACELOG->start_timer();    #metrics

    my $self = shift;

    my %opts = validate_with(
        params => \@_,
        spec   => {
            disaster_status => { type  => HASHREF,    optional => 1 },
            die_on_failure  => { regex => qr/^[01]$/, default  => 1, },
        },
    );

    my $disaster_status = $opts{disaster_status}
      || $self->_metrocluster()->get_disaster_status( include_HA => 1 );

    my $live_sites_ref =
        $opts{disaster_status}->{live_sites}
      ? $opts{disaster_status}->{live_sites}
      : $self->_metrocluster()->L_determine_live_sites();

    my $site = $self->get_site();
    my $this_site_alive = grep { $site eq $_ } @{$live_sites_ref};

    #do both sites, see what is up, or assume based on metrocluster operation?
    #for now, see what is up
    my $ci = $self->get_Metrocluster_component()->command_interface();

    #since we are switching to metrocluster operation history show, we need
    #a job id to get the relevant information. The job id is recorded
    #in DR_Coordinator.pm and is fetched here (in the below code). This
    #is then passed to the me op history showat line 3276
    my $dr_coordinator = $self->_DR_coordinator();
    my $job_info       = $dr_coordinator->_job_information();
    $Log->comment( 'job info ' . Dumper($job_info) );
    $dr_coordinator->_job_information(undef);
    my %cmd_opts;

    if ( defined $job_info ) {
        %cmd_opts = ( filter => { 'job-id' => $job_info->{job_id} } );
    }

    if ($this_site_alive) {
        ### Get all data/cluster operation
        #if job id is not passed then it will return the top most entry from
        #the me op history show output
        my $mcc_op_CS = NACL::CS::MetroclusterOperationHistory->fetch(
            command_interface => $ci,
            'method-timeout'  => 30,
            allow_empty       => 1,
            randomize_order   => 0,
            %cmd_opts
        );

        #$Log->comment(NACL::MTask::MCC::Utils->dump('met operation show:' .
        #Dumper $mcc_op_CS));

        $mcc_op_CS or return [];

        my $op_errors = [];

        $op_errors = $mcc_op_CS->error_list();

        # look at operation errors

        my @op_errors = grep { $_ && $_ ne q{-} } @{$op_errors};

        # Eliminate the results of 'metrocluster check run'.
        # Using a regex to be more resilient in the face of future
        # change.
        if (   $mcc_op_CS->operation() eq 'check'
            && $mcc_op_CS->state() eq 'successful' )
        {
            @op_errors = grep {
                $_ !~ m/^"?Use \s the \s
                                        "metrocluster \s check \s show" \s
                                        command \s to \s view \s the \s
                                        result \s of \s this \s operation\."?
                                       /xo
            } @op_errors;
        }

        if ( scalar @op_errors ) {
            my $op_err_str = join "\n", @op_errors;
            if ( $opts{die_on_failure} ) {
                NACL::MTask::MCC::Exceptions::Validation->throw(
                    "metrocluster operation show errors!:\n$op_err_str");
            }
            else {
                $Log->warn("metrocluster operation show errors!:\n$op_err_str");
            }
            return $op_errors;
        }
    }
    else {
        $Log->comment( "Site $site down, skipping metrocluster operation"
              . " show error validation" );
    }

    return;

}

=head2 validate_vservers()

Validate vservers for this cluster.  Validates vservers as follows:

  R=running, S=stopped, T=true, F=false
  admin-state/operational-state/config-lock
  normal (steady state): sync-source should be R/R/F, sync-dest should be R/S/T
  switchover: sync-* should be R/R/F
  heal aggrs: sync-* should be R/R/F
  heal roots: (survivor site): sync-* should be R/R/F
  heal roots: (disaster site): sync-* should be R/S/T


=over

=item Options

=item C<< disaster_status =E<gt> $mcc-E<gt>get_disaster_status() (optional) >>

(Optional) Pass in a previously generated disaster status structure if
you have one or a new one will be generated

=item C<< die_on_failure =E<gt> 0 | 1 (default = 1) >>

Whether to throw a NACL::MTask::MCC::Exceptions::Validation exception
on failure or to just warn

=back

=cut

sub validate_vservers {
    my $trace = NACL::MTask::MCC::Log::TRACELOG->start_timer();    #metrics

    my $self = shift;

    my %opts = validate_with(
        params => \@_,
        spec   => {
            disaster_status        => { type  => HASHREF,    optional => 1 },
            die_on_failure         => { regex => qr/^[01]$/, default  => 1, },
            initialization_timeout => { type  => SCALAR,     default  => 360 },
            initialization_sleep   => { type  => SCALAR,     default  => 10 }
        },
    );

    my $disaster_status = $opts{disaster_status}
      || $self->_metrocluster()->get_disaster_status( include_HA => 1 );

    #do both sites, see what is up, or assume based on metrocluster operation?
    #for now, self
    my $ci = $self->get_Metrocluster_component()->command_interface();

    ### Get all data/cluster Vservers
    my @vservers;

    my $timeout_start = time();

    # Retry vservers until they stop initializing
  RETRY: while (1) {

        @vservers = NACL::CS::Vserver->fetch(
            command_interface => $ci,
            filter            => {
                type    => 'data',
                subtype => 'sync-source|sync-destination'
            },
            allow_empty => 1
        );

        my @init = grep { $_->admin_state() eq 'initializing' } @vservers;

        if ( scalar @init == 0 ) {
            last RETRY;
        }
        else {
            my $time_elapsed = time - $timeout_start;
            my $time_left    = $opts{'initialization_timeout'} - $time_elapsed;
            if ( $time_left > 0 ) {
                $Log->comment( 'Vservers initializing; '
                      . "polling for $time_left remaining, "
                      . "snoozing for $opts{'initialization_sleep'}" );
                Tharn::snooze( $opts{'initialization_sleep'} );
                next RETRY;
            }
            else {
                $Log->warn("Vserver initialization timeout exceeded");
                last RETRY;
            }
        }

    }

    my %vserver_lock;
    try {
        my $output =
          $ci->apiset( interface => 'CLI' )
          ->debug_smdb_table_show( table => 'vserver_lock_db', );

        %vserver_lock =
          map { $_->{vserver} => $_->{owner} }
          @{ $output->get_parsed_output() };

    }
    catch NATE::BaseException with {

        #ignore if we can't get the table (may not work in non test builds)
        #so later on if we get a vserver lock error just warn - burt 815977
        $Log->comment( 'ignore problem getting vserver lock:' . shift );
    };
    $Log->comment( 'Validate all data sync-* vservers: ' . join q{,},
        ( map $_->{vserver}, @vservers ) );

    #$Log->comment( NACL::MTask::MCC::Utils->dump(\@vservers) );

    my $dr_state = $disaster_status->{DR_state};
    my $site     = $self->get_site();

    #check states.  R=running, S=stopped, T=true, F=false
    #admin-state/operational-state/config-lock
    #steady state: sync-source should be R/R/F, sync-dest should be R/S/T
    #SO: sync-* should be R/R/F
    #HA: sync-* should be R/R/F
    #HR (survivor site): sync-* should be R/R/F
    #HR (disaster site): sync-* should be R/S/T

    #some anonymous subroutines to validate various vserver states
    my $op_subtype_sub = sub {
        my $operational = shift || 'running';
        my $subtype     = shift || 'sync';

        #match subtype, but NOT this state is matched
        if (   $_->{'subtype'} =~ /$subtype/
            && $_->{'operational_state'} !~ /$operational/ )
        {
            #burt 847501, allow manually stopped vservers
            if (
                !(
                       $operational                             eq 'running'
                    && $_->{'operational_state'}                eq 'stopped'
                    && $_->{'operational_state_stopped_reason'} eq
                    'admin-state-stopped'
                )
              )
            {
                return {
                    'type'        => 'operational_state',
                    'vserver'     => $_->{vserver},
                    'vserver_obj' => $_,
                    'err'         => "site $site: "
                      . "operational state for $_->{subtype} "
                      . "vserver $_->{vserver} "
                      . "should be '$operational' but is "
                      . "'$_->{'operational_state'}'",
                };
            }
        }
    };
    my $admin_sub = sub {
        my $admin_state = shift || 'running';

        if ( $_->{'admin_state'} ne $admin_state ) {

            #burt 847501, allow manually stopped vservers
            if (
                !(
                       $admin_state eq 'running'
                    && $_->{'admin_state'} eq 'stopped'
                    && ( $_->{'operational_state_stopped_reason'} eq
                        'admin-state-stopped'
                        || $_->{'operational_state_stopped_reason'} eq
                        'sync-destination-and-switchover-not-done' )
                )
              )
            {
                return {
                    'type'        => 'admin_state',
                    'vserver'     => $_->{vserver},
                    'vserver_obj' => $_,
                    'err'         => "site $site: "
                      . "admin state for $_->{subtype} "
                      . "vserver $_->{vserver} "
                      . "should be '$admin_state' but is "
                      . "'$_->{'admin_state'}'",
                };
            }
        }
    };
    my $config_lock_sub = sub {
        my $config_lock = shift || 'false';
        my $subtype     = shift || 'sync';

        #match subtype, but NOT this state is matched
        if (
            (
                   $config_lock eq 'false'
                && $_->{'subtype'} =~ /$subtype/
                && $_->{'config_lock'} ne $config_lock
                &&

                #burt858249 ignore when owner eq vserver-dm
                #this will still complain if we can't
                #get vserver lock info.  if we want
                #to always ignore try adding
                #$vserver_lock{$_->{uuid}} eq 'vserverdm')
                $vserver_lock{ $_->{uuid} } ne 'vserver-dm'
            )
            ||

            (
                   $config_lock eq 'true'
                && $_->{'subtype'} =~ /$subtype/
                && $_->{'config_lock'} ne $config_lock
            )
          )
        {
            return {
                'type'        => 'config_lock',
                'vserver'     => $_->{vserver},
                'vserver_obj' => $_,
                'err'         => "site $site: "
                  . "config lock for $_->{subtype} "
                  . "vserver $_->{vserver} "
                  . "should be '$config_lock' but is "
                  . "'$_->{'config_lock'}'",
            };
        }
    };

    my $insert_err = sub {
        my $err_hashref = shift;
        my $type        = shift;

        if ( ref $_ eq 'HASH' ) {
            $type = $_->{type};
        }
        if ($_) {
            $err_hashref->{ $_->{vserver} }->{$type} = $_;
            return $_->{err};
        }
    };
    my $err = {};
    my @errs;
    if ( $dr_state eq 'normal' ) {

        #admin-state/operational-state/config-lock
        #steady state: sync-source should be R/R/F, sync-dest should be R/S/T

        push @errs, grep $_,
          map $insert_err->($err),
          map $op_subtype_sub->( 'running', 'sync-source' ), @vservers;
        push @errs, grep $_,
          map $insert_err->($err),
          map $op_subtype_sub->( 'stopped', 'sync-destination' ), @vservers;
        push @errs, grep $_,
          map $insert_err->($err),
          map $admin_sub->(), @vservers;
        push @errs, grep $_,
          map $insert_err->($err),
          map $config_lock_sub->( 'false', 'sync-source' ), @vservers;
        push @errs, grep $_,
          map $insert_err->($err),
          map $config_lock_sub->( 'true', 'sync-destination' ), @vservers;

    }
    elsif ( $dr_state =~ /switchover_completed|heal_aggrs_completed/ ) {
        push @errs, grep $_,
          map $insert_err->($err),
          map $op_subtype_sub->('running'), @vservers;
        push @errs, grep $_,
          map $insert_err->($err),
          map $admin_sub->(), @vservers;
        push @errs, grep $_,
          map $insert_err->($err),
          map $config_lock_sub->('false'), @vservers;
    }
    elsif ( $dr_state eq 'heal_roots_completed' ) {
        my $is_survivor_site = $site eq $disaster_status->{survivor_site};

        push @errs, grep $_,
          map $insert_err->($err),
          map $admin_sub->(), @vservers;

        my $operation_state;
        my $config_lock;

        if ($is_survivor_site) {
            $operation_state = 'running';
            $config_lock     = 'false';
        }
        else {
            $operation_state = 'stopped';
            $config_lock     = 'true';
        }

        push @errs, grep $_,
          map $insert_err->($err),
          map $op_subtype_sub->($operation_state), @vservers;
        push @errs, grep $_,
          map $insert_err->($err),
          map $config_lock_sub->( $config_lock, 'sync-source' ), @vservers;
    }
    else {
        $Log->warn( "dr state $dr_state not supported when validating vserver"
              . ' states' );
        return;

    }

    if ( scalar @errs ) {

        #$Log->comment("err hashref:" . Dumper $err);
        #$Log->comment("err strings:" . Dumper \@errs);
        my $errors = join "\n", @errs;

        if ( $opts{die_on_failure} ) {
            NACL::MTask::MCC::Exceptions::Validation->throw(
                "vserver validation errors (site $site, dr state $dr_state):"
                  . " \n$errors" );
        }
        else {
            $Log->warn(
                "vserver validation errors (site $site, dr state $dr_state):"
                  . " \n$errors" );
        }

        if (wantarray) {
            return ( $err, \@errs );
        }
        else {
            return $errors;
        }
    }
    else {
        $Log->comment(
            "vserver validation successful (dr state $dr_state): " . join q{,},
            ( map $_->{vserver}, @vservers )
        );
    }

    return;

}

=head2 validate_node()

Validate metrocluster node show for this cluster.  Validates node as follows:

  C=configured, U=unreachable, E=enabled
  configuration-state/dr-mirroring-state/dr-operation-state

  normal (steady state): (survivor or disaster site) survivor and disaster
              cluster should be C/E/normal
  switchover: survivor cluster should be C/E/switchover completed
  switchover: disaster cluster should be U/-/switched over ...
  heal aggrs: survivor cluster should be C/E/heal aggrs completed
  heal aggrs: disaster cluster should be U/-/switched over ...
  heal roots: (survivor site) survivor cluster should be C/E/heal roots
              completed
  heal roots: (survivor site) disaster cluster should be U/-/switched over ...
  heal roots: (survivor or disaster site) survivor cluster(booted) should be
              C/E/heal roots completed
  heal roots: (survivor or disaster site) disaster cluster(booted) should be
              C/E/waiting for switchback recovery

  rules:
  survivor cluster is C/E/dr state string
  disaster cluster is C/E/dr state string in steady state/normal
  disaster cluster is U/-/switched over ... in SO/HA/HR(unbooted)
  disaster cluster is C/E/waiting for switchback recovery in HR(booted)

=over

=item Options

=item C<< disaster_status =E<gt> $mcc-E<gt>get_disaster_status() (optional) >>

(Optional) Pass in a previously generated disaster status structure if
you have one or a new one will be generated

=item C<< die_on_failure =E<gt> 0 | 1 (default = 1) >>

Whether to throw a NACL::MTask::MCC::Exceptions::Validation exception
on failure or to just warn

=back

=cut

sub validate_node {
    my $trace = NACL::MTask::MCC::Log::TRACELOG->start_timer();    #metrics

    my $self = shift;

    my %opts = validate_with(
        params => \@_,
        spec   => {
            disaster_status => { type  => HASHREF,    optional => 1 },
            die_on_failure  => { regex => qr/^[01]$/, default  => 1, },
        },
    );

    my $disaster_status = $opts{disaster_status}
      || $self->_metrocluster()->get_disaster_status( include_HA => 1 );

    #do both sites, see what is up, or assume based on metrocluster operation?
    #for now, self
    my $ci = $self->get_Metrocluster_component()->command_interface();

    ### Get node show output
    my @node = NACL::MTask::MCC::Utils->metrocluster_node_show(
        command_interface => $ci,

        #requested_fields  => ['configuration-state',],
        allow_empty => 1
    );

    $Log->comment( 'Validate metrocluster node show output: ' . join q{,},
        ( map $_->{cluster}, @node ) );

    #$Log->comment( NACL::MTask::MCC::Utils->dump(\@node) );
    my $dr_state = $disaster_status->{DR_state};
    my $site     = $self->get_site();
    my $is_survivor_site;
    my $dr_type = '';
    my $survivor_name;
    my $disaster_name;
    my $disaster_site_booted;

    my $peer_name =
      ( grep { $self->get_name() eq $_->{cluster} } @node )[0]->{dr_cluster};
    if ( $disaster_status->{survivor_site} ) {
        $is_survivor_site = $site eq $disaster_status->{survivor_site};
        $dr_type       = $is_survivor_site ? 'survivor'        : 'disaster';
        $survivor_name = $is_survivor_site ? $self->get_name() : $peer_name;
        $disaster_name = $is_survivor_site ? $peer_name : $self->get_name();
        $disaster_site_booted =
          grep { $_ eq $disaster_status->{disaster_site} }
          @{ $disaster_status->{live_sites} };
    }

    #C=configured, U=unreachable, E=enabled
    #configuration-state/dr-mirroring-state/dr-operation-state
    #steady state: survivor and disaster cluster should be C/E/normal
    #SO: survivor cluster should be C/E/switchover completed
    #SO: disaster cluster should be U/-/switched over ...
    #HA: survivor cluster should be C/E/heal aggrs completed
    #HA: disaster cluster should be U/-/switched over ...
    #HR: (survivor site) survivor cluster should be C/E/heal roots completed
    #HR: (survivor site) disaster cluster should be U/-/switched over ...
    #HR: (survivor or disaster site) survivor cluster(booted) should be
    #    C/E/heal roots completed
    #HR: (survivor or disaster site) disaster cluster(booted) should be
    #    C/E/waiting for switchback recovery

    #configuration-state/dr-mirroring-state/dr-operation-state
    #boils down to:
    #survivor cluster is C/E/dr state string
    #disaster cluster is C/E/dr state string in steady state/normal
    #disaster cluster is U/-/switched over ... in SO/HA/HR(unbooted)
    #disaster cluster is C/E/waiting for switchback recovery in HR(booted)

    #some anonymous subroutines to validate various node states
    my $dr_operation_sub = sub {
        my $dr_operation_state = shift;
        my $cluster = shift || '.*';

        #match cluster, but NOT this state is matched
        if (   $_->{'cluster'} =~ /$cluster/
            && $_->{'dr_operation_state'} !~ /$dr_operation_state/ )
        {
            return {
                'type'     => 'dr_operation_state',
                'node'     => $_->{node},
                'node_obj' => $_,
                'err'      => "site $site: "
                  . "dr operation state for $dr_type "
                  . "node $_->{node} "
                  . "in cluster $cluster "
                  . "should be '$dr_operation_state' but " . 'is '
                  . "'$_->{'dr_operation_state'}'",
            };
        }
    };
    my $dr_mirroring_sub = sub {
        my $dr_mirroring_state = shift || 'enabled';
        my $cluster            = shift || '.*';

        #match cluster, but NOT this state is matched
        if (   $_->{'cluster'} =~ /$cluster/
            && $_->{'dr_mirroring_state'} ne $dr_mirroring_state )
        {
            return {
                'type'     => 'dr_mirroring_state',
                'node'     => $_->{node},
                'node_obj' => $_,
                'err'      => "site $site: "
                  . "dr mirroring state for $dr_type "
                  . "node $_->{node} "
                  . "in cluster $cluster "
                  . "should be '$dr_mirroring_state' but " . 'is '
                  . "'$_->{'dr_mirroring_state'}'",
            };
        }
    };
    my $config_state_sub = sub {
        my $config_state = shift || 'configured';
        my $cluster      = shift || '.*';

        #match cluster, but NOT this state is matched
        if (   $_->{'cluster'} =~ /$cluster/
            && $_->{'configuration_state'} ne $config_state )
        {
            return {
                'type'     => 'configuration_state',
                'node'     => $_->{node},
                'node_obj' => $_,
                'err'      => "site $site: "
                  . "config state for $dr_type "
                  . "node $_->{node} "
                  . "in cluster $cluster "
                  . "should be '$config_state' but is "
                  . "'$_->{'configuration_state'}'",
            };
        }
    };

    my $insert_err = sub {
        my $err_hashref = shift;
        my $type        = shift;

        if ( ref $_ eq 'HASH' ) {
            $type = $_->{type};
        }
        if ($_) {
            $err_hashref->{ $_->{node} }->{$type} = $_;
            return $_->{err};
        }
    };
    my $err = {};
    my @errs;
    ( my $dr_state_name = $dr_state ) =~ s/_/ /g;
    if ( $dr_state eq 'normal' ) {

        #configuration-state/dr-mirroring-state/dr-operation-state
        #survivor cluster is C/E/dr state string
        #disaster cluster is C/E/dr state string in steady state/normal

        push @errs, grep $_,
          map $insert_err->($err),
          map $config_state_sub->(), @node;
        push @errs, grep $_,
          map $insert_err->($err),
          map $dr_mirroring_sub->(), @node;
        push @errs, grep $_,
          map $insert_err->($err),
          map $dr_operation_sub->($dr_state_name), @node;

    }
    elsif (
        $dr_state =~ /switchover_completed|heal_aggrs_completed/
        ||

        #this state not technically supported in lib...
        ( $dr_state =~ /heal_roots_completed/ && !$disaster_site_booted )
      )
    {
        #configuration-state/dr-mirroring-state/dr-operation-state
        #survivor cluster is C/E/dr state string
        #disaster cluster is U/-/switched over ... in SO/HA/HR(unbooted)
        push @errs, grep $_,
          map $insert_err->($err),
          map $config_state_sub->( 'configured', $survivor_name ), @node;
        push @errs, grep $_,
          map $insert_err->($err),
          map $config_state_sub->( 'unreachable', $disaster_name ), @node;
        push @errs, grep $_,
          map $insert_err->($err),
          map $dr_mirroring_sub->( 'enabled', $survivor_name ), @node;
        push @errs, grep $_,
          map $insert_err->($err),
          map $dr_mirroring_sub->( '-', $disaster_name ), @node;
        push @errs, grep $_,
          map $insert_err->($err),
          map $dr_operation_sub->( $dr_state_name, $survivor_name ), @node;
        push @errs, grep $_,
          map $insert_err->($err),
          map $dr_operation_sub->( 'switched over', $disaster_name ), @node;
    }
    elsif ( $dr_state eq 'heal_roots_completed' ) {

        #configuration-state/dr-mirroring-state/dr-operation-state
        #survivor cluster is C/E/dr state string
        #disaster cluster is C/E/waiting for switchback recovery in HR(booted)

        push @errs, grep $_,
          map $insert_err->($err),
          map $config_state_sub->(), @node;
        push @errs, grep $_,
          map $insert_err->($err),
          map $dr_mirroring_sub->(), @node;
        push @errs, grep $_,
          map $insert_err->($err),
          map $dr_operation_sub->( $dr_state_name, $survivor_name ), @node;
        push @errs, grep $_, map $insert_err->($err), map $dr_operation_sub->(
            'waiting for switchback recovery',
            $disaster_name
        ), @node;

    }
    else {
        $Log->warn( "dr state $dr_state not supported when validating node"
              . ' states' );
        return;

    }

    if ( scalar @errs ) {

        #$Log->comment("err hashref:" . Dumper $err);
        #$Log->comment("err strings:" . Dumper \@errs);
        my $errors = join "\n", @errs;

        if ( $opts{die_on_failure} ) {
            NACL::MTask::MCC::Exceptions::Validation->throw(
                    "node validation errors (site $site, dr state $dr_state):"
                  . " \n$errors" );
        }
        else {
            $Log->warn(
                    "node validation errors (site $site, dr state $dr_state):"
                  . " \n$errors" );
        }

        if (wantarray) {
            return ( $err, \@errs );
        }
        else {
            return $errors;
        }
    }
    else {
        $Log->comment(
            "node validation successful (dr state $dr_state): " . join q{,},
            ( map $_->{node}, @node ) );
    }

    return;

}

=head2 L_get_cluster_mcc_nodes()

This library-only (L_) method finds all nodes in this cluster which are in MCC
relationships.  Returns a reference to an array of NACL::MTask::MCC::Node
objects.  This method should not find nodes in this cluster's DR partner
cluster, nor should it return nodes in this cluster which are not a part of any
MCC relationship.

=cut

sub L_get_cluster_mcc_nodes {
    my $trace = NACL::MTask::MCC::Log::TRACELOG->start_timer();    #metrics

    my $self = shift;

    #validate opts
    my %opts = validate_with(
        params => \@_,
        spec   => { dr_group_id => { type => SCALAR | UNDEF, optional => 1 }, },
        allow_extra => 1,
    );
    my $dr_group_id = $opts{dr_group_id};

    my @nodes     = ();
    my %dr_groups = %{ $self->_metrocluster()->get_DR_groups() };
    foreach my $dr_group ( keys %dr_groups ) {

        #skip if not one we want
        next if ( $dr_group_id && ( $dr_group != $dr_group_id ) );

        my @dr_partnerships = @{ $dr_groups{$dr_group}->get_DR_partnerships() };
        foreach my $partnership (@dr_partnerships) {
            push( @nodes,
                $partnership->get_partner_node( site => $self->_site() ) );
        }
    }

    return \@nodes;
}

# ATTO is 2-node, has bridges, but no switches
sub is_atto {
    my $self = shift;

    return ( $self->_metrocluster->is_2_node()
          && ( !$self->has_switches() )
          && $self->has_bridges() );
}

# Glowworm is 2-node and has neither switches nor bridges
sub is_glowworm {
    my $self = shift;

    return ( $self->_metrocluster->is_2_node()
          && ( !$self->has_switches() )
          && ( !$self->has_bridges() ) );
}

# FAB is 2-node, has bridges and switches
sub is_fab {
    my $self = shift;

    return ( $self->has_switches() && $self->has_bridges );

}

# This method is used to find what network interfaces on THIS cluster are
# associated with the intercluster addresses that the DR cluster uses to
# communicate to THIS cluster.  It returns an array of NACL::C::NetworkInterface
# objects.  (Usually it will only be one object, but is possible that
# multiple such interfaces could exist (I think.))
sub _find_DR_cluster_to_THIS_cluster_intercluster_interfaces {
    my $trace = NACL::MTask::MCC::Log::TRACELOG->start_timer();    #metrics

    my $self = shift;

    my $mcc_component = $self->get_Metrocluster_component();
    my $ci            = $mcc_component->command_interface();

    my @network_interfaces = ();

    my $this_cluster_name = $self->get_name();

    # be careful, this is confusing.  we are looking for the ClusterPeer
    # element on the DR site which has information about THIS site.  I.e.,
    # don't miss the get_peer() call in here!
    my $dr_site_ClusterPeer_component =
      $self->get_peer()
      ->_get_ClusterPeer_component( remote_cluster_name => $this_cluster_name );

    # these are the interface addresses on THIS cluster which we want
    my @IC_addresses_to_this_site_from_dr_site =
      $dr_site_ClusterPeer_component->get_one_state_attribute('peer-addrs');

    # find all the intercluster IFs on THIS cluster
    my @interface_components = NACL::C::NetworkInterface->find(
        command_interface => $ci,
        filter            => { role => 'intercluster' },
    );

    # if they match any of the DR cluster's remote addresses to THIS cluster
    foreach my $IC_IFs (@interface_components) {
        my $address = $IC_IFs->get_one_state_attribute('address');
        if ( any { $_ eq $address } @IC_addresses_to_this_site_from_dr_site ) {
            push( @network_interfaces, $IC_IFs );
        }
    }

    return @network_interfaces;
}

=head2 perform_switchback_execution_recovery()

    $cluster_A->perform_switchback_execution_recovery(
            # see BURT 559197
            timeout_for_slow_disk_SANOWN_updates           => 600,
            # probably obsolete use here
            do_peer_metrocluster_site_nodes_boot           => 0,
            do_intercluster_network_interface_enable       => 0,
            do_peer_metrocluster_command_interface_refresh => 0,
                                                            );

As of sn_warm chunk 5, most of this functionality has been replicated (and has
more valid usage) in prepare_for_switchback_execution().  The default values
for all options which were moved to prepare_for_switchback_execution()  in this
method are currently 0 (FALSE) at this point.

This method performs various actions to be performed after this site's
cluster executes a DR switchback command.

=over

=item Options

=item C<< timeout_for_slow_disk_SANOWN_updates => N (default = 600) >>

(Optional) [See BURT 559197] The library now actively checks to see if
disks complete the migration of their SANOWN states to the proper post-SB
states (normal).
This parameter sets the timeout.  If it takes longer than the timeout for
the states to migrate back to normal it will abort the wait and continue
anyway.

=item C<< do_intercluster_network_interface_enable => 0|1 (default = 0) >>

(Optional) [OBSOLETE usage here] This option will enable the intercluster
network interface on this cluster which allows this cluster to see its
Metrocluster peer cluster.

=item C<< do_peer_metrocluster_site_nodes_boot => 0|1 (default = 0) >>

(Optional) [OBSOLETE usage here] This option will attempt to boot all the nodes
in this cluster's Metrocluster peer cluster.  (I.e., if this called on cluster
A, it wiil try to boot all nodes in cluster B.)

=item C<< do_peer_metrocluster_command_interface_refresh => 0|1 (default = 0) >>

(Optional) [OBSOLETE usage here] This option will re-enable the command
interface to allow automation to access the disaster site CLI again.

=back

=cut

#TBD: Look later at this method and remove obsolete junk at some point.
sub perform_switchback_execution_recovery {
    my $trace = NACL::MTask::MCC::Log::TRACELOG->start_timer();    #metrics

    my $self     = shift;
    my %raw_opts = @_;

    # validate opts

    # need to know if this was explicitly passed or falling back on default
    my $timeout_for_slow_disk_SANOWN_updates_opt_was_passed =
      List::MoreUtils::any { $_ eq 'timeout_for_slow_disk_SANOWN_updates' } @_;

    my %opts = validate_with(
        params => \%raw_opts,
        spec   => {
            timeout_for_slow_disk_SANOWN_updates =>
              { regex => qr/^\d+$/, default => 6000, },
            do_peer_metrocluster_site_nodes_boot =>
              { regex => qr/^[01]$/, default => 0, },
            do_intercluster_network_interface_enable =>
              { regex => qr/^[01]$/, default => 0, },
            do_peer_metrocluster_command_interface_refresh =>
              { regex => qr/^[01]$/, default => 0, },
        },
    );
    unless (
        $self->_metrocluster()->workaround_expired(
            uichange => 'mcc_expire_workaround_640067',
            burt     => '640067'
        )
      )
    {
        $Log->comment('Part 3/3 of the BURT 645504 workaround.  BEGIN...');
        $Log->comment('Showing switched-back aggr status (re BURT 640067)...');
        my $until_recently_disaster_site_ci =
          $self->get_peer()->get_Metrocluster_component()->command_interface();
        $until_recently_disaster_site_ci->apiset( interface => 'CLI' )
          ->execute_command( command => 'disk show -v' );

        $Log->comment('Part 3/3 of the BURT 645504 workaround.  ...END.');
    }

    if ( $opts{do_peer_metrocluster_site_nodes_boot} ) {
        $Log->comment('Booting MCC associated nodes at disaster site...');
        $Log->warn( 'perform_switchback_execution_recovery('
              . 'do_peer_metrocluster_site_nodes_boot => 1) is probably an '
              . 'obsolete usage!' );
        my @subtest_handles = ();

        # transit objects could be empty in the case where we had to delete
        # them so we could externalize the object using Storable
        if ( !defined $self->_pre_switchover_halted_node_transit_objs() ) {

         #if we don't have the transit objects from before, try to make them now
            $Log->comment(
'Transit objects missing from _pre_switchover_halted_node_transit_objs(), attempting to recreate them'
            );
            $self->_pre_switchover_halted_node_transit_objs(
                $self->_create_peer_transit_objects() );
        }
        my $halted_node_transit_obj =
          $self->_pre_switchover_halted_node_transit_objs();
        foreach my $node_transit_obj ( @{$halted_node_transit_obj} ) {
            push @subtest_handles, subtest(
                sub {
                    $node_transit_obj->change_state(
                        to      => 'CLI',
                        timeout => 1500
                    );
                },
                '-bg',
                '-runid',
                $node_transit_obj->{conn}->{connhost}
                  . '_perform_switchback_execution_recovery_up'
            );
        }
        Subtest::wait_finish( subtests => \@subtest_handles );
        $self->_pre_switchover_halted_node_transit_objs(undef);
    }

    if ( $opts{do_intercluster_network_interface_enable} ) {
        $Log->comment('Enabling intercluster interfaces to partner site...');
        $Log->warn( 'perform_switchback_execution_recovery('
              . 'do_intercluster_network_interface_enable => 1) is probably an '
              . 'obsolete usage!' );
        my $interfaces_to_enable =
          $self->_pre_switchover_intercluster_network_interfaces_downed();

        foreach my $network_interface ( @{$interfaces_to_enable} ) {
            $network_interface->modify( 'status-admin' => 'up' );
        }
        $self->_pre_switchover_intercluster_network_interfaces_downed(undef);
    }

    if ( $opts{do_peer_metrocluster_command_interface_refresh} ) {
        $Log->comment('Refreshing disaster site command interface...');
        $Log->warn( 'perform_switchback_execution_recovery('
              . 'do_peer_metrocluster_command_interface_refresh => 1) is '
              . 'probably an obsolete usage!' );
        my $disaster_site = $self->get_peer()->_site();
        $self->_metrocluster()
          ->refresh_command_interfaces( site => $disaster_site );
    }

    $self->L_workaround_force_reserver_id_update();

    unless ( $self->_metrocluster()
        ->has_uichange( uichange => 'cluster-stream-down-on-switchover' ) )
    {
        $Log->comment( 'BURT 554679 + 662064 workaround.  BEGIN to bring up '
              . 'admin-state of crs stream' );

        #BURT 662064 mods:
        #
        #survivor RECEIVER stream up AFTER switcback
        #we are skipping disaster side stuff as dev _thinks_ it's OK.
        #put in if needed
        ##disaster SENDER stream up AFTER switchback

        my $survivor_site_ci =
          $self->get_Metrocluster_component()->command_interface();

        my $parsed_output =
          $survivor_site_ci->apiset( interface => 'CLI' )
          ->cluster_identity_show->get_parsed_output();

        my $cluster_uuid = $parsed_output->[0]->{'uuid'};

        $parsed_output =
          $survivor_site_ci->apiset( interface => 'CLI' )
          ->debug_crs_stream_show(
            'direction' => 'receiver',
            'synopsis'  => '"MetroCluster Cluster Domain Receiver Stream"',
          )->get_parsed_output();

        my $cmd =
            'debug smdb table mccStreamDBv2 modify -cluster-uuid '
          . $cluster_uuid
          . ' -stream-target-admin-state up ';

        foreach my $stream ( @{$parsed_output} ) {
            my $c_cmd = $cmd . '-stream-uuid ' . $stream->{'stream-uuid'};

            $survivor_site_ci->apiset( interface => 'CLI' )
              ->execute_command( command => $c_cmd );

            $survivor_site_ci->apiset( interface => 'CLI' )
              ->debug_crs_stream_modify_stream(
                'stream-uuid' => $stream->{'stream-uuid'},
                'admin-state' => 'up'
              );

        }

        # the stream may not come up immediately, retry for a while...
        my $end_time = time() + ( 2 * 60 );    # time allowed
        my $stream_done = 0;
        while ( !$stream_done && ( time() < $end_time ) ) {
            $parsed_output =
              $survivor_site_ci->apiset( interface => 'CLI' )
              ->debug_crs_stream_show(
                'direction' => 'receiver',
                'synopsis'  => '"MetroCluster Cluster Domain Receiver Stream"',
              )->get_parsed_output();

            my $check = 1;
            foreach my $stream ( @{$parsed_output} ) {
                if ( $stream->{'admin-state'} ne 'up' ) {
                    $check = 0;

                    # iff retry time ran out, throw an exception here
                    if ( time() > $end_time ) {
                        NACL::MTask::MCC::Exceptions::BaseException->throw(
                            'Failed to modify stream admin-state to up');
                    }
                }
            }
            $stream_done = $check;
            if ( !$stream_done ) {
                $Log->comment("sleeping 10 to retry stream up check...");
                Tharn::snooze(10);
            }
        }

        # shouldn't really need this except for corner cases maybe
        if ( !$stream_done ) {
            NACL::MTask::MCC::Exceptions::BaseException->throw(
                'Failed to modify stream admin-state to up');
        }

    }

    # see BURT 559197 / 541653

    my $dr_groups            = $self->_metrocluster()->get_DR_groups();
    my $sanown_timeout_left  = $opts{timeout_for_slow_disk_SANOWN_updates};
    my $sanown_timeout_start = time;
    my $sanown_timeout_end   = time + $sanown_timeout_left;

    #we don't want to wait more than timeout really (not timeout per call)
    $Log->comment("Verifying SANOWN disk state migrations...");

    foreach my $dr ( sort keys %{$dr_groups} ) {

        # timeout is total time, not each
        # L_wait_for_slow_SANOWN_state_migration logs warnings if needed.

        # if timeout_for_slow_disk_SANOWN_updates was not explicitly passed, and
        # this is a vsim system, we need to extend the timeout a lot.
        my $dr_group = $dr_groups->{$dr};
        my $timeout  = $sanown_timeout_end - time;
        if (   ( $dr_group->L_is_vsim_group() )
            && ( !$timeout_for_slow_disk_SANOWN_updates_opt_was_passed ) )
        {
            $timeout *= 10;
        }

        $dr_group->get_shared_storage()
          ->L_wait_for_slow_SANOWN_state_migration( timeout => $timeout );
    }

    return;
}

=head2 _is_remote_storage_reachable()

Checks remote storage connectivity from local cluster

=over

=item Options

=item C<< refresh_status => 0|1 (default = 0) >>

(Optional) Will force the library to query the latest status

=back

=cut

sub _is_remote_storage_reachable {
    my $trace = NACL::MTask::MCC::Log::TRACELOG->start_timer();    #metrics

    my $self = shift;
    my %opts = validate_with(
        params => \@_,
        spec   => { refresh_status => { type => BOOLEAN, default => 0 }, },
    );

    # fetch latest status if requested or unavailable
    if ( $opts{refresh_status} || !defined $self->_remote_storage_reachable() )
    {
        my @nodes = @{ $self->L_get_cluster_mcc_nodes() };
        my $node  = $nodes[0];

        # handle case where a node is unreachable due to SFO/HA
        if (
            scalar @nodes >= 2
            && $nodes[0]->get_Node_component->get_7m_or_nodescope_apiset(
                interface => 'CLI'
            )->api_is_session_alive()
          )
        {
            $node = $nodes[1];
        }

        my $response = $node->get_Node_component->get_7m_or_nodescope_apiset(
            interface => 'CLI' )->raid_config( name => 'check_links' );

        if (   $response->get_processed_output() =~ /ok\s+\(mirrored\)/
            && $response->get_processed_output() =~ /Remote storage links: Up/ )
        {
            $Log->comment('Remote storage is up and reachable');
            $self->_remote_storage_reachable(1);
        }
    }

    return $self->_remote_storage_reachable();
}

########## L_check_cluster_peer_connectivity ###################################
## Method to check cluster peer connectivity across two clusters
## The method polls for 6 mins before bailing out with an exception
##
## Input Params:
##   timeout => Timeout in seconds, default = 360 sec
##
#
sub L_check_cluster_peer_connectivity {
    my $trace = NACL::MTask::MCC::Log::TRACELOG->start_timer();    #metrics

    my $self = shift;
    my %opts = validate_with(
        params => \@_,
        spec   => { timeout => { type => SCALAR, default => 360 }, },
    );

    require NACL::CS::ClusterPeer;
    require NACL::CS::ClusterPeerHealth;
    require NACL::Exceptions::ClusterHealthCheckError;

    my @nodes   = @{ $self->L_get_cluster_mcc_nodes() };
    my $cpeer   = 0;
    my $timeout = time + $opts{timeout};

    # Check cluster peer connectivity from each cluster
    foreach my $mccnode (@nodes) {
        my $node = $mccnode->get_Node_component();
        do {
            $cpeer = NACL::CS::ClusterPeer->fetch( command_interface => $node );

            if ( lc $cpeer->availability() ne 'available'
                && time > $timeout )
            {
                my $msg =
                    'Remote Cluster is unreachable for more than 6 mins. '
                  . 'Please check all cluster peer connections from '
                  . $node->name();
                NACL::Exceptions::ClusterHealthCheckError->throw($msg);
            }
            elsif ( lc $cpeer->availability() ne 'available' ) {
                $Log->comment(
                        'Remote Cluster is unreachable. Retrying cluster '
                      . 'peer check in 10 secs...' );
                Tharn::snooze(10);
                NACL::CS::ClusterPeerHealth->fetch(
                    command_interface => $node,
                    allow_empty       => 1,
                    filter            => { 'bypass-cache' => 'true' }
                );
            }
            else {
                $Log->comment( 'Success!! Remote Cluster is now reachable from '
                      . $node->name() );
                next;
            }
          } while ( lc $cpeer->availability() ne 'available'
            && time < $timeout );
    }

    return;
}

1;
