# this is here due to our standard usage of @_ in Params::Validate methods.
##no critic (RequireArgUnpacking)
##no critic (ControlStructures::ProhibitUnlessBlocks)

# Copyright (c) 2011-2021 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.

## @summary MCC DR_Coordinator Module
## @author dl-mcc-dpg@netapp.com
## @status shared
## @pod here

package NACL::MTask::MCC::DR_Coordinator;

use strict;
use warnings;
use Carp;

use base qw (NACL::MTask::MTask);

use NACL::MTask::MCC::Log ();

use NACL::C::Metrocluster           ();
use NACL::CS::MetroclusterOperation ();
use NACL::CS::MetroclusterOperationHistory;
use NACL::STask::Vserver ();
use NACL::Exceptions::NoElementsFound qw(:try);
use NACL::APISet::Exceptions::JobFailedException qw(:try);

# for burt 667721
#use NACL::C::DebugVreport ();
use Tharn;
use NATE::Log qw(log_global);
my $Log       = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Params::Validate
  qw(validate validate_with BOOLEAN HASHREF OBJECT ARRAYREF SCALAR);
use Readonly;
use Data::Dumper;
use List::MoreUtils qw{ any };
use NACL::MTask::MCC::Exceptions::BaseException;
use NACL::MTask::MCC::Exceptions::State::CompletedWithManualRecovery;
use NACL::MTask::MCC::Exceptions::State::CompletedWithWarnings;
use NACL::MTask::MCC::Exceptions::State::OperationTimeOverrun;
use NACL::MTask::MCC::Utils;
use NACL::CS::Statistics;

use DateTime::Format::Strptime;

# the values are strings to attempt matching with the dr_state return
# value from 'metrocluster operation show'

Readonly my %DR_STATES => (
    normal            => { op => 'switchback', state => 'successful', },
    check             => { op => 'check',      state => 'successful', },
    check_in_progress => { op => 'check',      state => 'in[-_ ]progress', },
    check_failed      => { op => 'check',      state => 'failed', },
    switchover_in_progress =>
      { op => 'switchover', state => 'in[-_ ]progress', },
    switchover_completed => { op => 'switchover', state => 'successful', },
    switchover_failed    => { op => 'switchover', state => 'failed', },
    switchover_partially_successful =>
      { op => 'switchover', state => 'partially[-_ ]successful' },
    switchover_completed_with_manual_recovery_needed => {
        op    => 'switchover',
        state => 'completed[-_ ]with[-_ ]manual[-_ ]recovery[-_ ]needed'
    },
    switchover_completed_with_warnings =>
      { op => 'switchover', state => 'completed[-_ ]with[-_ ]warnings' },
    heal_aggrs_in_progress =>
      { op => 'heal[-_ ]aggregates', state => 'in[-_ ]progress', },
    heal_aggrs_completed =>
      { op => 'heal[-_ ]aggregates', state => 'successful', },
    heal_aggrs_failed => { op => 'heal[-_ ]aggregates', state => 'failed', },
    heal_aggrs_partially_successful =>
      { op => 'heal[-_ ]aggregates', state => 'partially[-_ ]successful' },
    heal_aggrs_completed_with_manual_recovery_needed => {
        op    => 'heal[-_ ]aggregates',
        state => 'completed[-_ ]with[-_ ]manual[-_ ]recovery[-_ ]needed'
    },
    heal_aggrs_completed_with_warnings => {
        op    => 'heal[-_ ]aggregates',
        state => 'completed[-_ ]with[-_ ]warnings'
    },
    heal_roots_in_progress =>
      { op => 'heal[-_ ]root[-_ ]aggregates', state => 'in[-_ ]progress', },
    heal_roots_completed =>
      { op => 'heal[-_ ]root[-_ ]aggregates', state => 'successful', },
    heal_roots_partially_successful => {
        op    => 'heal[-_ ]root[-_ ]aggregates',
        state => 'partially[-_ ]successful',
    },
    heal_roots_failed =>
      { op => 'heal[-_ ]root[-_ ]aggregates', state => 'failed', },

    #change above state to Failed after burt 680680 is fixed
    heal_roots_completed_with_manual_recovery_needed => {
        op    => 'heal[-_ ]root[-_ ]aggregates',
        state => 'completed[-_ ]with[-_ ]manual[-_ ]recovery[-_ ]needed'
    },
    heal_roots_completed_with_warnings => {
        op    => 'heal[-_ ]root[-_ ]aggregates',
        state => 'completed[-_ ]with[-_ ]warnings'
    },
    switchback_in_progress =>
      { op => 'switchback', state => 'in[-_ ]progress', },
    switchback_failed => { op => 'switchback', state => 'failed', },
    switchback_partially_successful =>
      { op => 'switchback', state => 'partially[-_ ]successful', },
    switchback_completed_with_manual_recovery_needed => {
        op    => 'switchback',
        state => 'completed[-_ ]with[-_ ]manual[-_ ]recovery[-_ ]needed'
    },
    switchback_completed_with_warnings =>
      { op => 'switchback', state => 'completed[-_ ]with[-_ ]warnings' },
    no_op => { op => 'META_no_op', state => 'META_no_state', },
    ignore__metrocluster_modify_unknown =>
      { op => 'modify', state => 'unknown', },
    ignore__metrocluster_modify => { op => 'modify', state => 'successful', },
    ignore__interface_create_successful =>
      { op => 'interface[-_ ]create', state => 'successful', },
    ignore__connect_successful => { op => 'connect', state => 'successful', },
    ignore__dr_group_create_successful =>
      { op => 'dr[-_ ]group[-_ ]create', state => 'successful', },
    ignore__metrocluster_configure_unknown =>
      { op => 'configure', state => 'unknown', },
    ignore__metrocluster_configure =>
      { op => 'configure', state => 'successful', },
    ignore__metrocluster_unconfigure =>
      { op => 'unconfigure', state => 'successful', },
    ignore__metrocluster_unconfigure_unknown =>
      { op => 'unconfigure', state => 'unknown', },
    ignore__switchback_continuation_agent =>
      { op => 'switchback[-_ ]continuation[-_ ]agent', state => 'successful', },
    ignore__switchback_continuation_agent_unknown =>
      { op => 'switchback[-_ ]continuation[-_ ]agent', state => 'unknown', },
    ignore__switchback_continuation_agent_warnings => {
        op    => 'switchback[-_ ]continuation[-_ ]agent',
        state => 'completed[-_ ]with[-_ ]warnings',
    },
    ignore__switchback_continuation_agent_manual_recovery => {
        op    => 'switchback[-_ ]continuation[-_ ]agent',
        state => 'completed[-_ ]with[-_ ]manual[-_ ]recovery[-_ ]needed',
    },
    ignore__switchback_continuation_agent_in_progress => {
        op    => 'switchback[-_ ]continuation[-_ ]agent',
        state => 'in[-_ ]progress'
    },
    ignore__unconfigure_continuation_agent => {
        op    => 'unconfigure[-_ ]continuation[-_ ]agent',
        state => 'successful',
    },
    ignore__unconfigure_continuation_agent_unknown =>
      { op => 'unconfigure[-_ ]continuation[-_ ]agent', state => 'unknown', },
    ignore__simulated_switchover =>
      { op => 'switchover[-_ ]simulate', state => 'successful|failed' },
    ignore__simulated_switchback =>
      { op => 'switchback[-_ ]simulate', state => 'successful|failed' },
);

Readonly my %DR_STATES__DEPRECATED => (
    normal                 => { op => 'Switchback', state => 'Successful', },
    switchover_in_progress => { op => 'Switchover', state => 'In Progress', },
    switchover_completed   => { op => 'Switchover', state => 'Successful', },
    switchover_failed      => { op => 'Switchover', state => 'Failed', },
    switchover_partially_successful =>
      { op => 'Switchover', state => 'Partially Successful' },
    heal_aggrs_in_progress =>
      { op => 'Heal Aggregates', state => 'In Progress', },
    heal_aggrs_completed => { op => 'Heal Aggregates', state => 'Successful', },
    heal_aggrs_failed    => { op => 'Heal Aggregates', state => 'Failed', },
    heal_aggrs_partially_successful =>
      { op => 'Heal Aggregates', state => 'Partially Successful' },
    heal_roots_in_progress =>
      { op => 'Heal Root Aggregates', state => 'In Progress', },
    heal_roots_completed =>
      { op => 'Heal Root Aggregates', state => 'Successful', },
    heal_roots_partially_successful =>
      { op => 'Heal Root Aggregates', state => 'Partially Successful', },
    heal_roots_failed => { op => 'Heal Root Aggregates', state => 'Unknown', },

    #change above state to Failed after burt 680680 is fixed
    switchback_in_progress => { op => 'Switchback', state => 'In Progress', },
    switchback_failed      => { op => 'Switchback', state => 'Failed', },
    switchback_partially_successful =>
      { op => 'Switchback', state => 'Partially Successful', },
    no_op => { op => 'META_no_op', state => 'META_no_state', },
    ignore__metrocluster_modify_unknown =>
      { op => 'Modify', state => 'Unknown', },
    ignore__metrocluster_modify => { op => 'Modify', state => 'Successful', },
    ignore__metrocluster_configure_unknown =>
      { op => 'Configure', state => 'Unknown', },
    ignore__metrocluster_configure =>
      { op => 'Configure', state => 'Successful', },
    ignore__metrocluster_unconfigure =>
      { op => 'Unconfigure', state => 'Successful', },
    ignore__metrocluster_unconfigure_unknown =>
      { op => 'Unconfigure', state => 'Unknown', },
    ignore__switchback_continuation_agent =>
      { op => 'Switchback Continuation Agent', state => 'Successful', },
    ignore__switchback_continuation_agent_unknown =>
      { op => 'Switchback Continuation Agent', state => 'Unknown', },
    ignore__switchback_continuation_agent_in_progress => {
        op    => 'Switchback Continuation Agent',
        state => 'In Progress'
    },
    ignore__unconfigure_continuation_agent =>
      { op => 'Unconfigure Continuation Agent', state => 'Successful', },
    ignore__unconfigure_continuation_agent_unknown =>
      { op => 'Unconfigure Continuation Agent', state => 'Unknown', },
    ignore__unconfigure_continuation_agent_in_progress => {
        op    => 'Unconfigure Continuation Agent',
        state => 'In Progress'
    },
);

=head1 NAME

NACL::MTask::MCC::DR_Coordinator

=head1 DESCRIPTION

This class is to be used for C-Mode Metrocluster testing.

    # Start test
    use NACL::MTask::MCC::Metrocluster;
    ...
    my $site_A_dr_coordinator =
                          $mcc->get_cluster(site => 'A')->get_DR_coordinator();
    ...
    # Perform test here
    ...
    # End test

=cut

=head1 Public Methods

There is no public constructor for this class.  A
NACL::MTask::MCC::DR_Coordinator object is obtained via calls such as

    my $site_A_dr_coordinator =
                          $mcc->get_cluster(site => 'A')->get_DR_Coordinator();

=cut

# Automatically generate the (non-public) constructor (includes a call to
# 'init') and also creates non-public members and non-public accesors for
# them.
use Class::MethodMaker [
    new => [ '-init', 'new' ],
    scalar =>
      [ { -type => 'NACL::MTask::MCC::Metrocluster' }, '_metrocluster', ],
    scalar => [

        # this is a flag that is true only when a test is first started and the
        # metrocluster operation table is seen to start out as empty.  this
        # SHOULD signify that no DR operation has ever occurred on this MCC
        # system since it was configured.
        qw/ _dr_operation_table_ok_to_be_empty /,
    ],
    scalar => [

        # this is a flag that indicates whether we should do a
        # statistics check after switchover and switchback
        qw/ _run_statistics /
    ],
    scalar => [
        qw/ _site
          _was_forced_on_disaster
          _stats_objects
          _job_information /
    ],

];

# validate the parameters passed in and what else?
sub init {
    my $trace = NACL::MTask::MCC::Log::TRACELOG->start_timer();    #metrics

    my $self = shift;
    my %opts = validate_with(
        params => \@_,
        spec   => {
            'metrocluster'   => { 'isa'   => 'NACL::MTask::MCC::Metrocluster' },
            'site'           => { 'regex' => qr/^[AB]$/ },
            'run_statistics' => { 'isa'   => BOOLEAN, 'default' => 1 },
            'stats_objects' => {
                'isa'     => ARRAYREF,
                'default' => [
                    qw/mcc_perf_cluster
                      mcc_perf_vserver
                      mcc_perf_node/
                ]
            },
        },
    );

    # set members per the passed constructor arguments.
    $self->_metrocluster( $opts{'metrocluster'} );
    $self->_site( $opts{'site'} );
    $self->_run_statistics( $opts{'run_statistics'} );
    $self->_stats_objects( $opts{'stats_objects'} );

    # when we start, we do not yet know if any DR op has ever occured on this
    # mcc system.  this flag is set to false after we have seen an actual
    # result after querying for a DR operation state.  once we have seen
    # an entry in this table, it should never be empty again until unconfig'd.
    $self->_dr_operation_table_ok_to_be_empty(1);

    #When we start there is no job information that we need to store.
    #hence setting the value to undef
    $self->_job_information(undef);

    return;
}

=head2 get_Metrocluster_component()

Utility to get the Metrocluster component associated with executing commands
for this cluster's DR Coordinator.  This is private because users probably
don't need to be getting the Metrocluster component from this class.

Returns a NACL::C::Metrocluster object.

=cut

sub get_Metrocluster_component {
    my $trace = NACL::MTask::MCC::Log::TRACELOG->start_timer();    #metrics

    my $self = shift;

    return $self->_metrocluster()
      ->get_Metrocluster_component( site => $self->_site() );
}

=head2 DR_switchover()

    $site_X_dr_coordinator->DR_switchover(timeout  => xxx;);

Execute a 'DR Switchover' command on the site X cluster command interface.

=over

=item Options:

=over

=item C<< expect-auso => 0|1 >>

Optional. Flag to indicate if MCC Lib should expect AUSO on 2-node MCC config
Defaults to true (1)

=item C<< timeout => xxx >>

Optional. See wait_for_DR_state() docs for default timeout value.

* Usage of this parameter is highly discouraged

=item C<< forced-on-disaster => 0|1 >>

Optional. Boolean. Defaults to false.

=item C<< force-nvfail-all => 0|1 >>

Optional. Boolean. Defaults to false.

=item C<< override-vetoes => 0|1 >>

Optional. Boolean. Defaults to false.

=item C<< retry-failed-nodes => \@nodes >>

Optional. List of names of failed nodes to retry.

=item C<< simulate => 0|1 >>

Optional. Boolean. Defaults to false.

=item C<< handle_SO_warnings_for_unm_aggr => 0|1 >>

Optional. Boolean. Defaults to true (1).

=back

=back



=cut

sub DR_switchover {
    my $trace = NACL::MTask::MCC::Log::TRACELOG->start_timer();    #metrics

    my $self = shift;
    my %opts = validate_with(
        params => \@_,
        spec   => {
            'timeout'            => { regex => qr/^\d+$/, optional => 1 },
            'state'              => { type  => SCALAR,    optional => 1 },
            'forced-on-disaster' => { type  => BOOLEAN,   default  => 1 },
            'force-nvfail-all'   => { type  => BOOLEAN,   default  => 0 },
            'override-vetoes'    => { type  => BOOLEAN,   default  => 0 },
            'retry-failed-nodes' => { type  => ARRAYREF,  optional => 1 },
            'reconnect_nodes'    => { type  => ARRAYREF,  optional => 1 },
            'wait_for_cluster_recovery' => { type => BOOLEAN, optional => 1 },
            'simulate'                  => { type => BOOLEAN, default  => 0 },
            'expect-auso'               => { type => BOOLEAN, default  => 1 },
            'handle_SO_warnings_for_unm_aggr' =>
              { type => BOOLEAN, default => 1 },

        },
    );

    # TODO 'state' option isn't documented
    my $mcc_component = $self->get_Metrocluster_component();
    my $ci            = $mcc_component->command_interface();
    $ci->refresh_command_interface();

    # get timing information
    my $time_start = time();

    my $default_state;
    if ( $opts{'simulate'} ) {
        $default_state = 'normal';
    }
    else {
        $default_state = 'switchover_completed';
    }
    my $state = $opts{state} || $default_state;
    my $timeout =
      exists( $opts{timeout} )
      ? $opts{timeout}
      : $self->_get_DR_state_timeout( state => $default_state );

    # for debugging
    $Log->comment( 'expect-auso value in SO is : '
          . NACL::MTask::MCC::Utils->dump( $opts{'expect-auso'} ) );

    if ( $opts{'expect-auso'} ) {
        $Log->comment( 'There is a possibility of A-USO happening on this MCC '
              . 'system. Waiting 40 seconds...' );

        Tharn::snooze(40);    # A-USO triggers automatically in 30 seconds max.
        if ( $self->get_DR_state() =~
            /switchover_in_progress|switchover_completed/ )
        {
            $Log->comment('The MCC system has successfully executed A-USO!');
            goto SKIP_SO_CMD_EXECUTION;
        }
        else {
            $Log->warn( 'MCC system did not execute A-USO. Proceeding with '
                  . 'manual MCC switchover operations...' );
        }
    }

    $Log->comment('Initiating metrocluster switchover...');

    #TODO just for testing this out.  We will generalize ui change strings
    #later, similar to workarounds

    #cdefs need to be updated for this to work TODO
    #$mcc_component->switchover(
    #'forced-on-disaster' => 'true',
    #'method-timeout' => $timeout,
    #);
    #for now we need to load our own cdefs
    #for APISet
    my $apiset = $ci->apiset();

    $apiset->set_timeout( 'connectrec-timeout' => $timeout );

    my %so_opts = (
          'forced-on-disaster' => $opts{'forced-on-disaster'}
        ? 'true'
        : 'false',
        'force-nvfail-all' => $opts{'force-nvfail-all'} ? 'true' : 'false',
        'override-vetoes'  => $opts{'override-vetoes'}  ? 'true' : 'false',
    );

    if ( defined $opts{'retry-failed-nodes'}
        && scalar @{ $opts{'retry-failed-nodes'} } )
    {
        $so_opts{'retry-failed-nodes'} =
          join( ' ', @{ $opts{'retry-failed-nodes'} } );
    }

    if ( $opts{'simulate'} ) {

        #simulate option cannot be used with any other SO option
        %so_opts = ( 'simulate' => 'true' );
    }

    my $run      = $apiset->metrocluster_switchover(%so_opts);
    my $op       = $run->get_parsed_output();
    my $data     = ${$op}[0];
    my %job_info = ( job_id => $data->{job_id} );

    $Log->comment( 'job info ' . Dumper( \%job_info ) );
    $self->_job_information( \%job_info );

  SKIP_SO_CMD_EXECUTION:

    # Pass along only the 'timeout' option to wait_for_DR_state
    my %wait_opts = (
        state => $state,
        (
            defined $opts{wait_for_cluster_recovery}
            ? ( wait_for_cluster_recovery => $opts{wait_for_cluster_recovery} )
            : ()
        ),
    );
    if ( exists $opts{'timeout'} ) {
        $wait_opts{'timeout'} = $opts{'timeout'};
    }
    if ( exists $opts{'reconnect_nodes'} ) {
        $wait_opts{'reconnect_nodes'} = $opts{'reconnect_nodes'};
    }
    if ( $opts{'handle_SO_warnings_for_unm_aggr'} ) {
        $wait_opts{'handle_SO_warnings_for_unm_aggr'} = 1;
    }

    $self->wait_for_DR_state(%wait_opts);

    my $execution_time = time() - $time_start;

    $self->_run_statistics_if_enabled();

    # Record if it was USO or NSO
    my $me_op_show =
      NACL::CS::MetroclusterOperation->fetch( command_interface => $ci );

    if ( lc $me_op_show->command_line() =~ /-forced-on-disaster true/ ) {
        $self->_was_forced_on_disaster(1);
    }
    else {
        $self->_was_forced_on_disaster(0);
    }

    return;
}

sub _run_statistics_if_enabled {
    my $self = shift;
    my (%opts) = @_;

    if ( $self->_run_statistics() ) {
        $self->validate_outage_window_latency(%opts);
    }

}

=head2 validate_outage_window_latency()

  $site_X_dr_coordinator->validate_outage_window_latency(
    cluster    => $cluster,
    timeout_ms => 120_000
  );

Checks the statistic 'outage_window_latency' against an expected
maximum time. Logs a comment if it's less, logs a warning if it's
more.

=over

=item Options:

=over

=item C<< cluster =E<gt> $cluster >>

Defaults to the Metrocluster component.

=item C<< timeout_ms =E<gt> $timeout_ms >>

If C<outage_window_latency> is larger than this value in milliseconds,
logs a warning. Defaults to 120,000 (120sec).

=back

=back

=cut

sub validate_outage_window_latency {
    my $trace = NACL::MTask::MCC::Log::TRACELOG->start_timer();    #metrics
    my $self  = shift;
    my %opts  = validate_with(
        params => \@_,
        spec   => {
            'cluster' => {
                type    => OBJECT,
                default => $self->get_Metrocluster_component()
            },
            'timeout_ms' => {
                type    => SCALAR,
                default => 120_000
            }
        }
    );

    my $state_instance;

    my $dr_state = $self->get_DR_state();

    if ( $dr_state eq 'normal' ) {
        $state_instance = 'switchback';
    }
    elsif ( $dr_state eq 'switchover_completed' ) {
        $state_instance = 'switchover';
    }
    else {
        $Log->comment(
            "validate_outage_window_latency skipped in state $dr_state");
        return;
    }

    my $ci = $opts{'cluster'}->command_interface();

    $ci->refresh_command_interface();

   #$ci->apiset(interface => 'CLI')->set_privilege('privilege-level' => 'test');

    my $stat = NACL::CS::Statistics->fetch(
        command_interface => $ci,
        filter            => {
            'object'   => 'mcc_perf_cluster',
            'raw'      => 'true',
            'counter'  => 'mcc_outage_window_latency',
            'instance' => $state_instance
        },
        'allow_empty' => 1,
    );

    if ( !defined $stat ) {
        $Log->comment('No mcc_outage_window_latency found');
    }
    else {
        my $stat_value = $stat->value();
        $stat_value =~ s/ms$//;

        if ( $stat_value > $opts{'timeout_ms'} ) {
            $Log->warn( "MCC stat 'mcc_outage_window_latency' '$stat_value' "
                  . "exceeded maximum $opts{'timeout_ms'} for $state_instance"
            );
        }
        else {
            $Log->comment( q{MCC stat 'mcc_outage_window_latency' was }
                  . "'$stat_value' for $state_instance" );
        }
    }

    $ci->apiset( interface => 'CLI' )->clear_privilege();

    return;

} ## end sub validate_outage_window_latency

=head2 DR_heal_aggrs()

    %args = {
                timeout  => xxx;
                interval => xxx;
            };

    $site_X_dr_coordinator->DR_heal_aggrs(%args);

Execute a 'DR heal -phase aggr' command on the site X cluster command interface

=over

=item Options:

=over

=item C<< timeout => xxx >>

Optional: See wait_for_DR_state() docs for default timeout value.

=item C<< interval => xxx >>

Optional: See wait_for_DR_state() docs for default timeout value.

=back

=back

=cut

sub DR_heal_aggrs {
    my $trace = NACL::MTask::MCC::Log::TRACELOG->start_timer();    #metrics

    my $self = shift;
    my %opts = validate_with(
        params => \@_,
        spec   => {
            timeout  => { regex => qr/^\d+$/, optional => 1 },
            'state'  => { type  => SCALAR,    optional => 1 },
            interval => { regex => qr/^\d+$/, optional => 1 },
            retry_attempts =>
              { regex => qr/^\d+$/, optional => 1, default => 1 },
            wait_for_cluster_recovery => { type => BOOLEAN, optional => 1 },
        },
    );
    my $retry_attempts = delete $opts{retry_attempts};
    my $default_state  = 'heal_aggrs_completed';
    my $state          = $opts{state} || $default_state;
    my $timeout =
      exists( $opts{timeout} )
      ? $opts{timeout}
      : $self->_get_DR_state_timeout( state => $default_state );

    my $mcc_component = $self->get_Metrocluster_component();
    my $ci            = $mcc_component->command_interface();
    $ci->refresh_command_interface();
    my $apiset = $mcc_component->command_interface()->apiset();

    $apiset->set_timeout( 'connectrec-timeout' => $timeout );
    $Log->comment('Initiating metrocluster heal -phase aggregates...');

    # get timing information
    my $time_start = time();
    my $run;

  RETRY_HEAL_AGGRS:

    $run = ();
    ## Commenting the try-catch block to avoid masking errors that occur
    ## during heal aggrs.
    try {
        $Log->warn('Heal Aggr retry mechanism in place');

        #my $return = $mcc_component->heal( phase            => 'aggregates',
        #                                   'method-timeout' => $timeout );

        $run = $apiset->metrocluster_heal( phase => 'aggregates', );
    }
    catch NACL::APISet::Exceptions::JobFailedException with {
        my $ex = shift;
        if ( $retry_attempts > 0 ) {
            $Log->warn('MCC heal aggregate failed! Retrying in 30 secs...');
            $retry_attempts--;
            $Log->warn("$retry_attempts retry attempts remaining...");
            Tharn::snooze(30);
            goto RETRY_HEAL_AGGRS;
        }
        else {
            $ex->throw();
        }
    };

#TODO look for unknown command in exception string before retry
#(or maybe an actual different exception related to invalid params)
#aggrs -> aggregates interface change.  Support deprecated UI for now...
#    if (! NACL::MTask::MCC::Utils->workaround_expired(
#                      expiration_date => '2013/05/06', burt => '630166')) {
#        $Log->comment('trying older UI, new interface likely not on this build...');
#        $mcc_component->heal(phase => 'aggrs',
#                             'method-timeout' => $timeout);
#    } else {
#rethrow exception, workaround has expired
#        my $exception_object = shift;
#        $exception_object->throw();
#    }
#};

    my $op       = $run->get_parsed_output();
    my $data     = ${$op}[0];
    my %job_info = ( job_id => $data->{job_id} );

    $Log->comment( 'job info ' . Dumper( \%job_info ) );
    $self->_job_information( \%job_info );

    $self->wait_for_DR_state(
        state => $state,
        %opts
    );

    my $execution_time = time() - $time_start;

    return;
}

=head2 DR_heal_root_aggrs()

    $site_X_dr_coordinator->DR_heal_root_aggrs(timeout  => xxx);

Execute a 'DR heal -phase root' command on the site X cluster command interface.

=over

=item Options:

=over

=item C<< timeout => xxx >>

=back

=back

Optional: See wait_for_DR_state() docs for default timeout value.

=cut

sub DR_heal_root_aggrs {
    my $trace = NACL::MTask::MCC::Log::TRACELOG->start_timer();    #metrics

    my $self = shift;
    my %opts = validate_with(
        params => \@_,
        spec   => {
            timeout                   => { regex => qr/^\d+$/, optional => 1 },
            'state'                   => { type  => SCALAR,    optional => 1 },
            retry_attempts            => { regex => qr/^\d+$/, default  => 1 },
            wait_for_cluster_recovery => { type  => BOOLEAN,   optional => 1 },
        },
    );

    my $retry_attempts = delete $opts{retry_attempts};
    my $mcc_component  = $self->get_Metrocluster_component();
    my $default_state  = 'heal_roots_completed';
    my $state          = $opts{state} || $default_state;
    my $timeout =
      exists( $opts{timeout} )
      ? $opts{timeout}
      : $self->_get_DR_state_timeout( state => $default_state );

    my $ci = $mcc_component->command_interface();
    $ci->refresh_command_interface();
    my $apiset = $mcc_component->command_interface()->apiset();

    $apiset->set_timeout( 'connectrec-timeout' => $timeout );
    $Log->comment('Initiating metrocluster heal -phase root-aggregates...');

    # get timing information
    my $time_start = time();
    my $run;

  RETRY_HEAL_ROOT_AGGRS:
    $run = ();
    ## Commenting the try-catch block to avoid masking errors that occur
    ## during heal root aggrs.
    try {
        $Log->warn('Heal Root Aggr retry mechanism in place');

       #my $return = $mcc_component->heal(phase            => 'root-aggregates',
       #                                  'method-timeout' => $timeout,
       #);

        $run = $apiset->metrocluster_heal( phase => 'root-aggregates', );

    }
    catch NACL::APISet::Exceptions::JobFailedException with {
        my $ex = shift;
        if ( $retry_attempts > 0 ) {
            $Log->warn('MCC heal root-aggr failed! Retrying in 30 secs...');
            $retry_attempts--;
            $Log->warn("$retry_attempts retry attempts remaining...");
            Tharn::snooze(30);
            goto RETRY_HEAL_ROOT_AGGRS;
        }
        else {
            $ex->throw();
        }
    };

#} catch NATE::BaseException with {
#TODO look for unknown command in exception string before retry
#(or maybe an actual different exception related to invalid params)
#aggrs -> aggregates interface change.  Support deprecated UI for now...
#     if (! NACL::MTask::MCC::Utils->workaround_expired(
#                       expiration_date => '2013/05/06', burt => '630166')) {
#         $Log->comment('trying older UI, new interface likely not on this build...');
#         $mcc_component->heal(phase => 'roots',
#                              'method-timeout' => $timeout);
#     } else {
#         #rethrow exception, workaround has expired
#         my $exception_object = shift;
#         $exception_object->throw();
#     }
# };

    my $op       = $run->get_parsed_output();
    my $data     = ${$op}[0];
    my %job_info = ( job_id => $data->{job_id} );

    $Log->comment( 'job info ' . Dumper( \%job_info ) );
    $self->_job_information( \%job_info );
    $self->wait_for_DR_state(
        state => $state,
        %opts
    );

    my $execution_time = time() - $time_start;

    # make sure disk states get updated after heal roots
    $self->_metrocluster()->get_cluster( site => 'A' )
      ->L_workaround_force_reserver_id_update();

    return;
}

=head2 DR_switchback()

    $site_X_dr_coordinator->DR_switchback(timeout  => xxx);

Execute a 'DR Switchback' command on the site X cluster command interface.

=over

=item Options:

=over

=item C<< override-cluster-stream-resync-errors => 0|1 (optional)>>

Optional. Boolean. Defaults to false.

=item C<< override-lost-config-updates-veto => 0|1 (optional)>>

Optional. Boolean. Defaults to false.

=item C<< override-vserver-stream-resync-errors => 0|1 (optional)>>

Optional. Boolean. Defaults to false.

=item C<< override-vetoes => 0|1 >>

Optional. Boolean. Defaults to false.

=item C<< simulate => 0|1 >>

Optional. Boolean. Defaults to false.

=item C<< timeout => xxx >>

=back

=back

Optional: See wait_for_DR_state() docs for default timeout value.

* Usage of this parameter is highly discouraged

=cut

sub DR_switchback {
    my $trace = NACL::MTask::MCC::Log::TRACELOG->start_timer();    #metrics

    my $self = shift;
    my %opts = validate_with(
        params => \@_,
        spec   => {
            timeout           => { regex => qr/^\d+$/, optional => 1 },
            'state'           => { type  => SCALAR,    optional => 1 },
            'override-vetoes' => { type  => BOOLEAN,   default  => 0 },
            'override-lost-config-updates-veto' =>
              { type => BOOLEAN, optional => 1 },
            'override-cluster-stream-resync-errors' =>
              { type => BOOLEAN, optional => 1 },
            'override-vserver-stream-resync-errors' =>
              { type => BOOLEAN, optional => 1 },
            wait_for_cluster_recovery => { type => BOOLEAN, optional => 1 },
            'simulate'                => {
                type    => SCALAR,
                regex   => qr/^[01]$/,
                default => 0
            },

        },
    );

    my $mcc_component = $self->get_Metrocluster_component();
    $Log->comment('Initiating metrocluster switchback...');

    # get timing information
    my $time_start = time();

    my $default_state;
    if ( $opts{'simulate'} ) {
        $default_state = 'heal_roots_completed';
    }
    else {
        $default_state = 'normal';
    }
    my $state = $opts{state} || $default_state;
    my $timeout =
      exists( $opts{timeout} )
      ? $opts{timeout}
      : $self->_get_DR_state_timeout( state => $default_state );

    my $ci = $mcc_component->command_interface();
    $ci->refresh_command_interface();
    my $apiset = $mcc_component->command_interface()->apiset();

    $apiset->set_timeout( 'connectrec-timeout' => $timeout );
    my %sb_opts;

    #we need 'true' here, not 1
    $sb_opts{'override-vetoes'} = 'true' if ( delete $opts{'override-vetoes'} );

    my $override_removed = $self->_metrocluster()->workaround_expired(
        uichange   => 'override-lost-config-updates-veto-removed',
        burt       => 818555,
        workaround => 0
    );
    unless ($override_removed) {
        $sb_opts{'override-lost-config-updates-veto'} = 'true'
          if ( delete $opts{'override-lost-config-updates-veto'} );
    }
    else {
        if ( $opts{'override-lost-config-updates-veto'} ) {
            $Log->warn(
                    'The switchback param "override-lost-config-updates-veto"'
                  . 'has been removed, please remove this param from your script'
            );
        }
    }
    $sb_opts{'override-cluster-stream-resync-errors'} = 'true'
      if ( delete $opts{'override-cluster-stream-resync-errors'} );
    $sb_opts{'override-vserver-stream-resync-errors'} = 'true'
      if ( delete $opts{'override-vserver-stream-resync-errors'} );

    if ( $opts{'simulate'} ) {
        $sb_opts{'simulate'} = 'true';
    }

    #my $return = $mcc_component->switchback(%sb_opts);
    my $run;
    try {
        $run = $apiset->metrocluster_switchback(%sb_opts);
    }
    catch NACL::APISet::Exceptions::JobFailedException with {
        my $ex  = shift;                                   # store the exception
        my $out = $apiset->metrocluster_operation_show();
        my $err = $out->get_parsed_output()->[0]->{'error-list'}->[1];
        if ( defined $err and $err =~ /There are failed disks present/ ) {
            $Log->warn( 'Retrying SB due to bad disks. '
                  . 'But file engsup ticket to get them replaced asap!' );
            $sb_opts{'override-vetoes'} = 'true';
            $run = $apiset->metrocluster_switchback(%sb_opts);

            # Set expected state as SB with manual recovery
            $state = 'switchback_completed_with_manual_recovery_needed';
        }
        else {
            $ex->throw();
        }
    };

    my $op       = $run->get_parsed_output();
    my $data     = ${$op}[0];
    my %job_info = ( job_id => $data->{job_id} );

    $Log->comment( 'job info ' . Dumper( \%job_info ) );
    $self->_job_information( \%job_info );

    #note switchback changes the state to normal
    $self->wait_for_DR_state(
        state => $state,
        defined $opts{timeout} ? ( timeout => $opts{timeout} ) : (),
        (
            defined $opts{wait_for_cluster_recovery}
            ? ( wait_for_cluster_recovery => $opts{wait_for_cluster_recovery} )
            : ()
        ),
    );

    my $execution_time = time() - $time_start;

    # make sure disk states get updated after switchback
    $self->_metrocluster()->get_cluster( site => 'A' )
      ->L_workaround_force_reserver_id_update();

    # Run statistics
    $self->_run_statistics_if_enabled();

    return;
}

=head2 get_DR_state()

    my $state = $site_X_dr_coordinator->get_DR_state();

    my ($state, $raw_state) =  $site_X_dr_coordinator->get_DR_state();


Returns one of the following strings, identifying the state of the DR operation
in the DR lifecycle:

    normal
    switchover_in_progress
    switchover_completed
    heal_aggrs_in_progress
    heal_aggrs_completed
    heal_roots_in_progress
    heal_roots_completed
    switchback_in_progress

If called in an array context, an array with the normal state and the
raw original state values are returned.  This allows access to the
standard equivalent state as well as the raw state the filer returns.

For example, the raw state of configure successful is functionally
equivalent to 'normal.'  'switchover_failed' is also equivalent to 'normal.'

In case the user wants to know the state for any reason they can get
it via the array context.

=over

=item Options:

=over

=item C<< skip_mcc_check_state => 0|1 >>

Optional: If set to 0, the method will ignore all 'check' jobs and return the first
          DR job (SO/Heal aggrs/Heal roots/SB) entry.
Default : 0

=back

=back

=cut

sub get_DR_state {
    my $trace = NACL::MTask::MCC::Log::TRACELOG->start_timer();    #metrics

    my $self = shift;

    my %opts = validate_with(
        params => \@_,
        spec   => {
            skip_mcc_check_state =>
              { type => SCALAR, regex => qr/^[01]$/, default => 1 },
        },
    );

    my $skip_check_state = $opts{skip_mcc_check_state};
    my $site_ci = $self->get_Metrocluster_component()->command_interface();
    $site_ci->refresh_command_interface()
      if !$site_ci->apiset()->get_connection()->is_responsive();

    # get timing information
    my $time_start = time();

    my $operation = undef;
    my $state     = undef;
    my $op_errors = undef;
    my $mcc_op_CS = undef;

    try {
        my @jobstates = NACL::CS::MetroclusterOperationHistory->fetch(
            command_interface => $site_ci,
            randomize_order   => 0,
        );

        my $latest_entry = shift @jobstates;

        $Log->comment(
            'Latest DR operation is : ' . $latest_entry->operation() );

        if ( $latest_entry->operation() =~ /check/ && $skip_check_state ) {
            $mcc_op_CS = $self->_get_the_first_non_check_state(
                job_history => \@jobstates );

        }
        elsif (( $latest_entry->operation() =~ /check/ && !$skip_check_state )
            || ( $latest_entry->operation() !~ /check/ ) )
        {
            $mcc_op_CS = $latest_entry;
        }

        if ( !defined($mcc_op_CS) ) {
            $operation = 'META_no_op';
            $state     = 'META_no_state';
        }
        else {
            $operation = $mcc_op_CS->operation();
            $state     = $mcc_op_CS->state();
            $op_errors = $mcc_op_CS->error_list();
        }

    }
    catch NACL::Exceptions::NoElementsFound with {

        # this is expected if no DR op has yet been performed on this setup.
        # but otherwise, it is unexpected.
        if ( $self->_dr_operation_table_ok_to_be_empty() ) {
            $operation = 'META_no_op';
            $state     = 'META_no_state';
        }
        else {
            NACL::MTask::MCC::Exceptions::BaseException->throw(
                    "ONTAP ERROR: metrocluster operation table empty after\n"
                  . "             previous DR operations have occured." );
        }
    }
    catch NACL::Exceptions::InvalidMethod with {

        # allow this to pass through to the workaround below.
        $operation = undef;    # (just reiterating this here.)
        $state     = undef;    # (just reiterating this here.)
    };

    $Log->comment( ' Value returned ~ '
          . ' operation : '
          . $operation . "\n"
          . ' state     : '
          . $state . "\n"
          . ' errors    : '
          . Dumper($op_errors)
          . "\n" );

    my $execution_time = time() - $time_start;

    my $found_state_string = undef;
    my $original_state_string;
    if ( defined($operation) && defined($state) ) {

        # strip any quotes
        $operation =~ s/^['"]//;
        $operation =~ s/['"]$//;
        $state     =~ s/^['"]//;
        $state     =~ s/['"]$//;

        foreach my $state_string ( keys(%DR_STATES) ) {
            if (   ( $operation =~ m/^$DR_STATES{$state_string}->{op}$/i )
                && ( $state =~ m/^$DR_STATES{$state_string}->{state}$/i ) )
            {
                $found_state_string = $state_string;
                last;
            }
        }

        # fall back to deprecated states

        if ( !defined $found_state_string ) {

            $operation =~ s/^['"]//;
            $operation =~ s/['"]$//;
            $state     =~ s/^['"]//;
            $state     =~ s/['"]$//;

            foreach my $state_string ( keys(%DR_STATES__DEPRECATED) ) {
                if (
                    (
                        $operation =~
                        m/^$DR_STATES__DEPRECATED{$state_string}->{op}$/i
                    )
                    && ( $state =~
                        m/^$DR_STATES__DEPRECATED{$state_string}->{state}$/i )
                  )
                {
                    $found_state_string = $state_string;
                    $Log->comment(
"WARNING! DR Operation/State [$operation/$state] is DEPRECATED!..."
                    );
                    last;
                }
            }

        }

        unless ( defined($found_state_string) ) {
            NACL::MTask::MCC::Exceptions::BaseException->throw(
                "DR Operation/State [$operation/$state] not expected.");
        }

        $original_state_string = $found_state_string;
        if ( $found_state_string eq 'no_op' ) {

            # no_op means no previous DR op has occurred on this setup.  it
            # must be in normal state
            $found_state_string = 'normal';
        }
        elsif (
            $found_state_string eq 'ignore__metrocluster_unconfigure_unknown' )
        {
            $found_state_string = 'normal';
        }
        elsif ( $found_state_string eq 'ignore__metrocluster_unconfigure' ) {
            $found_state_string = 'normal';
        }
        elsif (
            $found_state_string eq 'ignore__metrocluster_configure_unknown' )
        {
            $found_state_string = 'normal';
        }
        elsif ( $found_state_string eq 'ignore__metrocluster_configure' ) {
            $found_state_string = 'normal';
        }
        elsif ( $found_state_string eq 'ignore__interface_create_successful' ) {
            $found_state_string = 'normal';
        }
        elsif ( $found_state_string eq 'ignore__connect_successful' ) {
            $found_state_string = 'normal';
        }
        elsif ( $found_state_string eq 'ignore__dr_group_create_successful' ) {
            $found_state_string = 'normal';
        }
        elsif ( $found_state_string eq 'ignore__metrocluster_modify' ) {
            $found_state_string = 'normal';
        }
        elsif ( $found_state_string eq 'ignore__metrocluster_modify_unknown' ) {
            $found_state_string = 'normal';
        }
        elsif ( $found_state_string eq 'ignore__switchback_continuation_agent' )
        {
            $found_state_string = 'normal';
        }
        elsif ( $found_state_string eq
            'ignore__switchback_continuation_agent_unknown' )
        {
            $found_state_string = 'normal';
        }
        elsif ( $found_state_string eq
            'ignore__switchback_continuation_agent_in_progress' )
        {
            $found_state_string = 'switchback_in_progress';
        }
        elsif ( $found_state_string eq
            'ignore__switchback_continuation_agent_warnings' )
        {
            $found_state_string = 'normal';
        }
        elsif ( $found_state_string eq
            'ignore__switchback_continuation_agent_manual_recovery' )
        {
            $found_state_string = 'normal';
        }
        elsif (
            $found_state_string eq 'ignore__unconfigure_continuation_agent' )
        {
            $found_state_string = 'normal';
        }
        elsif ( $found_state_string eq 'switchover_failed' ) {
            $found_state_string = 'normal';
        }
        elsif ( $found_state_string eq 'switchback_failed' ) {
            $found_state_string = 'heal_roots_completed';
        }
        elsif ( $found_state_string eq
            'ignore__unconfigure_continuation_agent_unknown' )
        {
            $found_state_string = 'normal';
        }
        elsif ( $found_state_string eq 'ignore__simulated_switchover' ) {
            $found_state_string = 'normal';
        }
        elsif ( $found_state_string eq 'ignore__simulated_switchback' ) {
            $found_state_string = 'heal_roots_completed';
        }
        elsif ( $found_state_string eq 'switchback_completed_with_warnings' ) {

            #for now any "completed_with_warnings" state is equivalent to
            #success as we always look at the failure field anyway
            $found_state_string = 'normal';
        }
        elsif ( $found_state_string =~ /_completed_with_warnings$/ ) {

            #for now any "completed_with_warnings" state is equivalent to
            #success as we always look at the failure field anyway
            $found_state_string =~ s/_with_warnings$//;
        }
        elsif ( $found_state_string eq
            'switchback_completed_with_manual_recovery_needed' )
        {
            #for now any "completed_with_manual_recovery_needed" state is
            #equivalent to success as far as our state machine goes.
            #We will look for the 'real' value explicitly and decide
            #what to do.
            $found_state_string = 'normal';
        }
        elsif (
            $found_state_string =~ /completed_with_manual_recovery_needed$/ )
        {
            #for now any "completed_with_manual_recovery_needed" state is
            #equivalent to success as far as our state machine goes.
            #We will look for the 'real' value explicitly and decide
            #what to do.
            $found_state_string =~ s/_with_manual_recovery_needed$//;
        }
        elsif ($found_state_string eq 'heal_roots_in_progress_temp'
            || $found_state_string eq 'heal_roots_completed_temp'
            || $found_state_string eq 'heal_aggrs_in_progress_temp'
            || $found_state_string eq 'heal_aggrs_completed_temp' )
        {
      #this all needs to be upgraded, we need to support this for the short term
      #burt 655889
            $found_state_string =~ s/_temp$//;
        }
        else {
            # a previous op has been seen, should never see an empty table again
            # _FOR THIS SITE'S_ DR Coordinator
            $self->_dr_operation_table_ok_to_be_empty(0);
        }
    }

    # look at operation errors

    my @op_errors = grep { $_ && $_ ne q{-} } @{$op_errors};

    my $die_on_op_error =
      defined $ENV{DIE_ON_MCC_OPERATION_ERROR}
      ? $ENV{DIE_ON_MCC_OPERATION_ERROR}
      : 0;

    if ( scalar @op_errors ) {
        my $op_err_str = join "\n", @op_errors;
        if ($die_on_op_error) {
            NACL::MTask::MCC::Exceptions::BaseException->throw(
                "metrocluster operation show errors!:\n$op_err_str");
        }
        else {
            $Log->warn("metrocluster operation show errors!:\n$op_err_str");
        }
    }

    # For SO success/warn states record if it was USO or NSO
    if ( $found_state_string =~ /switchover_completed/ ) {
        $Log->comment( 'Found a "switchover_completed" DR state. '
              . 'Recording if it was USO or NSO' );
        my $me_op_show =
          NACL::CS::MetroclusterOperation->fetch(
            command_interface => $site_ci );

        if ( lc $me_op_show->command_line() =~ /-forced-on-disaster true/ ) {
            $Log->comment('Found a USO operation');
            $self->_was_forced_on_disaster(1);
        }
        else {
            $Log->comment('Found a NSO operation');
            $self->_was_forced_on_disaster(0);
        }
    }

    wantarray
      ? return ( $found_state_string, $original_state_string, $mcc_op_CS )
      : return $found_state_string;
}

# Note: Do not call this method directly. Always call get_DR_state()
# This method will return the first entry wherein the DR operation is not check
sub _get_the_first_non_check_state {

    my $self = shift;

    my %opts = validate_with(
        params => \@_,
        spec   => { job_history => { type => ARRAYREF } }
    );

    my $jobs                   = $opts{job_history};
    my $possible_current_state = undef;

    foreach my $job ( @{$jobs} ) {
        if ( $job->operation() eq 'check' ) {
            next;
        }
        else {
            $Log->comment( 'Found the job: '
                  . $job->operation()
                  . ' that we were looking for' );

            $possible_current_state = $job;
            last;
        }
    }

#The below code handles the scenario where the only entry in the history table is
# check. So the output will look something like:

#sti16-vsim-ucs414e4f_siteB::> metrocluster operation history show
# Operation                     State          Start Time          End Time
#----------------------------- -------------- ----------------   ----------------
# check                         successful     6/2/2015 13:12:06 6/2/2015 13:12:39
#1 entries were displayed.

#Given the above ouput and when we are attempting to ignore any operation that is check,
#the library still needs to derive the state of the system. The above output is
#seen is a newly configured system. If that is the case the library should ignore it and
#return normal. If not it should croak.

#In this case for a newly configured system, this method returns undef and will be
#handled in get_DR_state() where the mcc state will be returned as NORMAL
    if ( !defined($possible_current_state)
        && $self->_dr_operation_table_ok_to_be_empty() )
    {
        $Log->comment('Probably a newly configured MCC system');
    }
    elsif (!defined($possible_current_state)
        && !$self->_dr_operation_table_ok_to_be_empty() )
    {
        NACL::MTask::MCC::Exceptions::BaseException->throw(
            "ONTAP ERROR: metrocluster operation histroy table empty after \n"
              . "             previous DR operations have occured." );
    }

    return $possible_current_state;
}

=head2 _get_DR_state_timeout()

    my $timeout = $site_X_dr_coordinator->_get_DR_state_timeout(state => 'switchover_completed');

returns the default timeout for the specified dr operation:

                            switchover_completed => 300,
                            heal_aggrs_completed => 600,
                            heal_roots_completed => 600,
                            normal               => 120,

=cut

sub _get_DR_state_timeout {
    my $trace = NACL::MTask::MCC::Log::TRACELOG->start_timer();    #metrics

    my $self = shift;

    # constants
    my %DEFAULT_TIMEOUT = (
        switchover_completed => 600,
        heal_aggrs_completed => 600,
        heal_roots_completed => 600,
        normal               => 1200,
    );

    # more stringent specification timeouts
    my %SPECIFICATION_TIMEOUT = (
        switchover_completed => 300,    # 5 min
        heal_aggrs_completed => 600,    # no spec, use default
        heal_roots_completed => 600,    # no spec, use default
        normal               => 300,    # 5 min
    );

    #this just maps existing timeouts to the equivalents timeouts
    #for manual recovery and warning timeouts
    #this only works with *_completed keys
    %DEFAULT_TIMEOUT = map {
        my $t = $DEFAULT_TIMEOUT{$_};
        my $key = $_ eq 'normal' ? 'switchback_completed' : $_;
        $_                                      => $t,
          $key . '_with_manual_recovery_needed' => $t,
          $key . '_with_warnings'               => $t;
    } keys %DEFAULT_TIMEOUT;
    %SPECIFICATION_TIMEOUT = map {
        my $t = $SPECIFICATION_TIMEOUT{$_};
        my $key = $_ eq 'normal' ? 'switchback_completed' : $_;
        $_                                      => $t,
          $key . '_with_manual_recovery_needed' => $t,
          $key . '_with_warnings'               => $t;
    } keys %SPECIFICATION_TIMEOUT;

    #building state string for regex needed in option validation
    my $state_regex_string = join( '|', keys %DEFAULT_TIMEOUT );

    my $TIMEOUT_PARAM = $ENV{MCC_DR_STATE_TIMEOUT};
    if ($TIMEOUT_PARAM) {
        $Log->comment(
                "Using MCC_DR_STATE_TIMEOUT param $TIMEOUT_PARAM for "
              . 'dr state timeout' );
    }

    my %opts = validate_with(
        params => \@_,
        spec   => {

            #state   => { regex => qr/^($state_regex_string)$/ }, # or no regex?
            state => { type => SCALAR },                        # or no regex?
            timeout => { regex => qr/^\d+$/, optional => 1 },
        }
    );

    my $timeout =
         $TIMEOUT_PARAM
      || $opts{timeout}
      || $DEFAULT_TIMEOUT{ $opts{state} };
    my $spec_timeout = $SPECIFICATION_TIMEOUT{ $opts{state} };

    $Log->comment(
        "$opts{state} timeout: $timeout specification timeout:$spec_timeout");
    if (wantarray) {
        return ( $timeout, $spec_timeout );
    }
    else {
        return $timeout;
    }
}

=head2 wait_for_DR_state()

    my $state = $site_X_dr_coordinator->wait_for_DR_state(
                                     state  => 'heal_aggrs_completed');

This method waits until expected DR State is reached.DR_SOM check will be
skipped for in_progress dr-state's

=over

=item Options:

=over

=item C<< state => $state_to_wait_for >>

Required: Need to pass one of the following state values:

    normal
    switchover_in_progress
    switchover_completed
    heal_aggrs_in_progress
    heal_aggrs_completed
    heal_roots_in_progress
    heal_roots_completed
    switchback_in_progress

=item C<< wait_for_cluster_recovery => 0/1 >>

Optional: If in a particular test the DR operation itself is expected to
cause a cluster to become inaccessible, then this option may be set to 1 for the
wait to include a wait for the cluster to come back up and be reconnected to.

=item C<< handle_SO_warnings_for_unm_aggr => 0/1 >>

Optional: If set to 1 will handle SO completion warnings due to unmirrored aggrs
Default: 1

=item C<< timeout => xxx >>

Optional: The timeout in seconds for this method to wait through.  It is not
recommended that this option be passed for 'switchover_completed' or
'normal', as these are changes in state that have product specified timeouts.

=back

=back

=cut

sub wait_for_DR_state {
    my $trace = NACL::MTask::MCC::Log::TRACELOG->start_timer();    #metrics

    my $self = shift;

    #building state string for regex needed in option validation
    my @state_array_regex;
    for my $state_regex ( keys(%DR_STATES) ) {
        push( @state_array_regex, $state_regex );
    }
    my $state_regex_string = join( '|', @state_array_regex );
    my %opts = validate_with(
        params => \@_,
        spec   => {
            state   => { regex => qr/^($state_regex_string)$/ },
            timeout => { regex => qr/^\d+$/, optional => 1 },
            'wait_for_cluster_recovery' => { regex => qr/^\d+$/, default => 0 },
            'die_on_completed_with_manual_recovery' =>
              { type => BOOLEAN, default => 1 },
            'die_on_completed_with_warnings' =>
              { type => BOOLEAN, default => 1 },
            'handle_SO_warnings_for_unm_aggr' =>
              { type => BOOLEAN, default => 1 },
            'reconnect_nodes' => { type => ARRAYREF, optional => 1 },
        },
    );

# constants
# TODO these are in another method now...BUT we may want to have a shorter wait here,
# depending on how we deal with jobs now.  So for now, leaving them here
#my %DEFAULT_TIMEOUT = (
#switchover_completed => 300,
#heal_aggrs_completed => 600,
#heal_roots_completed => 600,
#normal               => 120,
#);

    my %INTERVAL = (
        switchover_completed => 5,
        heal_aggrs_completed => 15,
        heal_roots_completed => 5,
        normal               => 5,
    );

    my $desired_state = delete( $opts{state} );
    my $snooze_time =
      exists( $INTERVAL{$desired_state} ) ? $INTERVAL{$desired_state} : 5;
    my ( $default_timeout, $specification_timeout ) =
      $self->_get_DR_state_timeout( state => $desired_state );
    my $timeout =
      exists( $opts{timeout} )
      ? delete( $opts{timeout} )
      : $default_timeout;

    my $end_time = ( $timeout > 0 ) ? Tharn::timeout2time($timeout) : -1;
    my $no_time_limit = ( $timeout == 0 ) ? 1 : 0;

    my $wait_for_cluster_recovery = delete( $opts{wait_for_cluster_recovery} );
    my $die_on_completed_with_manual_recovery =
      $opts{die_on_completed_with_manual_recovery};
    my $die_on_completed_with_warnings = $opts{die_on_completed_with_warnings};
    my $handle_SO_warnings_for_unm_aggr =
      delete $opts{handle_SO_warnings_for_unm_aggr};

    $Log->comment("Waiting for State: $desired_state\n");
    my $found_desired_state = 0;
    my $current_state       = 'unknown';
    my $raw_state           = 'unknown';
    my $mcc_op_cs;
    my $attempts = 0;

    # time this block
    {
        my $tracer =
          NACL::MTask::MCC::Log::TRACELOG->start_timer(
            message => "WAIT_FOR: $desired_state" );    #metrics

        while ( $no_time_limit || ( time() <= $end_time ) ) {
            my $ci_connection_is_live = 0;
            try {
                #TODO - support the new completed warning|manual states here
                #...
                #we need to recognize the completed states and
                #support completing if it is expected (user passes in
                #something maybe) and if it's not expected potentially
                #we throw an exception here.  But we have to be able
                #to complete everything successfully.
                ( $current_state, $raw_state, $mcc_op_cs ) =
                  $self->get_DR_state();
                $ci_connection_is_live = 1;
            }
            catch NACL::MTask::MCC::Exceptions::BaseException with {

                # rethrow an MCC exception
                my $exception_object = shift;
                $attempts++;
                Tharn::snooze($snooze_time);
                if ( $attempts > 10 ) {
                    $exception_object->throw();
                }
            }
            catch NATE::BaseException with {

                # but this may signify a node is down
                my $exception_object = shift;
                if ( !$wait_for_cluster_recovery ) {

                    # unexpected
                    $attempts++;
                    Tharn::snooze($snooze_time);
                    if ( $attempts > 10 ) {
                        NACL::MTask::MCC::Exceptions::BaseException->throw(
                            $exception_object->text() );
                    }
                }
                else {
                    # expected
                    $Log->comment( 'Error is:' . $exception_object->text() );
                    $Log->comment('Trying to refresh CI if site live');
                    my $live_sites =
                      $self->_metrocluster()->L_determine_live_sites();
                    if ( any { $_ eq $self->_site() } @{$live_sites} ) {
                        $self->_metrocluster()
                          ->refresh_command_interfaces(
                            site => $self->_site() );
                    }

                    $ci_connection_is_live = 0;
                }
            };

            if ($ci_connection_is_live) {
                if (   $current_state eq $desired_state
                    || $raw_state eq $desired_state )
                {
                    $Log->comment(
                        "State found - Current State: $current_state\n");
                    $found_desired_state = 1;
                    last;
                }

                $Log->comment("Sleeping for: $snooze_time seconds\n");
                $Log->comment("Current State: $current_state\n");
                $Log->comment("Waiting for State: $desired_state\n");
            }

            Tharn::snooze($snooze_time);

            unless ($no_time_limit) {
                my $time_remaining = $end_time - time();
                $Log->comment("Time remaining: $time_remaining\n");
            }
        }

        # deal with special cases completed-with-manual-recovery-needed
        # and completed-with-warnings states.
        if (   ( $raw_state ne $desired_state )
            && ( $raw_state ne 'no_op' )
            && !$ENV{IGNORE_MCC_COMPLETED_STATES} )
        {
            my $errfield = join "\n", @{ $mcc_op_cs->error_list() };
            if ( $raw_state =~ /completed_with_manual_recovery_needed/ ) {
                my $errstr =
                    "Metrocluster Operation State Error:\n"
                  . "Current State: $raw_state\n"
                  . "was waiting for DR State: $desired_state\n"
                  . "Recovery text: $errfield";

               # Ignore SO completed with manual recovery needed warnings due to
               # Un-mirrored aggr offline and VServer offline - BURT 1003960
                my $unm_aggr_offline_warning =
                    'The following switched over aggregates '
                  . 'are missing or not online';
                my $vs_offline_warning =
                    'The following Vservers might have switched '
                  . 'over volumes that are missing or not mounted';
                if ( $raw_state =~
                    /switchover_completed_with_manual_recovery_needed/
                    && $errfield =~
                    /$unm_aggr_offline_warning|$vs_offline_warning/ )
                {
                    $Log->comment(
                            'Ignoring switchover completion warnings due to '
                          . 'Un-mirrored aggr offline and VServer offline' );
                }
                elsif ($die_on_completed_with_manual_recovery) {
                    NACL::MTask::MCC::Exceptions::State::CompletedWithManualRecovery
                      ->throw( $errstr, '-object' => $mcc_op_cs, );
                }
                else {
                    $Log->warn($errstr);
                }
            }
            elsif ( $raw_state =~ /completed_with_warnings/ ) {
                my $errstr =
                    "Metrocluster Operation State Error:\n"
                  . "Current State: $raw_state\n"
                  . "was waiting for DR State: $desired_state\n"
                  . "Warning text: $errfield";

              # Catch and ignore SO completion warnings due to un-mirrored aggrs
              # BURT 987929
                my $unm_aggr_warning =
                    'is not mirrored and its only plex "plex0" '
                  . 'uses storage from the remote pool';

                # Catch and ignore Heal Root Aggrs completion warnings due to
                # aggrs resyncing
                # BURT 997498
                my $HR_aggr_resync_warning =
                  'Some aggregates are still resyncing';

                # Catch and ignore SO completion warnings due to offlined plex
                #  BURT 1066771
                my $plex_offline_warning =
                    'is mirrored and its only online plex "plex0" '
                  . 'uses storage from the remote pool';

                if (   $handle_SO_warnings_for_unm_aggr
                    && $raw_state =~ /switchover_completed_with_warnings/
                    && $errfield  =~ /$unm_aggr_warning/ )
                {
                    $Log->comment(
'Ignoring SO completion warnings due to un-mirrored aggrs'
                    );
                }
                elsif ($raw_state =~ /heal_roots_completed_with_warnings/
                    && $errfield =~ /$HR_aggr_resync_warning/ )
                {
                    $Log->comment(
'Ignoring Heal Roots completion warnings due to aggregate '
                          . 'resyncing in progress' );
                }
                elsif ($raw_state =~ /switchover_completed_with_warnings/
                    && $errfield =~ /$plex_offline_warning/ )
                {
                    $Log->comment(
'Ignoring SO completion warnings due to plex offlined aggregate'
                    );
                }
                elsif ($die_on_completed_with_warnings) {
                    NACL::MTask::MCC::Exceptions::State::CompletedWithWarnings
                      ->throw( $errstr, '-object' => $mcc_op_cs, );
                }
                else {
                    $Log->warn($errstr);
                }
            }
        }

        unless ($found_desired_state) {
            NACL::MTask::MCC::Exceptions::BaseException->throw(
                    "Timed out waiting for DR State: $desired_state, "
                  . "Current State: $current_state\n" );
        }
    }

    # DR_SOM check will be skipped for in_progress dr-state

    if ( $current_state !~ /_in_progress/i ) {

        # now check that all nodes' DR_SOM's agree with this.
        # if not DR normal, we only care about nodes on the survivor site.
        my @sites_to_check =
          ( $current_state eq 'normal' ) ? (qw/ A B /) : ( $self->_site() );
        foreach my $site (@sites_to_check) {

            #refresh connections before
            my %extra = ();
            if ( $opts{reconnect_nodes} ) {
                $extra{nodes} = $opts{reconnect_nodes};
            }
            $self->_metrocluster()->refresh_command_interfaces(
                site => $site,
                %extra
            );
            my $cluster = $self->_metrocluster()->get_cluster( site => $site );
            my $sites_nodes = $cluster->L_get_cluster_mcc_nodes();

            my @nodes_to_validate = @{$sites_nodes};
            if ( $opts{reconnect_nodes} ) {

                # filter the node list so that it only contains the nodes
                # present in the reconnect list
                @nodes_to_validate =
                  grep { grep( /^$_$/, @{ $opts{reconnect_nodes} } ); }
                  @nodes_to_validate;
            }

            foreach my $mcc_node (@nodes_to_validate) {

                # this will throw an exception if there are any inconsistencies.
                $mcc_node->L_validate_DRSOM_FSM_state(
                    dr_state => $current_state );
            }
        }
    }

    if ( $raw_state ne 'no_op' && defined $mcc_op_cs ) {
        my $mcc_op_time = $self->_mcc_op_cs_time_delta($mcc_op_cs);

        if ( defined $specification_timeout
            && ( $mcc_op_time > $specification_timeout ) )
        {
            my $err =
                "dr operation took more time than spec allows: ($mcc_op_time"
              . " > $specification_timeout)";
            if ( $ENV{MCC_WARN_ONLY_ON_DR_OPERATION_TIME_OVERRUN} ) {
                $Log->warn($err);
            }
            else {
                NACL::MTask::MCC::Exceptions::State::OperationTimeOverrun
                  ->throw($err);
            }
        }
    }

    return $current_state;
}

sub _mcc_op_cs_time_delta {
    my $trace = NACL::MTask::MCC::Log::TRACELOG->start_timer();    #metrics

    my ( $self, $cs ) = @_;

    # 6/23/2014 22:57:32
    my $parse = DateTime::Format::Strptime->new( pattern => '%m/%d/%Y %T', );

    my $start = $parse->parse_datetime( $cs->start_time() );
    my $end   = $parse->parse_datetime( $cs->end_time() );

    if ( defined $start && defined $end ) {
        $Log->comment("metrocluster DR OP times:\nstart: $start\nend: $end");
        my $delta = $end - $start;
        return $delta->seconds();
    }
    else {
        $Log->warn( "Could not parse metrocluster operation show dates!\n"
              . Dumper $cs);
        return;
    }

}

sub L_was_forced_on_disaster {
    my $trace = NACL::MTask::MCC::Log::TRACELOG->start_timer();    #metrics
    my $self  = shift;

    return $self->_was_forced_on_disaster();
}

1;
