
# this is here due to our standard usage of @_ in Params::Validate methods.
##no critic (RequireArgUnpacking ProhibitFixedStringMatches)
##no critic (ControlStructures::ProhibitUnlessBlocks)
##no critic (ValuesAndExpressions::ProhibitInterpolationOfLiterals)

#
# Copyright (c) 2011-2014 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary MCC DR_Group Module
## @author dl-mcc-dpg@netapp.com
## @status shared
## @pod here

package NACL::MTask::MCC::DR_Group;

use strict;
use warnings;
use Carp;

use base qw (NACL::MTask::MTask);

use NACL::MTask::MCC::DR_Partnership ();
use NACL::MTask::MCC::HA_Group ();
use NACL::MTask::MCC::Shared_Storage ();
use NACL::MTask::MCC::Network::Topology ();
use NACL::MTask::MCC::Log ();

#use NACL::C::Metrocluster ();

use NATE::BaseException qw(:try);
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();

use Params::Validate
    qw(validate validate_with BOOLEAN HASHREF OBJECT ARRAYREF SCALAR UNDEF);

use Readonly;
use Data::Dumper;


=head1 NAME

NACL::MTask::MCC::DR_Group

=head1 DESCRIPTION

This class is to be used for C-Mode Metrocluster testing.

    # Start test
    use NACL::MTask::MCC::Metrocluster;
    ...
    my %dr_groups = %{ $mcc->get_DR_groups() };
    ...
    # Perform test here
    ...
    # End test

=cut

=head1 Public Methods

There is no public constructor for this class.  NACL::MTask::MCC::DR_Group
objects are obtained via calls such as

    my %dr_groups = %{ $mcc->get_DR_groups() };

    my $dr_group = $mcc->get_DR_groups()->{$dr_group_id};

    # see NACL::MTask::MCC::Metrocluster::get_singleton_DR_group() before use.
    my $dr_group = $mcc->get_singleton_DR_group();

In the initial implementation of MCC, there will only be a single DR Group in
a Metrocluster element, but future versions may have multiple DR Groups within
one Metrocluster element.

=cut

# Automatically generate the (non-public) constructor (includes a call to
# 'init') and also creates non-public members and non-public accesors for them.
use Class::MethodMaker [
    new    => [ '-init', 'new' ],
    scalar => [
        { -type => 'NACL::MTask::MCC::Metrocluster' },
        '_metrocluster',
              ],
    scalar => [
        { -type => 'NACL::MTask::MCC::Node' },
        '_node_A1',
              ],
    scalar => [
        { -type => 'NACL::MTask::MCC::Shared_Storage' },
        '_shared_storage',
              ],
    scalar => [
        { -type => 'NACL::MTask::MCC::Network::Topology' },
        '_network_topology',
              ],
    scalar => [
        qw/ _is_4_pack /,
              ],
    scalar => [
        qw/ _is_2_pack /,
              ],
    scalar => [
        qw/ _dr_partnerships /,
              ],
    scalar => [
        qw/ _ha_groups /,
              ],
    scalar => [
        qw/ _dr_group_id /,
              ],
                       ];

# validate the parameters passed in and what else?
sub init {
    my $trace = NACL::MTask::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;
    my %opts = validate_with(
        params => \@_,
        spec   => {
            metrocluster                          =>
                                 { isa   => 'NACL::MTask::MCC::Metrocluster' },
            site_A_Metrocluster_component         =>
                                 { isa   => 'NACL::C::Metrocluster' },
            A1_node_name                          =>
                                 { type  => SCALAR },
            constructor_log_path                  =>
                                 { type  => SCALAR|UNDEF },
                  },
                            );

    my $constructor_log_path = delete($opts{'constructor_log_path'});

    # set members per the passed constructor arguments.
    $self->_metrocluster($opts{'metrocluster'});

    my $dr_group_id = $self->_create_dr_group_id(
        site_A_Metrocluster_component =>
                                     $opts{site_A_Metrocluster_component},
        A1_node_name                  =>
                                     $opts{A1_node_name},

    );
    $self->_dr_group_id($dr_group_id);

    # initialize DR Partnerships pairing data
    $Log->reset_path();
    $Log->comment('    Determining DR Partnerships...');

    $Log->redirect(path => $constructor_log_path) if defined $constructor_log_path;
    my $dr_partnerships = $self->_create_DR_partnerships(
         site_A_Metrocluster_component => $opts{site_A_Metrocluster_component},
         A1_node_name                  => $opts{A1_node_name});
    $self->_dr_partnerships($dr_partnerships);
    $Log->reset_path();

    $Log->comment('    ...done');

    # set node A1
    $Log->comment('    Determining "A1" node...');

    $Log->redirect(path => $constructor_log_path) if defined $constructor_log_path;
    foreach my $partnership (@{ $dr_partnerships }) {
        my $site_A_node = $partnership->get_partner_node(site => 'A');
        if ($site_A_node->get_name() eq $opts{A1_node_name}) {
            $self->_node_A1($site_A_node);
            last;
        }
    }
    $Log->reset_path();

    $Log->comment('    ...done.  A1 = ' . $self->_node_A1()->get_name());

    # if there are 2 partnerships, this is a 4-pack
    $Log->comment('    Is this DR Group a 4-pack or a 2-node MCC?');

    $Log->redirect(path => $constructor_log_path) if defined $constructor_log_path;
    $self->_is_4_pack(scalar(@{ $dr_partnerships }) >= 2);


    # Only one DR partnership is a 2-pack
    $self->_is_2_pack(scalar(@{ $dr_partnerships }) == 1);

    $Log->reset_path();

    $Log->comment('    ...this is a ' . ($self->_is_4_pack() ?
                                 'standard 4-pack' : '2-node') . ' DR Group.');

    #  two HA Groups in a DR_Group (basically organizational entities here)
    $Log->comment('    Determining HA Groups...');

    $Log->redirect(path => $constructor_log_path) if defined $constructor_log_path;
    {
        my %ha_groups = ();
        foreach my $site ('A', 'B') {
            $ha_groups{$site} = NACL::MTask::MCC::HA_Group->new(
                dr_group => $self,
                cluster  => $self->_metrocluster()->get_cluster(site => $site));
        }
        $self->_ha_groups(\%ha_groups);
    }
    $Log->reset_path();

    $Log->comment('    ...done');

    # now that the MCC::Nodes in the DR partnerships have been constructed, we
    # can initialize their MCC::Network::Node handles.  This is necessary to do
    # before trying to build the storage network topology object...
    $Log->comment('    Initializing node shared storage network handles...');

    $Log->redirect(path => $constructor_log_path) if defined $constructor_log_path;
    foreach my $partnership (@{ $dr_partnerships }) {
        foreach my $site ('A', 'B') {
            my $mcc_node = $partnership->get_partner_node(site => $site);
            $mcc_node->L_init_network_node_handle();
        }
    }
    $Log->reset_path();

    $Log->comment('    ...done');

    # do network topology discovery for real groups
    #    (see BURT 642234 before making snide remarks about following line...)
    if ( ( ! $self->L_is_vsim_group() ) && $self->L_has_switches() ) {
        # users can use ENV var DISABLE_SHARED_STORAGE_NETWORK_VALIDATION to
        # disable this discovery/validation process
        my $skip_shared_storage_network_discovery =
                defined($ENV{DISABLE_SHARED_STORAGE_NETWORK_VALIDATION}) ?
                           $ENV{DISABLE_SHARED_STORAGE_NETWORK_VALIDATION} : 0;
        if ($skip_shared_storage_network_discovery) {
            $Log->warn(
                "    User set DISABLE_SHARED_STORAGE_NETWORK_VALIDATION=1\n" .
                '    Skipping recommended shared storage network validation!');
        } else {
            $Log->comment('    Starting network topology discovery...');

            $Log->redirect(path => $constructor_log_path);
            $self->_network_topology(
                  NACL::MTask::MCC::Network::Topology->new(dr_group => $self));
            $Log->reset_path();

            $Log->comment('    ...done');
        }
    }

    # record initial state of storage disks

    # this results in Shared_Storage objects that are not fully initialized yet.
    # the full init occurs in the first disk validation triggered in the
    # main metrocluster constructor.
    $Log->comment('    Initial setup of shared storage harness...');

    $Log->redirect(path => $constructor_log_path) if defined $constructor_log_path;
    $self->_shared_storage(NACL::MTask::MCC::Shared_Storage->new(
                                        metrocluster => $self->_metrocluster,
                                        dr_group     => $self));
    $Log->reset_path();

    $Log->comment('    ...done');

    return;
}

# _create_dr_group_id()
#
#
# Every DR group in a MCC setup now is identified by a unique id which maps to
# all the nodes that are part of the DR group. This method returns a scalar
# value. For latest builds it will return the dr_group_id for the A1 node or for
# older builds it will return the A1 node name

sub _create_dr_group_id{
    my $trace = NACL::MTask::MCC::Log::TRACELOG->start_timer();
    my $self = shift;

    #vaidate opts
    my %opts = validate_with(
        params => \@_,
        spec   => {
            site_A_Metrocluster_component => {isa  => 'NACL::C::Metrocluster'},
            A1_node_name                  => { type => SCALAR },
        },
    );

    my $site_A_mcc_component = $opts{site_A_Metrocluster_component};
    my $a1node = $opts{A1_node_name};
    my $returnvalue;
    my $mcc = $self->_metrocluster();

    # TODO Remove when safe
    $mcc->has_uichange_scaffold();

    my @responses = NACL::MTask::MCC::Utils->metrocluster_node_show(
        command_interface =>
            $site_A_mcc_component->command_interface(),
        requested_fields  => ['node','dr-group-id'],
        filter            => {
            node => $a1node,
        },);
    $returnvalue = $responses[0]->{'dr_group_id'};
    $Log->comment("dr group id for $a1node is: $returnvalue");

    return $returnvalue;
}

=head2 get_A1_node()

    my $node_A1 = $dr_group->get_A1_node();

This method returns the NACL::MTask::MCC::Node object which corresponds to the
A1 position in the DR Group.

=cut

sub get_A1_node {
    my $trace = NACL::MTask::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;

    return $self->_node_A1();
}

=head2 get_nodes_by_role()

    my $mcc_node_A1 = $dr_group->get_nodes_by_role()->{A1};

This method returns a hashref of the nodes (NACL::MTask::MCC::Node objects) vs
the MCC system role the node is playing, i.e.:

    {
        A1 => $A1_node,
        A2 => $A2_node,
        B1 => $B1_node,
        B2 => $B2_node,
    }

Where the nodes are defined as follows:

    A1 is the node that was passed as the A1 node for this DR Group.
    B1 is A1's DR Partner
    A2 is A1's HA Partner
    B2 is A2's DR Partner

=cut

sub get_nodes_by_role {
    my $trace = NACL::MTask::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;

    my @mcc_nodes = @{ $self->get_mcc_nodes() };
    my %node_roles = map { $_->get_role() => $_ } @mcc_nodes;

    return \%node_roles;
}

=head2 get_id()

Gets the unique identifier for a DR Group within the Metrocluster element.

    my $dr_group_id = $dr_group->get_id();

Returns a string identifier unique across all the DR Groups in a Metrocluster
system.  Currently this is just the name of the A1 node in this group.

=cut

sub get_id {
    my $trace = NACL::MTask::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;

    return $self->_dr_group_id();
}

# _create_DR_partnerships()
#
#
# Each node in the DR Group at site A is paired in a DR Partnership with a
# single node in the DR Group at site B.  This method returns an arrayref of
# NACL::MTask::MCC::DR_Partnership objects associated with this DR Group.
sub _create_DR_partnerships {
    my $trace = NACL::MTask::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;

    # validate opts
    my %opts = validate_with(
        params => \@_,
        spec   => {
          site_A_Metrocluster_component => { isa  => 'NACL::C::Metrocluster' },
          A1_node_name                  => { type => SCALAR },
                  },
                            );

    my $A1_node_name = $opts{A1_node_name};
    my $site_A_Metrocluster_component = $opts{site_A_Metrocluster_component};

    my @partnerships = ();
    my @responses = NACL::MTask::MCC::Utils->metrocluster_node_show(
            command_interface =>
                $site_A_Metrocluster_component->command_interface(),
            requested_fields => [ 'node', 'ha-partner', 'dr-partner', ]);

    my $A2_node_name = undef;
    foreach my $response (@responses) {
        if ($response->{node} eq $A1_node_name) {
             $A2_node_name = $response->{'ha_partner'};
             push(@partnerships, NACL::MTask::MCC::DR_Partnership->new(
                            dr_group            => $self,
                            site_A_partner_name => $A1_node_name,
                            site_B_partner_name => $response->{'dr_partner'},
                                                                      )
                 );
             last;
        }
    }

    foreach my $response (@responses) {
        if ($response->{node} eq $A2_node_name) {
             push(@partnerships, NACL::MTask::MCC::DR_Partnership->new(
                            dr_group            => $self,
                            site_A_partner_name => $A2_node_name,
                            site_B_partner_name => $response->{'dr_partner'}
                                                                      )
                 );
             last;
        }
    }

    return \@partnerships;
}

=head2 get_DR_partnerships()

    my @dr_partners = @{ $dr_group->get_DR_partnerships() };

    my $an_mcc_node_at_site_A = $dr_partners[0]->get_partner_node(site => 'A');

Each node in the DR Group at site A is paired in a DR Partnership with a
single node in the DR Group at site B.  This method returns an array of
NACL::MTask::MCC::DR_Partnership objects associated with this DR Group.

=cut

sub get_DR_partnerships {
    my $trace = NACL::MTask::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;

    return $self->_dr_partnerships();
}

=head2 get_HA_groups()

    my $ha_groups = $dr_group->get_HA_groups();
    my $ha_group_at_A = $dr_group->get_HA_groups()->{'A'};

This method returns a hashref containing the HA Groups associated with each
cluster in this DR Group.  The returned data structure has the following format:

    { 'A' => $ha_group_at_A (a NACL::MTask::MCC::HA_Group object),
      'B' => $ha_group_at_B (a NACL::MTask::MCC::HA_Group object)
    }

=cut

sub get_HA_groups {
    my $trace = NACL::MTask::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;

    return $self->_ha_groups();
}

=head2 is_4_pack()

   my $dr_group = $mcc->get_DR_groups()->[0];
   my $is_4_node = $dr_group->is_4_pack();

This method is used to determine if the testbed used for test is a 4
pack. Method returns boolean value true on determining that testbed is
a 4-pack and false if it's a 2-node.

=cut

sub is_4_pack {

    my $trace = NACL::MTask::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;
    return ( $self->_is_4_pack() );
}

=head2 is_2_pack()

   my $dr_group = $mcc->get_DR_groups()->[0];
   my $is_2_pack = $dr_group->is_2_pack();

Returns true if this is a 2-node cluster, false otherwise.

=cut

sub is_2_pack {
    my $self = shift;

    return $self->_is_2_pack();
}

=head2 is_real_4_pack()

   my $dr_group = $mcc->get_DR_groups()->[0];
   my $is_real_testbed = $dr_group->is_real_4_pack();

Deprecated, do not use.This method is used to determine if the testbed used for
test is real 4 pack or fake 2 node pack. Method returns boolean value true on
determining that testbed is real 4 pack and false when testbed is fake 2 node.

=cut

## This method will be removed later
sub is_real_4_pack {

    my $trace = NACL::MTask::MCC::Log::TRACELOG->start_timer(); #metrics

    $Log->comment(' Use of this method is_real_4_pack is Deprecated');
    my $self = shift;
    return ( $self->is_4_pack() );
}

=head2 L_is_vsim_group()

Returns true (1) if this DR Group is made up of vsim nodes.  False otherwise.

=cut

sub L_is_vsim_group {
    my $trace = NACL::MTask::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;

    return $self->_node_A1()->L_is_vsim();
}

sub L_has_switches {
    my $self = shift;

    return $self->_node_A1()->L_has_switches();
}

=head2 get_network_topology()

    my $dr_group = $mcc->get_DR_groups()->[0];
    my $dr_group_network_topology = $dr_group->get_network_topology();

This method returns the network topology of this DR Group.  It is in the form
of a NACL::MTask::MCC::Network::Topology object.

This method throws an exception if this DR_Group was constructed with the
following user set environment variable set to 1:

    DISABLE_SHARED_STORAGE_NETWORK_VALIDATION=1

=cut

sub get_network_topology {
    my $trace = NACL::MTask::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;

    unless ($self->_network_topology_isset()) {
        NACL::MTask::MCC::Exceptions::BaseException->throw(
                  "Shared storage network topology not available.:\n" .
                  "This DR Group was constructed without topology discovery.");
    }

    return $self->_network_topology();
}

=head2 get_shared_storage()

    my $shared_storage = $mcc->get_shared_storage();

Returns the NACL::MTask::MCC::Shared_Storage object.

=cut

sub get_shared_storage {
    my $trace = NACL::MTask::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;

    return $self->_shared_storage();
}

=head2 get_mcc_nodes()

    my @mcc_nodes = @{ $dr_group->get_mcc_nodes() };

    my @mcc_nodes = @{ $dr_group->get_mcc_nodes(site => 'A') };

This method returns arrayref NACL::MTask::MCC::Nodes in this DR-Group.  If site
is passed, it will return the nodes at that particular site in this DR group.

=over

=item Options

=over

=item C<< site => 'A | B' >>

(Optional) Specifies which site's node are desired.

=back

=back

=cut

sub get_mcc_nodes {
    my $trace = NACL::MTask::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;

    # validate opts
    my %opts = validate_with(
        params => \@_,
        spec   => {
            site => { regex => qr/^[AB]$/, optional => 1 },
                  },
                            );

    my @mcc_nodes = ();
    my @sites = defined($opts{site}) ? $opts{site} : qw(A B);
    my @dr_partnerships = @{ $self->get_DR_partnerships() };
    foreach my $dr_partnership (@dr_partnerships) {
        foreach my $site (@sites) {
            push(@mcc_nodes, $dr_partnership->get_partner_node(site => $site));
        }
    }

    return \@mcc_nodes;
}

1;
