# $Id $

# this is here due to our standard usage of @_ in Params::Validate methods.
##no critic (RequireArgUnpacking)

#
# Copyright (c) 2011-2021 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary MCC Network::Fabric Stub MTask Module
## @author dl-mcc-dpg-qa@netapp.com
## @status shared
## @pod here

package NACL::MTask::MCC::Network::Fabric;

use strict;
use warnings;
use Carp;

use base qw (NACL::MTask::MTask);

use NACL::MTask::MCC::Network::Switch ();

use NACL::Exceptions::NoElementsFound qw(:try);
use NATE::BaseException qw(:try);

use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();

use Params::Validate
    qw(validate validate_with BOOLEAN HASHREF OBJECT ARRAYREF SCALAR UNDEF);
use Readonly;
use Data::Dumper;


=head1 NAME

NACL::MTask::MCC::Network::Fabric

=head1 DESCRIPTION

This task is to be used for C-Mode MetroCluster testing.  This class helps to
define the network hardware mapping of an MCC DR Group's backend storage.

An MCC fabric is a container which organizes the inter/intra-site FC switches
and provides a public interface to their external access points.

Examples of such access points are connections to nodes and fiber bridges, but
not ISL connections which lie internal to the fabric.

The fabric also gives access to the individual switches inside the fabric for
more detailed control when needed.


    # Start test
    use NACL::MTask::MCC::Metrocluster;
    ...

    my $topology = $mcc->get_DR_groups()->{$group_id}->get_network_topology();
    my $fabric = $topology->get_intersite_fabrics()->[0];
    my $switch = $fabric->get_switch(site => 'A');

    ...
    # Perform test here
    ...

    # End test

Also please see the diagrams at

http://wikid.netapp.com/w/MCC_Automation/MCC_DPG_QA_Automation/Test_Library_Dev#MCC_Network_Topology_Discussion_and_Notes

for further clarification.

=cut

=head1 METHODS

There is no public constructor for this class.  A
NACL::MTask::MCC::Network::Fabric object is obtained via calls such as used in
conjunction with a NACL::MTask::MCC::Metrocluster object:

    my @fabrics = @{ $topology->get_intersite_fabrics() };

=cut

use Class::MethodMaker [
    new    => [ '-init', 'new' ],
    scalar => [
                { -type => 'NACL::MTask::MCC::Network::Topology' },
                '_topology',
                { -type => 'NACL::MTask::MCC::Network::Switch' },
                '_site_A_switch',
                { -type => 'NACL::MTask::MCC::Network::Switch' },
                '_site_B_switch',
              ],
                        ];

# validate the parameters passed in and what else?
sub init {
    $Log->enter() if $may_enter;
    my $self = shift;

    my %opts = validate_with(
        params => \@_,
        spec   => {
            site_A_switch => { isa => 'NACL::MTask::MCC::Network::Switch' },
            topology      => { isa => 'NACL::MTask::MCC::Network::Topology' },
                  },
                            );

    $self->_topology($opts{topology});

    my $site_A_switch = delete($opts{site_A_switch});

    $site_A_switch->_fabric($self);

    my $site_B_switch = $self->_determine_site_B_switch_in_this_fabric(
                                              site_A_switch => $site_A_switch);

    $self->_site_A_switch($site_A_switch);
    $self->_site_B_switch($site_B_switch);

    $Log->exit() if $may_exit;
    return;
}

=head2 _determine_site_B_switch_in_this_fabric()

    my $site_B_switch = $self->_determine_site_B_switch_in_this_fabric(
                                              site_A_switch => $site_A_switch);

Uses information we can get from the virtual/real switch $site_A_switch to
determine the virtual/real switch at site B which together form one of the
fabrics in the shared storage topology.

=over

=item Options

=over

=item C<< site_A_switch => $site_A_switch >>

(Required) Specifies the NACL::MTask::MCC::Network::Switch object that forms the
site A side of this fabric.

=back

=back

=cut

sub _determine_site_B_switch_in_this_fabric {
    my $trace = NACL::MTask::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;
    my %opts = validate_with(
        params => \@_,
        spec   => {
            site_A_switch => { isa => 'NACL::MTask::MCC::Network::Switch' },
                  },
                            );

    my $site_A_switch = delete($opts{site_A_switch});

    my ($site_B_switch_name,$vswitch_isl_target_spec) =
        $site_A_switch->get_vswitch_isl_target_spec();
    $Log->comment('switch site B target wwns:' .
                  Dumper $vswitch_isl_target_spec);

    $Log->comment("Site B switch name: $site_B_switch_name");
    return NACL::MTask::MCC::Network::Switch->new(
               vswitch_isl_target_spec => $vswitch_isl_target_spec,
               switch_name             => $site_B_switch_name,
               fabric                  => $self,
               );
}

=head2 get_switch()

    my $site_A_switch = $fabric->get_switch(site => 'A');

Returns the NACL::MTask::MCC::Network::Switch object associated with the switch
in this fabric located at the specified site.

=over

=item Options

=over

=item C<< site => 'A'|'B' >>

(Required) Specifies the site in which to get the switch from in this fabric.

=back

=back

=cut

sub get_switch {
    $Log->enter() if $may_enter;
    my $self = shift;
    my %opts = validate_with(
        params => \@_,
        spec   => {
            site => { regex => qr/^[AB]$/, }
                  },
                            );

    my $site = delete($opts{site});

    $Log->exit() if $may_exit;
    return ($site eq 'A') ? $self->_site_A_switch() : $self->_site_B_switch();
}

=head2 get_fcvi_ports()

    my $site_A_fcvi_ports = $fabric->get_fcvi_ports(site => 'A');

    # disable all FCVI ports on fabric at site A
    map { $_->disable() } @{ $site_A_fcvi_ports };

Returns an arrayref to the NACL::MTask::MCC::Network::Port objects associated
with the connections from the fabric to the fcvi ports on the nodes at the
specified site.

=over

=item Options

=over

=item C<< site => 'A'|'B' >>

(Required) Site at which to gather these ports from on this fabric.

=back

=back

=cut

sub get_fcvi_ports {
    $Log->enter() if $may_enter;
    my $self = shift;
    my %opts = validate_with(
        params => \@_,
        spec   => {
            site => { regex => qr/^[AB]$/, }
                  },
                            );

    my $site = delete($opts{site});

    $Log->exit() if $may_exit;
    return $self->get_switch(site => $site)->get_fcvi_ports();
}

=head2 get_storage_ports()

    my $site_A_storage_ports = $fabric->get_storage_ports(site => 'A');

    # disable all storage ports on fabric at site A
    map { $_->disable() } @{ $site_A_storage_ports };

Returns an arrayref to the NACL::MTask::MCC::Network::Port objects associated
with the connections from the fabric to the storage traffic ports on the nodes
at the specified site.

=over

=item Options

=over

=item C<< site => 'A'|'B' >>

(Required) Site at which to gather these ports from on this fabric.

=back

=back

=cut

sub get_storage_ports {
    $Log->enter() if $may_enter;
    my $self = shift;
    my %opts = validate_with(
        params => \@_,
        spec   => {
            site => { regex => qr/^[AB]$/, }
                  },
                            );

    my $site = delete($opts{site});

    $Log->exit() if $may_exit;
    return $self->get_switch(site => $site)->get_storage_ports();
}

=head2 get_fibreBridge_ports()

    my $site_B_fibreBridge_ports = $fabric->get_fibreBridge_ports(site => 'B');

    # enable all fibre-bridge ports on fabric at site B
    map { $_->enable() } @{ $site_B_fibreBridge_ports };

Returns an arrayref to the NACL::MTask::MCC::Network::Port objects associated
with the connections from the fabric to the fibre-bridge array at the specified
site.

=over

=item Options

=over

=item C<< site => 'A'|'B' >>

(Required) Site at which to gather these ports from on this fabric.

=back

=back

=cut

sub get_fibreBridge_ports {
    $Log->enter() if $may_enter;
    my $self = shift;
    my %opts = validate_with(
        params => \@_,
        spec   => {
            site => { regex => qr/^[AB]$/, }
                  },
                            );

    my $site = delete($opts{site});

    $Log->exit() if $may_exit;
    return $self->get_switch(site => $site)->get_fibreBridge_ports();
}

1;
