# $Id $

# this is here due to our standard usage of @_ in Params::Validate methods.
##no critic (RequireArgUnpacking)
##no critic (ControlStructures::ProhibitUnlessBlocks)
##no critic (ValuesAndExpressions::ProhibitInterpolationOfLiterals)

#
# Copyright (c) 2011-2021 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary MCC Network::Node Stub MTask Module
## @author dl-mcc-dpg-qa@netapp.com
## @status shared
## @pod here

package NACL::MTask::MCC::Network::Node;

use strict;
use warnings;
use Carp;

use base qw (NACL::MTask::MTask);

use NACL::APISet::Response;

use NACL::Exceptions::NoElementsFound qw(:try);
use NATE::BaseException qw(:try);

use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();

use Params::Validate
    qw(validate validate_with BOOLEAN HASHREF OBJECT ARRAYREF SCALAR);
use Readonly;
use Data::Dumper;


Readonly my $TO_BE_DETERMINED => undef;


=head1 NAME

NACL::MTask::MCC::Network::Node

=head1 DESCRIPTION

This task is to be used for C-Mode MetroCluster testing.

This class helps to define the network hardware mapping of an MCC DR Group.  An
MCC::Network::Node (not to be confused with the more general MCC::Node) is the
interface to get network specific information and objects associated with a
specific node in the DR Group, for example a node's set of FCVI ports used
in conjunction with the node's interconnect with its DR Partner.

    # Start test
    use NACL::MTask::MCC::Metrocluster;
    ...

    my $topology = $mcc->get_DR_groups()->{$group_id}->get_network_topology();
    my $network_node = $topology->get_network_nodes()->{'B'}->{'1'};

    my @fcvi_ports = @{ $network_node->get_fcvi_ports() };

    ...
    # Perform test here
    ...

    # End test

Also please see the diagrams at

http://wikid.netapp.com/w/MCC_Automation/MCC_DPG_QA_Automation/Test_Library_Dev#MCC_Network_Topology_Discussion_and_Notes

for further clarification.

=cut

=head1 Public METHODS

There is no public constructor for this class.  A
NACL::MTask::MCC::Network::Node object is obtained via calls such as

    my $network_node = $topology->get_network_nodes()->{'B'}->{'1'};

=cut

# automatically generate a (Non-public) constructor (includes a call to 'init')
# and (Non-public) member accessor methods
use Class::MethodMaker [
    new    => [ '-init', 'new' ],
    scalar => [
                { -type => 'NACL::MTask::MCC::DR_Group' },
                '_dr_group',
              ],
    scalar => [
                { -type => 'NACL::MTask::MCC::Node' },
                '_mcc_node',
              ],
    scalar => [
                { -type => 'NACL::APISet::Response' },
                { q{*}  => undef },  #don't autogenerate accessor method
                '_init_sysconfig_va_response',
              ],
    scalar => [
                { -type => 'NACL::APISet::Response' },
                { q{*}  => undef },  #don't autogenerate accessor method
                '_init_storage_show_bridge_response',
              ],
    scalar => [
                { -type => 'NACL::APISet::Response' },
                { q{*}  => undef },  #don't autogenerate accessor method
                '_init_storage_show_switch_response',
              ],
    scalar => [
        qw/ _dr_group_topology_role /,
              ],
    scalar => [
        qw/ _sysconfig_attempt_time /,
              ],
                        ];

# validate the parameters passed in and what else?
sub init {
    my $trace = NACL::MTask::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;

    my %opts = validate_with(
        params => \@_,
        spec   => {
            dr_group => { isa   => 'NACL::MTask::MCC::DR_Group' },
            mcc_node => { isa   => 'NACL::MTask::MCC::Node' },
        },
                            );

    my $dr_group = delete($opts{'dr_group'});
    my $mcc_node = delete($opts{'mcc_node'});

    $self->_dr_group($dr_group);
    $self->_mcc_node($mcc_node);

    # will be (A1|A2|B1|B2)
    $self->_dr_group_topology_role($self->_determine_network_topology_role());

    ## the local nodescope sysconfig -va results at initialization
    ##
    ## this call is taking way too long in some cases.  complain about it if so.
    #my $start_time = time();
    #$self->_init_sysconfig_va_response(
    #$mcc_node->get_nodescope_apiset()->sysconfig(
    #'verbose'              => '1',
    #'detailed-information' => '1'));
    #my $delta_t = time() - $start_time;
    #if ($delta_t > 55) {
    #my $msg = "'sysconfig -va' on node " . $mcc_node->get_name() .
    #" took $delta_t seconds!  Report to BURT 581107 if still " .
    #"occurring after July (?)";
    #$Log->warn($msg);
    #print "WARNING: $msg\n";
    #}

    $self->_sysconfig_attempt_time(0);

    #$Log->comment("create syconfig - in object init");
    my $sysconfig = $self->_create_sysconfig_var();
    $self->_init_sysconfig_va_response($sysconfig);

    $self->_init_storage_show_bridge_response(
        $self->_get_storage_show_bridge());
    $self->_init_storage_show_switch_response(
        $self->_get_storage_show_switch());

    return;
}

# in place of method maker's auto generated accessor...
sub _init_storage_show_switch_response {
    my $trace = NACL::MTask::MCC::Log::TRACELOG->start_timer(); #metrics
    my $self = shift;

    if (scalar @_) {
        $self->{_init_storage_show_switch_response} = shift; #set it if we need to
        return $self->{_init_storage_show_switch_response};
    } else {
        #return it if it's already here
        if (defined $self->{_init_storage_show_switch_response}) {
            return $self->{_init_storage_show_switch_response};
        } else {
            #try creating a new one and return it
            my $switch;
            #$Log->comment("create storage show switch - in init st sh s");
            if ($switch = $self->_get_storage_show_switch()) {
                $self->{_init_storage_show_switch_response} = $switch;
            }
            return $switch;
        }
    }
}

# in place of method maker's auto generated accessor...
sub _init_storage_show_bridge_response {
    my $trace = NACL::MTask::MCC::Log::TRACELOG->start_timer(); #metrics
    my $self = shift;

    if (scalar @_) {
        $self->{_init_storage_show_bridge_response} = shift; #set it if we need to
        return $self->{_init_storage_show_bridge_response};
    } else {
        #return it if it's already here
        if (defined $self->{_init_storage_show_bridge_response}) {
            return $self->{_init_storage_show_bridge_response};
        } else {
            #try creating a new one and return it
            my $bridge;
            #$Log->comment("create storage show bridge - in init st sh b");
            if ($bridge = $self->_get_storage_show_bridge()) {
                $self->{_init_storage_show_bridge_response} = $bridge;
            }
            return $bridge;
        }
    }
}

# in place of method maker's auto generated accessor...
sub _init_sysconfig_va_response {
    my $trace = NACL::MTask::MCC::Log::TRACELOG->start_timer(); #metrics
    my $self = shift;

    if (scalar @_) {
        $self->{_init_sysconfig_va_response} = shift; #set it if we need to
        return $self->{_init_sysconfig_va_response};
    } else {
        #return it if it's already here
        if (defined $self->{_init_sysconfig_va_response}) {
            return $self->{_init_sysconfig_va_response};
        } else {
            #try creating a new one and return it
            my $sysconfig;
            #$Log->comment("create syconfig - in init sc va r");
            if ($sysconfig = $self->_create_sysconfig_var()) {
                $self->{_init_sysconfig_va_response} = $sysconfig;
            }
            return $sysconfig;
        }
    }
}


# returns true if we're in a bad state and should wait to try again
sub _node_down_should_wait {
    my $trace = NACL::MTask::MCC::Log::TRACELOG->start_timer(); #metrics
    my $self = shift;

    my $delta_sysconfig_time = time - $self->_sysconfig_attempt_time();
    my $skip_time = 3 * 60;  #3 minutes?
    my $ignore_bad_state =
        $self->_dr_group()->_metrocluster()->_ignore_bad_state();

    #parent node object down?
    my $transit_wait = $self->_mcc_node()->_transit_attempt_should_wait();

    #pick either parent node wait time or ours
    my $wait_time =  $self->_sysconfig_attempt_time()     ?
                     ($skip_time - $delta_sysconfig_time) :
                     $transit_wait;

    if ($ignore_bad_state &&
        (($delta_sysconfig_time < $skip_time) || $transit_wait ) ) {
        return $wait_time;
    } else {
        return 0;
    }

}


# helper method to do sysconfig call
sub _create_sysconfig_var {
    my $trace = NACL::MTask::MCC::Log::TRACELOG->start_timer(); #metrics
    my $self = shift;
    $Log->comment("create syconfig - in create sysconfig var:" .
        $self->_mcc_node()->get_name()
        . ' [' . $self->_sysconfig_attempt_time() . ']');

    my $sysconfig;

    if (my $wait_time = $self->_node_down_should_wait()) {
        $Log->comment('It looks like we are trying to repeatedly call' .
            ' sysconfig on a node that is down while in bad start mode.' .
            " Wait $wait_time seconds before trying again");
        return;
    }

    try {
        # the local nodescope sysconfig -va results at initialization
        #
        # this call is taking way too long in some cases.
        # complain about it if so.
        my $start_time = time();
        # this is a temporary (?) fix to BURT 601682 where calls to sysconfig
        # are timing out in the APISet object
        $self->_mcc_node()->get_nodescope_apiset()->set_timeout('connectrec-timeout' => 300);
        $sysconfig =
                    $self->_mcc_node()->get_nodescope_apiset()->sysconfig(
                                              'verbose'              => '1',
                                              'detailed-information' => '1',
                                              'connectrec-timeout'   => 300);
        my $delta_t = time() - $start_time;
        if ($delta_t > 55) {
            my $msg =
                "'sysconfig -va' on node " . $self->_mcc_node()->get_name() .
                " took $delta_t seconds!  Report to BURT 581107 if still " .
                "occurring after July (?)";
            $Log->warn($msg);
            print "WARNING: $msg\n";
        }
        $self->_sysconfig_attempt_time(0);
    } catch NATE::BaseException with {
        $Log->warn("node " .  $self->_mcc_node()->get_name() .
            " may be down, can't call sysconfig...");
        $self->_sysconfig_attempt_time(time);
    };

    return $sysconfig;

}

#private method to get storage show bridge output
sub _get_storage_show_bridge {
    my $self = shift;

    if ($self->L_is_vsim()) {
        $Log->comment(
            "This is a vsim, don't show storage bridge since there isn't one."
        );
        return;

    } else {

        if (my $wait_time = $self->_node_down_should_wait()) {
            $Log->comment('It looks like we are trying to repeatedly call' .
                ' storage bridge show on a node that is down while in bad' .
                ' start mode.' .
                " Wait $wait_time seconds before trying again");
            return;
        }

        my $bridge;
        my $count = 5;
        while (not defined $bridge and $count > 0) {
            try {
                $bridge = $self->_mcc_node->get_nodescope_apiset()->
                    storage_show_bridge();
            } catch NATE::BaseException with {
                $Log->warn("node " .  $self->_mcc_node()->get_name() .
                           " may be down, can't call storage show bridge...");
            };
            $count--;
        }
        return $bridge;
    }
}

sub L_get_storage_show_bridge {
    my $self = shift;

    #poorly named internal methods...
    return $self->_init_storage_show_bridge_response();
    #return $self->_get_storage_show_bridge();
}


#private method to get storage show switch output
sub _get_storage_show_switch {
    my $self = shift;

    if ($self->L_is_vsim() ) {
        $Log->comment(
            "This is a vsim, don't show storage switch since there isn't one."
        );

        return;
    } else {

        if (my $wait_time = $self->_node_down_should_wait()) {
            $Log->comment('It looks like we are trying to repeatedly call' .
                ' storage switch show on a node that is down while in bad' .
                ' start mode.' .
                " Wait $wait_time seconds before trying again");
            return;
        }

        my $switch;
        my $count = 5;
        while (not defined $switch and $count > 0) {
            try {
                $switch = $self->_mcc_node->get_nodescope_apiset()->storage_show_switch();
            } catch NATE::BaseException with {
                $Log->warn("node " .  $self->_mcc_node()->get_name() .
                           " may be down, can't call storage show switch...");
            };
            $count--;
        }
        return $switch;
    }
}

sub L_get_storage_show_switch {
    my $self = shift;

    return $self->_get_storage_show_switch();
}


# determine the role (A1|A2|B1|B2) of this node in the dr group network topology
sub _determine_network_topology_role {
    my $trace = NACL::MTask::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;

    my $role = undef;
    my $A1_node = $self->_dr_group()->get_A1_node();
    if ($self->_mcc_node() == $A1_node) {
        $role = 'A1';
    } else {
        my @dr_partnerships = @{ $self->_dr_group()->get_DR_partnerships() };
        foreach my $partnership (@dr_partnerships) {
            if ($partnership->get_partner_node(site => 'A') == $A1_node) {
                if ($self->_mcc_node() ==
                                 $partnership->get_partner_node(site => 'B')) {
                    $role = 'B1';
                }
            } else {
                if ($self->_mcc_node() ==
                                 $partnership->get_partner_node(site => 'A')) {
                    $role = 'A2';
                } elsif ($self->_mcc_node() ==
                                 $partnership->get_partner_node(site => 'B')) {
                    $role = 'B2';
                }
            }
        }
    }
    unless (defined($role)) {
        my $msg = q{Could not determine node [} .
                  $self->_mcc_node()->get_name() . "]'s DR Group role\n" .
                  q{    (e.g., 'A1', 'A2', 'B1', or 'B2').};
        NACL::MTask::MCC::Exceptions::BaseException->throw($msg);
    }

    return $role;
}

=head2 L_is_initialized()

Returns true (1) if we can tell if this node is initialized.
This mostly just tells us if this node is up or if there was any error
doing sysconfig for some reason.

This is mainly here to help bringing things up in a bad/non-normal state

used for:
if it's not inited, we can't tell if this node is a vsim directly (as that
comes from sysconfig)

=cut

sub L_is_initialized {
    my $trace = NACL::MTask::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;

    if (ref $self->_init_sysconfig_va_response()) {
        return 1;
    } else {
        return 0;
    }

}

=head2 L_is_vsim()

Returns true (1) if this node is a node in a vsim 4-pack DR Group.  False
otherwise.

This is here because the sysconfig info is here.  Maybe it should be straight
off in Node instead of Network::Node.

=cut

sub L_is_vsim {
    my $trace = NACL::MTask::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;
    # seems only 1 entry, but for good style...
    if (ref $self->_init_sysconfig_va_response()) {
        my $entries = $self->_init_sysconfig_va_response()->get_parsed_output();
        foreach my $entry (@{ $entries }) {
            if (exists($entry->{slot_information})) {
                my @slot_entries = @{ $entry->{slot_information} };
                foreach my $slot_entry (@slot_entries) {
                    if (exists($slot_entry->{model_name}) &&
                       ($slot_entry->{model_name} =~ /SIMBOX/)) {
                        return 1;
                    }
                }
            }
        }
    } else {
        #site may be down.  Try seeing if partner knows.
        #no, this can call dr groups before they are defined for down node,
        #we can depend on metrocluster()->is_vsim() calling all up nodes to
        #figure this out (and it won't call us for a node that's down)
        #return $self->_dr_group()->_metrocluster()->is_vsim();
        return;

    };

    return 0;
}

# Returns true if this node has any associated switches

sub L_has_switches {
    my $self = shift;

    my $resp = $self->_init_storage_show_switch_response();

    return $resp->{'output'} !~ /^No switches found/;

} ## end sub L_has_switches

sub _parse_bridge_connections {
    my $self = shift;

    my @connections;

    my $bridge_entries;
    if (defined $self->_init_storage_show_bridge_response()) {
        $bridge_entries =
            $self->_init_storage_show_bridge_response()->get_parsed_output();
    }
    ENTRY: foreach my $entry (@{ $bridge_entries }) {
        if (exists($entry->{bridge})) {
            my ($switch_name,$port_descriptor) =
                $entry->{bridge} =~ /([^:]+):([^.]+)\./
                    or next ENTRY;
            my $vendor = $entry->{vendor};
            $Log->comment(
                "bridge switch: $switch_name port: $port_descriptor");
            push(@connections, {
                               switch_name     => $switch_name,
                               port_descriptor => $port_descriptor,
                               #do we want vendor? no wwn though
                               vendor          => $vendor,
                               #port_wwn        => $port_wwn,
                                                 } );

        }
    }

    return \@connections;

}


sub L_has_bridges {
    my $self = shift;


    my $resp = $self->_init_storage_show_bridge_response();

    return $resp->{'output'} !~ /^No bridges found/;

}


=head2 get_role()

    my $node_role = $mcc_network_node->get_role();

Returns the node's role in its DR group ('A1'|'A2'|'B1'|'B2').  Here A1 and
B1 are DR partners, and A2 and B2 are DR partners.

=cut

sub get_role {
    my $trace = NACL::MTask::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;

    return $self->_dr_group_topology_role();
}

=head2 get_shared_storage_connection_info()

    my $network_node = $mcc_node->L_get_network_node_functionality();
    my @fcvi_connection_info = @{ $network_node->
                    get_shared_storage_connection_info()->{fcvi_connections} };

Returns a hashref of data with the following format:

{
    fcvi_connections =>
        [
            {
                switch_name     => 'dpg-mcc-brcd6510-01-A-fabY',
                port_descriptor => '0',
                port_wwn        => '21:00:00:24:ff:32:06:90',
            },
            {
                switch_name     => 'dpg-mcc-brcd6510-01-A-fabX',
                port_descriptor => '0',
                port_wwn        => '21:00:00:24:ff:32:06:91',
            },
            ...
        ],
    storage_connections =>
        [
            {
                switch_name     => 'dpg-mcc-brcd6510-01-A-fabY',
                port_descriptor => '0',
                port_wwn        => '21:00:00:24:ff:32:06:90',
            },
            {
                switch_name     => 'dpg-mcc-brcd6510-01-A-fabX',
                port_descriptor => '0',
                port_wwn        => '21:00:00:24:ff:32:06:91',
            },
            ...
        ],
    fibrebridge_connections =>
        [
            {
              'switch_name' => 'dpg-mcc-brcd6510-01-A-fabX',
              'port_descriptor' => '7',
              'vendor' => 'ATTO'
            },
            {
              'switch_name' => 'dpg-mcc-brcd6510-01-B-fabX',
              'port_descriptor' => '7',
              'vendor' => 'ATTO'
            },
            ...
        ],
    ha_connections =>
        [
            # empty/undefined for now
        ],
}

=cut

sub get_shared_storage_connection_info {
    my $trace = NACL::MTask::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;

    my $info = { fcvi_connections        => [ ],
                 storage_connections     => [ ],
                 fibrebridge_connections => [ ],
                 ha_connections          => [ ], };

    # seems only 1 entry, but for good style...

    my $sysconfig = $self->_init_sysconfig_va_response();
    my $entries = $sysconfig ?
                  $self->_init_sysconfig_va_response()->get_parsed_output() :
                  undef;
$Log->comment('sysconf dump' . Dumper $entries);
    foreach my $entry (@{ $entries }) {
        if (exists($entry->{slot_information})) {
            my @slot_entries = @{ $entry->{slot_information} };
            SLOT: foreach my $slot_entry (@slot_entries) {
                if (defined($slot_entry->{device})) {
                    if (   ($slot_entry->{device} =~ /FCVI Host Adapter/) &&
                           ($slot_entry->{physical_link} eq 'UP') &&
                           (defined $slot_entry->{switch_port})
                        ) {
                        my ($switch_name, $port_descriptor) =
                              ($slot_entry->{switch_port} =~ /^([^:]+):(.+)$/)
                                  or next SLOT;

                        my $port_wwn = $slot_entry->{fc_node_name};
                        push(@{ $info->{fcvi_connections} }, {
                                           switch_name     => $switch_name,
                                           port_descriptor => $port_descriptor,
                                           port_wwn        => $port_wwn,
                                                             } );
                    } elsif ($slot_entry->{device} =~ /FC Host Adapter.+<UP>/) {
                        my ($switch_name, $port_descriptor) =
                              ($slot_entry->{switch_port} =~ /^([^:]+):(.+)$/)
                                  or next SLOT;
                        my $port_wwn = $slot_entry->{fc_port_name};
                        push(@{ $info->{storage_connections} }, {
                                           switch_name     => $switch_name,
                                           port_descriptor => $port_descriptor,
                                           port_wwn        => $port_wwn,
                                                             } );
                    }
                }
            }
        }
    }

    $info->{'fibrebridge_connections'} = $self->_parse_bridge_connections();

    return $info;
}


=head2 get_sas_host_adapters()

    my @sas_adapters = @{ $network_node->get_sas_host_adapters() }

Returns information on SAS host adapters from sysconfig. This is in the form
of an array of hashes in the form:

 [
   {
     'length' => '5',
     'shelf_ids' => ['21', '22'],
     'name' => '5d'
   },
]

=cut

sub get_sas_host_adapters {
    my $self = shift;

    my $sysconfig = $self->_init_sysconfig_va_response()->get_raw_output();

    my %adapters;

    my $adapter_name;

    open(my $sysconfig_fh, '<', \$sysconfig)
        or die "Couldn't open scalar: $!";

    LINE: while (my $line = <$sysconfig_fh>) {

        if ($line =~ /SAS\s+Host\s+Adapter\s+(\w+)/) {
            $adapter_name = $1;
            $Log->debug("Adapter name is $adapter_name");
            next;

            # Gather the cable length on that SAS adapter
        } elsif ($line =~ /QSFP\s+Type:.*\s+(\d+)m\s+ID:/) {
            defined $adapter_name
                or die "Parse error: QSFP Type before host adapter\n $line\n";
            if (defined $adapters{$adapter_name}{'length'}) {
                next LINE;
            }
            $Log->debug("Adapter length for '$adapter_name' is $1");
            $adapters{$adapter_name}{'length'} = $1;
            next;

            # Gather shelf IDs on that adapter

        } elsif ($line =~ /(\d+)[.]\d+.*NETAPP.*B\/sect/) {
            defined $adapter_name
                or die
                "Parse error: *NETAPP.*B\/sect before host adapter\n $line\n";
            $adapters{$adapter_name}{'shelf_ids'}{$1} = 1;
        } else {

            # Otherwise it's an unrelated line... move on
        }

    }

    return [
        map {
            {   name      => $_,
                length    => $adapters{$_}{'length'},
                shelf_ids => [sort keys %{$adapters{$_}{'shelf_ids'}}]
            }
            } keys %adapters
    ];

} ## end sub get_sas_host_adapters

=head2 get_sas_host_adapters_local_remote()

  my ($local, $remote) =
     $network_node->get_sas_host_adapters_local_remote();

Returns a list of SAS host adapters separated into local and remote
adapters based on cable lengths.

=cut

sub get_sas_host_adapters_local_remote {
    my $self = shift;

    my (@adapters) = @{ $self->get_sas_host_adapters() };

    my %by_length;

    foreach my $adapter ( @adapters ) {
        my $length = $adapter->{'length'};

        push(
            @{ $by_length{$length} },
            $adapter->{'name'}
        );
    }

    my @lengths = keys %by_length;

    if (@lengths == 1) {
        NATE::BaseException->throw(
            "There is only one length of SAS: $lengths[0]"
        );
    }

    my $middle = scalar(@lengths) / 2;

    my @local_adapters =
        map { @$_ } @by_length{ @lengths[ 0 .. ( $middle - 1 ) ] };

    my @remote_adapters =
        map { @$_ } @by_length{ @lengths[ ($middle ) .. $#lengths ] };


    return (\@local_adapters, \@remote_adapters);

}

=head2 get_intercluster_lifs()

Not yet Implemented.

    my @ic_lifs = @{ $network_node->get_intercluster_lifs() };

Returns an arrayref of NACL::MTask::MCC::Network::LIF objects associated with
intercluster lifs on this node.

=cut

sub get_intercluster_lifs {
    my $trace = NACL::MTask::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;

    return $TO_BE_DETERMINED;
}


1;

__END__
