# $Id $

# this is here due to our standard usage of @_ in Params::Validate methods.
##no critic (RequireArgUnpacking)
##no critic (ControlStructures::ProhibitUnlessBlocks)

#
# Copyright (c) 2011-2021 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary MCC Network::Topology Stub MTask Module
## @author dl-mcc-dpg-qa@netapp.com
## @status shared
## @pod here

package NACL::MTask::MCC::Network::Topology;

use strict;
use warnings;
use Carp;

use base qw (NACL::MTask::MTask);

use NACL::MTask::MCC::Network::Fabric ();
use NACL::MTask::MCC::Network::Node ();

use NACL::Exceptions::NoElementsFound qw(:try);
use NATE::BaseException qw(:try);

use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();

use Params::Validate
    qw(validate_with BOOLEAN HASHREF OBJECT ARRAYREF SCALAR UNDEF);
use Readonly;
use Data::Dumper;


Readonly my $TO_BE_DETERMINED => undef;


=head1 NAME

NACL::MTask::MCC::Network::Topology

=head1 DESCRIPTION

This task is to be used for C-Mode MetroCluster testing.  This class defines
the network hardware mapping of an MCC DR Group.  It is used to gain access
to network entities such as fabrics, switches, ports, etc.


    # Start test
    use NACL::MTask::MCC::Metrocluster;
    ...

    my $topology = $mcc->get_DR_groups()->{$group_id}->get_network_topology();
    my $fabric = $topology->get_intersite_fabrics()->[0];
    my $switch = $fabric->get_switch(site => 'A');

    ...
    # Perform test here
    ...

    # End test

=cut

=head1 Public METHODS

There is no public constructor for this class.  A
NACL::MTask::MCC::Network::Topology object is obtained via calls such as

    my $topology = $mcc->get_DR_groups()->{$group_id}->get_network_topology();

=cut

# automatically generate a (Non-public) constructor (includes a call to 'init')
# and (Non-public) member accessor methods
use Class::MethodMaker [
    new    => [ '-init', 'new' ],
    scalar => [
                { -type => 'NACL::MTask::MCC::DR_Group' },
                '_dr_group',
              ],
    scalar => [
        qw/ _fabrics _fiber_bridges _nodes_view_of_fabric /,
              ],
                        ];

# validate the parameters passed in and what else?
sub init {
    $Log->enter() if $may_enter;
    my $self = shift;

    my %opts = validate_with(
        params => \@_,
        spec   => {
            dr_group     => { isa   => 'NACL::MTask::MCC::DR_Group' },
        },
                            );

    $self->_dr_group($opts{'dr_group'});
    $self->_nodes_view_of_fabric($self->_discover_nodes_view_of_fabric());
    $self->_fabrics( $self->_discover_intersite_fabrics());

    $Log->exit() if $may_exit;
    return;
}

=head2 _discover_nodes_view_of_fabric()

This method gets connection information from all MCC::Network::Node
objects and constructs a view of ports that it thinks it is connected
to on the switches in the fabric.

This is important (now) for discovering the fabric because there
may be some switch ports that aren't used for the MCC node or have
various unknown network adapter names in the configuration.
Or, if a switch is shared between 2 or more testbeds there may
be some ports on the switch that are not configured in the standard
way and fail the standard parsing regexes (for example if the
other testbed is being gxbooted so the ports it uses don't
conform -- ie they don't list the node names as expected.)

Anyway, if we know what ports the testbed nodes care about then we
can use that information to make more intelligent decisions about
what to do when unknown switch ports are found.  If we can parse
them as we do now then great, if not then we will have to either
ignore or fail depending on if it's a port the node wants to use.

=cut

sub _discover_nodes_view_of_fabric {
    $Log->enter() if $may_enter;
    my $self = shift;

    my $node_view;

    my $A1_net_node = $self->_dr_group()->get_A1_node()->
                                            L_get_network_node_functionality();
    my $storage_net_info_as_seen_by_A1 =
                            $A1_net_node->get_shared_storage_connection_info();

    my $network_nodes = $self->get_network_nodes();

	foreach my $node (keys %{$network_nodes}) {
		my $connections =
		    $network_nodes->{$node}->get_shared_storage_connection_info();
        $Log->comment("connections for $node:" . Dumper $connections);
        foreach my $conn_type (keys %{$connections} ) {
            foreach my $conn ( @{ $connections->{$conn_type} } ) {
				my $switch = $conn->{switch_name};
				my $port   = $conn->{port_descriptor};

				if ($switch && defined $port) {
				  $Log->comment("OMG!@!") if $node_view->{$switch}->{$port};
				  #$node_view->{$switch}->{$port} = $conn;
				  foreach my $item (keys %{ $conn } ) {
				      $node_view->{$switch}->{$port}->{$item} =
					      $conn->{$item};
				  }

				  (my $type = $conn_type) =~ s/_connections//;
				  $node_view->{$switch}->{$port}->{type} = $type;

				  #save which node cares about this
				  #TODO maybe change this to hash instead?
				  push @{$node_view->{$switch}->{$port}->{from_node}}, $node;
			  }
            }
		}

	}
$Log->comment('node view of fabric:' . Dumper $node_view);
    $Log->exit() if $may_exit;
    return $node_view;
}

=head2 get_nodes_view_of_fabric()

    my $nodes_view = $topology->get_nodes_view_of_fabric();

    Returns a hashref with this format:
    {
        'dpg-mcc-brcd6510-01-A-fabX' => {
            '8' => {
                      'from_node' => [
                                       'A2'
                                     ],
                      'switch_name' => 'dpg-mcc-brcd6510-01-A-fabX',
                      'port_descriptor' => '8',
                      'port_wwn' => '21:00:00:24:ff:31:ec:eb',
                      'type' => 'fcvi'
                    },
             '1' => {
                      'from_node' => [
                                       'A1'
                                     ],
                      'switch_name' => 'dpg-mcc-brcd6510-01-A-fabX',
                      'port_descriptor' => '1',
                      'port_wwn' => '2:101:001b32:aeba8d',
                      'type' => 'storage'
                    },
             '7' => {
                      'from_node' => [
                                       'A1',
                                       'B1',
                                       'A2',
                                       'B2'
                                     ],
                      'switch_name' => 'dpg-mcc-brcd6510-01-A-fabX',
                      'port_descriptor' => '7',
                      'type' => 'fibrebridge',
                      'vendor' => 'ATTO'
                    },
             ....
           },
        'dpg-mcc-brcd6510-01-B-fabY' => {
             '8' => {
             ....
	       }
        ....
    }


    Basically this method returns the view of the fabric ports as seen
	by the nodes.  In this way we can validate port nodes based on this
	side as well. (And it can help with discovery in case some port
	nodes don't parse as expected or there are other issues)

=cut

sub get_nodes_view_of_fabric {
    $Log->enter() if $may_enter;
    my $self = shift;
    $Log->exit() if $may_exit;
	return $self->_nodes_view_of_fabric;
}


=head2 _discover_intersite_fabrics()

This method performs a discovery operation to determine the intersite fabrics
associated with this DR Group network topology.  Please see

http://wikid.netapp.com/w/MCC_Automation/MCC_DPG_QA_Automation/Test_Library_Dev#MCC_Network_Topology_Discussion_and_Notes

for (hopefully) clarification on what such a fabric is.

Returns an arrayref containing NACL::MTask::MCC::Network::Fabric objects.

Designer's Notes: This topology will consist of two redundant fabrics which
span between sites A and B.  Each of these fabrics will consist of a pair
of real or virtual switches, one at each site and connected to each other via
their ISLs.  Since these switches may be running any number of virtual
switches that may be used either in multiple DR Groups in the same MCC system,
or by multiple MCC systems, we need to discover which switches (real or
virtual) link to each other to form a particular fabric.  This method itself
makes use of the NACL::MTask::MCC::Network::Switch constructor which specifies
the parameter 'vswitch_fcvi_port_spec' in order to determine a switch that is
being used by a particular filer node (A1 in the DR Group owning this topology.)

This ties one end of one fabric to this shared storage topology.  In order to
find the Switch on the other end (i.e., at site B) of this fabric, we will pass
the site A switch to the NACL::MTask::MCC::Network::Fabric constructor.  The
NACL::MTask::MCC::Network::Fabric constructor will make use of the
NACL::MTask::MCC::Network::Switch constructor which specifies the parameter
'vswitch_isl_target_wwn' whose value can be found by querying the site A
switch.  See the NACL::MTask::MCC::Network::Fabric constructor for more
details on this process.

In the end, each fabric will be composed of the pair of linked real or virtual
switches spanning sites A and B.

This shared storage network topology is designed so that ALL FABRIC AND SWITCH
OPERATIONS ON THESE PAIRS OF VIRTUAL SWITCHES WILL MONITOR OR AFFECT ONLY THE
PORTS ASSOCIATED WITH THESE VIRTUAL SWITCHES.  ALL OTHER VIRTUAL SWITCHES
RUNNING ON THE ASSOCIATED REAL SWITCHES WILL REMAIN UNAFFECTED AND UNSEEN
WITHIN THIS TOPOLOGY.

If this is found to not be the case at any point, then this class must be fixed
so that the above statement is true.

=cut

sub _discover_intersite_fabrics {
    $Log->enter() if $may_enter;
    my $self = shift;

    my $A1_net_node = $self->_dr_group()->get_A1_node()->
                                            L_get_network_node_functionality();
    my $storage_net_info_as_seen_by_A1 =
                            $A1_net_node->get_shared_storage_connection_info();

    my $network_nodes = $self->get_network_nodes();

	foreach my $node (keys %{$network_nodes}) {
		my $connections =
		    $network_nodes->{$node}->get_shared_storage_connection_info();
        $Log->comment("connections for $node:" . Dumper $connections);
	}

    # A1 should see 2 switches at site A- one for each fabric
    my @fabrics = ();
    # track which switches have already been used to discover a fabric
    my %site_A_switches_done = ();
    my $switch_model = '';
    foreach my $fcvi_connection_info
                   (@{ $storage_net_info_as_seen_by_A1->{fcvi_connections} }) {
        my $site_A_switch_name = $fcvi_connection_info->{switch_name}
            or next;

        # don't create a fabric twice (in case there are multiple fcvi conns)
        if (exists($site_A_switches_done{$site_A_switch_name})) {
            next;
        } else {
            $site_A_switches_done{$site_A_switch_name} = 'done';
        }
		#TODO pass down the hash from Topology to just switch A
		#creation so that it can have access to the node_view_of_switches
		#thing....since there is sort of a cycle of news here. urg
        my $site_A_switch =
            NACL::MTask::MCC::Network::Switch->new(
                switch_name            => $site_A_switch_name,
                node_target_ports_spec => $storage_net_info_as_seen_by_A1,
				nodes_view_of_fabric   => $self->get_nodes_view_of_fabric(),
			);

        # Find the model of the switch to see if we'll allow a single fabric.
        $switch_model = $site_A_switch->get_model();

        # create the fabric object
        push(@fabrics, NACL::MTask::MCC::Network::Fabric->new(
				                             topology      => $self,
                                             site_A_switch => $site_A_switch));
    }

    # standard MCC shared storage network has 2 intersite fabrics
    # If we have an MCC with Cisco 9710s then a single fabric is supported
    if (scalar(@fabrics) == 1 and $switch_model =~ /9710/) {
        my @modules = $fabrics[0]->_site_A_switch()->get_modules();
        if (scalar(@modules) != 2) {
            local $Data::Dumper::MaxDepth = 2;
            NACL::MTask::MCC::Exceptions::BaseException->throw(
                "Not enough fabrics: detected " .
                scalar(@modules) . " modules, should be 2.\n" .
                "Fabric details:\n" . Dumper(@modules) . Dumper(@fabrics));
        }
    }
    # elsif (scalar(@fabrics) < 2) {
    #     local $Data::Dumper::Maxdepth = 2;
    #     NACL::MTask::MCC::Exceptions::BaseException->throw(
    #               "Not enough fabrics: detected " .
    #                   scalar(@fabrics) .
    #                       " fabrics, should be 2." .
    #                           " Fabric details:\n" .
    #                               Dumper(@fabrics));
    # }
    # elsif (scalar(@fabrics) != 2) {
    #     local $Data::Dumper::Maxdepth = 2;
    #     NACL::MTask::MCC::Exceptions::BaseException->throw(
    #               "Incorrect number of fabrics: detected " .
    #                   scalar(@fabrics) .
    #                       " fabrics, should be 2. Fabric details:\n" .
    #                           Dumper(@fabrics));
    # }

    $Log->exit() if $may_exit;
    return \@fabrics;
}

=head2 get_network_nodes()

    my $network_nodes = $topology->get_network_nodes();

Returns a hashref with this format:

    {
        A1 => $A1_net_node,
        A2 => $A2_net_node,
        B1 => $B1_net_node,
        B2 => $B2_net_node,

    }

where all $XX_net_node are NACL::MTask::MCC::Network::Node objects.

The nodes are defined as follows:

    A1 is the node that was passed as the A1 node for this DR Group.
    B1 is A1's DR Partner
    A2 is A1's HA Partner
    B2 is A2's DR Partner

=cut

sub get_network_nodes {
    $Log->enter() if $may_enter;
    my $self = shift;

    # these are MCC::Nodes
    my $nodes_by_role = $self->_dr_group()->get_nodes_by_role();

    # want the MCC::Network::Nodes
    my %net_nodes_by_role =
        map { $_ => $nodes_by_role->{$_}->L_get_network_node_functionality() }
                                                     keys(%{ $nodes_by_role });

    $Log->exit() if $may_exit;
    return \%net_nodes_by_role;
}

=head2 get_intersite_fabrics()

    my $fabrics = $topology->get_intersite_fabrics();

Returns a hashref with the following format:

    {
      'X' => $fabric_X,
      'Y' => $fabric_Y,
      . . .
    }

where the $fabric_C are NACL::MTask::MCC::Network::Fabric objects.  Bear in
mind that the order of these fabrics is arbitrary and does not indicate which
nodes access which fabrics in any primary fashion.  The fabrics are completely
interchangable when working as intended.

=cut

sub get_intersite_fabrics {
    $Log->enter() if $may_enter;
    my $self = shift;

    $Log->exit() if $may_exit;
    return $self->_fabrics();
}

=head2 get_fiber_bridge_arrays()

TBD: THIS METHOD NOT YET IMPLEMENTED.  BURT: 592069

    my $fiber_bridges = $topology->get_fiber_bridge_arrays();

Returns a hashref with this format:

    {
      'X' => {
                 'A' => [ $fiber_bridge_0, $fiber_bridge_1, . . . ],
                 'B' => [ $fiber_bridge_0, $fiber_bridge_1, . . . ]
             },
      'Y' => {
                 'A' => [ $fiber_bridge_0, $fiber_bridge_1, . . . ],
                 'B' => [ $fiber_bridge_0, $fiber_bridge_1, . . . ]
             },
    }

where X/Y and A/B are the fabric and site specification needed to define
which array of fiber switches are targetted, and all $fiber_bridge_n are
NACL::MTask::MCC::Network::Fiber_Bridge objects.

See the diagrams at

http://wikid.netapp.com/w/MCC_Automation/MCC_DPG_QA_Automation/Test_Library_Dev#MCC_Network_Topology_Discussion_and_Notes

for further clarification.

=cut

sub get_fiber_bridge_arrays {
    $Log->enter() if $may_enter;
    my $self = shift;

    my $a1_node = $self->_dr_group->get_A1_node();

    my $apiset = $self->get_nodescope_apiset();

    # Largely derived from Mandar's code
    my (%adapter_href, @adapters, @local_adapters, @remote_adapters, @local_shelves, @remote_shelves);

   # Gather disk adapters and shelf details
    my $response = $apiset->storage_show_disk(
                                                                specific => 1 );
    my @storage_show_disk_lines = split qr{\n}, $response->get_raw_output();
    foreach my $line ( @storage_show_disk_lines ) {
        # Gather list of online FC adapter names
        if ( $line =~ /^(\d\w)\.\S+\s+(\d+)\s+\d+\s+\S+\s+NETAPP/ ) {
            $adapter_href{$1}->{$2} = q{};
            if ( none { $_ eq $1 } @adapters ) {
                push @adapters, $1;
            }
        }
    }

    # Find unique adapters, viz. adapters that have first path to the shelf.
    # Skip the adapters that have second path to the shelf
    push my @uniq_adapters, $adapters[0];
    foreach my $adapter ( @adapters ) {
        # Skip if this adapter is already a unique adapter
        next if (any { $_ eq $adapter } @uniq_adapters);

        # Skip if there is a match of shelf on adapter with shelf on any
        # unique adapter
        my $match_found = 0;
        foreach my $uniq_adptr ( @uniq_adapters ) {
            my @adapter_shelves = keys %{$adapter_href{$adapter}};
            my @uniq_adptr_shelves = keys %{$adapter_href{$uniq_adptr}};
            if ( any { $_ eq $adapter_shelves[0] } @uniq_adptr_shelves ) {
                $match_found = 1;
                next;
            }
        }

        # Select the current adapter as uniq if its shelves did not match any
        # shelves on previously selected uniq adapters
        if (! $match_found ) {
            push @uniq_adapters, $adapter;
        }
    }

    # Sort the uniq adapters list
    @uniq_adapters = sort @uniq_adapters;

    # Divide top half of uniq adapters as local and bottom half as remote
    for my $index ( 0 .. $#uniq_adapters ) {
        if ($index le $#uniq_adapters/2) {
            push @local_adapters, $uniq_adapters[$index];
        } else {
            push @remote_adapters, $uniq_adapters[$index];
        }
    }

    # Gather local and remote shelf details
    @local_shelves = map { keys %{$adapter_href{$_} } } @local_adapters;
    @remote_shelves = map { keys %{$adapter_href{$_} } } @remote_adapters;

    return \@local_shelves, \@remote_shelves;

}

1;
