#
# Copyright (c) 2001-2015 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary This wrapper library for functionality already exported by NACL
## @author dl-mcc-dpg-qa-auto@netapp.com
## @status ready

package NACL::MTask::MCC::Traffic::TestbedSetupHelper;

use strict;
use warnings;
use Tharn;
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Data::Dumper;
use Params::Validate qw(validate validate_with
                        HASHREF ARRAYREF SCALAR UNDEF BOOLEAN OBJECT);

use NACL::C::SystemNode;
use NACL::C::VserverNfs;
use NACL::C::VserverExportPolicyRule;
use NACL::C::VolumeSnapshotAutodelete;
use NACL::C::VserverServicesDns;
use NACL::C::VserverCifsSuperuser;
use NACL::C::VserverIscsi;

use NACL::CS::NetworkInterface;
use NACL::CS::NetworkConnectionsListening;
use NACL::CS::VserverCifs;

use NACL::STask::VserverCifsShare;
use NACL::STask::NetworkInterface;
use NACL::STask::VserverNfs;
use NACL::STask::VserverServicesDns;
use NACL::STask::VserverCifsShare;
use NACL::STask::NetworkInterface;
use NACL::STask::Lun;
use NACL::STask::VserverCifsShareProperties;

use NACL::MTask::CIFS;
use NACL::MTask::MCC::Exceptions::BaseException;

use NACL::GeneralUtils qw(random_name_generator rand_words);
use NACL::Cleanup;


=head1 NAME

NACL::MTask::MCC::TestbedSetupHelper

=head1 DESCRIPTION

B<EXPERIMENTAL LIBRARY!>

TestbedSetupHelper is a wrapper library for the functionality already exported
by NACL to provide minimum lines of code writting in "thpl" scripts which is
required to call same function in multiple times from different scenarios
when Metrocluster Functional testing.

DISCLAIMER: This library is ONLY for internal use by the dpg-mcc-qa-auto team
            and will not be supported for ANY EXTERNAL users.

    #Start test
    use NACL::MTask::MCC::TestbedSetupHelper;
    ...

    my $obj =  NACL::MTask::MCC::TestbedSetupHelper->new (
                     nodes    => arrayref_of_NACL::C::Node_objects,
                     vservers => arrayref_of_vserver_objects_or_name,
                     label    => 'Any string' );

    ...

    #Perform test here
    ...

    #End test
    ....

=cut

=head1 METHODS

=head2 new (...)

=over

=item Options

=over

=item C<< nodes => one or mode NACL::C::Node objects in the mcc system >>

(Required.) An arrayref containing one or more node objects in an MCC system.
Node object should be of type : NACL::C::Node.

=back

=over

=item C<< vservers => one or more vserver names/objects in the mcc system >>

(Required.) An arrayref containing one or more vserver names/objects in an
MCC system which belong to the nodes passed in the 'nodes' parameter.
If a vserver is passed that does not belong to one or more nodes in the
'nodes' parameter the library will throw an exception.
If a vserver object is passed it should be of type : NACL::CS::Vserver.

=back

=over

=item C<< label => any user defined string label >>

(Optional) Any user defined string that can be attached to the object
Default: "No Label"

=back

=back

=cut

use Class::MethodMaker [
    scalar  => [ qw/ _iscsi_service_object_list
                     _aggregates_list
                     _nfs_volume_list
                     _cifs_volume_list
                     _san_volume_list
                     _san_vserver_list
                     _san_lun_list
                     _cifs_share_list
                     _vserver_list
                     _nodes_list
                     _cifs_username
                     _cifs_password
                     _igroup_list
                     _dummy_resource_list
                     _cleanup_object
                     _doCIFS
                     _doNFS
                     _doSAN
                     _is_mcc
                     _is_2_node
                     _skip_cleanup
                     _label
                   / ],
    new     => [ '-init', 'new' ],
];

sub init {
    my $self = shift;
    my %opts = validate_with(
        params => \@_,
        spec   => {
            nodes        => { type => ARRAYREF, },
            vservers     => { type => ARRAYREF, },
            skip_cleanup => { type => SCALAR, default => 0 },
            label        => { type => SCALAR, default => 'No Label' },
        },
    );


    #check if the nodes have been passed in and are of the expected type
    my $nodes = delete ($opts{'nodes'});
    my $vservers = delete ($opts{'vservers'});

    # Set the object label
    $self->_label( delete $opts{'label'} );

    unless (scalar(@{ $nodes }) ||
            scalar @{ $vservers }) {
         NACL::MTask::MCC::Exceptions::BaseException->throw(
               'Invalid constructor call. A nodes list and '
             . 'a vserver list must be passed');
    }

    #verify nodes passed are objects and not scalar
    foreach my $node (@{ $nodes }) {
        unless (ref($node) && $node->isa('NACL::C::Node')) {
           NACL::MTask::MCC::Exceptions::BaseException->throw(
                   'Invalid param type passed for nodes. '
                 . 'Expected type NACL::C::Node');
        }
    }

    $self->_nodes_list($nodes);

    #if a vserver name has been passed in instead of a vserver object
    #create an object out of it

    my $vs_list;

    foreach my $vs (@{ $vservers }) {
        unless (ref($vs) && $vs->isa('NACL::CS::Vserver')) {
            #We are doing this excercise in case an user has called the object
            #with an incorrect combination of node names and vserver.
            #e.g. Calling object with nodes [a1, b1] and vserver [vs1].
            #vs1 is not a valid vserver for b1 and the library should fail
            $Log->comment('Getting vserver object for: ' . $vs);
            my $obj;
            foreach my $node (@{  $self->_nodes_list() }) {
                my $vs_command_interface = $node;
                $Log->comment('The node name for fetching vserver is: ' .
                              $vs_command_interface->name());
                try {
                    $obj = NACL::CS::Vserver->fetch(
                              command_interface   => $vs_command_interface,
                              filter              => {
                                                 subtype       => 'sync-source|default',
                                                 'admin-state' => 'running',
                                                 vserver       => $vs
                                    },);
                } catch NACL::Exceptions::NoElementsFound with {
                    my $err_txt = 'No matching vserver\(s\) found';
                    my $exception = $_[0];
                    $Log->comment('vserver fetch failed with : '
                                . $exception->text());
                    if ($exception->text() =~ /$err_txt/i) {
                        NACL::MTask::MCC::Exceptions::BaseException->throw(
                              "No sync source veservers by the name $vs "
                            . 'were found on '
                            . $node->name());
                    } else {
                        NACL::MTask::MCC::Exceptions::BaseException->throw(
                             'Unexpected error encountered while fetching vs');
                    }
                };
            }
            $vs = $obj;
        }
        push(@{ $vs_list }, $vs);
    }

    $Log->comment(q{-} x 60);
    $Log->comment('Got vservers: ');
    foreach my $vs (@{ $vs_list }) {
        $Log->comment($vs->vserver());
    }
    $Log->comment(q{-} x 60);

    $self->_vserver_list($vs_list);

    $Log->comment('Creating the NACL::Cleanup object, a random db '
                . 'will be created in that process');
    $self->{_cleanup_object} = NACL::Cleanup->new( random_db => 1 );

    #library is called by the multi-protocol traffic which is used
    #is a non mcc environment also. When used in a non MCC system
    my $paramset = NATE::ParamSet->new( global => 1 );
    my $is_mcc =  $paramset->get('MCC', default => 1);
    $self->_is_mcc($is_mcc);

    if ($is_mcc) {
        $Log->comment(q{-} x 60);
        $Log->comment('Is this a 2 node system? ');

        $self->_is_2_node( $self->_is_2_node_system( node => ${$nodes}[0] ) );

        $Log->comment(q{-} x 60);
    } else {
        $Log->comment('Test is running on a NON MCC system!');
    }

    # Set skip cleanup flag
    $self->_skip_cleanup($opts{'skip_cleanup'});

    return;
}

=head2 create_resources (...)

This method creates the required objects which include volumes,aggreates,
shares, luns and igroups, specifications defined by user required for
starting traffic on nodes. It also will prepare the vserver which inlcude
starting NFS/CIFs/iSCSI service, creating export policies etc.

=over

=item Options

=over

=item C<< aggr_params => Hashref of aggregate specific parameters >>

(Optional.) Please look at the POD for _create_or_find_aggregates() to
see what parameters this hashref takes.

=back

=item C<< vol_params => Hashref of volume specific parameters >>

(Optional.) Please look at the POD for _create_volumes() to
see what parameters this hashref takes.

=back

=item C<< cifs_params => CIFS specific parameter that are required for
                         starting CIFS traffic >>

(Optional.) Please look at the POD for _prepare_tb_for_cifs to
see what parameters this hashref takes.

=back

=item C<< nfs_params => NFS specific parameters that are required for
                         starting NFS traffic >>

(Optional.) Please look at the POD for _prepare_tb_for_nfs to
see what parameters this hashref takes.

=back

=item C<< san_params => SAN specific parameters that are required for
                         starting iSCSI traffic >>

(Optional.) Please look at the POD for _prepare_tb_for_san to
see what parameters this hashref takes.

=back

=item C<< SAN => (0|1) If the user wants the library to prepare the testbed
                  to be able to do SAN traffic.

(Optional.) Default is 0.

=back
=item C<< NFS => (0|1) If the user wants the library to prepare the testbed
                  to be able to do NFS traffic.

(Optional.) Default is 0.

=back

=item C<< CIFS => (0|1) If the user wants the library to prepare the testbed
                  to be able to do CIFS traffic.

(Optional.) Default is 0.

=back

=item C<< Sample invocation of this method >>

$obj->create_resources(
                                   SAN   => 1,
                                   NFS   => 1,
                                   CIFS  => 1,

                     aggr_params => {
                                     aggregates_per_vs  => $aggregates_per_vs,
                                     reuse_aggrs        => $reuse_aggrs,
                                      diskcount         => $diskcount,
                               },
                     vol_params => {
                                     vols_per_aggr      => $vols_per_aggrs,
                                     per_vol_size       => $vol_size,
                               }
                     san_params => {
                                     san_clients        => $san_client_objects,
                                     san_vservers       => $san_vservers,
                                     luns_per_vol       => $lun_count,
                                     lun_size           => $lun_size,
                               },
                     cifs_params => {
                                  CIFS_USER             => $cifs_user,
                                  CIFS_PASSWD           => $cifs_pwd,
                                  shares_per_vol        => $shares_count,
                               },
                      nfs_params => {
                                 ENABLE_NFS_VER         => [ 3, 4 ],
                               },
                     );

=back

=back

=cut

sub create_resources {
    my $self = shift;

    my %args = validate_with(
                params => \@_,
                spec   => {
                    NFS          => { type => BOOLEAN, default => 0},
                    CIFS         => { type => BOOLEAN, default => 0},
                    SAN          => { type => BOOLEAN, default => 0},
                    vol_params   => { type => HASHREF|UNDEF, optional => 1 },
                    aggr_params  => { type => HASHREF|UNDEF, optional => 1 },
                    san_params   => { type => HASHREF|UNDEF, optional => 1 },
                    nfs_params   => { type => HASHREF|UNDEF, optional => 1 },
                    cifs_params  => { type => HASHREF|UNDEF, optional => 1 },
                } );

    $Log->comment(' Call to prepare filer objects');
    #First lets create or get the aggregates
    my $aggregate_list =
           $self->_create_or_find_aggregates(aggr_params => $args{aggr_params});

    $self->_aggregates_list($aggregate_list);
    $self->_doNFS($args{'NFS'});
    $self->_doCIFS($args{'CIFS'});
    $self->_doSAN($args{'SAN'});

    if ($args{'SAN'}) {

        my $nodes = $self->_nodes_list();
        #Lets enforce that two node objects (node A and its HA partner)
        #should be passed for SAN because of iSCSI multipathing in MCC.
        #We need to ensure that
        #iSCSI lifs are created on both nodes including the HA partner.
        #If that is not the case the lun mapping will fail.

        #This is only for iSCSI and we expect the call to be made in pairs of
        #2 for SAN

        #with 2 node introduced in FS.1 let us allow this if its a
        #2 node system.

        if ($self->_is_mcc() && @{$nodes} != 2 && !$self->_is_2_node() && @{$nodes} != 4) {
            NACL::MTask::MCC::Exceptions::BaseException->throw(
                        'Two nodes ~ node A and its HA partner is needed to '
                      . 'setup the testbed for SAN');
        }
        my %params;
        $params{san_params} = $args{san_params};
        $params{vol_params} = $args{vol_params};

        $self->_prepare_tb_for_san(\%params);
    }

    if ($args{'NFS'}) {

        my %params;
        $params{nfs_params} = $args{nfs_params};
        $params{vol_params} = $args{vol_params};

        $self->_prepare_tb_for_nfs(\%params);
    }

    if ($args{'CIFS'}) {

        my %params;
        $params{cifs_params} = $args{cifs_params};
        $params{vol_params} = $args{vol_params};

        $self->_prepare_tb_for_cifs(\%params);
    }


    return;

}

=head2 cleanup_resources (...)

Usage:
    use NACL::MTask::MCC::TestbedSetupHelper;
    ...

    my $obj =  NACL::MTask::MCC::TestbedSetupHelper->new (
                            nodes => arrayref_of_node_objects,
                         vservers => arrayref_of_vserver_objects_or_name,);
    ...

    $obj->cleanup_resources();
    ...

An object method which will delete all the filer objects or resources that
this library has created.

=cut

sub cleanup_resources {
    my $self = shift;

    my  %args = validate_with(
              params => \@_,
              spec   => {
                 max_connection_retry_count => { type => SCALAR, default => 3 },
            },
          );

    #return if skip_cleanup is  true
    if ($self->_skip_cleanup()) {
        $Log->comment('User has requested to skip test resource cleanup. '
                    . 'Returning ... ');
        return;
    }

    my $iscsi_service_object = $self->_iscsi_service_object_list();
    if (defined $iscsi_service_object &&
        scalar(@{ $iscsi_service_object }) > 0) {
        $Log->comment(q{-} x 60);
        $Log->comment('Deleting the iSCSI service created');
        $Log->comment(q{-} x 60);
        foreach my $obj (@{ $iscsi_service_object }) {
            try {
                $obj->stop();
            } catch NACL::APISet::Exceptions::CommandFailedException with {
                my $err_txt = 'Service, adapter, or operation not started';
                my $exception = $_[0];
                if ($exception->text() =~ /$err_txt/i) {
                    $Log->comment('iSCSI service already stopped..which is ok');
                } else {
                    NACL::MTask::MCC::Exceptions::BaseException->throw(
                                'Unexpected error encountered while stopping '
                              . 'iSCSI service: ' . $exception->text());
                }
            };
            $obj->delete();
        }
    }

    $Log->comment(q{-} x 60);
    $Log->comment('Cleaning up the filer resources');
    $Log->comment(q{-} x 60);

    my $CObj = $self->{_cleanup_object};
    $CObj->run(max_connection_retry_count => $args{max_connection_retry_count});

    return;
}

#Thanks to Peter C for suggesting this.
sub DESTROY {
    my $self = shift;

    $Log->comment('Calling the cleanup_resources() method from DESTROY');
    $self->cleanup_resources();
}

#--------------------------------------------#
#         Library private methods            #
#--------------------------------------------#

=head2 _prepare_tb_for_nfs (...)

Prepares the testbed so that the user can run NFS traffic. It does the following
- Created volumes
- Prepares the vserver to enable running NFS traffic

=item C<< nfs_params parameters >>

=over

=item C<< ENABLE_NFS_VER => Which NFS versions the users wants to enable in
                            the vserver.
(Optional.) Allowed values 3 and 4

=back

=cut

sub _prepare_tb_for_nfs {
    my $self = shift;

    my %opts = validate_with (
        params => \@_,
        spec   => {
                nfs_params => { type => HASHREF|UNDEF, optional => 1 },
                vol_params => { type => HASHREF|UNDEF, optional => 1 }
        },
    );

    my $vol_params = delete($opts{vol_params});
    my $nfs_params = delete($opts{nfs_params});

    $Log->comment('Preparing the vservers');
    $self->_prepare_vs_for_nfs(nfs_params => $nfs_params);

    $Log->comment('Creating volumes for NFS traffic');
    my $nfs_volumes = $self->_create_volumes(vol_params => $vol_params,
                                             protocol   => 'nfs');
    $self->_nfs_volume_list($nfs_volumes);

    return;
}

=head2 _prepare_tb_for_cifs (...)

Prepares the testbed so that the user can run CIFS traffic. It does the following
- Creates volumes
- Creates CIFS shares under the volumes
- Prepares vserver for CIFS traffic

=item C<< cifs_params parameters >>

=over

=item C<< CIFS_USER => Cifs user name to map cifs shares   >>

(Required. )

=back

=item C<< CIFS_PASSWD  => Cifs user's password to map cifs shares >>

(Required. )

=back

=item C<< shares_per_vol => # of shares to created per volume >>

(Required. )

=back

=cut

sub _prepare_tb_for_cifs  {
    my $self = shift;

    my %opts = validate_with (
        params => \@_,
        spec   => {
               cifs_params => { type => HASHREF|UNDEF, optional => 1 },
               vol_params  => { type => HASHREF|UNDEF, optional => 1 }
        },
    );

    my $vol_params = delete($opts{vol_params});
    my $cifs_params = delete($opts{cifs_params});

    $Log->comment('Preparing the vserver');
    $self->_prepare_vs_for_cifs(cifs_params => $cifs_params);

    $Log->comment('Creating volumes for CIFS traffic');
    my $cifs_volumes = $self->_create_volumes(vol_params => $vol_params,
                                              protocol   => 'cifs');
    $self->_cifs_volume_list($cifs_volumes);

    $Log->comment('Creating shares');
    my $shares_count = delete($cifs_params->{shares_per_vol});
    my $share_list = $self->_create_shares(shares_per_vol => $shares_count);
    $self->_cifs_share_list($share_list);

    return;
}

=head2 _prepare_tb_for_san (...)

Prepares the testbed so that the user can run iSCSI traffic. It does the following
- Creates volumes
- Creates luns
- Creates igroups
- Maps the lun to igroups
- Prepares the vserver for CIFS traffic

=item C<< san_params parameters >>

=over

=item C<< san_clients => ARRAYREF of san client objects. >>

(Optional.) This is required in order to create igroups for each
of the SAN clients. If this value it not passed igroups won't be created
and LUNs wont be mapped.

=back

=item C<< san_vservers => ARRAYREF of san vservers >>

(Optional.) If the user wants to specify which vserver to use of the list the was used while
creating the object. If a vserver is specified only that vserver will be
readied for SAN. If not all the vservers mainitained in the _vservers_list
will be used. The vservers prepared for SAN will be maintianed in
_san_vserver_list

=back

=item C<< luns_per_vol => # of luns per volume to be created >>

(Optional.) Default 0

=back

 =item C<< lun_size => The size of each lun that will be created >>

(Required.)

=back

=cut

sub _prepare_tb_for_san {
    my $self = shift;

    my %opts = validate_with (
        params => \@_,
        spec   => {
                san_params => { type => HASHREF|UNDEF },
                vol_params => { type => HASHREF|UNDEF, optional => 1 }
        },
    );

    my $san_params = delete($opts{san_params});
    my $vol_params = delete($opts{vol_params});

    $Log->comment('Preparing the vserver');
    $self->_prepare_vs_for_san(san_clients  => $san_params->{san_clients},
                               san_vservers => $san_params->{san_vservers});

    #Create vols for SAN
    $Log->comment('Creating volumes for SAN traffic');
    my $san_volumes = $self->_create_volumes(
                                    vol_params => $vol_params,
                                    vservers   => $self->_san_vserver_list(),
                                    protocol   => 'san');
    $self->_san_volume_list($san_volumes);

    #Create Luns
    $Log->comment('Creating LUNS for SAN traffic');
    my $igroups = $self->_igroup_list();
    my $san_luns = $self->_create_and_map_luns(
                                     igroups     => $igroups,
                                     lun_per_vol => $san_params->{luns_per_vol},
                                     lun_size    => $san_params->{lun_size});

    $self->_san_lun_list($san_luns);

    $Log->comment('Completed preparing the testbed for SAN');

    return;
}

=head2 _get_san_client_initiator_id (...)

Given a set of client object this method will return the intiator IQN for
the clients

=cut

sub _get_san_client_initiator_id {

    my %opts = validate_with (
        params => \@_,
        spec   => {
              clients => { type => ARRAYREF }
        },
    );

    my $clients = delete($opts{clients});
    my $initiator_ids;

    if (!defined($clients) || scalar(@{ $clients }  ) == 0) {
        $Log->comment('User did not pass any clients for SAN .. returning');
        return;
    }

    foreach my $client (@{ $clients }) {
        my  $initiator;
        if (!$client->isa('NACL::C::Client')) {
            NACL::MTask::MCC::Exceptions::BaseException->throw(
                                         'Param client type should be OBJECT.');
        }

        $Log->comment(q{-} x 60);
        $Log->comment('Gathering Intiator IDs for: ' . $client->name());
        $Log->comment(q{-} x 60);

        my $output = $client->execute(
                    command => 'cat /etc/iscsi/initiatorname.iscsi'
                );

        if ($output =~ /^InitiatorName\s*=\s*([-.:\d[:alpha:]]+)/m) {
            $initiator = lc $1; # Node names are not case-sensitive
        }

        if (!defined($initiator)) {
            NACL::MTask::MCC::Exceptions::BaseException->throw(
                'Client: ' . $client->name() . ' did not return a valid '
              . 'iqn. This could mean ISCSI is not correctly installed '
              . 'on the client. Please check the client and re-run the test');
        }

        $Log->comment('Client ' . $client->name() . " iqn is: $initiator");

        push @{$initiator_ids}, $initiator;
    }

    $Log->comment(q{-} x 60);
    $Log->comment('Completed gathering initiator IDs for clients as below');
    foreach my $id (@{$initiator_ids }) {
        $Log->comment($id);
    }
    $Log->comment(q{-} x 60);

    return $initiator_ids;

}

=head2 _create_iscsi_service (...)

Creates the iSCSI service and stops and starts it too

=cut

sub _create_iscsi_service {

    my %opts = validate_with (
        params => \@_,
        spec   => {
              vserver => { type => OBJECT }
        },
    );

    my $vserver = delete($opts{vserver});
    my $iscsi_obj;

    $Log->comment(q{-} x 60);
    $Log->comment('Creating iSCSI service on vserver: ' . $vserver->vserver());
    $Log->comment(q{-} x 60);

    try {
        $iscsi_obj = NACL::C::VserverIscsi->create(
                command_interface => $vserver->command_interface(),
                          vserver => $vserver->vserver(),
                    );
    } catch NACL::APISet::Exceptions::CommandFailedException with {
        my $err_txt = 'iSCSI service already exists';
        my $exception = $_[0];
        if ($exception->text() =~ /$err_txt/i) {
            $Log->comment('iSCSI service already exists .. which is ok');

            #if you hit this issue than just create an object
            #for the iSCSI service
            $iscsi_obj = NACL::C::VserverIscsi->new(
                          command_interface => $vserver->command_interface(),
                          vserver           => $vserver->vserver());
        } else {
            NACL::MTask::MCC::Exceptions::BaseException->throw(
                         'Unexpected error encountered: ' . $exception->text());
        }
    };

    #stop and start the iscsi service
    $Log->comment('Stopping the ISCSI service');
    $iscsi_obj->stop();

    $Log->comment('Starting the ISCSI service');
    $iscsi_obj->start();

    $Log->comment(q{-} x 60);
    $Log->comment('Completed creating iSCSI service on vserver: '
                . $vserver->vserver());
    $Log->comment(q{-} x 60);
    return $iscsi_obj ;
}

=head2 _modify_or_find_iscsi_lifs (...)

Method will verify if the vserver(s) have iSCSI lifs, if not will modify
one NAS lif to iSCSI.

NOTE: Code picked up from "dp/mcc/mcc_get_iscsi_lif.thpl" script

=cut

sub _modify_or_find_iscsi_lifs {

    my $self = shift;

    my %opts = validate_with (
        params => \@_,
        spec   => {
              nodes   => {type => ARRAYREF },
              vserver => { type => OBJECT },
        },
    );

    my $nodes = delete($opts{nodes});
    my $vserver = delete($opts{vserver});

    $Log->comment('Converting existing NFS/CIFS data lifs to iSCSI. '
                . 'Ignoring if already present');

    foreach my $node (@{ $nodes }) {
        #Code snippet below picked up from
        #/dp/mcc/mcc_get_iscsi_lif.thpl written by Mandar.
        $Log->comment(q{-} x 60);
        $Log->comment('Verifying for vserver: ' . $vserver->vserver()
                    . ' on node: ' . $node->name());
        $Log->comment(q{-} x 60);

        my $local_data_vserver = NACL::CS::Vserver->fetch(
                                command_interface => $node,
                                           filter => {
                                               vserver => $vserver->vserver(),
                                               subtype => 'sync-source|default',
                                               'allowed-protocols' => ['iscsi'],
                                                     });

        if ( !NACL::CS::NetworkInterface->fetch(
           command_interface => $node,
                      filter => {
                             'data-protocol' => ['iscsi'],
                                 'home-node' => $node->name(),
                                     vserver => $local_data_vserver->vserver,
                                },
                 allow_empty => 1 ) ) {

            my @lif_cs = NACL::CS::NetworkInterface->fetch(
               command_interface => $node,
                          filter => { 'role' => 'data',
                                 'home-node' => $node->name(),
                                     vserver => $local_data_vserver->vserver,
                                    });
            #proceed with modifying a lif to ISCSI under two conditions:
            ## 1) if NFS or CIFS traffic are also requested along with SAN
            ##    then the no. of data lifs found SHOULD BE > 1
            ## 2) if NFS AND CIFS traffic are NOT requested and no. of data lifs
            ##    available is NOT 0
            if (((scalar(@lif_cs) > 1) && ($self->_doCIFS() || $self->_doNFS()))
                ||
                (!scalar(@lif_cs) == 0 && !$self->_doCIFS() && !$self->_doNFS()))
            {
                my $lif = NACL::STask::NetworkInterface->find_data_interfaces(
                    command_interface => $node,
                               filter => { 'lif' => $lif_cs[0]->lif (),
                                   'status-oper' => 'up',
                                     'home-node' => $node->name(),
                                         vserver => $local_data_vserver->vserver,
                                         });

                $lif->purge();

                NACL::STask::NetworkInterface->create(
                                command_interface => $node,
                                              lif => $lif_cs[0]->lif (),
                                          vserver => $local_data_vserver->vserver(),
                                             role => 'data',
                                      'home-node' => $lif_cs[0]->home_node(),
                                      'home-port' => $lif_cs[0]->home_port(),
                                          address => $lif_cs[0]->address(),
                                          netmask => $lif_cs[0]->netmask(),
                                  'data-protocol' => ['iscsi'],
                                   'status-admin' => 'up',
                                  nacltask_verify => 1 );
            } else {
                NACL::MTask::MCC::Exceptions::BaseException->throw(
                    'System does not have enough data lifs to support NAS ' .
                    'and SAN traffic');
            }
        } else {
            $Log->comment('Found existing iSCSI DATA LIF. Skipping iSCSI LIF '
                        . 'creation for ' . $node->name());
        }

        $Log->comment(q{-} x 60);
        $Log->comment('Completed verifying iSCSI Lifs for vserver: '
                    . $vserver->vserver()
                    . ' on node: '
                    . $node->name());
        $Log->comment(q{-} x 60);
    }

    $Log->comment('Completed modifying or finding iSCSI Lifs for vserver: '
                . $vserver->vserver());

    return;
}

=head2 _prepare_vs_for_san (...)

Method to configure the vserver for iSCSI traffic. Involves following steps:

1) Create iSCSI service

2) Create igroups

3) Create SAN volumes

4) Create SAN luns and map the lun to the igroup

=cut

sub _prepare_vs_for_san {

    my $self = shift;

    my %opts = validate_with (
        params => \@_,
        spec   => {
                san_clients  => { type => ARRAYREF|UNDEF, optional => 1 },
                san_vservers => { type => ARRAYREF|UNDEF, optional => 1 },
        },
    );

    #check if the user has passed which vserver of the list that the object
    #maintains, should be used for iSCSI.
    #If nothing has been passed then use all the vservers maintianed by the
    #object

    my ($igroup_list, $iscsi_object_list);
    my $san_clients  = $opts{san_clients};

    $Log->comment(Dumper($san_clients));

    my $san_vservers = defined($opts{san_vservers})
                       ? $opts{san_vservers}
                       : $self->_vserver_list();

    $self->_san_vserver_list($san_vservers);

    #For all the clients (can be none) passed in get the intiator IDs
    my $initiator_groups;
    if (defined($san_clients)) {
        $Log->comment('Get client initiator IDs');
        $initiator_groups = _get_san_client_initiator_id(
                                                       clients => $san_clients);
    }

    #ensure that the vserver list are all objects.
    foreach my $vs (@{ $san_vservers }) {
        if (!(ref($vs)  && $vs->isa('NACL::CS::Vserver'))) {
            NACL::MTask::MCC::Exceptions::BaseException->throw(
                    'Invalid vserver param type passed in. Expected an object');
        }

        #modify or veirfy if iSCSI lifs exist
        $Log->comment('Verifying if the vserver has iSCSI lifs');
        my $nodes = $self->_nodes_list();
        $self->_modify_or_find_iscsi_lifs (
                            nodes  => $nodes,
                            vserver => $vs );

        $Log->comment('Creating iSCSI service on the vservers');
        #create iSCSI service for that vs
        my $iscsi_obj = _create_iscsi_service (vserver => $vs);
        push @{ $iscsi_object_list }, $iscsi_obj;

    }

    $self->_iscsi_service_object_list($iscsi_object_list);
    if ( defined( $initiator_groups ) ) {
        $Log->comment('Creating Igroups');
        $igroup_list = $self->_create_igroups (
                                     vservers => $san_vservers,
                                initiator_ids => $initiator_groups,
                                );
        $self->_igroup_list($igroup_list);
    }

    return;
}

=head2 _create_igroups (...)

Creates igroups required for starting iSCSI traffic on data VServers

=cut

sub _create_igroups {

    my $self = shift;

    my %opts = validate_with (
        params => \@_,
        spec   => {
                vservers      => { type => ARRAYREF },
                initiator_ids => { type => ARRAYREF|UNDEF, optional => 1 },
        },
    );

    my $vservers = $opts{vservers};
    my $initiators = $opts{initiator_ids};
    my $igroup_list;

    foreach my $vs (@{ $vservers }) {
        #create igroup for that vs and add the client initators that
        #can do traffic to that vserver
        my $igroup_name = random_name_generator(
                            prefix   => $vs->vserver(),
                            suffix   => 'ig',
                            size     => 1 );
        $Log->comment(q{-} x 60);
        $Log->comment("Creating igroup: $igroup_name"
                    . ' for vserver: ' . $vs->vserver());
        $Log->comment(q{-} x 60);

        #create a igroup for the vservser  and add all the initiator IDs to it
        my $igroup_obj = NACL::STask::LunIgroup->create(
                    command_interface  => $vs->command_interface(),
                    vserver            => $vs->vserver(),
                    igroup             => $igroup_name,
                    protocol           => 'iscsi',
                    ostype             => 'linux',
                    initiator          => $initiators,
                    nacltask_if_exists => 'purge',
              nacltask_cleanup_manager => $self->{_cleanup_object},
                   );

        $Log->comment(q{-} x 60);
        $Log->comment("Completed creating igroup: $igroup_name"
                    . ' for vserver: ' . $vs->vserver());
        $Log->comment(q{-} x 60);

        push @{ $igroup_list }, $igroup_obj;
    }

    return $igroup_list;
}

=head2 _prepare_vs_for_cifs (...)

Prepares the vserver so that CIFS traffic can be run on it

=cut

sub _prepare_vs_for_cifs {
    my $self = shift;

    my %opts = validate_with (
        params => \@_,
        spec   => {
                cifs_params => { type => HASHREF|UNDEF },
        },
    );

    my $cifs_params = delete($opts{cifs_params});
    my $vservers = $self->_vserver_list();

    # verify the required params are passed in
    if (!(defined($cifs_params->{CIFS_USER}) ||
         defined($cifs_params->{CIFS_PASSWD}))) {
        NACL::MTask::MCC::Exceptions::BaseException->throw(
            'The required params that the library needs were not passed '
          . 'Library expects the params: CIFS_USER and CIFS_PASSWD '
          . 'to be passed by the user when calling the method');
    }

    my $cifs_user = delete($cifs_params->{CIFS_USER});
    my $cifs_passwd = delete($cifs_params->{CIFS_PASSWD});

    foreach my $vs (@{ $vservers }) {

        $Log->comment(q{-} x 60);
        $Log->comment ('Preparing Vserver ' . $vs->vserver() . ' for CIFS');
        $Log->comment(q{-} x 60);

        # verify if CIFS is configured for the vserver using the command
        # vserver cifs show -vserver vs1
        $Log->comment('Verifying vserver cifs show -vserver '
                    . $vs->vserver() . ' status-admin up output');

        try {
             NACL::CS::VserverCifs->fetch(
                    command_interface => $vs->command_interface(),
                    filter            => {
                                            vserver        => $vs->vserver(),
                                            'status-admin' => 'up'
                                         });

        } catch NACL::Exceptions::NoElementsFound with {
             my $ex = shift;
             $Log->comment('NACL exception: ' . $ex->text());
             my $msg = 'CIFS is not configured for vserver : ' . $vs->vserver()
                     . '. Please run deux_init on the setup to configure CIFS!';

             NACL::MTask::MCC::Exceptions::BaseException->throw($msg);
        };

        $Log->comment('Verifying CIFS listening connections for '
                    . $vs->vserver() . ' vserver');

        try {
             NACL::CS::NetworkConnectionsListening->fetch(
                         command_interface => $vs->command_interface(),
                         filter            => {
                                                vserver => $vs->vserver(),
                                                service => q{cifs*},
                                              });

        } catch NACL::Exceptions::NoElementsFound with {
             my $ex = shift;
             $Log->comment('NACL exception: ' . $ex->text());

             my $msg = 'CIFS is not configured for vserver : ' . $vs->vserver()
                     . '. Please run deux_init on the setup to configure CIFS!';

             NACL::MTask::MCC::Exceptions::BaseException->throw($msg);
        };

        $Log->comment('Verifying VServer DNS services');
        try {
             NACL::CS::VserverServicesDns->fetch (
                        command_interface => $vs->command_interface(),
                        filter            => { vserver => $vs->vserver() },);

        } catch NACL::Exceptions::NoElementsFound with {
             my $ex = shift;
             $Log->comment('NACL exception: ' . $ex->text());
             my $msg = 'CIFS is not configured for vserver : ' . $vs->vserver()
                     . '. Please run deux_init on the setup to configure CIFS!';

             NACL::MTask::MCC::Exceptions::BaseException->throw($msg);
        };

        $Log->comment($vs->vserver() . ' is correctly configured for CIFS');

        #verify if the vserver has CIFS data lifs
        my $return_value = _has_working_data_lifs( vserver  => $vs,
                                                   protocol => 'cifs');

        if (!$return_value) {
            NACL::MTask::MCC::Exceptions::BaseException->throw(
                                   'Vserver "' . $vs->vserver()
                                 . '" does not have CIFS data lifs which is '
                                 . 'not a supported configuration');
        }

        $Log->comment($vs->vserver() . ' has CIFS data lifs');

        $self->_cifs_username($cifs_user);
        $self->_cifs_password($cifs_passwd);

        $Log->comment(q{-} x 60);
        $Log->comment ('Completed preparing Vserver '
                     . $vs->vserver() . ' for CIFS IO traffic');
        $Log->comment(q{-} x 60);
    }

    return;
}


=head2 _prepare_vs_for_nfs (...)

Prepares the vserver so that NFS traffic can be run
1) Creates NFS service
2) Enables NFS 3 and/or 4 if required
3) Creates volumes

=cut

sub _prepare_vs_for_nfs {
    my $self = shift;

    my %opts = validate_with (
        params => \@_,
        spec   => {
                nfs_params => { type => HASHREF|UNDEF, optional => 1 },
        },
    );

    my $nfs_params = delete($opts{nfs_params});
    my $vservers = $self->_vserver_list();

    foreach my $vs (@{ $vservers }) {

        $Log->comment(q{-} x 60);
        $Log->comment ('Preparing Vserver ' . $vs->vserver() . ' for NFS');
        $Log->comment(q{-} x 60);
        #first verify if the vserver has NFS data lifs
        my $return_value = _has_working_data_lifs( vserver  => $vs,
                                                   protocol => 'nfs');

        if (!$return_value) {
            NACL::MTask::MCC::Exceptions::BaseException->throw(
                                       'Vserver "' . $vs->vserver()
                                     . '" does not have NFS data lifs which is '
                                     . 'not a supported configuration');
        }

        my %vsnfsopts;
        my $nfs_status = NACL::C::VserverNfs->status(
                                command_interface => $vs->command_interface,
                                        'vserver' => $vs->vserver,);

        if (!$nfs_status) {
            $Log->comment('Marking the NFS service that will be created  ' .
                          'by the lib for deletion');
            $vsnfsopts{nacltask_cleanup_manager} = $self->{_cleanup_object};
        }

        $Log->comment('Creating NFS service on ' . $vs->vserver());
        #Turn on NFS service on the vserver
        my $nfs;
        my $vs_nfs = $vs->vserver();
        try {
           $nfs =  NACL::STask::VserverNfs->create(
              command_interface  => $vs->command_interface(),
              vserver            => $vs_nfs,
              tcp                => 'enabled',
              udp                => 'enabled',
              nacltask_if_exists => 'reuse',
              %vsnfsopts,
            );
        } catch NACL::APISet::Exceptions::CommandFailedException with {
            my $ex = shift;
            if ( $ex->text() =~ /the Vserver "$vs_nfs" is locked/ ) {
               $vs->start( force => 'true' );
            }

            $nfs =  NACL::STask::VserverNfs->create(
              command_interface  => $vs->command_interface(),
              vserver            => $vs_nfs,
              tcp                => 'enabled',
              udp                => 'enabled',
              nacltask_if_exists => 'reuse',
              %vsnfsopts,
            );
        };

        #Enable NFS version : 4,3
        my $enable_nfs = $nfs_params->{ENABLE_NFS_VER};
        if (defined $enable_nfs && scalar(@{ $enable_nfs }) > 0) {
            foreach my $ver (@{ $enable_nfs }) {
                if ( $ver !~ /[34]/ ) {
                    $Log->warn('Invalid NFS version passed. Ignoring it .. ');
                    next;
                }

                if ($ver == 3) {
                    $Log->comment ('Enabling NFS Ver 3.0');
                    $nfs->modify (
                        access => 'true',
                        'v3'   => 'true');
                }

                if ($ver == 4) {
                    $Log->comment ('Enabling NFS Ver 4.0');
                    $nfs->modify (
                        access => 'true',
                        'v4.0' => 'true',
                        'v4.1' => 'true');
                }
            }
        }

        #Create default export policy rule for the vservers
        $Log->comment('Creating default export policy rule for ' .
                        $vs->vserver());
        NACL::C::VserverExportPolicyRule->create(
                command_interface => $vs->command_interface(),
                vserver           => $vs->vserver(),
                policyname        => 'default',
                'clientmatch'     => '0.0.0.0/0',
                rwrule            => ['any'],
                rorule            => ['any'],
                protocol          => ['any'],
                superuser         => ['any'],);

        $Log->comment(q{-} x 60);
        $Log->comment ('Completed preparing Vserver ' . $vs->vserver()
                     . ' for NFS');
        $Log->comment(q{-} x 60);
    }

    return;
}

=head2 _has_working_data_lifs (...)

Private method to verify if the vserver have NFS/CIFS data lifs that can
be used for traffic.

In case of iSCSI if there are no iSCSI lifs then one NFS lif is modified to
support iSCSI

=cut

sub  _has_working_data_lifs {

    my %opts = validate_with (
        params => \@_,
        spec   => {
            vserver  => { type => OBJECT },
            protocol => { type => SCALAR },
        },
    );

    my $vs = delete($opts{vserver});
    my $protocol = delete($opts{protocol});
    my $return_value = 1;

    $Log->comment('Verifying if ' . $vs->vserver()
                . " has working $protocol data lifs");

    try {
        my $vserver_data_lif = NACL::CS::NetworkInterface->fetch(
            command_interface => $vs->command_interface(),
            filter => {
                           role => 'data',
                        vserver => $vs->vserver(),
                'status-oper'   => 'up',
                'data-protocol' => [$protocol] ,
            });
    } catch NACL::Exceptions::NoElementsFound with {
        my $exception = $_[0];
        my $err_txt = 'No matching network interface\(s\) found';

        $Log->comment('Network interface fetch failed for ' . $vs->vserver()
                    . ' with ' .  $exception->text());
        if ($exception->text() =~ /$err_txt/i) {
            $return_value = 0;
        }
   }

   return $return_value;
}

=head2 _create_shares (...)

Private method to create shares.

=cut

sub _create_shares {
    my $self = shift;

    my %opts = validate_with (
         params => \@_,
         spec   => {
                    shares_per_vol => { type => SCALAR|UNDEF, default => 0 },
                    volumes        => { type => ARRAYREF|UNDEF, optional => 1 },
         },
    );

    my $volumes = defined($opts{volumes})
                  ? $opts{volumes}
                  : $self->_cifs_volume_list();

    my $shares_count =  $opts{shares_per_vol};
    my $share_list;

    if ( $shares_count == 0 ) {
        $Log->comment('User did not request creation of shares ... returning');
        return;
    }

    $Log->comment($shares_count . ' shares per volume will be created');

    if ( !defined($volumes) ) {
        NACL::MTask::MCC::Exceptions::BaseException->throw(
                              'No CIFS volumes found to create CIFS shares on');
    }


    foreach my $vol (@{ $volumes }) {
        for (1 .. $shares_count) {
            my $share_name = random_name_generator(
                                prefix => $vol->volume(),
                                suffix => 'share',
                                size   => 0 );
            $Log->comment(q{-} x 60);
            $Log->comment("Creating share: $share_name on volume: "
                        . $vol->volume());
            $Log->comment(q{-} x 60);


            my $share = NACL::STask::VserverCifsShare->create(
                         volume                   => $vol,
                         vserver                  => $vol->vserver(),
                         'share-name'             => $share_name,
                         nacltask_cleanup_manager => $self->{_cleanup_object},
                        );

            push @{$share_list}, $share;

            $Log->comment(q{-} x 60);
            $Log->comment("Completed creating share: $share_name on volume: "
                        . $vol->volume());
            $Log->comment(q{-} x 60);
        }
    }

    $Log->comment('Modify share properties to support SMB 3.0');
    foreach my $share (@{ $share_list }) {
        $Log->comment(q{-} x 60);
        $Log->comment('Working on share: ' . $share->share_name());
        $Log->comment(q{-} x 60);

        try {
            NACL::STask::VserverCifsShareProperties->add(
                command_interface => $share->command_interface(),
                vserver           => $share->vserver(),
               'share-name'       => $share->share_name(),
               'share-properties' => ['continuously-available'] );
        } catch NATE::BaseException with {
            my $exception = shift;
            $Log->comment('Modifying share for SMB 3.0 threw an error: '
                        . $exception->text() );
        };
    }

    $Log->comment(q{-} x 60);
    $Log->comment('Total share resources created are:');
    foreach my $share (@{ $share_list }) {
        $Log->comment($share->share_name());
    }
    $Log->comment(q{-} x 60);

    return $share_list;
}

=head2 _create_and_map_luns (...)

Private method to create luns and map them to an igroup.

=cut

sub _create_and_map_luns {
    my $self = shift;

    my %opts = validate_with (
        params => \@_,
        spec   => {
                  igroups     => { type => ARRAYREF|UNDEF, optional => 1 },
                  lun_per_vol => { type => SCALAR, default => 0 },
                  lun_size    => { type => SCALAR, default => '20GB' },
                  volumes     => { type => ARRAYREF|UNDEF, optional => 1 },
        },
    );

    my $san_vols = defined($opts{volumes})
                   ? $opts{volumes}
                   : $self->_san_volume_list();

    my $no_of_luns_per_vol = delete($opts{lun_per_vol});
    my $per_vol_lun_size = delete($opts{lun_size});
    my $igroups = delete($opts{igroups});
    my $lun_list;

    if ($no_of_luns_per_vol == 0) {
        $Log->comment('The user did no request to create any luns '
                    . ' ... returning');
    }

    $Log->comment("Creating $no_of_luns_per_vol number of luns per vol of size"
                . $per_vol_lun_size);

    if (!defined($san_vols)) {
        NACL::MTask::MCC::Exceptions::BaseException->throw(
                                      'No SAN volumes found to create luns on');
    }


    foreach my $vol (@{ $san_vols }) {
        for (1 .. $no_of_luns_per_vol) {
            my $lun_name  = random_name_generator( prefix   => $vol->volume(),
                                                   suffix   => 'lun',
                                                   size     => 0 );
            $Log->comment(q{-} x 60);
            $Log->comment("Creating lun: $lun_name "
                        . 'on volume: ' . $vol->volume());
            $Log->comment(q{-} x 60);

            my $lun = NACL::STask::Lun->create(
                        command_interface        => $vol->command_interface,
                        lun                      => $lun_name,
                        volume                   => $vol->volume,
                        vserver                  => $vol->vserver,
                        ostype                   => 'linux',
                        size                     => $per_vol_lun_size,
                        if_exists                => 'purge',
                        nacltask_verify          => 1,
                        nacltask_cleanup_manager => $self->{_cleanup_object},
                    );

            push @{ $lun_list }, $lun;

            $Log->comment(q{-} x 60);
            $Log->comment("Completed creating and mapping lun: $lun_name "
                        . 'on volume: '
                        . $vol->volume());
            $Log->comment(q{-} x 60);
        }
    }

    if (!defined($igroups) || scalar(@{ $igroups }) == 0) {
        $Log->comment('No igroups exist to map the Luns .. continuing');
    } else {
        foreach my $lun (@{ $lun_list }) {
            foreach my $igroup (@{ $igroups }) {
                $Log->comment(q{-} x 60);
                $Log->comment('Mapping lun : '
                            . $lun->lun
                            . 'to igroup: '
                            . $igroup->igroup());
                $Log->comment(q{-} x 60);

                $lun->map(
                         igroup                    => $igroup,
                         nacltask_if_mapped_igroup => 'remap',
                         nacltask_verify           => 1,
                         nacltask_cleanup_manager  => $self->{_cleanup_object},
                    );

                $Log->comment(q{-} x 60);
                $Log->comment('Completed mapping lun : '
                            . $lun->lun()
                            . 'to igroup: '
                            . $igroup->igroup());
                $Log->comment(q{-} x 60);
            }
        }
    }

    $Log->comment(q{-} x 60);
    $Log->comment('Total lun resources created are: ');
    foreach my $lun (@{ $lun_list }) {
        $Log->comment($lun->lun());
    }
    $Log->comment(q{-} x 60);

    return $lun_list;
}

=head2 _create_volumes (...)

Private method to create volumes.

=item C<< vols_per_aggr => # of volumes per aggregate that will be created >>

(Optional.) Default is 0

=back

=item C<< per_vol_size => size of the volumes that will be created >>

(Optional.) Default is 1G

=back

=cut

sub _create_volumes {
    my $self = shift;

    my %opts = validate_with (
        params => \@_,
        spec   => {
                    aggregates => { type => ARRAYREF|UNDEF, optional => 1 },
                    vservers   => { type => ARRAYREF|UNDEF, optional => 1 },
                    vol_params => { type => HASHREF|UNDEF, optional => 1 },
                    protocol   => { type => SCALAR },
        },
    );

    my $params = delete $opts{vol_params};
    my $protocol = delete $opts{protocol};
    my $volume_list;

    my $vol_count  = defined($params->{vols_per_aggr})
                     ? $params->{vols_per_aggr}
                     : 0;

    my $vol_size   = defined($params->{per_vol_size})
                     ? $params->{per_vol_size}
                     : '1g';

    my $aggregates = defined($opts{aggregates})
                     ? $opts{aggregates}
                     : $self->_aggregates_list();

    my $vservers   = defined($opts{vservers})
                     ? $opts{vservers}
                     : $self->_vserver_list();

    if (!defined($aggregates) || scalar(@{$aggregates}) == 0) {
       NACL::MTask::MCC::Exceptions::BaseException->throw(
                                    'No aggregates found to create volumes on');
    }

    if (!defined($vservers) || scalar(@{ $vservers }) == 0) {
        NACL::MTask::MCC::Exceptions::BaseException->throw(
         'No vservers found to create volumes with');
    }

    if ($vol_count == 0) {
        $Log->comment('User did not request any volumes to be created ...'
                    . ' returning');
        return;
    }

    $Log->comment("For protocol: $protocol, $vol_count volumes per aggregate "
                . 'will be created');

    foreach my $vs (@{ $vservers }) {

        my $node = $vs->command_interface();

        foreach my $aggr (@{ $aggregates }) {

           for (1 .. $vol_count) {

#                # vol_name format => <node_name>_<random_word>_<protocol>
#                my $vol_name_prefix = $node->name();
#                $vol_name_prefix =~ s/[^a-zA-Z0-9]//g;
#                my $vol_name = random_name_generator(
#                                prefix => $vol_name_prefix,
#                                suffix => $protocol,
#                                size   => 1 );

                # new vol_name format => <random_word>_<random_word>_<protocol>
                my @rand_words = rand_words( size => 2 );
                my $vol_name = $rand_words[0]
                             . '_'
                             . $rand_words[1]
                             . '_'
                             . $protocol;

                $Log->comment(q{-} x 60);
                $Log->comment("Creating volume: $vol_name\n"
                            . 'On aggr: ' . $aggr->aggregate() . "\n"
                            . 'For starting '
                            . uc($protocol)
                            . 'IO traffic on it');
                $Log->comment(q{-} x 60);

                my $vol = NACL::STask::Volume->create(
                      command_interface        => $node,
                      aggregate                => $aggr->aggregate(),
                      volume                   => $vol_name,
                      vserver                  => $vs->vserver(),
                      size                     => $vol_size,
                      type                     => 'RW',
                      autosize                 => 'true',
                      'max-autosize'           => '200g',
                      'junction-active'        => 'true',
                      'junction-path'          => "/$vol_name",
                      'unix-permissions'       => '---rwxrwxrwx',
                      'security-style'         => 'mixed',
                      nacltask_cleanup_manager => $self->{_cleanup_object}, );

                NACL::C::VolumeSnapshotAutodelete->modify(
                     command_interface   => $node,
                     volume              => $vol->volume(),
                     vserver             => $vs->vserver(),
                     enabled             => 'true',
                     'defer-delete'      => 'none',
                     'target-free-space' => '5');

                $Log->comment(q{-} x 60);
                $Log->comment("Completed creating volume: $vol_name");
                $Log->comment(q{-} x 60);

                push(@{ $volume_list }, $vol);
            }
        }
    }

    $Log->comment(q{-} x 60);
    $Log->comment('Total volumes created are:');
    foreach my $vol (@{ $volume_list }) {
        $Log->comment($vol->volume());
    }
    $Log->comment(q{-} x 60);

    return $volume_list;
}

=head2 _create_or_find_aggregates (...)

Private method to create or find aggregates.

=item C<< aggregates_per_vs => # of aggregates to create per vserver >>

(Optional.) This number will determine how many aggregates to create per
vserver on each node. So if the user is calling this method with 2 vservers and
passes a value of 2 for this parameter, then the library will create
4 (# of aggregate_per_vs * # of vs) per node. By default 0 aggregates will
be created

=back

=item C<< reuse_aggrs => (0|1) decides whether to reuse the aggregates. >>

(Optional.) Boolean value which decides whether user wants to reuse the
aggregates that already exist in the system. If this options is on and the
system does not contain enough number of aggregates to match the users ask
then the library will go ahead and create the remaining aggregates.
By default value is 1.

=back

=item C<< diskcount => decides the size of the aggregate >>

(Optional.) The number of disks to use while creating aggregates.
By default value is 6.

=back

=cut

sub _create_or_find_aggregates {
    my $self = shift;

    my %opts = validate_with (
            params  => \@_,
            spec    => {
                    aggr_params   => { type => HASHREF|UNDEF, optional => 1 },
                    vserver_count => { type => SCALAR|UNDEF, optional => 1 },
            },
    );
    my $params = delete($opts{aggr_params});
    my $total_no_of_aggrs_per_node;

    my $aggregates_per_vs = defined($params->{aggregates_per_vs})
                            ? $params->{aggregates_per_vs}
                            : 0;

    my $reuse_aggrs =  defined($params->{reuse_aggrs})
                       ? $params->{reuse_aggrs}
                       : 1;

    my $diskcount = defined($params->{diskcount})
                    ? $params->{diskcount}
                    : 6;

    my $unmirrored_aggrs = defined($params->{unmirrored_aggrs})
                           ? $params->{unmirrored_aggrs}
                           : 0;

    my $skip_aggr_creation = defined($params->{skip_aggr_creation})
                           ? $params->{skip_aggr_creation}
                           : 0;	

    if ( $aggregates_per_vs > 0 ) {
        my $no_of_vservers = defined($opts{vserver_count})
                             ? $opts{vserver_count}
                             : scalar ( @{ $self->_vserver_list() } );

        if (defined($no_of_vservers) && $no_of_vservers == 0) {
            $total_no_of_aggrs_per_node = $aggregates_per_vs;
        } else {
            $total_no_of_aggrs_per_node = $no_of_vservers * $aggregates_per_vs;
        }

    } else {
        $Log->comment('The user did not request any aggregates .. returning');
        return;
    }

    $Log->comment("Creating $total_no_of_aggrs_per_node aggregates per node");

    # First lets see if the nodes already contain the required # of aggregates.
    # If yes create only the difference and re use the ones already present

    my @nodes = @{ $self->_nodes_list() };
    my $aggregate_list = ();

    foreach my $node (@nodes) {
        my @aggrs;
        my @unmirrored_aggrs;
        if ($reuse_aggrs) {
            $Log->comment('Looking for existing online non-root mirrored data '
                        . 'aggrs');
            @aggrs = NACL::CS::StorageAggregate->fetch(
                                 command_interface => $node,
                                 canned_filters    => [qw(online_not_root)],
                                 filter            => {
                                                        nodes     => [$node],
                                                        mirror    => 'true',
                                                        'is-home' => 'true',
                                                      },
                                 allow_empty       => 1 );

            if ($unmirrored_aggrs) {
                $Log->comment('Looking for existing online non-root unmirrored '
                            . 'data aggrs');
                @unmirrored_aggrs = NACL::CS::StorageAggregate->fetch(
                                 command_interface => $node,
                                 canned_filters    => [qw(online_not_root)],
                                 filter            => {
                                                        nodes     => [$node],
                                                        mirror    => 'false',
                                                        'is-home' => 'true',
                                                      },
                                 allow_empty       => 1 );
            }
        }

        # Create aggregates only if the reuse_aggrs pram is not set or if
        # the system has less than required existing aggrs
        if ( scalar @aggrs < $total_no_of_aggrs_per_node && !$skip_aggr_creation) {

            my $remaining_aggr_count = $total_no_of_aggrs_per_node -
                                       scalar(@aggrs);
            $Log->comment("Creating the remaining $remaining_aggr_count aggrs");

            for (1 .. $remaining_aggr_count) {
                my $aggr_was_created;
                my $disk_count_per_aggr = $diskcount;
                my $aggr_name_prefix = $node->name();
                $aggr_name_prefix =~ s/[^a-zA-Z0-9]//g;

                my $aggregate_name = random_name_generator(
                                                    prefix => $aggr_name_prefix,
                                                    suffix => 'aggr',
                                                    size   => 1 );
                $Log->comment(q{-} x 60);
                $Log->comment("Creating mirrored aggregate $aggregate_name "
                            . 'on node: '
                            . $node->name());
                $Log->comment(q{-} x 60);

                my $aggr = NACL::STask::Aggregate->create(
                           aggregate                => $aggregate_name,
                           command_interface        => $node,
                           diskcount                => $disk_count_per_aggr,
                           if_exists                => 'reuse',
                           mirror                   => 'true',
                           _was_created             => \$aggr_was_created,
                           nacltask_wait            => 1,
                           nacltask_cleanup_manager => $self->{_cleanup_object},
                           nacltask_to_check_mcc    => 1);

                $Log->comment(q{-} x 60);
                $Log->comment("Mirrored aggr $aggregate_name created on node: "
                            . $node->name());
                $Log->comment(q{-} x 60);

                push(@aggrs, $aggr);
            }
        }

        ## Create un-mirrored aggrs
        if ( scalar @unmirrored_aggrs < $unmirrored_aggrs && !$skip_aggr_creation ) {

            my $remaining_aggr_count = $unmirrored_aggrs -
                                                      scalar(@unmirrored_aggrs);
            $Log->comment("Creating $remaining_aggr_count unmirrored aggrs per "
                        . 'MCC node');

            for (1 .. $remaining_aggr_count) {
                my $aggr_was_created;
                my $disk_count_per_aggr = $diskcount/2;
                my $aggr_name_prefix = $node->name();
                $aggr_name_prefix =~ s/[^a-zA-Z0-9]//g;

                my $aggregate_name = random_name_generator(
                                                    prefix => $aggr_name_prefix,
                                                    suffix => 'un_aggr',
                                                    size   => 1 );
                $Log->comment(q{-} x 60);
                $Log->comment("Creating unmirrored aggregate $aggregate_name "
                            . 'on node '
                            . $node->name());
                $Log->comment(q{-} x 60);

                my $aggr = NACL::STask::Aggregate->create(
                           aggregate                => $aggregate_name,
                           command_interface        => $node,
                           diskcount                => $disk_count_per_aggr,
                           if_exists                => 'reuse',
                           mirror                   => 'false',
                           _was_created             => \$aggr_was_created,
                           nacltask_wait            => 1,
                           nacltask_cleanup_manager => $self->{_cleanup_object},
                           nacltask_to_check_mcc    => 1);

                $Log->comment(q{-} x 60);
                $Log->comment("Unmirrored aggr $aggregate_name created on node "
                            . $node->name());
                $Log->comment(q{-} x 60);

                push(@unmirrored_aggrs, $aggr);
            }
        }

        push @{$aggregate_list}, @aggrs;
        push @{$aggregate_list}, @unmirrored_aggrs;
    }

    $Log->comment(q{-} x 60);
    $Log->comment('Total aggregate resources created/reused are as below:');
    foreach my $aggr (@{ $aggregate_list }) {
        $Log->comment($aggr->aggregate());
    }
    $Log->comment(q{-} x 60);

    return $aggregate_list;
}

=head2 _create_vservers (...)

Creates new VServers with dummy cluster mgmt IPs

=cut

sub _create_vservers {
    my $self = shift;

    my %opts = validate_with (
                        params  => \@_,
                        spec    => {
                                        aggregate => { type => OBJECT },
                                        vserver_params   => { type => HASHREF },
                                   });

    my $vserver_list;
    my $vs_param = $opts{vserver_params};
    my $vs_count = $vs_param->{vserver_count};

    my $aggr = defined($opts{aggregate})
               ? $opts{aggregate}
               : $self->_aggregates_list();

    my $create_lifs = defined($vs_param->{create_data_lifs})
                      ? $vs_param->{create_data_lifs}
                      : 1;

    if ( !defined($aggr) || scalar(@{ $aggr }) == 0 ) {
        NACL::MTask::MCC::Exceptions::BaseException->throw(
                    'No aggregates were passed to create vserver root volumes');
    }

    for (1 .. $vs_count) {
        my $vs_name = random_name_generator( suffix => 'vs',
                                             size   => 1 );

        my $root_vol_name = 'r_' . $vs_name;
        $Log->comment(q{-} x 60);
        $Log->comment("Creating vserver: $vs_name");
        $Log->comment(q{-} x 60);

        my $vserver_obj = NACL::STask::Vserver->create(
                    command_interface            => $aggr->command_interface,
                    vserver                      => "$vs_name",
                    aggregate                    => $aggr->aggregate(),
                    rootvolume                   => $root_vol_name,
                    'ns-switch'                  => ['file','nis'],
                    'nm-switch'                  => ['file'],
                    'rootvolume-security-style'  => 'mixed',
                    'language'                   => 'en_US',
                    'snapshot-policy'            => 'default',
                    'antivirus-on-access-policy' => 'default',
                    nacltask_if_exists           => 'purge',
                        nacltask_cleanup_manager => $self->{_cleanup_object},
            );

        $Log->comment(q{-} x 60);
        $Log->comment("Completed creating vserver: $vs_name");
        $Log->comment(q{-} x 60);

        if ($create_lifs) {
            my $lif_count = defined($vs_param->{lif_count})
                            ? $vs_param->{lif_count}
                            : 1;

            my $lif_prefix = $vs_name;
            $lif_prefix =~ s/[^a-zA-Z0-9]//g;

            my $lif_name = random_name_generator(
                               prefix  => $lif_prefix,
                               suffix  => 'lif',
                               size    => 1 );

            my $lif_ip = _get_a_random_number();
            $lif_ip =~ s//./g;
            $lif_ip =~ s/^.//;

            $Log->comment(q{-} x 60);
            $Log->comment("Creating Lif: $lif_name with ip: $lif_ip for "
                        . "vserver: $vs_name");
            $Log->comment(q{-} x 60);

            NACL::STask::NetworkInterface->create(
                command_interface        => $aggr->command_interface(),
                lif                      => $lif_name,
                vserver                  => $vs_name,
                role                     => 'data',
                'home-node'              => $aggr->command_interface()->node(),
                'home-port'              => 'e0a',
                address                  => $lif_ip,
                netmask                  => '255.255.252.0',
                'data-protocol'          => ['nfs','cifs'],
                'status-admin'           => 'up',
                nacltask_verify          => 1,
                nacltask_cleanup_manager => $self->{_cleanup_object},
            );

            $Log->comment(q{-} x 60);
            $Log->comment("Completed creating Lif: $lif_name with ip: $lif_ip "
                        . "for vserver: $vs_name");
            $Log->comment(q{-} x 60);
        }

        push @{$vserver_list}, $vserver_obj;
    }

    return $vserver_list;
}

=head2 _get_a_random_number (...)

Returns a random number between 1000 and 8000

=cut

sub _get_a_random_number {
    my $range = 8000;
    my $minimum = 1000;

    my $random_number = int(rand($range)) + $minimum;

    return $random_number;
}

=head2 _isSubset (...)

Given two arrays determines if one array is a subset of the other

=cut

sub _isSubset {
    my ($littleSet, $bigSet) = @_;

    foreach my $item (@{$littleSet}) {
        return 0 unless grep {$_ eq $item} @{$bigSet};
    }

    return 1;
}

=head2 _is_2_node_system (...)

=item C<< node => NACL::C::Node object >>

(Required) The NACL:C::Node object to connect for verifying MCC configurations

=back

This will return 1 if it detects a 2-node setup

=cut

sub _is_2_node_system {
    my $self = shift;
    my %opts = validate_with (
        params => \@_,
        spec   => {
                node => { type => OBJECT },
        },
    );

    my @MetroclusterNode_status = NACL::CS::MetroclusterNode->fetch(
                                             command_interface => $opts{node} );

    scalar @MetroclusterNode_status > 2 ? return 0 : return 1;
}

=head2 label

Returns the label string used while creating the TestbedHelper object

=cut

sub label {
    my $self = shift;

    return $self->_label();
}

1;
