
#
# Copyright (c) 2014 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary MCC Validator Base MTask Module
## @author dl-mcc-dpg-qa@netapp.com
## @status shared
## @pod here

package NACL::MTask::MCC::Validator::Base;

use strict;
use warnings;

use Carp;

use Class::MethodMaker [
    new => [ qw/ -init new / ],
    scalar => [ qw/ prepared_data / ], #for use w/ prepare sub
    scalar =>
        [{'-type' => 'NACL::MTask::MCC::Metrocluster'}, qw/ metrocluster /],
    hash =>
        [ {'-default_ctor' => sub { $_[0]->_build_disaster_status(); } },
          qw/ disaster_status /
      ],
    array => [ qw/ errors / ]
];

use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();

use NACL::MTask::MCC::Exceptions::Validation;

use Params::Validate
    qw(validate_with BOOLEAN HASHREF OBJECT ARRAYREF SCALAR UNDEF);


=head1 NAME

NACL::MTask::MCC::Validator::Base

=head1 DESCRIPTION

This is an abstract base class for classes designed to validate
different aspects of the Metrocluster state. Developers should create
subclasses of this class that define checks on Metrocluster validity.

Developers should provide a 'validate' method in the subclass.

Assuming that 'MyValidator' is a subclass of this class, the basic
workflow is:

  my $validator = MyValidator->new(
      metrocluster   => $mcc,
      disaster_status => $disaster_status
  );

  # Calls "can_run_validation()", then "validate", and throws a
  # Validation exception if there are errors (since die_on_failure is
  # true)
  $validator->run(
     die_on_failure => 1
  );


To create the subclass, you'd do something like:

  package MyValidator;
  use strict;
  use warnings;

  use base 'NACL::MTask::MCC::Validator::Base';

  use NATE::BaseException qw(:try); #if class uses try/catch

  sub validate {
     my $self = shift;

     # Put validation code here
     # Call $self->errors_push() for any validation failures found
  }

  1;

=head1 METHODS

=head2 new()

  my $vtor = NACL::MTask::MCC::Validator::Base->new(
     metrocluster => $mcc,
     disaster_status => $disaster_status
  );

Creates a new NACL::MTask::MCC::Validator::Base object. Since this is
an abstract base class, there's little point to doing this-- create a
subclass instead.

=over

=item Options

=over

=item C<< metrocluster =E<gt> $mcc >>

(Required.) A NACL::MTask::MCC::Metrocluster object on which we'll do
validations.

=item C<< disaster_status >>

Hashref. This should be the disaster status as reported by
$mcc-E<gt>get_disaster_status( include_HA =E<gt> 1 ). If
C<disaster_status> is not provided or is undefined, the validator will
call get_disaster_status() on the Metrocluster object itself.

=back

=back

=cut

sub init {
    my $self = shift;

    my %opts = validate_with(
        params => \@_,
        spec   => {
            'metrocluster' => { 'type' => OBJECT },
            'disaster_status' => {'type' => HASHREF | UNDEF, 'optional' => 1},
        }
    );

    $self->metrocluster( $opts{'metrocluster'} );

    my $disaster_status = $opts{'disaster_status'};

    if (! defined $disaster_status ) {
        $disaster_status = $self->_build_disaster_status();
    }

    $self->disaster_status($disaster_status);

    $self->prepared_data({}); # default prepared data to hashref

    return;

}

=head2 prepare()

BEFORE the dr operation is performed, prepares the validation that will be
run after the dr operation occurs.

Most validations do not require a preparation step so it's not defined
in the base class.  If your validation requires a preparation step
(for example gathering data from before the dr operation to compare
after the dr operation is run) then define this method.

=cut

#sub prepare { return; }

=head2 can_prepare()

If a prepare method is defined then can_prepare should return 1;

You should not have to override this method in a subclass, it just makes
sure that a prepare method has been defined and that it is being
called prior to a dr operation.

=item Options

=over

=item C<< pre_dr_operation >>

boolean.  Set to true if calling directly before issuing a DR operation.
It should be false if directly after a DR operation.  The intention
is to be able to call prepare  before a dr operations but not after.


=back

=back

=cut

sub can_prepare {
    my $self = shift;

    my %in = validate_with(
        params => \@_,
        spec   => {
            pre_dr_operation => {regex => qr/^[01]$/, 'default' => 0},
        }
    );

    #if it is pre state AND prepare method exists, we can prepare
    return $in{pre_dr_operation} && $self->can('prepare');
}

=head2 run()

  $vtor->run(
    die_on_failure  => 1,
    extra_validate_options => { foo => 'bar' }
  );

Calls the can_run_validation() method, then, if it returns success,
the validate() method. If validate() returns any errors, it either
warns or throws a NACL::MTask::MCC::Exceptions::Validation exception
based on the value of C<die_on_failure>.

=over

=item Options

=over

=item C<< die_on_failure >>

Boolean. If this value is true, C<run()> throws a
NACL::MTask::MCC::Exceptions::Validation if there are any validation
failures returned by the C<validate()> method. Defaults to true.

=item C<< extra_validate_options >>

Hashref. Passes any options here to the validate() call.

=item C<< pre_validate >>

boolean.  Set to true if calling directly before issuing a DR operation.
It should be false if directly after a DR operation.  The intention
is to be able to call validations before or after dr operations and
some validations are designed to be compatible in certain situations.



=back

=back

If die_on_failure is false, returns true on success and false on
failure. Otherwise, returns true on success and throws a Validation
exception on failure.

=cut

sub run {
    my $self = shift;

    my %opts = validate_with(
        params => \@_,
        spec   => {
            'die_on_failure' =>
                {'type' => SCALAR, 'regex' => qr/^[01]$/, 'default' => 1},
            'pre_validate'   => {'regex' => qr/^[01]$/, 'default' => 0},
            'extra_validate_options' =>
            {'type' => HASHREF, default => {}}
        }
    );


    if (! $self->can('validate')) {
        NATE::BaseException->throw(
            q{Validator class '}
                . ref $self
                    . q{' does not implement a validate() method. }
                        . 'Please add one.'
                );
    }

    # Check to see if we can run the validation
    if (! $self->can_run_validation( pre_validate => $opts{pre_validate} )) {
        return;
    }

    # Run the validation
    $self->validate(%{ $opts{'extra_validate_options'} });

    # If there were errors
    if ( $self->errors_count() ) {
        my $summary = $self->format_error_summary();

        $self->metrocluster->log_validation_event(
            type => $self->validation_type,
            successful => 0,
            summary => $summary,
            results => [ $self->errors() ]
        );

        if ($opts{'die_on_failure'}) {
            NACL::MTask::MCC::Exceptions::Validation->throw($summary);
        } else {
            $Log->warn($summary);
        }

        return;

    } else {
        $self->metrocluster->log_validation_event(
            type => $self->validation_type,
            successful => 1,
            summary => $self->format_success_summary(),
        );
        $Log->comment('validation ' . $self->validation_type() .
            ' successful: ' . $self->format_success_summary());

        return 1;
    }


} ## end sub run

#errors_push defined by Class::MethodMaker (errors is defined as an array so
#the errors_push method is auto generated)

=head2 errors_push()

   $vtor->errors_push( $error );

Adds an error to the list of errors. Subclasses should use this for
every error discovered in the validate() method.

=head2 errors()

  my @errors = $vtor-E<gt>errors()

Returns the list of errors detected by the validate() method.

=head2 format_error_summary()

   $vtor-E<gt>format_error_summary();

Postprocesses error results from validate() into a string
summary. Defaults to: "Failed $validation_type validation in DR state
$dr_state: ", followed by C<errors()>, stringified and concatenated
with newlines.

=cut

sub format_error_summary {
    my $self = shift;

    my $message = join("\n", $self->errors() ) . "\n";

    return 'Failed '
        . $self->validation_type
            . q{ validation in DR state '}
                . $self->dr_state()
                    . "': $message";
}

=head2 format_success_summary()

  $vtor->format_success_summary()

Returns the message which should appear in the log on
success. Defaults to "Passed $validation_type validation in DR state
'foo'."

=cut

sub format_success_summary {
    my $self = shift;

    return 'Passed ' . $self->validation_type()
        . q{ validation in DR state '}
            . $self->dr_state()
                . q{'};
}

##
## _build_disaster_status()
##
## Gets the disaster status from the metrocluster object. Used to initialize
## the 'disaster_status' attribute.
##

sub _build_disaster_status {
    my $self = shift;

    return $self->metrocluster->get_disaster_status( include_HA => 1 );

}


=head2 validation_type()

  $vtor->validation_type()

Returns a human-friendly name for this validation type. This
validation type should be listed in C<%types> in
NACL::MTask::MCC::Metrocluster::validation_report(). The base
implementation returns a name which is the class's base name
translated from CamelCase into underscores. So a validator named
'ClusterShow' would be translated to 'cluster_show'.

If you want to customize your validation type, override this method to
return whatever string you want.

This method can be called in object or class context.

=cut

sub validation_type {
    my $self = shift;

    my $pkg = ref $self;
    if (! $pkg) {
        $pkg = $self;
    }

    # Chop off the beginning
    $pkg =~ s/^.*:://;

    # Translate from CamelCase to spaces
    $pkg =~ s/([[:lower:]])([[:upper:]])/$1_$2/g;

    return lc $pkg;

}

=head2 validation_description()

  $vtor->validation_description()

Returns a human-friendly description for this validation. Subclasses
should override this method to provide their own descriptions.

By default, returns "validation for '$validation_type'".

=cut

sub validation_description {
    my $self = shift;

    return q{validation for '}
        . $self->validation_type()
            . q{'};
}


=head2 extended()

Should return true if this validation is particularly time-consuming
and should only be run on request. Defined as false in the base
class. Override this method if your module should be considered 'extended'.

This method should be callable from either a class or object
perspective. (Since in almost all cases this will just return a simple
constant, that shouldn't be too difficult.)

=cut

sub extended { return; }

=head2 can_run_validation()

Should return true if we should run this validation, and false if we
should skip it. Should print a warning indicating why the validation
will be skipped if it returns false.

By default, returns true if we are in a stable state (normal,
switchover_completed, heal_aggrs_completed, heal_roots_completed) and
false otherwise.

=over

=item Options

=over

=item C<< pre_validate >>

boolean.  Set to true if calling directly before issuing a DR operation.
It should be false if directly after a DR operation.  The intention
is to be able to call validations before or after dr operations and
some validations are designed to be compatible in certain situations.


=back

=back


=cut

sub can_run_validation {
    my $self = shift;

    my %opts = validate_with(
        params => \@_,
        spec   => {
            'pre_validate'  => {'regex' => qr/^[01]$/, 'default' => 0},
        }
    );

    if ( $opts{pre_validate} ) {
        #by default we don't pre validate anything
        $Log->comment('Not pre validating '
                       . $self->validation_type()
                           . ': DR state '
                              . $self->dr_state()
                                  . ' not supported.');
        return;
    } elsif ( $self->is_state_completed() ) {
        return 1;
    }
    else {
        $Log->comment('Unable to run validation '
                       . $self->validation_type()
                           . ': DR state '
                              . $self->dr_state()
                                  . ' not supported.');

        return;
    }
}

=head2 validate()

  $vtor->validate();

ABSTRACT METHOD. The developer should define this method to do
whatever is necessary to validate whatever this class is trying to
validate.

If there are any validation failures, validate() should call
errors_push() to populate the errors list with strings describing
those failures.

Its return value is currently ignored.

=head2 dr_state()

Returns the disaster status DR state.

=cut

sub dr_state {
    my $self = shift;

    return $self->disaster_status()->{'DR_state'};
}

=head2 is_normal()

  $vtor->is_normal();

Returns true if the disaster status DR_state is 'normal'.

=cut

sub is_normal {
    my $self = shift;

    return $self->dr_state() eq 'normal';
}

=head2 is_switchover_completed()

  $vtor->is_switchover_completed();

Returns true if the disaster status DR_state is 'switchover_completed'.

=cut

sub is_switchover_completed {
    my $self = shift;

    return $self->dr_state() eq 'switchover_completed';
}

=head2 is_heal_aggrs_completed()

  $vtor->is_heal_aggrs_completed();

Returns true if the disaster_status DR_state is 'heal_aggrs_completed'.

=cut

sub is_heal_aggrs_completed {
    my $self = shift;

    return $self->dr_state() eq 'heal_aggrs_completed';
}

=head2 is_heal_roots_completed()

Returns true if the disaster status DR_state is 'heal_roots_completed'.

=cut

sub is_heal_roots_completed {
    my $self = shift;

    return $self->dr_state() eq 'heal_roots_completed';
}

=head2 is_state_completed()

Returns true if the disaster status DR_state is 'normal',
'switchover_completed', 'heal_aggrs_completed', or
'heal_roots_completed'.

=cut

sub is_state_completed {
    my $self = shift;

   return $self->is_normal() ||
       $self->is_switchover_completed() ||
            $self->is_heal_aggrs_completed() ||
                $self->is_heal_roots_completed();

}

=head2 is_run_by_default()

Indicates whether this validation should be run by default. The base
implementation returns true.

=cut

sub is_run_by_default {
    return 1;
}


=head1 ARCHITECTURAL GOALS

The goal for creating this architecture is to minimize duplicate code
while providing a common mechanism for running validations.

All validations have to get the state of some piece of the
metrocluster, compare it to an expected state, and log any
difference. Once we've finished checking everything, if there are any
unexpected results, we should either issue a warning or throw an
exception. We also need to record the results in the validation log.

Previously, the code to do this has been copied and pasted in multiple
locations. This architecture attempts to provide a structure that
abstracts away the parts that are commonly used. The Base class, here,
provides a run() method that handles logging validation events,
formatting errors, and other cross-cutting concerns for all validation
routines.

Using a class heirarchy also lets us set up further code reuse. For
example, many validations get a series of component state objects,
then compare subsets of those to expected results. One could set up a
pipe-and-filter architecture in a superclass, then only override the
methods specific to your particular class in a subclass.

=cut

1;
