#
# Copyright (c) 2014 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary MetroCluster Cluster Validation Base MTask Module
## @author dl-mcc-dpg-qa@netapp.com
## @status shared
## @pod here

package NACL::MTask::MCC::Validator::ClusterBase;

use strict;
use warnings;

use parent qw(NACL::MTask::MCC::Validator::Base);

use Params::Validate
    qw(validate_with BOOLEAN HASHREF OBJECT ARRAYREF SCALAR UNDEF);


=head1 NAME

NACL::MTask::MCC::Validator::ClusterBase

=head1 DESCRIPTION

This is an abstract base class for classes designed to validate
Metrocluster clusters. Developers should create subclasses of this
class that define checks on Metrocluster cluster validity by
overriding the C<validate()> method.

This class inherits all methods from NACL::MTask::MCC::Validator::Base.

=head2 new()

  my $vtor = NACL::MTask::MCC::Validator::MyVtor->new(
     metrocluster => $mcc,
     disaster_status => $disaster_status
  );

Creates a new NACL::MTask::MCC::Validator::ClusterBase object.

=over

=item Options

=over

=item C<metrocluster>

(Required.) A NACL::MTask::MCC::Metrocluster object on which we'll do
validations.

=item C<< disaster_status >>

Hashref. This should be the disaster status as reported by
$mcc-E<gt>get_disaster_status( include_HA =E<gt> 1 ). If
C<disaster_status> is not provided or is undefined, the validator will
call get_disaster_status() on the Metrocluster object itself.

=back

=back

(Copied from Base)

=head2 all_components()

  $vtor->all_components()

Returns cluster components for both sides.

=cut

sub all_components {
    my $self = shift;

    return (
        $self->metrocluster->get_cluster(site => q{A}),
        $self->metrocluster->get_cluster(site => q{B})
    );

}

=head2 survivor_components()

  $vtor->survivor_components()

Returns a list of surviving cluster components.

=cut

sub survivor_components {
    my $self = shift;

    my @survivors = 
        ( $self->disaster_status->{'survivor_site'} ) ||
        @{ $self->disaster_status->{'live_sites'} };

    my @clusters = map { $self->metrocluster->get_cluster(site => $_); }
                   @survivors;

    return @clusters;
} ## end sub survivor_clusters

=head2 disaster_components()

  $vtor->disaster_components()

Returns a list of disaster cluster components. In normal mode, this is
the empty list. Otherwise, it's the "disaster site" cluster.

=cut

sub disaster_components {
    my $self = shift;

    if ($self->is_normal()) {
        return;
    } else {
        return $self->metrocluster->get_cluster(
            site => $self->disaster_status->{'disaster_site'});
    }

} ## end sub disaster_clusters

=head2 random_component()

  $vtor->random_component()

Returns a randomly selected cluster.  If metrocluster is not
in a normal state it will return the cluster that is up.

=cut

sub random_component {
    my $self = shift;

    my @clusters = $self->survivor_components();
    #pick a random cluster
    my $random = $clusters[rand @clusters];
    return $random;

} ## end sub disaster_clusters

=head2 all_states()

Returns NACL::CS::Cluster objects for all clusters. See
C<all_components()>.

=cut

sub all_states {
    my $self = shift;

    return $self->_comps_to_states( $self->all_components() );
}

=head2 survivor_states()

Returns NACL::CS::Cluster objects for all surviving clusters. See
C<survivor_components()>.

=cut

sub survivor_states {
    my $self = shift;

    return $self->_comps_to_states( $self->survivor_components );
}

=head2 disaster_states()

Returns the NACL::CS::Cluster object, if any, for all disaster
clusters. See C<disaster_components()>.

=cut

sub disaster_states {
    my $self = shift;

    return $self->_comps_to_states( $self->disaster_components );
}


##
## _comps_to_states()
##
##   $vtor->_comps_to_states(@comps);
##
## Given a list of NACL::C::Cluster objects, transforms them into
## NACL::CS::Cluster objects by fetching their state.
##
sub _comps_to_states {
    my $self = shift;

    my (@comps) = @_;

    my @states;

    foreach my $comp ( @comps ) {

        my $ci = $comp->get_Cluster_component()->command_interface;

        my $cs = NACL::CS::Cluster->fetch(command_interface => $ci);

        push(@states, $cs);
    }

    return @states;

} ## end sub survivor_states

=head1 SEE ALSO

=over

=item *

L<NACL::MTask::MCC::Validator::Cluster::Base
|NACL::MTask::MCC::Validator::Cluster::Base>

=back

=cut

1;
