#
# Copyright (c) 2011-2014 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary MetroCluster Validator Manager
## @author dl-mcc-dpg-qa@netapp.com
## @status shared
## @pod here

package NACL::MTask::MCC::Validator::Manager;

use strict;
use warnings;

use Class::MethodMaker [
    new => [ qw/ new / ],
    scalar => [
        {'-default_ctor' => sub { return {}; },},
        qw/ prepared /,
    ],
];

use Module::Find ();

use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();

use NATE::BaseException qw(:try);
use NACL::MTask::MCC::Exceptions::Validation;

use Params::Validate
    qw(validate_with BOOLEAN HASHREF OBJECT ARRAYREF SCALAR UNDEF);

=head1 NAME

NACL::MTask::MCC::Validator::Manager

=head1 SYNOPSIS

  my $vmgr = NACL::MTask::MCC::Validator::Manager->new();

  my @validators = $vmgr->get_all_validations();

  my %validation_types_table = $vmgr->validation_types_table();

  $vmgr->run_all_validations(
    metrocluster => $mcc,
    disaster_status => $disaster_status,
    die_on_failure => 1,
    stop_on_failure => 1,
    run_extended => 1,
    include => [ 'foo', 'bar', qr/^baz/ ],
    exclude => [ 'pip', 'pop', qr/bim|bam/ ]
  );

=head1 DESCRIPTION

Validators are objects that can be used to validate the state of a
Metrocluster against its expected state. This module is used to find
out information about the validations and run them.

=head1 METHODS

=head2 new()

  NACL::MTask::MCC::Validator::Manager->new();

Creates a new NACL::MTask::MCC::Validator::Manager object.

=head2 get_all_validations()

  $vmgr->get_all_validations()

Returns the class names for all Validator classes not excluded by
C<excluded_classes)>.

=cut

sub get_all_validations {
    my $self = shift;

    my @classes = Module::Find::usesub('NACL::MTask::MCC::Validator');

    my @validators;

    CLASS: foreach my $class (@classes) {

        if (! $class->can('validate') ) {
            $Log->debug("Class $class does not have a validate() method");
            next CLASS;
        }

        if (! $class->isa('NACL::MTask::MCC::Validator::Base') ) {
            $Log->debug("Class $class does not extend Validator::Base");
            next CLASS;
        }

        push(@validators, $class);
    }

    return @validators;

}

=head2 validation_types_table()

  my %vtable = $vmgr->validation_types_table();

Returns a hash of validation types against validation type descriptions.

=cut

sub validation_types_table {
    my $self = shift;

    my %type_table;

    foreach my $validation_class ( $self->get_all_validations() ) {
        $type_table{ $validation_class->validation_type() } =
            $validation_class->validation_description();
    }

    return %type_table;
}

=head2 _get_all_validations_compatibility_list()

  my %vtable = $vmgr->_get_all_validations_compatibility_list();

Returns a list of validation types of the form:

( { name => 'validation_name_1',
    sub  => sub {  },
    extended => 0,
    manager => 1,
    default => 1
  },
)

This is intended only as a temporary integration layer with the
existing $mcc->get_all_validations() call

=cut

sub _get_all_validations_compatibility_list {
    my $self = shift;

    my @validation_list;
    foreach my $validation_class ( $self->get_all_validations() ) {
        push @validation_list, {
            name => $validation_class->validation_type(),
            sub  => sub {
                $Log->comment($validation_class->validation_type() .
                    ' run by validation manager...');
            },
            manager => 1, #integration thing, run in manager if here,
            default => $validation_class->is_run_by_default()
        };
    }
    return @validation_list;
}

=head2 run_all_validations()

  $vmgr->run_all_validations(
    metrocluster => $mcc,
    disaster_status => $disaster_status,
    die_on_failure => 1,
    stop_on_failure => 1,
    run_extended => 1,
    include => [ 'foo', 'bar', qr/^baz/ ],
    exclude => [ 'pip', 'pop', qr/bim|bam/ ],
    extra_validate_options => { foo => 'bar' }
  );

Runs all validations, then calls validation_report() on the
metrocluster.

=over

=item Options

=over

=item C< metrocluster >

The metrocluster to run this object on.

=item C< disaster_status >

The metrocluster's current disaster status as queried from
$mcc-E<gt>get_disaster_status( include_HA =E<gt> 1). If not provided, the
disaster status will be queried from the metrocluster.

=item C< die_on_failure >

If true, throws a Validation exception if there were any validation
failures. Defaults to true.

=item C< stop_on_failure >

If true, stops running validations after the first failed
validation. Defaults to false.

=item C< run_extended >

If true, runs validations marked as 'extended'. Puts a comment into
the log that since we are running extended validations, the
validations could take several minutes.

=item C< include >

Arrayref. If provided, we will only run validations whose validation
types match entries in this list. It can contain either strings or
regular expressions.

The include filter is applied before C< exclude >, so if you provide
arguments to both, anything in C< exclude > will be excluded from
items admitted by the C< include > filter.

=item C< exclude >

Arrayref. If provided, we will not run validations whose validation
types match items in the list. Items can be either plain strings or
regular expressions.

The exclude filter is applied after both C< include >, if provided,
and the C<exclude_classes> object attribute.

=item C< warn_only >

Arrayref. If provided, we will not throw an exception for any failed
validation whose validation type matches items in the list. The
validation will still be run. If the validation fails, a warning will
appear in the log with the error message, but no fatal exception will
be thrown. The validation report will show that validation as
failed. Items can be either plain strings or regular expressions.

=item C< extra_validate_options >

Hashref. Any options provided here are passed to the validate() call.

=item C< IGNORE_ALL_VALIDATION_FAILURES >
Global param to ignore all MCC validations failures. Default is set to 0

=back

=back

=cut

sub run_all_validations {
    my $self = shift;

    my %in = validate_with(
        params => \@_,
        spec   => {
            metrocluster     => {type       => OBJECT},
            disaster_status  => {optional   => 1},
            die_on_failure   => {regex      => qr/^[01]$/, default  => 1,},
            pre_dr_operation => {regex      => qr/^[01]$/, default  => 0,},
            stop_on_failure  => {regex      => qr/^[01]$/, default  => 0,},
            run_extended     => {regex      => qr/^[01]$/, default  => 0,},
            include          => {type       => ARRAYREF,   optional => 1,},
            exclude          => {type       => ARRAYREF,   optional => 1,},
            warn_only        => {type       => ARRAYREF,   default  => [], },
            extra_validate_options => { type => HASHREF,   default  => {}, },
        }
    );

    if ( defined $ENV{'IGNORE_ALL_VALIDATION_FAILURES'} ) {
        $in{die_on_failure} = 0;
        $in{stop_on_failure} = 0;
    }

    if ($in{'run_extended'}) {
        $Log->comment('run_all_validations() was called with run_extended, '
                . 'so it may take several minutes.');
    }

    my $disaster_status;
    if ($in{'disaster_status'}) {
        $disaster_status = $in{'disaster_status'};
    } else {
        $disaster_status = $in{'metrocluster'}
            ->get_disaster_status(include_HA => 1);
    }

    my @filtered_classes = $self->_get_filtered_classes(
        include => $in{'include'},
        exclude => $in{'exclude'},
        run_extended => $in{'run_extended'}
    );

    my @exceptions;

VALIDATION: foreach my $vtor_class (@filtered_classes) {

        my $vtor;
        if ($self->prepared()->{$vtor_class->validation_type()}) {
            $Log->comment('using prepared validation for ' .
                          $vtor_class->validation_type());
            #we prepared this type before, use that object instead of a new one
            $vtor = $self->prepared()->{$vtor_class->validation_type()};
            #update to current status rather than last dr state
            $vtor->disaster_status($disaster_status);
            #don't keep it around anymore
            delete $self->prepared()->{$vtor_class->validation_type()};
        } else {
            $vtor = $vtor_class->new(
                metrocluster    => $in{'metrocluster'},
                disaster_status => $disaster_status
            );
        }

        $Log->comment("Running validation " . $vtor->validation_type());

        try {
            $vtor->run(
                die_on_failure => $in{'die_on_failure'},
                pre_validate   => $in{'pre_dr_operation'},
                extra_validate_options => $in{'extra_validate_options'}
            );
        }
        catch NATE::BaseException with {
            my $exception = shift;

            $Log->warn(
                'Caught an exception in validation: ' . $exception->text());

            if ($self->_should_not_warn_only(
                    name => $vtor->validation_type(),
                    list => $in{'warn_only'}
                )
                ) {

                push(
                    @exceptions,
                    {   'name'    => $vtor->validation_type(),
                        'message' => $exception
                    }
                );

                if ($in{'stop_on_failure'}) {
                    last VALIDATION;
                }

            }
        }

        #prepare
        try {
            if ($vtor->can_prepare(
                    pre_dr_operation => $in{'pre_dr_operation'},)) {
                my $prepare = $vtor->prepare(
                    pre_dr_operation => $in{'pre_dr_operation'},
                    die_on_failure   => $in{'die_on_failure'},
                    extra_validate_options => $in{'extra_validate_options'}
                );
                #we need to grab this when running regular validation
                #instead of making a new one if it's there so save it
                $self->prepared()->{$vtor->validation_type()} = $vtor;
            }
        } catch NATE::BaseException with {
            my $exception = shift;

            $Log->warn(
                'Caught an exception in validation prepare: ' . $exception->text());

            if ($self->_should_not_warn_only(
                    name => $vtor->validation_type(),
                    list => $in{'warn_only'}
                )
                ) {

                push(
                    @exceptions,
                    {   'name'    => $vtor->validation_type(),
                        'message' => $exception
                    }
                );

                if ($in{'stop_on_failure'}) {
                    last VALIDATION;
                }

            }

        }
    }

    $in{'metrocluster'}->validation_report();

    if (@exceptions) {
        my $errmsg = 'Validation failed: '
            . join("\n",
            map {"Name: $_->{'name'}, Message: $_->{'message'}"} @exceptions);
        if ($in{'die_on_failure'}) {
            NACL::MTask::MCC::Exceptions::Validation->throw($errmsg);
        } else {
            $Log->warn($errmsg);
        }
    }

    if (wantarray) {
        return @exceptions;
    } else {
        # Return false if there are exceptions
        return ! scalar @exceptions;
    }

} ## end sub run_all_validations

sub _should_not_warn_only {
    my $self = shift;

    my %opts = validate_with(
        'params' => \@_,
        'spec'   => {
            'name' => { 'type' => SCALAR },
            'list' => { 'type' => ARRAYREF }
        }
    );

    # Return false if we match anything in the list
    foreach my $item ( @{ $opts{'list'} } ) {
        if ( ref($item) && ( ref $item eq 'Regexp' ) ) {
            $opts{'name'} =~ m/$item/ and return;
        } else {
            $opts{'name'} eq $item and return;
        }
    }

    # Otherwise, no match-- return true
    return 1;

}

sub _get_filtered_classes {
    my $self = shift;

    my %in = validate_with(
        params => \@_,
        spec   => {
            include           => {type => ARRAYREF|UNDEF, optional => 1},
            exclude           => {type => ARRAYREF|UNDEF, optional => 1},
            run_extended      => {regex      => qr/^[01]$/, default => 0,},
        }
    );

    my @vtor_classes = $self->get_all_validations();

    my @filtered_classes;

    VTOR_CLASS: foreach my $vtor_class ( @vtor_classes ) {

          # Skip extended classes unless called for
          if ( ! $in{'run_extended'} ) {
              $vtor_class->extended() and next VTOR_CLASS;
          }

          my $vtor_type = $vtor_class->validation_type();

        # Apply the include filter
        if ( defined $in{'include'} ) {

            $self->_apply_type_filter(
                type => $vtor_type,
                list => $in{'include'}
            ) or next VTOR_CLASS;

        }

        # Apply the exclude filter
        if ( defined $in{'exclude'} ) {

            $self->_apply_type_filter(
                type => $vtor_type,
                list => $in{'exclude'}
            ) and next VTOR_CLASS;

        }

        push(@filtered_classes, $vtor_class);

    }

    return @filtered_classes;
}

sub _apply_type_filter {
    my $self = shift;

        my %in = validate_with(
        params => \@_,
        spec   => {
            type => {type => SCALAR},
            list => {type => ARRAYREF}
        }
    );

    my $type = $in{'type'};

    foreach my $item ( @{ $in{'list'} } ) {
        if (ref $item eq 'Regexp') {
            $type =~ m/$item/ and return 1;
        }
        else {
            $type eq $item and return 1;
        }
    }

    return;

}

1;
