#
# Copyright (c) 2014 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary MetroCluster Storage Failover Show Validation
## @author dl-mcc-dpg-qa@netapp.com
## @status shared
## @pod here

package NACL::MTask::MCC::Validator::StorageFailoverShow;

use strict;
use warnings;

use parent qw(NACL::MTask::MCC::Validator::ClusterBase);

use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();

use Params::Validate
    qw(validate_with BOOLEAN HASHREF OBJECT ARRAYREF SCALAR UNDEF);

use NACL::CS::StorageFailover;

=head1 NAME

NACL::MTask::MCC::Validator::StorageFailoverShow

=head1 SYNOPSIS

  my $vtor = NACL::MTask::MCC::Validator::StorageFailoverShow->new(
    metrocluster => $mcc
  );

  $vtor->run();

=head1 DESCRIPTION

This class runs and validates the results of the 'storage failover
show' command.

=head2 Validation After SO

After SO, the "Takeover Possible" flag should be true for all nodes in the
surviving cluster.

=head2 Validation After HA

After HA, the "Takeover Possible" flag should be true for all nodes in the
surviving cluster.

=head2 Validation After HR

After HR, the "Takeover Possible" flag should be true for all nodes in the
surviving cluster.

=head2 Validation After SB

After SB, the 'Takeover Possible" flag should be true for all nodes in both
clusters.

=head1 METHODS

=head2 new()

  my $vtor = NACL::MTask::MCC::Validator::StorageFailoverShow->new(
     metrocluster => $mcc,
     disaster_status => $disaster_status
  );

Creates a new NACL::MTask::MCC::Validator::StorageFailoverShow object.

=over

=item Options

=over

=item C<metrocluster>

(Required.) A NACL::MTask::MCC::Metrocluster object on which we'll do
validations.

=item C<disaster_status>

Hashref. This should be the disaster status as reported by
$mcc-E<gt>get_disaster_status( include_HA =E<gt> 1 ). If
C<disaster_status> is not provided or is undefined, the validator will
call get_disaster_status() on the Metrocluster object itself.

=back

=back

(Copied from Base)

=head2 validate()

Runs the storage failover show command and validates the results.

If the Metrocluster's DR state is 'normal', checks that all nodes in
both clusters have 'Takeover Possible' set to true.

Otherwise, checks that all nodes in the surviving cluster has
'Takeover Possible' set to true.

=cut

sub validate {
    my $self = shift;

    if ( $self->is_normal() ) {
        $self->_validate_clusters(
            clusters => [ $self->all_components() ]
        );
    }
    else {
        $self->_validate_clusters(
            clusters => [ $self->survivor_components() ]
        );
    }

    return;
}

##
## _validate_clusters()
##
##   $vtor->validate_clusters(
##     clusters => \@cluster_components
##   );
##
## Runs 'storage failover show' on all clusters in
## @cluster_components, and records validation errors in errors() for
## any clusters without the 'Takeover Possible' flag set.
##
sub _validate_clusters {
    my $self = shift;

    my %opts = validate_with(
        params => \@_,
        spec   => {
            'clusters'        => {'type' => ARRAYREF }
        }
    );

    foreach my $cluster ( @{ $opts{'clusters'} } ) {

        my @storage_failover_cses =
            NACL::CS::StorageFailover->fetch(
                command_interface => $cluster->
                    get_Cluster_component()->
                        command_interface()
            );

        foreach my $cs ( @storage_failover_cses ) {
            $cs->enabled() eq 'true'
                or $self->errors_push(
                    q{Storage failover 'Takeover Enabled' flag for node '} .
                        $cs->node() . q{' was '} . $cs->enabled() . q{'}
                    );

        }

    }

    # Errors are recorded in errors(). We return void.
    return;

}

sub can_run_validation {

    my $self = shift;

    #don't run if this is a 2 node setup as there is no HA

    # arbitrary dr group
    my ($dr_group) = values(%{ $self->metrocluster()->get_DR_groups() });
    unless ($dr_group->is_4_pack()) {
        $Log->comment(
            '2 node testbed, skip storage failover validation...');
        return;
    }

    #do regular checks also
    return $self->SUPER::can_run_validation(@_);

}

=head1 SEE ALSO

=over

=item *

L<NACL::MTask::MCC::Validator::Base
|NACL::MTask::MCC::Validator::Base>

=item *

L<NACL::MTask::MCC::Validator::ClusterBase
|NACL::MTask::MCC::Validator::ClusterBase>

=back

=cut

1;
