# $Id: //depot/prod/test/nacldev/lib/NACL/MTask/SNMPTrapReceiver.pm $
#
## Copyright (c) 2001-2011 NetApp, Inc., All Rights Reserved
##   Any use, modification, or distribution is prohibited
##   without prior written consent from NetApp, Inc.
##
#

## @summary SNMPTrapReceiver Task Module
## @author bkprasha@netapp.com, dl-nacl-dev@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::MTask::SNMPTrapReceiver

=head1 DESCRIPTION

This library provides basic functionality to receive SNMP Traps

	use NACL::MTask::SNMPTrapReceiver;
    my $trap_receiver = NACL::MTask::SNMPTrapReceiver->new(command_interface => $client);
    my $single_trap = $trap_receiver->get_entry();
    my $traps = $trap_receiver->get_all();
    $trap_receiver->stop();
	
=head1 ATTRIBUTES

=head2 command_interface

(Required) See L<NACL::C::Component::command_interface|lib-NACL-C-Component-pm/command_interface>

=head2 traphost_ip

(Optional) The IP Address of the client listening to the SNMP Traps.  The IP 
Address can be specified in either ipv4 or ipv6 format

=head2 traphost_username

(Optional) The traphost username for the subtest to bind to the udp socket.
Defaults to 'root' in case not provided

=head2 port

(Optional) The port number on the client for SNMP Traps
	
=cut

package NACL::MTask::SNMPTrapReceiver;

use strict;
use warnings;

use Tharn qw(host snooze subtest $Parent  fatal_result);
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use NATE::Exceptions::Argument qw(:try);
use NATE::BaseException qw(:try);
use Params::Validate qw(validate validate_with OBJECT SCALAR ARRAYREF HASHREF);
use SNMP::TrapMgr qw(snmp_trap_mgr TRAP SNMP_TRAP_PORT);
use Data::Dumper;

=head1 METHODS

=head2 new

my $trap_receiver = NACL::MTask::SNMPTrapReceiver->new(command_interface => $client_ci);

Creates a new instance of the NACL::MTask::SNMPTrapReceiver object.  

=cut

use Class::MethodMaker [
    new    => [ '-hash', '-init', 'new' ],
    scalar => 'traphost_ip',
    scalar => 'traphost_username',
    scalar => [ { -default => '162' },     'port' ],
    scalar => 'domain',
    scalar => [ { -type    => 'Subtest' }, '_subtest' ],
    scalar => [ { -type => 'NACL::C::CommandInterface' }, 'command_interface' ],
    scalar => [ { -default => '120' }, 'snmp_message_timeout' ],
];

sub init {
    $Log->enter() if $may_enter;
    my ($self) = shift;

    if ( !( $self->command_interface() ) ) {
        $Log->exit() if $may_exit;
        NATE::Exceptions::Argument->throw(
            "Missing 'command_interface' argument!");
    }

    my ( $traphost_ipaddress, $port, $domain, @subTestArgs, $subtest,
        $clientobj, $traphost_username );

    if ( $self->traphost_ip() ) {
        $traphost_ipaddress = $self->traphost_ip();
    } else {
        $clientobj          = $self->command_interface()->hostrec();
        $traphost_ipaddress = $clientobj->default_ip();
    }

    $port = $self->port();
    if ( $traphost_ipaddress =~ /:/ ) {
        $domain = "udp/ipv6";
    } else {
        $domain = "udp/ipv4";
    }

    $self->domain($domain);
	
    if ( $self->traphost_username() ) {
		$traphost_username = $self->traphost_username();
	} else {
        $traphost_username = 'root';
    }
	
    my $code = '
	{
		use NACL::MTask::SNMPTrapReceiver;
				NACL::MTask::SNMPTrapReceiver->start(
										traphost_ip => "' . $traphost_ipaddress . '",
										domain => "' . $domain . '",
										port => ' . $port . ',
									);
	}';

    $subtest = subtest( $code, '-bg', '-client', $clientobj, '-username', $traphost_username );
    $self->_subtest($subtest);
	
    my $message =
      $self->_subtest()->message_get( type => "VERIFY", timeout => $self->snmp_message_timeout() );
  
    if ( $message !~ /SNMPREADY/ ) {
        $Log->debug("Error: Exiting as verifying 'SNMPREADY' failed!!! Message received is $message");
        $Log->exit() if $may_exit;
            NATE::BaseException->throw(
                "Error: Exiting as verifying 'SNMPREADY' failed!!! Message received is $message" );
    }
    
  $Log->exit() if $may_exit;
}

=head2 start

    $trap_receiver->start();

	NACL::MTask::SNMPTrapReceiver->start(
							traphost_ip => $traphost_ipaddress,
							domain => $domain,
							port => $port,
						);
						
(Class or Instance Method) The main NACL::MTask::SNMPTrapReceiver method. Calling this method causes
the trap to be collected and posted to the queue to be read.  This method works on the local client,
if called as a package call, to collect traps on a remote client it is suggested to make an instance
call

=over 

=item Options

=over 

=item C<< command_interface => $ci >>

(Required) See L<NACL::C::Component::command_interface|lib-NACL-C-Component-pm/command_interface>

=item C<< traphost_ip => $traphost_ip >>

(Optional for class methods, not applicable for instance method call) 
The IP Address of the client listening to the SNMP Traps.  The IP 
Address can be specified in either ipv4 or ipv6 format

=item C<< port => $port >>

(Optional for class methods, not applicable for instance method call) 
The port number on the client for SNMP Traps

=item C<< domain => $domain >>

(Optional for class methods, not applicable for instance method call)
Transport Domain to exchange SNMP messages between the local and remote devices
Accepts values such as 'udp6', 'udp/ipv6'; 'tcp', 'tcp4', 'tcp/ipv4'; 'tcp6', or 'tcp/ipv6'

=back

=back

=cut

sub start {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;

    my %opts = $pkg_or_obj->_common_validate_with( params => \@_ );

	if ( !( $opts{traphost_ip} ) ) {
        $Log->exit() if $may_exit;
        NATE::Exceptions::Argument->throw( "Missing 'traphost_ip' argument!"
              . " 'traphost_ip' argument is mandatory"
              . " if 'start' called via the package!!!" );
    }

    my ( $traphost_ip, $domain, $port );

    $traphost_ip = $opts{traphost_ip};
    $port        = $opts{port};
    $domain      = $opts{domain};
    if ( $traphost_ip =~ /:/ ) {
        $domain = "udp/ipv6";
    } else {
        $domain = "udp/ipv4";
    }

    $| = 1;
    $Log->comment(
"Listening for traps on addr($traphost_ip), port($port), domain($domain)"
    );

    if ( $domain =~ /ipv6/ ) {
        $Log->comment(
"    		Note: with domain($domain, SNMPv1 traps received will have 0.0.0.0 as agent-addr"
        );
    }

    $Parent->message_put( type => "VERIFY", message => "SNMPREADY" );

    # Start the trap manager
    snmp_trap_mgr(
        -localaddr => $traphost_ip,
        -domain    => $domain,
        -port      => $port,
        -callback  => [ \&_trap_received_cb ],
    );

    $Log->exit() if $may_exit;
}

sub _trap_received_cb {
    $Log->enter() if $may_enter;
    my ($self) = @_;

    my $ip = $self->hostname;
    my $received_time;

    if ( !$self->error ) {
        $received_time = _timestamp();
        if ( $self->type == TRAP ) {
            $Log->comment( "Received Trap-PDU from "
                  . $self->agent_addr
                  . " at $received_time :\n" );
        } else {
            $Log->comment( "Received SNMPv2-Trap-PDU from "
                  . $self->agent_addr
                  . " at $received_time :\n" );
        }

        my $trap = {
            received_time => $received_time,
            community     => $self->community,
            enterprise    => $self->enterprise,
            agent_address => $self->agent_addr,
            generic_trap  => $self->generic_trap,
            specific_trap => $self->specific_trap,
            up_time       => $self->time_stamp,
            var_binds     => $self->var_bind_list,
            error_status  => $self->error_status,
            error_index   => $self->error_index,
        };
        $Parent->message_put( type => "SNMP_Trap", message => $trap );
    } else {
        $Parent->message_put(
            type    => "SNMP_Trap_Error",
            message => $self->error,
        );
    }
    $Log->exit() if $may_exit;
}

sub _timestamp {
    $Log->enter() if $may_enter;
    my @t    = localtime;
    my $time = sprintf(
        "%02d-%02d-%04d %02d:%02d:%02d",
        $t[4] + 1,
        $t[3], $t[5] + 1900,
        $t[2], $t[1], $t[0]
    );
    $Log->exit() if $may_exit;
    return $time;

}

=head2 stop

    $trap_receiver->stop();

(Instance Method) Stops a NACL::MTask::SNMPTrapReceiver instance

=cut

sub stop {
    $Log->enter() if $may_enter;
    my ($pkg) = @_;
    $Log->trace("Stopping the server...\n");
    $pkg->_subtest()->kill();
    snooze(10)
      ;    #wait a few seconds to ensure the process dies and the port is free

    $Log->exit() if $may_exit;
    return;
}

=head2 get_entry

    $single_trap = $trap_receiver->get_entry();

	(Instance Method) Gets a single trap entry that was sent to the NACL::MTask::SNMPTrapReceiver
	
	This method returns a single trap message as a hash reference 
	which is of the format
	
	$single_trap = {
			'up_time' 		=> '12 days, 02:41:55.32',
			'enterprise' 	=> '1.3.6.1.4.1.789',
			'error_index' 	=> '0',
			'generic_trap' 	=> '6',
			'var_binds' 	=> {
					'1.3.6.1.4.1.789.1.1.12.0' => 'Volume newsnmp1234@vserver:02f52cbf-f58a-11e0-8c2e-123478563412 is offline. ',
					'1.3.6.1.4.1.789.1.1.9.0' => '123'
				},
			'received_time' => '10-13-2011 16:28:10',
			'community' 	=> 'public',
			'error_status' 	=> '0',
			'agent_address' => '10.72.42.176',
			'specific_trap' => '275'
		}

=cut

sub get_entry {
    $Log->enter() if $may_enter;
    my ($self) = @_;
    my $message = undef;

    #get the single message from the queue
    $message =
      $self->_subtest()->message_get( type => "SNMP_Trap", timeout => $self->snmp_message_timeout() );
    $Log->debug( sub { "Found raw message:  " . Dumper($message) } );
    $Log->exit() if $may_exit;
    return $message;
}

=head2 get_all

    $traps = $trap_receiver->get_all();

	(Instance Method) Get all entries that were sent to the NACL::MTask::SNMPTrapReceiver
	
	This method returns all the traps in the form of an array of hashes. 

	$traps = [
		{
			'up_time' 		=> '12 days, 02:41:55.32',
			'enterprise' 	=> '1.3.6.1.4.1.789',
			'error_index' 	=> '0',
			'generic_trap' 	=> '6',
			'var_binds' 	=> {
					'1.3.6.1.4.1.789.1.1.12.0' => 'Volume newsnmp1234@vserver:02f52cbf-f58a-11e0-8c2e-123478563412 is offline. ',
					'1.3.6.1.4.1.789.1.1.9.0' => '123'
				},
			'received_time' => '10-13-2011 16:28:10',
			'community' 	=> 'public',
			'error_status' 	=> '0',
			'agent_address' => '10.72.42.176',
			'specific_trap' => '275'
		},
		
	
		]
	
	
=cut

sub get_all {
    $Log->enter() if $may_enter;
    my ($self) = @_;
    my @msgConcat = ();
    my $msgTmp    = undef;

    while (1) {
        $msgTmp = undef;

        #get an entry that is on the queue, a message reported from the server
        eval { $msgTmp = $self->get_entry(); };

        #if no message available, then exit
        if ( !defined $msgTmp ) {
            last;
        }
        else {    #else add the message to the array
            push @msgConcat, $msgTmp;
        }
    }

    #return the array of messages
    $Log->exit() if $may_exit;
    return \@msgConcat;
}

=head2 get_error

    $error_message = $trap_receiver->get_error();

	(Instance Method) Gets a error message that was sent to the NACL::MTask::SNMPTrapReceiver
	
=cut

sub get_error {
    $Log->enter() if $may_enter;
    my ($self) = @_;
    my $error_message = undef;

    #get error message from the queue
    $error_message =
      $self->_subtest()->message_get( type => "SNMP_Trap_Error", timeout => $self->snmp_message_timeout() );
    $Log->debug( "Found error message:  " . $error_message );
    $Log->exit() if $may_exit;
    return $error_message;
}

sub _common_validate_with {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params => \@_,
        spec   => {
            params          => { type     => ARRAYREF },
            additional_spec => { type     => HASHREF, default => {} },
            allow_extra     => { optional => 1 },
        },
    );
    my %params = @{ $opts{params} };
    my %spec   = %{ $opts{additional_spec} };

    my @keys = qw (traphost_ip port domain);

    if ( ref $pkg_or_obj ) {
        foreach my $key (@keys) {
            my $key_isset = "${key}_isset";
            if ( $params{$key} && $pkg_or_obj->$key_isset() ) {
                $Log->exit() if $may_exit;
                NATE::Exceptions::Argument->throw(
                        "Option '$key' not allowed (it is already an "
                      . 'attribute of this object' );
            } ## end if ( $params{$key} )
            if ( $pkg_or_obj->$key_isset() ) {
                my $value = $pkg_or_obj->$key();
                $params{$key} = $value;
            }
        } ## end foreach my $key (@keys)
    } ## end if ( ref $pkg_or_obj )

    $Log->exit() if $may_exit;
    return validate_with(
        params => \%params,
        spec   => {
            %spec,
            traphost_ip => { type => SCALAR, optional => 0 },
            port        => { type => SCALAR, optional => 1 },
            domain      => { type => SCALAR, optional => 1 },
        },
        allow_extra => 1,
    );
} ## end sub _common_validate_with

1;
