# $Id: $
#
## Copyright (c) 2014 NetApp, Inc., All Rights Reserved
##   Any use, modification, or distribution is prohibited
##   without prior written consent from NetApp, Inc.
##

#
## @summary WAFL Ontap Setup MTask Module
## @author dl-awg-warp-svl
## @change
##
## @status shared
## @pod here

package NACL::MTask::Setup::Object::Snapmirror;

use strict;
use warnings;
use parent qw(NACL::MTask::Setup::ObjectFactory);

use Params::Validate qw(validate_with :types);

use NACL::STask::Snapmirror;

use NATE::Exceptions::Argument;

use NACL::GeneralUtils qw(random_name_generator);
use NATE::ParamSet qw(param_global);
use NATE::Log qw(log_global);

use Class::MethodMaker [
    array => [{-type => 'NACL::STask::Snapmirror'}, 'snapmirrors'],
    array => [{-type => 'NACL::STask::Volume'},     'source_volumes'],
    array => [{-type => 'NACL::STask::Volume'},     'destination_volumes'],
    scalar => [{ -type => 'NACL::Cleanup' }, 'cleanup_manager'],
    scalar => [qw/is_fanout node create_options/],
];

sub OBJECT_TYPE {return 'Snapmirror'}

my $Log = log_global;

sub new {
    $Log->enter;

    my $pkg = shift;

    my %opts = $pkg->_common_validate_with(
        params          => \@_,
        additional_spec => {
            node                => {type => OBJECT},
            cleanup_manager     => {isa => 'NACL::Cleanup', required => 1},
            source_volumes      => {type => ARRAYREF},
            destination_volumes => {type => ARRAYREF},
            create_options      => {type => HASHREF, default => {}},
        },
        ignore_primary_keys  => 1,
        no_command_interface => 1,
    );

    my $self = $pkg->SUPER::new(type => OBJECT_TYPE);

    $self->node(delete $opts{node});
    $self->source_volumes(@{delete $opts{source_volumes}});
    $self->destination_volumes(@{delete $opts{destination_volumes}});
    $self->cleanup_manager(delete $opts{cleanup_manager});

    $self->_create($opts{create_options});

    $Log->exit;
    return $self;
} ## end sub new

sub _create {
    my $self = shift;
    my ($create_options) = @_;

    $self->_set_fanout;

    #############################################
    $Log->step('Creating mirrors');
    #############################################
    my @dst_vols = $self->destination_volumes;

    my $creation_options = $self->_process_create_options($create_options);

    foreach my $i (0 .. scalar(@dst_vols) - 1) {
        my $src_vol_ind = ($self->is_fanout) ? 0 : $i;
        $self->SUPER::_create(
            create_options => {
                'source-path'      => $self->source_volumes->[$src_vol_ind],
                'destination-path' => $dst_vols[$i],
                nacltask_cleanup_manager => $self->cleanup_manager,
                %{$creation_options},
            }
        );
    }
} ## end sub _create

sub _set_fanout {
    my $self = shift;

    if (scalar(@{$self->source_volumes})
        < scalar(@{$self->destination_volumes})) {
        $Log->comment('Fanout set');
        $self->is_fanout(1);
    }
} ## end sub _set_fanout

1;

__END__

=pod

=head1 NAME

NACL::MTask::Setup::Object::Snapmirror

=head1 DESCRIPTION

This library creates one or more snapmiror relationships

=head1 ATTRIBUTES

=head2 nodes

STask Node object used for creating snapmirror relationships

=head2 source_volumes

STask volume objects used as source for relationship creation

=head2 destination_volumes

STask volume objects used as destination for relationship creation

=head1 METHODS

=head2 new

    my $mirror_obj = NACL::MTask::Setup::Object::Snapmirror->new(
        node => $nodes,
        source_volumes      => \@vols,
        destination_volumes => \@dest_vols,
        create_options => $create_options,
    );
(Instance method)

=over

=item Options

=over

=item C<< node => $node >>

(Required)
STask node object to be used as the command interface

=item C<< source_volumes => \@src_vols >>

(Optional, Arrayref)
List of Volume STask objects to be used as source

=item C<< destination_volumes => \@dst_vols >>

(Optional, Arrayref)
List of Volume STask objects to be used as destination
If number of destination vols is greater than source a fanout is created

=item C<< create_options => $mirror_create_options >>

(Optional, Hashref)
A hashref contianing the inputs for Snapmirror creation. It takes all the inputs given for
the create function of L<NACL::STask::Snapmirror|lib-NACL-STask-Snapmirror-pm>.
Refer L<NACL::STask::Snapmirror|lib-NACL-STask-Snapmirror-pm> for default values.

=back

=back

=cut
