# $Id: $
#
## Copyright (c) 2014 NetApp, Inc., All Rights Reserved
##   Any use, modification, or distribution is prohibited
##   without prior written consent from NetApp, Inc.
##

#
## @summary WAFL Ontap Setup MTask Module
## @author dl-awg-warp-svl
## @change
##
## @status shared
## @pod here

package NACL::MTask::Setup::Object::Volume;

use strict;
use warnings;
use parent qw(NACL::MTask::Setup::ObjectFactory);
use NACL::MTask::Setup::GlobalParams qw(:Volume);

use Params::Validate qw(:types);

use NACL::STask::Volume;
use NACL::STask::VolumeClone;

use NATE::Exceptions::Argument;

use NATE::ParamSet qw(param_global);
use NATE::Log qw(log_global);

use Class::MethodMaker [
    scalar => [{-type => 'NACL::STask::Node'},   'node',],
    scalar => [{ -type => 'NACL::Cleanup' }, 'cleanup_manager'],
    array  => [{-type => 'NACL::STask::Volume'}, 'volumes'],
    array  => [{-type => 'NACL::STask::VolumeClone'}, 'volume_clones'],
    scalar => [qw/count /],
];

sub OBJECT_TYPE {return 'Volume'}
sub DEFAULT_SIZE {return '20M'}

my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();

sub new {
    $Log->enter() if $may_enter;

    my $pkg = shift;

    my %opts = $pkg->_common_validate_with(
        params          => \@_,
        additional_spec => {
            node  => {type => SCALAR},
            cleanup_manager =>  { isa => 'NACL::Cleanup', required => 1},
            count => {
                type    => SCALAR,
                regex   => qr/^\d+$/,
                default => 1,
            },
            create_options => {type => HASHREF, default => {}},
            find_options   => {type => HASHREF, default => {}},
        },
        ignore_primary_keys  => 1,
        no_command_interface => 1,
    );

    return if ($opts{count} == 0);

    my $self = $pkg->SUPER::new(type => OBJECT_TYPE,);

    $self->node(delete $opts{node});
    $self->cleanup_manager(delete $opts{cleanup_manager});

    $self->_populate_count_and_size(\%opts);

    $self->_process_global_params;

    $self->_skip_find(1) if (!scalar(@{$self->names_from_global_param}));

    $self->_process_options_and_find_or_create($opts{create_options},
        $opts{find_options});

    $Log->exit() if $may_exit;
    return $self;
} ## end sub new

sub volume_clone {
    my $self = shift;

    $self->volume_clones(
        map { $_->clone_create(nacltask_cleanup_manager => $self->cleanup_manager) }
    $self->volumes);
}

sub volume_move {
    my $self = shift;

    map { $_->move(nacltask_cleanup_manager => $self->cleanup_manager) } ($self->volumes);
}

## private subroutines
sub _process_global_params {
    my $self = shift;

    $self->_cache_object_names_from_global_param($VOLUME_PARAM_FOR{NAMES});
}

sub _compute_vol_size {
    my $self = shift;
    my $opts = shift;

    my ($vol_count, $vol_size) = $self->_populate_count_and_size($opts);

    ## Separating the integers and unit and computing total size
    my ($size, $unit) = $vol_size =~ /(\d*)(.*)/;
    my $total_vol_size = $size * $vol_count . $unit;
    $Log->comment("Total volume size needed: $total_vol_size");

    return ($total_vol_size);
} ## end sub _compute_vol_size

sub _populate_count_and_size {
    my $self = shift;
    my $opts = shift;

    my $param = param_global();

    my $vol_count = $param->get($VOLUME_PARAM_FOR{COUNT}) || $opts->{count} || 1;
    $self->count($vol_count);

    # Setting default size for volume as 20M
    my $vol_size
        = $param->get($VOLUME_PARAM_FOR{SIZE})
        || $opts->{create_options}->{size}
        || DEFAULT_SIZE;
    $opts->{create_options}->{size} = $vol_size;

    return ($vol_count, $vol_size);
} ## end sub _populate_count_and_size

1;

__END__

=pod

=head1 NAME

NACL::MTask::Setup::Object::Volume

=head1 DESCRIPTION

This library finds or creates one or more volumes. Parameters to the library
can be overridden by global parameters. The list of global parameters for this
library are documented here
L<NACL::MTask::Setup::GlobalParams#Volume|lib-NACL-MTask-Setup-GlobalParams-pm/Volume>

=head1 ATTRIBUTES

=head2 node

STask Node object used for creating vserver

=head2 volumes

STask Volume STask objects found or created

=head1 METHODS

=head2 new

    my $vol_obj = NACL::MTask::Setup::Object::Volumes->new(
        node => $nodes,
        count => 2,
        create_options => $create_options,
        find_options => $find_options,
    );
(Instance method)

=over

=item Options

=over

=item C<< node => $node >>

(Required)
STask node object to be used as the command interface

=item C<< count => 2 >>

(Optional)
The number of Volumes needed

=item C<< find_options => $volume_find_options >>

(Optional, Hashref)
Parameters to find the volumes. These are parameters accepted by STask

See L<NACL::C::Volume::find()|lib-NACL-C-Volume-pm/find>
Ignore the command_interface as node is used as the command interface

=item C<< create_options => $volume_create_options >>

(Optional, Hashref)
A hashref contianing the inputs for Volume creation. These are parameters accepted by STask

See L<NACL::STask::Volume|lib-NACL-STask-Volume-pm>.
Ignore the command_interface as node is used as the command interface

=back

=back

=cut
