# $Id: $
#
## Copyright (c) 2014 NetApp, Inc., All Rights Reserved
##   Any use, modification, or distribution is prohibited
##   without prior written consent from NetApp, Inc.
##

#
## @summary WAFL Ontap Setup MTask Module
## @author dl-awg-warp-svl
## @change
##
## @status shared
## @pod here

package NACL::MTask::Setup::_Flexgroup;

use strict;
use warnings;
use parent qw(NACL::MTask::Setup);
use NATE::Log qw(log_global);
use Params::Validate qw(validate_with :types);

use NATE::Exceptions::Argument;

use Class::MethodMaker [ new => 'new', ];

my $Log = log_global;

sub create {
    my $self = shift;

    $self->_verify_invocation( style => 'instance_only' );
    my %opts = @_;

    $opts{aggregate_options}{is_flexgroup} = 1;
    $self->SUPER::create(%opts);

} ## end sub create

sub _set_opts_find_or_create_aggrs_set_attributes {
    my $self = shift;
    my ( $aggr_options, $node, $vol_create_options ) = @_;

    if ( $vol_create_options->{'aggr-list'} ) {
        $Log->comment( 'aggr-list is set in the volume create options'
              . 'Skipping aggreate find or create' );
        return;
    }
    else {
        $self->SUPER::_set_opts_find_or_create_aggrs_set_attributes(@_);
    }
}

sub _set_opts_find_or_create_vols {
    my $self = shift;
    my ( $node, $volume_options ) = @_;
    $self->_set_volume_create_options(
        $volume_options,
        {
            vserver     => $self->vservers->[0]->vserver,
            'aggr-list' => [ map { $_->aggregate } $self->aggregates ],
        }
    );

    return $self->_find_or_create_volumes( $volume_options, $node );

} ## end sub _set_opts_find_or_create_vols

1;
__END__

=pod

=head1 NAME

NACL::MTask::Setup::_Flexgroup

=head1 DESCRIPTION

This is library does a flexgroup setup. It sets up aggregates, vservers,
source volumes, desintation volumes and relationships

=head1 ATTRIBUTES

=head2 nodes

List of STask node objects found

=head2 aggregates

List of STask aggregate objects found or created

=head2 vservers

List of STask vserver objects found or created

=head2 volumes

List of STask volume objects(flexgroups) found or created

=head1 METHODS

=head2 create

    my $setup = NACL::MTask::Setup->new( type => $Flexgroup );

    $setup->create(
        run_pre_setup  => 0,
        volume_options => {
            ## To make sure 2 Aggrs are selected
            create_options => { size => $aggr_availsize + 10240 }
        },
    );
(Instance method)


This method does the following

=over

=item * Finds nodes

=item * Perform pre setup operations based on global params

=item * Find or create Aggregates

=item * Find or create Vservers

=item * Find or Create Flexgroups

=item * Perform post setup operations based on global params

=item * Perform pre cleanup operations based on global params

=back

=over

=item Options

=over

=item C<< node_options => $node_setup_options >>

(Optional, Hashref)
Parameters taken by node setup MTask.

See L<NACL::MTask::Setup::Object::Node::new|lib-NACL-MTask-Setup-Object-Node-pm/new>

=item C<< aggregate_options => $aggregate_options >>

(Optional, Hashref)
Parameters taken by aggregate setup MTask.

See L<NACL::MTask::Setup::Object::Aggregate::new|lib-NACL-MTask-Setup-Object-Aggregate-pm/new>

=item C<< vserver_options => $vserver_options >>

(Optional, Hashref)
Parameters taken by vserver setup MTask.

See L<NACL::MTask::Setup::Object::Vserver::new|lib-NACL-MTask-Setup-Object-Vserver-pm/new>

=item C<< volume_options => $volume_options >>

(Optional, Hashref)
Parameters taken by volume setup MTask.

See L<NACL::MTask::Setup::Object::Volume::new|lib-NACL-MTask-Setup-Object-Volume-pm/new>

=item C<< run_pre_setup => 0|1 >>

(Optional, Boolean)
Runs pre setup if set to 1. Default(1)

=back

=back

=head2 post_setup

$setup->post_setup;

(Instance method)

Post setup operations are controlled by global parameters. The 
post setup operations for objects can be found here L<NACL::MTask::Setup::GlobalParams#Post setup|lib-NACL-MTask-Setup-GlobalParams-pm/Post setup tasks>
Operations are performed on Nodes, Aggregates, Vservers, Volumes

=head1 Pre setup operations

Pre setup operations are controlled by global parameters. The 
pre setup operations can be found here L<NACL::MTask::Setup::GlobalParams#Pre setup|lib-NACL-MTask-Setup-GlobalParams-pm/Pre setup tasks>
Operations are performed on Nodes, Aggregates, Vservers, Volumes

=head1 Pre cleanup operations

Pre cleanup operations are controlled by global parameters. The 
pre cleanup operations can be found here
L<NACL::MTask::Setup::GlobalParams#Pre cleanup|lib-NACL-MTask-Setup-GlobalParams-pm/Pre cleanup tasks>
Operations are performed on Nodes, Aggregates, Vservers, Volumes

=cut
