#
# Copyright (c) 2011-2012 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary MTask module to manage/manipulate sets of snapmirror relationships
## @author rjray@netapp.com
## @status shared
## @pod here

package NACL::MTask::SnapmirrorSet;

# We're working in straight ASCII, so we don't need this one:
## no critic(ProhibitEnumeratedClasses)
# We want to use AUTOLOAD (for now)
## no critic(ProhibitAutoloading)
# The LS-specific hooks passed to calls to _execute() trigger this:
## no critic(ProhibitUnusedPrivateSubroutines)
# The NACL habit of using/exporting _-names triggers this:
## no critic(ProtectPrivateSubs)

use strict;
use warnings;
use base qw(NACL::MTask::MTask);

use Scalar::Util qw(blessed);
use Params::Validate qw(
    validate validate_with validation_options
    SCALAR ARRAYREF HASHREF OBJECT UNDEF BOOLEAN
);

use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use NACL::ComponentUtils qw(_verify_invocation);
use NACL::STask::Volume;
use NACL::STask::Snapmirror;
use NACL::CS::ClusterIdentity;
use NATE::BaseException qw(:try);
use NATE::Exceptions::Argument;
use NACL::Exceptions::SnapmirrorSetFailure;

# Used to test the "type" parameter in several methods, so just define it once
# up here:
my %SM_TYPES = (
    LS  => 1,
    DP  => 1,
    XDP => 1,
);
# Used in error messages. Must match the keys above:
my $SM_TYPES = 'LS, DP, XDP';

# Set the on_fail behavior for Params::Validate to throw an exception (of the
# NATE::Exceptions::Argument class) on errors instead of a plain croak().
validation_options(
    on_fail => sub {
        NATE::Exceptions::Argument->throw(shift);
    }
);

# These are spec-fragments that are used in multiple places in various calls
# to validate() and/or validate_with(). Putting them here cuts down on
# complexity of individual subroutines, and speeds up calls to validation by
# having the specs already computed.
#
# This is the spec for the "destinations" option to new(), add_to_set(), etc.
my $DESTINATIONS_SPEC = {
    type      => HASHREF,
    optional  => 1,
    callbacks => {
        'correct keys' => sub {
            my $ref = shift;
            # We've already established it's a hashref, now check that the keys
            # are limited to those of %SM_TYPES:
            for my $key (keys %{$ref}) {
                if (! $SM_TYPES{$key}) {
                    return 0;
                }
            }
            return 1;
        },
    },
};
# This is the spec for the "mirrors" option to new(), add_to_set(), etc.
my $MIRRORS_SPEC = {
    type      => ARRAYREF,
    optional  => 1,
    callbacks => {
        'content type check' => sub {
            my $ref = shift;
            # We've already established that it's an arrayref, now check that
            # all elements are Snapmirror objects:
            for my $elem (@{$ref}) {
                if (! ((blessed $elem) &&
                       $elem->isa('NACL::C::Snapmirror'))) {
                    return 0;
                }
            }
            return 1;
        },
    },
};
# This is the spec for options "LS", "DP", etc. to create(), add_to_set(), etc.
my $VOLUME_TYPE_SPEC = {
    type      => ARRAYREF | HASHREF,
    optional  => 1,
    callbacks => {
        'required type or contents' => sub {
            my $ref = shift;
            # If $ref isn't an arrayref, make it into one:
            if (ref($ref) ne 'ARRAY') {
                $ref = [ $ref ];
            }
            # Ensure that each element is a volume object or a volume-spec
            # hashref:
            for my $elem (@{$ref}) {
                return 0 if
                    (((blessed $elem) && ! $elem->isa('NACL::C::Volume')) &&
                     (ref($elem) ne 'HASH'));
            }
            # If we got through the loop, it's valid:
            return 1;
        },
    },
};
# This is for the spec for arguments "type", to mirrors(), destinations(), etc.
my $TYPE_SPEC = {
    type      => SCALAR,
    optional  => 1,
    callbacks => {
        'valid type' => sub {
            my $type = shift;
            # Is $type a key in %SM_TYPES?
            return $SM_TYPES{uc $type};
        },
    },
};
# This is for the spec for the argument "type" when it can be singular or
# plural (scalar or arrayref).
my $PLURAL_TYPES_SPEC = {
    type      => SCALAR | ARRAYREF,
    optional  => 1,
    callbacks => {
        valid => sub {
            my $value = shift;
            if (! ref($value)) {
                $value = [ $value ];
            }
            for my $type (@{$value}) {
                return 0 if (! $SM_TYPES{$type});
            }
            return 1;
        },
    },
};

# Ideally, I will eventually replace this autoloader with a set of dynamically-
# created methods based on a known list of valid snapmirror actions. For now,
# just use AUTOLOAD until we have that list.
our $AUTOLOAD;
sub AUTOLOAD {
    (my $method = $AUTOLOAD) =~ s/.*://;

    # This creates a closure around $method and makes it into the method
    # itself. A second or later call to $method won't even go through
    # AUTOLOAD, it will go directly to this new method:
    ## no critic (ProhibitNoStrict)
    no strict 'refs';
    *{$method} = sub {
        $Log->enter() if $may_enter;
        my $self = shift;

        $Log->exit() if $may_exit;
        return $self->execute(action => $method, @_);
    };

    goto &{$method};
}

=head1 NAME

NACL::MTask::SnapmirrorSet - Manage sets of snapmirror relationships

=head1 SYNOPSIS

    use NACL::STask::Volume;
    use NACL::STask::Snapmirror;
    use NACL::MTask::SnapmirrorSet;

    # Assume $source is already set to a volume created as RW, and
    # @dests is a list of one or more volumes created as DP.
    my $set = NACL::MTask::SnapmirrorSet->new(
        source       => $source,
        destinations => { LS => \@dests }
    );

    # $set now has created LS-type snapmirror relationships for each
    # volume in @dests, with $source as the source volume for all.

    # Initialize all the snapmirrors at once:
    $set->initialize;

    # Update all the snapmirrors at once:
    $set->update;

=head1 DESCRIPTION

The B<NACL::MTask::SnapmirrorSet> class is an adaptation of the GNOTE module
B<GNOTE::GX::SnapMirrorSet>. It is designed to facilitate the management of
groups of snapmirror relationships as a single entity, all with an identical
source-point, generally controlled and referenced by that source-point.  The
primary goal of the B<SnapMirrorSet> design was originally to enable management
of C<LS> mirror sets. However, it was later extended to allow and manage C<DP>
mirrors in the set. The B<NACL::MTask::SnapmirrorSet> class supports C<XDP>
mirrors as well.

A B<NACL::MTask::SnapmirrorSet> object is created from a single source object
of the B<NACL::STask::Volume> class or B<NACL::C::Volume> class, and zero or
more destination objects. The destination objects can be either other volume
objects, or a set of snapmirror objects (either B<NACL::C::Snapmirror> or
B<NACL::STask::Snapmirror>). Volumes and snapmirrors cannot currently be mixed
in a single creation call. See the documentation for L</new>, below, for more
detail.

The purpose of having a B<NACL::MTask::SnapmirrorSet> object is to perform the
various actions associated with snapmirror relationships (initialize, update,
etc.) on the set as a whole, rather than having to call the methods on the
individual B<NACL::STask::Snapmirror> objects. In the case of sets of C<LS>
relationships, some of the commands are called only once using the source-point
of the relationship and a slightly different command syntax. The goal is to
abstract these differences away from the user, and allow them to treat these
sets the way they would treat individual B<NACL::STask::Snapmirror> objects.

=head1 ATTRIBUTES

Each B<NACL::MTask::SnapmirrorSet> object has the following accessible
attributes.  All of these are immutable attributes; they cannot be changed via
these accessors.  They can only be changed through the methods described
later. The B<source> attribute cannot be changed for a
B<NACL::MTask::SnapmirrorSet> object at all (unless it is changed internally by
a call to the B<promote> method, see L</promote>).

=head2 source

This is the B<NACL::STask::Volume> that is the source side of all the
snapmirror relationships in the set. By definition of the set, no snapmirror
relationships in the set will have a different source-point than this.

=cut

###############################################################################
#
#   Sub Name:       source
#
#   Description:    Return the "source" attribute of this object
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#
#   Returns:        Success:    NACL::STask::Volume instance
#                   Failure:    throws exception
#
###############################################################################
sub source {
    $Log->enter() if $may_enter;
    my $self = shift;

    $Log->exit() if $may_exit;
    return $self->{source};
}

=head2 destinations

This is a collection of B<NACL::STask::Volume> objects that represent the
destination side of all the snapmirror relationships in the set. By definition
of the set, no destination B<NACL::STask::Volume> can represent more than one
relationship.  This should match ONTAP, where any given (source, destination)
tuple can only have one snapmirror relationship.

=over

=item Options

If the accessor is called with no arguments, the return value is a hash
reference in which the keys are the types (C<LS>, C<DP> and C<XDP>) and the
values are array references of B<NACL::STask::Volume> objects. All three keys
will be present, even if the array reference is empty.

=over

=item C<< type =E<gt> $mirror_type >>

If the accessor is called with an option of C<type>, the value of the option
must be one of C<LS>, C<DP> or C<XDP> (case is not important). The return value
is then an array reference containing zero or more B<NACL::STask::Volume>
objects that are the endpoints of the snapmirror relationships of that type.

=back

Note that while specifying C<type> is case-insensitive, the keys of the hash
reference returned by calling B<destinations> with no arguments uses
all-uppercase keys. Thus, the following three statements are identical:

    $ls_dests = $set->destinations(type => 'ls');

    $ls_dests = $set->destinations(type => 'LS');

    $ls_dests = $set->destinations->{LS};

All three return the array reference of C<LS> relationship end-points.

=item Exceptions

If the B<type> option is passed but is not a valid type, an exception of
class B<NATE::Exceptions::Argument> will be thrown. If an unknown option is
passed, a B<NATE::Exceptions::Argument> will be thrown.

=back

=cut

###############################################################################
#
#   Sub Name:       destinations
#
#   Description:    Return some or all of the volume objects that are the
#                   destination volumes for the snapmirror relationships
#                   managed by this object.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#                   @opts     in      array     Parameters to this call
#
#   Globals:        %SM_TYPES
#                   $TYPE_SPEC
#
#   Returns:        Success:    listref or hashref
#                   Failure:    throws exception
#
###############################################################################
sub destinations {
    $Log->enter() if $may_enter;
    my ($self, @opts) = @_;
    my $dests;

    my %opts = validate(
        @opts,
        {
            type => $TYPE_SPEC,
        }
    );

    if ($opts{type}) {
        # The only allowed option in this method is 'type', and if it was
        # passed it was validated earlier:
        $dests = [ @{$self->{destinations}->{uc $opts{type}}} ];
    } else {
        # If no options are passed, return a hash-table of all the destinations
        # keyed by the type:
        $dests = {};
        for my $type (keys %SM_TYPES) {
            $dests->{$type} = [ @{$self->{destinations}->{$type}} ];
        }
    }

    $Log->exit() if $may_exit;
    return $dests;
}

=head2 mirrors

This is a collection of B<NACL::STask::Snapmirror> objects that represent the
actual snapmirror relationships being managed by this set. The number of objects
here will always match the total number of B<NACL::STask::Volume> objects
contained in the B<destinations> set, described above.

=over

=item Options

If the accessor is called with no arguments, the return value is a hash
reference whose keys are the volume paths of the B<NACL::STask::Volume> objects
that represent the destination end-points. The value of each key is the
B<NACL::STask::Snapmirror> object that represents the relationship between that
end-point and B<source>.

=over

=item C<< type =E<gt> $mirror_type >>

If the accessor is called with an option of C<type>, the value of the option
must be one of C<LS>, C<DP> or C<XDP> (case is not important). The return value
is then an array reference containing zero or more B<NACL::STask::Snapmirror>
objects that are the relationships of that type in the set.

=item C<< destination =E<gt> $volume_or_uuid_or_path >>

If the accessor is called with an option of C<destination>, the value of the
option must be either a volume object, the UUID of a volume or a volume
path. The return value is then the single B<NACL::STask::Snapmirror> object
that represents a relationship between that object/path and B<source>. If there
is no such relationship in the set, an exception is thrown.

=back

If both C<type> and C<destination> are passed, C<destination> is used as it
is more specific.

=item Exceptions

If C<type> is passed but is not a valid mirror type, an exception of the class
B<NATE::Exceptions::Argument> is thrown. If C<destination> is passed but no
matching mirror is found, an exception of the same class is thrown.

=back

The following are equivalent:

    $ls_mirrors = $set->mirrors(type => 'LS');

    $ls_mirrors = [];
    for my $endpoint (@{$set->destinations(type => 'LS')}) {
        push @{$ls_mirrors}, $set->mirror(destination => $endpoint);
    }

Both result in C<$ls_mirrors> being an array reference of zero or more
B<NACL::STask::Snapmirror> objects that are LS mirrors of B<source>.

=cut

###############################################################################
#
#   Sub Name:       mirrors
#
#   Description:    Return all or a sub-set of the NACL::STask::Snapmirror
#                   objects that this set object is managing.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#                   @opts     in      array     Parameters to the call
#
#   Globals:        %SM_TYPES
#                   $SM_TYPES
#                   $TYPE_SPEC
#
#   Returns:        Success:    hashref, listref or NACL::STask::Snapmirror obj
#                   Failure:    throws exception
#
###############################################################################
sub mirrors {
    $Log->enter() if $may_enter;
    my ($self, @opts) = @_;
    my $value;

    my %opts = validate(
        @opts,
        {
            type        => $TYPE_SPEC,
            destination => {
                type     => SCALAR | HASHREF,
                optional => 1,
            },
        }
    );

    if ($opts{destination}) {
        # If they passed this parameter, it is treated as either a volume
        # object, a UUID or a path-string.
        my $key = $opts{destination};
        my $uuid;
        if ((blessed $key) && $key->isa('NACL::C::Volume')) {
            # For a Volume object, get the UUID from the state. We can't be
            # sure this object is the same referent as one of our internal
            # ones, so we can't use $self->{obj2uuid_map}:
            $uuid = $key->get_one_state_attribute('uuid');
        } elsif ($key =~ /^[-a-f0-9]+$/i) {
            # It's a straight UUID
            $uuid = $key;
        } else {
            # If it doesn't match the regexp above, it's a path. Get the UUID
            # from our internal map, throwing an exception if the path isn't
            # found:
            if (! ($uuid = $self->{path_map}->{$key})) {
                $Log->exit() if $may_exit;
                NATE::Exceptions::Argument->throw(
                    qq(Volume path "$key" not in the set)
                );
            }
        }

        # If we don't have the key in our table, throw an exception:
        if (! ($value = $self->{mirrors}->{$uuid})) {
            $Log->exit() if $may_exit;
            NATE::Exceptions::Argument->throw(
                qq(Volume UUID $uuid not in the set)
            );
        }
    } elsif ($opts{type}) {
        # Like destinations(), the type parameter has already been validated.
        $value = [];
        # Use the internal list for type $opts{type} to build the
        # new list of relationship objects.
        for my $endpt (@{$self->{destinations}->{uc $opts{type}}}) {
            my $uuid = $self->{obj2uuid_map}->{$endpt};
            push(@{$value}, $self->{mirrors}->{$uuid});
        }
    } else {
        # If no parameters were given, return a copy of the internal hash of
        # all mirrors. Only instead of being keyed by UUID, the returned hash
        # should be keyed by destination-path.
        $value = {};
        for my $path (keys %{$self->{path_map}}) {
            $value->{$path} = $self->{mirrors}->{$self->{path_map}->{$path}};
        }
    }

    $Log->exit() if $may_exit;
    return $value;
}

=head1 METHODS

In addition to the accessors, the following methods are provided by this class.
Some methods may only be called as class methods, these will be noted in their
individual sections.

=head2 new

    # Minimal construction option:
    $set = NACL::MTask::SnapmirrorSet->new(
        source => $source_volume
    );

    # Create from existing mirror relationships:
    $set = NACL::MTask::SnapmirrorSet->new(
        source  => $source_volume,
        mirrors => \@existing_mirrors
    );

    # Create from existing destination volumes-- creates the needed
    # snapmirror relationships:
    $set = NACL::MTask::SnapmirrorSet->new(
        source       => $source_volume,
        destinations => {
            LS  => \@volumes_for_ls,
            DP  => \@volumes_for_dp,
            XDP => \@volumes_for_xdp,
        }
    );

(May only be called as a class method.)

B<new> creates a new object of this class and returns a reference to it. The
only required argument is the C<source> parameter, as destinations can be
added dynamically in subsequent calls. Based on the parameters given, the
constructor may create some number of B<NACL::STask::Snapmirror> objects.

=over

=item Options

=over

=item C<< source =E<gt> $source_volume >>

(Required.) This is the volume object that represents the source end-point for
all the snapmirror relationships in the set. A set may have only one
source-point, and it cannot be changed after the set object is created. This is
the only strictly required parameter for B<new>, and must be either a
B<NACL::STask::Volume> object, a B<NACL::C::Volume> object, or a hash
reference. A B<NACL::C::Volume> object will be promoted to an STask object. If
it is a hash reference, the contents of the reference will be used to create an
object of the B<NACL::STask::Volume> class. If the creation of that object
fails, an exception will be thrown by B<new>. For detail on the keys that can
be in the hash reference, see L<NACL::C::Volume|lib-NACL-C-Volume-pm> and
L<NACL::STask::Volume|lib-NACL-STask-Volume-pm>.

=item C<< destinations =E<gt> $hash_reference >>

(Optional) This is the specification of destination volumes for the snapmirror
relationships that are to be maintained as part of this set object. The hash
reference specifies between one and three types of destinations: C<LS>, C<DP>
and C<XDP>. Any keys other than these three in the hash reference will trigger
an exception. Each key points to either an array reference or a hash reference:

If the value is a hash reference, the contents of the reference are used to
construct volumes to use as end-points. The exception is the key C<count>; if
this key is present in the hash reference it represents the number of volumes
to create. It is removed before creation is done. If C<count> is not present it
defaults to 1. For detail on the keys that can be in the hash reference, see
L<NACL::C::Volume|lib-NACL-C-Volume-pm> and
L<NACL::STask::Volume|lib-NACL-STask-Volume-pm>.

If the value is an array reference, the contents of the reference must be some
combination of volume objects (B<NACL::STask::Volume> or B<NACL::C::Volume>)
and hash references. Any hash references are taken as arguments for volume
creation, as above. Any B<NACL::C::Volume> objects are promoted to STask
objects before being stored.

For all destination volumes specified, a B<NACL::STask::Snapmirror> object will
be created, between the source volume and the individual destination volumes.
This will also create the snapmirror relationships themselves. The relationship
type will be determined by which hash-key this is a part of (C<LS>, C<DP> or
C<XDP>).

Any volume creation runs the risk of failure, and will result in an exception
being thrown. Likewise, creating the snapmirror objects can also cause an
exception to be thrown.

=item C<< mirrors =E<gt> $list_reference >>

(Optional) This is an alternative to specifying the destination volumes.
Instead, this specifies one or more existing B<NACL::STask::Snapmirror> or
B<NACL::C::Snapmirror> objects that should make up the set. Each is checked to
confirm that it uses the source volume as the source end-point, and each is
categorized by the type of snapmirror relationship it is (C<LS>, C<DP> or
C<XDP>). Any B<NACL::C::Snapmirror> objects are promoted to STask object before
being stored.

No new resources are created by this option, so the only source of exceptions
come from the calls to retrieve entity state for each of the mirror objects
passed in or from the attempt to attach an object to an existing volume.

=item C<< preserve =E<gt> 1|0 >>

(Optional, boolean.) If passed and true, this flag tells the newly-created
object to B<not> destroy any resources it creates. Normally, any objects
created by B<new> (both B<NACL::STask::Volume> and B<NACL::STask::Snapmirror>)
are cleaned-up and removed when the object is destroyed. Any volumes or
snapmirror relationships that already existed before being passed to B<new> are
not cleaned up. The default is C<0>, resources are automatically cleaned up
when the object is destroyed.

Note that if B<new> encounters an error I<during> the creation of resources,
it will destroy all resources it has created before throwing an exception.
This is not affected by this flag, as the expectation is that if B<new> fails
the filers should revert to the state they were in before B<new> began work.

=item C<< snapmirror_options =E<gt> $hash_reference >>

(Optional.) If passed, the hash reference given is interpreted as having
additional options to be used whenever the B<NACL::MTask::SnapmirrorSet>
object creates a B<NACL::STask::Snapmirror> object. The hash may have up to
four keys:

=over

=item C<LS>

=item C<DP>

=item C<XDP>

These three keys specify options that are only applied to the creation of
snapmirror objects of the given type. The value of each key is another
hash reference, with key/value pairs of options that are passed to any call
to B<NACL::STask::Snapmirror::create> for the given type.

=item C<all>

This key specifies a hash reference of key/value pairs that are passed to any
call to B<NACL::STask::Snapmirror::create> for I<any> type. These options are
passed first, before the type-specific options. You can specify general options
in this hash reference, then override them in the type-specific hash
reference. For example:

    snapmirror_options => {
        all => { schedule => 'default', tries => 8 },
        ls  => { schedule => 'ls_default' },
        ...
    }

This would pass the C<schedule> and C<tries> options to all calls, but the
calls to create a C<LS> snapmirror would get C<ls_default> as the value for
C<schedule>, rather than C<default>.

=back

Not all keys need to be specified, any that are not specified default to no
options. Any keys other than these four will be ignored.

=back

=item Exceptions

Any exceptions that can be thrown by the B<create> methods of
B<NACL::STask::Volume> or B<NACL::STask::Snapmirror> can appear here, as those
methods are called as part of the set-up of this constructor. Additionally, if
any of the input data to this method is invalid, a
B<NATE::Exceptions::Argument> exception will be thrown.

=back

When this method creates snapmirror relationships, they are I<not> initialized.
They must be explicitly initialized. See L</initialize>, below.

=cut

###############################################################################
#
#   Sub Name:       new
#
#   Description:    Class constructor.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $class    in      scalar    Class to bless into
#                   @opts     in      list      Additional options, converted
#                                                 to %opts by validate()
#
#   Globals:        %SM_TYPES
#                   $DESTINATIONS_SPEC
#                   $MIRRORS_SPEC
#
#   Returns:        Success:    new object of this class
#                   Failure:    throws exception
#
###############################################################################
sub new {
    $Log->enter() if $may_enter;
    my $class = shift;
    # This must me a static method invocation:
    $class->_verify_invocation(style => 'static_only');
    my %opts = validate(
        @_,
        {
            source => {
                type      => HASHREF | OBJECT,
                optional  => 0, # This is the only required parameter
                callbacks => {
                    'required type' => sub {
                        my $ref = shift;
                        # We've already established it's a hashref or a blessed
                        # ref, we just want to know (if it is blessed) if it is
                        # a Volume:
                        return (blessed $ref) ?
                            $ref->isa('NACL::C::Volume') : 1;

                    },
                },
            },
            destinations => $DESTINATIONS_SPEC,
            mirrors => $MIRRORS_SPEC,
            preserve => {
                # This is treated as a boolean, but I want the default to be
                # 0 rather than undef. Just in case I'd trigger any "use of
                # undefined value" warnings.
                type    => BOOLEAN,
                default => 0,
            },
            snapmirror_options => {
                type     => HASHREF,
                optional => 1,
            },
        }
    );

    # This is the structure of the SnapmirrorSet object. It is being
    # initialized here so that the hashref can be blessed and private
    # methods called on it to fill in these fields.
    my $self = bless {
        # The source attribute, a NACL::STask::Volume object
        source => undef,
        # Not sure if we'll need this, but cache it just in case:
        source_uuid => undef,
        # I'm dubious about caching these next two, as a change to the volume
        # name of the source would invalidate them without the object knowing
        # about it. But I don't want repeated calls to generate them, either.
        source_path => undef,
        source_path_sn => undef,
        # Table of the destination volumes, keyed by the snapmirror types
        destinations => {
            LS  => [],
            DP  => [],
            XDP => [],
        },
        # Default options to the creation of any snapmirror objects
        snapmirror_options => {
            LS  => {},
            DP  => {},
            XDP => {},
            all => {},
        },
        # Table of the NACL::STask::Snapmirror objects, keyed by dest UUID
        mirrors => {},
        # Mapping of destination paths to UUIDs, so that users can request a
        # snapmirror relationship by path instead of by object.
        path_map => {},
        # Map the actual NACL::STask::Volume objects we have to their UUIDs.
        # This saves us having to make repeated calls to $vol->state(), and
        # even if the volume name changes the UUID doesn't so it's safe to
        # cache. Uses the stringified referents as keys.
        obj2uuid_map => {},
        # Map UUIDs to objects, for when we need the reverse. Reversing the
        # above map won't work, because the keys are stringified referents,
        # not the referents themselves.
        uuid2obj_map => {},
        # Any volumes we create ourselves are also noted here, so they can be
        # destroyed when the object is destroyed.
        created_volumes => [],
        # Any snapmirror relationships we create are also noted, like volumes,
        # in case we need to remove them during clean-up or in case of failure
        # of the constructor.
        created_snapmirrors => [],
        # When this flag is true, the basic object has been through the
        # initialize() method. The flag serves to tell when newly-added
        # volumes should be immediately initialized or not.
        initialized => 0,
        # If this flag is true, then DESTROY should not clean up the contents
        # of created_snapmirrors and created_volumes. This does not apply if
        # we fail here in new(), however. In that case, we always clean up.
        preserve => $opts{preserve},

        # The remaining object properties are those that are more internal to
        # the object, those that are not dependent or directly affected by the
        # user input. These include properties such as the results and jobs
        # tables that execute()/_execute() manipulate, etc.

        # This table holds the results from the most-recent call to execute().
        _results => {},
        # This table holds any NACL::C::Job objects that resulted from the
        # most-recent call to execute().
        _jobs => {},
        # This indexes those snapmirror objects that have already been
        # initialized.
        _initialized => {},
    }, $class;

    # If $opts{snapmirror_options} is given, copy it. Don't use it directly,
    # as I don't trust people to not (accidentally) change the contents after
    # passing it here:
    if ($opts{snapmirror_options}) {
        for my $key (keys %{$opts{snapmirror_options}}) {
            # I'm not going to bother complaining about unknown keys. Just
            # ignore them in this case...
            if ($SM_TYPES{$key} || ($key eq 'all')) {
                %{$self->{snapmirror_options}->{$key}} =
                    %{$opts{snapmirror_options}->{$key}};
            }
        }
    }

    my $source_volume;
    if (blessed $opts{source}) {
        # If it's blessed we validated it in the validate() call, earlier.
        # Here we just check to see if it is already a STask, and if not we
        # promote the component to STask.
        $source_volume = $opts{source}->isa('NACL::STask::Volume') ?
            $opts{source} : $opts{source}->cast_component_to_task();
    } else {
        # A plain hash-reference. If you pass garbage here, you deserve what
        # you get back.
        $source_volume = $self->_create_volumes(
            %{$opts{source}},
            # Keep this one out of dest-specific tables and store it on the
            # 'source' internal key:
            source_volume => 1
        );
    }
    # Now that we have $source_volume, use it to update the internal "source"
    # attribute, which in turn derives and caches the volpath in both SN and
    # pre-SN formats:
    $self->_update_source(volume => $source_volume);

    # Process whichever of mirrors or destinations that were passed (assuming
    # either was passed). We now have to catch any exceptions that are thrown
    # by called methods, so that we can clean up after ourselves.
    if ($opts{destinations} || $opts{mirrors}) {
        try {
            my %add_opts;
            for (qw(destinations mirrors)) {
                if ($opts{$_}) {
                    $add_opts{$_} = $opts{$_};
                }
            }
            $self->add_to_set(%add_opts);
        } catch NATE::BaseException with {
            my $exception = shift;

            # Clean up any volumes we've created ourselves:
            $self->_cleanup_volumes_and_snapmirrors();

            $exception->throw;
        };
    }

    $Log->exit() if $may_exit;
    return $self;
}

=head2 create

    # Create a new object, just like calling new(), above:
    $set = NACL::MTask::SnapmirrorSet->create(
        source       => $source_volume,
        destinations => {
            LS  => \@volumes_for_ls,
            DP  => \@volumes_for_dp,
            XDP => \@volumes_for_xdp,
        }
    );

    # Augment an existing object with 3 new DP mirrors:
    $set->create(
        dp => {
            count             => 3,
            command_interface => $filer,
            aggregate         => $dp_dest_aggr,
            vserver           => 'vs0',
        }
    );

    # Or, add two new LS mirrors from different filers:
    $set->create(
        ls => [
            {
                command_interface => $filer_one,
                aggregate         => $ls_dest_aggr_one,
                vserver           => 'vs0',
            },
            {
                command_interface => $filer_two,
                aggregate         => $ls_dest_aggr_two,
                vserver           => 'vs0',
            },
        ]
    );

(May be called as a class method or an instance method.)

When called as a class method, the B<create> method is a synonym for B<new>,
above. It merely calls B<new> with the arguments list mostly unchanged. This is
provided for ease-of-use for those who are more accustomed to using C<create>
to instantiate objects rather than the more-traditional C<new>. See above for
the arguments and their disposition. The difference is that B<create> supports
some additional arguments that B<new> does not. When called as a constructor,
B<create> will massage the extra arguments into a format that B<new>
understands. The return value in this case is a new object reference, unless an
exception is thrown.

When called as an instance method, B<create> only accepts the options listed
below. The return value in this case is the object itself, unless an exception
is thrown. When called as an instance method, B<create> augments the existing
B<NACL::MTask::SnapmirrorSet> instance by adding new elements to the set, based
on the arguments passed. If no arguments are passed, or invalid arguments are
passed, a B<NATE::Exceptions::Argument> exception is thrown.

=over

=item Options

The following options are valid whether B<create> is called as a static or
instance method. Additionally, if called as a static method the options valid
for L</new> (see above) are also valid.

=over

=item C<< LS =E<gt> $listref | $hashref >>

=item C<< DP =E<gt> $listref | $hashref >>

=item C<< XDP =E<gt> $listref | $hashref >>

Each of these take the same style of arguments: a specification of volumes to
create and then use in creating snapmirror relationships. The specification is
the same as for the C<destinations> option in B<new>:

If the value is a hash reference, the contents of the reference are used to
construct volumes to use as end-points. The exception is the key C<count>; if
this key is present in the hash reference it represents the number of volumes
to create. It is removed before creation is done. If C<count> is not present it
defaults to 1. For detail on the keys that can be in the hash reference, see
L<NACL::C::Volume|lib-NACL-C-Volume-pm> and
L<NACL::STask::Volume|lib-NACL-STask-Volume-pm>.

If the value is an array reference, the contents of the reference must be some
combination of volume objects (B<NACL::STask::Volume> or B<NACL::C::Volume>)
and hash references. Any hash references are taken as arguments for volume
creation, as above. Any component objects are promoted to STask objects.

For all destination volumes specified, a B<NACL::STask::Snapmirror> object will
be created, between the source volume and the individual destination volumes.
This will also create the snapmirror relationships themselves. The relationship
type will be determined by which argument this is a part of (C<LS>, C<DP> or
C<XDP>). To maintain syntactic compatibility with B<GNOTE::GX::SnapMirrorSet>,
the options may be passed as lower-case (C<ls>, C<dp> or C<xdp>). If both C<LS>
C<ls> are present, the upper-case option takes precendence.

Any volume creation runs the risk of failure, and will result in an exception
being thrown. Likewise, creating the snapmirror objects can also cause an
exception to be thrown.

=back

=item Exceptions

Any exceptions that can be thrown by the B<create> methods of
B<NACL::STask::Volume> or B<NACL::STask::Snapmirror> can appear here, as those
methods are called as part of the actions of this method. Additionally, if
any of the input data to this method is invalid, a
B<NATE::Exceptions::Argument> exception will be thrown.

=back

When this method creates snapmirror relationships, they are I<not> initialized.
They must be explicitly initialized. See L</initialize>, below.

=cut

###############################################################################
#
#   Sub Name:       create
#
#   Description:    A slightly-different syntax for creation, meant to be
#                   compatible with the syntax of GNOTE::GX::SnapMirrorSet.
#                   Mostly massages arguments and passes off to new, unless
#                   called on an existing object in which case it augments
#                   the contents of an existing set.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#                   %opts     in      hash      Either options to new(), or
#                                                 sets of new volumes to add
#                                                 to the object
#
#   Globals:        %SM_TYPES
#                   $SM_TYPES
#
#   Returns:        Success:    $self
#                   Failure:    throws exception
#
###############################################################################
sub create {
    $Log->enter() if $may_enter;
    my ($self, %opts) = @_;

    # If this is a static call, then massage any extra arguments into the
    # format new() takes and pass off to it.
    if (! ref($self)) {
        # Just in case they already specified destinations, don't overwrite it
        $opts{destinations} ||= {};
        # These will auto-vivify if not already present on $opts{destinations}
        for my $type (keys %SM_TYPES) {
            my $spec = $opts{$type} || $opts{lc $type};
            next unless $spec;

            # Remove the options from %opts so new() doesn't choke on them
            delete $opts{$type}; delete $opts{lc $type};
            # If $spec is not a listref, make it one
            if (ref($spec) ne 'ARRAY') {
                $spec = [ $spec ];
            }
            # Push onto {destinations}. Will create the key if it doesn't
            # already exist. But if it does exist and isn't already a listref,
            # make it into one.
            if ($opts{destinations}->{$type} &&
                (ref($opts{destinations}->{$type}) ne 'ARRAY')) {
                $opts{destinations}->{$type} = [ $opts{destinations}->{$type} ];
            }
            push(@{$opts{destinations}->{$type}}, @{$spec});
        }

        # Note that we've already determined $self is the package, not a ref
        return $self->new(%opts);
    }

    # At this point we've established that this is an instance, not a fresh
    # creation. Thus, the only options allowed are the mirror types (upper
    # and lower case). Writing that in Params::Validate terms would be a mess
    # so just do it in a simpler way.
    for (keys %opts) {
        if (! $SM_TYPES{$_} && ! $SM_TYPES{uc $_}) {
            $Log->exit() if $may_exit;
            NATE::Exceptions::Argument->throw(
                qq(Unknown option "$_" to create, must be one of $SM_TYPES)
            );
        }
    }
    # Normalize the contents before we call add_to_set(). Any lower-case has
    # to become upper-case, unless the upper-case already exists in which case
    # it is just quietly deleted.
    for my $type (keys %SM_TYPES) {
        if ($opts{lc $type} && ! $opts{$type}) {
            $opts{$type} = $opts{lc $type};
        }
        delete $opts{lc $type};
    }

    # This handles all the add-a-volume logic:
    $self->add_to_set(%opts);

    $Log->exit() if $may_exit;
    return $self;
}

=head2 initialize

    # Create a new object, just like calling new() or create():
    $set = NACL::MTask::SnapmirrorSet->initialize(
        source       => $source_volume,
        destinations => {
            LS  => \@volumes_for_ls,
            DP  => \@volumes_for_dp,
            XDP => \@volumes_for_xdp,
        }
    );

    # Just initialize the elements on an object previously created
    # with create() or new():
    $set->initialize;

    # Augment an existing object with 3 new DP mirrors and then
    # initialize them:
    $set->initialize(
        dp => {
            count             => 3,
            command_interface => $filer,
            aggregate         => $dp_dest_aggr,
            vserver           => 'vs0',
        }
    );

(May be called as a class method or an instance method.)

When called as a class method, the B<initialize> method doubles as both a
constructor and an initializer. It accepts the same arguments that B<create>
does (which includes those arguments of B<new>) as well as the arguments
listed below. It passes the relevant arguments to B<create>, and if the
creation is successful performs initialization on the resulting set object.
The return value is the new object, unless an exception is thrown.

When called as an instance method, B<initialize> only accepts the options
specific to B<create> and the options listed below. The return value in this
case is the object itself, unless an exception is thrown. When called as an
instance method, B<initialize> will call B<create> if necessary (based on the
given parameters) to add new elements. It then initializes the existing set of
snapmirror relationships in the object.

=over

=item Options

The B<initialize> method accepts options for B<create>/B<new> when called as a
static method. In addition, it accepts the following options. When called as an
instance method, only the following options and options for B<create> are
accepted (options for B<new> are not accepted).

=over

=item C<< parallel =E<gt> 0|1 >>

Perform multiple C<initialize> operations in parallel, rather than
sequentially. Defaults to 0 (sequential).

=item C<< background =E<gt> 0|1 >>

Do everything in the background (that can be done-- not all commands support
backgrounding), and do not wait for jobs to complete before returning from
the call.

=item C<< destinations =E<gt> $array_ref >>

If passed, the value must be an array-reference containing one or more
end-points to initialize. The end-points may be volume or snapmirror objects.
If they are not the same objects that are stored in the set, they will be
used to look up the corresponding internal object, and that object will be
the one actually used to call the underlying B<initialize> method. Only one
of B<destinations> or C<type> may be specified in a single call.

B<IMPORTANT>: Note that this is a slightly different syntax than when the
B<destinations> argument is used with B<new> or B<create> (as a static
method). If you call B<initialize> as a static method you must use that
syntax. Only use this syntax when calling B<initialize> on an existing
B<NACL::MTask::SnapmirrorSet> object.

=item C<< type =E<gt> $type | $array_ref >>

If passed, the value must be one of C<LS>, C<DP>, C<XDP>, or an array-reference
of some combination of these. This tells B<initialize> to only operate on the
destinations of that type (or types). Only one of C<destinations> or C<type>
may be specified in a single call.

=back

=item Exceptions

In addition to any exceptions that can filter through from B<create> and B<new>,
this method may throw a B<NATE::Exceptions::Argument> exception if a bad
parameter is passed. It may also pass through any exceptions thrown by the
B<initialize> or B<initialize_ls_set> methods of B<NACL::STask::Snapmirror>.

=back

=cut

###############################################################################
#
#   Sub Name:       initialize
#
#   Description:    Initialize the snapmirror relationships contained within
#                   this set. May also add new elements to the set prior to
#                   initialization, per the semantics of create() or new().
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#                   %opts     in      hash      Additional options
#
#   Globals:        %SM_TYPES
#
#   Returns:        Success:    $self
#                   Failure:    throws exception
#
###############################################################################
sub initialize {
    $Log->enter() if $may_enter;
    my ($self, %opts) = @_;

    # If this is called as a static method, it has to conform to create()
    # argument syntax. Just pass everything up to create() and see what it
    # gives us back:
    if (! ref($self)) {
        # OK, not *everything*. Only pass those things that create() is going
        # to understand, and remove them from %opts. Then, what is left can be
        # passed to execute() later.
        my %args;
        # new() arguments:
        for my $arg (qw(source destinations mirrors preserve
                        snapmirror_options)) {
            next if (! $opts{$arg});
            $args{$arg} = delete $opts{$arg};
        }
        # create() arguments:
        for my $arg (keys %SM_TYPES) {
            if ($opts{$arg}) {
                $args{$arg} = delete $opts{$arg};
            }
            if ($opts{lc $arg}) {
                $args{lc $arg} = delete $opts{lc $arg};
            }
        }

        $self = $self->create(%args);
    } else {
        # Check to see if we have create() args. This could be a non-static
        # call that is still using create() args. This madness brought to you
        # by the requirement for GNOTE::GX::SnapMirrorSet compatibility...
        my %args;
        # create() arguments:
        for my $arg (keys %SM_TYPES) {
            if ($opts{$arg}) {
                $args{$arg} = delete $opts{$arg};
            }
            if ($opts{lc $arg}) {
                $args{lc $arg} = delete $opts{lc $arg};
            }
        }

        # Only call create() if we actually have some args for it:
        if (%args) {
            $self->create(%args);
        }
    }

    # OK, now what do we do? We have a special action for LS sets, but other
    # than that execute() can handle this for us. So call it with a pointer
    # to the special hook for LS.
    $self->execute(
        %opts,
        action  => 'initialize',
        LS_hook => '_initialize_ls'
    );
    # Set this flag. This tells future calls to some of the methods whether
    # they should initialize or not.
    $self->{initialized} = 1;

    $Log->exit() if $may_exit;
    return $self;
}

###############################################################################
#
#   Sub Name:       _initialize_ls
#
#   Description:    Perform initialize on the LS mirrors in a set, via the
#                   "initialize_ls_set" method of NACL::C::Snapmirror.
#                   This should only ever be called from _execute(), and makes
#                   assumptions about the setup/structure of private fields on
#                   $self based on being called from there.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#                   $targets  in      listref   Array-reference of the LS
#                                                 end-points being targeted
#                   %opts     in      hash      Additional options to the
#                                                 initialize_ls_set method
#
#   Returns:        Success:    Number of elements in $targets
#                   Failure:    Same, but sets $self->{_results} as well
#
###############################################################################
sub _initialize_ls {
    $Log->enter() if $may_enter;
    my ($self, $targets, %opts) = @_;

    # I don't know if there is special behavior for trying to initialize a
    # single LS snapmirror (or a subset of the whole LS slate), such as in
    # cases where additional LS mirrors are added. For now, just issue the
    # initialize_ls_set() command to the first snapmirror object mapped from
    # $targets.
    my %path_from_uuid = reverse %{$self->{path_map}};
    my $first = $targets->[0];
    my $uuid = $self->{obj2uuid_map}->{$first};
    my $mirror = $self->{mirrors}->{$uuid};
    if (! $mirror) {
        my $main_path = $path_from_uuid{$uuid};
        for my $target (@{$targets}) {
            my $path = $path_from_uuid{$self->{obj2uuid_map}->{$target}};
            $self->{_results}->{$path} = NATE::BaseException->new(
                "Panic! No mirror found for LS dest obj $main_path"
            );
        }
        return scalar(@{$targets});
    }

    try {
        my $job_component;
        $mirror->initialize_ls_set(
            %opts,
            job_component => \$job_component
        );
        # Flag all the targets as initialized
        for my $target (@{$targets}) {
            $self->{_initialized}->{$target} = 1;
             my $path = $path_from_uuid{$self->{obj2uuid_map}->{$target}};
             $self->{_results}->{$path} = 1;
        }
        # If there is a job_component, save it. But only under the one dest
        # key. Don't make N copies of it.
        if ($job_component) {
            my $path = $path_from_uuid{$uuid};
            $self->{_jobs}->{$path} = $job_component;
        }
    } catch NATE::BaseException with {
        my $exception = shift;

        # For now, just save this exception on all the target's keys in the
        # _results table:
        for my $target (@{$targets}) {
            my $path = $path_from_uuid{$self->{obj2uuid_map}->{$target}};
            $self->{_results}->{$path} = $exception;
        }
    };

    $Log->exit() if $may_exit;
    return scalar(@{$targets});
}

=head2 update

    # Update the entire set:
    $set->update;

    # Update only the LS mirrors:
    $set->update(type => 'LS');

    # Update the DP and XDP, but not the LS mirrors, using a specific
    # snapshot:
    $set->update(
        type              => [ 'DP', 'XDP' ],
        'source-snapshot' => $snapshot
    );

    # Update only some specific mirrors, as specified by their
    # end-points:
    $set->update(destinations => [ $dp_vol_1, $dp_vol_2 ]);

    # Update the entire set in parallel and return without waiting for
    # the tasks to complete:
    $set->update(parallel => 1, background => 1);

(May be called as an instance method only.)

Performs a C<snapmirror update> over all elements in the snapmirror set. If the
set includes any LS mirrors, those are updated with the C<update-ls-set>
command, instead.

Note that updates for C<XDP> relationships function slightly differently from
C<DP> or C<LS> relationships. An C<XDP> relationship will only update from a
specific snapshot, or from the most-recent snapshot that has been given a
label that matches one of the labels in the relationship's policy. See
L<NACL::C::SnapmirrorPolicy|lib-NACL-C-SnapmirrorPolicy-pm> for more on
snapmirror policies.

=over

=item Options

The B<update> method accepts the following options:

=over

=item C<< parallel =E<gt> 0|1 >>

Perform multiple C<update> operations in parallel, rather than
sequentially. Defaults to 0 (sequential).

=item C<< background =E<gt> 0|1 >>

Do everything in the background, and do not wait for jobs to complete before
returning from the call. Defaults to 0 (wait for jobs to complete).

=item C<< destinations =E<gt> $array_ref >>

If passed, the value must be an array-reference containing one or more
end-points to update. The end-points may be volume or snapmirror objects. If
they are not the same objects that are stored in the set, they will be used to
look up the corresponding internal object, and that object will be the one
actually used to call the underlying B<update> method. Only one of
B<destinations> or C<type> may be specified in a single call.

=item C<< type =E<gt> $type|$array_ref >>

If passed, the value must be one of C<LS>, C<DP>, C<XDP>, or an array-reference
of some combination of these. This tells B<update> to only operate on the
destinations of that type (or types). Only one of C<destinations> or C<type>
may be specified in a single call.

=back

In addition to these options, any additional options given are passed through
to B<update> and/or B<update_ls_set> methods in the B<NACL::STask::Snapmirror>
and B<NACL::C::Snapmirror> classes (see
L<NACL::STask::Snapmirror|lib-NACL-STask-Snapmirror-pm> and
L<NACL::C::Snapmirror|lib-NACL-C-Snapmirror-pm>). Only options that you would
pass to an instance call, not a static call, should be used here. Note,
however, that the options C<nacltask_wait> and C<foreground> are manipulated by
this B<update> method, so specifying them yourself will have no effect. Use
C<parallel> and C<background>, instead.

=item Exceptions

This method may throw a B<NATE::Exceptions::Argument> exception if a bad
parameter is passed. It may also pass through any exceptions thrown by the
B<update> or B<update_ls_set> methods of B<NACL::STask::Snapmirror> or
B<NACL::C::Snapmirror>.

=back

=cut

###############################################################################
#
#   Sub Name:       update
#
#   Description:    Front-end to execute() that adds a hook for LS sets. We
#                   have a special command for updating LS sets. Other than
#                   that, execute() can handled everything for us.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#                   %opts     in      hash      Options to be passed along to
#                                                 execute() (unchanged)
#
#   Returns:        Success:    $self
#                   Failure:    throws exception
#
###############################################################################
sub update {
    $Log->enter() if $may_enter;
    my ($self, %opts) = @_;

    # We have a special action for LS sets, but other than that execute() can
    # handle this for us. So call it with a pointer to the special hook for LS.
    $self->execute(
        %opts,
        action  => 'update',
        LS_hook => '_update_ls'
    );

    $Log->exit() if $may_exit;
    return $self;
}

###############################################################################
#
#   Sub Name:       _update_ls
#
#   Description:    Perform update on the LS elements of the snapmirror set by
#                   calling update_ls_set on the first element of $targets.
#                   This should only ever be called from _execute(), and makes
#                   assumptions about the setup/structure of private fields on
#                   $self based on being called from there.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#                   $targets  in      listref   Array-reference of the LS
#                                                 end-points being targeted
#                   %opts     in      hash      Additional options to the
#                                                 update_ls_set method
#
#   Returns:        Success:    Number of elements in $targets
#                   Failure:    Same, but sets $self->{_results} as well
#
###############################################################################
sub _update_ls {
    $Log->enter() if $may_enter;
    my ($self, $targets, %opts) = @_;

    # Issue the update_ls_set() command to the first snapmirror object mapped
    # from $targets.
    my %path_from_uuid = reverse %{$self->{path_map}};
    my $first = $targets->[0];
    my $uuid = $self->{obj2uuid_map}->{$first};
    my $mirror = $self->{mirrors}->{$uuid};
    if (! $mirror) {
        my $main_path = $path_from_uuid{$uuid};
        for my $target (@{$targets}) {
            my $path = $path_from_uuid{$self->{obj2uuid_map}->{$target}};
            $self->{_results}->{$path} = NATE::BaseException->new(
                "Panic! No mirror found for LS dest volume $main_path"
            );
        }
        $Log->exit() if $may_exit;
        return scalar(@{$targets});
    }

    try {
        my $job_component;
        $mirror->update_ls_set(
            %opts,
            job_component => \$job_component
        );
        # Flag all the targets as run
        for my $target (@{$targets}) {
            my $path = $path_from_uuid{$self->{obj2uuid_map}->{$target}};
            $self->{_results}->{$path} = 1;
        }
        # If there is a job_component, save it. But only under the one dest
        # key. Don't make N copies of it.
        if ($job_component) {
            my $path = $path_from_uuid{$uuid};
            $self->{_jobs}->{$path} = $job_component;
        }
    } catch NATE::BaseException with {
        my $exception = shift;

        # For now, just save this exception on all the target's keys in the
        # _results table:
        for my $target (@{$targets}) {
            my $path = $path_from_uuid{$self->{obj2uuid_map}->{$target}};
            $self->{_results}->{$path} = $exception;
        }
    };

    $Log->exit() if $may_exit;
    return scalar(@{$targets});
}

=head2 add_to_set

    # Add some volumes to an existing set:
    $set->add_to_set(
        destinations => {
            LS => {
                command_interface => $filer,
                aggregate         => $ls_dest_aggr,
                vserver           => $vserver,
                count             => 2,
            },
        }
    );

    # Or just:
    $set->add_to_set(
        LS => {
            command_interface => $filer,
            aggregate         => $ls_dest_aggr,
            vserver           => $vserver,
            count             => 2,
        },
    );

    # Use existing objects that are NACL::C::Volume or
    # NACL::STask::Volume, but do not initialize the new
    # snapmirror relationships:
    $set->add_to_set(
        LS            => \@volumes,
        do_initialize => 0
    );

    # Add some existing snapmirrors (NACL::C::Snapmirror or
    # NACL::STask::Snapmirror) to the object:
    $set->add_to_set(mirrors => \@snapmirrors);

This method lets you add additional relationships to an existing
B<NACL::MTask::SnapmirrorSet> object. Relationships may be added as either
volumes or snapmirrors. Whichever is passed, the corresponding other entity is
created and added as well (in the case of an existing snapmirror, a
B<NACL::STask::Volume> object is attached to existing destination volume, no
new volume is created).

With this method, you may add volumes (which trigger the creation of snapmirror
relationships and corresponding B<NACL::STask::Snapmirror> objects) or
snapmirrors (which trigger the creation of B<NACL::STask::Volume> objects to
represent the existing destination volumes), or both. The B<add_to_set> method
will reason out what is needed from the disposition of the options.

=over

=item Options

This method takes specification arguments in a format and fashion identical to
both the L</new> and L</create> methods. That is, the following parameters are
recognized by B<add_to_set> and have the exact same syntax:

=over

=item C<destinations>

A hash-reference specifying volumes. See L</new>.

=item C<mirrors>

An array-reference specifying existing snapmirrors. See L</new>.

=item C<LS>

=item C<DP>

=item C<XDP>

Specifications for volumes for the specified types. See L</create>.

=back

None of the options overrule any others. If all the above options are
provided, all will be processed.

In addition, the following boolean option is recognized:

=over

=item C<do_initialize>

Boolean, defaults to 1 (true). If true (the default), then newly-created
snapmirror relationships are initialized I<if> (and only if) this object
has already called the B<initialize> method at least once. If false, then
the user is responsible for initialization of the relationship at a later
point.

=back

=item Exceptions

Any exceptions that may occur from the creation of B<NACL::STask::Snapmirror>
or B<NACL::STask::Volume> objects may be thrown here, as well. In addition, any
attempt to add a volume or snapmirror relationship that is already a part of
this set will result in an exception of the class B<NATE::Exceptions::Argument>
to be thrown.

=back

When this method creates snapmirror relationships, they are not initialized
unless this object has already called B<initialize>. If the rest of the set has
been initialized already, then the new elements will be initialized as well
(unless the user also specifies the C<do_initialize> option with a false
value).

=cut

###############################################################################
#
#   Sub Name:       add_to_set
#
#   Description:    Add one or more existing snapmirror objects/relationships
#                   to the set. Inputs may be a combination of existing
#                   snapmirror objects or volume objects. The structure of
#                   the options is the same as for new(), above.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#                   @opts     in      list      Options to this, validated by
#                                                 a call to validate()
#
#   Globals:        %SM_TYPES
#                   $SM_TYPES
#                   $DESTINATIONS_SPEC
#                   $MIRRORS_SPEC
#                   $VOLUME_TYPE_SPEC
#
#   Returns:        Success:    $self
#                   Failure:    throws exception
#
###############################################################################
sub add_to_set {
    $Log->enter() if $may_enter;
    my ($self, @opts) = @_;

    my %opts = validate(
        @opts,
        {
            destinations  => $DESTINATIONS_SPEC,
            mirrors       => $MIRRORS_SPEC,
            LS            => $VOLUME_TYPE_SPEC,
            DP            => $VOLUME_TYPE_SPEC,
            XDP           => $VOLUME_TYPE_SPEC,
            do_initialize => {
                type    => BOOLEAN,
                default => 1,
            },
        }
    );

    # Check that we aren't trying to add an element that we already have in
    # the set, this should be done completely before any adds are done:
    if ($opts{mirrors}) {
        for my $mirror (@{$opts{mirrors}}) {
            if ($self->_mirror_in_set($mirror)) {
                my $path = $mirror->destination_path;
                $Log->exit() if $may_exit;
                NATE::Exceptions::Argument->throw(
                    'Attempt to add a snapmirror that already exists in this ' .
                    "SnapmirrorSet (destination-path $path)"
                );
            }
        }
    }
    for my $type (keys %SM_TYPES) {
        my @vols;

        if ($opts{destinations} && $opts{destinations}->{$type} &&
            (ref($opts{destinations}->{$type}) eq 'ARRAY')) {
            push(@vols, @{$opts{destinations}->{$type}});
        }
        if ($opts{$type} && (ref($opts{$type}) eq 'ARRAY')) {
            push(@vols, @{$opts{$type}});
        }

        for my $volume (@vols) {
            # Skip any hashrefs that are volume-creation specs
            next if (ref($volume) eq 'HASH');

            if ($self->_volume_in_set($volume)) {
                my %path_from_uuid = reverse %{$self->{path_map}};
                my $path =
                    $path_from_uuid{$volume->get_one_state_attribute('uuid')};
                $Log->exit() if $may_exit;
                NATE::Exceptions::Argument->throw(
                    'Attempt to add a volume that already exists in this ' .
                    "SnapmirrorSet ($path)"
                );
            }
        }
    }

    # Having confirmed that we have no duplicate elements that are slated for
    # add, do the actual adding.

    # Add mirrors first, if any. For now, I am assuming that if this object is
    # already initialized, then these new mirrors have been, too. I may have
    # to come back and update this...
    if ($opts{mirrors}) {
        for my $mirror (@{$opts{mirrors}}) {
            # The call to validate() already ensured that all of these objects
            # are at least NACL::C::Snapmirror. The _process_mirror method will
            # re-cast any NACL::C::Snapmirror to NACL::STask::Snapmirror.
            $self->_process_mirror(mirror => $mirror);
        }
    }
    # Add anything from the "destinations" option. These will create new
    # snapmirrors, so those will be initialized if $self has already been
    # through a general initialize.
    if ($opts{destinations}) {
        for my $type (qw(LS DP XDP)) {
            # Only bother with this if $opts{destinations}->{$type} exists:
            if (my $spec = $opts{destinations}->{$type}) {
                if (ref($spec) eq 'HASH') {
                    # Instead of treating a hashref and listref differently,
                    # just convert a hashref to a one-element listref...
                    $spec = [ $spec ];
                }

                # This takes the volume spec and populates the internal
                # {destinations}->{$type} listref with volumes, either
                # pre-existing or freshly-created.
                my $vols_added = $self->_process_volume_spec(
                    type => $type,
                    spec => $spec
                );
                # Create snapmirrors for the volumes we just added:
                my $created_snapmirrors = $self->_create_snapmirrors(
                    type    => $type,
                    volumes => $vols_added
                );
                # If this object has already gone through a general
                # initialize(), then initialize these new snapmirrors
                # explicitly.
                if ($self->{initialized} && $opts{do_initialize}) {
                    $self->initialize(destinations => $created_snapmirrors);
                }
            }
        }
    }
    # Process the type-specific options. These will create new snapmirrors, so
    # those will be initialized if $self has already been through a general
    # initialize.
    for my $type (keys %SM_TYPES) {
        my $spec = $opts{$type};
        next unless $spec;

        # If $spec is not a listref, make it one
        if (ref($spec) ne 'ARRAY') {
            $spec = [ $spec ];
        }
        my $volumes = $self->_process_volume_spec(
            type => $type,
            spec => $spec
        );
        # Rather than just call this method like we did in new(), we call it
        # with the volumes option so that it only creates snapmirrors for the
        # newly-added volumes.
        my $created_snapmirrors = $self->_create_snapmirrors(
            type    => $type,
            volumes => $volumes
        );
        # If this object has already gone through a general initialize(), then
        # initialize these new snapmirrors explicitly.
        if ($self->{initialized} && $opts{do_initialize}) {
            $self->initialize(destinations => $created_snapmirrors);
        }
    }

    $Log->exit() if $may_exit;
    return $self;
}

=head2 promote

    # promote the specified destination volume in LS mirror
    $set->promote(destination => $ls_dest_vol_obj);

    # promote the destination volume in specified LS mirror
    $set->promote(destination => $ls_mirror_obj);

(Can be called as an instance method only. Supports CMode CLI now)

Performs a C<snapmirror promote> of specified destination volume in LS
snapmirror set. On successful promote operation, the SnapmirrorSet
object will be updated with the new source volume information and old
source volume information will be cleaned up. Also, the promote operation
will change the source volume for DP or XDP mirrors in the set as well.

=over

=item Options

The B<promote> method accepts the following options:

=over

=item C<< destination =E<gt> $object >>

(Required) The end-point object to promote. The value must be a volume or LS
mirror object. The object passed in must be an object known to the set, not a
different object that just points to the same volume or snapmirror.

=back

=item Exceptions

This method may throw a B<NATE::Exceptions::Argument> exception if a bad
parameter is passed. It may also throw the exceptions thrown by
the B<promote> method of B<NACL::C::Snapmirror>, as well.

=back

=cut

sub promote {
    $Log->enter() if $may_enter;
    my ($self, @opts) = @_;
    my %opts = validate_with(
        params      => \@opts,
        spec        => {
            destination => {
                type      => OBJECT,
                callbacks => {
                    'object part of set' => sub {
                        my $obj = shift;

                        if ($obj->isa('NACL::STask::Volume')) {
                            # If the input is a volume is should be in the
                            # obj2uuid_map hash on $self:
                            return $self->{obj2uuid_map}->{$obj} ? 1 : 0;
                        } elsif ($obj->isa('NACL::STask::Snapmirror')) {
                            # If the input is a snapmirror we can just look
                            # at the values of the mirrors map:
                            my $found = grep { $_ eq $obj }
                                (values %{$self->{mirrors}});
                            return $found ? 1 : 0;
                        } else {
                            # If it isn't either of those, it's a bad input:
                            return 0;
                        }
                    },
                },
            },
        },
        allow_extra => 1
    );

    # Delete this one, the rest of %opts will be passed unchanged to promote()
    my $destination = delete $opts{destination};
    my ($volume, $snapmirror, $is_internal);

    # Based on what $destination is (volume or snapmirror), we need to get the
    # other:
    if ($destination->isa('NACL::STask::Volume')) {
        # This is easy. The volume's UUID is already cached, and that UUID is
        # the key in the mirrors table.
        $volume = $destination;
        $snapmirror = $self->{mirrors}->{$self->{obj2uuid_map}->{$volume}};
    } else {
        # This is a little harder. We have the snapmirror, but we need to
        # know which volume goes with it. Do this by reversing the mirrors
        # map to get the volume UUID associated with this snapmirror, then
        # use uuid2obj_map to get the volume object for that UUID.
        my %tmp_map = reverse %{$self->{mirrors}};
        $snapmirror = $destination;
        $volume = $self->{uuid2obj_map}->{$tmp_map{$snapmirror}}
    }

    # Before we remove $volume and $snapmirror from the set, we'll need to
    # know if $volume is one of the volumes we created-- if so, it needs to
    # go back in the list of created volumes for clean-up.
    for my $vol (@{$self->{created_volumes}}) {
        if ($vol == $volume) {
            $is_internal = 1;
            last;
        }
    }

    # OK, do the promote. If this fails, just let the exception propagate back
    # to the called of this method:
    $snapmirror->promote(%opts);

    # Remove the snapmirror pair from the set. Ignore the returned values
    # since we already have $volume and $snapmirror.
    $self->remove_from_set(mirrors => [ $snapmirror ]);

    # Update the set with new source volume
    $self->_update_source(volume => $volume);

    # Update the source_path attribute on all the mirrors the set has
    for my $mirror (values %{$self->{mirrors}}) {
        $mirror->source_path($self->{source_path});
    }

    # If $volume is an internally-created volume, add it back to the
    # created_volumes list so that it gets cleaned up at object destruction.
    if ($is_internal) {
        push(@{$self->{created_volumes}}, $volume);
    }

    $Log->exit() if $may_exit;
    return $self;
}

=head2 remove_from_set

    # Remove a single specific object from the set:
    ($vol, $mirror) = $set->remove_from_set(
        mirrors => [ $snapmirror ]
    );
    # or
    ($vol, $mirror) = $set->remove_from_set(
        volumes => [ $volume ]
    );

    # Remove several items at once:
    @pairs = $set->remove_from_set(
        volumes => \@vol_objects_or_paths,
        mirrors => \@sm_objects_or_dest_paths
    );

Remove one or more items from this set. Whether the value being removed is a
volume or a snapmirror, both related elements are removed and returned. No
deletion, purging, or other operations are performed on the resources that are
removed from the set. It is up to the user to use the returned values to call
methods such as B<purge>, etc. Also, when an item is removed from the set, it
will B<not> be automatically cleaned up when the set object is destroyed. The
user is now responsible for purging/destroying the snapmirror and the
destination volume.

The return value of the method is a list of those items removed from the set.
The list is a set of ordered pairs of (volume, snapmirror). The volume is
always first in the ordered pair.

B<IMPORTANT NOTE>: If you remove a volume/snapmirror pair that are of type
C<LS>, you really should at least call the B<delete> method on the snapmirror
object, to take it out of the LS rotation. Otherwise, some of the operations
this class performs may still affect it, despite it being removed from the
object's internal set.

=over

=item Options

=over

=item C<< mirrors =E<gt> $array_ref >>

An array-reference of snapmirror objects (B<NACL::STask::Snapmirror> or
B<NACL::C::Snapmirror>) to remove from the set. The corresponding volume
objects are removed as well.

=item C<< volumes =E<gt> $array_ref >>

An array-reference of volume objects (B<NACL::STask::Volume> or
B<NACL::C::Volume>) to remove from the set. The corresponding snapmirror
objects are removed as well.

=back

=item Exceptions

If the parameters do not match the required syntax, a
B<NATE::Exceptions::Argument> exception will be thrown. If any of the passed-in
values (volume or snapmirror) are not part of the set, an exception of the
same class is thrown and I<no> items are removed from the set.

=back

=cut

###############################################################################
#
#   Sub Name:       remove_from_set
#
#   Description:    Remove the given relationship(s) from the set. Return a
#                   list comprised of all the pairs of volume/snapmirror that
#                   get removed as a result of this.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#                   @opts     in      list      Options to this, validated by
#                                                 a call to validate()
#
#   Globals:        %SM_TYPES
#                   $MIRRORS_SPEC
#
#   Returns:        Success:    $self
#                   Failure:    throws exception
#
###############################################################################
sub remove_from_set {
    $Log->enter() if $may_enter;
    my ($self, @opts) = @_;
    my %mirror_map = reverse %{$self->{mirrors}};
    my %uuid2path_map = reverse %{$self->{path_map}};
    my ($mirror, $volume, $path, $uuid, $type, @volumes, @mirrors, @pairs);

    my %opts = validate(
        @opts,
        {
            mirrors => $MIRRORS_SPEC,
            # This is more like $MIRRORS_SPEC, and is not compatible with
            # $VOLUME_TYPE_SPEC.
            volumes => {
                type      => ARRAYREF,
                optional  => 1,
                callbacks => {
                    'content type check' => sub {
                        my $ref = shift;
                        # We've already established that it's an arrayref, now
                        # check that all elements are Volume objects:
                        for my $elem (@{$ref}) {
                            if (! ((blessed $elem) &&
                                   $elem->isa('NACL::C::Volume'))) {
                                return 0;
                            }
                        }
                        return 1;
                    },
                },
            },
        }
    );

    # Check that all the elements we are trying to remove are actually a part
    # of the set. This must be done completely before we proceed. Since the
    # *_in_set methods return the found object, save the returns for later:
    if ($opts{mirrors}) {
        for my $m (@{$opts{mirrors}}) {
            if (my $obj = $self->_mirror_in_set($m)) {
                push(@mirrors, $obj);
            } else {
                $path = $m->destination_path;
                $Log->exit() if $may_exit;
                NATE::Exceptions::Argument->throw(
                    'Attempt to remove a snapmirror that is not part of this ' .
                    "SnapmirrorSet (destination-path $path)"
                );
            }
        }
    }
    if ($opts{volumes}) {
        for my $v (@{$opts{volumes}}) {
            if (my $obj = $self->_volume_in_set($v)) {
                push(@volumes, $obj);
            } else {
                $path = NACL::STask::Snapmirror->construct_volpath(
                    vol_obj_or_volpath => $v
                );
                $Log->exit() if $may_exit;
                NATE::Exceptions::Argument->throw(
                    'Attempt to remove a volume that is not part of this ' .
                    "SnapmirrorSet ($path)"
                );
            }
        }
    }

    # Now process the lists we accumulated during the existence-check (the
    # arrays @mirrors and @volumes).
    for my $obj (@mirrors, @volumes) {
        # Because of the object-wrangling above, we know that this object is
        # the specific referent in our set. Whichever it is, derive the other
        # and remove everything from the relevant tables and arrays.
        if ($obj->isa('NACL::STask::Volume')) {
            # Have volume, need snapmirror

            $volume = $obj;
            $uuid = $self->{obj2uuid_map}->{$volume};
            # It's possible this has already been removed, so skip if it has:
            next if (! $uuid);
            $path = $uuid2path_map{$uuid};
            $mirror = $self->{mirrors}->{$uuid};
        } else {
            # Have snapmirror, need volume

            $mirror = $obj;
            $uuid = $mirror_map{$mirror};
            # It's possible this has already been removed, so skip if it has:
            next if (! $uuid);
            $path = $uuid2path_map{$uuid};
            $volume = $self->{uuid2obj_map}->{$uuid};
        }

        # Delete from the various maps, including our two local ones. Note
        # that we now also remove them from the internal lists of created
        # resources-- these are no longer going to be cleaned up by DESTROY.
        delete $self->{obj2uuid_map}->{$volume};
        delete $uuid2path_map{$uuid};
        delete $self->{mirrors}->{$uuid};
        delete $mirror_map{$mirror};
        delete $self->{path_map}->{$path};
        delete $self->{uuid2obj_map}->{$uuid};
        # Remove from the list in $self->{destinations}->{$type}
        for my $desttype (keys %SM_TYPES) {
            @{$self->{destinations}->{$desttype}} =
                grep { $_ != $volume } @{$self->{destinations}->{$desttype}};
        }
        # Remove $volume from the created_volumes list
        @{$self->{created_volumes}} =
            grep { $_ != $volume } @{$self->{created_volumes}};
        # Remove $mirror from the created_snapmirrors list
        @{$self->{created_snapmirrors}} =
            grep { $_ != $mirror } @{$self->{created_snapmirrors}};

        push(@pairs, $volume, $mirror);
    }

    $Log->exit() if $may_exit;
    return @pairs;
}

=head2 purge

    # Purge all snapmirrors, and internally-created volumes:
    $set->purge();

    # Purge everything, even externally-created volumes:
    $set->purge(purge_all => 1);

    # Purge only the LS mirrors, and only the internally-created
    # volumes, verifying the purge of each:
    $set->purge(
        type            => 'LS',
        nacltask_verify => 1
    );

(Instance method)

This method deletes all the mirrors in snapmirror set, aborting any existing
transfers that may be preventing ONTAP from deleting the relationship entry.
Volumes that were created internally are also deleted, except for the source
volume (if it was created internally). I<All> snapmirror relationships in the
set object are purged, even if they were pre-existing snapmirrors added by
B<new> or B<add_to_set>. If C<purge_all> is passed and true, then even volumes
created outside the set will be purged if their corresponding snapmirror is
purged.

The C<preserve> flag, if passed to B<new>, is ignored here. This is because
an explicit request to purge the snapmirror set is considered a higher
priority.

This method's behavior differs from the object destruction in that it will
delete existing snapmirror relationships that were created externally. It may
also purge volumes that were created externally, depending on whether
C<purge_all> is set.

The source volume is not purged, even if it was created internally, because a
purged snapmirror set object can still be used to add new snapmirror
relationships with the B<add_to_set> method.

Supports CMode CLI.

=over

=item Options

=over

=item C<< type =E<gt> $type|$listref >>

(Optional) Type of Snapmirror relationships to be purged. Value can be C<DP>,
C<LS>, C<XDP> or a list-reference containing one or more of those three. If
C<type> is not provided, by default all the Snapmirror relationships in the set
are purged.

=item C<< purge_all =E<gt> 0|1 >>

(Optional) If this parameter is passed and is a true value, then purge all
volumes whose associated Snapmirror relationships are purged. If this is not
passed, or is false, then only those volumes that were created internally
by this set object will be purged.

=item C<< nacltask_if_purged =E<gt> die|pass >>

(Optional, defaults to C<die>)

If C<die> (default), an exception is thrown if the snapmirror relationship to
be deleted does not exist.

If C<pass>, exits quietly when the snapmirror relationship to be deleted does
not exist.

Note that currently an error is thrown by ONTAP if a deleted relationship is
deleted again.

=item C<< nacltask_verify =E<gt> 0|1 >>

(Optional, defaults to 0)

A boolean value which specifies whether to verify whether the snapmirror
relationship does not exist after the purge has completed. A
L<NACL::Exceptions::VerifyFailure|lib-NACL-Exceptions-VerifyFailure-pm>
exception is thrown if the entry still exists (which is then caught and
gathered into a B<NACL::Exceptions::SnapmirrorSetFailure> exception which
gets thrown at the end of operations).

=back

All other options supported by
L<< NACL::STask::Snapmirror-E<gt>purge|lib-NACL-STask-Snapmirror-pm/purge >>
are supported by the B<NACL::MTask::SnapmirrorSet> B<purge> method as well.

=item Exceptions

If any exceptions are thrown by the underlying calls to B<purge> in
B<NACL::STask::Snapmirror>, this method will throw an exception of the
B<NACL::Exceptions::SnapmirrorSetFailure> class that encapsulates all the
gathered exceptions over the collected elements of this set. Because this
method is purging both snapmirrors and volumes, some of the elements in the
C<exception_objects> key of the exception may be volume objects rather than
snapmirror endpoint paths; check the value of each C<snapmirror> hash key
to see which (if any) are references rather than ordinary strings.

If a B<NACL::Exceptions::SnapmirrorSetFailure> exception is thrown, the
snapmirror relationships referenced by it will be in an indeterminate state;
they may have purged or they may not have. Any relationships I<not> referenced
by the exception should be assumed to have purged cleanly.

=back

=cut

sub purge {
    $Log->enter() if $may_enter;
    my $self = shift;
    my (@errors, $sm_fails, $vol_fails, $purge_all, $mirrors);

    my %opts = validate_with(
        params     => \@_,
        spec       => {
            type               => $PLURAL_TYPES_SPEC,
            purge_all          => { type => BOOLEAN, default => 0 },
            nacltask_verify    => { type => BOOLEAN, default => 0 },
            nacltask_if_purged =>
                NACL::STask::Snapmirror->
                      _if_action_completed_already_validate_spec(),
        },
        allow_extra => 1,
    );

    # If $opts{type} was passed, it is necessary to make (possibly) multiple
    # calls to mirrors() and flatten them into one listref.
    if ($opts{type}) {
        # $opts{type} already passed validation, but the mirrors() method only
        # accepts singular types, not plural. So loop over it and build up
        # @{$mirrors}.
        $mirrors = [];
        my $types = delete $opts{type};
        if (! ref($types)) {
            $types = [ $types ];
        }
        for my $type (@{$types}) {
            push(@{$mirrors}, @{$self->mirrors(type => $type)});
        }
    } else {
        # A plain call to mirrors() returns a hash table, and we only want
        # the values in a listref.
        $mirrors = $self->mirrors;
        $mirrors = [ values %{$mirrors} ];
    }

    # Here's the challenge: We have to purge all the specified snapmirror
    # relationships, but we have to purge *only* those volumes that were
    # created by this SnapmirrorSet object unless "purge_all" was passed. Also,
    # we don't purge the source volume even if it was created within
    # SnapmirrorSet. And we have multiple internal tables to attend to.
    #
    # Approach:
    #   1. Make an object map of the created_volumes list so that we know
    #      which objects are internal
    #   2. Use the derived list of mirrors as the list of things to remove
    #   3. Feed them to a single call to remove_from_set()
    #       * All the internal tables should now be empty
    #       * The returned list gives us the volumes and snapmirrors to purge
    #   4. For each (volume, snapmirror) pair in the list:
    #       * Purge the snapmirror
    #       * Purge the volume if it was created internally or if $purge_all
    #         is set

    $purge_all = delete $opts{purge_all};
    # %vol_opts is a copy of %opts with source_ci and nacltask_release taken
    # out, so that they don't confuse the purge() calls on volumes.
    my %vol_opts = %opts;
    delete @vol_opts{qw(source_ci nacltask_release)};
    $sm_fails = $vol_fails = 0;

    # 1. Make a map of the object referents from created_volumes
    my %objmap = map { $_ => 1 } @{$self->{created_volumes}};

    # 2/3. Use the list derived earlier and remove_from_set() to get the
    #      elements to remove
    my @pairs = $self->remove_from_set(mirrors => $mirrors);

    # 4. Loop over the pairs
    while (@pairs) {
        my ($volume, $mirror) = splice(@pairs, 0, 2);

        # Purge the snapmirror, catching any exception in the running tally
        # in @errors.
        try {
            $mirror->purge(%opts);
        } catch NATE::BaseException with {
            my $exception = shift;
            my $endpt = $mirror->destination_path;

            $sm_fails++;
            push(@errors, { snapmirror => $endpt, error => $exception });
        };

        # Purge the volume (if we should), catching any exception in the
        # running tally in @errors.
        try {
            if ($purge_all || $objmap{$volume}) {
                $volume->purge(%vol_opts);
            }
        } catch NATE::BaseException with {
            my $exception = shift;

            $vol_fails++;
            push(@errors, { snapmirror => $volume, error => $exception });
        };
    }

    if (@errors) {
        my $error_text = sprintf(
            'Exceptions thrown by purge for %d snapmirrors and %s volumes',
            $sm_fails,
            $vol_fails
        );
        $Log->exit() if $may_exit;
        NACL::Exceptions::SnapmirrorSetFailure->throw(
            text              => $error_text,
            exception_objects => \@errors
        );
    }

    $Log->exit() if $may_exit;
    return;
}

=head2 monitor

    # Wait until all mirrors in this set are in snapmirrored-idled state
    $sms->initialize();
    $sms->monitor();

    (or)

    # Wait until all ls mirrors are actively updating
    $sms->update(type => 'ls', background => 1);
    $sms->monitor(
        type                     => 'ls',
        nacltask_wait_for_status => 'transferring'
    );

(Instance method)

The method is used to wait for all of the SnapMirrors involved (which may be a
subset of the mirrors in the set) to reach a given state and/or status. Uses a
CMode-CLI.

=over

=item Options

=over

=item C<< nacltask_wait_for_state =E<gt> $expected_state >>

(Optional) State to monitor mirror for. Default is C<snapmirrored>.
This parameter can be specified as a scalar or as an array-reference.
If specified as a scalar, it is the value of the state which denotes
the completion of the operation.

If specified as an array-reference, it means that when the state reaches
any of the values specified, then the operation has completed.

=item C<< nacltask_wait_for_status =E<gt> $expected_status >>

(Optional) Status to monitor mirror for. Default is C<idle>.
This parameter can be specified as a scalar or as an array-reference.
If specified as a scalar, it is the value of the status which denotes
the completion of the operation.

If specified as an array-reference, it means that when the state reaches
any of the values specified, then the operation has completed.
If nacltask_wait_for_progress is specified and nacltask_wait_for_status
is not passed, nacltask_wait_for_status will be [ 'Preparing', 'Transferring',
'Finalizing']

=item C<< type =E<gt> $type|$listref >>

(Optional) Type of Snapmirror relationship to be monitored. Value can be C<DP>,
C<LS>, C<XDP> or a list-reference containing one or more of those three. If
C<type> is not provided, by default all the Snapmirror relationships in the set
are monitored.

=item C<< method-timeout =E<gt> $timeout >>

(Optional) How long in seconds to wait for the job to finish or for the state
to become $expected_state/$expected_status. Defaults to 1200 seconds.

=item C<< polling_interval =E<gt> $polling_value >>

(Optional) This is the interval (in seconds) after which the value of
the attributes will be polled. If not specified it defaults to 10 seconds.

=back

=item Exceptions

This method will throw a B<NACL::Exceptions::SnapmirrorSetFailure> exception if
there is any timeout or any other exception thrown from the underlying layers.

=back

=cut

sub monitor {
    $Log->enter() if $may_enter;
    my $self = shift;
    my %opts = validate_with(
        params => \@_,
        spec => {
            # task specific options
            'nacltask_wait_for_state' =>
                { type => SCALAR | ARRAYREF, default => 'snapmirrored' }, 
            'nacltask_wait_for_status' =>
                { type => SCALAR | ARRAYREF, default => 'idle' },
            type => $PLURAL_TYPES_SPEC,
        },
        allow_extra => 1,
    );

    my $type = delete $opts{type};
    my @errors;

    my ($mirrors, @mirrors);
    if ($type) {
        if (ref($type) eq 'ARRAY') {
            for my $elem (@{$type}) {
                # When type is passed as an arrayref.
                # Get snapmirror objects of one or more snapmirror type
                $mirrors = $self->mirrors(type => $elem);
                push(@mirrors, @{$mirrors});
            }
        } else {
            # Get snapmirror objects of a particular type
            $mirrors = $self->mirrors(type => $type);
            @mirrors = @{$mirrors};
        }
    } else {
        # If type is not passed, obtain all snapmirror objects in SnapmirrorSet
        $mirrors = $self->mirrors;
        @mirrors = values %{$mirrors};
    }

    # $set_flag is used in STask::Snapmirror->monitor(). This is to avoid sleep
    # of 10 sec during every stask monitor call which will cause latency while
    # obtaining snapmirror state/status. Only for the first call, when set_flag
    # is 0, sleep is executed.
    my $set_flag = 0;
    # Loop through the snapmirror objects and call monitor on them.
    for my $sm (@mirrors) {
        try {
            $sm->monitor(%opts, set_flag => $set_flag);
            $set_flag++;
        } catch NATE::BaseException with {
            my $ex = shift;
            my $err_str = $ex->text;
            my $sm_path =
                ($sm->get_primary_key_string =~ m{source-path=([-\w:/]+)})[0];
            push(@errors, { snapmirror => $sm_path, error => $ex });
        };
    }

    if (@errors) {
        my $text = sprintf(
            'monitor: Monitoring caught %d exception%s',
            scalar(@errors),
            (@errors == 1 ? q{} : 's')
        );
        $Log->exit() if $may_exit;
        NACL::Exceptions::SnapmirrorSetFailure->throw(
            text              => $text,
            exception_objects => \@errors
        );
    }

    $Log->exit() if $may_exit;
    return;
}

=head2 abort

    # Abort the entire set:
    $set->abort;

    # Abort only the LS mirrors:
    $set->abort(type => 'LS');

    # Abort the DP and XDP, but not the LS mirrors:
    $set->abort(type => [ 'DP', 'XDP' ]);

    # Abort only some specific mirrors, as specified by their end-points:
    $set->abort(destinations => [ $dp_vol_1, $dp_vol_2 ]);

    # Abort the entire set in parallel and return without waiting for the
    # tasks to complete:
    $set->abort(parallel => 1, background => 1);

(Can be called as an instance method only.)

Performs a C<snapmirror abort> over all elements in the snapmirror set. If the
set includes any LS mirrors, only one of those is aborted as that operation
applies to all the LS mirrors.

=over

=item Options

The B<abort> method accepts the following options:

=over

=item C<< parallel =E<gt> 0|1 >>

Perform multiple C<abort> operations in parallel, rather than
sequentially. Defaults to 0 (sequential).

=item C<< background =E<gt> 0|1 >>

Do everything in the background, and do not wait for jobs to complete before
returning from the call.

=item C<< destinations =E<gt> $array_ref >>

If passed, the value must be an array-reference containing one or more
end-points to abort. The end-points may be volume or snapmirror objects. If
they are not the same objects that are stored in the set, they will be used to
look up the corresponding internal object, and that object will be the one
actually used to call the underlying B<abort> method. Only one of
B<destinations> or C<type> may be specified in a single call.

=item C<< type =E<gt> $type|$array_ref >>

If passed, the value must be one of C<LS>, C<DP>, C<XDP>, or an array-reference
of some combination of these. This tells B<abort> to only operate on the
destinations of that type (or types). Only one of C<destinations> or C<type>
may be specified in a single call.

=back

In addition to these options, any additional options are passed through to
B<abort> method in the B<NACL::STask::Snapmirror> and
B<NACL::C::Snapmirror> classes (see
L<NACL::STask::Snapmirror|lib-NACL-STask-Snapmirror-pm> and
L<NACL::C::Snapmirror|lib-NACL-C-Snapmirror-pm>).

=item Exceptions

This method may throw a B<NATE::Exceptions::Argument> exception if a bad
parameter is passed. It may also pass through any exceptions thrown by the
B<abort> method of B<NACL::STask::Snapmirror> or
B<NACL::C::Snapmirror>.

=back

=cut

###############################################################################
#
#   Sub Name:       abort
#
#   Description:    Front-end to execute() that adds a hook for LS sets. We
#                   have a special behavior for LS mirrors in the set, where
#                   only one LS mirror needs to be aborted instead of all. 
#                   Other than that, execute() can handle everything for us.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#                   %opts     in      hash      Options to be passed along to
#                                                 execute() (unchanged)
#
#   Returns:        Success:    $self
#                   Failure:    throws exception
#
###############################################################################
sub abort {
    $Log->enter() if $may_enter;
    my ($self, %opts) = @_;

    # We have a special action for LS sets, but other than that execute() can
    # handle this for us. So call it with a pointer to the special hook for LS.
    $self->execute(
        %opts,
        action  => 'abort',
        LS_hook => '_abort_ls'
    );

    $Log->exit() if $may_exit;
    return $self;
}

###############################################################################
#
#   Sub Name:       _abort_ls
#
#   Description:    Perform abort on the LS mirrors in a set, via the
#                   "abort" method of NACL::STask::Snapmirror.
#                   This should only ever be called from _execute(), and makes
#                   assumptions about the setup/structure of private fields on
#                   $self based on being called from there.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#                   $targets  in      listref   Array-reference of the LS
#                                                 end-points being targeted
#                   %opts     in      hash      Additional options to the
#                                                 abort method
#
#   Returns:        Success:    Number of elements in $targets
#                   Failure:    Same, but sets $self->{_results} as well
#
###############################################################################
sub _abort_ls {
    $Log->enter;
    my ($self, $targets, %opts) = @_;

    my %path_from_uuid = reverse %{$self->{path_map}};
    my $first = $targets->[0];
    my $uuid = $self->{obj2uuid_map}->{$first};
    my $mirror = $self->{mirrors}->{$uuid};
    if (! $mirror) {
        my $main_path = $path_from_uuid{$uuid};
        for my $target (@{$targets}) {
            my $path = $path_from_uuid{$self->{obj2uuid_map}->{$target}};
            $self->{_results}->{$path} = NATE::BaseException->new(
                "Panic! No mirror found for LS dest volume $main_path"
            );
        }
    return scalar(@{$targets});
    }

    try {
        my $job_component;
        $mirror->abort(
            %opts,
            job_component => \$job_component
        );
        # Flag all the targets as run
        for my $target (@{$targets}) {
            my $path = $path_from_uuid{$self->{obj2uuid_map}->{$target}};
            $self->{_results}->{$path} = 1;
        }
        # If there is a job_component, save it. But only under the one dest
        # key. Don't make N copies of it.
        if ($job_component) {
            my $path = $path_from_uuid{$uuid};
            $self->{_jobs}->{$path} = $job_component;
        }
    } catch NATE::BaseException with {
        my $exception = shift;

        # For now, just save this exception on all the target's keys in the
        # _results table:
        for my $target (@{$targets}) {
            my $path = $path_from_uuid{$self->{obj2uuid_map}->{$target}};
            $self->{_results}->{$path} = $exception;
        }
    };

    $Log->exit;
    return scalar(@{$targets});

}

=head2 wait_on_jobs

    # Perform a task in parallel, in the background:
    $set->update(parallel => 1, background => 1);
    # Do something else...
    # Now make sure all the background tasks are finished:
    $set->wait_on_jobs;

(May only be called as an instance method.)

Currently supports CMode CLI.

This method iterates through any B<NACL::C::Job> objects created by an
operation that runs in parallel, and waits for the job to finish. If the
job is already finished, it looks it up in the system job history table to
see if it resulted in an error. If any of the jobs resulted in an error,
this method will throw an exception of type
B<NACL::Exceptions::SnapmirrorSetFailure>. If none of the jobs yielded an
error, the method returns with no value.

=over

=item Options

=over

=item C<< method-timeout =E<gt> $integer >>

(Optional, defaults to 120.) Specify the maximum number of seconds a single
call to B<NACL::C::Component::wait_on_job> should wait before timing out.
These calls are made by each individual B<NACL::C::Job> object.

=item C<< polling_interval =E<gt> $integer >>

(Optional, defaults to 10.) Specify the number of seconds that the underlying
call to B<NACL::C::Component::wait_on_job> should wait between attempts to
read the job status.

=back

=item Exceptions

If one or more jobs result in an error, a
B<NACL::Exceptions::SnapmirrorSetFailure> exception will be thrown that
encapsulates all the collected errors.

=back

=cut

###############################################################################
#
#   Sub Name:       wait_on_jobs
#
#   Description:    Poll all the jobs in $self->{_jobs} and make sure they
#                   complete. Save or report any exceptions.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#                   @opts     in      list      Additional options, checked by
#                                                 a call to validate_with()
#
#   Returns:        Success:    void
#                   Failure:    throws exception
#
###############################################################################
sub wait_on_jobs {
    $Log->enter() if $may_enter;
    my ($self, @opts) = @_;
    my %opts = validate(
        @opts,
        {
            'method-timeout' => {
                type    => SCALAR,
                regex   => qr/^\d+$/,
                default => 120,
            },
            polling_interval => {
                type    => SCALAR,
                regex   => qr/^\d+$/,
                default => 10,
            },
        }
    );

    my @errs;
    for my $path (keys %{$self->{_jobs}}) {
        my $job = delete $self->{_jobs}->{$path};
        try {
            $Log->debug(
                'wait_on_jobs: Polling job ID ' . $job->id
            );
            $job->wait_on_job(%opts);
            # If that didn't throw an exception, mark this path as
            # successful in the _results table:
            $self->{_results}->{$path} = 1;
            $Log->debug(
                'wait_on_jobs: Job ID ' . $job->id . ' finished cleanly'
            );
        } catch NATE::BaseException with {
            my $exception = shift;
            $Log->debug(
                'wait_on_jobs: Job ID ' . $job->id . ' caused exception: ' .
                $exception->text
            );
            push(@errs, { snapmirror => $path,
                          error      => $exception });
            # Save this on the _results table in case they want it later:
            $self->{_results}->{$path} = $exception;
        };
    }

    if (@errs) {
        my $text = sprintf(
            'wait_on_jobs: Job harvesting caught %d exception%s',
            scalar(@errs),
            (@errs == 1 ? q{} : 's')
        );
        $Log->exit() if $may_exit;
        NACL::Exceptions::SnapmirrorSetFailure->new(
            text              => $text,
            exception_objects => \@errs
        )->throw;
    }

    $Log->exit() if $may_exit;
    return;
}

###############################################################################
#
#   Sub Name:       execute
#
#   Description:    Execute the specified action as a snapmirror command, over
#                   all the mirrors in the set bounded by $self. The arguments
#                   "dest" or "type" can further restrict the set of mirrors
#                   that are operated upon.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#                   @opts     in      list      Additional options, checked by
#                                                 a call to validate_with()
#
#   Globals:        %SM_TYPES
#                   $SM_TYPES
#
#   Returns:        Success:    dependent on $opts{action}
#                   Failure:    throws exception
#
###############################################################################
sub execute {
    $Log->enter() if $may_enter;
    my ($self, @opts) = @_;
    my %opts = validate_with(
        params => \@opts,
        spec => {
            action => {
                type      => SCALAR,
                optional  => 0, # The only required option
                callbacks => {
                    'valid action' => sub {
                        my $action = shift;
                        return NACL::STask::Snapmirror->can($action) ? 1 : 0;
                    },
                },
            },
            destinations => {
                type      => ARRAYREF | OBJECT,
                optional  => 1,
                callbacks => {
                    'required type or contents' => sub {
                        my $ref = shift;
                        # If $ref isn't an arrayref, make it into one:
                        if (ref($ref) ne 'ARRAY') {
                            $ref = [ $ref ];
                        }
                        # Ensure that each element is a volume or snapmirror
                        # object:
                        for my $elem (@{$ref}) {
                            return 0 if
                                (! $elem->isa('NACL::C::Volume') &&
                                 ! $elem->isa('NACL::C::Snapmirror'));
                        }
                        # If we got through the loop, it's valid:
                        return 1;
                    },
                },
            },
            type => $PLURAL_TYPES_SPEC,
            parallel => {
                type    => BOOLEAN,
                default => 0,
            },
            background => {
                type    => BOOLEAN,
                default => 0,
            },
        },
        allow_extra => 1
    );

    # Only one of 'destinations' or 'type' are allowed:
    if ($opts{destinations} && $opts{type}) {
        $Log->exit() if $may_exit;
        NATE::Exceptions::Argument->throw(
            'Only one of "destinations" or "type" may be given in a call'
        );
    }

    # We know we got this one:
    my $action = delete $opts{action};

    # Massage the inputs into a single set
    my $targets;
    if ($opts{destinations}) {
        # This one is tricky-- if more than one destination was passed in, and
        # they are of different types, we have to make sure we set up $targets
        # correctly.
        my $spec = delete $opts{destinations};
        $targets = {};

        # Make sure $spec is a list so it's easier to process
        if (ref($spec) ne 'ARRAY') {
            $spec = [ $spec ];
        }

        for my $dest (@{$spec}) {
            my ($type, $obj);
            if ($dest->isa('NACL::C::Volume')) {
                # If it is a volume, we want to convert it to the internal
                # NACL::STask::Volume object that represents the same volume.
                # Then we can use our internal tables to get the relevant
                # snapmirror object at later points. We also have to derive
                # the type.

                # First look to see if it is one of our objects
                if ($self->{obj2uuid_map}->{$dest}) {
                    $obj = $dest;
                } else {
                    # It isn't. Try to match it to one of ours.
                    my $uuid = $dest->get_one_state_attribute('uuid');
                    if (! ($obj = $self->{uuid2obj_map}->{$uuid})) {
                        $Log->exit() if $may_exit;
                        NATE::Exceptions::Argument->throw(
                            qq{Volume passed for "destinations" (UUID $uuid)} .
                            ' not found in internal tables'
                        );
                    }
                }

                # At this point, if we haven't thrown an exception yet, $obj
                # should be one of our internal objects. Use it to get the
                # corresponding mirror object and use that to get $type.
                $type = $self->{mirrors}->{$self->{obj2uuid_map}->{$obj}}
                    ->get_one_state_attribute('type');
            } else {
                # If it is a snapmirror, we want to find the the internal
                # NACL::STask::Volume that matches its endpoint, so we can
                # store that and use it later to get the relevant snapmirror
                # object. We also have to derive the type, but that is easier
                # here than above (since we already have a snapmirror object).

                $type = $dest->get_one_state_attribute('type');
                if (! $SM_TYPES{$type}) {
                    # Not sure this can happen, but better to be safe
                    $Log->exit() if $may_exit;
                    NATE::Exceptions::Argument->throw(
                        'Mirror object passed to "destinations" of unknown ' .
                        "type ($type)"
                    );
                }
                my %mirror_map = reverse %{$self->{mirrors}};
                # Is it a mirror object we already have?
                if ($mirror_map{$dest}) {
                    # The value of this key will be the UUID of the destination
                    # NACL::STask::Volume object. Use that map to get the
                    # actual object.
                    my $uuid = $mirror_map{$dest};
                    if (! ($obj = $self->{uuid2obj_map}->{$uuid})) {
                        $Log->exit() if $may_exit;
                        NATE::Exceptions::Argument->throw(
                            qq(Sanity mis-match! No object for UUID $uuid!)
                        );
                    }
                } else {
                    # Not one of ours, we have to go digging
                    my $dest_path = $dest->destination_path;
                    if ($self->{path_map}->{$dest_path}) {
                        my $uuid = $self->{path_map}->{$dest_path};
                        $obj = $self->{uuid2obj}->{$uuid};
                    } else {
                        $Log->exit() if $may_exit;
                        NATE::Exceptions::Argument->throw(
                            'Mirror object passed to "destinations" does not' .
                            ' appear to be in this set (destination-path ' .
                            "$dest_path not found in internal tables)"
                        );
                    }
                }
            }

            push(@{$targets->{$type}}, $obj);
        }
    } elsif ($opts{type}) {
        # This is much easier, just use the destinations() method to get all
        # the internal objects for $opts{type}.
        my $type = delete $opts{type};
        if (ref($type)) {
            %{$targets} =
                map { $_ => $self->destinations(type => $_) } @{$type};
        } else {
            $targets = { $type => $self->destinations(type => $type) };
        }
    } else {
        # "$action" ALL THE THINGS!
        $targets = $self->destinations;
    }

    # All the real work is done by _execute(). It will setup/clear/populate
    # internal tables, execute the methods and collate results, etc.
    $self->_execute($action, $targets, %opts);

    $Log->exit() if $may_exit;
    return $self;
}

###############################################################################
#
#   Sub Name:       _execute
#
#   Description:    Execute $action as a method on all NACL::STask::Snapmirror
#                   objects associated with the NACL::STask::Volume objects
#                   passed in via $targets. Gather the results and return as
#                   a hashref keyed by target volpaths.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#                   $action   in      scalar    Action to take on targets
#                   $targets  in      hashref   The target destinations that
#                                                 will be the recipients of
#                                                 $action as a method-call
#                   %opts     in      hash      Options passed to all method
#                                                 calls
#
#   Globals:        %SM_TYPES
#                   %PARALLEL_CMDS
#
#   Returns:        Success:    $self
#                   Failure:    throws exception
#
###############################################################################
sub _execute {
    $Log->enter() if $may_enter;
    my ($self, $action, $targets, %opts) = @_;
    my (%uuid2path_map, $background, $parallel, $fanout, $count, $total);

    # Clear the internal tables used to store the results of individual
    # method calls.
    %{$self->{_results}} = ();
    %{$self->{_jobs}}    = ();

    # We'll need to map from UUIDs to paths for some of the things we do:
    %uuid2path_map = reverse %{$self->{path_map}};

    # Is there is special action to take on LS targets?
    my $ls_hook = delete $opts{LS_hook};

    # Don't want these two to propagate to any sub-calls:
    $parallel   = delete $opts{parallel};
    $background = delete $opts{background};

    # Are we supposed to run in parallel? The default is sequential.
    if (defined $parallel) {
        # It's not enough to test boolean truth-- I need $parallel to be
        # exactly 1 or 0.
        $parallel = $parallel ? 1 : 0;
    } else {
        $parallel = 0;
    }
    if ($background) {
        # Setting "background" forces "parallel"
        $parallel = 1;
    }
    $opts{nacltask_wait} = 1 - $parallel;
    $fanout = $parallel ? 'Parallel' : 'Sequential';

    $total = 0;
    for my $type (keys %SM_TYPES) {
        # Skip this $type if it wasn't passed or if the list is empty:
        next if ((! $targets->{$type}) || (! @{$targets->{$type}}));

        $Log->debug(
            "$fanout execution: Action $action, type $type, starting"
        );

        if (($type eq 'LS') && $ls_hook) {
            $Log->debug(
                "$fanout execution: Executing special LS hook for $action"
            );
            $count = $self->$ls_hook($targets->{$type}, %opts);
            $total += $count;
        } else {
            $count = 0;
            for my $dest (@{$targets->{$type}}) {
                $count++; $total++;
                my $uuid   = $self->{obj2uuid_map}->{$dest};
                my $path   = $uuid2path_map{$uuid};
                my $mirror = $self->{mirrors}->{$uuid};
                if (! $mirror) {
                    $self->{_results}->{$path} = NATE::BaseException->new(
                        "Panic! No mirror found for dest obj $path"
                    );
                } else {
                    # Run the command. Pass a job_component in case the task in
                    # question is backgrounded.
                    my $job_component;
                    try {
                        my @result = $mirror->$action(
                            %opts,
                            job_component => \$job_component
                        );
                        if ($job_component) {
                            $self->{_jobs}->{$path} = $job_component;
                        }
                        if (@result || (! defined($result[0]))) {
                            # This is a carry-over from the GNOTE SnapMirrorSet
                            # code. I'm not sure that any of these methods
                            # actually return a list, and I think the ones that
                            # return anything other than void just return
                            # $self...
                            $self->{_results}->{$path} = (@result > 1) ?
                                \@result : $result[0];
                        } else {
                            # Method call returned void (or it returned an
                            # explicit undef, which functions really shouldn't
                            # do). Just set the value here to true:
                            $self->{_results}->{$path} = 1;
                        }
                    } catch NATE::BaseException with {
                        my $exception = shift;
                        $Log->warn(
                            "_execute: For $type snapmirror relationship on " .
                            "dest volume $path, method $action threw an " .
                            'exception: ' . $exception->text
                        );
                        $self->{_results}->{$path} = $exception;
                    };

                    if (($action eq 'purge') || ($action eq 'delete')) {
                        # If the action they just performed was "purge" or
                        # "delete", then also remove the volume/snapmirror pair
                        # from the set:
                        $self->remove_from_set(mirrors => [ $mirror ]);
                    } elsif ($action eq 'initialize') {
                        # If it was "initialize", then mark the flag for this
                        # volume on $self->{_initialized}:
                        $self->{_initialized}->{$dest} = 1;
                    }
                }
            }
        }

        $Log->debug(
            "$fanout execution: Action $action, type $type, $count executed"
        );
    }

    $Log->debug(
        "$fanout execution: Action $action, $total snapmirrors executed"
    );

    # If this run was in parallel, and if they are *not* backgrounding it,
    # then we wait for all the jobs to complete.
    if ($parallel && !$background) {
        $Log->debug(
            "$fanout execution: Waiting for parallel jobs to complete");
        $self->wait_on_jobs(
            'method-timeout'   => $opts{'method-timeout'}   || 120,
            'polling_interval' => $opts{'polling_interval'} || 10,
        );
    }

    # Now look at the "_results" table and see if there were any exceptions
    # caught during the execution of this command across the targets. If so,
    # we need to throw a NACL::Exceptions::SnapmirrorSetFailure that holds
    # all of the per-mirror exceptions.
    my @ex_list;
    for my $path (keys %{$self->{_results}}) {
        if (blessed $self->{_results}->{$path} &&
                $self->{_results}->{$path}->isa('NATE::BaseException')) {
            push(@ex_list, { snapmirror => $path,
                             error      => $self->{_results}->{$path} });
        }
    }
    if (@ex_list) {
        my $text = sprintf(
            '_execute: Action "%s" triggered %d exception%s across %d ' .
                'snapmirror%s executed',
            $action,
            scalar(@ex_list),
            (@ex_list == 1 ? q{} : 's'),
            $total,
            ($total == 1 ? q{} : 's')
        );
        $Log->exit() if $may_exit;
        NACL::Exceptions::SnapmirrorSetFailure->new(
            text              => $text,
            exception_objects => \@ex_list
        )->throw;
    }

    $Log->exit() if $may_exit;
    return $self;
}

###############################################################################
#
#   Sub Name:       _update_source
#
#   Description:    Update the "source" attribute with the given volume. If
#                   volume is undef just update the other source-related
#                   attributes from the existing source.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#                   @opts     in      array     Options to the call
#
#   Returns:        Success:    $self
#                   Failure:    throws exception
#
###############################################################################
sub _update_source {
    $Log->enter() if $may_enter;
    my ($self, @opts) = @_;
    my %opts = validate(
        @opts,
        {
            volume => {
                type      => UNDEF | OBJECT,
                callbacks => {
                    class => sub {
                        my $obj = shift;
                        # We've established that it is either undef or blessed
                        return $obj ? $obj->isa('NACL::STask::Volume') : 1;
                    },
                },
            },
        }
    );

    my $new_source = $opts{volume};
    if ((! $new_source) && (! $self->{source})) {
        $Log->exit() if $may_exit;
        NATE::Exceptions::Argument->throw(
            'Cannot pass "undef" to _update_source if no source currently set'
        );
    }

    if ($new_source) {
        my $old_source = $self->{source};
        $self->{source} = $new_source;

        if ($old_source) {
            # If there is an old source, we need to remove it from the internal
            # list of volumes to clean up at object destruction. That's because
            # this is almost certainly happening because of a LS promote
            # operation, and ONTAP has already destroyed the old source volume.
            @{$self->{created_volumes}} =
                grep { $_ ne $old_source } @{$self->{created_volumes}};
        }
    }

    # Now that $self->{source} is set or updated, derive and cache the volpath
    # in both SN and pre-SN formats:
    my $source_path = NACL::STask::Snapmirror->construct_volpath(
        vol_obj_or_volpath => $self->{source}
    );
    # If we just got the SN-style path, save it as that and derive the RR
    # path. Otherwise save it as the "plain" path and derive the SN-style:
    if ($source_path =~ m{^[-\w]+://([-\w]+)/([-\w]+)$}) {
        # We got the RR-style path
        $self->{source_path} = $source_path;
        # The SN-style path is just the two elements we matched in the regexp:
        $self->{source_path_sn} = "$1:$2";
    } elsif ($source_path =~ /^([-\w]+):([-\w]+)$/) {
        # We got the SN-style path
        $self->{source_path_sn} = $source_path;
        # But to get the RR-style path, we need cluster name, and we don't
        # have that handy. We have to get it. But for the moment save the
        # parts we do have from the regexp match:
        $self->{source_path} = "://$1/$2";
        my $clus_cs_obj = NACL::CS::ClusterIdentity->fetch(
            command_interface => $self->{source}->command_interface,
            requested_fields  => [ 'name' ],
            allow_empty       => 1
        );
        if ($clus_cs_obj) {
            $self->{source_path} = $clus_cs_obj->name . $self->{source_path};
        } else {
            $Log->exit() if $may_exit;
            NATE::BaseException->throw(
                'Unable to derive the name of the source cluster'
            );
        }
    } else {
        $Log->exit() if $may_exit;
        NATE::BaseException->throw(
            "construct_volpath() built a path of unknown format: $source_path"
        );
    }
    # Cache the UUID:
    $self->{source_uuid} = $self->{source}->get_one_state_attribute('uuid');

    $Log->exit() if $may_exit;
    return $self;
}

###############################################################################
#
#   Sub Name:       _create_volumes
#
#   Description:    Create one or more volumes, using the values passed in as
#                   options to the NACL::STask::Volume::create call(s). All
#                   volumes we create in this way are marked for cleanup.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#                   %opts     in      hash      Options to the create calls,
#                                                 also a few of our own
#
#   Returns:        Success:    List reference of vols created or single vol
#                   Failure:    throws exception
#
###############################################################################
sub _create_volumes {
    $Log->enter() if $may_enter;
    my ($self, %opts) = @_;
    my $vols_created;

    # Strip off the arguments we care about:
    my $source_volume = delete $opts{source_volume} || 0;
    my $count         = delete $opts{count}         || 1;

    if ($source_volume) {
        $vols_created = NACL::STask::Volume->create(
            %opts,
            type          => 'RW',
            nacltask_wait => 1
        );
        # If no exception was thrown, the creation succeeded. Save to the
        # internal clean-up table:
        push(@{$self->{created_volumes}}, $vols_created)
    } else {
        $vols_created = [];
        for my $index (1 .. $count) {
            my $volume = NACL::STask::Volume->create(
                %opts,
                # These are always destination volumes, so type is always DP:
                type          => 'DP',
                nacltask_wait => 1
            );
            # If no exception was thrown, the creation succeeded. Save to the
            # internal clean-up table:
            push(@{$self->{created_volumes}}, $volume);
            # Put on the return-array as well:
            push(@{$vols_created}, $volume);
        }
    }

    $Log->exit() if $may_exit;
    return $vols_created;
}

###############################################################################
#
#   Sub Name:       _create_snapmirrors
#
#   Description:    Create NACL::STask::Snapmirror objects for the volumes
#                   passed in, or for all of the current volumes associated
#                   with the type, if no specific volumes are passed in.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#                   %opts     in      hash      Key/value argument pairs
#
#   Returns:        Success:    array reference of snapmirrors created
#                   Failure:    throws exception
#
###############################################################################
sub _create_snapmirrors {
    $Log->enter() if $may_enter;
    my ($self, %opts) = @_;
    my @sm_created;

    my $type = $opts{type};
    my $vols = $opts{volumes} || $self->{destinations}->{$type};

    for my $volume (@{$vols}) {
        # We index the snapmirror objects in our table by the UUID of the
        # destination volume:
        my $uuid = $self->{obj2uuid_map}->{$volume};
        my $snapmirror = NACL::STask::Snapmirror->create(
            %{$self->{snapmirror_options}->{all}},
            %{$self->{snapmirror_options}->{$type}},
            command_interface  => $volume->command_interface,
            'source-path'      => $self->{source},
            'destination-path' => $volume,
            type               => $type,
            # Always force this, as generic initialize doesn't work for LS:
            do_initialize      => 0,
            # Make sure the creation completes before returning:
            nacltask_wait      => 1
        );

        # Assuming we didn't just throw an exception, store it on {mirrors}
        # and on {created_snapmirrors}:
        $self->{mirrors}->{$uuid} = $snapmirror;
        push(@{$self->{created_snapmirrors}}, $snapmirror);
        # And one more, for the returned list:
        push(@sm_created, $snapmirror);
    }

    $Log->exit() if $may_exit;
    return \@sm_created;
}

###############################################################################
#
#   Sub Name:       _process_mirror
#
#   Description:    Process a single existing NACL::STask::Snapmirror object
#                   by learning what its type is, and attaching to the dest
#                   volume.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#                   %opts     in      hash      Argument pairs
#
#   Globals:        %SM_TYPES
#                   $SM_TYPES
#
#   Returns:        Success:    $self
#                   Failure:    throws exception
#
###############################################################################
sub _process_mirror {
    $Log->enter() if $may_enter;
    # As this is a private method, we've already validated the parameters
    my ($self, %opts) = @_;

    my $mirror = $opts{mirror};
    # If it is a component, cast it to a STask
    if (! $mirror->isa('NACL::STask::Snapmirror')) {
        $mirror = $mirror->cast_component_to_task();
    }

    # We'll be using this to attach to the destination volume so we can
    # add it to our internal tables:
    my $ci = $mirror->command_interface;
    # This will give us the type of the snapmirror, as well as the destination
    # volume:
    my $state = $mirror->state(
        requested_fields => [
            qw(type destination-vserver destination-volume source-path)
        ]
    );
    # Get some data we need from the state:
    my $type    = $state->type;
    my $vserver = $state->destination_vserver;
    my $volume  = $state->destination_volume;
    my $source  = $state->source_path;

    # Sanity-check the value of $type:
    if (! $SM_TYPES{$type}) {
        $Log->exit() if $may_exit;
        NATE::Exceptions::Argument->throw(
            qq(Snapmirror type "$type" not recognize, not one of $SM_TYPES)
        );
    }

    # Is this snapmirror using our $self->{source} as the source?
    if (($source ne $self->{source_path}) &&
        ($source ne $self->{source_path_sn})) {
        $Log->exit() if $may_exit;
        NATE::Exceptions::Argument->throw(
            qq(Snapmirror source "$source" does not match set source ) .
            "(one of '$self->{source_path_sn}' or '$self->{source_path}')"
        );
    }

    my $dest_volume = NACL::STask::Volume->find(
        command_interface => $ci,
        allow_empty       => 1,
        filter            => {
            vserver => $vserver,
            volume  => $volume,
        }
    );
    # Did we find it?
    if (! $dest_volume) {
        $Log->exit() if $may_exit;
        NATE::Exceptions::Argument->throw(
            "Failed to find dest volume for mirror ($vserver:$volume)"
        );
    }
    # Now that we have the volume, we need the UUID to store $mirror
    my $uuid = $dest_volume->get_one_state_attribute('uuid');
    # OK, we have everything we need at this point. Store things on $self in
    # the proper places. We didn't create this volume or snapmirror, so they
    # don't have to go on the lists for cleanup. Just call _process_volume to
    # set up the various mappings and store it on the proper internal tables.
    $self->_process_volume(
        volume => $dest_volume,
        type   => $type,
        uuid   => $uuid
    );
    $self->{mirrors}->{$uuid} = $mirror;

    $Log->exit() if $may_exit;
    return $self;
}

###############################################################################
#
#   Sub Name:       _process_volume
#
#   Description:    Process a single existing NACL::STask::Volume object
#                   by deriving anything we need from the state and pushing
#                   it onto the relevant internal lists.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#                   %opts     in      hash      Argument pairs
#
#   Globals:        %SM_TYPES
#                   $SM_TYPES
#
#   Returns:        Success:    $self
#                   Failure:    throws exception
#
###############################################################################
sub _process_volume {
    $Log->enter() if $may_enter;
    my ($self, %opts) = @_;

    # Because this is a private function, I'm assuming the args are pretty
    # much vetted. But I will check 'type' for sanity.
    my $volume = $opts{volume};
    my $type   = $opts{type};
    if (! $SM_TYPES{$type}) {
        $Log->exit() if $may_exit;
        NATE::Exceptions::Argument->throw(
            qq(Value "$type" for type not one of $SM_TYPES)
        );
    }

    # We need the UUID from $volume, and the volpath as NACL::STask::Snapmirror
    # sees it. We might have gotten the UUID in the call, if this was called
    # from _process_volume:
    my $uuid = $opts{uuid} || $volume->get_one_state_attribute('uuid');
    my $path = NACL::STask::Snapmirror->construct_volpath(
        vol_obj_or_volpath => $volume
    );
    # Quick, check to see if we already have this UUID. Trying to add the same
    # volume a second time probably means something is wrong somewhere:
    if ($self->{uuid2obj_map}->{$uuid}) {
        $Log->exit() if $may_exit;
        NATE::Exceptions::Argument->throw(
            qq(Attempt to add volume "$path" (UUID $uuid) more than once)
        );
    }

    # Save $volume on the proper destinations table, and add mappings for
    # UUID and path:
    push(@{$self->{destinations}->{$type}}, $volume);
    $self->{path_map}->{$path} = $uuid;
    $self->{uuid2obj_map}->{$uuid} = $volume;
    # Yes, I am using the object referent as a hash key.
    $self->{obj2uuid_map}->{$volume} = $uuid;

    $Log->exit() if $may_exit;
    return $self;
}

###############################################################################
#
#   Sub Name:       _process_volume_spec
#
#   Description:    Process the listref of one or more volumes/specs. This
#                   means just copying over those that are already volume
#                   STask objects, and creating volumes for those entries that
#                   are hashref specs.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#                   %opts     in      hash      Argument pairs
#
#   Returns:        Success:    listref of added volumes
#                   Failure:    throws exception
#
###############################################################################
sub _process_volume_spec
{
    $Log->enter() if $may_enter;
    # No need to validate %opts, as this is only called from within our public
    # methods where the data has already been validated.
    my ($self, %opts) = @_;
    my @added_volumes;

    my $type = $opts{type};
    my $spec = $opts{spec};

    for my $entry (@{$spec}) {
        if ((blessed $entry) && $entry->isa('NACL::C::Volume')) {
            # If $entry is a component, make it a STask:
            if (! $entry->isa('NACL::STask::Volume')) {
                $entry = $entry->cast_component_to_task();
            }
            $self->_process_volume(volume => $entry, type => $type);
            push(@added_volumes, $entry);
        } elsif (ref($entry) eq 'HASH') {
            my $created = $self->_create_volumes(
                %{$entry},
                # Make sure the new volumes are put into the right internal
                # tables:
                type => $type
            );
            # The above only created the volumes. This call puts each volume
            # on the internal tables and sets up some key mappings.
            for my $volume (@{$created}) {
                $self->_process_volume(volume => $volume, type => $type);
            }
            push(@added_volumes, @{$created});
        } else {
            $Log->exit() if $may_exit;
            NATE::Exceptions::Argument->throw(
                qq($type spec value "$entry" must be a hash reference or a ) .
                'NACL::STask::Volume object'
            );
        }
    }

    $Log->exit() if $may_exit;
    return \@added_volumes;
}

###############################################################################
#
#   Sub Name:       _cleanup_volumes_and_snapmirrors
#
#   Description:    Purge any snapmirror relationships and volumes that were
#                   created internally by this object.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#
#   Returns:        Success:    $self
#                   Failure:    throws exception
#
###############################################################################
sub _cleanup_volumes_and_snapmirrors {
    $Log->enter() if $may_enter;
    my ($self) = @_;

    for my $key (qw(created_snapmirrors created_volumes)) {
        while (@{$self->{$key}}) {
            my $object = pop(@{$self->{$key}});
            # Three cheers for polymorphism... we just have to make sure we
            # call purge() on the snapmirrors first, before calling it on any
            # volumes.
            try {
                # For some reason, there have been occasional time-outs on
                # purging volumes, particularly in the "offline" step...
                $object->purge('method-timeout' => 600);
            } catch NATE::BaseException with {
                my $exception = shift;
                my $type = ref($exception);

                # When we're calling this routine, we're either bailing out
                # from a failure or we're called from DESTROY when an object
                # is being torn down. Letting the exception bubble up won't
                # do anything, so just log a message about it for now.
                $Log->warn(
                    '_cleanup_volumes_and_snapmirrors: An exception of type ' .
                    "$type was caught during cleanup: " .
                    $exception->text
                );
            };
        }
    }

    $Log->exit() if $may_exit;
    return $self;
}

###############################################################################
#
#   Sub Name:       _mirror_in_set
#
#   Description:    Return a boolean as to whether the given snapmirror is a
#                   part of this SnapmirrorSet. Note that the object may be
#                   different than any of our referents but still point to an
#                   existing snapmirror.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#                   $mirror   in      ref       Object of NACL::C::Snapmirror
#                                                 or NACL::STask::Snapmirror
#                                                 classes
#
#   Returns:        Success:    Mirror object if it exists, 0 otherwise
#                   Failure:    throws exception
#
###############################################################################
sub _mirror_in_set {
    $Log->enter() if $may_enter;
    my ($self, $mirror) = @_;
    my $found = 0;

    # First check against $self->{mirrors}, the easy test:
    my %mirror_map = reverse %{$self->{mirrors}};
    if ($mirror_map{$mirror}) {
        $found = $mirror;
    } else {
        # Not that easy. Get the destination path and see if we have that:
        my $dest_path = $mirror->destination_path;
        my $uuid = $self->{path_map}->{$dest_path};
        if ($self->{mirrors}->{$uuid}) {
            $found = $self->{mirrors}->{$uuid};
        }
    }

    $Log->exit() if $may_exit;
    return $found;
}

###############################################################################
#
#   Sub Name:       _volume_in_set
#
#   Description:    Return a boolean as to whether the given volume is a part
#                   of this SnapmirrorSet. Note that the object may be
#                   different than any of our referents but still point to an
#                   existing volume.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#                   $volume   in      ref       Object of NACL::STask::Volume
#                                                 or NACL::C::Volume classes
#
#   Returns:        Success:    Volume object if it exists, 0 otherwise
#                   Failure:    throws exception
#
###############################################################################
sub _volume_in_set {
    $Log->enter() if $may_enter;
    my ($self, $volume) = @_;
    my $found = 0;

    # First check obj2uuid_map, the easiest test:
    if ($self->{obj2uuid_map}->{$volume}) {
        $found = $volume;
    } else {
        # Not that easy. Get the UUID for this volume and see if we have that:
        my $uuid = $volume->get_one_state_attribute('uuid');
        if ($self->{uuid2obj_map}->{$uuid}) {
            $found = $self->{uuid2obj_map}->{$uuid};
        }
    }

    $Log->exit() if $may_exit;
    return $found;
}

###############################################################################
#
#   Sub Name:       DESTROY
#
#   Description:    Class destructor. Called when an object of this class is
#                   finally destroyed.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#
#   Returns:        Success:    void
#                   Failure:    throws exception
#
###############################################################################
sub DESTROY {
    $Log->enter() if $may_enter;
    my ($self) = @_;

    # Unless the user specifically asked this object to retain all its
    # resources, call the method that cleans up the volumes and snapmirrors
    # that were created by us.
    local $@; # Preserve any existing errors
    if (! $self->{preserve}) {
        $self->_cleanup_volumes_and_snapmirrors();
    }

    $Log->exit() if $may_exit;
    return;
}

1;
