
# $Id: //depot/prod/test/nacldev/lib/NACL/MTask/SprayCore.pm#3 $

#
# Copyright (c) 2001-2016 NetApp, Inc., All Rights Reserved
# Any use, modification, or distribution is prohibited
# without prior written consent from NetApp, Inc.
#

## @summary MTask Module
## @author dl-nacl-vbu@netapp.com
## @status shared
## @pod here

=head1 NAME

    NACL::MTask::Spraycore

=head1 DESCRIPTION

    "NACL::MTask::Spraycore" provides a consolidated method for obtaining a
    core on a non HA pair and uploading it to a shared location using existing
    functionality accessible via

    C<NACL::C::SystemNodeCoredump>

    C<NACL::C::SystemNodeCoredumpConfig>

    C<NACL::C::Transit>

    C<NACL::APIset>

    The MTask will be concerned with -

        1) Force the system to panic.

        2) Dump and save the corefile.

        3) Upload to remote location (Only for C-Mode).


=head1 ATTRIBUTES

=head2 command_interface

(required) A component object that represents the host to which send commands to.
See L<NACL::C::Component::command_interface|lib-NACL-C-Component-pm/command_interface>.

=head1 METHODS

=head2 new

    (Cmode Method call)
    my $spraycoreObj = NACL::MTask::Spraycore->new(
        command_interface     => $command_interface
        delete_existing_cores => 0, # ( default => 1 )
        core_upload_location  => $core_upload_server
    );

    (7mode Method call)
    my $spraycoreObj = NACL::MTask::Spraycore->new(
        command_interface     => $command_interface
        delete_existing_cores => 0 , # ( default => 1 )
    );

    (Class method) This method will instantiate a spraycore object and set its
    attributes to their default values.

    Steps performed are :
    1) Set a core upload location. (Only for CMode)
    2) Delete any pre-exisiting cores in the system based on delete_existing_cores.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component::command_interface|lib-NACL-C-Component-pm/command_interface>.
A command interface object that will be used to communicate with the user specified node.

Applicable for both Cmode and 7Mode.

=item C<< delete_existing_cores => $delete_existing_cores >>

(optional, default 1 ) Option for the user to indicate whether to delete the core files existing in the node.
By default, it takes a value 1 and hence deletes all the existing core files.

Applicable for both Cmode and 7Mode.

=item C<< core_upload_location => $core_upload_location >>

Optional) By the default the core path is retrieved from the site.info file but can be overwritten by user input.

Applicable for Cmode.

=item All of the other various options described in,
L<< NACL::C::SystemNodeCoredump::delete_all     |lib-NACL-C-SystemNodeCoredump-pm/delete_all >>

=back

=back

=over

=item Exceptions

=over

=item C<< NATE::BaseException >>

=back

=back

=cut

## Package definition
package NACL::MTask::Spraycore;

## Compiler directives
use strict;
use warnings;

## Inherit the coredump components
use base qw (NACL::MTask::MTask);

## Module import
use NACL::ComponentUtils qw(Dumper);
use NACL::APISet;
use NACL::C::CommandInterface;
use NACL::Transit;
use Options;
use Params::Validate
  qw(validate validate_with BOOLEAN HASHREF OBJECT ARRAYREF SCALAR UNDEF);
use TestUtils qw(get_site_info);
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Utils qw(make_host);
use NACL::C::SystemNodeCoredump;
use NACL::CS::Cluster;
use NACL::CS::StorageFailover;
use NACL::C::Failover;
use NACL::C::SystemNodeCoredumpConfig;
use NACL::C::ClusterHa;
use NACL::CS::SystemNodeCoredumpStatus;
## Execption import
use NACL::APISet::Exceptions::ResponseException qw(:try);
use NACL::APISet::Exceptions::CommandFailedException;
use NACL::Exceptions::NoElementsFound;
use NATE::BaseException qw(:try);
use NACL::MTask::_DumpCore qw(run_coretool);
use NACL::STask::Node;
use NACL::C::Exceptions::SystemNodeCoredumpSegment::NotSupportedForThisPlatform ();

use Class::MethodMaker [
    scalar =>
      [ { -type => 'NACL::C::CommandInterface::ONTAP' }, 'command_interface' ],

];

###################################################
# Subroutine Name: new()                          #
# Objective:       Create a new SprayCore object  #
# Details:         See POD documentation below    #
# Returns : $self                                 #
###################################################

sub new {

    $Log->enter() if $may_enter;

    ## Grab the object handle from
    ## the input array @_
    my ( $class, @args ) = @_;

    ## Local variables
    my ( $core_upload_location, $hosttype, $self, $site, $upload_server );

    ## Validate the input
    my %args = validate_with(
        params => \@args,
        spec   => {
            command_interface 	  => { isa => 'NACL::C::CommandInterface::ONTAP' },
            core_upload_location  => { type => SCALAR | UNDEF,  optional => 1 },
            delete_existing_cores => { type => SCALAR, 		default  => 1 }
        }
    );

    ## Bless the class instance
    $self = bless {}, $class;

    ## Pass the caller's command interface ref to the
    ## spraycore object
    $self->{command_interface} = $args{command_interface};
    if ( defined $args{core_upload_location} ) {
        $self->{core_upload_location} = $args{core_upload_location};
    }

    ## Get the hosttype and site of the system
    ## from the tharnhost file
    $hosttype = $self->{command_interface}->hostrec()->hosttype();
    $site     = $self->{command_interface}{hostrec}{SITE};

    ## If the system is a simulator, throw an exception
    ## since SprayCore is not supported on sims
    if ( $hosttype =~ /vsim/i || $hosttype =~ /vsa/i ) {
        NATE::BaseException->throw(
            'Spraycore is not supported on vsim and vsa');
    }

    ## If the user specified an upload server
    ## use it for the upload
    if ( defined $args{core_upload_location} ) {

        ## store the user-specified upload location
        ## into the spraycore pbject
        $self->{core_upload_location} = $args{core_upload_location};
        $core_upload_location = $self->{core_upload_location};
        $Log->comment(
"The user specified core upload location will be used : $core_upload_location\n"
        );
    }

    ## If the user didn't specify an upload server,
    ## Fetch the dump server for the system's site.
    if ( !defined $args{core_upload_location} ) {

        $Log->comment("Reading the contents of the site info file...\n");
        my $options = new Options();
        $options->process();

        eval { %{ $self->{site_info} } = get_site_info(); };

        if ($@) {
            NATE::BaseException->throw(
                'Unable to read the contents of the site info file');
        }

        ## Read the dump server from the site info file
        ## and pass it to the spraycore object
        $self->{core_upload_server} =
          $self->{site_info}->{$site}->{netboot_server};
        $core_upload_location = $self->{core_upload_server};

        ## If there is no entry for the dump server
        ## throw an exception
        if ( !defined $core_upload_location ) {
            NATE::BaseException->throw(
                'Unable to find the dump server for site $site');
        }

        ## Set the upload server to the dump server that
        ## was fetched in from the site info file
        if ( defined $core_upload_location ) {
            $Log->comment(
"The default core upload location for site $site will be used : $core_upload_location\n"
            );

        }

    }    # end if()

    ## Cleanup pre-existing cores if the
    ## flag is set
    if ( $args{delete_existing_cores} == 1 ) {
        $self->_cleanup_cores();
		$self->cleanup_core_segment();
    }
    else {
        ## Keep pre-existing cores if the flag
        ## is not set and print a message
        $Log->comment( "delete_existing_cores was set to 0.\n"
              . "Keeping pre-existing cores ..." );
    }

    $Log->exit() if $may_exit;
    return $self;

}    # end new()

###########################################################
# Routine : start()                                       #
# Description : Subroutine to start the spraycore task    #
# Details: See POD documentation below                    #
# Returns : Nothing                                       #
###########################################################

=head2 start

    (Class call)
    my $spraycoreObj = NACL::MTask::Spraycore->start(command_interface => $command_interface);

    or

    (Instance call)
    $spraycoreObj->start();

    (Class or Instance method) Start the spraycore task.

    Steps peformed by this method :
	1) Verify whether disks are available that support sprayed dumpcore.
    2) Perform sparse core setting.
         It includes setting options coredump.metadata_only to on,
                                     cf disable,
    3) Panic the system and display dump information.
    4) Wait for node to reboot and wait until 'automatic' save core is done.
    5) Display core metadata after save and verify the core with coretool.
	6) Checks whether the upload switch was set to start the upload operation. (Only for CMode)

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
A component object that represents the host to send commands.
See L<NACL::C::Component::command_interface|lib-NACL-C-Component-pm/command_interface>.

Applicable for both Cmode and 7Mode.

=item C<< upload => boolean >>

(optional) Default is 1.
If set to 0, the upload operation will not happen.

Applicable for both Cmode and 7Mode.

=item C<< core-tool-loc => $core_tool_loc >>

(optional) Core tool location

Applicable for both Cmode and 7Mode.

=item C<< file-path => $file_path >>

(optional) Corefile location

Applicable for both Cmode and 7Mode.

=item C<< core-tool-path => $core_tool_path >>

(optional) Core tool utility path, where it will get downloaded.

Applicable for both Cmode and 7Mode.

=item C<< nacltask_verify => $nacltask_verify >>

(optional) Default is 1.
If set to 1, run_coretool() will be called and corefile will get verified.

If set to 0, verification will not be performed.

Applicable for both Cmode and 7Mode.

=item All of the other various options described in,

L<< NACL::C::SystemNodeCoredump::save_all       |lib-NACL-C-SystemNodeCoredump-pm/save_all >>
L<< NACL::C::SystemNodeCoredump::delete_all     |lib-NACL-C-SystemNodeCoredump-pm/delete_all >>
L<< NACL::C::SystemNodeCoredumpConfig::modify   |lib-NACL-C-SystemNodeCoredumpConfig-pm/modify >>
L<< NACL::CS::SystemNodeCoredumpStatus::fetch    |lib-NACL-CS-SystemNodeCoredumpStatus-pm/fetch >>
L<< NACL::C::StorageFailover::modify            |lib-NACL-C-StorageFailover-pm/modify >>
L<< NACL::CS::StorageFailover::fetch             |lib-NACL-CS-StorageFailover-pm/fetch >>

=back

=back

=over

=item Exceptions

=over

=item C<< NACL::APISet::Exceptions::MissingRequiredParamException >>

=item C<< NACL::APISet::Exceptions::CommandFailedException >>

=item C<< NATE::BaseException >>

=back

=back

=cut

sub start {

    ## Grab the object handle from
    ## the input array @_
    my ( $self, @args ) = @_;

    ## Validate the input
    my %args = validate_with(
        params => \@args,
        spec   => {
            upload          => { type => BOOLEAN, default  => 1 },
            'core-tool-loc' => { type => SCALAR,  optional => 1 },
            'file-path'     => { type => SCALAR, optional => 1 },
            'core-tool-path'=> { type => SCALAR, optional => 1 },
            nacltask_verify => { type => BOOLEAN, default  => 1 },
            }
    );

    ## check cluster ha and storage failover states.
    ## disable both if the're enabled. Also set
    ## save-onstartup to true
    $self->_check_config();

    ## Panic the system
    $self->_panic();

	## Connection may be down because of panic
	## Refresh the command interface

	$self->command_interface()->refresh_command_interface();

    NACL::MTask::_DumpCore->run_coretool(
        'core-tool-loc'   => $args{'core-tool-loc'},
        'core-id'         => $self->{corefile},
        'file-path'       => $args{'file-path'},
        'core-tool-path'  => $args{'core-tool-path'},
        command_interface => $self->{command_interface},
    ) if ($args{nacltask_verify});    

    ## Upload the core file if the flag
    ## is set
	$self->_upload() if (( $args{upload} == 1 ) && ($self->{command_interface}->is_cmode() ));

}    # end start()

###########################################################
# Start of Private/Local Subroutines ######################
# Note: The Routines Are Listed in Alphabetical Order #####
###########################################################

###########################################################
# Routine : _cleanup_cores()                              #
# Description : Local subroutine to delete pre-existing   #
# cores.                                                  #
# Returns : Nothing                                       #
###########################################################
sub _cleanup_cores {
    $Log->enter() if $may_enter;
    my $obj = shift;

    my $command_interface = $obj->{command_interface};
    my $coredump_type     = "kernel";
    NACL::C::SystemNodeCoredump->delete_all(
        command_interface => $command_interface,
        node              => $command_interface->node(),
        type              => $coredump_type,
    );
    $coredump_type = "unsaved-kernel";
    NACL::C::SystemNodeCoredump->delete_all(
        command_interface => $command_interface,
        node              => $command_interface->node(),
        type              => $coredump_type,
    );
	my $find_passed;
    try {
        NACL::C::SystemNodeCoredump->find(
            command_interface => $command_interface,
            filter            => { node => $command_interface->node(), },
        );
		$find_passed = 1;
    }
    catch NACL::Exceptions::NoElementsFound with { };
	NATE::BaseException->throw( "Corefiles are not deleted successfully !! ") if ($find_passed);
	
	$Log->exit() if $may_exit;
}    # end _cleanup_cores()

###########################################################
# Routine      : cleanup_core_segment()                   #
# Description  : Subroutine to delete segment which got   #
#                created during panic.                    #
# Returns      : Nothing                                  #
# Dependencies : CMode only                               #
###########################################################
sub cleanup_core_segment {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@_,
        additional_spec => {
            'node'              => { type => SCALAR, optional => 1 },
        },
        allow_extra => 1,
    );
    my %common_opts;
    $pkg_or_obj->_move_common_component_params(
        source => \%opts,
        target => \%common_opts
    );
    my $command_interface = delete $opts{command_interface};
    $opts{'node'} = $command_interface->node()
        if ( !( defined $opts{'node'} ) );
    my $command_supports = 0;


    try {
	NACL::C::SystemNodeCoredumpSegment->delete_all(
        command_interface => $command_interface,
        node              => $command_interface->node(),
    );
    }catch NACL::C::Exceptions::SystemNodeCoredumpSegment::NotSupportedForThisPlatform with { 
        $command_supports = 1;
    };
    if (! $command_supports) {
		my $coredump_type = "kernel";
		NACL::C::SystemNodeCoredump->delete_all(
			command_interface => $command_interface,
			node              => $opts{'node'},
			type              => $coredump_type,
			%common_opts
		);
        	my $find_passed;
        	try {
            	    NACL::C::SystemNodeCoredump->find(
                        command_interface => $command_interface,
                        filter            => { node => $command_interface->node(), },
            	    );
                    $find_passed = 1;
          	}
        	catch NACL::Exceptions::NoElementsFound with { };
                if ($find_passed) {
                    require NACL::Exceptions::VerifyFailure;
                    $Log->exit() if $may_exit;
                    NACL::Exceptions::VerifyFailure->throw( "Corefiles are not deleted successfully !! ");
                }
		try {
			NACL::C::SystemNodeCoredumpSegment->find(
				command_interface => $command_interface,
				filter            => { node => $opts{'node'}, },
                                allow_empty       => 1
			);

		} 
                catch NACL::C::Exceptions::SystemNodeCoredumpSegment::NotSupportedForThisPlatform with { };
    }
    $Log->exit() if $may_exit;
}    # end cleanup_core_segment()

###########################################################
# Routine : _check_config()                               #
# Description : Local subroutine to set the appropriate   #
# options for spraycore.                                  #
# Returns : Nothing                                       #
###########################################################
sub _check_config {

    ## Grab the object handle from
    ## the input array @_
    $Log->enter() if $may_enter;

    my ( $pkg_or_obj, @args ) = @_;
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => { 'node' => { type => SCALAR, optional => 1 }, },
    );
    my $command_interface = delete $opts{command_interface};

    ## Local variables
    my $api_set_obj            = undef;
    my $buf                    = undef;
    my $cluster_ha_state       = undef;
    my $coredump_type          = undef;
    my $cluster_command_output = undef;
    my $node                   = $pkg_or_obj->_get_node();
    my $save_on_startup        = undef;
    my $storage_failover_state = undef;

    ## Get the APISet ref from the spraycore
    ## object's command interface
    $api_set_obj = $pkg_or_obj->{command_interface}->apiset(
        category  => 'Node',
        interface => 'CLI',
        set       => $command_interface->mode()
    );

    ## check whether hostname is
    ## defined in the tharnhost
    if ( !defined $node ) {
        NACL::APISet::Exceptions::MissingRequiredParamException->throw(
            'hostname is not defined in the tharnhost file of the the system');
    }
    my $Not_Cluster = 0;

    if ( $pkg_or_obj->{command_interface}->is_cmode() ) {
        try {
            NACL::CS::Cluster->fetch( command_interface => $command_interface );
        }
        catch NACL::APISet::Exceptions::CommandFailedException with {
            $Not_Cluster = 1;
        };
        ## Get the state of cluster ha
        if ( !$Not_Cluster ) {
            $Log->comment("Verifying the state of cluster ha on node $node");
            $buf = NACL::CS::ClusterHa->fetch(
                command_interface => $command_interface,
                requested_fields  => [qw( configured )]
            );
            $cluster_ha_state = $buf->configured();

            if ( $cluster_ha_state eq 'true' ) {

                $Log->comment("Cluster ha was enabled, disabling it...");
                NACL::C::ClusterHa->modify(
                    command_interface => $pkg_or_obj->{command_interface},
                    configured        => 'false'
                );

            }
            elsif ( $cluster_ha_state eq 'false' ) {
                ## If cluster ha is already disabled proceed
                $Log->comment("Cluster ha was already disabled on node $node");
            }
            else {
                NATE::BaseException->throw(
"Cluster ha is neither true or false on node $node, it is : $cluster_ha_state"
                );
            }
        }
    }
    ## Reset $buf before using it for
    ## storage failover
    my $obj = NACL::CS::StorageFailover->fetch(
        command_interface => $command_interface,
        requested_fields  => [qw(enabled)],
    );
    my $enable = $obj->enabled();
    if ( $enable eq 'true' ) {
        $Log->comment("Storage failover was enabled, disabling it...");
        NACL::C::Failover->modify(
            command_interface => $command_interface,
            node              => $command_interface->node(),
            'enabled'         => 'false',
            'auto-giveback'   => 'false',
            'onpanic'         => 'false'
        );
    }
    $Log->comment(
        "Verifying that the note has enough disk space for spraycore");

    ## Reset $buf before using it for
    ## coredump status
    $buf = undef;
    $buf = NACL::C::SystemNodeCoredump->status(
        command_interface => $command_interface,
        filter            => { node => $command_interface->node(), },
        requested_fields  => [qw(type)]
    );
    $coredump_type = $buf->{type};

    if ( $coredump_type =~ /spray/i ) {
        $Log->comment("Spraycore is supported on node $node");
    }
    else {
        NATE::BaseException->throw(
            'Coredump type for this node is not "spare" ');
    }

    try {
        NACL::C::SystemNodeCoredumpConfig->modify(
            command_interface        => $command_interface,
            node                     => $command_interface->node(),
            "save-onstartup"         => "true",
            'coredump-metadata-only' => "on",
            'coredump-attempts'      => "2",
        );
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = shift;
        if ( $exception->text() !~
            /options coredump.metadata_only on contains invalid param value/i )
        {
            $Log->exit() if $may_exit;
            $exception->throw();
        }
    };
    $Log->exit() if $may_exit;
}    # end _check_config()

###########################################################
# Routine : _get_node()                                   #
# Description : Local subroutine to get the node  that is #
# being used                                              #
# Returns : Nothing                                       #
# Dependencies : CMode only                               #
###########################################################
sub _get_node {

    ## Grab the object handle from
    ## the input array @_
    my $self = shift;

    ## Local variables
    my $node = $self->{command_interface}{hostrec}{hostname};

    return $node;

}    # end _get_node()

###########################################################
# Routine : _panic()                                      #
# Description : Local subroutine to cause a sysctl        #
# panic                                                   #
# Returns : Nothing                                       #
###########################################################
sub _panic {

    ## Grab the object handle from
    ## the input array @_
    my $self   = shift;
    my $bsdObj = undef;
    my $cmd    = "sudo sysctl debug.panic=1";
    my $node   = $self->_get_node();

    $Log->comment("Causing a sysctl panic on node $node");
    NACL::STask::Node->panic(
        command_interface => $self->{command_interface},
        nacltask_reboot   => 0
    );

    # Reboot the node after
    ## the panic
    $self->_wait_for_reboot();

    ## Wait for the savecore
    ## operation to finish
    $self->_wait_for_savecore();

}    # end _panic()

###########################################################
# Routine : _wait_for_reboot()                            #
# Description : local subroutine to reboot the node       #
# Args : NONE                                             #
# Returns : Nothing                                       #
###########################################################
sub _wait_for_reboot {

    ## Grab the object handle from
    ## the input array @_
    my $self = shift;

    ## Local variables
    my $node = $self->_get_node();
    $self->{transitObj} = NACL::Transit->new( name => $node );
    my $transitObj = $self->{transitObj};

    ## Once the firmware state is reached
    ## after the panic, reboot the node
	if ( $self->{command_interface}->is_cmode() ) {
		$Log->comment("Changing state the state of node $node to FIRMWARE...");
		$transitObj->change_state( to => 'FIRMWARE', timeout => 600 );
	}
    $Log->comment("Changing state the state of node $node to CLI...");
    $transitObj->change_state( to => 'CLI', timeout => 600 );

}    # end _reboot()

###########################################################
# Routine : _upload()                                     #
# Description : Local subroutine to upload a core file    #
# to a server.                                            #
# Returns : Nothing                                       #
# Dependencies : CMode only                               #
###########################################################
sub _upload {

    ## Grab the object handle from
    ## the input array @_
    my $self = shift;

    ## Local varibales
    #my $api_set_Obj = $self->{cli_apiset_Obj};
    my $api_set_Obj           = undef;
    my $corefile              = $self->{corefile};
    my $core_upload_location  = undef;
    my $node_name             = $self->_get_node();
    my $default_upload_server = $self->{core_upload_server};

    ## Get the APISet ref from the spraycore
    ## object's command interface
    $api_set_Obj = $self->{command_interface}->apiset(
        category  => 'Node',
        interface => 'CLI',
        set       => 'CMode'
    );

    if ( defined $self->{core_upload_location} ) {
        $core_upload_location = $self->{core_upload_location};
    }
    else {
        $core_upload_location = 'http://' . $default_upload_server;
    }

    ## Start the upload operation
    $Log->comment(
        "The upload server that will be used is $core_upload_location");
    $Log->comment("The core file that will be uploaded is $corefile");
    $api_set_Obj->system_node_coredump_upload(
        node                     => $node_name,
        location                 => $core_upload_location,
        corename                 => $corefile,
        'connectrec-timeout'     => '3600',
        'connectrec-match_table' => [
            ("Enter User:")     => "",
            ("Enter Password:") => ""
        ]
    );

}    # end _upload()

###########################################################
# Routine : _wait_for_savecore()                          #
# Description : Local subroutine to wait for the savecore #
# operation to finish.                                    #
# Returns : Nothing                                       #
# Dependencies : CMode only                               #
###########################################################
sub _wait_for_savecore {

    ## Grab the object handle from
    ## the input array @_
    my $self = shift;

    ## Local variables
    my $api_set_Obj    = undef;
    my $buf            = undef;
    my $coredump_state = undef;
    my $corefile       = undef;
    my $el             = undef;
    my $i              = 0;
    my $node           = $self->_get_node();

    ## Get the APISet ref from the spraycore
    ## object's command interface
    $api_set_Obj = $self->{command_interface}->apiset(
        category  => 'Node',
        interface => 'CLI',
        set       => $self->{command_interface}->mode()
    );

    ## Restore the APISet connection after
    ## the panic and reboot
    $api_set_Obj->api_reconnect_session()
      if !$api_set_Obj->api_is_session_alive();
	## sleep for few seconds after the node is up before checking
	## that the save has started
	Tharn::snooze(30);
    ## wait for savecore to finish
    while (1) {
        $buf = undef;
        $Log->comment("Waiting for savecore to finish");
        try {
            $buf = NACL::C::SystemNodeCoredump->status(
                command_interface => $self->{command_interface},
                filter            => { node => $node, },
                requested_fields  => [qw(corename state)]
            );

            if ( $buf->corename() ne "-" ) {
                $corefile = $buf->corename();
            }

            ## if there is no core being saved right after the
            ## coredump/reboot, throw an execption
            if ( $buf->corename() eq "-" && $i == 0 ) {
                NATE::BaseException->throw('There is no core being saved');
            }
            ## Update the counter
            $i++;
            $coredump_state = $buf->state();
        }
        catch NACL::Exceptions::NoElementsFound with {
            $coredump_state = "";
        };

        ## Keep waiting for savecore to finish
        if ( $coredump_state =~ /saving/i ) {
            $Log->comment("Core file $corefile is being saved...");
            ## Sleep for 1s before
            ## polling
            Tharn::snooze(30);
            redo;
        }
        else {
            ## break when savecore is
            ## done
            $Log->comment("Savecore finished running");
            last;
        }
    }    # end while()
	if(! $corefile) {
		NATE::BaseException->throw('There is no core being saved');
	}
    $Log->comment(
        "The last core file that was generated and saved is $corefile");
    $self->{corefile} = $corefile;

}    # end _wait_for_savecore()

# end of Spraycore.pm #
1;
