# $Id$
#
# Copyright (c) 2001-2012 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary StorageAggregateRelocation Task Module
## @author dl-nacl-dev@netapp.com
## @status Public
## @pod here

package NACL::MTask::StorageAggregateRelocation;
use base qw(NACL::STask::StorageAggregateRelocation NACL::MTask::MTask);

use strict;
use warnings;
use Params::Validate qw(validate_with :types);
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use NACL::ComponentUtils qw(Dumper _hash_copy);
use NATE::Exceptions::Argument qw(:try);
use NATE::BaseException qw(:try);

use NACL::CS::StorageAggregate;
use NACL::C::StorageFailover;
use NACL::MTask::EventLogDetector;
use NACL::C::EventLog;
use NACL::STask::Node;

use Class::MethodMaker [ scalar => 'event_object' ];

# Validate the parameters passed here
sub init {
    $Log->enter() if $may_enter;
    my ( $self, @args ) = @_;
    my %opts = validate_with(
        params => \@args,
        spec   => {
            command_interface =>
                { isa => 'NACL::C::CommandInterface::ONTAP' },
            node           => { type => SCALAR },
            destination    => { type => SCALAR, optional => 1 },
            aggregate_list => { type => ARRAYREF, optional => 1 },
            aggr_vol_list  => { type => HASHREF, optional => 1 },
            event_object => {
                type     => OBJECT,
                isa      => 'NACL::MTask::EventLogDetector',
                optional => 1
            }
        },
        allow_extra => 1,
    );
    $self->aggregate_list( @{ $opts{aggregate_list} } );
    $self->aggr_vol_list( %{ $opts{aggr_vol_list} } );
    $Log->exit() if $may_exit;
} ## end sub init

=head1 NAME

NACL::MTask::StorageAggregateRelocation

=head1 DESCRIPTION

C<NACL::MTask::StorageAggregateRelocation> provides a number of well-defined but potentially
complex or multi-step methods related to Storage Aggregate Relocation task on ONTAP.

It builds on top of, and is a derived class of C<NACL::STask::StorageAggregateRelocation>.
See C<NACL::STask::StorageAggregateRelocation> for details.
This also means that a C<NACL::MTask::StorageAggregateRelocation> object may
generally be used in place of the respective stask object .

=head1 ATTRIBUTES

=head2 command_interface

(Required) As C<NACL::C::StorageAggregateRelocation>.  A component object that
represents the host to which to send commands.

=head1 METHODS

=head2 start

  my $Arl  = NACL::MTask::StorageAggregateRelocation->start(
      command_interface => $command_interface,
      node  => $node,
      destination => $destination_node,
      'aggregate-list' => $aggr_list,
      nacltask_verify => $boolean, #default '1',
      nacltask_prechecks => $boolean , #default '1',
      %other_options
  );

(Class method) start a Aggregate Relocation Operation.  This method provides additional
services beyond what's inherent to the product's Aggregate Relocation
commands, as controlled and described by the new C<nacltask_verify>,  C<nacltask_prechecks> options.

=over

=item Options

=over

=item C<< nacltask_prechecks => 0|1 >>
(optional) If 1 (the default), verify that the all dependencies for aggregate (s) relocation
are met before issuing the start command.If 0, do not carry out this precheck.

precheck includes verifying that :

(i)   all the aggregates are of type 'SFO'
(ii)  the source and destination nodes are HA Partners
(iii) all the aggregates are owned by source node
(iv)  all the aggregates are seen from both the nodes

=item C<< nacltask_verify => 0|1 >>

(Optional) If 1 (the default), verify that the aggregate relocation has successfully
completed.  If 0, do not verify.

This option verifies the following, using
L<< NACL::STask::StorageAggregateRelocation->taskverify_relocate | lib-NACL-STask-StorageAggregateRelocation-pm/taskverify_relocate >>

(i) if the home node and current node fields of the aggregates
are updated to that of destination node.

(ii) if the aggregates and the volumes in the aggregates are all online.

Additionally it also verifies that the EMS message is generated for event "sfo.aggr.relocated.perm" ,
corresponding to all the aggregates provided in the aggregate-list.

=item C<< nacltask_reverse_controller_upgrade => 0|1 >>

(Optional) Default is 0. This option should be provided as 1 only when performing controller upgrade
 in reverse direction. ( controller from Node A to Node B is already completed, and if aggregates has to
 returned back to the source node using -nod-controller-upgrade option then nacltask_reverse_controller_upgrade
 should be 1 )

=item C<< nacltask_log_retries => $no_of_retries >>

(Optional) Default is 3. Number of attempts to verify event log messages.

=item command_interface, apiset_must, apiset_should, method-timeout, node,
destination,aggregate-list,ndo-controller-upgrade,override-destination-checks,
relocate-to-higher-version etc.,

All of the other various options to
L<< NACL::STask::StorageAggregateRelocation->start | lib-NACL-STask-StorageAggregateRelocation-pm/start >>

=back

=back

=cut

sub start {
    $Log->enter() if $may_enter;
    my ( $pkg, @args ) = @_;
    my %opts = NACL::C::Component->_common_validate_with(
        params          => \@args,
        additional_spec => { %{ $pkg->_params_validate_spec() } },
        allow_extra     => 1,
    );

    my ( %arl_opts, %common_opts, $detector, @error );

    my $controller_upgrade = $opts{'ndo-controller-upgrade'};
    my $prechecks          = delete $opts{'nacltask_prechecks'};
    my $verify             = delete $opts{'nacltask_verify'};
    my $retries            = delete $opts{'nacltask_log_retries'} || 3;

    if ($prechecks) {
        $Log->comment(
            "Verify all the pre-requisite criteria before starting ARL ");
        $pkg->_verify_arl_feasible(
            command_interface => $opts{command_interface},
            node              => $opts{node},
            destination       => $opts{destination},
            'aggregate-list'  => $opts{'aggregate-list'},
        );
    }

    if ($verify) {

        $detector = NACL::MTask::EventLogDetector->new(
            command_interface      => $opts{command_interface},
            check_for_all_presence => [
                'sfo.aggr.relocated.perm', 'raid.vol.state.online',
                'arl.OpFinished'
            ]
        );
        # Ems is different for arl with ndo-controller-upgrade
        if ( defined $controller_upgrade && $controller_upgrade eq 'true' ) {
            $detector = NACL::MTask::EventLogDetector->new(
                command_interface      => $opts{command_interface},
                check_for_all_presence => [
                    'sfo.aggr.relocated.cu', 'raid.vol.state.online',
                    'arl.OpFinished'
                ]
            );
        }
        $opts{'nacltask_wait'} = '1';
        $detector->start();
    } else {
        $opts{'nacltask_wait'} = '0';
    }

    my @mtasks = $pkg->SUPER::start(%opts);
    foreach my $task (@mtasks) {
        $task->event_object($detector);
    }
    if ($verify) {
        if ( defined $controller_upgrade && $controller_upgrade eq 'true' ) {
            $detector->stop(
                check                    => 1,
                ignore_if_unsynchronized => 1
            );
            $Log->comment("Verifying the various EMS messages now ");
            $mtasks[0]->_verify_arl_ems_messages(
                detector_object => $detector,
                arl_type        => "ControllerUpgrade"
            );
        } else {
            my $flag = 1;
            my $exception;
            while ( $retries ) {
                try {
                    $detector->stop(
                        check                    => 1,
                        ignore_if_unsynchronized => 1
                    );
                    $mtasks[0]->_verify_arl_ems_messages();
                    $flag = 0;
                } catch NATE::BaseException with {
                    $exception = shift;
                    Tharn::snooze 10;
                };
                $retries = $flag ? $retries - 1 : 0;
            }
            if ($flag) {
                $Log->exit() if $may_exit;
                $exception->throw();
            }
        }
    }
    $Log->exit() if $may_exit;
    return wantarray ? @mtasks : $mtasks[0];
} ## end sub start

sub _verify_arl_feasible {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, %opts ) = @_;
    $Log->debug( sub { " These are the opts   " . Dumper( \%opts ) } );
    my $ci             = $opts{command_interface};
    my $node           = $opts{node};
    my $destination    = $opts{destination};
    my $aggregate_list = $opts{'aggregate-list'};

    my @ineligible_aggrs;
    my $msg;
    my $partner_node_ci;

    $partner_node_ci = NACL::C::Node->new(
        command_interface => $ci,
        node              => $destination
    );

    $Log->comment(
        "Verifying  ha_policy and owner_name for all the aggregates ");

    foreach my $aggr_ha (@$aggregate_list) {
        my $state = NACL::CS::StorageAggregate->fetch(
            command_interface => $ci,
            filter            => { nodes => [$node], aggregate => $aggr_ha },
            requested_fields  => [qw(ha-policy owner-name)]
        );

        if ( $state->ha_policy() !~ /sfo/ ) {
            push @ineligible_aggrs, $state->aggregate();
        }
        if ( $state->owner_name() !~ /$node/ ) {
            push @ineligible_aggrs, $state->aggregate();
        }
    }
    if (@ineligible_aggrs) {
        $msg = join ",", @ineligible_aggrs;
        $Log->exit() if $may_exit;
        NATE::Exceptions::Argument->throw( text => "Aggregates $msg either "
                . "\n do not have ha_policy as sfo or are not owned by $node"
        );
    }

    @ineligible_aggrs = ();

    $Log->comment(
        "Verifying  that all the aggregates are visible from the partner node  "
    );

    foreach my $aggr_ha (@$aggregate_list) {

        my $state = NACL::CS::StorageAggregate->fetch(
            command_interface => $partner_node_ci,
            filter            => { nodes => [$node], aggregate => $aggr_ha },
            allow_empty       => 1
        );

        if ( !$state ) {
            push @ineligible_aggrs, $state->aggregate();
        }
    }
    if (@ineligible_aggrs) {
        $msg = join ",", @ineligible_aggrs;
        $Log->exit() if $may_exit;
        NATE::Exceptions::Argument->throw( text => "Aggregates $msg either "
                . "\n are not visible from the destination node  $destination"
        );
    }

    my $partner_check_sub = sub {
        my $this_ci          = shift;
        my $this_node        = shift;
        my $this_destination = shift;

        my $cs = NACL::CS::StorageFailover->fetch(
            command_interface => $this_ci,
            requested_fields  => [qw(partner-name enabled)],
            filter            => { node => $this_node }
        );

        my $partner = $cs->partner_name();

        if ( $partner ne $this_destination ) {
            $msg
                = " The given destination "
                . $this_destination
                . " does not seem to be the ha partner \n"
                . " of the given node "
                . $this_node
                . ".\n ARL cannot happen for this reason ";
            $Log->comment("$msg");
            $Log->exit() if $may_exit;
            NATE::Exceptions::Argument->throw( text => $msg );
        }
        if ( $cs->enabled() ne 'true' ) {

            $msg = "Storage Failover is not enabled in the node $this_node "
                . "\n SFO needs to be enabled before ARL could start ";
            $Log->comment("$msg");
            $Log->exit() if $may_exit;
            NATE::Exceptions::Argument->throw( text => $msg );
        }
    };
    $Log->comment( "Verifying that the given nodes are mutual HA "
            . "partners and that SFO is enabled in both" );
    $partner_check_sub->( $ci,              $node,        $destination );
    $partner_check_sub->( $partner_node_ci, $destination, $node );

    $Log->comment(
        "The given parameters successfully meet the ARL pre-requisite criteria"
    );
    $Log->exit() if $may_exit;
} ## end sub _verify_arl_feasible

# Helper method which verifies various ems messages corresponding to ARL one by one.
# This in turn uses _process_arl_ems_message helper routine to process each event
# and comparing it to the expected pattern of messages for this ARL operation

sub _verify_arl_ems_messages {
    $Log->enter() if $may_enter;
    my $self = shift;
    my @args = @_;
    my %opts = validate_with(
        params => \@args,
        spec   => {
            detector_object => {
                type     => OBJECT,
                isa      => 'NACL::MTask::EventLogDetector',
                optional => 1
            },
            arl_type => { type => SCALAR, optional => 1 },
        }
    );

    my $detector    = $opts{detector_object};
    my $arl_type    = $opts{arl_type};
    my $ci          = $self->command_interface();
    my $destination = $self->destination();

    if ( defined $arl_type && $arl_type =~ /ControllerUpgrade/i ) {
        $Log->comment(
            "Verifying the presence of EMS 'sfo.aggr.relocated.cu' for all aggregates"
        );
        $self->_process_arl_ems_message( event => "sfo.aggr.relocated.cu" );
        $Log->comment("Verify event messages arl.OpFinished ");
        $self->_process_arl_ems_message( event => 'arl.OpFinished' );
    } else {
        $Log->comment(
            "Verifying the presence of EMS 'sfo.aggr.relocated.perm' for all aggregates"
        );
        $self->_process_arl_ems_message( event => "sfo.aggr.relocated.perm" );

        $Log->comment(
            "Verifying the presence of EMS 'raid.vol.state.online' for all aggregates"
        );

        my $partner_ci = NACL::C::Node->new(
            command_interface => $ci,
            node              => $destination
        );

        $self->_process_arl_ems_message(
            command_interface => $partner_ci,
            node              => $destination,
            event             => "raid.vol.state.online"
        );

        $Log->comment("Verify event messages arl.OpFinished ");
        $self->_process_arl_ems_message( event => 'arl.OpFinished' );
    }

} ## end sub _verify_arl_ems_messages

# Helper method to capture a particular event type of messages using the
# EventLogDetector instance ; It verifies every message for its expected
# pattern based on the event type

sub _process_arl_ems_message {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    my %opts = validate_with(
        params => \@args,
        spec   => {
            event_object => {
                type     => OBJECT,
                isa      => "NACL::MTask::EventLogDetector",
                optional => 1
            },
            event             => { type => SCALAR },
            command_interface => {
                type     => OBJECT,
                isa      => "NACL::C::CommandInterface::ONTAP",
                optional => 1
            },
            node             => { type => SCALAR,   optional => 1 },
            'aggregate-list' => { type => ARRAYREF, optional => 1 },
            destination      => { type => SCALAR,   optional => 1 }
        }
    );

    my $ci = $opts{command_interface} || $pkg_or_obj->command_interface();
    my $node = $opts{node} || $pkg_or_obj->node();
    my $destination = $opts{destination} || $pkg_or_obj->destination();

    my @aggregate_list
        = $opts{'aggregate-list'}
        ? @{ $opts{'aggregate-list'} }
        : $pkg_or_obj->aggregate_list();

    $Log->comment( "Aggregate " . Dumper( \@aggregate_list ) );

    my $detector   = $opts{event_object} || $pkg_or_obj->event_object();
    my $begin_time = $detector->event_begin_time();
    my $end_time   = $detector->event_end_time();
    my $date_range = $begin_time . '..' . $end_time;

    my $event_type;
    $event_type = $opts{event};
    my $found = 0;
    my %filter_opts;
    $filter_opts{'time'}  = $date_range;
    $filter_opts{'node'}  = $node;
    $filter_opts{'event'} = $event_type . '*';

    # First Capture all events of type "event" mentioned in the parameter
    my @event_state_objs = NACL::CS::EventLog->fetch(
        'command_interface' => $ci,
        'filter'            => \%filter_opts,
        'requested_fields'  => [qw(event)],
    );

    $detector->events( \@event_state_objs );

    # Now verify that the expected messages are all seen
    my $events = $detector->events();

    if ( scalar @{$events} == 0 ) {
        $Log->exit() if $may_exit;
        NACL::Exceptions::EventCheckFailure->throw(
            "EMS event $event_type not found");
    }

    if ( $event_type eq 'arl.OpFinished' ) {

        my $regex
            = "Aggregate relocation operation from the source node $node "
            . "to the destination node $destination finished";

        foreach my $event ( @{$events} ) {
            my $event_name = $event->event();
            if ( $event_name =~ /.*$regex.*/ ) {
                $found = 1;
                $Log->comment( "EMS Verification passed for the message\n "
                        . " arl.OpFinished Aggregate relocation operation \n"
                        . " from the source node $node to the destination "
                        . "node $destination finished" );
                last;
            }
        }

        if ( !$found ) {
            $Log->exit() if $may_exit;
            NACL::Exceptions::EventCheckFailure->throw(
                      "EMS message Not Found -- arl.OpFinished "
                    . "Aggregate relocation operation \n "
                    . "from the source node $node to the "
                    . "destination node $destination finished " );
        }
        $Log->exit() if $may_exit;
        return;
    }

    my ( @unexpected_results, $exception_member, $regex );

    foreach my $aggr (@aggregate_list) {
        $found = 0;
        if ( $event_type eq 'sfo.aggr.relocated.perm' ) {
            $regex = "Aggregate $aggr was permanently relocated from source "
                . "node $node to destination node $destination";
        } elsif ( $event_type eq 'raid.vol.state.online' ) {
            $regex = "vol_type=\"Aggregate\", vol=\"$aggr\"";
        } elsif ( $event_type eq 'sfo.aggr.relocated.temp' ) {
            $regex = "Aggregate $aggr was temporarily relocated from source "
                . "node $node to destination node $destination";
        }

        foreach my $event ( @{$events} ) {
            my $event_name = $event->event();
            $Log->debug( sub {"Event being verified $event_name "} );
            if ( $event_name =~ /.*$regex.*/ ) {
                $found = 1;
                $Log->comment("EMS Verification passed for $aggr");
                last;
            }    #if
        }    #foreach

        if ( !$found ) {
            $exception_member->{'aggregate'} = $aggr;
            $exception_member->{'text'}
                = "EMS message $event_type not found for aggregate $aggr";
            push @unexpected_results, $exception_member;

        }    #if
    }    #foreach

    if ( scalar @unexpected_results ) {
        $Log->exit() if $may_exit;
        my $text = join "\n",
            map {"$_->{aggregate} : $_->{text}"} @unexpected_results;
        NACL::STask::Exceptions::AggregateRelocationFailure->throw(
            "EMS message not found for one or more aggregates\n$text\n",
            unexpected_results => \@unexpected_results );
    }
    $Log->exit() if $may_exit;
} ## end sub _process_arl_ems_message

=head2 create_aggr_vol_list

   $arl_task_obj->create_aggr_vol_list (
           'aggregate-list' => \@Aggregate_list,
           command_interface => $ci );

( Instance method ) This method is used to add the attribute 'aggr_vol_list'
to the object with which its called. Ideally this should be called as soon as
the object is created using new() and before invoking aggregate relocation start
(either explicitly through the command or other means like TakeOver, giveback etc.,).
This is done so that the details of vserver and state of all the volumes under the
aggregates in aggregate-list parameter could be stored in the object's attribute
'aggr_vol_list', in the following structure.

'aggr_1' => {
                 'vol_1' => {vserver=>'vs0', state => 'online'},
                 'vol_2' => {vserver=>'vs0', state => 'online'},

}
'aggr_2' => {
                 'vol_3' => {vserver=>'vs1', state => 'online'},
                 'vol_4' => {vserver=>'vs1', state => 'online'},

}

=over

=item B<< 'aggregate-list' => $aggr_list >>

(Optional)

List of aggregates which would be relocated and hence whose volume's state
information needs to be calculated and stored in 'aggr_vol_list' .

=back

=cut

sub create_aggr_vol_list {
    $Log->enter() if $may_enter;
    my ( $self, @args ) = @_;
    my %opts = NACL::C::Component->_common_validate_with(
        params          => \@args,
        additional_spec => {
            command_interface =>
                { isa => 'NACL::C::CommandInterface::ONTAP', optional => 1 },
            'aggregate-list' => { type => ARRAYREF, optional => 1 }
        }
    );
    my $ci = $opts{command_interface} || $self->command_interface();
    my $aggregate_list = $opts{'aggregate-list'} || $self->aggregate_list();
    my %volumes_aggr;

    foreach my $aggr (@$aggregate_list) {
        my @aggr_vol = NACL::CS::Volume->fetch(
            command_interface => $ci,
            filter            => { aggregate => $aggr },
            requested_fields  => [qw(volume vserver state)],
            is_system_vol    => 0,
            allow_empty       => 1
        );
        foreach my $av (@aggr_vol) {
            my $volume = $av->volume();
            $volumes_aggr{$aggr}{$volume}{'vserver'} = $av->vserver();
            $volumes_aggr{$aggr}{$volume}{'state'}   = $av->state();
        }
    }

    $self->aggr_vol_list(%volumes_aggr);
    $Log->exit() if $may_exit;
} ## end sub create_aggr_vol_list

=head2 create_event_log_detector

   $arl_task_obj->create_event_log_detector (
                  command_interface => $ci,
                  nacltask_start_detector => '1' );

( Instance method ) This method is used to add the attribute 'event_object'
to the object with which its called. Ideally this should be called as soon as
the object is created using new() and before invoking aggregate relocation start
(either explicitly through the command or other means like TakeOver, giveback etc.,).
This method creates an event log detector instance of type
NACL::MTask::EventLogDetector with the option 'check_for_all_presence'
for the following events -  'sfo.aggr.relocated.perm',
                            'raid.vol.state.online',
                            'arl.OpFinished'
This in turn is stored as an attribute 'event_object' of the calling object .

=over

=item B<< 'nacltask_start_detector' => '0/1' >>
(Optional)

Option to indicate whether the event log detector needs to be started .
Default is 1, in which case the detector is started . If Option is set to
0,  the detector is not started via this method.

=back

=cut

sub create_event_log_detector {
    $Log->enter() if $may_enter;
    my ( $self, @args ) = @_;
    my %opts = NACL::C::Component->_common_validate_with(
        params          => \@args,
        additional_spec => {
            command_interface =>
                { isa => 'NACL::C::CommandInterface::ONTAP', optional => 1 },
            nacltask_start_detector => { type => BOOLEAN, default => 1 }
        },
    );
    my $ci = $opts{command_interface} || $self->command_interface();
    my $detector
        = NACL::MTask::EventLogDetector->new( command_interface => $ci, );
    $self->event_object($detector);
    if ( $opts{nacltask_start_detector} ) {
        $detector->start();
    }
    $Log->exit() if $may_exit;
} ## end sub create_event_log_detector

=head2 perform_arl_verifications

   $arl_task_obj->perform_arl_verifications(
           "method-timeout"  =>  $timeout,
            nacltask_arl_type    => $arl_type,
            nacltask_polling_interval => $poll_delta,
           %other_options
    );

(instance method)
The method does the following ARL verifications -

(i) if the aggregates are all online after relocation to destination.

(ii) if the state of all volumes under each aggregate is same as before relocation.

(iii) if the home node and current node fields of all the aggregates are appropriate.

(iv)  Presence of EMS messages 'sfo.aggr.relocated.perm', 'raid.vol.state.online',
      'arl.OpFinished' are verified .


=over

=item Options

=over

=item B<< "method-timeout"=> $scalar >>

(Optional) Timeout to be used per aggregate relocate .
Default value is 60 seconds. If an aggregate takes more than 60 seconds
a AggregateRelocationFailure exception is thrown by the method.

=item B<< "nacltask_polling_interval"=> $scalar >>

(Optional) The polling time interval value for waiting for relocation to complete.
default 5 sec

=item B<< "nacltask_arl_type" => $scalar >>

(Optional)

This option is used to specify the type of ARL that has occured . It defaults to 'ARL'.

The valid values for this option are -

'ARL'     -  which implies that this aggregate relocate operation was explicitly started by
             'aggregate relocation start' command.
             In the verification step, the method checks that home-node = destination node and
             current-node = destination node.

'TakeOver' - which implies that the ARL was initiated by a TakeOver operation.
             In the verification step, the method checks that home-node = source node and
             current-node = destination node.

'ControllerUpgrade' - which implies that the ARL was initiated with special options
             -ndo-controller-upgrade true
             In the verification step, the method checks that home-node = source node
             and current-node = destination node.

'GiveBack' - This is a placeholder for the timebeing. Will be filled in the near future.


=item B<< "nacltask_verify_ems" => 0|1 >>

(Optional)

This option is used to specify whether ems verification needs to be performed.
This defaults to 1.

=item B<< node => $node >>

(optional)

The name of the node from which the aggregate(s) is relocated .

=item B<< destination => $destination_node >>

(optional)

The name of the destination node to which the aggregate is relocated

=item B<< 'aggregate-list' => $aggr_list >>

(Optional)

List of aggregates which will be verified for relocation

=item B<< 'aggregate-vol-details' => $aggr_vol_list >>

(Optional)

This is a hash of hash containing the details of the aggregates and
volumes belonging to each aggregate . The data is expected to be
provided in the following format.

'aggr_1' => {
                 'vol_1' => {vserver=>'vs0', state => 'online'},
                 'vol_2' => {vserver=>'vs0', state => 'online'},

}
'aggr_2' => {
                 'vol_3' => {vserver=>'vs1', state => 'online'},
                 'vol_4' => {vserver=>'vs1', state => 'online'},

}

=item B<< event_object => $ev_obj >>

(optional)

An event log detector object which will be stopped and checked for the presence
of EMS messages.

=back

=back

=cut

sub perform_arl_verifications {
    $Log->enter() if $may_enter;
    my ( $self, @args ) = @_;
    my %opts = NACL::C::Component->_common_validate_with(
        params          => \@args,
        additional_spec => {
            command_interface =>
                { isa => 'NACL::C::CommandInterface::ONTAP', optional => 1 },
            node                    => { type => SCALAR,   optional => 1 },
            destination             => { type => SCALAR,   optional => 1 },
            'aggregate-list'        => { type => ARRAYREF, optional => 1 },
            'aggregate-vol-details' => { type => HASHREF,  optional => 1 },
            event_object            => {
                type     => OBJECT,
                isa      => 'NACL::MTask::EventLogDetector',
                optional => 1
            },
            nacltask_arl_type   => { type => SCALAR,  default => 'ARL' },
            nacltask_verify_ems => { type => BOOLEAN, default => '1' }
        },
    );

    my $ci          = $opts{command_interface} || $self->command_interface();
    my $node        = $opts{node}              || $self->node();
    my $destination = $opts{destination}       || $self->destination();
    my $arl_type    = $opts{nacltask_arl_type};
    my @aggregate_list
        = $opts{'aggregate-list'}
        ? @{ $opts{'aggregate-list'} }
        : $self->aggregate_list();

    my $aggregate_vol_list = $opts{'aggregate-vol-details'}
        || $self->aggr_vol_list();
    my $ems_verify = $opts{nacltask_verify_ems};
    my $event_detector;

    if ($ems_verify) {
        $event_detector = $opts{event_object} || $self->event_object();

        if ( !$event_detector ) {
            $Log->exit() if $may_exit;
            NATE::Exceptions::Argument->throw(
                "A NACL::MTask::EventLogDetector object needs to be either "
                    . "passed in the argument options or needs to be set in "
                    . "the calling object through create_event_log_detector method"
            );
        }
    }
    my $pkg = ref $self;
    my %stask_verify_opts;
    $pkg->_hash_copy(
        source => \%opts,
        copy   => [
            qw(event_object nacltask_arl_type nacltask_polling_interval apiset_must apiset_should )
        ],
        target => \%stask_verify_opts
    );

    #node destination aggregate-list command_interface

    %stask_verify_opts = (
        %stask_verify_opts,
        command_interface       => $ci,
        node                    => $node,
        destination             => $destination,
        'aggregate-list'        => \@aggregate_list,
        'aggregate-vol-details' => $aggregate_vol_list
    );

    if ( $opts{'method-timeout'} ) {
        $stask_verify_opts{'method-timeout'} = $opts{'method-timeout'};
    }

    $Log->debug(
        sub {
            "stask_verify_opts Dumper " . Dumper( \%stask_verify_opts );
        }
    );

    $Log->comment(
        "Verify that ARL is complete, aggr/vol attributes are as expected ");
    $pkg->taskverify_relocate(%stask_verify_opts);

    if ($ems_verify) {
        $Log->comment("Stop Detector");
        if ( $arl_type =~ /ARL/i ) {
            $event_detector->stop(
                check_for_all_presence => [
                    'sfo.aggr.relocated.perm', 'raid.vol.state.online',
                    'arl.OpFinished'
                ]
            );
        } elsif ( $arl_type =~ /ControllerUpgrade/i ) {
            $event_detector->stop(
                check_for_all_presence => [
                    'sfo.aggr.relocated.cu', 'raid.vol.state.online',
                    'arl.OpFinished'
                ]
            );
        }

        $self->_verify_arl_ems_messages(
            detector_object => $event_detector,
            arl_type        => $arl_type
        );
    }
    $Log->exit() if $may_exit;
} ## end sub perform_arl_verifications

sub _params_validate_spec {
    return {
        command_interface  => { isa  => 'NACL::C::CommandInterface::ONTAP' },
        node               => { type => SCALAR },
        destination        => { type => SCALAR },
        'aggregate-list'   => { type => ARRAYREF },
        nacltask_prechecks => { type => BOOLEAN, default => 1 },
        nacltask_verify    => { type => BOOLEAN, default => 1 },
        nacltask_reverse_controller_upgrade =>
            { type => BOOLEAN, default => 0 }
    };
} ## end sub _params_validate_spec

=head2 verify_fanta_to_ems_messages

   $arl_task_obj->verify_fanta_to_ems_messages();

   OR

   $arl_task_obj->verify_fanta_to_ems_messages(
       event_object => $event_log_detector_obj
   );

(Instance method)
This method is used to verify ARL ems message generated during FANTA takeover.
Ideally this should be called once giveback is done and takenover node is up.
This method will verify temporary aggregate relocation ems(sfo.aggr.relocated.temp)
for all SFO aggregates.

    # FANTA Takeover
    my $arl_task_obj = NACL::MTask::StorageAggregateRelocation->new(
        command_interface => $node,
        node              => $partner->node(),
        destination       => $node->node(),
        ...
    );

    # Create aggr-vol list for arl verification
    $arl_task_obj->create_aggr_vol_list( 'aggregate-list' => .. );
    $arl_task_obj->create_event_log_detector();

    # Issue takeover cmd
    NACL::STask::StorageFailover->takeover(..);
    ...

    # Giveback
    NACL::STask::StorageFailover->giveback(..);

    # Verify ARL ems msg
    $arl_task_obj->verify_fanta_to_ems_messages();

=over

=item Options

=over

=item B<< event_object => $ev_obj >>

(optional)
An event log detector object which will be stopped and checked for the presence
of EMS messages.

=back

=back

=cut

sub verify_fanta_to_ems_messages {
    $Log->enter() if $may_enter;
    my ( $self, @args ) = @_;
    my %opts = validate_with(
        params => \@args,
        spec   => {
            event_object => {
                type     => OBJECT,
                isa      => 'NACL::MTask::EventLogDetector',
                optional => 1
            },
        }
    );
    my $detector = $opts{event_object} || $self->event_object();
    $detector->stop(
        check_for_all_presence   => ['sfo.aggr.relocated.temp'],
        ignore_if_unsynchronized => 1
    );

    $self->_process_arl_ems_message( event => "sfo.aggr.relocated.temp" );
    $Log->exit() if $may_exit;

} ## end sub verify_fanta_to_ems_messages

1;
