#
# Copyright (c) 2013-2017 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Traffic Task Module
## @author dl-nacl-dev@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::MTask::Traffic - To start and stop I/O traffic to or on a filer 

=head1 SYNOPSIS

  use NACL::MTask::Traffic;

  # Start the traffic
  my $path = NACL::MTask::Traffic->get_abs_path(
    $mount_stask_or_c_object, 'subdir1'
  );
  my $traffic_obj = NACL::MTask::Traffic->start(
                              command_interface => $ci,
                              paths => $path,
                              tool => $tool,
                              %apiset_options_to_be_sent_for_tool
                    );
  # Run the testcases which needs to be tested under I/O workload
  ...

  # Stop the traffic
  $traffic_obj->stop();


=head1 DESCRIPTION

C<NACL::MTask::Traffic> provides methods related to start/stop
I/O workload to a filer from any client or locally on the filer.

=head1 ATTRIBUTES

=head2 command_interface

(Required, isa 'NACL::C::Client' || isa 'NACL::C::SystemNode')
A command interface object which represents the client or filer which we
are going to generate the I/O traffic against.
See L<NACL::C::Component::command_interface|lib-NACL-C-Component-pm/command_interface>.

=head2 paths

(Optional)
An array input which takes a list of absolute directory paths or absolute
file paths or paths to raw devices on which I/O should happen.

For file system based I/O, it takes a list of paths which are present under
same/different mount points. If the underlying tool operates on a single
directory or file, then absolute path for the same should be specified.
You can provide it as simple scalar or array reference having only one value.
Whereas if the I/O tool operates on list of directories or files, then
you will have to specify the absolute paths of the same by passing an
array reference to this input.
Eg: paths => 'Y:\dir', paths => ['/mount/dir1/file1', '/mount/dir1/file2']

You can use
L<< NACL::MTask::Traffic->get_abs_path|lib-NACL-ComponentUtils-pm/get_abs_path >>
to construct the absolute path in a client agnostic way.

For performing I/O on raw disks, you should specify an array reference having
the list of I/O device paths.
Eg: paths => ['/dev/sda1', '/dev/sda2']

Note that if this input is specified, then the corresponding APISet option
to specify the directory or file path(s) is not required.
Eg: APIset defines 'dir' as the programmatic option name to specify the directory
path for running hammer tool. If 'paths' is specified as input for this task,
then you don't have to specify the 'dir' option.

=head2 _rpc_worker_object

(Optional)
This is a private attribute which this library would populate to store
L<NACL::STask::Client::RPC::Worker|lib-NACL-STask-Client-RPC-Worker-pm> object.
This represents the worker process started in the specified client to start
the I/O tool.
This should not be set/unset by the user.

=cut

package NACL::MTask::Traffic;

use strict;
use warnings;

use base qw(NACL::MTask::MTask);

use Tharn qw( sleep);
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Params::Validate qw(validate_with CODEREF OBJECT SCALAR ARRAYREF BOOLEAN);

use constant NACL_MTASK_TRAFFIC_TIMEOUT => 9**9**9;

use NATE::BaseException qw(:try);
use NACL::Exceptions::TrafficFailed;
use NATE::Exceptions::Argument;
use NATE::PerlRPC::Worker::Process;
use NACL::ComponentUtils qw(Dumper _verify_invocation get_abs_path);
use NATE::ParamSet;
use feature qw(state);
use NACL::GeneralUtils qw(
    create_file
    delete_file
    random_name_generator
);

my %Path_Map = (
    hammer => 'dir',
    sio    => 'devices_list',
    iozone => 'filename',
    dwim   => 'filename',
    random => 'scratch-dir',
    flg    => 'path',
    dt     => 'of',
    spt    => 'dsf',
    fops   => 'working-dir',
    crud   => 'filename',
);

=head1 METHODS

=head2 new

  my $traffic_obj = NACL::MTask::Traffic->new (
                            command_interface => $ci,
                            paths             => \@paths # Specify multiple paths
                    );
                    or
  my $traffic_obj = NACL::MTask::Traffic->new (
                            command_interface => $ci,
                            paths             => $path # Specify a single path
                    );

Return a new NACL::MTask::Traffic object with the given attributes.

=cut

use Class::MethodMaker [
    new    => [ '-hash', '-init', 'new' ],
    array  => 'paths',
    scalar => [
        { '-type' => 'NATE::PerlRPC::Worker::Process', },
        '_rpc_worker_object'
    ],
    scalar =>
        [ { '-type' => 'NATE::PerlRPC::Future', }, '_future' ],
    scalar => [ { '-type' => 'NACL::C::CommandInterface' }, 'command_interface' ],
    scalar => 'purge_generated_file',
    scalar => 'tool',
    scalar => 'fifo',
    scalar => 'stop_gracefully',
    scalar => 'filer_io_tool_path',
];

sub init {
    $Log->enter() if $may_enter;
    my $self = shift;
    validate_with(
        params => \@_,
        spec   => {
            paths => { type => SCALAR | ARRAYREF, optional => 1 },
            command_interface => { type => OBJECT, isa => 'NACL::C::CommandInterface' },
        },
    );

    if ( $self->paths_isset() ) {
        my @paths = $self->paths();
        if ( ref( $paths[0] ) eq 'ARRAY' ) {
            $self->paths( @{ $paths[0] } );
        }
    }
    $Log->exit() if $may_exit;
}

#Following will be used in creating runid for subtests spawned by Traffic.
#The counter will be incremented for every spawned subtest.
our $traffic_counter = '001';

=head2 start

  # Provide command_interface and the directory or file paths (Must be absolute
  # if specified) in which the I/O workload should be generated.
  my $traffic_obj =  NACL::MTask::Traffic->start(
                              command_interface => $command_interface,
                              paths => $path, # Must be absolute path
                              tool => $tool,
                              %apiset_options_to_be_sent_for_tool
                     );
                     (or)
  # Instance method call and so command_interface and paths should already
  # be present in the object as its attributes.
  $traffic_obj->start(tool => $tool, %apiset_options_to_be_sent_for_tool );

 (Class or Instance method)
 This method starts the I/O traffic to the filer.
 This is an asynchronous call that it starts the I/O utility with the specified
 options and immediately returns back NACL::MTask::Traffic object.

 Now its also possible to run traffic on filers. Filer's Systemshell will be used
 to run traffic. Since IO_TOOL's directory path will not be mounted on filer, tools
 binary needs to be copied to the filer. Option C << filer_io_tool_path >>
 can be passed to Traffic object to specify the its location on the filer.
 For more details see option C << filer_io_tool_path >> below.
 
 Also, check out the following unittest script for Filer Traffic

 NACL/UnitTest/MTask/TrafficFiler.thpl

=head3 Dependencies handled

1. time-to-run or runtime sort of paramters taken by the I/O tool need
not be specified. The task ensures to run the I/O tool till stop() is invoked.
If you still specify this parameter and if the tool stops running after
completing the specified duration and before you could invoke
$traffic_object->stop, then this would result in
NACL::Exceptions::TrafficFailed. But this default behavior can be
customized using C<< on_done >> option.

=head3 Options

=over

=item C<< "tool =>  $tool_name" >>

(Required)
Name of the I/O tool to be invoked to start the traffic.
Supported tools for isa 'NACL::C::Client' are hammer, dt, iozone, sio, flg, random, fops.
Supported tools for isa 'NACL::C::SystemNode' are hammer, dt, iozone.

=item  C<< command_interface => $ci >>

(Optional)
The component object representing the client or filer on which the traffic needs
to be started. For isa 'NACL::C::Client' this is also the client on which the 
mount_point is located. 

=item  C<< paths => \@absolute_path >> or C<< paths => $absolute_path >>

(Optional)
Refer to description of C<< paths >> under attributes section.

=item C<< %apiset_options_to_be_sent_for_tool >>

(Optional)
The input options with which the specified I/O tool should be invoked.
These option names are the programmatic option names as exposed by the CDEFs
of the APIset and the values are the corresponding values for those options.

In case if the tool is 'dt' then user can provide following option to choose from the 
various tool behavior like dt,crud,hammer,linkrename,sio :

   iobehavior=type       Specify the I/O behavior to use.
   Where type is:
     dt                The dt I/O behavior (default).
     crud              The crud I/O behavior.
     hammer            The hammer I/O behavior.
     linkrename        The linkrename I/O behavior.
     sio               The simple I/O (sio) behavior.

for eg: the call for iobehavior linkrename looks like : 
     my $traffic_obj = NACL::MTask::Traffic->start(
        command_interface => $client,
        tool => 'dt',
        iobehavior => 'linkrename',
        %options_to_be_sent_for_linkrename
     );

for more details you can use help as:
/usr/software/test/bin/dt  iobehavior=linkrename help

=item C<< on_done => $on_done_coderef >>

(Optional)
This parameter takes a code ref as the input specifying what the task should do
when the invocation of the I/O utility (which would be started asynchronously)
completes at any instance. Note that the I/O utility may complete either because
of a failure or on completing the specified work.

By default, the task throws NACL::MTask::TrafficFailed exception for both the
cases indicating that the I/O is not happening currently and it doesn't
differentiate whether the tool has completed successfully or not.
Note that this might cause failure in whatever code you
are executing in the foreground when this exception is thrown. Specifying a
code reference to this input would help you to customize this behavior.

This option is useful only when you would like to run the I/O tool indefinitely
till 'stop' method is invoked. If you are specifying,

 * 'runtime' sort of  parameters

 * Options which would convey the tool, when it should complete the I/O
   Eg: 'dt' uses an option called 'stopon' which accepts the name of the file
   as input. 'dt' watches for the existence of this file and
   stops running I/O as soon as it sees the presence of the file

then the best fit is C<< on_success >> and C<< on_failure >> parameters.

An object of type
L<NACL::STask::Client::RPC::Future|lib-NACL-STask-Client-RPC-Future-pm>
would be sent as input argument to the coderef specified.
For quick reference, refer to the section
L<"Future object, a quick overview"|lib-NACL-MTask-Traffic-pm/Quick Overview of RPC Future object>

Eg:
        my $on_done = sub {
            my $future = shift;
            # The I/O tool is not running currently
            # and I don't expect this
            Place your customized action here.
        }

        my $traffic = NACL::MTask::Traffic->start(
            command_interface => $client,  paths => $path,
            tool => 'random',
            # Specify on_done handler
            on_done => $on_done,
            %other_apiset_args
        );


Note that C<< on_done >> option is mutually exclusive to C<< on_success >>
and C<< on_failure >>.
Specifying both will result in NATE::Exceptions::Argument.

=item C<< on_success => $on_success_coderef >>

(Optional)
This parameter takes a code ref as the input specifying the callback to
be invoked when the invocation of the I/O utility started asynchronously
completes successfully.
Success criteria is generally determined based on the exit status of the tool.

An object of type
L<NACL::STask::Client::RPC::Future|lib-NACL-STask-Client-RPC-Future-pm>
would be sent as input argument to the coderef specified.
For quick reference, refer to the section
L<"Future object, a quick overview"|lib-NACL-MTask-Traffic-pm/Quick Overview of RPC Future object>

Eg:
    on_success => sub {
        my $future = shift;
        my $apiset_response_object = $future->result;
        my $raw_output = $response->get_raw_output();
        ...
        Place your customized action here.
    }

    my $traffic = NACL::MTask::Traffic->start(
        command_interface => $client,  paths => $path,
        tool => 'dt',
        # Specify on_success handler
        on_success => $on_success,
        %other_apiset_args
    );

Note that C<< on_success >> option is mutually exclusive to C<< on_done >>
and specifying both will result in NATE::Exceptions::Argument.

=item C<< on_failure => $on_failure_coderef >>

(Optional)
This parameter takes a code ref or string as the input specifying the callback to
be invoked when the invocation of the I/O utility started asynchronously
fails at any instance.
Failure criteria is generally determined based on the exit status of the tool.

The valid string values are "die", "ignore". Specifying any other strings would
result in NATE::Exceptions::Argument

=over

=item C<< die >>

NACL::MTask::TrafficFailed exception will be thrown

=item C<< ignore >>

No exception will be thrown and failure will not be reported.

=back

An object of type
L<NACL::STask::Client::RPC::Future|lib-NACL-STask-Client-RPC-Future-pm>
would be sent as input argument to the coderef specified.
For quick reference, refer to the section
L<"Future object, a quick overview"|lib-NACL-MTask-Traffic-pm/Quick Overview of RPC Future object>

Eg:
    on_failure => sub {
        my $future = shift;
        my $exception_string = $future->exception();
        ...
        Place your customized action here.
    }

    my $traffic = NACL::MTask::Traffic->start(
        command_interface => $client,  paths => $path,
        tool => 'dt',
        on_success => $on_success,
        # Specify on_failure handler
        on_failure => $on_failure,
        %other_apiset_args
    );

Note that C<< on_failure >> option is mutually exclusive to C<< on_done >>
and specifying both will result in NATE::Exceptions::Argument.

If you specify on_success option alone, then on_failure will default to 'die'.
If you specify on_failure option alone, then no defaults will be considered
for on_success and on_done.
If neither of on_success, on_failure or on_done is specified, then on_done
option will be provided its default value.

=item C<< "execute_remotely => $boolean" >>

(Optional)
Defaults to 0. This parameter determines whether the traffic task should start
the process (which runs traffic asynchronously in the background) in the test
controller (the client from which ntest is started) or in the test client (the
client from which the load has to be generated). The latter choice gives you
scale out design as the process is started in the remote client and thus takes
away from consuming the CPU and memory resources of the test controller. Note
that this approach would require your test client to be set appropriately with
STAF running and the agnostic libraries be available there. To obtain the
scaleout approach, you have to specify the value of this parameter as '1'.

On enabling this option, an additional optimization performed for non-Windows
test clients is that the I/O tool would be invoked locally in the test client
without establishing a rsh/ssh/rlogin sort of connection to itself (the test
client). Thus the valuable network resources are not consumed unnecessarily.

For isa 'NACL::C::SystemNode' this option is ignored as the traffic task is
always started on the filer.

=item C<< "stop_gracefully => $boolean" >>

(Optional)
Defaults to 0. If this option is set, then the tool will be stopped gracefully
for infinite I/O streams. As of now, this option is supported only for 'sio'
and specifying this option for other tools will make the task to complain
with NATE::BaseException.

Certain I/O tools support options such as (stopfile or termfile) through which
one can convey the tool, when it should stop running the I/O for indefinite
I/O streams (there is no definite runtime). If you enable this option in the
start() call and then later when you invoke stop(), the task will gracefully
stop the I/O tool by creating the stop file and thus signalling the I/O tool
to stop running the I/O. A graceful stop would ensure that you get the results
of the I/O tool invocation through on_* handlers. The task waits 60 sec for the
I/O tool to react for the presence of stopfile and finally kills
the process which runs the I/O tool.

If the I/O tool hadn't stopped even after specified graceful_shutdown_timeout,
then the aggressive measure of killing the process (which runs the I/O tool)
at the end would make sure to stop the I/O tool.

For tools for which this option is not supported yet, the task will stop the
invocation of the tool by killing the process.

Note:

1. The stopfile used to stop the invocation of the I/O tool will finally be
cleaned up by the task itself.
2. The task creates the stop file in the current working directory of the test
client (client indicated by command_interface) and it avoids any possible clash
with multiple independent test runs on the same test client by using the
resource name (file path or device path) in the stop file name.
The stop file name will be of the form <tool>_stopfile_<normalized_resource_name>.

Example:

    my $response;
    my $traffic = NACL::MTask::Traffic->start(
        %opts,
        # Stop gracefully
        stop_gracefully => 1,
        # show how to use on_success handler to obtain the response object
        on_success => sub {
            my $future = shift;
            $response  = $future->result();
        }
    );
    ...
    $traffic->stop;
    # At this point, $response will be populated by on_done handler.
    my $parsed_output = $response->get_parsed_output();
    ...

=item C<< runid => $runid >>

(Optional)
Runid of the subprocess which starts traffic.

=item C<< filer_io_tool_path => $filer_io_tool_path >>

(Optional)
This option is useful for running Traffic on filers. It can be used to
specify the location of io_tool's binary on the filer. If nothing is passed
then it will be defaulted to '/mroot'.

NACL_<COMMAMD>_<SET>_FULL_PATH ntest command line param can also be used
to specify the location. In this case the SET is going to be SYSTEMSHELL.
COMMAND will be io_tool's name.
e.g. if dt is in /mroot and sio is in /mroot/vol0 then you need to pass
as follows.

NACL_DT_SYSTEMSHELL_FULL_PATH=/mroo/dt
NACL_SIO_SYSTEMSHELL_FULL_PATH=/mroot/vol0/sio

If both NACL_<COMMAMD>_<SET>_FULL_PATH and filer_io_tool_path are set/passed
then filer_io_tool_path is preferred.
NACL_<COMMAMD>_<SET>_FULL_PATH is preferred over default value of filer_io_tool_path

Things to note is that value of filer_io_tool_path will not include binary name,
it will be path to the binary whereas  NACL_<COMMAND>_<SET>_FULL_PATH will be 
complete path include binary name. 
So suppose you want traffic to pick dt from /mroot then this is how you should do it.

filer_io_tool_path => '/mroot',

OR 

NACL_DT_SYSTEMSHELL_FULL_PATH => '/mroo/dt',


=back

=head2 Quick Overview of RPC Future object

The following is the list of methods from Future object commonly referred by
the users of this task. For complete description, you can refer to
L<NACL::STask::Client::RPC::Future|lib-NACL-STask-Client-RPC-Future-pm>

1. $future->exception -> Returns exception text if there is a failure, else it
returns a undef.

2. $future->result -> Returns the response object returned by APISet on
invoking the I/O utility. This will have raw output (whatever emitted by the
tool on the console), parsed_output, execution status (or return status), etc.,
Refer to APISet user guide
http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/APILayerUsersGuide/index.html#Using_the_Response_object
for more information on accessing these fields. Invoking this method in case
of failure would result in NATE::BaseException being thrown whose exception
text is same as $future->exception.

3. $future->result->get_execution_status() - This returns the exit or return
status of the tool.

=cut

sub start {

    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    $Log->debug( "Opts to 'start':\n" . Dumper( \@_ ) )
        if ( $Log->may_debug() );
    my %opts = validate_with(
        params => \@_,
        spec   => {
            tool  => { type => SCALAR },
            paths => { type => SCALAR | ARRAYREF, optional => 1 },
            command_interface => {
                type     => OBJECT,
                 isa      => 'NACL::C::CommandInterface',
                optional => 1
            },

            # deprecated
            on_exception => {
                type     => CODEREF,
                optional => 1
            },
            on_done => {
                type     => CODEREF,
                optional => 1
            },
            on_success => {
                type     => CODEREF,
                optional => 1
            },
            on_failure => {
                type      => CODEREF | SCALAR,
                optional  => 1,
                callbacks => {
                    "Value of 'on_failure' must be either coderef or "
                        . "'die' or 'ignore' " => sub {
                        unless ( ref $_[0] ) { $_[0] =~ /^(die)|(ignore)$/ }
                        }
                    }
            },

            # default is 0 - We still haven't reached the standardization
            # where everyone configures clients in a way that ntest can be
            # started from them
            execute_remotely => { type => BOOLEAN, optional => 1 },
            # Certain I/O tools like sio, dt, dwim allow us to stop infinite
            # I/O streams gracefully by supporting options such as 'stopfile'
            # But doing a graceful stop comes with additional overhead of
            # bookkeeping some variables, creating additional apiset/worker
            # for the creation of stop file. So this feature is not turned on
            # by default. Those users who need scalability (the above mentioned
            # overhead will not be tolerated by such users) and not interested in
            # capturing the tool's output when stopped gracefully, will benefit
            # from this. Further this is backward compatible.
            stop_gracefully => {type => BOOLEAN, default => 0},
            purge_generated_file => {type => SCALAR, default => 0},
            runid => {type => SCALAR, optional => 1},
            filer_io_tool_path => {type => SCALAR, optional => 1},
        },
        allow_extra => 1
    );
    
    # on_exception is the old option name for 'on_done' and both
    # cannot be specified.
    if ( defined $opts{on_exception} ) {
        if ( defined $opts{on_done} ) {
            $Log->exit() if $may_exit;
            NATE::Exceptions::Argument->throw(
                      '"on_exception" option is renamed to be "on_done"'
                    . ' and specifying both the options is invalid' );
        }
        $opts{on_done} = delete $opts{on_exception};
    }
    
    %opts = $pkg_or_obj->_handle_inputs(%opts);

    my $command_interface = delete $opts{command_interface};
    my $paths_ref         = $opts{paths};
    my $self;

    unless ( ref $pkg_or_obj ) {

        # Class method call, so create an instance
        $self = NACL::MTask::Traffic->new(
            command_interface => $command_interface, );
        $self->paths(@$paths_ref) if ($paths_ref);
    } else {

        # Instance method call. Need not create a new instance
        $self = $pkg_or_obj;
    }
    
    $pkg_or_obj->purge_generated_file(delete $opts{purge_generated_file});
    
    my $stop_gracefully   = delete $opts{stop_gracefully};
    my $execute_remotely  = delete $opts{execute_remotely};

    if ($command_interface->isa("NACL::C::SystemNode")) {
        $Log->debug( 
            "Command_Interface is NACL::C::SystemNode " .
            "so ignore stop_gracefully and execute_remotely." 
        );
        $stop_gracefully  = 0;
        $execute_remotely = 0;
    }

    $self->stop_gracefully($stop_gracefully);
    if ($stop_gracefully) {
        $self->{execute_remotely} = $execute_remotely;
    }

    # For the new design of starting the process remotely,
    # Refer http://wikid.netapp.com/w/NATE/Project/TestingAtScale#Phase_3
    # for more details.

    # For starting the process locally (default behavior):
    # Start the subtest locally and make the connection to remote
    # client using login/exec connid to start the utility. This way,
    # we don't need to have STAF running in remote client but this
    # may not provide the necessary scale-out
    my $client_for_worker = ($execute_remotely) ? $command_interface->hostrec : undef;

    my $is_windows = "";
    if($client_for_worker && $client_for_worker->isa("NACL::C::Client")) {
       $is_windows = $client_for_worker->is_windows();
    } else {
        # Is this OS windows ?
        $is_windows = $^O =~/Win/i ;
    }

    if(!$self->filer_io_tool_path_isset()) {
        my $filer_io_tool_path = delete $opts{filer_io_tool_path};
        if ($command_interface->isa("NACL::C::CommandInterface::ONTAP")) {
            my $paramset = NATE::ParamSet->new(global => 1);
            # For node command interface hardcoding the value of
            # 'set' as systemshell
            my $key = "NACL_" . uc($opts{tool}) . "_SYSTEMSHELL_FULL_PATH";
            my $value;

            if(defined $filer_io_tool_path) {
                $value = $filer_io_tool_path . '/' . $opts{tool};
            } elsif (defined $paramset->get($key)) {
                $value = $paramset->get($key);
            } else {
                $value = '/mroot/' . $opts{tool};
            }
            $self->filer_io_tool_path($value);
            $paramset->set(key => $key, value => $value);
        }
    }
    
    my $worker = NATE::PerlRPC::Worker::Process->new(
        hostrec => $client_for_worker,
        start  => 0,
    );

    $self->_process_inputs( \%opts );
    
    # Push options which are needed for graceful exit of the tool.
    if ($stop_gracefully) {
        $self->_insert_options_for_graceful_termination(opts => \%opts, tool => $opts{tool});
    }
    
    $self->_insert_options_for_interaction(
        opts => \%opts,
    );

    my @paths = $self->paths();
    my $first_path_str = "";
    if (@paths) {
        if ( ref( $paths[0] ) eq 'ARRAY' ) {
            $first_path_str = $paths[0]->[0];
        } else {
            $first_path_str = $paths[0];
        }
    }

    my $machine_name = _get_machine_name($command_interface);
    my $r_id = delete $opts{runid};
    if (!defined $r_id) {
        $r_id = $traffic_counter."_".
                   $opts{tool}."_".
                   $machine_name."_".
                   $first_path_str;

        $traffic_counter++;
        # Covert /,:,\ to _
        $r_id =~ s#/|\\|:#_#g;
    }

    #Windows file/directory name limitation is 255 characters.
    #So limiting run_id to 20 chars
    $r_id = substr($r_id,0,20) if ($is_windows);

    $worker->runid($r_id);
    
    # Start the worker.
    $worker->start();    

    # Store the value in the object if the worker process is
    # started successfully.
    $self->_rpc_worker_object($worker);

    # Process Inputs
    my $custom_on_done = delete $opts{on_done};
    my $on_success     = delete $opts{on_success};
    my $on_failure     = delete $opts{on_failure};

    my $tool = delete $opts{tool};
    $self->tool($tool);

    # Process on_success and on_failure options
    if ( defined $on_success && !( defined $on_failure ) ) {
        $Log->debug( "Defaulting on_failure handler to 'die' as "
                . "on_success option is specified" );
        $on_failure = 'die';
    }
    if ( defined $on_failure && !( ref $on_failure ) ) {
        if ( $on_failure eq 'die' ) {
            $on_failure = sub {
                my $future = shift;
                _default_on_failure_exception( $future->exception() )->throw;
            };
        } else {

            # on_failure is passed a value of 'ignore'
            $on_failure = sub {

                # No-ops
            };
        }
    }

    # construct on_done to be passed to the RPC worker call
    my $on_done = sub {

      ## It may be the case when worker object stop is being called twice
      ## In such a case worker object existence must be checked
      if($self->_rpc_worker_object_isset){
        my $future = shift;
        $self->_reset_exception_if_needed($future->exception());    
        my $exception = $future->exception();
        $Log->trace("Killing the remote worker");

        # Traffic is already stopped running. delete stopfile so that
        # it wont try to stop traffic gracefully
        delete $self->{stopfilename};
        
        # Setting this variable after stopping the load, so as to avoid any race
        # conditions with user invoked stop() in case of graceful exit.
        if ($stop_gracefully) {
            $self->{stopped} = 1;
        }
        
        if ($self->purge_generated_file) {
            $self->_purge_file_generated_by_tool();
        }
        
        $self->_remove_fifo();
        
        my $runid = $self->_rpc_worker_object->runid;
        
        my $warning = "User has passed custom on_* handler as callback. " . 
        "Subtest $runid has failed to run IO. We dont know " . 
        "if this failure is handled in callback OR not so just want to warn " . 
        "user about it. You can ignore this warning if you are intentionally " . 
        "ignoring the failure or it is handled by callback on_* handlers.";
        try {
            if ( defined $custom_on_done ) {

                # user defined callback
                $Log->comment("Invoking user defined on_done handler");
                if ($self->_is_subtest_failed($exception)) {
                    $Log->warn($warning);
                }

                $custom_on_done->($future);
            } elsif ( defined($on_success) || defined($on_failure) ) {
                if ($exception) {
                    $Log->trace( "Invoking on_failure handler, "
                            . "exception is,\n$exception" );
                    if ($self->_is_subtest_failed($exception)) {
                        $Log->warn($warning);
                    }
                    $on_failure->($future);
                } else {
                    if ( defined $on_success ) {
                        $Log->trace("Invoking user defined on_success handler");
                        $on_success->($future);
                    }
                }
            } else {

                # default callback
                $Log->comment("Invoking default on_done handler");

                if ($self->_is_subtest_failed($exception)) {
                    my $msg = "Traffic utility failed to run IO in subtest $runid and the error is:\n"
                        . $exception;
                    
                    $self->_log_failure_in_subtest($exception->text);
                    
                    if ($self->_future_isset) {
                        my $ex = $self->_future->exception();
                        $self->_reset_exception_if_needed($ex);
                    }
                    
                    $Log->exit() if $may_exit;
                    NACL::Exceptions::TrafficFailed->throw($msg);
                }
            }
        } finally {
            # IO worker object has stopped. Resetting the value.
            if($self->_rpc_worker_object_isset) {
                $self->_rpc_worker_object()->stop();
            }
            $self->_rpc_worker_object_reset();
        };
      }
    };

    # The start() should atleast be blocked till APISet object
    # is created and the command line for invoking I/O tool is
    # issued as a method call on it.

    $worker->call(
        assign => '$::Client',
        value  => $command_interface,
    );
    
    my $future = $worker->submit(
        eval => 
            'use NACL::MTask::Traffic;
            my ($execute_remotely, $tool, %opts) = @_;
            my $apiset = NACL::MTask::Traffic->_get_apiset($_[0], $::Client, $tool);
            return $apiset->$tool(%opts);',
        on_done     => $on_done,
        args        => [$execute_remotely, $tool, %opts],
        wrap_result => 1,
    );

    $self->_write_to_fifo('');
        
    # Store the future object to wait on it later.
    $self->_future($future);

    $Log->exit() if $may_exit;
    return $self;
}

=head2 stop

    $traffic_obj->stop()

 (Instance method) This method stops the I/O traffic started earlier.
 It attemps to perform a graceful stop (only for the tools for which graceful
 termination is supported) and then kill the process which runs the I/O tool.

=over

=item Options

=over

=item if_stopped

(Optional, defaults to "die")

If 'die' (default), exception is thrown if traffic is not running currently.
If 'pass', exits quietly when it detects that traffic is not running currently.

=item graceful_shutdown_timeout

(Optional, defaults to 60 seconds)

This option specifies how long the task should wait for the I/O tool to stop
gracefully. This option is applicable only if you have specified
stop_gracefully => 1 in the call to start() and graceful termination
is supported for the I/O tool.

If the I/O tool has failed to respond in the specified timeout, then
NATE::BaseException will be thrown indicating that graceful termination
could not be achieved.

=back

=back

=cut

sub stop {
    $Log->enter() if $may_enter;
    my ( $self, %opts ) = @_;
    $self->_verify_invocation( style => 'instance_only' );
    %opts = validate_with(
        params => \%opts,
        spec   => {
            'if_stopped' => $self->_if_action_completed_already_validate_spec(
                name => 'if_stopped'
            ),
            graceful_shutdown_timeout => {type => SCALAR, default => 60},
            purge_generated_file => {type => SCALAR, optional => 1},
        }
    );
    my $timeout = delete $opts{graceful_shutdown_timeout};
    unless ( $self->_rpc_worker_object_isset ) {
        if ( $opts{if_stopped} eq 'die' ) {
            $Log->exit() if $may_exit;
            NATE::BaseException->throw(
                "You have to invoke start() before invoking stop()");
        }
    } else {
        # First, try graceful exit.
        my $exception;
        if ($self->_can_be_stopped_gracefully and
           (!defined $self->{stopped} || 
           $self->{stopped} != 1)) {
            # deleting the value from $self is important to avoid any race
            # conditions
            my $file = delete $self->{stopfilename};
            my $tool = $self->tool();
            if ($tool =~ /dwim/) {
                $self->_write_to_fifo("-done");
            } else {
                create_file(
                    filepath => $file,
                    client   => $self->command_interface,
                );
            }

            # wait till the tool actually stops. on_done sets $self->{stopped}
            # indicating that the invocation of the tool has completed.
            my $wait_end = time() + $timeout;
            # Using Tharn::sleep is important here, so that NATE can invoke
            # on_done handler while this code is sleeping, otherwise, this code
            # will be waiting for on_done to complete which doesn't even get a
            # chance to execute as this code is running!
            while (!$self->{stopped} && (time() < $wait_end)) { sleep 2;}
            unless ($self->{stopped}) {
                # Even after wait time timeout, the tool has not responded and
                # stopped its execution. kill the process and thus stop
                # the tool aggressively as the next step.
                $exception = NATE::BaseException->new(
                    'The tool is still running even after creating the'.
                    ' stop file and waited for ' . $timeout . ' seconds');
            }
            delete $self->{stopped};
            # Delete the stopfile and cleanup.
            if($tool !~ /dwim/) {
                delete_file(
                    filepath => $file,
                    client   => $self->command_interface,
                );
            }
        }
        # This check is needed again to be safer with on_done handler
        # which also invokes stop()
        if ($self->_rpc_worker_object_isset) {
            # Kill the worker. This retains the old behavior as well.
            $Log->trace('stop(): Killing the worker process which runs '.
                'the I/O tool');
            $self->_rpc_worker_object()->stop();
            if ($self->_future_isset) {
                my $ex = $self->_future->exception();
                $self->_reset_exception_if_needed($ex);
            }
            
            $self->_rpc_worker_object_reset();
        }
    
        if (exists $opts{purge_generated_file} and defined $opts{purge_generated_file}) {
            $self->purge_generated_file($opts{purge_generated_file});
        }
        
        if ($self->purge_generated_file) {
            $self->_purge_file_generated_by_tool();
            # on_done handler also tries to purge this file.
            # Make sure on_done doesnt try it after its already deleted
            $self->purge_generated_file(0);
        }
        
        # Throwing the exception after attempting the aggressive stop by
        # killing the process.
        if ($exception) {
            $Log->exit() if $may_exit;
            $exception->throw;
        }
    }
    $Log->exit() if $may_exit;
}

sub set_dynamically {
    $Log->enter() if $may_enter;
    my ($self, %opts) = @_;
    my $tool = $self->tool();
    if($tool =~ /dwim/) {
        $self->_set_dynamically_dwim(%opts);
    } else {
        $Log->warn("Currently changing value of a option dynamically " .
                   "is only supported for dwim IO tool");
    }
    
    $Log->exit() if $may_exit;
}
        
sub purge {
    $Log->enter() if $may_enter;
    my ( $self, %opts ) = @_;

    if (exists $opts{nacltask_if_purged}) {
        $opts{if_stopped} = delete $opts{nacltask_if_purged};
    }
    
    $self->stop(%opts);
    
    $Log->exit() if $may_exit;
}

=head2 wait_for_completion

    Example:1 (Generic)
    $traffic_obj->start(%args, %stop_criteria);
    ... Do something ...
    my $response = $traffic_obj->wait_for_completion();

    Example:2 (Specific)
    $traffic_obj->start(
        tool => 'sio',
        read => '0', rand => '100', block_size => '4k', start_byte => 0, 'file_size' => '100m', num_of_threads => 3,
        duration_in_secs => 60 # This is the stop criteria
    );
    ... Do something ...

    # If the tool has completed successfully, it returns APISet response object,
    # whereas if the tool has failed, it will result in
    # NACL::Exceptions::TrafficFailed
    my $response = $traffic_obj->wait_for_completion();


(Instance method) This method waits in the foreground for the I/O tool
to finish the execution(i.e., "blocking wait")  and returns back as
soon as the prompt is seen.

If the execution was successful,
it will return back APISet response object which holds the raw output,
parsed output (not supported for all the I/O tools yet), command executed,
execution status, execution time, etc., Refer to
L<APISet user guide|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/APILayerUsersGuide/index.html#Using_the_Response_object>
for more information on accessing these fields.

In the call to start(), if you have specified options like on_done,
on_success, on_failure, then those callbacks will also get executed
before this method returns.

=head3 Points to note:

=over

=item stop() should not be called with this method

stop() should not be called before invoking this method
as the I/O tool would have completed the execution by the time,
this method returns back.
stop() itself is being called at the end of this mehtod.
So no explicitly call for stop() is required if this method is invoked.

=item specific stop criteria must be used while using this method

If you haven't specified options like time duration or options that would
stop the tool by itself on meeting some specific stop criteria, then this
method shouldn't be used, as it will result in "infinite wait".

=back

=head3 More details for advanced users

If you are an advanced user who overrides C<on_success> or C<on_failure> or
C<on_done> handlers in the call to start() and using this method, then the
following section would matter for you.

If you override C<on_done> or C<on_success> handler to throw exception
and if the tool has completed successfully, then invoking wait_for_completion
after the actual completion of tool would return back response object, whereas
invoking it before the completion of the tool would result in the exception
obtained from C<on_success> or C<on_done> callback.

=over

=item Exceptions

=over

=item NACL::Exceptions::TrafficFailed

=over

=item Tool failed

Tool failed, user hasn't overridden on_exception or on_failure handlers.

=item Tool failed

Tool failed, user has overridden on_failure handler to its default value
'die'.

=back

=item NATE::BaseException

=over

=item Tool failed

Tool failed, user has overridden on_failure or on_done handler not to
throw exception.

=item Tool failed

Tool failed, user has overridden on_failure or on_done handler to throw some
other exception and this method was invoked after the completion of the tool.

=back

Ideally, we should have thrown NACL::Exceptions::TrafficFailed, this support
may be added later.

=item Other Exceptions decided by the caller.

=over

=item Tool failed

Tool failed, user has overridden either of on_done or on_failure
handler to throw exception and this method was invoked
before the completion of I/O tool.

=item Tool passed

Tool passed, user has overridden on_done or on_success handler
to throw exception and this method was invoked before the completion of I/O
tool.

=back

=back

=back

=cut

sub wait_for_completion {
    $Log->enter() if $may_enter;
    my ( $self, %opts ) = @_;
    $self->_verify_invocation( style => 'instance_only' );
    if (! $self->is_io_tool_running())
    {
        $Log->debug("Info : The tool is not running and worker has been stopped");
        $Log->exit() if $may_exit;
        return;
    }

    unless ( $self->_future_isset ) {
        $Log->exit() if $may_exit;
        NATE::BaseException->throw(
                  "Couldn't find response object to obtain the "
                . "completion status." );
    }
    my $future = $self->_future;
    my $response;
    try {

        # This is a blocking wait on the traffic call to complete
        $response = $future->result();
    }
    catch NACL::Exceptions::TrafficFailed with {

        # This exception will be obtained for default handlers
        # of on_done or on_failure
        my $ex = shift;
        $Log->debug( 'Future->result declares error, but we '
                . 'still need to check for response object' );
        $response = $ex->_response;

        # When the tool has completed successfully (because of reaching the
        # specified time duration ,etc.,), then $traffic_object->start
        # throws exception from on_done (default behavior, if user hasn't
        # overridden) indicating that the traffic is not running, but
        # for wait_for_completion, we are interested in propagating the
        # exception only if its an actual tool failure. This is because
        # the callers of this method are those who specify tool options
        # which stops the tool on reaching specific condition.
        unless ($response) {
            $Log->debug('Failure in I/O tool completion');

            # If response object is not obtained, then there is some failure
            # in the invocation of the tool. So rethrow the exception
            $ex->throw;
        }
    }
    finally {
        $self->stop() if ( $self->is_io_tool_running );
    };

    $Log->exit() if $may_exit;
    return $response;
}

=head2 is_io_tool_running

    my $is_running = $traffic_obj->is_io_tool_running();

(Instance method) This method returns back 1/0 based on whether the
I/O tool is currently running or not.

=cut

sub is_io_tool_running {
    $Log->enter() if $may_enter;
    my ( $self, %opts ) = @_;
    $self->_verify_invocation( style => 'instance_only' );
    my $is_running = ( $self->_rpc_worker_object_isset ) ? 1 : 0;
    $Log->exit() if $may_exit;
    return $is_running;
}

####################################
#       Private Methods            #
####################################

sub _process_inputs {
    $Log->enter() if $may_enter;
    my ( $self, $args_ptr ) = @_;
    my $traffic_path;
    my $paths_ref = delete $args_ptr->{paths};
    if ( defined $paths_ref ) {
        $traffic_path = join( ' ', @{$paths_ref} );
    }
    
    # if 'iobehavior => <tool>' option is passed it means that tool passed
    # to iobehavior option will be used to generate traffic.
    # dt will be using tool's parser to parse the options.
    # We also need to use names for 'path' and 'runtime' like
    # options that tool's parser understands
    my $tool = $args_ptr->{iobehavior} // $args_ptr->{tool};

    
    # Specify the work dir or file path for the I/O tool
    if ( defined $Path_Map{$tool}
        && !defined $args_ptr->{ $Path_Map{$tool} } )
    {
        $args_ptr->{ $Path_Map{$tool} } = $traffic_path;
    }

    if ( ( $tool eq 'dt' ) && ( defined $paths_ref ) ) {
        unless ( defined( $args_ptr->{'if'} )
            || defined( $args_ptr->{'of'} ) )
        {
            $args_ptr->{'of'} = join( ',', @{$paths_ref} );
        }
    }

    $tool = $args_ptr->{tool};
    # Set the timeout values for the I/O tools which takes time-to-run
    # the I/O tool as mandatory input.
    # hammer, random, iozone, fops don't expect this param
    # flg has its own defaults.
    my %timeout_map = ( 
        sio => 'duration_in_secs',
    );

    if ( defined $timeout_map{$tool}
        && !defined $args_ptr->{ $timeout_map{$tool} } )
    {
        $args_ptr->{ $timeout_map{$tool} } = NACL_MTASK_TRAFFIC_TIMEOUT;
    }

    # Set the timeout if not provided by the user so that APISet will not fail
    # saying TimeoutException
    my $timeout = delete $args_ptr->{timeout}
        || delete $args_ptr->{'connectrec-timeout'};
    unless ( defined $timeout ) {

        # Set the timeout to -1, indicating no time limit
        # NACL_MTASK_TRAFFIC_TIMEOUT gets evaluated to zero in certain Windows
        # clients, may be its running out of bit range. So we use the NATE
        # support here.
        $args_ptr->{'connectrec-timeout'} = -1;
    }

    $Log->exit() if $may_exit;
}

sub _handle_inputs {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, %opts ) = @_;

    if ( ref $pkg_or_obj ) {

        # Instance method call. We should get command_interface and paths
        # from the object itself.
        my $error_keys;
        if ( defined $opts{command_interface} ) {
            $Log->exit() if $may_exit;
            NATE::Exceptions::Argument->throw(
                "'command_interface' is already an attribute of this object "
                    . "and must not be specified to NACL::MTask::Traffic::start"
            );
        }
        if ( $pkg_or_obj->_rpc_worker_object_isset ) {
            my $runid = $pkg_or_obj->_rpc_worker_object->runid;
            $Log->exit() if $may_exit;
            NACL::Exceptions::TrafficFailed->throw(
                "IO from previous iteration is still running as part of subtest $runid. Stop it first and then try again"
            );
        }
        foreach my $key (qw(command_interface paths)) {

            # Populate the input hash
            my $key_isset = "${key}_isset";
            if ( $pkg_or_obj->$key_isset() ) {
                $opts{$key} = $pkg_or_obj->$key();
            }
        }
    } else {

        # static method call
        unless ( defined $opts{command_interface} ) {
            $Log->exit() if $may_exit;
            NATE::Exceptions::Argument->throw(
                "Mandatory parameter 'command_interface' missing in call "
                    . "to NACL::MTask::Traffic::start" );
        }
    }

 # check if valid combination of on_done, on_success, on_failure is specified.
    if ( $opts{on_done} && ( $opts{on_success} || $opts{on_failure} ) ) {
        $Log->exit() if $may_exit;
        NATE::Exceptions::Argument->throw( 'on_done option is mutually '
                . 'exclusive with on_success and on_failure' );
    }

    # Handles if scalar value is stored in the 'paths' input for
    # package method calls and instance method calls where the user
    # has specified this input in the call to start() rather than in new()
    if ( defined $opts{paths} && ( !ref( $opts{paths} ) ) ) {
        $opts{paths} = [ $opts{paths} ];
    }

    $Log->exit() if $may_exit;
    return %opts;
}

# This is to be used only by start() method to obtain
# an APISet object.
sub _get_apiset {
    my ( $pkg, $do_scale, $ci, $tool ) = @_;
    my %apiset_opts;
    my $apiset;
    
    $apiset_opts{interface} = 'CLI';
    if ( $do_scale && ($ci->isa("NACL::C::CommandInterface::ONTAP") || !$ci->is_windows()) ) {
        my $conn = $ci->hostrec()->connect(
            template => "local_persist",
            conntype => "rlogin",
            conncmd  => [ "/bin/sh", "-i" ]
        );
        $apiset_opts{connection} = $conn;
    }
    if ($ci->isa("NACL::C::CommandInterface::ONTAP")) {
        $apiset_opts{set} = "Systemshell";
    }
    $apiset = $ci->apiset(%apiset_opts);

    if ( ($tool =~ /spt/ ) ||
         ($ci->isa("NACL::C::CommandInterface::ONTAP")) ) {
        #spt and on-filer tools requires root privilege
        $apiset->set_privilege('privilege-level' => "root");
    }
    return $apiset;
}

sub _default_on_failure_exception {
    $Log->enter() if $may_enter;
    my $exception = shift;
    my $msg
        = 'The I/O tool has reported failure and is no longer running.'
        . "The reason is,\n"
        . $exception;
    my $obj = NACL::Exceptions::TrafficFailed->new($msg);
    $Log->exit() if $may_exit;
    return $obj;
}

sub _insert_options_for_graceful_termination {
    $Log->enter() if $may_enter;
    my ($self, %args) = @_;
    my $opts_ref = $args{opts};
    my $tool = $args{tool};
    my $file;
    my %stop_file = (
        'sio' => 'stop',
        'dt'  => 'stopon',
    );

    if ($self->_is_graceful_termination_supported($tool)) {
        unless ($tool =~ /dwim/ or exists $opts_ref->{$stop_file{$tool}}) {
            $file = $self->_get_unique_name_from_resource_list($opts_ref->{ $Path_Map{$tool} }, $tool);
            ##Fix for burt725044
            ##Get the logdirectory
            my $logpath = $self->_get_logpath();
            if ( defined $logpath ) {
                if ( $self->command_interface->is_windows ) {
                    $file = $logpath . "\\$file";
                } else {
                    $file = $logpath . "/$file";
                }
            }
            $Log->debug( sub {"This is the log path $logpath"} );
            ## Stop file should be created in the top logdirectory path itself
            $Log->debug( sub {"This is the file $file"} );

            # sio expects a hyphen to be the first character in the filename.
            if ($tool eq "sio") {
                $opts_ref->{$stop_file{$tool}} = "-$file";
            } else {
                $opts_ref->{$stop_file{$tool}} = "$file";
            }
        }
    } else {
        $Log->exit() if $may_exit;
        NATE::BaseException->throw(
            'Graceful termination (stop_gracefully => 1) is supported only '.
            'for sio and dt as of now.'
        );
    }
    if (defined $file) {
        $self->{stopfilename} = $file;
    }
    $Log->exit() if $may_exit;
}

sub _get_logpath {
    my ($self) = @_;
    my $logpath;
    my $client      = $self->command_interface();
    my $client_type = $client->hostrec()->hosttype();
    if ( $client_type =~ /unix/ ) {
        $logpath = NATE::ParamSet->param_global->get('TOP_LOGDIR');
    } elsif ( $client_type =~ /windows/ ) {
        $logpath = NATE::ParamSet->param_global->get('TOP_LOGDIR_WINDOWS');
    }

    return $logpath;
}

sub _get_unique_name_from_resource_list {
    $Log->enter() if $may_enter;
    my ($self, $devices_list, $tool) = @_;
    # If there are multiple devices (or files/directories), select the first one
    my @list;
    if($tool eq 'sio'){
	@list = split(/\s* \s*/ , $devices_list);
    } else{
	@list = split(/\s*,\s*/ , $devices_list);
    }
    $devices_list = $list[0];
    # Convert the path or device name into a valid file name.
    $devices_list =~ s/^([a-zA-Z]):/${1}_drive/;
    $devices_list =~ s/[\\\/]/_/g;
    # To avoid any possible clash in stop file name when multiple
    # independent tests run on the same test client, the devices_list
    # name is added to the stop file name. The uniqueness comes from the
    # fact that same resource (file or device or dir) cannot be shared
    # by more than one test simultaneously.
    my $file = "${tool}_stopfile_${devices_list}";
    $Log->exit() if $may_exit;
    return $file;
}

sub _log_failure_in_subtest {
    $Log->enter() if $may_enter;
    my ($self, $message) = @_;
    $self->_rpc_worker_object()->call(
        eval => 'my ($msg) = @_;
        NATE::Result::Fail->new($msg)->log();',
        args => [$message],
    );
    
    $Log->exit() if $may_exit;
}

sub _is_graceful_termination_supported {
    my ($self, $tool) = @_;

    my %graceful_term_hash = (
        sio => 1,
        dt  => 1,
        dwim => 1,
    );
    
    return (exists $graceful_term_hash{$tool});
}

# An exception is raised even if user has explicitly called $worker->stop
# Dont want to fail in this case so resetting the exception
sub _reset_exception_if_needed {
    my ($self, $ex) = @_;
    if (! $self->_is_subtest_failed($ex)) {
        $self->_future->_exception_reset();
    }
}

sub _is_subtest_failed {
    my ($self, $ex) = @_;
    my $failed = 0;
    if (defined $ex and $ex->text !~ /Worker->stop was called/) {
        $failed = 1; 
    }
    
    return $failed;
}

sub _purge_file_generated_by_tool {
    $Log->enter() if $may_enter;
    my ($self) = @_;
    
    my $ci = $self->command_interface;
    my @paths = $self->paths();
    foreach my $file_or_dir (@paths) {
        $ci->delete_file_or_directory(
            filepath => $file_or_dir,
        );
    }
    
    $Log->exit() if $may_exit;
}

sub _set_dynamically_dwim {
    $Log->enter() if $may_enter;
    my ($self, @args) = @_;
    # Only 'throughput' can be changed dynamically
    state $spec = {
        throughput => { type => SCALAR },
    };
    my %opts = validate_with(
        params => \@args,
        spec   => $spec,
    );
    my $ci = $self->command_interface();
    my $apiset = $ci->apiset(interface => 'CLI');
    my $fifo = $self->fifo;
    my $output = $apiset->xlate_dwim(%opts);
    my @cmd_output = split(/\s+/, $output, 2);
    $output = $cmd_output[1];
    $apiset->execute_command(command => "echo $output > $fifo");
    $Log->exit() if $may_exit;
}

sub _insert_options_for_interaction {
    $Log->enter() if $may_enter;
    my ($self, %args) = @_;
    my $opts_ref = $args{opts};
    my $tool = $opts_ref->{tool};
    if ($tool =~ /dwim/) {
        my $apiset = $self->command_interface->apiset();
        my $fifo = random_name_generator(suffix => "fifo");
        my $logpath = $self->_get_logpath();
        if ( defined $logpath ) {
            if ( $self->command_interface->is_windows ) {
                $fifo = $logpath . "\\$fifo";
            } else {
                $fifo = $logpath . "/$fifo";
            }
        }
        my $out = $apiset->execute_command(
            command => "mkfifo $fifo",
        );

        $opts_ref->{fifo} = $fifo;
        $self->fifo($fifo);
    }

    $Log->exit() if $may_exit;
}

sub _write_to_fifo {
    $Log->enter() if $may_enter;
    my ($self, $content) = @_;
    # only dwim tool supports intraction
    if ($self->tool =~ /dwim/) {
        my $ci = $self->command_interface();
        my $apiset = $ci->apiset(interface => 'CLI');
        my $fifo = $self->fifo;

        $apiset->execute_command(command => "echo $content > $fifo");
    }

    $Log->exit() if $may_exit;
}

sub _remove_fifo {
    $Log->enter() if $may_enter;
    my ($self, $content) = @_;
    # fifo will only exists if tool is dwim
    if ($self->fifo_isset() and $self->tool =~ /dwim/) {
        my $ci = $self->command_interface();
        my $apiset = $ci->apiset(interface => 'CLI');
        my $fifo = $self->fifo;

        $apiset->execute_command(command => "rm $fifo");
    }

    $Log->exit() if $may_exit;
}

sub _can_be_stopped_gracefully {
    my ($self) = @_;
    my $tool = $self->tool();
    return ($self->stop_gracefully and $self->_is_graceful_termination_supported($tool));
}

sub _get_machine_name {
    my ($self) = @_;
    my $ci = $self->command_interface();
    my $name;
    if ($ci->isa("NACL::C::CommandInterface::ONTAP")) {
        $name = $ci->node();
    } else {
        $name = $ci->client();
    }
    
    return $name;
}

1;
