# $Id:
# Copyright (c) 2011 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary NACL Task for ttcp utility on filer and client
## @author  kathar.hidayath@netapp.com, dl-nacl-dev@netapp.com
## @status public
## @pod here

=head1 NAME

    NACL::MTask::Ttcp

=head1 DESCRIPTION

    "NACL::MTask::Ttcp" provides a number of well-defined methods
    related to TTCP functionality on ONTAP as well as Host systems.

    The MTask will mainly be concerned with ttcp transmitter and receiver.
    ttcp transmitter and receiver can be start on both filer and linux client.

=head1 ATTRIBUTES

=head2 command_interface

    (Required) Represents the filer/host command interface on which the ttcp utility should be run .

=cut

package NACL::MTask::Ttcp;

use strict;
use warnings;
use base qw/NACL::MTask::MTask/;

use Tharn qw( connect host subtest);
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use NACL::ComponentUtils qw(Dumper);
use NATE::Exceptions::Argument qw(:try);
use Params::Validate qw(validate_with SCALAR SCALARREF BOOLEAN OBJECT);
use NACL::APISet::Exceptions::CommandFailedException;

use Class::MethodMaker [
    new => [ '-hash', 'new' ],
    scalar => [
        { -type => 'NACL::C::CommandInterface' },
        'command_interface',
    ],
];
=head1 METHODS

=head2 is_installed

   NACL::MTask::Ttcp->is_installed(
             'command_interface'              => $Node,
             );

(Class Method) Based on the command interface, this method checks whether the
ttcp functionality is installed on a Filer or Client . This method supports linux
client & Filer (7mode/Nodescope).
( Returns 1 if the ttcp command is found on the host/filer specified through
the connection , else it returns zero.)

=over

=item  Options

=over

=item C<< "command_interface=>$Node" >>

(Required) : Contains the corresponding(client/filer) command interface on which the ttcp functionality needs to be checked .

=back

=back

=cut

sub is_installed() {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = validate_with(
        params => \@_,
        spec   => {
            'command_interface' => { type => OBJECT },
        },
        allow_extra => 1,
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset;
    my $verify = 0;

    $apiset = _get_apiset($command_interface);
    try {
        $apiset->ttcp();
    } catch NACL::APISet::Exceptions::ResponseException with {
        my $e     = shift;
        # If ttcp is there then we expect to see a usage message
        if ( $e->text =~ /^usage/i ) {
            $verify = 1;
        } else {
            $verify = 0;
        }
    };
    $Log->exit() if $may_exit;
    return $verify;
}

=head2 start_receiver

   NACL::MTask::Ttcp->start_receiver(
              'command_interface'        => $Node,
              nacltask_bind_retry        => $bind_time ,
              length-of-bufs-read        => $size,
              port                       => $port,
              address-family             => $af,
              nacltask_protocol          => $protocol,
              nacltask_method_timeout    => $timeout,
              nacltask_run_in_background => $default,
             );

(Class method) Starts ttcp in receive mode. If the command interface is either a node/host ,
               the ttcp functionality will start in the corresponding object. This method supports linux
               client & Filer (7mode/Nodescope).Returns a hash reference containing the parsed output
               returned after executing the command.

=over

=item Options

=over

=item C<< "command_interface=>$Node" >>

(Required) : Contains the host/node command interface on which the ttcp receiver should be started .

=item C<< "nacltask_bind_retry=>$bind_time" >>

(Optional) : Number of times to retry the ttcp command if it
returns "bind: Address already in use" error. Default : 5.

=item C<< "length-of-bufs-read=>$size" >>

(Optional) : Length of buffers to use. Default : 1024.

=item C<< "port=>$port" >>

(Optional) : Port number to listen on. Default : 5001.

=item C<< "number-of-source-bufs=>$source_bufs" >>

(Optional) : Number of source buffers transmitted (default 2048).

=item C<< "host=>$host_ip" >>

(Optional) : In trasmit mode, ip address where packets need to transmit

=item C<< "source=>$boolean" >>

(Optional) : If transmitting, source a data pattern to network; if receiving,
sink (discard) the data. the default is to transmit data from stdin or print
the received data to stdout.

=item C<< "nacltask_run_in_background=> $default" >>

(Optional) : This option is used to run the start_receiver method in background .
Default : 1

=back

=item For other options which is used in receiver mode , check the respective host object's ttcp.cdefs.

=back

=cut

sub start_receiver() {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = $pkg->_common_validate_with(
        params          => \@_,
        additional_spec => {
            nacltask_bind_retry => { type => SCALAR, default => '5',
                      callbacks => {
                      "Value of 'bind_entry' should be non zero" => sub {
                                                $_[0] > 0 }
                        }
                    },
            nacltask_run_in_background => { type => BOOLEAN, default => 1 },
            'sink'                     => { type => SCALAR, default => 1 },
            'length-of-bufs-read'      => { type => SCALAR, default => 1024 },
            'port'                     => { type => SCALAR, default => 5001 },
        },
        allow_extra => 1,
    );

    my $command_interface = delete $opts{command_interface};
    my $nacltask_bind_retry = delete $opts{nacltask_bind_retry};
    my $nacltask_run_in_background = delete $opts{nacltask_run_in_background};
    delete $opts{apiset_must};
    delete $opts{apiset_should};
    my $apiset;
    my $result;
    if($command_interface->hostrec->hosttype() =~ /linux/i) {
        if(($opts{'address-family'}) && ($opts{'address-family'} =~ /inet6/)) {
            $opts{'ipv6'} = 1;
        } else {
            $opts{'ipv4'} = 1;
        }
        delete $opts{'address-family'};
    }
    $opts{'receive-mode'} = '1';
    $apiset = _get_apiset($command_interface);

AGAIN: {
        use warnings qw(exiting);
        try {
            if ($nacltask_run_in_background) {
                $result = _background_run($apiset, %opts);
            } else {
                $result = $apiset->ttcp(%opts);
                $result = $result->get_parsed_output();
            }
        } catch NACL::APISet::Exceptions::ResponseException with {
            my $e = shift;
            if ( $e->text() =~ /bind/i ) {
                if($nacltask_bind_retry) {
                    sleep(15);
                    $nacltask_bind_retry--;
                    no warnings qw(exiting);
                    redo AGAIN;
                } else {
                    $Log->exit() if $may_exit;
                    $e->throw();
                }
            } else {
                $Log->exit() if $may_exit;
                $e->throw();
            }
        };
    }
    $Log->exit() if $may_exit;
    return $result;
}

=head2 start_transmitter

   NACL::MTask::Ttcp->start_transmitter(
             'command_interface'        => $Node,
              nacltask_bind_retry       => $bind_time ,
              length-of-bufs-read       => $size,
              port                      => $port,
              address-family            => $af,
              protocol                  => $protocol,
              output-buffer-size        => $outputsize,
              nacltask_method_timeout   => $timeout,
             );

(Class method) Starts ttcp in transmit mode. If the command interface is either a node/host ,
the ttcp functionality will start in the corresponding object. This method supports linux
client & Filer (7mode/Nodescope).

Returns a hash reference containing the parsed output returned after executing the command.

=over

=item Options

=over

=item C<< "command_interface=>$Node" >>

(Required) : Contains the host/node command interface on which the ttcp receiver should be started .

=item C<< "nacltask_bind_retry=>$bind_time" >>

(Optional) : Number of times to retry the ttcp command if it returns "bind: Address already in use" error. Default : 5.

=item C<< "length-of-bufs-read=>$size" >>

(Optional) : Length of buffers to use. Default : 1024.

=item C<< "port=>$port" >>

(Optional) : Port number to listen on. Default : 5001.

=item C<< "address-family=>$af" >>

(Optional) : address family(inet or inet6). Default : "inet".

=item C<< "protocol=>$protocol" >>

(Optional) : Protocol to use ('tcp' or 'udp'). Default : 'tcp'.

=item C<< "number-of-source-bufs=>$source_bufs" >>

(Optional) : Number of source buffers transmitted (default 2048).

=item C<< "host=>$host_ip" >>

(Optional) : In trasmit mode, ip address where packets need to transmit

=item C<< "address-family=>(inet/inet6)" >>

(Optional) : In trasmit mode, address family to be specified.(default ipv4)

=item C<< "source=>$boolean" >>

(Optional) : If transmitting, source a data pattern to network; if receiving,
sink (discard) the data. the default is to transmit data from stdin or print
the received data to stdout.

=item C<< "output-buffer-size=> $outputsize" >>

(Optional) : Output buffer size in bytes. This option is only supported on a filer.

=item C<< "nacltask_run_in_background=> $default" >>

(Optional) : This option is used to run the start_receiver method in background .
Default : 0

=back

=item For other options which is used in transmitter mode , check the respective host object's ttcp.cdefs .

=back

=cut

sub start_transmitter() {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = $pkg->_common_validate_with(
        params          => \@_,
        additional_spec => {
            nacltask_bind_retry => { type => SCALAR, default => '5',
                      callbacks => {
                      "Value of 'bind_entry' should be non zero" => sub {
                                                $_[0] > 0 }
                        }
                    },
            nacltask_run_in_background => { type => BOOLEAN, default => 0 },
            'source'                   => { type => SCALAR, default => 1 },
            'address-family'           => { type => SCALAR, default => 'inet' },
            'length-of-bufs-read'      => { type => SCALAR, default => 1024 },
            'port'                     => { type => SCALAR, default => 5001 },
            'number-of-source-bufs'    => { type => SCALAR, default => 2048 },
        },
        allow_extra => 1,
    );

    $opts{'transmit-mode'} = 1;

    my $command_interface          = delete $opts{command_interface};
    my $nacltask_bind_retry        = delete $opts{nacltask_bind_retry};
    my $nacltask_run_in_background = delete $opts{nacltask_run_in_background};
    delete $opts{apiset_must};
    delete $opts{apiset_should};
    my $apiset;
    my $result;
    if($command_interface->hostrec->hosttype() =~ /linux/i) {
        if($opts{'address-family'} =~ /inet6/) {
            $opts{'ipv6'} = 1;
        } else {
            $opts{'ipv4'} = 1;
        }
        delete $opts{'address-family'};
    }
    $apiset = _get_apiset($command_interface);
REDO: {
        use warnings qw(exiting);
        try {
            if ($nacltask_run_in_background) {
                $result = _background_run($apiset, %opts);
            } else {
                $result = $apiset->ttcp(%opts);
                $result = $result->get_parsed_output();
            }
        } catch NACL::APISet::Exceptions::ResponseException with {
            my $e = shift;
            if ( $e->text() =~ /bind/i ) {
                if($nacltask_bind_retry) {
                    sleep(15);
                    $nacltask_bind_retry--;
                    no warnings qw(exiting);
                    redo REDO;
                } else {
                    $Log->exit() if $may_exit;
                    $e->throw();
                }
            } else {
                $Log->exit() if $may_exit;
                $e->throw();
            }
        };
    }
    $Log->exit() if $may_exit;
    return $result;
}

#===============================================================================
#                              Helper Methods
#===============================================================================

sub _get_apiset {
    my ($command_interface) = @_;
    my $apiset = undef;
    my $hosttype = $command_interface->hostrec->hosttype();
    if ( $hosttype =~ /filer/i ) {
        $apiset = $command_interface->get_7m_or_nodescope_apiset();
    } elsif ( $hosttype =~ /linux/i ) {
		$apiset = NACL::APISet->new(
			hostobj   => $command_interface->hostrec,
			category  => "Host",
			interface => "CLI",
			set       => "Linux",
		);
    } else {
        NATE::BaseException->throw( "TTCP not supported on Hosttype: $hosttype" );
    }
    return $apiset;
}
sub _background_run {
    my ($apiset, %args) = @_;
    my $subset_obj = Tharn::subtest( sub {
                                        my $string = undef;
                                        eval {
                                            my $obj = $apiset->ttcp(%args);
                                            my $output = $obj->get_parsed_output();
                                            map $string.="$_=$output->[0]->{$_}\n", keys %{$output->[0]};
                                        };
                                        Tharn::logresult('FATAL', "EVAL_ERROR=$@") if $@;
                                        Tharn::logresult('PASS', $string);
                                    }, '-bg', '-onexit', sub {}, );
    return $subset_obj;
}
1;
