#
# Copyright (c) 2001-2011 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Component Module
## @author dl-nacl-dev@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::MTask::VerifyVolID

=head1 DESCRIPTION

This task is to be used for Volume UUID testing.

A typical test would have the Volume UUIDs checked at the beginning of the
test, then perform volume operations, and then at the end of the test re-reads
the Volume UUIDs and checks them against the original values.
(Description of functionality and options accepted by L<new|lib-NACL-MTask-VerifyVolID-pm/new>.
Description of functionality and options accepted by L<verify|lib-NACL-MTask-VerifyVolID-pm/verify>.)

    # Start test
    use NACL::MTask::VerifyVolID;
    ...
    my $vvid = NACL::MTask::VerifyVolID->new(
        command_interface => $ci,
        volumes => [$vol1, $vol2, ... ],
        vserver => $vserver
    );
    ...
    # Perform vol ops here
    ...
    $vvid->verify();
    # End test
	
=cut

package NACL::MTask::VerifyVolID;

use strict;
use warnings;

use Params::Validate
    qw(validate validate_with BOOLEAN HASHREF OBJECT ARRAYREF SCALAR UNDEF);
use NACL::C::Volume;
use NACL::Exceptions::NoElementsFound qw(:try);
use NATE::BaseException qw(:try);
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use NACL::ComponentUtils qw(Dumper);
require Data::Dumper;

use base qw (NACL::MTask::MTask);

=head1 METHODS

=head2 new

    my $vvid = NACL::MTask::VerifyVolID->new(
        command_interface => $ci,
        volumes => [ $vol1, $vol2, ... ],
        vserver => $vserver,
        [require_provenance => 0|1,]
        [id_pairs_match => 0|1,]
        [apiset_must => $apiset_must_hashref,]
        [apiset_should => { $apiset_should_hashref},]
    );

Creates a new "VerifyVolID" object, reads the Instance-UUID and
Provenance-UUID of all volumes provided and checks whether they satisfy the
provided criteria. If any of the criteria were not matched, then a 
NATE::BaseException is thrown, with the exception text detailing the error(s)
that occurred.

=over

=item Options

=over

=item C<< command_interface => $ci >>

(Required, isa NACL::C::CommandInterface)
The command interface on which to read and verify the Volume UUIDs.
See L<NACL::C::Component::command_interface|lib-NACL-C-Component-pm/command_interface>

=item C<< volumes => [ $vol1, $vol2, ... ] >>

(Required, ARRAYREF)
An arrayref of volumes on which we want to verify the UUIDs. Each element of
the arrayref can be a volume name (scalar) or a volume component object (isa
'NACL::C::Volume')

=item C<< vserver => $vserver >>

(Required if run on a CMode host, optional if run on a 7Mode host. If provided
when run on a 7Mode host, then it is simply ignored)
The vserver on which the volumes should exist. This can be either the vserver
name or a Vserver component object. (isa 'NACL::C::Vserver')

=item C<< require_provenance => 0|1 >>

(Optional, defaults to 1 if not provided)
If this is set as 1, then it requires all volumes to have a Provenance-UUID. 
(i.e. absence of a Provenance-UUID value for any of the volumes will result
in failure).

If this is 0, it allows one or more volumes to have a blank Provenance-UUID
value.

=item C<< id_pairs_match => 0|1 >>

(Optional, defaults to 1 if not provided)
If set as 1, this option will mandate that the Provenance-UUID and
Instance-UUID for a particular volume match. However, for volumes which
have a blank Provenance-UUID, this condition is not applied.

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=back

=back

In addition to the checks performed as a result of the "require_provenance"
and "id_pairs_match", checks are also performed to ensure that all the
specified volumes exist on the specified vserver. If one or more of these
volumes does not exist then a NATE::BaseException is thrown.

=cut

use Class::MethodMaker [
    new    => [ '-hash', '-init', 'new' ],
    scalar => 'vserver',
    scalar => 'volumes',
    scalar => [ { -default => 1 }, 'require_provenance' ],
    scalar => [ { -default => 1 }, 'id_pairs_match' ],
    scalar => [
        { -type => 'NACL::C::CommandInterface::ONTAP' },
        'command_interface'
    ],
    scalar => '_old_uuids',
];

# Validate the parameters passed in and also read the VolIDs.
sub init {
    $Log->enter() if $may_enter;
    my $self = shift;
    my %opts = validate_with(
        params => \@_,
        spec   => {
            volumes            => { type => ARRAYREF },
            vserver            => { type => SCALAR | OBJECT | UNDEF, optional => 1 },
            command_interface  => { isa  => 'NACL::C::CommandInterface' },
            require_provenance => { type => BOOLEAN, optional => 1 },
            id_pairs_match     => { type => BOOLEAN, optional => 1 }
        },
    );

    # Validate that the arguments passed in volumes are either names,
    # or component or task objects

    my @volumes = @{ $self->volumes() };
    foreach my $each_vol (@volumes) {
        my $ref = ref $each_vol;
        if ($ref) {
            $Log->exit() if $may_exit;
            if (   !$each_vol->isa('NACL::C::Volume')
                && !$each_vol->isa('NACL::STask::Volume') )
            {
                NATE::Exceptions::Argument->throw( 'Volumes names (scalar),'
                        . "volume component objects (isa 'NACL::C::Volume' or "
                        . "volume task objects (isa 'NACL::STask::Volume') "
                        . "can be specified in the 'volumes' parameter "
                        . "but one of the inputs was of type '$ref'. "
                        . "Dumper of all options provided in 'volumes':\n"
                        . Data::Dumper::Dumper( \@volumes ) );
            }
        }
    }

    my %common_args;
    $self->_copy_common_component_params(
        source => \%opts,
        target => \%common_args,
    );
    $self->_read_and_update_volids(%common_args);
    $Log->exit() if $may_exit;
}

=head2 verify

    $vvid->verify(
        [changed_instance   => 0|1,]
        [changed_provenance => 0|1,]
		[apiset_must        => $ruleset,]
		[apiset_should      => $ruleset]
    );
	
(Instance method)
Re-reads the VolIDs (Instance-UUID and Provenance-UUID) for all the volumes,
checks whether they match the initial rules ("require_provenance" and
"id_pairs_match"), and verify the new values against the intial values.
(Refer to the explanations of "changed_instance" and "changed_provenance"
below for the checks performed)

This does not have any return value (silent success). Instead, on failure a 
NATE::BaseException is thrown, with the exception text detailing the error(s) 
that occurred.

=over

=item Options

=over

=item C<< changed_instance => 0|1 >>

(Optional, defaults to 0 if not provided)
If set to 1, require that the Instance-UUID for all volumes has changed.
If set to 0, require that the Instance-UUID for all volumes is the same.

=item C<< changed_provenance => 0|1 >>

(Optional, defaults to 0 if not provided)
If set to 1, require that the Provenance-UUID for all volumes has changed.
If set to 0, require that the Provenance-UUID for all volumes is the same.

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=back

=back

=cut

sub verify {
    $Log->enter() if $may_enter;
    my $self = shift;
    my %opts = validate @_,
        {
        changed_instance => {
            type    => BOOLEAN,
            default => 0
        },
        changed_provenance => {
            type    => BOOLEAN,
            default => 0
        },
        %{ $self->_component_common_validate_spec_without_ci() }
        };
    my $changed_instance   = $opts{changed_instance};
    my $changed_provenance = $opts{changed_provenance};

    my %common_args;
    $self->_copy_common_component_params(
        source => \%opts,
        target => \%common_args
    );

    my $old_hashref          = $self->_old_uuids();
    my @old_instance_array   = @{ $old_hashref->{instance_uuid} };
    my @old_provenance_array = @{ $old_hashref->{provenance_uuid} };

    $Log->comment("\nVerifying VolIDs");

    my $new_hashref
        = $self->_read_volids( %common_args, _log_read_result => 0 );
    my @new_instance_array   = @{ $new_hashref->{instance_uuid} };
    my @new_provenance_array = @{ $new_hashref->{provenance_uuid} };

    my $full_error_text;
    my $msg
        = "Results of verifying VolIDs:\n"
        . "(changed_instance = $changed_instance; "
        . "changed_provenance = $changed_provenance)\n\n";
    my $error_num = 1;
    my @volumes   = $self->_get_volume_names();
    foreach my $volume (@volumes) {
        my $old_instance   = shift @old_instance_array;
        my $old_provenance = shift @old_provenance_array;
        my $new_instance   = shift @new_instance_array;
        my $new_provenance = shift @new_provenance_array;

        $msg .= NATE::Util::indent( 4,
                  "For volume '$volume':\n"
                . "Old Instance-UUID: $old_instance\n"
                . "New Instance-UUID: $new_instance\n"
                . "Old Provenance-UUID: $old_provenance\n"
                . "New Provenance-UUID: $new_provenance\n" );

        my $has_instance_changed = 0;
        if ( $old_instance ne $new_instance ) {
            $has_instance_changed = 1;
        }

        my $has_provenance_changed = 0;
        if ( $old_provenance ne $new_provenance ) {
            $has_provenance_changed = 1;
        }

        if ( $has_instance_changed != $changed_instance ) {
            my $error_msg;
            if ($changed_instance) {
                $error_msg
                    = 'The Instance-UUID should have changed '
                    . '(changed_instance=1), however it is the same for '
                    . "volume '$volume'.\n"
                    . "Old and New Instance-UUID: $old_instance\n";
            } else {
                $error_msg
                    = 'The Instance-UUID should not have changed '
                    . '(changed_instance=0), however it has changed '
                    . "for volume '$volume'.\n"
                    . "Old Instance-UUID: $old_instance\n"
                    . "New Instance-UUID: $new_instance\n";
            }
            $msg .= _indent( 4, "Error! $error_msg" );
            $full_error_text .= "$error_num. $error_msg";
            $error_num++;
        }

        if ( $has_provenance_changed != $changed_provenance ) {
            my $error_msg;
            if ($changed_provenance) {
                $error_msg
                    = 'The Provenance-UUID should have changed '
                    . '(changed_provenance=1), however it is the same for '
                    . "volume '$volume'.\n"
                    . "Old and New Provenance-UUID: $old_provenance\n";
            } else {
                $error_msg
                    = 'The Provenance-UUID should not have changed '
                    . '(changed_provenance=0), however it has changed '
                    . "for volume '$volume'.\n"
                    . "Old Provenance-UUID: $old_provenance\n"
                    . "New Provenance-UUID: $new_provenance\n";
            }
            $msg .= _indent( 4, "Error! $error_msg" );
            $full_error_text .= "$error_num. $error_msg";
            $error_num++;
        }
        $msg .= "\n";
    }

    $Log->comment($msg);
    $Log->comment();

    if ($full_error_text) {
        $full_error_text
            = 'One or more errors occurred while verifying '
            . "VolIDs. Errors:\n"
            . $full_error_text;
        $Log->exit() if $may_exit;
        NATE::BaseException->throw(
            NATE::Util::indent( 4, $full_error_text ) );
    }

    $Log->exit() if $may_exit;
}

=begin COMMENTED_OUT

    $self->_read_and_update_volids(
        [apiset_must => $ruleset,]
        [$apiset_should => $ruleset,]
    );
	
(For developers of VerifyVolID task)
Reads VolIDs (by calling _read_volids()) and stores them in the object
as the "old" values. (Stored in $self->_old_uuids()). These Old UUIDs are
compared against during verification.

=end COMMENTED_OUT

=cut

sub _read_and_update_volids {
    $Log->enter() if $may_enter;
    my $self = shift;
    $Log->debug( "Opts to 'read_and_update_volids':\n" . Dumper( {@_} ) );
    my %common_args = validate @_,
        $self->_component_common_validate_spec_without_ci();
    $self->_old_uuids( $self->_read_volids(%common_args) );
    $Log->debug( "New VolIDs:\n" . Dumper( $self->_old_uuids() ) );
    $Log->exit() if $may_exit;
}

=begin COMMENTED_OUT

    my @volume_names = $self->_get_volume_names();
	
(For developers of VerifyVolID task)
The 'volumes' attribute of the object is allowed to be a combination of names,
volume component objects and volume task objects. This returns an array
of their corresponding volume names.

=end COMMENTED_OUT

=cut

sub _get_volume_names {
    $Log->enter() if $may_enter;
    my $self    = shift;
    my @volumes = @{ $self->volumes() };
    my @volume_names;
    foreach my $each_vol (@volumes) {
        if ( ref $each_vol ) {

            # Has to be a component or task object
            # Invoke the accessor to get the name
            push @volume_names, $each_vol->volume();
        } else {

            # The volume name is stored
            push @volume_names, $each_vol;
        }
    }
    $Log->exit() if $may_exit;
    return @volume_names;
}

=begin COMMENTED_OUT

    my $uuids = $self->_read_volids(
        [apiset_should    => $ruleset,]
        [apiset_must      => $ruleset,]
        [_log_read_result => 1]
    );

(For developers of the VerifyVolID task)
This takes on the heavy lifting of reading the Instance-UUID and Provenance-UUID
for all volumes, checking whether they match the criteria of "require_provenance"
and "id_pairs_match" (see the documentation of new for an explanation of these).
It also checks whether all of the specified volumes exist on the vserver,
and if the Instance-UUIDs are unique to volumes. Any error(s) results in a 
NATE::BaseException being thrown. 

This method is internally called by init (read the VolIDs at object creation
time) and by verify (read VolIDs and then verify the new values against the old
ones).

This returns a hashref which has two keys: "provenance_uuid" and "instance_uuid".
Each of these is in turn a hashref with each entry keyed by the volume name.
Sample Dumper:
    {
        provenance_uuid => [ 'prov_uuid1', 'prov_uuid2', ... ],
        instance_uuid   => [ 'inst_uuid0', 'inst_uuid1', ... ]
    }
	
The options it accepts are apiset_must, apiset_should and _log_read_result.
apiset_(must|should) have their usual meanings.

_log_read_result is optional and defaults to 1. It specifies whether to log
the result of reading the VolIDs. When called by verify, we should delay
logging till after the verification is done, so we send _log_read_result
as 0 to delay the logging in this case.

=end COMMENTED_OUT

=cut

sub _read_volids {
    $Log->enter() if $may_enter;
    my $self = shift;
    my %opts = validate @_,
        {
        %{ $self->_component_common_validate_spec_without_ci() },
        _log_read_result => { type => BOOLEAN, default => 1 }
        };

    my $_log_read_result = delete $opts{_log_read_result};

    my $need_to_compare = $self->id_pairs_match();
    my $req_prov        = $self->require_provenance();
    my ( $msg, $full_error_text );
    my $error_num = 1;

    $msg
        = "Result of reading VolIDs\n"
        . "(id_pairs_match = $need_to_compare; "
        . "require provenance = $req_prov)\n\n";
    my @volumes = $self->_get_volume_names();
    my $vserver = $self->vserver();
    my %instance_uuid_seen;
    my ( @provenance_uuid, @instance_uuid );
    my $vol_str = join ', ', @volumes;

    $Log->comment("\nReading VolIDs of volumes $vol_str\n");

    foreach my $volume (@volumes) {
        try {
            my $volume_state = NACL::CS::Volume->fetch(
                command_interface => $self->command_interface(),
                filter => { volume => $volume, vserver => $vserver },
                requested_fields => [qw(instance-uuid provenance-uuid)],
                %opts
            );
            my $inst_uuid = $volume_state->instance_uuid();
            my $prov_uuid = $volume_state->provenance_uuid();

            $msg .= NATE::Util::indent( 4,
                      "For volume '$volume':\n"
                    . "Instance-UUID: $inst_uuid\n"
                    . "Provenance-UUID: $prov_uuid\n" );

            push @instance_uuid,   $inst_uuid;
            push @provenance_uuid, $prov_uuid;

            my $need_to_compare = $self->id_pairs_match();

            if ($req_prov) {
                if ( !defined $prov_uuid || $prov_uuid eq '-' ) {
                    my $error_msg
                        = 'A Provenance UUID value is '
                        . 'required  (since require_provenance has been set '
                        . 'to 1) but it is blank  '
                        . "for volume '$volume'\n";
                    $msg .= _indent( 4, "Error! $error_msg" );
                    $full_error_text .= "$error_num. $error_msg";
                    $error_num++;
                }
            } else {
                $need_to_compare = 0;
            }

            if ($need_to_compare) {
                if ( $inst_uuid ne $prov_uuid ) {
                    my $error_msg
                        = 'The Instance-UUID '
                        . "($inst_uuid) does not match the "
                        . "Provenance-UUID ('$prov_uuid') for volume "
                        . "'$volume'. (These ID pairs must match since "
                        . "'id_pairs_match' has been set to 1)\n";
                    $msg .= _indent( 4, "Error! $error_msg" );
                    $full_error_text .= "$error_num. $error_msg";
                    $error_num++;
                }
            }

            if ( exists $instance_uuid_seen{$inst_uuid} ) {
                my $error_msg
                    = "The Instance-UUID ($inst_uuid) "
                    . "of volume '$volume' matches the Instance-UUID "
                    . "of volume '$instance_uuid_seen{$inst_uuid}'. "
                    . "(Instance-UUIDs must be unique to a volume)\n";
                $msg .= _indent( 4, "Error! $error_msg" );
                $full_error_text .= "error_num. $error_msg";
                $error_num++;
            } else {
                $instance_uuid_seen{$inst_uuid} = $volume;
            }

        }
        catch NACL::Exceptions::NoElementsFound with {
            my $error_msg = "The volume '$volume' was not found ";
            if (defined $vserver) {
                my $vserver_name
                    = ( ref $vserver )
                    ? $vserver->vserver()
                    : $vserver;
                    $error_msg .= "on vserver '$vserver_name'\n";
            }
            $msg .= _indent( 4, "Error! $error_msg\n" );
            $full_error_text .= $error_msg;
        }
        finally {
            $msg .= "\n";
            $full_error_text .= "\n" if $full_error_text;
        };
    }

    my $log_read_msg = sub {
        $Log->comment("$msg\n");
    };

    $log_read_msg->() if $_log_read_result;

    if ($full_error_text) {
        $full_error_text
            = 'One or more errors occurred while reading '
            . "VolIDs. Errors:\n"
            . $full_error_text;

        $log_read_msg->() unless $_log_read_result;
        $Log->exit() if $may_exit;
        NATE::BaseException->throw(
            NATE::Util::indent( 4, $full_error_text ) );
    } else {

        $Log->exit() if $may_exit;
        return {
            instance_uuid   => \@instance_uuid,
            provenance_uuid => \@provenance_uuid
        };
    }
}

# my $new_msg = _indent( $count, $msg );
# Indent all lines of $msg by $count and return this new message
sub _indent {
    my ( $count, $msg ) = @_;
    my $new_msg = ( ' ' x $count ) . $msg;
    return NATE::Util::indent( $count, $new_msg );
}

1;
