#
#
# Copyright (c) 2013 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#
#
## @summary MTask support Module
## @author asada@netapp.com, dl-nacl-dev@netapp.com
## @status shared
## @pod here
##############################################################################
package NACL::MTask::VolumeFileIterator::LocalNASFile;
use Moose;
use MTF::MooseCompatibleException qw(:all);

# perl file stat conver
use Fcntl qw(S_ISDIR S_ISREG S_ISLNK S_ISBLK S_ISCHR S_ISFIFO S_ISSOCK);

# NATE

use Data::Dumper;
use File::Spec::Functions qw(catdir);
use NACL::C::Client::Directory;
use NATE::BaseException;

#NACL
use NACL::C::VolumeFile;

#
# Function: _type
#   Type of file can be: file, directory and so on
# Arguments:
#   [0] - String file type
# Returns:
#   Returns the type of file
#
has _type => (
    is => 'rw',
    isa => 'Str',
);

#
# Function: volume
#
# Arguments:
#   [0] - NACL::C:Volume object instance
# Returns:
#   Returns NACL::C::Volume object instance
#
has volume => (
    is => 'ro',
    isa => 'NACL::C::Volume',
    required => 1,
);

#
# Function: path
#
# Arguments:
#   [0] - path within the volume
# Returns:
#   Returns path within the volume to the file
#
has path => (
    is => 'ro',
    isa => 'Str',
    clearer   => 'clear_path',
    predicate => 'has_path',
    default => '',
);

#
# mount object to be used to get the client side mount point
#
has 'mount' => (
    is => 'ro',
    isa => 'Object',
);
#
# Function: _name
#
# Arguments:
#   [0] - file/directory name
# Returns:
#   Returns file/directory name.
#
has _name => (
    is => 'ro',
    isa => 'Str',
    default => '',
);


sub raw_name {
    my $self = $_[0];
    my $raw_name = catdir ($self->mount()->mount_point(),$self->path(),"/",$self->_name());
    $raw_name = _agnostic_catdir_conversion($raw_name);
    return $raw_name;
}

sub standard_type {
    my $self = $_[0];
    return $self->_type();
}

sub is_directory {
    my $self = $_[0];
    if ($self->standard_type() =~ qr/directory/xi) {
        return 1; #TRUE
    }
    return 0; # FALSE
}

sub enumerate_directory {
    $Log->enter();
    my $self = $_[0];

    # Runs on this client itself
    # use the path provided to call NACL:C:Client::Directory open and read
    # raw name is mount point ($self->mount()->mount_point()), path and name (dir name) concatenated
    # throws a base exception if it cant open
    my $cli_dir_obj = NACL::C::Client::Directory->open(dir => $self->raw_name());

    # throws a base exception if it cant read
    my @files = $cli_dir_obj->read();

    # Call close right after
    $cli_dir_obj->close();

    my @local_nas_file_objs = ();

    foreach my $file (@files) {
        if (
            $file =~ /^\.snapshot$/ ||
            $file =~ /^(")?\.(")?$/ ||
            $file =~ /^(")?\.\.(")?$/
            )
        {
            next;
        };

        $Log->debug ("File: $file");
        my $file_path =  catdir($self->mount->mount_point(),$self->path(),$file);
        $file_path = _agnostic_catdir_conversion($file_path);

        my $type = $self->_get_file_type($file_path);

        if (!defined $type) {
            next;
        }
        if (!defined $self->path()) {
            $self->path("");
        }

        my $baby_nas_file_obj = NACL::MTask::VolumeFileIterator::LocalNASFile->new (
            path => $self->path(),
            mount => $self->mount(),
            volume => $self->volume(),
            _type => $type,
            _name => $file,
        );

        push @local_nas_file_objs, $baby_nas_file_obj;
    }

    $Log->debug(sub { "Files found under the specfied path: " . Dumper(\@local_nas_file_objs) });
    $Log->exit();
    return @local_nas_file_objs;
}

sub make_volume_file {
    my $self = $_[0];
    $Log->enter();

    my $volume = $self->volume();
    my $path =  catdir("","vol",$volume->volume,$self->path(),$self->_name());
    $path = _agnostic_catdir_conversion($path);

    # Convert the file to NACL::C::Volume file object
    my $vol_file_obj = NACL::C::VolumeFile->new(
        command_interface => $volume->command_interface(),
        vserver => $volume->vserver(),
        path => $path,  #volume/directory/file path and its name
    );

    $Log->exit();
    return $vol_file_obj;
}

sub _get_file_type {
    my $self = shift;
    my $file_path = shift;
    $Log->enter();

    my $type = undef;
    my @attributes = ();
    # Lets Stat() the file and set the type
    eval {
        @attributes = stat($file_path);
    };

    if (!($@)) {
        my $st_mode = $attributes[2];
        # call marcos to check if it is a directory or regular file else next
        if(S_ISREG($st_mode)) {
            $type = "file";
        }
        elsif (S_ISDIR($st_mode)) {
            $type = "directory";
        }
        elsif (S_ISLNK($st_mode)) {
            $type = "symlink";
        }
        elsif (S_ISBLK($st_mode)) {
            $type = "blockdev";
        }
        elsif (S_ISCHR($st_mode)) {
            $type = "chardev";
        }
        elsif (S_ISFIFO($st_mode)) {
            $type = "fifo";
        }
        elsif (S_ISSOCK($st_mode)) {
            $type = "socket";
        }
    } else {
        $Log->enter();
        # stat() failed
        NATE::BaseException->throw("Error stating $file_path: $@");
    }

    return $type;
}

sub _agnostic_catdir_conversion {
    $Log->enter();
    my $val = $_[0];
    $val =~ s/\\/\//g;
    $Log->exit();
    return $val;
}

1;

=head1 NAME

Note: This is intended to iterate files from a local client to which the mount point is provided.

NACL::MTask::VolumeFileIterator::LocalNASFiles - LocalNas Files handles processing LocalNas file types
under volume file iterator

=head1 SYNOPSIS

    use NACL::MTask::VolumeFileIterator::LocalNASFiles;

    my $zp_files_obj = NACL::MTask::VolumeFileIterator::LocalNASFile->new(
        volume => $volume,
        path => $mount_point,
    );


=head1 DESCRIPTION

This LocalNas Files library is used by Volume File Iterator for understanding and operating on File LocalNas's.

=head1 FUNCTIONS

=head2 new

Build a new LocalNASFiles object instance
Arguments:
volume - NACL::C::Volume instance
path - mount point to where the files/directories reside

=head3 Returns

object instance is returned

=head2 raw_name

Builds the complete file/directory path (using mount point) and name
Arguments: n.a

=head3 Returns

Concatenated file/directory path and name are returned.

=head2 enumerate_directory

Builds a list of (LocalNASFiles object) files/directories underneath the specified instance directory
Arguments: n.a (uses the instance path)

=head3 Returns

A list of LocalNas Files object is returned.

=head2 make_volume_file

Builds a NACL::C::VolumeFile object for LocalNas File object
Arguments:
command_interface - (required) command interface object

=head3 Returns

A NACL::C::VolumeFile object for this instance of  LocalNas Files object.

=head2 standard_type

File type to return
Arguments: n.a

=head3 Returns

Returns the file object type (types: file, directory)

=head2 is_directory

Returns TRUE 1 if object is a directory
Arguments:

=head3 Returns

TRUE or FALSE if file is dir or not.

=head1 SEE ALSO

=cut
