#
#
# Copyright (c) 2013 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#
#
## @summary MTask support Module
## @author asada@netapp.com, dl-nacl-dev@netapp.com
## @status shared
## @pod here
##############################################################################
package NACL::MTask::VolumeFileIterator::ZapiFile;
use Moose;
use MTF::MooseCompatibleException qw(:all);

# NATE
use Data::Dumper;
use Params::Validate qw(validate SCALAR ARRAYREF HASHREF SCALARREF );
use File::Spec::Functions qw(catdir);
use NACL::CS::VolumeListFilesInDirectory;

#NACL
use NACL::C::VolumeFile;

#
# Function: _type
#   Type of file can be: file, directory and so on
# Arguments:
#   [0] - String file type
# Returns:
#   Returns the type of file
#
has _type => (
    is => 'rw',
    isa => 'Str',
);

#
# Function: volume
#
# Arguments:
#   [0] - NACL::C:Volume object instance
# Returns:
#   Returns NACL::C::Volume object instance
#
has volume => (
    is => 'ro',
    isa => 'NACL::C::Volume',
    required => 1,
);

#
# Function: path
#
# Arguments:
#   [0] - path to the directory/file
# Returns:
#   Returns the path to file/directory.
#
has path => (
    is => 'ro',
    isa => 'Str',
    clearer   => 'clear_path',
    predicate => 'has_path',
    required  => 1,
);

#
# Function: _name
#
# Arguments:
#   [0] - file/directory name
# Returns:
#   Returns file/directory name.
#
has _name => (
    is => 'ro',
    isa => 'Str',
    default => '.',
);

sub raw_name {
    my $self = $_[0];
    my $raw_name = catdir ($self->path(),$self->_name());
    $raw_name = _agnostic_catdir_conversion($raw_name);
    return $raw_name;
}

sub standard_type {
    my $self = $_[0];
    return $self->_type();
}

sub is_directory {
    my $self = $_[0];
    if ($self->_type() =~ qr/directory/xi) {
        return 1; #TRUE
    }
    return 0; # FALSE
}

sub enumerate_directory {
    $Log->enter();
    my $self = $_[0];

    my @zapi_file_objs = ();

    my $volume = $self->volume();
    my $dir_path = catdir($self->path(),$self->_name()); # Volume downwards e.g: /dir_name  or /dir_name/sub_dir
    $dir_path = _agnostic_catdir_conversion($dir_path);

    my $apiset = $volume->command_interface()->apiset(
        category => 'Node',
        interface => 'ZAPI',
        set => 'CMode',
    );
    my $complete_path = catdir('/vol',$volume->volume(),$dir_path);
    $complete_path = _agnostic_catdir_conversion($complete_path);
    $Log->debug("File Path in ZapiFile.pm  ". $complete_path);
    # Returns Array of NACL::CS::VolumeFile objects
    my @files = NACL::CS::VolumeListFilesInDirectory->fetch(
        command_interface => $volume->command_interface(),
        filter            => {
            'directory-path'    => $complete_path,
            vserver => $volume->vserver(),
        },
        requested_fields => ['is-junction', 'file-type',],
    );

    foreach my $vol_file_cs_obj (@files) {
        # The (")? is a workaround for 727682
        if (
            $vol_file_cs_obj->name() =~ /^\.snapshot$/ ||
            $vol_file_cs_obj->name() =~ /^(")?\.(")?$/ ||
            $vol_file_cs_obj->name() =~ /^(")?\.\.(")?$/
            )
        {
            next;
        };

        if ( ($vol_file_cs_obj->file_type() =~ qr/junction/xi) ||
             ($vol_file_cs_obj->is_junction() =~ /true/i) ){
            next;
        }

        # Burt: 737872 - Storing file type to what ZAPI is returning should be one of the
        # following types: file, directory, blockdev, chardev, symlink, socket, fifo, stream, lun
        $Log->debug ("File type: ". $vol_file_cs_obj->file_type());

        # Lets create a zapi file object for each entity (file or directory) under this
        # specified directory for 'this' (java love) Zapi File Object
        my $baby_zapi_file_obj = NACL::MTask::VolumeFileIterator::ZapiFile->new (
            volume => $volume,
            path => $dir_path,
            _type => $vol_file_cs_obj->file_type(),
            _name => $vol_file_cs_obj->name(),
        );

        push @zapi_file_objs, $baby_zapi_file_obj;
    }

    $Log->debug(sub { "Files found under the specfied path: " . Dumper(\@zapi_file_objs) });
    $Log->exit();
    return @zapi_file_objs;
}

sub make_volume_file {
    my $self = $_[0];
    $Log->enter();

    my $volume = $self->volume();

    my $path =  catdir("","vol",$volume->volume,$self->raw_name());
    $path = _agnostic_catdir_conversion($path);

    # Convert the file to NACL::C::Volume file object
    my $vol_file_obj = NACL::C::VolumeFile->new(
        command_interface => $volume->command_interface(),
        vserver => $volume->vserver(),
        path => $path,  #directory/file path and its name
    );

    $Log->exit();
    return $vol_file_obj;
}

sub _agnostic_catdir_conversion {
    $Log->enter();
    my $val = $_[0];
    $val =~ s/\\/\//g;
    $Log->exit();
    return $val;
}

1;

=head1 NAME

NACL::MTask::VolumeFileIterator::ZapiFiles - Zapi Files handles processing zapi file types
under volume file iterator

=head1 SYNOPSIS

    use NACL::MTask::VolumeFileIterator::ZapiFiles;

    my $zp_files_obj = NACL::MTask::VolumeFileIterator::ZapiFile->new(
        volume => $volume,
        path => $path_to_file_directory,
    );

    alternatively

    my $zp_files_obj = NACL::MTask::VolumeFileIterator::ZapiFile->new(
        volume => $volume,
        path => $path_to_file_directory,
        );

=head1 DESCRIPTION

This Zapi Files library is used by Volume File Iterator for understanding and operating on File Zapi's.

=head1 FUNCTIONS

=head2 new

Build a new ZapiFiles object instance
Arguments:
volume - NACL::C::Volume instance
path - Path of the directory where the files/further directories reside

=head3 Returns

object instance is returned

=head2 raw_name

Builds the complete file/directory path and name
Arguments: n.a

=head3 Returns

Concatenated file/directory path and name are returned.

=head2 enumerate_directory

Builds a list of (ZapiFiles object) files/directories underneath the specified instance directory
Arguments: n.a (uses the instance path)

=head3 Returns

A list of Zapi Files object is returned.

=head2 make_volume_file

Builds a NACL::C::VolumeFile object for Zapi File object
Arguments:
command_interface - (required) command interface object

=head3 Returns

A NACL::C::VolumeFile object for this instance of  Zapi Files object.

=head2 standard_type

File type to return
Arguments: n.a

=head3 Returns

Returns the file object type (types: file, directory, blockdev, chardev, symlink, socket, fifo, stream, lun)

=head2 is_directory

Returns TRUE 1 if object is a directory
Arguments:

=head3 Returns

TRUE or FALSE if file is dir or not.

=head1 SEE ALSO

=cut
