# $Id: //depot/prod/test/nacldev/lib/NACL/MTask/VolumeQuota.pm#1 $
#
# Copyright (c) 2001-2011 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Volume Quota MTask module
## @author swati.bajaj@netapp.com, dl-nacl-dev@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::MTask::VolumeQuota

=head1 DESCRIPTION

C<NACL::MTask::VolumeQuota> provides a number of well-defined but
potentially complex or multi-step methods related to Quotas in
ONTAP.
It builds on top of and is a derived class of C<NACL::C::VolumeQuota>. This
means that any functionality present in 
L<NACL::C::VolumeQuota|lib-NACL-C-VolumeQuota-pm> will also be
present in this MTask object.

=head1 ATTRIBUTES

=head2 command_interface

A component object that represents the host to which to send commands.

=head2 volume

The name of the volume for which quota applies.

=head2 vserver

The name of the vserver for which quota policy is applied (Applicable for CMode only)

=head2 unix_user

Represents the NACL::STask::VserverServicesUnixUser instance that was created during the setup
(Applicable only for CMode and if the type is 'user')

=head2 unix_group

Represents the NACL::STask::VserverServicesUnixGroup instance that was created during the setup
(Applicable only for CMode and if the type is 'group')

=head2 quota_policy

Represents the NACL::C::VolumeQuotaPolicy instance that was created during 
the setup (Applicable only for CMode)

=head2 quota_policy_rule

Represents the NACL::C::VolumeQuotaPolicyRule instance that was created 
during the setup

=head1 METHODS

=head2 setup

    # Example usage for tree quota
    my $quota_obj = NACL::MTask::VolumeQuota->setup(
        command_interface => $command_interface,
        volume => $volume_name,
        vserver => $vserver_name,
        type => "tree", 
        nacltask_if_exists => $action, #default is die
        nacltask_quota_on => $boolean, #default is 1
        %other_options
    );

    # Example usage for user quota
    my $quota_obj = NACL::MTask:Quota->setup(
        command_interface => $command_interface,
        volume => $volume_obj,
        qtree => $qtree_name,
        type => $type,
        disk-limit => $disk_limit,
        file-limit => $file_limit,
        threshold => $threshold,
        soft-disk-limit => $soft_disk_limit,
        soft-file-limit => $soft_file_limit,
        user => $user,
        'id' => "2",
        'primary-gid' => 2,
        nacltask_create_user_or_group => $boolean, #default is 1
        nacltask_if_exists => $action, #default is die
        %other_options
    );
                
(Class method)
This method will mainly be concerned with setting up of quotas.
Either the volume or qtree should be provided to setup quotas.

In case of CMode it does the following:
1) Assumes the volume/qtree specified is already created, throws an exception otherwise.
2) Create a quota policy using volume quota policy
3) Create a rule in above policy using volume quota policy rule
4) Apply quota policy on the vserver
5) Turn on the quota on the given volume if nacltask_quota_on is set.
6) If the 'type' specified is 'user' and nacltask_create_user_or_group is set to 1, 
   a UNIX user is created before applying quota
7) If the 'type' specified is 'group' and nacltask_create_user_or_group is set to 1,
   a UNIX group is created before applying quota. 

In case of 7Mode it does the following:
2) Creates new quota rules in /etc/quotas
2) Turn on the quotas for the given volume if nacltask_quota_on is set

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required)

This is the component object that represents the host to which the commands are 
sent.

=item C<< volume => $volume_name >>

(Optional)

The name of the volume or volume Component/STask object for which the quota has 
to be setup. 
Either the volume or qtree should be provided to setup quotas.

=item C<< qtree => $qtree_name >>

(Optional)

A qtree name. If the intention is to setup quota for a qtree.
Either the qtree or volume should be provided to setup quotas.

=item C<< vserver => $vserver_name >>

(Required if run on a CMode host, optional if run on a 7Mode host. If provided
when run on a 7Mode host, then it is simply ignored)
The vserver on which the volumes should exist. This can be either be the vserver
name or a Vserver component object. (isa 'NACL::C::Vserver')

=item C<< type => $type >>

(Required)

Type of quota rule to be created. It could be 'tree', 'user', 'group'.

=item C<< nacltask_create_user_or_group => $boolean >>

(Optional) - when set to 1, Unix User or Unix Group will be created.
           - when set to 0, Unix User or Unix Group will not be created..
           - by default it is set to 1.

=item  C<< nacltask_if_exists=>$action >>

(Optional) Specifies an action to be taken if the quota has already been 
created.

     $action 
    - if set to  "die"(default), an exception is raised.
    - if set to "reuse", no action is performed
    - if set to "purge", deletes the existing quota rule before creating it.
    
=item C<< nacltask_quota_on => $boolean >>

(Optional) - when set to 1, quota created will be turned on.
           - when set to 0, the quota created will not be turned on.
           - by default it is set to 1.

=back

=item apiset_must, apiset_should, method-timeout,
all the other options supported by
NACL::STask::VserverServicesUnixUser->create,
NACL::STask::VserverServicesUnixGroup->create,
NACL::C::VolumeQuotaPolicy->create 
NACL::C::VolumeQuotaPolicyRule->create
NACL::C::VolumeQuota->on are supported by
NACL::MTask::VolumeQuota->setup as well.

=back

=cut

package NACL::MTask::VolumeQuota;

use strict;
use warnings;

use base qw/NACL::STask::VolumeQuota NACL::MTask::MTask/;

use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use NACL::STask::STask;
use NACL::STask::VolumeQuota;
use NACL::C::Volume;
use NACL::C::VolumeQuotaPolicy;
use NACL::C::VolumeQuotaPolicyRule;
use NACL::STask::VserverServicesUnixUser;
use NACL::STask::VserverServicesUnixGroup;
use NACL::C::Vserver;
use NACL::CS::VolumeQuota;
use NACL::ComponentUtils qw(Dumper);
use NATE::Exceptions::Argument qw(:try);
use Params::Validate qw(validate_with SCALAR SCALARREF BOOLEAN OBJECT);
use NACL::APISet::Exceptions::CommandFailedException;

use Class::MethodMaker [
    scalar => [
        { -type => 'NACL::STask::VserverServicesUnixUser' },
        'unix_user'
    ],
    scalar => [
        { -type => 'NACL::STask::VserverServicesUnixGroup' },
        'unix_group'
    ],
    scalar => [
        { -type => 'NACL::C::VolumeQuotaPolicy' },
        'quota_policy'
    ],
    scalar => [
        { -type => 'NACL::C::VolumeQuotaPolicyRule' },
        'quota_policy_rule'
    ],
];


sub setup {
    $Log->enter();
    my $pkg  = shift;
    
    if ( ref($pkg) ) {
        $Log->exit();
        NATE::BaseException->throw(
            "This Setup can be invoked only as a  NACL::MTask::VolumeQuota class method"
        );
    } ## end if ( ref($pkg) )
        
    my %opts = $pkg->_common_validate_with(
        params => \@_,
        additional_spec   => {
            nacltask_if_exists =>  $pkg->_if_exists_validate_spec,
            nacltask_quota_on => { type => SCALAR, default => 1 },
            nacltask_create_user_or_group => { type => SCALAR, default => 1 },
            volume => { type     => SCALAR | OBJECT, optional => 1,},
            vserver => { type => SCALAR | OBJECT, optional => 1,},
            qtree => { type => SCALAR , optional => 1,},
        },
        allow_extra => 1
    );
 
    $Log->debug( sub {"This is the opts in initial Setup " . Dumper( \%opts )} );
   
    my $volume = $opts{volume};
    my $qtree = $opts{qtree};
    my $ci = $opts{'command_interface'};

    # if the volume or qtree is not provided then error out
    if ( !$volume && !$qtree ) {
        $Log->exit();
        NATE::Exceptions::Argument->throw('Either of Volume or qtree are '.
                                          'required but neither were provided');
    }
    
    if (ref $volume ) {
        $opts{volume} = $volume->volume();
        if (!($opts{'vserver'}) and $ci->mode() eq 'CMode' ) {
            $opts{vserver} = $volume->vserver();
        }
    }       
    
    my $obj = $pkg->new(
        command_interface => $ci,
        volume            => $opts{volume},
        vserver           => $opts{vserver},
    );

    my $vserver = $opts{vserver}; 
    my $type = $opts{type};
    my $policy_name = $opts{'policy-name'};
    my $verify = delete $opts{nacltask_verify};
    my $quota_on = delete $opts{'nacltask_quota_on'};
    my $user_or_group = delete $opts{'nacltask_create_user_or_group'};

    my %component_call_opts = %opts;    

    # these are the additional steps required to setup quotas in CMode
    if($ci->mode() eq 'CMode') {
        # For creating user/group Quota we need to do this additional step
        if ((defined $user_or_group) && ($type =~ /^user|group$/)) {

            my (%common_opts, %stask_option, %create_option, $method, $pkg_name);

            # Get the corresponding package
            if ($type eq "user") {
                $method = "unix_user";
                $pkg_name = "NACL::STask::VserverServicesUnixUser";
            } else {
                $method = "unix_group";
                $pkg_name = "NACL::STask::VserverServicesUnixGroup";
            }

            # Copy common parameters
            $pkg_name->_copy_common_component_params_with_ci(
                source => \%opts,
                target => \%common_opts
            );

            # Get the STask specific option
            $pkg_name->_hash_copy(
               source => \%opts,
               target => \%stask_option,
               copy   => ['nacltask_if_exists', 'purge_if_duplicate_by_id']
            );

            # Get the corresponding command options
            my $apiset = $opts{command_interface}->apiset(interface => 'CLI');
            my $api = "vserver_services_".$method."_create";
            my @options = $apiset->get_allowed_options_for_command( command => $api );
            $pkg_name->_hash_copy(
                source => \%opts,
                target => \%create_option,
                copy   => \@options,
            );

            # Call the corresponding package
            my $create_obj = $pkg_name->create(
                %create_option,
                %common_opts,
                %stask_option,
                nacltask_wait => 1,
            );

            $obj->$method($create_obj);
        }

        $Log->trace("Create a Quota Policy");
        my $quota_policy =  _handle_creation(
            %component_call_opts,
            component     => 'NACL::C::VolumeQuotaPolicy',
            method        => 'create',
            delete_method => 'delete',
        );
        $obj->quota_policy($quota_policy);

        $Log->trace("Modify quota policy for vserver");
        NACL::C::Vserver->modify(
            command_interface=>$ci,
            vserver =>$vserver,
            "quota-policy" => $policy_name,
        );
    } # end CMode specific setup

    # Common to both CMode and 7Mode
    $Log->trace("Create a Quota Policy Rule");
    my $quota_policy_rule = _handle_creation(
        %component_call_opts,
        component     => 'NACL::C::VolumeQuotaPolicyRule',
        method        => 'create',
        delete_method => 'delete',
    );
    $obj->quota_policy_rule($quota_policy_rule);
    
    if ($quota_on) {    
        $Log->trace("Enable Quota on the volume");
        my %common_opts;
        $pkg->_copy_common_component_params(
            source => \%opts,
            target => \%common_opts
        );
        # If quotas are already on for the volume and a new rule is being 
        # added that changes something other than limits, quotas must be 
        # turned off, then turned back on to make the rule active
        
        NACL::STask::VolumeQuota->off(
                                    command_interface=>$ci,
                                    volume => $volume,
                                    vserver => $vserver,
                                    %common_opts
                                );

        NACL::STask::VolumeQuota->on(
                                    command_interface=>$ci,
                                    volume => $volume,
                                    vserver => $vserver,
                                    %common_opts
                                );
    }
        
    $Log->exit();
    return $obj;
}

############################################################################################################
# Name        : _handle_creation
# Description : Invoked from setup. If there are duplicate entries already existing on the filer, based on
#               nacltask_if_exists option which might be to die,purge or reuse, appropriate action is taken. 
#               If e.g. nacltask_if_exists => purge, It will call purge action which deletes the entries 
#               and recreates the entries. If that configuration is not there, it simply creates the entry. 
#               Returns an object of the requested component type only if this method creates one so that 
#               this can be stored as an attribute for deletion during cleanup.
############################################################################################################

sub _handle_creation {
    $Log->enter();

    my %opts = @_;
    
    # get the name of the component and the name of the method required for
    # create/add
    my $pkg    = $opts{component};
    my $method = $opts{method};

    # extract other necessary details
    my $command_interface = $opts{command_interface};
    my $if_exists = $opts{nacltask_if_exists};
    my $existing = undef;
    
    # We will not handle the nacltask_if_exists => 'die' case here since the
    # create call will anyway throw an error if the element exists.
    if ( $if_exists ne 'die' ) {
        $Log->trace('Attempt to find the element');
        $existing = $pkg->find(
            command_interface => $command_interface,
            filter            => $pkg->get_primary_keys_options(%opts),
            allow_empty       => 1
        );
    }

    if ($existing) {
        $Log->trace('Found a matching element');
        if ( $if_exists eq 'purge' ) {
            _purge(%opts);
        } elsif ( $if_exists eq 'reuse' ) {
            $Log->exit();
            return $existing;
        }
    }

    # Create the element here if not found above, and if nacltask_if_exists is set to 'die'
    my $self = $pkg->$method(%opts, 
                             command_interface => $command_interface,
                             allow_extra       => 1);

    $Log->exit();
    return $self;
}

############################################################################################################
# Name        : _purge
# Description : Invoked from _handle_creation. It takes the appropriate delete method to be used and the
#               component that has requested for purge, frames the method with appropriate parameters
#               to be called and invokes the same.
############################################################################################################

sub _purge {
    $Log->enter();
    my %opts = @_;
    my $pkg    = $opts{component};
    my $method = $opts{delete_method};
    my $command_interface = $opts{command_interface};

    # call either delete/remove on the package provided
    try {
        $pkg->$method( command_interface => $command_interface, %opts, allow_extra => 1 );
    } catch NACL::APISet::Exceptions::CommandFailedException with {
        my $ex = shift;
        if ( $ex->text() =~ /command failed: invalid operation/ && $pkg->isa('NACL::C::VolumeQuotaPolicy') ) {
            $Log->debug( "There may be vservers associated with this policy." . 
                         "We will modify the vserver to default to the ".
                         "'default' quota-policy" );
            
            NACL::C::Vserver->modify(
                command_interface=>$command_interface,
                vserver =>$opts{vserver},
                "quota-policy" => 'default',
            );
            # Call delete/purge method again
            $pkg->$method( command_interface => $command_interface, 
                           %opts,
                           allow_extra => 1 
            );
        } else {
            $Log->exit();
            $ex->throw();
        }
    };

    $Log->exit();
}

1;
