#
# Copyright (c) 2011 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Multiple Vserver Task Module
## @author bruce.blinn@netapp.com
## @status shared
## @pod here

package NACL::MTask::VserverMulti;

use strict;
use warnings;

use base qw(NACL::STask::Vserver);

use NACL::ComponentUtils qw(Dumper);
use List::Util qw(min max);
use NACL::C::SecurityLogin;
use NACL::MTask::VserverSetup;
use NACL::STask::NetworkInterface;
use NACL::STask::Volume;
use NATE::BaseException qw(:try);
use NATE::Exceptions::Argument;
use NACL::C::VserverServicesUnixGroup;
use NACL::C::VserverServicesUnixUser;
use Params::Validate qw(validate SCALAR SCALARREF ARRAYREF BOOLEAN);
use Net::Netmask;
use Tharn qw( host subtest logresult);
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();

=head1 NAME

NACL::MTask::VserverMulti

=head1 DESCRIPTION

C<NACL::MTask::VserverMulti> adds methods to C<NACL::STask::Vserver> to make it
easier to create or purge multiple VServers in a single method call.

This module will be used in tests and utility scripts needed to test very
large configurations, such as those being tested in Panamax.

=head1 METHODS

=head2 create_multi

    my @vsvr_names = NACL::MTask::VserverMulti->create_multi(
        command_interface           => $command_interface,
        vserver                     => $vserver,
        aggregate                   => $aggregate,
        nacltask_modify_root_vol    => $boolean,
        nacltask_num_subtests       => $num_subtests,
        nacltask_num_vservers       => $num_vservers,
        nacltask_start_index        => $start_index,
        nacltask_subtest_delay      => $num_seconds,
        nacltask_subtests_per_second => $num_subtests,
        %other_options
    );

(Class method) Create multiple VServers in a single method call.  The standard
VServer attributes can be passed to this method, in which case, they will be
passed through to the create() method when each VServer is created.  However,
the rootvolume attribute is automatically set by this method, and therefore,
cannot be specified by the caller.  (The name of the root volume will be set to
"root_" appended with the name of the vserver.)

The following steps will be performed for each VServer:

    Create the VServer.
    Increase the number of INODEs in the root volume.
    Create a data LIF for the VServer.
    Initialize the VServer administrative user.

If this method is called in array context, it will return an array containing
the names of the VServers that were created (or started to be created, if there
was an error); it cannot return the VServer objects because they are created in
a separate process.  If this method is called in a scalar context, it will
return the number of VServers created.

If the VServer is created with a default root volume, the root volume will be
modified so that it can contain at least 2000 files and directories (a default
root volume is limited to about 550 files and directories).  This behavior can
be disabled by setting the "nacltask_modify_root_vol" parameter to 0.

If you want LIFs created for the VServers, you need to create a NATE host file
for each VServer.  If there is no NATE host file for a VServer, that VServer
will not have a LIF.  If the NATE host file exists for a VServer, the following
entries will be used.

    hosttype            # Required entry.
    DATA_IP1
    DATA_MASK1          # Default 255.255.255.0
    DATA_PORT1          # Default e0c
    DATA_NODE1          # Default $command_interface->node()
    default_username    # Default vsadmin
    default_password    # Default netapp1!

The "hosttype" entry is not used by this module, but it is required in all NATE
host files.  The other entries listed above are optional, and any additional
entries in the NATE host file will be ignored (at least, they are not used when
creating the VServer).

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) A component object that represents the host
to which to send commands.
See L<NACL::C::Component::command_interface|
lib-NACL-C-Component-pm/command_interface>.

=item C<< vserver => $vserver >>

(Required) This parameter either contains a reference to an array of VServer
names, or it contains a string that will be the beginning of each VServer name.

If this parameter contains a reference to an array, then it contains the actual
names of the VServers that will be created.  Therefore, the number of VServers
that will be created is the number of entries in the array.  For example,

    vserver => ['vs_coke', 'vs_pepsi', 'vs_water']

If the vserver parameter contains a string, it will be the prefix of the
VServer names.  If the number of VServers to create is greater than one, the
actual names will be appended with an underscore and a 5-digit number.  (In
this case, the nacltask_start_index parameter will be used to specify the index
of the first VServer, and the nacltask_num_vsvrs parameter will specify the
number of VServers to create.)  For example, if the vserver parameter is
"my_vsvr", the name of the first VServer will be "my_vsvr_00000.

=item C<< aggregate => $aggregate >>

(Required) As C<NACL::C::Vserver>.  The name of the aggregate where the root
volumes of the VServers will be created.  The aggregate must already exist.

=item C<< nacltask_modify_root_vol => $boolean >>

(Optional) This is a boolean parameter can be used to prevent the method from
automatically increasing the number of files and directories (INODEs) that the
root volume can hold. See the description above.  The default is to
automatically increase the number of files and directories that the root volume
can hold to at least 2000.  To disable this behavior, set this parameter to 0.

=item C<< nacltask_num_subtests => $num_subtests >>

(Optional) This is the number of parallel processes, or subtests, that will be
used to create the VServers.  The number of subtests that can be used is
restricted by the maximum number of SSH connections that can be made.
The default value for this option is 50 subtests.

=item C<< nacltask_num_vsvrs => $num_vservers >>

(Optional) If the vserver parameter contains the VServer name prefix, this
parameter will specify the number of VServers to create.  The default is to
create one VServer.

=item C<< nacltask_start_index => $start_index >>

(Optional) If the vserver parameter contains the VServer name prefix, this
parameter will specify the starting value for the VServer number that will be
appended to the name of the each VServer.  The default value is 0.

=item C<< nacltask_subtest_delay => $num_seconds >>

(Optional) By default, filers are configured to allow a maximum of 10 new SSH
connections per second.  A separate subtest is used to create each volume, and
each subtest starts another SSH connection.  This parameter is the number of
seconds to wait between every "nacltask_subtests_per_second" subtests that are
started.  The default is two seconds.

Note: A value a one second does not always work because even though the
subtests are not started at more than 10 per second, it is possible for more
that 10 requests to arrive at the filer in less than one second.  Setting this
value to two seconds seems to clear up that intermittent failure.  This failure
shows up in the /mroot/etc/log/mlog/messages.log file as the following error
from the xinetd daemon: "Deactivating service ssh due to excessive incoming
connections."

=item C<< nacltask_subtests_per_second => $num_subtests >>

(Optional) This parameter is the number of subtests to start before waiting
"nacltask_subtest_delay" seconds.  The default is 10 subtests.  See also
the nacltask_subtest_delay parameter.

=item C<< nacltask_use_ipv6_address => $nacltask_use_ipv6_address >>

(Optional) This is a boolean parameter.
  - when set to 1, IPv6 address will be used for lif, routing-group creation.
  - when set to 0, IPv4 address will be used for lif, routing-group creation.
  - by default its set to 0.

=back

=back

=head2 purge_multi

    my $num_purged = NACL::MTask::VserverMulti->purge_multi(
        command_interface           => $command_interface,
        vserver                     => $vserver_names,
        nacltask_num_subtests       => $num_subtests,
        nacltask_if_error           => $action,
        nacltask_subtest_delay      => $num_seconds,
        nacltask_subtests_per_second => $num_subtests,
        %other_options
    );

(Class method) This method will purge multiple VServers in a single method
call.  The standard VServer attributes can be passed to this method, in which
case, they will be passed through to the purge() method when each VServer is
purged.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) A component object that represents the host to which to send 
commands.
See L<NACL::C::Component::command_interface|
lib-NACL-C-Component-pm/command_interface>.

=item C<< vserver => $vserver_names >>

(Required) This is either a VServer name or a reference to an array of VServer
names, which are the names of the VServers to purge.  The VServer name(s) may
contain wildcard characters to allow the name to match more than one VServer.
The wildcard characters are those defined by the filer's command line
interpreter.  For example, if you pass:

    vserver => ['eng_vsvrs*', 'mkt_vsvrs*']

This method will purge all the VServers with names that begin with the string
"eng_vsvrs" or "mkt_vsvrs".

=item C<< nacltask_num_subtests => $num_subtests >>

(Optional) This is the number of parallel processes, or subtests, that will be
used to create the VServers.  The number of subtests that can be used is
restricted by the maximum number of SSH connections that can be made.
The default value for this option is 50 subtests.

=item C<< nacltask_if_error => $action >>

(Optional) This parameter specifies the action to take if there is an error
trying to remove one of the VServers.  The possible values are "die" or
"continue".  These strings are not case sensitive.  The default is "die".

=item C<< nacltask_subtest_delay => $num_seconds >>

(Optional) By default, filers are configured to allow a maximum of 10 new SSH
connections per second.  A separate subtest is used to create each volume, and
each subtest starts another SSH connection.  This parameter is the number of
seconds to wait between every "nacltask_subtests_per_second" subtests that are
started.  The default is two seconds.

Note: A value a one second does not always work because even though the
subtests are not started at more than 10 per second, it is possible for more
that 10 requests to arrive at the filer in less than one second.  Setting this
value to two seconds seems to clear up that intermittent failure.  This failure
shows up in the /mroot/etc/log/mlog/messages.log file as the following error
from the xinetd daemon: "Deactivating service ssh due to excessive incoming
connections."

=item C<< nacltask_subtests_per_second => $num_subtests >>

(Optional) This parameter is the number of subtests to start before waiting
"nacltask_subtest_delay" seconds.  The default is 10 subtests.  See also
the nacltask_subtest_delay parameter.

=back

=back

=cut

sub create_multi {
    $Log->enter() if $may_enter;
    my $pkg = shift;
    my %opts = $pkg->_common_validate_with(
        params          => \@_,
        additional_spec => {
            command_interface       => { type => SCALAR,   optional => 0 },
            vserver                 => { type => ARRAYREF | SCALAR, optional => 0 },
            aggregate               => { type => SCALAR,   optional => 0 },
            nacltask_modify_root_vol=> { type => BOOLEAN,  default  => 1 },
            nacltask_num_subtests   => { type => SCALAR,   default  => 50 },
            nacltask_num_vservers   => { type => SCALAR,   default  => 1 },
            nacltask_start_index    => { type => SCALAR,   default  => 0 },
            nacltask_subtest_delay  => { type => SCALAR, default  => 2 },
            nacltask_subtests_per_second => { type => SCALAR, default  => 10 },
            nacltask_use_ipv6_address => { type => SCALAR,  default => 0 },
        },
        allow_extra => 1,
    );

    my $num_subtests            = delete $opts{nacltask_num_subtests};
    my $num_vservers            = delete $opts{nacltask_num_vservers};
    my $start_index             = delete $opts{nacltask_start_index};
    my $subtest_delay           = delete $opts{nacltask_subtest_delay};
    my $subtests_per_second     = delete $opts{nacltask_subtests_per_second};
    my $vserver                 = delete $opts{vserver};
    my $ci                      = $opts{command_interface};
    my @vsvr_names;

    $Log->trace("num_subtests   = $num_subtests");
    $Log->trace("num_vservers   = $num_vservers");
    $Log->trace("start_index    = $start_index");
    $Log->trace("subtest_delay  = $subtest_delay");
    $Log->trace("subtests_per_second = $subtests_per_second");
    $Log->trace("vserver        = $vserver");
    $Log->trace("aggregate      = $opts{aggregate}");

    if ($ci->mode() ne 'CMode') {   # VServers are only used in C-mode.
        $Log->exit() if $may_exit;
        if (wantarray()) {
            return ();
        }
        return 0;
    }

    # The "rootvolume" option is controlled by this subroutine, so make sure it
    # is not specified by the caller.
    if (defined($opts{rootvolume})) {
        $Log->exit() if $may_exit;
        NATE::Exceptions::Argument->throw(
            "The \"rootvolume\" cannot be used with this method.");
    }

    # See if the vserver parameter is the prefix of the names or an array of
    # names.
    if (ref($vserver)) {
        @vsvr_names = @{$vserver};
    } elsif ($num_vservers == 1) {
        @vsvr_names = ($vserver);
    } else {
        foreach my $i (0..$num_vservers - 1) {
            push(@vsvr_names,
                $vserver . '_' . sprintf('%05d', $i + $start_index));
        }
    }

    $num_vservers = scalar(@vsvr_names);

    if ($num_subtests <= 0) {
        $num_subtests = 50;
    }

    my $i = 0;
    while ($i < $num_vservers) {
        #
        # Start a batch of subtests.
        #
        my $num_left    = scalar(@vsvr_names) - $i;
        my $batch_size  = min($num_left, $num_subtests);
        my @subtests    = ();

        foreach (1..$batch_size) {
            my $vsvr_name = $vsvr_names[$i];

            if (($i++ % $subtests_per_second) == 0) {
                # By default, filers are configured to allow a maximum of 10
                # ssh connections per second.  This prevents intermittent
                # connection failures on fast filers.
                sleep $subtest_delay;
            }

            srand; # Give each subtest a different seed.

            $Log->trace("Creating $vsvr_name");
            my $subtest = subtest(\&_create_vserver, -bg, -runid => "cr_$vsvr_name",
                '--',
                $vsvr_name,
                %opts);

            push(@subtests, $subtest);
        }

        #
        # Wait for the current batch of subtests to finish.
        #
        $Log->trace('Waiting for subtests to finish.');
        try {
            Subtest::wait_finish(subtests => \@subtests);
        }
        otherwise {
            my $e = shift;
            $Log->exit() if $may_exit;
            if ($e->isa("NATE::Result::Fatal")) {
                NATE::BaseException->throw("@{$e->{messages}}");
            } else {
                $e->throw();
            }
        };
    }

    $Log->exit() if $may_exit;
    if (wantarray()) {
        # VServers are created in a separate process, so you cannot return the
        # VServer object to the parent process.
        return @vsvr_names;
    }

    return scalar(@vsvr_names);
}

sub purge_multi {
    $Log->enter() if $may_enter;
    my $pkg = shift;
    my %opts = $pkg->_common_validate_with(
        params          => \@_,
        additional_spec => {
            command_interface       => { type => SCALAR, optional => 0 },
            vserver                 => { type => ARRAYREF | SCALAR, optional => 0 },
            nacltask_num_subtests   => { type => SCALAR, default  => 0 },
            nacltask_if_error       => { type => SCALAR, default  => 'die' },
            nacltask_subtest_delay  => { type => SCALAR, default  => 2 },
            nacltask_subtests_per_second => { type => SCALAR, default  => 10 },
        },
        allow_extra => 1,
    );

    my $ci                      = delete $opts{command_interface};
    my $vserver                 = delete $opts{vserver};
    my $num_subtests            = delete $opts{nacltask_num_subtests};
    my $subtest_delay           = delete $opts{nacltask_subtest_delay};
    my $subtests_per_second     = delete $opts{nacltask_subtests_per_second};
    my $if_error                = delete $opts{nacltask_if_error};

    $Log->trace("vserver        = $vserver");
    $Log->trace("num_subtests   = $num_subtests");
    $Log->trace("subtest_delay  = $subtest_delay");
    $Log->trace("subtests_per_second = $subtests_per_second");
    $Log->trace("if_error       = $if_error");

    if ($ci->mode() ne 'CMode') {   # VServers are only used in 7-mode.
        $Log->exit() if $may_exit;
        if (wantarray()) {
            return ();
        }
        return 0;
    }

    if ($num_subtests <= 0) {
        $num_subtests = 50;
    }

    my @vsvrs = _find_vservers($ci, $vserver);
    $Log->trace('Found ' . scalar(@vsvrs) . ' VServers.');

    my $vsvr_id = 0;
    my $purged  = 0;
    while (scalar(@vsvrs)) {
        #
        # Start a batch of subtests.
        #
        my $num_vsvrs   = scalar(@vsvrs);
        my $batch_size  = min($num_vsvrs, $num_subtests);
        my @subtests    = ();

        foreach (1..$batch_size) {
            my $vsvr        = shift(@vsvrs);
            my $vsvr_name   = $vsvr->vserver();
            if (($vsvr_id++ % $subtests_per_second) == 0) {
                # By default, filers are configured to allow a maximum of 10
                # ssh connections per second.  This prevents intermittent
                # connection failures on fast filers.
                sleep $subtest_delay;
            }

            $Log->trace("Removing VServer $vsvr_name");
            $purged++;
            my $subtest = subtest(\&_purge_vserver, -bg, -runid => "rm_$vsvr_name",
                '--',
                $vsvr,
                %opts);

            my $subtest_info = {};
            $subtest_info->{subtest}    = $subtest;
            $subtest_info->{vsvr_name}  = $vsvr_name;

            push(@subtests, $subtest_info);
        }

        #
        # Wait for the current batch of subtests to finish.
        #
        $Log->trace('Waiting for subtests to finish.');
        foreach my $subtest_info (@subtests) {
            my $subtest     = $subtest_info->{subtest};
            my $vsvr_name   = $subtest_info->{vsvr_name};

            try {
                $Log->trace("Waiting for VServer $vsvr_name to be removed.");
                $subtest->wait_finish();
            }
            otherwise {
                if ($if_error =~ /continue/i) {
                    $Log->warn("Purge VServer $vsvr_name failed; continuing.");
                    $purged--;
                } else {
                    my $e = shift;
                    $Log->exit() if $may_exit;
                    if ($e->isa("NATE::Result::Fatal")) {
                        NATE::BaseException->throw("@{$e->{messages}}");
                    } else {
                        $e->throw();
                    }
                }
            };
        }
    }

    $Log->exit() if $may_exit;
    return $purged;
}

########################################################################
#     P R I V A T E   M E T H O D S
########################################################################

sub _create_vserver {
    $Log->enter() if $may_enter;
    my ($vsvr_name, %opts) = @_;

    try {
        my $ci              = delete $opts{command_interface};
        my $modify_root_vol = delete $opts{nacltask_modify_root_vol};
        my $is_ipv6         = delete $opts{nacltask_use_ipv6_address};

        # create the hashes defining the subtasks

        # %opts is going to be the vserver params
        if (!defined($opts{'nm-switch'})) {
            $opts{'nm-switch'} = [ 'file' ];
        }
        if (!defined($opts{'ns-switch'})) {
            $opts{'ns-switch'} = [ 'file' ];
        }
        if (!defined($opts{'rootvolume-security-style'})) {
            $opts{'rootvolume-security-style'} = 'mixed';
        }
        $opts{'rootvolume'} = "root_$vsvr_name";

        my $vsvr;
        try {
            $opts{'vserver'} = $vsvr_name;
            my $aggr_parms = {
                aggregate          => $opts{aggregate},
                nacltask_if_exists => 'reuse',
                diskcount          => '3',
            };
            my $vsvr_host = host($vsvr_name);
            $Log->debug("**** dump of $vsvr_name hostrec: ".Dumper(\$vsvr_host));
            my ( $netif_parms, $rtgrp_parms, $export_rule_parms );

            if ($is_ipv6) {
                my $params = _get_ipv6_params(
                    vserver_host      => $vsvr_host,
                    command_interface => $ci
                );
                $netif_parms       = $params->{netif_parms};
                $rtgrp_parms       = $params->{rtgrp_parms};
                $export_rule_parms = $params->{export_rule_parms};
            } else {
                my $ip_addr = $vsvr_host->DATA_IP1();
                my $netmask = $vsvr_host->DATA_MASK1();
                $Log->trace("**** ip: $ip_addr, netmask: $netmask");
                # See if netmask length is appended to the IP address.
                if ($ip_addr =~ /^\s*(\d+.\d+.\d+.\d+)\/(\d+)/) {
                    $ip_addr = $1;
                    $netmask = _make_netmask($2);
                }
                my $block = new Net::Netmask($ip_addr,$netmask);
                my $bits = $block->bits();
                $Log->trace("**** ip: $ip_addr, netmask: $netmask, bits: $bits");
                $netif_parms = {
                    lif         => $vsvr_name . "_lif0",
                    vserver     => $vsvr_name,
                    role        => 'data',
                    address     => $ip_addr,
                    netmask     => $netmask ||= '255.255.255.0',
                    'home-node' => $vsvr_host->DATA_NODE1() || $ci->node() ,
                    'home-port' => $vsvr_host->DATA_PORT1() || 'e0c',
                };
                { $Data::Dumper::Maxdepth = 2;
                  $Log->trace("**** dump netif_parms: ".Dumper(\$netif_parms));
                };
                $rtgrp_parms = {
                    vserver         => $vsvr_name,
                    'routing-group' => "d" . $block->base() . "/$bits",
                    gateway         => (
                        $vsvr_host->gateway() || $vsvr_host->DATA_GATEWAY1()
                    ),
                    destination => '0.0.0.0/0',
                };
                $export_rule_parms = {
                    policyname  => 'default',
                    clientmatch => '0.0.0.0/0',
                    rorule      => ['any'],
                    rwrule      => ['any'],
                };
            }
            my $nfs_parms = {
                if_exists   => 'reuse',
                v3          => 'true',
                v4          => 'true',
            };
            my $export_policy_parms = {
                policyname  => 'default',
            };

            # make security login param defaults to match old behavior
            my $newpwd = $vsvr_host->default_password() ||
                $ci->hostrec()->default_password() || 'netapp1!';
            my $sec_login_parms = {
                username => $vsvr_host->default_username() || 'vsadmin',
                newpwd   => $newpwd,
                newpwd2  => $newpwd,
                force => 1,
            };

            $Log->trace("Create and setup VServer $vsvr_name.");
            my $vsvrsu = NACL::MTask::VserverSetup->setup(
                command_interface         => $ci,
                vserver_params            => \%opts,
                aggregate_params          => $aggr_parms,
                network_interface_params  => $netif_parms,
                nfs_params                => $nfs_parms,
                routing_grp_params        => $rtgrp_parms,
                export_policy_params      => $export_policy_parms,
                export_policy_rule_params => $export_rule_parms,
                security_login_params     => $sec_login_parms,
            );
            $vsvr = $vsvrsu->{vserver_obj};

            # all of the unix user/group needs to be done still
            $Log->trace("Create the unix user/group for vserver $vsvr_name");
            NACL::C::VserverServicesUnixGroup->create(
                command_interface => $ci,
                vserver => $vsvr_name,
                name => 'root', id => 0,
            );
            NACL::C::VserverServicesUnixUser->create(
                command_interface => $ci,
                vserver => $vsvr_name,
                user => 'root', id => 0, 'primary-gid' => 0,
            );
            NACL::C::VserverServicesUnixGroup->create(
                command_interface => $ci,
                vserver => $vsvr_name,
                name => 'nobody', id => 65534,
            );
            NACL::C::VserverServicesUnixUser->create(
                command_interface => $ci,
                vserver => $vsvr_name,
                user => 'nobody', id => 65534, 'primary-gid' => 65534,
            );
            # and unconditionally modify root volume unix perms:
            $Log->trace("Modifying the root_vol perms for vserver $vsvr_name");
            NACL::STask::Volume->modify(
                command_interface => $ci,
                volume => $vsvr->get_one_state_attribute('rootvolume'),
                'unix-permissions' => '0777',
            );

        } catch NACL::Exceptions::SetupFailure::VserverSetupFailure with {
            my $e = shift;
            $Log->warn( "VserverSetup failed: ".$e->text() );
            $e-> throw();       # just re-throw this exception
        } otherwise {
            # catch any error, assume means no host info.
            # fall back to create bare vserver (old behavior)
            my $e = shift;
            my $text = $e->isa("NATE::BaseException") ? $e->text() : $e;
            $Log->trace("caught and ignored exception: $text");
            if ( ! $vsvr ) {
                $Log->trace("Creating vserver $vsvr_name");
                $vsvr = NACL::STask::Vserver->create(
                    command_interface => $ci,
                    vserver           => $vsvr_name,
                    %opts,
                );
                _setup_vserver_admin($vsvr);
            }
        };


        if ($modify_root_vol) {
            # If the VServer is created with a default root volume, it will be
            # limited to about 550 files and directories.  The following
            # subroutine will make sure the limit is at least 2000.
            $Log->trace('Increasing inodes on root volume.');
            _increase_vsvr_root_inodes($vsvr, 2000);
        }

    }
    catch NATE::BaseException with {
        my $e = shift;
        logresult(
            type    => 'FAIL',
            msg     => "Creating VServer $vsvr_name failed: " . $e->text(),
        );
    };

    $Log->exit() if $may_exit;
    return 0;
}

sub _find_vservers {
    my ($ci, $vsvr_name) = @_;

    # The vsvr_name can be a string or a reference to an array of strings.
    my @vsvr_names = ($vsvr_name);
    if (ref $vsvr_name eq 'ARRAY') {
       @vsvr_names = @$vsvr_name;
    }

    my @vservers;
    foreach my $name (@vsvr_names) {
        try {
            my @found = NACL::STask::Vserver->find(
                command_interface   => $ci,
                filter              => { vserver => $name },
            );

            push(@vservers, @found);
        }
        catch NACL::Exceptions::NoElementsFound with {
        };
    }

    return @vservers;
}

sub _increase_vsvr_root_inodes {
    my ($vsvr, $need) = @_;
    #
    # Make sure the VServer's root volume has enough free INODEs to create the
    # number of files and directories specified by the $need parameter.  A
    # default volume is 20Mb and is limited to about 550 files.
    #
    my $rootvol = NACL::STask::Volume->find(
        command_interface   => $vsvr->command_interface(),
        filter              => { vserver    => $vsvr->vserver(),
                                 volume     => $vsvr->state()->rootvolume(),
                               },
    );

    my $max_files   = $rootvol->state->files();
    my $files_used  = $rootvol->state->files_used();
    my $avail       = $max_files - $files_used;

    if ($avail < $need) {
        $rootvol->modify('files' => $need + $files_used);
        return $need;
    }

    return $avail;
}

sub _make_netmask {
    my ($len) = @_;
    #
    # Given the length in bits of the netmask, create the netmask.
    #
    my $netmask = ~ ((2 ** (32-$len)) - 1);

    my $str = sprintf('%d.%d.%d.%d',
        (($netmask & 0xFF000000)>>24),
        (($netmask & 0x00FF0000)>>16),
        (($netmask & 0x0000FF00)>>8),
        ($netmask & 0x000000FF));

    $Log->trace("netmask = " . sprintf('0x%x (%s)', $netmask, $str));

    return $str;
}

sub _purge_vserver {
    $Log->enter() if $may_enter;
    my ($vsvr, %opts) = @_;

   PURGE: {
     use warnings;
     my $known_error = 0;

     # For burt 742868
     # These are to handle cases where vservers which are peered together are
     # being deleted in parallel. The result is that the same vserver peer
     # will be attempted to be deleted by both parallel processes, which
     # could result in either a VserverPeer::DoesNotExist or a NoElementsFound
     # exception to be thrown.
     try {
       $Log->trace('Removing VServer ' . $vsvr->vserver());
       $vsvr->purge(%opts);
     } catch NACL::C::Exceptions::VserverPeer::DoesNotExist with {
       $known_error = 1;
     } catch NACL::Exceptions::NoElementsFound with {
       $known_error = 1;
     };

     if ($known_error) {
       no warnings;
       redo PURGE;
     }
   }

   $Log->exit() if $may_exit;
}

sub _setup_vserver_admin {
    my ($vsvr) = @_;

    my $hostrec;
    my $username;
    my $password;

    try {
        $hostrec = host($vsvr->vserver());
    } otherwise {};

    if ($hostrec) {
        $username = $hostrec->default_username();
        $password = $hostrec->default_password();
    }

    $username ||= 'vsadmin';
    $password ||= 'netapp1!';

    $Log->trace("Setting up admin (user=$username, pass=$password).");
    try {
        NACL::C::SecurityLogin->password(
            command_interface       => $vsvr->command_interface(),
            vserver                 => $vsvr->vserver(),
            username                => $username,
            newpwd                  => $password,
            force                   => 1,
        );
    } catch NACL::APISet::Exceptions::ResponseException with {
        my $e = shift;
        if ($e->text() !~ /New password must be different than old password./i) {
            $e->throw();
        }
    };

    NACL::C::SecurityLogin->unlock(
        command_interface       => $vsvr->command_interface(),
        vserver                 => $vsvr->vserver(),
        username                => $username,
    );
}

sub _get_ipv6_params {
    $Log->enter() if $may_enter;
    my %opts = @_;
    my %params;
    my $vserver_host = $opts{vserver_host};
    my $ci           = $opts{command_interface};
    my $exclude      = [ { family => 'inet4' } ];
    my $include      = [ { nettype => 'data' } ];
    my @nets         = $vserver_host->net_filter(
        exclude => $exclude,
        include => $include
    );
    my $ip_addr     = $nets[0]->{address};
    my $netmask     = $nets[0]->{netmask};
    my $vsvr_name   = $vserver_host->hostname();
    my $home_port   = $nets[0]->{interface} || 'e0c';
    require NetAddr::IP;
    my $subnet      = new NetAddr::IP->new6( $ip_addr, $netmask )->network();
    my $block_base  = lc $subnet->short();
    my $masklen     = $subnet->masklen();
    my $destination = '::/0';
    my $r_group     = "d" . $block_base . "/$masklen";
    $params{netif_parms} = {
        lif              => $vsvr_name . "_lif0_ipv6",
        vserver          => $vsvr_name,
        role             => 'data',
        address          => $ip_addr,
        'netmask-length' => $netmask,
        'home-node'      => $ci->node(),
        'home-port'      => $home_port,
    };
    $params{rtgrp_parms} = {
        vserver         => $vsvr_name,
        'routing-group' => $r_group,
        gateway         => ( $nets[0]->{gateway} ),
        destination     => $destination,
    };
    $params{export_rule_parms} = {
        policyname  => 'default',
        clientmatch => $destination,
        rorule      => ['any'],
        rwrule      => ['any'],
    };

    $Log->exit() if $may_exit;
    return \%params;
} ## end sub _get_ipv6_params

1;
