#
# Copyright (c) 2001-2013 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary VserverSecurityFileDirectory Task Module
## @author snagaraj@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::MTask::VserverSecurityFileDirectory

=head1 DESCRIPTION

C<NACL::MTask::VserverSecurityFileDirectory> provides a number of well-defined but
potentially complex or multi-step methods related to VserverSecurityFileDirectory in
ONTAP.

The MTask will mainly be concerned with setting up of Vserver Security
file-directory using the parameters that are given below.

=head1 ATTRIBUTES

=head2 vserver

(Required) The name of the Vserver. This is used to create policy, ntfs, sacl,
dacl and policy_task.

=head2 path

(Required) Paths for which sacls has to be set.(Arrayref, even if only a single value is to be sent it)

=head2 ntfs_mode

(Required) A Scalar value of one of these value ( 'propagate'/'ignore'/'replace')

=head2 dacl

(Optional) An Array ref containing hashes with account, access-type,
and other_options similar to call of NACL::STask::VserverSecurityFileDirectoryNtfsDacl->add.

=head2 sacl

(Optional) An Array ref containing hashes with account, access-type,
and other_options similar to call of NACL::STask::VserverSecurityFileDirectoryNtfsSacl->add.

=head2 cleanup

(Optional) A boolean value of either 1 or 0. If '1' the created objects are deleted else if "0" they are not deleted. Default is 1.

=head1 METHODS

=head2 set_ntfs_acl

    # Example usage for VserverSecurityFileDirectory set_ntfs_acl
        NACL::MTask::VserverSecurityFileDirectory->set_ntfs_acl(
            command_interface => $command_interface,
            vserver   => $vserver_name,
            path      => [ $pathname, $pathname1, .. ],
            ntfs_mode => 'propagate'/'ignore'/'replace',
            sacl     => [
                            {
                                'account'       => $account_name1,
                                'access-type'   => 'success'/'failure',
                                %other_options,
                            },
                            {
                                'account'       => $account_name2,
                                'access-type'   => 'success'/'failure',
                                %other_options,
                            },
                        ],
            dacl     => [
                            {
                                'account'       => $account_name1,
                                'access-type'   => 'allow'/'deny',
                                %other_options,
                            },
                            {
                                'account'       => $account_name2,
                                'access-type'   => 'allow'/'deny',
                                %other_options,
                            },
                        ],
            cleanup    => $boolean #default  1
        );

    (Class method) This method performs the setup for the VserverSecurityFileDirectory command in CMode:

     Step 1: Create the VserverSecurityFileDirectory policy on the vserver using
             NACL::STask::VserverSecurityFileDirectoryPolicy->create.
     Step 2: Creates a ntfs object using the
             STask::VserverSecurityFileDirectoryNtfs->create
     Step 3: Creates a list of SACLs using the
             NACL::STask::VserverSecurityFileDirectoryNtfsSacl->add.
     Step 4: Creates a list of DACLs using the
             NACL::STask::VserverSecurityFileDirectoryNtfsDacl->add.
     Step 5: Creates a list of Policy Task using the
             NACL::STask::VserverSecurityFileDirectoryPolicyTask->add.
     Step 6: Applies the Policy using NACL::C::VserverSecurityFileDirectory->apply
     Step 7: Cleans up the created object based on cleanup value.

=over

=item Options

=over

=item C<< vserver => $vserver_name >>

(Required) An scalar value containing the vserver name.

=item C<< 'path' => [$path_name1, $path_name2, ....] >>

(Required) (Arrayref, even if only a single value is to be sent it should
specified as an arrayref containing that single value)

=item C<< 'ntfs_mode' => $ntfs_mode >>

(Required) A Scalar value of one of these value (
'propagate'/'ignore'/'replace')

=item C<< 'sacl' => [ {  } , { } ] >>

(Optional) An Array ref containing one or more hash ref with account, access-type,
and other_options similar to call of NACL::STask::VserverSecurityFileDirectoryNtfsSacl->add.
indicating the parameters used to create a sacl.

=item C<< 'dacl' => [ {  } , { } ] >>

(Optional) An Array ref containing one or more hash ref with account, access-type,
and other_options similar to call of NACL::STask::VserverSecurityFileDirectoryNtfsDacl->add.
indicating the parameters used to create a dacl.

=item C<< cleanup => $cleanup_value >>

(Optional) A boolean value of either 1 or 0. If '1' the created objects are deleted else if "0" they are not deleted. Default is 1.

=cut

package NACL::MTask::VserverSecurityFileDirectory;

use strict;
use warnings;

use base qw(NACL::MTask::MTask NACL::STask::STask );

use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use NACL::STask::VserverSecurityFileDirectory;
use NACL::STask::VserverSecurityFileDirectoryPolicy;
use NACL::STask::VserverSecurityFileDirectoryPolicyTask;
use NACL::STask::VserverSecurityFileDirectoryNtfs;
use NACL::STask::VserverSecurityFileDirectoryNtfsSacl;
use NACL::STask::VserverSecurityFileDirectoryNtfsDacl;
use NACL::STask::STask;
use NATE::Exceptions::Argument qw(:try);
use NACL::APISet::Exceptions::ResponseException;
use NACL::APISet::Exceptions::CommandFailedException qw(:try);
use Params::Validate qw/validate_with SCALAR ARRAYREF BOOLEAN HASHREF/;
use NACL::ComponentUtils qw(_optional_scalars Dumper);

my @objs;

############################################################################################################
# Name        : setup
# Description : Actual task function to be invoked by test scripts. POD above has complete details
#               Create File-Directory policy, ntfs, sacl, dacl, task, and then
#               apply the policy
############################################################################################################
sub set_ntfs_acl {

    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    NATE::BaseException->throw(
        "Setup can only be invoked with a proper package call")
        if ref $pkg;

    my %opts = validate_with(

        params => \@args,
        spec   => {
            %{NACL::C::Component->_common_validate_spec()},
            vserver   => {type => SCALAR,},
            path      => {type => ARRAYREF,},
            ntfs_mode => {type => SCALAR,},
            sacl      => {type => ARRAYREF, optional => 1, default => [],},
            dacl      => {type => ARRAYREF, optional => 1, default => [],},
            cleanup   => {type => SCALAR, optional => 1, default => 1},
        },
        allow_extra => 1,
    );

    my %common_args;
    $pkg->_copy_common_component_params_with_ci(
        source => \%opts,
        target => \%common_args,
    );

    # extract necessary details from the object
    my $command_interface = $opts{command_interface};
    my $vserver           = $opts{vserver};
    my @paths             = @{$opts{path}};
    my $ntfs_mode         = $opts{ntfs_mode};
    my @sacl              = @{$opts{sacl}};
    my @dacl              = @{$opts{dacl}};
    my $cleanup           = $opts{cleanup};

    NATE::BaseException->throw("This method should be called only for CMode")
        if $command_interface->mode() ne 'CMode';

    my $obj = {};
    my $policy_objs;
    my $policy = $pkg->random_name_generator();
    my $sd     = $pkg->random_name_generator();

    try {
        $policy_objs
            = NACL::STask::VserverSecurityFileDirectoryPolicy->create(
            %common_args,
            'policy-name'   => $policy,
            vserver         => $vserver,
            nacltask_verify => 1
            );

        $$obj{policy} = $policy_objs;

        my $sd_obj = NACL::STask::VserverSecurityFileDirectoryNtfs->create(
            %common_args,
            'ntfs-sd'       => $sd,
            vserver         => $vserver,
            nacltask_verify => 1
        );
        $$obj{ntfs_sd} = $sd_obj;

        foreach my $sacl_t (@sacl) {
            NACL::STask::VserverSecurityFileDirectoryNtfsSacl->add(
                %common_args,
                'ntfs-sd' => $sd,
                %{$sacl_t},
                vserver         => $vserver,
                nacltask_verify => 1
            );
        }

        foreach my $dacl_t (@dacl) {
            NACL::STask::VserverSecurityFileDirectoryNtfsDacl->add(
                %common_args,
                'ntfs-sd' => $sd,
                %{$dacl_t},
                vserver         => $vserver,
                nacltask_verify => 1
            );
        }

        my $index = 100;
        foreach my $path (@paths) {
            NACL::STask::VserverSecurityFileDirectoryPolicyTask->add(
                %common_args,
                'path'          => $path,
                'ntfs-sd'       => [$sd],
                'ntfs-mode'     => $ntfs_mode,
                vserver         => $vserver,
                'policy-name'   => $policy,
                'index'         => $index,
                nacltask_verify => 1
            );
        }

        NACL::STask::VserverSecurityFileDirectory->apply(
            command_interface => $command_interface,
            'vserver'         => $vserver,
            'policy-name'     => $policy_objs,
        );
    }
    catch NATE::BaseException with {
        my $exception = shift;
        push(@objs, $obj);
        NACL::MTask::VserverSecurityFileDirectory->cleanup();
        $Log->exit() if $may_exit;
        $exception->throw();
    };
    if ($cleanup == 1) {
        cleanup($obj);
    } else {
        push(@objs, $obj);
    }
    $Log->exit() if $may_exit;

} ## end sub set_ntfs_acl

=head2 cleanup

    NACL::MTask::VserverSecurityFileDirectory->cleanup();

    (Class or Instance Method) Performs the cleanup operation for
    the VserverSecurityFileDirectory set_ntfs_acl

    Step 1: Deletes the ntfs_sd.
    Step 2: Deletes the policy.

    Uses a CMode CLI.

=back

=cut

############################################################################################################
# Name        : cleanup
# Description : Deletes the policy and ntfs-sd created.
#
############################################################################################################

sub cleanup {
    my $obj = shift;
    unless ($obj eq "NACL::MTask::VserverSecurityFileDirectory") {
        if (defined($$obj{ntfs_sd})) {
            $$obj{ntfs_sd}->purge();
        }
        if (defined($$obj{policy})) {
            $$obj{policy}->purge();
        }
    } else {
        $Log->enter() if $may_enter;
        foreach $obj (@objs) {
            if (defined($$obj{ntfs_sd})) {
                $$obj{ntfs_sd}->purge();
            }
            if (defined($$obj{policy})) {
                $$obj{policy}->purge();
            }
        }
        @objs = undef;
        $Log->exit() if $may_exit;
    }
} ## end sub cleanup

1;
