# Copyright (c) 2012 Network Appliance, Inc.
# All rights reserved.

## @summary Setup Evaluation Matcher
## @author  Adam.Ciapponi@netapp.com
## @pod     here

=head1 NAME

NACL::Matchers::EvaluationMatcher

=head1 DESCRIPTION

This module implements a matcher that determines whether it matches by calling eval

=head1 METHODS

=cut

package NACL::Matchers::EvaluationMatcher;

# Compiler directives.
use strict;
use warnings;
use CR::Setup::Constants;
use CR::Setup::Exception;
use Data::Dumper;
use Params::Validate qw(:all);
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use base qw(NACL::Matcher);

=head2 matches

Builds up a string to evaluate comparisons/expressions and return a boolean value.

=cut

sub matches {
    validate_pos(@{[@_[1..$#_]]}, NACL::Matcher::matcher_value_validate_spec());
    $_[0]->_set_state_from_matcher_value( matcher_value => $_[1], which_value => 'actual' );

    my $str = $_[0]->{'prefix'};
    if ( $_[0]->{'actual_dereferenced'} && ref( $_[0]->{'actual'} ) ) {
        $str .= $_[0]->get_dereference_char('actual');
        $str .= '{$_[0]->{\'actual\'}}';
    }
    else {
        $str .= '$_[0]->{\'actual\'}';
    }
    if ( $_[0]->{'expected_in_expression'} ) {
        $str .= $_[0]->{'middle'};
        if ( $_[0]->{'expected_dereferenced'} && ref( $_[0]->{'expected'} ) )
        {
            $str .= $_[0]->get_dereference_char('expected');
            $str .= '{$_[0]->{\'expected\'}}';
        }
        else {
            $str .= '$_[0]->{\'expected\'}';
        }
    }
    $str .= $_[0]->{'suffix'};
    if ( eval($str) ) {
        # eval returns undef on exception so it can't be true
        # and have an exception
        $TRUE;
    }
    else {
        my $e = $@;
        if ( $e ) {
            CR::Setup::Exception->throw('Eval threw exception: ' . Dumper($e));
        } else {
            $FALSE;
        }
    }
}

=head2 with_prefix

Sets the prefix for the eval string before actual.

=cut

my @with_prefix_validate_spec = (
    { type => SCALAR }
);
sub with_prefix {
    validate_pos(@{[@_[1..$#_]]}, @with_prefix_validate_spec);
    $_[0]->{'prefix'} = $_[1];
    $_[0];
}

=head2 with_middle

Sets the middle part for the eval string between actual and expected. Only used if expected is in the expression.

=cut

my @with_middle_validate_spec = (
    { type => SCALAR }
);
sub with_middle {
    validate_pos(@{[@_[1..$#_]]}, @with_middle_validate_spec);
    $_[0]->{'middle'} = $_[1];
    $_[0];
}

=head2 with_suffix

Sets the suffix for the eval string after actual or after expected if it is in the expression.

=cut

my @with_suffix_validate_spec = (
    { type => SCALAR }
);
sub with_suffix {
    validate_pos(@{[@_[1..$#_]]}, @with_suffix_validate_spec);
    $_[0]->{'suffix'} = $_[1];
    $_[0];
}

=head2 with_expected_in_expression

Sets a flag to use the expected value in the eval string.

=cut

sub with_expected_in_expression {
    validate_pos(@{[@_[1..$#_]]});
    $_[0]->{'expected_in_expression'} = $TRUE;
    $_[0];
}

=head2 with_actual_dereferenced

Sets a flag to dereference the actual value if it is a reference.

=cut

sub with_actual_dereferenced {
    validate_pos(@{[@_[1..$#_]]});
    $_[0]->{'actual_dereferenced'} = $TRUE;
    $_[0];
}

=head2 with_expected_dereferenced

Sets a flag to dereference the expected value if it is a reference.

=cut

sub with_expected_dereferenced {
    validate_pos(@{[@_[1..$#_]]});
    $_[0]->{'expected_dereferenced'} = $TRUE;
    $_[0];
}

=head2 get_dereference_char

Gets the character needed to dereference a property of the matcher (actual/expected).

=cut

my @get_dereference_char_validate_spec = (
    { type => SCALAR }
);
sub get_dereference_char {
    validate_pos(@{[@_[1..$#_]]}, @get_dereference_char_validate_spec);
    my $char;
    if ( ref( $_[0]->{$_[1]} ) eq 'SCALAR' ) {
        $char = '$';
    }
    elsif ( ref( $_[0]->{$_[1]} ) eq 'ARRAY' ) {
        $char = '@';
    }
    elsif ( ref( $_[0]->{$_[1]} ) eq 'HASH' ) {
        $char = '%';
    # CODE refs are handled by decoderef in MatcherTarget
    } else {
        CR::Setup::Exception->throw('Cannot dereference type: ' . ref($_[0]->{$_[1]}));
    }
    $char;
}
1;
