
# Copyright (c) 2012 Network Appliance, Inc.
# All rights reserved.

## @summary Setup Mock Object Matcher
## @author  Adam.Ciapponi@netapp.com
## @pod     here

=head1 NAME

NACL::Matchers::MockObjectMatcher

=head1 DESCRIPTION

This module implements a matcher that determines whether it matches by looking at whether or not the specified calls were made to the Test::MockObject.

=head1 METHODS

=cut

package NACL::Matchers::MockObjectMatcher;

# Compiler directives.
use strict;
use warnings;
use CR::Setup::Constants;
use CR::Setup::Exception;
use Data::Dumper;
use Params::Validate qw(:all);
use Scalar::Util qw(blessed);
use Test::Deep::NoTest qw(eq_deeply);
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();

use base qw(NACL::Matcher);

=head2 matches

Determines matching against whether or not the mock object was called correctly and returns a boolean value.

=cut

sub matches {
    validate_pos( @{[@_[1..$#_]]}, NACL::Matcher::matcher_value_validate_spec() );
    $_[1] = $_[0]->_set_state_from_matcher_value( matcher_value => $_[1], which_value => 'actual' );

    my $is_matcher_value_a_mock_object = blessed($_[1]) && $_[1]->can('_calls');
    if ( ! $is_matcher_value_a_mock_object ) {
        CR::Setup::Exception->throw(join("\n", 'Target specified for MockObjectMatcher was not a mock object:', Dumper($_[1])));
    }

    $_[0]->{'actual'} = {};

    # Semantics for defaults
    if ( defined $_[0]->{'qualifier'} ) {
        # If number is defined, everything is provided so there is no need to do anything
        # The number undefined case is not possible without violation of encapsulation as
        # input validation throws an exception if undef is provided
    } else {
        if ( defined $_[0]->{'number'} ) {
            # This is looking for a specific number of times
            # Have->received('blah')->one()->time()
            $_[0]->{'qualifier'} = 'exactly';
        } else {
            # This is essentially looking for whether or not it was called at all
            # Have->received('blah')
            $_[0]->{'qualifier'} = 'at least';
            $_[0]->{'number'} = 1;
        }
    }

    my $method = $_[0]->{'expected'};
    my $qualifier = $_[0]->{'qualifier'};
    my $number = $_[0]->{'number'};
    my $args = $_[0]->{'args'};

    $_[0]->{'expected'} = {
        method => $method,
        qualifier => $qualifier,
        number => $number,
        args => $args,
    };

    if ($qualifier eq 'at position') {
        my $call = $_[1]->_calls->[$number];
        if (!defined $call) {
            $_[0]->{'actual'}->{'call at position'} = 'No call at position ' . $number;
            $FALSE;
        } else {
            $_[0]->{'actual'}->{'call at position'} = $call;
            my $actual_method = $call->[0];
            my @actual_args = @{$call->[1]};
            shift @actual_args; #self
            $_[0]->{'actual'}->{'number'} = $number;
            $_[0]->{'actual'}->{'method'} = $actual_method;
            if ($method eq $actual_method) {
                if ( defined $args ) {
                    if ( ref($args) eq 'ARRAY' ) {
                        $_[0]->{'actual'}->{'args'} = \@actual_args;
                        if (eq_deeply($args, \@actual_args)) {
                            $TRUE;
                        } else {
                            $FALSE;
                        }
                    } else {
                        # HASH
                        my %actual_args_hash = @actual_args;
                        $_[0]->{'actual'}->{'args'} = \%actual_args_hash;
                        if (eq_deeply($args, \%actual_args_hash)) {
                            $TRUE;
                        } else {
                            $FALSE;
                        }
                    }
                } else {
                    $TRUE;
                }
            } else {
                $FALSE;
            }
        }
    } else {
        my @calls = @{ $_[1]->_calls };
        $_[0]->{'actual'}->{'call history'} = \@calls;
        my $num_times = 0;
        foreach my $call (@calls) {
            my $actual_method = $call->[0];
            my @actual_args = @{$call->[1]};
            shift @actual_args; #self
            if ($method eq $actual_method) {
                if ( defined $args ) {
                    if ( ref($args) eq 'ARRAY' ) {
                        if (eq_deeply($args, \@actual_args)) {
                            $num_times++;
                        } else {
                            # Doesn't Match
                            next;
                        }
                    } else {
                        # HASH
                        my %actual_args_hash = @actual_args;
                        if (eq_deeply($args, \%actual_args_hash)) {
                            $num_times++;
                        } else {
                            # Doesn't Match
                            next;
                        }
                    }
                } else {
                    $num_times++;
                }
            } else {
                # Doesn't Match
                next;
            }
        }
        $_[0]->{'actual'}->{'number'} = $num_times;
        if ($qualifier eq 'exactly') {
            if ($num_times == $number) {
                $TRUE;
            } else {
                $FALSE;
            }
        } elsif ($qualifier eq 'at most') {
            if ($num_times <= $number) {
                $TRUE;
            } else {
                $FALSE;
            }
        } else {
            # at least
            if ($num_times >= $number) {
                $TRUE;
            } else {
                $FALSE;
            }
        }
    }
}

=head2 at_most

Sets the qualifier for the number of matching calls to be at most the specified value.

=cut

my @at_most_validate_spec = ({ type => SCALAR });
sub at_most {
    validate_pos( @{[@_[1..$#_]]}, @at_most_validate_spec);
    $_[0]->{'number'} = $_[1];
    $_[0]->{'qualifier'} = 'at most';
    $_[0];
}

=head2 at_least

Sets the qualifier for the number of matching calls to be at least the specified value.

=cut

my @at_least_validate_spec = ({ type => SCALAR });
sub at_least {
    validate_pos( @{[@_[1..$#_]]}, @at_least_validate_spec);
    $_[0]->{'number'} = $_[1];
    $_[0]->{'qualifier'} = 'at least';
    $_[0];
}

=head2 exactly

Sets the qualifier for the number of matching calls to be exactly the specified value.

=cut

my @exactly_validate_spec = ({ type => SCALAR });
sub exactly {
    validate_pos( @{[@_[1..$#_]]}, @exactly_validate_spec);
    $_[0]->{'number'} = $_[1];
    $_[0]->{'qualifier'} = 'exactly';
    $_[0];
}

=head2 zero

For convenience, sets the number of matching calls or the position of the call to 0.

=cut

sub zero {
    validate_pos(@{[@_[1..$#_]]});
    $_[0]->{'number'} = 0;
    $_[0];
}

=head2 one

For convenience, sets the number of matching calls or the position of the call to 1.

=cut

sub one {
    validate_pos(@{[@_[1..$#_]]});
    $_[0]->{'number'} = 1;
    $_[0];
}

=head2 two

For convenience, sets the number of matching calls or the position of the call to 2.

=cut

sub two {
    validate_pos(@{[@_[1..$#_]]});
    $_[0]->{'number'} = 2;
    $_[0];
}

=head2 three

For convenience, sets the number of matching calls or the position of the call to 3.

=cut

sub three {
    validate_pos(@{[@_[1..$#_]]});
    $_[0]->{'number'} = 3;
    $_[0];
}

=head2 time

Does nothing but is here so that the matcher can be used in a manner that is a proper english statement.

=cut

sub time {
    validate_pos(@{[@_[1..$#_]]});
    $_[0];
}

=head2 times

Does nothing but is here so that the matcher can be used in a manner that is a proper english statement.

=cut

sub times {
    validate_pos(@{[@_[1..$#_]]});
    $_[0];
}

=head2 with_args

Sets the args that need to be matched for the specified call.

=cut

my @with_args_validate_spec = ({ type => ARRAYREF|HASHREF });
sub with_args {
    validate_pos( @{[@_[1..$#_]]}, @with_args_validate_spec);
    $_[0]->{'args'} = $_[1];
    $_[0];
}

=head2 at_position

Sets the qualifier for the call to be at the specified position. This results in only single call comparison.

=cut

my @at_position_validate_spec = ({ type => SCALAR });
sub at_position {
    validate_pos( @{[@_[1..$#_]]}, @at_position_validate_spec);
    $_[0]->{'number'} = $_[1];
    $_[0]->{'qualifier'} = 'at position';
    $_[0];
}
1;
