#
# Copyright (c) 2017 NetApp, Inc., All Rights Reserved.
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

package NACL::Plugin::Driver::Resource::Clusters;
use strict;
use warnings;

use Carp             qw(confess);
use File::Spec::Functions qw(catdir);
use Params::Validate qw(validate SCALAR);

use NATE::Util::Param qw(get_param_enumeration_list);

use NACL::STask::Node;
use NACL::C::Cluster;
use NACL::Plugin::Driver::Resource::Util qw(redirect_logger reset_loggers);

use parent qw(NATE::Driver::Plugin::Resource);

use constant RESOURCE_TYPE_PARAMETER_NAME => 'CLUSTERS';

sub new {
    my $class = shift;
    my $self = {};
    my $obj = bless $self, $class;
    return $obj;
}

##########################################################################
# This method discovers nodes and will be grouped by generic cluster name.
# Nodes will be assigned to relative cluster parameter names
# like 'CLUSTER1',...CLUSTERN and CLUSTERA,...CLUSTERZ.
#########################################################################
sub discover_resources {
    my $self = shift;
    redirect_logger(resource_obj => $self);
    $self->{_log}->log("Started getting clusters info...");
    my @nodes = ();
    eval {
        @nodes = NACL::STask::Node->find();
    };
    if (my $error = $@) {
       $self->_confess("Encountered error while getting nodes: $error");
    }
    my $clusters_max_count = scalar(@nodes);
    eval {
        # cluster parameter enumeration list
        # CLUSTER1, CLUSTER2,..., CLUSTERN
        my $cluster_params_alpbt_suffix = get_param_enumeration_list(parameter_name => $self->RESOURCE_TYPE_PARAMETER_NAME,
                                                                parameter_count     => $clusters_max_count,
                                                                parameter_suffix    => NATE::Util::Param::NUMBER_SUFFIX
        );
        # CLUSTERA, CLUSTERB,..., CLUSTERZ
        my $cluster_params_num_suffix = get_param_enumeration_list(parameter_name   => $self->RESOURCE_TYPE_PARAMETER_NAME,
                                                                parameter_count     => $clusters_max_count,
                                                                parameter_suffix    => NATE::Util::Param::ALPHABET_SUFFIX
        );

        shift(@{$cluster_params_alpbt_suffix});   # remove first item 'CLUSTER'
        pop(@{$cluster_params_alpbt_suffix});     # remove last item 'CLUSTERS'
        shift(@{$cluster_params_num_suffix});
        pop(@{$cluster_params_num_suffix});
        my %cluster_first_node = ();
        my $i = 0;
        foreach my $node(@nodes) {
            my $cg_first_node = NACL::C::Cluster->find( command_interface => $node );
            if(!exists($cluster_first_node{$cg_first_node->name})) {
                my @cg_nodes = NACL::C::Cluster->nodes( command_interface => $node );
                $self->{cluster_nodes_index}->{$i} = \@cg_nodes;
                $self->{cluster_param_index_map}->{$cluster_params_alpbt_suffix->[$i]} = $i;
                $self->{cluster_param_index_map}->{$cluster_params_num_suffix->[$i]} = $i;
                # Used cluster name same as cluster parameter.
                push(@{$self->{cluster_param_cluster_name_map}}, {$cluster_params_num_suffix->[$i] => $cluster_params_num_suffix->[$i]});
                $self->{cluster_param_index_map}->{'CLUSTER'} = $i if($i == 0);
                $cluster_first_node{$cg_first_node->name} = $i;
                $i++;
            }
        }
        $self->{_log}->log("Completed getting clusters information.");
    };
    if (my $error = $@) {
       $self->_confess("Encountered error discovering clusters: $error");
    }
    $self->{_log}->log("Completed getting clusters info.");
    reset_loggers(resource_obj => $self);
}

#########################################################################
# This method needs to be implemented when cluster resource_type
# allocation is required to be supported directly in NATE Driver testcases.
# Currently cluster nodes allocation is supported via NODES resource type.
#########################################################################
sub get_resources {
    my $self = shift;
    my %opts = @_;
    my $rc_type = $opts{resource_type}//'';
    my $alloc_scheme = $opts{allocation_scheme}//'';
    my $plugin_opts = $opts{resource_plugin_opts}//{};
    my $testname = $opts{testname}//'';
    my $driver = $opts{driver}//undef;
    $self->_confess("get_resources() - Method not implemented\n");
}

sub get_all_resources {
    my $self = shift;
    return $self->_find_resources();
}

sub get_resource_type_parameter_list {
    my $self = shift;
    my %opts = validate(@_, {
        parameters_count => {type => SCALAR},
    });
    $self->_confess("get_resource_type_parameter_list() - Method not implemented\n");
}

#########################################################################
# Returns cluster parameter names
# [cluster1,..,clusterN, clusterA,..,clusterZ]
#########################################################################
sub _get_cluster_parameters {
    my $self = shift;
    my @param_list = ();
    if(defined($self->{cluster_param_index_map})) {
        foreach my $param(keys(%{$self->{cluster_param_index_map}})) {
            push(@param_list, $param);
        }
    }
    return \@param_list;
}

###################################################################
# Input: param_name - A single cluster parameter like
#        CLUSTER, CLUSTERS, CLUSTER1, CLUSTERA
# Return: An array ref containing member nodes of a cluster group.
###################################################################
sub _find_resource {
    my $self = shift;
    my %opts = validate (@_, {
            resource_name  => {tyep => SCALAR},
        }
    );
    my $cluster_param = $opts{resource_name};
    $cluster_param =~ s/^\s+|\s+$//;
    $cluster_param = uc($cluster_param);
    if(exists($self->{cluster_param_index_map}->{$cluster_param})) {
        my $cluster_index = $self->{cluster_param_index_map}->{$cluster_param};
        return $self->{cluster_nodes_index}->{$cluster_index};
    }elsif($cluster_param =~ m/CLUSTERS|all/i) {
        return $self->_find_resources();
    }else {
        return [];
    }
}

###################################################################
# Helper method get nodes belonging to all of a clusters.
###################################################################
sub _find_resources {
    my $self = shift;
    my $clusters = $self->_get_all_clusters();
    my @all_nodes = ();
    foreach my $c(@{$clusters}) {
        my $cg_nodes = $self->_find_resource(resource_name => $c);
        push(@all_nodes, @{$cg_nodes});
    }
    return \@all_nodes;
}
###################################################################
# Method to names of all the cluster groups.
###################################################################
sub _get_all_clusters {
    my $self = shift;
    my @clusters = ();
    my $cluster_maps = $self->{cluster_param_cluster_name_map}//[];
    foreach my $cluster_map(@{$cluster_maps}) {
        while(my ($cluster_param, $cluster_name) = each(%{$cluster_map})) {
            push(@clusters, $cluster_name);
        }
    }
    return \@clusters;
}

###################################################################
# Helper method to validate cluster parameter
# Input: Cluster parameter like cluster<N> or cluster<A|B|..Z>
###################################################################
sub _is_valid_cluster_param {
    my $cluster = shift;
    my @clstrs = split(',', $cluster);
    my $is_valid = 1;
    foreach my $c(@clstrs) {
        if($c !~ m/^cluster\d+$|^cluster[a-z]$|^all$/i) {
            $is_valid = 0;
            last;
        }
    }
    return $is_valid
}

###################################################################
# Before throwing an exception, reset logger back to previous state
###################################################################
sub _confess {
    my $self = shift;
    my $error = shift;
    reset_loggers(resource_obj => $self);
    confess $error;
}

1;