#
# Copyright (c) 2017 NetApp, Inc., All Rights Reserved.
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

package NACL::Plugin::Driver::Resource::Nodes;
use strict;
use warnings;

use autouse 'Data::Dumper' => qw(Dumper);
use Carp             qw(confess);
use File::Spec::Functions qw(catdir);
use Params::Validate qw(validate SCALAR HASHREF ARRAYREF);

use Hostrec;
use NATE::Util::Param qw(get_param_enumeration_list find_nodes);
use NACL::Plugin::Driver::Resource::Util qw(redirect_logger reset_loggers);


use parent qw(NATE::Driver::Plugin::Resource);

use constant RESOURCE_TYPE_PARAMETER_NAME => 'NODES';
use constant RESOURCE_TYPE_PARAMETER_NAME_ALIAS1 => 'FILERS';
use constant DEFAULT_GROUP => 'all_nodes';
use constant NO_RESOURCE => 'no_resource'; # should have same value as NATE::Driver::NO_RESOURCE

sub new {
    my $class = shift;
    my $self = { is_clusters_discovered => 0, is_nodes_discovered => 0, nodes_count => {}};
    my $obj = bless $self, $class;
    return $obj;
}

##########################################################################
# This method discovers all the nodes. If 'clusters' parameter value
# is defined then, it segregates discovered nodes by cluster names.
# Relative cluster names like 'cluster1',...'cluster<N>
# Input: 'clusters' - Comma separated relative cluster names.
# (The individual resource type item specification defined in NATE Driver
# testcase 'resources' will be passed.)
##########################################################################
sub discover_resources {
    my $self = shift;
    my %opts = @_;
    my $cluster = $opts{clusters}//undef;
    my $log_context = $opts{log_context}//'nodes_discovery';
    redirect_logger(resource_obj => $self);
    if(defined($cluster)) {
        if(_is_valid_cluster_param($cluster)) {
            if(!$self->{is_clusters_discovered}) {
                require NACL::Plugin::Driver::Resource::Clusters;
                my $clusters_obj = NACL::Plugin::Driver::Resource::Clusters->new();
                $clusters_obj->discover_resources();
                my $c_param_names = $clusters_obj->_get_cluster_parameters();
                if(scalar(@{$c_param_names})) {
                    foreach my $c_name(@{$c_param_names}) {
                        my $nodes = $clusters_obj->_find_resource(resource_name => $c_name);
                        $self->_construct_parameter_maps(nodes => $nodes, cluster => $c_name);
                    }
                }
                $self->{is_clusters_discovered} = 1;
            }else {
                $self->{_log}->log("$log_context: Clusters '$cluster' resources already discovered");
            }
        }else {
            $self->confess("$log_context: Invalid clusters value '$cluster'\n"
            ."Resource item specification:\n"
            .Dumper(\%opts)."\n"
            );
        }
    }

    if(!$self->{is_nodes_discovered}) {
        my $nodes = find_nodes();
        $self->_construct_parameter_maps(nodes => $nodes);
        $self->{is_nodes_discovered} = 1;
    }else {
        $self->{_log}->log("Default resources already discovered");
    }
    reset_loggers(resource_obj => $self);
}

##########################################################################
# Returns an array reference of the available nodes.
##########################################################################
sub get_all_resources {
    my $self = shift;
    my $nodes = find_nodes();
    return $nodes;
}

##########################################################################
# Gets an array ref of parameter enumeration for the resource type NODES
# Return: [FILER, FILERA,...FILERZ, FILER_LIST, NODES]
##########################################################################
sub get_resource_type_parameter_list {
    my $self = shift;
    my %opts = validate(@_, {
        parameters_count => {type => SCALAR},
    });
    my $nodes_count = $opts{parameters_count};
    my @parameter_enum_list = ();
    if($nodes_count > 0) {
        # FILERS parameter enumeration list
        # FILER, FILERA, FILERB,..., FILERZ, FILERS
        my $filers_params_alpbt_suffix = get_param_enumeration_list(parameter_name  => $self->RESOURCE_TYPE_PARAMETER_NAME_ALIAS1,
                                                                parameter_count     => $nodes_count,
                                                                parameter_suffix    => NATE::Util::Param::ALPHABET_SUFFIX
        );
        my $filer_list = 'FILER_LIST';
        push(@parameter_enum_list, @{$filers_params_alpbt_suffix});
        push(@parameter_enum_list, $filer_list);
        push(@parameter_enum_list, $self->RESOURCE_TYPE_PARAMETER_NAME);
    }else {
        $self->_confess("'parameters_count' - arguement value should be > 0\n");
    }
    return \@parameter_enum_list;
}

##########################################################################
# This methods gets the NODE resource values based on the specification
# for 'allocation_scheme' and 'clusters' group.
##########################################################################
sub get_resources {
    my $self = shift;
    my %opts = validate(@_, {
        resource_type           => {type => SCALAR},
        allocation_scheme       => {type => SCALAR},
        resource_plugin         => {type => SCALAR, optional => 1},
        clusters                => {type => SCALAR, optional => 1},
        log_context             => {type => SCALAR,  optional => 1},
    });
    my @resources = ();
    my $rc_type = $opts{resource_type};
    my $alloc_scheme = $opts{allocation_scheme};
    my $clusters_filter = $opts{clusters}//'';
    my $testname = $opts{log_context}//'no_context';
    $alloc_scheme =~s/^\s+|\s+$//g;

    # allocation_scheme => none|any|all|node<N>|node<A|B|...|Z>
    # clusters          => { cluster => cluster<N>|cluster<A|B|..|Z>}
    # 1. allocation_scheme == 'none' then, no node need to be allocated.
    # 2. Get nodes based on input filter in 'clusters' and perform
    #    allocation using those nodes.
    # 3. allocation_scheme == 'any', select a random node.
    # 4. allocation_scheme == 'all', select all nodes.
    # 5. allocation_scheme == 'node<N>|node<A|,..|Z>, select a Nth node from nodes set.
    if(_is_valid_node_param($alloc_scheme)) {
        if($alloc_scheme =~ m/^none$/i) {
            push(@resources, $self->NO_RESOURCE);
        }else {
            my $resources_new = [];
            if($clusters_filter) {
                my @clusters = split(',', $clusters_filter);
                foreach my $clusterN(@clusters) {
                    $clusterN =~s/^\s+|\s+$//g;
                    my $c_grp_nodes = $self->_find_resources(cluster => $clusterN);
                    if(scalar(@{$c_grp_nodes}) == 0) {
                        $self->_confess("$testname: Not able to find any nodes.\n"
                        ."Resource spec:\n".Dumper(\%opts)."\n"
                        ."Available resources:\n".Dumper($self->get_resources_info())."\n"
                        );
                    }
                    push(@{$resources_new}, @{$c_grp_nodes});
                }
            }else {
                $resources_new = $self->get_all_resources();
            }
            if(scalar(@{$resources_new}) == 0) {
                $self->_confess("$testname: Not able to find any nodes.\n"
                ."Resource spec:\n".Dumper(\%opts)."\n"
                ."Available resources:\n".Dumper($self->get_resources_info())."\n"
                );
            }
            @resources = ();
            if($alloc_scheme =~ m/^any$/i) {
                if(scalar(@{$resources_new})) {
                    my $index = int(rand(scalar(@{$resources_new})));
                    my $any_rc = $resources_new->[$index];
                    push(@resources, $any_rc);
                }else {
                    $self->_confess("Testcase '$testname': No nodes found to assign to testcase for allocation_scheme '".$alloc_scheme."'\n"
                    ."Resource spec:\n".Dumper(\%opts)."\n"
                    );
                }
            }elsif($alloc_scheme =~ m/^all$/i) {
                push(@resources, @{$resources_new});
            }elsif($alloc_scheme =~ m/^node\d+$|node[a-z]$/i) {
                my $node = '';
                my %resources = ();
                my $index = 1;
                # 52 nodes parameters.
                my @suffix = ('A'..'Z', 'AA'..'AZ');
                if(scalar(@{$resources_new}) > scalar(@suffix)) {
                   $self->_confess("Testcase '$testname': Nodes allocation parameter is out of the range.\n"
                   ."Supoorted range is: nodeA,..,nodeZ, nodeAA,.., nodeAZ or node1,.., node52\n"
                   ."Resource spec:\n".Dumper(\%opts)."\n"
                   );
                }
                foreach my $nd(@{$resources_new}) {
                    my $param1 = 'NODE'.$index;
                    my $param2 = 'NODE'.$suffix[$index-1];
                    $resources{$param1} = $nd;
                    $resources{$param2} = $nd;
                    $index++;
                }
                my $node_select = uc($alloc_scheme);
                if(exists($resources{$node_select})) {
                    push(@resources, $resources{$node_select});
                }else {
                    $self->_confess("Testcase '$testname': Not able to find node value for '".$alloc_scheme."'\n"
                    ."Resource spec:\n".Dumper(\%opts)."\n"
                    );
                }
            }else {
                $self->_confess("Testcase '$testname': Invalid '".$rc_type."' resource type allocation scheme.\n"
                ."Resource spec:\n".Dumper(\%opts)."\n"
                );
            }
        }
    }else {
        $self->_confess("Invalid allocation_scheme '".$opts{allocation_scheme}."\n"
        ."Resource spec:\n".Dumper(\%opts)."\n"
        );
    }
    return \@resources;
}

##########################################################################
# Store nodes based on cluster names they belong to.
##########################################################################
sub _construct_parameter_maps {
    my $self = shift;
    my %opts = validate(@_, {
        cluster => {type => SCALAR, optional => 1},
        nodes   => {type => ARRAYREF},
    });
    my $cluster = $opts{cluster}//$self->DEFAULT_GROUP;
    $cluster = uc($cluster);
    my $nodes = $opts{nodes};
    $self->{cluster_nodes}->{$cluster} = $nodes;
}

##########################################################################
# This gets nodes by their cluster parameter names.
# If there is no cluster group then, all the nodes will be considered belong
# to a default group parameter 'ALL_NODES'. This method is helpful to get
# info for logging discovered resources.
# Return : Hashref containing cluster group names as key and node names
#          separated by comma(',') belonging to the respective cluster group.
##########################################################################
sub get_resources_info {
    my $self = shift;
    my %cluster_info = ();
    if(defined($self->{cluster_nodes}) && %{$self->{cluster_nodes}}) {
        while(my($c_name, $c_grp_nodes) = each(%{$self->{cluster_nodes}})) {
            if(scalar(@{$c_grp_nodes})) {
                $cluster_info{$c_name} = join(',', @{$c_grp_nodes});
            }else {
                $cluster_info{$c_name} = '';
            }
        }
    }
    return \%cluster_info;
}



##########################################################################
# Gets nodes belonging to a given cluster group cluster<N>.
# If no matching cluster found an empty array ref is returned.
##########################################################################
sub _find_resources {
    my $self = shift;
    my %opts = validate (@_, {
            cluster => {type => SCALAR},
        }
    );
    my $cluster = $opts{cluster}//$self->DEFAULT_GROUP;
    $cluster = uc($cluster);
    if(exists($self->{cluster_nodes}->{$cluster})) {
        return $self->{cluster_nodes}->{$cluster};
    }else {
        return [];
    }
}

###################################################################
# Helper method to validate cluster parameter
# Input: Cluster parameter names seperated by comma(',')
# Return: 1 for valid parameter and 0 for non-valid parameter.
###################################################################
sub _is_valid_cluster_param {
    my $cluster = shift;
    my @clstrs = split(',', $cluster);
    my $is_valid = 1;
    foreach my $c(@clstrs) {
        if($c !~ m/^cluster\d+$|^cluster[a-z]$|^all$/i) {
            $is_valid = 0;
            last;
        }
    }
    return $is_valid
}

###################################################################
# Helper method to validate cluster parameter
# Input: Nodes parameter name
###################################################################
sub _is_valid_node_param {
    my $node_param = shift;
    if($node_param =~ m/^all$|^node\d+$|^any$|^none$|^node[a-z]$/i) {
        return 1;
    }else {
        return 0;
    }
}

###################################################################
# Before throwing an exception, reset logger back to previous state
###################################################################
sub _confess {
    my $self = shift;
    my $error = shift;
    reset_loggers(resource_obj => $self);
    confess $error;
}

1;

__END__

# Local Variables:
# perl-indent-level: 4
# indent-tabs-mode: nil
# End:

#############################################################################
#                             POD Documentation
#############################################################################

=head1 NAME

NACL::Plugin::Driver::Resource::Nodes

=head1 SYNOPSIS

use NACL::Plugin::Driver::Resource::Nodes;

my $node_rcs = NACL::Plugin::Driver::Resource::Nodes->new();
my %plugin_opts = ('clusters' => 'cluster<N>');

# Discover clusters group nodes
$node_rcs->discover_resources(%plugin_opts);

my $rcs = $node_rcs->get_resources(
                                'resource_tye'          => 'NODES',
                                'allocation_scheme'     => 'ALL|ANY|NONE|NODE<N>',
                                'clusters'              => 'cluster<N>',
                                'log_context'           => 'context',
);

=head1 DESCRIPTION

A C<NACL::Plugin::Driver::Resource::Nodes> object represents collection of NODE/FILER resources.
It helps to discover nodes available in testbed/environment and seggregate them cluster
wise if cluster groups are present. 'NACL::Plugin::Driver::Resource::Nodes' is a subclass of
'NATE::Driver::Plugin::Resource'. This class designed specifically to be used as a plugin in NATE
Driver to discover and to allocate node/filer resources to a testcase. So it may not be
best fit to be used seperately.

=head1 METHODS

=head2 new

This is a constructor method which returns an instance of 'NACL::Plugin::Driver::Resource::Nodes' class.

=over

=item Synopsis

use NACL::Plugin::Driver::Resource::Nodes;
my $node_rcs = NACL::Plugin::Driver::Resource::Nodes->new();

=item Arguments

No arguments

=item Return values

An object of type 'NACL::Plugin::Driver::Resource::Nodes'.

=back

=head2 discover_resources

This method helps to discover node/filer resources. Optionally it seggregates nodes as cluster groups
like cluster1 nodes, clusterN nodes if 'clusters' parameter has value defined. When clusters is
specified, it discovers nodes belonging to all the cluster groups.

=over

=item Synopsis

my %plugin_opts = ('clusters' => 'cluster<N>');

# Discover clusters group nodes
$node_rcs->discover_resources(%plugin_opts);

=item Arguments

=over

=item C<< clusters => $hashref >>

(optional) This parameter takes a cluster names the resources to be discovered.
If clusters parameter has value defined then, it discover cluster groups and
have cluster parameters cluste1,...,clusterN updated with the nodes belonging
to respective groups.

Nodes will be assigned to cluster<N> parameters in the order they are discovered.

Example: Assume NODES discovery order is: n11, n31, n32, n21, n22, n12

Assume above nodes form clusters C1, C2 and C3 as follows

C1 = (n11, n12), C2 = (n21, n22), C3 = (n31, n32)

The discovered cluster group nodes will be assigned to cluster<N> parameters like below
cluster1 = n11,n12
cluster2 = n31,n32
cluster3 = n21,n22

If no arguement is passed then, all the nodes discovered will be considered to belong to
a single default group.

=over

=item Example: 1

my %plugin_opts = ('clusters' => 'cluster<N>');
# Discover clusters group nodes
$node_rcs->discover_resources(%plugin_opts);

=item Example: 2

$node_rcs->discover_resources();

=back

=back

=item Return values

No return value.

=back

=head2 get_all_resources

This method gets all the nodes available irrespective of arguments to 'discover_resources()'.
Its wrapper on 'NATE::Util::Param::find_nodes()'

=head2 get_resource_type_parameter_list

This methods gets an array ref of parameter names for the NODES resource type from
where the node/filer parameter value can be read.

=over

=item Synopsis

my $param_list = $node_rcs->get_resource_type_parameter_list( parameters_count => $count);

Above method returns $count no. of NODES parameter names like  FILERA, FILERB,..., FILERZ

In addition to above parameter names, FILER, FILERS, FILER_LIST and NODES parameters will be
part of of the parameter list.

=item Arguments

=over

=item C<< parameters_count => $integer >>

Number of parameter names for NODE resource type to be returned. FILER, FILER_LIST and NODES parameters
will be added in addition to the parameter list returned.

=back

=item Return values

An array reference.

=back

=head2 get_resources

This methods gets the NODE resource values based on the specification for 'allocation_scheme' and
'clusters' parameter values.

=over

=item Synopsis

my $rcs = $node_rcs->get_resources(
                                'resource_type'         => 'NODES',
                                'allocation_scheme'     => 'ALL|ANY|NONE|NODE<N>',
                                'clusters'              => 'cluster<N>',
                                'log_context            => 'context',
);

=item Arguments

=over

=item C<< resource_type => $scalar >>

(Required) A string representing resource type name. For node/filer resources its value is 'NODES'.

=item C<< allocation_scheme => $scalar >>

(Required) Specifies the resource quantum to be selected from the discovered nodes resource set.
If 'clusters' parameter has any cluster specified then, allocation scheme considers nodes belonging
to that cluster group only. Otherwise by default all the nodes will be considered.

NODES resource type 'allocation_scheme' supports below values

=over

=item * ALL - All the nodes will be selected.

=item * ANY - A random node from available nodes set will be selected.

=item * NONE - No node will be selected. Instead a constant 'no_resource' will be returned.

=item * NODE<N> - Nth relative node from nodes resource set will be selected.

=back

=item C<< clusters => $hashref >>

(optional) Specifies the filter on resources to be considered while getting nodes for the specified allocation scheme.
Like cluster to be considered like 'cluster<N>'.

=item C<< log_context => $scalar >>

(optional) Specifies the context like which testcase trying to get node. This is helps for logging/debugging.

=back

=item Return values

An array reference.

=back

=head1 Use Case

'NACL::Plugin::Driver::Resource::Nodes' has been implemeneted to be used as a plugin in NATE Driver. This plugin
takes care allocating node/filer resources to testcases in NATE Driver. NATE Driver testcase can use
testcase's 'resources' option to define node/filer resources specification for the testcase as shown
in the below example.

=over

=item Example:1

     use NATE::Driver;
     use NATE::Util::Param qw(find_nodes);

     my %group_spec = (
        tests => [
            {
                name => 'per_serial_cluster1',
                codespec => \&code_ref,
                resources => [
                    {
                        'resource_type'         => 'NODES',
                        'allocation_scheme'     => 'ALL',
                        'clusters'              => 'cluster1',
                    }
                ],
            }
        ]
     );

     sub main {
        $DB::single = 2;
        my $driver = NATE::Driver->new();
        $driver->group_add(name => 'group1', %group_spec);
        $driver->run();
     }

     sub code_ref {
         my $driver = shift;
         my $nodes = find_nodes();
         my $msg = "Nodes belonging to cluste1:".join(',', @{$nodes});
         $driver->log($msg);
         $driver->PASS("Explicit PASS in ".$driver->testname);
     }

=back
