package NACL::Plugin::TestLifeCycle::ASUPTrigger::Deployment;

use v5.14;
use strict;
use warnings FATAL => 'all';
use Data::Dumper::Concise qw/Dumper/;
use Carp qw/confess/;
use Time::HiRes qw/time/;
use List::Util qw/max/;

use NATE::Log      qw/log_global/;
use NATE::ParamSet qw/param_global/;
use NATE::OS       qw/login_name/;
use NATE::Database::Test qw/testdb_global/;
use Smoke qw//;

use Exporter 'import';
our @ISA = qw(Exporter);
our @EXPORT_OK = qw/is_enabled/;

my %Statistics;
sub get_statistics {
    # Return a shallow copy to protect from outside
    return %Statistics;
}

# There are two levels of enablement: plugin and asup
# The plugin may be enabled or disabled
# if the plugin is enabled, the asup functionality may be enabled or disabled
# The subroutine below returns two flags: enable_plugin, enable_asup

# Important: IF YOU ARE LOOKING TO DISABLE THE ASUPTRIGGER MECHANISM BECAUSE
# IT IS CAUSING YOU TROUBLE, PLEASE TRY DISABLE_ASUPTRIGGER_ASUP=1 FIRST, AND
# CONTACT jroussin@netapp.com

sub is_enabled {
    my @flags = _is_enabled();
    log_global->comment(
        "Results:\n".
        "ASUPTrigger plugin enabled: " . $flags[0] ."\n" .
        "ASUPTrigger ASUP enabled  : " . $flags[1] ."\n"
    );
    return @flags;
}

sub _is_enabled {
    my $params          = param_global();
    my $login           = login_name();
    my $submitted_by    = $params->get('SUBMITTED_BY') //'';
    my $site            = $params->get('SITE') //'';
    my $testbed         = $params->get('TESTBED') //'';
    my $jobid           = $params->get('JOBID') //'';
    my $buildroot       = $params->get('BUILDROOT') //'';
    my $test_plan       = $params->get('test_plan') //''; # test_plan path will be either absolute or relative to NATE_LIB
    my $disable_plugin  = $params->get('DISABLE_ASUPTRIGGER_PLUGIN') //'';
    my $disable_asup    = $params->get('DISABLE_ASUPTRIGGER_ASUP') //'';

    my %tests = %ntest::Tests;
    my @normal_tests;
    if ( keys %tests && $tests{NORMAL} && @{$tests{NORMAL}}){
        @normal_tests = map { $_->resolve_fqname } @{$tests{NORMAL}};
    }

    log_global->comment(
        "Evaluating whether the ASUPTrigger plugin should be enabled using these params:\n"
        . 'NATE::OS::login_name()     = '. $login         ."\n"
        . 'SUBMITTED_BY               = '. $submitted_by  ."\n"
        . 'SITE                       = '. $site          ."\n"
        . 'TESTBED                    = '. $testbed       ."\n"
        . 'JOBID                      = '. $jobid         ."\n"
        . 'BUILDROOT                  = '. $buildroot     ."\n"
        . 'test_plan                  = '. $test_plan     ."\n"
        . 'DISABLE_ASUPTRIGGER_PLUGIN = '. $disable_plugin."\n"
        . 'DISABLE_ASUPTRIGGER_ASUP   = '. $disable_asup  ."\n"
        . 'Normal tests               = '. Dumper(\@normal_tests) ."\n"
    );

    return (0,0) if (!@normal_tests);
    return (0,0) if ($disable_plugin);

    my @return;

    # Detect when the run is purely of ctl stuff and not a testsuite
    my $non_ctl_tools = scalar(grep {
        ! m|^/x/eng/ctl/test-tools/| &&
        ! m|^/x/eng/ctl/ctl5/thpl/| &&
        ! m|/tools/results_mail\.thpl$|
    } @normal_tests);
    if (!$non_ctl_tools){
        log_global->comment("Only CTL tools detected as normal tests. Turning off ASUPTrigger plugin");
        return (0,0);
    }


    if ($site eq 'sti' || $site eq 'ctl'){

        # 8.2 is causing plugin to not compile. 9.1 compiles but spews out errors. Don't know about the others.
        if ($buildroot && ( $buildroot =~ m{^/x/eng/.*/R?8\.\d} || $buildroot =~ m{^/x/eng/.*/R?9\.[01]} ) ){
            log_global()->comment("Detected unsupported ONTAP build. Disabling ASUPTrigger plugin");
            return (0,0);
        }

        if ( $testbed eq 'pushbutton'){
            log_global->comment("Detected CTL-PB environment");

            if (!$submitted_by){
                log_global()->comment('No SUBMITTED_BY seen, assuming "setup" step. Disabling ASUPTrigger plugin.');
                return (0,0);
            }
        }
        else{
            if ($test_plan){
                log_global->comment("Detected CTL5 environment");
            }
            else{
                log_global->comment("Detected CTL environment");
            }
        }
    }
    elsif (Smoke::runningInBubble()){
        log_global->comment("Detected Smoke bubble");
        # We turn it off for now, unless specifically told otherwise
        if ( ! defined($disable_plugin) || $disable_plugin ne '0' ){
            return (0,0);
        }
    }
    else{
        # See if we are running in OpenLab by any chance
        require Net::Ping;
        my $pinger = Net::Ping->new(undef,2); #2s timeout
        my $openlab_host = 'cyclortp.rtp.openenglab.netapp.com';
        my $rtp_host     = 'cyclrtp.rtp.netapp.com';

        # Speed up the common case of running in a lab that can see RTP englab
        my $success = $pinger->ping($rtp_host);
        if ($success){
            log_global()->comment("Not running in OpenLab");
        }
        else{
            $success = $pinger->ping($openlab_host);
            if ($success){
                log_global()->comment("Detected OpenLab. Disabling ASUPTrigger plugin");

                # Since determining we are in OpenLab is expensive
                # Make a note of that fact somewhere accessible from any process
                # The KeyVal service would be great for this, but it doesnt work
                # in Cloud tests. Use the NATE DB instead
                my $db = testdb_global();
                $db->set_parameter(
                    subtest_id  => 0,
                    param_key   => 'RUNNING_IN_OPENLAB',
                    param_value => 1,
                    origin      => 'asuptrigger', # Tag the param differently from real parameters
                );
                return (0,0);
            }
            else{
                log_global()->comment("Not running in OpenLab");
            }
        }
    }

    # Turn off if run by non-QRE folks for now
    if ($submitted_by){
        $login = $submitted_by;
    }
    my $chain = _get_management_chain($login);
    @return = (1,1);
    if ($chain && $chain !~ /blineman/){

        # Outside of QRE, we only need to keep track of hardware usage for now
        # so only send ASUPs if a non-vsim node is detected and only send
        # TEST ASUPs
        my $db = testdb_global();
        my $commit = 0;
        my $asuptrigger_asup_type = $params->get('ASUPTRIGGER_ASUP_TYPE');
        if (! $asuptrigger_asup_type){
            $db->set_parameter(
                subtest_id  => 0,
                param_key   => 'ASUPTRIGGER_ASUP_TYPE',
                param_value => 'test',
                origin      => 'asuptrigger', # Tag the param differently from real parameters
            );
            $commit = 1;
        }
        my $asuptrigger_skip_vsims = $params->get('ASUPTRIGGER_SKIP_VSIMS');
        if (! defined $asuptrigger_skip_vsims){
            $db->set_parameter(
                subtest_id  => 0,
                param_key   => 'ASUPTRIGGER_SKIP_VSIMS',
                param_value => 1,
                origin      => 'asuptrigger', # Tag the param differently from real parameters
            );
            $commit = 1;
        }
        # This is currently only used for reporting purposes. The bit will be sent to graylog
        my $asuptrigger_user_not_qre =  $params->get('ASUPTRIGGER_USER_IS_NOT_QRE');
        if (! defined $asuptrigger_user_not_qre){
            $db->set_parameter(
                subtest_id  => 0,
                param_key   => 'ASUPTRIGGER_USER_IS_NOT_QRE',
                param_value => 1,
                origin      => 'asuptrigger', # Tag the param differently from real parameters
            );
            $commit = 1;
        }
        $db->db_commit if ($commit);
    }

    $return[1] = 0 if ($disable_asup);
    return @return;
}

# It's going to be pretty taxing calling manager_of on every call of ntest in ctl
# Make sure to filter by suite name first
sub _get_management_chain {
    my $login = shift;
    my $cmd = '/usr/software/bin/perl-5.8.8 /usr/software/rats/bin/manager_of -c -o login ' . $login . ' 2>&1';
    my @managers;
    my $start_time = time();
    log_global()->comment("Computing management chain of '$login'");
    my @lines = `$cmd`;
    my $runtime = time() - $start_time;
    $Statistics{max_runtime} //= $runtime;
    $Statistics{max_runtime} = max($runtime, $Statistics{max_runtime});
    my $tries = 5;
    while (!@lines && $tries > 0){
        my $sleep_time = 2 + (5 - $tries)*5;
        log_global()->comment("No lines received from '$cmd'. Sleeping $sleep_time seconds and trying again");
        sleep $sleep_time;
        $tries --;
        $start_time = time();
        @lines = `$cmd`;
        $runtime = time() - $start_time;
        $Statistics{max_runtime} = max($runtime, $Statistics{max_runtime});
        $Statistics{retries} ++;
    }
    if ( $lines[0] =~ /not found/ ){
        log_global()->warn("User '$login' not found with manager_of");
        return;
    }
    if (!@lines || scalar @lines < 3){
        log_global->warn("Unexpected output of '$cmd':\n". join("\n", @lines));
        return;
    }

    my $management_chain = join(',', map { my $a = $_; $a =~ s/\s+//g; $a } @lines);
    log_global()->comment("Management chain for login '$login': $management_chain");
    return $management_chain;
}

1;
