#
# Copyright (c) 2001-2016 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Resource List Module for Cleanup Subsystem
## @author dl-nacl-dev@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::Realm::HelpXmlStore;

=head1 METHODS

=head2 new

(Class method) Instantiates a new Realm::HelpxmlStore object.

Example:

my $helpxml_store_obj = NACL::Realm::HelpXmlStore->new( build_root => $build_root );

=head1 ATTRIBUTES

=over

=item C<< help_xml_path >>

path of the help xml files

=item C<< help_xml_accessible >>

returns the true (boolean) if the 
help xml path is accessible, else returns false.

=item C<< cli_to_zapi_path >>

path of the xml file containing cli to 
zapi command mapping.

=item C<< cli_to_zapi_accessible >>

returns the true (boolean) if the 
cli to zapi path is accessible, 
else returns false.

=item C<< cli_to_zapi_mapping_possible >>

returns the true (boolean) if the both cli to 
zapi path and help xml paths are accessible,
else returns false.

=item C<< cli_to_zapi >>

perl data structure (hashref) containing cli 
to zapi command mapping data.

=item C<< help_xml_contents >>

perl data containing "command" as a key and its 
"attribute mappings" as values.

=item C<< build_root >>

build root

=item C<< ui_details >>

A hash-reference, keyed by the API (like "volume_show") with the value
being a hash-reference containing details of the command as obtained
by issuing a '?' on the CMode-CLI UI.

=back

=cut

package NACL::Realm::HelpXmlStore;

use strict;
use warnings;

use Params::Validate qw(validate validate_with :types);
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use XML::Simple;
use NACL::Tools::CGT::CgtXml qw(parse_xmlfile);
use NACL::Tools::CGT::CGTUtils qw(parse_leaf_level_cmd _typedef_cli_zapi determine_pk);
use NACL::CS::ComponentState::ZapiSkip qw(make_zapi_skip);
use NACL::CS::ComponentState::ZapiArray qw(make_zapi_array);
use NACL::Global;

use Data::Dumper;

use Class::MethodMaker [
    new    => [ '-hash', '-init', 'new' ],
    scalar => [
        {   '-default_ctor' => sub {
                $_[0]->_get_help_xml_path();
            },
        },
        'help_xml_path'
    ],
    scalar => [
        {   '-default_ctor' => sub {
                defined $_[0]->help_xml_path() && -e $_[0]->help_xml_path();
            },
        },
        'help_xml_accessible'
    ],
    scalar => [
        {   '-default_ctor' => sub {
                $_[0]->_get_cli_to_zapi_path(
                    build_path => $_[0]->build_root() );
            },
        },
        'cli_to_zapi_path'
    ],
    scalar => [
        {   '-default_ctor' => sub {
                defined $_[0]->cli_to_zapi_path() && -e $_[0]->cli_to_zapi_path();
            },
        },
        'cli_to_zapi_accessible'
    ],
    scalar => [
        {   '-default_ctor' => sub {
                $_[0]->help_xml_accessible()
                    && $_[0]->cli_to_zapi_accessible();
            },
        },
        'cli_to_zapi_mapping_possible'
    ],
    scalar => [
        {   '-default_ctor' => sub {
                my $self = $_[0];
                if ( $self->cli_to_zapi_accessible() ) {
                    return NACL::Tools::CGT::CgtXml::cli_to_zapi(
                        $self->build_root() );
                }
                }
        },
        'cli_to_zapi'
    ],
    hash   => [qw/help_xml_contents/],
    scalar => 'build_root',
    hash   => 'ui_details',
    scalar => [
        {-type => 'NACL::APISet::Version'},
        'version_manager'
    ],
];

sub init {
    $Log->enter() if $may_enter;
    my ( $self, @args ) = @_;
    my %opts = validate_with(
        params => \@args,
        spec   => {
            build_root      => { type => SCALAR, },
            version_manager => {type => OBJECT, isa => 'NACL::APISet::Version'},
        },
    );
    $Log->comment('HelpXmlStore created with build_root: ' . $opts{build_root});
    $Log->exit() if $may_exit;
}

=head2 get_attribute_map

(Instance method) Returns the data zapi attribute mapping 
perl data structure for the given command.

Example:

use NACL::Realm;
my $realm_obj = NACL::Realm_new();
my $help_xml_obj = $realm_obj->get_help_xml_objw();
my $cmd        = "storage aggregate create";
my $attribute_map = $help_xml_obj->get_attribute_map(command => $cmd,
                                                     alias_map => $alias_map);

=over

=item Options

=back

=over

=item C<<command>>

(Mandatory) Command for which the zapi attribute mapping needs to be objtained.
Example" "storage aggregate create"

=back

=over

=item Exceptions thrown

=back

=cut

sub get_attribute_map {
    $Log->enter() if $may_enter;
    my ( $self, @args ) = @_;
    my %opts = validate( @args, { command => { type => SCALAR },
                                  alias_map => { type => HASHREF|UNDEF, optional => 1 },} );
    my $command = $opts{command};
    my $alias_map = $opts{alias_map};
    my $has_alias_fields = 0;
    if ($alias_map
        && keys %{ $alias_map } ) {
        $has_alias_fields = 1;
    } 
         

    if ( $self->help_xml_contents_exists($command) ) {
        my $ds = $self->help_xml_contents_index($command);
        if ( !$has_alias_fields ) {
            return $ds;
        } else {
            if ( $ds->{has_alias_fields} == $has_alias_fields ) {
                return $ds;
            }
        }
    }

    my $help_xml_path = $self->help_xml_path();
    my $xml_struct = $self->_read_xml_content( command => $command,
                                               alias_map => $alias_map);

    $self->help_xml_contents_set( $command => $xml_struct );

    $Log->exit() if $may_exit;
    return $xml_struct;
}

# read the xml content
sub _read_xml_content {
    $Log->enter() if $may_enter;
    my ( $self, @args ) = @_;
    my %opts = validate( @args, { command => { type => SCALAR },
                                  alias_map => { type => HASHREF|UNDEF },
                                  } );
    my $help_xml_path = $self->help_xml_path();
    my $cmd = $opts{command};
    my $alias_map = $opts{alias_map};
    $cmd =~ s/\s+/\//g;
    $help_xml_path .= $cmd . '.xml';
    my $xml_struct = parse_xmlfile(xml_file => $help_xml_path,
                                   alias_map => $alias_map);

    $Log->exit() if $may_exit;
    return $xml_struct;
}

# get the help_xml path

sub _get_help_xml_path {
    $Log->enter() if $may_enter;
    my ($self) = @_;

    my $build_path = $self->build_root();
    $build_path =~ s/\/$//;
    my $path = $build_path . '/ng_docs/bedrock/export/common/';
    if ($self->version_manager()->is_version_ge(
        release => $NACL::Global::ONTAP_Release_Longboard_0)) {
        $path .= 'gendocs/';
    } else {
        $path .= 'dist/root/mroot_late/etc/clihelp/help_xml/';
    }
    $Log->comment("HelpXmlStore: help_xml_path: $path");

    if ( !( -e $path ) || !( -d $path ) ) {
        $Log->comment("help_xml path: $path is inaccessible");
        $self->help_xml_accessible(0);
        $self->cli_to_zapi_accessible(0);
        $Log->exit() if $may_exit;
        return;
    }
    $self->help_xml_path($path);
    $Log->exit() if $may_exit;
    return $path;
}

# get the cli to zapi command mapping

sub _get_cli_to_zapi_path {
    $Log->enter() if $may_enter;
    my ( $self, @args ) = @_;
    my %opts = validate_with(
        params => \@args,
        spec   => { build_path => { type => SCALAR, } }
    );

    my $build_path = delete $opts{build_path};
    $build_path =~ s/\/$//;
    my $path;
    if ($self->version_manager()->is_version_ge(
        release => $NACL::Global::ONTAP_Release_Longboard_0)) {
        $path = $build_path . "/final/bedrock/export/common/cli_zapi/";
    } else {
        $path = $build_path . "/adminapi/bedrock/internal/common/smf/zapi/cli_zapi/";
    }
    $Log->comment("HelpXmlStore: cli_to_zapi_path: $path");
    $self->cli_to_zapi_path($path);
    $Log->exit() if $may_exit;
    return $path;
}

=head2 get_ui_details

    my $ui_details_hashref = $help_xml_obj->get_ui_details(
        command_interface => $ci,
        api               => $api,
    );

Returns a hash-reference containing details obtained by issuing a '?'
at the CMode-CLI UI. If an entry for the specified API already exists,
it simply returns the existing entry rather than re-querying.

=cut

sub get_ui_details {
    $Log->enter() if $may_enter;
    my ($self, @args) = @_;
    my %opts = validate_with(
        params => \@args,
        spec   => {
            command_interface => {
                type     => OBJECT,
                isa      => 'NACL::C::CommandInterface::ONTAP',
            },
            api => {type => SCALAR},
        }
    );
    my $api = $opts{api};
    if ($self->ui_details_exists($api)) {
        $Log->exit() if $may_exit;
        return $self->ui_details_index($api);
    }
    my $valid;
    my $ci = $opts{command_interface};
    my $apiset = $ci->apiset(interface => 'CLI', set => 'CMode',version_info => { compatibility => 0 });
    my $cmd_ref = parse_leaf_level_cmd( apiset => $apiset,
        api => $api, want_error_msg => 0, valid_command => \$valid);
    $self->ui_details_set($api => $cmd_ref);

    if ( $valid
         && $api =~ /show$/ ) {
        my $pks = determine_pk(apiset => $apiset,
                     show_cmd => $api);
        %{ $cmd_ref } = (%{ $cmd_ref }, %{ $pks });
    }
    $Log->exit() if $may_exit;
    return $cmd_ref;
}

=head2 get_zapi_field_mapping

(Instance method) Returns the zapi mapping for fields in ngsh 

Example:

use NACL::Realm;
my $realm_obj = NACL::Realm->realm();
my $help_xml_obj = $realm_obj->get_help_xml_obj();
my $zapi_mapping = $help_xml_obj->get_zapi_field_mapping(command => 'vserver modify',
                                                     fields   => [ 'allowed-protocols']);

=over

=item Options

=back

=over

=item command

(Mandatory) The ngsh command for which the zapi mapping needs to be obtained.
Example "vserver modify"

=item fields 

(Mandatory) Arrayref containing ngsh command option for which the zapi mapping needs to be obtained.

=item mapping_as_ds

(Optional) Returns the zapi mapping as a data structure rather than a string

=back

=cut

sub get_zapi_field_mapping {
    $Log->enter() if $may_enter;
    my ($self, @args) = @_;
    my %opts = validate_with(
        params => \@args,
        spec   => {
            command => {type => SCALAR},
            new_fields   => {type => ARRAYREF},
            mapping_as_ds => {type => BOOLEAN, default => 0},
            alias_map => { type => HASHREF|UNDEF, optional => 1 }
        }
    );
    my $command = $opts{command};
    my $new_fields   = $opts{new_fields};
    my $alias_map = $opts{alias_map};
    my ( %mapping, $map);
    my $dump = $self->get_attribute_map(command => $command,
                                        alias_map => $alias_map);
    my $fields = $dump->{fields};
    foreach my $field ( @{ $new_fields } ) {
        if ($fields->{$field}
            && $fields->{$field}->{'zapi-mapping'} ) {
            $map = $fields->{$field}->{'zapi-mapping'};
            if ( $opts{mapping_as_ds} 
                 && $map =~ /^\[/ ) {
                 my $eval_str = eval "$map";
                 if ( $@ ) {
                     $Log->exit() if $may_exit;
                     NATE::BaseException->throw("Unable to eval $map :".$@);
                 }                 
                 $map = $eval_str;
            }
            $mapping{$field} = $map;
            $Log->debug(sub { "\n Updated mapping for $field is ".Dumper($map) });
        }
    }
    $Log->exit() if $may_exit;
    return \%mapping;
}

=head2 get_zapi_translation

(Instance method) Returns an hashref containing zapi fields translations for a 
given ngsh command

Example:

use NACL::Realm;
my $realm_obj = NACL::Realm->realm();
my $help_xml_obj = $realm_obj->get_help_xml_obj();
my $zapi_mapping = $help_xml_obj->get_zapi_translation(command => 'vserver modify',);

=over

=item Options

=back

=over

=item command

(Mandatory) The ngsh command for which the zapi fields translation  needs to be obtained.
Example "vserver modify"

=back

=cut

sub get_zapi_translation {
    $Log->enter() if $may_enter;
    my ($self, @args) = @_;
    my %opts = validate_with(
        params => \@args,
        spec   => {
            command => {type => SCALAR},
            alias_map => {type => HASHREF|UNDEF}
        }
    );
    my $command = $opts{command};
    my $alias_map = $opts{alias_map};
    my $dump = $self->get_attribute_map(command => $command,
                                        alias_map => $alias_map);
    my $translation = $dump->{zapi_value_conversion};
    $Log->exit() if $may_exit;
    return $translation;
}

=head2 get_cli_to_zapi_mapping

    my $zapi_arrayref = $self->get_cli_to_zapi_mapping(command => $command);

Determines the ZAPI(s) equivalent to the specified CLI command. If there is
a ZAPI equivalent, it returns an array-reference with all of the equivalent
ZAPIs. If there is no ZAPI equivalent, then undef is returned.

=cut

sub get_cli_to_zapi_mapping {
    $Log->enter() if $may_enter;
    my ($self, @args) = @_;
    my %opts = validate(@args, {command => {type => SCALAR}});
    my $command = $opts{command};
    my $cli_to_zapi_hash = $self->cli_to_zapi();
    my $zapi_cmds = $cli_to_zapi_hash->{$command};
    if (defined $zapi_cmds && !ref $zapi_cmds) {
        # If there's only a single ZAPI equivalent, cli_to_zapi() will store
        # it as a scalar, but if there are multiple ZAPIs, it will be stored
        # as an array. For this reason, we normalize everything to an
        # array-reference.
        $zapi_cmds = [$zapi_cmds];
    }
    $Log->exit() if $may_exit;
    return $zapi_cmds;
}

=head2 get_field_datatype

(Instance method) Returns the datatype  for fields in ngsh

Example:

use NACL::Realm;
my $realm_obj = NACL::Realm->realm();
my $help_xml_obj = $realm_obj->get_help_xml_obj();
my $datatype = $help_xml_obj->get_field_datatype(command => 'vserver modify',
                                                 fields   => [ 'allowed-protocols']);

=over

=item Options

=back

=over

=item command

(Mandatory) The ngsh command for which the datatype needs to be obtained for the passedin fields.
Example "vserver modify"

=item fields

(Mandatory) Arrayref containing ngsh command option for which the data type needs to be obtained.

=back

=cut


sub get_field_datatype {
    $Log->enter() if $may_enter;
    my ($self, @args) = @_;
    my %opts = validate(@args, {command => {type => SCALAR},
                                fields => {type => ARRAYREF},
                                alias_map => { type => HASHREF|UNDEF} });
    my $command = $opts{command};
    my $alias_map = $opts{alias_map};
    my $fields = $opts{fields};
    my $attr_map = $self->get_attribute_map(command => $command,
                                            alias_map => $alias_map);
    my %datatype;
    foreach my $field ( @{ $fields } ) {
        if ( $attr_map->{fields}->{$field} ) {
            if ( $attr_map->{fields}->{$field}->{is_array} == 1 ) {
                $datatype{$field}="array";
            } else {
                $datatype{$field}="scalar";
            }
        }
    }
    $Log->exit() if $may_exit;
    return \%datatype;
}

sub get_field_datatype_ui {
    $Log->enter() if $may_enter;
    my ($self, @args) = @_;
    my %opts = validate(@args, {
                            command_interface => {
                                type     => OBJECT,
                                isa      => 'NACL::C::CommandInterface::ONTAP',
                            },
                            api => {type => SCALAR},
                            fields => {type => ARRAYREF},
                         });
    my $fields = delete $opts{fields};
    my $attr_map = $self->get_ui_details(%opts);
    my %datatype;
    foreach my $field ( @{ $fields } ) {
        if ( $attr_map->{fields}->{$field} ) {
            if ( $attr_map->{fields}->{$field}->{is_array} == 1 ) {
                $datatype{$field}="array";
            } else {
                $datatype{$field}="scalar";
            }
        }
    }
    $Log->exit() if $may_exit;
    return \%datatype;
}
1;
