#
# Copyright (c) 2001-2016 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Aggregate Task Module
## @author dl-nacl-dev@netapp.com
## @status shared
## @pod here

package NACL::STask::Aggregate;

use strict;
use warnings;

use base qw(NACL::C::Aggregate NACL::STask::STask);

use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Params::Validate
    qw(validate validate_with SCALAR BOOLEAN SCALARREF HASHREF OBJECT UNDEF ARRAYREF);
use Readonly;
use NATE::Exceptions::Argument;
use NACL::Exceptions::NoElementsFound;
use NACL::Exceptions::HighAvailabilityError;
use NATE::BaseException qw(:try);
use NACL::Exceptions::VerifyFailure ();
use NACL::APISet::Exceptions::CommandFailedException;
use NACL::ComponentUtils qw(_optional_scalars);
use NACL::APISet::Exceptions::ResponseException           ();
use NACL::APISet::Exceptions::TimeoutException            ();
use NACL::APISet::Exceptions::InvalidOptionException      ();
use NATE::BaseException                                   ();
use NACL::Exceptions::VserverHasVolumesOnAnotherAggregate ();
use NACL::C::Exceptions::StorageAggregate::NoMatchingDisksAvailable ();
use NACL::C::Exceptions::StorageAggregate::UnmirroredAggregatesNotSupported ();
use NACL::C::Exceptions::StorageAggregate::FreeSpaceReallocationEnabled ();
use NACL::C::Exceptions::StorageAggregate::ContainsVolumes;
use NACL::C::Exceptions::StorageAggregate::DoesNotExist;
use NACL::STask::Volume;
use NACL::STask::RaidGroup;
use NACL::STask::Vserver;
use NACL::C::VolumeClone;
use NACL::STask::StorageDisk;
use NACL::CS::StorageAggregateShowResyncStatus;
use NACL::CS::RaidConfigInfoListDisk;
use NACL::CS::StorageAggregateShowStatus;
use NACL::ComponentUtils qw(Dumper);
use NACL::GeneralUtils qw(arrange_by_create_time);
use NACL::C::Volume;
use NACL::C::DebugVreport;
use NACL::STask::StoragePool;
use NACL::STask::StorageAggregateObjectStore;
use NACL::GeneralUtils qw(nacl_method_retry);
use NACL::STask::Exceptions::SpaceComparisonFailure;
use NACL::STask::Exceptions::MVRcheckFailure;
use NACL::CS::StoragePoolShowAggregate;
use NACL::C::Exceptions::StorageAggregate::StoragePoolBusy;
use NACL::Global;
use NATE::ParamSet qw(param_global);

use Class::MethodMaker [
    'scalar' => [ { -type => 'NACL::C::Job' }, 'job_component', ] ];

# Same routine for aggregates and volumes.  Pull implementation from
# _AggregateVolume.pm.
use NACL::STask::_AggregateVolume qw(:all);
use constant NACL_AGGREGATE_VERIFY_TIMEOUT => 1800;
use constant MEDIA_SCRUB_TIMEOUT           => 3600;
use constant MEDIA_SCRUB_POLL_TIME         => 20;
use constant NACL_AGGREGATE_MIRROR_TIMEOUT => 180;

Readonly my %MIN_DISK_COUNT => (
    non_debug_build => {
        raid_tec => 7,
        raid_dp => 5,
        raid4 => 3,
    },
    debug_build => {
        raid_tec => 7,
        raid_dp => 3,
        raid4 => 2,
    }
);
my $VALID_SIZE_RE = qr/^[+-]?\d+(?:\.\d+)?(?:(?i)(b|[kmgt]b?))?$/i;

my $Infinity = 9**9**9;

=head1 NAME

NACL::STask::Aggregate

=head1 DESCRIPTION

C<NACL::STask::Aggregate> provides a number of well-defined but
potentially complex or multi-step methods related to Aggregates in
ONTAP.

It builds on top of, and is a derived class of C<NACL::C::Aggregate>,
and so it also provides methods that are more in the scope of
individual Aggregate-related commands.  See C<NACL::C::Aggregate> for
details.  This also means that a C<NACL::STask::Aggregate> object may
generally be used in place of a component object.

=head1 CLEANUP METHODS

Cleanup can be registered for the following methods, 'create', 'modify',
'online', 'offline', 'restrict', shared_create, recyclable_create.

Cleanup methods are,

    Aggregate Method      Cleanup Method
    --------------------------------------
    create                purge
    modify                modify
    online                modify
    offline               modify
    restrict              modify
    rename                rename (TBD)
    shared_create         shared_undo
    recyclable_create     recyclable_undo

=head1 ATTRIBUTES

=head2 command_interface

(Required) A component object that represents the host to which to send commands.

=head2 aggregate

(Required) The name of the aggregate.

=head2 job_component

(Optional) A job component (NACL::C::Job) representing the last job that
operated on this aggregate. At the moment this will only get filled in for
the job used to create this aggregate (if this task object was used to create
this aggregate, and if C-mode was used so that there is/was a job ID)

=head2 last_job

B<DEPRECATED>. The C<job_component> attribute should be used instead, this
has been retained for backwards compatibility.

=head1 METHODS

=head2 create

  my $aggregate  = NACL::STask::Aggregate->create(
      command_interface  => $command_interface,
      aggregate          => $aggregate,
      nacltask_if_exists => $action,      # default 'die'
      nacltask_wait      => $boolean,     # default '1'
      nacltask_to_cleanup => $boolean,
      nacltask_cleanup_manager => $object,
      %other_options
  );

  or

  my $aggregate  = NACL::STask::Aggregate->create(
      command_interface  => $command_interface,
  );

(Class method) Create an aggregate.  This method provides additional
services beyond what's inherent to the product's aggregate creation
commands, as controlled and described by the new C<if_exists> and
C<wait> options.

Note that the argument C<aggregate> is optional for this method.
If this is not specified, a random name is generated using
L<NACL::GeneralUtils::random_name_generator|lib-NACL-GeneralUtils-pm/random_name_generator>.
(See C<nacltask_rng_args> for how to customize arguments to
random_name_generator())

=over

=item Options

=over

=item C<< if_exists=>$action >>

B<DEPRECATED>.
Use C<nacltask_if_exists> instead.

=item C<< nacltask_if_exists => $action >>

(Optional) What to do if the aggregate to be created already exists.
If $action is "die" (the default), then fail with an exception (in the
same way that the aggregate component would have: by trying the
aggregate create and letting the product complain about the aggregate
already existing).  If action is "reuse", then do nothing (return a
task object referring to the existing aggregate).  If action is
"purge", then delete the aggregate (see the C<purge> method, below)
before creating a new one.

=item C<< nacltask_wait => 0|1 >>

(Optional) If 1 (the default), wait for this aggregate to be
created (see "wait_for_creation" method, below).  If 0, do not
wait any longer than necessary (any longer than the aggregate creation
command waits).

=item C<< _was_created => \$scalar >>

(Optional)
When this option is provided a reference to a scalar variable, the variable
gets filled in with a boolean value describing whether the aggregate
was found (value will be 0; this scenario is possible when
if_exists => "reuse") or whether the aggregate was created (value will be 1).
This is necessary to determine whether the aggregate needs to be
cleaned up later.

    my $was_created;
    my $aggr_obj = NACL::STask::Aggregate->create(
        if_exists    => 'reuse',
        _was_created => \$was_created,
        %other_opts
    );
    # Operate on $aggr_obj here
    # ...
    # Now determine whether to clean up the aggregate, since we're not sure
    # whether we reused an existing aggregate or created a new one
    if ($was_created) {
        # New aggregate was created. Clean it up.
        $aggr_obj->purge();
    }

=item C<< "method-timeout" => 7200 >>

(Optional) As component method-timeout, controls how long the command
will wait before completing.  However, the default timeout has been
raised to 7200 seconds.

=item C<< nacltask_enable_hybrid => 1|0 >>

(Optional, default => 0)
If nacltask_enable_hybrid is passed with the value 1, after the aggregate is
successfully created, aggregate will be modified to be hybrid enabled.
If this option is set, nacltask_wait will be overridden because the
aggregate can't be modifies while being created.

=item C<< nacltask_enable_composite => 1|0 >>

(Optional, default => 0)
If nacltask_enable_composite is passed with the value 1, after the aggregate is
successfully created, aggregate will be modified to be composite.
If this option is set, nacltask_wait will be overridden to wait (1) because the
aggregate can't be modified while being created.
This method assumes that the object-store needed for the attach is already created.

=item C<< nacltask_object_store_name >>

(Optional) Name of object store. This parameter will be used to create composite
aggregates, which is enabled by parameter 'nacltask_enable_composite'.
This method assumes that the object-store needed for the attach is already created.

=item C<< nacltask_object_store_size_limit >>

(Optional) Size limit of object store. This parameter will be used to create composite
aggregates, which is enabled by parameter 'nacltask_enable_composite'.
This method assumes that the object-store needed for the attach is already created.

=item C<< nacltask_add_disks => \%add_disks_opts >>

(Optional)
If this option is provided, once the aggregate is successfully created, the disks
specified through this option will be added as per the parameters passed. This
parameter is reference to a hash, which is same
as L<< NACL::C::Aggregate->add_disks|lib-NACL-C-Aggregate-pm/add_disks >>

=item C<< nacltask_min_size => $min_size >>

(Optional)
When this parameter is specified with the valid aggregate size, once the aggregate
is created, disks are added till the size of the aggregate reaches $nacltask_min_size.
Also, if this option is specified, "diskcount" parameter is ignored and the aggregate
is created with the minimum disks required relevant to the disktype specified and
nacltask_wait will be overridden to 1. grow_size() is called at the end of the create
method, hence If the create() method is called with the options, "nacltask_enable_hybrid"
or "nacltask_add_disks", aggregate will be converted to hybrid, disks will be added as
per the "nacltask_add_disks" option and then grow_size() will be called. If there are
not sufficient spare disks available to grow the aggregate size, CommandFailedException
will be thrown and the aggregate will be purged (no aggregate will be created) before
returning the control.

=item C<< nacltask_no_ssd => $boolean >>

(Optional default: 0)
This option is to specify that the aggregate chooses non-SSD disks while creating an
aggregate overriding what ONTAP chooses. The disktype for which the maximum number of
spare disks are available will be considered. When this option is choosen, option
"diskcount" is mandatory and if not given NATE::Exceptions::Argument will be thrown.
Option "disklist" or/and "disktype" with "nacltask_no_ssd" is also invalid hence this
combination of options is also results in NATE::Exceptions::Argument.
And also choosing non-SSDs is applicable only for create method and not for "grow".


=item C<< polling_interval => $interval >>

(Optional)
If nacltask_wait is set to 1, this is the interval at which to show the progress
of the job (if we're watching the progress of the job) or the interval at
which to poll the filer for the value of the "state" field.

=item C<< nacltask_add_to_aggr_list_of_vserver => $vserver_name >>

(Optional)
If this option is provided, once the aggregate is successfully created,
Created aggregate should be added to the aggr-list of the vserver , which is
specified through this option. If this option is not passed then no aggregates
will be added to the aggr-list of the vserver.

=item C<< nacltask_to_cleanup => 0|1 >>

(Optional, defaults to 0(no to cleanup))
Flag indicating if this operation is to be cleaned up or not.

=item C<< nacltask_cleanup_manager >>

Cleanup manager to use for registering.
Default : Will use the default cleanup manager.

=item C<nacltask_use_partitions>

If this flag is set to 1 then the aggr create function will look for only partitioned disks.
If it is set to 1, then the aggr create function will look for only physical disks.
If a value is not specified, then ONTAP chooses the disks to use".
(when nacltask_use_partitions is not specified, 'undef' will be considered as default).

=item C<< nacltask_rng_args => { args_for_rng_here } >>

(Optional, used only if a random aggregate name is generated)
If no aggregate name is supplied, a random name is generated using
L<NACL::GeneralUtils::random_name_generator|lib-NACL-GeneralUtils-pm/random_name_generator>
with two random words in it and with the suffix being "_aggr". This argument
can be used to customize the arguments to that method. For example, if we'd
like to specify that the random generated aggregate name should be 255 characters
length, then this could be:

    NACL::STask::Aggregate->create(
        %other_opts,
        nacltask_rng_args => { length => '255' }
    );

See L<NACL::GeneralUtils::random_name_generator|lib-NACL-GeneralUtils-pm/random_name_generator>
for a description of the arguments it accepts.

=item C<MCC options>

The options accepted for MCC configuration replication verification is
documented at L<MCC options|lib-NACL-STask-STask-pm/mcc_options>.

=item C<< blob => $blob_value >>

(Optional, defaults to 0)
This parameter contains the value of the blob which is one of the attribute in
mailbox entry. If the value is passed, it is verified with the dblade table
entry. This parameter is applicable only if we create a mirrored aggregate.

=item C<< nacltask_verify_mvr => $boolean >>

(Optional, defaults to 0)
This option is to check whether the MVR(mirror vote records) check needs to be done
 for the created aggregate.

=item C<< nacltask_storage_pool >>

(Optional) This parameter contains the name of storage pool to be added to the aggregate.

=item C<< nacltask_allocation_unit >>

(Optional) This parameter contains the value of  allocation unit to be
added to create hybrid aggregate. Default value is 1. and maximum value is 4.

=item C<additional_disk_count>

If specified, the diskcount of the aggregate created will be <minimum_disk_count> + <additional_disk_count>.

=item C<< nacltask_initialize_compliance_clock => 0|1 >>

(Defaults to 1)
This option specifies whether to initialize the Snaplock Compliance Clock (which
is necessary for creating snaplock aggregates). If set to 0 and the compliance
clock is not already set, a
C<NACL::C::Exceptions::StorageAggregate::ComplianceClockNotSet> exception is
thrown.

=item command_interface, apiset_must, apiset_should, aggregate, diskcount, disklist, voltype, etc.

All of the other various options to
L<< NACL::C::Aggregate->create|lib-NACL-C-Aggregate-pm/create >>

=item Arguments defaulted/auto-determined

=over

=item C<aggregate>

Note that if this is not specified, a random name is generated using
L<NACL::GeneralUtils::random_name_generator|lib-NACL-GeneralUtils-pm/random_name_generator>.
(See C<nacltask_rng_args> for how to customize arguments to
random_name_generator())

=item C<diskcount>

Note that if both diskcount and disklist are not specified, diskcount will be autoassigned
based on the raidtype(when raidtype is not specified, 'raid_dp' will be considered as default).

=item C<mirror>

For sDOT systems with more than one node in the cluster, the STask defaults to
creating mirrored aggregates (i.e. by setting mirror => 'true').
If necessary, this can be disabled by passing mirror => 'false' in the call.

=back

=item Parameters

The following command-line parameters can be specified to affect the behaviour of the call at run-time.

=over

=item NACL_AGGR_RAIDTYPE

This value tells raidtype and sets the same for aggregate creation.

=item NACL_AGGR_MIRROR

This value overides mirror value for aggregate creation.

=item NACL_AGGR_SIDL

This value enables single instance data logging on aggreggate

=back

=back

=back

=over

=item Exceptions

=over

=item C<NACL::C::Exceptions::Aggregate::AlreadyExists>

This type of exception is thrown when an attempt is made to create an aggregate that already exists.

=item C<NATE::Exceptions::Argument>

This type of exception is thrown when disklist and disktype values are passed while nacltask_no_ssd is true or when diskcount is not defined.

=item C<NACL::C::Exceptions::StorageAggregate::NotEnoughMatchingDisksAvailable>

This type of exception is thrown when there are not enough disks available while creating an aggregate.

=back

=back

=cut

sub create {
    $Log->enter() if $may_enter;
    my $pkg = shift;

    # We cannot directly use _if_exists_validate_spec since it defaults
    # the value to "die". This would mean that we would not know whether
    # if_exists or nacltask_if_exists was provided.
    my %opts = $pkg->_common_validate_with(
        params          => \@_,
        additional_spec => {

            # Kept for backwards compatibility
            if_exists => {
                type      => SCALAR,
                callbacks => $pkg->_if_exists_validate_callback(),
                optional  => 1
            },

            nacltask_if_exists => {
                type      => SCALAR,
                callbacks => $pkg->_if_exists_validate_callback(),
                optional  => 1
            },
            nacltask_wait          => { type => SCALAR,    default  => 1 },
            _was_created           => { type => SCALARREF, optional => 1 },
            nacltask_add_disks     => { type => HASHREF,   optional => 1 },
            nacltask_enable_hybrid => { type => SCALAR,    default  => 0 },
            nacltask_enable_composite => { type => SCALAR, default  => 0 },
            nacltask_object_store_name => { type => SCALAR, optional => 1 },
            nacltask_object_store_size_limit => { type => SCALAR, optional => 1 },
            nacltask_no_ssd        => { type => BOOLEAN,   optional => 1 },
            nacltask_min_size      => { type => SCALAR,    default  => 0 },
            nacltask_add_to_aggr_list_of_vserver =>
                { type => SCALAR, optional => 1 },
            $pkg->_cleanup_validate_spec(),
            nacltask_rng_args      => $pkg->_rng_args_validate_spec(),
            $pkg->_mcc_validate_spec(),
            nacltask_use_partitions =>
              { type => SCALAR, default => undef},
            blob =>
              { type => SCALAR, default => 0 },
            nacltask_verify_mvr =>
              { type => BOOLEAN, default => 0 },
            nacltask_storage_pool       => { type => SCALAR, optional => 1},
            nacltask_allocation_unit    => { type => SCALAR, default => 1, optional => 1},
            nacltask_reuse_storage_pool => { type => SCALAR, optional => 1},
            additional_disk_count       => { type => SCALAR, default  => 0},
            nacltask_initialize_compliance_clock => {type => BOOLEAN, default => 1},
        },
        ignore_primary_keys => 1,
        allow_extra => 1,
    );

    # Transform %opts from the options we received into the options to
    # pass to the base class method.

    my $if_exists
        = delete $opts{nacltask_if_exists}
        || $opts{if_exists}
        || 'die';

    # Having the "delete $opts{if_exists}" on the previous line would not
    # necessarily delete it, because it would not get evaluated if
    # the "delete $opts{nacltask_if_exists}" returned some value. Hence,
    # we're explicitly deleting it here separately.
    delete $opts{if_exists};
    my $blob                 = delete $opts{blob};
    my $wait                 = delete $opts{nacltask_wait};
    my $was_created          = delete $opts{_was_created};
    my $aggr_min_size        = delete $opts{nacltask_min_size};
    my $nacltask_no_ssd      = delete $opts{nacltask_no_ssd};
    my $nacltask_verify_mvr  = delete $opts{nacltask_verify_mvr};
    my $vserver_aggr_list
        = delete $opts{nacltask_add_to_aggr_list_of_vserver};
    my $nacltask_ssp_name       = delete $opts{nacltask_storage_pool};
    my $nacl_allocation_unit    = delete $opts{nacltask_allocation_unit};
    my $nacltask_reuse_storage_pool
        = delete $opts{nacltask_reuse_storage_pool};
    my $nacltask_to_check_mcc  = $opts{nacltask_to_check_mcc};
    my $initialize_compliance_clock = delete $opts{nacltask_initialize_compliance_clock};
    my $enable_composite = delete $opts{nacltask_enable_composite};
    my $nacl_object_store_name = delete $opts{nacltask_object_store_name};
    my $nacl_object_store_size_limit = delete $opts{nacltask_object_store_size_limit};

    # If a nacltask_storage_pool was specified we should default the aggregate to enable hybrid.
    # Only hybrid aggregates can have storage pool's added.
    my $enable_hybrid;
    if ($nacltask_ssp_name) {
        $enable_hybrid = 1;
        delete $opts{nacltask_enable_hybrid}
    } else {
        $enable_hybrid = delete $opts{nacltask_enable_hybrid};
    }

    # If the aggregate is specified for reuse, the aggregate create subroutine
    # should continue execution of the create routine if there is additional
    # logic to run after aggr is selected for reuse
    my $continue_after_reuse = 1 if ($if_exists eq 'reuse' and
                                     $aggr_min_size ||
                                     $enable_hybrid ||
                                     $enable_composite ||
                                     $nacltask_verify_mvr ||
                                     $nacltask_to_check_mcc);

    # The node could be specified through either node/nodes, and either
    # option can be specified as a scalar or a list, so handle all cases.
    my $node_name;
    foreach my $node_field (qw(node nodes)) {
        my $opt_val = $opts{$node_field};
        if (defined $opt_val) {
            if (ref $opt_val eq 'ARRAY') {
                $node_name = $opt_val->[0];
            } else {
                $node_name = $opt_val;
            }
            last;
        }
    }
    $node_name ||= 'local';

    my ( %opts_for_cleanup, %opts_for_register, $nacltask_to_cleanup, %mcc_opts );

    $pkg->_move_common_mcc_opts(
        source => \%opts,
        target => \%mcc_opts,
    );

    $pkg->_move_common_cleanup_opts(
        source => \%opts,
        target => \%opts_for_cleanup,
    );

    my %common_opts;
    $pkg->_copy_common_component_params(
        source => \%opts,
        target => \%common_opts
    );

    if ($nacltask_no_ssd) {
        if ($opts{disklist} || $opts{disktype}) {
            $Log->exit() if $may_exit;
            NATE::Exceptions::Argument->throw(
                "Option disklist and disktype are invalid when nacltask_no_ssd is \"true\""
            );
        } elsif (!$opts{diskcount}) {
            $Log->exit() if $may_exit;
            NATE::Exceptions::Argument->throw(
                "Option diskcount is mandatory when nacltask_no_ssd is \"true\""
            );
        } else {
            try {
                my $node;
                if ( defined $opts{'node'} && ref $opts{'node'} eq 'ARRAY' ) {
                    $node = $opts{'node'}->[0];
                } else {
                    $node = $opts{'command_interface'}->node();
                }

                $opts{disktype} = $pkg->_disktype_by_max_availability(
                    command_interface => $opts{command_interface},
                    not_of_type       => "SSD",
                    node              => $node,
                );
            }

            catch NACL::Exceptions::NoElementsFound with {
                my $exception = shift;
                my $aggr = NACL::CS::StorageAggregate->fetch(
                    command_interface => $opts{'command_interface'},
                    filter            => { aggregate => $opts{aggregate} },
                    allow_empty       => 1,
                    %common_opts,
                );

                if(!$aggr) {
                    $Log->exit() if $may_exit;
                    $exception->throw();
                }
            };
        } ## end else [ if ( $opts{disklist} ||...)]
    } ## end if ($nacltask_no_ssd)

#---------------------------------------------------------------------------
#Create aggregate according to the nacltask_use_partitions flag, if flag is set
#   aggr creation will happens with only Partitions  else with Physical disks.
# If the filer is having partition and physical disks then aggr will get created
#   with disk list else with disk count.
# if the user specified the disk list then aggr creation will happens with
#   provided disk list.
# If user passes STORAGE_TYPE,CHECKSUN_TYPE and RPM then it will look for only
#   specified disks, if not exception will be generated.
# If disk list is not provided and node is having mixed disk (i.e.partition+physical),
#   then only new code
#   will comes into picture else earlier Task code will handle the aggr create.
#If a value is not specified, then ONTAP chooses the disks to use
    %opts = $pkg->_add_partition_disklist(
        opts => {%opts, 'nodescope-node-name' => $node_name});
    $$was_created = 0;
    my %wait_for_creation_opts;
    $pkg->_copy_common_component_params(
        source => \%opts,
        target => \%wait_for_creation_opts
    );

    $wait = 1 if ($aggr_min_size);
    my $additional_disk_count = delete $opts{additional_disk_count};

    if (!$opts{'diskcount'} && !$opts{'disklist'} && !$opts{'partitionlist'}) {
        $opts{diskcount} = $pkg->get_min_diskcount(
            'command_interface' => $opts{'command_interface'},
            'raidtype'          => $opts{'raidtype'}
        );
    }
    # diskcount will be added with additional_disk_count
    if ( $additional_disk_count ) {
        if ( $opts{diskcount} && !$opts{'disklist'} && !$opts{'partitionlist'}) {
            $opts{diskcount} = $opts{diskcount} + $additional_disk_count;
        } else {
            $Log->exit() if $may_exit;
            NATE::Exceptions::Argument->throw(
                "Option additional_disk_count is invalid when either disklist or partitionlist is passed"
            );
        }
    }

    my $rng_args = delete $opts{nacltask_rng_args};
    my $random_name_generated;
    if ( !defined $opts{aggregate} ) {
        $opts{aggregate} = $pkg->random_name_generator(%$rng_args);
        $random_name_generated = 1;
    }

    my %convert_to_hybrid_opts;
    $pkg->_copy_common_component_params_with_ci(
        source => \%opts,
        target => \%convert_to_hybrid_opts,
    );
    $pkg->_hash_copy(
        source => \%opts,
        target => \%convert_to_hybrid_opts,
        copy   => [qw( polling_interval aggregate)]
    );
    $pkg->_hash_move(
        source => \%opts,
        target => \%convert_to_hybrid_opts,
        move   => [qw(nacltask_add_disks )]
    );
    $pkg->_hash_move(
        source => \%opts,
        target => \%wait_for_creation_opts,
        move   => [qw(polling_interval)]
    );
# w.r.t burt #998839, For sDOT setup with cluster size > 1 create mirrored aggr
# w.r.t burt #1004317, enforce the aggregate creation to be of mirrored by passing NACL_AGGR_MIRROR=1

    my $global_mirrored = param_global->get('NACL_AGGR_MIRROR');
    if (($opts{command_interface}->is_sdot_multiple_nodes()) && (!defined $opts{disklist}) && (!defined $opts{"mirror-disklist"})) {
        if ((!defined $opts{mirror}) || $global_mirrored ) {
            $opts{mirror} = 'true';
            $opts{diskcount} *= 2 if(defined $opts{diskcount}) ;
        }
    }
# w.r.t burt #1000739,  create raidtype as per passed in "NACL_AGGR_RAIDTYPE"
# raidtype should be passed as NACL_AGGR_RAIDTYPE=raid0 or NACL_AGGR_RAIDTYPE=raid_dp
    my $global_raidtype = param_global->get('NACL_AGGR_RAIDTYPE');
    if($global_raidtype) {
        $Log->comment("Setting raidtype to '$global_raidtype' because "
                . "parameter 'NACL_AGGR_RAIDTYPE' was provided");
        $opts{raidtype} = $global_raidtype ;
    }

# w.r.t burt1019997, for sDOT/AWS setup, enforce aggregate to have sidl enabled by passing NACL_AGGR_SIDL=1

    my $global_option_sidl = param_global->get('NACL_AGGR_SIDL');
    if ($global_option_sidl && (($global_option_sidl != 0) && ($opts{raidtype} !~ /raid0/i))) {
        NATE::Exceptions::Argument->throw("Sidl option is applicable for raid0 aggrs only");
    }
    $wait = 1 if ($global_option_sidl);

# If the nacltask_enable_hybrid => 1, or if nacltask_enable_composite => 1, we need to wait till the aggregate creation
# is complete. Hence overriding the $wait if nacltask_enable_hybrid => 1 or if nacltask_enable_composite => 1
    if ($enable_hybrid || $enable_composite) {
        $wait = 1;
    }

    # We want to know the job ID of the job performed by
    # SUPER::create, whether or not our caller does.
    my $dummy;
    $opts{job_component} ||= \$dummy;

    $opts{'method-timeout'} ||= AGGREGATE_TIMEOUT;

    my $command_interface = $opts{command_interface};

    my $self;
    my $found;
CREATE: {
    use warnings qw(exiting);
    my $sub_handle_if_exists = sub {
        my $exception = shift;
        $self = $pkg->_element_exists_handler(
            create_opts        => \%opts,
            nacltask_if_exists => $if_exists,
            exception          => $exception
        );
        if ($self) {
            $found = 1;
        }
        else {
            no warnings qw(exiting);
            redo CREATE;
        }
    };
    try {
        $self         = $pkg->SUPER::create(%opts);
        $$was_created = 1;
        $self->job_component( ${ $opts{job_component} } );
        if ($wait) {
            $self->wait_for_creation(%wait_for_creation_opts);
        }
    } ## end try
    catch NACL::C::Exceptions::StorageAggregate::AlreadyExists with {
        my $exception = shift;
        if ($random_name_generated) {
            $Log->comment( 'Random name collision, regenerate a '
                    . 'new name and retry' );
            $opts{volume} = $pkg->random_name_generator(%$rng_args);
            no warnings qw(exiting);
            redo CREATE;
        }
        else {
            $Log->trace('The aggregate already exists');
            $sub_handle_if_exists->($exception);
        }
    }
    catch
        NACL::C::Exceptions::StorageAggregate::NotEnoughMatchingDisksAvailable
        with {
        my $exception = shift;
        if ( $command_interface->is_7mode() )
        {
            my %common_component_options;
            $pkg->_copy_common_component_params_with_ci(
                source => \%opts,
                target => \%common_component_options,
            );
            my $aggrs = $pkg->find(
                %common_component_options,
                filter      => { aggregate => $opts{aggregate} },
                allow_empty => 1,
            );
            if ($aggrs) {
                $sub_handle_if_exists->($exception);
            } else {
                $Log->exit() if $may_exit;
                $exception->throw();
            }
        }
        else {
            $Log->exit() if $may_exit;
            $exception->throw();
        }

    }
    catch
        NACL::C::Exceptions::StorageAggregate::UnmirroredAggregatesNotSupported
            with {
        my $exception_obj = $_[0];
        if(defined($opts{'mirror'})) { #this value will be false
            $Log->exit() if $may_exit;
            $exception_obj->throw();
        } else {
            $opts{'mirror'} = "true";
            # The disk count is doubled for mirrored aggregates.
            $opts{'diskcount'} *= 2;
        }
        no warnings qw(exiting);
        redo CREATE;
    }
    catch NACL::C::Exceptions::StorageAggregate::SnaplockNotEnabled with {
        my $exception = $_[0];
        $exception->license_feature();
        no warnings qw(exiting);
        redo CREATE;
    }
    catch NACL::C::Exceptions::StorageAggregate::SnaplockEnterpriseNotEnabled with {
        my $exception = $_[0];
        if ( $command_interface->version_manager()->is_version_lt(
            release => $NACL::Global::ONTAP_Release_Longboard_0) ) {
            $exception->license_feature();
            no warnings qw(exiting);
            redo CREATE;
        } else {
            $Log->exit() if $may_exit;
            $exception->throw();
        }
    } catch NACL::C::Exceptions::StorageAggregate::ComplianceClockNotSet with {
        my $exception = $_[0];
        if ($initialize_compliance_clock) {
            # If $opts{node} is provided, that is the node on which the
            # aggr is being created.
            my $node = $opts{node} // $opts{command_interface}->node();
            require NACL::STask::SnaplockComplianceClock;
            NACL::STask::SnaplockComplianceClock->initialize(
                %common_opts,
                command_interface       => $opts{command_interface},
                node                    => $node,
                nacltask_if_initialized => 'pass',
            );
            no warnings;
            redo CREATE;
        } else {
            $Log->exit() if $may_exit;
            $exception->throw();
        }
    };
} ## end CREATE:

    if($global_option_sidl) {
    $Log->comment("Enabling single_instance_data_logging on aggregate ".
        "because parameter 'NACL_AGGR_SIDL' was provided");
        my $apiset = $command_interface->get_7m_or_nodescope_apiset();
    my $aggr_name = $self->aggregate();
        my $retries = 5;

SIDL_RETRY: {
    try {
        $apiset->aggr_options(
        aggregate => $aggr_name,
        'option-name' => "single_instance_data_logging",
        'option-value' => "on",
        );
        } catch NACL::APISet::Exceptions::ResponseException with {
            my $ex    = shift;
            my $error = $ex->text();
            if ( $error =~ /currently busy/s ) {
                --$retries;
                if ($retries) {
                    Tharn::snooze(10);
                    no warnings qw(exiting);
                    redo SIDL_RETRY;
                }
            } else {
                $ex->throw();
            }
        };
} #SIDL_RETRY
    }

    if ($vserver_aggr_list) {
        $self->add_to_aggr_list_of_vserver( vserver => $vserver_aggr_list );
    }
    if ($found and not $continue_after_reuse) {
        $Log->exit() if $may_exit;
        return $self;
    }

    $self->_check_in_metrocluster_node(
        command_interface => $opts{command_interface},
        method_used       => "create",
        %mcc_opts,
    );

    if($nacltask_verify_mvr) {
        my %pass_opts;
        $pass_opts{blob} = $blob if($blob);
        $pkg->verify_mirror_votes(
            command_interface => $opts{command_interface},
            aggregate         => $self->aggregate(),
            _method           => "create",
            %pass_opts,
        );
    }
# After the aggregate is online after creation, it will be in the state of dynamic
# grow for a while. During this stage, no operation aggregate is allowed.
# Since there are no field to query for this state, we can add some snooze after
# the aggregate comes online hoping it will fix this problem.
    Tharn::snooze(10) if ($enable_hybrid or $enable_composite or $aggr_min_size);

    if ($enable_hybrid) {
        if ($nacltask_ssp_name || $nacltask_reuse_storage_pool){
            $convert_to_hybrid_opts{nacltask_storage_pool}       = $nacltask_ssp_name;
            $convert_to_hybrid_opts{nacltask_allocation_unit}    = $nacl_allocation_unit;
            $convert_to_hybrid_opts{nacltask_reuse_storage_pool} = $nacltask_reuse_storage_pool;
            $convert_to_hybrid_opts{node}                        = $node_name;
            $convert_to_hybrid_opts{continue_after_reuse}        = $continue_after_reuse;
        }
        $pkg->convert_to_hybrid(%convert_to_hybrid_opts,%opts_for_cleanup);
    } ## end if ($enable_hybrid)

    if ($enable_composite) {
        my %object_store_attach_opts;
        $object_store_attach_opts{'command_interface'} = $opts{command_interface};
        $object_store_attach_opts{'aggregate'} = $self->aggregate();
        $object_store_attach_opts{'object-store-name'} = $nacl_object_store_name if ($nacl_object_store_name);
        $object_store_attach_opts{'size-limit'} = $nacl_object_store_size_limit if ($nacl_object_store_size_limit);
        NACL::STask::StorageAggregateObjectStore->attach(%object_store_attach_opts);

    } ## end if ($enable_composite)

    if ($aggr_min_size) {
        delete $opts{raidtype};
        $pkg->grow_size(
            %opts,
            nacltask_min_size => $aggr_min_size,
            allow_extra       => 1
        );
    } ## end if ($aggr_min_size)

    # Register this resource with the Cleanup Manager for cleanup
    $pkg->_copy_common_opts_for_cleanup(
        'source'              => {%opts, %opts_for_cleanup},
        'target'              => \%opts_for_register,
        'nacltask_to_cleanup' => \$nacltask_to_cleanup,
        'to_cleanup'          => 'purge'
    );
    $pkg->_register_for_cleanup(%opts_for_register)
        if ($nacltask_to_cleanup && $$was_created);

    $Log->exit() if $may_exit;
    return $self;
} ## end sub create

=head2 convert_to_hybrid

    NACL::STask::Aggregate->convert_to_hybrid(
        nacltask_wait      => $boolean,     # default '1'
        polling_interval   => $interval,    # default '10'
        command_interface  => $ci,
        'method-timeout'   => 300,
        aggregate          => $aggr,
        nacltask_add_disks => {
            'disklist' => [qw(v4.17 v4.18 v4.19)],
            'disktype' => 'SSD',
        }
    );

    or

    $aggr_task_obj->convert_to_hybrid(
        nacltask_wait      => $boolean,     # default '1'
        polling_interval   => $interval,    # default '10'
        'method-timeout'   => 300,
        nacltask_add_disks => {
            'disklist' => [qw(v4.17 v4.18 v4.19)],
            'disktype' => 'SSD',
        }
    );

(Class or instance method) This method will convert an existing aggregate
to a hybrid aggregate. This consists of making the aggregate hybrid-enabled
and adding the disks(SSD's) as specified. It also waits till all the
requested disks are added to aggregate. If it fails to modify the aggregate to
hybird_enabled in a given 'method-timeout' or default timeout (7200 seconds),
it will throw NACL::APISet::Exceptions::CommandFailedException.

=over

=item Options

=over

=item C<< "method-timeout" => 7200 >>

(Optional) As component method-timeout, controls how long the command
will wait before completing.  However, the default timeout has been
raised to 7200 seconds.

=item C<< aggregate=>$aggr_name >>

(Required for class method, Not Applicable for instance method)
Name of aggregate.

=item C<< nacltask_add_disks => $disks_to_add >>

(Optional)
If nacltask_add_disks is provided, the disks specified will be added.
This parameter is reference to a hash, which is same
as L<< NACL::C::Aggregate->add_disks|lib-NACL-C-Aggregate-pm/add_disks >>

=item C<< polling_interval => $interval >>

(Optional)
When nacltask_add_disks is provided, this is the interval at which the
disks added to the aggregate will be checked for to be in the 'INUSE'
state. As the "aggregate add-disks" command doesn't makes an entry in
the job table, this is the way we can ensure if the disks are successfully
added to the aggregate. Default value is 10 seconds.

=item command_interface, apiset_must, apiset_should, aggregate etc.

All of the other various options to
L<< NACL::C::Aggregate->modify|lib-NACL-C-Aggregate-pm/modify >>

=back

=back

=over

=item Exceptions

=over

=item C<NACL::APISet::Exceptions::CommandFailedException>

This type of exception is thrown when an attempt is made to modify the aggregate to hybrid_enabled in a given $timeout seconds.

=back

=back

=cut

sub convert_to_hybrid {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my $pkg        = ref($pkg_or_obj) || $pkg_or_obj;
    my %opts       = $pkg_or_obj->_common_validate_with(
        params          => \@_,
        additional_spec => {
            nacltask_add_disks                 => { type => HASHREF, default  => {} },
            polling_interval                   => { type => SCALAR,  optional => 1 },
            nacltask_wait                      => { type => SCALAR,  default  => 1 },
            nacltask_storage_pool              => { type => SCALAR|UNDEF,  optional => 1 },
            nacltask_allocation_unit           => { type => SCALAR,  default  => 1 },
            nacltask_reuse_storage_pool        => { type => SCALAR|UNDEF,  optional => 1 },
            $pkg->_cleanup_validate_spec(),
        },
        allow_extra => 1,
    );

   my %common_component_opts;
    $pkg_or_obj->_copy_common_component_params_with_ci(
        source => \%opts,
        target => \%common_component_opts
    );

    my $continue_after_reuse = delete $opts{continue_after_reuse};
    my $storage_pool_added = 0;
    my $is_hybrid =  $pkg->is_hybrid(
                         aggregate => $opts{aggregate},
                         %common_component_opts
                     );

    # If the aggregate is already hybrid active do nothing (burt520394)
    # We should proceed however if we have determined that additional logic should be run
    if ($is_hybrid and !$continue_after_reuse)
    {
        $Log->debug(
            "Aggregate is already hybrid, hence returning without converting and adding disks"
        );
        return;
    } ## end if ( $pkg->is_hybrid( ...))

    my %add_disks_opts = %{ delete $opts{nacltask_add_disks} };
    my $raid_type;

    #If nacltask_storage_pool flag is set, then create SSPs .
    #if SSP is already available then check the SSP is having the required spare capacity
    #if SSP is not having enough spare capacity then create SSP with new name.
    if ($opts{nacltask_storage_pool} || $opts{nacltask_reuse_storage_pool}) {
        $raid_type = $add_disks_opts{raidtype};
        my %ssp_create_hash;

        $pkg->_move_common_cleanup_opts(
            source => \%opts,
            target => \%ssp_create_hash,
        );

        if (defined $add_disks_opts{'disk-list'}) {
            $ssp_create_hash{'disk-list'}
                = $add_disks_opts{'disk-list'};
        } else {
            my %determine_disk_list;
            $pkg->_hash_copy(
                source => \%add_disks_opts,
                target => \%determine_disk_list,
                copy   => [qw(diskcount raidtype)],
            );
            $determine_disk_list{node} = $opts{node};
            $ssp_create_hash{nacltask_determine_disk_list}
                = {%determine_disk_list};
        }

        if (defined $opts{nacltask_storage_pool}) {
            $ssp_create_hash{'storage-pool'} = $opts{nacltask_storage_pool};
        }

        if ($opts{nacltask_storage_pool} && !$opts{nacltask_reuse_storage_pool}) {
            my $pool = NACL::C::StoragePool->find(
                %common_component_opts,
                filter            => {'storage-pool' => $opts{nacltask_storage_pool}},
                allow_empty       => 1,
            );
            if (!defined $pool) {
                NACL::STask::StoragePool->create(%ssp_create_hash,
                    %common_component_opts);
            }
        } else {
            # If we are reusing the storage pool, we need to check to see if the pool
            # has already been added to the aggregate
            my $pool_component_state = NACL::CS::StoragePoolShowAggregate->fetch(%common_component_opts,
                                                                                 filter => {'aggregate' => $opts{aggregate},
                                                                                            'storage-pool' => $opts{'nacltask_storage_pool'}},
                                                                                 allow_empty => 1);
            $storage_pool_added = 1 if ($pool_component_state);

            my $pool = NACL::C::StoragePool->show_spare_capacity(
                %common_component_opts,
                filter => {
                    'allocation-unit-count' => '>=' . $opts{nacltask_allocation_unit},
                    node                    => $opts{node},
                },
                allow_empty => 1
            );
            if (defined $pool) {
                $Log->trace("Info: Reusing the existing SSP");
                $ssp_create_hash{'storage-pool'} = $pool->storage_pool();
            } else {
                  my $new_pool = NACL::STask::StoragePool->create(%ssp_create_hash,
                  %common_component_opts,
                  nacltask_if_exists => 'reuse');
                  $ssp_create_hash{'storage-pool'} = $new_pool->storage_pool();
            }
        }

        %add_disks_opts = ();
        $add_disks_opts{'allocation-units'} = $opts{nacltask_allocation_unit};
        $add_disks_opts{'storage-pool'}     = $ssp_create_hash{'storage-pool'};
        $add_disks_opts{'raidtype'}         = $raid_type if(defined($raid_type));
        delete $opts{node};
    }

    my $add_disks = keys %add_disks_opts;
    $pkg_or_obj->_copy_common_component_params_with_ci(
        source => \%opts,
        target => \%add_disks_opts
    );
    $pkg_or_obj->_hash_move(
        source => \%opts,
        target => \%add_disks_opts,
        move   => [qw(polling_interval nacltask_wait)]
    );
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%add_disks_opts,
        copy   => [ $pkg_or_obj->get_primary_keys() ]
    );
    my $wait = $add_disks_opts{nacltask_wait};

    try {
        if (!$is_hybrid) {
            $pkg->modify( %opts, 'hybrid-enabled' => 'true' );
            $is_hybrid = 1;
        }
    }
    catch NACL::APISet::Exceptions::CommandFailedException with {
        my $ex = shift;
        $Log->debug(
            "Attempt to modify the aggregate failed:\n" . $ex->text() );

        # Ignoring the error if the aggregate is already hybrid-enabled
        $ex->throw() if ( $ex->text() !~ /is already a hybrid-enabled/ );
    };

    if ($add_disks and !$storage_pool_added) {
        $pkg->add_disks(%add_disks_opts);
    }
    $Log->exit() if $may_exit;
    return;
} ## end sub convert_to_hybrid

=head2 add_disks

    NACL::STask::Aggregate->add_disks(
        nacltask_wait     => $boolean,     # default '1'
        polling_interval  => $interval,    # default '10'
        command_interface => $ci,
        aggregate         => $aggr,
        diskcount         => "3"
    );

    or

    aggr_task_obj->add_disks(
        nacltask_wait      => $boolean,     # default '1'
        polling_interval   => $interval,    # default '10'
        diskcount          => "3",
        disktype           => "SSD",
    );

(Class or instance method) Add disks to an existing aggregate.
This method provides additional service of waiting till the disks
get added. This wait is helpful as add-disks job doesn't gets registered
in the job table.

Cleanup for this action cannot be done, because once added the disks
cannot be removed from the aggregate.

=over

=item Options

=over

=item C<< nacltask_wait => 0|1 >>

(Optional) If 1 (the default), wait for this aggregate to be
created (see "wait_for_add_disks" method, below).  If 0, do not
wait any longer than necessary (any longer than the aggregate add
disk command waits).

=item C<< "method-timeout" => 120 >>

(Optional) As component method-timeout, controls how long the add-disks command
will wait before completing.  Default timeout is 120 seconds. This timeout is
not applicable for used for the wait_for_add_disks method (that uses an
infinite timeout)

=item C<< polling_interval => $interval >>

(Optional)
If nacltask_wait is set to 1, this is the interval at which to poll the status
of the aggregate to know if all the disks have been added. or the interval at
which to poll the filer for the value of the "raidstatus" field of the aggregate.

=item command_interface, apiset_must, apiset_should, aggregate, diskcount, disklist, voltype, etc.

All of the other various options to
L<< NACL::C::Aggregate->add_disks|lib-NACL-C-Aggregate-pm/add_disks >>

=back

=back

=cut

sub add_disks {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my $pkg        = ref($pkg_or_obj) || $pkg_or_obj;
    my %opts       = $pkg_or_obj->_common_validate_with(
        params          => \@_,
        additional_spec => {
            nacltask_wait    => { type => SCALAR, default => 1 },
            polling_interval => { type => SCALAR, default => POLL_DELTA }
        },
        allow_extra => 1,
    );

    # Transform %opts from the options we received into the options to
    # pass to the base class method.

    my $wait = delete $opts{nacltask_wait};

    my $retries = 5;

    my %wait_for_add_disks_opts;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%wait_for_add_disks_opts,
        copy   => [qw(method-timeout aggregate command_interface)]
    );
    $pkg_or_obj->_hash_move(
        source => \%opts,
        target => \%wait_for_add_disks_opts,
        move   => [qw(polling_interval)]
    );
    $opts{'method-timeout'} ||= 120;

AGAIN: {

        # try adding the disks. If aggregate is busy (dynamic grow), retry
        # five times.
        try {
            $pkg->SUPER::add_disks(%opts);
        } catch NACL::C::Exceptions::StorageAggregate::StoragePoolBusy with {
            my $ex     = shift;
            my $job_id = $ex->get_job_id();
            if (defined $job_id) {
                my %common_opts;
                $pkg->_copy_common_component_params(
                    source => \%opts,
                    target => \%common_opts,
                );
                my $job_component = NACL::C::Job->new(
                    command_interface => $opts{command_interface},
                    id                => $job_id,
                );
                $job_component->wait_on_job(%common_opts,
                    'method-timeout' => 1800);
                no warnings qw(exiting);
                redo AGAIN;
            } else {
                $Log->exit() if $may_exit;
                $ex->throw();
            }
        }
        catch NACL::APISet::Exceptions::ResponseException with {
            my $ex    = shift;
            my $error = $ex->text();
            if ( $error =~ /currently busy/s ) {
                --$retries;
                if ($retries) {
                    Tharn::snooze(10);
                    no warnings qw(exiting);
                    redo AGAIN;
                }
            } else {
                $ex->throw();
            }
        };
    } ## end AGAIN:
    if ($wait) {
        $pkg->wait_for_add_disks(%wait_for_add_disks_opts);
    }
    $Log->exit() if $may_exit;
    return;
} ## end sub add_disks

=head2 wait_for_add_disks

    NACL::STask::Aggregate->wait_for_add_disks(
      command_interface => $command_interface,
      aggregate         => $aggregate,
      polling_interval  => 20,
    );

    or

    $aggr_task_obj->wait_for_add_disks( polling_interval  => 20 );

(Class or instance method) This method waits till all the requested disks
are added to aggregate. The method will wait for the raidstatus of the given
aggregate and when it reaches the value 'normal', it is considered that the
disks have been added.

=over

=item Options

=over

=item C<< aggregate=>$aggr_name >>

(Required for class method, Not Applicable for instance method)
Name of aggregate.

=item C<< polling_interval => $interval >>

(Optional)
This is the interval at which the raidstatus of the aggregate is polled till
it reaches the value 'normal' As the "aggregate add-disks"
command doesn't makes an entry in the job table, this is the way we can
wait for add-disks.
default value is 10 seconds.

=item command_interface, apiset_must, apiset_should, aggregate etc.

=back

=back

=cut

sub wait_for_add_disks {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my $pkg        = $pkg_or_obj->get_package_name();
    my %opts       = $pkg_or_obj->_common_validate_with(
        params => \@_,
        additional_spec =>
            { polling_interval => { type => SCALAR, default => 10 } },
    );

    if ( !$opts{'method-timeout'} ) {
        my %calculate_method_timeout_opts;
        $pkg->_hash_copy(
            source => \%opts,
            target => \%calculate_method_timeout_opts,
            copy   => [qw(aggregate command_interface)]
        );
        $opts{'method-timeout'}
            = $pkg->_calculate_method_timeout(%calculate_method_timeout_opts);
    } ## end if ( !$opts{'method-timeout'...})

    $pkg->wait_on_attribute(
        attribute_to_check => 'raidstatus',
        till_not_value     => qr/\bgrowing\b/,
        %opts
    );

    $Log->exit() if $may_exit;
} ## end sub wait_for_add_disks

=head2 purge

    $aggregate->purge(%other_options);

    NACL::STask::Aggregate->purge(
      command_interface => $command_interface,
      aggregate         => $aggregate,
      %other_options
    );

(Class or instance method) This method deletes an aggregate,
Once the aggregate is deleted , the Ontap will automatically
removes from the vserver's aggr-list, as well as detaching or
deleting anything preventing this aggregate from being
deleted.  Currently this includes:


=over

=item *

If the aggregate cannot be deleted because it contains flexible
volumes, the flexible volumes on the aggregate will be purged as well ( see
L<< NACL::STask::Volume->purge | lib-NACL-STask-Volume-pm/purge >> )

=back

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "method-timeout" => 7200 >>

(Optional) As component method-timeout, controls how long the command
will wait before completing.  However, the default timeout has been
raised to 7200 seconds.

=item C<< aggregate=>$aggr_name >>

(Required for class method, Not Applicable for instance method)
Name of aggregate.

=item C<< _was_deleted => \$scalar >>

(Optional)
When this option is provided a reference to a scalar variable, the variable
gets filled in with a boolean value describing whether the aggregate
did not exist (value will be 0; this scenario is possible when
nacltask_if_purged => "pass") or whether the aggregate was deleted
(value will be 1).

=item C<< if_purged => die | pass >>

B<DEPRECATED>.
This option is deprecated. Use C<nacltask_if_purged> instead.

=item C<< nacltask_if_purged => die | pass >>

(Optional, defaults to "die")
Specifies what to do if the aggregate we're trying to purge no longer
exists. A value of C<die> (the default) will result in an exception
being thrown. A value of C<pass> will mean that the exception that would
normally occur when we're trying to delete a aggregate that does not exist
is suppressed.

=item C<< nacltask_verify=>$nacltask_verify_boolean >>

(Optional)
If '0' (default), verification will not be performed.
If '1', verification will be performed to ensure that the deletion did happen successfully.

=item C<MCC options>

The options accepted for MCC configuration replication verification is
documented at L<MCC options|lib-NACL-STask-STask-pm/mcc_options>.

=item C<< nacltask_verify_mvr => $boolean >>

(Optional, defaults to 0)
This option is to check whether the MVR(mirror vote records) check needs to be done
for the aggregate to be deleted.

=item C<< purge_flexgroup_vols => $boolean >>

(Optional, defaults to 0)
Flexgroup is deleted before purging an aggregate if value of this option is set to 1
and purging an aggregate which is hosting a flexgroup will throw an exception if value of this option is set to 0

=back

=back

=over

=item Exceptions

=over

=item C<NACL::Exceptions::VerifyFailure>

This type of exception is thrown when verification fails for the deleted aggregate.

=item C<NACL::Exceptions::VserverHasVolumesOnAnotherAggregate>

This type of exception is thrown when vserver has volumes on another aggregate.

=back

=back

=cut

sub purge {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %orig_opts  = @_;

    # Cannot use _if_action_completed_already_validate_spec directly since it
    # defaults the value to 'die' so we won't know whether if_purged or
    # nacltask_if_purged was provided by the caller.
    my $additional_spec = {

        # if_purged kept for backwards compatibility
        if_purged => {
            type     => SCALAR,
            optional => 1,
            callbacks =>
                $pkg_or_obj->_if_action_completed_already_validate_callback()
        },
        nacltask_if_purged => {
            type     => SCALAR,
            optional => 1,
            callbacks =>
                $pkg_or_obj->_if_action_completed_already_validate_callback()
        },
        nacltask_verify => { type => BOOLEAN, default => 0 },
        $pkg_or_obj->_mcc_validate_spec(),
        nacltask_verify_mvr => { type => BOOLEAN, default => 0 },
        _was_deleted        => { type => SCALARREF, optional => 1 },
        purge_flexgroup_vols => { type => BOOLEAN, default => 0 },
    };

    # Throw away options specific to the task
    $pkg_or_obj->_hash_move(
        source => \%orig_opts,
        target => {},
        move   => [ keys %{$additional_spec} ]
    );
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@_,
        additional_spec => $additional_spec,
    );

    my $nacltask_verify_mvr = delete $opts{nacltask_verify_mvr};
    my $purge_flexgroup_vols = delete $opts{purge_flexgroup_vols};
    my %mcc_opts;
    $pkg_or_obj->_move_common_mcc_opts(
        source => \%opts,
        target => \%mcc_opts,
    );

    my %common_opts;
    $pkg_or_obj->_copy_common_component_params_with_ci(
        source => \%opts,
        target => \%common_opts
    );
    my $if_purged = $opts{nacltask_if_purged} || $opts{if_purged} || 'die';
    my $was_deleted = delete $opts{_was_deleted};
    $$was_deleted = 0;
    my $already_purged;
    my $aggr_state;

    # Check if the aggr is in failed state
    my @requested_fields = ('state');
    push @requested_fields, 'uuid' if ($nacltask_verify_mvr);
    my $aggr_state_obj = NACL::CS::Aggregate->fetch(
        %common_opts,
        filter             => { 'aggregate' => $opts{'aggregate'} },
        'requested_fields' => \@requested_fields,
        'allow_empty'      => 1,
    );

    my $aggr_uuid;
    if ($aggr_state_obj) {
        $aggr_state = $aggr_state_obj->state();

        # if Cmode, do "vreport fix" for all the volume of that aggregate
        # if it has volumes.
        if (   $aggr_state eq "failed"
            && $common_opts{'command_interface'}->is_cmode() )
        {
            $pkg_or_obj->_handle_failed_aggr_for_purge( %common_opts,
                'aggregate' => $opts{'aggregate'} );
        } ## end if ( $aggr_state eq "failed"...)

        $aggr_uuid = $aggr_state_obj->uuid() if ($nacltask_verify_mvr);
    } ## end if ($aggr_state_obj)
    my $is_sys_volume_present = 0;
AGAIN: {
        use warnings qw(exiting);
        my $job;
        try {

            # Skip offlining the aggregate before destroying the aggregate.
            # if the aggregate is in the failed state.
            if ( $is_sys_volume_present == 0) {
                $pkg_or_obj->modify(

                    # apparently this can take a while if, for
                    # example, the aggr is still being created.
                    "method-timeout" => AGGREGATE_TIMEOUT,
                    %orig_opts,
                    state                => "offline",
                    nacltask_if_modified => 'pass'
                ) unless ($aggr_state =~ /failed|offline/i);
                $pkg_or_obj->delete( %orig_opts, job_component => \$job );
            }
            else {
                $pkg_or_obj->delete( %orig_opts, job_component => \$job );
                delete $orig_opts{force};
            }
            $$was_deleted = 1;
            if ($job) {
                $Log->trace("Wait on the job ID");
                $pkg_or_obj->wait_on_job(
                    job_component    => $job,
                    'method-timeout' => AGGREGATE_TIMEOUT,
                    %orig_opts
                );
            } ## end if ($job)
        } ## end try
        catch NACL::C::Exceptions::StorageAggregate::ContainsVolumes with {

                # Can't offline or delete volumes on an offline aggregate
                # in C-mode, so online it just in case that helps

                try {
                    $pkg_or_obj->modify( %orig_opts, state => 'online' );
                }
                otherwise {

                # Probably already online.  Even if not, let's keep
                # going and maybe volume purge will still succeed.
                };

                # Volume->find() works in a mode-agnostic fashion, but
                # is not the optimal way of finding all volumes within
                # an aggregate in 7Mode. This is because in 7Mode it will
                # end up listing all volumes in verbose mode. However,
                # "aggr status" has a field ("volumes") which is the list
                # of all volumes within that aggregate.

                # CMode returns a "create-time" field for volumes. If we
                # arrange the volumes in reverse order of creation (i.e.
                # descending order for the create-time field), then we should
                # be able to delete them without the delete command failing
                # because of there being clones or it being a root-volume
                my @volumes;
                my $command_interface = $common_opts{command_interface};
                if ( $command_interface->mode() eq 'CMode' ) {
                    my @requested_fields;
                    @requested_fields = ($purge_flexgroup_vols)? [qw(create-time is-constituent flexgroup-msid)] : [qw(create-time)];
                    my @cs_objects = NACL::CS::Volume->fetch(
                        %common_opts,
                        filter => {
                            aggregate => $opts{aggregate},
                            type      => '!DEL',
                        },
                        requested_fields => @requested_fields
                    );
                    foreach my $vol (@cs_objects) {
                        if($vol->volume() =~ 'MDV_*') {
                            $orig_opts{force} = 'true';
                            $is_sys_volume_present = 1;
                            redo AGAIN;
                        }
                    }
                    my @arranged_cs_objects = arrange_by_create_time(
                        volume_cs_objects => \@cs_objects );
                    foreach my $cs (@arranged_cs_objects) {
                        if($purge_flexgroup_vols) {
                            my $constituent;
                            $constituent = $cs->is_constituent() if (defined $cs->is_constituent());
                            if($constituent eq 'true'){
                                my $fg_msid = $cs->flexgroup_msid();
                                my $vol = NACL::STask::Volume->find(
                                        %common_opts,
                                        filter => {
                                           'flexgroup-msid' => $fg_msid,
                                           'is-constituent' => '!true',
                                        },
                                    );
                                push @volumes,$vol;
                            }
                            else{
                                push @volumes, $cs->get_task_instance();
                            }
                        }
                        else{
                            push @volumes, $cs->get_task_instance();
                        }
                    }
                } else {
                    my $aggr_cs = NACL::CS::Aggregate->fetch(
                        %common_opts,
                        filter => {aggregate => $opts{aggregate}},
                        requested_fields => [qw(volumes)],
                    );
                    my @volume_names = $aggr_cs->volumes();
                    foreach my $volume_name (@volume_names) {
                        my $volume = NACL::STask::Volume->new(
                            command_interface => $command_interface,
                            volume            => $volume_name
                        );
                        push @volumes, $volume;
                    } ## end foreach my $volume_name (@volume_names)
                } ## end else [ if ( $command_interface...)]

                foreach my $vol (@volumes) {
                    my $pkg = $pkg_or_obj->get_package_name();
                    $pkg->_purge_volumes_and_clones(
                        %common_opts,
                        volume_obj => $vol,
                        aggregate  => $opts{aggregate}
                    );
                } ## end foreach my $vol (@volumes)
                no warnings qw(exiting);
                redo AGAIN;
        }
        catch NACL::C::Exceptions::StorageAggregate::DoesNotExist with {
            my $e = shift;
            $already_purged = 1;

            # Check if the command failed because the aggregate
            # does not exist
            if ( $if_purged !~ /pass/i ) {
                $Log->exit() if $may_exit;
                   $e->throw();
            }
        };
    } ## end AGAIN:

    if ( !$already_purged ) {
        if ( $opts{nacltask_verify} ) {
            $pkg_or_obj->taskverify_purge(%orig_opts);
        }
        try {

            # Start zeroing spares so that the next attempt to create
            # an aggregate can use the already zeroed spares.
            NACL::STask::StorageDisk->zerospares(
                command_interface => $opts{command_interface} );
        } ## end try
        catch NACL::APISet::Exceptions::ResponseException with {

            # Failure of the "disk zerospares" command should not be
            # reported as a failure of purge().
            my $exception = shift;
            $Log->warn( "Call to zerospares failed:\n" . $exception->text() );
        };
    } ## end if ( !$already_purged )

    $pkg_or_obj->_check_in_metrocluster_node(
        command_interface => $common_opts{'command_interface'},
        method_used       => "purge",
        aggregate         => $opts{aggregate},
        %mcc_opts,
    );

    if($nacltask_verify_mvr) {
        my $return_hash = $pkg_or_obj->get_mirror_vote(
            command_interface => $common_opts{command_interface},
            aggregate         => $opts{'aggregate'},
            _method           => "purge",
            uuid              => $aggr_uuid,
        );

        if(!defined($return_hash->{dblade_info})) {
            $Log->comment("The mirror vote records are deleted successfully ".
                "from the dblade table");
        } else {
            $Log->exit() if $may_exit;
            NACL::STask::Exceptions::MVRcheckFailure
                ->throw("The dblade entry for the aggregate".
                    " is still persistent even after its deletion");
        }
    }

    $Log->exit() if $may_exit;
} ## end sub purge

# Purge all volumes and clones of it.
# Turns out that clones of a clone can be created, so this method has been
# implemented recursively.
# This logic is needed for 7Mode and in case the logic of deleting volumes
# in reverse order of creation for CMode is not full proof
sub _purge_volumes_and_clones {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    my %opts = $pkg->_common_validate_with(
        params              => \@args,
        ignore_primary_keys => 1,
        additional_spec     => {
            volume_obj => { isa  => 'NACL::STask::Volume', type => OBJECT },
            aggregate  => { type => SCALAR },
            command_interface => { isa => 'NACL::C::CommandInterface::ONTAP',
                type => OBJECT, optional => 1 }
        }
    );
    my $vol_obj         = delete $opts{volume_obj};
    my $aggregate       = delete $opts{aggregate};
    unless (exists $opts{command_interface}) {
        $opts{command_interface} = $vol_obj->command_interface;
    }
    my %opts_without_ci = %opts;
    delete $opts_without_ci{command_interface};

    # By this point %opts has only the common options (with ci)

    try {

        # Need to send nacltask_if_purged as pass because the order in
        # which the volumes get deleted can get re-ordered dynamically
        # based on which volumes are clones and root-volumes
        $vol_obj->purge(%opts_without_ci, nacltask_if_purged => 'pass');
    }
    catch NACL::C::Exceptions::Volume::FailedToDeleteJunction with {
        # Unmount failed because vserver's root-volume is offline.
        # Online it and retry it.
        NACL::STask::Volume->modify(
            %opts,
            extended_query => {
                vserver => $vol_obj->vserver(),
                vsroot  => 'true',
                type    => '!LS',
            },
            state => 'online',
        );
        $pkg->_purge_volumes_and_clones(
            %opts,
            volume_obj => $vol_obj,
            aggregate  => $aggregate
        );
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = shift;
        if ($exception->text() =~ /has clone volumes/            # 7Mode
            || $exception->text() =~ /has one or more clones/    # CMode
            )
        {

            # Need to purge clones of a volume before the volume can be
            # purged

            # NACL::C::VolumeClone->find() works for both CMode and 7Mode.
            # However, in 7Mode this requires getting details of all volumes.
            # Instead, 7Mode directly returns a field with all the clone
            # volumes contained within a volume. We use that in 7Mode to
            # shorten the size of the log.
            my @vol_clones;
            if ( $opts{command_interface}->mode() eq 'CMode' ) {
                my @clone_objs = NACL::C::VolumeClone->find(
                    %opts,
                    filter => {
                        'parent-volume' => $vol_obj->volume(),
                        vserver         => $vol_obj->vserver()
                    }
                );

                foreach my $each_clone (@clone_objs) {
                    my $clone_obj = NACL::STask::Volume->new(
                        command_interface => $each_clone->command_interface(),
                        volume            => $each_clone->flexclone(),
                        vserver           => $each_clone->vserver()
                    );
                    push @vol_clones, $clone_obj;
                } ## end foreach my $each_clone (@clone_objs)
            } else {
                my @clones_names
                    = $vol_obj->get_one_state_attribute('clones-list');
                foreach my $each_clone (@clones_names) {
                    my $clone_obj = NACL::STask::Volume->new(
                        command_interface => $vol_obj->command_interface(),
                        volume            => $each_clone
                    );
                    push @vol_clones, $clone_obj;
                } ## end foreach my $each_clone (@clones_names)
            } ## end else [ if ( $opts{command_interface...})]

            foreach my $vol_clone (@vol_clones) {
                $pkg->_purge_volumes_and_clones(
                    %opts,
                    volume_obj => $vol_clone,
                    aggregate  => $aggregate
                );
            } ## end foreach my $vol_clone (@vol_clones)

            # Now that all the clones have been purged, re-try deleting the
            # volume
            $vol_obj->purge(%opts_without_ci);
        } elsif ( $exception->text()
            =~ /Cannot delete root volume because the Vserver has volumes/ )
        {

          # Need to delete all other volumes of a vserver to be able to delete
          # its root volume. Check whether all the other volumes of this
          # vserver are on this aggregate, error out if not
            my @state_objs = NACL::CS::Volume->fetch(
                %opts,
                filter => {
                    vserver => $vol_obj->vserver(),
                    vsroot  => 'false',
                    type    => '!DEL',
                },
                requested_fields => [qw(aggregate create-time)]
            );
            my @not_on_aggr;
            foreach my $state_obj (@state_objs) {
                if ( $state_obj->aggregate() ne $aggregate ) {
                    push @not_on_aggr, $state_obj;
                }
            }
            if (@not_on_aggr) {
                my $msg
                    = "Cannot delete the volume '"
                    . $vol_obj->volume()
                    . "' which is the rootvolume of vserver '"
                    . $vol_obj->vserver()
                    . "' because one or more of the "
                    . 'volumes in this vserver is not present on '
                    . "aggregate '$aggregate'\n";
                my @volumes;
                foreach my $state_obj (@not_on_aggr) {
                    $msg
                        .= "Volume '"
                        . $state_obj->volume()
                        . "' is on aggregate '"
                        . $state_obj->aggregate() . "'\n";
                    push @volumes,
                        {
                        volume    => $state_obj->volume(),
                        aggregate => $state_obj->aggregate()
                        };
                } ## end foreach my $state_obj (@not_on_aggr)
                $Log->exit() if $may_exit;

                NACL::Exceptions::VserverHasVolumesOnAnotherAggregate->throw(
                    text              => $msg,
                    command_interface => $opts{command_interface},
                    vserver           => $vol_obj->vserver(),
                    aggregate         => $aggregate,
                    volumes           => \@volumes
                );
            } else {

                # All volumes in the vserver are in this aggregate, purge
                # each of them
                my @arranged_cs_objects;
                if ( $opts{command_interface}->mode() eq 'CMode' ) {
                    @arranged_cs_objects = arrange_by_create_time(
                        volume_cs_objects => \@state_objs );
                } else {
                    @arranged_cs_objects = @state_objs;
                }
                foreach my $state_obj (@arranged_cs_objects) {
                    my $vol_stask = $state_obj->get_task_instance();
                    $pkg->_purge_volumes_and_clones(
                        volume_obj => $vol_stask,
                        aggregate  => $aggregate
                    );
                } ## end foreach my $state_obj (@arranged_cs_objects)

                # Now that all other volumes in the vserver have been purged
                # we can delete the rootvolume
                $vol_obj->purge(%opts_without_ci);
            } ## end else [ if (@not_on_aggr) ]
        } else {
            $Log->exit() if $may_exit;
            $exception->throw();
        }
    };

    $Log->exit() if $may_exit;
} ## end sub _purge_volumes_and_clones

=head2 purge_aggrs_not_containing_vsroot_vols

    # Purges all aggregates that do not contain a rootvolume os a vserver
    NACL::STask::Aggregate->purge_aggrs_not_containing_vsroot_vols(
        command_interface => $node);

    # Accepts all of the options of "find_aggrs_not_containing_vsroot_vols"
    # so it is also possible to further filter the aggregates to purge.
    # This example purges aggrs not containing any vsroot vol AND which
    # are smaller than 1TB.
    NACL::STask::Aggregate->purge_aggrs_not_containing_vsroot_vols(
        command_interface => $node,
        filter            => { size => '<1TB' }
    );

(Class method only, not applicable as an instance method)

Purges all aggregates that do not contain the rootvolume of a vserver.
It is also possible to further filter down the list of which of these
aggregates to purge (all of the arguments accepted by
L<find_aggrs_not_containing_vsroot_vols|lib-NACL-C-StorageAggregate-pm/find_aggrs_not_containing_vsroot_vols>
are accepted by this method.)

=cut

sub purge_aggrs_not_containing_vsroot_vols {
    $Log->enter() if $may_enter;
    my ($pkg, %opts) = @_;

    my @aggrs = $pkg->find_aggrs_not_containing_vsroot_vols(
        allow_empty => 1,
        %opts
    );

    my %common_params;
    $pkg->_copy_common_component_params(
        source => \%opts,
        target => \%common_params
    );

    map { $_->purge(%common_params) } @aggrs;

    $Log->exit() if $may_exit;
}

=head2 wait_for_creation

    $aggregate->wait_for_creation(
      'method-timeout'  => $timeout,
    );

    NACL::STask::Aggregate->wait_for_creation(
        'method-timeout'  => $timeout,
        command_interface => $command_interface,
        aggregate         => $aggregate_name,
        job_component     => $job_component,
    );

Wait for the aggregate creation to complete.  If a job for creating
this aggregate is known (because the mode is C-mode and
NACL::STask::Aggregate->create was used to create this object, or
because it was passed in using the "job_component" option) then this
waits for that job.

Otherwise, it polls until the "state" of the aggregate is "online".
If the state of the aggregate becomes something unexpected (currently
something other than "creating" or "online") then a
L<NACL::Exceptions::UnexpectedAVState|lib-NACL-Exceptions-UnexpectedAVState-pm>
exception is thrown.

=over

=item Options

=over

=item C<< job=>$job_component >>

(Optional for class method, Not Applicable for instance method)
A NACL::C::Job object representing the C-mode Job used to
create this aggregate.

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< aggregate=>$aggr_name >>

(Required for class method, Not Applicable for instance method)
Name of aggregate.

=item C<< "method-timeout"=>$timeout >>

(Optional) How long in seconds to wait for the job to finish or for
the state to become "online".  Defaults to 7200 (2 hours).

=item C<< polling_interval => $interval >>

(Optional)
The interval, in seconds, at which to either show the progress of the job (if
we're watching its progress) or the interval at which to poll the value of the
"state" field. This defaults to 10 seconds if not provided.

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=back

=back

=over

=item Exceptions

=over

=item C<NACL::APISet::Exceptions::TimeoutException>

This type of exception is thrown when newly created aggregate does not come online even after waiting for $timeout seconds.

=back

=back

=cut

sub wait_for_creation {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts = $pkg_or_obj->_common_validate_with(
        params => \@args,
        additional_spec   => {
            $pkg_or_obj->_job_validate_spec(),
            polling_interval => { type => SCALAR, default => POLL_DELTA },
        },
    );
    %opts       = $pkg_or_obj->_extract_job(%opts);
    my $was_method_timeout_passed;
    my $pkg = $pkg_or_obj->get_package_name();
    if ( defined $opts{'method-timeout'} ) {
        $was_method_timeout_passed = 1;
    } else {
        $opts{'method-timeout'} = AGGREGATE_TIMEOUT;
    }

    AGAIN: {
        try {
            use warnings;
            $pkg->wait_for_completion(
                %opts,
                attribute_to_check        => 'state',
                till_value                => 'online',
                valid_intermediate_states => $pkg->_valid_intermediate_states(),
            );
        } ## end try
        catch NACL::Exceptions::Timeout with {
            my $ex = shift;
            if ($was_method_timeout_passed) {
                $ex->throw();
                $Log->exit() if $may_exit;
            } else {
                my %calculate_method_timeout_opts;
                $pkg->_hash_copy(
                    source => \%opts,
                    target => \%calculate_method_timeout_opts,
                    copy   => [qw(aggregate command_interface)]
                );
                $opts{'method-timeout'} = $pkg->_calculate_method_timeout(
                    %calculate_method_timeout_opts);
                no warnings;
                redo AGAIN;
            } ## end else [ if ($was_method_timeout_passed)]
        };
    } ## end AGAIN:
    $Log->exit() if $may_exit;
} ## end sub wait_for_creation

=head2 modify

Same as L<NACL::C::Aggregate::modify|lib-NACL-C-Aggregate-pm/modify> with the
addition of support for verification of the modification (through the new
C<nacltask_verify> option). This verification is performed by checking whether
all the fields have been set to their new values.
There are some cases where the aggregate will be busy in some operation and will
not let the modification happen until its busy. This scenario is handled by
retrying a maximum of 5 times at an interval of 30 seconds. If the operation
doesn't succeed even after 5 retries, then L<NACL::C::Exceptions::StorageAggregate::CurrentlyBusy|lib-NACL-C-Exceptions-StorageAggregate-CurrentlyBusy-pm>
L<NACL::C::Exceptions::StorageAggregate::AggregateHasMountedVolumes|lib-NACL-C-Exceptions-StorageAggregate-AggregateHasMountedVolumes-pm>
will be thrown.

=over

=item Options

=over

=item *

All of the options accepted by the Aggregate component's modify. (See
L<NACL::C::Aggregate::modify|lib-NACL-C-Aggregate-pm/modify>)

=back

=over

=item C<< nacltask_verify => 0|1 >>

(Optional, defaults to being 0)
This new option (not supported by the component call) can be used to specify
whether we want to verify the operation or not. It defaults to 0, meaning that
the default is to not verify.

=item C<< nacltask_to_cleanup => 0|1 >>

(Optional, defaults to 0(no to cleanup))
Flag indicating if this operation is to be cleaned up or not.

=item C<< nacltask_cleanup_manager >>

Cleanup manager to use for registering.
Default : Will use the default cleanup manager.

=back

=back

=over

=item Exceptions

=over

=item C<NACL::Exceptions::VerifyFailure>

This type of exception is thrown when verification for aggregate modify has failed.

=item C<NACL::C::Exceptions::StorageAggregate::AlreadyModified>

This type of exception is thrown when an attempt is made to modify an aggregate which is already modified.

=item C<NACL::C::Exceptions::StorageAggregate::AggregateHasMountedVolumes>

This type of exception is thrown when an attempt is made to modify an aggregate to an offline state.

=back

=back

=cut

sub modify {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    if ($Log->may_debug()) {
        my $me = (caller(0))[3];
        $Log->debug("Options to '$me' are:\n" . Dumper({@args}));
    }
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            nacltask_if_modified =>
                $pkg_or_obj->_if_action_completed_already_validate_spec(
                    name => 'nacltask_if_modified'
                ),
            nacltask_verify => { type => SCALAR, default => 0 },
            $pkg_or_obj->_cleanup_validate_spec(),
            nacltask_retry_tries_count => { type => SCALAR, default => 10 },
            nacltask_retry_sleep_time  => { type => SCALAR, default => 5 },
            nacltask_retry_num_times_code_ran =>
                { type => SCALARREF, optional => 1 },
        },
        allow_extra => 1
    );

    my %opts_for_cleanup;
    $pkg_or_obj->_move_common_cleanup_opts(
        source => \%opts,
        target => \%opts_for_cleanup,
    );

    my %nacltask_options;
    $pkg_or_obj->_move_nacltask_options(
        source => \%opts,
        target => \%nacltask_options
    );

    my ($opts_for_register, $nacltask_to_cleanup);
    if (   $opts_for_cleanup{'nacltask_to_cleanup'}
        || $opts_for_cleanup{'nacltask_cleanup_manager'})
    {
        $opts_for_register
            = $pkg_or_obj->_common_modify_for_cleanup(%opts_for_cleanup,
            %opts);
        $nacltask_to_cleanup = 1;
    }

    my $if_modified = $nacltask_options{nacltask_if_modified};


    my %retry_opts;
    $pkg_or_obj->_copy_retry_params(
        source => \%nacltask_options,
        target => \%retry_opts
    );

    try {
        nacl_method_retry(
            code       => sub { $pkg_or_obj->SUPER::modify(%opts); },
            exceptions => [
                qw(NACL::C::Exceptions::StorageAggregate::CurrentlyBusy NACL::C::Exceptions::StorageAggregate::AggregateHasMountedVolumes)
            ],
            %retry_opts
        );
    }
    catch NACL::C::Exceptions::StorageAggregate::AlreadyModified with {
        my $exception = $_[0];
        if ($if_modified =~ /pass/i) {
            $Log->debug("Exception suppressed because 'if_modified' "
                    . "was sent as '$if_modified'");
        } else {
            $Log->exit() if $may_exit;
            $exception->throw();
        }
    };

    if ($nacltask_options{nacltask_verify}) {
        $pkg_or_obj->taskverify_modify(%opts);
    }

    $pkg_or_obj->_register_for_cleanup(%{$opts_for_register}) if ($nacltask_to_cleanup);
    $Log->exit() if $may_exit;
} ## end sub modify

=head2 mirror

  NACL::STask::Aggregate->mirror(
      command_interface     => $command_interface,
      aggregate             => $aggregate,
      nacltask_if_mirrored  => $action,      # default 'die'
      nacltask_wait         => $boolean,     # default '1'
      %other_options
  );

  $aggregate->mirror(
      nacltask_if_mirrored  => $action,      # default 'die'
      nacltask_wait         => $boolean,     # default '1'
      %other_options
  );

(Class and Instance method) To mirror an existing aggregate. This method
provides additional services beyond what's inherent to the product's
aggregate mirror command, as controlled and described by the new
"nacltask_if_mirrored" and "nacltask_wait" options.

Cleanup for this action cannot be done, because disks are added to the
aggregate for mirroring. Disks once added cannot be removed from the
aggregate.

Supports CMode CLI and ZAPI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for Class method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< aggregate=>$aggregate_name >>

(Required for Class method)
Name of the aggregate to be mirrored.

=item C<< nacltask_wait => 0|1 >>

(Optional, defaults to being 1)
This new option (not supported by the component call) can be used to wait
till the aggregate becomes completely mirrored and the raidstatus says
"mirrored, normal". If 0, do not wait any longer than necessary (any longer
than the aggregate mirror command waits).

=item C<< nacltask_if_mirrored=>$if_mirrored_value >>
(Optional)
 If "die" (default), exception is thrown if the aggregate is already mirrored.
 If "pass", exits quietly if the specified aggregate is already mirrored.

=item C<< "method-timeout" => 7200 >>

(Optional) As component method-timeout, controls how long the command
will wait before completing.  However, the default timeout has been
raised to 7200 seconds.

=item C<< polling_interval => $interval >>

(Optional)
If nacltask_wait is set to 1, this is the interval at which
to poll the filer for the value of the "raidstatus" field.
Default is 10 seconds.

=item C<< blob => $blob_value >>

(Optional, defaults to 0)
This parameter contains the value of the blob which is one of the attribute in
mailbox entry. If the value is passed, it is verified with the dblade table
entry.

=item C<< nacltask_verify_mvr => $boolean >>

(Optional, defaults to 0)
This option is to check whether the MVR(mirror vote records) check needs to be done
for the mirrored aggregate.

=back

=over

=item *

All of the options accepted by the Aggregate component's mirror. (See
L<NACL::C::StorageAggregate::mirror|lib-NACL-C-StorageAggregate-pm/mirror>)

=back

=back

=cut

sub mirror {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = $pkg_or_obj->_common_validate_with(
        params          => \@_,
        additional_spec => {
            nacltask_if_mirrored =>
                $pkg_or_obj->_if_action_completed_already_validate_spec(
                name => 'nacltask_if_mirrored'
                ),
            nacltask_wait    => { type => SCALAR, default => 1 },
            polling_interval => { type => SCALAR, default => 10 },
            blob             => { type => SCALAR, default => 0 },
            nacltask_verify_mvr => { type => BOOLEAN, default => 0 },
        },
        allow_extra => 1,
    );

    my %common_opts;
    $pkg_or_obj->_copy_common_component_params_with_ci(
        source => \%opts,
        target => \%common_opts
    );
    my $if_mirrored      = delete $opts{nacltask_if_mirrored};
    my $wait             = delete $opts{nacltask_wait};
    my $polling_interval = delete $opts{polling_interval};
    my $blob             = delete $opts{blob};
    my $nacltask_verify_mvr = delete $opts{nacltask_verify_mvr};
    $opts{'method-timeout'} ||= NACL_AGGREGATE_MIRROR_TIMEOUT;

    my $pkg = ref $pkg_or_obj || $pkg_or_obj;

MIRROR: {
        use warnings qw(exiting);
        try {
            $pkg->SUPER::mirror(%opts);
        }
        catch NACL::C::Exceptions::StorageAggregate::FreeSpaceReallocationEnabled with {
            $Log->warn("free-space-realloc is enabled, disable it and then ".
                        "try mirror again");
            $pkg->modify(
                aggregate            => $opts{aggregate},
                "free-space-realloc" => "off",
                %common_opts
            );
            no warnings qw(exiting);
            redo MIRROR;
        }
        catch NACL::APISet::Exceptions::ResponseException with {
            my $ex = shift;
            $Log->debug( "Error is: " . $ex->text() );
            unless ( ( $ex->text() =~ /already mirrored/i )
                && ( $if_mirrored eq 'pass' ) )
            {
                # Unknown error or user wants us to throw exception if it is
                # already mirrored.
                $ex->throw();
            } ## end unless ( ( $ex->text() =~ ...))
        };
    }

    if ($wait) {
        $Log->debug("Wait for mirroring to complete");
        my $wait_options = $pkg->get_primary_keys_options(%opts);
        $wait_options->{polling_interval} = $polling_interval;
        $wait_options->{blob} = $blob if($blob);
        $wait_options->{nacltask_verify_mvr} = $nacltask_verify_mvr
            if($nacltask_verify_mvr);
        $Log->debug(
            sub {
                "Options to wait_for_mirror:\n" . Dumper($wait_options);
            }
        );
        $pkg->wait_for_mirror( %$wait_options, %common_opts );
    } ## end if ($wait)

    $Log->exit() if $may_exit;
} ## end sub mirror

=head2 wait_for_mirror

    NACL::STask::Aggregate->wait_for_mirror(
      command_interface => $command_interface,
      aggregate         => $aggregate,
      polling_interval  => 20,
    );

    or

    $aggr_task_obj->wait_for_mirror( polling_interval  => 20 );

(Class or instance method) This method waits till the aggregate is completely
mirrored by polling for the 'raidstatus'. When it becomes 'mirrored, normal',
it understands that the mirroring operation is completed.

Supports CMode CLI and ZAPI.

=over

=item Options

=over

=item C<< "command_interface=>$command_interface" >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< aggregate=>$aggr_name >>

(Required for class method, Not Applicable for instance method)
Name of aggregate.

=item C<< polling_interval => $interval >>

(Optional)
This is the interval at which the raidstatus of the aggregate is polled till
it reaches the value 'mirrored, normal' As the "aggregate mirror" command in
CMode doesn't makes an entry in the job table, this is the way we can
wait for mirroring operation to complete.
default value is 10 seconds.

=item C<< "method-timeout=>$timeout" >>

(Optional) How long in seconds to wait for the mirroing operation
to complete. Defaults to 7200 (2 hours).

=item C<< blob => $blob_value >>

(Optional, defaults to 0)
This parameter contains the value of the blob which is one of the attribute in
mailbox entry. If the value is passed, it is verified with the dblade table
entry. This is applicable when the mirror vote records are being verfied in
the dblade & smdb tables.

=item C<< nacltask_verify_mvr => $boolean >>

(Optional, defaults to 0)
This option is to check whether the MVR(mirror vote records) check needs to be done
for the mirrored aggregate.

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=back

=back

=over

=item Exceptions

=over

=item C<NACL::APISet::Exceptions::TimeoutException>

This type of exception is thrown when mirroring has not happened completly even after waiting for $timeout seconds.

=back

=back

=cut

sub wait_for_mirror {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = $pkg_or_obj->_common_validate_with(
        params => \@_,
        additional_spec =>
            { polling_interval => { type => SCALAR, default => 10 },
              blob             => { type => SCALAR, default => 0 },
              nacltask_verify_mvr => { type => BOOLEAN, default => 0 }, },
    );

    my $blob = delete $opts{blob};
    my $nacltask_verify_mvr = delete $opts{nacltask_verify_mvr};
    $opts{'method-timeout'} ||= AGGREGATE_TIMEOUT;
    my $pkg = ref $pkg_or_obj || $pkg_or_obj;
    my $end_time = time() + $opts{'method-timeout'};
    while (1) {

       # TODO - we need to go away from raid_config and use ngsh commands only
        my $state = $pkg->_get_raid_config_info_listaggr(%opts);
        if ( $state =~ /mirrored/i ) {
            $Log->debug("Aggregate has completed mirroring");
            last;
        }
        if ( time() > $end_time ) {
            my $error_text
                = "After waiting for $opts{'method-timeout'} seconds "
                . "the value for the attribute 'mirror_state_name' "
                . "was still '$state', which is not the "
                . "final expected value\n";
            $Log->exit() if $may_exit;
            NACL::APISet::Exceptions::TimeoutException->throw($error_text);
        } ## end if ( time() > $end_time)
        Tharn::snooze( $opts{polling_interval} );
    } ## end while (1)

    if($nacltask_verify_mvr) {
        my %pass_opts;
        $pass_opts{blob} = $blob if($blob);
        $pkg->verify_mirror_votes(
            command_interface => $opts{command_interface},
            aggregate         => $opts{aggregate},
            _method           => "wait_for_mirror",
            %pass_opts,
        );
    }

    $Log->exit() if $may_exit;
} ## end sub wait_for_mirror

# TODO - we need to go away from raid_config and use ngsh commands only

sub _get_raid_config_info_listaggr {
    $Log->enter() if $may_enter;
    my $pkg               = shift;
    my %opts              = @_;
    my $command_interface = $opts{command_interface};
    my $aggregate         = $opts{aggregate};
    my %common_opts_with_ci;
    $pkg->_copy_common_component_params_with_ci(
        source => \%opts,
        target => \%common_opts_with_ci
    );
    my $aggr_on_node = $pkg->get_local_node_name( %common_opts_with_ci,
        'aggregate' => $aggregate );

    my $apiset = $command_interface->get_7m_or_nodescope_apiset();
    my $resp   = $apiset->raid_config(
        info                  => "listaggr $aggregate",
        'nodescope-node-name' => $aggr_on_node
    );
    my $state = $resp->get_parsed_output()->[0]->{mirror_state_name};
    $Log->debug(
        "Obtained mirror_state_name of aggregate $aggregate as $state");
    $Log->exit() if $may_exit;
    return $state;
}

=head2 verify

  NACL::STask::Aggregate->verify(
      command_interface     => $command_interface,
      aggregate             => $aggregate,
      action                => $action_string, #start or resume
      nacltask_wait         => $boolean,
      %other_options
  );

  $aggregate->verify(
      action            => $action_string, # stop or suspend
      nacltask_verify   => $boolean,
      %other_options
  );

(Class and Instance method) This method starts/stops/suspends/resumes
the RAID mirror verification on the specified aggregate and keeps
polling for the status of the verification operation until the specified
'action' is completed. For action being 'start' and 'resume', this method
provides additional services to wait till the verification
completes. For action being 'stop' and 'suspend', this method provides
additional services to verify whether the specified action is performed
successfully or not.

Note that to get the status of verify operation, the component method
L<NACL::C::StorageAggregate::verify_status|lib-NACL-C-StorageAggregate-pm/verify_status>
can be used.

If plex name is specified, then verify command will be invoked with specifying
plex name which will result in fixing the specified plex if there is a
mismatch.

Supports CMode CLI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for Class method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< aggregate=>$aggregate_name >>

(Required for Class method)
Name of the aggregate to be verified.

=item C<< 'action' => $string >>
The possible values are 'start', stop', 'suspend', 'resume'. To specify
the action as 'status', use
See L<verify_status|lib-NACL-C-StorageAggregate-pm/verify_status>

=item C<< "method-timeout" => 1800 >>

(Optional) As component method-timeout, controls how long to wait
for verification to complete.  However, the default timeout has been
raised to 1800 seconds.

=item C<< nacltask_wait => 0|1 >>

(Optional, defaults to being 1)
This new option (not supported by the component call) can be used to wait
till the verify operation completes. If 0, this method will just invoke
the specified 'action'('start' or 'resume') and will not wait any longer.
Note: When specifying action as 'stop' or 'resume', the verification will not
running and so specifying nacltask_wait will make no sense.

=item C<< polling_interval => $interval >>

(Optional)
This is the interval at which to poll the status of the verification operation
by running 'verify' command with the action as 'status'. It defaults to 10 sec.
This option is effective only when nacltask_wait option is set to 1.

=item C<< nacltask_verify => 0|1 >>

(Optional, defaults to being 0)
This new option (not supported by the component call) can be used to specify
whether we want to verify the specified 'action' is invoked successfully or
not. It defaults to 0, meaning that the default is to not verify.

=back

=over

=item *

All of the options accepted by the Aggregate component's mirror. (See
L<NACL::C::StorageAggregate::verify|lib-NACL-C-StorageAggregate-pm/verify>)

=back

=back

=over

=item Exceptions

=over

=item C<NACL::Exceptions::VerifyFailure>

This type of exception is thrown when RAID mirror verification on a particular aggregate fails.

=item C<NATE::BaseException>

This type of exception is thrown when verify operation is not in expected state (starts/stops/suspends/resume).

=back

=back

=cut

sub verify {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = $pkg_or_obj->_common_validate_with(
        params          => \@_,
        additional_spec => {
            nacltask_wait    => { type => SCALAR, default => 1 },
            nacltask_verify  => { type => SCALAR, default => 0 },
            polling_interval => { type => SCALAR, default => 10 }
        },
        allow_extra => 1,
    );
    my %common_opts;
    $pkg_or_obj->_copy_common_component_params_with_ci(
        source => \%opts,
        target => \%common_opts
    );
    my $wait             = delete $opts{nacltask_wait};
    my $verify           = delete $opts{nacltask_verify};
    my $polling_interval = delete $opts{polling_interval};
    my $action           = $opts{action};

    my $pkg = ref $pkg_or_obj || $pkg_or_obj;

    # Invoke component verify method
    $pkg->SUPER::verify(%opts);

    my $pri_keys = $pkg->get_primary_keys_options(%opts);

    if ($verify) {
        $Log->debug("verify that '$action' is performed successfully");
        my $verify_status
            = $pkg->SUPER::verify_status( %$pri_keys, %common_opts );
        my $is_suspended;
        $is_suspended = $verify_status->is_suspended()
            if ( defined $verify_status );
        if ( $action =~ /^(start|resume)$/i ) {
            if ( $is_suspended !~ /false/i ) {

                # Verification failed
                $Log->exit() if $may_exit;
                NACL::Exceptions::VerifyFailure->throw(
                    "Verify operation is in suspended state",
                    unexpected_values => { is_suspended => $is_suspended } );
            } ## end if ( $is_suspended !~ ...)
        } elsif ( $action =~ /^suspend$/i ) {
            if ( $is_suspended !~ /true/i ) {

                # Verification failed
                $Log->exit() if $may_exit;
                NACL::Exceptions::VerifyFailure->throw(
                    "Verify operation is not in suspended state",
                    unexpected_values => { is_suspended => $is_suspended }
                );
            } ## end if ( $is_suspended !~ ...)
        } elsif ( $action =~ /^stop$/i ) {
            if ( defined $verify_status ) {

                # When verification is not running, verify_status should
                # return undef
                $Log->exit() if $may_exit;
                $verify_status->command_interface_reset();
                NACL::Exceptions::VerifyFailure->throw(
                    "Verify operation is still running",
                    unexpected_values => {%$verify_status}
                );
            } ## end if ( defined $verify_status)
        } else {
            $Log->exit() if $may_exit;
            NATE::BaseException->throw("Unknown action '$action'");
        }
    } ## end if ($verify)

    if ( ( $action =~ /^(start|resume)$/i ) && ($wait) ) {
        $Log->debug("Wait for 'verify' to complete");
        my $wait_options
            = { %$pri_keys, polling_interval => $polling_interval };
        $Log->debug(
            sub {
                "Options to wait_for_verify:\n" . Dumper($wait_options);
            }
        );
        $pkg->wait_for_verify( %$wait_options, %common_opts );
    } ## end if ( ( $action =~ /^(start|resume)$/i...))

    $Log->exit() if $may_exit;
} ## end sub verify

=head2 wait_for_verify

    NACL::STask::Aggregate->wait_for_verify(
      command_interface => $command_interface,
      aggregate         => $aggregate,
      polling_interval  => 20,
    );

    or

    $aggr_task_obj->wait_for_verify( polling_interval  => 20 );

(Class or instance method) This method waits till the aggregate is completely
mirrored by polling for the 'raidstatus'. When it becomes 'mirrored, normal',
it understands that the mirroring operation is completed.

Supports CMode CLI.

=over

=item Options

=over

=item C<< "command_interface=>$command_interface" >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< aggregate=>$aggr_name >>

(Required for class method, Not Applicable for instance method)
Name of aggregate.

=item C<< polling_interval => $interval >>

(Optional)
This is the interval at which the verify command be invoked with the action as
'status'. This method polls till the verify operation completes
("Verify is not running" for CMode CLI). Default value is 10 seconds.

=item C<< "method-timeout=>$timeout" >>

(Optional) How long in seconds to wait for the mirroing operation
to complete. Defaults to 1800 (30 min).

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=back

=back

=over

=item Exceptions

=over

=item C<NACL::APISet::Exceptions::TimeoutException>

This type of exception is thrown when verify operation is still running even after waiting for 'method-timeout' seconds.

=back

=back

=cut

sub wait_for_verify {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = $pkg_or_obj->_common_validate_with(
        params => \@_,
        additional_spec =>
            { polling_interval => { type => SCALAR, default => 10 } }
    );
    my $method_timeout = delete $opts{'method-timeout'}
        || NACL_AGGREGATE_VERIFY_TIMEOUT;
    my $polling_interval = delete $opts{'polling_interval'};
    my $end_time         = time() + $method_timeout;
    my $pkg              = ref $pkg_or_obj || $pkg_or_obj;

    while (1) {
        my $status = $pkg->SUPER::verify_status(%opts);
        $Log->debug( sub { Dumper($status) } );
        unless ( defined $status ) {

            # verification is not running
            $Log->debug('verify is not running');
            last;
        } ## end unless ( defined $status )
        if ( time() > $end_time ) {
            my $error_text
                = "After waiting for $method_timeout seconds "
                . "the verify operation is still running\n. The "
                . "status is\n"
                . Dumper($status);
            $Log->exit() if $may_exit;
            NACL::APISet::Exceptions::TimeoutException->throw($error_text);
        } ## end if ( time() > $end_time)
        Tharn::snooze($polling_interval);
    } ## end while (1)

    $Log->exit() if $may_exit;
} ## end sub wait_for_verify

=head2 wait_for_resync

    NACL::STask::Aggregate->wait_for_resync(
      command_interface => $command_interface,
      aggregate         => $aggregate,
      polling_interval  => 20,
    );

    or

    $aggr_task_obj->wait_for_resync( polling_interval  => 20 );

(Class or instance method) This method waits till the resync operation
completes on an aggregate by polling for 'resyncing_percent' field in the
output of NACL::CS::StorageAggregateShowResyncStatus to reach '-'.

Supports only CMode CLI as of now.

=over

=item Options

=over

=item C<< "command_interface=>$command_interface" >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< aggregate=>$aggr_name >>

(Required for class method, Not Applicable for instance method)
Name of aggregate.

=item C<< polling_interval => $interval >>

(Optional)
This is the interval at which the
NACL::C::StorageAggregate->show_resync_status() be executed to check
for the status of resync operation. Default value is 10 seconds.

=item C<< "method-timeout=>$timeout" >>

(Optional) How long in seconds to wait for the mirroring operation
to complete. Defaults to 1800 (30 min).

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=back

=back

=over

=item Exceptions

=over

=item C<NACL::APISet::Exceptions::TimeoutException>

This type of exception is thrown when resync operation is still running even after waiting for 'method-timeout' seconds.

=back

=back

=cut

sub wait_for_resync {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = $pkg_or_obj->_common_validate_with(
        params => \@_,
        additional_spec =>
            { polling_interval => { type => SCALAR, default => 10 } }
    );
    my $method_timeout = delete $opts{'method-timeout'}
        || NACL_AGGREGATE_VERIFY_TIMEOUT;
    my $polling_interval = delete $opts{'polling_interval'};
    my $end_time         = time() + $method_timeout;
    my $pkg              = ref $pkg_or_obj || $pkg_or_obj;

    my %common_opts = ();
    $pkg_or_obj->_copy_common_component_params_with_ci(
        source => \%opts,
        target => \%common_opts
    );

    my $pri_keys = $pkg->get_primary_keys_options(%opts);

    while (1) {
        my @state_objs = NACL::CS::StorageAggregateShowResyncStatus->fetch(
            %common_opts,
            filter           => {%$pri_keys},
            requested_fields => ['is-resyncing']
        );
        $Log->debug( sub { Dumper( \@state_objs ) } );
        my ($is_resyncing);

        my $is_last = 1;
        foreach my $state_obj (@state_objs) {
            $is_resyncing = $state_obj->is_resyncing();
            if ( $is_resyncing !~ /false/) {
                $Log->debug('resync is still going on');
                $is_last = 0;
                last;
            }
        } ## end foreach my $state_obj (@state_objs)
        last if ($is_last);
        if ( time() > $end_time ) {
            my $error_text
                = "After waiting for $method_timeout seconds "
                . "the resync operation is still running\n. The "
                . "resync status is\n"
                . Dumper( \@state_objs );
            $Log->exit() if $may_exit;
            NACL::APISet::Exceptions::TimeoutException->throw($error_text);
        } ## end if ( time() > $end_time)
    } ## end while (1)

    $Log->exit() if $may_exit;
} ## end sub wait_for_resync

=head2 media_scrub (Class Method only)

    NACL::STask::Aggregate->media_scrub(
       command_interface  => $command_interface,
       disk_name   => $disk_name,
       nacltask_wait  => 0,
       nacltask_wait_interval  => 600
    );

This method performs one of the operations - media_scrub_start or media_scrub_stop
on any of aggregate, plex, raid group, traditional volume or disk.
By default task starts media_scrub on all raid groups and spare drives
in the given node, but if any of the options from aggregate, plex_name,
raid group, trad volume or disk_name is specified then it starts media_scrub
on specified option. Supports only CMode/7Mode CLI as of now.

Cleanup for this action of scrubbing cannot be done.

The options are as given below -

=over

=item Options

=over

=item C<< aggregate=> scalar >>

(Optional) Name of aggregate on which media scrub 'nacltask_action' is to be performed

=item C<< plex_name => scalar >>

(Optional) Name of the plex on which media scrub 'nacltask_action' is to be performed

=item C<< rg_name => scalar >>

(Optional) Name of the raid group on which media scrub 'nacltask_action' is to be performed

=item C<< disk_name => scalar >>

(Optional) Name of the disk on which media scrub 'nacltask_action' is to be performed

=item C<< trad_vol => scalar >>

(Optional) Name of the traditional volume on which media scrub 'nacltask_action' is to
be performed

=item C<< nacltask_action => scalar >>

(Optional) media scrub action to be performed (start|stop)
(Default - start)

=item C<< nacltask_wait => boolean >>

(Optional)If nacltask_wait is 1, wait till media scrub is completed
(Default - 1)By default task will wait for the completion for particular action.

=item C<< command_interface=>$command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< nacltask_wait_interval => 3600 >>

(Optional) It is the time needed to completing the media scrub, the default
timeout has been set to 3600 seconds.

=item C<<nacltask_polling_time => scalar>>

(Optional) Polling time interval

=back

=back

=over

=item Exceptions

=over

=item C<NATE::BaseException>

This type of exception is thrown if raid.media_scrub.enable flag is not set.

=back

=back

=cut

sub media_scrub {
    $Log->enter() if $may_enter;
    my $pkg = shift;
    my ( $opts, $temp_opts ) = $pkg->_validate_media_options(@_);
    my $command_interface = $opts->{command_interface};
    my $nodescope_or_7m_apiset
        = $command_interface->get_7m_or_nodescope_apiset();

    # Set the media_scrub option 'on'
    $nodescope_or_7m_apiset->options( 'raid.media_scrub.enable' => "on" );
    my $text
        = $nodescope_or_7m_apiset->options( 'raid.media_scrub.enable' => "" )
        ->get_parsed_output();

    my $flag_status = $text->[0]->{"raid.media_scrub.enable"};

    # Verify the status of media scrub option
    if ( $flag_status ne "on" ) {
        $Log->exit() if $may_exit;
        NATE::BaseException->throw(
            "The raid.media_scrub.enable flag is not set on the system");
    }
    $Log->debug("Media scrubbing is currently enabled on the node");

    my $action                 = $opts->{"nacltask_action"} || "start";
    my $nacltask_wait          = $opts->{"nacltask_wait"};
    my $nacltask_wait_interval = $opts->{"nacltask_wait_interval"}
        || MEDIA_SCRUB_TIMEOUT;
    my $nacltask_polling_time = $opts->{"nacltask_polling_time"}
        || MEDIA_SCRUB_POLL_TIME;
    my $cmd = "aggr_media_scrub_" . $action;
    my $expected_value;

    if ( $action eq "start" ) {
        $expected_value = "suspended";
    } else {
        $expected_value = "Media scrub disabled";
    }

    my $cmd_output = $nodescope_or_7m_apiset->$cmd(%$temp_opts);

    if ($nacltask_wait) {
        $pkg->wait_for_scrub_completion(
            command_interface      => $command_interface,
            attribute_to_check     => "status",
            expected_value         => $expected_value,
            element                => $temp_opts,
            nacltask_wait_interval => $nacltask_wait_interval,
            nacltask_polling_time  => $nacltask_polling_time,
        );
    } ## end if ($nacltask_wait)

} ## end sub media_scrub

=head2 wait_for_scrub_completion (Class or Instance Method)

=over

=item Examples

$pkg->wait_for_scrub_completion(
    command_interface      => $command_interface,
    attribute_to_check     => "status",
    expected_value         => $expected_value,
    element                => \%parameter,
    nacltask_wait_interval => $time_out
 );

This methods waits for media scrub to complete on
any of aggregate, tradvol, disk, raid group or plex.

=back

=over

=item Options

=over

=item C<< attribute_to_check => scalar >>

(Required) The value of attribute needs to be check

=item C<< expected_value => scalar >>

(Required) The name of the attribute that needs to be checked

=item C<< element => hashref >>

(Required) Reference to a hash containing the type of raid object as key and
the name of the object as value.

=item C<<nacltask_wait_interval => scalar>>

(Optional) Maximum time out limit before throwing an exception in case of failure

=item C<<nacltask_polling_time => scalar>>

(Optional) Polling time interval

=back

=back

=over

=item Exceptions

=over

=item C<NACL::APISet::Exceptions::TimeoutException>

This type of exception is thrown when media scrub is not completed in the given time.

=back

=back

=cut

sub wait_for_scrub_completion {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = validate_with(
        params => \@_,
        spec   => {
            command_interface =>
                { type => OBJECT, isa => 'NACL::C::CommandInterface' },
            attribute_to_check     => { type => SCALAR },
            expected_value         => { type => SCALAR },
            element                => { type => HASHREF | UNDEF },
            nacltask_wait_interval => { type => SCALAR, optional => 1 },
            nacltask_polling_time  => { type => SCALAR, optional => 1 },
        },
    );

    my $command_interface  = $opts{command_interface};
    my $element            = $opts{element};
    my $attribute_to_check = $opts{attribute_to_check};
    my $expected_value     = $opts{expected_value};
    my $method_timeout     = $opts{'nacltask_wait_interval'}
        || MEDIA_SCRUB_TIMEOUT;
    my $polling_interval = $opts{nacltask_polling_time}
        || MEDIA_SCRUB_POLL_TIME;
    my $end_time = time() + $method_timeout;
    my $nodescope_or_7m_apiset
        = $command_interface->get_7m_or_nodescope_apiset();

    my $key   = ( keys %{$element} )[0];
    my $value = $element->{$key};

    while (1) {
        my $scrub_output
            = $nodescope_or_7m_apiset->aggr_media_scrub_status(%$element)
            ->get_parsed_output();
        my $scrub_status = $scrub_output->[0]->{status};

        last
            if ( $scrub_status eq $expected_value );
        $Log->trace(
            "Current value of media scrub $attribute_to_check on $key is $scrub_status"
        );
        if ( time() > $end_time ) {
            $Log->exit() if $may_exit;
            NACL::APISet::Exceptions::TimeoutException->throw(
                      " Media scrub has not been completed in the given time "
                    . "and the current status is $scrub_status\n" );
        } ## end if ( time() > $end_time)
        Tharn::snooze($polling_interval);
    } ## end while (1)

    $Log->exit() if $may_exit;
} ## end sub wait_for_scrub_completion

=head2 media_scrub_status (Class Method only)

    NACL::STask::Aggregate->media_scrub_status(
       command_interface  => $command_interface,
       disk_name   => $disk_name,
    );

This method gets the media scrub status for any one of - aggregate,plex,raid group,
traditional volume or disk and returns the status of the element.
Supports only CMode CLI as of now.
The options are as given below -

=over

=item Options

=over

=item C<< aggregate=> scalar >>

(Optional) Name of aggregate on which media scrub 'nacltask_action' is to be performed

=item C<< plex_name => scalar >>

(Optional) Name of the plex on which media scrub 'nacltask_action' is to be performed

=item C<< rg_name => scalar >>

(Optional) Name of the raid group on which media scrub 'nacltask_action' is to be performed

=item C<< disk_name => scalar >>

(Optional) Name of the disk on which media scrub 'nacltask_action' is to be performed

=item C<< trad_vol => scalar >>

(Optional) Name of the traditional volume on which media scrub 'nacltask_action' is to
be performed

=item C<< command_interface=>$command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=back

=back

=cut

sub media_scrub_status {
    $Log->enter() if $may_enter;
    my $pkg = shift;
    my ( $opts, $temp_opts ) = $pkg->_validate_media_options(@_);
    my $command_interface = $opts->{command_interface};
    my $nodescope_or_7m_apiset
        = $command_interface->get_7m_or_nodescope_apiset();
    my $scrub_output
        = $nodescope_or_7m_apiset->aggr_media_scrub_status(%$temp_opts)
        ->get_parsed_output();
    my $scrub_status = $scrub_output->[0]->{status};
    $Log->exit() if $may_exit;

    # Return the status of element
    return $scrub_status;
} ## end sub media_scrub_status

sub _validate_media_options {
    $Log->enter() if $may_enter;
    my $pkg = shift;
    $pkg->_verify_invocation( style => 'static_only' );
    my %opts = $pkg->_common_validate_with(
        params          => \@_,
        additional_spec => {
            _optional_scalars(
                qw(aggregate plex_name rg_name disk_name trad_vol nacltask_wait_interval)
            ),
            nacltask_action => { type => SCALAR, default => 'start' },
            nacltask_wait   => { type => SCALAR, default => 1 }
        },
        ignore_primary_keys => 1
    );

    my %temp_opts;
    $pkg->_hash_copy(
        source => \%opts,
        target => \%temp_opts,
        map    => {
            "plex_name" => "plex-name",
            "rg_name"   => "rg-name",
            "disk_name" => "disk-name",
            "aggregate" => "atv-name",
            "trad_vol"  => "atv-name",
        }
    );

    if ( scalar( keys %temp_opts ) > 1 ) {
        $Log->exit() if $may_exit;
        NATE::BaseException->throw( "Only one of the parameter among "
                . "(aggregate, plex_name, rg_name, disk_name, trad_vol) can "
                . "be passed " );
    } ## end if ( scalar( keys %temp_opts...))
    $Log->exit() if $may_exit;
    return ( \%opts, \%temp_opts );
} ## end sub _validate_media_options

=head2 modify

    NACL::STask::Aggregate->modify(
        command_interface    => $ci,
        aggregate            => $aggr_name,
        nacltask_if_modified => $if_modified  # Default is "die"
        %other_opts
    );

                        or

    $aggr_stask_obj->modify(%other_opts, nacltask_if_modified => $if_modified);

(Class or instance method)
Modifies properties of the aggregate. This method provides additional
services beyond what's inherent to the product's aggregate modification
commands, as controlled and described by the new C<nacltask_if_modified> and
C<nacltask_verify> options.

=over

=item Options

=over

=item C<< nacltask_if_modified => die|pass >>

(Optional, defaults to "die")
This option controls what to do if the aggregate is already in the state
we are trying to modify it to. This can be one of two values:

=over

=item C<die>

If this is the value of nacltask_if_modified, then if the aggregate is
already in the state we want to modify it to be, then an exception is thrown.

=item C<pass>

If this is the value of nacltask_if_modified, then the call succeeds even if
the aggregate is already in the state we want to modify it to be.

=back

For example, if the aggregate represented by $aggr_obj is of raidtype raid_dp,
this call will B<not> result in an exception:
$aggr_obj->modify(raidtype => 'raid_dp', nacltask_if_modified => 'pass');

=item C<< nacltask_verify => 0|1 >>

(Optional, defaults to 0)
Specifies whether to verify that all the fields got modified to the values
specified in the call. If one or more of the fields don't match, then a
L<NACL::Exceptions::VerifyFailure|lib-NACL-Exceptions-VerifyFailure-pm>
exception is thrown.

=item C<< nacltask_to_cleanup => 0|1 >>

(Optional, defaults to 0(no to cleanup))
Flag indicating if this operation is to be cleaned up or not.

=item C<< nacltask_cleanup_manager >>

Cleanup manager to use for registering.
Default : Will use the default cleanup manager.

=item Other options

All of the other options accepted by
L<NACL::C::StorageAggregate::modify|lib-NACL-C-StorageAggregate-pm/modify>.

=back

=back

=cut

# Definition is from _AggregateVolume.

=head2 wait_for_raidio (Class|Instance Method)

    NACL::STask::Aggregate->wait_for_raidio(
       command_interface  => $command_interface,
       aggregate   => $aggr_name,
       raid_operation  => $operation,
       'method-timeout' => 600,
       nacltask_polling_time => 20
    );
    or
    $aggr_obj->wait_for_raidio(raid_operation  => $operation);

This method waits for particular RAID IO operation like
(Recons/Reparity/Scrub/Verify) to complete on the specified
aggregate. Supports for CMode/7Mode CLI as of now.
The options are as given below -

=over

=item Options

=over

=item C<< aggregate=> scalar >>

(Required) Name of aggregate on which operation is performed
Optional for instance method

=item C<< raid_operation => scalar >>

(Required) Name of the raid IO operation performed on aggregate
Values can be any of among recons,reparity,scrub or verify.

=item C<< method-timeout => scalar >>

(Optional) It is the maximum time that needed to complete the operation,
the default timeout has been set to 3600 seconds.

=item C<<nacltask_polling_time => scalar>>

(Optional) Polling time interval,
the default polling time is 20 seconds

=item C<< command_interface=>$command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=back

=back

=over

=item Exceptions

=over

=item C<NATE::BaseException>

This type of exception is thrown when raid operation is not completed with in maximum wait time interval.

=back

=back

=cut

sub wait_for_raidio {

    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = $pkg->_common_validate_with(
        params          => \@_,
        additional_spec => {
            raid_operation        => { type => SCALAR },
            'method-timeout'      => { type => SCALAR, default => 3600 },
            nacltask_polling_time => { type => SCALAR, default => 20 }
        },
    );
    my $command_interface  = $opts{command_interface};
    my $raid_io            = lc $opts{raid_operation};
    my $aggregate          = $opts{aggregate};
    my $method_timeout     = $opts{'method-timeout'};
    my $polling_time       = $opts{nacltask_polling_time};
    my $end_time           = time() + $method_timeout;
    my $attribute_to_check = $raid_io . "_progress";
    my $nodescope_or_7m_apiset
        = $command_interface->get_7m_or_nodescope_apiset();
    my $raid_io_status;

    if ( $raid_io =~ /recons|reparity|scrub/ig ) {
        my @aggr_group_list = NACL::STask::RaidGroup->list(
            command_interface => $command_interface,
            tree              => $aggregate,
        );
        $Log->debug( "aggr_group_list" . Dumper( \@aggr_group_list ) );
        my %rg_hash = %{ $aggr_group_list[0] };
        while (scalar(keys(%rg_hash))) {
            foreach my $raid_group_key (  keys %rg_hash ) {
                $Log->debug( "raid_group_key" . Dumper($raid_group_key) );

                my $rg_params = NACL::STask::RaidGroup->list(
                    command_interface => $command_interface,
                    raid_name         => [$raid_group_key],
                );

                $raid_io_status
                    = $rg_params->{$raid_group_key}->{$attribute_to_check};
                $Log->debug( "rg_params" . Dumper($rg_params) );
                if ( $raid_io_status == -1 ) {
                    delete $rg_hash{$raid_group_key};
                } else {
                    if ( time() > $end_time ) {
                        $Log->exit() if $may_exit;
                        NATE::BaseException->throw( "Raid operation $raid_io "
                                . "is not completed in maximum wait time interval"
                        );
                    } ## end if ( time() > $end_time)
                    Tharn::snooze($polling_time);
                } ## end else [ if ( $raid_io_status ==...)]
            } ## end while ( my ( $raid_group_key...))
            $Log->debug("Raid Operation $raid_io is completed");
        } ## end while (1)

    } elsif ( $raid_io =~ /verify/ig ) {
        while (1) {
            my $rg_output = $nodescope_or_7m_apiset->raid_config(
                info => "listaggr $aggregate" )->get_parsed_output();
            $Log->debug( "rg_output" . Dumper($rg_output) );
            $raid_io_status = $rg_output->[0]->{verifypct};

            # When $raid_io_status = -1 , the operation will complete
            last if ( $raid_io_status == -1 );
            if ( time() > $end_time ) {
                $Log->exit() if $may_exit;
                NATE::BaseException->throw( "Raid operation $raid_io "
                        . "is not completed in maximum wait time interval" );
            }
            Tharn::snooze($polling_time);
        } ## end while (1)
    } ## end elsif ( $raid_io =~ /verify/ig)
} ## end sub wait_for_raidio

=head2 grow_size

    my $aggregate = NACL::STask::Aggregate->create(
    command_interface => $command_interface,
    aggregate         => $aggregate,
    if_exists         => $action,              # default 'die'
    nacltask_wait     => $boolean,             # default '1'
    %other_options
    );

    $aggregate->grow_size( nacltask_min_size => '200g' );

    or

    NACL::STask::Aggregate->grow_size(
    command_interface => $command_interface,
    aggregate         => $aggregate,
    nacltask_min_size => '200g',
    %other_options
    );

(Class or instance method) This method will grow the size of and
existing aggregate till the size becomes specified nacltask_min_size. It
adds up one disk at action time and checks the size and this cycles till
the aggregate gows at least nacltask_min_size. If the aggregate is already
meeting the size criteria, it will just return without adding any disks.
If there are not sufficient spare disks available to grow the aggregate,
CommandFailedException will be thrown and the aggregate will be purged
(Only if this method is called from NACL::STask::Aggregate->create)
before returning the control.

Cleanup for this action cannot be done, since disks are added to the
aggregate to grow its size. Once disks are added they cannot be removed
from the aggregate.

=over

=item Options

=over

=item C<< aggregate=>$aggr_name >>

(Required for class method, Not Applicable for instance method)
Name of aggregate.

=item C<< nacltask_min_size => $min_size >>

(Mandatory, SCALAR)
This param defines the minimum size to which the aggregate will at least be grown.

=item C<< polling_interval => $interval >>

(Optional, SCALAR)
This is the interval at which the aggregate will be polled to ensure the disk
is been added. Default value is 10 seconds.

=item command_interface, apiset_must, apiset_should, aggregate etc.

All of the other various options to
L<< NACL::C::Aggregate->add_disks|lib-NACL-C-Aggregate-pm/add_disks >> except diskcount.

=back

=back

=over

=item Exceptions

=over

=item C<NACL::APISet::Exceptions::CommandFailedException>

This type of exception is thrown when there is failure to create the aggregate with minimum size.

=item C<NACL::C::Exceptions::StorageAggregate::InsufficientDisksToStartNewRaidGroup>

This type of exception is thrown when there is attempt made to create new raid group with insufficient disks.

=item C<NACL::C::Exceptions::StorageAggregate::NoMatchingDisksAvailable>

This type of exception is thrown when an attempt is made to create the aggregate with no matching available disks.

=back

=back

=cut

sub grow_size {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my $pkg        = ref($pkg_or_obj) || $pkg_or_obj;
    my %opts       = $pkg_or_obj->_common_validate_with(
        params          => \@_,
        additional_spec => {
            polling_interval  => { type  => SCALAR, optional => 1 },
            nacltask_min_size => { regex => $VALID_SIZE_RE },
        },
        allow_extra => 1
    );
    my $caller            = ( caller(1) )[3];
    my $nacltask_min_size = delete $opts{nacltask_min_size};

    # handle if the "disklist" is given, delete it from %opts as it is invalid
    # for this method.
    delete $opts{disklist};
    my $min_size
        = NACL::C::UnitNormalization::convert_to_bytes($nacltask_min_size);

    my %common_component_opts;
    $pkg->_copy_common_component_params_with_ci(
        source => \%opts,
        target => \%common_component_opts,
    );

    my $aggr_size = sub {
        return NACL::CS::Aggregate->fetch(
            %common_component_opts,
            filter           => { aggregate => $opts{aggregate} },
            requested_fields => [qw(size)]
        )->size();
    };

    my $size = &$aggr_size;

    # If the aggregate is already of the specified min size, just return
    if ( $size >= $min_size ) {
        $Log->exit() if $may_exit;
        return;
    }

    # Calculate the total size of all the available spare disks considerig
    # the disktype if specified, and check the possibility of growing the
    # aggregate to the minimum si requested.
    my %spare_disk_options = %common_component_opts;
    $spare_disk_options{disktype} = $opts{'disktype'}
        if defined $opts{'disktype'};
    my $possible_grow_size
        = $pkg->_get_size_of_available_spare_disks(%spare_disk_options);

    # take of the current size of an aggregate from the specified min size
    # for comparison

    my $min_size_to_compare = $min_size - &$aggr_size();
    if ( $possible_grow_size < $min_size_to_compare ) {

        # destroy the aggregate if the caller is "create()"
        if ( $caller eq "NACL::STask::Aggregate::create" ) {
            $pkg->purge( %common_component_opts,
                aggregate => $opts{aggregate} );
        }
        $Log->exit() if $may_exit;
        NACL::APISet::Exceptions::CommandFailedException->throw( text =>
                "Failed to create the aggregate with minimum size. Not enough spare disks available!"
        );
    } ## end if ( $possible_grow_size...)

    # ignore the diskcount as it is invalid for growing the aggregate
    $Log->debug("Parameter \"diskcount\" is ignored for this method")
        if delete $opts{diskcount};

    # Handle the disktype option for mixed_raid_type aggregates
    my $raid_type = NACL::CS::Aggregate->fetch(
        %common_component_opts,
        filter           => {aggregate => $opts{aggregate}},
        requested_fields => [qw(raidtype)]
    )->raidtype();
    my %disk_opts;
    my @disks;
    if ($raid_type eq "mixed_raid_type") {
        my $disk_type;
        @disks = NACL::CS::StorageAggregateShowStatus->fetch(
            %common_component_opts,
            'filter' => {'aggregate' => $opts{aggregate}, 'disk-type' => '!SSD'},
            'requested_fields' => [qw/raidgroup disk disk-type/],
        );
        $disk_type = $disks[-1]->disk_type();

        %disk_opts = ('disktype' => $disk_type);
    }

    while ( $size < $min_size ) {
        my $disk_count = 1;

        # Try adding one disk at a time and when got an error "Failed to add
        # disks to aggregate <aggr_name>. Reason: Starting a new raid4 RAID
        # group; The number of disks specified must be at least twice the
        # number of plexes in the target aggregate '<aggr_name>', catch the
        # exception and add ($diskcount + 1) and continue this till command
        # succeeds
    AGAIN: {
            use warnings 'exiting';
            try {
                $disk_opts{'diskcount'} = $disk_count;
                $pkg->add_disks(
                    %opts,
                    %disk_opts,
                    nacltask_wait => 1,
                );
            } ## end try
            catch
                NACL::C::Exceptions::StorageAggregate::InsufficientDisksToStartNewRaidGroup
                with {
                my $ex = shift;
                my $raidtype = NACL::CS::Aggregate->fetch(
                    %common_component_opts,
                    filter           => {aggregate => $opts{aggregate}},
                    requested_fields => [qw(raidtype)]
                )->raidtype();
               # my $disk_count_to_start = ($raidtype eq "raid4") ? 2 : 3;
                my $disk_count_to_start
                    = NACL::STask::Aggregate->_min_diskcount_to_start_new_rg(
                    'aggregate'     => $opts{aggregate},
                    'raidtype'      => $raidtype,
                    '_disk_cs_objs' => \@disks,
                    %common_component_opts,
                    );

                if ($disk_count > $disk_count_to_start) {
                    my $error
                        = 'Should not have to add more than '
                        . $disk_count_to_start
                        . ' disks when '
                        . 'starting a new RAID group.  Something unexpected '
                        . 'has happened.'
                        . "\n";
                    $ex->set_text($error . $ex->text());
                    $Log->exit() if $may_exit;
                    $ex->throw();
                } else {
                    $disk_count = $disk_count_to_start;
                }
                no warnings 'exiting';
                redo AGAIN;
            }
            catch
                NACL::C::Exceptions::StorageAggregate::NoMatchingDisksAvailable
                with {
                my $ex = shift;
                if ($caller eq "NACL::STask::Aggregate::create") {

                    $pkg->purge(%common_component_opts,
                        aggregate => $opts{aggregate});
                }
                my $error
                    = "Failed to create the aggregate with minimum size: ";
                $error .= $nacltask_min_size . "\n" . $ex->text();
                $ex->set_text($error . $ex->text());
                $Log->exit() if $may_exit;
                $ex->throw();
            }
            otherwise {
                my $ex = shift;
                $Log->exit() if $may_exit;
                $ex->throw();
            };

        } ## end AGAIN:

        $size = &$aggr_size;
    } ## end while ( $size < $min_size)
    $Log->exit() if $may_exit;
    return;
} ## end sub grow_size

# Returns the total size of all the spare disks

sub _get_size_of_available_spare_disks {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = $pkg_or_obj->_common_validate_with(
        params              => \@_,
        ignore_primary_keys => 1,
        allow_extra         => 1,
    );

    my %spare_disk_opts;
    $spare_disk_opts{command_interface} = $opts{command_interface};
    my %filter_opts;
    $filter_opts{state} = 'spare';
    $filter_opts{type} = $opts{disktype} if defined $opts{disktype};

    my $size = 0;
    try {
        my @disks = NACL::CS::StorageDisk->fetch(
            %spare_disk_opts,
            canned_filters => [qw(spare)],
            requested_fields => [qw(usable-size)]
        );

        foreach my $disk_info (@disks) {
            $size += $disk_info->usable_size();
        }
        $size .= 'mb';
    } ## end try
    catch NACL::Exceptions::NoElementsFound with {
        # Partitioned disks found
        my $exception = shift;
        my $text      = $exception->text();
        my $node = $opts{command_interface};

        my @disks = NACL::CS::StorageAggregateShowSpareDisks->fetch(
            %spare_disk_opts,
            filter => {'owner-name' => $node->node},
            requested_fields => [qw(local-usable-data-size)],
            allow_empty => 1,
        );

        foreach my $disk_info (@disks) {
            if ($disk_info->local_usable_data_size() =~ /\d+/) {
                $size += $disk_info->local_usable_data_size();
            }
        }
    };

    $Log->exit() if $may_exit;
    return NACL::C::UnitNormalization::convert_to_bytes($size);
} ## end sub _get_size_of_available_spare_disks

# Helper method for NACL::STask::Aggregate->create(), which returns a
# disktype with the largest number of spare disks.
# If 'not_of_type' is provided, method will exclude that disktype.
# For example, If there are 5 SSDs, 4 FCALs and 3 SATA spare disks,
# method will return "FCAL" though SSD's are more in number when
# 'not_of_type' => "SSD". If "not_of_type" is not specified, method
# will default "not_of_type" to "SSD"s.

sub _disktype_by_max_availability {

    $Log->enter() if $may_enter;

    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params => \@_,
        spec   => {
            command_interface => { type => OBJECT },
            not_of_type       => { type => SCALAR, default => "SSD" },
            node              => { type => SCALAR }
        },
    );

    my $not_of_type = delete $opts{not_of_type};
    my $node        = delete $opts{'node'};
    my ( %no_of_disks_per_type, %filter );

    %filter = ( 'container-type' => 'spare', 'owner' => $node );

    my @disks = NACL::CS::StorageDisk->fetch(
        %opts,
        filter           => \%filter,
        requested_fields => [qw(type)]
    );

    # Following is the regex to exclude the disktype.
    # This can be changed if more of the SSD types  are added in future

    my $not_of_type_regex = qr/$not_of_type/i;

    foreach my $disk (@disks) {
        my $type = $disk->type();
        next if ( defined $type && $type =~ $not_of_type_regex );
        ++$no_of_disks_per_type{$type} if ( defined $type );
    }

    my $max_avil_disk_type;
    my $max_val = 0;
    foreach my $type ( keys %no_of_disks_per_type ) {
        if ( $no_of_disks_per_type{$type} > $max_val ) {
            $max_avil_disk_type = $type;
            $max_val            = $no_of_disks_per_type{$type};
        }
    } ## end foreach my $type ( keys %no_of_disks_per_type)
    $Log->exit() if $may_exit;
    return $max_avil_disk_type;
} ## end sub _disktype_by_max_availability

=head2 shared_create

    $Shared_Aggr = NACL::STask::Aggregate->shared_create(
      command_interface => $command_interface,
      aggregate         => qr/$aggregate/i,
      raidtype          => { 'or' => ['raid4', 'raid_dp', 'raid_tec'] },
      diskcount         => {'>=' => 3},
      disksize          => { '>=' => 50, '<' => 512 },
      raidgroup         => "rg1",
      nacltask_cleanup_manager => $object,
    );

(Class method) Creates a shared aggregate prefixed with 'AG'.
Instead of purge use 'shared_undo' to release the shared aggregate.
The 'shared_undo' method is registered for cleanup automatically,
but it needs to be invoked using the cleanup manager's run method.

Current limitation is, the criteria can be specified for only those attributesi
which will hold (in the state class) the scalar or arrayref as their values.

The criteria has to be either logical or relational and the supported ones are,

=over

=item Logical "or", "not"

Supports only these type of logical criteria.

Example:  raidtype => { 'or' => ['raid4', 'raid_dp', 'raid_tec', ...], 'not' => ['raid3', ...] }

Multiple conditions for the same attribute gets "AND"ed to ensure criteria match.

=item Relational [==, <, <=, >, >= !=]

Example:  size     => {">" => 1415577600}

=item nacltask_cleanup_manager

Cleanup manager to use for registering.
Default : Will use the default cleanup manager.

=back

=cut

=head2 recyclable_create

    $Recycl_Aggr = NACL::STask::Aggregate->recyclable_create(
      command_interface => $command_interface,
      aggregate         => qr/$aggregate/i,
      raidtype          => { 'or' => ['raid4', 'raid_dp', 'raid_tec'] },
      diskcount         => {'>=' => 3},
      disksize          => { '>=' => 50, '<' => 512 },
      raidgroup         => "rg1",
      nacltask_cleanup_manager => $object,
    );

(Class method) Creates a recyclable aggregate prefixed with 'ag'.
Instead of purge use 'recyclable_undo' to release the aggregate.
The 'recyclable_undo' method is registered for cleanup automatically,
but it needs to be invoked using the cleanup manager's run method.

Current limitation is, the criteria can be specified for only those attributes
which will hold (in the state class) the scalar or arrayref as their values.

The criteria has to be either logical or relational and the supported ones are,

=over

=item Logical "or", "not"

Supports only these type of logical criteria.

Example:  raidtype => { 'or' => ['raid4', 'raid_dp', 'raid_tec', ...], 'not' => ['raid3', ...] }

Multiple conditions for the same attribute gets "AND"ed to ensure criteria match.

=item Relational [==, <, <=, >, >= !=]

Example:  size     => {">" => 1415577600}

=item nacltask_cleanup_manager

Cleanup manager to use for registering.
Default : Will use the default cleanup manager.

=back

=cut

=head2 shared_undo

    $Shared_Aggr = NACL::STask::Aggregate->shared_undo(
      command_interface => $command_interface,
      aggregate         => qr/$aggregate/i,
      raidtype          => { 'or' => ['raid4', 'raid_dp', 'raid_tec'] },
      diskcount         => {'>=' => 3},
      disksize          => { '>=' => 50, '<' => 512 },
      raidgroup         => "rg1",
    );

(Class method) Reverts the aggregate to its shared state by reinitializing its
contained element. The freed shared aggregate is prefixed by '_AG'.

Current limitation is, the criteria can be specified for only those
attributes which will hold (in the state class) the scalar or
arrayref as their values.Resource name provided will be matched
with the free shared resources available and if found, will be undone.

The criteria has to be either logical or relational and the supported ones are,

=over

=item Logical "or", "not"

Supports only these type of logical criteria.

Example:  raidtype => { 'or' => ['raid4', 'raid_dp', 'raid_tec', ...], 'not' => ['raid3', ...] }

Multiple conditions for the same attribute gets "AND"ed to ensure criteria match.

=item Relational [==, <, <=, >, >= !=]

Example:  size     => {">" => 1415577600}

=back

=cut

=head2 recyclable_undo

    $Recycl_Aggr = NACL::STask::Aggregate->recyclable_undo(
      command_interface => $command_interface,
      aggregate         => qr/$aggregate/i,
      raidtype          => { 'or' => ['raid4', 'raid_dp', 'raid_tec'] },
      diskcount         => {'>=' => 3},
      disksize          => { '>=' => 50, '<' => 512 },
      raidgroup         => "rg1",
    );

(Class method) Reverts the aggregate to its recyclable state by reinitializing
its contained element. The freed recyclable aggregate is prefixed by '_ag'.

Current limitation is, the criteria can be specified for only those
attributes which will hold (in the state class) the scalar or arrayref
as their values. Resource name provided will be matched with the free
recyclable resources available and if found, will be undone.

The criteria has to be either logical or relational and the supported ones are,

=over

=item Logical "or", "not"

Supports only these type of logical criteria.

Example:  raidtype => { 'or' => ['raid4', 'raid_dp', 'raid_tec', ...], 'not' => ['raid3', ...] }

Multiple conditions for the same attribute gets "AND"ed to ensure criteria match.

=item Relational [==, <, <=, >, >= !=]

Example:  size     => {">" => 1415577600}

=back

=cut

=head2 cleanup

    $Stask_obj->cleanup();

(Instance Only Method)
Performs the registered cleanup operations for the element.

=cut

# Helper method to calculate the 'method-timeout' based on the "zest" of the disks being added/created with.

sub _calculate_method_timeout {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;

    my %opts = validate_with(
        params => \@_,
        spec   => {
            command_interface =>
                { isa => 'NACL::C::CommandInterface::ONTAP', type => OBJECT },
            'aggregate' => { type => SCALAR },
        }
    );

    my @disklist;
    my @disk_objs;
     @disk_objs = NACL::CS::StorageAggregateShowStatus->fetch(
        'command_interface' => $opts{command_interface},
        'filter'            => {'aggregate' => $opts{'aggregate'}},
        'requested_fields'  => [qw(disk)],
    );

    foreach my $disk_obj (@disk_objs) {
            push @disklist, $disk_obj->disk();
    }

    my $calculated_method_timeout
        = NACL::STask::StorageDisk->_get_zeroing_estimate(
        command_interface => $opts{command_interface},
        'disklist'        => \@disklist,
        );

    $Log->exit() if $may_exit;
    return $calculated_method_timeout;
} ## end sub _calculate_method_timeout

=head2 wait_for_aggr_zeroing_completion

    NACL::STask::Aggregate->wait_for_aggr_zeroing_completion(
        command_interface    => $ci,
        aggregate            => $aggr_name,
        polling_interval     => $polling_interval,
        'method-timeout'     => 900,
    );

                        or

    $aggr_stask_obj->wait_for_aggr_zeroing_completion(
                    polling_interval => $polling_interval,
                    'method-timeout' => 900,);

(Class or instance method)
Waits till all the disks belonging to the given aggregate are zeroed.
If the disks are already zeroed, method just returns.

=over

=item C<< polling_interval => scalar >>

(Optional) Polling time interval,
the default polling time is 20 seconds

=item C<< method-timeout => scalar >>

(Optional) Timeout value for wait.
If not specified, timeout will be calculated based on
data provided by the disk vendor.

=back

=over

=item Exceptions

=over

=item C<NACL::APISet::Exceptions::TimeoutException>

This type of exception is thrown when disk zeroing did not complete with in $timeout seconds.

=back

=back

=cut

sub wait_for_aggr_zeroing_completion {
    $Log->enter() if $may_enter;

    my $pkg_or_obj = shift;
    my $pkg        = $pkg_or_obj->get_package_name();
    my %opts       = $pkg_or_obj->_common_validate_with(
        params => \@_,
        additional_spec =>
            { polling_interval => { type => SCALAR, default => 20 }, },
    );
    my $aggregate = $opts{aggregate};
    my $ci        = $opts{command_interface};
    my $timeout = $opts{'method-timeout'} || $pkg->_calculate_method_timeout(
        command_interface => $ci,
        'aggregate'       => $aggregate
    );
    my $polling_interval = $opts{polling_interval};
    my $apiset           = $ci->get_7m_or_nodescope_apiset();
    $timeout += time();

    while ( time() <= $timeout ) {
        my $zeroing       = 0;
        my $parsed_output = $apiset->raid_config(
            'info'               => 'listaggr',
            'name'               => $aggregate,
            'connectrec-timeout' => 90
        )->get_parsed_output();
        foreach my $disk_info ( @{$parsed_output} ) {
            my $fsm_state = $disk_info->{'fsm_state'};
            if (   $fsm_state eq "ZEROING"
                || $fsm_state eq "NOTZEROED"
                || $fsm_state =~ /NEEDFS/ )
            {
                $Log->trace( "disk id: "
                        . $disk_info->{'id'}
                        . " is being zeroed..." );
                Tharn::snooze($polling_interval);
                $zeroing = 1;
                last;
            } ## end if ( $fsm_state eq "ZEROING"...)
        } ## end foreach my $disk_info ( @{$parsed_output...})

        if ( !$zeroing ) {
            $Log->exit() if $may_exit;
            return;
        }
    } ## end while ( time() <= $timeout)
    $Log->exit() if $may_exit;
    NACL::APISet::Exceptions::TimeoutException->throw(
        "Failed to wait for disks to get zeroed in $timeout seconds");
} ## end sub wait_for_aggr_zeroing_completion

=head2 get_root()

    my $root_aggr = NACL::STask::Aggregate->get_root(
        command_interface => $command_interface);

    or

    my $root_aggr = $aggregate->get_root( );

    or

    my %root_aggrs = NACL::STask::Aggregate->get_root(
       command_interface => $command_interface);

(Class or instance method) This method will return the root aggregate/s.
When called in scalar context, it returns the first aggregate which gets
displayed by the "storage aggregate show -fields aggregate,node -root true"
command. In array(hash) context, it will return the aggregates in the cluster
along with node information.

Example:
my %root_aggrs = NACL::STask::Aggregate->get_root(
   command_interface => $command_interface);

$VAR1 = {
          'test-vsim2' => 'aggr0_testcluster_1_02_0',
          'test-vsim1' => 'aggr0_test_vsim1_0'
        };

=over

=item Options

=over

=item command_interface, apiset_must, apiset_should, aggregate etc.

=back

=back

=cut

sub get_root {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts       = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {_optional_scalars qw(aggregate)},
    );

    my %common_options;
    $pkg_or_obj->_copy_common_component_params_with_ci(
        source => \%opts,
        target => \%common_options
    );

    my @root_aggr_objs = NACL::CS::StorageAggregate->fetch(
        %common_options,
        'filter'           => {root => "true"},
        'requested_fields' => [qw(aggregate node)]
    );

    my %root_aggrs;
    for my $aggr_obj (@root_aggr_objs) {
        $root_aggrs{$aggr_obj->node()->[0]} = $aggr_obj->aggregate();
    }

    $Log->exit() if $may_exit;
    return wantarray ? %root_aggrs : $root_aggr_objs[0]->aggregate();
}

=head2 add_to_aggr_list_of_vserver

    my $aggregate = NACL::STask::Aggregate->create(
    command_interface                    => $command_interface,
    aggregate                            => $aggregate,
    if_exists                            => $action,              # default 'die'
    nacltask_wait                        => $boolean,             # default '1'
    nacltask_add_to_aggr_list_of_vserver => $vserver_name,
    %other_options
    );
        (or)

    $aggregate->add_to_aggr_list_of_vserver( vserver => $vserver_name );
    where $aggregate is an object of "NACL::STask::Aggregate".

(Instance method) This method will add the aggr-list to the specified
vserver using the method add_to_aggr_list from the vserver.

=over

=item Options

=over

=item C<< vserver => $vserver_name >>

(Required)
Name of vserver.

=back

=back

=cut

sub add_to_aggr_list_of_vserver {
    $Log->enter() if $may_enter;
    my $self = shift;
    $self->_add_remove_aggr_list_helper( @_, api => 'add_to_aggr_list' );
    $Log->exit() if $may_exit;
    return;
} ## end sub add_to_aggr_list_of_vserver

=head2 remove_from_aggr_list_of_vserver

    $aggregate->remove_from_aggr_list_of_vserver( vserver => $vserver_name );

(Instance method) This method will remove the aggr-list from the specified
vserver using the method remove_from_aggr_list from the vserver.

=over

=item Options

=over

=item C<< vserver => $vserver_name >>

(Required)
Name of vserver.

=back

=back

=cut

sub remove_from_aggr_list_of_vserver {
    $Log->enter() if $may_enter;
    my $self = shift;
    $self->_add_remove_aggr_list_helper( @_, api => 'remove_from_aggr_list' );
    $Log->exit() if $may_exit;
    return;
} ## end sub remove_from_aggr_list_of_vserver

#Helper method for adding and removing the aggr-list to/from the vserver
#This method will add or remove aggr-list to/from the vserver, And this method will
#call the appropriate "api" based on the "api" option passed.
sub _add_remove_aggr_list_helper {
    $Log->enter() if $may_enter;
    my $self = shift;
    my %opts = $self->_common_validate_with(
        params          => \@_,
        additional_spec => {
            vserver => { type => SCALAR },
            api     => { type => SCALAR },
        },
    );
    my $api       = delete $opts{api};
    my $aggregate = delete $opts{aggregate};
    if ( $opts{'command_interface'}->mode() eq '7Mode' ) {
        $Log->exit() if $may_exit;
        return;
    } else {
        NACL::STask::Vserver->$api( %opts, 'aggr-list' => [$aggregate], );
    }
    $Log->exit() if $may_exit;
    return;
} ## end sub _add_remove_aggr_list_helper

=head2 vreport_fix()

    NACL::STask::Aggregate->vreport_fix(
        command_interface => $command_interface,
        'aggregate'       => $aggregate
    );

    or

    $aggregate_stask_obj->vreport_fix( );

(Class or instance method)
This method is used to fix vreport discrepancy by updating VLDB tables.
Supported only for CMode CLI.

=over

=item Options

=over

=item command_interface, apiset_must, apiset_should, aggregate etc.

=back

=back

=cut

=head2 undestroy

    my $aggr_obj = NACL::STask::Aggregate->undestroy(
        command_interface => $ci,
        aggregate         => $aggr,
        raidtree_id       => $raidtree_id
    );

    or

    my $aggr_obj = $aggr_task_obj->undestroy(
        raidtree_id          => $raidtree_id,
    );

(Class or instance method)
Undestroy an not existing aggregate.
This functionality is only available for 7mode filer.

=over

=item Options

=over

=item C<< nacltask_if_undestroyed => $action >>

(Optional) What to do if the aggregate to be undestroyed already exists.
If $action is "die" (the default), then fail with an NATE::BaseException exception.
If action is "pass", then do nothing.

=item C<< nacltask_verify => $nacltask_verify_boolean >>

(Optional)
If '0' (default), verification will not be performed.
If '1', verification will be performed to ensure that the undestroy operation did happen successfully.
If aggregate does not exists after the undestroy operation, then a
L<NACL::Exceptions::VerifyFailure|lib-NACL-Exceptions-VerifyFailure-pm>
exception is thrown.

=item command_interface, apiset_must, apiset_should, aggregate, raidtree_id etc.

All of the other various options to
L<< NACL::C::Aggregate->undestroy|lib-NACL-C-Aggregate-pm/undestroy >>

=back

=back

=over

=item Exceptions

=over

=item C<NACL::Exceptions::VerifyFailure>

This type of exception is thrown when aggregate does not exists after the undestroy operation.

=item C<NATE::BaseException>

This type of exception is thrown if the functionality is not supported on C-Mode.

Or

This type of exception is thrown if aggregate is already exists for 7mode.

=back

=back

=cut

sub undestroy {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my $pkg        = ref($pkg_or_obj) || $pkg_or_obj;
    my %opts       = $pkg_or_obj->_common_validate_with(
        params          => \@_,
        additional_spec => {
            raidtree_id => { type => SCALAR, optional => 1 },
            nacltask_if_undestroyed =>
                $pkg->_if_action_completed_already_validate_spec(),
            nacltask_verify => { type => BOOLEAN, default => 0 }
        },
        allow_extra => 1,
    );

    $Log->debug(
        sub {
            "Opts to NACL::STask::Aggregate->undestroy()" . Dumper( \%opts );
        }
    );

    my %common_opts;
    my $nacltask_if_undestroyed = delete $opts{nacltask_if_undestroyed};
    my $nacltask_verify         = delete $opts{nacltask_verify};
    my $aggregate               = $opts{aggregate};

    if ( $opts{command_interface}->is_cmode() ) {
        $Log->exit() if $may_exit;
        NATE::BaseException->throw(
            "This functionality is not supported on C-Mode");
    }
    $pkg->_copy_common_component_params_with_ci(
        source => \%opts,
        target => \%common_opts,
    );

    my $obj;
    try {
        $obj = $pkg->SUPER::undestroy(%opts);
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = shift;
        my $text      = $exception->text();
        if ($text =~ /aggregate .* already exists/i    # 7Mode
            )
        {
            if ( $nacltask_if_undestroyed ne 'pass' ) {
                $Log->exit() if $may_exit;
                NATE::BaseException->throw(
                    "Aggregate $aggregate is already exists. So exiting... ");
            }
        } else {
            $exception->throw();
        }
    };
    if ($nacltask_verify) {
        try {
            $pkg->find(
                filter => { aggregate => $aggregate },
                %common_opts
            );
        } ## end try
        catch NACL::Exceptions::NoElementsFound with {
            my $e = shift;
            $Log->exit() if $may_exit;
            NACL::Exceptions::VerifyFailure->throw(
                "Aggregate $aggregate does not exists after the undestroy operation."
                    . $e->text() );
        };
    } ## end if ($nacltask_verify)
    $Log->exit() if $may_exit;
    return $obj;
} ## end sub undestroy

=head2 sum_total_size

    $aggregate->sum_total_size(key => $key);

    $scalar = NACL::STask::Aggregate->sum_total_size(
      command_interface => $command_interface,
      aggregate         => $aggregate_name,
      key               => 'total',
    );

    %hash = NACL::STask::Aggregate->sum_total_size(
      command_interface => $command_interface,
      aggregate         => $aggregate_name,
      key               => 'all',
    );

(Class or Instance method)
This method gets the total size of an aggregate (usable space + snapshot space).
Can be used to obtain any of the measurable spaces in the aggregate depending on
the value of $key, including the total, used, reserved, or available space. If
desired, a hash of all the values can be returned by setting $key to "all".

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< aggregate => $aggregate_name >>

(Required for class method, Not Applicable for instance method)
Name of aggregate under test.

=item C<< key => "total" | "used" | "reserved" | "avail" | "all" >>

(Optional) The space measurement to return. The default is total. A reference
to a hash is returned if $key is set to "all", otherwise a scalar. Values are
measured in Bytes.

=item C<< calculate_spill => 0|1 >>

(Optional) If "1" (default) calculate the snapshot spill and add it to the
total. This is no longer done by default on the product on SN and later
releases.

This parameter is ignored on earlier releases, and the used space
reported is as follows:

If snapshot space used is greater than the aggregate snapshot reserve
then the space used is reported both in the Active File System (AFS) as well
as snapshot.  In order to avoid double counting that space, the total space
used is:
    AFS->{used} + Snapshot->{total}
Otherwise, the equation is
    AFS->{used} +  Snapshot->{used}

=item C<< determine_appropriate_nodescope_node => 0 | 1 >>

(Optional, defaults to 1)

=item C<< nodescope-node-name => $nodescope_node_name >>

(Optional)

Allows the caller to override the task and set the nodescope node name
explicitly.

=back

=back

=cut

sub sum_total_size {
    $Log->enter() if $may_enter;

    #my $pkg_or_obj = shift;
    my ($pkg_or_obj, @args) = @_;

    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            key => {type => SCALAR, default => "total"},
            determine_appropriate_nodescope_node =>
                {type => SCALAR, default => 1},
            'nodescope-node-name' => {type => SCALAR, optional => 1},
            calculate_spill       => {type => SCALAR, default  => 1},
        },
    );

    my $ci                 = $opts{command_interface};
    my $aggregate_name     = $opts{aggregate};
    my $get_nodescope_node = $opts{determine_appropriate_nodescope_node};
    my $calculate_spill    = $opts{calculate_spill};

    # There is support for df in CMode, but it's a hidden command so it's not
    # (yet) supported at the component layer. Using the nodescope APISet
    # allows us to use simpler code, but we have to get the correct Node
    my $apiset;
    my %options;

    $apiset = $ci->get_7m_or_nodescope_apiset();
    if (   !defined $opts{'nodescope-node-name'}
        && $ci->is_cmode()
        && $get_nodescope_node) {
        my %node_call_opts = (
            command_interface => $ci,
            aggregate         => $aggregate_name,
        );
        my $node = $pkg_or_obj->get_local_node_name(%node_call_opts);
        $options{'nodescope-node-name'} = $node;
    }

    my $parsed = $apiset->df(
        aggregate         => 1,
        name              => $aggregate_name,
        'reserved-space'  => 1,
        "scale-kilobytes" => 1,
        %options
    )->get_parsed_output();

    # We're using scale-kilobytes, and this provides the "KB" suffix
    # but we need to convert it to bytes
    foreach my $line (@{$parsed}) {
        foreach my $index ('total', 'used', 'avail', 'reserved') {
            $line->{$index} = NACL::C::UnitNormalization::convert_to_bytes(
                $line->{$index});
        }
    }
    my $parsed_afs  = $parsed->[0];
    my $parsed_snap = $parsed->[1];

    # Since we need the sum total volume size, adding the AFS + snapshot
    my $ret_hash = {};
    $ret_hash->{total} = $parsed_afs->{total} + $parsed_snap->{total};
    $ret_hash->{avail} = $parsed_afs->{avail} + $parsed_snap->{avail};
    $ret_hash->{reserved}
        = $parsed_afs->{reserved} + $parsed_snap->{reserved};

    # All reserved snapshot space is "used"
    $ret_hash->{used} = $parsed_afs->{used} + $parsed_snap->{total};

    if ($calculate_spill) {
        my $spill;
        my $name = "aggregate:$aggregate_name:wv_fsinfo_blks_used";
        my $blks_used
            = $apiset->stats_show(object_name => $name)->get_parsed_output()
            ->[0]->{value};

        $name = "aggregate:$aggregate_name:wv_fsinfo_blks_used_by_plane0";
        my $plane_0
            = $apiset->stats_show(object_name => $name)->get_parsed_output()
            ->[0]->{value};
        my $snap = $blks_used - $plane_0;

        # Blocks are 4K, df reports in 1K, so normalize
        $snap *= 4;
        if ($parsed_snap->{total} > $snap) {
            $spill = 0;
        } else {
            $spill = $snap - $parsed_snap->{total};
            $Log->debug("snap spill calculated as " . $spill . "K");
        }
        $ret_hash->{used} += $spill;
    }

    $Log->exit() if $may_exit;
    if ($opts{key} eq 'all') {
        return $ret_hash;
    } else {
        return $ret_hash->{$opts{key}};
    }
} ## end sub sum_total_size

sub _handle_failed_aggr_for_purge {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;

    my %opts = $pkg_or_obj->_common_validate_with(
        params      => \@args,
        allow_extra => 1,
    );

    my @vol_list = NACL::C::Volume->find(
        command_interface => $opts{'command_interface'},
        filter            => { 'aggregate' => $opts{'aggregate'} },
        allow_empty       => 1,
    );
    foreach my $vol_obj (@vol_list) {
        my $object = $vol_obj->vserver() . ":" . $vol_obj->volume();
        NACL::C::DebugVreport->fix(
            command_interface => $opts{'command_interface'},
            'object'          => $object,
            'type'            => 'volume'
        );
    } ## end foreach my $vol_obj (@vol_list)
    $Log->exit() if $may_exit;
} ## end sub _handle_failed_aggr_for_purge

sub _default_rng_suffix {
    return 'aggr';
}

=head2 online

    NACL::STask::Aggregate->online(
        command_interface    => $ci,
        volume               => $vol_name,
        vserver              => $vserver_name,
        nacltask_to_cleanup  => $boolean,
        nacltask_cleanup_manager => $object,
        %other_opts
    );

    or

    $vol_stask_obj->online(%other_opts,
                        nacltask_to_cleanup  => $boolean,
                        nacltask_cleanup_manager => $object,);

(Class or instance method)
Offlines a volume. This method provides additional services beyond what's
inherent to the product's volume online commands, as controlled and
described by C<nacltask_to_cleanup> and C<nacltask_cleanup_manager> options.

=over

=item Options

=over

=item C<< nacltask_to_cleanup => 0|1 >>

(Optional, defaults to 0(no to cleanup))
Flag indicating if this operation is to be cleaned up or not.

=item C<< nacltask_cleanup_manager >>

Cleanup manager to use for registering.
Default : Will use the default cleanup manager.

=item Other options

All of the other options accepted by
L<NACL::C::Aggregate::online|lib-NACL-C-Aggregate-pm/online>.

=back

=back

=cut

# Definition is from _AggregateVolume.

=head2 offline

    NACL::STask::Aggregate->offline(
        command_interface    => $ci,
        volume               => $vol_name,
        vserver              => $vserver_name,
        nacltask_to_cleanup  => $boolean,
        nacltask_cleanup_manager => $object,
        %other_opts
    );

    or

    $vol_stask_obj->offline(%other_opts,
                        nacltask_to_cleanup  => $boolean,
                        nacltask_cleanup_manager => $object,);

(Class or instance method)
Offlines a volume. This method provides additional services beyond what's
inherent to the product's volume offline commands, as controlled and
described by C<nacltask_to_cleanup> and C<nacltask_cleanup_manager> options.

=over

=item Options

=over

=item C<< nacltask_to_cleanup => 0|1 >>

(Optional, defaults to 0(no to cleanup))
Flag indicating if this operation is to be cleaned up or not.

=item C<< nacltask_cleanup_manager >>

Cleanup manager to use for registering.
Default : Will use the default cleanup manager.

=item Other options

All of the other options accepted by
L<NACL::C::Aggregate::offline|lib-NACL-C-Aggregate-pm/offline>.

=back

=back

=cut

# Definition is from _AggregateVolume.

=head2 restrict

    NACL::STask::Aggregate->restrict(
        command_interface    => $ci,
        volume               => $vol_name,
        vserver              => $vserver_name,
        nacltask_to_cleanup  => $boolean,
        nacltask_cleanup_manager => $object,
        %other_opts
    );

    or

    $vol_stask_obj->restrict(%other_opts,
                        nacltask_to_cleanup  => $boolean,
                        nacltask_cleanup_manager => $object,);

(Class or instance method)
Offlines a volume. This method provides additional services beyond what's
inherent to the product's volume restrict commands, as controlled and
described by C<nacltask_to_cleanup> and C<nacltask_cleanup_manager> options.

=over

=item Options

=over

=item C<< nacltask_to_cleanup => 0|1 >>

(Optional, defaults to 0(no to cleanup))
Flag indicating if this operation is to be cleaned up or not.

=item C<< nacltask_cleanup_manager >>

Cleanup manager to use for registering.
Default : Will use the default cleanup manager.

=item Other options

All of the other options accepted by
L<NACL::C::Aggregate::restrict|lib-NACL-C-Aggregate-pm/restrict>.

=back

=back

=cut

# Definition is from _AggregateVolume.

# This task method is just a wrapper around the "rename" method of component
# and it provides an option to register the rename operation for
# cleanup
# Note: If the cleanup is registered with STask instance, since the medhotd to
# rename happened through component and component updates the newname on the instance,
# Cleanup will work fine but the purge on that STask object will fail

sub rename {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    $pkg_or_obj->_rename(@args, 'element' => 'aggregate');
    $Log->exit() if $may_exit;
}


# Helper routine to determine the minimum number of disks required to start a new
# raid group based on the raidtype of an aggregate

sub _min_diskcount_to_start_new_rg {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            raidtype      => {type => SCALAR},
            aggregate     => {type => SCALAR},
            _disk_cs_objs => {type => ARRAYREF|UNDEF, optional => 1},
        },
        allow_extra => 1
    );

    my $raidtype = $opts{raidtype};
    my $tree     = $opts{aggregate};

    if ($raidtype eq "mixed_raid_type") {
        my %common_component_opts;
        $pkg_or_obj->_copy_common_component_params_with_ci(
            source => \%opts,
            target => \%common_component_opts,
        );
        my $ci = $common_component_opts{'command_interface'};

        # Get all the raidgroup <=> raidtype info
        my $raid_groups = NACL::STask::RaidGroup->list(
            'command_interface' => $ci,
            'tree'              => $tree,
        );

        my @disks = @{delete $opts{'_disk_cs_objs'}};

        my %raidtype;
        foreach my $disk (@disks) {
            my $raidgroup = $disk->raidgroup();
            # The raidgroup field in 7Mode is represented as 'aggr/plex/rg#' and the
            # raidgroup in CMode is represented just as 'rg#',
            # we need to extract the correct raid group. That is why the following
            # if($ci->is_cmode()){...} is necessary.
            my @rg;
            if($ci->is_cmode()){
                @rg = grep {$_ =~ /$raidgroup/} (keys %{$raid_groups});
                $raidgroup = $rg[0];
            }
            my $_raid_type = $raid_groups->{$raidgroup}->{raid_type};
            $_raid_type = "raid_dp" if($_raid_type eq "raidrd");
            $raidtype{$_raid_type} = $raidgroup;
        }
        if (scalar(keys %raidtype) > 1) {
            $Log->exit() if $may_exit;
            NATE::BaseException->throw(
                "Aggregate can't have multiple \"data - raidtypes\"");
        } else {
            $raidtype = (keys %raidtype)[0];
        }
    }

    my $min_diskcount = $pkg_or_obj->get_min_diskcount(
        'command_interface' => $opts{command_interface},
        'raidtype'          => $raidtype
    );

    $Log->exit() if $may_exit;
    return $min_diskcount;
}

=head2 get_min_diskcount

    NACL::STask::Aggregate->get_min_diskcount(
      command_interface => $command_interface,
      raidtype          => $raidtype,
    );

    or

    $aggr_task_obj->get_min_diskcount( raidtype => $raidtype, );

(Class or instance method) This method return the minimum diskcount
required to create an aggregate based on the raidtype provided.

Supports CMode CLI/ZAPI.

=over

=item Options

=over

=item C<< "command_interface=>$command_interface" >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< raidtype=>$raidtype >>

(Optional, default => raid_dp)
Raidtype of the aggregate to create

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=back

=back

=cut

sub get_min_diskcount {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my $pkg = ref($pkg_or_obj) || $pkg_or_obj;
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            raidtype => {type => SCALAR | UNDEF, optional => 1},
            _optional_scalars(qw/aggregate/),
        },
        allow_extra => 1
    );

    my $raidtype          = $opts{'raidtype'} || 'raid_dp';
    my $command_interface = $opts{'command_interface'};

    # force the 'force-small-aggregate' only when,
    my $version = ($pkg->_has_new_diskcount('command_interface' => $command_interface)) ? 'non_debug_build' : 'debug_build';
    $Log->exit() if $may_exit;
    return $MIN_DISK_COUNT{$version}{$raidtype};
}

# This is helper function to identify if the build has new mininum diskcount
# defined in it.This subroutine will return 1 if the build has new minimum
# diskcount defined, else it returns 0

sub _has_new_diskcount {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts = validate_with(
        params => \@args,
        spec   => {'command_interface' => {type => OBJECT}},
    );

    my $command_interface = delete $opts{'command_interface'};
    my $version_manager = $command_interface->version_manager();
    if ($command_interface->is_cmode()      # if CMode
        && !$version_manager->is_debug()    # non-debug build
        )
    {
        $Log->exit() if $may_exit;
        return 1;
    } else {
        $Log->exit() if $may_exit;
        return 0;
    }
}

=head2 space_comparison

    NACL::STask::Aggregate->space_comparison(
        command_interface => $command_interface,
        aggregate         => $Aggregate,
        %other_options
    );

    (or)

    $Aggr_obj->space_comparison();

(Class and Instance method)
Performs the space comparison for the aggregate
using different commands(df, show-space, show-footprint).

1. CS objects would be obtained for the aggregate show-space command to
get the "volume footprints", "total_used" attributes and for the volume
show-footprint command to get the "total footprint" value and for the
aggregate to get the values populated by "df -A" command.

2. The value of "volume_footprints" of aggregate show-space will be
compared to the total of all volume's "total_footprint" value of volume
show-footprint command.
volume Footprints(aggr show-space) = sum_all_vols(Total Footprint(volume show-footprint)).
If values are not equal but the difference of these values are less than or equals the
nacltask_threshold value then it will not return anything but if the values are beyond the
threshold value then the comparison will perform again based on the nacltask_retry value.
After retrying the comparison, if values are still not equal then it will throw an exception
of type NACL::STask::Exceptions::SpaceComparisonFailure.

3. Total of "used" and "total(snapshot total space)" will be compared
to the "total_used" space of aggregate show_space command.
Df –A used + snapshot total = Total Used( command output of aggr show-space).
If values are not equal but the difference of these values are less than or equals the
nacltask_threshold value then it will not return anything but if the values are beyond the
threshold value then the comparison will perform again based on the nacltask_retry value.
After retrying the comparison, if values are still not equal then it will throw an exception
of type NACL::STask::Exceptions::SpaceComparisonFailure.

Supports Cmode CLI/ZAPI & 7Mode CLI

=over

=item Options

=over

=item C<< volume_footprints >>

Volume Data and Metadata Footprint, possible value(s) are
<integer>[KB,MB,GB,TB,PB]
Filled in for CMode CLI, 7Mode/Nodescope CLI

=item C<< nacltask_threshold => $threshold_limit >>

(Optional, default => 1%) This option will provide the threshold limit for the values
If the comparison got failed then atleast the difference of the values should not be beyond the
thresholkd value.

=item C<< nacltask_retry => $num_of_retries >>

(Optional) If this option is provided, once the comparison failed between the values
and the difference is beyond the threshold value then the comparison has to be done
iteratively. The value for this option should be more than the default value '1'.
If '1' (default), comparison of the values will iterate the loop for one time.

=item C<< polling_interval => $interval >>

(Optional)
The interval, in seconds, the interval at which method has to wait
for the next iteration of comparison.

=item C<< total_footprint >>

Total Footprint, possible value(s) are <integer>[KB,MB,GB,TB,PB]
Filled in for CMode CLI/ZAPI, 7Mode CLI/ZAPI.

=item Other options

All of the other options accepted by
L<NACL::C::Aggregate::show_space|lib-NACL-C-Aggregate-pm/show_space>.
L<NACL::C::Aggregate::find|lib-NACL-C-Aggregate-pm/find>.
L<NACL::C::Volume::show_footprint|lib-NACL-C-Volume-pm/show_footprint>.

=back

=back

=over

=item Exceptions

=over

=item C<NACL::STask::Exceptions::SpaceComparisonFailure>

This exception is thrown when verification of field values does not pass, i.e
the field value for one or more fields does not match what was provided.

=back

=back

=cut

sub space_comparison {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            nacltask_threshold_value => {type => SCALAR, default => 1},
            nacltask_retries         => {type => SCALAR, default => 1},
            polling_interval         => {type => SCALAR, default => 10},
        },
    );

    my ($failure_msg, $mismatched_value);
    my $num_of_retries   = 0;
    my $threshold_limit  = delete $opts{nacltask_threshold_value};
    my $max_retry          = delete $opts{nacltask_retries};
    my $polling_interval = delete $opts{polling_interval};

    $pkg_or_obj ||= ref $pkg_or_obj;

    my $fooprint_hash;
    while ($num_of_retries <= $max_retry) {
        $fooprint_hash = $pkg_or_obj->_get_cs_objects(%opts, footprint => 1);
        if (defined($fooprint_hash->{total_footprint})) {
            if (_space_range_check_failed($fooprint_hash->{volume_footprint},
                    $fooprint_hash->{total_footprint}, $threshold_limit )) {
                $num_of_retries++;
                $Log->comment("\n As the calculated threshold value is ".
                "less than the specified threshold limit retrying to compare the values");
                Tharn::snooze $polling_interval;
            } else {
                # If the threshold limit is beyond the calculated limit
                # then exit from the loop
                last;
            }
        } else {
            $Log->comment(
                "\nNo total volume footprint found in Footprint table.".
                " This indicates no volumes found in the aggregate."
            );
            last;
        }
    }
    if ($num_of_retries > $max_retry ) {
        $failure_msg .= sprintf("\n%-20s%-30s\n", "Total_Footprint",
            "Volume_Footprints");
        $failure_msg .= sprintf("\n%-20s%-30s\n",
            $fooprint_hash->{total_footprint},
            $fooprint_hash->{volume_footprint});
        $mismatched_value->{total_footprint}
            = $fooprint_hash->{total_footprint};
        $mismatched_value->{volume_footprints}
            = $fooprint_hash->{volume_footprint};
    }
    my $space_hash;
    $num_of_retries = 0;
    while ($num_of_retries <= $max_retry) {
        $space_hash = $pkg_or_obj->_get_cs_objects(%opts, footprint => 0);
        if (_space_range_check_failed($space_hash->{used_including_snapshot_reserve},
            $space_hash->{total_space}, $threshold_limit )) {
            $num_of_retries++;
            $Log->comment("\n As the calculated threshold value is ".
            "less than the specified threshold limit retrying to compare the values");
            Tharn::snooze $polling_interval;
        } else {
            # If the threshold limit is beyond the calculated limit
            # then exit from the loop
            last;
        }
    }
    if ($num_of_retries > $max_retry ) {
        $failure_msg .= sprintf("\n%-20s%-30s\n",
            "Total_Space", "Used_Including_Snapshot_Reserve");
        $failure_msg .= sprintf("\n%-20s%-30s\n",
            $space_hash->{total_space},
            $space_hash->{used_including_snapshot_reserve});
        $mismatched_value->{total_space} = $space_hash->{total_space};
        $mismatched_value->{used_including_snapshot_reserve}
            = $space_hash->{used_including_snapshot_reserve};
    }
    if (defined($failure_msg)) {
        $Log->exit() if $may_exit;
        NACL::STask::Exceptions::SpaceComparisonFailure->throw(
            "\n\nSpace Comparison failed for the aggregate $opts{aggregate}\n".$failure_msg,
            mismatched_values  => $mismatched_value
        );
    }
    $Log->exit() if $may_exit;
} ## end sub space_comparison

=head2 get_mirror_vote

    NACL::STask::Aggregate->get_mirror_vote(
        command_interface => $command_interface,
        aggregate         => $Aggregate,
        %other_options
    );

    (or)

    $Aggr_obj->get_mirror_vote();

(Class and Instance method)
Supports Cmode CLI. This method is used only for mirrored aggregates.
This method is used to fetch the aggregate's mirror vote information
from the dblade and smdb table.
For dblade table information, the command used is
"run local raid_config info listmvm pretty". For smdb table information,
the command used is "debug smdb table aggr_vote show".

The return value will be in hash reference format as ,

    {
        aggregate   => "aggregate_name",
        dblade_info => {
                            "aggr_name" => "aggregate_name",
                            id          => 0,
                            rdb_dirty   => 0,
                            ..
                            ...
                    },
        smdb_info   => {
                         "aggrid" => 'aggr-uuid',
                         'cluster_uuid' => 'cluster-uuid',
                         ..
                         ...
                    },
    }

=over

=item Options

=over

=item C<< command_interface => $ci >>

(Required for class call) This contains a value of type C<NACL::C::CommandInterface::ONTAP>
object on which we need to fetch the mirror vote records.

=item C<< aggregate => $aggr >>

(Required for class call) This is the name of the aggregate for which the dblade & smdb table
entries are to be consolidated.

=item C<< _method => $operation >>

(Required) This is the name of the operation which is done before the verification.
Possible values are create, purge & wait_for_mirror.

=item C<< uuid => $uuid >>

(Optional) This is the aggregate uuid. This is should be used mandatory if the
_method value is purge or delete.

=item Other options

All of the other options such as command_interface, apiset_must, apiset_should , etc
are supported.

=back

=back

=cut

sub get_mirror_vote {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            aggregate => {type => SCALAR},
            command_interface => {
                 type => OBJECT,
                 isa  => 'NACL::C::CommandInterface::ONTAP'},
           _method    => {type => SCALAR},
           uuid       => {type => SCALAR, optional => 1},
        },
        allow_extra => 1,
    );

    my $command_interface = $opts{command_interface}
        || $pkg_or_obj->command_interface();
    my $aggr              = $opts{aggregate}
        || $pkg_or_obj->aggregate();

    my $method = $opts{_method};
    my ( $return, $return_hash, %common_opts, @smdb_parsed_op,
        %new_hash, $uuid );

    $pkg_or_obj->_copy_common_component_params_with_ci(
        source => \%opts,
        target => \%common_opts
    );

    if($method !~ /(purge|delete)/i) {
        $return = $pkg_or_obj->find(
            command_interface => $command_interface,
            filter            => {
                aggregate    => $aggr,
            },
            %common_opts,
            allow_empty => 1,
        );

        if( $return->state()->mirror() !~ /true/ ) {
            $Log->exit() if $may_exit;
            NACL::STask::Exceptions::MVRcheckFailure
                ->throw("This method is only for mirrored aggregates");
        }
        $uuid = $return->state()->uuid();
    }

    my $aggr_uuid = $uuid || $opts{uuid};
    $return_hash->{aggregate} = $aggr;

    my $apiset       = $command_interface->get_7m_or_nodescope_apiset();
    my $output       = $apiset->raid_config( info => "listmvm");
    my $array_output = $output->get_parsed_output();
    my $cnt = 0;

    foreach my $list (@$array_output) {
        if( (defined $list->{'aggr_name'})
         && ($list->{'aggr_name'} eq $aggr) ) {
            if( ++$cnt > 1 ) {
                $Log->exit() if $may_exit;
                NACL::STask::Exceptions::MVRcheckFailure->throw(
                    "Multiple blob mirror vote entries listed".
                    " for the aggregate"
                );
            } else {
                $return_hash->{"dblade_info"} = $list;
            }
        }
    }

    #Fetch the info from RDB
    if($command_interface->is_vldb_online()) {
        my $smdb_output = $command_interface->apiset(interface => "CLI")
            ->debug_smdb_table_show( table =>"aggr_vote",
            "aggrID" => $aggr_uuid );

        my $hash_ref = $smdb_output->get_parsed_output()->[0];
        foreach (keys %$hash_ref) {
            my $key = $_; my $val = $$hash_ref{$key};
            $key =~ s/-/_/;
            $new_hash{$key} = $val;
        }
        $return_hash->{"smdb_info"} = \%new_hash;
    } else {
        $Log->comment("vldb is offline hence we cannot fetch smdb details");
    }

    $Log->exit() if $may_exit;
    return $return_hash;
}

=head2 verify_mirror_votes

    NACL::STask::Aggregate->verify_mirror_votes(
        command_interface => $command_interface,
        aggregate         => $Aggregate,
        _method           => $operation,
        %other_options,
    );

(Class Method)
Supports Cmode CLI.
This method is used to compare the values passed from smdb & dblade tables.
If the values are not matching C<NACL::STask::Exceptions::MVRcheckFailure> exception is thrown.
If the vldb is online and the rdb_dirty flag is 1 or if the vldb is offline
and the rdb_dirty flag is 0, C<NACL::STask::Exceptions::MVRcheckFailure> is thrown since the
mirror vote not updated in RDB mirror vote table. No verification will be
performed in such cases.

=over

=item Options

=over

=item C<< aggregate => $aggr >>

(Required) This is the name of the aggregate for which the dblade & smdb tables
entries are to be matched with. Only mirrored aggregate needs to be passed.

=item C<< _method => $operation >>

(Required) This is the name of the operation which is done before the verification.
Possible values are create, purge & wait_for_mirror.

=item C<< blob => $blob >>

(Optional) This option contains the value of the blob which needs to be verified
with the blob value present in dblade entry. If its not equal,
C<NACL::STask::Exceptions::MVRcheckFailure> exception is thrown.
Defaults to 0, so no check is done.

=item C<< uuid => $uuid >>

(Optional) This is the aggregate uuid. This is should be used mandatory if the
_method value is purge or delete.

=item Other options

All of the other options such as command_interface, apiset_must, apiset_should , etc
are supported.

=back

=back

=over

=item Exceptions

=over

=item C<NACL::STask::Exceptions::MVRcheckFailure>

This exception is thrown when the values present in the mvr_info fields are not
equal to the values present in the smdb & dblade table fields.

=item C<NACL::STask::Exceptions::MVRcheckFailure>

This exception is thrown when the vldb is online and the rdb_dirty flag is 1 or
if the vldb is offline and the rdb_dirty flag is 0.

=back

=back

=cut

sub verify_mirror_votes {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            aggregate         => {type => SCALAR},
            _method           => {type => SCALAR},
            command_interface => {type => OBJECT,isa  => 'NACL::C::CommandInterface::ONTAP'},
            blob              => {type => SCALAR, default  => 0},
            uuid              => {type => SCALAR, optional => 1},
        },
        allow_extra => 1,
    );

    my $ci     = $opts{command_interface};
    my $aggr   = $opts{aggregate};
    my $method = $opts{_method};
    my $vldb_online = $ci->is_vldb_online();

    my $rdb_dirty_flag;
    if( $vldb_online ) {

        my %pass_opts;
        $pass_opts{uuid} = $opts{uuid} if(defined $opts{uuid});
        my $return_hash = $pkg->get_mirror_vote(
            command_interface => $ci,
            aggregate         => $aggr,
            _method           => $method,
            %pass_opts,
        );

        $rdb_dirty_flag      = $return_hash->{'dblade_info'}->{'rdb_dirty'};
        my $dblade_hash_info = $return_hash->{'dblade_info'};
        my $smdb_hash_info   = $return_hash->{'smdb_info'};

        if(  $rdb_dirty_flag eq '0'  ) {
            #check mvr_info with dblade info & smdb info tables
            $Log->comment("Vldb is online as well as the rdb_dirty flag is set to 0,".
                " hence we need to check in both the dblade & smdb tables");
            if($opts{blob}) {
                if($opts{blob} eq $return_hash->{'dblade_info'}->{'blob'}) {
                    $Log->comment("Blob passed has same value as that ".
                    "present in the dblade table");
                } else {
                    $Log->exit() if $may_exit;
                    NACL::STask::Exceptions::MVRcheckFailure->throw("Blob value is".
                        " not equal in the dblade table entries");
                }
            }
            ## as per burt#869335 some fields can be skipped while comparing the
            ## table entries . Entries to be verified : plex0_ctime plex1_id nplexes plex0_id plex1_ctime
            $pkg->_check_with_table_info(
                dblade_hash  => $dblade_hash_info,
                smdb_hash    => $smdb_hash_info,
        field_to_check => [qw(plex0_ctime plex1_id nplexes plex0_id plex1_ctime)]
            );
        } else {
            $Log->exit() if $may_exit;
            NACL::STask::Exceptions::MVRcheckFailure->throw("Mirror vote not updated".
                " in RDB mirror vote table even though VLDB is online");
        }
    } elsif ((!$vldb_online) && ( $rdb_dirty_flag eq '0' )) {
        $Log->exit() if $may_exit;
        NACL::STask::Exceptions::MVRcheckFailure->throw("RDB dirty flag is not 1 even".
            " though VLDB is offline and mailbox mirror vote data is dirtied");
    }

    $Log->exit() if $may_exit;
}

sub _check_with_table_info {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    my %opts = validate_with(
        params => \@args,
        spec   => {
            dblade_hash => {type => HASHREF},
            smdb_hash   => {type => HASHREF},
        field_to_check => {type => ARRAYREF, optional => 1}
        },
    );

    my $dblade_hash = $opts{dblade_hash};
    my $smdb_hash   = $opts{smdb_hash};
    my $field_to_check = $opts{field_to_check} ;

    my ( @fields_eq_both_tables, @fields_ne_both_tables,
         @fields_not_def_both_table );

    foreach (keys %$dblade_hash) {
        my $dblade_field = $_;
        if(defined($$smdb_hash{$dblade_field})) {
            if( $$dblade_hash{$dblade_field}
                eq $$smdb_hash{$dblade_field} ) {
                push @fields_eq_both_tables, $dblade_field;
            } else {
                push @fields_ne_both_tables, $dblade_field;
            }
        } else {
            next;
        }
    }
    my @fields_ne;
    foreach my $chk_field (@{$field_to_check}){
       if(grep {$chk_field eq $_} @fields_ne_both_tables ){
            push @fields_ne , $chk_field
       }
    }
    if(@fields_ne) {
        $Log->exit() if $may_exit;
        NACL::STask::Exceptions::MVRcheckFailure->throw("MVR verification failed in "
            ."both dblade & smdb table for the following entries:"
        .Dumper @fields_ne,
            unmatched_fields => \@fields_ne);
    }

    $Log->comment("Fields ".join(",",@fields_eq_both_tables)." are having equal values in "
        ." both the dblade & smdb table entries") if(@fields_eq_both_tables);
    $Log->exit() if $may_exit;
}

sub _get_cs_objects {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {footprint => {type => SCALAR}},
        allow_extra     => 1
    );
    my (%common_opts, $total_footprint);
    $pkg_or_obj->_copy_common_component_params_with_ci(
        source => \%opts,
        target => \%common_opts
    );
    my $aggr_show_space_obj
        = NACL::C::StorageAggregate->show_space(%common_opts,
        filter => {aggregate => $opts{aggregate}},);

    my %hash;
    if ($opts{footprint} == 1) {
        my @vol_objs = NACL::STask::Volume->find(
            %common_opts,
            filter      => {aggregate => $opts{aggregate}},
            allow_empty => 1,
        );

        foreach my $vol_obj (@vol_objs) {
            my $foot_printobj = NACL::C::Volume->show_footprint(%common_opts,
                filter => {volume => $vol_obj->volume()},);
            $foot_printobj->{total_footprint}
                = NACL::C::UnitNormalization::convert_to_bytes(
                $foot_printobj->{total_footprint});
            $total_footprint += $foot_printobj->{total_footprint};
        }
        $hash{volume_footprint} =
            NACL::C::UnitNormalization::convert_to_bytes($aggr_show_space_obj->{volume_footprints});
        $hash{total_footprint}  = $total_footprint;
    } else {
        my $used_space
            = $aggr_show_space_obj->{used_including_snapshot_reserve};

        $used_space
            = NACL::C::UnitNormalization::convert_to_bytes($used_space);
        $hash{used_including_snapshot_reserve} = $used_space;

        my $node;
        my $apiset
            = $common_opts{command_interface}->get_7m_or_nodescope_apiset();
        if ($common_opts{command_interface}->is_cmode()) {
            my %node_call_opts = (
                command_interface => $common_opts{command_interface},
                aggregate         => $opts{aggregate},
            );
            $node = $pkg_or_obj->get_local_node_name(%node_call_opts);
        }
        my $parsed = $apiset->df(
            aggregate             => 1,
            name                  => $opts{aggregate},
            'reserved-space'      => 1,
            "scale-kilobytes"     => 1,
            'nodescope-node-name' => $node,
        )->get_parsed_output();

        foreach my $line (@{$parsed}) {
            foreach my $index ('total', 'used', 'avail', 'reserved') {
                $line->{$index}
                    = NACL::C::UnitNormalization::convert_to_bytes(
                    $line->{$index});
            }
        }
        my $parsed_afs  = $parsed->[0];
        my $parsed_snap = $parsed->[1];

        $hash{total_space} = $parsed_afs->{used} + $parsed_snap->{total};
    }
    $Log->exit() if $may_exit;
    return \%hash;
} ## end sub _get_cs_objects

# Consider the higher value for which the comparison is happening and
# calculate the threshold value with the threshold_limit specified by the user
# and compare this threshold value with the difference of two values, If threshold value
# is less than the difference then this method returns 1 else returns 0

sub _space_range_check_failed {
    my @opts = @_;
    my $threshold_value = 0;
    my $target_value = 0;
    my $limit = $opts[2];
    if((defined $opts[0]) && (defined $opts[1])) {
        if ($opts[0] < $opts[1]) {
            $target_value = $opts[1] - $opts[0];
            $threshold_value = $opts[1] * ($limit/100);
        } else {
            $target_value = $opts[0] - $opts[1];
            $threshold_value = $opts[0] * ($limit/100);
        }
    }
    ($threshold_value < $target_value) ? return 1 : return 0;
} ## end sub _space_range_check_failed


#----------------------------------------------------------------------------
# This subroutine will return options hash to the aggregate creation according
# to the "nacltask_use_partitions" flag and if the node doesn't have required
# type of logical disks then exception will be thrown.
#----------------------------------------------------------------------------
# %options = $pkg->_add_partition_disklist(
#             opts => \%opts);
# Options
# C<< opts => %Hash_ref >>
#(Required for class method,)
# Hash reference of aggregate create options passed by user at main subroutine.

sub _add_partition_disklist {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %options    = validate_with(
        params => \@_,
        spec   => { opts => { type => HASHREF }, },
    );
    my %opts                    = %{ $options{opts} };
    my $command_interface       = $opts{command_interface};
    my $nacltask_use_partitions = delete $opts{nacltask_use_partitions};

    #Consider the node parameter for finding the disks from the node
    my %node = ('nodescope-node-name' => delete $opts{'nodescope-node-name'});

    #Which partition type to use to get disks list
    my $logical_type = ($nacltask_use_partitions) ? 'partition' : 'physical';

    #lets not do anything if disk list or partition list  is mentioned
    return %opts
      if ( ( defined $opts{'disklist'} )
        or ( defined $opts{'partitionlist'} )
        or ( !defined $nacltask_use_partitions ) );

    # return %opts unless ( defined $nacltask_use_partitions );
    my %disk_opts = ();
    $pkg_or_obj->_hash_copy_defined(
        source => \%opts,
        target => \%disk_opts,
        map    => {
            disktype    => 'disk_type',
            chksumstyle => 'checksum_type',
            diskrpm     => 'rpm'
        },
    );
    my $partition_index = 1 if ( $logical_type =~ /partition/i );
    my $diskcount =
        (defined($opts{'mirror'}) and $opts{'mirror'} =~ /true/i)
      ? $opts{'diskcount'} / 2
      : $opts{'diskcount'};

    $Log->debug(
        sub { 'Collecting disks of following types: ' . Dumper \%disk_opts } );

    #Get All disks properties:
    my @cs_disk_objs = NACL::CS::RaidConfigInfoListDisk->fetch(
        command_interface => $command_interface,
        %node
    );

    # Find all spare partitions
    my @partitions = NACL::CS::RaidConfigInfoListDisk->get_matching_objects(
        state_objs => \@cs_disk_objs,
        filter     => {
            "logical_disk_type" => "partition",
            spare               => 1,
            partition_index     => $partition_index,
            %disk_opts,
        },
    );

    # Find all spare physicals
    my @physicals = NACL::CS::RaidConfigInfoListDisk->get_matching_objects(
        state_objs => \@cs_disk_objs,
        filter     => {
            "logical_disk_type" => "physical_disk",
            spare               => 1,
            %disk_opts,
        },
    );

    # Find all spare container_disk
    my @containers = NACL::CS::RaidConfigInfoListDisk->get_matching_objects(
        state_objs => \@cs_disk_objs,
        filter     => { "logical_disk_type" => "container_disk", },
    );

#Separate disks based on logical disk type. [partition_disks, physical_disks, container_disk]
    my %disk_logical_types = (
        partition      => \@partitions,
        physical       => \@physicals,
        container_disk => \@containers
    );

    #Check disks availability
    my $partition_disk_count = @partitions;
    my $physical_disk_count  = @physicals;

    ## Check the node has mixed logical disks
    my $mixed_logical_type =
      ( $partition_disk_count && $physical_disk_count ) ? 1 : 0;

    #filter required disks based on logical type
    #repeat for checksum_type and disk_type
    my @cs_disk_objs_logical_type = @{ $disk_logical_types{$logical_type} };

   # Replace partition disks with container_disks if(partition disks to be used)
    my %pool_disks = ( 0 => [], 1 => [] );
    if ($nacltask_use_partitions) {
        ## Get the container disk for the partition
        foreach my $disk (@cs_disk_objs_logical_type) {
            my @cont_disk =
              NACL::CS::RaidConfigInfoListDisk->get_matching_objects(
                state_objs => \@containers,
                filter => { container_disk_id => $disk->container_disk_id() },
              );
            push(
                @{ $pool_disks{ $disk->{pool} } },
                $cont_disk[0]->cluster_name()
            ) if ( scalar @cont_disk > 0 );
        }
    } else {
        map { push( @{ $pool_disks{ $_->{pool} } }, $_->cluster_name() ) }
          @cs_disk_objs_logical_type;
    }
    $Log->debug( sub { "Avilable $logical_type are: " . Dumper \%pool_disks } );
    my @pool_0           = @{ $pool_disks{0} };
    my @pool_1           = @{ $pool_disks{1} };
    my $pool1_disk_count = @pool_1;
    my $pool0_disk_count = @pool_0;
    ## Check for the both pools has required amount of logical_type disks
    if ( $pool0_disk_count >= $diskcount && $pool1_disk_count >= $diskcount ) {
        $Log->comment(
            "disk count $diskcount $pool0_disk_count and $pool1_disk_count");
        $Log->trace(
                "Sufficient $logical_type disks are available on both pools "
              . " Pool0 Disks : @pool_0 and Pool1 Disks: @pool_1 " );
        @pool_0 = @pool_0[ 0 .. $diskcount - 1 ];
        @pool_1 = @pool_1[ 0 .. $diskcount - 1 ];
        ## Check for the sufficient logical_type disks are available from combination of both pool
        ## If sufficient disks are available from both pool throw warning message
    } elsif ( ( $pool0_disk_count + $pool1_disk_count ) >= $opts{'diskcount'} )
    {
        if ( !exists $opts{'ignore-pool-checks'} ) {
            $Log->warn(
"Sufficient $logical_type disks are available on combination of both pools  this "
                  . "might cause the aggregate creation fail (To force the aggregate"
                  . " creation use -ignore-pool-checks option with disklist) "
                  . " Pool0 Disks : @pool_0 and Pool1 Disks: @pool_1 " );
        }
        my @all_disks = ( @pool_0, @pool_1 );
        @pool_0 = @all_disks[ 0 .. $diskcount - 1 ];
        @pool_1 =
          @all_disks[ $diskcount .. ( ( $diskcount + $diskcount ) - 1 ) ];
        ## Throw exception in case combination of both pool doen't has sufficient disks
    } elsif ( ( $pool0_disk_count + $pool1_disk_count ) < $opts{'diskcount'} ) {
        $Log->exit() if $may_exit;
        NATE::BaseException->throw(
                "Sufficient $logical_type disks are not available "
              . "Pool0 disks -> @pool_0 and Pool1 -> @pool_1  \n parameters passed :"
              . Dumper \%disk_opts );
    }

    if ($mixed_logical_type) {
        $opts{'disklist'} = \@pool_0
          ; #we should be able to use pool1 disks if there are not enough disks in pool0
        $opts{'mirror-disklist'} = \@pool_1 if ( $opts{'mirror'} =~ /true/i );
        delete $opts{'mirror'};
        delete $opts{'diskcount'};
        delete $opts{'diskrpm'};
        delete $opts{'disktype'};
        delete $opts{'chksumstyle'};
        delete $opts{'node'};
    } else {
        delete $opts{'ignore-pool-checks'};
    }
    $Log->exit() if $may_exit;
    return %opts;
}

1;

