#
# Copyright (c) 2014 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Ifconfig Task Module
## @author pateln@netapp.com, dl-nacl-dev@netapp.com
## @status shared
## @pod here

package NACL::STask::Client::Ifconfig;

use strict;
use warnings;
use feature 'state';

use base qw/NACL::C::Client::Ifconfig NACL::STask::STask/;
use Params::Validate qw(validate SCALAR UNDEF SCALARREF BOOLEAN);
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Net::IP qw(:PROC);
use NACL::STask::Exceptions::Client::Ifconfig::NotCreated;
use NACL::STask::Exceptions::Client::Ifconfig::NotDeleted;
use NACL::C::Exceptions::Client::Ifconfig::AlreadyExists;
use NACL::APISet::Exceptions::CommandFailedException qw(:try);

=head1 NAME

NACL::STask::Client::Ifconfig

=head1 DESCRIPTION

C<NACL::STask::Client::Ifconfig> provides a number of well-defined but
potentially complex or multi-step methods related to Client-side Ifconfig.

It builds on top of and is a derived class of C<NACL::C::Client::Ifconfig>. This
means that any functionality present in
L<NACL::C::Client::Ifconfig|lib-NACL-C-Client-Ifconfig-pm> will also be
present in this STask object. (functionality which is multi-step will have
overridden implementations defined here, single step functionality will be
directly re-used from the component)

=head1 ATTRIBUTES

=head2 command_interface

(Required, isa 'NACL::C::Client') A command interface object which represents
the client on which we are manipulating the mount points.
See L<NACL::C::Component::command_interface|lib-NACL-C-Component-pm/command_interface>.

=head2 interface

(Required) Name of network interface on client.

=head1 METHODS

=head2 add_address

       NACL::STask::Client::Ifconfig->add_address(
        command_interface  => $Client,
        interface          => $Interface,
        address            => $ipv6_address,
        prefixlen          => $Prefixlen,
        nacltask_if_exists => 'reuse',
        nacltask_verify    => 1,
    );

(Class method) Request to add Ipv6/Ipv4 address to an interface
This method provides additional services beyond what's inherent to the
product's Ifconfig add_address method, as controlled and described
by the "nacltask_if_exists" option.

=over

=item Options

=item C<< command_interface => $Client >>

(Required) A component object that represents the host to which
to send commands. See NACL::C::Component::command_interface.

=item C<< interface => $string >>

(Required) Name of the interface to which ipv6 needs to be added.

=item C<< address  => $string >>

(Required) IPV4 or IPV6 address value needs to be added to the interface.

=item C<< netmask => $string >>

(optional) Netmask value to be assigned to the interface if it is IPV4 address.

=item C<< prefixlen => $string >>

(optional) prefixlen value to be assigned to the interface if it is IPV6 address.

=item C<< _was_created => \$scalar >>

(Optional) When this option is provided a reference to a scalar
variable, the variable gets filled in with a boolean value
describing whether the IP address was found (value will be 0;
this scenario is possible when nacltask_if_exists => "reuse") or
whether the IP address was added (value will be 1). This is necessary
to determine whether the IP address needs to be deleted up later.

        my $was_created;

        NACL::STask::Client::Ifconfig->add_address(
            command_interface  => $Client,
            interface          => $Interface,
            address            => $ipv6_address,
            prefixlen          => $Prefixlen,
            nacltask_if_exists => 'reuse',
            nacltask_verify    => 1,
            '_was_created'     => \$was_created,
        );

       # Now determine whether to delete the IP address, since we're not sure
       # whether we reused an existing IP address or created a new one
       if ($was_created) {
           NACL::STask::Client::Ifconfig->remove_address(
                command_interface => $Client,
                interface         => $Interface,
                address           => $ipv6_address,
                prefixlen         => $Prefixlen,
          );

       }


=item C<< nacltask_if_exists => $action >>

(Optional) Specifies an action to be taken if the IP address
already exists. If set to die an exception is thrown if IP address
already exists. If set to reuse, no action is performed. And if set to
remove_address it will delete an existing IP address before creating a new one.

=item C<< nacltask_verify => 0|1 >>

(Optional) If '0' (default), verification will not be performed.
If '1', verification will be performed to ensure that the
IP address creation did happen successfully.

=item Other options

All of the other options such as command_interface, apiset_must,
apiset_should , etc are supported. And it also supports all the
options as described in
L<< NACL::C::Client::Ifconfig->create|lib-NACL-C-Client-Ifconfig-pm/create >>

=back

=over

=item Exceptions


=item C<NACL::STask::Exceptions::Client::Ifconfig::NotCreated>

This type of exception is thrown if the requested IP address
was not added on the interface.

=item NACL::C::Exceptions::Client::Ifconfig::AlreadyExists

This type of exception is thrown when an attempt is made to add an IP address
that already is added to that interface

=item Ifconfig Component exceptions

=back

=cut


sub add_address {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;

    state $spec = {
        nacltask_if_exists => $pkg_or_obj->_if_exists_validate_spec,
        nacltask_verify    => {type => BOOLEAN, default => 0},
        _was_created       => {type => SCALARREF, optional => 1},
    };

    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => $spec,
        allow_extra     => 1,
    );

    # set options to bring up the interface
    my %up_opts = ();
    if (!ref($pkg_or_obj)) {
        my $common_param_with_ci
            = $pkg_or_obj->_copy_common_component_params_with_ci(
            source => \%opts,);
        my $primary_keys = $pkg_or_obj->get_primary_keys_options(%opts);
        %up_opts = (%{$common_param_with_ci}, %{$primary_keys});
    }

    my $self;
    $up_opts{inet6} = 1 if (Net::IP::ip_is_ipv6($opts{'address'}));
    my $was_created = delete $opts{_was_created};
    $$was_created = 0;
    my $if_exists = delete $opts{nacltask_if_exists};
    my $verify    = delete $opts{nacltask_verify};


ADD: {
        use warnings qw(exiting);
        try {

            $pkg_or_obj->up(%up_opts);
            $pkg_or_obj->SUPER::add_address(%opts);
            $$was_created = 1;
        }

        catch NACL::C::Exceptions::Client::Ifconfig::AlreadyExists with {
            my $exception = shift;
            $Log->trace('IP address already added to the Interface');

            $self = $pkg_or_obj->_element_exists_handler(
                create_opts         => \%opts,
                purge_opts          => \%opts,
                nacltask_if_exists  => $if_exists,
                exception           => $exception,
            );
            if (!$self) {
                no warnings 'exiting';
                redo ADD;
            }
        };
    }

    if ($verify && $$was_created) {
        my %common_opts;
        $pkg_or_obj->_copy_common_component_params(
            source => \%opts,
            target => \%common_opts
        );

        my $found = NACL::CS::Client::Ifconfig->fetch(
            command_interface => $opts{command_interface},
            filter            => {'interface' => $opts{interface},},
            allow_empty       => 1,
            %common_opts,
        );
        my $flag = _verify($found,%opts);
        if (!$flag) {
        NACL::STask::Exceptions::Client::Ifconfig::NotCreated->throw(
            "Interface was not created");
        }
    }
    $Log->exit() if $may_exit;
} ## end sub add_address

=head2 remove_address

    NACL::STask::Client::Ifconfig->remove_address(
        command_interface => $Client,
        interface         => $Interface,
        address           => $ipv6_address,
        prefixlen         => $Prefixlen,
    );

(Class Method) This task method will remove the IPV4/IPV6
address from a interface in a  Client. It also verifies if
the address was deleted.

=over

=item C<< command_interface => $ci >>

(Required) A component object that represents the host to which
to send commands. See NACL::C::Component::command_interface.

=item C<< interface => $string >>

(Required) Name of the interface to which ipv6 needs to be added.

=item C<< address  => $string >>

(Required) IPV4 or IPV6 address value needs to be added to the interface.

=item C<< netmask => $string >>

(optional) Netmask value to be assigned to the interface if it is IPV4 address.

=item C<< prefixlen => $string >>

(optional) prefixlen value to be assigned to the interface if it is IPV6 address.

=item C<< nacltask_verify => 0 | 1 >>

(Optional) If '0' (default), verification will not be performed.
If '1', verification will be performed to ensure that the
deletion did happen successfully.

=back

=over

=item Exceptions

=item C<NACL::STask::Exceptions::Client::Ifconfig::NotDeleted>

This type of exception is thrown when the requested IP address
is not deleted from an interface.

=item Ifconfig Component exceptions

=back

=cut


sub remove_address {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;

    state $spec = {
        nacltask_verify => {type => BOOLEAN, default  => 0},
    };

    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => $spec,
        allow_extra     => 1,
    );

    my $verify = delete $opts{nacltask_verify};
    $pkg_or_obj->SUPER::remove_address(%opts);

    if ($verify) {
        my %common_opts;
        $pkg_or_obj->_copy_common_component_params(
            source => \%opts,
            target => \%common_opts
        );

        my $found = NACL::CS::Client::Ifconfig->fetch(
            command_interface => $opts{command_interface},
            filter            => {'interface' => $opts{interface}},
            allow_empty       => 1,
            %common_opts,
        );

        my $flag = _verify($found,%opts);
        if ($flag) {
        NACL::STask::Exceptions::Client::Ifconfig::NotDeleted->throw(
            "IPV6 address was not deleted");
        }
    }
    $Log->exit() if $may_exit;
} ## end sub remove_address

sub _verify {
    $Log->enter() if $may_enter;
    my ($found,%opts)   = @_;
    my $flag    = 0;
    my %up_opts = ();
    $up_opts{inet6} = 1 if (Net::IP::ip_is_ipv6($opts{'address'}));
    if (defined $found) {
        if (defined $up_opts{inet6}) {

            #IPV6
            my $inet_addr = $opts{address} . "/" . $opts{prefixlen};
            my $ipv6      = $found->inet6_info();
            foreach my $val (@{$ipv6}) {
                if ($val->{addr} eq $inet_addr) {
                    $flag = 1;
                    last;
                }
            }
        } else {

            #IPV4
            my $inet_addr = $opts{address};
            my $ipv4      = $found->inet_addr();
            $flag = 1 if ($ipv4 eq $inet_addr);
        }
    }
    $Log->exit() if $may_exit;
    return $flag;
} ## end sub _verify


# Should not be used by external callers, hence is not documented. This is a
# wrapper around the remove_address method.
# Being a private routine, it does away with the validation, assuming that
# everything is passed in correctly.
sub purge {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->remove_address(@args);
    $Log->exit() if $may_exit;
}

1;
