## @summary Mount Task Module
## @author dl-nacl-dev@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::STask::Client::Mount

=head1 DESCRIPTION

C<NACL::STask::Client::Mount> provides a number of well-defined but
potentially complex or multi-step methods related to Client-side mount points.
At the moment only Volumes or Qtrees can be mounted using this library.
(For Linux clients it uses "do_mount" to perform an NFS mount, for Windows
clients it uses "net use" to map a CIFS share)

It builds on top of and is a derived class of C<NACL::C::Client::Mount>. This
means that any functionality present in
L<NACL::C::Client::Mount|lib-NACL-C-Client-Mount-pm> will also be
present in this STask object. (functionality which is multi-step will have
overridden implementations defined here, single step functionality will be
directly re-used from the component)

This library only performs the client side steps required to mount along with
figuring out information from the server (if required). However,it does not
do any kind of setup on the server itself for creating an export point.

Cleanup can be registered for the following methods

Cleanup methods are,
     Mount Method         Cleanup Method
    ----------------------------------------
     create               purge

=head1 ATTRIBUTES

=head2 command_interface

(Required, isa 'NACL::C::Client') A command interface object which represents
the client on which we are manipulating the mount points.
See L<NACL::C::Component::command_interface|lib-NACL-C-Component-pm/command_interface>.

=head2 mount_point

(Required) The mount point on the client. (For Linux this would be the absolute
path at which the volume/qtree is mounted; for windows this would be the drive letter)

=head2 _create_opts

Options/Params that are passed to create method.
These will be used in 'remount' method to remount the volume/qtree.

=cut

package NACL::STask::Client::Mount;

use strict;
use warnings;

use NACL::C::Volume;
use NACL::C::Qtree;
use NACL::C::VserverCifsShare;
use NACL::CS::Interface;
use Net::IP qw(:PROC);
use NACL::Util::IPv6 qw(convert_ipv6_address_format);
use Storable qw(freeze thaw);
use NATE::Exceptions::Argument qw(:try);
use NACL::Exceptions::NoElementsFound qw/:try/;
use Params::Validate
    qw(validate validate_with SCALAR UNDEF BOOLEAN SCALARREF ARRAYREF HASHREF OBJECT);
use NATE::BaseException qw(:try);
use Scalar::Util qw/blessed/;
use NACL::ComponentUtils qw(Dumper _dump_one);
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use NACL::GeneralUtils qw(nacl_method_retry);
use NACL::C::Exceptions::Client::FileDirOps::ReadTestFailed;
use NACL::STask::Exceptions::Client::Mount::NotMounted ();
use NACL::C::Exceptions::Client::Mount::DoesNotExist;
use NACL::Exceptions::VerifyFailure;
use NACL::C::Exceptions::Client::FileDirOps::WriteTestFailed;
use NACL::MTask::PacketCapture;
use NATE::ParamSet qw(param_global);
use NACL::C::Client::FileDirOps ;
use NACL::C::VserverCifsSession;
use NATE::Util::Sockets qw(get_hostname);

use base qw/NACL::C::Client::Mount NACL::STask::STask/;

my @nfs_versions = qw (4.1 4 3);

use Class::MethodMaker [
    hash => [qw/_create_opts/],
];

=head1 METHODS

=head2 new

    my $mount_obj = NACL::STask::Client::Mount->new(
        command_interface => $client,
        mount_point       => $mount_point
    );

This returns a new Mount STask object. Note that this does not actually
create a mount point on the client, it merely returns a Perl object.

=head2 find

    my @mount_objs = NACL::STask::Client::Mount->find(
        command_interface => $client,
        %other_opts
    );

(Class method)
A constructor of sorts, which discovers which mount point are present on the
client. In an array context it creates an STask object for each present mount
point and returns them. In a scalar context it creates an STask object for just
one mount point.

See <L<NACL::C::Component::find|lib-NACL-C-Component-pm/find> for a more
detailed description along with explanation of all the options accepted.

=head2 state

    my $mount_state = $mount_obj->state();

(Instance method)
Fetch the current state of the mount point associated with this task object, and
return a Component State object corresponding to this task (the returned CS
object is of type L<NACL::CS::Client::Mount|lib-NACL-CS-Client-Mount-pm>.

See <L<NACL::C::Component::state|lib-NACL-C-Component-pm/state> for a more
detailed description along with explanation of all the options accepted.

=head2 create

    1. Recommended way for Unix mounts:
    my $mount = NACL::STask::Client::Mount->create(
        command_interface => $client,
        volume            => $vol_C_object,
        nacltask_to_cleanup      => 1,          # default 0,
        nacltask_cleanup_manager => $CleanupObj,
        %other_options
    );

    2. Recommended way for CIFS share mapping:
    my $mount = NACL::STask::Client::Mount->create(
        command_interface => $client,
        share             => $cifs_share_C_object,
        nacltask_to_cleanup      => 1,          # default 0,
        nacltask_cleanup_manager => $CleanupObj,
        %other_options
    );

    3. For Unix mount, user can specify the path option,
    which represents directory path on the server that has
    to be mounted. Similarly, for CIFS share mapping, share
    name can be specified.

    For Unix Mount:
    ---------------
    my $mount = NACL::STask::Client::Mount->create(
        command_interface => $client,
        path              => $path,
        server            => $server_ip,
        #server          => [$server_ip1,$server_ip2], also allowed 
        nacltask_to_cleanup      => 1,          # default 0,
        nacltask_cleanup_manager => $CleanupObj,
        %other_options
    );

    For CIFS share mapping:
    -----------------------
    my $mount = NACL::STask::Client::Mount->create(
        command_interface => $client,
        share             => $name,
        server            => $server_ip, #can be a list for ex [$server_ip1,$server_ip2], 
        nacltask_to_cleanup      => 1,          # default 0,
        nacltask_cleanup_manager => $CleanupObj,
        %other_options
    );

    4. This method also accepts server_command_interface as
    input, which represents the host that is used to run the
    server side commands.

    For Unix Mount:
    ---------------
    my $mount = NACL::STask::Client::Mount->create(
        command_interface        => $client,
        server_command_interface => $ci
        path                     => $path,
        vserver                  => $vserver_name, # for CMode
        nacltask_to_cleanup      => 1,          # default 0,
        nacltask_cleanup_manager => $CleanupObj,
        %other_options
    );

    For CIFS share mapping:
    -----------------------
    my $mount = NACL::STask::Client::Mount->create(
        command_interface        => $client,
        server_command_interface => $ci
        share                    => $share_name,
        vserver                  => $vserver_name,  # for CMode
        nacltask_to_cleanup      => 1,          # default 0,
        nacltask_cleanup_manager => $CleanupObj,
        %other_options
    );

    For SAMBA Mount:
    -------------------
    my $mount = NACL::STask::Client::Mount->create(
        command_interface        => $client,
        server_command_interface => $ci
        volume                   => $volume,
        'source-path'            => $source_path,
        'mount-point'            => $mount_point,
        nacltask_to_cleanup      => 1,          # default 0,
        nacltask_cleanup_manager => $CleanupObj,
        %other_options
    );

(Class method)
Creates a mount point, mounts a volume or qtree at that mount point and returns
an STask Mount object. Performs an NFS mount for Linux clients and a CIFS
mount for Windows clients.

=over

=item Options

=over

=item  C<< command_interface => $ci >>

(Required) This is component object representing the client on which the mount
point has to be created.

=item  C<< server => $ip | \@ips >>

(Optional) The ip or server name of the file server. Use array reference
if multiple entries are provided.
If this is not set, then the C<server_command_interface> option is
required.  For 7-mode, the default will be the C<default_ip> of the
C<hostrec> of that C<server_command_interface> and no commands will
need to be run.  For C-mode, the C<vserver> option is also then
required, and the default will be all of the allowed IP addresses for
the vserver (found by sending commands to C<server_command_interface>)
where each address will be tried before giving up, and where the
"allowed" addresses are those which are enabled for the protocol being
mounted (NFS, CIFS) and which are not disallowed by any
C<select_interface> and C<nacltask_address_family> option settings.

=item C<< path => $path >>

(Optional) This parameter represents a directory path on the server that has
to be mounted. Required if none of volume, qtree, share are provided. We must
also provide enough information to identify the path (vserver is required for
CMode if server is not passed in). For a Windows client, this parameter will
be used to figure out a share that corresponds to that path

=item C<< vers => 4.1|4|3 >>

(Optional) This parameter represents NFS version.
May be used to test specific NFS version.
When this parameter is specified mounting is tried specifically using this version
only. If this version is not supported, mount command will error out.
When this option is not supplied, first default version is tried & on failure,
subsequent versions from latest to oldest are tried till successful operation.

=item C<< volume => $vol >>

(Optional) Not required if 'path' (for Linux) or 'share_name' (for Windows)
is given. A L<Volume component object|lib-NACL-C-Volume-pm> or
L<Flexclone Component/STask object|lib-NACL-C-VolumeClone-pm> or
volume name of the volume that has to be mounted. If a volume/flexclone object
is given the necessary details are picked up from it. Otherwise, individual
details such as vserver name and server_command_interface (required to
uniquely identify the volume) have to be provided.

=item C<< qtree => $qtree >>

(Optional) Not required if 'path' (for Linux) or 'share_name' (for Windows)
is given. A L<Qtree component object|lib-NACL-C-Qtree-pm> or qtree name of
the qtree that has to be mounted. If an object is given the necessary details
are picked up from it. Otherwise, individual details such as volume name,
vserver name and server_command_interface (required to uniquely identify the
qtree) have to be provided.

=item C<< share => $share_name >>

(Optional) The share name on the server which has to be mapped.

=item C<< protocol => $protocol >>

(Optional) The protocol used for mount.Value need to be set to 'nfsv3win'
for mounting volume on windows using nfs. 

=item C<< server_command_interface => $ci >>

(Optional) A L<command interface object|lib-NACL-C-CommandInterface-pm>
which represents the server (ONTAP node or cluster) to be mounted.  It
is required to determine defaults for the C<server> and C<path> options,
and isn't required if both of those options are set.

=item C<< mount_point => $mount_point >>

(Optional) A mount point to use. Providing this is not recommended since
this value is not OS agnostic. For Windows, it is a drive letter, and
for Unix it is a directory name.

=item C<< select_interface => $local >>

(Optional) This option can be used to limit which IP addresses will be
included when determining a default for the C<server> option for a
C-mode server:

=item C<< options => $string >>

(Optional) Used for SAMBA mount. It contains comma-separated list of mount options.

=item C<< 'file-system-type' => $string  >>

Used for SAMBA mount. This option limit the set of filesystem types.
Specifying the filesystem, for eg, cifs

=item C<< 'source-path' => $string >>

(Required) Used for SAMBA mount. It is the source which contains the share to be mounted.

=item C<< 'mount-point'  => $string >>

(Required) Used for SAMBA mount. It is the destination where the share has to be mounted.

=over

=item *

If set to C<local>, will only include addresses that are on the same
node as the volume (addresses of interfaces whose C<curr-node> is the
node of the C<volume> to be mounted)

=item *

If set to C<remote>, will only include addresses that are on a
different node than the volume (addresses of interfaces whose
C<curr-node> is not the node of the C<volume> to be mounted)

=item *

(if not set, there is no such limit)

=back

=item  C<< nacltask_address_family => $address_family >>

(Optional) This option can be used to limit which IP addresses will be
included when determining a default for the C<server> option for a
C-mode server:

=over

=item *

If set to C<ipv4>, will include only addresses of network interfaces
whose C<address-family> is "ipv4".

=item *

If set to C<ipv6>, will include only addresses of network interfaces
whose C<address-family> is "ipv6".

=item *

(if not set, there is no such limit)

=back

=item C<< _was_created => \$scalar >>

(Optional) When this option is provided a reference to a scalar variable,
the variable gets filled in with a boolean value describing whether the
mount point was found (value will be 0; this scenario is possible when
nacltask_if_exists => "reuse") or whether the mount point was created
(value will be 1). This is necessary to determine whether the mount point
needs to be cleaned up later.

=item C<< nacltask_verify >>

When the mount_point name is not specified in the list of arguments passed to
this method, verification is done based on whatever mount_point is present in
the mount object after creation.

=over

=item *

If set to C<0> (default), the verification step is skipped.

=item *

If set to C<1>, mount point creation will be verified.

=back

=item  C<< nacltask_if_exists => $action >>

(Optional) Specifies an action to be taken if the share already exists.
This value will be ignored if the mount_point is not present in the list of
arguments passed to this method since we will automatically generate a mount_point
name in that case.

=item C<< nacltask_to_cleanup => 0|1 >>

(Optional, default to 0(no to cleanup)
Flag indicating if this operation needs to be registered for clean up or not.

=item C<< nacltask_cleanup_manager >>

Cleanup manager to be used for registering.
Default : Will use the default cleanup manager.

=over

=item *

if set to C<die> (default), an exception is raised.

=item *

if set to C<reuse>, no action is performed

=item *

if set to C<purge>, deletes the existing mount point before creating it.

=item C<< extra_diagnosis_options >>

Its a HASHREF which enables the api to perform diagnosis in the case of
mount failure.
The diagnosis tool offers the following capabilities : 
1. Handle if cifs/nfs mapping already exist. 
2. Enable the capture on the interface which is being mounted.
3. Test the share for R/W capabilities. 

the options are provided in the following manner : 
NACL::STask::Client::Mount->create(
   command_interface => $ci,
   server => [$server1, $server2],
   extra_diagnosis_options => {
                    enable_capture => 1,
                    share_tests => "rw",
                    handle_if_map_exists => 1
                 }
);
If enable_capture is provided as value 1 then packet capture will be enabled using :
NACL::MTask::PacketCapture->new(packet_capture_details => $details);
Here user can also provide the customized capture filter(For reference to filter please 
refer to NACL::MTask::PacketCapture).

So the call now will look like as : 
   NACL::Stask::Client::Mount->create(
   command_interface => $ci,
   server => [$server1, $server2],
   extra_diagnosis_options => {
           enable_capture => 1,
           capture_filter => $capture_filter, 
           share_tests => "rw",
           handle_if_map_exists => 1,
           enable_all_tools => 1
         }
   );

if share_tests is provided as value 1 then read / write test will be performed on the share.
Default is 'rw'. 
The share_test will also be performed to newly created mount if any of the diagnose_option is defined & nacltask_verify is set to 1.

if handle_if_map_exists is enabled then following steps are performed.
unmap the share in case if :
## 1. mapping exists with the specified 'drive letter'/'mount point' on a different share path.
## 2. mapping exist with the same share path but status is unavailable.
## 3. A different drive is mounted to the share path than the one provided.
## 4. nacltask_if_exists is given as "purge" & mapping exists with the same path 
##     but with different drive.
## 5. nacltask_if_exists is given as "purge" & mapping exists with the same path 
##     but share has driveless mapping.
    
Reuse the existing mapping if share path matches and :
## 1. if 'reuse' is specified & mount_point is '*' & existing drive is some drive letter
## 2. if 'reuse' is specified & mount_point is '*' & existing drive is null
## 3. if 'reuse' is specified & mount_point is not '*'.

Apart from the above steps it also performs the 'rw' test on the share if 'reuse' & 'share_tests' both are specified .

if enable_all_tools is set to 1 then all tools (capture,share_test,handle_if_map_exists) will be enabled & run by default.

=item C<< mount-prefix => $string >>

(Optional) If specified, mount-prefix string will be concatenated with mount command.
This allows user to run variety of troubleshooting tools, such as strace.

=back

All of the other parameters command_interface, mount_point etc (except C<server>
and C<path>) accepted by
L<NACL::C::Client::Mount::create|lib-NACL-C-Client-Mount-pm/create> are accepted
by this method.

=back

=back

=cut

sub create {
    $Log->enter() if $may_enter;
    my $pkg = shift;

    my %opts = $pkg->_common_validate_with(
        params          => \@_,
        additional_spec => {
            server => { type => SCALAR|ARRAYREF|UNDEF, optional => 1 },
            nacltask_additional_path => { type => SCALAR, default => "" },
            nacltask_if_exists => $pkg->_if_exists_validate_spec,
            _was_created       => { type => SCALARREF, optional => 1 },
            nacltask_verify    => { type => SCALAR, default => 0 },
            $pkg->_cleanup_validate_spec(),
            nacltask_address_family =>
                { regex => qr/^(ipv4|ipv6)$/, optional => 1 },
            select_interface => {
                type      => SCALAR,
                optional  => 1,
                callbacks => {
                    "Value of 'select_interface' should be 'local' or 'remote'"
                        => sub {
                        defined $_[0] && $_[0] =~ /^(local)|(remote)$/;
                        }
                }
            },
            server_command_interface =>
                { isa => 'NACL::C::CommandInterface::ONTAP', optional => 1 },
            qtree => {type => SCALAR|OBJECT, optional => 1 },
            share => {type => SCALAR, optional => 1 },
            protocol => { type => SCALAR, optional => 1 },
            extra_diagnosis_options => { type => HASHREF, optional => 1 },
            'mount-prefix' => {type => SCALAR, optional => 1 },
        },
        allow_extra         => 1,
        ignore_primary_keys => 1
    );

    my $diagnosis_opts = delete $opts{extra_diagnosis_options} ;
    my %orig_opts = %opts;
    my $client = $opts{command_interface};
    my $select_interface = delete $opts{select_interface};
    my (%opts_for_cleanup, %opts_for_register, $nacltask_to_cleanup );
        $pkg->_move_common_cleanup_opts(
          source  => \%opts,
          target  => \%opts_for_cleanup,
    );

    my $t = {};
    $pkg->_move_nacltask_options(
        source => \%opts,
        target => $t
    );
    my $sleep = $t->{nacltask_retry_sleep_time} || 60;
    my $was_created = delete $opts{_was_created};
    $$was_created = 0;

    my ( $path, $ci );

    # by the end of this we should have a path

    # The user may pass volume name / object or qtree name or object
    # At the end of this step, we should have the primary keys of the
    # volume / qtree object

    # If a flexclone object is passed, let's cast it to a volume object first
    my $vol = $opts{volume};
    if ( defined($vol) && blessed($vol) && $vol->isa('NACL::C::VolumeClone') )
    {
        $opts{volume} = $opts{volume}->cast_to_volume();
    }

    _populate_keys(
        hash => \%opts,
        key  => 'volume',
        isa  => 'NACL::C::Volume'
    );
    _populate_keys(
        hash => \%opts,
        key  => 'qtree',
        isa  => 'NACL::C::Qtree'
    );

    $ci = $opts{server_command_interface};
    my $volume = $opts{volume};
    my $node;

    my $qtree;
    if (defined $opts{'qtree-path'}) {
        $qtree = (split( /\//, $opts{'qtree-path'} ))[-1] || "";
        delete $opts{'qtree-path'};
    } else {
        $qtree = delete $opts{qtree} || "";
    }

   # handling a special case in CMode
   # the default qtree is named '""' (a string that contains two double quotes
   # as its contents)
    $qtree = "" if $qtree eq '""';
   # if path is not provided, only then construct one from volume, qtree etc.
    my $protocol = $opts{protocol} // '';
    if ( !$opts{path} && (!$opts{share} || $protocol eq 'nfsv3win') ) {
        if ( !$ci ) {
            NATE::Exceptions::Argument->throw( 'The server_command_interface '
                    . 'parameter is required since '
                    . 'the none of qtree, volume '
                    . 'object were passed in' );
        } ## end if ( !$ci )
        $path = NACL::CS::Volume->fetch(
            command_interface => $ci,
            requested_fields  => ['junction-path'],
            filter => NACL::C::Volume->get_primary_keys_options(%opts),
            %{ $pkg->_copy_common_component_params( source => \%opts ) }
        )->junction_path();

        # if its a qtree, then append the qtree name to the path
        $path .= '/' . $qtree;
        $path .= '/' . $t->{nacltask_additional_path};
        $path =~ s/\/\/+/\//;
        $path =~ s/\/$// if $path ne '/';
        # add path to %opts
        $opts{path} = $path;
    } ## end if ( !$opts{path} && !...)
    delete $opts{volume};
    my $client_type = $client->type();
    if ( $client_type =~ /windows/i || defined $opts{share}) {
        if ( $protocol !~ /nfsv3win/ ) {

            # get a share name from the path
            if ( $opts{path} && !$opts{share} ) {

                # if path is present, then we need to find a cifs share on that
                # path with appropriate permissions
                _get_share_from_path( \%opts );
            } ## end if ( $opts{path} && !$opts...)

            # we are supporting both a share object and share name using the same
            # key (share). First we'll rename share to share-name.
            # This is the key that will be filled by _populate_keys should an
            # object be passed
            $opts{'share-name'} = delete $opts{share};

            # The user may pass in a CIFS share object or just the share name
            # At the end of this step we should have all the primary keys of the
            # share
            _populate_keys(
                hash => \%opts,
                key  => 'share-name',
                isa  => 'NACL::C::VserverCifsShare'
            );

            # the component accepts the argument as share_name
            # So before making the component call, we'll rename share-name to share_name
            $opts{share_name} = delete $opts{'share-name'};
        }
     } elsif ( !$client->hostrec->is_unix() ) {
        NATE::Exceptions::Argument->throw(
            'Host type: ' . $client_type . ' not supported' );
    }
    $ci = $opts{server_command_interface};
    my $vserver = delete $opts{vserver};

    # if server is not specified in %opts, we have to figure it out from the
    # vserver / hostrec's default ip
    # at the end of this if {}, we should have a server ip

    if ( !$ci && !$opts{server} ) {
        $Log->exit() if $may_exit;
        NATE::Exceptions::Argument->throw( 'The server_command_interface '
                . 'parameter is required when the '
                . '"server" parameter is not provided' );
    } ## end if ( !$ci && !$opts{server...})

    my @interface_cs;

    ## check if diagnose_opts is defined
    if(defined $diagnosis_opts && $diagnosis_opts->{enable_all_tools}){
        $diagnosis_opts->{handle_if_map_exists} = 1;
        $diagnosis_opts->{enable_capture} = 1;
        $diagnosis_opts->{test_share} = 'rw';
    }
    my $handle_map_param;
    if(defined $diagnosis_opts && $diagnosis_opts->{handle_if_map_exists}){
       $handle_map_param = 1 ;
    }

   my $other_opts = {
	     vserver => $vserver,
	     select_interface => $select_interface,
	     volume => $volume,
	     nacl_task_opts => $t
    };

    $Log->comment('Populate the servers');
    my $servers = [];
    $servers = $pkg->_populate_servers(\%opts, $other_opts );
    $Log->comment("Servers obtained are: ".Dumper(@$servers));

    ## handle if map is already shared
    if ( ($client_type =~ /windows/i || defined $opts{share}) && $protocol !~ /nfsv3win/) {

	 my $self =
              $pkg->_handle_if_map_exists(%opts,
                  nacltask_options => $t,
                  diagnosis_param => $diagnosis_opts
              );
         ## check if self is defined
         if($self){
            $$was_created = 1;
            $opts{mount_point} = $self->mount_point() if(!$opts{mount_point});
            
            # Register this resource with the cleanup manager for cleanup.
            $pkg->_copy_common_opts_for_cleanup(
               'source'             => {%opts, %opts_for_cleanup},
               'target'             => \%opts_for_register,
               'nacltask_to_cleanup'=> \$nacltask_to_cleanup,
               'to_cleanup'         => 'purge',
            );

            $pkg->_register_for_cleanup(%opts_for_register)
               if ($nacltask_to_cleanup);
            $self->_create_opts(%orig_opts);
            $Log->exit() if $may_exit;
            return $self ;
         }else{
            $Log->comment('Continuing with Mount');
         }
    }

    # The nacltask_if_exists value will be ignored if the mount_point
    # (primary key for this Component / STask) is not provided in the
    # list of arguments provided to this ( create() ) method call.
    # This is because for this task, we automatically generate a mount
    # point name if it is not provided
    if ( !exists $opts{mount_point} ) {
        $t->{nacltask_if_exists} = "die";
    }

    my $self;
    my @all_exceptions;
    my %create_opts;
    my $mount = sub {
        my %extra_opts = @_;
        %create_opts = ();
        $pkg->_hash_copy(
            source => \%opts,
            target => \%create_opts,
            copy =>
                [qw(share_name mount_point server path command_interface retries)],
        );

        # now %opts has only those params which are required for creation

        ## check if handle map already executed
        if(!$handle_map_param){
            $self = $pkg->_handle_if_exists_action(
                if_exists   => $t->{nacltask_if_exists},
                create_opts => \%create_opts
            );
        }
        my $nate_params = NATE::ParamSet->new(global => 1);

        ### The MTU_SIZE parameter defined from ntest command-line is to override the Mount default values and mount with user-specified MTU values.

        my $mtu_size = $nate_params->get('MTU_SIZE');
        my $mount_opts = "";
        if(defined $mtu_size) {
            require NACL::C::UnitNormalization;
            $mtu_size = NACL::C::UnitNormalization::convert_to_bytes($mtu_size);
            $mount_opts = delete $opts{'mountopts'};
            if ( defined $mount_opts ){
                   if ( $mount_opts !~ /rsize/ && $mount_opts !~ /wsize/) {
                        $mount_opts = $mount_opts . ",rsize=". $mtu_size . ",wsize=". $mtu_size;
                        $opts{'mountopts'} = $mount_opts;
                   }
                   else {
                        $opts{'mountopts'} = $mount_opts;
                   }
            }
            else {
                  $opts{'mountopts'} = "rsize=". $mtu_size . ",wsize=". $mtu_size;
           }
        }
        if ( !$self ) {
            $Log->debug( 'The options beings sent to create are: ' .Dumper( %extra_opts ));
            $self         = $pkg->SUPER::create(%opts,%extra_opts);
            $$was_created = 1;
        }
    };

    my $error = 0;
    # if vers is passed by user, other nfs version will not be tried.
    my $vers_defined = defined $opts{vers} ? 1:0;

    # $opt{vers} is used in defining mountpoint.
    # for example if $opt{vers}=4 mountpoint will be /t/10.53.6.47/root-4
    # if $opt{vers}=4.1 mountpoint will be /t/10.53.6.47/root-4.1
    # if vers=4 and mountopt is passed as "minorversion=1" it is desired that
    # mountpoint should be /t/10.53.6.47/root-4.1 & not /t/10.53.6.47/root-4
    # passing vers=4.1 solves this problem.
    if ($vers_defined) {
        if (defined $opts{mountopts} && ($opts{vers} !~ /\./)) {
            if ($opts{mountopts} =~ /minorversion\s*=\s*(\d)/) {
                $opts{vers}.=".$1";
            }
        }
    }

    # following variable is to track of nfs versions tried.
    my %nfs_vers_tried = map {$_=>0} @nfs_versions;

    my @nfs_ver_error_strings = (
    "RPC Error: Program not registered",
    "Command exited with status 32",
    "unknown nfs mount parameter: minorversion",
    "mount.nfs4: Protocol not supported"
    );
    ## diagnosis will be performed only in case of failures
    if(defined $diagnosis_opts) {
      $diagnosis_opts->{enable_diagnosis} = 0 ;
      $diagnosis_opts->{diagnosis_count} = 0 ;
    }

    my %retry_c_opts = ();
    my $max_retries = 0;
    my $current_retry = 1;
    ## NACL::C::Client::Mount->create() default value of retries is '2'
    ## calculating the appropriate value of retry count
    if($client_type =~ /windows|unix/i){
       if(defined $t->{nacltask_retry_tries_count} || defined $opts{retries}){
         $max_retries = $t->{nacltask_retry_tries_count} ;
         $max_retries //= $opts{retries} ;
       }elsif(defined $diagnosis_opts && defined $diagnosis_opts->{enable_capture}){
         $max_retries = 2;
       }else{
         $max_retries = 1;
       }
       %retry_c_opts = ('retries' => $max_retries);
    }

    AGAIN: {
        ## start the capture on the current server
        ## before mount
        $pkg->_capture_start(
            command_interface => $client,
            server => $servers,
            diagnosis_param => $diagnosis_opts
        );

        foreach my $curr_server (@$servers) {
            if (defined $protocol && $protocol eq 'nfsv3win') {
                $opts{server} = $curr_server. "\\". $volume  ;
            } else {
                $opts{server} = $curr_server;
            }
            try {
                $mount->(%retry_c_opts);
            }
            catch NATE::BaseException with {
                my $e = shift;
                if ( $e->text() =~ /Bad file descriptor/) {
                    $client->refresh_command_interface();
                    throw NACL::APISet::Exceptions::TimeoutException->new
                        ("Timed out attempting to mount:\n$e->text()");
                }
                $Log->debug(
                    'Exception found : ' . ref $e . ' : ' . $e->text );
                push @all_exceptions, $e;
                $error = 1;

            };
            last if ($self);
        }
        ## end the capture after the mount is created
        $pkg->_capture_stop(
             command_interface => $client,
             diagnosis_param   => $diagnosis_opts
        );
    }

    # if user has supplied vers, do not try other nfs versions
    if (!$vers_defined && $error && !$self) {
        my $latest_error_text = $all_exceptions[-1]->text;
        my $combined_search = join("|",@nfs_ver_error_strings);
        if ($latest_error_text =~ /($combined_search)/ ) {
            foreach my $nfs_ver (@nfs_versions) {
                # make sure if $nfs_ver is already tried.
                if (!$nfs_vers_tried{$nfs_ver}) {
                    #reset error
                    $error = 0;
                    #so that same version is not tried again
                    $nfs_vers_tried{$nfs_ver}++;
                    $opts{vers} = $nfs_ver;
                    $Log->debug('Trying again with NFS version $nfs_ver');
                    goto AGAIN;
                }
            }
            $Log->debug('All NFS versions tried. Mount failed.');
        }
    }

    if ( !$self ) {
        my $e = _combine_exceptions(@all_exceptions);

        ## end capture
        $pkg->_capture_stop(
           command_interface => $client,
           diagnosis_param  =>  $diagnosis_opts
        ) if(defined $diagnosis_opts);

        if(++$current_retry <= $max_retries){
           $diagnosis_opts->{max_retry} = $max_retries;
           $diagnosis_opts->{current_retry} = $current_retry;
           $pkg->_set_diagnose(diagnosis_param  =>  $diagnosis_opts);
           $Log->comment('Sleep b/w retry for '.$sleep.' seconds');
           Tharn::snooze($sleep);
           $Log->comment("Start Retry $current_retry of Mount");
           goto AGAIN;
        }
        $Log->exit() if $may_exit;
        $e->throw();

    }

    if ($$was_created) {

        # If mount point does not exist in the list of arguments provided
        # to this method call, then verify from the mount_point value
        # stored in whatever is returned by $pkg->SUPER::create( )

        if ( !exists $opts{mount_point} || $opts{mount_point} eq "*" ) {
            $create_opts{mount_point} = $self->mount_point();
        }

        # For IPv6 addresses, nfs mount point has extra braces in
        # server IP. Ex: [fd20:8b1e:b255:8305:3170:190a:d100:0000]:/vol_dvsN1
        # So in _generic_create_modify_verify method, the param 'server'
        # needs to be provided as, server => "[" . $server_ip. "]" for
        # Linux/Solaris. For Windows, convert the IPv6 address to Microsofts
        # .ip6-literal.net format.


        if ( $t->{nacltask_verify} ) {
            if ( Net::IP::ip_is_ipv6( $create_opts{server} ) ) {
                $create_opts{server} = convert_ipv6_address_format(
                    ipv6_address => $create_opts{server},
                    client_type  => $client_type
                );
            } ## end if ( Net::IP::ip_is_ipv6...)

            # Remove the single quotes before passing the create options to
            #_generic_create_modify_verify routine
            $create_opts{server} =~ s/^'(.*)'$/$1/;
            try {
                $Log->comment('Verifying the created Mount');
                my $no_of_retries = delete $create_opts{retries};
                $pkg->_generic_create_modify_verify(%create_opts);
                $create_opts{retries} = $no_of_retries;
                if(defined $diagnosis_opts){
                  $Log->comment('Performing read/write test');
                  $pkg->_test_share(%create_opts,
                     diagnosis_param => $diagnosis_opts,
                     protocol => $protocol 
                  );
                }
            } catch NACL::Exceptions::VerifyFailure with {
                $create_opts{server} = $1 if ( $create_opts{server} =~ /\[(.*)\]/ ) ;
                $pkg->_generic_create_modify_verify(%create_opts);
            };
        } ## end if ( $t->{nacltask_verify...})
    } ## end if ($$was_created)

    $opts{mount_point} = $self->mount_point() if(!$opts{mount_point});

    # Register this resource with the cleanup manager for cleanup.
    $pkg->_copy_common_opts_for_cleanup(
        'source'             => {%opts, %opts_for_cleanup},
        'target'             => \%opts_for_register,
        'nacltask_to_cleanup'=> \$nacltask_to_cleanup,
        'to_cleanup'         => 'purge',
    );

    $pkg->_register_for_cleanup(%opts_for_register)
          if ($nacltask_to_cleanup);

    $self->_create_opts(%orig_opts);

    $Log->exit() if $may_exit;
    return $self;
} ## end sub create

sub _set_diagnose{
    my ($pkg, %opts) = @_;
    
    my $diagnosis_opts = $opts{diagnosis_param};
    if($diagnosis_opts->{capture_on_retry}){
       if($diagnosis_opts->{capture_on_retry} == $diagnosis_opts->{current_retry}){
         $Log->comment('Retrying with Capture Enabled');
         $diagnosis_opts->{enable_diagnosis} = 1;
       }else{
         $diagnosis_opts->{enable_diagnosis} = 0;
       }
    }elsif($diagnosis_opts->{enable_capture} && !$diagnosis_opts->{capture_on_retry}){
       if($diagnosis_opts->{current_retry} == $diagnosis_opts->{max_retry} - 1){
         $Log->comment('Retrying with Capture Enabled');
         $diagnosis_opts->{enable_diagnosis} = 1 ;
       }
    }
}

=head2 _handle_if_map_exists

This private subroutine performs the following steps:.

unmap the share in case if :
## 1. mapping exists with the specified drive letter on a different share path.
## 2. mapping exist with the same share path but status is unavailable.
## 3. A different drive is mounted to the share path than the one provided.
## 4. nacltask_if_exists is given as "purge" & mapping exists with the same path
##     but with different drive.
## 5. nacltask_if_exists is given as "purge" & mapping exists with the same path
##     but share has driveless mapping.

Reuse the existing mapping if share path matches & :
## 1. if 'reuse' is specified & mount_point is '*' & existing drive is some drive letter
## 2. if 'reuse' is specified & mount_point is '*' & existing drive is null
## 3. if 'reuse' is specified & mount_point is not '*'.

Apart from the above steps it also performs the 'rw' test on the share if 'reuse' & 'share_tests' both are specified .

=cut

sub _handle_if_map_exists {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;

    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@_,
        additional_spec => {
            share_name            => {type => SCALAR},
            mount_point           => {type => SCALAR, optional => 1},
            server               => {type => SCALAR|ARRAYREF},
            nacltask_options     => {type => HASHREF,
                                     default => {nacltask_if_exists => "reuse"}}
        },
        allow_extra => 1,
    );

    my $command_interface  = $opts{command_interface};
    my $apiset             = $command_interface->apiset();
    my $nacltask_opts      = delete $opts{nacltask_options};
    my $nacltask_if_exists = $nacltask_opts->{nacltask_if_exists} ;
    my $mount_point        = $opts{mount_point} ;
    my $share              = $opts{share_name};
    my $driveless_mapping  = delete $opts{driveless_mapping};

    my $diagnosis_opts    = $opts{diagnosis_param};

    if(!defined $diagnosis_opts || 
          (defined $diagnosis_opts && !$diagnosis_opts->{handle_if_map_exists})){
       $Log->exit() if $may_exit;
       return
    }
    my %retry_opts ;
    $pkg_or_obj->_copy_retry_params(
        source => $nacltask_opts,
        target => \%retry_opts
    );
    ## generate probable share path to compare with existing mapped path
    my @actual_path = ();
    my $path ;
    if(defined $opts{path}){
       $path = $opts{path};
       push @actual_path, $opts{path} ;
    }else{
       if(ref $opts{server}){
          foreach my $server (@{$opts{server}}){
            $path = '\\\\'.$server.'\\'.$share ;
            push @actual_path, $path;
          }
       }else{
          $path = '\\\\'.$opts{server}.'\\'.$share ;
          push @actual_path, $path;
       }
    }

    my $filter = {};
    $filter = {filter => {mount_point => $mount_point} }   if(defined $mount_point) ;

    $Log->comment('Check if map exists');
    $Log->comment('Listing the current mappings on client: '.$command_interface->name());
    my @mount_fetch = NACL::CS::Client::Mount->fetch(
	command_interface => $command_interface,
        %{$filter},
	requested_fields  => [qw(mount_point share_name server status )],
	allow_empty       => 1
    );

    if(!$mount_point && !$driveless_mapping){
        $mount_point = "*";
    }elsif(!$mount_point && $driveless_mapping){
        $mount_point = "";
    }

    my $check_ping_to_server = 0;

    my %test_share_opts = (
        command_interface => $command_interface,
        server  => $opts{server},
        diagnosis_param => $diagnosis_opts
    );

    my $reused_obj_ref = {};
    my $reused_obj ;

    ## The block below performs the following check :
    ## unmap the share in case if :
    ## 1. mapping exists with the specified drive letter on a different share path.
    ## 2. mapping exist with the same share path but status is unavailable.
    ## 3. A different drive is mounted to the share path than the one provided.
    ## 4. nacltask_if_exists is given as "purge" & mapping exists with the same path 
    ##     but with different drive.
    ## 5. nacltask_if_exists is given as "purge" & mapping exists with the same path 
    ##     but share has driveless mapping.
    
    ## Reuse the existing mapping if share path matches & :
    ## 1. if 'reuse' is specified & mount_point is '*' & existing drive is some drive letter
    ## 2. if 'reuse' is specified & mount_point is '*' & existing drive is null
    ## 3. if 'reuse' is specified & mount_point is not '*'.

    foreach my $map (@mount_fetch){
      $Log->debug("\n"
                . "Share  . . . . : ".$map->share_name
                . "Requested Share: $share \n"
                . "Drive  . . . . : ".$map->drive()
                . "Requested Drive: $mount_point");

      my $path = $map->path();
      $path =~ s/(\\|\.)/\\$1/g;

      if (   $mount_point =~ /\w/
            && $map->drive() =~ /$mount_point/i
            && !grep (/$path/, @actual_path )) {
            $Log->comment("Existing mapping of ".$map->path()." is using the "
                    . "specified drive letter ($mount_point), removing");
            $pkg_or_obj->purge(
                 command_interface => $command_interface,
                 mount_point   => $map->drive(),
                 nacltask_verify => 1,
                 nacltask_if_purged => 'pass'
            );
            $check_ping_to_server =1 ;
        } elsif (grep (/$path/, @actual_path )) {
            if ($map->status() =~ /Unavailable|Disconnected/) {
                $Log->comment(
                    "Existing mapping of ".$map->path()." is unavailable,
                          removing $mount_point"
                );
                $pkg_or_obj->purge(
                    command_interface => $command_interface,
                    mount_point   => $mount_point,
                    nacltask_verify => 1,
                    nacltask_if_purged => 'pass'
                );
                $check_ping_to_server = 1;
            } elsif ($mount_point =~ /\w/
                && $map->drive() !~ /$mount_point/i) {
                $Log->comment(
                    "Existing mapping of ".$map->path()."does not have the "
                        . "correct drive letter ($mount_point), removing ".$map->drive());
                $pkg_or_obj->purge(
                     command_interface => $command_interface,
                     mount_point   => $map->drive(),
                     nacltask_verify => 1,
                     nacltask_if_purged => 'pass'
                 );
                 $check_ping_to_server = 1;
            } elsif ($nacltask_if_exists eq "reuse"
                && $mount_point   =~ /\*/
                && $map->drive() =~ /\w/i) {
                $Log->comment(
                    "Reusing existing mapping ".$map->drive()." , ".$map->path()
                );
                $Log->comment('Performing read/write test'); 
                $pkg_or_obj->_test_share(
                     path => $map->path(),
                     %test_share_opts
                );
                $reused_obj = 1;
            } elsif ($nacltask_if_exists eq "reuse" && $mount_point !~ /\*/) {
                $Log->comment(
                    "Reusing existing driveless mapping on path: ".$map->path);
                $Log->comment('Performing read/write test');
                $pkg_or_obj->_test_share(
                     path => $map->path,
                     %test_share_opts
                );
                $reused_obj = 1;
            } elsif( $nacltask_if_exists eq "reuse"
                && $mount_point =~ /\*/
                && !$map->drive()) {
                $Log->comment(
                    "Forcing creation of new drive mapping to ".$map->path()
                );
                last;
            } elsif ($nacltask_if_exists eq "purge"
                && $map->drive()
                && $mount_point =~ /\w/) {
                $Log->comment(
                    "Removing existing mapping ".$map->drive().", ".$map->path()
                );
                $pkg_or_obj->purge(
                     command_interface => $command_interface,
                     mount_point   => $map->drive(),
                     nacltask_verify => 1,
                     nacltask_if_purged => 'pass'
                );
                $check_ping_to_server = 1;
            } elsif ($nacltask_if_exists eq "purge" && !$map->drive()) {
                $Log->comment(
                    "Removing existing driveless mapping ".$map->path());
                $pkg_or_obj->purge(
                     command_interface => $command_interface,
                     mount_point   => $map->mount_point(),
                     nacltask_if_purged => 'pass'
                 );
                 $check_ping_to_server = 1;
            }
        }
        if($reused_obj){
            $pkg_or_obj->_hash_copy(
            source => $map,
            target => $reused_obj_ref,
            copy =>
                [qw(mount_point command_interface)],
            );
            last;
        }

    }

    if($reused_obj){
        $Log->debug('The reused obj ref found is:'.Dumper($reused_obj_ref));
        $Log->exit() if $may_exit;
        return $pkg_or_obj->new(
           %{$reused_obj_ref}
        );
    }

    # Check for ping to servers
    my @ping_exceptions ;
    if($check_ping_to_server){
      $Log->comment('Check for ping to all the servers');
      foreach my $server (@{$opts{server}}){
        try{
	  nacl_method_retry(
            code => sub { $apiset->ping('destination' => $server); },
	    exceptions => ['NACL::APISet::Exceptions::CommandFailedException'],
	    %retry_opts
	  );
        }catch NACL::APISet::Exceptions::CommandFailedException with{
	  my $e = shift;
	  $Log->debug('Exception found while trying to ping : '.$server
                    .' and exception details are '. ref $e . ' : ' . $e->text
          );
	push @ping_exceptions, $e;
        };
      }
    }
    if(@ping_exceptions){
        my $e = _combine_exceptions(@ping_exceptions);
        if(scalar(@ping_exceptions) == scalar(@{$opts{server}})){
            $Log->comment("None of the available servers could be pinged.");
            $Log->exit() if $may_exit;
            $e->throw();
        }else{
            $Log->comment('The following servers couldnt be pinged: '.$e->text());
        }
    }
    $Log->exit() if $may_exit;
} ##end sub _handle_if_map_exists

=head2 purge

    NACL::STask::Client::Mount->purge(
        command_interface => $ci,
        %other_options
    );

    or

    $mount_object->purge( nacltask_verify => 1 );

(Instance and class method) Deletes a mount point.

=over

=item Options

=over

=item C<< command_interface => $cmd_interface >>

(Required only for class method call but not for instance method call)
This is the component object that represents the host to which the
commands are sent.

=item C<< nacltask_verify >>

(Optional) - when set to 1, mount point deletion will be verified.
           - when set to 0, the verification step is skipped.
           - by default its set to 0.

=item C<< retries => $num >>

(Optional, defaults to 3)
The number of times to retry unmounting in case the resource is busy when
attempting to unmount.

=back

All of the other parameters accepted by
L<NACL::C::Client::Mount::delete|lib-NACL-C-Client-Mount-pm/delete> are
accepted by this method.

=back

=cut

sub purge {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;

    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@_,
        additional_spec => {
            nacltask_if_purged => { type => SCALAR, default => "die" },
            nacltask_verify    => { type => SCALAR, default => 0 },
            retries            => { type => SCALAR, default => 3 },
        },
        allow_extra => 1,
    );

    my $is_windows = $opts{command_interface}->is_windows();
    my $command_interface = $opts{command_interface};
    my $mount_point = $opts{mount_point};

    my $t = {};
    $pkg_or_obj->_move_nacltask_options(
        source => \%opts,
        target => $t
    );

    # for anything other than 'die' check if the share exists
    if ( $t->{nacltask_if_purged} ne 'die' ) {

        my $self = $pkg_or_obj->_handle_if_exists_action(
            if_exists   => 'reuse',
            create_opts => \%opts
        );
        # if the mount point doesn't exist, simply return
        return if !$self;
    } ## end if ( $t->{nacltask_if_purged...})

    if ( ref $pkg_or_obj ) {

        # delete the primary keys from %opts since we're going to make an
        # instance call passing %opts
        my %hash = $pkg_or_obj->get_pk_hash_with_ci();
        $pkg_or_obj->_hash_move(
            source => \%opts,
            target => {},
            move   => [ keys %hash ]
        );
    } ## end if ( ref $pkg_or_obj )

    my ( $share_name, $server, %extra_opts  );
    if ( $t->{nacltask_verify} && $is_windows ) {
	my %filter_hash = ( mount_point => $mount_point );
	my $mount_states = NACL::CS::Client::Mount->fetch(
		command_interface => $command_interface,
		filter            => {%filter_hash},
		allow_empty       => 1,
		requested_fields  => [qw(share_name)],
	);
	if( defined $mount_states ){
	    $share_name = $mount_states->share_name();
	    $server     = $mount_states->server();
	}
    }
    my $win_hostip = "";
    my $hostname = "";
    my @active_sessions;
    my $node_cmd_interface = "";
    my $sessionid          = "";
    my $sharedhostip       = "";
    if ($is_windows) {
        $win_hostip = $command_interface->default_ip();
        my $response     = "";
        #Fetching Node IP from the net use mount point
        eval { $response = $command_interface->apiset()->net_use( sharename => $mount_point ); };
	if ( ref $response ) {
            my $output = $response->get_parsed_output();
            my $shareip = $output->[0]{remote} || '';
            if ( $shareip =~ /(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})/ ) {
                $sharedhostip = $1;
            }

            #getting the node name
            if ( defined($sharedhostip) && ( $sharedhostip ne "" ) ) {
                my $mount_hostname = get_hostname($sharedhostip);
                my $name_temp = substr $mount_hostname, 0, index( $mount_hostname, '.' ) || '';
                $hostname = substr( $name_temp, 0, -3 );
            }
            if ( defined($hostname) && ( $hostname ne "" ) ) {
                eval {
                    $node_cmd_interface = NACL::C::Node->new( name => $hostname );

                    #getting active sessions before mount deletion
                    @active_sessions = NACL::CS::VserverCifsSession->fetch(
                    command_interface => $node_cmd_interface,
                    filter            => { address      => $win_hostip,
					  'lif-address' => $sharedhostip },
                    allow_empty       => 1
                );
               };
            }
	    my $noofsessions = @active_sessions;
            if ( ( $noofsessions == 1 ) && ( $active_sessions[0]->{shares} == 1 ) ) {
                $sessionid = $active_sessions[0]->{session_id};
            }
	} else {
	    $Log->comment('There is an issue while fetching the drive details.');
	}
    }
    try {
        $pkg_or_obj->delete( 'force' => 'true', %opts );
    }catch NACL::C::Exceptions::Client::Mount::DoesNotExist with {
        my $exception = shift;
        if ( $t->{nacltask_if_purged} ne 'die' ) {
            $Log->exit() if $may_exit;
            $exception->throw();
        }
    }catch NATE::BaseException with {
	my $exception = shift;
	if ( $exception->text() =~ /Bad file descriptor/i ) {
	    $command_interface->refresh_command_interface();
	    $Log->exit() if $may_exit;
	    NACL::APISet::Exceptions::TimeoutException->throw("Timed out attempting to mount:\n".$exception->text());
	} elsif ( $exception->text() =~ /network connection could not be found/i ) {
	    if ( $t->{nacltask_if_purged} eq 'pass' ) {
		# suppress here, mount already purged
	    }
	} else{
	    $Log->exit() if $may_exit;
	    $exception->throw();
	}
    };
   if ( $is_windows && ( defined($sessionid) && ( $sessionid ne "" ) ) ) {
        if ( defined($node_cmd_interface) && ( $node_cmd_interface ne "" ) ) {
            sleep 5;

            #checking if the session is stale or not
            @active_sessions = NACL::CS::VserverCifsSession->fetch(
                command_interface => $node_cmd_interface,
                filter            => { address       => $win_hostip ,
				       'lif-address' => $sharedhostip },
                allow_empty       => 1
            );
	my $noofsessions = @active_sessions;
            #closing the stale sessions
            if ( ( $noofsessions) && ( $active_sessions[0]->{session_id} eq $sessionid ) ) {
                eval {
                    NACL::C::VserverCifsSession->close(
                        command_interface => $node_cmd_interface,
                        node              => $active_sessions[0]->{node},
                        vserver           => $active_sessions[0]->{vserver},
                        'connection-id'   => $active_sessions[0]->{connection_id},
                        'session-id'      => $active_sessions[0]->{session_id}
                    );
                };
            }
        }
        @active_sessions = NACL::CS::VserverCifsSession->fetch(
            command_interface => $node_cmd_interface,
            filter            => { address => $win_hostip ,
				   'lif-address' => $sharedhostip },
            allow_empty       => 1
        );
        if (@active_sessions) {
            $Log->warn("Unable to close stale sessions.");
        }
    }
    if ( $t->{nacltask_verify} ) {
        my %common_options = ();
        $pkg_or_obj->_copy_common_component_params_with_ci(
            source => \%opts,
            target => \%common_options
        );
        my $pri_keys = $pkg_or_obj->get_primary_keys_options(%opts);
	if ( ( defined $share_name ) && ( defined $server ) ) {
		$extra_opts{share_name} = $share_name;
		$extra_opts{server}     = $server;
	}
	$extra_opts{mount_point}     = $mount_point;
	$extra_opts{command_interface}     = $command_interface;

	$pkg_or_obj->_mount_purge_verify( %$pri_keys, %extra_opts );

    } ## end if ( $t->{nacltask_verify...})
    $Log->exit() if $may_exit;
} ## end sub purge

=head2 remount

    1. Recommended way for Unix remounts:
    my $mount = NACL::STASK::Client::Mount->create(
        command_interface => $client,
        volume            => $vol_C_object,
        %other_options
    );

    $mount->remount();

    2. Recommended way for CIFS share re-mapping:

    my $mount = NACL::STASK::Client::Mount->create(
        command_interface => $client,
        share             => $cifs_share_C_object,
        %other_options
    );

    $mount->remount();

    3. For Unix mount, user can specify the path option,
    which represents directory path on the server that has
    to be mounted. Similarly, for CIFS share mapping, share
    name can be specified.

    For Unix Mount:
    ---------------
    my $mount = NACL::STask::Client::Mount->create(
        command_interface => $client,
        path              => $path,
        server            => $server_ip,
        %other_options
    );

    $mount->remount();

    For CIFS share mapping:
    -----------------------
    my $mount = NACL::STask::Client::Mount->create(
        command_interface => $client,
        share             => $name,
        server            => $server_ip,
        %other_options
    );

    $mount->remount();

    4. This method also accepts server_command_interface as
    input, which represents the host that is used to run the
    server side commands.

    For Unix Mount:
    ---------------
    my $mount = NACL::STask::Client::Mount->create(
        command_interface        => $client,
        server_command_interface => $ci
        path                     => $path,
        vserver                  => $vserver_name, # for CMode
        %other_options
    );

    $mount->remount();

    For CIFS share mapping:
    -----------------------
    my $mount = NACL::STask::Client::Mount->create(
        command_interface        => $client,
        server_command_interface => $ci
        share                    => $share_name,
        vserver                  => $vserver_name,  # for CMode
        %other_options
    );

    $mount->remount();

(Instance method only)
Removes existing mount point and recreates it, mounts a volume or
qtree at that mount point and returns an STask Mount object.
Performs an NFS mount for Linux clients and a CIFS mount for
Windows clients.

=over

=item Options

none

=back

=over

=item Exceptions

=over

=item C<< NACL::STask::Exception::Client::Mount::NotMounted >>

Exception is thrown if specified mount point does not exist
on the client.

=back

All other exceptions thrown by NACL::C::Client::Mount are expected
be thrown.

=back

=cut

sub remount {
    $Log->enter() if $may_enter;
    my $pkg = $_[0];
    $pkg->_verify_invocation( style => 'instance_only' );

    my $create_opts         = $pkg->_create_opts();
    my $ci                  = $create_opts->{command_interface};
    my $mount_point         = $pkg->{mount_point};
    my $mount;

    # Check whether the mount point exists or not.
    $mount = NACL::STask::Client::Mount->find(
        command_interface => $ci,
        filter => {
            mount_point => $mount_point
        },
        allow_empty => 1,
    );
    if (!$mount) {
        my $error = "Specified mount point '$mount_point' is not found";
        $Log->exit() if $may_exit;
        NACL::STask::Exceptions::Client::Mount::NotMounted
            ->throw($error);
    }

    # Remove the existing mount
    NACL::STask::Client::Mount->purge(
        command_interface => $create_opts->{command_interface},
        mount_point       => $mount_point
    );

    # Do the mount operation
    my $mount_create = NACL::STask::Client::Mount->create(
        mount_point => $mount_point,
        %$create_opts,
    );

    $Log->exit() if $may_exit;
    return $mount_create;
} ## end sub remount

# populate the primary keys and server_command_interface for a component
# in the hash reference passed in as 'hash' with command_interface stored in
# server_command_interface

sub _populate_keys {
    my %hash = @_;
    my $opts = $hash{hash};
    my $key  = $hash{key};
    my $isa  = $hash{isa};
    if ( blessed $opts->{$key} && $opts->{$key}->isa($isa) ) {
        my %other_opts = $opts->{$key}->get_primary_keys_hash();
        $opts->{server_command_interface}
            = $opts->{$key}->command_interface();

        # delete the key volume or qtree from the original hash passed to
        # create() since it is going to be filled in from %other_opts
        # with the value as the element's name
        delete $opts->{$key};

        foreach ( keys %other_opts ) {

            # if the primary key is already present in %$opts, then don't
            # overwrite it
            if ( !defined $opts->{$_} ) {
                $opts->{$_} = $other_opts{$_};
            }
        } ## end foreach ( keys %other_opts )
    } ## end if ( blessed $opts->{$key...})
} ## end sub _populate_keys

sub _get_share_from_path {
    my $hash = shift;
    my $ci   = $hash->{server_command_interface};
    if ( !$ci ) {
        NATE::Exceptions::Argument->throw( 'The server_command_interface '
                . 'parameter is required to find a'
                . ' share which has the necessary '
                . 'permissions but it is not present' );
    } ## end if ( !$ci )
    my %filter_hash = ( path => $hash->{path} );
    if ( $hash->{vserver} ) {
        $filter_hash{vserver} = $hash->{vserver};
    }

    my @shares = NACL::CS::VserverCifsShare->fetch(
        filter            => {%filter_hash},
        command_interface => $ci,
        allow_empty       => 1
    );

    my ( $share, $i ) = ( undef, 0 );
    my @conditions = (
        '(Full_Control|Full Control|Read|Change)',
        '(Full_Control|Full Control|Change)',
        '(Full_Control|Full Control)'
    );

    # try to get the share with the least restrictive permissions for everyone
    foreach my $state (@shares) {
        my $condition = $conditions[$i];
        my @str       = $state->acl();
        foreach (@str) {
            if ( $_ =~ /[Ee]veryone \// && $_ =~ /$condition/ ) {
                if ( $i < 3 ) {
                    $i++;
                }
                $share = $state->get_component_instance();
                last;
            } ## end if ( $_ =~ /[Ee]veryone \//...)
        } ## end foreach (@str)

      # we have found a share with the least strict permissions if i becomes 3
      # there is no need of continuing our search
        if ( $i == 3 ) {
            last;
        }
    } ## end foreach my $state (@shares)

    if ( !$share ) {
        NACL::Exceptions::NoElementsFound->throw(
            "No cifs shares found for accessing " . $hash->{path} );
    }
    $hash->{share} = $share;
    delete $hash->{path};
} ## end sub _get_share_from_path

sub _combine_exceptions {
    $Log->enter() if $may_enter;
    my (@exceptions) = @_;
    my $new = thaw(freeze(shift @exceptions)) ;
    while ( my $e = shift @exceptions ) {
        $new->{'-text'} .= "\n" . $e->{'-text'};
    }
    $Log->exit() if $may_exit;
    return $new;
} ## end sub _combine_exceptions

sub _mount_purge_verify {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, %opts ) = @_;
    my $pkg        = $pkg_or_obj->get_package_name();
    my $command_interface = $pkg_or_obj->command_interface() || $opts{command_interface};

    delete $opts{command_interface};

    my $mount_info = $pkg->find( command_interface => $command_interface, filter => {%opts}, allow_empty => 1 );

    if ( $mount_info ) {
        my $mount_exists = 1;
        # Double Check if the mount path exists
        try {
#In case of ipv6, mount output still showing the unmounted mount even if unmount was successful.
#Using ls command to verify that mount point is deleted successfully after unmouting.
	    my $path = $command_interface->apiset()->ls( "paths" => "$opts{'mount_point'}" );

#An extra check (check for more than just the Drive letter) w.r.t Burt 930304
            if ( $opts{'share_name'} && $opts{'server'} ) {
                if (   ( $opts{'share_name'} ne $mount_info->get_one_state_attribute('share_name') )
                    || ( $opts{'server'} ne $mount_info->get_one_state_attribute('server') ) )
                {
                    $mount_exists = 0;
                }
            }

            # Should have been deleted, but still exists. Exception out here
            $Log->exit() if $may_exit;
            if ($mount_exists eq 1){
                NACL::Exceptions::VerifyFailure->throw(
                        'Verification failed! The element '
                         ._dump_one( \%opts )
                        . ' still exists (it should have been purged)' );
            }
        }
        catch NACL::APISet::Exceptions::InvalidOptionException with {
            ## Do nothing it is deleted succesfully
        };
    }

    # If we've gotten here, then all good
    $Log->exit() if $may_exit;
}

=head2 _test_share

This private subroutine performs the read/write test on a specified
share path

=cut

sub _test_share{
    $Log->enter() if $may_enter;
    my $pkg = shift;

    my %opts = $pkg->_common_validate_with(
        params          => \@_,
        additional_spec => {
            path              => {type => SCALAR, optional => 1},
            share_name        => {type => SCALAR, optional => 1},
	    tests             => {type => SCALAR, default => 'rw'},
	    diagnosis_param   => {type => HASHREF, optional => 1},
	    mount_point       => {type => SCALAR, optional => 1},
            test_max_retries  => {type => SCALAR, default => 2},
            test_snooze       => {type => SCALAR, default => 10}
	},
        allow_extra => 1,
    );

    if(!defined $opts{path} &&  !defined $opts{server} && !defined $opts{share_name}){
        $Log->exit() if $may_exit;
        NATE::Exceptions::Argument->throw("Path could not be determined."
                       ." Either of 'path' or 'server' & 'share_name' is required");
    }
    my $path  = $opts{path};
    if ( (defined $opts{protocol} && $opts{protocol} ne 'nfsv3win')) {
       $path //= "\\\\" . $opts{server}
                     . '\\' . $opts{share_name};
    } else {
       $path  //= "\\\\" . $opts{server};
    }
    my $tests = delete $opts{tests};
    my $diagnosis_opts = $opts{diagnosis_param};

    delete $diagnosis_opts->{capture_on_retry};
    $tests = $diagnosis_opts->{share_tests} if(defined $diagnosis_opts->{share_tests});
    my ($command_interface, $test_max_retries, $servers, $test_snooze, $test_failed);
    $command_interface = $opts{command_interface};

    if ($command_interface->is_linux()) {
        $path = $opts{'mount_point'};
    }

    $servers = $opts{server};
    $test_max_retries = delete $opts{test_max_retries};
    $diagnosis_opts->{max_retry} = $test_max_retries;
    $test_snooze  = delete $opts{test_snooze};

    my $test_retry_count = 1;
TEST_SHARE:  {
    $test_failed = 0;
    $pkg->_capture_start(
            command_interface => $command_interface,
            server => $servers,
            diagnosis_param => $diagnosis_opts
        );

    try{
	if($tests =~ /r/){
          $Log->comment("Read Test on path: $path");
          require NACL::C::Client::FileDirOps;
          NACL::C::Client::FileDirOps->read_test(
		command_interface => $command_interface,
		path         => $path,
	  );
        }
        if($tests =~ /w/){
          $Log->comment("Write Test on path: $path");
	  NACL::C::Client::FileDirOps->write_test(
		command_interface => $command_interface,
		path         => $path,
          );
        }
     }catch NACL::C::Exceptions::Client::FileDirOps::WriteTestFailed with{
	    my $exception = shift;
            $test_failed = 1;
            $Log->comment("Write Test failed on path: $path with error: ".$exception->text());
     }catch NACL::C::Exceptions::Client::FileDirOps::ReadTestFailed with{
            my $exception = shift;
            $test_failed = 1;
            $Log->comment("Read Test failed on path: $path with error: ".$exception->text());
     };
     Tharn::snooze($test_snooze) if($test_failed);

}
     if($test_failed && (++$test_retry_count <= $test_max_retries)){
        ## enable diagnose for last retry
        if($test_retry_count == $test_max_retries - 1 ){
          $diagnosis_opts->{current_retry} = $test_retry_count;
          $pkg->_set_diagnose(diagnosis_param  =>  $diagnosis_opts);
          $Log->comment("The Read/Write test failed, Enabling Packet Capture 
             & reiterating the test") if($diagnosis_opts->{enable_capture});
        }
        goto TEST_SHARE;
     }
     ## stop packet capture if running.
     $pkg->_capture_stop(
           command_interface => $command_interface,
           diagnosis_param   => $diagnosis_opts
     ) ;

     $Log->exit() if $may_exit;
} ## end sub _test_share

=head2 _populate_servers

This private subroutine looks for the availbale servers
which is to be mounted on clients

=cut

sub _populate_servers{
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, $opts , $other_opts) = @_;
    my $client_ci = $opts->{command_interface};
    my $server_ci = $opts->{server_command_interface};
    my $vserver  = $other_opts->{vserver};
    my $volume   = $other_opts->{volume};
    my $select_interface = $other_opts->{select_interface};
    my $client_type      = $client_ci->type();
    my $protocol         = $opts->{protocol} // '';
    my $t                = $other_opts->{nacl_task_opts} ;

    my @interface_cs ;
    my ($server, $share, $address, $node);

    if ( !$opts->{server} ) {
       my $mode = $server_ci->mode();
       if ( $mode eq 'CMode' ) {
	if ( !$vserver ) {
	  $Log->exit() if $may_exit;
	  NATE::Exceptions::Argument->throw(
		"vserver input required but not provided");
        }

	# We have to select an interface which supports the protocol
	# that we want to use while mounting. For NFS mounts, "nfs" must be
	# among the supported data protocols
	my %protocol_filter = ( 'data-protocol' => ["nfs"] );

	# We require cifs to be supported either if we're using a Windows
	# client or if we're mounting on a linux client using cifs
	if ( ( $opts->{type} || '' ) eq 'cifs'
		|| ($client_type =~ /windows/i && $protocol ne 'nfsv3win'))
	{
	    $protocol_filter{'data-protocol'} = ["cifs"];
	}
	if ( $select_interface ) {
           if ($volume) {
		$node = NACL::C::Volume->get_local_node_name(
			command_interface => $server_ci,
			volume            => $volume,
			vserver           => $vserver
		);
	   } else {
		if ($client_type =~ /unix/i ) {
		 $Log->exit() if $may_exit;
		 NATE::Exceptions::Argument->throw("volume input required but not provided");
                }
	   }
           if ($node) {
		if ( $select_interface eq "remote" ) {
                        $protocol_filter{'curr-node'} = "!$node";
		} elsif ( $select_interface eq "local" ) {
			$protocol_filter{'curr-node'} = $node;
		}
           }
	} ## end if ( $select_interface...)
		# If nacltask_address_family param is defined,
		# pass this as filter option to network-interface-fatch
		my %addr_family;
		if ( defined $t->{nacltask_address_family} ) {
		    %addr_family	= ( 'address-family' => $t->{nacltask_address_family} );
		}
		@interface_cs = NACL::CS::Interface->fetch(
			command_interface => $server_ci,
			filter            => {
			  vserver => $vserver,
			  %protocol_filter,
			  %addr_family
		        },
		        requested_fields => [qw(address)],
			allow_empty      => 1,
		);
		unless (@interface_cs) {
		   $Log->exit() if $may_exit;
		   NATE::BaseException->throw( "The vserver "
                        . $vserver
                        . ' should have an interface defined for it, but it does not'
		   );
		} ## end unless (@interface_cs)
	  } else {
		$opts->{server} =$server_ci->hostrec()->default_ip();
	}
     } ## end if ( !$opts{server} )
    my $servers = [];

    if (defined $opts->{server}) {
        if (ref $opts->{server}) {
            $servers = $opts->{server};
        } else {
            $servers = [$opts->{server}];
        }
    } else {
        foreach (@interface_cs) {
            push (@$servers,$_->address());
        }
    }
    $opts->{server} = $servers;
    return $servers ;
    $Log->exit() if $may_exit;
} ## end sub _populate_servers

=head2 _capture_start

The private subroutine starts the packet capture on the 
specified share path

=cut

sub _capture_start {
    my ($pkg_or_obj, %opts) = @_;
    my $diagnosis_opts = $opts{diagnosis_param};
    $Log->comment("diagnostic opts are" . Data::Dumper::Dumper($diagnosis_opts));
    return if(!($diagnosis_opts->{enable_diagnosis}));

    my $Capture;
    my $client = NACL::C::Client->new(name => $opts{command_interface}->hostrec->hostname());

    my $details = [{target_machine => $client}];
    ## check if server is an array ref or scalar
    my $servers;
    if (defined $opts{server} && ref $opts{server}) {
	$servers = $opts{server};
    } else {
	$servers = [$opts{server}];
    }

    if(defined $diagnosis_opts->{capture_filter}) {
      $details->[0]->{capture_filter} = $diagnosis_opts->{capture_filter};
    }else{
      my @filter_addresses = @{$servers};
      my $filter;
      foreach my $ip (@filter_addresses) {
         $filter .= "host $ip or ";
      }
      $filter =~ s/or\s+$//;
      $details->[0]->{capture_filter} = $filter;
    }

    $Log->comment(
	"Enabling packet capture on ".$opts{command_interface}->hostrec->hostname()." with the following parameters\n"
		. Dumper($details));
    $Capture   = NACL::MTask::PacketCapture->new(packet_capture_details => $details);
    $Capture->start();

    Tharn::snooze(2);
    $diagnosis_opts->{capture_obj} = $Capture;
    return $Capture;
} ## end sub _capture_start

=head2 _capture_stop

The private subroutine stops the packet capture on the
specified share path

=cut

sub _capture_stop {
    my ($pkg_or_obj, %opts) = @_;

    my $diagnosis_opts = $opts{diagnosis_param};

    return if(!$diagnosis_opts->{enable_diagnosis});
    my $command_interface = $opts{'command_interface'};
    my $Capture = $diagnosis_opts->{capture_obj} ;
    my $capture_file_copy_path = $diagnosis_opts->{capture_file_copy_path};

    $Log->comment("Stopping packet capture on ".$opts{command_interface}->hostrec->hostname());
    try{
       if(defined $Capture){

         $Capture->stop() ;
         my $ret;
         $ret = $Capture->show();
         if (split(/\n/, Dumper($ret)) > 250) {
            $Log->comment('Disabling display of capture because it exceeds 250 lines');
         } else {
            $Log->comment(Data::Dumper::Dumper($ret));
         }
         if(defined $capture_file_copy_path){
           $Log->comment('Copying capture to path: '.$capture_file_copy_path);
           $Capture->get(path => $capture_file_copy_path);
         }
       }
    }finally{
       $Capture->cleanup() if(defined $Capture);
       $diagnosis_opts->{capture_obj} = undef;
    };
} ## end sub _capture_stop

1;
