#
# Copyright (c) 2011 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary ...
## @author dl-nacl-dev@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::STask::Client::RPC::Future

=head1 SYNOPSIS

   my $call = NACL::STask::Client::RPC::Future->new(worker=>$worker,eval=>"2+2");
   $call->start();
   my $answer = $call->wait();  # returns 4


   my $call = NACL::STask::Client::RPC::Future->new(worker=>$worker,
                                                  package=>"NACL::C::Node",
                                                  method=>"find",
                                                  args=>[],
                                                  wrap_result=>1);
   $all->start();
   $call->wait()->get_one_state_attribute("availsize");

=head1 DESCRIPTION

=head1 ATTRIBUTES

=head1 METHODS

=cut

package NACL::STask::Client::RPC::Future;

use strict;
use warnings;
use Params::Validate
    qw(validate SCALAR ARRAYREF BOOLEAN CODEREF HASHREF OBJECT);
use NACL::ComponentUtils qw(_hash_copy);
use NATE::Events qw(listener_add eventloop_once);
use MIME::Base64 qw(decode_base64 encode_base64);
use Storable qw(nfreeze thaw);
use NATE::BaseException qw(:try);
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use NATE::Time qw(timeout2time);

our $MessageId = 0;

# This is set only temporarily while thawing received messages, in
# order to make sure that Wrapper objects that are received from
# remote sources are able to attach themselves to the relevant worker
# being received from

our $Worker = undef;

use Class::MethodMaker [
    new    => [ '-hash', 'new' ],
    scalar => 'package',
    scalar => 'object',
    scalar => 'method',
    scalar => 'args',
    scalar => 'eval',
    scalar => 'worker',
    array  => '_notify_list',
    scalar => [ { -default => 0 }, 'wrap_return' ],
    scalar => '_result',
    scalar => '_exception',
    scalar => 'on_done',
    scalar => '_status',
];

=head2 new

   my $call = NACL::STask::Client::RPC::Future->new(worker=>$worker,eval=>"2+2");

=cut

# send request to Worker.

sub _start {
    $Log->enter() if $may_enter;
    my $self = shift;
    my %opts = validate @_, { process => { type => OBJECT }, };

    my $rpc_request = $self->_hash_copy(
        source => $self,
        copy   => [qw(package object method args eval wrap_return)]
    );

    # message_id to track STAF message
    my $message_id = "NACL/STask/Client/RPC/Reply-$MessageId";
    $rpc_request->{message_id} = $message_id;
    $MessageId++;

    $opts{process}->message_put(
        message => encode_base64( nfreeze($rpc_request) ),
        type    => "NACL/STask/Client/RPC/Request"
    );

    my $listener = {
        filter => sub { $_[0]->{type} eq $message_id },
        once => 1,
        callback => sub {
            local $Worker = $self->worker();
            my $rpc_response = thaw( decode_base64( $_[0]->{message} ) );
            %$self = ( %$self, %$rpc_response );
            $self->_status('done');
            $self->notify();
        },
    };
    listener_add($listener);
    $Log->exit() if $may_exit;
} ## end sub _start

sub notify {
    my $self = shift;
    foreach my $callback ( $self->_notify_list() ) {
        $callback->($self);
    }
    if ( $self->{on_done} ) {
        $self->{on_done}->($self);
    }
} ## end sub notify

=head2 result

   my $result = $future->result();

=over

=item Options

=over

=item timeout, on_timeout

How long ->result() should wait for the call to complete; if this
timeout is reached, then "return &on_timeout($call)" is performed.

(The timeout/on_timeout passed to the wait() call does not in any way
cancel the call once it triggers; it's meant to control how interested
the ->result() caller is in having to wait around for a result, not how
"in error" it is for the call to have taken this long)

If there's a timeout/on_timeout attached to the call itself, then both
of them may be performed.

=back

=back

=cut

sub result {
    $Log->enter() if $may_enter;
    my $self = shift;
    my %opts = validate @_,
        {
        timeout    => { type => SCALAR,  optional => 1 },
        on_timeout => { type => CODEREF, optional => 1 },
        };
    if (defined $opts{timeout}) {
        my $end_time = time + $opts{timeout};
        while ((time <= $end_time) && !defined $self->_status() ){
            my $avl_time = $end_time - time;
            # While eventloop_once returns back as soon as an event is detected
            # (need not signal the completion of required process), eventloop waits completely
            # for the specified time (even if the required process completes before that time)
            # Hence the following hybrid approach of polling and eventloop is
            # used to wait not more than the specified time, but detects the
            # completion as soon as it happens.
            eventloop_once(timeout => $avl_time);
        }
        return $opts{on_timeout}->($self) if (  !defined $self->_status());
    } else {
        while ( !defined $self->_status()) {
            eventloop_once();
        }
    }

    if ( $self->_exception() ) {
        NATE::BaseException->throw( $self->_exception() );
    }

    $Log->exit() if $may_exit;
    return $self->_result();
} ## end sub result

sub exception {
    $Log->enter() if $may_enter;
    my $self = shift;
    my %opts = validate @_,
        {
        timeout    => { type => SCALAR,  optional => 1 },
        on_timeout => { type => CODEREF, optional => 1 },
        };

    my $end_time = timeout2time( $opts{timeout} || -1 );
    while ( !defined $self->_status() ) {
        my $count = eventloop_once( timeout => $end_time );
        if ( !$count ) {
            $Log->exit() if $may_exit;
            return $opts{on_timeout}->($self)
                if ( exists $opts{on_timeout} );
        }
    } ## end while ( !defined $self->_status...)

    $Log->exit() if $may_exit;
    return $self->_exception();
} ## end sub exception

########################## Helper Functions ##############################

sub _add_to_notify_list {
    $Log->enter() if $may_enter;
    my $self              = shift;
    my $callback_code_ref = shift;
    my @callback_array    = $self->_notify_list();
    push @callback_array, $callback_code_ref;
    if (@callback_array) {
        $self->_notify_list(@callback_array);
    } else {
        $self->_notify_list_reset();
    }
    $Log->exit() if $may_exit;
} ## end sub _add_to_notify_list

1;
