#
# Copyright (c) 2011 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary ...
## @author dl-nacl-dev@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::STask::Client::RPC::Wrapper

=head1 DESCRIPTION

This package allows a process (generally an RPC worker process) to
keep another value (let's call it the "real" object) pinned locally so
that future requests by remote processes can indirectly refer to the
real value instead of a copy of the real value being forwarded to the
remote process and then a copy of that forwarded back again.

The way this works is:

=over

=item *

Users don't generally call ->new on this package.  Instead, they
usually add "wrap_return=>1" in their calls to $pool->call or
$worker->call, which will have the worker wrap its real return value
using this package and then return the wrapper instead.

=item *

The real object can be "registered", which causes it to be given a
unique key and stored somewhere, and which creates an instance of this
object referring indirectly to that real object.

=item * 

Instances of this wrapper contain the key and a process id (whatever
is enough to distinguishes this process from other processes that this
value may appar in e.g. registered STAF handle or subtest id)

=item *

When this object is frozen, it contains the key and process id.  (See
STORABLE_freeze in Storable)

=item *

When this object is thawed on the process containing the real object
(identified by process id), then calling arbitrary methods on this
reference will (via AUTOLOAD) call the corresponding method on the
real object.

=item *

When this object is thawed in some other process, by an RPC parent
process receiving from a particular worker, the future/worker/pool
doing the thawing will add the worker as a property of this referece.
A call to an arbitrary method on this reference will (via AUTOLOAD)
call $worker->call to invoke the corresponding method and argumnts in
the corresponding worker.

[Implementation notes: this can be done by STORABLE_attach using
tricks with "local".  Consider if the worker should be a weakened
reference (?).  STORABLE_freeze should not include the worker in the
frozen value]

=back

=head1 ATTRIBUTES

=head1 METHODS

=cut

package NACL::STask::Client::RPC::Wrapper;
use Sys::Hostname qw(hostname);
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Params::Validate qw(validate OBJECT);
use NATE::BaseException qw(:try);

our %Objects;
our $Hostname = hostname();
our $Id       = 0;

use Class::MethodMaker [ new => [ '-init', 'new' ], ];

sub init {
    my $self = shift;
    my %opts = validate @_, { "value" => { type => OBJECT }, };
    $self->{_hostname}       = $Hostname;
    $self->{_pid}            = $$;
    $self->{_id}             = $Id++;
    $self->{_worker}         = undef;
    $Objects{ $self->{_id} } = $opts{value};
    $Log->exit() if $may_exit;

} ## end sub init

sub AUTOLOAD {
    $Log->enter() if $may_enter;
    my $self = shift;
    our $AUTOLOAD;
    my $autoload = $AUTOLOAD;
    return if ( $AUTOLOAD =~ /DESTROY/ );
    $autoload =~ s/.*:://;

    if ( exists $self->{_value} ) {
        $Log->exit() if $may_exit;
        return $self->{_value}->$autoload(@_);
    }
    if ( !$self->{_worker} ) {
        $Log->exit() if $may_exit;
        NATE::BaseException->throw(
            "No worker defined on this wrapper.  ->_rpc_foo methods can only be called on wrapper objects received via an RPC worker "
        );
    } ## end if ( !$self->{_worker})

    if ( ( my $rpc_method ) = ( $autoload =~ m/^_rpc_(.*)/ ) ) {
        $Log->exit() if $may_exit;
        return $self->{_worker}->$rpc_method( object => $self, @_ );
    } else {
        $Log->exit() if $may_exit;
        return $self->{_worker}
            ->call( object => $self, method => $autoload, args => \@_ );
    }
} ## end sub AUTOLOAD

sub STORABLE_freeze {
    $Log->enter() if $may_enter;
    my ( $self, $cloning ) = @_;
    return if $cloning;    # Regular default serialization
    my $data = {%$self};
    delete $data->{_worker};
    my $dummy = "";
    $Log->exit() if $may_exit;
    return ( $dummy, $data );
} ## end sub STORABLE_freeze

sub STORABLE_thaw {
    $Log->enter() if $may_enter;
    my ( $self, $cloning, $dummy, $data ) = @_;
    %$self = %$data;
    if ( $self->{_hostname} eq $Hostname && $self->{_pid} eq $$ ) {
        $self->{_value} = $Objects{ $self->{_id} };
    }
    my $worker = $NACL::STask::Client::RPC::Future::Worker;
    $self->{_worker} = $worker if $worker;
    $Log->exit() if $may_exit;
} ## end sub STORABLE_thaw

1;
