#
# Copyright (c) 2001-2013 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Transition Task Module
## @author rohit4@netapp.com,dl-nacl-dev@netapp.com
## @change     2015-02-04 santosh6 : Added QBT supporting subroutines

## @status Shared
## @pod here
################################################################################
package NACL::STask::Client::Transition;

use strict;
use warnings;

use base qw/NACL::C::Client::Transition NACL::STask::STask/;

use NATE::Log qw(log_global);
my $Log = log_global();
use Params::Validate
    qw(validate validate_with SCALAR BOOLEAN SCALARREF HASHREF OBJECT UNDEF ARRAYREF);
use NATE::Exceptions::Argument qw(:try);
use NACL::Exceptions::NoElementsFound qw(:try);
use NATE::BaseException qw(:try);
use NACL::APISet::Exceptions::TimeoutException ();
use NACL::Exceptions::UnexpectedState;
use NACL::GeneralUtils qw(nacl_method_retry);

=head1 NAME

NACL::STask::Client::Transition

=head1 DESCRIPTION

C<NACL::STask::Client::Transition> provides a number of well-defined but
potentially complex or multi-step methods related to transition session.

It builds on top of, and is a derived class of C<NACL::C::Client::Transition>,
and so it also provides methods that are more in the scope of
individual transition-related commands.  See C<NACL::C::Client::Transition> for
details.  This also means that a C<NACL::STask::Client::Transition> object may
generally be used in place of a component object. If any of start, complete fails
in below methods then Task retries thrice with an interval of 10 seconds after which
it throws NACL::APISet::Exceptions::CommandFailedException.
These defaults can be changed using nacltask_retry_tries_count and nacltask_retry_sleep_time .

=head1 ATTRIBUTES

=head2 command_interface

(Required) A component object that represents the host to which to send commands.

=head2 session-name

(Required) The name of the transition session used to identify it uniquely .
(Session names are unique within a given client on which 7-mode transition tool is installed)

=head1 METHODS

=head2 get_transitioned_volume

  my $trans_obj = NACL::STask::Client::Transition->get_transitioned_volume(
        command_interface               => $hostwindows,
        nacltask_rng_args               => { length => '12' },
        "method-timeout"                => $method_timeout,
        S_mode_credentials              => \%S_mode_credentials,
        C_mode_credentials              => \%C_mode_credentials,
        create_args                     => \%create,
        schedule_add_args               => \@schedule,
        volumepair_add_args             => \%volume,
        qtreepair_add_args              => \%qtreepair,
        stagingsystems_add_args         => \%stagingsystems,
        volumeaggregatemapping_add_args => \%volumeaggregate,
        precheck_args                   => \%precheck,
        start_args                      => \%start,
        precutover_args                 => \%precutover,
        cutover_args                    => \%cutover,
    );

(Class method) Creates transitioned volumes on the C-mode.

This method

1) Adds 7 mode and C mode credentials

2) Creates a transition session

3) Adds volume-pairs to session

4) Adds qtree-pairs to session

5) Adds staging-system to session

6) Adds volume-aggregate-mapping to session

7) Adds schedules to session

8) Runs starts on the session

9) Waits till the session state reaches Copy with status Copy-update

10)Runs pre-cutover on the session

11)Waits till the session state reaches Precutover with status Copy-update

12)Runs cutover on the session

13)Waits till the session state reaches Completed

Note that session-name has to be supplied as shown above and not inside %create .The
time for which the task waits for the session status/state to reach Copy-update/Completed after
start/cutover operations respectively  depends upon method-timeout value passed.
method-timeout has to be passed to the STask as above and not inside any of the hashes .

User should increase the value of method-timeout depending upon the configurations
transitioned . If its a huge configuration then start and cutover operations
will take huge time and in that case user should adjust the method-timeout accordingly.

Note that the argument C<session-name> is optional for this method.
If this is not specified, a random name is generated using
L<NACL::GeneralUtils::random_name_generator|lib-NACL-GeneralUtils-pm/random_name_generator>.
(See C<nacltask_rng_args> for how to customize arguments to
random_name_generator())

=over

=item Options

=over

=item C<< S_mode_credentials  => \%S_mode_credentials  >>

(Required) Its a hash reference to a hash which contains the arguments required
for adding 7-mode credentials. It accepts all the arguments accepted by
L<< NACL::C::Client::Transition->credentials_add|lib-NACL-C-Client-Transition-pm/credentials_add >>.
As mentioned above this hash should not contain  method-timeout and other common options.

    my %S_mode_credentials=(
        "host-name"       => "10.227.202.110",
        username          => "root",
        password          => "netapp1!"
    );

=item C<< C_mode_credentials  => \%C_mode_credentials  >>

(Required) Its a hash reference to a hash which contains the arguments required
for adding C-mode credentials. It accepts all the arguments accepted by
L<< NACL::C::Client::Transition->credentials_add|lib-NACL-C-Client-Transition-pm/credentials_add >>.
As mentioned above this hash should not contain  method-timeout and other common options.

    my %C_mode_credentials=(
        "host-name"       => "10.227.202.110",
        username          => "admin",
        password          => "netapp1!"
    );

=item C<< S_mode_staging_credentials  => \%S_mode_staging_credentials  >>

(Required) Its a hash reference to a hash which contains the arguments required
for adding 7-mode staging credentials. It accepts all the arguments accepted by
L<< NACL::C::Client::Transition->credentials_add|lib-NACL-C-Client-Transition-pm/credentials_add >>.
As mentioned above this hash should not contain  method-timeout and other common options.

    my %S_mode_staging_credentials=(
        "host-name"       => "10.227.202.112",
        username          => "root",
        password          => "netapp1!"
    );

=item C<< "create_args" => \%create >>

(Required) Its a hash reference to a hash which contains the arguments required
for creating the session. It accepts all the arguments accepted by
L<< NACL::C::Client::Transition->create|lib-NACL-C-Client-Transition-pm/create >>.
As mentioned above this hash should not contain session-name , command_interface ,
method-timeout , nacltask_rng_args and other common options.

    my %create=(
        "session-type"              => "standalone",
        "7-mode-host-name"          => "10.227.202.110",
        "data-copy-ipaddress"       => "10.227.202.110",
        vfiler                      => "vfiler0",
        "c-mode-host-name"          => "10.230.25.178",
        "svm-name"                  => "vs_sec",
        "schedule-name"             => "daily",
    );

=item C<< "volumepair_add_args" => \%volume >>

(Required) Its a hash reference to a hash which contains the arguments specific to
volumepair adding. It accepts all the arguments accepted by
L<< NACL::C::Client::Transition->volumepair_add|lib-NACL-C-Client-Transition-pm/volumepair_add >>.
As mentioned above this hash should not contain session-name , command_interface ,
method-timeout and other common options.

    my %volume=(
        "volume-name-7-mode"  => \@arr1,
        "volume-name-c-mode"  => \@arr2,
        "aggregate-name"      => "aggr_trans",
    );

=item C<< "qtreepair_add_args" => \%qtreepair >>

(Required) Its a hash reference to a hash which contains the arguments specific to
qtreepair adding. It accepts all the arguments accepted by
L<< NACL::C::Client::Transition->qtreepair_add|lib-NACL-C-Client-Transition-pm/qtreepair_add >>.
As mentioned above this hash should not contain session-name , command_interface ,
method-timeout and other common options.

    my %qtreepair=(
        "volume-name-7-mode"  => "vol1",
        "qtree-name-7-mode"   => \@qtree_7mode,
        "volume-name-c-mode"  => \@vol_cmode,
        "qtree-name-c-mode"   => \@qtree_cmode,
    );

=item C<< "stagingsystems_add_args" => \%stagingsystems >>

(Required) Its a hash reference to a hash which contains the arguments specific to
qtreepair adding. It accepts all the arguments accepted by
L<< NACL::C::Client::Transition->stagingsystems_add|lib-NACL-C-Client-Transition-pm/stagingsystems_add >>.
As mentioned above this hash should not contain session-name , command_interface ,
method-timeout and other common options.

    my %stagingsystems=(
        "host-name"            => "10.227.202.112",
        "vfiler-name"          => "vfiler0",
        "aggregates"           => "aggr_trans1,aggr_trans2",
        "datacopy-ip-address"  => "10.227.202.112",
    );

=item C<< "volumeaggregatemapping_add_args" => \%volumeaggregate >>

(Required) Its a hash reference to a hash which contains the arguments specific to
qtreepair adding. It accepts all the arguments accepted by
L<< NACL::C::Client::Transition->volumeaggregatemapping_add|lib-NACL-C-Client-Transition-pm/volmeaggregatemappping_add >>.
As mentioned above this hash should not contain session-name , command_interface ,
method-timeout and other common options.

    my %volumeaggregate=(
        "volume-name-c-mode"       => \@vol_cmode,
        "target-cluster-aggregate" => "aggr_trans",
        "staging-system-host-name" => "10.227.202.112",
        "staging-aggregate"        => "aggr_trans1",
    );

=item C<< "schedule_add_args" => \@schedule >>

(Required) Its a reference to an array of hash references where each hash contains
separate schedule information. It accepts all the arguments accepted by
L<< NACL::C::Client::Transition->schedule_add|lib-NACL-C-Client-Transition-pm/schedule_add >>.
As mentioned above this hash should not contain session-name , command_interface ,
method-timeout and other common options.

    my %schedule1=(
        "schedule-name-SN1.3"           => "new1",
        "days-range"                    => "1-4",
        "start-time"                    => "12:00",
        "duration"                      => "20:30",
        "available-transfers-percentage"=> "4",
        "session-snapmirror-throttle"   => "230",
        "update-frequency"              => "1:30",
    );
    my %schedule2=(
        "schedule-name-SN1.3"           => "new2",
        "days-range"                    => "5",
        "start-time"                    => "12:00",
        "duration"                      => "20:30",
        "available-transfers-percentage"=> "4",
        "session-snapmirror-throttle"   => "230",
        "update-frequency"              => "1:30",
    );
    my @schedule=(\%schedule1,\%schedule2);

=item C<< start_args          => \%start >>

(Optional) Its a reference to a hash which contains the arguments specific to
transition start. It accepts all the arguments accepted by
L<< NACL::C::Client::Transition->start|lib-NACL-C-Client-Transition-pm/start >>.
As mentioned above this hash should not contain session-name , command_interface ,
method-timeout and other common options.

    my %start=(
        "no-precheck" => 0
    );

=item C<< precutover_args       => \%precutover >>

(Optional) Its a hash reference to a hash which contains the arguments specific to
transition precutover. It accepts all the arguments accepted by
L<< NACL::C::Client::Transition->precutover|lib-NACL-C-Client-Transition-pm/precutover >>.
As mentioned above this hash should not contain session-name , command_interface ,
method-timeout and other common options.

    my %precutover=(
        mode                       => "rw_test",
        ## time_to_wait_in_rw(in seconds) is applicable only if mode selected is rw_test
        ## After this time a resync will be issued automatically
        time_to_wait_in_rw         => "240",
        nacltask_retry_tries_count => 10 ,
    );

=item C<< cutover_args       => \%cutover >>

(Optional) Its a hash reference to a hash which contains the arguments specific to
transition cutover. It accepts all the arguments accepted by
L<< NACL::C::Client::Transition->cutover|lib-NACL-C-Client-Transition-pm/cutover >>.
As mentioned above this hash should not contain session-name , command_interface ,
method-timeout and other common options.

    my %cutover=(
        nacltask_retry_tries_count => 10 ,
    );

=item C<< "method-timeout" => 10800 >>

(Optional) As component method-timeout, controls how long the command
will wait before completing. Default timeout is 10800 seconds.

=item C<< nacltask_rng_args => { args_for_rng_here } >>

(Optional, used only if a random session name is generated)
If no session name is supplied, a random name is generated using
L<NACL::GeneralUtils::random_name_generator|lib-NACL-GeneralUtils-pm/random_name_generator>
with two random words in it . This argument
can be used to customize the arguments to that method. For example, if we'd
like to specify that the random generated session name should be 255 characters
length, then this could be:

    NACL::STask::Client::Transition->get_transitioned_volume(
        %other_opts,
        nacltask_rng_args => { length => '255' }
    );

See L<NACL::GeneralUtils::random_name_generator|lib-NACL-GeneralUtils-pm/random_name_generator>
for a description of the arguments it accepts.

=item Arguments defaulted/auto-determined

=over

=item C<session-name>

Note that if this is not specified, a random name is generated using
L<NACL::GeneralUtils::random_name_generator|lib-NACL-GeneralUtils-pm/random_name_generator>.
(See C<nacltask_rng_args> for how to customize arguments to
random_name_generator())

=back

=back

=back

=over

=item Exceptions

=over

=item C<NACL::C::Exceptions::Client::Transition::AlreadyExists>

This type of exception is thrown when an attempt is made to create a session that already exists.

=back

=back

=cut


sub get_transitioned_volume {

    $Log->enter();
    my ($pkg,@args) = @_;
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            nacltask_rng_args               => $pkg->_rng_args_validate_spec(),
            S_mode_credentials              => { type => HASHREF,  optional => 0 },
            C_mode_credentials              => { type => HASHREF,  optional => 0 },
            S_mode_staging_credentials      => { type => HASHREF,  optional => 1 },
            create_args                     => { type => HASHREF,  optional => 0 },
            volumepair_add_args             => { type => HASHREF,  optional => 1 },
            qtreepair_add_args              => { type => HASHREF,  optional => 1 },
            stagingsystems_add_args         => { type => HASHREF,  optional => 1 },
            volumeaggregatemapping_add_args => { type => HASHREF,  optional => 1 },
            schedule_add_args               => { type => ARRAYREF, optional => 1 },
            precheck_args                   => { type => HASHREF,  optional => 1 },
            start_args                      => { type => HASHREF,  optional => 1 },
            precutover_args                 => { type => HASHREF,  optional => 1 },
            cutover_args                    => { type => HASHREF,  optional => 1 },
            prim_create_args                => { type => HASHREF,  optional => 1 },
            sec_create_args                 => { type => HASHREF,  optional => 1 },
            prim_S_mode_credentials         => { type => HASHREF,  optional => 1 },
            ignorableerrors_args            => { type => HASHREF,  optional => 1 },
        },
        ignore_primary_keys => 1,
        allow_extra => 1,
    );

    # Transform %opts from the options we received into the options to
    # pass to the base class method.

    my %common_opts;
    $pkg->_copy_common_component_params_with_ci(
        source => \%opts,
        target => \%common_opts
    );

    my %rng_args;
    $pkg->_hash_move(
        source => \%opts,
        target => \%rng_args,
        move   => [ qw( nacltask_rng_args ) ]
    );

    my $S_mode_credentials          = delete $opts{S_mode_credentials};
    my $C_mode_credentials          = delete $opts{C_mode_credentials};
    my $S_mode_staging_credentials  = delete $opts{S_mode_staging_credentials};
    my $create_args                 = delete $opts{create_args};
    my $volumepair_add_args         = delete $opts{volumepair_add_args};
    my $qtreepair_add_args          = delete $opts{qtreepair_add_args};
    my $stagingsystems_add_args     = delete $opts{stagingsystems_add_args};
    my $volumeaggregatemapping_add_args = delete $opts{volumeaggregatemapping_add_args};
    my $schedule_add_args           = delete $opts{schedule_add_args};
    my $precheck_args               = delete $opts{precheck_args};
    my $start_args                  = delete $opts{start_args};
    my $precutover_args             = delete $opts{precutover_args};
    my $cutover_args                = delete $opts{cutover_args};
    my $session_name                = delete $opts{"session-name"};
    my $prim_create_args            = delete $opts{prim_create_args};
    my $sec_create_args             = delete $opts{sec_create_args};
    my $prim_S_mode_credentials     = delete $opts{prim_S_mode_credentials};
    my $ignorableerrors_args        = delete $opts{ignorableerrors_args};

    $pkg->credentials_add( %$S_mode_credentials , %common_opts );

    $pkg->credentials_add( %$C_mode_credentials , %common_opts );

    $pkg->credentials_add( %$S_mode_staging_credentials , %common_opts ) if ( $S_mode_staging_credentials );

    if($create_args->{"session-type"} eq "secondary"){
        $pkg->credentials_add( %$prim_S_mode_credentials , %common_opts );
    }

    my $self = $pkg->create( %$create_args , %common_opts , %rng_args ,
        "session-name" => $session_name
    );

    my %add_sec_cluster_mode;
    my %add_pri_seven_mode;
    if($create_args->{"session-type"} eq "primary"){
        $add_sec_cluster_mode{"c-mode-host-name"}       = $sec_create_args->{"c-mode-host-name"};
        $add_sec_cluster_mode{"svm-name"}    = $sec_create_args->{"svm-name"};
        $self->add_secondary_cluster_mode_system( %add_sec_cluster_mode , %common_opts );
    }
    if($create_args->{"session-type"} eq "secondary"){
        $add_pri_seven_mode{"source-host"} = $prim_create_args->{"source-host"} || $prim_create_args->{"7-mode-host-name"};
        $add_pri_seven_mode{"management-ipaddress"}         = $prim_create_args->{"7-mode-host-name"};
        $add_pri_seven_mode{"data-copy-ipaddress-staggered"}= $prim_create_args->{"data-copy-ipaddress"};
        if($prim_create_args->{"multipathing-ipaddress"} ne undef){
            $add_pri_seven_mode{"multipathing-ipaddress"}   = $prim_create_args->{"multipathing-ipaddress"};
        }
        $self->add_primary_seven_mode_system( %add_pri_seven_mode , %common_opts );
    }

    $self->volumepair_add( %$volumepair_add_args , %common_opts ) if ( $volumepair_add_args );

    $self->qtreepair_add( %$qtreepair_add_args , %common_opts ) if ( $qtreepair_add_args );

    $self->stagingsystems_add( %$stagingsystems_add_args , %common_opts ) if ( $stagingsystems_add_args );

    $self->volumeaggregatemapping_add( %$volumeaggregatemapping_add_args , %common_opts ) if ( $volumeaggregatemapping_add_args );

    if ( $schedule_add_args ) {
        my @schedule_add_args=@{$schedule_add_args};
        foreach (@schedule_add_args){
            $self->schedule_add( %{$_} , %common_opts );
        }
    }

    if ($ignorableerrors_args) {
        $self->ignorableerrors_add(%$ignorableerrors_args,%common_opts);
    }

    $self->precheck( %$precheck_args, %common_opts );

    $self->start( %$start_args , %common_opts );

    ## By default we wait for 60 secs in RW mode before we fire a resync
    my $time_to_wait_in_rw = delete $precutover_args->{time_to_wait_in_rw};
    $time_to_wait_in_rw = 60 if (! defined $time_to_wait_in_rw );

    $self->precutover( %$precutover_args , %common_opts );

    if($precutover_args->{mode} =~ /rw_test/i){
        sleep $time_to_wait_in_rw;
        my %wait_opts;
        $self->_hash_move(
            source => $precutover_args,
            target => \%wait_opts,
            move   => [qw(polling_interval)]
        );
        my %retry_opts;
        $self->_hash_move(
            source => $precutover_args,
            target => \%retry_opts,
            move   => [qw(nacltask_retry_tries_count nacltask_retry_sleep_time)]
        );
        $self->resync(%common_opts,%wait_opts,%retry_opts);
    }

    $self->cutover( %$cutover_args , %common_opts ) if ( $cutover_args );

    $Log->exit();
    return $self;

} ## end sub get_transitioned_volume

=head2 get_transitioned_volume_staggered

  NACL::STask::Client::Transition->get_transitioned_volume_staggered(
        nacltask_rng_args   =>{ length = '12' },
        command_interface   => $Client,
        secondary_info      => \%secondary_info,
        primary_info        => \%primary_info
    );

(Class method) Creates transitioned volumes on primary and secondary C-mode

This method

 1)Adds secondary 7 mode and secondary C mode credentials

 2)Adds primary 7 mode and primary C mode credentials

 3)Creates a transition secondary session

 4)Adds primary 7 mode system to the secondary session

 5)Adds volume-pairs to secondary session

 6)Adds schedules to secondary session

 7)Runs start on the secondary session

 8)Waits till the secondary session state reaches Copy with status Copy-update

 9)Runs pre-cutover on the secondary session

10)Waits till the secondary session state reaches Precutover with status Copy-update

11)Runs cutover on the secondary session

12)Waits till the secondary session state reaches Completed

13)Creates a transition primary session

14)Adds secondary C mode system to the primary session

15)Adds volume-pairs to primary session

16)Adds schedules to primary session

16)Runs start on the primary session

17)Waits till the primary session state reaches Copy with status Copy-update

18)Runs pre-cutover on the primary session

19)Waits till the primary session state reaches Precutover with status Copy-update

20)Runs cutover on the primary session

21)Waits till the primary session state reaches Completed

Note that before invoking this get_transitioned_volume_staggered() ,
there should exists a snapmirror relationship between primary 7-mode volumes
and secondary 7-mode volumes.

=over

=item Options

=over

=item C<< secondary_info  =>  \%secondary_info  >>

(Required) Its a hash reference which contains the arguments required
for creating a secondary session and a transitioned volume at secondary C-mode

    my %secondary_info=(
        "session-name"      => "Sec",
        "method-timeout"    => 10800,
        S_mode_credentials  => \%S_mode_credentials_sec,
        C_mode_credentials  => \%C_mode_credentials_sec,
        create_args         => \%create_sec,
        volumepair_add_args => \%volume_sec,
        schedule_add_args   => \@schedule,
        precheck_args       => \%precheck,
        start_args          => \%start_sec,
        precutover_args     => \%precutover_sec,
        cutover_args        => \%cutover_sec,
    );

Note that session-name has to be supplied as shown above and not inside %create_sec. The
time for which the task waits for the session status/state to reach Copy-update/Completed after
start/cutover operations respectively depends upon method-timeout value passed.
method-timeout has to be passed to the STask as above and not inside any of the hashes.

User should increase the value of method-timeout depending upon the configurations
transitioned. If its a huge configuration then start and cutover operations
will take huge time and in that case user should adjust the method-timeout accordingly.

Note that the argument C<session-name> is optional for this hash.
If this is not specified, a random name is generated using
L<NACL::GeneralUtils::random_name_generator()|lib-NACL-GeneralUtils-pm/random_name_generator>.
(See C<nacltask_rng_args> for how to customize arguments to
random_name_generator())

=over

=item C<< S_mode_credentials  => \%S_mode_credentials_sec  >>

(Required) Its a hash reference which contains the arguments required
for adding secondary 7-mode credentials. It accepts all the arguments accepted by
L<< NACL::C::Client::Transition->credentials_add()|lib-NACL-C-Client-Transition-pm/credentials_add >>.
As mentioned above this hash should not contain method-timeout and other common options.

    my %S_mode_credentials_sec=(
        "host-name"       => $Host7Mode_SEC->{MGMT_IP},
        username          => "root",
        password          => $Host7Mode_SEC->{default_password}
    );

=item C<< C_mode_credentials  => \%C_mode_credentials_sec  >>

(Required) Its a hash reference which contains the arguments required
for adding secondary C-mode credentials. It accepts all the arguments accepted by
L<< NACL::C::Client::Transition->credentials_add()|lib-NACL-C-Client-Transition-pm/credentials_add >>.
As mentioned above this hash should not contain method-timeout and other common options.

    my %C_mode_credentials_sec=(
        "host-name"      => $ip_sec,
        username         => "admin",
        password         => $HostCMode->{default_password}
    );

=item C<< create_args  => \%create_sec >>

(Required) Its a hash reference which contains the arguments required
for creating the session. It accepts all the arguments accepted by
L<< NACL::C::Client::Transition->create()|lib-NACL-C-Client-Transition-pm/create >>.
As mentioned above the hash should not contain session-nmae , comman_interface ,
method-timeout , nacltask_rng_args and other common options.

    my %create_sec=( 
        "session-type"              => "secondary",
        "7-mode-host-name"          => $Host7Mode_SEC->{MGMT_IP},
        "data-copy-ipaddress"       => $Host7Mode_SEC->{MGMT_IP},
        vfiler                      => $sec_filer,
        "c-mode-host-name"          => $ip_sec,
        "svm-name"                  => "vserver_sec",
        "schedule-name"             => $schedule,
    );

=item C<< volumepair_add_args  => \%volume_sec >>

(Required) Its a hash reference which contains the arguments specific to
volumepair adding. It accepts all the arguments accepted by
L<< NACL::C::Client::Transition->volumepair_add()|lib-NACL-C-Client-Transition-pm/volumepair_add >>.
As mentioned above this hash should not contain session-name , command_interface ,
method-timeout and other common options.

    my %volume_sec=(
        "volume-name-7-mode"  => \@arr1,
        "volume-name-c-mode"  => \@arr2,
        "aggregate-name"      => $aggr,
    );

=item C<< "schedule_add_args" => \@schedule >>

(Required) Its a reference to an array of hash references where each hash contains
separate schedule information. It accepts all the arguments accepted by
L<< NACL::C::Client::Transition->schedule_add|lib-NACL-C-Client-Transition-pm/schedule_add >>.
As mentioned above this hash should not contain session-name , command_interface ,
method-timeout and other common options.

    my %schedule1=(
        "schedule-name-SN1.3"           => "new1",
        "days-range"                    => "1-4",
        "start-time"                    => "12:00",
        "duration"                      => "20:30",
        "available-transfers-percentage"=> "4",
        "session-snapmirror-throttle"   => "230",
        "update-frequency"              => "1:30",
    );
    my %schedule2=(
        "schedule-name-SN1.3"           => "new2",
        "days-range"                    => "5",
        "start-time"                    => "12:00",
        "duration"                      => "20:30",
        "available-transfers-percentage"=> "4",
        "session-snapmirror-throttle"   => "230",
        "update-frequency"              => "1:30",
    );
    my @schedule=(\%schedule1,\%schedule2);

=item C<< start_args          => \%start_sec >>

(Optional) Its a hash reference which contains the arguments specific to
transition start. It accepts all the arguments accepted by
L<< NACL::C::Client::Transition->start()|lib-NACL-C-Client-Transition-pm/start >>.
As metioned above this hash should not contain session-name , command_interface ,
method-timeout and other common options.

    my %start_sec=(
        "no-precheck" => 0,
        nacltask_retry_tries_count => 1000000 ,
    );

=item C<< precutover_args       => \%precutover_sec >>

(Optional) Its a hash reference which contains the arguments specific to
transition precutover. It accepts all the arguments accepted by
L< NACL::C::Client::Transition->precutover()|lib-NACL-C-Client-Transition-pm/precutover >.
As metioned above this hash should not contain session-name , command_interface ,
method-timeout and other common options.

    my %precutover_sec=(
        mode                       => "rw_test",
        ## time_to_wait_in_rw(in seconds) is applicable only if mode selected is rw_test
        ## After this time a resync will be issued automatically
        time_to_wait_in_rw         => "240",
        nacltask_retry_tries_count => 1000000 ,
    );

=item C<< cutover_args       => \%cutover_sec >>

(Optional) Its a hash reference which contains the arguments specific to
transition cutover. It accepts all the arguments accepted by
L< NACL::C::Client::Transition->cutover()|lib-NACL-C-Client-Transition-pm/cutover >.
As metioned above this hash should not contain session-name , command_interface ,
method-timeout and other common options.

    my %cutover_sec=(
        nacltask_retry_tries_count => 1000000 ,
    );

=item C<< "method-timeout" => 10800 >>

(Optional) As component method-timeout, controls how long the command
will wait before completing. default timeout is 10800 seconds.

=item C<session-name>

Note that if this is not specified, a random name is generated using
L<NACL::GeneralUtils::random_name_generator()|lib-NACL-GeneralUtils-pm/random_name_generator>.
random_name_generator())

=back

=item C<< primary_info  =>  \%primary_info  >>

(Required) Its a hash reference which contains the arguments required
for creating a primary session and a transitioned volume at primary C-mode

     my %primary_info=(
        "method-timeout"    => 10800,
        "session-name"      => "Pri",
        S_mode_credentials  => \%S_mode_credentials_pri,
        C_mode_credentials  => \%C_mode_credentials_pri,
        create_args         => \%create_pri,
        volumepair_add_args => \%volume_pri,
        schedule_add_args   => \@schedule,
        precheck_args       => \%precheck,
        start_args          => \%start_pri,
        precutover_args     => \%precutover_pri,
        cutover_args        => \%cutover_pri,
    );

Note that session-name has to be supplied as shown above and not inside %create_pri. The
time for which the task waits for the session status/state to reach Copy-update/Completed after
start/cutover operations respectively depends upon method-timeout value passed.
method-timeout has to be passed to the STask as above and not inside any of the hashes.

User should increase the value of method-timeout depending upon the configurations
transitioned. If its a huge configuration then start and cutover operations
will take huge time and in that case user should adjust the method-timeout accordingly.

Note that the argument C<session-name> is optional for this hash.
If this is not specified, a random name is generated using
L<NACL::GeneralUtils::random_name_generator()|lib-NACL-GeneralUtils-pm/random_name_generator>.
(See C<nacltask_rng_args> for how to customize arguments to
random_name_generator())

=over

=item C<< S_mode_credentials  => \%S_mode_credentials_pri  >>

(Required) Its a hash reference which contains the arguments required
for adding primary 7-mode credentials. It accepts all the arguments accepted by
L<< NACL::C::Client::Transition->credentials_add()|lib-NACL-C-Client-Transition-pm/credentials_add >>.
As mentioned above this hash should not contain method-timeout and other common options.

    my %S_mode_credentials_pri=(
        "host-name"      => $Host7Mode->{MGMT_IP},
        username         => "root",
        password         => $Host7Mode->{default_password}
    );

=item C<< C_mode_credentials  => \%C_mode_credentials_pri  >>

(Required) Its a hash reference which contains the arguments required
for adding primary C-mode credentials. It accepts all the arguments accepted by
L<< NACL::C::Client::Transition->credentials_add()|lib-NACL-C-Client-Transition-pm/credentials_add >>.
As mentioned above this hash should not contain method-timeout and other common options.

    my %C_mode_credentials_pri=(
        "host-name"       => $ip_pri,
         username         => "admin",
         password         => $HostCMode_SEC->{default_password}
     );

=item C<< create_args  => \%create_pri >>

(Required) Its a hash reference which contains the arguments required
for creating the session. It accepts all the arguments accepted by
L<< NACL::C::Client::Transition->create()|lib-NACL-C-Client-Transition-pm/create >>.
As mentioned above the hash should not contain session-nmae , comman_interface ,
method-timeout , nacltask_rng_args and other common options.

    my %create_pri=(
        "session-type"              => "primary",
        "7-mode-host-name"          => $Host7Mode->{MGMT_IP},
        "data-copy-ipaddress"       => $Host7Mode->{MGMT_IP},
        vfiler                      => $prim_filer,
        "c-mode-host-name"          => $ip_pri,
        "svm-name"                  => $vserver,
        "schedule-name"             => $schedule,
    );

=item C<< volumepair_add_args  => \%volume_pri >>

(Required) Its a hash reference which contains the arguments specific to
volumepair adding. It accepts all the arguments accepted by
L<< NACL::C::Client::Transition->volumepair_add()|lib-NACL-C-Client-Transition-pm/volumepair_add >>.
As mentioned above this hash should not contain session-name , command_interface ,
method-timeout and other common options.

    my %volume_pri=(
        "volume-name-7-mode"  => \@arr1,
        "volume-name-c-mode"  => \@arr2,
        "aggregate-name"      => $aggr,
    );

=item C<< "schedule_add_args" => \@schedule >>

(Required) Its a reference to an array of hash references where each hash contains
separate schedule information. It accepts all the arguments accepted by
L<< NACL::C::Client::Transition->schedule_add|lib-NACL-C-Client-Transition-pm/schedule_add >>.
As mentioned above this hash should not contain session-name , command_interface ,
method-timeout and other common options.

    my %schedule1=(
        "schedule-name-SN1.3"           => "new1",
        "days-range"                    => "1-4",
        "start-time"                    => "12:00",
        "duration"                      => "20:30",
        "available-transfers-percentage"=> "4",
        "session-snapmirror-throttle"   => "230",
        "update-frequency"              => "1:30",
    );
    my %schedule2=(
        "schedule-name-SN1.3"           => "new2",
        "days-range"                    => "5",
        "start-time"                    => "12:00",
        "duration"                      => "20:30",
        "available-transfers-percentage"=> "4",
        "session-snapmirror-throttle"   => "230",
        "update-frequency"              => "1:30",
    );
    my @schedule=(\%schedule1,\%schedule2);

=item C<< start_args          => \%start_pri >>

(Optional) Its a hash reference which contains the arguments specific to
transition start. It accepts all the arguments accepted by
L<< NACL::C::Client::Transition->start()|lib-NACL-C-Client-Transition-pm/start >>.
As metioned above this hash should not contain session-name , command_interface ,
method-timeout and other common options.

     my %start_pri=(
        "no-precheck" => 0,
        nacltask_retry_tries_count => 1000000 ,
    );

=item C<< precutover_args       => \%precutover_pri >>

(Optional) Its a hash reference which contains the arguments specific to
transition precutover. It accepts all the arguments accepted by
L< NACL::C::Client::Transition->precutover()|lib-NACL-C-Client-Transition-pm/precutover >.
As metioned above this hash should not contain session-name , command_interface ,
method-timeout and other common options.

    my %precutover_pri=(
        mode                       => "rw_test",
        ## time_to_wait_in_rw(in seconds) is applicable only if mode selected is rw_test
        ## After this time a resync will be issued automatically
        time_to_wait_in_rw         => "240",
        nacltask_retry_tries_count => 1000000 ,
    );

=item C<< cutover_args       => \%cutover_pri >>

(Optional) Its a hash reference which contains the arguments specific to
transition cutover. It accepts all the arguments accepted by
L< NACL::C::Client::Transition->cutover()|lib-NACL-C-Client-Transition-pm/cutover >.
As metioned above this hash should not contain session-name , command_interface ,
method-timeout and other common options.

    my %cutover_pri=(
        nacltask_retry_tries_count => 1000000 ,
    );


=item C<< "method-timeout" => 10800 >>

(Optional) As component method-timeout, controls how long the command
will wait before completing. default timeout is 10800 seconds.

=item C<session-name>

Note that if this is not specified, a random name is generated using
L<NACL::GeneralUtils::random_name_generator()|lib-NACL-GeneralUtils-pm/random_name_generator>.
random_name_generator())

=back

=item C<<nacl_task_rng_args => { args_for_rng_here } >>

(Optional, used only if a random session name is generated)
If no session name is supplied, a random name is generated using
L<NACL::GeneralUtils::random_name_generator|lib-NACL-GeneralUtils-pm/random_name_genrator>
with two random words in it . This argument can be used to customize the
arguments to that method . For example, if we'd like to spwcify that the random
generated session name should be 255 characters length, then this could be:

    NACL::STask::Client::Transition->get_transitioned_volume_staggered(
        %other_opts,
        nacltask_rng_args => { length => '255' }
    )

See L<NACL::GeneralUtils::random_name_generator|lib-NACL-GeneralUtils-pm/random_name_generator>
for a description of the arguments it accepts.

=back

=back

=over

=item Exceptions

=over

=item C<NACL::C::Exceptions::Client::Transition::AlreadyExists>

This type of exception is thrown when an attempt is made to create a session that already exists.

=back

=back

=cut

sub get_transitioned_volume_staggered{

    $Log->enter();
    my ($pkg,@args) = @_;
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            nacltask_rng_args        => $pkg->_rng_args_validate_spec(),
            secondary_info           => { type => HASHREF,optional => 0 },
            primary_info             => { type => HASHREF,optional => 0 },
        },
        ignore_primary_keys => 1,
        allow_extra => 1,
    );

    my %sec_info = %{$opts{secondary_info}};
    my %pri_info = %{$opts{primary_info}};

    my %common_opts;
    $pkg->_copy_common_component_params_with_ci(
        source => \%opts,
        target => \%common_opts
    );

    my %prim_create_args        = (prim_create_args=>$pri_info{"create_args"});
    my %sec_create_args         = (sec_create_args=>$sec_info{"create_args"});
    my %prim_S_mode_credentials = (prim_S_mode_credentials=>$pri_info{"S_mode_credentials"});

    $pkg->get_transitioned_volume(
        %sec_info ,
        %common_opts,
        %prim_create_args,
        %prim_S_mode_credentials,
    );
    delete $pri_info{create_args}->{"source-host"};
    $pkg->get_transitioned_volume(
        %pri_info ,
        %common_opts,
        %sec_create_args
    );

    $Log->exit();
    return 1;

}## end sub get_transitioned_volume_staggered

=head2 create

  my $trans_obj = NACL::STask::Client::Transition->create(
        command_interface         => $Client,
        "session-name"            => $session_name,
        %other_options
  );

  or

  my $trans_obj = NACL::STask::Client::Transition->create(
        command_interface         => $Client,
        %other_options
  );

(Class method) Create a transition session.

Note that the argument C<session-name> is optional for this method.
If this is not specified, a random name is generated using
L<NACL::GeneralUtils::random_name_generator|lib-NACL-GeneralUtils-pm/random_name_generator>.
(See C<nacltask_rng_args> for how to customize arguments to random_name_generator())

=over

=item Options

=over

=item C<< "method-timeout" => 60 >>

(Optional) As component method-timeout, controls how long the command
will wait before completing. Default timeout is 60 seconds.

=item C<< nacltask_rng_args => { args_for_rng_here } >>

(Optional, used only if a random session name is generated)
If no session name is supplied, a random name is generated using
L<NACL::GeneralUtils::random_name_generator|lib-NACL-GeneralUtils-pm/random_name_generator>
with two random words in it . This argument can be used to customize the arguments
to that method. For example, if we'd like to specify that the random generated
session name should be 255 characters length, then this could be:

    NACL::STask::Client::Transition->create(
        %other_opts,
        nacltask_rng_args => { length => '255' }
    );

See L<NACL::GeneralUtils::random_name_generator|lib-NACL-GeneralUtils-pm/random_name_generator>
for a description of the arguments it accepts.

All of the other various options to
L<< NACL::C::Client::Transition->create|lib-NACL-C-Client-Transition-pm/create >>

=item Arguments defaulted/auto-determined

=over

=item C<session-name>

Note that if this is not specified, a random name is generated using
L<NACL::GeneralUtils::random_name_generator|lib-NACL-GeneralUtils-pm/random_name_generator>.
(See C<nacltask_rng_args> for how to customize arguments to random_name_generator())

=back

=back

=back

=over

=item Exceptions

=over

=item C<NACL::C::Exceptions::Client::Transition::AlreadyExists>

This type of exception is thrown when an attempt is made to create a session that already exists.

=back

=back

=cut

sub create {

    $Log->enter();
    my ($pkg, @args) = @_;
    my %opts = $pkg->_common_validate_with(
        params => \@args,
        additional_spec => {
            nacltask_rng_args   => $pkg->_rng_args_validate_spec(),
            "session-name"      => { optional => 1 },
            "group-name"        => { optional => 1 },
        },
        allow_extra => 1
    );

    my $rng_args = delete $opts{nacltask_rng_args};
    my $random_name_generated;
    if ( !defined $opts{"session-name"} ) {
        $opts{"session-name"} = $pkg->random_name_generator(%$rng_args);
        $random_name_generated = 1;
    }

    my $self;
CREATE: {
        try {
            $self = $pkg->SUPER::create(%opts);
        } catch NACL::C::Exceptions::Client::Transition::AlreadyExists with {
            my $exception = shift;
            if ($random_name_generated) {
                $Log->comment( 'Random name collision, regenerate a '
                                . 'new name and retry' );
                $opts{"session-name"} = $pkg->random_name_generator(%$rng_args);
                no warnings qw(exiting);
                redo CREATE;
            } else {
                $Log->trace('The session already exists');
                $Log->exit();
                $exception->throw();
            }
        };
    }
    $Log->exit();
    return $self;

} ## end sub create

=head2 start

  NACL::STask::Client::Transition->start(
        command_interface         => $Client,
        "session-name"            => $session_name,
        'method-timeout'          => 10800,
        nacltask_wait             => $boolean,     # default '1'
        polling_interval          => $interval,    # default '10'
        %other_options
    );

  or

  $trans_obj->start(
        nacltask_wait             => $boolean,     # default '1'
        polling_interval          => $interval,    # default '10'
        'method-timeout'          => 10800,
  );

(Class or instance method) Runs start on a transition session.This method provides
additional service of waiting till the session status reaches copy-update.

=over

=item Options

=over

=item C<< "method-timeout" => 10800 >>

(Optional) As component method-timeout, controls how long the command
will wait before the session reaches Copy-update status. Default timeout is 10800 seconds.

=item C<< nacltask_wait => 0|1 >>

(Optional) If 1 (the default), wait for the session to reach Copy-update status
(see "wait_for_operation" method, below).  If 0, do not wait any longer than
necessary.

=item C<< polling_interval => $interval >>

(Optional)
When nacltask_wait is provided, this is the interval at which session status will be
polled until it reaches Copy-update . Default value is 10 seconds.

All of the other various options to
L<< NACL::C::Client::Transition->start|lib-NACL-C-Client-Transition-pm/start >>

=back

=back

=cut

sub start {

    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            nacltask_retry_tries_count => { type => SCALAR, default => 3 },
            nacltask_retry_sleep_time  => { type => SCALAR, default => 10 },
            nacltask_retry_num_times_code_ran =>
                { type => SCALARREF, optional => 1 },
            nacltask_wait => { type => SCALAR, default => 1 },
            polling_interval    => { type => SCALAR,  optional => 1 },
            "parsed-output"     => { optional => 1 },
            "run-background"    => { optional => 1 },
            "group-name"        => { optional => 1 },
        },
        allow_extra => 1,
    );

    my $wait = delete $opts{nacltask_wait};
    my %retry_params;
    $pkg_or_obj->_move_retry_params(
        source => \%opts,
        target => \%retry_params
    );

    my %wait_opts;
    $pkg_or_obj->_hash_move(
        source => \%opts,
        target => \%wait_opts,
        move   => [qw(polling_interval)]
    );

    nacl_method_retry(
        code        => sub { $pkg_or_obj->SUPER::start(%opts) },
        exceptions  => 'NACL::APISet::Exceptions::ResponseException',
        %retry_params,
    );

    if ($wait) {
        my %primary_keys;
        $pkg_or_obj->_copy_primary_keys(
            source => \%opts,
            target => \%primary_keys
        );
        my %common_opts;
        $pkg_or_obj->_copy_common_component_params_with_ci(
            source => \%opts,
            target => \%common_opts
        );
        $pkg_or_obj->wait_for_operation(
            %primary_keys,
            %common_opts,
            %wait_opts,
            till_value         => 'Copy-update',
            attribute_to_check => "session_status",
        );
    }
    $Log->exit();

} ## end sub start

=head2 precutover

  NACL::STask::Client::Transition->precutover(
        command_interface         => $Client,
        "session-name"            => $session_name,
        'method-timeout'          => 10800,
        nacltask_wait             => $boolean,     # default '1'
        polling_interval          => $interval,    # default '10'
        %other_options
    );

  or

  $trans_obj->precutover(
        nacltask_wait             => $boolean,     # default '1'
        polling_interval          => $interval,    # default '10'
        'method-timeout'          => 10800,
  );

(Class or instance method) Runs precutover on a transition session.This method when
run with mode=~/rw_test/i waits till session status reaches Read-write, else when
ran with ro_test or no_test, waits till session status reaches Copy-update

=over

=item Options

=over

=item C<< "method-timeout" => 10800 >>

(Optional) As component method-timeout, controls how long the command
will wait before the session status reaches Read-write/Copy-update . Default timeout is 10800 seconds.

=item C<< nacltask_wait => 0|1 >>

(Optional) If 1 (the default), wait for the session to reach Read-write/Copy-update status
(see "wait_for_operation" method, below).  If 0, do not wait any longer than
necessary.

=item C<< polling_interval => $interval >>

(Optional)
When nacltask_wait is provided, this is the interval at which session status will be
polled until it reaches Read-write/Copy-update . Default value is 10 seconds.

All of the other various options to
L<< NACL::C::Client::Transition->precutover|lib-NACL-C-Client-Transition-pm/precutover >>

=back

=back

=cut

sub precutover {

    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            nacltask_retry_tries_count => { type => SCALAR, default => 3 },
            nacltask_retry_sleep_time  => { type => SCALAR, default => 10 },
            nacltask_retry_num_times_code_ran =>
                { type => SCALARREF, optional => 1 },
            nacltask_wait => { type => SCALAR, default => 1 },
            polling_interval    => { type => SCALAR,  optional => 1 },
            "parsed-output"     => { optional => 1 },
            "run-background"    => { optional => 1 },
            "group-name"        => { optional => 1 },
        },
        allow_extra => 1,
    );
    my $wait = delete $opts{nacltask_wait};
    my %retry_params;
    $pkg_or_obj->_move_retry_params(
        source => \%opts,
        target => \%retry_params
    );

    my %wait_opts;
    $pkg_or_obj->_hash_move(
        source => \%opts,
        target => \%wait_opts,
        move   => [qw(polling_interval)]
    );

    nacl_method_retry(
        code        => sub { $pkg_or_obj->SUPER::precutover(%opts) },
        exceptions  => 'NACL::APISet::Exceptions::ResponseException',
        %retry_params,
    );

    if ($wait) {
        my %primary_keys;
        $pkg_or_obj->_copy_primary_keys(
            source => \%opts,
            target =>\%primary_keys
        );
        my %common_opts;
        $pkg_or_obj->_copy_common_component_params_with_ci(
            source => \%opts,
            target => \%common_opts
        );

        my $till_value;
        if($opts{'mode'} =~ /rw_test/i){
            $till_value = "Read-write";
        }else{
            $till_value = "Copy-update";
        }

        $pkg_or_obj->wait_for_operation(
            %primary_keys,
            %common_opts,
            %wait_opts,
            till_value         => $till_value,
            attribute_to_check => "session_status",
        );
    }
    $Log->exit();

}

=head2 resync

  NACL::STask::Client::Transition->resync(
        command_interface         => $Client,
        "session-name"            => $session_name,
        'method-timeout'          => 10800,
        nacltask_wait             => $boolean,     # default '1'
        polling_interval          => $interval,    # default '10'
        %other_options
    );

  or

  $trans_obj->resync(
        nacltask_wait             => $boolean,     # default '1'
        polling_interval          => $interval,    # default '10'
        'method-timeout'          => 10800,
    );

(Class or instance method) Runs resync on a transition session.This method provides
additional service of waiting till the session reaches Precutover/Copy-update state/status.

=over

=item Options

=over

=item C<< "method-timeout" => 10800 >>

(Optional) As component method-timeout, controls how long the command
will wait before the session reaches Precutover/Copy-update state/status. Default timeout is 10800 seconds.

=item C<< nacltask_wait => 0|1 >>

(Optional) If 1 (the default), wait for the session to reach Precutover/Copy-update state/status
(see "wait_for_operation" method, below). If 0, do not wait any longer than
necessary.

=item C<< polling_interval => $interval >>

(Optional)
When nacltask_wait is provided, this is the interval at which session status will be
polled until it reaches Precutover/Copy-update state/status . Default value is 10 seconds.

All of the other various options to
L<< NACL::C::Client::Transition->resync|lib-NACL-C-Client-Transition-pm/resync >>

=back

=back

=cut

sub resync {

    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            nacltask_retry_tries_count => { type => SCALAR, default => 3 },
            nacltask_retry_sleep_time  => { type => SCALAR, default => 10 },
            nacltask_retry_num_times_code_ran =>
                { type => SCALARREF, optional => 1 },
            nacltask_wait => { type => SCALAR, default => 1 },
            polling_interval    => { type => SCALAR,  optional => 1 },
            "parsed-output"     => { optional => 1 },
            "run-background"    => { optional => 1 },
            "group-name"        => { optional => 1 },
        },
        allow_extra => 1,
    );

    my $wait = delete $opts{nacltask_wait};
    my %retry_params;
    $pkg_or_obj->_move_retry_params(
        source => \%opts,
        target => \%retry_params
    );

    my %wait_opts;
    $pkg_or_obj->_hash_move(
        source => \%opts,
        target => \%wait_opts,
        move   => [qw(polling_interval)]
    );

    nacl_method_retry(
        code        => sub { $pkg_or_obj->SUPER::resync(%opts) },
        exceptions  => 'NACL::APISet::Exceptions::ResponseException',
        %retry_params,
    );

    if ($wait) {
        my %primary_keys;
        $pkg_or_obj->_copy_primary_keys(
            source => \%opts,
            target =>\%primary_keys
        );
        my %common_opts;
        $pkg_or_obj->_copy_common_component_params_with_ci(
            source => \%opts,
            target => \%common_opts
        );

        $pkg_or_obj->wait_for_operation(
            %primary_keys,
            %common_opts,
            %wait_opts,
            till_value         => "Copy-update",
            attribute_to_check => "session_status",
        );
    }
    $Log->exit();

}

=head2 cutover

  NACL::STask::Client::Transition->cutover(
        command_interface         => $Client,
        "session-name"            => $session_name,
        'method-timeout'          => 10800,
        nacltask_wait             => $boolean,     # default '1'
        polling_interval          => $interval,    # default '10'
        %other_options
    );

  or

  $trans_obj->cutover(
        nacltask_wait             => $boolean,     # default '1'
        polling_interval          => $interval,    # default '10'
        'method-timeout'          => 10800,
  );

(Class or instance method) Runs cutover on a transition session.This method provides
additional service of waiting till the session reaches Completed state.

=over

=item Options

=over

=item C<< "method-timeout" => 10800 >>

(Optional) As component method-timeout, controls how long the command
will wait before the session reaches Completed state. Default timeout is 10800 seconds.

=item C<< nacltask_wait => 0|1 >>

(Optional) If 1 (the default), wait for the session to reach Completed state
(see "wait_for_operation" method, below).  If 0, do not wait any longer than
necessary.

=item C<< polling_interval => $interval >>

(Optional)
When nacltask_wait is provided, this is the interval at which session state will be
polled until it reaches Completed . Default value is 10 seconds.

All of the other various options to
L<< NACL::C::Client::Transition->cutover|lib-NACL-C-Client-Transition-pm/cutover >>

=back

=back

=cut

sub cutover {

    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            nacltask_retry_tries_count => { type => SCALAR, default => 3 },
            nacltask_retry_sleep_time  => { type => SCALAR, default => 10 },
            nacltask_retry_num_times_code_ran =>
                { type => SCALARREF, optional => 1 },
            nacltask_wait => { type => SCALAR, default => 1 },
            polling_interval    => { type => SCALAR,  optional => 1 },
            "parsed-output"     => { optional => 1 },
            "run-background"    => { optional => 1 },
            "group-name"        => { optional => 1 },
        },
        allow_extra => 1,
    );
    my $wait = delete $opts{nacltask_wait};
    my %retry_params;
    $pkg_or_obj->_move_retry_params(
        source => \%opts,
        target => \%retry_params
    );

    my %wait_opts;
    $pkg_or_obj->_hash_move(
        source => \%opts,
        target => \%wait_opts,
        move   => [qw(polling_interval)]
    );

    nacl_method_retry(
        code        => sub { $pkg_or_obj->SUPER::cutover(%opts) },
        exceptions  => 'NACL::APISet::Exceptions::ResponseException',
        %retry_params,
    );

    if ($wait) {
        my %primary_keys;
        $pkg_or_obj->_copy_primary_keys(
            source => \%opts,
            target => \%primary_keys
        );
        my %common_opts;
        $pkg_or_obj->_copy_common_component_params_with_ci(
            source => \%opts,
            target => \%common_opts
        );

        $pkg_or_obj->wait_for_operation(
            %primary_keys,
            %common_opts,
            %wait_opts,
            till_value         => "Completed",
            attribute_to_check => "session_state",
        );
    }
    $Log->exit();

}

=head2 precheck

  NACL::STask::Client::Transition->precheck(
        command_interface         => $Client,
        "session-name"            => $session_name,
        'method-timeout'          => 10800,
        %other_options
    );

  or

  $trans_obj->precheck(
        'method-timeout'          => 10800,
  );

(Class or instance method) Runs precheck on a transition session.

=over

=item Options

=over

=item C<< "method-timeout" => 10800 >>

(Optional) As component method-timeout, controls how long the command
will wait before precheck completes. Default timeout is 10800 seconds.

All of the other various options to
L<< NACL::C::Client::Transition->precheck|lib-NACL-C-Client-Transition-pm/precheck >>

=back

=back

=cut

sub precheck {

    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            nacltask_retry_tries_count => { type => SCALAR, default => 3 },
            nacltask_retry_sleep_time  => { type => SCALAR, default => 10 },
            nacltask_retry_num_times_code_ran =>
                { type => SCALARREF, optional => 1 },
            "parsed-output"     => { optional => 1 },
            "run-background"    => { optional => 1 },
            "group-name"        => { optional => 1 },
        },
        allow_extra => 1,
    );

    my %retry_params;
    $pkg_or_obj->_move_retry_params(
        source => \%opts,
        target => \%retry_params
    );

    nacl_method_retry(
        code        => sub { $pkg_or_obj->SUPER::precheck(%opts) },
        exceptions  => 'NACL::APISet::Exceptions::ResponseException',
        %retry_params,
    );

    $Log->exit();

}

=head2 wait_for_operation

  NACL::STask::Client::Transition->wait_for_operation(
        command_interface         => $Client,
        "session-name"            => $session_name,
        till_value                => "Completed",
        attribute_to_check        => "session_state",
        polling_interval          => $interval,    # default '10'
        'method-timeout'          => 10800,
  );

  or

  $trans_obj->wait_for_operation(
        till_value                => 'Read-write',
        attribute_to_check        => "session_status",
        polling_interval          => $interval,    # default '10'
        'method-timeout'          => 10800,
  );

(Class or instance method) This method is used to wait till the session reaches a
desired value specified by "till_value" field and the attribute which is getting polled
is "attribute_to_check" .If the session reaches "Copy-unknown" or "Delete" state
then an NACL::Exceptions::UnexpectedState exception is raised.

=over

=item Options

=over

=item C<< "till_value" => $state >>

(Required) This is the session's "attribute_to_check" field which will be waited for until method-timeout.

=item C<< "attribute_to_check" => $field >>

(Required) This is the field which will be polled for until method-timeout or until till_value.

=item C<< "method-timeout" => 10800 >>

(Optional) As component method-timeout, controls how long the command
will wait before the session reaches the state specified by "till_value" field.
Default timeout is 10800 seconds.

=item C<< polling_interval => $interval >>

(Optional)
When nacltask_wait is provided, this is the interval at which attribute_to_check will be
polled until it reaches the state specified by "till_value" field. . Default value is 10 seconds.

=back

=back

=cut

sub wait_for_operation {

    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    my %opts = $pkg_or_obj->_common_validate_with(
        params => \@args,
        additional_spec   => {
            polling_interval         => { type => SCALAR, default => '10' },
            attribute_to_check       => { type => SCALAR, optional => 0 },
            till_value               => { type => SCALAR, optional => 0 },
        },
    );

    my $pkg = $pkg_or_obj->get_package_name();

    # Hard-coding the timeout to 3 hours if user does not specify method-timeout
    $opts{'method-timeout'} = '10800' if(not defined $opts{'method-timeout'});
    # Pulling out the attribute to poll for
    my $attribute_to_check   = delete $opts{"attribute_to_check"};

    my @invalid_states=(qr/Copy\-unknown/i,"Delete");
    $pkg->wait_on_attribute(
        %opts,
        attribute_to_check => $attribute_to_check,
        invalid_states     => \@invalid_states,
    );

} ## end sub wait_for_operation

1;
