#
# Copyright (c) 2013-16 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Cluster Task Module
## @author dl-nacl-dev@netapp.com
## @status shared
## @pod here

package NACL::STask::Cluster;

use strict;
use warnings;
use base qw(NACL::C::Cluster NACL::STask::STask);
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use NATE::Exceptions::Argument qw(:try);
use NACL::Exceptions::VerifyFailure;
use NACL::Exceptions::ClusterRingError;
use NATE::BaseException;
use NACL::Exceptions::ClusterHealthCheckError;
use NACL::APISet::Exceptions::CommandFailedException;
use NACL::APISet::Exceptions::ResponseException;
use NACL::C::Exceptions::Cluster::PreClusterMode;
use NACL::C::Exceptions::ClusterIdentity::PreClusterMode;
use NACL::C::Exceptions::TimeNotSynchronised qw(:try);
use NACL::STask::Exceptions::NodeAlreadyInCluster;
use NACL::Exceptions::HighAvailabilityError qw(:try);

use NACL::C::Interface;
use NACL::C::Vserver;
use NACL::C::ClusterPeer;
use NACL::CS::Spm;
use NACL::CS::ClusterIdentity;
use NACL::CS::ClusterRing;
use NACL::CS::EventLog;
use NACL::CS::NetworkPort;
use NACL::C::SystemNodeDate;
use NACL::STask::NetworkRoutingGroupsRoute;
use NACL::STask::NetworkRoute;
use NACL::GeneralUtils qw(nacl_method_retry);

use Data::Dumper;

use NACL::STask::Node;
use NATE::Time qw(timeout2time);

use NACL::ComponentUtils qw(_optional_scalars);
use Params::Validate
    qw(validate_with validate :types );
use Scalar::Util qw(blessed);

use Class::MethodMaker [
    new    => [ '-hash', '-init', 'new' ],
    scalar => [
        {   '-default_ctor' => sub {
                my $self = $_[0];
                my $name = $self->get_one_identity_attr('name');
                return $name;
                }
        },
        'name'
    ]
];

=head1 NAME

NACL::STask::Cluster

=head1 DESCRIPTION

C<NACL::STask::Cluster> provides a number of well-defined but
potentially complex or multi-step methods related to Cluster in
ONTAP.

It builds on top of, and is a derived class of C<NACL::C::Cluster>,
and so it also provides methods that are more in the scope of
individual Cluster-related commands.  See C<NACL::C::Cluster> for
details.  This also means that a C<NACL::STask::Cluster> object may
generally be used in place of a component object.

=head1 ATTRIBUTES

=head2 command_interface

(Required) As C<NACL::C::Cluster>.  A component object that
represents the node to which to send commands.

=head2 name

(Optional) The name of the cluster.

=head1 METHODS

=cut

=head2 cluster_health_check

    NACL::STask::Cluster->cluster_health_check(
                    'command_interface'  => $command_interface,
                    'time'               => $time,
                    'health-check-mode'  => "simple|complete"
        );

(Class method)This method is called for CMode only .
It will do the following steps:

1)Check the health of all the nodes of cluster are true.
2)Verify the network connectivity to remote cluster interfaces, and
  perform the RPC server check.
3)Verify the Peer relationship using cluster peer ping
4)Verify the cluster's ring-replication status and their masters.
5)Verify the state of processes managed by SPM.
6)Check for any critrical or emergency level events for the specified amount of time

This method will check for the health of all nodes in a cluster. It will
throw exception if any of the node health status is 'False' . From
one node , it will try to ping all the other nodes in the cluster to ensure
connectivity. It will throw exception if node is unreachable. Next , it
checks overall cluster health by verifying that all of the individual cluster
rings are healthy. Along with this , it will check for critical or emergency
level events and the state of process managed by SPM . Exception will be thrown
if any critical/emergency event occur in particular period op time or if any
process is not in running state.This method works for a non-HA setup as well
where there are no peers.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) A component object that represents the host to which to send commands.

=item C<< 'health-check-mode' => $check_mode >>

(optional) the health check mode to be performed. Value may be simple or complete
simple health performs the check of only the health of node
complete health check performs the other checks also like
ping, cluster ring replication as mentioned above from steps 1 to 6

Default : complete

=item C<< nacltask_ignore => 0|1 >>

(Optional) specifies an action to be taken if any of the 4 steps failed.

     $action
    - if set to  "0"(default), an exception is raised.
    - if set to  "1" , then it will go to next step

=item C<< time => $time >>

(Optional) specifies the time range to check for event logs.
            Time should be in the format MM/DD/YYYY HH:MM:SS
            Range can be specified like 11/4/2011 04:54:38..11/4/2011 04:54:38 or
            >11/4/2011 04:54:38 or <11/4/2011 04:54:38
            By default , It will take the time when method started execution.
            ie >$current_date time

=item C<< node => $node >>

(Optional)
Cluster health check by default is performed on all the nodes in
the cluster. If node option is provided then the cluster health check performs
health check on only the node and its partner.

=item C<< "method-timeout" => 300 >>

(Optional) As component method-timeout, controls how long the command
will wait before completing.  However, the default timeout has been
raised to 300 seconds.

=back

=back

=over

=item Exceptions

=over

=item C<NACL::APISet::Exceptions::CommandFailedException>

This type of exception is failed when pinging to a cluster fails.

=item C<NACL::Exceptions::ClusterHealthCheckError>

This type of exception is thrown when there is any failure while performing cluster health check.

=back

=back

=cut

sub cluster_health_check {
    $Log->enter() if $may_enter;
    my $pkg = shift;
    my ( @peer_nodes_in_cluster, @error, $mesg );
    my %opts = validate_with(
        params      => \@_,
        spec        => {
              %{ NACL::C::Component->_common_validate_spec() },
              'health-check-mode' => { regex => qr/^simple|complete$/ , default => "complete" },
              'node' => {type => SCALAR|OBJECT|UNDEF, optional => 1},
	      'method-timeout'     => { type => SCALAR , default => '300'},
        },
        allow_extra => 1,
    );
    my $command_interface = $opts{command_interface};
    my $time              = $opts{time}
        || ">"
        . $pkg->_get_sysdate( command_interface => $command_interface );

    my $check_mode = delete $opts{'health-check-mode'};

    my (@nodes, $node, $filter) ;
    $node = $opts{node};
    
    try{
       if(defined $node){
         if(blessed $node){
           push @nodes, $node->node() ;
           push @nodes, NACL::STask::Node->get_partner(
                     command_interface    => $node,
                );
         }else{
           push @nodes,  $node;
           my $node_obj = NACL::C::Node->new(node => $node);
           push @nodes, NACL::STask::Node->get_partner(
                     command_interface    => $node_obj,
                );
         }
         $filter = { filter => {node => join(',',@nodes)}} ;
      }
    }catch NACL::Exceptions::HighAvailabilityError with{
        ## just suppressing this exception
        $Log->debug('HA partner of node is not available');
    };

    my @nodes_in_cluster
        = NACL::CS::Cluster->fetch(
              command_interface => $command_interface,
               %{$filter}
        );
    foreach (@nodes_in_cluster) {

        if ( $_->health() !~ /true/ ) {
            $mesg = "Health of node " . $_->node() . "is" . $_->health();
            push( @error, $mesg );
        }
    }

    if($check_mode =~ /complete/i){
        # Check network connectivity to remote cluster interfaces,
        # and perform the RPC server check

        my %extra_opts;
        $pkg->_hash_copy(
                source => \%opts,
                target => \%extra_opts,
                copy   => [qw(method-timeout)]
        );

        foreach (@nodes_in_cluster) {
           try {
                NACL::C::Cluster->ping_cluster(
                   command_interface => $command_interface,
                   node              => $_->node(),
                   'skip-rpccheck'   => 'false',
                    %extra_opts,
                );
           }
           catch NACL::APISet::Exceptions::CommandFailedException with {
                my $err_msg = shift;
                $err_msg
                   = "ping-cluster from" . $_->node() . " failed\n" . $err_msg;
                push( @error, $err_msg );
           };
        }
        # Check Peer relationship,

        try {

            #Checking whether peer relation ship is established or not
            @peer_nodes_in_cluster = NACL::C::ClusterPeer->find(
                                command_interface => $command_interface );
        } ## end try
        catch NACL::Exceptions::NoElementsFound with {
           $Log->exit() if $may_exit;
        };
        # If intercluster relationship is present , then do peer ping()
        if ( scalar @peer_nodes_in_cluster ) {
            my @cp = NACL::C::ClusterPeer->ping(
                command_interface => $command_interface, );
            foreach my $cp (@cp) {
                if ( $cp->status() !~ /interface_reachable/ ) {
                  $mesg
                        = "cluster-peer-ping from "
                        . $cp->originating_node() . "to "
                        . $cp->destination_node() . "is "
                        . $cp->status();
                  push( @error, $mesg );
                }
            } ## end foreach my $cp (@cp)
        }
        # Check cluster's ring-replication status and their masters
        $pkg->check_rings( command_interface => $command_interface );

        # Verify the state of processes managed by SPM
        foreach (@nodes_in_cluster) {
          my @spm_details = NACL::CS::Spm->fetch(
                command_interface => $command_interface,
                filter            => { node => $_->node(), },
                requested_fields  => [qw(state)]
          );
          foreach my $entry (@spm_details) {
             if ( $entry->{state} !~ /running/ ) {
                 $mesg
                  = $entry->{name}
                  . " process of "
                  . $_->node()
                  . "is not running";
                push( @error, $mesg );
             }
          }
        }
        # Check for any critrical or emergency level events for the specified amount of time
        my @events_log=();
        try {
            @events_log = NACL::CS::EventLog->fetch(
                'command_interface' => $command_interface,
                'filter'            => {
                    severity => "CRITICAL|EMERGENCY",
                    time     => $time
                },
                'allow_empty' => 1,
            );
        }
        catch NACL::C::Exceptions::TimeNotSynchronised with {
                my $ex = shift;
                $Log->debug(
                        "Ignoring the time does not appear to be synchronised across this cluster error"
                );
                @events_log = $ex->state_objects();
        };

        if (@events_log) {
                foreach my $log_name (@events_log) {
                $mesg .= $log_name->event();
                }
                $mesg
                   = "EventLog shows the following CRITICAL messages \n" . $mesg;
                push( @error, $mesg );
        }
    }
    $Log->exit() if $may_exit;
    if (@error) {
        if ( $opts{nacltask_ignore} ) {
            $Log->warn(@error);
        } else {
            $Log->debug(@error);
            NACL::Exceptions::ClusterHealthCheckError->throw(@error);
        }
    }

}

sub _get_sysdate {
    $Log->enter() if $may_enter;
    my $self = shift;
    require NACL::CS::ClusterDate;
    my %opts = validate_with(
        params => \@_,
        spec => { 'command_interface' => { type => OBJECT, optional => 1 } },
        allow_extra => 1
    );
    my $ci = $opts{command_interface} || $self->command_interface();
    $Log->exit() if $may_exit;
    return NACL::CS::ClusterDate->fetch(
        'command_interface' => $ci,
        'requested_fields'  => ['date'],
    )->date();
}

=head2 check_rings

  $cluster->check_rings(
    "method-timeout"  => $timeout,
  );

  NACL::STask::Cluster->check_rings(
    "method-timeout"  => $timeout,
    command_interface => $command_interface,
  );

(Class or instance method) This method checks overall cluster health
by verifying that all of the individual cluster rings are healthy.
Note: All entries which share the same unitname comprise one ring.

  Given the following Cmode command output from "cluster ring show":

  node  unitname online    epoch master local db-epoch db-trnxs num-online
  ----- -------- ------    ----- ------ ----- -------- -------- ----------
  vsim5 mgmt     master    132   vsim5  vsim5 132      9        2
  vsim5 vldb     master    96    vsim5  vsim5 96       1        2
  vsim5 vifmgr   master    96    vsim5  vsim5 96       54       2
  vsim5 bcomd    master    96    vsim5  vsim5 96       46       2
  vsim6 mgmt     secondary 132   vsim5  vsim6 132      9        2
  vsim6 vldb     secondary 96    vsim5  vsim6 96       1        2
  vsim6 vifmgr   secondary 96    vsim5  vsim6 96       54       2
  vsim6 bcomd    secondary 96    vsim5  vsim6 96       46       2

  Perform the following checks:

    1) Must be an entry/unitname for each node in the cluster
    2) The num_online counter must == number of nodes in the cluster
    3) For each entry, node must == local
    4) Master is the same for each entry that has the same unitname
       (though not necessarily the same for ALL entries), and not "-"
    5) if master == node, online must == "master"
    6) if master != node, online must == "secondary"
    7) The following values, epoch, db_epoch and num-online
       must not be "-" and must be <= the corresponding value
       for the master of the ring

If any of the previous checks fail then the following exception
will be thrown:

L<NACL::Exceptions::ClusterRingError|lib-NACL-Exceptions-ClusterRingError-pm>

Contained within the text of the exception will be all of the checks which have
failed.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< "method-timeout"=>$timeout >>

(Optional) How long in seconds to wait for the component calls to wait.
Defaults to 60 seconds.

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=back

=back

=over

=item Exceptions

=over

=item C<NACL::Exceptions::ClusterRingError>

This type of exception is thrown when there is any failure while fetching ClusterRing information.

=back

=back

=cut

sub check_rings {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;

    my %orig_opts  = (
        command_interface => undef,
        @_
    );

    my @verify_params = %orig_opts;
    my %opts          = $pkg_or_obj->_common_validate_with(
        params      => \@verify_params,
        additional_spec => {
             'node' => {type => SCALAR|OBJECT|UNDEF , optional => 1}
        },
        allow_extra => 1
    );

    if ( ref($pkg_or_obj) && defined $pkg_or_obj->{'node'} ) {
         delete $opts{'node'};
    }

    my %common_opts;
    $pkg_or_obj->_copy_common_component_params(
        source => \%opts,
        target => \%common_opts
    );

    # need to retrieve command_interface after
    # _common_validate_with() because if
    # we are calling as an instance method, it
    # won't be filled in before then
    my $command_interface = $opts{command_interface};

    my @counters = qw( num_online epoch db_epoch db_trnxs );
    my %master;    # hash of master entry data keyed by ring
    my %node;      # hash of nodes keyed by ring
    my @error;     # any errors we've encountered
    my @nodes ;
    my $node_filter;
    my @nodes_arr ;
    if (defined $opts{node}) {
        if(blessed $opts{node}){
            $node_filter = { filter => {node => $opts{node}->node() }};
            push @nodes_arr, $opts{node}->node() ;
        } else {
            my $node_obj = NACL::C::Node->new(node => $opts{node});
            $node_filter = { filter => {node => $node_obj->node() } };
            push @nodes_arr, $opts{node} ;
        }
    }

    # Get nodes() from Component Class
    my $pkg = ref($pkg_or_obj) || $pkg_or_obj;
    @nodes = $pkg->nodes(
        command_interface => $command_interface,
        %common_opts
    );

    @nodes_arr = @nodes if(!defined $opts{node});
    my @rings;
    my @secondary_rings;

    try {
        ## w.r.t Burt963397 the counters values may vary as expected,
        ## so fetch cluster rings through filter online => secondary
	## for later comparisons of counter.

	@secondary_rings = NACL::CS::ClusterRing->fetch(
            command_interface => $command_interface,
            %common_opts,
            filter      => {online => "secondary"},
            allow_empty => 1
        );

	@rings = NACL::CS::ClusterRing->fetch(
            command_interface => $command_interface,
            %common_opts,
            %{$node_filter}
        );
    } ## end try
    catch NACL::APISet::Exceptions::ResponseException with {

        # Note: the error from BR is resulting in a ResponseException
        #       but in RR, the same condition is not.  Submitted
        #       BURT 505208 to address this
        my $exp = shift;

        # In case we can't parse this, at least report the raw
        # exception text
        my $msgstr = $exp->text();

        # We'd like to narrow this down a bit for the user as the
        # ResponseException contains bits of command output as well
        # as the error string
        if ( $msgstr =~ /(rdb_ring_info: .+ ring .+ for .*)\n/i ) {

            # In this case $raw should at least contain only the error text
            my $raw = $1;
            if ( $raw =~ /rdb_ring_info: .+? ring .+? for (\S+?) /i ) {
                $msgstr = "ring:$1, $raw";
            } else {
                $msgstr = $raw;
            }
        } ## end if ( $msgstr =~ /(rdb_ring_info: .+ ring .+ for .*)\n/i)
        $msgstr = 'node:' . $command_interface->node() . " $msgstr";
        $Log->debug($msgstr);
        NACL::Exceptions::ClusterRingError->throw( $msgstr . "\n" );
    };

    # possible values for online are "-", "master", "offline", or "secondary"
    @rings = sort { $a->online() cmp $b->online() } @rings;

    # For each entry, do basic validation (fields != '-')
    # and does consistency checking:
    #   node == local
    #   if master == node then online == 'master'
    #   if master != node then online == 'secondary'
    #   num_online == number of nodes in cluster
    #   each ring reports the same master
    # Then save master ring counters to be used in further checking.
    my ($expected, $actual) ;
    foreach my $entry (@rings) {
        # Skip the "localhost" entry. See burt 780072.
        # Explanation from burt:
        # The "cluster ring show" shows the same row twice. Once with
        # "localhost" and once with  the real hostname.
        # Under normal conditions the row is displayed only once with the
        # proper hostname.
        # Under transitions (nodes coming online/offline and in/out of
        # cluster) the node name may be "localhost".
        # So the name was "localhost" and cluster ring show got that row. Then
        # the transition completed while "cluster ring show" was running and
        # it picked up the second row.
        next if ($entry->node() eq 'localhost');
        my $unitname = $entry->unitname();

        if ( !defined $node{$unitname} ) {
            $node{$unitname} = [ $entry->node() ];
        } else {
            push( @{ $node{$unitname} }, $entry->node() );
        }

        # Check local (node must equal local)
        if ( $entry->node() ne $entry->local() ) {
	    $expected = "Expected : node must be equal to \"local\" \nActual:";
            push( @error, $expected ) if(! grep( /$expected/, @error ));

            my $msgstr
                = 'node:'
                . $entry->node()
                . ' != local:'
                . $entry->local()
                . ' for ring:'
                . $entry->unitname();
            push( @error, $msgstr );
        } ## end if ( $entry->node() ne...

       # Check master and online (if master() == node(), online() == 'master')
        if ( $entry->master() eq '-' ) {
	    $expected = "Expected : master should not be null \nActual:";
	    push( @error, $expected ) if(! grep( /$expected/, @error ));

            my $msgstr
                = 'node:'
                . $entry->node()
                . ' master:-, master should not be -';
            push( @error, $msgstr );
        } elsif ( $entry->master() eq $entry->node() ) {
            if ( $entry->online() ne "master" ) {

                $expected = "Expected: for each ring if master == node, then "
	            ."online should be \"master\" \nActual:";
                push( @error, $expected ) if(! grep( /$expected/, @error ));

                my $msgstr
                    = 'node:'
                    . $entry->node()
                    . ' master:'
                    . $entry->master()
                    . ' online:'
                    . $entry->online()
                    . ' ring:'
                    . $entry->unitname();
                push( @error, $msgstr );
            } ## end if ( $entry->online() ...
        } else {
            if ( $entry->online() ne "secondary" ) {

                $expected = "Expected: node must be equal to \"secondary\" \nActual:" ;
		push( @error, $expected ) if(! grep( /$expected/, @error ));

                my $msgstr
                    = 'node:'
                    . $entry->node()
                    . ' master:'
                    . $entry->master()
                    . ' online:'
                    . $entry->online()
                    . ' ring:'
                    . $entry->unitname();
                push( @error, $msgstr );
            } ## end if ( $entry->online() ...
        } ## end else [ if ( $entry->master() ...

        # Check that each entry for a given ring reports the same master
        if ( defined $master{$unitname}{master} ) {
            if (   ( $entry->master() ne $master{$unitname}{master} )
                && ( $master{$unitname}{master} ne "-" ) )
            {
                $expected = "Expected: Each entry for a given ring "
	           . "should report the same master $master{$unitname}{master} "
		   ."\nActual:";
		push( @error, $expected ) if(! grep( /$expected/, @error ));

                my $msgstr
                    = 'node:'
                    . $entry->node()
                    . " ring:$unitname master:"
                    . $entry->master();
                push( @error, $msgstr );
            } ## end if ( ( $entry->master(...
        } else {
            $master{$unitname}{master} = $entry->master();
        }

        # Save master counters
        foreach my $counter (@counters) {

            # Check that counters aren't "-"
            if ( $entry->$counter() eq "-" ) {

                $expected = "Expected: ring counters should not be equal to - \nActual:";
		push( @error, $expected ) if(! grep( /$expected/, @error ));

                my $msgstr
                    = 'node:'
                    . $entry->node()
                    . " ring: $unitname, $counter:- $counter ";
                push( @error, $msgstr );
            } else {

                # Save value of "master" counter
                if ( !defined $master{$unitname}{$counter} ) {
                    if ( defined $master{$unitname} ) {
                        $master{$unitname}{$counter} = $entry->$counter();
                    }
                }

                if ( $counter eq "num_online" ) {
                    if ( $entry->$counter() ne scalar(@nodes) ) {
			$expected = "Expected: $counter should be equal to total nodes "
                                . "in the cluster which is "
				. scalar(@nodes)
				. "\nActual:" ;
			push( @error, $expected ) if(! grep( /$expected/, @error ));

                        my $msgstr
                            = 'node:'
                            . $entry->node()
                            . " ring: $unitname,"
                            . " num_online: ".$entry->$counter();
                        push( @error, $msgstr );
                    } ## end if ( $entry->$counter(...
                } ## end if ( $counter eq "num_online")
            } ## end else [ if ( $entry->$counter(...
        } ## end foreach my $counter (@counters)
    } ## end foreach my $entry (@rings)

    # Verify that the counters for each ring are <= the counters of the
    # master of that ring.
    foreach my $entry (@secondary_rings) {
        my $unitname = $entry->unitname();

        # All counters must be <= the master counter
        foreach my $counter (@counters) {

            # Since we are sorted by online, "master" entries will come first
            if ( defined $master{$unitname}{$counter} ) {
                if (   ( $entry->$counter() ne '-' )
                    && ( $entry->$counter() > $master{$unitname}{$counter} ) )
                {
		    $expected = "Expected: counters for each ring should be <= the counters of the master of that ring \nActual:" ;
                    push( @error, $expected ) if(! grep( /$expected/, @error ));

                    my $msgstr
                        = 'node:'
                        . $entry->node()
                        . " ring: $unitname, $counter:"
                        . $entry->$counter()
                        . " should be <= "
                        . $master{$unitname}{$counter};
                    push( @error, $msgstr );
                } ## end if ( ( $entry->$counter...
            } else {

                $expected = "Expected: each entry should have a master defined \nActual:" ;
                push( @error, $expected ) if(! grep( /$expected/, @error ));
                my $msgstr
                    = 'node:'
                    . $entry->node()
                    . " ring: $unitname, No master found for counter $counter";
                push( @error, $msgstr );
            } ## end else [ if ( defined $master{$unitname...
        } ## end foreach my $counter (@counters)
    } ## end foreach my $entry (@rings)

    # Make sure we have all of the ring entries for each node
    my @unitnames = ( keys %node );
    foreach my $node (@nodes_arr) {
        foreach my $unitname (@unitnames) {

            # verify that all nodes are in each ring
            my @sorted = sort @{ $node{$unitname} };
            $node{$unitname} = \@sorted;

            # Are all nodes represented for all rings?
            if ( !grep( /^$node$/, @{ $node{$unitname} } ) ) {

                $expected = "Expected: All nodes should be represented by rings \nActual:" ;
		push( @error, $expected ) if(! grep( /$expected/, @error ));

                my $msgstr
                    = 'node:' . $node . " No entry found for ring: $unitname";
                push( @error, $msgstr );
            }
        } ## end foreach my $unitname (@unitnames)
    } ## end foreach my $node (@nodes)

    $Log->exit() if $may_exit;
    if (@error) {
        $Log->debug(@error);
        NACL::Exceptions::ClusterRingError->throw(
            CORE::join( "\n", @error ) . "\n" );
    }
} ## end sub check_rings

=head2 identity

    $Cluster_task_obj->identity(%options);
                     (or)
    NACL::STask::Cluster->identity(
            command_interface => $command_interface,
                                 %options);

(Class or instance method) This method returns the state object of
ClusterIdentity component, corresponding to the cluster of which this node is
a member.

=over

=item options

=over

=item C<< filter=>{ attribute1=>$value1, attribute2=>$value2, ... } >>

(Optional) A hashref that limits the components returned to be only
those whose attributes have the given values (If possible, the
commands/requests sent will be optimized to limit discovery
similarly).  The attribute names should match the fields used in the
ComponentState object corresponding to ClusterIdentity Component.

=item C<< requested_fields=>['name1', 'name2', ...] >>

(Optional) An array reference containing the names of component state
fields that must be filled in; others are allowed (but not required)
to be undefined.  If the requested_fields option is undef or not
given, it is assumed that all component state fields should be filled
in.  An optimal component state implementation will limit the commands
it sends, their verbosity, their columns, etc., when doing so will
still fetch the given fields, but simpler component implementations
may just ignore this option and always query for all of the state of
this element that they can.

=item C<< allow_empty=>$boolean >>

(Optional) If false (the default) then throw a NACL::Exceptions::NoElementsFound
exception if no elements are found.  If true, then return undef or an empty list
if no elements are found.

=item C<< apiset_must=>$ruleset >>

(Optional) A hashref constraining which API (mode, interface, etc.) on
the given command_interface will be used.
(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional) A hashref suggesting which API (mode, interface, etc.) on
the given command_interface is preferred.
(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'method-timeout' => $timeout_in_seconds >>

(Optional) We can use this option to specify a large timeout value, which will
be used by all command(s) run, in case the command(s) take a long time to run
due to  large amounts of output.

Other options (if any) which are applicable to ClusterIdentity->fetch() method
as well
L<< NACL::CS::ClusterIdentity |lib-NACL-CS-ClusterIdentity-pm/fetch >>

=back

=back

=cut

sub identity {

    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;

    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@_,
        additional_spec => {
            %{ NACL::CS::ComponentState::_fetch_validate_spec() },
            _optional_scalars(qw(node)),
        },
        ignore_primary_keys => 1
    );
    delete $opts{node};
    my $cluster_id_state = NACL::CS::ClusterIdentity->fetch(%opts);

    $Log->exit() if $may_exit;
    return $cluster_id_state;

} ## end sub identity

=head2 get_one_identity_attr

    $Cluster_task_obj->get_one_identity_attr(attribute_name,%other_options);
                     (or)
    NACL::STask::Cluster->get_one_identity_attr(attribute_name,
            command_interface => $command_interface,
                                 %other_options);

(Class or instance method) This method returns the value of the attribute
given in attribute_name.The attribute names should match the fields used in the
ComponentState object corresponding to ClusterIdentity Component.

=over

=item other_options

=over

=item C<< apiset_must=>$ruleset >>

(Optional) A hashref constraining which API (mode, interface, etc.) on
the given command_interface will be used.
(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional) A hashref suggesting which API (mode, interface, etc.) on
the given command_interface is preferred.
(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'method-timeout' => $timeout_in_seconds >>

(Optional) We can use this option to specify a large timeout value, which will
be used by all command(s) run, in case the command(s) take a long time to run
due to  large amounts of output.

=back

=back

=cut

sub get_one_identity_attr {

    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my $attr       = shift;

    my $cluster_id_state = $pkg_or_obj->identity(
        requested_fields => [$attr],
        @_
    );

    $Log->exit() if $may_exit;
    return $cluster_id_state->$attr();

} ## end sub get_one_identity_attr

=head2 create

    my $Cluster  = NACL::STask::Cluster->create(
                command_interface  =>  $command_interface,
                cluster => $cluster_name,
                if_exists => $action,      # default 'die'
                lifs_setup => $boolean,    # default '1'
                %other_options
           );

(Class method) creates cluster on node specified by command_interface attribute.
This method provides additional services beyond what's inherent to the product's
cluster create command,as controlled and described by the "if_exists" option.
It also takes care of cluster-create related dependencies such as base licensing,
node mgmt lif creation, intra cluster lif creation, cluster management lif
creation and renaming nodes taking the  details from tharnhost file of the node .

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required) Node on which to create the cluster.
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< clustername=>$cluster_name >>

    (Optional) The name of the Cluster. If not specified, for create operation,
    generates a random cluster name  as $createnode_time().

=item C<< if_exists=>$action >>

    (Optional) What to do if the cluster is setup already. If $action is "die"
    the default), then fail with an exception (in the same way that the cluster
    component would have: by trying the cluster create and letting the product
    complain about the existence of cluster setup). If action is "reuse",
    then do nothing (return a task object referring to the existing cluster).

=item C<< lifs_setup=>0|1 >>

    (Optional)if it is set to 1 Creates cluster lifs and cluster-management lif.

=item C<< "method-timeout"=>$timeout >>

(Optional) How long in seconds to wait for the component calls to wait.
Defaults to 60 seconds.

=item C<< %other_options >>

All of the other various options to
L<< NACL::C::Cluster->create|lib-NACL-C-Cluster-pm/create >>

=back

=back

=over

=item Exceptions

=over

=item C<NACL::APISet::Exceptions::CommandNotFoundException>

This type of exception is thrown when the particular node already has a cluster created on it .

=back

=back

=cut

sub create {

    $Log->enter() if $may_enter;
    my $me   = ( caller(0) )[3];
    my ($pkg, %opts)  = @_;

    # populate the node name which is the primary key
    if ( !ref($pkg) ) {
        %opts = $pkg->_populate_node_name_in_static(%opts);
    }

    %opts = $pkg->_common_validate_with(

        # The default 60 second timeout is insufficient.  Unless the caller
        # specifies otherwise, bump it up. See burt 813506 for discussion
        # about how cluster create has no performance requirement and can take
        # pretty much as long as it pleases.
        params          => [ %{ { 'method-timeout' => 600, %opts } } ],
        additional_spec => {
            if_exists => {
                type      => SCALAR,
                default   => 'die',
                callbacks => {
                    "Value of 'if_exists' should be 'die' or 'reuse'" =>
                        sub { $_[0] =~ /^(?:die|reuse)$/ }
                    }

            },
            clustername => { type => SCALAR, optional => 1 },
            lifs_setup  => { type => SCALAR, default  => 1 },
        },
        allow_extra => 1,
    );

    my $node        = $opts{command_interface};
    my $clustername = delete $opts{clustername} || $node->node() . time();
    my $if_exists   = delete $opts{if_exists};
    my $lifs_setup  = delete $opts{lifs_setup};

    if ($node->version_manager()->is_version_eq(release => "9.1")) {
        if ( ! defined $opts{license}) {
            $opts{license} = "ETYYFLXUQUMADFAAAAAAAAAAAAAA";
        }
    }

    delete $opts{node};
    my $self;
    my %common_options;

    $pkg->_copy_common_component_params(
        source => \%opts,
        target => \%common_options
    );

    # if_exists will only ever equal reuse or die,
    # _common_validate_with assures this
    try {
        $self = $pkg->find(
            command_interface => $node,
            filter            => $pkg->get_primary_keys_options(%opts),
            allow_empty       => 1,
            %common_options
        );
    } ## end try
    catch NACL::C::Exceptions::Cluster::PreClusterMode with {

        # pre-cluster, find will throw the CommandNotFoundException
        # this is expected because the "show" command will not exist
        $Log->comment(
            "$me - Pre-cluster the \"cluster show\" command will not exist, this is expected"
        );
    };

    if ( $self && ( $if_exists eq 'die' ) ) {
        NACL::STask::Exceptions::NodeAlreadyInCluster->throw( "Node "
                . $node->node()
                . ' already has a cluster created on it and if_exists option was \"die\"'
        );
    } ## end if ( $self && ( $if_exists...

    if ( !$self ) {
        my $lifinfo;
        if ($lifs_setup) {
            $lifinfo = NACL::STask::Node->get_hostrec_lif_info(%opts);
            $pkg->_setup_lifs( %opts, lifinfo => $lifinfo, %common_options );
        }

        my $apiset;
        $self = $pkg->C_SUPER(
            _method     => 'create',
            clustername => $clustername,
            apiset_chosen => \$apiset,
            %opts
        );

        if ($apiset->{interface} eq 'ZAPI') {
            $pkg->_cluster_create_join_progress_get(
                 apiset           => $apiset,
                 'method-timeout' => 1600,
            );
        }

        if ($lifs_setup) {
            $pkg->_setup_clus_mgmt_lif( %opts, lifinfo => $lifinfo );
        }
    } ## end if ( !$self )
    $self->name($clustername);
    $Log->exit() if $may_exit;

    return $self;
} ## end sub create

=head2 join

  $cluster = NACL::STask::Cluster->join(
      command_interface => $command_interface,
      create_node => NACL::C::Node,
      if_exists => $action,   # default 'die'
      %other_options
  );

(Class method) Joins the node specified to the cluster. This
method provides additional services beyond what's inherent to the product's
cluster join command, as controlled and described by the "if_exists" option.

L<NACL::APISet::Exceptions::CommandFailedException|lib-NACL-APISet-Exceptions-CommandFailedException-pm>

Contained within the text of the exception will be all of the checks which
have failed.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required) Node to join into the existing cluster.
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< create_node=>NACL::C::Node >>

(Required) Node on which the cluster was created.

=item C<< if_exists=>$action >>

(Optional) What to do if the node is already a member of the cluster. If
$action is "die" (the default), then fail with an exception (in the same way
that the cluster component would have: by trying the cluster join and letting
the product complain). If action is "reuse", then do nothing (return a task
object referring to the existing cluster).

=item C<< %other_options >>

All the other options supported by NACL::C::Cluster->join are supported by
NACL::STask::Cluster->join as well.

=item C<< nacltask_to_cleanup => 0|1 >>

(Optional, defaults to 0(no to cleanup))
Flag indicating if this operation is to be cleaned up or not.

=item C<< nacltask_cleanup_manager >>

Cleanup manager to use for registering.
Default : Will use the default cleanup manager.

=back

=back

=over

=item Exceptions

=over

=item C<NACL::APISet::Exceptions::CommandFailedException>

This type of exception is thrown when an attempt is made to join a node that is already a member of the cluster.

=back

=back

=cut

sub join {
    $Log->enter() if $may_enter;
    my ($pkg, %opts)  = @_;

    # populate the node name which is the primary key
    if ( !ref($pkg) ) {
        %opts = $pkg->_populate_node_name_in_static(%opts);
    }

    # command_interface for NACL::C::Cluster::join refers to the node
    # node being added.
    %opts = $pkg->_common_validate_with(

        # The default 60 second timeout is insufficient.  Unless the caller
        # specifies otherwise, bump it up. See burt 813506 for discussion
        # about how cluster create has no performance requirement and can take
        # pretty much as long as it pleases.
        params          => [%{{'method-timeout' => 600, %opts}}],
        additional_spec => {
            if_exists => {
                type      => SCALAR,
                default   => 'die',
                callbacks => {
                    "Value of 'if_exists' should be 'die' or 'purge'" =>
                        sub { $_[0] =~ /^(?:die|reuse)$/ }
                }
            },
            create_node => {isa  => 'NACL::C::Node'},
            lifs_setup  => {type => BOOLEAN, default => 1},
            check_rings => {type => BOOLEAN, default => 1},
            $pkg->_cleanup_validate_spec(),
        },
        allow_extra => 1,
    );

    my (%opts_for_cleanup, $nacltask_to_cleanup);
    $pkg->_copy_common_opts_for_cleanup(
        'source'              => \%opts,
        'target'              => \%opts_for_cleanup,
        'nacltask_to_cleanup' => \$nacltask_to_cleanup,
        'to_cleanup'          => '_unjoin'
    );

    my $node        = $opts{command_interface};
    my $create_node = delete $opts{create_node};
    my $if_exists   = delete $opts{if_exists};
    my $lifs_setup  = delete $opts{lifs_setup};
    my $check_rings = delete $opts{check_rings};
    delete $opts{node};

    my $self;
    my %common_options;

    $pkg->_copy_common_component_params(
       source => \%opts,
       target => \%common_options,
    );
    # See if we're already a member of the specified cluster.
    if ( $pkg->_in_same_cluster( node1 => $node, node2 => $create_node ) ) {

        # NACL::C::join does not fail if the node is already a member of the
        # cluster, so we have to check manually and die if appropriate.
        if ( $if_exists eq 'die' ) {
            NACL::STask::Exceptions::NodeAlreadyInCluster->throw(
                    'text' => $node->node()
                    . ' is already a member of cluster '
                    . $create_node->node() );
        } else {
            $self = $pkg->find(
                command_interface => $node,
                filter            => { node => $node->node() },
                %common_options,
            );
        } ## end else [ if ( $if_exists eq 'die')

    } else {
        my $lifinfo;
        if ($lifs_setup) {
            $lifinfo = NACL::STask::Node->get_hostrec_lif_info( %opts,
                required_fields => [qw(NAME IP MASK PORT)] );
                $pkg->_setup_lifs( %opts, lifinfo => $lifinfo, %common_options );
        }

         my $clus_lif = NACL::C::Interface->find(
            command_interface => $create_node,
            filter            => {
                role        => 'cluster',
                'home-node' => $create_node->node()
            },
            %common_options,
        );

        my $apiset;
        $self = $pkg->C_SUPER(
            _method => 'join',
            %opts,
            clusteripaddr => $clus_lif->state()->address(),
            apiset_chosen => \$apiset,
        );
        if ($apiset->{interface} eq 'ZAPI') {
             $pkg->_cluster_create_join_progress_get(
                  apiset           => $apiset,
                  'method-timeout' => 1600,
             );
        }

        if ($lifs_setup) {
            $pkg->_setup_clus_mgmt_lif( %opts, lifinfo => $lifinfo );
        }
    } ## end else [ if ( $pkg->_in_same_cluster...

    if ($check_rings) {
        $self->_execute_method_in_loop(method => \&check_rings);
    }

    if ($nacltask_to_cleanup) {
        my $other_opts = {'create_node' => $create_node};
        $self->_register_for_cleanup(%opts_for_cleanup, 'other_opts' => $other_opts);
    }

    $Log->exit() if $may_exit;
    return $self;
} ## end sub join

sub _unjoin {
	$Log->enter() if $may_enter;
	my $self = $_[0];
    require NACL::MTask::Cluster;
	
	NACL::MTask::Cluster->unjoin(
        command_interface => $self->{'create_node'},
        unjoin_node       => $self->{'command_interface'},
	);
	
	$Log->exit() if $may_exit;
}


=head2 _setup_lifs

  $cluster->_setup_lifs(
    lifinfo => $lifinfo,
    netmask => '255.255.248.0',
  );

  NACL::STask::Cluster->_setup_lifs(
    command_interface => $command_interface,
    lifinfo => $lifinfo,
    netmask => '255.255.248.0',
  );

(Class or instance method) This method uses previously retrieved LIF data
from NACL::STask::Node::get_hostrec_lif_info() to create cluster lifs.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< lifinfo=>$lifinfo >>

(Required) Data structure as returned NACL::STask::Node::get_hostrec_lif_info().  Used to provide the
attributes for the lifs.

=item C<< netmask=>$netmask >>

(Optional) Provides the netmask for the created lifs.  If not specified,
the value is derived from 'lifinfo'.

=back

=back

=cut

sub _setup_lifs {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = @_;

    my $command_interface = $opts{command_interface};
    my $lifinfo           = $opts{lifinfo};
    my $netmask           = $opts{netmask} || $lifinfo->{CLUSTER}[0]{MASK};

    my $name = $command_interface->node();

    my $cluster_info = $lifinfo->{CLUSTER};
    my $max          = scalar @$cluster_info;
    for ( my $i = 1; $i <= $max; $i++ ) {
        my $info = $cluster_info->[ $i - 1 ];
        next if !defined $info;
        my $ipaddr       = $info->{IP};
        my $cluster_port = $info->{PORT};
        my %hash;
        $hash{vserver} = 'Cluster';
        my $clus_lif     = NACL::C::Interface->find(
            command_interface => $command_interface,
            filter            => {
                'home-port' => $cluster_port,
                lif         => $info->{NAME},
                %hash,
            },
            allow_empty => 1
        );
        $name = $command_interface->has_uichange(uichange => 'ipspaces-ms4')? 'Cluster': $name;
        if ( !$clus_lif ) {
            NACL::C::Interface->create(
                command_interface => $command_interface,
                server            => $name,
                address           => $ipaddr,
                lif               => $info->{NAME},
                role              => 'cluster',
                'home-node'       => $command_interface->node(),
                'home-port'       => $cluster_port,
                netmask           => $netmask,
            );
        } ## end if ( !$clus_lif )
    } ## end for ( my $i = 1; $i <= ...
    $Log->exit() if $may_exit;
} ## end sub _setup_lifs

=head2 _setup_clus_mgmt_lif

  $cluster->_setup_clus_mgmt_lif(
    lifinfo => $lifinfo,
    ip => $address,
    port => 'e0c',
    netmask => '255.255.248.0',
  );

  NACL::STask::Cluster->_setup_clus_mgmt_lif(
    command_interface => $command_interface,
    lifinfo => $lifinfo,
    ip => $address,
    port => 'e0c',
    netmask => '255.255.248.0',
  );

(Class or instance method) This method uses previously retrieved LIF data
from the NACL::STask::Node::get_hostrec_lif_info() to create the cluster management lif.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< lifinfo=>$lifinfo >>

(Required) Data structure as returned NACL::STask::Node::get_hostrec_lif_info().  Used to provide the
attributes for the lif.

=item C<< ip=>$address >>

(Optional) Provides the address for the created lif.  If not specified,
the value is derived from 'lifinfo'.

=item C<< port=>$port >>

(Optional) Provides the port for the created lif.  If not specified,
the value is derived from 'lifinfo'.

=item C<< netmask=>$netmask >>

(Optional) Provides the netmask for the created lif.  If not specified,
the value is derived from 'lifinfo'.

=item C<< nacltask_add_route=> 1|0 >>

(Optional) Add the route for cluster management interface.
Default to '0'

=back

=back

=cut

sub _setup_clus_mgmt_lif {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = validate_with(
        params => \@_,
        spec   => {
            'command_interface' =>
              { type => OBJECT, isa => 'NACL::C::CommandInterface' },
            'lifinfo'            => { type => HASHREF, optional => 1 },
            'ip'                 => { type => SCALAR,  optional => 1 },
            'port'               => { type => SCALAR,  optional => 1 },
            'netmask'            => { type => SCALAR,  optional => 1 },
            'nacltask_add_route' => { type => BOOLEAN, default  => 0 },
        },
        allow_extra => 1
    );
    my $command_interface = $opts{command_interface};
    my $lifinfo           = $opts{lifinfo};
    my $cluster_mgmt      = $lifinfo->{CLUSTER_MGMT}[0];
    my $mgmt_ip           = $opts{ip} || $cluster_mgmt->{IP};
    my $mgmt_port         = $opts{port} || $cluster_mgmt->{PORT};
    my $nacltask_add_route= $opts{nacltask_add_route};    

    my $clus_mgmt_lif = NACL::C::Interface->find(
        command_interface => $command_interface,
        filter            => { role => 'cluster-mgmt' },
        allow_empty       => 1
    );
    if ( !$clus_mgmt_lif ) {
        my $vs = NACL::C::Vserver->find(
            command_interface => $command_interface,
            filter            => { type => 'admin' }
        );
        my $mgmt_netmask = $opts{netmask} || $cluster_mgmt->{MASK};
        my $lif=NACL::C::Interface->create(
            command_interface => $command_interface,
            server            => $vs->vserver(),
            address           => $mgmt_ip,
            lif               => 'clus_mgmt',
            role              => 'cluster-mgmt',
            'home-node'       => $command_interface->node(),
            'home-port'       => $mgmt_port,
            netmask           => $mgmt_netmask,
        );

        if ($nacltask_add_route &&  $cluster_mgmt->{GATEWAY} ) {
            my %opt = (
                command_interface  => $command_interface,
                nacltask_if_exists => 'reuse',
                vserver            => $vs->vserver(),
                destination        => '0.0.0.0/0',
                gateway            => $cluster_mgmt->{GATEWAY},
            );

            if (
                $command_interface->has_uichange( uichange => 'ipspaces-ms4' ) )
            {
                NACL::STask::NetworkRoute->create(%opt);
            }
            else {
                NACL::STask::NetworkRoutingGroupsRoute->create( %opt,
                    'routing-group' =>
                      $lif->get_one_state_attribute('routing-group') );
            }
        }
    } ## end if ( !$clus_mgmt_lif )
    $Log->exit() if $may_exit;
} ## end sub _setup_clus_mgmt_lif

=head2 _get_cluster_identity

  $cluster->_get_cluster_identity($command_interface);

  NACL::STask::Cluster->_get_cluster_identity($command_interface);

(Class or instance method) This method returns a NACL::C::ClusterIdentity
for the given command interface.  If not a member of a cluster, undef is
returned.

=over

=item Options

=over

=item C<< $command_interface >>

(Required)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=back

=back

=cut

sub _get_cluster_identity {
    my ($pkg, $ci) = @_;

    my $cid;
    try {
        $cid = $pkg->identity( command_interface => $ci );
    }
    catch NACL::C::Exceptions::ClusterIdentity::PreClusterMode with {

        # If a node isn't part of a cluster, this will fail as the 'cluster
        # identity' command won't be available.  Trap only that exception.
    };
    return $cid;
} ## end sub _get_cluster_identity

=head2 _in_same_cluster

  $cluster->_in_same_cluster(
    node1 => NACL::C::Node,
    node2 => NACL::C::Node,
  );

  NACL::STask::Cluster->_in_same_cluster(
    node1 => NACL::C::Node,
    node2 => NACL::C::Node,
  );

(Class or instance method) Determines whether the provided nodes are members
of the same cluster.  Returns 1 if they are, 0 or undef if either is not a
cluster member or if they belong to different clusters.

=over

=item Options

=over

=item C<< node1=>NACL::C::Node >>

(Required) First node to check.

=item C<< node2=>NACL::C::Node >>

(Required) Second node to check.

=back

=back

=cut

sub _in_same_cluster {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = @_;
    my ( $node1, $node2 ) = ( $opts{node1}, $opts{node2} );
    my $cid1 = $pkg->_get_cluster_identity($node1);
    if ( !defined $cid1 || !defined $cid1->uuid() ) {
        return undef;
    }
    my $cid2 = $pkg->_get_cluster_identity($node2);
    if ( !defined $cid2 || !defined $cid2->uuid() ) {
        return undef;
    }

    $Log->exit() if $may_exit;
    return $cid1->uuid() eq $cid2->uuid();
} ## end sub _in_same_cluster

=head2 _execute_method_in_loop

  $cluster->_execute_method_in_loop(
            method => \&method,
            polling_interval => $poll_int,
            exec_times => $num_times
            );

  NACL::STask::Cluster->_execute_method_in_loop(
               method => \&method,
               polling_interval => $poll_int,
               exec_times => $num_times
               );

(Class or instance method) Executes a given method for 'n' number of times,
unless it doesnot throw an exception. If no exception is caught in an ith
iteration where i < n , then the method returns successfully. If it continues
to catch an exception till the nth iteration , then the exception is thrown
back to the caller .

=over

=item Options

=over

=item C<< method >>

(Required) Reference to the method in the package which needs to be executed

=item C<< polling_interval >>

(Optional) Number of seconds to wait before the next iteration starts. By default
waits for 10 seconds before the next iteration.

=item C<< exec_times >>

(Optional) Maximum number of times the method needs to be executed in the loop
untill it doesnot throw an exception. By default the method is executed five
number of times.

Returns nothing on success. On Failure , the exception caught is re-thrown
to the caller.

=back

=back

=cut

sub _execute_method_in_loop {

    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;

    my %opts = validate_with(
        params => \@_,
        spec   => {
            method           => { type => CODEREF },
            polling_interval => { type => SCALAR, default => 10 },
            exec_times       => { type => SCALAR, default => 5 }
        },
        allow_extra => 1
    );

    my $method           = $opts{'method'};
    my $polling_interval = $opts{'polling_interval'};
    my $num_exec         = $opts{'exec_times'};
    my $exception        = undef;

    foreach ( my $i = 1; $i <= $num_exec; $i++ ) {
        try {
            $pkg_or_obj->$method;
        }
        catch NATE::BaseException with {
            $exception = shift;
        };

        # An exception was thrown and there are more iterations
        if ( ref($exception) && $i < $num_exec ) {
            Tharn::snooze $polling_interval ;
            $exception = undef;
        } elsif ( $i < $num_exec ) {
            last;
        }
    } ## end foreach ( my $i = 1; $i <= ...
    if ($exception) {
        $Log->exit() if $may_exit;
        $exception->throw();
    }

    $Log->exit() if $may_exit;
} ## end sub _execute_method_in_loop

=head2 unique_clusters

    if( NACL::STask::Cluster->unique_clusters('nodes_to_check' => \@nodes)) {
        # Proceed to test
    }

Given an array of STask or Component Node objects, this routine will return 1
if the supplied nodes are located on different clusters, and 0 otherwise.
This method can only be invoked as a Class method.

=over

=item Options

=over

=item C<< nodes_to_check => [$node1, $node2, $node3] >>

 (Required parameter): A key value pair, assigned to an array reference
 containing the list of node objects.

=back

=back

=over

=item Exceptions

=over

=item C<NATE::Exceptions::Argument>

This type of exception is thrown when atleast 2 nodes are required to compare Clusters.

=item C<NATE::BaseException>

This type of exception is thrown when there is any failure while retrieving cluster ID.

=back

=back

=cut

sub unique_clusters {
    $Log->enter() if $may_enter;
    my $pkg = shift;

    validate_with(
        params => \@_,
        spec   => { nodes_to_check => { type => ARRAYREF, optional => 0 }, },
    );

    my %opts             = @_;
    my @nodes_to_compare = @{ $opts{'nodes_to_check'} };
    my %hash_to_compare;
    my $cid;
    my $uuid;

    if ( @nodes_to_compare < 2 ) {
        $Log->exit() if $may_exit;
        NATE::Exceptions::Argument->throw(
            qq(Atleast 2 nodes are required to compare Clusters) );
    }
    foreach my $node (@nodes_to_compare) {
        $cid = $pkg->_get_cluster_identity($node);
        if ( !defined $cid || !defined $cid->uuid() ) {
            $Log->exit() if $may_exit;
            NATE::BaseException->throw(
                qq(Could not retrieve the Cluster ID) );
        }
        $uuid = $cid->uuid();
        $hash_to_compare{$uuid} = 1;
    }
    $Log->exit() if $may_exit;
    return scalar keys %hash_to_compare == @nodes_to_compare;
}

=head2 group_nodes_by_unique_clusters

   # build out a lookup hash that contains
   my $cluster_hash_ref = NACL::STask::Cluster->group_nodes_by_unique_clusters(nodes => [$node1, $node2, $node3, $node4]);

   # Assuming that node1 and node2 are in the same cluster and node3 and
   # node4 are in another, this will be the value of $cluster_hash_ref:
   #
   # $cluster_hash_ref = {
   #   '$cluster1_uuid' => [$node1, $node2],
   #   '$cluster2_uuid' => [$node3, $node4],
   # }

Given an array of STask or Component Node objects, this routine will return
a hash reference where the keys are cluster identities (cluster uuids) and the
values are array references of NACL::C::Node objects that belong to each cluster.
This method can only be invoked as a Class method.

=over

=item Options

=over

=item C<< nodes => [$node1, $node2, $node3, etc] >>

 (Required parameter): A key value pair, assigned to an array reference
 containing the list of node objects.

=back

=back

=over

=item Exceptions

=over

=item C<NATE::Exceptions::Argument>

This type of exception is thrown when at least 1 node is required to group nodes by unique clusters.

=item C<NATE::BaseException>

This type of exception is thrown when could not retrieve the Cluster ID.

=back

=back

=cut

sub group_nodes_by_unique_clusters {
    $Log->enter() if $may_enter;
    my $pkg = shift;

    validate_with(
        params => \@_,
        spec   => { nodes => { type => ARRAYREF, optional => 0 }, },
    );

    my %opts         = @_;
    my @nodes        = @{ $opts{'nodes'} };
    my $cluster_hash = {};
    my $cid;
    my $uuid;

    if ( not scalar(@nodes) ) {
        $Log->exit() if $may_exit;
        NATE::Exceptions::Argument->throw(
            qq(At least 1 node is required to group nodes by unique clusters)
        );
    }

    # build the cluster hash reference in this format:
    # {
    #   'cluster_uuid1' => [node_obj1, node_obj2, ...],
    #   'cluster_uuid2' => [node_obj3, node_obj4, ...],
    #   ...
    # }
    foreach my $node (@nodes) {
        $cid = $pkg->_get_cluster_identity($node);
        if ( !defined $cid || !defined $cid->uuid() ) {
            $Log->exit() if $may_exit;
            NATE::BaseException->throw(
                qq(Could not retrieve the Cluster ID) );
        }
        $uuid = $cid->uuid();
        push( @{ $cluster_hash->{$uuid} }, $node );
    }

    $Log->debug($cluster_hash);
    $Log->exit() if $may_exit;
    return $cluster_hash;
}

=head2 find_unique_clusters

   # obtain an array of unique NACL::C::Cluster objects
   my @cluster_objs = NACL::STask::Cluster->find_unique_clusters(nodes => [$node1, $node2, $node3, $node4]);

Given an array of STask or Component Node objects, this routine will return
an array or array reference of unique NACL::C::Cluster objects, that are associated to
node objects specified by the 'nodes' input.
This method can only be invoked as a Class method.

=over

=item Options

=over

=item C<< nodes => [$node1, $node2, $node3, etc] >>

 (Optional parameter): A key value pair, assigned to an array reference
 containing the list of node objects. By default, all NACL::C::Node objects will be used.
 (Default): [NACL::C::Node->find()]

=back

=back

=over

=item Exceptions

=over

=item C<NATE::Exceptions::Argument>

This type of exception is thrown when at least 1 node is required to find unique clusters.

=back

=back

=cut

sub find_unique_clusters {
    $Log->enter() if $may_enter;
    my $pkg = shift;

    validate_with(
        params => \@_,
        spec   => {
            nodes => { type => ARRAYREF, optional => 1, default => undef },
        },
    );

    my %opts = @_;
    my @nodes;
    if ( not defined( $opts{'nodes'} ) ) {
        @nodes = NACL::C::Node->find();
    } else {
        @nodes = @{ $opts{'nodes'} };
    }

    if ( not scalar(@nodes) ) {
        $Log->exit() if $may_exit;
        NATE::Exceptions::Argument->throw(
            qq(At least 1 node is required to find unique clusters) );
    }

    # first build the cluster_hash
    my $cluster_hash
        = $pkg->group_nodes_by_unique_clusters( nodes => \@nodes );
    my @cluster_objs = ();

    # build a NACL::C::Cluster object for each cluster uuid in the hash
    foreach my $cluster_uuid ( keys %{$cluster_hash} ) {

        # grab the first node from the list to make the cluster object
        my $node_obj = $cluster_hash->{$cluster_uuid}->[0];
        my $cluster_obj
            = NACL::STask::Cluster->new( node => $node_obj->node() );
        push( @cluster_objs, $cluster_obj );
    }

    $Log->exit() if $may_exit;
    return ( wantarray() ) ? @cluster_objs : \@cluster_objs;
}

=head2 sync_clocks

B<DEPRECATED.> ONTAP now uses a cluster-wide date with the
date synchronized across the nodes by NTP, so using this method should
be unnecessary. If it is really necessary to change the date/time for
CMode, use L<NACL::C::ClusterDate::modify|lib-NACL-C-ClusterDate-pm/modify>.

=cut

 sub sync_clocks {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, %opts ) = @_;
    require NACL::STask::Node;
    if ( ref $pkg_or_obj && ( !exists( $opts{command_interface} ) ) ) {
        $opts{command_interface} = $pkg_or_obj->command_interface;
    }
    validate_with(
        params => \%opts,
        spec   => $pkg_or_obj->_common_validate_spec
    );
    NACL::STask::Node->set_current_datetime(%opts);
    $Log->exit() if $may_exit;
}

sub _cluster_create_join_progress_get {
    $Log->enter() if $may_enter;
    my ($pkg, %opts) = @_;
    my $apiset = $opts{apiset};
    my $status = '';
    my $end_time = timeout2time($opts{'method-timeout'});
    while ($status ne 'true') {
        my ($response, $progress_get_succeeded, $times_get_ran);
        while (!$progress_get_succeeded) {
            try {
                $response = $apiset->cluster_create_join_progress_get();
                $progress_get_succeeded = 1;
            } catch NACL::APISet::Exceptions::ResponseException with {
                my $exception = $_[0];
                # See burts 953464, 817095, 798709
                if ($exception->text() =~
                           /cannot connect to socket|No response received|failed to connect SSL/i) {
                    if ($times_get_ran++ < 5) {
                        Tharn::snooze(1);
                    } else {
                        $exception->set_text('Even after retrying for 5 '
                            . 'seconds, the attempt to connect to the socket '
                            . 'to execute the ZAPI was unsuccessful. This '
                            . 'probably means that Apache took longer than '
                            . "expected to restart. Original failure:\n"
                            . $exception->text());
                        $Log->exit() if $may_exit;
                        $exception->throw();
                    }
                } else {
                    # Some other error, so propagate it
                    $Log->exit() if $may_exit;
                    $exception->throw();
                }
            };
        }
        my $output = $response->get_parsed_output();
        $status = $output->[0]{attributes}[0]->{'cluster-create-join-progress-info'}[0]->{'is-complete'};
        Tharn::sleep(30);
        if (time() > $end_time) {
            $Log->exit() if $may_exit;
            NACL::Exceptions::Timeout->throw("Timed out waiting for the cluster to get created/joined.");
        }
    }
    $Log->exit() if $may_exit;
}

=head2 group_vservers_by_unique_clusters

   # returns a hash that contains cluster uuid as key and vserver(objects) as value
   my $cluster_hash_ref = NACL::STask::Vserver->group_vservers_by_unique_clusters(
           cservers => [$cserver1, $cserver2],
           vservers => [$vserver1, $vserver2, $vserver3, $vserver4]);

   # Assuming that vserver1 and vserver2 are in the same cluster and vserver3 and
   # vserver4 are in another, this will be the value of $cluster_hash_ref:
   #
   # $cluster_hash_ref = {
   #   '$cluster1_uuid' => [$vserver1, $vserver2],
   #   '$cluster2_uuid' => [$vserver3, $vserver4],
   # }

Given an array of STask or Component Vserver objects, this routine will return
a hash reference where the keys are cluster identities (cluster uuids) and the
values are array references of NACL::C::Vserver objects that belong to each cluster.
This method can only be invoked as a Class method.

=over

=item Options

=over

=item C<< vservers => [$vserver1, $vserver2, $vserver3, etc] >>
Vserver is an ARRAYREF containing the value of $vserver1, $vserver2 .
where $vserver1 ,$vserver2 .. are object

=item C<< cservers => [$cserver1, $cserver2,  etc] >>
cserver is an ARRAYREF containing the value of $cserver1, $cserver2 .
where $cserver1 ,$cserver2 are object.

=back

=back

=cut
sub group_vservers_by_unique_clusters {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    my %opts = validate_with(
                  params        => \@args,
                  spec => {
                     cservers => { type => ARRAYREF, optional => 1 },
                     vservers => { type => ARRAYREF, optional => 0 },
                  },
                  ignore_primary_keys => 1,
    );

    my @cservers = @{ $opts{'cservers'} };
    my @vservers = @{ $opts{'vservers'} };
    my $cluster_hash = {};
    my ($obj,$caught_exception) ;

    #Using cserver as command_interface and looping each vserver on each
    #cserver. If vserver is getting then add cluster uid as key.

    foreach my $ci (@cservers) {
       foreach my $vserver (@vservers){
          $caught_exception = 0;
          try {
              $obj = NACL::CS::Vserver->fetch(command_interface => $ci ,
                                    filter => {vserver => $vserver->vserver() });
          }
          catch NACL::Exceptions::NoElementsFound with {
              $caught_exception = 1 ;
          };

          next if($caught_exception == 1);

          my $aggr = $obj->aggregate();
          my $aggr_obj = NACL::CS::Aggregate->fetch(command_interface => $ci ,
                                      filter => {aggregate => $aggr }, apiset_must => { interface => 'CLI'});

          my $uuid = $aggr_obj->cluster_id() ;
          if ( !defined $uuid  ) {
             $Log->exit() if $may_exit;
             NATE::BaseException->throw(
                qq(Could not retrieve the Cluster ID) );
          }
          push( @{ $cluster_hash->{$uuid} }, $vserver );
       }
    }

    $Log->debug( sub { Dumper $cluster_hash } );
    $Log->exit() if $may_exit;
    return $cluster_hash;
}

=head2 wait_for_healthy_state

    NACL::STask::Cluster->wait_for_healthy_state(
          'command_interface'  => $command_interface,
           'node'              => $node,
          'health-check-mode'  => "simple|complete"
    );
    $Cluster_obj->wait_for_healthy_state();
This method waits until the Cluster becomes healthy

=item C<< node => $node >>

(Optional)
Cluster health check by default is performed on all the nodes in
the cluster. If node option is provided then the cluster health check performs
health check on only the node and its partner.

=item C<< nacltask_retry_tries_count => $tries_count >>

(Optional, defaults to 3)
The number of times to retry to check the health of the cluster.

=item C<< nacltask_retry_sleep_time => $sleep_time >>

(Optional, defaults to 180 seconds)
The time to sleep between retries of the cluster_health_check .

All of the other various options to
L<< NACL::STask::Cluster->cluster_health_check|lib-NACL-STask-Cluster-pm/cluster_health_check >>

=cut

sub wait_for_healthy_state {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;

    my %opts = $pkg_or_obj->_common_validate_with(
        params             => \@args,
        additional_spec    => {
            %{ NACL::C::Component->_common_validate_spec() },
            'node' => {type => SCALAR|OBJECT|UNDEF , optional => 1},
            'health-check-mode' => { regex => qr/^simple|complete$/ , default => "complete" },
            nacltask_retry_tries_count => { type => SCALAR, default => 3 },
            nacltask_retry_sleep_time  => { type => SCALAR, default => 180 },
        },
        ignore_primary_keys => 1,
		allow_extra => 1
    );

    my %retry_opts;
    $pkg_or_obj->_move_retry_params(
        source => \%opts,
        target => \%retry_opts
    );
    my $command_interface = $opts{command_interface} // $pkg_or_obj->command_interface();
    my $health_check_mode = delete $opts{'health-check-mode'};

    ## wait till node gets healthy
    nacl_method_retry(
        code        => sub {
            NACL::STask::Cluster->cluster_health_check(
             command_interface => $opts{command_interface},
             node           => $opts{node},
             'health-check-mode'  => $health_check_mode
            );
        },
        exceptions  => [
            'NACL::Exceptions::ClusterRingError',
            'NACL::Exceptions::ClusterHealthCheckError',
            'NACL::APISet::Exceptions::TimeoutException'
        ],
        %retry_opts
    );
    $Log->exit() if $may_exit;
}

1;
