#
# Copyright (c) 2012 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Cluster Task Module
## @author Deepthi Shetty, dl-mpo-nacl-dev@netapp.com
## @status shared
## @pod here

package NACL::STask::DFM::Cluster;

use strict;
use warnings;

use base qw(NACL::C::DFM::Cluster NACL::STask::STask);

use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Params::Validate qw(validate SCALAR SCALARREF ARRAYREF BOOLEAN OBJECT);
use NATE::BaseException qw(:try);
use NATE::Exceptions::Argument qw(:try);
use NACL::ComponentUtils qw(_optional_scalars);
use NACL::MTask::DFM::Monitor;
use Hostrec;
use Net::IP qw(ip_is_ipv4 ip_is_ipv6);

use Data::Dumper;

=head1 NAME

NACL::STask::DFM::Cluster

=head1 DESCRIPTION

C<NACL::STask::DFM::Cluster> provides a number of well-defined but
potentially complex or multi-step methods related to Clusters in
OnCommand.

It builds on top of, and is a derived class of C<NACL::C::DFM::Cluster>,
and so it also provides methods that are more in the scope of
individual Cluster-related commands.  See C<NACL::C::DFM::Cluster> for
details.  This also means that a C<NACL::STask::DFM::Cluster> object may
generally be used in place of a component object.

=head1 METHODS

=head2 add

  my $cluster  = NACL::STask::DFM::Cluster->add(
      command_interface        => $command_interface,
      'cluster-hostname-or-ip' => $cluster_name,
      username		       => $username,
      password 		       => $password,
      nacltask_if_exists       => $action,      # default 'reuse'
      refresh_monitors         => 0|1,          # default 0
      %other_options
  );

(Class method) Add a cluster.  This method provides additional
services beyond what's inherent to the product's cluster addition
commands, as controlled and described by the new C<if_exists> option.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>
(Required)

(Required) As C<NACL::C::DFM::DFMClient>.  A component object that
represents the host to which to send commands.

=item C<< cluster_hostname_or_ip => $cluster_name_or_ip >>

(Required) The hostname or ip of the cluster to be added.

=item C<< nacltask_if_exists => $action >>

(Optional) What to do if the cluster to be added already exists.
If action is "reuse" (the default), then do nothing (return a
task object referring to the existing cluster). If $action is "die",
then fail with an exception (in the same way that the cluster component
would have: by trying the cluster add and letting the product
complain about the cluster already existing). If action is "purge",
then delete the cluster (see the C<purge> method, below) before
creating a new one.

=item C<< refresh_monitors => 0|1 >>

(Optional) If 0 (the default), do not run a refresh on the added cluster.
If 1, then run a monitor refresh on it.

=item C<< _was_added => \$scalar >>

=item C<< hostobj => $hostobj

(Optional) Cluster Hostrec object needs to be passed.The precedence is given to this object.
If this is passed, then the username,password and cluster-hostname-or-ip keys (if provided explicitly),
will be overridden by the corresponding values extracted from the hostrec object.
If not provided, then the username,password and the cluster-hostname-or-ip needs to be explicitly provided.

=item C<< _was_added => \$scalar >>

(Optional)
When this option is provided a reference to a scalar variable, the variable
gets filled in with a boolean value describing whether the cluster
was found (in which case the value will be '0'; possible with
if_exists => "reuse") or whether the cluster was added (value will be 1;
possible with if_exists => "purge" or when cluster wasn't found). This
will be useful to determine whether the cluster needs to be cleaned up later.

    my $was_added;
    my $cluster  = NACL::STask::DFM::Cluster->add(
	nacltask_if_exists  => 'reuse',
	_was_added          => \$was_created,
	%other_opts
    );

    # To decide for cleanup
    if( $was_added ) {
	# Purge
	$cluster->purge( ... );

=item C<< monitor_list => $array_ref >>

( Optional )
This option is meaningful only when refresh_monitors => 1 is given.
Refer L<< NACL::MTask::DFM::Monitor | lib-NACL-MTask-DFM-Monitor-pm >>. 

=item C<< complete_on_states => $array_ref  >>

( Optional )
This option is meaningful only when refresh_monitors => 1 is given. 
Refer L<< NACL::MTask::DFM::Monitor | lib-NACL-MTask-DFM-Monitor-pm >>. 

=item C<< wait_time => $scalar >>

( Optional)
This option is meaningful only when refresh_monitors => 1 is given.
Refer L<< NACL::MTask::DFM::Monitor | lib-NACL-MTask-DFM-Monitor-pm >>.

=item C<< polling_interval => $scalar >>

( Optional )
This option is meaningful only when refresh_monitors => 1 is given.
Refer L<< NACL::MTask::DFM::Monitor | lib-NACL-MTask-DFM-Monitor-pm >>.


=back

=item command_interface, apiset_must, apiset_should, method-timeout, password,
username, admin-transport, admin-port, etc.

All of the other various options to
L<< NACL::C::DFM::Cluster->add | lib-NACL-C-DFM-Cluster-pm/add >>

=back

=cut

sub add {

    $Log->enter() if $may_enter;

    my $pkg = shift;

    my %opts = $pkg->_common_validate_with(

        params          => \@_,
        additional_spec => {
            nacltask_if_exists =>
                { %{ $pkg->_if_exists_validate_spec() }, default => 'reuse' },
            _was_added           => { type => SCALARREF, optional => 1 },
            'refresh_monitors'   => { type => SCALAR,    default  => '0' },
            'monitor_list'       => { type => ARRAYREF,  optional => 1, },
            'wait_time'          => { type => SCALAR,    optional => 1, },
            'polling_interval'   => { type => SCALAR,    optional => 1, },
            'complete_on_states' => { type => ARRAYREF, optional => 1 },
            hostobj => { isa => 'Hostrec::Appliance', optional => 1, },
        },
        ignore_primary_keys => 1,
        allow_extra         => 1,
    );

    my $cmd_interface = $opts{command_interface};
    if ( exists $opts{hostobj} ) {
        my $host = delete( $opts{hostobj} );
        $opts{'password'}               = $host->default_password;
        $opts{'username'}               = $host->default_username;
        $opts{'cluster-hostname-or-ip'} = $host->default_ip;
    }
    my $cluster_name_or_ip = $opts{'cluster-hostname-or-ip'};
    my $cluster_username   = $opts{'username'};
    my $cluster_password   = $opts{'password'};

# Task specific options, not used for component add
# nacltask_if_exists, refresh_monitors, monitor_list, wait_time, polling_interval
    my $if_exists = delete $opts{nacltask_if_exists};
    my $was_added = delete $opts{_was_added};

    my $refresh = delete $opts{refresh_monitors};
    my %refresh_monitor_options;

    # Inititalize it to 0 (false)
    $$was_added = 0;

    # Take options for monitor refresh if provided.
    $pkg->_hash_move(
        source => \%opts,
        target => \%refresh_monitor_options,
        move   => [qw( complete_on_states monitor_list wait_time polling_interval )],
    );

    my $self;

    # Call the component add()
    try {

        $self = $pkg->SUPER::add(%opts);

        # Set flag
        $$was_added = 1;

    }
    catch NACL::APISet::Exceptions::ResponseException with {

        my $exception = shift;
        my $text      = $exception->text();

        my $expected_error_message = "Cluster.*already exists";

       # Throw the exception if it's not the exception message we're expecting
       # Or if the if_exists option is die
        if (   ( $text !~ $expected_error_message )
            || ( $if_exists =~ /die/i ) )
        {

            # Rethrow the exception
            $exception->throw();
        } else {
     
            my $cluster_name;
     
             # Create an APISet object of the Host, to query for it's hostname
                my $hostrec_obj = Hostrec->new(
                    id               => "cluster-name",
                    default_username => $cluster_username,
                    default_password => $cluster_password,
                    default_ip       => $cluster_name_or_ip,
                    hosttype         => "filer-ng",
                    reuse            => 0,
                    save             => 0,
                );
                my $cluster_apiset = NACL::APISet->new(
                    category  => "Node",
                    interface => "CLI",
                    set       => "CMode",
                    hostobj   => $hostrec_obj,
                );
                my $response = $cluster_apiset->execute_raw_command(
                    command => "hostname" );
                chomp($response);
                chomp($response);
                $cluster_name = $response;

           
            try {

                $self = $pkg->find(
                    command_interface => $cmd_interface,
                    filter            => { 'cluster-name' => $cluster_name },
                );
            };

            # If if_exists is 'purge' then call 'purge' on the object
            if ( $if_exists =~ /purge/i ) {

                $self->purge();

                # Call add again
                $self = $pkg->SUPER::add(%opts);

                # Set flag
                $$was_added = 1;

            } ## end of $if_exists eq purge

        } ## end of else of if_exists eq die

    };    ## end of catch

    # If monitor has to be refreshed
    if ($refresh) {

        # Refresh monitors
        my $object_id = $self->cluster_name_or_id();

        # Add to the refresh monitor option
        $refresh_monitor_options{object_name_or_ids} = [$object_id];

        $pkg->_hash_copy(
            source => \%opts,
            target => \%refresh_monitor_options,
            copy   => [qw ( command_interface )],
        );

        $Log->debug("Invoking Refresh monitor Task.");

        NACL::MTask::DFM::Monitor->run(%refresh_monitor_options);

    } ## end if $refresh

    $Log->exit() if $may_exit;

    # Return the STask object
    return $self;
}

=head2 purge

    $cluster->purge(%other_options);

    NACL::STask::DFM::Cluster->purge(
      command_interface   => $command_interface,
      'cluster-name-or-id'  => $cluster_name,
      nacltask_if_purged  => $action,     # default 'die'
      nacltask_verify     => 0|1,	  # default 0
      %other_options
    );

(Class or instance method) This method deletes a cluster.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< cluster_name_or_id => $cluster_name_or_id  >>

(Required) The name or id of the cluster to be added.

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< nacltask_verify => $nacltask_verify_boolean >>

(Optional)
If '0' (default), verification will not be performed.
If '1', verification will be performed to ensure that
the deletion did happen successfully.

=item C<< nacltask_if_purged => die | pass >>

(Optional, defaults to "die")
Specifies what to do if the cluster we're trying to purge no longer
exists. A value of C<die> (the default) will result in an exception
being thrown. A value of C<pass> will mean that the exception that would
normally occur when we're trying to delete a cluster that does not exist
is suppressed.

=back

=back

=cut

sub purge {

    $Log->enter() if $may_enter;

    my $pkg_or_obj = shift;

    my %opts = $pkg_or_obj->_common_validate_with(

        params          => \@_,
        additional_spec => {
            nacltask_if_purged =>
                $pkg_or_obj->_if_action_completed_already_validate_spec(),
            nacltask_verify => { type => SCALAR, default => 0 },
        },
        allow_extra => 1,
    );

    my $cmd_interface      = $opts{command_interface};
    my $cluster_name_or_id = $opts{'cluster-name-or-id'};

    # Task specific options, not used for purge
    # nacltask_if_purged, nacltask_verify
    my $if_purged   = delete $opts{nacltask_if_purged};
    my $verify_flag = delete $opts{nacltask_verify};

    # Flag to check if purged or not
    my $is_purged_flag = 0;

    try {

        $pkg_or_obj->SUPER::purge(%opts);

        # Purge successful, hence set $is_purged_flag
        $is_purged_flag = 1;

        $Log->debug("Cluster purge successful.");

    }
    catch NACL::APISet::Exceptions::ResponseException with {

        my $exception = shift;
        my $text      = $exception->text();

        my $expected_error_message = "Cluster.*does not exist";

        if (   ( $text !~ $expected_error_message )
            || ( $if_purged !~ /pass/i ) )
        {

            $Log->exit() if $may_exit;

            # Rethrow the exception
            $exception->throw();

        }
    };

    # If need to verify the cluster purge (i.e. if it was deleted here)
    if ( $verify_flag && $is_purged_flag ) {

        # Do a find
        try {

            my $search_field;

            if ( $cluster_name_or_id =~ /^\d+$/ ) {
                $search_field = "cluster-id";
            } else {
                $search_field = "cluster-name";
            }

            $pkg_or_obj->find(
                command_interface => $cmd_interface,
                filter            => { $search_field => $cluster_name_or_id },
            );

            $Log->exit() if $may_exit;

            # If no exception is thrown, it means verify FAILED
            NACL::Exceptions::VerifyFailure->throw(
                      "Verification of cluster-purge "
                    . " for $cluster_name_or_id FAILED!" );

        }
        catch NACL::Exceptions::NoElementsFound with {

            # This exception means that the purge was successful
            $Log->debug("Cluster purge verified successfully.");

        }
    }

    $Log->exit() if $may_exit;
}

1;
