# $Id: //depot/prod/test/UM6.0_integ/lib/NACL/STask/DFM/Datasource.pm#8 $
#
# Copyright (c) 2012 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Datasource Task Module
## @author Vinothg, dl-mpo-nacl-dev@netapp.com
## @status shared
## @pod here

package NACL::STask::DFM::Datasource;

use strict;
use warnings;

use base qw(NACL::C::DFM::Datasource NACL::STask::STask);

use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Params::Validate qw(validate SCALAR SCALARREF ARRAYREF BOOLEAN OBJECT);
use NATE::BaseException qw(:try);
use NATE::Exceptions::Argument qw(:try);
use NACL::ComponentUtils qw(_optional_scalars);
use NACL::CS::DFM::Cluster;
use Hostrec;
use Data::Dumper;

use constant DEFAULT_WAIT_TIME        => 900;
use constant DEFAULT_POLLING_INTERVAL => 60;

=head1 NAME

NACL::STask::DFM::Datasource

=head1 DESCRIPTION

C<NACL::STask::DFM::datasource> provides a number of well-defined but
potentially complex or multi-step methods related to datasource in
OnCommand.

It builds on top of, and is a derived class of C<NACL::C::DFM::Datasource>,
and so it also provides methods that are more in the scope of
individual datasource-related commands.  See C<NACL::C::DFM:Datasource:
details.  This also means that a C<NACL::STask::DFM::Datasource> object may
generally be used in place of a component object.

=head1 METHODS

=head2 add

  my $datasource_id  = NACL::STask::DFM::Datasource->add(
      command_interface        => $command_interface,
      'hostname-or-ip' => $hostname_or_ip,
      username               => $username,
      password                => $password,
      nacltask_if_exists       => $action,      # default 'reuse'
      %other_options
  );
                     OR
  my $datasource_id  = NACL::STask::DFM::Datasource->add(
      command_interface     => $command_interface,
      hostobj               => $datasource->hostrec(),
      nacltask_if_exists    => $action,      # default 'reuse'
      %other_options
  );

(Class method) Add a datasource.  This method provides additional
services beyond what's inherent to the product's datasource addition
commands, as controlled and described by the new C<if_exists> option.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>
(Required)

(Required) As C<NACL::C::DFM::DFMClient>.  A component object that
represents the host to which to send commands.

=item C<< hostname_or_ip => $hostname_or_ip >>

(Required) The hostname or ip of the datasource to be added.

=item C<< nacltask_if_exists => $action >>

(Optional) What to do if the datasource to be added already exists.
If action is "reuse" (the default), then do nothing (return a
task object referring to the existing datasource). If $action is "die",
then fail with an exception (in the same way that the datasource component
would have: by trying the datasource add and letting the product
complain about the datasource already existing). If action is "purge",
then delete the datasource (see the C<purge> method, below) before
creating a new one.

=item C<< hostobj => $hostobj

(Optional) Datasource Hostrec object needs to be passed.The precedence is given to this object.
If this is passed, then the username,password and hostname-or-ip keys (if provided explicitly),
will be overridden by the corresponding values extracted from the hostrec object.
If not provided, then the username,password and the hostname-or-ip needs to be explicitly provided.

=item C<< wait_till_object_exists => $object_type >>

(Optional)
This is an optional input given. The valid values for $object_type
currently is "cluster". When "cluster" is given as input, a cluster-list
is executed and checked for the datasource to be listed there.
It waits for the datasource to get listed for $wait_time amount of time, polling
every $polling_interval time.
If this option is no provided, no such check is performed. If the option is provided
with valid inputs, then on failure of the object to get listed within the $wait_time
as exception is thrown.

=item C<< wait_time => $time_in_seconds >>

(Optional, defaults to 900s)
This is the amount of time (in SECONDS) for which it waits for the datasource
to get listed.

=item C<< polling_interval => $time_in_seconds >>

(Optional, defaults to 60s)
This is the frequency at which the list command is run to check if the datasource
is getting listed.

=item C<< _was_added => \$scalar >>

(Optional)
When this option is provided a reference to a scalar variable, the variable
gets filled in with a boolean value describing whether the datasource
was found (in which case the value will be '0'; possible with
if_exists => "reuse") or whether the datasource was added (value will be 1;
possible with if_exists => "purge" or when datasource wasn't found). This
will be useful to determine whether the datasource needs to be cleaned up later.

    my $was_added;
    my $datasource_id  = NACL::STask::DFM::Datasource->add(
    nacltask_if_exists  => 'reuse',
    _was_added          => \$was_created,
    %other_opts
    );

    # To decide for cleanup
    if( $was_added ) {
    # Purge
    $datasource_id->purge( ... );

=back

=item command_interface, apiset_must, apiset_should, method-timeout, password,
username, admin-transport, admin-port, etc.

All of the other various options to
L<< NACL::C::DFM::Datasource->add | lib-NACL-C-DFM-Datasource-pm/add >>

=back

=cut

sub add {

    $Log->enter() if $may_enter;

    my $pkg = shift;

    my %opts = $pkg->_common_validate_with(

        params          => \@_,
        additional_spec => {
            nacltask_if_exists => { %{ $pkg->_if_exists_validate_spec() }, default => 'reuse' },
            _was_added         => { type => SCALARREF,            optional => 1 },
            hostobj            => { isa  => 'Hostrec::Appliance', optional => 1, },
            'polling_interval' => { type => SCALAR,               default  => DEFAULT_POLLING_INTERVAL, },
            'wait_time'        => { type => SCALAR,               default  => DEFAULT_WAIT_TIME, },
        },
        ignore_primary_keys => 1,
        allow_extra         => 1,
    );

    my $cmd_interface = $opts{command_interface};
    if ( exists $opts{hostobj} ) {
        my $host = delete( $opts{hostobj} );
        $opts{'password'}       = $host->default_password;
        $opts{'username'}       = $host->default_username;
        $opts{'hostname-or-ip'} = $host->default_ip;
    }
    my $hostname_or_ip      = $opts{'hostname-or-ip'};
    my $datasource_username = $opts{'username'};
    my $datasource_password = $opts{'password'};

    # Task specific options, not used for component add
    # nacltask_if_exists, refresh_monitors, monitor_list, wait_time, polling_interval
    my $if_exists        = delete $opts{nacltask_if_exists};
    my $was_added        = delete $opts{_was_added};
    my $obj_name         = delete $opts{wait_till_object_exists};
    my $polling_interval = delete $opts{polling_interval};
    my $wait_time        = delete $opts{wait_time};

    # Inititalize it to 0 (false)
    $$was_added = 0;

    # Take options for monitor refresh if provided.

    my ( $self, $cluster_obj, $datasourceid, $found, $job_obj, $datasource_added );
    $datasource_added = 0;

    # Call the component add()
    try {
        ( $job_obj, $self ) = $pkg->SUPER::add(%opts);
        $datasource_added = 1;
        $datasourceid     = $self->datasource_id();

        # Set flag
        $$was_added = 1;
    }
    catch NACL::APISet::Exceptions::ResponseException with {

        my $exception = shift;
        my $text      = $exception->text();

        my $expected_error_message = "Cluster.*already exists";

        # Throw the exception if it's not the exception message we're expecting
        # Or if the if_exists option is die
        if (   ( $text !~ $expected_error_message )
            || ( $if_exists =~ /die/i ) )
        {

            # Rethrow the exception
            $exception->throw();
        } else {

            # Create an APISet object of the Host, to query for it's hostname
            my $hostrec_obj = Hostrec->new(
                id               => "datasource-id",
                default_username => $datasource_username,
                default_password => $datasource_password,
                default_ip       => $hostname_or_ip,
                hosttype         => "filer-ng",
                reuse            => 0,
                save             => 0,
            );
            my $datasource_apiset = NACL::APISet->new(
                category  => "Node",
                interface => "CLI",
                set       => "CMode",
                hostobj   => $hostrec_obj,
            );
            my $response = $datasource_apiset->execute_raw_command( command => "hostname" );
            chomp($response);
            chomp($response);
            my $hostname = $response;
            $response = $datasource_apiset->network_interface_show( role => "cluster-mgmt" );
            my $hostip = $response->get_parsed_output()->[0]->{'address'};

            try {
                $self = $pkg->find(
                    command_interface => $cmd_interface,
                    filter            => { 'hostname-or-ip' => "$hostname|$hostip" },
                );
            };
            $datasourceid = $self->datasource_id();

            # If if_exists is 'purge' then call 'purge' on the object
            if ( $if_exists =~ /purge/i ) {

                $self->purge();

                # Call add again
                ( $job_obj, $self ) = $pkg->SUPER::add(%opts);
                $datasource_added = 1;

                # Set flag
                $$was_added = 1;

            }

        } ## end of else of if_exists eq die

    };    ## end of catch

    if ( $datasource_added == 1 ) {
        $Log->debug("Datasource added newly. Id $datasourceid");
    } else {
        $Log->debug("Datasource found with id $datasourceid ");
    }
    $Log->debug('Calling datasource-object-refresh');
	sleep(5); # adding 5 second sleep as a workaround for 674223
    $self->object_refresh();

    $Log->exit() if $may_exit;

    # Return the STask object
    return $self;
}

=head2 purge

    $datasource->purge(%other_options);

    NACL::STask::DFM::Datasource->purge(
      command_interface   => $command_interface,
      'datasource-id'     => $datasource_id,
      nacltask_if_purged  => $action,     # default 'die'
      nacltask_verify     => 0|1,      # default 0
      %other_options
    );

(Class or instance method) This method deletes a datasource.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< datasource_id => $datasource_id  >>

(Required) The name or id of the datasource to be added.

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< nacltask_verify => $nacltask_verify_boolean >>

(Optional)
If '0' (default), verification will not be performed.
If '1', verification will be performed to ensure that
the deletion did happen successfully.

=item C<< nacltask_if_purged => die | pass >>

(Optional, defaults to "die")
Specifies what to do if the datasource we're trying to purge no longer
exists. A value of C<die> (the default) will result in an exception
being thrown. A value of C<pass> will mean that the exception that would
normally occur when we're trying to delete a datasource that does not exist
is suppressed.

=back

=back

=cut

sub purge {

    $Log->enter() if $may_enter;

    my $pkg_or_obj = shift;

    my %opts = $pkg_or_obj->_common_validate_with(

        params          => \@_,
        additional_spec => {
            nacltask_if_purged => $pkg_or_obj->_if_action_completed_already_validate_spec(),
            nacltask_verify    => { type => SCALAR, default => 0 },
        },
        allow_extra => 1,
    );

    my $cmd_interface = $opts{command_interface};
    my $datasource_id = $opts{'datasource-id'};

    # Task specific options, not used for purge
    # nacltask_if_purged, nacltask_verify
    my $if_purged   = delete $opts{nacltask_if_purged};
    my $verify_flag = delete $opts{nacltask_verify};

    # Flag to check if purged or not
    my $is_purged_flag = 0;

    try {

        $pkg_or_obj->SUPER::delete(%opts);

        # Purge successful, hence set $is_purged_flag
        $is_purged_flag = 1;

        $Log->debug("Datasource purge successful.");

    }
    catch NACL::APISet::Exceptions::ResponseException with {

        my $exception              = shift;
        my $text                   = $exception->text();
        my $expected_error_message = "Failed to remove datasources";

        if (   ( $text !~ $expected_error_message )
            || ( $if_purged !~ /pass/i ) )
        {

            $Log->exit() if $may_exit;

            # Rethrow the exception
            $exception->throw();

        }
    };

    # If need to verify the datasource purge (i.e. if it was deleted here)
    if ( $verify_flag && $is_purged_flag ) {

        # Do a find
        try {

            $pkg_or_obj->find(
                command_interface => $cmd_interface,
                filter            => { "datasource-id" => $datasource_id },
            );

            $Log->exit() if $may_exit;

            # If no exception is thrown, it means verify FAILED
            NACL::Exceptions::VerifyFailure->throw( "Verification of datasource-delete " . " for $datasource_id FAILED!" );

        }
        catch NACL::APISet::Exceptions::ResponseException with {
            my $exception = shift;
            my $text      = $exception->text();

            # This exception means that the purge was successful
            $Log->debug("Datasource purge verified successfully.");

        }
    }

    $Log->exit() if $may_exit;
}

## This method is an alias for purge
sub delete {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    $pkg_or_obj->purge(@_);
    $Log->exit() if $may_exit;
}

=head2 object_refresh

    $job_object = NACL::STask::DFM::Datasource->object_refresh(
      command_interface         => $command_interface,
      'datasource-id'           => $datasource_id,
    [ nacltask_check_completion => 0|1, # default 1
      [ success_states          => $array_ref_of_states, ] ]
      %other_options
    );
        OR
    $job_object = $datasource->object_refresh(
        nacltask_check_completion => 0|1, # default 1
        %other_options
    );

(Class or instance method) This method request that the monitored attributes of the given object (and optionally child objects) be "refreshed".
Additionally it will also check the status of the job.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< datasource_id => $datasource_id  >>

(Required for class method, Not Applicable for instance method)

=item C<< nacltask_check_completion => 0|1 >>
(Optional)
If '1' (default), the job-wait-for-completion will be run for the
job started for the refresh operation. It will check for the "status" of
the job at the end. If it is one of the success states ( default normal)
then the operation is considered successful. Otherwise a NACL::Exceptions::OperationFailed
will be thrown.
The success states can be configured using the success_states input.
If '0', no check will be done on the job started for the refresh operation.

=item C<< success_states => $array_ref_of_states >>
(Optional)
This option is considered only when nacltask_check_completion => 1 is present.
This is an array ref representing the list of states of the job's "status",
on reaching any of which, the job started for the refresh operation will be considered complete.
Valid values are: normal(default), warning, partial_failures, error.

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=back

=back

=cut

sub object_refresh {
    $Log->enter() if $may_enter;

    my $pkg_or_obj = shift;

    my %opts = $pkg_or_obj->_common_validate_with(

        params          => \@_,
        additional_spec => {
            nacltask_check_completion => { type => SCALAR,   default  => 1 },
            wait_time                 => { type => SCALAR,   default  => DEFAULT_WAIT_TIME, },
            success_states            => { type => ARRAYREF, optional => 1, default => ['normal'] },
        },
        allow_extra => 1,
    );

    my $cmd_interface = $opts{command_interface};
    my $datasource_id = $opts{'datasource-id'};

    # Task specific options, not used for object-refresh
    my $check_completion          = delete $opts{nacltask_check_completion};
    my $wait_time                 = delete $opts{wait_time};
    my @success_states_from_input = @{ delete $opts{success_states} };
    my $job                       = $pkg_or_obj->SUPER::object_refresh(%opts);

    my $job_state = $job->wait_for_state( 'wait-for-state-timeout' => $wait_time, );
    if ($check_completion) {
        my $status = $job_state->state->status();
        if ( grep ( /^$status$/, @success_states_from_input ) ) {

            $Log->comment( 'Status of the object-refresh is  :' . $status );
        } else {
            NATE::BaseException->throw( 'Status of the object-refresh is  :' . $status );

        }

        $Log->exit() if $may_exit;

    }
    $Log->exit() if $may_exit;
    return $job;
}

1;
