# $Id: //depot/prod/test/nacldev/lib/NACL/STask/DFM/Eventclass.pm#1 $
#
# Copyright (c) 2012 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Eventclass Task Module
## @author Ramya M R, dl-mpo-nacl-dev@netapp.com
## @status shared
## @pod here

package NACL::STask::DFM::Eventclass;

use strict;
use warnings;

use base qw(NACL::C::DFM::Eventclass NACL::STask::DFM);

use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Params::Validate qw(validate SCALAR SCALARREF ARRAYREF BOOLEAN OBJECT);
use NATE::BaseException qw(:try);
use NATE::Exceptions::Argument qw(:try);
use NACL::ComponentUtils qw(_optional_scalars);

use Data::Dumper;

=head1 NAME

NACL::STask::DFM::Eventclass

=head1 DESCRIPTION

C<NACL::STask::DFM::Eventclass> provides a number of well-defined but
potentially complex or multi-step methods related to Eventclass in
OnCommand.

It builds on top of, and is a derived class of C<NACL::C::DFM::Eventclass>,
and so it also provides methods that are more in the scope of
individual Eventclass-related commands.  See C<NACL::C::DFM::Eventclass> for
details.  This also means that a C<NACL::STask::DFM::Eventclass> object may
generally be used in place of a component object.

=head1 METHODS

=head2 add_custom

  my $Eventclass  = NACL::STask::DFM::Eventclass->add_custom(
      command_interface  => $command_interface,
      'event-namespace-name' => $event_namespace_name,
	  'about-message' => $about_message,
	  'event-names' => $event_names,
	  'event-class-name' => $event_class_name, 
      nacltask_if_exists => $action,      # default 'reuse'
      %other_options
  );

(Class method) Create a Custom Eventclass.  This method provides additional
services beyond what's inherent to the product's Eventclass add custom
command, as controlled and described by the new C<nacltask_if_exists> option.

=over

=item Options

=over

=item C<< command_interface => $command_interface >> 
(Required)

(Required) As C<NACL::C::DFM::DFMClient>.  A component object that
represents the host to which to send commands.

=item C<< 'event-namespace-name' => $event_namespace_name >>

(Required)

=item C<< 'about-message' => $about_message >>

(Required)

=item C<< 'event-names' => $event_names >>

(Required)

=item C<< 'event-class-name' => $event_class_name >>

(Required for class method, Not Applicable for instance method)

=item C<< nacltask_if_exists => $action >>

(Optional) What to do if the eventclass to be created already exists.
If action is "reuse" (the default), then do nothing (return a 
task object referring to the existing role). If $action is "die",
then fail with an exception . 

item C<< _was_created => \$scalar >>

(Optional)
When this option is provided a reference to a scalar variable, the variable
gets filled in with a boolean value describing whether the eventclass
was found (value will be 0; this scenario is possible when 
if_exists => "reuse") or whether the eventclass was created (value will be 1).
This is necessary to determine whether the created eventclass needs to be
cleaned up later.

    my $was_created;
    my $Eventclass = NACL::STask::DFM::Eventclass->add_custom(
        nacltask_if_exists    => 'reuse',
        _was_created => \$was_created,
        %other_opts
    );
    # Operate on $Eventclass here
    # ...
    # Now determine whether to destroy the eventclass, since we're not sure
    # whether we reused an existing eventclass or created a new one
    if ($was_created) {
       
        $Eventclass->delete_custom();
    }
=back

=item command_interface, apiset_must, apiset_should, method-timeout,etc.

All of the other various options to
L<< NACL::C::DFM::Eventclass->add_custom | lib-NACL-C-DFM-Eventclass-pm/add_custom >>

=back

=cut

sub add_custom {

    $Log->enter() if $may_enter;

    my $pkg_or_obj = shift;
    my $pkg        = $pkg_or_obj->get_package_name();
    my $self;

    my %opts = $pkg_or_obj->_common_validate_with(

        params          => \@_,
        additional_spec => {
            nacltask_if_exists => { type => SCALAR,    default  => 'reuse' },
            _was_created       => { type => SCALARREF, optional => 1 },
        },
        ignore_primary_keys => 1,
        allow_extra         => 1,
    );
    my $if_exists   = delete $opts{nacltask_if_exists};
    my $was_created = delete $opts{_was_created};

CREATE: {
        use warnings qw(exiting);
        try {
            $self         = $pkg->SUPER::add_custom(%opts);
            $$was_created = 1;
        }
        catch NACL::C::Exceptions::DFM::Eventclass::AlreadyExists with {
            my $exception = shift;
            my $text      = $exception->text();
            $Log->log('The Eventclass already exists');
            $self = $pkg->_element_exists_handler(
                create_opts        => \%opts,
                nacltask_if_exists => $if_exists,
                exception          => $exception
            );
            if ( !$self ) {
                no warnings qw(exiting);
                redo CREATE;
            }
        };

    }

    $Log->exit() if $may_exit;

    ## Return the STask object
    return $self;

}

=head2 purge

    $Eventclass->purge(%other_options);

    NACL::STask::DFM::Eventclass->purge(
      command_interface   => $command_interface,
      'event-class-name' => $event_class_name
      nacltask_if_purged  => $action,     # default 'die' 
      nacltask_verify     => 0|1,	  # default 0  
      %other_options
    );

(Class or instance method) This method deletes an eventclass

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< 'event-class-name' => $event_class_name >>

(Required for class method, Not Applicable for instance method)

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< nacltask_verify => $nacltask_verify_boolean >>

(Optional)
If '0' (default), verification will not be performed.
If '1', verification will be performed to ensure that
the deletion did happen successfully.

=item C<< nacltask_if_purged => die | pass >>

(Optional, defaults to "die")
Specifies what to do if the eventclass we're trying to delete no longer
exists. A value of C<die> (the default) will result in an exception
being thrown. A value of C<pass> will mean that the exception that would
normally occur when we're trying to destroy an eventclass which does not exist
is suppressed.

=back

=back

=cut

sub purge {

    $Log->enter() if $may_enter;

    my $pkg_or_obj = shift;
    my $pkg        = $pkg_or_obj->get_package_name();

    my %opts = $pkg_or_obj->_common_validate_with(

        params          => \@_,
        additional_spec => {
            nacltask_if_purged => { type => SCALAR, default => 'die' },
            nacltask_verify    => { type => SCALAR, default => 0 },
        },
        allow_extra => 1,
    );

    my $command_interface = $opts{command_interface};

    # Task specific options, not used for purge
    # nacltask_if_purged, nacltask_verify
    my $if_purged   = delete $opts{nacltask_if_purged};
    my $verify_flag = delete $opts{nacltask_verify};

    try {

        $pkg->SUPER::delete_custom(%opts);

        if ($verify_flag) {

            $pkg->taskverify_purge(%opts);

        }
        $Log->trace("Eventclass deletion successful.");

    }
    catch NACL::C::Exceptions::DFM::Eventclass::DoesNotExist with {

        my $exception = shift;
        my $text      = $exception->text();

        if ( $if_purged !~ /pass/i ) {
            $Log->exit() if $may_exit;

            # Rethrow the exception
            $exception->throw();
        }
    };

    $Log->exit() if $may_exit;
}

## This method is an alias for purge()
sub delete_custom {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    $pkg_or_obj->purge(@_);
    $Log->exit() if $may_exit;
}

=head2 taskverify_purge

    $rbac->taskverify_purge(%other_options);

    NACL::STask::DFM::Eventclass->taskverify_purge(
      command_interface   => $command_interface,
	  'event-class-name' => $event_class_name, 
      %other_options
    );

(Class or instance method) This method verifies if a eventclass is deleted or not.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< event-class-name => $event_class_name  >>

(Required for Class method, Not Applicable for instance method) The name of the eventclass which is deleted.

=back

=back

=cut

sub taskverify_purge {
    my ( $pkg_or_obj, %opts ) = @_;
    $pkg_or_obj->_generic_purge_verify(%opts);
}
1;

