# $Id: //depot/prod/test/nacldev/lib/NACL/STask/DFM/Job.pm $
#
# Copyright (c) 2012 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Job Task Module
## @author poornima, dl-mpo-nacl-dev@netapp.com
## @change 06-12-2012 deepthis: Adding filtering option for tasks
## @status shared
## @pod here

package NACL::STask::DFM::Job;

use strict;
use warnings;

use base qw(NACL::C::DFM::Job NACL::STask::STask);

use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Params::Validate
    qw(validate SCALAR SCALARREF ARRAYREF HASHREF BOOLEAN OBJECT);
use NATE::BaseException qw(:try);
use NATE::Exceptions::Argument qw(:try);
use NACL::ComponentUtils qw(_optional_scalars);
use NACL::CS::DFM::JobTask;
use NACL::STask::Exceptions::DFM::JobWaitForStateFailure;
use NACL::STask::Exceptions::DFM::JobFailure;

use Data::Dumper;

=head1 NAME

NACL::STask::DFM::Job

=head1 DESCRIPTION

C<NACL::STask::DFM::Job> provides a number of well-defined but
potentially complex or multi-step methods related to Job in
OnCommand.

It builds on top of, and is a derived class of C<NACL::C::DFM::Job>,
and so it also provides methods that are more in the scope of
individual Job-related commands.  See C<NACL::C::DFM::Job> for
details.  This also means that a C<NACL::STask::DFM::Job> object may
generally be used in place of a component object.

=head1 METHODS

=head2 wait_for_stete

    NACL::STask::DFM::Job->wait_for_state(  command_interface => $ci,
                                        'job-list' => $job_list,
                                        'wait-for-state-timeout' => $timeout,
                                        ...
                                    );

(Class method) This method blocks until a specified timeout period elapses 
or until a job state is reached by at least one job. It returns an array of
 Job components for ids which have reached the job state.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< 'job-list' => $job_list >>
(Array)
(Required for class method, Not Applicable for instance method)

=item C<< 'status' => $job_status >>

(Optional) Status of the job once terminal

=item C<< 'state' => $job_state >>

(Optional) State of the job to be reached.

=item C<< 'wait-for-state-timeout' => $timeout >>

(Required) Number of seconds to wait for job(s) to reach the specified state.

=item C<< 'task-status-filter-on-fail' => $task_status >>

(Optional) Status of the tasks which have to be populated in the exception object,
in case the job fails. By default "error" is considered.

=item C<< 'method-timeout' => $time_in_seconds >>

(Optional, Not required if 'wait-for-state-timeout' is provided)
The NATE timeout value. If no value is specified, it is set to a value
slightly greater than the 'timeout' value (if provided) to the wait-for-completion
ZAPI. If no 'timeout' value is provided, then it is set to the NATE default value
of 150.

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=back

=back

=cut

use constant DEFAULT_FAILURE_TASK_STATUS => 'error';
use constant DEFAULT_SUCCESS_JOB_STATUS  => 'normal';

sub wait_for_state {

    $Log->enter() if $may_enter;

    my $pkg = shift;
    my (@exception_obj);
    my %opts = $pkg->_common_validate_with(
        params          => \@_,
        allow_extra     => 1,
        additional_spec => {
            'task-status-filter-on-fail' =>
                { type => SCALAR, default => DEFAULT_FAILURE_TASK_STATUS },
        },
        ignore_primary_keys => 1,
    );

    my $command_interface  = delete $opts{command_interface};
    my $task_status_filter = delete $opts{'task-status-filter-on-fail'};

    my %job_task_fetch_filter = ();
    if ( $task_status_filter ne 'all' ) {

        $job_task_fetch_filter{'status'} = $task_status_filter;
    }

    my $job_success_state
        = ( defined $opts{'status'} )
        ? delete $opts{'status'}
        : DEFAULT_SUCCESS_JOB_STATUS;

    my @job_wait_obj = $pkg->SUPER::wait_for_state(
        command_interface => $command_interface,
        %opts
    );

    my $count = 0;

    foreach my $job_wait_obj (@job_wait_obj) {

        my $job_state_obj = $job_wait_obj->state();
        my $job_status    = $job_state_obj->status();

        my %job_tasks_info = ();

        if ( $job_status ne $job_success_state ) {

            try {
                my @jobtask_obj = NACL::CS::DFM::JobTask->fetch(
                    command_interface => $command_interface,
                    filter            => {
                        'job-id' => $job_wait_obj->job_id(),
                        %job_task_fetch_filter,
                    },
                );

                $job_tasks_info{'job_tasks_info'} = \@jobtask_obj;

            }
            catch NACL::Exceptions::NoElementsFound with {

                my $exp_obj = shift;
                $Log->log( $exp_obj->text );
            };

            $job_tasks_info{'job_id'}    = $job_state_obj->job_id();
            $job_tasks_info{'job_state'} = $job_state_obj->state();
            $job_tasks_info{'job_type'}  = $job_state_obj->type();

            $Log->error( "Details of job failure #" . ++$count );
            $Log->log( "Job id: " . $job_tasks_info{'job_id'} );
            $Log->log( "Job state: " . $job_tasks_info{'job_state'} );
            $Log->log( "Job type: " . $job_tasks_info{'job_type'} );
            $Log->log("Task dump with status("
                    . $task_status_filter . "):"
                    . Dumper( $job_tasks_info{'job_tasks_info'} ) )
                if ( defined $job_tasks_info{'job_tasks_info'} );

            my $exp_obj
                = NACL::STask::Exceptions::DFM::JobWaitForStateFailure->new(
                "List of all job-wait-for-state failures",
                %job_tasks_info, );
            push( @exception_obj, $exp_obj );

        }

    }

    if ( scalar(@exception_obj) > 0 ) {

        NACL::STask::Exceptions::DFM::JobFailure->throw(
            "list of all job related info",
            failures => \@exception_obj, );

    }
    $Log->exit() if $may_exit;
}

1;
