# $Id: //depot/prod/test/main/lib/NACL/STask/DFM/StorageService.pm#1 $
#
# Copyright (c) 2012 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary StorageService Task Module
## @author Revathi N, dl-mpo-nacl-dev@netapp.com
## @status shared
## @pod here

package NACL::STask::DFM::StorageService;

use strict;
use warnings;

use base qw(NACL::C::DFM::StorageService NACL::STask::DFM);

use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Params::Validate qw(validate SCALAR SCALARREF ARRAYREF BOOLEAN OBJECT);
use NATE::BaseException qw(:try);
use NACL::ComponentUtils qw(_optional_scalars);
use NACL::MTask::DFM::Monitor;
use NACL::STask::DFM::Job;
use constant DEFAULT_METHOD_TIMEOUT => 2000;
use constant DEFAULT_CLIENT_TAG => 'STASK_PARTNER_APP';

use Data::Dumper;

=head1 NAME

NACL::STask::DFM::StorageService

=head1 DESCRIPTION

C<NACL::STask::DFM::StorageService> provides a number of well-defined but
potentially complex or multi-step methods related to StorageService in
OnCommand.

It builds on top of, and is a derived class of C<NACL::C::DFM::StorageService>,
and so it also provides methods that are more in the scope of
individual StorageService-related commands.  See C<NACL::C::DFM::StorageService> for
details.  This also means that a C<NACL::STask::DFM::StorageService> object may
generally be used in place of a component object.

=head1 METHODS

=head2 create

    my $StorageService  = NACL::STask::DFM::StorageService->create(
        command_interface  => $command_interface,
        'storage-service-name'    => $storage_service_name,
        'storage-service-topology-info'   => \%storage_service_topology_info,
        nacltask_if_exists => $action,      # default 'reuse'
        %other_options
    );

(Class method) Create a StorageService.  This method provides additional
services beyond what's inherent to the product's storageservice creation
commands, as controlled and described by the new C<nacltask_if_exists> option.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

(Required) As C<NACL::C::DFM::DFMClient>.  A component object that
represents the host to which to send commands.

=item C<< storage_service_name => $storage_service_name >>

(Required) The name of the storageservice that is to be created.

=item C<< 'storage-service-topology-info' => $hashref >>

(Required) (Hashref, can be a reference to a hash or hashofarrays or hashofhashes)

=item C<< nacltask_if_exists => $action >>

(Optional) What to do if the storageservice to be created already exists.
If $action is "reuse" (the default), then do nothing (return a
task object referring to the existing storageservice). If $action is "die",
then fail with an exception . If $action is "purge",
then delete the storageservice (see the C<destroy> method, below) before
creating a new one.

item C<< _was_created => \$scalar >>

(Optional)
When this option is provided a reference to a scalar variable, the variable
gets filled in with a boolean value describing whether the role
was found (value will be 0; this scenario is possible when
if_exists => "reuse") or whether the role was created (value will be 1).
This is necessary to determine whether the created needs to be
cleaned up later.

    my $was_created;
    my $StorageService = NACL::STask::DFM::StorageService->create(
        nacltask_if_exists    => 'reuse',
        _was_created => \$was_created,
        %other_opts
    );
    # Operate on $storageservice here
    # ...
    # Now determine whether to destroy the storageservice, since we're not sure
    # whether we reused an existing storageservice or created a new one
    if ($was_created) {
        # New Role was created. Clean it up.
        $StorageService->destroy();
    }

=item command_interface, apiset_must, apiset_should, method_timeout, storage-service-resource-key etc.

All of the other various options to
L<< NACL::C::DFM::StorageService->create | lib-NACL-C-DFM-StorageService-pm/create >>

=back

=back

=cut

sub create {

    $Log->enter() if $may_enter;

    my $pkg_or_obj = shift;
    my $pkg        = $pkg_or_obj->get_package_name();
    my $self;

    my %opts = $pkg_or_obj->_common_validate_with(

        params          => \@_,
        additional_spec => {
            nacltask_if_exists =>
              { %{ $pkg->_if_exists_validate_spec() }, default => 'reuse' },
            _was_created => { type => SCALARREF, optional => 1 },
        },
        ignore_primary_keys => 1,
        allow_extra         => 1,
    );

    my $storage_service_name = $opts{'storage-service-name'};
    $opts{'storage-service-client-tag'} ||= DEFAULT_CLIENT_TAG;
    my $cmd_interface        = $opts{command_interface};

    my $if_exists   = delete $opts{nacltask_if_exists};
    my $was_created = delete $opts{_was_created};

    my $refresh = delete $opts{refresh_monitors};
    my %refresh_monitor_options;

    # Inititalize it to 0 (false)
    $$was_created = 0;

    # Call the component
    try {

        $self         = $pkg->SUPER::create(%opts);
        $$was_created = 1;
    }
    catch NACL::APISet::Exceptions::ResponseException with {

        my $exception              = shift;
        my $text                   = $exception->text();
        my $expected_error_message = "Storage service.*already exists";

        # Throw the exception if it's not the exception message we're expecting
        # Or if the nacltask_if_exists option is die
        if (   ( $text !~ $expected_error_message )
            || ( $if_exists =~ /die/i ) )
        {

            # Rethrow the exception
            $exception->throw();

        }
        else {
            try {

                $self = $pkg->find(
                    command_interface => $cmd_interface,
                    filter =>
                      { 'storage-service-name' => $storage_service_name },
                );
            };

            # If if_exists is 'purge' then call 'purge' on the object
            if ( $if_exists =~ /purge/i ) {

                $self->purge();

                # Call add again
                $self = $pkg->SUPER::create(%opts);

                # Set flag
                $$was_created = 1;

            } ## end of $if_exists eq purge

        } ## end of else of if_exists eq purge

    };    ## end of catch

    $Log->exit() if $may_exit;

    ## Return the STask object
    return $self;
}

=head2 purge

    $StorageService->purge(%other_options);

    NACL::STask::DFM::StorageService->purge(
      command_interface   => $command_interface,
      'storage-service-resource-key'  => $storage_service_resource_key,
      nacltask_if_purged  => $action,     # default 'die'
      nacltask_verify     => 0|1,         # default 0
      %other_options
    );

(Class or instance method) This method destroys a StorageService.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<<storage_service_resource_key => $storage_service_resource_key  >>

(Required) The name or id of the StorageService to be destroyed.

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< nacltask_verify => $nacltask_verify_boolean >>

(Optional)
If '0' (default), verification will not be performed.
If '1', verification will be performed to ensure that
the deletion did happen successfully.

=item C<< nacltask_if_purged => die | pass >>

(Optional, defaults to "die")
Specifies what to do if the StorageService we're trying to delete no longer
exists. A value of C<die> (the default) will result in an exception
being thrown. A value of C<pass> will mean that the exception that would
normally occur when we're trying to destroy a StorageService which does not exist
is suppressed.

=back

=back

=cut

sub purge {

    $Log->enter() if $may_enter;

    my $pkg_or_obj = shift;
    my $pkg        = $pkg_or_obj->get_package_name();

    my %opts = $pkg_or_obj->_common_validate_with(

        params          => \@_,
        additional_spec => {
            method_timeout => { type => SCALAR, default => DEFAULT_METHOD_TIMEOUT, optional => 1 },
            nacltask_if_purged => { type => SCALAR, default => 'die' },
            nacltask_verify    => { type => SCALAR, default => 0 },
        },
        allow_extra => 1,
    );

    my $command_interface          = $opts{command_interface};
    my $storage_service_resource_key = $opts{'storage-service-resource-key'};

    # Task specific options, not used for purge
    # nacltask_if_purged, nacltask_verify
    my $if_purged      = delete $opts{nacltask_if_purged};
    my $verify_flag    = delete $opts{nacltask_verify};
    my $method_timeout = delete $opts{method_timeout};

    try {

        my $job_component = $pkg->SUPER::destroy(%opts);
        my $jobid         = $job_component->{'job_id'};

        ##Wait for the job to get complete
        NACL::STask::DFM::Job->wait_for_state(
            command_interface        => $command_interface,
            'job-list'               => [$jobid],
            'wait-for-state-timeout' => $method_timeout,
        );

        # If need to verify the storage service deletion
        if ($verify_flag) {

            $pkg->taskverify_destroy(%opts);

        }
        $Log->log("StorageService deletion successful.");

    }
    catch NACL::APISet::Exceptions::ResponseException with {

        my $exception = shift;
        my $text      = $exception->text();

        my $expected_error_message =
          "Storage service with resource key.*was not found";

        if (   ( $text !~ $expected_error_message )
            || ( $if_purged !~ /pass/i ) )
        {

            $Log->exit() if $may_exit;

            # Rethrow the exception
            $exception->throw();

        }
    };

    $Log->exit() if $may_exit;
}

## This method is an alias for purge
sub destroy {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    $pkg_or_obj->purge(@_);
    $Log->exit() if $may_exit;
}

=head2 subscribe

    $storageservice->subscribe(%other_options);

    NACL::STask::DFM::StorageService->subscribe(
        command_interface  => $command_interface,
        'storage-service-resource-key'  =>  $storage_service_resource_key,
        'subscription-context' => $string,
        'subscription-members' => @members,
        nacltask_if_object_subscribed  => $action      # default 'pass'
        %other_options
    );

=for html (Class or instance method)
This method is used to  subscribes objects to a storage service. Currently, volumes are subscribable
objects.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>
(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< 'subscription-members' => $subscription_members >>

(Required)

=item C<< 'storage-service-resource-key' => $storage_service_resource_key >>

(Required for class method, Not Applicable for instance method)

=item C<< 'subscription-context' => $subscription_context >>

(Required)

=item C<< $nacltask_if_object_subscribed  => $action >>

(Optional) What to do if the object already subscribed to the storage service.
If action is "pass" (the default), then do nothing . If $action is "die",
then fail with an exception .

item C<< _was_subscribed => \$scalar >>

(Optional)
When this option is provided a reference to a scalar variable, the variable
gets filled in with a boolean value describing whether the object
was already subsribed (value will be 0; this scenario is possible when
$if_object_subscribed => "pass") or whether the object was subscribed (value will be 1).
This is necessary to determine whether the subscribed members needs to be
cleaned up later.

    my $was_subscribed
    NACL::STask::DFM::StorageService->subscribe(
        nacltask_if_object_subscribed => 'pass',
        _was_subscribed                        => \$was_subscribed
        %other_opts
    );
    # Operate on $storageservice here
    # ...
    # Now determine whether to unsubscribe the object, since we're not sure
    # whether we reused the existing object or subscribed the object to storage service.
    if ($was_subscribed) {
        # New object was subscribed. Clean it up.
        $storageservice->unsubscribe();
    }

=back

=item command_interface, apiset_must, apiset_should, method-timeout, storage-service-resource-key etc.

All of the other various options to
L<< NACL::C::DFM::StorageService->subscribe | lib-NACL-C-DFM-StorageService-pm/subscribe >>

=back

=cut

sub subscribe {

    $Log->enter() if $may_enter;

    my $pkg_or_obj = shift;
    my $pkg        = $pkg_or_obj->get_package_name();
    my $self;

    my %opts = $pkg_or_obj->_common_validate_with(

        params          => \@_,
        additional_spec => {
            method_timeout => { type => SCALAR, default => DEFAULT_METHOD_TIMEOUT, optional => 1 },
            nacltask_if_object_subscribed =>
              { type => SCALAR, default => 'pass' },
            _was_subscribed => { type => SCALARREF, optional => 1 },
        },
        ignore_primary_keys => 1,
        allow_extra         => 1,
    );

    my $command_interface          = $opts{command_interface};
    my $storage_service_resource_key = $opts{'storage-service-resource-key'};
    my $if_object_subscribed =
      delete $opts{nacltask_if_object_subscribed};
    my $was_subscribed = delete $opts{_was_subscribed};
    my $method_timeout = delete $opts{method_timeout};

    $$was_subscribed = 0;

    # Call the component subscribe()
    try {

        my $job_component = $pkg->SUPER::subscribe(%opts);
        my $jobid         = $job_component->job_id();
        $$was_subscribed = 1;

        ##Wait for the job to get complete
        NACL::STask::DFM::Job->wait_for_state(
            command_interface        => $command_interface,
            'job-list'               => [$jobid],
            'wait-for-state-timeout' => $method_timeout,
        );

    }
    catch NACL::APISet::Exceptions::ResponseException with {

        my $exception = shift;
        my $text      = $exception->text();
        my $expected_error_message =
          "Object.*is already subscribed to storage service.*";

        # Throw the exception if it's not the exception message we're expecting
        # Or if the nacltask_if_member_already_added option is die
        if (   ( $text !~ $expected_error_message )
            || ( $if_object_subscribed =~ /die/i ) )
        {
            $Log->exit() if $may_exit;

            # Rethrow the exception
            $exception->throw();

        }
        else {
            $Log->log(
"The proposed object is already subscribed to the storage service. Hence,reusing the existing one"
            );
        }

    };    ## end of catch

    $Log->exit() if $may_exit;
}

=head2 unsubscribe

    $storageservice->unsubscribe(%other_options);

    NACL::STask::DFM::StorageService->unsubscribe(
        command_interface  => $command_interface,
        'storage-service-resource-key'  =>  $storage_service_resource_key,
        'subscription-context' => $string,
        'subscription-members' => @members,
        nacltask_if_object_already_unsubscribed  => $action      # default 'pass'
        %other_options
    );

=for html (Class or instance method)
This method is used to  unsubscribes members from an existing storage service. If member is
unsubscribed, it remains in the storage service but is marked disabled. When a
member is marked disabled, all of its protection artifacts, such as
relationships and storage objects, are preserved; however, it does not actively
participate in future replication updates or conformance checks.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>
(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>


=item C<< 'storage-service-resource-key' => $storage_service_resource_key >>

(Required for class method, Not Applicable for instance method)

=item C<< 'subscription-context' => $subscription_context >>

(Required)

=item C<< $nacltask_if_object_already_unsubscribed  => $action >>

(Optional) What to do if the object is already unsubscribed from the storage service.
If action is "pass" (the default), then do nothing . If $action is "die",
then fail with an exception .

=back

=item command_interface, apiset_must, apiset_should, method-timeout, storage-service-resource-key etc.

All of the other various options to
L<< NACL::C::DFM::StorageService->unsubscribe | lib-NACL-C-DFM-StorageService-pm/unsubscribe >>

=back

=cut

sub unsubscribe {

    $Log->enter() if $may_enter;

    my $pkg_or_obj = shift;
    my $pkg        = $pkg_or_obj->get_package_name();
    my $self;

    my %opts = $pkg_or_obj->_common_validate_with(

        params          => \@_,
        additional_spec => {
            method_timeout => { type => SCALAR, default => DEFAULT_METHOD_TIMEOUT, optional => 1 },
            nacltask_if_object_already_unsubscribed =>
              { type => SCALAR, default => 'pass' },
        },
        ignore_primary_keys => 1,
        allow_extra         => 1,
    );

    my $command_interface          = $opts{command_interface};
    my $storage_service_resource_key = $opts{'storage-service-resource-key'};
    my $if_object_unsubscribed =
      delete $opts{nacltask_if_object_already_unsubscribed};
    my $method_timeout = delete $opts{method_timeout};

    # Call the component unsubscribe()
    try {
        my $job_component = $pkg->SUPER::unsubscribe(%opts);
        my $jobid         = $job_component->job_id();

        ##Wait for the job to get complete
        NACL::STask::DFM::Job->wait_for_state(
            command_interface        => $command_interface,
            'job-list'               => [$jobid],
            'wait-for-state-timeout' => $method_timeout,
        );

    }
    catch NACL::APISet::Exceptions::ResponseException with {

        my $exception = shift;
        my $text      = $exception->text();
        my $expected_error_message =
          "Storage service subscription for context.*was not found";

        # Throw the exception if it's not the exception message we're expecting
        # Or if the nacltask_if_member_already_added option is die
        if (   ( $text !~ $expected_error_message )
            || ( $if_object_unsubscribed =~ /die/i ) )
        {
            $Log->exit() if $may_exit;

            # Rethrow the exception
            $exception->throw();

        }
        else {
            $Log->log(
"The proposed object is already unsubscribed from existing storage service."
            );
        }

    };    ## end of catch

    $Log->exit() if $may_exit;

}

=head2 conform

    $storageservice->conform(%other_options);

    NACL::STask::DFM::StorageService->conform(
        command_interface  => $command_interface,
        'storage-service-resource-key'  =>  $storage_service_resource_key,
        'subscription-context' => $string,
        'members' => \@members,
        %other_options
    );

=for html (Class or instance method)
This method is used to  conforms objects to a storage service. Currently, volumes are subscribable
objects.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>
(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< 'members' => $members >>

(Required)

=item C<< 'storage-service-resource-key' => $storage_service_resource_key >>

(Required for class method, Not Applicable for instance method)

=item C<< 'subscription-context' => $subscription_context >>

(Required)

=back

=item command_interface, apiset_must, apiset_should, method-timeout, storage-service-resource-key etc.

All of the other various options to
L<< NACL::C::DFM::StorageService->conform | lib-NACL-C-DFM-StorageService-pm/conform >>

=back

=cut

sub conform {

    $Log->enter() if $may_enter;

    my $pkg_or_obj = shift;
    my $pkg        = $pkg_or_obj->get_package_name();
    my $self;

    my %opts = $pkg_or_obj->_common_validate_with(

        params          => \@_,
        additional_spec => {
            method_timeout => { type => SCALAR, default => DEFAULT_METHOD_TIMEOUT, optional => 1 },
        },
        ignore_primary_keys => 1,
        allow_extra         => 1,
    );

    my $command_interface          = $opts{command_interface};
    my $storage_service_resource_key = $opts{'storage-service-resource-key'};
    my $method_timeout = delete $opts{method_timeout};

    # Call the component conform()
    try {

        my $job_component = $pkg->SUPER::conform(%opts);
        my $jobid         = $job_component->job_id();

        ##Wait for the job to get complete
        NACL::STask::DFM::Job->wait_for_state(
            command_interface        => $command_interface,
            'job-list'               => [$jobid],
            'wait-for-state-timeout' => $method_timeout,
        );

    }
    catch NACL::APISet::Exceptions::ResponseException with {

        my $exception = shift;
        $Log->exit() if $may_exit;
        # Rethrow the exception
        $exception->throw();

    };    ## end of catch

    $Log->exit() if $may_exit;
}

=head2 protection_update_start

    $storageservice->protection_update_start(%other_options);

    NACL::STask::DFM::StorageService->protection_update_start(
        command_interface  => $command_interface,
        'storage-service-resource-key'  =>  $storage_service_resource_key,
        'subscription-context' => $string,
        'members' => \@members,
        'member-snapshots' => \@snapshots
        %other_options
    );

=for html (Class or instance method)
This method is used to  protection_update_starts objects to a storage service. Currently, volumes are subscribable
objects.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>
(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< 'members' => $members >>

(Optional)

=item C<< 'storage-service-resource-key' => $storage_service_resource_key >>

(Required for class method, Not Applicable for instance method)

=item C<< 'subscription-context' => $subscription_context >>

(Optional)

=item C<< 'member-snapshots' => \@snapshots >>
Reference to an array of snapshot resource keys

(Optional)

=back

=item command_interface, apiset_must, apiset_should, method-timeout, storage-service-resource-key etc.

All of the other various options to
L<< NACL::C::DFM::StorageService->protection_update_start | lib-NACL-C-DFM-StorageService-pm/protection_update_start >>

=back

=cut

sub protection_update_start {

    $Log->enter() if $may_enter;

    my $pkg_or_obj = shift;
    my $pkg        = $pkg_or_obj->get_package_name();
    my $self;

    my %opts = $pkg_or_obj->_common_validate_with(

        params          => \@_,
        additional_spec => {
            method_timeout => { type => SCALAR, default => DEFAULT_METHOD_TIMEOUT, optional => 1 },
            nacltask_verify_job_status => { type => SCALAR, default => 'normal' },
        },
        ignore_primary_keys => 1,
        allow_extra         => 1,
    );

    my $command_interface          = $opts{command_interface};
    my $storage_service_resource_key = $opts{'storage-service-resource-key'};
    my $method_timeout = delete $opts{method_timeout};
    my $job_status = delete $opts{nacltask_verify_job_status};

    # Call the component protection_update_start()
    try {

        my $job_component = $pkg->SUPER::protection_update_start(%opts);
        my $jobid         = $job_component->job_id();

        ##Wait for the job to get complete
        NACL::STask::DFM::Job->wait_for_state(
            command_interface        => $command_interface,
            'job-list'               => [$jobid],
            'wait-for-state-timeout' => $method_timeout,
            status                   => $job_status,
        );

    }
    catch NACL::APISet::Exceptions::ResponseException with {

        my $exception = shift;
        $Log->exit() if $may_exit;
        # Rethrow the exception
        $exception->throw();

    };    ## end of catch

    $Log->exit() if $may_exit;
}

=head2 cleanup

    $storageservice->cleanup(%other_options);

    NACL::STask::DFM::StorageService->cleanup(
        command_interface  => $command_interface,
        'storage-service-resource-key'  =>  $storage_service_resource_key,
        %other_options
    );

=for html (Class or instance method)
This method is used to remove the specified member, all the relevant data protection
relationships and relevant provisioned storage objects from the storage service.
Specified members must not have any active subscriptions.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>
(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>


=item C<< 'storage-service-resource-key' => $storage_service_resource_key >>

(Required for class method, Not Applicable for instance method)

=back

=item command_interface, apiset_must, apiset_should, method-timeout, storage-service-resource-key etc.

All of the other various options to
L<< NACL::C::DFM::StorageService->cleanup | lib-NACL-C-DFM-StorageService-pm/cleanup >>

=back

=cut

sub cleanup {

    $Log->enter() if $may_enter;

    my $pkg_or_obj = shift;
    my $pkg        = $pkg_or_obj->get_package_name();
    my $self;

    my %opts = $pkg_or_obj->_common_validate_with(

        params          => \@_,
        additional_spec => {
            method_timeout => {
                type     => SCALAR,
                default  => DEFAULT_METHOD_TIMEOUT,
                optional => 1
            },
        },
        ignore_primary_keys => 1,
        allow_extra         => 1,
    );

    my $command_interface            = $opts{command_interface};
    my $storage_service_resource_key = $opts{'storage-service-resource-key'};
    my $method_timeout               = delete $opts{method_timeout};

    # Call the component unsubscribe()
    try {
        my $job_component = $pkg->SUPER::cleanup(%opts);
        my $jobid         = $job_component->job_id();

        ##Wait for the job to get complete
        NACL::STask::DFM::Job->wait_for_state(
            command_interface        => $command_interface,
            'job-list'               => [$jobid],
            'wait-for-state-timeout' => $method_timeout,
        );

    }
    catch NACL::APISet::Exceptions::ResponseException with {

        my $exception = shift;

        # Rethrow the exception
        $exception->throw();

    };    ## end of catch

    $Log->exit() if $may_exit;

}

=head2 import_relationships

    $storageservice->import_relationships(%other_options);

    NACL::STask::DFM::StorageService->import_relationships(
        command_interface  => $command_interface,
        'storage-service-resource-key'  =>  $storage_service_resource_key,
        %other_options
    );

=for html (Class or instance method)
This method is used to import non-managed storage objects into an existing
storage service. A relationship can only be imported into one storage service.
Source and destination volumes of the relationship can be associated with only
 a single node of the same storage service.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>
(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>


=item C<< 'storage-service-resource-key' => $storage_service_resource_key >>

(Required for class method, Not Applicable for instance method)

=back

=item command_interface, apiset_must, apiset_should, method-timeout, storage-service-resource-key etc.

All of the other various options to
L<< NACL::C::DFM::StorageService->import_relationships | lib-NACL-C-DFM-StorageService-pm/import_relationships >>

=back

=cut

sub import_relationships {

    $Log->enter() if $may_enter;

    my $pkg_or_obj = shift;
    my $pkg        = $pkg_or_obj->get_package_name();
    my $self;

    my %opts = $pkg_or_obj->_common_validate_with(

        params          => \@_,
        additional_spec => {
            method_timeout => {
                type     => SCALAR,
                default  => DEFAULT_METHOD_TIMEOUT,
                optional => 1
            },
        },
        ignore_primary_keys => 1,
        allow_extra         => 1,
    );

    my $command_interface            = $opts{command_interface};
    my $storage_service_resource_key = $opts{'storage-service-resource-key'};
    my $method_timeout               = delete $opts{method_timeout};

    # Call the component unsubscribe()
    try {
        my $job_component = $pkg->SUPER::import_relationships(%opts);
        my $jobid         = $job_component->job_id();

        ##Wait for the job to get complete
        NACL::STask::DFM::Job->wait_for_state(
            command_interface        => $command_interface,
            'job-list'               => [$jobid],
            'wait-for-state-timeout' => $method_timeout,
        );

    }
    catch NACL::APISet::Exceptions::ResponseException with {

        my $exception = shift;

        # Rethrow the exception
        $exception->throw();

    };    ## end of catch

    $Log->exit() if $may_exit;

}

=head2 modify

    $storageservice->modify(%other_options);

    NACL::STask::DFM::StorageService->modify(
        command_interface  => $command_interface,
        'storage-service-resource-key'  =>  $storage_service_resource_key,
        %other_options
    );

=for html (Class or instance method)
This method is used to modify the attributes of an existing storage service.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>
(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>


=item C<< 'storage-service-resource-key' => $storage_service_resource_key >>

(Required for class method, Not Applicable for instance method)

=back

=item command_interface, apiset_must, apiset_should, method-timeout, storage-service-resource-key etc.

All of the other various options to
L<< NACL::C::DFM::StorageService->modify | lib-NACL-C-DFM-StorageService-pm/modify >>

=back

=cut

sub modify {

    $Log->enter() if $may_enter;

    my $pkg_or_obj = shift;
    my $pkg        = $pkg_or_obj->get_package_name();
    my $self;

    my %opts = $pkg_or_obj->_common_validate_with(

        params          => \@_,
        additional_spec => {
            method_timeout => {
                type     => SCALAR,
                default  => DEFAULT_METHOD_TIMEOUT,
                optional => 1
            },
        },
        ignore_primary_keys => 1,
        allow_extra         => 1,
    );

    my $command_interface            = $opts{command_interface};
    my $storage_service_resource_key = $opts{'storage-service-resource-key'};
    my $method_timeout               = delete $opts{method_timeout};

    # Call the component modify()
    try {
        my $job_component = $pkg->SUPER::modify(%opts);
        my $jobid         = $job_component->job_id();

        ##Wait for the job to get complete
        NACL::STask::DFM::Job->wait_for_state(
            command_interface        => $command_interface,
            'job-list'               => [$jobid],
            'wait-for-state-timeout' => $method_timeout,
        );

    }
    catch NACL::APISet::Exceptions::ResponseException with {

        my $exception = shift;
        $Log->exit() if $may_exit;

        # Rethrow the exception
        $exception->throw();

    };    ## end of catch

    $Log->exit() if $may_exit;

}

=head2 subscribe_and_conform

    $storageservice->subscribe_and_conform(%other_options);

    NACL::STask::DFM::StorageService->subscribe_and_conform(
        command_interface  => $command_interface,
        'storage-service-resource-key'  =>  $storage_service_resource_key,
        'subscription-context' => $string,
        'subscription-members' => @members,
        %other_options
    );

=for html (Class or instance method)
This method is used to  subscribe and conforms objects to a storage service.
Currently, volumes are subscribable
objects.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>
(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< 'subscription-members' => $subscription_members >>

(Required)

=item C<< 'storage-service-resource-key' => $storage_service_resource_key >>

(Required for class method, Not Applicable for instance method)

=item C<< 'subscription-context' => $subscription_context >>

(Required)

=back

=item command_interface, apiset_must, apiset_should, method-timeout, storage-service-resource-key etc.

All of the other various options to
L<< NACL::C::DFM::StorageService->subscribe_and_conform | lib-NACL-C-DFM-StorageService-pm/subscribe_and_conform >>

=back

=cut

sub subscribe_and_conform {

    $Log->enter() if $may_enter;

    my $pkg_or_obj = shift;
    my $pkg        = $pkg_or_obj->get_package_name();
    my $self;

    my %opts = $pkg_or_obj->_common_validate_with(

        params          => \@_,
        additional_spec => {
            method_timeout => { type => SCALAR, default => DEFAULT_METHOD_TIMEOUT, optional => 1 },
        },
        ignore_primary_keys => 1,
        allow_extra         => 1,
    );

    $pkg_or_obj->subscribe(%opts);
    $opts{members} = delete( $opts{'subscription-members'} );
    delete( $opts{'subscription-context'} );
    $pkg_or_obj->conform(%opts);

    $Log->exit() if $may_exit;
}

=head2 import_and_conform

    $storageservice->import_and_conform(%other_options);

    NACL::STask::DFM::StorageService->import_and_conform(
        command_interface  => $command_interface,
        'storage-service-resource-key'  =>  $storage_service_resource_key,
        'subscription-context' => $string,
        'subscription-members' => @members,
        %other_options
    );

=for html (Class or instance method)
This method is used to import and conforms objects to a storage service.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>
(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< 'subscription-members' => $subscription_members >>

(Required)

=item C<< 'storage-service-resource-key' => $storage_service_resource_key >>

(Required for class method, Not Applicable for instance method)

=item C<< 'subscription-context' => $subscription_context >>

(Required)

=back

=item command_interface, apiset_must, apiset_should, method-timeout, storage-service-resource-key etc.

All of the other various options to
L<< NACL::C::DFM::StorageService->import_and_conform | lib-NACL-C-DFM-StorageService-pm/import_and_conform >>

=back

=cut


sub import_and_conform {

    $Log->enter() if $may_enter;

    my $pkg_or_obj = shift;
    my $pkg        = $pkg_or_obj->get_package_name();
    my $self;

    my %opts = $pkg_or_obj->_common_validate_with(

        params          => \@_,
        additional_spec => {
            method_timeout => {
                type     => SCALAR,
                default  => DEFAULT_METHOD_TIMEOUT,
                optional => 1
            },
        },
        ignore_primary_keys => 1,
        allow_extra         => 1,
    );

    $pkg_or_obj->import_relationships(%opts);
    $opts{members} = delete( $opts{'subscription-members'} );
    delete( $opts{'subscription-context'} );
    $pkg_or_obj->conform(%opts);

    $Log->exit() if $may_exit;
}

###########################################################################
# name :taskverify_destroy
# This method is used to verify whether the storage service is purged or not

###########################################################################

sub taskverify_destroy {
    my $pkg_or_obj = shift;
    my %opts       = $pkg_or_obj->_common_validate_with( params => \@_ );
    my $pkg        = ref $pkg_or_obj || $pkg_or_obj;
    my %common_params;
    $pkg->_move_common_component_params_with_ci(
        source => \%opts,
        target => \%common_params
    );

    try {
        $pkg->find( %common_params, filter => {%opts}, allow_empty => 1 );
    }
    catch NACL::APISet::Exceptions::ResponseException with {

        my $exception = shift;
        my $text      = $exception->text();

        my $expected_error_message = "Storage service with resource key.*was not found";
        if ( $text !~ $expected_error_message ) {
            $Log->log('Verification failed. Storage service still exist');

            # Rethrow the exception
            $exception->throw();
            $Log->exit() if $may_exit;
        }

    };
}

1;
