#
# Copyright (c) 2014 NetApp, Inc., All Rights Reserved
# Any use, modification, or distribution is prohibited
# without prior written consent from NetApp, Inc.

## @summary DebugCrsServiceLevel Task Module
## @author dl-nacl-dev@netapp.com
## @status shared
## @pod here


package NACL::STask::DebugCrsServiceLevel;

use strict;
use warnings;

use base qw(NACL::C::DebugCrsServiceLevel NACL::STask::STask);
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Params::Validate qw(validate SCALAR OBJECT SCALARREF HASHREF BOOLEAN ARRAYREF);
use NACL::ComponentUtils qw(_optional_scalars Dumper);

use NACL::CS::DebugCrsServiceLevelModules;
use NACL::Exceptions::Timeout;
use NATE::BaseException;
use NACL::APISet::Exceptions::ResponseException qw(:try);

=head1 NAME

NACL::STask::DebugCrsServiceLevel

=head1 DESCRIPTION

C<NACL::STask::DebugCrsServiceLevel>provides a few helper methods that
are related to the CRS Service level in ONTAP.

CRS stands for Configuration Replication Service which is an internally
used for Metro-cluster and Vserver DR scenarios of backing up configurations.

=head1 METHODS

=head2 wait_for_module_state

  NACL::STask::DebugCrsServiceLevel->wait_for_crs_agent(
        command_interface => $command_interface,
      );

(Class Method) Method that will check the service level module state of
the passed in 'module-name' for the 'oper-state' to occur. This defaults
to the CrsAgent module and the acquired_master_duties state.

=over

=item C<< command_interface => $command_interface >>

(Required) A component object that represents the host that is used for
sending messages to the other side.

=item C<< "method-timeout" => 300 >>

(Optional) As component method-timeout, controls how long the command
will wait before completing.  However, the default timeout has been
raised to 300 seconds.

=item C<< "poll-interval" => 10 >>

(Optional) This interval will be the basis for sleeping in between
waiting for the acquiring of master duties for the CrsAgent process.

=item C<< "module-name" => 'CrsAgent' >>

(Optional) This is the module to wait for.

=item C<< "oper-state" => 'acquired_master_duties' >>

(Optional) This is the state to wait for the module to be in.

=back

=cut

sub wait_for_module_state {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = $pkg->_common_validate_with(
        ignore_primary_keys => 1,
        params          => \@_,
        additional_spec => {
            'method-timeout'    => { type => SCALAR, default => 300},
            'poll-interval'     => { type => SCALAR, default => 10 },
            'module-name'       => { type => SCALAR, default => 'CrsAgent'},
            'oper-state'        => { type => SCALAR, default => 'acquired_master_duties'},
        },
        allow_extra => 1,
    );

    my %common_opts;
    $pkg->_copy_common_component_params(
        source => \%opts,
        target => \%common_opts
    );

    my %filter = ();
    $pkg->_hash_copy(
            source => \%opts,
            target => \%filter,
            copy   => [qw(oper-state module-name)]
        );

    my @interfaces = NACL::C::Node->find(command_interface => $opts{command_interface});
    my $found = 0;
    my $start = [ Time::HiRes::gettimeofday( ) ];
    while(Time::HiRes::tv_interval( $start ) < $opts{'method-timeout'} ) {
        
        my @states = ();
        foreach my $interface (@interfaces) {
            @states = NACL::CS::DebugCrsServiceLevelModules->fetch(
                    command_interface => $interface,
                    'filter'          => \%filter,
                    allow_empty => 1,
                    %common_opts,
                );

            if(@states) {
                last;
            }
        }

        if(@states) {
            $found = 1;
            last;
        } else {
            sleep($opts{'poll-interval'});
        }
    }

    if(!$found ) {
        $Log->exit() if $may_exit;
        NACL::Exceptions::Timeout->throw("wait_for_module_state timed out over $opts{'method-timeout'}", timeout => $opts{'method-timeout'});
    }

    $Log->exit() if $may_exit;
}

=head2 wait_for_master_duties

  NACL::STask::DebugCrsServiceLevel->wait_for_master_duties(
        command_interface => $command_interface,
      );

(Class Method) Method that is used to check all the service level for the
'acquired_master_duties'. Also will wait for it to be acquired.

=over

=item C<< command_interface => $command_interface >>

(Required) A component object that represents the host that is used for
sending messages to the other side.

=item C<< "method-timeout" => 300 >>

(Optional) As component method-timeout, controls how long the command
will wait before completing.  However, the default timeout has been
raised to 300 seconds.

=item C<< "poll-interval" => 10 >>

(Optional) This interval will be the basis for sleeping in between
waiting for the acquiring of master duties.

=back

=cut

sub wait_for_master_duties {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = $pkg->_common_validate_with(
        ignore_primary_keys => 1,
        params          => \@_,
        additional_spec => {
            'poll-interval'     => { type => SCALAR, default => 10 },
            'method-timeout'    => { type => SCALAR, default => 300},
        },
        allow_extra => 1,
    );

    my %common_opts;
    $pkg->_copy_common_component_params(
        source => \%opts,
        target => \%common_opts
    );
    my @interfaces = NACL::C::Node->find(command_interface => $opts{command_interface});
    my $found = 0;
    my $start = [ Time::HiRes::gettimeofday( ) ];
    while(Time::HiRes::tv_interval( $start ) < $opts{'method-timeout'} ) {

        my @states = ();
        foreach my $interface (@interfaces) {
            @states = NACL::CS::DebugCrsServiceLevel->fetch(
                    command_interface => $interface,
                    'filter'          => {
                        'oper-state'  => 'acquired_master_duties',
                    },
                    allow_empty => 1,
                    %common_opts,
                );

            if(@states) {
                last;
            }
        }

        if(@states) {
            $found = 1;
            last;
        } else {
            sleep($opts{'poll-interval'});
        }
    }

    if(!$found ) {
        $Log->exit() if $may_exit;
        NACL::Exceptions::Timeout->throw("wait_for_master_duties timed out over $opts{'method-timeout'}", timeout => $opts{'method-timeout'});
    }

    $Log->exit() if $may_exit;
}

1;
