#
# Copyright (c) 2009-2011 NetApp Inc.
# All rights reserved.
#
## @summary Provide an exception class that is a subclass of
##          NACL::Exceptions::NoElementsFound that will be thrown when the
##          call to NACL::STask::Volume->create fails because there are
##          no aggregates with sufficient space for the volume
##
## @author madhavs@netapp.com, dl-nacl-dev@netapp.com
## @status Shared
## @pod here

=head1 NAME

NACL::STask::Exceptions::NoAggregatesWithSufficientSpaceForVolume

=head1 DESCRIPTION

If a call to L<< NACL::STask::Volume->create|lib-NACL-STask-Volume-pm/create >>
is made without the C<aggregate> argument, then the method attempts to find
online aggregates with sufficient available space to hold the aggregate. If
there are none found, then this type of exception is thrown.

=head1 EXAMPLES

=head2 Throwing exception

    NACL::STask::Exceptions::NoAggregatesWithSufficientSpaceForVolume qw(:try);
    
    NACL::STask::Exceptions::NoAggregatesWithSufficientSpaceForVolume->throw(
        text => 'No aggregates exist with sufficient size',
        size => $size
    );

=head2 Catching and handling exception

    try {
        NACL::STask::Volume->create(command_interface => $ci);
    } catch NACL::STask::Exceptions::NoAggregatesWithSufficientSpaceForVolume with {
        my $exception = shift;
        
        # Accessing the error message
        $Log->comment('Error: ' . $exception->text());
        
        # Get the size required
        $Log->comment('Error: ' . $exception->size());
    };

=cut

package NACL::STask::Exceptions::NoAggregatesWithSufficientSpaceForVolume;

use strict;
use warnings;

use Params::Validate qw(validate SCALAR);
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();

use base 'NACL::Exceptions::NoElementsFound';

use Class::MethodMaker [ scalar => 'size', ];

sub new {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = validate(
        @_,
        {   text => { type => SCALAR },
            size => { type => SCALAR }
        }
    );
    my $text = delete $opts{text};
    my $self = $pkg->SUPER::new( $text, %opts );

    $Log->exit() if $may_exit;
    return $self;
}

1;
