#
# Copyright (c) 2009-2011 NetApp Inc.
# All rights reserved.
#
## @summary Provide an exception class that is a subclass of
##          NATE::BaseException that will be thrown when the
##          call to NACL::STask::Volume->create fails because no vservers
##          with the suitable aggr-list can be found.
##
## @author madhavs@netapp.com, dl-nacl-dev@netapp.com
## @status Shared
## @pod here

=head1 NAME

NACL::STask::Exceptions::NoVserversWithSuitableAggrList

=head1 DESCRIPTION

If a call to L<< NACL::STask::Volume->create|lib-NACL-STask-Volume-pm/create >>
is made without the C<vserver> argument, then the method attempts to find a
vserver which can host the volume. This is done by analyzing the aggr-list
of each of the available data-vservers and choosing a vserver which
contains in its aggr-list an aggregate which can be used to host the
volume (has sufficient space and is online). If no such vservers could be
found, this this type of exception is thrown.

=head1 EXAMPLES

=head2 Throwing exception

    NACL::STask::Exceptions::NoVserversWithSuitableAggrList->throw(
        text => 'No vservers with suitable aggr-list',
        
        # The names of aggregates "suitable" for the vserver ("suitable"
        # being defined as those aggregates which have sufficient space
        # and are online)
        aggregates      => \@aggr_names,
        
        # An array of Vserver CS objects which are expected to, at a minimum,
        # contain the value for "aggr-list"
        vserver_cs_objs => \@vserver_cs_objs
    );

=head2 Catching and handling exception

    try {
        $vol = NACL::STask::Volume->create(command_interface => $node);
    } catch NACL::STask::Exceptions::NoVserversWithSuitableAggrList with {
        my $exception = shift;
        
        # Getting the error text is the same as for other exceptions
        $Log->comment('Error text: ' . $exception->text());
        
        # Getting the list of "suitable" aggregates
        $Log->comment('Suitable aggregates: ' . Dumper($exception->aggregates()));
        
        # Getting the Vserver CS objects and the aggr-list of each.
        my @vserver_cs_objs = $exception->vserver_cs_objs();
        foreach my $cs_obj (@vserver_cs_objs) {
            $Log->comment('Vserver: ' . $cs_obj->vserver());
            $Log->comment('Aggr-list: ' . Dumper($cs_obj->aggr_list()));
        }
    };

=head1 ATTRIBUTES

The values of each of the attributes can be accessed using the accessor
methods. Using the C<aggregates> attribute as an example, its value can be
accessed by doing:

    my $aggregates = $exception->aggregates();

Note that those fields listed as array fields, can be accessed in either
scalar or list context:

    # Scalar context, returns an array-ref
    my $aggregates_arrayref = $exception->aggregates();
    
    # List context, returns an array
    my @aggregates = $exception->aggregates();

=over

=item C<aggregates>

(Array)
The list of "suitable" aggregates for this volume. ("suitable" being defined
as those aggregates which have sufficient space and are online)

=item C<vserver_cs_objs>

(Array)
An array of Vserver CS objects which are expected to, at a minimum,
contain the value for "aggr-list"

=back

=cut

package NACL::STask::Exceptions::NoVserversWithSuitableAggrList;

use strict;
use warnings;

use Params::Validate qw(validate SCALAR ARRAYREF);
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();

use base 'NATE::BaseException';

use Class::MethodMaker [
    array => 'aggregates',
    array => 'vserver_cs_objs'
];

sub new {
    $Log->enter();
    my $pkg  = shift;
    my %opts = validate(
        @_,
        {   text            => { type => SCALAR },
            aggregates      => { type => ARRAYREF },
            vserver_cs_objs => { type => ARRAYREF }
        }
    );
    my $text = delete $opts{text};
    my $self = $pkg->SUPER::new( $text, %opts );

    $Log->exit();
    return $self;
}

1;
