# $Id$
#
# Copyright (c) 2009-2011 NetApp Inc.
# All rights reserved.
#
## @summary Provide an exception class that is a subclass of
##          NATE::BaseException that will be thrown when the
##          call to NACL::STask::Volume->create fails because no vservers
##          with the suitable allowed protocols can be found.
##
## @author chethanm@netapp.com, dl-nacl-dev@netapp.com
## @status Shared
## @pod here

=head1 NAME

NACL::STask::Exceptions::NoVserversWithSuitableAllowedProtocols

=head1 DESCRIPTION

If a call to L<< NACL::STask::Volume->create|lib-NACL-STask-Volume-pm/create >>
is made without the C<vserver> argument, then the method attempts to find a
vserver which can host the volume. This is done by analyzing the allowed-protocols
of each of the available data-vservers and choosing a vserver which
matches specified allowed-protocols which can be used to host the
volume (has sufficient space and is online). If no such vservers could be
found, this this type of exception is thrown.

=head1 EXAMPLES

=head2 Throwing exception

    NACL::STask::Exceptions::NoVserversWithSuitableAllowedProtocols->throw(
        text => 'No vservers with suitable allowed protocols',
    );

=head2 Catching and handling exception

    try {
        $vol = NACL::STask::Volume->create(
            command_interface         => $node,
            nacltask_usable_protocols => ['nfs', 'cifs']
        );
    }
    catch NACL::STask::Exceptions::NoVserversWithSuitableAllowedProtocols with {
        my $exception = shift;

        # Getting the error text is the same as for other exceptions
        $Log->comment('Error text: ' . $exception->text());
    };

=cut

package NACL::STask::Exceptions::NoVserversWithSuitableAllowedProtocols;

use strict;
use warnings;

use Params::Validate qw(validate SCALAR);
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();

use base 'NATE::BaseException';

sub new {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = validate( @_, { text => { type => SCALAR }, } );
    my $text = delete $opts{text};
    my $self = $pkg->SUPER::new( $text, %opts );

    $Log->exit() if $may_exit;
    return $self;
}

1;
