# Copyright (c) 2014 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#
## @summary Fiji Task Module
## @author dl-nacl-dev@netapp.com
## @status shared
## @pod here

package NACL::STask::Fiji;

use strict;
use warnings;
use NACL::ComponentUtils qw(Dumper);
use NACL::STask::Cluster;
use base qw(NACL::C::Component NACL::STask::STask);
use NACL::C::Fiji;

use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Params::Validate qw(validate OBJECT SCALAR BOOLEAN ARRAYREF UNDEF HASHREF SCALARREF);
use NATE::BaseException qw(:try);
use NACL::Exceptions::Timeout qw(:try);
use NACL::STask::Exceptions::FijiClearFailure;
use NATE::Database::Test;
use NACL::GeneralUtils qw(nacl_method_retry);

# TASK related constants;
use constant WAIT_TIME => 10;
my $FIJI_COMPLETION_TIMEOUT = 300;
my $default_tag;
my $testcase_id;

=head1 NAME

NACL::STask::Fiji

=head1 DESCRIPTION

This module provides the task layer methods related to the fiji operations
on the Node. These operations consist of fiji set(), clear(), get_all_slots(),
 active() and remaining_attempts().
Fiji is 7mode functionality and works through nodeshell in CMode. The fiji
commands are available on debug or fiji enabled non debug builds.

find() and state() methods are not supported in this STask.
Please use NACL::C::Fiji->find() or NACL::CS::Fiji->fetch().

=head1 METHODS

=head2 new

Create a FIJI object

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, not applicable for instance method.) See
L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>.

=item C<< tag => string >>

A descriptive name of the FIJI (case is not significant)

=item C<< program => string >>

The actual fiji program, with sprintf-like format descriptors where arguments
should be substituted.Program that will be construct using tag and nacltask_param
Presence of program value with in double quotes makes it to log in to the node shell.
For example, program =>
            "\"(rsm.nvlog.write.fail.no.space(eq \@vol %s) (return 1))\""
Presence of program value with in single quotes makes it to log in to the NGSH.
For example, program =>
            "\'(rsm.nvlog.write.fail.no.space(eq \@vol %s) (return 1))\'"

=item C<< slot => scalar >>

Slot name where fiji needs to be set

=item C<< nodescope-node-name => $Node >>

The actual nodename on which fiji is executed. If not specified fiji will be
executed on command_interface provided.

=back

=item Examples

    $Fiji_obj = NACL::STask::Fiji->new(
        command_interface => $Node,
        tag               => 'Simulate NVLOG out of space',
        program =>
            "\'(rsm.nvlog.write.fail.no.space(eq \@vol %s) (return 1))\'"
        'nodescope-node-name' => $Node->node(),
    );

    $Fiji_obj = NACL::STask::Fiji->new(
        command_interface => $Node,
        tag    => 'My generic FIJI',
        program   => '(%s (%s) (%s))',
     );

=back

Note that this does NOT set the FIJI on the filer, it only
creates the object.

=cut

use Class::MethodMaker [
    new    => [ '-hash', '-init', 'new' ],
    scalar => [
        { -type => 'NACL::C::CommandInterface::ONTAP' },
        'command_interface'
    ],
    scalar => [
        {   '-default_ctor' => sub {
                $testcase_id = $NATE::Result::Id_Path[-1];
                $default_tag = NATE::Database::Test::testdb_global()
                    ->get_testcase(testcase_id => $testcase_id)->{name};
                $Log->trace("default_tag is $default_tag");
                return $default_tag;
                }
        },
        'tag'
    ],
    scalar => 'program',
    scalar => 'slot',
    scalar => 'nodescope_node_name',
];

=head2 set

(Instance or Class Method)

Set a FIJI on Node.

This will activate the FIJI on node with which it was associated
when the object was created (see ::new).

=over

=item Options

=over

=item C<< nacltask_param => string >>

Further arguments are used to assemble the FIJI. They should be
the parameters specified by the C<nacltask_param> argument with the
FIJI was registered.
If value of nacltask_param is a scalar, it is taken to mean that the FIJI program
has a single, non-optional parameter, and the value of nacltask_param is its name.
For example: nacltask_param => 'volname'.

If the value of nacltask_param is a hash with a single element, it is taken to mean
that the FIJI program has a single parameter, but it is optional.
The name of the parameter is the hash key and the default value is the
hash value. For example: nacltask_param => { action => 'return 1' }.
This hash may contain only one element.

If the value of nacltask_param is an array, the FIJI may have more than one parameter.
Each element of the array may be either a scalar or a hash with a single
element (as described above). The order of the elements of the array
specify the order of the parameter substitution.

=item C<< nacltask_nvram => boolean >>

(Optional)If nvram is 1, set the fiji in nvram

=item C<< nacltask_wait => boolean >>

(Optional)If nacltask_wait is 1, wait till fiji is set

=item C<< nacltask_wait_interval => boolean >>

(Optional)If nacltask_wait_interval is time of wait for setting fiji

=item C<< 'nodescope-node-name' => string >>

(Optional) run fiji command on specified node. (Default = 'local')

=item C<< nacltask_retry_tries_count => $tries_count >>

(Optional, defaults to 5)
The number of times to retry the create command if it fails because
the file is temporarily busy.

Note that it is possible to make it not retry the create command by setting
the value of this argument to 1.

=item C<< nacltask_retry_sleep_time => $sleep_time >>

(Optional, defaults to 30 seconds)
The time to sleep between retries of the create command

=item C<< 'method-timeout' => 60 >>

(Optional) As component method-timeout, controls how long the command
will wait before completing.Default timeout is 60 seconds.

=item C<< nacltask_retry_num_times_code_ran => \$variable >>

(Optional)
A reference to a scalar variable which gets filled in with the number
of times the resume command was invoked.

=back

=item Examples

    $fiji_obj->set(nacltask_param => ['component_test_vol']);

    or

    NACL::STask::Fiji->set(
        command_interface => $Node,
        tag =>
            'Cause resync to fail by making marking of volume as a duplicate fail',
        program => "\'(idump.resync.fail.mark (eq \@vol %s)(return 1))\'",
        nacltask_param   => [ 'component_test_vol'],
        nacltask_nvram => 1
    );

=item Exceptions

=over

=item C<NATE::BaseException>

This type of exception is thrown when there is any failure during setting FIJIi or
When too many arguments are passed for FIJI or
When there is any failure to assemble fiji string.

=back

=back

=cut

sub set {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my ( $command_interface, $tag, $fiji_prog );

    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@_,
        additional_spec => {
            nacltask_param         => { type => ARRAYREF, optional => 1 },
            nacltask_nvram         => { type => SCALAR,   default  => 0 },
            nacltask_wait          => { type => BOOLEAN,  default  => 0 },
            nacltask_wait_interval => { type => SCALAR,   optional => 1 },
            nacltask_retry_tries_count => { type => SCALAR,  default => 5 },
            nacltask_retry_sleep_time  => { type => SCALAR,  default => 30 },
            'method-timeout'             => { type => SCALAR,  default => 60 },
            nacltask_retry_num_times_code_ran =>
                { type => SCALARREF, optional => 1 },
        },
        allow_extra => 1,
    );
    $Log->debug( "Opts to 'set' fiji are %opts:\n" . Dumper( \%opts ) )
        if ( $Log->may_debug() );

    $fiji_prog = $pkg_or_obj->program( $pkg_or_obj->_assemble(%opts) );
    if ( ref $pkg_or_obj ) {
        $tag               = $pkg_or_obj->tag();
        $command_interface = $pkg_or_obj->{command_interface};
    } else {
        $tag               = $opts{tag};
        $command_interface = $opts{command_interface};
    }

    my %common_opts;
    $pkg_or_obj->_copy_common_component_params_with_ci(
        source => \%opts,
        target => \%common_opts
    );

    my $node_name = $opts{'nodescope-node-name'};
    my $nacltask_wait = delete $opts{nacltask_wait};
    my $nvram         = delete $opts{nacltask_nvram};
    my $filter        = $nvram ? 'nvfilter' : 'filter';
    my $cmd           = "fiji_" . $filter;
    my $status_key    = $nvram ? "filter_status" : "status";
    my %fiji_args     = ( filter => $fiji_prog );

    $fiji_args{'nodescope-node-name'} = $node_name;
    $fiji_args{'tag'} = $tag;

    $Log->debug(
        "Setting FIJI $filter with $tag and program => \n $fiji_prog");
    my $nodescope_apiset = $command_interface->get_7m_or_nodescope_apiset();

    my %retry_opts;
    $pkg_or_obj->_copy_retry_params(
        source => \%opts,
        target => \%retry_opts
    );

    my $text;
    nacl_method_retry(
        code => sub { $text = $nodescope_apiset->$cmd( 'connectrec-timeout' => $opts{'method-timeout'},%fiji_args )
                   ->get_parsed_output();
                },
        exceptions => 'NACL::APISet::Exceptions::UnexpectedOutputException',
        %retry_opts,
    );

    my $slot        = $text->[0]->{slot_number};
    my $fiji_status = $text->[0]->{$status_key};
    $Log->debug("fiji set on slot number $slot and $fiji_status");
    NATE::BaseException->throw("Error setting FIJI: $fiji_status")
        if ( $fiji_status ne "set" );

    if ( ref $pkg_or_obj ) {
		$pkg_or_obj->slot($slot);
    }

    my $time_out = $opts{'nacltask_wait_interval'}
        || $FIJI_COMPLETION_TIMEOUT;
    if ($nacltask_wait) {
        Tharn::snooze(WAIT_TIME);
        $pkg_or_obj->wait_for_fiji(
            %common_opts,
            program => $fiji_prog, 
            slot                     => $slot,
            'nacltask_wait_interval' => $time_out,
            'nodescope-node-name'    => $node_name,
        );
    } ## end if ($nacltask_wait)
    $Log->exit() if $may_exit;
} ## end sub set

=head2 get_all_slots

(Class method)

Returns the reference of array slots in the FIJI table that contains the FIJI,
and undef if the FIJI is not in the table.

=over

=item Options

=over

=item C<< program => $program >>

(Optional) Fiji Program whose slot needs to be find out

=item C<< 'nodescope-node-name' => string >>

(Optional) run fiji command on specified node. (Default = 'local')

=back

=item Examples

    NACL::STask::Fiji->get_all_slots(command_interface = $ci,
                                   program => $prog);

=back

=cut

sub get_all_slots {
    $Log->enter() if $may_enter;
    my $pkg = shift;
    $pkg->_verify_invocation( style => 'static_only' );
    my %opts = $pkg->_common_validate_with(
        params          => \@_,
        additional_spec => {
            tag                   => { type => SCALAR, optional => 1 },
        },
    );

    if (!(defined ($opts{tag}))) {
        $opts{tag} = NATE::Database::Test::testdb_global()
                            ->get_testcase(testcase_id => $testcase_id)->{name};
    }

    my $node_name = delete $opts{'nodescope-node-name'};
    my $expected_prog = $opts{program}." -tag ".$opts{tag};
    $expected_prog =~ s/\'//g;
    my $command_interface = $opts{command_interface};

    my $nodescope_apiset  = $command_interface->get_7m_or_nodescope_apiset();
    my @slot;
    my $slot_found = 0;

    my $text = $nodescope_apiset->
                    fiji_show('nodescope-node-name'=>$node_name)
                            ->get_parsed_output();

    foreach my $row (@$text) {
        my $actual_prog = $row->{filter};
        my $slot        = $row->{slot};
        $actual_prog =~ s/\"//g;
        $Log->trace(
            "expected program is $expected_prog and actual is $actual_prog");
        $Log->trace("actual program is $actual_prog and slot is $slot");
        if ( $expected_prog eq $actual_prog ) {

            # If we don't know exactly which slot it should be in
            # (older versions of ONTAP), but the fiji program matches,
            # return the current slot.
            $Log->trace("Returned slot is $slot ");
            push @slot, $slot;
            $slot_found = 1;

        } ## end if ( $expected_prog eq...)

    } ## end foreach my $row (@$text)
    if ($slot_found) {
        $Log->exit() if $may_exit;
        return \@slot;
    } else {
        $Log->exit() if $may_exit;
        return undef;
    }

    # not found
} ## end sub get_all_slots

=head2 active

(Instance method)

Returns true if the FIJI is active (i.e. it has not triggered yet).
Returns false if the FIJI has been cleared or has triggered.

=over

=item Examples

     $Fiji_obj->active();
     $Fiji_obj->active('nodescope-node-name' => '<nodename>');

=back

=cut

sub active {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,%args) = @_;
    if($pkg_or_obj->_get_fiji_status(%args) =~ /in progress/){
        $Log->exit() if $may_exit;
        return 1;
    }

    $Log->exit() if $may_exit;
    return 0;
} ## end sub active

=head2 remaining_attempts

(Instance Method)

This method returns the remaining attempts of the fiji.

=cut

sub remaining_attempts {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,%opts) = @_;
    my ( $command_interface, $obj, $count_final );

    $command_interface = $opts{command_interface};
    $obj = NACL::CS::Fiji->fetch(command_interface => $command_interface);
    $count_final = $obj->maxcount - $obj->count;

    $Log->comment("fiji_remaining_attempts :"
			.Dumper($count_final));
    $Log->exit() if $may_exit;
    return $count_final;
} ## end sub remaining_attempts


=head2 clear

(Class or Instance method)

Clear a FIJI.

This will clear the FIJI whether it is in NVRAM or not.

=over

=item Options

=over

=item C<< slot => $slot >>

(Required for class method) Slot where fiji needs to be cleared

=item C<< tag => string >>

(Optional)A descriptive name of the FIJI (case is not significant)

=item C<< program => string >>

(Optional)The actual fiji program, with sprintf-like format descriptors where arguments
should be substituted.Program that will be construct using tag and nacltask_param

=item C<< nacltask_nvram => BOOLEAN >>

(Optional) if nvram is 1,fiji will be clear in nvram (Default = 0)

=item C<< nacltask_fiji_clear_on_all_nodes => BOOLEAN >>

(Optional) if nacltask_fiji_clear_on_all_nodes is 1,fiji will be clear on all nodes (Default = 0)

=item C<< 'nodescope-node-name' => string >>

(Optional) run fiji clear on specified node. (Default = 'local')

=item C<< nacltask_retry_tries_count => $tries_count >>

(Optional, defaults to 5)
The number of times to retry the create command if it fails because
the file is temporarily busy.

Note that it is possible to make it not retry the create command by setting
the value of this argument to 1.

=item C<< nacltask_retry_sleep_time => $sleep_time >>

(Optional, defaults to 30 seconds)
The time to sleep between retries of the create command

=item C<< nacltask_retry_num_times_code_ran => \$variable >>

(Optional)
A reference to a scalar variable which gets filled in with the number
of times the resume command was invoked.

=back

=item Examples

    $Fiji_obj->clear();

    or

    NACL::STask::Fiji->clear(
        command_interface => $Node,
        slot              => $slot,
        tag =>
            'Cause resync to fail by making marking of volume as a duplicate fail',
        program =>
            '(idump.resync.fail.mark (eq @vol component_test_vol)(return 1))',
    );

=item Exceptions

=over

=item C<NACL::STask::Exceptions::FijiClearFailure>

This type of exception is thrown when any failure happens while doing fiji clear.

=back

=back

=cut

sub clear {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts;
    my %common_option = (
        nacltask_nvram                   => { type => BOOLEAN, default => 0 },
        nacltask_fiji_clear_on_all_nodes => { type => BOOLEAN, default => 0 },
        nacltask_retry_tries_count => { type => SCALAR,  default => 5 },
        nacltask_retry_sleep_time  => { type => SCALAR,  default => 30 },
        nacltask_retry_num_times_code_ran =>
            { type => SCALARREF, optional => 1 },
    );
    if ( ref($pkg_or_obj) ) {
        %opts = $pkg_or_obj->_common_validate_with(
            params          => \@_,
            additional_spec => \%common_option,
        );
    } else {
        %opts = $pkg_or_obj->_common_validate_with(
            params          => \@_,
            additional_spec => {
                slot    => { type => SCALAR },
                tag     => { type => SCALAR, optional => 1 },
                program => { type => SCALAR, optional => 1 },
                %common_option,
            },
        );
    } ## end else [ if ( ref($pkg_or_obj) )]

    if ( ref ($pkg_or_obj) && !(defined $pkg_or_obj->slot())) {
        NATE::BaseException->throw("caught undefined value of slot."
                ."instance Fiji->clear can not be called before Fiji->set"
        );
    }

    my ( $slot, $nvram, $nacltask_fiji_clear_on_all_nodes, $exception_obj );
    my $command_interface;
    if ( ref $pkg_or_obj ) {
        $slot              = $pkg_or_obj->{slot};
        $command_interface = $pkg_or_obj->{command_interface};
    } else {
        $slot              = $opts{slot};
        $command_interface = $opts{command_interface};
    }

    my $nodescope_apiset = $command_interface->get_7m_or_nodescope_apiset();
    $nvram = delete $opts{"nacltask_nvram"};
    $nacltask_fiji_clear_on_all_nodes
        = delete $opts{"nacltask_fiji_clear_on_all_nodes"};

    my $nodescope_node = delete $opts{"nodescope-node-name"};

    my $status_key;

    $Log->debug("Clear FIJI on $slot slot ");

    my $clear = $nvram ? "nvclear" : "clear";
    if ( $slot eq "all" ) {
        $status_key = "status";
    } else {
        $status_key = "filter_status";
    }

    my $cmd = "fiji_" . $clear;

    my %retry_opts;
    $pkg_or_obj->_copy_retry_params(
        source => \%opts,
        target => \%retry_opts
    );

    my ( $text, $fiji_status, @text_array, @fiji_status );
    if ($nacltask_fiji_clear_on_all_nodes) {
        my @nodes = NACL::STask::Cluster->node_objs(
            command_interface => $command_interface );
        $Log->debug(
            sub { "Nodes in the cluster are: \n" . Dumper( \@nodes ) } );
        foreach my $node (@nodes) {
            my $node_name = $node->name();
            $Log->debug("Clear the fiji on node $node_name");

            nacl_method_retry(
                code => sub { $text = $nodescope_apiset->$cmd(
                                  'slot'                => $slot,
                                  'nodescope-node-name' => $node_name
                              )->get_parsed_output();
                        },
                exceptions => 'NACL::APISet::Exceptions::CommandFailedException',
                %retry_opts,
            );

            $fiji_status = $text->[0]->{$status_key};

            if (   defined $nvram
                && defined $fiji_status
                && ( $fiji_status ne "cleared" ) )
            {
                if ( !defined $exception_obj ) {
                    $exception_obj
                        = NACL::STask::Exceptions::FijiClearFailure->new(
                        text => "Fiji Clear fail for nvram" );
                }
                $exception_obj->fiji_failure_set(
                    $node->name() => "Error clearing FIJI: $fiji_status" );
            } ## end if ( defined $nvram &&...)
        } ## end foreach my $node (@nodes)
    } else {
        nacl_method_retry(
            code => sub {$text
                             = $nodescope_apiset->
                             $cmd( 'slot' => $slot,'nodescope-node-name' => $nodescope_node )
                                  ->get_parsed_output();
                    },
            exceptions => 'NACL::APISet::Exceptions::CommandFailedException',
            %retry_opts,
        );
        $fiji_status = $text->[0]->{$status_key};
        if (   defined $nvram
            && defined $fiji_status
            && ( $fiji_status ne "cleared" ) )
        {    # nvclear doesn't print anything at all
            if ( !defined $exception_obj ) {
                $exception_obj
                    = NACL::STask::Exceptions::FijiClearFailure->new(
                    text => "Fiji Clear fail for nvram" );
            }
            $exception_obj->fiji_failure_set( $command_interface->name() =>
                    "Error clearing FIJI: $fiji_status" );
        } ## end if ( defined $nvram &&...)
    } ## end else [ if ($nacltask_fiji_clear_on_all_nodes)]

    $exception_obj->throw() if ( defined $exception_obj );
    $Log->exit() if $may_exit;
} ## end sub clear

=head2 wait_for_fiji

(Class or Instance Method)

This methods waits for trigger of fiji on the particular node.

=over

=item Options

=over

=item C<< slot =>$slot >>

(Required for Class method) Slot name.

=item C<<nacltask_fiji_operation => $ops>>

(optional) expected operation on fiji.
Default - set

=item C<<nacltask_wait_interval => $timeout>>

(Required) Maximum timeout limit.

=item C<<nacltask_polling_interval => $poll_time>>

(Optional) This is the interval (in seconds) after which the value of
the attributes will be polled. If not specified it defaults to 10 seconds.

=item C<< 'nodescope-node-name' => string >>

(Optional) run fiji command on specified node. (Default = 'local')

=item command_interface, nacltask_wait_interval etc.

=back

=item Examples

    $self->wait_for_fiji(
        command_interface => $command_interface,
        slot              => $slot,
        "nacltask_wait_interval"  => $FIJI_COMPLETION_TIMEOUT
    );

=item Exceptions

=over

=item C<NACL::Exceptions::Timeout>

This type of exception is thrown when fiji has not get triggered with in $timeout seconds or
When Fiji has not hit on specified slot and maximum time limit is crossed.

=back

=back

=cut

sub wait_for_fiji {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = $pkg_or_obj->_common_validate_with(
        params          => \@_,
        additional_spec => {
            slot => { type => SCALAR, optional => 1},
            nacltask_fiji_operation =>
                { regex => qr/^hit|set$/, default => 'set' },
            nacltask_wait_interval    => { type => SCALAR },
            nacltask_polling_interval => { type => SCALAR, default => 10 },
        },
    );

    $opts{slot} //= $pkg_or_obj->slot();
    my $fiji_ops = delete $opts{nacltask_fiji_operation};

    if ( $fiji_ops eq 'hit' ) {
        $pkg_or_obj->_wait_for_fiji_hit(%opts);
    } else {
        $pkg_or_obj->_wait_for_fiji_set(%opts);
    }
    $Log->exit() if $may_exit;
} ## end sub wait_for_fiji

sub _wait_for_fiji_set {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, %args ) = @_;

    my $method_timeout    = delete $args{"nacltask_wait_interval"};
    my $end_time          = time() + $method_timeout;
    my $polling_interval  = delete $args{'nacltask_polling_interval'};

    ## for set we need to check if the fiji has been hit or in progress
    while (  $pkg_or_obj->_get_fiji_status(%args) !~ /in progress|fiji hit/ )  {
       if ( time() > $end_time ) {
         $Log->exit() if $may_exit;
         NACL::Exceptions::Timeout->throw(
		" Fiji has not triggered and maximum time limit is crossed\n"
         );
       } ## end if ( time() > $end_time)
       Tharn::snooze($polling_interval);
    } ## end while ( ( $pkg_or_obj->_get_fiji_status...))

    $Log->exit() if $may_exit;
} ## end sub _wait_for_fiji_set

sub _get_fiji_status{
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, %opts ) = @_;

    my $node_name         = $opts{'nodescope-node-name'} // $pkg_or_obj->nodescope_node_name();

    my $command_interface = $opts{command_interface} // $pkg_or_obj->command_interface();
    my $fiji_slot         = $opts{slot} // $pkg_or_obj->{slot};
    my $program           = $pkg_or_obj->{program};
    my $nodescope_apiset  = $command_interface->get_7m_or_nodescope_apiset();
    my $fiji_hit = 0;

    my $parsed_output     = $nodescope_apiset->fiji_dump('nodescope-node-name'=>$node_name,'slot' => $fiji_slot)->get_parsed_output();

    my @fiji_arr  = ();
    my $slot_int;

    ## using map to filter out the parsed output
    ## here $fiji_slot would be numeric & $_->{slot_number}
    ## would be alphanumeric . so extra check is there for
    ## comparison.
    ## for eg. $fiji_slot is 0
    ## and $_->{slot_number} is '0 (nvram)'

     map{
        if(exists $_->{status} && $program =~ /$_->{event}/  ){
	   ($slot_int) = $_->{slot_number} =~ /(\d+)/ ;
           push @fiji_arr , $_ if($slot_int eq $fiji_slot);
	}
     }@{$parsed_output};

    $Log->exit() if $may_exit;
    return $fiji_arr[0]->{status} if(@fiji_arr);
}

sub _wait_for_fiji_hit {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, %args ) = @_;

    my $method_timeout    = delete $args{"nacltask_wait_interval"};
    my $end_time          = time() + $method_timeout;
    my $polling_interval  = delete $args{'nacltask_polling_interval'};

    while ( $pkg_or_obj->_get_fiji_status(%args) !~ /fiji hit/) {
	if ( time() > $end_time ) {
	  $Log->exit() if $may_exit;
	  NACL::Exceptions::Timeout->throw(
		" Fiji has not hit on specified slot and maximum time limit is crossed\n"
	  );
	} ## end if ( time() > $end_time)
	Tharn::snooze($polling_interval);
    } ## end while ($pkg_or_obj->_get_fiji_status..)

    $Log->exit() if $may_exit;
} ## end sub _wait_for_fiji_hit

#############################################
# Method : _assemble
# Description :  Assemble the fiji program when params
#                         are provided in fiji->set()
##############################################

sub _assemble {
    my $self = shift;
    my %opts = @_;

    my $ftemp     = $opts{nacltask_param};
    my $fprogram  = $opts{program};
    my @arguments = ();

    $Log->debug( "Fiji::_assemble: ftemp:" . Dumper($ftemp) )
        if ( $Log->may_debug() );
    if ( !$ftemp ) {

        # this point takes no params, so nothing to process
    } elsif ( !ref $ftemp ) {
        push @arguments, $opts{$ftemp};
    } elsif ( ref $ftemp eq 'ARRAY' ) {
        my $params = $ftemp;
        foreach my $ptemp ( @{$params} ) {
            if ( !ref $ptemp ) {
                push @arguments, $ptemp;
            } elsif ( ref $ptemp eq 'HASH' ) {

                # parameter, optional with default
                NATE::BaseException->throw(
                    "Too many parameters in argument hash for FIJI")
                    if ( scalar %{$ptemp} > 1 );

                # single parameter, optional with default
                my ( $key, $value ) = %{$ptemp};
                push @arguments, $value;
            } else {
                NATE::BaseException->throw("Bad type for $ptemp");
            }
        } ## end foreach my $ptemp ( @{$params...})
    } else {
        NATE::BaseException->throw( "Bad type for " . $ftemp );
    }
    $Log->debug( "Program arguments are :" . Dumper(@arguments) )
        if ( $Log->may_debug() );
    my $fiji_string = sprintf( $fprogram, @arguments );
    $Log->debug("fiji_string is : \n $fiji_string");
    NATE::BaseException->throw("Can't assemble fiji string")
        unless $fiji_string;
    return $fiji_string;
} ## end sub _assemble

# This method should return the validation spec for the primary keys
# Note that here, we do not include hostrec, command_interface in this
# validation spec
sub _primary_keys_validate_spec {
    return (
        tag     => { type => SCALAR, default => $default_tag },
        'nodescope-node-name' => { type => SCALAR, default => "local" },
        program => { type => SCALAR },
    );
} ## end sub _primary_keys_validate_spec

1;
