#
# Copyright (c) 2013 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Ipspace Task Module
## @author dl-nacl-dev@netapp.com
## @status shared
## @pod here

package NACL::STask::NetworkPortBroadcastDomain;

use strict;
use warnings;
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use NACL::C::Node;
use base qw(NACL::C::NetworkPortBroadcastDomain NACL::STask::STask);
use NACL::STask::NetworkInterface;
use NACL::C::NetworkPort;

# import Exceptions
use NATE::BaseException qw(:try);
use NACL::APISet::Exceptions::CommandFailedException;
use NACL::C::Exceptions::NetworkPortBroadcastDomain::AlreadyExists;
use NACL::C::Exceptions::NetworkPortBroadcastDomain::PortsAssignedHasLifAssociation;
use Params::Validate qw(validate SCALAR ARRAYREF UNDEF SCALARREF);

=head1 NAME

NACL::STask::NetworkPortBroadcastDomain

=head1 DESCRIPTION

C<NACL::STask::NetworkPortBroadcastDomain> provides a number of well-defined but
potentially complex or multi-step methods related to Network Port broadcast-domain in ONTAP.
It builds on top of C<NACL::C::NetworkPortBroadcastDomain> component. Since it is
derived class of C<NACL::C::NetworkPortBroadcastDomain>, we can use all the methods of
C<NACL::C::NetworkPortBroadcastDomain> from the object of this task.

=head1 CLEANUP METHODS

Cleanup can be registered for the following method, 'modify'.

Cleanup methods are,

    NetworkPortBroadcastDomain Method         Cleanup Method
    -----------------------------------------------
    modify                                    modify

=head1 ATTRIBUTES

=head2 command_interface

(Required) A component object that represents the host to which to send commands.
See L<NACL::C::Component::command_interface|lib-NACL-C-Component-pm/command_interface>.

=head2 ipspace
(Required) Name of the ipspace.

=head2 broadcast-domain
(Required) Name of the broadcast domain.

=head1 METHODS

=head2 add_ports

L<< NACL::STask::STask->create_method_builder(c_method_name => 'add_ports')|lib-NACL-STask-STask-pm/create_method_builder >>

=head2 remove_ports

L<< NACL::STask::STask->purge_method_builder(c_method_name => 'remove_ports')|lib-NACL-STask-STask-pm/purge_method_builder >>

=cut

__PACKAGE__->create_method_builder(c_method_name => 'add_ports', stask_method_name => 'add_ports', c_exception_name => 'PortAlreadyAssigned');
__PACKAGE__->purge_method_builder(c_method_name => 'remove_ports', stask_method_name => 'remove_ports', c_exception_name => 'PortAlreadyRemoved');

=head2 create

    my $NetworkPortBroadcastDomain_obj    = NACL::STask::NetworkPortBroadcastDomain->create(
                                command_interface => $ci,
                                'ipspace' => $ipspace,
                                'broadcast-domain' => $broadcast_domain,
                                ...);

(Class method)
This method is used to create a new layer 2 broadcast domain.
Uses CMode CLI/ZAPI .
STask adds functionality of _was_created and nacl_task_if_exists.

=over

=item Options

=over

=item C<< nacltask_if_exists=>$action >>

(Optional) What to do if the broadcast domain to be created already exists.
If $action is "die" (the default), then fail with an exception (in the
same way that the broadcast domain component would have: by trying the
broadcast domain create and letting the product complain about the broadcast domain
already existing).  If action is "reuse", then do nothing (return a
task object referring to the existing broadcast domain).  If action is
"purge", then delete the broadcast domain (see the C<purge> method, below)
before creating a new one.

=item C<< _was_created => \$scalar >>

(Optional)
When this option is provided a reference to a scalar variable, the variable
gets filled in with a boolean value describing whether the broadcast domain
was found (value will be 0; this scenario is possible when
nacltask_if_exists => "reuse") or whether the broadcast domain
was created (value will be 1).
This is necessary to determine whether the broadcast domain needs to be
cleaned up later.

    my $was_created;
    my $broadcast_domain_obj = NACL::STask::NetworkPortBroadcastDomain->create(
        nacltask_if_exists    => 'reuse',
        _was_created => \$was_created,
        %other_opts
    );
    # Operate on $broadcast_domain_obj here
    # ...
    # Now determine whether to clean up the broadcast domain, since we're not sure
    # whether we reused an existing broadcast domain or created a new one
    if ($was_created) {
        # New broadcast domain was created. Clean it up.
        $broadcast_domain_obj->purge();
    }

=item Component options

Same options as those accepted by
L<< NACL::C::NetworkPortBroadcastDomain->create|lib-NACL-C-NetworkPortBroadcastDomain-pm/create >>

=back

=back

=over

=item Exceptions

=over

=item C<NACL::C::Exceptions::NetworkPortBroadcastDomain::AlreadyExists>

This type of exception is thrown when an attempt is made to create a Broadcast Domain that already exists.

=back

=back

=cut

sub create {
    $Log->enter();
    my ($pkg, @opts) = @_;
    my $self;
    my %opts = $pkg->_common_validate_with(
        params          => \@opts,
        additional_spec => {
            nacltask_if_exists => {
                type      => SCALAR,
                callbacks => $pkg->_if_exists_validate_callback(),
                optional  => 1
            },
            _was_created         => { type => SCALARREF, optional => 1 }
        },
        allow_extra => 1
    );
    my $if_exists           = delete $opts{nacltask_if_exists};
    my $was_created         = delete $opts{_was_created};
    $$was_created = 0;
  CREATE: {
        use warnings;
        try {
            # Calling network port broadcast domain create
            $self = $pkg->SUPER::create(%opts);
            $$was_created = 1;
        }
        catch NACL::C::Exceptions::NetworkPortBroadcastDomain::AlreadyExists with {
            my $exception = shift;
            # covers CMode CLI and ZAPI
            $Log->trace('The broadcast domain already exists');
            $self = $pkg->_element_exists_handler(
                create_opts        => \%opts,
                nacltask_if_exists => $if_exists,
                exception          => $exception
            );
            if ( !$self ) {
                no warnings qw(exiting);
                redo CREATE;
            }
        };
    } ## end CREATE:

    $Log->exit();
    return $self;
} ## end sub create

=head2 purge

    $bd_obj->purge();

    or

    NACL::STask::NetworkPortBroadcastDomain->purge(
        'command_interface'     => $command_interface,
        'ipspace'               => $ipspace,
        'broadcast-domain;      => $bd,
    );


(Class or instance method) This method deletes broadcast domain.
It will purge lifs associated with the ports that belong to it, Ports will automatically
get assigned to Default ipspace.

=over

=item Options

=over

=item C<< _was_deleted => \$scalar >>

(Optional)
When this option is provided a reference to a scalar variable, the variable
gets filled in with a boolean value describing whether the broadcast domain
did not exist (value will be 0; this scenario is possible when
nacltask_if_purged => "pass") or whether the broadcast domain was deleted
(value will be 1).

=item C<< nacltask_if_purged=>die|pass >>

(Optional)
If "die", exception is thrown when broadcast domain to be deleted does not exist
If "pass", exits quietly when broadcast domain to be deleted does not exist

=item Component options

Same options as those accepted by 
L<< NACL::C::NetworkPortBroadcastDomain->delete|lib-NACL-C-NetworkPortBroadcastDomain-pm/delete >>

=back

=back

=over

=item Exceptions

=over

=item C<NACL::C::Exceptions::NetworkPortBroadcastDomain::DoesNotExist>

This type of exception is thrown when an attempt is made to delete Broadcast Domain that does not exists.

=back

=back

=cut

sub purge {
    $Log->enter();
    my ($pkg_or_obj, @opts) = @_;
    my %opts       = $pkg_or_obj->_common_validate_with(
        params          => \@opts,
        additional_spec => {
            nacltask_if_purged => {type => SCALAR, default => 'pass'},
            _was_deleted       => { type => SCALARREF, optional => 1 }
        },
        allow_extra     => 1
    );
    my $was_deleted = delete $opts{_was_deleted};
    $$was_deleted = 0;

    my %common_opts;
    $pkg_or_obj->_copy_common_component_params_with_ci(
        source => \%opts,
        target => \%common_opts
    );

    my $if_purged = delete $opts{nacltask_if_purged};

  DELETE: {
        try {
            $pkg_or_obj->delete(%opts);
            $$was_deleted = 1;
        } catch NACL::C::Exceptions::NetworkPortBroadcastDomain::DoesNotExist with {
            my $exception = shift;
            if ( $if_purged !~ /pass/i ) {
                $Log->exit();
                $exception->throw();
            }
        } catch NACL::C::Exceptions::NetworkPortBroadcastDomain::PortsAssignedHasLifAssociation with {
            # If any port assigned to a broadcast domain also has lif associated with it,
            # then delete lif associated with each of the port. No need to reassign the port
            # to other ipspace/brocast-domain because all ports will automatically get
            # assigned to ipspace Default.
            $pkg_or_obj->_delete_all_lifs(
                %common_opts,
                ipspace            => $opts{ipspace},
                'broadcast-domain' => $opts{'broadcast-domain'},
            );
            goto DELETE;
        };
    }

    $Log->exit();
}


=head2 modify

    NACL::STask::NetworkPortBroadcastDomain->modify(
       'command_interface'     => $command_interface,
        'ipspace'               => $ipspace,
        'broadcast-domain;      => $bd,
        'nacltask_to_cleanup'   => 1,          # default 0,
        'nacltask_cleanup_manager' => $CleanupObj,
        'nacltask_verify'         => 1,        # default 0,
        %other_opts,
    );
                        or

    $bd_obj->modify(%other_opts);

(Class or instance method)
This method modifies the properties of the network port (mtu)
through NetworkPortBroadcastDomain.

Uses CMode CLI/ZAPI .

=over

=item Options

=over

=item C<< nacltask_verify => 0|1 >>

(Optional, defaults to 0)
Specifies whether to verify that all the fields got modified to the values
specified in the call. If one or more of the fields don't match, then a
L<NACL::Exceptions::VerifyFailure|lib-NACL-Exceptions-VerifyFailure-pm>
exception is thrown.

=item C<< nacltask_to_cleanup => 0|1 >>

(Optional, default to 0(no to cleanup)
Flag indicating if this operation needs to be registered for clean up or not.

=item C<< nacltask_cleanup_manager >>

Cleanup manager to be used for registering.
Default : Will use the default cleanup manager.

=item Other options

All of the other options accepted by
L<NACL::C::NetworkPortBroadcastDomain::modify|lib-NACL-C-NetworkPortBroadcastDomain-pm/modify>.

=back

=back

=over

=item Exceptions

=over

=item C<NACL::Exceptions::VerifyFailure>

This type of exception is thrown when verification for NetworkPortBroadcastDomain modify has failed.

=back

=back

=cut

sub modify {
    $Log->enter();
    my $pkg_or_obj = shift;
    my %opts       = $pkg_or_obj->_common_validate_with(
        params          => \@_,
        additional_spec => {
            nacltask_verify => { type => SCALAR, default => 0 },
            $pkg_or_obj->_cleanup_validate_spec(),
        },
        allow_extra => 1
    );

    my $verify = delete $opts{nacltask_verify};
    my ( %opts_for_cleanup, $opts_for_register, $register_for_cleanup );
    $pkg_or_obj->_move_common_cleanup_opts(
        source              => \%opts,
        target              => \%opts_for_cleanup,
        nacltask_to_cleanup => \$register_for_cleanup,
    );

    if ($register_for_cleanup) {
        $opts_for_register =
          $pkg_or_obj->_common_modify_for_cleanup( %opts_for_cleanup, %opts );
    }

    $pkg_or_obj->SUPER::modify(%opts);

    #Register for cleanup if $nacltask_to_cleanup is set
    $pkg_or_obj->_register_for_cleanup( %{$opts_for_register} )
      if ($register_for_cleanup);

    # Verify the modifications is done or not
    if ($verify) {
        $pkg_or_obj->taskverify_modify(%opts);
    }

    $Log->exit();
} ## end sub modify

sub _delete_all_lifs {
    $Log->enter();
    my ( $pkg, %opts ) = @_;
    
    my $ipspace = delete $opts{ipspace};
    my $bd = delete $opts{'broadcast-domain'};
    my @port_states = NACL::CS::NetworkPort->fetch(
        %opts,
        filter => {
            'broadcast-domain' => $bd,
            ipspace            => $ipspace,
        },
        requested_fields => [ qw(node) ],
    );

    my @lifs_to_be_deleted;
    foreach my $state (@port_states) {
        my @lifs = NACL::STask::NetworkInterface->find(
            %opts,
            filter => {
                'curr-port' => $state->port(),
                'curr-node' => $state->node(),
            }
        );
        push(@lifs_to_be_deleted, @lifs);
    }

    foreach my $lif (@lifs_to_be_deleted) {
        $lif->purge(%opts);
    }

    $Log->exit();
}

# Overwrite the element_exists_handler for add_ports help.
sub _element_exists_handler {
    my ($self,%args) = @_;
    my $exception = $args{exception};
    # If the Port is already assigned then we need to see if it was assigned to the right place.
    if(ref($exception) eq 'NACL::C::Exceptions::NetworkPortBroadcastDomain::PortAlreadyAssigned' and $args{'nacltask_if_exists'} eq 'reuse') {
        $exception->text() =~ /is already assigned to broadcast domain "(\S+)" in IPspace "(\S+)"/i;
        my $assigned_bd = $1;
        my $assigned_ipspace = $2;
        # If it was assigned to the correct place then we do normal process.
        if ($assigned_bd eq $args{create_opts}->{'broadcast-domain'} and 
            $assigned_ipspace eq $args{create_opts}->{ipspace}) {
            return $self->SUPER::_element_exists_handler(%args);      
        # If it was not assigned to the correct place then fail
        } else {
            $exception->throw();
        }
    } else {
        return $self->SUPER::_element_exists_handler(%args);    
    } 
}

1;
