#
#   Copyright (c) 2013-2017 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary NetworkRoute Task Module
## @author dl-nacl-dev@netapp.com
## @status shared
## @pod here

package NACL::STask::NetworkRoute;

use base qw(NACL::C::NetworkRoute NACL::STask::STask);
use strict;
use warnings;
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Net::IP qw(:PROC);
use NACL::APISet::Response::CLI::Freebsd;
use NACL::APISet::Host::CLI::Freebsd::default::netstat;
use NACL::C::NetworkOptionsIpv6;
use NACL::C::Vserver;
use NATE::Exceptions::Argument qw(:try);
use NACL::Exceptions::VerifyFailure;
use Params::Validate qw(:all);
use Class::MethodMaker [ scalar => '_was_route_created' ];
use NACL::CS::NetworkInterface;
use NACL::Global;

=head1 NAME

NACL::STask::NetworkRoute

=head1 DESCRIPTION

C<NACL::STask::NetworkRoute> provides a number of well-defined but
potentially complex or multi-step methods related to network routes in
ONTAP.

It builds on top of, and is a derived class of C<NACL::C::NetworkRoute>,
and so it also provides methods that are more in the scope of individual Network Route-related commands.
See C<NACL::C::NetworkRoute> for details.  This also means that a C<NACL::STask::NetworkRoute>
object may generally be used in place of a component object.

=head1 ATTRIBUTES

=head2 command_interface

(Required) A component object that represents the host to which to send commands.

=head2 vserver

(Required) The name of the vserver for which network route has to be created.

=head2 gateway

(Required) Gateway details to create network route.

=head2 destination

(Required) The destination represented as IpAddress/Mask.

=head1 METHODS

=head2 create

  my $route  = NACL::STask::NetworkRoute->create(
      command_interface   => $command_interface,
      vserver             => $vserver,
      destination         => $destination,
      gateway             => $gateway,
      nacltask_if_exists  => $action,      # default 'die'
      %other_options
  );

(Class method) Creates a network route.  This method provides additional
services beyond what's inherent to the product's network route creation
commands, as controlled and described by the new C<nacltask_if_exists> options.

=over

=item Options

=over

=item C<< nacltask_if_exists=>$action >>

(Optional) What to do if the route to be created already exists.
If $action is "die" (the default), then fail with an exception (in the
same way that the route component would have: by trying the
route create and letting the product complain about the route
already existing).  If action is "reuse", then do nothing (return a
task object referring to the existing route).  If action is
"purge", then delete the route (see the C<purge> method, below)
before creating a new one.

=item C<< nacltask_verify=>$nacltask_verify_boolean >>

(Optional)
 If '0' (default), verification will not be performed.
 If '1', verification will be performed to ensure that network route addition
 did happen successfully(check the addition of route in ngsh
 & nodescope & BSD).

 =item C<< nacltask_verify_only_ngsh=>$nacltask_verify_only_ngsh >>

(Optional)
 If '0' (default), verification will not be performed.
 If '1', verification will be performed to ensure that the creation
 did happen successfully in ngsh.

=item C<< nacltask_verify_only_nodescope=>$nacltask_verify_only_nodescope >>

(Optional)
 If '0' (default), verification will not be performed.
 If '1', verification will be performed to ensure that network route addition
 did happen successfully in nodescope.

 =item C<< nacltask_verify_only_bsd=>$nacltask_verify_only_bsd >>

(Optional)
 If '0' (default), verification will not be performed.
 If '1', verification will be performed to ensure that network route addition
 did happen successfully in bsd.

 =item C<< _was_created => \$scalar >>

(Optional)
When this option is provided a reference to a scalar variable, the variable
gets filled in with a boolean value describing whether the route
was found (value will be 0; this scenario is possible when
nacltask_if_exists => "reuse") or whether the network route was created (value will
 be 1).
This is necessary to determine whether the network route needs to be
cleaned up later.

    my $was_created;
    my $route_obj = NACL::STask::NetworkRoute->create(
        nacltask_if_exists    => 'reuse',
        _was_created => \$was_created,
        %other_opts
    );
    # Operate on $route_obj here
    # ...
    # Now determine whether to clean up the route
    if ($was_created) {
        # New route was created. Clean it up.
        $route_obj->purge();
    }

=item C<< nacltask_to_cleanup => 0|1 >>

(Optional, defaults to 0(no to cleanup))
Flag indicating if this operation is to be cleaned up or not.

=item C<< nacltask_cleanup_manager >>

Cleanup manager to be used for registering.
Default : Will use the default cleanup manager.

=item C<MCC options>

The options accepted for MCC configuration replication verification is
documented at L<MCC options|lib-NACL-STask-STask-pm/mcc_options>.
In case of IPV6 route it is required to manually enable IPV6 option
MCC on the remote MCC site if its not already enabled.
To enable IPV6 use :
NACL::C::NetworkOptionsIpv6->modify(command_interface => $ci,
enabled => 'true')).
If ipv6 option is not enabled on remote mcc the replication will fail.

=back

=item command_interface, apiset_must, apiset_should, method-timeout, routing-group, destination, gateway, etc.

All of the other various options to
L<< NACL::C::NetworkRoute->create | lib-NACL-C-NetworkRoute-pm/create >>

=back

=over

=item Exceptions

=over

=item C<NACL::Exceptions::VerifyFailure>

This type of exception is thrown when there is a failure in verifying network route creation
in ngsh, sk and bsd.

=back

=cut

sub create {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            nacltask_if_exists        => $pkg->_if_exists_validate_spec(),
            nacltask_verify           => {type => SCALAR, default => 0},
            nacltask_verify_only_ngsh => {type => SCALAR, default => 0},
            nacltask_verify_only_nodescope => {type => SCALAR, default => 0},
            nacltask_verify_only_bsd => {type => SCALAR,    default  => 0},
            _was_created             => {type => SCALARREF, optional => 1},
            polling_interval         => {type => SCALAR,    default  => 20},
            $pkg->_cleanup_validate_spec(),
	    $pkg->_mcc_validate_spec(),
        },
        allow_extra => 1
    );

    my $if_exists         = delete $opts{nacltask_if_exists};
    my $was_created       = delete $opts{_was_created};
    my $verify            = delete $opts{nacltask_verify};
    my $verify_ngsh       = delete $opts{nacltask_verify_only_ngsh};
    my $verify_bsd        = delete $opts{nacltask_verify_only_bsd};
    my $verify_nodescope  = delete $opts{nacltask_verify_only_nodescope};
    my $polling_interval  = delete $opts{polling_interval};
    my $command_interface = $opts{command_interface};

    my %mcc_opts;
    $pkg->_move_common_mcc_opts(
        source => \%opts,
        target => \%mcc_opts,
    );

    $$was_created = 0;
    my $self;
    my (%common_args, %opts_for_cleanup, %opts_for_register, $nacltask_to_cleanup);
    $pkg->_copy_common_component_params_with_ci(
        source => \%opts,
        target => \%common_args,
    );

    $pkg->_move_common_cleanup_opts(
        source => \%opts,
        target => \%opts_for_cleanup,
    );
CREATE: {
        use warnings qw(exiting);
        try {
            if (Net::IP::ip_is_ipv6($opts{'address'})) {
                NACL::C::NetworkOptionsIpv6->modify(
                    command_interface => $command_interface,
                    enabled           => "true",
                );
            }
            $self = $pkg->SUPER::create(%opts);
            Tharn::snooze($polling_interval);
            $$was_created = 1;
            $self->_was_route_created(1);
        }
        catch NACL::C::Exceptions::NetworkRoute::AlreadyExists with {
            my $exception = $_[0];
            $Log->trace('The network route already exists');
            $self = $pkg->_element_exists_handler(
                create_opts        => \%opts,
                nacltask_if_exists => $if_exists,
                exception          => $exception
            );
            if (!$self) {
                no warnings qw(exiting);
                redo CREATE;
            }
        };
    }

    ## check for replication in mcc
    $self->_check_in_metrocluster_node(
        command_interface => $opts{command_interface},
        method_used       => "create",
        %mcc_opts,
    );

    if ($$was_created) {
        my $vsid;

        if ($verify || $verify_nodescope || $verify_bsd) {
            $vsid = NACL::C::Vserver->get_vsid(
                command_interface => $self->command_interface(),
                vserver           => $self->vserver()
            );
        }

        #Verify that the route is created in ngsh
        if (($verify) || ($verify_ngsh)) {
            $self->taskverify_create(%opts);
        }

        #Skip validation for HN above release, refer burt#1062014
        if( $command_interface->version_manager()->is_version_lt(
           release => $NACL::Global::ONTAP_Release_Harpoon_0)){

            #Verify that the route is created in nodescope
            if (($verify) || ($verify_nodescope)) {
                $self->_taskverify_nodescope(
                    command_interface => $command_interface,
                    destination       => $opts{destination},
                    action            => 'delete',
                    gateway           => $opts{gateway},
                    vsid              => $vsid,
                    vserver           => $opts{vserver},
                    action            => 'create'
                );
            }
            #Verify that the route is created in systemshell
            if (($verify) || ($verify_bsd)) {
                # Checking the route existence on the BSD
                $self->_taskverify_systemshell(
                    command_interface => $command_interface,
                    destination       => $opts{destination},
                    gateway           => $opts{gateway},
                    action            => 'create',
                    vsid              => $vsid,
                );
            }
        }
        # Register this resource with the Cleanup Manager for cleanup
        $pkg->_copy_common_opts_for_cleanup(
            'source'              => {%opts, %opts_for_cleanup},
            'target'              => \%opts_for_register,
            'nacltask_to_cleanup' => \$nacltask_to_cleanup,
            'to_cleanup'          => 'purge'
        );
        $pkg->_register_for_cleanup(%opts_for_register)
            if ($nacltask_to_cleanup && $$was_created);
    } ## end if ($$was_created)
    $Log->exit() if $may_exit;
    return $self;
} ## end sub create

=back

=head2 purge

    $route->purge(%other_options);

    NACL::STask::NetworkRoute->purge(
      command_interface => $command_interface,
       gateway          => $gateway,
       vserver          => $Vserver,
       destination      => $destination,
      %other_options
    );

(Class or instance method) This method deletes a network route.

=over

=item Options

=over

=item C<< vsid=>$vsid >>

(Optional)
 This represents the vserver id for which route has been created.
 In nodescope , routing group name starts with this vsid followed by
 routing group name.
 In systemshell, vsid and netif values will be same.

 If not provided, this will be determined by running "debug smdb table
 vserver_id_to_name show"

=item C<< nacltask_verify=>$nacltask_verify_boolean >>

(Optional)
 If '0' (default), verification will not be performed.
 If '1', verification will be performed to ensure that
 the deletion did happen successfully(check the deletion of route in ngsh
  & nodescope & BSD).

=item C<< nacltask_verify_only_ngsh=>$nacltask_verify_only_ngsh >>

(Optional)
 If '0' (default), verification will not be performed.
 If '1', verification will be performed to ensure that the deletion
 did happen successfully in ngsh.

 =item C<< nacltask_verify_only_nodescope=>$nacltask_verify_only_nodescope >>

(Optional)
 If '0' (default), verification will not be performed.
 If '1', verification will be performed to ensure that the deletion
 did happen successfully in nodescope.

=item C<< nacltask_verify_only_bsd=>$nacltask_verify_only_bsd >>

(Optional)
 If '0' (default), verification will not be performed.
 If '1', verification will be performed to ensure that the deletion
 did happen successfully in bsd.

 =item C<< nacltask_if_purged => die | pass >>

(Optional, defaults to "die")
Specifies what to do if the route we're trying to purge no longer
exists. A value of C<die> (the default) will result in an exception
being thrown. A value of C<pass> will mean that the exception that would
normally occur when we're trying to delete a route that does not exist
is suppressed.

=item C<< "method-timeout" => $timeout >>

(Optional) As component method-timeout, controls how long the command
will wait before completing.  However, the default timeout has been
raised to 120 seconds.

=back

=item command_interface, apiset_must, apiset_should, method-timeout, routing-group, destination, gateway, etc.

All of the other various options to
L<< NACL::C::NetworkRoute->delete | lib-NACL-C-NetworkRoute-pm/delete >>

=back

=over

=item Exceptions

=over

=item C<NACL::C::Exceptions::NetworkRoute::DoesNotExist>

This type of exception is thrown when an attempt is made to delete an route which does not exists.

=item C<NACL::Exceptions::VerifyFailure>

This type of exception is thrown when there is a failure in verifying network route deletion
in ngsh, sk and bsd.

=back

=cut

sub purge {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my $pkg = ref($pkg_or_obj) || $pkg_or_obj;
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            'method-timeout'          => {type => SCALAR, default => 120},
            nacltask_verify           => {type => SCALAR, default => 0},
            nacltask_verify_only_ngsh => {type => SCALAR, default => 0},
            nacltask_verify_only_nodescope => {type => SCALAR, default => 0},
            nacltask_verify_only_bsd       => {type => SCALAR, default => 0},
            nacltask_if_purged =>
            $pkg_or_obj->_if_action_completed_already_validate_spec(),
	    $pkg_or_obj->_mcc_validate_spec(),
        },
        allow_extra => 1
    );
    my $command_interface = $opts{command_interface};
    my $if_purged         = delete $opts{nacltask_if_purged};
    my $verify            = delete $opts{nacltask_verify};
    my $verify_ngsh       = delete $opts{nacltask_verify_only_ngsh};
    my $verify_bsd        = delete $opts{nacltask_verify_only_bsd};
    my $verify_nodescope  = delete $opts{nacltask_verify_only_nodescope};
    my $already_purged    = 0;
    my $vsid              = delete $opts{vsid};

    my %mcc_opts;
    $pkg_or_obj->_move_common_mcc_opts(
        source => \%opts,
        target => \%mcc_opts,
    );

    my %primary_keys;
    $pkg_or_obj->_copy_primary_keys(
        source => \%opts,
        target => \%primary_keys
    );

    my %common_opts;
    $pkg_or_obj->_copy_common_component_params_with_ci(
        source => \%opts,
        target => \%common_opts
    );

    # Call the component delete. If route doesn't exist
    # then it would throw an exeception which is handled below.
    try {
        $pkg_or_obj->delete(%opts);
    } ## end try
    catch NACL::C::Exceptions::NetworkRoute::DoesNotExist with {
        my $e = $_[0];
        if ($if_purged =~ m[^die$]i) {
            $Log->exit() if $may_exit;
            $e->throw();
        } else {
            $already_purged = 1;
        }
    };
    return if $already_purged;
    if (($verify) || ($verify_ngsh)) {
        $pkg_or_obj->_generic_purge_verify(%opts);
    }

    if (!defined $vsid && ($verify || $verify_nodescope || $verify_bsd)) {
        $vsid = NACL::C::Vserver->get_vsid(
            command_interface => $command_interface,
            vserver           => $opts{vserver}
        );
    }

    #Skip validation for HN above release, refer burt#1062014
    if( $command_interface->version_manager()->is_version_lt(
	release => $NACL::Global::ONTAP_Release_Harpoon_0)){

	if (($verify) || ($verify_nodescope)) {
            # Checking the route deletion on the nodescope
            $pkg_or_obj->_taskverify_nodescope(
		command_interface => $command_interface,
		gateway           => $opts{gateway},
		destination       => $opts{destination},
		action            => 'delete',
		vsid              => $vsid,
		vserver           => $opts{vserver},
	    );
	} ## end if (($verify) || ($verify_nodescope...
    }
    if (($verify) || ($verify_bsd)) {
        # Checking the route deletion on the BSD
        $pkg_or_obj->_taskverify_systemshell(
            command_interface => $command_interface,
            destination       => $opts{destination},
            action            => 'delete',
            vsid              => $vsid,
        );
    }

    ## check for replication in mcc
    $pkg_or_obj->_check_in_metrocluster_node(
        command_interface => $opts{'command_interface'},
        method_used       => "purge",
        %mcc_opts,
        %primary_keys,
    );

    $Log->exit() if $may_exit;
} ## end sub purge

=back

=head2 _taskverify_systemshell

   $task_obj->_taskverify_systemshell(
           command_interface => $node,
           destination       => $destination,
           action            => $action,
    );

   or

   NACL::STask::NetworkRoute->_taskverify_systemshell(
           command_interface => $node,
           destination       => $destination,
           action            => $action,
   );

(Class or Instance method)
This function takes destination & action as input and checks
the exact match of address from the apiset (route -sng)
to the destination on the systemshell & according to the
action throws the error message.

=over

=item Options

=over

=item C<< action => $action >>

(Required) The action to be done.
The value for action can be create or delete.

=back

=back

=cut

sub _taskverify_systemshell {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my $pkg = ref($pkg_or_obj) || $pkg_or_obj;
    my %opts = validate_with(
        params => \@args,
        spec   => {
            'command_interface' => {type => OBJECT},
            'destination'       => {type => SCALAR},
            'gateway'           => {type => SCALAR, default => undef},
            'action'            => {type => SCALAR},
            'vsid'              => {type => SCALAR},
        },
        allow_extra => 1
    );
    my $command_interface = $opts{command_interface};
    my $destination       = $opts{destination};

    ## the gateway may be of the format fd20:8b1e:b255:9112:0000:0000:0000:0001
    ## which can be compressed to fd20:8b1e:b255:9112::1
    ## So using Net::IP::ip_compress_address
    my $gateway           = Net::IP::ip_compress_address($opts{gateway}, 6);

    my $action            = $opts{action};
    my $vsid              = $opts{vsid};
    my $apiset            = $command_interface->apiset(
        category  => 'Node',
        interface => 'CLI',
        set       => 'Systemshell'
    );
    $apiset->set_privilege("privilege-level" => "root");

    ## FuseN changes require the validations to be vserverized due to routes
    ## being added per ipspace in different vnets
    my $netstat_cmd = "sudo vcontext -v ". $vsid
                        ." /usr/bin/netstat netstat -rnGD";
    my $raw_output = $apiset->execute_command(command => $netstat_cmd);
    my $response = NACL::APISet::Response::CLI::Freebsd->new(
                    output => $raw_output,
                    command => $netstat_cmd,
                    parser_file =>
                       'lib/NACL/APISet/Host/CLI/Freebsd/default/netstat.pm',
                    cdef_path=>
                       'lib/NACL/APISet/Host/CLI/Freebsd/default/netstat.cdef',
                    method_name => 'netstat',
                    execution_time => 1 ,
                    actual_command => $netstat_cmd,
                    execution_status => 0,
                    );
    my $output = $response->get_parsed_output();
    my $check  = 0;

    foreach my $inet (@$output) {
        if (Net::IP::ip_is_ipv6($gateway)) {
            $inet = $inet->{'addr-family-inet6'};
        } else {
            $inet = $inet->{'addr-family-inet'};
        }
        my $dest = _dest_calc($destination);
        if ($dest =~ m/(::\/0|0\.0\.0\.0)/) {
            $dest = "default";
        }
        my $interface;
        foreach my $row (@$inet) {
            if ($row->{'destination'} eq 'default') {
                $interface = $row->{'interface'};
            } elsif ($row->{'destination'} =~ /(\S+?)\//) {
                my $dest_only = $1;

                if ($destination =~ /$dest_only/) {
                    $interface = $row->{'interface'};
                }
            }
            no warnings 'uninitialized';
            my $interface_name = 'interface';
            if (   ($row->{'destination'} eq $dest)
                && ($row->{$interface_name} eq $interface)
                && ($row->{netif} == $vsid)) {
                $check = 1;
                if ($row->{'gateway'} =~ /(\S+?)%/) {
                    $row->{'gateway'} = $1;
                }
                if (($action eq 'create') && ($row->{'gateway'} eq $gateway))
                {
                    last;
                } else {
                    $check = 0;
                }
            }
        }
    } ## end foreach my $inet (@$output)
    my $msg;
    if (($action eq "create") && ($check == 0)) {
        $msg = 'Verification failed on Systemshell! Destination '
            . "not present in the routing table\n";
    }
    if (($action eq "delete") && ($check == 1)) {
        $msg = 'Verification failed on Systemshell! Destination '
            . "present in the routing table\n";
    }
    $Log->exit() if $may_exit;
    NACL::Exceptions::VerifyFailure->throw($msg) if ($msg);
} ## end sub _taskverify_systemshell

=head2 _taskverify_nodescope

   $task_obj->_taskverify_nodescope(
           command_interface => $node,
           destination       => $destination,
           action            => $action,
    );

   or

   NACL::STask::NetworkRoute->_taskverify_nodescope(
           command_interface => $node,
           destination       => $destination,
           action            => $action,
   );

(Class or Instance method)
This function takes destination & action as input and checks
the exact match of address from the apiset (route -sng)
to the destination on the nodescope & according to the
action throws the error message.

=over

=item Options

=over

=item C<< action => $action >>

(Required) The action to be done.
The value for action can be create or delete.

=back

=back

=cut

sub _taskverify_nodescope {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my $pkg = ref($pkg_or_obj) || $pkg_or_obj;
    my %opts = validate_with(
        params => \@args,
        spec   => {
            'command_interface' => {type => OBJECT},
            'destination'       => {type => SCALAR},
            'gateway'           => {type => SCALAR, default => undef},
            'action'            => {type => SCALAR},
            'vsid'              => {type => SCALAR},
            'vserver'           => {type => SCALAR},
        },
        allow_extra => 1
    );

    my %common_opts;
    $pkg_or_obj->_copy_common_component_params_with_ci(
        source => \%opts,
        target => \%common_opts
    );

    my $command_interface = $opts{command_interface};
    my $nodescope_apiset  = $command_interface->get_7m_or_nodescope_apiset();

    my $int_obj = NACL::CS::NetworkInterface->fetch(
        %common_opts,
        filter => { vserver => $opts{vserver}, },
        requested_fields => [ qw(home-node) ],
    );

    my $home_node = $int_obj->home_node();

    my $response
        = $nodescope_apiset->route_show('num-mode' => 1, 'group' => 1,
              'nodescope-node-name' => $home_node);

    my $output = $response->get_parsed_output();

    $opts{output} = $output; 

    if (Net::IP::ip_is_ipv6($opts{gateway})) {
        $pkg_or_obj->_taskverify_ipv6(%opts);
    } else {
        $pkg_or_obj->_taskverify_ipv4(%opts);
    }
    $Log->exit() if $may_exit;
} ## end sub _taskverify_nodescope

sub _taskverify_ipv4 {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    my $command_interface = $opts{command_interface};
    my $destination       = $opts{destination};
    my $gateway           = $opts{gateway};
    my $action            = $opts{action};
    my $vsid              = $opts{vsid};
    my $output            = delete $opts{output};
    
    my $check             = 0;

    my ($dest_pass, $mask_pass, $dest_only, $mask, $interface);
    $destination = "default" if ($destination =~ m/(::\/0|0\.0\.0\.0)/);

    if ($destination =~ /(\S+?)\/(\S+)/) {
        $dest_pass = $1;
        $mask_pass = $2;
    } else {
        $dest_pass = $destination;
    }
    foreach my $row (@$output) {
        my $row_modified;
        my $inet = $row->{'internet'};
        foreach my $inet_details (@$inet) {
            if ($inet_details->{'routing_group'} =~ /^$vsid[_].*/) {
                if ($inet_details->{'destination'} =~ /(\S+?)\/(\S+)/) {
                    $dest_only = $1;
                    $mask      = $2;
                } else {
                    $dest_only = $inet_details->{'destination'};
                }
                $interface = $inet_details->{'interface'}
                    if ($inet_details->{'destination'} eq 'default');
                if (   ($dest_pass && $dest_only)
                    && ($dest_pass =~ /$dest_only/)
                    && ($inet_details->{'interface'} =~ /$interface/)) {
                    if (   ($action =~ /create/)
                        && ($inet_details->{'gateway'} !~ /$gateway/)) {
                        $check = 0;
                        last;
                    }
                    if ($mask && $mask_pass) {
                        if ($mask eq $mask_pass) {
                            $check = 1;
                            last;
                        } else {
                            $check = 0;
                        }
                    } else {
                        $check = 1;
                        last;
                    }
                } ## end if (($dest_pass && $dest_only...
            }
        } ## end foreach my $inet_details (@$inet)
    } ## end foreach my $row (@$output)
    my $msg;
    if (($action eq "create") && ($check == 0)) {
        $msg = 'Verification failed on Nodescope! Destination '
            . "not present in the routing table\n";
    }
    if (($action eq "delete") && ($check == 1)) {
        $msg = 'Verification failed on Nodescope! Destination '
            . "present in the routing table\n";
    }
    $Log->exit() if $may_exit;
    NACL::Exceptions::VerifyFailure->throw($msg) if ($msg);
} ## end sub _taskverify_ipv4

=head2 _taskverify_ipv6

   $task_obj->_taskverify_ipv6(
           command_interface => $node,
           destination       => $destination,
           action            => $action,
    );

   or

   NACL::STask::NetworkRoute->_taskverify_ipv6(
           command_interface => $node,
           destination       => $destination,
           action            => $action,
   );

(Class or Instance method)
This function takes destination & action as input and checks
the exact match of ipv6 address from the apiset (route -sng)
to the destination & according to the action throws the error
message.

=over

=item Options

=over

=item C<< action => $action >>

(Required) The action to be done.
The value for action can be create or delete.

=back

=back

=cut

sub _taskverify_ipv6 {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    my $command_interface = $opts{command_interface};
    my $destination       = $opts{destination};

    ## the gateway may be of the format fd20:8b1e:b255:9112:0000:0000:0000:0001
    ## which can be compressed to fd20:8b1e:b255:9112::1
    ## So using Net::IP::ip_compress_address
    my $gateway           = Net::IP::ip_compress_address($opts{gateway}, 6);

    my $action            = $opts{action};
    my $vsid              = $opts{vsid};
    my $output            = delete $opts{output};

    my $check  = 0;
    $destination = "default" if ($destination =~ m/(::\/0|0\.0\.0\.0)/);
    foreach my $row (@$output) {
        my $row_modified;
        my $inet = $$row{'internet_v6'};
        my $interface;
        foreach my $inet_details (@$inet) {
            if ($inet_details->{'routing_group'} =~ /$vsid[_].*/) {
                if ($inet_details->{'destination'} eq 'default') {
                    $interface = $inet_details->{'interface'};
                } elsif ($inet_details->{'destination'} =~ /(\S+?)\//) {
                    my $dest_only = $1;
                    $interface = $inet_details->{'interface'}
                        if ($destination =~ /$dest_only/);
                }
                if (   ($inet_details->{'destination'} eq $destination)
                    && ($inet_details->{'interface'} eq $interface)) {
                    $check = 1;
                    if (   ($action eq 'create')
                        && ($inet_details->{'gateway'} eq $gateway)) {
                        last;
                    } else {
                        $check = 0;
                    }
                }
            }
        }
    }
    my $msg;
    if (($action eq "create") && ($check == 0)) {
        $msg = 'Verification failed ! Destination for ipv6 '
            . "not present in the routing table\n";
    }
    if (($action eq "delete") && ($check == 1)) {
        $msg = 'Verification failed ! Destination for ipv6 '
            . "present in the routing table\n";
    }
    $Log->exit() if $may_exit;
    NACL::Exceptions::VerifyFailure->throw($msg) if ($msg);
} ## end sub _taskverify_ipv6

sub _dest_calc {
    my $dest = shift;
    my $final_dest;

    if (($dest =~ /(\S+?)\/(\S+)/) && ($dest !~ /.*:/)) {
        my $destination = $1;
        my $netmask     = $2;
        my $quotient    = int($netmask / 8);
        my $rem         = $netmask % 8;
        my $mask;
        my @each_bit = split(/\./, $destination);
        my @each_addr;
        my $start;

        for (my $i = 0; $i < $quotient; $i++) {
            push(@each_addr, $each_bit[$i]);
        }
        my $subnet = 0;
        if ($rem ne 0) {
            my $bin_bits = 8 - $rem;
            for (my $j = 7; $j >= $bin_bits; $j--) {
                $mask   = 2**$j;
                $subnet = $subnet + $mask;
            }
            if ($subnet > $each_bit[$quotient]) {
                push(@each_addr, $each_bit[$quotient]);
            } else {
                push(@each_addr, $subnet);
            }
            $start = $quotient + 1;
        } else {
            $start = $quotient;
        }

        for (my $k = $start; $k < 4; $k++) {
            push(@each_addr, 0);
        }
        $final_dest = join(".", @each_addr);
        $final_dest = $final_dest . "\/" . "$netmask";
    } else {
        $final_dest = $dest;
    }
    return $final_dest;
} ## end sub _dest_calc

1;

