# $Id: //depot/prod/test/nacldev/lib/NACL/STask/NetworkRoutingGroupsRoute.pm#1 $
#
#   Copyright (c) 2013 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary NetworkRoutingGroupsRoute Task Module
## @author dl-nacl-dev@netapp.com
## @status shared
## @pod here

package NACL::STask::NetworkRoutingGroupsRoute;

use strict;
use warnings;

use base qw(NACL::C::Route NACL::STask::STask);

use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Net::IP qw(:PROC);
use Params::Validate qw(validate SCALAR SCALARREF validate_with OBJECT);
use NATE::Exceptions::Argument qw(:try);
use NACL::C::CommandInterface;
use Data::Dumper;
use NACL::C::Exceptions::NetworkRoutingGroupsRoute::AlreadyExists ();

use constant DEFAULT_TIMEOUT => 120;
use constant POLL_DELTA      => 20;

use Class::MethodMaker [ scalar => '_was_route_created' ];

=head1 NAME

NACL::STask::NetworkRoutingGroupsRoute

=head1 DESCRIPTION

C<NACL::STask::NetworkRoutingGroupsRoute> provides a number of well-defined but
potentially complex or multi-step methods related to routes in
ONTAP.

It builds on top of, and is a derived class of C<NACL::C::Route>,
and so it also provides methods that are more in the scope of
individual Route-related commands.  See C<NACL::C::Route> for
details.  This also means that a C<NACL::STask::NetworkRoutingGroupsRoute> object may
generally be used in place of a component object.

=head1 ATTRIBUTES

=head2 command_interface

(Required) As C<NACL::C::Route>.  A component object that
represents the host to which to send commands.

=head2 vserver

(Required for C-mode, ignored for 7-mode).  The name of the vserver
containing the route

=head2 'routing-group'

(Required for C-mode, ignored for 7-mode). The name of the routing group

=head2 destination

(Required) The destination represented as IpAddress/Mask.

=head1 METHODS

=head2 create

  my $route  = NACL::STask::NetworkRoutingGroupsRoute->create(
      command_interface   => $command_interface,
      vserver             => $vserver,
      'routing-group'     => $routing_group,
      destination         => $destination,
      gateway             => $gateway,
      nacltask_if_exists  => $action,      # default 'die'
      %other_options
  );

(Class method) Create a route.  This method provides additional
services beyond what's inherent to the product's route creation
commands, as controlled and described by the new C<nacltask_if_exists> options.

=over

=item Options

=over

=item C<< gateway => $gateway >>

(Required) The gateway is an IpAddress

=item C<< polling_interval => $interval >>

(Optional) Checks the progress of the routing-table upgrade every few seconds.
Also checks the deletion of routing-table entry on systemshell for ping method.
This option allows you to specify how many seconds to wait between checks.  If
this option is not specified, it will check every twenty seconds.

=item C<< "method-timeout" => $timeout >>

(Optional) Controls how long the command will wait before completing.
Default is 120 seconds.

=item C<< nacltask_if_exists=>$action >>

(Optional) What to do if the route to be created already exists.
If $action is "die" (the default), then fail with an exception (in the
same way that the route component would have: by trying the
route create and letting the product complain about the route
already existing).  If action is "reuse", then do nothing (return a
task object referring to the existing route).  If action is
"purge", then delete the route (see the C<purge> method, below)
before creating a new one.

=item C<< nacltask_verify=>$nacltask_verify_boolean >>

(Optional)
 If '0' (default), verification will not be performed.
 If '1', verification will be performed to ensure that the addition
 did happen successfully(check the addition of route in ngsh
 & nodescope & BSD).

=item C<< nacltask_verify_only_ngsh=>$nacltask_verify_only_ngsh >>

(Optional)
 If '0' (default), verification will not be performed.
 If '1', verification will be performed to ensure that the addition
 did happen successfully in ngsh.

=item C<< nacltask_verify_only_bsd=>$nacltask_verify_only_bsd >>

(Optional)
 If '0' (default), verification will not be performed.
 If '1', verification will be performed to ensure that the addition
 did happen successfully in bsd.

=item C<< _was_created => \$scalar >>

(Optional)
When this option is provided a reference to a scalar variable, the variable
gets filled in with a boolean value describing whether the route
was found (value will be 0; this scenario is possible when
nacltask_if_exists => "reuse") or whether the network route was created (value will
 be 1).
This is necessary to determine whether the network route needs to be
cleaned up later.

    my $was_created;
    my $route_obj = NACL::STask::NetworkRoutingGroupsRoute->create(
        nacltask_if_exists    => 'reuse',
        _was_created => \$was_created,
        %other_opts
    );
    # Operate on $route_obj here
    # ...
    # Now determine whether to clean up the route
    if ($was_created) {
        # New route was created. Clean it up.
        $route_obj->purge();
    }

=back

=item command_interface, apiset_must, apiset_should, method-timeout, routing-group, destination, gateway, etc.

All of the other various options to
L<< NACL::C::Route->create | lib-NACL-C-Route-pm/create >>

=back

=over

=item Exceptions

=over

=item C<NACL::C::Exceptions::NetworkRoutingGroupsRoute::AlreadyExists>

This type of exception is thrown when an attempt is made to create a routing group that already exists.

=item C<NACL::Exceptions::VerifyFailure>

This type of exception is thrown when verification fails for the created NetworkRoutingGroupsRoute in ngsh and bsd.

=item C<NACL::APISet::Exceptions::TimeoutException>

This type of exception is thrown when there is a failure in updating routing table with in specified timeout seconds.

=back

=cut

sub create {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = $pkg->_common_validate_with(
        params          => \@_,
        additional_spec => {
            method_timeout   => { type => SCALAR, default => DEFAULT_TIMEOUT },
            polling_interval => { type => SCALAR, default => POLL_DELTA },
            nacltask_if_exists        => $pkg->_if_exists_validate_spec(),
            nacltask_verify           => { type => SCALAR, default => 0 },
            nacltask_verify_only_ngsh => { type => SCALAR, default => 0 },
            nacltask_verify_only_bsd  => { type => SCALAR, default => 0 },
            _was_created              => { type => SCALARREF, optional => 1 },
            destination               => { type => SCALAR, optional => 1 },
        },
        allow_extra => 1
    );
    my $self;
    $self = $pkg->_redirect_call( _method => 'create', %opts );
    if ($self) {
        $Log->exit() if $may_exit;
        return $self;
    }
    my $if_exists         = delete $opts{nacltask_if_exists};
    my $was_created       = delete $opts{_was_created};
    my $command_interface = $opts{command_interface};
    my $verify            = delete $opts{nacltask_verify};
    my $verify_bsd        = delete $opts{nacltask_verify_only_bsd};
    my $verify_ngsh       = delete $opts{nacltask_verify_only_ngsh};
    my $polling_interval  = delete $opts{polling_interval};
    my $timeout           = delete $opts{method_timeout};
    $$was_created = 0;

    if ( Net::IP::ip_is_ipv6( $opts{gateway} ) ) {
        $opts{gateway} = Net::IP::ip_compress_address( $opts{gateway}, 'IPv6' );
    }

    if ( $command_interface->is_cmode ) {
        if ( Net::IP::ip_is_ipv6( $opts{gateway} ) ) {
            $opts{destination} = '::/0' if ( !defined $opts{destination} );
        }
        else {
            $opts{destination} = '0.0.0.0/0' if ( !defined $opts{destination} );
        }
    }

CREATE: {
            use warnings;
            try {
                $self = $pkg->SUPER::create(%opts);
                $$was_created = 1;
                $self->_was_route_created(1);
            }
            catch NACL::C::Exceptions::NetworkRoutingGroupsRoute::AlreadyExists with {
                my $exception = shift;
                $self = $pkg->_element_exists_handler(
                    create_opts        => \%opts,
                    nacltask_if_exists => $if_exists,
                    exception          => $exception,
                );
                if ( !$self ) {
                    no warnings qw(exiting);
                    redo CREATE;
                }
            }
            catch NACL::C::Exceptions::NetworkRoutingGroupsRoute::AddressNotFound with {
                my $exception = shift;
                require NACL::C::Options;
                NACL::C::Options->option(
                    command_interface => $command_interface,
                    'option-name'     => 'ip.v6.enable',
                    'option-value'    => 'on'
                );
                Tharn::snooze($polling_interval);
                my $end_time = time() + $timeout;

                while (1) {
                    if ( time() > $end_time ) {
                        require NACL::APISet::Exceptions::TimeoutException;
                        my $msg = "Timeout in updating the routing table\n";
                        NACL::APISet::Exceptions::TimeoutException->throw($msg)
                          if ($msg);
                    }
                    if ( $command_interface->mode() eq '7Mode' ) {
                        my $obj = $pkg->find(
                            command_interface => $command_interface,
                            filter            => { destination => 'default' }
                        );
                        last if $obj;

                    }
                    Tharn::snooze($polling_interval);
                } ## end while (1)
                if ( $command_interface->mode() eq 'CMode' ) {
                    require NACL::C::NetworkOptionsIpv6;
                    NetworkOptionsIpv6->modify(
                        command_interface => $command_interface,
                        enabled           => 'true'
                    );
                }
                no warnings qw(exiting);
                redo CREATE;
            };
} ## end CREATE:
    my $type = delete $opts{'destination-type'};
    my $action = 'create';

    if ($$was_created) {
        # Verify that the route was created
        if ( ($verify) || ($verify_ngsh) ) {
            ## Checking the route existence on the ngsh
            if ( $self->is_ipv6() ) {
                $pkg->_taskverify_ipv6(
                    command_interface => $command_interface,
                    destination       => $opts{destination},
                    gateway           => $opts{gateway},
                    action            => $action
                );
            }
            elsif ( $type eq "net" ) {
                $self->_taskverify_nodescope(
                    command_interface => $command_interface,
                    destination       => $opts{destination},
                    gateway           => $opts{gateway},
                    action            => $action
                );
            }
            else {
                $pkg->taskverify_create(%opts);
                if ( !$verify_ngsh ) {
                    if ( $command_interface->mode() eq "CMode" ) {
                        $self->_taskverify_nodescope(
                            command_interface => $command_interface,
                            destination       => $opts{destination},
                            gateway           => $opts{gateway},
                            action            => $action
                        );
                    }
                }
            } ## end else [ if ($self->is_ipv6(destination...
        } ## end if (($verify) || ($verify_ngsh...

        if ( ($verify) || ($verify_bsd) ) {
            ## Checking the route existence on the BSD
            $self->_taskverify_systemshell(
                command_interface => $command_interface,
                destination       => $opts{destination},
                gateway           => $opts{gateway},
                action            => $action
            );

        }
    }
    $Log->exit() if $may_exit;
    return $self;
} ## end sub create

=head2 purge

    $route->purge(%other_options);

    NACL::STask::NetworkRoutingGroupsRoute->purge(
      command_interface => $command_interface,
      'routing-group'   => $routing_group,
       vserver            => $Vserver,
       destination       => $destination,
      %other_options
    );

(Class or instance method) This method deletes a route, as well as
detaching or deleting other feature (ping to an ipaddress assigned to an interface ,
route gets added to the routing table in systemshell in cmode)
that may be preventing this route from being deleted from the systemshell:

=over

=item Options

=over

=item C<< nacltask_verify=>$nacltask_verify_boolean >>

(Optional)
 If '0' (default), verification will not be performed.
 If '1', verification will be performed to ensure that
 the deletion did happen successfully(check the deletion of route in ngsh
  & nodescope & BSD).

=item C<< nacltask_verify_only_ngsh=>$nacltask_verify_only_ngsh >>

(Optional)
 If '0' (default), verification will not be performed.
 If '1', verification will be performed to ensure that the deletion
 did happen successfully in ngsh.

=item C<< nacltask_verify_only_bsd=>$nacltask_verify_only_bsd >>

(Optional)
 If '0' (default), verification will not be performed.
 If '1', verification will be performed to ensure that the deletion
 did happen successfully in bsd.

=item C<< nacltask_if_purged => die | pass >>

(Optional, defaults to "die")
Specifies what to do if the route we're trying to purge no longer
exists. A value of C<die> (the default) will result in an exception
being thrown. A value of C<pass> will mean that the exception that would
normally occur when we're trying to delete a route that does not exist
is suppressed.

=back

=item command_interface, apiset_must, apiset_should, method-timeout, routing-group, destination, gateway, etc.

All of the other various options to
L<< NACL::C::Route->delete | lib-NACL-C-Route-pm/delete >>

=back

=over

=item Exceptions

=over

=item C<NACL::C::Exceptions::NetworkRoutingGroupsRoute::DoesNotExist>

This type of exception is thrown when an attempt is made to delete routing groups route that does not exists.

=item C<NACL::Exceptions::VerifyFailure>

This type of exception is thrown when verification fails for the deleted routing group route in ngsh and bsd.

=item C<NACL::APISet::Exceptions::TimeoutException>

This type of exception is thrown when there is a failure in deleting route entry from the routing table with in specified timeout seconds.

=back

=cut

sub purge {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my $pkg        = ref($pkg_or_obj) || $pkg_or_obj;
    my %opts       = $pkg_or_obj->_common_validate_with(
        params          => \@_,
        additional_spec => {
            method_timeout   => { type => SCALAR, default => DEFAULT_TIMEOUT },
            polling_interval => { type => SCALAR, default => POLL_DELTA },
            nacltask_verify  => { type => SCALAR, default => 0 },
            nacltask_verify_only_ngsh => { type => SCALAR, default => 0 },
            nacltask_verify_only_bsd  => { type => SCALAR, default => 0 },
            nacltask_if_purged =>
              $pkg_or_obj->_if_action_completed_already_validate_spec()
        },
        allow_extra => 1
    );
    my $status = 0;
    $pkg_or_obj->_redirect_call(
        _method => 'purge',
        status  => \$status,
        %opts
    );
    if ($status) {
        $Log->exit() if $may_exit;
        return;
    }
    my $command_interface = $opts{command_interface};
    my $if_purged         = delete $opts{nacltask_if_purged};
    my $verify            = delete $opts{nacltask_verify};
    my $verify_bsd        = delete $opts{nacltask_verify_only_bsd};
    my $verify_ngsh       = delete $opts{nacltask_verify_only_ngsh};
    my $polling_interval  = delete $opts{polling_interval};
    my $timeout           = delete $opts{method_timeout};
    my $action            = 'delete';
    my $already_purged    = 0;
    delete $opts{gateway} if ( defined $opts{gateway} );

    # Call the component delete. If route doesn't exist
    # then it would throw an exeception which is handled below.
    try {
        ## Purging of the ipaddress in routing table, that gets added to the bsd only,
        ## after pinging that ipaddress.
        if (   ( $opts{destination} !~ /\// )
            && ( $command_interface->mode() eq "CMode" ) )
        {
            require NACL::CS::NetworkInterface;
            my $cs = NACL::CS::NetworkInterface->fetch(
                command_interface => $command_interface,
                filter            => { address => $opts{destination}, },
                requested_fields  => [qw(curr-port)],
            );
            my $port = $cs->curr_port();
            my $nodescope_apiset =
              $command_interface->get_7m_or_nodescope_apiset();
            my $end_time = time() + $timeout;

            while (1) {
                if ( time() > $end_time ) {
                    require NACL::APISet::Exceptions::TimeoutException;
                    my $msg =
                      "Timeout in deleting entry from the routing table\n";
                    NACL::APISet::Exceptions::TimeoutException->throw($msg)
                      if ($msg);
                }
                $nodescope_apiset->ifconfig( 'interface' => $port, down => 1 );
                $nodescope_apiset->ifconfig( 'interface' => $port, up   => 1 );
                try {
                    $pkg->_taskverify_systemshell(
                        command_interface => $command_interface,
                        destination       => $opts{destination},
                        action            => $action
                    );
                }
                catch NACL::Exceptions::VerifyFailure with {
                    my $e   = shift;
                    my $msg = 'Verification failed on Systemshell! Destination '
                      . "present in the routing table\n";

                    if ( $e->text() =~ m[$msg]i ) {
                        Tharn::snooze($polling_interval);
                    }
                    else {
                        $Log->exit() if $may_exit;
                        $e->throw();
                    }
                };
                last if ( !$@ );
            } ## end while (1)

        } ## end if (($opts{destination...
        else {
            ## Purging ipv4 address on 7mode & cmode
            ##delete $opts{'destination-type'};
            $pkg->delete(%opts);
        }
    } ## end try
    catch NACL::C::Exceptions::NetworkRoutingGroupsRoute::DoesNotExist with {
        my $e = shift;
        if ( $if_purged =~ m[^die$]i ) {
            $Log->exit() if $may_exit;
            $e->throw();
        }
        else {
            $already_purged = 1;
        }
    };

    return if $already_purged;
    delete $opts{'destination-type'};
    if ( ($verify) || ($verify_ngsh) ) {
        ## Checking the route deletion on the ngsh
        if ( $pkg->is_ipv6( destination => $opts{'destination'} ) ) {
            $pkg->_taskverify_ipv6(
                command_interface => $command_interface,
                destination       => $opts{destination},
                action            => $action
            );
        }
        else {
            $pkg->_generic_purge_verify(%opts);
            if ( !$verify_ngsh ) {
                if ( $command_interface eq "CMode" ) {
                    $pkg->_taskverify_nodescope(
                        command_interface => $command_interface,
                        destination       => $opts{destination},
                        action            => $action
                    );
                }
            }
        } ## end else [ if ($pkg->is_ipv6(destination...
    } ## end if (($verify) || ($verify_ngsh...

    if ( ($verify) || ($verify_bsd) ) {
        ## Checking the route deletion on the BSD
        $pkg->_taskverify_systemshell(
            command_interface => $command_interface,
            destination       => $opts{destination},
            action            => $action
        );
    }

    $Log->exit() if $may_exit;

} ## end sub purge

=head2 _taskverify_ipv6

   $task_obj->_taskverify_ipv6(
           command_interface => $node,
           destination       => $destination,
           action            => $action,
    );

   or

   NACL::STask::NetworkRoutingGroupsRoute->_taskverify_ipv6(
           command_interface => $node,
           destination       => $destination,
           action            => $action,
   );

(Class or Instance method)
This function takes destination & action as input and checks
the exact match of ipv6 address from the apiset (route -sng)
to the destination & according to the action throws the error
message.

=over

=item Options

=over

=item C<< action => $action >>

(Required) The action to be done.
The value for action can be create or delete.

=back

=back

=cut

sub _taskverify_ipv6 {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my $pkg        = ref($pkg_or_obj) || $pkg_or_obj;
    my %opts       = validate_with(
        params => \@_,
        spec   => {
            'command_interface' => { type => OBJECT },
            'destination'       => { type => SCALAR },
            'gateway'           => { type => SCALAR, default => undef },
            'action'            => { type => SCALAR },
        },
    );

    my $command_interface = $opts{command_interface};
    my $destination       = $opts{destination};
    my $gateway           = $opts{gateway};
    my $action            = $opts{action};

    my $nodescope_apiset = $command_interface->get_7m_or_nodescope_apiset();
    my $response =
      $nodescope_apiset->route_show( 'num-mode' => 1, 'group' => 1 );

    my $check  = 0;
    my $output = $response->get_parsed_output();
    if ( $destination =~ m/(::\/0|0\.0\.0\.0)/ ) {
        $destination = "default";
    }

    foreach my $row (@$output) {
        my $row_modified;
        my $inet = $$row{'internet_v6'};
        my $interface;
        foreach my $inet_details (@$inet) {
            if ( $inet_details->{'destination'} eq 'default' ) {
                $interface = $inet_details->{'interface'};
            }
            elsif ( $inet_details->{'destination'} =~ /(.+?)\// ) {
                my $dest_only = $1;
                if ( $destination =~ /$dest_only/ ) {
                    $interface = $inet_details->{'interface'};
                }
            }
            if (   ( $inet_details->{'destination'} eq $destination )
                && ( $inet_details->{'interface'} eq $interface ) )
            {
                $check = 1;
                if (   ( $action eq 'create' )
                    && ( $inet_details->{'gateway'} eq $gateway ) )
                {
                    last;
                }
                else {
                    $check = 0;
                }
            }
        }
    }
    my $msg;
    if ( ( $action eq "create" ) && ( $check == 0 ) ) {
        $msg = 'Verification failed ! Destination for ipv6 '
          . "not present in the routing table\n";
    }
    if ( ( $action eq "delete" ) && ( $check == 1 ) ) {
        $msg = 'Verification failed ! Destination for ipv6 '
          . "present in the routing table\n";
    }
    $Log->exit() if $may_exit;
    require NACL::Exceptions::VerifyFailure;
    NACL::Exceptions::VerifyFailure->throw($msg) if ($msg);

} ## end sub _taskverify_ipv6

=head2 _taskverify_systemshell

   $task_obj->_taskverify_systemshell(
           command_interface => $node,
           destination       => $destination,
           action            => $action,
    );

   or

   NACL::STask::NetworkRoutingGroupsRoute->_taskverify_systemshell(
           command_interface => $node,
           destination       => $destination,
           action            => $action,
   );

(Class or Instance method)
This function takes destination & action as input and checks
the exact match of address from the apiset (route -sng)
to the destination on the systemshell & according to the
action throws the error message.

=over

=item Options

=over

=item C<< action => $action >>

(Required) The action to be done.
The value for action can be create or delete.

=back

=back

=cut

sub _taskverify_systemshell {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my $pkg        = ref($pkg_or_obj) || $pkg_or_obj;
    my %opts       = validate_with(
        params => \@_,
        spec   => {
            'command_interface' => { type => OBJECT },
            'destination'       => { type => SCALAR },
            'gateway'           => { type => SCALAR, default => undef },
            'action'            => { type => SCALAR },
        },
        allow_extra => 1    # To allow the common options through
    );

    my $command_interface = $opts{command_interface};
    my $destination       = $opts{destination};
    my $gateway           = $opts{gateway};
    my $action            = $opts{action};

    #my ($command_interface, $destination, $action) = values(%opts);
    my $apiset = $command_interface->apiset(
        category  => 'Node',
        interface => 'CLI',
        set       => 'Systemshell'
    );
    $apiset->set_privilege( "privilege-level" => "root" );
    my $response =
      $apiset->netstat( 'routing-tables' => 1, 'numeric-addr' => 1 );
    my $output = $response->get_parsed_output();
    my $check  = 0;
    my $inet_one;

    foreach my $inet (@$output) {
        if ( $pkg->is_ipv6( destination => $destination ) ) {
            $inet = $inet->{'addr-family-inet6'};
        }
        else {
            $inet = $inet->{'addr-family-inet'};
        }
        my $dest = _dest_calc($destination);
        if ( $dest =~ m/(::\/0|0\.0\.0\.0)/ ) {
            $dest = "default";
        }
        my $interface;
        foreach my $row (@$inet) {
            if ( $row->{'destination'} eq 'default' ) {
                $interface = $row->{'interface'};
            }
            elsif ( $row->{'destination'} =~ /(.+?)\// ) {
                my $dest_only = $1;
                if ( $destination =~ /$dest_only/ ) {
                    $interface = $row->{'interface'};
                }
            }
            my $interface_name = 'interface';
            if ( $command_interface->mode() eq '7Mode' ) {
                $interface      = 'default-ipspace';
                $interface_name = 'netif';
            }
            if (   ( $row->{'destination'} eq $dest )
                && ( $row->{$interface_name} eq $interface ) )
            {
                $check = 1;
                if ( $row->{'gateway'} =~ /(.+?)%/ ) {
                    $row->{'gateway'} = $1;
                }
                if (   ( $action eq 'create' )
                    && ( $row->{'gateway'} eq $gateway ) )
                {
                    last;
                }
                else {
                    $check = 0;
                }
            }
        }
    } ## end foreach my $inet (@$output)
    my $msg;
    require NACL::Exceptions::VerifyFailure;
    if ( ( $action eq "create" ) && ( $check == 0 ) ) {
        $msg = 'Verification failed on Systemshell! Destination '
          . "not present in the routing table\n";
    }
    if ( ( $action eq "delete" ) && ( $check == 1 ) ) {
        $msg = 'Verification failed on Systemshell! Destination '
          . "present in the routing table\n";
    }
    $Log->exit() if $may_exit;
    NACL::Exceptions::VerifyFailure->throw($msg) if ($msg);

} ## end sub _taskverify_systemshell

=head2 _taskverify_nodescope

   $task_obj->_taskverify_nodescope(
           command_interface => $node,
           destination       => $destination,
           action            => $action,
    );

   or

   NACL::STask::NetworkRoutingGroupsRoute->_taskverify_nodescope(
           command_interface => $node,
           destination       => $destination,
           action            => $action,
   );

(Class or Instance method)
This function takes destination & action as input and checks
the exact match of address from the apiset (route -sng)
to the destination on the nodescope & according to the
action throws the error message.

=over

=item Options

=over

=item C<< action => $action >>

(Required) The action to be done.
The value for action can be create or delete.

=back

=back

=cut

sub _taskverify_nodescope {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my $pkg        = ref($pkg_or_obj) || $pkg_or_obj;
    my %opts       = validate_with(
        params => \@_,
        spec   => {
            'command_interface' => { type => OBJECT },
            'destination'       => { type => SCALAR },
            'gateway'           => { type => SCALAR, default => undef },
            'action'            => { type => SCALAR },
        },
        allow_extra => 1    # To allow the common options through
    );

    my $command_interface = $opts{command_interface};
    my $destination       = $opts{destination};
    my $gateway           = $opts{gateway};
    my $action            = $opts{action};
    my $check             = 0;
    my $nodescope_apiset  = $command_interface->get_7m_or_nodescope_apiset();
    my $response =
      $nodescope_apiset->route_show( 'num-mode' => 1, 'group' => 1 );
    my $output = $response->get_parsed_output();
    my ( $dest_pass, $mask_pass, $dest_only, $mask, $interface );

    if ( $destination =~ m/(::\/0|0\.0\.0\.0)/ ) {
        $destination = "default";
    }

    if ( $destination =~ /(.+?)\/(.+)/ ) {
        $dest_pass = $1;
        $mask_pass = $2;
    }
    else {
        $dest_pass = $destination;
    }
    foreach my $row (@$output) {
        my $row_modified;
        my $inet = $row->{'internet'};
        foreach my $inet_details (@$inet) {
            if ( $inet_details->{'destination'} =~ /(.+?)\/(.+)/ ) {
                $dest_only = $1;
                $mask      = $2;
            }
            else {
                $dest_only = $inet_details->{'destination'};
            }
            if ( $inet_details->{'destination'} eq 'default' ) {
                $interface = $inet_details->{'interface'};
            }
            if (   ( $dest_pass && $dest_only )
                && ( $dest_pass =~ /$dest_only/ )
                && ( $inet_details->{'interface'} eq $interface ) )
            {
                if (   ( $action eq 'create' )
                    && ( $inet_details->{'gateway'} ne $gateway ) )
                {
                    $check = 0;
                    last;
                }
                if ( $mask && $mask_pass ) {
                    if ( $mask eq $mask_pass ) {
                        $check = 1;
                        last;
                    }
                    else {
                        $check = 0;
                    }
                }
            } ## end if (($dest_pass && $dest_only...
        } ## end foreach my $inet_details (@$inet)
    } ## end foreach my $row (@$output)
    my $msg;
    if ( ( $action eq "create" ) && ( $check == 0 ) ) {
        $msg = 'Verification failed on Nodescope! Destination '
          . "not present in the routing table\n";
    }
    if ( ( $action eq "delete" ) && ( $check == 1 ) ) {
        $msg = 'Verification failed on Nodescope! Destination '
          . "present in the routing table\n";
    }
    $Log->exit() if $may_exit;
    require NACL::Exceptions::VerifyFailure;
    NACL::Exceptions::VerifyFailure->throw($msg) if ($msg);

} ## end sub _taskverify_nodescope

sub _dest_calc {
    my $dest = shift;
    my $final_dest;
    if ( ( $dest =~ /(.+?)\/(.+)/ ) && ( $dest !~ /.*:/ ) ) {
        my $destination = $1;
        my $netmask     = $2;
        my $quotient    = int( $netmask / 8 );
        my $rem         = $netmask % 8;
        my $mask;
        my @each_bit = split( /\./, $destination );
        my @each_addr;
        my $start;

        for ( my $i = 0 ; $i < $quotient ; $i++ ) {
            push( @each_addr, $each_bit[$i] );
        }
        my $subnet = 0;
        if ( $rem ne 0 ) {
            my $bin_bits = 8 - $rem;
            for ( my $j = 7 ; $j >= $bin_bits ; $j-- ) {
                $mask   = 2**$j;
                $subnet = $subnet + $mask;
            }
            if ( $subnet > $each_bit[$quotient] ) {
                push( @each_addr, $each_bit[$quotient] );
            }
            else {
                push( @each_addr, $subnet );
            }
            $start = $quotient + 1;
        }
        else {
            $start = $quotient;
        }

        for ( my $k = $start ; $k < 4 ; $k++ ) {
            push( @each_addr, 0 );
        }
        $final_dest = join( ".", @each_addr );
        $final_dest = $final_dest . "\/" . "$netmask";
    }
    else {
        $final_dest = $dest;
    }
    return $final_dest;
} ## end sub _dest_calc

1;

