#
# Copyright (c) 2013 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary SecurityLogin  Task Module
## @author dl-nacl-dev@netapp.com
## @status shared
## @pod here

package NACL::STask::SecurityLogin;

use strict;
use warnings;

use base qw(NACL::C::SecurityLogin NACL::STask::STask);
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Params::Validate
  qw(validate validate_with SCALAR SCALARREF OBJECT HASHREF BOOLEAN ARRAYREF);

use NATE::Exceptions::Argument qw(:try);
use NATE::BaseException                         ();
use NACL::APISet::Exceptions::ResponseException ();

use NACL::C::Exceptions::SecurityLogin::AlreadyExists     ();
use NACL::C::Exceptions::SecurityLogin::DoesNotExist ();

use NACL::GeneralUtils qw(nacl_method_retry);

=head1 NAME

NACL::STask::SecurityLogin

=head1 DESCRIPTION

C<NACL::STask::SecurityLogin> provide methods to create, purge
security login in ONTAP.

It is created on top of C<NACL::C::SecurityLogin> component.
Since it is derived class of C<NACL::C::SecurityLogin>, we can use all the
methods of C<NACL::C::SecurityLogin> from the object of this task.

=head1 ATTRIBUTES

=head2 command_interface

(Required) A component object that represents the host to which to send commands.
See L<NACL::C::Component::command_interface|lib-NACL-C-Component-pm/command_interface>.

=head2 vserver

(Required) The name of the Vserver to be used.

=head2 username

(Required) Username to be created for the vserver.

=head1 METHODS

=head2 create

   my $SecurityLogin = NACL::STask::SecurityLogin->create(
                                  command_interface => $ci,
                                  authmethod        => $authmethod,
                                  vserver           => $vserver,
                                  application       => $application,
                                  username          => $username,
                                  password          => $password
                                  %other_options,
                                );

(Class Method) This method is used to create a security login for the given vserver.
If login exists, it will perform the action based on "nacltask_if_exists" parameter.
Default behavior would be "die".

It supports only CMode.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< 'vserver' => $vserver >>

(Required) Name of the vserver to be used.

=item C<< 'username' => $username >>

(Required) User name for the given vserver

=item C<< 'password' => $password >>

(Required) Password for the  given user

=item C<< 'authmethod' => $vserver >>

(Required) Name of the vserver to be used.

=item C<< 'application' => $username >>

(Required) Specifies the application name.

=item C<< 'role' => $role >>

(Optional for CLI , Required for ZAPI) Specifies the role name.

=item C<< "nacltask_if_exists => $action" >>

(Optional) What to do if the security login to be created already exists.
If $action is "die", then fail with an exception. If action is
"purge", then purge the security login(see the "purge" method, below)
before creating a new one. If $action is "reuse", It will return
the object of existing login.

=item C<< "nacltask_verify => $action" >>

(Optional) The user of this library can specify to verify whether the security
login is created or not. If the action is 0, which is default, it wont
verify the creation. If the user sets the action to 1, it will verify
using the component state of this library.

=item C<< nacltask_retry_tries_count => $tries_count >>

(Optional, defaults to 3)
The number of times to retry the create command if it fails because
the DbFileBusy is temporarily busy.

Note that it is possible to make it not retry the delete command by setting
the value of this argument to 1.

=item C<< nacltask_retry_sleep_time => $sleep_time >>

(Optional, defaults to 10 seconds)
The time to sleep between retries of the create command.

=item C<< "_was_created => \$scalar" >>

(Optional) When this option is provided with a reference to a scalar
variable, the variable gets filled in with a boolean value
describing whether the security login  was available (value will be 0;
this scenario is possible when if_exists => "reuse") or whether the
security login was created (value will be 1). This is necessary to
determine whether the security login needs to be cleaned up later.

command_interface, apiset_must, apiset_should, mode, etc. All of
the other various options supported by
L<< NACL::C::SecurityLogin->create| lib-NACL-C-SecurityLogin-pm/create >>

=back

=back

=over

=item Exceptions

=over

=item C<NACL::C::Exceptions::SecurityLogin::AlreadyExists>

This type of exception is thrown when an attempt is made to create a security login that already exists.

=item C<NACL::Exceptions::VerifyFailure>

This type of exception is thrown when verification fails for the created security login.

=back

=back

=cut

sub create {
    $Log->enter() if $may_enter;
    my $pkg = shift;

    my %opts = $pkg->_common_validate_with(
        params          => \@_,
        additional_spec => {
            nacltask_if_exists => $pkg->_if_exists_validate_spec(),
            _was_created       => { type => SCALARREF, optional => 1 },
            nacltask_verify    => { type => BOOLEAN, default => 0 },
            nacltask_retry_tries_count => { type => SCALAR, default => 3 },
            nacltask_retry_sleep_time  => { type => SCALAR, default => 10 },
        },
    allow_extra => 1,
    );
    # Transform %opts from the options we received into the options to
    # pass to the base class method.
    my ( $self, %common_opts, %nacltask_opts );
    $pkg->_copy_common_component_params_with_ci(
        source => \%opts,
        target => \%common_opts
    );

    my %retry_opts;
    $pkg->_move_retry_params(
        source => \%opts,
        target => \%retry_opts
    );

    $pkg->_move_nacltask_options(
        source => \%opts,
        target => \%nacltask_opts
    );

    my $was_created        = delete $opts{_was_created};
    my $nacltask_if_exists = delete $nacltask_opts{nacltask_if_exists};
    my $nacltask_verify    = delete $nacltask_opts{nacltask_verify};

  CREATE: {
        use warnings;
        try {
            nacl_method_retry(
                code => sub { $self = $pkg->SUPER::create( %opts ) },
                exceptions => 'NACL::C::Exceptions::SecurityLogin::DbFileBusy',
                %retry_opts
            );
            $$was_created = 1;
        }
        catch NACL::C::Exceptions::SecurityLogin::AlreadyExists with {
                my $exception = shift;
                $self = $pkg->_element_exists_handler(
                    create_opts        => \%opts,
                    nacltask_if_exists => $nacltask_if_exists,
                    exception          => $exception
                );
                if ( !$self ) {
                    no warnings qw(exiting);
                    redo CREATE;
                }
        };
    }

    if ($nacltask_verify) {
        delete $opts{password};
        $pkg->verify_state(%opts);
    }
    $Log->exit() if $may_exit;
    return $self;

} ## end sub create

=head2 purge

    $SecurityLogin->purge();

            (or)

    NACL::STask::SecurityLogin->purge(
              'command_interface'   => $ci,
              authmethod            => $authmethod,
              vserver               => $vserver,
              application           => $application,
              username              => $username,
    );

(Class or instance method)
This method is used to delete the security login.
It supports only CMode.

=over

=item Options

=over

=item C<< command_interface => $ci >>

(Required for class method, Not Applicable for instance method)
A component object that represents the host which to send commands.
See NACL::C::Component::command_interface

=item C<< 'vserver' => $vserver >>

(Required) Name of the vserver to be used.

=item C<< 'username' => $username >>

(Required) User name for the given vserver

=item C<< 'authmethod' => $vserver >>

(Required) Name of the vserver to be used.

=item C<< 'application' => $username >>

(Required) Specifies the application name.

=item C<< nacltask_verify => $nacltask_verify_boolean >>

(Optional)
If '0' (default), verification will not be performed.
If '1', verification will be performed to ensure that the deletion did happen
successfully.

=back

=back

=over

=item Exceptions

=over

=item C<NACL::C::Exceptions::SecurityLogin::DoesNotExist>

This type of exception is thrown when an attempt is made to delete a security login that does not exists.

=item C<NACL::Exceptions::VerifyFailure>

This type of exception is thrown when verification fails for the deleted security login. 

=back

=back

=cut

sub purge {
    $Log->enter() if $may_enter;
    my $pkg_or_obj      = shift;
    my %orig_opts       = @_;
    my $additional_spec = {
        nacltask_verify    => { type => BOOLEAN, default => 0 },
        nacltask_if_purged => { type => BOOLEAN, default => 'die' },
        _was_deleted       => { type => SCALARREF, optional => 1 },
    };

    # Throw away options specific to the task
    $pkg_or_obj->_hash_move(
        source => \%orig_opts,
        target => {},
        move   => [ keys %{$additional_spec} ]
    );
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@_,
        additional_spec => $additional_spec,
    );

    my %common_opts;
    $pkg_or_obj->_copy_common_component_params_with_ci(
        source => \%opts,
        target => \%common_opts
    );
    my $if_purged = delete $opts{nacltask_if_purged};

    my $was_deleted = delete $opts{_was_deleted};
    $$was_deleted = 0;

    use warnings;
    try {
        $pkg_or_obj->SUPER::delete( %common_opts, %orig_opts );
        $$was_deleted = 1;
    }
    catch NACL::C::Exceptions::SecurityLogin::DoesNotExist with {
        my $exception = shift;
        if ( $if_purged !~ /pass/i ) {
            $Log->exit() if $may_exit;
            $exception->throw();
        }
    };
    if ( $opts{nacltask_verify} ) {
        $pkg_or_obj->_generic_purge_verify(%orig_opts);
    }
    $Log->exit() if $may_exit;
}

=head2 password

    NACL::STask::SecurityLogin->password(
              command_interface     => $ci,
              vserver               => $vserver,
              username              => $username,
              newpwd                => $password1,
              newpwd2               => $password2,
              ...
    );

(Class method)
This method is used to to modify a password for a user.

Uses a CMode CLI/ZAPI APISet.
  Invokes security-login-modify-password for CMode ZAPI.
  Invokes 'useradmin diaguser password' command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface => $ci >>

(Required) A component object that represents the host which to send commands.
See L<< NACL::C::Component::command_interface|lib-NACL-C-Component-pm/command_interface >>

=item C<< 'vserver' => $vserver >>

(Required) Name of the vserver to be used.

=item C<< 'username' => $username >>

(Required) User name for the given vserver

=item C<< nacltask_retry_tries_count => $tries_count >>

(Optional, defaults to 3) The number of times to retry the security login password
command if it fails because the DbFileBusy is temporarily busy.

Note that it is possible to make it not retry the security login password command by setting
the value of this argument to 1.

=item C<< nacltask_retry_sleep_time => $sleep_time >>

(Optional, defaults to 10 seconds)
The time to sleep between retries of the security login password command.


=item C<< other => %others >>

command_interface, apiset_must, apiset_should, mode, etc. All of
the other various options supported by
L<< NACL::C::SecurityLogin->lock| lib-NACL-C-SecurityLogin-pm/lock >>


=back

=back

=over

=item Exceptions

=over

=item C<NACL::C::Exceptions::SecurityLogin::DoesNotExist>

This type of exception is thrown when an attempt is made to change the password of a security login that does not exists.

=back

=back

=cut

sub password {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;

    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            nacltask_retry_tries_count => { type => SCALAR, default => 3 },
            nacltask_retry_sleep_time  => { type => SCALAR, default => 10 },
        },
        allow_extra => 1,
    );

    my %retry_opts;
    $pkg->_move_retry_params(
        source => \%opts,
        target => \%retry_opts
    );

    nacl_method_retry(
        code => sub { $pkg->SUPER::password(%opts) },
        exceptions => 'NACL::C::Exceptions::SecurityLogin::DbFileBusy',
        %retry_opts
    );

    $Log->exit() if $may_exit;

} ## end sub password

=head2 lock

    NACL::STask::SecurityLogin->lock(
              command_interface     => $ci,
              vserver               => $vserver,
              username              => $username,
              ...
    );

(Class method)
This method is used to lock a user account with password auth method.

Uses a CMode CLI/ZAPI or 7Mode CLI APISet.
  Invokes 'security-login-lock' API for CMode ZAPI.
  Invokes 'useradmin diaguser lock' command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface => $ci >>

(Required for class method, Not Applicable for instance method)
A component object that represents the host which to send commands.
See L<< NACL::C::Component::command_interface|lib-NACL-C-Component-pm/command_interface >>

=item C<< 'vserver' => $vserver >>

(Required) Name of the vserver to be used.

=item C<< 'username' => $username >>

(Required) User name for the given vserver

=item C<< nacltask_retry_tries_count => $tries_count >>

(Optional, defaults to 3) The number of times to retry the security login lock command if it fails because
the DbFileBusy is temporarily busy.

Note that it is possible to make it not retry the security login lock command by setting
the value of this argument to 1.

=item C<< nacltask_retry_sleep_time => $sleep_time >>

(Optional, defaults to 10 seconds)
The time to sleep between retries of the security login lock command.

=item C<< nacltask_verify => 0|1 >>

(Optional, defaults to 0)
The user of this library can specify to verify whether the security
login is locked or not. If the action is 0, which is default, it wont
verify the lock operation. If the user sets the action to 1, it will verify
using the component state of this library.

=item C<< other => %others >>

command_interface, apiset_must, apiset_should, mode, etc. All of
the other various options supported by
L<< NACL::C::SecurityLogin->lock| lib-NACL-C-SecurityLogin-pm/lock >>

=back

=back

=over

=item Exceptions

=over

=item C<NACL::C::Exceptions::SecurityLogin::DoesNotExist>

This type of exception is thrown when an attempt is made lock the security login that does not exists.

=item C<NACL::Exceptions::VerifyFailure>

This type of exception is thrown when verification fails for the security login lock.

=back

=back

=cut

sub lock {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;

    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            nacltask_retry_tries_count => { type => SCALAR, default => 3 },
            nacltask_retry_sleep_time  => { type => SCALAR, default => 10 },
            nacltask_verify            => { type => SCALAR, default => 0 },
        },
        allow_extra => 1,
    );

    my %retry_opts;
    $pkg->_move_retry_params(
        source => \%opts,
        target => \%retry_opts
    );

    my $verify = delete $opts{nacltask_verify};

    nacl_method_retry(
        code => sub { $pkg->SUPER::lock(%opts) },
        exceptions => 'NACL::C::Exceptions::SecurityLogin::DbFileBusy',
        %retry_opts
    );

    if ($verify) {
        $pkg->verify_state( %opts, 'acctlocked' => 'yes' );
    }

    $Log->exit() if $may_exit;

} ## end sub lock

=head2 unlock

    NACL::STask::SecurityLogin->unlock(
              command_interface     => $ci,
              vserver               => $vserver,
              username              => $username,
              ...
    );

(Class method)
This method is used to unlock a user account with password auth method.

Uses a CMode CLI/ZAPI or 7Mode CLI APISet.
  Invokes 'security-login-unlock' API for CMode ZAPI.
  Invokes 'useradmin diaguser unlock' command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface => $ci >>

(Required for class method, Not Applicable for instance method)
A component object that represents the host which to send commands.
See L<< NACL::C::Component::command_interface|lib-NACL-C-Component-pm/command_interface >>

=item C<< 'vserver' => $vserver >>

(Required) Name of the vserver to be used.

=item C<< 'username' => $username >>

(Required) User name for the given vserver

=item C<< nacltask_retry_tries_count => $tries_count >>

(Optional, defaults to 3) The number of times to retry the security login lock command if it fails because
the DbFileBusy is temporarily busy.

Note that it is possible to make it not retry the security login unlock command by setting
the value of this argument to 1.

=item C<< nacltask_retry_sleep_time => $sleep_time >>

(Optional, defaults to 10 seconds)
The time to sleep between retries of the security login unlock command.

=item C<< nacltask_verify => 0|1 >>

(Optional, defaults to 0)
The user of this library can specify to verify whether the security
login is unlocked or not. If the action is 0, which is default, it wont
verify the unlock operation. If the user sets the action to 1, it will verify
using the component state of this library.

=item C<< other => %others >>

command_interface, apiset_must, apiset_should, mode, etc. All of
the other various options supported by
L<< NACL::C::SecurityLogin->lock| lib-NACL-C-SecurityLogin-pm/lock >>

=back

=back

=over

=item Exceptions

=over

=item C<NACL::C::Exceptions::SecurityLogin::DoesNotExist>

This type of exception is thrown when an attempt is made unlock security login that does not exists.

=item C<NACL::Exceptions::VerifyFailure>

This type of exception is thrown when verification fails for the security login unlock.

=back

=back

=cut

sub unlock {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;

    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            nacltask_retry_tries_count => { type => SCALAR, default => 3 },
            nacltask_retry_sleep_time  => { type => SCALAR, default => 10 },
            nacltask_verify            => { type => SCALAR, default => 0 },
        },
        allow_extra => 1,
    );

    my %retry_opts;
    $pkg->_move_retry_params(
        source => \%opts,
        target => \%retry_opts
    );

    my $verify = delete $opts{nacltask_verify};

    nacl_method_retry(
        code => sub { $pkg->SUPER::unlock(%opts) },
        exceptions => 'NACL::C::Exceptions::SecurityLogin::DbFileBusy',
        %retry_opts
    );

    if ($verify) {
        $pkg->verify_state( %opts, 'acctlocked' => 'no' );
    }

    $Log->exit() if $may_exit;

} ## end sub unlock

=head2 modify

   NACL::C::SecurityLogin->modify(
                                  command_interface => $ci,
                                  authmethod => $authmethod,
                                  vserver => $vserver,
                                  application => $application,
                                  username => $username,
                                  ...,
                                 );

         or

  $SecurityLogin_object->modify(...);

(Class or instance method)
This method is used to modify a login method.

Uses a CMode CLI/ZAPI APISet.
  Invokes 'security-login-modify-iter' API for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface => $ci >>

(Required for class method, Not Applicable for instance method)
A component object that represents the host which to send commands.
See L<< NACL::C::Component::command_interface|lib-NACL-C-Component-pm/command_interface >>

=item C<< 'vserver' => $vserver >>

(Required) Name of the vserver to be used.

=item C<< 'username' => $username >>

(Required) User name for the given vserver

=item C<< "authmethod" => $authmethod >>

(Required for class method, Not Applicable for instance method)

=item C<< 'password' => $password >>

(Required for class method, Not Applicable for instance method) Password for the  given user

=item C<< nacltask_verify => 0|1 >>

(Optional, defaults to 0)
The user of this library can specify to verify whether the security
login is modified or not. If the action is 0, which is default, it wont
verify the modify operation. If the user sets the action to 1, it will verify
using the component state of this library.

=item C<< other => %others >>

command_interface, apiset_must, apiset_should, mode, etc. All of
the other various options supported by
L<< NACL::C::SecurityLogin->lock| lib-NACL-C-SecurityLogin-pm/lock >>

=back

=back

=over

=item Exceptions

=over

=item C<NACL::C::Exceptions::SecurityLogin::DoesNotExist>

This type of exception is thrown when an attempt is made lock the security login that does not exists.

=item C<NACL::Exceptions::VerifyFailure>

This type of exception is thrown when verification fails for the security login lock.

=back

=back

=cut

sub modify {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;

    my %opts = $pkg_or_obj->_common_validate_with(
        params => \@args,
        additional_spec =>
          { nacltask_verify => { type => SCALAR, default => 0 }, },
        allow_extra => 1,
    );

    my $verify = delete $opts{nacltask_verify};

    $pkg_or_obj->SUPER::modify(%opts);

    if ($verify) {
        $pkg_or_obj->_generic_create_modify_verify(%opts);
    }

    $Log->exit() if $may_exit;

} ## end sub modify

1;
