#
# Copyright (c) 2001-2016 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Composite Aggregate Object Store Task Module
## @author ng-compaggr-smg-qa@netapp.com sureshb@netapp.com
## @status shared
## @pod here

package NACL::STask::StorageAggregateObjectStore;

use strict;
use warnings;
use base qw(NACL::C::Component NACL::STask::STask);
use List::Compare;
use NACL::C::SystemNode;
use NACL::ComponentUtils qw(Dumper);
use NACL::GeneralUtils qw(nacl_method_retry);
use NACL::InstantComponent qw(instant_component generate_pod);
use NATE::Log qw(log_global);
use NATE::BaseException qw(:try);
use NACL::Exceptions::NoElementsFound;
use Params::Validate qw(validate OBJECT SCALAR ARRAYREF UNDEF SCALARREF);

my $Log = log_global();

use Class::MethodMaker [
    scalar =>
        [{-type => 'NACL::C::CommandInterface::ONTAP'}, 'command_interface'],
    scalar => 'node',
];

sub config_create {
    $Log->enter();
    my (@args, $node, %opts, $pkg);
    ($pkg, @args) = @_;
    %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            command_interface     => {type => OBJECT, isa => 'NACL::C::Node'},
            'object-store-name'   => {type => SCALAR},
            'provider-type'       => {type => SCALAR, reqex => /AWS_S3|SGWS|S3EMU/},
            'ip-address'          => {type => SCALAR, optional => 1},
            's3-name'             => {type => SCALAR, optional => 1},
            'server'              => {type => SCALAR, optional => 1},
            'container-name'      => {type => SCALAR, optional => 1},
            'port'                => {type => SCALAR, optional => 1, default => 80},
            'access-key'          => {type => SCALAR, optional => 1},
            'secret-password'     => {type => SCALAR, optional => 1},
            'encrypt'             => {type => SCALAR, optional => 1},
            'encryption-key-id'   => {type => SCALAR, optional => 1},
            'object-checker-node' => {type => SCALAR, optional => 1},
            'ssl-enabled'         => {type => SCALAR, optional => 1},
            'ipspace'             => {type => SCALAR, optional => 1},
            _was_created          => {type => SCALARREF, optional => 1},
            nacltask_if_exists => {
                type      => SCALAR,
                callbacks => $pkg->_if_exists_validate_callback(),
                optional  => 1
            },
            nacltask_retry_tries_count => { type => SCALAR, default => 10 },
            nacltask_retry_sleep_time  => { type => SCALAR, default => 15 },
            nacltask_retry_num_times_code_ran =>
                { type => SCALARREF, optional => 1 },
        },
        allow_extra => 1,
    );

    my $ip_address = delete $opts{'ip-address'};
    my $server = $opts{'server'};
    my $s3_name = delete $opts{'s3-name'};
    my $container_name = $opts{'container-name'};

    if($ip_address ne ""){
        $Log->warn("This field 'ip-address' is now changed to 'server', support for ip-address is only available on debug builds for short time, update your script");
    }
    if($server eq "" and $ip_address ne ""){
        $opts{server} = $ip_address;
        $Log->warn("This field 'server' is now changed to accept only FQDN (like aws.netapp.com), support for ip-address=172.xx.yy.zz format is only available on debug builds for short time, update your script");
    }
    if($server eq "" and $ip_address eq ""){
        NATE::BaseException->throw("The fields -server and/or -ip-address are not defined, you must provide -server=<SGWS/AWS-S3 FDQN>");
    }
    if($s3_name ne ""){
        $Log->warn("This field 's3-name' is now changed to 'container-name', support for container-name is only available on debug builds for short time, update your script");
    }
    if($container_name eq "" and $s3_name ne ""){
        $opts{'container-name'} = $s3_name;
        $Log->warn("This field 's3-name' is now changed to 'container-name', support for this auto-conversion is only available on debug builds for short time, update your script");
    }
    if(!$opts{'container-name'} or !$opts{server}){
        NATE::BaseException->throw("The fields -container-name and/or -s3-name are not defined, you must provide -conainter-name=<bucket-name>");
    }

    my $if_exists = delete $opts{nacltask_if_exists};
    $node = delete($opts{command_interface});
    my $was_created          = delete $opts{_was_created};

    instant_component(
        component_name    => 'NACL::C::StorageAggregateObjectStoreConfig',
        command_interface => $node
    );

    my %nacltask_options;
    $pkg->_move_nacltask_options(
        source => \%opts,
        target => \%nacltask_options
    );

    my %retry_opts;
    $pkg->_copy_retry_params(
        source => \%nacltask_options,
        target => \%retry_opts
    );

    require NACL::CS::NetworkInterface;
    NACL::CS::NetworkInterface->fetch(
        command_interface => $node,
        filter            => {role => 'intercluster'},
    );

    # Initialize to 0 and set if we create the object store
    $$was_created = 0;

    my $seach_by_server_container = 0;

    try {
        NACL::C::StorageAggregateObjectStoreConfig->create(
            command_interface => $node,
            %opts
        );
        # If we get here then we created the object store.
        $$was_created = 1;
    }
    catch NATE::BaseException with {
        my $exception = $_[0];
        $Log->debug($exception->text());
        my $already_exists_text  = "already another object store config";
        my $duplicate_entry_text = "duplicate entry";
        if ($exception->text() !~ /$already_exists_text|$duplicate_entry_text/i) {
            nacl_method_retry(
                code       => sub {
                                NACL::C::StorageAggregateObjectStoreConfig->create(
                                    command_interface => $node,
                                    %opts
                                );
                              },
                exceptions => [
                    qw(NACL::APISet::Exceptions::CommandFailedException)
                ],
                %retry_opts
            );
            # If we get here then we created the object store.
            $$was_created = 1;
        }

        if ($exception->text() =~ /$already_exists_text/i) {
            # We potentially are using an object-store with a different
            # object-store name as this error message only means that the
            # server ip and bucket name are already in use
            $seach_by_server_container = 1;
        }
    };
    my $ret;

    if ($seach_by_server_container) {
        $ret = $pkg->config_show(
            command_interface => $node,
            server            => $opts{'server'},
            'container-name'  => $opts{'container-name'},

        );
    } else {
        $ret = $pkg->config_show(
            command_interface   => $node,
            'object-store-name' => $opts{'object-store-name'}
        );
    }

    return $ret;
} ## end sub config_create

sub config_show {
    $Log->enter();
    my (@args, $node, %opts, $pkg);
    ($pkg, @args) = @_;
    %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            command_interface     => {type => OBJECT, isa => 'NACL::C::Node'},
            'object-store-name'   => {type => SCALAR, optional => 1},
            'provider-type'       => {type => SCALAR, optional => 1},
            'ip-address'          => {type => SCALAR, optional => 1},
            's3-name'             => {type => SCALAR, optional => 1},
            'server'              => {type => SCALAR, optional => 1},
            'container-name'      => {type => SCALAR, optional => 1},
            'port'                => {type => SCALAR, optional => 1},
            'access-key'          => {type => SCALAR, optional => 1},
            'secret-password'     => {type => SCALAR, optional => 1},
            'encrypt'             => {type => SCALAR, optional => 1},
            'encryption-key-id'   => {type => SCALAR, optional => 1},
            'object-checker-node' => {type => SCALAR, optional => 1},
            'ssl-enabled'         => {type => SCALAR, optional => 1},
            'ipspace'             => {type => SCALAR, optional => 1},
        },
        allow_extra => 1,
    );
    $node = delete($opts{command_interface});
    instant_component(
        component_name    => 'NACL::C::StorageAggregateObjectStoreConfig',
        command_interface => $node
    );
    my $ip_address = delete $opts{'ip-address'};
    my $server = $opts{'server'};
    my $s3_name = delete $opts{'s3-name'};
    my $container_name = $opts{'container-name'};

    if(!defined($server) and defined($ip_address)){
        $opts{'server'} = $ip_address;
        $Log->warn("This field 'ip-address' is now changed to 'server' and accepts only FQDN (like aws.netapp.com), support for ip-address=172.xx.yy.zz format is only available on debug builds for short time, update your script");
    }
    if(!defined($container_name) and defined($s3_name)){
        $opts{'container-name'} = $s3_name;
        $Log->warn("This field 's3-name' is now changed to 'container-name', support for this auto-conversion is only available on debug builds for short time, update your script");
    }
    my $ca_configs;
    try {
        my %myopts = ();
        $myopts{'object-store-name'} = $opts{'object-store-name'}
            if (defined $opts{'object-store-name'});
        $myopts{'provider-type'} = $opts{'provider-type'}
            if (defined $opts{'provider-type'});
        $myopts{'server'} = $opts{'server'}
            if (defined($opts{'server'}));
        $myopts{'container-name'} = $opts{'container-name'}
            if (defined($opts{'container-name'}));
        $myopts{'encrypt'} = $opts{'encrypt'}
            if (defined $opts{'encrypt'});
        $myopts{'ssl-enabled'} = $opts{'ssl-enabled'}
            if (defined $opts{'ssl-enabled'});
        $myopts{'ipspace'} = $opts{'ipspace'}
            if (defined $opts{'ipspace'});
        $myopts{'port'} = $opts{'port'}
            if (defined $opts{'port'});

        $ca_configs = NACL::C::StorageAggregateObjectStoreConfig->find(
            command_interface => $node,
            filter            => {%myopts},
        );
    }
    catch NACL::Exceptions::NoElementsFound with {
        my $exception = $_[0];
        $Log->debug($exception->text());
        return 0;
    };
    require Data::Dumper;
    $Data::Dumper::Maxdepth = 1;
    print "Object-store-name=" . Dumper(\$ca_configs) . "\n";
    return $ca_configs;
} ## end sub config_show

sub config_modify {
    $Log->enter();
    my (@args, $node, %opts, $pkg);
    ($pkg, @args) = @_;
    %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            command_interface          => {type => OBJECT, isa => 'NACL::C::Node'},
            'object-store-name'        => {type => SCALAR},
            'new-object-store-name '   => {type => SCALAR, optional => 1},
            'ip-address'               => {type => SCALAR, optional => 1},
            's3-name'                  => {type => SCALAR, optional => 1},
            'server'                   => {type => SCALAR, optional => 1},
            'container-name'           => {type => SCALAR, optional => 1},
            'port'                     => {type => SCALAR, optional => 1},
            'access-key'               => {type => SCALAR, optional => 1},
            'secret-password'          => {type => SCALAR, optional => 1},
            'object-checker-node'      => {type => SCALAR, optional => 1},
            'ssl-enabled'              => {type => SCALAR, optional => 1},
            'ipspace'                  => {type => SCALAR, optional => 1},
            nacltask_retry_tries_count => { type => SCALAR, default => 10 },
            nacltask_retry_sleep_time  => { type => SCALAR, default => 15 },
            nacltask_retry_num_times_code_ran =>
                { type => SCALARREF, optional => 1 },
        },
        allow_extra => 1,
    );

    my $ip_address     = delete $opts{'ip-address'};
    my $server         = $opts{'server'};
    my $s3_name        = delete $opts{'s3-name'};
    my $container_name = $opts{'container-name'};
    $node              = delete $opts{command_interface};

    if(!defined($server) and defined($ip_address)){
        $opts{'server'} = $ip_address;
        $Log->warn("This field 'ip-address' is now changed to 'server' and accepts only FQDN (like aws.netapp.com), support for ip-address=172.xx.yy.zz format is only available on debug builds for short time, update your script");
    }
    if(!defined($container_name) and defined($s3_name)){
        $opts{'container-name'} = $s3_name;
        $Log->warn("This field 's3-name' is now changed to 'container-name', support for this auto-conversion is only available on debug builds for short time, update your script");
    }

    instant_component(
        component_name    => 'NACL::C::StorageAggregateObjectStoreConfig',
        command_interface => $node
    );

    my %nacltask_options;
    $pkg->_move_nacltask_options(
        source => \%opts,
        target => \%nacltask_options
    );

    my %retry_opts;
    $pkg->_copy_retry_params(
        source => \%nacltask_options,
        target => \%retry_opts
    );

    nacl_method_retry(
        code       => sub {
                        NACL::C::StorageAggregateObjectStoreConfig->modify(
                            command_interface => $node,
                            %opts
                        );
                      },
        exceptions => [
            qw(NACL::APISet::Exceptions::CommandFailedException)
        ],
        %retry_opts
    );

    my %ret = $pkg->config_show(command_interface => $node);
    return %ret;
} ## end sub config_modify

sub config_rename {
    $Log->enter();
    my (@args, $node, %opts, $pkg);
    ($pkg, @args) = @_;
    %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            command_interface   => {type => OBJECT, isa => 'NACL::C::Node'},
            'object-store-name' => {type => SCALAR},
            'new-object-store-name' => {type => SCALAR},
        },
        allow_extra => 1,
    );

    $node = delete($opts{command_interface});

    instant_component(
        component_name    => 'NACL::C::StorageAggregateObjectStoreConfig',
        command_interface => $node
    );

    NACL::C::StorageAggregateObjectStoreConfig->rename(
        command_interface => $node,
        %opts
    );

    my %ret = $pkg->config_show(command_interface => $node);
    return %ret;
} ## end sub config_rename

sub config_delete {
    $Log->enter();
    my (@args, $node, %opts, $pkg);
    ($pkg, @args) = @_;
    %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            command_interface   => {type => OBJECT, isa => 'NACL::C::Node'},
            'object-store-name' => {type => SCALAR},
            'force' => {type => SCALAR, default => 'false', optional => 1},
            _was_deleted        => { type => SCALARREF, optional => 1 },
            nacltask_if_purged => {
                type     => SCALAR,
                optional => 1,
                callbacks =>
                    $pkg->_if_action_completed_already_validate_callback()
            },
        },
        allow_extra => 1,
    );

    my $if_purged = delete $opts{nacltask_if_purged};
    $node = delete($opts{command_interface});
    my $was_deleted = delete $opts{_was_deleted};
    $$was_deleted = 0;
    instant_component(
        component_name    => 'NACL::C::StorageAggregateObjectStoreConfig',
        command_interface => $node
    );

    NACL::C::StorageAggregateObjectStoreConfig->delete(
        command_interface => $node,
        %opts
    );
    $$was_deleted = 1;
    return 1;
} ## end sub config_delete

sub object_checker_start {
    $Log->enter();
    my (@args, $node, %opts, $pkg);
    ($pkg, @args) = @_;
    %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            command_interface   => {type => OBJECT, isa => 'NACL::C::Node'},
            'object-store-name' => {type => SCALAR},
            'node' => {type => SCALAR, optional => 1},
        },
        allow_extra => 1,
    );

    $node = delete($opts{command_interface});

    instant_component(
        component_name => 'NACL::C::StorageAggregateObjectStoreObjectChecker',
        command_interface => $node
    );

    NACL::C::StorageAggregateObjectStoreObjectChecker->start(
        command_interface => $node,
        %opts
    );

    my %ret = $pkg->object_checker_show_status();
    return %ret;
} ## end sub object_checker_start

sub object_checker_show_status {
    $Log->enter();
    my (@args, $node, %opts, $pkg);
    ($pkg, @args) = @_;
    %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            command_interface => {type => OBJECT, isa => 'NACL::C::Node'},
            'object-store-name'    => {type => SCALAR, optional => 1},
            'status'               => {type => SCALAR, optional => 1},
            'objects_checked'      => {type => SCALAR, optional => 1},
            'orphan_objects_found' => {type => SCALAR, optional => 1},
            'non_data_objects'     => {type => SCALAR, optional => 1},
        },
        allow_extra => 1,
    );

    $node = delete($opts{command_interface});

    instant_component(
        component_name => 'NACL::C::StorageAggregateObjectStoreObjectChecker',
        command_interface => $node
    );

    NACL::CS::StorageAggregateObjectStoreObjectChecker->show_status(
        command_interface => $node,
        %opts
    );

    return 1;
} ## end sub object_checker_show_status

sub object_checker_dump {
    $Log->enter();
    my (@args, $node, %opts, $pkg);
    ($pkg, @args) = @_;
    %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            command_interface   => {type => OBJECT, isa => 'NACL::C::Node'},
            'object-store-name' => {type => SCALAR},
            'id' => {type => SCALAR, optional => 1},
        },
        allow_extra => 1,
    );

    $node = delete($opts{command_interface});

    instant_component(
        component_name => 'NACL::C::StorageAggregateObjectStoreObjectChecker',
        command_interface => $node
    );

    my %ret = NACL::C::StorageAggregateObjectStoreObjectChecker->dump(
        command_interface => $node,
        %opts
    );

    return %ret;
} ## end sub object_checker_dump

sub object_checker_review {
    $Log->enter();
    my (@args, $node, %opts, $pkg);
    ($pkg, @args) = @_;
    %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            command_interface   => {type => OBJECT, isa => 'NACL::C::Node'},
            'object-store-name' => {type => SCALAR},
            'uuid'                  => {type => SCALAR, optional => 1},
            'is-uuid-known'         => {type => SCALAR, optional => 1},
            'list-data-objects'     => {type => SCALAR, optional => 1},
            'list-non-data-objects' => {type => SCALAR, optional => 1},
        },
        allow_extra => 1,
    );

    $node = delete($opts{command_interface});

    instant_component(
        component_name => 'NACL::C::StorageAggregateObjectStoreObjectChecker',
        command_interface => $node
    );

    my %ret = NACL::C::StorageAggregateObjectStoreObjectChecker->review(
        command_interface => $node,
        %opts
    );

    return %ret;
} ## end sub object_checker_review

sub object_checker_delete {
    $Log->enter();
    my (@args, $node, %opts, $pkg);
    ($pkg, @args) = @_;
    %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            command_interface   => {type => OBJECT, isa => 'NACL::C::Node'},
            'object-store-name' => {type => SCALAR},
            'uuid'   => {type => SCALAR, optional => 1},
            'object' => {type => SCALAR, optional => 1},
        },
        allow_extra => 1,
    );

    $node = delete($opts{command_interface});

    instant_component(
        component_name => 'NACL::C::StorageAggregateObjectStoreObjectChecker',
        command_interface => $node
    );

    NACL::C::StorageAggregateObjectStoreObjectChecker->delete(
        command_interface => $node,
        %opts
    );

    return 1;
} ## end sub object_checker_delete

sub object_checker_cleanup {
    $Log->enter();
    my (@args, $node, %opts, $pkg);
    ($pkg, @args) = @_;
    %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            command_interface   => {type => OBJECT, isa => 'NACL::C::Node'},
            'object-store-name' => {type => SCALAR},
        },
        allow_extra => 1,
    );

    $node = delete($opts{command_interface});

    instant_component(
        component_name => 'NACL::C::StorageAggregateObjectStoreObjectChecker',
        command_interface => $node
    );

    NACL::C::StorageAggregateObjectStoreObjectChecker->cleanup(
        command_interface => $node,
        %opts
    );

    return 1;
} ## end sub object_checker_cleanup

sub find {
    $Log->enter();
    my (@args, $node, %opts, $pkg);
    ($pkg, @args) = @_;
    %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            command_interface   => {type => OBJECT, isa => 'NACL::C::Node'},
            aggregate           => {type => SCALAR, optional => 1},
            'object-store-name' => {type => SCALAR, optional => 1},
        },
        allow_extra => 1,
    );

    $node = delete($opts{command_interface});

    instant_component(
        component_name    => 'NACL::C::StorageAggregateObjectStore',
        command_interface => $node
    );

    NACL::C::StorageAggregateObjectStore->find(
        command_interface => $node,
        %opts,
    );
}

sub attach {
    $Log->enter();
    my (@args, $node, %opts, $pkg);
    ($pkg, @args) = @_;
    %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            command_interface   => {type => OBJECT, isa => 'NACL::C::Node'},
            aggregate           => {type => SCALAR},
            'object-store-name' => {type => SCALAR},
            'size-limit' => {type => SCALAR, optional => 1},
            nacltask_retry_tries_count => { type => SCALAR, default => 10 },
            nacltask_retry_sleep_time  => { type => SCALAR, default => 15 },
            nacltask_retry_num_times_code_ran =>
                { type => SCALARREF, optional => 1 },
        },
        allow_extra => 1,
    );

    $node = delete($opts{command_interface});

    require NACL::STask::NetworkInterface;
    my $ic_lif = NACL::STask::NetworkInterface->find(
        command_interface => $node,
        filter      => {role => 'intercluster', 'home-node' => $node->name},
        allow_empty => 1,
    );
    if(!$ic_lif){
        $Log->comment("There is no IC lif on node ". $node->name);
    }

    instant_component(
        component_name    => 'NACL::C::StorageAggregateObjectStore',
        command_interface => $node
    );

    my %nacltask_options;
    $pkg->_move_nacltask_options(
        source => \%opts,
        target => \%nacltask_options
    );

    my %retry_opts;
    $pkg->_copy_retry_params(
        source => \%nacltask_options,
        target => \%retry_opts
    );

    try {
        NACL::C::StorageAggregateObjectStore->attach(
            command_interface => $node,
            %opts
        );
    } catch NATE::BaseException with {
        my $exception = $_[0];
        my $text      = $exception->text();
        $Log->debug($text);
        if ($text !~ /duplicate entry|already a/i) {
            nacl_method_retry(
                code       => sub {
				  try {
				      NACL::C::StorageAggregateObjectStore->attach(
				          command_interface => $node,
                                          %opts
                                      );
			          } catch NATE::BaseException with {
				      my $exception = shift;
				      if ($exception->text() =~ /duplicate entry|already a/i) {
				          $Log->comment("already exsits");
				      } else {
				        $exception->throw();
				      }
			          };
			      },
                exceptions => [
                    qw(NACL::APISet::Exceptions::CommandFailedException)
                ],
                %retry_opts
            );
        }
    };

    return 1;
} ## end sub attach

sub ClientBucketclient_bucket_show {
    $Log->enter();
    my (@args, $node, %opts, $pkg);
    ($pkg, @args) = @_;
    %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            command_interface   => {type => OBJECT, isa => 'NACL::C::Node'},
            'handle'            => {type => SCALAR, optional => 1},
            'bucket'            => {type => SCALAR, optional => 1},
            'ipaddress'         => {type => SCALAR, optional => 1},
            'hostname'          => {type => SCALAR, optional => 1},
            'port'              => {type => SCALAR, optional => 1},
            'provider-type'     => {type => SCALAR, optional => 1},
            'provider-protc'    => {type => SCALAR, optional => 1},
            'access-key'        => {type => SCALAR, optional => 1},
            'secret-key'        => {type => SCALAR, optional => 1},
            'tls-encrypted'     => {type => SCALAR, optional => 1},
            'auth-version'      => {type => SCALAR, optional => 1},
            'loopback'          => {type => SCALAR, optional => 1},
            'sys-vsid'          => {type => SCALAR, optional => 1},
            'ipspace'           => {type => SCALAR, optional => 1},
            'server-data-encrypt' => {type => SCALAR, optional => 1},
            'alive'             => {type => SCALAR, optional => 1},
            'session-token'     => {type => SCALAR, optional => 1},
            'test'              => {type => SCALAR, optional => 1},
            'instance'          => {type => SCALAR, optional => 1},
        },
        allow_extra => 1,
    );

    $node = delete($opts{command_interface});

    instant_component(
        component_name    => 'NACL::C::StorageAggregateObjectStoreClientBucket',
        command_interface => $node
    );

    my %ret = NACL::CS::StorageAggregateObjectStoreClientBucket->fetch(
        command_interface => $node,
        %opts
    );

    return %ret;
}

sub show {
    $Log->enter();
    my (@args, $node, %opts, $pkg);
    ($pkg, @args) = @_;
    %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            command_interface   => {type => OBJECT, isa => 'NACL::C::Node'},
            'object-store-name' => {type => SCALAR},
            'bin-uuid'                    => {type => SCALAR, optional => 1},
            'bin-number'                  => {type => SCALAR, optional => 1},
            'size-limit'                  => {type => SCALAR, optional => 1},
            'object-store-availability'   => {type => SCALAR, optional => 1},
            'space-nearly-full-threshold' => {type => SCALAR, optional => 1},
            'space-full-threshold'        => {type => SCALAR, optional => 1},
        },
        allow_extra => 1,
    );

    $node = delete($opts{command_interface});

    instant_component(
        component_name    => 'NACL::C::StorageAggregateObjectStore',
        command_interface => $node
    );

    my %ret = NACL::CS::StorageAggregateObjectStore->fetch(
        command_interface => $node,
        %opts
    );

    return %ret;
} ## end sub show

sub modify {
    $Log->enter();
    my (@args, $node, %opts, $pkg);
    ($pkg, @args) = @_;
    %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            command_interface   => {type => OBJECT, isa => 'NACL::C::Node'},
            'object-store-name' => {type => SCALAR},
            'aggregate'                   => {type => SCALAR, optional => 1},
            'size-limit'                  => {type => SCALAR, optional => 1},
            'space-nearly-full-threshold' => {type => SCALAR, optional => 1},
            'space-full-threshold'        => {type => SCALAR, optional => 1},
        },
        allow_extra => 1,
    );

    $node = delete($opts{command_interface});

    instant_component(
        component_name    => 'NACL::C::StorageAggregateObjectStore',
        command_interface => $node
    );

    NACL::C::StorageAggregateObjectStore->modify(
        command_interface => $node,
        %opts
    );

    my %ret = $pkg->show();
    return %ret;
} ## end sub modify

sub df_AC {
    $Log->enter();
    my (@args, $node, %opts, $pkg);
    ($pkg, @args) = @_;
    %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            command_interface => {type => OBJECT, isa => 'NACL::C::Node'},
            aggregate => {type => SCALAR, optional => 1},
        },
        allow_extra => 1,
    );

    $node = delete($opts{command_interface});
    my $aggr = delete($opts{aggregate});

    #$opts{composite} = 'true';
    require NACL::C::StorageAggregate;

    my %ret = NACL::C::StorageAggregate->show_df(
        command_interface => $node,
        filter            => {aggregate => $aggr},
    );
    my $aggr_obj = NACL::C::StorageAggregate->find(
        command_interface => $node,
        filter            => {aggregate => $aggr},
    );
    my $aggr_home_node = $aggr_obj->get_one_state_attribute('home-name');
    require NACL::C::Node;
    my $home_node_obj = NACL::C::Node->find(
        command_interface => $node,
        filter            => {node => $aggr_home_node},
    );
    my $apiset = $home_node_obj->get_7m_or_nodescope_apiset(

        #category => 'Node',
        #set => '7Mode',
        #interface => 'CLI',
    );
    %ret = $apiset->execute_command(command => "df -AC $aggr");

    $Log->exit();
    return %ret;
} ## end sub df_AC

sub check_depenedencies {
    $Log->enter();
    my (@args, $node, %opts, $pkg);
    ($pkg, @args) = @_;
    %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            command_interface => {type => OBJECT, isa => 'NACL::C::Node'},
            aggregate => {type => SCALAR, optional => 1},
        },
        allow_extra => 1,
    );

    $Log->comment("Checking Feature Dependencies");
    $pkg->_check_create_IC_LIF(%opts);
    $pkg->_debug_build_config(%opts);
    $pkg->_check_ca_capability(%opts);
    $Log->exit();
    return 1;
} ## end sub check_depenedencies

#################################
######## PRIVATE MODULES ########
#################################

sub _check_create_IC_LIF {
    $Log->enter();
    my (@args, $node, %opts, $pkg);
    ($pkg, @args) = @_;
    %opts = $pkg->_common_validate_with(
        params => \@args,
        additional_spec =>
            {command_interface => {type => OBJECT, isa => 'NACL::C::Node'},},
        allow_extra => 1,
    );

    $node = delete($opts{command_interface});
    my $hostrec = $node->{hostrec};

    #print "\nhostrec" . Dumper(\$hostrec);

    my ($ic_exists, $intf);
    require NACL::C::Vserver;
    my $vs_admin = NACL::C::Vserver->find(
        command_interface => $node,
        filter            => {type => 'admin'},
    );
    $Log->comment("Using admin vserver " . $vs_admin->vserver);

    my @nodes = NACL::C::Node->find();
    $Log->comment("Nodes=" . Dumper(\@nodes));
    require NACL::STask::NetworkInterface;
    NACL::STask::NetworkInterface->find(
        command_interface => $node,
        allow_empty       => 1,
    );
    foreach my $n (@nodes) {
        $Log->comment("Check IC lif for node ". $n->name);
        my $ic_exists;
        try {
            $ic_exists =
            NACL::STask::NetworkInterface->find(
                command_interface => $n,
                filter      => {role => 'intercluster', 'home-node' => $n->name},
                allow_empty => 1,
            );
        } catch NATE::BaseException with {
            my $exception = $_[0];
            my $text      = $exception->text();
            $Log->debug($text);
            # on XCDR this fails due to CI on different cluster
            # and home-node on different cluster
            if ($text !~ /invalid value for field/i) {
                NATE::BaseException->throw($exception);
            }
                $ic_exists = 1;
        };

        $Log->comment("IC-lif". Dumper($ic_exists));
        if ($ic_exists and defined $ic_exists->{lif}) {
            $Log->comment("This node ". $n->name ." has intercluster lif, move on with other nodes");
            next;
        }

        $Log->comment("This node ". $n->name ." does not have intercluster lif, trying to create one");
        my $intf =
        $pkg->_create_new_ic_lif(
            command_interface => $n,
            vserver           => $vs_admin->vserver
        );
        $intf =
        NACL::STask::NetworkInterface->find(
            command_interface => $n,
            filter      => {role => 'intercluster', 'home-node' => $n->name},
            allow_empty => 1,
        );
        last if ($intf and defined $intf->{role});
        $intf =
        $pkg->_delete_create_ic_lif(
            command_interface => $n,
            vserver           => $vs_admin->vserver
        );
        $pkg->_verify_ic_lif(
            command_interface => $n,
            vserver           => $vs_admin->vserver
        );
    }
    $Log->exit();
    return 1;
} ## end sub _check_create_IC_LIF

sub _verify_ic_lif {
    $Log->enter();
    my (@args, $node, %opts, $pkg);
    ($pkg, @args) = @_;
    %opts = $pkg->_common_validate_with(
        params => \@args,
        additional_spec =>
            {command_interface => {type => OBJECT, isa => 'NACL::C::Node'},},
        allow_extra => 1,
    );
    $node = $opts{command_interface};
    my $ic_exists = NACL::STask::NetworkInterface->find(
        command_interface => $node,
        filter            => {role => 'intercluster', 'home-node' => $node->name},
    );
}

sub _create_new_ic_lif {
    $Log->enter();
    my (@args, $node, %opts, $pkg);
    ($pkg, @args) = @_;
    %opts = $pkg->_common_validate_with(
        params => \@args,
        additional_spec =>
            {command_interface => {type => OBJECT, isa => 'NACL::C::Node'},},
        allow_extra => 1,
    );

    $node = $opts{command_interface};
    $Log->comment("Node params:" . Dumper($node));
    my $hostrec    = $node->{hostrec};
    my $crt_ic_lif = 0;
    $Log->comment(
        "Check for available IPs for creating IC LIFs using DATA_IP2,3");
    foreach my $id (qw(2 3)) {
        next unless ($hostrec->{"DATA_IP$id"});
        my $port = $hostrec->{"DATA_PORT$id"};
        my $addr = $hostrec->{"DATA_IP$id"};
        my $mask = $hostrec->{"DATA_MASK$id"};
        $Log->comment("Creating IC LIF for node="
                . $node->name
                . ", Port=$port, Addr=$addr, Mask=$mask");
        my $ic_name = $node->name . '_ic';
        $ic_name =~ s/-/_/g;
        my %parms = (
            command_interface => $node,
            lif_name          => $ic_name,
            vserver           => $opts{vserver},
            role              => 'intercluster',
            home_node         => $node->name,
            home_port         => $port,
            address           => $addr,
            netmask           => $mask,
        );
        $Log->comment("parms=" . Dumper(\%parms));
        my $intf_e = $pkg->_nw_intf_check_addr(%parms);
        next if($intf_e);
        my $intf = $pkg->_nw_intf_create(%parms);
        $crt_ic_lif = ($intf) ? 1 : 0;
        last if ($crt_ic_lif);
    }
    return $crt_ic_lif;
} ## end sub _create_new_ic_lif

sub _delete_create_ic_lif {
    $Log->enter();
    my (@args, $node, %opts, $pkg);
    ($pkg, @args) = @_;
    %opts = $pkg->_common_validate_with(
        params => \@args,
        additional_spec =>
            {command_interface => {type => OBJECT, isa => 'NACL::C::Node'},},
        allow_extra => 1,
    );

    $node = $opts{command_interface};
    my $hostrec = $node->{hostrec};
    $Log->comment("Delete an existing data lif for ic lif from DATA_IP3,2");
    my $ret = 0;
    foreach my $id (qw(2 3 4 5)) {
        my $port = $hostrec->{"DATA_PORT$id"};
        my $addr = $hostrec->{"DATA_IP$id"};
        my $mask = $hostrec->{"DATA_MASK$id"};
        $Log->comment("DATA_IP$id=$addr, DATA_PORT$id=$port, DATA_MASK$id=$mask");
        $Log->comment("This ID=$id, DATA_IP$id is not defined on this testbed, move on to another ID")
            if(!$addr);
        next if (!$addr);
        $Log->comment("IC LIF for node="
                . $node->name
                . ", Port=$port, Addr=$addr, Mask=$mask");
        my $ic_name = $node->name . '_ic';
        $ic_name =~ s/-/_/g;
        require NACL::CS::NetworkInterface;
        my $intfs_cs = NACL::CS::NetworkInterface->fetch(
            command_interface => $node,
            filter            => {address => $addr},
            requested_fields => [ 'home-node', 'home-port', 'address', 'netmask', 'lif' ],
        );
        $Log->comment("Intfs_cs" . Dumper($intfs_cs));
        require NACL::C::NetworkInterface;
        my $intfs = NACL::C::NetworkInterface->find(
            command_interface => $node,
            filter            => {address => $addr},
        );
        $Log->comment("Intfs" . Dumper($intfs));
        if ($intfs) {
            $Log->comment("Delete the existing interface to reuse "
                    . " for creating admin vserver IC lif");
            try {
                $intfs->delete();
            } catch NACL::C::Exceptions::NetworkInterface::LifEnabled with {
                $intfs->modify('status-admin' => 'down',);
                $intfs->delete();
            };
        } else {
            NATE::BaseException->throw(
                "Unable to find interface with IP-address=$addr");
        }
        my %vs_data;

        $Log->comment("Creating admin vserver IC lif");
        my %parms = (
            command_interface => $node,
            lif_name          => $intfs_cs->{lif},
            vserver           => $opts{vserver},
            role              => 'intercluster',
            home_node         => $node->name,
            home_port         => $port,
            address           => $addr,
            netmask           => $mask,
        );
        $Log->comment("params=" . Dumper(\%parms));
        $ret = $pkg->_nw_intf_create(%parms);
        last if ($ret);
    }
    if (!$ret) {
        NATE::BaseException->throw(
            "Unable to create IC lif, check if you have sufficient IP addresses for node="
                . $node->name);
    }
    return $ret;
} ## end sub _delete_create_ic_lif

sub _nw_intf_check_addr {
    $Log->enter();
    my (@args, $node, %opts, $pkg);
    ($pkg, @args) = @_;
    %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            command_interface => {type => OBJECT, isa => 'NACL::C::Node'},
            address           => {type => SCALAR},
            lif_name          => {type => SCALAR, optional => 1},
            vserver           => {type => SCALAR, optional => 1},
            role              => {type => SCALAR, optional => 1},
            home_node         => {type => SCALAR, optional => 1},
            home_port         => {type => SCALAR, optional => 1},
            netmask           => {type => SCALAR, optional => 1},
        },
        allow_extra => 1,
    );
    my $intf_e =
    NACL::STask::NetworkInterface->find(
        command_interface => $opts{command_interface},
        filter => { address => $opts{address}, },
        allow_empty => 1,
    );
    my $found = ($intf_e)? 1: 0;
    $Log->comment("Network intf with same address ". $opts{address} . " exists(". $found .')');
    return $found;
}

sub _nw_intf_create {
    $Log->enter();
    my (@args, $node, %opts, $pkg);
    ($pkg, @args) = @_;
    %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            command_interface => {type => OBJECT, isa => 'NACL::C::Node'},
            lif_name          => {type => SCALAR},
            vserver           => {type => SCALAR},
            role              => {type => SCALAR},
            home_node         => {type => SCALAR},
            home_port         => {type => SCALAR},
            address           => {type => SCALAR},
            netmask           => {type => SCALAR},
        },
        allow_extra => 1,
    );
    require NACL::C::Vserver;
    my $vs_admin = NACL::C::Vserver->find(
        command_interface => $opts{command_interface},
        filter            => {type => 'admin'},
    );
    $Log->comment("Using admin vserver " . $vs_admin->vserver);
    my $ret = 0;
    try {
        my $e_intf_same_name =
        NACL::STask::NetworkInterface->find(
            command_interface => $opts{command_interface},
            filter => { lif => $opts{lif_name}, },
            allow_empty => 1,
        );
        if($e_intf_same_name){
            $Log->comment("A LIF with same name exists, rename it: ". Dumper($e_intf_same_name));
            my $e_intf_same_name_ren =
            NACL::STask::NetworkInterface->rename(
                command_interface => $opts{command_interface},
                lif               => $opts{lif_name},
                vserver           => $e_intf_same_name->vserver,
                newname           => $opts{lif_name}. '_1',
            );
        }

        my $intf =
        NACL::STask::NetworkInterface->create(
            command_interface => $opts{command_interface},
            lif               => $opts{lif_name},
            vserver           => $vs_admin,
            role              => $opts{role},
            'home-node'       => $opts{home_node},
            'home-port'       => $opts{home_port},
            address           => $opts{address},
            netmask           => $opts{netmask},
        );
        $ret = 1;
    }
    catch NATE::BaseException with {
        my $exception = $_[0];
        my $text      = $exception->text();
        my $exp_text  = 'A LIF with the same name already exists';
        $Log->debug($text);
        if ($text !~ /Duplicate ip address/i or $text !~ /$exp_text/i) {
            NATE::BaseException->throw($exception);
        }
    };
    $Log->exit();
    return $ret;
} ## end sub _nw_intf_create

sub _nw_intf_delete {
    $Log->enter();
    my (@args, $node, %opts, $pkg);
    ($pkg, @args) = @_;
    %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            node     => {type => OBJECT, isa => 'NACL::C::Node'},
            lif_name => {type => SCALAR},
            vserver  => {type => SCALAR},
        },
        allow_extra => 1,
    );
    try {
    require NACL::STask::NetworkInterface;
        my $intf = NACL::STask::NetworkInterface->purge(
            command_interface => $opts{node},
            lif               => $opts{lif_name},
            vserver           => $opts{vserver},

            #role              => $opts{role},
            #'home-node'       => $opts{home_node},
            #'home-port'      => $opts{home_port},
            #address          => $opts{address},
            #netmask          => $opts{netmask},
        );
    }
    catch NATE::BaseException with {
        my $exception = $_[0];
        my $text      = $exception->text();
        $Log->debug($text);
        if ($text !~ /in use/i) {
            NATE::BaseException->throw($exception);
        }
    };
    $Log->exit();
    return 1;
} ## end sub _nw_intf_delete

sub _debug_build_config {
    $Log->enter();
    my (@args, $node, %opts, $pkg);
    ($pkg, @args) = @_;
    %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            command_interface => {type => OBJECT, isa => 'NACL::C::Node'},

            #node => {type => SCALAR, optional => 1},
        },
        allow_extra => 1,
    );

    my $ci = delete($opts{command_interface});

    require NACL::STask::Node;
    my @nodes = NACL::C::Node->find();
    foreach my $n (@nodes) {
        my $is_aff = $n->get_one_state_attribute('is-all-flash-optimized');
        $Log->comment("Node " . $n->name . " is-all-flash-optimized=$is_aff");
        NACL::STask::Node->set_bootargs(
            command_interface => $ci,
            node              => $n->name,
            flags => [{'bootarg.composite_aggr_disk_type_check' => 'false'}]
        ) unless ($is_aff =~ /true/);
    }

    $Log->exit();
    return 1;
} ## end sub _debug_build_config

sub _check_ca_capability {
    $Log->enter();
    my (@args, $node, %opts, $pkg);
    ($pkg, @args) = @_;
    %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            command_interface => {type => OBJECT, isa => 'NACL::C::Node'},
        },
        allow_extra => 1,
    );

    my $ci = delete($opts{command_interface});

    require NACL::STask::Node;
    my @nodes = NACL::STask::Node->find();
    foreach my $n (@nodes) {
        my $node = $n->name();
        my $version_manager = $n->version_manager();
        require NACL::CS::SystemCapability;
        try {
            my $cap = NACL::CS::SystemCapability->fetch(
                command_interface => $ci,
                filter => {capability => 'composite_aggregate', node => $node}
            );

            if ($cap->{enabled} !~ /true/) {
                if ($version_manager->is_promoted()) {
                    $Log->warn("Composite aggregate capability is disabled on $node. "
                               ."Cannot modify on promoted builds");
                } else {
                    $Log->comment("Enabling Capability for comp-aggr on node $node");
                    require NACL::C::SystemCapability;
                    NACL::C::SystemCapability->modify(
                        command_interface => $ci,
                        capability        => 'composite_aggregate',
                        node              => $node,
                        enabled           => 'true',
                    );
                }
            }
        } catch NATE::BaseException with {
            my $exception = $_[0];
            my $text      = $exception->text();
            $Log->debug($text);
            # make this ignore the error for XCDR config
            if ($text !~ /is an invalid value for field/i) {
                $exception->throw();
            }
        };
    }
} ## end sub _check_ca_capability

1;
