#
# Copyright (c) 2001-2016 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary StorageFailover Task Module
## @author dl-nacl-dev@netapp.com
## @status shared
## @pod here

package NACL::STask::StorageFailover;

use strict;
use warnings;

use base qw(NACL::C::StorageFailover NACL::STask::STask);
use HAutils qw(aws_bring_filer_up);

use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Params::Validate qw(validate SCALAR UNDEF OBJECT BOOLEAN ARRAYREF);
use NACL::Exceptions::InvalidChoice qw(:try);
use NACL::C::Cluster;
use NACL::C::ClusterHa;
use NACL::CS::StorageFailover();
use NACL::CS::StorageFailoverInternalOptions();
use NACL::CS::StorageFailoverShowGiveback;
use NACL::CS::StorageFailoverShowTakeover;
use NACL::CS::EventLog;
use NACL::Transit;
use NACL::STask::Node;
use NACL::APISet::Exceptions::TimeoutException;
use NACL::APISet::Exceptions::UnexpectedOutputException;
use NACL::STask::Exceptions::UnexpectedSFOState;
use NACL::Transit::Exceptions::TransitException;
use NACL::C::Exceptions::StorageFailover::UnableToReadPartnerState;
use NACL::C::Exceptions::StorageFailover::DiskInventoryNotReceived;
use NACL::C::Exceptions::StorageFailover::PartnerNotFullyBooted;
use NACL::C::Exceptions::StorageFailover::GivebackAlreadyInProgress;
use NATE::BaseException;
use NACL::Exceptions::VerifyFailure;
use NACL::ComponentUtils qw(Dumper);
use NACL::GeneralUtils qw(nacl_method_retry);
use NACL::CS::SystemNodeAutosupport;
use NACL::C::SystemNodeAutosupport;
use NACL::C::Options;
use NACL::C::SystemNode;
use NATE::Exceptions::Argument;
use NACL::Exceptions::NoElementsFound;
use Scalar::Util qw(blessed);
use constant TAKEOVER_TIMEOUT          => 600;
use constant POLL_DELTA                => 10;
use constant POLL_DELTA_MBWAIT         => 2;
use constant DEFAULT_TIMEOUT           => 600;
use constant MAX_WFG_TIME              => 300;
use constant MAX_WFG_TO_UP             => 600;
use constant MAX_VETO_TIME             => 10;
use constant ONE_MIN                   => 60;
use constant TWO_MINS                  => 120;
use constant POLLING_INTERVAL          => 10;
use constant TRUE                      => 1;
use constant FALSE                     => 0;
use constant TEN_SEC_TIMEOUT           => 10;
use constant NODE_HEALTH_CHECK_TIMEOUT => 20;
use constant MAX_DIS_TIME              => 300;
use constant AWS_MODEL_REGEX           => 'CDvM200';

=head1 NAME

NACL::STask::StorageFailover

=head1 DESCRIPTION

C<NACL::STask::StorageFailover> provides a number of well-defined but
potentially complex or multi-step methods related to Storage Failover in
ONTAP.

It builds on top of, and is a derived class of C<NACL::C::StorageFailover>,
and so it also provides methods that are more in the scope of individual
Storage failover-related commands.  See C<NACL::C::StorageFailover> for
details.  This also means that a C<NACL::STask::StorageFailover> object
may generally be used in place of a component object.

=head1 CLEANUP METHODS

Cleanup can be registered for the following methods

Cleanup methods are,

     StorageFailover Method    Cleanup Method
    ------------------------------------------
      modify                   _modify

=head1 ATTRIBUTES

=head2 command_interface

(Required) As C<NACL::C::StorageFailover>.  A component object that
represents the host to which to send commands.


=head1 METHODS

=head2 enable

  NACL::STask::StorageFailover->enable(
      command_interface => $command_interface,
      node              => $node_object,
      nacltask_modify_mode_option => 1,
      %other_options
  );

(Class method) Enables storage-failover. This method provides
additional services beyond what's inherent to the product's storage-failover
commands.

(For CMode) ,this method will issue the appropriate commands to enable
storage-failover for one or multi storage-failover pairs in the cluster.

The method will not complain if storage-failover is already enabled on the
requested storage-failover pairs.

The method will throw an exception if the number of nodes in the cluster
is odd.  It will also throw an exception if there is a failure while enabling
storage-failover on any of the storage-failover pairs.

(For 7Mode), the method enables controller failover through the 'cf enable'
command in the node . If the node option is not passed in the input ,
it enables controller failover in the node belonging to the given
command_interface.

The method will not complain if controller failover is already enabled in the
node(s).

=over

=item Options

=over

=item C<< node=>$node_object >>

(Optional) (For CMode), this indicates the Node on which storage-failover is to be
enabled. Enabling storage-failover on one node of a storage-failover pair will
automatically enable it on the other.
If omitted, the method will enable all the storage-failover
pairs in the cluster.
(For 7Mode), this represents the node on which the controller failover is to
be enabled

=item C<< nacltask_modify_mode_option => $boolean >>

(Optional) This flag indicates if the user wants to modify
the storage failover 'ha_mode' option. Default is 0, which means that the
method does not have to modify the flag.
If the user wishes to modify the mode option to 'ha', using this flag,
then a reboot will be issued on the node . Also in case of two node or multi node
cluster 'refresh_command_interface' is done only for the node whose
'command_interface' is provided .

=item C<< nacltask_healthcheck_retries_count => $scalar >>

(Optional) This value indicates the number of retries count the health check should be performed
before modifying the Cluster ha value

=item C<< nacltask_healthcheck_sleep_time => $scalar >>

(Optional) This value indicates the sleep time while retrying for the health check status
before modifying the Cluster ha value

=item command_interface, apiset_must, apiset_should etc.

All of the other various options are described in
L<NACL::C::StorageFailover|lib-NACL-C-StorageFailover-pm>.

=back

=item Exceptions

=over

=item C<NACL::Exceptions::InvalidChoice>

This type of exception is thrown when singleton node is present or odd number of nodes are present in a particular cluster.

=back

=back

=cut

sub enable {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj , %opts ) = @_;

    $opts{action} = 'true';
    $pkg_or_obj->_enable_disable( %opts );

    $Log->exit() if $may_exit;
} ## end sub enable

=head2 disable

  NACL::STask::StorageFailover->disable(
      command_interface => $command_interface,
      node              => $node_object,
      nacltask_modify_mode_option => 1
      %other_options
  );

(Class method) Disables storage-failover. This method provides
additional services beyond what's inherent to the product's storage-failover
commands.

(For CMode), This method will issue the appropriate commands to disable
storage-failover for one or multi storage-failover pairs in the cluster.

The method will not complain if storage-failover is already disabled on the
requested storage-failover pairs.

The method will throw an exception if the number of nodes in the cluster
is odd.  It will also throw an exception if there is a failure while disabling
storage-failover on any of the storage-failover pairs.

(For 7Mode), the method disables controller failover through the 'cf disable'
command in the node . If the node option is not passed in the input ,
it disables controller failover in the node belonging to the given
command_interface.

The method will not complain if controller failover is already disabled in the
node(s).

=over

=item Options

=over

=item C<< node=>$node_object >>

(Optional) For CMode, Node on which storage-failover is to be disabled.
Disabling storage-failover on one node of a storage-failover pair will
automatically disable it on the other.
If omitted, the method will identify and disable all the storage-failover
pairs in the cluster.
(For 7Mode), this represents the node on which the controller failover is to
be disabled

=item C<< nacltask_modify_mode_option => $boolean >>

(Optional) This flag indicates if the user wants to modify
the storage failover 'ha_mode' option. Default is 0, which means that the
method does not have to modify the flag .
If the user wishes to modify the mode option to 'non_ha', using this flag,
then a reboot will be issued on the node . Also in case of two node or multi node
cluster 'refresh_command_interface' is done only for the node whose
'command_interface' is provided .

=item C<< nacltask_healthcheck_retries_count => $scalar >>

(Optional) This value indicates the number of retries count the health check should be performed
before modifying the Cluster ha value

=item C<< nacltask_healthcheck_sleep_time => $scalar >>

(Optional) This value indicates the sleep time while retrying for the health check status
before modifying the Cluster ha value

=item command_interface, apiset_must, apiset_should etc.

All of the other various options are described in
L<NACL::C::StorageFailover|lib-NACL-C-StorageFailover-pm>.

=back

=item Exceptions

=over

=item C<NACL::Exceptions::InvalidChoice>

This type of exception is thrown when singleton node is present or odd number of nodes are present in a particular cluster.

=back

=back

=cut

sub disable {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj , %opts ) = @_;

    $opts{action} = 'false';
    $pkg_or_obj->_enable_disable( %opts );

    $Log->exit() if $may_exit;
} ## end sub disable

# local routine
sub _enable_disable {
    $Log->enter() if $may_enter;

    my ( $pkg_or_obj, @args ) = @_;
    my $pkg        = ref($pkg_or_obj) || $pkg_or_obj;

    # validate the parameters
    my %opts = $pkg->_common_validate_with(
        params             => \@args,
        only_static_method => 1,
        additional_spec    => {
            node => { type => SCALAR | OBJECT | UNDEF, optional => 1 },
            nacltask_modify_mode_option => { type => BOOLEAN, default => 0 },
            nacltask_healthcheck_retries_count => { type => SCALAR, default => 5 },
            nacltask_healthcheck_sleep_time  => { type => SCALAR, default => 10 },
        },
        ignore_primary_keys => 1,
    allow_extra => 1
    );

    my $ci         = $opts{command_interface};
    my $node       = $opts{node};
    $Log->debug("Node is '$node'") if defined($node);
    my $action = $opts{action};    # action : enable = true disable = false
    my $action_str = ( $action eq 'true' ) ? 'enable' : 'disable';
    my $mode = $ci->mode();        # 7 or C mode
    $Log->debug("Action is '$action', mode is '$mode'");
    my $modify_mode_option = delete $opts{nacltask_modify_mode_option};

    my $healthcheck_retries_count = delete $opts{nacltask_healthcheck_retries_count} ;
    my $healthcheck_sleep_time = delete $opts{nacltask_healthcheck_sleep_time} ;

    # copy apiset_must_should so we can pass it down
    my %apiset_must_should;
    $pkg->_hash_copy(
        source => \%opts,
        target => \%apiset_must_should,
        copy   => [qw(apiset_must apiset_should)]
    );

    if ( $mode eq '7Mode' ) {

        # 7mode support
        if ( !( defined $node ) ) {

            #If the user has not specified the node name in the method
            #take the node name from the command interface
            $node = $ci->node();
        }

        if ( $action eq 'true' && $modify_mode_option ) {
            # Verify and set mode option if available
            # in this build or if passed by the user
            $pkg->_set_mode_in_param( %opts, '_node_to_set' => $node );
        }

        $pkg->modify(
            command_interface => $ci,
            node              => $node,
            enabled           => $action,
            %apiset_must_should
        );
        if ( $action eq 'false' && $modify_mode_option ) {
            # Verify and set mode option if available
            # in this build or if passed by the user
            $pkg->_set_mode_in_param( %opts, '_node_to_set' => $node );
        }

    } else {

        # Cmode support
        # get the nodes in the cluster
        my @nodes = NACL::C::Cluster->find(
            command_interface => $ci,
            %apiset_must_should
        );
        my $nodenum = $#nodes + 1;
        $Log->debug("Num of nodes on cluster '$nodenum'");

        #$nodenum=1; # force nodenum for testing
        if ( $nodenum % 2 ) {
                my $msg = 'Odd number of nodes ' . $nodenum . ' on cluster';
                $Log->exit() if $may_exit;
                NACL::Exceptions::InvalidChoice->throw($msg);
        } elsif ( $nodenum == 2 ) {
            $Log->debug('Two node cluster');
            my @nodes;
            if ( !defined($node) ) {
                if ($modify_mode_option) {
                    @nodes = NACL::C::Cluster->node_objs(
                        command_interface => $ci );
                }
                $node = @nodes ? $nodes[0] : $ci->node();
            }

            if ( $action eq 'true' && $modify_mode_option ) {

                # Verify and set mode option if available in this
                # build or if passed by the user
                if (@nodes) {
                    foreach my $each_node (@nodes) {
                        $pkg->_set_mode_in_param( %opts,
                            _node_to_set => $each_node );
                    }
                } else {
                    $pkg->_set_mode_in_param( %opts,
                        '_node_to_set' => $node );
                }
            }

            ## as per burt#848945 before modifying cluster HA the
            ## cluster health check should be performed.
            ## Performing 'cluster health check' and 'cluster ring show'
            ## as part of heath check up

            my $health_check = sub {
                NACL::STask::Cluster->cluster_health_check(
                    'health-check-mode' => "simple",
                    command_interface => $ci
                );
                NACL::STask::Cluster->check_rings(command_interface => $ci);
            };

            nacl_method_retry(
               code => $health_check,
               tries_count => $healthcheck_retries_count,
               sleep_time  => $healthcheck_sleep_time,
               exceptions  => [
                 'NACL::Exceptions::ClusterRingError',
                 'NACL::Exceptions::ClusterHealthCheckError',
                 'NACL::APISet::Exceptions::TimeoutException'
               ],
            );

            # enable/disable cluster ha
            # issue "cluster ha modify -configured true/false"

            NACL::C::ClusterHa->modify(
                command_interface => $ci,
                configured        => $action,
                %apiset_must_should
            );

            # enable/disable storage failover
            # issue "storage failover modify -node <> -enabled true/false"

            # Wait for node's health is "true". See Burt656971.
            $pkg->_wait_till_healthy_and_modify( %opts, node => $node);

            if ( $action eq 'false' && $modify_mode_option ) {

                # Verify and set mode option if available in
                # this build or if passed by the user
                if (@nodes) {
                    foreach my $each_node (@nodes) {
                        $pkg->_set_mode_in_param( %opts,
                            _node_to_set => $each_node );
                    }
                } else {
                    $pkg->_set_mode_in_param( %opts,
                        '_node_to_set' => $node );
                }
            }
        } else {    # > 2 nodes
            $Log->debug("Multi '$nodenum' node cluster");

            if ($node) {

                # enable/disable only the pair that was requested
                # issue "storage failover modify -node <> -enabled true/false"
                $Log->debug(
                    "Will '$action_str' storage-failover on '$node' and its partner"
                );

                if ( $action eq 'true' && $modify_mode_option ) {

                    # Verify and set mode option if available in
                    # this build or if passed by the user
                    $pkg->_set_mode_in_param( %opts,
                        '_node_to_set' => $node );
                }
               $pkg->_wait_till_healthy_and_modify(
             %opts, node => $node
        );
                if ( $action eq 'false' && $modify_mode_option ) {

                    # Verify and set mode option if available in
                    # this build or if passed by the user
                    $pkg->_set_mode_in_param( %opts,
                        '_node_to_set' => $node );
                }

            } else {

                # enable/disable each pair in the cluster
                my @states = NACL::CS::StorageFailover->fetch(
                    command_interface => $ci,
                    requested_fields  => [qw(node partner-name)],
                    %apiset_must_should
                );
                my @performed_list = ();
                foreach my $state (@states) {
                    my $mynode  = $state->node();
                    my $partner = $state->partner_name();
                    if ( grep( /$mynode/, @performed_list ) ) {
                        $Log->debug(
                            "Skipping '$mynode' since its already in desired state"
                        );
                    } else {
                        $Log->debug( "Will '$action_str' storage-failover "
                                . "on '$mynode' and its partner '$partner'" );
                        if ( $action eq 'true' && $modify_mode_option ) {

#Verify and set mode option if available in this build or if passed by the user
                            $pkg->_set_mode_in_param( %opts,
                                '_node_to_set' => $mynode );
                        }
                        ## wait till healthy and modify sfo
            $pkg->_wait_till_healthy_and_modify(%opts, node => $mynode);

                        if ( $action eq 'false' && $modify_mode_option ) {

#Verify and set mode option if available in this build or if passed by the user
                            $pkg->_set_mode_in_param( %opts,
                                '_node_to_set' => $mynode );
                        }

       # partner gets enabled / disabled when the node is enabled / disabled
       # push partner into performed list so we dont peform action on it again
                        push( @performed_list, $partner );
                    } ## end else [ if ( grep( /$mynode/, ...
                } ## end foreach my $state (@states)
            } ## end else [ if ($node)
        } ## end else [ if ( $nodenum < 2 )
    } ## end else [ if ( $mode eq '7Mode' )

    $Log->exit() if $may_exit;

} ## end sub _enable_disable

=head2 modify

    NACL::STask::StorageFailover->modify(
        command_interface => $command_interface,
        onfailure => 'true',
        onpanic   => 'true',
        %other_options
    );

    or

    NACL::STask::StorageFailover->modify(
        command_interface => $command_interface,
        ha_enable => '1',
        %other_options
    );

(Class method) Modify StorageFailover Options. This method will
modify the specified Storage Failover attributes and also verify that the
attribute is modified as expected.

=over

=item Options

=over

=item C<< nacltask_verify => 1  >>

(Optional)

  - would verify that the option is modified at the end, if set to 1
  - by default it is set to 0

This option is not considered if ha_enable or ha_disable are set to true.

Note that it is also possible to perform the verification later by calling
L<taskverify_create|lib-NACL-STask-STask-pm/taskverify_create>. This can
be used for cases where ONTAP chooses to not set all attributes to the exact
values specified. In the call we can specify those attributes we want
verified.

=item C<< ha_enable => '1'  >>

(Optional)
If the parameter is set to '1' then the call enables ha on the filer.

=item C<< ha_disable => '1'  >>

(Optional)

  - if the parameter is set to '1' then the call disables ha on the filer

=item C<< nacltask_to_cleanup => 0|1 >>

(Optional, default to 0(no to cleanup)
Flag indicating if this operation needs to be registered for clean up or not.

=item C<< nacltask_cleanup_manager >>

Cleanup manager to be used for registering.
Default : Will use the default cleanup manager.

=item All of the other various options are described in

L<< NACL::C::StorageFailover->modify|lib-NACL-C-StorageFailover-pm/modify >>

=back

=item Exceptions

=over

=item C<NACL::Exceptions::VerifyFailure>

This type of exception is thrown when verification for storage failover modify has failed.

=back

=back

=cut

sub modify {

    $Log->enter() if $may_enter;

    my $pkg = shift;

    # validate the parameters
    my %opts = $pkg->_common_validate_with(
        params => \@_,
        additional_spec => {
            node => { type => OBJECT | UNDEF | SCALAR, optional => 1 },
            nacltask_healthcheck_retries_count => { type => SCALAR, default => 5 },
            nacltask_healthcheck_sleep_time  => { type => SCALAR, default => 10 },
            $pkg->_cleanup_validate_spec(),
        },
        ignore_primary_keys => 1,
        allow_extra         => 1
    );

    my (%opts_for_cleanup, $opts_for_register, $register_for_cleanup );
    $pkg->_move_common_cleanup_opts(
        source  => \%opts,
        target  => \%opts_for_cleanup,
        nacltask_to_cleanup  => \$register_for_cleanup,
    );

    my $ha_enable  = delete $opts{ha_enable}  if defined $opts{ha_enable};
    my $ha_disable = delete $opts{ha_disable} if defined $opts{ha_disable};
    my $onpanic = $opts{'onpanic'} if defined $opts{'onpanic'};
    my $auto_giveback_after_panic = $opts{'auto-giveback-after-panic'}
        if defined $opts{'auto-giveback-after-panic'};
    my $verify = delete $opts{nacltask_verify};
    my $healthcheck_retries_count = delete $opts{nacltask_healthcheck_retries_count} ;
    my $healthcheck_sleep_time = delete $opts{nacltask_healthcheck_sleep_time} ;

    if ($ha_enable) {

        # Perform the extra code for enabling/disabling
        # if the option has been provided
        $pkg->_enable_disable(
            action => 'true',
            %opts,
            nacltask_healthcheck_retries_count => $healthcheck_retries_count,
            nacltask_healthcheck_sleep_time    => $healthcheck_sleep_time
        );
    } elsif ($ha_disable) {

        # Perform the extra code for enabling/disabling
        # if the option has been provided
        $pkg->_enable_disable(
            action => 'false',
            %opts,
            nacltask_healthcheck_retries_count => $healthcheck_retries_count,
            nacltask_healthcheck_sleep_time    => $healthcheck_sleep_time
        );
    }
    # Remove onpanic and auto-giveback-after-panic if HA is non-shared burt 1009539
    if (($onpanic && $onpanic =~ /false/i) ||
        ($auto_giveback_after_panic && $auto_giveback_after_panic =~ /false/i)) {
        # copy apiset_must_should to pass it into sfo fetch
        my %apiset_must_should;
        $pkg->_hash_copy(
            source => \%opts,
            target => \%apiset_must_should,
            copy   => [qw(apiset_must apiset_should)]
        );
        # check mode for non-shared HA
        my $ha_type = NACL::CS::StorageFailover->fetch(
            command_interface => $opts{command_interface},
            filter            => {
                node => 'local',
            },
            requested_fields  => [qw(ha-type)],
            %apiset_must_should,
        )->ha_type();
        if ($ha_type =~ /non_shared_storage/){
            $Log->warn('onpanic and auto-giveback-after-panic are not supported ' .
                'in a non-shared HA configuration. Reference BURT:1005111. Removing ' .
                'arguments from modify call');
            delete $opts{'onpanic'} if ($onpanic && $onpanic =~ /false/i);
            delete $opts{'auto-giveback-after-panic'}
                if ($auto_giveback_after_panic && $auto_giveback_after_panic =~ /false/i);
        }
    }
    my %common_opts;
    $pkg->_move_common_component_params_with_ci(
        source => \%opts,
        target => \%common_opts
    );
    $Log->debug('These are the common opts ' . Dumper( \%common_opts ));

    my $node_name;
    if ( defined $opts{node} ) {
        if ( blessed( $opts{node} )
            && $opts{node}->isa('NACL::C::SystemNode') )
        {
            $node_name = $opts{node}->node();
        } else {
            $node_name = $opts{node};
        }
        delete $opts{node};
    } else {
        $node_name = $common_opts{command_interface}->node();
    }

    $Log->debug('These are the modify opts ' . Dumper( \%opts ));

    if ($register_for_cleanup) {
        my %pass_opts;
        $pass_opts{node} = $node_name;

        $pass_opts{command_interface} = $common_opts{command_interface};
        $opts_for_register
          = $pkg->_common_modify_for_cleanup(%opts_for_cleanup,%opts,%pass_opts);
    }

    # Call the component's version to modify all other attributes if specified
    if ( keys %opts ) {
        $pkg->SUPER::modify( %opts, %common_opts, node => $node_name );

        if ( defined $verify ) {
            $pkg->verify_state(
                'node' => $node_name,
                %opts,
                %common_opts
            );

        }
    }

    # Register this resource with the cleanup manager for cleanup.
    foreach my $key ( keys (%{ $opts_for_register->{'orig_opts'}}) ) {
       delete $opts_for_register->{'orig_opts'}->{$key} if ( $opts_for_register->{'orig_opts'}->{$key} eq '-' );
    }
    $register_for_cleanup = (scalar ( keys (%{$opts_for_register->{'orig_opts'}}) ) > 0) ? 1 : 0 ;

    $pkg->_register_for_cleanup(%{$opts_for_register})
        if ($register_for_cleanup);
    $Log->exit() if $may_exit;

} ## end sub modify

=head2 check_state

  NACL::STask::StorageFailover->check_state(
      command_interface     => $command_interface,
      node                  => $node_object,
      partner               => $partner_object,
      node_state            => $node_state,
      partner_state         => $partner_state,
      %other_options
  );

(Class method) This method is used to check the state of the
storage-failover or the controller failover pair.

The state of the storage-failover or the controller failover pair is determined
and compared with expected state values.

An exception NACL::STask::Exceptions::UnexpectedSFOState will be thrown if either
of the expected states are not reached.

The method supports both CMode and 7Mode

=over

=item Options

=over

=item C<< node=>$node_object >>

(Required) Node Object.
Object of type NACL::C::Node of the first node in the storage-failover pair.

=item C<< partner=>$partner_object >>

(Required) Partner Object.
Object of type NACL::C::Node of  partner node in the storage-failover pair.

=item C<< node_state=>$node_state >>

This is the state in which the first node in the storage-failover or the
controller failover pair is expected to be in.

Default value is "up".

(For CMode) Possible state values are,

    "up" - state when the node is up with the OS booted.

    "takeover" - state when the node is in takeover mode.

    "takeover_autogiveback_scheduled" - state when the node is in takeover and
    autogiveback is scheduled

    "any_takeover" - state when the node is any takeover state

    "wfg" - state when the node is in "waiting for giveback"(wfg) prompt.

    "wfg_or_givewait" - state when the node is in *any* of the "waiting for giveback" prompts.

    "partial_giveback" - state when node has not sent all of the storage-failover
    aggregates with storage-failover policy to the partner.

    "sparecore" - state when partner selects sparecore style for dumping core.

    "dumpcore" - state when partner selects spraycore style for dumping core.

    "givewait" - state when the node is at "Press ctrl-C for maintainance mode".
    This is a transient state.  The node doesn't wait in this state at a prompt.

    "node_unreachable" - state when the partner is unreachable

    "system_id_changed_on_partner" - when the partner system id is changed after headswap

(For 7Mode) Possible state values are,

    "up" - state when the node is up with the OS booted.

    "takeover" - state when the node is in takeover mode.

    "wfg" - state when the node is in "waiting for giveback"(wfg) prompt.

    "sparecore" - state when partner selects sparecore style for dumping core.

    "dumpcore" - state when partner selects spraycore style for dumping core.

    "givewait" - state when the node is at "Press ctrl-C for maintainance mode".
    This is a transient state.  The node doesn't wait in this state at a prompt.

=item C<< partner_state=>$partner_state >>

(Optional) This is the state in which the second node in the storage-failover or
the controller failover pair is expected to be in.

Default value is "up".

(For CMode) Possible state values are,

    "up" - state when the node is up with the OS booted.

    "takeover" - state when the node is in takeover mode.

    "takeover_autogiveback_scheduled" - state when the node is in takeover and
    autogiveback is scheduled

    "any_takeover" - state when the node is any takeover state

    "wfg" - state when the node is in "waiting for giveback"(wfg) prompt.

    "partial_giveback" - state when node has not sent all of the storage-failover
    aggregates with storage-failover policy to the partner.

    "sparecore" - state when partner selects sparecore style for dumping core.

    "dumpcore" - state when partner selects spraycore style for dumping core.

    "givewait" - state when the node is at "Press ctrl-C for maintainance mode".
    This is a transient state.  The node doesn't wait in this state at a prompt.

    "partner_inhibit_takeover" - Partner node halted after disabling takeover due to
    which takeover is not possible

    "failed_due_to_diskcheck" - giveback failed due to presence of failed disk in the system

    "failed_due_to_autosupport" - giveback failed due to asup generation

    "waiting_for_cluster_applications" - state when the node is booting up and waiting for
    cluster application to come online

    "system_id_changed_on_node" - when the node system id is changed after headswap

    "unknown" - node is in unknown state

    "nocheck" - ignore the state of the node

    "relocating_sfo_aggregates" - state when partner node starts relocating SFO
    aggregates in preparation of takeover

    "sfo_phase_done" - state when SFO aggregates are relocated.

    "disk_inventory_not_received" - state of node when disk inventory hasn't been
    exchanged between the TO node and WFG node

(For 7Mode) Possible state values are,

    "up" - state when the node is up with the OS booted.

    "takeover" - state when the node is in takeover mode.

    "wfg" - state when the node is in "waiting for giveback"(wfg) prompt.

    "sparecore" - state when partner selects sparecore style for dumping core.

    "dumpcore" - state when partner selects spraycore style for dumping core.

    "givewait" - state when the node is at "Press ctrl-C for maintainance mode".
    This is a transient state.  The node doesn't wait in this state at a prompt.

=item command_interface, apiset_must, apiset_should etc.

All of the other various options are described in
L<NACL::C::StorageFailover|lib-NACL-C-StorageFailover-pm>.

=back

=item Exceptions

=over

=item C<NACL::STask::Exceptions::UnexpectedSFOState>

This type of exception is thrown when actual state of a particular node is not equal to the expected state.

=item C<NACL::Exceptions::InvalidChoice>

This type of exception is thrown when an invalid state is passed.

=item C< NACL::APISet::Exceptions::UnexpectedOutputException>

This type of exception is thrown when current state of the node is unexpected.

=back

=back

=cut

sub check_state {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my $pkg = ref($pkg_or_obj) || $pkg_or_obj;

    # validate the parameters
    my %opts = $pkg->_common_validate_with(
        params             => \@_,
        only_static_method => 1,
        additional_spec    => {
            node          => { isa  => 'NACL::C::Node', type     => OBJECT },
            partner       => { isa  => 'NACL::C::Node', type     => OBJECT },
            node_state    => { type => SCALAR },
            partner_state => { type => SCALAR,          optional => 1 },
        },
        ignore_primary_keys => 1,
    );

    my $ci   = $opts{command_interface};
    my $mode = $ci->mode();                # 7 or C mode

    $Log->debug(
        "Node is $opts{node}, Partner is $opts{partner}, mode is $mode");
    $Log->debug("Expected Node state is $opts{node_state} ");
    $Log->debug("Expected Partner state is $opts{partner_state}")
        if defined $opts{partner_state};

    if ( $mode eq '7Mode' ) {
        $pkg->_check_state_7Mode(%opts);
    } else {

        # copy apiset_must_should so we can pass it down
        my %apiset_must_should;
        $pkg->_hash_copy(
            source => \%opts,
            target => \%apiset_must_should,
            copy   => [qw(apiset_must apiset_should)]
        );

        my $node_name    = $opts{node}->node();
        my $partner_name = $opts{partner}->node();
        my %expected_state;
        tie %expected_state, "Tie::IxHash";
        %expected_state = ( $node_name => $opts{node_state} );
        if ( defined $opts{partner_state} ) {
            $expected_state{$partner_name} = $opts{partner_state};
        }
        $Log->debug(
            "Values in expected_state hash " . Dumper( \%expected_state ) );

        my %actual_state
            ; ## Used for creating a hash of nodenames and its storage-failover state
        my %partner_name;

        # Cmode support
        # Create a hash of nodenames and its storage-failover state
        my @sfo = NACL::CS::StorageFailover->fetch(
            command_interface => $ci,
            requested_fields  => [qw(node partner-name state-description)],
            filter            => {
                node => "$node_name,$partner_name",
            },
            %apiset_must_should
        );

        foreach my $sf (@sfo) {
            my $nodename = $sf->node();
            $actual_state{$nodename} = $sf->state_description();
            $partner_name{$nodename} = $sf->partner_name();
        }

        foreach my $nodename ( keys %expected_state ) {
            my $msg
                = "Node $nodename not in \"$expected_state{$nodename}\" state, current state is \"$actual_state{$nodename}\"";
            $Log->comment("Verifying the state for the node $nodename ");
            my $expected_state = $expected_state{$nodename};
            my $actual_state   = $actual_state{$nodename};

            my $throw_exception = sub {
                NACL::STask::Exceptions::UnexpectedSFOState->throw(
                    $msg,
                    monitoring_fields => ['state'],
                    state             => $actual_state
                );
            };
            if ( $expected_state eq "up" ) {
                if ( $actual_state !~
                    /^\"Connected to $partner_name{$nodename}(, Takeover is not possible: Storage failover is disabled)*\"/i )
                {
                    $Log->exit() if $may_exit;
                    $throw_exception->();
                }
            } elsif ( $expected_state eq "up_mailbox_unhealthy" ) {
                if ( $actual_state !~
                    /^\"Connected to $partner_name{$nodename}(, Takeover is not possible: Mailbox disks are not healthy)*\"/i )
                {
                    $Log->exit() if $may_exit;
                    $throw_exception->();
                }
            } elsif ( $expected_state eq "takeover" ) {
                if ( $actual_state ne "\"In takeover\"" ) {
                    $Log->exit() if $may_exit;
                    $throw_exception->();
                }
            } elsif ( $expected_state eq "takeover_autogiveback_scheduled" ) {
                my $regex1
                    = qr/In takeover, Auto giveback will be initiated in \d+ seconds/;
                my $regex2 = qr/In takeover, Auto giveback deferred/;

                $Log->debug(
                    "Expected state $nodename $regex1 \n" . " or $regex2 " );
                $Log->debug("current state is \'$actual_state{$nodename}\'");

                if (   ( $actual_state !~ /$regex1/si )
                    && ( $actual_state !~ /$regex2/si ) )
                {
                    $Log->exit() if $may_exit;
                    $throw_exception->();
                }
            } elsif ( $expected_state eq "any_takeover" ) {

                # match any takeover state
                my $regex = qr/In takeover/i;

                $Log->debug("Expected state $nodename $regex");
                $Log->debug("current state is \'$actual_state{$nodename}\'");

                if ( $actual_state !~ $regex ) {
                    $Log->exit() if $may_exit;
                    $throw_exception->();
                }
            } elsif ( $expected_state eq "wfg" ) {
                if ( !($actual_state eq
                    "\"Waiting for giveback (HA mailboxes)\""
                    || $actual_state eq
                    "\"Waiting for giveback\"") )
                {
                    $Log->exit() if $may_exit;
                    $throw_exception->();
                }
            } elsif ( $expected_state eq "wfg_or_givewait" ) {
                if ( $actual_state !~ /Waiting for giveback/i ) {
                    $Log->exit() if $may_exit;
                    $throw_exception->();
                }
            } elsif ( $expected_state eq "partial_giveback" ) {
                if ( $actual_state !~
                    /(Connected to|Waiting for) $partner_name{$nodename}, Partial giveback/ )
                {
                    $Log->exit() if $may_exit;
                    $throw_exception->();
                }
            } elsif ( $expected_state eq "sparecore" ) {
                if ( $actual_state ne "\"Dumping sparecore\"" ) {
                    $Log->exit() if $may_exit;
                    $throw_exception->();
                }
            } elsif ( $expected_state eq "dumpcore" ) {
                if ( $actual_state ne "\"Dumping core\"" ) {
                    $Log->exit() if $may_exit;
                    $throw_exception->();
                }
            } elsif ( $expected_state eq "givewait" ) {
                if ( !($actual_state eq
                    "\"Waiting for giveback (disk reservations)\""
                    || $actual_state eq
                    "\"Waiting for reservations to clear\"") )
                {
                    $Log->exit() if $may_exit;
                    $throw_exception->();
                }
            } elsif ( $expected_state eq "partner_inhibit_takeover" ) {
                if ( $actual_state !~
                    /"Waiting for $partner_name{$nodename}, Takeover is not possible: / .
                    /Partner node halted after disabling takeover(, Mailbox disks are not healthy)*/ .
                    /(, Local node missing partner disks)*(, NVRAM log not synchronized)*"/)
                {
                    $Log->exit() if $may_exit;
                    $throw_exception->();
                }
            } elsif ( $expected_state eq "failed_due_to_diskcheck" ) {
                if ( $actual_state ne
                    "\"Previous giveback failed in module: disk check\"" )
                {
                    $Log->exit() if $may_exit;
                    $throw_exception->();
                }
            } elsif ( $expected_state eq "failed_due_to_autosupport" ) {
                if ( $actual_state ne
                    "\"Previous giveback failed in module: autosupport\"" )
                {
                    $Log->exit() if $may_exit;
                    $throw_exception->();
                }
            } elsif ( $expected_state eq "node_unreachable" ) {
                if ( $actual_state ne "\"Node unreachable\"" ) {
                    $Log->exit() if $may_exit;
                    $throw_exception->();
                }
            } elsif ( $expected_state eq "waiting_for_cluster_applications" )
            {
                if ( $actual_state ne "\"Waiting for cluster applications\"" )
                {
                    $Log->exit() if $may_exit;
                    $throw_exception->();
                }
            } elsif ( $expected_state eq "system_id_changed_on_partner" ) {
                if ( $actual_state !~ /System ID changed on partner/ ) {
                    $Log->exit() if $may_exit;
                    $throw_exception->();
                }
            } elsif ( $expected_state eq "system_id_changed_on_local" ) {
                if ( $actual_state !~ /System ID changed on local/ ) {
                    $Log->exit() if $may_exit;
                    $throw_exception->();
                }
            } elsif ( $expected_state eq "sfo_phase_done" ) {
                if ( $actual_state !~ /Relocated SFO aggregates/ ) {
                    $Log->exit() if $may_exit;
                    $throw_exception->();
                }
            } elsif ( $expected_state eq "relocating_sfo_aggregates" ) {
                if ( $actual_state !~ /Relocating SFO aggregates/ ) {
                    $Log->exit() if $may_exit;
                    $throw_exception->();
                }
            } elsif ( $expected_state eq "unknown" ) {
                if ( $actual_state ne "Unknown" ) {
                    $Log->exit() if $may_exit;
                    $throw_exception->();
                }
            } elsif ( $expected_state eq "disk_inventory_not_received" ) {
                if ( $actual_state !~ /Disk Inventory information not yet recieved/ ) {
                    $Log->exit() if $may_exit;
                    $throw_exception->();
                }
            } elsif ( $expected_state eq "nocheck" ) {

                # Ignore the sfo state of the node
            } else {
                $Log->exit() if $may_exit;
                NACL::Exceptions::InvalidChoice->throw(
                    "Invalid state $expected_state{$nodename}");
            }
        }
    }
    $Log->exit() if $may_exit;
}

# Private Method to handle 7Mode check_state method call
sub _check_state_7Mode {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = @_;

    # 7mode support
    my ( $ci, $sfo, $node_name, $partner_name, $local_node_state,
        $partner_node_state );
    my ( %expected_state, $msg, %apiset_must_should );

    # Used for creating a hash of nodenames and its storage-failover state
    my %actual_state;

    # copy apiset_must_should so we can pass it down
    $pkg->_copy_common_component_params(
        source => \%opts,
        target => \%apiset_must_should,
    );

    # Set the command interface here
    $ci = $opts{command_interface};

    # fetch the name of the nodes from the Node object passed in the parameter
    $node_name    = $opts{node}->node();
    $partner_name = $opts{partner}->node();

    # Push the values in expected_state hash
    # Key=name of the node Value=expected state the node should be in ( given
    # by the caller ). This is done for local node and partner node
    tie %expected_state, "Tie::IxHash";
    %expected_state = ( $node_name => $opts{node_state} );
    if ( defined $opts{partner_state} ) {
        $expected_state{$partner_name} = $opts{partner_state};
    }
    $Log->debug(
        "Values in expected_state hash " . Dumper( \%expected_state ) );

    # Fetch the actual states of the node and print the same at debug level
    $sfo = NACL::CS::StorageFailover->fetch(
        command_interface => $ci,
        requested_fields =>
            [qw(local-firmware-state partner-firmware-state )],
        %apiset_must_should
    );

    $local_node_state   = $sfo->local_firmware_state();
    $partner_node_state = $sfo->partner_firmware_state();
    $Log->debug( "Actual Firmware state of the node " . $local_node_state );
    $Log->debug(
        "Actual Firmware state of the partner " . $partner_node_state );
    # Fill the name of the node and the actual state of the node in
    # another has %actual_state
    $actual_state{$node_name}    = $local_node_state;
    $actual_state{$partner_name} = $partner_node_state;
    $Log->debug( "Values in actual_state hash " . Dumper( \%actual_state ) );
    # Verify if the expected-state and actual-state of the nodes are equal
    # Otherwise throw an exception
    foreach my $node_in_hash ( keys %expected_state ) {
        my $msg
            = "Node $node_in_hash is not in \"$expected_state{$node_in_hash}\" state, current state is \"$actual_state{$node_in_hash}\"";
        $Log->comment("Verifying the state for the node $node_in_hash ");

        if ( $expected_state{$node_in_hash} eq "up" ) {
            if ( $actual_state{$node_in_hash} ne "SF_UP" ) {
                $Log->exit() if $may_exit;
                NACL::STask::Exceptions::UnexpectedSFOState->throw(
                    $msg,
                    monitoring_fields    => ['local_firmware_state'],
                    local_firmware_state => $actual_state{$node_in_hash}
                );
            } ## end if ( $actual_state{$node_in_hash} ne "SF_UP"
        } ## end if ( $expected_state{$node_in_hash}eq "up"
        elsif ( $expected_state{$node_in_hash} eq "takeover" ) {
            if ( $actual_state{$node_in_hash} ne "SF_TO" ) {
                $Log->exit() if $may_exit;
                NACL::STask::Exceptions::UnexpectedSFOState->throw(
                    $msg,
                    monitoring_fields    => ['local_firmware_state'],
                    local_firmware_state => $actual_state{$node_in_hash}
                );
            } ## end if ( $actual_state{$node_in_hash} ne "SF_TO"
        } ## end if ( $expected_state{$node_in_hash}eq "takeover"
        elsif ( $expected_state{$node_in_hash} eq "wfg" ) {
            if ( $actual_state{$node_in_hash} ne "SF_MBWAIT" ) {
                $Log->exit() if $may_exit;
                NACL::STask::Exceptions::UnexpectedSFOState->throw(
                    $msg,
                    monitoring_fields    => ['local_firmware_state'],
                    local_firmware_state => $actual_state{$node_in_hash}
                );
            } ## end if ( $actual_state{$node_in_hash} ne "SF_MBWAIT"
        } ## end if ( $expected_state{$node_in_hash}eq "wfg"
        elsif ( $expected_state{$node_in_hash} eq "givewait" ) {
            if ( $actual_state{$node_in_hash} ne "SF_GIVEWAIT" ) {
                $Log->exit() if $may_exit;
                NACL::STask::Exceptions::UnexpectedSFOState->throw(
                    $msg,
                    monitoring_fields    => ['local_firmware_state'],
                    local_firmware_state => $actual_state{$node_in_hash}
                );
            } ## end if ( $actual_state{$node_in_hash} ne "SF_GIVEWAIT"
        } ## end if ( $expected_state{$node_in_hash}eq "givewait"
        elsif ( $expected_state{$node_in_hash} eq "sparecore" ) {
            if ( $actual_state{$node_in_hash} ne "SF_SPARECORE" ) {
                $Log->exit() if $may_exit;
                NACL::STask::Exceptions::UnexpectedSFOState->throw(
                    $msg,
                    monitoring_fields    => ['local_firmware_state'],
                    local_firmware_state => $actual_state{$node_in_hash}
                );
            } ## end if ( $actual_state{$node_in_hash} ne "SF_SPARECORE"
        } ## end if ( $expected_state{$node_in_hash}eq "sparecore"
        elsif ( $expected_state{$node_in_hash} eq "dumpcore" ) {
            if ( $actual_state{$node_in_hash} ne "SF_DUMPCORE" ) {
                $Log->exit() if $may_exit;
                NACL::STask::Exceptions::UnexpectedSFOState->throw(
                    $msg,
                    monitoring_fields    => ['local_firmware_state'],
                    local_firmware_state => $actual_state{$node_in_hash}
                );
            } ## end if ( $actual_state{$node_in_hash} ne "SF_DUMPCORE"
        } ## end if ( $expected_state{$node_in_hash}eq "dumpcore"

        elsif ($expected_state{$node_in_hash} eq "partial_giveback"
            || $expected_state{$node_in_hash} eq
            "takeover_autogiveback_scheduled" ) {
            $Log->exit() if $may_exit;
            NACL::Exceptions::InvalidChoice->throw(
                      "The state $expected_state{$node_in_hash} "
                    . " is not yet supported in 7Mode " );
        } else {
            $Log->exit() if $may_exit;
            NACL::Exceptions::InvalidChoice->throw(
                "Invalid state $expected_state{$node_in_hash}");
        }
    }    ## End foreach hash
} ## end sub _check_state_7Mode

=head2 wait_for_state

  NACL::STask::StorageFailover->wait_for_state(
      command_interface     => $command_interface,
      node                  => $node_object,
      partner               => $partner_object,
      node_state            => $node_state,
      partner_state         => $partner_state,
      'method-timeout'      => $method_timeout,
      nacltask_poll_interval => $poll_interval,
      %other_options
  );

(Class method) This method is used to wait on the state of the SFO/CFO operation
on the given storage-failover or controller failover pair . It does so by
calling the method check_state until the desired state is reached or timeout
occurs.

The state of the storage-failover/controller failover pair is determined and
compared with expected state values, using the check_state package method .

A Timeout exception will be thrown if either of the expected states are
not reached within the given wait time..

The method supports both CMode and 7Mode

=over

=item Options

=over

=item C<< node=>$node_object >>

(Required) Node Object.
Object of type NACL::C::Node of the first node in the storage-failover pair.

=item C<< partner=>$partner_object >>

(Required) Partner Node object.
Object of type NACL::C::Node of the partner node of the storage-failover pair.

=item C<< node_state=>$node_state >>

(Required) Node State

This is the state in which the first node in the storage-failover
pair is expected to be in.

(For CMode) Possible state values are,

    "up" - state when the node is up with the OS booted.

    "takeover" - state when the node is in takeover mode.

    "any_takeover" - state when the node is any takeover state

    "takeover_autogiveback_scheduled" - state when the node is in takeover and
    autogiveback is scheduled

    "wfg" - state when the node is in "waiting for giveback"(wfg) prompt.

    "partial_giveback" - state when node has not sent all of the storage-failover
    aggregates with storage-failover policy to the partner.

    "sparecore" - state when partner selects sparecore style for dumping core.

    "dumpcore" - state when partner selects spraycore style for dumping core.

    "givewait" - state when the node is at "Press ctrl-C for maintainance mode".
    This is a transient state.  The node doesn't wait in this state at a prompt.

(For 7Mode) Possible state values are,

    "up" - state when the node is up with the OS booted.

    "takeover" - state when the node is in takeover mode.

    "wfg" - state when the node is in "waiting for giveback"(wfg) prompt.

    "sparecore" - state when partner selects sparecore style for dumping core.

    "dumpcore" - state when partner selects spraycore style for dumping core.

    "givewait" - state when the node is at "Press ctrl-C for maintainance mode".
    This is a transient state.  The node doesn't wait in this state at a prompt.

=item C<< partner_state=>$partner_state >>

(Optional) This is the state in which the second node in the storage-failover
pair is expected to be in.

Default value is "up".

(For CMode) Possible state values are,

    "up" - state when the node is up with the OS booted.

    "relocating_sfo_aggregates" - state when partner node starts relocating SFO
    aggregates in preparation of takeover

    "sfo_phase_done" - state when SFO aggregates are relocated.

    "takeover" - state when the node is in takeover mode.

    "any_takeover" - state when the node is any takeover state

    "takeover_autogiveback_scheduled" - state when the node is in takeover and
    autogiveback is scheduled

    "wfg" - state when the node is in "waiting for giveback"(wfg) prompt.

    "partial_giveback" - state when node has not sent all of the storage-failover
    aggregates with storage-failover policy to the partner.

    "sparecore" - state when partner selects sparecore style for dumping core.

    "dumpcore" - state when partner selects spraycore style for dumping core.

    "givewait" - state when the node is at "Press ctrl-C for maintainance mode".
    This is a transient state.  The node doesn't wait in this state at a prompt.

    "partner_inhibit_takeover" - Partner node halted after disabling takeover due to
    which takeover is not possible

    "failed_due_to_diskcheck" - giveback failed due to presence of failed disk in the system

    "failed_due_to_autosupport" - giveback failed due to asup generation

    "nocheck" - ignore the state of the node

(For 7Mode) Possible state values are,

    "up" - state when the node is up with the OS booted.

    "takeover" - state when the node is in takeover mode.

    "wfg" - state when the node is in "waiting for giveback"(wfg) prompt.

    "sparecore" - state when partner selects sparecore style for dumping core.

    "dumpcore" - state when partner selects spraycore style for dumping core.

    "givewait" - state when the node is at "Press ctrl-C for maintainance mode".
    This is a transient state.  The node doesn't wait in this state at a prompt.

=item C<< nacltask_poll_interval=>$poll_interval >>

(Optional) A value for the polling interval to verify the status of the storage failover.
The default value is 10 seconds.

=item C<< method-timeout=>$timeout >>

(Optional) A timeout value after which the TimeoutException will be thrown if the
status of the Storage Failover pair does not reach the desired values.

=item command_interface, apiset_must, apiset_should etc.

All of the other various options are described in
L<NACL::C::StorageFailover|lib-NACL-C-StorageFailover-pm>.

=back

=item Exceptions

=over

=item C<NACL::APISet::Exceptions::TimeoutException>

This type of exception is thrown when one or both the nodes in the SFO pair did not reach the desired state.

=back

=back

=cut

sub wait_for_state {

    my $pkg = shift;

    my %opts = $pkg->_common_validate_with(
        params             => \@_,
        only_static_method => 1,
        additional_spec    => {
            node          => { isa  => 'NACL::C::Node', type     => OBJECT },
            partner       => { isa  => 'NACL::C::Node', type     => OBJECT },
            node_state    => { type => SCALAR },
            partner_state => { type => SCALAR,          optional => 1 },
            nacltask_poll_interval => { type => SCALAR, optional => 1 }
        },
        ignore_primary_keys => 1,
    );

    my $time_to_wait  = delete $opts{'method-timeout'} || DEFAULT_TIMEOUT;

    my $node          = $opts{node};
    my $ver_mgr = $node->get_version_manager();
    my $build_type = $ver_mgr->get_version_attribute( attribute => "build_options" );  # Type of build Eg: x86_64.debug.gcov

    $Log->trace("build_type is : $build_type");
    $Log->trace("time_to_wait is : $time_to_wait");
    if ($build_type =~ /gcov/i ) {
        $time_to_wait = 6 * $time_to_wait;
        $Log->trace("time_to_wait is setting for gcov: $time_to_wait");
    } else {
        $Log->trace("time_to_wait is setting for non-gcov: $time_to_wait");
    }
    my $system_node_auto_support_modified = 0;
LOOP:{
      my $end_time      = time() + $time_to_wait;
      my $poll_interval = delete $opts{'nacltask_poll_interval'} || POLL_DELTA;
      my $exit_flag     = 0;
      while ( !$exit_flag ) {
          try {
              $pkg->check_state(%opts);
              $Log->debug(
                  "Both the nodes in the SFO pair reached the expected state");
              $exit_flag = 1;
              if ($system_node_auto_support_modified) {
                  # enable asup back, since it was disabled earlier
                  NACL::C::SystemNodeAutosupport->modify(
                      command_interface => $opts{command_interface},
                      node              => $node,
                      state             => "enable",
                      'local-collection'  => "true",
                  );
                  #reset system_node_auto_support_modified value
                  $system_node_auto_support_modified = 0;
              }
          } ## end try
          catch NACL::APISet::Exceptions::UnexpectedOutputException with {
              my $exception = shift;
              $Log->debug("Reached the Catch for Unexpected Exceptions ");
              if ( time() > $end_time ) {
                  $Log->debug( "Exception text from \'check_state\' "
                        . $exception->text() );
                  if ( $exception->text() =~ /Previous giveback failed in module: autosupport/i) {
                      $system_node_auto_support_modified = 1;

                      # disable asup and retry wait_for_state
                      NACL::C::SystemNodeAutosupport->modify(
                          command_interface => $opts{command_interface},
                          node              => $node,
                          state             => "disable",
                          'local-collection'  => "false",
                      );
                      sleep 5;
                      redo LOOP;
                  }
                  my $error_message
                      = "After waiting for $time_to_wait seconds "
                      . "one or both the nodes in the SFO pair did not reach the "
                      . "expected state.\n " . $exception->text();
                  $Log->exit() if $may_exit;
                  NACL::APISet::Exceptions::TimeoutException->throw(
                      $error_message);
              } ## end if ( time() > $end_time)

              Tharn::snooze $poll_interval ;
          };
      }      ## while loop ends here
     } ## end LOOP
} ## end sub wait_for_state

=head2 takeover

  NACL::STask::StorageFailover->takeover(
      command_interface     => $command_interface,
      node                  => $node_object,
      partner               => $partner_object,
      takeovertype          => $takeover_type,
      nacltask_verify       => $nacltask_verify,
      %other_options
  );


(Class method) Initiate takeover. This method provides
additional services that are inherited from
L<NACL::C::StorageFailover|lib-NACL-C-StorageFailover-pm>
to perform takeover with the various options provided.

This method performs the takeover on the specified node depending on the
flag takeovertype.

=over

=item Options

=over

=item C<< node=>$node_object >>

(Required) Node initiating takeover.
Specifies the node that is to take over its partner's storage.

=item C<< partner=>$partner_object >>

(Required) Node to takeover.
Specifies the node that is taken over. It is shut down and
its partner takes over its storage.

=item C<< takeovertype=>$takeover_type >>

(Required) Type of takeover to perform.
Possible values are "takeover_command", "panic", "reboot", "halt",
"system_powercycle", "system_power_off, "system_reset", "watchdog_reset".
Default is "takeover_command".

(For 7Mode) Presently the following option of takeover is supported -
takeover_command. The remaining types of takeover panic,reboot,halt,powercycle,
watchdog_reset will be implemented in the future.

=item C<< nacltask_inhibit_takeover => $inhibit >>

(Optional) ( For CMode )Boolean value, 0 or 1. Defaults to 0. If the value is set to 1,
then reboot command is sent with inhibit-takeover flag set to true.
(For 7Mode) the option is ignored

=item C<< nacltask_skip_lif_migration_before_action => $skip_lif_migration >>

(Optional) Option to indicate if skip-lif-migration should be performed. This
option obviates the need to uniquely specify the skip-lif-migration option that
is specific to a takeover type.

This option is only relevant for takeover types of halt, reboot, and takeover_command.
It will be ignored for all other takeover types.

=item C<< nacltask_boot_after_action => $boot_after_action >>

(Optional) Option to indicate if the partner node needs to be booted using
boot_ontap after the takeover has been initiated. Default is "0". If AUTOBOOT is
disabled, this means that the partner should be waiting at the Firmware prompt
when the takeover action completes.

=item C<< (Deprecated) nacltask_boot_after_halt => $boot_after_halt >>

This option has been deprecated in favor of nacltask_boot_after_action.

(Optional) Option to indicate if the partner node needs to be booted using
boot_ontap after issuing a halt. Default is "0" which means that the
partner will be waiting in the Firmware prompt

=item C<< (Deprecated) nacltask_boot_after_panic => $boot_after_panic >>

This option has been deprecated in favor of nacltask_boot_after_action.

(Optional) Option to indicate if the partner node needs to be booted using
boot_ontap after issuing a panic. Default is "1" which means that the
partner is booted using boot_ontap after it reaches Firmware Prompt after panic.

=item C<< nacltask_verify=>$nacltask_verify >>

(Optional) Boolean value, 0 or 1. Defaults to 0 (do not perform
verification). If 1, then check_state() will be invoked to
verify complete takeover scenario.

=item C<< nacltask_expect_takeover => $expect >>

(Optional) (For CMode) Boolean value, 0 or 1. Defaults to 1. Option to indicate if takeover
is expected to happen or not. Based on this option the states of the node and
its partner are set, when nacltask_verify option = 1. By default it takes the
value as 1 , indicating that takeover is expected to happen. If the value is
set to 0 , then the takeover is not expected to have occured. This option is
relevant when takeovertype is equal to panic . If the caller has set
the value of 'onpanic' flag to false in the filer then he needs to set this task
option nacltask_expect_takeover  to '0' .
(For 7Mode) the option is ignored .

=item C<< nacltask_check_partner_state => $check_partner_state >>

(Optional) Boolean value , 0 or 1 . Defaults to 1. Option to indicate whether
to check the state of the partner after takeover. The state of the partner is
verified if this flag and nacltask_verify are set to 1 .

=item C<< nacltask_powercycle_after_panic => $powercycle >>

(Optional) Boolean value, 0 or 1. Defaults to 0. If this option is set to 1,
the partner node will be powercycled after issuing panic command on the node.
When the value is 0, the routine ensures the partner reaches the firmware prompt
after panic.
The option is considered and valid if the takeovertype = 'panic'

(For 7Mode) the option is ignored

=item C<< nacltask_poll_interval=>$nacltask_poll_interval >>

(Optional) This is the interval (in seconds) after which the status of TO will
be polled. If not specified it defaults to 10 seconds.

=item C<< nacltask_retry_tries_count => $tries_count >>

(Optional, defaults to 3)
The number of times to retry the takeover command if it fails because
the LIF migration failed.

Note that it is possible to make it not retry the takeover command by setting
the value of this argument to 1.

=item C<< nacltask_retry_sleep_time => $sleep_time >>

(Optional, defaults to 180 seconds)
The time to sleep between retries of the takeover command.

=item C<< nacltask_system_power_off_wait => $nacltask_system_power_off_wait >>

(Optional, defaults to 180 seconds)
The time to sleep between system_power_off and syatem_power_on of the takeover command.

=item C<< 'method-timeout' => $time_to_wait >>

(Optional)
Time in seconds to wait for the completion of the Takeover operation

=item C<< nacltask_pre_takeover_callbacks => [ <coderef>, ... ] >>

(Optional) Ignored for 7-mode. The code references passed in will be called in
the order received right before takeover starts. The following arguments will be
passed to each callback: node, partner, command_interface, event_type

=item C<< nacltask_post_takeover_callbacks => [ <coderef>, ... ] >>

(Optional) Ignored for 7-mode. The code references passed in will be called in
the order received after takeover completes. The following arguments will be
passed to each callback: node, partner, command_interface, event_type.

=item command_interface, apiset_must, apiset_should etc.

All of the other various options are described in
L<NACL::C::StorageFailover|lib-NACL-C-StorageFailover-pm>.

=back

=item Assumptions

Takeover can occur in the following cases.

    a. takeover due to panic

    b. takeover due to reboot

    c. takeover due to halt

    d. takeover due to power cycle

    e. takeover due to watchdog reset

    f. takeover due to takeover command

    g. takeover due to system power off

    h. takeover due to system reset

=item Exceptions

=over

=item C<NACL::Exceptions::VerifyFailure>

This type of exception is thrown when verification fails for the complete of takeover scenarior.

=item C<NACL::Exceptions::InvalidChoice>

This type of exception is thrown when an invalid value is passed for takeover type.

=back

=back

=cut

sub takeover {
    $Log->enter() if $may_enter;
    my (
        %apiset_must_should, $build_type, %opts, $pkg, %takeover_opts, $takeover_started,
    );
    $pkg = shift;

    %opts = $pkg->_common_validate_with(
        params             => \@_,
        only_static_method => 1,
        additional_spec    => {
            node    => { isa => 'NACL::C::Node', type => OBJECT },
            partner => { isa => 'NACL::C::Node', type => OBJECT },
            takeovertype => { type => SCALAR, default => 'takeover_command' },
            option       => { type => SCALAR, default => 'normal' },
            'bypass-optimization' => { type => SCALAR, optional => 1 },
            nacltask_verify =>
                { type => SCALAR | UNDEF, optional => 1, default => 0 },
            nacltask_poll_interval    => { type => SCALAR,  optional => 1 },
            nacltask_inhibit_takeover => { type => BOOLEAN, default  => 0 },
            nacltask_expect_takeover  => { type => BOOLEAN, default  => 1 },
            nacltask_check_partner_state => { type => BOOLEAN, default => 1 },
            nacltask_powercycle_after_panic =>
                { type => BOOLEAN, default => 0 },
            nacltask_boot_after_action => {type => BOOLEAN, default => 0 },
            nacltask_retry_tries_count => { type => SCALAR, default => 3 },
            nacltask_retry_sleep_time  => { type => SCALAR, default => 180 },
            nacltask_system_power_off_wait => { type => SCALAR, default => 60 },
            nacltask_pre_takeover_callbacks  => { type => ARRAYREF, optional => 1},
            nacltask_post_takeover_callbacks => { type => ARRAYREF, optional => 1},
        },
        allow_extra => 1
    );

    my $wait_time = delete $opts{'method-timeout'} || TAKEOVER_TIMEOUT;
    my $poll_interval = delete $opts{'nacltask_poll_interval'} || POLL_DELTA;
    my $inhibit       = delete $opts{'nacltask_inhibit_takeover'};
    my $to_expected   = delete $opts{'nacltask_expect_takeover'};
    my $check_partner_state = delete $opts{'nacltask_check_partner_state'};
    my $powercycle          = delete $opts{'nacltask_powercycle_after_panic'};
    my $nacltask_system_power_off_wait = $opts{'nacltask_system_power_off_wait'};

    $takeover_opts{boot_after_action}                =
        delete($opts{nacltask_boot_after_action});
    $takeover_opts{skip_lif_migration_before_action} =
        delete($opts{nacltask_skip_lif_migration_before_action});

    # These could be deprecated in the when 9.4 branches off of devN
    DEPRECATED_OPTS: {
        my (@boot_after_options, @skip_lif_options, );
        @skip_lif_options = qw(
            skip-lif-migration
            skip-lif-migration-before-reboot
            skip-lif-migration-before-shutdown
            skip-lif-migration-before-takeover
        );
        foreach my $skip_lif (@skip_lif_options) {
            if (defined($opts{$skip_lif})) {
                $Log->warn(
                    "The parameter '$skip_lif' is being deprecated. " .
                    'Please use \'nacltask_skip_lif_migration_before_action\'.'
                );
                $takeover_opts{skip_lif_migration_before_action} = delete($opts{$skip_lif});
            }
        } # End foreach (@skip_lif_options)

        @boot_after_options = qw(
            nacltask_boot_after_halt
            nacltask_boot_after_panic
        );
        foreach my $boot_after (@boot_after_options) {
            if (defined($opts{$boot_after})) {
                $Log->warn(
                    "The parameter '$boot_after' is being deprecated. Please " .
                    'use \'nacltask_boot_after_action\'.'
                );
                $takeover_opts{boot_after_action} = delete($opts{$boot_after});
            }
        } # End foreach (@boot_after_options)
    } # End DEPRECATED_OPTS

    # Check to see if we're on a GCOV build. If so, increase 'method-timeout'
    # to accomodate increased system shutdown times for coverage dumps.
    $build_type = $opts{node}->get_version_manager()->get_version_attribute(
        attribute => 'build_options',
    );
    $Log->trace(
        "build_type is     : $build_type\n" .
        "method-timeout is : $wait_time"
    );
    if ($build_type =~ /gcov/i ) {
        $wait_time *= 6;
        $Log->trace("method-timeout updated for gcov to : $wait_time");
    } else {
        $Log->trace("method-timeout not updated for non-gcov: $wait_time");
    }

    # Copy apiset must and should so we can pass it down
    $pkg->_hash_copy(
        source => \%opts,
        target => \%apiset_must_should,
        copy   => [qw(apiset_must apiset_should)]
    );

    my $takeover_type         = $opts{takeovertype};
    my $to_node               = $opts{node};
    my $partner_node          = $opts{partner};
    my $verify_required       = $opts{nacltask_verify};
    my $ci                    = $opts{command_interface};
    my $option                = $opts{option};
    my $CS_package            = $pkg->get_CS_package_name();
    my $mode                  = $ci->mode();    # 7 or C mode
    my $partner_node_name     = $partner_node->node();
    my $task_verify_completed = 0;
    $takeover_started         = 0;

    my $pre_takeover_callbacks = delete $opts{nacltask_pre_takeover_callbacks};
    my $post_takeover_callbacks = delete $opts{nacltask_post_takeover_callbacks};

    for my $callbacks ($pre_takeover_callbacks, $post_takeover_callbacks){
        next if ( !$callbacks );
        for my $c ( @$callbacks ){
            my $ref_c = ref($c);
            if ( !$ref_c eq 'CODE'){
                $Log->exit;
                NATE::Exceptions::Argument->throw(
                    "Each callback specified in nacltask_pre_takeover_callbacks"
                    ." or nacltask_post_takeover_callbacks must be a code"
                    ." reference and not a '$ref_c'"
                );
            }
        }
    }

    $Log->debug("Mode is '$mode'");
    if ( $mode eq '7Mode' ) {
        $pkg->_takeover_7Mode(
            %opts,
            'method-timeout'    => $wait_time,
            poll_interval       => $poll_interval,
            check_partner_state => $check_partner_state,
        );
    } elsif ( $mode eq 'CMode' ) {
        my (
            $auto_giveback, $auto_giveback_after_panic, $expected_partner_state,
            $expected_takeover_state,
        );
        # C Mode Support

        # If nacltask_verify option is set, then set the expected
        # states of the nodes here
        if ($verify_required) {
            # set the takeover state based on the autogiveback option
            my ($state, );
            $state = $CS_package->fetch(
                command_interface => $ci,
                filter            => { node => $to_node },
                requested_fields  => [qw(auto-giveback auto-giveback-after-panic)]
            );
            $auto_giveback_after_panic = $state->auto_giveback_after_panic();
            $auto_giveback = $state->auto_giveback();
            if ( $auto_giveback eq 'false' ) {
                $expected_takeover_state = 'takeover';
            } else {
                $expected_takeover_state = 'takeover_autogiveback_scheduled';
            }
            $expected_partner_state = 'wfg';
        }    ## if ( $verify_required )

        # Check for type of takeover command, then Initiate the takeover
        # Initiate the takeover due to takeover command - Cmode-CLI Command
        # Print date and time before initiating takeover for debugging purpose
        DATE_PRINT: {
            my ($system_node_date, );
            $system_node_date = NACL::CS::ClusterDate->fetch(
                'command_interface' => $ci,
                'requested_fields'  => [qw(dateandtime timezone)]
            );
            $Log->comment('Date     : ' . $system_node_date->dateandtime());
            $Log->comment('TimeZone : ' . $system_node_date->timezone());
        } # End DATE_PRINT

        my $is_AWS_filer = 0;
        my $filer_model;
        try {
            # This will fail if node is not running, but is needed for HA objects lib
            $filer_model = $ci->get_one_state_attribute('model');
        } catch NATE::BaseException with {
            # Catching any other issues to avoid failing scripts.
            # One scenario is when the systems are not online or are in the loader prompt
            $filer_model = $ci->{hostrec}->{MODEL};
            if (!defined($filer_model) && keys( %{ $ci->{hostrec} } ) == 0) {
                # if MODEL doesn't exist it will create an empty hash
                # statement needed to avoid autovivification
                $Log->comment('Could not determine filer model, returning non AWS model type');
                delete $ci->{hostrec};
                $filer_model = 'FAS';
            }
        };

        my $conn_id = 'sp_login';
        # assuming AURORA testbeds do share common model: 'AFF-A700s'
        $conn_id='bmc_login' if($filer_model eq 'AFF-A700s');

        if ( ( defined $filer_model && $filer_model =~ AWS_MODEL_REGEX ) || ( !defined $filer_model && $pkg->_is_cloud($ci) )){
            $is_AWS_filer = 1;
            $Log->comment(
                'Model Number is of an AWS filer and will take an alternate' .
                ' code path than typical FAS '
            );
            $Log->comment("Platform model: $filer_model") if (defined($filer_model));
        }

        if ($is_AWS_filer && $takeover_type ne 'takeover_command') {
            $pkg->_takeover_AWS(
                %opts,
                wait_time                          => $wait_time,
                poll_interval                      => $poll_interval,
                check_partner_state                => $check_partner_state,
                inhibit                            => $inhibit,
                skip_lif_migration_before_reboot   => $takeover_opts{skip_lif_migration_before_action},
                skip_lif_migration_before_shutdown => $takeover_opts{skip_lif_migration_before_action},
                to_expected                        => $to_expected,
                powercycle                         => $powercycle,
                boot_after_halt                    => $takeover_opts{boot_after_action},
                expected_takeover_state            => $expected_takeover_state,
                expected_partner_state             => $expected_partner_state,
                boot_after_panic                   => $takeover_opts{boot_after_action},
                auto_giveback                      => $auto_giveback,
                auto_giveback_after_panic          => $auto_giveback_after_panic,
                apiset_must_should                 => \%apiset_must_should,
                inhibit                            => $inhibit,
            );
            # Verification was done in _takeover_AWS for separately handled
            # takeover types. So, no verification is needed which is done at the
            # end of this method.
            return;
        } # End if(AWS && $takeover_type ne 'takeover_command')

        if ( $takeover_type eq 'takeover_command' ) {

            # Check for callbacks to be executed before takeover
            if ($to_expected && $pre_takeover_callbacks && @$pre_takeover_callbacks) {
                $pkg->_execute_event_callbacks(
                    callbacks         => $pre_takeover_callbacks,
                    command_interface => $ci,
                    takeover_node     => $to_node,
                    partner_node      => $partner_node,
                    event_type        => $takeover_type,
                );
            }

            # Invoke storage failover takeover command to initiate takeover
            # partner node (Surrender node) goes down once takeover is initiated
            # This command should return soon after takeover is initiated
            # There is seperate method to be called to verify the takeover state
            my (%retry_opts, %to_opts, );
            $pkg->_move_retry_params(
                source => \%opts,
                target => \%retry_opts
            );

            # Options to NACL::C::StorageFailover. Must do like this as 'undef'
            # is not valid to pass. This will either set options if they were
            # passed in to us, or not set the key otherwise.
            $pkg->_hash_move(
                source => \%opts,
                target => \%to_opts,
                move   => [qw(bypass-optimization allow-disk-inventory-mismatch)],
            );
            if (defined($takeover_opts{skip_lif_migration_before_action})) {
                $to_opts{'skip-lif-migration-before-takeover'} =
                    $takeover_opts{skip_lif_migration_before_action};
            }

            nacl_method_retry(
                code        => sub {
                    $pkg->SUPER::takeover(
                        command_interface => $ci,
                        bynode            => $to_node,
                        option            => $option,
                        %apiset_must_should,
                        %to_opts,
                    );
                },
                exceptions  => 'NACL::C::Exceptions::StorageFailover::SkipLifMigrationBeforeTakeover',
                %retry_opts,
            );

            $takeover_started = 1;
        } # End if ($takeover_type eq 'takeover_command')

        # Check if the takeover is due to PANIC, then panic the partner node
        if ( $takeover_type eq 'panic' ) {

            # Check for callbacks to be executed before takeover
            if ($to_expected && $pre_takeover_callbacks && @$pre_takeover_callbacks) {
                $pkg->_execute_event_callbacks(
                    callbacks         => $pre_takeover_callbacks,
                    command_interface => $ci,
                    takeover_node     => $to_node,
                    partner_node      => $partner_node,
                    event_type        => $takeover_type,
                );
            }

            $Log->comment('Initiating Takeover on panic(Induced panics) ');
            my ($autoboot_val, );
            $autoboot_val = $partner_node->get_7m_or_nodescope_apiset()->
                bootargs_get(arg => 'AUTOBOOT')->get_raw_output();

            NACL::STask::Node->panic(
                command_interface => $partner_node,
                nacltask_reboot   => 0
            );
            if ($powercycle) {
                # support added if nacltask_powercycle_after_panic=1
                my $api_set_obj = $partner_node->apiset(
                    connid    => $conn_id,
                    set       => 'SP',
                    interface => 'CLI',
                );
                $pkg->check_state(
                    command_interface => $ci,
                    node              => $partner_node,
                    partner           => $to_node,
                    node_state        => 'sparecore',
                    'method-timeout'  => $wait_time,
                );
                #Added 10 seconds delay to make sure the dump is in progress.
                Tharn::sleep TEN_SEC_TIMEOUT;

                $Log->debug('Creating SP connection handle');
                #Added prompts to handle Razor platforms
                my @prompts_answers
                    = (
                    '.*This will cause a dirty shutdown of your appliance.*' => 'y', '.*This command will trigger a destage of ONTAP and will take some time to complete after which the system will restart. Use "system reset" or "system reset current" to avoid the delay.*' => 'y');
                $api_set_obj->system_power(
                    cycle                    => 1,
                    'connectrec-match_table' => \@prompts_answers,
                );

            }
            # Need to boot the partner now.
            if ( $autoboot_val !~ /AUTOBOOT == true/i ){
                my ($transit_obj, );
                $transit_obj = NACL::Transit->new( name => $partner_node_name );
                $transit_obj->wait_for_state(
                    wait_for               => 'FIRMWARE',
                    wait_for_state_timeout => $wait_time,
                );
                if ($takeover_opts{boot_after_action}) {
                    $transit_obj->{conn}->send("boot_ontap\r");
                }
            }

            if ($verify_required) {
                $Log->comment('In \'panic\' private verify routine.');
                my %call_opts;
                if ($to_expected) {

                    # For this one case calling check_for_state() is not suitable
                    # The state of the takenover node is highly transient when it changes
                    # from wfg to booting, because autogiveback is already deferred
                    # in the other node.
                    # Hence checking the state of the takeover node using wait_for_state
                    # Then calling transit_obj->change_state to capture the state of
                    # the partner if the caller wants to verify the state of partner too.

                    # If auto-giveback-after-panic is set to true
                    # Auto-giveback will be scheduled irrespective of option
                    # cf.giveback.auto.enable in case of takeover due to panic

                    $expected_takeover_state
                        = 'takeover_autogiveback_scheduled' if($auto_giveback_after_panic eq 'true');

                    %call_opts = (
                        command_interface      => $ci,
                        node                   => $to_node,
                        partner                => $partner_node,
                        node_state             => $expected_takeover_state,
                        'method-timeout'       => $wait_time,
                        nacltask_poll_interval => $poll_interval,
                        %apiset_must_should,
                    );
                    $pkg->wait_for_state(%call_opts);

                    if ( $check_partner_state && !$takeover_opts{boot_after_action} ) {
                        $Log->debug(
                            "The option boot_after_action was provided as 0.Hence the Partner Node is in \n"
                                . "FirmWare Prompt after panic.The option check_partner_state is provided \n "
                                . "as 1. Skipping another verification for partner state, since already \n"
                                . 'verified that the partner is in FirmWare State using transit_obj->change_state.'
                        );
                    } elsif ($check_partner_state) {
                        my ($transit_obj, );
                        $transit_obj = NACL::Transit->new( name => $partner_node_name );
                        $transit_obj->change_state(
                            to      => 'TAKEN_OVER_WAITING_ONTAP',
                            timeout => $wait_time,
                        );
                    }
                    # Verification for takeovertype=panic and takeover_expected
                    # is completed in the previous codeblock.
                    # Setting the flag task_verify_completed to 1 so that
                    # no verification is needed in the end of this method again.
                    $task_verify_completed = 1;
                } else {
                    # This can be handled in the general verification block
                    $expected_takeover_state = 'up';
                    $expected_partner_state  = 'up';
                }
            }    # if ( $verify_required )
            $takeover_started = 1;
        } # End if ($takeover_type eq 'panic')

        # Check if the takeover is due to REBOOT then reboot the partner node
        if ( $takeover_type eq 'reboot' ) {
            my ($autoboot_val, %reboot_opts, );
            $expected_takeover_state = 'takeover_autogiveback_scheduled';
            $Log->comment('Initiating Takeover on reboot command');
            $autoboot_val = $partner_node->get_7m_or_nodescope_apiset()->
                bootargs_get(arg => 'AUTOBOOT')->get_raw_output();

            # build options for reboot command
            %reboot_opts = (
                'ignore-quorum-warnings'           => $opts{command_interface}->
                    version_manager()->is_version_ge(release => '8.3.1') ? 'true' : 'false',
                'inhibit-takeover'                 => $inhibit ? 'true' : 'false',
                'node'                             => 'local',
                'skip-lif-migration-before-reboot' =>
                    $takeover_opts{skip_lif_migration_before_action} ? 'true' : 'false',
            );
            try {
                $partner_node->apiset()->system_node_reboot(%reboot_opts);
            } catch NATE::BaseException with {
                # We expect the connection to potentially drop after the reboot
                # command is issued. If we catch an exception, but that's not
                # the reason, rethrow it.
                my ($ex, );
                $ex = shift;
                if ($ex->text() !~ /Unexpected EOF/) {
                    $ex->throw();
                }
            };

            if ($verify_required && (!$to_expected || $inhibit)) {
                $expected_takeover_state = 'up';
                $expected_partner_state  = 'up';
            }    ## if ($verify_required)

            # Check for callbacks to be executed before takeover
            if ($to_expected && $pre_takeover_callbacks && @$pre_takeover_callbacks) {
                $pkg->_execute_event_callbacks(
                    callbacks         => $pre_takeover_callbacks,
                    command_interface => $ci,
                    takeover_node     => $to_node,
                    partner_node      => $partner_node,
                    event_type        => $takeover_type,
                );
            }

            # If AUTOBOOT is not true, need to boot node.
            if ( $autoboot_val !~ /AUTOBOOT == true/i ){
                my ($transit_obj, );
                $transit_obj = NACL::Transit->new( name => $partner_node_name );
                $transit_obj->wait_for_state(
                    wait_for               => 'FIRMWARE',
                    wait_for_state_timeout => $wait_time,
                );
                $transit_obj->{conn}->send("boot_ontap\r");
            }

            $Log->debug(' takeover via reboot completed ');
            $takeover_started = 1;
        } # End if ($takeover_type eq 'reboot')

        # Check if takeover is due to WATCHDOG_RESET then issue the watchdog-reset command on partner node
        if ( $takeover_type eq 'watchdog_reset' ) {
            my ($apiset_sp, $autoboot_val, $transit_obj, $transit_to_state, );
            $Log->comment('Initiating HW-Assisted Takeover due to watchdog-reset');
            $autoboot_val = $partner_node->get_7m_or_nodescope_apiset()->
                bootargs_get(arg   => 'AUTOBOOT')->get_raw_output();
            $apiset_sp = $partner_node->apiset(
                interface => 'CLI',
                set       => 'Systemshell',
            );

            $Log->debug('Creating System connection handle');
            $apiset_sp->watchdog(
                'reset'           => 1,
                'privilege-level' => 'root',
            );

            $transit_obj = NACL::Transit->new( name => $partner_node_name );

            # Check for callbacks to be executed before takeover
            if ($to_expected && $pre_takeover_callbacks && @$pre_takeover_callbacks) {
                $pkg->_execute_event_callbacks(
                    callbacks         => $pre_takeover_callbacks,
                    command_interface => $ci,
                    takeover_node     => $to_node,
                    partner_node      => $partner_node,
                    event_type        => $takeover_type,
                );
            }

            if ( $autoboot_val =~ /AUTOBOOT == true/i ){
                $transit_to_state = 'TAKEN_OVER_WAITING_ONTAP';
            }else{
                $transit_to_state = 'FIRMWARE';
            }
            try {
                $transit_obj->wait_for_state(
                    wait_for               => $transit_to_state,
                    wait_for_state_timeout => $wait_time,
                );
            } catch NACL::Transit::Exceptions::Timeout with {
                my ($exception, $new_ex_text, );
                $exception = shift;
                $new_ex_text = '\'watchdog reset\' failed to execute- '.$exception->text();
                $Log->exit();
                NACL::Transit::Exceptions::Timeout->throw($new_ex_text);
            };
            $transit_obj->{conn}->send("boot_ontap\r") if ($transit_to_state eq 'FIRMWARE');

            $Log->debug(' takeover via watchdog reset completed ');
            $takeover_started = 1;
        } # End if($takeover_type eq 'watchdog_reset')

        # Check if takeover is due to HALT Command then issue the halt command on partner node
        if ( $takeover_type eq 'halt' ) {
            my ($halt_cmd, $halt_opts, $transit_obj, $transit_to_state, );
            $Log->comment('Initiating Takeover on halt command');
            $transit_obj = NACL::Transit->new( name => $partner_node_name, );

            # build options for halt command
            $halt_opts=q();
            $halt_opts .= ' -skip-lif-migration-before-shutdown true'
                if ($takeover_opts{skip_lif_migration_before_action});
            $halt_opts .= ' -ignore-quorum-warnings true'
                if ($opts{command_interface}->version_manager()->is_version_ge(release => '8.3.1'));

            if ($inhibit) {
                $halt_cmd
                    = "set -confirmations off ; system node halt -node local -inhibit-takeover true $halt_opts";
                $transit_to_state = 'FIRMWARE';
                $transit_obj->change_state(
                    to      => $transit_to_state,
                    haltcmd => $halt_cmd,
                    timeout => $wait_time,
                );
                $expected_takeover_state = 'partner_inhibit_takeover';
                $expected_partner_state  = 'nocheck';

                if ( $takeover_opts{boot_after_action} == 1 ) {
                    $transit_obj->{conn}->send("boot_ontap\r");
                    $expected_takeover_state = 'up';
                    $expected_partner_state  = 'up';
                }
            } else {

                # Check for callbacks to be executed before takeover
                if ($to_expected && $pre_takeover_callbacks && @$pre_takeover_callbacks) {
                    $pkg->_execute_event_callbacks(
                        callbacks         => $pre_takeover_callbacks,
                        command_interface => $ci,
                        takeover_node     => $to_node,
                        partner_node      => $partner_node,
                        event_type        => $takeover_type,
                    );
                }

                $halt_cmd
                    = "set -confirmations off ; system node halt -node local $halt_opts";
                $transit_to_state = 'FIRMWARE';
                $transit_obj->change_state(
                    to      => $transit_to_state,
                    haltcmd => $halt_cmd,
                    timeout => $wait_time,
                );
                $expected_partner_state = 'nocheck';

                if ( $takeover_opts{boot_after_action} == 1 ) {
                    $transit_obj->{conn}->send("boot_ontap\r");
                    $expected_partner_state = 'wfg';
                }
            }

            $Log->comment(' takeover via halt completed ');
            $takeover_started = 1;
        } # End if($takeover_type eq 'halt')

        # Check if takeover is due to SYSTEM_POWERCYCLE then issue system-power cycle command from RLM/SP on partner node
        if ( $takeover_type eq 'system_powercycle' ) {
            my (
                $apiset_sp, $autoboot_val, @prompts_answers, $transit_obj,
                $transit_to_state,
            );
            $Log->comment('Initiating Takeover due to System-power cycle');

            $autoboot_val = $partner_node->get_7m_or_nodescope_apiset()->bootargs_get(
                arg   => 'AUTOBOOT',
            )->get_raw_output();
            $transit_obj  = NACL::Transit->new( name => $partner_node_name );

            $Log->debug('Creating SP connection handle');
            $apiset_sp = $partner_node->apiset(
                connid    => $conn_id,
                set       => 'SP',
                interface => 'CLI',
            );

            # Check for callbacks to be executed before takeover
            if ($to_expected && $pre_takeover_callbacks && @$pre_takeover_callbacks) {
                $pkg->_execute_event_callbacks(
                    callbacks         => $pre_takeover_callbacks,
                    command_interface => $ci,
                    takeover_node     => $to_node,
                    partner_node      => $partner_node,
                    event_type        => $takeover_type,
                );
            }

            @prompts_answers = (
                '.*This will cause a dirty shutdown of your appliance.*' => 'y',
                '.*This command will trigger a destage of ONTAP and will take some time to complete after which the system will restart. Use "system reset" or "system reset current" to avoid the delay.*' => 'y'
            );
            $apiset_sp->system_power(
                cycle                    => 1,
                'connectrec-match_table' => \@prompts_answers,
            );

            if ( $autoboot_val =~ /AUTOBOOT == true/i ){
                $transit_to_state = 'TAKEN_OVER_WAITING_ONTAP';
            } else {
                $transit_to_state = 'FIRMWARE';
            }
            try{
                $transit_obj->wait_for_state(
                    wait_for               => $transit_to_state,
                    wait_for_state_timeout => $wait_time,
                );
            } catch NACL::Transit::Exceptions::Timeout with {
                my ($exception, $new_ex_text, );
                $exception   = shift;
                $new_ex_text = '\'system power cycle\' command could not bring node down- '.$exception->text();
                $Log->exit() if $may_exit;
                NACL::Transit::Exceptions::Timeout->throw($new_ex_text);
            };
            $transit_obj->{conn}->send("boot_ontap\r") if ($transit_to_state eq 'FIRMWARE');

            $Log->comment(' takeover via system_powercycle completed ');
            $takeover_started = 1;
        } # End if($takeover_type eq 'system_powercycle)

        # Check if takeover is due to SYSTEM_POWER_OFF then issue system-power off command from RLM/SP on partner node
        if ( $takeover_type eq 'system_power_off' ) {
            my (
                $apiset_sp, $autoboot_val, @prompts_answers, $transit_obj,
                $transit_to_state,
            );
            $Log->comment('Initiating Takeover due to System-power off');
            $autoboot_val = $partner_node->get_7m_or_nodescope_apiset()->
                bootargs_get(arg => 'AUTOBOOT')->get_raw_output();
            $transit_obj  = NACL::Transit->new( name => $partner_node_name );

            $Log->debug('Creating SP connection handle');
            $apiset_sp = $partner_node->apiset(
                connid    => $conn_id,
                set       => 'SP',
                interface => 'CLI',
            );

            # Check for callbacks to be executed before takeover
            if ($to_expected && $pre_takeover_callbacks && @$pre_takeover_callbacks) {
                $pkg->_execute_event_callbacks(
                    callbacks         => $pre_takeover_callbacks,
                    command_interface => $ci,
                    takeover_node     => $to_node,
                    partner_node      => $partner_node,
                    event_type        => $takeover_type,
                );
            }

            @prompts_answers = (
                '.*This will cause a dirty shutdown of your appliance.*' => 'y',
                '.*This command will trigger a destage of ONTAP and will take some time to complete.*' => 'y',
            );
            $apiset_sp->system_power(
                off                      => 1,
                'connectrec-match_table' => \@prompts_answers,
            );
            $Log->comment( 'Sleeping for ' . $nacltask_system_power_off_wait . ' seconds');
            Tharn::sleep($nacltask_system_power_off_wait);
            $apiset_sp->system_power( on => 1, );

            if ( $autoboot_val =~ /AUTOBOOT == true/i ){
                $transit_to_state = 'TAKEN_OVER_WAITING_ONTAP';
            } else {
                $transit_to_state = 'FIRMWARE';
            }
            try{
                $transit_obj->wait_for_state(
                    wait_for               => $transit_to_state,
                    wait_for_state_timeout => $wait_time,
                );
            } catch NACL::Transit::Exceptions::Timeout with {
                my ($exception, $new_ex_text, );
                $exception   = shift;
                $new_ex_text = '\'system power off\' command could not bring node down- '.$exception->text();
                $Log->exit() if $may_exit;
                NACL::Transit::Exceptions::Timeout->throw($new_ex_text);
            };
            $transit_obj->{conn}->send("boot_ontap\r") if ($transit_to_state eq 'FIRMWARE');

            $Log->comment(' takeover via system_power_off completed ');
            $takeover_started = 1;
        } # End if($takeover_type eq 'system_power_off)

        # Check if takeover is due to SYSTEM_RESET then issue system reset command from RLM/SP on partner node
        if ( $takeover_type eq 'system_reset' ) {
            my (
                $apiset_sp, $autoboot_val, @prompts_answers, $transit_obj,
                $transit_to_state,
            );
            $Log->comment('Initiating Takeover due to System reset');
            $autoboot_val = $partner_node->get_7m_or_nodescope_apiset()->
                bootargs_get(arg => 'AUTOBOOT')->get_raw_output();
            $transit_obj  = NACL::Transit->new( name => $partner_node_name );

            $Log->debug('Creating SP connection handle');
            $apiset_sp = $partner_node->apiset(
                connid    => $conn_id,
                set       => 'SP',
                interface => 'CLI',
            );

            # Check for callbacks to be executed before takeover
            if ($to_expected && $pre_takeover_callbacks && @$pre_takeover_callbacks) {
                $pkg->_execute_event_callbacks(
                    callbacks         => $pre_takeover_callbacks,
                    command_interface => $ci,
                    takeover_node     => $to_node,
                    partner_node      => $partner_node,
                    event_type        => $takeover_type,
                );
            }

            @prompts_answers = (
                '.*This will cause a dirty shutdown of your appliance.*' => 'y',
                '.*This command will trigger a destage of ONTAP and will take some time to complete.*' => 'y',
            );
            $apiset_sp->system_reset(
                'connectrec-match_table' => \@prompts_answers, );

            if ( $autoboot_val =~ /AUTOBOOT == true/i ){
                $transit_to_state = 'TAKEN_OVER_WAITING_ONTAP';
            } else {
                $transit_to_state = 'FIRMWARE';
            }
            try{
                $transit_obj->wait_for_state(
                    wait_for               => $transit_to_state,
                    wait_for_state_timeout => $wait_time,
                );
            } catch NACL::Transit::Exceptions::Timeout with {
                my ($exception, $new_ex_text, );
                $exception   = shift;
                $new_ex_text = '\'system reset\' command could not bring node down- '.$exception->text();
                $Log->exit() if $may_exit;
                NACL::Transit::Exceptions::Timeout->throw($new_ex_text);
            };
            $transit_obj->{conn}->send("boot_ontap\r") if ($transit_to_state eq 'FIRMWARE');

            $Log->debug(' takeover via system_reset completed ');
            $takeover_started = 1;
        } # End of ugly if(takeovertype)/elsif/elsif/ad nauseum

        if (!$takeover_started) {
            # Unknown option for takeover_type
            my ($err_msg, );
            $err_msg = "Unknown Option for takeover_type: $takeover_type";
            $Log->exit() if $may_exit;
            NACL::Exceptions::InvalidChoice->throw($err_msg);
        }

        # Check if Takeover Verification is required, if so then check takeover state
        if ( $verify_required && !$task_verify_completed ) {
            $Log->comment(' Executing the nacltask verify condition ');

            my %call_opts = (
                command_interface      => $ci,
                node                   => $to_node,
                partner                => $partner_node,
                node_state             => $expected_takeover_state,
                'method-timeout'       => $wait_time,
                nacltask_poll_interval => $poll_interval,
                %apiset_must_should,
            );
            if ($check_partner_state) {
                $call_opts{'partner_state'} = $expected_partner_state;
            }
            $pkg->wait_for_state(%call_opts);
        } ## end if ($verify_required)

        # Check for callbacks to be executed after takeover command
        if (   $verify_required && !$inhibit && $to_expected
            && $post_takeover_callbacks
            && @$post_takeover_callbacks) {
            $pkg->_execute_event_callbacks(
                callbacks         => $post_takeover_callbacks,
                command_interface => $ci,
                takeover_node     => $to_node,
                partner_node      => $partner_node,
                event_type        => $takeover_type
            );
        }

    } else {
        # Not a valid mode option
        $Log->exit() if $may_exit;
        NACL::Exceptions::InvalidChoice->throw("Mode ($mode) is not supported.");
    } ## end else [ if ( $mode eq '7Mode' )
    $Log->exit() if $may_exit;
    return;
} # End takeover()

################################################################################
# Function     : _is_cloud()
# Description  :
# Arguments    : ??
# Returns      : ??
################################################################################
sub _is_cloud {
    my ($pkg , $node_obj ) = @_;
    my $is_cloud = 0;

    try {
        NACL::C::SystemNode->find(
            command_interface => $node_obj,
            filter => {model => 'CDvM*00'},
        );
        $is_cloud = 1;
    } catch NACL::Exceptions::NoElementsFound with {
    };
    return $is_cloud;
} # End _is_cloud()

################################################################################
# Function     : _takeover_7Mode()
# Description  : Implementation of 7Mode takeover
# Arguments    : ??
# Returns      : None
################################################################################
sub _takeover_7Mode {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = @_;

    my ($takeover_type,           $verify_required,
        $expected_takeover_state, $expected_partner_state,
        $ci,                      %apiset_rules,
        $to_node,                 $partner_node,
        $wait_time,               $poll_interval,
        $check_partner_state
    );

    $ci                  = $opts{command_interface};
    $takeover_type       = $opts{takeovertype};
    $to_node             = $opts{node};
    $partner_node        = $opts{partner};
    $verify_required     = $opts{nacltask_verify};
    $wait_time           = $opts{'method-timeout'};
    $poll_interval       = $opts{'poll_interval'};
    $check_partner_state = $opts{'check_partner_state'};
    # copy common apiset rules
    $pkg->_copy_common_component_params(
        source => \%opts,
        target => \%apiset_rules
    );

    # 7mode support for panic,reboot,halt,system_powercycle,watchdog_reset
    # does not exist
    if (   $takeover_type eq 'panic'
        || $takeover_type eq 'reboot'
        || $takeover_type eq 'halt'
        || $takeover_type eq 'system_powercycle'
        || $takeover_type eq 'watchdog_reset' ) {
        my $invalid_choice_msg
            = 'takeover type '
            . "$takeover_type is not supported in "
            . '7Mode Takeover ';
        $Log->exit() if $may_exit;
        NACL::Exceptions::InvalidChoice->throw($invalid_choice_msg);
    }

    # If user wants to verify the states of the node at the end of takeover
    # then set the values of expected state of local node and partner node
    if ($verify_required) {
        $expected_takeover_state = 'takeover';
        $expected_partner_state  = 'wfg';
    }
    if ( $takeover_type eq 'takeover_command' ) {

        # Invoke component storage failover takeover command  to initiate the takeover
        # partner node (Surrender node) goes down once the takeover is being initiated

        # This command should return soon after takeover is initiated
        # There is seperate method to be called to verify the takeover state

        $pkg->SUPER::takeover(
            command_interface => $ci,
            bynode            => $to_node,
            option            => 'normal',
            %apiset_rules
        );
    } else {
        $Log->exit() if $may_exit;
        NACL::Exceptions::InvalidChoice->throw(
            "Unknown option $takeover_type passed to the takeover method"
        );
    }

    # If user wants to verify the states of the node
    # verify the same using wait_for_state
    if ($verify_required) {
        $Log->comment(' Executing the nacltask verify condition ');

        my %call_opts = (
            command_interface      => $ci,
            node                   => $to_node,
            partner                => $partner_node,
            node_state             => $expected_takeover_state,
            'method-timeout'       => $wait_time,
            nacltask_poll_interval => $poll_interval,
            %apiset_rules,
        );

        # Verify the state of the partner too if the caller has set the flag
        if ($check_partner_state) {
            $call_opts{'partner_state'} = $expected_partner_state;
        }
        $pkg->wait_for_state(%call_opts);
    } ## end if ($verify_required)
    return;
} ## end sub _takeover_7Mode

################################################################################
# Function     : _takeover_AWS()
# Description  : Implementation of amazon web services takeover
# Arguments    : ??
# Returns      : None
################################################################################
sub _takeover_AWS {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = @_;

    my $ci                        = $opts{command_interface};
    my $takeover_type             = $opts{takeovertype};
    my $to_node                   = $opts{node};
    my $partner_node              = $opts{partner};
    my $verify_required           = $opts{nacltask_verify};
    my $to_expected               = $opts{to_expected};
    my $wait_time                 = $opts{wait_time};
    my $poll_interval             = $opts{poll_interval};
    my $check_partner_state       = $opts{check_partner_state};
    my $powercycle                = $opts{powercycle};
    my $expected_takeover_state   = $opts{expected_takeover_state};
    my $expected_partner_state    = $opts{expected_partner_state};
    my $boot_after_panic          = $opts{boot_after_panic};
    my $boot_after_halt           = $opts{boot_after_halt};
    my $auto_giveback             = $opts{auto_giveback};
    my $auto_giveback_after_panic = $opts{auto_giveback_after_panic};
    my $apiset_must_should        = $opts{apiset_must_should};
    my $inhibit                   = $opts{inhibit};
    my $skip_lif_migration_before_reboot = $opts{skip_lif_migration_before_reboot};
    my $skip_lif_migration_before_shutdown = $opts{skip_lif_migration_before_shutdown};
    my $task_verify_completed = 0;

    if ( $takeover_type eq 'system_power_off'
        || $takeover_type eq 'system_reset'
        || $takeover_type eq 'system_powercycle'
        || $takeover_type eq 'watchdog_reset' )
    {
        my $invalid_choice_msg
            = 'takeover type '
            . "$takeover_type is not supported for "
            . 'AWS StorageFailover->takeover ';
        $Log->exit() if $may_exit;
        NACL::Exceptions::InvalidChoice->throw($invalid_choice_msg);
    } elsif ($takeover_type eq 'panic') {
        $Log->comment('Initiating takeover on panic for AWS filer(induced panic) ');
        if ($powercycle) {
            $Log->comment('Powercycle is set to true, but this is not currently supported for AWS platforms');
        }
        if (!$boot_after_panic) {
            $Log->comment(
                "The option boot_after_panic was passed in as 0. For AWS platforms, this is not a \n"
                . 'supported option.'
            );
        }
        # Verify node is set for autoboot before panicking
        $ci->apiset->system_node_run(
            'node'            => $partner_node->node(),
            'command'         => 'bootargs set AUTOBOOT true',
            'privilege-level' => 'test',
        );
        NACL::STask::Node->panic(
            command_interface => $partner_node,
            nacltask_reboot   => 0,
        );

        if ($verify_required) {
            my %call_opts;
            if ($to_expected) {
                $Log->comment('Executing the nacltask verify condition when TO is expected');

                $expected_takeover_state
                    = 'takeover_autogiveback_scheduled' if($auto_giveback_after_panic eq 'true');

                # if 'auto_giveback' is set to false then it overrides
                # 'auto_giveback_after_panic' option, see burt1038112.
                $expected_takeover_state
                    = 'takeover' if($auto_giveback eq 'false');

                %call_opts = (
                    command_interface      => $ci,
                    node                   => $to_node,
                    partner                => $partner_node,
                    node_state             => $expected_takeover_state,
                    'method-timeout'       => $wait_time,
                    nacltask_poll_interval => $poll_interval,
                    %$apiset_must_should,
                );

                $pkg->wait_for_state(%call_opts);

                if ($check_partner_state) {
                    $Log->comment('Verifying the nacltask_check_partner_state condition that partner is WFG');
                    %call_opts = (
                        command_interface => $ci,
                        node              => $to_node,
                        partner           => $partner_node,
                        node_state        => $expected_takeover_state,
                        partner_state     => 'wfg',
                        'method-timeout'  => $wait_time,
                        nacltask_poll_interval => $poll_interval,
                        %$apiset_must_should,
                    );
                    $pkg->wait_for_state(%call_opts);
                }
                $task_verify_completed = 1;
            } else {
                # This can be handled in the general verification block
                $expected_takeover_state = 'up';
                $expected_partner_state  = "up";
            }
        }    # if ( $verify_required )
    } elsif ($takeover_type eq 'reboot') {
        $expected_takeover_state = 'takeover_autogiveback_scheduled';
        $Log->comment('Initiating AWS takeover on reboot command');

        # build options for reboot command
        my $skip_lif_migration_option = 'false';
        my $ignore_quorum_option = 'false';
        $skip_lif_migration_option = 'true'
            if ($skip_lif_migration_before_reboot);
        $ignore_quorum_option = 'true'
            if ($opts{command_interface}->version_manager()->is_version_ge(release => '8.3.1'));

        # Verify node is set for autoboot before panicking
        $ci->apiset->system_node_run(
            'node' => $partner_node->node(),
            'command' => 'bootargs set AUTOBOOT true',
            'privilege-level'    => 'test',
        );

        if ($inhibit) {
            # Reboot using C::SystemNode to avoid use of transit library
            NACL::C::SystemNode->reboot(
                command_interface        => $ci,
                node                     => $partner_node,
                'inhibit-takeover'       => 'true',
                'skip-lif-migration'     => $skip_lif_migration_option,
                'ignore-quorum-warnings' => $ignore_quorum_option,
            );

            if ($verify_required) {
                $expected_takeover_state = 'up';
                $expected_partner_state  = 'up';
            }    ## if ($verify_required)
        } else {
            NACL::C::SystemNode->reboot(
                command_interface        => $ci,
                node                     => $partner_node,
                'inhibit-takeover'       => 'false',
                'skip-lif-migration'     => $skip_lif_migration_option,
                'ignore-quorum-warnings' => $ignore_quorum_option,
            );
            # means the flag onreboot is false in the filer and hence
            # nacltask_expect_takeover is 0
            if ( $verify_required && !$to_expected ) {
                $expected_takeover_state = 'up';
                $expected_partner_state  = 'up';
            }    ## if ($verify_required)
        }

        $Log->debug('Reboot has been executed');
    } elsif ($takeover_type eq 'halt') {
        $Log->comment('Initiating Takeover on halt command');

        # Set opts for halt command
        my $skip_lif_migration_option = $skip_lif_migration_before_shutdown ? 'true' : 'false';
        my $ignore_quorum_option = ($opts{command_interface}->version_manager()->is_version_ge(release => '8.3.1')) ? 'true' : 'false';

        # Check if inhibit is set
        if ($inhibit) {
            NACL::C::SystemNode->halt(
                command_interface        => $ci,
                node                     => $partner_node,
                'inhibit-takeover'       => 'true',
                'skip-lif-migration'     => $skip_lif_migration_option,
                'ignore-quorum-warnings' => $ignore_quorum_option,
            );

            $expected_takeover_state = 'partner_inhibit_takeover';
            $expected_partner_state  = 'nocheck';

            # Need some delay after AWS node is halted. See burt1033834
            Tharn::sleep(ONE_MIN);

            if ( $boot_after_halt == 1 ) {
                #Wait for aws state to be running
                aws_bring_filer_up(
                    node                  => $partner_node,
                    wait_for_health_check => 0,
                );
                $expected_takeover_state = 'up';
                $expected_partner_state  = 'up';
            }
        } else {
            NACL::C::SystemNode->halt(
                command_interface        => $ci,
                node                     => $partner_node,
                'inhibit-takeover'       => 'false',
                'skip-lif-migration'     => $skip_lif_migration_option,
                'ignore-quorum-warnings' => $ignore_quorum_option,
            );

            $expected_partner_state = 'nocheck';

            # Need some delay after AWS node is halted. See burt1033834
            Tharn::sleep(ONE_MIN);

            if ( $boot_after_halt == 1 ) {
                #Wait for aws state to be running
                aws_bring_filer_up(
                    node                  => $partner_node,
                    wait_for_health_check => 0,
                );
                $expected_partner_state = 'wfg';
            }
        }

        $Log->comment('Change state in halt on AWS is complete');
    }
    # This is a general verification for options that don't have a unique verification
    if ( $verify_required && !$task_verify_completed ) {
            $Log->comment('Executing the general nacltask verify condition');

            my %call_opts = (
                command_interface      => $ci,
                node                   => $to_node,
                partner                => $partner_node,
                node_state             => $expected_takeover_state,
                'method-timeout'       => $wait_time,
                nacltask_poll_interval => $poll_interval,
                %$apiset_must_should
            );
            if ($check_partner_state) {
                $call_opts{'partner_state'} = $expected_partner_state;
            }
            $pkg->wait_for_state(%call_opts);
        } ## end if ($verify_required)
    $Log->exit() if $may_exit;
    return;
} ## end sub _takeover_AWS

=head2 giveback

  NACL::STask::StorageFailover->giveback(
      command_interface     => $command_interface,
      node                  => $node_object,
      partner               => $partner_object,
      'require-partner-waiting' =>"true",
      'override-vetoes'=>"true",
      'only-cfo-aggregates'=>"true",
      'wait-for-asup-generation' => "0",
      'failure-reason' => 'giveback_failed_due_to_failed_disk',
      method-timeout        => $method-timeout,
      %other_options
  );

(Class method) Initiate giveback. This method provides additional
services that are inherited from
L<NACL::C::StorageFailover|lib-NACL-C-StorageFailover-pm> to perform
giveback based on certain storage-failover flags.

The type of giveback is determined by "auto-giveback" flag
along with user input options "require-partner-waiting",
"override-vetoes" and "only-cfo-aggregates". User input options
"require-partner-waiting", "override-vetoes" and "only-cfo-aggregates"
is valid only when "auto-giveback" flag is false.

=over

=item Options

=over

=item C<< node=>$node_object >>

(Required) Node initiating giveback.
Specifies the node object of type NACL::C::Node, that currently holds the storage that is to be
returned to the partner node.

=item C<< partner=>$partner_object >>

(Required) Node to which control is givenback.
Specifies the partner object of type NACL::C::Node, whose storage is currently taken over
by its partner and will be given back by this giveback operation.

=item C<< method-timeout=>$method-timeout >>

(Optional) Time in seconds to perform giveback. When the caller passes a value for this
           method-timeout it will be used to wait for the partner to go to WFG
           State. Also the same value is used to wait for the partner to be back to UP
           state after the giveback is issued.

Default :  For partner to reach WFG  prompt , this is set to 300 seconds
           Again for the partner to come back to UP state from WFG , this is set to 220 seconds.

=item C<< 'require-partner-waiting'=>"true" >>

(Optional) user input flag that makes the takeover node giveback
only after the partner is at waiting for giveback prompt
Default : 'true'

=item C<< 'override-vetoes'=>"true" >>

(Optional) user input flag that allows giveback to happen
even when giveback is vetoed due to any reason
Default : 'true'

=item C<< nacltask_skip_auto_giveback => 1 >>

(Optional) user input flag that is used to prevent to wait for auto giveback
to be happened. It can't be used along with auto_giveback_expected option
Default : 0

=item C<< nacltask_skip_resync_check => 1 >>

(Optional) user input flag that is used to prevent checking that the root
aggregate has completed resync before initiating giveback. This is a common
issue in AWS that fails giveback due to a resync in progress.
Default : 0

=item C<< nacltask_resync_timeout => 3600 >>

(Optional) user input time before resync check is aborted. Units are in
seconds and default to 3600, which equal to 1 hour.
Default : 3600

=item C<< 'only-cfo-aggregates'=>"true" >>

(Optional) user input flag that is used to giveback only cfo style aggregates
Default : 'true'

=item C<< wait_for_partner => 1 >>

(Optional) user input flag that is used to wait for partner to come to wfg
state before issuing a giveback
Default : 1

=item C<< auto_giveback_expected => 0 >>

(Optional) user input flag that is used to decide whether an auto giveback
is expected or not. Example, if takeover happens due to partner reboot then
even if the autogiveback flag is disabled, a autogiveback is expected to
happen
Default : 0

=item C<< 'wait-for-asup-generation' => "0" >>

(Optional) user input flag that is used to wait as along as autosupport
is being generated if giveback is vetoed due to autosupport.
Default : 'false'

=item C<<  'failure-reason' => 'giveback_failed_due_to_failed_disk' >>

(Optional) user input flag that is used to exit giveback in case
of failure. To avoid the auto-giveback failure due failed disks,
'failure-reason' => 'giveback_failed_due_to_failed_disk'
needs to be specified.

    NACL::STask::StorageFailover->giveback(
        command_interface => $command_interface,
        node              => $node_object,
        partner           => $partner_object,
        'failure-reason'  => 'giveback_failed_due_to_failed_disk',
        'override-vetoes' => 'true'
    );

If auto-giveback fails due to disk failure and if override-vetoes is true,
task will perform manual giveback with 'override-vetoes' = true.

=item C<< nacltask_verify => $nacltask_verify >>

(Optional) Boolean value, 0 or 1. Defaults to 1 (perform
verification). If 1, then check_state() will be invoked to
verify complete giveback scenario. Otherwise its not.

=item C<< nacltask_poll_interval => $poll_interval >>

(Optional) A value for the polling interval to verify the status of giveback.
The default value is 10 seconds.

=item C<< nacltask_ignore_partial_giveback  => $nacltask_ignore_partial_giveback  >>

(Optional) Boolean value, 0 or 1. Defaults to 0 ( Don't ignore the partial giveback).
If 1, then nacltask_verify will ignore the partial giveback state.

=item C<< nacltask_pre_giveback_callbacks >>

(Optional) (For CMode) Provides ability to execute callbacks before a
giveback event
(For 7Mode) the option is ignored

=item C<< nacltask_post_giveback_callbacks >>

(Optional) (For CMode) Provides ability to execute callbacks after a
giveback event
(For 7Mode) the option is ignored

=item command_interface, apiset_must, apiset_should etc.

All of the other various options are described in
L<NACL::C::StorageFailover|lib-NACL-C-StorageFailover-pm>.

=back

=item Assumptions

Setting storage-failover flags is not a part of this method,
and has to be handled before this method.

Storage-failover flags Storage-failover flag auto-giveback along with
user input options "require-partner-waiting","override-vetoes" and "only-cfo-aggregates"
will determine the type of giveback.

=item Exceptions

=over

=item C<NACL::Exceptions::VerifyFailure>

This type of exception is thrown when verification fails for giveback.

=item C<NATE::Exceptions::Argument>

This type of exception is thrown when invalid params are passed.

=item C<NACL::C::Exceptions::StorageFailover::UnableToReadPartnerState>

This type of exception is thrown when there is failure to read partner state.

=item C<NACL::C::Exceptions::StorageFailover::DiskInventoryNotReceived>

This type of exception is thrown when disk inventory not received by node.

=item C<NACL::Exceptions::InvalidChoice>

This type of exception is thrown when an invalid option is passed.

=item C<NATE::BaseException>

This type of exception is thrown when autosupport support did not finish even after 40 mins.

=back

=back

=cut

sub giveback {
    $Log->enter() if $may_enter;
    my $pkg = shift;

    my %opts = $pkg->_common_validate_with(
        params             => \@_,
        only_static_method => 1,
        additional_spec    => {
            node    => { isa => 'NACL::C::Node', type => OBJECT },
            partner => { isa => 'NACL::C::Node', type => OBJECT },
            'require-partner-waiting' =>
                { type => SCALAR | UNDEF, optional => 1 },
            'override-vetoes' => { type => SCALAR | UNDEF, optional => 1 },
            'only-cfo-aggregates' =>
                { type => SCALAR | UNDEF, optional => 1 },
            nacltask_verify             => { type => SCALAR,  default => 1 },
            nacltask_skip_auto_giveback => { type => BOOLEAN, default => 0 },
            nacltask_skip_resync_check  => { type => BOOLEAN, default => 0 },
            nacltask_resync_timeout     => { type => SCALAR,  default => 3600},
            wait_for_partner            => { type => BOOLEAN, default => 1 },
            auto_giveback_expected      => { type => BOOLEAN, default => 0 },
            'wait-for-asup-generation'  => { type => BOOLEAN, default => 0 },
            nacltask_ignore_partial_giveback => { type => BOOLEAN, default => 0},
            'failure-reason' => { type => SCALAR | UNDEF, optional => 1 },
            nacltask_poll_interval =>
                { type => SCALAR, default => POLLING_INTERVAL },
            nacltask_pre_giveback_callbacks  => { type => ARRAYREF, optional => 1},
            nacltask_post_giveback_callbacks => { type => ARRAYREF, optional => 1},
        },
    );

    my $node         = $opts{node};
    my $partner      = $opts{partner};
    my $node_name    = $opts{node}->node();
    my $partner_name = $opts{partner}->node();

    # copy apiset_must_should so we can pass it down
    my %apiset_must_should;
    $pkg->_hash_copy(
        source => \%opts,
        target => \%apiset_must_should,
        copy   => [qw(apiset_must apiset_should)]
    );

    my $msg;
    my $ci                     = $opts{command_interface};
    my $mode                   = $ci->mode();                # 7 or C mode
    my $expected_node_state    = "up";
    my $expected_partner_state = "up";
    my $retry_giveback_waiting_for_asup = 0;
    my $retry_giveback_waiting_for_coredump  = 0;
    my $verify_required                 = $opts{nacltask_verify};
    my $nacltask_skip_auto_giveback
        = delete $opts{nacltask_skip_auto_giveback};
    my $nacltask_skip_resync_check
        = delete $opts{nacltask_skip_resync_check};
    my $nacltask_resync_timeout = delete $opts{nacltask_resync_timeout};
    my $method_timeout = delete $opts{'method-timeout'};
    my $wait_time      = $method_timeout || MAX_WFG_TO_UP;
    my $poll_interval  = delete $opts{nacltask_poll_interval};
    my $pre_giveback_callbacks = delete $opts{nacltask_pre_giveback_callbacks};
    my $post_giveback_callbacks = delete $opts{nacltask_post_giveback_callbacks};

    if ( $nacltask_skip_auto_giveback && $opts{auto_giveback_expected} ) {
        NATE::Exceptions::Argument->throw( "Invalid params - Either specify"
                . " the option nacltask_skip_auto_giveback or auto_giveback_expected"
        );
    }

    my %args;
    $args{'require-partner-waiting'} = $opts{'require-partner-waiting'}
        if ( defined $opts{'require-partner-waiting'} );
    $args{'override-vetoes'} = $opts{'override-vetoes'}
        if ( defined $opts{'override-vetoes'} );
    $args{'only-cfo-aggregates'} = $opts{'only-cfo-aggregates'}
        if ( defined $opts{'only-cfo-aggregates'} );

    $Log->debug("Mode is '$mode'");

    if ( $mode eq '7Mode' ) {

        # 7-Mode support
        $pkg->_giveback_7Mode( %opts, 'method-timeout' => $wait_time );
    } else {

        # C-Mode support
        # First check for auto-giveback delay time
        # in failover state-description
        my $auto_giveback_delay_sec;
        my $auto_gv_delay_found = 0;
        my $failo_cs            = NACL::CS::StorageFailover->fetch(
            command_interface => $opts{command_interface},
            filter            => { 'node' => $node },
            requested_fields  => [qw(state-description)]
        );

        if ( $failo_cs->state_description
            =~ m[Auto\s+giveback\s+will\s+be\s+initiated\s+in\s+(\d+)\s+seconds]i
            ) {
            $auto_giveback_delay_sec = $1;
            $auto_gv_delay_found     = 1;
        }

        # Get giveback flag values
        my $autogiveback_options
            = NACL::CS::StorageFailoverInternalOptions->fetch(
            command_interface => $opts{command_interface},
            filter            => { 'node' => $node },
            requested_fields  => [qw(auto-giveback delay-seconds)],
            %apiset_must_should
            );

        if ( !$auto_gv_delay_found ) {
            $auto_giveback_delay_sec = $autogiveback_options->delay_seconds();
        }

        # Auto-giveback processing
        if ((      $opts{auto_giveback_expected}
                || $autogiveback_options->auto_giveback() eq "true"
            )
            && !$nacltask_skip_auto_giveback
            ) {

            # Check for callbacks to be executed before giveback
            if ($pre_giveback_callbacks && @$pre_giveback_callbacks) {
                $pkg->_execute_event_callbacks(
                    callbacks         => $pre_giveback_callbacks,
                    command_interface => $ci,
                    takeover_node     => $node,
                    partner_node      => $partner,
                    event_type        => 'auto-giveback',
                );
            }

            $Log->comment( "Sleeping for auto-giveback delay seconds "
                    . "$auto_giveback_delay_sec" );
            Tharn::sleep($auto_giveback_delay_sec);

            # Check if auto-giveback exceeds the announced auto-giveback delay
            $failo_cs = NACL::CS::StorageFailover->fetch(
                command_interface => $opts{command_interface},
                filter            => { 'node' => $node },
                requested_fields  => [qw(state-description)]
            );
            if ( $failo_cs->state_description
                =~ m[Auto\s+giveback\s+will\s+be\s+initiated\s+in\s+(\d+)\s+seconds]i
                ) {
                $auto_giveback_delay_sec = $1 + 5;
                $Log->comment(
                    'Sleeping again for auto-giveback delay seconds : '
                        . $auto_giveback_delay_sec );
                Tharn::sleep($auto_giveback_delay_sec);
            }
        }    # End of Autogiveback processing

        # Manual Giveback processing (Issue giveback command)
        else {

            # Wait for partner to come to "Waiting for Giveback" state before
            # issuing giveback
            if ( $opts{wait_for_partner} == 1 ) {

                my $is_partial_giveback = 0;

                # No need to wait for partner to come to "wfg" if partner
                # state is "partial-giveback"
                try {
                    $pkg->check_state(
                        command_interface => $ci,
                        node              => $node,
                        partner           => $partner,
                        node_state        => "partial_giveback",
                        partner_state     => "nocheck",
                    );
                    $is_partial_giveback = 1;
                }
                catch NACL::STask::Exceptions::UnexpectedSFOState with {
                    my $e = shift;
                    $Log->debug( "Unexpected SFO state : " . $e->text );
                };

                if ( !$is_partial_giveback ) {

                    try {
                        $pkg->check_state(
                            command_interface => $ci,
                            node              => $node,
                            partner           => $partner,
                            node_state        => "disk_inventory_not_received",
                            partner_state     => "wfg",
                        );
                        Tharn::sleep MAX_DIS_TIME;
                    }
                    catch NACL::STask::Exceptions::UnexpectedSFOState with {
                        my $e = shift;
                        $Log->debug( "Unexpected SFO state : " . $e->text );
                    };

                    try {

                        # Unless the user has specified a method-timeout
                        # the timeout for partner to come to wfg is MAX_WFG_TIME = 300 seconds

                        my $partner_wfg_time = $method_timeout
                            || MAX_WFG_TIME;
                        $pkg->wait_for_state(
                            command_interface      => $ci,
                            node                   => $node,
                            partner                => $partner,
                            node_state             => "nocheck",
                            partner_state          => "wfg",
                            nacltask_poll_interval => $poll_interval,
                            'method-timeout'       => $partner_wfg_time,
                        );
                        $Log->comment("Verifying that the partner has reached SF_MBWAIT");
                        my $end_time      = time() + ONE_MIN ;
                        my $poll_interval = POLL_DELTA_MBWAIT;
                        my $exit_flag     = 0;
                        while ( !$exit_flag || (time() > $end_time)) {
                            my $state = NACL::CS::StorageFailover->fetch(
                                 command_interface => $ci,
                                 requested_fields  => [qw(partner-firmware-state )],
                                 filter            => { node => $node }
                            )->partner_firmware_state();
                            if ( $state eq 'SF_MBWAIT' ) {
                                $Log->debug(
                                     "Partner reached SF_MBWAIT state");
                                $exit_flag = 1;
                            } else {
                                Tharn::snooze $poll_interval;
                            }
                        }## while loop ends here
                        if ( !$exit_flag ) {
                            my $error_message
                                = "After waiting for ONE_MIN seconds "
                                . "partner did not reach SF_MBWAIT sate ";
                            $Log->exit() if $may_exit;
                            NATE::BaseException->throw($error_message);
                        } #end if
                    }
                    catch NACL::APISet::Exceptions::TimeoutException with {
                        my $exception_object = shift;
                        my $exception_text   = $exception_object->text();
                        $Log->exit() if $may_exit;
                        NATE::BaseException->throw($exception_text);
                    };
                    # Verify resync of root aggregates has completed,
                    # else will fail GB in AWS environment via raid module.
                    if (!$nacltask_skip_resync_check) {
                        # Find all the root aggregates
                        my @root_aggregates = NACL::CS::StorageAggregate->fetch(
                            command_interface => $ci,
                            filter            => {
                                'root'  => 'true',
                                'node'  => "$node_name|$partner_name",
                            },
                        );
                        $pkg->_wait_for_resync(
                            %opts,
                            aggregates     => \@root_aggregates,
                            resync_timeout => $nacltask_resync_timeout,
                        );
                    } ## end if nacltask_skip_resync_check
                }
            }
            #Waiting for 2 seconds before initiating giveback
            ## More information is available @ burt:796364"
            Tharn::sleep 2;

            #Print date and time before giveback for debugging purpose
            my $system_node_date = NACL::CS::ClusterDate->fetch(
                'command_interface' => $node,
                'requested_fields'  => [qw(dateandtime timezone)]
            );
            $Log->comment("Date     : " . $system_node_date->dateandtime());
            $Log->comment("TimeZone : " . $system_node_date->timezone());

            # Check for callbacks to be executed before giveback
            if ($pre_giveback_callbacks && @$pre_giveback_callbacks) {
                $pkg->_execute_event_callbacks(
                    callbacks         => $pre_giveback_callbacks,
                    command_interface => $ci,
                    takeover_node     => $node,
                    partner_node      => $partner,
                    event_type        => 'giveback',
                );
            }

            my $no_of_retries = 0;
            RETRY: {
                use warnings qw(exiting);
                try {
                    NACL::C::StorageFailover->giveback(
                        command_interface => $node,
                        ofnode            => $partner_name,
                        %args,
                        %apiset_must_should,
                    );
                }

                # Verify if the giveback command went through successfully
                catch NACL::C::Exceptions::StorageFailover::UnableToReadPartnerState with {
                    my $exception_object = shift;

                    # Check if failure is due to "Unable to read partner state"
                    if ( $opts{'failure-reason'}
                        =~ m/unable_to_read_partner_state/ ) {

                        # giveback failure is expected, no verification needed
                        $verify_required = 0;
                    } else {
                        $Log->exit() if $may_exit;
                        $exception_object->throw();
                    }
                }
                catch NACL::C::Exceptions::StorageFailover::DiskInventoryNotReceived with {
                    my $exception_object = shift;

                    # Check if failure is due to "Disk inventory not received"
                    if ( $opts{'failure-reason'}
                        =~ m/disk_inventory_information_not_yet_received/ ) {

                        # giveback failure is expected, no verification needed
                        $verify_required = 0;
                    } else {
                        $Log->exit() if $may_exit;
                        $exception_object->throw();
                    }
                }
                catch NACL::C::Exceptions::StorageFailover::PartnerNotFullyBooted with {
                    my $exception_object = shift;
                    # retry giveback if no_of_retries less than 5
                    if($no_of_retries++ < 5) {
                        Tharn::snooze(10);
                        redo RETRY;
                    } else {
                        $Log->exit() if $may_exit;
                        $exception_object->throw();
                    }
                }
                catch NACL::C::Exceptions::StorageFailover::GivebackAlreadyInProgress with {
                    my $exception_object = shift;

                    # Check if failure is due to "A giveback is already in progress"
                    if ( $exception_object->text() =~ /Failed to initiate giveback. Reason: A giveback is already in progress./ ) {
                        # giveback failure is expected, no verification needed
                        $verify_required = 0;
                    } else {
                           $Log->exit() if $may_exit;
                           $exception_object->throw();
                    }
                };
            }  #end of RETRY
        }    # end of Manual giveback processing

        # Verify that filers have reached expected states after giveback
        if ($verify_required) {
        LOOP: {
                my $verify_state = 1;

                # Check if giveback has been vetoed
                Tharn::sleep MAX_VETO_TIME;

                my $sfo = NACL::CS::StorageFailover->fetch(
                    command_interface => $ci,
                    filter            => { node => $node_name },
                    requested_fields  => [
                        qw(node partner-name
                            state-description)
                    ],
                    %apiset_must_should
                );
                my $failure         = $sfo->state_description();
                my $giveback_state  = $sfo->giveback_state();
                my $giveback_module = $sfo->current_giveback_module();

                # Check if failure is due to "Failed Disk"
                if (($failure =~
                    /Previous giveback failed in module: disk check/i)
                    || (   defined $giveback_state
                        && defined $giveback_module
                        && $giveback_state  =~ m[previous_giveback_failed]i
                        && $giveback_module =~ m[disk check]i )
                    ) {
                    my $exception = 0;
                    if ( $opts{'failure-reason'}
                        =~ m/giveback_failed_due_to_failed_disk/ ) {

                        # Failure expected
                        $Log->comment( "Giveback failed due to disk check "
                                . "as expected" );
                        $verify_state = 0;
                    } else {
                        $msg = "$opts{'failure-reason'} did not match "
                            . "giveback_failed_due_to_failed_disk";
                        $exception = 1;
                    }

               # Manual giveback should happen when auto-giveback gets vetoed.
                    if ( defined $opts{'override-vetoes'}
                        && $opts{'override-vetoes'} eq 'true' ) {
                        NACL::C::StorageFailover->giveback(
                            command_interface => $node,
                            ofnode            => $partner_name,
                            %args,
                            %apiset_must_should,
                        );
                        $verify_state = 1;
                    }

                    if ($exception) {
                        $Log->exit() if $may_exit;
                        NATE::BaseException->throw($msg);
                    }
                }

                # Check if failure is due to "Autosupport veto"
                elsif (
                    ($failure =~
                    /Previous giveback failed in module: autosupport/i)
                    || (   defined $giveback_state
                        && defined $giveback_module
                        && $giveback_state  =~ m[previous_giveback_failed]i
                        && $giveback_module =~ m[autosupport]i )
                ) {
                    $Log->comment(
                        "Giveback failed due to autosupport veto \n");
                    if ( $opts{'wait-for-asup-generation'} ) {
                        if ( $retry_giveback_waiting_for_asup < 20 ) {

                            # Retry giveback every two min for 40min
                            # till asup is collected
                            Tharn::sleep TWO_MINS;
                            $Log->comment( "Redoing the LOOP to peform "
                                    . "giveback as autosupport is getting generated"
                            );
                            $retry_giveback_waiting_for_asup++;

                            # Issue giveback and verify again
                            try {
                                NACL::C::StorageFailover->giveback(
                                    command_interface => $node,
                                    ofnode            => $partner_name,
                                    %args,
                                    %apiset_must_should,
                                );
                            } catch NACL::APISet::Exceptions::ResponseException with {
                                my $exception_object = shift;
                                $Log->exit() if $may_exit;
                                NATE::BaseException->throw(
                                    $exception_object->text() );
                                };

                            #Verify again
                            redo LOOP;
                        } else {
                            $msg = "Autosupport did not finish generating "
                                . "even after 40 mins";
                            $Log->exit() if $may_exit;
                            NATE::BaseException->throw($msg);
                        }
                    }    # End of "wait-for-asup-generation"
                     # "wait-for-asup-generation" is not set, so retry giveback
                     # after disabling asup
                    else {
                        $Log->comment("Disable asup and retry giveback");
                        $retry_giveback_waiting_for_asup++;
                        if ( $retry_giveback_waiting_for_asup <= 1 ) {
                            $Log->comment(
                                "Performing manual giveback after disabling asup "
                                . "since giveback was vetoed due to asup and "
                                . "the node will not retry giveback" );

                            # disable asup and retry giveback just once
                            NACL::C::SystemNodeAutosupport->modify(
                                command_interface => $ci,
                                node              => $node,
                                state             => "disable",
                                'local-collection'  => "false",
                                %apiset_must_should,
                            );
                            sleep 5;
                            # Issue giveback and verify again
                            try {
                                NACL::C::StorageFailover->giveback(
                                    command_interface => $node,
                                    ofnode            => $partner_name,
                                    %args,
                                    %apiset_must_should,
                                );
                            } catch NACL::APISet::Exceptions::ResponseException with {
                                my $exception_object = shift;
                                $Log->exit() if $may_exit;
                                NATE::BaseException->throw(
                                    $exception_object->text() );
                                };
                            #Verify again
                            redo LOOP;
                        } else {
                            $Log->comment(
                                "Enabling the asup, since it was disabled earlier");
                            # enable asup back, since it was disabled earlier
                            NACL::C::SystemNodeAutosupport->modify(
                                command_interface => $ci,
                                node              => $node,
                                state             => "enable",
                                'local-collection'  => "true",
                                %apiset_must_should,
                            );
                            $msg = "Giveback got vetoed by asup even "
                                . "after disabling asup";
                            $Log->exit() if $may_exit;
                            NATE::BaseException->throw($msg);
                        }
                    }
                }    # End of Check if failure is due to "Autosupport veto"
                elsif (
                    ($failure =~
                    /Previous giveback failed in module: coredump/i)
                    || (   defined $giveback_state
                        && defined $giveback_module
                        && $giveback_state  =~ m[previous_giveback_failed]i
                        && $giveback_module =~ m[coredump]i )
                ) {
                    if($retry_giveback_waiting_for_coredump < 3) {
                          $Log->comment(
                            "Performing manual giveback  "
                            . "since giveback was vetoed due to coredump and "
                            . "the node will not retry giveback" );

                            # w.r.t burt#710055 the failure :
                        # 'Previous giveback failed in module: coredump'
                        # should be retried with giveback after two min

                        Tharn::sleep TWO_MINS;
                        $Log->debug( "Redoing the LOOP to peform "
                            . "giveback as coredump is being generated"
                        );
                        $retry_giveback_waiting_for_coredump++;

                        # Issue giveback and verify again
                        try {
                            NACL::C::StorageFailover->giveback(
                                command_interface => $node,
                                ofnode            => $partner_name,
                                %args,
                                %apiset_must_should,
                            );
                        } catch NACL::APISet::Exceptions::ResponseException with {
                            my $exception_object = shift;
                            $Log->exit() if $may_exit;
                            NATE::BaseException->throw(
                                $exception_object->text() );
                        };

                        #Verify again
                        redo LOOP;
                    } else {
                        $msg = "Coredump did not finish generating "
                            . "even after 3 giveback attempts made
                                in interval of 2 mins";
                        $Log->exit() if $may_exit;
                        NATE::BaseException->throw($msg);
                    }
                } ## End of Check if failure is due to "coredump"

                # Verify States
                if ($verify_state) {

                    # Set the verify options in case of "only cfo aggrs"
                    if ( defined $args{'only-cfo-aggregates'}
                        && $args{'only-cfo-aggregates'} eq 'true' ) {
                        $expected_node_state    = 'partial_giveback';
                        $expected_partner_state = 'up';
                    }

                    # Verify that the nodes have come to appropriate states
                    try {
                        $pkg->wait_for_state(
                            command_interface      => $ci,
                            node                   => $node,
                            partner                => $partner,
                            node_state             => $expected_node_state,
                            partner_state          => $expected_partner_state,
                            nacltask_poll_interval => $poll_interval,
                            'method-timeout'       => $wait_time,
                        );
                    } catch NACL::APISet::Exceptions::TimeoutException with {
                        my $exception_object = shift;
                        my $exception_text   = $exception_object->text();
                        if($opts{nacltask_ignore_partial_giveback} && ($exception_text =~ /Partial giveback/i) ) {
                            $Log->debug($exception_text);
                            $Log->exit() if $may_exit;
                        } elsif (!$nacltask_skip_resync_check &&
                            ($exception_text =~
                            /Giveback of one or more SFO aggregates failed|Previous giveback failed in module: raid/i)) {
                            # Verify resync of all aggregates has completed,
                            # else will fail GB in AWS/sDOT environment via errors above.
                            $Log->comment('Previous giveback failed. ' .
                                'Checking for resync in progress and re-verifying');
                            # Find all mirrored aggregates belonging to HA pair
                            my @all_aggregates = NACL::CS::StorageAggregate->fetch(
                                command_interface => $ci,
                                allow_empty       => 1,
                                filter            => {
                                    'mirror' => 'true',
                                    'state'  => 'online',
                                    'node'   => "$node_name|$partner_name",
                                },
                            );
                            # If there are no mirrored aggregates rethrow exception
                            if (!(@all_aggregates)){
                                $exception_object->throw();
                            }
                            # If resyncing, wait for completion
                            $pkg->_wait_for_resync(
                                %opts,
                                aggregates     => \@all_aggregates,
                                resync_timeout => $nacltask_resync_timeout,
                            );
                            # Issue giveback after resync
                            NACL::C::StorageFailover->giveback(
                                command_interface => $node,
                                ofnode            => $partner_name,
                                %args,
                                %apiset_must_should,
                            );
                            # Only go through this section once
                            $nacltask_skip_resync_check = 1;
                            #Verify again
                            redo LOOP;
                        } else {
                            # Before throwing error, lets run a few debug commands
                            try {
                                $Log->comment('Expected state was not reached, running debug' .
                                    ' commands before failing:');
                                NACL::CS::StorageFailoverShowGiveback->fetch(
                                    command_interface => $ci,
                                );
                                NACL::CS::StorageFailoverShowTakeover->fetch(
                                    command_interface => $ci,
                                );
                                NACL::CS::EventLog->fetch(
                                    command_interface => $ci,
                                    allow_empty => 1,
                                    filter => {
                                        'time'     => '>15m',
                                        'severity' => 'ALERT|EMERGENCY|ERROR',
                                    }
                                );
                                $Log->comment('The system failed to reach the state it was expecting.' .
                                    ' This is not always caused by script issues. Please review the' .
                                    ' above logs before filing a burt.');
                            } catch NATE::BaseException with {
                                # We don't want to do anything to cover up the original failure
                            }; # End of try-catch for debug commands
                            $Log->exit() if $may_exit;
                            NATE::BaseException->throw($exception_text);
                        }
                    };    # End of try-catch wait_for_state

                    # Enable asup if it was disabled
                    $Log->comment('Enable asup if it was disabled by testcase');
                    if ( $retry_giveback_waiting_for_asup > 0 ) {
                        NACL::C::SystemNodeAutosupport->modify(
                            command_interface   => $ci,
                            node                => $node,
                            state               => 'enable',
                            'local-collection'  => 'true',
                            %apiset_must_should,
                        );
                    }
                }    # End of Verify States

            }    # End of LOOP
        }    # End of Verify

    }    # end of C-mode

    # Check for callbacks to be executed after giveback
    if (   $verify_required
        && $post_giveback_callbacks
        && @$post_giveback_callbacks) {
        $pkg->_execute_event_callbacks(
            callbacks         => $post_giveback_callbacks,
            command_interface => $ci,
            takeover_node     => $node,
            partner_node      => $partner,
            event_type        => 'giveback',
        );
    }

    $Log->exit() if $may_exit;
    return;
}    # end sub giveback

# Giveback implementation for 7Mode
sub _giveback_7Mode {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = @_;
    my ( $failover_options, %apiset_rules, $ci, $node, $partner, $msg,
        $transit_obj, $state );

    $ci      = $opts{command_interface};
    $node    = $opts{node};
    $partner = $opts{partner};
    # copy the apiset rules from opts
    $pkg->_copy_common_component_params(
        source => \%opts,
        target => \%apiset_rules
    );

    # get check partner enabled and auto giveback enabled flag values
    $failover_options = NACL::CS::StorageFailover->fetch(
        command_interface => $ci,
        filter            => { 'node' => $node, },
        requested_fields  => [qw(check-partner auto-giveback)],
        %apiset_rules
    );

    # Currently check-partner=true and auto-giveback=true are supported
    if (   $failover_options->check_partner() =~ /true/
        && $failover_options->auto_giveback() =~ /true/ ) {
        # If the autogiveback is enabled, ensure that the
        # state of the partner node is 'TAKEN_OVER_WAITING_ONTAP'
        # Giveback will follow. Otherwise throw an exception
        $transit_obj = NACL::Transit->new( name => $partner->hostrec() );
        $transit_obj->change_state(
            to      => 'TAKEN_OVER_WAITING_ONTAP',
            timeout => $opts{'method-timeout'},
        );
        $state
            = $transit_obj->get_state( timeout => $opts{'method-timeout'} );
        if ( $state ne 'TAKEN_OVER_WAITING_ONTAP' ) {
            $msg = ('TAKEN_OVER_WAITING_ONTAP could not be reached');
            $Log->exit() if $may_exit;
            NATE::BaseException->throw($msg);
        } ## end if ( $state ne 'TAKEN_OVER_WAITING_ONTAP')
    }
    if (   $failover_options->auto_giveback() =~ /false/i
        && $failover_options->check_partner() =~ /true/i ) {
        my %args;
        $args{'override-vetoes'} = $opts{'override-vetoes'}
            if ( defined $opts{'override-vetoes'} );
        NACL::C::StorageFailover->giveback(
            command_interface => $ci,
            %args
        );
        $Log->exit() if $may_exit;
    }
    # Throw exceptions for other combination of values of check-partner and
    # auto-giveback flags
    elsif ($failover_options->check_partner() =~ /false/i
        && $failover_options->auto_giveback() =~ /true/i ) {
        $msg
            = 'Check Partner Enabled flag is: '
            . $failover_options->check_partner()
            . ' and Auto Giveback Enabled flag is: '
            . $failover_options->auto_giveback()
            . ' and is currently not supported.';
        $Log->exit() if $may_exit;
        NACL::Exceptions::InvalidChoice->throw($msg);

    } elsif ( $failover_options->check_partner() =~ /false/i
        && $failover_options->auto_giveback() =~ /false/i ) {
        $msg
            = 'Check Partner Enabled flag is: '
            . $failover_options->check_partner()
            . ' and Auto Giveback Enabled flag is: '
            . $failover_options->auto_giveback()
            . ' and is currently not supported.';
        $Log->exit() if $may_exit;
        NACL::Exceptions::InvalidChoice->throw($msg);
    } ## end elsif ( $failover_options...
    return;
} ## end sub _giveback_7Mode

# Check the Params passed to the enable()/disable() method.
# Then decide the value which needs to be set for ha_mode.
# ha_mode = 'ha' for enable(); ha_mode = 'non_ha' for disable()
# Then calls modify_ha_mode with the appropriate ha_mode set.

sub _set_mode_in_param {

    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = @_;

    my %common_opts;
    $pkg->_move_common_component_params(
        source => \%opts,
        target => \%common_opts
    );

    my $ci     = $opts{command_interface};
    my $action = delete $opts{'action'};
    my $node   = delete $opts{_node_to_set};

    my %mode_opts = (
        command_interface => $ci,
        node              => $node,
        %common_opts,
        nacltask_verify_ha_mode => 1,
    );

    if ( !exists( $opts{'mode'} ) ) {

        if ( $action eq 'true' ) {

            $mode_opts{'nacltask_desired_ha_mode'} = 'ha';
            $pkg->modify_ha_mode(%mode_opts);

        } elsif ( $action eq 'false' ) {

            $mode_opts{'nacltask_desired_ha_mode'} = 'non_ha';
            $pkg->modify_ha_mode(%mode_opts);

        }

    } else {
        $mode_opts{'nacltask_desired_ha_mode'} = $opts{mode};
        $pkg->modify_ha_mode(%mode_opts);
    }
    $Log->exit() if $may_exit;
    return;
} ## end sub _set_mode_in_param

# Call reboot on the specified node using NACL::Transit
# Takes a commandinterface and the hostrec object of the node
# to be rebooted

sub _reboot_node {
    $Log->enter() if $may_enter;
    my $pkg = shift;
    my ( $ci, $hostobj ) = @_;

    my $is_AWS_filer = 0;
    my $filer_model;
    #attempting to get the model without connecting to the filer
    if (exists($ci->{hostrec})) {
        #if statement needed to avoid autovivification
        $filer_model = $ci->{hostrec}->{MODEL};
    } else {
        $Log->comment('Could not determine filer model, returning non AWS model type');
        $filer_model = 'FAS';
    }
    if ( $filer_model =~ AWS_MODEL_REGEX ) {
        $is_AWS_filer = 1;
        $Log->comment('Model Number is of an AWS filer and will take an alternate' .
            ' code path then typical FAS platforms model: ' . $ci->{hostrec}->{MODEL});
    }

    if ($is_AWS_filer) {
        # Reboot using C::SystemNode to avoid use of transit library
        my $partner_name = NACL::STask::Node->get_partner(command_interface => $ci, );
        my $partner_object = NACL::STask::Node->new( name => $partner_name, );
        NACL::C::SystemNode->reboot(
            command_interface      => $partner_object,
            node                   => $ci->node(),
            'inhibit-takeover'     => 'true',
        );
        $pkg->wait_for_state(
            command_interface     => $partner_object,
            node                  => $ci,
            partner               => $partner_object,
            node_state            => 'up',
            partner_state         => 'up',
        );
    } else {
        my $transit_obj = NACL::Transit->new( name => $hostobj );
        $transit_obj->reboot();
    }
    $ci->refresh_command_interface( no_version_change => 1 );

    $Log->exit() if $may_exit;
    return;
} ## end sub _reboot_node

=head2 modify_ha_mode

  NACL::STask::StorageFailover->modify_ha_mode(
      command_interface => $command_interface,
      node              => $node_object,
      nacltask_desired_ha_mode => 'ha',
      nacltask_verify_ha_mode => 1,
      %other_options
  );

(Class method) The method modifies storage failover ha_mode of the filer to ha or non-ha.
Works for both CMode and 7Mode .
(For CMode)
Method verifies if the option 'mode' is available in the 'storage failover
modify' command in the build .If the option is available, modifies the
ha mode to value given by the caller  and reboots the node.
(For 7Mode)
Method verifies if the option 'options cf.mode' is defined in the build.
If the option is available, modifies the
ha mode to value given by the caller  and reboots the node.

In both the cases the method does not complain if the option is not defined
in the build.

=over

=item Options

=over

=item C<< node=>$node_object >>

(Optional) Specifies the node in which the option need to be set .
This node will be rebooted. If it is not specified the node which is represented
in the command_interface is considered for the value.

=item C<< nacltask_desired_ha_mode => $string >>

(Required) Specifies the ha_mode which needs to be set on the node.
Valid values are - 'ha', 'non_ha'

=item C<< nacltask_verify_ha_mode => $boolean >>

(Optional) Specifies if the verification has to follow the modification.
Defaults to '0' which means that verification is not done.

=item command_interface, apiset_must, apiset_should etc.

All of the other various options are described in
L<NACL::C::StorageFailover|lib-NACL-C-StorageFailover-pm>.

=back

=item Exceptions

=over

=item C<NACL::Exceptions::VerifyFailure>

This type of exception is thrown when verification ha mode modify has failed.

=back

=back

=cut

sub modify_ha_mode {

    $Log->enter() if $may_enter;
    my $pkg = shift;

    # validate the parameters
    my %opts = $pkg->_common_validate_with(
        params             => \@_,
        only_static_method => 1,
        additional_spec    => {
            node => { type => OBJECT | SCALAR | UNDEF, optional => 1 },
            nacltask_desired_ha_mode => { type => SCALAR },
            nacltask_verify_ha_mode  => {
                type      => BOOLEAN,
                default   => 0,
                callbacks => {
                    'The value of nacltask_verify_ha_mode should be 1 or 0 '
                        => sub { $_[0] == 0 || $_[0] == 1 }
                }
            }
        },
    );

    my %common_opts;
    $pkg->_move_common_component_params(
        source => \%opts,
        target => \%common_opts
    );

    my $ci   = delete $opts{command_interface};
    my $node = delete $opts{node};
    my $mode_option_present;

    #CMode verification
    if ( $ci->mode =~ /cmode/i ) {
        my $apiset = $ci->apiset( interface => 'CLI' );
        $mode_option_present = $apiset->is_option_present(
            api   => 'storage_failover_modify',
            field => 'mode'
        );
        if ( !$mode_option_present ) {
            $Log->debug('ha-mode is not available in the build ');
            $Log->exit() if $may_exit;
            return;
        }
    }
    # 7mode verification
    else {
        $mode_option_present = NACL::CS::Options->fetch(
            command_interface => $ci,
            filter            => { 'name' => 'cf.mode' },
            allow_empty       => 1
        );
        if ( !$mode_option_present ) {
            $Log->debug('ha-mode is not available in the build ');
            $Log->exit() if $may_exit;
            return;

        }
    }

    my $target_mode = delete $opts{nacltask_desired_ha_mode};
    my $verify      = delete $opts{nacltask_verify_ha_mode};

    # If node name is not given then going to convert mode of the
    # node belonging to command interface
    unless ($node) {
        $node = $ci->node();
    }

    # Need to populate these two values for further ops.
    my ( $node_name, $hostrec );

    if ( blessed($node) && $node->isa('NACL::C::SystemNode') ) {
        $node_name = $node->node();
        $hostrec   = $node->hostrec();

    } else {
        $node_name = $node;

        $hostrec = NACL::C::SystemNode->new( name => $node_name )->hostrec();
    }

    my $cs_pkg      = $pkg->get_CS_package_name();
    my $actual_mode = $cs_pkg->fetch(
        command_interface => $ci,
        requested_fields  => [qw(mode)],
        filter            => { node => $node_name }
    )->mode();

    $Log->debug( 'Actual mode :' . $actual_mode );
    $Log->debug( 'Expected mode:' . $target_mode );

    # Is it already in the expected ha mode
    if ( $actual_mode eq $target_mode ) {
        $Log->debug(
            "The node $node_name is already in $target_mode. \n"
                . ' Hence not modifying the storage failover option'
        );
        $Log->exit() if $may_exit;
        return;
    }

    $Log->debug("Modifying the ha_mode for node $node_name to $target_mode");
    $pkg->modify(
        command_interface => $ci,
        node              => $node_name,
        mode              => $target_mode,
        %common_opts
    );

    $Log->comment("Rebooting the node $node_name ");
    $pkg->_reboot_node( $ci, $hostrec );

    $Log->debug(
        "Successfully modified the ha_mode flag to $target_mode for node $node_name"
    );

    if ($verify) {
        my ($target_mode_verify, );
        $target_mode_verify = $target_mode;

        # If using ZAPIs and $target_mode is 'non_ha', we need to change this
        # to 'non-ha' for verification, as the status text returned in the ZAPI
        # is different than text seen in CLI response.
        if (
            $target_mode eq 'non_ha'
            && defined($ci->apiset_should())
            && $ci->apiset_should()->{interface} eq 'ZAPI'
        ) {
            $target_mode_verify = 'non-ha';
        }

        try {
            $pkg->verify_state(
                command_interface => $ci,
                node              => $node_name,
                mode              => $target_mode_verify,
                %common_opts
            );
        }
        catch NACL::Exceptions::VerifyFailure with {
            my $e = shift;
            $Log->debug(
                "The node $node-name is not in ha_mode '$target_mode' after modification "
            );
            $Log->exit() if $may_exit;
            $e->throw();
        };
    }
    $Log->exit() if $may_exit;
    return;
} ## end sub modify_ha_mode

################################################################################
# Function     : _modify()
# Description  :
# Arguments    : ??
# Returns      : None
################################################################################
sub _modify {
    $Log->enter() if $may_enter;
    my (@args, %opts, $pkg, $pkg_or_obj, $retry_count, );
    ( $pkg_or_obj, @args ) = @_;
    %opts = $pkg_or_obj->_common_validate_with(
        params => \@args,
        additional_spec =>
            { node => { type => OBJECT | UNDEF | SCALAR, optional => 1 }, },
        ignore_primary_keys => 1,
        allow_extra         => 1
    );
    $retry_count = 1;
    $pkg = ref($pkg_or_obj) || $pkg_or_obj;

    MODIFY: {
        my ($must_redo, );
        $must_redo = 0;
        try {
            $pkg->modify(%opts);
        }
        catch NATE::BaseException with {
            my $ex = shift;
            if ( $ex->text() =~ /stale\s+connection/i ) {

                # Since this method is being used in the MTask::Revert where
                # the version might change hence 'no_version_change' is not
                # being used here
                $opts{'command_interface'}->refresh_command_interface();
                if($retry_count--) { $must_redo = 1; }
            } else {
                $ex->throw();
            }
        };
        if ($must_redo) { redo MODIFY; }
    }
    $Log->exit() if $may_exit;
    return;
} # End _modify()

################################################################################
# Function     : _common_modify_for_cleanup()
# Description  :
# Arguments    : ??
# Returns      : ??
################################################################################
sub _common_modify_for_cleanup {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my $options_for_register =
        $pkg_or_obj->SUPER::_common_modify_for_cleanup(@args, _to_cleanup_method => '_modify');
    $Log->exit() if $may_exit;
    return $options_for_register;
} # End _common_modify_for_cleanup()

################################################################################
# Function     : _wait_till_healthy_and_modify()
# Description  : ???
# Arguments    : ???
# Returns      : None
################################################################################
sub _wait_till_healthy_and_modify {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;

    my %opts = $pkg_or_obj->_common_validate_with(
        params             => \@args,
        additional_spec    => {
            node => { type => SCALAR | OBJECT | UNDEF, optional => 1 },
            action => { type => BOOLEAN , default => 'true'},
            nacltask_retry_tries_count => { type => SCALAR, default => 3 },
            nacltask_retry_sleep_time  => { type => SCALAR, default => 180 },
        },
        ignore_primary_keys => 1,
        allow_extra => 1
    );

    my $command_interface = $opts{command_interface} // $pkg_or_obj->command_interface();

    my %retry_opts;
    $pkg_or_obj->_move_retry_params(
        source => \%opts,
        target => \%retry_opts
    );

    my %apiset_must_should;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%apiset_must_should,
        copy   => [qw(apiset_must apiset_should)]
    );

    ## wait till cluster gets healthy
    NACL::STask::Cluster->wait_for_healthy_state(
        command_interface => $command_interface,
        node => $opts{node},
        'health-check-mode'  => 'simple',
        %retry_opts
    );

    if(defined $opts{node}){
        ## modify the storage failover
        NACL::C::StorageFailover->modify(
            command_interface => $command_interface,
            node              => $opts{node},
            enabled           => $opts{action},
            %apiset_must_should
        );
    }
    $Log->exit() if $may_exit;
    return;
} # End _wait_till_healthy_and_modify()

################################################################################
# Function     : _wait_for_resync()
# Description  : ???
# Arguments    : ???
# Returns      : None
################################################################################
sub _wait_for_resync {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    my %opts = $pkg_or_obj->_common_validate_with(
        params => \@args,
        additional_spec => {
            aggregates      => { type => ARRAYREF},
            resync_timeout  => { type => SCALAR},
        },
        ignore_primary_keys => 1,
        allow_extra         => 1
    );

    my $command_interface         = $opts{command_interface};
    my $aggregates                = delete $opts{aggregates};
    my $resync_timeout            = delete $opts{resync_timeout};

    foreach my $aggr (@{$aggregates}) {
        # Find all the aggregate plexs
        my @resync_aggr_info = NACL::CS::StorageAggregateShowResyncStatus->fetch(
            command_interface => $command_interface,
            filter            => {
                aggregate         => $aggr->aggregate(),
            },
            requested_fields  => ['status', 'resyncing-percent'],
        );
        foreach my $resync_obj (@resync_aggr_info) {
            my $timeout = $resync_timeout;
            while ($resync_obj->status() =~ /resyncing/ && $timeout > 0) {
                $Log->comment('Aggregate ' . $resync_obj->aggregate() .
                    ' is currently resyncing. Percent complete:' .
                    $resync_obj->resyncing_percent() .
                    ". Will timeout in $timeout seconds");
                $timeout -= 20;
                sleep(20);
                $resync_obj = NACL::CS::StorageAggregateShowResyncStatus->fetch(
                    command_interface => $command_interface,
                    filter            => {
                        aggregate         => $resync_obj->aggregate(),
                        plex              => $resync_obj->plex(),
                    },
                    requested_fields  => ['status', 'resyncing-percent'],
                );
            }
            $Log->comment('Resync time for aggregate: ' . $resync_obj->aggregate() .
                ' plex: ' . $resync_obj->plex() . ' took: ' .
                ($resync_timeout - $timeout) . ' seconds');
        } ## end foreach resync_aggr_info
    } ## end foreach aggr
    $Log->exit() if $may_exit;
    return;
} # End sub _wait_for_resync()

sub _execute_event_callbacks {
    $Log->enter();
    my $pkg  = shift;
    my %opts = validate(
        @_,
        {   takeover_node => {isa => 'NACL::C::Node', type => OBJECT},
            partner_node  => {isa => 'NACL::C::Node', type => OBJECT},
            command_interface =>
                {isa => 'NACL::C::CommandInterface', type => OBJECT},
            event_type    => {type => SCALAR},
            callbacks     => {type => ARRAYREF}
        },
    );
    my $event_callbacks = delete $opts{callbacks};
    my %callback_opts;
    my %opts_map = (
        'takeover_node' => 'node',
        'partner_node'  => 'partner',
    );
    $pkg->_hash_copy(
        source => \%opts,
        target => \%callback_opts,
        copy   => [qw/command_interface event_type/],
        map    => \%opts_map,
    );
    foreach my $callback (@$event_callbacks) {
        $callback->(%callback_opts);
    }

    $Log->exit();
} ## end sub _execute_event_callbacks

1;
