# Copyright (c) 2014 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary StoragePool Task Module
## @author dl-nacl-dev@netapp.com
## @status planned
## @pod here

package NACL::STask::StoragePool;

use strict;
use warnings;
use feature 'state';

use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use base qw(NACL::C::StoragePool NACL::STask::STask);
use Params::Validate qw(validate validate_with HASHREF SCALARREF BOOLEAN);
use NACL::ComponentUtils qw(Dumper);
use NACL::STask::Aggregate;
use NATE::BaseException;
use NATE::Exceptions::Argument qw(:try);
use NACL::C::Exceptions::StoragePool::AlreadyExists;

=head1 NAME

NACL::STask::StoragePool

=head1 DESCRIPTION

C<NACL::STask::StoragePool> provide methods to create and
purge a shared storage pool

It is created on top of C<NACL::C::StoragePool> component.
Since it is derived class of C<NACL::C::StoragePool>,
we can use all the methods of C<NACL::C::StoragePool>
from the object of this task.

=head1 CLEANUP METHODS

Cleanup can be registered for the following methods,

     StoragePool Method         Cleanup Method
     -----------------------------------------
     create                      purge

=head1 ATTRIBUTES

=head2 command_interface

(Required) A component object that represents the node to which to send commands.
See L<NACL::C::Component::command_interface|lib-NACL-C-Component-pm/command_interface>

=head2 storage_pool

(Optional) Shared Storage Pool Name.

=head1 METHODS

=head2 create

  my $StoragePool = NACL::STask::StoragePool->create(
      command_interface          => $ci,
      'storage-pool'      => $shared_storage_pool,
      'disk-list'                => \@disk_list,
      'nacltask_to_cleanup'      => 1, #default 0
      'nacltask_cleanup_manager' => $CleanupObj,
      %other_options,
  );

  or

  my $StoragePool = NACL::STask::StoragePool->create(
      command_interface => $ci,
      nacltask_determine_disk_list => {
          node      => $node_name,
          diskcount => $diskcount,
          raidtype  => $raidtype
      },
      nacltask_if_exists  => 'reuse',
      nacltask_verify     => 1,
      nacltask_to_cleanup => 1, #default 0
      'nacltask_cleanup_manager' => $CleanupObj,
      %other_options,
  );

# The STask can determine the appropriate disk-list if one was not supplied.
# If we'd like to control the disks that get chosen, 
# the "nacltask_determine_disk_list" option can be provided.
# It is a hash-reference that accepts "node", "diskcount" and "raidtype" as keys.

  or 

  my $StoragePool = NACL::STask::StoragePool->create(
     command_interface => $ci,
     ); 

# STask automatically picks a name. When combined with the above functionality of
# automatically being able to pick a disk-list, we can make calls like this
# where no options are supplied.

(Class Method) This method is used to create a new shared storage pool.
And also provides a mechanism of cleanup for the objects that are created by
this method.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< 'storage-pool' => $shared_storage_pool >>

Shared Storage Pool Name.
Note that if this is not specified, a random name is generated using
L<NACL::GeneralUtils::random_name_generator|lib-NACL-GeneralUtils-pm/random_name_generator>.
(See C<nacltask_rng_args> for how to customize arguments to
random_name_generator())

=item C<< nacltask_if_exists => $action >>

(Optional) What to do if the Storage Pool to be created already exists.
If $action is "die" (the default), then fail with an exception (in the
same way that the storage pool component would have: by trying the
storage pool create and letting the product complain about the storage pool
already existing).  If action is "reuse", then do nothing (return a
task object referring to the existing storage pool).  If action is
"purge", then delete the storage pool (see the C<purge> method, below)
before creating a new one.


=item C<< _was_created => \$scalar >>

(Optional)
When this option is provided a reference to a scalar variable, the variable
gets filled in with a boolean value describing whether the storage pool
was found (value will be 0; this scenario is possible when
if_exists => "reuse") or whether the storage pool was created (value will be 1).
This is necessary to determine whether the storage pool needs to be
cleaned up later.

    my $was_created;
    my $storage_pool_obj = NACL::STask::StoragePool->create(
        if_exists    => 'reuse',
        _was_created => \$was_created,
        %other_opts
    );
    # Operate on $storage_pool_obj here
    # ...
    # Now determine whether to clean up the storage pool, since we're not sure
    # whether we reused an existing storage pool or created a new one
    if ($was_created) {
        # New storage pool was created. Clean it up.
        $storage_pool_obj->purge();
    }

=item C<< nacltask_wait => 0|1 >>

(Defaults to 1)
This argument can be used to specify whether to wait for creation of the storage
pool.

=item C<< nacltask_determine_disk_list => $hashref >>

(Optional, considered only if disk-list is not provided)
Determine the disk-list to use for the storage pool.
If we'd like to control the disks that get chosen, the 
"nacltask_determine_disk_list" option can be provided. It is a hash-reference
that accepts "node", "diskcount" and "raidtype" as keys.

=item C<< nacltask_to_cleanup => 0|1 >>

(Optional, default to 0(no to cleanup)
Flag indicating if this operation needs to be registered for clean up or not.

=item C<< nacltask_cleanup_manager >>

Cleanup manager to be used for registering.
Default : Will use the default cleanup manager.

=item other options

See L<< NACL::C::StoragePool->create |lib-NACL-C-StoragePool-pm/create >>
for all the other options accepted by this method.

=back

=back

=over

=item Exceptions

=item C<NACL::STask::Exceptions::StoragePool::NotCreated>

This type of exception is thrown if the requested storage pool
was not created.

=item C<NACL::STask::Exceptions::StoragePool::NotCreated>

This type of exception is thrown if the requested storage pool
was not created.

=item NACL::C::Exceptions::StoragePool::AlreadyExists

This type of exception is thrown when an attempt is made to add an IP address
that already is added to that interface

=item Storage Pool Component exceptions

=back

=cut

sub create {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;

    state $spec = {
        $pkg->_cleanup_validate_spec(),
        nacltask_determine_disk_list => {type => HASHREF, optional => 1},
        nacltask_if_exists => $pkg->_if_exists_validate_spec,
        nacltask_wait      => {type => BOOLEAN, default => 1},
        nacltask_verify    => {type => BOOLEAN, default => 0},
        nacltask_rng_args  => $pkg->_rng_args_validate_spec(),
        _was_created       => {type => SCALARREF, optional => 1},
    };

    my %opts = $pkg->_common_validate_with(
        params              => \@args,
        additional_spec     => $spec,
        allow_extra         => 1,
        ignore_primary_keys => 1,
    );

    my $if_exists   = delete $opts{nacltask_if_exists};
    my $verify      = delete $opts{nacltask_verify};

    # If verify => 1, then we have to wait regardless of value of nacltask_wait
    my $wait        = delete $opts{nacltask_wait} || $verify;
    my $was_created = delete $opts{_was_created};
    $$was_created = 0;

    my (%opts_for_cleanup, %opts_for_register, $nacltask_to_cleanup);
    $pkg->_move_common_cleanup_opts(
        source => \%opts,
        target => \%opts_for_cleanup,
    );

    #Assign a random name if pool name is not provided
    my $rng_args = delete $opts{nacltask_rng_args};
    if (!defined $opts{'storage-pool'}) {
        $opts{'storage-pool'} = $pkg->random_name_generator(%$rng_args);
        my $name = $opts{'storage-pool'};
    }

    #Determine the disk list if not passed by the user.
    if (!defined $opts{'disk-list'}) {
        $opts{'disk-list'} = $pkg->_determine_disk_list(%opts);
    }
    delete $opts{'nacltask_determine_disk_list'};

    #Create the Storage pool using the disk list.
    my $self;
    
    # We need the job_component to wait for creation
    my $job;
    $opts{job_component} ||= \$job;
CREATE: {
        use warnings qw(exiting);
        try {
            # Call the component StoragePool create method
            $self = $pkg->SUPER::create(%opts);
            $$was_created = 1;
            
            my $job_returned = ${$opts{job_component}};
            if ($wait && $job_returned) {
                my %common_opts;
                $pkg->_copy_common_component_params(
                    source => \%opts,
                    target => \%common_opts,
                );
                $job_returned->wait_on_job(%common_opts);
            }
        }
        catch NACL::C::Exceptions::StoragePool::AlreadyExists with {
            my $err_msg = shift;
            $self = $pkg->_element_exists_handler(
                create_opts        => \%opts,
                nacltask_if_exists => $if_exists,
                exception          => $err_msg
            );
            if (!$self) {
                no warnings 'exiting';
                redo CREATE;
            }
        }
	catch NATE::BaseException with {
	    my $ex = shift;
	    if (($ex-> text() =~/Error: Job (\d+) is running in the background./i)) {
		my $jobid = $1;
		my $job = NACL::C::Job->new(
		    command_interface=>$opts{command_interface},
		    id =>$jobid,
		);
                if ($wait && $job) {
		    $job->wait_on_job();
                }
		$self = $pkg->SUPER::find(
			    command_interface    => $opts{command_interface},
			    filter               => {
				'storage-pool'	 => $opts{storage_pool}
			    },
			    allow_empty          => 1,
			);
	    }
	    else {
		$ex->throw();
	    }
	};
    }
    delete $opts{'disk-list'};
    delete $opts{'job_component'};
    delete $opts{'simulate'};
    $pkg->taskverify_create(%opts) if $verify;


    if ($$was_created) {
        $pkg->_copy_common_opts_for_cleanup(
            'source'              => {%opts, %opts_for_cleanup},
            'target'              => \%opts_for_register,
            'nacltask_to_cleanup' => \$nacltask_to_cleanup,
            'to_cleanup'          => 'purge',
        );
        $pkg->_register_for_cleanup(%opts_for_register)
            if ($nacltask_to_cleanup);
    }

    $Log->exit() if $may_exit;
    return $self;
} ## end sub create

=head2 purge

  NACL::STask::StoragePool->purge(
      command_interface     => $ci,
      'storage-pool' => $shared_storage_pool,
      %other_options,
  );
                or

  $StoragePool_obj->purge(...);

(Class or instance method)
This method is used to delete an existing shared storage pool.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< 'storage-pool' => $shared_storage_pool >>

(Required for class method, Not Applicable for instance method)
Shared Storage Pool Name.

=item other options

See L<< NACL::C::StoragePool->delete|lib-NACL-C-StoragePool-pm/delete >>
for all the other options accepted by this method.

=back

=back

=over

=item Exceptions

=over

=item C<NACL::C::Exceptions::StoragePool::DoesNotExist>

This type of exception is thrown when an attempt is made to delete a non existing
storage pool.

=back

=back

=cut

sub purge {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;

    state $spec = {nacltask_verify => {type => BOOLEAN, default => 0},};

    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => $spec,
    );

    my $verify = delete $opts{nacltask_verify};
    $pkg_or_obj->delete(%opts);

    $pkg_or_obj->taskverify_purge(%opts) if ($verify);

    $Log->exit() if $may_exit;
} ## end sub purge

sub _determine_disk_list {
    $Log->enter() if $may_enter;
    my ($pkg, %opts) = @_;

    my %nacltask;
    if (defined $opts{'nacltask_determine_disk_list'}) {
        %nacltask = %{$opts{'nacltask_determine_disk_list'}};
    }

    #If node is not provided, default it to "local".
    if (!defined $nacltask{'node'}) {
        $nacltask{'node'} = $opts{'command_interface'}->node();
    }

    my $command_interface = $opts{'command_interface'};

    my $partner_name = ();
    my (@all_ssd_disks, @partner_ssd_disks) = ();
    my $spare_ssd_count = ();
    my $nodescope_obj   = $command_interface->get_7m_or_nodescope_apiset();

    #Get the minimum diskcount for that raidtype
    my $new_min_diskcount = NACL::STask::Aggregate->get_min_diskcount(
        'command_interface' => $opts{'command_interface'},
        'raidtype'          => $nacltask{'raidtype'}
    );

    # If diskcount is not provided then use this minimum value
    # if diskcount is provided, then compare it against the
    # minimum value. If it is lower than the minimum value,
    # set it to the minimum

    if (!defined $nacltask{diskcount}) {
        $nacltask{diskcount} = $new_min_diskcount;
    } else {
        if ($nacltask{diskcount} < $new_min_diskcount) {
            $nacltask{diskcount} = $new_min_diskcount;
        }
    }
    my $disk_count = $nacltask{diskcount};

    # Now, we know the number of disks required for the
    # storage pool and which node these disks should be on.
    # Get the list of all spare physical SSDs on node
    my @spare_ssd_disks = NACL::CS::RaidConfigInfoListDisk->fetch(
        command_interface => $opts{'command_interface'},
        filter            => {
            spare             => 1,
            disk_type         => 'SSD',
            logical_disk_type => 'physical_disk',
        },
        allow_empty           => 1,
        'nodescope-node-name' => $nacltask{node},
    );

    if (scalar(@spare_ssd_disks) < $disk_count) {
        try {
            my %common_opts;
            $pkg->_copy_common_component_params(
                source => \%opts,
                target => \%common_opts
            );
            $partner_name = NACL::STask::Node->get_partner(
                command_interface => $command_interface,
                %common_opts,
            );
        }
        catch NACL::Exceptions::HighAvailabilityError with {
            $Log->comment('Partner node not available');
        };

        if ($partner_name) {

            #Get All disks properties from partner node:
            @partner_ssd_disks = NACL::CS::RaidConfigInfoListDisk->fetch(
                command_interface => $command_interface,
                allow_empty       => 1,
                filter            => {
                    logical_disk_type => 'physical_disk',
                    spare             => 1,
                    disk_type         => 'SSD',
                },
                'nodescope-node-name' => $partner_name,
            );
        }
    }

    # if num_spare_physical_ssds >= diskcount
    # if num disks in pool0 <= diskcount
    # Assign disks from pool1 to pool0 till num disks in pool0 == diskcount
    if ((scalar @spare_ssd_disks >= $disk_count)) {

        # Find all spare ssd disks from current node.
        # If pool0 is not having enough disks, then assign the SSD disks
        # from pool1 to pool0 and then create SSP
        my %pool_disks = (0 => [], 1 => []);
        map { push(@{$pool_disks{$_->pool()}}, $_->name()) } @spare_ssd_disks;
        $Log->debug(
            sub {
                '_determine_disk_list: SSD disks:' . Dumper \%pool_disks;
            }
        );

        my @pool_0 = @{$pool_disks{0}};
        my @pool_1 = @{$pool_disks{1}};

        if (scalar @pool_0 < $disk_count && scalar @pool_1 > 0) {
            for (my $i = 0; (scalar @pool_0 + $i) < $disk_count; $i++) {
                $nodescope_obj->disk_assign(
                    'disk-list'  => $pool_1[$i],
                    force        => 'true',
                    'owner-name' => $command_interface->node,
                    pool         => 0
                );
            }
        }

        @all_ssd_disks = NACL::CS::RaidConfigInfoListDisk->fetch(
            command_interface => $command_interface,
            allow_empty       => 1,
            filter            => {
                spare             => 1,
                disk_type         => 'SSD',
                logical_disk_type => 'physical_disk',
                pool              => 0,
            },
        );
    } elsif ((scalar @spare_ssd_disks < $disk_count)
        && (scalar @partner_ssd_disks >= $disk_count)) {
        $Log->comment(
            '_determine_disk_list: Not enough disks are available in current node so checking in partner node'
        );

        my %partner_pool_disks = (0 => [], 1 => []);
        map { push(@{$partner_pool_disks{$_->pool()}}, $_->name()) }
            @partner_ssd_disks;
        $Log->debug(
            sub {
                '_determine_disk_list: Partner SSD disks:'
                    . Dumper \%partner_pool_disks;
            }
        );

        my @partner_pool0_ssd_disks = @{$partner_pool_disks{0}};
        my @partner_pool1_ssd_disks = @{$partner_pool_disks{1}};

        if (   (scalar @partner_pool0_ssd_disks < $disk_count)
            && (scalar @partner_pool1_ssd_disks > 0)) {
            for (
                my $i = 0;
                (scalar @partner_pool0_ssd_disks + $i) < $disk_count;
                $i++
                ) {
                $nodescope_obj->disk_assign(
                    'disk-list'           => $partner_pool1_ssd_disks[$i],
                    force                 => 'true',
                    'nodescope-node-name' => $partner_name,
                    pool                  => 0,
                );
            }
        }
        @all_ssd_disks = NACL::CS::RaidConfigInfoListDisk->fetch(
            command_interface => $command_interface,
            allow_empty       => 1,
            filter            => {
                spare             => 1,
                disk_type         => 'SSD',
                logical_disk_type => 'physical_disk',
                pool              => 0,
            },
            'nodescope-node-name' => $partner_name,
        );

    } elsif (
        (      (scalar @spare_ssd_disks < $disk_count)
            && (scalar @spare_ssd_disks > 1)
        )
        && (   (scalar @partner_ssd_disks < $disk_count)
            && (scalar @partner_ssd_disks > 1))
        && (((scalar @spare_ssd_disks) + (scalar @partner_ssd_disks))
            >= $disk_count)
        ) {
        for (my $i = 0; (scalar @spare_ssd_disks + $i) < $disk_count; $i++) {
            $nodescope_obj->disk_assign(
                'disk-list'           => $partner_ssd_disks[$i]->name(),
                force                 => 'true',
                'owner-name'          => $command_interface->node(),
                'nodescope-node-name' => $partner_name,
                pool                  => 0
            );
        }

        foreach my $disk (@spare_ssd_disks) {
            if ($disk->pool() == 1) {
                $nodescope_obj->disk_assign(
                    'disk-list'  => $disk->name(),
                    force        => 'true',
                    'owner-name' => $command_interface->node(),
                    pool         => 0
                );
            }
        }

        @all_ssd_disks = NACL::CS::RaidConfigInfoListDisk->fetch(
            command_interface => $command_interface,
            allow_empty       => 1,
            filter            => {
                spare             => 1,
                disk_type         => 'SSD',
                logical_disk_type => 'physical_disk',
                pool              => 0,
            },
        );
    }    # end  if (scalar @spare_ssd_disks < $disk_count
    my @ssd_disk_list;
    foreach my $disk (@all_ssd_disks) {
        push(@ssd_disk_list, $disk->cluster_name());
    }
    $Log->debug(
        sub {
            "_determine_disk_list: SSD disks from both pools: @ssd_disk_list";
        }
    );
    my $retry=3;
    RETRY:
    if (scalar(@ssd_disk_list) >= $disk_count) {
        my @disk_lists = @ssd_disk_list[0 .. $disk_count - 1];
        return \@disk_lists;
    } else {
        if ( $retry > 0 ){
	    ## some delay might be required before the disks can be re-assigned
	    Tharn::snooze(5);
	    $retry-- ;
	    goto RETRY;
	}
        $Log->exit() if $may_exit;
        NATE::BaseException->throw(
            'Sufficient SSD disks are not available in both pool');
    }
} ## end sub _determine_disk_list

sub _default_rng_suffix {
    return 'pool';
}

1;
