#
#   Copyright (c) 2012 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Zone Task Module for Switch Zone
## @author nayankumar.patel@netapp.com
## @status shared
## @pod here

package NACL::STask::Switch::Zone;

use strict;
use warnings;

use base qw(NACL::C::Switch::Zone NACL::STask::STask);

use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Params::Validate qw(validate SCALAR BOOLEAN SCALARREF);
use NACL::ComponentUtils qw(_optional_scalars);

# Import Exceptions
use NACL::APISet::Exceptions::ResponseException qw(:try);
use NACL::C::Exceptions::Switch::ElementAlreadyExists;
use NACL::C::Exceptions::Switch::DoesNotExists;

=head1 NAME

NACL::STask::Switch::Zone

=head1 DESCRIPTION

C<NACL::STask::Switch::Zone> provide methods to create Zone on FC switch.

It is created on top of C<NACL::C::Switch::Zone> component. Since it is 
derived class of C<NACL::C::Switch::Zone>, we can use all the methods of 
C<NACL::C::Switch::Zone> from the object of this task.

=head1 ATTRIBUTES

=head2 command_interface

(Required) A component object that represents the host to which to send 
commands.
See L<NACL::C::Component::command_interface|lib-NACL-C-Component-pm/command_interface>.

=head2 zone-name

(Required) As C<NACL::C::Switch::Zone>. Name of zone to create.

=cut


=head1 METHODS

=head2 create

my $zone = NACL::STask::Switch::Zone->create(
    command_interface   => $command_interface,
    "zone-name"           => my_zone,
    "vsan-id"             => vsan_id,
    "member"              => zone_members,
);

(Class Method) create requested Zone on switch. Not all the
options are mandatory. What options are required for a particular
type of SET is mentioned in C<options> section below.
if zone creation fails, it will throw an exception.

=over

=item Options

=over

=item C<< nacltask_if_exists => $action >>

(Optional, defaults to "die") 
Specifies an action to be taken if the zone already exists.

=over

=item C<die>

When set to "die", an exception is thrown if the zone already exists.

=item C<reuse>

If a zone of the same name already exists, reuse it.

=item C<purge>

If a zone of the same name already exists, purge it and create a new one.

=back

=item C<< nacltask_verify => 0|1 >>

(Optional, defaults to 0)

When set to 0, verification is not performed. When set to 1, it is verified that 
the created zone has all of the attributes specified.

Note that it is also possible to perform the verification later by calling
L<taskverify_create|lib-NACL-STask-STask-pm/taskverify_create>.
           
=item C<< _was_created => \$scalar >>

(Optional)
When this option is provided a reference to a scalar variable, the variable
gets filled in with a boolean value describing whether the zone
was found (value will be 0; this scenario is possible when 
if_exists => "reuse") or whether the zone was created (value will be 1).
This is necessary to determine whether the zone needs to be
cleaned up later.

    my $was_created;
    my $zone_obj = NACL::STask::Switch::Zone->create(
        if_exists    => 'reuse',
        _was_created => \$was_created,
        %other_opts
    );
    # Operate on $zone_obj here
    # ...
    # Now determine whether to clean up the zone, since we're not sure
    # whether we reused an existing zone or created a new one
    if ($was_created) {
        # New zone was created. Clean it up.
        $zone_obj->purge();
    }

=item other options

All of the other various options are described in
L<< NACL::C::Switch::Zone->create|lib-NACL-C-Switch-Zone-pm/create >>

=back

=back

=cut

sub create {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            nacltask_if_exists => $pkg->_if_exists_validate_spec(),
            nacltask_verify => { type => SCALAR,    default  => 0 },
            _was_created    => { type => SCALARREF, optional => 1 },
        },
        allow_extra => 1
    );

    my $if_exists 
        = delete $opts{nacltask_if_exists};
        
    my $command_interface = $opts{command_interface};
    my $verify            = delete $opts{nacltask_verify};
    my $was_created       = delete $opts{_was_created};
    $$was_created = 0;

    my $self;
CREATE: {
        use warnings qw(exiting);
        try {
            $self = $pkg->SUPER::create(%opts);
            $$was_created = 1;
            if ($verify) {      
                 my @members = split /;/, $opts{'member'};
                 $opts{members} = \@members;
                 delete $opts{member};
                 $pkg->taskverify_create(%opts);
            }
        }
        catch NACL::C::Exceptions::Switch::ElementAlreadyExists with {
            my $exception = shift;
            $Log->trace('Zone already exists');
            $self = $pkg->_element_exists_handler(
                create_opts        => \%opts,
                nacltask_if_exists => $if_exists,
                exception          => $exception
            );
            if ( !$self ) {
                no warnings qw(exiting);
                redo CREATE;
            }
             
        };
    }
    
    $Log->exit() if $may_exit;
   
    return $self;
}

=head2 purge

    NACL::STask::Switch::Zone->purge(
        command_interface  => $command_interface,
        'zone-name'        => $zone,
        nacltask_if_purged => $action, #default "die"
    );

      or

    $zone_obj->purge();

(Class or instance method)

This method deletes a zone on the given switch. It also waits
for the Zone to be deleted before returning (this can be
configured using the C<nacltask_wait> option)

=over

=item Options

=over

=item C<< nacltask_if_purged=> 'die' | 'pass' >>

(Optional, defaults to 'die')

If "die", exception is thrown when Zone to be deleted does not exist.
If "pass", exits quietly when Zone to be deleted does not exist.

=back

=back

=cut

sub purge {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my $pkg        = $pkg_or_obj->get_package_name();
    my %opts       = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            nacltask_if_purged => $pkg_or_obj
                            ->_if_action_completed_already_validate_spec(),
        },
        allow_extra => 1,
    );
    my $command_interface = $opts{command_interface};
    my $wait              = delete $opts{nacltask_wait};

    # if nacltask_if_purged is not defined default it to die.
    my $if_purged 
        = delete $opts{nacltask_if_purged};

    # Call the component delete method blindly
    # if the zone does not exist then it would 
    # throw an exeception which is handled below.
    try {
        # Delete zone
        $pkg->delete( %opts );
    } catch NACL::C::Exceptions::Switch::DoesNotExists with {
        my $error = shift;
        $Log->trace("Zone does not exist.");
        if ( $if_purged =~ /die/ ) {
            $Log->exit() if $may_exit;
            $error->throw();
        }
    };
    
    $Log->exit() if $may_exit;
    
}

1;
