#
#   Copyright (c) 2001-2011 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary common routine for Interface and PortChannel Task Module for Switch
## @author nayankumar.patel@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::STask::Switch::_Interface

=head1 DESCRIPTION

common code for Interface and PortChannel STask implementation.

=cut

package NACL::STask::Switch::_Interface;

use strict;
use warnings;

use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Params::Validate qw(validate_with SCALAR ARRAYREF UNDEF SCALARREF);
use NACL::ComponentUtils qw(Dumper);
use NATE::Exceptions::Argument qw(:try);

BEGIN {
    use Exporter qw(import);
    our %EXPORT_TAGS = (
        all => [
            qw(_configure _configure_cisconetwork_cli _add_to_vlan
            _add_to_vlan_cisconetwork_cli _remove_from_vlan 
            _remove_from_vlan_cisconetwork_cli)
        ]
    );
    our @EXPORT_OK = @{ $EXPORT_TAGS{all} };
}

#This is a AUTOLOAD method from NACL/C/Switch/Interface.pm and PortChannel.pm
#please check POD for configure() method in Interface.pm or PortChannel.pm
sub _configure {
    $Log->enter() if $may_enter;
    my $pkg = shift;
    my %opts = validate_with(
        params      => \@_,
        spec        => {
            speed        => { type => SCALAR, optional => 1 },
            duplex       => { type => SCALAR, optional => 1 },
            mtu          => { type => SCALAR, optional => 1 },
            'fc-send'    => { type => SCALAR, optional => 1 },
            'fc-receive' => { type => SCALAR, optional => 1 },
            switchport   => { type => SCALAR, optional => 1 },
            status       => { type => SCALAR, default  => 'up' },
        },
        allow_extra => 1
    );
    
    $Log->exit(); 
    return $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => "_configure_cisconetwork_cli",
                interface => "CLI",
                set       => "CiscoNetwork"
            },
        ],
    );
}    #end of sub configure

sub _configure_cisconetwork_cli {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );

    my %c_opts;

    $pkg->_hash_move(
        source => \%opts,
        target => \%c_opts,
        move   => [qw(interface-type interface-num command_interface)],
    );

    my $apiset = delete $opts{apiset};

    $Log->debug(
        sub {
            "Options to configure:\n" . Dumper( \%opts );
        }
    );

    ## set the cli for requested interface.
    $pkg->_set_ifc_cisconetwork_cli( %c_opts, apiset => $apiset );

    ##set configuration if requested
    if ( defined( $opts{switchport} ) and $opts{switchport} ) {
        $apiset->switchport();
    }

    if ( defined( $opts{speed} ) ) {
        $apiset->speed( speed => $opts{speed} );
    }

    if ( defined( $opts{duplex} ) ) {
        $apiset->duplex( duplex => $opts{duplex} );
    }

    if ( defined( $opts{mtu} ) ) {
        if (   $c_opts{command_interface}->hostrec()->hosttype =~ /nxos/
            && $c_opts{command_interface}->hostrec()->model() !~ /9332/ ) {
            $Log->warn("Setting mtu size is not supported on Nexus OS");
        } else {
            if (   $c_opts{command_interface}->hostrec()->model() =~ /4948/
                && $opts{mtu} == 9216 ) {
                $Log->comment(
                    "Cisco Ethernet 4948 switch does not support MTU 9216, configuring MTU to 9198"
                );
                $opts{mtu} = 9198;
            }
            $apiset->mtu( mtu => $opts{mtu} );
        }
    }
    if ( defined( $opts{'fc-send'} ) ) {
        if ( $c_opts{command_interface}->hostrec()->model() =~ /4948/ ) {
            $Log->comment(
                "Cisco Ethernet 4948 switch does not support flowcontrol send configuration"
            );
        } else {
            $apiset->flowcontrol(
                'flow-direction' => 'send',
                'flow-operation' => $opts{'fc-send'}
            );
        }
    }

    if ( defined( $opts{'fc-receive'} ) ) {
        $apiset->flowcontrol(
            'flow-direction' => 'receive',
            'flow-operation' => $opts{'fc-receive'}
        );
    }

    if ( $opts{status} =~ /up/i ) {
        $apiset->shutdown( deconfigure => 1 );
    } elsif ( $opts{status} =~ /down/i ) {
        $apiset->shutdown();
    } else {
        $Log->exit() if $may_exit;
        NATE::Exceptions::Argument->throw("status should be 'up' or 'down'");
    }

    $Log->exit() if $may_exit;
} ## end sub _configure_cisconetwork_cli


#This is a AUTOLOAD method from NACL/C/Switch/Interface.pm and PortChannel.pm
#please check POD for add_to_vlan() method in Interface.pm or PortChannel.pm
sub _add_to_vlan {
    $Log->enter() if $may_enter;
    my $pkg = shift;
    my %opts = validate_with(
        params          => \@_,
        spec => {
            vlan                => { type => SCALAR, default => 'all' },
            'native-vlan'       => { type => SCALAR, default => 1 },
            nacltask_append     => { type => SCALAR, default => 1 },
        },
        allow_extra => 1
    );
    
    $Log->exit() if $may_exit;
    
    return $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => "_add_to_vlan_cisconetwork_cli",
                interface => "CLI",
                set       => "CiscoNetwork"
            },
        ],
    );
}    #end of sub add_to_vlan

sub _add_to_vlan_cisconetwork_cli {
    $Log->enter() if $may_enter;
    my $pkg = shift;
     my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );
     
    my %c_opts;

    $pkg->_hash_move(
        source => \%opts,
        target => \%c_opts,
        move   => [qw(interface-type interface-num command_interface)],
    );
    
    my $apiset = delete $opts{apiset}; 

    my $hosttype = $c_opts{command_interface}->hostrec()->hosttype();
    $Log->debug(
        sub {
            "Options to add_to_vlan:\n" . Dumper(\%opts);
        }
    );

    $Log->debug("Host type is $hosttype");

    ## set the cli for requested interface.
    $pkg->_set_ifc_cisconetwork_cli( %c_opts, apiset => $apiset );
    
    if ( $hosttype =~ /ios/i ) {
        $apiset->switchport();
        $apiset->switchport_trunk( encapsulation => 'dot1q' );
    }

    $apiset->switchport( mode => 'trunk' );
    $apiset->switchport_trunk( 
        'native-vlan' => 1, 
        'vlan-list' => $opts{'native-vlan'}
    );
    
    if ( $opts{vlan} =~ /^all$/i ) {
        $apiset->switchport_trunk(
            'allowed-vlan' => 1,
            'vlan-list' => $opts{vlan}
        );
    } elsif ( $opts{nacltask_append} ) {
        $apiset->switchport_trunk(
            'allowed-vlan' => 1,
            add => 1,
            'vlan-list' => "$opts{vlan}"
        );
    } else {
        $apiset->switchport_trunk(
            'allowed-vlan' => 1,
            'vlan-list' => "$opts{'native-vlan'},$opts{vlan}"
        );
    }

    $Log->exit() if $may_exit;
}  

#This is a AUTOLOAD method from NACL/C/Switch/Interface.pm and PortChannel.pm
#please check POD for remove_from_vlan() method in Interface.pm or PortChannel.pm
sub _remove_from_vlan {
    $Log->enter() if $may_enter;
    my $pkg = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => { vlan => { type => SCALAR, default => 'all' } },
        allow_extra => 1
    );
    
    $Log->exit() if $may_exit;
    
    return $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => "_remove_from_vlan_cisconetwork_cli",
                interface => "CLI",
                set       => "CiscoNetwork"
            },
        ],
    );
}    #end of sub delete_from_vlan

sub _remove_from_vlan_cisconetwork_cli {
    $Log->enter() if $may_enter;
    my $pkg = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );
     
    my %c_opts;

    $pkg->_hash_move(
        source => \%opts,
        target => \%c_opts,
        move   => [qw(interface-type interface-num command_interface)],
    );
    
    my $apiset = delete $opts{apiset}; 

    my $hosttype = $c_opts{command_interface}->hostrec()->hosttype();
    $Log->debug(
        sub {
            "Options to remove_from_vlan:\n" . Dumper(\%opts);
        }
    );

    $Log->debug("Host type is $hosttype");

    ## set the cli for requested interface.
    $pkg->_set_ifc_cisconetwork_cli( %c_opts, apiset => $apiset );
    
    if ( $opts{vlan}  =~ /^all$/i ) {
        $apiset->switchport( mode => 'access' );
        $apiset->switchport_trunk(
            'allowed-vlan'  => 1,
            'vlan-list'     => $opts{vlan},
            deconfigure     => 1,
        );
        if ( $hosttype =~ /ios/i ) {
            $apiset->switchport_trunk( 
                encapsulation => 'dot1q', 
                deconfigure => 1 
            );
        }
    } else {
        $apiset->switchport_trunk(
            'allowed-vlan'  => 1,
            'vlan-list'     => $opts{vlan},
            remove          => 1,
        );
    }
    $Log->exit() if $may_exit;
}

1;
