#
# Copyright (c) 2001-2015 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary SystemHaInterconnectLink Task Module
## @author ng-interconnect-qa
## @status shared
## @pod here

package NACL::STask::SystemHaInterconnectLink;

use strict;
use warnings;
use base qw(NACL::C::Component NACL::STask::STask);
use List::Compare;
use NACL::C::SystemNode;
use NACL::CS::SystemHaInterconnectLinkInfo;
use NACL::ComponentUtils qw(Dumper);
use NACL::InstantComponent qw(instant_component);
use NATE::Log qw(log_global);
use NATE::BaseException qw(:try);
use Params::Validate qw(validate OBJECT SCALAR ARRAYREF UNDEF);
use Readonly;

my $Log = log_global();

# TASK related constants;
Readonly my $Method_Timeout    => 120;
Readonly my $Pollwait_Interval => 10;

use Class::MethodMaker [
    scalar => [
        { -type => 'NACL::C::CommandInterface::ONTAP' },
        'command_interface'
    ],
    scalar => 'node',
    array  => 'links',
    scalar => 'method_timeout',
    scalar => 'pollwait_interval',
];

=head1 NAME

NACL::STask::SystemHaInterconnectLink

=head1 DESCRIPTION

C<NACL::STask::SystemHaInterconnectLink> provides a well-defined set of methods
to turn on/off the Interconnect link(s) on a system.

It uses the C<NACL::CS::SystemHaInterconnectLinkInfo> to derive information,
and provides methods that are more in the scope of individual
system ha interconnect link-related commands.
See C<NACL::CS::SystemHaInterconnectLinkInfo> for details.

=head1 ATTRIBUTES

=head2 command_interface

(Required) A component object that represents the host to which to send commands.
See L<NACL::C::Component::command_interface|lib-NACL-C-Component-pm/command_interface>

=head1 METHODS

=for :later

=head2 new

    $stask_link_obj = NACL::STask::SystemHaInterconnectLink->new(
        command_interface => $command_interface,
    );

(Class method)

(For CMode) Creates a SystemHaInterconnectLink object.

P.S: This method does NOT perform any link on/off tasks.

=over

=item Options

=over

=item C<< node => $node_name >>

(Optional) This indicates the node which will be operated on.

If omitted, the method will operate on both the nodes in the HA pair for the
I<'command_interface'> specified.

=item C<< links => @links >>

(Optional) (ArrayRef) This option accepts a list of link numbers to execute against.

If not specified, this option will default to a list containing all available link numbers.

=item C<< method_timeout => $time_to_wait >>

(Optional) Time in seconds to wait for the verification of operation.

If not specified, I<'method_timeout'> defaults to 600 seconds (10 minutes).

=item C<< polling_interval => $interval >>

(Optional) This is the interval (in seconds) after which the value of the attribute will be polled.

If not specified, I<'polling_interval'> defaults to 10 seconds.

=back

=back

=over

=item Exceptions

=over

=item C<NACL::Exceptions::InvalidChoice exception>

This type of exception is thrown when an invalid choice is passed for the I<'links> option.

Examples of invalid choices include 'active', 'inactive', invalid link number, etc.

=back

=back

=head2 on

    NACL::STask::SystemHaInterconnectLink->on(
        command_interface => $command_interface,
    );

=for :later

=begin :instance

    # or

    $stask_link_obj->on();  (See ::new)

=end :instance

(Class method) Turns HA-Interconnect link(s) on.

(For CMode) This method will issue the appropriate command to turn on a single,
multiple or all HA-Interconnect links on the requested node(s).

By default, it will turn on all the links on both nodes in the HA Pair that
the I<'command_interface'> belongs to.

In addition, this method, by default, will also verify that the HA interconnect
link was actually turned on.

The method will not complain if the link is already 'up' on the requested node.

The method will throw an exception if there is a failure while turning on a link
on the requested node(s).

=over

=item Options

=over

=item C<< node => $node_name >>

(Optional) (Scalar) This indicates the node which will be operated on.

If omitted, the method will operate on both the nodes in the HA pair for the
I<'command_interface'> specified.

=item C<< links => @links >>

(Optional) (ArrayRef) This option accepts a list of link numbers to execute against.
By default all available links will be operated on.

Example:

    NACL::STask::SystemHaInterconnectLink->on(
        command_interface => $command_interface,
        node              => $node_name,
        links             => [0],
    );

=item C<< method_timeout => $time_to_wait >>

(Optional) Time in seconds to wait for the verification after operation.

If not specified, I<'method_timeout'> defaults to 600 seconds (10 minutes).

=item C<< polling_interval => $interval >>

(Optional) This is the interval (in seconds) after which the value of the attribute will be polled.

If not specified, I<'polling_interval'> defaults to 10 seconds.

=item C<< nacltask_verify => $boolean >>

(Optional) This flag is used to verify if the link was actually turned on.

By default, the flag is set to 1, meaning verification will take place.

=back

=back

=over

=item Exceptions

=over

=item C<NACL::Exceptions::InvalidChoice exception>

This type of exception is thrown when an invalid choice is passed for the I<'links'> option.

=item C<NACL::Exceptions::VerifyFailure>

This type of exception is thrown when verification fails for the link that was turned on.

=back

=back

=cut

sub on {
    $Log->enter();
    my ($pkg , %opts ) = @_;

    $opts{action} = 'on';
    $pkg->_on_off(%opts);

    $Log->exit();
} # End on()

=head2 off

    NACL::STask::SystemHaInterconnectLink->off(
        command_interface => $command_interface,
    );

=for :later

=begin :instance

    # or

    $stask_link_obj->off();  (See ::new)

=end :instance

(Class method) Turns HA-Interconnect link(s) off.

(For CMode) This method will issue the appropriate command to turn off a single,
multiple or all HA-Interconnect links off the requested node(s).
By default, it will turn off all the links on both nodes in the HA Pair that
the I<'command_interface'> belongs to.

In addition, this method will also verify that the ha interconnect link actually went down.

The method will not complain if the link is already 'down' on the requested node.

The method will throw an exception if there is a failure while turning off a link
on the requested node(s).

=over

=item Options

=over

=item C<< node => $node_name >>

(Optional) (Scalar) This indicates the node which will be operated on.

If omitted, the method will operate on both the nodes in the HA pair for the
I<'command_interface'> specified.

=item C<< links => @links >>

(Optional) (Arrayref) This option accepts a list of link numbers to execute against.
By default all available links will be operated on.

Example:

    NACL::STask::SystemHaInterconnectLink->off(
        command_interface => $command_interface,
        node              => $node_name,
        links             => [0],
    );

    Acceptable list of link number options can also include 'active' or 'inactive'.

=item C<< method_timeout => $time_to_wait >>

(Optional) Time in seconds to wait for the verification after operation.

If not specified, I<'method_timeout'> defaults to 600 seconds (10 minutes).

=item C<< polling_interval => $interval >>

(Optional) This is the interval (in seconds) after which the value of the attribute will be polled.

If not specified, I<'polling_interval'> defaults to 10 seconds.

=back

=back

=over

=item Exceptions

=over

=item C<NACL::Exceptions::InvalidChoice exception>

This type of exception is thrown when an invalid choice is passed for the I<'links'> option.

=item C<NACL::Exceptions::VerifyFailure>

This type of exception is thrown when verification fails for the link that was turned off.

=back

=back

=cut

sub off {
    $Log->enter();
    my ($pkg , %opts ) = @_;

    $opts{action} = 'off';
    $pkg->_on_off(%opts);

    $Log->exit();
} # End off()

################################################################################
#                              LOCAL ROUTINES
################################################################################

################################################################################
# Function    : _action_on_link()
# Description :
#     Turns the link 'on' or 'off'
# Arguments   :
#     command_interface - NACL::C::SystemNode - (Required) CI to work on
#     action            - Scalar              - (Required) on|off
#     link              - Scalar              - (Required) HA IC link number
#     node              - Scalar              - (Required) Node name
# Returns     : None
################################################################################
sub _action_on_link {
    $Log->enter();
    my ($action, $cmd, $apiset, $ci, $err, $link, $node, %opts, $pkg, $verify,);
    ($pkg, %opts ) = @_;
    $action        = $opts{action};
    $apiset        = $opts{command_interface}->apiset();
    $link          = $opts{link};
    $node          = $opts{node};
    $verify        = delete($opts{nacltask_verify});
    $err           = 'Link ' . $link . '(.*)' . ($action eq 'on'? 'up':'down');

    $Log->log("== Turn link $link '$action' on node $node ==");
    $cmd = "system_ha_interconnect_link_$action";

    try {
        $apiset->$cmd(node => $node, link => $link, );
    } catch NACL::APISet::Exceptions::CommandFailedException with {
        my $ex  = shift;
        if ($ex->text() =~ /$err/) {
            $err =~ s/\(\.\*\)/ is already /;
            $Log->log("== $err. Proceed... ==");
        } else {
            $Log->log('== Unknown exception caught. Re-throwing... ==');
            $Log->exit();
            $ex->throw();
        }
    };

    $Log->exit();
    return;
} # End _action_on_link()

################################################################################
# Function    : _get_links()
# Description :
#     Get all available HA Interconnect link numbers for the HA pair belonging
#     to the 'command_interface'.
# Arguments   :
#     command_interface - NACL::C::SystemNode - (Required) CI to work on
#     state             - Scalar              - (Optional) active|inactive
# Returns     : (ArrayRef) List of HA Interconnect link numbers
# Exception   : NATE::Exceptions::Argument - It requested links not found
################################################################################
sub _get_links {
    $Log->enter();
    use NACL::Expectation;
    my (
        $ci, @links, @objs, %opts, $pkg, @requested_fields, $state,
        $want_active,
    );
    ($pkg, %opts) = @_;
    $ci           = $opts{command_interface};
    $state        = $opts{state} || '';

    @requested_fields = qw(port-number);
    if ($state =~ /active/) {
        push(@requested_fields, 'is-active-link');
        $want_active = ($state eq 'active') ? 'true' : 'false';
    }

    @objs = NACL::CS::SystemHaInterconnectLinkInfo->fetch(
        command_interface => $ci,
        requested_fields  => [@requested_fields],
    );

    # Depending on which links are needed, add them to the list
    foreach my $link (@objs) {
        if ($state =~ /active/) {
            if ($link->is_active_link() eq $want_active) {
                push(@links, $link->port_number());
            }
        } else {
            push(@links, $link->port_number());
        } # End if-else
    }

    # If 'links' is empty, throw an exception
    if (!scalar(@links)) {
        $Log->exit();
        NATE::Exceptions::Argument->throw(
            "$state link(s) not found."
        );
    }

    $Log->exit();
    return \@links;
} # End sub _get_links()

################################################################################
# Function    : _get_nodes()
# Description :
#     Get all the nodes belonging to the HA pair 'command_interface' belongs to.
# Arguments   :
#     command_interface - NACL::C::SystemNode - (Required) Node to work on
# Returns     : (ArrayRef) List of node names
# Exceptions  : NACL::Exceptions::HighAvailabilityError - If no partner found
################################################################################
sub _get_nodes {
    $Log->enter();
    my ($ci, @nodes, %opts, $pkg, );
    $pkg  = shift;
    %opts = @_;
    $ci   = $opts{command_interface};
    push @nodes, $ci->name() ;
    push @nodes, NACL::STask::Node->get_partner(
        command_interface => $ci,
    );
    $Log->exit();
    return \@nodes;
} # End sub _get_nodes()

################################################################################
# Function    : _on_off()
# Description :
#     Turns HA Interconnect link(s) on/off and verifies action if
#     nacltask_veriy is set to '1'
# Arguments   :
#     command_interface - NACL::C::SystemNode - (Required) CI to work on
#     action            - Scalar          - (Required) on|off
#     links             - ArrayRef        - (Optional) List of HA Interconnect
#                                           link numbers to be operated on
#     node              - Scalar|ArrayRef - (Optional) List of nodes
#     pollwait_interval - Scalar          - (Optional) Time in seconds to wait
#                                           before polling
#     method_timeout    - Scalar          - (Optional) Time in seconds before
#                                           timeout
# Returns     : None
################################################################################
sub _on_off {
    $Log->enter();
    my($action, @args, $ci, $links, $node, %opts, $pkg, $verify, );
    ($pkg, @args) = @_;
    %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            links             => {type => ARRAYREF, optional => 1},
            node              => {type => SCALAR|ARRAYREF|UNDEF, optional => 1},
            pollwait_interval => {optional => 1, default => $Pollwait_Interval},
            method_timeout    => {optional => 1, default => $Method_Timeout},
            nacltask_verify   => {optional => 1, default => 1 },
        },
        allow_extra     => 1,
    );

    $ci     = $opts{command_interface};
    $action = $opts{action};
    $verify = $opts{nacltask_verify};
    $links  = delete($opts{links});
    $node   = delete($opts{node});

    # If 'links' is not defined, populate with all available link numbers for
    # the HA pair for the given 'command_interface'
    if (!defined $links) {
        $links = $pkg->_get_links(command_interface => $ci,);
    } else {
        # If 'links' is defined, verify if the links are valid
        $links = $pkg->_verify_links(
            command_interface => $ci,
            action            => $action,
            links             => $links,
        );
    }

    # If 'node' is not defined, by default, it will be populated with the
    # HA pair nodes belonging to the 'command_interface' specified
    if (!defined $node) {
        $node = $pkg->_get_nodes(command_interface => $ci,);
    }

    # If $node is a scalar convert to ArrayRef for simplicity
    if (ref($node) ne 'ARRAY') {
        $node = [$node];
    }

    # Create component object for 'status show' if nacltask_verify is '1'
    # Let's do this here so that the component is installed only once
    if ($verify) {
        instant_component(
            command_interface =>  $ci,
            component_name    => 'NACL::C::SystemHaInterconnectStatus',
        );
    }

    # For each of the link numbers in the list of valid links,
    # 1) Perform the on/off action on each node
    # 2) If nacltask_verify is '1',
    #    verify if the link is in the expected state for each node
    foreach my $link (@$links) {
        foreach my $node_name (@$node) {
            $pkg->_action_on_link(
                node   => $node_name,
                link   => $link,
                %opts,
            );
        }
        if ($verify) {
            foreach my $node_name (@$node) {
                $pkg->_verify_action(
                    node   => $node_name,
                    link   => $link,
                    %opts,
                );
            }
        } # End if ($verify)
    }

    $Log->exit();
    return;
} # End sub _on_off()

################################################################################
# Function    : _verify_action()
# Description :
#     Verify the on/off action performed on a link
# Arguments   :
#     command_interface - NACL::C::SystemNode - (Required) CI to work on
#     action            - Scalar    - (Required) on|off
#     links             - Scalar    - (Required) List of HA Interconnect Link
#                                     numbers
#     node              - Scalar    - (Required) Node name
#     pollwait_interval - Scalar    - (Required) Time in seconds before polling
#     method_timeout    - Scalar    - (Required) Time in seconds before timeout
# Returns     : None
# Exception   : NACL::Exceptions::VerifyFailure - If verification fails
################################################################################
sub _verify_action {
    $Log->enter();
    my (
        $action, $apiset, $ci, $err, $link, $method_timeout, $node, %opts,
        $pkg, $pollwait,
    );
    ($pkg, %opts)   = @_;
    $action         = $opts{action};
    $ci             = $opts{command_interface};
    $link           = $opts{link};
    $node           = $opts{node};
    $method_timeout = $opts{method_timeout};
    $pollwait       = $opts{pollwait_interval};
    $action         = $action eq 'on'? 'up':'down';

    $Log->log("== Verify if link $link is now '$action' on node $node ==");
    try {
        NACL::C::SystemHaInterconnectStatus->wait_on_attributes(
            command_interface => $ci,
            node              => $node,
            attributes        => [
                {
                    attribute_to_check => 'link-status',
                    till_value         => qr/-|$action/i
                },
                {
                    attribute_to_check => "link$link-status",
                    till_value         => qr/-|$action/i
                }
            ],
            'method-timeout'  => $method_timeout,
            polling_interval  => $pollwait,
        );
    } catch NACL::APISet::Exceptions::TimeoutException with {
        my $exception = shift;
        $Log->exit();
        NACL::Exceptions::VerifyFailure->throw($exception->text());
    };

    $Log->exit();
    return;
} # End _verify_action()

################################################################################
# Function    : _verify_links()
# Description :
#     Verify if HA Interconnect link numbers are valid for the platform
# Arguments   :
#     command_interface - NACL::C::SystemNode - (Required) CI to work on
#     action            - Scalar    - (Required) - on|off
#     links             - ArrayRef  - (Required) - List of HA IC Link numbers
# Returns     : (ArrayRef) List of HA Interconnect link numbers
# Exceptions  : NATE::Exceptions::Argument - If links numbers are invalid
################################################################################
sub _verify_links {
    $Log->enter();
    require NACL::CS::Metrocluster;
    require NACL::CS::SystemHaInterconnectInfo;
    my (
        $action, @all_links, $ci, $err, @invalid_links, $lc, $links, $mcc,
        %opts, $pkg, $list, $valid_links,
    );
    ($pkg, %opts) = @_;
    $action       = $opts{action};
    $ci           = $opts{command_interface};
    $links        = $opts{links};
    $valid_links  = $pkg->_get_links(command_interface => $ci);

    # Now, if the action is 'off', we can allow only 'active' in the list of
    # valid link numbers if,
    # - If the system has only one link,
    # - If the system is an MCC
    if ($action eq 'off') {
        $mcc = NACL::CS::Metrocluster->fetch(
                command_interface => $ci,
                requested_fields  => [qw(local-configuration-state)],
            )->local_configuration_state();
        if (@$valid_links == 1 || $mcc eq 'configured') {
            push(@$valid_links, 'active');
        } else {
            push(@$valid_links, ('active', 'inactive'));
        }
    }

    # Compare the list of valid links and the passed in list of link numbers
    $lc = List::Compare->new($valid_links, $links);

    # Get the link numbers that exist only in the passed in list of link numbers
    @invalid_links = $lc->get_Ronly;

    # If invalid link numbers are found, throw an exception
    if (scalar(@invalid_links)) {
        $err = join(', ',@invalid_links);
        $Log->exit();
        NATE::Exceptions::Argument->throw("Invalid link(s): [$err]");
    }

    # For 'active' or 'inactive' in the list,
    # add corresponding link numbers to the list of valid links
    foreach my $link (@$links) {
        if ($link =~ /active|inactive/) {
            $list = $pkg->_get_links(command_interface => $ci, state => $link);
            @all_links = List::Compare->new(\@all_links,$list)->get_union();
            next;
        }

        # Add to list of valid links if link number does not already exist
        push(@all_links, $link) unless grep{$_ == $link} @all_links;
    }

    $Log->exit();
    return \@all_links;
} # End sub _verify_links()
