#
# Copyright (c) 2001-2013 NetApp, Inc., All Rights Reserved
# Any use, modification, or distribution is prohibited
# without prior written consent from NetApp, Inc.
#

## @summary SystemLicense Task Module
## @author dl-nacl-dev@netapp.com
## @status shared
## @pod here

package NACL::STask::SystemLicense;

use strict;
use warnings;

use base qw(NACL::C::SystemLicense NACL::STask::STask);
use NACL::STask::SystemLicense::Default;
use Params::Validate qw(validate_with SCALAR SCALARREF ARRAYREF BOOLEAN HASHREF);
use NACL::Exceptions::NoElementsFound qw(:try);
use NACL::APISet::Exceptions::ResponseException;
use NACL::Exceptions::VerifyFailure;
use NACL::C::Exceptions::SystemLicense::DoesNotExist;
use NACL::ComponentUtils qw(Dumper);
use NACL::C::Exceptions::Component;
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();

use Class::MethodMaker [ 'scalar' => '_was_added' ];

=head1 NAME

NACL::STask::SystemLicense

=head1 DESCRIPTION

"NACL::STask::SystemLicense" provides a number of well-defined but
potentially complex or multi-step methods related to Licenses
in ONTAP.

This builds on top of, and is a derived class of "NACL::C::SystemLicense",
and so it also provides methods that are more in the scope of
individual License-related commands. For example, method find
from NACL::C::SystemLicense can be used directly.

See "NACL::C::SystemLicense" for details.  This also means that a
"NACL::STask::SystemLicense" object may generally be used in place of
a component object.

=head1 LICENSING FORMATS

See the L<Component|lib-NACL-C-SystemLicense-pm/LICENSING_FORMATS> to understand
what "Licensing V1" and "Licensing V2" mean.

=head1 ATTRIBUTES

=head2 command_interface

(Required) As C<NACL::C::SystemLicense>. A component object that
represents the host to which to send commands.

=head2 package

The name of the license package. This attribute is only applicable for licensing v2.

=head2 serial_number

Cluster Base license serial number. This attribute is only applicable for licensing v2.

=head2 feature

The name of the license feature, used to identify it uniquely. This attribute is only
applicable for licensing v1 (SystemLicense feature names are assumed to be unique
within a node or within a cluster).

=head1 METHODS

=head2 add

    # For Licensing V1, use argument "feature":
    my $stask = NACL::STask::SystemLicense->add(
        command_interface => $command_interface,
        feature           => $feature,
        nacltask_verify   => 1,           # default is 0
        _was_added        => \$was_added,
        nacltask_if_licensed => "pass",   # default is pass
        %other_options
    );

    # For Licensing V2, use argument "package":

    # Adding a single package can be done by providing it as a scalar
    my $stask = NACL::STask::SystemLicense->add(
        command_interface => $command_interface,
        package           => $package,
        nacltask_verify   => 1,           # default is 0
        _was_added        => \$was_added,
        nacltask_if_licensed => "pass",   # default is pass
        %other_options
    );

    # Adding multiple packages can be done by providing package as an array
    # The return value will be an array of objects:
    my @stasks = NACL::STask::SystemLicense->add(
        command_interface   => $command_interface,
        package             => [ $package, $package, .. ],
        'license-code'      => [ $license_code1, $license_code2, .. ],
        %other_options
    );


    # Adding multiple packages can be done by providing package as an array
    # If license-code is not provided, like in this example, the STask
    # will automatically determine the license-code values for each package
    # The return value will be an array of objects:
    my @stasks = NACL::STask::SystemLicense->add(
        command_interface   => $command_interface,
        package             => [ $package, $package, .. ],
        %other_options
    );

	my $license_obj = NACL::STask::SystemLicense->add(
        command_interface => $Node,
        package           => $package,
        nacltask_to_cleanup => 1,
		nacltask_cleanup_manager => $CleanupObj,
	);

(Class method) Adds a feature license. This method provides
additional services beyond what's inherent to the product's
license commands.  Basically it will determine the license
code for the desired feature based on the model of the filer.
It also provides CMode/7Mode mapping for below features

    1. flexclone (CMode) <=> flex_clone (7Mode)
    2. For C-mode run:
       a> Map "snapmirror" to "snapmirror_dp".

       For 7-mode run:
       a> Map "snapmirror_dp" to "snapmirror".
       b> Map "snapmirror_ls" to "snapmirror".

=over

=item command_interface

(Required) As NACL::C::Node. A component object that represents
the host to which to send commands.

=item feature

(Required; applicable for Licensing V1) The name of the license feature, used to identify it
uniquely. (License feature names are assumed to be unique within
a node or within a cluster). The feature names accepted by this
task are case insensitive.

CMode features for licensing v1:

Base, Mirror, FlexVol_HPO, CIFS, SnapRestore, NFS, SnapMirror_DP
SnapMirror_LS (for BR), iSCSI, FCP, FlexClone

CMode features for licensing v2:

Base, CDMI, CIFS, FCP, FlexClone, iSCSI, NFS, SnapLock, SnapManagerSuite,
SnapMirror, SnapProtectApps, SnapRestore, SnapVault

7Mode license feature names are:

a_sis, cifs, cf, cf_remote, compression, compression, dafs,
disk_sanitization, env_exception, fcp, flash_cache, flex_clone,
flex_scale, flexcache_nfs, http, iscsi, nearstore_option,
nearstore_asis1, nfs, operations_manager, persistent_archive,
protection_manager, provisioning_manager, rapid_restore,
smdomino, smsql, snapdrive_unix, snapdrive_windows, snaplock,
snaplock_enterprise, snapmanagerexchange, snapmanager_hyperv,
snapmanager_oracle, snapmanager_sap, snapmanager_sharepoint,
snapmirror, snapmirror_sync, snapmover, snaprestore, snapvalidator,
storage_services, sv_linux_pri, sv_ontap_pri, sv_ontap_sec,
sv_unix_pri, sv_vi_pri, sv_windows_pri, syncmirror_local,
vfiler, vld, volcopy

7Mode features for licensing v2:

CDMI, CIFS, FCP, FlexClone, iSCSI, NFS, SnapLock, SnapManagerSuite,
SnapMirror, SnapProtectApps, SnapRestore, SnapVault

=item license-code

(Optional) SystemLicense code for the feature we want to
license. If specified, this method will try to add the license
code without verifying the mode, model and serial-number.
If not specified, the STask will automatically determine the appropriate
license-code value for the package(s) (this is done by looking up the
mode, model and serial-number)

=item Options

=item (Optional) B<< nacltask_verify => 1 >>

  - when set to 1, license addition will be verified.
  - when set to 0, the verification step is skipped.
  - by default its set to 0.

=item (Optional) B<< _was_added => \$scalar >>

When this option is provided a reference to a scalar variable,
the variable gets filled in with a boolean value describing
whether the license was added (value will be 1) or not (value
will be 0).

    my $was_added;
    my $license_obj = NACL::STask::SystemLicense->add(
        _was_added => \$was_added,
        %other_opts
    );

    # Operate on $license_obj here
    # ...

    # License was added. Clean it up.
    if ($was_added) {
        $license_obj->purge();
    }

    or

    my $was_added;
    my $license_obj = NACL::STask::SystemLicense->add(
        _was_added => \$was_added,
        %other_opts
    );

    # Operate on $license_obj here
    # ...

    # License was added. Clean it up.
    if ($license_obj->_was_added()) {
        $license_obj->purge();
    }

=item (Optional) B<< _was_licensed => 0 | 1 >>

    - when set to 1, checks if the features are already licensed.
    - when set to 0, the verification step is skipped.
    - by default its set to 0.

This option is applicable only when multiple features are provided.

    my @license_objs = NACL::STask::SystemLicense->add(
        command_interface   => $command_interface,
        feature             => "$feature1, $feature2,..",
        _was_licensed => 1,
        %other_options
    );

    # Operate on $license_obj here
    # ...

    # Licenses were added. Clean it up.
    foreach my $license_obj ( @license_objs ) {
        if ($license_obj->_was_added()) {
            $license_obj->purge();
        }
    }

=item (Optional) B<< nacltask_if_licensed => die | pass >>

  - If "die", exception is thrown when we try to add a license
    that isn't needed anymore.
  - If "pass", returns the license task object. Default is "pass".


=item (Optional) B<< nacltask_check_node_locked_license => 0 | 1 >>

  - when set to 0, node locked license check is skipped.
  - when set to 1, node locked license check will be performed.
  - by default its set to 1. Applicable only for 7Mode v2 licenses.

All of the other various options are described in
L<< NACL::C::SystemLicense->add|lib-NACL-C-SystemLicense-pm/add >>

=item (Optional) B<< nacltask_to_cleanup => 0 | 1 >>

  - when set to 0, cleanup is disabled.
  - when set to 1, appropriate cleanup is invoked.
  - by default its set to 0.

=item (Optional) B<< nacltask_cleanup_manager => $CleanupManagerObj, >>

This option when provided contains the cleanup manager object,
which will be used for cleanup operations

=back

=over

=item Exceptions

=over

=item C<NACL::C::Exceptions::SystemLicense::AlreadyExists>

This type of exception is thrown when an attempt is made to add SystemLicense that already exists.

=item C<NACL::Exceptions::VerifyFailure>

This type of exception is thrown when verification fails for the added System license.

=item C<NACL::Exceptions::NoElementsFound>

This type of exception is thrown to verification fails for system license that does not exists.

=item C<NACL::C::Exceptions::SystemLicense::NoLicenseCodeIsNecessary>

This type of exception is thrown when no system license code is necessary.

=back

=cut

sub add {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = $pkg->_common_validate_with(
        params          => \@_,
        additional_spec => {
            nacltask_verify => { type => SCALAR, default => 0 },
            nacltask_check_node_locked_license =>
              { type => BOOLEAN, default => 1 },
            _was_added           => { type => SCALARREF, optional => 1 },
            _was_licensed        => { type => SCALAR,    default  => 0 },
            nacltask_if_licensed => {
                type      => SCALAR,
                default   => "pass",
                callbacks => {
                        "Value of 'nacltask_if_licensed' should"
                      . " be 'die' or 'pass'" =>
                      sub { $_[0] =~ /^(?:die|pass)$/ }
                }
            },
            $pkg->_cleanup_validate_spec(),
        },
        allow_extra => 1
    );

    # move the cleanup related options to seperate hash
    my ( %opts_for_cleanup, %opts_for_register, $nacltask_to_cleanup );
    $pkg->_move_common_cleanup_opts(
        source => \%opts,
        target => \%opts_for_cleanup,
    );

    my $ci = $opts{command_interface};
    my $is_v2 = $ci->is_feature_supported( feature => 'license_v2' );
    my $stasks =
        $is_v2
      ? $pkg->_add_v2_licenses(%opts)
      : $pkg->_add_license(%opts);
    my @stask_objs = ref $stasks eq 'ARRAY' ? @$stasks : ($stasks);

    $pkg->_copy_common_opts_for_cleanup(
        'source'              => { %opts, %opts_for_cleanup },
        'target'              => \%opts_for_register,
        'nacltask_to_cleanup' => \$nacltask_to_cleanup,
        'to_cleanup'          => 'purge'
    );
    $pkg->_register_for_cleanup(%opts_for_register) if ($nacltask_to_cleanup);

    $Log->exit() if $may_exit;
    return wantarray ? @stask_objs : $stask_objs[0];
} ## end sub add

sub taskverify_add {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = $pkg_or_obj->_common_validate_with(
        params      => \@_,
        allow_extra => 1
    );

    my $pkg   = ref $pkg_or_obj || $pkg_or_obj;
    my $ci    = $opts{command_interface};
    my $is_v1 = $ci->is_feature_supported( feature => 'license_v1' );
    my $license_cs;
    my %common_opts;

    my $package_or_feature =
      defined $opts{package}
      ? delete $opts{package}
      : delete $opts{feature};
    $pkg_or_obj->_copy_common_component_params_with_ci(
        source => \%opts,
        target => \%common_opts
    );

    # For 7Mode, check is-licensed attribute
    if ( $ci->is_7mode() && $is_v1 ) {
        $license_cs = NACL::CS::SystemLicense->fetch(
            %common_opts,
            filter      => { feature => $package_or_feature },
            allow_empty => 1
        );
        if ( $license_cs->is_licensed() ne 'true' ) {
            $Log->exit() if $may_exit;
            NACL::Exceptions::VerifyFailure->throw(
                "Verification failed for $package_or_feature");
        }
    } else {
        my $attribute = $pkg->_get_attribute_name(%opts);

        my @package_or_feature_arr = $package_or_feature;
        if (ref $package_or_feature) {
            @package_or_feature_arr = @$package_or_feature;
        } else {
            @package_or_feature_arr = $package_or_feature;
        }

        # Packages/features might be provided in upper/lower/mixed-case,
        # so standardize on lower-case to allow for comparison.
        @package_or_feature_arr = map {lc $_} @package_or_feature_arr;
        my $package_or_feature_str = join '|', @package_or_feature_arr;
        my @found_license_objs = $pkg->find(
            %common_opts,
            filter      => {$attribute => $package_or_feature_str},
            allow_empty => 1,
        );

        # Packages/features might be provided in upper/lower/mixed-case,
        # so standardize on lower-case to allow for comparison.
        my @found_licenses = map {lc $_->$attribute()} @found_license_objs;
        require List::Compare;
        my $list_compare = List::Compare->new('--unsorted', '--accelerated',
            \@package_or_feature_arr, \@found_licenses);
        my @licenses_not_found = $list_compare->get_Lonly();
        if (@licenses_not_found) {
            $Log->exit() if $may_exit;
            NACL::Exceptions::VerifyFailure->throw('After add, the following '
                . 'licenses were not found '
                . join (',', @licenses_not_found)
            );
        }
    } ## end else [ if ( $opts{command_interface...
    $Log->exit() if $may_exit;
} ## end sub taskverify_add

=head2 purge

    NACL::STask::SystemLicense->purge(
        command_interface  => $command_interface,
        feature            => $feature,
        nacltask_verify    => 1,       # default is 0
        nacltask_if_purged => $action, # default is 'die'
        %other_options
    );

    or

    NACL::STask::SystemLicense->purge(
        command_interface  => $command_interface,
        package            => $package,
        'serial-number'    => $serial_number,
        nacltask_verify    => 1,       # default is 0
        nacltask_if_purged => $action, # default is 'die'
        %other_options
    );

    or

    NACL::STask::SystemLicense->purge(
        command_interface  => $command_interface,
        package            => [$package1, $package2, .. ],
        'serial-number'    => [$serial_number1,$serial_number2,..],
        nacltask_verify    => 1,       # default is 0
        nacltask_if_purged => $action, # default is 'die'
        %other_options
    );

    or

    $license_obj->purge();

(Class or instance method) This method deletes license on the given node.

=over

=item (Optional) B<< nacltask_if_purged => die | pass >>

  - If "die", exception is thrown when license to be deleted
    does not exist. Default id "die".
  - If "pass", exits quietly when license to be deleted does
    not exist

=item (Optional) B<< nacltask_verify => 1  >>

  - would verify that the license is deleted at the end, if set to 1.
  - by default it is set to 0

=back

=over

=item Exceptions

=over

=item C<NACL::C::Exceptions::SystemLicense::DoesNotExist>

This type of exception is thrown when an attempt is made to delete SystemLicense that does not exists.

=item C<NACL::Exceptions::VerifyFailure>

This type of exception is thrown when verification fails for the deleted System license.

=item C<NACL::C::Exceptions::Component>

This type of exception is thrown to verify component delete for the license that does not exists.

=back

=cut

sub purge {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my $pkg        = ref($pkg_or_obj) || $pkg_or_obj;
    my %opts       = $pkg_or_obj->_common_validate_with(
        params          => \@_,
        additional_spec => {
            nacltask_verify => { type => SCALAR, default => 0 },
            nacltask_if_purged =>
              $pkg_or_obj->_if_action_completed_already_validate_spec(),
            $pkg->_cleanup_validate_spec(),
        },
        allow_extra => 1
    );

    # move the cleanup related options to seperate hash
    my ( %opts_for_cleanup, %opts_for_register, $nacltask_to_cleanup );
    $pkg->_move_common_cleanup_opts(
        source => \%opts,
        target => \%opts_for_cleanup,
    );

    my %nacltask_options;
    my %common_opts;
    my @serial_number;
    my $ci            = $opts{command_interface};
    my $attribute     = $pkg->_get_attribute_name(%opts);
    my $is_v2         = $ci->is_feature_supported( feature => 'license_v2' );
    my $doesnot_exist = 'NACL::C::Exceptions::SystemLicense::DoesNotExist';

    # License features which are not supported for V2, Log warning
    # message and return;
    if ($is_v2) {
        my $v1_feature_only =
          $pkg->_check_if_feature_not_supported( %opts, method => 'purge' );

        if ($v1_feature_only) {
            $Log->exit() if $may_exit;
            return;
        }
    }

    $pkg->_copy_common_component_params_with_ci(
        source => \%opts,
        target => \%common_opts
    );

    $pkg_or_obj->_move_nacltask_options(
        source => \%opts,
        target => \%nacltask_options
    );

    # Call the component delete. If license doesn't exist
    # then it would throw an exeception which is handled below.
    try {
        $pkg->delete(%opts);
    }
    catch NACL::C::Exceptions::Component with {
        my $e = shift;

        # If the exception is of type MultipleCommandFailied,
        # check all the failures.
        if ( $nacltask_options{nacltask_if_purged} =~ m[^pass$]i ) {
            if ( $e->isa('NACL::C::Exceptions::MultipleCommandFailed') ) {
                my @all_failures = $e->all_failures();
                foreach my $failure (@all_failures) {
                    if ( !$doesnot_exist->detect( exception => $failure ) ) {
                        $Log->exit() if $may_exit;
                        $e->throw();
                    }
                }
            }
            elsif (
                $e->isa('NACL::C::Exceptions::SystemLicense::DoesNotExist') )
            {

                # don't throw
            }
            else {
                $Log->exit() if $may_exit;
                $e->throw();
            }
        }
        else {
            $Log->exit() if $may_exit;
            $e->throw();
        }
    };

    if ( $nacltask_options{nacltask_verify} ) {
        $pkg->taskverify_purge(%opts);
    }

    $pkg->_copy_common_opts_for_cleanup(
        'source'              => { %opts, %opts_for_cleanup },
        'target'              => \%opts_for_register,
        'nacltask_to_cleanup' => \$nacltask_to_cleanup,
        'to_cleanup'          => 'add'
    );

    $pkg->_register_purge_for_cleanup(%opts_for_register)
      if ($nacltask_to_cleanup);

    $Log->exit() if $may_exit;
} ## end sub purge

sub _register_purge_for_cleanup {
    $Log->enter() if $may_enter;
    my ( $pkg, %opts ) = @_;

    if ( exists $opts{'package'} ) {
        $opts{'feature'} = delete $opts{'package'};
    }
    delete $opts{'serial-number'};

    $pkg->_register_for_cleanup(%opts);
    $Log->exit() if $may_exit;
}

sub taskverify_purge {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = $pkg_or_obj->_common_validate_with(
        params      => \@_,
        allow_extra => 1
    );

    my $ci    = $opts{command_interface};
    my $is_v1 = $ci->is_feature_supported( feature => 'license_v1' );
    my $pkg   = ref $pkg_or_obj || $pkg_or_obj;
    my %common_opts;
    $pkg_or_obj->_copy_common_component_params_with_ci(
        source => \%opts,
        target => \%common_opts
    );
    my $attribute = $pkg->_get_attribute_name(%opts);
    my @packages =
      ref $opts{$attribute} ne 'ARRAY'
      ? split /,/, $opts{$attribute}
      : @{ $opts{$attribute} };
    my @license_cs = NACL::CS::SystemLicense->fetch(
        %common_opts,
        filter      => { $attribute => join '|', @packages },
        allow_empty => 1
    );

    foreach my $cs (@license_cs) {

        # For 7Mode, check is-licensed attribute, otherwise
        # check serial number.
        if ($cs) {
            if ( $ci->is_7mode() && $is_v1 ) {
                if ( $cs->is_licensed() ne 'false' ) {
                    $Log->exit() if $may_exit;
                    NACL::Exceptions::VerifyFailure->throw(
                            "Verification failed! Failed to "
                          . "purge the license for "
                          . $cs->feature() );
                } ## end if ( $license_state_obj...
            }
            else {
                if ( $cs->serial_number() ) {
                    $Log->exit() if $may_exit;
                    NACL::Exceptions::VerifyFailure->throw(
                            "Verification failed! Failed to "
                          . "purge the license for "
                          . $cs->$attribute() );
                } ## end if ( $cs...
            } ## end else [ if ( $opts{command_interface...
        } ## end if ($cs)
    }

    $Log->exit() if $may_exit;
} ## end sub taskverify_purge

=head2 get_valid_license_packages

    NACL::STask::SystemLicense->get_valid_license_packages(
        command_interface  => $command_interface,
        nacltask_filter => {
                'mode' => 'CMode',
		'platform' => 'filer_or_vsim',
                'serial-number' => '1-80-000011'
        },
    );

(Class method) This method returns the list of
valid features/packages available for the combination of
mode+platform+serial-number on the given command_interface.
Returns the array of features/packages available.

=over

=item (Required) B<< nacltask_filter => $filter_hashref >>

  HASHREF that contains the values of mode,platform, serial-number
  nacltask_filter => {
        'mode' => 'CMode',
	'platform' => 'filer_or_vsim',
        'serial-number' => '1-80-000011'
  },
  mode default value is 'CMode'
  Based on this filter the list of valid license/packages
  are searched.

=back

=cut

sub get_valid_license_packages{
    $Log->enter() if $may_enter;

    my $pkg  = shift;
    my %opts = $pkg->_common_validate_with(
        params          => \@_,
        additional_spec => {
            nacltask_filter => { type => HASHREF },
        },
    );

    my $ci            = $opts{command_interface} ;
    my $is_v2         = $ci->is_feature_supported( feature => 'license_v2' );

    my %license_hash =
        $is_v2
      ? %NACL::STask::SystemLicense::Default::Default_Licenses_V2
      : %NACL::STask::SystemLicense::Default::Default_Licenses_V1;

    my $mode_filter = $opts{nacltask_filter}->{mode} // 'CMode' ;
    my $platform_filter = $opts{nacltask_filter}->{platform} ;
    my $serial_number_filter = $opts{nacltask_filter}->{'serial-number'};

    if(!defined $platform_filter){
        NATE::BaseException->throw('
                mode & platform filters are both required .
                Filter provided :'.Dumper($opts{nacltask_filter}));
    }

    my (%features_hash , @features );

    %features_hash =  %{$license_hash{$mode_filter}->{$platform_filter}};

    if($serial_number_filter){
	map {push @features, $_ } keys %{$features_hash{$serial_number_filter}} ;
    }else{
	## check if serial number not provided
	## then @features should not contain any
	## reference data structure
	my $contains_ref ;
	foreach (keys %features_hash){
	   if(ref $features_hash{$_}){
             $contains_ref = 1;
	     last;
	   }
	}
	map {push @features, $_} keys %features_hash if(! $contains_ref);
    }

    my $err_msg = 'No features found for this combination.
                May be the combination is incorrect or
                an extra filter of serial-number is required.
                Filter provided : '.Dumper($opts{nacltask_filter}) ;

    $Log->debug($err_msg) if(!@features);

    $Log->exit() if $may_exit;
    return @features;
}## end sub get_valid_license_packages

sub _add_v2_licenses {
    $Log->enter() if $may_enter;
    my ( $pkg, %opts ) = @_;
    foreach my $param ( 'license-code', 'package' ) {
        if (defined $opts{$param} && !ref $opts{$param}) {
            $opts{$param} = [split /,/, $opts{$param}];
        }
    }
    my @licenses;
    my @packages = @{$opts{package}};

    if ( defined $opts{'license-code'} ) {
        @licenses = @{$opts{'license-code'}};
    }
    my $obj_or_objs;
    if ( scalar(@packages) > 1 ) {
        $opts{'license-code'} = \@licenses;
        $opts{package} = \@packages;
        $obj_or_objs = $pkg->_add_multiple_licenses(%opts);
    }
    else {
        $obj_or_objs = $pkg->_add_license(%opts);
    }
    if ($opts{command_interface}->is_7mode()) {
        my @lc_packages = map {lc $_} @packages;
        my @snaplock_packages = grep {/snaplock/} @lc_packages;
        foreach my $snaplock_pkg (@snaplock_packages) {
            my $option_name = "licensed_feature.${snaplock_pkg}.enable";
            $opts{command_interface}->get_7m_or_nodescope_apiset()->options(
                $option_name => 'on'
             );
        }
    }
    $Log->exit() if $may_exit;
    return $obj_or_objs;
} ## end sub _add_v2_licenses

sub _add_license {
    $Log->enter() if $may_enter;
    my ( $pkg, %opts ) = @_;
    my $was_added = delete $opts{_was_added};
    my $package_or_feature =
      defined $opts{package} ? $opts{package} : $opts{feature};
    my $ci = $opts{command_interface};
    my $is_v2 = $ci->is_feature_supported( feature => 'license_v2' );

    # _was_licensed option is not required
    delete $opts{_was_licensed};

    my $self;
    my %nacltask_options;

    $pkg->_move_nacltask_options(
        source => \%opts,
        target => \%nacltask_options
    );

    # License features which are not supported for V2, Log warning
    # message return license object
    if ($is_v2) {
        my $v1_feature_only =
          $pkg->_check_if_feature_not_supported( %opts, method => 'add' );
        if ($v1_feature_only) {
            my $serial_number = $pkg->find_base_serial_number(
                command_interface => $opts{command_interface}
            );

            # An earlier method might have turned it into an array-ref, so handle this case.
            $package_or_feature = $package_or_feature->[0] if (ref $package_or_feature);
            $self = $pkg->new(
                command_interface => $opts{command_interface},
                feature           => $package_or_feature,
                serial_number     => $serial_number,
            );
            $Log->exit() if $may_exit;
            return $self;
        }
    }

    my %additional_args;
    if ($opts{command_interface}->is_cmode()) {
        my %common_opts_ci;
        $pkg->_copy_common_component_params_with_ci(
            source => \%opts,
            target => \%common_opts_ci,
        );
        $additional_args{nacl_serial_number} =
            $pkg->find_base_serial_number_cm(%common_opts_ci);
    }

    # If license-code is not specified, call _get_license_code method
    $opts{'license-code'} //= $pkg->_get_license_code(%opts, %additional_args);

    # 7mode doesn't allow a site license to overwrite a node
    # locked license. So, for a node locked license, we need
    # to check if it is already installed. See burt602756.
    my $is_node_locked = 0;
    my $is_node_locked_check_req =
      $nacltask_options{nacltask_check_node_locked_license};
    if ( $ci->is_7mode() && $is_v2 && $is_node_locked_check_req ) {
        $is_node_locked =
          $pkg->_is_node_locked_license( 'package' => $package_or_feature );
    }

    if ( defined $was_added || $is_node_locked ) {
        $Log->debug( "Either _was_added option is provided or"
              . " the license is a node locked license. Going"
              . " to check if the license is already available" );
        $self = $pkg->_check_if_already_licensed(%opts);
    }

    # Try to add the license and check if license-code
    # is necessary for the feature.
    if ( !defined $self ) {
        try {
            $$was_added = 0;
            $self       = $pkg->SUPER::add(%opts, %additional_args);
            $$was_added = 1;
            $self->_was_added($$was_added);

            # Verify license add
            if ( $nacltask_options{nacltask_verify} ) {
                $pkg->taskverify_add(%opts);
            }
        } ## end try
        catch NACL::C::Exceptions::SystemLicense::NoLicenseCodeIsNecessary
          with {
            my $e = shift;
            if ( $nacltask_options{nacltask_if_licensed} =~ m[^pass$]i ) {
                if ( $opts{command_interface}
                    ->is_feature_supported( feature => 'license_v2' ) )
                {
                    my $cs        = $pkg->get_CS_package_name();
                    my $cs_object = $cs->fetch(
                        command_interface => $opts{command_interface},
                        filter            => { package => $package_or_feature },
                        requested_fields  => [qw(serial-number)]
                    );
                    $self = $pkg->new(
                        command_interface => $opts{command_interface},
                        feature           => $package_or_feature,
                        serial_number     => $cs_object->serial_number()
                    );
                }
                else {
                    $self = $pkg->new(
                        command_interface => $opts{command_interface},
                        feature           => $package_or_feature
                    );
                } ## end else [ if ( $opts{command_interface...
            }
            else {
                $Log->debug( "Error msg : " . $e->text() );
                $Log->exit() if $may_exit;
                $e->throw();
            }
          };
    } ## end if ( !defined $self )

    $Log->exit() if $may_exit;
    return $self;
} ## end sub _add_license

sub _add_multiple_licenses {
    $Log->enter() if $may_enter;
    my ( $pkg, %opts ) = @_;

    my $was_added_multiple = delete $opts{_was_licensed};
    my $license_packages   = delete $opts{package};
    my $license_codes      = delete $opts{'license-code'};

    # _was_added option is not required
    delete $opts{_was_added};

    my %nacltask_options;
    my @stasks;
    my @already_licensed;

    $pkg->_move_nacltask_options(
        source => \%opts,
        target => \%nacltask_options
    );

    my %package_hash;
    @package_hash{@$license_packages} = @$license_codes;
    my $is_node_locked_check_req =
      $nacltask_options{nacltask_check_node_locked_license};

    my $is_cmode = $opts{command_interface}->is_cmode();
    my $serial_number;
    my $get_serial_num = sub {
        if (!defined $serial_number) {
            my %common_opts_ci;
            $pkg->_copy_common_component_params_with_ci(
                source => \%opts,
                target => \%common_opts_ci,
            );
            $serial_number = $pkg->find_base_serial_number_cm(%common_opts_ci);
        }
        return $serial_number;
    };

    foreach my $package ( keys %package_hash ) {

        my $is_node_locked = 0;
        my $self;

        # 7mode doesn't allow a site license to overwrite a node
        # locked license. So, for a node locked license, we need
        # to check if it is already installed. See burt602756.
        if (   $opts{command_interface}->is_7mode()
            && $is_node_locked_check_req )
        {
            $is_node_locked =
              $pkg->_is_node_locked_license( 'package' => $package );
        }

        if ( $was_added_multiple || $is_node_locked ) {
            $Log->debug( "Either _was_licensed option is provided"
                  . " or the license is a node locked license."
                  . " Going to check if the license is already"
                  . " available" );
            $self =
              $pkg->_check_if_already_licensed( %opts, package => $package );
        }

        if ( defined $self ) {
            push @already_licensed, $self;
            delete $package_hash{$package};
        } elsif ( !defined $package_hash{$package} ) {
            # If license-code is not specified, call
            # _get_license_code method
            my %additional_args;
            if ($is_cmode) {
                $additional_args{nacl_serial_number} = $get_serial_num->();
            }
            $package_hash{$package} =
                $pkg->_get_license_code(%opts, package => $package, %additional_args);
        }
    }

    my @package_array = keys %package_hash;
    @$license_codes = values %package_hash;

    if ( scalar @package_array ) {
        %opts = (
            %opts,
            package        => \@package_array,
            'license-code' => $license_codes
        );

        my %additional_args;
        if ($is_cmode && defined $serial_number) {
            $additional_args{nacl_serial_number} = $serial_number;
        }

        # Call license add
        @stasks = $pkg->SUPER::add(%opts, %additional_args);
        foreach my $stask (@stasks) { $stask->_was_added(1) }

        # Verify license add
        if ( $nacltask_options{nacltask_verify} ) {
            $pkg->taskverify_add(%opts);
        }
    }

    my @stask_objs;
    foreach my $stask ( @stasks, @already_licensed ) {
        foreach my $package (@$license_packages) {
            if ( lc $stask->package() eq lc $package ) {
                push @stask_objs, $stask;
            }
        }
    }

    $Log->exit() if $may_exit;
    return \@stask_objs;
} ## end sub _add_multiple_licenses

sub _guess_model {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = $pkg_or_obj->_common_validate_with(
        params      => \@_,
        allow_extra => 1
    );
    my $model;
    my $version_manager = $opts{command_interface}->get_version_manager();

    # If 7-mode simulator with promoted build, return 'promoted_simulator'.
    # Since 7-mode and C-mode development filer and simulator licenses
    # are same, return filer_or_vsim

    if ($opts{command_interface}->is_7mode()
        && $version_manager->is_promoted() &&
        $version_manager->get_version_attribute( attribute => 'modelinfo' ) =~ m[SIMBOX]i)
    {
        $model = 'promoted_simulator';
    }
    else {
        $model = 'filer_or_vsim';
    }

    $Log->exit() if $may_exit;
    return $model;
} ## end sub _guess_model

sub _get_license_code {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = $pkg_or_obj->_common_validate_with(
        params      => \@_,
        allow_extra => 1
    );

    my $model = $pkg_or_obj->_guess_model(%opts);
    my $ci    = $opts{command_interface};
    my $mode  = $ci->mode();
    my $pkg   = ref $pkg_or_obj || $pkg_or_obj;
    my $package_or_feature =
      defined $opts{package}
      ? delete $opts{package}
      : delete $opts{feature};


    my $serial = $opts{nacl_serial_number};

    my $attribute       = $pkg->_get_attribute_name(%opts);
    my $is_v2           = $ci->is_feature_supported( feature => 'license_v2' );
    my $license_version = $is_v2 ? 'V2' : 'V1';
    my %license_hash =
        $is_v2
      ? %NACL::STask::SystemLicense::Default::Default_Licenses_V2
      : %NACL::STask::SystemLicense::Default::Default_Licenses_V1;

    if ($mode eq 'CMode' && !defined $serial) {
        my %common_opts;
        $pkg_or_obj->_copy_common_component_params_with_ci(
            source => \%opts,
            target => \%common_opts
        );
        $serial = $pkg->find_base_serial_number_cm(%common_opts);
    } ## end if ( $mode eq 'CMode' )

    # Verify Default_Licenses{$mode}->{$model} hash.
    my $license = $license_hash{$mode};
    if ( !exists $license->{$model} ) {
        $Log->exit() if $may_exit;
        NACL::Exceptions::NoElementsFound->throw(
                "Entry doesn't exist for $model ($model) in"
              . " Default_Licenses hash" );
    } ## end if ( !exists $license->...

    # Verify Default_Licenses{$mode}->{$model}->{$serial} hash
    if ( $mode ne '7Mode' && !exists $license->{$model}->{$serial} ) {
        $Log->exit() if $may_exit;
        NACL::Exceptions::NoElementsFound->throw(
                "Entry doesn't exist for serial $serial"
              . " ($mode, $model) in Default_Licenses hash" );
    } ## end if ( $mode ne '7Mode' ...

    # Now check for the license-code
    my $license_hash = $license->{$model};
    my $license_code =
        $mode eq '7Mode'
      ? $license_hash->{$package_or_feature}
      : $license_hash->{$serial}->{$package_or_feature};

    if ( !defined $license_code ) {
        $Log->exit() if $may_exit;
        NACL::Exceptions::NoElementsFound->throw(
                "Either feature $package_or_feature is not valid for "
              . "Licensing $license_version or license code is not "
              . "present in Default_Licenses hash" );
    } ## end if ( !defined $license_code)

    $Log->exit() if $may_exit;
    return $license_code;
} ## end sub _get_license_code

sub _common_validate_with {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = $pkg_or_obj->SUPER::_common_validate_with(@_);
    my $pkg        = ref($pkg_or_obj) || $pkg_or_obj;
    my $attribute  = $pkg->_get_attribute_name(%opts);
    my $package_or_feature =
      defined $opts{package}
      ? delete $opts{package}
      : delete $opts{feature};
    $opts{$attribute} = $package_or_feature;
    my $ci = $opts{command_interface};
    my $is_v2 = $ci->is_feature_supported( feature => 'license_v2' );

    if ($is_v2) {
        my @package_names;
        my @packages =
          ref $opts{$attribute} eq 'ARRAY'
          ? @{ $opts{$attribute} }
          : split /,/, $opts{$attribute};
        if ( scalar @packages > 1 ) {
            foreach my $package (@packages) {
                push @package_names,
                  $pkg_or_obj->_map_feature_name( %opts,
                    $attribute => $package );
            }
            $opts{$attribute} = \@package_names;
        }
        else {
            $opts{$attribute} = $packages[0];
            $opts{$attribute} = $pkg_or_obj->_map_feature_name(%opts);
        }
    }
    else {
        $opts{$attribute} = $pkg_or_obj->_map_feature_name(%opts);
    }

    $Log->exit() if $may_exit;
    return %opts;
} ## end sub _common_validate_with

sub _map_feature_name {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, %opts ) = @_;
    my $ci = $opts{command_interface};
    my $v2 = $ci->is_feature_supported( feature => 'license_v2' );
    my $package_or_feature =
      defined $opts{package}
      ? lc delete $opts{package}
      : lc delete $opts{feature};

    my %map_old_to_new_v2_feature =
      %NACL::STask::SystemLicense::Default::Map_Old_Feature_To_New_Feature_V2;
    my %map_features_v1_to_v2_cm =
      %NACL::STask::SystemLicense::Default::Map_V1_To_V2_Cm;
    my %map_features_v2_to_v1_cm = reverse %map_features_v1_to_v2_cm;
    my %map_features_v1_to_v2_7m =
      %NACL::STask::SystemLicense::Default::Map_V1_To_V2_7m;
    my %map_features_v2_to_v1_7m =
      %NACL::STask::SystemLicense::Default::Map_V2_To_V1_7m;
    my %map_features_cm_to_7m =
      %NACL::STask::SystemLicense::Default::Map_Features_Cm_To_7m;
    my %map_features_7m_to_cm = reverse %map_features_cm_to_7m;
    my %map_features_dev_to_promo =
      %NACL::STask::SystemLicense::Default::Map_Features_Dev_To_Promo_7m;
    my %map_features_promo_to_dev = reverse %map_features_dev_to_promo;

    my $mapped_license_feature;

    if ($v2) {
        if ( $ci->is_7mode() ) {
            $mapped_license_feature = _mapping(
                mapping_tables =>
                  [ \%map_features_cm_to_7m, \%map_features_v1_to_v2_7m ],
                feature => $package_or_feature
            );
        } else {
            $mapped_license_feature = _mapping(
                mapping_tables =>
                  [ \%map_features_7m_to_cm, \%map_features_v1_to_v2_cm ],
                feature => $package_or_feature
            );
        }

        $mapped_license_feature = _mapping(
            mapping_tables => [ \%map_old_to_new_v2_feature ],
            feature        => $mapped_license_feature
        );

    } else {
        if ( $ci->is_7mode() ) {
            if ( $ci->get_version_manager()->is_promoted() ) {
                $mapped_license_feature = _mapping(
                    mapping_tables => [
                        \%map_features_cm_to_7m,
                        \%map_features_v2_to_v1_7m,
                        \%map_features_dev_to_promo
                    ],
                    feature => $package_or_feature
                );
            } else {
                $mapped_license_feature = _mapping(
                    mapping_tables => [
                        \%map_features_cm_to_7m,
                        \%map_features_v2_to_v1_7m,
                        \%map_features_promo_to_dev
                    ],
                    feature => $package_or_feature
                );
            }
        } else {
            $mapped_license_feature = _mapping(
                mapping_tables =>
                  [ \%map_features_7m_to_cm, \%map_features_v2_to_v1_cm ],
                feature => $package_or_feature
            );
        }
    }

    $Log->exit() if $may_exit;
    return $mapped_license_feature;
} ## end sub _map_feature_name

sub _check_if_already_licensed {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, %opts ) = @_;
    my $existing = undef;
    my $ci       = $opts{command_interface};
    my $is_v1    = $ci->is_feature_supported( feature => 'license_v1' );

    my %common_opts;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%common_opts,
        copy => [ keys( %{ $pkg_or_obj->_common_validate_spec_without_ci() } ) ]
    );

    # For V2, package could be specified as an array, but the CS package
    # expects it to be a scalar, so normalize to a scalar by joining on |
    my $filter = $pkg_or_obj->get_primary_keys_options(%opts);
    if (exists $filter->{package} && ref $filter->{package}) {
        $filter->{package} = join '|', @{$filter->{package}};
    }
    $existing = $pkg_or_obj->find(
        command_interface => $ci,
        filter            => $filter,
        allow_empty       => 1,
        %common_opts
    );

    # For 7Mode License V1, check is-licensed attribute
    if ( $ci->is_7mode() && $is_v1 && defined $existing ) {
        if ( $existing->get_one_state_attribute('is-licensed') ne 'true' ) {
            undef $existing;
        }
    }
    $Log->exit() if $may_exit;
    return $existing;
} ## end sub _check_if_already_licensed

sub _check_if_feature_not_supported {
    $Log->enter() if $may_enter;
    my ( $pkg, %opts ) = @_;
    my $mode   = $opts{command_interface}->mode();
    my $method = delete $opts{method};
    my $features =
      $NACL::STask::SystemLicense::Default::Unsupported_Features{$mode};
    my $package_or_feature =
      defined $opts{package}
      ? $opts{package}
      : $opts{feature};

    # An earlier method might have turned it into an array-ref, so handle this case.
    $package_or_feature = $package_or_feature->[0] if (ref $package_or_feature);
    my $only_v1 = $features->{ lc $package_or_feature };
    my $value   = 0;

    # if caller is purge method, return nothing
    if ($only_v1) {
        $value = 1;
        if ( $method =~ m[purge]i ) {
            $Log->warn( "$package_or_feature is not supported feature for "
                  . "Licensing V2. So, skipping license delete and "
                  . "verification" );
        }
        elsif ( $method =~ m[add]i ) {
            $Log->warn( "$package_or_feature is not supported feature for "
                  . "Licensing V2. So, skipping license add and verification" );
        }
    }

    $Log->exit() if $may_exit;
    return $value;
} ## end sub _check_if_feature_not_supported

sub _get_attribute_name {
    $Log->enter() if $may_enter;
    my ( $pkg, %opts ) = @_;
    my $ci        = $opts{command_interface};
    my $is_v2     = $ci->is_feature_supported( feature => 'license_v2' );
    my $attribute = $is_v2 ? 'package' : 'feature';
    $Log->exit() if $may_exit;
    return $attribute;
} ## end sub _get_attribute_name

sub _is_node_locked_license {
    $Log->enter() if $may_enter;
    my ( $pkg, %opts ) = @_;
    my $package = lc $opts{package};
    my $is_node_locked =
      $NACL::STask::SystemLicense::Default::Node_Locked_License{$package};
    $Log->exit() if $may_exit;
    return $is_node_locked;
} ## end sub _is_node_locked_license

sub _mapping {
    $Log->enter() if $may_enter;
    my %opts = validate_with(
        params => \@_,
        spec   => {
            mapping_tables => { type => ARRAYREF },
            feature        => { type => SCALAR },
        },
    );

    my @mapping_tables = @{ $opts{mapping_tables} };
    my $feature        = $opts{feature};
    foreach my $map (@mapping_tables) {
        my $mapped_feature = $map->{$feature};
        $feature = defined $mapped_feature ? $mapped_feature : $feature;
    }
    $Log->exit() if $may_exit;
    return $feature;
} ## end sub _mapping

1;
