#
# Copyright (c) 2001-2011 NetApp, Inc., All Rights Reserved
# Any use, modification, or distribution is prohibited
# without prior written consent from NetApp, Inc.
#

## @summary Storage for Default Licenses
## @status Shared
## @author dl-nacl-dev
## @pod here

package NACL::STask::SystemLicense::Default;
use strict;
use warnings;

=head1 NAME

NACL::STask::SystemLicense::Default - Storage for Default Licenses

=head1 SYNOPSIS

  use NACL::STask::SystemLicense::Default;

=head1 DESCRIPTION

This was separated from NACL::STask::SystemLicense for internal
housekeeping, and therefore there should be absolutely nothing
of interest for the end user.

All the default licenses are listed in %Default_Licenses hash.
Also, all the feature-names which need CMode to 7Mode mapping
or 7Mode to CMode mapping, should be listed in %Map_Features
hash. Currently listed features in %Map_Features_Cm_To_7m
hash are:

    1. flexclone     ( flex_clone for 7Mode )
    2. snapmirror_ls ( snapmirror for 7Mode )
    3. snapmirror_dp ( snapmirror for 7Mode )

License features which need V1 to V2 mapping for CMode are
listed in %Map_V1_To_V2_Cm hash. Currently listed features
in %Map_V1_To_V2_Cm hash are:

    1. snapmirror_ls     ( snapmirror for Licensing V2 )
    2. snapmirror_dp     ( snapmirror for Licensing V2 )
    3. spa               ( snapprotectapps for Licensing V2 )
    4. SnapManager_Suite ( snapmanagersuite for Licensing V2 )

License features which need V1 to V2 mapping for 7Mode are
listed in %Map_V1_To_V2_7m hash. Currently listed features
in %Map_V1_To_V2_7m hash are:

    1.  nfs                    ( nfs for Licensing V2 )
    2.  cifs                   ( cifs for Licensing V2 )
    3.  iscsi                  ( iscsi for Licensing V2 )
    4.  fcp                    ( fcp for Licensing V2 )
    5.  snaprestore            ( snaprestore for Licensing V2 )
    6.  snapmirror             ( snapmirror for Licensing V2 )
    7.  snapmirror_sync        ( snapmirror for Licensing V2 )
    8.  flex_clone             ( flexclone for Licensing V2 )
    9.  sv_ontap_sec           ( snapvault for Licensing V2 )
    10. sv_ontap_pri           ( snapvault for Licensing V2 )
    11. snaplock               ( snaplock for Licensing V2 )
    12. snapmanagerexchange    ( snapmanagersuite for Licensing V2 )
    13. snapmanager_hyperv     ( snapmanagersuite for Licensing V2 )
    14. snapmanager_oracle     ( snapmanagersuite for Licensing V2 )
    15. snapmanager_sap        ( snapmanagersuite for Licensing V2 )
    16. snapmanager_sharepoint ( snapmanagersuite for Licensing V2 )
    17. snapmanager_vi         ( snapmanagersuite for Licensing V2 )
    18. snapdrive_windows      ( snapmanagersuite for Licensing V2 )
    19. smsql                  ( snapmanagersuite for Licensing V2 )
    20. sv_applications_pri    ( snapprotectapps for Licensing V2 )

Licenses listed in the hash %Map_V2_To_V1_7m need mapping from V2 to
V1. snapmanagersuite is excluded from this list as multiple V1
features are maaped to feature snapmanagersuite. Currently listed
features in %Map_V2_To_V1_7m hash are:

    1. flexclone       ( flex_clone for Licensing V1 )
    2. snapprotectapps ( sv_applications_pri for Licensing V1 )

SnapMirror_LS license code is available only for serial-number
1-80-123456.

Please visit http://wikid.netapp.com/w/Licensing/Internal_Licenses
for reference.

=cut

our %Map_Features_Cm_To_7m = (
    flexclone     => 'flex_clone',
    snapmirror_ls => 'snapmirror',
    snapmirror_dp => 'snapmirror',
);

our %Map_V1_To_V2_Cm = (
    snapmirror_ls     => 'snapmirror',
    snapmirror_dp     => 'snapmirror',
    spa               => 'snapprotectapps',
    snapmanager_suite => 'snapmanagersuite',
);

our %Map_V1_To_V2_7m = (
    nfs                    => 'nfs',
    cifs                   => 'cifs',
    iscsi                  => 'iscsi',
    fcp                    => 'fcp',
    snaprestore            => 'snaprestore',
    snapmirror             => 'snapmirror',
    snapmirror_sync        => 'snapmirror',
    flex_clone             => 'flexclone',
    sv_ontap_sec           => 'snapvault',
    sv_ontap_pri           => 'snapvault',
    snaplock               => 'snaplock',
    snapmanagerexchange    => 'snapmanagersuite',
    snapmanager_hyperv     => 'snapmanagersuite',
    snapmanager_oracle     => 'snapmanagersuite',
    snapmanager_sap        => 'snapmanagersuite',
    snapmanager_sharepoint => 'snapmanagersuite',
    snapmanager_vi         => 'snapmanagersuite',
    snapdrive_windows      => 'snapmanagersuite',
    smsql                  => 'snapmanagersuite',
    sv_applications_pri    => 'snapprotectapps',
);

our %Map_V2_To_V1_7m = (
    flexclone       => 'flex_clone',
    snapprotectapps => 'sv_applications_pri',
);

our %Map_Features_Dev_To_Promo_7m = ( cf => 'cluster' );

our %Unsupported_Features = (
    'CMode' => {
        mirror         => 1,
        flexvol_hpo    => 1,
        striped_volume => 1,
    },

    '7Mode' => {
        a_sis                => 1,
        cf                   => 1,
        cf_remote            => 1,
        compression          => 1,
        dafs                 => 1,
        disk_sanitization    => 1,
        env_exception        => 1,
        flash_cache          => 1,
        flex_scale           => 1,
        flexcache_nfs        => 1,
        http                 => 1,
        multistore           => 1,
        nearstore_option     => 1,
        nearstore_asis1      => 1,
        operations_manager   => 1,
        persistent_archive   => 1,
        protection_manager   => 1,
        provisioning_manager => 1,
        rapid_restore        => 1,
        smdomino             => 1,
        snapdrive_unix       => 1,
        snapmover            => 1,
        snapvalidator        => 1,
        storage_services     => 1,
        sv_linux_pri         => 1,
        sv_unix_pri          => 1,
        sv_vi_pri            => 1,
        sv_windows_pri       => 1,
        syncmirror_local     => 1,
        vfiler               => 1,
        vld                  => 1,
        volcopy              => 1,
    },
);

our %Default_Licenses_V1 = (
    '7Mode' => {
        'filer_or_vsim' => {
            a_sis                  => 'OSFYRVH',
            cifs                   => 'GTLYRVH',
            cf                     => 'KDQLGBN',
            cf_remote              => 'GTFJHSI',
            compression            => 'QTEUWOJ',
            dafs                   => 'OBOLDQC',
            disk_sanitization      => 'CPUPBIL',
            env_exception          => 'KUVPBIL',
            fcp                    => 'ICRNCZG',
            flash_cache            => 'KRFUWOJ',
            flex_clone             => 'QNFYRVH',
            flex_scale             => 'MZGLGBN',
            flexcache_nfs          => 'OSRAREM',
            http                   => 'AODUWOJ',
            insight_balance        => 'WIOLGBN',
            iscsi                  => 'IKVAREM',
            multistore             => 'SPQLGBN',
            nearstore_option       => 'SAXCNCG',
            nearstore_asis1        => 'EJCFMLK',
            nfs                    => 'UYTCNCG',
            operations_manager     => 'QNVRXFF',
            persistent_archive     => 'EHLAREM',
            protection_manager     => 'QVDWSMD',
            provisioning_manager   => 'IDIJHSI',
            rapid_restore          => 'QBBSXFF',
            smdomino               => 'CFZRXFF',
            smsql                  => 'SHEWSMD',
            snapdrive_unix         => 'WOFLDQC',
            snapdrive_windows      => 'UVGYRVH',
            snaplock               => 'KJHHIJE',
            snaplock_enterprise    => 'AOXRXFF',
            snapmanagerexchange    => 'GFMYRVH',
            snapmanager_hyperv     => 'QDSCNCG',
            snapmanager_oracle     => 'EPTEMLK',
            snapmanager_sap        => 'KYLAOTB',
            snapmanager_sharepoint => 'IFNNCZG',
            snapmirror             => 'UUNLDQC',
            snapmirror_sync        => 'IEOAOTB',
            snapmover              => 'KBNNOFF',
            snaprestore            => 'EOYPBIL',
            snapvalidator          => 'MCGWSMD',
            storage_services       => 'AJPPBIL',
            sv_linux_pri           => 'GROPYWA',
            sv_ontap_pri           => 'YAYCNCG',
            sv_ontap_sec           => 'KVPYRVH',
            sv_unix_pri            => 'IUILGBN',
            sv_vi_pri              => 'ARNPYWA',
            sv_windows_pri         => 'SLGLDQC',
            syncmirror_local       => 'UELJHSI',
            vfiler                 => 'SPQLGBN',
            vld                    => 'QYOAREM',
            volcopy                => 'AAWPYWA'
        },
        promoted_simulator => {    # this is for promoted filer simulator only
            a_sis                  => 'MTVVGAF',
            cifs                   => 'DZDACHD',
            cluster                => 'JQIGWWF',
            cluster_remote         => 'BCZZBHD',
            compression            => 'CEVIVFK',
            dafs                   => 'TBPGWWF',
            disk_sanitization      => 'PZKEAZL',
            fcp                    => 'BKHEXNB',
            flex_cache             => 'HFAYFJJ',
            flexcache_nfs          => 'ADIPPVM',
            flex_clone             => 'ANLEAZL',
            http                   => 'NAZOMKC',
            http_compression       => 'SEDAFSN',
            iscsi                  => 'BSLRLTG',
            rapid_restore          => 'CMRGWWF',
            nearstore_option       => 'ELNRLTG',
            nfs                    => 'BQOEAZL',
            operations_manager     => 'CYLGWWF',
            protection_manager     => 'CGUKRDE',
            provisioning_manager   => 'UYNXFJJ',
            smdomino               => 'RKBAFSN',
            smsql                  => 'HNGEAZL',
            snapdrive_unix         => 'IZVZBHD',
            snapdrive_windows      => 'GGXMQMI',
            snaplock               => 'ZOJPPVM',
            snaplock_enterprise    => 'PTZZESN',
            snapmanagerexchange    => 'BCJEAZL',
            snapmanager_hyperv     => 'COIRLTG',
            snapmanager_oracle     => 'QZJTKCL',
            snapmanager_sap        => 'WICPMKC',
            snapmanager_sharepoint => 'UPDCBQH',
            snapmirror             => 'DFVXFJJ',
            snapmirror_sync        => 'XJQIVFK',
            snaprestore            => 'DNDCBQH',
            snapvalidator          => 'JQAACHD',
            socks                  => 'UXBNQMI',
            sv_linux_pri           => 'ZYICXLC',
            sv_ontap_pri           => 'PVOIVFK',
            sv_ontap_sec           => 'PDXMQMI',
            sv_unix_pri            => 'RQAYBFE',
            sv_windows_pri         => 'ZOPRKAM',
            syncmirror_local       => 'RIQTKCL',
            vfiler                 => 'NQBYFJJ',
            vld                    => 'JGFRLTG'
        }
    },
    'CMode' => {
        'filer_or_vsim' => {
            '1-80-000011' => {
                base              => 'VMZIVZGKWVMADF',
                mirror            => 'TBVMWZGKWVMADF',
                flexvol_hpo       => 'FWSBXZGKWVMADF',
                cifs              => 'RQQQXZGKWVMADF',
                snaprestore       => 'DLOFYZGKWVMADF',
                insight_balance   => 'HNUJDAHKWVMADF',
                nfs               => 'BAKJZZGKWVMADF',
                snapmanager_suite => 'JYYFCAHKWVMADF',
                snapmirror_dp     => 'NUHYZZGKWVMADF',
                striped_volume    => 'LJDCBAHKWVMADF',
                spa               => 'VSWUCAHKWVMADF',
                iscsi             => 'ZOFNAAHKWVMADF',
                fcp               => 'XDBRBAHKWVMADF',
                flexclone         => 'PFMUYZGKWVMADF'
            },
            '1-80-000013' => {
                base              => 'XBOLYLDPMZSEBG',
                mirror            => 'VQJPZLDPMZSEBG',
                flexvol_hpo       => 'HLHEAMDPMZSEBG',
                cifs              => 'TFFTAMDPMZSEBG',
                snaprestore       => 'FADIBMDPMZSEBG',
                insight_balance   => 'JCJMGMDPMZSEBG',
                nfs               => 'DPYLCMDPMZSEBG',
                snapmanager_suite => 'LNNIFMDPMZSEBG',
                snapmirror_dp     => 'PJWADMDPMZSEBG',
                spa               => 'XHLXFMDPMZSEBG',
                striped_volume    => 'NYREEMDPMZSEBG',
                iscsi             => 'BEUPDMDPMZSEBG',
                fcp               => 'ZSPTEMDPMZSEBG',
                flexclone         => 'RUAXBMDPMZSEBG'
            },
            '1-80-000017' => {
                base              => 'ZYSOQSYLTDIFMA',
                mirror            => 'XNOSRSYLTDIFMA',
                flexvol_hpo       => 'JIMHSSYLTDIFMA',
                cifs              => 'VCKWSSYLTDIFMA',
                snaprestore       => 'HXHLTSYLTDIFMA',
                insight_balance   => 'LZNPYSYLTDIFMA',
                nfs               => 'FMDPUSYLTDIFMA',
                snapmanager_suite => 'NKSLXSYLTDIFMA',
                snapmirror_dp     => 'RGBEVSYLTDIFMA',
                spa               => 'ZEQAYSYLTDIFMA',
                striped_volume    => 'PVWHWSYLTDIFMA',
                iscsi             => 'DBZSVSYLTDIFMA',
                fcp               => 'BQUWWSYLTDIFMA',
                flexclone         => 'TRFAUSYLTDIFMA'
            },
            '1-80-123456' => {
                base           => 'NWUZZJPJYBFDAA',
                mirror         => 'LLQDBKPJYBFDAA',
                flexvol_hpo    => 'XFOSBKPJYBFDAA',
                cifs           => 'JAMHCKPJYBFDAA',
                snaprestore    => 'VUJWCKPJYBFDAA',
                nfs            => 'TJFAEKPJYBFDAA',
                striped_volume => 'DTYSFKPJYBFDAA',
                snapmirror_dp  => 'FEDPEKPJYBFDAA',
                snapmirror_ls  => 'RYAEFKPJYBFDAA',
                iscsi          => 'RYAEFKPJYBFDAA',
                fcp            => 'PNWHGKPJYBFDAA',
                flexclone      => 'BYKBBPAMUBDAAA'
            },
            '1-80-000008' => {
                base           => 'JWFJEXMWZWYQSD',
                cifs           => 'FAXQGXMWZWYQSD',
                flexclone      => 'DPSUHXMWZWYQSD',
                flexvol_hpo    => 'TFZBGXMWZWYQSD',
                snaprestore    => 'RUUFHXMWZWYQSD',
                nfs            => 'PJQJIXMWZWYQSD',
                snapmirror_dp  => 'BEOYIXMWZWYQSD',
                striped_volume => 'ZSJCKXMWZWYQSD',
                iscsi          => 'NYLNJXMWZWYQSD',
                fcp            => 'LNHRKXMWZWYQSD'
            }
        }
    }
);

our %Default_Feature_Licenses_V2 = (
#   cdmi                => 'WSQFRWDUSAFDAAAAAAAAAAAAAAAA', Deprecated in FS.0
    cifs                => 'MJXMPWDUSAFDAAAAAAAAAAAAAAAA',
    fcp                 => 'KYSQQWDUSAFDAAAAAAAAAAAAAAAA',
    flexclone           => 'GCKYSWDUSAFDAAAAAAAAAAAAAAAA',
    iscsi               => 'YDVBQWDUSAFDAAAAAAAAAAAAAAAA',
    nfs                 => 'APZXOWDUSAFDAAAAAAAAAAAAAAAA',
    snaplock            => 'ERFCUWDUSAFDAAAAAAAAAAAAAAAA',
    snapmanagersuite    => 'QLDRUWDUSAFDAAAAAAAAAAAAAAAA',
    snapmirror          => 'UHMJSWDUSAFDAAAAAAAAAAAAAAAA',
    snapprotectapps     => 'CGBGVWDUSAFDAAAAAAAAAAAAAAAA',
    snaprestore         => 'INOURWDUSAFDAAAAAAAAAAAAAAAA',
    snapvault           => 'SWHNTWDUSAFDAAAAAAAAAAAAAAAA',
    v_storageattach     => 'AXFTWDBBVAAAAAAAAAAAAAAAAAAA',
    insight_balance     => 'YLBXXDBBVAAAAAAAAAAAAAAAAAAA',
    ocshift             => 'YJSNXWDUSAFDAAAAAAAAAAAAAAAA',
    tpm                 => 'OQVEAEBBVAAAAAAAAAAAAAAAAAAA',
    ve                  => 'ALTTAEBBVAAAAAAAAAAAAAAAAAAA',
);

our %Cloud_License = (
    cloud             => '',
);

# Refer to this link to update the Demo licenses:
# http://licensing.netapp.com/securecc/internal-licenses/feature-demo-licenses-v2-0/
# The demo license will expire on 29-Sep-2025
our %Demo_Licenses_V2 = (
#   cdmi                => 'SOLHQNFLWWPCPFAAAAAAAAAAAAAA', Deprecated in FS.0
    cifs             => 'EJDJUOTPDVHXCFAAAAAAAAAAAAAA',
    fcp              => 'CYYMVOTPDVHXCFAAAAAAAAAAAAAA',
    flexclone        => 'YBQUXOTPDVHXCFAAAAAAAAAAAAAA',
    iscsi            => 'QDBYUOTPDVHXCFAAAAAAAAAAAAAA',
    nfs              => 'SOFUTOTPDVHXCFAAAAAAAAAAAAAA',
    snaplock         => 'WQLYYOTPDVHXCFAAAAAAAAAAAAAA',
    snapmanagersuite => 'ILJNZOTPDVHXCFAAAAAAAAAAAAAA',
    snapmirror       => 'MHSFXOTPDVHXCFAAAAAAAAAAAAAA',
    snapprotectapps  => 'UFHCAPTPDVHXCFAAAAAAAAAAAAAA',
    snaprestore      => 'ANUQWOTPDVHXCFAAAAAAAAAAAAAA',
    snapvault        => 'KWNJYOTPDVHXCFAAAAAAAAAAAAAA',
    v_storageattach  => 'GAFRAPTPDVHXCFAAAAAAAAAAAAAA',
    insight_balance  => 'EPAVBPTPDVHXCFAAAAAAAAAAAAAA',
    ocshift          => 'QJYJCPTPDVHXCFAAAAAAAAAAAAAA',
    tpm              => 'ATRCEPTPDVHXCFAAAAAAAAAAAAAA',
    ve               => 'MNPREPTPDVHXCFAAAAAAAAAAAAAA',
);

our %Default_Licenses_V2 = (
    'CMode' => {
        'filer_or_vsim' => {
            '1-80-000011' => {
                base => 'ETYYFLXUQUMADFAAAAAAAAAAAAAA',
                %Default_Feature_Licenses_V2
            },
            '1-80-000013' => {
                base => 'GINBJXTZGYSEBGAAAAAAAAAAAAAA',
                %Default_Feature_Licenses_V2
            },
            '1-80-000017' => {
                base => 'IFSEBEPWNCIFMAAAAAAAAAAAAAAA',
                %Default_Feature_Licenses_V2
            },
            '1-80-000099' => {
                base => 'UJBGVLVQJHOJKBAAAAAAAAAAAAAA',
                %Demo_Licenses_V2
            }
        },
	'AWS' => {%Cloud_License},
	'AZURE' => {%Cloud_License},
	'SDOT' => {%Cloud_License}
    },
    '7Mode' => {
        'filer_or_vsim' => {%Default_Feature_Licenses_V2},
        'promoted_simulator' => {%Default_Feature_Licenses_V2},
    }
);

our %Node_Locked_License = (
    nfs   => 1,
    cifs  => 1,
    iscsi => 1,
);

# Virtual Attached Storage license, previously named V_Series
# in early V2 implementations
our %Map_Old_Feature_To_New_Feature_V2 = ( v_series => 'v_storageattach', );

1;
