#
# Copyright (c) 2014 NetApp, Inc., All Rights Reserved
# Any use, modification, or distribution is prohibited
# without prior written consent from NetApp, Inc.
#

## @summary VolumeFileClone Task Module
## @author sshaik@netapp.com, dl-nacl-dev@netapp.com
## @status shared
## @pod here

package NACL::STask::VolumeFileClone;

use strict;
use warnings;

use base qw(NACL::C::VolumeFileClone NACL::STask::STask);

use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Params::Validate qw(validate_with SCALAR SCALARREF BOOLEAN OBJECT UNDEF);
use NACL::APISet::Exceptions::TimeoutException;
use NATE::Time qw(timeout2time);

use constant DEFAULT_CLONE_TIMEOUT => 500;
use constant POLL_DELTA            => 5;

use NACL::STask::Volume;
use NACL::GeneralUtils qw(nacl_method_retry);
use NACL::C::Exceptions::VolumeFileClone::CurrentlyBusy qw(:try);
use NACL::C::Exceptions::VolumeFileClone::CloneFileExists qw(:try);

=head1 NAME

NACL::STask::VolumeFileClone

=head1 DESCRIPTION

C<NACL::STask::VolumeFileClone> provides a number of well-defined but potentially
complex or multi-step methods related to Volume File Clone task on ONTAP.

It builds on top of, and is a derived class of C<NACL::C::VolumeFileClone>,
and so it also provides methods that are more in the scope of
individual VolumeFileClone-related commands.  See C<NACL::C::VolumeFileClone> for
details.  This also means that a C<NACL::STask::VolumeFileClone> object may
generally be used in place of a component object.

=head1 METHODS

=head2 create

    my $vol_fileclone_obj = NACL::STask::VolumeFileClone->create(
          command_interface   => $command_interface,
          vserver             => $vserver,
          "source-path"       => $Lun,
          "destination-path"  => $Clone,
          nacltask_wait       => $boolean,  # default '1'
          %other_options
    );



(Class method) creates file or LUN full/sub file clone.. This method returns an
object of type NACL::STask::VolumeFileClone. This method provides
additional services beyond what's inherent to the product's volume file clone
creation commands, as controlled and described by the new "nacltask_wait" option.

Applicable for CMode CLI/ZAPI and 7Mode CLI.

=over

=item Options

=over

=item C<< nacltask_wait => 0|1 >>

(Optional) If 1 (the default), wait for volume clone create job completion
If 0, do not wait.

=item C<< nacltask_retry_tries_count => $tries_count >>

(Optional, defaults to 5)
The number of times to retry the create command if it fails because
the file is temporarily busy.

Note that it is possible to make it not retry the create command by setting
the value of this argument to 1.

=item C<< nacltask_retry_sleep_time => $sleep_time >>

(Optional, defaults to 20 seconds)
The time to sleep between retries of the create command

=item command_interface, apiset_must, apiset_should, method-timeout,etc,.
All of the other various options described in
L<< NACL::C::VolumeFileClone->create | lib-NACL-C-VolumeFileClone-pm/create >>

=back

=back

=cut

sub create {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;

    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            nacltask_wait => {type => SCALAR, default => 1},
            nacltask_retry_tries_count => { type => SCALAR,  optional => 1 },
            nacltask_retry_sleep_time  => { type => SCALAR,  default => 20 },
        },
        allow_extra     => 1,
    );

    # Transform %opts from the options we received into the options to
    # pass to the base class method.

    my $wait = delete $opts{nacltask_wait};
    $opts{'method-timeout'} ||= DEFAULT_CLONE_TIMEOUT;

    #see burt #869112
    $opts{'nacltask_retry_tries_count'} //= int ($opts{'method-timeout'} / 20) + 1;

    my %nacltask_options;
    $pkg->_move_nacltask_options(
         source => \%opts,
         target => \%nacltask_options,
    );

    my %retry_opts;
    $pkg->_copy_retry_params(
        source => \%nacltask_options,
        target => \%retry_opts
    );

    my ($self,$num_times_code_ran);
    $retry_opts{num_times_code_ran} = \$num_times_code_ran;

    try{
        nacl_method_retry(
            code => sub { $self = $pkg->SUPER::create(%opts); },
            exceptions => ['NACL::C::Exceptions::VolumeFileClone::CurrentlyBusy'],
            %retry_opts
            );
    }catch NACL::C::Exceptions::VolumeFileClone::CloneFileExists with {
        my $exception = shift;
        # See burt 869112.
        # If the create originally failed saying "device busy", then when it finally completes, it says "clone file exists"
        # Check if the call to create() was retried, if so then this error is valid
        if ($num_times_code_ran > 1) {
            $Log->exit() if $may_exit;
            $exception->throw();
        }
    };

    if ($wait) {
        my %wait_for_creation_opts;
        $pkg->_hash_copy(
            source => \%opts,
            target => \%wait_for_creation_opts,
            copy   => [qw(volume vserver)]
        );
        $pkg->_copy_common_component_params_with_ci(
            source => \%opts,
            target => \%wait_for_creation_opts
        );
        if (!$opts{volume}) {
            if ($opts{'source-path'} =~ /\/\w+\/(\S+?)\//) {
                $wait_for_creation_opts{volume} = $1;
            }
        }
        $pkg->_hash_move(
            source => \%opts,
            target => \%wait_for_creation_opts,
            move   => [qw(polling_interval)]
        );
        $pkg->wait_for_creation(%wait_for_creation_opts);
    }

    $Log->exit() if $may_exit;
    return $self;
} ## end sub create

=head2 wait_for_creation

    # Class method: wait for volume file clone to get created and come online
    NACL::STask::VolumeFileClone->wait_for_creation(
        "method-timeout"  => $timeout,
        command_interface => $command_interface,
        volume            => $volume,
        vserver           => $vserver,
    );

Wait for the volume file clone creation to complete. If a file clone is creating
using the snapshots of the volume then it waits until the value of "snaprefs" of
the "clone list"(nodescope command) is "undef".

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< "method-timeout"=>$timeout >>

(Optional) How long in seconds to wait for the job to finish or for
the state to become "online".  Defaults to 7200 (2 hours).

=item C<< polling_interval => $interval >>

(Optional)
The interval, in seconds, at which to either show the progress of the job (if
we're watching its progress) or the interval at which to poll the value of the
"state" field. This defaults to 10 seconds if not provided.

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=back

=back

=over

=item Exceptions

=over

=item C<NACL::Exceptions::Timeout>

This type of exception is thrown when clone operation is not completed and the maximum time limit is crossed.

=back

=back

=cut

sub wait_for_creation {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            'nodescope-node-name' => {type => SCALAR, optional => 1},
            polling_interval      => {type => SCALAR, default  => POLL_DELTA},
            volume                => {type => SCALAR},
            vserver               => {type => SCALAR, optional => 1},
        },
    );

    my $apiset   = $opts{command_interface}->get_7m_or_nodescope_apiset();
    my $timeout  = $opts{method_timeout} || DEFAULT_CLONE_TIMEOUT;
    my $end_time = timeout2time($timeout);
    my ($done, %args,%common_opts);

    if ($opts{command_interface}->is_cmode()) {
        my @requested_fields = ('name-ordinal');
        if ($opts{'nodescope-node-name'}) {
            $args{'nodescope-node-name'} = $opts{'nodescope-node-name'};
        } else {
            push(@requested_fields, 'node');
        }
        $pkg_or_obj->_copy_common_component_params_with_ci(
               source => \%opts,
               target => \%common_opts,
        );
        my $vol_ordinal = NACL::CS::Volume->fetch(
            %common_opts,
            filter            => {
                volume  => $opts{volume},
                vserver => $opts{vserver},
            },
            requested_fields => \@requested_fields
        );
        $args{'nodescope-node-name'} = $vol_ordinal->node() if (!$opts{'nodescope-node-name'});
        my $name_ordinal = $vol_ordinal->name_ordinal();
        if ($name_ordinal !~ /base/) {
            $args{volume} = $opts{volume} . "(" . $name_ordinal . ")";
        } else {
            $args{volume} = $opts{volume};
        }
        $args{header} = 1;
    } else {
        $args{header} = 1;
        $args{volume} = $opts{volume};
    }
    while (!$done) {
        my $response;
        $response = $apiset->clone_list(%args);
        my $output = $response->get_parsed_output();
        my $snaprefs_counter
            = $output->[0]->{clone_map_file_header}->[0]->{'snaprefs'};
        if (!defined($snaprefs_counter)) {
            $Log->trace("Clone operation has completed successfully");
            $done = 1;
        } else {
            if (time() > $end_time) {
                $Log->exit() if $may_exit;
                NACL::Exceptions::Timeout->throw(
                    "Clone operation is not completed and the maximum time limit $timeout is crossed",
                    timeout => $timeout
                );
            } else {
                Tharn::snooze($opts{polling_interval});
            }
        }
    }
    $Log->exit() if $may_exit;
} ## end sub wait_for_creation
1;
