#
# Copyright (c) 2015 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary VolumeFile Task Module
## @author dl-nacl-dev@netapp.com
## @status shared
## @pod here

package NACL::STask::VolumeFileFingerprint;

use strict;
use warnings;

use base qw(NACL::C::VolumeFileFingerprint NACL::STask::STask);

use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();

=head1 NAME

NACL::STask::VolumeFileFingerprint

=head1 DESCRIPTION

C<NACL::STask::VolumeFileFingerprint> provides a number of well-defined but
potentially complex or multi-step methods related to Volume Files in
ONTAP.

It builds on top of, and is a derived class of C<NACL::C::VolumeFileFingerprint>,
and so it also provides methods that are more in the scope of
individual VolumeFileFingerprint related commands.  See C<NACL::C::VolumeFileFingerprint> for
details.  This also means that a C<NACL::STask::VolumeFileFingerprint> package may
generally be used in place of C<NACL::C::VolumeFileFingerprint> package.

=head1 METHODS

=head2 start_and_dump

    my $dump_cs_obj = NACL::STask::VolumeFileFingerprint->start_and_dump(
        command_interface => $ci,
        vserver           => $vs,
        file              => $file_path,
        ...
    );

B<NOTE:> If you have a C<NACL::STask::VolumeFile> object available in the
script, then this can be invoked directly by doing:

    # Command_interface, vserver and file path are extracted from $file_obj
    # and do not need to be specified
    $file_obj->show_fingerprint();

This method is a synchronous API for getting the file fingerprint details. In
other words, this method performs the following:

* Starts the file fingerprint operation (using the C<start> method).
* Waits till it completes (using the L<wait_for_completion|lib-NACL-STask-VolumeFileFingerprint-pm/wait_for_completion> method).
* Queries for the fingerprint (using the C<dump> method).

It returns a
L<NACL::CS::VolumeFileFingerprint|lib-NACL-CS-VolumeFileFingerprint-pm>
object.

The options accepted by this method are the same as those accepted by
the L<start|lib-NACL-C-VolumeFileFingerprint-pm/start> method.

By default, this method waits for 600 seconds for the fingerprint operation
to complete. If the operation might take longer, then the higher
timeout value should be specified through C<method-timeout>.

B<FOR 7Mode:> This method works for 7Mode as well. For 7Mode, this simply
invokes the C<dump> method directly since file fingerprint is not
asynchronous for 7Mode.

=cut

sub start_and_dump {
    $Log->enter();
    my ($pkg, @args) = @_;
    my %opts = $pkg->_common_validate_with(
        params              => \@args,
        ignore_primary_keys => 1,
        allow_extra         => 1,
    );
    my $dump_obj;
    if ($opts{command_interface}->is_cmode()) {
        my $fingerprint_obj = $pkg->start(%opts);
        my %common_opts;
        $pkg->_copy_common_component_params(
            source => \%opts,
            target => \%common_opts,
        );
        $fingerprint_obj->wait_for_completion(%common_opts);
        $dump_obj = $fingerprint_obj->dump(%common_opts);
    } else {
        my %common_opts_ci;
        $pkg->_move_common_component_params_with_ci(
            source => \%opts,
            target => \%common_opts_ci,
        );
        $dump_obj = $pkg->dump(%common_opts_ci, filter => {%opts});
    }
    $Log->exit();
    return $dump_obj;
}

=head2 wait_for_completion

    # Package call
    NACL::STask::VolumeFileFingerprint->wait_for_completion(
        command_interface => $ci,
        'session-id'      => $session_id,
        ...
    );
    
    # Instance call
    $fingerprint_stask_obj->wait_for_completion();

(Applicable only for CMode)
Wait for completion of the file fingerprint operation. This is done by
waiting for the file fingerprint's C<progress-percentage> to reach 100. By
default, this method waits for 600 seconds for the fingerprint operation to
complete. If the operation might take longer, then a higher timeout value should
be specified through C<method-timeout>.

The options this method accepts are C<command_interface>, C<session-id>,
C<method-timeout>, C<polling_interval> along with options common to all
Component/STask methods such as C<apiset_must>, C<apiset_should>.

=cut

sub wait_for_completion {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    
	# Do not have this wait on progress-percentage because when it fails
	# it just shows 0% and Completed, so would get stuck infinitely.
	# Just break out on "Completed", the "dump" will take care of
	# checking if it actually succeeded.
	# Burt 922818.
	$pkg_or_obj->wait_on_attributes(
        attributes => [
            {
                attribute_to_check  => 'operation-status',
                till_value          => 'Completed',
                valid_intermediate_states => [qw(Unknown In-Progress)],
            },
        ],
        'method-timeout'   => 600,
        @args,
    );
    $Log->exit();
}

1;
