=head1 NAME

NACL::STask::VserverCifsUsersAndGroupsLocalGroup

=head1 DESCRIPTION

C<NACL::STask::VserverCifsUsersAndGroupsLocalGroup> provides a number of well-defined but
potentially complex or multi-step methods related to Vserver CIFS Users and
Group in ONTAP.

It builds on top of, and is a derived class of C<NACL::C::VserverCifsUsersAndGroupsLocalGroup>,
and so it also provides methods that are more in the scope of
individual vserver-related commands.  See C<NACL::C::VserverCifsUsersAndGroupsLocalGroup> for
details.  This also means that a C<NACL::STask::VserverCifsUsersAndGroupsLocalGroup> object may
generally be used in place of a component object.


=head1 METHODS

=head2 create

This accepts the same options and provides the same facilities as
L<NACL::C::VserverCifsUsersAndGroupsLocalGroup::create|lib-NACL-C-VserverCifsUsersAndGroupsLocalGroup-pm/create>.
Additionally, the STask provides the following facilities: C<Cleanup>,
C<nacltask_verify>, C<MCC checks>. The description of these features are
present at:
L<Common Create options|lib-NACL-STask-STask-pm/Common_Create_Options>.

=head2 purge

This accepts the same options and provides the same facilities as
L<NACL::C::VserverCifsUsersAndGroupsLocalGroup::delete|lib-NACL-C-VserverCifsUsersAndGroupsLocalGroup-pm/delete>.
Additionally, the STask provides the following facilities: C<nacltask_verify>,
C<MCC checks>. The description of these features are
present at:
L<Common Purge options|lib-NACL-STask-STask-pm/Common_Purge_Options>.

=head2 modify

This accepts the same options and provides the same facilities as
L<NACL::C::VserverCifsUsersAndGroupsLocalGroup::modify|lib-NACL-C-VserverCifsUsersAndGroupsLocalGroup-pm/modify>.
Additionally, the STask provides the following facilities: C<nacltask_verify>.
The description of these features are present at:
L<Common modify options|lib-NACL-STask-STask-pm/Common_Modify_Options>.

=cut

package NACL::STask::VserverCifsUsersAndGroupsLocalGroup;

use strict;
use warnings;
use base qw(NACL::C::VserverCifsUsersAndGroupsLocalGroup NACL::STask::STask);

use NATE::Log qw( log_global );
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Params::Validate qw(validate SCALAR BOOLEAN SCALARREF);
use NACL::ComponentUtils qw (_optional_scalars Dumper);
use NACL::Exceptions::VerifyFailure                     ();
use NATE::Exceptions::Argument qw(:try);
use NACL::APISet::Exceptions::ResponseException ();
use NACL::C::Exceptions::VserverCifsUsersAndGroupsLocalGroup::MemberAlreadyExists qw(:try);
use NACL::C::Exceptions::VserverCifsUsersAndGroupsLocalGroup::MemberDoesNotExist  ();

__PACKAGE__->create_method_builder();
__PACKAGE__->purge_method_builder();
__PACKAGE__->modify_method_builder();


=head2 add_members

   NACL::STask::VserverCifsUsersAndGroupsLocalGroup->add_members(
                                'command_interface'  => $command_interface,
                                'group-name'         => $name,
                                'member-names'       => $members
                                'nacltask_if_exists' => $action, # default die
                                %other_options,
                    );

(Class Method) This method is used to add members to loacl group.
If member already exists, it will perform the action based on
"nacltask_if_exists" parameter. Default behavior would be "die".

Uses CMode CLI/ZAPI APISet.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< 'group-name' => $name >>

(Required) Name of the local group.

=item C<< 'member-names' => $members >>

(Required) An Arrayref of member names.

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "nacltask_if_exists => $action" >>

(Optional) What to do if the members were already added.
If $action is "die", then fail with an exception. If action is
"purge", then remove the members from the group  
before adding them. If $action is "reuse", It will return
the object of existing members.

=item C<< "nacltask_verify => $action" >>

(Optional) The user of this library can specify to verify whether the members
is created or not. If the action is 0, which is default, it wont
verify the creation. If the user sets the action to 1, it will verify the
it using the component state of this library.

=item C<< "_was_created => \$scalar" >>

(Optional) When this option is provided with a reference to a scalar
variable, the variable gets filled in with a boolean value
describing whether the members were available (value will be 0; this
scenario is possible when if_exists => "reuse") or whether the
members were created (value will be 1). This is necessary to
determine whether the members need to be cleaned up later.

my $was_created;

NACL::STask::VserverCifsUsersAndGroupsLocalGroup->add_members(
                        command_interface       => $command_interface,
                        $nacltask_if_exists     => 'reuse',
                        _was_created            => \$was_created,
                        nacltask_to_cleanup     => $boolean,
                        %other_opts
                    );

# Now determine whether to clean up the members, since we're not sure
# whether we reused an existing members or created a new one
if ($was_created) {
   # New members were created. Clean it up.
   NACL::STask::VserverCifsUsersAndGroupsLocalGroup->remove_members(
                        command_interface       => $command_interface,
                        %other_opts
                    );
}

=item C<< nacltask_to_cleanup => 0|1 >>

(Optional, defaults to 0(no to cleanup))
Flag indicating if this operation is to be cleaned up or not.

=item C<< nacltask_cleanup_manager >>

Cleanup manager to use for registering.
Default : Will use the default cleanup manager.

command_interface, apiset_must, apiset_should, mode, etc.
All of the other various options to
L<< NACL::C::VserverCifsUsersAndGroupsLocalGroup->create|lib-NACL-C-VserverCifsUsersAndGroupsLocalGroup-pm/create >>

=back

=over

=item Exceptions

=over

=item C<NACL::C::Exceptions::VserverCifsUsersAndGroupsLocalGroup::MemberAlreadyExists>

This type of exception is thrown when an attempt is made to add members who already exist.

=item C<NACL::Exceptions::VerifyFailure>

This type of exception is thrown when verification fails for the added members.

=back

=back

=back

=cut

sub add_members {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = $pkg->_common_validate_with(
        params          => \@_,
        additional_spec => {
            nacltask_if_exists => $pkg->_if_exists_validate_spec(),
            nacltask_verify    => { type => BOOLEAN, default => 0 },
            _was_created       => { type => SCALARREF, optional => 1 },
            $pkg->_cleanup_validate_spec(),
        },
        allow_extra     => 1
    );

    # Transform %opts from the options we received into the options to
    # pass to the base class method.
    my ( $self, %common_opts, %nacltask_opts );
    $pkg->_copy_common_component_params_with_ci(
        source => \%opts,
        target => \%common_opts
    );

    my ( %opts_for_cleanup, %opts_for_register, $nacltask_to_cleanup );
    $pkg->_move_common_cleanup_opts(
        source => \%opts,
        target => \%opts_for_cleanup,
    );

    $pkg->_move_nacltask_options(
        source => \%opts,
        target => \%nacltask_opts
    );

    my $was_created = delete $opts{_was_created};
    $$was_created = 0;
    my $if_exists = $nacltask_opts{'nacltask_if_exists'};
    my $verify    = $nacltask_opts{'nacltask_verify'};

ADD: {
        use warnings;
        try {
            $self         = $pkg->SUPER::add_members(%opts);
            $$was_created = 1;
        }
        catch NACL::C::Exceptions::VserverCifsUsersAndGroupsLocalGroup::MemberAlreadyExists with {
            my $exception = shift;
            $Log->comment("Member already Exists!");
            if ( $if_exists eq 'die' ) {
                $Log->exit() if $may_exit;
                $exception->throw();
            } elsif ($if_exists eq 'resue') {
                $Log->debug("Exception suppressed because 'if_exists' "
                                    . "was sent as '$if_exists'");
            } elsif ($if_exists eq 'purge') {
                $pkg->remove_members(
                                        'command_interface' => $opts{command_interface},
                                        'vserver'           => $opts{vserver},
                                        'group-name'        => $opts{'group-name'},
                                        'members'           => $opts{'members'},
                                    );
                
                no warnings qw(exiting);
                redo ADD;
            }
        };
    }
    if ($verify) {
        $pkg->verify_state(%opts);
    }

    $Log->exit() if $may_exit;
    return;
} ## end sub add_members

=head2 remove_members

    NACL::STask::VserverCifsUsersAndGroupsLocalGroup->remove_members(
                'command_interface' => $command_interface,
                'vserver'           => $vserver,
                'group-name'        => $name,
                'members'           => $members,
                'nacltask_if_purged => $action,
                );

(Class or instance method)
This method is used to delete the loacl group members.

Uses a CMode CLI/ZAPI APISet.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
A component object that represents the host which to send commands.
See NACL::C::Component::command_interface

=item C<< 'vserver' => $vserver >>

(Required for class method, Not Applicable for instance method)
The name of the vserver for the snap mirror policy.

=item C<< 'group-name' => $name >>

(Required) Name of the local group.

=item C<< 'member-names' => $members >>

(Required) An Arrayref of member names.

=item C<< _was_deleted => \$scalar >>

(Optional)
When this option is provided a reference to a scalar variable, the variable
gets filled in with a boolean value describing whether the member
did not exist (value will be 0; this scenario is possible when
nacltask_if_purged => "pass") or whether the member was deleted
(value will be 1).

=item C<< nacltask_if_purged => $nacltask_if_purged >>

(Optional)
If 'pass' (default), It will pass if the member is already deleted.
If 'fail', It will fail if the member is already deleted.

All of the other various options to
L<< NACL::C::VserverCifsUsersAndGroupsLocalGroup->remove_members|lib-NACL-C-VserverCifsUsersAndGroupsLocalGroup-pm/remove_members >>

=back

=over

=item Exceptions

=over

=item C<NACL::C::Exceptions::VserverCifsUsersAndGroupsLocalGroup::MemberDoesNotExist>

This type of exception is thrown when an attempt is made to delete a member that does not exists.

=item C<NACL::Exceptions::VerifyFailure>

This type of exception is thrown when verification fails for the deleted member.

=back

=cut

sub remove_members {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my $additional_spec = {
        nacltask_if_purged => $pkg_or_obj->_if_action_completed_already_validate_spec(),
        _was_deleted       => { type => SCALARREF, optional => 1 },
    };

    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => $additional_spec,
        allow_extra => 1,
    );

    my %nacltask_opts;
    $pkg_or_obj->_move_nacltask_options(
        source => \%opts,
        target => \%nacltask_opts,
    );

    my $if_purged   = $nacltask_opts{nacltask_if_purged};

    my %common_opts;
    $pkg_or_obj->_copy_common_component_params_with_ci(
        source => \%opts,
        target => \%common_opts
    );

    my $was_deleted = delete $opts{_was_deleted};
    $$was_deleted = 0;

    use warnings qw(exiting);
    try {
        $pkg_or_obj->SUPER::remove_members(%common_opts, %opts );
        $$was_deleted = 1;
    }
    catch NACL::C::Exceptions::VserverCifsUsersAndGroupsLocalGroup::MemberDoesNotExist with {
        my $exception = shift;
        $Log->comment("Entry Does Not Exists!");
        if ( $if_purged !~ /pass/i ) {
            $Log->exit() if $may_exit;
            $exception->throw();
        }
    };

    $Log->exit() if $may_exit;
} ## end sub remove_members

1;
