#
# Copyright (c) 2014 NetApp, Inc.
# All rights reserved.
#
## @summary VserverSecurityTraceTraceResult Task Module
## @author namitha.k@netapp.com
## @status shared
## @pod here


package NACL::STask::VserverSecurityTraceTraceResult;

use strict;
use warnings;

use base qw(NACL::C::VserverSecurityTraceTraceResult NACL::STask::STask);

use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Params::Validate qw(validate SCALAR SCALARREF BOOLEAN ARRAYREF);
use NATE::Exceptions::Argument qw(:try);

use NACL::APISet::Exceptions::ResponseException ();



=head1 NAME

NACL::STask::VserverSecurityTraceTraceResult

=head1 DESCRIPTION

C<NACL::STask::VserverSecurityTraceTraceResult> provides a number of well-defined but
potentially complex or multi-step methods related to VserverSecurityTraceTraceResult in
ONTAP.

It builds on top of, and is a derived class of C<NACL::C::VserverSecurityTraceTraceResult>,
and so it also provides methods that are more in the scope of
individual VserverSecurityTraceTraceResult-related commands.  See C<NACL::C::VserverSecurityTraceTraceResult>
for details.  This also means that a C<NACL::STask::VserverSecurityTraceTraceResult> object may
generally be used in place of a component object.

=head1 ATTRIBUTES

=head2 command_interface

(Required) A component object that represents the host to which to send commands.
See L<NACL::C::Component::command_interface|lib-NACL-C-Component-pm/command_interface>.

=head1 METHODS

=head2 purge

    NACL::C::VserverSecurityTraceTraceResult->purge(
                            command_interface  => $command_interface,
                            seqnum             => $seqnum,
                            node               => $node,
                            vserver            => $vserver,
                            nacltask_verify    => 0|1, #default 0
                            nacltask_if_purged => $if_purged,    
                                ...);

                                 or

    $VserverSecurityTraceTraceResult_obj->purge(...);

(Class or instance method)
This method is used to delete Sectrace results.
Uses a CMode CLI.

=over

=item Options

=over

=item C<< 'seqnum' => $seqnum >>
(Required for class method, Not Applicable for instance method)
Sequence number

=item C<< 'vserver' => $vserver >>
(Required for class method, Not Applicable for instance method)
Vserver

=item C<< 'node' => $node >>
(Required for class method, Not Applicable for instance method)
Node

=item C<< "nacltask_verify => 0|1" >>
(Optional)when set to 1, Vserver Security Trace Trace-Results deletion will be verified.
          when set to 0, the verification step is skipped.
          by default its set to 0.

=item C<< "nacltask_if_purged => $if_purged" >> 
(Optional)If trace-result has already been deleted or does not exist
          when set to die, fails with an exception.
		  when set to pass, no exception is thrown.


=back

=back

=over

=item Exceptions

=over

=item C<NACL::APISet::Exceptions::ResponseException>

This type of exception is thrown when an attempt is made to delete vserver security trace results that does not exists.

=item C<NATE::Exceptions::VerifyFailure>

This type of exception is thrown when verification fails for the deleted vserver security trace results.

=back

=back

=cut

sub purge {
    $Log->enter();
    my $pkg_or_obj      = shift;
    my %orig_opts       = @_;
    my $additional_spec = {
        nacltask_verify    => { type => BOOLEAN, default => 0 },
        nacltask_if_purged => { type => BOOLEAN, default => 0 }
    };

    # Throw away options specific to the task
    $pkg_or_obj->_hash_move(
        source => \%orig_opts,
        target => {},
        move   => [ keys %{$additional_spec} ]
    );
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@_,
        additional_spec => $additional_spec,
    );

    my %common_opts;
    $pkg_or_obj->_copy_common_component_params_with_ci(
        source => \%opts,
        target => \%common_opts
    );
    my $if_purged = $opts{nacltask_if_purged};
	
	AGAIN: {
        use warnings;
        try {
            $pkg_or_obj->delete( %common_opts, %orig_opts );
        }
        catch NACL::APISet::Exceptions::ResponseException with {
            my $e = shift;
            if (   $e->text =~ /.*command\s+failed.*entry doesn't exist/i
                || $e->text
                =~ /.*Either\s+specify.*or set at least one key to.*./i
                || $e->text
                =~ /.*no entries matching your query.*/i ) {

                # Check if the command failed because the sectrace filter
                # does not exist
                if ( $if_purged !~ /pass/i ) {
                    $Log->exit();
                    $e->throw();
                }
            } else {
                $Log->exit();
                $e->throw();
            }
        };
    }

    if ( $opts{nacltask_verify} ) {
        $pkg_or_obj->_generic_purge_verify(%orig_opts);
    }
    $Log->exit();
} ## end sub purge

1;



			
			
			
			
