#
# Copyright (c) 2013 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary VserverServicesNameServiceNsSwitch Task Module
## @author sshaik@netapp.com, dl-nacl-dev
## @status shared
## @pod here

package NACL::STask::VserverServicesNameServiceNsSwitch;

use strict;
use warnings;

use base qw(NACL::C::VserverServicesNameServiceNsSwitch NACL::STask::STask);
use Params::Validate qw(validate SCALAR validate_with SCALARREF BOOLEAN );
use NACL::ComponentUtils qw(_optional_scalars);
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use NATE::Exceptions::Argument qw(:try);
use NATE::BaseException  ();
use NACL::C::Exceptions::VserverServicesNameServiceNsSwitch::AlreadyExists  ();
use NACL::C::Exceptions::VserverServicesNameServiceNsSwitch::DoesNotExist   ();
use NACL::STask::_VserverServices qw(:all);


=head1 NAME

NACL::STask::VserverServicesNameServiceNsSwitch

=head1 DESCRIPTION

C<NACL::STask::VserverServicesNameServiceNsSwitch> provide methods to create, modify
and purge ns-switch entry for vserver in ONTAP.

It is created on top of C<NACL::C::VserverServicesNameServiceNsSwitch> component
Since it is derived class of C<NACL::C::VserverServicesNameServiceNsSwitch>,
we can use all the methods of C<NACL::C::VserverServicesNameServiceNsSwitch>
from the object of this task.

=head1 ATTRIBUTES

=head2 command_interface

(Required) A component object that represents the node to which to send commands.
See L<NACL::C::Component::command_interface|lib-NACL-C-Component-pm/command_interface>.

=head2 vserver

(Required) The name of the Vserver to be used.

=head2 database

(Required) The Name service switch database.

=head1 EXCEPTIONS

    1. NACL::C::Exceptions::VserverServicesNameServiceNsSwitch::AlreadyExists
       (If another ns-switch entry is already present for the vserver & nacltask_if_exists option is set to "die").

    2. NACL::C::Exceptions::VserverServicesNameServiceNsSwitch::DoesNotExist
      (Exception is raised if a ns-switch entry for the particular vserver does not exists).

    3. NACL::Exceptions::VerifyFailure
      (Exception is raised if the value of database or sources are not found in the conf file).

=head1 METHODS

=head2 create

   my $ns_switch_obj = NACL::STask::VserverServicesNameServiceNsSwitch->create(
                                'command_interface'  => $ci,
                                'vserver'            => $vserver,
                                database             => $database,
                                'sources'            => \@Sources,
                                nacltask_verify      => $action, # default to 0
                                'nacltask_if_exists' => $action, # default die
                                _was_created         => \$scalar,
                                nacltask_verify_using_state_only  => $boolean, #default to 1
                                nacltask_verify_in_conf_file_only => $boolean, #default to 0
                                nacltask_verify_file_version      => $boolean, #default to 0
                                nacltask_verify_file_version_retry => $retry, #default to 6
                                nacltask_verify_file_version_interval => $interval, #default to 10
                                nacltask_conf_file_path  => $file_path,
                                nacltask_to_cleanup      => 1, #default to 0
                                nacltask_cleanup_manager => $cleanupobj,
                                %other_options,
                    );

(Class Method) This method is used to create a ns-switch entry on the particular vserver.
If ns-switch entry already exists, it will perform the action based on
"nacltask_if_exists" parameter. Default behavior would be "die".

Uses CMode CLI/ZAPI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component::command_interface|lib-NACL-C-Component-pm/command_interface>.

=item C<< 'vserver' => $vserver >>

(Required) Name of the vserver to be used.

=item C<< "database" => $database >>

(Required) This parameter is used to specify the name
service database for which the order of the source lookup is being specified.
Possible values for this parameter are (hosts|group|passwd|netgroup|namemap).

=item C<< "sources" => [$source1, $source2,..] >>

This parameter used to specify the name service sources and the order
in which to look them up for the specified vserver and name service database.
Possible values for this parameter are (files|dns|ldap|nis).

=item C<< "nacltask_if_exists => $action" >>

(Optional) What to do if the DNS to be configured already exists.
If $action is "die" (default), then fail with NACL::C::Exceptions::VserverServicesNameServiceNsSwitch::AlreadyExists.
If $action is "purge", then purge the ns-switch entry(see the "purge" method below)
before creating a new one.
If $action is "reuse", It will return the object of existing ns-switch entry for the vserver.

=item C<< "nacltask_verify => $action" >>

(Optional) The user of this library can specify to verify whether the ns-switch entry created or not.
If the action is 0(default), verification is not done.
If the user sets the action to 1, it will verify it using the component state of this library
and also check whether it exists in nsswitch.conf files of that particular vserver.

=item C<< "_was_created => \$scalar" >>

(Optional) When this option is provided with a reference to a scalar
variable, the variable gets filled in with a boolean value
describing whether the ns-switch entry was available (value will be 0; this
scenario is possible when if_exists => "reuse") or whether the
ns-switch entry was created (value will be 1). This is necessary to
determine whether the ns-switch needs to be cleaned up later.

my $was_created;
my $nsswitch_obj = NACL::STask::VserverServicesNameServiceNsSwitch->create(
                        nacltask_if_exists     => 'reuse',
                        _was_created            => \$was_created,
                        %other_opts
                    );
# Operate on $nsswitch_obj here
# Now determine whether to clean up the ns-switch entry, since we're not sure
# whether we reused an existing ns-switch entry or created a new one
if ($was_created) {
   # Created a new one, Clean it up.
   $nsswitch_obj->purge();
}

=item C<< nacltask_verify_using_state_only => $boolean  >>

(Optional) The user of this library can specify to verify
the ns-switch entry created or not through component state method or not.
If the action is 1 (default), verification will be done through component state.
If the user sets the action to 0, it will not verify through state method.

=item C<< nacltask_verify_in_conf_file_only => $boolean >>

(Optional) The user of this library can specify to verify
the ns-switch entry created or not through ns-switch conf file.
If the action is 1, verification will be done by checking the  nsswitch.conf
file of that particular vserver.
If the user sets the action to 0(default), it will not verify by checking the  nsswitch.conf
file of that particular vserver.
(This option is deprecated)

=item C<<nacltask_verify_file_version => $boolean >>

This option verifies whether the file-version and rdb-version are in sync
within the same node and across all nodes in a cluster.
It defaults to 0 , verification will not  be done.
If set to 1, it will verify.

=item C<<nacltask_verify_file_version_retry => $scalar >>

The user of this library can verify the file-version command based on number of retries.
This option will be set to 6(default),means a user can do 6 retries to verify
file-verison and rdb-version are in sync.

=item C<<nacltask_verify_file_version_interval => $scalar >>

The user of this library can verify the file-version command based on an interval of 10seconds.
This option will be set to 10(default),means a user can execute the command with an interval of 10seconds
in 6 retries..

=item C<< nacltask_conf_file_path => $file_path >>

(Optional) The user of this library can specify the file path to be verified for
the created ns-switch entry.

=item C<< nacltask_to_cleanup => 0|1 >>

(Optional, defaults to 0(no to cleanup))
Flag indicating if this operation is to be cleaned up or not.

=item C<MCC options>

The options accepted for MCC configuration replication verification is
documented at L<MCC options|lib-NACL-STask-STask-pm/mcc_options>.

=item C<< nacltask_cleanup_manager >>

Cleanup manager to use for registering.
Default : Will use the default cleanup manager.

command_interface, apiset_must, apiset_should, etc. All of
the other various options,
See L<< NACL::C::VserverServicesNameServiceNsSwitch::create|lib-NACL-C-VserverServicesDns-pm/create >>

=back

=back

=cut

sub create {
    $Log->enter() if $may_enter;
    my ( $pkg, @args ) = @_;
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            nacltask_if_exists       => $pkg->_if_exists_validate_spec(),
            nacltask_verify          => { type => BOOLEAN, default => 0 },
            _was_created             => { type => SCALARREF, optional => 1 },
            nacltask_verify_using_state_only  => { type => BOOLEAN, default => 1 },
            nacltask_verify_in_conf_file_only => { type => BOOLEAN, default => 0 },
            nacltask_conf_file_path           => { type => SCALAR, optional => 1 },
            nacltask_verify_file_version      => { type => BOOLEAN, default => 0 },
            nacltask_verify_file_version_retry => { type => SCALAR, default => 6 },
            nacltask_verify_file_version_interval => { type => SCALAR, default => 10 },
            $pkg->_cleanup_validate_spec(),
            $pkg->_mcc_validate_spec(),
        },
        allow_extra => 1
    );

    if ( $opts{enabled} ) {
       delete $opts{enabled};
       $Log->warn("\n Passing enabled option to NACL::STask::VserverServicesNameServiceNsSwitch".
                   "->create is not valid anymore, removing this option from input. Please update your scripts");
    } 

    # Transform %opts from the options we received into the options to
    # pass to the base class method.
    my ( $self, %common_opts, %nacltask_opts, %opts_for_cleanup,
    $nacltask_to_cleanup, %opts_for_register );

    $pkg->_move_common_cleanup_opts(
        source              => \%opts,
        target              => \%opts_for_cleanup,
        nacltask_to_cleanup => \$nacltask_to_cleanup,
    );

    my %mcc_opts;
    $pkg->_move_common_mcc_opts(
        source => \%opts,
        target => \%mcc_opts,
    );

    $pkg->_copy_common_component_params_with_ci(
        source => \%opts,
        target => \%common_opts
    );
    $pkg->_move_nacltask_options(
        source => \%opts,
        target => \%nacltask_opts
    );

    my $was_created = delete $opts{_was_created};
    $$was_created = 0;
    my $if_exists                = $nacltask_opts{'nacltask_if_exists'};
    my $verify                   =  $nacltask_opts{'nacltask_verify'};

    CREATE: {
        use warnings;
        try {
            $self         = $pkg->SUPER::create(%opts);
            $$was_created = 1;
        } catch NACL::C::Exceptions::VserverServicesNameServiceNsSwitch::AlreadyExists with {
            my $exception = $_[0];
            $Log->debug(
                "STask The ns-switch entry already exists $nacltask_opts{'nacltask_if_exists'}"
            );
            $self = $pkg->_element_exists_handler(
                create_opts        => \%opts,
                nacltask_if_exists => $if_exists,
                exception          => $exception
            );
            if ( !$self ) {
                no warnings qw(exiting);
                redo CREATE;
            }

        };
    }

    $self->_check_in_metrocluster_node(
        command_interface => $opts{command_interface},
        method_used       => "create",
        %mcc_opts,
    );

    if ( $nacltask_opts{'nacltask_verify_in_conf_file_only'} eq '1' ) {
        $Log->warn(" nacltask_verify_in_conf_file_only Option is deprecated."
                   . "Update your script to not to pass this option");
    }
    if ( $nacltask_opts{'nacltask_verify_file_version'} eq '1' ) {
          $pkg->_file_version_show(%opts);
    }

    if($$was_created) {
        # Verification based on nacltask_opts
        $pkg->verify_state(%opts)
            if ($verify || $nacltask_opts{nacltask_verify_using_state_only});

        # Register this resource with the Cleanup Manager for cleanup
        if ($nacltask_to_cleanup) {
            $pkg->_copy_common_opts_for_cleanup(
                'source'              => {%opts, %opts_for_cleanup},
                'target'              => \%opts_for_register,
                'nacltask_to_cleanup' => \$nacltask_to_cleanup,
                'to_cleanup'          => 'purge'
            );

            $pkg->_register_for_cleanup(%opts_for_register);
        }
    }
    $Log->exit() if $may_exit;
    return $self;
} ## end sub create

=head2 purge

    $nsswitch_obj->purge();

            (or)

    NACL::STask::VserverServicesNameServiceNsSwitch->purge(
            'command_interface' => $ci,
            'vserver'           => $vserver_name,
            database            => $Database,
            'nacltask_verify'   => 1,
            nacltask_verify_using_state_only  => $boolean, #default to 1
            nacltask_verify_in_conf_file_only => $boolean, #default to 0 
            nacltask_conf_file_path => $file_path,
            nacltask_if_purged      => pass|fail,
    );

(Class or instance method)
This method is used to remove ns-switch entry for a particular vserver.

Uses CMode CLI/ZAPI.

=over

=item Options

=over

=item C<< command_interface => $ci >>

(Required for class method, Not Applicable for instance method)
A component object that represents the host which to send commands.
See L<NACL::C::Component::command_interface|lib-NACL-C-Component-pm/command_interface>.

=item C<< 'vserver' => $vserver >>

(Required for class method, Not Applicable for instance method)
Name of the vserver to be used.

=item C<< "database" => $database >>

(Required) This parameter is used to specify the name
service database of the vserver, for which the name service switch
entry is to be deleted.
Possible values for this parameter are (hosts|group|passwd|netgroup|namemap).

=item C<< nacltask_verify => $nacltask_verify_boolean >>

(Optional)
If '0' (default), verification will not be performed.
If '1', verification will be performed to ensure that the deletion did happen
successfully and also checks whether it got deleted from nsswitch.conf
file of that particular vserver.

=item C<< nacltask_if_purged => $nacltask_if_purged >>

(Optional) Default value is 'fail'.
If 'pass', It will pass if the ns-switch entry is already deleted.
If 'fail', NACL::C::Exceptions::VserverServicesNameServiceNsSwitch::DoesNotExist type of exception is raised.

=item C<< nacltask_verify_using_state_only => $boolean  >>

(Optional) The user of this library can specify to verify
the ns-switch entry deleted or not through component state method or not.
If the action is 1 (default), verification will be done through component state.
If the user sets the action to 0, it will not verify through state method.

=item C<< nacltask_verify_in_conf_file_only => $boolean  >>

(Optional) The user of this library can specify to verify
the ns-switch entry deleted or not through ns-switch conf file.
If the action is 1 , verification will be done by checking the  nsswitch.conf
file of that particular vserver.
If the user sets the action to 0(default), it will not verify by checking the  nsswitch.conf
file of that particular vserver.
(This option is deprecated)

=item C<MCC options>

The options accepted for MCC configuration replication verification is
documented at L<MCC options|lib-NACL-STask-STask-pm/mcc_options>.

=item C<< nacltask_conf_file_path => $file_path >>

(Optional) The user of this library can specify the file path to be verified for
the deleted ns-switch entry.

All of the other various options,
See L<< NACL::C::VserverServicesNameServiceNsSwitch::delete|lib-NACL-C-VserverServicesDns-pm/delete >>

=back

=back

=cut

sub purge {
    $Log->enter() if $may_enter;
   my ( $pkg_or_obj, @args ) = @_;
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            nacltask_if_purged       => { type => SCALAR, default => 'pass' },
            nacltask_verify          => { type => BOOLEAN, default => 0 },
            nacltask_conf_file_path  => { type => SCALAR, optional => 1 },
            nacltask_verify_using_state_only  => { type => BOOLEAN, default => 1 },
            nacltask_verify_in_conf_file_only => { type => BOOLEAN, default => 0 },
            $pkg_or_obj->_mcc_validate_spec(),
        },
        allow_extra => 1
    );

    # Transform %opts from the options we received into the options to
    # pass to the base class method.
    my ( %common_opts, %nacltask_opts );
    $pkg_or_obj->_copy_common_component_params_with_ci(
        source => \%opts,
        target => \%common_opts
    );

    my %mcc_opts;
    $pkg_or_obj->_move_common_mcc_opts(
        source => \%opts,
        target => \%mcc_opts,
    );

    $pkg_or_obj->_move_nacltask_options(
        source => \%opts,
        target => \%nacltask_opts
    );

    my %primary_keys;
    $pkg_or_obj->_copy_primary_keys(
        source => \%opts,
        target => \%primary_keys
    );

    my $was_created = delete $opts{_was_created};
    $$was_created = 0;
    my $if_purged                = delete $nacltask_opts{'nacltask_if_purged'};
    my $verify                   =  delete $nacltask_opts{'nacltask_verify'};

    try {
        $pkg_or_obj->delete(%opts);
    } catch NACL::C::Exceptions::VserverServicesNameServiceNsSwitch::DoesNotExist with {
        my $exception = $_[0];
        $Log->debug(
            "STask The ns-switch entry does not exists $if_purged"
        );
        if ( $if_purged !~ /pass/i ) {
            $Log->exit() if $may_exit;
            $exception->throw();
        }
    };

    $pkg_or_obj->_check_in_metrocluster_node(
        command_interface => $opts{'command_interface'},
        method_used       => "purge",
        %mcc_opts,
        %primary_keys,
    );

    if ( $nacltask_opts{'nacltask_verify_in_conf_file_only'} eq '1' ) {
        $Log->warn(" nacltask_verify_in_conf_file_only Option is deprecated."
                   . "Update your script to not to pass this option");
    }

    $pkg_or_obj->taskverify_purge(%opts)
        if ($verify || $nacltask_opts{nacltask_verify_using_state_only});

    $Log->exit() if $may_exit;
} ## end sub purge

=head2 modify

   my $ns_switch_obj = NACL::STask::VserverServicesNameServiceNsSwitch->modify(
                                'command_interface'  => $ci,
                                'vserver'            => $vserver,
                                database             => $database,
                                'sources'            => \@Sources,
                                nacltask_verify      => $boolean,
                                nacltask_verify_using_state_only  => $boolean, #default to 1
                                nacltask_verify_in_conf_file_only => $boolean, #default to 0
                                nacltask_verify_file_version      => $boolean, #default to 0
                                nacltask_verify_file_version_retry => { type => SCALAR, default => 6 },
                                nacltask_verify_file_version_interval => { type => SCALAR, default => 10 },
                                nacltask_conf_file_path  => $file_path,
                                nacltask_to_cleanup      => 1,
                                nacltask_cleanup_manager => $cleanup_obj,
                                %other_options,
                    );

(Class or Instance Method) This method is used to modify a ns-switch entry on the vserver.

Uses CMode CLI/ZAPI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component::command_interface|lib-NACL-C-Component-pm/command_interface>.

=item C<< 'vserver' => $vserver >>

(Required) Name of the vserver on which the name service switch entry
needs to be modifed.

=item C<< "database" => $database >>

(Required) This parameter used to specify the name service database,
of the given Vserver,for which to modify the name service switch entry.
Possible values for this parameter are (hosts|group|passwd|netgroup|namemap).

=item C<< "sources" => [$source1, $source2,..] >>

This parameter used to specify the name service sources and the
order in which look up for the specified vserver and name service database.
Possible values for this parameter are (files|dns|ldap|nis).

=item C<< nacltask_verify => 0|1  >>

(Optional, defaults to 0)
Verifies whether all of the attributes got set to the specified value
by using component state method and also check whether it modified in
nsswitch.conf file of that particular vserver.

Note that it is also possible to perform the verification later by calling
L<taskverify_modify|lib-NACL-STask-STask-pm/taskverify_modify>. This can
be used for cases where ONTAP chooses to not set all attributes to the exact
values specified. In the call we can specify those attributes we want
verified.

=item C<< nacltask_verify_using_state_only => $boolean >>

(Optional) The user of this library can specify to verify
the ns-switch entry modified or not through component state method or not.
If the action is 1 (default), verification will be done through component state.
If the user sets the action to 0, it will not verify through state method.

=item C<< nacltask_verify_in_conf_file_only => $boolean >>

(Optional) The user of this library can specify to verify
the ns-switch entry modified or not through ns-switch conf file.
If the action is 1, verification will be done by checking the  nsswitch.conf
file of that particular vserver.
If the user sets the action to 0(default), it will not verify by checking the  nsswitch.conf
file of that particular vserver.
(This option is deprecated)

=item C<<nacltask_verify_file_version => $boolean >>

This option verifies whether the file-version and rdb-version are in sync
within the same node and across all nodes in a cluster.
It defaults to 0 , verification will not  be done.
If set to 1, it will verify.

=item C<<nacltask_verify_file_version_retry => $scalar >>

The user of this library can verify the file-version command based on number of retries.
This option will be set to 6(default),means a user can do 6 retries to verify
file-verison and rdb-version are in sync.

=item C<<nacltask_verify_file_version_interval => $scalar >>

The user of this library can verify the file-version command based on an interval of 10seconds.
This option will be set to 10(default),means a user can execute the command with an interval of 10seconds
in 6 retries..

=item C<< nacltask_conf_file_path => $file_path >>

(Optional) The user of this library can specify the file path to be verified for
the modified ns-switch entry.

=item C<< nacltask_to_cleanup => 0|1 >>

(Optional, defaults to 0(no to cleanup))
Flag indicating if this operation is to be cleaned up or not.

=item C<< nacltask_cleanup_manager >>

Cleanup manager to use for registering.
Default : Will use the default cleanup manager.

=item other options

See L<NACL::C::Lun->modify|lib-NACL-C-Lun-pm/modify> for
all the other options accepted by this method.

=item C<MCC options>

The options accepted for MCC configuration replication verification is
documented at L<MCC options|lib-NACL-STask-STask-pm/mcc_options>.

=back

=back

=cut

sub modify {
    $Log->enter() if $may_enter;
    my ( $pkg_or_obj, @args ) = @_;
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            nacltask_verify          => { type => BOOLEAN, default => 0 },
            nacltask_conf_file_path  => { type => SCALAR, optional => 1 },
            nacltask_verify_using_state_only  => { type => BOOLEAN, default => 1 },
            nacltask_verify_in_conf_file_only => { type => BOOLEAN, default => 0 },
            nacltask_verify_file_version      => { type => BOOLEAN, default => 0 },
            nacltask_verify_file_version_retry => { type => SCALAR, default => 6 },
            nacltask_verify_file_version_interval => { type => SCALAR, default => 10 },
            $pkg_or_obj->_cleanup_validate_spec(),
            $pkg_or_obj->_mcc_validate_spec(),
        },
        allow_extra => 1
    );

    if ( $opts{enabled} ) {
       delete $opts{enabled};
       $Log->warn("\n Passing enabled option to NACL::STask::VserverServicesNameServiceNsSwitch".
                   "->modify is not valid anymore, removing this option from input. Please update your scripts");
    } 
    #Transform %opts from the options we received into the options to
    #pass to the base class method.
    my ( %common_opts, %nacltask_opts, %opts_for_cleanup,
    $nacltask_to_cleanup, $opts_for_register,$register_for_cleanup );

    $pkg_or_obj->_move_common_cleanup_opts(
        source => \%opts,
        target => \%opts_for_cleanup,
        nacltask_to_cleanup => \$register_for_cleanup
    );

    my %mcc_opts;
    $pkg_or_obj->_move_common_mcc_opts(
        source => \%opts,
        target => \%mcc_opts,
    );

    my %primary_keys;
    $pkg_or_obj->_copy_primary_keys(
        source => \%opts,
        target => \%primary_keys
    );

    $pkg_or_obj->_copy_common_component_params_with_ci(
        source => \%opts,
        target => \%common_opts
    );
    $pkg_or_obj->_move_nacltask_options(
        source => \%opts,
        target => \%nacltask_opts
    );
    my $verify                   =  $nacltask_opts{'nacltask_verify'};

    if ($register_for_cleanup) {
        $opts_for_register
            = $pkg_or_obj->_common_modify_for_cleanup(%opts_for_cleanup,
            %opts);
    }

    $pkg_or_obj->SUPER::modify(%opts);

    $pkg_or_obj->_check_in_metrocluster_node(
        command_interface => $opts{'command_interface'},
        method_used       => "modify",
        %mcc_opts,
        %primary_keys,
    );

    if ( $nacltask_opts{'nacltask_verify_in_conf_file_only'} eq '1' ) {
        $Log->warn(" nacltask_verify_in_conf_file_only Option is deprecated."
                   . "Update your script to not to pass this option");
    }
    if ( $nacltask_opts{'nacltask_verify_file_version'} eq '1' ) {
          $pkg_or_obj->_file_version_show(%opts);
    }


    # Verification based on the nacltask_opts
    $pkg_or_obj->verify_state(%opts)
        if ($verify && $nacltask_opts{nacltask_verify_using_state_only});

    $pkg_or_obj->_register_for_cleanup(%{$opts_for_register})
        if ($register_for_cleanup);
    $Log->exit() if $may_exit;
} ## end sub modify

sub _verify_config_file {
    my ($pkg,@args) = @_;
    my %opts = validate_with(
        params => \@args,
        spec   => {
            _optional_scalars(qw(_method)),
        },
        allow_extra => 1
    );

    my $apiset = $opts{command_interface}->apiset(interface => 'CLI');
    my $method = delete $opts{_method};
    my $regex = qr/(\w+):\s+(.*)/;
    my $file_path;

    if(defined $opts{nacltask_conf_file_path}) {
        $file_path = $opts{nacltask_conf_file_path};
    } else {
        my $response = $apiset->debug_smdb_table_show(
                'table'           => 'vserver_by_name',
                vserver           => $opts{'vserver'},
        );
        my $output = $response->get_parsed_output();
        if (defined ($output->[0]->{'id-unsigned'})) {
            $file_path = "/clus/.vserver_$output->[0]->{'id-unsigned'}/.vsadmin/config/etc/nsswitch.conf";
        } else {
            $Log->exit() if $may_exit;
            NATE::BaseException->throw("Not Able to get the Vserver Id for the vserver");
        }
    }
    my $systemshell_apiset = $opts{command_interface}->apiset(
        interface => 'CLI',
        set       => 'Systemshell'
    );
    my $config_output = $systemshell_apiset->execute_raw_command(
        command => "sudo cat $file_path | grep $opts{database}"
    );
    my $msg;
    if(($method ne 'purge') && ($config_output =~/$regex/)) {
        my $database = $1;
        my $sources = $2;
        $sources =~s/\s$//;
        $opts{sources} = join(" ", @{$opts{sources}});
        if ($opts{database} !~/$database/) {
            $msg  = "\n* The value of $opts{database} is not found in the conf file\n" ;
        }
        if ($opts{sources} !~/$sources/){
            $msg .= "\n* The value of $opts{sources} is not found in the conf file\n" ;
        }
        if ($msg) {
            $Log->exit() if $may_exit;
            NACL::Exceptions::VerifyFailure->throw(
                  "Verification failed! $msg" );
        }
    } else {
        if ($config_output =~/$regex/) {
            $Log->exit() if $may_exit;
            NACL::Exceptions::VerifyFailure->throw(
                  "Verification failed! Found the ns-switch entry which should have been purged" );
        }
    }
}

=head2 parents

    @parents = $ns_switch_obj->parents();

    ( Class or Instance method ) This method returns the list containing the names of
    the parent objects of this resource. Used for finding dependencies between resources.

=cut

sub parents {
    $Log->enter() if $may_enter;
    $Log->exit() if $may_exit;
    return qw(NACL::STask::Vserver);
}

1;
