#
# Copyright (c) 2014 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary WAFL scanner related Task Module
## @author kvishal@netapp.com
## @status shared
## @pod here

package NACL::STask::WaflScanner;

use strict;
use warnings;
use Time::HiRes qw( gettimeofday );

use Params::Validate qw( SCALAR BOOLEAN );
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use NATE::STAF::Peer qw($Parent);
use parent qw(NACL::C::Component::ONTAP NACL::STask::STask);
use NACL::GeneralUtils qw(random_name_generator);
use NACL::STask::Volume;

use NATE::BaseException qw(:try);
use NATE::Exceptions::Argument;
use NACL::Exceptions::Timeout;

use Readonly;

Readonly my %Scanner_regex_for => (
    snapshot_create             => qr/snap\s+create\s+summary\s+update/,
    vbn_to_inode                => qr/VBN\s+to\s+inode\s+translation/,
    file_fold                   => qr/file\s+folding/,
    container_block_reclamation => qr/container\s+block\s+reclamation/,

    ## TBD populate more scanner regexes. Use NACRE on wafl_scan_status
);

use Class::MethodMaker [
    'scalar' => [{-type => 'NATE::Process'}, 'scanner_process'],
    'scalar' => 'volume',
    'scalar' => 'vserver',
    'scalar' => 'aggregate',
    'scalar' => 'wait_for_start',
    'scalar' => 'scanner_type',
    'scalar' => 'scan_speed',
    'scalar' => 'scanner_start_time',
    'scalar' => 'start_time_var_name',
];

=head1 NAME

NACL::STask::WaflScanner

=head1 DESCRIPTION

C<NACL::STask::WaflScanner> provides functionalities around WAFL scanners.
It also has the standard regex for scanners to be used on scan status output

=over

=item Supported scanner types

    snapshot_create => qr/snap\s+create\s+summary\s+update/,
    vbn_to_inode    => qr/VBN\s+to\s+inode\s+translation/,
    file_fold       => qr/file\s+folding/,
    container_block_reclamation => qr/container\s+block\s+reclamation/,

=back

=head1 ATTRIBUTES

=head2 scanner_start_time

Time at which scanner started. Set by scan_status_stop method

=head1 METHODS

=head2 scan_status_start

    $scan_check = NACL::STask::WaflScanner->scan_status_start(
        command_interface => $command_interface,
        volume            => $volume,
        scanner_type      => 'snapshot_create',
        scan_speed        => $reduced_scan_speed,
    );

    $scan_check = NACL::STask::WaflScanner->scan_status_start(
        command_interface => $command_interface,
        aggregate         => $aggregate,
        scanner_type      => 'vbn_to_inode',
    );

(Class method) Starts a background process which tracks the wafl scanner status
    based on the scanner_type passed in. Then scan happens on the node scope
    shell. The command used is diag command 'wafl scan status'.
    wafl scan status [-S] [-A|-V] [aggr|vol]

=over

=item Options

=over

=item C<< volume => $volume >>

(Optional) This is the volume on which wafl scan status is done. Either
    aggregate or volume needs to be given not both.

=item C<< vserver => $vserver >>

(Optional for CMode, Ignored for 7Mode)
The vserver on which the volume resides.

=item C<< aggregate => $aggregate >>

(Optional) This is the Aggregate on which wafl scan status is done. Either
    aggregate or volume needs to be given not both.

=item C<< scanner_type => $scanner_type >>

(Required) Type of scanner to be tracked.

=item C<< scan_speed => $scan_speed >>

(Optional) This is the scan speed to be set after the status check has begun.

=item C<< include_subscan => 0 | 1 >>

(Optional) This will lead to wafl status output having subscan ids.

=item C<< determine_appropriate_nodescope_node => 0 | 1 >>

(Optional, default to 1)
If set to 1 (the default) determines the appropriate node on which to invoke
the Nodescope commands.

=item C<< nodescope-node-name => 'nodea' >>

(Optional )
node to be used for nodescope connection. If this is not given node is auto
determined

=back

=back

=over

=item Exceptions

=over

=item C<NATE::Exceptions::Argument>

    This type of exception is thrown for below scenarios

    vserver argument is missing while running in CMode.
    Both aggregate and volume inputs are given.
    Either aggregate or volume input is not given.
    Unsupported scanner type is passed.

=back

=back

=cut

sub scan_status_start {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;

    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            volume                => {type => SCALAR,  optional => 1},
            vserver               => {type => SCALAR,  optional => 1},
            aggregate             => {type => SCALAR,  optional => 1},
            scanner_type          => {type => SCALAR},
            scan_speed            => {type => SCALAR,  optional => 1},
            include_subscan       => {type => BOOLEAN, default  => 0},
            'nodescope-node-name' => {type => SCALAR,  optional => 1},
            determine_appropriate_nodescope_node =>
                {type => BOOLEAN, default => 1},
        },
    );

    ## Validating the scanner_type
    my $me           = (caller(0))[3];
    my $scanner_type = $opts{scanner_type};
    if (!exists($Scanner_regex_for{$scanner_type})) {
        $Log->exit() if $may_exit;
        NATE::Exceptions::Argument->throw(
            "$me: $scanner_type is not yet"
                . ' supported. Add support for this scanner in the lib.'
                . ' Supported types are: '
                . join "\n",
            (keys(%Scanner_regex_for))
        );
    }

    my $ci        = $opts{command_interface};
    my $volume    = $opts{volume};
    my $vserver   = $opts{vserver};
    my $aggregate = $opts{aggregate};

    if (defined $volume && defined $aggregate) {
        $Log->exit() if $may_exit;
        NATE::Exceptions::Argument->throw("$me: Either volume or aggregate"
                . ' needs to be passed, not both');
    } ## end if ( $ci->is_cmode() &&...)

    if (!defined $volume && !defined $aggregate) {
        $Log->exit() if $may_exit;
        NATE::Exceptions::Argument->throw(
            "$me: Either volume or aggregate" . ' needs to be passed');
    } ## end if ( $ci->is_cmode() &&...)

    ## Checking if vserver is passed along with volume
    if ($ci->is_cmode() && defined $volume && !defined $vserver) {
        $Log->exit() if $may_exit;
        NATE::Exceptions::Argument->throw("$me: The 'vserver' argument is "
                . 'required in CMode but was not provided');
    } ## end if ( $ci->is_cmode() &&...)

    my $scan_speed      = $opts{scan_speed};
    my $include_subscan = delete $opts{include_subscan};
    my $determine_node_scope
        = delete $opts{determine_appropriate_nodescope_node};
    my $nodescope_node = delete $opts{'nodescope-node-name'};
    ## Setting a unique variable so that variable names don't clash when
    ## lib is run in parallel.
    my $var_name = random_name_generator(prefix => 'start_time');
    $opts{start_time_var_name} = $var_name;
    my %common_param;
    $pkg->_move_common_component_params(
        source => \%opts,
        target => \%common_param,
    );

    my $self = $pkg->new(%opts);

    my %nodescope_override;
    if ($determine_node_scope && !$nodescope_node) {
        $nodescope_node = $self->_local_node_name($ci);
    }
    %nodescope_override = ('nodescope-node-name' => $nodescope_node);

    my $nodescope_apitest = $ci->get_7m_or_nodescope_apiset();

    my %scan_opts
        = ($volume) ? (volume => $volume) : (aggregate => $aggregate);
    $scan_opts{'subscan-id'} = $include_subscan;

    #############################################
    $Log->comment('Starting tracking scanner in background');
    #############################################
    my $proc = NATE::Process->new(
        codespec => $pkg->can('_check_wafl_scan_running'),
        runid    => "Tracking scanner: $scanner_type" . random_name_generator,
        args     => [
            $nodescope_apitest, $scanner_type,
            $var_name, (%scan_opts, %nodescope_override)
        ],
        onexit => \&NATE::Process::on_exit_die_worst_result,
    );
    $proc->start;
    $self->scanner_process($proc);

    if (defined $scan_speed) {
        $nodescope_apitest->wafl_scan_speed(
            speed => $scan_speed,
            %nodescope_override
        );
    }

    $Log->exit() if $may_exit;
    return $self;
} ## end sub scan_status_start

=head2 scan_status_stop

    $scan_check->scan_status_stop(
        timeout => 60,
        nacltask_wait_for_completion => 1,
        nacltask_wait_for_completion_timeout => 60,
    );

(Instance method) Stops the background scan status check.
    It will die if scanner has not started. It will also
    optionally wait for scanner to complete. It'll reset the wafl scan speed
    if it was changed during start

=over

=item Options

=over

=item C<< scanner_start_timeout => 60 >>

(Optional) Maximum amount of time to wait before we stop doing scan status.
    Default is 60.

=item C<< nacltask_wait_for_completion => 1|0 >>

(Optional) if 1 (default) waits for the scanner to complete.

=item C<< nacltask_wait_for_completion_timout => 60 >>

(Optional) Maximum amount of time to wait for scanner to complete once started.
    process. default is 60

=back

=back

=over

=item Exceptions

=over

=item C<NACL::Exceptions::Timeout>

This type of exception is thrown when scanner wait for completion times out.
Also thrown if scanner start does not happen

=back

=back

=cut

sub scan_status_stop {
    $Log->enter() if $may_enter;
    my ($self, @args) = @_;

    $self->_verify_invocation(style => 'instance_only');

    my %opts = $self->_common_validate_with(
        params          => \@args,
        additional_spec => {
            scanner_start_timeout        => {type => SCALAR,  default => 60},
            nacltask_wait_for_completion => {type => BOOLEAN, default => 1},
            nacltask_wait_for_completion_timout =>
                {type => SCALAR, default => 60},
        },
    );

    my $process_timeout     = $opts{scanner_start_timeout};
    my $not_run             = $opts{nacltask_if_scanner_not_run};
    my $wait_for_completion = $opts{nacltask_wait_for_completion};
    my $wait_for_completion_timeout
        = $opts{nacltask_wait_for_completion_timout};
    my $ci = $opts{command_interface};

    my $proc          = $self->scanner_process;
    my $scanner_regex = $Scanner_regex_for{$self->scanner_type};

    ## This timeout will take care of timeout for background scan status check
    eval { $proc->wait(timeout => $process_timeout); };
    if (my $error = $@) {
        if ($error =~ /Timed out/) {
            NACL::Exceptions::Timeout->throw(
                'Scanner did not start in ' . "$process_timeout seconds");
        } else {
            die "$error\n";
        }
    }

    $proc->destroy();

    my %nodescope_override
        = ('nodescope-node-name' => $self->_local_node_name($ci));

    my $nodescope_apiset = $ci->get_7m_or_nodescope_apiset();

    $nodescope_apiset->wafl_scan_speed(speed => 0, %nodescope_override)
        if ($self->scan_speed);
    if ($wait_for_completion) {
        $Log->comment('Waiting for scanner to complete');
        my $done;
        my %scan_opts
            = ($self->volume)
            ? (volume => $self->volume)
            : (aggregate => $self->aggregate);

        my $end_time = NATE::Time::timeout2time($wait_for_completion_timeout);
        while (time < $end_time) {
            my $response = $nodescope_apiset->wafl_scan_status(%scan_opts,
                %nodescope_override);
            if ($response->get_processed_output() =~ /$scanner_regex/) {
                $done = 1;
                last;
            } else {
                Tharn::snooze(1);
            }
        }
        NACL::Exceptions::Timeout->throw('Scanner did not complete in '
                . "$wait_for_completion_timeout seconds")
            if (!$done);
    }

    ## Store the scanner start timeout so that parent can access it.
    my $staf = NATE::STAF->new(global => 1);
    my $start_time = $staf->var_get(var => $self->start_time_var_name,);
    $self->scanner_start_time($start_time);
    $Log->debug("Scanner start time: $start_time");

    $Log->exit;
} ## end sub scan_status_stop

sub _check_wafl_scan_running {
    $Log->step('Scanning WAFL');
    my ($apiset, $scanner_type, $var_name, %scan_opts) = @_;

    my $staf = NATE::STAF->new(global => 1);
    $staf->var_set(
        handle   => $Parent->handle(),
        endpoint => $Parent->endpoint(),
        var      => {$var_name => 0},
    );

    my $response;
    do {
        $response
            = $apiset->wafl_scan_status(%scan_opts)->get_processed_output();
    } while ($response !~ /$Scanner_regex_for{$scanner_type}/);

    my $time = gettimeofday;
    $staf->var_set(
        handle   => $Parent->handle(),
        endpoint => $Parent->endpoint(),
        var      => {$var_name => $time},
    );
    $Log->comment("'$scanner_type' scanner is running");

} ## end sub _check_wafl_scan_running

sub _local_node_name {
    my ($self, $ci) = @_;

    if ($self->volume) {
        return NACL::STask::Volume->get_local_node_name(
            command_interface => $ci,
            vserver           => $self->vserver,
            volume            => $self->volume,
        );
    } else {
        return NACL::STask::Aggregate->get_local_node_name(
            command_interface => $ci,
            aggregate         => $self->aggregate,
        );
    }
} ## end sub _local_node_name

=head1 NAME

InodeUpgradeStatus

=head1 DESCRIPTION

C<InodeUpgrade_Status> provides functionalities for V4Inode upgrade status

=over

=back

=head1 ATTRIBUTES

=head2 scanner_start_time

Time at which v4 inode upgrade started.

=head1 METHODS

=head2 inode_upgrade_scan_status_start

(Class method) Starts a background process which tracks the v4 inode upgrade
    For now, by setting wafl flag:
        setflag wafl_larger_larger_inodes_enable 1
        setflag fl_v4ino_upgrade_disable 0
    After later milestone, the upgrade will automaticly triggered after filer upgrade is done

     my $upgrade = InodeUpgradeStatus->inode_upgrade_start(
                command_interface  => $ci,
            );

=over

=item Options

=over

=item C<< volume => $volume >>

(Optional) This is the volume on which v4 inode upgrade status is done.
    Either aggregate or volume needs to be given not both.

=item C<< vserver => $vserver >>

(Optional) The vserver on which the volume resides
    If volume is given, vserver need to be given too

=item C<< aggregate => $aggregate >>

(Optional) This is the Aggregate on which v4 inode upgrade status is done. Either
    aggregate or volume needs to be given not both.

=back

=back

=over

=back

=cut

sub inode_upgrade_scan_status_start {

    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;

    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            volume         => {type => SCALAR, optional => 1},
            vserver        => {type => SCALAR, optional => 1},
            aggregate      => {type => SCALAR, optional => 1},
            wait_for_start => {type => SCALAR, default  => 1},
        },
    );

    my $ci             = $opts{command_interface};
    my $volume         = $opts{volume};
    my $vserver        = $opts{vserver};
    my $aggregate      = $opts{aggregate};
    my $wait_for_start = $opts{wait_for_start};

    if (defined $volume && defined $aggregate) {
        $Log->exit() if $may_exit;
        NATE::Exceptions::Argument->throw(
            "Either volume or aggregate needs to be passed, not both");
    }

    if (!defined $volume && !defined $aggregate) {
        $Log->exit() if $may_exit;
        NATE::Exceptions::Argument->throw(
            "Either volume or aggregate needs to be passed");
    }

    if (defined $volume && !defined $vserver) {
        $Log->exit() if $may_exit;
        NATE::Exceptions::Argument->throw(
            "If volume is passed, vserver need to be passed too");
    }

    my $command
        = ($volume)
        ? "volume inode-upgrade show -vserver $vserver -volume $volume -fields status"
        : "aggr inode-upgrade -aggregate $aggregate -fields status";

    ## Setting a unique variable so that variable names don't clash when
    ## lib is run in parallel.
    my $var_name = random_name_generator(prefix => 'start_time');
    $opts{start_time_var_name} = $var_name;
    my %common_param;
    $pkg->_move_common_component_params(
        source => \%opts,
        target => \%common_param,
    );

    my $self = $pkg->new(%opts);

    #############################################
    $Log->comment('Starting to track upgrade in background');
    #############################################

    my $proc = NATE::Process->new(
        codespec => $pkg->can('_track_v4ino_upgrade_scanner_start'),
        runid    => "Tracking v4ino upgrade: " . random_name_generator,
        args     => [$ci->apiset(), $var_name, $command, $wait_for_start],
        onexit   => \&NATE::Process::on_exit_die_worst_result,
    );

    $proc->start;
    $self->scanner_process($proc);

    $Log->comment('Set wafl_larger_larger_inodes_enable to 1');
    my $nodescope_apitest
        = $ci->get_7m_or_nodescope_apiset(interface => 'CLI');
    $nodescope_apitest->execute_command(
        command           => "setflag wafl_larger_larger_inodes_enable 1",
        'privilege-level' => "test",
    );

    $Log->comment('Set wafl_v4ino_upgrade_disable to 0');
    $nodescope_apitest->execute_command(
        command           => "setflag wafl_v4ino_upgrade_disable 0",
        'privilege-level' => "test",
    );

    $Log->exit() if $may_exit;

    return $self;

} ## end sub inode_upgrade_scan_status_start

=head2 inode_upgrade_scan_status_stop

(Instance method) Stops the background v4 inode upgrade status check.
    It will die if v4 inode upgrade has not started withine timeout. It will also
    optionally wait for upgrade to complete.

    my $upgrade = InodeUpgradeStatus->inode_upgrade_start(
                command_interface => $ci,
                volume => $vol_name
                nacltask_wait_for_completion => 1,
                nacltask_wait_for_completion_timeout => 100,
            );

    or
    $upgrade->scan_status_stop(
        aggregate => $aggr->aggregate
        nacltask_wait_for_completion => 1,
        nacltask_wait_for_completion_timeout => 100,
    );

=over

=item Options

=over

=item Options

=over

=item C<< volume => $volume >>

(Optional) This is the volume on which v4 inode upgrade status is done. Either
    aggregate or volume needs to be given not both.

=item C<< vserver => $vserver >>

(Optional) The vserver on which the volume resides

=item C<< aggregate => $aggregate >>

(Optional) This is the Aggregate on which v4 inode upgrade status is done. Either
    aggregate or volume needs to be given not both.

=item C<< nacltask_wait_for_completion => 1|0 >>

(Optional) if 1 (default) waits for the upgrade to complete.

=item C<< nacltask_wait_for_completion_timout => 3600 >>

(Optional) Maximum amount of time to wait for upgrade to complete once started.
    process. default is 3600

=back

=back

=over

=item Exceptions

=over

=item C<NATE::Exceptions::Argument>

This type of exception is thrown for below scenarios:
        Both aggregate and volume inputs are given
        Either aggregate or volume input is not given
        vserver argument is missing while volume is given

=back

=item C<NACL::Exceptions::Timeout>

This type of exception is thrown when upgrade wait for completion times out.

=back

=back

=cut

sub inode_upgrade_scan_status_stop {

    $Log->enter() if $may_enter;
    my ($self, @args) = @_;

    $self->_verify_invocation(style => 'instance_only');
    my %opts = $self->_common_validate_with(
        params          => \@args,
        additional_spec => {
            upgrade_timeout => {type => SCALAR, default => 3600},
            nacltask_wait_for_completion => {type => BOOLEAN, default => 1},
            nacltask_wait_for_completion_timout =>
                {type => SCALAR, default => 3600},
        },
    );

    my $volume              = $self->volume;
    my $vserver             = $self->vserver;
    my $aggregate           = $self->aggregate;
    my $process_timeout     = $opts{upgrade_timeout};
    my $wait_for_completion = $opts{nacltask_wait_for_completion};
    my $wait_for_completion_timeout
        = $opts{nacltask_wait_for_completion_timout};
    my $ci = $opts{command_interface};

    my $proc = $self->scanner_process;

    ## This timeout will take care of timeout for background scan status check
    eval { $proc->wait(timeout => $process_timeout); };
    if (my $error = $@) {
        if ($error =~ /Timed out/) {
            NACL::Exceptions::Timeout->throw(
                'Upgrade did not start in ' . "$process_timeout seconds");
        } else {
            die "$error\n";
        }
    }

    $proc->destroy();

    my $command
        = ($volume)
        ? "volume inode-upgrade show -vserver $vserver -volume $volume -fields status"
        : "aggregate inode-upgrade show -aggregate $aggregate -fields status";

    if ($wait_for_completion) {
        my $response;
        my $api = $ci->apiset();

        ## Store the upgrade start timeout so that parent can access it.
        my $staf = NATE::STAF->new(global => 1);
        my $start_time = $staf->var_get(var => $self->start_time_var_name,);
        $self->scanner_start_time($start_time);

        $Log->debug("Upgrade start time: $start_time");
        my $done;
        my $end_time = NATE::Time::timeout2time($wait_for_completion_timeout);
        while (time < $end_time) {

            $Log->comment('Waiting for v4ino upgrade to finish');
            $response = $api->execute_raw_command(command => $command,);

            if ($response =~ /There are no entries matching your query/) {
                $done = 1;
                last;
            } else {
                Tharn::snooze(1);
            }
        }

        if (!$done) {
            NACL::Exceptions::Timeout->throw(
                "Upgrade did not complete in $wait_for_completion_timeout seconds"
            );
        }
    }

    $Log->exit;

} ## end sub inode_upgrade_scan_status_stop

sub _track_v4ino_upgrade_scanner_start {

    $Log->step('Start to check upgrade status');
    my ($apiset, $var_name, $command, $wait_for_start) = @_;

    my $staf = NATE::STAF->new(global => 1);

    $staf->var_set(
        handle   => $Parent->handle(),
        endpoint => $Parent->endpoint(),
        var      => {$var_name => 0},
    );

    if ($wait_for_start) {
        my $response;
        do {
            $response = $apiset->execute_raw_command(command => $command);
        } while ($response =~ /pending/);
    }
    my $time = gettimeofday;
    $staf->var_set(
        handle   => $Parent->handle(),
        endpoint => $Parent->endpoint(),
        var      => {$var_name => $time},
    );

    $Log->comment(" v4ino upgrade started");

} ## end sub _track_v4ino_upgrade_scanner_start

1;
