#
# Copyright (c) 2013 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Wreck Task Module
## @author dl-nacl-dev@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::STask::Wreck

=head1 DESCRIPTION

The wreck command is used to inject persistent corruption into the file system.
This is useful for testing wafliron, the file system consistency checkers.
This command is only available at the 'test' command level, and hence, is not
available in promoted builds.

    ## START (Running Wreck on aggregates)
    use NACL::STask::Wreck;
    my $wreck = NACL::STask::Wreck->run(
        nacltask_aggregates          => ['aggr1'],
        'nacltask_inject_commands'   => [{command=>'pseudo_bad', inum => 100, level => 0, bno => 0}],
        'method-timeout'             => 300,              
        nacltask_modify_flags        => { wafl_iron_panic_on_corruption => 0 }
    );

    #perform some operation  here like start wafliron etc.
    ...
    ...
    # Restore back the all flags
    $wreck->wreck_cleanup();

    # Get the objects of generated event logs
    $EventLog_objs = $wreck->event_logs();
    @log_objects = $EventLog_objs->{aggregate}->{aggr1};

    ## END

    or 

    ## START (Running Wreck on volumes)
    use NACL::STask::Wreck;
    my $wreck = NACL::STask::Wreck->run(
        nacltask_volumes             => ['vol1'],
        'nacltask_inject_commands'   => [{command=>'pseudo_bad', inum => 100, level => 0, bno => 0}],
        'method-timeout'             => 300,              
        nacltask_modify_flags        => { wafl_iron_panic_on_corruption => 0 }
    );

    #perform some operation  here like start wafliron etc.
    ...
    ...
    # Restore back the all flags and deletes the files created on volume
    $wreck->wreck_cleanup();

    # Get the objects of generated event logs
    $EventLog_objs = $wreck->event_logs();
    @log_objects = $EventLog_objs->{volume}->{vol1};

    # Get the inode value of volume
    my $inodes = $wreck->volume_inode();
    my $inode = $inodes->{vol1}->{inode};

    ## END


 Steps involved in run method:
 Task will throw the exception if step(1 to 3 & 5) didn't match.
 Step 1 : Aggr/Vol should be online.
 Step 2 : pseudo_bad and unverified both should not be 1.
 Step 3 : If the corruption string/corruption flag has "pseudo_bad or unverified",
          Then aggregate should have parity disks initialized.
 Step 4 : All disks should be zeroed on volume/aggregate.
 Step 5 : Creates the files on Volume.
 Step 6 : Get the inode of Volume.
 Step 7 : run wreck command
 Step 8 : Check for Finished or Wrote vbn. for successful operation.
 Step 9 : make a hash of flag initial value
 Step 10: Set/Unset flags.

=cut

package NACL::STask::Wreck;

use strict;
use warnings;

use base qw(NACL::C::Wreck NACL::STask::STask);

use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Params::Validate
    qw(validate validate_with validate_pos HASHREF SCALAR ARRAYREF OBJECT BOOLEAN );
use NATE::Exceptions::Argument qw(:try);
use NACL::APISet::Exceptions::ResponseException;
use NACL::Exceptions::UnexpectedState;
use NACL::STask::Aggregate;
use NACL::STask::Volume;
use NACL::CS::Volume;
use NACL::CS::RaidConfigInfoListDisk;
use NACL::MTask::SystemLogDetector;
use NACL::CS::EventLog;
use NACL::CS::SystemNodeDate;
use NACL::C::VolumeFile;
use Time::Local;

use NACL::ComponentUtils qw(Dumper);

use Class::MethodMaker [
    new    => [ '-hash', 'new' ],
    scalar => 'raw_output',
    scalar => 'flags_default_values',
    scalar => 'command_interface',
    scalar => 'event_logs',
    scalar => 'volume_inode',
    scalar => 'created_files_on_volume',
];

=head1 METHODS

=head2 run

    my $wreck = NACL::STask::Wreck->run(
        command_interface                => $command_interface,
        nacltask_volumes                 => [ list of the volumes name/object ],
        'nacltask_inject_commands'       => [ array of hash of commands ],
        pseudo_bad                       => 1,                               #optional
        'method-timeout'                 => N,                               #optional
        nacltask_modify_flags            => { list of flag to set }          #optional
        nacltask_file_creation_on_volume => 0                                #optional
    );

    or

    my $wreck = NACL::STask::Wreck->run(
        command_interface           => $command_interface,
        nacltask_aggregates         => [ list of the aggregates name/object ],
        'nacltask_inject_commands'  => [ array of hash of commands ],
        pseudo_bad                  => 1,                               #optional
        'method-timeout'            => N,                               #optional
        nacltask_modify_flags       => { list of flag to set }          #optional
    );


(Class Method)

Run the wreck command on list of volume/aggregate.
It is mandatory to specify atleast one of volume or aggregate parameter.


=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See NACL::C::Component::command_interface

=item C<< 'nacltask_inject_commands' => ARRAYREF >>

(Required) specifies a single or multiple injection command used to corrupt
the volume/aggregate as AOH. Each hash contains the command name and all supporting
options fot this command.
The Format of nacltask_inject_commands looks like:
    'nacltask_inject_commands'   => [
            { command => 'pseudo_bad', inum => 100, level => 0, bno => 0},
            { command => 'volinfo_snap_delete_mask', bits => 1, index => 0 }
            ];

If nacltask_inject_commands is more than one,It is written to /etc/inject-file before running wreck.
Thease 'nacltask_inject_commands' are common for all volumes/aggregate.

=item C<< 'method-timeout'=> SCALAR >>

(Optional) As component method-timeout, controls how long the command
will wait before completing.

=item C<< nacltask_aggregates => ARRAYREF >>

(Optional) List of aggregates name/object(See C<< NACL::C::Aggregate >>) to be corruped by wreck.
It is mandatory to specify atleast one of volume or aggregate parameter.

=item C<< nacltask_volumes => ARRAYREF >>

(Optional) List of volumes name/object(See C<< NACL::C::Volume >>) to be corruped by wreck.
It is mandatory to specify atleast one of volume or aggregate parameter.

=item C<< pseudo_bad => BOOLEAN >>

(Optional) See C<<NACL::C::Wreck>>

=item C<< unverified => BOOLEAN >>

(Optional) See C<<NACL::C::Wreck>>

=item C<< nacltask_modify_flags => HASHREF >>

(Optional) After executing wreck command, Users typically call wafliron to fix the corruption. 
therefore disable the 'panic on corruption mode' flag.
This method used to set/unset the flags.

=item C<< nacltask_file_creation_on_volume => BOOLEAN >>

(Optional) default value '1'
This Stask creates the temporary files on volume, before running the wreck command.
User can skip this file creation operation by setting this flag to '0'.

=back

=over

=item Exceptions

=over

=item C<NATE::Exceptions::Argument>

This type of exception is thrown if there is illegal use of switch pseudo_bad and unverified together.

=item C<NACL::Exceptions::UnexpectedState>

This type of exception is thrown when wreck fails to corrupt filesystem or
when disks belongs to aggregates/volumes are not zeroed or
when pseudo_bad or unverified raid status of aggregate is parity uninit'd.

=back

=cut

sub run {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = validate_with(
        params => \@_,
        spec   => {
            command_interface =>
                { type => OBJECT, isa => 'NACL::C::CommandInterface' },
            nacltask_inject_commands => { type => ARRAYREF },
            'method-timeout'         => { type => SCALAR, optional => 1 },
            nacltask_aggregates => {
                type      => ARRAYREF,
                optional  => 1,
                callbacks => {
                    "All of the objects provided must be of type 'NACL::C::Agregate'"
                        => sub {
                        my $objs_or_name = $_[0];
                        foreach my $aggr_obj_name ( @{$objs_or_name} ) {
                            if ( ref($aggr_obj_name) || !ref($aggr_obj_name) )
                            {
                                return 0
                                    if (
                                    ref($aggr_obj_name)
                                    && !$aggr_obj_name->isa(
                                        "NACL::C::Aggregate")
                                    );
                            } ## end if ( ref($aggr_obj_name...))
                        } ## end foreach my $aggr_obj_name (...)
                        return 1;
                        }
                }
            },
            nacltask_volumes => {
                type      => ARRAYREF,
                optional  => 1,
                callbacks => {
                    "All of the objects provided must be of type 'NACL::C::Volume'"
                        => sub {
                        my $objs_or_name = $_[0];
                        foreach my $vol_obj_name ( @{$objs_or_name} ) {
                            if ( ref($vol_obj_name) || !ref($vol_obj_name) ) {
                                return 0
                                    if ( ref($vol_obj_name)
                                    && !$vol_obj_name->isa("NACL::C::Volume")
                                    );
                            } ## end if ( ref($vol_obj_name...))
                        } ## end foreach my $vol_obj_name ( ...)
                        return 1;
                        }
                }
            },
            nacltask_modify_flags => { type => HASHREF, optional => 1 },
            nacltask_file_creation_on_volume =>
                { type => BOOLEAN, default => 1 },
            pseudo_bad => { type => BOOLEAN, optional => 1 },
            unverified => { type => BOOLEAN, optional => 1 },
            %{ NACL::C::Component->_common_validate_spec_without_ci() }
        },
    );

    $Log->debug( sub { "Args to Wreck->run()" . Dumper( \%opts ) } );

    my $aggr_obj_or_name   = delete( $opts{nacltask_aggregates} );
    my $vol_obj_or_name    = delete( $opts{nacltask_volumes} );
    my $inject_script      = delete( $opts{'nacltask_inject_commands'} );
    my $flags_to_set_unset = delete( $opts{nacltask_modify_flags} );
    my $file_creatation_flag
        = delete( $opts{nacltask_file_creation_on_volume} );
    my $command_interface = delete( $opts{command_interface} );
    my ( %comp_call_opt, $pseudo_bad, $unverified ) = ();
    if ( defined( $opts{pseudo_bad} ) ) {
        $pseudo_bad = delete( $opts{pseudo_bad} );
        $comp_call_opt{'pseudo-bad'} = $pseudo_bad;
    }
    if ( defined( $opts{unverified} ) ) {
        $unverified = delete( $opts{unverified} );
        $comp_call_opt{unverified} = $unverified;
    }

    my %common_opts;
    $pkg->_copy_common_component_params(
        source => \%opts,
        target => \%common_opts
    );

    my $aggr_list = [];
    my $vol_list  = [];

    # local variables
    my $flag_restore        = {};
    my $run_wreck           = { aggregate => [], volume => [] };
    my %inject_method       = ();
    my $wreck_raw_output    = {};
    my $wreck_event_logs    = {};
    my $wreck_volume_inode  = {};
    my $wreck_created_files = {};

    # Getting the aggregate name
    if ( defined($aggr_obj_or_name) ) {
        foreach my $aggr ( @{$aggr_obj_or_name} ) {
            if ( ref($aggr) ) {
                push @{$aggr_list}, $aggr->aggregate();
            } else {
                push @{$aggr_list}, $aggr;
            }
        } ## end foreach my $aggr ( @{$aggr_obj_or_name...})
    } ## end if ( defined($aggr_obj_or_name...))

    # Getting the volume name
    if ( defined($vol_obj_or_name) ) {
        foreach my $vol ( @{$vol_obj_or_name} ) {
            if ( ref($vol) ) {
                push @{$vol_list}, $vol->volume();
            } else {
                push @{$vol_list}, $vol;
            }
        } ## end foreach my $vol ( @{$vol_obj_or_name...})
    } ## end if ( defined($vol_obj_or_name...))

    # Illegal use of switch pseudo_bad and unverified together
    # Both should not be '1'
    if ( $pseudo_bad && $unverified ) {
        $Log->exit() if $may_exit;
        NATE::Exceptions::Argument->throw(
            "Illegal use of switch pseudo_bad and unverified together");
    }

    my $check_raid_status;

    $check_raid_status = 1
        if ( $pseudo_bad
        || $unverified
        || grep( /pseudo_bad|unverified/, @{$inject_script} ) );

    # Verifying the aggregate should be online and parity uninit'd
    if ( $#$aggr_list >= 0 ) {
        foreach my $name (@$aggr_list) {
            my $aggr = NACL::STask::Aggregate->find(
                command_interface => $command_interface,
                filter            => { aggregate => $name },
                %opts,
            );
            if ( $aggr->get_one_state_attribute( 'state', %common_opts ) ne
                "online" )
            {
                $aggr->modify( state => "online", %common_opts );
            }
            if (   $check_raid_status
                && $aggr->{raidstatus} =~ /parity uninit'd/ )
            {
                $Log->exit() if $may_exit;
                NACL::Exceptions::UnexpectedState->throw(
                    "if pseudo_bad or unverified (flags or in inject_script),"
                        . " raid status of aggergate should not be parity uninit'd"
                        . Dumper($aggr) );
            } ## end if ( $check_raid_status...)
            # Checking all disks should be zeroed on aggregate
            my @disks_obj = NACL::CS::RaidConfigInfoListDisk->fetch(
                command_interface => $command_interface,
                filter            => { tree => $name, zero => '!-1' },
                allow_empty       => 1,
            );

            if (@disks_obj) {
                $Log->exit() if $may_exit;
                NACL::Exceptions::UnexpectedState->throw(
                    "These disks of Aggregate $name is not zeroed"
                      . Dumper( \@disks_obj) );
            } ## end if  
            push @{ $run_wreck->{aggregate} }, $name;
        } ## end foreach my $name (@$aggr_list)
    } ## end if ( $#$aggr_list >= 0)

    # Verifying the volume should be online
    if ( $#$vol_list >= 0 ) {
        my $inode = {};
        foreach my $name (@$vol_list) {

            # Checking volume should be online
            my $vol = NACL::STask::Volume->find(
                command_interface => $command_interface,
                filter            => { volume => $name },
                %common_opts
            );
            my $vol_status = NACL::CS::Volume->fetch(
                command_interface => $command_interface,
                filter            => { volume => $name },
                %common_opts
            );

            if ( $vol_status->state() ne "online" ) {
                $vol->modify( state => "online", %common_opts );
            }
            my $tree;
            if ( $vol_status->volume_style() =~ /trad/ ) {
                $tree = $vol_status->volume();
            } else {
                $tree = $vol_status->aggregate();
            }

            # Checking all disks should be zeroed on aggregate
            my @disks_cs = NACL::CS::RaidConfigInfoListDisk->fetch(
                command_interface => $command_interface,
                filter            => { tree  => $tree, zero => '!-1' },
                allow_empty       => 1,
            );
            if(@disks_cs) {
                $Log->exit() if $may_exit;
                NACL::Exceptions::UnexpectedState->throw(
                    "These disks of volume $name is not zeroed"
                        . Dumper( \@disks_cs ) );
            } ## end if ( $#non_zeroed_disk...)

            $wreck_created_files->{$name}
                = _create_file_on_volume( $command_interface, $vol)
                if ($file_creatation_flag);
            $wreck_volume_inode->{$name}
                = _get_inode_of_volume( $command_interface, $name );
            push @{ $run_wreck->{volume} }, $name;
        } ## end foreach my $name (@$vol_list)
    } ## end if ( $#$vol_list >= 0 )

    # write the inject-commandss into system file
    my $apiset_obj;
    if ( $#$inject_script > 0 ) {
        $apiset_obj = $command_interface->get_7m_or_nodescope_apiset();
        try {
            my $response = $apiset_obj->rm( 'paths' => "/etc/inject-script" );
        }
        catch NACL::APISet::Exceptions::ResponseException with {
            my $exception_object = shift;
            my $resp             = $exception_object->get_response_object();
            $Log->trace("Expected: /etc/inject-script may be not present");
            $Log->trace( "Raw output was:\n" . $resp->get_raw_output() );
        };
        foreach my $ic_hash (@$inject_script) {
            my $cmd = '';
            while ( my ( $key, $value ) = each(%$ic_hash) ) {
                if ( $key eq 'command' ) {
                    $cmd = "$value " . $cmd;
                } elsif ( !defined($value) || $value eq '' ) {
                    $cmd = $cmd . "$key ";
                } else {
                    $cmd = $cmd . "$key=$value ";
                }
            } ## end while ( my ( $key, $value...))
            $cmd =~ s/^\s*|\s*$//g;
            $cmd =~ s/^|$/'/g;

            $apiset_obj->wrfile(
                'append'  => 1,
                'file'    => "/etc/inject-script",
                'content' => $cmd
            );
        } ## end foreach my $ic_hash (@$inject_script)
        $inject_method{'inject-file'} = "/etc/inject-script";
    } else {
        my $cmd = '';
        while ( my ( $key, $value ) = each( %{ $inject_script->[0] } ) ) {
            if ( $key eq 'command' ) {
                $cmd = "$value " . $cmd;
            } elsif ( !defined($value) || $value eq '' ) {
                $cmd = $cmd . "$key ";
            } else {
                $cmd = $cmd . "$key=$value ";
            }
        } ## end while ( my ( $key, $value...))
        $cmd =~ s/^\s*|\s*$//g;
        $cmd =~ s/^|$/'/g;
        $inject_method{'inject-command'} = $cmd;
    } ## end else [ if ( $#$inject_script ...)]

    $Log->trace(
        sub { "Running wreck on these vol/aggr" . Dumper($run_wreck) } );

    # Run the Wreck
    while ( my ( $type, $list ) = each(%$run_wreck) ) {
        my $wreck_op;
        foreach my $name (@$list) {
            $Log->trace("Running Wreck on $name $type");

            # Getting the start time for EventLog
            my $event_start_time = NACL::CS::SystemNodeDate->fetch(
                command_interface => $command_interface,
                requested_fields  => ['date'],
                %opts
            )->date();

            $wreck_op = $pkg->SUPER::wreck(
                command_interface => $command_interface,
                $type             => $name,
                %inject_method, %comp_call_opt,
                %common_opts
            );
            $Log->trace( sub { $wreck_op . Dumper($wreck_op) } );

            # Getting the end time for EventLog

            my $event_end_time = $pkg->_get_sysdate(
                'command_interface' => $command_interface,
                %opts
            );
            my %filter_opts;
            if ( $command_interface->mode() eq "CMode" ) {
                $filter_opts{'time'}
                    = $event_start_time . '..' . $event_end_time;
            } else {

         # Calculate the time period between event start time and current time
                my $current_time = $pkg->_get_sysdate(
                    'command_interface' => $command_interface,
                    %opts
                );
                my $time_interval = $pkg->_calculate_time_diff(
                    time_1 => $current_time,
                    time_2 => $event_start_time
                );
                $filter_opts{'time-interval'} = $time_interval;

            } ## end else [ if ( $command_interface...)]
            my $node = $command_interface->node();
            $filter_opts{'node'} = $node if defined $node;

            # Fetching the Eventlogs
            my @event_state_objs = NACL::CS::EventLog->fetch(
                'command_interface' => $command_interface,
                'filter'            => \%filter_opts,
                allow_empty         => 1,
                %common_opts
            );

            $wreck_raw_output->{$type}->{$name} = $wreck_op;
            $wreck_event_logs->{$type}->{$name} = \@event_state_objs;
            if ( $wreck_op !~ /Finished/s || $wreck_op !~ /Wrote vbn .*/s ) {
                $Log->exit() if $may_exit;
                NACL::Exceptions::UnexpectedState->throw(
                    "Wreck failed to corrupt filesystem: $wreck_op");
            }

        } ## end foreach my $name (@$list)
    } ## end while ( my ( $type, $list...))

    if ( defined($flags_to_set_unset) ) {
        if ( !$apiset_obj ) {
            $apiset_obj = $command_interface->get_7m_or_nodescope_apiset();
        }
        foreach my $flag ( keys %$flags_to_set_unset ) {
            my $api_op = $apiset_obj->printflag( flags => "$flag" );

            $flag_restore->{$flag} = $api_op->{parsed_output}->[0]->{$flag};
        }

        while ( my ( $flag, $value ) = each(%$flags_to_set_unset) ) {
            $Log->trace("Setting flags $flag to $value");
            $apiset_obj->setflag(
                'flag'  => "$flag",
                'value' => "$value"
            );

        } ## end while ( my ( $flag, $value...))
    } ## end if ( defined($flags_to_set_unset...))

    $Log->exit() if $may_exit;
    return $pkg->new(
        command_interface       => $command_interface,
        raw_output              => $wreck_raw_output,
        event_logs              => $wreck_event_logs,
        volume_inode            => $wreck_volume_inode,
        flags_default_values    => $flag_restore,
        created_files_on_volume => $wreck_created_files
    );
} ## end sub run

=head2 wreck_cleanup

    $wreck->wreck_cleanup();
     
(Instance Method)

Restore the flags to its origional values and deletes all files
created by Wreck task on volume.
This method restore the values of flags which are changed by the 
NACL::STask::Wreck::run method and deletes the files on volume which 
are created by NACL::STask::Wreck::run method.

=cut

sub wreck_cleanup {
    $Log->enter() if $may_enter;
    my $pkg = shift;

    my %opts = validate_with(
        params => \@_,
        spec =>
            { %{ NACL::C::Component->_common_validate_spec_without_ci() }, },
    );

    my $apiset_obj = $pkg->command_interface()->get_7m_or_nodescope_apiset();

    # Restoring the flags
    my $flags_to_set_unset = $pkg->{flags_default_values};

    $Log->trace(
        sub {
            "flags to restore" . Dumper($flags_to_set_unset);
        }
    );

    while ( my ( $flag, $value ) = each(%$flags_to_set_unset) ) {
        $Log->trace("Resetting flags $flag to $value");
        $apiset_obj->setflag(
            'flag'  => "$flag",
            'value' => "$value"
        );

    } ## end while ( my ( $flag, $value...))

    # Deleting the files added by Wreck Task
    my $files_to_be_deleted = $pkg->{created_files_on_volume};
    $Log->trace(
        sub {
            'Files to be deleted on volumes' . Dumper($files_to_be_deleted);
        }
    );
    while ( my ( $volume, $files ) = each(%$files_to_be_deleted) ) {
        foreach my $file (@$files) {
            $Log->trace("Deleting '$file' file on $volume volume");
            $apiset_obj->rm( "paths" => $file );
        }
    } ## end while ( my ( $volume, $files...))

    $Log->exit() if $may_exit;
} ## end sub wreck_cleanup

=head2 event_logs

    my $Eventlog_objs = $wreck->event_logs();
     
    my $generated_eventlog_objs_agr = $Eventlog_objs->{aggregate}->{'aggregate_name'}
    my $generated_eventlog_objs_vol = $Eventlog_objs->{volume}->{'volume_name'}
    
(Instance Method)

Get the all events objs NACL::CS::EventLog generated via Wreck command.
This method grep the EventLog obj which are generated by the 
NACL::STask::Wreck::run method.

Return Data Structure Format:

  $VAR1 ={
          aggregate => {
                         aggr1 => [EventLog Objects],
                         aggr2 => [EventLog Objects],
          },
          volume => {
                         vol1  => [EventLog Objects],
          }
  };

=cut

=head2 volume_inode

    my $inodes = $wreck->volume_inode();
    
    my $inode_value = $inodes->{'volume_name'}->{inode}; 

(Instance Method)

Get the inode value of volume.
This method return the inode value of nacltask_volumes.  

Return Data Structure Format:

  $VAR1 ={
          vol1 => {
                    inode => '13',
          },
          vol2 =. {
                    inode => '25',
          }
  }

=cut

# subroutine to get the inode value of the volume.

sub _get_inode_of_volume {
    $Log->enter() if $may_enter;
    my ( $command_interface, $volume ) = @_;
    my $inode      = {};
    my $apiset_obj = $command_interface->get_7m_or_nodescope_apiset();
    if ( $volume !~ /^\/vol/i ) {
        $volume = '/vol/' . $volume;
    }
    my $res_obj = $apiset_obj->showfh( 'filename' => $volume );
    $inode->{inode} = hex("$res_obj->{parsed_output}->[0]->{fileid}");
    $inode->{dsid}  = hex("$res_obj->{parsed_output}->[0]->{dsid}")
        if ( exists( $res_obj->{parsed_output}->[0]->{dsid} ) );
    $Log->exit() if $may_exit;
    return $inode;
} ## end sub _get_inode_of_volume

# subroutine to create files on volume

sub _create_file_on_volume {
    $Log->enter() if $may_enter;
    my ( $command_interface, $vol_obj ) = @_;

    # Create a file on the volume.
    my @files = ();
    for ( my $i = 1; $i <= 5; $i++ ) {
        my $file = $vol_obj->file_create(size => $i * 1024, filename => "file$i");
        push @files,$file->path();
    } ## end for ( my $i = 1; $i <= ...)
    $Log->exit() if $may_exit;
    return \@files;
} ## end sub _create_file_on_volume

# Helper method to return the sysdate of the filer

sub _get_sysdate {
    $Log->enter() if $may_enter;
    my $self = shift;
    my %opts = validate_with(
        params => \@_,
        spec => { 'command_interface' => { type => OBJECT, optional => 1 } },
        allow_extra => 1
    );
    my $ci = $opts{command_interface} || $self->command_interface();
    $Log->exit() if $may_exit;
    return NACL::CS::SystemNodeDate->fetch(
        'command_interface' => $ci,
        'requested_fields'  => ['date'],
    )->date();
} ## end sub _get_sysdate

sub _calculate_time_diff {
    my $self = shift;
    my %opts = validate_with(
        params => \@_,
        spec   => {
            'time_1' => {
                regex =>
                    qr/^\d{1,2}\/\d{1,2}\/\d{4}\s\d{1,2}:\d{1,2}:\d{1,2}$/
            },
            'time_2' => {
                regex =>
                    qr/^\d{1,2}\/\d{1,2}\/\d{4}\s\d{1,2}:\d{1,2}:\d{1,2}$/
            },
        },
    );
    my @time1 = split( /\/|\s|:/, $opts{time_1} );
    my @time2 = split( /\/|\s|:/, $opts{time_2} );
    $time1[0] = $time1[0] - 1;
    $time2[0] = $time2[0] - 1;

    my $epoch_time_1
        = timelocal( ( map { s/^0*//; $_ } ( @time1[ 5, 4, 3, 1, 0, 2 ] ) ) );
    my $epoch_time_2
        = timelocal( ( map { s/^0*//; $_ } ( @time2[ 5, 4, 3, 1, 0, 2 ] ) ) );
    return $epoch_time_1 - $epoch_time_2;
} ## end sub _calculate_time_diff

1;

