#
# Copyright (c) 2014-2015 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Mix-in for the CifsConfigReplayUtils STasks
## @author ng-cifs-qa-nb@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::STask::_Mixins::CifsConfigReplayUtils

=head1 DESCRIPTION

This mix-in contains the implementation for the CIFS Replay check
for VserverCifsShare and VserverCifsShareProperties STasks.

=cut

package NACL::STask::_Mixins::CifsConfigReplayUtils;
use strict;
use warnings;

BEGIN {
    use Exporter qw(import);
    our @EXPORT_OK = qw(
        wait_for_cifs_share_replay
    );
}

use Params::Validate qw(:all);
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use NATE::BaseException qw(:try);
use NACL::ComponentUtils qw(_hash_move _hash_copy);
use Data::Dumper;
use NACL::C::SystemNode;
use NACL::C::Node;

# function : wait_for_cifs_share_replay
# description : This method is to check the replay is completed for the configuration made for
# cifs share and properties create/delete/add/remove, the replay completeness is checked on
# all the nodes in the cluster for the configuration replication for a Total of nacltask_replay_timeout seconds
# at a time interval of nacltask_replay_interval seconds. the default value of nacltask_replay_timeout is 30 seconds
# and nacltask_replay_interval is 5 seconds.

sub wait_for_cifs_share_replay {
    $Log->enter() if $may_enter;

    my ($obj, @args) = @_;

    my %validate_spec = (
        %{NACL::C::Component->_common_validate_spec()},
        nacltask_replay_timeout  => {type => SCALAR, default => 30},
        nacltask_replay_interval => {type => SCALAR, default => 5},
        method_action            => {type => SCALAR}
    );

    my %opts = validate_with(
        params      => \@args,
        spec        => \%validate_spec,
        allow_extra => 1
    );

    my $nacltask_options = {};
    $obj->_move_nacltask_options(
        source => \%opts,
        target => $nacltask_options
    );

    my $method_action = $opts{'method_action'};

    # get rid of extra options (if any)

    my %new_opts;

    $obj->_hash_move(
        source => \%opts,
        target => \%new_opts,
        move   => [keys %validate_spec]
    );

    # now %opts should not have any of the options specific to the component
    # involved (the one whose object / package name is used to invoke this
    # method)

    my $command_interface = $new_opts{command_interface};
    my $mode              = $command_interface->mode();

    # fetch the values for timeout and interval
    my $timeout  = $nacltask_options->{nacltask_replay_timeout};
    my $interval = $nacltask_options->{nacltask_replay_interval};

    #If the Command interface is 7 Mode, the replay check is not needed
    if ($mode ne 'CMode' || $timeout <= 0) {
        $Log->exit() if $may_exit;
        return;
    }

# The Configuration replay needs to be verified on all the nodes and the command diag nblade cifs shares
# takes node as a parameter and hence we need all the node names in the cluster.
    my @nodes = ();

# The system node and diag nblade commands will not be available in vserver context hence we need to fetch it from the
# NACL::C::Node
    if ($command_interface->isa('NACL::C::Vserver')) {
        @nodes             = NACL::C::Node->find();
        $command_interface = $nodes[0];
    } else {
        @nodes = NACL::C::SystemNode->find(
            command_interface => $command_interface);
    }

    my @node_names = map { $_->node() } @nodes;

    my $elapsed_time = 0;
    my ($start_time, $end_time);

# In order to avoid more iterations we will start with wait for a minimum of nacltask_replay_interval seconds
    Tharn::snooze($interval);
    my $replay_status = 0;
    $elapsed_time = 5;
    foreach my $node_name (@node_names) {
        $replay_status = 0;
        ## The total time taken for all configuration replay on all the nodes should not exceed nacltask_replay_timeout value
        while ($elapsed_time <= $timeout) {
            $start_time = time();

#_check_nblade_cifs_share_config returns 1 if the replay is succeeded in the node.
            $replay_status = _check_nblade_cifs_share_config(
                command_interface => $command_interface,
                node              => $node_name,
                method_action     => $method_action,
                %opts
            );
            if ($replay_status) {
                last;
            }
            Tharn::snooze($interval);
            $end_time = time();
            $elapsed_time += ($end_time - $start_time);
        }
    }

    $Log->exit() if $may_exit;
    NACL::Exceptions::ReplayNotComplete->throw(
        "Replay did not complete within $timeout seconds on all the nodes in the cluster"
    ) if (!$replay_status);

    $Log->exit() if $may_exit;
} ## end sub wait_for_cifs_share_replay

# This method checks the configuration replayed to nblade on the node passed as input parameter
# and returns 1 if the replay is success else returns 0
sub _check_nblade_cifs_share_config {
    my %opts              = @_;
    my $command_interface = delete $opts{command_interface};
    my $node              = delete $opts{node};
    my $method_action     = delete $opts{method_action};
    my $vserver           = delete $opts{vserver};
    my $share             = delete $opts{'share-name'};
    my $apiset            = $command_interface->apiset();
    my $result            = 0;
    my @share_properties   = ();
    my %acl_to_verify      = ();
    my $nblade_cifs_config = undef;

    if (exists($opts{'share-properties'})) {
        @share_properties = @{delete $opts{'share-properties'}};
    }

    my %input_parameter = (vserver => $vserver, node => $node);

    if ( $command_interface->version_manager()->is_version_ge(release =>'9.0') ) {
        if ($share !~ /\*/) {
            $input_parameter{'share'} = $share;
        }
    }

    my $is_node_down = 0;

    try {
        $nblade_cifs_config
            = $apiset->diag_nblade_cifs_shares(%input_parameter);
    }
    catch NACL::APISet::Exceptions::CommandFailedException with {

#Handling the node down case. For replay success, replay should complete in
#all nodes in the cluster for test to continue. But in node down case in SFO or
#non-SFO,the cifs share configurations will be success is other nodes and replay
#cannot be checked on the node which is down. In this case, replay success is
#returned since test should not fail with replay incomplete error.

        my $cluster = NACL::C::Cluster->find(
            command_interface => $command_interface,
            filter            => {node => $node}
        );
        my $node_health = $cluster->get_one_state_attribute('health');
        if ($node_health eq 'false') {
            $is_node_down = 1;
        } else {

#retry is needed as there can be a race condition where the node was coming up
#and the diag command returned RPC time out and health becoming true immediately
#to avoid this we retry once again when the health is not false.
            $nblade_cifs_config
                = $apiset->diag_nblade_cifs_shares(%input_parameter);
        }
    };

    return 1 if ($is_node_down);

    my %properties_to_verify = ();
    my %method_action_to_function_map = (
        purge => sub {
            if ($share =~ /\*/) {
                my @cifs_shares = @{$nblade_cifs_config->get_parsed_output()};
                $result = 1;
                foreach my $nblade_shares (@cifs_shares) {
                    if ($nblade_shares->{'share'} =~ /^$share/i) {
                        $result = 0;
                        last;
                    }
                }

            } else {
                my $processed_output = $nblade_cifs_config->get_processed_output();
                if ($processed_output =~ /No shares exist/i) {
                    $result = 1;
                }
            }
        },
        remove => sub {
            %properties_to_verify
                = _hash_populate(keys => \@share_properties, value => 'off');
        },
        add => sub {
            %properties_to_verify
                = _hash_populate(keys => \@share_properties, value => 'on');
        },
        create => sub {
            %properties_to_verify
                = _hash_populate(keys => \@share_properties, value => 'on');

# if its a share create apart from the input parameter passed to the cifs share create
# we need to check for the acecount as well, because by default when a share is created
# an ACL will be created for the share.
            $acl_to_verify{acecount} = '1';
        }
    );

    if (exists $method_action_to_function_map{$method_action}) {
		$method_action_to_function_map{$method_action}->();
    } else {
        NATE::BaseException->throw(
                  "No action defined for "
                . $method_action
                . " If you think this needs to be supported,
		create a BURT with type=NACL, subtype=nacl_core"
        );

    }

    if ($method_action eq 'purge') {

        # This is to handle the purge case.
        # The sub routine ref sets the result
        # value.
        return $result;
    }

    my $share_attributes = _map_share_attributes_for_replay_check(%opts);
    my $share_properties
        = _map_share_properties_for_replay_check(%properties_to_verify);
    my %share_config_search_for_replay
        = (%{$share_properties}, %{$share_attributes}, %acl_to_verify);

    my @replay_search_results
        = $nblade_cifs_config->search(%share_config_search_for_replay);

    if (@replay_search_results == 1) {
        return 1;
    } else {
        $Log->comment(
            "Unable to find a Replay configurations in the cifs data store. Dumping the parsed output\n"
                . Dumper $nblade_cifs_config->get_parsed_output());
        $Log->comment(
            "Dumping the share configurations searched in the cifs data store for replay check\n"
                . Dumper(\%share_config_search_for_replay));
        return 0;
    }

} ## end sub _check_nblade_cifs_share_config

#populate the key value pair for the share properties passed in the Configuration
sub _hash_populate {
    my %args  = @_;
    my @keys  = @{$args{keys}};
    my $value = $args{value};
    my %hash;

# This is to handle for situation when the share-properties are passed as ["oplocks,browsable,changenotify"]
    my $key = join(',', @keys);
    @keys = split(',', $key);

    foreach my $key (@keys) {
        $key =~ s/^\s+|\s+$//g;
        $key =~ s/-/_/g;
        $hash{$key} = $value;
    }
    return %hash;
} ## end sub _hash_populate

# This methods is to fill the share properties in proper key value pair as the there is
# difference in the names stored in the diag_nblade_cifs_shares and the share configuration commands
sub _map_share_properties_for_replay_check {
    my (%opts) = @_;

# This mapping is required since the name mismatch in the diag_nblade_cifs_shares
# and the cifs share create/modify and cifs share properties add/remove commands
    my %mapped_properties = (
        'continuously_available'   => "continuously_avail",
        'namespace_caching'        => "namespacecaching",
        'show_previous_versions'   => "showpreviousversions",
        'encrypt_data'             => "encryptdata",
        'showsnapshot'             => "show_snapshot",
        'attributecache'           => "attribute_cache",
        'access_based_enumeration' => "accessbasedenumeration",
        'directory_leasing'        => "directoryleasing",
        'browsable'                => 'browsable',
        'homedirectory'            => 'homedirectory',
        'branchcache'              => 'branchcache',
        'changenotify'             => 'changenotify',
        'oplocks'                  => 'oplocks',
    );

    my $share_properties_for_replay_check = {};

    foreach my $property (keys(%opts)) {
        if (exists($mapped_properties{$property})) {
            $share_properties_for_replay_check->{$mapped_properties{$property}
            } = $opts{$property};
        } else {
            $share_properties_for_replay_check->{$property}
                = $opts{$property};
        }
    }
    return $share_properties_for_replay_check;
} ## end sub _map_share_properties_for_replay_check

# This methods is to fill the share attributes in proper key value pair as the there is
# difference in the names stored in the diag_nblade_cifs_shares and the share configuration commands
sub _map_share_attributes_for_replay_check {
    my %opts = @_;

# This mapping is required since the name mismatch in the diag_nblade_cifs_shares
# and the cifs share create/modify and cifs share properties add/remove commands
    my %mapped_attributes = (
        'file-umask'                => 'file_umaskoctal',
        'dir-umask'                 => 'dir_umaskoctal',
        'symlink-properties'        => 'symlinks',
        'path'                      => 'path',
        'attribute-cache-ttl'       => 'attributecachettl',
        'offline-files'             => 'offlinefiles',
        'vscan-fileop-profile'      => 'vscanfileopprofile',
        'max-connections-per-share' => 'maxconnectionspershare',
        'force-group-for-create'    => 'forcegroupenabled',

    );
    my $share_attribute_for_replay_check = {};

# The symlink values in CLI and in the output of the diag_nblade_cifs_shares is
# is different
    my %symlink_values = (
        'symlinks'               => 'symlinks',
        'enable'                 => 'enable',
        'hide'                   => 'hide',
        'disable'                => 'disable',
        'symlinks_and_widelinks' => 'symlinks_and_widelinks',
        'read_only'              => 'read_only',
        'read-only'              => 'read_only',
        'symlinks-and-widelinks' => 'symlinks_and_widelinks',
        'no_strict_security'     => 'no_strict_security',
        'no-strict-security'     => 'no_strict_security',

    );

    foreach my $attributes (keys(%opts)) {
        if ($attributes eq 'symlink-properties') {
            my @values = @{$opts{$attributes}};
            for (my $i = 0; $i < scalar(@values); $i++) {
                $share_attribute_for_replay_check
                    ->{$mapped_attributes{$attributes}}->[$i]
                    = $symlink_values{$values[$i]};
            }
        } elsif ($attributes eq 'force-group-for-create') {
            my $value = "Disabled";
            $value = "Enabled" if ($opts{$attributes} =~ /\w+|\d+/);
            $share_attribute_for_replay_check
                ->{$mapped_attributes{$attributes}} = $value;
        } elsif (exists($mapped_attributes{$attributes})) {
            $share_attribute_for_replay_check
                ->{$mapped_attributes{$attributes}} = $opts{$attributes};
        } else {
            $share_attribute_for_replay_check->{$attributes}
                = $opts{$attributes};
        }
    }

    return $share_attribute_for_replay_check;
} ## end sub _map_share_attributes_for_replay_check

1;
