#
# Copyright (c) 2013 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Mix-in for the VserverFpolicyPolicy STasks
## @author sshaik@netapp.com, dl-nacl-dev@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::STask::_Mixins::VserverFpolicyPolicy

=head1 DESCRIPTION

This mix-in contains the implementation for create and purge methods.

=head1 CLEANUP METHODS

Cleanup can be registered for the following methods, '_create_helper'.

Cleanup methods are,
     VserverFpolicyPolicy Method     Cleanup Method
    -----------------------------------------------
     create                           purge

= head2 create

=item C<< nacltask_to_cleanup => 0|1 >>

(Optional, default to 0(no to cleanup)
Flag indicating if this operation needs to be registered for clean up or not.

=item C<< nacltask_cleanup_manager >>

Cleanup manager to be used for registering.
Default : Will use the default cleanup manager.

=cut

package NACL::STask::_Mixins::VserverFpolicyPolicy;

use strict;
use warnings;

BEGIN {
    use Exporter qw(import);
    our @EXPORT_OK = qw( create purge );
    our %EXPORT_TAGS = ( 'all' => \@EXPORT_OK );
}

use strict;
use warnings;
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use Params::Validate qw(validate SCALAR BOOLEAN SCALARREF);
use NATE::BaseException qw(:try);

sub create {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = $pkg->_common_validate_with(
        params          => \@_,
        additional_spec => {
            nacltask_if_exists => $pkg->_if_exists_validate_spec(),
            _was_created       => { type => SCALARREF, optional => 1 },
            $pkg->_cleanup_validate_spec(),
        },
        allow_extra => 1
    );

    # Transform %opts from the options we received into the options to
    # pass to the base class method.
    my ( $self, %common_opts, %nacltask_opts, %opts_for_cleanup, %opts_for_register, $nacltask_to_cleanup  );
    $pkg->_copy_common_component_params_with_ci(
        source => \%opts,
        target => \%common_opts
    );
    $pkg->_move_common_cleanup_opts(
        source  => \%opts,
        target  => \%opts_for_cleanup,
    );
    $pkg->_move_nacltask_options(
        source => \%opts,
        target => \%nacltask_opts
    );
    my $was_created = delete $opts{_was_created};
    my $if_exists   = $nacltask_opts{'nacltask_if_exists'};
    my $verify      = $nacltask_opts{'nacltask_verify'};
    $$was_created = 0;

CREATE: {
        use warnings;
        try {
            $self         = $pkg->C_SUPER::create(%opts);
            $$was_created = 1;
        }
        catch NACL::C::Exceptions::AlreadyExists with {
            my $exception = $_[0];
            $Log->debug(
                "STask The Policy already exists $nacltask_opts{'nacltask_if_exists'}"
            );
            $self = $pkg->_element_exists_handler(
                create_opts        => \%opts,
                nacltask_if_exists => $if_exists,
                exception          => $exception
            );
            if ( !$self ) {
                no warnings qw(exiting);
                redo CREATE;
            }
        };
    }
    if ($verify) {
        $pkg->verify_state(%opts);
    }
    # Register this resource with the cleanup manager for cleanup.
    $pkg->_copy_common_opts_for_cleanup(
        'source'             => {%opts, %opts_for_cleanup},
        'target'             => \%opts_for_register,
        'nacltask_to_cleanup'=> \$nacltask_to_cleanup,
        'to_cleanup'         => 'purge',
    );

    $pkg->_register_for_cleanup(%opts_for_register)
         if ($nacltask_to_cleanup && $$was_created);

    $Log->exit() if $may_exit;
    return $self;

} ## end sub create
sub purge {
    $Log->enter() if $may_enter;
    my $pkg_or_obj      = shift;
    my %orig_opts       = @_;
    my $additional_spec = {
        nacltask_verify    => { type => BOOLEAN, default => 0 },
        nacltask_if_purged => { type => BOOLEAN, default => 0 }
    };

    # Throw away options specific to the task
    $pkg_or_obj->_hash_move(
        source => \%orig_opts,
        target => {},
        move   => [ keys %{$additional_spec} ]
    );
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@_,
        additional_spec => $additional_spec,
    );

    my %common_opts;
    $pkg_or_obj->_copy_common_component_params_with_ci(
        source => \%opts,
        target => \%common_opts
    );
    my $if_purged = $opts{nacltask_if_purged};

AGAIN: {
        use warnings;
        try {
            $pkg_or_obj->delete( %common_opts, %orig_opts );
        }
        catch NACL::C::Exceptions::DoesNotExist with {
            my $e = $_[0];
            # Check if the command failed because the policy scope
            # does not exist
            if ( $if_purged !~ /pass/i ) {
                $Log->exit() if $may_exit;
                $e->throw();
            }
        };
    }
    if ( $opts{nacltask_verify} ) {
        $pkg_or_obj->_generic_purge_verify(%orig_opts);
    }
    $Log->exit() if $may_exit;
} ## end sub purge

1;
