#
# Copyright (c) 2017 NetApp, Inc., All Rights Reserved.
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

package NACL::Service::ASUPTrigger::Logger;

use v5.14;
use strict;
use warnings FATAL => 'all';
use Carp qw/confess/;
use Data::Dumper::Concise qw/Dumper/;
use MooseX::Params::Validate qw/validated_list validated_hash/;
use Data::UUID;
use List::MoreUtils qw/uniq/;
use Sys::Hostname qw/hostname/;

use Tharn_UsageService qw/log_to_syslog/;
use TestUtils qw(get_site_info);
use Smoke qw//;
use NATE::Log qw/log_global/;
use Hostrec;


use NATE::ParamSet qw//;
use NATE::Database::Test qw/testdb_global/;
use SystemTest::ASUPMessage;

use Moose::Role;

has syslog_server => (
    is          => 'rw',
    isa         => 'Maybe[Str]',
    lazy        => 1,
    builder     => '_populate_syslog_info',
    predicate   => 'has_syslog_server',
);

has syslog_port => (
    is          => 'rw',
    isa         => 'Maybe[Int]',
    predicate   => 'has_syslog_port',
);

has enable_syslog => (
    is      => 'rw',
    isa     => 'Bool',
    default => 1,
);

has top_logdir => (
    is      => 'rw',
    isa     => 'Str',
    lazy    => 1,
    builder => '_populate_top_logdir',
);

has nate_uuid => (
    is      => 'rw',
    isa     => 'Str',
    lazy    => 1, # Important
    builder => '_populate_nate_uuid',
);

has user => (
    is      => 'rw',
    isa     => 'Str',
    lazy    => 1, # Important
    builder => '_populate_user',
);

has user_is_not_qre => (
    is      => 'rw',
    isa     => 'Bool',
);

# Errors that happen in timers need to be buffered and communicated at the
# next API call
has errors => (
    is          => 'rw',
    isa         => 'ArrayRef[Str]',
    traits      => ['Array'],
    clearer     => 'clear_errors',
    handles => {
        add_error => 'push',
        has_errors=> 'count',
    }
);

has site => (
    is      => 'rw',
    isa     => 'Str',
    lazy    => 1,
    builder => '_populate_site',
    clearer => 'clear_site',
    predicate => 'has_site',
);

has testbed => (
    is      => 'rw',
    isa     => 'Maybe[Str]',
    lazy    => 1,
    builder => '_populate_testbed',
);

has test_plan => (
    is      => 'rw',
    isa     => 'Maybe[Str]',
    lazy    => 1,
    builder => '_populate_testplan',
);

has service_worker_file => (
    is      => 'rw',
    isa     => 'Str',
    lazy    => 1,
    builder => '_populate_service_worker_file',
);

has buildroot => (
    is      => 'rw',
    isa     => 'Str',
    lazy    => 1,
    builder => '_populate_buildroot',
);

has force_error_logging => (
    is      => 'rw',
    isa     => 'Bool',
    default => 0,
);

has log_nate_errors => (
    is      => 'rw',
    isa     => 'Bool',
    default => 1,
);

has syslog_buffer => (
    is          => 'rw',
    isa         => 'ArrayRef',
    traits      => ['Array'],
    handles => {
        _add_to_syslog_buffer   => 'push',
        get_next_syslog         => 'shift',
        return_syslog_to_buffer => 'unshift',
    },
    default     => sub{[]},
);

has mock_graylog => (
    is      => 'rw',
    isa     => 'Bool',
    default => 0,
);

has mock_graylog_port => (
    is      => 'rw',
    isa     => 'Int',
);

my $Log = log_global();

sub get_syslog_info {
    my ($self) = @_;
    return {
        server => $self->syslog_server,
        port   => $self->syslog_port,
    };
}

sub report_error_to_syslog {
    my ($self, $err, %additional_payload) = @_;
    $self->_report_to_syslog(
        message => $err,
        level   => 'err',
        %additional_payload,
    );
}

sub report_comment_to_syslog {
    my ($self, $message, %additional_payload) = @_;
    $Log->comment($message);
    $self->_report_to_syslog(
        message => $message,
        level   =>'info',
        %additional_payload
    );
}

sub add_to_syslog_buffer {
    my ($self, %syslog) = validated_hash(\@_,
        payload   => { isa => 'HashRef', optional => 0 },
        log_level => { isa => 'Str',     optional => 0 },
    );
    $syslog{creation_time} = time();
    $self->_add_to_syslog_buffer(\%syslog);
    return;
}

sub empty_syslog_buffer {
    my $self = shift;

    # retry_count is per syslog. max_time is total.
    my ($retry_count, $max_time) = validated_list(
        \@_,
        retry_count => { isa => 'Int', optional => 1, default => 0 },
        max_time    => { isa => 'Int', optional => 1, default => 60},
    );
    my $end_time   = time() + $max_time;

    while ( (my $syslog = $self->get_next_syslog) && time < $end_time ){
        $syslog->{log_name} = 'SUAMP';
        $syslog->{server}   = $self->syslog_server;
        $syslog->{port}     = $self->syslog_port;

        if ($self->mock_graylog){
            # The mock syslogd seems to only work with udp
            $syslog->{protocol} = 'udp';
        }

        # Using Dumper turns numbers into strings and graylog cant create the right
        # graphs. Make a copy of the args.
        my %c = %$syslog;
        $Log->comment("Sending to syslog:\n" . Dumper(\%c));

        # Try sending the syslogs FIFO. If it fails, put it back in the buffer
        # and stop all processing
        my $success;
        my $err;
        my $min_sleep_time = 1;
        my $syslog_retry_count = $retry_count;
        RETRY: while ( time() < $end_time ){
            $success = eval { log_to_syslog(%$syslog); 1 };
            $err = $@;
            last RETRY if ($success);
            last RETRY if ($syslog_retry_count <= 0);
            my $sleep_time = $min_sleep_time + ($retry_count - $syslog_retry_count)**2;
            $Log->comment("Unable to send message to syslog. Sleeping $sleep_time seconds before next attempt\nError:\n$err");
            sleep $sleep_time;
            $syslog_retry_count --;
        }
        if (!$success){
            if ($self->log_nate_errors){
                $Log->error($err);
            }
            else{
                $Log->comment($err);
            }
            # Convert exceptions to string. Exceptions typically do the right thing
            # when in a string
            $self->add_error("$err");
            $self->return_syslog_to_buffer($syslog);

            # In case we are able later to send syslog, add a message to the buffer
            my $error_syslog = $self->_build_syslog(
                message => 'Unable to send syslog message',
                level   => 'err',
            );
            $self->add_to_syslog_buffer( %$error_syslog );
            last;
        }
    }
    return;
}

sub _build_syslog {
    my ($self, %args) = @_;
    my $str     = delete $args{message};
    my $level   = delete $args{level};

    confess("Missing arguments:\n" . Dumper(\%args))
        if ( !defined($str) || !defined($level));

    my %syslog = (
        payload   => {
            message             => $str,
            site                => $self->site // '',
            testbed             => $self->testbed // '',
            test_plan           => $self->test_plan // '',
            top_logdir          => $self->top_logdir,
            nate_uuid           => $self->nate_uuid,
            user                => $self->user,
            service_worker_file => $self->service_worker_file,
            buildroot           => $self->buildroot,
            asup_type           => $self->asup_type,
            skip_vsims          => $self->skip_vsims,
            user_is_not_qre     => $self->user_is_not_qre // '',
            %args,
        },
        log_level => $level,
    );
    return \%syslog;
}

sub _report_to_syslog {
    my $self = shift;
    return if (!$self->enable_syslog);
    my $syslog = $self->_build_syslog(@_);
    $self->add_to_syslog_buffer(%$syslog);
    $self->empty_syslog_buffer;
    return;
}

sub get_errors {
    my ($self) = @_;
    my $response = {
        errors              => $self->errors,
        force_error_logging => $self->force_error_logging,
    };
    $self->clear_errors;
    return $response;
}

# Send a syslog to Graylog
# This will allow us to detect errors happening in test runs without
# having to rely on test teams filing burts to tell us
sub log_error {
    my ($self, $err) = @_;
    if ($self->log_nate_errors){
        $Log->error($err);
    }
    else{
        $Log->comment($err);
    }
    # Convert exceptions to string. Exceptions typically do the right thing
    # when in a string
    $self->add_error("$err");
    $self->report_error_to_syslog("$err");
    return;
}

sub disable_syslog {
    my ($self) = @_;
    $self->enable_syslog(0);
    $self->force_error_logging(1);
    return;
}

# We need the site to pick the right syslog server
sub _populate_site {
    my ($self) = @_;
    my $site = '';
        # If we are running in the smoke bubble, figure out which one we are in
    if (Smoke::runningInBubble()) {
        my $bubble = Smoke::whichBubble();
        if ($bubble =~ /rtp/i) {
            $site = "rtp";
        } elsif ($bubble =~ /svl/i) {
            $site = "svl";
        } elsif ($bubble =~ /nbs/i) {
            $site = "btc";
        } elsif ($bubble =~ /ptc/i) {
            $site = "ptc";
        } else {
            $self->log_error("Could not determine which smoke bubble we are running out of, skipping syslog reporting.");
            return '';
        }
        $Log->info("We're running in a $site bubble");
    }
    else{
        $Log->info("Not running in a bubble");
    }

    if (!$site){
        my $params = NATE::ParamSet->new( global => 1);
        $site = $params->get('SITE');
        if (!$site){
            $Log->info("No SITE found in parameters");
            if (!$self->no_nodes_detected){
                $Log->info("Looking at nodes to find site information");
                my $nodes = $self->nodes;
                if ($nodes && keys %$nodes){
                    # Use the site of the first node we see with a SITE
                    my %sites;
                    for my $node_name ( keys %$nodes ){
                        my $hostrec = eval { Hostrec->new(id => $node_name) };
                        if ($hostrec){
                            my $SITE = $hostrec->{SITE};
                            if ($SITE){
                                $sites{$SITE} = 1;
                            }
                        }
                    }
                    my @sites = keys %sites;
                    if (@sites){
                        $Log->info("Sites detected in nodes:\n" .Dumper(\@sites));
                        $site = $sites[0];
                    }
                }
            }

            if (!$site){
                my $hostname = hostname();
                if ($hostname !~ /\./){
                    my @lines = `host $hostname`;
                    for my $line (@lines){
                        my @fields = split(/\s/, $line);
                        if ($fields[0] =~ /^$hostname\./){
                            $hostname = $fields[0];
                        }
                    }
                }
                if ($hostname =~ /^sti/){
                    $site = 'sti';
                }
                elsif ($hostname =~ /\.rtp\./){
                    $site = 'rtp';
                }
                elsif ($hostname =~ /\.gdl\./){
                    $site = 'rtp';
                }
                elsif ($hostname =~ /\.nane\./){
                    $site = 'nane';
                }
                elsif ($hostname =~ /\.spin\./){
                    $site = 'ptc';
                }
                elsif ($hostname =~ /\.(btc|nbs)\./){
                    $site = 'btc';
                }
                elsif( $hostname =~ /\.eng\.netapp\.com/){
                    $site = 'svl';
                }
                if ($site){
                    $Log->comment("Site found after looking at local hostname '$hostname': $site");
                }
                else {
                    $Log->comment("No site found after looking at local hostname '$hostname'");
                }
            }
        }
        else{
            $Log->info("SITE found as a parameter");
        }
    }
    $site //= '';
    $Log->info("Site found: $site");
    if (!$site){
        $self->log_error("Unable to figure out what site we're running from");
    }
    # All sites should be lowercase to have any hope of being found in site_info.dat
    return lc($site);
}

# The testbed is used to identify pushbutton runs
sub _populate_testbed {
    my $self = shift;
    my $params = NATE::ParamSet->new( global => 1);
    return $params->get('TESTBED');
}

# The testplan is used to identify ctl5 runs
sub _populate_testplan {
    my $self = shift;
    my $params = NATE::ParamSet->new( global => 1);
    return $params->get('test_plan');
}

# The testplan is used to identify ctl5 runs
sub _populate_buildroot {
    my $self = shift;
    my $params = NATE::ParamSet->new( global => 1);
    my $b = $params->get('BUILDROOT') // '';
    return $b;
}

sub _populate_service_worker_file {
    my $self = shift;
    my $file = $INC{'NACL/Service/ASUPTrigger/Worker.pm'} // '';
    return $file;
}

sub _populate_syslog_info {
    my $self = shift;
    my $site = $self->site;

    if (!$site){
        # Nothing we can do
        $self->log_error("No site found, unable to enable syslog. Contact jroussin\@netapp.com when you see this error message and provide a path to these logs");
        $self->disable_syslog;
        return;
    }

    # get_site_info only contains lowercase site names. But uppercase is prevalent
    # in some parts of NetApp
    $site = lc($site);

    # if we found a valid site,
    # parse out the syslog server data for this site
    # force the site info dat file to be sourced from this workspace
    $Log->comment("Looking for syslog info for site '$site'");
    my %siteinfo_local  = get_site_info(force_workspace_site_info_file => 1);
    my $syslog_server   = $siteinfo_local{$site}->{syslog_server};
    my $syslog_port     = $siteinfo_local{$site}->{syslog_port};

    if (!$syslog_server || !$syslog_port){
        $self->log_error("Missing syslog entries for site '$site'. Contact jroussin\@netapp.com when you see this error message and provide a path to these logs");
        $self->disable_syslog;
        return;
    }
    else{
        $Log->comment("Syslog server: " . $syslog_server);
        $Log->comment("Syslog port  : " . $syslog_port);

    }
    $self->syslog_port($syslog_port);

    # Only used by CITs to avoid failed tests when graylog goes down
    if ($self->mock_graylog){
        if (!$self->mock_graylog_port){
            confess("Missing mock graylog port");
        }
        $syslog_server = '127.0.0.1';
        $self->syslog_port($self->mock_graylog_port);
    }

    return $syslog_server;
}

sub _populate_top_logdir {
    my ($self) = @_;
    my $ParamSet = NATE::ParamSet->new(global=>1);
    my $logdir = $ParamSet->get('TOP_LOGDIR');
    return $logdir;
}

sub _populate_nate_uuid {
    my ($self) = @_;
    my $db = testdb_global();
    my $ug = Data::UUID->new();
    my $uuid_binary = $db->get_ntest_uuid();
    my $uuid = $ug->to_string($uuid_binary);
    return $uuid;
}

sub _populate_user {
    my ($self) = @_;
    my $db = testdb_global();
    my $user = SystemTest::ASUPMessage::identify_test_runner($db);
    return $user;
}

1;
