#
# Copyright (c) 2017 NetApp, Inc., All Rights Reserved.
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

package NACL::Service::ASUPTrigger::TestWatcher;

use v5.14;
use strict;
use warnings FATAL => 'all';
use Carp qw/confess/;
use Data::Dumper::Concise qw/Dumper/;
use MooseX::Params::Validate qw/validated_list validated_hash/;

use Moose::Role;

# There should only be one running normal test at any given time
has running_test => (
    is        => 'rw',
    isa       => 'HashRef',
    clearer   => 'clear_running_test',
    predicate => 'has_running_test',
);

has completed_tests => (
    is      => 'rw',
    isa     => 'HashRef[HashRef]',
    traits  => ['Hash'],
    default => sub{{}},
    clearer => 'clear_completed_tests',
    handles => {
        is_test_completed   => 'exists',
        add_completed_test  => 'set',
        completed_tests_list=> 'values',
    }
);

# There could be 0 to many testcases running
has running_testcases => (
    is      => 'rw',
    isa     => 'HashRef[HashRef]',
    traits  => ['Hash'],
    default => sub{{}},
    clearer => 'clear_running_testcases',
    handles => {
        is_testcase_running     => 'exists',
        add_running_testcase    => 'set',
        remove_running_testcase => 'delete',
        running_testcases_list  => 'values',
        has_running_testcases   => 'count',
    }
);

# Keep track of testcases that came and went without having been
# associated with an ASUP payload
# Per node backlog
# node => { id => 1 }
has completed_testcase_backlog => (
    is      => 'rw',
    isa     => 'HashRef[HashRef]',
    clearer => 'clear_completed_testcase_backlog',
    default => sub{{}},
);

has completed_testcases => (
    is      => 'rw',
    isa     => 'HashRef[HashRef]',
    traits  => ['Hash'],
    default => sub{{}},
    clearer => 'clear_completed_testcases',
    handles => {
        is_testcase_completed   => 'exists',
        add_completed_testcase  => 'set',
    }
);

sub is_test_running {
    my ($self, $id) = @_;
    confess "You must specify test id" if (!defined $id);
    return 0 if ( !$self->has_running_test);
    return 1 if ($self->running_test->{id} == $id);
    return 0;
}

sub set_running_test {
    my ($self, %info) = validated_hash(
        \@_,
        id      => { isa => 'Int', optional => 0 },
        name    => { isa => 'Str', optional => 0 },
    );
    if ($self->has_running_test){
        confess "There is already a running test";
    }
    $self->running_test(\%info);
    return;
}

sub clear_test_state {
    my ($self) = @_;
    $self->clear_completed_tests;
    $self->clear_running_test;
    $self->clear_completed_testcases;
    $self->clear_running_testcases;
    $self->clear_completed_testcase_backlog;
    return;
}

sub dump_test_status {
    my ($self) = @_;
    return join("\n",
        "Running test:\n".          Dumper($self->running_test),
        "Completed tests:\n".       Dumper($self->completed_tests),
        "Running testcases:\n".     Dumper($self->running_testcases),
        "Completed testcases:\n".   Dumper($self->completed_testcases),
        "Backlog testcases:\n".     Dumper($self->completed_testcase_backlog),
    );
}

sub current_test_id {
    my ($self) = @_;
    return if (!$self->has_running_test);
    return $self->running_test->{id};
}

sub running_testcase_ids {
    my ($self) = @_;
    return if (!$self->has_running_testcases);
    return [map { $_->{id} } $self->running_testcases_list];
}

sub completed_testcase_backlog_ids {
    my ($self, $node) = validated_list(\@_,
        node     => { isa => 'Str', optional => 0 },
    );
    return if (!$self->has_completed_testcase_backlog_for_node(node => $node));
    return [ keys %{$self->completed_testcase_backlog()->{$node}} ];
}

sub has_completed_testcase_backlog {
    my ($self) = @_;
    my $backlog = $self->completed_testcase_backlog;
    return 0 if (!$backlog);
    return 0 if (!keys %$backlog);
    return 1;
}

sub has_completed_testcase_backlog_for_node {
    my ($self, $node) = validated_list(\@_,
        node     => { isa => 'Str', optional => 0 },
    );
    return 0 if ( ! $self->completed_testcase_backlog() );
    my $backlog = $self->completed_testcase_backlog()->{$node};
    return 0 if (!$backlog);
    return 0 if (! keys %$backlog);
    return 1;
}

sub add_completed_testcase_to_backlog {
    my ($self, $id, $node) = validated_list(\@_,
        testcase => { isa => 'Int', optional => 0 },
        node     => { isa => 'Str', optional => 0 },
    );
    my $backlog = $self->completed_testcase_backlog;
    if (!$backlog){
        $backlog = { $node => { $id => 1 } };
    }
    else{
        $backlog->{$node}->{$id} = 1;
    }
    $self->completed_testcase_backlog($backlog);
    return;
}

sub clear_completed_testcase_backlog_for_node {
    my ($self, $node) = validated_list(\@_,
        node     => { isa => 'Str', optional => 0 },
    );
    if ( $self->completed_testcase_backlog ){
        delete $self->completed_testcase_backlog()->{$node};
    }
    return;
}

1;
