# Any use, modification, or distribution is prohibited
# without prior written consent from NetApp, Inc.

## @summary Worker code for discovery requests
## @author bellizzi@netapp.com
## @pod here

=head1 NAME

NACL::Service::Discover::Worker - Worker code for discovery requests

=cut

package NACL::Service::Discover::Worker;
use strict;
use warnings;
use v5.14;

# Seems to be required by NACL
use Tharn;

use Hostrec;
use NATE::Log qw(log_global);
use NATE::ParamSet;
use NATE::Inc qw(find_file);
use NATE::ServiceAPI::Resource qw(resource_global);

use parent qw(NACL::Service::Worker);
use NACL::C::Node;
use NACL::C::Volume;
use NACL::STask::Cluster;
use NACL::InstantComponent qw(instant_component);
use NACL::STask::Cluster;
use NATE::BaseException qw(:try);
use NATE::Exceptions::Argument;
use NACL::Exceptions::InvalidChoice;
use NACL::Exceptions::NoElementsFound;
use NACL::APISet::Exceptions::ResponseException;
use NACL::APISet::Exceptions::MethodNotFoundException;

use POSIX qw(strftime);
use Data::Dumper::Concise;
use XML::Simple;
use Devel::Size qw(total_size);
use Sort::Key::Natural qw(natsort natkeysort);
use Params::Validate
    qw(validate_with BOOLEAN HASHREF OBJECT ARRAYREF SCALAR UNDEF);

$Data::Dumper::Sortkeys = 1;

my $log = log_global();

sub init {
    my $self = shift;
    $self->SUPER::init(@_);

    # Call new on the resource service to ensure it is started
    my $rs = NATE::ServiceAPI::Resource->new();

} ## end sub init

sub discover {
    my $self     = shift;
    my %incoming = validate_with(
        params => \@_,
        spec   => {
            discover_homedirs         => {type => BOOLEAN, default => 0},
            discover_local_cifs_users => {type => BOOLEAN, default => 0},
            discover_local_unix_users => {type => BOOLEAN, default => 0},
            import_remote_users       => {type => BOOLEAN, default => 0},
            nacl_objects              => {type => HASHREF, default => {}},
            std_objects               => {type => ARRAYREF, default => []},
            max_processes             => {type => SCALAR,  default => 10},
            remove_existing_users     => {type => BOOLEAN, default => 0},
            user_specifications       => {type => HASHREF, default => {}}
        }
    );

    # Sort the node objects by cluster UUID
    my @nodes = NACL::C::Node->find();
    my $nodes_by_cluster
        = NACL::STask::Cluster->group_nodes_by_unique_clusters(
        nodes => \@nodes);

    # Discover the requested ONTAP objects
    $self->_discover_ontap(
        max_processes    => $incoming{max_processes},
        nacl_objects     => $incoming{nacl_objects},
        std_objects      => $incoming{std_objects},
        nodes_by_cluster => $nodes_by_cluster
    );

    if (   $incoming{discover_homedirs}
        || $incoming{discover_local_cifs_users}
        || $incoming{discover_local_unix_users}
        || keys(%{$incoming{user_specifications}})) {

        $self->_create_discover_lock(resource_type => 'user');

        if ($incoming{remove_existing_users}) {
            $self->_remove_resources(resource_type => 'user');
        }

        # Discover the directories/qtrees in all CIFs homedir search paths
        my $homedirs = {};
        if ($incoming{discover_homedirs}) {
            foreach my $cluster_uuid (keys(%{$nodes_by_cluster})) {

                my $node = ${$nodes_by_cluster->{$cluster_uuid}}[0];
                $homedirs = $self->_discover_cifs_home_directories(
                    cluster_uuid => $cluster_uuid,
                    node         => $node->node()
                );
            }
        }

        # Import remote users and match them to their homedirs
        if (keys(%{$incoming{user_specifications}})) {
            $self->_import_users(
                homedirs            => $homedirs,
                user_specifications => $incoming{user_specifications},
            );
        }

        foreach my $cluster_uuid (keys(%{$nodes_by_cluster})) {

            my @users;

            # Import ONTAP local CIFS users and match them to their homedirs
            if ($incoming{discover_local_cifs_users}) {
                $self->_discover_cifs_local_users(
                    cluster_uuid => $cluster_uuid,
                    homedirs     => $homedirs,
                );
            }

            # Import ONTAP local unix  users
            if ($incoming{discover_local_unix_users}) {
                $self->_discover_unix_local_users(
                    cluster_uuid => $cluster_uuid,);
            }
        }

        $self->_remove_discover_lock(resource_type => 'user');
    }

    $log->comment('Discover request complete');
} ## end sub discover

sub _discover_ontap {

    my $self     = shift;
    my %incoming = validate_with(
        params => \@_,
        spec   => {
            nacl_objects     => {type => HASHREF},
            std_objects      => {type => ARRAYREF},
            max_processes    => {type => SCALAR, default => 10},
            nodes_by_cluster => {type => HASHREF}
        },
    );

    my $nacl_objects     = $incoming{nacl_objects};
    my $std_objects      = $incoming{std_objects};
    my $max_processes    = $incoming{max_processes};
    my $nodes_by_cluster = $incoming{nodes_by_cluster};

    # Sort the NACL object names
    my @objects = grep { $_ ne "Vserver" } natsort(keys(%{$nacl_objects}));

    if (scalar(@{$std_objects})) {
        # Remove any resources in objects that is also an std_object
        my @tmp = grep { my $resource = $_; !grep { $resource eq $_ } @{$std_objects} } @objects;
        @objects = @tmp;
    }

    # Create an index for sub process uniqueness
    my $proc_index = 1;

    # Create and epoch timestamp for process uniqueness
    my $time = time();

    # Discover the requested NACL objects for each cluster
    foreach my $cluster_uuid (keys(%{$nodes_by_cluster})) {

        my @nodes = @{$nodes_by_cluster->{$cluster_uuid}};
        my $node  = shift(@nodes);

        if (grep { $_ eq 'Vserver' } keys(%{$nacl_objects})) {

            # Get the vserver object first since we will add the UUID of the
            # owning vserver if an object has a vserver attribute
            my $runid = $time . '_0_Vserver';
            my $proc  = $self->_create_ontap_discover_process(
                cluster_uuid => $cluster_uuid,
                nacl_objects => $nacl_objects,
                node         => $node->node(),
                object       => 'Vserver',
                runid        => $runid
            );
            $log->comment('Waiting for discover process to complete');
            $proc->wait();
        } 
        
        # Create an array to store process objects
        my @procs;

        # Spawn a process for each NACL object
        foreach my $object (@objects) {
            # Here we build a list of node names. Whenever we create a
            # a process we pass that process a node name so that
            # process connects only to that node thereby distributing
            # the discover load across the nodes in a cluster
            if (!@nodes) {
                @nodes = @{$nodes_by_cluster->{$cluster_uuid}};
            }
            $node = shift(@nodes);

            # Create a unique runid
            my $runid = $time . '_' . $proc_index . '_' . $object;

            my $proc = $self->_create_ontap_discover_process(
                cluster_uuid => $cluster_uuid,
                nacl_objects => $nacl_objects,
                node         => $node->node(),
                object       => $object,
                runid        => $runid
            );

            # Store the process object
            push(@procs, $proc);

            # Increment the process index
            $proc_index++;

            # If we started enough procs wait for all of them to complete
            if (($proc_index % $max_processes) == 0
                || $object eq $objects[$#objects]) {
                $log->comment('Waiting for discover processes to complete');

                # This will wait for all processes to complete
                foreach (@procs) { $_->wait(); }
                @procs = ();
            }
        }

        # Spawn a process for each NACL object
        foreach my $std_object (@{$std_objects}) {
            # Here we build a list of node names. Whenever we create a
            # a process we pass that process a node name so that
            # process connects only to that node thereby distributing
            # the discover load across the nodes in a cluster
            if (!@nodes) {
                @nodes = @{$nodes_by_cluster->{$cluster_uuid}};
            }
            $node = shift(@nodes);

            # Create a unique runid
            my $runid = $time . '_' . $proc_index . '_' . $std_object;

            my $proc = $self->_create_ontap_discover_process(
                cluster_uuid => $cluster_uuid,
                nacl_objects => $nacl_objects,
                node         => $node->node(),
                object       => $std_object,
                runid        => $runid
            );

            # Increment the process index
            $proc_index++;

            $proc->wait();
        }
    }

} ## end sub _discover_ontap

sub _create_ontap_discover_process {

    my $self     = shift;
    my %incoming = validate_with(
        params => \@_,
        spec   => {
            cluster_uuid => {type => SCALAR},
            nacl_objects => {type => HASHREF},
            node         => {type => SCALAR},
            object       => {type => SCALAR},
            runid        => {type => SCALAR},
        },
    );

    # Prepare the coderef
    my $code = '{
        use NACL::Service::Discover::Worker;
        NACL::Service::Discover::Worker->discover_nacl_ontap_object(@_);
    }';

    $log->comment("Creating process for $incoming{runid}");

    my $proc = $self->process_start(
        codespec       => $code,
        runid          => $incoming{runid},
        inherit_params => 1,
        args           => [
            nacl_objects => $incoming{nacl_objects},
            cluster_uuid => $incoming{cluster_uuid},
            node         => $incoming{node},
            object       => $incoming{object},
        ],
        on_termination =>
            [\&process_exit_status, $incoming{node}, $incoming{object}]
    );
    return $proc;
} ## end sub _create_ontap_discover_process

sub discover_nacl_ontap_object {

    my $self     = shift;
    my %incoming = validate_with(
        params => \@_,
        spec   => {
            cluster_uuid => {type => SCALAR},
            node         => {type => SCALAR},
            nacl_objects => {type => HASHREF},
            object       => {type => SCALAR}
        },
    );

    # Assign variables from incoming arguments
    my $node         = $incoming{node};
    my $nacl_objects = $incoming{nacl_objects};
    my $cluster_uuid = $incoming{cluster_uuid};
    my $object       = $incoming{object};

    $self->_create_discover_lock(resource_type => $object);

    $log->comment("Discovering attributes for $object");

    # Connect to the node
    $log->comment("Establishng command interface to $node");
    my $ci = NACL::C::Node->new(node => $node);

    # Force NACL to connect now
    my $apiset = $ci->apiset();
    $apiset->set_timeout('timeout' => 600);

    # Call instant component because it will work for file
    # and instant NACL objects
    my $c = "NACL::C::$object";
    $log->comment("Calling instant component for $c");
    instant_component(
        component_name    => $c,
        command_interface => $ci
    );

    # Get the attributes for the object
    my $valid_attributes = $self->_get_nacl_attributes(
        nacl_objects => $nacl_objects,
        object       => $object
    );

    # Call fetch on the NACL component state object
    my $cs = "NACL::CS::$object";
    $log->comment("Fetching NACL data for $cs");
    my @result;
    try {
        @result = $cs->fetch(command_interface => $ci);
    }
    catch NATE::Exceptions::Argument with {
        my $ex = shift;
        die $cs . " Argument error: " . $ex->text() . "\n";
    }
    catch NACL::Exceptions::InvalidChoice with {
        my $ex = shift;
        die $cs . " InvalidChoice error: " . $ex->text() . "\n";
    }
    catch NACL::Exceptions::NoElementsFound with {
        $log->debug($cs . ": no data found");
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $ex = shift;
        die $cs . " Response error: " . $ex->text() . "\n";
    }
    catch NACL::APISet::Exceptions::MethodNotFoundException with {
        my $ex = shift;
        die $cs . " MethodNotFound error: " . $ex->text() . "\n";
    }
    otherwise {
        my $ex = shift;
        die $cs . " Unknown error: " . $ex->text() . "\n";
    };

    my $vservers = $self->get_resources(
        resource_type => 'Vserver',
        filter        => {cluster_uuid => {$cluster_uuid => '='}}
    );

    my @resources;
    foreach my $r (@result) {

        # Create a hash to hold our key value pairs
        my %entry;
        $entry{resource_type} = $object;
        $entry{max_instances} = 9999;

        # Add only the key value pairs we want
        my %values;
        foreach my $a (keys(%{$valid_attributes})) {
            if (defined($r->{$a})) {
                if ($valid_attributes->{$a} eq 'array') {
                    $values{$a} = join(',', @{$r->{$a}});
                } else {
                    $values{$a} = $r->{$a};
                }
                if ($values{$a} eq '-') { $values{$a} = ''; }
            }
        }

        # If it does not exist add the cluster UUID
        if (!defined($values{'cluster_uuid'})) {
            $values{'cluster_uuid'} = $cluster_uuid;
        }

        # If the vserver attribute exists add the vserver UUID
        if (defined($values{'vserver'})) {
            if ($object eq 'Vserver') {

                # Since we are the vserver object use the UUID value
                # directly
                $values{'vserver_uuid'} = $values{uuid};
            } else {

                # pull uuid from lookup table
                my @match = grep { $_->{'vserver'} eq $values{vserver} }
                    @{$vservers};
                $values{'vserver_uuid'} = $match[0]->{uuid};
                if (!$values{'vserver_uuid'}) {
                    die "received no vserver uuid from resource discovery";
                }
            }
        }

        # Add the attributes to the request hash
        $entry{attributes} = \%values;

        push(@resources, \%entry);

    }

    $self->_remove_resources(
        resource_type => $object,
        filter        => {cluster_uuid => {$cluster_uuid => '='}}
    );

    $self->_register_resources(
        resource_type => $object,
        resources     => \@resources
    );

    $self->_remove_discover_lock(resource_type => $object);

    $log->comment("$cs discovery complete");
} ## end sub discover_nacl_ontap_object

sub _get_nacl_attributes {

    my $self     = shift;
    my %incoming = validate_with(
        params => \@_,
        spec   => {
            nacl_objects => {type => HASHREF},
            object       => {type => SCALAR}
        },
    );

    my $ci           = $incoming{command_interface};
    my $nacl_objects = $incoming{nacl_objects};
    my $object       = $incoming{object};

    $log->comment("Getting attributes for $object");
    my $cs = "NACL::CS::$object";
    my $attributes;
    my $err;

    # Report exceptions but do not die
    try {
        $attributes = $cs->get_attributes();
    }
    catch NATE::Exceptions::Argument with {
        my $ex = shift;
        $err = $cs . " Argument error: " . $ex->text();
    }
    catch NACL::Exceptions::InvalidChoice with {
        my $ex = shift;
        $err = $cs . " InvalidChoice error: " . $ex->text();
    }
    catch NACL::Exceptions::NoElementsFound with {
        $log->debug($cs . ": no data found");
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $ex = shift;
        $err = $cs . " Response error: " . $ex->text();
    }
    catch NACL::APISet::Exceptions::MethodNotFoundException with {
        my $ex = shift;
        $err = $cs . " MethodNotFound error: " . $ex->text();
    };

    if ($log->may_debug()) { $log->debug(Dumper($attributes)); }

    my %filtered_attributes;
    foreach my $a (keys(%{$attributes})) {

        # If there are explict attributes specified by the requestor then
        # return only those attributes
        if (defined($nacl_objects->{$object}->{attributes})
            && !grep { $_ eq $a } @{$nacl_objects->{$object}->{attributes}}) {
            next;
        }

        # If there are ignored attributes specified by the requestor then
        # return everything but those attributes
        if (defined($nacl_objects->{$object}->{ignored_attributes})
            && grep { $_ eq $a }
            @{$nacl_objects->{$object}->{ignored_attributes}}) {
            next;
        }
        $filtered_attributes{$a} = $attributes->{$a};
    }
    return \%filtered_attributes;

} ## end sub _get_nacl_attributes

sub _discover_cifs_home_directories {

    my $self     = shift;
    my %incoming = validate_with(
        params => \@_,
        spec   => {
            cluster_uuid => {type => SCALAR},
            node         => {type => SCALAR},
        },
    );

    my $cluster_uuid = $incoming{cluster_uuid};
    my $node         = $incoming{node};

    $log->comment("Establishng command interface to $node");
    my $ci = NACL::C::Node->new(node => $node);

    my $cluster_name
        = $self->_get_cluster_name(cluster_uuid => $cluster_uuid);

    # Get the discovered home directory search paths
    my $search_paths = $self->get_resources(
        resource_type => 'VserverCifsHomeDirectorySearchPath',
        filter        => {cluster_uuid => {$cluster_uuid => '='}}
    );

    my $vservers = $self->get_resources(
        resource_type => 'Vserver',
        filter        => {cluster_uuid => {$incoming{cluster_uuid} => '='},}
    );

    my $volumes = $self->get_resources(
        resource_type => 'Volume',
        filter        => {cluster_uuid => {$incoming{cluster_uuid} => '='},}
    );

    my $qtrees = $self->get_resources(
        resource_type => 'VolumeQtree',
        filter        => {cluster_uuid => {$incoming{cluster_uuid} => '='},}
    );

    my $homedirs = {};

    foreach my $homedir (@{$qtrees}) {
        my $volume  = $homedir->{volume};
        my $vserver = $homedir->{vserver};
        my $qtree   = $homedir->{qtree};

        my @vserver_match = grep { $_->{'vserver'} eq $vserver } @{$vservers};
        my $vserver_uuid = $vserver_match[0]->{uuid};
        if (!$vserver_uuid) {
            die "received no vserver uuid from resource discovery";
        }

        my @volume_match = grep { $_->{'volume'} eq $volume } @{$volumes};
        my $junction_path = $volume_match[0]->{junction_path};

        $homedirs->{$cluster_uuid}->{$vserver_uuid}->{$qtree}
            = $junction_path . '/' . $qtree;
    }

    return $homedirs;
} ## end sub _discover_cifs_home_directories

sub _match_user_homedirs {

    my $self     = shift;
    my %incoming = validate_with(
        params => \@_,
        spec   => {
            cluster_uuid => {type => SCALAR | UNDEF, optional => 1},
            homedirs     => {type => HASHREF},
            user         => {type => HASHREF},
            vserver_uuid => {type => SCALAR | UNDEF, optional => 1},
        },
    );

    my $cluster_uuid = $incoming{cluster_uuid};
    my $homedirs     = $incoming{homedirs};
    my $user         = $incoming{user};
    my $vserver_uuid = $incoming{vserver_uuid};

    my @matched;
    foreach my $cuuid (keys(%{$homedirs})) {
        if (defined($cluster_uuid) && $cuuid ne $cluster_uuid) { next; }
        foreach my $vuuid (keys(%{$homedirs->{$cuuid}})) {
            if (defined($vserver_uuid) && $vuuid ne $vserver_uuid) {
                next;
            }
            foreach my $dir (keys(%{$homedirs->{$cuuid}->{$vuuid}})) {

                # Skip if the user name does not match the user name
                if ($dir eq $user->{user_name}) {

                    # Add the homedir path and vserver to the user info
                    my $path = $homedirs->{$cuuid}->{$vuuid}->{$dir};

                    $log->comment(
                        "User name $user->{user_name} matches $cuuid:$vuuid:$path/$dir"
                    );

                    push(
                        @matched,
                        {   cluster_uuid => $cuuid,
                            homedir      => $path,
                            vserver_uuid => $vuuid
                        }
                    );
                }
            }
        }
    }
    $log->comment($user->{user_name} . ' matched ' . @matched . ' homedirs');
    return \@matched;
} ## end sub _match_user_homedirs

sub _import_users {

    my $self     = shift;
    my %incoming = validate_with(
        params => \@_,
        spec   => {
            homedirs            => {type => HASHREF},
            user_specifications => {type => HASHREF}
        },
    );

    my $homedirs = $incoming{homedirs};
    my $users    = $self->parse_user_specifications(
        user_specifications => $incoming{user_specifications});

    $log->comment('Importing ' . @{$users} . ' remote users');

    my @resources;
    foreach my $u (@{$users}) {

        my $matched = $self->_match_user_homedirs(
            homedirs => $homedirs,
            user     => $u
        );
        if (@{$matched}) {
            my @matched_cluster_uuids;
            my @matched_vserver_uuids;
            my @matched_homedirs;
            foreach my $match (@{$matched}) {
                if (   defined($match->{cluster_uuid})
                    && defined($match->{vserver_uuid})
                    && defined($match->{homedir})) {
                    push(@matched_cluster_uuids, $match->{cluster_uuid});
                    push(@matched_vserver_uuids, $match->{vserver_uuid});
                    push(@matched_homedirs,      $match->{homedir});
                }
            }
            $u->{cluster_uuid} = join(';', @matched_cluster_uuids);
            $u->{vserver_uuid} = join(';', @matched_vserver_uuids);
            $u->{homedir}      = join(';', @matched_homedirs);
        }

        my %entry;
        $entry{resource_type} = 'user';
        $entry{max_instances} = 9999;
        $entry{attributes}    = $u;
        push(@resources, \%entry);

    }

    $self->_register_resources(
        resource_type => 'user',
        resources     => \@resources
    );

} ## end sub _import_users

sub parse_user_specifications {

    my $self     = shift;
    my %incoming = validate_with(
        params => \@_,
        spec   => {user_specifications => {type => HASHREF}}
    );

    if ($log->may_debug()) {
        $log->debug("Generating users from:\n"
                . Dumper($incoming{user_specifications}));
    }

    my $users = [];
    foreach my $key (keys(%{$incoming{user_specifications}})) {
        my @user_specifications;
        if (ref($incoming{user_specifications}->{$key}->{users}) eq 'ARRAY') {
            @user_specifications
                = @{$incoming{user_specifications}->{$key}->{users}};
        } elsif (defined($incoming{user_specifications}->{$key}->{users})) {
            push(@user_specifications,
                $incoming{user_specifications}->{$key}->{users});
        }

        $log->comment("Detected following user specifications:\n"
                . Dumper(\@user_specifications));

        if ($log->may_debug()) {
            $log->debug("Detected following user specifications:\n"
                    . Dumper(\@user_specifications));
        }

        foreach my $us (@user_specifications) {
            my $ad           = $us->{ad}           || "";
            my $count        = $us->{count}        || "";
            my $startcount   = $us->{startcount}   || "";
            my $host         = $us->{host}         || "";
            my $hosttype     = $us->{hosttype}     || "";
            my $nis          = $us->{nis}          || "";
            my $password     = $us->{password}     || "";
            my $user_name    = $us->{user_name}    || "";
            my $zero_padding = $us->{zero_padding} || "";
            my $protocol     = $us->{protocol}     || "";
            my $domain       = $us->{domain}       || "";
            my $services     = $us->{services}     || "";

            if (!defined($startcount) || ($startcount eq "")) {
                $startcount = 1;
            }

            # Create each users entry in the array
            if ($count) {

                # If the count tag is specified do the work to
                # create that many users
                my $endcount = $startcount + $count;
                for (my $i = $startcount; $i < $endcount; $i++) {

                    # Copy the counter var so we can pad it later if
                    # needed
                    my $idx = $i;

                    # Pad the user index if required
                    if ($zero_padding) {
                        $idx = sprintf("%0*d", $zero_padding, $i);
                    }

                    # If the password tag is not specified in the
                    # XML then use the user name as the password
                    if (!defined($us->{password})) {
                        $password = $user_name . $idx;
                    }

                    # Push the user info into the users AoH
                    push(
                        @{$users},
                        {   cluster_uuid => '',
                            user_name    => $user_name . $idx,
                            ad           => $ad,
                            host         => $host,
                            hosttype     => $hosttype,
                            nis          => $nis,
                            password     => $password,
                            protocol     => $protocol,
                            domain       => $domain,
                            services     => $services
                        }
                    );
                }
            } else {

                # If the password tag is not specified in the XML then
                # use the user name as the password
                if (!$password) {
                    $password = $user_name;
                }

                # Push the user info into the users AoH
                push(
                    @{$users},
                    {   cluster_uuid => '',
                        user_name    => $user_name,
                        ad           => $ad,
                        host         => $host,
                        hosttype     => $hosttype,
                        nis          => $nis,
                        password     => $password,
                        protocol     => $protocol,
                        domain       => $domain,
                        services     => $services
                    }
                );
            }
        }
    }
    return $users;
} ## end sub parse_user_specifications

sub _discover_cifs_local_users {

    my $self     = shift;
    my %incoming = validate_with(
        params => \@_,
        spec   => {
            cluster_uuid => {type => SCALAR},
            homedirs     => {type => HASHREF},
        },
    );

    my $cluster_uuid = $incoming{cluster_uuid};
    my $homedirs     = $incoming{homedirs};
    my $id           = $incoming{id};

    $log->comment("Discovering CIFS local users for cluster $cluster_uuid");

    # Get the disacovered cifs server objects
    my $cifs_servers = $self->get_resources(
        resource_type => 'VserverCifs',
        filter        => {cluster_uuid => {$cluster_uuid => '='}}
    );

    # Create a hash for easy lookup
    my %cifs_server_lookup;
    foreach my $cs (@{$cifs_servers}) {
        if (defined($cs->{vserver}) && defined($cs->{cifs_server})) {
            $cifs_server_lookup{$cs->{vserver}} = $cs->{cifs_server};
        }
    }

    # Get the discovered CIFS local users
    my $local_users = $self->get_resources(
        resource_type => 'VserverCifsUsersAndGroupsLocalUser',
        filter        => {
            cluster_uuid        => {$cluster_uuid => '='},
            is_account_disabled => {false         => '='},
        }
    );

    $log->comment('Processing ' . @{$local_users} . ' CIFS local users');

    if (@{$local_users}) {

        my $vservers = $self->get_resources(
            resource_type => 'Vserver',
            filter        => {cluster_uuid => {$cluster_uuid => '='}}
        );

        # Walk through the list of local users
        my @resources;
        foreach my $lu (@{$local_users}) {

            # Strip the host from the user_name value
            my $user_name = $lu->{user_name};
            my $vserver   = $lu->{vserver};

            # pull uuid from lookup table
            my @match = grep { $_->{'vserver'} eq $vserver } @{$vservers};
            my $vserver_uuid = $match[0]->{uuid};
            if (!$vserver_uuid) {
                die "received no vserver uuid from resource discovery";
            }

            my $host = $cifs_server_lookup{$vserver};
            $user_name =~ s/$host\\//g;

            # Push the user info into the users AoH
            my %user = (
                cluster_uuid => $cluster_uuid,
                vserver_uuid => $vserver_uuid,
                user_name    => $user_name,
                ad           => '',
                host         => $host,
                hosttype     => 'filer-ng-cfe',
                nis          => '',
                password     => $lu->{full_name},
                protocol     => 'cifs',
                domain       => '',
                services     => ''
            );
            my $matched = $self->_match_user_homedirs(
                homedirs     => $homedirs,
                cluster_uuid => $cluster_uuid,
                vserver_uuid => $vserver_uuid,
                user         => \%user
            );
            if (@{$matched}) {
                $user{homedir} = $matched->[0]->{homedir};
            }
            my %entry;
            $entry{resource_type} = 'user';
            $entry{max_instances} = 9999;
            $entry{attributes}    = \%user;

            push(@resources, \%entry);

        }

        $self->_register_resources(
            resource_type => 'user',
            resources     => \@resources
        );

    }

} ## end sub _discover_cifs_local_users

sub _discover_unix_local_users {

    my $self     = shift;
    my %incoming = validate_with(
        params => \@_,
        spec   => {cluster_uuid => {type => SCALAR},},
    );

    my $cluster_uuid = $incoming{cluster_uuid};

    $log->comment("Discovering unix local users");

    # Get the discovered unix local users
    my $local_users = $self->get_resources(
        resource_type => 'VserverServicesNameServiceUnixUser',
        filter        => {cluster_uuid => {$cluster_uuid => '='}}
    );

    $log->comment('Processing ' . @{$local_users} . ' unix local users');

    if (@{$local_users}) {

        my $vservers = $self->get_resources(
            resource_type => 'Vserver',
            filter        => {cluster_uuid => {$cluster_uuid => '='}}
        );

        # Walk through the list of local users
        my @resources;
        foreach my $lu (@{$local_users}) {

            if (!defined($lu->{user_name}) || $lu->{user_name} !~ /\w+/) {
                next;
            }

            my $vserver = $lu->{vserver};

            # pull uuid from lookup table
            my @match = grep { $_->{'vserver'} eq $vserver } @{$vservers};
            my $vserver_uuid = $match[0]->{uuid};
            if (!$vserver_uuid) {
                die "received no vserver uuid from resource discovery";
            }

            # Push the user info into the users AoH
            my %user = (
                cluster_uuid => $cluster_uuid,
                vserver_uuid => $vserver_uuid,
                user_name    => $lu->{user_name},
                ad           => '',
                host         => $vserver,
                hosttype     => 'filer-ng-cfe',
                nis          => '',
                password     => $lu->{full_name},
                protocol     => 'nfs',
                domain       => '',
                services     => ''
            );

            my %entry;
            $entry{resource_type} = 'user';
            $entry{max_instances} = 9999;
            $entry{attributes}    = \%user;

            push(@resources, \%entry);

        }

        $self->_register_resources(
            resource_type => 'user',
            resources     => \@resources
        );

    }

} ## end sub _discover_unix_local_users

sub _get_relative_volume_path {

    my $self     = shift;
    my %incoming = validate_with(
        params => \@_,
        spec   => {
            path    => {type => SCALAR},
            volumes => {type => ARRAYREF}
        },
    );

    my $path    = $incoming{path};
    my $volumes = $incoming{volumes};

    $log->debug("Getting volume from homedir search path $path");

    # Try the exact match
    $log->debug("Searching volume junction paths for an exact match");
    my @matches = grep { $_->{junction_path} eq $path } @{$volumes};

    # Keep deconstructing the path and comparing it against known
    # junction-paths until we find a matching volume. I'm assuming
    # that junction paths are unique
    if (!@matches) {
        $log->debug("No exact junction_path match found "
                . "for $path, deconstructing");
        my @elements = split(/\//, $path);
        if (@elements) {
            while (@elements) {
                pop(@elements);
                my $current_path = join('/', @elements);

                # If the current path is undef, this means we are down to the
                # two elements / and <dir>. Reset the current path to be /
                # so we match the root volume.
                if (!$current_path) { $current_path = '/'; }
                $log->debug(
                    "Checking for junction paths matching $current_path");
                @matches = grep { $_->{junction_path} eq $current_path }
                    @{$volumes};
                if (@matches) { last; }
            }
        }
    }

    # Now that we have a volume, create the relative path
    my @elements = split(/\//, $path);
    while (@elements && $elements[0] ne $matches[0]->{volume}) {
        shift(@elements);
    }

    # Shift off the volume
    my $volume = shift(@elements);

    my $relative_path = join('/', @elements);
    if ($relative_path !~ /\/$/) { $relative_path .= '/'; }
    return ($volume, $relative_path);
} ## end sub _get_relative_volume_path

sub _get_vserver_uuid {

    my $self     = shift;
    my %incoming = validate_with(
        params => \@_,
        spec   => {
            cluster_uuid => {type => SCALAR},
            vserver      => {type => SCALAR}
        },
    );

    # Get the discovered unix local users
    my $vservers = $self->get_resources(
        resource_type => 'Vserver',
        filter        => {
            cluster_uuid => {$incoming{cluster_uuid} => '='},
            vserver      => {$incoming{vserver}      => '='}
        }
    );

    my $vserver_uuid = $vservers->[0]->{uuid};
    if (!$vserver_uuid) {
        die "received no vserver uuid from resource discovery";
    }
    return $vserver_uuid;
} ## end sub _get_vserver_uuid

sub _get_cluster_name {

    my $self     = shift;
    my %incoming = validate_with(
        params => \@_,
        spec   => {cluster_uuid => {type => SCALAR}},
    );

    my $clusters = $self->get_resources(
        resource_type => 'ClusterIdentity',
        filter        => {cluster_uuid => {$incoming{cluster_uuid} => '='},}
    );

    my $cluster_name = $clusters->[0]->{name};
    return $cluster_name;
} ## end sub _get_cluster_name

sub _register_resources {

    my $self     = shift;
    my %incoming = validate_with(
        params => \@_,
        spec   => {
            resources      => {type => ARRAYREF},
            resource_type  => {type => SCALAR},
            max_block_size => {type => SCALAR, default => 15999999}
        },
    );

    $log->comment('Registering ' . @{$incoming{resources}} . ' resources');

    my $max_block_size  = $incoming{max_block_size};
    my $available_space = $incoming{max_block_size};
    my @entries         = ();
    my $request_count   = 0;
    my $block_count     = 0;
    foreach my $r (@{$incoming{resources}}) {

        # Create a hash to hold our key value pairs
        my %entry = %{$r};

        # Get the size of the request hash
        my $bytes = total_size([\%entry]);

        $available_space -= $bytes;

        # If there is space then add request to block
        if ($available_space >= 0) {
            my $current_block = $block_count + 1;
            $log->comment(
                "Adding entry of size ($bytes) to request ($current_block)");
            push @entries, \%entry;
            $request_count++;
        }

        # If there is no more space, register the block entries
        if ($available_space <= 0) {

            # Perform the resource service register request
            my $rs = resource_global();
            $rs->register(entries => [@entries]);
            @entries = ();
            $block_count++;
            my $tmp = $available_space;
            $available_space = $max_block_size;

            # Current request was never added to previous block so redo it
            if ($tmp < 0) {
                redo;
            }
        }
    }

    # If there is any information left that hasnt been registered, do so
    if ($available_space != $max_block_size) {
        my $rs = resource_global();
        $rs->register(entries => [@entries]);
        $block_count++;
    }

    $log->comment("Sent ($block_count) requests");

} ## end sub _register_resources

sub _remove_resources {

    my $self     = shift;
    my %incoming = validate_with(
        params => \@_,
        spec   => {
            resource_type => {type => SCALAR},
            filter        => {type => HASHREF}
        },
    );

    my $filters = $incoming{filter};
    $filters->{resource_type} = {$incoming{resource_type} => '='};

    my $rs = resource_global();
    my $found_resources = $rs->get_resources(filter => $filters);

    $log->comment('Removing '
            . @{$found_resources}
            . " exising $incoming{resource_type} resources matching:\n"
            . Dumper($incoming{filter}));

    foreach my $resource (@{$found_resources}) {
        $rs->remove(resources => [$resource]);
    }

} ## end sub _remove_resources

sub _create_discover_lock {

    my $self     = shift;
    my %incoming = validate_with(
        params => \@_,
        spec   => {resource_type => {type => SCALAR}}
    );

    $log->comment("Creating discover lock for $incoming{resource_type}");

    # Log the hash if debug
    my %entry = (
        resource_type => 'discover_lock',
        max_instances => 1,
        attributes    => {lock => $incoming{resource_type}}
    );

    # Get the size of the request hash
    my $bytes = total_size([\%entry]);

    if ($log->may_debug()) {
        $log->comment("Creating register request ($bytes)");
        $log->comment(Dumper(\%entry));
    }

    my $rs = resource_global();
    $rs->register(entries => [\%entry]);

} ## end sub _create_discover_lock

sub _remove_discover_lock {

    my $self     = shift;
    my %incoming = validate_with(
        params => \@_,
        spec   => {resource_type => {type => SCALAR}}
    );

    my $resource_type = $incoming{resource_type};

    $log->comment("Removing discover lock for $resource_type");

    $self->_remove_resources(
        resource_type => 'discover_lock',
        filter        => {lock => {$resource_type => '='}}
    );

} ## end sub _remove_discover_lock

sub get_resources {

    my $self     = shift;
    my %incoming = validate_with(
        params => \@_,
        spec   => {
            resource_type => {type => SCALAR},
            filter        => {type => UNDEF | HASHREF, default => {}},
            attempt_wait  => {type => SCALAR, default => 5},
            max_wait      => {type => SCALAR, default => 120}
        },
    );

    my $rs = resource_global();

    my $elapsed_time = 0;
    if (!defined($incoming{filter}->{resource_type})) {
        $incoming{filter}->{resource_type}
            = {$incoming{resource_type} => '='};
    } ## end

    if ($log->may_debug()) {
        $log->comment("Retrieving $incoming{resource_type} resources "
                . "matching the following filters:\n"
                . Dumper($incoming{filter}));
    } else {
        $log->comment("Retrieving $incoming{resource_type} resources");
    }
    my $resources = $rs->get_resources(filter => $incoming{filter});
    my @array_of_attrs;
    if (@{$resources}) {
        @array_of_attrs = map { $_->attrs() } @{$resources};
    }

    return \@array_of_attrs;
} ## end sub get_resources

sub process_exit_status {
    my ($proc, $node, $object) = @_;
    my $es = NATE::Child::decode_status($proc->exit_status());
    $log->comment("$object discovery exited with $es");
    if (   $es !~ /status 0$/
        && $es =~ /status 25[543]$|signal/) {
        die "Discovery process failed with $es on $node for $object";
    }
} ## end sub process_exit_status

1;
