#
# Copyright (c) 2015 NetApp, Inc., All Rights Reserved.
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#
## @summary Event Worker Library
## @author fujii@netapp.com
## @status shared
## @pod here

package NACL::Service::Events::Worker;

use strict;
use warnings;

use parent    qw(NACL::Service::Worker);
use Data::Dumper;

use Params::Validate qw(validate_with UNDEF OBJECT HASHREF
    SCALAR SCALARREF ARRAYREF BOOLEAN);

use NATE::Events qw(eventloop);
use NATE::Log;
use NACL::GeneralUtils qw(random_name_generator);

use NACL::C::Node;
use NACL::C::Snapshot;
use NACL::CS::Snapmirror;
use NACL::CS::SnapmirrorShowHistory;

use NACL::STask::Lun;
use NACL::STask::Volume;
use NACL::STask::Cluster;
use NACL::STask::NetworkPort;
use NACL::STask::VolumeFile;
use NACL::STask::VolumeClone;
use NACL::STask::VolumeFileClone;
use NACL::STask::StorageFailover;
use NACL::STask::StorageAggregateRelocation;

use NATE::BaseException qw(:try);
use NACL::APISet::Exceptions::ResponseException;
use NACL::STask::Exceptions::UnexpectedSFOState;

my $COLON = q{:};

=head1 NAME

NACL::Service::Events::Worker

=head1 DESCRIPTION

Provides implementations for the Events service.

=cut

=head1 NAME

dummyOp

=head1 DESCRIPTION

Dummy event call. This event does nothing except return.

=cut

sub dummyOp {

    return 1;

}

=head1 NAME

takeover

=head1 DESCRIPTION

Provides the ability to perform a storage failover takeover.

=head1 ATTRIBUTES

=head2 node

(Required) String value - the natehost record of the node that will
perform the takeover for its partner.

=head2 node_partner

(Required) String value - the natehost record of the partner node
that will be taken over

=head2 takeovertype

(Optional) Defaults to 'takeover_command'. See the documentation for
L<NACL::STask::StorageFailover-E<gt>takeover|lib-NACL-STask-StorageFailover-pm/takeover>
command for other possible values for other values for takeovertype.

=head2 timeout

(Optional) Defaults to 600 seconds.

=head2 %opts

Passed directly to the underlying implementation in
L<NACL::STask::StorageFailover-E<gt>takeover|lib-NACL-STask-StorageFailover-pm/takeover>

=head1 EXAMPLE

    C<$client->takeover(
        node            => $nodes[0]->name(),
        node_partner         => $nodes[1]->name(),
        takeovertype    => $takeovertype,
    );>

=cut

sub takeover {

    my $self = shift;

    my %opts = validate_with(
        params      => \@_,
        allow_extra => 1,
        spec        => {
            node         => { type => SCALAR, optional => 0, },
            node_partner      => { type => SCALAR, optional => 0, },
            takeovertype => {
                type     => SCALAR,
                default  => 'takeover_command',
                optional => 1,
            },
            timeout => {
                type     => SCALAR,
                default  => 600,
                optional => 1,
            },
        },
    );

    my $node_name    = $opts{node};
    my $partner_name = $opts{node_partner};
    my $takeovertype = $opts{takeovertype};

    my $node    = NACL::C::Node->new( name => $node_name );
    my $partner = NACL::C::Node->new( name => $partner_name );
    my @cluster
        = NACL::STask::Cluster->find_unique_clusters( nodes => [$node] );

    NACL::STask::StorageFailover->takeover(
        command_interface            => $cluster[0],
        node                         => $node,
        partner                      => $partner,
        takeovertype                 => $takeovertype,
        nacltask_verify              => 1,
        nacltask_check_partner_state => 1,
        'method-timeout'             => $opts{timeout},
    );

    return 1;
}

=head1 NAME

giveback

=head1 DESCRIPTION

Provides the ability to perform a storage failover giveback.

=head1 ATTRIBUTES

=head2 node

(Required) String value - the natehost record of the node that will
perform the giveback to its partner.

=head2 node_partner

(Required) String value - the natehost record of the partner node
that will be givenback control

=head2 require-partner-waiting

(Optional) String value of true|false. Default is true.

=head2 override-vetoes

(Optional) String value of true|false. Default is true.

=head2 timeout

(Optional) Defaults to 300 seconds
Passed directly to the underlying implementation in
L<NACL::STask::StorageFailover-E<gt>giveback|lib-NACL-STask-StorageFailover-pm/giveback>

=head1 EXAMPLE

   C<$client->giveback(
      node            => $nodes[0]->name(),
      node_partner         => $nodes[1]->name(),
   );>

=cut

sub giveback {

    my $self = shift;

    my %opts = validate_with(
        params      => \@_,
        allow_extra => 1,
        spec        => {
            node         => { type => SCALAR, optional => 0, },
            node_partner      => { type => SCALAR, optional => 0, },
            'require-partner-waiting' => {
                type     => SCALAR,
                default  => 'true',
                optional => 1,
            },
            'override-vetoes' => {
                type     => SCALAR,
                default  => 'true',
                optional => 1,
            },
            timeout => {
                type     => SCALAR,
                default  => 300,
                optional => 1,
            },
        },
    );

    my $node_name    = $opts{node};
    my $partner_name = $opts{node_partner};

    my $node    = NACL::C::Node->new( name => $node_name );
    my $partner = NACL::C::Node->new( name => $partner_name );
    my @cluster
        = NACL::STask::Cluster->find_unique_clusters( nodes => [$node] );

    return NACL::STask::StorageFailover->giveback(
        command_interface         => $cluster[0],
        node                      => $node,
        partner                   => $partner,
        nacltask_verify           => 1,
        'require-partner-waiting' => $opts{require_partner_waiting},
        'override-vetoes'         => $opts{'override-vetoes'},
        'method-timeout'          => $opts{timeout},
    );

}


=head1 NAME

TOGB

=head1 DESCRIPTION

Provides the ability to perform a series of storage failover takeover
followed by a giveback

=head1 ATTRIBUTES

=head2 node

(Required) String value - the natehost record of the node that will
perform the takeover and giveback to the partner.

=head2 node_partner

(Required) String value - the natehost record of the partner node
that will be takenover and givenback control

=head2 to_args

(Optional) Hash reference of name-value pairs that are passed to the takeover
call.

=head3 takeovertype

(Optional) Defaults to 'takeover_command'. See the documentation for
L<NACL::STask::StorageFailover-E<gt>takeover|lib-NACL-STask-StorageFailover-pm/takeover>
command for other possible values for other values for takeovertype.

=head3 timeout

(Optional) Defaults to 600 seconds.

=head3 %opts

Passed directly to the underlying implementation in
L<NACL::STask::StorageFailover-E<gt>takeover|lib-NACL-STask-StorageFailover-pm/takeover>

=head2 gb_args

(Optional) Hash reference of name-value pairs that are passed to the giveback
call.

=head3 require-partner-waiting

(Optional) String value of true|false. Default is true.

=head3 override-vetoes

(Optional) String value of true|false. Default is true.

=head3 timeout

(Optional) Defaults to 300 seconds

=head3 %opts

Other name-value pairs are passed directly to the underlying implementation in
L<NACL::STask::StorageFailover-E<gt>giveback|lib-NACL-STask-StorageFailover-pm/giveback>

=head2 iterations

(Optional) Integer value - number of times to execute the takeover/giveback
combination.  Default is 1.

=head2 'iteration-delay'

(Optional*) Integer value - the delay in seconds between consecutive
takeover/giveback calls.  Default is 300 seconds.

=head2 'event-delay'

(Optional*) Integer value - the delay in seconds between takeover
and giveback calls.  Default is 60 seconds.

=cut

sub TOGB {

    my $self = shift;
    my %opts = validate_with(
        params      => \@_,
        allow_extra => 0,
        spec        => {
            node    => { type => SCALAR, optional => 0, },
            node_partner => { type => SCALAR, optional => 0, },
            'iteration-delay' =>
                { type => SCALAR, default => 300, optional => 1, },
            'event-delay' => { type => SCALAR, default => 60, optional => 1, },
            iterations    => { type => SCALAR, default => 1,  optional => 1, },
            to_args => { type => HASHREF, default => {}, optional => 1, },
            gb_args => { type => HASHREF, default => {}, optional => 1, },
        },
    );

    my $iteration_delay = $opts{'iteration-delay'};
    my $event_delay     = $opts{'event-delay'};
    my $iterations      = $opts{iterations};

    my %to_args = %{ $opts{to_args} };
    my %gb_args = %{ $opts{gb_args} };

    delete $opts{'iteration-delay'};
    delete $opts{'event-delay'};
    delete $opts{interval};
    delete $opts{to_args};
    delete $opts{gb_args};

    for ( 1 .. $iterations ) {

        $self->takeover( %opts, %to_args, request_timeout => -1, );
        Tharn::snooze($event_delay);
        $self->giveback( %opts, %gb_args, request_timeout => -1, );
        Tharn::snooze($iteration_delay) unless $_ == 1;

    }

    return 1;
}

=head1 NAME

volumeClone

=head1 DESCRIPTION

Provides the ability to create a set of volume clones. Clone names will be
autogenerated and returned as an array.

=head1 ATTRIBUTES

=head2 node

(Required) String value - the natehost record of the node that the
command will be sent to.

=head2 vserver

(Required) String value of owning vserver for parent volume

=head2 parent-volume

(Required) String value of the parent volume

=head2 quantity

(Optional) Integer value - number of clones to create from
the source volume.  Default is 1.

=head2 %opts

(Optional) Options passed directly to the underlying implementation in
L<NACL::STask::VolumeClone-E<gt>create|lib-NACL-STask-VolumeClone-pm/create>

=head1 RETURNS

Returns an array of clone names.

=head1 EXAMPLE

   C<my @flexclone_names = $client->volumeClone(
      node            => $nodes[0]->name(),
      vserver         => 'vs0',
      'parent-volume' => 'src_vol_3',
      quantity        => 3,
   );>

=cut

sub volumeClone {

    my $self = shift;

    my %opts = validate_with(
        params      => \@_,
        allow_extra => 1,
        spec        => {
            node            => { type => SCALAR, optional => 0, },
            vserver         => { type => SCALAR, optional => 0, },
            'parent-volume' => { type => SCALAR, optional => 0, },
            quantity =>
                { type => SCALAR, default => 1, optional => 1, },
        },
    );

    my $node_name = $opts{node};
    my $quantity  = $opts{quantity};

    my $node = NACL::C::Node->new( name => $node_name );
    my @cluster
        = NACL::STask::Cluster->find_unique_clusters( nodes => [$node] );

    delete $opts{node};
    delete $opts{quantity};

    my @vol_clones;

    for ( 1 .. $quantity ) {
        push(
            @vol_clones,
            NACL::STask::VolumeClone->create(
                command_interface => $cluster[0],
                %opts,
            )->volume()
        );
    }
    return \@vol_clones;
}

=head1 NAME

volumeFileClone

=head1 DESCRIPTION

Provides the ability to perform a volume file (LUN) clone

=head1 ATTRIBUTES

=head2 node

(Required) String value - the natehost record of the node that the
command will be sent to.

=head2 vserver

(Required) String value of owning vserver for parent volume

=head2 volume

(Required) String value of the parent volume name.

=head2 source-path

(Optional) String value - volume file or LUN source path. If not specified,
the service will create a linux ostype 4 megabyte LUN with a random name
to utilize as the source-path LUN

=head2 destination-path

(Optional) String value - volume file or LUN destination path. If not
specified, the service will create a random name as a member of the specified
volume parameter.

=head2 quantity

(Optional) Integer value - The number of clones to create. This option will
only be honored if the destination-path is NOT specified.

=head2 _create_LUN

(Optional) Integer value 0 or 1. Default is 0. If value is 1 and source-path
is not specified, this will create a random LUN. otherwise a random file.

=head2 %opts

(Optional) Options passed directly to the underlying implementation in
L<NACL::stask::VolumeFileClone-E<gt>create|lib-NACL-STask-VolumeFileClone-pm/create>

=head1 RETURNS

Returns an array of strings representing the objects created during this
service call. They are of the form: 'vserver:volume:full_path_to_file.
Example: vs0:src_vol_11:/vol/src_vol_11/lun1

=head1 EXAMPLE

    create 5 clones of a LUN created for you.

    C<$client->volumeFileClone(
        node     => $nodes[0]->name(),
        vserver  => 'vs0',
        volume   => 'src_vol_11',
        quantity => 5,
    );>

    create a clone of the specified LUN.

    C<$client->volumeFileClone(
        node               => $nodes[0]->name(),
        vserver            => 'vs0',
        volume             => 'src_vol_11',
        'source-path'      => '/vol/src_vol_11/lun1',
    );>

    create a clone of the specified LUN with the specified destination name.

    C<$client->volumeFileClone(
        node               => $nodes[0]->name(),
        vserver            => 'vs0',
        volume             => 'src_vol_11',
        'source-path'      => '/vol/src_vol_11/lun1',
        'destination-path' => '/vol/src_vol_11/lun2',
    );>

=cut

sub volumeFileClone {

    my $self = shift;

    my %opts = validate_with(
        params      => \@_,
        allow_extra => 1,
        spec        => {
            node        => { type => SCALAR, optional => 0, },
            vserver     => { type => SCALAR, optional => 0, },
            volume      => { type => SCALAR, optional => 0, },
            _create_LUN => { type => SCALAR, default  => 0, optional => 1, },
            quantity    => { type => SCALAR, default  => 1, optional => 1, },
        },
    );

    my $node_name   = $opts{node};
    my $volume      = $opts{volume};
    my $quantity    = $opts{quantity};
    my $_create_LUN = $opts{_create_LUN};

    my $node = NACL::C::Node->new( name => $node_name );
    my @cluster
        = NACL::STask::Cluster->find_unique_clusters( nodes => [$node] );

    delete $opts{node};
    delete $opts{volume};
    delete $opts{quantity};
    delete $opts{_create_LUN};

    my @volume_files;

    #
    # if no source path is declared, create a zero byte source first.
    #
    if ( !exists( $opts{'source-path'} ) ) {

        my $lun_name = NACL::GeneralUtils::random_name_generator( size => 2, );

        $opts{'source-path'} = "/vol/$volume/$lun_name";

        #
        # return the source path as a courtesy
        # because it was randomly generated.
        #
        push( @volume_files,
                  $opts{vserver}
                . $COLON
                . $volume
                . $COLON
                . $opts{'source-path'} );

        if ( $_create_LUN ) {
            my $lun = NACL::STask::Lun->create(
                command_interface => $cluster[0],
                vserver           => $opts{vserver},
                volume            => $volume,
                lun               => $lun_name,
                size              => '4m',
                ostype            => 'linux',
            );
        } else {
            my $file_obj = NACL::STask::VolumeFile->create(
                command_interface => $cluster[0],
                path              => $opts{'source-path'},
                vserver           => $opts{vserver},
                nacltask_wait     => 1,
            );
        }
    }

    #
    # if no destination path is declared, randomize one.
    #
    if ( !exists( $opts{'destination-path'} ) ) {

        for ( 1 .. $quantity ) {

            my $lun_name = NACL::GeneralUtils::random_name_generator(
                size   => 2,
                prefix => "/vol/$volume/",
            );

            #
            # return the destination paths
            # because they are randomly generated
            #
            push( @volume_files,
                $opts{vserver} . $COLON . $volume . $COLON . $lun_name );

            NACL::STask::VolumeFileClone->create(
                command_interface  => $cluster[0],
                'destination-path' => $lun_name,
                %opts
            );
        }

    } else {

        #
        # return the destination path as a courtesy.
        #
        push( @volume_files,
                  $opts{vserver}
                . $COLON
                . $volume
                . $COLON
                . $opts{'source-path'} );

        NACL::STask::VolumeFileClone->create(
            command_interface => $cluster[0],
            %opts
        );
    }

    return \@volume_files;

}

=head1 NAME

volumeCreate

=head1 DESCRIPTION

Provides the ability to create a volume.

=head1 ATTRIBUTES

=head2 node

(Required) String value - the natehost record of the node that the
command will be sent to. Converted to cluster interface.

=head2 %opts

(Optional) Options passed directly to the underlying implementation in
L<NACL::STask::Volume-E<gt>create|lib-NACL-STask-Volume-pm/create>

=head1 RETURNS

Scalar of the form vserver : volume_name that represents the volume created.

=head1 EXAMPLE

   C<$client->volumeCreate(
      node            => $nodes[0]->name(),
   );>

=cut

sub volumeCreate {

    my $self = shift;
    my %opts = validate_with(
        params      => \@_,
        allow_extra => 1,
        spec        => {
            node    => { type => SCALAR, optional => 0, },
        },
    );

    my $node_name = $opts{node};
    delete $opts{node};
    delete $opts{request_timeout};

    my $node = NACL::C::Node->new( name => $node_name );
    my @c = NACL::STask::Cluster->find_unique_clusters( nodes => [$node] );

    my $vol = NACL::STask::Volume->create(
        command_interface => $c[0],
        %opts
    );

    return $vol->vserver() . $COLON . $vol->volume();

}

=head1 NAME

volumeDelete

=head1 DESCRIPTION

Provides the ability to delete a volume.

=head1 ATTRIBUTES

=head2 node

(Required) String value - the natehost record of the node that the
command will be sent to. Converted to cluster interface.

=head2 vserver

(Required) String value - the vserver name that the volume resides in.

=head2 volume

(Required) String value - the volume name to delete.

=head2 %opts

(Optional) Options passed directly to the underlying implementation in
L<NACL::STask::Volume-E<gt>purge|lib-NACL-STask-Volume-pm/purge>

=head1 EXAMPLE

   C<$client->volumeDelete(
      node            => $nodes[0]->name(),
      vserver         => 'vs0',
   );>

=cut

sub volumeDelete {

    my $self = shift;
    my %opts = validate_with(
        params      => \@_,
        allow_extra => 1,
        spec        => {
            node    => { type => SCALAR, optional => 0, },
            vserver => { type => SCALAR, optional => 0, },
            volume =>  { type => SCALAR, optional => 0, },
        },
    );

    my $node_name = $opts{node};
    delete $opts{node};
    delete $opts{request_timeout};

    my $node = NACL::C::Node->new( name => $node_name );
    my @c = NACL::STask::Cluster->find_unique_clusters( nodes => [$node] );

    NACL::STask::Volume->purge(
        command_interface => $c[0],
        %opts
    );

    return 1;

}

=head1 NAME

volumeCreateDelete

=head1 DESCRIPTION

Provides the ability to create a volume, wait a specificed number of seconds
and then delete the volume.

=head1 ATTRIBUTES

=head2 node

(Required) String value - the natehost record of the node that the
command will be sent to. Converted to cluster interface.

=head2 'event-delay'

(Required) Integer value - the delay in seconds between creation of the volume
and the deletion. Default is 0 seconds.

=head2 iterations

(Optional) Integer value - number of times to execute the volume create/delete
combination.  Default is 1.

=head2 'iteration-delay'

(Optional*) Integer value - the delay in seconds between consecutive
volume create/delete calls.  Default is 0 seconds.

=head1 EXAMPLE

   C<$client->volumeCreateDelete(
      node            => $nodes[0]->name(),
      'event-delay'   => 42,
   );>

=cut

sub volumeCreateDelete {

    my $self = shift;
    my %opts = validate_with(
        params      => \@_,
        allow_extra => 1,
        spec        => {
            node          => { type => SCALAR, optional => 0, },
            'event-delay' => { type => SCALAR, default  => 0, optional => 1, },
            iterations    => { type => SCALAR, default  => 1, optional => 0, },
            'iteration-delay' =>
                { type => SCALAR, default => 0, optional => 1, },
        },
    );

    my $event_delay     = $opts{'event-delay'};
    my $iteration_delay = $opts{'iteration-delay'};
    my $iterations      = $opts{iterations};

    delete $opts{'event-delay'};
    delete $opts{'iteration-delay'};
    delete $opts{iterations};

    for ( 1 .. $iterations ) {

        my @components = split( m/[:]/,
            $self->volumeCreate( %opts, request_timeout => -1, ) );

        Tharn::snooze($event_delay);

        $self->volumeDelete(
            %opts,
            request_timeout => -1,
            volume          => $components[1],
            vserver         => $components[0],
        );

        Tharn::snooze($iteration_delay) unless $_ == 1;

    }

    return 1;

}

=head1 NAME

lunOffline

=head1 DESCRIPTION

Provides the ability to offline a LUN.

=head1 ATTRIBUTES

=head2 node

(Required) String value - the natehost record of the node that will be used to
issue the offline

=head2 path

(Required) String value - the path of the LUN to offline, wildcards accepted.

=head2 vserver

(Optional) String value - the host vserver of the LUN to offline, wildcards
accepted. Defaults to 'vs0'.

=head1 EXAMPLE

   C<$client->lunOffline(
      node    => $nodes[0]->name(),
      path    => '/vol/src_vol_3/DPG_QA_quick_setup_lun*',
      vserver => 'vs*',
   );>

=cut

sub lunOffline {

    my $self = shift;
    my %opts = validate_with(
        params      => \@_,
        allow_extra => 1,
        spec        => {
            node     => { type => SCALAR, optional => 0, },
            path     => { type => SCALAR, optional => 0, },
            vserver  => { type => SCALAR, default  => 'vs0', optional => 1, },
        },
    );

    $opts{state} = 'offline';

    my $node = NACL::C::Node->new( name => $opts{node} );

    NACL::STask::Lun->modify(
        'command_interface' => $node,
        'path'              => $opts{path},
        'vserver'           => $opts{vserver},
        'state'             => 'offline',
    );

    return 1;
}

=head1 NAME

lunOnline

=head1 DESCRIPTION

Provides the ability to online a LUN.

=head1 ATTRIBUTES

=head2 node

(Required) String value - the natehost record of the node that will be used to
issue the online

=head2 path

(Required) String value - the path of the LUN to online, wildcards accepted.

=head2 vserver

(Optional) String value - the host vserver of the LUN to online, wildcards
accepted. Defaults to 'vs0'.

=head1 EXAMPLE

   C<$client->lunOnline(
      node    => $nodes[0]->name(),
      path    => '/vol/src_vol_3/DPG_QA_quick_setup_lun*',
      vserver => 'vs*',
   );>

=cut

sub lunOnline {

    my $self = shift;
    my %opts = validate_with(
        params      => \@_,
        allow_extra => 1,
        spec        => {
            node     => { type => SCALAR, optional => 0, },
            path     => { type => SCALAR, optional => 0, },
            vserver  => { type => SCALAR, default  => 'vs0', optional => 1, },
        },
    );

    my $node = NACL::C::Node->new( name => $opts{node} );

    NACL::STask::Lun->modify(
        'command_interface' => $node,
        'path'              => $opts{path},
        'vserver'           => $opts{vserver},
        'state'             => 'online',
    );

    return 1;
}

=head1 NAME

lunOfflineOnline

=head1 DESCRIPTION

Provides the ability to offline a LUN, wait a specificed number of seconds
and then online the LUN.

=head1 ATTRIBUTES

=head2 node

(Required) String value - the natehost record of the node that will be used to
issue the offline

=head2 path

(Required) String value - the path of the LUN to offline, wildcards accepted.

=head2 'event-delay'

(Required) Integer value - the delay in seconds between offline of the LUN
and the online

=head2 iterations

(Optional) Integer value - number of times to execute the takeover/giveback
combination.  Default is 1.

=head2 'iteration-delay'

(Optional*) Integer value - the delay in seconds between consecutive
offline/online calls.  Default is 300 seconds.

=head2 vserver

(Optional) String value - the host vserver of the LUN to offline, wildcards
accepted. Defaults to 'vs0'.

=head1 EXAMPLE

    C<$client->lunOfflineOnline(
        node     => $nodes[0]->name(),
        path     => '/vol/src_vol_3/DPG_QA_quick_setup_lun*',
        vserver  => 'vs*',
        'event-delay' => 42,
    );>

=cut

sub lunOfflineOnline {

    my $self = shift;

    my %opts = validate_with(
        params      => \@_,
        allow_extra => 1,
        spec => {
            node    => { type => SCALAR, optional => 0, },
            path    => { type => SCALAR, optional => 0, },
            vserver => { type => SCALAR, default  => 'vs0', optional => 1, },
            'event-delay' => { type => SCALAR, optional => 0, },
            iterations    => { type => SCALAR, default  => 1, optional => 0, },
            'iteration-delay' =>
                { type => SCALAR, default => 300, optional => 1, },
            },
    );

    my $iterations      = $opts{iterations};
    my $iteration_delay = $opts{'iteration-delay'};
    my $event_delay     = $opts{'event-delay'};

    delete $opts{interval};
    delete $opts{'iteration-delay'};
    delete $opts{'event-delay'};

    for ( 1 .. $iterations ) {
        $self->lunOffline(
            request_timeout => -1,
            %opts,
        );

        Tharn::snooze( $event_delay );

        $self->lunOnline(
            request_timeout => -1,
            %opts,
        );

        Tharn::snooze( $iteration_delay ) unless $_ == 1;
    }

    return 1;

}

=head1 NAME

snapmirrorShow

=head1 DESCRIPTION

Provides the ability to execute a snapmirror show on a simple schedule.

=head1 ATTRIBUTES

=head2 node

(Required) String value - the natehost record of the node that will be used to
issue the offline

=head2 filter

(Optional) Hash reference - filter options to be passed to snapmirror show

=head2 iterations

(Optional*) Integer value - number of times to execute the snapmirror show
command. Default is 1. Specifying 'period' requires you to specify 'period.'
Not specifying the complete tuple E<lt>iterations,periodE<gt> will throw an
error.

=head2 period

(Optional*) Integer value - the delay in seconds between consecutive snapmirror
show commands. Default is 60 seconds Specifying 'period' requires you to
specify 'iterations.' Not specifying the complete tuple
E<lt>iterations,periodE<gt> will throw an error.

=head1 EXAMPLE

    Exactly one snapmirror show execution.
    C < $client->snapmirrorShow( node => $nodes[0]->name(), );
    >

    Exactly one snapmirror show execution but only list them from vserver 'vso'.
    C < $client->snapmirrorShow(
        node   => $nodes[0]->name(),
        filter => { vserver => 'vs0', },
    );
    >

    25 snapmirror show executions 5 seconds apart.
    C < $client->snapmirrorShow(
        node       => $nodes[0]->name(),
        iterations => 25,
        period     => 5
    );
    >

=cut

sub snapmirrorShow {

    my $self = shift;

    my %opts = validate_with(
        params      => \@_,
        allow_extra => 1,
        spec => {
            node => { type => SCALAR, optional => 0, },
            filter =>
                { type => HASHREF, default => {}, optional => 1, },
            iterations => {
                type     => SCALAR,
                optional => 1,
                default  => 1,
                depends  => ['period'],
            },
            period => {
                type     => SCALAR,
                optional => 1,
                default  => 60,
                depends  => ['iterations'],
            },
            },
    );

    my $iterations = $opts{iterations};
    my $period     = $opts{period};

    delete $opts{iterations};
    delete $opts{period};

    my $node = NACL::C::Node->new( name => $opts{node} );
    my @cluster
        = NACL::STask::Cluster->find_unique_clusters( nodes => [$node] );


    for ( 1 .. $iterations ) {

        NACL::CS::Snapmirror->fetch(
            command_interface => $cluster[0],
            filter            => $opts{filter},
            allow_empty       => 1,
        );

        Tharn::snooze($period) unless $iterations == 1;
    }

    return 1;
}

=head1 NAME

snapmirrorShowHistory

=head1 DESCRIPTION

Provides the ability to execute a snapmirror show history on a simple schedule.

=head1 ATTRIBUTES

=head2 node

(Required) String value - the natehost record of the node that will be used to
issue the offline

=head2 filter

(Optional) Hash reference - filter options to be passed to snapmirror show


=head2 iterations

(Optional*) Integer value - number of times to execute the snapmirror show
command. Default is 1. Specifying 'period' requires you to specify 'period.'
Not specifying the complete tuple E<lt>iterations,periodE<gt> will throw an
error.

=head2 period

(Optional*) Integer value - the delay in seconds between consecutive snapmirror
show commands. Default is 60 seconds Specifying 'period' requires you to
specify 'iterations.' Not specifying the complete tuple
E<lt>iterations,periodE<gt> will throw an error.

=head1 EXAMPLE


    Exactly one snapmirror show history execution.
    C < $client->snapmirrorShowHistory( node => $nodes[0]->name(), );
    >

    Exactly one snapmirror show history execution but only list
    them from vserver 'vso'.
    C < $client->snapmirrorShowHistory(
        node   => $nodes[0]->name(),
        filter => { vserver => 'vs0', },
    );
    >

    25 snapmirror show history executions 5 seconds apart.
    C < $client->snapmirrorShowHistory(
        node       => $nodes[0]->name(),
        iterations => 25,
        period     => 5
    );
    >


=cut


sub snapmirrorShowHistory {

    my $self = shift;

    my %opts = validate_with(
        params      => \@_,
        allow_extra => 1,
        spec => {
            node => { type => SCALAR, optional => 0, },
            filter =>
                { type => HASHREF, default => {}, optional => 1, },
            iterations => {
                type     => SCALAR,
                optional => 1,
                default  => 1,
                depends  => ['period'],
            },
            period => {
                type     => SCALAR,
                optional => 1,
                default  => 60,
                depends  => ['iterations'],
            },
            },
    );

    my $iterations = $opts{iterations};
    my $period     = $opts{period};

    delete $opts{iterations};
    delete $opts{period};

    my $node = NACL::C::Node->new( name => $opts{node} );
    my @cluster
        = NACL::STask::Cluster->find_unique_clusters( nodes => [$node] );

    for ( 1 .. $iterations ) {
        NACL::CS::SnapmirrorShowHistory->fetch(
            command_interface => $cluster[0],
            filter            => $opts{filter},
            allow_empty       => 1,
        );
        Tharn::snooze($period) unless $iterations == 1;
    }

    return 1;
}

=head1 NAME

volOfflineOnline

=head1 DESCRIPTION

Perform a volume offline and wait for a specified amount of time before
onlining the volume.

=head1 ATTRIBUTES

=head2 node_list

(Required) Comma separated string value - the names of the nodes that
make up the cluster where the volume to be operated on resides

=head2 vol_list

(Required) Array (of strings) reference - A reference to the array that contains
the names of the volumes to perform the offline/online upon.

=head2 vol_pattern

(Optional) String value - A regular expression pattern that will be used to
determine which of the volume names will be operated on. In other words,
this function will only operate on volumes in vol_list that match this
regular expression pattern.

=head2 iterations

(Optional) scalar - The number of times to perform a volume offline
and volume online on all the volumes. Defaults to 1.  Use this
parameter if you know exactly how many volume offline/onlines you'd
like to do.

=head2 runlength

(Optional) scalar - The number of seconds to run this volume offline/
online operation in a loop.  This parameter overrides iterations
and has no default.  Use this parameter if you have a time duration
that you'd like to run continuous volume offline/onlines on a set of
volumes.

=head2 'event-delay'

(Optional) scalar - The number of seconds to wait after offlining the
volume before onlining it.  Defaults to 30.

=head2 'iteration-delay'

(Optional) scalar - The number of seconds to wait between volume
offline/online operations if iterations is greater than 1 or if
runlength is specified.  Defaults to 60.


=head1 EXAMPLE

 simple usage
 C<$client->volOfflineOnline(
        node_list => "dpg-3270-node1,dpg-3270-node1",
        vol_list => [ "test_vol1" ],
 );>

 set iterations and pattern matching
 C<$client->volOfflineOnline(
        node_list => "dpg-3270-node1,dpg-3270-node1",
        vol_list => [ "test_vol1", "super_important_online_vol", "test_vol2" ],
        vol_pattern => 'test',
        iterations => 10, #run 10 times on each volume
        'event-delay' => 1, #only leave each volume offline for 1s
        'iteration-delay' => 10, #only wait 10s before starting the next cycle
 );>

 runlength based
 C<$client->volOfflineOnline(
        node_list => "dpg-3270-node1,dpg-3270-node1",
        vol_list => [ "test_vol1", "super_important_online_vol", "test_vol2" ],
        vol_pattern => 'test',
        runlength => 3600, #run in a loop for 1hr
        'event-delay' => 1, #only leave each volume offline for 1s
        'iteration-delay' => 10, #only wait 10s before starting the next cycle
 );>

=cut

sub volOfflineOnline {
    my $self = shift;

    my %opts = validate_with(
        params      => \@_,
        allow_extra => 1,
        spec        => {
            node_list       => { type => SCALAR,   optional => 0, },
            vol_list    => { type => ARRAYREF, optional => 0, },
            vol_pattern => { type => SCALAR,   optional => 1, },
            iterations  => { type => SCALAR,   default  => 1, optional => 1, },
            runlength   => { type => SCALAR,   optional => 1, },
            'iteration-delay' =>
                { type => SCALAR, default => 60, optional => 1, },
            'event-delay' => { type => SCALAR, default => 30, optional => 1, },
        },
    );

    my @cluster = $self->_getClusters( nodes => $opts{node_list}, );

    my $vol_list_aref = $opts{vol_list};

    my @vol_objs = _create_filtered_vol_objs_array(
        vol_list          => $opts{vol_list},
        vol_pattern       => $opts{vol_pattern},
        command_interface => $cluster[0],
    );

    #
    # by default vol will be offline for 30s
    #
    my $event_delay = $opts{'event-delay'};

    #
    # time between running events, if running more than one
    #
    my $iteration_delay = $opts{ 'iteration-delay' };

    my $iterations = $opts{iterations};

    my $runlength;
    if ( $opts{runlength} ) {
        $runlength = $opts{runlength};

        # runlength overrides iterations, so set cycle
        # count to something really high.
        $iterations = $runlength;
    } else {

        # if no runlength is provided, let it run long
        # enough to finish one complete cycle
        $runlength = $iteration_delay;
    }

    my $start_time = time();
    while ( $iterations > 0 && ( ( $runlength + $start_time ) > time() ) ) {
        foreach my $vol_obj (@vol_objs) {
            if ( $vol_obj->get_one_state_attribute('junction-active')
                =~ /true/ )
            {
                #
                # if mounted, unmount
                # store junction-path for later mounting
                #
                $vol_obj->{_old_junction_path}
                    = $vol_obj->get_one_state_attribute('junction-path');
                $vol_obj->unmount();
            }
            $vol_obj->offline();
        }

        Tharn::snooze($event_delay);

        foreach my $vol_obj (@vol_objs) {
            $vol_obj->online();
            if ( $vol_obj->{_old_junction_path} ) {

                #ONTAP call is no-op, if already mounted
                $vol_obj->mount(
                    'junction-path' => $vol_obj->{_old_junction_path} );
            }
        }

        if ( $iterations > 1 ) {
            Tharn::snooze($iteration_delay);
        }
        $iterations = $iterations - 1;
    }

    return 1;
}


=head1 NAME

volOffline

=head1 DESCRIPTION

Perform a volume offline

=head1 ATTRIBUTES

=head2 node_list

(Required) String value - the names of the nodes that
make up the cluster where the volume to be operated on resides

=head2 vol_list

(Required) Array (of strings) reference - A reference to the array that contains
the names of the volumes to perform the offline/online upon.


=head1 EXAMPLE

 simple usage
 C<$client->volOffline(
        node => "dpg-3270-node1",
        vol_list => [ "test_vol1" ],
 );>

=cut

=head1 NAME

volOnline

=head1 DESCRIPTION

Perform a volume offline

=head1 ATTRIBUTES

=head2 node

(Required) String value - the names of the nodes that
make up the cluster where the volume to be operated on resides

=head2 vol_list

(Required) Array (of strings) reference - A reference to the array that contains
the names of the volumes to perform the offline/online upon.


=head1 EXAMPLE

 simple usage
 C<$client->volOffline(
        node => "dpg-3270-node1",
        vol_list => [ "test_vol1" ],
 );>

=cut

sub volOffline {
    my $self = shift;

    my %opts = validate_with(
        params      => \@_,
        allow_extra => 1,
        spec        => {
            node     => { type => SCALAR,   optional => 0, },
            vol_list => { type => ARRAYREF, optional => 0, },
        }
    );

    my @cluster = $self->_getClusters( nodes => $opts{node}, );

    foreach my $vol_name ( @{ $opts{vol_list} } ) {

        my $vol_obj = NACL::STask::Volume->find(
            command_interface => $cluster[0],
            filter            => { volume => $vol_name },
        );

        if ( $vol_obj->get_one_state_attribute('junction-active') =~ /true/ ) {
            $vol_obj->unmount();
        }
        $vol_obj->offline();
    }

    return 1;
}

=head1 NAME

volOnline

=head1 DESCRIPTION

Perform a volume online

=head1 ATTRIBUTES

=head2 node

(Required) String value - the names of the nodes that
make up the cluster where the volume to be operated on resides

=head2 volname

(Required) Strings - volume name to bring online

=head2 junction-path

(Required) String - junction path to mount the volume

=head1 EXAMPLE

 simple usage
 C<$client->volOnline(
        node => "dpg-3270-node1",
        volname => "test_vol1",
        'junction-path' => "/",
 );>

=cut

sub volOnline {
    my $self = shift;

    my %opts = validate_with(
        params      => \@_,
        allow_extra => 1,
        spec        => {
            node            => { type => SCALAR, optional => 0, },
            volname         => { type => SCALAR, optional => 0, },
            'junction-path' => { type => SCALAR, optional => 0, },
        }
    );

    my @cluster = $self->_getClusters( nodes => $opts{node}, );

    my $vol_obj = NACL::STask::Volume->find(
        command_interface => $cluster[0],
        filter            => { volume => $opts{volname} },
    );

    $vol_obj->online();
    $vol_obj->mount( 'junction-path' => $opts{'junction-path'} );

    return 1;
}

=head1 NAME

interclusterFailure

=head1 DESCRIPTION

Simulate an intercluster network failure by downing the intercluster lifs
and then bringing them back up after a specified time interval

=head1 ATTRIBUTES

=head2 node_list

(Required) Comma separated string value - the names of the nodes that
make up the cluster where the lifs to be operated on reside

=head2 iterations

(Optional) scalar - The number of times to perform an intercluster
network failure. Defaults to 1.  Use this parameter if you know
exactly how many incluster network failures you'd like to do.

=head2 'iteration-delay'

(Optional) scalar - The number of seconds to wait between intercluster
failure operations if iterations is greater than 1 or if
runlength is specified.  Defaults to 60.

=head2 runlength

(Optional) scalar - The number of seconds to run this intercluster
failure operation in a loop.  This parameter overrides iterations
and has no default.  Use this parameter if you have a time duration
that you'd like to run continuous intercluster network toggles

=head2 'event-delay'

(Optional) scalar - The number of seconds to wait after downing the lifs
before bringing them back up.  Defaults to 30.

=head1 EXAMPLE

 simple usage
 C<$client->interclusterFailure(
        node_list => "dpg-3270-node1,dpg-3270-node1",
 );>

 set iterations
 C<$client->interclusterFailure(
        node_list => "dpg-3270-node1,dpg-3270-node1",
        iterations => 10, #run 10 times
        'event-delay' => 1, #only leave the lifs down for 1s
        'iteration-delay' => 10, #only wait 10s before starting the next cycle
 );>

 runlength based
 C<$client->interclusterFailure(
        node_list => "dpg-3270-node1,dpg-3270-node1",
        runlength => 3600, #run in a loop for 1hr
 );>

=cut

sub interclusterFailure {

    my $self = shift;

    my %opts = validate_with(
        params      => \@_,
        allow_extra => 1,
        spec        => {
            node_list      => { type => SCALAR, optional => 0, },
            iterations => { type => SCALAR, default  => 1, optional => 1, },
            'iteration-delay' =>
                { type => SCALAR, default => 60, optional => 1, },
            runlength => { type => SCALAR, optional => 1, },
            'event-delay' => { type => SCALAR, default => 30, optional => 1, },
        },
    );

    my $event_delay = $opts{'event-delay'};
    my $iteration_delay = $opts{'iteration-delay'};

    #time between running events, if running more than one
    my $iterations = $opts{iterations};

    my $runlength;
    if ( $opts{runlength} ) {
        $runlength   = $opts{runlength};
        # runlength overrides iterations, so
        # set cycle count to something really high.
        $iterations = $runlength;
    } else {
        # if no runlength is provided, let it run
        # long enough to finish one complete cycle
        $runlength = $iteration_delay;
    }

    my @cluster = $self->_getClusters(
        nodes => $opts{node_list},
    );

    my $start_time = time();
    while ( $iterations > 0 && ( ( $runlength + $start_time ) > time() ) ) {
        my @lif_objs = NACL::STask::NetworkInterface->find(
            command_interface => $cluster[0],
            filter            => { role => 'intercluster', }
        );

        foreach my $lif_obj (@lif_objs) {
            $lif_obj->modify( 'status-admin' => 'down' );
        }

        Tharn::snooze( $event_delay );

        foreach my $lif_obj (@lif_objs) {
            $lif_obj->modify( 'status-admin' => 'up' );
        }

        if ( $iterations > 1 ) {
            Tharn::snooze( $iteration_delay );
        }
        $iterations = $iterations - 1;
    }

    return 1;
}

=head1 NAME

volMove

=head1 DESCRIPTION

Perform a volume move to specified and wait for a specified
amount of time before moving the volume back to the original
aggregate.

=head1 ATTRIBUTES

=head2 node_list

(Required) Comma separated string value - the names of the nodes that
make up the cluster where the volume(s) to be operated on reside.

=head2 vol_list

(Required) Array (of strings) reference - A reference to the array that contains
the names of the volumes to perform the vol move upon.

=head2 vol_pattern

(Optional) String value - A regular expression pattern that will be used to
determine which of the volume names will be operated on. In other words,
this function will only operate on volumes in vol_list that match this
regular expression pattern.

=head2 dst_aggr

(Optional) String value - The name of the aggregate that you'd like the
volumes to move to.  If not specified NACL will pick the first eligible
aggregate in the list returned from the "volume move target-aggr" command.

=head2 iterations

(Optional) scalar - The number of times to perform a vol move and
a vol move back on all the volumes. Defaults to 1.  Use this
parameter if you know exactly how many volume moves you'd
like to do.

=head2 'iteration-delay'

(Optional) scalar - The number of seconds to wait between volume
move operations if iterations is greater than 1 or if
runlength is specified.  Defaults to 60.

=head2 runlength

(Optional) scalar - The number of seconds to run this volume move
operation in a loop.  This parameter overrides iterations
and has no default.  Use this parameter if you have a time duration
that you'd like to run continuous volume move on a set of
volumes.

=head2 'event-delay'

(Optional) scalar - The number of seconds to wait after moving the
volume before moving it back to it's original aggr.  Defaults to 30.

=head1 EXAMPLE

 #simple
 C<$client->volMove(
        node_list => "dpg-3270-node1,dpg-3270-node1",
        vol_list => [ "test_vol1" ],
 );>

 #set iterations and pattern matching
 C<$client->volMove(
        node_list => "dpg-3270-node1,dpg-3270-node1",
        vol_list => [ "test_vol1", "super_important_never_move_vol" ],
        vol_pattern => 'test',
        iterations => 10, #run 10 times on each volume
        vol_return_wait_time => 1,
        'iteration-delay' => 10, #only wait 10s before starting the next cycle
 );>

 #runlength based
 C<$client->volMove(
        node_list => "dpg-3270-node1,dpg-3270-node1",
        vol_list => [ "test_vol1" ],
        runlength => 3600, #run in a loop for 1hr
 );>

=cut

sub volMove {
    my $self = shift;

    my %opts = validate_with(
        params      => \@_,
        allow_extra => 1,
        spec        => {
            node_list       => { type => SCALAR,   optional => 0, },
            vol_list    => { type => ARRAYREF, optional => 0, },
            vol_pattern => { type => SCALAR,   optional => 1, },
            dst_aggr    => { type => SCALAR,   optional => 1, },
            iterations  => { type => SCALAR,   default  => 1, optional => 1, },
            'iteration-delay' =>
                { type => SCALAR, default => 60, optional => 1, },
            runlength => { type => SCALAR, optional => 1, },
            'event-delay' => { type => SCALAR, default => 30, optional => 1, },
        },
    );

    my $event_delay       = $opts{'event-delay'};
    my $inter_event_delay = $opts{inter_event_delay};
    my $iterations        = $opts{iterations};
    my $iteration_delay   = $opts{'iteration-delay'};
    my $runlength;

    if ( $opts{runlength} ) {
        $runlength   = $opts{runlength};

        # runlength overrides iterations, so set cycle
        # count to something really high.
        $iterations = $runlength;
    } else {
        # if no runlength is provided, set it to really long
        # to give enough time for all the cycles to complete
        $runlength = time();
    }

    my @cluster = $self->_getClusters( nodes => $opts{node_list}, );

    my @vol_objs = _create_filtered_vol_objs_array(
        command_interface => $cluster[0],
        vol_list          => $opts{vol_list},
        vol_pattern       => $opts{vol_pattern},
    );

    my %vol_move_opts;

    if ( $opts{dst_aggr} ) {

        # dst_aggr is optional as NACL vol move doesn't
        # need destination-aggregate

        # if no destination-aggregate, one is automatically picked by
        # using NACL::CS::VolumeMoveTargetAggr
        $vol_move_opts{'destination-aggregate'} = $opts{dst_aggr};
    }
    my $start_time = time();
    my $log = NATE::Log::log_global();
    while ( $iterations > 0 && ( ( $runlength + $start_time ) > time() ) ) {
        foreach my $vol_obj (@vol_objs) {
            my $original_aggr = $vol_obj->get_one_state_attribute('aggregate');
            $vol_obj->{_original_aggr} = $original_aggr;   #store original aggr
            my $vol_name = $vol_obj->volume;
            try {
                $vol_obj->move_start( %vol_move_opts );
            }
            catch NATE::BaseException with {
                $log->comment(Dumper(shift));
            };
        }
        Tharn::snooze( $event_delay );

        foreach my $vol_obj (@vol_objs) {
            my $vol_name      = $vol_obj->volume;
            my $original_aggr = $vol_obj->{_original_aggr};
            try {
                $vol_obj->move_start( 'destination-aggregate' => $original_aggr, );
            }
            catch NATE::BaseException with {
                $log->comment(Dumper(shift));
            };
        }

        if ( $iterations > 1 ) {
            Tharn::snooze( $iteration_delay );
        }
        $iterations = $iterations - 1;
    }

    return 1;
}

=head1 NAME

reboot

=head1 DESCRIPTION

Provides the ability to perform a node reboot

=head1 ATTRIBUTES

=head2 node

(Required) String value - the natehost record of the node that will
be rebooted.

=head1 EXAMPLE

 C<$client->reboot(
       node            => $nodes[0]->name(),
 );>

=cut

sub reboot {

    my $self = shift;

    my %opts = validate_with(
        params      => \@_,
        allow_extra => 1,
        spec        => { node => { type => SCALAR, optional => 0, }, },
    );

    my $node    = NACL::C::Node->new( name => $opts{node} );

    return $node->reboot();

}

=head1 NAME

ARL

=head1 DESCRIPTION

Provides the ability to perform an aggregrate relocation.

Verifies that the aggregate relocation has successfully completed.

=head1 ATTRIBUTES

=head2 node

(Required) String value - the natehost record of the node the
aggregate currently resides on.

=head2 node_dest

(Required) String value - the natehost record of the node the
aggregate is being moved to.

=head2 aggr_list

(Required) Array of strings of aggregate names to be relocated from the source
node to the destination node.

=head1 EXAMPLE

 C<$client->ARL(
         node => $nodes[1]->name(),
         node_dest => $nodes[0]->name(),
         aggr_list => \@aggr_names,
 );>

=cut

sub ARL {

    my $self = shift;

    my %opts = validate_with(
        params      => \@_,
        allow_extra => 1,
        spec        => {
            node      => { type => SCALAR,   optional => 0, },
            node_dest      => { type => SCALAR,   optional => 0, },
            aggr_list => { type => ARRAYREF, optional => 0, },
            'iteration-delay' =>
                { type => SCALAR, default => 300, optional => 1, },
            'event-delay' => { type => SCALAR, default => 60, optional => 1, },
            iterations    => { type => SCALAR, default => 1,  optional => 1, },
        },
    );

    my $iteration_delay = delete $opts{'iteration-delay'};
    my $event_delay     = delete $opts{'event-delay'};
    my $iterations      = delete $opts{iterations};

    my $node = NACL::C::Node->new( name => $opts{node});
    my $dest = NACL::C::Node->new( name => $opts{node_dest});
    my @cluster
        = NACL::STask::Cluster->find_unique_clusters( nodes => [$node] );

    while ( $iterations > 0 ) {
        NACL::STask::StorageAggregateRelocation->start(
            command_interface => $cluster[0],
            node => $node,
            destination =>$dest,
            'aggregate-list' => $opts{aggr_list},
        );
        Tharn::snooze($event_delay);
        NACL::STask::StorageAggregateRelocation->start(
            command_interface => $cluster[0],
            node => $dest,
            destination =>$node,
            'aggregate-list' => $opts{aggr_list},
        );
        if ( $iterations > 1 ) {
            Tharn::snooze( $iteration_delay );
        }
        $iterations--;
    }
#$log->reset_path();
    return 1;
}

=head1 NAME

snapshotCreateDelete

=head1 DESCRIPTION

Provides the ability to create a snapshot, wait a specificed number of seconds
and then delete the snapshot.

=head1 ATTRIBUTES

=head2 node

(Required) String value - the natehost record of the node that the
command will be sent to. Converted to cluster interface.

=head2 vserver

(Required) String value - the name of the vserver that has access to the
target volume that the snapshot will be created on.

=head2 volume

(Required) String value - the name of the volume that
the snapshot will be created on.

=head2 snapshot

(Optional) String value - the name of the snapshot that be created.
If not specified, a random name will be chosen.

=head2 'event-delay'

(Required) Integer value - the delay in seconds between creation of the
snapshot and the deletion. Default is 0 seconds.

=head2 iterations

(Optional) Integer value - number of times to execute the snapshot
create/delete combination.  Default is 1.

=head2 'iteration-delay'

(Optional*) Integer value - the delay in seconds between consecutive
snapshot create/delete calls.  Default is 0 seconds.

=head1 EXAMPLE

    C<$client->snapshotCreateDelete(
        node              =E<gt> $nodes[0]->name(),
        vserver           =E<gt> 'vs0',
        volume            =E<gt> 'myvolume',
        'event-delay'     =E<gt> 42,
        iterations        =E<gt> 5,
        'iteration-delay' =E<gt> 60,
    );>

=cut

sub snapshotCreateDelete {

    my $self = shift;
    my %opts = validate_with(
        params      => \@_,
        allow_extra => 1,
        spec        => {
            node          => { type => SCALAR, optional => 0, },
            vserver       => { type => SCALAR, optional => 0, },
            volume        => { type => SCALAR, optional => 0, },
            snapshot      => { type => SCALAR, optional => 1, },
            'event-delay' => { type => SCALAR, default  => 0, optional => 1, },
            iterations    => { type => SCALAR, default  => 1, optional => 0, },
            'iteration-delay' =>
                { type => SCALAR, default => 0, optional => 1, },
        },
    );

    my $event_delay     = $opts{'event-delay'};
    my $iteration_delay = $opts{'iteration-delay'};
    my $iterations      = $opts{iterations};

    delete $opts{'event-delay'};
    delete $opts{'iteration-delay'};
    delete $opts{iterations};

    my $node = NACL::C::Node->new( name => $opts{node});
    my @cluster
        = NACL::STask::Cluster->find_unique_clusters( nodes => [$node] );

    for ( 1 .. $iterations ) {

        my $snapshot_name;
        if ( !exists( $opts{'snapshot'} ) ) {
            $snapshot_name
                = NACL::GeneralUtils::random_name_generator( size => 2, );
        } else {
            $snapshot_name = $opts{snapshot};
        }

        my $snapshot = NACL::C::Snapshot->create(
            command_interface => $cluster[0],
            vserver           => $opts{vserver},
            volume            => $opts{volume},
            snapshot          => $snapshot_name,
        );

        Tharn::snooze($event_delay);

        NACL::C::Snapshot->delete(
            command_interface => $cluster[0],
            vserver           => $opts{vserver},
            volume            => $opts{volume},
            snapshot          => $snapshot_name,
        );

        Tharn::snooze($iteration_delay) unless $_ == 1;

    }

    return 1;

}

=head1 NAME

sfcod

=head1 DESCRIPTION

Provides the ability to perform an single file copy on demand (SFCOD)

=head1 ATTRIBUTES

=head2 node

(Required) String value - the natehost record of the node the
aggregate currently resides on.

=head2 quantity

(Optional) Integer value - The number of files to create.
Specifying quantity will require you to specify the following:
    - 'source-vserver'
    - 'source-volume'
    - 'destination-vserver'
    - 'destination-volume'
    - 'size'

=head2 'source-vserver'

(Optional*) String value = The name of the source vserver.
This parameter is mandatory if you use the 'quantity' option.

=head2 'source-volume'

(Optional*) String value = The name of the source volume.
This parameter is mandatory if you use the 'quantity' option.

=head2 'destination-vserver'

(Optional*) String value = The name of the destination vserver.
This parameter is mandatory if you use the 'quantity' option.

=head2 'destination-volume'

(Optional*) String value = The name of the destination volume.
This parameter is mandatory if you use the 'quantity' option.

=head2 size

(Optional*) Integer value = The size of the files to create
on the source and copied to the destination.
This parameter is mandatory if you use the 'quantity' option.

=head2 'source-paths'

(Required) Array reference of vserver:volume/paths

=head2 'destination-paths'

(Required) Array reference of vserver:volume/paths

=head1 EXAMPLE

 C<$client->sfcod(
    node                => $nodes[1]->name(),
    'source-paths'      => \@aggr_names,
    'destination-paths' => \@aggr_names,
 );>

=cut

sub sfcod {

    my $self = shift;

    my %opts = validate_with(
        params      => \@_,
        allow_extra => 1,
        spec => {
            node     => { type => SCALAR, optional => 0, },
            quantity => {
                type     => SCALAR,
                default  => 1,
                optional => 1,
                depends  => [
                    'source-vserver',      'source-volume',
                    'destination-vserver', 'destination-volume',
                ],
            },
            'source-vserver'      => { type => SCALAR, optional => 1, },
            'source-volume'       => { type => SCALAR, optional => 1, },
            'destination-vserver' => { type => SCALAR, optional => 1, },
            'destination-volume'  => { type => SCALAR, optional => 1, },
            'size' => { type => SCALAR, default => 4096, optional => 1, },
            'source-paths' =>
                { type => ARRAYREF, default => [], optional => 1, },
            'destination-paths' =>
                { type => ARRAYREF, default => [], optional => 1, },
            },
    );

    my $node = NACL::C::Node->new( name => $opts{node});
    my @volume_files;
    my @cluster
        = NACL::STask::Cluster->find_unique_clusters( nodes => [$node] );

    my $source_paths      = $opts{'source-paths'};
    my $destination_paths = $opts{'destination-paths'};

    if ( ! @{ $source_paths } ) {

        for ( 1 .. $opts{quantity} ) {

            my $filename
                = NACL::GeneralUtils::random_name_generator( size => 2, );

            push(
                $source_paths,
                $opts{'source-vserver'} . $COLON . "$opts{'source-volume'}/$filename"
            );
                #$opts{'source-vserver'} . $COLON . "/vol/$opts{'source-volume'}/$filename"

            my $file_obj = NACL::STask::VolumeFile->create(
                command_interface => $cluster[0],
                path              => "/vol/$opts{'source-volume'}/$filename",
                vserver           => $opts{'source-vserver'},
                nacltask_wait     => 1,
                size              => $opts{size},
            );

        }
    }

    if ( ! @{ $destination_paths } ) {

        for ( 1 .. $opts{quantity} ) {

            my $filename
                = NACL::GeneralUtils::random_name_generator( size => 2, );

            push(
                $destination_paths,
                $opts{'destination-vserver'}
                    . $COLON
                    . "$opts{'destination-volume'}/$filename"
                    #. "/vol/$opts{'destination-volume'}/$filename"
            );

        }

    }

    NACL::STask::VolumeFileCopy->start(
        command_interface   => $cluster[0],
        'source-paths'      => $opts{'source-paths'},
        'destination-paths' => $opts{'destination-paths'},
        'report-completed'  => 'true',
        'nacltask_wait'     => 1,
    );

    return 1;

}

=head1 NAME

sfmod

=head1 DESCRIPTION

Provides the ability to perform an single file modify on demand(SFMOD)

=head2 quantity

(Optional) Integer value - The number of files to create.
Specifying quantity will require you to specify the following:
    - 'source-vserver'
    - 'source-volume'
    - 'destination-vserver'
    - 'destination-volume'

=head2 'source-vserver'

(Optional*) String value = The name of the source vserver.
This parameter is mandatory if you use the 'quantity' option.

=head2 'source-volume'

(Optional*) String value = The name of the source volume.
This parameter is mandatory if you use the 'quantity' option.

=head2 'destination-vserver'

(Optional*) String value = The name of the destination vserver.
This parameter is mandatory if you use the 'quantity' option.

=head2 'destination-volume'

(Optional*) String value = The name of the destination volume.
This parameter is mandatory if you use the 'quantity' option.

=head2 size

(Optional*) Integer value = The size of the files to create
on the source and copied to the destination.
The default value is 4096 bytes.

=head2 'source-paths'

(Optional) Array reference of vserver:volume/paths. If not
supplied - random names will be used to create sources.
You must supply the source-vserver and destination-volume
if you are not supplying this argument. Note that the
format for this is vserver:volume/path and not the usual
format of vserver:/vol/volname/path.

=head2 'destination-paths'

(Optional) Array reference of vserver:volume/paths. If not
supplied, random names will be created for destinations.
You must supply the destination-vserver and
destination-volume if you are not supplying this argument.
Note that the format for this is vserver:volume/path
and not the usual format of vserver:/vol/volname/path.

=head1 EXAMPLE

 C<$client->sfmod(
    node                => $nodes[1]->name(),
    'source-paths'      => \@aggr_names,
    'destination-paths' => \@aggr_names,
 );>

=cut

sub sfmod {

    my $self = shift;

    my %opts = validate_with(
        params      => \@_,
        allow_extra => 1,
        spec => {
            node     => { type => SCALAR, optional => 0, },
            quantity => {
                type     => SCALAR,
                default  => 1,
                optional => 1,
                depends  => [
                    'source-vserver',      'source-volume',
                    'destination-vserver', 'destination-volume',
                ],
            },
            'source-vserver'      => { type => SCALAR, optional => 1, },
            'source-volume'       => { type => SCALAR, optional => 1, },
            'destination-vserver' => { type => SCALAR, optional => 1, },
            'destination-volume'  => { type => SCALAR, optional => 1, },
            'size' => { type => SCALAR, default => 4096, optional => 1, },
            'source-paths' =>
                { type => ARRAYREF, default => [], optional => 1, },
            'destination-paths' =>
                { type => ARRAYREF, default => [], optional => 1, },
            },
    );

    my $node = NACL::C::Node->new( name => $opts{node});
    my @volume_files;
    my @cluster
        = NACL::STask::Cluster->find_unique_clusters( nodes => [$node] );

    my $source_paths      = $opts{'source-paths'};
    my $destination_paths = $opts{'destination-paths'};

    if ( ! @{ $source_paths } ) {

        for ( 1 .. $opts{quantity} ) {

            my $filename
                = NACL::GeneralUtils::random_name_generator( size => 2, );

            push(
                $source_paths,
                $opts{'source-vserver'} . $COLON . "$opts{'source-volume'}/$filename"
            );
                #$opts{'source-vserver'} . $COLON . "/vol/$opts{'source-volume'}/$filename"

            my $file_obj = NACL::STask::VolumeFile->create(
                command_interface => $cluster[0],
                path              => "/vol/$opts{'source-volume'}/$filename",
                vserver           => $opts{'source-vserver'},
                nacltask_wait     => 1,
                size              => $opts{size},
            );

        }
    }

    if ( ! @{ $destination_paths } ) {

        for ( 1 .. $opts{quantity} ) {

            my $filename
                = NACL::GeneralUtils::random_name_generator( size => 2, );

            push(
                $destination_paths,
                $opts{'destination-vserver'}
                    . $COLON
                    . "$opts{'destination-volume'}/$filename"
                    #. "/vol/$opts{'destination-volume'}/$filename"
            );

        }

    }

    NACL::STask::VolumeFileMove->start(
        command_interface   => $cluster[0],
        'source-paths'      => $source_paths,
        'destination-paths' => $destination_paths,
        'report-completed'  => 'true',
        'nacltask_wait'     => 1,
    );

    return 1;

}

=head1 NAME

portToggle

=head1 DESCRIPTION

Simulate a port failure by bringing the port down administratively
and then bringing it back up after a specified time interval.

=head1 ATTRIBUTES

=head2 node

(Required) String value - the natehost record of the node that owns
the physical network port.

=head2 ports

(Required) Array reference. An array reference composed of individual
ports. e.g [ 'e0c', 'e0d' ].

=head2 iterations

(Optional) scalar - The number of times to perform the port toggle.
Defaults to 1.

=head2 'iteration-delay'

(Optional) scalar - The number of seconds to wait between port
failure/recovery operations. Defaults to 60.

=head2 'event-delay'

(Optional) scalar - The number of seconds to wait after downing the ports
before bringing them back up.  Defaults to 30.

=head1 EXAMPLE

 C<$client->portToggle(
        node => 'dpg-3270-node1',
        ports => [ 'e3c', 'e3d' ],
        iterations => 10,
        'event-delay' => 1,
        'iteration-delay' => 10,
 );>


=cut

sub portToggle {

    my $self = shift;

    my %opts = validate_with(
        params      => \@_,
        allow_extra => 1,
        spec        => {
            node       => { type => SCALAR,   optional => 0, },
            ports      => { type => ARRAYREF, optional => 0, },
            iterations => { type => SCALAR,   default  => 1, optional => 1, },
            'iteration-delay' =>
                { type => SCALAR, default => 60, optional => 1, },
            'event-delay' => { type => SCALAR, default => 30, optional => 1, },
        },
    );

    my $event_delay     = $opts{'event-delay'};
    my $iteration_delay = $opts{'iteration-delay'};

    #time between running events, if running more than one
    my $iterations = $opts{iterations};

    my $node = NACL::C::Node->new( name => $opts{node} );
    my @cluster
        = NACL::STask::Cluster->find_unique_clusters( nodes => [$node] );

    for ( 1 .. $iterations ) {

        foreach my $port ( @{ $opts{ports} } ) {

            NACL::C::NetworkPort->modify(
                command_interface => $cluster[0],
                node              => $node,
                port              => $port,
                'up-admin'        => 'false',
            );

        }
        Tharn::snooze($event_delay);

        foreach my $port ( @{ $opts{ports} } ) {

            NACL::C::NetworkPort->modify(
                command_interface => $cluster[0],
                node              => $node,
                port              => $port,
                'up-admin'        => 'true',
            );

        }

        Tharn::snooze($iteration_delay) unless $_ == 1;

    }

    return 1;
}



=head1 NAME

_create_filtered_vol_objs_array

=head1 DESCRIPTION

Accepts:
 1. array reference of volume names.
 2. (Optional) string to pattern match against.
 3. command interface to use.

Returns:
 1. An array of NACL::STask::Volume volume objects

=head2 vol_list

(Required) An array reference of volume names.

=head2 vol_pattern

(Optional) A string to use for regular expression matching. If
not provided, no filtering will be done.

=head2 command_interface

(Required) The command interface to use to find the volume objs.

=cut

sub _create_filtered_vol_objs_array {
    #my $self = shift;
    my %opts = validate_with(
        params      => \@_,
        allow_extra => 1,
        spec        => {
            vol_list          => { type => ARRAYREF,        optional => 0, },
            command_interface => { type => OBJECT | SCALAR, optional => 0, },
            vol_pattern       => { type => SCALAR | UNDEF,  optional => 1, },
        },
    );

    my $vol_list_aref = $opts{vol_list};
    my $vol_pattern   = $opts{vol_pattern};
    my $ci            = $opts{command_interface};

    my @vol_list_pattern_match;
    if ($vol_pattern) {
        @vol_list_pattern_match
            = grep { $_ =~ qr/$vol_pattern/i } @{ $vol_list_aref };
    } else {
        @vol_list_pattern_match = @{ $vol_list_aref };
    }

    my @vol_objs;
    foreach my $vol_name (@vol_list_pattern_match) {
        my $vol = NACL::STask::Volume->find(
            command_interface => $ci,
            filter            => { volume => $vol_name },
        );
        if ($vol) {
            push( @vol_objs, $vol );
        }
    }
    return @vol_objs;
}

=head1 NAME

_getClusters

=head1 DESCRIPTION

Pass in a list of node names and return an array of cluster objects to use
as a command interface.

=head1 EXAMPLE

    C<my $ci = _getClusters( nodes =E<gt> "dpg-3270-n1,dpg-3270-n2" );>

=head2 nodes

(Required) Comma seperated list of node names

=cut

sub _getClusters {

    my $self = shift;
    my %opts = @_;

    my @node_names = split( m/[,]/, $opts{nodes} );
    my $node = NACL::C::Node->new( name => $node_names[0] );

    return NACL::STask::Cluster->find_unique_clusters( nodes => [$node] );

}

1;

__END__
