#
# Copyright (c) 2015 NetApp, Inc., All Rights Reserved.
#   Any other use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @pod here
## @author dl-nacl-dev@netapp.com
## @summary An example config module for Xgen

package NACL::Service::Xgen::Examples::ExampleConfig;

use strict;
use warnings;

###############################################
### All configs must directly or indirectly ###
### be derived from the base config module. ###
###############################################
use parent qw(NACL::Service::Xgen::Config);

###############################################################
### All configurations must implement a "build" method.     ###
### This method is responsible for creating and registering ###
### resource objects.                                       ###
###############################################################
sub build {
    my ($self, %args) = @_;

    # invoke the base config's build routine:
    $self->SUPER::build(%args);

    # build some resources objects
    my @simple_resources = ();
    require NACL::Service::Xgen::Examples::ExampleResource;
    foreach (1..3) {
        my $resource = NACL::Service::Xgen::Examples::ExampleResource->new(foo => "bar",
                                                                           bar => [1,2,3]);
        push(@simple_resources, $resource);
    }

    # register the resources with the configuration
    $self->add_resources(resources => \@simple_resources);

    # make some complex resources
    my @complex_resources = ();
    require NACL::Service::Xgen::Examples::ExampleResourceComplex;
    foreach (1..5) {
        my $resource =
            NACL::Service::Xgen::Examples::ExampleResourceComplex->new(required_param1 => "bar",
                                                                       required_param2 => [1,2,3]);
        push(@complex_resources, $resource);
    }

    # register the resources with the configuration
    $self->add_resources(resources => \@complex_resources);

    # configuration is complete!
    return;
}

1;

__END__

=head1 NAME

NACL::Service::Xgen::Examples::ExampleConfig

=head1 DESCRIPTION

This is a simple example implementation of an Xgen Config module. See the base class for all APIs
L<C<NACL::Service::Xgen::Config>|lib-NACL-Service-Xgen-Config-pm> that are available.

=head1 IMPLEMENTATION

To create a configuration module, the very minimum that is needed is to make it a derived class
(indirectly or directly) of L<C<NACL::Service::Xgen::Config>|lib-NACL-Service-Xgen-Config-pm>
and to implement the C<build> method.

=head2 Code

=over

 package NACL::Service::Xgen::Examples::ExampleConfig;

 use strict;
 use warnings;

 ###############################################
 ### All configs must directly or indirectly ###
 ### be derived from the base config module. ###
 ###############################################
 use parent qw(NACL::Service::Xgen::Config);

 ###############################################################
 ### All configurations must implement a "build" method.     ###
 ### This method is responsible for creating and registering ###
 ### resource objects.                                       ###
 ###############################################################
 sub build {
     my ($self, %args) = @_;

     # invoke the base config's build routine:
     $self->SUPER::build(%args);

     # build some resources objects
     my @simple_resources = ();
     require NACL::Service::Xgen::Examples::ExampleResource;
     foreach (1..3) {
         my $resource = NACL::Service::Xgen::Examples::ExampleResource->new(foo => "bar",
                                                                            bar => [1,2,3]);
         push(@simple_resources, $resource);
     }

     # register the resources with the configuration
     $self->add_resources(value => \@simple_resources);

     # make some complex resources
     my @complex_resources = ();
     require NACL::Service::Xgen::Examples::ExampleResourceComplex;
     foreach (1..5) {
         my $resource =
             NACL::Service::Xgen::Examples::ExampleResourceComplex->new(required_param1 => "bar",
                                                                        required_param2 => [1,2,3]);
         push(@complex_resources, $resource);
     }

     # register the resources with the configuration
     $self->add_resources(value => \@complex_resources);

     # configuration is complete!
     return;
 }

=back

=head1 SEE ALSO

For information on the resource modules used in this config, see

L<C<NACL::Service::Xgen::Examples::ExampleResource>|lib-NACL-Service-Xgen-Examples-ExampleResource-pm>

L<C<NACL::Service::Xgen::Examples::ExampleResourceComplex>|lib-NACL-Service-Xgen-Examples-ExampleResourceComplex-pm>

=head1 AUTHOR/MAINTAINER

=over

=item NACL Development (dl-nacl-dev@netapp.com)

=back

=cut
