#
# Copyright (c) 2001-2016 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Helper functions for CGT
## @author madhav.shekar@netapp.com, dl-nacl-dev@netapp.com
## @status shared
## @pod here

package NACL::Tools::CGT::CGTUtils;

use strict;
use warnings;

use Params::Validate qw(validate validate_with SCALAR SCALARREF HASHREF ARRAYREF UNDEF);
use NATE::Log qw(log_global);
use NACL::APISet;
use NACL::APISet::Exceptions::ResponseException qw(:try);
use NACL::APISet::Exceptions::TimeoutException qw(:try);
use NACL::APISet::Response::CLI::ParserUtils;
use NATE::BaseException;
use Data::Dumper qw(Dumper);
use XML::Simple;
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();

BEGIN {
    use Exporter qw(import);
    our @EXPORT_OK = qw(_determine_split_list _convert_to_underscored_form
        perltidy natelint is_edited _remove_trailing_newlines
        parse_leaf_level_cmd determine_pk processCmds processCmd
        _update_method_name _typedef_cli_zapi _supported_zapis
        _convert_hashrefs_array _set_zapi_on_object _get_check_method _get_choices
        _load_event_class _determine_cli_only_fields
    );
}

# This helper function takes in an array of variable names and returns an
# array of lines, each of which is of length 70 or less
# If the component method accepts a lot of options, then simply joining them
# together on a space could lead to very long _optional_scalars() and
# $allowed_cmode_options lines. perltidy does not split extremely
# large lines into multiple lines, so the line could be much longer than the
# 80 character limit. By splitting it into lines of length 70 or less
# we can ensure that these lines fit within the 80 character width limit
sub _determine_split_list {
    my @input_array = @_;
    my @arr2;
    $arr2[0] = '';
    my $length = 0;
    my $i      = 0;
    my $j      = 0;
    foreach my $elem (@input_array) {
        if ($length + length($elem) > 70) {
            $j++;
            $arr2[$j] = '';
            $length = 0;
        }
        if (length $arr2[$j] == 0) {
            $arr2[$j] .= $elem;
        } else {
            $arr2[$j] .= " $elem";
        }
        $length += length $elem;
    }
    return @arr2;
} ## end sub _determine_split_list

## Helper function to extract all those fields which are CLI-only

sub _determine_cli_only_fields {
    my $fields_info = shift;

    my @cli_only_fields = ();
    foreach my $fields (keys %{$fields_info}){
       if(!defined $fields_info->{$fields}->{'zapi-mapping'}
              && !defined $fields_info->{$fields}->{'zapi-typedef-info-pod'}){
          push @cli_only_fields, $fields;
       }
    }

    return @cli_only_fields ;
} ## end sub _determine_cli_only_fields

# Helper function to convert variable names to their underscored form
# This involves converting hyphens and dots to underscores
# It accepts a single (un-named) parameter which is the variable name
# to be converted, it returns the underscored form
sub _convert_to_underscored_form {
    my $variable_name_to_convert = shift;
    $variable_name_to_convert =~ s/-/_/g;
    $variable_name_to_convert =~ s/\./_/g;
    return $variable_name_to_convert;
} ## end sub _convert_to_underscored_form

sub perltidy {
    my %opts = validate @_, {file_path => {type => SCALAR}};
    my $file_path = $opts{file_path};

    my $ptidy_command
        = 'perltidy --perl-best-practices --cuddled-else '
        . '--nooutdent-keywords --notabs --output-line-ending=unix '
        . '--nostandard-output --nostandard-error-output --square-bracket-tightness=2 '
        . '--brace-tightness=2 --paren-tightness=2 -b';
    system("$ptidy_command $file_path");

    # Remove the backup (.bak) file perltidy creates
    unlink "$file_path.bak";
} ## end sub perltidy

sub natelint {
    my %opts = validate @_,
        {
        nate_lib           => {type => SCALAR},
        file_path_from_lib => {type => SCALAR},
        };
    my $error;
    my $nate_lib = $opts{nate_lib};
    my $file     = $opts{file_path_from_lib};

    my $natelint_output = `natelint -prepend_inc $nate_lib $file 2>&1`;
    unless ($natelint_output =~ /OK/) {
        $error
            = 'natelint failed on '
            . "$file. Error was:\n"
            . $natelint_output;
    }
    return $error;
} ## end sub natelint

#
# This function checks whether the component is manually
# updated or not
#
sub is_edited {
    my $file       = shift;
    my $return_val = 0;
    if (!-e $file) {
        $Log->warn(
            "$file is a newly generated file, be sure to " . "p4 add it");
        return 0;
    }
    $/ = ';';
    open(HANDLE, "<$file") or die $!;
    my @content = <HANDLE>;
    if (grep(/set\s*=>\s*\'7Mode\'/, @content)) {
        $Log->comment("$file: modified for 7Mode support");
        $/ = "\n";
        return 1;
    }
    if (grep(/interface\s*=>\s*\'ZAPI\'/, @content)) {
        $Log->comment("$file: modified for ZAPI support");
        $/ = "\n";
        return 1;
    }
    if (grep(/connectrec-match_table/, @content)) {
        $Log->comment("$file: modified to handle prompts");
        $/ = "\n";
        return 1;
    }
    my @spec = grep(/_primary_keys_validate_spec/, @content);
    if (@spec && grep(/optional/, $spec[0])) {
        $Log->comment("$file: primary key validate spec updated");
        $/ = "\n";
        return 1;
    }
    if (!grep(/auto-generated by CGT/, @content)) {
        $Log->comment("$file: \"auto-generated by CGT\" is missing  ");
        $/ = "\n";
        return 1;
    }
    $/ = "\n";
    return 0;
} ## end sub is_edited

sub _remove_trailing_newlines {
    my $array      = shift;
    my @array      = @$array;
    my $last_index = $#array;
    my $i;
    for ($i = $last_index; $i >= 0; $i--) {
        if ($array[$i] ne "\n") {
            next if ($array[$i] eq ";");
            last;
        }
    }
    splice @array, $i if ($i != $last_index);
} ## end sub _remove_trailing_newlines

sub parse_leaf_level_cmd {
    my %opts = validate_with(
        params => \@_,
        spec   => {
            apiset         => {isa  => 'NACL::APISet'},
            api            => {type => SCALAR},
            want_error_msg => {type => SCALAR, default => 1},
            valid_command  => {type => SCALARREF, optional => 1},
        },
    );
    my ($cmdsRef, $caught_exception);
    my ($output, $response, $error);
    my $apiset    = $opts{apiset};
    my $cmd       = $opts{api};
    my $error_msg = $opts{want_error_msg};
    my $valid_command = $opts{valid_command};
    $$valid_command = 1;
    try {
        $cmdsRef = $apiset->cli_command_helper(api => $cmd,
                                               valid_command => $opts{valid_command});
        $cmdsRef->{command} = $cmd;
        $cmdsRef->{actual_cmd} = $apiset->translate_api_to_command(
            api => $cmd);
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = shift;
        my $message   = "Error occurred for '$cmd ?': " . $exception->text();
        if ( $message =~ /Ambiguous argument/ ) {
           $$valid_command = 0;
        }
        if ($error_msg) {
            $Log->comment($message);
        } else {
            $Log->debug($message);
        }
        $caught_exception = 1;
    }
    catch NATE::BaseException with {
        my $exception = shift;
        my $message
            = "Caught a "
            . ref($exception)
            . " for cmd: $cmd"
            . " Exception Text "
            . $exception->text();
        if ($error_msg) {
            $Log->comment($message);
        } else {
            $Log->debug($message);
        }
        $caught_exception = 1;
    };

    if (!$caught_exception) {
        my $vs_context = _check_vserver_context(
            apiset => $apiset,
            cmd    => $cmd
        );
        $cmdsRef->{vserver_context} = $vs_context;
    }
    return $cmdsRef;
} ## end sub parse_leaf_level_cmd

sub determine_pk {
    my %opts = validate_with(
        params => \@_,
        spec   => {
            apiset       => {isa  => 'NACL::APISet'},
            show_options => {type => HASHREF, default => undef},
            show_cmd     => {type => SCALAR},
        },
    );
    my $show_command = $opts{show_cmd};
    my $apiset       = $opts{apiset};
    my $show_options = $opts{show_options};
    my (@show_options, @cdef_derived_primary_keys, @optional_pks, @pks);
    my ($pk, $conversion_hashref ) = {};
    my $cdef = $apiset->get_command_definition(command => $show_command);
    my $alias_map_ref = $cdef->get_alias_map();
    my %reverse_alias_map = ();
    my @cdef_keys = ();
    if ( $alias_map_ref ) {
        %reverse_alias_map = reverse %{ $alias_map_ref };
    }


    # The %Keys hash in the CDEF file will generally be the primary
    # keys for the component. However, in the cases of there being
    # mutually exclusive primary keys (example: for Lun either path or
    # (qtree+volume+lun) should be specified), %Keys will contain all
    # of these keys and hence the primary keys generated will not be
    # entirely accurate. There is a way of using the product to inform
    # us of what exactly the primary keys are (this is explained in
    # the following piece of code). However, for that code to work we
    # need to know the primary keys as specified by the CDEF file
    # (called "CDEF derived primary keys")
    my $schema        = $apiset->{schema};
    my $CLI_separator = $apiset->{session}->{CLI_separator};

    # Determine the CDEF file containing this command's definition,
    # read it, and extract the %Keys hash
    my $cdef_file = $schema->get_command_definition(command => $show_command)
        ->get_file();
    my %CommandDefinitions;
    eval NATE::Inc::get_file("$cdef_file");
    my $keys_ref = $CommandDefinitions{$show_command}->{keys};

    if (defined $keys_ref) {
        my @keys = keys %$keys_ref;
        if ( $alias_map_ref ) {
             require List::Compare;
             my @alias_values = values %reverse_alias_map; 
             my $lc = List::Compare->new(\@keys, \@alias_values);
             @cdef_keys = $lc->get_Lonly();
        } else {
             @cdef_keys = @keys;
        }
        map { push @cdef_derived_primary_keys, $_ if (length($_) > 1); }
            @cdef_keys;
        
        if ( keys %reverse_alias_map ) {
            foreach my $cdef_key (@cdef_derived_primary_keys) {
               if ( $reverse_alias_map{$cdef_key} ) {
                   push( @optional_pks, $cdef_key);
                   push( @optional_pks, $reverse_alias_map{$cdef_key});
               } elsif ( !$alias_map_ref->{$cdef_key} ) {
                   push( @pks, $cdef_key);
               }
            }
        } else {
            @pks = @cdef_derived_primary_keys
        }
    } else {
        return $pk;
    }

    # When any show command is invoked with the fields option
    # specified as a non-primary key, the fields displayed in the
    # output are the primary keys along with the specified non-primary

    # key. For example, invoking "volume show -fields files" will have
    # the output contain the fields volume, vserver and files. Hence,
    # all the fields other that files (i.e. volume and vserver) are
    # the primary keys.
    # The next few lines determine any one non-primary key
    if ($show_options) {
        @show_options = keys %{$show_options};
    } else {
        my $cdef_fields = $CommandDefinitions{$show_command}->{options};
        @show_options = keys %{$cdef_fields};
    }
    my %show_option_hash;
    map { $show_option_hash{$_} = 1 } @show_options;
    map { delete $show_option_hash{$_} } ( @cdef_derived_primary_keys,
                                           keys %reverse_alias_map,
                                           values %reverse_alias_map ) ;
    my @non_primary_keys = keys %show_option_hash;
    my $non_primary_key  = $non_primary_keys[0];
    unless (defined $non_primary_key) {
        return $pk;
    }

    # Subroutine reference which determines the primary keys from the
    # show command output. It does this by considering all fields
    # except the non-primary key. The fields of the output should be
    # provided as parameter
    my $determine_pk_from_output = sub {
        my %opts = validate(@_, {fields => {type => ARRAYREF},});
        my @fields = @{$opts{fields}};
        my %primary_keys_hash;
        foreach my $field (@fields) {
            $primary_keys_hash{$field} = 1;
        }
        delete $primary_keys_hash{$non_primary_key};
        my @primary_keys = keys %primary_keys_hash;
        
        if ( @primary_keys && keys %reverse_alias_map ) {
            @optional_pks = (); @pks = ();
            foreach my $primary_key (@primary_keys) {
               if ( $reverse_alias_map{$primary_key} ) {
                   push( @optional_pks, $primary_key);
                   push( @optional_pks, $reverse_alias_map{$primary_key});
               } elsif ( !$alias_map_ref->{$primary_key} ) {
                   push( @pks, $primary_key);
               }
            }
        } else {
            @pks = @primary_keys
        }
    };

    try {
        my $response = $apiset->$show_command(fields => $non_primary_key);
        my $parsed_output = $response->get_parsed_output();
        my @fields        = (keys %{$parsed_output->[0]});
        $determine_pk_from_output->(fields => \@fields);
    }
    catch NACL::APISet::Exceptions::ResponseException with {

        # This generally occurs if there are no elements of the
        # particular type present on the service
        $Log->warn("Could not run '$show_command'");
    }
    catch NACL::APISet::Exceptions::TimeoutException with {

        # Some show commands might take more than 60 seconds to
        # complete. ("statistics show" is a particularly extreme case)
        # When the command times out we have enough information to
        # determine the primary keys.
        $Log->warn("'$show_command' timed out");

        # Temporary hack to get "debug log show" to abort
        $apiset->get_connection()->send_intr();
        $apiset->get_connection()->send_intr();
        $apiset->get_connection()->resync();

        # The text of the timeout exception is the raw output
        my $timeout = shift;
        my $output  = $timeout->text();

        # Determine the command run by using APISet's xlate
        # functionality
        if ($output) {
            my $xlate = "xlate_$show_command";
            my $command_invoked = $apiset->$xlate(fields => $non_primary_key);

            # Use ParserUtils' remove_echo() to remove command echo
            remove_echo(
                output_ptr => \$output,
                command    => $command_invoked
            );

            # Split into lines, and keep shifting out till we find a line
            # that has the CLI separator. This will be the line with the
            # field names. This should be robust enough to handle any
            # future changes to the text of the TimeoutException
            my @lines = split /\n/, $output;
            while (@lines) {
                last if ($lines[0] =~ /$CLI_separator/);
                shift @lines;
            }

            if (@lines) {
                my $first_line = $lines[0];
                my @fields = split $CLI_separator, $first_line;
                $determine_pk_from_output->(fields => \@fields);
            }
        }
    };
    $pk->{primary_keys} = \@pks;
    $pk->{optional_primary_keys} = \@optional_pks;
    return $pk;
} ## end sub determine_pk

########################################################################
# Function:    processCmds()
#
# Description: Start at the passed command, drill into commands using the
#              '?'.  Split the output into the command and description.
#              Call processCmd on each found command.
#
# Parameters:  cmd        - Command used to start the search
#              cmdHashRef - hash built to contain all discovered commands
#              walk_sub_command - Defaults to 1 if this is set to 0 , it wont walk the
#                                 UI for subcommand, this is required for instant
#                                 component project
# Returns:     Nothing, but does update the passed cmdHashRef
#
# Notes:
########################################################################
sub processCmds {
    $Log->enter() if $may_enter;
    my %params = validate_with(
        params => \@_,
        spec   => {
            cmd              => {type => SCALAR | UNDEF},
            cmdHashRef       => {type => HASHREF},
            walk_sub_command => {type => SCALAR, default => 1},
            apiset           => {isa  => 'NACL::APISet'},
        },
    );

    # Assign the inputs
    my $cmd          = $params{cmd};
    my $cmdsRef      = $params{cmdHashRef};
    my $walk_sub_cmd = $params{walk_sub_command};
    my $apiset       = $params{apiset};

    my ($output, $line);
    $output = $apiset->execute_raw_command(command => "$cmd ?");
    remove_echo("output_ptr" => \$output, "command" => "$cmd ?");

    if ($output =~ s/(\cG)+\r*\n*\s*(INFO|WARNING(.|\n)*)?ERROR:\s//) {

        # When we issue a '?' against a directory with no commands listed
        # under it, we get an error saying
        # ERROR: "" is not a recognized command
        # We simply skip these directories
        $Log->warn("Error for $cmd ?.It appears the  $cmd "
                . "directory is empty. Error message was \n$output");
    } else {

        # Loop through the command's output processing each possible subcmd
        my @lines = split(/\n/, $output);
        foreach $line (@lines) {

            if ($line =~ /^\s*(\S+)\s*(.*)\s*$/) {

                processCmd(
                    cmd              => "$cmd $1",
                    description      => $2,
                    cmdHashRef       => $cmdsRef,
                    walk_sub_command => $walk_sub_cmd,
                    apiset           => $apiset,
                );
            } else {
                $Log->debug("Skipping line: $line");
            }
        }
    }
    $Log->exit() if $may_exit;
} ## end sub processCmds

########################################################################
# Function:    processCmd()
#
# Description: Process the passed command and description.  If the
#              command ends with a '>', there are more subcmds to process,
#              call parent method processCmds to do so.
#
# Parameters:  cmd         - Command to process
#              description - Description of the command
#              cmdHashRef  - ref to a hash to build up with discovered commands
#              walk_sub_command - Defaults to 1 if this is set to 0 , it wont walk the
#                                 UI for subcommand, this is required for instant
#                                 component project
#
# Returns:     Nothing, but does update the passed cmdHashRef
#
# Notes:
########################################################################
sub processCmd {
    $Log->enter() if $may_enter;
    my %params = validate_with(
        params => \@_,
        spec   => {
            cmd              => {type => SCALAR},
            cmdHashRef       => {type => HASHREF},
            description      => {type => SCALAR, optional => 1},
            walk_sub_command => {type => SCALAR, default => 1},
            apiset           => {isa  => 'NACL::APISet'},
        },
    );

    # Assign the passed inputs
    my $cmdsRef         = $params{cmdHashRef};
    my $cmd_description = $params{description};
    my $cmd             = $params{cmd};
    my $walk_sub_cmd    = $params{walk_sub_command};
    my $apiset          = $params{apiset};

    my ($has_commands, $output);

    # Remove leading and trailing spaces and check to see if this
    #   command has subcmds
    $cmd =~ s/^\s*//;
    $cmd =~ s/\s*$//;
    if ($cmd =~ s/>$//) {
        $has_commands = 1;
    }

    # Trim spaces and check if this is a non-admin command (description
    #   preceded with an '*')
    $cmd_description =~ s/^\s*//;
    $cmd_description =~ s/\s*$//;
    $cmd_description =~ s/\*//;

    # If this is a parent node, process the subcmds
    if ($has_commands) {
        if ($walk_sub_cmd) {
            $cmdsRef->{$cmd} = {
                has_commands => $has_commands,
                description  => "$cmd_description",
            };

            # Add the command the cmdHashRef

            processCmds(
                cmdHashRef       => \%{$cmdsRef->{$cmd}{subcmds}},
                cmd              => $cmd,
                apiset           => $apiset,
                walk_sub_command => $walk_sub_cmd
            );
        }
    } else {

        # Issue a '?' against the leaf command
        my $fields = {};
        $cmd =~ s/(\s|\-)/_/g;
        $apiset->{version_info}->{compatibility} = 0;
        my $ds = parse_leaf_level_cmd(apiset => $apiset, api => $cmd);
        if (defined $ds) {
            $fields = $ds->{fields};
            $cmdsRef->{$cmd} = {
                description => "$cmd_description",
                fields      => $fields,
            };
            if ($ds->{is_singleton}) {
                $cmdsRef->{$cmd}->{is_singleton} = 1;
            } else {
                $cmdsRef->{$cmd}->{is_CS} = $ds->{is_CS};
            }
            $cmdsRef->{$cmd}->{vserver_context} = $ds->{vserver_context};
        }
    }
    $Log->exit() if $may_exit;
} ## end sub processCmd

sub _check_vserver_context {
    $Log->enter() if $may_enter;
    my %params = validate_with(
        params => \@_,
        spec   => {
            cmd    => {type => SCALAR},
            apiset => {isa  => 'NACL::APISet'},
        },
    );
    my $vs_context = 0;
    my ($ref, $walk_ui);
    my $apiset = $params{apiset};
    my $cmd    = $params{cmd};
    my $broot  = $apiset->get_version_manager()
        ->get_version_attribute(attribute => 'build_path');
    $broot
        .= "/final/bedrock/export/x86_64/image/mroot/etc/clihelp/help_xml/";
    my $org_cmd = $cmd;
    $cmd =~ s/_/\//g;
    my $xml_file .= "$broot$cmd.xml";
    eval { $ref = XMLin($xml_file); };

    if ($@) {
        $walk_ui = 1;
    }
    if ($walk_ui) {
        $org_cmd =~ s/_/ /g;
        $cmd = "debug smdb table vs_cmds show -command \"" . $org_cmd . "\"";
        my $output = $apiset->execute_raw_command(command => $cmd);
        if ($output !~ /no entries matching/) {
            $vs_context = 1;
        }
    } else {
        if ($ref->{vserver} eq "true") {
            $vs_context = 1;
        }
    }
    return $vs_context;
} ## end sub _check_vserver_context

#Sets method name in component method datastructure,determine whether method name
# clashes with primay key attribute if it clashes ,set method to api name
sub _update_method_name {
    $Log->enter() if $may_enter;
    my %opts = validate_with(
        params => \@_,
        spec   => {
            component_name => {type => SCALAR},
            method_ds      => {type => HASHREF},
            primary_keys   => {type => ARRAYREF, default => []},
        },
    );

    my $api             = $opts{component_name};
    my $command_hashref = $opts{method_ds};
    my @primary_keys    = @{$opts{primary_keys}};
    $api =~ s/-/_/g;
    $api =~ s/ /_/g;
    my $method = $command_hashref->{command};
    $method =~ s/${api}_//g;
    if (   (scalar @primary_keys && grep /^$method$/, @primary_keys)
        || ($method =~ /^\d/)) {
        $method = $command_hashref->{command};
    }
    $command_hashref->{method_name} = $method;
    $Log->exit() if $may_exit;
} ## end sub _update_method_name

#
# This function traverses a nested hash structure
# and it returns a list of keys in that structure
# Input: hash reference
# Output: list of keys in the input hash
# This function is called in many places
#
# Input:
#  $ref = { 'lun-transition-7mode-info' => {
#                       'serial' => '$value'
#                       }
# Output:
#  ( lun-transition-7mode-info serial )
#
sub _convert_hashrefs_array {
    $Log->enter() if $may_enter;
    my ($hashref) = @_;
    my (%temp_hash, @keys, $elem);
    while (defined $hashref
        && ref $hashref) {
        %temp_hash = %{$hashref};
        $elem      = (keys %temp_hash)[0];
        push(@keys, $elem);
        $hashref = $hashref->{$elem};
    }
    $Log->exit() if $may_exit;
    return @keys;
} ## end sub _convert_hashrefs_array

#
#Private function which returns mapping string for typedefs
# Input: hash reference which contain info on a field
# Output: returns a mapping string
#
#Example 1:
#
#Input:
# 'zapi-typedef-info' => {
#            'cluster-peer-info' => {
#                      'peer-addresses' => {
#                            'remote-inet-address' => '@values'
#                      }
#            }
#  },
#
#
#Output:
#            [
#                make_zapi_array('peer-addresses'),
#                make_zapi_skip('remote-inet-address'),
#            ],
#
#Example2 :
#
#Input:
#
#'protocol' => {
#           'zapi-typedef-info' => {
#                     'iscsi-stats-info' => {
#                          'iscsi-error-stats' => {
#                              'iscsi-error-stats-info' => {
#                                  'protocol' => '$value'
#                                   }
#                               }
#                       }
#           }
#
#
#Output:
#
#            [
#              'iscsi-error-stats', make_zapi_skip('iscsi-error-stats-info'),
#              'protocol'
#            ],
#
sub _typedef_cli_zapi {
    $Log->enter() if $may_enter;
    my ($field) = @_;
    my @arr = _convert_hashrefs_array($field->{'zapi-typedef-info-pod'});
    shift @arr;
    my ($str);
    my $size = @arr;
    if ($size == 0) {
        $Log->debug("\n Unable to get the mapping for $field->{field} ");
        return;
    } elsif ($size == 1) {
        $str = $arr[0];
    } else {
        if ($field->{is_array} == 1) {
            $str = "[ make_zapi_array('" .$arr[0]. 
            "'), make_zapi_skip('" .$arr[1]. "'),";
            if ( $arr[2] ) {
                $str .= " '".$arr[2]."'";         
            }
            $str .= " ]";
        } else {
            $str = "[ '". $arr[0]."',".
                   " make_zapi_skip('".$arr[1]."'), '".$arr[2]."',";
            if ( $arr[3] && $arr[4] ) {
               $str .= " make_zapi_skip('".$arr[3]."'), '".$arr[4]."'";
            } elsif ( $arr[3] ) {
                $str .= " '".$arr[2]."'";         
            }
            $str .= " ]";
        }
    }
    return $str;
} ## end sub _typedef_cli_zapi

#
# param1       : An object of ComponentMethod or ComponentStatefile
# param2       : Array reference containing zapis
# Description  : If -iter zapi is contained in the arrayref it sets
#                iter_zapi attribute on the passed in object, similarly
#                for non_iter_zapi it sets non_iter_zapi attribute
#

sub _set_zapi_on_object {
    my ($self, $ds) = @_;
    my (@zapis, $warn);
    my $zapis = $ds->{'zapi-command'};
    if ($zapis) {
        if (ref $zapis) {
            @zapis = @{$zapis};
        } else {
            push @zapis, $zapis;
        }
        my @iter_zapi = grep(/-iter/, @zapis);
        if (@iter_zapi > 1) {
            $warn
                = "\n Unable to add iter ZAPI support for "
                . $ds->{command}
                . " command \n"
                . " equivalent ZAPIs are "
                . join(",", @iter_zapi);
        } elsif (@iter_zapi == 1) {
            $self->iter_zapi($iter_zapi[0]);
        }
        my @non_iter_zapi = grep({$_ !~ /-iter$/} @zapis);
        if (@non_iter_zapi > 1) {
            $warn
                .= "\n Unable to add non-iter ZAPI support for "
                . $ds->{command}
                . " command \n"
                . " equivalent ZAPIs are "
                . join(",", @non_iter_zapi);
        } elsif (@non_iter_zapi == 1) {
            $self->non_iter_zapi($non_iter_zapi[0]);
        }
        if (   $self->nate_lib_isset()
            && $warn) {
            $Log->warn($warn);
        }
    }
} ## end sub _set_zapi_on_object

sub _populate_params_for_check_method{
    $Log->enter() if $may_enter;
    my ($self, $param_ref) = @_ ;

    my ($fields_hashref, $method_name, $vs_context, @pk_keys);

    if(ref($self) eq 'NACL::Tools::CGT::ComponentStateFile'){
       $fields_hashref = $self->ds() ;
       $method_name    = "fetch";
    }else{
       $fields_hashref = $self->comp_method_hashref() ;
       $method_name    = $self->method_name();
       $vs_context     = $self->vserver_context();
       @pk_keys        = $self->primary_keys();
    }
    $vs_context //= $fields_hashref->{vserver_context} if(defined $fields_hashref->{vserver_context});
    @pk_keys     = @{$fields_hashref->{primary_keys}} if(! @pk_keys);

    $param_ref->{vs_context}      = $vs_context ;
    $param_ref->{pk_keys}         = \@pk_keys ;
    my @cli_fields                 = $self->cli_only_fields();
    $param_ref->{cli_fields}      = \@cli_fields ;
    $param_ref->{method_name}    = $method_name ;
}

sub _get_check_method {
    $Log->enter() if $may_enter;
    my ($self) = @_ ;

    my %params = ();
    _populate_params_for_check_method($self, \%params);

    my $str = "";
    my @methods ;
    my $method_name = $params{method_name};
    push @methods , "_check_${method_name}_cmode_zapi_iter"  if($self->iter_zapi_isset());
    push @methods, "_check_${method_name}_cmode_zapi_non_iter"  if($self->non_iter_zapi_isset()) ;
    my $code_generated = 0;
    foreach my $method_name (@methods) {
       $code_generated = 0;
       my $iter_str = "";
       if($method_name =~ /non_iter/){
         $params{method_name} = $method_name ;
         _generate_check_method_for_non_iter(\$str, \$code_generated, \%params);
         $iter_str = "non-iter";
       }else{
          $params{method_name} = $method_name ;
          _generate_check_method_for_iter(\$str, \$code_generated, \%params);
          $iter_str = "iter" ;
       }
       if($code_generated && @{$params{cli_fields}}){
         $str .= <<"HEREDOC";
         if(\@invalid_fields) {
         my \$msg = 'The following fields provided '
                . 'in the call to '.\$pkg
                . ' are not supported by the '
                . '$iter_str ZAPI, so the ZAPI back-end is being '
                . "skipped:\n"
                . join (', ', \@invalid_fields);
         \$Log->comment(\$msg);
         \$Log->exit() if \$may_exit;
         NACL::Exceptions::InvalidChoice->throw(\$msg);
         }
        \$Log->exit() if \$may_exit;
       }
HEREDOC
       }elsif($code_generated && ! @{$params{cli_fields}}) {
         $str .= '$Log->exit() ;'."\n".'}';
       }
    }#end foreach my $method_name (@methods) { ..

    $Log->exit ;
    return $str;
} ## end sub _get_check_method

sub _generate_check_method_for_iter{
    my ($str, $code_generated, $param_hash_ref) = @_;
    my $cli_str  = join(" ", @{$param_hash_ref->{cli_fields}} ) ;
    my $convert_str =  'qw(' . $cli_str . ')';
    if($cli_str){
        $$code_generated = 1;
        $$str .= <<"HEREDOC";
sub $param_hash_ref->{method_name}  {
HEREDOC
        $$str .= <<'HEREDOC';
    $Log->enter() if $may_enter;
    my ($pkg, %opts) = @_;
HEREDOC
    if($param_hash_ref->{method_name} =~ /fetch/){
      $$str .= <<'HEREDOC';
      my @invalid_fields = $pkg->_invalid_fields_check(
           filter           => $opts{filter},
           requested_fields => $opts{requested_fields},
           _fields          =>
HEREDOC
      $$str .= <<"HEREDOC";
[ $convert_str ]  );
HEREDOC
    }
    else{
       $$str .= <<"HEREDOC";
       my \@invalid_fields = ();
       ## following are the CLI-only fields
       my \@cli_only_fields =
HEREDOC
       $$str .= <<"HEREDOC";
$convert_str ;
       foreach my \$field(\@cli_only_fields) {
         push \@invalid_fields, \$field if(defined \$opts{\$field});
       }
HEREDOC
      } ##end else
    } ##end if ($cli_str){
}#end sub _generate_check_method_for_iter{ ..

sub _generate_check_method_for_non_iter{
    my ($str, $code_generated, $param_hash_ref) = @_;
    my $cli_str  = join(" ", @{$param_hash_ref->{cli_fields}} ) ;
    my $convert_str =  'qw(' . $cli_str . ')';
     if (@{$param_hash_ref->{pk_keys}} ) {
        $$code_generated = 1 ;
        $$str .= "sub $param_hash_ref->{method_name} {" ;
        $$str .= <<'HEREDOC';
        $Log->enter() if $may_enter;
        my ($pkg, @opts) = @_;
        $pkg->_base_check_non_iter(
         @opts,
HEREDOC
        my $pk_str = join(" ", @{$param_hash_ref->{pk_keys}});
        $$str .= <<"HEREDOC";

        _primary_keys => [qw($pk_str)],
HEREDOC
      if ( $param_hash_ref->{vs_context} ) {
         $$str .= <<"HEREDOC";
            _vs_context   => $param_hash_ref->{vs_context}

HEREDOC
      }## end if ( $param_hash_ref->{vs_context} ) {
      $$str .= <<"HEREDOC";
    );
HEREDOC
     } ## end if (@{$param_hash_ref->{pk_keys}} ) {

    if($cli_str){
      $$code_generated = 1 ;
      if(! @{$param_hash_ref->{pk_keys}}){
        $$str .= "sub $param_hash_ref->{method_name} {" ;
        $$str .= '$Log->enter();' ;
        $$str .= 'my ($pkg, @opts) = @_;' ;
      }

    if($param_hash_ref->{method_name} =~ /fetch/){
      $$str .=  'my %opts_hashref = @opts;'. "\n" ;

      $$str .= <<'HEREDOC';
      my @invalid_fields = $pkg->_invalid_fields_check(
           filter           => $opts_hashref{filter},
           requested_fields => $opts_hashref{requested_fields},
           _fields          =>
HEREDOC
      $$str .= <<"HEREDOC";
[ $convert_str ]  );
HEREDOC
    }else{
      $$str .= <<"HEREDOC";
    my \@invalid_fields = ();
    ## following are the CLI-only fields
    my \@cli_only_fields =
HEREDOC
    $$str .= <<"HEREDOC";
$convert_str
        ;
    foreach my \$field(\@cli_only_fields) {
        push \@invalid_fields, \$field if(grep (/^\$field\$/ , \@opts));
}
HEREDOC
     }## end else 
    }# end if( $cli_str}){ ...
}# end sub _generate_check_method_for_non_iter{ ...

sub _get_choices {
    $Log->enter() if $may_enter;
    my ($self, $method_name) = @_;
    my $is_fh_set = $self->filehandle_isset();

    ## get cli fields
    my @cli_fields = ();
    @cli_fields = $self->cli_only_fields() ;

    ## get primary_keys
    my ($fields_hashref, $vs_context, @primary_keys, $param_hash);

    if(ref($self) eq 'NACL::Tools::CGT::ComponentStateFile'){
       $fields_hashref = $self->ds() ;
    }else{
       $fields_hashref = $self->comp_method_hashref() ;
       @primary_keys   = $self->primary_keys();
    }
    @primary_keys      = @{$fields_hashref->{primary_keys}} if(!@primary_keys);

    my $str = <<"HEREDOC";
            {   method    => '_${method_name}_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
                zapi_type => 'none',
HEREDOC
    if ( !$is_fh_set ) {
        $str .= <<'HEREDOC';
            version_info => {cdefs_from_build => 1},
HEREDOC
    }
    $str .= <<'HEREDOC';
        },
HEREDOC
    if ($self->iter_zapi_isset()) {
        $str .= <<"HEREDOC";
            {   method    => '_${method_name}_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
                zapi_type => 'iter',
HEREDOC
        if(@cli_fields){
          $str .= <<"HEREDOC";
		check     => \'_check_${method_name}_cmode_zapi_iter\',
HEREDOC
        }
        if ( !$is_fh_set ) {
            $str .= <<'HEREDOC';
                version_info => {cdefs_from_build => 1},
HEREDOC
        }
        $str .= <<'HEREDOC';
        },
HEREDOC
    }
    if ($self->non_iter_zapi_isset()) {
        $str .= <<"HEREDOC";
            {   method    => '_${method_name}_cmode_zapi_non_iter',
                interface => 'ZAPI',
                set       => 'CMode',
                zapi_type => 'non-iter',
HEREDOC
        if( @cli_fields || @primary_keys ){
            $str .= "    check     => '_check_${method_name}_cmode_zapi_non_iter',\n";
        }
        if ( !$is_fh_set ) {
            $str .= <<'HEREDOC';
                version_info => {cdefs_from_build => 1},
HEREDOC
        }
        $str .= "},";
    }
    $Log->exit() if $may_exit;
    return $str;
} ## end sub _get_choices

sub _load_event_class {
    $Log->enter() if $may_enter;
    $Log->exit() if $may_exit;
    return "__PACKAGE__->_load_event_class();";
}
1;

__END__

=head1 NAME

NACL::Tools::CGT::CGTUtils

=head1 DESCRIPTION

This package contains utility functions used internally by CGT.

=cut
