
#
# Copyright (c) 2001-2011 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Class representing each component built
## @author madhav.shekar@netapp.com, dl-nacl-dev@netapp.com
## @status shared
## @pod here

package NACL::Tools::CGT::ComponentDS;

use strict;
use warnings;

use NACL::Tools::CGT::ComponentMethod;
use NACL::Tools::CGT::ComponentStateFile;
use NATE::Log qw(log_global);
use Data::Dumper qw(Dumper);
use NACL::Tools::CGT::CGTUtils
    qw(_determine_split_list _convert_to_underscored_form _remove_trailing_newlines _supported_zapis _set_zapi_on_object
       _load_event_class);
use NATE::ParamSet;
use Tie::File;

use Class::MethodMaker [
    new   => [ '-hash', 'new' ],
    array => 'component_ds',
    scalar => [ { -default => 0 }, 'contains_CS_file' ],
    array  => 'primary_keys',
    array  => 'optional_primary_keys',
    scalar => 'component_name',
    scalar => 'filename',
    scalar => 'C_package_name',
    scalar => 'C_FH',
    scalar => 'CS_package_name',
    hash   => 'CS_ds',
    hash   => 'warnings',
    array  => 'CS_fields',
    scalar => 'CS_command',
    scalar => [ { -default => 0 }, 'CS_is_singleton' ],
    scalar => 'CS_FH',
    scalar => 'nate_lib',
    scalar => 'should_perltidy',
    scalar => 'should_natelint',
    scalar => 'found_init',
    scalar => 'generate_pod_alone',
    scalar => [ { -default => 0 }, 'contains_CS_zapi' ],
    hash   => 'cs_method_hashref'
];
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();

sub build_packages {
    my $self = shift;
    my @errors;
    $self->determine_package_name();

    my @component_ds = $self->component_ds();

    my @show_array = grep { $_->{method_name} eq 'show' } @component_ds;
    $self->determine_CS_package_name();
    my $cs_file;
    my $should_perltidy = $self->should_perltidy();
    my $should_natelint = $self->should_natelint();
    if ( scalar @show_array ) {
        $self->contains_CS_file(1);
        $cs_file = NACL::Tools::CGT::ComponentStateFile->new(
            nate_lib        => $self->nate_lib(),
            C_package_name  => $self->C_package_name(),
            filename        => $self->filename(),
            component_name  => $self->component_name(),
            should_natelint => $self->should_natelint(),
            should_perltidy => $self->should_perltidy(),
        );
        $cs_file->ds(%{ $show_array[0]});
        $cs_file->warnings( %{ $self->warnings() } );
        $cs_file->command( $show_array[0]->{command} );
        if ( defined $show_array[0]->{'zapi-command'} ) {
            _set_zapi_on_object($cs_file, $show_array[0]);
            $self->contains_CS_zapi(1);
        }
        my @CS_fields = keys %{ $cs_file->ds()->{fields} };
        $cs_file->fields(@CS_fields);

        $cs_file->is_singleton( $show_array[0]->{is_singleton} );

        my @primary_keys;
        if ( $cs_file->ds()->{primary_keys} ) {
            @primary_keys = @{ $cs_file->ds()->{primary_keys} };
        } else {
            @primary_keys = ();
        }
        $self->primary_keys(@primary_keys);
        my @optional_primary_keys;
        if ( $cs_file->ds()->{optional_primary_keys} ) {
            @optional_primary_keys = @{ $cs_file->ds()->{optional_primary_keys} };
        } else {
            @optional_primary_keys = ();
        }
        $self->optional_primary_keys(@optional_primary_keys);
        $cs_file->CS_package_name( $self->CS_package_name() );
    }
    
    my $p4_client_lib_root = $self->nate_lib();
    my $file               = $self->filename();

    if ( $self->contains_CS_file() ) {
        $cs_file->build_cs_package();
        if ( $should_natelint ) {
            my $error = $cs_file->natelint();
            push ( @errors, $error)
                if ( $error );
        }
    }

    $self->build_c_package();
    $self->perltidy() if $should_perltidy;
    if ( $should_natelint ) {
       my $error = $self->natelint();
       push ( @errors, $error)
           if ( $error );
    }
    return \@errors;
}

sub determine_package_name {
    my $self               = shift;
    my $component_name     = $self->component_name();
    my @package_name_elems = split( /\s|-/, $component_name );
    my $filename           = '';
    map { $filename .= ucfirst $_ } @package_name_elems;
    $self->filename($filename);
    my $package_name = $filename;
    $self->C_package_name("NACL::C::$package_name");
}

sub determine_CS_package_name {
    my $self            = shift;
    my $CS_package_name = $self->C_package_name();
    $CS_package_name =~ s/::C::/::CS::/;
    $self->CS_package_name($CS_package_name);
}

sub build_c_package {
    my $self               = shift;
    my $p4_client_lib_root = $self->nate_lib();
    my $file               = $self->filename();
    my $path               = "$p4_client_lib_root/NACL/C/$file.pm";
    if ( !-e $path ) {
        $Log->comment("Writing the component file NACL::C::$file");
        open( my $C_FH, '>', $path ) or die $!;
        $self->C_FH($C_FH);
        $self->build_c_header();
        $self->build_c_attribute_list();
        $self->build_c_package_name_includes();
        $self->find_init_method();
        $self->build_c_new();
        $self->build_vserver_init();
        $self->build_c_methods();
        my $found_init = $self->found_init();

        if ($found_init) {
            $self->write_custom_new();
        }
        $self->add_event_class();
        $self->build_c_primary_keys_validate_spec();
        $self->build_c_return_true();
        close $C_FH;
        $Log->warn(
            "$file is a newly generated file, be sure to " . "p4 add it" );
    } elsif ( -w $path ) {
        my @component_ds = $self->component_ds();
        my ( %existing_method_hashref, $frontend_index, @array );
        my @missing_method_list      = ();
        my @missing_method_name      = ();
        my %warnings                 = $self->warnings();
        tie @array, 'Tie::File', $path, recsep => ';';
        _remove_trailing_newlines( \@array );
        my $method_maker;

        foreach my $line (@array) {
            if ( $line =~ /Class::MethodMaker/s ) {
                $method_maker = $line;
                last;
            }
        }
        my @primary_keys_from_ds = $self->primary_keys();

        my ( @new_keys, @existing_keys );
        foreach (@primary_keys_from_ds) {
            my $underscored_keys = _convert_to_underscored_form($_);
            if ( $method_maker !~ /scalar\s*=>\s*['"]$underscored_keys['"]/s )
            {
                push @new_keys, $_;
            }
        }

        if ($method_maker) {

            #Find the existing keys from the C file
            #Parsing the MethodMaker to figure-out
            foreach ( split( /,/, $method_maker ) ) {
                if ( $_ =~ /.*scalar\s*=>\s*['"](.+)['"]$/s ) {
                    my $match = $1;
                    $match =~ s/_/-/g;
                    push @existing_keys, $match;
                }
            }
        } else {
            @existing_keys = ();
        }

        foreach my $comp_method_hashref (@component_ds) {
            my $method_name = $comp_method_hashref->{method_name};
            if ( defined $warnings{$file}{$method_name}
                 && $warnings{$file}{$method_name} eq "do_not_add" ) {
               next;
            }
            #Handle the case of import method
            $method_name = "do_" . $method_name
                if ( $method_name eq "import" );
            $frontend_index = 0;
            if ( $method_name eq 'show' ) {
                $self->cs_method_hashref($comp_method_hashref);
                next;
            }
            for ( my $i = 0; $i < $#array; $i++ ) {
                if ( $array[$i] =~ /sub\s+$method_name\s*\{/s ) {
                    $frontend_index = $i;
                    last;
                }
            }
            if ( $frontend_index == 0 ) {
                if ( $method_name eq "init" ) {
                    $Log->warn("\ninit needs to be added manually to $path");
                } else {
                    push @missing_method_list, $comp_method_hashref;
                    push @missing_method_name, $method_name;
                }
            } else {
                $existing_method_hashref{$method_name} = $frontend_index;
            }
        }

        if (@new_keys) {
            my $flag = 0;
            if ( exists $warnings{$file}{new_primary_keys} ) {
                my $ignore_primary_keys = $warnings{$file}{new_primary_keys};
                my @primary_keys = split( /\s+/, $ignore_primary_keys );
                foreach my $key (@new_keys) {
                    if ( !grep { $key eq $_ } @primary_keys ) {
                        $flag = 1;
                        last;
                    }
                }
            } else {
                $flag = 1;
            }

            my $temp = join( " ", @new_keys );
            if ($flag) {
                if (@missing_method_list) {
                    my $new_methods = join( " ", @missing_method_name );
                    $Log->warn( "New Primary keys: $temp found, "
                            . $new_methods
                            . "\n methods in $path has to be updated manually to support  the \n"
                            . "new primary keys and also check all the existing methods to see whether \n"
                            . "the new primary keys are supported or not" );
                } else {
                    $Log->warn(
                        "New Primary keys: $temp found, check all the methods in $path \n"
                            . "to see whether the new primary keys are supported or not"
                    );
                }
            }
        }
        my $should_perltidy = $self->should_perltidy();
        $self->perltidy() if $should_perltidy;
        $self->update_method_with_missing_options( \%existing_method_hashref,
            \@existing_keys );
        $self->build_missing_c_methods( \@missing_method_list )
            if (@missing_method_list);
    } else {
        $Log->warn("$path is not editable ");
    }
}

sub find_init_method {
    my $self                     = shift;

    my @component_ds = $self->component_ds();
    my $found        = 0;
    foreach my $comp_method_hashref (@component_ds) {
        my $method_name = $comp_method_hashref->{method_name};
        if ( $method_name eq "init" ) {
            $found = 1;
            last;
        }
    }
    $self->found_init($found);
}

# Build the header of the Component file
# This would include the copyright lines, leading tharndoc, along with the
# POD for Name and Description
sub build_c_header {
    my $self = shift;

    my $file = $self->filename();
    my $str  = <<"HEREDOC";
# \$Id\$

#
# Copyright (c) 2001-2014 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## \@summary $file Component Module (auto-generated by CGT)
## \@author dl-nacl-dev\@netapp.com
## \@status shared
## \@pod here

HEREDOC

    my $package_name = $self->C_package_name();
    $str .= <<"HEREDOC";
=head1 NAME

$package_name

=head1 DESCRIPTION

HEREDOC

    my $CS_package_name            = $self->CS_package_name();
    my $hyphenated_CS_package_name = $CS_package_name;
    $hyphenated_CS_package_name =~ s/::/-/g;
    $hyphenated_CS_package_name .= '-pm';
    my $contains_CS  = $self->contains_CS_file();
    my @primary_keys = $self->primary_keys();

    if ( $contains_CS && @primary_keys ) {
        $str .= <<"HEREDOC";
C<$package_name> is a derived class of
L<NACL::C::Component::ONTAP|lib-NACL-C-Component-ONTAP-pm>. It represents access
to an ONTAP $file.  A related class is
L<$CS_package_name|lib-$hyphenated_CS_package_name>, which represents
the past state of an ONTAP $file.

HEREDOC
    } elsif ( !$contains_CS ) {
        $str .= <<"HEREDOC";
C<$package_name> is a derived class of
L<NACL::C::Component::ONTAP|lib-NACL-C-Component-ONTAP-pm>. 
Unlike other components this package does not model any element/service 
but is present to provide component-level support for functionality. 
Since it does not model any element/service there is no ComponentState 
module for this component. For the same reason, 'find' and 'state' will
not work for this component.
Instead $CS_package_name->fetch() should be used.
(See L<fetch|lib-NACL-CS-ComponentState-pm/fetch> for a full description)

=cut

HEREDOC
    } elsif ( $contains_CS && !@primary_keys ) {
        $str .= <<"HEREDOC";
C<$package_name> is a derived class of
L<NACL::C::Component::ONTAP|lib-NACL-C-Component-ONTAP-pm>. It represents access
to an ONTAP $file.  A related class is
L<$CS_package_name|lib-$hyphenated_CS_package_name>, which represents
the past state of an ONTAP $file.
However, since it does not have any primary keys, 'find' and 'state' do not
work for this component. (This component is considered a 'singleton',
i.e. there is only one row shown by the relevant 'show' command)

=cut

HEREDOC
    }
    if ( $self->generate_pod_alone() ) {
        return $str;
    } else {
        my $C_FH = $self->C_FH();
        print $C_FH $str;
    }
}

sub build_c_attribute_list {
    my $self         = shift;
    my @primary_keys = $self->primary_keys();
    my $str;
    if (@primary_keys) {
        $str = <<"HEREDOC";
=head1 ATTRIBUTES

=head2 command_interface

See L<NACL::C::Component::command_interface|lib-NACL-C-Component-pm/command_interface>

HEREDOC

        foreach my $primary_key (@primary_keys) {

          # We can't directly modify $primary_key since it's an alias and it
          # would end up modifying the contents of @primary_keys.
          # Hence, it's moved into a separate variable, which is then modified
          # to have hyphens changed to underscores.
            my $hyphenated_form = $primary_key;
            my $underscored_primary_key
                = _convert_to_underscored_form($hyphenated_form);
            $str .= <<"HEREDOC";
=head2 $underscored_primary_key

HEREDOC
        }
    } else {
        $str = <<"HEREDOC";
=head1 ATTRIBUTES

None.

HEREDOC
    }

    $str .= <<"HEREDOC";
=cut

HEREDOC

    if ( $self->generate_pod_alone() ) {
        return $str;
    } else {
        my $C_FH = $self->C_FH();
        print $C_FH $str;
    }
}

sub build_c_package_name_includes {
    my $self = shift;

    my $package_name = $self->C_package_name();
    my $str          = <<"HEREDOC";
package $package_name;

use strict;
use warnings;

HEREDOC

    if ( $self->contains_CS_file() ) {
        my $CS_package_name = $self->CS_package_name();
        $str .= <<"HEREDOC";
use $CS_package_name;
HEREDOC
    }

    $str .= <<"HEREDOC";
use NACL::ComponentUtils qw (_optional_scalars Dumper);

use base 'NACL::C::Component::ONTAP';
use Params::Validate qw(validate_with SCALAR ARRAYREF SCALARREF);
use NATE::Log qw(log_global);
use NACL::CS::ComponentState::ZapiSkip qw(make_zapi_skip);
use NACL::CS::ComponentState::ZapiArray qw(make_zapi_array);
use feature 'state';
HEREDOC

    my $C_FH = $self->C_FH();
    print $C_FH $str;
}

sub build_c_new {
    my $self = shift;

    my @primary_keys = $self->primary_keys();
    my @optional_primary_keys = $self->optional_primary_keys();
    @primary_keys = (@primary_keys, @optional_primary_keys);
    my $file         = $self->filename();
    my $package_name = $self->C_package_name();
    my $str;
    $str .= <<"HEREDOC";
=head1 METHODS

=head2 new

  my \$$file = $package_name->new(
                                   command_interface => \$ci,
HEREDOC

    my %underscored_form;
    foreach my $primary_key (@primary_keys) {
        my $hyphenated_form = $primary_key;
        my $underscored_primary_key
            = _convert_to_underscored_form($hyphenated_form);
        $underscored_form{"$hyphenated_form"} = $underscored_primary_key;
        $str .= <<"HEREDOC";
                                   '$primary_key' => \$$underscored_primary_key,
HEREDOC
    }
    $str .= <<"HEREDOC";
								  );

Return a new $file component object with the given attributes.

=cut

HEREDOC

    if ( !$self->generate_pod_alone() ) {
        $str .= <<"HEREDOC";
use Class::MethodMaker [
HEREDOC

        foreach my $primary_key (@primary_keys) {
            my $underscored_primary_key = $underscored_form{"$primary_key"};
            $str .= <<"HEREDOC";
    scalar => '$underscored_primary_key',
HEREDOC
        }

        $str .= <<"HEREDOC";
];
    my \$Log = log_global();
$may_enter = $Log->may_enter();
$may_exit  = $Log->may_exit();
HEREDOC
    }
    if ( $self->contains_CS_file() ) {
        my $file_obj        = "${file}_obj";
        my $file_objs       = $file_obj . 's';
        my $state           = $file . '_state';
        my $CS_package_name = $self->CS_package_name();
        $str .= <<"HEREDOC";
=head2 find

    my \$$file_obj  = $package_name->find(command_interface => \$ci, ...);
	my \@$file_objs = $package_name->find(command_interface => \$ci, ...); 

(Class method)
A constructor of sorts, which discovers which elements are present (by
interacting with the given command interface).  In an array context it
creates a component object for each present element and returns them.
In a scalar context it creates a component object for just one
element.

This method is generally not implemented in individual components. The base
class (L<NACL::C::Component|lib-NACL-C-Component-pm> provides an implementation.
See <L<NACL::C::Component::find|lib-NACL-C-Component-pm/find> for a more
detailed description along with explanation of all the options accepted.

HEREDOC
    if ( $self->contains_CS_zapi() ) {
        $str .= <<HEREDOC;
Uses a CMode CLI/ZAPI APISet.

HEREDOC
    } else {
        $str .= <<HEREDOC;
Uses a CMode CLI APISet.

HEREDOC
    }
$str .= <<HEREDOC;
=head2 state

    my $state = \$$file_obj->state();

(Instance method)
Fetch the current state of the element associated
with this component, and return a Component State object corresponding
to this component (calling this on $package_name would return 
a $CS_package_name object)

This method is generally not implemented in individual components. The base
class (L<NACL::C::Component|lib-NACL-C-Component-pm> provides an implementation.
See <L<NACL::C::Component::state|lib-NACL-C-Component-pm/state> for a more
detailed description along with explanation of all the options accepted.

HEREDOC

    if ( $self->contains_CS_zapi() ) {
         $str .= <<HEREDOC;
Uses a CMode CLI/ZAPI APISet.

HEREDOC
    } else {
        $str .= <<HEREDOC;
Uses a CMode CLI APISet.

HEREDOC
    }
    $str .= <<HEREDOC
=cut

HEREDOC
    }

    if ( $self->generate_pod_alone() ) {
        return $str;
    } else {
        my $C_FH = $self->C_FH();
        print $C_FH $str if $str;
    }
}

sub build_vserver_init {
    my $self         = shift;
    my @component_ds = $self->component_ds();
    my $C_FH         = $self->C_FH();
    my $str;
    if ( $component_ds[0]->{vserver_context} ) {
        $str .= <<'HEREDOC';
sub init {
    $Log->enter() if $may_enter;
    my ( $self, @args ) = @_;
    $self->SUPER::init(@args);
    $self->_vserver_context_init(@args);
    $Log->exit() if $may_exit;
}
HEREDOC
        print $C_FH $str;
    }
}

sub build_c_methods {
    my $self                     = shift;
    my @component_ds             = $self->component_ds();
    my @primary_keys             = $self->primary_keys();
    my @optional_primary_keys    = $self->optional_primary_keys();
    my $generate_pod_alone       = $self->generate_pod_alone();
    my $pod_str                  = "";
    my $component_method;
    my $generate_check_method    = 0;
    my @show_array = grep { $_->{method_name} eq 'show' } @component_ds;
    foreach my $comp_method_hashref (@component_ds) {
        my $api_method  = $comp_method_hashref->{command};
        my $method_name = $comp_method_hashref->{method_name};
        #Handle the case of import method
        $method_name = "do_" . $method_name
            if ( $method_name eq 'import' );

        next
            if ( $method_name eq 'show' );
        my $package_var;
        my $create_object_flag = 0;
        if (( $method_name =~ /create/ )
            or (    ( $method_name eq 'start' )
                and ( $comp_method_hashref->{is_static} ) )
            or ( $method_name eq 'copy' )
            )
        {
            $create_object_flag = 1;
        }
        $comp_method_hashref->{is_static} = 1 unless (@primary_keys);

        if ( ( $comp_method_hashref->{is_static} ) or ($create_object_flag) )
        {
            $package_var = '$pkg';
        } else {
            $package_var = '$pkg_or_obj';
        }

        my $fields = $comp_method_hashref->{fields};
        my ( @options, @array_list, @optional_scalars );
        while ( my ( $field_name, $field_hashref ) = each %$fields ) {
            push @options, $field_name;
            if ( $field_hashref->{is_array} ) {
                push @array_list, $field_name;
            } else {
                if ( !$comp_method_hashref->{can_use_cvw} ) {
                    push @optional_scalars, $field_name;
                } elsif (
                    !grep {
                        $_ eq $field_name
                    } @primary_keys
                    )
                {
                    push @optional_scalars, $field_name;
                }
            }
        }

        $component_method = NACL::Tools::CGT::ComponentMethod->new(
            method_name         => $method_name,
            package_var         => $package_var,
            create_object_flag  => $create_object_flag,
            comp_method_hashref => $comp_method_hashref,
            api_method          => $api_method,
            package_name        => $self->C_package_name(),
            filename            => $self->filename(),
            filehandle          => $self->C_FH(),
            generate_pod_alone  => $generate_pod_alone,
            vserver_context     => $comp_method_hashref->{vserver_context},
            nate_lib            => $self->nate_lib()
        );
        $component_method->options(@options);
        $component_method->array_list(@array_list);
        $component_method->optional_scalars(@optional_scalars);
        $component_method->primary_keys(@primary_keys);
        $component_method->optional_primary_keys(@optional_primary_keys);
        $component_method->cs_method_hashref($show_array[0])
            if ( @show_array );
        _set_zapi_on_object($component_method, $comp_method_hashref);
        if ($generate_pod_alone) {
            if ( $comp_method_hashref->{is_CS} ) {
                my $filename = $self->return_cslike_filename($method_name);
                $pod_str .= $component_method->build_cs_like_frontend_method(
                    $filename);
            } else {
                $pod_str .= $component_method->build_c_frontend_method();
            }
            $pod_str .= "\n\n";
            next;
        }
        if ( $comp_method_hashref->{is_CS} ) {
            my $filename = $self->return_cslike_filename($method_name);
            $component_method->build_cs_like_frontend_method($filename);
            $self->build_or_update_cslike_state_file( $filename, $method_name,
                    $comp_method_hashref );
        } else {
            $component_method->build_c_frontend_method();
            $component_method->build_c_backend_method();
        }
        if ( $component_method->non_iter_zapi_isset()
             && !$generate_check_method
             && @primary_keys
             && !$generate_pod_alone ) {
            $component_method->build_c_check_method();
            $generate_check_method = 1;
        }
    }
    if ($generate_pod_alone) {
        return $pod_str;
    }
}

#
#This method will add the missing methods to the C file
#It is an instance method ,called by an object of class
#componentDS
# arg1 is an arrayref of component data struture
# arg2 is an arrayref ,which contains the missing method hashrefs
#

sub build_missing_c_methods {
    my $self                     = shift;
    my $p4_client_lib_root       = $self->nate_lib();
    my $file                     = $self->filename();
    my $path                     = "$p4_client_lib_root/NACL/C/$file\.pm";
    my $component_ds             = shift;
    my @primary_keys             = $self->primary_keys();
    my $cs_method_hashref        = $self->cs_method_hashref();
    my @content;
    tie( @content, 'Tie::File', $path );
    my $str_frontend_backend;

    foreach my $comp_method_hashref (@$component_ds) {
        my $api_method  = $comp_method_hashref->{command};
        my $method_name = $comp_method_hashref->{method_name};

        #Handle the case of import method
        $method_name = "do_" . $method_name
            if ( $method_name eq 'import' );
        $Log->comment("$method_name will be  added to $file component");
        my $package_var;
        my $create_object_flag = 0;
        if (( $method_name =~ /create/ )
            or (    ( $method_name eq 'start' )
                and ( $comp_method_hashref->{is_static} ) )
            or ( $method_name eq 'copy' )
            )
        {
            $create_object_flag = 1;
        }
        $comp_method_hashref->{is_static} = 1 unless (@primary_keys);

        if ( ( $comp_method_hashref->{is_static} ) or ($create_object_flag) )
        {
            $package_var = '$pkg';
        } else {
            $package_var = '$pkg_or_obj';
        }

        my $fields = $comp_method_hashref->{fields};
        my ( @options, @array_list, @optional_scalars );
        while ( my ( $field_name, $field_hashref ) = each %$fields ) {
            push @options, $field_name;
            if ( $field_hashref->{is_array} ) {
                push @array_list, $field_name;
            } else {
                if ( !$comp_method_hashref->{can_use_cvw} ) {
                    push @optional_scalars, $field_name;
                } elsif (
                    !grep {
                        $_ eq $field_name
                    } @primary_keys
                    )
                {
                    push @optional_scalars, $field_name;
                }
            }
        }

        my $component_method = NACL::Tools::CGT::ComponentMethod->new(
            method_name         => $method_name,
            package_var         => $package_var,
            create_object_flag  => $create_object_flag,
            comp_method_hashref => $comp_method_hashref,
            api_method          => $api_method,
            package_name        => $self->C_package_name(),
            filename            => $self->filename(),
            vserver_context     => $comp_method_hashref->{vserver_context},
        );
        $component_method->options(@options);
        $component_method->array_list(@array_list);
        $component_method->optional_scalars(@optional_scalars);
        $component_method->primary_keys(@primary_keys);
        $component_method->cs_method_hashref($cs_method_hashref);
        _set_zapi_on_object($component_method, $comp_method_hashref); 
        my $str_frontend;
        my $str_backend;

        if ( $comp_method_hashref->{is_CS} ) {
            my $filename = $self->return_cslike_filename($method_name);
            $str_frontend
                = $component_method->build_cs_like_frontend_method($filename);
            $self->build_or_update_cslike_state_file( $filename, $method_name,
                $comp_method_hashref );
        } else {
            $str_frontend = $component_method->build_c_frontend_method();
            $str_backend  = $component_method->build_c_backend_method();
        }
        $str_frontend_backend = "";
        if ( !$str_backend ) {
            $str_frontend_backend .= $str_frontend;
        } else {
            $str_frontend_backend .= $str_frontend . "\n" . $str_backend;
        }
        my $str_frontend_backend_check = $str_frontend_backend;
        if ( $component_method->iter_zapi_isset()
             && $component_method->non_iter_zapi_isset()
             && !grep { $_ =~ /sub\s+_check_non_iter/ } @content ) {
             $str_frontend_backend_check .= "\n" . $component_method->build_c_check_method();
        }
        my $found_index = 0;
        for ( my $i = $#content; $i >= 0; $i-- ) {
            if ( $content[$i] =~ /sub\s+_primary_keys_validate_spec\s*{/ ) {
                $found_index = $i;
                last;
            }
        }
        if ( !$found_index ) {
            for ( my $i = $#content; $i >= 0; $i-- ) {
                if ( $content[$i] =~ /^1;/ ) {
                    $found_index = $i;
                    last;
                }
            }
        }
        splice @content, $found_index, 0, $str_frontend_backend_check;
    }
}

sub update_method_with_missing_options {
    my $self = shift;
    my ( $frontend_hash, $existing_keys ) = @_;
    my $p4_client_lib_root       = $self->nate_lib();
    my $file                     = $self->filename();
    my @component_ds             = $self->component_ds();
    my @primary_keys             = $self->primary_keys();
    my $path                     = "$p4_client_lib_root/NACL/C/$file\.pm";
    my ( @array, $api_index, $backend_option_index, $frontend_option_index,
        $pod_insert_index );
    my ( $i, $j, $k, $l, %primary_keys_hash );

    foreach ( @primary_keys, @$existing_keys ) {
        $primary_keys_hash{$_} = 1;
    }
    tie @array, 'Tie::File', $path, recsep => ';';
    foreach my $comp_method_hashref (@component_ds) {
        my $api_method  = $comp_method_hashref->{command};
        my $method_name = $comp_method_hashref->{method_name};

        #Handle the case of import method
        $method_name = "do_" . $method_name
            if ( $method_name eq 'import' );

        next if ( !$frontend_hash->{$method_name} );
        my $frontend_index = $frontend_hash->{$method_name};
        next
            if ( ( $method_name eq 'show' ) or ( $method_name eq 'status' ) );
        if ( $comp_method_hashref->{is_CS} ) {
            my $filename = $self->return_cslike_filename($method_name);
            $self->build_or_update_cslike_state_file( $filename, $method_name,
                $comp_method_hashref );
            next;
        }

        $backend_option_index = $frontend_option_index = $pod_insert_index
            = 0;
        my $backend_method_name = "_" . $method_name . "_cmode_cli";
        my $backend_index       = 0;
        for ( my $m = $frontend_index; $m < $#array; $m++ ) {
            if ( $array[$m] =~ /sub\s+$backend_method_name\s*{/s ) {
                $backend_index = $m;
                last;
            }
        }

        #If frontend and backend is not found
        #dont attempt to update
        if ( $frontend_index != 0 && $backend_index != 0 ) {
            my $fields = $comp_method_hashref->{fields};
            my ( @array_list, @optional_scalars );
            while ( my ( $field_name, $field_hashref ) = each %$fields ) {
                if ( $field_hashref->{is_array} ) {
                    push @array_list, $field_name;
                } else {
                    if ( !$comp_method_hashref->{can_use_cvw} ) {
                        push @optional_scalars, $field_name;
                    } elsif (
                        !grep {
                            $_ eq $field_name
                        } @primary_keys
                        )
                    {
                        push @optional_scalars, $field_name;
                    }
                }
            }

            #POD insert index is the same as frontend index
            $pod_insert_index = $frontend_index;

            #Traverse  forward to get the frontend validate spec
            for (
                $k = $frontend_index;
                $array[$k] !~ /sub\s+$backend_method_name\s*{/s;
                $k++
                )
            {
                if ( $array[$k] =~ /_common_validate_with/s ) {
                    $frontend_option_index = $k;
                    last;
                }
                if ( $k == $#array ) {
                    $Log->warn(
                        "\n Unable to find frontend options for $method_name in $path"
                    );
                    last;
                }
            }
            my $instant_component_method = 0;
            if ( $array[$frontend_option_index] =~ /api\s+=>/s ) {
                $instant_component_method = 1;
                ## for instant Component the fields should be
                ## checked-in state $additional_spec = {additional_spec => {}}
                ## which is one index lesser than _common_validate_with
                $frontend_option_index = $frontend_option_index -1 ;
            }
            if ( $instant_component_method == 0 ) {

                #Traverse forward to get backend options
                for (
                    $j = $backend_index;
                    $array[$j] !~ /_common_validate_with/s;
                    $j++
                    )
                {
                    if ( $array[$j] =~ /allowed_cmode_options\s+=/s ) {
                        $backend_option_index = $j;
                        last;
                    }
                    if ( $j == $#array ) {
                        $Log->warn( "\n Unable to find backend options for "
                                . " $backend_method_name in $path" );
                        last;
                    }
                }
            }

            #
            # If any of the above three index is zero do not update
            # If the frontend validation spec is returned by a function
            # call do not update
            #
            if (   ( $backend_option_index || $instant_component_method )
                && $frontend_option_index
                && $pod_insert_index
                && $array[$frontend_option_index] !~ /.+\(\)/s )
            {

                #can update the method
                my ( $scalar_str, $array_frontend_str, @new_arrays, $pod_str,
                    $scalar );
                my @new_scalars = ();
                $scalar_str = $array_frontend_str = $pod_str = $scalar = "";
                @new_arrays = ();
                my $pod      = $array[$pod_insert_index];
                my $frontend = $array[$frontend_option_index];
                my $backend  = $array[$backend_option_index];

                my $diff_scalar
                    = _check_scalar_in_array( \@optional_scalars, $frontend );
                if (@$diff_scalar) {
                    $Log->warn(
                        "Unable to update $method_name method in $path ,\n because "
                            . join( " ", @$diff_scalar )
                            . " options has changed from array to scalar" );
                    next;
                }

                my $diff_array
                    = _check_array_in_scalar( \@array_list, $frontend );
                if (@$diff_array) {
                    $Log->warn(
                        "Unable to update $method_name method in $path ,\n because "
                            . join( " ", @$diff_array )
                            . " options has changed from scalar to array" );
                    next;
                }

                foreach (@optional_scalars) {
                    next if ( exists $primary_keys_hash{$_} );
                    if ( $array[$frontend_option_index] !~ /qw.*[(\s]$_[\s)]/s
                        && $array[$frontend_option_index]
                        !~ /['"]?$_['"]?\s*=>\s*{\s*type\s*=>.*?SCALAR.*?,/s
                        && $array[$frontend_option_index]
                        !~ /['"]?$_['"]?\s*=>\s*{\s*type\s*=>.*?BOOLEAN.*?,/s
                        )
                    {
                        push @new_scalars, $_;
                    }
                }

                my %warnings = $self->warnings();
                foreach (@array_list) {

# This is just a workaround to handle an ui bug, where a field in the ui looks like
# an array ,but being listed in keys hash of the cdef file
# Example: In vserver fcp wwpn-alias remove command alias shows up as array ,but
# it is also listed in the keys hash
                    if ( exists $primary_keys_hash{$_} ) {

                        if ( !exists $warnings{$file}{$method_name}
                            || $warnings{$file}{$method_name} ne
                            "has_ui_bug" )
                        {
                            $Log->warn(
                                "$_ option for $method_name method in $path is an array and also the primary\n"
                                    . "key this could be a potential UI bug on ONTAP"
                            );
                        }
                        next;
                    }

                    if ( $array[$frontend_option_index]
                        !~ /['"]?$_['"]?\s*=>\s*{\s*type\s*=>.*?ARRAYREF.*?,/s
                        )
                    {
                        $array_frontend_str .= "\n\"$_\""
                            . " => {type => ARRAYREF, optional => 1},";
                        push @new_arrays, $_;
                        my $array_option_single = $_;
                        if ( $_ =~ /(.*)list$/ ) {
                            $array_option_single = $1;
                        } elsif ( $_ =~ /(.*)s$/ ) {
                            $array_option_single = $1;
                        }
                        $pod_str .= <<"HEREDOC";
=item C<< '$_' => [ \$${array_option_single}1, \$${array_option_single}2, ... ] >>

HEREDOC
                        my $desc_array = $comp_method_hashref->{fields}->{$_}
                            ->{description};
                        foreach my $desc (@$desc_array) {
                            $pod_str .= <<"HEREDOC";
$desc
HEREDOC
                        }
                        $pod_str .= <<"HEREDOC"
(Arrayref, even if only a single value is to be sent it should specified
as an arrayref containing that single value)
Applicable for CMode CLI

HEREDOC
                    }
                }

                if ( scalar @new_scalars ) {
                    foreach (@new_scalars) {
                        $pod_str .= <<"HEREDOC";
=item C<< '$_' => \$string >>

HEREDOC
                        my $desc_array = $comp_method_hashref->{fields}->{$_}
                            ->{description};
                        foreach my $desc (@$desc_array) {
                            $pod_str .= <<"HEREDOC";
$desc
HEREDOC
                        }
                        $pod_str .= <<"HEREDOC";
Applicable for CMode CLI

HEREDOC
                    }
                    my @split_array = _determine_split_list(@new_scalars);
                    foreach my $line (@split_array) {
                        $scalar_str .= <<"HEREDOC";
            $line
HEREDOC
                    }

                }
                #Update scalar list in frontend spec
                if ($scalar_str) {
                    if ( $frontend =~ /(.*qw\s*\()(.*?)(\).*)$/s ) {
                        $frontend = $1 . $2 . "\n$scalar_str" . $3;
                        $array[$frontend_option_index] = $frontend;
                    } elsif (
                        $frontend =~ /(.+additional_spec\s*=>\s*{)(.+$)/s )
                    {
                        my $prematch  = $1;
                        my $postmatch = $2;
                        $scalar_str =~ s/^\s*//;
                        $scalar
                            = "_optional_scalars( qw( " . $scalar_str . ")),";
                        $frontend = $prematch . $scalar . $postmatch;
                        $array[$frontend_option_index] = $frontend;
                    } elsif ( $frontend =~ /(.*)(\)$)/s ) {
                        my $prematch  = $1;
                        my $postmatch = $2;
                        $scalar_str =~ s/^\s*//;
                        $prematch .= "," if ( $prematch !~ /,$/s );
                        $scalar
                            = $prematch
                            . " additional_spec => {_optional_scalars( qw( "
                            . $scalar_str . ")) },";
                        $frontend = $scalar . $postmatch;
                        $array[$frontend_option_index] = $frontend;
                    } else {
                        $Log->warn(
                            "Unable to add $array_frontend_str to $method_name method in $path"
                        );
                    }
                }

                #Update arrayref list in frontend spec
                if ($array_frontend_str) {
                    if ( $frontend =~ /(.*additional_spec\s*=>\s*{)(.+$)/s ) {
                        $frontend = $1 . "$array_frontend_str" . $2;
                        $array[$frontend_option_index] = $frontend;
                    } elsif ( $frontend =~ /(.*)(\)$)/s ) {
                        my $prematch  = $1;
                        my $postmatch = $2;
                        $prematch .= "," if ( $prematch !~ /,$/ );
                        $frontend
                            = $prematch
                            . " additional_spec => {"
                            . $array_frontend_str
                            . $postmatch;
                        $array[$frontend_option_index] = $frontend;
                    } else {
                        $Log->warn(
                            "Unable to add $array_frontend_str to $method_name method in $path"
                        );
                    }

                }

                #Update backend option
                if ( $instant_component_method == 0 ) {
                    if ($scalar_str) {
                        if ( $backend =~ /(.*)\).*]/s ) {
                            $backend = $1 . "\n$scalar_str" . ")]";
                            $array[$backend_option_index] = $backend;
                        } else {
                            $Log->warn(
                                "Unable to add $scalar_str to $backend_method_name method in $path"
                            );
                        }
                    }

                    if (@new_arrays) {
                        my $array_list_string = join ' ', @new_arrays;
                        if ( $backend =~ /(.*)\).*]/s ) {
                            $backend = $1 . "\n$array_list_string" . ")]";
                            $array[$backend_option_index] = $backend;
                        } else {
                            $Log->warn( "Unable to add $array_list_string to"
                                    . "$backend_method_name method in $path"
                            );
                        }
                        my $backend_method_str .= <<"HEREDOC";
    foreach my \$array_element (qw($array_list_string)) {
HEREDOC
                        $backend_method_str .= <<'HEREDOC';
        if ( defined $opts{$array_element} ) {
                    $opts{$array_element} = join ',', @{$opts{$array_element}};
                }
        }
HEREDOC
                        my $hash_copy_index = 0;
                        for (
                            my $a = $backend_option_index;
                            $array[$a] !~ /sub/;
                            $a++
                            )
                        {
                            if ( $array[$a] =~ /_hash_copy/ ) {
                                $hash_copy_index = $a;
                                last;
                            }
                            if ( $a == $#array ) {
                                last;
                            }
                        }
                        if ( $hash_copy_index != 0 ) {

                            #
                            # The hash_copy string can span multiple lines
                            # It needs to insert the string just before the
                            # _hash_copy
                            #
                            my $hash_copy_str = $array[$hash_copy_index];
                            my @hash_copy_array
                                = split( /\n/, $hash_copy_str );
                            my $insert_index = 0;
                            for (
                                my $index = 0;
                                $index < $#hash_copy_array;
                                $index++
                                )
                            {
                                if ($hash_copy_array[$index] =~ /_hash_copy/ )
                                {
                                    $insert_index = $index;
                                    last;
                                }
                            }
                            splice @hash_copy_array, $insert_index, 0,
                                $backend_method_str;
                            my $new_hash_copy_str
                                = join( "\n", @hash_copy_array );
                            $array[$hash_copy_index] = $new_hash_copy_str;
                        } else {
                            $Log->warn(
                                "Unable to update $path::$backend_method_name "
                            );
                        }
                    }
                }

                #Update POD
                if ($pod_str) {
                    if ( $pod =~ /(.*?)(=item C<< '\*.*)/s ) {
                        $pod = $1 . $pod_str . $2;
                        $array[$pod_insert_index] = $pod;
                    } elsif ( $pod =~ /(.*?)(=back.*)/s ) {
                        $pod = $1 . $pod_str . $2;
                        $array[$pod_insert_index] = $pod;
                    } else {
                        $Log->warn(
                            "Unable to update the POD for $method_name method in $path"
                        );
                    }
                }
            }
        }
    }
}

sub add_event_class {
    my $self = shift;
    my $str = _load_event_class();
    my $C_FH = $self->C_FH();
    print $C_FH $str;
}

sub build_c_primary_keys_validate_spec {
    my $self = shift;
    my $str;
    my @primary_keys = $self->primary_keys();
    my @optional_primary_keys = $self->optional_primary_keys();
    if (@primary_keys
        || @optional_primary_keys ) {
        $str = <<"HEREDOC";
sub _primary_keys_validate_spec {
	return (
HEREDOC

        foreach my $primary_key (@primary_keys) {
            $str .= <<"HEREDOC";
        '$primary_key' => { type => SCALAR },
HEREDOC
        }

        foreach my $optional_primary_key (@optional_primary_keys) {
            $str .= <<"HEREDOC";
        '$optional_primary_key' => { type => SCALAR, optional => 1 },
HEREDOC
        }
        $str .= <<"HEREDOC";
\t);
}

HEREDOC

        my $C_FH = $self->C_FH();
        print $C_FH $str;
    }
}

sub build_c_return_true {
    my $self = shift;
    my $str  = <<'HEREDOC';
1;
HEREDOC

    my $C_FH = $self->C_FH();
    print $C_FH $str;
}

sub natelint {
    my $self     = shift;
    my $C_file   = $self->C_package_name();
    my $nate_lib = $self->nate_lib();
    $C_file =~ s/::/\//g;
    $C_file .= '.pm';
    my $ret = NACL::Tools::CGT::CGTUtils::natelint(
        nate_lib           => $nate_lib,
        file_path_from_lib => $C_file
    );
    return $ret;
}

sub perltidy {
    my $self = shift;
    my $C_file_path
        = $self->nate_lib() . '/NACL/C/' . $self->filename() . '.pm';
    NACL::Tools::CGT::CGTUtils::perltidy( file_path => $C_file_path );
}

sub _check_array_in_scalar {
    my $array1   = shift;
    my $frontend = shift;
    my @found    = ();
    foreach (@$array1) {
        if ( $frontend =~ /qw.*[(\s]$_[\s)]/s ) {
            push @found, $_;
        }

        if ( $frontend =~ /['"]?$_['"]?\s*=>\s*{\s*type\s*=>(.*?,)/s ) {
            my $type = $1;
            push @found, $_ if ( $type !~ /ARRAYREF/ );
        }
    }
    return \@found;
}

sub _check_scalar_in_array {
    my $array1   = shift;
    my $frontend = shift;
    my @found    = ();
    foreach (@$array1) {
        if ( $frontend =~ /['"]?$_['"]?\s*=>\s*{\s*type\s*=>\s*(.*?,)/s ) {
            my $type = $1;

            # handle the case of SCALAR|ARRARYREF or SCALAR or BOOLEAN
            push @found, $_ if ( $type !~ /SCALAR/ && $type !~ /BOOLEAN/ );
        }
    }
    return \@found;
}

sub return_cslike_filename {
    my $CS_filename;
    my $self        = shift;
    my $method_name = shift;
    my $filename    = $self->filename();
    my @temp        = split( /_+/, $method_name );
    map { $CS_filename .= ucfirst $_ } @temp;
    return "$filename" . $CS_filename;
}

#
# This method generates or updates cslike state files
# Pass : filename, methodname and reference to component method hashref
#

sub build_or_update_cslike_state_file {
    my $self = shift;
    my ( $filename, $method_name, $comp_method_hashref, $zapi_command ) = @_;
    my $CS_package_name = 'NACL::CS::' . $filename;
    my $cs_file         = NACL::Tools::CGT::ComponentStateFile->new(
        nate_lib        => $self->nate_lib(),
        filename        => $filename,
        is_CS_like      => 1,
        C_method_name   => $method_name,
        CS_package_name => $CS_package_name,
        C_package_name  => $self->C_package_name(),
        ds              => $comp_method_hashref,
        is_singleton    => $comp_method_hashref->{is_singleton},
        command         => $comp_method_hashref->{command},
        component_name  => $self->component_name(),
        should_natelint => $self->should_natelint,
        should_perltidy => $self->should_perltidy,
    );
    my $fields  = $comp_method_hashref->{fields};
    my @options = keys %$fields;
    $cs_file->fields(@options);
    _set_zapi_on_object($cs_file, $comp_method_hashref);
    $cs_file->build_cs_package();
}

sub write_custom_new {
    my $self = shift;
    my $str;
    $str = <<"HEREDOC";

sub new {
    my (\$pkg, \%opts) = \@_;
    my \$self = bless {}, \$pkg;
    while (my (\$key, \$value) = each \%opts) {
        \$self->\$key(\$value);
    }
    \$self->SUPER::init(\%opts);
    return \$self;
}
HEREDOC
    my $C_FH = $self->C_FH();
    print $C_FH $str if $str;
}

1;

__END__

=head1 NAME

NACL::Tools::CGT::ComponentDS

=head1 DESCRIPTION

This class stores the attributes required to build components and contains
methods which build the component libraries. (This library is to be used
only by CGT)

=head1 ATTRIBUTES

=over

=item C<< component_ds >>

This is the data structure built by WalkUI for each component (array-reference).
It contains multiple hash references, each of which contain details for the
various methods of the component.

=item C<< contains_CS_file >>

This is a boolean which specifies whether there is a CS file for this 
component.

=item C<< primary_keys >>

(Array)
This stores the primary keys of the component.

=item C<< component_name >>

The name of the component being generated. (Example: for the directory
"antivirus on-demand command", the component name is "AntivirusOnDemandCommand")

=item C<< C_package_name >>

The package name of the component. For component "AntivirusOnDemandCommand", its
package name is "NACL::C::AntivirusOnDemandCommand"

=item C<< C_FH >>

The filehandle for the file into which the component is being written.

=back

=cut
