#
# Copyright (c) 2001-2011 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary A class representing a single component method
## @author madhav.shekar@netapp.com, dl-nacl-dev@netapp.com
## @status shared
## @pod here

package NACL::Tools::CGT::ComponentMethod;

use strict;
use warnings;
use NATE::Log qw(log_global);
use Data::Dumper qw(Dumper);
use NACL::Tools::CGT::CGTUtils
    qw(_determine_split_list _convert_to_underscored_form
    _supported_zapis _get_check_method _get_choices _determine_cli_only_fields);
use XML::Simple;

use Class::MethodMaker [
    scalar => 'method_name',
    scalar => 'package_var',
    array  => 'options',
    array  => 'array_list',
    array  => 'optional_scalars',
    scalar => 'create_object_flag',
    hash   => 'comp_method_hashref',
    hash   => 'cs_method_hashref',
    scalar => 'api_method',
    array  => 'primary_keys',
    array  => 'optional_primary_keys',
    scalar => 'package_name',
    scalar => 'filename',
    scalar => 'filehandle',
    scalar => 'generate_method_alone',
    scalar => 'generate_pod_alone',
    scalar => 'vserver_context',
    scalar => 'iter_zapi',
    scalar => 'non_iter_zapi',
    scalar => 'nate_lib',
    array  => 'cli_only_fields'
];
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();

#This method defines the attributes to be stored in
#object of type NACL::Tools::CGT::ComponentMethod

sub new {
    my ($pkg, %opts) = @_;
    my $self = bless {}, $pkg;

    ##populate cli_only_fields
    my @cli_fields = _determine_cli_only_fields($opts{comp_method_hashref}->{fields}) ;
    # if ZAPI runs in vserver_context, do not include 'vserver' in invalid_fields.
    if($opts{comp_method_hashref}->{vserver_context} ){@cli_fields = grep { $_  !~ /^vserver$/ } @cli_fields; }
    $self->cli_only_fields(@cli_fields);

    while (my ($key, $value) = each %opts) {
        $self->$key($value);
    }
    return $self;
}

#
#This method generates the frontend method .If filehandle is passed ,
#it writes the generated content to the file,else it returns  the content
#as a string
#
#If generate_method_alone is set 1 in the object ,it would generate the
#frontend method alone , no POD, this option will be used for instant component
#
#If generate_pod_alone is set 1 in the object ,it would generate the
#POD alone , no method, this option will be used for instant component
#

sub build_c_frontend_method {
    my $self                = shift;
    my $method_name         = $self->method_name();
    my $package_var         = $self->package_var();
    my $create_object_flag  = $self->create_object_flag();
    my $comp_method_hashref = $self->comp_method_hashref();
    my $cs_method_hashref   = $self->cs_method_hashref();
    my $is_static           = $comp_method_hashref->{is_static};
    my $fields              = $comp_method_hashref->{fields};
    my $api_method          = $self->api_method();
    my $zapi_api            = $comp_method_hashref->{'zapi-command'};
    my $command             = $api_method;
    $command =~ s/_/ /g;
    my (%options, %primary_keys_hash, $pod);
    my $str     = "";
    my @options = $self->options();

    foreach my $option (@options) {
        $options{$option} = 1;
    }

    my @primary_keys = $self->primary_keys();
    my @optional_primary_keys = $self->optional_primary_keys();
    foreach my $primary_key (@primary_keys) {
        $primary_keys_hash{$primary_key} = 1;
    }

    my @optional_scalars = $self->optional_scalars();
    my @array_list       = $self->array_list();

    my $package_name          = $self->package_name();
    my $generate_method_alone = $self->generate_method_alone();
    my $generate_pod_alone    = $self->generate_pod_alone();
    if (!$generate_method_alone) {
        $str .= <<"HEREDOC";
=head2 $method_name

HEREDOC

        if ($create_object_flag) {
            my $file = $self->filename();
            $str .= <<"HEREDOC";
    my \$${file}_obj	= $package_name->$method_name(
HEREDOC
        } else {
            $str .= <<"HEREDOC";
    $package_name->$method_name(
HEREDOC
        }

        $str .= <<"HEREDOC";
                                command_interface => \$ci,
HEREDOC

        foreach my $primary_key (@primary_keys) {
            if (!$is_static
                || ($is_static && exists $options{$primary_key})) {
                my $hyphenated_form = $primary_key;
                my $underscored_primary_key
                    = _convert_to_underscored_form($hyphenated_form);
                $str .= <<"HEREDOC";
	                            '$primary_key' => \$$underscored_primary_key,
HEREDOC
            }
        }
        $str .= <<"HEREDOC";
                                ...);
HEREDOC

        if ($is_static) {
            $str .= <<"HEREDOC";
   
(Class method)
HEREDOC
        } else {
            my $file = $self->filename();
            $str .= <<"HEREDOC";
	                             
								 or

    \$${file}_obj->$method_name(...);

(Class or instance method)
HEREDOC
        }

        if ( $comp_method_hashref->{description} ) {
            my $description = lcfirst($comp_method_hashref->{description});
            $str .= <<"HEREDOC";
This method is used to $description.
HEREDOC
        }

        if ( $self->iter_zapi_isset() 
            || $self->non_iter_zapi_isset()) {
            $str .= <<"HEREDOC";
Uses CMode CLI/ZAPI .
HEREDOC
            if ( $self->iter_zapi_isset() ) {
                my $iter =  $self->iter_zapi();
                $str .= <<"HEREDOC";
Invokes  $iter for CMode Zapi iter
HEREDOC
            }
            if ( $self->non_iter_zapi_isset() ) {
                my $non_iter = $self->non_iter_zapi();
                $str .= <<"HEREDOC";
Invokes  $non_iter  for CMode Zapi non-iter
HEREDOC
            }
        } else {
            $str .= <<"HEREDOC";
Uses CMode CLI.

HEREDOC
        }

        $str .= <<"HEREDOC";

=over

=item Options

=over

=item C<< command_interface => \$command_interface >>

HEREDOC

        if ($is_static) {
            $str .= <<"HEREDOC";
(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

HEREDOC
        } else {
            $str .= <<"HEREDOC";
(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

HEREDOC
        }
        $str .= <<'HEREDOC';
=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

HEREDOC

        foreach my $primary_key (@primary_keys) {
            if (!$is_static
                || ($is_static && exists $options{$primary_key})) {
                my $hyphenated_form = $primary_key;
                my $underscored_primary_key
                    = _convert_to_underscored_form($hyphenated_form);
                $str .= <<"HEREDOC";

=item C<< '$primary_key' => \$$underscored_primary_key >>

HEREDOC
                if ($is_static) {
                    $str .= <<"HEREDOC";
(Required)
HEREDOC
                } else {
                    $str .= <<"HEREDOC";
(Required for class method, Not Applicable for instance method)
HEREDOC
                }

                my $desc = $comp_method_hashref->{fields}->{$primary_key}
                    ->{description};
                foreach (@$desc) {
                    $str .= <<"HEREDOC";
$_
HEREDOC
                }
                $pod = $self->_pod_for_option($comp_method_hashref->{fields}->{$primary_key});
                $str .= <<"HEREDOC" 
$pod
HEREDOC
            }
        }
        foreach my $optional_scalar (@optional_scalars) {
            if (  !length $optional_scalar
                || exists $primary_keys_hash{$optional_scalar}) {
                next;
            }
            $str .= <<"HEREDOC";

=item C<< '$optional_scalar' => \$string >>

HEREDOC
            if ($comp_method_hashref->{fields}->{$optional_scalar}
                ->{description}) {
                my $desc = $comp_method_hashref->{fields}->{$optional_scalar}
                    ->{description};
                foreach (@$desc) {
                    $str .= <<"HEREDOC";
$_
HEREDOC
                }
            }

            $pod = $self->_pod_for_option($comp_method_hashref->{fields}->{$optional_scalar});
            $str .= <<"HEREDOC"
$pod
HEREDOC
        }

        foreach my $array_option (@array_list) {
            my $array_option_single = $array_option;
            if ($array_option =~ /(.*)list$/) {
                $array_option_single = $1;
            } elsif ($array_option =~ /(.*)s$/) {
                $array_option_single = $1;
            }
            $str .= <<"HEREDOC";

=item C<< '$array_option' => [ \$${array_option_single}1, \$${array_option_single}2, ... ] >>

HEREDOC

            if ($comp_method_hashref->{$array_option}->{description}) {
                foreach (@$comp_method_hashref->{fields}->{$array_option}
                    ->{description}) {
                    $str .= <<"HEREDOC";
$_
HEREDOC
                }
            }
            $pod = $self->_pod_for_option($comp_method_hashref->{fields}->{$array_option});
            $str .= <<"HEREDOC";
(Arrayref, even if only a single value is to be sent it should specified
as an arrayref containing that single value)
$pod
HEREDOC
        }

        $str .= <<"HEREDOC";

=item C<< '*' >>

This method takes all other options supported by "$command" command in ONTAP

=item C<< other_options >>

This method takes other common options described in L<Common method options|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#Other_options_valid_on_component_methods>

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

HEREDOC

        if ($generate_pod_alone) {
            return $str;
        }
    }
    my $vserver_context = $self->vserver_context();
    my $exception_name;

    if ( $method_name eq "create" ) {
       $exception_name = "AlreadyExists";
    } elsif ( $method_name eq "delete" ) {
       $exception_name = "DoesNotExist";
    }

    $str .= <<"HEREDOC";
sub $method_name {
    \$Log->enter() if \$may_enter;
	my ($package_var, \@args) = \@_;
        $package_var->_frontend_log_debug_opts(\@args);
HEREDOC
     if (@optional_scalars or @array_list) {
        $str .= <<'HEREDOC';
        state $additional_spec = {
        additional_spec => {
HEREDOC

        foreach my $array_elem (@array_list) {
            $str .= <<"HEREDOC";
            '$array_elem' => { type => ARRAYREF, optional => 1 },
HEREDOC
        }
        if (scalar @optional_scalars) {
            my @split_array = _determine_split_list(@optional_scalars);

            my $first_elem = shift @split_array;
            $str .= <<"HEREDOC";
                        _optional_scalars( qw( $first_elem
HEREDOC
            foreach my $line (@split_array) {
                $str .= <<"HEREDOC";
            $line
HEREDOC
            }

            $str .= ") ),";
        }

        $str .= '} 
           };
        ' ;

    }

    $str .= <<"HEREDOC";
	my \%opts = $package_var->_common_validate_with(
	    params => \\\@args,
HEREDOC
    if(@optional_scalars or @array_list){
      $str .= <<"HEREDOC";
            %{\$additional_spec},
HEREDOC
    }

    unless ($comp_method_hashref->{can_use_cvw}) {
        $str .= <<"HEREDOC"
       ignore_primary_keys => 1,
HEREDOC
    }
    my $cli_cmd = $comp_method_hashref->{actual_cmd};
    $str .= <<"HEREDOC";
        cli_cmd => \'$cli_cmd\',
	);
HEREDOC

    if ($vserver_context) {
        $str .= <<"HEREDOC";
        my \$vserver;
        \$opts{vserver_operated_on} = \\\$vserver;
HEREDOC
    }
    my $choices = _get_choices( $self, $method_name);
    $str .= <<"HEREDOC";
        $package_var->call_on_apiset_with_exceptions(
	    \%opts,
		choices => [
                            $choices
		],
HEREDOC
    if ( $exception_name ) {
        $self->_generate_exception_class()
            if ($self->nate_lib_isset());
        $str .= <<"HEREDOC";
            nacl_exceptions => [qw($exception_name)],
HEREDOC
    }
    $str .= <<"HEREDOC";
        frontend => '${package_name}::${method_name}',
	);
	
HEREDOC
    if ($create_object_flag) {
        $str .= <<'HEREDOC';
        my %obj_opts =  (
            command_interface => $opts{command_interface},
HEREDOC

        foreach my $primary_key (@primary_keys) {
            my $hyphenated_form = $primary_key;
            my $underscored_primary_key
                = _convert_to_underscored_form($hyphenated_form);
            if (   $vserver_context
                && $underscored_primary_key eq "vserver") {
                $str .= <<"HEREDOC";
        $underscored_primary_key => \$vserver,
HEREDOC
            } else {
                $str .= <<"HEREDOC";
        $underscored_primary_key => \$opts{'$primary_key'},
HEREDOC
            }
        }
        $str .= <<'HEREDOC';
        );
HEREDOC
        foreach my $optional_primary_key (@optional_primary_keys) {
            my $hyphenated_form = $optional_primary_key;
            my $underscored_primary_key
                = _convert_to_underscored_form($hyphenated_form);
            $str .= <<"HEREDOC";
            \$obj_opts{$underscored_primary_key} = \$opts{'$optional_primary_key'}
                 if ( \$opts{'$optional_primary_key'} );
HEREDOC
        }
    $str .= <<'HEREDOC';
        my $obj = $pkg->new(%obj_opts);
        $Log->exit() if $may_exit;
        return $obj;
HEREDOC
    } else {
        $str .= <<'HEREDOC';
       $Log->exit() if $may_exit;
HEREDOC
    }

    if ($method_name eq 'rename' && $package_var eq '$pkg_or_obj') {
        my (@primary_keys_accepted, @remaining_options);
        foreach my $option (@options) {
            if (exists $primary_keys_hash{$option}) {
                push @primary_keys_accepted, $option;
            } else {
                push @remaining_options, $option;
            }
        }
        if (@primary_keys_accepted != 1 || @remaining_options != 1) {
            my $all_pk         = join ',', @primary_keys;
            my $pk_accepted    = join ',', @primary_keys_accepted;
            my $remaining_opts = join ',', @remaining_options;
            $Log->comment("Command '$api_method' generated a 'rename' "
                    . 'method but it could not be determined which primary '
                    . "key was to be renamed.\n"
                    . 'Primary keys are: '
                    . $all_pk
                    . 'Primary keys accepted are: '
                    . $pk_accepted
                    . 'Remaining options accepted are: '
                    . $remaining_opts);
        } else {
            my $primary_key   = $primary_keys_accepted[0];
            my $rename_option = $remaining_options[0];

            $primary_key = _convert_to_underscored_form($primary_key);
            $str .= <<"HEREDOC";
    $package_var->${primary_key}( \$opts{'$rename_option'} ) if ref( $package_var );
HEREDOC
        }
    }
    $str .= <<'HEREDOC';
}
HEREDOC

    my $FH = $self->filehandle();
    if ($FH) {
        print $FH $str;
    } else {
        return $str;
    }
} ## end sub build_c_frontend_method

#
#This method generates the backend method .If filehandle is passed ,
#it writes the generated content to the file,else it returns  the content
#as a string
#
sub build_c_backend_method {
    my $self = shift;
    my (%primary_keys_hash);
    my $method_name         = $self->method_name();
    my $package_var         = $self->package_var();
    my @options             = $self->options();
    my $comp_method_hashref = $self->comp_method_hashref();
    my $create_object_flag = $self->create_object_flag();
    my $api_method         = $self->api_method();

    # Backend method
    my $backend_method_str = <<"HEREDOC";
sub _${method_name}_cmode_cli {
    \$Log->enter() if \$may_enter;
	my ($package_var, \@opts) = \@_;
            $package_var->_base_backend_cmode_cli(
                \@opts,
                api     => \'$api_method\',
            );
    \$Log->exit() if \$may_exit;
}
HEREDOC
    if ($self->iter_zapi_isset()) {
        $backend_method_str .= $self->_generate_iter_zapi();
    }
    if ($self->non_iter_zapi_isset()) {
        $backend_method_str .= $self->_generate_non_iter_zapi();
    }

    my $FH = $self->filehandle();
    if ($FH) {
        print $FH $backend_method_str;
    } else {
        return $backend_method_str;
    }
} ## end sub build_c_backend_method

#
#This method generates the cs like frontend  method .If filehandle is passed ,
#it writes the generated content to the file,else it returns  the content
#as a string
#
#If generate_method_alone is set 1 in the object ,it would generate the
#frontend method alone , no POD, this option will be used for instant component
#
#If generate_pod_alone is set 1 in the object ,it would generate the
#POD alone , no method, this option will be used for instant component
#
#
sub build_cs_like_frontend_method {
    my $self = shift;

    my $package       = $self->package_name();
    my $method        = $self->method_name();
    my $method_obj    = "${method}_obj";
    my $method_objs   = $method_obj . 's';
    my $filename      = shift;
    my $CS_link       = "lib-NACL-CS-$filename-pm";
    my $CS_string     = "NACL::CS::" . $filename;
    my @component     = split(/::/, $package);
    my $component_obj = $component[$#component];
    my $str           = "";

    if (!$self->generate_method_alone()) {
        $str = <<"HEREDOC";
=head2 $method

    my \$$method_obj = $package->$method(command_interface => \$ci,
                          [ filter => { \%filter } ,]
			  [ requested_fields => [ \@req_fields ]);
						  
                    or
					
    my \@$method_objs = $package->$method(command_interface => \$ci,
                          [ filter => { \%filter } ,]
			  [ requested_fields => [ \@req_fields ]);

                    or
  
   my \$$method_obj = \$${component_obj}_obj->$method(); 
						  
(Class or Instance method)

This method is a CS-like fetch method which accepts all the parameters of 
L<fetch|lib-NACL-ComponentState-pm/fetch> (requested_fields, filter, allow_empty) 
and has all of the other properties of fetch methods (callable in scalar or list 
context, throws a NoElementsFound exception if no matching elements are found).
The requested fields and filter takes L<ATTRIBUTES|$CS_link/ATTRIBUTES>
of $CS_string as input. 
This method returns objects of the type L<$CS_string|$CS_link>.

=cut

HEREDOC
    }

    if ($self->generate_pod_alone()) {
        return $str;
    }

    $str .= <<"HEREDOC";

sub $method {
    \$Log->enter() if \$may_enter;
	my (\$pkg_or_obj, \@args) = \@_;
        my \%opts = \$pkg_or_obj->_handle_obj_cs_like_calls(\@args);
	require $CS_string;
	my \@state_objs = $CS_string->fetch(\%opts);
	\$Log->exit() if \$may_exit;
	return wantarray ? \@state_objs : \$state_objs[0];
}

HEREDOC

    my $C_FH = $self->filehandle();
    if ($C_FH) {
        print $C_FH $str;
    } else {
        return $str;
    }
} ## end sub build_cs_like_frontend_method

#
# This function takes field hashref and it returns
# the POD for the field( in non iter POD format)
#
sub _pod_for_c_option_non_iter {
    my $field = shift;
    my $str;
    if ($field->{'zapi-mapping'}) {
        if ($field->{'zapi-mapping'} ne $field->{field}) {
            $str = "Mapping for non-iter: $field->{'zapi-mapping'}";
        }
    }
    return $str;
} ## end sub _pod_for_c_option_non_iter

#
# This function takes field hashref and it returns
# the POD for the field( in  iter POD format)
#
sub _pod_for_c_option_iter {
    my $field = shift;
    local $XML::Simple::PREFERRED_PARSER = 'XML::Parser';
    my $str;
    my $typedef = $field->{'zapi-typedef-info-pod'};
    if ($typedef) {
        my $xs = XML::Simple->new(
            ForceArray => 1,
            KeepRoot   => 1,

            #  ForceContent => 1
        );
        my $xml = $xs->XMLout($typedef);
        $xml =~ s/\s+$//g;
        $str = "Mapping for iter-zapi: $xml";
    }
    return $str;
} ## end sub _pod_for_c_option_iter

#
# This function takes component method DS as input and it generates
# the backend for non iter zapi
#
sub _generate_non_iter_zapi {
    my $self                = shift;
    my $comp_method_hashref = $self->comp_method_hashref();
    my $method_name         = $self->method_name();
    my $package_var         = $self->package_var();
    my $zapi_api            = $self->non_iter_zapi();
    my (@copy, %map_hash, $copy_str, $map_str);
    my %fields = %{$comp_method_hashref->{fields}};
    foreach my $field (keys %fields) {
        if ($fields{$field}->{'zapi-mapping'}) {
            if ($fields{$field}->{'zapi-mapping'} eq $field) {
                push @copy, $field;
            } else {
                $map_hash{$field} = $fields{$field}->{'zapi-mapping'};
            }
        }
    }
    if (@copy) {
        my @split_array = _determine_split_list(@copy);
        $copy_str = join("\n", @split_array);
        $copy_str = "[qw(" . $copy_str . ")];";
    }
    if (keys %map_hash) {
        foreach my $key (keys %map_hash) {
            $map_str .= <<"HEREDOC";
 \'$key\' => \'$map_hash{$key}\',
HEREDOC
        }
        $map_str = "{" . $map_str . "};";
    }
    my $cs_method_hashref = $self->cs_method_hashref();
    my $backend_method_str = <<"HEREDOC";
sub _${method_name}_cmode_zapi_non_iter {
    \$Log->enter() if \$may_enter;
      my ($package_var, \@opts) = \@_;
HEREDOC
    $backend_method_str .= "my \$copy = $copy_str "
        if ($copy_str);

    $backend_method_str .= "my \$map = $map_str "
        if ($map_str);
    $backend_method_str .= <<"HEREDOC";
            $package_var->_base_backend_cmode_zapi(
                \@opts,
                api     => \'$zapi_api\',
HEREDOC
    $backend_method_str .= "copy => \$copy, "
        if ($copy_str);
    $backend_method_str .= "map => \$map, "
        if ($map_str);
    if ( $comp_method_hashref->{zapi_value_conversion} ) {
        my $convert_str = _get_zapi_convert_str($comp_method_hashref->{zapi_value_conversion});
        $backend_method_str .= $convert_str
    }
    $backend_method_str .= <<"HEREDOC";
            );
    \$Log->exit() if \$may_exit;
}
HEREDOC
    return $backend_method_str;
} ## end sub _generate_non_iter_zapi

#
# This function takes component method DS as input and it generates
# the backend for iter zapi
#
sub _generate_iter_zapi {
    my $self                = shift;
    my $comp_method_hashref = $self->comp_method_hashref();
    my $method_name         = $self->method_name();
    my $package_name        = $self->package_name();
    $package_name =~ s/.*::(.*)$/$1/g;
    my $package_var       = $self->package_var();
    my $zapi_api          = $self->iter_zapi();
    my $cs_method_hashref = $self->cs_method_hashref();
    my (@copy, %map_hash, $copy_str, $map_str);
    my %fields    = %{$comp_method_hashref->{fields}};
    my %cs_fields = %{$cs_method_hashref->{fields}};

    foreach my $field (keys %cs_fields) {
        my $str = $cs_fields{$field}->{'zapi-mapping'};
        if ($str) {
            if ($str ne $field) {
                $str = "'". $str ."'"
                   if($str !~ /,/);
                $str =  "'".$field."' => ".$str;
                $map_str .= " $str,";
            } else {
                push(@copy, $str);
            }
        }
    }
    if (@copy) {
        my @split_array = _determine_split_list(@copy);
        $copy_str = join("\n", @split_array);
        $copy_str = "[qw(" . $copy_str . ")];";
    }
    if ($map_str) {
        $map_str = "{" . $map_str . "};";
    }
    my $exception_pkg
        = "NACL::C::Exceptions::" . $package_name . "::DoesNotExist";
    my $backend_method_str = <<"HEREDOC";
sub _${method_name}_cmode_zapi {
    \$Log->enter() if \$may_enter;
     my ($package_var, \@opts) = \@_;
     require $exception_pkg;
HEREDOC
    $backend_method_str .= "my \$copy = $copy_str "
        if ($copy_str);
    $backend_method_str .= "my \$map = $map_str "
        if ($map_str);
    my $show_cmd = $cs_method_hashref->{actual_cmd};
    $backend_method_str .= <<"HEREDOC";
            $package_var->_iter_helper(
                \@opts,
                api     => \'$zapi_api\',
                show_cmd => \'$show_cmd\',
HEREDOC
    $backend_method_str .= "copy => \$copy, "
        if ($copy_str);
    $backend_method_str .= "map => \$map, "
        if ($map_str);
    if ( $cs_method_hashref->{zapi_value_conversion} ) {
       my $convert_str = _get_zapi_convert_str($cs_method_hashref->{zapi_value_conversion});
       $backend_method_str .= $convert_str;
    }
    $backend_method_str .= <<"HEREDOC";
         does_not_exist_exception => \'$exception_pkg\',
            );
    \$Log->exit() if \$may_exit;
}
HEREDOC
    return $backend_method_str;
} ## end sub _generate_iter_zapi

sub _generate_exception_class {
    my $self = shift;
    my ( $class_name, $exception_file, $base_exception, $pod_str );
    my $pkg_name = $self->package_name();
    $pkg_name =~ s/.*::(.*)$/$1/g;
    my $p4_client_path      = $self->nate_lib();
    my $comp_method_hashref = $self->comp_method_hashref();
    my $method_name = $comp_method_hashref->{method_name};
    my $element_name = $comp_method_hashref->{command};
    $element_name =~ m/(.*)_/g;
    $element_name = $1;
    $element_name =~ s/_/ /g;
    my $exception_dir  = "$p4_client_path/NACL/C/Exceptions/$pkg_name";
    if ( $method_name eq 'delete' ) {
        $base_exception = "DoesNotExist";
        $exception_file = $exception_dir . "/DoesNotExist.pm";
        $class_name = "NACL::C::Exceptions::" . $pkg_name . "::DoesNotExist";
        $pod_str = "does not exist";
    } elsif ( $method_name eq 'create' ) {
        $base_exception = "AlreadyExists";
        $exception_file = $exception_dir . "/AlreadyExists.pm";
        $class_name = "NACL::C::Exceptions::" . $pkg_name . "::AlreadyExists";
        $pod_str = "already exists";
    } 
    if (!-e $exception_file) {
        mkdir $exception_dir
            if !-d $exception_dir;
        open(FH, '>', $exception_file) or die $!;
        my $str = <<"HEREDOC";
#
# Copyright (c) 2014 NetApp Inc.
# All rights reserved.
#
## \@summary An exception for, when an attempt is made to operate on $element_name element
## which is non-existent.
##
## \@author dl-nacl-dev\@netapp.com
## \@status Shared
## \@pod here

=head1 NAME

$class_name

=head1 DESCRIPTION

This exception is thrown when a attempt is made to perform an operation on a $element_name element
which $pod_str .

=cut

package $class_name;


use Tharn qw(\);
use NATE::Log qw(log_global);
my \$Log = log_global();
\$may_enter = \$Log->may_enter();
\$may_exit  = \$Log->may_exit();
use base qw(
    NACL::C::Exceptions::$base_exception
);
1;
HEREDOC
        print FH $str;
        NACL::Tools::CGT::CGTUtils::perltidy(file_path => $exception_file);
        $Log->warn(
            "\n $exception_file is a newly generated file do p4 add on it");
    }
} ## end sub _generate_exception_class

sub _pod_for_option {
    my ( $self, $option ) = @_;
    my $str;
    if ( !$self->non_iter_zapi_isset() 
         && !$self->iter_zapi_isset() ) {
         $str = "Applicable for CMode CLI";
         return $str;
     } else {
         $str = "Applicable for CMode CLI/ZAPI";
     }
          
     if ( $self->iter_zapi_isset() ) {
        $str .= " iter"; 
     } 
     if ( $self->non_iter_zapi_isset() ) {
        $str .= ", non-iter."; 
     } 

     if (  $self->iter_zapi_isset() ) {
         my $cs_method_hashref = $self->cs_method_hashref();
         my $field_name = $option->{field};
         my $iter_pod = _pod_for_c_option_iter($cs_method_hashref->{fields}->{$field_name});
         if ($iter_pod) {
             $str .= "\n\n";
             $str .= $iter_pod;
         }
     }

     if (  $self->non_iter_zapi_isset() ) {
         my $non_iter_pod = _pod_for_c_option_non_iter($option);
         if ($non_iter_pod) {
             $str .= "\n\n";
             $str .= $non_iter_pod;
         }
     }
    return $str;
}
sub build_c_check_method {
    my $self = shift;
    my $str = _get_check_method($self);
    my $CS_FH = $self->filehandle();
    if ( $CS_FH ) {
        print $CS_FH $str;
    } else {
        return $str;
    }
}
sub _get_zapi_convert_str {
    my $convert_ref = shift;
    my $convert_str;
    while( my ($key, $value) = each %{ $convert_ref } ) {
        my @values = _determine_split_list(@{ $value });
        my $str  = join("\n", @values ) ;
        $convert_str .= $key."=> [qw(" . $str . ")],";
    }
    $convert_str = "zapi_field_translations => {". $convert_str . "},";
    return $convert_str;
}

1;
