#
# Copyright (c) 2001-2011 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary A class representing a CS file
## @author madhav.shekar@netapp.com, dl-nacl-dev@netapp.com
## @status shared
## @pod here

package NACL::Tools::CGT::ComponentStateFile;

use strict;
use warnings;
use NATE::Log qw(log_global);
use Data::Dumper qw(Dumper);
use NACL::Tools::CGT::CGTUtils
    qw(_determine_split_list _convert_to_underscored_form
    _remove_trailing_newlines _convert_hashrefs_array
    _get_check_method _get_choices _determine_cli_only_fields);
use Tie::File;
use NATE::BaseException qw(:try);
use XML::Simple;

use Class::MethodMaker [
    scalar => 'nate_lib',
    scalar => 'filename',
    scalar => 'filehandle',
    scalar => 'is_CS_like',
    scalar => 'C_method_name',
    scalar => 'CS_package_name',
    scalar => 'C_package_name',
    hash   => 'ds',
    hash   => 'warnings',
    array  => 'fields',
    scalar => 'is_singleton',
    scalar => 'command',
    scalar => 'component_name',
    scalar => 'should_natelint',
    scalar => 'should_perltidy',
    scalar => 'generate_method_alone',
    scalar => 'iter_zapi',
    scalar => 'non_iter_zapi',
    array  => 'cli_only_fields'
];
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();


#This method defines the attributes to be stored in
#object of type NACL::Tools::CGT::ComponentStateFile

sub new {
    my ($pkg, %opts) = @_;
    my $self = bless {}, $pkg;

    ##populate cli_only_fields
    my @cli_fields = _determine_cli_only_fields($opts{ds}->{fields}) ;
    $self->cli_only_fields(@cli_fields) ;

    while (my ($key, $value) = each %opts) {
        $self->$key($value);
    }
    return $self;
}

sub build_cs_package {
    my $self               = shift;
    my $p4_client_lib_root = $self->nate_lib();
    my $file               = $self->filename();
    my $path               = "$p4_client_lib_root/NACL/CS/$file.pm";
    if (!-e $path) {
        open(my $CS_FH, '>', $path) or die $!;
        $self->filehandle($CS_FH);
        if ($self->is_CS_like()) {
            $Log->comment("Writing the CS-like file NACL::CS::$file");
            $self->build_cs_like_header();
            $self->build_cs_like_name_desc_attr();
        } else {
            $Log->comment("Writing the CS file NACL::CS::$file");
            $self->build_cs_header();
            $self->build_cs_name_desc_attr();
        }
        $self->build_attributes_includes();
        $self->build_cs_new();
        my $ds = $self->ds();
        if ( $ds->{extra_filter_fields} ) {
            $self->generate_extra_filter_fields();
        }
        $self->build_cs_check_method();
        $self->build_cs_fetch_frontend();
        my $value_convert_method = $self->build_update_state_obj_method();
        if ( $value_convert_method ) {
            print $CS_FH $value_convert_method;
        } 
        $self->build_cs_fetch_backend();
        $self->perltidy() if $self->should_perltidy();
        close $CS_FH;
        $Log->warn(
            "$path is a newly generated file, be sure to " . "p4 add it");
        $self->natelint() if $self->should_natelint();
    } else {
        if (-w $path) {
            $self->perltidy() if $self->should_perltidy();
            $self->add_missing_statefields();
        } else {
            $Log->warn("$path is not editable ");
        }
    }
} ## end sub build_cs_package

#
#This method will add the missing state fields in the CS file
#

sub add_missing_statefields {
    my $self                = shift;
    my @state_fields        = $self->fields();
    my $show_fields_hashref = $self->ds();
    my $p4_client_lib_root  = $self->nate_lib();
    my $file                = $self->filename();
    my $path                = "$p4_client_lib_root/NACL/CS/$file.pm";
    my (@missing_in_pod, @missing_in_methodmaker, $starts_with_digit);
    my ($str1, $str2, $underscored_state_option, $pod, $field_type);
    my $fields = {};
    tie my @array, 'Tie::File', $path;

    _remove_trailing_newlines(\@array);

    my $cur_cs_file = $self->CS_package_name();
    try {
        eval "require $cur_cs_file";
    }
    otherwise {
        $Log->exit() if $may_exit;
        NATE::Exceptions::Argument->throw(
            'Could not import package ' . "'$cur_cs_file'");
    };

    $fields = $cur_cs_file->get_attributes();
    foreach my $state_option (@state_fields) {
        my $hyphenated_state_option = $state_option;
        next unless (length $hyphenated_state_option);
        ($underscored_state_option, $pod)
            = _handle_state_options($hyphenated_state_option);
        my $is_array
            = $show_fields_hashref->{fields}->{$state_option}->{is_array};
        my $type = $is_array ? 'array' : 'scalar';
        my %warnings = $self->warnings();
        if (defined $fields
            && $fields->{$underscored_state_option}) {
            if (!(defined $warnings{$file} && $warnings{$file}{is_singleton})
                && defined $fields->{$underscored_state_option}
                && $type ne $fields->{$underscored_state_option}) {
                $Log->warn(
                    "Type Conflict: $underscored_state_option is $fields->{$underscored_state_option} "
                        . "in $path ,it should be $type");
            }
        } else {
            if (   $underscored_state_option =~ /(.*)_reset$/
                && defined $fields
                && $fields->{$1}) {
                $Log->warn(
                    "$underscored_state_option of type $type, need to be added "
                        . " manually before $1 state field in $cur_cs_file");
                next;
            } elsif ( $warnings{$file}{show}{$underscored_state_option} eq "do_not_add" ) {
                next; 
            }
            $str1 = "$type => \'$underscored_state_option\',";
            push @missing_in_methodmaker, $str1;
            my $str2 = "";
            $str2 .= <<"HEREDOC";
=item C<< $underscored_state_option >>

HEREDOC
            if ($show_fields_hashref->{fields}->{$hyphenated_state_option}
                ->{description}) {
                my $desc = $show_fields_hashref->{fields}
                    ->{$hyphenated_state_option}->{description};
                foreach my $desc (@$desc) {
                    $str2 .= <<"HEREDOC";
$desc
HEREDOC
                }
            }

            if ($is_array) {
                $field_type = "array";
                $str2 .= <<"HEREDOC";
(Array)
Note that for array fields, the accessor method can be invoked in either
scalar or list context.

my \$$underscored_state_option = \$obj->$underscored_state_option();
 
# \$$underscored_state_option contains a reference to the array of values
 
my \@$underscored_state_option = \$obj->$underscored_state_option();
 
# \@$underscored_state_option contains the array of values
 
If this field needs to be passed to the filter hash,
the value for this field should be passed in as an arrayref

# filter => { $underscored_state_option = [ value1, value2...] }
HEREDOC
            } else {
                $field_type = "scalar";
            }

            if ($pod) {
                $str2 .= <<"HEREDOC";
$pod
HEREDOC
            }

            $str2 .= <<"HEREDOC";

Filled in for CMode CLI.
HEREDOC
            push @missing_in_pod, $str2;
            push @missing_in_pod, "\n";
            $fields->{$underscored_state_option} = $field_type;
        }
    }

    if (@missing_in_methodmaker) {
        my $last;
        for (my $i = 0; $i < $#array; $i++) {
            if ($array[$i] =~ /^=back/) {
                my $j = $i - 1;
                splice @array, $j, 0, '', @missing_in_pod;
                $last = $i;
                last;
            }
        }

        # Don't start iterating through the array from the beginning
        # Simply continue from where the "=back" tag was found
        for (my $i = $last; $i < $#array; $i++) {
            if ($array[$i] =~ /^];/) {
                for (my $j = $i;; $j--) {

                    # Walk backwards through the array till we find
                    # a field definition
                    if ($array[$j] =~ /=>\s*['"]\w+['"]/) {

                       # Check if it has a comma at the end, append one if not
                        $array[$j] .= ',' if ($array[$j] !~ /,/);
                        last;
                    }
                }
                splice @array, $i, 0, @missing_in_methodmaker;
                last;
            }
        }
        $self->perltidy() if $self->should_perltidy();
        $self->natelint() if $self->should_natelint();
    }
} ## end sub add_missing_statefields

sub build_cs_header {
    my $self = shift;
    my $file = $self->filename();
    my $str2 = <<"HEREDOC";
# \$Id\$

#
# Copyright (c) 2001-2014 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## \@summary $file ComponentState Module (auto-generated by CGT)
## \@author dl-nacl-dev\@netapp.com
## \@status shared
## \@pod here

HEREDOC

    my $CS_FH = $self->filehandle();
    print $CS_FH $str2;
} ## end sub build_cs_header

sub build_cs_like_header {
    my $self = shift;
    my $file = $self->filename();
    my $method
        = $self->C_package_name() . '->' . $self->C_method_name() . '()';
    my $str2 = <<"HEREDOC";
# \$Id\$

#
# Copyright (c) 2001-2014 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## \@summary ComponentState Module for the method $method (auto-generated by CGT)
## \@author dl-nacl-dev\@netapp.com
## \@status shared
## \@pod here

HEREDOC

    my $CS_FH = $self->filehandle();
    print $CS_FH $str2;
} ## end sub build_cs_like_header

sub build_cs_name_desc_attr {
    my $self                    = shift;
    my $hyphenated_package_name = $self->C_package_name();
    $hyphenated_package_name =~ s/::/-/g;
    $hyphenated_package_name .= '-pm';

    my $CS_package_name = $self->CS_package_name();
    my $file            = $self->filename();
    my $package_name    = $self->C_package_name();
    my $str2 .= <<"HEREDOC";
=head1 NAME

$CS_package_name

=head1 DESCRIPTION

C<$CS_package_name> is a derived class of
L<NACL::CS::ComponentState::ONTAP|lib-NACL-CS-ComponentState-ONTAP-pm>. It represents the
state of an ONTAP $file.  A related class is
L<$package_name|lib::$hyphenated_package_name>, which represents access
to an ONTAP $file.

=head1 ATTRIBUTES

The individual pieces of data that are part of the state of the $file
element are the attributes of the $file ComponentState. 

Additionally, the command_interface used to obtain the ComponentState object is 
also an attribute of the object. This makes it easier to obtain the component
object corresponding to the CS object, using 
L<get_component_instance|lib-NACL-CS-ComponentState-pm/get_component_instance>.

=over

=item command_interface

The command_interface with which the CS object was obtained. (Available in
all CS objects regardless of requested_fields and the mode/interface)

HEREDOC

    my $CS_FH = $self->filehandle();
    print $CS_FH $str2;
} ## end sub build_cs_name_desc_attr

sub build_cs_like_name_desc_attr {
    my $self                    = shift;
    my $hyphenated_package_name = $self->C_package_name();
    $hyphenated_package_name =~ s/::/-/g;
    $hyphenated_package_name .= '-pm';

    my $CS_package_name = $self->CS_package_name();
    my $file            = $self->filename();
    my $method
        = $self->C_package_name() . '->' . $self->C_method_name() . '()';
    my $str2 .= <<"HEREDOC";
=head1 NAME

$CS_package_name

=head1 DESCRIPTION

C<$CS_package_name> is a derived class of
L<NACL::CS::ComponentState::ONTAP|lib-NACL-CS-ComponentState-ONTAP-pm>. Object(s) of this type
are returned when $method is invoked.

((This module does not represent the state of any element, but is an object
repesentation of the output obtained when $method is invoked.)

=head1 ATTRIBUTES

The fields of the output are fields of the ComponentState object.

=over

HEREDOC

    my $CS_FH = $self->filehandle();
    print $CS_FH $str2;
} ## end sub build_cs_like_name_desc_attr

#
#If generate_pod_alone is set 1 in the object ,it would generate the
#POD alone and it returns the generated POD as a string
#This option will be used for instant component
#
sub build_attributes_includes {
    my $self                = shift;
    my $CS_package_name     = $self->CS_package_name();
    my @state_fields        = $self->fields();
    my $show_fields_hashref = $self->ds();
    my $str2;
    my ($underscored_state_option, $pod);

    if (!$self->generate_method_alone()) {
        foreach my $state_option (@state_fields) {
            next unless (length $state_option);
            my $starts_with_digit;
            my $hyphenated_state_option = $state_option;
            ($underscored_state_option, $pod)
                = _handle_state_options($hyphenated_state_option);
            my $is_array
                = $show_fields_hashref->{fields}->{$state_option}{is_array};
            $str2 .= <<"HEREDOC";

=item C<< $underscored_state_option >>

HEREDOC
            if ($show_fields_hashref->{fields}->{$hyphenated_state_option}
                ->{description}) {
                my $desc = $show_fields_hashref->{fields}
                    ->{$hyphenated_state_option}->{description};
                foreach my $desc (@$desc) {
                    $str2 .= <<"HEREDOC";
$desc
HEREDOC
                }
            }

            if ($is_array) {
                $str2 .= <<"HEREDOC";
(Array)
Note that for array fields, the accessor method can be invoked in either
scalar or list context.

my \$$underscored_state_option = \$obj->$underscored_state_option();
 
# \$$underscored_state_option contains a reference to the array of values
 
my \@$underscored_state_option = \$obj->$underscored_state_option();
 
# \@$underscored_state_option contains the array of values
 
If this field needs to be passed to the filter hash,
the value for this field should be passed in as an arrayref

# filter => { $underscored_state_option = [ value1, value2...] }

HEREDOC
            }
            if ($pod) {
                $str2 .= <<"HEREDOC";
$pod
HEREDOC
            }
            my $str = $self->_pod_for_state_field(
                $show_fields_hashref->{fields}->{$hyphenated_state_option});
            $str2 .= <<"HEREDOC";
$str
HEREDOC
        }

        $str2 .= <<"HEREDOC";
=back

=cut

HEREDOC
    }

    $str2 .= <<"HEREDOC";
package $CS_package_name;

use strict;
use warnings;
use Params::Validate qw(validate);
use NATE::Log qw(log_global);

use NACL::Exceptions::NoElementsFound qw(:try);

HEREDOC

    if ($self->iter_zapi_isset()
        || $self->non_iter_zapi_isset()) {
        $str2 .= <<"HEREDOC";
use NACL::CS::ComponentState::ZapiSkip qw(make_zapi_skip);
use NACL::CS::ComponentState::ZapiArray qw(make_zapi_array);
HEREDOC
    }
    $str2 .= <<"HEREDOC";
use base 'NACL::CS::ComponentState::ONTAP';

HEREDOC

    if ($self->generate_method_alone()) {
        return $str2;
    }
    my $CS_FH = $self->filehandle();
    print $CS_FH $str2;
} ## end sub build_attributes_includes

sub build_cs_new {
    my $self = shift;

    my $str2 .= <<"HEREDOC";
use Class::MethodMaker [
HEREDOC

    my @state_fields        = $self->fields();
    my $show_fields_hashref = $self->ds();
    foreach my $state_option (@state_fields) {
        my $hyphenated_state_option = $state_option;
        next unless (length $hyphenated_state_option);
        my ($underscored_state_option,)
            = _handle_state_options($hyphenated_state_option);
        my $is_array
            = $show_fields_hashref->{fields}->{$state_option}->{is_array};
        my $type = $is_array ? 'array' : 'scalar';
        $str2 .= <<"HEREDOC";
\t$type => '$underscored_state_option',
HEREDOC
    }

    $str2 .= <<"HEREDOC";
];
    my \$Log = log_global();
my \$may_enter = \$Log->may_enter();
my \$may_exit  = \$Log->may_exit();
HEREDOC
    if ($self->generate_method_alone()) {
        return $str2;
    } else {
        my $CS_FH = $self->filehandle();
        print $CS_FH $str2;
    }
} ## end sub build_cs_new

sub build_cs_fetch_frontend {
    my $self                 = shift;
    my $CS_package_name      = $self->CS_package_name();
    my $element              = $CS_package_name;
    $element =~ /NACL::CS::(.*)$/;
    $element = $1;
    my $not_matching_element = $element . '(s)';
    my $states = $element . '_states';
    my $state  = $element . '_state';
    my $str2;
    my $is_singleton = $self->is_singleton();

    if (!$self->generate_method_alone()) {
        $str2 .= <<"HEREDOC";
=head1 METHODS

=head2 fetch

    my \$$state  = $CS_package_name->fetch(command_interface => \$ci, ...);
HEREDOC

        if (!$is_singleton) {
            $str2 .= <<"HEREDOC";
	my \@$states = $CS_package_name->fetch(command_interface => \$ci, ...);

HEREDOC
        }

        if ($is_singleton) {
            $str2 .= <<'HEREDOC';
(Class method)
Discovers which elements are present and returns their state in ComponentState
objects. This being a "singleton" table (i.e. one which can only have one
row) is callable only in scalar context.
HEREDOC
        } else {
            $str2 .= <<'HEREDOC';
(Class method)
Discovers which elements are present and returns their state in ComponentState
objects. Called in scalar context it returns only one state object, in list
context it returns all state objects.
HEREDOC
        }

        $str2 .= <<'HEREDOC';
See L<NACL::CS::ComponentState|lib-NACL-CS-ComponentState-pm/fetch> for a
more detailed description along with a complete explanation of the options
it accepts.

HEREDOC
        if ($self->iter_zapi_isset()
            || $self->non_iter_zapi_isset()) {
            $str2 .= <<"HEREDOC";
Uses CMode CLI/ZAPI.

HEREDOC
            if ($self->iter_zapi_isset()) {
                my $iter_zapi = $self->iter_zapi();
                $str2 .= <<"HEREDOC";
Invokes $iter_zapi API for CMode ZAPI iter.

HEREDOC
            }
            if ($self->non_iter_zapi_isset()) {
                my $non_iter_zapi = $self->non_iter_zapi();
                $str2 .= <<"HEREDOC";
Invokes $non_iter_zapi API for CMode ZAPI non-iter.

HEREDOC
            }

        } else {
            $str2 .= <<"HEREDOC";
Uses CMode CLI.

HEREDOC
        }

        $str2 .= <<"HEREDOC";
            
=over

=item Exceptions

=over

=item C<NACL::Exceptions::NoElementsFound>

When there are no elements matching the query specified or elements of that type
doesn't exist, then this exception will be thrown.

=back

=back

=cut

HEREDOC
    }

    my $choices = _get_choices($self, "fetch");
    my $ds = $self->ds();
    my $actual_cmd = $ds->{'actual_cmd'};
    $str2 .= <<"HEREDOC";
sub fetch {
    \$Log->enter() if \$may_enter;
    my (\$pkg, \@args) = \@_;
    my \@state_objs = \$pkg->SUPER::fetch(
        \@args,
        show_cmd => '$actual_cmd',
        choices => [
                    $choices
HEREDOC
    $str2 .= <<"HEREDOC";
        ],
        exception_text => 'No matching $not_matching_element found',
	frontend => '${CS_package_name}::fetch',
HEREDOC

    if ($is_singleton) {
        $str2 .= "is_singleton   => 1",;
    }
    $str2 .= <<'HEREDOC';
	);
    $Log->exit() if $may_exit;
    return wantarray ? @state_objs : $state_objs[0];
}
HEREDOC

    if ($self->generate_method_alone()) {
        return $str2;
    } else {
        my $CS_FH = $self->filehandle();
        print $CS_FH $str2;
    }
} ## end sub build_cs_fetch_frontend

sub build_cs_fetch_backend {
    my $self = shift;
    my ($copy_str, $map_str, @copy);
    my $show                = $self->command();
    my $is_singleton        = $self->is_singleton();
    my $show_fields_hashref = $self->ds();
    my $str3                = <<'HEREDOC';

sub _fetch_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    my @state_objs = $pkg->SUPER::_fetch_cmode_cli(@args, 
HEREDOC
    $str3 .= "api => '$show',";

    if ($is_singleton) {
        $str3 .= "is_singleton => 1,";
    }
    $str3 .= ");";
    $str3 .= <<'HEREDOC';
    $Log->exit() if $may_exit;
	return @state_objs;
}
HEREDOC
    if ($self->iter_zapi_isset()) {
        $map_str = $copy_str = "";
        my $zapi_api = $self->iter_zapi();
        $str3 .= <<'HEREDOC';

sub _fetch_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    my @state_objs = $pkg->SUPER::_fetch_cmode_zapi(@args, 
HEREDOC
    my $copy_map = $self->_get_copy_map_str();
    $str3 .= $copy_map
        if ($copy_map);
    $str3 .= <<"HEREDOC";
    api => "$zapi_api",
    );
    \$Log->exit() if \$may_exit;
    return \@state_objs;
    }
HEREDOC

    }
    if ($self->non_iter_zapi_isset()) {
        my $zapi_api = $self->non_iter_zapi();
        @copy = ();
        $copy_str = $map_str = "";
        $str3 .= <<'HEREDOC';
sub _fetch_cmode_zapi_non_iter {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    my @state_objs = $pkg->SUPER::_fetch_cmode_zapi_non_iter(@args,
HEREDOC
    my $copy_map = $self->_get_copy_map_str();
    $str3 .= $copy_map
        if ($copy_map);
    $str3 .= <<"HEREDOC";
    api => "$zapi_api",
HEREDOC
    if ( $show_fields_hashref->{primary_keys} ) {
        my %pks;
        map { $pks{$_}=1 } @{ $show_fields_hashref->{primary_keys} };
        if ( $show_fields_hashref->{vserver_context}
            && $pks{vserver} ) {
            delete $pks{vserver};
            $str3 .= "vserver_operated_on => \\my \$vserver,"; 
        }
        my $pk_str = join( " ", keys %pks);
        $str3 .= <<"HEREDOC";
        primary_keys => [ qw( $pk_str ) ],
HEREDOC
    }
$str3 .= <<"HEREDOC";
    );
    \$Log->exit() if \$may_exit;
    return \@state_objs;
    }
HEREDOC

    }
    if ($self->generate_method_alone()) {
        return $str3;
    } else {
        $str3 .= <<'HEREDOC';
1;
HEREDOC
        my $CS_FH = $self->filehandle();
        print $CS_FH $str3;
    }
} ## end sub build_cs_fetch_backend

sub natelint {
    my $self     = shift;
    my $file     = $self->CS_package_name();
    my $nate_lib = $self->nate_lib();
    $file =~ s/::/\//g;
    $file .= '.pm';
    my $ret = NACL::Tools::CGT::CGTUtils::natelint(
        nate_lib           => $nate_lib,
        file_path_from_lib => $file
    );
    return $ret;
} ## end sub natelint

sub perltidy {
    my $self = shift;
    my $file_path
        = $self->nate_lib() . '/NACL/CS/' . $self->filename() . '.pm';
    NACL::Tools::CGT::CGTUtils::perltidy(file_path => $file_path);
} ## end sub perltidy

sub _handle_state_options {
    my $hyphenated_state_option = shift;
    my ($starts_with_digit, $fetch_attribute, %hash, $str2);
    my $underscored_state_option
        = _convert_to_underscored_form($hyphenated_state_option);

    if ($underscored_state_option =~ /^\d/) {
        $starts_with_digit        = 1;
        $underscored_state_option = "_" . "$underscored_state_option";
    }

    if ($underscored_state_option =~ /^fetch$/) {
        $fetch_attribute          = 1;
        $underscored_state_option = "c_fetch";
    }

    if ($starts_with_digit) {
        $str2 .= <<"HEREDOC";
This is the accessor for the field '$hyphenated_state_option'. Since it starts
with a digit, we cannot have an accessor of the same name (perl doesn't
allow subroutine whose names start with a digit). So the convention we
follow is to have the accessor methods for such fields start with an underscore.
HEREDOC
    }

    if ($fetch_attribute) {
        $str2 .= <<"HEREDOC";
This attribute stores the value of the "fetch" field that's present
in the show output. Note that there is no "c-fetch" field present
in the output, this attribute has been renamed so that there is
no clash with the C<fetch> method.
HEREDOC
    }

    return ($underscored_state_option, $str2);
} ## end sub _handle_state_options

sub _generate_pod_for_iter {
    my $field = shift;
    my $pod;
    local $XML::Simple::PREFERRED_PARSER = 'XML::Parser';
    my $typedefinfo = $field->{'zapi-typedef-info-pod'};
    if ($typedefinfo) {
        my $xs = XML::Simple->new(
            ForceArray => 1,
            KeepRoot   => 1,

            #  ForceContent => 1
        );
        my $xml = $xs->XMLout($typedefinfo);
        $xml =~ s/\s+$//g;
        $pod = <<"HEREDOC";
Mapping for CMode ZAPI iter: $xml
HEREDOC
    }
    return $pod;
} ## end sub _generate_pod_for_iter

sub _generate_pod_for_non_iter {
    my $field = shift;
    my ($pod, $map);
    my $typedefinfo = $field->{'zapi-typedef-info-pod'};
    if ($typedefinfo) {
        my @keys = keys %{$typedefinfo};
        $map = $typedefinfo->{$keys[0]};
        my @elem = _convert_hashrefs_array($map);
        if (   @elem == 1
            && $field->{field} ne $elem[0]) {
            $pod = <<"HEREDOC";
Mapping for CMode ZAPI non-iter: $elem[0]
HEREDOC
        }
        if (@elem > 1) {
            my $xs = XML::Simple->new(
                ForceArray => 1,
                KeepRoot   => 1,
            );
            my $xml = $xs->XMLout($map);
            $xml =~ s/\s+$//g;
            $pod = <<"HEREDOC";
Mapping for CMode ZAPI non-iter: $xml
HEREDOC
        }
    }
    return $pod;
} ## end sub _generate_pod_for_non_iter

sub _pod_for_state_field {
    my ($self, $option) = @_;
    my $str;
    if (   !$self->non_iter_zapi_isset()
        && !$self->iter_zapi_isset()) {
        $str = "Filled in for CMode CLI";
        return $str;
    } else {
        $str = "Filled in for CMode CLI/ZAPI";
    }
    if ($self->iter_zapi_isset()) {
        $str .= " iter";
    }
    if ($self->non_iter_zapi_isset()) {
        $str .= ", non-iter.";
    }
    if ($self->iter_zapi_isset()) {
        my $iter_pod = _generate_pod_for_iter($option);
        if ($iter_pod) {
            $str .= "\n\n";
            $str .= $iter_pod;
        }
    }
    if ($self->non_iter_zapi_isset()) {
        my $non_iter_pod = _generate_pod_for_non_iter($option);
        if ($non_iter_pod) {
            $str .= "\n";
            $str .= $non_iter_pod;
        }
    }
    return $str;
} ## end sub _pod_for_state_field

sub build_cs_check_method {
    my $self = shift;
    my $str = _get_check_method($self) ;
    my $CS_FH = $self->filehandle();
    if ( $CS_FH ) {
        print $CS_FH $str;
    } else {
        return $str;
    }
}

sub build_update_state_obj_method {
    my $self = shift;
    my $show_fields_hashref = $self->ds();
    my $convert_str = "";
    my $method_str;
    if ( $show_fields_hashref->{zapi_value_conversion} ) {
       my %convert = %{ $show_fields_hashref->{zapi_value_conversion} };
       while( my ($key, $value) = each %convert ) {
           my @values = _determine_split_list(@{ $value });
           my $str  = join("\n", @values ) ;
           $convert_str .= $key."=> [qw(" . $str . ")],";
       }
       $method_str = <<'HEREDOC';
sub _update_state_objs_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->SUPER::_update_state_objs_cmode_zapi(
        @args,
        zapi_field_translations => {
HEREDOC
    $method_str .= <<"HEREDOC";
$convert_str
        },
    );
    \$Log->exit() if \$may_exit;
}
HEREDOC
    }
    return $method_str;
}

sub _get_copy_map_str {
    $Log->enter() if $may_enter;
    my ($self) = @_;
    my $show_fields_hashref = $self->ds();
    my %fields = %{$show_fields_hashref->{fields}};
    my ($copy_str, $map_str, @copy, $copy_map);
    if ( $show_fields_hashref->{extra_filter_fields} ) {
        %fields = (%fields, %{ $show_fields_hashref->{extra_filter_fields} });
    }
    foreach my $field (keys %fields) {
        my $str = $fields{$field}->{'zapi-mapping'};
        if ($str) {
            if ($str ne $field) {
                $str = "'$str'"
                    if($str !~ /,/);
                $str =  "'$field' => ".$str;
                $map_str .= "$str,";
            } else {
                push(@copy, $str);
            }
        }
    }
    if (@copy) {
        my @split_array = _determine_split_list(@copy);
        foreach my $line (@split_array) {
            $copy_str .= <<"HEREDOC";
$line
HEREDOC
        }
        $copy_str = "[qw(" . $copy_str . ")],";
        $copy_map .= " copy => $copy_str";
    }
    if ( $map_str) {
        $map_str = "{" . $map_str . "},";
        $copy_map .= " map => $map_str";
    }
    $Log->exit() if $may_exit;
    return $copy_map;
}

sub generate_extra_filter_fields {
    my $self = shift;
    my ($copy_str, $map_str, @copy);
    my $show_fields_hashref = $self->ds();
    my $extra_filter_fields=$show_fields_hashref->{extra_filter_fields};
    my @fields = keys %{ $extra_filter_fields };
    my $fields_str = join(" ", @fields);
    my $method = <<"HEREDOC";
sub _extra_filter_fields {
    \$Log->enter() if \$may_enter;
    \$Log->exit() if \$may_exit;
    return [ qw( $fields_str ) ];
}
HEREDOC
    my $CS_FH = $self->filehandle();
    if ( $CS_FH ) {
        print $CS_FH $method;
    } else {
        return $method;
    }
}

1;
