=pod 

=head1  NAME

NACL::Tools::httpd::control

=head1 DESCRIPTION

Provides start(), check(), stop() methods to control NACL::Tools::httpd

=head1 EXAMPLES

my $httpd = NACL::Tools::httpd::control->new();

my $url = $httpd->start(
            hostrec => $client,
            peer    => $opts{command_interface}->hostrec()->default_ip(),
            gets    => { "/$basename" => $opts{package} },
            );

$httpd->check();

$httpd->stop();

=cut

package NACL::Tools::httpd::control;

use strict;
use warnings;

use Hostrec ();
use NATE::Inc qw(find_file);
use Params::Validate qw(validate SCALAR HASHREF ARRAYREF UNDEF CODEREF);
use Class::MethodMaker [
    new    => [ '-hash', 'new' ],
    scalar => [
        {   '-default_ctor' => sub {
                find_file( "NACL/Tools/httpd/http_server.pl",
                    search => [ grep { !m@^http://@ } @INC ] )
                    or die $@;
            },
        },
        'httpd_path'
    ],
    scalar => [
        {
            '-default_ctor' => sub {
                find_file( "NACL/Tools/httpd/http_proxy.pl",
                    search => [ grep { !m@^http://@ } @INC ] )
                    or die $@;
            },
        },
        'http_proxy_path'
    ],

    scalar => [ { '-type' => 'Connectrec' }, 'conn' ],

    scalar => 'url',
];

sub start {
    &start_httpd;
}

sub start_httpd {
    my $self = shift;
    my %opts = validate @_,
        {
         hostrec => { isa  => "Hostrec" },
         peer    => { type => SCALAR },
         gets    => { type => HASHREF, default => {} },
         puts    => { type => HASHREF, default => {} },
         posts   => { type => HASHREF, default => {} },
         verbose => { type => SCALAR, default => Tharn::param("NACL_VERBOSE_HTTP_SERVER"), optional => 1 },
        };
    my @args;
    if ($opts{verbose}) {
        push @args, "-v";
    }
    foreach my $pair (
        [ "GET",  $opts{gets} ],
        [ "PUT",  $opts{puts} ],
        [ "POST", $opts{posts} ],
      )
    {
        my ( $request, $optval ) = @$pair;
        while ( my ( $key, $value ) = each %$optval ) {
            push @args, ( $request, $key, $value );
        }
    }

    return $self->_start_process(
        hostrec => $opts{hostrec},
        command => [
            $self->httpd_path,
            "-peer" => $opts{peer},
            @args,
        ],
    );
}

sub start_proxy {
    my $self = shift;
    my %opts = validate @_,
      {
        peer       => { type => SCALAR },
        repository => { type => SCALAR },
        verbose => { type => SCALAR, default => Tharn::param("NACL_VERBOSE_HTTP_PROXY"), optional => 1 },
      };
    my @args = (
                "-peer"       => $opts{peer},
                "-repository" => $opts{repository},
               );
    if ($opts{verbose}) {
        unshift @args, "-v";
    }
    return $self->_start_process(
        hostrec => Tharn::host("localhost"),
        command => [
            $self->http_proxy_path,
            @args,
        ]
    );
}

sub _start_process {
    my $self = shift;
    my %opts = validate @_,
      {
        hostrec => { isa  => "Hostrec" },
        command => { type => ARRAYREF },
        cwd     => { type => SCALAR | UNDEF, default => undef },
      };

    # Here, we fail one way or another if we see it exit, successfully
    # or otherwise.  What should happen, however, is that it prints
    # out what URL it's listening on, and we capture that and jump out
    # of waiting for any more output (leaving the command running;
    # we'll check up on it later using execute_expect)

    my $hostrec = $opts{hostrec};
    my @connopts = ( template => "login" );
    if ( $hostrec->id() eq "localhost" ) {

        # NATE should do this for exec and login connections to localhost
        @connopts = (
            template => "local",
            conntype => "rsh-persay",
            conncmd  => [ "/bin/sh", "-c" ]
        );
    }
    my $conn = $opts{hostrec}->connect(@connopts);
    $self->conn($conn);
    my $url;

  OUT: {
        my $command = $opts{command};
        $conn->execute(
            command      => $command,
            check_status => 1,
            match_table  => [
                "^URL: (.*)\n" => sub {
                    $url = $1;
                    no warnings qw(exiting);
                    last OUT;
                },
            ],
        );
        die( $command->[0]
              . " was supposed to serve forever, but it completed successfully?\n"
        );
    } ## end OUT:
    $self->url($url);
    return $url;
}

sub check {
    my $self = shift;
    my $conn = $self->conn();
    $conn->execute_expect(
        check_status => 1,
        timeout      => 0,
        on_timeout   => sub { },
    );
    1;
} ## end sub check

# Specific to http_proxy; doesn't make sense for http_server.  Look in
# the output of http_proxy for the next file that it has saved.  If
# there are none, possibly wait for more (based on timeout and
# on_timeout options).  Also, throw an exception if http_proxy exited
# unexpectedly, like "check".

sub wait_for_file {
    my $self = shift;
    my %opts = validate @_,
      {
        timeout    => { type => SCALAR },
        on_timeout => { type => CODEREF },
      };
    my $conn = $self->conn();
    my $new_filename;
  OUT: {
        $conn->execute_expect(
            check_status => 1,
            timeout      => $opts{timeout},
            on_timeout   => $opts{on_timeout},
            match_table  => [
                "Saved file: (.*)\n" => sub {
                    $new_filename = $1;
                    no warnings qw(exiting);
                    last OUT;
                },
            ],
        );
    }
    return $new_filename;
}

sub stop {
    my $self = shift;
    my $conn = $self->conn();
    $conn->send_intr();
    $conn->disconnect();
    $self->conn_reset();
    $self->url_reset();
} ## end sub stop

1;
