# Copyright (c) 2016 NetApp, Inc., All Rights Reserved
# Any use, modification, or distribution is prohibited
# without prior written consent from NetApp, Inc.

## @summary HECR DataSet Unit Tests
## @author archana@netapp.com
## @status shared
## @pod here

package NACL::UnitTest::AppBuilder::DpgWorkflows::HECR;

use strict;
use warnings;

use Params::Validate qw(validate_with :types);
use Data::Dumper;

use Tharn qw(snooze);
use NATE::Log qw(log_global);
use NATE::BaseException qw(:try);
use NATE::Result;

use NACL::AppBuilder::DpgWorkflows::Workload qw(
                        Wf_create_multiplesize_filesnclones
                        DataSet_GrowClones_PunchHoles
                        DataSet_GrowParents_PunchHoles
                        );

# dummy method
sub no_op {
    return 1;
}

=head1 NAME

HECR DataSets UnitTests

=head1 DESCRIPTION

This library contains the unit tests for HECR Data Sets.

=head1 TESTCASES

=head2 HECR_3_n_4

=head3 DEPENDENCIES

None.

=head3 CONFIG

    ntest
    NFS_CLIENTS=foo \
    CONFIG_FILE='lib/NACL/UnitTest/AppBuilder/DpgWorkfows/HECR.yml' \
    lib/NACL/AppBuilder/reIgniter.thpl

=cut

sub HECR_3_n_4 {
    shift; # remove pkg name
    my %args = validate_with(
        params => \@_,
        spec => {
            sm_obj      => {type => ARRAYREF},
            client      => {type => ARRAYREF|SCALAR, }, # nfs or san client depending on file_type
            file_count  => {type => SCALAR, default => 1}, # number of files to modify
            file_type   => {type => SCALAR, regex => qr/(file|lun)/, default => 'file'},
            holes_in_begin => {type => SCALAR, default => 1, regex => qr/0|1/},
            holes_in_end   => {type => SCALAR, default => 0, regex => qr/0|1/},
            holes_in_mid   => {type => SCALAR, default => 0, regex => qr/\d+/},
            hole_size_in_begin => {type => SCALAR, default => '4K'},
            hole_size_in_end   => {type => SCALAR, default => '4K'},
            hole_size_in_mid   => {type => SCALAR, default => '4K'},
            offset_between_holes => {type => SCALAR, default => '4K'}, # should match default-extent-size
            grow_size   => {type => SCALAR, default => '20K'},
            #snapshot_name => {type => SCALAR, default => 's3'},
            snapmirror_label => {type => SCALAR, default => 'weekly'},
        },
    );
    my $log = log_global(1);

    my $sm_obj = @{$args{sm_obj} }[0]; # use first snapmirror
    my ($vserver_src,$vol_src_name) = ($sm_obj->{source_path} =~ /^(.*):(.*)$/);
    $log->comment('Got source vserver:'.$vserver_src.' src vol:'.$vol_src_name);

    my $vol_src_obj = NACL::C::Volume->find(command_interface => $sm_obj->{source_ci},
                                         filter => {
                                                volume  => $vol_src_name,
                                                vserver => $vserver_src,
                                                 },
                                      );
    my @vol_srcs;
    push @vol_srcs, $vol_src_obj;

    my @traffic_client;
    if (ref($args{client}) eq 'ARRAY') {
        @traffic_client = map( NACL::C::Client->new( name => $_ ), @{$args{client}} );
    } else {
        @traffic_client = map( NACL::C::Client->new( name => $_ ), $args{client} );
    }
    my $client_name = $traffic_client[0]->client();

    $log->step('Create HECR dataset s1');
    my $ret = Wf_create_multiplesize_filesnclones(src_vol_obj  => \@vol_srcs,
                                                snapmirror_label => $args{snapmirror_label},
                                                nfsclient        => $client_name, #$args{client}
                                                );
    my $file_objs_ref = $ret->{'file_obj'};
    my $clone_objs_ref = $ret->{'clone_obj'};

    $log->step('Create HECR dataset s3');
    my %dataset_args = %args;
    delete $dataset_args{sm_obj};
    $dataset_args{file_obj} = $clone_objs_ref;
    DataSet_GrowClones_PunchHoles(%dataset_args);

    $log->step('Create HECR dataset s4');
    my %dataset_args2 = %args;
    delete $dataset_args2{sm_obj};
    $dataset_args2{file_obj} = $file_objs_ref;
    DataSet_GrowParents_PunchHoles(%dataset_args);

    $log->log(NATE::Result::Pass->new('PASS'));
}

1;
