
#
# Copyright (c) 2015 NetApp, Inc., All Rights Reserved.
#   Any other use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

package NACL::UnitTest::AppBuilder::Xgen::Config::Lif;

use strict;
use warnings;

use parent qw(NACL::UnitTest::AppBuilder::Xgen::Config::Vserver);

use NACL::AppBuilder::Xgen::Resource::Vserver;
use NACL::AppBuilder::Xgen::Resource::NetworkRoute;
use NACL::AppBuilder::Xgen::Resource::Lif;
use constant ROLE => 'data';
use constant DATA_PROTOCOLS => [ 'iscsi' ];

sub build {
    my ($self, %args) = @_;

    # We only want to do this if something did not pass in any Ethernet fabric definitions.
    # This fabric will include all the ips from all nodes and not sort specific to clusters.
    # This means that IPs will not be owned by a specific cluster and can be used by any cluster.
    unless (scalar keys %{$args{eth_fabric_data}}) {
        $args{eth_fabric_data} = $self->generate_eth_fabric_data_from_natehost_files(%args);
    }

    $self->SUPER::build(%args);

    #Create LIFS
    $self->generate_lif_resources(%args);
}

sub generate_lif_resources {
    my ($self, %args) = @_;

    my $nodes = $self->get_resources(type => "NACL::AppBuilder::Xgen::Resource::Node");

    my $vservers = $self->get_resources(type => "NACL::AppBuilder::Xgen::Resource::Vserver");
    my $vs = $vservers->[0];

    my $eth_fabric_data = $args{eth_fabric_data};

    # make an iscsi lif on each node (on each fabric)
    my $i = 1;
    my $tmp;
    foreach my $node (@{$nodes}) {
        my $node_name = $node->get_attribute('node');
        my $node_id = $self->get_node_id($node_name);

        # obtain the ethenet fabric table for this node
        my $fabric_table = $self->get_fabric_lookup_table_by_node($node_name, 'eth');
        foreach my $fabric (keys %{$fabric_table}) {
            my $address_range_v4 = $eth_fabric_data->{$fabric}->{address_range_v4};
            my $netmask_v4       = $eth_fabric_data->{$fabric}->{netmask_v4};
            if (!$address_range_v4 or !$netmask_v4) {
                NATE::Exceptions::Argument->throw("No address ranges and/or netmasks were provided for fabric '$fabric'");
            }
            # if this fabric has a gateway defined,
            # make a route resource on this vserver if we havn't built one yet for this fabric
            $tmp->{$fabric} //= {};
            if (!$tmp->{$fabric}->{route_resource} and $eth_fabric_data->{$fabric}->{gateway_v4}) {
                my $route_resource = NACL::AppBuilder::Xgen::Resource::NetworkRoute->new(vserver => $vs->get_attribute('vserver'),
                                                                                         gateway => $eth_fabric_data->{$fabric}->{gateway_v4});
                $route_resource->add_parent_resources(resources => [$vs]);
                $self->add_resources(resources => [$route_resource]);
                $tmp->{$fabric}->{route_resource} = $route_resource;
            }

            # choose a port on this fabric for this node
            my $port_name = shift(@{$fabric_table->{$fabric}});
            push(@{$fabric_table->{$fabric}}, $port_name);
            my $port_id = $self->get_port_id($node_name, $port_name);

            # make the lif resource
            my $lif = NACL::AppBuilder::Xgen::Resource::Lif->new(vserver => $vs->get_attribute('vserver'),
                                                                 'role' => $self->ROLE,
                                                                 'data-protocol' => $self->DATA_PROTOCOLS,
                                                                 address_range => $address_range_v4,
                                                                 netmask => $netmask_v4,
                                                                 lif => "lf_n" . $node_id . "_p" . $port_id,
                                                                 'home-node' => $node_name,
                                                                 'home-port' => $port_name);
            $lif->add_parent_resources(resources => [$vs]);
            $self->add_resources(resources => [$lif]);
        }
    }

}

sub generate_eth_fabric_data_from_natehost_files {
    my ($self, %args) = @_;

    my @NODES = NACL::C::Node->find();
    my @ips;
    my $netmask = undef;
    my $gateway = undef;
    foreach my $node (@NODES) {
        my $hostrec = Hostrec->new(id=>$node->node());
        my %networks = $hostrec->param_fields();
        foreach my $key (keys %networks) {
            if($key =~ /DATA_IP/i) {
                push (@ips, $networks{$key});
            } elsif ($key =~ /DATA_GATEWAY/i) {
                $gateway = $networks{$key};
            } elsif ($key =~ /DATA_MASK/i) {
                $netmask = $networks{$key};
            }
        }
    }
    my %eth_fabric_data = ();
    $eth_fabric_data{'eth1'} = {
        'ipspace'           => 'Default',
        'broadcast_domain'  => 'Default',
        'address_range_v4'  => \@ips,
        'netmask_v4'        => $netmask,
        'gateway_v4'        => $gateway,
    };

    return \%eth_fabric_data
}
1;

