#
# Copyright (c) 2015 NetApp, Inc., All Rights Reserved.
#   Any other use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

package NACL::UnitTest::AppBuilder::Xgen::Config::SimpleFcpConfig;

use parent qw(NACL::AppBuilder::Xgen::Config::ONTAP);

# these are all the resources that we are going to create
use NACL::AppBuilder::Xgen::Resource::Vserver;
use NACL::AppBuilder::Xgen::Resource::Volume;
use NACL::AppBuilder::Xgen::Resource::Lun;
use NACL::AppBuilder::Xgen::Resource::LunMap;
use NACL::AppBuilder::Xgen::Resource::Igroup;
use NACL::AppBuilder::Xgen::Resource::Igroup::Initiator;
use NACL::AppBuilder::Xgen::Resource::Fcp;
use NACL::AppBuilder::Xgen::Resource::Lif;

use NATE::Exceptions::Argument;

use constant NUM_FCP_LIFS_PER_NODE_PER_FABRIC => 2;

sub build {
    my ($self, %args) = @_;

    $self->SUPER::build(%args);

    $self->generate_vserver_resources(%args);

    $self->generate_fcp_lif_resources(%args);

    $self->generate_volume_lun_igroup_resources(%args);
}

sub generate_vserver_resources {
    my ($self, %args) = @_;

    my $aggregates = $self->get_resources(type => "NACL::AppBuilder::Xgen::Resource::Aggregate");

    # make a vserver with an fcp service
    my $vs = NACL::AppBuilder::Xgen::Resource::Vserver->new(vserver => "xgen_vserver",
                                                            aggregate => $aggregates->[0]->get_attribute('aggregate'),
                                                            'rootvolume-security-style' => "mixed");
    my $fcp = NACL::AppBuilder::Xgen::Resource::Fcp->new(vserver => "xgen_vserver");

    # make the parent of the vserver be the cluster that this aggregate belongs to
    $vs->add_parent_resources(resources => [$aggregates->[0]->get_ancestors(type => "NACL::AppBuilder::Xgen::Resource::Cluster")]);
    $fcp->add_parent_resources(resources => [$vs]);
    $self->add_resources(resources => [$fcp, $vs]);
}

sub generate_fcp_lif_resources {
    my ($self, %args) = @_;

    my $nodes = $self->get_resources(type => "NACL::AppBuilder::Xgen::Resource::Node");

    my $vservers = $self->get_resources(type => "NACL::AppBuilder::Xgen::Resource::Vserver");
    my $vs = $vservers->[0];

    # create lifs per node per fabric
    foreach (1..$self->NUM_FCP_LIFS_PER_NODE_PER_FABRIC) {

        # make an fcp lifs on each node (on each fabric)
        foreach my $node (@{$nodes}) {
            my $node_name = $node->get_attribute('node');
            my $node_id = $self->get_node_id($node_name);

            # obtain the fcp fabric table for this node
            my $fabric_table = $self->get_fabric_lookup_table_by_node($node_name, 'fcp');

            foreach my $fabric (keys %{$fabric_table}) {
                # choose a port on this fabric for this node
                my $port_name = shift(@{$fabric_table->{$fabric}});
                push(@{$fabric_table->{$fabric}}, $port_name);
                my $port_id = $self->get_port_id($node_name, $port_name);

                # make the lif resource
                my $lif = NACL::AppBuilder::Xgen::Resource::Lif->new(vserver => $vs->get_attribute('vserver'),
                                                                     'data-protocol' => ['fcp'],
                                                                     lif => "lf_n" . $node_id . "_p" . $port_id,
                                                                     'home-node' => $node_name,
                                                                     'home-port' => $port_name);
                $lif->add_parent_resources(resources => [$vs]);
                $self->add_resources(resources => [$lif]);
            }
        }
    }
}

sub generate_volume_lun_igroup_resources {
    my ($self, %args) = @_;

    # make a volume on one aggr, with a lun
    # make an igroup,
    # assign igroup to client
    # map lun to igroup

    my $vservers = $self->get_resources(type => "NACL::AppBuilder::Xgen::Resource::Vserver");
    my $vs = $vservers->[0];

    my $vol = NACL::AppBuilder::Xgen::Resource::Volume->new(vserver => $vs->get_attribute('vserver'),
                                                            volume => "test_vol",
                                                            size => "80g",
                                                            aggregate => $vs->get_attribute('aggregate'));
    $vol->add_parent_resources(resources => [$vs]);
    $self->add_resources(resources => [$vol]);

    my $clients = $self->get_resources(type => "NACL::AppBuilder::Xgen::Resource::Client");

    foreach my $client (@{$clients}) {
        # skip clients that are not doing fcp
        next if (!grep("fcp", @{$client->get_attribute('protocols')}));

        my $client_name = $client->get_attribute('name');
        my $ostype = $client->get_attribute('ostype');

        my $lun = NACL::AppBuilder::Xgen::Resource::Lun->new(vserver => $vs->get_attribute('vserver'),
                                                             lun => "test_lun_" . $client_name,
                                                             volume => "test_vol",
                                                             size => "4g",
                                                             ostype => $ostype);
        my $lunmap = NACL::AppBuilder::Xgen::Resource::LunMap->new(vserver => $vs->get_attribute('vserver'),
                                                                   lun => "test_lun_" . $client_name,
                                                                   volume => "test_vol",
                                                                   igroup => $client_name);

        my $igroup = NACL::AppBuilder::Xgen::Resource::Igroup->new(vserver => $vs->get_attribute('vserver'),
                                                                   igroup => $client_name,
                                                                   protocol => "fcp",
                                                                   ostype => $ostype =~ /windows/ ? "windows" : $ostype);
        my $initiator = NACL::AppBuilder::Xgen::Resource::Igroup::Initiator->new(vserver => $vs->get_attribute('vserver'),
                                                                                 igroup => $client_name,
                                                                                 protocol => "fcp",
                                                                                 client_name => $client_name);

        $lun->add_parent_resources(resources => [$vol]);
        $igroup->add_parent_resources(resources => [$vs]);
        $lunmap->add_parent_resources(resources => [$lun, $igroup]);
        $initiator->add_parent_resources(resources => [$igroup]);
        $self->add_resources(resources => [$lun, $lunmap, $igroup, $initiator]);
    }
}

1;

__END__

=head1 NAME

NACL::UnitTest::AppBuilder::Xgen::Config::SimpleFcpConfig

=head1 AUTHOR/MAINTAINER

=over

=item NACL (dl-nacl-dev@netapp.com)

=back

   Last revised $Date: $ by $Author: $

=cut
