#
# Copyright (c) 2015 NetApp, Inc., All Rights Reserved.
#   Any other use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

package NACL::UnitTest::AppBuilder::Xgen::Config::SimpleiSCSIConfig;

use parent qw(NACL::AppBuilder::Xgen::Config::ONTAP);

# these are all the resources that we are going to create
use NACL::AppBuilder::Xgen::Resource::Vserver;
use NACL::AppBuilder::Xgen::Resource::Volume;
use NACL::AppBuilder::Xgen::Resource::Lun;
use NACL::AppBuilder::Xgen::Resource::LunMap;
use NACL::AppBuilder::Xgen::Resource::Igroup;
use NACL::AppBuilder::Xgen::Resource::Igroup::Initiator;
use NACL::AppBuilder::Xgen::Resource::Iscsi;
use NACL::AppBuilder::Xgen::Resource::Lif;
use NACL::AppBuilder::Xgen::Resource::NetworkRoute;

use NATE::Exceptions::Argument;

sub build {
    my ($self, %args) = @_;

    $self->SUPER::build(%args);

    $self->generate_vserver_resources(%args);
    
    $self->generate_iscsi_lif_resources(%args);
    
    $self->generate_volume_lun_igroup_resources(%args);
}

sub generate_vserver_resources {
    my ($self, %args) = @_;

    my $aggregates = $self->get_resources(type => "NACL::AppBuilder::Xgen::Resource::Aggregate");

    # make a vserver with an iSCSI service
    my $vs = NACL::AppBuilder::Xgen::Resource::Vserver->new(vserver => "xgen_vserver",
                                                            aggregate => $aggregates->[0]->get_attribute('aggregate'),
                                                            'rootvolume-security-style' => "mixed");
    my $iscsi = NACL::AppBuilder::Xgen::Resource::Iscsi->new(vserver => "xgen_vserver");

    # make the parent of the vserver be the cluster that this aggregate belongs to
    $vs->add_parent_resources(resources => [$aggregates->[0]->get_ancestors(type => "NACL::AppBuilder::Xgen::Resource::Cluster")]);
    $iscsi->add_parent_resources(resources => [$vs]);
    $self->add_resources(resources => [$iscsi, $vs]);
}

sub generate_iscsi_lif_resources {
    my ($self, %args) = @_;

    my $nodes = $self->get_resources(type => "NACL::AppBuilder::Xgen::Resource::Node");

    my $vservers = $self->get_resources(type => "NACL::AppBuilder::Xgen::Resource::Vserver");
    my $vs = $vservers->[0];

    my $eth_fabric_data = $args{eth_fabric_data};

    # make an iscsi lif on each node (on each fabric)
    my $i = 1;
    my $tmp;
    foreach my $node (@{$nodes}) {
        my $node_name = $node->get_attribute('node');
        my $node_id = $self->get_node_id($node_name);

        # obtain the ethenet fabric table for this node
        my $fabric_table = $self->get_fabric_lookup_table_by_node($node_name, 'eth');
        foreach my $fabric (keys %{$fabric_table}) {
            my $address_range_v4 = $eth_fabric_data->{$fabric}->{address_range_v4};
            my $netmask_v4       = $eth_fabric_data->{$fabric}->{netmask_v4};
            if (!$address_range_v4 or !$netmask_v4) {
                NATE::Exceptions::Argument->throw("No address ranges and/or netmasks were provided for fabric '$fabric'");
            }
            # if this fabric has a gateway defined, 
            # make a route resource on this vserver if we havn't built one yet for this fabric
            $tmp->{$fabric} //= {};
            if (!$tmp->{$fabric}->{route_resource} and $eth_fabric_data->{$fabric}->{gateway_v4}) {
                my $route_resource = NACL::AppBuilder::Xgen::Resource::NetworkRoute->new(vserver => $vs->get_attribute('vserver'),
                                                                                         gateway => $eth_fabric_data->{$fabric}->{gateway_v4});
                $route_resource->add_parent_resources(resources => [$vs]);
                $self->add_resources(resources => [$route_resource]);
                $tmp->{$fabric}->{route_resource} = $route_resource;
            }

            # choose a port on this fabric for this node
            my $port_name = shift(@{$fabric_table->{$fabric}});
            push(@{$fabric_table->{$fabric}}, $port_name);
            my $port_id = $self->get_port_id($node_name, $port_name);
            
            # make the lif resource
            my $lif = NACL::AppBuilder::Xgen::Resource::Lif->new(vserver => $vs->get_attribute('vserver'),
                                                                 'data-protocol' => ['iscsi'],
                                                                 address_range => $address_range_v4,
                                                                 netmask => $netmask_v4,
                                                                 lif => "lf_n" . $node_id . "_p" . $port_id,
                                                                 'home-node' => $node_name,
                                                                 'home-port' => $port_name);
            $lif->add_parent_resources(resources => [$vs]);
            $self->add_resources(resources => [$lif]);
        }
    }

}

sub generate_volume_lun_igroup_resources {
    my ($self, %args) = @_;

    # make a volume on one aggr, with a lun
    # make an igroup,
    # assign igroup to client
    # map lun to igroup

    my $vservers = $self->get_resources(type => "NACL::AppBuilder::Xgen::Resource::Vserver");
    my $vs = $vservers->[0];

    my $vol = NACL::AppBuilder::Xgen::Resource::Volume->new(vserver => $vs->get_attribute('vserver'),
                                                            volume => "test_vol",
                                                            size => "500m",
                                                            aggregate => $vs->get_attribute('aggregate'));
    $vol->add_parent_resources(resources => [$vs]);
    $self->add_resources(resources => [$vol]);

    my $clients = $self->get_resources(type => "NACL::AppBuilder::Xgen::Resource::Client");

    foreach my $client (@{$clients}) {
        # skip clients that are not doing iSCSI
        next if (!grep("iscsi", @{$client->get_attribute('protocols')}));

        my $client_name = $client->get_attribute('name');
        my $ostype = $client->get_attribute('ostype');

        my $lun = NACL::AppBuilder::Xgen::Resource::Lun->new(vserver => $vs->get_attribute('vserver'),
                                                             lun => "test_lun_" . $client_name,
                                                             volume => "test_vol",
                                                             size => "50m",
                                                             ostype => $ostype);
        my $lunmap = NACL::AppBuilder::Xgen::Resource::LunMap->new(vserver => $vs->get_attribute('vserver'),
                                                                   lun => "test_lun_" . $client_name,
                                                                   volume => "test_vol",
                                                                   igroup => $client_name);

        my $igroup = NACL::AppBuilder::Xgen::Resource::Igroup->new(vserver => $vs->get_attribute('vserver'),
                                                                   igroup => $client_name,
                                                                   protocol => "iscsi",
                                                                   ostype => $ostype =~ /windows/ ? "windows" : $ostype);
        my $initiator = NACL::AppBuilder::Xgen::Resource::Igroup::Initiator->new(vserver => $vs->get_attribute('vserver'),
                                                                                 igroup => $client_name,
                                                                                 protocol => "iscsi",
                                                                                 client_name => $client_name);

        $lun->add_parent_resources(resources => [$vol]);
        $igroup->add_parent_resources(resources => [$vs]);
        $lunmap->add_parent_resources(resources => [$lun, $igroup]);
        $initiator->add_parent_resources(resources => [$igroup]);
        $self->add_resources(resources => [$lun, $lunmap, $igroup, $initiator]);
    }
}

1;

__END__

=head1 NAME

NACL::UnitTest::AppBuilder::Xgen::Config::SimpleiSCSIConfig

=head1 AUTHOR/MAINTAINER

=over

=item NACL (dl-nacl-dev@netapp.com)

=back

   Last revised $Date: $ by $Author: $

=cut
