#
# Copyright (c) 2014 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#


# We don't want to set a package here so that whoever includes these methods is
# in the same package as this file.
#package NACL::UnitTest::MTask::DataSetStandardTests;

use strict;
use warnings;

use Tharn qw($Log);

use NACL::STask::Volume;
use NACL::STask::Snapmirror;
use NATE::BaseException qw(:try);
use NACL::MTask::VolumeFileIterator;
use data_mobility::ondemand::lib::RemoteFile;

BEGIN {
    use Exporter qw(import);
    our %EXPORT_TAGS = (
        all => [
            qw(
                framework_create_find_purge
                framework_full_create
                framework_multiple_instances
                framework_new_create_purge
                framework_static_create_purge
                framework_find_on_mirror
            )
        ]
    );
    our @EXPORT_OK = @{ $EXPORT_TAGS{all} };
}

our ($Framework, $PKG);

sub framework_full_create {
    my $topology = $Framework->topology();
    my $volume = $topology->volume();

    my $before_file_count = $volume->state()->files_used();

    my $dataset = $PKG->new(
        volume => $volume,
        clients => [$topology->clients()],
    );
    $dataset->create();
    $dataset->validate();
    $dataset->purge();

    my $after_file_count = $volume->state()->files_used();

    if ($before_file_count != $after_file_count) {
        $Log->error("Had $before_file_count files before create, and ".
            "$after_file_count files after purge. They should be the same ".
            "if purge is working properly "
        );
    }

    $topology->clean_volume();

    return $TCD::PASS;
}

sub framework_new_create_purge {
    my $topology = $Framework->topology();
    my $volume = $topology->volume();

    my $dataset = $PKG->new(
        volume => $volume,
        clients => [$topology->clients()],
        fast_mode => 1,
    );
    $dataset->create();
    $dataset->validate();
    $dataset->purge();

    $topology->clean_volume();

    return $TCD::PASS;
}

sub framework_static_create_purge {
    my $topology = $Framework->topology();
    my $volume = $topology->volume();

    my $dataset = $PKG->create(
        volume => $volume,
        clients => [$topology->clients()],
        fast_mode => 1,
    );
    $dataset->validate();
    $dataset->purge();

    $topology->clean_volume();

    return $TCD::PASS;
}

sub framework_multiple_instances {
    my $topology = $Framework->topology();
    my $volume = $topology->volume();

    my @datasets;
    foreach (1..2) {
        my $dataset = $PKG->new(
            volume => $volume,
            clients => [$topology->clients()],
            fast_mode => 1,
        );
        $dataset->create();
        push @datasets, $dataset;
    }

    map { $_->validate() } @datasets;
    map { $_->purge() } @datasets;

    $topology->clean_volume();

    return $TCD::PASS;
}

sub framework_create_find_purge {
    my $test = $_[0];
    my $topology = $Framework->topology();
    my $volume = $topology->volume();

    $test->step("Creating the DataSet");
    my $dataset = $PKG->new(
        volume => $volume,
        clients => [$topology->clients()],
        fast_mode => 1,
    );
    $dataset->create();

    $test->step("Validating the DataSet");
    $dataset->validate();

    $test->step("Finding the DataSet");
    $dataset = $PKG->find(
        volume => $volume,
        clients => [$topology->clients()],
    );
    $dataset->validate();

    $test->step("Purging the DataSet");
    $dataset->purge();

    $topology->clean_volume();

    return $TCD::PASS;
}

# We will have to figure out how this works if the DataSet prevents snapmirror
# from initializing. This could include SFOD destinations or files with L1
# corruption. Those kinds of DataSets might be useful for someone but clearly
# won't pass this test.
sub framework_find_on_mirror {
    my $topology = $Framework->topology();
    my $volume = $topology->volume();

    my $dataset = $PKG->new(
        volume => $volume,
        clients => [$topology->clients()],
        fast_mode => 1,
    );
    $dataset->create();
    $dataset->validate();

    my $dp_volume = NACL::STask::Volume->create(
        command_interface => $volume->command_interface(),
        vserver => $volume->vserver(),
        size => $volume->state()->size(),
        type => 'DP',
        nacltask_to_cleanup => 1,
    );
    my $mirror = NACL::STask::Snapmirror->create(
        command_interface => $dp_volume->command_interface(),
        'source-path' => $volume,
        'destination-path' => $dp_volume,
        nacltask_to_cleanup => 1,
    );
    $dp_volume->mount('junction-path' => '/'.$dp_volume->volume());

    $dataset = $PKG->find(
        volume => $dp_volume,
        clients => [$topology->clients()],
    );
    $dataset->validate();
    try {
        $dataset->purge();
        $Log->error("Purging a DataSet from a DP volume should have thrown ".
            "some kind of exception"
        );
    } catch NATE::BaseException with {
        my $ex = shift;
        $Log->comment("Caught exception as expected while purging from a ".
            "DP volume\n".$ex->text()
        );
    };

    $topology->clean_volume();

    return $TCD::PASS;
}

sub framework_files_writeable_from_client {
    my $test = $_[0];
    my $topology = $Framework->topology();
    my $volume = $topology->volume();

    $test->step("Creating the DataSet");
    my $dataset = $PKG->new(
        volume => $volume,
        clients => [$topology->clients()],
        fast_mode => 1,
    );
    $dataset->create();

    foreach my $client ($topology->get_unix_client(), $topology->get_windows_client()) {
        $test->step("Checking for access via client ".$client->client());
        my $file_iterator = NACL::MTask::VolumeFileIterator->new(
            volume => $volume,
        );

        my $remote_file = data_mobility::ondemand::lib::RemoteFile->new(
            client => $client,
        );

        while ($file_iterator->has_next()) {
            my $file = $file_iterator->next();

            my $file_state = $file->state();
            if ($file_state->file_type() !~ qr/^file$/i) {
                $Log->comment("Skipping write test on file ".$file->path().
                    " of type ".$file_state->file_type()
                );
                next;
            }

            # Attempt to write to the file using a client
            my $fh = $remote_file->open(
                file_name => $file,
                access_mode => 'write',
                open_mode => 'open-existing',
            );
            $fh->write(
                offset => 0,
                contents => "A"x4096,
            );
            $remote_file->close(
                file_handle => $fh,
            );
        }
    }

    return $TCD::PASS;
}

1;
