# $Id$

#
# Copyright (c) 2011-2021 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary MetroCluster Unit Tests Utility Module
## @author dl-mcc-dpg@netapp.com
## @status shared
## @pod here

package NACL::UnitTest::MTask::MCC::TestUtils;

use strict;
use warnings;
use Carp;

use Params::Validate
    qw(validate validate_with BOOLEAN HASHREF OBJECT ARRAYREF SCALAR UNDEF);


use Exporter 'import';
use Params::Validate ':all';

BEGIN {

     $SIG{'__WARN__'} = sub {
         if ( $_[0] =~
                  m/^Called UNIVERSAL::isa\(\) as a function, not a method/
              ) {
             return;
         }
         else {
             CORE::warn(@_);
         }
     };

}

our @EXPORT_OK = qw( make_mock_log make_bootstrap_util
  make_cluster_component mock_cs_metrocluster make_mcc_without_init
  make_mock_mcc mock_subtest mock_aggregates mock_many_aggregates
  make_mock_aggregate make_mock_network_node mock_wafliron
  mock_network_interfaces test_use_mcc set_mcc_state_normal
  set_mcc_state_switchover set_mcc_state_heal_aggregates
  set_mcc_state_heal_roots mock_cs_cluster mock_component_state_fetch
  make_dr_coordinator);

our %EXPORT_TAGS = ( 'all' => \@EXPORT_OK );

use Test::MockObject;
use Test::MockObject::Extends;
use Test::MockModule;
use NACL::UnitTest::MTask::MCC::MockObject;

sub test_use_mcc {
    local $SIG{'__WARN__'} = sub {
        CORE::warn( $_[0] )
          unless ( $_[0] =~ m{^Subroutine \s _check_volume_state \s 
                              redefined \s at
                         |
                             ^Unrecognized \s escape \s \\s
                         }x );
    };
    Test::More::use_ok('NACL::MTask::MCC::Metrocluster');
    return;
}


sub make_mock_log {

    my $log = Test::MockObject::Extends->new('NATE::Log');
    $log->fake_new('NATE::Log');
    $log->set_true('comment');
    $log->set_true('debug');
    $log->set_true('reset_path');
    $log->set_true('redirect');

    return $log;
}

sub make_bootstrap_util {

    my $bootstrap_util = Test::MockObject->new();
    $bootstrap_util->fake_module(
        'NACL::MTask::MCC::Metrocluster::_Bootstrap_Util');
    $bootstrap_util->fake_new(
        'NACL::MTask::MCC::Metrocluster::_Bootstrap_Util');
    $bootstrap_util->set_always( 'get_site_A_Metrocluster_component',
        make_cluster_component('A') );
    $bootstrap_util->set_always( 'get_site_B_Metrocluster_component',
        make_cluster_component('B') );

    $bootstrap_util->set_always( 'get_A1_node_names', [qw/foo bar/] );

    $bootstrap_util->set_isa('NACL::MTask::MCC::Metrocluster::_Bootstrap_Util');

    return $bootstrap_util;

}

sub mock_apiset {

    my $apiset = NACL::UnitTest::MTask::MCC::MockObject->new();
    $apiset->set_true('set_privilege');
    $apiset->set_true('clear_privilege');
    return $apiset;

}

sub mock_command_interface {

    my $command_interface = NACL::UnitTest::MTask::MCC::MockObject->new();
    $command_interface->set_isa('NACL::C::Cluster', 'NACL::C::CommandInterface');
    $command_interface->set_false('refresh_command_interface');
    $command_interface->set_always('command_interface', $command_interface);
    $command_interface->set_always('apiset', mock_apiset());

    my $version_manager = NACL::UnitTest::MTask::MCC::MockObject->new();
    $command_interface->set_always('version_manager', $version_manager);

    return $command_interface;

}

sub make_cluster_component {
    my ($site) = @_;

    my $cluster = NACL::UnitTest::MTask::MCC::MockObject->new();
    $cluster->set_isa('NACL::C::Cluster');

    my $ci = mock_command_interface();

    $cluster->set_always( 'command_interface', $ci );

    $ci->set_always('_site', $site);

    return $cluster;
}

sub mock_cs_cluster {

    my %in = validate_with(
        params => \@_,
        spec   => {
            'health' => {
                'type' => SCALAR,
                'default' => 'true'
            }
        }
    );

    my $cs_cluster = NACL::UnitTest::MTask::MCC::MockObject->new();
    $cs_cluster->set_isa('NACL::CS::Cluster');

    $cs_cluster->set_always('health', 'true');

    $cs_cluster->set_always('node', 'mock node');

    my $cs_mod = Test::MockModule->new('NACL::CS::Cluster');

    $cs_mod->mock(
        'fetch',
        sub {
            return $cs_cluster;
        }
    );

    return ($cs_cluster, $cs_mod);
}


sub mock_cs_metrocluster {
    my $cs_mcc = Test::MockModule->new('NACL::CS::Metrocluster');
    $cs_mcc->mock(
        'fetch',
        sub {
            return (
                {
                    configuration_state => 'enabled',
                    dr_operation_state  => 'Normal'
                }
            ) x 4;
        }
    );

    return $cs_mcc;

}

sub make_mcc_without_init {

    my $mcc_module = Test::MockModule->new('NACL::MTask::MCC::Metrocluster');

    $mcc_module->mock(
        'init',
        sub {
            return;
        }
    );

    my $mcc =
      NACL::MTask::MCC::Metrocluster->new( nodes => [qw/foo bar baz bap/] );

    $mcc_module->unmock('init');

    my $mock_mcc = Test::MockObject::Extends->new($mcc);

    $mock_mcc->set_always( '_site_A_metrocluster_component',
        make_cluster_component('A') );

    $mock_mcc->set_always( '_site_B_metrocluster_component',
        make_cluster_component('B') );

    return $mock_mcc;

}

sub make_mock_network_node {

    my %in = validate_with(
        params => \@_,
        spec   => {
            'bridge_output' => {
                'type'     => SCALAR | UNDEF,
                'optional' => 1
            },
            'switch_output' => {
                'type'     => SCALAR | UNDEF,
                'optional' => 1
            },
            'sysconfig_output' => {
                'type'     => ARRAYREF | UNDEF,
                'optional' => 1
            },

        }
    );

    my $netnode
        = Test::MockObject::Extends->new('NACL::MTask::MCC::Network::Node');

    $netnode->set_false('_init_sysconfig_va_response');

    $netnode->set_always('_determine_network_topology_role', 'A1');

    $netnode->set_always('_get_storage_show_bridge',
        Test::MockObject->new()
            ->set_always('get_parsed_output', [$in{'bridge_output'}]));

    $netnode->set_always('_get_storage_show_switch',
        Test::MockObject->new()
            ->set_always('get_parsed_output', [$in{'switch_output'}]));

    my $nodescope_apiset = NACL::UnitTest::MTask::MCC::MockObject->new();

    $nodescope_apiset->set_always(
            'sysconfig',
            {   'parsed_output' =>
                    [{'slot_information' => $in{'sysconfig_output'} }]
            }
        );

    $nodescope_apiset->set_true('set_timeout');


    $netnode->set_always(
        'get_nodescope_apiset',
        $nodescope_apiset
    );

    $netnode->set_false('L_is_vsim');

    return $netnode;
} ## end sub make_mock_network_node

sub make_mock_mcc {
    my $mcc = make_mcc_without_init();

    my $cluster_a = Test::MockObject::Extends->new(
        NACL::MTask::MCC::Cluster->new(
            site         => 'A',
            metrocluster => $mcc
        )
    );

    my $cluster_b = Test::MockObject::Extends->new(
        NACL::MTask::MCC::Cluster->new(
            site         => 'B',
            metrocluster => $mcc
        )
    );

    $mcc->_cluster_A($cluster_a);

    $mcc->_cluster_B($cluster_b);

    my $dr_group = NACL::UnitTest::MTask::MCC::MockObject->new();

    $mcc->_dr_groups([$dr_group]);

    $dr_group->set_true('is_4_pack');

    $dr_group->set_false('is_2_pack');

    $dr_group->set_always('get_id', 'A');

    set_mcc_state_normal($mcc);

    return $mcc;
}

sub mock_subtest {
    my ($class) = @_;

    $class //= 'NACL::MTask::MCC::Metrocluster';

    my $mock_subtest = Test::MockObject->new();
    my $mcc_mod = Test::MockModule->new($class);
    $mcc_mod->mock(
        'subtest',
        sub (@) {
            my ($sub) = @_;

            my @results;

            my $rc = eval {
                @results = &{$sub}();
                return 1;
            };
            if ($rc) {
                $mock_subtest->set_list('results', @results);
            }
            else {
                $mock_subtest->set_false('results');
            }
            return $mock_subtest;
        }
    );
    my $subtest      = Test::MockModule->new('Subtest');
    $subtest->mock( 'new',         sub ($;%) { return $mock_subtest } );
    $subtest->mock( 'wait_finish', sub       { } );

    return ( $mcc_mod, $subtest, $mock_subtest );
}

sub make_mock_aggregate {

    my %opts = validate_with(
        params => \@_,
        spec => {
            'state_ref' => {
                'type'    => SCALARREF,
                'default' => \my $def_state_ref
            },
            'verify_ref' => {
                'type'    => SCALARREF,
                'default' => \my $def_verify_ref
            },
            'raid_status_ref' => {
                'type'    => SCALARREF,
                'default' => \my $def_raid_status_ref
            },
            'is_home_ref' => {
                'type'    => SCALARREF,
                'default' => \my $def_is_home_ref
            },
            'is_scrub_ref' => {
                'type'    => SCALARREF,
                'default' => \my $def_is_scrub_ref
            },
            'raidgroups' => {
                'type' => ARRAYREF,
                default => [
                    '/sti6250mcc_006_aggr1/plex0/rg0 (block)',
                    '/sti6250mcc_006_aggr1/plex1/rg0 (block)'
                ]
            },
        }
    );

    if (! defined ${ $opts{'state_ref'} } ) {
        ${ $opts{'state_ref'} } = 'OK';
    }

    my %attr_refs_table = (
        'state'       => $opts{'state_ref'},
        'raid_status' => $opts{'raid_status_ref'},
        'is_home'     => $opts{'is_home_ref'}
    );


    my $aggr = Test::MockObject->new();

    $aggr->mock(
        'get_one_state_attribute',
        sub {
            my $ref = $attr_refs_table{$_[1]} or return '';
            return ${$ref};
        }
    );


    $aggr->mock('verify', sub { ${ $opts{'verify_ref'} }++ });
    $aggr->set_always( 'aggregate',               'mock aggregate' );

    $aggr->mock('scrub', sub { ${ $opts{'scrub_ref'} }++ });

    my $command_interface = NACL::UnitTest::MTask::MCC::MockObject->new();
    $command_interface->set_isa('NACL::C::Cluster');
    $command_interface->set_false('refresh_command_interface');
    $aggr->set_always( 'command_interface', $command_interface );

    my $version_manager = NACL::UnitTest::MTask::MCC::MockObject->new();
    $command_interface->set_always('version_manager', $version_manager);

    my $aggr_state = Test::MockObject->new();
    $aggr_state->set_bound('state', $opts{'state_ref'});
    $aggr_state->set_bound('raidstatus', $opts{'raid_status_ref'});
    $aggr_state->set_bound('is_home', $opts{'is_home_ref'});

    $aggr_state->set_list('raidgroups', @{ $opts{'raidgroups'} } );

    $aggr->set_always('state', $aggr_state);

    $aggr->set_isa( 'NACL::STask::Aggregate', 'NACL::C::Aggregate', 'NACL::C::StorageAggregate' );

    return $aggr;
}

sub mock_wafliron {

    my $wafliron_mock_module = Test::MockModule->new('NACL::MTask::Wafliron');
    my $wafliron = Test::MockObject->new();
    $wafliron_mock_module->mock('new', sub {
                                    my (@args) = @_;

                                    return $wafliron;
                                });

    $wafliron->set_true('start');
    $wafliron->set_false('get_wafliron_messages');
    $wafliron->set_false('corruption_found');

    return ($wafliron_mock_module, $wafliron);

}

sub mock_network_interfaces {

    my %opts = validate_with(
        params => \@_,
        spec => {
            'mock_results' => { type => ARRAYREF, optional => 1 }
        }
    );

    my @mock_results;

    if ( defined $opts{'mock_results'} ) {
        @mock_results = @{ $opts{'mock_results'} };
    }
    else {
        my $mock_ni = Test::MockObject->new();
        $mock_ni->set_always('get_one_state_attribute', 'up');

        @mock_results = ($mock_ni, $mock_ni, $mock_ni, $mock_ni);
    }

    my $mock_module = Test::MockModule->new('NACL::CS::NetworkInterface');
    $mock_module->mock(
        'fetch',
        sub {
            return @mock_results;
        }
    );

    return $mock_module;

}

# We should eventually merge mock_many with mock_aggregates
sub mock_many_aggregates {
    my %opts = validate_with(
        'params' => \@_,
        'spec' => {
            'mcc' => {'type' => OBJECT},
            'aggr_states' => {'type' => ARRAYREF }
        },
    );

    my @mock_aggrs;

    foreach my $state ( @{ $opts{'aggr_states'} } ) {

        my %aggr_opts = (
            state_ref => \ $state->{'state'},
            raid_status_ref => \ $state->{'raidstatus'},
            is_home_ref => \ $state->{'is_home'}
        );

        if ( defined $state->{'raidgroups'} ) {
            $aggr_opts{'raidgroups'} = $state->{'raidgroups'};
        }

        my $aggr = make_mock_aggregate(
            %aggr_opts
        );

        push(@mock_aggrs, $aggr);
    }

    my $mock_module = Test::MockModule->new('NACL::C::StorageAggregate');
    $mock_module->mock(
        'find',
        sub { 
            my ($self, %opts) = @_;

            defined $opts{'command_interface'} or die "Command interface required";
            return @mock_aggrs 
       }
    );

    return $mock_module;

}

sub mock_aggregates {
    my %opts = validate_with(
        'params' => \@_,
        'spec'   => {
            'mcc'       => {'type' => OBJECT},
            'state_ref' => {
                'type'    => SCALARREF,
                'default' => \my $def_state_ref
            },
            'verify_ref' => {
                'type'    => SCALARREF,
                'default' => \my $def_verify_ref
            },
            'raid_status_ref' => {
                'type'    => SCALARREF,
                'default' => \my $def_raid_status_ref
            }
        }
    );

    my ($state_ref, $verify_ref, $raid_status_ref)
        = @opts{'state_ref', 'verify_ref', 'raid_status_ref'};

    my $mock_module = Test::MockModule->new('NACL::C::StorageAggregate');
    $mock_module->mock(
        'find',
        sub {
            my $self = shift;

            my $mock_aggr = make_mock_aggregate(
                'state_ref' => $opts{'state_ref'},
                'verify_ref' => $opts{'verify_ref'},
                'raid_status_ref' => $opts{'raid_status_ref'}
            );

            return $mock_aggr;
        }
    );

    return $mock_module;
} ## end sub mock_aggregates

sub mock_component_state_fetch {

    my %opts = validate_with(
        'params' => \@_,
        'spec'   => {
            'cs_class' => {
                'type'    => SCALAR
            },
            'cs_states' => {
                'type'    => HASHREF
            }
        }
    );

    # Create a mock object that pretends to be the appropriate
    # component state
    my $mock_cs = Test::MockObject->new();
    $mock_cs->set_isa( $opts{'cs_class'}, 'NACL::CS::ComponentState' );

    foreach my $cs_state ( keys %{ $opts{'cs_states'} } ) {
        $mock_cs->set_always( $cs_state, $opts{'cs_states'}{$cs_state} );
    }

    # Mock the component state's fetch() method to return our mock object
    my $cs_mod = Test::MockModule->new( $opts{'cs_class'} );
    $cs_mod->mock('fetch', sub {
                      return $mock_cs;
                  });

    # Return the mock module (to keep it in scope) and the mock object
    # (in case the value needs to be changed)
    return ($cs_mod, $mock_cs);

}

sub set_mcc_state_normal {
    my ($mcc) = @_;

    $mcc->set_always(
        'get_disaster_status',
        {   DR_state        => 'normal',
            'survivor_site' => 'A',
            'disaster_site' => 'B',
            'live_sites'    => ['A', 'B']
        }
    );

}

sub set_mcc_state_switchover {
    my ($mcc) = @_;

    $mcc->set_always(
        'get_disaster_status',
        {   DR_state        => 'switchover_completed',
            'survivor_site' => 'A',
            'disaster_site' => 'B',
            'live_sites'    => ['A', 'B']
        }
    );

}

sub set_mcc_state_heal_aggregates {
    my ($mcc) = @_;

    $mcc->set_always(
        'get_disaster_status',
        {   DR_state        => 'heal_aggrs_completed',
            'survivor_site' => 'A',
            'disaster_site' => 'B',
            'live_sites'    => ['A', 'B']
        }
    );

}

sub set_mcc_state_heal_roots {
    my ($mcc) = @_;

    $mcc->set_always(
        'get_disaster_status',
        {   DR_state        => 'heal_roots_completed',
            'survivor_site' => 'A',
            'disaster_site' => 'B',
            'live_sites'    => ['A', 'B']
        }
    );

}

##
## make_dr_coordinator()
##
## Creates a real DR_Coordinator object initialized with mock
## objects.
##
## Returns: Test::MockObject extension of DR_Coordinator.
##
sub make_dr_coordinator {

    my $mcc = make_mock_mcc();

    my $dr = NACL::MTask::MCC::DR_Coordinator->new(
        metrocluster => $mcc,
        site         => 'A'
    );

    my $mock_dr = Test::MockObject::Extends->new($dr);

    return $mock_dr;
}



INIT {

    our $Log                  = make_mock_log();
    our $Mock_CS_Metrocluster = mock_cs_metrocluster();
}

1;

__END__

=head1 NAME

NACL::UnitTest::MTask::MCC::TestUtils - Utilities for writing unit tests for MCC

=head1 SYNOPSIS

  use NACL::UnitTest::MTask::MCC::TestUtils qw(:all);

  # Override log functions with a mock log
  $Log = make_mock_log();

  # Override bootstrap util functions
  $Bootstrap = make_bootstrap_util();

  # Override CS::Metrocluster calls
  $Mock_CS_Metrocluster = mock_cs_metrocluster();

  # Get a Metrocluster MTask without init() called
  my $mcc = make_mcc_without_init();

  # Make a mock MCC
  my $mcc = make_mock_mcc();

  # Make subtests immediately execute in this process
  my ($mcc_mod, $subtest) = mock_subtest();

  # Make a mock NACL::STask::Aggregate
  my $aggr = make_mock_aggregate();

  # Hijack Aggregate->find() to return mock aggregates
  my $mock_module = mock_aggregates();

=head1 DESCRIPTION

This module contains routines that allow you to create mocked objects
useful for unit tests on MCC QA library code.

=cut
