#!/usr/software/bin/perl

use strict;
use warnings;

use Test::More import => ['!subtest'];
use Test::Warn;
use UNIVERSAL::isa;

use Test::MockObject;
use Test::MockObject::Extends;
use Test::MockModule;
use NACL::UnitTest::MTask::MCC::TestUtils ':all';
use NACL::UnitTest::MTask::MCC::MockObject;

our $Log;

MAIN: {

    test_use_DR_Coordinator();
    test_get_DR_state();
    test_get_DR_state_obsolete();
    test_DR_switchover();
    done_testing();
}

#############################################################
#
# Tests
#
# These subroutines test the NACL::MTask::MCC::DR_Coordinator
# object.
##############################################################

##
## test_use_DR_Coordinator()
##
## Loads NACL::MTask::MCC::DR_Coordinator and prints "ok" if it loads
## correctly.
##

sub test_use_DR_Coordinator {
    ##no critic (ErrorHandling::RequireCarping)

    # Skip warnings caused by UNIVERSAL:isa (which we need to avoid
    # problems with Params::Validate and Test::MockObject)
    local $SIG{'__WARN__'} = sub {
        $_[0] =~ /^Called UNIVERSAL::(isa|can)|^Unrecognized escape/ and return;
        warn( $_[0] );
    };

    use_ok('NACL::MTask::MCC::DR_Coordinator');

    return;
}

##
## test_get_DR_state()
##
## Tests get_DR_state against some known operation results.
##
sub test_get_DR_state {

    # Set up MetroclusterOperation::fetch() to return a mock object
    my $result = Test::MockObject->new();
    my $cs_mcc = Test::MockModule->new('NACL::CS::MetroclusterOperation');
    $cs_mcc->mock(
        'fetch',
        sub {
            return $result;
        }
    );

    # Create a DR Coordinator to which we can add mock methods
    my $mock_dr = make_dr_coordinator();

    my $test_no_warnings_sub = sub {
        my ($operation, $state, $expected_found_state) = @_;

        # Test that asking for state when the machine reports 'heal root
        # aggregates'/'partially successful' returns
        # 'heal_roots_partially_successful' with no warnings;
        $result->set_always( 'operation', $operation );
        $result->set_always( 'state', $state );
        $result->set_false( 'error_list' );

        my $found_state = $mock_dr->get_DR_state();

        warnings_are {

            is( $found_state, $expected_found_state,
                "get_DR_state ($operation, $state) -> $expected_found_state" );

        } [], 'No warnings for non-deprecated state';
    };

    $test_no_warnings_sub->('heal root aggregates',
                            'partially successful',
                            'heal_roots_partially_successful'
                        );

    $test_no_warnings_sub->('switchover-simulate',
                            'successful',
                            'normal'
                        );

    $test_no_warnings_sub->('switchback-simulate',
                            'successful',
                            'heal_roots_completed'
                        );

    return;
}

##
## test_get_DR_state_warnings()
##
## Tests get_DR_state against some deprecated operation results.
##
sub test_get_DR_state_obsolete {

    test_operation('switchback-continuation-agent',
                   'in-progress',
                   'switchback_in_progress');

    return;
}

sub test_operation {
    my ($operation, $state, $expected_found_state) = @_;

    # Set up MetroclusterOperation::fetch() to return a mock object
    my $result = Test::MockObject->new();
    my $cs_mcc = Test::MockModule->new('NACL::CS::MetroclusterOperation');
    $cs_mcc->mock(
        'fetch',
        sub {
            return $result;
        }
    );

    # Create a DR Coordinator to which we can add mock methods
    my $mock_dr = make_dr_coordinator();

    # Test that asking for state when the machine reports 'heal root
    # aggregates'/'partially successful' returns
    # 'heal_roots_partially_successful' with no warnings;
    $result->set_always( 'operation', $operation );
    $result->set_always( 'state', $state );
    $result->set_false( 'error_list' );

    my $found_state = $mock_dr->get_DR_state();

    is( $found_state, $expected_found_state,
        "get_DR_state ($operation, $state) -> $expected_found_state" );

}



##
## test_DR_switchover()
##
## Tests that DR_switchover makes the appropriate calls.
##
sub test_DR_switchover {

    # Return a mock object for the Metrocluster component,
    # return an arbitrary number for the state timeout,
    my $mock_dr = make_dr_coordinator();

    $mock_dr->_run_statistics(0);

    # Mock subtest
    my ($mcc_mod, $subtest, $mock_subtest) = mock_subtest('NACL::MTask::MCC::DR_Coordinator');

    $mock_dr->_metrocluster->mock(
        'has_uichange',
        sub {
            my ( $self, %opts ) = @_;
            if ( $opts{'uichange'} eq 'mcc-switchover-ui-change' ) {
                return 1;
            } elsif ( $opts{'uichange'} eq 'cluster-stream-down-on-switchover' )
            {
                return 1;
            } else {
                diag("Got args: @_");
            }
            return;
        }
    );
    $mock_dr->_metrocluster->set_false('refresh_command_interfaces');

    $mock_dr->set_always( 'get_DR_state', 'switchover_completed' );

    my $mock_mcc_c = NACL::UnitTest::MTask::MCC::MockObject->new();

    $mock_mcc_c->{'command_interface'}{'version_manager'}{'hostobj'} =
      NACL::UnitTest::MTask::MCC::MockObject->new();

    $mock_dr->set_always( 'get_Metrocluster_component', $mock_mcc_c );

    # Set up command interface mocking
    my $ci = Test::MockObject->new();
    $ci->set_false('refresh_command_interface');
    my ( $new_method, $apiset ) = setup_apiset();
    $ci->set_always( 'apiset', $apiset );
    $ci->set_isa('NACL::C::CommandInterface');

    my $version_manager = NACL::UnitTest::MTask::MCC::MockObject->new();
    $ci->set_always('version_manager', $version_manager);
    $version_manager->set_true('hostobj');

    $mock_mcc_c->set_always( 'command_interface', $ci );
    $apiset->set_always( 'cluster_identity_show',
        Test::MockObject->new()
          ->set_always( 'get_parsed_output', [ { uuid => 'test-uuid' } ] ) );
    $apiset->set_always(
        'debug_crs_stream_show',
        Test::MockObject->new()->set_always(
            'get_parsed_output', [ { synopsis => 'test-synopsis' } ]
        )
    );
    $apiset->set_true('set_privilege');

    # Mocks with tests

    my ($dr_state_opts, $set_timeout_opts, $metrocluster_switchover_opts);

    $mock_dr->mock('wait_for_DR_state', sub {
                       my ($self, %opts) = @_;

                       $dr_state_opts = \%opts;
                   });

    $apiset->mock(
        'set_timeout',
        sub {
            my ( $self, %opts ) = @_;

            $set_timeout_opts = \%opts;

            return;
        }
    );

    $apiset->mock(
        'metrocluster_switchover',
        sub {
            my ( $self, %opts ) = @_;

            $metrocluster_switchover_opts = \%opts;

            return;
        }
    );

    # Run with defaults
    $mock_dr->DR_switchover();

    is_deeply( $dr_state_opts, { 'state' => 'switchover_completed' } );
    is( $set_timeout_opts->{'connectrec-timeout'}, 600, 'connectrec_timeout' );
#    is( $metrocluster_switchover_opts->{'privilege-level'},
#        'test', 'privilege level' );
    is( $metrocluster_switchover_opts->{'forced-on-disaster'},
        'true', 'forced-on-disaster' );
    is( $metrocluster_switchover_opts->{'force-nvfail-all'}, 'false', 'force-nvfail-all default' );
    is( $metrocluster_switchover_opts->{'override-vetoes'}, 'false', 'override-vetoes default' );
    ok( ! exists $metrocluster_switchover_opts->{'retry-failed-nodes'}, 'retry-failed-nodes default unset' );

    # Run with custom timeout
    $mock_dr->DR_switchover( timeout => 200 );

    is( $set_timeout_opts->{'connectrec-timeout'}, 200, 'connectrec_timeout custom' );
    is_deeply( $dr_state_opts, { 'state' => 'switchover_completed', 'timeout' => 200 } );

    # Customize forced-on-disaster
    $mock_dr->DR_switchover( 'forced-on-disaster' => 0 );
    is ( $metrocluster_switchover_opts->{'forced-on-disaster'},
         'false', 'forced-on-disaster false'
         );
    is_deeply( $dr_state_opts, { 'state' => 'switchover_completed' } );

    # Make force-nvfail-all true
    $mock_dr->DR_switchover( 'force-nvfail-all' => 1 );
    is( $metrocluster_switchover_opts->{'force-nvfail-all'},
        'true', 'force-nvfail-all true' );

    # Make override-vetoes true
    $mock_dr->DR_switchover( 'override-vetoes' => 1 );
    is( $metrocluster_switchover_opts->{'override-vetoes'}, 'true', 'override-vetoes true' );

    # Set retry-failed-nodes
    $mock_dr->DR_switchover( 'retry-failed-nodes' => [ 'foo', 'bar' ] );
    is( $metrocluster_switchover_opts->{'retry-failed-nodes'},
        'foo bar', 'retry-failed-nodes set' );

    return;
}


#############################################################
#
# Subroutines
#
# These are utility subroutines used by the test routines.
#
#############################################################


##
## make_mock_mcc_metrocluster()
##
## Creates a mock MCC::Metrocluster object with some mocked
## methods. This is to pass to the DR_Coordinator.
##
## Returns: Test::MockObject
##
sub make_mock_mcc_metrocluster {
    my $mock_mcc = Test::MockObject->new();
    $mock_mcc->set_isa('NACL::MTask::MCC::Metrocluster');

    my $mock_mcc_c = Test::MockObject->new();
    my $ci         = Test::MockObject->new();
    $ci->set_isa('NACL::C::CommandInterface');
    $mock_mcc_c->set_always( 'command_interface', $ci );

    $mock_mcc->set_always( 'get_Metrocluster_component', $mock_mcc_c );

    $mock_mcc->set_false('show_detailed_perf_stats');

    return $mock_mcc;
}

##
## setup_apiset()
##
## Creates a fake NACL::APISet object and sets up the APISet
## constructor to return it.
##
sub setup_apiset {

    my $apiset = Test::MockObject->new();

    my $mock_new = Test::MockModule->new('NACL::APISet');
    $mock_new->mock('new', $apiset);

    return ($mock_new, $apiset);
}

__END__

=head1 NAME

02_dr_coordinator.t - Unit tests for DR_Coordinator

=head1 SYNOPSIS

    perl 02_dr_coordinator.t

=head1 DESCRIPTION

Runs a series of independent unit tests on  NACL::MTask::MCC::DR_Coordinator.
For each test, prints out 'OK' if the test passes, or 'not ok' if it fails.

=cut

