#!/usr/software/bin/perl5.14.0

use strict;
use warnings;

use Test::Most import => ['!subtest'];

use Test::MockModule;
use NACL::UnitTest::MTask::MCC::MockObject;
use NACL::UnitTest::MTask::MCC::TestUtils ':all';
use Carp;
use Params::Validate
    qw(validate validate_with BOOLEAN HASHREF OBJECT ARRAYREF SCALAR UNDEF);
use NACL::APISet::Response::CLI::CMode; #for get_parsed_output()
MAIN: {
    # Suppress logging
    $Tharn::Log = make_mock_log();
    test_use_mcc();
    test_vserver_validate();
    test_vserver_validate_error();
    test_vserver_validate_config_lock();
    test_vserver_validate_config_lock_owner();
    test_vserver_validate_config_lock_owner_die();
    test_vserver_validate_config_lock_dest_ok();
    done_testing();
}

sub make_mock_vserver {

    my $vserver = Test::MockObject->new();

    $vserver->set_always('vserver', 'mock vserver');
    $vserver->{'vserver'} = 'mock vserver';
    $vserver->{'subtype'} = 'sync';
    $vserver->{'operational_state'} = 'initializing';
    $vserver->{'config_lock'} = 'config_lock';
    $vserver->{'uuid'} = 'mock-uuid';

    $vserver->{'admin_state'} = 'running';

    return $vserver;

}


our $start_time;
our $total_snoozed;

BEGIN {

    $total_snoozed = 0;

     *{CORE::GLOBAL::time} = sub () {

         if (! defined $start_time) {
             $start_time = CORE::time();
         }

         my $fake_time = $start_time + $total_snoozed;

         return $fake_time;
     };

     sub mock_snooze {

         my $mock_tharn    = Test::MockModule->new('Tharn');
         $mock_tharn->mock('snooze', sub {
                               $total_snoozed += $_[0];
                           }
                       );
         $total_snoozed = 0;
         $start_time = undef;
         return $mock_tharn;
     }

     sub clear_total_snoozed {
         $total_snoozed = 0;
     }

}

sub mock_debug_smdb_table_show {

    my %opts = validate_with(
        'params' => \@_,
        'spec'   => {
            'mcc' => {
                'type'    => OBJECT,
            },
            'uuid' => {
                type => SCALAR,
                default => 'mock-uuid'
            },
            'owner' => {
                type => SCALAR,
                default => 'mcc'
            }
        }
    );

    my $mcc = $opts{'mcc'};

    foreach my $site (qw/A B/) {

        my $apiset = $mcc->get_Metrocluster_component(site => $site)
            ->command_interface()->apiset();

        $apiset->set_always(
            'debug_smdb_table_show',

            bless( {
                    'parsed_output' => [
                        {   'owner' => $opts{'owner'},
                            'vserver' =>
                                $opts{'uuid'}
                        }
                    ],
                },
                'NACL::APISet::Response::CLI::CMode'
            )
        );

    }

    return $mcc;

} ## end sub make_mcc_with_debug_smdb_table_show

sub test_vserver_validate {

    my $mcc = make_mock_mcc();

    mock_debug_smdb_table_show( mcc => $mcc );

    # Set normal state
    set_mcc_state_normal($mcc);

    my $mock_vserver_mod = Test::MockModule->new('NACL::CS::Vserver');

    my $vserver = make_mock_vserver();

    $vserver->set_series(
        'admin_state',
        'initializing',
        'initializing',
        'running',
        'running'
    );

    $mock_vserver_mod->mock(
        'fetch',
        sub {
            return $vserver;
        }
    );

    my $mock_tharn = mock_snooze();

    # Validate vservers
    lives_ok {
        $mcc->validate_vservers();
    }
    'validate_vservers initializing';

} ## end sub test_vserver_validate

sub test_vserver_validate_config_lock {

    my $mcc = make_mock_mcc();

    mock_debug_smdb_table_show( mcc => $mcc, owner => 'vserverdm' );

    # Set normal state
    set_mcc_state_normal($mcc);

    my $mock_vserver_mod = Test::MockModule->new('NACL::CS::Vserver');

    my $vserver = make_mock_vserver();
    $vserver->{'config_lock'} = 'true';
    $vserver->{'subtype'} = 'sync-source';

    $vserver->set_always(
        'admin_state',
        'running'
    );

    $vserver->{'admin_state'} = 'running';

    $mock_vserver_mod->mock(
        'fetch',
        sub {
            return $vserver;
        }
    );

    my $mock_tharn = mock_snooze();

    # Validate vservers
    dies_ok {
        $mcc->validate_vservers();
    }
    'validate_vservers config lock dies yay';

} ## end sub test_vserver_validate_config_lock

sub test_vserver_validate_config_lock_owner {

    my $mcc = make_mock_mcc();

    mock_debug_smdb_table_show( mcc => $mcc, owner => 'vserver-dm' );

    # Set normal state
    set_mcc_state_normal($mcc);

    my $mock_vserver_mod = Test::MockModule->new('NACL::CS::Vserver');

    my $vserver = make_mock_vserver();
    $vserver->{'config_lock'} = 'true';
    $vserver->{'subtype'} = 'sync-source';

    $vserver->set_always(
        'admin_state',
        'running'
    );

    $vserver->{'admin_state'} = 'running';
    $vserver->{'operational_state'} = 'running';

    $mock_vserver_mod->mock(
        'fetch',
        sub {
            return $vserver;
        }
    );

    my $mock_tharn = mock_snooze();

    # Validate vservers, ignore lock owned by mcc
    # should NOT be locked, but it is - it is OK because owner is mcc
    lives_ok {
        $mcc->validate_vservers();
    }
    'validate_vservers config lock lives when lock owner is mcc ';

} ## end sub test_vserver_validate_config_lock

sub test_vserver_validate_config_lock_owner_die {

    my $mcc = make_mock_mcc();

    mock_debug_smdb_table_show( mcc => $mcc, owner => 'vserverdm' );

    # Set normal state
    set_mcc_state_normal($mcc);

    my $mock_vserver_mod = Test::MockModule->new('NACL::CS::Vserver');

    my $vserver = make_mock_vserver();
    $vserver->{'config_lock'} = 'true';
    $vserver->{'subtype'} = 'sync-source';

    $vserver->set_always(
        'admin_state',
        'running'
    );

    $vserver->{'admin_state'} = 'running';
    $vserver->{'operational_state'} = 'running';

    $mock_vserver_mod->mock(
        'fetch',
        sub {
            return $vserver;
        }
    );

    my $mock_tharn = mock_snooze();

    # Validate vservers, fail lock NOT owned by mcc/owned by vserverdm
    # should NOT be locked, but it is - it is not OK because owner is not mcc
    dies_ok {
        $mcc->validate_vservers();
    }
    'validate_vservers config lock dies when lock owner is vserverdm';

} ## end sub test_vserver_validate_config_lock

sub test_vserver_validate_config_lock_dest_ok {

    my $mcc = make_mock_mcc();

    mock_debug_smdb_table_show( mcc => $mcc, owner => 'vserverdm' );

    # Set heal-roots state
    set_mcc_state_heal_roots($mcc);

    my $mock_vserver_mod = Test::MockModule->new('NACL::CS::Vserver');

    my $vserver_a = make_mock_vserver();
    $vserver_a->{'config_lock'} = 'false';
    $vserver_a->{'subtype'} = 'sync-source';

    $vserver_a->set_always(
        'admin_state',
        'running'
    );

    $vserver_a->{'admin_state'} = 'running';
    $vserver_a->{'operational_state'} = 'running';

    my $vserver_b = make_mock_vserver();
    $vserver_b->{'config_lock'} = 'true';
    $vserver_b->{'subtype'} = 'sync-destination';

    $vserver_b->set_always(
        'admin_state',
        'stopped'
    );

    $vserver_b->{'admin_state'} = 'running';
    $vserver_b->{'operational_state'} = 'stopped';

    $mock_vserver_mod->mock(
        'fetch',
        sub {
            my ($vserver, %opts) = @_;

            my $ci = $opts{'command_interface'};

            if ( $ci->_site eq 'A' ) {
                return $vserver_a;
            } else {
                return $vserver_b;
            }
        }
    );

    my $mock_tharn = mock_snooze();

    # Validate vservers, fail lock NOT owned by mcc/owned by vserverdm
    # should NOT be locked, but it is - it is not OK because owner is not mcc
    lives_ok {
        $mcc->validate_vservers();
    }
    'validate_vservers config lock lives with locked dest vserver';

} ## end sub test_vserver_validate_config_lock


sub test_vserver_validate_error {

    $total_snoozed = 0;

    # Create mock MCC
    my $mcc = make_mock_mcc();

    mock_debug_smdb_table_show( mcc => $mcc );

    # Set normal state
    set_mcc_state_normal($mcc);

    my $mock_vserver_mod = Test::MockModule->new('NACL::CS::Vserver');

    my $vserver = make_mock_vserver();

    $vserver->set_always('admin_state', 'initializing');

    $vserver->{'admin_state'} = 'initializing';

    $mock_vserver_mod->mock(
        'fetch',
        sub {
            return $vserver;
        }
    );

    my $mock_tharn = mock_snooze();

    # Validate vservers
    dies_ok {
        $mcc->validate_vservers();
    }
    'validate_vservers times out';

} ## end sub test_vserver_validate_error
