#!/usr/software/bin/perl5.14.0

use strict;
use warnings;

use Test::Most import => ['!subtest'];
use Test::MockModule;

use NACL::UnitTest::MTask::MCC::MockObject;
use NACL::UnitTest::MTask::MCC::TestUtils ':all';
use Carp;
use Params::Validate
    qw(validate validate_with BOOLEAN HASHREF OBJECT ARRAYREF SCALAR UNDEF);

my ($ran_foo, $ran_bar) = (0, 0);

MAIN: {
    use_ok('NACL::MTask::MCC::Validator::Manager');

    # Suppress logging
{
    no warnings 'once';
    $NACL::MTask::MCC::Validator::Base::Log = make_mock_log();
    $NACL::MTask::MCC::Validator::Manager::Log = make_mock_log();
}

    test_constructor();
    test_get_all();
    test_table();
    test_run_all_validations();
    test_metrocluster_compat();
    test_metrocluster_run_all();
    done_testing();
}

sub test_constructor {

    # Test default exclusion
    my $mgr = NACL::MTask::MCC::Validator::Manager->new();

    isa_ok( $mgr,  'NACL::MTask::MCC::Validator::Manager' );


} ## end sub test_MY_TEST

sub test_get_all {
    my $mock_loader = mock_loader();

    my $mgr = NACL::MTask::MCC::Validator::Manager->new();

    my @vtors = $mgr->get_all_validations();

    is_deeply( \@vtors, [qw/Foo Bar/], 'get_all_validations() correct' );
}

sub test_table {
    my $mock_loader = mock_loader();

    my $mgr = NACL::MTask::MCC::Validator::Manager->new();

    my %vtable = $mgr->validation_types_table();

    is_deeply( \%vtable, { 
        foo => 'Description for foo',
        bar => 'Description for bar'
    }, 'validation types table correct' );
}

sub test_run_all_validations {
    my $mock_loader = mock_loader();
    my $metrocluster = make_mock_mcc();

    # Suppress validation_report() output
    $metrocluster->set_true('validation_report');

    my $mgr = NACL::MTask::MCC::Validator::Manager->new();

    # Test running validations
    ($ran_foo, $ran_bar) = (0, 0);
    $mgr->run_all_validations(
        run_extended => 0,
        metrocluster => $metrocluster
    );
    is_deeply([$ran_foo, $ran_bar], [1, 0],
              'Ran non-extended validations');

    # Test running extended
    ($ran_foo, $ran_bar) = (0, 0);
    $mgr->run_all_validations(
        run_extended => 1,
        metrocluster => $metrocluster
    );
    is_deeply([$ran_foo, $ran_bar], [1, 1],
              'Ran extended validations');

    # Test plain string exclusions
    ($ran_foo, $ran_bar) = (0, 0);
    $mgr->run_all_validations(
        run_extended => 1,
        metrocluster => $metrocluster,
        exclude      => [ 'foo' ]
    );
    is_deeply([$ran_foo, $ran_bar], [0, 1],
              'Plain string exclusions');

    # Test plain string exclusions
    ($ran_foo, $ran_bar) = (0, 0);
    $mgr->run_all_validations(
        run_extended => 1,
        metrocluster => $metrocluster,
        exclude      => [ qr/oo/ ]
    );
    is_deeply([$ran_foo, $ran_bar], [0, 1],
              'Regex exclusions');

    # Test plain string inclusions
    ($ran_foo, $ran_bar) = (0, 0);
    $mgr->run_all_validations(
        run_extended => 1,
        metrocluster => $metrocluster,
        include      => [ 'bar' ]
    );
    is_deeply([$ran_foo, $ran_bar], [0, 1],
              'String inclusion');

   # Test regex inclusions
    ($ran_foo, $ran_bar) = (0, 0);
    $mgr->run_all_validations(
        run_extended => 1,
        metrocluster => $metrocluster,
        include      => [ qr/ar$/ ]
    );
    is_deeply([$ran_foo, $ran_bar], [0, 1],
              'Regex inclusion');

    # Throwing exceptions
    throws_ok {
        $mgr->run_all_validations(
            metrocluster => $metrocluster,
            die_on_failure => 1,
            run_extended => 1,
            extra_validate_options => { 'errors_bar' => 1 }
        );
    } 'NACL::MTask::MCC::Exceptions::Validation', 'throws exceptions';

}

sub test_metrocluster_compat {
    my $mock_loader = mock_loader();
    my $mgr = NACL::MTask::MCC::Validator::Manager->new();

    my @all_validations = $mgr->_get_all_validations_compatibility_list();

    delete $_->{'sub'} foreach @all_validations;

    is_deeply(\@all_validations, [{'name' => 'foo', default => 1, manager => 1}, {'name' => 'bar', default => 0, manager => 1}] );

}

sub test_metrocluster_run_all {
    my $mock_loader = mock_loader();
    my $mgr = NACL::MTask::MCC::Validator::Manager->new();
    my $metrocluster = make_mock_mcc();

    ($ran_foo, $ran_bar) = (0, 0);

    $metrocluster->set_list(
        'get_all_validations',
        {'name' => 'foo', default => 0, manager => 1}, {'name' => 'bar', default => 1, manager => 1}
    );

    $metrocluster->_validator($mgr);

    $metrocluster->run_all_validations();

    is_deeply([$ran_foo, $ran_bar], [1, 0], 'Tested metrocluster run all');
}


sub mock_loader {
    my $mock_loader = Test::MockModule->new('Module::Find');
    $mock_loader->mock('usesub', sub (*) { return qw/Foo Bar FooBase/ });

    return $mock_loader;
}

{
package Foo;

use base 'NACL::MTask::MCC::Validator::Base';

sub validation_type { 'foo' };

sub validation_description { 'Description for foo' }

sub validate { $ran_foo = 1 };
}

{
package Bar;

use base 'NACL::MTask::MCC::Validator::Base';

sub validation_type { 'bar' };

sub validation_description { 'Description for bar' }

sub extended { 1; }

sub is_run_by_default { 0; }

sub validate {
    my $self = shift;

    my %opts = @_;

    $ran_bar = 1;

    $self->errors_push('Error from bar') if $opts{'errors_bar'};
}


}

