#
#Copyright (c) 2014-2015 NetApp, Inc.
#All rights reserved.
#

#
## @summary ParserTester UnitTest Module
## @author dl-nacl-dev@netapp.com
## @status shared
## @pod here
#

package NACL::UnitTest::ParserTester;
use strict;
use warnings;
use Params::Validate qw(:all);
use Tharn qw{$Log};
use NATE::BaseException;
use NACL::APISet;
use NACL::APISet::Schema;
use Data::Dumper;
use Data::Compare;
###############################################################################
# Method:    execute
# Objective: To run a pre-defined output on a parser.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub execute {
    my $self = shift @_;
    my (%args)
        = validate( @_,
                    { output    => { type => SCALAR },
                      method    => { type => SCALAR },
                      cdef_path => { type => SCALAR },
                      parser_path => { type => SCALAR, optional => 1 },
                      options   => { type => SCALAR, optional => 1 },
                      set       => { type => SCALAR, optional => 1, },
                      execution_status       => { type => SCALAR, default => 0, },
                    }
        );
    my $Cdef_path = $args{'cdef_path'};
    if ( $Cdef_path =~ /Host/ and !exists $args{set} ) {
        NATE::BaseException->throw( "Please specify the 'set'\n."
                                  . "Possible values for 'Host' category are "
                                  . "Linux, Windows, Solaris etc" );
    }

    # 'Set' defaults to 7Mode for backward compatibility reasons.
    my $set = $args{'set'} || '7Mode';
    my $constructor = 'NACL::APISet::Response::CLI::' . "$set";
    eval "use $constructor";
    if ($@) {
        NATE::BaseException->throw( "Response class could not be loaded: ",
                                    $@ );
    }
    my $command = $args{'method'};
    $Cdef_path = NATE::Inc::find_file($Cdef_path);
    my $parser_path = NATE::Inc::find_file($args{'parser_path'});
    my $schema_obj = NACL::APISet::Schema->new();
    my @files = @{ $schema_obj->read_cdef_dir( directory => $Cdef_path ) };
    $schema_obj->load( files => \@files, path => $Cdef_path );
    my $parser_file
        = $schema_obj->get_command_definition( command => $command )
        ->get_file();
    $parser_file =~ s/cdefs$/pm/;
    $parser_file = $parser_path."/".$command.".pm" if ( ! -e $parser_file );
    $command = $schema_obj->get_command_definition( command => $command )
        ->get_command();

    # Append options to end of command
    $command .= ' ' . $args{'options'} if ( $args{'options'} );
    my $response_object =
        $constructor->new( output           => $args{output},
                           command          => $command,
                           execution_time   => '12',
                           method_name      => $args{method},
                           actual_command   => $command,
                           cdef_path        => $Cdef_path,
                           execution_status => $args{execution_status},
                           parser_file      => $parser_file,
        );
}

sub execute_and_compare {
    my ($pkg, @opts) = @_;
    my %args = validate_with(
        params      => \@opts,
        spec        => {expected_parsed_output => {type => ARRAYREF},},
        allow_extra => 1
    );
    my $expected_parsed_output = delete $args{expected_parsed_output};
    my $response_object = $pkg->execute(%args);
    my $parsed_output   = $response_object->get_parsed_output();
    $pkg->compare(
        obtained_parsed_output => $parsed_output,
        expected_parsed_output => $expected_parsed_output
    );
    return $response_object;
}

sub compare {
    my ($pkg, @opts) = @_;
    my %args = validate_with(
        params => \@opts,
        spec   => {
            obtained_parsed_output => {type => ARRAYREF},
            expected_parsed_output => {type => ARRAYREF},
        },
        allow_extra => 1
    );
    local $Data::Dumper::Terse    = 1;
    local $Data::Dumper::Indent   = 3;
    local $Data::Dumper::Sortkeys = 1;
    my $expected_output = Dumper delete $args{expected_parsed_output};
    my $parsed_output   = Dumper delete $args{obtained_parsed_output};
    if (!(Data::Compare->new()->Cmp($expected_output, $parsed_output))) {
        $Log->comment("Expected parsed output\n" . $expected_output);
        $Log->comment("Obtained parsed output\n" . $parsed_output);
        NATE::BaseException->throw(
            "Expected parsed output did not match the parsed output obtained");
    }
    return 1;
} ## end sub compare


1;
##############################################################################
#                                    POD here
##############################################################################

=head1 NAME

NACL::UnitTest::ParserTester - used to test the reachability of certain regions of code in a parser.

=head1 SYNOPSIS

  use NACL::UnitTest::ParserTester;

=head1 DESCRIPTION

  A C<NACL::UnitTest::ParserTester> is used to verify whether appropriate statements are being 
  executed when previously seen output is fed as input to the parser through this module.

=head1 Methods

=head2 execute

This method is used to run the previously seen output on the specified parser file.

=over

=item Synopsis

execute(cdef_path=>$cdef_path, method=>$method, output=>$output)

=item Arguments

=over

=item cdef_path

(mandatory)
CDEF path of the parser that needs to be tested.

=item parser_path

(optional)
Parser path of the parser that needs to be tested.
If parser is not present at provided cdef_path location then script looks for it at parser_path

For example: Method 'ifconfig' parser present under dir 'NACL/APISet/Host/CLI/Linux/Redhat' referes to corresponding cdef present under path 'NACL/APISet/Host/CLI/Unix'

=item method

(mandatory)
Method name in <name>_<options> format.
For example: If the command to be issued is vol offline on the filer, then 'method' in this context would have to be specified as vol_offline.

=item output

(mandatory)
Previously seen output should be passed as a string.

=item options

(optional)
Additional options which are part of the command but not part of the method
name.  For example:  If the command to be issued is 'aggr status -S', then
'method' in this case would be 'aggr_status' and 'options' would be '-S'.

=item set

(mandatory)
The 'set' override specifies which APISet needs to be used.
Possible values for 'set' include 7Mode and all the supported Host APISets
(like Linux, Windows etc).

=back

=item Returns

An APISet Response object.

=back

=head2 execute_and_compare

=over

This is a wrapper around and execute and compare methods. It compares the
expected parsed output (hard-coded in the script) with the parsed output
obtained by using get_parsed_output() on the response object obtained returned
from the call to execute().

=item Arguments

=over

Accepts all the arguments accepted by execute() and takes the following additional
argument:

=item expected_parsed_output

(mandatory)
The parsed output that is hard-coded in the script.

=back

=item Returns

Throws NATE::BaseException if the expected_parsed_output does not match the
parsed output obtained by calling get_parsed_output() on the hard-coded
raw output. Otherwise, an APISet Response object is returned.

=back

=head2 compare

=over

Compares two ARRAYREFS and determines if both of them are the same.

=item Arguments

=over

=item expected_parsed_output

(mandatory)
The parsed output that is hard-coded in the script.

=item obtained_parsed_output

(mandatory)
The parsed output obtained by calling get_parsed_output() on the raw output.

=back

=back

=over

=item Returns

=back

1 if the two ARRAYREFS are the same. Otherwise, throws a NATE::BaseException

=cut


