package NACL::UnitTest::TestCleanup;
use TCD;
use vars qw($Test);
use NACL::Cleanup;
use NATE::BaseException qw(:try);
use NACL::Exceptions::CleanupFailed qw(:try);
use NACL::ComponentUtils qw(Dumper);
use Tharn qw($Log);
use Params::Validate qw(validate validate_with  :types);
use Data::Compare;

BEGIN {
    use Exporter qw(import);
    our @EXPORT_OK = qw(
        create_cleanup_manager
        create_an_element
        run_cleanup
        test_cleanup
        _verify_if_element_purged
        _verify_modify
        _verify_rename
    );
}
$Test = new TCD( -testcases => [] );

sub create_cleanup_manager {
    $Log->enter();
    my (@args) = @_;
    $Log->enter();

    my %opts = validate_with(
        params => \@args,
        spec   => {
            'db_name'   => { type => SCALAR, optional => 1 },
            'random_db' => { type => SCALAR, optional => 1 },
        }
    );

    my $cleanup_mgr;
    try {
        $Test->step( "Creating the cleanup db with the options : "
                . Dumper( \%opts ) );
        $cleanup_mgr = NACL::Cleanup->new(%opts);
        my $db_name = $cleanup_mgr->resourceList_filename();
        $Test->nlog("Cleanup db - $db_name is been created successfully!");

    }
    catch NATE::BaseException with {
        my $ex  = shift;
        my $err = $ex->text();
        $Test->nlog(
            "Creation of the cleanup object failed with an error: " . $err );
        $ex->throw();
    };

    $Log->exit();
    return $cleanup_mgr;
}

sub create_an_element {
    $Log->enter();
    my (@args) = @_;
    $Log->enter();

    my %opts = validate_with(
        params => \@args,
        spec   => {
            'package'   => { type => SCALAR },
            'arguments' => { type => HASHREF },
            'cleanup_manager' =>
                { type => OBJECT, 'isa' => 'NACL::Cleanup', default => 0 },
            'nacltask_to_cleanup' => { type => SCALAR, default => 0 }
        }
    );
    my %arguments           = %{ $opts{'arguments'} };
    my $package             = $opts{'package'};
    my $cleanup_manager     = $opts{'cleanup_manager'};
    my $nacltask_to_cleanup = $opts{'nacltask_to_cleanup'};

    eval "require $package";

    if ($@) {
        my $error = "Could not load the package - " . $package . ".\n";
        $error .= "Error: " . $@;
        NATE::BaseException->throw($error);
    }

    my $element;
    my %opts_to_create = (%arguments);

    if ($cleanup_manager) {
        $opts_to_create{'nacltask_cleanup_manager'} = $cleanup_manager;
    } elsif ($nacltask_to_cleanup) {
        $opts_to_create{'nacltask_to_cleanup'} = $nacltask_to_cleanup;
    } else {
    }

    try {
        $Test->step( "Creating the element with the following options : "
                . Dumper( \%opts_to_create ) );
        $element = $package->create(%opts_to_create);
        $Test->nlog("Created the element successfully!");

    }
    catch NATE::BaseException with {
        my $ex  = shift;
        my $err = $ex->text();
        $Test->nlog(
            "Creation of the element failed with an error: " . $err );
        $ex->throw();
    };

    $Log->exit();
    return $element;
}

sub run_cleanup {
    $Log->enter();
    my (@args) = @_;
    $Log->enter();

    my %opts = validate_with(
        params => \@args,
        spec   => {
            'cleanup_manager' => { type => OBJECT, 'isa' => 'NACL::Cleanup' },
        },
    );
    my $cleanup_mgr = $opts{'cleanup_manager'};

    try {
        $Test->step("Running the cleanup for the registered elements.");
        $cleanup_mgr->run();
    }
    catch NACL::Exceptions::CleanupFailed with {
        my $ex = shift;
        $Test->nlog("No elements registered for cleanup");
        $ex->throw();
    }
    catch NATE::BaseException with {
        my $ex  = shift;
        my $err = $ex->text();
        $Test->nlog( "Cleanup run failed with the error: " . $err );
        $ex->throw();
    };

    $Log->exit();
}

sub test_cleanup {
    my (@args) = @_;
    $Log->enter();

    my %opts = validate_with(
        params => \@args,
        spec   => {
            'package'               => { type => SCALAR },
            'method'                => { type => SCALAR },
            'create_args'           => { type => HASHREF, default => 0 },
            'modify_args'           => { type => HASHREF, default => 0 },
            'rename_args'           => { type => HASHREF, default => 0 },
            'test_create_for_reuse' => { type => SCALAR, default => 1 },
            'reuse_attribute_maps'  => { type => HASHREF, default => {} },
        },
    );

    my $package               = $opts{'package'};
    my $method                = $opts{'method'};
    my %create_args           = %{ $opts{'create_args'} };
    my %modify_args           = %{ $opts{'modify_args'} };
    my %rename_args           = %{ $opts{'rename_args'} };
    my $test_create_for_reuse = delete $opts{'test_create_for_reuse'};
    my $cleanup_mgr           = create_cleanup_manager( 'random_db' => 1 );
    my %reuse_attribute_maps  = %{ delete $opts{'reuse_attribute_maps'} };

    if ( !scalar %{ $opts{'create_args'} } ) {
        NATE::BaseException->throw("\"create_args\" parameter is mandatory");
    }
    my $element1 = create_an_element(
        'package'         => $package,
        'arguments'       => \%create_args,
        'cleanup_manager' => $cleanup_mgr
    );

    if ( $method eq "create" ) {

        if ($test_create_for_reuse) {
            my $cleanup_mgr_for_reuse
                = create_cleanup_manager( 'random_db' => 1 );
            my %create_reuse_args
                = ( %create_args, nacltask_if_exists => 'reuse' );
            while ( my ( $key, $value ) = each %create_reuse_args ) {
                if ( defined $reuse_attribute_maps{$key} ) {
                    delete $create_reuse_args{$key};
                    $create_reuse_args{ $reuse_attribute_maps{$key} }
                        = $value;
                }
            }
            my $element2 = create_an_element(
                'package' => $package,
                'arguments' => \%create_reuse_args,
                'cleanup_manager' => $cleanup_mgr_for_reuse
            );

            run_cleanup( 'cleanup_manager' => $cleanup_mgr_for_reuse );

            _verify_if_element_purged(
                'element'        => $element2,
                'fail_if_purged' => 1
            );
        }

        run_cleanup( 'cleanup_manager' => $cleanup_mgr, );
        _verify_if_element_purged(
            'element'        => $element1,
            'fail_if_exists' => 1
        );
    }

    if ( $method eq "modify" ) {
        if ( !scalar %{ $opts{'modify_args'} } ) {
            NATE::BaseException->throw(
                "For \"modify\" method, \"modify_args\" parameter is mandatory"
            );
        }
        _verify_modify(
            'element'         => $element1,
            'modify_args'     => \%modify_args,
            'cleanup_manager' => $cleanup_mgr
        );
    }
    if ( $method eq "rename" ) {
        if ( !scalar %{ $opts{'rename_args'} } ) {
            NATE::BaseException->throw(
                "For \"rename\" method, \"rename_args\" parameter is mandatory"
            );
        }
        _verify_rename(
            'element'         => $element1,
            'rename_args'     => \%rename_args,
            'cleanup_manager' => $cleanup_mgr
        );
    }

    $Log->exit();
    return $TCD::PASS;
}

sub _verify_if_element_purged {
    $Log->enter();
    my (@args) = @_;
    $Log->enter();

    my %opts = validate_with(
        params => \@args,
        spec   => {
            'element'        => { type => OBJECT },
            'fail_if_exists' => { type => SCALAR, default => 0 },
            'fail_if_purged' => { type => SCALAR, default => 0 }
        }
    );

    $element        = $opts{'element'};
    $fail_if_exists = $opts{'fail_if_exists'};
    $fail_if_purged = $opts{'fail_if_purged'};

    try {
        $Test->step("Verify if the element exists after cleanup is run...");
        $element->find();
        if ($fail_if_exists) {
            NATE::BaseException->throw(
                "Cleanup failed to delete the element which is not expected!"
            );
        }
    }
    catch NACL::Exceptions::NoElementsFound with {
        if ($fail_if_purged) {
            NATE::BaseException->throw(
                "Cleanup deleted the element which is not expected!");
        }
    }

    $Log->exit();
}

sub _verify_modify {
    $Log->enter();
    my (@args) = @_;
    $Log->enter();

    my %opts = validate_with(
        params => \@args,
        spec   => {
            'element'         => { type => OBJECT },
            'modify_args'     => { type => HASHREF },
            'cleanup_manager' => { type => OBJECT, 'isa' => 'NACL::Cleanup' },
        }
    );

    my $element                = $opts{'element'};
    my $cleanup_mgr            = $opts{'cleanup_manager'};
    my %modify_args            = %{ $opts{'modify_args'} };
    my $cleanup_mgr_for_modify = create_cleanup_manager( 'random_db' => 1 );
    my %attribs_before_modify;

    my $state_obj_before_modify
        = $element->state( requested_fields => [ keys %modify_args ] );

    $Test->step( "Call \"modify\" for the following attributes: "
            . Dumper( \%modify_args ) );
    $element_after_modify = $element->modify( %modify_args,
        'nacltask_cleanup_manager' => $cleanup_mgr_for_modify );

    $Test->step("Run cleanup and verify if the attributes are reverted.");
    run_cleanup( 'cleanup_manager' => $cleanup_mgr_for_modify );
    my $state_obj_after_cleanup
        = $element->state( requested_fields => [ keys %modify_args ] );

    $Test->nlog( "Here is \$state_obj_before_modify: "
            . Dumper($state_obj_before_modify) );
    $Test->nlog( "Here is \$state_obj_after_cleanup: "
            . Dumper($state_obj_after_cleanup) );


    foreach my $key (keys %modify_args) {
        my $before_modify = $state_obj_before_modify->$key();
        my $after_modify  = $state_obj_after_cleanup->$key();
        if (!(  Data::Compare->new()
                ->Cmp($before_modify, $after_modify)
            )
            ) {
            run_cleanup('cleanup_manager' => $cleanup_mgr);
            NATE::BaseException->throw(
                "Cleanup failed to revert the attribute $key, which is not expected!"
            );
        }

    }
    run_cleanup( 'cleanup_manager' => $cleanup_mgr );
    $Log->exit();
}

sub _verify_rename {
    $Log->enter();
    my (@args) = @_;
    $Log->enter();

    my %opts = validate_with(
        params => \@args,
        spec   => {
            'element'         => { type => OBJECT },
            'rename_args'     => { type => HASHREF },
            'cleanup_manager' => { type => OBJECT, 'isa' => 'NACL::Cleanup' }
            , # rename_attribute is the attribute of the element which is being renamed and it is mandatory argument.
        }
    );

    my $element          = $opts{'element'};
    my $cleanup_mgr      = $opts{'cleanup_manager'};
    my %rename_args      = %{ $opts{'rename_args'} };
    my $rename_attribute = delete $rename_args{'rename_attribute'};

    if ( !$rename_attribute ) {
        NATE::BaseException->throw(
            "\"rename_attribute\" is mandatory attribute!");
    }
    my $cleanup_mgr_for_rename = create_cleanup_manager( 'random_db' => 1 );
    my $rename_attribute_before_rename = $element->$rename_attribute;

    $Test->step("Call \"rename\" and register for cleanup");
    $element->rename( %rename_args,
        'nacltask_cleanup_manager' => $cleanup_mgr_for_rename );
    my $rename_attribute_after_rename = $element->$rename_attribute;

    $Test->step(
        "Run cleanup and verify if the element is renamed back to its original name."
    );
    run_cleanup( 'cleanup_manager' => $cleanup_mgr_for_rename );
    $cleanup_mgr_for_rename->run();
    try {
        _verify_if_element_purged(
            'element'        => $element,
            'fail_if_exists' => 1
        );
    }
    catch NATE::BaseException with {
        $ex = shift;
        $Test->nlog(
            "Cleanup failed to rename the element back to its original name: $rename_attribute_before_rename, which is not expected!"
        );
        run_cleanup( 'cleanup_manager' => $cleanup_mgr );
        $ex->throw();

    };
    run_cleanup( 'cleanup_manager' => $cleanup_mgr );

    $Log->exit();
}

1;
