# $Id$
#
#Copyright (c) 2011 Network Appliance, Inc.
#All rights reserved.

package NACL::UnitTest::TestRequestedFields;

use strict;
use warnings;

use Params::Validate qw(validate SCALAR HASHREF ARRAYREF BOOLEAN);

use Tharn qw($Log);
use NACL::ComponentUtils qw(_dump_one);
use NATE::BaseException ();

BEGIN {
    use Exporter qw(import);
    our @EXPORT_OK=qw(
                      test_requested_fields
                     );
};

# A helper routine to compare various different kinds of 'fetch' calls
# with each other.

# This tries to do so in a way that is independent of exactly which
# component state object is being tested, though the approach may only
# work for some fields (hopefully most) of only some component states
# (hopefully most).

# One general requirement is that the fields on the element being
# tested are ones that aren't changing as test is being performed.

# The test begins by doing a base fetch call with the options given in
# 'common_options' and 'base_options' options (these options could
# include a 'filter' option, such as if there is a particular element
# you wish to test against.  They could include 'requested_fields'
# option, such as if you wish to make sure the base fetch includes
# certain hidden fields, but you will want to include all the values
# that you pass as part of the 'fields' option or else the comparisons
# described below will naturally fail), and the first component state
# returned determines the element and component state object against
# which future 'test' fetches will be compared.

# Then, for each field in the 'fields' option, a test fetch will be
# done with requested_fields set to just this field, and with 'filter'
# set to include the primary keys of the element under test, and with
# other options as passed in 'common_options' and 'test_options'
# options.  The test fetch of the particular field will be compared
# with the value gotten during the 'base' fetch call, and if any tests
# fail then an exception will be thrown (containing all the rolled up
# failures).

# common_options would usually set the command interface, plus any
# other options the base and test fetches have in common.  One way
# that the "base_options" and "test_options" may be used is to compare
# two different interfaces (so one could e.g. have base options
# include apiset_should=>{interface=>'CLI'} and test options include
# apiset_should=>{interface=>'ZAPI'} to compare CLI and ZAPI).  The
# "base_label" and "test_label" options let one customize the label
# given to each of the fetches in log messages and errors, such as to
# say "CLI" and "ZAPI" instead of the default labels "Base" and
# "Current".

sub test_requested_fields {
    my (%opts) = validate @_,
      {
        fetch_package   => { type => SCALAR },
        common_options  => { type => HASHREF, default => {} },
        base_options    => { type => HASHREF, default => {} },
        test_options    => { type => HASHREF, default => {} },
        base_label      => { type => SCALAR, default => "Base" },
        test_label      => { type => SCALAR, default => "Current" },
        fields          => { type => ARRAYREF },
      };
    my $base_state =$opts{fetch_package}->fetch
        (
         %{ $opts{common_options} },
         %{ $opts{base_options} },
        );
    my %base_primary_keys =
      $base_state->get_component_instance()->get_primary_keys_hash();
    my @fields = @{$opts{fields}};
    $Log->comment("Fields to test: @fields");

    my $message="";
    foreach my $field (@fields) {
        $Log->comment("Testing field $field\n");
	my $base_field_value = $base_state ? $base_state->$field() : "No component state object was returned";
        my $base_str = _dump_one( $base_field_value );
        $Log->comment("Base value: $base_str\n");
        my @test_state_array = $opts{fetch_package}->fetch(
            %{ $opts{common_options} },
            %{ $opts{test_options} },
            filter => { %base_primary_keys, $field => $base_field_value, },
            requested_fields => [$field],
	    allow_empty=>1,
        );
	if (@test_state_array != 1) {
	    $message .= "While fetching with field $field ($opts{test_label}), should have gotten 1 component state, not ".scalar(@test_state_array)."\n";
	}
	my $test_field_value = $test_state_array[0] ? $test_state_array[0]->$field() : "No component state object was returned";
        my $test_str = _dump_one( $test_field_value );
        $Log->comment("Test value: $test_str\n");
        if ( $test_str ne $base_str ) {
            $message .= ("Field $field has a different value when requested in a fetch than in the previous baseline fetch:\n"
                         . "  $opts{test_label} value: $test_str\n"
                         . "  $opts{base_label} value: $base_str\n" );
        }
    }
    if (length($message)) {
        NATE::BaseException->throw($message);
    }
}

1;
