package NACL::UnitTest::Utils;

use strict;
use warnings;
use feature 'state';
use Tharn qw($Log);


BEGIN {
    use Exporter qw(import);
    our @EXPORT_OK = qw(_object_check check_zapi_invalid_method
        create_node_with_invalid_build_path linked_component_test
        mcc_verification_builder pause_mcc_queue start_mcc_queue
        mock_internal_c_method_for_zapi 
        mock_zapi_fetch mock_c_method_zapi fixup_mockmodule);
    our %EXPORT_TAGS = ('all' => \@EXPORT_OK);
}

use Params::Validate qw(:all);
use Scalar::Util qw(blessed);
use NACL::ComponentUtils qw(_dump_one);
use NACL::C::DebugCrsConfiguration;
use NATE::Process;
use Test::MockModule;
use NACL::Expectation;
fixup_mockmodule();

use NATE::BaseException qw(:try);

sub _object_check {
    my %opts = validate_with(
        params => \@_,
        spec   => {
            object => {required => 1},
            type   => {type     => SCALAR}
        }
    );
    my $object = $opts{object};
    my $type   = $opts{type};

    unless (blessed($object) && $object->isa($type)) {
        NATE::BaseException->throw('Should have returned a blessed '
                . "object of type '$type' but instead returned a "
                . ref($object)
                . "'");
    }
}

sub mock_internal_c_method_for_zapi {
    $Log->enter();
    my @objects;
    my $mock_base_component = Test::MockModule->new('NACL::C::Component::ONTAP');
    my $mock_populate_jobs = sub { return 1;};
    $mock_base_component->mock('_populate_job_components', $mock_populate_jobs);
    push @objects, $mock_base_component;
    $Log->exit();
    return \@objects;
}

sub check_zapi_invalid_method {
    state $spec = {
        package => {type => SCALAR},
        method  => {type => SCALAR},
        args    => {type => HASHREF},
        tcd_obj => {type => OBJECT},
    };
    my %opts = validate_with(
        params => \@_,
        spec   => $spec,
    );
    my $pkg    = $opts{package};
    my $method = $opts{method};
    my $args   = $opts{args};
    my $tcd    = $opts{tcd_obj};
    my %args_copy = %$args;
    my $ci = delete $args_copy{command_interface};
    my $opts_str = "command_interface => $ci";
    if (keys %args_copy) {
        my $dump_one = _dump_one({%args_copy});
        $dump_one =~ s/^{//;
        $dump_one =~ s/}$//;
        $opts_str .= ", $dump_one";
    }    

    $tcd->nlog("Invoking $pkg->$method($opts_str)");    
    try {
        $pkg->$method(%$args, apiset_must => {interface => 'ZAPI'});
        NATE::BaseException->throw('Should have failed when attempted to '
            . 'run through ZAPI, but did not');
    } catch NACL::Exceptions::InvalidChoice with {
        my $exception = $_[0];
        $tcd->nlog("Could not run ZAPI as expected. Full error text:\n"
            . $exception->text());
    };
}

sub create_node_with_invalid_build_path {
    state $spec = {node => {type => OBJECT, isa => 'NACL::C::SystemNode'}};
    my %opts = validate_with(params => \@_, spec => $spec);
    require NACL::C::Node;
    my $new_node = NACL::C::Node->new(node => $opts{node}->node());

    # Create an APISet so that the version manager gets created.
    $new_node->apiset(interface => 'CLI');
    
    # Setting a dummy build path so that the help_xml is not used 
    $new_node->version_manager()->{build_path} = 'a';
    
    return $new_node;
}

sub linked_component_test {
    state $spec = {
        code => {type => CODEREF},
        expected_command  => {type => SCALAR},
    };
    my %opts = validate_with(
        params => \@_,
        spec   => $spec,
    );

    my $code = $opts{code};
    my $expected_command = $opts{expected_command};

    try {
        $code->();
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = shift;
        my $command  = $exception->get_response_object()->get_command();

        if ($command !~ /$expected_command/) {
            NATE::BaseException->throw("Should have run $expected_command ".
                 "but instead ran $command");
        }
    };
}

sub pause_mcc_queue {_mcc_queue_handler(@_, _val => 'true')}
sub start_mcc_queue {_mcc_queue_handler(@_, _val => 'false')}
sub _mcc_queue_handler {
    my %opts = @_;
    my $val  = delete $opts{_val};
    NACL::C::DebugCrsConfiguration->modify(%opts,
        'keep-sender-streams-paused' => $val);
}

sub mcc_verification_builder {
    state $spec = {
        package => {type => SCALAR},
        command_interface => {type => OBJECT, isa => 'NACL::C::CommandInterface::ONTAP'},
        tcd_obj => {type => OBJECT},
        create_opts => {type => HASHREF, default => {}},
    };
    my %opts = validate_with(params => \@_, spec => $spec);
    my $pkg = $opts{package};
    my $ci  = $opts{command_interface};
    my $tcd = $opts{tcd_obj};
    my $create_opts = $opts{create_opts};
    
    my %mcc_opts = (
        nacltask_to_check_mcc => 1,
        nacltask_mcc_verify_retry_count => 1,
        polling_interval_mcc_check => 1,
    );

    my $run_test = sub {
        my $description = $_[0];
        $tcd->description($description);
        my $cleanup = NACL::Cleanup->new(random_db => 1);
        try {
            pause_mcc_queue(command_interface => $ci);
            try {
                $pkg->create(
                    command_interface        => $ci,
                    nacltask_cleanup_manager => $cleanup,
                    nacltask_cleanup_args    => {nacltask_if_purged => 'pass'},
                    %mcc_opts,
                    %$create_opts,
                );
            } catch NACL::C::Exceptions::MCCVerificationFailed with {
                my $exception = $_[0];
                $tcd->nlog("MCC Verification failed as expected. Error:\n"
                    . $exception->text());
                my %primary_keys = $exception->primary_keys();
                
                start_mcc_queue(command_interface => $ci);
                $pkg->create_mcc_verification(
                    command_interface => $ci,
                    remote_interface  => $ci->mcc_remote_interface(),
                    %primary_keys,
                );
                
                pause_mcc_queue(command_interface => $ci);
                try {
                    $pkg->purge(
                        command_interface => $ci,
                        %primary_keys,
                        %mcc_opts
                    );
                } catch NACL::C::Exceptions::MCCVerificationFailed with {
                    my $inner_exception = $_[0];
                    $tcd->nlog("MCC Verification failed as expected. Error:\n"
                        . $inner_exception->text());
                };
            };
        } finally {
            start_mcc_queue(command_interface => $ci);
            $cleanup->run();
        };
    };
    
    $run_test->('Test for nacltask_to_check_mcc argument');
    
    delete $mcc_opts{nacltask_to_check_mcc};
    $pkg =~ /NACL::STask::(.*)/;
    my $obj_name = $1;
    NATE::Process->new(
        codespec   => $run_test,
        args       => ['Test for NACL_MCC_CHECK command-line parameter'],
        params     => {NACL_MCC_CHECK => 1},
        background => 0,
        runid      => "${obj_name}_NACL_MCC_CHECK",
        onexit     => \&NATE::Process::on_exit_propagate_worst_result,
    )->start();
}

sub mock_zapi_fetch {
    my %opts = validate_with(
        params => \@_,
        spec   => {
            api => {type => SCALAR},
            expected_apiset_input => {type => HASHREF, default => {}},
            typedef_name => {type => SCALAR, optional => 1},
            partial_parsed_output => {type => ARRAYREF, default => []},
            full_parsed_output => {type => ARRAYREF, optional => 1},
            verify_fields => {type => HASHREF, default => {}},
            fetch_inputs => {type => HASHREF, default => {}},
            tcd_obj      => {type => OBJECT, isa => 'TCD'},
            cs_pkg       => {type => SCALAR},
        },
    );
    my $api = $opts{api};
    my $expected_apiset_input = $opts{expected_apiset_input};
    my $typedef_name = $opts{typedef_name};
    my $partial_parsed_output = $opts{partial_parsed_output};
    my $full_parsed_output = $opts{full_parsed_output};
    if (!defined $full_parsed_output) {
        # If full_parsed_output is not provided, then
        # typedef_name is mandatory.
        validate_with(
            params => \@_,
            spec => {typedef_name => {type => SCALAR}},
            allow_extra => 1
        );
    }
    
    my $verify_fields = $opts{verify_fields};
    my %fetch_inputs = %{$opts{fetch_inputs}};
    my $tcd_obj = $opts{tcd_obj};
    my $cs_pkg = $opts{cs_pkg};
    my @mocks;
    
    my $mock_apiset = mock_zapi_apiset_input_check(
        expected_apiset_input => $expected_apiset_input,
        api                   => $api,
        tcd_obj               => $tcd_obj,
    );
    push @mocks, $mock_apiset;
    
    my $mock_response = Test::MockModule->new('NACL::APISet::Response::ZAPI::CMode');
    push @mocks, $mock_response;
    $mock_response->mock(
        'get_parsed_output',
        sub {
            my $expected_parsed_output = [];
            if (defined $full_parsed_output) {
                $expected_parsed_output = $full_parsed_output;
            } else {            
                $expected_parsed_output->[0]{'attributes-list'}[0]{$typedef_name} =
                    $partial_parsed_output;
                $expected_parsed_output->[0]{'num-records'} = scalar @$partial_parsed_output;
            }
            $tcd_obj->nlog("Setting parsed output as:\n" . Data::Dumper::Dumper $expected_parsed_output);
            return $expected_parsed_output;
        }
    );
    
    my @cs;
    try {
        my %step_inputs = %fetch_inputs;
        delete $step_inputs{command_interface};
        $tcd_obj->step('Test for call to fetch with options:' . _dump_one(\%step_inputs));
        
        $fetch_inputs{apiset_should}{interface} = 'ZAPI';
        @cs = $cs_pkg->fetch(%fetch_inputs);
        if (keys %$verify_fields) {
            $tcd_obj->nlog('Verifying fields on CS object: ' .
                Data::Dumper::Dumper $verify_fields);
            $cs[0]->verify_fields(%$verify_fields);
            $tcd_obj->nlog('CS fields are verified');
        }
    } finally {
        map {$_->unmock_all()} @mocks;
    };

    return @cs;
}

sub mock_c_method_zapi {
    my %opts = validate_with(
        params => \@_,
        spec   => {
            api => {type => SCALAR},
            expected_apiset_input => {type => HASHREF, default => {}},
            c_inputs => {type => HASHREF, default => {}},
            tcd_obj      => {type => OBJECT, isa => 'TCD'},
            c_pkg       => {type => SCALAR},
            method      => {type => SCALAR},
        },
    );
    my $api = $opts{api};
    my $expected_apiset_input = $opts{expected_apiset_input};
    
    my %c_inputs = %{$opts{c_inputs}};
    my $tcd_obj = $opts{tcd_obj};
    my $c_pkg = $opts{c_pkg};
    my $method = $opts{method};
    
    my $mock_apiset = mock_zapi_apiset_input_check(
        expected_apiset_input => $expected_apiset_input,
        api                   => $api,
        tcd_obj               => $tcd_obj,
    );
    
    my @ret;
    try {
        my %step_inputs = %c_inputs;
        delete $step_inputs{command_interface};
        $tcd_obj->step("Test for call to $c_pkg->$method() with options:". _dump_one(\%step_inputs));
        
        $c_inputs{apiset_should}{interface} = 'ZAPI';
        @ret = $c_pkg->$method(%c_inputs);
    } finally {
        $mock_apiset->unmock_all();
    };

    return wantarray ? @ret : $ret[0];
}

# Test::MockModule fails for hyphenated subroutine names, even though
# they're valid, so override the check here.
sub fixup_mockmodule {
    no strict 'refs';
    no warnings 'redefine';
    *{"Test::MockModule::_valid_subname"} = sub { return 1 };
    use strict 'refs';
    use warnings 'redefine';    
}

sub mock_zapi_apiset_input_check {
    my %opts = validate_with(
        params => \@_,
        spec   => {
            expected_apiset_input => {type => HASHREF, default => {}},
            api                   => {type => SCALAR},
            tcd_obj               => {type => OBJECT, isa => 'TCD'},
        },
    );
    my $expected_apiset_input = $opts{expected_apiset_input};
    my $api                   = $opts{api};
    my $tcd_obj               = $opts{tcd_obj};
    my $mock_apiset = Test::MockModule->new('NACL::APISet::Node::ZAPI::CMode');
    my @mocks;
    push @mocks, $mock_apiset;
    $mock_apiset->mock(
        $api,
        sub {
            my ($self, %opts) = @_;
            if (keys %$expected_apiset_input) {
                $tcd_obj->nlog('Checking that APISet input is ' .
                    Data::Dumper::Dumper ($expected_apiset_input));
                this(\%opts)->should(Be->deeply_equal_to($expected_apiset_input));
                $tcd_obj->nlog('APISet input is as expected');
            }
            require NACL::APISet::Response::ZAPI::CMode;
            return bless {}, 'NACL::APISet::Response::ZAPI::CMode';
        }
    );
    return $mock_apiset;
}

1;
