package NACL::UnitTest::_Mixins;

use strict;
use warnings;

use Tharn qw($Log);
use Params::Validate qw(validate_with :types);
use NACL::Expectation;

BEGIN {
    use Exporter qw(import);
    our @EXPORT_OK = qw(_test_idle_timeout_param_helper);
    our %EXPORT_TAGS = ('all' => \@EXPORT_OK);
}

# Test for the idle_timeout parameter (NACL_CONNECTION_IDLE_TIMEOUT)
# The approach is to create an APISet object in a sub-process with this
# parameter specified. We snooze after APISet creation is complete,
# so that the idle_timeout is hit. After the snooze, we check that the connection
# has been disconnected. We also run a command after it to ensure the
# connection has been re-established as expected.
sub _test_idle_timeout_param_helper {
    my %opts = validate_with(
        params => \@_,
        spec   => {
            tcd        => {type => OBJECT, isa => 'TCD'},
            apiset_sub => {type => CODEREF}
        }
    );
    my $tcd        = $opts{tcd};
    my $apiset_sub = $opts{apiset_sub};
    require NATE::Process;
    
    # Don't set this to too low a value, else it will end up getting
    # disconnected in Connectrec->new() itself, so the call will fail
    # Setting to a value of 10 so that if the prompt takes some time to be
    # returned (was intermittently happening in Smokes; see burt 955086), then
    # it does not fail.
    my $idle_time = 10;
    NATE::Process->new(
        codespec => sub {
            $tcd->nlog("Create an APISet object with idle_timeout=$idle_time");
            my $apiset = $apiset_sub->();
            
            $tcd->nlog('Get original cluster identity details');
            my $orig_identity = $apiset->cluster_identity_show()->get_parsed_output();
            
            my $snooze_time = $idle_time + 1;
            $tcd->nlog("Snooze for $snooze_time seconds so that idle_timeout is hit");
            Tharn::snooze($snooze_time);
            
            $tcd->nlog('Check if connection is disconnected');
            my $is_connected = $apiset->get_connection()->is_connected(
                verify => 0);
            this ($is_connected)->should(Be->false());
            $tcd->nlog('Connection is disconnected, as expected');
            
            $tcd->step('Basic sanity test that the APISet is  still usable '
                . 'even after the idle_timeout has been hit');
            my $final_identity = $apiset->cluster_identity_show()->get_parsed_output();
            
            # If the APISet session correctly reapplied all the settings, this
            # parsed output should match the original parsed output.
            this ($final_identity)->should(Be->deeply_equal_to($orig_identity));
        },
        background => 0,
        runid      => 'idle_timeout_param',
        onexit     => \&NATE::Process::on_exit_propagate_worst_result,
        params     => {NACL_CONNECTION_IDLE_TIMEOUT => $idle_time},
        inherit    => 1,
    )->start();
}

1;
