
#
# Copyright (c) 2016 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Singleton containing quota library configuration
## @author bare@netapp.com, ng-quota-qa@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::Util::QuotaLibSingleton

=head1 DESCRIPTION

C<NACL::Util::QuotaLibSingleton> is a singleton class that manages and stores
quota library configuration.

=head1 PARAMETERS

=over

=item C<< QUOTA_LIB_APISET_SHOULD_... >>

=item C<< QUOTA_LIB_APISET_MUST_... >>

NATE parameters in the form above will cause the fetch methods for
quota-related subclasses of L<NACL::CS::ComponentState|lib-NACL-CS-ComponentState-pm>
(e.g., L<NACL::CS::VolumeQtree|lib-NACL-CS-VolumeQtree-pm>,
L<NACL::CS::VolumeQuota|lib-NACL-CS-VolumeQuota-pm>, and
L<NACL::CS::VolumeQuotaReport|lib-NACL-CS-VolumeQuotaReport-pm>) to pick
up the given rules as the defaults for their apiset_should and
apiset_must rulesets.

For example, the following invocation:

  ntest QUOTA_LIB_APISET_SHOULD_INTERFACE=ZAPI \
        QUOTA_LIB_APISET_SHOULD_ZAPI_TYPE=non-iter \
        path/to/my/test.thpl

will cause the quota-related component state classes to prefer the non-iter
ZAPI interface when executing the test script.

For more information on the form of these parameters and the apiset_should
and apiset_must rulesets, see the
L<Rule Sets|http://brewery.netapp.com/Brewery/QA/Tools/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#rule-sets>
section of the
L<NACL Component Layer User's Guide|http://brewery.netapp.com/Brewery/QA/Tools/NACL/UsersGuide/ComponentLayerUsersGuide/index.html>.
Specifically, the parameters work as described in the
L<As a Test Parameter|http://brewery.netapp.com/Brewery/QA/Tools/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#as-a-test-parameter>
sub-section, but will apply to all quota-related component state classes,
rather than to specific command interfaces.

=item C<< QUOTA_LIB_SNMP_REPORT_TABLE >>

Selects which table will be used for SNMP-based quota reports.  This may
either be C<qrV2Table> or C<qrV3Table>.  The default is C<qrV3Table>.

=back

=cut

package NACL::Util::QuotaLibSingleton;

use strict;
use warnings;
use NACL::ResourceRecords;
use NATE::Log qw(log_global);
use NATE::ParamSet qw(param_global);
use NATE::Result;
use Params::Validate qw(HASHREF OBJECT validate_pos validate_with);

my $Log = log_global();

use Class::MethodMaker [
    new    => [qw(-hash _new)],
    scalar => 'snmp_report_table',
];

=head1 ATTRIBUTES

=head2 snmp_report_table

Name of the SNMP table to be used for SNMP-based quota reports.

=head1 METHODS

=cut

sub _build_ruleset_helper {
    my $self          = shift;
    my $which_ruleset = shift;
    my %opts          = validate_with(
        params      => \@_,
        spec        => {
            command_interface => { type => OBJECT, optional => 1, },
            $which_ruleset    => { type => HASHREF, default => {}, },
        },
        allow_extra => 1,
    );

    my %out_ruleset   = %{$opts{$which_ruleset}};
    my $quota_ruleset = $self->{resource_record}->$which_ruleset();
    my $ci_ruleset    = $opts{command_interface}
                            ? $opts{command_interface}->$which_ruleset()
                            : {};
    foreach my $key (NACL::ChooseAPISet::ruleset_keys()) {
        foreach my $source ($quota_ruleset, $ci_ruleset) {
            if (!exists($out_ruleset{$key}) && exists($source->{$key})) {
                $out_ruleset{$key} = $source->{$key};
                last;
            }
        }
    }
    return \%out_ruleset;
}

=head2 apiset_must

  my $quota_singleton = NACL::Util::QuotaLibSingleton->get();
  my $apiset_must     = $quota_singleton->apiset_must(%opts);

Builds an apiset_must ruleset to be applied in the quota libraries.  This
resolves the ruleset based on up to three inputs:  any explicit apiset_must
ruleset passed as an argument, any default apiset_must ruleset configured
for the quota library (via the C<QUOTA_LIB_APISET_MUST_...> parameters), and
any default apiset_must ruleset configured for a command_interface passed as
an argument.  The ruleset is returned as a hash reference.

=cut

sub apiset_must {
    my $self = shift;
    return $self->_build_ruleset_helper('apiset_must', @_);
}

=head2 apiset_should

  my $quota_singleton = NACL::Util::QuotaLibSingleton->get();
  my $apiset_should   = $quota_singleton->apiset_should(%opts);

Builds an apiset_should ruleset to be applied in the quota libraries.  This
resolves the ruleset based on up to three inputs:  any explicit apiset_should
ruleset passed as an argument, any default apiset_should ruleset configured
for the quota library (via the C<QUOTA_LIB_APISET_SHOULD_...> parameters), and
any default apiset_should ruleset configured for a command_interface passed as
an argument.  The ruleset is returned as a hash reference.

=cut

sub apiset_should {
    my $self = shift;
    return $self->_build_ruleset_helper('apiset_should', @_);
}

=head2 get

  my $quota_singleton = NACL::Util::QuotaLibSingleton->get();

Retrieves the NACL::Util::QuotaLibSingleton, constructing it if necessary.

=cut

my $Singleton;

sub get {
    my $pkg = shift;
    validate_pos(@_);

    if (defined($Singleton)) {
        return $Singleton;
    }

    my $snmp_report_table = param_global()->get(
        'QUOTA_LIB_SNMP_REPORT_TABLE',
        default => 'qrV3Table',
    );
    if ($snmp_report_table !~ /^qrV[23]Table$/) {
        $Log->log(NATE::Result::Warn->new("The value of the "
              . "QUOTA_LIB_SNMP_REPORT_TABLE parameter ($snmp_report_table) "
              . 'is invalid; using the default value (qrV3Table).'
        ));
        $snmp_report_table = 'qrV3Table';
    }
    my $self = $pkg->_new(snmp_report_table => $snmp_report_table);
    $self->{resource_record} = NACL::ResourceRecords::_find_rules(
        prefix  => 'QUOTA_LIB',
    );
    return $self;
}

=head1 EXAMPLE

Code similar to the following can be used in the fetch methods of classes
derived from L<NACL::CS::ComponentState|lib-NACL-CS-ComponentState-pm>,
and will cause the fetch method to pick up any apiset_must/apiset_should
rulesets configured via the C<QUOTA_LIB_...> parameters.

  sub fetch {
      my ($self, %opts)    = @_;
      my $quota_singleton  = NACL::Util::QuotaLibSingleton->get();
      $opts{apiset_must}   = $quota_singleton->apiset_must(%opts);
      $opts{apiset_should} = $quota_singleton->apiset_should(%opts);
      my @state_objs = $self->SUPER::fetch(
          %opts,
          choices => [
              ...
          ],
      );
      return @state_objs;
  }

=cut

1;
