# -*-perl-*-
#
# Copyright (c) 2015,.. NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary DPG Workloads
## @author ng-smgr-auto@netapp.com
## @status shared
## @pod here

package NACL::AppBuilder::DpgWorkflows::Workload;

use strict;
use warnings;

use Params::Validate qw(validate_with :types);
use Data::Dumper;
use Carp;

use Tharn;
use NATE::BaseException qw( :try );
use NATE::Log qw(log_global);

use NACL::Exceptions::InvalidParam;
use NACL::C::UnitNormalization qw(convert_to_bytes);
use NACL::C::Client;
use NACL::C::Cluster;
use NACL::CS::Snapmirror;
use NACL::CS::LunIgroup;
use NACL::STask::Volume;
use NACL::STask::VolumeSnapshot;
use NACL::STask::VolumeFile;
use NACL::STask::VolumeFileClone;
use NACL::MTask::Traffic;
use NACL::AppBuilder::WFC::IO;
use NACL::STask::Client::Mount;
use NACL::C::Slist;

use NACL::AppBuilder::DpgWorkflows::Fiji qw(
    fiji_unsplit_clones
    fiji_wait
    fiji_clear_all
    fiji_wafl_lsa_punch_holes_worker_delay
);

use NATE::Process;

use NACL::GeneralUtils qw(
    random_name_generator
    nacl_method_retry
);

use San::Task::IOWorkloadHelper;
use IOMan;
use dp::snapmirror::paloma::vault::lib::VaultCommonUtils qw( dataset );
use NACL::AppBuilder::DpgWorkflows::Validation qw(
    get_composite_aggr_stats
    lsa_validation
    lsa_verify_in_progress
    lsa_count_verify
);
use dp::snapmirror::paloma::vault::lib::VaultCommonUtils
    qw( create_snapshot
);

use base qw(Exporter);

our @EXPORT_OK = qw(
        _get_inum
        _punch_holes_in_files
        nfs_metadata_ops
        trunc_file
        modify_file_attrs
        mount_nfs_vol
        start_nfs_traffic
        file_punch_hole
        create_empty_file_with_size
        start_traffic
        start_read_on_dst_cg_files
        prepare_for_traffic
        prepare_dst_luns
        Wf_start_traffic
        manipulate_hardlinks
	      datagen
        Wf_create_fileclones_filestodirectories
        Wf_create_multiplesize_filesnclones
        DataSet_GrowClones_PunchHoles
        DataSet_GrowParents_PunchHoles
        populate_hecr_dataset
        Wf_create_dirty_ss
        Wf_data_transitions
        Wf_cad_test
        cad_test
        cadchk
        Wf_EID_Holes
        compression_dataset
        movedirs
        modify_files2dir_vc
        populate_modify_symlinks
        populate_lcp_dataset
        _execute_iofence_cmd
        fg_set_placement
        fg_populate
        populate_with_datagen
	start_fg_read_traffic
	stop_fg_read_traffic
	files_populate
	files_modify
	ldr_deterministic_dataset
	fg_create_hundreds_of_files_and_links
        fg_modify_hundreds_of_files_and_links
	find_files_dir_links
        create_delete_fifo
        create_delete_devices
        populate_create_tree
        appcg_populate_overwrite
        appcg_dataset_secondary
        appcg_dataset_adaptive
        LSA_file_truncate
        LSA_file_creation_IO_DT
        populate_lsa_dataset
);

use POSIX;
use vars qw (
            $counter );

our %EXPORT_TAGS = (
    'all' => \@EXPORT_OK,
);

=head1 NAME

NACL::AppBuilder::DpgWorkflows::Workload

=head1 DESCRIPTION

This library contains i/o workloads

=head1 METHODS

=cut

=head2 nfs_metadata_ops

This method contains the nfs metadata ops like truncate file and creating,removing symlinks

=over

=item example

=item options

=over

=item mount_point

(required) scalar value - the mount point of volume

=item size_to

(required)  integer value - the percent value to which file should should be reduced .

=item Client

(required) Client Object - the nfs client object.

=item files

(required)  Array - array of file names.

=back

=back

=cut

sub nfs_metadata_ops {

    my %args = validate_with(
        params      => \@_,
        allow_extra => 0,
        spec        => {
            mount_point => { type => SCALAR, optional => 0 },
            size_to => { type => SCALAR, optional => 0 },
            Client  => {type => OBJECT, optional => 0 },
            files  =>  {type => OBJECT, optional => 0 },
        },
    );

    my $mount_point = $args{'mount_point'};
    my $size_to     = $args{'size_to'};
    my $Client    = $args{Client};
    my @files     = $args{files};

    trunc_file($mount_point,$size_to,@files);
    modify_file_attrs($mount_point,$Client,@files);
}

=head1 NAME

NACL::AppBuilder::DpgWorkflows::Workload

=head1 DESCRIPTION

This library contains i/o workloads

=head1 METHODS

=cut

=head2 trunc_file

This method contains the nfs metadata ops truncate file.
Given a size this method will truncate the file to a that size.

=over

=item example

=item options

=over

=item mount_point

(required) scalar value - the mount point of volume

=item size_to

(required)  integer value - the percent value to which file should should be reduced .

=item files

(required)  Array - array of file names.

=item client
(required) OBJECT - the client object on which the truncate command should be run

=item flag_on_fail
(Optional) SCALAR - Whether or not to log an error if the truncate fails.
(Default): 1

=back

=back

=cut
sub trunc_file {
    my %args = validate_with(
        params      => \@_,
        allow_extra => 0,
        spec        => {
            client          => { type => OBJECT, },
            mount_point     => { type => SCALAR, optional => 0 },
            size_to         => { type => SCALAR, optional => 0 },
            files           => {type => ARRAYREF, optional => 0},
            flag_on_fail    => { type => SCALAR, optional => 1, default => 1 },
        },
    );

    my $client      = $args{client};
    my $mount_point = $args{'mount_point'};
    my $size_to     = $args{'size_to'};
    my @files       = @{$args{files}};
    my $flag_on_fail = $args{flag_on_fail};
    my $log         = log_global();

    foreach my $file(@files){
        $log->comment("Truncating file $file to size : $size_to");
        if ($size_to =~ /%^/) {
           my $percent = $size_to;
           chop $percent;
           my $size = (stat "$mount_point/$file")[7];
           $size_to = int(($percent / 100) * $size);
        }
        my $fh =  $mount_point .'/' . $file;
        my $cmd = "truncate -s $size_to $fh";
        $log->comment('Running command: ' . $cmd);
        $client->apiset( interface => "CLI")->execute_raw_command(
                                command     => $cmd,
                       'connectrec-timeout' => 300,);


        $log->comment('Confirm that the size of the file is: ' . $size_to);
        my $op =  $client->apiset( interface => "CLI")->execute_raw_command(
                                   command   => "du -b $fh");

        $log->comment('File size after the truncate operation: ' . Dumper($op));

        if ($op =~ /$size_to\s+$fh/) {
            $log->comment("Truncated file" . "$mount_point/$file" ."to size : $size_to\n");
        } else {
           if ($flag_on_fail) {
               $log->error("truncate failed on file". "$mount_point/$file" .": $!");
           } else {
                $log->warn("truncate failed on file". "$mount_point/$file" .": $!");
           }
        }

    }
    return;
}

=head1 NAME

NACL::AppBuilder::DpgWorkflows::Workload

=head1 DESCRIPTION

This library contains i/o workloads

=head1 METHODS

=cut

=head2 modify_file_attrs

This method contains the nfs metadata ops - creating,removing symlinks

=over

=item example

=item options

=over

=item mount_point

(required) scalar value - the mount point of volume

=item Client

(required) Client Object - the nfs client object.

=item files

(required)  Array - array of file names.

=back

=back

=cut
sub modify_file_attrs {
##create symlinks, remove symlinks
my %args = validate_with(
        params      => \@_,
        allow_extra => 0,
        spec        => {
            mount_point => { type => SCALAR, optional => 0 },
            Client  => {type => OBJECT, optional => 0},
            files  =>  {type => ARRAYREF, optional => 0},
        },
    );

    my $mount_point = $args{mount_point};
    my $Client    = $args{Client};
    my @files     = @{$args{files}};

my $data_info;
my $Client_apiset = $Client->apiset();
my @slink_list;
foreach my $file (@files) {
    my $slink = "$mount_point/more_sl_$file";
    push @slink_list, $slink;
    $Client_apiset->set_privilege( 'privilege-level' => 'root', );
    $Client_apiset->touch( 'path' => "$mount_point/$file",
                           'connectrec-timeout' => 3600,);
    $Client_apiset->ln(files => "$mount_point/$file",
                       'link-name' => $slink,
                       'symlink' => 1,
                       'connectrec-timeout' => 3600,
                      );
    }
    $data_info->{slinks} = \@slink_list;

    foreach my $sl (@{$data_info->{slinks}}) {
                $Client_apiset->rm( paths => $sl,
                                    'connectrec-timeout' => 3600,);
            }

}


=head2 mount_nfs_vol

This method is to mount a nfs volume.

It will create the junction-path for the volume if not created and mount the volume on the client provided.
It returns the mount point created.

=over

=item options

=over

=item volume

(required) object - volume to mount

=item client_name

(required)  scalar - nfs client name

=item nfsvers

(optional)  scalar - nfs version for the mount point. Defaults to 3.

=back

=back

=cut

sub mount_nfs_vol {
    my %args = validate_with(
        params      => \@_,
        allow_extra => 0,
        spec        => {
                volume       => { type => OBJECT, optional => 0 },
                client_name  => { type => SCALAR, optional => 0 },
                nfsvers      => { type => SCALAR, optional => 1, default => '3' },
        },
    );
    my $log = log_global();

    my $volume = $args{'volume'};
    my $vers = $args{'nfsvers'};

    my $client_obj = NACL::C::Client->new( name => $args{'client_name'} );

    if ($volume->get_one_state_attribute('junction-path') eq "-") {
        $log->log('Will create junction-path');
        $volume->mount( 'junction-path' => "/".$volume->volume() );
    }

    my $mount = NACL::STask::Client::Mount->create(
            command_interface   => $client_obj,
            server_command_interface => $volume->command_interface,
            vserver             => $volume->vserver,
            volume              => $volume,
            vers                => $vers,
    );

    return $mount->mount_point();
}

# similar to mount_nfs_vol but will return mount obj an instance of NACL::C::Client::Mount
sub mount_nfs_vol2 {
    my %args = validate_with(
        params      => \@_,
        allow_extra => 0,
        spec        => {
                volume       => { type => OBJECT, optional => 0 },
                client_name  => { type => SCALAR, optional => 0 },
                nfsvers      => { type => SCALAR, optional => 1, default => '3' },
        },
    );
    my $log = log_global();
    my $mount_point = mount_nfs_vol(@_);
    $log->log('Mount point is '.$mount_point);
    my $client_obj = NACL::C::Client->new( name => $args{'client_name'} );
    my $mount_obj = NACL::C::Client::Mount->find(command_interface => $client_obj,
                                                filter => { mount_point => $mount_point },
                                                );
    return $mount_obj;
}


=head2 start_nfs_traffic

This method is to run sio on NFS volume files.

=over

=item options

=over

=item required

(required) object - nfs client to run IOs.
(required) array - absolute file paths for the files to run IOs

=item optional

    tool           - io tool. default is SIO.
    file_size      - file size for IOs. default is 64k.
    block_size     - IO block size, default is 64k.
    num_of_threads - SIO threads, default is 1.
    rand           - random or sequential. 0-100. default is 0.
    read           - read percentage. 0-100. default is 0.
    start_byte     - start byte. default is 0.

=back

=back

=cut

sub start_nfs_traffic {

    my %args = validate_with(
        params => \@_,
        spec   => {
            client         => { type => OBJECT, optional => 1 },
            paths          => { type => ARRAYREF, optional => 1 },
            tool           => { type => SCALAR, optional => 1, default => 'sio' },
            file_size      => { type => SCALAR, optional => 1, default => '500m' },
            block_size     => { type => SCALAR, optional => 1, default => '4k' },
            num_of_threads => { type => SCALAR, optional => 1, default => '1' },
            rand           => { type => SCALAR, optional => 1, default => '0' },
            read           => { type => SCALAR, optional => 1, default => '0' },
            start_byte     => { type => SCALAR, optional => 1, default => '0' },
            traffic_object => { type => OBJECT, optiona => 1, default => undef },
        },
    );

    my @traffic_objs;
    my $log = log_global(1);
    if($args{'tool'} eq 'sio') {
        if (!$args{traffic_object} &&
            (!defined $args{client} || !defined $args{paths})) {
            die "Manadatory params client and/or paths missint \n";
        }
        my %sio_args = (
            verify         => '1',
            break_on_dc    => '1',
            tool              => $args{'tool'},
            file_size         => $args{'file_size'},
            block_size        => $args{'block_size'},
            read              => $args{'read'},
            num_of_threads    => $args{'num_of_threads'},
            rand              => $args{'rand'},
            'privilege-level' => 'root',
            start_byte        => $args{'start_byte'},
            rep_latency    => '60',
            direct          => '1',
        );
        if (!defined $args{traffic_object}) {
            foreach my $path (@{$args{'paths'}}) {
                my $traffic_obj = NACL::MTask::Traffic->start(
                    command_interface => $args{'client'},
                    paths             => $path,
                    %sio_args,
                );
                push @traffic_objs, $traffic_obj;
            }
        } else {
            $log->comment('sio opts ' . Dumper(\%sio_args));
            $log->comment('Traffic object provided resuing it to start traffic');
            my $tobj = $args{traffic_object};
            $log->trace('Traffic object in start_nfs_traffic ' . Dumper($tobj));
            $tobj->start(%sio_args);
            push @traffic_objs, $tobj;
        }
    } else {
        die "unsupported IO tool - $args{'tool'}.\n";
    }
    $log->exit();
    return @traffic_objs;
}


=head2 file_punch_hole

This method is to punch holes on NFS volume file.

=over

=item options

=over

=item required

(required) node - object, nfs client to run IOs.
(required) file_path - SCALAR, absolute file path.
(required) offset - SCALAR, offset to punch hole.
(required) length - SCALAR, length to punch hole.It must be multiple of 4096

=back

=back

=cut

sub file_punch_hole {

    my %args = validate_with(
        params => \@_,
        spec   => {
            node      => { type => OBJECT, optional => 0 },
            file_path => { type => SCALAR, optional => 0 },
            offset    => { type => SCALAR, optional => 0 },
            length    => { type => SCALAR, optional => 0 }, #multiple of 4096
        },
    );

    my $cmd = "file punchhole " . "$args{file_path} " . "$args{offset} ". "$args{length}";
    my $api = $args{node}->get_7m_or_nodescope_apiset();
    $args{node}->privilege_level('test');

    return $api->execute_command( command => $cmd, );

}


=head2 create_empty_file_with_size

This method is to create an empty file using host tool truncate.

=over

=item options

=over

=item required

(required) client - SCALAR, name of the client to mount volume and create file
(required) file_path - SCALAR, absolute file path.
(required) size - SCALAR, file size.

=back

=back

=cut

sub create_empty_file_with_size{

    my %args = validate_with(
        params => \@_,
        spec   => {
            client    => { type => SCALAR, optional => 0 },
            file_path => { type => SCALAR, optional => 0 },
            size      => { type => SCALAR, optional => 0 },
        },
    );

    my $Client = NACL::C::Client->new( name => $args{client} );
    my $cmd = "/usr/software/bin/sudo truncate --size " . "$args{size} " . $args{file_path};
    $Client->apiset(interface => "CLI")->execute_raw_command(
            command             => $cmd,
           'connectrec-timeout' => 300);

}

=head2 Wf_start_traffic

A wrapper which internally calls prepare_for_traffic() and

    - creates start and stop traffic handles
    - actually starts the traffic
    - return the start and stop traffic handles to users for control.
    - can call this method to start traffic on src and/or dst

Note: Calling this method is really useful when you want to start traffic to both
src and dst.

=over

=item Example

    my $traffic_res = Wf_start_traffic (
                                      node_src  => $node_src,
                                      node_dst  => $node_dst,
                                      cg_type   => 'lun',
                                      lun_src   => $args{lun_src},
                                      lun_dst   => $args{lun_dst},
                                      sanclients => $args{sanclients});

Returns an hashref of start and stop traffic handles for both src and dst.

Example output:

    $VAR1 = {
        'src'   => {
            '_tstart' => sub { },
            '_tstop'   => sub { },
        },
        'dst'   => {
            '_tstart' => sub { },
            '_tstop'   => sub { },
        }
    }


=item Options

=over

=item node_src

(Optional) The source node objects
Required if lun_src is defined.

=item node_dst

(Optional) The dst  node objects
Required if lun_dst is defined.

=item sanclients

(Rquired) The list of clients for doing SAN traffic.


=item SAN_workload_type

(Optional) The type of SAN workload to invoke
(Default) SAN_IO_Random_BS


=item lun_src

(Optional) The list of src luns to start IO on

=item lun_dst

(Optional) The list of dst luns to start IO on

=item SAN_workload_type

(Optional) The type of SAN workload to invoke
(Default) SAN_IO_Random_BS

=item cg_type

(Optional) The cg type on which IO will be started - lun/file
(Default) lun

=back

=back

=cut

sub Wf_start_traffic {

    my %args = validate_with(
        params => \@_,
        spec   => {
            node_src    => { type => SCALAR|HASHREF, optional => 1 },
            node_dst    => { type => SCALAR|HASHREF, optional => 1 },
            cg_type     => { type => SCALAR, regex => qr/^(file|lun)$/, default => 'lun', },
            lun_src     => { type => ARRAYREF, optional => 1 },
            lun_dst     => { type => ARRAYREF, optional => 1 },
            sanclients  => {  type => ARRAYREF|SCALAR, optional=> 0 },
            SAN_workload_type =>  =>  { type => SCALAR, optional => 1, default => 'SAN_IO_Random_BS'},
        },
    );

    my $node_src = $args{node_src};
    my $node_dst = $args{node_dst};
    my $cg_type  = $args{cg_type};
    my $lun_src  = $args{lun_src};
    my $lun_dst  = $args{lun_dst};
    my $sanclients  = $args{sanclients};
    my $workload = $args{SAN_workload_type};
    my $traffic_res;
    my @clients;

    if ( ($node_src) && (ref($node_src) !~ /Node/i) ) {
        $node_src =  NACL::STask::Node->new( name => $node_src);
    }
    if ( ($node_dst) && (ref($node_dst) !~ /Node/i) ) {
        $node_dst = NACL::STask::Node->new( name => $node_dst);
    }

    if (ref($sanclients) eq "ARRAY") {
        @clients =  @{$sanclients};
    } else {
        @clients = split ',', $sanclients;
    }

    my $log = log_global(1);
    $log->enter();
    my %traffic_params;
    my %resources;

    if (defined $lun_dst) {
        if (!defined $node_dst) {
            NACL::Exceptions::InvalidParam->throw('node_dst is required param when lun_dst is defined');
        }

        #setting the required traffic params
        $traffic_params{dst}{command_interface} = $node_dst;
        $traffic_params{dst}{cg_item_type}      = $cg_type;
        $traffic_params{dst}{sanclients}        = \@clients;
        $resources{dst}{luns}                   = $lun_dst;
    }

    if (defined $lun_src) {
        if (!defined $node_src) {
            NACL::Exceptions::InvalidParam->throw('node_src is required param when lun_src is defined');
        }

        #setting the required traffic params
        $traffic_params{src}{command_interface} = $node_src;
        $traffic_params{src}{cg_item_type}      = $cg_type;
        $traffic_params{src}{sanclients}        = \@clients;
        $resources{src}{luns}                   = $lun_src;
    }

    $log->comment('Attempting to start traffic');
    foreach my $target (keys %traffic_params) {
       my %args = %{ $traffic_params{$target} };
       $log->comment('.. preparing on ' . $target . ' with args: ' . Dumper(%args));
       ($traffic_res->{$target}->{_tstart}, $traffic_res->{$target}->{_tstop}) = prepare_for_traffic (%args);

        $log->comment('... starting on ' . $target);
        $traffic_res->{$target}->{_tstart}->( luns      =>  $resources{$target}{luns});
    }

    $log->trace('Traffic hash being returned: ' . Dumper($traffic_res));
    $log->exit();

    return $traffic_res;
}

=head1 NAME

NACL::AppBuilder::DpgWorkflows::Workload

=head1 DESCRIPTION

This library contains i/o workloads

=head1 METHODS

=cut

=head2 manipulate_hardlinks

This method provides the option of creating, adding or deleting hardlinks to a file.
When creating hardlinks for the first time, use the 'create' operation. The return value is an array of hardlinks to a file.
If hardlinks are already created for a file and you want to add more hardlinks, use the 'add' operation. When using this operation, pass the existing array of hardlinks to the file.
Use 'delete' operation to delete hardlinks to a file. When using this operation, pass the existing array of hardlinks to the file.

=over

=item example
        my $ret = manipulate_hardlinks(
                'client'       => $client,
                'file_path'    => $src_file1_path,
                'dir_path'     => $src_dir_path,
                'hardlink_list'=> \@hl_array
                'count'        => 2,
                'operation'    => 'add',
            );

    Example output:
        $VAR1 = ['...',
                 '...',
                 '...',];
=item options

=over

=item Client

(required) Client Object - the nfs client object.

=item file_path

(required)  scalar value - hardlink will point to this file path.

=item dir_path

(required) scalar value - the directory path where hardlink will be created.

=item hardlink_list

(optional) arrayref - list of hardlinks pointing to the same file.

=item count

(optional) scalar value - number of hardlinks to be created, added or deleted. (default count=1)

=item operation

(optional) scalar value - the operation to be performed on the hardlinks (default operation=create)
                        create: when creating hardlinks for the first time
                        add   : use this option when hardlinks are already present for a file and you would like to add more hardlinks pointing to the same file
                        delete: delete hardlink to a file
                        modify: Run the touch command on the parent file (which we can get from file_path). This will automatically modify all the hardlinks
                                pointing to that parent file.

=back

=back

=cut
sub manipulate_hardlinks {
    my %args = validate_with(
        params      => \@_,
        allow_extra => 0,
        spec        => {
            client        => {type => OBJECT,   optional => 0},
            file_path     => {type => SCALAR,   optional => 0},
            dir_path      => {type => SCALAR,   optional => 0},
            hardlink_list => {type => ARRAYREF, optional => 1},
            count         => {type => SCALAR,   optional => 1, default => 1},
            operation     => {type => SCALAR, optional => 1, default => 'create', regex => qr/create|add|delete|modify/, },
            prefix        => {type => SCALAR|UNDEF, optional => 1, },
        },
    );

    my $dir_path = $args{dir_path};
    my $client      = $args{client};
    my $file_path = $args{file_path};
    my $count     = $args{count};
    my $operation = $args{operation};

    my $data_info;
    my $Client_apiset = $client->apiset();
    my @hardlink_list;
    my $log = log_global(1);
    $log->enter();

    if ($operation =~ /create/) {
        $log->step('Create hardlinks');
        for (my $i = 0; $i < $count; $i++) {
            my $hl;
            if (defined $args{prefix}) {
                $hl = "$dir_path/" . $args{prefix} . $i;
            } else {
                $hl = "$dir_path/hl_" . int(rand(2**32));
            }
            $log->step('Creating hardlink of name: '  . $hl);
            push @hardlink_list, $hl;
            $Client_apiset->ln(
                files                => $file_path,
                'link-name'          => $hl,
                'symlink'            => 0,
                'connectrec-timeout' => 600,
                'privilege-level'    => 'root'
            );
        }
    } elsif ($operation =~ /add/) {
        if (!$args{hardlink_list}) {
            my $error
                = 'Hardlinks must be created before adding more hardlinks. Use operation = create instead of add when creating hardlinks for the first time.';
            NATE::BaseException->throw($error);
        }
        $log->step('Add more hardlinks');
        @hardlink_list = @{$args{hardlink_list}};
        for (my $i = 0; $i < $count; $i++) {
            my $hl = "$dir_path/hl_" . int(rand(2**32));
            push @hardlink_list, $hl;
            $Client_apiset->ln(
                files                => $file_path,
                'link-name'          => $hl,
                'symlink'            => 0,
                'connectrec-timeout' => 600,
                'privilege-level'    => 'root'
            );
        }
    } elsif ($operation =~ /delete/) {
        if (!$args{hardlink_list}) {
            my $error = 'There are no hardlinks to be deleted';
            NATE::BaseException->throw($error);
        }
        $log->step('Delete hardlinks');
        @hardlink_list = @{$args{hardlink_list}};
        if ($count >= scalar(@hardlink_list)) {
            my $error
                = 'You are trying to delete '
                . $count
                . ' hardlinks when only '
                . scalar(@hardlink_list)
                . ' hardlinks are persent';
            NATE::BaseException->throw($error);
        }

        for (my $i = 0; $i < $count; $i++) {
            my $hl = shift @hardlink_list;
            $log->comment('Unlinking hardlink ' . Dumper($hl));
            $Client_apiset->rm(
                paths                => $hl,
                'connectrec-timeout' => 600,
                'privilege-level'    => 'root'
            );
        }

    } elsif ($operation =~ /modify/) {
        #it is sufficient to touch the 'parent file' because that will automatically change the
        #time for all the hardlinks pointing to that file.
        #hence nothing to return here
        $log->step('Modifying hardlinks by modifying the timestamp on the src file: ' . $file_path);
        $Client_apiset->touch (
                    path                 =>   $file_path,
                    'connectrec-timeout' => 600,
                    'privilege-level'    => 'root',);
    }

    $log->comment('Hardlinks for file '
            . $file_path
            . 'are : '
            . Dumper(\@hardlink_list));
    $log->exit();
    return \@hardlink_list;
} ## end sub manipulate_hardlinks



=head2 populate_create_tree
This method does read write/ read traffic on volume using create tree.

=over

=item example

    populate_create_tree(operation_type => $operation_type,
                          vol_obj => $vol_obj,
                          client_name => $client_nm,
                        );

=item options

=over

=item operation_type

(required) Operation type can be populate/modify/read

=item vol_obj

(required) ARRAYREF - the volume object to run create tree on.

=item client_name

(required) ARRAYREF - nfs client name.

=back

=cut

sub populate_create_tree{
   my %args = validate_with(
        params => \@_,
        spec => {
            operation_type   => { type => SCALAR,  regex => qr/populate|modify|read/, default => 'populate' }, # populate or modify or read
            vol_obj  => { type => OBJECT, optional => 1 },
            file_size => { type => SCALAR, default => '100MB' },
            file_cnt => {type => SCALAR, default => '10'},
            client_name => {type => SCALAR},
            mod_delete_pct => {type => SCALAR, default => '10%'},
            mod_create_pct => {type => SCALAR, default => '10%'},
            sleep_after_modify => {type => SCALAR, default => '1800'},
        },
    );

    my $log = log_global(1);
    $log->enter();
    my $operation_type     = $args{operation_type};
    my $vol_obj            = $args{vol_obj};
    my $file_size          = $args{file_size};
    my $file_cnt           = $args{file_cnt};
    my $client_name        = $args{client_name};
    my $mod_delete_pct     = $args{mod_delete_pct};
    my $mod_create_pct     = $args{mod_create_pct};
    my $sleep_after_modify = $args{sleep_after_modify};
    my $vol_src_name = $vol_obj->volume;
    my @subtests;
    my $pm;
    my @subtests_read1;
    my $node_obj = NACL::STask::Node->new(name => $vol_obj->get_local_node_name());
    my $node_src_name = $node_obj->name();

    if ($operation_type eq 'populate'){
        $log->step('Populate the source volume using Create_Tree.thpl');
        my $script = 'DPSAS/IO/run_create_tree.thpl';
        push @subtests,
            NATE::Process->new(
                codespec => $script,
                runid => "run_create_tree",
                params => {
                    'TST_CLUSTER_NODES'  => $node_src_name,
                    'OPERATION_TYPE'     => 'populate',
                    'TEST_VOLS'          => $vol_obj->volume,
                    'CLIENTS'            => $client_name,
                    'BYTES_PER_SEC'      => 'nolimit',
                    'DATA_PROTOCOLS'     => 'nfs',
                    'FILE_SIZE'          => $file_size,
                    'FILE_CNT'           => $file_cnt,
                },
                inherit_params => 1,
                onexit=> \&NATE::Process::on_exit_propagate_worst_result,
            );
    }elsif($operation_type eq 'modify'){
        $log->step('Modify the source volume using Create_Tree.thpl');
        my $script = 'DPSAS/IO/run_create_tree.thpl';
        push @subtests,
            NATE::Process->new(
                codespec => $script,
                runid => "run_create_tree",
                params => {
                    'TST_CLUSTER_NODES'  => $node_src_name,
                    'OPERATION_TYPE'     => 'modify',
                    'TEST_VOLS'          => $vol_obj->volume,
                    'CLIENTS'            => $client_name,
                    'BYTES_PER_SEC'      => 'nolimit',
                    'DATA_PROTOCOLS'     => 'nfs',
                    'FILE_SIZE'          => $file_size,
                    'FILE_CNT'           => $file_cnt,
                    'MOD_DELETE_PCT'     => $mod_delete_pct,
                    'MOD_CREATE_PCT'     => $mod_create_pct,
                    'READ_PCT'           => 0,
                },
                inherit_params => 1,
                onexit=> \&NATE::Process::on_exit_propagate_worst_result,
            );
    }elsif($operation_type eq 'read'){
          $log->step('Read volume using Create_Tree.thpl');
          my $script = 'DPSAS/IO/run_create_tree.thpl';
          push @subtests,
            NATE::Process->new(
                codespec => $script,
                runid => "run_create_tree_read_from_dst",
                params => {
                    'TST_CLUSTER_NODES'  => $node_src_name,
                    'OPERATION_TYPE'     => 'read',
                    'TEST_VOLS'          => $vol_obj->volume,
                    'CLIENTS'            => $client_name,
                    'BYTES_PER_SEC'      => 'nolimit',
                    'DATA_PROTOCOLS'     => 'nfs',
                    'READ_PCT'           => '100%',
                    'NUM_RUNS'           => '1',
                },
                inherit_params => 1,
                onexit=> \&NATE::Process::on_exit_propagate_worst_result,
               );
            sleep(300);
            my $show_footprint_obj_after_initialize = $vol_obj->show_footprint();
            $log->comment('value of capacity tier after initialize ' . Dumper($show_footprint_obj_after_initialize));
    }
    $pm = NATE::ProcManager->new(proc_info=> \@subtests);
    $pm->start();
    $pm->wait();
    $pm->destroy();
    if($operation_type eq 'modify'){
        my $src_nodescope_apiset = $node_obj->get_7m_or_nodescope_apiset();

        $log->step("Run wafl scan redirect on volume $vol_src_name  - this will remove the stale pvbn entries");
        $src_nodescope_apiset->execute_command(command => "wafl scan redirect $vol_src_name");

        $log->step('sync');
        $src_nodescope_apiset->execute_command(command => 'sync');

        $log->step('Flush the wafl cache');
        $src_nodescope_apiset->execute_command(command => 'wafl flush cache', 'connectrec-timeout' => 3600,);

        sleep($sleep_after_modify);
        my $show_footprint_obj_after_modify = $vol_obj->show_footprint();
        $log->comment('value of capacity tier after modify ' . Dumper($show_footprint_obj_after_modify));
    }

    my $Aggr_name = $vol_obj->get_one_state_attribute('aggregate');
    my $aggr = NACL::STask::Aggregate->find(
                command_interface => $node_obj,
                filter => { aggregate => $Aggr_name },
            );
    my($comp_t1,$perf_t1) = get_composite_aggr_stats(aggr=>$aggr);
    $log->comment('value of comp_t1 ' . $comp_t1);
    $log->comment('value of perf_t1 ' . $perf_t1);

}

=head2 populate_hecr_dataset

This method populates all hecr data with default values.
Incase values other than default is needed please call individual methods.

=over

=item example

    populate_hecr_dataset( sm_obj => $sm_obj,
                           client => $nfsclient);

=item options

=over

=item sm_obj

(required) Snapmirror Object - the snapmirror object.

=item client

(required) ARRAYREF - the snapmirror client to be used to run traffic on files.

=item volume_objs

(required) ARRAYREF - volume to hold datasets.

=item ss_to_create

(optional) ARRAYREF - snapshots to populate.Note s1 or s2 is mandatory as one of the params since s3 and
s4 depends on that.

=item data_file_size

(optional) ARRAYREF - the list of file size( which also act as file names ) to create


=back

=cut

sub populate_hecr_dataset {
    my %args = validate_with(
        params => \@_,
        spec => {
            sm_obj       => { type => ARRAYREF },
            client       => { type => ARRAYREF|SCALAR }, # nfs or san client depending on file_type
            ss_to_create => { type => ARRAYREF, optional => 1 },
            data_file_size => { type => ARRAYREF, default => ['100KB', '100MB', '200MB'] },
            extra_params => {type => HASHREF | UNDEF, optional =>1 }
        },
    );

    my $log = log_global(1);
    $log->enter();

    my %extra_params;
    if(!$args{extra_params}){
       %extra_params = ('filecount' => '1',);
    } else{
        %extra_params = %{$args{extra_params}};
    }
    $log->comment('Extra params passed to populate hecr : ' . Dumper(\%extra_params));
    my @snapmirrors = @{$args{sm_obj}};
    my $sm_obj = @{$args{sm_obj} }[0]; # use first snapmirror
    my $ss_to_create = $args{ss_to_create};

    my (@vols_src, @vols_dst);
    $log->step('Get node and src vol information from the sm objects');
    foreach my $snapmirror (@snapmirrors) {
        my $source_path      = $snapmirror->{source_path};
        my $destination_path = $snapmirror->{destination_path};
        my ($vserver_src, $vol_src_name) = ($source_path =~ /^(.*):(.*)$/);
        my ($vserver_dst, $vol_dst_name)
            = ($destination_path =~ /^(.*):(.*)$/);

        # Get vol objs
        my $vol_src = NACL::STask::Volume->new(
            command_interface => $snapmirror->{source_ci},
            volume            => $vol_src_name,
            vserver           => $vserver_src
        );
        push(@vols_src, $vol_src);

        my $vol_dst = NACL::STask::Volume->new(
            command_interface => $snapmirror->{command_interface},
            volume            => $vol_dst_name,
            vserver           => $vserver_dst,
        );
        push(@vols_dst, $vol_dst);
    }
    my $ret;

    my @traffic_client;
    if (ref($args{client}) eq 'ARRAY') {
        @traffic_client = map( NACL::C::Client->new( name => $_ ), @{$args{client}} );
    } else {
        @traffic_client = map( NACL::C::Client->new( name => $_ ), $args{client} );
    }
    my $client_name = $traffic_client[0]->client();
    if(!$args{ss_to_create}){
       $ss_to_create = [ 's1' , 's2'];
    }
    my $data_file_size = $args{data_file_size} || ['100KB', '100MB', '200MB'];

    if (grep {$_ eq 's1'}@{$ss_to_create}){
       $log->step('Create HECR dataset s1');
       $ret = Wf_create_multiplesize_filesnclones( src_vol_obj => \@vols_src,
                                                   nfsclient   => $client_name,
                                                   data_file_size => $data_file_size,
                                                   %extra_params, );
       $log->trace('s1 dataset return object' . Dumper($ret));
    }

    if (grep {$_ eq 's2'}@{$ss_to_create}){
        $log->step('Create HECR dataset s2');
        Wf_create_fileclones_filestodirectories(src_vol_obj => \@vols_src,
                                                nfsclient    => $client_name,
                                                %extra_params, );
    }

    if ((grep { $_ =~ 's3' } @{$ss_to_create})
            && (defined $ret)) {
            $log->step('Create HECR dataset s3');
            my $ret_ss3;
            foreach my $v (@vols_src) {
                my $vname         = $v->volume();
                my $file_objs_ref = $ret->{$vname}->{'file_obj'};
                my $file_cnt      = floor(scalar(@{$file_objs_ref}) / 2);
                $ret_ss3 = Wf_data_transitions(
                    file_obj        => $file_objs_ref,
                    client          => $client_name,
                    snapshot_prefix => 'ss3',
                    sleep_sec       => 180,
                    file_count      => $file_cnt,
                    %extra_params,
                );
            }
        }

    if (grep {$_ eq 's7'}@{$ss_to_create}) {
        $log->step('Create HECR dataset s7');
        Wf_EID_Holes('src_vol_obj' => \@vols_src, nfsclient => $client_name,);

    }

    if (grep { $_ =~ 'dirty_clone_ss' } @{$ss_to_create}) {
            $log->step('Create dirty clone snapshot');
            Wf_create_dirty_ss(
                'src_vol_obj' => \@vols_src,
                'nfsclient'   => $client_name,
                'clone_count'   => 3,
            );
    }
    foreach my $vol (@vols_src){
        my $vname = $vol->volume();
        $ret->{$vname}->{snapshots} = $ss_to_create;
    }
    return $ret;
}


=head2 Wf_create_multiplesize_filesnclones

This method creates files of various sizes - tiny files, small files, medium and large size files and their clones and take a snapshot 's1' defined in HECR dataset.
It returns a hashref of arrayrefs of filenames, file_obj and clone_obj.

=over

=item example

    my $hashref = Wf_create_multiplesize_filesnclones(src_vol_obj => $volume);
    #hash will contain references to 2 lists - one for clone and one for files
    foreach my $key (keys(%hash)) {
        foreach my $file (@{$hash{$key}}) {
                print Dumper $file;
        }
    }
    my $filenames=$hashref->{file_names};
    my $files=$hashref->{file_obj};
    my $clones=$hashref->{clone_obj};

=item options

=over

=item src_vol_obj

(required) Volume Object - the volume object on which multiple files of various sizes needs to be created.

=item snapshot_name

(optional) SCALAR  - the snapshot name to be created after dataset creation.Default value is s1.

=item snapmirror_label

(optional) SCALAR  - the snapmirror label for snapshot.Default value is weekly.

=item nfsclient

(required) SCALAR - the snapmirror client to be used to run traffic on files.

=item data_file_size

 (optional) ARRAYREF - the list of file size( which also act as file names ) to create

=back

=cut

sub Wf_create_multiplesize_filesnclones{
   my %args = validate_with(
        params => \@_,
        allow_extra => 1,
        spec   => {
            src_vol_obj  => { type => ARRAYREF, optional => 0 },
            snapshot_name => { type => SCALAR, optional => 1, default =>'s1' },
            snapmirror_label => { type => SCALAR, optional=> 1 ,default => 'weekly' },
            nfsclient        => { type => SCALAR, optional=>0 },
            data_file_size   => { type => ARRAYREF, optional => 1, default => ['50KB', '200KB', '1GB', '1MB', '500KB', '500MB', '2MB']},
            filecount        => { type => SCALAR, optional=>1, default => 1 },
            timeout          => { type => SCALAR, optional => 1, default => 900 },
            traffic_timeout  => { type => SCALAR, optional => 1, default => 120 },
        },
    );

    my $src_vol_obj        = $args{src_vol_obj};
    my $ss_name            = $args{snapshot_name};
    my $snapmirror_lbl     = $args{snapmirror_label};
    my $nfsclient          = $args{nfsclient};
    my $log                = log_global();
    my @subtests;

    # set file sizes
    my @Data_File_Sizes    = @{ $args{data_file_size} };

    foreach my $v (@{ $src_vol_obj }) {
        my %args = (
                volume          => $v,
                client_name     => $nfsclient,
                file_sizes      => \@Data_File_Sizes,
                timeout         => $args{timeout},
                traffic_timeout => $args{traffic_timeout},
                filecount       => $args{filecount},);
        push (@subtests, create_parallel_processes('_create_filesnclones', \%args));
    }

    run_processes_bg('file_and_clones', \@subtests);


    my %ret;
    foreach my $v (@{ $src_vol_obj }) {
       my $cmd_int = NACL::C::Node->new(name => $v->get_local_node_name());
       $log->comment('Collect all the files created as part of datset d1 ' .
                     'with prefix d1_file in the volume: ' . $v->volume());
       my $fnames = _get_filenames_from_vol (
                        volume      =>  $v,
                        prefix      =>  '^/d1_file',
                        );

       $log->comment('Get file objects for : ' . Dumper($fnames));
       my @fileobjs;
       foreach my $f (@{ $fnames }) {
            my $p = "/vol/" . $v->volume . "$f";
            push @fileobjs, NACL::STask::VolumeFile->find (
                                command_interface       =>   $cmd_int,
                                filter                  => {
                                             path        => $p,
                                             vserver     => $v->vserver()});
       }
       $ret{$v->volume}{'file_names'} = $fnames;
       $ret{$v->volume}{'file_obj'} = \@fileobjs;

       $log->comment('Collect all the clone files created as part of dataset d1 ' .
                     'with prefix d1_clone in the volume ' .  $v->volume());
       my $clonenames = _get_filenames_from_vol (
                        volume      =>  $v,
                        prefix      =>  '^/d1_clone',
                        );
       my @cloneobjs;
       foreach my $f (@{ $clonenames }) {
            my $p = "/vol/" . $v->volume . "$f";
            push @cloneobjs, NACL::STask::VolumeFile->find (
                                command_interface       =>   $cmd_int,
                                filter                  => {
                                        path        => $p,
                                        vserver     => $v->vserver()});
       }
       $ret{$v->volume}{'clone_obj'} = \@cloneobjs;

       $log->comment('Cut a ss ' . $ss_name . ' on vol: ' . $v->volume());
       NACL::STask::VolumeSnapshot->create(
           command_interface  => NACL::C::Node->new(name => $v->get_local_node_name()),
           volume             => $v,
           vserver            => $v->vserver(),
           snapshot           => $ss_name,
           'snapmirror-label' => $snapmirror_lbl,
      );
    }

    $log->trace('Return value : ' . Dumper(\%ret));
    $log->exit();
    return \%ret;
}

sub _get_filenames_from_vol {

   my %args = validate_with(
        params => \@_,
        spec   => {
            volume        => { type => OBJECT, optional => 0 },
            prefix        => { type => SCALAR, optional => 0 },});

    my $v = $args{volume};
    my $prefix = $args{prefix};

    my @op = NACL::C::Volume->explore(
                command_interface =>  NACL::C::Node->new(name => $v->get_local_node_name()),
                format            => 'inode',
                find              => 'type!=0',
                dump              => 'name',
                scope             => $v->volume);

    my @files;
    foreach my $rec (@op) {
        push @files, $rec->name;
    }

    my @ret;
    foreach my $f (@files) {
        if ($f =~ /$prefix/) {
                push @ret, $f;
        }
    }

    return \@ret;
}

sub _create_filesnclones {

   my %args = validate_with(
        params => \@_,
        spec   => {
            volume        => { type => OBJECT, optional => 0 },
            client_name   => { type => SCALAR, optional => 0 },
            file_sizes    => { type => ARRAYREF, optional => 0 },
            filecount     => { type => SCALAR, optional => 1, default => 1 },
            timeout       => { type => SCALAR, optional => 1, default => 900 },
            traffic_timeout => { type => SCALAR, optional => 1, default => 120 },
        },
    );

    my $v              = $args{volume};
    my $vname          = $v->volume();
    my $vs             = $v->vserver();
    my $cmd_interface  =  NACL::C::Node->new(name => $v->get_local_node_name());
    my $client_name    = $args{client_name};
    my $files          = $args{file_sizes};
    my $filecnt        = $args{filecount};
    my $timeout        = $args{timeout};
    my $traffic_duration = $args{traffic_timeout};
    my $log            = log_global();
    my @subtests;

    $log->comment('Mount the volume');
    my $mtpt = mount_nfs_vol (
            volume      => $v,
            client_name => $client_name,
            nfsvers     => '3');

    $log->comment('Divide the file size array by half');
    my $half = ceil (scalar(@{ $files })/2);
    my @files_half = splice (@{ $files }, 0, $half);
    my %size_hash;
    $size_hash{0} = \@files_half;
    $size_hash{1} = $files;

    #$log->comment('Spawning multiple proceeses to create files of different sizes');
    foreach my $k (keys %size_hash) {
        my $size = $size_hash{$k};
        $log->comment('Working on sizes: ' . Dumper($size));
        my %args = (
                mount_pt          =>  $mtpt,
                volume            =>  $v,
                size              =>  $size,
                filecount         =>  $filecnt,
                client_name       =>  $client_name,
                traffic_timeout   =>  $traffic_duration,
                'method-timeout'  =>  $timeout,);
        #XXX will re-enable once the issue is fixed
        #push (@subtests, create_parallel_processes('_create_actual_files', \%args));
        _create_actual_files(\%args);
    }
    #XXX will re-enable once the issue is fixed
    #run_processes_bg('create_files', \@subtests);

    $log->exit();

    return;
}

sub _create_actual_files {

   my %args = validate_with(
        params => \@_,
        spec   => {
            volume          => { type => OBJECT, optional => 0 },
            size            => { type => ARRAYREF, optional => 0 },
            filecount       => { type => SCALAR, optional => 0 },
            client_name     => { type => SCALAR, optional => 0 },
            mount_pt        => { type => SCALAR, optional => 0 },
            'method-timeout'=> { type => SCALAR, optional => 1, default => 900 },
            traffic_timeout => { type => SCALAR, optional => 1, default => 120 },
            io_tool         => { type => SCALAR, default => 'comp' },
        },
    );

    my $mountPoint  = $args{mount_pt};
    my $mtimeout    = $args{'method-timeout'};
    my $ttimeout    = $args{traffic_timeout};
    my $size        = $args{size};
    my $filecnt     = $args{filecount};
    my $vol         = $args{volume};
    my $vname       = $vol->volume();
    my $log         = log_global();
    my $Client      = NACL::C::Client->new( name => $args{client_name} );

    my @file_objs;
    foreach my $s (@{ $size }) {
        my $mkfilepath = "/vol/$vname/d1_file_$s";
        $log->comment('Creating ' . $filecnt . ' files with path ' . $mkfilepath);
        $log->comment('Creating for size: ' . $s);
        my @f_objs = NACL::STask::VolumeFile->create(
                            command_interface => NACL::C::Node->new(name => $vol->get_local_node_name()),
                            vserver           => $vol->vserver(),
                            size              => $s,
                            path              => $mkfilepath,
                            filecount         => $filecnt,
                            'method-timeout'  => $mtimeout,
                           );

        my @paths;
        foreach my $file_obj (@f_objs) {
            my $fstate = $file_obj->state();
            if ($fstate->file_size() !~ /^0(K|M|G)B$/i){ #we could have used $size here but thats fine
                my $name = $fstate->name;
                push (@paths, "$mountPoint/$name");

                if($args{'io_tool'} !~ /sio/i){
                    $log->comment("Writting compression dataset to file $name");
                    compression_dataset (
                        operation   => 'populate',
                        client_name => $args{client_name},
                        volume_objs => [$vol],
                        file_paths  => ["$name"],
                        file_size   => $fstate->file_size(),
                    );
                }
            }
        }

        if($args{'io_tool'} =~ /sio/i){
            $log->comment('Starting IO to all paths: ' . Dumper(\@paths));
            my @t_obj   = start_nfs_traffic ( client => $Client,
                                              paths  => \@paths,
                                              file_size  => $s,
                                              block_size => '4k' );

            $log->comment('Sleeping for ' . $ttimeout . ' to fill up the files');
            sleep($ttimeout);

            $log->comment('Stopping traffic to all paths: ' . Dumper(\@paths));
            foreach my $t (@t_obj) {
                $t->stop();
            }
        }

        $log->comment('Sleeping for 15 sec for CP');
        sleep(15);

        push (@file_objs, @f_objs);
    }

    $log->comment('Creating clone files');
    foreach my $file_obj (@file_objs) {
        my $fstate = $file_obj->state();
        my $fname = $fstate->name;
        my $clone = "/vol/$vname/d1_clone_$fname";
        $log->comment('.... of name : ' . $clone);
        NACL::STask::VolumeFileClone->create(
            command_interface   =>  NACL::C::Node->new(name => $vol->get_local_node_name()),
            vserver             => $vol->vserver(),
            'source-path'       => $file_obj->path(),
            'destination-path'  => $clone);
    }

    $log->exit();

    return;
}


=head2 Wf_create_fileclones_filestodirectories

Create multiple source files, create their clones, delete some files and create dir with same name, take snapshot s2.

=over

=item example

    Wf_create_fileclones_filestodirectories(src_vol_obj => $volume);

=item options

=over

=item src_vol_obj

(required) Volume Object - the volume object on which multiple files of various sizes needs to be created.

=item create_file_count

(optional) SCALAR - the number of files to be created.The default value is 5.

=item delete_file_count

(optional) SCALAR - the number of files to be deleted.The default value is 2.Note this vaulue should be lesser than create_file_count.

=item snapshot_name

(optional) SCALAR  - the snapshot name to be created after dataset creation.Default value is s2.

=item snapmirror_label

(optional) SCALAR  - the snapmirror label for snapshot.Default value is weekly.

=item nfsclient

(required) SCALAR - the snapmirror client to be used to run traffic on files.

=back

=cut

sub Wf_create_fileclones_filestodirectories{
     my %args = validate_with(
        params => \@_,
        allow_extra => 1,
        spec   => {
            src_vol_obj  => { type => ARRAYREF, optional => 0 },
            create_file_count   => { type => SCALAR, optional => 1, default => 2 },
            delete_file_count   => { type => SCALAR, optional => 1 , default => 1 },
            snapshot_name => { type => SCALAR, optional => 1, default =>'s2' },
            snapmirror_label => { type => SCALAR, optional=> 1 ,default => 'weekly' },
            nfsclient        => { type => SCALAR, optional=>0 },
            io_tool          => { type => SCALAR, default => 'comp' },
        },
    );

    my $src_vol_obj        = $args{src_vol_obj};
    my $create_file_count  = $args{create_file_count};
    my $delete_file_count  = $args{delete_file_count};
    my $ss_name            = $args{snapshot_name};
    my $snapmirror_lbl     = $args{snapmirror_label};
    my $nfsclient          = $args{nfsclient};
    my $log                = log_global();
    my $Client = NACL::C::Client->new( name => $nfsclient );
    my $client_name = $Client->client();

    if($delete_file_count > $create_file_count){
        NACL::Exceptions::InvalidParam->throw("Please provide files to be deleted $delete_file_count lesser than files to be created $create_file_count");
    }

    $log->comment('Creating no of files and their clones depending on file_count'. $create_file_count);

    my (@file_objs,@clone_objs);

    foreach my $src_vol(@{$src_vol_obj}){
        my $vol_name  =  $src_vol->volume;
        my $src_vol_vserver    = $src_vol->vserver;
        my $src_vol_ci         =  NACL::C::Node->new(name => $src_vol->get_local_node_name());
        my $mountPoint = mount_nfs_vol (
            volume      => $src_vol,
            client_name => $client_name,
            nfsvers     => '3',
        );

        my $mkfile_path = "/vol/$vol_name/file";
        my $file_obj = NACL::STask::VolumeFile->create(
                command_interface => $src_vol_ci,
                vserver           => $src_vol_vserver,
                size              => '500KB',
                path              => $mkfile_path,
                filecount         => $create_file_count,
               'method-timeout'   => 600,
        );
        push @file_objs, $file_obj;

        my @paths;
        for (my $i=0; $i<$create_file_count; $i++){
            my $file_name = "file" . ".$i";
            push @paths, "$mountPoint/$file_name";

            if($args{'io_tool'} !~ /sio/i){
                $log->comment("Writting compression dataset to file $file_name");
                compression_dataset (
                    operation   => 'populate',
                    client_name => $client_name,
                    volume_objs => [$src_vol],
                    file_paths  => ["$file_name"],
                    file_size   => '500KB',
                );
            }
        }

        if($args{'io_tool'} =~ /sio/i){
            my @traffic_obj = start_nfs_traffic(client => $Client,
                                                paths  => \@paths,
                                                file_size  => '500KB',
                                                block_size => '32k');
            #sleep for 2 minutes to fill up files
            sleep(120);
            $traffic_obj[0]->stop();
        }

        #sleep for a while for CP to happen
        sleep(15);

        for (my $i=0; $i<$create_file_count; $i++){
            my $Clone = $mkfile_path . ".$i" . '_clone';
            my $vol_fileclone_obj = NACL::STask::VolumeFileClone->create(
               command_interface   => $src_vol_ci,
               vserver             => $src_vol_vserver,
               "source-path"       => $mkfile_path . ".$i",
               "destination-path"  => $Clone,
            );
            push @clone_objs, $vol_fileclone_obj;
        }

       $log->comment("Deleting $delete_file_count files and creating directories with same name as files");
       for (my $i=0; $i<$delete_file_count; $i++){
           my $file_obj = pop @file_objs; # XXX - do we care which end we pop from ?
           $log->trace('File obj is '.Dumper($file_obj));
           my $file_path = $file_obj->path();
           $log->log('Deleting file '.$file_path);
           $file_obj->delete(); # TBD purge instead ?

            NACL::C::VolumeDirectory->create(
                command_interface => $src_vol_ci,
                path              => $file_path,
                #path              => $mkfile_path . ".$i",
                vserver           => $src_vol_vserver  # Required for CMode
            );
        }

        NACL::STask::VolumeSnapshot->create(
                command_interface  => $src_vol_ci,
                volume             => $src_vol,
                vserver            => $src_vol_vserver,
                snapshot           => "$ss_name",
               'snapmirror-label' => $snapmirror_lbl,
        );
    }

    my %ret;
    $ret{'file_obj'} = \@file_objs;
    $ret{'clone_obj'} = \@clone_objs;
    return \%ret;

}

=head2 Wf_create_dirty_ss

Create multiple source files, run io on files, set fiji to capture unsplit clones, create their clones, create clones out of clones, take snapshot dirty clone snapshot.

=over

=item example

    Wf_create_dirty_ss(src_vol_obj => $volume);

=item options

=over

=item src_vol_obj

(required) Volume Object - ARRAYREF -the volume object on which multiple files of various sizes needs to be created.





=item clone_count

(optional) SCALAR - the number of clone files to be created.The default value is 1.

=item snapshot_name

(optional) SCALAR  - the snapshot name to be created after dataset creation.Default value is dirty_clone_ss.

=item snapmirror_label

(optional) SCALAR  - the snapmirror label for snapshot.Default value is weekly.

=item nfsclient

(required) SCALAR - the snapmirror client to be used to run traffic on files.

=back

=cut

sub Wf_create_dirty_ss{
     my %args = validate_with(
        params => \@_,
        spec   => {
            src_vol_obj  => { type => ARRAYREF, optional => 0 },
            file_size => {type => ARRAYREF, optional => 0, default => ['50KB', '200KB']} ,
            clone_count => {type => SCALAR, optional => 1, default => 1},
            snapshot_name => { type => SCALAR, optional => 1, default =>'dirty_clone_ss'},
            snapmirror_label => { type => SCALAR, optional=> 1 ,default => 'weekly'},
            nfsclient        => {type => SCALAR, optional=>0},
        },
    );

    my $src_vol_obj        = $args{src_vol_obj};
    my $file_size          = $args{file_size};
    my $clone_count        = $args{clone_count};
    my $ss_name            = $args{snapshot_name};
    my $snapmirror_lbl     = $args{snapmirror_label};
    my $nfsclient          = $args{nfsclient};
    my $log                = log_global();
    my $Client = NACL::C::Client->new( name => $nfsclient );
    my @file_objs;
    my @traffic_obj;

    my @subtests;
    $counter=0;
    foreach my $src_vol(@{$src_vol_obj}) {
        my %args_func = (src_vol_obj  => $src_vol,
                          file_size => $file_size,
                        clone_count => $clone_count,
                      snapshot_name => $ss_name,
                   snapmirror_label => $snapmirror_lbl,
                   nfsclient        => $nfsclient,);
        push (@subtests, create_parallel_processes('Wf_create_dirtyclones_ss', \%args_func));

    }

    run_processes_bg('Wf_create_dirtyclones_ss', \@subtests);
}



=head2 Wf_create_dirtyclones_ss

Create multiple source files, run io on files, set fiji to capture unsplit clones, create their clones, create clones out of clones, take snapshot dirty clone snapshot.

=over

=item example

    Wf_create_dirtyclones_ss(src_vol_obj => $volume);

=item options

=over

=item src_vol_obj

(required) Volume Object - the volume object on which multiple files of various sizes needs to be created.

=item create_file_count

(optional) SCALAR - the number of files to be created.The default value is 5.

=item clone_count

(optional) SCALAR - the number of clone files to be created.The default value is 1.

=item snapshot_name

(optional) SCALAR  - the snapshot name to be created after dataset creation.Default value is dirty_clone_ss.

=item snapmirror_label

(optional) SCALAR  - the snapmirror label for snapshot.Default value is weekly.

=item nfsclient

(required) SCALAR - the snapmirror client to be used to run traffic on files.

=back

=cut

sub Wf_create_dirtyclones_ss{
     my %args = validate_with(
        params => \@_,
        spec   => {
            src_vol_obj  => { type => OBJECT, optional => 0 },
            file_size => { type => ARRAYREF, optional => 0},
            clone_count => { type => SCALAR, optional => 1, default => 1},
            clone_of_clone_count => { type => SCALAR, optional => 1, default => 6},
            snapshot_name => { type => SCALAR, optional => 1, default =>'dirty_clone_ss'},
            snapmirror_label => { type => SCALAR, optional=> 1 ,default => 'weekly'},
            nfsclient        => { type => SCALAR, optional=>0},
        },
    );

    my $src_vol_obj        = $args{src_vol_obj};
    my $file_size          = $args{file_size};
    my $clone_count        = $args{clone_count};
    my $ss_name            = $args{snapshot_name};
    my $snapmirror_lbl     = $args{snapmirror_label};
    my $nfsclient          = $args{nfsclient};
    my $c_o_c_cnt          = $args{clone_of_clone_count};
    my $log                = log_global();
    my $Client = NACL::C::Client->new( name => $nfsclient );
    my $client_name = $Client->client();
    my @file_objs;
    my @traffic_obj;
    my @io_objs;

    my $create_file_count;
    $create_file_count = scalar(@{$file_size});
    $log->comment('Creating no of files and thier clones depending on file_count'. $create_file_count);

    my $vol_name  =  $src_vol_obj->volume;
    my $src_vol_vserver    = $src_vol_obj->vserver;
    my $src_vol_fsid      = $src_vol_obj->get_one_state_attribute('fsid');
    my $src_vol_ci         =  NACL::C::Node->new(name => $src_vol_obj->get_local_node_name());
    my $mountPoint = mount_nfs_vol (
            volume      => $src_vol_obj,
            client_name => $client_name,
            nfsvers     => '3',
    );

    my $mkfile_path = "/vol/$vol_name/filee";

    fiji_unsplit_clones(node=>$src_vol_ci,fsid=>$src_vol_fsid);

    my $extent_size = $src_vol_obj->get_one_state_attribute('default-extent-size');
    foreach my $datasizename(@{$file_size}){
        @file_objs = NACL::C::VolumeFile->create(
            command_interface => $src_vol_ci,
            vserver           => $src_vol_vserver,
            size              => $datasizename,
            path              => $mkfile_path,
            filecount         => $create_file_count,
            'method-timeout'  => 600,
        );

        my @paths;
        for (my $i=0; $i<$create_file_count; $i++){
            my $file_name = "filee" . ".$i";
            push @paths, "$mountPoint/$file_name";
        }

        @traffic_obj = start_nfs_traffic(client => $Client,
                                          paths => \@paths,
                                      file_size => $datasizename,
                                     block_size => $extent_size);
        push @io_objs, @traffic_obj;
    }

    for (my $j =0; $j<$create_file_count; $j++){
        for(my $i=0; $i<$clone_count; $i++){
                my $Clone = $mkfile_path . "$i" . "$j". '_clone';
                my $source_path = $mkfile_path . ".$j";
                my $destination_path = $Clone;
                my $vol_fileclone_obj = _create_clone(command_interface => $src_vol_ci,
                                                       vserver=>$src_vol_vserver,
                                                       source_path=>$source_path,
                                                       destination_path => $destination_path
                                                       );
                my $src_path = $destination_path;
                foreach my $k (1..$c_o_c_cnt){
                    my $dest_path = $src_path . "COC$k";

                    my $obj   =  _create_clone(command_interface => $src_vol_ci,
                                               vserver=>$src_vol_vserver,
                                               source_path=>$src_path,
                                               destination_path => $dest_path
                                  );
                    $src_path  = $dest_path;
                }
        }
    }

    $log->step('Verify that we hit the fiji on src node');
    fiji_wait(node=>$src_vol_ci);

    NACL::STask::VolumeSnapshot->create(
               command_interface  => $src_vol_ci,
               volume             => $src_vol_obj,
               vserver            => $src_vol_vserver,
               snapshot           => "$ss_name",
              'snapmirror-label' => $snapmirror_lbl,
    );

    $log->comment('Stop the Traffic');
    foreach my $io_obj(@io_objs){
        $io_obj->stop();
    }

    $log->step('Clear all fijis');
    fiji_clear_all(node=>$src_vol_ci);
}

sub _create_clone{
    my %args = validate_with(
        params => \@_,
        spec   => {
            command_interface  => { type => OBJECT, optional => 0 },
            vserver            => { type => SCALAR, optional => 0},
            source_path        => {type  => SCALAR, optional => 0},
            destination_path   => {type  => SCALAR, optional => 0},
        },
    );

    my $src_vol_ci       = $args{command_interface};
    my $src_vol_vserver  = $args{vserver};
    my $source_path      = $args{source_path};
    my $destination_path = $args{destination_path};

    my $vol_fileclone_obj = NACL::STask::VolumeFileClone->create(
                   command_interface   => $src_vol_ci,
                   vserver             => $src_vol_vserver,
                   "source-path"       => $source_path,
                   "destination-path"  => $destination_path,
               );

   return $vol_fileclone_obj;
}
=head2 Wf_Cad_test

Delete and recreate all the clones and Test the CRM file entries.

=over

=item example

    Wf_Cad_test(src_vol_obj => $volume);

=item options

=over

=item src_vol_obj

(required) Volume Object - ARRAYREF -the volume object on which multiple files of various sizes needs to be created.

=item create_file_count

(optional) SCALAR - the number of files to be created.The default value is 5.

=item clone_count

(optional) SCALAR - the number of clone files to be created.The default value is 1.

=item snapshot_name

(optional) SCALAR  - the snapshot name to be created after dataset creation.Default value is dirty_clone_ss.

=item snapmirror_label

(optional) SCALAR  - the snapmirror label for snapshot.Default value is weekly.

=item nfsclient

(required) SCALAR - the snapmirror client to be used to run traffic on files.

=back

=cut

sub Wf_cad_test{
     my %args = validate_with(
        params => \@_,
        spec   => {
            src_vol_obj  => { type => ARRAYREF, optional => 0 },
            file_size => {type => ARRAYREF, optional => 1,default => ['50KB', '200KB']},
            clone_count => {type => SCALAR, optional => 1, default => 1},
            snapshot_name => { type => SCALAR, optional => 1, default =>'cad_test_ss'},
            snapmirror_label => { type => SCALAR, optional=> 1 ,default => 'weekly'},
            nfsclient        => {type => SCALAR, optional=>0},
        },
    );

    my $src_vol_obj        = $args{src_vol_obj};
    my $file_size          = $args{file_size};
    my $clone_count        = $args{clone_count};
    my $ss_name            = $args{snapshot_name};
    my $snapmirror_lbl     = $args{snapmirror_label};
    my $nfsclient          = $args{nfsclient};
    my $log                = log_global();
    my $Client = NACL::C::Client->new( name => $nfsclient );
    my @file_objs;
    my @traffic_obj;

    my @subtests;
    $counter=0;
    foreach my $src_vol(@{$src_vol_obj}) {
        my %args_func = (src_vol_obj  => $src_vol,
                  file_size => $file_size,
                        clone_count => $clone_count,
                      snapshot_name => $ss_name,
                   snapmirror_label => $snapmirror_lbl,
                   nfsclient        => $nfsclient,);
        push (@subtests, create_parallel_processes('cad_test', \%args_func));
  }

    run_processes_bg('cad_test', \@subtests);
}



=head2 cad_test

Create multiple source files, create their clones,take a snapshot cad_test_ss1 delete and recreate clones with same name, take snapshot cad_test_ss2.
After transfer of second snapshot cad_test_ss2 the entry will be deleted from CRM file.

=over

=item example

    cad_test(src_vol_obj => $volume);

=item options

=over

=item src_vol_obj

(required) Volume Object - the volume object on which multiple files of various sizes needs to be created.

=item create_file_count

(optional) SCALAR - the number of files to be created.The default value is 5.

=item clone_count

(optional) SCALAR - the number of clone files to be created.The default value is 1.

=item snapshot_name

(optional) SCALAR  - the snapshot name to be created after dataset creation.Default value is cad_ss.It will be appended with 1 and 2.

=item snapmirror_label

(optional) SCALAR  - the snapmirror label for snapshot.Default value is weekly.

=item nfsclient

(required) SCALAR - the snapmirror client to be used to run traffic on files.

=back

=cut

sub cad_test{
     my %args = validate_with(
        params => \@_,
        spec   => {
            src_vol_obj  => { type => OBJECT, optional => 0 },
            file_size => {type => ARRAYREF, optional => 0},
            clone_count => {type => SCALAR, optional => 1, default => 1},
            snapshot_name => { type => SCALAR, optional => 1, default =>'cad_test_ss'},
            snapmirror_label => { type => SCALAR, optional=> 1 ,default => 'weekly'},
            nfsclient        => {type => SCALAR, optional=>0},
        },
    );

    my $src_vol_obj        = $args{src_vol_obj};
    my $file_size          = $args{file_size};
    my $clone_count        = $args{clone_count};
    my $ss_name            = $args{snapshot_name};
    my $snapmirror_lbl     = $args{snapmirror_label};
    my $nfsclient          = $args{nfsclient};
    my $Client = NACL::C::Client->new( name => $nfsclient );
    my $client_name = $Client->client();
    my @file_objs;
    my %clone_hash;
    my @clone_nms;
    my @inums;
    my $log = log_global(1);
    $log->enter();

    $log->comment('Creating no of files and their clones' );

    my $vol_name  =  $src_vol_obj->volume;
    my $src_vol_vserver    = $src_vol_obj->vserver;
    my $src_vol_ci         =  NACL::C::Node->new(name => $src_vol_obj->get_local_node_name());
    my $mountPoint = mount_nfs_vol (
            volume      => $src_vol_obj,
            client_name => $client_name,
            nfsvers     => '3',
    );

    my $mkfile_path = "/vol/$vol_name/fillees";

    foreach my $datasizename(@{$file_size}){
        my $mkfile_path = "/vol/$vol_name/$datasizename" . $$;
        my $file_obj = NACL::C::VolumeFile->create(
                          command_interface => $src_vol_ci,
                          vserver           => $src_vol_vserver,
                          size              => $datasizename,
                          path              => $mkfile_path,
                         'method-timeout'   => 600,
                         );
        push @file_objs,$file_obj;

      }

    foreach my $file_obj(@file_objs){
        my $path = $file_obj->path();
        for(my $i=0; $i<$clone_count; $i++){
            my $Clone =  $path . '_clone';
            my $vol_fileclone_obj = NACL::STask::VolumeFileClone->create(
                   command_interface   => $src_vol_ci,
                   vserver             => $src_vol_vserver,
                   "source-path"       => $path,
                   "destination-path"  => $Clone,
               );
               push @clone_nms, $Clone;
             }
    }

    my @cad_ss_arr;

    my $ss_1 =  NACL::STask::VolumeSnapshot->create(
                   command_interface  => $src_vol_ci,
                   volume             => $src_vol_obj,
                   vserver            => $src_vol_vserver,
                   snapshot           => "$ss_name" . "_1",
                   'snapmirror-label' => $snapmirror_lbl,
                );

    my $ss_1_name = $ss_1->snapshot();

    foreach my $c(@clone_nms){
             # before deletion get the inum of the clones so as to look
             # up for missing inum in the cad_test output
             my $inum =  _get_inum(clone => $c,src_node_obj => $src_vol_ci,vol_src_obj =>$src_vol_obj);
             $log->comment("The inode number of clone getting deleted is $inum");
             push @inums, $inum;
             #my $src_Nodescope_Apiset = $src_vol_ci->get_7m_or_nodescope_apiset;
             #$src_vol_ci->privilege_level('test');
             #my $cmd = "rm $c";
             #$src_Nodescope_Apiset->execute_command(
             #    command => $cmd );
             NACL::STask::VolumeFile->purge(command_interface => $src_vol_ci,
                                            vserver           => $src_vol_vserver, # Required for CMode
                                            path              => $c,
                                           );

    }

    my $ss_2 =   NACL::STask::VolumeSnapshot->create(
                  command_interface  => $src_vol_ci,
                  volume             => $src_vol_obj,
                  vserver            => $src_vol_vserver,
                  snapshot           => "$ss_name" . "_2",
                  'snapmirror-label' => $snapmirror_lbl,
                );

    my $ss_2_name = $ss_2->snapshot();

   push(@cad_ss_arr, ("$ss_1_name", "$ss_2_name"));

   ##cad_test shouldnt show that entry
    $log->comment(
"The inums deleted are @inums\n"
    );

   return (\@inums,\@cad_ss_arr);

}

=head2 Wf_EID_Holes

Create multiple source files only with holes, create their clones and increase the size of clones using setattr,take snapshot s7.
Verify that EID_Hole will be sent. This can be validated only update update is complete and counters get incremented so covering
this in workflow.pm
=over

=item example

    Wf_EID_Holes(src_vol_obj => $volume);

=item options

=over

=item src_vol_obj

(required) Volume Object - ARRAYREF -the volume object on which multiple files of various sizes needs to be created.

=item create_file_count

(optional) SCALAR - the number of files to be created.The default value is 5.

=item snapshot_name

(optional) SCALAR  - the snapshot name to be created after dataset creation.Default value is s7.

=item snapmirror_label

(optional) SCALAR  - the snapmirror label for snapshot.Default value is weekly.

=item nfsclient

(required) SCALAR - the snapmirror client to be used to run traffic on files.

=back

=cut

sub Wf_EID_Holes{
     my %args = validate_with(
        params => \@_,
        spec   => {
            src_vol_obj  => { type => ARRAYREF, optional => 0 },
            create_file_count   => { type => SCALAR, optional => 1, default => 5},
            snapshot_name => { type => SCALAR, optional => 1, default =>'s7'},
            snapmirror_label => { type => SCALAR, optional=> 1 ,default => 'weekly'},
            nfsclient        => {type => SCALAR, optional=>0},
        },
    );

    my $src_vol_obj        = $args{src_vol_obj};
    my $create_file_count  = $args{create_file_count};
    my $file_size          = $args{file_size};
    my $clone_count        = $args{clone_count};
    my $ss_name            = $args{snapshot_name};
    my $snapmirror_lbl     = $args{snapmirror_label};
    my $nfsclient          = $args{nfsclient};
    my $log                = log_global();
    my $Client = NACL::C::Client->new( name => $nfsclient );
    my @file_objs;
    my @traffic_obj;

    my @subtests;
    $counter=0;
    foreach my $src_vol(@{$src_vol_obj}) {
        my %args_func = (src_vol_obj  => $src_vol,
                create_file_count   => $create_file_count,
                      snapshot_name => $ss_name,
                   snapmirror_label => $snapmirror_lbl,
                   nfsclient        => $nfsclient,);
        push (@subtests, create_parallel_processes('Wf_EID_Hole', \%args_func));

    }

    run_processes_bg('Wf_EID_Hole', \@subtests);
}



=head2 Wf_EID_Hole

Create multiple source files only with holes, create their clones and increase the size of clones using setattr,take snapshot s7.
Verify that EID_Hole will be sent. This can be validated only update update is complete and counters get incremented so covering
this in workflow.pm

=over

=item example

    Wf_EID_Hole(src_vol_obj => $volume);

=item options

=over

=item src_vol_obj

(required) Volume Object - the volume object on which multiple files of various sizes needs to be created.

=item create_file_count

(optional) SCALAR - the number of files to be created.The default value is 2.

=item snapshot_name

(optional) SCALAR - the snapshot name to be created.The default value is s7.

=item snapmirror_label

(optional) SCALAR - the snapmirror label to be created.The default value is weekly.

=item nfsclient

(optional) SCALAR - nfs client.

=back

=cut

sub Wf_EID_Hole{
     my %args = validate_with(
        params => \@_,
        spec   => {
            src_vol_obj  => { type => OBJECT, optional => 0 },
            create_file_count   => { type => SCALAR, optional => 1, default => 1},
            snapshot_name => { type => SCALAR, optional => 1, default =>'s7'},
            snapmirror_label => { type => SCALAR, optional=> 1 ,default => 'weekly'},
            new_clone_size   => {type => SCALAR, optional=>1, default => '1000KB'},
            nfsclient        => {type => SCALAR, optional=>0},
        },
    );

    my $src_vol_obj        = $args{src_vol_obj};
    my $create_file_count  = $args{create_file_count};
    my $ss_name            = $args{snapshot_name};
    my $snapmirror_lbl     = $args{snapmirror_label};
    my $nfsclient          = $args{nfsclient};
    my $new_clone_size     = $args{new_clone_size};
    my $log                = log_global();
    my $Client = NACL::C::Client->new( name => $nfsclient );
    my $client_name = $Client->client();
    my @file_objs;
    my @clone_objs;


    my $vol_name  =  $src_vol_obj->volume;
    my $src_vol_vserver    = $src_vol_obj->vserver;
    my $src_vol_ci         =  NACL::C::Node->new(name => $src_vol_obj->get_local_node_name());
    my $mountPoint = mount_nfs_vol (
            volume      => $src_vol_obj,
            client_name => $client_name,
            nfsvers     => '3',
  );

    my $mkfile_path = "/vol/$vol_name/files";

    $log->comment('Creating no of files with 0KB length and their clones depending on file_count'. $create_file_count);

    @file_objs = NACL::STask::VolumeFile->create(
                command_interface => $src_vol_ci,
                vserver           => $src_vol_vserver,
                size              => '0KB',
                path              => $mkfile_path,
                filecount         => $create_file_count,
               'method-timeout'   => 600,
     );

     foreach my $file_obj(@file_objs){
         my $file_state = $file_obj->state();
         my $file_size = $file_state->file_size();
         my $file_path = $file_obj->path();
         my $new_file_size = '500KB';

         $log->comment('Using testsetattr command will grow it and fill it just with holes');

         my $file_attributes = $file_obj->testsetattr('get-attributes' => 1,
                                                           'file-size' => $new_file_size
                                                      );
       $log->trace('File Attributes are '.Dumper($file_attributes));
     }


     $log->comment('Create clones from those files with just holes');

     for (my $i =0; $i<$create_file_count; $i++){
           my $Clone = $mkfile_path . ".$i" . '_clone';
           my $vol_fileclone_obj = NACL::STask::VolumeFileClone->create(
               command_interface   => $src_vol_ci,
               vserver             => $src_vol_vserver,
               "source-path"       => $mkfile_path . ".$i",
               "destination-path"  => $Clone,
               );


          $log->comment('Fill clones with more holes');

          my $clone_attributes = $vol_fileclone_obj->testsetattr('get-attributes' => 1,
                                                                'file-size' => $new_clone_size);
          push @clone_objs, $vol_fileclone_obj;

      }

    $log->comment('Take the Snapshot');

    NACL::STask::VolumeSnapshot->create(
               command_interface => $src_vol_ci,
               volume            => $src_vol_obj,
               vserver           => $src_vol_vserver,
                snapshot         => "$ss_name",
              'snapmirror-label' => $snapmirror_lbl,
    );


}

=head2 cadchk

This method check CRM entries on given node.It shows which clone is linked to which parent file.
This method returns the inode numbers of the clones present on the node.

=over

=item example

    cadchk(vol_obj => $volume,node_obj=>$node_obj);

=item options

=over

=item vol_obj

(required) Volume Object - the volume object.

=item node_obj

(optional) Node Obj - the node object.

=back

=cut

sub cadchk {
     my %args = validate_with(
        params => \@_,
        spec   => {
            vol_obj  => { type => OBJECT, optional => 0 },
            node_obj => { type => OBJECT, optional => 0 },
            entries  => { type => ARRAYREF, optional => 0 },
            cad_enabled_value => { type => SCALAR, optional => 1, default => 'true'},
            verify_cad_value  => { type => SCALAR, optional => 1, default => 0 },
        },
    );

    my $vol_obj  = $args{vol_obj};
    my $node_obj = $args{node_obj};
    my $entries  = $args{entries};
    my $cadval   = $args{cad_enabled_value};
    my $verify   = $args{verify_cad_value};
    my $vol = $vol_obj->volume;
    my @dst;
    my $log = log_global(1);
    $log->enter();
    my $Dst_Nodescope_Apiset = $node_obj->get_7m_or_nodescope_apiset();
    my $out = $Dst_Nodescope_Apiset->execute_command(
        command => 'cad_test ' . $vol . ' list' );
    if ( $out =~ /num_entries:\s*(\d+)/i ) {
        $log->comment("The number of cad entries found is $1");
    }

    if (!$verify &&
        ($1 != scalar(@{$entries}))) {
        croak('Incorrect number of entries found in the cad file');
    }

    my @arr = split( /\n/, $out );
    my (@not_found, @incorr_value);
    foreach (@arr) {
        if ( $_ =~
/Inum\s*=\s*(\d+).*cad-enabled\s*=\s*(\w+).*generation\s*=\s*(\d+).*/i
          )
        {
            push( @dst, $1);
            #check incorrect entries only if user is explictly asking for it.
            #if user is only wanting to verify the CAD_enabled value .. skip this
            if (!$verify &&
                 (! grep { $1 == $_ } @{$entries})) {
                push @not_found, $1;
            }

            #check only for the CAD_enabled field value for the entries passed in
            if ($verify &&
                (grep { $2 eq $_ } @{$entries})){
                $log->comment('Verify CAD_enabled value for inode: ' . $1);
                if ($2 ne $cadval) {
                    push @incorr_value, $1;
                }
            }
        }

    }
    $log->debug(
"The inum cad enabled option and generation number from destination is @dst\n"
    );

    my $errtxt;
    if (scalar @not_found) {
        $errtxt .= 'Found unexpected entries in the CAD file: ' . Dumper(\@not_found) ;
    }

    if (scalar @incorr_value) {
       $errtxt .= "\n" . 'Found CAD entries with incorrect CAD_enabled value: '. Dumper(\@incorr_value);
    }

    if (defined $errtxt) {
        croak($errtxt);
    } else {
        $log->comment('CAD entries were as expected');
    }

    $log->exit();
    return;
}

# private method
sub _get_inum {

    my %args = validate_with(
        params => \@_,
        spec   => {
            src_node_obj  => { type => OBJECT, optional => 0 },
            vol_src_obj   => { type => OBJECT, optional => 0},
            clone         => {type =>  SCALAR, optional => 0},
        },
    );

    my $src_vol_obj        = $args{vol_src_obj};
    my $src_node_obj       = $args{src_node_obj};
    my $clone              = $args{clone};

    my($clone_path,$clone_nm) = ( $clone =~ /(.*)\/(.*)$/ );

   my $log = log_global(1);
   $log->enter();

   $log->comment("The clone for which inum is needed is $clone_nm");

   my $src_Nodescope_Apiset = $src_node_obj->get_7m_or_nodescope_apiset();
   my $file1inum;
   my $out = $src_Nodescope_Apiset->execute_command(
        command => 'ls -l /vol/' . $src_vol_obj->volume );
   if ( $out =~ /No such file or directory/i ){

   }

   my @arr = split( /\n/, $out );
   foreach (@arr) {
      if ( ( $_ =~ /\b$clone_nm\b/ ) && ( $_ =~ /\d+\s+(\d+)/ ) ) {
            $file1inum = $1;
       }
   }

  $log->comment('inum num returned for : ' . $clone_nm . ' is: ' . $file1inum);
  return $file1inum;

}

=head2 prepare_for_traffic

Provides handles to closures which will let you start/stop traffic to one or more luns.
The prepare method take params that will remain constant through out the workflow for e.g.
    - command_inerfaces
    - nfsclients/sanclients
    - cg_item_type
    - runtime

Parameters that can vary through the workflow can be directly passed while invoking the closures.

The method will maintain an internal hash of the luns that it has started IO on and the
corresponding IOMan object (returend from IOWH* library) and the iorecord that needs to be passed to IOMan while stopping IO on that lun.

The library does:
    - start IO on multiple luns
    - start read ONLY IO on dst luns
    - stop IO on select luns
    - stop IO on all luns

*IOWH - IO Workload helper library

=over

=item Example

    my ($start_traffic, $stop_traffic) = prepare_for_traffic (
                        command_interface       =>  cmd_int1,
                        cg_item_type            =>  'lun',
                        sanclients              =>  [client1, client2],
                        runtime                 => '7200');

     $start_traffic->(luns                     => [$luns[0], $luns[1]],
                      SAN_workload_type        => 'SAN_IO_Read_Only',);
                        OR
     $start_traffic->(luns                     => [$luns[0], $luns[1]]);

     $stop_traffic->(itemnames                 => [$luns[0]->path()]);
                        OR
     $stop_traffic->();


=item Options

=over

=item command_interface

(Required) The source or dst node objects

=item vols

(Optional) The list of volumes that will be used while starting NFS traffic. Once the support to pre-create files is provided we can discontinue this param.

=item sanclients

(Optional) if not doing SAN. The list of clients for doing SAN traffic.

=item nfsclients

(Optional if not doing NFS. The list of clients for doing NFS traffic.

=item

(Optional) Set if you want the method to mount files for NFS traffic.
(Default) 1

=item Param definitions for start_traffic

=over

=item files

(Optional) if not doing NFS. The list of files to start IO on.


=item luns

(Optional) if not doing SAN. The list of luns to start IO on.

=item SAN_workload_type

(Optional) The type of SAN workload to invoke
(Default) SAN_IO_Random_BS

=back

=item Param definitions for stop_traffic

=over

=item itemnames

(Optional) The list of lun paths/file names to stop IO on.
If none are passed then IO will be stopped on all luns.

=item Presumptions

    - the luns are already mapped to the appropriate igroups
    - the iscsi targets have been logged in and the luns have been discovered on the clients.

=item Note:

    - No support for NFS traffic.
    - List of different kind of SAN worloads can be found : http://gilroy.lab.netapp.com/x/eng/localtest/noarch/dev/lib/San/Task/IOWorkloadHelper.pm#_initialize_workloads-
    - The SAN_IO_Read_Only workload does NOT do real time comparision of data that gets written on the src with the data that is read from the dst. See burt #951832 for more details.

=back

=back

=back


=cut

sub prepare_for_traffic {

    my %args = validate_with (
        params => \@_,
        spec   => {
            command_interface   =>  { type => HASHREF,   optional => 0 },
            vols                =>  { type => ARRAYREF, optional => 1 },
            cg_item_type        =>  { type => SCALAR,   regex => qr/file|lun/, default => 'lun' },
            runtime             =>  { type => SCALAR,   optional => 1,         default => 7200 },
            nfsclients          =>  { type => ARRAYREF, optional => 1 },
            sanclients          =>  { type => ARRAYREF, optional => 1 },
            do_mount            =>  { type => SCALAR,   optional => 1,         default => 1 },
        }
    );

    my $cmd_interface = $args{command_interface};
    my $vols          = $args{vols};
    my $cg_item_type  = $args{cg_item_type};
    my $runtime       = $args{runtime};
    my $nfsclients    = $args{nfsclients};
    my $sanclients    = $args{sanclients};
    my $do_mount      = $args{do_mount};
    my $active_io_map;
    my $log           = log_global();

    #make sure sanclients/nfsclients are passed in when cg-item is lun/file
    if ($cg_item_type =~ /lun/) {
        if (!defined $sanclients) {
                NACL::Exceptions::InvalidParam->throw('Argument sanclients should be passed in to prepare_for_traffic when cg_item_type is lun');
        }
    } elsif ($cg_item_type =~ /file/) {
        if (!defined $nfsclients) {
                NACL::Exceptions::InvalidParam->throw('Argument nfsclients should be passed in to prepare_for_traffic when cg_item_type is file');
        }
    }

    $log->comment('Create the IOWorkloadHelper object');
    my $IOWorkloadHelper_obj = new San::Task::IOWorkloadHelper();

    my $start_traffic = sub {
         my %args = validate_with (
            params => \@_,
            spec   => {
                files               =>  { type => ARRAYREF, optional => 1 },
                luns                =>  { type => ARRAYREF, optional => 1 },
                SAN_workload_type   =>  { type => SCALAR, default => 'SAN_IO_Random_BS'},
                rescan              =>  { type => SCALAR, default => 0, optional => 1},
            },
         );

        my $files        = $args{files};
        my $luns         = $args{luns};
        my $san_workload = $args{SAN_workload_type};
        my $dorescan     = $args{rescan};
        my @lunpaths;
        my @iorecords;

        if ($cg_item_type =~ /lun/) {
            if (!defined $luns) {
                NACL::Exceptions::InvalidParam->throw('Argument luns should be passed in to start_traffic when cg_item_type is lun');
            }


            $log->comment('Getting the lunpaths');
            foreach my $l (@{ $luns }) {
                 push (@lunpaths, $l->path());
            }

            if ($dorescan) {
                $log->comment('Doing a rescan on the clients for new luns or for luns that have changed to RW from read-only');
                 my @dev = $IOWorkloadHelper_obj->get_client_devices (
                                clients         => $sanclients,
                                rescan          => 1,
                                command_ints    => [$cmd_interface],
                                lun_paths        => \@lunpaths);
                if (scalar (@dev) == 0) {
                    NATE::BaseException->throw ('A rescan of the luns failed');
                }
            }

            $log->comment('Starting SAN IO on luns: ' . Dumper(\@lunpaths)) ;

            my $IOManObj =  $IOWorkloadHelper_obj->run_san_workloads (
                    clients         => $sanclients,
                    workload        => $san_workload,
                    runtime         => $runtime,
                    background      =>  1,
                    command_ints    => [$cmd_interface],
                    lun_paths        => \@lunpaths);

            $log->comment('Get IO man records');
            foreach my $cl (@{$sanclients}) {
                $log->comment('Get records for client name: ' . $cl);
                push @iorecords, $IOManObj->get_io_records( -client  => $cl );
            }

            $log->comment('Creating the lun<->trafficobject dictonary');
            foreach my $rec (@iorecords) {
                $active_io_map->{$rec->{lun_path}}->{iorecord} = $rec; #store the io record
                $active_io_map->{$rec->{lun_path}}->{iomanobj} = $IOManObj; #store the ioman object
            }

            $log->comment('The dictonary created is: ' . Dumper(keys %{$active_io_map}));
            $log->comment('Method done: Finished starting IO');
        } else {
            if (!defined $files) {
                NACL::Exceptions::InvalidParam->throw('Argument files should be passed in to start_traffic when cg_item_type is file');
            }
            $log->comment('Start NFS IO');
        }
    };


    my $stop_traffic = sub {
        my %args = validate_with (
            params => \@_,
            spec   => {
                itemnames =>  { type => ARRAYREF, optional => 1 },
            }
        );

        my @items ;
        my @subtests;


        #if user has not specified a specific item to stop traffic on; then
        #do the operation on all active io objects.

        if (!defined $args{itemnames}) {
            $log->comment('IO will be stopped on all active items');
            @items = keys %{ $active_io_map };
        } else {
            @items = @{ $args{itemnames}};
        }

        $log->comment('IO will be stopped on the following items: ' . Dumper(\@items));
        foreach my $it (@items) {
            $log->comment('Stopping IO on item : ' . $it);
            my $iomanobj = $active_io_map->{$it}->{iomanobj};
            my $iorecord = $active_io_map->{$it}->{iorecord};

            try {
                $iomanobj->stop ( -io_records => [$iorecord] );
            } catch NATE::BaseException with {
                my $ex = shift;
                $log->warn("Failure while trying to stop IO on item: $it" .
                           '. Exception: ' . $ex->text());
            };

            $log->comment(' ... success');
            $log->comment('Updating the active IO hash');

            delete $active_io_map->{$it};
        }

        $log->comment('The active IO updated hash is: ' .  Dumper(keys %{$active_io_map}));
        $log->comment('Method done: Finished stopping IO');
    };

    $log->exit();
    #return the two subroutine references to user.
    return ($start_traffic, $stop_traffic);
}

=head2 prepare_dst_luns

Given a list of src luns that are mirrored via a SM relationship, this method will:

    - find the corresponding mirrored luns on the dst
    - map the luns to the same igroup that the source lun is mapped to.
      This is required so that DT can do real time comparision of data that gets written on the src with data that gets read from the destination.
    - discover the devices on the clients,

Returns an ARRAYREF of dst luns that have been mapped and discovered.

=over

=item Example


    my $dst_luns = prepare_dst_luns (
                                  dst_node         => $n,
                                  src_luns         => [$src_lun_obj1, $src_lun_obj2],
                                  sanclients       => ['scspr0049419006', 'scspr0049419003'],);

=item Options

=over

=item dst_node

(Required) The dst node object to be used to find the luns that need to be prepared for read traffic.

=item src_luns

(Required) The list of src luns objects so that we can query the dst for the same luns. The src luns are also required to get the igroup that they are mapped to so that the dst luns can
be mapped to the same igroup.

=item sanclients

(Required) The list of clients that will be used to start traffic on.

=item Presumptions

    - the dst node iscsi targets have already been discovered and logged in on the clients that will be passed in to this method.
    - the dst should have an igroup with the same initiator list as that on the src igroup. It is ok if the igroup name is different.


=back

=back

=cut

sub prepare_dst_luns {

    my %args = validate_with (
        params => \@_,
        spec   => {
            dst_node            =>  { type => OBJECT, optional => 0 },
            src_luns            =>  { type => ARRAYREF, optional => 0 },
            sanclients          =>  { type => ARRAYREF, optional => 0 },
        }
    );

    my $cmd_interface = $args{dst_node};
    my $src_luns      = $args{src_luns};
    my $sanclients    = $args{sanclients};
    my $log           = log_global();
    my @dst_luns;

    foreach my $l (@{ $src_luns }) {
        $log->comment('Looking for lun : ' . $l->lun() .
                      ' on the src node to get the igroup information');
        #doing this call here because NACL::C::Lun does not give back the
        #igroup information
        my $srcobj = NACL::CS::LunMapped->fetch (
                        command_interface       => $l->command_interface(),
                        filter                  => {
                                lun             => $l->lun() });

        $log->comment('Lun: ' . $srcobj->lun . ' is mapped to igroup: ' . $srcobj->igroup());
        $log->comment('Get the initator name for the igroup: ' . $srcobj->igroup());

        my $src_ig = NACL::CS::LunIgroup->fetch (
                            command_interface       => $l->command_interface(),
                            filter                  => {
                                igroup              => $srcobj->igroup() });

        my @initiators = @{ $src_ig->initiator() };
        $log->comment('Igroup: ' . $src_ig->igroup() . ' is mapped to inititator: ' .
                      Dumper(@initiators));


        $log->comment('Find the igroup with the initiator: ' . Dumper(@initiators) .
                      ' on the dst');

        my $dst_ig = NACL::CS::LunIgroup->fetch (
                        command_interface       =>  $cmd_interface,
                        filter                  => {
                                initiator       => \@initiators });

        $log->comment('found the igroup: ' . $dst_ig->igroup() . ' on the dst');

        $log->comment('Looking for the same lun: ' . $l->lun() . ' on the dst node');
        my $obj =  NACL::C::Lun->find (
                        command_interface       => $cmd_interface,
                        filter                  => {
                                lun             => $l->lun(),
                                });

        if (! defined $obj) {
            NATE::BaseException->throw ('did not find lun with name: ' . $l->lun .
                                        'on the destination node');
        }

        $log->comment('Mapping the lun to igroup : ' . $dst_ig->igroup());
        $obj->map ( igroup => $dst_ig->igroup() );
        push (@dst_luns, $obj);
    }

    $log->comment('Discover the dst luns on the client');
    #note the expectation is that the igroups with the respective initiators have
    #already been created and the clients are logged into the target.

    $log->comment('Getting the lunpaths');
    my @lunpaths;
    foreach my $l (@dst_luns) {
         push (@lunpaths, $l->path());
    }

    my $iowh =  new San::Task::IOWorkloadHelper();
    my @dev = $iowh->get_client_devices (
                    clients         => $sanclients,
                    command_ints    => [$cmd_interface],
                    lun_paths       => \@lunpaths,
                    rescan          => 1);

    if (scalar (@dev) == 0) {
        NATE::BaseException->throw ('Failed to discover the dst luns on the clients. ' .
                                    'Make sure the clients are logged in to the iscsi target\(s\) ' .
                                    'before calling this method');
    }

    $log->comment('Successfully discovered the luns. The prepared dst luns are: ' .
                  Dumper(\@dst_luns));

    return \@dst_luns;
}



=head2 start_read_on_dst_cg_files

This method is to run SIO read on dst CG files. It returns traffic objects.

=over

=item options

=over

=item required

(required) client_name - SCALAR, name of the client to mount volume and create file
(required) volume - OBJECT, volume containing CG items
(required) cg_items - ARRAYREF, file names for the files to run SIO.

=item optional

(optional) cg_tyoe - files or LUNs. Currently supports files only.
(optional) io_size - io block size.
(optional) nfsvers - nfs version.

=back

=back

=cut


sub start_read_on_dst_cg_files{

    my %args = validate_with(
        params => \@_,
        spec   => {
            client_name => { type => SCALAR, optional => 0 },
            volume      => { type => OBJECT, optional => 0 },
            cg_items    => { type => ARRAYREF, optional => 0 },
            cg_type     => { type => SCALAR, optional => 1, regex => qr/file|lun/, default => 'file'},
            io_size     => { type => SCALAR|UNDEF, default => '4k' },
            nfsvers     => { type => SCALAR, default => '3' },
        },
    );

    my @io_objs;
    my $client_obj = NACL::C::Client->new( name => $args{'client_name'} );

    if ($args{'cg_type'} eq 'file') {

        my $mount_point = mount_nfs_vol (
            volume      => $args{'volume'},
            client_name => $args{'client_name'},
            nfsvers     => $args{'nfsvers'},
        );

        foreach my $file (@{$args{'cg_items'}}) {

            my $file_path = '/vol/' . $args{'volume'}->volume() . '/' . $file;

            my $file_obj = NACL::CS::VolumeFile->fetch(
                command_interface => $args{'volume'}->command_interface(),
                filter            => {
                    path    => $file_path,
                    vserver => $args{'volume'}->vserver(),
                },
                'requested_fields' => ['file-size'],
            );

            $file_path = $mount_point . '/' . $file;

            # read only SIO
            my @io_obj = start_nfs_traffic (
                client => $client_obj,
                read   => 100,
                paths  => [$file_path],
                file_size  => $file_obj->file_size(),
                block_size => $args{'io_size'},
            );

            push @io_objs, @io_obj;
        }
    }

    return @io_objs;
}


# private method
sub _stagger {
    my $delay = shift;
    $delay = int(rand($delay));
    my $log = log_global();
    $log->enter();
    $log->comment('Staggered start delay is: ' . $delay);
    $log->exit();
    Tharn::snooze($delay);
}

#private method
sub _start_dt_traffic {

    my %args = validate_with(
        params => \@_,
        spec   => {
            traffic_obj => { type => OBJECT, optional => 0 },
            lun_size    => { type => SCALAR, optional => 0 },
            on_failure =>
                { type => CODEREF | SCALAR, default => 'die', optional => 1 },
            on_success => { type => CODEREF, optional => 1 },
        },
    );

    my $traffic_obj    = $args{traffic_obj};
    my $lun_size_bytes = $args{lun_size};

    #
    # Use 1% of the total lun size for each dt pass. There will be
    # multiple dt passes
    #
    my $dt_limit_bytes = int( $lun_size_bytes * 0.01 );
    my $modulo         = $dt_limit_bytes % 512;
    $dt_limit_bytes   -= $modulo;

    if ( $dt_limit_bytes < 512 ) {
        $dt_limit_bytes = 512;
    }

    if ( defined( $args{on_success} ) ) {
        return $traffic_obj->start(
            tool   => 'dt',
            iotype => 'random',
            limit  => $dt_limit_bytes,

            # This puts a 5 second delay at the end of each pass
            cdelay            => '1',
            oflags            => 'trunc',
            bs                => '4k',
            incr              => 'var',
            min               => '2k',
            max               => '256k',
            pattern           => 'iot',
            runtime           => '-1',
            dispose           => 'keeponerror',
            prefix            => "'%d@%h'",
            disable           => 'pstats',
            enable            => 'compare,lbdata,debug,syslog',
            'privilege-level' => 'root',
            on_failure        => $args{on_failure},
            on_success        => $args{on_success},
        );

    }
    return $traffic_obj->start(
        tool   => 'dt',
        iotype => 'random',
        limit  => $dt_limit_bytes,

        # This puts a 5 second delay at the end of each pass
        cdelay            => '1',
        oflags            => 'trunc',
        bs                => '4k',
        incr              => 'var',
        min               => '2k',
        max               => '256k',
        pattern           => 'iot',
        runtime           => '-1',
        dispose           => 'keeponerror',
        prefix            => "'%d@%h'",
        disable           => 'pstats',
        enable            => 'compare,lbdata,debug,syslog',
        'privilege-level' => 'root',
        on_failure        => $args{on_failure},
    );
}

=head2 start_traffic

This method is to run IOs on files or LUNs.SIO is used for files and dt is used for
LUNs.

=over

=item options

=over

=item required

(required) client_name - SCALAR, name of the client to mount volume and create file
(required) volume - OBJECT, volume containing CG items
(required) cg_type - SCALAR, CG item type: file or lun.

=item optional

(optional) lun_set - ARRAYREF, a set of LUNs to run DT IO
(optional) file_paths - SCALAR, absolute file paths for NFS IO.
(optional) file_size - SCALAR, file size for NFS IO.

=back

=back

=cut


sub start_traffic {

    my %args = validate_with(
        params => \@_,
        spec   => {
            client_name => { type => SCALAR, optional => 0 },
            volume      => { type => OBJECT, optional => 0 },
            file_paths  => { type => ARRAYREF, optional => 1 },
            file_size   => { type => SCALAR, optional => 1, default => '500m', },
            lun_set     => { type => ARRAYREF, optional => 1 },
            cg_type     => { type => SCALAR, optional => 0, regex => qr/file|lun/, },
        },
    );

    my @io_objs;

    my $Client = NACL::C::Client->new( name => $args{client_name} );

    if($args{cg_type} eq 'file') {

        my @file_paths = @{$args{file_paths}};

        @io_objs = start_nfs_traffic (
            client    => $Client,
            paths     => \@file_paths,
            tool      => 'sio',
            file_size => $args{file_size},
        );

    } else {

        my @lunset = @{$args{lun_set}};

        my @lifs = NACL::C::NetworkInterface->find_up_data_iscsi_interfaces(
            command_interface => $args{volume}->command_interface, );

        my %LUNS;
        foreach my $lun ( @lunset ) {
            my $lun_path = NACL::MTask::Client::LunMount->lun_discovery(
                command_interface       => $Client,
                filer_command_interface => $args{volume}->command_interface,
                vserver                 => $args{volume}->vserver(),
                filer_lif               => $lifs[0],
                'lun-path'              => $lun->path(),
            );

            $LUNS{$lun->lun()}{traffic_obj} = NACL::MTask::Traffic->new(
                command_interface => $Client,
                paths             => $lun_path,
            );
        }

        foreach my $lun ( @lunset ) {
            _start_dt_traffic(
                lun_size    => $lun->state->size(),
                traffic_obj => $LUNS{$lun->lun()}{traffic_obj},
            );
            push @io_objs, $LUNS{$lun->lun()}{traffic_obj};
        }
    }

    return @io_objs;
}

=head2 datagen

Executes Datagen_CMode.thpl with the provided parameters
This is designed to work with the yaml interface of ice king by passing in
*volume objects (for example)

Yaml example:
    - task:
        testcase: NACL::AppBuilder::DpgWorkflows::Workload->datagen(vol)
        type: testcode
        param:
            vol:
                resources: *Vol


=over

=item options

=over

=item required

 vol - arrayref of vol objects (picks the 1st) or a volume name

=item optional

 client - client to run datagen on (defaults to localhost if not provided)
 datagen - custom thpl path to use, defaults to dp/tools/DataGen/DataGen_CMode.thpl
 vserver - vserver for volume above, or it'll find it in the vol object
 aggr - aggr for volume above, or it'll find it in the vol object
 filer - node for volume above, or it'll find it in the vol object

 If only the volume is passed, the aggr,vserver, and node will be taken from the vol object.

 Any other parameters you can pass to DataGen can be passed in and they will be
 directly passed down to the thpl

=back

=back

=cut


sub datagen {
	my $self = shift;

    my %args = validate_with(
        params => \@_,
        allow_extra => 1,
        spec   => {
            datagen => { type => SCALAR, default => 'dp/tools/DataGen/DataGen_CMode.thpl', },
            client  => { type => ARRAYREF | SCALAR, optional => 1 },
            vserver => { type => ARRAYREF | SCALAR, optional => 1 },
            aggr    => { type => ARRAYREF | SCALAR, optional => 1 },
            vol     => { type => ARRAYREF | SCALAR, required => 1 },
            filer   => { type => ARRAYREF | SCALAR, optional => 1 },
        },
    );

    my $log = log_global();

    $log->trace('datagen raw args:'. Dumper \%args);



    #delete to remove from args we are passing down
	my $datagen = delete $args{datagen};
	#add client
	my $client = delete $args{client} if defined $args{client};


	my %lookup = (
		'aggr' => 'aggregate',
		'filer' => 'node',
		'vserver' => 'vserver',
		'vol' => 'volume',

	);

	my $volstate;
	if (ref $args{vol} eq 'ARRAY' && (ref $args{vol}->[0]) =~ /volume/i) {
	    $volstate = $args{vol}->[0]->state();
	}
    $log->trace('volstate:'. Dumper $volstate);

	foreach my $key (qw ( vol aggr vserver filer ) ) {
		my $val = delete $args{$key};
		if (ref $val eq 'ARRAY') {
			$val = $val->[0]->{$lookup{$key}};

		} else {
			if ($val) {
			    #scalar value passed, assume correct
			} else {
				#grab these from the volume object if needed
				if ($key eq 'vserver') {
					$val = $volstate->{vserver};
				} elsif ($key eq 'aggr') {
					$val = $volstate->{aggregate};
					if ($val eq '-') {
					    my $array = $volstate->{aggr_list};
						$val = $array->[0];
				    }
				} elsif ($key eq 'filer') {
					$val = $volstate->{node};
					if ($val eq '-') {
					    my $array = $volstate->{nodes};
						$val = $array->[0];
				    }
				}
			}
		}
		$args{uc $key} = $val;
	}

    $log->comment('datagen processed args:'. Dumper \%args);

    my $proc = NATE::Process->new(
		codespec => $datagen,
		params => \%args,
	    env => { GNOTE_NOSYSCONFIG => 1,},
		( $client ? (client => $client) : () ),
    );
    $proc->start();
    $proc->wait();
    $proc->destroy();

}

#Use some of the clone files in s1, punch holes in existing data blocks (BLOCK TO HOLE) + grow file with data and punch holes in the region growed. Make sure to punch multiple holes in the clone file, some at the begining, some in middle and some at the end. Take snapshot s3
sub DataSet_GrowClones_PunchHoles {
     _DataSet_GrowFiles_PunchHoles(@_,snapshot_name=>'s3');
}

#Use some of the source files in s1, punch holes in existing data blocks (BLOCK TO HOLE) + grow file with data and punch holes in the region growed. Make sure to punch multiple holes in the clone file, some at the begining, some in middle and some at the end. Take snapshot s4. This one tests that NO send_hold will be invoked. The clone files are unchanged. Data blocks are only sent for source files, no clones ops.
sub DataSet_GrowParents_PunchHoles {
    _DataSet_GrowFiles_PunchHoles(@_,snapshot_name=>'s4');
}

# Assumptions
# All files should be on the same volume
# Will pick the first set of file_count files to modify
# file_count should be lesser than the # of files in file_obj - if its lesser will not die but will use all the files
# there should be enough room to grow the files
sub _DataSet_GrowFiles_PunchHoles {
    my %args = validate_with(
        params      => \@_,
        allow_extra => 0,
        spec        => {
            client      => {type => ARRAYREF|SCALAR}, # nfs or san client depending on file_type
            file_obj    => {type => ARRAYREF},
            file_count  => {type => SCALAR, default => 1}, # number of files to modify
            file_type   => {type => SCALAR, regex => qr/(file|lun)/, default => 'file'},
            holes_in_begin => {type => SCALAR, default => 1, regex => qr/0|1/},
            holes_in_end   => {type => SCALAR, default => 1, regex => qr/0|1/},
            holes_in_mid   => {type => SCALAR, default => 1, regex => qr/\d+/},
            hole_size_in_begin => {type => SCALAR, default => '4K'},
            hole_size_in_end => {type => SCALAR, default => '4K'},
            hole_size_in_mid => {type => SCALAR, default => '4K'},
            offset_between_holes => {type => SCALAR, default => '4K'}, # should match default extent size
            grow_size => {type => SCALAR, default => '20k'},
            snapshot_name => {type => SCALAR},
            snapmirror_label => {type => SCALAR, default => 'weekly'},
            clones_to_create => {type => SCALAR, default => 3},
            sleep_sec        => {type => SCALAR, default => 120,},
        },
    );

    my @file_objs = @{$args{file_obj}};
    my $log       = log_global();
    my $clone_cnt = $args{clones_to_create};
    my $ssecs     = $args{sleep_sec};
    my @new_clones;

    # splice/pick file_count # of files from file_obj - start from first
    my @chosen_file_objs = splice (@file_objs, 0, $args{file_count});
    $log->trace('Chosen files are ...'.Dumper(@chosen_file_objs));
    my $extent_size;
    foreach my $file_obj ( @chosen_file_objs ) {
        my $vol_stask_obj = $file_obj->get_containing_volume_obj();
        my $cmd_interface = $vol_stask_obj->command_interface();
        $log->trace('FileObj is...'.Dumper $file_obj);
        $log->log('Path='.$file_obj->path());
        $log->comment('Create ' . $clone_cnt . ' clones for ' . $file_obj->path());
        for (my $i = 0; $i < $clone_cnt; $i++) {
            my $clone = $file_obj->path() . '_prime' . $i;
            $log->comment( '... Creating clone: ' . $clone);
            my $clone_file_obj = NACL::STask::VolumeFileClone->create(
                        command_interface   => $cmd_interface,
                        vserver             => $vol_stask_obj->vserver(),
                        'source-path'       => $file_obj->path(),
                        'destination-path'  => $clone);
            push (@new_clones, $clone_file_obj);
        }
        $extent_size = $vol_stask_obj->get_one_state_attribute('default-extent-size') if !$extent_size;
    }

    push (@chosen_file_objs, @new_clones);
    $log->comment('Final list');
    foreach my $file_obj ( @chosen_file_objs ) {
       $log->log('Path='.$file_obj->path());
    }

    # Grow files
    _grow_files(file_obj  => \@chosen_file_objs,
                grow_size => $args{grow_size},
                client    => $args{client},
                sleep_sec => $ssecs,
                );

    # Punch holes in the files
    my $hole_range_info = _punch_holes_in_files(file_obj       => \@chosen_file_objs,
                                                client         => $args{client},
                                                holes_in_begin => $args{holes_in_begin},
                                                holes_in_end   => $args{holes_in_end},
                                                holes_in_mid   => $args{holes_in_mid},
                                                hole_size_in_begin => $extent_size,
                                                hole_size_in_end   => $extent_size,
                                                hole_size_in_mid   => $extent_size,
                                                offset_between_holes => $extent_size,
                                                );
    $log->comment('Holes are punched. Hole range info is '.Dumper($hole_range_info));

    # Take Snapshot
    _take_snapshot(file_obj => \@chosen_file_objs,
                    snapshot_name => $args{snapshot_name},
                    snapmirror_label => $args{snapmirror_label},
                    );

    return(\@chosen_file_objs);
}

# This method will pulls the info from file_obj and take SS
sub _take_snapshot {
    my %args = validate_with(
        params      => \@_,
        allow_extra => 0,
        spec        => {
            file_obj         => {type => ARRAYREF},
            snapshot_name    => {type => SCALAR},
            snapmirror_label => {type => SCALAR, default => 'weekly'},
        },
    );
    my $log = log_global();

    # Extract command_interface, vserver name, vol name from the file object.
    # Can use any one file since the assumption is that all files are on same volume
    my $first_fileobj = @{ $args{file_obj} }[0]; # just pick first file

    my $volume_stask_obj = $first_fileobj->get_containing_volume_obj();

    $log->log("Take snapshot snapshot=$args{snapshot_name} ".
                "on volume=$volume_stask_obj->volume vserver=$first_fileobj->{vserver} ".
                "using label=$args{snapmirror_label}");
    my $ss_obj = $volume_stask_obj->snapshot_create(
                                        snapshot           => $args{snapshot_name},
                                        'snapmirror-label' => $args{snapmirror_label},
                                    );
    return $ss_obj;
}

# this will punch holes in files using nodescope command
# ideally would like to punch holes via the client - TBD for now
sub _punch_holes_in_files {
    my %args = validate_with(
        params      => \@_,
        allow_extra => 0,
        spec        => {
            file_obj       => {type => ARRAYREF},
            client         => {type => ARRAYREF|SCALAR, optional=> 1}, # right now not used but want to use it going fwd
            holes_in_begin => {type => SCALAR, default => 1, regex => qr/0|1/},
            holes_in_end   => {type => SCALAR, default => 1, regex => qr/0|1/},
            holes_in_mid   => {type => SCALAR, default => 2, regex => qr/\d+/},
            hole_size_in_begin => {type => SCALAR, default => '4K'}, # max 4M
            hole_size_in_end   => {type => SCALAR, default => '4K'}, # max 4M
            hole_size_in_mid   => {type => SCALAR, default => '4K'}, # max 4M
            offset_between_holes => {type => SCALAR, default => '4K'}, # should match extent size for now
        },
    );
    my $log = log_global();

    NACL::Exceptions::InvalidParam->throw('hole_size_in_begin should be specified if holes_in_begin is used')
        if ($args{holes_in_begin} && !$args{hole_size_in_begin});
    NACL::Exceptions::InvalidParam->throw('hole_size_in_mid should be specified if holes_in_mid is used')
        if ($args{holes_in_mid} && !$args{hole_size_in_mid});
    NACL::Exceptions::InvalidParam->throw('hole_size_in_end should be specified if holes_in_end is used')
        if ($args{holes_in_end} && !$args{hole_size_in_end});

    # assumption is that the size of the holes are within the file size
    # XXX add check later
    # 50KB / 200KB are typical sizes

    my $hole_size_in_begin = convert_to_bytes($args{hole_size_in_begin});
    my $hole_size_in_mid = convert_to_bytes($args{hole_size_in_mid});
    my $hole_size_in_end = convert_to_bytes($args{hole_size_in_end});
    my $offset_between_holes = convert_to_bytes($args{offset_between_holes});
    $log->log("hole size in begin=$hole_size_in_begin ".
              "hole size in mid=$hole_size_in_mid ".
              "hole size in end=$hole_size_in_end".
              "offset size between holes=$offset_between_holes"
            );

    # Rules for punching holes for metawafl
    # hole should be aligned to extent size
    # hole size can be a multiple of extent size
    # file punch hole <filepath> <offset> <length> ... offset and length should be a multiple of block size
    my %hole_ranges;
    foreach my $file_obj (@{$args{file_obj}}) {
        my @hole_range_info;
        my $file_state = $file_obj->state();
        my $file_size = $file_state->file_size();
        my $file_path = $file_obj->path();

        my $file_attributes = $file_obj->testsetattr('get-attributes' => 1);
        $log->trace('File Attributes are '.Dumper($file_attributes));

        # XXX - will not work with Object packing
        my $allocation_size = $file_attributes->{allocation_size}; # this is the size that has been allocated
        $log->log('Allocation size for file is '.$allocation_size);

        my $offset;
        if ($args{holes_in_begin}) {
            push @hole_range_info, { 'hole-start' => 0, 'hole-size' => $hole_size_in_begin };
            $offset += $hole_size_in_begin; # offset the first hole
        }
        foreach (my $i=0; $i<$args{holes_in_mid}; $i++) {
            if (!defined $offset) { $offset = 0; }
            if (($offset + $hole_size_in_mid) >= $file_size) {
                $log->log('Skip. Offset + hole size for mid is beyond file size');
                last;
            }
            $offset += $offset_between_holes; # offset between holes
            push @hole_range_info, { 'hole-start' => $offset, 'hole-size' => $hole_size_in_mid };
            $offset+=$hole_size_in_mid;
        }
        if ($args{holes_in_end}) {
            my $hole_end_start = $allocation_size-$hole_size_in_end;
            push @hole_range_info, { 'hole-start' => $hole_end_start, 'hole-size' => $hole_size_in_end };
        }
        $log->log("Will punch holes. File Path=$file_path. File Size=$file_size. Hole Range info=".Dumper(@hole_range_info));

        # TBD punch in parallel
        $file_obj->punch_hole( determine_appropriate_nodescope_node => 1,
                                'hole-range-info' => \@hole_range_info,
                                apiset_must => { interface => 'CLI' }, # use CLI for now burt 979522
                            );
        $hole_ranges{$file_state->name} = \@hole_range_info;
    }
    return \%hole_ranges;
}


=head2 Wf_data_transitions

This method excercises the different data transitions using source/parent files
passed in by the user or by creating new files. Data transitions include:

HOLE TO DATA - grow parent with holes; grow clones with data
NONE TO DATA - do nothing to parent; grow clones with data
DATA TO NONE - grow parent with data; do nothing to clones
DATA TO HOLE - grow parent with data; grow clones with holes
NONE TO HOLE - do nothing to parent; grow clones with holes
HOLE TO NONE - grow parent with holes; do nothing to clones
DATA CHANGED BETWEEN SRC and CLONE - Change the same region of the parent and clone file with different data

The method will perform the following sequence.

- create a clone
- perform DATA_HOLE and HOLE_DATA
- cut a ss

- create a clone
- perform DATA_NONE and HOLES_NONE
- cut a ss

- create a clone
- perform NONE_DATA and NONE_HOLES
- cut a ss

- create a clone
- perform DATA_CHANGE between parent and clone
- cut a ss

=over

=item Options

=over

=item file_obj

(required) The list of source files objects to excercise data transitions on.
The method will create a clone file before each transition and cut a ss after each
transition.

=item client

(required) client name to use to do IO

=item grow_size

(Optional) By how much to grow the file (clones/parent) to excercise the transitions

=item file_count

(Optional) How many files to choose from the list to perform the data transitions on.
(Default) 1

=item snapmirror_label

(Optional) What snapmirror label to apply on the ss.
(Default) weekly.

=item sleep_sec

(Optional) How long to run IO when we need to fill a grown region with data.
(Default) 180

=item file_type

(Optional) Whether the transitions are being performed on a file/lun
(Default) file

=item snapshot_prefix

(Required) SS prefix to use for the snapshots this method will cut

=back

=back

=cut

sub Wf_data_transitions {
    my %args = validate_with(
        params      => \@_,
        allow_extra => 1,
        spec        => {
            client               => {type => ARRAYREF|SCALAR}, # nfs or san client depending on file_type
            file_obj             => {type => ARRAYREF},
            file_count           => {type => SCALAR, default => 1}, # number of files to modify
            file_type            => {type => SCALAR, regex => qr/(file|lun)/, default => 'file'},
            grow_size            => {type => SCALAR, default => '20k'},
            snapmirror_label     => {type => SCALAR, default => 'weekly'},
            snapshot_prefix      => {type => SCALAR},
            sleep_sec            => {type => SCALAR, default => '120', optional => 1},
        },
    );

    my $log = log_global();
    my @file_objs = @{$args{file_obj}};
    $log->trace('File objects: ' . Dumper(@file_objs));
    my @chosen_file_objs = splice (@file_objs, 0, $args{file_count});
    $log->log('Chosen files are ...');
    foreach my $file_obj ( @chosen_file_objs ) {
        $log->trace('FileObj is...'.Dumper $file_obj);
        $log->log('Path='.$file_obj->path());
    }

    my @subtests;
    $counter=0;
    foreach my $sf (@chosen_file_objs) {
        my %args_func = (file_obj             => $sf,
                         client               => $args{client},
                         grow_size            => $args{grow_size},
                         sleep_sec            => $args{sleep_sec},);
        push (@subtests, create_parallel_processes('_perform_transition', \%args_func));

    }

    run_processes_bg('data_transitions', \@subtests);

    my @vols;
    foreach my $f (@chosen_file_objs) {
        my $vol_stask_obj    =  $f->get_containing_volume_obj();
        if (! grep {$_->volume() eq $vol_stask_obj->volume()} @vols) {
            push (@vols, $vol_stask_obj);
        }
    }

    my %ret;
    foreach my $v (@vols) {
        my $ss_name = 's3';
        $log->log('Cut a ss ' . $ss_name);
        my $ss = $v->snapshot_create (
                    snapshot           => $ss_name,
                    'snapmirror-label' => $args{snapmirror_label});
        $ret{$v->volume()}  = $ss->snapshot();
    }

    return \%ret;
}


sub _perform_transition {

    my %args = validate_with(
        params      => \@_,
        allow_extra => 0,
        spec        => {
            file_obj             => { type => OBJECT, },
            client               => { type => ARRAYREF|SCALAR, optional => 1},
            grow_size            => { type => SCALAR },
            grow_size            => {type => SCALAR, default => '20k'},
            sleep_sec            => {type => SCALAR, default => '120', optional => 1},
        },
    );
    my $log = log_global();
    my $file             = $args{file_obj};
    my $client           = $args{client};
    my $grow_size        = $args{grow_size};
    my $file_path        = $file->path();
    my $vol_stask_obj    = $file->get_containing_volume_obj();
    my $cmd_interface    = $vol_stask_obj->command_interface();
    my $extent_size      = $vol_stask_obj->get_one_state_attribute('default-extent-size');

    $log->comment('extent size : ' . $extent_size);

    #NOTE:for very small sized files the punch hole is not very effective.
    #you might end punching holes at the same offsets even if you are trying
    #to punch in the middle, end and begin.
    #so its best to pass in a bunch of different sized files to really excercise this
    #code.
    $log->step('Starting DATA_HOLE transition.' . "\n" .
               'Create a new clone file and punch holes while leaving the parent file as is');
    my $clone_name =  $file_path . "_clone0";
    $log->comment('Creating a clone file ' . $clone_name . ' of the source file ' . $file_path);
    my $file_clone0 = NACL::STask::VolumeFileClone->create(
                                command_interface  => $cmd_interface,
                                'source-path'      => $file_path,
                                'destination-path' => $clone_name,
                                vserver            => $vol_stask_obj->vserver(),
                                );

    $log->comment('Punch holes in the start and the middle of the clone file');
    _punch_holes_in_files(
                        file_obj             => [$file_clone0],
                        client               => $client,
                        holes_in_begin       => 1,
                        holes_in_end         => 0,
                        holes_in_mid         => 1,
                        hole_size_in_begin   => $extent_size,
                        hole_size_in_mid     => $extent_size,
                        offset_between_holes => $extent_size);

    $log->comment('...done');
    $log->step('Starting HOLE_DATA transition.' . "\n" .
               ' Reuse clone file ' . $clone_name . '.' . "\n" .
               ' Punch holes in the parent file at offsets which contain data in the clone file' . "\n");
    $log->comment('Punch holes in the middle and end of the parent file');
    _punch_holes_in_files(
                        file_obj             => [$file],
                        client               => $client,
                        holes_in_begin       => 0,
                        holes_in_end         => 1,
                        holes_in_mid         => 1,
                        hole_size_in_end     => $extent_size,
                        hole_size_in_mid     => $extent_size,
                        offset_between_holes => $extent_size);

    $log->step('Starting DATA_NONE and HOLES_NONE transition.' . "\n" .
               ' Grow the parent file with holes and fill portion of the' . "\n" .
               ' grown area with data leaving the clone file as is');
    $clone_name =  $file_path . "_clone1";
    $log->comment('Create a new clone file ' . $clone_name . ' of the parent');
    my $file_clone1 = NACL::STask::VolumeFileClone->create(
                                command_interface  => $cmd_interface,
                                'source-path'      => $file_path,
                                'destination-path' => $clone_name,
                                vserver            => $vol_stask_obj->vserver(),
                                );

    $log->comment('Grow the parent file ' . $file_path . ' and fill it with holes and data');
    _grow_files ( file_obj     => [$file],
                  client       => $client,
                  grow_size    => $grow_size,
                  fill         => 'both',
                  sleep_sec    => $args{sleep_sec});


    $log->step('Starting NONE_DATA and NONE_HOLES transition.' . "\n" .
               ' Grow the clone file with holes and fill portion of the' . "\n" .
               'grown area with data leaving the parent file as is');
    $clone_name =  $file_path . "_clone2";
    $log->comment('Create a new clone file ' . $clone_name . ' of the parent');
    my $file_clone2 = NACL::STask::VolumeFileClone->create(
                                command_interface  => $cmd_interface,
                                'source-path'      => $file_path,
                                'destination-path' => $clone_name,
                                vserver            => $vol_stask_obj->vserver(),
                                );
    $log->comment('Grow the clone file ' . $clone_name. ' and fill it with holes and data');
    _grow_files ( file_obj     => [$file_clone2],
                  client       => $client,
                  grow_size    =>  $grow_size,
                  fill         => 'both',
                  sleep_sec    => $args{sleep_sec},);

    $log->step('Starting DATA_CHANGE transition.' . "\n" .
               ' Grow both the clone and parent files with different data');
    $clone_name =  $file_path . "_clone3";
    $log->comment('Creating a clone file ' . $clone_name . ' of the source file ' . $file_path);
    my $file_clone3 = NACL::STask::VolumeFileClone->create(
                                command_interface  => $cmd_interface,
                                'source-path'      => $file_path,
                                'destination-path' => $clone_name,
                                vserver            => $vol_stask_obj->vserver(),
                                );

    $log->step('Grow both the parent file: ' . $file_path . ' and the clone file: ' .
                $clone_name . " by $grow_size and fill same regions with different data \n");
    $log->comment('.... working on parent file: ' . $file_path);
    _grow_files ( file_obj     => [$file],
                  client       => $client,
                  grow_size    => $grow_size,
                  fill         => 'both',
                  unique_data  => 1,);

    $log->comment('.... working on clone file: ' . $clone_name);
    _grow_files ( file_obj     => [$file_clone3],
                  client       => $client,
                  grow_size    => $grow_size,
                  fill         => 'both',
                  unique_data  => 1,);

    return;
}

# limitation: if fill eq 'data' this method works on file_objs that have files of the same size

=head2 _grow_files

This method will grow a file by $grow_size and fill it with data/holes/both.
If 'both' is specified it will first grow the file by $grow_size with holes and then
fill half the grown region with data.

=over

=item Options

=over

=item file_obj
(required) The list of source files objects.

=item client
client name to use to do IO

=item grow_size
(Optional) By how much to grow the file (clones/parent) to excercise the transitions
(Default) 1m

=item fill
(Optional) Whether to fill the grown region with holes/data/both.
When 'both' is specified it will fill half the grown region with data and half with holes.
(Default) data

=item sleep_sec
(Optional) How long to run IO when we need to fill a grown region with data.
(Default) 180

=item unique_data
(Optional) Some tests require unique data to be filled between files. SIO does not have the
functionality where the user can specify a 'pattern'. Hence if this param is set then the method will use 'dd' to fill the grown region with unique data using input file=/dev/urandom.
(Defualt) 0

=item extent_size
(Optional) The extent size of the volume the file belongs to. This is important since the
extent size controls the file size when we grow the file. This is a caveat for MS3.

=back

=back

=cut

sub _grow_files {
    my %args = validate_with(
        params      => \@_,
        allow_extra => 0,
        spec        => {
            file_obj    => {type => ARRAYREF},
            client      => {type => ARRAYREF|SCALAR, optional => 1}, # not using now - plan to use later
            grow_size   => {type => SCALAR, default => '1m'},
            fill        => {type => SCALAR, default => 'data', regex => qr/^(data|hole|both)$/},
            sleep_sec   => {type => SCALAR, default => '120', optional => 1},
            unique_data => {type => SCALAR, default => 0, optional => 1},
            io_tool     => {type => SCALAR, default => 'comp' },
        },
    );
    my $log = log_global();
    $log->log("Will grow all files by $args{grow_size}");

    my @traffic_client;
    if (ref($args{client}) eq 'ARRAY') {
        @traffic_client = map( NACL::C::Client->new( name => $_ ), @{$args{client}} );
    } else {
        @traffic_client = map( NACL::C::Client->new( name => $_ ), $args{client} );
    }
    my $client_name = $traffic_client[0]->client();
    $log->log('Using client '.$client_name);
    my $client_api = $traffic_client[0]->apiset();
    $client_api->set_timeout('timeout' => 600);

    # get the vol obj
    # assuming all files are on same vol
    my $first_fileobj = @{ $args{file_obj} }[0]; # just pick first file
    my $file_path = $first_fileobj->path();
    my $elements_obj = $first_fileobj->_extract_volume_file_from_path(path => $file_path);
    my $vol_name = $elements_obj->volume();
    $log->log('using volume '.$vol_name);
    my $vol_obj = NACL::STask::Volume->find(command_interface => $first_fileobj->{command_interface},
                                            filter            => {
                                                volume  => $vol_name,
                                                vserver => $first_fileobj->{vserver},
                                            }
                                        );
    my $mount_obj = mount_nfs_vol2(volume => $vol_obj, client_name => $client_name);

    my @traffic_objs;
    my $extent_size = $vol_obj->get_one_state_attribute('default-extent-size');
    $extent_size = convert_to_bytes($extent_size);

    foreach my $file_obj (@{$args{file_obj}}) {

        my $file_state = $file_obj->state(); # issues ZAPI file-get-file-info
        my $old_file_size = $file_state->file_size();
        my $file_path = $file_obj->path();
        my $new_file_size = $old_file_size + convert_to_bytes($args{grow_size});
        $new_file_size = ceil($new_file_size/$extent_size) * $extent_size;
        $log->log("Will grow file=$file_path existing_size=$old_file_size new_size=$new_file_size grow_size=$args{grow_size}");
        $file_obj->testsetattr('file-size' => $new_file_size);

        # check if the size got updated - try via zapi file-get-file-info and testsetattr -l
        $file_state = $file_obj->state(); # refresh the state
        my $file_size = $file_state->file_size();
        $log->log("New size of file $file_path is $file_size");
        $log->warn("Current size did not match expected size. Expected:$new_file_size Got:$file_size") if ($file_size ne $new_file_size);
        my $file_attributes = $file_obj->testsetattr('get-attributes' => 1);
        $log->trace('File Attributes are '.Dumper($file_attributes));
        $file_size = $file_attributes->{size};
        $log->error("Current size did not match expected size. Expected:$new_file_size Got:$file_size") if ($file_size ne $new_file_size);

        my %size_args;
        my $grow_size;
        if ($args{fill} eq 'data') {
            $size_args{file_size} = $new_file_size,
            $size_args{start_byte} = $old_file_size;
        } elsif ($args{fill} eq 'both') {
            $grow_size = convert_to_bytes($args{grow_size})/2;
            my $end_size = $old_file_size + $grow_size;
            $end_size = ceil($end_size/$extent_size) * $extent_size;
            $size_args{file_size} = $end_size,
            $size_args{start_byte} = $old_file_size;
        }

        $log->trace('start and end byte args are: ' . Dumper(\%size_args));
        if ($args{fill} eq 'data' ||
            $args{fill} eq 'both' ) {

            $log->log('Filling the grown area with data as requested.');
            if($args{'io_tool'} !~ /sio/i){
                my $fn = $file_obj->state->name;
                $log->comment('Writting compressed dataset to file ' . $fn);
                compression_dataset (
                    operation   => 'populate',
                    client_name => $client_name,
                    volume_objs => [$vol_obj],
                    file_paths  => ["$fn"],
                    file_size   => $size_args{file_size},
                    start_offset => $old_file_size,
                );
            } else {
                my $client_path_to_file = $file_obj->get_client_path_to_file( mount_obj => $mount_obj );
                if (! $args{unique_data}) {
                   $log->log("Starting traffic on Client path to file=$client_path_to_file");
                   my $traffic_obj = NACL::MTask::Traffic->start( command_interface => $traffic_client[0],
                                                                tool              => 'sio',
                                                                paths             => $client_path_to_file,
                                                                block_size        => '4k',
                                                                read              => 0,
                                                                num_of_threads    => 1,
                                                                rand              => 0,
                                                                'privilege-level' => 'root',
                                                                verify            => '1',
                                                                break_on_dc       => '1',
                                                                rep_latency       => '60',
                                                                %size_args
                                                            );
                    push @traffic_objs, $traffic_obj;
                } else {
                    $log->log('User has requested unique data to be written. ' .
                              'Since sio supports only fixed pattern using dd instead');
                    my $block_size = 1 * 1024; #1k
                    my $count = $grow_size/$block_size; #count = amount of data to write/block size
                    my $seek_value = $old_file_size/$block_size;
                    my $cmd = "dd conv=notrunc  if=/dev/urandom of=$client_path_to_file bs=$block_size seek=$seek_value count=$count";
                    $log->log('Running command ' . $cmd . ' on the client');
                    $client_api->execute_command('connectrec-timeout' => 600,
                                                 'privilege-level'    => 'root',
                                                 command              => $cmd);
                }
            }
        }

        $file_attributes = $file_obj->testsetattr('get-attributes' => 1);
        $log->trace('File Attributes are '.Dumper($file_attributes));
    }


    # sleep for a bit and then stop the traffic
    if ($args{fill} eq 'data' ||
        $args{fill} eq 'both') {
        if (!$args{unique_data}) {
            $log->log("Sleeping for $args{sleep_sec} sec to fill up files");
            sleep($args{sleep_sec});
            foreach my $traffic_obj (@traffic_objs) {
                $traffic_obj->stop();
            }
        }
    } else {
        $log->log('Skipping filling the grown area with data as requested.');
    }

    return;
}

sub run_processes_bg{
    my($func_name,$subtest)=@_;
    my @subtests  = @{$subtest};

    if( scalar @subtests == 0 ) {
        NATE::BaseException->throw('No processes in @subtests.');
    }

    my $pm = NATE::ProcManager->new(
            runid               => 'PM_' . $func_name,
            max_active_children => 16,
            proc_info           => \@subtests,
            );
    $pm->start();
    $pm->wait();
    $pm->destroy();

}

sub create_parallel_processes{
    my($func_name,$params)=@_;
    my %params_hash = %$params;
    my @subtests;
    my $subtest = NATE::Process->new(
            codespec => \&$func_name,
            runid => sprintf("%03d_$func_name", ++$counter),
            args => [%params_hash],
            inherit_params => 1,
            onexit => \&NATE::Process::on_exit_die_worst_result,
            );
    return $subtest;
}



=head2 populate_lcp_dataset

This method is to generate compression and uncompression datasets in MW volumes.
and cut a snapshot after each data generation.

(lcp_ss1):
    dir1: Pattern of U, C and H in every 64 fbns (one WAFL_SNAP_READ)
    dir2: Completely compressed data
    dir3: Completely uncompressed data

(lcp_ss2):
    dir1: Modify and change to pattern of C and H in every 64 fbns
    dir2: Modify and change it to U and C
    dir3: Modify and change it to all C

(lcp_ss3):
    dir1: Grow the file and have U, C and H in the grown region
    dir2: Grow the file and have all C in the grown region
    dir3: Truncate the file to lesser size

(lcp_ss4):
    grow the files just got truncated in lcp_ss3

(lcp_ss5):
    create clones of the files.

(lcp_ss6):
    Modify and change to pattern of C+U+H.

=over

=item options
    file_paths - file paths for datasets,
    file_count - number of files to write, default to '1',
    file_size  - file size to write, default to '2m',
    dir_depth  - dir depth, default to '1',
    dir_count  - number of dirs to create, default to '1',
    dedupe_pct - deduplication percentage of the datasets, default to '0',
    debug_prt  - printout of wafl showvnb, default to '0',
    data_type  - 0: old create_tree behavior; 1: mix of u & c, 2: c; default to '0',
    file_fill_pct        - percentage of data fill. rest will be holes. default to '100',
    snapmirror_label     - snapmirror label, default to 'weekly',
    mod_punchhole_pct    - percantage of files to punch holes. default to '0',
    mod_punchhole_by_pct - percantage of each file to punch holes, default to '25',
    mod_overwrite_pct    - percantage of files to modify. default to '100',
    mod_overwrite_by_pct - percantage of each file to modify, default to '50',
    ss_to_create         - snapshot to create, default to all.
    block_type           - fixed block size or random block size. default to fixed.

=over

=item required

volume_objs - OBJECT, volume to hold datasets.
client_name - SCALAR, client to drive IOs.

=back

=back

=cut

sub populate_lcp_dataset {
    my %args = validate_with(
        params => \@_,
        spec => {
            volume_objs => { type => ARRAYREF },
            client_name => { type => SCALAR },
            file_paths  => { type => ARRAYREF, optional => 1 },
            file_count  => { type => SCALAR, default => '2' },
            file_size   => { type => SCALAR, default => '75m' },
            dir_depth   => { type => SCALAR, default => '1' },
            dir_count   => { type => SCALAR, default => '1' },
            dedupe_pct  => { type => SCALAR, default => '0' },
            debug_prt   => { type => SCALAR, default => '0' },
            data_type   => { type => SCALAR, default => '1' },
            block_type  => { type => SCALAR, default => 'fixed' },
            ss_to_create         => { type => ARRAYREF|UNDEF, optional => 1 },
            file_fill_pct        => { type => SCALAR, default => '100' },
            snapmirror_label     => { type => SCALAR, default => 'weekly' },
            mod_overwrite_pct    => { type => SCALAR, default => '0' },
            mod_overwrite_by_pct => { type => SCALAR, default => '75' },
            mod_punchhole_pct    => { type => SCALAR, default => '0' },
            mod_punchhole_by_pct => { type => SCALAR, default => '25' },
        },
    );

    my $volume_objs = $args{volume_objs};
    my $client_name = $args{client_name};
    my $file_paths  = $args{file_paths};
    my $file_count  = $args{file_count};
    my $file_size   = $args{file_size};
    my $dir_depth   = $args{dir_depth};
    my $dir_count   = $args{dir_count};
    my $dedupe_pct  = $args{dedupe_pct};
    my $debug_prt   = $args{debug_prt};
    my $data_type   = $args{data_type};
    my $file_fill_pct        = $args{file_fill_pct};
    my $snapmirror_label     = $args{snapmirror_label};
    my $mod_overwrite_pct    = $args{mod_overwrite_pct};
    my $mod_overwrite_by_pct = $args{mod_overwrite_by_pct};
    my $mod_punchhole_pct    = $args{mod_punchhole_pct};
    my $mod_punchhole_by_pct = $args{mod_punchhole_by_pct};

    my $log = log_global();

    my $ss_to_create = $args{ss_to_create} ||
                    ['lcp_ss1' , 'lcp_ss2' , 'lcp_ss3', 'lcp_ss4'];

    $log->log('COMP: dataset to create - ' . Dumper($ss_to_create));
    $log->log('COMP: file size changed from ' . convert_to_bytes($file_size));
    $file_size = convert_to_bytes($file_size) + 1024 * (rand(time) % 32); # random allignment
    $log->log('COMP: to ' . $file_size);

    my $ret;
    # use random block size to generate data: partial extent writes
    if($args{block_type} eq 'random'){
        $log->step('COMP: use DT with random block sizes to populate data');
        _wafl_flush_cache(node => @{$volume_objs}[0]->get_local_node_name());
        my (%rtn, @subtests);
        my $client_obj = NACL::C::Client->new( name => $client_name );
        foreach my $vol (@{$volume_objs}){
            my $mount_obj = mount_nfs_vol2(volume => $vol, client_name => $client_name);
            my $vname = $vol->volume;
            my $mkfilepath = "/vol/$vname/$vname" . "_mwfile";
            $log->log('...create dt files in volume ' . $vname);
            my @f_objs = NACL::STask::VolumeFile->create(
                            command_interface => $vol->command_interface,
                            vserver           => $vol->vserver,
                            size              => $file_size,
                            path              => $mkfilepath,
                            filecount         => $file_count,
            );
            my @dt_traffic_objs;
            push(
                @subtests,
                NATE::Process->new(
                    codespec => sub {
                        foreach my $fobj (@f_objs){
                            my $client_path_to_file = $fobj->get_client_path_to_file(mount_obj => $mount_obj);
                            try {
                                my $dt_traffic_obj = NACL::MTask::Traffic->start(
                                    command_interface => $client_obj,
                                    paths             => $client_path_to_file,
                                    tool              => 'dt',
                                    iotype            => 'random',
                                    limit             => $fobj->state->file_size,
                                    bs                => 'random',
                                    incr              => 'variable',
                                    min               => '1k',
                                    max               => '256k',
                                    pattern           => 'iot',
                                    runtime           => '-1',
                                    dispose           => 'keep',
                                    prefix            => '"%u@%h"',
                                    disable           => 'pstats',
                                    flags             => 'direct',
                                    'privilege-level' => 'root',
                                );
                                push @dt_traffic_objs, $dt_traffic_obj;
                            } catch NATE::BaseException with {
                                my $exception = shift;
                                $log->error('Unexpected: Failed to start DT!!! ' . $exception->text());
                            };
                        }
                        foreach my $ss_name (@{$ss_to_create}){
                            $log->log('COMP: ...sleep 60s');
                            Tharn::snooze(60);

                            $log->log('...take a snapshot ' . $ss_name . ' on volume ' . $vol->volume());
                            NACL::STask::VolumeSnapshot->create(
                                command_interface  => $vol->command_interface,
                                volume             => $vol,
                                vserver            => $vol->vserver,
                                snapshot           => $ss_name,
                                'snapmirror-label' => $snapmirror_label,
                            );
                        }
                        foreach my $dt_traffic_obj (@dt_traffic_objs){
                            $dt_traffic_obj->stop();
                        }
                    },
                    runid      => $vname,
                    background => 1,
                    onexit     => \&NATE::Process::on_exit_propagate_worst_result,
                )
            );
            foreach my $fobj (@f_objs){
                push @{$rtn{$vname}{'file_names'}}, $fobj->state->name;
            }
            $rtn{$vname}{'file_obj'} = \@f_objs;
        }
        if (@subtests) {
            my $pm = NATE::ProcManager->new(
                proc_info => \@subtests,
            );
            $pm->start();
            $pm->wait();
            $pm->destroy();
        }
        $ret = \%rtn;
    } else { # use fixed block size to generate dataset
        $log->step('COMP : use create_tree to create specific datasets for LCP');
        # create 3 dirs to hold 3 type of files: comp, uncomp, and mixed.
        my (%fps, $rtn);
        foreach my $vol (@{$volume_objs}){
            my @temp_fps;
            my $vname = $vol->volume();
            $log->log('COMP: Create 3 dirs in volume ' . $vname);
            for my $i (0...2) {
                my $mkfile_path = "/vol/$vname/mw_dir$i";
                NACL::C::VolumeDirectory->create(
                    command_interface => $vol->command_interface(),
                    path              => $mkfile_path,
                    vserver           => $vol->vserver,
                    perm              => '777'
                );
                push @temp_fps, "mw_dir$i";
            }
            $fps{$vname}{dir_paths} = \@temp_fps;
        }

        my $uncomp_dir = shift @{$fps{@{$volume_objs}[0]->volume}{dir_paths}};
        my $comp_dir   = shift @{$fps{@{$volume_objs}[0]->volume}{dir_paths}};
        my $mix_dir    = $fps{@{$volume_objs}[0]->volume}{dir_paths};

        $log->log('COMP: uncomp_dir: ' . $uncomp_dir .  "\n" .
                  '        comp_dir: ' . $comp_dir . "\n" .
                  '         mix_dir: ' . Dumper($mix_dir) . "\n");

        my %opts = (
            client_name => $client_name,
            volume_objs => \@{$volume_objs},
            file_count  => $file_count,
            file_size   => $file_size,
            dir_depth   => $dir_depth,
            dir_count   => $dir_count,
            debug_prt   => $debug_prt,
        );

        # always create base dataset
        $log->step('COMP: create lcp_ss1');
        my $ss_name = 'lcp_ss1';
        $log->step("COMP: creating uncompressed files in dir $uncomp_dir");
        $rtn = compression_dataset (
            operation       => 'populate',
            file_paths      => [$uncomp_dir],
            compression_pct => 0, # for uncompressd dataset
            file_fill_pct   => 100,
            %opts,
        );
        $ret = $rtn;

        $log->step("COMP: creating compressed files in dir $comp_dir");
        $rtn = compression_dataset (
            operation     => 'populate',
            file_paths    => [$comp_dir],
            data_type     => 2,  # for compressed dataset
            file_fill_pct => 100,
            %opts,
        );
        foreach my $vol (@{$volume_objs}){
            my $vname = $vol->volume();
            push @{$ret->{$vname}->{'file_names'}}, @{$rtn->{$vname}->{'file_names'}};
            push @{$ret->{$vname}->{'file_obj'}}, @{$rtn->{$vname}->{'file_obj'}};
        }

        $log->step("COMP: creating uncomp+comp+hole files in dir " . Dumper($mix_dir));
        $rtn = compression_dataset (
            operation     => 'populate',
            file_paths    => $mix_dir,
            data_type     => 1, # for uncompressed & compressed dataset
            snapshot      => $ss_name,
            file_fill_pct => 75, # <100, holes
            %opts,
        );
        foreach my $vol (@{$volume_objs}){
            my $vname = $vol->volume();
            push @{$ret->{$vname}->{'file_names'}}, @{$rtn->{$vname}->{'file_names'}};
            push @{$ret->{$vname}->{'file_obj'}}, @{$rtn->{$vname}->{'file_obj'}};
        }

        # modify the base dataset
        if (grep {$_ eq 'lcp_ss2'} @{$ss_to_create}){
            $log->step('COMP: create lcp_ss2');
            $ss_name = 'lcp_ss2';
            $log->step("COMP: modifying uncompressed files to compressed in dir $uncomp_dir");
            push my @subtests_hdl, subtest(
                sub {
                    compression_dataset (
                        operation            => 'modify',
                        file_paths           => [$uncomp_dir],
                        data_type            => 2,
                        mod_overwrite_pct    => 100, # all files
                        mod_overwrite_by_pct => 100, # whole file
                        %opts,
                    );
                },  '-bg', '-runid', 'lcp_ss2_ucomp_to_comp'
            );
            Subtest::wait_finish(subtests => \@subtests_hdl);
            $log->step("COMP: modifying compressed files to u+c in dir $comp_dir");
            push @subtests_hdl, subtest(
                sub {
                    compression_dataset (
                        operation            => 'modify',
                        file_paths           => [$comp_dir],
                        data_type            => 1,
                        mod_overwrite_pct    => 100,
                        mod_overwrite_by_pct => (50 + rand(time) % 30),
                        %opts,
                    );
                },  '-bg', '-runid', 'lcp_ss2_comp_to_ucomp&comp'
            );
            Subtest::wait_finish(subtests => \@subtests_hdl);
            $log->step("COMP: modifying u+c+h files to c+h in dir " . Dumper($mix_dir));
            push @subtests_hdl, subtest(
                sub {
                    $rtn = compression_dataset (
                        operation            => 'modify',
                        file_paths           => $mix_dir,
                        data_type            => 2,
                        mod_overwrite_pct    => 100,
                        mod_overwrite_by_pct => (50 + rand(time) % 30),
                        %opts,
                    );
                    foreach my $vol (@{$volume_objs}){
                        $log->log('COMP: punching holes for files in volume ' . $vol->volume);
                        my $vname = $vol->volume();
                        my $extent_size = $vol->get_one_state_attribute('default-extent-size');
                        foreach my $file_obj (@{$rtn->{$vname}->{'file_obj'}}){
                            my $fsize = $file_obj->state->file_size;
                            my $block_cnts = floor(($fsize/convert_to_bytes($extent_size))/3);
                            $log->comment('COMP: punch holes on volume ' . $vname . ': file - ' .
                                          $file_obj->state->name . ', size - ' .
                                          $fsize . ', cnt - ' . $block_cnts);
                            _punch_holes_in_files(
                                file_obj             => [$file_obj],
                                client               => $client_name,
                                holes_in_begin       => 1,
                                holes_in_end         => 1,
                                holes_in_mid         => $block_cnts,
                                hole_size_in_begin   => $extent_size,
                                hole_size_in_mid     => $extent_size,
                                hole_size_in_end     => $extent_size,
                                offset_between_holes => $extent_size, );
                        }
                        NACL::STask::VolumeSnapshot->create(
                            command_interface  => $vol->command_interface,
                            volume             => $vol,
                            vserver            => $vol->vserver,
                            snapshot           => $ss_name,
                            'snapmirror-label' => $snapmirror_label,
                        );
                    }
                },  '-bg', '-runid', 'lcp_ss2_u&c&h_to_comp&holes'
            );
            Subtest::wait_finish(subtests => \@subtests_hdl);
        }

        # modify dataset with u+c
        if (grep {$_ eq 'lcp_ss2_1'} @{$ss_to_create}){
            $log->step('COMP: create lcp_ss2_1');
            $ss_name = 'lcp_ss2_1';
            $log->log("COMP: modifying files in volumes");
            push my @subtests_hdl, subtest(
                sub {
                    compression_dataset (
                        operation            => 'modify',
                        snapshot             => $ss_name,
                        data_type            => 1,
                        mod_overwrite_pct    => 100,
                        mod_overwrite_by_pct => 25,
                        %opts,
                    );
                },  '-bg', '-runid', 'lcp_ss2_1_u&c'
            );
            Subtest::wait_finish(subtests => \@subtests_hdl);
        }

        # modify dataset with u+c
        if (grep {$_ eq 'lcp_ss2_2'} @{$ss_to_create}){
            $log->step('COMP: create lcp_ss2_2');
            $ss_name = 'lcp_ss2_2';
            $log->log("COMP: modifying files in volumes");
            push my @subtests_hdl, subtest(
                sub {
                    compression_dataset (
                        operation            => 'modify',
                        snapshot             => $ss_name,
                        data_type            => 1,
                        mod_overwrite_pct    => 100,
                        mod_overwrite_by_pct => 50,
                        %opts,
                    );
                },  '-bg', '-runid', 'lcp_ss2_2_u&c'
            );
            Subtest::wait_finish(subtests => \@subtests_hdl);
        }

        # modify dataset with u+c
        if (grep {$_ eq 'lcp_ss2_3'} @{$ss_to_create}){
            $log->step('COMP: create lcp_ss2_3');
            $ss_name = 'lcp_ss2_3';
            $log->log("COMP: modifying files in volumes");
            push my @subtests_hdl, subtest(
                sub {
                    compression_dataset (
                        operation            => 'modify',
                        snapshot             => $ss_name,
                        data_type            => 1,
                        mod_overwrite_pct    => 100,
                        mod_overwrite_by_pct => 75,
                        %opts,
                    );
                },  '-bg', '-runid', 'lcp_ss2_3_u&c'
            );
            Subtest::wait_finish(subtests => \@subtests_hdl);
        }

        # modify the dataset : grow and truncate
        if (grep {$_ eq 'lcp_ss3'} @{$ss_to_create}){
            $log->step('COMP: create lcp_ss3');
            $ss_name = 'lcp_ss3';
            $log->step("COMP: truncating files by ~25% in dir $uncomp_dir");
            push my @subtests_hdl, subtest(
                sub {
                    my $rtn = compression_dataset (
                        operation         => 'modify',
                        file_paths        => [$uncomp_dir],
                        mod_shrink_pct    => 100,
                        mod_shrink_by_pct => (10 + rand(time) % 20),
                        %opts,
                    );
                    foreach my $vol (@{$volume_objs}){
                        my $vname = $vol->volume();
                        foreach my $file_obj (@{$rtn->{$vname}->{'file_obj'}}){
                            my $fsize = $file_obj->state->file_size;
                            $log->comment('COMP: volume ' . $vname . ': file - ' .
                                        $file_obj->state->name . ', size is now - ' .
                                        $fsize . ', used to be - ' . convert_to_bytes($file_size));
                        }
                    }
                },  '-bg', '-runid', 'lcp_ss3_truncate'
            );
            Subtest::wait_finish(subtests => \@subtests_hdl);
            $log->step("COMP: growing files by ~25% and fill compressed in grow region in dir $comp_dir");
            push @subtests_hdl, subtest(
                sub {
                    my $rtn = compression_dataset (
                        operation       => 'modify',
                        file_paths      => [$comp_dir],
                        data_type       => 2,
                        mod_grow_pct    => 100,
                        mod_grow_by_pct => (10 + rand(time) % 20),
                        %opts,
                    );
                    foreach my $vol (@{$volume_objs}){
                        my $vname = $vol->volume();
                        foreach my $file_obj (@{$rtn->{$vname}->{'file_obj'}}){
                            my $fsize = $file_obj->state->file_size;
                            $log->comment('COMP: volume ' . $vname . ': file - ' .
                                        $file_obj->state->name . ', size is now - ' .
                                        $fsize . ', used to be - ' . convert_to_bytes($file_size));
                        }
                    }
                },  '-bg', '-runid', 'lcp_ss3_grow_w_comp'
            );
            Subtest::wait_finish(subtests => \@subtests_hdl);
            $log->step("COMP: growing files by 25% and fill u+c+h in grow region in dir " . Dumper($mix_dir));
            push @subtests_hdl, subtest(
                sub {
                    $rtn = compression_dataset (
                        operation            => 'modify',
                        file_paths           => $mix_dir,
                        data_type            => 1,
                        mod_grow_pct         => 100,
                        mod_grow_by_pct      => (10 + rand(time) % 20),
                        %opts,
                    );
                    foreach my $vol (@{$volume_objs}){
                        my $vname = $vol->volume();
                        my $extent_size = $vol->get_one_state_attribute('default-extent-size');
                        foreach my $file_obj (@{$rtn->{$vname}->{'file_obj'}}){
                            my $fsize = $file_obj->state->file_size;
                            my $block_cnts = floor(($fsize/convert_to_bytes($extent_size))/3);
                            $log->comment('Punch holes on volume ' . $vname . ': file - ' .
                                          $file_obj->state->name . ', size - ' .
                                          $fsize . ', cnt - ' . $block_cnts);
                            _punch_holes_in_files(
                                file_obj             => [$file_obj],
                                client               => $client_name,
                                holes_in_begin       => 1,
                                holes_in_end         => 1,
                                holes_in_mid         => $block_cnts,
                                hole_size_in_begin   => $extent_size,
                                hole_size_in_mid     => $extent_size,
                                hole_size_in_end     => $extent_size,
                                offset_between_holes => $extent_size, );
                        }
                        NACL::STask::VolumeSnapshot->create(
                            command_interface  => $vol->command_interface,
                            volume             => $vol,
                            vserver            => $vol->vserver,
                            snapshot           => $ss_name,
                            'snapmirror-label' => $snapmirror_label,
                        );
                    }
                },  '-bg', '-runid', 'lcp_ss3_grow_w_holes'
            );
            Subtest::wait_finish(subtests => \@subtests_hdl);
        }

        # modify the dataset
        if (grep {$_ eq 'lcp_ss4'} @{$ss_to_create}){
            $log->step('COMP: create lcp_ss4');
            $ss_name = 'lcp_ss4';
            $log->step('COMP: grow the files just got truncated in dir $uncomp_dir');
            push my @subtests_hdl, subtest(
                sub {
                    my $rtn = compression_dataset (
                        operation       => 'modify',
                        file_paths      => [$uncomp_dir],
                        data_type       => 2,
                        mod_grow_pct    => 100,
                        mod_grow_by_pct => (10 + rand(time) % 20),
                        %opts,
                    );
                    foreach my $vol (@{$volume_objs}){
                        my $vname = $vol->volume();
                        foreach my $file_obj (@{$rtn->{$vname}->{'file_obj'}}){
                            my $fsize = $file_obj->state->file_size;
                            $log->comment('COMP: volume ' . $vname . ': file - ' .
                                        $file_obj->state->name . ', size is now - ' .
                                        $fsize . ', used to be - ' . convert_to_bytes($file_size));
                        }
                    }
                },  '-bg', '-runid', 'lcp_ss4_grow_the_truncated'
            );
            Subtest::wait_finish(subtests => \@subtests_hdl);
            $log->step('COMP: truncate the files just got grown in dir $comp_dir');
            push @subtests_hdl, subtest(
                sub {
                    compression_dataset (
                        operation  => 'modify',
                        file_paths => [$comp_dir],
                        mod_shrink_pct    => 100,
                        mod_shrink_by_pct => (5 + rand(time) % 10),
                        %opts,
                    );
                },  '-bg', '-runid', 'lcp_ss4_truncate_the_grown'
            );
            Subtest::wait_finish(subtests => \@subtests_hdl);
            $log->step('COMP: modify the files just got grown in dir' . Dumper($mix_dir));
            push @subtests_hdl, subtest(
                sub {
                    my $rtn = compression_dataset (
                        operation         => 'modify',
                        file_paths        => $mix_dir,
                        snapshot          => $ss_name,
                        mod_overwrite_pct    => 100,
                        mod_overwrite_by_pct => 75,
                        %opts,
                    );
                },  '-bg', '-runid', 'lcp_ss4_modify_the_grown'
            );
        }

        # create clones of the files
        if (grep {$_ eq 'lcp_ss5'} @{$ss_to_create}){
            $log->step('COMP: create lcp_ss5');
            $ss_name = 'lcp_ss5';
            push my @subtests_hdl, subtest(
                sub {
                    foreach my $vol (@{$volume_objs}){
                        my $vname = $vol->volume();
                        foreach my $fobj (@{$ret->{$vname}->{'file_obj'}}){
                            $log->log('COMP: create clone of file ' . $fobj->path());
                            my $clone_path = $fobj->path() . '_mwclone';
                            NACL::STask::VolumeFileClone->create(
                                command_interface   => $vol->command_interface,
                                vserver             => $vol->vserver(),
                                'source-path'       => $fobj->path(),
                                'destination-path'  => $clone_path);
                        }
                        $log->log('COMP: cut a snapshot lcp_ss5');
                        NACL::STask::VolumeSnapshot->create(
                            command_interface  => $vol->command_interface,
                            volume             => $vol,
                            vserver            => $vol->vserver,
                            snapshot           => $ss_name,
                            'snapmirror-label' => $snapmirror_label,
                        );
                    }
                },  '-bg', '-runid', 'lcp_ss5_clones'
            );
            Subtest::wait_finish(subtests => \@subtests_hdl);
        }

        # modify dataset with u+c+h
        if (grep {$_ eq 'lcp_ss6'} @{$ss_to_create}){
            $log->step('COMP: create lcp_ss6');
            foreach my $vol (@{$volume_objs}){
                my $vname = $vol->volume();
                my $extent_size = $vol->get_one_state_attribute('default-extent-size');
                push my @subtests_hdl, subtest(
                    sub {
                        foreach my $file_obj (@{$ret->{$vname}->{'file_obj'}}){
                            my $fsize = $file_obj->state->file_size;
                            my $block_cnts = floor(($fsize/convert_to_bytes($extent_size))/3);
                            $log->comment('COMP: punch holes on volume ' . $vname . ': file - ' .
                                          $file_obj->state->name . ', size - ' .
                                          $fsize . ', cnt - ' . $block_cnts);
                            _punch_holes_in_files(
                                file_obj             => [$file_obj],
                                client               => $client_name,
                                holes_in_begin       => 0,
                                holes_in_end         => 0,
                                holes_in_mid         => $block_cnts,
                                hole_size_in_mid     => $extent_size,
                                offset_between_holes => $extent_size, );
                        }
                    },  '-bg', '-runid', 'lcp_ss6_punchholes_all_' . $vol->volume()
                );
                Subtest::wait_finish(subtests => \@subtests_hdl);
            }
            $ss_name = 'lcp_ss6';
            $log->log("COMP: modifying files in volumes");
            push my @subtests_hdl, subtest(
                sub {
                    compression_dataset (
                        operation            => 'modify',
                        snapshot             => $ss_name,
                        data_type            => 1,
                        mod_overwrite_pct    => 100,
                        mod_overwrite_by_pct => 25,
                        %opts,
                    );
                },  '-bg', '-runid', 'lcp_ss6_u&c&h'
            );
            Subtest::wait_finish(subtests => \@subtests_hdl);
        }
    }

    return $ret;
}

=head2 modify_files2dir_vc

This method will modify all files under a given path to a directory and
vice versa.

=over

=item options
    volume_objs       - The list of vol objects you want to create the dataset
    client_name       - The client on which the volume is mounted
    dir_to_file_cnt   - The number of directories to convert to files
    file_to_dir_cnt   - The number of files to convert to directories.
    cut_a_ss          - cut a snapshot for each volume
    snapshot          - snapshot name
    snapmirror_label  - snapmirror label to apply to the snapshot
    mountpts          - a hash of mount points for each volume. Structure is:
                        $VAR = {
                            'vol1' => ['mt1', 'mt2', 'mt3'],
                            'vol2' => ['mt1']
                          }

=over

=item required

volume_objs - ARRAYREF of vol objects, volumes to hold datasets.
client_name - SCALAR, client to drive IOs.
mountpts    - HASHREF, the hash of arrays represenation of the mount pts for the vol.

=back

=back

=cut

sub modify_files2dir_vc {
    my %args = validate_with(
        params => \@_,
        spec => {
            volume_objs      => { type => ARRAYREF },
            client_name      => { type => SCALAR },
            dir_to_file_cnt  => {type => SCALAR|UNDEF, optional => 1},
            file_to_dir_cnt  => {type => SCALAR|UNDEF, optional => 1},
            cut_a_ss         => {type => SCALAR|UNDEF, optional => 1, default => 0},
            snapshot         => {type => SCALAR|UNDEF, optional => 'modify_files_2_dir_ss',},
            snapmirror_label => { type => SCALAR, default => 'weekly' },
            mountpts         => { type => HASHREF}});

    my @vols  = @{ $args{volume_objs} };
    my $cname = $args{client_name};
    my $mtpts = $args{mountpts};
    my $dtof  = $args{dir_to_file_cnt};
    my $ftod  = $args{file_to_dir_cnt};
    my $log   = log_global();
    my $clobj = NACL::C::Client->new(name => $cname);
    my $api   = $clobj->apiset;
    my $snapshot          = $args{snapshot};
    my $snapmirror_label  = $args{snapmirror_label};
    my %ret;


    foreach my $v (@vols) {
        my @pts = @{$mtpts->{$v->volume}};
        foreach my $pt (@pts) {
            $log->comment('Run find on mount pt: ' . $pt);
            my $op = $api->ls (paths => $pt);
            my $resp = $op->get_processed_output();
            my @resp  = split ('\s+', $resp);
            my $dcnt = 0;
            my $fcnt = 0;

            foreach my $it (@resp) {
                if (($fcnt >= $ftod) &&
                    ($dcnt >= $dtof)) {
                    $log->comment('...done..existing loop');
                    last;
                }

                $it = $pt . "/" . $it ;
                $log->comment("processing $it");
                $log->comment('Determine if its a directory or a file');
                my $op = $clobj->execute (command => "file $it");
                if ($op =~ /.:\s+directory.*/) {
                    #Enter the loop only if the user has requested conversion
                    #of directories to files. And we have not already done
                    #the required number of conversion

                    if (defined $dtof && $dcnt <= $dtof) {
                        $log->comment('Delete the dir');
                        $api->rm (force             =>  1,
                                  recursive         =>  1,
                                  'privilege-level' => 'root',
                                  paths             =>  $it);
                        $log->comment('Recreate it as a file');
                        $clobj->execute (command => "sudo touch $it");
                        $dcnt++;
                    }
                } else {
                    #Enter the loop only if the user has requested conversion
                    #of files to directories. And we have not already done
                    #the required number of conversion
                    if (defined $fcnt && $fcnt <= $ftod) {
                        $log->comment('Delete the file');
                        $api->rm (force             =>  1,
                                  recursive         =>  1,
                                  'privilege-level' => 'root',
                                   paths            =>  $it);
                        $log->comment('Recreate it as a directory');
                        $api->mkdir(paths             => $it,
                                    'privilege-level' => 'root',
                        );
                        $fcnt++;
                    }
                }
                $log->comment('Done processing item: ' . $it);
            }
        }
        $log->comment('...done processing vol: ' . $v->volume());
        if($args{cut_a_ss}) {
            $log->comment('Cutting a ss: ' . $snapshot);
            NACL::STask::VolumeSnapshot->create(
                    command_interface       =>   $v->command_interface(),
                    volume                  =>   $v,
                    vserver                 =>   $v->vserver(),
                    snapshot                =>   $snapshot,
                    'snapmirror-label'      =>   $snapmirror_label,);
            push @{$ret{$v->volume()}{'snapshots'}}, $snapshot;
        }
    }

    $log->exit();
    return \%ret;
}

=head2 movedirs

This method will move one dir under another directory path.
It will scan the mount point and create directory pairs like below:

$VAR = {
        'dir1' => 'dir2',
        'dir3' => 'dir4',
        'dir5' => undef,
      }

Based on the above mapping it will move for example; dir1 under dir2 ignorning
directory entries that have no mapping.

=over

=item options
    volume_objs         - The list of vol objects you want to create the dataset
    client_name         - The client on which the volume is mounted.
    cut_a_ss            - cut a snapshot for each volume
    snapshot            - snapshot name
    snapmirror_label    - snapmirror label to apply to the snapshot
    mountpts            - a hash of mount points for each volume. Structure is:
                          $VAR = {
                              'vol1' => ['mt1', 'mt2', 'mt3'],
                              'vol2' => ['mt1']
                            }

=over

=item required

volume_objs   - ARRAYREF of vol objects, volumes to hold datasets.
client_name   - SCALAR, client to drive IOs.
mountpts      - HASHREF, the hash of arrays represenation of the mount pts for the vol.

=back

=back

=cut

sub movedirs {
    my %args = validate_with(
        params => \@_,
        spec => {
            volume_objs      => { type => ARRAYREF },
            client_name      => { type => SCALAR },
            cut_a_ss         => {type => SCALAR|UNDEF, optional => 1, default => 0},
            snapshot         => {type => SCALAR|UNDEF, optional => 'move_dirs_ss',},
            snapmirror_label => { type => SCALAR, default => 'weekly' },
            mountpts         => { type => HASHREF}});

    my @vols  = @{ $args{volume_objs} };
    my $cname = $args{client_name};
    my $mtpts = $args{mountpts};
    my $log   = log_global();
    my $clobj = NACL::C::Client->new(name => $cname);
    my $api   = $clobj->apiset;
    my $snapshot = $args{snapshot};
    my $snapmirror_label     = $args{snapmirror_label};
    my %ret;

    foreach my $v (@vols) {
        my @pts = @{$mtpts->{$v->volume}};
        foreach my $pt (@pts) {
            $log->comment('Run ls on mount pt: ' . $pt);
            my $op = $api->ls (paths => $pt);
            my $resp = $op->get_processed_output();
            my @resp  = split ('\s+', $resp);
            my $first_dir;
            my $second_dir;
            my $gotpair = 0;
            my %pairs;

            foreach my $it (@resp) {
                $it = $pt . "/" . $it ;
                $log->comment("processing $it");
                $log->comment('Determine if its a directory or a file');
                my $op = $clobj->execute (command => "file $it");
                #if a pair has been obtained .. reset the counter and search
                #for the next pair
                if ($gotpair) {
                    $gotpair = 0;
                    $first_dir = undef;
                }

                #set firstdir value.
                if ($op =~ /.:\s+directory.*/) {
                   if (!defined $first_dir) {
                        $first_dir = $it;
                        $pairs{$first_dir} = undef;
                   } else {
                        #assign the dir as the value
                        $pairs{$first_dir} = $it;
                        $gotpair = 1;
                   }
                }
            }

            #no dumping of data structures.
            $log->comment('Pairs obtained are: ' . Dumper(\%pairs));
            foreach my $k (keys %pairs) {
                $log->comment('Processing ' . $k);
                if (!defined $pairs{$k}) {
                    $log->comment('No destination directory found .. ignore');
                    next();
                } else {
                    my $fdir  = $k;
                    my $sdir  = $pairs{$k};
                    $log->comment("Move dir $fdir under $sdir");
                    $api->mv( source      => $fdir,
                              target      => $sdir,
                              'privilege-level'    => 'root',
                              interactive => 1,);
                }
            }
            $log->comment('Run ls post move on: ' . $pt);
            $op = $api->ls (
                                "long-list" => 1,
                                "recursive" => 1,
                                paths => $pt);

        }
        $log->comment('...done processing vol: ' . $v->volume());
        if($args{cut_a_ss}){
            $log->comment('Cutting a ss: ' . $snapshot);
            NACL::STask::VolumeSnapshot->create(
                    command_interface       =>   $v->command_interface(),
                    volume                  =>   $v,
                    vserver                 =>   $v->vserver(),
                    snapshot                =>   $snapshot,
                    'snapmirror-label'      =>   $snapmirror_label,);
            push @{$ret{$v->volume()}{'snapshots'}}, $snapshot;
        }
    }

    $log->exit();
    return \%ret;
}

=head2 populate_modify_symlinks

This method allows you to perform following operations give a mount point.

- populate: create X number of symlinks for Y files. If the method cannot find
            Y files under the mount point. It will create the remaining. It
            will use the 'symlinks' param to determine the value of X.
- modify_delete: Delete X number of symlinks. It will use the 'symlinks' param to
                 determin how many to delete.
- modify_data: Modify data for X symlinks. X = value of 'symlinks' param. The
               modification is done on the parent file.

=over

=item options


    volume_objs       - The list of vol objects you want to create the dataset
    client_name       - The client on which the volume is mounted.
    cut_a_ss          - cut a snapshot for each volume
    snapshot          - snapshot name
    snapmirror_label  - snapmirror label to apply to the snapshot
    filecount         - the number of files for which to create symlinks. If the expected
                        number are not found the method will create the remaining files.
                        Default 10.
    symlinks          - The number of symlinks to create per file. Default 2.
    symlinks_as_files - While gathering files to create symlinks of should we
                        consider other symlinks. Default no.
    operation         - What type of operation would you like to perform.
                        Default: populate
    mountpts          - a hash of mount points for each volume. Structure is:
                        $VAR = {
                            'vol1' => ['mt1', 'mt2', 'mt3'],
                            'vol2' => ['mt1']
                          }

=over

=item required

volume_objs - ARRAYREF of vol objects, volumes to hold datasets.
client_name - SCALAR, client to drive IOs.
mountpts    - HASHREF, the hash of arrays represenation of the mount pts for the vol.

=back

=back

=cut

sub populate_modify_symlinks {
    my %args = validate_with(
        params => \@_,
        spec => {
            volume_objs      => { type => ARRAYREF },
            client_name      => { type => SCALAR },
            cut_a_ss         => { type => SCALAR|UNDEF, optional => 1, default => 0},
            snapshot         => { type => SCALAR|UNDEF, optional => 1,},
            mountpts         => { type => HASHREF|UNDEF},
            filecount        => { type => SCALAR|UNDEF, optional => 1, default => 100 },
            symlinks         => { type => SCALAR|UNDEF, optional => 1, default => 2},
            symlinks_as_files => { type => SCALAR|UNDEF, optional => 1, default => 0 },
            snapmirror_label => { type => SCALAR|UNDEF, default => 'weekly' },
            operation        => { type => SCALAR|UNDEF,  regex => qr/^(populate|modify_delete|modify_data)$/, default => 'populate'},
            name_prefix      => { type => SCALAR|UNDEF, optional => 1},
            });


    my @vols  = @{ $args{volume_objs} };
    my $cname = $args{client_name};
    my $mtpts = $args{mountpts};
    my $log   = log_global();
    my $clobj = NACL::C::Client->new(name => $cname);
    my $api   = $clobj->apiset;
    my $snapshot = $args{snapshot};
    my $filecnt  = $args{filecount};
    my $symlinks = $args{symlinks};
    my $operation = $args{operation};
    my $symllink_pat = 'mwsymllink';
    my $snapmirror_label     = $args{snapmirror_label};


    if ($operation =~ /populate/) {
        foreach my $v (@vols) {
            $log->comment("Collect $filecnt file(s) from mount point to create symlinks of");
            my @pts = @{$mtpts->{$v->volume}};
            foreach my $pt (@pts) {
                $log->comment('Run find on mount pt: ' . $pt);
                my @files;
                my $fcnt = 0;
                my $pending = 1;
                my $op = $api->find (path => $pt,
                                     name => "'*'");
                my $resp = $op->get_processed_output();
                my @resp  = split ('\s+', $resp);
                foreach my $it (@resp) {
                    if ($fcnt == $filecnt) {
                        $log->comment('Received the number of files required');
                        $pending = 0;
                        last;
                    }
                    if ($it =~ /\.snapshot/) {
                        $log->comment('... ignoring');
                        next;
                    }
                    $log->comment('Determine if ' . $it . ' is a directory or a file');
                    my $op = $clobj->execute (command => "file $it");
                    print "op is: $op \n";
                    if ($op =~ /.*:\s+symbolic link/ &&
                        !$args{symlinks_as_files} ) {
                        $log->comment('Skipping symbolic links');
                        next;
                    }
                    if ($op !~ /.:\s+directory.*/) {
                        push (@files, $it);
                        $fcnt++;
                    }
                }
                #if we have not been able to gather the required number of
                #files. Calcualte the diff and create the remaining files and add
                #some data to the files.
                if ($pending){
                   my $pcnt = $filecnt - scalar(@files) ;
                   $log->comment('Creating pending files of count: ' . $pcnt .
                                 ' with prefix: ' . $pt . "/pendingfiles_");

                   foreach (my $i = 0; $i < $pcnt; $i++) {
                        my $file = $pt . '/pendingfiles_' . $i . '_' . time();
                        my $size  = int(rand (1024)) * 512;
                        my $bytesrate = int (rand(32)) + 10;
                        my $count = int($size/$bytesrate);
                        $log->comment('... file : ' . $file .
                                      ' of size: ' . $size . 'KB' .
                                      ' with bytes rate: ' . $bytesrate .
                                      ' with count: ' . $count);
                        $api->touch ('path' => $file,
                                     'connectrec-timeout' => 120);
                        $api->dd ( "input-file"     => '/dev/urandom',
                                   "output-file"    =>  $file,
                                   "bytes-rate"     =>  $bytesrate,
                                   "block-count"    =>  $count,
                                   'privilege-level'=> 'root',
                                   'connectrec-timeout' => 1800
                                   );
                        push (@files, $file);
                   }
                }

                #create the symlinks
                foreach my $f (@files) {
                    $log->comment('Creating ' . $symlinks . ' sym links for file : ' . $f);
                    my $lcnt = 0;
                    $f =~ m/(^.*\/)(.*$)/;
                    my $path = $1;
                    while ($lcnt < $symlinks) {
                        my $linkname;
                        if ($args{name_prefix}) {
                            $linkname = $path . $args{name_prefix} . $lcnt . 0;
                        } else {
                            $linkname =  $f . '_' . $symllink_pat . '_'  . $lcnt . '_' . time();
                        }
                        $log->comment('...symlink: ' . $linkname);
                        $api->ln (files                =>  $f,
                                  "link-name"          => $linkname,
                                  "symlink"            => 1,
                                  'connectrec-timeout' => 1800,
                                  'privilege-level'    => 'root',);

                        $lcnt++;
                    }
                }
            }
        }
    } elsif ($operation =~ /modify/) {
        foreach my $v (@vols) {
            my @pts = @{$mtpts->{$v->volume}};
            foreach my $pt (@pts) {
                my $cmd = "ls -lR $pt | grep ^l";
                $log->comment('Run ' . $cmd . ' on the client');
                my $op = $clobj->execute (command => $cmd);

                my @lines = split /\n/, $op;
                my (@link_paths, @symlink_e, @excercised_links);

                foreach my $l (@lines) {
                    $l =~ /.*\s(.*)\s->\s(.*)/;
                    push (@link_paths, $2);
                    push (@symlink_e, $1);
                }

                #doing this here because there could be multiple
                #symlinks pointing to the same file. So we want to gather only
                #the unique files.
                if ($symlinks > $#link_paths) {
                    $symlinks = $#link_paths;
                }

                if ($operation =~ /modify_data/) {
                    $log->comment('Modify symllinks');
                    splice (@link_paths, $symlinks, $#link_paths);

                    foreach my $sl (@link_paths) {
                        my $size  = int(rand (1024)) * 512;
                        my $bytesrate = int (rand(32)) + 10; #this will avoid 0's
                        my $count = int($size/$bytesrate);
                        $bytesrate .= 'KB';
                        $log->comment('... file : ' . $sl.
                                          ' of size: ' . $size . 'KB' .
                                          ' with bytes rate: ' . $bytesrate .
                                          ' with count: ' . $count);
                        $api->dd ( "input-file"      =>   '/dev/urandom',
                                   "output-file"     =>   $sl,
                                   "bytes-rate"      =>   $bytesrate,
                                   "block-count"     =>   $count,
                                   'connectrec-timeout' => 900,
                                   'privilege-level' =>   'root',);
                    }
                } elsif ($operation =~ /modify_delete/) {
                    $log->comment('Delete symllinks');
                    my $cnt = 0;
                    foreach my $sl (@symlink_e) {
                        $log->comment('Find the full link path');
                        my $output = $api->find (path    =>  $pt,
                                               name    => $sl);
                        my $path = $output->get_processed_output();
                        if (!grep {$path eq $_} @excercised_links) {
                            if ($cnt == $symlinks) { $log->comment('...done'); last; }
                            $log->comment('Deleting symlink: '  . $path);
                            $api->rm( paths => $path,
                                     'connectrec-timeout' => 120,);
                            push (@excercised_links, $path);
                            $cnt++;
                        }
                    }
                }
            }
        }
    }

    my %ret;
    $log->comment('Cut a ss');
    if($args{cut_a_ss}){
        $snapshot = 'symlinks_create_ss' if (!$snapshot);
    }


    if ($snapshot) {
        foreach my $v (@vols) {
            my $vname = $v->volume();
            $log->log('Take a snapshot ' . $snapshot . ' on volume ' . $vname);
            NACL::STask::VolumeSnapshot->create(
                command_interface  => $v->command_interface,
                volume             => $v,
                vserver            => $v->vserver,
                snapshot           => $snapshot,
                'snapmirror-label' => $snapmirror_label,
            );
            push @{$ret{$vname}{'snapshots'}}, $snapshot;
        }
    }

    return \%ret;
}

=head2 compression_dataset

This method is to generate compression and uncompression datasets in MW volumes.

=over

=item options
    file_paths - file paths for datasets,
    file_count - number of files to write, default to '1',
    file_size  - file size to write, default to '2m',
    dir_depth  - dir depth, default to '1',
    dir_count  - number of dirs to create, default to '1',
    dedupe_pct - deduplication percentage of the datasets, default to '0',
    debug_prt  - printout of wafl showvnb, default to '0',
    data_type   - generate mixed of compressed and uncompressed data with various ratios, default to '0',
    file_fill_pct        - percentage of data fill. rest will be holes. default to '100',
    start_offset         - offset to begin writes, default to 0,
    snapmirror_label     - snapmirror label, default to 'weekly',
    mod_punchhole_pct    - percantage of files to punch holes. default to '0',
    mod_punchhole_by_pct - percantage of each file to punch holes, default => '25',
    mod_overwrite_pct    - percantage of files to modify. default to '100',
    mod_overwrite_by_pct - percantage of each file to modify, default => '50',
    mod_grow_pct         - percantage of files to grow size. default to '0',
    mod_grow_by_pct      - percantage of each file to grow size, default => '25',
    mod_shrink_pct       - percantage of files to shrink size. default to '0',
    mod_shrink_by_pct    - percantage of each file to shrink size, default => '25',

=over

=item required

volume_objs - OBJECT, volume to hold datasets.
client_name - SCALAR, client to drive IOs.

=back

=back

=cut

sub compression_dataset {
    my %args = validate_with(
        params => \@_,
        spec => {
            extra_param => { type => HASHREF|UNDEF, optional => 1},
            volume_objs => { type => ARRAYREF },
            client_name => { type => SCALAR },
            operation   => { type => SCALAR, default => 'populate' },
            file_paths  => { type => ARRAYREF, optional => 1 },
            file_count  => { type => SCALAR, default => '2' },
            file_size   => { type => SCALAR, default => '75m' },
            dir_depth   => { type => SCALAR, default => '1' },
            dir_count   => { type => SCALAR, default => '2' },
            dedupe_pct  => { type => SCALAR, default => '0' },
            debug_prt   => { type => SCALAR, default => '0' },
            data_type   => { type => SCALAR, default => '0' },
            snapshot    => { type => SCALAR, optional => 1 },
            cut_a_ss    => { type => SCALAR, default => 0 },
            flush_cache => { type => SCALAR, default => 1 },
            start_offset         => { type => SCALAR, default => '0' },
            file_fill_pct        => { type => SCALAR, default => '100' },
            snapmirror_label     => { type => SCALAR, default => 'weekly' },
            compression_pct      => { type => SCALAR, default => '25' },
            mod_punchhole_pct    => { type => SCALAR, default => '0' },
            mod_punchhole_by_pct => { type => SCALAR, default => '25' },
            mod_overwrite_pct    => { type => SCALAR, default => '0' },
            mod_overwrite_by_pct => { type => SCALAR, default => '50' },
            mod_grow_pct         => { type => SCALAR, default => '0' },
            mod_grow_by_pct      => { type => SCALAR, default => '25' },
            mod_shrink_pct       => { type => SCALAR, default => '0' },
            mod_shrink_by_pct    => { type => SCALAR, default => '25' },
            timeout              => { type => SCALAR, default => '1800' },
            file_list            => { type => SCALAR, default => '1' },
        },
    );

    my $volume_objs = $args{volume_objs};
    my $client_name = $args{client_name};
    my $operation   = $args{operation};
    my $file_paths  = $args{file_paths};
    my $file_count  = $args{file_count};
    my $file_size   = $args{file_size};
    my $dir_depth   = $args{dir_depth};
    my $dir_count   = $args{dir_count};
    my $dedupe_pct  = $args{dedupe_pct};
    my $snapshot    = $args{snapshot};
    my $data_type   = $args{data_type};
    my $debug_prt   = $args{debug_prt};
    my $file_fill_pct        = $args{file_fill_pct};
    my $start_offset         = $args{start_offset};
    my $snapmirror_label     = $args{snapmirror_label};
    my $compression_pct      = $args{compression_pct};
    my $mod_punchhole_pct    = $args{mod_punchhole_pct};
    my $mod_punchhole_by_pct = $args{mod_punchhole_by_pct};
    my $mod_overwrite_pct    = $args{mod_overwrite_pct};
    my $mod_overwrite_by_pct = $args{mod_overwrite_by_pct};
    my $mod_grow_pct         = $args{mod_grow_pct};
    my $mod_grow_by_pct      = $args{mod_grow_by_pct};
    my $mod_shrink_pct       = $args{mod_shrink_pct};
    my $mod_shrink_by_pct    = $args{mod_shrink_by_pct};
    my $extra_param          = $args{extra_param};
    my $timeout              = $args{timeout};
    my $file_list            = $args{file_list};

    my $log = log_global();
    my $ctmw_file_pattern = 'CTmwLCP';

    # flash cache before generating dataset
    _wafl_flush_cache(node => @{$volume_objs}[0]->get_local_node_name())
        if ($args{flush_cache});

    my $CrTree_Bin = 'dp/metawafl/create_tree_mw.pl';
    my $nate_lib = param('NATE_LIB');
    my @paths = split (/;/, $nate_lib);
    foreach my $path (@paths) {
        if ( -f "${path}/$CrTree_Bin"){
            $CrTree_Bin = "${path}/$CrTree_Bin";
            last;
        }
    }

    # can not exceed 100%
    if(($mod_overwrite_pct + $mod_punchhole_pct) > 100) {
        $mod_overwrite_pct = 100 - $mod_punchhole_pct;
    }

    my $ct_opts =
                "OPERATION_TYPE=$operation " .
                "DEDUPE_PCT=$dedupe_pct " .
                "COMPRESSION_PCT=$compression_pct " .
                "FILE_FILL_PCT=$file_fill_pct " .
                "FILE_SIZE=$file_size " .
                "START_OFFSET=$start_offset " .
                "MW_DATASET=$data_type ";

    if ($operation =~ /modify/i) {
        $ct_opts .=
                "MOD_OVERWRITE_PCT=$mod_overwrite_pct " .
                "MOD_OVERWRITE_BY_PCT=$mod_overwrite_by_pct " .
                "MOD_PUNCHHOLE_PCT=$mod_punchhole_pct " .
                "MOD_PUNCHHOLE_BY_PCT=$mod_punchhole_by_pct " .
                "MOD_GROW_PCT=$mod_grow_pct " .
                "MOD_GROW_BY_PCT=$mod_grow_by_pct " .
                "MOD_SHRINK_PCT=$mod_shrink_pct " .
                "MOD_SHRINK_BY_PCT=$mod_shrink_by_pct ";
    } else {
        $ct_opts .=
                "FILE_CNT=$file_count " .
                "DIR_DEPTH=$dir_depth " .
                "DIR_CNT=$dir_count ";
    }

    #append the extra params
    foreach my $key (keys %{$extra_param}) {
        $ct_opts .=  "$key=" . $extra_param->{$key} . " ";
    }
    $log->log('All params that will be passed to create tree are: ' . Dumper($ct_opts));

    my $client_obj = NACL::C::Client->new( name => $client_name );

    my %ret;
    foreach my $vol (@{$volume_objs}) {
        my (@client_cmds, $opts, $cmd);
        my $extent_size = $vol->get_one_state_attribute('default-extent-size');
        my $vname = $vol->volume();
        $opts = $ct_opts . "IO_SIZE=$extent_size ";

        $log->log("COMP: $operation compression dataset on volume " . $vol->volume);
        my $IOHandle = NACL::AppBuilder::WFC::IO->new(
            volume  => $vol,
            client  => $client_obj,
        );
        my $mount_pnt = $IOHandle->create_mount(access => 'RW')->mount_point();
        $ret{$vname}{mount_point} = $mount_pnt;

        if(!$file_paths){
            # command to generate dataset
            $cmd = "$CrTree_Bin $opts" . $mount_pnt;
            push @client_cmds, $cmd;
        } else {
            foreach my $file (@{$file_paths}){
                # command to generate dataset
                $cmd = "$CrTree_Bin $opts" . "$mount_pnt/$file";
                push @client_cmds, $cmd;
            }
        }

        # generating datasets
        foreach my $cmd (@client_cmds) {
            $client_obj->apiset()->execute_command(
                command              => $cmd,
                'connectrec-timeout' => $timeout,
                'privilege-level'    => 'root'
            );
        }

        # wafl flush cache
        _wafl_flush_cache(node => @{$volume_objs}[0]->get_local_node_name())
            if ($args{flush_cache});

        if($args{cut_a_ss}){
            $snapshot = 'comp_ss_' . time if (!$snapshot);
        }
        if ($snapshot) {
            $log->log('COMP: take a snapshot ' . $snapshot . ' on volume ' . $vol->volume());
            NACL::STask::VolumeSnapshot->create(
                command_interface  => $vol->command_interface,
                volume             => $vol,
                vserver            => $vol->vserver,
                snapshot           => $snapshot,
                'snapmirror-label' => $snapmirror_label,
            );
            push @{$ret{$vname}{'snapshots'}}, $snapshot;
        }

        if ($file_list) {
            $log->log('COMP: get all files in volume ' . $vname);
            my $fobjs = NACL::C::Volume->explore(
                            command_interface => $vol->command_interface,
                            format            => 'inode',
                            find              => 'type!=0',
                            dump              => 'name',
                            scope             => $vname,
                            'method-timeout'  => $timeout,
            );
            my (@fps, @fileobjs);
            if(!$file_paths) {
                foreach my $f_name (@{$fobjs->name}) {
                    if($f_name =~ /$ctmw_file_pattern/) {
                        push @fps, $f_name;
                        my $p = "/vol/$vname/$f_name";
                        push @fileobjs, NACL::STask::VolumeFile->find (
                                    command_interface => $vol->command_interface,
                                    filter            => {
                                            path    => $p,
                                            vserver => $vol->vserver()
                                    }
                        );
                    }
                }
            } else {
                foreach my $fp (@{$file_paths}) {
                    my $p = "/vol/$vname/$fp";
                    my $fobj = NACL::STask::VolumeFile->find (
                                    command_interface => $vol->command_interface,
                                    filter            => {
                                            path    => $p,
                                            vserver => $vol->vserver()
                                    },);
                    if($fobj->state->file_type !~ qr/^directory$/i) { # file/lun
                        $log->log('existing file: ' . $fp);
                        push @fps, $fp;
                        push @fileobjs, $fobj;
                    } else { # can be dir
                        $log->log('existing dir: ' . $fp);
                        foreach my $f_name (@{$fobjs->name}) {
                            if($f_name =~ /$fp.*$ctmw_file_pattern/) {
                                push @fps, $f_name;
                                my $p = "/vol/$vname/$f_name";
                                push @fileobjs, NACL::STask::VolumeFile->find (
                                            command_interface => $vol->command_interface,
                                            filter            => {
                                                    path    => $p,
                                                    vserver => $vol->vserver()
                                            }
                                );
                            }
                        }
                    }
                }
            }
            $ret{$vname}{'file_names'} = \@fps;
            $ret{$vname}{'file_obj'} = \@fileobjs;
            $log->log('COMP: files are: ' . Dumper($ret{$vname}{'file_names'}));
        }

        # for debugging: printout of compression details of the 1st file
        if($debug_prt) {
            $log->log('COMP: -- debug output: showvbns of a file --');
            my $node_obj = NACL::C::Node->new(name => @{$volume_objs}[0]->get_local_node_name());
            my $api = $node_obj->get_7m_or_nodescope_apiset();
            my $file = @{$ret{$vname}{'file_names'}}[0];
            my $cmd = "wafl showvbns /vol/$vname/$file";
            $api->execute_command(command => $cmd, 'connectrec-timeout' => 600);
            _wafl_flush_cache(node => @{$volume_objs}[0]->get_local_node_name())
                if ($args{flush_cache});
        }
    }

    return \%ret;
}

=head2 appcg_dataset_adaptive

This method is to write the data using adaptive dataset provided by SE team.

=over

=item options
    src_vol_name - the name of src vol,
    client - the name of name,
    node_mgmt_ip - the node mangement ip addr,
    cg_size - compression group size,
    path - path for mount point,

=over

=item required

volume_objs - OBJECT, volume to hold datasets.
client_name - SCALAR, client to drive IOs.

=back

=back

=cut

sub appcg_dataset_adaptive {
    my %args = validate_with(
        params => \@_,
        spec => {
            src_vol_name => { type => SCALAR },
            client       => { type => OBJECT },
            node_mgmt_ip => { type => SCALAR },
            cg_size      => { type => SCALAR },
            path         => { type => SCALAR },
            count        => { type => SCALAR, optional => 1, default => '1' },
            num_files    => { type => SCALAR, optional => 1, default => '8' },
            file_size    => { type => SCALAR, optional => 1, default => '1k,5k,15k,100k,500k,700k,2M,5M' },
            vbn_bad      => { type => SCALAR, optional => 1, default => '1' },
            timeout      => { type => SCALAR, optional => 1, default => '1800' },
        },
    );
    my $tool_path_option = $NACL::AppBuilder::DpgWorkflows::StringConstants::DATASET_TOOL_PATH_ADAPTIVE;
    #my $tool_path_option = '/usr/software/test/common/dev/dblade/wafl/compression/tools/cmpr_dataset_adaptive.pl';
    my $status ;
    my $log = log_global();
    my $output = $args{client}->execute(
        command =>
            "/usr/bin/sudo $tool_path_option -cg_size $args{cg_size} \-filer $args{node_mgmt_ip} \-volume \"$args{src_vol_name}\"  \-mnt_pt  $args{path} -num_files $args{num_files} \-file_name foo \-overwrite 0 \-size \"$args{file_size}\" \-vbn_bad $args{vbn_bad}",
        get_status => \$status,
        timeout    => $args{timeout},
    );
    if ( !defined $output ) {
        NACL::APISet::Exceptions::CommandFailedException->throw($output);
    }
    if (   $output =~ /Read-only file system/i
        || $output =~ /error opening file/i
        || $output =~ /No such file or directory/i ) {
        $log->step("Failed to write data on source volume");
        NATE::BaseException->throw($output);
    }


}#end of appcg_dataset_adaptive

=head2 appcg_dataset_secondary

This method is to write the data using secondary dataset provided by SE team.

=over

=item options
    src_vol_name - the name of src vol,
    client - the name of name,
    node_mgmt_ip - the node mangement ip addr,
    cg_size - compression group size,
    path - path for mount point,

=over

=item required

volume_objs - OBJECT, volume to hold datasets.
client_name - SCALAR, client to drive IOs.

=back

=back

=cut

sub appcg_dataset_secondary {
    my %args = validate_with(
        params => \@_,
        spec => {
            src_vol_name => { type => SCALAR },
            client       => { type => OBJECT },
            node_mgmt_ip => { type => SCALAR },
            path         => { type => SCALAR },
            count        => { type => SCALAR, optional => 1, default => '1' },
            num_files    => { type => SCALAR, optional => 1, default => '8' },
            file_size    => { type => SCALAR, optional => 1, default => '1k,5k,15k,100k,500k,700k,2M,5M' },
            vbn_bad      => { type => SCALAR, optional => 1, default => '1' },
            timeout      => { type => SCALAR, optional => 1, default => '1800' },
        },
    );
    my $tool_path_option = $NACL::AppBuilder::DpgWorkflows::StringConstants::DATASET_TOOL_PATH_SECONDARY;
    #my $tool_path_option = '/usr/software/test/common/dev/dblade/wafl/compression/tools/cmpr_dataset.pl';
    my $status ;
    my $log = log_global();
    my $output = $args{client}->execute(
        command =>
            "/usr/bin/sudo $tool_path_option \-filer $args{node_mgmt_ip} \-volume \"$args{src_vol_name}\"  \-mnt_pt  $args{path} -num_files $args{num_files} \-file_name foo \-overwrite 0 \-compress_partial_cg_inline 1 \-full_cgs 1  \-size \"$args{file_size}\" \-vbn_bad $args{vbn_bad}",
        get_status => \$status,
        timeout    => $args{timeout},
    );
    if ( !defined $output ) {
        NACL::APISet::Exceptions::CommandFailedException->throw($output);
    }
    if (   $output =~ /Read-only file system/i
        || $output =~ /error opening file/i
        || $output =~ /No such file or directory/i ) {
        $log->step("Failed to write data on source volume");
        NATE::BaseException->throw($output);
    }


}#end of appcg_dataset_secondary

=head2 appcg_populate_overwrite

This method is to populate and overwrite the data based on provided compression group.

=over

=item options
    src_vol_name - the name of src vol,
    client - the name of name,
    node_mgmt_ip - the node mangement ip addr,
    cg - compression group ,
    path - path for mount point,
    count - count to have diff run_id for each iteration,

=over

=item required

volume_objs - OBJECT, volume to hold datasets.
client_name - SCALAR, client to drive IOs.

=back

=back

=cut

sub appcg_populate_overwrite {
    my %args = validate_with(
        params => \@_,
        spec => {
            src_vol_name => { type => SCALAR },
            client       => { type => OBJECT },
            node_mgmt_ip => { type => SCALAR },
            cg           => { type => SCALAR },
            path         => { type => SCALAR },
            count        => { type => SCALAR, optional => 1, default => '1' },
            timeout      => { type => SCALAR, optional => 1, default => '1800' },
        },
    );
    my $cg = $args{cg};
    my $src_vol_name = $args{src_vol_name};
    my $client = $args{client};
    my $node_mgmt_ip = $args{node_mgmt_ip};
    my $path = $args{path};
    my $count = $args{count} ;  #count is used for creating logs with diff run_id for each overwrite
    my $log = log_global();
    my $proc;

    if ($cg eq 'Sec32k'){
        $log->log("Overwrite the data with ". $cg . ' dataset' );
        $proc = NATE::Process->new(
            codespec => sub { appcg_dataset_secondary( src_vol_name => $src_vol_name, client => $client ,
                                                       node_mgmt_ip => $node_mgmt_ip, path => $path ,
                                                       count => $count ) },
            runid    => 'populate_' . $count,
            inherit  => 1,
            onexit => \&NATE::Process::on_exit_propagate_worst_result
        );
    }
    elsif ($cg eq 'Adap8k' ){
        $log->log("Overwrite the data with $cg dataset");
        $proc = NATE::Process->new(
            codespec => sub { appcg_dataset_adaptive( src_vol_name => $src_vol_name, client => $client ,
                                                       node_mgmt_ip => $node_mgmt_ip, cg_size => '8K' ,
                                                       path => $path, count => $count ) },
            runid    => 'populate_' . $count,
            inherit  => 1,
            onexit => \&NATE::Process::on_exit_propagate_worst_result
        );
    }
    elsif ($cg eq 'Adap16k' ){
        $log->log("Overwrite the data with $cg dataset");
        $proc = NATE::Process->new(
            codespec => sub { appcg_dataset_adaptive( src_vol_name => $src_vol_name, client => $client ,
                                                       node_mgmt_ip => $node_mgmt_ip, cg_size => '16K' ,
                                                       path => $path , count => $count ) },
            runid    => 'populate_' . $count,
            inherit  => 1,
            onexit => \&NATE::Process::on_exit_propagate_worst_result
        );
    }
    elsif ($cg eq 'Adap32k' ){
        $log->log("Overwrite the data with $cg dataset");
        $proc = NATE::Process->new(
            codespec => sub { appcg_dataset_adaptive( src_vol_name => $src_vol_name, client => $client ,
                                                       node_mgmt_ip => $node_mgmt_ip, cg_size => '32K' ,
                                                       path => $path , count => $count ) },
            runid    => 'populate_' . $count,
            inherit  => 1,
            onexit => \&NATE::Process::on_exit_propagate_worst_result
        );
    }
    else {
        $log->log("Overwrite the data with $cg dataset");
        $proc = NATE::Process->new(
            codespec => sub { appcg_dataset_secondary( src_vol_name => $src_vol_name, client => $client ,
                                                       node_mgmt_ip => $node_mgmt_ip, path => $path ,
                                                       count => $count ) },
            runid    => 'populate_' . $count,
            inherit  => 1,
            onexit => \&NATE::Process::on_exit_propagate_worst_result
        );
    }
    $proc->start;
    $proc->wait();
    $proc->destroy;

    $count++;
    return $count;
}#end of appcg_populate_overwrite

sub _wafl_flush_cache {
    my %args = validate_with(
        params => \@_,
        spec => {
            node => { type => SCALAR },
        },
    );

    my $node = $args{node};
    my $node_obj = NACL::C::Node->new(name => $node);
    $node_obj->privilege_level('test');
    my $api = $node_obj->get_7m_or_nodescope_apiset();
    my $cmd = "wafl flush cache";
    $api->execute_command(command => $cmd ,'connectrec-timeout' => 1200);
    $api->api_reconnect_session();
    $node_obj->refresh_command_interface();
    sleep(30);
}


sub populate_with_datagen{

   my %args = validate_with(
        params => \@_,
        spec => {
            volume_objs => { type => ARRAYREF, optional=>0 },
            ss_to_create => { type=> ARRAYREF, optional=>1, default=> ['s1','s2']},
            snapmirror_label => { type => SCALAR, optional=>1, default=> 'weekly' },
        },
    );

  my $vol_objs = $args{'volume_objs'};
  my $ss_to_create = $args{'ss_to_create'};
  my @snapshots = @{$ss_to_create};
  my $snapmirror_lbl = $args{'snapmirror_label'};

  foreach my $vol(@{$vol_objs}){
     my $Src_vol_name =  $vol->volume;
     my $Src_Vserver = $vol->vserver;
     my $Src_Aggr_name = $vol->get_one_state_attribute('aggregate');
     my $Host_A = $vol->get_local_node_name();

     my $subtest_handle = subtest(
            'dp/tools/DataGen/DataGen_CMode.thpl', "VOL=$Src_vol_name",
            "FILER=$Host_A",                       "VSERVER=$Src_Vserver",
            "AGGR=$Src_Aggr_name",                 'SPARSE_FILE_SIZE=200m',
            'SPARSE_DIR_WIDTH=2',                  'SPARSE_DIR_DEPTH=2',
            'SPARSE_FILE_COUNT=5',                 'SPARSE_PERCENT=70',
            'DATASET_TYPE=SPARSE'
        );
     Subtest::wait_finish( subtests => [$subtest_handle] );

     my $ss_to_initialize = shift @snapshots;

     $vol->snapshot_create(snapshot           => $ss_to_initialize,
                          'snapmirror-label' => $snapmirror_lbl,
                           );

     foreach my $ss( @snapshots) {
          my $subtest_handle = subtest(
             'dp/tools/DataGen/DataGen_CMode.thpl', "VOL=$Src_vol_name",
             "FILER=$Host_A",                       "VSERVER=$Src_Vserver",
             "AGGR=$Src_Aggr_name",                 'DEDUPE_FILE_SIZE=200m',
             'DEDUPE_DIR_WIDTH=2',                  'DEDUPE_DIR_DEPTH=2',
             'DEDUPE_FILE_COUNT=2',                 'DEDUPE_PER=40',
             'DATASET_TYPE=DEDUPE',
             "OPER=modify",
             "PERCENT_MODIFICATION=20",
          );
          Subtest::wait_finish( subtests => [$subtest_handle] );

          $vol->snapshot_create(snapshot           => $ss,
                                'snapmirror-label' => $snapmirror_lbl,
                                );
     }
  }
}


=head2 _execute_iofence_cmd

This method is to create iofence on a given file path.

=item required

=over

cmd - SCALAR, to do set/clear/get iofence
nodescope - OBJECT,
path - SCALAR.
type - type of iofence.
action - SCALAR, fail/suspend.
client - SCALAR, client name

=back

=cut

sub _execute_iofence_cmd {

   #--------------------------------------------------------------------------
   # Usage: iofence <set | clear | get>^M
   # {file_path_name |-i <inode num> -g <gen num> -f <fsid> -s <snapid>}^M
   # [-t <1=ondisk_rw (default) 2=incore_rw 3=incore_w>]^M
   # [-e <1=fail (default) 2=suspend>]^M
   #--------------------------------------------------------------------------

    my %args = validate_with(
        params => \@_,
        spec => {
            cmd => { type => SCALAR },
            nodescope => { type => OBJECT },
            path   => { type => SCALAR },
            type     => { type => SCALAR, optional=>1 },
            action  => { type => SCALAR, optional=>1 },
            client  => { type => ARRAYREF|SCALAR, optional=>1 },
            volume  => { type => OBJECT, optional => 1},
            check_RW => {type => SCALAR, optional=>1 , default=>'true', },
        },
    );


    my $cmd = $args{cmd};
    my $nodescope = $args{nodescope};
    my $path = $args{path};
    my $type = $args{type};
    my $action = $args{action};
    my $v = $args{volume};
    my @traffic_client;


    my $log = log_global();
    $log->enter();

    my $iofence_cmd;
    if ( $cmd =~ /get/ ) {
        $iofence_cmd = "iofence get $path";
    } elsif ( $cmd =~ /set/ ) {
        $iofence_cmd = "iofence set $path -t $type -e $action";
    } elsif ( $cmd =~ /clear/ ) {
        $iofence_cmd = "iofence clear $path -t $type";
    }

    my $out = $nodescope->execute_command( command => $iofence_cmd );
    $log->comment( 'Result iofence cmd ' . $cmd . ' : ' . $out );

    if ( $cmd =~ /set/ ) {
        sleep(30);
        try {

           if (ref($args{client}) eq 'ARRAY') {
             @traffic_client = map( NACL::C::Client->new( name => $_ ), @{$args{client}} );
         } else {
             @traffic_client = map( NACL::C::Client->new( name => $_ ), $args{client} );
         }

         my $client_name = $traffic_client[0]->client();
         my $Client_Obj = $traffic_client[0];

         my @fields = split /\//, $path;
         my $file_name = $fields[3];

         $log->comment("file name is" .  $file_name);

         my $mtpt = mount_nfs_vol (
              volume      => $v,
              client_name => $client_name,
              nfsvers     => '3');


         my $client_path = "$mtpt/$file_name";

         my $Client_apiset = $Client_Obj->apiset(interface => "CLI",
                                                  );
         $Client_apiset->ls('long-list'=>'1',
                            paths=>$client_path);

         $Client_apiset->set_timeout('connectrec-timeout' => '600');

         sleep(30);

          my $response_obj = $Client_apiset->sio(
            "privilege-level" => "root",
            duration_in_secs  => '10',
            devices_list      => $client_path,
            rand              => '0',
            block_size        => '4k',
            file_size         => '10m',
            start_byte        => '0',
            num_of_threads    => '3',
            read              => '0',
        );
          $log->error('SIO was able to write to the fenced file.');
          $log->comment(
            'Response sio ' . Dumper( $response_obj->{parsed_output}->[0] ) );
        } catch NATE::BaseException with {

              my $exception_obj = shift;
              $log->comment( 'SIO write command failed: ' . $exception_obj->text() );
              if ($exception_obj->text() =~ /file access error|Error locking file|Timeout/ ) {
                  $log->comment( 'SIO reports failure - as expected.'
                             . $exception_obj->text() );
              } else {
                  $log->comment('SIO failed for unexpected reason: '.$exception_obj->text());
              }
          };
      #}
    }


    # Get can be whatever it is - our set and clear cmds should succeed.
    return 1 if ( $cmd =~ /set|clear/ && $out =~ /Success/ );
    return $out if ( $cmd =~ /get/ );
    NATE::BaseException->throw('IOFence command ' . $cmd . 'failed.' . $out );
  #}
} ## end sub _execute_iofence_cmd



# set the placement for flexgroup to be manual do on all nodes
#Reference: https://wikid.netapp.com/w/FlexGroup/QA/TestingFAQ#How_do_I_determine_where_files_get_placed_during_ingest.3F

sub fg_set_placement {
    my %args = validate_with(
        params      => \@_,
        allow_extra => 0,
        spec        => {
            node   => { type => OBJECT },
            action => { type => SCALAR, regex => qr/^(auto|manual)$/ },
        },
    );
    my $log = log_global();

    my @node_objs = NACL::C::Cluster->node_objs( command_interface => $args{node} );

    my $error;
    foreach my $nodeobj (@node_objs) {
        $log->log('Setting flexgroup placement to '.$args{action}.' for node '. $nodeobj->name() );
        my $response = $nodeobj->get_7m_or_nodescope_apiset->execute_command(command => "priv set -q test; flexgroup set placement=$args{action}");
        if ($response !~ /ok/) {
            my $msg = "Was not able to set flexgroup placement to $args{action}. Expected 'ok' in response. Got $response";
            $log->error($msg);
            $error.=$msg;
        }
    }
    if ($error) {NATE::BaseException->throw($error)};
    return;
}

##This subroutine creates qtrees on a FlexGroup volume. It takes three params:
## 1. FlexGroup Volume Object
## 2. Location of the qtree(s) to be created: root, non-root-local and non-root-remote
##    > root location denotes a qtree being placed on root constituent
##    > non-root-local location denotes a qtree being placed on same node as root
##      constituent but on a different constituent
##    > non-root-remote location denotes a qtree being placed on a node other than
##      root constituent node
## 3. Number of qtree(s) to be created
sub fg_create_qtrees{
    my %args = validate_with(
        params => \@_,
        spec   => {
            vol_obj  => { type => OBJECT, isa => 'NACL::STask::Volume' },
            location => { type => SCALAR },
            count    => { type => SCALAR },
        },
    );
    my $log        = log_global(1);
    my $flexgroup  = $args{vol_obj};
    my $location   = $args{location};
    my $num_of_qts = $args{count};
    my $ci         = $flexgroup->command_interface();
    my $vserver    = $flexgroup->vserver();
    my $volume     = $flexgroup->volume();
    $log->step('Identifying target constituent');
    my $root_constituent = $flexgroup->get_flexgroup_constituents(
        filter  => {
            'constituent-role' => 'root',
        },
    );
    my $root_node = $root_constituent->get_one_state_attribute('node');

    my @target_constituent = ($root_constituent);
    if ($location eq 'non-root-local') {
        try {
            @target_constituent = $flexgroup->get_flexgroup_constituents(
                filter => {
                    node               => $root_node,
                   'constituent-role' => '!root',
                },
            );
        } catch NATE::BaseException with {
            my $ex = shift;
            if ( $ex->text =~ /No matching volume/i ) {
                $log->log(
                    "No matching constituent volume found due to ARL or VolMove"
                  . ", skipping qtree creation"
                );
                return;
            }
        };
    } elsif ($location eq 'non-root-remote') {
        try {
            @target_constituent = $flexgroup->get_flexgroup_constituents(
                filter => {
                    node => "!$root_node",
                },
            );
        } catch NATE::BaseException with {
            my $ex = shift;
            if ( $ex->text =~ /No matching volume/i ) {
                $log->log(
                    "No matching constituent volume found due to ARL or VolMove"
                  . ", skipping qtree creation"
                );
                return;
            }
        };
    }
    fg_set_placement( node => $ci, action => 'manual');

    my @securities = qw(default  mixed    unix     ntfs);
    my @oplocks    = qw(enable   default  disable  default);
    while($num_of_qts > 0){
        my @security_oplock_arg;
        my $security = $securities[int(rand(4))];
        my $oplock   = $oplocks[int(rand(4))];
        if ( $security ne 'default' ) {
            push @security_oplock_arg, 'security-style' => $security;
        }
        if ( $oplock ne 'default' ) {
            push @security_oplock_arg, 'oplock-mode' => $oplock;
        }
        my $target_fgindex = $target_constituent[int(rand(scalar(@target_constituent)))]
            ->get_one_state_attribute('flexgroup-index');
        my $qtree_name = random_name_generator(
            prefix => $location . '_' . $security . '_' . $oplock,
            size   => 1,
            suffix => "fgindex${target_fgindex}"
        );
        nacl_method_retry(
            code        => sub {
                NACL::STask::Qtree->create(
                    command_interface => $ci,
                    vserver           => $vserver,
                    volume            => $flexgroup->volume(),
                    qtree             => $qtree_name,
                    @security_oplock_arg,
                );
            },
            tries_count => 6,
            sleep_time  => 10,
            exceptions  => 'NATE::BaseException',
        );
        --$num_of_qts;
        $log->log($num_of_qts.' left to be created.');
    }

    fg_set_placement( node => $ci, action => 'auto');
} ## end sub fg_create_qtrees

## This subroutine populates all the qtrees present in a FlexGroup Volume.
## It takes four params:
## 1. FlexGroup volume Object
## 2. WFC::IO object
## 3. Type of qtrees to be populated: root, non-root-local and non-root-remote
## 4. Dataset Size
sub fg_populate_qtrees {
    my %args = validate_with(
        params => \@_,
        spec   => {
            vol_obj                        => { type => OBJECT,   isa      => 'NACL::STask::Volume' },
            io_handle                      => { type => ARRAYREF, optional => 1 },
            populate_qtree_root            => { type => SCALAR,   optional => 1, default => 1 },
            populate_qtree_non_root_local  => { type => SCALAR,   optional => 1, default => 1 },
            populate_qtree_non_root_remote => { type => SCALAR,   optional => 1, default => 1 },
            dataset_size                   => { type => SCALAR,   default  => '1m' },
        },
    );
    my $log                            = log_global(1);
    my $flexgroup                      = $args{vol_obj};
    my $populate_qtree_root            = $args{populate_qtree_root};
    my $populate_qtree_non_root_local  = $args{populate_qtree_non_root_local};
    my $populate_qtree_non_root_remote = $args{populate_qtree_non_root_remote};
    my $dataset_size                   = $args{dataset_size};
    my $ci                             = $flexgroup->command_interface();
    my $vserver                        = $flexgroup->vserver();
    my $volume                         = $flexgroup->volume();
    my $io_handle                      = @{ $args{io_handle} }[0];

    my @qtrees_to_be_populated;
    if ( $populate_qtree_root ) {
        $log->step('Fetch the root qtrees');
        my @qtree_root = NACL::CS::Qtree->fetch(
            command_interface => $ci,
            filter            => {
                vserver => $vserver,
                volume  => $flexgroup,
                qtree   => 'root*',
            },
        );
        splice @qtrees_to_be_populated, scalar @qtrees_to_be_populated, 0, @qtree_root;
    }

    if ( $populate_qtree_non_root_local ) {
        $log->step('Fetch the non root local qtrees');
        my @qtree_non_root_local = NACL::CS::Qtree->fetch(
            command_interface => $ci,
            filter            => {
                vserver => $vserver,
                volume  => $flexgroup,
                qtree   => 'non-root-local*',
            },
        );
        splice @qtrees_to_be_populated, scalar @qtrees_to_be_populated, 0, @qtree_non_root_local;
    }

    if ( $populate_qtree_non_root_remote ) {
        $log->step('Fetch the non root remote qtrees');
        my @qtree_non_root_remote = NACL::CS::Qtree->fetch(
            command_interface => $ci,
            filter            => {
                vserver => $vserver,
                volume  => $flexgroup,
                qtree   => 'non-root-remote*',
            },
        );
        splice @qtrees_to_be_populated, scalar @qtrees_to_be_populated, 0, @qtree_non_root_remote;
    }

    foreach ( @qtrees_to_be_populated ) {
        if ( $_->security_style !~ /ntfs/i && $io_handle->client()->type =~ /linux/i ) {
            $log->log(
                'Populating qtree ' . $_->qtree() . ' using client '
              . $io_handle->client()->name
            );
            $io_handle->data_populate(
                size => $dataset_size,
                path => $_->qtree() . "/file_" . time(),
            );
        } elsif ( $_->security_style !~ /unix/i && $io_handle->client()->type =~ /win/i ) {
            $log->log(
                 'Populating qtree ' . $_->qtree() . ' using client '
               . $io_handle->client()->name
            );
            $io_handle->data_populate(
                size => $dataset_size,
                path => $_->qtree() . "/file_" . time(),
            );
        } else {
            $log->log(
                'Skipping population of qtree: ' . $_->qtree() . ' as the '
              . 'security style: ' . $_->security_style . ' and client type: '
              . $io_handle->client()->type . ' of client: '
              . $io_handle->client()->name . ' are not compatible!'
            );
        }
    }
} ## end sub fg_populate_qtrees

# populate a FG using create tree
sub fg_populate {
    my %args = validate_with(
        params      => \@_,
        allow_extra => 0,
        spec        => {
            volume           => { type => OBJECT },
            client           => { type => ARRAYREF|SCALAR},
            dataset_size     => { type => SCALAR },
            data_module      => { type => SCALAR, regex => qr/create_tree/, default => 'create_tree' },
            filename_prefix  => { type => SCALAR, default => 'file' },
            qtree_count      => { type => SCALAR, default => '1' },
            iterations       => { type => SCALAR, default => 1},
            iteration_snooze => { type => SCALAR, default => 1},
        },
    );
    my $log = log_global();

    my $vol_obj      = $args{volume};
    my $dataset_size = $args{dataset_size};
    my $qtree_count  = $args{qtree_count};

    # The client may come as a scalar or a list. If it's a list just use the first one
    my $client = (ref $args{client}) ? $args{client}->[0] : $args{client};
    # There are cases where the scalar value might be a comma-separated list:
    if ($client =~ /,/) {
        $client = (split /,/ => $client)[0];
    }
    my $client_obj = NACL::C::Client->new( name => $client );

    $log->log('Mount volume '.$vol_obj->volume().' on client '.$client);
    my $mount_obj = NACL::STask::Client::Mount->create( command_interface  => $client_obj,
                                                        volume             => $vol_obj,
                                                        nacltask_if_exists => 'purge',
                                                        nacltask_verify    => 1,
                                                    );

    $log->log('Create IO handle');
    my $IOHandle = NACL::AppBuilder::WFC::IO->new(
        volume => $vol_obj,
        client => $client_obj
    );

    $log->log('Create mount point');
    $IOHandle->create_mount( access => 'RW' );

    $log->log('Create some qtrees and populate them');
    foreach ( 'root', 'non-root-local', 'non-root-remote' ) {
        fg_create_qtrees(
            vol_obj  => $vol_obj,
            location => $_,
            count    => $qtree_count
        );
    }

    fg_populate_qtrees(
        vol_obj      => $vol_obj,
        io_handle    => [ $IOHandle ],
        dataset_size => '1m'
    );

    # Set the flexgroup placement to manual
    my $node = NACL::C::Node->new(name => $vol_obj->get_local_node_name());
    fg_set_placement(node=>$node, action => 'manual');

    # to make it interesting for flexgroup make files so they will land on different constituents
    my @constituents = $vol_obj->get_flexgroup_constituents();
    my $num_constituents = scalar (@constituents);
    my $num_files = $num_constituents;

    # convert dataset_size to bytes, divide by # of constuitents
    my ($unit) = ($dataset_size =~ /\d*(\S*)/);
    my $dataset_size_in_bytes = convert_to_bytes($dataset_size);
    $log->log("dataset_size $dataset_size in_bytes=$dataset_size_in_bytes unit=$unit");

    # create tree fails if size is not integer so collect the mod and add it to the last file_size
    my $file_size_in_bytes = int($dataset_size_in_bytes / $num_constituents);
    my $file_size = $file_size_in_bytes;
    my $left_over_size_in_bytes = $dataset_size_in_bytes % $num_constituents;
    my $last_file_size_in_bytes = $file_size_in_bytes + $left_over_size_in_bytes if ($left_over_size_in_bytes);
    my $last_file_size = $last_file_size_in_bytes;
    $log->log("Got $num_constituents for the flexgroup. Setting num_file=$num_files file_size=$file_size last_file_size=$last_file_size");

    # create the file names so that they will land on different constituents
    my @filenames;
    foreach my $iteration (1..$args{iterations}){
	$log->log("Starting iteration ".$iteration);
	for (my $i = 1; $i <= $num_files; $i++) {
            $file_size  = $last_file_size if (($i == ($num_files)) && $last_file_size);
            my $filename = $args{filename_prefix}.'_'.$file_size.'_fgindex'.$i; # should end in fgindexN where N is member #
            $log->log("Populate flexgroup $vol_obj->volume file $filename with data $file_size");
            dataset( operation => 'populate',
                volume    => $vol_obj,
                mount     => $mount_obj,
                client    => $client_obj,
                file_size => $file_size,
                file_list => ["/$filename"],
            );
            push @filenames, $filename;
	}
	Tharn::snooze($args{iteration_snooze});
    }

    # set the placement back to auto (default)
    fg_set_placement(node=>$node, action => 'auto');

    return @filenames;
}

sub start_fg_read_traffic {
    my %args = validate_with(
        params => \@_,
        spec   => {
            vol_obj   => { type => OBJECT },
            client    => { type => SCALAR|ARRAYREF },
            filenames => { type => ARRAYREF },
        }
    );
    my $log = log_global();

    my $vol_obj = $args{vol_obj};
    # The client may come as a scalar or a list. If it's a list just use the first one
    my $client = (ref $args{client}) ? $args{client}->[0] : $args{client};
    # There are cases where the scalar value might be a comma-separated list:
    if ($client =~ /,/) { $client = (split /,/ => $client)[0]; }
    $log->log("Mount $vol_obj->volume on $client");
    my $io_handle = NACL::AppBuilder::WFC::IO->new( volume => $vol_obj,
                                                    client => $client
                                                    );
    $io_handle->create_mount( access => 'RW' );
    my @traffic_ids;
    my $filenames_ref = $args{filenames};
    foreach my $filename (@$filenames_ref) {
        #----------------------------------------------------------
        $log->step("Start read traffic on volume $vol_obj->volume file $filename");
        #----------------------------------------------------------
        my $id = $io_handle->traffic_start( traffic_file => $filename,);
        $log->log("Got traffic id $id");
        push(@traffic_ids, $id); # not used really
    }
    return $io_handle;
}

sub stop_fg_read_traffic {
    my %args = validate_with(
        params => \@_,
        spec   => {
            io_handle => { type => OBJECT }, # WFC::IO handle
        },
    );
    $args{io_handle}->traffic_stop();
}


=head2 files_populate

This method is to populate volume files with
/usr/software/test/bin/populate.

=item required

    client - OBJECT,
    path   - SCALAR,

=back

=cut

sub files_populate {
    my %args = validate_with(
        params => \@_,
	allow_extra => 1,
        spec   => {
            client => { type => OBJECT },
	    path   => { type => SCALAR },
        },
    );

    my $client_obj = $args{client};
    my $path = $args{path};

    my $populate_bin = "/usr/software/test/bin/populate";
    my $opts;
    foreach my $key ( keys %args ){
        $opts = $opts . " $key=$args{$key}" if !($key =~ /client|path/);
    }

    my $cmd = "$populate_bin $path" . $opts;
    $client_obj->apiset()->execute_command(
	command              => $cmd,
	'connectrec-timeout' => 3600,
	'privilege-level'    => 'root'
    );

    return;
}


=head2 files_modify

This method is to modify volume files with
/usr/software/test/bin/modify.

=item required

    client - OBJECT,
    path   - SCALAR.

=back

=cut

sub files_modify {
    my %args = validate_with(
        params => \@_,
	allow_extra => 1,
        spec   => {
            client => { type => OBJECT },
	    path   => { type => SCALAR },
        },
    );

    my $client_obj = $args{client};
    my $path = $args{path};

    my $modify_bin = "/usr/software/test/bin/modify";
    my $opts;
    foreach my $key ( keys %args ){
        $opts = $opts . " $key=$args{$key}" if !($key =~ /client|path/);
    }

    my $cmd = "$modify_bin $path" . $opts;
    $client_obj->apiset()->execute_command(
	command              => $cmd,
	'connectrec-timeout' => 3600,
	'privilege-level'    => 'root'
    );

    return;
}

sub ldr_deterministic_dataset {
    my %args = validate_with(
        params => \@_,
        allow_extra => 1,
        spec   => {
            volume => {type => OBJECT },
	    client => { type => SCALAR },
	},
    );

    my $log = log_global();

    my $vol_obj = $args{volume};
    my $client = $args{client};

    my $client_obj = NACL::C::Client->new( name => $client );
    my $IOHandle = NACL::AppBuilder::WFC::IO->new(
	volume	=> $vol_obj,
	client	=> $client_obj,
	);

    $IOHandle->create_mount('access' => 'RW');
    my $mount_pt = $IOHandle->mount_point();

    $log->step('Creating Direcories, files and links on the src volume.');
    try{
        my $chmod_cmd = "chmod 777 $mount_pt";
        $client_obj->apiset()->execute_command(
            command              => $chmod_cmd,
            'connectrec-timeout' => 60,
            'privilege-level'    => 'root'
        );

        my $drename_cmd = "drename setup -v -d 8 -c 8 -t 1 -f 1 -p $mount_pt";
        $client_obj->apiset()->execute_command(
            command              => $drename_cmd,
            'connectrec-timeout' => 120,
            'privilege-level'    => 'root'
        );

        $log->step('Creating local symbolic links, local hard links, remote hard links');
        my @cmds = ("ln -s $mount_pt/dir0_fgindex1/dir1_fgindex2/filewithfairlylongname~0 $mount_pt/dir0_fgindex1/dir1_fgindex2/filewithfairlylongname~0_symlink", "ln $mount_pt/dir0_fgindex1/dir1_fgindex2/filewithfairlylongname~0 $mount_pt/dir0_fgindex1/dir1_fgindex2/filewithfairlylongname~0_hardlink", "ln $mount_pt/dir0_fgindex1/dir1_fgindex2/dir2_fgindex3/filewithfairlylongname~0 $mount_pt/dir0_fgindex1/dir1_fgindex2/filewithfairlylongname~0_rhardlink");
        foreach my $cmd(@cmds){
            $client_obj->apiset()->execute_command(
		command	=> $cmd,
		'connectrec-timeout' => 60,
	'privilege-level'    => 'root'
		);
        }
    }catch NATE::BaseException with{
	my $ex = shift;
	NATE::BaseException->throw('Dataset creation failed due to reason: '.$ex->text());
    };
    return;
}

sub fg_create_hundreds_of_files_and_links {
    my %args = validate_with(
        params      => \@_,
        allow_extra => 0,
        spec        => {
            volume         => { type => OBJECT },
	    iohandle       => { type => OBJECT },
            client         => { type => OBJECT },
	    create_files   => { type => SCALAR, default => 1 },
	    file_cnt       => { type => SCALAR, default => 200 },
            create_lnks    => { type => SCALAR, default => 1 },
            min_lnk_cnt    => { type => SCALAR, default => 500 },
            max_lnk_cnt    => { type => SCALAR, default => 1000 },
	    blk_repeat_cnt => { type => SCALAR, default => 1000},
	    wild_character_array => {
                type => ARRAYREF,
                default =>
                  [ '~', '^', '=', '_', '-', ':', '@', '$', '#', '%', '+' ]
            },
        },
    );
    my $log = log_global();

    my $vol_obj		  = $args{volume};
    my $dataset_size      = $args{dataset_size};
    my $create_files_flag = $args{create_files};
    my $create_links_flag = $args{create_lnks};
    my $client_obj	  = $args{client};
    my $IOHandle	  = $args{iohandle};
    my $mount_pt	  = $IOHandle->mount_point();

    # Set the flexgroup placement to manual
    my $node = NACL::C::Node->new(name => $vol_obj->get_local_node_name());
    fg_set_placement(node=>$node, action => 'manual');

    # to make it interesting for flexgroup make files so they will land on different constituents
    my @constituents = $vol_obj->get_flexgroup_constituents();
    my $num_constituents = scalar (@constituents);
    my $lnk_cnt_delta = $args{max_lnk_cnt} - $args{min_lnk_cnt};
    my $num_links = $args{min_lnk_cnt} + int(rand($lnk_cnt_delta));
    my @wld_char_array = @{ $args{wild_character_array} };
    my (@filenames, @filepaths);
    foreach my $index(1..$num_constituents){
        #Create a directory with the fg_index as suffix
        try{
            my $chmod_cmd = "chmod 777 $mount_pt";
            $client_obj->apiset()->execute_command(
                command              => $chmod_cmd,
                'connectrec-timeout' => 60,
                'privilege-level'    => 'root'
            );
            my $dir_path = $mount_pt.'/ldr_dir'.$index.'_fgindex'.$index;
            my $dir_creat_cmd = "mkdir $dir_path";
            $client_obj->apiset()->execute_command(
                command              => $dir_creat_cmd,
                'connectrec-timeout' => 60,
                'privilege-level'    => 'root'
            );

	    if($create_files_flag){
		my $file_num = $args{file_cnt};
                #inside each directory create files
                my $num_of_files = 1 + int(rand($file_num));
                while($num_of_files){
                    my $wldchar_index = int(rand(scalar @wld_char_array));
                    my $file_index = ($num_of_files % $num_constituents)+1;
                    my $file_name = "ldr_file".$file_index.$wld_char_array[$wldchar_index]."_".time()."_fgindex$file_index";
                    my $file_path = "$dir_path/$file_name";
                    my $count          = int(rand($args{'blk_repeat_cnt'}));#Control avg file size by varying blk_repeat_cnt
                    my $bs             = int(rand(1024));
                    my $file_creat_cmd = "dd if=/dev/urandom bs=$bs count=$count of=$file_path";
                    $client_obj->apiset()->execute_command(
                        command              => $file_creat_cmd,
                        'connectrec-timeout' => 60,
                        'privilege-level'    => 'root'
                    );
                    push(@filenames, $file_name);
                    push(@filepaths, $file_path);
                    $num_of_files--;
                }
	    }
        }catch NATE::BaseException with{
            my $ex = shift;
            NATE::BaseException->throw('Creation of Directories and files failed due to reason: '.$ex->text());
        };
    }

    #Now create remote and local hardlinks
    if($create_links_flag){
	if(!$create_files_flag){	#meaning, no files were created by this subroutine
	    $log->log('Since no files were created, no links will be created. Exiting.');
	    return;
	}
        my (@rhls, @lhls);
        while(scalar @rhls <= $num_links){
            foreach my $dir_index(1..$num_constituents){
                foreach my $i(1..scalar(@filenames)){
                    my $file_name = $filenames[$i];
                    my $file_path = $filepaths[$i];
                    my $ln_suffix;
                    if($file_name =~ /_fgindex$dir_index$/){
                        $ln_suffix = '_lhl_'.time();
                    }else{
                        $ln_suffix = '_rhl_'.time();
                    }
                    my $ln_path = $mount_pt.'/ldr_dir'.$dir_index.'_fgindex'.$dir_index."/$file_name$ln_suffix";
                    $log->log('Creating the hardlink: '.$ln_path);
                    my $link_create_cmd = "ln $file_path $ln_path";
                    $client_obj->apiset()->execute_command(
                        command              => $link_create_cmd,
                        'connectrec-timeout' => 60,
                        'privilege-level'    => 'root'
                    );
                    if($ln_suffix =~ /_rhl_/){
                        push(@rhls, $ln_path);
                    }else{
                        push(@lhls, $ln_path);
                    }
                }
            }
        }
        $log->log('Created the following links:'."\n".'Remote hardlinks -'."\n".Dumper(@rhls)."\n".'Local hardlinks -'."\n".Dumper(@lhls));
    }

    # set the placement back to auto (default)
    fg_set_placement(node=>$node, action => 'auto');
    return;
}
sub find_files_dir_links {
    my %args = validate_with(
        params      => \@_,
        allow_extra => 0,
        spec        => {
            mount_point => { type => SCALAR },
            client      => { type => OBJECT },
	    cmd_timeout	=> { type => SCALAR, default => '1800' },
	},
    );
    my $log = log_global();
    my $mount_pt = $args{mount_point};
    my $client_obj = $args{client};
    my $timeout = $args{cmd_timeout};

    #Find the files, directories and links present on the specified volume
    my $find_cmd = "find $mount_pt";
    my $find_output = $client_obj->apiset()->execute_command(
        command              => $find_cmd,
        'connectrec-timeout' => $timeout,
        'privilege-level'    => 'root'
    );

    #Segregate the directories, files and links
    my(@dirs, @files, @rhls, @lhls);
    my @find_op_arr = split /\n/, $find_output;
    foreach my $path(@find_op_arr){
        if($path =~ /\.snapshot/){
            next;       #ignore
	}elsif($path =~ /_rhl_/){
            push(@rhls, $path);
        }elsif($path =~ /_lhl_/){
            push(@lhls, $path);
        }elsif($path =~ /ldr_file/){
            push(@files, $path);
        }elsif($path =~ /ldr_dir/){
            push(@dirs, $path);
        }else{
            $log->log('Found an entry that doesnt belong to any category: '.$path);
        }
    }

    #$log->log('Found the following:'."\n".'Directories: '."\n".Dumper(@dirs)."\n".'Files: '."\n".Dumper(@files).
    #    "\n".'Remote hardlinks: '."\n".Dumper(@rhls)."\n".'Local hardlinks: '."\n".Dumper(@lhls));
    return (\@dirs, \@files, \@rhls, \@lhls);
}

sub fg_modify_hundreds_of_files_and_links {
    my %args = validate_with(
        params      => \@_,
        allow_extra => 0,
        spec        => {
            volume      => { type => OBJECT },
	    mount_point => { type => SCALAR },
            client      => { type => OBJECT },
	    op		=> { type => SCALAR },
	    cmd_timeout => { type => SCALAR, default => '1800'},
        },
    );
    my $log = log_global();
    my $mount_pt = $args{mount_point};
    my $vol = $args{volume};
    my $client_obj = $args{client};
    my $timeout = $args{cmd_timeout};
    my $op_type = (ref $args{op}) ? $args{op}->[0] : $args{op};
    my @op_type_arr;
    # There are cases where the scalar value might be a comma-separated list:
    if ($op_type =~ /,/) { @op_type_arr = split /,/, $op_type; }

    # Set the flexgroup placement to manual
    my $node = NACL::C::Node->new(name => $vol->get_local_node_name());
    fg_set_placement(node=>$node, action => 'manual');

    # to make it interesting for flexgroup make files so they will land on different constituents
    my @constituents = $vol->get_flexgroup_constituents();
    my $num_constituents = scalar (@constituents);

    foreach my $op(@op_type_arr){
        my ($dir_ref, $file_ref, $rhl_ref, $lhl_ref) = find_files_dir_links(
	    mount_point	=> $mount_pt,
	    client	=> $client_obj,
	    cmd_timeout => $timeout,
	);

        #Find the files, directories and links present on the specified volume
        my @files = @$file_ref;
        my @dirs = @$dir_ref;
        my @rhls = @$rhl_ref;
        my @lhls = @$lhl_ref;

	if($op eq 'create_file'){
	    $log->step('Create a few new files.');
	    my $createcnt = 1 + int(rand(10));
	    while($createcnt){
		my $file_index = ($createcnt % $num_constituents)+1;
		my $file_name= "ldr_file".$file_index."_".time()."_fgindex$file_index";
		foreach my $dir_path(@dirs){
		    my $file_path = "$dir_path/$file_name";
		    my $file_creat_cmd = "touch $file_path";
                    $client_obj->apiset()->execute_command(
                        command              => $file_creat_cmd,
                        'connectrec-timeout' => $timeout,
                        'privilege-level'    => 'root'
                    );
		}
		$createcnt--;
	    }
	}elsif($op eq 'create_lnk'){
	    $log->step('Create a few new links.');
	    my $createcnt = 1 + int(rand(10));
            #while($createcnt){
            foreach (1..$createcnt){
		foreach my $dir_index(1..$num_constituents){
		    foreach my $i(1..scalar(@files)){
			my $file_name = $files[$i];
			my $ln_suffix;
			if($file_name =~ /_fgindex$dir_index$/){
                            $ln_suffix = '_lhl_'.time();
                        }else{
                            $ln_suffix = '_rhl_'.time();
                        }
			my $ln_path = $file_name.$ln_suffix;
                        $log->log('Creating the hardlink: '.$ln_path);
                        my $link_create_cmd = "ln $file_name $ln_path";
                        $client_obj->apiset()->execute_command(
                            command              => $link_create_cmd,
                            'connectrec-timeout' => $timeout,
                            'privilege-level'    => 'root'
                        );
		    }
		}
	    }
	}elsif($op eq 'del_file'){
	    $log->step('Delete a few files and links.');
	    my $delcnt = 1 + int(rand(scalar @files));
	    foreach my $cnt(1..$delcnt){
            #Get a random file name and delete
                my $file_name = $files[int(rand(scalar @files))];
                my $del_cmd= "rm $file_name";
                $client_obj->apiset()->execute_command(
                    command              => $del_cmd,
		    'connectrec-timeout' => $timeout,
		    'privilege-level'    => 'root'
	        );
	    }
        }elsif($op eq 'del_dir'){
	    $log->step('Delete a few directories.');
	    my $delcnt = 1 + int(rand(scalar @dirs));
	    foreach my $cnt(1..$delcnt){
            #Get a random directory name and delete
                my $dir_name = $dirs[int(rand(scalar @dirs))];
	        my $del_cmd= "rm -rf $dir_name";
	        $client_obj->apiset()->execute_command(
                    command              => $del_cmd,
                    'connectrec-timeout' => $timeout,
                    'privilege-level'    => 'root'
                );
            }
	}elsif($op eq 'del_lnk'){
	    $log->step('Delete a few remote hard links.');
	    my $delcnt = 1 + int(rand(scalar @rhls));
	    foreach my $cnt(1..$delcnt){
            #Get a random links name and delete
                my $lnk_name = $rhls[int(rand(scalar @rhls))];
		my $del_cmd = "rm $lnk_name";
		$client_obj->apiset()->execute_command(
                    command              => $del_cmd,
                    'connectrec-timeout' => $timeout,
                    'privilege-level'    => 'root'
                );
            }
	    $log->step('Delete a few local hard links.');
            $delcnt = 1 + int(rand(scalar @lhls));
	    foreach my $cnt(1..$delcnt){
            #Get a random links name and delete
                my $lnk_name = $lhls[int(rand(scalar @lhls))];
		my $del_cmd = "rm $lnk_name";
                $client_obj->apiset()->execute_command(
                    command              => $del_cmd,
                    'connectrec-timeout' => $timeout,
                    'privilege-level'    => 'root'
                );
            }
        }elsif($op eq 'rn_file'){
	    $log->step('Rename a few files.');
	    my $rn_cnt = 1 + int(rand(scalar @files));
	    foreach my $cnt(1..$rn_cnt){
	        #Get a random file name and rename
	        my $file_name = $files[int(rand(scalar @files))];
                my $mv_cmd= "mv $file_name renamed_$file_name";
                $client_obj->apiset()->execute_command(
                    command              => $mv_cmd,
                    'connectrec-timeout' => $timeout,
                    'privilege-level'    => 'root'
                );
            }
        }elsif($op eq 'rn_dir'){
	    $log->step('Rename a few dirs.');
	    my $rn_cnt = 1 + int(rand(scalar @dirs));
	    foreach my $cnt(1..$rn_cnt){
	        #Get a random file name and rename
	        my $dir_name = $dirs[int(rand(scalar @dirs))];
                my $mv_cmd= "mv $dir_name renamed_$dir_name";
	        $client_obj->apiset()->execute_command(
                    command              => $mv_cmd,
                    'connectrec-timeout' => $timeout,
                    'privilege-level'    => 'root'
                );
            }
        }elsif($op eq 'rn_lnk'){
	    $log->step('Rename a few remote hardlinks.');
	    my $rn_cnt = 1 + int(rand(scalar @rhls));
            foreach my $cnt(1..$rn_cnt){
            #Get a random links name and delete
                my $lnk_name = $rhls[int(rand(scalar @rhls))];
                my $mv_cmd = "mv $lnk_name renamed_$lnk_name";
                $client_obj->apiset()->execute_command(
                    command              => $mv_cmd,
                    'connectrec-timeout' => $timeout,
                    'privilege-level'    => 'root'
                );
            }
            $log->step('Rename a few local hard links.');
            my $delcnt = 1 + int(rand(scalar @lhls));
            foreach my $cnt(1..$delcnt){
                #Get a random links name and rename
                my $lnk_name = $lhls[int(rand(scalar @lhls))];
                my $mv_cmd = "mv $lnk_name rename_$lnk_name";
                $client_obj->apiset()->execute_command(
                    command              => $mv_cmd,
                    'connectrec-timeout' => $timeout,
                    'privilege-level'    => 'root'
                );
	    }
	}
    }

    # Set the flexgroup placement back to auto
    $node = NACL::C::Node->new(name => $vol->get_local_node_name());
    fg_set_placement(node=>$node, action => 'auto');

    return;
}

sub create_delete_fifo {
    my %args  = validate_with  (
        params =>  \@_,
        spec   =>  {
                vol_objects   =>   { type => ARRAYREF },
                client_obj    =>   { type => OBJECT },
                count         =>   { type => SCALAR, default => 20 },
                nfsversion    =>   { type => SCALAR, default => '3'},
                fifo_ds       =>   { type => HASHREF, optional => 1},
                op            =>   { type => SCALAR, regex => qr/^(create|delete)$/, default => 'create' }
        },
    );

    my $vols        = $args{vol_objects};
    my $client_obj  = $args{client_obj};
    my $cli_api     = $client_obj->apiset();
    my $vers        = $args{nfsversion};
    my $log         = log_global();
    my $cnt         = $args{count};
    my $op          = $args{op};
    my $fifo_ds     = $args{fifo_ds};
    my @modes       = ('0777','0770','0077','0666');
    my $ret;
    $log->enter();

    foreach my $v (@{ $vols }) {
        $log->comment('Mounting volume: ' . $v->volume());
        my $mount = NACL::STask::Client::Mount->create(
                        command_interface    =>    $client_obj,
                        server_command_interface =>  $v->command_interface,
                        vserver                  =>  $v->vserver,
                        volume                   =>  $v,
                        vers                     =>  $vers,);
        if ($op =~ /create/) {

            my $mount_pt = $mount->mount_point();

            for (my $i = 0; $i<$cnt; $i++) {
                my $fifo_path = $mount_pt . '/fifo_' . $i . '_' . time();
                my $m = shift @modes;
                push @modes, $m;
                $log->comment('Creating fifo ' . $fifo_path . ' under vol: ' . $v->volume() .
                              ' using mode: ' . $m);
                $cli_api->mkfifo( '-m'              =>  $m,
                                  'privilege-level' => 'root',
                                  namepipe          => $fifo_path);
                my $vname = $v->volume();
                push @{ $ret->{$vname} }, $fifo_path;

            }
        } elsif ($op =~ /delete/) {
            if (!defined $fifo_ds) {
                croak('Please provide a list of fifo\'s that you want to delete');
            }
            my $vname = $v->volume();
            my @fifos = @{$fifo_ds->{$vname}};
            if ($cnt > scalar(@fifos)) {
                $cnt = scalar(@fifos);
            }
            my @target_fifos  = splice @fifos, 0, $cnt;
            $log->comment('Fifos that will be deleted are: ' . Dumper(\@target_fifos));
            foreach my $f (@target_fifos) {
                $cli_api->rm ('privilege-level'    => 'root',
                              paths                => $f,
                              'connectrec-timeout' => 600,);
            }

            $ret->{$vname} = \@fifos;
        }
    }

    $log->exit();
    return $ret;
}

sub create_delete_devices {
    my %args = validate_with (
        params   =>   \@_,
        spec     => {
                vol_objects     =>    { type => ARRAYREF },
                client_obj      =>    { type => OBJECT },
                count           =>    { type => SCALAR, default => 10 },
                nfsversion      =>    { type => SCALAR, default => '3'},
                device_ds       =>    { type => HASHREF, optional => 1},
                op              =>    { type => SCALAR, regex => qr/^(create|delete)$/, default => 'create'}},
    );

    my $vols        = $args{vol_objects};
    my $client_obj  = $args{client_obj};
    my $cli_api     = $client_obj->apiset();
    my $vers        = $args{nfsversion};
    my $log         = log_global();
    my $cnt         = $args{count};
    my $op          = $args{op};
    my $dev_ds      = $args{device_ds};
    my @modes       = ('0777','0770','0077','0666');
    my @types       = ('b', 'c', 'u');
    my $ret;
    $log->enter();

    foreach my $v (@{ $vols }) {
        $log->comment('Mounting volume: ' . $v->volume());
        my $mount = NACL::STask::Client::Mount->create(
                        command_interface    =>    $client_obj,
                        server_command_interface =>  $v->command_interface,
                        vserver                  =>  $v->vserver,
                        volume                   =>  $v,
                        vers                     =>  $vers,);
        if ($op =~ /create/) {
            my $mount_pt = $mount->mount_point();
            foreach my $t (@types) {
                my %opts;

                for (my $j = 0; $j < $cnt; $j++) {
                    my $dev_file = $mount_pt . '/dev_' . $t . $j  . '_' . time();
                    my $m = shift @modes;
                    push @modes, $m;
                    $log->comment('Creating device file of type: ' . $t .
                                  ' under vol: ' . $v->volume() .
                                  ' using mode: '  . $m);
                    my $cmd = "sudo mknod -m $m $dev_file $t 10 2";
                    $cli_api->execute_raw_command (
                                command => $cmd);

                    my $vname = $v->volume();
                    push @{ $ret->{$vname} }, $dev_file;
                }
            }
        } elsif ($op =~ /delete/) {
            if (!defined $dev_ds) {
                croak('Please provide a list of device files that you want to delete');
            }

            my $vname = $v->volume();
            my @devs = @{ $dev_ds->{$vname}};
            if ($cnt > scalar(@devs)) {
                $cnt = scalar(@devs);
            }
            my @target_devs = splice @devs, 0, $cnt;
            $log->comment('Device files that will be deleted are: ' . Dumper(\@target_devs));
            foreach my $f (@target_devs) {
                $cli_api->rm ('privilege-level'    => 'root',
                              paths                => $f,
                              'connectrec-timeout' => 600,);
            }

            $ret->{$vname} = \@devs;
        }
    }

    $log->exit();
    return $ret;
}

=head2 LSA_file_creation_IO_DT

    This method is to create the file and it will perform punch hole .

=item required

=item required

    (required) volume - OBJECT
    (required) client - OBJECT  # example truncate_file_size => 45 or truncate_file_size => 20
    (required) node - OBJECT
    (required) file_name - SCALAR.

=item optional

    file_size      - SCALAR. (by deafult 1gb)
    dt_limit_bytes - SCALAR. (by deafult 250m)

=back

=cut

sub LSA_file_creation_IO_DT {
    my %args = validate_with(
        params => \@_,
        spec   => {
            client    => { type => OBJECT, optional => 0 },
            volume    => { type => OBJECT },
            file_size => { type => SCALAR , optional => 1 , default => '1g'},
            node      => { type => OBJECT },
            dt_limit_bytes => {type => SCALAR , optional => 1 , default => '250m'},
            file_name     =>  {type => SCALAR , optional => 0 }
        }
    );
    my $Log = log_global();
    $Log->enter();

    my $vol_obj     = $args{volume};
    my $client_obj  = $args{client};
    my $client_host = $client_obj->name();
    my $node_src    = $args{node};
    my $file_size   = $args{file_size};
    my $dt_limit_bytes = $args{dt_limit_bytes};
    my $file_name   = $args{file_name};

    #file creation
    my $file = '/vol/' . $vol_obj->volume() . "/$file_name";
    $Log->step( 'Create file '
            . $file
            . ' of size: '
            . $file_size
            . ' on SRC volume  '
            . $vol_obj->volume() );
    my @file_obj = NACL::STask::VolumeFile->create(
        command_interface => $node_src,
        vserver           => $vol_obj->vserver(),
        size              => $file_size,
        path              => $file,
        filecount         => 1,
    );

    #NFS Mount
    $Log->step('Mount src volume');
    my $Mount = NACL::STask::Client::Mount->create(
        command_interface  => $client_obj,
        volume             => $vol_obj,
        nacltask_if_exists => 'purge',
        nacltask_verify    => 1,
        mountopts          => 'nfsvers=3',
    );

    my $NFSmount_point = $Mount->mount_point();

    #DT_IO_TRAFFIC
    my $host1         = host($client_host);
    my $client_apiset = NACL::APISet->new(
        hostobj   => $host1,
        category  => "Host",
        interface => "CLI",
        set       => "Linux",
        connid    => "ssh",
        conntype  => "ssh-persist",
    );

    $Log->step("set permission chmod 777 for file $NFSmount_point/$file_name");
    my $command = "chmod 777 $NFSmount_point/$file_name";
    $client_apiset->execute_command( command => $command );

    $Log->step('NFS DT Traffic');
    my $cmd
        = "/usr/software/test/bin/dt of=$NFSmount_point/$file_name dispose=keep bs=random limit=$dt_limit_bytes flags=direct procs=1 iotype=sequential pf=/usr/software/test/noarch/dtdata/pattern_all enable=syslog enable=compare";
    $client_apiset->execute_command(
        command              => $cmd,
        'connectrec-timeout' => 80000
    );

    #punch holes
    $Log->step('Punching holes');
    _punch_holes_in_files(
        file_obj             => \@file_obj,
        holes_in_mid         => 200,
        hole_size_in_mid     => '4k',
        offset_between_holes => '0k',
        holes_in_end         => 0
    );
}

=head2 LSA_file_truncate

    This method is to truncate the file.

=item required

=item required

    (required) volume - OBJECT
    (required) truncate_file_size - SCALAR  # example truncate_file_size => 45 or truncate_file_size => 20
    (required) node - OBJECT
    (required) file_name - SCALAR.

=back

=cut

sub LSA_file_truncate {
    my %args = validate_with(
        params => \@_,
        spec   => {
            volume             => { type => OBJECT },
            truncate_file_size => { type => SCALAR },
            node               => { type => OBJECT },
            file_name          => { type => SCALAR , optional => 0 },
        }
    );
    my $Log = log_global();
    $Log->enter();
    my $volume             = $args{volume};
    my $truncate_file_size = $args{truncate_file_size};
    my $file               = $args{file_name };
    my $node               = $args{node};

    my $volumeFile;
    my $actual_file_size;
    my $volume_name = $volume->volume();
    try {
        $volumeFile = NACL::CS::VolumeFile->fetch(
                command_interface => $node,
                filter            => {
                    vserver => $volume->{vserver},
                    path    => "/vol/$volume_name/$file",
                },
            );
            $actual_file_size = $volumeFile->{'file_size'};
    } ## end try
    catch NACL::Exceptions::NoElementsFound with {
            $Log->error("The Given Input File not found.");
            $Log->exit();
    };

    my $size_to = int(($truncate_file_size / 100) * $actual_file_size);

    my $response = $node->apiset->system_node_run(
        command =>
            "\"priv set -q test; file truncate /vol/$volume_name/$file $size_to \"",
        node => $node->node,
    );

    my $slist_output = NACL::C::Slist->slist(
                        command_interface => $node,
                        path              => "/vol/$volume_name/$file",
                    )->get_parsed_output();
    my $file_size = $slist_output->[0]->{file_size};
    if($size_to == $file_size) {
        $Log->log('truncate passed on file ' . $file );
    } else {
        $Log->warn('truncate failed on file ' . $file);

    }

    return $size_to;
} ## end sub LSA_file_truncate

=head2 populate_lsa_dataset

    This method is to Populate LSA DATA.

=item required

    (required) volume - OBJECT
    (required) truncate_file_size - SCALAR  # example truncate_file_size => 45 or truncate_file_size => 20
    (required) node - OBJECT

=item optional

    file_name      - SCALAR.
    file_size      - SCALAR. (by deafult 1gb)
    fiji_sleep     - SCALAR. (by deafult 300)
    dt_limit_bytes - SCALAR. (by deafult 250m)
    client         - OBJECT.

=item Example
    Without file name it will create a new file
    populate_lsa_dataset (
            volume => $vol_src_1,
            truncate_file_size => 45,
            node => $Nodes[0],
            fiji_sleep => '400',
            client => $Client);

    with file name, it uses existing file (Note : File should have Punch holes for LSA functionality)
    populate_lsa_dataset (
            volume => $vol_src_1,
            truncate_file_size => 45,
            node => $Nodes[0],
            file_name => 'file0'
            fiji_sleep => '400',
            client => $Client);

=back

=cut

sub populate_lsa_dataset {
    my %args = validate_with(
        params => \@_,
        spec   => {
            volume             => { type => OBJECT },
            truncate_file_size => { type => SCALAR },
            node               => { type => OBJECT },
            file_name          => { type => SCALAR, optional => 1 },
            file_size => { type => SCALAR, optional => 1, default => '1gb' },
            fiji_sleep => { type => SCALAR, optional => 1, default => '300' },
            dt_limit_bytes =>
                { type => SCALAR, optional => 1, default => '250m' },
            client => { type => OBJECT , optional => 1 },
        }
    );

    my $Log = log_global();
    $Log->enter();

    my $volume             = $args{volume};
    my $truncate_file_size = $args{truncate_file_size};
    my $file               = $args{file_name};
    my $node               = $args{node};
    my $fiji_sleep         = $args{fiji_sleep};
    my $dt_limit_bytes     = $args{dt_limit_bytes};
    my $client             = $args{client};
    my $file_size          = $args{file_size};

    my $volume_name = $volume->volume();
    my $volumeFile;
    my $actual_file_size;
    if ( defined($file) ) {
        try {
            $volumeFile = NACL::CS::VolumeFile->fetch(
                command_interface => $node,
                filter            => {
                    vserver => $volume->{vserver},
                    path    => "/vol/$volume_name/$file",
                },
            );
            $actual_file_size = $volumeFile->{'file_size'};
        } ## end try
        catch NACL::Exceptions::NoElementsFound with {
            $Log->error("The Given Input File not found.");
            $Log->exit();
        };
    } else {
        $Log->step( 'Creating a File file0 of size  ' . $file_size );
        $file = 'file0';
        LSA_file_creation_IO_DT(
            client         => $client,
            volume         => $volume,
            file_size      => $file_size,
            node           => $node,
            dt_limit_bytes => $dt_limit_bytes,
            file_name      => $file
        );

        try {
            $volumeFile = NACL::CS::VolumeFile->fetch(
                command_interface => $node,
                filter            => {
                    vserver => $volume->{vserver},
                    path    => "/vol/$volume_name/$file",
                },
            );
            $actual_file_size = $volumeFile->{'file_size'};
        } ## end try
        catch NACL::Exceptions::NoElementsFound with {
            $Log->error("File Creation Failed.");
            $Log->exit();
        };
    } ## end else [ if ( defined($file_name...))]

    my $initial_count = lsa_count_verify( node => $node );

    my $slist_output = NACL::C::Slist->slist(
        command_interface => $node,
        path              => "/vol/$volume_name/$file",
    )->get_parsed_output();
    my $file_inode = $slist_output->[0]->{inode};

    $Log->step(
        'setting the fiji wafl.lsa.punch.holes.worker.delay on volume '
            . $volume_name );
    my $fiji_delay = fiji_wafl_lsa_punch_holes_worker_delay(
        node         => $node,
        'sleep'      => $fiji_sleep,
        'file_inode' => $file_inode,
        volume       => $volume_name,
    );

    $Log->step('fiji dump');
    $fiji_delay->wait_for_fiji(
        slot                     => $fiji_delay->slot(),
        'nacltask_wait_interval' => 60,
    );

    #truncate operation
    $Log->step('file truncate');
    my $size_to_truncate = LSA_file_truncate(
        volume             => $volume,
        truncate_file_size => $truncate_file_size,
        node               => $node,
        file_name          => $file,
    );

    $Log->step('wait for fiji to hit');
    $fiji_delay->wait_for_fiji(
        slot                        => $fiji_delay->slot(),
        'nacltask_wait_interval'    => 1200,
        'nacltask_fiji_operation'   => 'hit',
        'nacltask_polling_interval' => 5,
    );

    $Log->step( 'verfying lsa count on volume ' . $volume_name );
    my $lsa_count = lsa_count_verify( node => $node );

    if ( $lsa_count > $initial_count ) {
        $Log->log('LSA is applied as expected');
    }

    $Log->step( 'verfying lsa is active on volume ' . $volume_name );
    lsa_verify_in_progress(
        node     => $node,
        vol_name => $volume_name
    );

    my $snap1 = "lsa_snapshot_$$";

    $Log->step( 'Cut a snapshot '
            . $snap1
            . ' on source volume '
            . $volume->volume() );
    my $SNAP1 = create_snapshot( $volume, $snap1 );

    $Log->step( 'LSA validation using command testsetattr on volume '
            . $volume_name );
    my $result = lsa_validation(
        'logical_file_size'  => $size_to_truncate,
        'physical_file_size' => $actual_file_size,
        'node'               => $node,
        'volume_name'        => $volume_name,
        'file_name'          => $file,
    );

    my %hash;
    $hash{'snapshot_obj'} = $SNAP1;
    $hash{'file_obj'}     = $volumeFile;

    return %hash;
} ## end sub populate_lsa_dataset

1;
