#
# Copyright (c) 2015 NetApp, Inc., All Rights Reserved.
#   Any other use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

package NACL::AppBuilder::Xgen::Config::DMG::Base_YML;

use strict;
use warnings;

use Tie::IxHash;
use Params::Validate qw(validate_with :types);
use Data::Dumper;

use parent qw( NACL::AppBuilder::Xgen::Config::ONTAP );
use NATE::Log qw( log_global );
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();

use NATE::ParamSet qw(param_global);

use NATE::BaseException qw(:try);
use NACL::AppBuilder::Xgen::Resource::ClusterPeer;
use NACL::AppBuilder::Xgen::Resource::VserverPeer;
use NACL::AppBuilder::Xgen::Resource::Vserver;
use NACL::AppBuilder::Xgen::Resource::Volume;
use NACL::AppBuilder::Xgen::Resource::Volume::Efficiency;
use NACL::AppBuilder::Xgen::Resource::Volume::EfficiencyPolicy;
use NACL::AppBuilder::Xgen::Resource::Lun;
use NACL::AppBuilder::Xgen::Resource::LunMap;
use NACL::AppBuilder::Xgen::Resource::File;
use NACL::AppBuilder::Xgen::Resource::Iscsi;
use NACL::AppBuilder::Xgen::Resource::Igroup;
use NACL::AppBuilder::Xgen::Resource::Igroup::Initiator;
use NACL::AppBuilder::Xgen::Resource::Lif;
use NACL::AppBuilder::Xgen::Resource::NetworkRoute;
use NACL::AppBuilder::Xgen::Resource::Nfs;
use NACL::AppBuilder::Xgen::Resource::Snapmirror;
use NACL::AppBuilder::Xgen::Resource::SnapmirrorInitialize;
use NACL::AppBuilder::Xgen::Resource::SnapmirrorPolicy;
use NACL::AppBuilder::Xgen::Resource::SnapmirrorPolicyRule;
use NACL::AppBuilder::Xgen::Resource::SnapshotPolicy;
use NACL::AppBuilder::Xgen::Resource::JobScheduleCron;
use NACL::Service::Xgen::Util qw(convert_percent_layout_to_count_layout
                                 get_next_layout_element
                                 dumper
                                 positive_filter);
use NACL::C::AVANode;
use dp::xplatform::lib::SMAgent;

use NATE::BaseException;
use POSIX qw(ceil floor);
use Hostrec;

# These are the constant values that determine what happens in the config.
# This constant controls the number of aggregates per node for the config.
use constant AGGREGATES_PER_NODE => 0;
# This constant controls the total number of vservers in the config.
use constant TOTAL_VSERVER_COUNT => 4;
# This constant controls the total number of volumes in the config.
use constant TOTAL_VOLUME_COUNT => 88;
# This constant controls the total number of luns in the confg.
use constant TOTAL_LUN_COUNT => 176;
# This constant controls the total number of files in the config.
use constant TOTAL_FILE_COUNT => 176;
# This constant controls the total number of snapmirrors in the config.
use constant TOTAL_SNAPMIRROR_COUNT => 44;
# This constant controls the number of disks per aggregate.
use constant NUMBER_DISKS_PER_AGGREGATE => 3;
# This constant controls the size of all the volumes in the config.
use constant VOLUME_SIZE => "500m";
# This constant controls the size of all the luns in the config.
use constant LUN_SIZE => "35m";
# This constant controls the size of all the files in the config.
use constant FILE_SIZE => "20m";
# This is used to control the percentage of volumes with luns
use constant PERCENTAGE_OF_VOLUMES_WITH_LUNS => 100;
# This is used to control the percentage of volumes with files.
use constant PERCENTAGE_OF_VOLUMES_WITH_FILES => 0;
# This constant will control the number of cascade configs
use constant TOTAL_CASCADE_CONFIG_COUNT => 2;
# This constant will control the layouts of the cascade configs based on percents.
use constant CASCADE_TYPE_PERCENT_LAYOUT => {'ADR;VAULT'       => 50,
                                             'ADR;ADR;VAULT'   => 50};
# This constant will control the number of fanin configs
use constant TOTAL_FANIN_CONFIG_COUNT => 2;
# This constant will control the layouts of the fanin configs based on percents.
use constant FANIN_TYPE_PERCENT_LAYOUT => {'ADR;VAULT'       => 50,
                                           'ADR;ADR;VAULT'   => 50};
# This constant will control the number of fanin configs
use constant TOTAL_FANOUT_CONFIG_COUNT => 2;
# This constant will control the layouts of the fanin configs based on percents.
use constant FANOUT_TYPE_PERCENT_LAYOUT => {'ADR;VAULT'       => 50,
                                            'ADR;ADR;VAULT'   => 50};
# This constant will control the number of fanin configs
use constant TOTAL_BIDIRECTIONAL_CONFIG_COUNT => 2;
# This constant will control the layouts of the fanin configs based on percents.
use constant BIDIRECTIONAL_TYPE_PERCENT_LAYOUT => {'ADR;ADR'       => 50,
                                                   'VAULT;VAULT'   => 50};
# This constant controls the percentage of all Snapmirrors that will
# be each type of Snapmirror relationship.
use constant UNIDIRECTIONAL_SNAPMIRROR_TYPE_PERCENT_LAYOUT => {SMGR                            => 20,
                                                              ADR                             => 4,
                                                              ADR_NC                          => 4,
                                                              VAULT                           => 4,
                                                              VAULT_NC                        => 4,
                                                              VAULT_SRCSE                     => 4,
                                                              VAULT_SRCCOMPRESSION            => 4,
                                                              VAULT_ADDDEDUP                  => 4,
                                                              VAULT_LRE                       => 4,
                                                              LRSE_DR_MIRROR_LATEST           => 4,
                                                              LRSE_DR_MIRROR_ALL              => 4,
                                                              LRSE_DR_MIRROR_AND_VAULT        => 4,
                                                              LRSE_DR_MIRROR_LATEST_LRE       => 4,
                                                              LRSE_DR_MIRROR_ALL_LRE          => 4,
                                                              LRSE_DR_MIRROR_AND_VAULT_LRE    => 4,
                                                              LRSE_DR_MIRROR_LATEST_NC        => 4,
                                                              LRSE_DR_MIRROR_ALL_NC           => 4,
                                                              LRSE_DR_MIRROR_AND_VAULT_NC     => 4,
                                                              LRSE_DR_MIRROR_LATEST_NC_LRE    => 4,
                                                              LRSE_DR_MIRROR_ALL_NC_LRE       => 4,
                                                              LRSE_DR_MIRROR_AND_VAULT_NC_LRE => 4,};
# This constant controls the percentage of SMGR relationships that
# will be lun versus file.
use constant SMGR_CG_ITEM_PERCENT_LAYOUT => {'lun'   => 100,
                                             'file'  => 0};

# This constant controls the percentage of luns or files that are
# used in the SMGR relationships
use constant PERCENTAGE_OF_LUNS_FILES_USED_FOR_CG_ITEMS => 50;

# This constant controls the percentage of volumes that will have
# a vm-align-suffix and what the value will be.
# :NONE: means that the volumes will not have a vm align suffix.
use constant VOLUME_VM_ALIGN_SUFFIX_PERCENT_LAYOUT => {':NONE:' => 100,
                                                       '.vmdk'  => 0};
# This constant controls the percentage of volumes that will have
# a vm-align-sector and what the value will be.
# This percentage will only affect volumes that have a
# vm-align-suffix.
use constant VOLUME_VM_ALIGN_SECTOR_PERCENT_LAYOUT => {'1' => 50,
                                                       '2' => 50};
# The Following 2 constants control the minimum and maximum
# percent of files in a vm aligned volume that have the correct
# suffix. If the values are equal then the value is to percent
# that all volumes will recieve. If the values are not equal
# then the percent will be chosen randomly between the
# minimum and the maximum.  The random_seed if fixed by default.
# A user can pass in a random value for 'random_seed' in args to
# the config to set the random seed to a random value. This means
# that by default a config with the same constant values will always
# build the same thing unless random_seed is passed in.
use constant MINIMUM_PERCENT_VM_ALIGNED_FILES => 50;
use constant MAXIMUM_PERCENT_VM_ALIGNED_FILES => 50;
# This constant is used to turn on and off setup and cleanup for
# snapmirrors in EnGenX for this configuration. The snapmirror
# resources objects are still created but will not be created
# or deleted in ontap by EnGenX
use constant SETUP_CLEANUP_SNAPMIRRORS => 1;
# This constant is used to turn on and off snapmirror initialization.
use constant INITIALIZE_SNAPMIRRORS => 1;
# This constant is used to start or not start volume_efficiency
# Volume Efficiency will not start regardless for destinations of SMs.
use constant START_VOLUME_EFFICIENCY => 1;

#aggr name format: node_name, node_id, aggr_only
use constant AGGR_NAME_FORMAT => 'aggr_only' ;

#volume name format: vserver (append vserver name), volume_only
use constant VOLUME_NAME_FORMAT => 'volume_only' ;
###############################################################################
# Description: The parent generates aggregates based on the override method
#              generate_aggregate_resources.  Calls modular functions to
#              generate resources.
# Input: Arguments to pass to the module methods and other Xgen controls.
# Output: NONE
# Exceptions: None thrown directly by this method.
###############################################################################
sub build {
    my ($self, %args) = @_;

    #save xgen params
    $self->xgen_config_params($args{xgen_params} || {});
    my $resource_hash = delete $args{resources};

    $Log->comment("NODES:" . param_global()->get("NODES"));
    $Log->comment("args xgen:" . dumper \%args);
    $Log->comment("keys passed in to xgen dump:" . dumper $self->xgen_config_params());

    # We only want to do this if something did not pass in any Ethernet fabric definitions.
    # This fabric will include all the ips from all nodes and not sort specific to clusters.
    # This means that IPs will not be owned by a specific cluster and can be used by any cluster.
    unless (scalar keys %{$args{eth_fabric_data}}) {
        $Log->comment("Generating Ethernet Fabric data from Natehost files.");
        $args{eth_fabric_data} = $self->generate_eth_fabric_data_from_natehost_files(%args);
    }

    $Log->comment("Generating Parent Configuration.");
    $self->SUPER::build(%args);

    my $iscsi_clients = $args{iscsi_clients};
    my $auto_nfs_lifs = $self->xgen_config_params()->{AUTO_NFS_LIF_CREATE} || 0;

    #TODO: Add it to the the 'loop glue'
    #
if (0) {
    $Log->step("Generating Cluster Peers.");
    # Call to generate cluster peer resources.
    #$self->generate_cluster_peer_resources(%args);
    $self->generate_one_cluster_peer(src_cluster => 'dpgvlt32700506',
                                    dst_cluster => 'dpgvlt32700708');

    $Log->step("Generating Vservers.");
    # Call to generate vserver_peer resources.
    $self->generate_vserver_resources(%args);

    $Log->step("Generating ISCSI Lifs.");
    # Call to generate iscsi lif resources.
    $self->generate_iscsi_lif_resources(%args);

    $Log->step("Generating NFS Lifs.");
    # Call to generate nfs lif resources.
    $self->generate_nfs_lif_resources(%args);

    $Log->step("Generating Igroups.");
    # Call to genereate iscsi service resources.
    $self->generate_igroup_resources(%args);

    $Log->step("Generating Initiators.");
    # Call to genereate iscsi service resources.
    $self->generate_igroup_initiator_resources(%args);

    $Log->step("Generating ISCSI Services.");
    # Call to genereate iscsi service resources.
    $self->generate_iscsi_services_resources(%args);

    $Log->step("Generating Volumes.");
    # Call to generate volume resources.
    $self->generate_volume_resources(%args);

    $Log->step("Generating Snapmirror Resources.");
    # Call to generate snapmirror resources.
    #$self->generate_snapmirror_resources(%args);

    $Log->step("Generating Luns.");
    # Call to generate lun resources.
    $self->generate_lun_resources(%args);

    $Log->step("Generating Lun Maps.");
    # Call to generate lun resources.
    $self->generate_lun_map_resources(%args);

    $Log->step("Generating Files.");
    # Call to generate file resources.
    $self->generate_file_resources(%args);

    $Log->step("Adding CG Items to SMGR Snapmirrors.");
    # Call to add_cg_items_to_SMGR_snapmirror_resources
    # This must be after luns and files are created
    $self->add_cg_items_to_SMGR_snapmirror_resources(%args);

    $Log->step("Generating Volume Efficiency Options.");
    # Call to generate_volume_efficiency_resources
    # This must be after snapmirrors are created since
    # they control the volumes we want efficiency on.
    $self->generate_volume_efficiency_resources(%args);

    $Log->step("Generating Vserver Peers.");
    # Call to generate vserver_peer resources.
    $self->generate_vserver_peer_resources(%args);
}

if (01) {
    my @resource_list = ('aggregate', 'vserver',
        'vserver_nfs',
        'vserver_peer',
        'snapshot_policy',
        'snapmirror_policy', 'snapmirror_policy_rule', 'job_schedule',
        'iscsi_stuff', #original base.pm code
        'volume', 'lun',
		'btrfs',
		'btrfs_volume',
		'snapmirror',
        'volume_efficiency_policy',
        'volume_efficiency',
    );

    foreach my $key (@resource_list) {
        my $entry = $resource_hash->{$key}; #get the aggregate key
		$Log->log("KEY:$key");
        foreach my $childkey (keys %{ $entry }) { #get all the aggrs under the aggregate key
            my $child_entry = $entry->{$childkey};
            if ($key =~ /aggregate/) {
                # Call to generate ONE aggregate resource
                $Log->log('Generating ONE aggregate resource.');
                $self->generate_one_aggregate_resource(%{$child_entry});
            } elsif ($key =~ /^volume$/) {
                $Log->step('Generating ONE volume resource.');
                $self->generate_one_volume_resource(%{$child_entry});
            } elsif ($key =~ /lun/) {
                #Call to generate ONE lun resource
                $Log->step('Generating ONE lun resource.');
                $self->generate_one_lun_resource (%{$child_entry});
            } elsif ($key =~ /^vserver$/) {
                # Call to generate ONE vserver resource
                $Log->step('Generating ONE vserver resource.');
                $self->generate_one_vserver_resource(%{$child_entry});
            } elsif ($key =~ /^snapmirror$/) {
                # Call to generate ONE snapmirror resource
                $Log->step('Generating ONE snapmirror resource.');
                $self->generate_one_snapmirror_resource(
                    %{$child_entry},
                    xgen_metadata_args  => {
                        sm_config => 'unidirectional',
                        #sm_config_layout => $config_layout,
                        #sm_config_id => $$config_id);
                    },
                );

            } elsif ($key =~ /^snapmirror_policy$/) {
                # Call to generate ONE snapmirror policy resource
                $Log->step('Generating ONE snapmirror policy resource.');

                my $vserver_sm_policy_hash = {}; # A mapping of vservers to snapmirror policies.
                $self->generate_one_snapmirror_policy_resource (
                    %{$child_entry},
                );
            } elsif ($key =~ /snapmirror_policy_rule/) {
                # Call to generate ONE snapmirror policy rule resource
                $Log->step('Generating ONE snapmirror policy rule resource.');

                $self->generate_one_snapmirror_policy_rule_resource(
                    %{$child_entry},
                );
            } elsif ($key =~ /snapshot_policy/) {
                # Call to generate ONE snapshot policy  resource
                $Log->step('Generating ONE snapshot policy resource.');

                my $vserver_snapshot_policy_hash = {}; # A mapping of vservers to snapshot policies.
                $self->generate_one_snapshot_policy_resource(
                    %{$child_entry},
                    vserver_snapshot_policy_hash => $vserver_snapshot_policy_hash,
                );
            } elsif ($key =~ /job_schedule/) {
                # Call to generate ONE job schedule resource
                $Log->step('Generating ONE job schedule resource.');

                my $cluster_job_schedule_cron_hash = {}; # A mapping of clusters to job schedules.
                $self->generate_one_job_schedule_cron_resource (
                    %{$child_entry},
                    cluster_job_schedule_cron_hash => $cluster_job_schedule_cron_hash,
                    #cluster => 'dpgvlt32700708',
                    #name => 'archjob',
                    #minute => 5,
                );
            } elsif ($key =~ /^volume_efficiency$/) {
                # Call to generate ONE volume efficiency resource
                $Log->step('Generating ONE volume efficiency resource.');
                my $vserver_volume_efficiency_policy_hash = {};

                $self->generate_one_volume_efficiency_resource(
                    %{$child_entry},
                    vserver_volume_efficiency_policy_hash => $vserver_volume_efficiency_policy_hash,
                    #vserver => 'vs1',
                    #volume => 'vol2',
                    #policy => 'foo',
                    #start => '0',
                );
            } elsif ($key =~ /^volume_efficiency_policy$/) {
                # Call to generate ONE volume efficiency policy resource
                $Log->step('Generating ONE volume efficiency policy resource.');
                my $vserver_volume_efficiency_policy_hash = {};


                $self->generate_one_volume_efficiency_policy_resource(
                    %{$child_entry},
                    vserver_volume_efficiency_policy_hash => $vserver_volume_efficiency_policy_hash,
                    #vserver => 'vs1',
                    #policy => 'foo',
                    #schedule => 'daily',
                    #duration => '100',
                );
            } elsif ($key =~ /vserver_peer/) {
                # Call to generate ONE vserver peer resource
                $Log->step('Generating ONE vserver peer resource.');
                $self->generate_one_vserver_peer_resource(%{$child_entry});
            } elsif ($key =~ /vserver_nfs/) {
                # Call to generate ONE vserver nfs resource
                $Log->step('Generating ONE vserver nfs resource.');
                $self->generate_one_vserver_nfs_resource(%{$child_entry});
            } elsif ($key =~ /^btrfs$/) {
                # Call to generate ONE btrfs  resource
                $Log->step('Generating ONE btrfs resource.');
                $self->generate_one_btrfs_resource(%{$child_entry});
            } elsif ($key =~ /^btrfs_volume$/) {
                # Call to generate ONE btrfs_volume  resource
                $Log->step('Generating ONE btrfs_volume resource.');
                $self->generate_one_btrfs_volume_resource(%{$child_entry});
            } elsif ($key =~ /iscsi_stuff/) {
             #leaving it empty for now.
            }
        }
    }

    if ($iscsi_clients && scalar @{$iscsi_clients}) {

        #Invoking all the non 'generate_one*' calls
        $Log->step("Generating ISCSI Lifs.");
        # Call to generate iscsi lif resources.
        $self->generate_iscsi_lif_resources(%args);

        #TODO ideally move this but unclear about ordering dependecies for now
        if ($auto_nfs_lifs) {
            $Log->step("Generating NFS Lifs.");
            # Call to generate nfs lif resources.
            $self->generate_nfs_lif_resources(%args);
        } else {
            $Log->step("Auto nfs lif create off: Don't Generate NFS Lifs.");
        }

        $Log->step("Generating Igroups.");
        # Call to genereate iscsi service resources.
        $self->generate_igroup_resources(%args);

        $Log->step("Generating Initiators.");
        # Call to genereate iscsi service resources.
        $self->generate_igroup_initiator_resources(%args);

        $Log->step("Generating ISCSI Services.");
        # Call to genereate iscsi service resources.
        $self->generate_iscsi_services_resources(%args);

        $Log->step("Generating Lun Maps.");
        # Call to generate lun resources.
        $self->generate_lun_map_resources(%args);
    } else {
        $Log->step("No ISCSI clients, skipping ISCSI generation.");

        if ($auto_nfs_lifs) {
            $Log->step("Generating NFS Lifs.");
            # Call to generate nfs lif resources.
            $self->generate_nfs_lif_resources(%args);
        } else {
            $Log->step("Auto nfs lif create off: Don't Generate NFS Lifs.");
        }

    }
}


if (0) {
    $Log->step('Generating ONE aggregate resource on DST.');
    $self->generate_one_aggregate_resource(node=>'dpgqa-vlt-funct-3270-07',
                                        diskcount => 3,
                                        aggregate => 'dst_aggr',
                                        );
    $Log->step('Generating ONE vserver resource on DST.');
    $self->generate_one_vserver_resource(vserver => 'vs1',
                                        aggregate => 'dst_aggr',
                                        );
    $Log->step('Generating SOME volume resource on DST.');
    $self->generate_one_volume_resource(vserver => 'vs1',
                                        aggregate => 'dst_aggr',
                                        volume => 'vol1',
                                        size => '100m',
                                        );
    $self->generate_one_volume_resource(vserver => 'vs1',
                                        aggregate => 'dst_aggr',
                                        volume => 'vol2',
                                        size => '100m',
                                        );
    $Log->step('Generating volume efficiency policy');
    $self->generate_one_volume_efficiency_policy_resource(vserver => 'vs1',
                                        policy => 'foo',
                                        schedule => 'daily',
                                        duration => '100',
                                        );
    $Log->step('Generating volume efficiency');
    $self->generate_one_volume_efficiency_resource(vserver => 'vs1',
                                        volume => 'vol1',
                                        start => '1',
                                        compression => 'true',
                                        'inline-compression' => 'true',
                                        );
    $self->generate_one_volume_efficiency_resource(vserver => 'vs1',
                                        volume => 'vol2',
                                        policy => 'foo',
                                        start => '0',
                                        );

}
if (0) {
    #my $cluster_vserver_hash = {}; # A mapping of vservers to clusters.
    my $vserver_sm_policy_hash = {}; # A mapping of vservers to snapmirror policies.
    my $vserver_snapshot_policy_hash = {}; # A mapping of vservers to snapshot policies.
    my $cluster_job_schedule_cron_hash = {}; # A mapping of clusters to job schedules.
    # Call to generate ONE aggregate resource
    $Log->step('Generating ONE aggregate resource.');
    $self->generate_one_aggregate_resource(node=>'dpgqa-vlt-funct-3270-05',
                                        diskcount => 3,
                                        aggregate => 'src_aggr',
                                        );
    # Call to generate ONE aggregate resource
    $Log->step('Generating ONE aggregate resource on DST.');
    $self->generate_one_aggregate_resource(node=>'dpgqa-vlt-funct-3270-07',
                                        diskcount => 3,
                                        aggregate => 'dst_aggr',
                                        );
    # Call to generate ONE vserver resource
    $Log->step('Generating ONE vserver resource on SRC.');
    $self->generate_one_vserver_resource(vserver => 'vs0',
                                        aggregate => 'src_aggr',
                                        );
    # Call to generate ONE vserver resource
    $Log->step('Generating ONE vserver resource on DST.');
    $self->generate_one_vserver_resource(vserver => 'vs1',
                                        aggregate => 'dst_aggr',
                                        );
    # Create Snapshot policy
    $Log->step('Generate Snapshot policy resource');
    $self->generate_one_snapshot_policy_resource( policy => 'archpolicy',
                                            vserver => 'vs0',
                                            enabled => 'true',
                                            schedule1 => 'hourly',
                                            count1 => 16,
                                            vserver_snapshot_policy_hash => $vserver_snapshot_policy_hash,
                                        );
    # Call to generate ONE volume resource
    $Log->step('Generating SOME volume resource on SRC.');
    $self->generate_one_volume_resource(vserver => 'vs0',
                                        aggregate => 'src_aggr',
                                        volume => 'src_vol',
                                        size => '100m',
                                        );
    $self->generate_one_volume_resource(vserver => 'vs0',
                                        aggregate => 'src_aggr',
                                        volume => 'dst_vol2',
                                        size => '100m',
                                        );
    $self->generate_one_volume_resource(vserver => 'vs0',
                                        aggregate => 'src_aggr',
                                        volume => 'vlt_src_vol',
                                        size => '100m',
                                        'snapshot-policy' => 'archpolicy',
                                        );
    $self->generate_one_volume_resource(vserver => 'vs0',
                                        aggregate => 'src_aggr',
                                        volume => 'smgr_src_vol',
                                        size => '100m',
                                        );
    $self->generate_one_volume_resource(vserver => 'vs0',
                                        aggregate => 'src_aggr',
                                        volume => 'lrse_dr_src_vol',
                                        size => '100m',
                                        );
    # Call to generate ONE volume resource
    $Log->step('Generating SOME volume resource on DST.');
    $self->generate_one_volume_resource(vserver => 'vs1',
                                        aggregate => 'dst_aggr',
                                        volume => 'src_vol2',
                                        size => '100m',
                                        );
    $self->generate_one_volume_resource(vserver => 'vs1',
                                        aggregate => 'dst_aggr',
                                        volume => 'dst_vol',
                                        size => '100m',
                                        );
    $self->generate_one_volume_resource(vserver => 'vs1',
                                        aggregate => 'dst_aggr',
                                        volume => 'vlt_dst_vol',
                                        size => '100m',
                                        );
    $self->generate_one_volume_resource(vserver => 'vs1',
                                        aggregate => 'dst_aggr',
                                        volume => 'smgr_dst_vol',
                                        size => '500m',
                                        );
    $self->generate_one_volume_resource(vserver => 'vs1',
                                        aggregate => 'dst_aggr',
                                        volume => 'lrse_dr_dst_vol',
                                        size => '100m',
                                        );
    #Call to generate ONE lun resource
    $Log->step('Generating SOME lun resource.');
    $self->generate_one_lun_resource (vserver   => 'vs0',
                                      volume    => 'src_vol',
                                      lun       => 'lun1',
                                      size      => '50m',
                                        );
    $self->generate_one_lun_resource (vserver   => 'vs0',
                                      volume    => 'smgr_src_vol',
                                      lun       => 'src_lun1',
                                      size      => '50m',
                                        );
    $self->generate_one_lun_resource (vserver   => 'vs1',
                                      volume    => 'smgr_dst_vol',
                                      lun       => 'dst_lun1',
                                      size      => '50m',
                                        );

    #Call to generate ONE snapmirror resource
    $Log->step('Generating SOME snapmirror resource.');

    $self->generate_one_snapmirror_policy_resource ( policy => 'archpolicy',
                                                    vserver => 'vs1',
                                                    type => 'async-mirror',
                                                    vserver_sm_policy_hash => $vserver_sm_policy_hash,
                                                    );
    $self->generate_one_job_schedule_cron_resource ( cluster_job_schedule_cron_hash => $cluster_job_schedule_cron_hash,
                                                    cluster => 'dpgvlt32700708',
                                                    name => 'archjob',
                                                    minute => 5,
                                                    );
    $self->generate_one_snapmirror_policy_rule_resource( policy => 'archpolicy',
                                                    'snapmirror-label' => 'sm_created',
                                                    #'snapmirror-label' => 'archlabel',
                                                    keep => 1,
                                                    vserver => 'vs1',
                                                    );
    $Log->step('Generating SOME snapmirror resource.');
    $self->generate_one_snapmirror_resource(src_vol         => 'smgr_src_vol',
                                        dst_vol             => 'smgr_dst_vol',
                                        sm_type             => 'SMGR',
                                        cg_items            => ['src_lun1'],
                                        src_cg_name         => 'src_cg_arch',
                                        dst_cg_name         => 'dst_cg_arch',
                                        initialize          => 0,
                                        );
    $self->generate_one_snapmirror_resource(src_vol         => 'lrse_dr_src_vol',
                                        dst_vol             => 'lrse_dr_dst_vol',
                                        sm_type             => 'LRSE_DR_MIRROR_LATEST',
                                        initialize          => 1,
                                        xgen_metadata_args  => {
                                                                 sm_config => 'unidirectional',
                                                                 }
                                        );
    $self->generate_one_snapmirror_resource(src_vol         => 'vlt_src_vol',
                                        dst_vol             => 'vlt_dst_vol',
                                        sm_type             => 'VAULT',
                                        initialize          => 1,
                                        xgen_metadata_args  => {
                                                                 sm_config => 'unidirectional',
                                                                 }
                                        );
    $self->generate_one_snapmirror_resource(src_vol         => 'src_vol',
                                        dst_vol             => 'dst_vol',
                                        sm_type             => 'ADR',
                                        initialize          => 1,
                                        schedule            => 'archjob',
                                        policy              => 'archpolicy',
                                        cluster_job_schedule_cron_hash => $cluster_job_schedule_cron_hash,
                                        xgen_metadata_args  => {
                                                                 sm_config => 'unidirectional',
                                                                 }
                                        );
    $self->generate_one_snapmirror_resource(src_vol         => 'src_vol2',
                                        dst_vol             => 'dst_vol2',
                                        sm_type             => 'ADR',
                                        initialize          => 1,
                                        xgen_metadata_args  => {
                                                                 sm_config => 'unidirectional',
                                                                 }
                                        );
}

}


# generate cluster and node resources
sub generate_cluster_and_node_resources {
    my($self, %args) = @_;
    $Log->enter() if $may_enter;
    $Log->comment("Base_YML Generate cluster and node resources");

    my $nacl_node_objs = $self->{nacl_node_objs};
    my @all_node_objs;

    foreach my $node (@{ $nacl_node_objs }) {
        if (!grep {$node->node() eq $_->node()} @all_node_objs) {
            #first push the one passed in my user.
            #But the node could already have been identified if its
            #HA partner was passed in previously.
            push (@all_node_objs, $node);
        }

        #now get all the nodes in that cluster which the passed in node
        #belongs to
        my @n = NACL::CS::Cluster->fetch (
                        command_interface => $node,
                        requested_fields  => [qw(node)]);

        $self->make_cluster_hostrec( node => $node);

        #Now push all the nodes that belong to the cluster
        foreach my $entry (@n) {
            my $nobj = NACL::C::Node->new (name => $entry->node);
            if (!grep {$nobj->node() eq $_->node()} @all_node_objs) {
                push (@all_node_objs, $nobj);
            }
        }
    }

    tie my %cluster_names, 'Tie::IxHash' or die "Could not Tie Hash";

    foreach my $nacl_node_obj (@all_node_objs) {
        # figure out what cluster $nacl_node_obj belongs to
        # once you do that update the lookup table for this node
        my @clusters =
            NACL::CS::ClusterIdentity->fetch(command_interface => $nacl_node_obj);

        my $node_name = $nacl_node_obj->node();
        my $cluster_name = $clusters[0]->name();

        # associating this node with its cluster name
        $self->set_cluster_name($node_name, $cluster_name);

        # hash of unique cluster names that points to an array
        # of nodes in that cluster
        push(@{$cluster_names{$cluster_name}}, $node_name);
    }

    foreach my $cluster_name (keys %cluster_names) {
        my $nodes = $cluster_names{$cluster_name};
        my $cluster_resource =
            new NACL::AppBuilder::Xgen::Resource::Cluster(cluster => $cluster_name,
                                                nodes   => $nodes);

        $cluster_resource->set_shared_attributes(cluster_name => $cluster_name);
        $self->add_resources(resources => [$cluster_resource]);

        foreach my $node (@$nodes) {
            my $node_resource = new NACL::AppBuilder::Xgen::Resource::Node(node => $node);
            $node_resource->add_parent_resources(resources => [$cluster_resource]);

            $self->add_node_resource_to_lookup_table($node_resource);
            $self->add_resources(resources => [$node_resource]);
        }
    }

    $Log->exit() if $may_exit;
}

sub get_cluster_hostrec {
    my $self = shift;
    my %args = validate_with(
        params => \@_,
        spec => {
            cluster => { type => SCALAR, },
        },
        #allow_extra => 1,
    );

	my $cluster_name = $args{cluster};

    if (my $cluster_ci = $self->{cluster_ci}->{$cluster_name}){
		return $cluster_ci;
	} else {
        $Log->log("unknown cluster ci $cluster_name");
	}
}

sub make_cluster_hostrec {
    my $self = shift;
    my %args = validate_with(
        params => \@_,
        spec => {
            node => { type => OBJECT, },
        },
        #allow_extra => 1,
    );

    my $node_component = $args{node};

    my $clusteridentity = NACL::CS::ClusterIdentity->fetch(
        command_interface => $node_component,
        requested_fields  => [ 'name' ] );

    my $cluster_name = $clusteridentity->name();

    if (my $cluster_ci = $self->{cluster_ci}->{$cluster_name}){
		return $cluster_ci;
	}


    my $clus_mgmt_lif = NACL::CS::NetworkInterface->fetch(
        command_interface => $node_component,
        filter => { role => 'cluster-mgmt' },
    );

    # cannot rely on a cluster nate_host file being available, so create
    # cluster mgmt hostrec from info in node nate_host file
    my $node_hostrec    = $node_component->{hostrec};
    my $cluster_hostrec = Hostrec->new(
        id               => $cluster_name,
        hosttype         => $node_hostrec->{'hosttype'},
        hostname         => $cluster_name,
        default_ip       => $clus_mgmt_lif->address(),
        default_ip_gx    => $clus_mgmt_lif->address(),,
        default_username => $node_hostrec->{'default_username'} || q{admin},
        default_password => $node_hostrec->{'default_password'} ||
                                                               q{netapp1!},
    );
    my $cluster_component = NACL::C::Cluster->new(hostrec => $cluster_hostrec);
    $self->{cluster_ci}->{$cluster_name} = $cluster_component;
	return $cluster_component;
}

# generate 'one' aggregate resource
sub generate_one_aggregate_resource {
    my $self = shift;
    my %args = validate_with(
        params => \@_,
        spec => {
            #node_resource_obj => { type => OBJECT, isa =>'NACL::AppBuilder::Xgen::Resource::Node'},
            node => { type => SCALAR, }, # parent
            diskcount => { type => SCALAR, default => 3 },
            aggregate => { type => SCALAR, optional => 1 }, # name of aggr, if not specified this method will decide name
            aggr_name_format => { type => SCALAR, default => 'node_name', regex => qr/^(node_name|node_id|aggr_only)$/},
            aggr_number => { type => SCALAR, default => 0 }, # used in the aggr naming
        },
        allow_extra => 1, # allow additional args that OnTap accepts
    );

    #my $node_resource_obj = $args{node_resource_obj};
    #my $node_name = $node_resource_obj->get_attribute('node');
    my $node = delete $args{node};
    my $aggregate = delete $args{aggregate};
    my $aggr_name_format = delete $args{aggr_name_format};
    my $aggr_number = delete $args{aggr_number};
    my $diskcount = $args{diskcount};

    my $node_resource_obj_array = $self->get_resources(# category => qr/.*/ do we need this ?
                                                        type => 'NACL::AppBuilder::Xgen::Resource::Node',
                                                        data => { node => $node },
                                                        );
    #my $node_resource_obj = @{ $node_resource_obj_array }[0]; # get the first elem out
    my $node_resource_obj;
    foreach my $ele (@{ $node_resource_obj_array }) {
        if ($ele->get_attribute('node') eq $node) {
            $node_resource_obj = $ele;
        }
    }
    $Log->debug('My node resource obj is '.Dumper($node_resource_obj));

    if ($aggregate) {
        # we use user provided name
        # do we need to add number suffix? check
        $Log->log('Using user provided aggr name '.$aggregate);
    } else {
        # we decide name of aggregate
        #$aggr_name_format = $self->param('AGGR_NAME_FORMAT') || 'node_name';
        if ($aggr_name_format eq 'node_name') {
            $aggregate = $node . '_aggr' . $aggr_number;
        } elsif ($aggr_name_format eq 'node_id') {
            $aggregate = 'node' . $self->get_node_id($node) . '_aggr' . $aggr_number;
        } elsif ($aggr_name_format eq 'aggr_only') {
            $aggregate = 'aggr_' . $aggr_number;
        } else {
            $aggregate = 'aggr_' . $aggr_number;
        }
    }

    # make the aggregate resource
    $Log->log('Adding aggr resource name='.$aggregate.' diskcount='.$diskcount.' node='.$node);
    my $aggregate_resource_obj = NACL::AppBuilder::Xgen::Resource::Aggregate->new(aggregate  => $aggregate,
                                                                                 nodes => [$node],
                                                                                 %args
                                                                                );
    # connect the aggregate resource to the nodes parent resource
    $aggregate_resource_obj->add_parent_resources(resources => [$node_resource_obj]);
    $self->add_resources(resources => [$aggregate_resource_obj]);
}

# generate 'one' vserver resource
sub generate_one_vserver_resource {
    my $self = shift;
    my %args = validate_with(
        params => \@_,
        spec => {
            vserver => { type => SCALAR, }, # name of vserver
            aggregate => { type => SCALAR, }, # parent
        },
        allow_extra => 1, # allow additional args that OnTap accepts
    );
    my $aggregate = $args{aggregate};
    my $vserver = $args{vserver};

    my $aggregate_resource_obj_array = $self->get_resources( type => 'NACL::AppBuilder::Xgen::Resource::Aggregate',
                                                            data => { aggregate => $aggregate },
                                                            );
    #my $aggregate_resource_obj = @{ $aggregate_resource_obj_array }[0]; # get the first elem out
    my $aggregate_resource_obj;
    foreach my $ele (@{ $aggregate_resource_obj_array}) {
        if ($ele->get_attribute('aggregate') eq $aggregate) {
            $aggregate_resource_obj = $ele;
        }
    }
    $Log->debug('My aggregate resource obj is '.Dumper($aggregate_resource_obj));

    # make the vserver resource
    $Log->log('Adding vserver resource name='.$vserver.' aggregate='.$aggregate);
    $Log->trace('Adding vserver resource args:' . Dumper \%args);
    my $vserver_resource_obj = NACL::AppBuilder::Xgen::Resource::Vserver->new(%args);

    # connect the vserver resource to the aggr's cluster parent resource
    $vserver_resource_obj->add_parent_resources(resources => [$aggregate_resource_obj]);
    $self->add_resources(resources => [$vserver_resource_obj]);
}

# generate 'one' btrfs resource
sub generate_one_btrfs_resource {
    my $self = shift;
    my %args = validate_with(
        params => \@_,
        spec => {
            btrfs => { type => SCALAR, }, # name of btrfs
			#btrfs_volume => { type => SCALAR, }, # parent
        },
        allow_extra => 1, # allow additional args that OnTap accepts
    );
	$Log->log( "ONE BTRFS:" . Dumper \%args);



#my $aggregate = $args{aggregate};
	my $btrfs = $args{btrfs};
    my $sma = dp::xplatform::lib::SMAgent->new(client=>$btrfs);

	#TODO - start
	#$sma->start(@_);

	#my $aggregate_resource_obj_array = $self->get_resources( type => 'NACL::AppBuilder::Xgen::Resource::Aggregate',
	#data => { aggregate => $aggregate },
	#);
	##my $aggregate_resource_obj = @{ $aggregate_resource_obj_array }[0]; # get the first elem out
	#my $aggregate_resource_obj;
	#foreach my $ele (@{ $aggregate_resource_obj_array}) {
	#if ($ele->get_attribute('aggregate') eq $aggregate) {
	#$aggregate_resource_obj = $ele;
	#}
	#}
	#$Log->log('My aggregate resource obj is '.Dumper($aggregate_resource_obj));
	#
	## make the btrfs resource
	#$Log->log('Adding btrfs resource name='.$btrfs.' aggregate='.$aggregate);
	#$Log->trace('Adding btrfs resource args:' . Dumper \%args);
	#my $btrfs_resource_obj = NACL::AppBuilder::Xgen::Resource::btrfs->new(%args);
	#
	## connect the btrfs resource to the aggr's cluster parent resource
	#$btrfs_resource_obj->add_parent_resources(resources => [$aggregate_resource_obj]);
	#$self->add_resources(resources => [$btrfs_resource_obj]);
	#$self->add_resources(resources => [$btrfs]);
}

# generate 'one' btrfs_volume resource
sub generate_one_btrfs_volume_resource {
    my $self = shift;
    my %args = validate_with(
        params => \@_,
        spec => {
            btrfs => { type => SCALAR, }, # name of btrfs
            name => { type => SCALAR, }, # parent
            create_lun => { type => SCALAR, default => 1}, # parent
        },
        allow_extra => 1, # allow additional args that OnTap accepts
    );
    $Log->log( "ONE BTRFS VOLUME:" . Dumper \%args);

    my $btrfs = $args{btrfs};
    my $volume = $args{name};
    my $create_lun = $args{create_lun};
    my $sma = dp::xplatform::lib::SMAgent->new(client=>$btrfs);

    my $ip = $sma->ip();
    $Log->log('My btrfs ip is '.$ip);
    #$Log->log('My btrfs resource obj is '.Dumper($sma));
    $Log->log('Creating btrfs volume for SMA: '.$volume);
    my $btrfs_vol_id = $sma->vol_create(vol_name => $volume);
    if ($create_lun) {
        $Log->log('create lun');
        $sma->create_lun(vol_name => $volume); #needed for initialize
	} else {
        $Log->log('DON\'T create lun');
	}
    my $source_volume_resource_obj = "$ip:/lun/$btrfs_vol_id";
    $Log->log('source path: '.$source_volume_resource_obj);

}

# generate 'one' volume resource
sub generate_one_volume_resource {
    my $self = shift;
    my %args = validate_with(
        params => \@_,
        spec => {
            aggregate => { type => SCALAR, optional => 1 }, # name of aggregate - used as parent
            vserver => { type => SCALAR }, # name of vserver
            volume => { type => SCALAR, optional => 1 }, # name of volume - if not specified, this method will create name
            volume_name_format => { type => SCALAR, default => 'vserver', regex => qr/^(vserver|volume_only|volumeonly)$/},
            volume_number => { type => SCALAR, default => 0 }, # used in the vol naming
            volume_suffix => { type => SCALAR, optional => 1 }, # used in the vol naming
            size => { type => SCALAR, optional => 1 }, # if not specified, OnTap will default to min size
        },
        allow_extra => 1, # allow additional args that OnTap accepts
    );
    my $aggregate = $args{aggregate};
    my $aggr_list = $args{'aggr-list'};
    if (!$aggregate) {
        if ( $aggr_list && scalar @{$aggr_list}) {
            #Just grab the 1st aggr from the list. In the future we MAY want
            #to put the vols as children of all the aggrs but I'm not sure
            #how that will affect the creation of things in engenx
            $aggregate = $aggr_list->[0];
        } else {
            NATE::Exceptions::Argument->throw("Either aggregate or aggr-list needs to be passed to generate_one_volume_resource");
        }
    }
    my $vserver = $args{vserver};
    my $volume = $args{volume};
    my $volume_name_format = delete $args{volume_name_format};
    my $volume_number = delete $args{volume_number};
    my $volume_suffix = delete $args{volume_suffix};

    my $vserver_resource_obj_array = $self->get_resources( type => 'NACL::AppBuilder::Xgen::Resource::Vserver',
                                                            data => { vserver => $vserver },
                                                            );
    #my $vserver_resource_obj = @{ $vserver_resource_obj_array }[0]; # get the first elem out
    my $vserver_resource_obj;
    foreach my $ele (@{ $vserver_resource_obj_array}) {
        if ($ele->get_attribute('vserver') eq $vserver) {
            $vserver_resource_obj = $ele;
        }
    }
    $Log->debug('My vserver resource obj is '.Dumper($vserver_resource_obj));

    my $aggregate_resource_obj_array = $self->get_resources( type => 'NACL::AppBuilder::Xgen::Resource::Aggregate',
                                                            data => { aggregate => $aggregate },
                                                            );
    #my $aggregate_resource_obj = @{ $aggregate_resource_obj_array }[0]; # get the first elem out
    my $aggregate_resource_obj;
    foreach my $ele (@{ $aggregate_resource_obj_array}) {
        if ($ele->get_attribute('aggregate') eq $aggregate) {
            $aggregate_resource_obj = $ele;
        }
    }
    $Log->debug('My aggregate resource obj is '.Dumper($aggregate_resource_obj));

    if ($volume) {
        # we use user provided name
        # do we need to add number suffix? check
        $Log->log('Using user provided vol name '.$volume);
    } else {
        # we decide vol name
        if ($volume_name_format eq 'vserver') {
            $volume = 'vol_'.$volume_number.'_'.$vserver;
        } elsif ($volume_name_format eq 'volumeonly') {
            $volume = 'vol'.$volume_number;
        } elsif ($volume_name_format eq 'volume_only') {
            $volume = 'vol_'.$volume_number;
        } else {
            $volume = 'vol_'.$volume_number;
        }
        # add suffix if user specified
        $volume .= $volume_suffix;
    }

    # make the volume resource
    $Log->log('Adding volume resource name='.$volume.' aggregate='.$aggregate.' vserver='.$vserver.' ARGS='.Dumper(\%args));
    #$Log->log('snapshot policy='.$args{'snapshot-policy'}) if ($args{'snapshot-policy'};
    my $volume_resource_obj = NACL::AppBuilder::Xgen::Resource::Volume->new(%args);

    # connect the volume resource to the vserver and aggregate
    $volume_resource_obj->add_parent_resources(resources => [$vserver_resource_obj,$aggregate_resource_obj]);
    $self->add_resources(resources => [$volume_resource_obj]);
}

# generate 'one' lun resource
sub generate_one_lun_resource {
    my $self = shift;
    my %args = validate_with(
        params => \@_,
        spec => {
            vserver => { type => SCALAR, },
            volume  => { type => SCALAR, },
            lun     => { type => SCALAR, optional => 1},
            size    => { type => SCALAR, },
            ostype  => { type => SCALAR, default => 'linux'},
        },
        'allow_extra' => 1,
    );

    my $vserver = $args{vserver};
    my $volume  = $args{volume};
    my $lun     = $args{lun};
    my $size    = $args{size};
    my $ostype  = $args{ostype};

    my $vol_resource_obj_array = $self->get_resources(
                                                      type => 'NACL::AppBuilder::Xgen::Resource::Volume',
                                                      data => { volume => $volume },
                                                    );
    #my $vol_resource_obj = @{ $vol_resource_obj_array }[0]; # get the first elem out
    my $vol_resource_obj ;
    foreach my $ele (@{ $vol_resource_obj_array}) {
        if ($ele->get_attribute('volume') eq $volume) {
            $vol_resource_obj = $ele;
        }
    }
    $Log->debug('My vol resource obj is '.Dumper($vol_resource_obj));

    $Log->log('Adding lun resource name='.$lun .
              ' volume='.$volume .
              ' vserver='.$vserver .
              ' size=' . $size .
              ' ostype=' . $ostype);
    my $lun_resource_obj = NACL::AppBuilder::Xgen::Resource::Lun->new(vserver =>$vserver,
                                                             volume =>$volume,
                                                             lun => $lun,
                                                             size => $size,
                                                             ostype => $ostype,
                                                             %args);


    # connect the lun resource to the volume parent resource
    $lun_resource_obj->add_parent_resources(resources => [$vol_resource_obj]);
    $self->add_resources(resources => [$lun_resource_obj]);
}

# generate 'one' cluster peer relationship
sub generate_one_cluster_peer {

    my $self = shift;
    my %args = validate_with(
        params => \@_,
        spec   => {
            src_cluster => { type => SCALAR, required => 1 },
            dst_cluster => { type => SCALAR, required => 1 },
        },
        allow_extra => 1,    # allow additional args that OnTap accepts
    );

    my $src_cluster = delete $args{src_cluster};
    my $dst_cluster = delete $args{dst_cluster};

    my ( $source_cluster_obj, $destination_cluster_obj );

    # Get the clusters so that we can alternate vserver creation between multiple clusters.
    my $cluster_resources = $self->get_resources(type => "NACL::AppBuilder::Xgen::Resource::Cluster");

    #
    # will get two clusters every time, figure out
    # which is the src and the dst
    #
    if ( $src_cluster eq @{$cluster_resources}[0]->get_attribute('cluster_name') )
    {
        $source_cluster_obj      = @{$cluster_resources}[0];
        $destination_cluster_obj = @{$cluster_resources}[1];
    } else {
        $source_cluster_obj      = @{$cluster_resources}[1];
        $destination_cluster_obj = @{$cluster_resources}[0];
    }


    $Log->debug( 'My src cluster resource objects is ' . Dumper($source_cluster_obj) );
    $Log->debug( 'My dst cluster resource objects is ' . Dumper($destination_cluster_obj) );

    my $cluster_peer_resource_obj
        = NACL::AppBuilder::Xgen::Resource::ClusterPeer->new(
        cluster  => $source_cluster_obj->get_attribute('cluster_name'),
        rcluster => $destination_cluster_obj->get_attribute('cluster_name')
        );


    $cluster_peer_resource_obj->add_parent_resources(
        resources => [ $destination_cluster_obj, $source_cluster_obj ] );

    $self->add_resources(resources => [ $cluster_peer_resource_obj ] );

}

# generate one vserver peer relationship
sub generate_one_vserver_peer_resource {
    my $self = shift;
    my %args = validate_with(
        params => \@_,
        spec   => {
            vserver => { type => SCALAR, required => 1 },
            vserver_peer => { type => SCALAR, required => 1 },
        },
        allow_extra => 1,    # allow additional args that OnTap accepts
    );

    my @vserver_peers;
    my $vservers = $self->get_resources(type => "NACL::AppBuilder::Xgen::Resource::Vserver");

    my ($src_vserver,$dst_vserver);

    foreach my $vserver (@{ $vservers }) {
        if ($vserver->get_attribute('vserver') eq $args{vserver}) {
            $src_vserver = $vserver;
        } elsif ($vserver->get_attribute('vserver') eq $args{vserver_peer}) {
            $dst_vserver = $vserver;
        }
    }

    my %vserver_peer_args = (vserver        => $args{vserver},
                             'peer-vserver' => $args{vserver_peer});
    if ($src_vserver->get_attribute('cluster_name') ne $dst_vserver->get_attribute('cluster_name')) {
        $vserver_peer_args{'peer-cluster'} = $dst_vserver->get_attribute('cluster_name');
    }

    $Log->log( 'Vserver Peer args ' . Dumper(\%vserver_peer_args) );

    my $vserver_peer_res = NACL::AppBuilder::Xgen::Resource::VserverPeer->new(%vserver_peer_args);
    # Add parents with source cluster last so that the command is run from the source cluster.
    $vserver_peer_res->add_parent_resources(resources => [$dst_vserver,$src_vserver]);

    push(@vserver_peers, $vserver_peer_res);
    $self->add_resources(resources => \@vserver_peers);
}

###############################################################################
# DESCRIPTION: Builds an ethernet fabric data hash from the nate_host files
#              for all the nodes found by NACL::C::Node->find();  Uses the
#              attributes DATA_IP, DATA_GATEWAY, DATA_MASK from the nate_host
#              files.
# INPUT: NONE
# OUTPUT: A hashref of the built ethernet fabric data hash.
# Exceptions: None thrown directly by this method.
###############################################################################
# get param, either passed in or the constant default
#
sub param {
    my $self  = shift;
    my $param = shift;

    my $value = $self->xgen_config_params()->{$param} // $self->get_constant($param);

    $Log->comment("xgen config param $param value: $value"); #TODO set to trace or remove
    #$Log->comment("xgen config params dump:" . dumper $self->xgen_config_params()); #TODO set to trace or remove

    return $value;
}

sub get_constant {
    my $self  = shift;
    my $param = shift;

    my $constant;

    my $method = $self->can($param);
    if (defined $method) {
        $constant = $self->$method;
        $Log->comment("constant param $param value: $constant\n"); #TODO set to trace or remove
    } else {
        $Log->comment("constant param $param NOT defined as a constant\n"); #TODO set to trace or remove
    }

    return $constant;
}

#setter/getter for xgen config params that are passed in
sub xgen_config_params {
    my $self = shift;

    my $param = shift;

    if (defined $param) {
        $self->{xgen_config_params} = $param;
    }

    return $self->{xgen_config_params};
}
#sub paramset {
#my $self = shift;
#$Log->comment("paramset dump:" . Dumper $Paramset); #TODO set to trace or remove
#return $Paramset;
#}

sub generate_eth_fabric_data_from_natehost_files {
    my ($self, %args) = @_;

    my @NODES = NACL::C::Node->find();
    my @ips;
    my $netmask = undef;
    my $gateway = undef;
    foreach my $node (@NODES) {
        my $hostrec = Hostrec->new(id=>$node->node());
        my %networks = $hostrec->param_fields();
        foreach my $key (keys %networks) {
            if($key =~ /DATA_IP/i) {
                push (@ips, $networks{$key});
            } elsif ($key =~ /DATA_GATEWAY/i) {
                $gateway = $networks{$key};
            } elsif ($key =~ /DATA_MASK/i) {
                $netmask = $networks{$key};
            }
        }
    }
    my %eth_fabric_data = ();
    $eth_fabric_data{'eth1'} = {
        'ipspace'           => 'Default',
        'broadcast_domain'  => 'Default',
        'address_range_v4'  => \@ips,
        'netmask_v4'        => $netmask,
        'gateway_v4'        => $gateway,
    };

    $Log->comment("The fabric data that was generated is " . dumper(\%eth_fabric_data));
    return \%eth_fabric_data
}

###############################################################################
# DESCRIPTION: Overrides the generate_aggregate_resources method from ONTAP.pm
#              so that it does not query for aggregates but instead creates
#              aggregates based on the AGGREGATES_PER_NODE and
#              NUMBER_DISKS_PER_AGGREGATE constants. The naming convention
#              use is n#_aggr#. Where both # start at 0 for each cluster.
#              TODO: Currently the N# does not start at 0 for each cluster.
#                    ONTAP.pm may need to be updated.
# INPUT: NONE
# OUTPUT: A hashref of the built ethernet fabric data hash.
# Exceptions: None thrown directly by this method.
###############################################################################
sub generate_aggregate_resources {
    my ($self, %args) = @_;

    my $nodes = $self->get_resources(type => "NACL::AppBuilder::Xgen::Resource::Node");

    #aggr name format: node_name, node_id, only_aggr
    my $aggr_name_format = $self->param('AGGR_NAME_FORMAT') || 'node_name';
    my %cluster_aggr_count_hash = ();
    for (my $i = 0; $i < $self->param('AGGREGATES_PER_NODE'); $i++) {
        foreach my $node (@{$nodes}) {
            my $node_name = $node->get_attribute('node');
            my $aggr_number = $cluster_aggr_count_hash{$node->get_attribute('cluster_name')} // 0;
            $cluster_aggr_count_hash{$node->get_attribute('cluster_name')}++;

            my $aggr_name;
            if ($aggr_name_format eq 'node_name') {
                $aggr_name = $node_name . "_aggr" . $aggr_number;
            } elsif ($aggr_name_format eq 'node_id') {
                $aggr_name = "node" . $self->get_node_id($node_name) . "_aggr" . $aggr_number;
            } elsif ($aggr_name_format eq 'aggr_only') {
                $aggr_name = "aggr_" . $aggr_number;
            } else {
                $aggr_name = "aggr_" . $aggr_number;
            }

            # make the aggregate resource
            my $aggregate = NACL::AppBuilder::Xgen::Resource::Aggregate->new('aggregate'    => $aggr_name,
                                                                             'diskcount'    => $self->param('NUMBER_DISKS_PER_AGGREGATE'),
                                                                             'nodes'        => [$node_name],
                                                                             'maxraidsize'  => $self->param('NUMBER_DISKS_PER_AGGREGATE'));
            $aggregate->add_parent_resources(resources => [$node]);
            $self->add_resources(resources => [$aggregate]);
        }
    }
}

###############################################################################
# DESCRIPTION: Generates NFS lif resources and adds them to the configuration.
#              N Lif resources are created in every vserver per fabric defined,
#              where N is the number of nodes in the cluster that the vserver
#              is in. The lifs use the IPs, Netmask, and Gateway described in
#              the fabrics in the eth_data_fabric hash. This method currently
#              supports IPV4.
# INPUT: eth_fabric_data - a hash ref of ethernet fabric data including
#                          ips, netmask, gateway(optional) for the different
#                          ethernet fabrics for the testbed.
# OUTPUT: NONE
# Exceptions: None thrown directly by this method.
###############################################################################
sub generate_nfs_lif_resources {
    my ($self, %args) = @_;

    my $vservers = $self->get_resources(type => "NACL::AppBuilder::Xgen::Resource::Vserver");

    foreach my $vs (@{$vservers}) {
        my $eth_fabric_data = $args{eth_fabric_data};

        # Each vserver should only have one ancestor that is a cluster.
        my ($cluster) = $vs->get_ancestors(type => "NACL::AppBuilder::Xgen::Resource::Cluster");
        # Get all the nodes that are children of the vservers cluster so that we get all the nodes.
        my @nodes = $cluster->get_descendants(type => "NACL::AppBuilder::Xgen::Resource::Node");
        my $already_create_routes;
        # make an iscsi lif on each node (on each fabric)
        foreach my $node (@nodes) {
            my $node_name = $node->get_attribute('node');
            my $node_id = $self->get_node_id($node_name);

            # obtain the ethenet fabric table for this node
            my $fabric_table = $self->get_fabric_lookup_table_by_node($node_name, 'eth');
            foreach my $fabric (keys %{$fabric_table}) {
                my $address_range_v4 = $eth_fabric_data->{$fabric}->{address_range_v4};
                my $netmask_v4       = $eth_fabric_data->{$fabric}->{netmask_v4};
                if (!$address_range_v4 or !$netmask_v4) {
                    NATE::Exceptions::Argument->throw("No address ranges and/or netmasks were provided for fabric '$fabric'");
                }
                # if this fabric has a gateway defined,
                # make a route resource on this vserver if we havn't built one yet for this fabric
                $already_create_routes->{$fabric} //= {};
                if (!$already_create_routes->{$fabric}->{route_resource} and $eth_fabric_data->{$fabric}->{gateway_v4}) {
                    my $route_resource = NACL::AppBuilder::Xgen::Resource::NetworkRoute->new(vserver => $vs->get_attribute('vserver'),
                                                                                             gateway => $eth_fabric_data->{$fabric}->{gateway_v4});
                    $route_resource->add_parent_resources(resources => [$vs]);
                    $self->add_resources(resources => [$route_resource]);
                    $already_create_routes->{$fabric}->{route_resource} = $route_resource;
                }

                # choose a port on this fabric for this node
                my $port_name = shift(@{$fabric_table->{$fabric}});
                push(@{$fabric_table->{$fabric}}, $port_name);
                my $port_id = $self->get_port_id($node_name, $port_name);

                # make the lif resource
                my $lif = NACL::AppBuilder::Xgen::Resource::Lif->new(vserver => $vs->get_attribute('vserver'),
                                                                     'data-protocol' => ['nfs'],
                                                                     address_range => $address_range_v4,
                                                                     netmask => $netmask_v4,
                                                                     lif => "nfs_lf_n" . $node_id . "_p" . $port_id,
                                                                     'home-node' => $node_name,
                                                                     'home-port' => $port_name);
                $lif->add_parent_resources(resources => [$vs]);
                $self->add_resources(resources => [$lif]);
            }
        }
    }
}

###############################################################################
# DESCRIPTION: Generates ISCSI lif resources and adds them to the configuration.
#              N Lif resources are created in every vserver per fabric defined,
#              where N is the number of nodes in the cluster that the vserver
#              is in. The lifs use the IPs, Netmask, and Gateway described in
#              the fabrics in the eth_data_fabric hash. This method currently
#              supports IPV4.
# INPUT: eth_fabric_data - a hash ref of ethernet fabric data including
#                          ips, netmask, gateway(optional) for the different
#                          ethernet fabrics for the testbed.
# OUTPUT: NONE
# Exceptions: None thrown directly by this method.
###############################################################################
sub generate_iscsi_lif_resources {
    my ($self, %args) = @_;

    my $vservers = $self->get_resources(type => "NACL::AppBuilder::Xgen::Resource::Vserver");

    foreach my $vs (@{$vservers}) {
        my $eth_fabric_data = $args{eth_fabric_data};

        # Each vserver should only have one ancestor that is a cluster.
        my ($cluster) = $vs->get_ancestors(type => "NACL::AppBuilder::Xgen::Resource::Cluster");
        # Get all the nodes that are children of the vservers cluster so that we get all the nodes.
        my @nodes = $cluster->get_descendants(type => "NACL::AppBuilder::Xgen::Resource::Node");
        my $already_create_routes;
        # make an iscsi lif on each node (on each fabric)
        foreach my $node (@nodes) {
            my $node_name = $node->get_attribute('node');
            my $node_id = $self->get_node_id($node_name);

            # obtain the ethenet fabric table for this node
            my $fabric_table = $self->get_fabric_lookup_table_by_node($node_name, 'eth');
            foreach my $fabric (keys %{$fabric_table}) {
                my $address_range_v4 = $eth_fabric_data->{$fabric}->{address_range_v4};
                my $netmask_v4       = $eth_fabric_data->{$fabric}->{netmask_v4};
                if (!$address_range_v4 or !$netmask_v4) {
                    NATE::Exceptions::Argument->throw("No address ranges and/or netmasks were provided for fabric '$fabric'");
                }
                # if this fabric has a gateway defined,
                # make a route resource on this vserver if we havn't built one yet for this fabric
                $already_create_routes->{$fabric} //= {};
                if (!$already_create_routes->{$fabric}->{route_resource} and $eth_fabric_data->{$fabric}->{gateway_v4}) {
                    my $route_resource = NACL::AppBuilder::Xgen::Resource::NetworkRoute->new(vserver => $vs->get_attribute('vserver'),
                                                                                             gateway => $eth_fabric_data->{$fabric}->{gateway_v4});
                    $route_resource->add_parent_resources(resources => [$vs]);
                    $self->add_resources(resources => [$route_resource]);
                    $already_create_routes->{$fabric}->{route_resource} = $route_resource;
                }

                # choose a port on this fabric for this node
                my $port_name = shift(@{$fabric_table->{$fabric}});
                push(@{$fabric_table->{$fabric}}, $port_name);
                my $port_id = $self->get_port_id($node_name, $port_name);

                # make the lif resource
                my $lif = NACL::AppBuilder::Xgen::Resource::Lif->new(vserver => $vs->get_attribute('vserver'),
                                                                     'data-protocol' => ['iscsi'],
                                                                     address_range => $address_range_v4,
                                                                     netmask => $netmask_v4,
                                                                     lif => "iscsi_lf_n" . $node_id . "_p" . $port_id,
                                                                     'home-node' => $node_name,
                                                                     'home-port' => $port_name);
                $lif->add_parent_resources(resources => [$vs]);
                $self->add_resources(resources => [$lif]);
            }
        }
    }
}

###############################################################################
# DESCRIPTION: Generates file resources and adds them to the configuration.
#              Generates a number of files equal to the TOTAL_FILE_COUNT
#              constant.  The files will only be generated on non DP type
#              volumes. The PERCENTAGE_OF_VOLUMES_WITH_FILES controls the
#              percent of the non DP type volumes that get files. If a
#              volume has a vm-align-suffix the MAXIMUM_PERCENT_VM_ALIGNED_FILES
#              and MINIMUM_PERCENT_VM_ALIGNED_FILES constants control the range
#              of random percentages of files that have the vm-align-suffix in
#              each volume. The file size is controlled by the FILE_SIZE
#              constant.
# INPUT: NONE
# OUTPUT: NONE
# Exceptions: None thrown directly by this method.
###############################################################################
sub generate_file_resources {
    my ($self, %args) = @_;

    # Get the clusters so that we can alternate vserver creation between multiple clusters.
    my $volume_resources = $self->get_resources(type => "NACL::AppBuilder::Xgen::Resource::Volume");

    # Filter the volumes resources by security style mixed.
    my $matches_hash = positive_filter(resources => $volume_resources,
                                       filter    => {'type' => 'DP'});

    my @volumes = @{$matches_hash->{'not_matches'}};

    my $volume_count = ceil((scalar(@volumes) * $self->param('PERCENTAGE_OF_VOLUMES_WITH_FILES')) / 100);
    my @files;

    return unless ($volume_count > 0);

    # Format is cluster->vserver->volume->desired_percentage
    #                                     aligned_files
    #                                     total_files
    my %volume_vm_aligned_hash = ();
    for (my $i = 0; $i < $self->param('TOTAL_FILE_COUNT'); $i++) {
        my $volume_index = $i%$volume_count;

        my $volume = $volumes[$volume_index];

        my $vm_align_suffix;

        try {
            $vm_align_suffix = $volume->get_attribute('vm-align-suffix');
        } otherwise {
            $vm_align_suffix = undef;
        };

        my $path;
        # If we have a suffix then we want to perform the suffix logic.
        if (defined $vm_align_suffix) {
            # Get the desired percent of aligned files for this volume.
            my $desired_percent_of_aligned = $volume_vm_aligned_hash{$volume->get_attribute('cluster_name')}
                                        {$volume->get_attribute('vserver')}
                                            {$volume->get_attribute('volume')}
                                                {'desired_percentage'};
            # If the desired percent of aligned files for the volume is undefined,
            # then this is the first file for the volume and we need to calculated the desired percent.
            unless (defined $desired_percent_of_aligned) {
                $desired_percent_of_aligned = $self->param('MINIMUM_PERCENT_VM_ALIGNED_FILES') +
                                                                     int(rand($self->param('MAXIMUM_PERCENT_VM_ALIGNED_FILES') -
                                                                              $self->param('MINIMUM_PERCENT_VM_ALIGNED_FILES')));
                $volume_vm_aligned_hash{$volume->get_attribute('cluster_name')}
                                        {$volume->get_attribute('vserver')}
                                            {$volume->get_attribute('volume')}
                                                {'desired_percentage'} = $desired_percent_of_aligned;
            }
            # Add to the total count for this volume.
            $volume_vm_aligned_hash{$volume->get_attribute('cluster_name')}
                                        {$volume->get_attribute('vserver')}
                                            {$volume->get_attribute('volume')}
                                                {'total_files'}++;
            my $total_count  = $volume_vm_aligned_hash{$volume->get_attribute('cluster_name')}
                                                            {$volume->get_attribute('vserver')}
                                                                {$volume->get_attribute('volume')}
                                                                    {'total_files'};
            my $aligned_count = $volume_vm_aligned_hash{$volume->get_attribute('cluster_name')}
                                                            {$volume->get_attribute('vserver')}
                                                                {$volume->get_attribute('volume')}
                                                                    {'aligned_files'} // 0;

            $path = '/vol/' . $volume->get_attribute("volume") . "/file$i";
            if ((($aligned_count / $total_count) * 100) < $desired_percent_of_aligned) {
                $volume_vm_aligned_hash{$volume->get_attribute('cluster_name')}{$volume->get_attribute('vserver')}{$volume->get_attribute('volume')}{'aligned_files'}++;
                $path .= $vm_align_suffix;
            }
        } else {
            $path = '/vol/' . $volume->get_attribute("volume") . "/file$i";
        }

        my $file_res = NACL::AppBuilder::Xgen::Resource::File->new('vserver' =>$volume->get_attribute('vserver'),
                                                                   'path'  => $path,
                                                                   'file-size'    =>$self->param('FILE_SIZE'),
                                                                   'xgen_metadata' => {retry_strings => ['Mounting unjunctioned Volume',
                                                                                                         'number of allowed concurrent mkfile threads exceeded']});

        # connect the vs resource to the aggr's cluster parent resource
        $file_res->add_parent_resources(resources => [$volume]);

        push(@files, $file_res);
    }

    $self->add_resources(resources => \@files);
}

###############################################################################
# DESCRIPTION: Generates lun resources and adds them to the configuration.
#              Generates a number of luns equal to the TOTAL_LUN_COUNT
#              constant.  The luns will only be generated on non DP type
#              volumes. The PERCENTAGE_OF_VOLUMES_WITH_LUNS controls the
#              percent of the non DP type volumes that get luns. The luns
#              ostype defaults to linux but will be changes if the lun is
#              mapped. The lun size is controlled by the LUN_SIZE constant.
# INPUT: NONE
# OUTPUT: NONE
# Exceptions: None thrown directly by this method.
###############################################################################
sub generate_lun_resources {
    my ($self, %args) = @_;

    # Get the clusters so that we can alternate vserver creation between multiple clusters.
    my $volume_resources = $self->get_resources(type => "NACL::AppBuilder::Xgen::Resource::Volume");

    # Filter the volumes resources by security style mixed.
    my $matches_hash = positive_filter(resources => $volume_resources,
                                       filter    => {'type' => 'DP'});

    my @volumes = @{$matches_hash->{'not_matches'}};

    my $volume_count = ceil((scalar(@volumes) * $self->param('PERCENTAGE_OF_VOLUMES_WITH_LUNS')) / 100);
    my @luns;

    return unless ($volume_count > 0);

    for (my $i = 0; $i < $self->param('TOTAL_LUN_COUNT'); $i++) {
        my $volume_index = $i%$volume_count;

        my $volume = $volumes[$volume_index];

        my $lun_res = NACL::AppBuilder::Xgen::Resource::Lun->new(vserver =>$volume->get_attribute('vserver'),
                                                                 volume  =>$volume->get_attribute("volume"),
                                                                 lun     => "lun$i",
                                                                 size    =>$self->param('LUN_SIZE'),
                                                                 ostype  => 'linux'); # Defaulting the ostype to linux. Will overwrite when maps created.

        # connect the vs resource to the aggr's cluster parent resource
        $lun_res->add_parent_resources(resources => [$volume]);

        push(@luns, $lun_res);
    }
    $self->add_resources(resources => \@luns);
}

###############################################################################
# DESCRIPTION: Generates lun map resources and adds them to the configuration.
#              Generates a lun map on every lun in the configuration.  Each lun
#              will be mapped to a single igroup.  In a vserver the lun
#              will be evenly distributed between igroups for maps. The luns
#              ostype will be changed to the ostype of the igroups client.
# INPUT: NONE
# OUTPUT: NONE
# Exceptions: None thrown directly by this method.
###############################################################################
sub generate_lun_map_resources {
    my ($self, %args) = @_;

    # Get the clusters so that we can alternate vserver creation between multiple clusters.
    my $vserver_resources = $self->get_resources(type => "NACL::AppBuilder::Xgen::Resource::Vserver");
    my @vservers = @{$vserver_resources};

    my @lun_maps;
    foreach my $vserver (@vservers) {
        my @luns = $vserver->get_descendants(type => "NACL::AppBuilder::Xgen::Resource::Lun");

        next unless (scalar @luns);

        my @igroups = $vserver->get_descendants(type => "NACL::AppBuilder::Xgen::Resource::Igroup");

        next unless (scalar @igroups);

        #(cl1,cl2)  vserver1:vol1:lun1  vserver2:vol1:lun1  might not work in this case so need better logic.
        foreach my $lun (@luns) {
            my $igroup = shift(@igroups);
            push(@igroups,$igroup);

            # Change the os_type of the lun to match the map.
            $lun->set_private_attributes(ostype => $igroup->get_attribute('client_os_type'));

            my $lunmap_res = NACL::AppBuilder::Xgen::Resource::LunMap->new(vserver => $lun->get_attribute('vserver'),
                                                                           lun => $lun->get_attribute('lun'),
                                                                           volume => $lun->get_attribute('volume'),
                                                                           igroup => $igroup->get_attribute('igroup'));

            # We want the vserver to be last.
            $lunmap_res->add_parent_resources(resources => [$igroup,$lun]);

            push(@lun_maps, $lunmap_res);
        }
    }
     $self->add_resources(resources => \@lun_maps);
}

###############################################################################
# DESCRIPTION: Generates igroup resources and adds them to the configuration.
#              Generates an igroup per ISCSI client on every vserver in the
#              configuration.  This method currently only supports ISCSI. The
#              igroups are named after the client they represent.
# INPUT: NONE
# OUTPUT: NONE
# Exceptions: None thrown directly by this method.
###############################################################################
sub generate_igroup_resources {
    my ($self, %args) = @_;

    # Get the clusters so that we can alternate vserver creation between multiple clusters.
    my $vserver_resources = $self->get_resources(type => "NACL::AppBuilder::Xgen::Resource::Vserver");
    my @vservers = @{$vserver_resources};

    # Get the clusters so that we can alternate vserver creation between multiple clusters.
    my $client_resources = $self->get_resources(type => "NACL::AppBuilder::Xgen::Resource::Client");
    my @clients = @{$client_resources};

    my @igroup;
    foreach my $vserver (@vservers) {
        foreach my $client (@clients) {
            # skip clients that are not doing iSCSI
            next if (!grep("iscsi", @{$client->get_attribute('protocols')}));

            my $igroup_res = NACL::AppBuilder::Xgen::Resource::Igroup->new(vserver => $vserver->get_attribute('vserver'),
                                                                           igroup => $client->get_attribute('name'),
                                                                           protocol => "iscsi",
                                                                           ostype => $client->get_attribute('ostype') =~ /windows/ ? "windows" : $client->get_attribute('ostype'),
                                                                           xgen_metadata => {client_os_type => $client->get_attribute('ostype')}); # used later for lun ostype.

            # We want the vserver to be last.
            $igroup_res->add_parent_resources(resources => [$client,$vserver]);

            push(@igroup, $igroup_res);
        }
    }
    $self->add_resources(resources => \@igroup);
}

###############################################################################
# DESCRIPTION: Generates igroup initiator resources and adds them to the
#              configuration. Generates an igroup initiator per igroup. Since
#              igroups are named after the client they represent we use the name
#              of the igroup for the client name.
# INPUT: NONE
# OUTPUT: NONE
# Exceptions: None thrown directly by this method.
###############################################################################
sub generate_igroup_initiator_resources {
    my ($self, %args) = @_;

    # Get the clusters so that we can alternate igroup creation between multiple clusters.
    my $igroup_resources = $self->get_resources(type => "NACL::AppBuilder::Xgen::Resource::Igroup");
    my @igroups = @{$igroup_resources};

    my @igroup_initiators;
    foreach my $igroup (@igroups) {
        my $initiator_res = NACL::AppBuilder::Xgen::Resource::Igroup::Initiator->new(vserver => $igroup->get_attribute('vserver'),
                                                                                     igroup => $igroup->get_attribute('igroup'),
                                                                                     protocol => "iscsi",
                                                                                     client_name => $igroup->get_attribute('igroup'));

        # We want the vserver to be last.
        $initiator_res->add_parent_resources(resources => [$igroup]);

        push(@igroup_initiators, $initiator_res);
    }

    $self->add_resources(resources => \@igroup_initiators);
}

###############################################################################
# DESCRIPTION: Generates ISCSI service resources and adds them to the
#              configuration. Generates an ISCSI service per vserver.
# INPUT: NONE
# OUTPUT: NONE
# Exceptions: None thrown directly by this method.
###############################################################################
sub generate_iscsi_services_resources {
    my ($self, %args) = @_;

    # Get the clusters so that we can alternate vserver creation between multiple clusters.
    my $vserver_resources = $self->get_resources(type => "NACL::AppBuilder::Xgen::Resource::Vserver");
    my @vservers = @{$vserver_resources};

    my @iscsi_services;

    foreach my $vserver (@vservers) {
        my $iscsi_res = NACL::AppBuilder::Xgen::Resource::Iscsi->new(vserver => $vserver->get_attribute('vserver'));

        # Add vserver as parent to iscsi service.
        $iscsi_res->add_parent_resources(resources => [$vserver]);

        push(@iscsi_services, $iscsi_res);
    }
    $self->add_resources(resources => \@iscsi_services);
}

###############################################################################
# DESCRIPTION: Generates vserver nfs service resources and adds them to the
#              configuration.
# INPUT: vserver
# OUTPUT: NONE
# Exceptions: NATE::Exceptions::Argument if vserver is invalid
###############################################################################
sub generate_one_vserver_nfs_resource {
    my $self = shift;
    my %args = validate_with(
        params => \@_,
        spec   => {
            vserver => { type => SCALAR, required => 1 },
            #vserver_peer => { type => SCALAR, required => 1 },
        },
        allow_extra => 1,    # allow additional args that OnTap accepts
    );

    my $vserver_resources = $self->get_resources(type => "NACL::AppBuilder::Xgen::Resource::Vserver");
    my @vservers = @{$vserver_resources};

    my $vserver_parent;

    foreach my $vserver (@vservers) {
        if ($vserver->get_attribute('vserver') eq $args{vserver}) {
            $vserver_parent = $vserver;
        }
    }
    if (!$vserver_parent) {
        NATE::Exceptions::Argument->throw("vserver '$args{vserver}' does not exist");
    }
    my $nfs_res = NACL::AppBuilder::Xgen::Resource::Nfs->new(
        %args,
        xgen_nacl_cs_validation => 0, # need to turn off validation until
                                      # params in default cdef
    );
    # Add vserver as parent to nfs service.
    $nfs_res->add_parent_resources(resources => [$vserver_parent]);
    $self->add_resources(resources => [ $nfs_res ]);
}

###############################################################################
# DESCRIPTION: Generates cluster peer resources and adds them to the
#              configuration. Generates cluster peers in a cascade fasion.
#              C1->C2->C3. The order of the peers is based on the order off
#              clusters returned from the configuration and has no known
#              control for the order.
# INPUT: NONE
# OUTPUT: NONE
# Exceptions: None thrown directly by this method.
###############################################################################
sub generate_cluster_peer_resources {
    my ($self, %args) = @_;

    # Get the clusters so that we can alternate vserver creation between multiple clusters.
    my $cluster_resources = $self->get_resources(type => "NACL::AppBuilder::Xgen::Resource::Cluster");
    my @clusters = @{$cluster_resources};

    my @cluster_peers;

    return unless (scalar @clusters > 1);

    for (my $i = 1; $i < scalar @clusters; $i++) {
        my $source_cluster = $clusters[$i-1];
        my $destination_cluster = $clusters[$i];

        # May need to get a node from each cluster and use that as the name instead of the cluster_name
        # since nate_host files may not always exist for the clusters.
        my $cluster_peer_res = NACL::AppBuilder::Xgen::Resource::ClusterPeer->new(cluster => $source_cluster->get_attribute('cluster_name'),
                                                                                  rcluster => $destination_cluster->get_attribute('cluster_name'));

        # Add parents with source cluster last so that the command is run from the source cluster.
        $cluster_peer_res->add_parent_resources(resources => [$destination_cluster,$source_cluster]);

        push(@cluster_peers, $cluster_peer_res);
    }
    $self->add_resources(resources => \@cluster_peers);
}

###############################################################################
# DESCRIPTION: Generates vserver peer resources and adds them to the
#              configuration. Generates vserver peers based on snapmirrors in
#              the configuration. Gets all the snapmirrors from the
#              configurations and builds 1 vserver peer per pair of vservers
#              that need to be peered for snapmirrors to work.
# INPUT: NONE
# OUTPUT: NONE
# Exceptions: None thrown directly by this method.
###############################################################################
sub generate_vserver_peer_resources {
    my ($self, %args) = @_;

    # Get the clusters so that we can alternate vserver creation between multiple clusters.
    my $snapmirror_resources = $self->get_resources(type => "NACL::AppBuilder::Xgen::Resource::Snapmirror");
    my @snapmirrors = @{$snapmirror_resources};

    my @vserver_peers;

    return unless (scalar @snapmirrors > 0);

    my %vserver_peer_hash;
    for (my $i = 0; $i < scalar @snapmirrors; $i++) {
        my $source_vserver = $snapmirrors[$i]->get_attribute("source-vserver");
        my $destination_vserver = $snapmirrors[$i]->get_attribute("destination-vserver");
        my @vserver_resources = $snapmirrors[$i]->get_ancestors(type => "NACL::AppBuilder::Xgen::Resource::Vserver");
        if (defined $vserver_peer_hash{$source_vserver} and
            defined $vserver_peer_hash{$source_vserver}{$destination_vserver}) {
               # This will have already been added to the resource list.
               next;
        } elsif (defined $vserver_peer_hash{$destination_vserver} and
                 defined $vserver_peer_hash{$destination_vserver}{$source_vserver}) {
            # This will have already been added to the resource list.
            next;
        } else {
            # This combination was not found so we need to add this resource.
            $vserver_peer_hash{$source_vserver}{$destination_vserver} = 1;
            my %vserver_peer_args = (vserver => $source_vserver,
                                     'peer-vserver' => $destination_vserver);

            foreach my $vserver (@vserver_resources) {
                if ( $vserver->get_attribute('vserver') eq $source_vserver) {
                    $source_vserver = $vserver;
                } elsif ( $vserver->get_attribute('vserver') eq $destination_vserver) {
                    $destination_vserver = $vserver;
                }
            }

            if($source_vserver->get_attribute('cluster_name') ne $destination_vserver->get_attribute('cluster_name')) {
                $vserver_peer_args{'peer-cluster'} = $destination_vserver->get_attribute('cluster_name');
            }
            my $vserver_peer_res = NACL::AppBuilder::Xgen::Resource::VserverPeer->new(%vserver_peer_args);
            # Add parents with source cluster last so that the command is run from the source cluster.
            $vserver_peer_res->add_parent_resources(resources => [$destination_vserver,$source_vserver]);

            push(@vserver_peers, $vserver_peer_res);
        }
    }
    $self->add_resources(resources => \@vserver_peers);
}


###############################################################################
# DESCRIPTION: Generates vserver resources and adds them to the configuration.
#              Generates a number of vservers based on the TOTAL_VSERVER_COUNT
#              constant. The vservers will be named vs# where # starts at 0 and
#              counts up to TOTAL_VSERVER_COUNT-1. The vservers will alternate
#              between the clusters provided and inside each cluster will
#              alternate between the aggregates in that cluster that are part
#              of the configuration.
# INPUT: NONE
# OUTPUT: NONE
# Exceptions: None thrown directly by this method.
###############################################################################
sub generate_vserver_resources {
    my ($self, %args) = @_;

    # Get the clusters so that we can alternate vserver creation between multiple clusters.
    my $cluster_resources = $self->get_resources(type => "NACL::AppBuilder::Xgen::Resource::Cluster");
    my @clusters = @{$cluster_resources};
    my %cluster_aggregate_hash = ();
    foreach my $cluster_resource (@$cluster_resources) {
        @{$cluster_aggregate_hash{$cluster_resource->get_attribute('cluster_name')}} =
                $cluster_resource->get_descendants(type => "NACL::AppBuilder::Xgen::Resource::Aggregate");
        # {clus1 => [aggr1_obj,aggr2_obj], clus2 => [aggr1_obj,aggr2_obj]}
    }

    my @vservers;

    for (my $i = 0; $i < $self->param('TOTAL_VSERVER_COUNT'); $i++) {
        my $cluster = shift(@clusters);
        push (@clusters, $cluster);

        my $aggr = shift(@{$cluster_aggregate_hash{$cluster->get_attribute('cluster_name')}});
        push(@{$cluster_aggregate_hash{$cluster->get_attribute('cluster_name')}}, $aggr);

        my $vs_res = NACL::AppBuilder::Xgen::Resource::Vserver->new(vserver =>"vs$i",
                                                                    aggregate => $aggr->get_attribute("aggregate"));

        # connect the vs resource to the aggr's cluster parent resource
        $vs_res->add_parent_resources(resources => [$aggr]);

        push(@vservers, $vs_res);
    }
    $self->add_resources(resources => \@vservers);
}

###############################################################################
# DESCRIPTION: Generates volume resources and adds them to the configuration.
#              Generates a number of volumes based on the TOTAL_VOLUME_COUNT
#              constant. The volumes will be named Vol# where # starts at 0 and
#              counts up to TOTAL_VOLUME_COUNT-1. The volumes will alternate
#              between the clusters provided and inside each cluster will
#              alternate between the aggregates and vservers in that cluster
#              that are part of the configuration. The
#              VOLUME_VM_ALIGN_SUFFIX_PERCENT_LAYOUT will control the
#              percentage of the total volumes that have each of the
#              vm-align-suffix attributes defined in the layout. The
#              VOLUME_VM_ALIGN_SECTOR_PERCENT_LAYOUT will control the
#              percentage of the volumes with vm-align-sufix that have
#              each of the vm-align-sector attributes defined in the layout.
#              The volume size is controlled by the VOLUME_SIZE constant.
# INPUT: NONE
# OUTPUT: NONE
# Exceptions: None thrown directly by this method.
###############################################################################
sub generate_volume_resources {
    my ($self, %args) = @_;
  # Get the clusters so that we can alternate vserver creation between multiple clusters.
    my $cluster_resources = $self->get_resources(type => "NACL::AppBuilder::Xgen::Resource::Cluster");
    my @clusters = @{$cluster_resources};
    my %cluster_aggregate_hash = ();
    my %cluster_vserver_hash = ();
    foreach my $cluster_resource (@$cluster_resources) {
        @{$cluster_aggregate_hash{$cluster_resource->get_attribute('cluster_name')}} =
                $cluster_resource->get_descendants(type => "NACL::AppBuilder::Xgen::Resource::Aggregate");
        # {clus1 => [aggr1_obj,aggr2_obj], clus2 => [aggr1_obj,aggr2_obj]}
        @{$cluster_vserver_hash{$cluster_resource->get_attribute('cluster_name')}} =
                $cluster_resource->get_descendants(type => "NACL::AppBuilder::Xgen::Resource::Vserver");
        # {clus1 => [vs0_obj,vs2_obj], clus2 => [vs1_obj,vs3_obj]}
    }

    my @volumes;

    # Create a suffix layout based on the total number of volume and the percentage suffix layout constant.
    my $suffix_layout = convert_percent_layout_to_count_layout(layout        => $self->param('VOLUME_VM_ALIGN_SUFFIX_PERCENT_LAYOUT'),
                                                               total_objects => $self->param('TOTAL_VOLUME_COUNT'));
    # Get the number of none suffix volumes to use for the total for the sector layout.
    my $volumes_without_suffix_count = $suffix_layout->{':NONE:'} // 0;
    # Create a sector layout based on the total number of volumes that have a suffix and the percentage sector layout constant.
    my $sector_layout = convert_percent_layout_to_count_layout(layout        => $self->param('VOLUME_VM_ALIGN_SECTOR_PERCENT_LAYOUT'),
                                                               total_objects => $self->param('TOTAL_VOLUME_COUNT') - $volumes_without_suffix_count);

    for (my $i = 0; $i < $self->param('TOTAL_VOLUME_COUNT'); $i++) {
        my $cluster = shift(@clusters);
        push (@clusters, $cluster);

        my $aggr = shift(@{$cluster_aggregate_hash{$cluster->get_attribute('cluster_name')}});
        push(@{$cluster_aggregate_hash{$cluster->get_attribute('cluster_name')}}, $aggr);

        my $vserver = shift(@{$cluster_vserver_hash{$cluster->get_attribute('cluster_name')}});
        push(@{$cluster_vserver_hash{$cluster->get_attribute('cluster_name')}}, $vserver);

        my $vs_name = $vserver->get_attribute("vserver"),

        my $volume_name_format = $self->param('VOLUME_NAME_FORMAT') || 'vserver';
        my $volname;
        if ($volume_name_format eq 'vserver') {
            $volname = "vol_${i}_${vs_name}";
        } elsif ($volume_name_format eq 'volume_only') {
            $volname = "vol_$i";
        } elsif ($volume_name_format eq 'volumeonly') {
            $volname = "vol$i";
        } else {
            $volname = "vol_$i";
        }

        my %volume_arguments = (vserver   => $vs_name,
                                aggregate => $aggr->get_attribute("aggregate"),
                                volume    => $volname,
                                size      => $self->param('VOLUME_SIZE'));

        # Get the next suffix element to use from the suffix layout
        my $suffix = get_next_layout_element(iterator => $i,
                                             layout => $suffix_layout);

        # If there is a suffix then set the vm-align-suffix and get and set the vm-align-sector
        unless($suffix eq ':NONE:') {
            $volume_arguments{'vm-align-suffix'} = $suffix;
            $volume_arguments{'vm-align-sector'} = get_next_layout_element(iterator => $i,
                                                                           layout => $sector_layout);
        }

        my $volume_res = NACL::AppBuilder::Xgen::Resource::Volume->new(%volume_arguments);

        # connect the vs resource to the aggr's cluster parent resource
        $volume_res->add_parent_resources(resources => [$vserver,$aggr]);

        push(@volumes, $volume_res);
    }
    $self->add_resources(resources => \@volumes);
}

###############################################################################
# DESCRIPTION: Adds SMGR related data to SMGR type snapmirrors in the
#              configuration.  The SMGR_CG_ITEM_PERCENT_LAYOUT constant will
#              control the percentage of SMGR snapmirrors that get each type
#              defined in the layout. Current types known are file and lun.
#              Based on the type of cg items this method generates and add
#              source-path, destination-path, and cg-item-mappings to the
#              SMGR snapmirror resource and the SMGR snapmirror initialize
#              resource if it exists.
# INPUT: NONE
# OUTPUT: NONE
# Exceptions: None thrown directly by this method.
###############################################################################
sub add_cg_items_to_SMGR_snapmirror_resources {
    my ($self, %args) = @_;

    my $snapmirror_resources = $self->get_resources(type => "NACL::AppBuilder::Xgen::Resource::Snapmirror");

    # Filter the snapmirrors resources by snapmirror type of smgr.
    my $matches_hash = positive_filter(resources => $snapmirror_resources,
                                       filter    => {'sm_type' => 'smgr'});

    my @snapmirrors = @{$matches_hash->{'matches'}};

    # Create a SMGR layout based on the number of SMGR snapmirror required.
    my $smgr_layout = convert_percent_layout_to_count_layout(layout        => $self->param('SMGR_CG_ITEM_PERCENT_LAYOUT'),
                                                             total_objects => scalar(@snapmirrors));

    my $index = 0;
    foreach my $snapmirror (@snapmirrors) {
            # Get the next cg item type element to use from the smgr layout
        my $cg_item_type = get_next_layout_element(iterator => $index,
                                                   layout => $smgr_layout);
        $index++;
        my $cg_source_path = $snapmirror->get_attribute('source-vserver') . ':/cg/srccg_' . $snapmirror->get_attribute('source-volume');
        my $cg_destination_path = $snapmirror->get_attribute('destination-vserver') . ':/cg/dstcg_' . $snapmirror->get_attribute('destination-volume');

        if($cg_item_type =~ /file/i) {
            $cg_source_path .= '_file';
            $cg_destination_path .= '_file';
        }
        if($cg_item_type =~ /lun/i) {
            $cg_source_path .= '_lun';
            $cg_destination_path .= '_lun';
        }

        my @volumes = $snapmirror->get_ancestors(type => "NACL::AppBuilder::Xgen::Resource::Volume");
        my ($source_volume,$destination_volume);
        foreach my $volume (@volumes) {
            my $source = 0;
            my $destination = 0;
            if($volume->get_attribute('vserver') eq $snapmirror->get_attribute('source-vserver') and
               $volume->get_attribute('volume') eq $snapmirror->get_attribute('source-volume')) {
                $source_volume = $volume;
                next;
            } elsif($volume->get_attribute('vserver') eq $snapmirror->get_attribute('destination-vserver') and
               $volume->get_attribute('volume') eq $snapmirror->get_attribute('destination-volume')) {
               $destination_volume = $volume;
               next;
            } else {
                NATE::BaseException->throw("Found a child volume of an SMGR relationship when trying to add cg-item-mappings that is not a source or destination volume.")
            }
        }

        my %cg_item_mappings = ();

        if($cg_item_type =~ /file/i) {
            my @files = $source_volume->get_descendants(type => "NACL::AppBuilder::Xgen::Resource::File");
            # use only specified % of files for cg
            my $file_count = ceil((scalar(@files) * $self->param('PERCENTAGE_OF_LUNS_FILES_USED_FOR_CG_ITEMS')) / 100);
            my @files_for_cg = splice(@files, 0, $file_count);
            foreach my $file (@files_for_cg) {
                # Files were built using the path so we use the path.
                my $source_cg_item = $file->get_attribute('path');
                # The destination path will be the same as the source but we need to replace the source
                # volume name with the destination volume name.
                my $destination_cg_item = $source_cg_item;
                my $source_volume_name = $source_volume->get_attribute('volume');
                my $destination_volume_name = $destination_volume->get_attribute('volume');
                $destination_cg_item =~ s/$source_volume_name/$destination_volume_name/;
                $cg_item_mappings{$source_cg_item} = $destination_cg_item;
            }
        }
        if($cg_item_type =~ /lun/i) {
            my @luns = $source_volume->get_descendants(type => "NACL::AppBuilder::Xgen::Resource::Lun");
            # use only specified % of luns for cg
            my $lun_count = ceil((scalar(@luns) * $self->param('PERCENTAGE_OF_LUNS_FILES_USED_FOR_CG_ITEMS')) / 100);
            my @luns_for_cg = splice(@luns, 0, $lun_count);

            foreach my $lun (@luns_for_cg) {
                # Luns were created using the individual parameters so we need to buld the path.
                my $source_cg_item = '/vol/' . $lun->get_attribute('volume') . '/' . $lun->get_attribute('lun');
                my $destination_cg_item = '/vol/' . $destination_volume->get_attribute('volume') . '/' . $lun->get_attribute('lun');
                $cg_item_mappings{$source_cg_item} = $destination_cg_item;
            }
        }
        $snapmirror->set_private_attributes('cg-item-mappings' => \%cg_item_mappings,
                                            'source-path' => $cg_source_path,
                                            'destination-path' => $cg_destination_path);
        $snapmirror->set_metadata_attributes('cg_item_type' => $cg_item_type);
        my ($snapmirror_initialize) = $snapmirror->get_descendants(type => "NACL::AppBuilder::Xgen::Resource::SnapmirrorInitialize");
        if (defined $snapmirror_initialize) {
            $snapmirror_initialize->set_private_attributes('cg-item-mappings' => \%cg_item_mappings,
                                                'source-path' => $cg_source_path,
                                                'destination-path' => $cg_destination_path);
            $snapmirror_initialize->set_metadata_attributes('cg_item_type' => $cg_item_type);
        }
    }
}

# add user provided cg items to 'ONE' snapmirror resource
sub add_cg_items_to_one_SMGR_snapmirror_resources {
    my $self = shift;
    my %args = validate_with(
        params => \@_,
        spec => {
            snapmirror => { type => OBJECT, }, # add isa ???
            src_cg_name => { type => SCALAR, optional => 1, }, # if not specified this method will choose a name
            dst_cg_name => { type => SCALAR, optional => 1, }, # if not specified this method will choose a name
            cg_items => { type => ARRAYREF, },
        },
        allow_extra => 0,
    );
    my $snapmirror = $args{snapmirror};
    my $src_cg_name = $args{src_cg_name};
    my $dst_cg_name = $args{dst_cg_name};
    my $cg_items = $args{cg_items};

    if (!$src_cg_name) {
        $src_cg_name = 'srccg_'.$snapmirror->get_attribute('source-volume');
        $Log->log('Setting cg name for src to '.$src_cg_name);
    } else {
        $Log->log('Using user provided cg name for src '.$src_cg_name);
    }
    if (!$dst_cg_name) {
        $dst_cg_name = 'dstcg_'.$snapmirror->get_attribute('source-volume');
        $Log->log('Setting cg name for dst to '.$dst_cg_name);
    } else {
        $Log->log('Using user provided cg name for dst '.$dst_cg_name);
    }

    my $cg_source_path = $snapmirror->get_attribute('source-vserver') . ':/cg/'.$src_cg_name;
    my $cg_destination_path = $snapmirror->get_attribute('destination-vserver') . ':/cg/'.$dst_cg_name;
    $Log->log('cg src path='.$cg_source_path.' cg dst path='.$cg_destination_path);

    # do we need this block - leaving it in for now
    my @volumes = $snapmirror->get_ancestors(type => "NACL::AppBuilder::Xgen::Resource::Volume");
    my ($source_volume,$destination_volume);
    foreach my $volume (@volumes) {
        my $source = 0;
        my $destination = 0;
        if($volume->get_attribute('vserver') eq $snapmirror->get_attribute('source-vserver') and
           $volume->get_attribute('volume') eq $snapmirror->get_attribute('source-volume')) {
            $source_volume = $volume;
            next;
        } elsif($volume->get_attribute('vserver') eq $snapmirror->get_attribute('destination-vserver') and
           $volume->get_attribute('volume') eq $snapmirror->get_attribute('destination-volume')) {
           $destination_volume = $volume;
           next;
        } else {
            NATE::BaseException->throw("Found a child volume of an SMGR relationship when trying to add cg-item-mappings that is not a source or destination volume.")
        }
    }

    # pick only the luns that user specified
    my %cg_items_hash = map { $_ => 1 } @{$cg_items};
    my @luns = $source_volume->get_descendants(type => 'NACL::AppBuilder::Xgen::Resource::Lun');
    my %cg_item_mappings = ();
    foreach my $lun (@luns) {
        my $lun_name = $lun->get_attribute('lun');
        if(exists($cg_items_hash{$lun_name})) {
            $Log->log('Found lun name='.$lun_name);
            # Luns were created using the individual parameters so we need to buld the path.
            my $source_cg_item = '/vol/' . $lun->get_attribute('volume') . '/' . $lun->get_attribute('lun');
            my $destination_cg_item = '/vol/' . $destination_volume->get_attribute('volume') . '/' . $lun->get_attribute('lun');
            $cg_item_mappings{$source_cg_item} = $destination_cg_item;
        } else {
            $Log->log('Lun name='.$lun_name.' not in user cg items');
        }
    }
    $snapmirror->set_private_attributes('cg-item-mappings' => \%cg_item_mappings,
                                        'source-path' => $cg_source_path,
                                        'destination-path' => $cg_destination_path);
    #$snapmirror->set_metadata_attributes('cg_item_type' => $cg_item_type);

    # TBD do same for file as well

    my ($snapmirror_initialize) = $snapmirror->get_descendants(type => 'NACL::AppBuilder::Xgen::Resource::SnapmirrorInitialize');
    if (defined $snapmirror_initialize) {
        $snapmirror_initialize->set_private_attributes('cg-item-mappings' => \%cg_item_mappings,
                                                        'source-path' => $cg_source_path,
                                                        'destination-path' => $cg_destination_path);
        #$snapmirror_initialize->set_metadata_attributes('cg_item_type' => $cg_item_type);
    }
}

###############################################################################
# DESCRIPTION: Generates Volume Efficiency Resources and Volume Efficiency
#              Policy resources in the configuration. This will generate
#              A volume efficiency resource on each volume that was
#              designated during snapmirror resource generation to need
#              volume efficiency. Uses the START_VOLUME_EFFICIENCY constant
#              to know whether to tell EnGenX to start volume efficiency.
#              Volume efficiency will never be started on destinations of
#              snapmirrors. The arguments for volume efficiency are based on a
#              template corresponding to the volumes volume_efficency_options
#              parameter.  If the Volume efficiency arguments contain a policy,
#              then a policy is created by calling
#              generate_volume_efficiency_policy_resource
# INPUT: NONE
# OUTPUT: NONE
# Exceptions: NATE::BaseException based on the volume_efficiency_options
#             template not existing.
###############################################################################
sub generate_volume_efficiency_resources {
    my ($self, %args) = @_;

    my $volume_resources = $self->get_resources(type => "NACL::AppBuilder::Xgen::Resource::Volume");

    # Filter the volumes resources by security style mixed.
    my $matches_hash = positive_filter(resources => $volume_resources,
                                       filter    => {'volume_efficiency_options' => undef});

    # We want the volumes that have volume efficiency options defined.
    my @volumes = @{$matches_hash->{'not_matches'}};

    return unless (scalar(@volumes) > 0);

    my $vserver_volume_efficiency_policy_hash = {};
    foreach my $volume (@volumes) {
        my $volume_efficiency_options_template = $volume->get_attribute('volume_efficiency_options');
        my $volume_efficiency_options_template_method = uc($volume_efficiency_options_template) . '_VOLUME_EFFICIENCY_OPTIONS';
        my $start;
        try {
            $volume->get_attribute('destination_for_sm_type');
            # If this is a destination then we do not want to start the service.
            $start = 0;
        } otherwise {
            $start = $self->param('START_VOLUME_EFFICIENCY');
        }; # Do Nothing.
        my $method = $self->can($volume_efficiency_options_template_method);
        if (defined $method) {
            #TODO - param check for these volume efficiency args
            my %volume_efficiency_args = %{$self->$method};
            my $volume_efficiency = NACL::AppBuilder::Xgen::Resource::Volume::Efficiency->new('vserver' => $volume->get_attribute('vserver'),
                                                                                              'volume' => $volume->get_attribute('volume'),
                                                                                              'start'  => $start,  # Put start here so it can be overridden by args
                                                                                              %volume_efficiency_args);
            $volume_efficiency->add_parent_resources(resources => [$volume]);
            $self->add_resources(resources => [$volume_efficiency]);

            my $volume_efficiency_policy = $volume_efficiency_args{'policy'};
            # Add the job_schedule to the destination cluster if it has not already been added.
            $self->generate_volume_efficiency_policy_resource('volume'                                => $volume,
                                                              'volume_efficiency_policy'              => $volume_efficiency_policy,
                                                              'vserver_volume_efficiency_policy_hash' => $vserver_volume_efficiency_policy_hash);
        } else {
            NATE::BaseException->throw("Tried to use a volume efficiency options template $volume_efficiency_options_template that this config does not understand.");
        }
    }
}

# generate 'ONE' volume efficiency resource
sub generate_one_volume_efficiency_resource {
    my $self = shift;
    my %args = validate_with(
        params => \@_,
        spec => {
                vserver => { type => SCALAR, },
                volume => { type => SCALAR, }, # parent
                start => { type => SCALAR, },
                vserver_volume_efficiency_policy_hash => { type => HASHREF, default => {}, },
        },
        allow_extra => 1, # allow additional args that OnTap accepts
    );
    my $vserver = $args{vserver};
    my $volume = $args{volume};
    my $start = delete $args{start};
    my $vserver_volume_efficiency_policy_hash = delete $args{vserver_volume_efficiency_policy_hash};

    my $vol_resource_obj_array = $self->get_resources( type => 'NACL::AppBuilder::Xgen::Resource::Volume',
                                                      data => { volume => $volume },
                                                    );
    my $vol_resource_obj ;
    foreach my $ele (@{ $vol_resource_obj_array}) {
        if ($ele->get_attribute('volume') eq $volume) {
            $vol_resource_obj = $ele;
        }
    }
    $Log->debug('My vol resource obj is '.Dumper($vol_resource_obj));
    $Log->log('call xgen volume efficiency new');
    my $start_time = time();

    my $volume_efficiency_resource_obj = NACL::AppBuilder::Xgen::Resource::Volume::Efficiency->new(start => $start,
                                                                                                %args);
    $Log->log('after xgen volume efficiency.  Call took ' . (time() - $start_time) . ' seconds');
    $volume_efficiency_resource_obj->add_parent_resources(resources => [$vol_resource_obj]);
    $self->add_resources(resources => [$volume_efficiency_resource_obj]);

}

###############################################################################
# DESCRIPTION: Generates a single Volume Efficiency Policy resource
#              in the configuration. This will generate
#              a volume efficiency policy resource on the vserver of the
#              volume that is passed in. The arguments for volume efficiency
#              policy are based on a template corresponding to the
#              volume_efficiency_policy argument.
# INPUT: volume - the volume with volume efficiency options that require
#           this policy
#        volume_efficiency_policy - the name of the policy
#        vserver_volume_efficiency_policy_hash - a hash mapping policy name
#           to cluster and vserver. This is used to ensure no duplicate
#           policies.
# OUTPUT: NONE
# Exceptions: NATE::BaseException based on the volume_efficiency_policy
#             template not existing.
###############################################################################
sub generate_volume_efficiency_policy_resource {
    my ($self, %args) = @_;
    my $volume = $args{'volume'};
    my $policy_name = $args{'volume_efficiency_policy'};
    my $vserver_volume_efficiency_policy_hash = $args{'vserver_volume_efficiency_policy_hash'};

    if (not defined $policy_name or
        defined $vserver_volume_efficiency_policy_hash->{$volume->get_attribute('cluster_name')}
                                                      ->{$volume->get_attribute('vserver')}
                                                      ->{$policy_name}) {
        return; # Do nothing because the volume efficiency policy already exists.
    }

    $vserver_volume_efficiency_policy_hash->{$volume->get_attribute('cluster_name')}
                                          ->{$volume->get_attribute('vserver')}
                                          ->{$policy_name} = 1;
    my ($vserver) = $volume->get_ancestors(type => "NACL::AppBuilder::Xgen::Resource::Vserver");

    my $volume_efficiency_policy_method = uc($policy_name) . '_VOLUME_EFFICIENCY_POLICY';

    my $method = $self->can($volume_efficiency_policy_method);
    if (defined $method) {
        #TODO - param check for these volume efficiency args
        my %volume_efficiency_policy_args = %{$self->$method};
        my $volume_efficiency_policy = NACL::AppBuilder::Xgen::Resource::Volume::EfficiencyPolicy->new('vserver' => $vserver->get_attribute('vserver'),
                                                                                                       %volume_efficiency_policy_args);
        $volume_efficiency_policy->add_parent_resources(resources => [$vserver]);
        $self->add_resources(resources => [$volume_efficiency_policy]);
        return $volume_efficiency_policy;
    } else {
        NATE::BaseException->throw("Tried to use a volume efficiency policy $policy_name that this config does not understand.");
    }
}
# get and/or generate vserver resource object
sub get_vserver_resource_object {
    my $self = shift;
    my %args = validate_with(
        params => \@_,
        spec => {
                vserver_name => { type => SCALAR, },
            },
    ); 

    my $vserver = $args{vserver_name};

    if (!$self->{vserver_lookup}) {
        my $vserver_resource_obj_array = $self->get_resources(
            type => 'NACL::AppBuilder::Xgen::Resource::Vserver',
            #data => { vserver => $source_vserver_name },
        );
        foreach my $ele (@{ $vserver_resource_obj_array}) {
            $Log->log('Got vserver='.$ele->get_attribute('vserver'));
            my $name = $ele->get_attribute('vserver');
            #save it by name
            $self->{vserver_lookup}->{$name} = $ele;
        }
    }

    return $self->{vserver_lookup}->{$vserver};

}

# generate 'ONE' volume efficiency policy resource
sub generate_one_volume_efficiency_policy_resource {
    my $self = shift;
    my %args = validate_with(
        params => \@_,
        spec => {
                vserver => { type => SCALAR, }, # parent
                policy => { type => SCALAR, },
                vserver_volume_efficiency_policy_hash => { type => HASHREF, default => {}, },
        },
        allow_extra => 1, # allow additional args that OnTap accepts
    );
    my $vserver = $args{vserver};
    my $policy = $args{policy};
    my $vserver_volume_efficiency_policy_hash = delete $args{'vserver_volume_efficiency_policy_hash'};

    #my $vserver_resource_obj_array = $self->get_resources( type => 'NACL::AppBuilder::Xgen::Resource::Vserver',
    #data => { vserver => $vserver },
    #);
    #my $vserver_resource_obj;
    #foreach my $ele (@{ $vserver_resource_obj_array}) {
    #if ($ele->get_attribute('vserver') eq $vserver) {
    #$vserver_resource_obj = $ele;
    #}
    #}
    my $vserver_resource_obj = $self->get_vserver_resource_object(vserver_name => $vserver);
    $Log->debug('My vserver resource obj is '.Dumper($vserver_resource_obj));

    if (not defined $policy or
        defined $vserver_volume_efficiency_policy_hash->{$vserver_resource_obj->get_attribute('cluster_name')}
                                                      ->{$vserver}
                                                      ->{$policy}) {
        return; # Do nothing because the volume efficiency policy already exists.
    }

    $vserver_volume_efficiency_policy_hash->{$vserver_resource_obj->get_attribute('cluster_name')}
                                          ->{$vserver}
                                          ->{$policy} = 1;

    my $volume_efficiency_policy_resource_obj = NACL::AppBuilder::Xgen::Resource::Volume::EfficiencyPolicy->new(%args);
    $volume_efficiency_policy_resource_obj->add_parent_resources(resources => [$vserver_resource_obj]);
    $self->add_resources(resources => [$volume_efficiency_policy_resource_obj]);
}

###############################################################################
# DESCRIPTION: Generates a single Job Schedule Cron resource
#              in the configuration. This will generate
#              a Job Schedule Cron resource on the cluster of the
#              volume that is passed in. The arguments for job schedule
#              cron are based on a template corresponding to the
#              job_schedule_cron argument.
# INPUT: volume - the volume from a snapmirror that requires this schedule
#        job_schedule_cron - the name of the schedule
#        cluster_job_schedule_cron_hash - a hash mapping schedule name
#           to cluster. This is used to ensure no duplicate schedules.
# OUTPUT: NONE
# Exceptions: NATE::BaseException based on the job_schedule_cron
#             template not existing.
###############################################################################

sub generate_job_schedule_cron_resource {
    my ($self, %args) = @_;
    my $volume = $args{'volume'};
    my $schedule_name = $args{'job_schedule_cron'};
    my $cluster_job_schedule_cron_hash = $args{'cluster_job_schedule_cron_hash'};

    if (not defined $schedule_name or
        defined $cluster_job_schedule_cron_hash->{$volume->get_attribute('cluster_name')}
                                               ->{$schedule_name}) {
        return; # Do nothing because the schedule already exists.
    }

    my $cluster_name = $volume->get_attribute('cluster_name');
    $cluster_job_schedule_cron_hash->{$volume->get_attribute('cluster_name')}
                                   ->{$schedule_name} = 1;
    my ($cluster) = $volume->get_ancestors(type => "NACL::AppBuilder::Xgen::Resource::Cluster");

    my $schedule_method = uc($schedule_name) . '_JOB_SCHEDULE';

    my $method = $self->can($schedule_method);
    if (defined $method) {
        #TODO - param check for these JOB SCHEDULE args
        my %job_schedule_args = %{$self->$method};

        # double-check to see if there is already an entry for job schedule in the resource list for the cluster $cluster
        my $job_schedule_resources = $self->get_resources(type => "NACL::AppBuilder::Xgen::Resource::JobScheduleCron");
        $Log->trace('Job schedule resources are '.Dumper($job_schedule_resources));

        # Filter the job schedule resources by name
        my $job_schedule_matches_hash = positive_filter(resources => $job_schedule_resources,
                                                        filter    => {'name' => $schedule_name});
        $Log->trace('Job schedule matches hash is '.Dumper($job_schedule_matches_hash));

        if (!$job_schedule_matches_hash) {
            # no entry in resource list for job schedule so continue and add it to the Resource List
            $Log->log("No entry in resource list for job schedule $schedule_name. Will add it to the resource list");
        } else {
            # Get the matches
            my $job_schedule_by_name = $job_schedule_matches_hash->{'matches'};

            # Loop through the matching job_schedules that matched
            for my $job_schedule (@{$job_schedule_by_name}) {

                # get the cluster for the job schedule
                my ($cluster_resource_obj) = $job_schedule->get_ancestors(type => "NACL::AppBuilder::Xgen::Resource::Cluster");
                if ($cluster_resource_obj->get_attribute('cluster_name') eq $cluster_name) {
                    $Log->warn("Entry in resource list for job schedule $schedule_name already exists. Will skip adding one more entry to the resource list");
                    return;
                }
            }
            $Log->log("No entry in resource list for job schedule $schedule_name for cluster $cluster_name. Will add it to the resource list");
        }

        $Log->log('Creating job schedule Args='.Dumper(%job_schedule_args));
        my $job_schedule = NACL::AppBuilder::Xgen::Resource::JobScheduleCron->new(%job_schedule_args);
        $job_schedule->add_parent_resources(resources => [$cluster]);
        $self->add_resources(resources => [$job_schedule]);
    } else {
        # TODO - check for a list of possible default schedules
        $Log->warn("schedule $schedule_name has no defined schedule method so we are assuming it is a default schedule and not making a schedule with that name.");
    }
}

# generate 'ONE' job schedule resource
sub generate_one_job_schedule_cron_resource {
    my $self = shift;
    my %args = validate_with(
        params => \@_,
        spec => {
                cluster => { type => SCALAR, },  # parent
                name => { type => SCALAR, },
                minute => { type => SCALAR, },
                cluster_job_schedule_cron_hash => { type => HASHREF, default => {}, },
        },
        allow_extra => 1, # allow additional args that OnTap accepts
    );
    my $cluster = delete $args{cluster};
    my $name = $args{name};
    my $minute = $args{minute};

    my $cluster_job_schedule_cron_hash = delete $args{'cluster_job_schedule_cron_hash'};
    if (not defined $name or
        defined $cluster_job_schedule_cron_hash->{$cluster}->{$name}) {
        return; # Do nothing because the schedule already exists.
    }
    $cluster_job_schedule_cron_hash->{$cluster}->{$name} = 1;

    # double-check to see if there is already an entry for job schedule in the resource list for the cluster $cluster
    my $job_schedule_resources = $self->get_resources(type => "NACL::AppBuilder::Xgen::Resource::JobScheduleCron");
    $Log->trace('Job schedule resources are '.Dumper($job_schedule_resources));

    # Filter the job schedule resources by name
    my $job_schedule_matches_hash = positive_filter(resources => $job_schedule_resources,
                                                    filter    => {'name' => $name});
                                                    # XXX - can I filter on the cluster too ? - that will save the checking later
    $Log->trace('Job schedule matches hash is '.Dumper($job_schedule_matches_hash));

    if (!$job_schedule_matches_hash) {
        # no entry in resource list for job schedule so continue and add it to the Resource List
        $Log->log("No entry in resource list for job schedule $name. Will add it to the resource list");
    } else {
        # Get the matches
        my $job_schedule_by_name = $job_schedule_matches_hash->{'matches'};

        # Loop through the matching job_schedules that matched
        for my $job_schedule (@{$job_schedule_by_name}) {

            # get the cluster for the job schedule
            my ($cluster_resource_obj) = $job_schedule->get_ancestors(type => "NACL::AppBuilder::Xgen::Resource::Cluster");
            if ($cluster_resource_obj->get_attribute('cluster_name') eq $cluster) {
                $Log->warn("Entry in resource list for job schedule $name already exists. Will skip adding one more entry to the resource list");
                return;
            }
        }
        $Log->log("No entry in resource list for job schedule $name for cluster $cluster. Will add it to the resource list");
    }

    my $cluster_resource_obj_array = $self->get_resources( type => 'NACL::AppBuilder::Xgen::Resource::Cluster',);
    my $cluster_resource_obj;
    foreach my $ele (@{ $cluster_resource_obj_array }) {
        if ($ele->get_attribute('cluster_name') eq $cluster) {
            $cluster_resource_obj = $ele;
        }
    }
    $Log->debug('My cluster resource obj is '.Dumper($cluster_resource_obj));

    $Log->log('Creating job schedule name='.$name.' Args='.Dumper(%args));
    my $job_schedule_resource_obj = NACL::AppBuilder::Xgen::Resource::JobScheduleCron->new(%args);
    $job_schedule_resource_obj->add_parent_resources(resources => [$cluster_resource_obj]);
    $self->add_resources(resources => [$job_schedule_resource_obj]);
}

###############################################################################
# DESCRIPTION: Generates a single Snapshot Policy resource
#              in the configuration. This will generate
#              a Snapshot Policy resource on the cluster of the
#              volume that is passed in. The arguments for job policy
#              cron are based on a template corresponding to the
#              snapshot_policy argument.
# INPUT: volume - the volume that requires this policy
#        snapshot_policy - the name of the policy
#        cluster_snapshot_policy_hash - a hash mapping policy name
#           to vserver and cluster. This is used to ensure no duplicate
#           policies.
# OUTPUT: NONE
# Exceptions: NATE::BaseException based on the snapshot_policy
#             template not existing.
###############################################################################
sub generate_snapshot_policy_resource {
    my ($self, %args) = @_;
    my $volume = $args{'volume'};
    my $policy_name = $args{'snapshot_policy'};
    my $vserver_snapshot_policy_hash = $args{'vserver_snapshot_policy_hash'};

    if (not defined $policy_name or
        defined $vserver_snapshot_policy_hash->{$volume->get_attribute('cluster_name')}
                                             ->{$volume->get_attribute('vserver')}
                                             ->{$policy_name}) {
        return; # Do nothing because the snapshot policy already exists.
    }

    $vserver_snapshot_policy_hash->{$volume->get_attribute('cluster_name')}
                                 ->{$volume->get_attribute('vserver')}
                                 ->{$policy_name} = 1;
    my ($vserver) = $volume->get_ancestors(type => "NACL::AppBuilder::Xgen::Resource::Vserver");

    my $policy_method = uc($policy_name) . '_SNAPSHOT_POLICY';

    my $method = $self->can($policy_method);
    if (defined $method) {
        #TODO - param check for these SNAPSHOT POLICY args
        my %snapshot_policy_args = %{$self->$method};
        my $snapshot_policy = NACL::AppBuilder::Xgen::Resource::SnapshotPolicy->new('vserver' => $vserver->get_attribute('vserver'),
                                                                                    'xgen_nacl_cs_validation' => 0,
                                                                                    %snapshot_policy_args);
        $snapshot_policy->add_parent_resources(resources => [$vserver]);
        $self->add_resources(resources => [$snapshot_policy]);
    } else {
        $Log->comment("Snapshot Policy $policy_name has no defined policy method so we are assuming it is a default policy and not making a policy with that name.");
    }
}

# generate 'ONE' snapshot policy
sub generate_one_snapshot_policy_resource {
    my $self = shift;
    my %args = validate_with(
        params => \@_,
        spec => {
            policy => { type => SCALAR, },
            vserver => { type => SCALAR, optional => 1 }, # parent
            enabled => { type => SCALAR, },
            schedule1 => { type => SCALAR, },
            count1 => { type => SCALAR, },
            vserver_snapshot_policy_hash => { type => HASHREF, default => {} },
        },
        allow_extra => 1, # allow additional args that OnTap accepts
    );
    my $vserver = $args{vserver};
    my $policy = $args{policy};
    my $vserver_snapshot_policy_hash = delete $args{'vserver_snapshot_policy_hash'};

    my $vserver_resource_obj_array = $self->get_resources( type => 'NACL::AppBuilder::Xgen::Resource::Vserver',
                                                            data => { vserver => $vserver },
                                                            );
    my $vserver_resource_obj;
    foreach my $ele (@{ $vserver_resource_obj_array}) {
        if ($ele->get_attribute('vserver') eq $vserver) {
            $vserver_resource_obj = $ele;
        }
    }
    $Log->debug('My vserver resource obj is '.Dumper($vserver_resource_obj));

    if (not defined $policy or
        defined $vserver_snapshot_policy_hash->{$vserver_resource_obj->get_attribute('cluster_name')}
                                             ->{$vserver}
                                             ->{$policy}) {
        return; # Do nothing because the snapshot policy already exists.
    }
    $vserver_snapshot_policy_hash->{$vserver_resource_obj->get_attribute('cluster_name')}->{$vserver}->{$policy} = 1;

    my $snapshot_policy_resource_obj = NACL::AppBuilder::Xgen::Resource::SnapshotPolicy->new(
                                                                            'xgen_nacl_cs_validation' => 0,
                                                                            %args);

    $snapshot_policy_resource_obj->add_parent_resources(resources => [$vserver_resource_obj]);
    $self->add_resources(resources => [$snapshot_policy_resource_obj]);
}

###############################################################################
# DESCRIPTION: Generates a single snapmirror Policy resource
#              in the configuration. This will generate
#              a snapmirror Policy resource on the cluster of the
#              volume that is passed in. The arguments for job policy
#              cron are based on a template corresponding to the
#              snapmirror_policy argument. Also generates Snapmirror Policy
#              Rule resources in the configuration based on rules defined in the
#              snapmirror policy arguments.
# INPUT: volume - the volume from a snapmirror that requires this policy
#        snapmirror_policy - the name of the policy
#        vserver_sm_policy_hash - a hash mapping policy name
#           to vserver and cluster. This is used to ensure no duplicate
#           policies.
# OUTPUT: NONE
# Exceptions: NATE::BaseException based on the snapmirror_policy
#             template not existing.
###############################################################################
sub generate_snapmirror_policy_resource {
    my ($self, %args) = @_;
    my $volume = $args{'volume'};
    my $policy_name = $args{'snapmirror_policy'};
    my $vserver_sm_policy_hash = $args{'vserver_sm_policy_hash'};
    if (not defined $policy_name or
        defined $vserver_sm_policy_hash->{$volume->get_attribute('cluster_name')}
                                             ->{$volume->get_attribute('vserver')}
                                             ->{$policy_name}) {
        return; # Do nothing because the snapmirror policy already exists.
    }

    $vserver_sm_policy_hash->{$volume->get_attribute('cluster_name')}
                                 ->{$volume->get_attribute('vserver')}
                                 ->{$policy_name} = 1;
    my ($vserver) = $volume->get_ancestors(type => "NACL::AppBuilder::Xgen::Resource::Vserver");

    my $policy_method = uc($policy_name) . '_SNAPMIRROR_POLICY';

    my $method = $self->can($policy_method);
    if (defined $method) {
        #TODO - param check for these SNAPSHOT POLICY args
        my %snapmirror_policy_args = %{$self->$method(%args)};
        my $snapmirror_policy_rules = delete $snapmirror_policy_args{'xgen_snapmirror_policy_rules'};
        my $snapmirror_policy = NACL::AppBuilder::Xgen::Resource::SnapmirrorPolicy->new('vserver' => $vserver->get_attribute('vserver'),
                                                                                        %snapmirror_policy_args);
        $snapmirror_policy->add_parent_resources(resources => [$vserver]);
        $self->add_resources(resources => [$snapmirror_policy]);

        foreach my $snapmirror_policy_rule_args (@{$snapmirror_policy_rules}) {
            my $snapmirror_policy_rule = NACL::AppBuilder::Xgen::Resource::SnapmirrorPolicyRule->new('vserver' => $vserver->get_attribute('vserver'),
                                                                                                     'policy'  => $policy_name,
                                                                                                     %{$snapmirror_policy_rule_args});
            $snapmirror_policy_rule->add_parent_resources(resources => [$snapmirror_policy]);
            $self->add_resources(resources => [$snapmirror_policy_rule]);
        }
    } else {
        $Log->comment("Snapmirror Policy $policy_name has no defined policy method so we are assuming it is a default policy and not making a policy with that name.");
    }
}

# generate 'ONE' snapmirror policy resource
sub generate_one_snapmirror_policy_resource {
    my $self = shift;
    my %args = validate_with(
        params => \@_,
        spec => {
            policy => { type => SCALAR, },
            vserver => { type => SCALAR, }, # parent
            vserver_sm_policy_hash => { type => HASHREF, default => {} },
        },
        allow_extra => 1, # allow additional args that OnTap accepts
    );
    my $vserver = $args{vserver};
    my $vserver_resource_obj_array = $self->get_resources( type => 'NACL::AppBuilder::Xgen::Resource::Vserver',
                                                            data => { vserver => $vserver },
                                                            );
    #my $vserver_resource_obj = @{ $vserver_resource_obj_array }[0]; # get the first elem out
    my $vserver_resource_obj;
    foreach my $ele (@{ $vserver_resource_obj_array}) {
        if ($ele->get_attribute('vserver') eq $vserver) {
            $vserver_resource_obj = $ele;
        }
    }
    $Log->debug('My vserver resource obj is '.Dumper($vserver_resource_obj));
    my ($cluster_resource_obj) = $vserver_resource_obj->get_ancestors(type => 'NACL::AppBuilder::Xgen::Resource::Cluster');
    $Log->debug('My cluster resource obj is '.Dumper($cluster_resource_obj));

    my $policy = $args{policy};
    my $vserver_sm_policy_hash = delete $args{vserver_sm_policy_hash};
    my $cluster_name = $cluster_resource_obj->get_attribute('cluster_name');
    if (not defined $policy or
        defined $vserver_sm_policy_hash->{$cluster_name}->{$vserver}->{$policy}) {
        return; # Do nothing because the snapmirror policy already exists.
    }
    $vserver_sm_policy_hash->{$cluster_name}->{$vserver}->{$policy} = 1;

    $Log->log('Creating snapmirror policy policy='.$policy.' Args='.Dumper(%args));
    my $snapmirror_policy_resource_obj = NACL::AppBuilder::Xgen::Resource::SnapmirrorPolicy->new(%args);
    $snapmirror_policy_resource_obj->add_parent_resources(resources => [$vserver_resource_obj]);
    $self->add_resources(resources => [$snapmirror_policy_resource_obj]);
}


# generate 'ONE' snapmirror policy rule resource
sub generate_one_snapmirror_policy_rule_resource {
    my $self = shift;
    my %args = validate_with(
        params => \@_,
        spec => {
            policy => { type => SCALAR, }, # parent
            'snapmirror-label' => { type => SCALAR, },
            keep => { type => SCALAR, },
            vserver => { type => SCALAR, },
        },
        allow_extra => 1, # allow additional args that OnTap accepts
    );
    my $policy = $args{policy};
    my $snapmirror_policy_resource_obj_array = $self->get_resources( type => 'NACL::AppBuilder::Xgen::Resource::SnapmirrorPolicy',
                                                            data => { policy => $policy },
                                                            );
    #my $snapmirror_policy_resource_obj = @{ $snapmirror_policy_resource_obj_array }[0]; # get the first elem out
    my $snapmirror_policy_resource_obj;
    foreach my $ele (@{ $snapmirror_policy_resource_obj_array}) {
        if ($ele->get_attribute('policy') eq $policy) {
            $snapmirror_policy_resource_obj = $ele;
        }
    }
    $Log->debug('My snapmirror policy resource obj is '.Dumper($snapmirror_policy_resource_obj));

    $Log->log('Creating snapmirror policy rule. rule='.$args{'snapmirror-label'}.' policy='.$policy.' Args='.Dumper(%args));
    my $snapmirror_policy_rule_resource_obj = NACL::AppBuilder::Xgen::Resource::SnapmirrorPolicyRule->new(%args);
    $snapmirror_policy_rule_resource_obj->add_parent_resources(resources => [$snapmirror_policy_resource_obj]);
    $self->add_resources(resources => [$snapmirror_policy_rule_resource_obj]);
}

###############################################################################
# DESCRIPTION: Generates snapmirror resources
#              in the configuration. This method calls all the config
#              generation methods to generate snapmirrors in the configuration.
#              This also creates and holds the references to global data used
#              by some other methods to ensure no duplicates.
# INPUT: NONE
# OUTPUT: NONE
# Exceptions: None thrown directly by this method.
###############################################################################
sub generate_snapmirror_resources {
    my ($self, %args) = @_;

    my $volume_resources = $self->get_resources(type => "NACL::AppBuilder::Xgen::Resource::Volume");
    my @temp_volumes = @{$volume_resources};
    my $available_volumes = \@temp_volumes;
    # A mapping of vservers to clusters.
    my $cluster_vserver_hash = {};
    # A mapping of vservers to snapmirror policies.
    my $vserver_sm_policy_hash = {};
    # A mapping of vservers to snapshot policies.
    my $vserver_snapshot_policy_hash = {};
    # A mapping of clusters to job schedules.
    my $cluster_job_schedule_cron_hash = {};

    my $config_id = 0;
    my $created_snapmirrors = 0;

    $Log->comment("Generating FANIN Snapmirror configs.");
    # Create all the cascade snapmirror configs that were defined.
    $self->generate_fanin_snapmirror_configs(available_volumes                 => $available_volumes,
                                             created_snapmirrors               => \$created_snapmirrors,
                                             'vserver_sm_policy_hash'          => $vserver_sm_policy_hash,
                                             'vserver_snapshot_policy_hash'    => $vserver_snapshot_policy_hash,
                                             'cluster_job_schedule_cron_hash'  => $cluster_job_schedule_cron_hash,
                                             config_id                         => \$config_id);

    $Log->comment("Generating FANOUT Snapmirror configs.");
    # Create all the cascade snapmirror configs that were defined.
    $self->generate_fanout_snapmirror_configs(available_volumes                 => $available_volumes,
                                              created_snapmirrors               => \$created_snapmirrors,
                                              'vserver_sm_policy_hash'          => $vserver_sm_policy_hash,
                                              'vserver_snapshot_policy_hash'    => $vserver_snapshot_policy_hash,
                                              'cluster_job_schedule_cron_hash'  => $cluster_job_schedule_cron_hash,
                                              config_id                         => \$config_id);

    $Log->comment("Generating BIDIRECTIONAL Snapmirror configs.");
    # Create all the cascade snapmirror configs that were defined.
    $self->generate_bidirectional_snapmirror_configs(available_volumes                 => $available_volumes,
                                                     created_snapmirrors               => \$created_snapmirrors,
                                                     'vserver_sm_policy_hash'          => $vserver_sm_policy_hash,
                                                     'vserver_snapshot_policy_hash'    => $vserver_snapshot_policy_hash,
                                                     'cluster_job_schedule_cron_hash'  => $cluster_job_schedule_cron_hash,
                                                     config_id                         => \$config_id);

    $Log->comment("Generating CASCADE Snapmirror configs.");
    # Create all the cascade snapmirror configs that were defined.
    $self->generate_cascade_snapmirror_configs(available_volumes                 => $available_volumes,
                                               created_snapmirrors               => \$created_snapmirrors,
                                               'vserver_sm_policy_hash'          => $vserver_sm_policy_hash,
                                               'vserver_snapshot_policy_hash'    => $vserver_snapshot_policy_hash,
                                               'cluster_job_schedule_cron_hash'  => $cluster_job_schedule_cron_hash,
                                               config_id                         => \$config_id);

    $Log->comment("Generating UNIDIRECTIONAL Snapmirror config.");
    # Create all the unidirectional snapmirrors that were not taken by other configs.
    # Unidirectional does not mean all the snapmirrors will be c1-c2 as some can be c1-c1 or a mix.
    # It just means that we do not specifically choose the direction except that we favor cluster to cluster snapmirrors.
    $self->generate_unidirectional_snapmirror_config(snapmirror_count                 => $self->param('TOTAL_SNAPMIRROR_COUNT') - $created_snapmirrors,,
                                              available_volumes                 => $available_volumes,
                                              created_snapmirrors               => \$created_snapmirrors,
                                              'vserver_sm_policy_hash'          => $vserver_sm_policy_hash,
                                              'vserver_snapshot_policy_hash'    => $vserver_snapshot_policy_hash,
                                              'cluster_job_schedule_cron_hash'  => $cluster_job_schedule_cron_hash,
                                              config_id                         => \$config_id);



    # Some snapmirror src/dst volumes may have incorrect names if there are
    # multiple snapmirrors related with a particular volume (src/dst).
    # After snapmirrors are created, go back and make sure they have the correct volume names
    #
    $Log->comment("Updating src/dst Snapmirror volume names.");

    my @volumes = @{$volume_resources};

    foreach my $vol (@volumes) {
        #$Log->comment('volume:' . $vol->get_attribute('volume'));
        my $volname = $vol->get_attribute('volume');

        my @sm = $vol->get_descendants(type => "NACL::AppBuilder::Xgen::Resource::Snapmirror");
        foreach my $sm (@sm) {
            #$Log->comment('  sm:'  . dumper($sm));
            my $source_vol = $sm->get_attribute('source-volume');
            my $dest_vol   = $sm->get_attribute('destination-volume');
            #$Log->comment(' srcvol:' . $source_vol);
            #$Log->comment(' dstvol:' . $dest_vol);

            foreach my $smvol ($source_vol, $dest_vol) {
                if ($volname =~ /$smvol/) {
                    if ($smvol ne $volname) {
                        $Log->comment("updating snapmirror volume name $smvol to $volname");
                        $smvol = $volname;
                    }
                }
            }

            $sm->set_metadata_attributes('source-volume' => $source_vol);
            $sm->set_metadata_attributes('destination-volume' => $dest_vol);
        }


    }

    #need to do it for snapmirror initialize also
    $Log->comment("Updating src/dst Snapmirror Initialize volume names.");

    @volumes = @{$volume_resources};

    foreach my $vol (@volumes) {
        #$Log->comment('volume:' . $vol->get_attribute('volume'));
        my $volname = $vol->get_attribute('volume');

        my @sm = $vol->get_descendants(type => "NACL::AppBuilder::Xgen::Resource::SnapmirrorInitialize");
        foreach my $sm (@sm) {
            #$Log->comment('  sm init:'  . dumper($sm));
            my $source_vol = $sm->get_attribute('source-volume');
            my $dest_vol   = $sm->get_attribute('destination-volume');
            #$Log->comment(' srcvol:' . $source_vol);
            #$Log->comment(' dstvol:' . $dest_vol);

            foreach my $smvol ($source_vol, $dest_vol) {
                if ($volname =~ /$smvol/) {
                    if ($smvol ne $volname) {
                        $Log->comment("updating snapmirror initialize volume name $smvol to $volname");
                        $smvol = $volname;
                    }
                }
            }

            $sm->set_metadata_attributes('source-volume' => $source_vol);
            $sm->set_metadata_attributes('destination-volume' => $dest_vol);
        }


    }


}

###############################################################################
# DESCRIPTION: Generates snapmirror resources in the configuration. This method
#              generates snapmirrors in the configuration based on the rules
#              for a bidirectional config.  Bidirectional in this method means
#              two snapmirrors that reverse the source and destination vservers
#              VS1->VS2 and VS2->VS1.
#              It also prefers vservers on different clusters.
#              Creates a number of configs based on the
#              TOTAL_BIDIRECTIONAL_CONFIG_COUNT constant and the percentage of
#              configs with each type of layout is based on the
#              BIDIRECTIONAL_TYPE_PERCENT_LAYOUT constant.
# INPUT: snapmirror_count - number of snapmirrors left to be created
#        available_volumes - reference to an array of available volumes.
#        created_snapmirrors - a scalar reference to the number of already
#           created snapmirrors.
#        config_id - The next config id to use for snapmirror configs.
#        vserver_sm_policy_hash - a hashref to a mapping of snapmirror
#           policy name to vserver and cluster.
#        vserver_snapshot_policy_hash - a hashref to a mapping of snapshot
#           policy name to vserver and cluster.
#        cluster_job_schedule_cron_hash - a hashref to a mapping of schedules
#           to cluster.
# OUTPUT: NONE
# Exceptions: None thrown directly by this method.
###############################################################################
sub generate_bidirectional_snapmirror_configs {
    my ($self, %args) = @_;
    my $snapmirror_count = $args{'snapmirror_count'};
    my $available_volumes = $args{'available_volumes'};
    my $created_snapmirrors = $args{'created_snapmirrors'};
    my $config_id           = $args{'config_id'};
    # A mapping of vservers to snapmirror policies.
    my $vserver_sm_policy_hash = $args{'vserver_sm_policy_hash'};
    # A mapping of vservers to snapshot policies.
    my $vserver_snapshot_policy_hash = $args{'vserver_snapshot_policy_hash'};
    # A mapping of clusters to job schedules.
    my $cluster_job_schedule_cron_hash = $args{'cluster_job_schedule_cron_hash'};

    # This will get the layout in terms of numbers instead of percents based on the total object count passed in.
    my $layout = convert_percent_layout_to_count_layout('layout'        => $self->param('BIDIRECTIONAL_TYPE_PERCENT_LAYOUT'),
                                                        'total_objects' => $self->param('TOTAL_BIDIRECTIONAL_CONFIG_COUNT'));
    # For the destination we start with no must use clusters.
    my @destination_must_use_clusters = ();
    # For the destination we start with no must use vservers.
    my @destination_must_use_vservers = ();
    # Loop through TOTAL_CASCADE_CONFIG_COUNT and create those configs.
    for (my $i = 0; $i < $self->param('TOTAL_BIDIRECTIONAL_CONFIG_COUNT'); $i++) {
        my $config_layout = get_next_layout_element('iterator' => $i,       # This is not required but allows the get_next_layout_element to
                                                                            # rotate through the layout instead of exhausting the first, then
                                                                            # the second, etc.
                                                    'layout'   => $layout); # layout is the numbers layout we got earlier.

        # Config the config layout into an array of the types of snapmirrors in this cascade config.
        my @config_type_array = split(';',$config_layout);

        my %last_resort_vservers_hash = ();
        my %last_resort_clusters_hash = ();


        # For the first destination volume it can come from anywhere.
        my @first_vservers_volumes = $self->get_available_snapmirror_volume('available_volumes'   => $available_volumes,
                                                                            'all_from_same_vserver' => 1,
                                                                            'volume_count'          => 2);

        my $first_vservers_cluster = $first_vservers_volumes[0]->get_attribute('cluster_name');
        my $first_vserver = $first_vservers_volumes[0]->get_attribute('vserver');

        # For the first destination volume it can come from anywhere.
        my @second_vservers_volumes = $self->get_available_snapmirror_volume('available_volumes'   => $available_volumes,
                                                                             'all_from_same_vserver' => 1,
                                                                             'volume_count'          => 2,
                                                                             'unuseable_vservers'    => [$first_vserver],
                                                                             'last_resort_clusters'  => [$first_vservers_cluster]);

        #Loop through and set the metadata on all the volumes.
        for (my $j = 0; $j < 2; $j++) {

            my $first_vservers_volume  = $first_vservers_volumes[$j];

            my $second_vservers_volume = $second_vservers_volumes[$j];

            # Update some metadata on the volume.
            $first_vservers_volume->set_metadata_attributes('sm_config'        => 'bidirectional',
                                                            'sm_config_layout' => $config_layout,
                                                            'sm_config_id'     => $$config_id,);

            # Update some metadata on the volume.
            $second_vservers_volume->set_metadata_attributes('sm_config'     => 'bidirectional',
                                                             'sm_config_layout' => $config_layout,
                                                             'sm_config_id'  => $$config_id,);
        }

        # Now we make our bidirectional config.
        # v1->v2 and v2->v1 based on the config type.
        # The SM Type is the config type arrays current index
        my $SM_TYPE = $config_type_array[0];
        $self->generate_snapmirror_resource('source_volume'                  => $first_vservers_volumes[0],
                                            'destination_volume'             => $second_vservers_volumes[0],
                                            'sm_type'                        => $SM_TYPE,
                                            'vserver_sm_policy_hash'         => $vserver_sm_policy_hash,
                                            'vserver_snapshot_policy_hash'   => $vserver_snapshot_policy_hash,
                                            'cluster_job_schedule_cron_hash' => $cluster_job_schedule_cron_hash,,
                                            'xgen_metadata_args'             => {'sm_config'      => 'bidirectional',
                                                                                 'sm_config_layout' => $config_layout,
                                                                                 'sm_config_id' => $$config_id});
        $$created_snapmirrors++;

        $SM_TYPE = $config_type_array[1];
        $self->generate_snapmirror_resource('source_volume'                  => $second_vservers_volumes[1],
                                            'destination_volume'             => $first_vservers_volumes[1],
                                            'sm_type'                        => $SM_TYPE,
                                            'vserver_sm_policy_hash'         => $vserver_sm_policy_hash,
                                            'vserver_snapshot_policy_hash'   => $vserver_snapshot_policy_hash,
                                            'cluster_job_schedule_cron_hash' => $cluster_job_schedule_cron_hash,,
                                            'xgen_metadata_args'             => {'sm_config'      => 'bidirectional',
                                                                                 'sm_config_layout' => $config_layout,
                                                                                 'sm_config_id' => $$config_id});
        $$created_snapmirrors++;
        # We finished the sm config so we need to update the config id
        $$config_id++;
    }
}

###############################################################################
# DESCRIPTION: Generates snapmirror resources in the configuration. This method
#              generates snapmirrors in the configuration based on the rules
#              for a fanin config.  Fanin in this method means
#              multiple snapmirrors that go to the same destination vserver
#              VS2->VS1 VS3->VS1.
#              It also prefers vservers on different clusters.
#              Creates a number of configs based on the
#              TOTAL_FANIN_CONFIG_COUNT constant and the percentage of
#              configs with each type of layout is based on the
#              FANIN_TYPE_PERCENT_LAYOUT constant.
# INPUT: snapmirror_count - number of snapmirrors left to be created
#        available_volumes - reference to an array of available volumes.
#        created_snapmirrors - a scalar reference to the number of already
#           created snapmirrors.
#        config_id - The next config id to use for snapmirror configs.
#        vserver_sm_policy_hash - a hashref to a mapping of snapmirror
#           policy name to vserver and cluster.
#        vserver_snapshot_policy_hash - a hashref to a mapping of snapshot
#           policy name to vserver and cluster.
#        cluster_job_schedule_cron_hash - a hashref to a mapping of schedules
#           to cluster.
# OUTPUT: NONE
# Exceptions: None thrown directly by this method.
###############################################################################
sub generate_fanin_snapmirror_configs {
    my ($self, %args) = @_;
    my $snapmirror_count = $args{'snapmirror_count'};
    my $available_volumes = $args{'available_volumes'};
    my $created_snapmirrors = $args{'created_snapmirrors'};
    my $config_id           = $args{'config_id'};
    # A mapping of vservers to snapmirror policies.
    my $vserver_sm_policy_hash = $args{'vserver_sm_policy_hash'};
    # A mapping of vservers to snapshot policies.
    my $vserver_snapshot_policy_hash = $args{'vserver_snapshot_policy_hash'};
    # A mapping of clusters to job schedules.
    my $cluster_job_schedule_cron_hash = $args{'cluster_job_schedule_cron_hash'};

    # This will get the layout in terms of numbers instead of percents based on the total object count passed in.
    my $layout = convert_percent_layout_to_count_layout('layout'        => $self->param('FANIN_TYPE_PERCENT_LAYOUT'),
                                                        'total_objects' => $self->param('TOTAL_FANIN_CONFIG_COUNT'));
    # For the destination we start with no must use clusters.
    my @destination_must_use_clusters = ();
    # For the destination we start with no must use vservers.
    my @destination_must_use_vservers = ();
    # Loop through TOTAL_CASCADE_CONFIG_COUNT and create those configs.
    for (my $i = 0; $i < $self->param('TOTAL_FANIN_CONFIG_COUNT'); $i++) {
        my $config_layout = get_next_layout_element('iterator' => $i,     # This is not required but allows the get_next_layout_element to
                                                                          # rotate through the layout instead of exhausting the first, then
                                                                          # the second, etc.
                                                    'layout'   => $layout); # layout is the numbers layout we got earlier.

        # Config the config layout into an array of the types of snapmirrors in this cascade config.
        my @config_type_array = split(';',$config_layout);

        my %last_resort_vservers_hash = ();
        my %last_resort_clusters_hash = ();


        # For the first destination volume it can come from anywhere.
        my @destination_volumes = $self->get_available_snapmirror_volume('available_volumes'   => $available_volumes,
                                                                         'all_from_same_vserver' => 1,
                                                                         'volume_count'          => scalar (@config_type_array));

        my $destination_cluster = $destination_volumes[0]->get_attribute('cluster_name');
        my $destination_vserver = $destination_volumes[0]->get_attribute('vserver');

        my @source_volumes = ();
        # Now we need to find our source_volumes.
        # We will make the destination cluster always a last_resort
        # and also try to rotate through any remaining clusters.
        for (my $j = 0; $j < scalar (@config_type_array); $j++) {
            my %clusters_hash = ();
            foreach my $volume (@{$available_volumes}) {
                $clusters_hash{$volume->get_attribute('cluster_name')}++;
            }
            my @clusters = keys (%clusters_hash);
            # If there are clusters in the
            my %temp_hash = ();
            foreach my $cluster (@clusters) {
                if(defined $last_resort_clusters_hash{$cluster}) {
                    $temp_hash{$cluster} = $last_resort_clusters_hash{$cluster};
                }
            }
            %last_resort_clusters_hash = %temp_hash;
            my @last_resort_clusters = ();
            # Build the last_resort_clusters based on the home many times each has been used.
            my $times_used = 0;
            foreach my $cluster (keys %last_resort_clusters_hash) {
                if ($last_resort_clusters_hash{$cluster} == $times_used) {
                    # If this is used the same number of times then add it to the list
                    push(@last_resort_clusters,$cluster);
                } elsif ($last_resort_clusters_hash{$cluster} > $times_used) {
                    # If this is used more number of times then make it the list and update
                    # the times used. This way we can have the last resorts be the highest used
                    # clusters.
                    @last_resort_clusters = ($cluster);
                    $times_used = $last_resort_clusters_hash{$cluster};
                } # Else we ignore since it is not the most used.
            }

            # make sure to always have the destination cluster in the last resort list.
            push(@last_resort_clusters,$destination_cluster);

            # if the last resort clusters is >= the available clusters that means that all clusters
            # are in the last resort so we can assume that we only want to have the destination cluster
            # as the last resort.
            if(scalar (@last_resort_clusters) >= scalar (@clusters)) {
                @last_resort_clusters = ($destination_cluster);
            }
            my @last_resort_vservers = ();
            # Build the last_resort_clusters based on the home many times each has been used.
            $times_used = 0;
            foreach my $vserver (keys %last_resort_vservers_hash) {
                if ($last_resort_vservers_hash{$vserver} == $times_used) {
                    # If this is used the same number of times then add it to the list
                    push(@last_resort_vservers,$vserver);
                } elsif ($last_resort_vservers_hash{$vserver} > $times_used) {
                    # If this is used more number of times then make it the list and update
                    # the times used. This way we can have the last resorts be the highest used
                    # vservers.
                    @last_resort_vservers = ($vserver);
                    $times_used = $last_resort_vservers_hash{$vserver};
                } # Else we ignore since it is not the most used.
            }

            # get a volume based on the criteria from the previous volumes.
            # If last_resort and unuseable have the same vservers unuseable takes precendence.
            my ($volume) = $self->get_available_snapmirror_volume('available_volumes'  => $available_volumes,
                                                                  'last_resort_clusters' => \@last_resort_clusters,
                                                                  'last_resort_vservers' => \@last_resort_vservers,
                                                                  'unuseable_vservers'   => [$destination_vserver]);
            # push the volume into our list of volumes for the chain.
            push(@source_volumes,$volume);

            # Now we need to update the criteria based on the volume found.
            $last_resort_vservers_hash{$volume->get_attribute('vserver')}++;
            $last_resort_clusters_hash{$volume->get_attribute('cluster_name')}++;
        }

        for (my $j = 0; $j < scalar (@config_type_array); $j++) {
            # The source volume is the volume at the current index.
            my $source_volume      = $source_volumes[$j];
            # The destination volume is the volume at the current index plus 1.
            my $destination_volume = $destination_volumes[$j];

            # Update some metadata on the volume.
            $source_volume->set_metadata_attributes('sm_config'        => 'fanin',
                                                    'sm_config_layout' => $config_layout,
                                                    'fanin_index'      => $j,
                                                    'sm_config_id'     => $$config_id,);

            # Update some metadata on the volume.
            $destination_volume->set_metadata_attributes('sm_config'     => 'fanin',
                                                         'sm_config_layout' => $config_layout,
                                                         'fanin_index' => $j,
                                                         'sm_config_id'  => $$config_id,);

            # The SM Type is the config type arrays current index
            my $SM_TYPE = $config_type_array[$j];
            $self->generate_snapmirror_resource('source_volume'                  => $source_volume,
                                                'destination_volume'             => $destination_volume,
                                                'sm_type'                        => $SM_TYPE,
                                                'vserver_sm_policy_hash'         => $vserver_sm_policy_hash,
                                                'vserver_snapshot_policy_hash'   => $vserver_snapshot_policy_hash,
                                                'cluster_job_schedule_cron_hash' => $cluster_job_schedule_cron_hash,,
                                                'xgen_metadata_args'             => {'sm_config'      => 'fanin',
                                                                                     'sm_config_layout' => $config_layout,
                                                                                     'fanin_index'  => $j,
                                                                                     'sm_config_id' => $$config_id});
            $$created_snapmirrors++;
        }
        # We finished the sm config so we need to update the config id
        $$config_id++;
    }
}

###############################################################################
# DESCRIPTION: Generates snapmirror resources in the configuration. This method
#              generates snapmirrors in the configuration based on the rules
#              for a fanout config.  Fanout in this method means
#              multiple snapmirrors that also have the same source volume.
#              VS1:Vol1->VS2:Vol2 and VS1:Vol1->VS3:Vol3
#              It also prefers volumes on different clusters, vserver, and
#              aggregates.
#              Creates a number of configs based on the
#              TOTAL_FANOUT_CONFIG_COUNT constant and the percentage of
#              configs with each type of layout is based on the
#              FANOUT_TYPE_PERCENT_LAYOUT constant.
# INPUT: snapmirror_count - number of snapmirrors left to be created
#        available_volumes - reference to an array of available volumes.
#        created_snapmirrors - a scalar reference to the number of already
#           created snapmirrors.
#        config_id - The next config id to use for snapmirror configs.
#        vserver_sm_policy_hash - a hashref to a mapping of snapmirror
#           policy name to vserver and cluster.
#        vserver_snapshot_policy_hash - a hashref to a mapping of snapshot
#           policy name to vserver and cluster.
#        cluster_job_schedule_cron_hash - a hashref to a mapping of schedules
#           to cluster.
# OUTPUT: NONE
# Exceptions: None thrown directly by this method.
###############################################################################
sub generate_fanout_snapmirror_configs {
    my ($self, %args) = @_;
    my $snapmirror_count = $args{'snapmirror_count'};
    my $available_volumes = $args{'available_volumes'};
    my $created_snapmirrors = $args{'created_snapmirrors'};
    my $config_id           = $args{'config_id'};
    # A mapping of vservers to snapmirror policies.
    my $vserver_sm_policy_hash = $args{'vserver_sm_policy_hash'};
    # A mapping of vservers to snapshot policies.
    my $vserver_snapshot_policy_hash = $args{'vserver_snapshot_policy_hash'};
    # A mapping of clusters to job schedules.
    my $cluster_job_schedule_cron_hash = $args{'cluster_job_schedule_cron_hash'};

    # This will get the layout in terms of numbers instead of percents based on the total object count passed in.
    my $layout = convert_percent_layout_to_count_layout('layout'        => $self->param('FANOUT_TYPE_PERCENT_LAYOUT'),
                                                        'total_objects' => $self->param('TOTAL_FANOUT_CONFIG_COUNT'));
    # For the destination we start with no must use clusters.
    my @destination_must_use_clusters = ();
    # For the destination we start with no must use vservers.
    my @destination_must_use_vservers = ();
    # Loop through TOTAL_CASCADE_CONFIG_COUNT and create those configs.
    for (my $i = 0; $i < $self->param('TOTAL_FANOUT_CONFIG_COUNT'); $i++) {
        my $config_layout = get_next_layout_element('iterator' => $i,     # This is not required but allows the get_next_layout_element to
                                                                          # rotate through the layout instead of exhausting the first, then
                                                                          # the second, etc.
                                                    'layout'   => $layout); # layout is the numbers layout we got earlier.

        # Config the config layout into an array of the types of snapmirrors in this cascade config.
        my @config_type_array = split(';',$config_layout);

        my %last_resort_aggregates_hash = ();
        my %last_resort_vservers_hash = ();
        my %last_resort_clusters_hash = ();


        # For the source volume it can come from anywhere.
        my ($source_volume) = $self->get_available_snapmirror_volume('available_volumes' => $available_volumes,);

        # Update some metadata on the volume.
        $source_volume->set_metadata_attributes('sm_config'        => 'fanout',
                                                'sm_config_layout' => $config_layout,
                                                'fanout_index'     => 0,
                                                'sm_config_id'     => $$config_id,);

        my $source_cluster = $source_volume->get_attribute('cluster_name');
        my $source_vserver = $source_volume->get_attribute('vserver');
        my $source_aggregate = $source_volume->get_attribute('aggregate');
        my @destination_volumes = ();
        # Now we need to find our destination_volumes.
        # We will make the source cluster always a last_resort
        # and also try to rotate through any remaining clusters.
        for (my $j = 0; $j < scalar (@config_type_array); $j++) {
            my %clusters_hash = ();
            foreach my $volume (@{$available_volumes}) {
                $clusters_hash{$volume->get_attribute('cluster_name')}++;
            }
            my @clusters = keys (%clusters_hash);
            # If there are clusters in the
            my %temp_hash = ();
            foreach my $cluster (@clusters) {
                if(defined $last_resort_clusters_hash{$cluster}) {
                    $temp_hash{$cluster} = $last_resort_clusters_hash{$cluster};
                }
            }
            %last_resort_clusters_hash = %temp_hash;
            my @last_resort_clusters = ();
            # Build the last_resort_clusters based on the home many times each has been used.
            my $times_used = 0;
            foreach my $cluster (keys %last_resort_clusters_hash) {
                if ($last_resort_clusters_hash{$cluster} == $times_used) {
                    # If this is used the same number of times then add it to the list
                    push(@last_resort_clusters,$cluster);
                } elsif ($last_resort_clusters_hash{$cluster} > $times_used) {
                    # If this is used more number of times then make it the list and update
                    # the times used. This way we can have the last resorts be the highest used
                    # clusters.
                    @last_resort_clusters = ($cluster);
                    $times_used = $last_resort_clusters_hash{$cluster};
                } # Else we ignore since it is not the most used.
            }

            # make sure to always have the source cluster in the last resort list.
            push(@last_resort_clusters,$source_cluster);

            # if the last resort clusters is >= the available clusters that means that all clusters
            # are in the last resort so we can assume that we only want to have the source cluster
            # as the last resort.
            if(scalar (@last_resort_clusters) >= scalar (@clusters)) {
                @last_resort_clusters = ($source_cluster);
            }

            my @last_resort_vservers = ();
            # Build the last_resort_clusters based on the home many times each has been used.
            $times_used = 0;
            foreach my $vserver (keys %last_resort_vservers_hash) {
                if ($last_resort_vservers_hash{$vserver} == $times_used) {
                    # If this is used the same number of times then add it to the list
                    push(@last_resort_vservers,$vserver);
                } elsif ($last_resort_vservers_hash{$vserver} > $times_used) {
                    # If this is used more number of times then make it the list and update
                    # the times used. This way we can have the last resorts be the highest used
                    # vservers.
                    @last_resort_vservers = ($vserver);
                    $times_used = $last_resort_vservers_hash{$vserver};
                } # Else we ignore since it is not the most used.
            }

            my @last_resort_aggregates = ();
            # Build the last_resort_clusters based on the home many times each has been used.
            $times_used = 0;
            foreach my $aggregate (keys %last_resort_aggregates_hash) {
                if ($last_resort_aggregates_hash{$aggregate} == $times_used) {
                    # If this is used the same number of times then add it to the list
                    push(@last_resort_aggregates,$aggregate);
                } elsif ($last_resort_aggregates_hash{$aggregate} > $times_used) {
                    # If this is used more number of times then make it the list and update
                    # the times used. This way we can have the last resorts be the highest used
                    # aggregates.
                    @last_resort_aggregates = ($aggregate);
                    $times_used = $last_resort_aggregates_hash{$aggregate};
                } # Else we ignore since it is not the most used.
            }

            # get a volume based on the criteria from the previous volumes.
            # If last_resort and unuseable have the same vservers unuseable takes precendence.
            my ($volume) = $self->get_available_snapmirror_volume('available_volumes'  => $available_volumes,
                                                                  'last_resort_clusters' => \@last_resort_clusters,
                                                                  'last_resort_vservers' => \@last_resort_vservers,
                                                                  'last_resort_aggregates' => \@last_resort_aggregates,
                                                                  'unuseable_vservers'   => [$source_vserver],
                                                                  'unuseable_aggregates' => [$source_aggregate],);
            # push the volume into our list of volumes for the chain.
            push(@destination_volumes,$volume);

            # Now we need to update the criteria based on the volume found.
            $last_resort_vservers_hash{$volume->get_attribute('vserver')}++;
            $last_resort_clusters_hash{$volume->get_attribute('cluster_name')}++;
        }

        for (my $j = 0; $j < scalar (@config_type_array); $j++) {
            # The destination volume is the volume at the current index plus 1.
            my $destination_volume = $destination_volumes[$j];

            # Update some metadata on the volume.
            $destination_volume->set_metadata_attributes('sm_config'     => 'fanout',
                                                         'sm_config_layout' => $config_layout,
                                                         'fanout_index' => $j,
                                                         'sm_config_id'  => $$config_id,);

            # The SM Type is the config type arrays current index
            my $SM_TYPE = $config_type_array[$j];
            $self->generate_snapmirror_resource('source_volume'                  => $source_volume,
                                                'destination_volume'             => $destination_volume,
                                                'sm_type'                        => $SM_TYPE,
                                                'vserver_sm_policy_hash'         => $vserver_sm_policy_hash,
                                                'vserver_snapshot_policy_hash'   => $vserver_snapshot_policy_hash,
                                                'cluster_job_schedule_cron_hash' => $cluster_job_schedule_cron_hash,,
                                                'xgen_metadata_args'             => {'sm_config'      => 'fanout',
                                                                                     'sm_config_layout' => $config_layout,
                                                                                     'fanout_index'  => $j+1,
                                                                                     'sm_config_id' => $$config_id});
            $$created_snapmirrors++;
        }
        # We finished the sm config so we need to update the config id
        $$config_id++;
    }
}

###############################################################################
# DESCRIPTION: Generates snapmirror resources in the configuration. This method
#              generates snapmirrors in the configuration based on the rules
#              for a cascade config.  Cascade in this method means
#              multiple snapmirrors in a change of volumes.
#              VOl1->Vol2->Vol3
#              It also prefers volumes on different clusters, and vservers
#              Creates a number of configs based on the
#              TOTAL_CASCADE_CONFIG_COUNT constant and the percentage of
#              configs with each type of layout is based on the
#              CASCADE_TYPE_PERCENT_LAYOUT constant.
# INPUT: snapmirror_count - number of snapmirrors left to be created
#        available_volumes - reference to an array of available volumes.
#        created_snapmirrors - a scalar reference to the number of already
#           created snapmirrors.
#        config_id - The next config id to use for snapmirror configs.
#        vserver_sm_policy_hash - a hashref to a mapping of snapmirror
#           policy name to vserver and cluster.
#        vserver_snapshot_policy_hash - a hashref to a mapping of snapshot
#           policy name to vserver and cluster.
#        cluster_job_schedule_cron_hash - a hashref to a mapping of schedules
#           to cluster.
# OUTPUT: NONE
# Exceptions: None thrown directly by this method.
###############################################################################
sub generate_cascade_snapmirror_configs {
    my ($self, %args) = @_;
    my $snapmirror_count = $args{'snapmirror_count'};
    my $available_volumes = $args{'available_volumes'};
    my $created_snapmirrors = $args{'created_snapmirrors'};
    my $config_id           = $args{'config_id'};
    # A mapping of vservers to snapmirror policies.
    my $vserver_sm_policy_hash = $args{'vserver_sm_policy_hash'};
    # A mapping of vservers to snapshot policies.
    my $vserver_snapshot_policy_hash = $args{'vserver_snapshot_policy_hash'};
    # A mapping of clusters to job schedules.
    my $cluster_job_schedule_cron_hash = $args{'cluster_job_schedule_cron_hash'};

    # This will get the layout in terms of numbers instead of percents based on the total object count passed in.
    my $layout = convert_percent_layout_to_count_layout('layout'        => $self->param('CASCADE_TYPE_PERCENT_LAYOUT'),
                                                        'total_objects' => $self->param('TOTAL_CASCADE_CONFIG_COUNT'));

    # Loop through TOTAL_CASCADE_CONFIG_COUNT and create those configs.
    for (my $i = 0; $i < $self->param('TOTAL_CASCADE_CONFIG_COUNT'); $i++) {
        my $config_layout = get_next_layout_element('iterator' => $i,     # This is not required but allows the get_next_layout_element to
                                                                          # rotate through the layout instead of exhausting the first, then
                                                                          # the second, etc.
                                                    'layout'   => $layout); # layout is the numbers layout we got earlier.

        # Config the config layout into an array of the types of snapmirrors in this cascade config.
        my @config_type_array = split(';',$config_layout);

        my @volumes = ();

        # For the first volume we set all the filters to nothing since the first
        # volume can go anywhere.
        my @unuseable_vservers = ();
        my %last_resort_vservers_hash = ();
        my %last_resort_clusters_hash = ();
        # We need to have snapmirrors+1 volumes so we use <=
        for (my $j = 0; $j <= scalar (@config_type_array); $j++) {
            my @last_resort_clusters = ();
            # Build the last_resort_clusters based on the home many times each has been used.
            my $times_used = 0;
            foreach my $cluster (keys %last_resort_clusters_hash) {
                if ($last_resort_clusters_hash{$cluster} == $times_used) {
                    # If this is used the same number of times then add it to the list
                    push(@last_resort_clusters,$cluster);
                } elsif ($last_resort_clusters_hash{$cluster} > $times_used) {
                    # If this is used more number of times then make it the list and update
                    # the times used. This way we can have the last resorts be the highest used
                    # clusters.
                    @last_resort_clusters = ($cluster);
                    $times_used = $last_resort_clusters_hash{$cluster};
                } # Else we ignore since it is not the most used.
            }

            my @last_resort_vservers = ();
            # Build the last_resort_clusters based on the home many times each has been used.
            $times_used = 0;
            foreach my $vserver (keys %last_resort_vservers_hash) {
                if ($last_resort_vservers_hash{$vserver} == $times_used) {
                    # If this is used the same number of times then add it to the list
                    push(@last_resort_vservers,$vserver);
                } elsif ($last_resort_vservers_hash{$vserver} > $times_used) {
                    # If this is used more number of times then make it the list and update
                    # the times used. This way we can have the last resorts be the highest used
                    # vservers.
                    @last_resort_vservers = ($vserver);
                    $times_used = $last_resort_vservers_hash{$vserver};
                } # Else we ignore since it is not the most used.
            }

            # get a volume based on the criteria from the previous volumes.
            # If last_resort and unuseable have the same vservers unuseable takes precendence.
            my ($volume) = $self->get_available_snapmirror_volume('available_volumes'  => $available_volumes,
                                                                  'last_resort_clusters' => \@last_resort_clusters,
                                                                  'last_resort_vservers' => \@last_resort_vservers,
                                                                  'unuseable_vservers'   => \@unuseable_vservers);
            # push the volume into our list of volumes for the chain.
            push(@volumes,$volume);

            # Now we need to update the criteria based on the volume found.
            # Can not use the same vserver for both source and destination of a single SM.
            @unuseable_vservers = ($volume->get_attribute('vserver'));
            $last_resort_vservers_hash{$volume->get_attribute('vserver')}++;
            $last_resort_clusters_hash{$volume->get_attribute('cluster_name')}++;

            # Update some metadata on the volume.
            $volume->set_metadata_attributes('sm_config'     => 'cascade',
                                             'sm_config_layout' => $config_layout,
                                             'cascade_index' => $j,
                                             'sm_config_id'  => $$config_id,)
        }

        # Now loop through and create all the Snapmirrors in the chain.
        for (my $j = 0; $j < scalar (@config_type_array); $j++) {
            # The source volume is the volume at the current index.
            my $source_volume      = $volumes[$j];
            # The destination volume is the volume at the current index plus 1.
            my $destination_volume = $volumes[$j+1];

            # The SM Type is the config type arrays current index
            my $SM_TYPE = $config_type_array[$j];
            $self->generate_snapmirror_resource('source_volume'                  => $source_volume,
                                                'destination_volume'             => $destination_volume,
                                                'sm_type'                        => $SM_TYPE,
                                                'vserver_sm_policy_hash'         => $vserver_sm_policy_hash,
                                                'vserver_snapshot_policy_hash'   => $vserver_snapshot_policy_hash,
                                                'cluster_job_schedule_cron_hash' => $cluster_job_schedule_cron_hash,,
                                                'xgen_metadata_args'             => {'sm_config'      => 'cascade',
                                                                                     'sm_config_layout' => $config_layout,
                                                                                     'cascade_index'  => $j,
                                                                                     'snapmirror_initialize_order' => $j,
                                                                                     'sm_config_id' => $$config_id,});
            $$created_snapmirrors++;
        }
        # We finished the sm config so we need to update the config id
        $$config_id++;
    }
}

###############################################################################
# DESCRIPTION: Generates snapmirror resources in the configuration. This method
#              generates snapmirrors in the configuration based on the rules
#              for a unidirectional config.  Unidirectional in this method means
#              each snapmirror creates is an atomic snapmirror with no
#              deliberate relationship to other snapmirrors.
#              It also prefers volumes on different clusters, and vservers.
#              Creates a number of snapmirrors based on the
#              snapmirror_count argument and the percentage of
#              snapmirrors with each type of layout is based on the
#              UNIDIRECTIONAL_TYPE_PERCENT_LAYOUT constant. This method
#              fills in snapmirrors up to the TOTAL_SNAPMIRROR_COUNT.
# INPUT: snapmirror_count - number of snapmirrors left to be created
#        available_volumes - reference to an array of available volumes.
#        created_snapmirrors - a scalar reference to the number of already
#           created snapmirrors.
#        config_id - The next config id to use for snapmirror configs.
#        vserver_sm_policy_hash - a hashref to a mapping of snapmirror
#           policy name to vserver and cluster.
#        vserver_snapshot_policy_hash - a hashref to a mapping of snapshot
#           policy name to vserver and cluster.
#        cluster_job_schedule_cron_hash - a hashref to a mapping of schedules
#           to cluster.
# OUTPUT: NONE
# Exceptions: None thrown directly by this method.
###############################################################################
sub generate_unidirectional_snapmirror_config {
    my ($self, %args) = @_;
    my $snapmirror_count    = $args{'snapmirror_count'};
    my $available_volumes   = $args{'available_volumes'};
    my $created_snapmirrors = $args{'created_snapmirrors'};
    my $config_id           = $args{'config_id'};
    # A mapping of vservers to snapmirror policies.
    my $vserver_sm_policy_hash = $args{'vserver_sm_policy_hash'};
    # A mapping of vservers to snapshot policies.
    my $vserver_snapshot_policy_hash = $args{'vserver_snapshot_policy_hash'};
    # A mapping of clusters to job schedules.
    my $cluster_job_schedule_cron_hash = $args{'cluster_job_schedule_cron_hash'};
    # Create a SMGR layout based on the number of SMGR snapmirror required.
    my $snapmirror_type_layout = convert_percent_layout_to_count_layout('layout'        => $self->param('UNIDIRECTIONAL_SNAPMIRROR_TYPE_PERCENT_LAYOUT'),
                                                                        'total_objects' => $snapmirror_count);

    for (my $i = 0; $i < $snapmirror_count; $i++) {
        # The source volume can be anywhere.
        my ($source_volume)      = $self->get_available_snapmirror_volume('available_volumes' => $available_volumes);
        # The destination volume cannot be on the same vserver and should try to be on a different cluster.
        my ($destination_volume) = $self->get_available_snapmirror_volume('available_volumes' => $available_volumes,
                                                                          'unuseable_vservers'   => [$source_volume->get_attribute('vserver')],
                                                                          'last_resort_clusters' => [$source_volume->get_attribute('cluster_name')]);
        $source_volume->set_metadata_attributes('sm_config'     => 'unidirectional',
                                                'sm_config_id'  => $$config_id,);
        $destination_volume->set_metadata_attributes('sm_config'     => 'unidirectional',
                                                     'sm_config_id'  => $$config_id,);
        # Get the next cg item type element to use from the smgr layout
        my $SM_TYPE = get_next_layout_element('iterator' => $i,
                                              'layout'   => $snapmirror_type_layout);
        $self->generate_snapmirror_resource('source_volume'                  => $source_volume,
                                            'destination_volume'             => $destination_volume,
                                            'sm_type'                        => $SM_TYPE,
                                            'vserver_sm_policy_hash'         => $vserver_sm_policy_hash,
                                            'vserver_snapshot_policy_hash'   => $vserver_snapshot_policy_hash,
                                            'cluster_job_schedule_cron_hash' => $cluster_job_schedule_cron_hash,
                                            'xgen_metadata_args'             => {'sm_config'     => 'unidirectional',
                                                                                 'sm_config_id' => $$config_id});
        $$created_snapmirrors++;
    }
    # Finished the unidirectional config so update the config_id.
    $$config_id++;
 }

###############################################################################
# DESCRIPTION: Generates a single snapmirror resource in the configuration.
#              This will generate a snapmirror resource using the volumes
#              passed in. The arguments for snapmirror are based on a
#              template corresponding to the snapmirror type.
#              snapmirror_policy argument.
#              Also calls method to generate snapmirror policies and
#              job schedule crons based on whether they are needed for the
#              snapmirror type.
#              Will Create Snapmirror Initialize Resources based on the
#              INITIALIZE_SNAPMIRRORS constant.
#              Will set metadata to turn off EnGenX setup and cleanup
#              for the snapmirror and snapmirror initialize based on the
#              SETUP_CLEANUP_SNAPMIRRORS constant.
# INPUT: source_volume - the source volume of the snapmirror
#        destination_volume - the destination volume of the snapmirror
#        vserver_sm_policy_hash - a hashref to a mapping of snapmirror
#           policy name to vserver and cluster.
#        vserver_snapshot_policy_hash - a hashref to a mapping of snapshot
#           policy name to vserver and cluster.
#        cluster_job_schedule_cron_hash - a hashref to a mapping of schedules
#           to cluster.
#        xgen_metadata_args - arguments to store on the snapmirror as metadata.
# OUTPUT: NONE
# Exceptions: None thrown directly by this method.
###############################################################################
sub generate_snapmirror_resource {
    my ($self, %args) = @_;

    my $source_volume = $args{'source_volume'};
    my $destination_volume = $args{'destination_volume'};
    my $SM_TYPE = $args{'sm_type'};

    my $source_volume_name = $source_volume->get_attribute('volume') . '_src_' . lc $SM_TYPE;
    my $dest_volume_name = $destination_volume->get_attribute('volume') . '_dst_' . lc $SM_TYPE;

    $source_volume->set_private_attributes(volume => $source_volume_name);
    $destination_volume->set_private_attributes(volume => $dest_volume_name);


    # A mapping of vservers to snapmirror policies.
    my $vserver_sm_policy_hash = $args{vserver_sm_policy_hash};
    # A mapping of vservers to snapshot policies.
    my $vserver_snapshot_policy_hash = $args{vserver_snapshot_policy_hash};
    # A mapping of clusters to job schedules.
    my $cluster_job_schedule_cron_hash = $args{cluster_job_schedule_cron_hash};
    my $xgen_metadata_args             = $args{xgen_metadata_args};
    my $method = $self->can($SM_TYPE);
    NATE::BaseException->throw("$SM_TYPE is not a supported Snapmirror Type in this configuration.\nPlease check your SM definition parameters.") unless(defined $method);
    my $snapmirror_args = $self->$method(
                   'source_volume' => $source_volume,
                   'destination_volume' => $destination_volume,
                   'vserver_snapshot_policy_hash' => $vserver_snapshot_policy_hash, # This is used for snapmirrors that require a snapshot policy.
                   );

    my ($source_aggr) = $source_volume->get_ancestors(type => 'NACL::AppBuilder::Xgen::Resource::Aggregate');
    my ($source_node) = $source_aggr->get_ancestors(type => 'NACL::AppBuilder::Xgen::Resource::Node');
    my ($destination_aggr) = $destination_volume->get_ancestors(type => 'NACL::AppBuilder::Xgen::Resource::Aggregate');
    my ($destination_node) = $destination_aggr->get_ancestors(type => 'NACL::AppBuilder::Xgen::Resource::Node');
    $snapmirror_args->{xgen_metadata}->{source_node}  = $source_node->get_attribute('node');
    $snapmirror_args->{xgen_metadata}->{destination_node}  = $destination_node->get_attribute('node');
    if(not $self->param('SETUP_CLEANUP_SNAPMIRRORS')) {
        $snapmirror_args->{xgen_metadata}->{setup}   = 0;
        $snapmirror_args->{xgen_metadata}->{cleanup} = 0;
    }

    if(defined $xgen_metadata_args) {
        foreach my $key (keys %{$xgen_metadata_args}) {
            $snapmirror_args->{xgen_metadata}->{$key} = $xgen_metadata_args->{$key};
        }
    }

    # If the volumes are on different clusters then we want to add the source-cluster as an argument.
    if($source_volume->get_attribute('cluster_name') ne $destination_volume->get_attribute('cluster_name')) {
        $snapmirror_args->{'source-cluster'} = $source_volume->get_attribute('cluster_name');
    }

    # Create the snapmirror resource.
    my $snapmirror = NACL::AppBuilder::Xgen::Resource::Snapmirror->new(%{$snapmirror_args});
    # The destination is required to be second since we need the destination cluster to be
    # the cluster used for these snapmirror commands.
    $snapmirror->add_parent_resources(resources => [$source_volume,$destination_volume]);
    $self->add_resources(resources => [$snapmirror]);

    my $snapmirror_policy = $snapmirror_args->{'policy'};
    # Add the snapmirror_policy to the destination vserver if it has not already been added.
    $self->generate_snapmirror_policy_resource('volume'                 => $destination_volume,
                                               'snapmirror_policy'      => $snapmirror_policy,
                                               'vserver_sm_policy_hash' => $vserver_sm_policy_hash);


    my $job_schedule_cron = $snapmirror_args->{'schedule'};
    # Add the job_schedule to the destination cluster if it has not already been added.
    $self->generate_job_schedule_cron_resource('volume'                         => $destination_volume,
                                               'job_schedule_cron'              => $job_schedule_cron,
                                               'cluster_job_schedule_cron_hash' => $cluster_job_schedule_cron_hash);


    # Add the job_schedule to the source cluster if it has not already been added.
    $self->generate_job_schedule_cron_resource('volume'                         => $source_volume,
                                               'job_schedule_cron'              => $job_schedule_cron,
                                               'cluster_job_schedule_cron_hash' => $cluster_job_schedule_cron_hash);

    if($self->param('INITIALIZE_SNAPMIRRORS')) {
        # Initialize does not have schedule so delete it.
        delete $snapmirror_args->{'schedule'};

        # Create the snapmirror initialize resource so the snapmirror gets initialized.
        my $snapmirror_initialize = NACL::AppBuilder::Xgen::Resource::SnapmirrorInitialize->new(%{$snapmirror_args});
        $snapmirror_initialize->add_parent_resources(resources => [$snapmirror]);
        $self->add_resources(resources => [$snapmirror_initialize]);
    }

}

# get and/or generate vol resource object
sub get_volume_resource_object {
    my $self = shift;
    my %args = validate_with(
        params => \@_,
        spec => {
                volume_name => { type => SCALAR, },
            },
    ); 

    my $vol = $args{volume_name};

    if (!$self->{volume_lookup}) {
        my $src_vol_resource_obj_array = $self->get_resources(
            type => 'NACL::AppBuilder::Xgen::Resource::Volume',
            #data => { volume => $source_volume_name },
        );
        my $source_volume_resource_obj ;
        foreach my $ele (@{ $src_vol_resource_obj_array}) {
            $Log->log('Got volume='.$ele->get_attribute('volume'));
            my $name = $ele->get_attribute('volume');
            #save it by name
            $self->{volume_lookup}->{$name} = $ele;
        }
    }

    return $self->{volume_lookup}->{$vol};

}


# generate 'ONE' snapmirror resource
sub generate_one_snapmirror_resource {
    my $self = shift;
    my %args = validate_with(
        params => \@_,
        spec => {
                sm_type => { type => SCALAR, },
                #type => { type => SCALAR, optional => 1, },
                schedule => { type => SCALAR, optional => 1 },
                policy => { type => SCALAR, optional => 1, },
                src_vol => { type => SCALAR, }, # will be the parents
                dst_vol => { type => SCALAR, }, # will be the parents
                cg_items => { type => ARRAYREF, optional => 1, },
                src_cg_name => { type => SCALAR, optional => 1, }, # if not specified this method will choose a name
                dst_cg_name => { type => SCALAR, optional => 1, }, # if not specified this method will choose a name
                create => { type => SCALAR, default => 1 },
                initialize => { type => SCALAR, default => 0 },
				altavault => { type => SCALAR, optional => 1, }, #ontap->altavault SM?
                altavault_whitelist => { type => SCALAR, default => 1 },
				btrfs => { type => SCALAR, optional => 1, }, #btrfs->ontap SM?
				solidfire => { type => SCALAR, optional => 1, }, #solidfire->ontap SM?
                vserver_sm_policy_hash => { type => HASHREF, default => {} },
                vserver_snapshot_policy_hash => { type => HASHREF, default => {} },
                cluster_job_schedule_cron_hash => { type => HASHREF, default => {} },
        },
        allow_extra => 1, # allow additional args that OnTap accepts and metadata_args
    );

    my $source_volume_name = $args{src_vol};
    my $destination_volume_name = $args{dst_vol};
    my $SM_TYPE = $args{sm_type};
    my $ava = $args{altavault};
    my $btrfs = $args{btrfs};
    my $solidfire = $args{solidfire};
	my $ontap_dest = 1;
	my $ontap_source = 1;
	if ($ava) {
		#is destination not ontap?
		$ontap_dest = 0;
	}
	if ($btrfs) {
		#is source not ontap?
		$ontap_source = 0;
	}
	if ($solidfire) {
		#is source not ontap?
		$ontap_source = 0;
	}




    $Log->log('Source volume name='.$source_volume_name.' Destination volume name='.$destination_volume_name);
    $Log->log("altavault:$ava");
    $Log->log("solidfire:$solidfire");
    $Log->log("btrfs:$btrfs");
    #my $src_vol_resource_obj_array = $self->get_resources(
    #type => 'NACL::AppBuilder::Xgen::Resource::Volume',
    #data => { volume => $source_volume_name },
    #);
    #my $source_volume_resource_obj ;
    #foreach my $ele (@{ $src_vol_resource_obj_array}) {
    #$Log->log('Got volume='.$ele->get_attribute('volume'));
    #if ($ele->get_attribute('volume') eq $source_volume_name) {
    #$source_volume_resource_obj = $ele;
    #}
    #}
    my $source_volume_resource_obj = $self->get_volume_resource_object(volume_name => $source_volume_name);
	if (!$source_volume_resource_obj && $btrfs) {
        $Log->log("$source_volume_name is a btrfs vol source");
        my $sma = dp::xplatform::lib::SMAgent->new(client=>$btrfs);
		my $ip = $sma->ip();
        $Log->log('My btrfs ip is '.$ip);
        $Log->log('My btrfs resource obj is '.Dumper($sma));
		#$Log->log('Creating btrfs volume for SMA: '.$source_volume_name);
        $Log->log('Get btrfs volume id for SMA: '.$source_volume_name);
		#my $btrfs_vol_id = $sma->vol_create(vol_name => $source_volume_name);
		my $btrfs_vol_id = $sma->get_vol_id(vol_name => $source_volume_name);
		$sma->create_lun(vol_name => $source_volume_name); #needed for initialize
        $source_volume_resource_obj = "$ip:/lun/$btrfs_vol_id";
	}

    $Log->debug('My src vol resource obj is '.Dumper($source_volume_resource_obj));

    my $dst_vol_resource_obj_array = $self->get_resources(
                                                      type => 'NACL::AppBuilder::Xgen::Resource::Volume',
                                                      data => { volume => $destination_volume_name },
                                                    );
    my $destination_volume_resource_obj ;
    foreach my $ele (@{ $dst_vol_resource_obj_array}) {
        $Log->log('Got volume='.$ele->get_attribute('volume'));
        if ($ele->get_attribute('volume') eq $destination_volume_name) {
            $destination_volume_resource_obj = $ele;
        }
    }

	#it's an ontap->ava snapmirror
	my $avanode;
	if (!$destination_volume_resource_obj && $ava) {
        $avanode = NACL::C::AVANode->new(name => $ava);
		my $ip = $avanode->hostrec->default_ip;
        $Log->log('My ava ip is '.$ip);
        $Log->log('My ava resource obj is '.Dumper($avanode));
        $destination_volume_resource_obj = "$ip:/share/$destination_volume_name";

		#we need to whitelist the snapmirror nodes
		if ($args{altavault_whitelist}) {
            $Log->log('Automatically adding whitelist IPs on AVA for ontap source nodes');

            my ($cluster_obj) = $source_volume_resource_obj->get_ancestors(type => "NACL::AppBuilder::Xgen::Resource::Cluster");
            my $cluster  = $cluster_obj->get_attribute('cluster_name');
            $Log->comment("source volume cluster: $cluster");
            $Log->trace("source volume cluster: $cluster" . Dumper $cluster_obj);
            #my $nacl_node_objs = $self->{nacl_node_objs};
			#this won't work without cluster natehost files
            #my $ci = NACL::C::Cluster->new (
            #name => $cluster,);
			my $ci = $self->get_cluster_hostrec(cluster => $cluster);
            $Log->trace("source volume cluster obj: $cluster" . Dumper $ci);
            my @interfaces = NACL::C::NetworkInterface->find(
                command_interface => $ci,
                filter            => {
                    'role'           => 'intercluster',
                    'address-family' => 'ipv4',
                },
            );

            $Log->trace('lifs:' . join ',',@interfaces);

            my $whitelist_result = $avanode->show_snapmirror_whitelist();

            my $whitelist = join '', $whitelist_result->get_raw_output();
            my @address;
            foreach my $iface (@interfaces) {
                push @address, $iface->get_one_state_attribute('address');
            }


            foreach my $addr (@address) {
                if ( $whitelist =~ qr/$addr/ ) {
                    $Log->comment("AVA already accepts $addr");
                } else {
                    $Log->comment("AVA not yet accepting $addr, add:");
                    $avanode->snapmirror_whitelist_add( ip => $addr );
                }
            }

		} else {
            $Log->log('Not adding whitelist IPs on AVA, disabled');
		}

	}
    $Log->debug('My dst vol resource obj is '.Dumper($destination_volume_resource_obj));
    #don't dump parent/children
    #$Log->log('My dst vol resource obj is '.Dumper($destination_volume_resource_obj->{shared}));
    #$Log->log(Dumper($destination_volume_resource_obj->{private}));
    #$Log->log(Dumper($destination_volume_resource_obj->{metadata}));

	if ($ontap_source) {
        $source_volume_resource_obj->set_private_attributes(volume => $source_volume_name);
	}
	if ($ontap_dest) {
	    $destination_volume_resource_obj->set_private_attributes(volume => $destination_volume_name);
    }

    # A mapping of vservers to snapmirror policies.
    my $vserver_sm_policy_hash = $args{vserver_sm_policy_hash};
    # A mapping of vservers to snapshot policies.
    my $vserver_snapshot_policy_hash = $args{vserver_snapshot_policy_hash};
    # A mapping of clusters to job schedules.
    my $cluster_job_schedule_cron_hash = $args{cluster_job_schedule_cron_hash};
    my $xgen_metadata_args             = $args{xgen_metadata_args};
    my $method = $self->can($SM_TYPE);
    NATE::BaseException->throw("$SM_TYPE is not a supported Snapmirror Type in this configuration.\nPlease check your SM definition parameters.") unless(defined $method);
    my $snapmirror_args = $self->$method(
                   'source_volume' => $source_volume_resource_obj,
				   'destination_volume' => $destination_volume_resource_obj,
				   #'destination_path' => $destination_volume_resource_obj,
                   'vserver_snapshot_policy_hash' => $vserver_snapshot_policy_hash, # This is used for snapmirrors that require a snapshot policy.
                   );

    my $source_aggr;
    my $source_node;
	if ($ontap_source) {
        ($source_aggr) = $source_volume_resource_obj->get_ancestors(type => 'NACL::AppBuilder::Xgen::Resource::Aggregate');
        ($source_node) = $source_aggr->get_ancestors(type => 'NACL::AppBuilder::Xgen::Resource::Node');
	}

	my $destination_aggr;
	my $destination_node;

	if ($ontap_dest) {
	    ($destination_aggr) = $destination_volume_resource_obj->get_ancestors(type => 'NACL::AppBuilder::Xgen::Resource::Aggregate');
	    ($destination_node) = $destination_aggr->get_ancestors(type => 'NACL::AppBuilder::Xgen::Resource::Node');
    }
	if ($ontap_source) {
        $snapmirror_args->{xgen_metadata}->{source_node}  = $source_node->get_attribute('node');
	} else {
	    $snapmirror_args->{xgen_metadata}->{source_volume}  = $source_volume_name;
	    $snapmirror_args->{xgen_metadata}->{source_node}  = $solidfire;
		if ($solidfire) {
	        $snapmirror_args->{xgen_metadata}->{solidfire}  = $solidfire;
		}
	}
	if ($ontap_dest) {
	    $snapmirror_args->{xgen_metadata}->{destination_node}  = $destination_node->get_attribute('node');
	} else {
	    $snapmirror_args->{xgen_metadata}->{destination_node}  = $ava;
		if ($ava) {
	        $snapmirror_args->{xgen_metadata}->{altavault}  = $ava;
		}
    }
    if(not $self->param('SETUP_CLEANUP_SNAPMIRRORS')) {
        $snapmirror_args->{xgen_metadata}->{setup}   = 0;
        $snapmirror_args->{xgen_metadata}->{cleanup} = 0;
    }

    if(defined $xgen_metadata_args) {
        foreach my $key (keys %{$xgen_metadata_args}) {
            $snapmirror_args->{xgen_metadata}->{$key} = $xgen_metadata_args->{$key};
        }
    }

    # If the volumes are on different clusters then we want to add the source-cluster as an argument.
    if ($ontap_dest && $ontap_source &&
        ($source_volume_resource_obj->get_attribute('cluster_name') ne
         $destination_volume_resource_obj->get_attribute('cluster_name')) ) {
        $snapmirror_args->{'source-cluster'} = $source_volume_resource_obj->get_attribute('cluster_name');
    }

    # use the job schedule and snapmirror policy that user specified - let it over ride the defaults
    my $job_schedule_cron = $args{schedule};
    if (!$job_schedule_cron) {
        $job_schedule_cron = $snapmirror_args->{'schedule'};
    } else {
        $snapmirror_args->{'schedule'} = $job_schedule_cron;
    }
    my $snapmirror_policy = $args{policy};
    if (!$snapmirror_policy) {
        $snapmirror_policy = $snapmirror_args->{'policy'};
    } else {
        $snapmirror_args->{'policy'} = $snapmirror_policy;
    }

    # Create the snapmirror resource.
    if($args{create}) {
    $Log->log('Snapmirror resource will be created with snapmirror args '.Dumper($snapmirror_args));
    my $snapmirror = NACL::AppBuilder::Xgen::Resource::Snapmirror->new(%{$snapmirror_args});
    $Log->debug('Snapmirror xgen  resource '.Dumper($snapmirror));
    # The destination is required to be second since we need the destination cluster to be
    # the cluster used for these snapmirror commands.
    $snapmirror->add_parent_resources(resources => [
			($ontap_source ? ($source_volume_resource_obj) : ()),
			($ontap_dest   ? ($destination_volume_resource_obj) : ()),
		]);
    $self->add_resources(resources => [$snapmirror]);
    $Log->log('AFTER Snapmirror add  resources will be created with snapmirror args '.Dumper($snapmirror_args));

    # Add the snapmirror_policy to the destination vserver if it has not already been added.
    $Log->log('Generate snapmirror policy resource volume='.$destination_volume_name.' snapmirror_policy='.$snapmirror_policy);
	if ($ontap_dest) {
        $self->generate_snapmirror_policy_resource('volume'                 => $destination_volume_resource_obj,
                                               'snapmirror_policy'      => $snapmirror_policy,
                                               'vserver_sm_policy_hash' => $vserver_sm_policy_hash);
    } elsif ($ontap_source) {
		#create on source not dest if no ontap dest
        $self->generate_snapmirror_policy_resource('volume'                 => $source_volume_resource_obj,
                                               'snapmirror_policy'      => $snapmirror_policy,
                                               'vserver_sm_policy_hash' => $vserver_sm_policy_hash);
    }


    if ($job_schedule_cron) {
        # Add the job_schedule to the destination cluster if it has not already been added.
        $Log->log('Generate job schedule cron resource to destination cluster volume='.$destination_volume_name.' job_schedule_cron='.$job_schedule_cron);
	    if ($ontap_dest) {
		    $self->generate_job_schedule_cron_resource('volume'                     => $destination_volume_resource_obj,
                                               'job_schedule_cron'              => $job_schedule_cron,
                                               'cluster_job_schedule_cron_hash' => $cluster_job_schedule_cron_hash);
        } elsif ($ontap_source) {
		    #create on source not dest if no ontap dest
            $self->generate_job_schedule_cron_resource('volume'                     => $source_volume_resource_obj,
                                               'job_schedule_cron'              => $job_schedule_cron,
                                               'cluster_job_schedule_cron_hash' => $cluster_job_schedule_cron_hash);
        }
    }

    # TBD do we want the INITIALIZE_SNAPMIRRORS to override the initialize
    if($args{initialize}) {
    #if($self->param('INITIALIZE_SNAPMIRRORS')) {
        # Initialize does not have schedule so delete it.
        delete $snapmirror_args->{'schedule'};

        # Create the snapmirror initialize resource so the snapmirror gets initialized.
        my $snapmirror_initialize = NACL::AppBuilder::Xgen::Resource::SnapmirrorInitialize->new(%{$snapmirror_args});
        $snapmirror_initialize->add_parent_resources(resources => [$snapmirror]);
        $self->add_resources(resources => [$snapmirror_initialize]);
    }

    # this should go after initialize since we will update the initialize resource object with more info in add_cg_items_to_one_SMGR_snapmirror_resouce
    # if its SMGR then add the cg-item-mappings
    if ($SM_TYPE eq 'SMGR') {
        $Log->log('Add cg items to the SMGR snapmirror resource');
        my %smgr_args;
        $smgr_args{src_cg_name} = $args{src_cg_name} if ($args{src_cg_name});
        $smgr_args{dst_cg_name} = $args{dst_cg_name} if ($args{dst_cg_name});
        $self->add_cg_items_to_one_SMGR_snapmirror_resources(snapmirror => $snapmirror,
                                                            cg_items => $args{cg_items},
                                                            %smgr_args
                                                        );
    }
} else {
    $Log->log('DO NOT create Snapmirror resource with snapmirror args '.Dumper($snapmirror_args));
}

}

###############################################################################
# VOLUME EFFICIENCY POLICY TEMPLATES BELOW
# Create new template by making the name of the method
# the policy name uppercased and suffixed with _VOLUME_EFFICIENCY_POLICY
# The method should return the arguments wanted for policy create.
###############################################################################

###############################################################################
# se_hourly policy that is just a default policy.
###############################################################################
sub SE_HOURLY_VOLUME_EFFICIENCY_POLICY {
    return {'policy' => 'se_hourly'};
}
###############################################################################
# VOLUME EFFICIENCY OPTION TEMPLATES BELOW
# Create new template by making the name of the method
# the template name uppercased and suffixed with _VOLUME_EFFICIENCY_OPTIONS
# The method should return the arguments wanted for volume efficiency modify.
###############################################################################

###############################################################################
# Default Storage Efficiency Turned on using SE_HOURLY policy and started
###############################################################################
sub STORAGE_EFFICIENCY_ON_VOLUME_EFFICIENCY_OPTIONS {
    return {'policy' => 'se_hourly'};
}

###############################################################################
# Storage efficiency on with compression
# Does not use a volume efficiency policy.
###############################################################################
sub COMPRESSION_ON_VOLUME_EFFICIENCY_OPTIONS {
    return {'compression'        => 'true',};
}

###############################################################################
# JOB SCHEDULE TEMPLATES BELOW
# Create new template by making the name of the method
# the Job Schedule name uppercased and suffixed with _JOB_SCHEDULE
# The method should return the arguments wanted for job schedule cron create.
###############################################################################

###############################################################################
# Job Schedule for 10 minutes after the hour every hour.
###############################################################################
sub HOURLY_10_JOB_SCHEDULE {
    return {'name'  =>'hourly_10',
            'minute'=>'10'};
}

###############################################################################
# Job Schedule for 15 minutes after the hour every hour.
###############################################################################
sub HOURLY_15_JOB_SCHEDULE {
    return {'name'  =>'hourly_15',
            'minute'=>'15'};
}

###############################################################################
# Job Schedule for 30 minutes after the hour every hour.
###############################################################################
sub SE_HOURLY_JOB_SCHEDULE {
    return {'name'  =>'se_hourly',
            'minute'=>'30'};
}

###############################################################################
# SNAPSHOT POLICY TEMPLATES BELOW
# Create new template by making the name of the method
# the Snapshot Policy name uppercased and suffixed with _SNAPSHOT_POLICY
# The method should return the arguments wanted for snapshot policy create.
###############################################################################

###############################################################################
# Snapshot policy to be used with some
###############################################################################
sub XDP_POLICY_SNAPSHOT_POLICY {
    return {'policy'=>'xdp_policy',
            'enabled'=>'true',
            'schedule1'=>'hourly', 'prefix1'=>'hourly', 'count1'=>'16', 'snapmirror-label1'=>'rpo_hourly',
            'schedule2'=>'daily',  'prefix2'=>'daily',  'count2'=>'2',  'snapmirror-label2'=>'rpo_daily',
            'schedule3'=>'weekly', 'prefix3'=>'weekly', 'count3'=>'2',  'snapmirror-label3'=>'rpo_weekly'};
}

###############################################################################
# SNAPMIRROR POLICY TEMPLATES BELOW
# Create new template by making the name of the method
# the Snapmirror Policy name uppercased and suffixed with _SNAPMIRROR_POLICY
# The method should return the arguments wanted for snapshot policy create.
# Also should return xgen_snapmirror_policy_rules as part of the return hash
# if snapmirror policy rules need to be created.
###############################################################################

###############################################################################
# Policy for SMGR snapmirrors.
###############################################################################
sub SMGR_SNAPMIRROR_POLICY {
    return {'tries'                   => 8,
            'policy'                  => 'Sync',
            'user-snapshot-timeout'   => 5,
            'type'                    => 'sync-mirror',
            'xgen_nacl_cs_validation' => 0};
}

###############################################################################
# Policy for ADR snapmirrors with net compression enabled.
###############################################################################
sub ADR_NETCOMP_SNAPMIRROR_POLICY {
    return {'policy'                          => 'ADR_netcomp',
            'type'                            => 'async-mirror',
            'is-network-compression-enabled'  => 'true'};
}

###############################################################################
# Policy for Vault Snapmirrors.
# Include snapmirror policy rules
###############################################################################
sub XDPRPO_SNAPMIRROR_POLICY {
    return {'policy' => 'XDPRPO',
            'xgen_snapmirror_policy_rules' =>
                [{'snapmirror-label'=>'rpo_hourly', 'keep'=>'24'},
                 {'snapmirror-label'=>'rpo_daily',  'keep'=>'7'},
                 {'snapmirror-label'=>'rpo_weekly', 'keep'=>'4'},
                 {'snapmirror-label'=>'weekly',     'keep'=>'4'}]};
}

###############################################################################
# Policy for Vault Snapmirrors with net compression enabled.
# Include snapmirror policy rules
###############################################################################
sub XDPRPO_NETCOMP_SNAPMIRROR_POLICY {
    return {'policy' => 'XDPRPO_netcomp',
            'is-network-compression-enabled'  => 'true',
            'xgen_snapmirror_policy_rules' =>
                [{'snapmirror-label'=>'rpo_hourly', 'keep'=>'24'},
                 {'snapmirror-label'=>'rpo_daily',  'keep'=>'7'},
                 {'snapmirror-label'=>'rpo_weekly', 'keep'=>'4'}]};
}

###############################################################################
# Policy for Mirror Latest Snapmirrors.
# Include snapmirror policy rules
###############################################################################
sub MIRRORLATEST_NC_SNAPMIRROR_POLICY {
    my ($self, %args) = @_;
    return {'policy'                          => 'MirrorLatest_NC',
            'type'                            => 'async-mirror',
            'is-network-compression-enabled'  => 'true'
            };
}

###############################################################################
# Policy for Mirror All Snapshots Snapmirrors.
# Include snapmirror policy rules
###############################################################################
sub MIRRORALLSNAPSHOTS_NC_SNAPMIRROR_POLICY {
    my ($self, %args) = @_;
    return {'policy' => 'MirrorAllSnapshots_NC',
            'type'                            => 'async-mirror',
            'is-network-compression-enabled'  => 'true',
            'xgen_snapmirror_policy_rules' =>
                [{'snapmirror-label'=>'all_source_snapshots', 'keep'=>'1',},]};
}

###############################################################################
# Policy for Vault Snapmirrors with net compression enabled.
# Include snapmirror policy rules
###############################################################################
sub MIRRORANDVAULT_NC_SNAPMIRROR_POLICY {
    my ($self, %args) = @_;
    return {'policy' => 'MirrorAndVault_NC',
            'xgen_snapmirror_policy_rules' =>
                [{'snapmirror-label'=>'rpo_hourly', 'keep'=>'24',},
                 {'snapmirror-label'=>'rpo_daily',  'keep'=>'7',},
                 {'snapmirror-label'=>'rpo_weekly', 'keep'=>'4',},
                 {'snapmirror-label'=>'daily',      'keep'=>'7',},
                 {'snapmirror-label'=>'weekly',     'keep'=>'52',},]};
}

###############################################################################
# Policy for Vault Snapmirrors with net compression enabled.
# Include snapmirror policy rules
###############################################################################
sub MIRRORANDVAULT_SNAPMIRROR_POLICY {
    my ($self, %args) = @_;
    return {'policy' => 'MirrorAndVault',
            'vserver' => $args{'volume'}->get_attribute('cluster_name'),
            'xgen_metadata' => {'setup' => 0,
                                'cleanup' => 0,},
            'xgen_snapmirror_policy_rules' =>
                [{'snapmirror-label'=>'rpo_hourly', 'keep'=>'24', 'vserver' => $args{'volume'}->get_attribute('cluster_name')},
                 {'snapmirror-label'=>'rpo_daily',  'keep'=>'7', 'vserver' => $args{'volume'}->get_attribute('cluster_name')},
                 {'snapmirror-label'=>'rpo_weekly', 'keep'=>'4', 'vserver' => $args{'volume'}->get_attribute('cluster_name')}]};
}

###############################################################################
# SNAPMIRROR TEMPLATES BELOW
# Create new template by making the name of the method
# the snapmirror template name uppercased
# The method should return the arguments wanted for snapmirror create, and
# set volume attributes if needed for the snapmirror.
###############################################################################

###############################################################################
# Creates the basic arguments for snapmirrors between volumes.
# Inputs:   source_volume - the source volume for the relationship.
#           destination_volume - the destination volume for the relationship.
#           sm_type - the type of snapmirror being created.
#           sm_policy - the name of the snapmirror policy to use
# Outputs:  A hash containing the snapmirror arguments.
#           Also updates the source and destination volumes with some metadata.
###############################################################################
sub generate_basic_snapmirror_args {
    my ($self, %args) = @_;
    my $source_volume = $args{'source_volume'};
    my $destination_volume = $args{'destination_volume'};
    my $sm_type = $args{'sm_type'};
    my $type    = $args{'type'};
    my $sm_policy = $args{'sm_policy'};

    my %optional_args = ();
    $optional_args{'policy'} = $sm_policy if (defined $sm_policy);
    $optional_args{'schedule'} = $args{'schedule'} if (defined $args{'schedule'});
    # Set metadata that can be queried for later.  When this gets to engenx all the fields will
    # be prepended with underscore.
    $source_volume->set_metadata_attributes('source_for_sm_type' => $sm_type);
    $destination_volume->set_metadata_attributes('destination_for_sm_type' => $sm_type);
    # Setup the source, destination, and type arguments.
    return {'source-vserver'    => $source_volume->get_attribute('vserver'),
            'source-volume'         => $source_volume->get_attribute('volume'),
            'destination-vserver'   => $destination_volume->get_attribute('vserver'),
            'destination-volume'    => $destination_volume->get_attribute('volume'),
            'type'                  => $type,
            %optional_args,
            xgen_metadata           => {'sm_type' => $sm_type} # set some metadata to be used later for filtering.
            };
}

###############################################################################
# Creates an smgr type Snapmirror Resource.
# The cg-item-mappings will be added later by the method
#     add_cg_items_to_SMGR_snapmirror_resources
# Defines each SMGR as a type file|lun
# Creates source_path/destination_path based on type.
# Uses policy smgr
###############################################################################
sub SMGR {
    my ($self, %args) = @_;
    my $source_volume = $args{'source_volume'};
    my $destination_volume = $args{'destination_volume'};

    # Set metadata that can be queried for later.  When this gets to engenx all the fields will
    # be prepended with underscore.
    $source_volume->set_metadata_attributes('source_for_sm_type' => 'smgr');
    $destination_volume->set_metadata_attributes('destination_for_sm_type' => 'smgr');
    # Setup the source, destination, and type arguments.  Many of the other required arguments will be added
    # by another method. This method just creates the shell that points to the correct volumes,type, and policy.
    return {'type'                => 'XDP',
            'policy'              => 'Sync',
            xgen_metadata         => {'sm_type' => 'smgr',
                                      'source-vserver'      => $source_volume->get_attribute('vserver'),
                                      'destination-vserver' => $destination_volume->get_attribute('vserver'),
                                      'source-volume'       => $source_volume->get_attribute('volume'),
                                      'destination-volume'  => $destination_volume->get_attribute('volume'),}
            };
}

sub AVA {
    my ($self, %args) = @_;
    my $source_volume = $args{'source_volume'};
    my $destination_volume = $args{'destination_volume'};

    # Set metadata that can be queried for later.  When this gets to engenx all the fields will
    # be prepended with underscore.
	#$source_volume->set_metadata_attributes('source_for_sm_type' => 'smgr');
    $source_volume->set_metadata_attributes('source_for_sm_type' => 'ava');
    # Setup the source, destination, and type arguments.  Many of the other required arguments will be added
    # by another method. This method just creates the shell that points to the correct volumes,type, and policy.
	#
        $Log->log('Generating snapshot policy resource for policy=xdp_policy');
        $self->generate_snapshot_policy_resource('volume'          => $args{source_volume},
                                                 'snapshot_policy' => 'xdp_policy',
                                                 %args,);
    return {'type'                => 'XDP',
		    'policy'              => 'XDPRPO',
                                      'source-path'       => $source_volume->get_attribute('vserver') . ':' .
									                         $source_volume->get_attribute('volume'),
                                      'destination-path'  => $destination_volume,
            xgen_metadata         => {'sm_type' => 'ava',
				                      #'source-vserver'      => $source_volume->get_attribute('vserver'),
									  #'source-volume'       => $source_volume->get_attribute('volume'),
                                      'source-path'       => $source_volume->get_attribute('vserver') . ':' .
									                         $source_volume->get_attribute('volume'),
                                      'destination-path'  => $destination_volume,}
            };
}

sub SOLIDFIRE {
    my ($self, %args) = @_;
    my $source_volume = $args{'source_volume'};
    my $destination_volume = $args{'destination_volume'};

    # Set metadata that can be queried for later.  When this gets to engenx all the fields will
    # be prepended with underscore.
    $destination_volume->set_metadata_attributes('source_for_sm_type' => 'solidfire');
    # Setup the source, destination, and type arguments.  Many of the other required arguments will be added
    # by another method. This method just creates the shell that points to the correct volumes,type, and policy.
	#
        $Log->log('Generating snapshot policy resource for policy=xdp_policy');
        $self->generate_snapshot_policy_resource('volume'          => $args{destination_volume},
                                                 'snapshot_policy' => 'xdp_policy',
                                                 %args,);
    return {'type'                => 'XDP',
		    'policy'              => 'MirrorLatest',
                                      'source-path'       => $source_volume,
                                      'destination-path'  => $destination_volume->get_attribute('vserver') . ':' .
									                         $destination_volume->get_attribute('volume'),
            xgen_metadata         => {'sm_type' => 'solidfire',
                                      'source-path'       => $source_volume,
                                      'destination-path'  => $destination_volume->get_attribute('vserver') . ':' .
									                         $destination_volume->get_attribute('volume'),
                                      }
            };
}

sub SMA_BTRFS {
    my ($self, %args) = @_;
    my $source_volume = $args{'source_volume'};
    my $destination_volume = $args{'destination_volume'};

    # Set metadata that can be queried for later.  When this gets to engenx all the fields will
    # be prepended with underscore.
    $destination_volume->set_metadata_attributes('source_for_sm_type' => 'sma_btrfs');
    # Setup the source, destination, and type arguments.  Many of the other required arguments will be added
    # by another method. This method just creates the shell that points to the correct volumes,type, and policy.
	#
        $Log->log('Generating snapshot policy resource for policy=xdp_policy');
        $self->generate_snapshot_policy_resource('volume'          => $args{destination_volume},
                                                 'snapshot_policy' => 'xdp_policy',
                                                 %args,);
    return {'type'                => 'XDP',
		    'policy'              => 'MirrorLatest',
                                      'source-path'       => $source_volume,
                                      'destination-path'  => $destination_volume->get_attribute('vserver') . ':' .
									                         $destination_volume->get_attribute('volume'),
            xgen_metadata         => {'sm_type' => 'solidfire',
                                      'source-path'       => $source_volume,
                                      'destination-path'  => $destination_volume->get_attribute('vserver') . ':' .
									                         $destination_volume->get_attribute('volume'),
                                      }
            };
}

###############################################################################
# Creates an ADR snapmirror
# type is DP
# Destination volume is type DP
# Uses the DPDefault policy
###############################################################################
sub ADR {
    my ($self, %args) = @_;
    $args{'destination_volume'}->set_private_attributes(type => 'DP',
                                                        'snapshot-policy' => 'none');
    my $snapmirror_args = $self->generate_basic_snapmirror_args('sm_type'   => 'adr',
                                                                'sm_policy' => 'DPDefault',
                                                                'type'      => 'DP',
                                                                'schedule'  => 'hourly_15',
                                                                %args);
    return $snapmirror_args;
}

###############################################################################
# Creates an ADR snapmirror with net compression enabled
# type is DP
# Destination volume is type DP
# Uses the ADR_netcomp policy
###############################################################################
sub ADR_NC {
    my ($self, %args) = @_;
    # Same as ADR but with a policy having net compression turned on
    my $snapmirror_args = $self->ADR('sm_type'   => 'adr_nc',
                                     'sm_policy' => 'ADR_netcomp',
                                     %args);
    return $snapmirror_args;
}


###############################################################################
# Creates an Vault snapmirror
# type is XDP
# Uses the XDPRPO policy
# Uses hourly_10 schedule
# Destination volume is type DP
# The source volume has a snapshot policy.
###############################################################################
sub VAULT {
    my ($self, %args) = @_;
    $args{'destination_volume'}->set_private_attributes(type => 'DP',
                                                        'snapshot-policy' => 'none');

    try {
        $args{'source_volume'}->get_attribute('snapshot-policy');
        $Log->log('Keeping snapshot policy set by user to '.$args{'source_volume'}->get_attribute('snapshot-policy'));
    } otherwise {
        $Log->log('Setting snapshot policy to xdp_policy');
        try {
            my $type = $args{source_volume}->get_attribute('type');
            # Only set the snapshot-policy if the type is not DP.
            if ($type ne 'DP') {
                $args{'source_volume'}->set_private_attributes('snapshot-policy' => 'xdp_policy');
            }
        } otherwise {
            # Only set the snapshot-policy if the type is not DP.
            $args{'source_volume'}->set_private_attributes('snapshot-policy' => 'xdp_policy');
        };
        $Log->log('Generating snapshot policy resource for policy=xdp_policy');
        $self->generate_snapshot_policy_resource('volume'          => $args{source_volume},
                                                 'snapshot_policy' => 'xdp_policy',
                                                 %args,);
    };
    my $snapmirror_args = $self->generate_basic_snapmirror_args('sm_type'   => 'vault',
                                                                'sm_policy' => 'XDPRPO',
                                                                'type'      => 'XDP',
                                                                'schedule'  => 'hourly_10',
                                                                %args);
    return $snapmirror_args;
}

###############################################################################
# Creates an Vault where the source volume has storage efficiency on.
###############################################################################
sub VAULT_SRCSE {
    my ($self, %args) = @_;
    $args{'source_volume'}->set_metadata_attributes('volume_efficiency_options' => 'storage_efficiency_on');
    my $snapmirror_args = $self->VAULT('sm_type'   => 'vault_srcse',
                                       %args);
    return $snapmirror_args;
}

###############################################################################
# Creates an Vault where the source volume has compression on.
###############################################################################
sub VAULT_SRCCOMPRESSION {
    my ($self, %args) = @_;
    $args{'source_volume'}->set_metadata_attributes('volume_efficiency_options' => 'compression_on');
    my $snapmirror_args = $self->VAULT('sm_type'   => 'vault_srccompression',
                                       %args);
    return $snapmirror_args;
}

###############################################################################
# Creates an Vault where the destination volume has storage efficiency on.
###############################################################################
sub VAULT_ADDDEDUP {
    my ($self, %args) = @_;
    $args{'destination_volume'}->set_metadata_attributes('volume_efficiency_options' => 'storage_efficiency_on');
    my $snapmirror_args = $self->VAULT('sm_type'   => 'vault_adddedup',
                                       %args);
    return $snapmirror_args;
}

###############################################################################
# Creates an Vault where the destination volume has compression on.
###############################################################################
sub VAULT_LRE {
    my ($self, %args) = @_;
    $args{'destination_volume'}->set_metadata_attributes('volume_efficiency_options' => 'compression_on');
    my $snapmirror_args = $self->VAULT('sm_type'   => 'vault_srccompression',
                                       %args);
    return $snapmirror_args;
}

###############################################################################
# Creates an Vault with network compression on.
# Uses the XDPRPO_netcomp policy
###############################################################################
sub VAULT_NC {
    my ($self, %args) = @_;
    # Same as VAULT but with a policy having net compression turned on
    my $snapmirror_args = $self->VAULT('sm_type'   => 'vault_nc',
                                       'sm_policy' => 'XDPRPO_netcomp',
                                       %args);
    return $snapmirror_args;
}

###############################################################################
# Creates an Vault NC where destination volume has compression on.
###############################################################################
sub VAULT_NC_LRE {
    my ($self, %args) = @_;
    $args{'destination_volume'}->set_metadata_attributes('volume_efficiency_options' => 'compression_on');
    my $snapmirror_args = $self->VAULT_NC('sm_type'   => 'vault_srccompression',
                                          %args);
    return $snapmirror_args;
}

###############################################################################
# Creates an Vault with using a different snapmirror policy
# Uses the MirrorLatest
###############################################################################
sub LRSE_DR_MIRROR_LATEST {
    my ($self, %args) = @_;
    # Same as VAULT but with a policy having MirrorLatest policy
    my $snapmirror_args = $self->VAULT('sm_type'   => 'lrse_dr_mirror_latest',
                                       'sm_policy' => 'MirrorLatest',
                                       %args);
    return $snapmirror_args;
}

###############################################################################
# Creates an LRSE-DR-MIRROR-LATEST where the destination volume has
#   compression on.
###############################################################################
sub LRSE_DR_MIRROR_LATEST_LRE {
    my ($self, %args) = @_;
    $args{'destination_volume'}->set_metadata_attributes('volume_efficiency_options' => 'compression_on');
    # Same as MirrorLatest
    my $snapmirror_args = $self->LRSE_DR_MIRROR_LATEST('sm_type'   => 'lrse_dr_mirror_latest_lre',
                                                       %args);
    return $snapmirror_args;
}

###############################################################################
# Creates an Vault with using a different snapmirror policy
# Uses the MirrorAllSnapshots
###############################################################################
sub LRSE_DR_MIRROR_ALL {
    my ($self, %args) = @_;
    # Same as VAULT but with a policy having MirrorAllSnapshots policy
    my $snapmirror_args = $self->VAULT('sm_type'   => 'lrse_dr_mirror_all',
                                       'sm_policy' => 'MirrorAllSnapshots',
                                       %args);
    return $snapmirror_args;
}

###############################################################################
# Creates an LRSE_DR_MIRROR_ALL where the destination volume has
#   compression on
###############################################################################
sub LRSE_DR_MIRROR_ALL_LRE {
    my ($self, %args) = @_;
    $args{'destination_volume'}->set_metadata_attributes('volume_efficiency_options' => 'compression_on');
    # Same as LRSE_DR_MIRROR_ALL
    my $snapmirror_args = $self->LRSE_DR_MIRROR_ALL('sm_type'   => 'lrse_dr_mirror_all_lre',
                                                    %args);
    return $snapmirror_args;
}
###############################################################################
# Creates an Vault with using a different snapmirror policy
# Uses the MirrorAndVault
###############################################################################
sub LRSE_DR_MIRROR_AND_VAULT {
    my ($self, %args) = @_;
    # Same as VAULT but with a policy having MirrorAllSnapshots policy
    my $snapmirror_args = $self->VAULT('sm_type'   => 'lrse_dr_mirror_and_vault',
                                       'sm_policy' => 'MirrorAndVault',
                                       %args);
    return $snapmirror_args;
}

###############################################################################
# Creates an LRSE_DR_MIRROR_AND_VAULT where the destination volume has
#   compression on
###############################################################################
sub LRSE_DR_MIRROR_AND_VAULT_LRE {
    my ($self, %args) = @_;
    $args{'destination_volume'}->set_metadata_attributes('volume_efficiency_options' => 'compression_on');
    # Same as LRSE_DR_MIRROR_AND_VAULT
    my $snapmirror_args = $self->LRSE_DR_MIRROR_AND_VAULT('sm_type'   => 'lrse_dr_mirror_and_vault_lre',
                                                          %args);
    return $snapmirror_args;
}

###############################################################################
# Creates an LRSE-DR-MIRROR-LATEST with network compression on
# Uses the MirrorLatest_NC policy
###############################################################################
sub LRSE_DR_MIRROR_LATEST_NC {
    my ($self, %args) = @_;
    # Same as VAULT but with a policy having MirrorLatest policy
    my $snapmirror_args = $self->VAULT('sm_type'   => 'lrse_dr_mirror_latest_nc',
                                       'sm_policy' => 'MirrorLatest_NC',
                                       %args);
    return $snapmirror_args;
}

###############################################################################
# Creates an LRSE_DR_MIRROR_LATEST_NC where the destination volume has
#   compression on
###############################################################################
sub LRSE_DR_MIRROR_LATEST_NC_LRE {
    my ($self, %args) = @_;
    $args{'destination_volume'}->set_metadata_attributes('volume_efficiency_options' => 'compression_on');
    # Same as MirrorLatest
    my $snapmirror_args = $self->LRSE_DR_MIRROR_LATEST_NC('sm_type'   => 'lrse_dr_mirror_latest_nc_lre',
                                                       %args);
    return $snapmirror_args;
}

###############################################################################
# Creates an LRSE-DR-MIRROR-ALL with network compression on
# Uses the MirrorAllSnapshots_NC policy
###############################################################################
sub LRSE_DR_MIRROR_ALL_NC {
    my ($self, %args) = @_;
    # Same as VAULT but with a policy having MirrorAllSnapshots policy
    my $snapmirror_args = $self->VAULT('sm_type'   => 'lrse_dr_mirror_all_nc',
                                       'sm_policy' => 'MirrorAllSnapshots_NC',
                                       %args);
    return $snapmirror_args;
}

###############################################################################
# Creates an LRSE_DR_MIRROR_ALL_NC where the destination volume has
#   compression on
###############################################################################
sub LRSE_DR_MIRROR_ALL_NC_LRE {
    my ($self, %args) = @_;
    $args{'destination_volume'}->set_metadata_attributes('volume_efficiency_options' => 'compression_on');
    # Same as LRSE_DR_MIRROR_ALL
    my $snapmirror_args = $self->LRSE_DR_MIRROR_ALL_NC('sm_type'   => 'lrse_dr_mirror_all_nc_lre',
                                                    %args);
    return $snapmirror_args;
}
###############################################################################
# Creates an LRSE-DR-MIRROR-ALL with network compression on
# Uses the MirrorAndVault_NC policy
###############################################################################
sub LRSE_DR_MIRROR_AND_VAULT_NC {
    my ($self, %args) = @_;
    # Same as VAULT but with a policy having MirrorAllSnapshots policy
    my $snapmirror_args = $self->VAULT('sm_type'   => 'lrse_dr_mirror_and_vault_nc',
                                       'sm_policy' => 'MirrorAndVault_NC',
                                       %args);
    return $snapmirror_args;
}

###############################################################################
# Creates an LRSE_DR_MIRROR_AND_VAULT_NC where the destination volume has
#   compression on
###############################################################################
sub LRSE_DR_MIRROR_AND_VAULT_NC_LRE {
    my ($self, %args) = @_;
    $args{'destination_volume'}->set_metadata_attributes('volume_efficiency_options' => 'compression_on');
    # Same as LRSE_DR_MIRROR_AND_VAULT
    my $snapmirror_args = $self->LRSE_DR_MIRROR_AND_VAULT_NC('sm_type'   => 'lrse_dr_mirror_and_vault_nc_lre',
                                                          %args);
    return $snapmirror_args;
}

###############################################################################
# HELPER METHODS
###############################################################################

###############################################################################
# Gets Available Volumes for Snapmirrors based on filters.
# Inputs:   volume_count - The number of volumes to return matching the filters
#           available_volumes - a mapping of available volumes
#                               to cluster and vserver.
#           all_from_same_vserver - whether all the volumes must come from the
#                                   same vserver
#           must_use_clusters - array of cluster names required for the volumes
#           last_resort_clusters - array of cluster names that should only
#                                  be used if there are no volumes in other
#                                  clusters that match the criteria.
#           unuseable_clusters - array of cluster names that can not be used
#                                for the volumes
#           must_use_vservers - array of vserver names required for the volumes
#           last_resort_vservers - array of vserver names that should only
#                                  be used if there are no volumes in other
#                                  vservers that match the criteria.
#           unuseable_vservers - array of vserver names that can not be used
#                                for the volumes
#           must_use_aggregates - array of aggregate names required for the volumes
#           last_resort_aggregates - array of aggregate names that should only
#                                  be used if there are no volumes in other
#                                  aggregates that match the criteria.
#           unuseable_aggregates - array of aggregate names that can not be used
#                                for the volumes
# Outputs:  A hash containing the snapmirror arguments.
#           Also updates the source and destination volumes with some metadata.
###############################################################################
sub get_available_snapmirror_volume {
    my ($self, %args) = @_;

    # The number of volumes to return match the filters.
    # Defaults to 1.
    my $volume_count           = $args{volume_count} // 1;
    # A reference to a hash of the volumes mapped to clusters and vserver.
    my $available_volumes      = $args{available_volumes};

    # These are the filters for the method.
    # CLUSTER FILTERS
    my $must_use_clusters      = $args{must_use_clusters};
    my $unuseable_clusters     = $args{unuseable_clusters};
    my $last_resort_clusters   = $args{last_resort_clusters};
    # VSERVER FILTERS
    my $must_use_vservers      = $args{must_use_vservers};
    my $unuseable_vservers     = $args{unuseable_vservers};
    my $last_resort_vservers   = $args{last_resort_vservers};
    # AGGREGATE FILTERS
    my $must_use_aggregates    = $args{must_use_aggregates};
    my $unuseable_aggregates   = $args{unuseable_aggregates};
    my $last_resort_aggregates = $args{last_resort_aggregates};


    my $all_from_same_vserver  = $args{all_from_same_vserver} // 0;

    my @return_volumes = ();

    my $filtered_hash = {};
    ###############################################################################
    # The following section of code will build the filtered_hash.
    # The filtered hash is built in the following format based on passed in filters.
    # {
    # useable->cluster_name->{useable->vserver_name->useable->[array of volumes],
    #                                                unuseable->[array of volumes],
    #                                                last_resort->[array of volumes],
    #                         unuseable->vserver_name->useable->[array of volumes],
    #                                                  unuseable->[array of volumes],
    #                                                  last_resort->[array of volumes],
    #                         last_resort->vserver_name->useable->[array of volumes],
    #                                                    unuseable->[array of volumes],
    #                                                    last_resort->[array of volumes],
    # last_resort->cluster_name->{useable->vserver_name->useable->[array of volumes],
    #                                                    unuseable->[array of volumes],
    #                                                    last_resort->[array of volumes],
    #                             unuseable->vserver_name->useable->[array of volumes],
    #                                                      unuseable->[array of volumes],
    #                                                      last_resort->[array of volumes],
    #                             last_resort->vserver_name->useable->[array of volumes],
    #                                                        unuseable->[array of volumes],
    #                                                        last_resort->[array of volumes],
    # unuseable->cluster_name->{useable->vserver_name->useable->[array of volumes],
    #                                                  unuseable->[array of volumes],
    #                                                  last_resort->[array of volumes],
    #                           unuseable->vserver_name->useable->[array of volumes],
    #                                                    unuseable->[array of volumes],
    #                                                    last_resort->[array of volumes],
    #                           last_resort->vserver_name->useable->[array of volumes],
    #                                                      unuseable->[array of volumes],
    #                                                      last_resort->[array of volumes],
    #}
    ###############################################################################
    while (scalar @{$available_volumes}) {
        my $current_volume = shift @{$available_volumes};
        my $current_cluster = $current_volume->get_attribute('cluster_name');
        my $current_vserver = $current_volume->get_attribute('vserver');
        my $current_aggregate = $current_volume->get_attribute('aggregate');
        my $cluster_useability;
        my $vserver_useability;
        my $volume_useability;
        if (defined $must_use_clusters ) {
            if(grep(/^$current_cluster$/,@{$must_use_clusters})) {
                $cluster_useability = 'useable';
            } else {
                $cluster_useability = 'unuseable';
            }
        } elsif (defined $unuseable_clusters and grep(/^$current_cluster$/,@{$unuseable_clusters})) {
            $cluster_useability = 'unuseable';
        } elsif (defined $last_resort_clusters and grep(/^$current_cluster$/,@{$last_resort_clusters})) {
            $cluster_useability = 'last_resort';
        } else {
            $cluster_useability = 'useable';
        }

        if (defined $must_use_vservers ) {
            if(grep(/^$current_vserver$/,@{$must_use_vservers})) {
                $vserver_useability = 'useable';
            } else {
                $vserver_useability = 'unuseable';
            }
        } elsif (defined $unuseable_vservers and grep(/^$current_vserver$/,@{$unuseable_vservers})) {
            $vserver_useability = 'unuseable';
        } elsif (defined $last_resort_vservers and grep(/^$current_vserver$/,@{$last_resort_vservers})) {
            $vserver_useability = 'last_resort';
        } else {
            $vserver_useability = 'useable';
        }

        if (defined $must_use_aggregates ) {
            if(grep(/^$current_volume$/,@{$must_use_aggregates})) {
                $volume_useability = 'useable';
            } else {
                $volume_useability = 'unuseable';
            }
        } elsif (defined $unuseable_aggregates and grep(/^$current_volume$/,@{$unuseable_aggregates})) {
            $volume_useability = 'unuseable';
        } elsif (defined $last_resort_aggregates and grep(/^$current_volume$/,@{$last_resort_aggregates})) {
            $volume_useability = 'last_resort';
        } else {
            $volume_useability = 'useable';
        }

        push(@{$filtered_hash->{$cluster_useability}->{$current_cluster}->{$vserver_useability}->{$current_vserver}->{$volume_useability}}, $current_volume);
    }

    ###############################################################################
    # The following code will search through the filter hash for the best volumes
    # to use.
    # CU = Cluster Useable
    # VSU = Vserver Useable
    # VU  = Volume Useable
    # CL = Cluster Last Resort
    # VSL = Vserver Last Resort
    # VL  = Volume Resort
    # The search happens in the following order.
    # CU VSU VU
    # CU VSU VL
    # CU VSL VU
    # CU VSL VL
    # CL VSU VU
    # CL VSU VL
    # CL VSL VU
    # CL VSL VL
    ###############################################################################

    # Look for cluster useable, vserver useable, volume useable.
    if (scalar (@return_volumes) < $volume_count and
        defined $filtered_hash->{useable}) {
        my @clusters = keys (%{$filtered_hash->{useable}});
        # Look for vserver useable, volume useable
        foreach my $cluster (@clusters) {
            if (defined $filtered_hash->{useable}->{$cluster}->{useable}) {
                my @vservers = keys (%{$filtered_hash->{useable}->{$cluster}->{useable}});
                # Look for volume useable.
                foreach my $vserver (@vservers) {
                    if($all_from_same_vserver) {
                        # If we want all from same vserver and there are not enough useable then we skip this vserver for now.
                        next if (scalar @{$filtered_hash->{useable}->{$cluster}->{useable}->{$vserver}->{useable}} < $volume_count);
                        # If there are enough volumes then add the useable volumes
                        while ((scalar @{$filtered_hash->{useable}->{$cluster}->{useable}->{$vserver}->{useable}}) and
                               (scalar (@return_volumes) < $volume_count)) {
                            push (@return_volumes, (shift @{$filtered_hash->{useable}->{$cluster}->{useable}->{$vserver}->{useable}}));
                        }
                    # Otherwise we will just start adding volumes from the useable space of this vserver.
                    } else {
                        while ((scalar @{$filtered_hash->{useable}->{$cluster}->{useable}->{$vserver}->{useable}}) and
                               (scalar (@return_volumes) < $volume_count)) {
                            push (@return_volumes, (shift @{$filtered_hash->{useable}->{$cluster}->{useable}->{$vserver}->{useable}}));
                        }
                    }
                }
            }
        }
    }

    # Look for cluster useable, vserver useable, volume last_resort.
    if (scalar (@return_volumes) < $volume_count and
        defined $filtered_hash->{useable}) {
        my @clusters = keys (%{$filtered_hash->{useable}});
        # Look for vserver useable, volume last_resort
        foreach my $cluster (@clusters) {
            if (defined $filtered_hash->{useable}->{$cluster}->{useable}) {
                my @vservers = keys (%{$filtered_hash->{useable}->{$cluster}->{useable}});
                # Look for volume last_resort.
                foreach my $vserver (@vservers) {
                    if($all_from_same_vserver) {
                        # If we want all from same vserver and there are not enough useable and last_resort then we skip this vserver
                        next if (scalar @{$filtered_hash->{useable}->{$cluster}->{useable}->{$vserver}->{useable}} +
                                 scalar @{$filtered_hash->{useable}->{$cluster}->{useable}->{$vserver}->{last_resort}}  < $volume_count);
                        # Add all the useable volumes first, then the last resorts.
                        while ((scalar @{$filtered_hash->{useable}->{$cluster}->{useable}->{$vserver}->{useable}}) and
                               (scalar (@return_volumes) < $volume_count)) {
                            push (@return_volumes, (shift @{$filtered_hash->{useable}->{$cluster}->{useable}->{$vserver}->{useable}}));
                        }
                        while ((scalar @{$filtered_hash->{useable}->{$cluster}->{useable}->{$vserver}->{last_resort}}) and
                               (scalar (@return_volumes) < $volume_count)) {
                            push (@return_volumes, (shift @{$filtered_hash->{useable}->{$cluster}->{useable}->{$vserver}->{last_resort}}));
                        }
                    # Otherwise we will just start adding volumes from the useable space of this vserver.
                    } else {
                        while ((scalar @{$filtered_hash->{useable}->{$cluster}->{useable}->{$vserver}->{last_resort}}) and
                               (scalar (@return_volumes) < $volume_count)) {
                            push (@return_volumes, (shift @{$filtered_hash->{useable}->{$cluster}->{useable}->{$vserver}->{last_resort}}));
                        }
                    }
                }
            }
        }
    }

    # Look for cluster useable, vserver last_resort, volume useable.
    if (scalar (@return_volumes) < $volume_count and
        defined $filtered_hash->{useable}) {
        my @clusters = keys (%{$filtered_hash->{useable}});
        # Look for vserver useable, volume useable
        foreach my $cluster (@clusters) {
            if (defined $filtered_hash->{useable}->{$cluster}->{last_resort}) {
                my @vservers = keys (%{$filtered_hash->{useable}->{$cluster}->{last_resort}});
                # Look for volume useable.
                foreach my $vserver (@vservers) {
                    if($all_from_same_vserver) {
                        # If we want all from same vserver and there are not enough useable then we skip this vserver for now.
                        next if (scalar @{$filtered_hash->{useable}->{$cluster}->{last_resort}->{$vserver}->{useable}} < $volume_count);
                        # If there are enough volumes then add the useable volumes
                        while ((scalar @{$filtered_hash->{useable}->{$cluster}->{last_resort}->{$vserver}->{useable}}) and
                               (scalar (@return_volumes) < $volume_count)) {
                            push (@return_volumes, (shift @{$filtered_hash->{useable}->{$cluster}->{last_resort}->{$vserver}->{useable}}));
                        }
                    # Otherwise we will just start adding volumes from the useable space of this vserver.
                    } else {
                        while ((scalar @{$filtered_hash->{useable}->{$cluster}->{last_resort}->{$vserver}->{useable}}) and
                               (scalar (@return_volumes) < $volume_count)) {
                            push (@return_volumes, (shift @{$filtered_hash->{useable}->{$cluster}->{last_resort}->{$vserver}->{useable}}));
                        }
                    }
                }
            }
        }
    }

    # Look for cluster useable, vserver last_resort, volume last_resort.
    if (scalar (@return_volumes) < $volume_count and
        defined $filtered_hash->{useable}) {
        my @clusters = keys (%{$filtered_hash->{useable}});
        # Look for vserver useable, volume last_resort
        foreach my $cluster (@clusters) {
            if (defined $filtered_hash->{useable}->{$cluster}->{last_resort}) {
                my @vservers = keys (%{$filtered_hash->{useable}->{$cluster}->{last_resort}});
                # Look for volume last_resort.
                foreach my $vserver (@vservers) {
                    if($all_from_same_vserver) {
                        # If we want all from same vserver and there are not enough useable and last_resort then we skip this vserver
                        next if (scalar @{$filtered_hash->{useable}->{$cluster}->{last_resort}->{$vserver}->{useable}} +
                                 scalar @{$filtered_hash->{useable}->{$cluster}->{last_resort}->{$vserver}->{last_resort}}  < $volume_count);
                        # Add all the useable volumes first, then the last resorts.
                        while ((scalar @{$filtered_hash->{useable}->{$cluster}->{last_resort}->{$vserver}->{useable}}) and
                               (scalar (@return_volumes) < $volume_count)) {
                            push (@return_volumes, (shift @{$filtered_hash->{useable}->{$cluster}->{last_resort}->{$vserver}->{useable}}));
                        }
                        while ((scalar @{$filtered_hash->{useable}->{$cluster}->{last_resort}->{$vserver}->{last_resort}}) and
                               (scalar (@return_volumes) < $volume_count)) {
                            push (@return_volumes, (shift @{$filtered_hash->{useable}->{$cluster}->{last_resort}->{$vserver}->{last_resort}}));
                        }
                    # Otherwise we will just start adding volumes from the useable space of this vserver.
                    } else {
                        while ((scalar @{$filtered_hash->{useable}->{$cluster}->{last_resort}->{$vserver}->{last_resort}}) and
                               (scalar (@return_volumes) < $volume_count)) {
                            push (@return_volumes, (shift @{$filtered_hash->{useable}->{$cluster}->{last_resort}->{$vserver}->{last_resort}}));
                        }
                    }
                }
            }
        }
    }

    # Look for cluster last_resort, vserver useable, volume useable.
    if (scalar (@return_volumes) < $volume_count and
        defined $filtered_hash->{last_resort}) {
        my @clusters = keys (%{$filtered_hash->{last_resort}});
        # Look for vserver useable, volume useable
        foreach my $cluster (@clusters) {
            if (defined $filtered_hash->{last_resort}->{$cluster}->{useable}) {
                my @vservers = keys (%{$filtered_hash->{last_resort}->{$cluster}->{useable}});
                # Look for volume useable.
                foreach my $vserver (@vservers) {
                    if($all_from_same_vserver) {
                        # If we want all from same vserver and there are not enough useable then we skip this vserver for now.
                        next if (scalar @{$filtered_hash->{last_resort}->{$cluster}->{useable}->{$vserver}->{useable}} < $volume_count);
                        # If there are enough volumes then add the useable volumes
                        while ((scalar @{$filtered_hash->{last_resort}->{$cluster}->{useable}->{$vserver}->{useable}}) and
                               (scalar (@return_volumes) < $volume_count)) {
                            push (@return_volumes, (shift @{$filtered_hash->{last_resort}->{$cluster}->{useable}->{$vserver}->{useable}}));
                        }
                    # Otherwise we will just start adding volumes from the useable space of this vserver.
                    } else {
                        while ((scalar @{$filtered_hash->{last_resort}->{$cluster}->{useable}->{$vserver}->{useable}}) and
                               (scalar (@return_volumes) < $volume_count)) {
                            push (@return_volumes, (shift @{$filtered_hash->{last_resort}->{$cluster}->{useable}->{$vserver}->{useable}}));
                        }
                    }
                }
            }
        }
    }

    # Look for cluster last_resort, vserver useable, volume last_resort.
    if (scalar (@return_volumes) < $volume_count and
        defined $filtered_hash->{last_resort}) {
        my @clusters = keys (%{$filtered_hash->{last_resort}});
        # Look for vserver useable, volume last_resort
        foreach my $cluster (@clusters) {
            if (defined $filtered_hash->{last_resort}->{$cluster}->{useable}) {
                my @vservers = keys (%{$filtered_hash->{last_resort}->{$cluster}->{useable}});
                # Look for volume last_resort.
                foreach my $vserver (@vservers) {
                    if($all_from_same_vserver) {
                        # If we want all from same vserver and there are not enough useable and last_resort then we skip this vserver
                        next if (scalar @{$filtered_hash->{last_resort}->{$cluster}->{useable}->{$vserver}->{useable}} +
                                 scalar @{$filtered_hash->{last_resort}->{$cluster}->{useable}->{$vserver}->{last_resort}}  < $volume_count);
                        # Add all the useable volumes first, then the last resorts.
                        while ((scalar @{$filtered_hash->{last_resort}->{$cluster}->{useable}->{$vserver}->{useable}}) and
                               (scalar (@return_volumes) < $volume_count)) {
                            push (@return_volumes, (shift @{$filtered_hash->{last_resort}->{$cluster}->{useable}->{$vserver}->{useable}}));
                        }
                        while ((scalar @{$filtered_hash->{last_resort}->{$cluster}->{useable}->{$vserver}->{last_resort}}) and
                               (scalar (@return_volumes) < $volume_count)) {
                            push (@return_volumes, (shift @{$filtered_hash->{last_resort}->{$cluster}->{useable}->{$vserver}->{last_resort}}));
                        }
                    # Otherwise we will just start adding volumes from the useable space of this vserver.
                    } else {
                        while ((scalar @{$filtered_hash->{last_resort}->{$cluster}->{useable}->{$vserver}->{last_resort}}) and
                               (scalar (@return_volumes) < $volume_count)) {
                            push (@return_volumes, (shift @{$filtered_hash->{last_resort}->{$cluster}->{useable}->{$vserver}->{last_resort}}));
                        }
                    }
                }
            }
        }
    }

    # Look for cluster last_resort, vserver last_resort, volume useable.
    if (scalar (@return_volumes) < $volume_count and
        defined $filtered_hash->{last_resort}) {
        my @clusters = keys (%{$filtered_hash->{last_resort}});
        # Look for vserver useable, volume useable
        foreach my $cluster (@clusters) {
            if (defined $filtered_hash->{last_resort}->{$cluster}->{last_resort}) {
                my @vservers = keys (%{$filtered_hash->{last_resort}->{$cluster}->{last_resort}});
                # Look for volume useable.
                foreach my $vserver (@vservers) {
                    if($all_from_same_vserver) {
                        # If we want all from same vserver and there are not enough useable then we skip this vserver for now.
                        next if (scalar @{$filtered_hash->{last_resort}->{$cluster}->{last_resort}->{$vserver}->{useable}} < $volume_count);
                        # If there are enough volumes then add the useable volumes
                        while ((scalar @{$filtered_hash->{last_resort}->{$cluster}->{last_resort}->{$vserver}->{useable}}) and
                               (scalar (@return_volumes) < $volume_count)) {
                            push (@return_volumes, (shift @{$filtered_hash->{last_resort}->{$cluster}->{last_resort}->{$vserver}->{useable}}));
                        }
                    # Otherwise we will just start adding volumes from the useable space of this vserver.
                    } else {
                        while ((scalar @{$filtered_hash->{last_resort}->{$cluster}->{last_resort}->{$vserver}->{useable}}) and
                               (scalar (@return_volumes) < $volume_count)) {
                            push (@return_volumes, (shift @{$filtered_hash->{last_resort}->{$cluster}->{last_resort}->{$vserver}->{useable}}));
                        }
                    }
                }
            }
        }
    }

    # Look for cluster last_resort, vserver last_resort, volume last_resort.
    if (scalar (@return_volumes) < $volume_count and
        defined $filtered_hash->{last_resort}) {
        my @clusters = keys (%{$filtered_hash->{last_resort}});
        # Look for vserver useable, volume last_resort
        foreach my $cluster (@clusters) {
            if (defined $filtered_hash->{last_resort}->{$cluster}->{last_resort}) {
                my @vservers = keys (%{$filtered_hash->{last_resort}->{$cluster}->{last_resort}});
                # Look for volume last_resort.
                foreach my $vserver (@vservers) {
                    if($all_from_same_vserver) {
                        # If we want all from same vserver and there are not enough useable and last_resort then we skip this vserver
                        next if (scalar @{$filtered_hash->{last_resort}->{$cluster}->{last_resort}->{$vserver}->{useable}} +
                                 scalar @{$filtered_hash->{last_resort}->{$cluster}->{last_resort}->{$vserver}->{last_resort}}  < $volume_count);
                        # Add all the useable volumes first, then the last resorts.
                        while ((scalar @{$filtered_hash->{last_resort}->{$cluster}->{last_resort}->{$vserver}->{useable}}) and
                               (scalar (@return_volumes) < $volume_count)) {
                            push (@return_volumes, (shift @{$filtered_hash->{last_resort}->{$cluster}->{last_resort}->{$vserver}->{useable}}));
                        }
                        while ((scalar @{$filtered_hash->{last_resort}->{$cluster}->{last_resort}->{$vserver}->{last_resort}}) and
                               (scalar (@return_volumes) < $volume_count)) {
                            push (@return_volumes, (shift @{$filtered_hash->{last_resort}->{$cluster}->{last_resort}->{$vserver}->{last_resort}}));
                        }
                    # Otherwise we will just start adding volumes from the useable space of this vserver.
                    } else {
                        while ((scalar @{$filtered_hash->{last_resort}->{$cluster}->{last_resort}->{$vserver}->{last_resort}}) and
                               (scalar (@return_volumes) < $volume_count)) {
                            push (@return_volumes, (shift @{$filtered_hash->{last_resort}->{$cluster}->{last_resort}->{$vserver}->{last_resort}}));
                        }
                    }
                }
            }
        }
    }

    if (scalar (@return_volumes) < $volume_count){
        NATE::BaseException->throw("Could not find an available volume for snapmirroring that fit the following criteria.\n" .
                                   "volume_count = " . $volume_count . "\n" .
                                   "all_from_same_vserver = " . $all_from_same_vserver . "\n" .
                                   "must_use_clusters = \n" . dumper($must_use_clusters) . "\n" .
                                   "must_use_vservers = \n" . dumper($must_use_vservers) . "\n" .
                                   "unuseable_clusters = \n" . dumper($unuseable_clusters) . "\n" .
                                   "unuseable_vservers = \n" . dumper($unuseable_vservers) . "\n" .
                                   "last_resort_clusters = \n" . dumper($last_resort_clusters) . "\n" .
                                   "last_resort_vservers = \n" . dumper($last_resort_vservers) . "\n");
    }

    ###############################################################################
    # This code rebuilds the available volumes array from what is left in the filtered hash.
    # The volumes that we return from this method should no longer exist in the filtered_hash
    # or the available volumes array
    ###############################################################################
    foreach my $cluster_useability (keys (%{$filtered_hash})) {
        foreach my $cluster (keys (%{$filtered_hash->{$cluster_useability}})) {
            foreach my $vserver_useability (keys (%{$filtered_hash->{$cluster_useability}->{$cluster}})) {
                foreach my $vserver (keys (%{$filtered_hash->{$cluster_useability}->{$cluster}->{$vserver_useability}})) {
                    foreach my $volume_useability (keys (%{$filtered_hash->{$cluster_useability}->{$cluster}->{$vserver_useability}->{$vserver}})) {
                        push(@{$available_volumes}, @{$filtered_hash->{$cluster_useability}
                                                                                  ->{$cluster}
                                                                                  ->{$vserver_useability}
                                                                                  ->{$vserver}->{$volume_useability}});
                    }
                }
            }
        }
    }

    return (wantarray()) ? @return_volumes : \@return_volumes;
}

1;

__END__

=head1 NAME

PTEFIT::Xgen::Config::Example

=head1 AUTHOR/MAINTAINER

=over

=item PTEFIT AUTOMATION (ng-ptefit-automation-core@netapp.com)

=back

   Last revised $Date: $ by $Author: $

