#
#
# Copyright (c) 2016 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#
#
## @summary Transition Component Module
## @author rohit4@netapp.com, dl-nacl-dev@netapp.com

## @change     2014-12-11 ppranav  : Added 13 new components for CLI usage
## @change     2014-12-12 dinoc    : Added 9 subroutines
## @change     2015-02-04 santosh6 : Added QBT supporting subroutines
## @change     2015-02-20 dinoc    : Added 10 subroutines for CFT
## @change     2015-02-23 ppranav  : Added 3 subroutines for CFT
## @change     2015-04-02 dinoc    : Added Revert and cabling subroutines for CFT
## @change     2015-07-13 santosh6 : Added cft delete command for GUI
## @change     2016-03-21 ashan    : Added changes for Plan Configuration
## @change     2016-04-28 dinoc    : Added GUI transition_credentials_remove
## @chnage     2016-05-17 dinoc    : Added GUI transition_state_machine_status
## @change     2016-06-01 ashan    : Added support for Edit Project from Dashboard
## @status Review
## @pod here
################################################################################

=head1 NAME

NACL::C::Client::Transition

=head1 DESCRIPTION

C<NACL::C::Client::Transition> is a derived class of
L<NACL::C::Component|lib-NACL-C-Component-pm>.

=head1 ATTRIBUTES

=head2 command_interface

See L<NACL::C::Component::command_interface|lib-NACL-C-Component-pm/command_interface>

=head2 session-name

The name of the transition session used to identify it uniquely . (Session names
are unique within a given client on which 7-mode transition tool is installed)

=cut

###################
# Package
package NACL::C::Client::Transition;

###################
# Everytime
use strict;
use warnings;

###################
# Module includes
use Params::Validate qw(:all);
use Params::Validate qw(SCALAR validate_with);
use NACL::ComponentUtils qw(_optional_scalars Dumper);

use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use NACL::Exceptions::InvalidFilterField qw(:try);
use NACL::APISet::Exceptions::CommandFailedException ();
use NACL::Exceptions::InvalidChoice;
use NATE::BaseException qw(:try);
use NATE::Exceptions::Argument ();
use NACL::C::Exceptions::Client::Transition::AlreadyExists;

use base qw( NACL::C::Component::Client );

use NACL::CS::Client::Transition;

=head1 METHODS

=head2 new

    my $trans_obj = NACL::C::Client::Transition->new(
        command_interface => $Client,
        "project-name"    => $project_name,
        "group-name"      => $group_name,
    );

Return a new transition component object with the given attributes.
Project name and group name are the primary keys and must be supplied while creating a transition
object.

=cut

use Class::MethodMaker [
    scalar => "project_name",
    scalar => "group_name",
    scalar => "session_name",
];

sub _primary_keys_validate_spec {
    return (
        "project-name" => { type => SCALAR ,optional => 1 },
        "group-name"   => { type => SCALAR ,optional => 1 },
        "session-name" => { type => SCALAR ,optional => 1 },
    );
}

=head2 credentials_add

    $resp = NACL::C::Client::Transition->credentials_add(
        command_interface => $Client,
        "host-name"       => $hostname,
        username          => $username,
        password          => $password
    );

(Class method) This method is used to add the credentials of either 7 or C-mode.
This method takes care of handling interactive prompt for providing the password
which is not handled by API layer.
"Returns the response object."

"Supports CLI, GUI".

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< host-name => $hostname >>

(Required) Hostname or IP address of 7mode or Cmode.

=item C<< username => $username >>

(Required) Corresponding username for 7mode or Cmode.

=item C<< password => $password >>

(Optional) Corresponding password for 7mode or Cmode .The default value is netapp1!

=back

=back

=cut

sub credentials_add {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            password => {type => SCALAR, default => 'netapp1!'},
            _optional_scalars qw( host-name username ),
        },
        ignore_primary_keys => 1,
    );
    my $obj = $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_credentials_add_smtt_cli',
                interface => 'CLI',
                set       => 'SMTT',
            },
            {   method    => '_credentials_add_smtt_gui',
                interface => 'GUI',
                set       => 'SMTT',
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $obj;
} ## end sub credentials_add

sub _credentials_add_smtt_cli {
    $Log->enter() if $may_enter;
    my ($pkg,@args) = @_;
    my %opts = $pkg->_validate_backend_opts(@args);
    my $password=delete $opts{password} ;
    my $apiset = delete $opts{apiset};
    my %target;
    $pkg->_hash_copy(
        source => \%opts,
        target => \%target,
        copy   => [qw(host-name username)]
    );
    my $force_prompt = ["Enter password for" => $password ];
    my $response = $apiset->transition_credentials_add(
        %target,
        'connectrec-match_table' => $force_prompt,
    );
    $Log->exit() if $may_exit;
    return $response;
} ## end sub _credentials_add_smtt_cli

sub _credentials_add_smtt_gui {
    $Log->enter() if $may_enter;
    my ($pkg,@args) = @_;
    my %opts = $pkg->_validate_backend_opts(@args);
    my $password=delete $opts{password} ;
    my $apiset = delete $opts{apiset};
    my %target;
    $pkg->_hash_copy(
        source => \%opts,
        target => \%target,
        copy   => [qw(host-name username)]
    );
    my $response = $apiset->transition_credentials_add(
        %target,
        password => $password
    );
    $Log->exit() if $may_exit;
    return $response;
} ## end sub _credentials_add_smtt_gui

=head2 credentials_modify

    $resp = NACL::C::Client::Transition->credentials_modify(
        command_interface => $Client,
        "host-name"       => $hostname,
        username          => $username,
        password          => $password
    );

(Class method) This method is used to modify the credentials of either 7 or
C-mode. This method takes care of handling interactive prompt for providing the
password which is not handled by API layer.
"Returns the response object."

"Supports CLI".

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< host-name => $hostname >>

(Required) Hostname or IP address of 7mode or Cmode.

=item C<< username => $username >>

(Required) Corresponding username for 7mode or Cmode.

=item C<< password => $password >>

(Optional) Corresponding password for 7mode or Cmode .The default value is netapp1!

=back

=back

=cut

sub credentials_modify {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            password => {type => SCALAR, default => 'netapp1!'},
            _optional_scalars qw( host-name username ),
        },
        ignore_primary_keys => 1,
    );
    my $obj = $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_credentials_modify_smtt_cli',
                interface => 'CLI',
                set       => 'SMTT',
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $obj;
} ## end sub credentials_modify

sub _credentials_modify_smtt_cli {
    $Log->enter() if $may_enter;
    my ($pkg,@args) = @_;
    my %opts = $pkg->_validate_backend_opts(@args);
    my $password=delete $opts{password} ;
    my $apiset = delete $opts{apiset};
    my %target;
    $pkg->_hash_copy(
        source => \%opts,
        target => \%target,
        copy   => [qw(host-name username)]
    );
    my $force_prompt = ["Enter password for" => $password ];
    my $response = $apiset->transition_credentials_modify(
        %target,
        'connectrec-match_table' => $force_prompt,
    );
    $Log->exit() if $may_exit;
    return $response;
} ## end sub _credentials_modify_smtt_cli

=head2 credentials_get

    $resp = NACL::C::Client::Transition->credentials_get(
        command_interface => $Client,
        "host-name"       => $hostname,
    );

(Class method) This method is used to get the credentials of either 7 or C-mode.
"Returns the response object."

"Supports CLI".

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< host-name => $hostname >>

(Required) Hostname or IP address of 7mode or Cmode.
=back

=back

=cut

sub credentials_get {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars qw( host-name
            ),
        },
        ignore_primary_keys => 1,
    );
    my $obj = $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_credentials_get_smtt_cli',
                interface => 'CLI',
                set       => 'SMTT',
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $obj;
} ## end sub credentials_get

sub _credentials_get_smtt_cli {
    $Log->enter() if $may_enter;
    my ($pkg,@args) = @_;
    my %opts = $pkg->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my %target;
    $pkg->_hash_copy(
        source => \%opts,
        target => \%target,
        copy   => [qw(host-name)]
    );
    my $response = $apiset->transition_credentials_get(%target);
    $Log->exit() if $may_exit;
    return $response;
} ## end sub _credentials_get_smtt_cli

=head2 credentials_remove

    $resp = NACL::C::Client::Transition->credentials_remove(
        command_interface => $Client,
        "host-name"       => $hostname,
    );

(Class method) This method is used to remove the credentials of either 7 or
C-mode.
"Returns the response object."

"Supports CLI".

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< host-name => $hostname >>

(Required) Hostname or IP address of 7mode or Cmode.
=back

=back

=cut

sub credentials_remove {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars qw( host-name
            ),
        },
        ignore_primary_keys => 1,
    );
    my $obj = $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_credentials_remove_smtt_cli',
                interface => 'CLI',
                set       => 'SMTT',
            },
            {   method    => '_credentials_remove_smtt_gui',
                interface => 'GUI',
                set       => 'SMTT',
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $obj;
} ## end sub credentials_remove

sub _credentials_remove_smtt_cli {
    $Log->enter() if $may_enter;
    my ($pkg,@args) = @_;
    my %opts = $pkg->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my %target;
    $pkg->_hash_copy(
        source => \%opts,
        target => \%target,
        copy   => [qw(host-name)]
    );
    my $response = $apiset->transition_credentials_remove(%target
    );
    $Log->exit() if $may_exit;
    return $response;
} ## end sub _credentials_remove_smtt_cli

sub _credentials_remove_smtt_gui {
    $Log->enter() if $may_enter;
    my ($pkg,@args) = @_;
    my %opts = $pkg->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my %target;
    $pkg->_hash_copy(
        source => \%opts,
        target => \%target,
        copy   => [qw(host-name)]
    );
    my $response = $apiset->transition_credentials_remove(%target
    );
    $Log->exit() if $may_exit;
    return $response;
} ## end sub _credentials_remove_smtt_gui

=head2 create

    $trans_obj=NACL::C::Client::Transition->create(
        command_interface         => $Client,
        "session-name"            => $session_name,
        "session-type"            => $type,
        "7-mode-host-name"        => $host_7,
        "data-copy-ipaddress"     => $data_copy_ip,
        vfiler                    => $vfiler,
        "c-mode-host-name"        => $host_C,
        "svm-name"                => $vserver,
    );

(Class method) This method is used to create the transition session .

"Supports CLI, GUI".

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< session-name => $name >>

(Required) Name of the transition session .

=item C<< session-type => $type >>

(Required) Transition session type . It can be standalone , primary or secondary .

=item C<< 7-mode-host-name => $host_7 >>

(Required) Hostname or IP address of 7mode .

=item C<< data-copy-ipaddress => $data_copy_ip >>

(Required) Data copy IP address

=item C<< vfiler => $vfiler >>

(Required) Name of the vfiler from which volume will be transitioned .

=item C<< multipathing-ipaddress => $multipath >>

(Optional) Multipathing IP address .

=item C<< c-mode-host-name => $host_C >>

(Required) Hostname or IP address of Cmode .

=item C<< svm-name => $vser >>

(Required) Name of the vserver to which volumes will be transitioned .

=item C<< svm-rootvol-aggr => $vser_rootvol_aggr >>

(Optional) Vserver root volume aggregate name .

=item C<< schedule-name => $sched >>

(Optional) Snapmirror schedule for 7-C updates .

=back

=back

=cut

sub create {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            "S_mode_credentials"        => { optional => 1 },
            "C_mode_credentials"        => { optional => 1 },
            "volumepair_add_args"       => { optional => 1 },
            "schedule_add_args"         => { type => ARRAYREF, optional => 1 },
            "seven-mode-only"           => { type => SCALAR,   optional => 1 },
            "plan_configuration"        => { type => HASHREF,  optional => 1 },
            "create_new_grp"            => {type => SCALAR,    optional=>1   },
            _optional_scalars qw(
                session-type 7-mode-host-name data-copy-ipaddress
                vfiler multipathing-ipaddress c-mode-host-name svm-name
                svm-rootvol-aggr group-name
            ),
        },
    );
    my $obj;
    try{
        $obj = $pkg->call_on_apiset(
            %opts,
            choices => [
                {   method    => '_create_smtt_cli',
                    interface => 'CLI',
                    set       => 'SMTT',
                },
                {   method    => '_create_smtt_gui',
                    interface => 'GUI',
                    set       => 'SMTT',
                },
            ],
        );
    } catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = $_[0];
        NACL::C::Exceptions::Client::Transition::AlreadyExists
            ->detect_convert_and_throw(exception => $exception);
        $Log->exit() if $may_exit;
        $exception->throw();
    };
    $Log->exit() if $may_exit;
    return $obj;
} ## end sub create

sub _create_smtt_cli {
    $Log->enter() if $may_enter;
    my ($pkg,@args) = @_;
    my %opts = $pkg->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my %target;
    $pkg->_hash_copy(
        source => \%opts,
        target => \%target,
        copy   => [qw( session-name session-type 7-mode-host-name vfiler
            multipathing-ipaddress c-mode-host-name svm-name
            svm-rootvol-aggr data-copy-ipaddress )
        ]
    );
    $apiset->transition_create(%target);
    $Log->exit() if $may_exit;
    return $pkg->new(
        command_interface => $opts{command_interface},
        'session-name'    => $opts{'session-name'}
    );
} ## end sub _create_smtt_cli

sub _create_smtt_gui {
    $Log->enter() if $may_enter;
    my ($pkg,@args) = @_;
    my %opts = $pkg->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my %target;
    $pkg->_hash_copy(
        source => \%opts,
        target => \%target,
        copy   => [qw( session-name session-type 7-mode-host-name vfiler
            multipathing-ipaddress c-mode-host-name svm-name
            svm-rootvol-aggr data-copy-ipaddress S_mode_credentials
            C_mode_credentials volumepair_add_args schedule_add_args seven-mode-only
            plan_configuration group-name create_new_grp)
        ]
    );
    $apiset->transition_create(%target);
    $Log->exit() if $may_exit;
    return $pkg->new(
        command_interface => $opts{command_interface},
        'session-name'    => $opts{'session-name'}
    );
} ## end sub _create_smtt_gui

=head2 volumepair_add

    $resp = $trans_obj->volumepair_add(
        "volume-name-7-mode"  => [qw(vol1 vol2)],
        "volume-name-c-mode"  => [qw(vol_compo1 vol_compo2)],
        "aggregate-name"      => [qw(aggr1 aggr2)],
    );

                OR

    $resp = $trans_obj->volumepair_add(
        "volume-name-7-mode"  => \@arr1,
        "volume-name-c-mode"  => \@arr2,
        "aggregate-name"      => [qw(aggr1 aggr2)],
    );


(Class or instance method) This method is used to add volumepairs to the session .
"Returns the response object."

"Supports CLI, GUI".

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< session-name => $name >>

(Required for class method, Not Applicable for instance method)
Name of the transition session .

=item C<< volume-name-7-mode => $7_vol_arr_ref >>

(Required) Its a reference to an array which contains 7mode volumes.

=item C<< volume-name-c-mode => $C_vol_arr_ref >>

(Required) Its a reference to an array which contains Cmode volumes.

=item C<< aggregate-name => $aggr >>

(Required) This is the cmode aggregate name where the transitioned volumes will be placed .

=back

=back

=cut

sub volumepair_add {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            "volume-name-7-mode"    => { type => ARRAYREF },
            "volume-name-c-mode"    => { type => ARRAYREF, optional => 1 },
            "aggregate-name"        => { type => ARRAYREF, optional => 1 },
            "transition-as-standalone"
        },
    );
    my @obj = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_volumepair_add_smtt_cli',
                interface => 'CLI',
                set       => 'SMTT',
            },
            {   method    => '_volumepair_add_smtt_gui',
                interface => 'GUI',
                set       => 'SMTT',
            },
        ],
    );
    $Log->exit() if $may_exit;
    return wantarray ? @obj : $obj[0];
} ## end sub volumepair_add

sub _volumepair_add_smtt_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my $volume_name_7_mode = delete $opts{"volume-name-7-mode"};
    my $volume_name_c_mode = delete $opts{"volume-name-c-mode"};
    my $aggregate_name     = delete $opts{"aggregate-name"};
    if($opts{'transition-as-standalone'}) {$opts{'transition-as-standalone'}="";}
    else {delete $opts{'transition-as-standalone'};}
    my (%target, $response, @response_objs);
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%target,
        copy   => [qw (session-name transition-as-standalone) ]
    );
    foreach (0..@$volume_name_7_mode-1){
        $target{"volume-name-7-mode"}=$volume_name_7_mode->[$_];
        $target{"volume-name-c-mode"}=$volume_name_c_mode->[$_] if ( $volume_name_c_mode->[$_] );
        $target{"aggregate-name"}=$aggregate_name->[$_] if ( $aggregate_name->[$_] );
        $response = $apiset->transition_volumepair_add(%target);
        push @response_objs,$response;
    }
    $Log->exit() if $may_exit;
    return @response_objs;
} ## end sub _volumepair_add_smtt_cli

sub _volumepair_add_smtt_gui {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my $volume_name_7_mode = delete $opts{"volume-name-7-mode"};
    my $volume_name_c_mode = delete $opts{"volume-name-c-mode"};
    my $aggregate_name     = delete $opts{"aggregate-name"};
    my (%target, $response, @response_objs);
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%target,
        copy   => [qw (session-name) ]
    );
    $target{"volume-name-7-mode"}=$volume_name_7_mode;
    $target{"volume-name-c-mode"}=$volume_name_c_mode if ( $volume_name_c_mode );
    $target{"aggregate-name"}=$aggregate_name if ( $aggregate_name );
    $response = $apiset->transition_volumepair_add(%target);
    push @response_objs,$response;

    $Log->exit() if $may_exit;
    return @response_objs;
} ## end sub _volumepair_add_smtt_gui

=head2 volumepair_modify

    $resp = $trans_obj->volumepair_modify(
        "volume-name-7-mode"  => [qw(vol1 vol2)],
        "volume-name-c-mode"  => [qw(vol_compo3 vol_compo4)],
        "aggregate-name"      => [qw(aggr1 aggr2)],
    );

                OR

    $resp = $trans_obj->volumepair_modify(
        "volume-name-7-mode"  => \@arr1,
        "volume-name-c-mode"  => \@arr2,
        "aggregate-name"      => [qw(aggr1 aggr2)],
    );


(Class or instance method) This method is used to modify volumepairs which are
existing in the transition session.
"Returns the response object."

"Supports CLI".

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< session-name => $name >>

(Required for class method, Not Applicable for instance method)
Name of the transition session .

=item C<< volume-name-7-mode => $7_vol_arr_ref >>

(Required) Its a reference to an array which contains 7mode volumes.

=item C<< volume-name-c-mode => $C_vol_arr_ref >>

(Required) Its a reference to an array which contains Cmode volumes.

=item C<< aggregate-name => $aggr >>

(Required) This is the cmode aggregate name where the transitioned volumes will
be placed .

=back

=back

=cut

sub volumepair_modify {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            "volume-name-7-mode"    => { type => ARRAYREF, optional => 0 },
            "volume-name-c-mode"    => { type => ARRAYREF, optional => 1 },
            "aggregate-name"        => { type => ARRAYREF, optional => 1 },
            "transition-as-standalone" => {type => SCALAR,optional=>1},
        },
    );
    my @obj = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_volumepair_modify_smtt_cli',
                interface => 'CLI',
                set       => 'SMTT',
            },
        ],
    );
    $Log->exit() if $may_exit;
    return wantarray ? @obj : $obj[0];
} ## end sub volumepair_modify

sub _volumepair_modify_smtt_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my $volume_name_7_mode = delete $opts{"volume-name-7-mode"};
    my $volume_name_c_mode = delete $opts{"volume-name-c-mode"};
    my $aggregate_name     = delete $opts{"aggregate-name"};
    my $transition_as_standalone     = delete $opts{"transition-as-standalone"};
    my (%target, $response, @response_objs);
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%target,
        copy   => [qw (session-name) ]
    );
    foreach (0..@$volume_name_7_mode-1){
        $target{"volume-name-7-mode"} = $volume_name_7_mode->[$_];
        $target{"volume-name-c-mode"} = $volume_name_c_mode->[$_]
            if ( $volume_name_c_mode->[$_] );
        $target{"aggregate-name"} = $aggregate_name->[$_]
            if ( $aggregate_name->[$_] );
        $target{"transition-as-standalone"} = $transition_as_standalone->[$_]
            if ( $transition_as_standalone->[$_] );
        $response = $apiset->transition_volumepair_modify(%target);
        push @response_objs,$response;
    }
    $Log->exit() if $may_exit;
    return @response_objs;
} ## end sub _volumepair_modify_smtt_cli

=head2 volumepair_show

    $resp = $trans_obj->volumepair_show(
        "volume-name-7-mode"  => "vol1",
    );
               AND
    $resp = $trans_obj->volumepair_show(
        "volume-name-7-mode"  => "vol2",
    );

(Class or instance method) This method is used to display volumepairs which are
existing in the transition session.
"Returns the response object."

"Supports CLI".

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< session-name => $name >>
(Required for class method, Not Applicable for instance method)
Name of the transition session .

=item C<< volume-name-7-mode => $7_vol_arr_ref >>

(Optional) Names of 7mode volumes; one name at a time

=back

=back

=cut

sub volumepair_show {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            "volume-name-7-mode"    => { type => SCALAR, optional => 1 },
        },
    );
    my @obj = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_volumepair_show_smtt_cli',
                interface => 'CLI',
                set       => 'SMTT',
            },
        ],
    );
    $Log->exit() if $may_exit;
    return wantarray ? @obj : $obj[0];
} ## end sub volumepair_show

sub _volumepair_show_smtt_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my $volume_name_7_mode = delete $opts{"volume-name-7-mode"};
    my (%target, $response, @response_objs);
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%target,
        copy   => [qw (session-name) ]
    );
    $target{"volume-name-7-mode"} = $volume_name_7_mode
        if ( $volume_name_7_mode );
    $response = $apiset->transition_volumepair_show(%target);
    $Log->exit() if $may_exit;
    return $response;
} ## end sub _volumepair_show_smtt_cli

=head2 volumepair_remove

    $resp = $trans_obj->volumepair_remove(
        "volume-name-7-mode"  => [qw(vol1 vol2)],
    );

                OR

    $resp = $trans_obj->volumepair_remove(
        "volume-name-7-mode"  => \@arr1,
    );


(Class or instance method) This method is used to remove volumes from the session .
"Returns the response object."

"Supports CLI, GUI".

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< session-name => $name >>

(Required for class method, Not Applicable for instance method)
Name of the transition session .

=item C<< volume-name-7-mode => $7_vol_arr_ref >>
=item C<< group-name => $group-name

(Required) Its a reference to an array which contains 7mode volumes.

=back

=back

=cut

sub volumepair_remove {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            "volume-name-7-mode"    => { type => ARRAYREF },
              _optional_scalars qw( group-name ),
        },
    );
    my @obj = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_volumepair_remove_smtt_cli',
                interface => 'CLI',
                set       => 'SMTT',
            },
            {   method    => '_volumepair_remove_smtt_gui',
                interface => 'GUI',
                set       => 'SMTT',
            },
        ],
    );
    $Log->exit() if $may_exit;
    return wantarray ? @obj : $obj[0];
} ## end sub volumepair_remove

sub _volumepair_remove_smtt_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my $volume_name_7_mode = delete $opts{"volume-name-7-mode"};
    my (%target, $response, @response_objs);
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%target,
        copy   => [qw (session-name) ]
    );
    foreach (0..@$volume_name_7_mode-1){
        $target{"volume-name-7-mode"}=$volume_name_7_mode->[$_];
        $response = $apiset->transition_volumepair_remove(%target);
        push @response_objs,$response;
    }
    $Log->exit() if $may_exit;
    return @response_objs;
} ## end sub _volumepair_remove_smtt_cli

sub _volumepair_remove_smtt_gui {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my $volume_name_7_mode = delete $opts{"volume-name-7-mode"};
    my (%target, $response, @response_objs);
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%target,
        copy   => [qw (session-name  group-name) ]
    );
    $target{"volume-name-7-mode"}=$volume_name_7_mode;
    $response = $apiset->transition_volumepair_remove(%target);
    push @response_objs,$response;

    $Log->exit() if $may_exit;
    return @response_objs;
} ## end sub _volumepair_remove_smtt_gui


=head2 lif_add

    $resp = $trans_obj->lif_add(
        "schedule-name"            => $sc,
        "ip-address"               => $target_ip_address,
        "default-gateway"          => $gateway,
        "netmask"                  => $netmask,
        "home-node"                => $node,
        "home-port"                => $port,
        "origin-ip-address"        => $origin_ip_address,
    );
(Class or instance method )
This method is used to add lif to the transition session.
"Returns the response object."

"Supports CLI, GUI".

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< session-name => $name >>

(Required for class method, Not Applicable for instance method)
Name of the transition session .

=item C<< ip-address => $ip_address >>

(Optional)

=item C<< default-gateway => $gateway >>

(Optional)

=item C<< netmask => $netmask >>

(Optional)

=item C<< home-node => $node >>

(Optional)

=item C<< home-port => $port >>

(Optional)

=item C<< origin-ip-address => $origin_ip_address >>

=back

=back

=cut

sub lif_add{
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
             _optional_scalars qw( ip-address default-gateway
                netmask home-node home-port origin-ip-address
            ),
        },
    );
    my $obj = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_lif_add_smtt_cli',
                interface => 'CLI',
                set       => 'SMTT',
            },
            {   method    => '_lif_add_smtt_gui',
                interface => 'GUI',
                set       => 'SMTT',
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $obj;
} ## end sub lif_add

sub _lif_add_smtt_cli{
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my (%target, $response);
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%target,
        copy   => [qw (session-name ip-address default-gateway
                netmask home-node home-port origin-ip-address ) ]
    );
    $response = $apiset->transition_lif_add(%target);
    $Log->exit() if $may_exit;
    return $response;
} ## end sub _lif_add_smtt_cli

sub _lif_add_smtt_gui{
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my (%target, $response);
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%target,
        copy   => [qw (session-name ip-address default-gateway
                netmask home-node home-port origin-ip-address ) ]
    );
    $response = $apiset->transition_lif_add(%target);
    $Log->exit() if $may_exit;
    return $response;
} ## end sub _lif_add_smtt_gui

=head2 operation_history

    $resp = $trans_obj->operation_history(
        "operation-name" => "Precheck",
    );

(Class or instance method) This method is used to get operation details for a session.
"Returns the response object."

"Supports  GUI".

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< operation-name => $operation >>
(Required for class method, Not Applicable for instance method)
Name of the transition session .

(Required) Its a reference to a hash contained parsed output of a particular operation.

=back

=back

=cut

sub operation_history {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            "operation-name"    => { type => SCALAR },
			_optional_scalars qw( group-name ),
        },
    );
    my @obj = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_operation_history_smtt_gui',
                interface => 'GUI',
                set       => 'SMTT',
            },
        ],
    );
    $Log->exit() if $may_exit;
    return wantarray ? @obj : $obj[0];
} ## end sub operation_history

sub _operation_history_smtt_gui {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my $operation = delete $opts{"operation-name"};
    my (%target, $response, @response_objs);
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%target,
        copy   => [qw (session-name group-name) ]
    );
    $target{"operation-name"}=$operation;
    $response = $apiset->transition_operation_history(%target);
    push @response_objs,$response;

    $Log->exit() if $may_exit;
    return @response_objs;
} ## end sub _operation_history_smtt_gui


=head2 lif_modify

    $resp = $trans_obj->lif_modify(
        "ip-address"               => $target_ip_address,
        "default-gateway"          => $gateway,
        "netmask"                  => $netmask,
        "home-node"                => $node,
        "home-port"                => $port,
    );
(Class or instance method )
This method is used to modify lif to the transition session.
"Returns the response object."

"Supports CLI".

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< session-name => $name >>

(Required for class method, Not Applicable for instance method)
Name of the transition session .

=item C<< ip-address => $ip_address >>

(Optional)

=item C<< default-gateway => $gateway >>

(Optional)

=item C<< netmask => $netmask >>

(Optional)

=item C<< home-node => $node >>

(Optional)

=item C<< home-port => $port >>

=back

=back

=cut

sub lif_modify{
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
             _optional_scalars qw( ip-address default-gateway
                netmask home-node home-port
            ),
        },
    );
    my $obj = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_lif_modify_smtt_cli',
                interface => 'CLI',
                set       => 'SMTT',
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $obj;
} ## end sub lif_modify

sub _lif_modify_smtt_cli{
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my (%target, $response);
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%target,
        copy   => [qw (session-name ip-address default-gateway
                netmask home-node home-port ) ]
    );
    $response = $apiset->transition_lif_modify(%target);
    $Log->exit() if $may_exit;
    return $response;
} ## end sub _lif_modify_smtt_cli

=head2 lif_show

    $resp = $trans_obj->lif_show(
        "ip-address"  => "1.12.255.3",
    );
Shows details of lif for given lif IP

                OR
    $resp = $trans_obj->lif_show( );
Shows details of lifs to the transition session.

(Class or instance method) This method is used to remove lif from the session .
"Returns the response object."

"Supports CLI".

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< session-name => $name >>

(Required for class method, Not Applicable for instance method)
Name of the transition session .

=item C<< ip-address => $ip_address >>
(Required)If a particular lif details are needed.
(Optional)If details of all lifs are needed at a time

=back

=back

=cut

sub lif_show {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            "ip-address"    => { type => SCALAR, optional => 1 },
        },
    );
    my @obj = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_lif_show_smtt_cli',
                interface => 'CLI',
                set       => 'SMTT',
            },
        ],
    );
    $Log->exit() if $may_exit;
    return wantarray ? @obj : $obj[0];
} ## end sub lif_show

sub _lif_show_smtt_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my $ipaddress = delete $opts{"ip-address"};
    my (%target, $response, @response_objs);
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%target,
        copy   => [qw (session-name) ]
    );
    $target{"ip-address"} = $ipaddress if ( $ipaddress );
    $response = $apiset->transition_lif_show(%target);
    $Log->exit() if $may_exit;
    return $response;
} ## end sub _lif_show_smtt_cli

=head2 lif_remove

    $resp = $trans_obj->lif_remove(
        "ip-address"  => [qw(1.12.255.2 1.12.255.3)],
    );

                OR

    $resp = $trans_obj->lif_remove(
        "ip-address"  => \@arr1,
    );


(Class or instance method) This method is used to remove lif from the session .
"Returns the response object."

"Supports CLI, GUI".

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< session-name => $name >>

(Required for class method, Not Applicable for instance method)
Name of the transition session .

=item C<< ip-address => $ip_arr_ref >>

(Required) Its a reference to an array which contains LIFs.

=back

=back

=cut

sub lif_remove {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            "ip-address"    => { type => ARRAYREF },
        },
    );
    my @obj = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_lif_remove_smtt_cli',
                interface => 'CLI',
                set       => 'SMTT',
            },
            {   method    => '_lif_remove_smtt_gui',
                interface => 'GUI',
                set       => 'SMTT',
            },
        ],
    );
    $Log->exit() if $may_exit;
    return wantarray ? @obj : $obj[0];
} ## end sub lif_remove

sub _lif_remove_smtt_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my $ipaddress = delete $opts{"ip-address"};
    my (%target, $response, @response_objs);
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%target,
        copy   => [qw (session-name) ]
    );
    foreach (0..@$ipaddress-1){
        $target{"ip-address"}=$ipaddress->[$_];
        $response = $apiset->transition_lif_remove(%target);
        push @response_objs,$response;
    }
    $Log->exit() if $may_exit;
    return @response_objs;
} ## end sub _lif_remove_smtt_cli

sub _lif_remove_smtt_gui {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my $ipaddress = delete $opts{"ip-address"};
    my (%target, $response, @response_objs);
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%target,
        copy   => [qw (session-name) ]
    );
    $target{"ip-address"}=$ipaddress;
    $response = $apiset->transition_lif_remove(%target);
    push @response_objs,$response;

    $Log->exit() if $may_exit;
    return @response_objs;
} ## end sub _lif_remove_smtt_gui

=head2 editproject_volumepair_add

    $resp = $trans_obj->editproject_volumepair_add(
        "volume-name-7-mode"  => [qw(vol1 vol2)],
        "volume-name-c-mode"  => [qw(vol_compo1 vol_compo2)],
        "aggregate-name"      => [qw(aggr1 aggr2)],
    );

                OR

    $resp = $trans_obj->editproject_volumepair_add(
        "volume-name-7-mode"  => \@arr1,
        "volume-name-c-mode"  => \@arr2,
        "aggregate-name"      => [qw(aggr1 aggr2)],
    );


(Class or instance method) This method is used to add volumepairs to the session 
using Project edit button from the Project Dashboard.
"Returns the response object."

"Supports CLI, GUI".

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< session-name => $name >>

(Required for class method, Not Applicable for instance method)
Name of the transition session .

=item C<< volume-name-7-mode => $7_vol_arr_ref >>

(Required) Its a reference to an array which contains 7mode volumes.

=item C<< volume-name-c-mode => $C_vol_arr_ref >>

(Required) Its a reference to an array which contains Cmode volumes.

=item C<< aggregate-name => $aggr >>

(Required) This is the cmode aggregate name where the transitioned volumes will be placed .

=back

=back

=cut

sub editproject_volumepair_add {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            "volume-name-7-mode"    => { type => ARRAYREF },
            "volume-name-c-mode"    => { type => ARRAYREF, optional => 1 },
            "aggregate-name"        => { type => ARRAYREF, optional => 1 },
        },
    );
    my @obj = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_editproject_volumepair_add_smtt_gui',
                interface => 'GUI',
                set       => 'SMTT',
            },
        ],
    );
    $Log->exit();
    return wantarray ? @obj : $obj[0];
} ## end sub editproject_volumepair_add


sub _editproject_volumepair_add_smtt_gui {
    $Log->enter();
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my $volume_name_7_mode = delete $opts{"volume-name-7-mode"};
    my $volume_name_c_mode = delete $opts{"volume-name-c-mode"};
    my $aggregate_name     = delete $opts{"aggregate-name"};
    my (%target, $response, @response_objs);
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%target,
        copy   => [qw (session-name) ]
    );
    $target{"volume-name-7-mode"}=$volume_name_7_mode;
    $target{"volume-name-c-mode"}=$volume_name_c_mode if ( $volume_name_c_mode );
    $target{"aggregate-name"}=$aggregate_name if ( $aggregate_name );
    $response = $apiset->transition_editproject_volumepair_add(%target);
    push @response_objs,$response;

    $Log->exit();
    return @response_objs;
} ## end sub _editproject_volumepair_add_smtt_gui

=head2 editproject_volumepair_remove

    $resp = $trans_obj->editproject_volumepair_remove(
        "volume-name-7-mode"  => [qw(vol1 vol2)],
    );

                OR

    $resp = $trans_obj->editproject_volumepair_remove(
        "volume-name-7-mode"  => \@arr1,
    );


(Class or instance method) This method is used to remove volumes from the session .
"Returns the response object."

"Supports CLI, GUI".

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< session-name => $name >>

(Required for class method, Not Applicable for instance method)
Name of the transition session .

=item C<< volume-name-7-mode => $7_vol_arr_ref >>

(Required) Its a reference to an array which contains 7mode volumes.

=back

=back

=cut

sub editproject_volumepair_remove {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            "volume-name-7-mode"    => { type => ARRAYREF },
        },
    );
    my @obj = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_editproject_volumepair_remove_smtt_gui',
                interface => 'GUI',
                set       => 'SMTT',
            },
        ],
    );
    $Log->exit();
    return wantarray ? @obj : $obj[0];
} ## end sub editproject_volumepair_remove

sub _editproject_volumepair_remove_smtt_gui {
    $Log->enter();
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my $volume_name_7_mode = delete $opts{"volume-name-7-mode"};
    my (%target, $response, @response_objs);
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%target,
        copy   => [qw (session-name) ]
    );
    $target{"volume-name-7-mode"}=$volume_name_7_mode;
    $response = $apiset->transition_editproject_volumepair_remove(%target);
    push @response_objs,$response;

    $Log->exit();
    return @response_objs;
} ## end sub _editproject_volumepair_remove_smtt_gui

=head2 editproject_data_ip_modify

    $resp = $trans_obj->editproject_data_ip_modify(
        "datacopy-ip-address"           => $data_copy_ip,
        "multipath-ip-address"          => $multipath_ip,
    );
(Class or instance method )
This method is used to add data copy IP and Multipath IP addresses.
"Returns the response object."

"Supports GUI".

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< session-name => $name >>

(Required for class method, Not Applicable for instance method)
Name of the transition session .

=item C<< datacopy-ip-address => $data_copy_ip >>

(Required)

=item C<< multipath-ip-address => $multipath_ip >>

(Optional)

=back

=back

=cut

sub editproject_data_ip_modify{

	$Log->enter();
	my ($pkg_or_obj, @args) = @_;
	$pkg_or_obj->_frontend_log_debug_opts(@args);
	my %opts = $pkg_or_obj->_common_validate_with(
		params 			=> \@args,
		additional_spec => {
			_optional_scalars qw( session-name group-name datacopy-ip-address multipath-ip-address			
			),
		},
	);
	my $obj = $pkg_or_obj->call_on_apiset(
		%opts,
		choices => [
			{	method 		=> '_editproject_data_ip_modify_gui',
				interface 	=> 'GUI',
				set			=> 'SMTT',
			},
		],
	);
}#end editproject_data_ip_modify

sub _editproject_data_ip_modify_gui{
	$Log->enter();
	my ($pkg_or_obj,@args) = @_;
	my %opts = $pkg_or_obj->_validate_backend_opts(@args);
	my $apiset = delete $opts{apiset};
	my %target;
	my $response;
	$pkg_or_obj->_hash_copy(
		source => \%opts,
		target => \%target,
		copy   => [qw ( session-name group-name datacopy-ip-address multipath-ip-address )]
	);
	$response = $apiset->transition_editproject_data_ip_modify(%target);
	$Log->exit();
	return $response;
}

=head2 editproject_svm_modify

    $resp = $trans_obj->editproject_svm_modify(
        "svm-name"           => $svm_name,
    );
(Class or instance method )
This method is used to Select the target cluster and SVM to which
 the 7-Mode volumes must be transitioned.
"Returns the response object."

"Supports GUI".

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< session-name => $name >>

(Required for class method, Not Applicable for instance method)
Name of the transition session .

=item C<< svm-name => $svm_name >>

(Required)

=back

=back

=cut

sub editproject_svm_modify{

	$Log->enter();
	my ($pkg_or_obj, @args) = @_;
	$pkg_or_obj->_frontend_log_debug_opts(@args);
	my %opts = $pkg_or_obj->_common_validate_with(
		params 			=> \@args,
		additional_spec => {
			_optional_scalars qw( session-name svm-name
			),
		},
	);
	my $obj = $pkg_or_obj->call_on_apiset(
		%opts,
		choices => [
			{	method 		=> '_editproject_svm_modify_gui',
				interface 	=> 'GUI',
				set			=> 'SMTT',
			},
		],
	);
}

sub _editproject_svm_modify_gui{
	$Log->enter();
	my ($pkg_or_obj,@args) = @_;
	my %opts = $pkg_or_obj->_validate_backend_opts(@args);
	my $apiset = delete $opts{apiset};
	my %target;
	my $response;
	$pkg_or_obj->_hash_copy(
		source => \%opts,
		target => \%target,
		copy   => [qw ( session-name svm-name )]
	);
	$response = $apiset->transition_editproject_svm_modify(%target);
	$Log->exit();
	return $response;
}

=head2 editproject_schedule_modify

    $resp = $trans_obj->editproject_schedule_modify(
        "schedule-name"           => $sc,
        "days-range"                    => $days_range,
        "start-time"                    => $start_time,
        "duration"                      => $duration,
        "available-transfers-percentage"=> $available-transfers-percentage,
        "session-snapmirror-throttle"   => $throttle,
    );
(Class or instance method)
This method is used to modify the schedule of the transition session using Edit Configuration
button from the Project Dashboard
"Returns the response object."

"Supports CLI".

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< session-name => $name >>

(Required for class method, Not Applicable for instance method)
Name of the transition session .

=item C<< schedule-name => $sc >>

(Required)

=item C<< days-range => $days_range >>

(Optional)

=item C<< start-time => $start_time >>

(Optional)

=item C<< duration => $duration >>

(Optional)

=item C<< available-transfers-percentage => $available-transfers-percentage >>

(Optional)

=item C<< session-snapmirror-throttle => $throttle >>

(Optional)

=item C<< update-frequency => $update_frequency >>

=back

=back

=cut

sub editproject_schedule_modify{
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars qw( schedule-name 7-Mode-host days-range
                start-time duration available-transfers-percentage session-snapmirror-throttle
                update-frequency max-cap-concurrent-transfers
            ),
        },
    );
    my $obj = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
			{   method    => '_editproject_schedule_modify_smtt_gui',
                interface => 'GUI',
                set       => 'SMTT',
            },
        ],
    );
    $Log->exit();
    return $obj;
} ## end sub schedule_modify

sub _editproject_schedule_modify_smtt_gui{
    $Log->enter(); 
	print "Entered _schedule_modify_smtt_gui \n";
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my %target;
    my $response;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%target,
        copy   => [qw (session-name schedule-name 7-Mode-host days-range start-time duration
            available-transfers-percentage session-snapmirror-throttle update-frequency
            max-cap-concurrent-transfers ) ]
    );
    $response = $apiset->transition_editproject_schedule_modify(%target);
    $Log->exit();
    return $response;

} ##end sub _schedule_modify_smtt_gui

=head2 editproject_schedule_remove

    $resp = $trans_obj->editproject_schedule_remove(
        "schedule-name"           => $sc
    );
(Class or instance method)
This method is used to remove the schedule from the transition session.
"Returns the response object."

"Supports CLI".

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< session-name => $name >>

(Required for class method, Not Applicable for instance method)
Name of the transition session .

=item C<< schedule-name => $sc >>

(Required)

=back

=back

=cut

sub editproject_schedule_remove{
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars qw( schedule-name 7-Mode-host ),
        },
    );
    my $obj = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_editproject_schedule_remove_smtt_gui',
                interface => 'GUI',
                set       => 'SMTT',
            },
        ],
    );
    $Log->exit();
    return $obj;
} ## end sub editproject_schedule_remove

sub _editproject_schedule_remove_smtt_gui{
    $Log->enter();
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my %target;
    my $response;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%target,
        copy   => [qw (session-name schedule-name 7-Mode-host ) ]
    );
    $response = $apiset->transition_editproject_schedule_remove(%target);
    $Log->exit();
    return $response;
} ## end sub _editproject_schedule_remove_smtt_gui

=head2 editproject_plan_config

    $resp = $trans_obj->editproject_plan_config(
        "plan_configuration" => \%plan_config,
    );

(Class or instance method)
This method is used to edit the plan configuration for transition session .
"Returns the response object."

"Supports CLI, GUI".

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< group-name => $group_name >>

(Required) Project Group name

=item C<< session-name => $name >>

(Required for class method, Not Applicable for instance method)
Name of the transition session .

=item C<< plan_configuration => \%plan_config >>

(Required) It is a hash reference.
Configuration that needs to be ignored while transition. Default value is none is ignored.

=back

=back

=cut

sub editproject_plan_config {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            "plan_configuration"        => { type => HASHREF,  optional => 1 },
            _optional_scalars qw(
                group-name
            ),
        },
    );
    my $obj= $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_editproject_plan_config_smtt_gui',
                interface => 'GUI',
                set       => 'SMTT',
            },
        ],
    );
    $Log->exit();
    return $obj;
} ## end sub editproject_plan_config

sub _editproject_plan_config_smtt_gui {
    $Log->enter();
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my %target;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%target,
        copy    => [qw( session-name group-name plan_configuration)]
    );
    $apiset->transition_editproject_plan_config(%target);
    $Log->exit();
} ## end sub _editproject_planconfig_smtt_gui

=head2 export_lunmap

    $resp = $trans_obj->export_lunmap(
        "output-file"  => "file\xyz.txt",
    );

(Class or instance method) This method is used to copy lunmap information to
given output file.
"Returns the response object."

"Supports CLI".

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< session-name => $name >>
(Required for class method, Not Applicable for instance method)
Name of the transition session .

=item C<< output-file => $output-file >>

(Required) output-file path

=back

=back

=cut

sub export_lunmap {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            "output-file"    => { type => SCALAR, optional => 0 },
        },
    );
    my $obj = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_export_lunmap_smtt_cli',
                interface => 'CLI',
                set       => 'SMTT',
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $obj;
} ## end sub export_lunmap

sub _export_lunmap_smtt_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my $output_file = delete $opts{"output-file"};
    my (%target, $response, @response_objs);
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%target,
        copy   => [qw (session-name) ]
    );
    $target{"output-file"} = $output_file;
    $response = $apiset->transition_export_lunmap(%target);
    $Log->exit() if $may_exit;
    return $response;
} ## end sub _export_lunmap_smtt_cli

=head2 version

    $resp = NACL::C::Client::Transition->version(
        command_interface => $Client,
        verbose           => " ",
    );

(Class method) This method is used to display transition tool version
"Returns the response object."

"Supports CLI".

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< verbose => "blank space" >>

(Optional) verbose mode.

=back

=back

=cut

sub version {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            verbose => {type => SCALAR, optional => 1},
        },
        ignore_primary_keys => 1,
    );
    my $obj = $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_version_smtt_cli',
                interface => 'CLI',
                set       => 'SMTT',
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $obj;
} ## end sub version

sub _version_smtt_cli {
    $Log->enter() if $may_enter;
    my ($pkg,@args) = @_;
    my %opts = $pkg->_validate_backend_opts(@args);
    my $verbose = delete $opts{verbose} ;
    my $apiset = delete $opts{apiset};
    my %target;
    $pkg->_hash_copy(
        source => \%opts,
        target => \%target,
    );
    $target{"verbose"} = $verbose if ($verbose =~ /^\s*$/);
    my $response = $apiset->transition_version( %target );
    $Log->exit() if $may_exit;
    return $response;
} ## end sub _version_smtt_cli

=head2 add_primary_seven_mode_system

    $resp = $trans_obj->add_primary_seven_mode_system(
        "session-name"          => $session_name,
        "source-host"           => $host_7,
        "management-ipaddress"  => $management_ip,
        "data-copy-ipaddress"   => $data_copy_ip,
        "multipathing-ipaddress"=> $multipath,
    );
(Class or instance method) This method is used to add
primary seven mode system to the secondary session.
"Returns the response object."

"Supports CLI".

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< session-name => $session_name >>

(Required) Name of the transition session .

=item C<< source-host => $host_7 >>

(Required) Hostname or IP address of 7mode .

=item C<< management-ipaddress => $management_ip >>

(Required) Management IP address of the source .

=item C<< data-copy-address => $data_copy_ip >>

(Required) Data copy IP address .

=item C<< multipathing-ipaddress => $multipath >>

(Optional) Multipathing IP address .

=back

=back

=cut
sub add_primary_seven_mode_system {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars qw( source-host management-ipaddress
                data-copy-ipaddress-staggered multipathing-ipaddress
            ),
        },
    );
    my $obj=$pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_add_primary_seven_mode_system_smtt_cli',
                interface => 'CLI',
                set       => 'SMTT',
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $obj;
} ## end sub add_primary_seven_mode_system

sub _add_primary_seven_mode_system_smtt_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my %target;
    $pkg_or_obj->_hash_copy(
      source => \%opts,
        target => \%target,
        copy    => [qw( session-name source-host management-ipaddress
            data-copy-ipaddress-staggered multipathing-ipaddress
        )]
    );
    my $response = $apiset->transition_add_primary_seven_mode_system(%target);
    $Log->exit() if $may_exit;
    return $response;
} ## end sub _add_primary_seven_mode_system_smtt_cli

=head2 add_secondary_cluster_mode_system

    $resp = $trans_obj->add_secondary_cluster_mode_system(
        "session-name"          => $session_name,
        "c-mode-host-name"      => $host_C,
        "vserver-name"          => $c_mode_vserver,
    );
(Class or instance method) This method is used to add
secondary cluster mode system to the primary session.
"Returns the response object."

"Supports CLI".

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< session-name => $session_name >>

(Required) Name of the transition session .

=item C<< c-mode-host-name => $host_C >>

(Required) Hostname or IP address of Cmode .

=item C<< vserver-name => $c_mode_vserver >>

(Required) Name of the vserver to which volumes will be transitioned .

=back

=back

=cut

sub add_secondary_cluster_mode_system {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars qw( c-mode-host-name
                svm-name ),
        },
    );
    my $obj= $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_add_secondary_cluster_mode_system_smtt_cli',
                interface => 'CLI',
                set       => 'SMTT',
            },
         ],
    );
    $Log->exit() if $may_exit;
    return $obj;
} ## end sub add_secondary_cluster_mode_system

sub _add_secondary_cluster_mode_system_smtt_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my %target;
    $pkg_or_obj->_hash_copy(
       source => \%opts,
       target => \%target,
       copy    => [qw( session-name c-mode-host-name svm-name )]
    );
    my $response = $apiset->transition_add_secondary_cluster_mode_system(%target);
    $Log->exit() if $may_exit;
    return $response;
} ## end sub _add_secondary_cluster_mode_system_smtt_cli

=head2 schedule_add

    $resp = $trans_obj->schedule_add(
        "schedule-name"                 => $sc,
        "days-range"                    => $days_range,
        "start-time"                    => $start_time,
        "duration"                      => $duration,
        "available-transfers-percentage"=> $available-transfers-percentage,
        "session-snapmirror-throttle"   => $throttle,
    );
(Class or instance method )
This method is used to add schedule to the transition session.
"Returns the response object."

"Supports CLI, GUI".

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< session-name => $name >>

(Required for class method, Not Applicable for instance method)
Name of the transition session .

=item C<< schedule-name => $sc >>

(Required)

=item C<< days-range => $days_range >>

(Required)

=item C<< start-time => $start_time >>

(Required)

=item C<< duration => $duration >>

(Required)

=item C<< available-transfers-percentage => $available-transfers-percentage >>

(Required)

=item C<< session-snapmirror-throttle => $throttle >>

(Optional)

=item C<< update-frequency => $update_frequency >>

(Optional)

=back

=back

=cut

sub schedule_add{
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
             _optional_scalars qw( schedule-name 7-Mode-host days-range
                start-time duration available-transfers-percentage session-snapmirror-throttle
                update-frequency max-cap-concurrent-transfers
            ),
        },
    );
    my $obj = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_schedule_add_smtt_cli',
                interface => 'CLI',
                set       => 'SMTT',
            },
            {   method    => '_schedule_add_smtt_gui',
                interface => 'GUI',
                set       => 'SMTT',
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $obj;
} ## end sub schedule_add

sub _schedule_add_smtt_cli{
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my %target;
    my $response;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%target,
        copy   => [qw (session-name schedule-name 7-Mode-host days-range start-time
            duration available-transfers-percentage session-snapmirror-throttle update-frequency
            max-cap-concurrent-transfers ) ]
    );
    $response = $apiset->transition_schedule_add(%target);
    $Log->exit() if $may_exit;
    return $response;
} ## end sub _schedule_add_smtt_cli

sub _schedule_add_smtt_gui{
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my %target;
    my $response;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%target,
        copy   => [qw (session-name schedule-name 7-Mode-host days-range start-time
            duration available-transfers-percentage session-snapmirror-throttle update-frequency
            max-cap-concurrent-transfers ) ]
    );
    $response = $apiset->transition_schedule_add(%target);
    $Log->exit() if $may_exit;
    return $response;
} ## end sub _schedule_add_smtt_gui

=head2 schedule_modify

    $resp = $trans_obj->schedule_modify(
        "schedule-name"           => $sc,
        "days-range"                    => $days_range,
        "start-time"                    => $start_time,
        "duration"                      => $duration,
        "available-transfers-percentage"=> $available-transfers-percentage,
        "session-snapmirror-throttle"   => $throttle,
    );
(Class or instance method)
This method is used to modify the schedule of the transition session.
"Returns the response object."

"Supports CLI".

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< session-name => $name >>

(Required for class method, Not Applicable for instance method)
Name of the transition session .

=item C<< schedule-name => $sc >>

(Required)

=item C<< days-range => $days_range >>

(Optional)

=item C<< start-time => $start_time >>

(Optional)

=item C<< duration => $duration >>

(Optional)

=item C<< available-transfers-percentage => $available-transfers-percentage >>

(Optional)

=item C<< session-snapmirror-throttle => $throttle >>

(Optional)

=item C<< update-frequency => $update_frequency >>

=back

=back

=cut

sub schedule_modify{
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars qw( schedule-name 7-Mode-host days-range
                start-time duration available-transfers-percentage session-snapmirror-throttle
                update-frequency max-cap-concurrent-transfers
            ),
        },
    );
    my $obj = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_schedule_modify_smtt_cli',
                interface => 'CLI',
                set       => 'SMTT',
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $obj;
} ## end sub schedule_modify

sub _schedule_modify_smtt_cli{
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my %target;
    my $response;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%target,
        copy   => [qw (session-name schedule-name 7-Mode-host days-range start-time duration
            available-transfers-percentage session-snapmirror-throttle update-frequency
            max-cap-concurrent-transfers ) ]
    );
    $response = $apiset->transition_schedule_modify(%target);
    $Log->exit() if $may_exit;
    return $response;
} ## end sub _schedule_modify_smtt_cli

=head2 schedule_remove

    $resp = $trans_obj->schedule_remove(
        "schedule-name"           => $sc
    );
(Class or instance method)
This method is used to remove the schedule from the transition session.
"Returns the response object."

"Supports CLI".

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< session-name => $name >>

(Required for class method, Not Applicable for instance method)
Name of the transition session .

=item C<< schedule-name => $sc >>

(Required)

=back

=back

=cut

sub schedule_remove{
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars qw( schedule-name 7-Mode-host ),
        },
    );
    my $obj = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_schedule_remove_smtt_cli',
                interface => 'CLI',
                set       => 'SMTT',
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $obj;
} ## end sub schedule_remove

sub _schedule_remove_smtt_cli{
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my %target;
    my $response;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%target,
        copy   => [qw (session-name schedule-name 7-Mode-host ) ]
    );
    $response = $apiset->transition_schedule_remove(%target);
    $Log->exit() if $may_exit;
    return $response;
} ## end sub _schedule_remove_smtt_cli

=head2 precheck

    $resp = $trans_obj->precheck(
        "seven-mode-only" => $bool,
    );

(Class or instance method)
This method is used to run precheck on the transition session .
"Returns the response object."

"Supports CLI, GUI".

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< session-name => $name >>

(Required for class method, Not Applicable for instance method)
Name of the transition session .

=item C<< seven-mode-only => $bool >>

(Optional) It is a boolean value of 0 or 1.
When 1 it means that only 7mode precheck will be informed . Default value is 0 .

=back

=back

=cut

sub precheck {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars qw(
                seven-mode-only wait_for_operation parsed-output run-background
                group-name
            ),
        }
    );
    my $obj= $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_precheck_smtt_cli',
                interface => 'CLI',
                set       => 'SMTT',
            },
            {   method    => '_precheck_smtt_gui',
                interface => 'GUI',
                set       => 'SMTT',
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $obj;
} ## end sub precheck

sub _precheck_smtt_cli {
    $Log->enter() if $may_enter;
    my
    ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my %target;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%target,
        copy    => [qw( seven-mode-only session-name )]
    );
    my $response = $apiset->transition_precheck(%target);
    $Log->exit() if $may_exit;
    return $response;
} ## end sub _precheck_smtt_cli

sub _precheck_smtt_gui {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my %target;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%target,
        copy    => [qw( seven-mode-only session-name wait_for_operation group-name 
            parsed-output run-background )
        ]
    );
    my $response = $apiset->transition_precheck(%target);
    $Log->exit() if $may_exit;
    return $response;
} ## end sub _precheck_smtt_cli

=head2 start

    $resp = $trans_obj->start(
        "no-precheck" => $bool,
    );

(Class or instance method) This method is used to run start on the transition session .
"Returns the response object."

"Supports CLI, GUI".

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< session-name => $name >>

(Required for class method, Not Applicable for instance method)
Name of the transition session .

=item C<< no-precheck => $bool >>

(Optional) It is a boolean value of 0 or 1.
When 1 it means that no prechecks will be performed as a part of start operation
. Default value is 0 .

=back

=back

=cut

sub start {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars qw(
                no-precheck wait_for_operation parsed-output run-background 
                group-name
            ),
        }
    );
    my $obj = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_start_smtt_cli',
                interface => 'CLI',
                set       => 'SMTT',
            },
            {   method    => '_start_smtt_gui',
                interface => 'GUI',
                set       => 'SMTT',
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $obj;
} ## end sub start

sub _start_smtt_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my %target;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%target,
        copy   => [qw( no-precheck session-name )]
    );
    my $response = $apiset->transition_start(%target);
    $Log->exit() if $may_exit;
    return $response;
} ## end sub _start_smtt_cli

sub _start_smtt_gui {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my %target;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%target,
        copy   => [qw( no-precheck session-name wait_for_operation parsed-output 
            run-background group-name)
        ]
    );
    my $response = $apiset->transition_start(%target);
    $Log->exit() if $may_exit;
    return $response;
} ## end sub _start_smtt_gui

=head2 update

    $resp = $trans_obj->update(
                "available-transfers-percentage"   => $available-transfers-percentage,
                "session-snapmirror-throttle"      => $throttle,
            );
(Class or instance method)
This method is used to run update on the transition session.
"Returns the response object."

"Supports CLI, GUI".

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< session-name => $name >>

(Required for class method, Not Applicable for instance method)
Name of the transition session .

=item C<< available-transfers-percentage => $available-transfers-percentage >>

(Required)

=item C<< session-snapmirror-throttle => $throttle >>

(Optional)

=back

=back

=cut

sub update{
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars qw(
                available-transfers-percentage session-snapmirror-throttle
                max-cap-concurrent-transfers group-name parsed-output
                run-background
            ),
        },
    );
    my $obj = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_update_smtt_cli',
                interface => 'CLI',
                set       => 'SMTT',
            },
            {   method    => '_update_smtt_gui',
                interface => 'GUI',
                set       => 'SMTT',
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $obj;
} ## end sub update

sub _update_smtt_cli{
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my %target;
    my $response;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%target,
        copy   => [qw (
            session-name available-transfers-percentage session-snapmirror-throttle
            max-cap-concurrent-transfers) ]
    );
    $response = $apiset->transition_update(%target);
    $Log->exit() if $may_exit;
    return $response;
} ## end sub _update_smtt_cli

sub _update_smtt_gui{
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my %target;
    my $response;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%target,
        copy   => [qw (
            session-name available-transfers-percentage session-snapmirror-throttle
            max-cap-concurrent-transfers group-name run-background parsed-output)
        ]
    );
    $response = $apiset->transition_update(%target);
    $Log->exit() if $may_exit;
    return $response;
} ## end sub _update_smtt_gui

=head2 pause

    $resp = $trans_obj->pause();

(Class or instance method) This method is used to run pause on the transition session .
"Returns the response object."

"Supports CLI, GUI".

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< session-name => $name >>

(Required for class method, Not Applicable for instance method)
Name of the transition session .

=back

=back

=cut

sub pause {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars qw( wait_for_operation parsed-output run-background group-name),
        },
    );
    my $obj = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_pause_smtt_cli',
                interface => 'CLI',
                set       => 'SMTT',
            },
            {   method    => '_pause_smtt_gui',
                interface => 'GUI',
                set       => 'SMTT',
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $obj;
} ## end sub pause

sub _pause_smtt_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my %target;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%target,
        copy   => [qw(session-name)]
    );
    my $response = $apiset->transition_pause(%target);
    $Log->exit() if $may_exit;
    return $response;
} ## end sub _pause_smtt_cli

sub _pause_smtt_gui {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my %target;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%target,
        copy   => [qw( session-name wait_for_operation parsed-output run-background 
            group-name)
        ]
    );
    my $response = $apiset->transition_pause(%target);
    $Log->exit() if $may_exit;
    return $response;
} ## end sub _pause_smtt_gui

=head2 resume

    $resp = $trans_obj->resume();

(Class or instance method) This method is used to run resume on the transition session .
"Returns the response object."

"Supports CLI, GUI".

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< session-name => $name >>

(Required for class method, Not Applicable for instance method)
Name of the transition session .

=back

=back

=cut

sub resume {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars qw( wait_for_operation parsed-output run-background group-name),
        },
    );
    my $obj = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_resume_smtt_cli',
                interface => 'CLI',
                set       => 'SMTT',
            },
            {   method    => '_resume_smtt_gui',
                interface => 'GUI',
                set       => 'SMTT',
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $obj;
} ## end sub resume

sub _resume_smtt_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my %target;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%target,
        copy   => [qw(session-name)]
    );
    my $response = $apiset->transition_resume(%target);
    $Log->exit() if $may_exit;
    return $response;
} ## end sub _resume_smtt_cli

sub _resume_smtt_gui {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my %target;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%target,
        copy   => [qw( session-name wait_for_operation parsed-output run-background
            group-name)
        ]
    );
    my $response = $apiset->transition_resume(%target);
    $Log->exit() if $may_exit;
    return $response;
} ## end sub _resume_smtt_gui

=head2 precutover

    $resp = $trans_obj->precutover(
        "mode"       => $mode,
        "available-transfers-percentage"    => $available-transfers-percentage,
    );
(Class or instance method)
This method is used to run precutover on the transition session.
"Returns the response object."

"Supports CLI, GUI".

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< session-name => $name >>

(Required for class method, Not Applicable for instance method)
Name of the transition session .

=item C<< mode => $mode >>

(Required)

=item C<< available-transfers-percentage => $available-transfers-percentage >>

(Required)

=item C<< snapmirror-throttle => $throttle >>

(Optional)

=item C<< ad-server-host-name => $ad_server_host_name >>

(Optional)

=item C<< ad-sever-ip-address => $ad_server_ip_address >>

(Optional)

=item C<< ad-server-admin-user-name => $ad_server_admin_user_name >>

(Optional)

=item C<< ad-server-password => $ad_server_password >>

(Optional)

=item C<< kdc-server-host-name => $kdc_server_host_name >>

(Optional)

=item C<< kdc-server-admin-user-name => $kdc_server_admin_user_name >>

(Optional)

=item C<< kdc-server-password => $kdc_server_password >>

(Optional)

=back

=back

=cut

sub precutover{
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars qw(
                mode available-transfers-percentage snapmirror-throttle ad-server-host-name
                ad-server-ip-address ad-server-admin-user-name
                ad-server-password kdc-server-host-name
                kdc-server-admin-user-name kdc-server-password
                wait_for_operation max-cap-concurrent-transfers
                parsed-output run-background group-name
            ),
        },
    );
    my $obj = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_precutover_smtt_cli',
                interface => 'CLI',
                set       => 'SMTT',
            },
            {   method    => '_precutover_smtt_gui',
                interface => 'GUI',
                set       => 'SMTT',
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $obj;
} ## end sub precutover

sub _precutover_smtt_cli{
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my %target;
    my $response;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%target,
        copy   => [qw (
            session-name mode available-transfers-percentage snapmirror-throttle ad-server-host-name
            ad-server-ip-address ad-server-admin-user-name ad-server-password
            kdc-server-host-name kdc-server-admin-user-name kdc-server-password
            max-cap-concurrent-transfers) ]
    );
    $response = $apiset->transition_precutover(%target);
    $Log->exit() if $may_exit;
    return $response;
} ## end sub _precutover_smtt_cli

sub _precutover_smtt_gui{
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my %target;
    my $response;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%target,
        copy   => [qw (
            session-name mode available-transfers-percentage snapmirror-throttle ad-server-host-name
            ad-server-ip-address ad-server-admin-user-name ad-server-password
            kdc-server-host-name kdc-server-admin-user-name kdc-server-password
            wait_for_operation max-cap-concurrent-transfers parsed-output 
            run-background group-name) 
        ]
    );
    $response = $apiset->transition_precutover(%target);
    $Log->exit() if $may_exit;
    return $response;
} ## end sub _precutover_smtt_gui

=head2 resync

$resp = $trans_obj->resync();

(Class or instance method)
This method is used to run resync on the transition session.
"Returns the response object."

"Supports CLI, GUI".

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< session-name => $name >>

(Required for class method, Not Applicable for instance method)
Name of the transition session .

=back

=back

=cut

sub resync{
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars qw( wait_for_operation parsed-output run-background 
                group-name 
            ),
        },
    );
    my $obj = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_resync_smtt_cli',
                interface => 'CLI',
                set       => 'SMTT',
            },
            {   method    => '_resync_smtt_gui',
                interface => 'GUI',
                set       => 'SMTT',
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $obj;
} ## end sub resync

sub _resync_smtt_cli{
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my %target;
    my $response;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%target,
        copy   => [qw (
            session-name ) ]
    );
    $response = $apiset->transition_resync(%target);
    $Log->exit() if $may_exit;
    return $response;
} ## end sub _resync_smtt_cli

sub _resync_smtt_gui{
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my %target;
    my $response;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%target,
        copy   => [qw (
            session-name wait_for_operation parsed-output run-background 
            group-name ) 
        ]
    );
    $response = $apiset->transition_resync(%target);
    $Log->exit() if $may_exit;
    return $response;
} ## end sub _resync_smtt_gui

=head2 cutover

    $resp = $trans_obj->cutover(
        "available-transfers-percentage"    => $available-transfers-percentage,
    );
(Class or instance method)
This method is used to run cutover on the transition session.
"Returns the response object."

"Supports CLI, GUI".

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< session-name => $name >>

(Required for class method, Not Applicable for instance method)
Name of the transition session .

=item C<< available-transfers-percentage => $available-transfers-percentage >>

(Required)

=item C<< session-snapmirror-throttle => $throttle >>

=item C<< offline-source-volumes => $bool >>
Specify either true or false .

(Optional)

=back

=back

=cut

sub cutover{
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars qw(
                available-transfers-percentage session-snapmirror-throttle wait_for_operation
                max-cap-concurrent-transfers offline-source-volumes parsed-output
                run-background group-name
            )
        }
    );
    my $obj = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_cutover_smtt_cli',
                interface => 'CLI',
                set       => 'SMTT',
            },
            {   method    => '_cutover_smtt_gui',
                interface => 'GUI',
                set       => 'SMTT',
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $obj;
} ## end sub cutover

sub _cutover_smtt_cli{
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my %target;
    my $response;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%target,
        copy   => [qw (session-name available-transfers-percentage session-snapmirror-throttle
                   max-cap-concurrent-transfers offline-source-volumes) ]
    );
    $response = $apiset->transition_cutover(%target);
    $Log->exit() if $may_exit;
    return $response;
} ## end sub _cutover_smtt_cli

sub _cutover_smtt_gui{
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my %target;
    my $response;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%target,
        copy   => [qw (session-name available-transfers-percentage session-snapmirror-throttle
            wait_for_operation max-cap-concurrent-transfers offline-source-volumes
            parsed-output run-background group-name)
        ]
    );
    $response = $apiset->transition_cutover(%target);
    $Log->exit() if $may_exit;
    return $response;
} ## end sub _cutover_smtt_gui

=head2 state_machine_status

    $resp = $trans_obj->state_machine_status(
        "operation"       => "precheck",
        "group-name"      => $group_name,
    );
             Or
    $resp = NACL::C::Client::Transition->state_machine_status(
        command_interface => $Client,
        "project-name"    => $project_name,
        "operation"       => "precheck",
        "group-name"      => $group_name,
    );
(Class or instance method)
This method is used to return status of GUI transition operation
"Returns the response object."

"Supports GUI".

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< projetct-name => $name >>

(Required for class method, Not Applicable for instance method)
Name of the transition project .

=item C<< group-name => $group_name >>

(Optional)

=item C<< operation => $operation >>
operation can have value as precheck,start,incremental, precutover,cutover,
apply_svm_config,export,cabling,import,commit,rollback_precheck,rollback_verify_7mode
(Required)

=back

=back

=cut

sub state_machine_status{
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            "operation"   => { type => SCALAR, optional => 0 },
            _optional_scalars qw( group-name
            ),
        },
    );
    my $obj = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_state_machine_status_smtt_gui',
                interface => 'GUI',
                set       => 'SMTT',
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $obj;
} ## end sub state_machine_status

sub _state_machine_status_smtt_gui{
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my %target;
    my $response;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%target,
        copy   => [ qw ( project-name operation group-name) 
        ]
    );
    $response = $apiset->transition_state_machine_status(%target);
    $Log->exit() if $may_exit;
    return $response;
} ## end sub _state_machine_status_smtt_gui

=head2 abort

    $resp = $trans_obj->abort();

(Class or instance method)
This method is used to run abort on the transition session .
"Returns the response object."

"Supports CLI, GUI".

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< session-name => $name >>

(Required for class method, Not Applicable for instance method)
Name of the transition session .

=back

=back

=cut

sub abort {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
	additional_spec => {
		 _optional_scalars qw(
			wait_for_operation parsed-output run-background group-name
                 )
	}
     );
    my $obj = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_abort_smtt_cli',
                interface => 'CLI',
                set       => 'SMTT',
            },
            {   method    => '_abort_smtt_gui',
                interface => 'GUI',
                set       => 'SMTT',
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $obj;
} ## end sub abort

sub _abort_smtt_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my %target;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%target,
        copy   => [qw(session-name)]
    );
    my $response = $apiset->transition_abort(%target);
    $Log->exit() if $may_exit;
    return $response;
} ## end sub _abort_smtt_cli

sub _abort_smtt_gui {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my %target;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%target,
        copy   => [qw( session-name wait_for_operation parsed-output 
            run-background group-name)
        ]
    );
    my $response = $apiset->transition_abort(%target);
    $Log->exit() if $may_exit;
    return $response;
} ## end sub _abort_smtt_gui

=head2 delete

    $resp = $trans_obj->delete();

(Class or instance method)
This method is used to run delete on the transition session .
"Returns the response object."

"Supports CLI".

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< session-name => $name >>

(Required for class method, Not Applicable for instance method)
Name of the transition session .

=back

=back

=cut

sub delete {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
    );
    my $obj = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_delete_smtt_cli',
                interface => 'CLI',
                set       => 'SMTT',
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $obj;
} ## end sub delete

sub _delete_smtt_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my %target;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%target,
        copy   => [qw(session-name)]
    );
    my $response = $apiset->transition_delete(%target);
    $Log->exit() if $may_exit;
    return $response;
} ## end sub _delete_smtt_cli

=head2 add_cluster_info

    $res = NACL::C::Client::Transition->add_cluster_info(
        "command_interface"         => $HostWindows,
        "session-name"              => $SESSION,
        "host-name"                 => $host_C,
        "svm-name"                  => $svm_c,
        "aggregate-name"            => $aggr_c,
    );
(Class or instance method) This method is used to add
cluster information to the session.
"Returns the response object."

"Supports CLI".

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< session-name => $session_name >>

(Required) Name of the transition session .

=item C<< host-name => $host_c >>

(Required) Hostname or IP address of Cmode .

=item C<< svm-name => $svm_c >>

(Required) Name of svm on C mode System .

=item C<< aggregate-name => $aggr_c >>

(Optional) Name of aggregate on C mode System .

=back

=back

=cut
sub add_cluster_info {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            "host-name"     => { type => SCALAR },
            "svm-name"      => { type => SCALAR },
            _optional_scalars qw( aggregate-name ),
        },
    );
    my $obj=$pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_add_cluster_info_smtt_cli',
                interface => 'CLI',
                set       => 'SMTT',
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $obj;
} ## end sub add_cluster_info

sub _add_cluster_info_smtt_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my %target;
    $pkg_or_obj->_hash_copy(
      source => \%opts,
        target => \%target,
        copy    => [qw( session-name host-name svm-name aggregate-name )]
    );
    my $response = $apiset->transition_add_cluster_info(%target);
    $Log->exit() if $may_exit;
    return $response;
} ## end sub _add_cluster_info_smtt_cli

=head2 modify

    $res = NACL::C::Client::Transition->modify(
        "command_interface"         => $HostWindows,
        "session-name"              => $SESSION,
        "data-copy-ipaddress"       => $Host7Mode->{MGMT_IP},
        "multipathing-ipaddress"    => $MULTIPATH_IP,
        "svm-name"                  => $VSERVER,
        "c-mode-host-name"          => $HostCMode->{MGMT_IP},
        "aggregate-name"            => $AGGR,
    );
(Class or instance method) This method is used to add
cluster information to the session.
"Returns the response object."

"Supports CLI".

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< session-name => $session_name >>

(Required) Name of the transition session .

=item C<< data-copy-ipaddress => $host_c >>

(Optional) Data copy IP address

=item C<< multipathing-ipaddress => $host_c >>

(Optional) Multipathing IP address .

=item C<< svm-name => $svm_c >>

(Optional) Name of the vserver to which volumes will be transitioned .

=item C<< c-mode-host-name => $host_C >>

(Optional) Hostname or IP address of Cmode .

=item C<< aggregate-name => $aggr_c >>

(Optional) Vserver root volume aggregate name on Cmode system .

=back

=back

=cut
sub modify {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars qw( data-copy-ipaddress multipathing-ipaddress
                                    svm-name c-mode-host-name aggregate-name
            ),
        },
    );
    my $obj=$pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_modify_smtt_cli',
                interface => 'CLI',
                set       => 'SMTT',
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $obj;
} ## end sub modify

sub _modify_smtt_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my %target;
    $pkg_or_obj->_hash_copy(
      source => \%opts,
        target => \%target,
        copy    => [qw( session-name data-copy-ipaddress multipathing-ipaddress
                        svm-name c-mode-host-name aggregate-name
        )]
    );
    my $response = $apiset->transition_modify(%target);
    $Log->exit() if $may_exit;
    return $response;
} ## end sub _modify_smtt_cli

=head2 property_set

    $comp_resp = NACL::C::Client::Transition->property_set(
        "command_interface"         => $HostWindows,
        "session-name"              => "ses1",
        "property-name"             => "cluster-audit-saveas-path",
        "property-value"            => "/vol/$VOL_NAME",
    );
(Class or instance method) This method is used to add
cluster information to the session.
"Returns the response object."

"Supports CLI".

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< session-name => $session_name >>

(Required) Name of the transition session .

=item C<< property-name => "cluster-audit-saveas-path" >>

(Required) Name of the property to set value .

=item C<< property-value => "/vol/vol1" >>

(Required) Value of property .

=back

=back

=cut
sub property_set {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            "property-name"         => { type => SCALAR },
            "property-value"        => { type => SCALAR },
        },
    );
    my $obj=$pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_property_set_smtt_cli',
                interface => 'CLI',
                set       => 'SMTT',
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $obj;
} ## end sub property_set

sub _property_set_smtt_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my %target;
    $pkg_or_obj->_hash_copy(
      source => \%opts,
        target => \%target,
        copy    => [qw( session-name property-name property-value
        )]
    );
    my $response = $apiset->transition_property_set(%target);
    $Log->exit() if $may_exit;
    return $response;
} ## end sub _property_set_smtt_cli

=head2 property_get

    $comp_resp = NACL::C::Client::Transition->property_get(
        "command_interface"         => $HostWindows,
        "session-name"              => "ses1",
        "property-name"             => "cluster-audit-saveas-path",
    );
(Class or instance method) This method is used to add
cluster information to the session.
"Returns the response object."

"Supports CLI".

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< session-name => $session_name >>

(Required) Name of the transition session .

=item C<< property-name => "cluster-audit-saveas-path" >>

(Required) Name of the property to get value .

=back

=back

=cut
sub property_get {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            "property-name"         => { type => SCALAR },
        },
    );
    my $obj=$pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_property_get_smtt_cli',
                interface => 'CLI',
                set       => 'SMTT',
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $obj;
} ## end sub property_get

sub _property_get_smtt_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my %target;
    $pkg_or_obj->_hash_copy(
      source => \%opts,
        target => \%target,
        copy    => [qw( session-name property-name )]
    );
    my $response = $apiset->transition_property_get(%target);
    $Log->exit() if $may_exit;
    return $response;
} ## end sub _property_get_smtt_cli

=head2 property_reset

    $comp_resp = NACL::C::Client::Transition->property_reset(
        "command_interface"         => $HostWindows,
        "session-name"              => "ses1",
        "property-name"             => "cluster-audit-saveas-path",
    );
(Class or instance method) This method is used to add
cluster information to the session.
"Returns the response object."

"Supports CLI".

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< session-name => $session_name >>

(Required) Name of the transition session .

=item C<< property-name => "cluster-audit-saveas-path" >>

(Required) Name of the property to reset value .

=back

=back

=cut
sub property_reset {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            "property-name"         => { type => SCALAR },
        },
    );
    my $obj=$pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_property_reset_smtt_cli',
                interface => 'CLI',
                set       => 'SMTT',
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $obj;
} ## end sub property_reset

sub _property_reset_smtt_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my %target;
    $pkg_or_obj->_hash_copy(
      source => \%opts,
        target => \%target,
        copy    => [qw( session-name property-name )]
    );
    my $response = $apiset->transition_property_reset(%target);
    $Log->exit() if $may_exit;
    return $response;
} ## end sub _property_reset_smtt_cli


=head2 remove_primary_seven_mode_system

    $res = NACL::C::Client::Transition->remove_primary_seven_mode_system(
        "command_interface"         => $HostWindows,
        "session-name"              => $SESSION,
        "source-host"               => $Host7Mode->{MGMT_IP},
    );
(Class or instance method) This method is used to add
cluster information to the session.
"Returns the response object."

"Supports CLI".

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< session-name => $session_name >>

(Required) Name of the transition session .

=item C<< source-host => $host_7 >>

(Required) Hostname or IP address of 7mode .

=back

=back

=cut
sub remove_primary_seven_mode_system {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            "source-host"           => { type => SCALAR },
        },
    );
    my $obj=$pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_remove_primary_seven_mode_system_smtt_cli',
                interface => 'CLI',
                set       => 'SMTT',
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $obj;
} ## end sub remove_primary_seven_mode_system

sub _remove_primary_seven_mode_system_smtt_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my %target;
    $pkg_or_obj->_hash_copy(
      source => \%opts,
        target => \%target,
        copy    => [qw( session-name source-host )]
    );
    my $response = $apiset->transition_remove_primary_seven_mode_system(%target);
    $Log->exit() if $may_exit;
    return $response;
} ## end sub _remove_primary_seven_mode_system_smtt_cli

=head2 remove_secondary_cluster_mode_system

    $comp_resp = NACL::C::Client::Transition->
                                        remove_secondary_cluster_mode_system(
        "command_interface"         => $HostWindows,
        "session-name"              => "ses_prima",
        "secondary-svm"             => $VSERVER
    );
(Class or instance method) This method is used to add
cluster information to the session.
"Returns the response object."

"Supports CLI".

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< session-name => $session_name >>

(Required) Name of the transition session .

=item C<< secondary-svm => $VSERVER >>

(Required) Name of secondary svm on CMode system .

=back

=back

=cut
sub remove_secondary_cluster_mode_system {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            "secondary-svm"         => { type => SCALAR },
        },
    );
    my $obj=$pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_remove_secondary_cluster_mode_system_smtt_cli',
                interface => 'CLI',
                set       => 'SMTT',
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $obj;
} ## end sub remove_secondary_cluster_mode_system

sub _remove_secondary_cluster_mode_system_smtt_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my %target;
    $pkg_or_obj->_hash_copy(
      source => \%opts,
        target => \%target,
        copy    => [qw( session-name secondary-svm )]
    );
    my $response = $apiset->transition_remove_secondary_cluster_mode_system(%target);
    $Log->exit() if $may_exit;
    return $response;
} ## end sub _remove_secondary_cluster_mode_system_smtt_cli

=head2 bundle_tool_logs

    $res = NACL::C::Client::Transition->bundle_tool_logs(
        "command_interface"         => $HostWindows,
    );
(Class or instance method) This method is used to add
cluster information to the session.
"Returns the response object."

"Supports CLI".

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=back

=back

=cut
sub bundle_tool_logs {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => { },
        ignore_primary_keys => 1,
    );
    my $obj=$pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_bundle_tool_logs_smtt_cli',
                interface => 'CLI',
                set       => 'SMTT',
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $obj;
} ## end sub bundle_tool_logs

sub _bundle_tool_logs_smtt_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my %target;
    $pkg_or_obj->_hash_copy(
      source => \%opts,
        target => \%target,
        copy    => [ ]
    );
    my $response = $apiset->transition_bundle_tool_logs(%target);
    $Log->exit() if $may_exit;
    return $response;
} ## end sub _bundle_tool_logs_smtt_cli

=head2 jobs

    $res = NACL::C::Client::Transition->jobs(
        "command_interface"         => $HostWindows,
        "session-name"              => $SESSION,
    );
(Class or instance method) This method is used to add
cluster information to the session.
"Returns the response object."

"Supports CLI".

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< session-name => $session_name >>

(Required) Name of the transition session .

=item C<< operation-name => "start" >>

(Optional) Name of the operation .

=back

=back

=cut
sub jobs {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars qw( operation-name ),
        },
    );
    my $obj=$pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_jobs_smtt_cli',
                interface => 'CLI',
                set       => 'SMTT',
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $obj;
} ## end sub jobs

sub _jobs_smtt_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my %target;
    $pkg_or_obj->_hash_copy(
      source => \%opts,
        target => \%target,
        copy    => [qw( session-name operation-name )]
    );
    my $response = $apiset->transition_jobs(%target);
    $Log->exit() if $may_exit;
    return $response;
} ## end sub _jobs_smtt_cli

=head2 show_primary_seven_mode_systems

    $comp_resp = NACL::C::Client::Transition->show_primary_seven_mode_systems(
            "command_interface"         => $HostWindows,
            "session-name"              => "ses_secon",
        );
(Class or instance method) This method is used to add
cluster information to the session.
"Returns the response object."

"Supports CLI".

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< session-name => $session_name >>

(Required) Name of the transition session .

=back

=back

=cut
sub show_primary_seven_mode_systems {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
    );
    my $obj=$pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_show_primary_seven_mode_systems_smtt_cli',
                interface => 'CLI',
                set       => 'SMTT',
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $obj;
} ## end sub show_primary_seven_mode_systems

sub _show_primary_seven_mode_systems_smtt_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my %target;
    $pkg_or_obj->_hash_copy(
      source => \%opts,
        target => \%target,
        copy    => [qw( session-name )]
    );
    my $response = $apiset->transition_show_primary_seven_mode_systems(%target);
    $Log->exit() if $may_exit;
    return $response;
} ## end sub _show_primary_seven_mode_systems_smtt_cli

=head2 show_secondary_cluster_mode_systems

    $comp_resp = NACL::C::Client::Transition->
                                            show_secondary_cluster_mode_systems(
            "command_interface"         => $HostWindows,
            "session-name"              => "ses_prima",
        );
(Class or instance method) This method is used to add
cluster information to the session.
"Returns the response object."

"Supports CLI".

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< session-name => $session_name >>

(Required) Name of the transition session .

=back

=back

=cut
sub show_secondary_cluster_mode_systems {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
    );
    my $obj=$pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_show_secondary_cluster_mode_systems_smtt_cli',
                interface => 'CLI',
                set       => 'SMTT',
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $obj;
} ## end sub show_secondary_cluster_mode_systems

sub _show_secondary_cluster_mode_systems_smtt_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my %target;
    $pkg_or_obj->_hash_copy(
      source => \%opts,
        target => \%target,
        copy    => [qw( session-name )]
    );
    my $response = $apiset->transition_show_secondary_cluster_mode_systems(%target);
    $Log->exit() if $may_exit;
    return $response;
} ## end sub _show_secondary_cluster_mode_systems_smtt_cli

=head2 job_status

    $res = NACL::C::Client::Transition->job_status(
        "command_interface"         => $HostWindows,
        "job-id" => $job_id,
    );
(Class or instance method) This method is used to add
cluster information to the session.
"Returns the response object."

"Supports CLI".

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< job-id => $job-id >>

(Required) Value of property .

=back

=back

=cut
sub job_status {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            "job-id"         => { type => SCALAR },
        },
        ignore_primary_keys => 1,
    );
    my $obj=$pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_job_status_smtt_cli',
                interface => 'CLI',
                set       => 'SMTT',
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $obj;
} ## end sub job_status

sub _job_status_smtt_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my %target;
    $pkg_or_obj->_hash_copy(
      source => \%opts,
        target => \%target,
        copy    => [qw( job-id )]
    );
    my $response = $apiset->transition_job_status(%target);
    $Log->exit() if $may_exit;
    return $response;
} ## end sub _job_status_smtt_cli

=head2 job_results

    $res = NACL::C::Client::Transition->job_results(
        "command_interface"         => $HostWindows,
        "job-id"                    => $job_id,
    );
(Class or instance method) This method is used to add
cluster information to the session.
"Returns the response object."

"Supports CLI".

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< job-id => $job_id >>

(Required) Unique job id of the job .

=back

=back

=cut
sub job_results {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            "job-id"         => { type => SCALAR },
        },
        ignore_primary_keys => 1,
    );
    my $obj=$pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_job_results_smtt_cli',
                interface => 'CLI',
                set       => 'SMTT',
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $obj;
} ## end sub job_results

sub _job_results_smtt_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my %target;
    $pkg_or_obj->_hash_copy(
      source => \%opts,
        target => \%target,
        copy    => [qw( job-id )]
    );
    my $response = $apiset->transition_job_results(%target);
    $Log->exit() if $may_exit;
    return $response;
} ## end sub _job_results_smtt_cli

=head2 qtreepair_add

    $resp = $trans_obj->qtreepair_add(
        "volume-name-7-mode"  => vol1,
        "qtree-name-7-mode"   => [qw(qtree1 qtree2 )],
        "volume-name-c-mode"  => [qw(vol_compo1 vol_compo2)],
        "qtree-name-c-mode"   => [qw(qtree_comp1 qtree_comp2)],
    );

                OR

    $resp = $trans_obj->qtreepair_add(
        "volume-name-7-mode"  => $vol_name,
        "qtree-name-7-mode"   => \@qtree_7mode,
        "volume-name-c-mode"  => \@vol_cmode,
        "qtree-name-c-mode"   => \@qtree_cmode,
    );


(Class or instance method) This method is used to add qtreepairs to the session .
"Returns the response object."

"Supports CLI".

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< session-name => $name >>

(Required for class method, Not Applicable for instance method)
Name of the transition session .

=item C<< volume-name-7-mode => $7_vol_arr_ref >>

(Required) Its a reference to an array which contains 7mode volumes.

=item C<< qtree-name-7-mode => $7_qtree_arr_ref >>

(Required) Its a reference to an array which contains 7mode qtree.

=item C<< volume-name-c-mode => $C_vol_arr_ref >>

(Optional) Its a reference to an array which contains Cmode volumes.

=item C<< qtree-name-c-mode => $C_qtree_arr_ref >>

(Optional) Its a reference to an array which contains Cmode volumes.

=back

=back

=cut

sub qtreepair_add {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            "volume-name-7-mode"    => { type => SCALAR,   optional => 0 },
            "qtree-name-7-mode"     => { type => ARRAYREF, optional => 0 },
            "volume-name-c-mode"    => { type => ARRAYREF, optional => 1 },
            "qtree-name-c-mode"     => { type => ARRAYREF, optional => 1 },
        },
    );
    my @obj = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_qtreepair_add_smtt_cli',
                interface => 'CLI',
                set       => 'SMTT',
            },
        ],
    );
    $Log->exit() if $may_exit;
    return wantarray ? @obj : $obj[0];
} ## end sub qreepair_add

sub _qtreepair_add_smtt_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my $volume_name_c_mode = delete $opts{"volume-name-c-mode"};
    my $qtree_name_7_mode  = delete $opts{"qtree-name-7-mode"};
    my $qtree_name_c_mode  = delete $opts{"qtree-name-c-mode"};
    my (%target, $response, @response_objs);
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%target,
        copy   => [qw (session-name volume-name-7-mode ) ]
    );
    foreach (0..@$qtree_name_7_mode-1){
        $target{"qtree-name-7-mode"}=$qtree_name_7_mode->[$_];
        $target{"volume-name-c-mode"}=$volume_name_c_mode->[$_] if ( $volume_name_c_mode->[$_] );
        $target{"qtree-name-c-mode"}=$qtree_name_c_mode->[$_] if ( $qtree_name_c_mode->[$_] );
        $response = $apiset->transition_qtreepair_add(%target);
        push @response_objs,$response;
    }
    $Log->exit() if $may_exit;
    return @response_objs;
} ## end sub _qtreeepair_add_smtt_cli

=head2 qtreepair_modify

    $resp = $trans_obj->qtreepair_modify(
        "volume-name-7-mode"  => vol1,
        "qtree-name-7-mode"   => [qw(qtree1 qtree2 )],
        "volume-name-c-mode"  => [qw(vol_compo1 vol_compo2)],
        "qtree-name-c-mode"   => [qw(qtree_comp1 qtree_comp2)],
    );

                OR

    $resp = $trans_obj->qtreepair_modify(
        "volume-name-7-mode"  => $vol_name,
        "qtree-name-7-mode"   => \@qtree_7mode,
        "volume-name-c-mode"  => \@vol_cmode,
        "qtree-name-c-mode"   => \@qtree_cmode,
    );


(Class or instance method) This method is used to modify qtreepairs of the session .
"Returns the response object."

"Supports CLI".

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< session-name => $name >>

(Required for class method, Not Applicable for instance method)
Name of the transition session .

=item C<< volume-name-7-mode => $7_vol_arr_ref >>

(Required) Its a reference to an array which contains 7mode volumes.

=item C<< qtree-name-7-mode => $7_qtree_arr_ref >>

(Required) Its a reference to an array which contains 7mode qtree.

=item C<< volume-name-c-mode => $C_vol_arr_ref >>

(Optional) Its a reference to an array which contains Cmode volumes.

=item C<< qtree-name-c-mode => $C_qtree_arr_ref >>

(Optional) Its a reference to an array which contains Cmode volumes.

=back

=back

=cut

sub qtreepair_modify {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            "volume-name-7-mode"    => { type => SCALAR,   optional => 0 },
            "qtree-name-7-mode"     => { type => ARRAYREF, optional => 0 },
            "volume-name-c-mode"    => { type => ARRAYREF, optional => 1 },
            "qtree-name-c-mode"     => { type => ARRAYREF, optional => 1 },
        },
    );
    my @obj = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_qtreepair_modify_smtt_cli',
                interface => 'CLI',
                set       => 'SMTT',
            },
        ],
    );
    $Log->exit() if $may_exit;
    return wantarray ? @obj : $obj[0];
} ## end sub qtreepair_modify

sub _qtreepair_modify_smtt_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my $volume_name_c_mode = delete $opts{"volume-name-c-mode"};
    my $qtree_name_7_mode  = delete $opts{"qtree-name-7-mode"};
    my $qtree_name_c_mode  = delete $opts{"qtree-name-c-mode"};
    my (%target, $response, @response_objs);
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%target,
        copy   => [qw (session-name volume-name-7-mode ) ]
    );
    foreach (0..@$qtree_name_7_mode-1){
        $target{"qtree-name-7-mode"}=$qtree_name_7_mode->[$_];
        $target{"volume-name-c-mode"}=$volume_name_c_mode->[$_] if ( $volume_name_c_mode->[$_] );
        $target{"qtree-name-c-mode"}=$qtree_name_c_mode->[$_] if ( $qtree_name_c_mode->[$_] );
        $response = $apiset->transition_qtreepair_modify(%target);
        push @response_objs,$response;
    }
    $Log->exit() if $may_exit;
    return @response_objs;
} ## end sub _qtreepair_modify_smtt_cli

=head2 qtreepair_remove

    $resp = $trans_obj->qtreepair_remove(
        "volume-name-7-mode"  => vol1,
        "qtree-name-7-mode"   => [qw(qtree1 qtree2 )],
    );

                OR

    $resp = $trans_obj->qtreepair_remove(
        "volume-name-7-mode"  => $vol_name,
        "qtree-name-7-mode"   => \@qtree_7mode,
    );


(Class or instance method) This method is used to remove qtreepairs from the session .
"Returns the response object."

"Supports CLI".

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< session-name => $name >>

(Required for class method, Not Applicable for instance method)
Name of the transition session .

=item C<< volume-name-7-mode => $7_vol_arr_ref >>

(Required) Its a reference to an array which contains 7mode volumes.

=item C<< qtree-name-7-mode => $7_qtree_arr_ref >>

(Required) Its a reference to an array which contains 7mode qtree.

=back

=back

=cut

sub qtreepair_remove {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            "volume-name-7-mode"    => { type => SCALAR,   optional => 0 },
            "qtree-name-7-mode"     => { type => ARRAYREF, optional => 0 },
        },
    );
    my @obj = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_qtreepair_remove_smtt_cli',
                interface => 'CLI',
                set       => 'SMTT',
            },
        ],
    );
    $Log->exit() if $may_exit;
    return wantarray ? @obj : $obj[0];
} ## end sub qtreepair_remove

sub _qtreepair_remove_smtt_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my $qtree_name_7_mode  = delete $opts{"qtree-name-7-mode"};
    my (%target, $response, @response_objs);
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%target,
        copy   => [qw (session-name volume-name-7-mode ) ]
    );
    foreach (0..@$qtree_name_7_mode-1){
        $target{"qtree-name-7-mode"}=$qtree_name_7_mode->[$_];
        $response = $apiset->transition_qtreepair_remove(%target);
        push @response_objs,$response;
    }
    $Log->exit() if $may_exit;
    return @response_objs;
} ## end sub _qtreepair_remove_smtt_cli

=head2 stagingsystems_add

    $resp = $trans_obj->stagingsystems_add(
        "host-name"            => $host_name,
        "vfiler-name"          => $vfiler,
        "aggregates"           => $list_aggregates,
        "datacopy-ip-address"  => $data_ip,
        "multipath_ip_address" => $multipath_ip,
    );


(Class or instance method) This method is used to add staging system to the session .
"Returns the response object."

"Supports CLI".

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< session-name => $name >>

(Required for class method, Not Applicable for instance method)
Name of the transition session .

=item C<< host-name => $host_name >>

(Required) Hostname or IP address of Staging system.

=item C<< vfiler-name => $vfiler >>

(Required) Vfiler of Staging system

=item C<< aggregates => $aggr_list >>

(Required) List of staging system aggregates .

=item C<< datacopy-ip-address => $data_ip >>

(Required) Datacopy IP address.

=item C<< multipath-ip-address => $data_ip >>

(Optional) Multipath IP address.

=back

=back

=cut

sub stagingsystems_add {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            "host-name"            => { type => SCALAR,   optional => 1 },
            "vfiler-name"          => { type => SCALAR,   optional => 1 },
            "aggregates"           => { type => SCALAR,   optional => 1 },
            "datacopy-ip-address"  => { type => SCALAR,   optional => 1 },
            "multipath-ip-address" => { type => SCALAR,   optional => 1 },
        },
    );
    my $obj = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_stagingsystems_add_smtt_cli',
                interface => 'CLI',
                set       => 'SMTT',
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $obj;
} ## end sub stagingsystems_add

sub _stagingsystems_add_smtt_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my (%target, $response);
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%target,
        copy   => [qw (session-name host-name vfiler-name aggregates datacopy-ip-address
                   multipath-ip-address ) ]
    );
    $response = $apiset->transition_stagingsystems_add(%target);
    $Log->exit() if $may_exit;
    return $response;
} ## end sub _stagingsystems_add_smtt_cli

=head2 stagingsystems_modify

    $resp = $trans_obj->stagingsystems_modify(
        "host-name"            => $host_name,
        "vfiler-name"          => $vfiler,
        "aggregates"           => $list_aggregates,
        "datacopy-ip-address"  => $data_ip,
        "multipath_ip_address" => $multipath_ip,
    );

(Class or instance method) This method is used to modify staging system of the session .
"Returns the response object."

"Supports CLI".

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< session-name => $name >>

(Required for class method, Not Applicable for instance method)
Name of the transition session .

=item C<< host-name => $host_name >>

(Required) Hostname or IP address of Staging system.

=item C<< vfiler-name => $vfiler >>

(Optional) Vfiler of Staging system

=item C<< aggregates => $aggr_list >>

(Optional) List of staging system aggregates .

=item C<< datacopy-ip-address => $data_ip >>

(Optional) Datacopy IP address.

=item C<< multipath-ip-address => $data_ip >>

(Optional) Multipath IP address.

=back

=back

=cut

sub stagingsystems_modify {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            "host-name"            => { type => SCALAR,   optional => 1 },
            "vfiler-name"          => { type => SCALAR,   optional => 0 },
            "aggregates"           => { type => SCALAR,   optional => 0 },
            "datacopy-ip-address"  => { type => SCALAR,   optional => 0 },
            "multipath-ip-address" => { type => SCALAR,   optional => 1 },
        },
    );
    my $obj = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_stagingsystems_modify_smtt_cli',
                interface => 'CLI',
                set       => 'SMTT',
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $obj;
} ## end sub stagingsystems_modify

sub _stagingsystems_modify_smtt_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my (%target, $response);
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%target,
        copy   => [qw (session-name host-name vfiler-name aggregates datacopy-ip-address
                   multipath-ip-address ) ]
    );
    $response = $apiset->transition_stagingsystems_modify(%target);
    $Log->exit() if $may_exit;
    return $response;
} ## end sub _stagingsystems_modify_smtt_cli

=head2 stagingsystems_remove

    $resp = $trans_obj->stagingsystems_remove(
        "host-name" => $host_name,
    );


(Class or instance method) This method is used to remove staging system from the session .
"Returns the response object."

"Supports CLI".

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< session-name => $name >>

(Required for class method, Not Applicable for instance method)
Name of the transition session .

=item C<< host-name => $host_name >>

(Required) Hostname or IP address of Staging system.

=back

=back

=cut

sub stagingsystems_remove {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            "host-name" => { type => SCALAR,   optional => 0 },
        },
    );
    my $obj = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_stagingsystems_remove_smtt_cli',
                interface => 'CLI',
                set       => 'SMTT',
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $obj;
} ## end sub stagingsystems_remove

sub _stagingsystems_remove_smtt_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my (%target, $response);
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%target,
        copy   => [qw (session-name host-name ) ]
    );
    $response = $apiset->transition_stagingsystems_remove(%target);
    $Log->exit() if $may_exit;
    return $response;
} ## end sub _stagingsystems_remove_smtt_cli

=head2 volumeaggregatemapping_add

    $resp = $trans_obj->volumeaggregatemapping_add(
        "volume-name-c-mode"       => [qw(vol_compo1 vol_compo2)],
        "target-cluster-aggregate" => $clus_aggr,
        "volume-size"              => $vol_size,
        "staging-system-host-name" => $staging_host,
        "staging-aggregate"        => $staging_aggr,
    );

                OR

    $resp = $trans_obj->volumeaggregatemapping_add(
        "volume-name-c-mode"       => \@vol_cmode,
        "target-cluster-aggregate" => $clus_aggr,
        "volume-size"              => $vol_size,
        "staging-system-host-name" => $staging_host,
        "staging-aggregate"        => $staging_aggr,
    );

(Class or instance method) This method is used to add volumeaggregatemapping to the session .
"Returns the response object."

"Supports CLI".

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< session-name => $name >>

(Required for class method, Not Applicable for instance method)
Name of the transition session .

=item C<< volume-name-c-mode => $C_vol_arr_ref >>

(Required) Its a reference to an array which contains Cmode volumes.

=item C<< target-cluster-aggregate => $clus_aggr >>

(Required) Aggregate of target CMode.

=item C<< volume-size => $vol_size >>

(Optional) Size of Cmode volume .

=item C<< staging-system-host-name => $staging_host >>

(Required) Hostname or IP address of Staging system .

=item C<< staging-aggregate => $staging_aggr >>

(Required) Aggregate of staging system .

=back

=back

=cut

sub volumeaggregatemapping_add{
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            "volume-name-c-mode"       => { type => ARRAYREF, optional => 0 },
            "target-cluster-aggregate" => { type => SCALAR,   optional => 0 },
            "volume-size"              => { type => SCALAR,   optional => 1 },
            "staging-system-host-name" => { type => SCALAR,   optional => 0 },
            "staging-aggregate"        => { type => SCALAR,   optional => 0 },
        },
    );
    my $obj = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_volumeaggregatemapping_add_smtt_cli',
                interface => 'CLI',
                set       => 'SMTT',
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $obj;
} ## end sub precutover

sub _volumeaggregatemapping_add_smtt_cli{
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my $volume_name_c_mode = delete $opts{"volume-name-c-mode"};
    my (%target,$response,@response_objs);
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%target,
        copy   => [qw (
            session-name target-cluster-aggregate volume-size
                staging-system-host-name staging-aggregate) ]
    );
    foreach(0..@$volume_name_c_mode-1){
        $target{"volume-name-c-mode"} = $volume_name_c_mode->[$_];
        $response = $apiset->transition_volumeaggregatemapping_add(%target);
        push @response_objs,$response;
    }
    $Log->exit() if $may_exit;
    return @response_objs;
} ## end sub _volumeaggregatemapping_smtt_cli

=head2 volumeaggregatemapping_modify

    $resp = $trans_obj->volumeaggregatemapping_modify(
        "volume-name-c-mode"       => [qw(vol_compo1 vol_compo2)],
        "target-cluster-aggregate" => $clus_aggr,
        "volume-size"              => $vol_size,
        "staging-system-host-name" => $staging_host,
        "staging-aggregate"        => $staging_aggr,
    );

                OR

    $resp = $trans_obj->volumeaggregatemapping_modify(
        "volume-name-c-mode"       => \@vol_cmode,
        "target-cluster-aggregate" => $clus_aggr,
        "volume-size"              => $vol_size,
        "staging-system-host-name" => $staging_host,
        "staging-aggregate"        => $staging_aggr,
    );

(Class or instance method) This method is used to modify volumeaggregatemapping of the session .
"Returns the response object."

"Supports CLI".

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< session-name => $name >>

(Required for class method, Not Applicable for instance method)
Name of the transition session .

=item C<< volume-name-c-mode => $C_vol_arr_ref >>

(Required) Its a reference to an array which contains Cmode volumes.

=item C<< target-cluster-aggregate => $clus_aggr >>

(Optional) Aggregate of target CMode.

=item C<< volume-size => $vol_size >>

(Optional) Size of Cmode volume .

=item C<< staging-system-host-name => $staging_host >>

(Optional) Hostname or IP address of Staging system .

=item C<< staging-aggregate => $staging_aggr >>

(Optional) Aggregate of staging system .

=back

=back

=cut

sub volumeaggregatemapping_modify{
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            "volume-name-c-mode"       => { type => ARRAYREF, optional => 0 },
            "target-cluster-aggregate" => { type => SCALAR,   optional => 0 },
            "volume-size"              => { type => SCALAR,   optional => 1 },
            "staging-system-host-name" => { type => SCALAR,   optional => 0 },
            "staging-aggregate"        => { type => SCALAR,   optional => 0 },
        },
    );
    my $obj = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_volumeaggregatemapping_modify_smtt_cli',
                interface => 'CLI',
                set       => 'SMTT',
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $obj;
} ## end sub volumeaggregatemapping_modify

sub _volumeaggregatemapping_modify_smtt_cli{
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my $volume_name_c_mode = delete $opts{"volume-name-c-mode"};
    my (%target,$response,@response_objs);
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%target,
        copy   => [qw (
            session-name target-cluster-aggregate volume-size
                staging-system-host-name staging-aggregate) ]
    );
    foreach(0..@$volume_name_c_mode-1){
        $target{"volume-name-c-mode"} = $volume_name_c_mode->[$_];
        $response = $apiset->transition_volumeaggregatemapping_modify(%target);
        push @response_objs,$response;
    }
    $Log->exit() if $may_exit;
    return @response_objs;
} ## end sub _volumeaggregatemapping_modify_smtt_cli

=head2 volumeaggregatemapping_remove

    $resp = $trans_obj->volumeaggregatemapping_remove(
        "volume-name-c-mode"       => [qw(vol_compo1 vol_compo2)],
    );

                OR

    $resp = $trans_obj->volumeaggregatemapping_remove(
        "volume-name-c-mode"       => \@vol_cmode,
    );

(Class or instance method) This method is used to remove volumeaggregatemapping from the session .
"Returns the response object."

"Supports CLI".

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< session-name => $name >>

(Required for class method, Not Applicable for instance method)
Name of the transition session .

=item C<< volume-name-c-mode => $C_vol_arr_ref >>

(Required) Its a reference to an array which contains Cmode volumes.

=back

=back

=cut

sub volumeaggregatemapping_remove{
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            "volume-name-c-mode"       => { type => ARRAYREF, optional => 0 },
        },
    );
    my $obj = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_volumeaggregatemapping_remove_smtt_cli',
                interface => 'CLI',
                set       => 'SMTT',
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $obj;
} ## end sub volumeaggregatemapping_remove

sub _volumeaggregatemapping_remove_smtt_cli{
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my $volume_name_c_mode = delete $opts{"volume-name-c-mode"};
    my (%target,$response,@response_objs);
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%target,
        copy   => [qw ( session-name) ]
    );
    foreach(0..@$volume_name_c_mode-1){
        $target{"volume-name-c-mode"} = $volume_name_c_mode->[$_];
        $response = $apiset->transition_volumeaggregatemapping_remove(%target);
        push @response_objs,$response;
    }
    $Log->exit() if $may_exit;
    return @response_objs;
} ## end sub _volumeaggregatemapping_smtt_cli

=head2 cft_show

    $resp = NACL::C::Client::Transition->cft_show(
        command_interface => $Client,
        project-name      => $Project_name,
        verbose           => " ",
    );

(Class method) This method is used to display transition CFT projects' status
"Returns the response object."

"Supports CLI".

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< project-name => $Project_name >>

(Required) project Name.

=item C<< verbose => "blank space" >>

(Optional) verbose mode.

=back

=back

=cut

sub cft_show {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            "project-name" => {type => SCALAR, optional => 1},
        },
        ignore_primary_keys => 1,
    );
    my $obj = $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_cft_show_smtt_cli',
                interface => 'CLI',
                set       => 'SMTT',
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $obj;
} ## end sub cft_show

sub _cft_show_smtt_cli {
    $Log->enter() if $may_enter;
    my ($pkg,@args) = @_;
    my %opts = $pkg->_validate_backend_opts(@args);
    my $project_name = delete $opts{"project-name"};
    my $apiset = delete $opts{apiset};
    my %target;
    $pkg->_hash_copy(
        source => \%opts,
        target => \%target,
    );
    $target{"project-name"} = $project_name if ($project_name);
    my $response = $apiset->transition_cft_show( %target );
    $Log->exit() if $may_exit;
    return $response;
} ## end sub _cft_show_smtt_cli

=head2 cft_svmmapping_show

    $resp = NACL::C::Client::Transition->cft_svmmapping_show(
        command_interface => $Client,
        "project-name"  => $Project_name,
        svm             => $VSERVER.1,
    );
Shows details of cft svm mapped to the given project name
                        OR
    $resp = NACL::C::Client::Transition->cft_svmmapping_show(
        command_interface => $Client,
        "project-name"    => $Project_name
    );
Shows details of list of svm mapped to the given project name

(Class or instance method) This method is used to show svm mapping details of a CFT project .
"Returns the response object."

"Supports CLI".

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< project-name => $name >>

(Required) Name of the transition cft project .

=item C<< svm => $svm_name >>

(Required)If a particular svm mapping details are needed.
(Optional)If details of all mapped svm to the project are needed at a time

=back

=back

=cut

sub cft_svmmapping_show {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            "project-name" => { type => SCALAR, optional => 0 },
            "svm-name"     => { type => SCALAR, optional => 1 },
        },
        ignore_primary_keys => 1,
    );
    my @obj = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_cft_svmmapping_show_smtt_cli',
                interface => 'CLI',
                set       => 'SMTT',
            },
        ],
    );
    $Log->exit() if $may_exit;
    return wantarray ? @obj : $obj[0];
} ## end sub cft_svmmapping_show

sub _cft_svmmapping_show_smtt_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my $svm_name = delete $opts{"svm-name"};
    my (%target, $response, @response_objs);
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%target,
        copy   => [qw (project-name) ]
    );
    $target{"svm-name"} = $svm_name if ( $svm_name );
    $response = $apiset->transition_cft_svmmapping_show(%target);
    $Log->exit() if $may_exit;
    return $response;
} ## end sub _cft_svmmapping_show_smtt_cli

=head2 cft_aggregate_show

    $resp = NACL::C::Client::Transition->cft_aggregate_show(
        command_interface => $Client,
        "project-name"    => $Project_name
    );
Shows details of list of aggregates in the given project name

(Class or instance method) This method is used to show aggregates details .
"Returns the response object."

"Supports CLI".

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< project-name => $name >>

(Required)

=back

=back

=cut

sub cft_aggregate_show {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            "project-name" => { type => SCALAR, optional => 0 },
        },
        ignore_primary_keys => 1,
    );
    my @obj = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_cft_aggregate_show_smtt_cli',
                interface => 'CLI',
                set       => 'SMTT',
            },
        ],
    );
    $Log->exit() if $may_exit;
    return wantarray ? @obj : $obj[0];
} ## end sub cft_aggregate_show

sub _cft_aggregate_show_smtt_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my (%target, $response, @response_objs);
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%target,
        copy   => [qw (project-name) ]
    );
    $response = $apiset->transition_cft_aggregate_show(%target);
    $Log->exit() if $may_exit;
    return $response;
} ## end sub _cft_aggregate_show_smtt_cli

=head2 cft_volumepair_show

    $resp = NACL::C::Client::Transition->cft_volumepair_show(
        command_interface => $Client,
        "project-name"    => $Project_name
    );
Shows details of list of volumepairs in the given project name

(Class or instance method) This method is used to show volume pairs details .
"Returns the response object."

"Supports CLI".

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< project-name => $name >>

(Required)

=back

=back

=cut

sub cft_volumepair_show {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            "project-name" => { type => SCALAR, optional => 0 },
        },
        ignore_primary_keys => 1,
    );
    my @obj = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_cft_volumepair_show_smtt_cli',
                interface => 'CLI',
                set       => 'SMTT',
            },
        ],
    );
    $Log->exit() if $may_exit;
    return wantarray ? @obj : $obj[0];
} ## end sub cft_volumepair_show

sub _cft_volumepair_show_smtt_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my (%target, $response, @response_objs);
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%target,
        copy   => [qw (project-name) ]
    );
    $response = $apiset->transition_cft_volumepair_show(%target);
    $Log->exit() if $may_exit;
    return $response;
} ## end sub _cft_volumepair_show_smtt_cli

=head2 cft_delete

    NACL::C::Client::Transition->cft_delete(
        command_interface => $Client,
        "project-name"    => $Project_name
    );
Runs 'delete' operation on the given CFT project

(Class or instance method) This method is used to Run 'delete' operation.

"Supports CLI,GUI".

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< project-name=>$Project_name >>

(Required for class method, Not Applicable for instance method)

=item C<< group-name=>$group_name >>

(Required for class method, Not Applicable for instance method)

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< project-name => $name >>

(Required)

=back

=back

=cut

sub cft_delete {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            "project-name" => { type => SCALAR, optional => 1 },
            "group-name"   => { type => SCALAR, optional => 1 },
        },
        ignore_primary_keys => 1,
    );
    my @obj= $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_cft_delete_smtt_cli',
                interface => 'CLI',
                set       => 'SMTT',
            },
            {   method    => '_cft_delete_smtt_gui',
                interface => 'GUI',
                set       => 'SMTT',
            },
        ],
    );
    $Log->exit() if $may_exit;
    return wantarray ? @obj : $obj[0];
} ## end sub cft_delete

sub _cft_delete_smtt_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my (%target, $response, @response_objs);
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%target,
        copy   => [qw (project-name) ]
    );
    my $force_prompt = ["Do you want to" => "Yes" ];
    $response = $apiset->transition_cft_delete(
        %target,
        'connectrec-match_table' => $force_prompt,
    );
    $Log->exit() if $may_exit;
    return $response;
} ## end sub _cft_delete_smtt_cli

sub _cft_delete_smtt_gui {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my (%target, $response, @response_objs);
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%target,
        copy   => [qw (project-name group-name ) ]
    );
    $response = $apiset->transition_cft_delete(%target);
    $Log->exit() if $may_exit;
    return $response;
} ## end sub _cft_delete_smtt_gui


=head2 cft_precheck

    $resp = NACL::C::Client::Transition->cft_precheck(
        command_interface => $Client,
        "project-name"    => $Project_name
    );
Runs 'Precheck' operation on the given CFT project

(Class or instance method) This method is used to Run 'Precheck' operation.
"Returns the response object."

"Supports CLI,GUI".

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< project-name=>$Project_name >>

(Required for class method, Not Applicable for instance method)

=item C<< group-name=>$group_name >>

(Required for class method, Not Applicable for instance method)

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< project-name => $name >>

(Required)

=back

=back

=cut

sub cft_precheck {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            "project-name" => { type => SCALAR, optional => 1 },
            "group-name"   => { type => SCALAR, optional => 1 },
            "parsed-output"=> { type => SCALAR, optional => 1 },
            "run-background"=> { type => SCALAR , optional => 1},
        },
        ignore_primary_keys => 1,
    );
    my @obj= $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_cft_precheck_smtt_cli',
                interface => 'CLI',
                set       => 'SMTT',
            },
            {   method    => '_cft_precheck_smtt_gui',
                interface => 'GUI',
                set       => 'SMTT',
            },
        ],
    );
    $Log->exit() if $may_exit;
    return wantarray ? @obj : $obj[0];
} ## end sub cft_precheck

sub _cft_precheck_smtt_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my (%target, $response, @response_objs);
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%target,
        copy   => [qw (project-name) ]
    );
    $response = $apiset->transition_cft_precheck(%target);
    $Log->exit() if $may_exit;
    return $response;
} ## end sub _cft_precheck_smtt_cli

sub _cft_precheck_smtt_gui {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my (%target, $response, @response_objs);
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%target,
        copy   => [qw (project-name group-name parsed-output run-background) ]
    );
    $response = $apiset->transition_cft_precheck(%target);
    $Log->exit() if $may_exit;
    return $response;
} ## end sub _cft_precheck_smtt_gui

sub cft_check_readiness {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            "project-name" => { type => SCALAR, optional => 1 },
            "group-name"   => { type => SCALAR, optional => 1 },
            "parsed-output"=> { type => SCALAR, optional => 1 },
            "run-background"=> { type => SCALAR , optional => 1},
        },
        ignore_primary_keys => 1,
    );
    my @obj= $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_cft_check_readiness_smtt_gui',
                interface => 'GUI',
                set       => 'SMTT',
            },
        ],
    );
    $Log->exit() if $may_exit;
    return wantarray ? @obj : $obj[0];
} ## end sub cft_precheck

sub _cft_check_readiness_smtt_gui {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my (%target, $response, @response_objs);
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%target,
        copy   => [qw (project-name group-name parsed-output run-background) ]
    );
    $response = $apiset->transition_cft_check_readiness(%target);
    $Log->exit() if $may_exit;
    return $response;
} ## end sub _cft_precheck_smtt_gui



=head2 cft_apply_svm_config

    $resp = NACL::C::Client::Transition->cft_apply_svm_config(
        command_interface => $Client,
        "project-name"    => $Project_name
    );
Runs 'apply svm config' operation on the given CFT project

(Class or instance method) This method is used to Run 'apply svm config' operation.
"Returns the response object."

"Supports CLI,GUI".

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< project-name=>$Project_name >>

(Required for class method, Not Applicable for instance method)

=item C<< group-name=>$group_name >>

(Required for class method, Not Applicable for instance method)

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=back

=back

=cut

sub cft_apply_svm_config {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            "project-name" => { type => SCALAR, optional => 1 },
            "group-name"   => { type => SCALAR, optional => 1 },
            "parsed-output"=> { type => SCALAR, optional => 1 },
            "run-background"=> { type => SCALAR , optional => 1},
        },
        ignore_primary_keys => 1,
    );
    my @obj= $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_cft_apply_svm_config_smtt_cli',
                interface => 'CLI',
                set       => 'SMTT',
            },
            {   method    => '_cft_apply_svm_config_smtt_gui',
                interface => 'GUI',
                set       => 'SMTT',
            },
        ],
    );
    $Log->exit() if $may_exit;
    return wantarray ? @obj : $obj[0];
} ## end sub cft_precheck

sub _cft_apply_svm_config_smtt_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my (%target, $response, @response_objs);
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%target,
        copy   => [qw (project-name) ]
    );
    $response = $apiset->transition_cft_apply_svm_config(%target);
    $Log->exit() if $may_exit;
    return $response;
} ## end sub _cft_apply_svm_config_smtt_cli

sub _cft_apply_svm_config_smtt_gui {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my (%target, $response, @response_objs);
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%target,
        copy   => [qw (project-name group-name parsed-output run-background) ]
    );
    $response = $apiset->transition_cft_apply_svm_config(%target);
    $Log->exit() if $may_exit;
    return $response;
} ## end sub _cft_apply_svm_config_smtt_gui



=head2 cft_start

    $resp = NACL::C::Client::Transition->cft_start(
        command_interface => $Client,
        "project-name"    => $Project_name
    );
Runs 'Start' operation on the given CFT project

(Class or instance method) This method is used to Run 'cft start' operation.
"Returns the response object."

"Supports CLI,GUI".

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< project-name=>$Project_name >>

(Required for class method, Not Applicable for instance method)

=item C<< group-name=>$group_name >>

(Required for class method, Not Applicable for instance method)

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=back

=back

=cut

sub cft_start {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            "project-name" => { type => SCALAR, optional => 1 },
            "group-name"   => { type => SCALAR, optional => 1 },
            "parsed-output"=> { type => SCALAR, optional => 1 },
            "run-background"=> { type => SCALAR , optional => 1},
        },
        ignore_primary_keys => 1,
    );
    my @obj= $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_cft_start_smtt_cli',
                interface => 'CLI',
                set       => 'SMTT',
            },
            {   method    => '_cft_start_smtt_gui',
                interface => 'GUI',
                set       => 'SMTT',
            },
        ],
    );
    $Log->exit() if $may_exit;
    return wantarray ? @obj : $obj[0];
} ## end sub cft_start

sub _cft_start_smtt_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my (%target, $response, @response_objs);
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%target,
        copy   => [qw (project-name) ]
    );
    my $force_prompt = ["Do you want to start the transition" => "Yes" ];
    $response = $apiset->transition_cft_start(
        %target,
        'connectrec-match_table' => $force_prompt,
    );
    $Log->exit() if $may_exit;
    return $response;
} ## end sub _cft_start_smtt_cli

sub _cft_start_smtt_gui {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my (%target, $response, @response_objs);
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%target,
        copy   => [qw (project-name group-name parsed-output run-background) ]
    );
    $response = $apiset->transition_cft_start(%target);
    $Log->exit() if $may_exit;
    return $response;
} ## end sub _cft_start_smtt_gui

=head2 cft_import

    $resp = NACL::C::Client::Transition->cft_import(
        command_interface => $Client,
        "project-name"    => $Project_name
    );
Runs 'Import' operation on the given CFT project

(Class or instance method) This method is used to Run 'cft import' operation.
"Returns the response object."

"Supports CLI,GUI".

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< project-name=>$Project_name >>

(Required for class method, Not Applicable for instance method)

=item C<< group-name=>$group_name >>

(Required for class method, Not Applicable for instance method)

=back

=back

=cut

sub cft_import {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            "project-name" => { type => SCALAR, optional => 1 },
            "group-name"   => { type => SCALAR, optional => 1 },
            "parsed-output"=> { type => SCALAR, optional => 1 },
            "run-background"=> { type => SCALAR , optional => 1},
        },
        ignore_primary_keys => 1,
    );
    my @obj= $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_cft_import_smtt_cli',
                interface => 'CLI',
                set       => 'SMTT',
            },
            {   method    => '_cft_import_smtt_gui',
                interface => 'GUI',
                set       => 'SMTT',
            },
        ],
    );
    $Log->exit() if $may_exit;
    return wantarray ? @obj : $obj[0];
} ## end sub cft_import

sub _cft_import_smtt_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my (%target, $response, @response_objs);
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%target,
        copy   => [qw (project-name) ]
    );
    $response = $apiset->transition_cft_import(%target);
    $Log->exit() if $may_exit;
    return $response;
} ## end sub _cft_import_smtt_cli

sub _cft_import_smtt_gui {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my (%target, $response, @response_objs);
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%target,
        copy   => [qw (project-name group-name parsed-output run-background) ]
    );
    $response = $apiset->transition_cft_import(%target);
    $Log->exit() if $may_exit;
    return $response;
} ## end sub _cft_import_smtt_gui

sub cft_commit {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            "project-name" => { type => SCALAR, optional => 1 },
            "group-name"   => { type => SCALAR, optional => 1 },
            "parsed-output"=> { type => SCALAR, optional => 1 },
            "run-background"=> { type => SCALAR , optional => 1},
        },
        ignore_primary_keys => 1,
    );
    my @obj= $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_cft_commit_smtt_cli',
                interface => 'CLI',
                set       => 'SMTT',
            },
            {   method    => '_cft_commit_smtt_gui',
                interface => 'GUI',
                set       => 'SMTT',
            },
        ],
    );
    $Log->exit() if $may_exit;
    return wantarray ? @obj : $obj[0];
} ## end sub cft_import

sub _cft_commit_smtt_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my (%target, $response, @response_objs);
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%target,
        copy   => [qw (project-name) ]
    );
    my $force_prompt = ["Do you want to" => "Yes" ];
    $response = $apiset->transition_cft_commit(
        %target,
        'connectrec-match_table' => $force_prompt,
    );
    $Log->exit() if $may_exit;
    return $response;
} ## end sub _cft_import_smtt_cli

sub _cft_commit_smtt_gui {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my (%target, $response, @response_objs);
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%target,
        copy   => [qw (project-name group-name parsed-output run-background) ]
    );
    $response = $apiset->transition_cft_commit(%target);
    $Log->exit() if $may_exit;
    return $response;
} ## end sub _cft_import_smtt_gui

=head2 cft_svmmapping_add

    $res = NACL::C::Client::Transition->cft_svmmapping_add(
        "command_interface"         => $HostWindows,
        "project-name"              => $pr_name,
        "7-mode-ip-or-fqdn"         => $host_ip,
        "svm-name"                  => $svm_name,
        "svm-type"                  => $svm_type
    );
(Class or instance method) This method is used to add
svm mapping to the CFT project.
"Returns the response object."

"Supports CLI,GUI".

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< project-name=>$Project_name >>

(Required for class method, Not Applicable for instance method)

=item C<< group-name=>$group_name >>

(Required for class method, Not Applicable for instance method)

=item C<< 7-mode-ip-or-fqdn => $host_ip >>

(Required) IP address of the 7Mode node .

=item C<< svm-name => $svm_name >>

(Required) svm_name for the CFT project .

=item C<< svm-type => $svm_type >>

(Required) svm_type for the CFT project .

=back

=back

=cut
sub cft_svmmapping_add {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            "project-name"              => { type => SCALAR },
            "7-mode-host-name"          => { optional => 0 },
            "svm-name"                  => { type => SCALAR },
            "svm-type"                  => { type => SCALAR,optional => 1 },
            "group-name"                => { type => SCALAR, optional => 1 },
            _optional_scalars qw( vfiler-name ),
        },
        ignore_primary_keys => 1,
    );
    my $obj = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_cft_svmmapping_add_smtt_cli',
                interface => 'CLI',
                set       => 'SMTT',
            },
            {   method    => '_cft_svmmapping_add_smtt_gui',
                interface => 'GUI',
                set       => 'SMTT',
            },
        ],
    );

    $Log->exit() if $may_exit;
    return $obj;
} ## end sub cft_svmmapping_add

sub _cft_svmmapping_add_smtt_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my %target;
    $pkg_or_obj->_hash_copy(
      source => \%opts,
        target => \%target,
        copy    =>  [
            qw( project-name 7-mode-host-name vfiler-name svm-name svm-type )
        ]
    );
    my $response = $apiset->transition_cft_svmmapping_add(%target);
    $Log->exit() if $may_exit;
    return $response;
} ## end sub _cft_svmmapping_add_smtt_cli

sub _cft_svmmapping_add_smtt_gui {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my (%target, $response, @response_objs);
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%target,
        copy   => [qw (project-name 7-mode-host-name vfiler-name svm-name) ]
    );
    $response = $apiset->transition_cft_svmmapping_add(%target);
    $Log->exit() if $may_exit;
    return $response;
} ## end sub _cft_svmmapping_add_smtt_gui

=head2 cft_svmmapping_modify

    $res = NACL::C::Client::Transition->cft_svmmapping_modify(
        "command_interface"         => $HostWindows,
        "project-name"              => $pr_name,
        "7-mode-ip-or-fqdn"         => $host_ip,
        "svm-name"                  => $svm_name,
        "svm-type"                  => $svm_type
    );
(Class or instance method) This method is used to modify info. in the
svm mapping of the CFT project.
"Returns the response object."

"Supports CLI".

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< project-name => $project_name >>

(Required) Unique project name of the CFT project .

=item C<< 7-mode-ip-or-fqdn => $host_ip >>

(Required) IP address of the 7Mode node .

=item C<< svm-name => $svm_name >>

(Required) svm_name for the CFT project .

=item C<< svm-type => $svm_type >>

(Required) svm_type for the CFT project .

=back

=back

=cut
sub cft_svmmapping_modify {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            "project-name"              => { type => SCALAR },
            "7-mode-ip-or-fqdn"         => { optional => 0 },
            "svm-name"                  => { type => SCALAR },
            "svm-type"                  => { type => SCALAR },
            _optional_scalars qw( vfiler-name ),
        },
        ignore_primary_keys => 1,
    );
    my $obj=$pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_cft_svmmapping_modify_smtt_cli',
                interface => 'CLI',
                set       => 'SMTT',
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $obj;
} ## end sub cft_svmmapping_modify

sub _cft_svmmapping_modify_smtt_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my %target;
    $pkg_or_obj->_hash_copy(
      source => \%opts,
        target => \%target,
        copy    =>  [
            qw( project-name 7-mode-ip-or-fqdn vfiler-name svm-name svm-type )
        ]
    );
    my $response = $apiset->transition_cft_svmmapping_modify(%target);
    $Log->exit() if $may_exit;
    return $response;
} ## end sub _cft_svmmapping_modify_smtt_cli

=head2 cft_volumepair_modify

    $res = NACL::C::Client::Transition->cft_volumepair_modify(
        "command_interface"         => $HostWindows,
        "project-name"              => $project,
        "7-mode-ip-or-fqdn"         => "10.63.34.76",
        "vfiler-name"               => $VFILER,
        "source-volume-name"        => $vol.1,
        "target-volume-name"        => $vol.1
    );
(Class or instance method) This method is used to modify volumepair in cft
cluster information to the session.
"Returns the response object."

"Supports CLI".

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< project-name => $project >>

(Required) Unique project name of the cft project .

=item C<< 7-mode-ip-or-fqdn => $7Mode-ip >>

(Required) Ip address of 7mode HA node.

=item C<< vfiler-name => $VFILER >>

(Required) vfiler name of HA node.

=item C<< source-volume-name => $vol.1 >>

(Required) volume name of source volume in HA node.

=item C<< target-volume-name => $vol.1 >>

(Required) volume name of target volume in the cluster node.

=back

=back

=cut
sub cft_volumepair_modify {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            "project-name"              => { type => SCALAR },
            "7-mode-ip-or-fqdn"         => { optional => 0 },
            "vfiler-name"               => { type => SCALAR },
            "source-volume-name"        => { type => SCALAR },
            "target-volume-name"        => { type => SCALAR },
        },
        ignore_primary_keys => 1,
    );
    my $obj=$pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_cft_volumepair_modify_smtt_cli',
                interface => 'CLI',
                set       => 'SMTT',
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $obj;
} ## end sub cft_volumepair_modify

sub _cft_volumepair_modify_smtt_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my %target;
    $pkg_or_obj->_hash_copy(
      source => \%opts,
        target => \%target,
        copy    =>  [
            qw( project-name 7-mode-ip-or-fqdn vfiler-name
                source-volume-name target-volume-name
            )
        ]
    );
    my $response = $apiset->transition_cft_volumepair_modify(%target);
    $Log->exit() if $may_exit;
    return $response;
} ## end sub _cft_volumepair_modify_smtt_cli

=head2 cft_create

    $res = NACL::C::Client::Transition->cft_create(
        "command_interface"         => $HostWindows,
        "project-name"              => $project_name,
        "7-mode1-host-name"         => $7_mode1_host_name,
        "7-mode2-host-name"         => $7_mode2_host_name,
        "group-name"                => $grp_name
    );
(Class or instance method) This method is used to add
cluster information to the session.
"Returns the response object."

"Supports CLI,GUI".

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< project-name => $project_name >>

(Required) CFT Project name .

=item C<< group-name => $group_name >>

(Optional) CFT Group name .

=item C<< 7-mode1-host-name => $7_mode1_host_name >>

(Required) 7Mode ha1 node name.

=item C<< 7-mode2-host-name => $7_mode2_host_name >>

(Required) 7Mode ha2 node name.

=item C<< group-name => $project_group_name >>

(Optional) CFT project group name.

=back

=back

=cut

sub cft_create {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            "project-name"            => { type => SCALAR },
            "group-name"              => { type => SCALAR, optional => 1},
            "7-mode1-host-name"       => { type => SCALAR },
            "7-mode2-host-name"       => { type => SCALAR },
            "S_mode_credentials"      => { optional => 1 },
            "S_mode_credentials_HA"   => { optional => 1 },
            "C_mode_credentials"      => { optional => 1 },
            "cft_clusterinfo_add"     => { optional => 1 },
            "cft_svm_mapping_add_pri" => { optional => 1 },
            "cft_svm_mapping_add_sec" => { optional => 1 },
            "cft_nodemapping_set"     => { optional => 1 },
            "plan_configuration"      => { optional => 1 },
            _optional_scalars qw( group-name ),
        },
        ignore_primary_keys => 1,
    );
    my $obj = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_cft_create_smtt_cli',
                interface => 'CLI',
                set       => 'SMTT',
            },
            {   method    => '_cft_create_smtt_gui',
                interface => 'GUI',
                set       => 'SMTT',
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $obj;
} ## end sub cft_create

sub _cft_create_smtt_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my %target;
    $pkg_or_obj->_hash_copy(
      source => \%opts,
        target => \%target,
        copy    =>  [
            qw( project-name 7-mode1-host-name 7-mode2-host-name
                group-name
            )
        ]
    );
    $apiset->transition_cft_create(%target);
    $Log->exit() if $may_exit;
    return $pkg_or_obj->new(
        command_interface => $opts{command_interface},
        'project-name'    => $opts{'project-name'},
        'group-name'      => $opts{'group-name'},
    );
} ## end sub _cft_create_smtt_cli

sub _cft_create_smtt_gui {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my %target;
    $pkg_or_obj->_hash_copy(
      source => \%opts,
        target => \%target,
        copy    =>  [
            qw( project-name 7-mode1-host-name 7-mode2-host-name
                group-name S_mode_credentials C_mode_credentials
                S_mode_credentials_HA cft_clusterinfo_add cft_svm_mapping_add_pri
                cft_svm_mapping_add_sec cft_nodemapping_set plan_configuration
            )
        ]
    );

    $apiset->transition_cft_create(%target);
    $Log->exit() if $may_exit;
    return $pkg_or_obj->new(
        command_interface => $opts{command_interface},
        'project-name'    => $opts{'project-name'},
        'group-name'      => $opts{'group-name'},
    );
    #return $response;
} ## end sub _cft_create_smtt_cli

=head2 cft_ignorableerrors_add

    $res = NACL::C::Client::Transition->cft_ignorableerrors_add(
        "command_interface"         => $HostWindows,
        "project-name"              => $Project_name,
        "ignorableerrors"           => $ignorableerrors,
    );
(Class or instance method) This method is used to add
cluster information to the session.
"Returns the response object."

"Supports CLI".

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

item C<< project-name => $Project_name >>

(Required) CFT project name .

=item C<< ignorableerrors => $ignorableerrors >>

(Required) CFT ignorable errors .

=back

=back

=cut
sub cft_ignorableerrors_add {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            "project-name"          => { type => SCALAR },
            "ignorableerrors"       => { type => SCALAR },
        },
        ignore_primary_keys => 1,
    );
    my $obj=$pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_cft_ignorableerrors_add_smtt_cli',
                interface => 'CLI',
                set       => 'SMTT',
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $obj;
} ## end sub cft_ignorableerrors_add

sub _cft_ignorableerrors_add_smtt_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my %target;
    $pkg_or_obj->_hash_copy(
      source => \%opts,
        target => \%target,
        copy    =>  [
            qw( project-name ignorableerrors )
        ]
    );
    my $response = $apiset->transition_cft_ignorableerrors_add(%target);
    $Log->exit() if $may_exit;
    return $response;
} ## end sub _cft_ignorableerrors_add_smtt_cli


=head2 ignorableerrors_add

    $res = NACL::C::Client::Transition->ignorableerrors_add(
        "command_interface"         => $HostWindows,
        "session-name"              => $Session_name,
        "ignorableerrors"           => $ignorableerrors,
    );
(Class or instance method) This method is used to add
cluster information to the session.
"Returns the response object."

"Supports CLI".

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

item C<< project-name => $Session_name >>

(Required) CBT Session name .

=item C<< ignorableerrors => $ignorableerrors >>

(Required) CBT ignorable errors .

=back

=back

=cut

sub ignorableerrors_add {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            "session-name"          => { type => SCALAR, optional => 1 },
            "ignorableerrors"       => { type => ARRAYREF },
        },
        ignore_primary_keys => 1,
    );
    my @obj = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_ignorableerrors_add_smtt_cli',
                interface => 'CLI',
                set       => 'SMTT',
            },
        ],
    );
    $Log->exit() if $may_exit;
    return wantarray ? @obj : $obj[0];
} ## end sub ignorableerrors_add

sub _ignorableerrors_add_smtt_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my $ignorableerrors    = delete $opts{"ignorableerrors"};
    my (%target, $response, @response_objs);
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%target,
        copy   => [qw (session-name) ]
    );
    foreach (0..@$ignorableerrors-1){
        $target{"ignorableerrors"}=$ignorableerrors->[$_];
        $response = $apiset->transition_ignorableerrors_add(%target);
        push @response_objs,$response;
    }
    $Log->exit() if $may_exit;
    return @response_objs;
} ## end sub _ignorableerrors_add_smtt_cli

=head2 cft_clusterinfo_add

    $res = $trans_obj->cft_clusterinfo_add(
        "c-mode-host-name"          => $CMode_hostname,
    );
(Class or instance method) This method is used to add
cluster information to the session.
"Returns the response object."

"Supports CLI,GUI".

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< project-name=>$Project_name >>

(Required for class method, Not Applicable for instance method)

=item C<< group-name=>$group_name >>

(Required for class method, Not Applicable for instance method)

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< project-name=>$Project_name >>

(Required for class method, Not Applicable for instance method)

=item C<< group-name=>$group_name >>

(Required for class method, Not Applicable for instance method)

=item C<< c-mode-host-name => $CMode_hostname >>

(Required) Cmode hostname/IPAddress.

=back

=back

=cut

sub cft_clusterinfo_add {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            "project-name"      => { type => SCALAR },
            "c-mode-host-name"  => { type => SCALAR },
            "group-name"        => { type => SCALAR, optional => 1 },
        },
        ignore_primary_keys => 1,
    );
    my $obj= $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_cft_clusterinfo_add_smtt_cli',
                interface => 'CLI',
                set       => 'SMTT',
            },
            {   method    => '_cft_clusterinfo_add_smtt_gui',
                interface => 'GUI',
                set       => 'SMTT',
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $obj;
} ## end sub cft_clusterinfo_add

sub _cft_clusterinfo_add_smtt_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my %target;
    $pkg_or_obj->_hash_copy(
      source => \%opts,
        target => \%target,
        copy    =>  [
            qw( project-name c-mode-host-name )
        ]
    );
    my $response = $apiset->transition_cft_clusterinfo_add(%target);
    $Log->exit() if $may_exit;
    return $response;
} ## end sub _cft_clusterinfo_add_smtt_cli

sub _cft_clusterinfo_add_smtt_gui {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my (%target, $response, @response_objs);
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%target,
        copy   => [qw (project-name c-mode-host-name group-name) ]
    );
    $response = $apiset->transition_cft_clusterinfo_add(%target);
    $Log->exit() if $may_exit;
    return $response;
} ## end sub _cft_clusterinfo_add_smtt_gui


=head2 cft_nodemapping_set

    $res = NACL::C::Client::Transition->cft_nodemapping_set(
        "command_interface"         => $HostWindows,
        "project-name"              => $Project_name,
        "cluster-node1-name"        => $cluster_node1_name,
        "cluster-node2-name"        => $cluster_node2_name,
    );
(Class or instance method) This method is used to add
cluster information to the session.
"Returns the response object."

"Supports CLI".

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< project-name=>$Project_name >>

(Required for class method, Not Applicable for instance method)

=item C<< group-name=>$group_name >>

(Required for class method, Not Applicable for instance method)

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< project-name=>$Project_name >>

(Required for class method, Not Applicable for instance method)

=item C<< group-name=>$group_name >>

(Required for class method, Not Applicable for instance method)

=item C<< cluster-node1-name => $cluster_node1_name  >>

(Required) Node1 name in the CMode cluster .

=item C<< cluster-node2-name => $cluster_node2_name >>

(Required) Node2 name in the CMode cluster .

=back

=back

=cut
sub cft_nodemapping_set {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            "project-name"        => { type => SCALAR },
            "cluster-node1-name"  => { type => SCALAR },
            "cluster-node2-name"  => { type => SCALAR },
            "group-name"          => { type => SCALAR, optional => 1 },
        },
        ignore_primary_keys => 1,
    );
    my $obj= $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_cft_nodemapping_set_smtt_cli',
                interface => 'CLI',
                set       => 'SMTT',
            },
            {   method    => '_cft_nodemapping_set_smtt_gui',
                interface => 'GUI',
                set       => 'SMTT',
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $obj;
} ## end sub cft_nodemapping_set

sub _cft_nodemapping_set_smtt_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my %target;
    $pkg_or_obj->_hash_copy(
      source => \%opts,
        target => \%target,
        copy    =>  [
            qw( project-name cluster-node1-name cluster-node2-name)
        ]
    );
    my $response = $apiset->transition_cft_nodemapping_set(%target);
    $Log->exit() if $may_exit;
    return $response;
} ## end sub _cft_nodemapping_set_smtt_cli

sub _cft_nodemapping_set_smtt_gui {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my (%target, $response, @response_objs);
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%target,
        copy   => [qw (project-name cluster-node1-name cluster-node2-name group-name) ]
    );
    $response = $apiset->transition_cft_nodemapping_set(%target);
    $Log->exit() if $may_exit;
    return $response;
} ## end sub _cft_nodemapping_set_smtt_gui

=head2 cft_verify_cabling

    $resp = NACL::C::Client::Transition->cft_verify_cabling(
        command_interface => $Client,
        "project-name"    => $Project_name
    );
Runs 'verify cabling' operation on the given CFT project

(Class or instance method) This method is used to Run 'cft verify-cabling' operation.
"Returns the response object."

"Supports CLI,GUI".

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< project-name=>$Project_name >>

(Required for class method, Not Applicable for instance method)

=item C<< group-name=>$group_name >>

(Required for class method, Not Applicable for instance method)

=back

=back

=cut

sub cft_verify_cabling {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            "project-name" => { type => SCALAR, optional => 1 },
            "group-name"   => { type => SCALAR, optional => 1 },
            "parsed-output"=> { type => SCALAR, optional => 1 },
            "run-background"=> { type => SCALAR , optional => 1},
        },
        ignore_primary_keys => 1,
    );
    my @obj= $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_cft_verify_cabling_smtt_cli',
                interface => 'CLI',
                set       => 'SMTT',
            },
            {   method    => '_cft_verify_cabling_smtt_gui',
                interface => 'GUI',
                set       => 'SMTT',
            },
        ],
    );
    $Log->exit() if $may_exit;
    return wantarray ? @obj : $obj[0];
} ## end sub cft_verify_cabling

sub _cft_verify_cabling_smtt_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my (%target, $response, @response_objs);
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%target,
        copy   => [qw (project-name) ]
    );
    $response = $apiset->transition_cft_verify_cabling(%target);
    $Log->exit() if $may_exit;
    return $response;
} ## end sub _cft_verify_cabling_smtt_cli

sub _cft_verify_cabling_smtt_gui {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my (%target, $response, @response_objs);
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%target,
        copy   => [qw (project-name group-name parsed-output run-background) ]
    );
    $response = $apiset->transition_cft_verify_cabling(%target);
    $Log->exit() if $may_exit;
    return $response;
} ## end sub _cft_verify_cabling_smtt_gui


=head2 cft_revert_precheck

    $resp = NACL::C::Client::Transition->cft_revert_precheck(
        command_interface => $Client,
        "project-name"    => $Project_name
    );
Runs 'revert precheck' operation on the given CFT project

(Class or instance method) This method is used to Run 'cft revert precheck' operation.
"Returns the response object."

"Supports CLI,GUI".

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< project-name=>$Project_name >>

(Required for class method, Not Applicable for instance method)

=item C<< group-name=>$group_name >>

(Required for class method, Not Applicable for instance method)

=back

=back

=cut

sub cft_revert_precheck {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            "project-name" => { type => SCALAR, optional => 1 },
            "group-name"   => { type => SCALAR, optional => 1 },
            "parsed-output"=> { type => SCALAR, optional => 1 },
        },
        ignore_primary_keys => 1,
    );
    my @obj= $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_cft_revert_precheck_smtt_cli',
                interface => 'CLI',
                set       => 'SMTT',
            },
            {   method    => '_cft_revert_precheck_smtt_gui',
                interface => 'GUI',
                set       => 'SMTT',
            },
        ],
    );
    $Log->exit() if $may_exit;
    return wantarray ? @obj : $obj[0];
} ## end sub cft_revert_precheck

sub _cft_revert_precheck_smtt_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my (%target, $response, @response_objs);
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%target,
        copy   => [qw (project-name) ]
    );
    $response = $apiset->transition_cft_revert_precheck(%target);
    $Log->exit() if $may_exit;
    return $response;
} ## end sub _cft_revert_precheck_smtt_cli

sub _cft_revert_precheck_smtt_gui {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my (%target, $response, @response_objs);
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%target,
        copy   => [qw (project-name group-name parsed-output) ]
    );
    $response = $apiset->transition_cft_revert_precheck(%target);
    $Log->exit() if $may_exit;
    return $response;
} ## end sub _cft_revert_precheck_smtt_gui

=head2 cft_revert_generate_manual_steps

    $resp = NACL::C::Client::Transition->cft_revert_generate_manual_steps(
        command_interface => $Client,
        "project-name"    => $Project_name
    );
Runs 'revert generate-manual-steps' operation on the given CFT project

(Class or instance method) This method is used to Run 'cft revert generate-manual-steps'
    operation.
"Returns the response object."

"Supports CLI,GUI".

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< project-name=>$Project_name >>

(Required for class method, Not Applicable for instance method)

=item C<< group-name=>$group_name >>

(Required for class method, Not Applicable for instance method)

=back

=back

=cut

sub cft_revert_generate_manual_steps {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            "project-name" => { type => SCALAR, optional => 1 },
            "group-name"   => { type => SCALAR, optional => 1 },
            "parsed-output"=> { type => SCALAR, optional => 1 },
        },
        ignore_primary_keys => 1,
    );
    my @obj= $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_cft_revert_generate_manual_steps_smtt_cli',
                interface => 'CLI',
                set       => 'SMTT',
            },
            {   method    => '_cft_revert_generate_manual_steps_smtt_gui',
                interface => 'GUI',
                set       => 'SMTT',
            },
        ],
    );
    $Log->exit() if $may_exit;
    return wantarray ? @obj : $obj[0];
} ## end sub cft_revert_generate_manual_steps

sub _cft_revert_generate_manual_steps_smtt_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my (%target, $response, @response_objs);
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%target,
        copy   => [qw (project-name) ]
    );
    $response = $apiset->transition_cft_revert_generate_manual_steps(%target);
    $Log->exit() if $may_exit;
    return $response;
} ## end sub _cft_revert_generate_manual_steps_smtt_cli

sub _cft_revert_generate_manual_steps_smtt_gui {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my (%target, $response, @response_objs);
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%target,
        copy   => [qw (project-name group-name parsed-output) ]
    );
    $response = $apiset->transition_cft_revert_generate_manual_steps(%target);
    $Log->exit() if $may_exit;
    return $response;
} ## end sub _cft_revert_generate_manual_steps_smtt_gui

=head2 cft_revert_verify7_mode

    $resp = NACL::C::Client::Transition->cft_revert_verify7_mode(
        command_interface => $Client,
        "project-name"    => $Project_name
    );
Runs 'revert verify7-mode' operation on the given CFT project

(Class or instance method) This method is used to Run 'cft revert verify7-mode'
    operation.
"Returns the response object."

"Supports CLI,GUI".

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< project-name=>$Project_name >>

(Required for class method, Not Applicable for instance method)

=item C<< group-name=>$group_name >>

(Required for class method, Not Applicable for instance method)

=back

=back

=cut

sub cft_revert_verify7_mode {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            "project-name" => { type => SCALAR, optional => 1 },
            "group-name"   => { type => SCALAR, optional => 1 },
            "parsed-output"=> { type => SCALAR, optional => 1 },
        },
        ignore_primary_keys => 1,
    );
    my @obj= $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_cft_revert_verify7_mode_smtt_cli',
                interface => 'CLI',
                set       => 'SMTT',
            },
            {   method    => '_cft_revert_verify7_mode_smtt_gui',
                interface => 'GUI',
                set       => 'SMTT',
            },
        ],
    );
    $Log->exit() if $may_exit;
    return wantarray ? @obj : $obj[0];
} ## end sub cft_revert_verify7_mode

sub _cft_revert_verify7_mode_smtt_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my (%target, $response, @response_objs);
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%target,
        copy   => [qw (project-name) ]
    );
    $response = $apiset->transition_cft_revert_verify7_mode(%target);
    $Log->exit() if $may_exit;
    return $response;
} ## end sub _cft_revert_verify7_mode_smtt_cli

sub _cft_revert_verify7_mode_smtt_gui {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my (%target, $response, @response_objs);
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%target,
        copy   => [qw (project-name group-name parsed-output) ]
    );
    $response = $apiset->transition_cft_revert_verify7_mode(%target);
    $Log->exit() if $may_exit;
    return $response;
} ## end sub _cft_revert_verify7_mode_smtt_gui

=head2 cft_revert_abort

    $resp = NACL::C::Client::Transition->cft_revert_abort(
        command_interface => $Client,
        "project-name"    => $Project_name
    );
Runs 'revert abort' operation on the given CFT project

(Class or instance method) This method is used to Run 'cft revert abort'
   operation.
"Returns the response object."

"Supports CLI".

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< project-name => $name >>

(Required) CFT Project name

=back

=back

=cut

sub cft_revert_abort {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            "project-name" => { type => SCALAR, optional => 0 },
        },
        ignore_primary_keys => 1,
    );
    my @obj= $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_cft_revert_abort_smtt_cli',
                interface => 'CLI',
                set       => 'SMTT',
            },
            {   method    => '_cft_revert_abort_smtt_gui',
                interface => 'GUI',
                set       => 'SMTT',
            },
        ],
    );
    $Log->exit() if $may_exit;
    return wantarray ? @obj : $obj[0];
} ## end sub cft_revert_abort

sub _cft_revert_abort_smtt_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my (%target, $response, @response_objs);
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%target,
        copy   => [qw (project-name) ]
    );
    $response = $apiset->transition_cft_revert_abort(%target);
    $Log->exit() if $may_exit;
    return $response;
} ## end sub _cft_revert_abort_smtt_cli

sub _cft_revert_abort_smtt_gui {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my (%target, $response, @response_objs);
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%target,
        copy   => [qw (project-name) ]
    );
    $response = $apiset->transition_cft_revert_abort(%target);
    $Log->exit() if $may_exit;
    return $response;
} ## end sub _cft_revert_abort_smtt_gui


sub cft_confirm_execution_revert_manual_steps {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            "project-name"       => { type => SCALAR , optional => 1},
            "group-name"         => { type => SCALAR , optional => 1},
            'method-timeout'     => { type => SCALAR , optional => 1},
            'seven-mode'         => { type => SCALAR , optional => 1},
            'cabling'            => { type => SCALAR , optional => 1},
            'cluster-mode'       => { type => SCALAR , optional => 1},
            'parsed-output'      => { type => SCALAR , optional => 1},
        },
        ignore_primary_keys => 1,
    );
    my @obj= $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_cft_confirm_execution_revert_manual_steps_smtt_gui',
                interface => 'GUI',
                set       => 'SMTT',
            },
        ],
    );
    $Log->exit() if $may_exit;
    return wantarray ? @obj : $obj[0];
} ## end sub cft_confirm_execution_revert_manual_steps

sub _cft_confirm_execution_revert_manual_steps_smtt_gui {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my (%target, $response, @response_objs);
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%target,
        copy   => [qw (project-name group-name parsed-output
                       seven-mode cabling cluster-mode) ]
    );
    $response = $apiset->transition_cft_confirm_execution_revert_manual_steps(%target);
    $Log->exit() if $may_exit;
    return $response;
} ## end sub _cft_confirm_execution_revert_manual_steps_smtt_gui

=head2 cft_edit_config

    $resp = NACL::C::Client::Transition->cft_edit_config(
        command_interface => $Client,
        "project-name"    => $Project_name,
        "plan_configuration"=>$plan_configuration
    );
Runs 'cft_edit_config' operation on the given CFT project

(Class or instance method) This method is used to edit  cft plan configuration page

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< project-name=>$Project_name >>

(Required for class method, Not Applicable for instance method)

=item C<< group-name=>$group_name >>

(Required for class method, Not Applicable for instance method)

=item C<< plan_configuration=>$plan_config >>

(Required for class method, Not Applicable for instance method)

=back

=back

=cut

sub cft_edit_config {
$Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            "plan_configuration"        => { type => HASHREF,  optional => 1 },
            _optional_scalars qw(
                group-name project-name
            ),
        },
    );
    my $obj= $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_editconfig_cft_smtt_gui',
                interface => 'GUI',
                set       => 'SMTT',
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $obj;
} ## end sub cft_edit_config

sub _editconfig_cft_smtt_gui {
    $Log->enter() if $may_enter;
    print "Entered _editconfig_cft_smtt_gui\n";
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my %target;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%target,
        copy    => [qw( project-name group-name plan_configuration)]
    );
    print Dumper(\%target);
    $apiset->transition_cft_editconfig(%target);
    $Log->exit() if $may_exit;
} ## end sub _editconfig_cft_smtt_gui




=head2 edit_tabs

    my $res = NACL::C::Client::Transition->edit_tabs(
            "method-timeout"            => $METHOD_TIMEOUT,
            command_interface           => $HostWindows,
            "project-name"              => $PROJECT_NAME,
            "c-mode-host-name"          => "10.237.222.73",
            "goto_tab"                  => "CFT SVM Mapping",
            "svm-name"                  => "gui_vserver",
            "remove_schedule"           => "sc1",
            "networking_edit"           => \%netw,
            "volumepair_add_args"       => \%volume_pri,
            "cluster_volume_map_args"   => \%vol_map,
            schedule_edit_args          => \@schedule,
            "data-copy-ipaddress"       => $Host7Mode[0]->{DATA_IP2},
            "apiset_should"             => { interface => $INTERFACE },
            S_mode_credentials          => \%S_mode_credentials,
            C_mode_credentials          => \%C_mode_credentials,
            S_mode_credentials_HA       => \%S_mode_credentials_HA,
            cft_clusterinfo_add         => \%cft_clusterinfo_add,
            cft_svm_mapping_add_pri     => \%cft_svmmapping_add_pri,
            cft_svm_mapping_add_sec     => \%cft_svmmapping_add_sec,
            cft_nodemapping_set         => \%cft_nodemapping_set,
    );

    my %schedule1 = (
        "schedule-name"                 => $SCHEDULE_NAME,
        "days-range"                    => $DAYS_RANGE,
        "start-time"                    => $START_TIME,
        "duration"                      => $DURATION,
        "available-transfers-percentage"=> $AVAILABLE_TRANSFERS_PERCENTAGE,
        "session-snapmirror-throttle"   => $SESSION_SNAPMIRROR_THROTTLE,
        "update-frequency"              => $UPDATE_FREQUENCY,
    );
    my @schedule = ( \%schedule1 );

    my %volume_pri=(
        "volume-name-add"     => \@arr2,
        "volume-name-7-mode"  => \@vo,
        "aggregate-name"      => \@agg,
        "volume-name-c-mode"  => \@vol,
    );
    my %netw = (
        "ip-address"  => "10.237.37.85",
        "default-gateway" => "10.237.37.1",
        "netmask"      => "255.255.255.0",
        "home-node"    => "sti8060-268",
        "home-port"    => "e0j",
    );
    %vol_map = (
        $CLUS_AGGR    => [ $7_VOL.'1', $7_VOL.'2' ],
        $CLUS_AGGR.'1'        => [ $7_VOL.'3' ],
        $CLUS_AGGR.'2'       => [ $7_VOL.'4'],
    );

(Class method) This method is used to create the transition session .

"Supports CLI, GUI".

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< project-name => $name >>

(Required) Name of the transition session .

=item C<< 7-mode-host-name => $host_7 >>

(Required) Hostname or IP address of 7mode .

=item C<< goto_tab => $tab >>

(Required) goto_tab

=item C<< data-copy-ipaddress => $data_copy_ip >>

(Required) Data copy IP address

=item C<< vfiler => $vfiler >>

(Required) Name of the vfiler from which volume will be transitioned .

=item C<< multipathing-ipaddress => $multipath >>

(Optional) Multipathing IP address .

=item C<< c-mode-host-name => $host_C >>

(Required) Hostname or IP address of Cmode .

=item C<< svm-name => $vser >>

(Required) Name of the vserver to which volumes will be transitioned .

=item C<< svm-rootvol-aggr => $vser_rootvol_aggr >>

(Optional) Vserver root volume aggregate name .

=item C<< remove_schedule => $sched >>

(Optional) schedule name which has to be deleted .

=item C<< schedule_edit_args => \%sched_edit >>

(Optional) inputs for editing schedule .

=item C<< networking_edit => %\netw >>

(Optional) input for network lif edit .

=item C<< S_mode_credentials => %\cred >>

(Optional) 7Mode credentials .

=item C<< C_mode_credentials => %\cred >>

(Optional) CMode credentials .

=item C<< cft_nodemapping_set => %\cluster_map >>

(Optional) CFT cluster map input .

=item C<< cft_svm_mapping_add_pri => %\svmmap_pri >>

(Optional) input for primary 7Mode vfiler map .

=item C<< cft_svm_mapping_add_sec => %\svmmap_sec >>

(Optional) input for secondary 7Mode vfiler map .

=back

=back

=cut

sub edit_tabs {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
	print Dumper("In Component");
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            "S_mode_credentials"        => { optional => 1 },
            "C_mode_credentials"        => { optional => 1 },
            "S_mode_credentials_HA"     => { optional => 1 },
            "group-name"                => { optional => 1 },
            "cft_clusterinfo_add"       => { optional => 1 },
            "cft_svm_mapping_add_pri"   => { optional => 1 },
            "cft_svm_mapping_add_sec"   => { optional => 1 },
            "cft_nodemapping_set"       => { optional => 1 },
            "volumepair_add_args"       => { type => HASHREF,  optional => 1 },
            "cluster_volume_map_args"   => { type => HASHREF,  optional => 1 },
            "mount_policy"              => { type => SCALAR,   optional => 1 },
            "schedule_add_args"         => { type => ARRAYREF, optional => 1 },
            "schedule_edit_args"        => { type => ARRAYREF, optional => 1 },
            "seven-mode-only"           => { type => SCALAR,   optional => 1 },
            "plan_configuration"        => { type => HASHREF,  optional => 1 },
            "goto_tab"                  => { type => SCALAR},
            "remove_schedule"           => { type => SCALAR},
            "networking_edit"           => { type => HASHREF,  optional => 1},
            "networking_add"            => { type => HASHREF,  optional => 1},
            "networking_delete"         => { type => HASHREF,  optional => 1},
            "add_smode_trans_lif"       => { type => HASHREF,  optional => 1},
            _optional_scalars qw(
                7-mode-host-name data-copy-ipaddress session-type
                vfiler multipathing-ipaddress c-mode-host-name svm-name
                svm-rootvol-aggr remove_schedule
            ),
        },
    );
    my $obj;
    try{
        $obj = $pkg->call_on_apiset(
            %opts,
            choices => [
                {   method    => '_edit_tabs_smtt_gui',
                    interface => 'GUI',
                    set       => 'SMTT',
                },
            ],
        );
    } catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = $_[0];
        NACL::C::Exceptions::Client::Transition::AlreadyExists
            ->detect_convert_and_throw(exception => $exception);
        $Log->exit() if $may_exit;
        $exception->throw();
    };
    $Log->exit() if $may_exit;
    return $obj;
} ## end sub edit_tabs

sub _edit_tabs_smtt_gui {
    $Log->enter() if $may_enter;
    my ($pkg,@args) = @_;
    my %opts = $pkg->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my %target;
    $pkg->_hash_copy(
        source => \%opts,
        target => \%target,
        copy   => [qw( project-name  7-mode-host-name vfiler goto_tab group-name
            multipathing-ipaddress c-mode-host-name svm-name session-type
            svm-rootvol-aggr data-copy-ipaddress S_mode_credentials
            C_mode_credentials volumepair_add_args schedule_add_args seven-mode-only
            plan_configuration remove_schedule schedule_edit_args networking_edit
            networking_add networking_delete add_smode_trans_lif S_mode_credentials_HA
            cft_clusterinfo_add cft_svm_mapping_add_pri cft_svm_mapping_add_sec
            cft_nodemapping_set cluster_volume_map_args mount_policy)
        ]
    );
    my $response = $apiset->transition_edit_tabs(%target);
    $Log->exit() if $may_exit;
    return $response;
} ## end sub _edit_tabs_smtt_gui

=head2 move_project_group

    $resp = NACL::C::Client::Transition->move_project_group(
        command_interface => $Client,
        "project-name"    => $Project_name,
        "group-name"      => $group_name,
        "new-group"       => $new_group,
    );
Runs 'move_project_group' operation on the given CFT/CBT project

(Class or instance method) This method is used to click 'move_project_group'
   button.
"Returns the response object."

"Supports GUI".

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< project-name => $name >>

(Required) CFT Project name

=item C<< group-name => $group_name >>

(Required) CFT Project group name

=item C<< new-group => $new_name >>

(Required) CFT Project new group name

=back

=back

=cut

sub move_project_group {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            "project-name" => { type => SCALAR, optional => 0 },
            "group-name"   => { type => SCALAR, optional => 0 },
            "new-group"    => { type => SCALAR, optional => 0 },
        },
        ignore_primary_keys => 1,
    );
    my @obj= $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_move_project_group_smtt_gui',
                interface => 'GUI',
                set       => 'SMTT',
            },
        ],
    );
    $Log->exit();
    return wantarray ? @obj : $obj[0];
} ## end sub move_project_group

sub _move_project_group_smtt_gui {
    $Log->enter();
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my (%target, $response, @response_objs);
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%target,
        copy   => [qw (project-name group-name new-group) ]
    );
    $response = $apiset->transition_move_project_group(%target);
    $Log->exit();
    return $response;
} ## end sub _move_project_group_smtt_gui

=head2 volume_mount_modify

    $resp = NACL::C::Client::Transition->volume_mount_modify(
        command_interface => $Client,
        "project-name"    => $Project_name,
        "mount_policy"    => $policy,
    );
Runs 'volume_mount_modify' operation on the given CFT/CBT project

(Class method) This method is used to click/select 'volume mount policy' list 
 in the project dashboard.
"Returns the response object."

"Supports GUI".

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< project-name => $name >>

(Required) CFT/CBT Project name

=item C<< group-name => $group_name >>

(Required) CBT/CFT Project group name

=item C<< mount_policy => $policy >>

(Required) Project CMode volume mount policy

=back

=back

=cut

sub volume_mount_modify {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            "project-name"  => { type => SCALAR, optional => 0 },
            "group-name"    => { type => SCALAR, optional => 1 },
            "mount_policy"  => { type => SCALAR },
        },
        ignore_primary_keys => 1,
    );
    my @obj= $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_volume_mount_modify_smtt_gui',
                interface => 'GUI',
                set       => 'SMTT',
            },
        ],
    );
    $Log->exit();
    return wantarray ? @obj : $obj[0];
} ## end sub volume_mount_modify

sub _volume_mount_modify_smtt_gui {
    $Log->enter();
    my ($pkg_or_obj,@args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = delete $opts{apiset};
    my (%target, $response, @response_objs);
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%target,
        copy   => [qw (project-name group-name mount_policy) ]
    );
    $response = $apiset->transition_volume_mount_modify(%target);
    $Log->exit();
    return $response;
} ## end sub _volume_mount_modify_smtt_gui

1;
