#
# Copyright (c) 2001-2017 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Snapmirror Component Module
## @author benjaram@netapp.com, Chethan.M@netapp.com, pragyan@netapp.com, dl-nacl-dev@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::C::Snapmirror

=head1 DESCRIPTION

C<NACL::C::Snapmirror> is a derived class of
L<NACL::C::Component::ONTAP|lib-NACL-C-Component-ONTAP-pm>.  It represents access
to an ONTAP Snapmirror on a volume.  A related class is
L<NACL::CS::Snapmirror|lib-NACL-CS-Snapmirror-pm>, which represents
the past state of an ONTAP Snapmirror.

=head1 ATTRIBUTES

=head2 command_interface

See L<NACL::C::Component::command_interface|lib-NACL-C-Component-pm/command_interface>

=head2 destination_path

Destination path of the volume

=head2 source_path

Source path of the volume

=head2 source_ci

Represents command interface of the source cluster or node.
This option is required if you are having intercluster snapmirror relationship
and if volume object is not passed for source-path. Required when the vserver
command_interface is passed to the method rather than node ci.
NOTE: source_ci is required in places where source-volume, source-vserver and
source-cluster parameters are passed to method call rather than source-path. This
is required since construction of source-path needs source node's version to
determine if the path should be in RR or SN style.

See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>.

=head2 is_granular

This is a boolean flag that indicates whether the relationship represented
by the object is a granular relationship or not. This cannot be explicitly set
by the user when creating a relationship, it is set internally based on the
logic around granular relationships.

=cut

package NACL::C::Snapmirror;

use strict;
use warnings;
use feature 'state';

# This block is a hack to make "perl -wc" work despite the circular dependency
# on this module:
BEGIN {
    use File::Spec;
    $INC{'NACL/C/Snapmirror.pm'} ||= File::Spec->rel2abs(__FILE__);
}

use NACL::Global;
use NACL::CS::Snapmirror;
use NACL::ComponentUtils qw (_optional_scalars Dumper);
use NACL::C::Exceptions::Snapmirror::TransferInProgress;
use NACL::C::Exceptions::Snapmirror::DoesNotExist;
use NACL::C::Exceptions::Snapmirror::SnapshotInUse;
use NACL::C::Exceptions::Snapmirror::RelationshipInfoBeingUpdated;
use NACL::C::Exceptions::Snapmirror::DeleteIsPending;
use NACL::C::Exceptions::Snapmirror::RelationshipInfoBeingPropagated;
use NACL::C::Exceptions::Snapmirror::VolumeNotInitializedorRestricted;
use NACL::C::Exceptions::Snapmirror::SubsystemNotFullyOperational;
use NACL::C::Exceptions::Snapmirror::BreakNotAllowedForLS;
use NACL::C::Exceptions::Snapmirror::VolumeIsOffline;
use NACL::C::Exceptions::Snapmirror::DestinationAlreadyExists;

use base 'NACL::C::Component::ONTAP';
use NACL::C::Job;
use Params::Validate qw(:all);
use NATE::Log qw(log_global);
my $Log = log_global();
use NATE::BaseException qw(:try);
use NACL::APISet::Exceptions::ResponseException;
use NATE::Exceptions::Argument qw(:try);
use NACL::Exceptions::InvalidChoice;
use NATE::ParamSet qw(param_global);
use NACL::CS::VserverPeer;
use NACL::C::Cluster;
use NACL::C::Vserver;
use NACL::C::Volume;

use NACL::C::_Mixins::Common qw(:iv_methods);
use NACL::C::_Mixins::Snapmirror qw(:all);

use NACL::C::Exceptions::Volume::AlreadyRestricted;
use NACL::C::Exceptions::Snapmirror::VolumeIsOffline                ();
use NACL::C::Exceptions::Snapmirror::CannotConnectToSourceFiler     ();
use NACL::C::Exceptions::Snapmirror::DataProtectionVolume           ();
use NACL::C::Exceptions::Snapmirror::NoReleasableDestinationFound   ();
use NACL::C::Exceptions::Snapmirror::NotQuiescedOrQuiescing         ();
use NACL::C::Exceptions::Snapmirror::NoTransferToAbort              ();
use NACL::C::Exceptions::Snapmirror::ServiceNotEnabledOnSource      ();
use NACL::C::Exceptions::Snapmirror::SnapmirrorDpLicenseCheckFailed ();
use NACL::C::Exceptions::Snapmirror::SnapmirrorIsTurnedOff          ();
use NACL::C::Exceptions::Snapmirror::TransferFromSourceNotPossible  ();
use NACL::C::Exceptions::Snapmirror::PolicyLookupFailed             ();
use NACL::C::Exceptions::Snapmirror::BreakNotAllowedForVault        ();
use NACL::C::Exceptions::Snapmirror::StaleCache                     ();
use NACL::C::Exceptions::Snapmirror::SystemInitialization           ();
use NACL::C::Exceptions::Snapmirror::NotGranular                    ();

=head1 METHODS

=head2 new

my $snapmirror = NACL::C::Snapmirror->new(command_interface=>$ci,
                                         source-path=>$source_path,
                                         destination-path=>$destination_path);

Return a new Snapmirror component object with the given attributes.

=cut

use Class::MethodMaker [
    scalar => 'destination_path',
    scalar => 'source_path',
    scalar => [{-type => 'NACL::C::CommandInterface::ONTAP'}, 'source_ci'],

    # Not a primary key, but used to determine which ZAPI to invoke.
    # Hence, is a private attribute and is undocumented.
    scalar => '_is_infinivol',

    # This is a new flag, to identify snapmirror relationships that are
    # granular. It is not a primary key. It is a private attribute that should
    # be accessed by the is_granular() method.
    scalar => [
        {
            -default_ctor => sub {
                my $self = shift;
                my $class = ref $self;

                # This basically computes "is_granular" and caches it on
                # $self. We force the call to be a static call so that
                # is_granular() doesn't try to use the (not-yet-set) attribute.
                return $class->is_granular(
                    'source-path'      => $self->source_path,
                    'destination-path' => $self->destination_path
                );
            },
        },
        '_is_granular',
    ],
];

=head2 create

  my $snapmirror  = NACL::C::Snapmirror->create(command_interface=>$ci,
                                        "destination-vserver"=>$vserver,
                                        "destination-volume"=>$volume,
                                        "source-vserver"=>$vserver,
                                        "source-volume"=>$volume,
                                        type=>$mirrortype);
  (or)

  my $snapmirror = NACL::C::Snapmirror->create(command_interface=>$ci,
                                        source-path=>$source_path,
                                        destination-path=>$destination_path);

  (or)

  my $snapmirror = NACL::C::Snapmirror->create(command_interface=>$ci,
                                        source-vserver=>$vserver,
                                        source-volume=>$volume,
                                        destination-path=>$destination_path);


(Class method) Request creation of a new snapmirror element, and return a new
snapmirror component object that refers to it.

NOTE: This method B<always> returns a new object, even if called as an instance
method (i.e. for adding elements to an existing consistency group).

Uses a CMode CLI/ZAPI or 7Mode-CLI/ZAPI APISet.

Invokes 'snapmirror-create' API for CMode ZAPI.

Invokes 'snapmirror-initialize' API for 7Mode ZAPI.

Invokes 'snapmirror initialize' command for 7Mode CLI.

=over

=item NACL_SVMDR_XDP

Global flag that affect the behavior of snapmirror create command
If this param is set, then all the snapmirror relation created would be of type xdp.

=over

=item NACL_SVMDR_MIRRORVAULT

Global flag that affect the behavior of snapmirror create command
If this param is set in conjunction with NACL_SVMDR_XDP, then all the snapmirror relation of type XDP created will be set with policy MirrorAndVault.
This param functions only when NACL_SVMDR_XDP is set to 1
if this param is not set, then default policy used for XDP relation will be "MirrorAllSnapshot" 

=over

=item NACL_SNAPMIRROR_DP  

(Optional, defaults to 1)

Global flag that affect the behavior of snapmirror create command
If this param is set, then all the snapmirror relation created would be of type dp.
Applicable to c-mode only.

=over

=item NACL_SNAPMIRROR_DP_OPTION

(Optional, defaults to 1)

Global flag that affect the behavior of snapmirror create command
If this param is set in conjunction with NACL_SNAPMIRROR_DP, then the option replication.create_data_protection_rels.enable will be turned on.  
If this param is not set, then the option option replication.create_data_protection_rels.enable will not be turned on.
This param should be off for parallel tests.
Applicable to c-mode only.

=over


=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "destination-cluster"=>$string >>

(Optional) Name of the Cluster where the destination volume resides

Applicable for CMode CLI/ZAPI.

CM ZAPI:

destination-path input would be constructed from this input in the form (RR: cluster://volume/vserver, SN and above: not applicable) and passed to destination-location option.

=item C<< "destination-vserver"=>$string >>

(Optional) Name of the Vserver  where the destination volume resides

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

destination-path input would be constructed from this input in the form
(RR: cluster://volume/vserver, SN and above: vserver:volume) and passed to
'destination-location' option.

=item C<< "destination-volume" =>$string >>

(Optional) Name of the destination volume  which will be in snapmirror relationship

Applicable for CMode CLI/ZAPI, 7Mode CLI/ZAPI.

Maps to:

7M CLI:

Since this option is not passed to the apilayer call ,there is no mapping
information provided ,destination-path will be constructed using this argument
and passed to the apilayer call.

7M ZAPI:

Since this option is not passed to the apilayer call ,there is no mapping
information provided ,destination-location will be constructed using this argument
and passed to the apilayer call.

CM ZAPI:

destination-path input would be constructed from this input in the form
(RR: cluster://volume/vserver, SN and above: vserver:volume) and passed to
'destination-location' option.

=item C<< destination-cg =E<gt> $string >>

(Optional) Name of destination consistency group for a granular relationship.

Applicable only to CMode CLI/ZAPI.

This option does not map to an API-layer option. If passed (and if
C<destination-path> is not already specified), this is used with the
C<destination-vserver> value to form the destination path as
C<$vserver:/cg/$cg_name>. If this parameter is present, it is given
precedence over C<destination-volume> (but only for CMode).

=item C<< "source-cluster" =>$string >>

(Optional) Name of cluster where the source volume resides

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

source-path input would be constructed from this input in the form
(RR: cluster://volume/vserver, SN and above: not applicable) and passed to
source-location option.

=item C<< "source-vserver" =>$string >>

(Optional) Name of the vserver where the source volume resides

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

source-path input would be constructed from this input in the form
(RR: cluster://volume/vserver, SN and above: vserver:volume) and passed to
'source-location' option.

=item C<< "source-volume" =>$string >>

(Optional) Name of the source volume for which the snapmirror relation is created

Applicable for CMode CLI/ZAPI, 7Mode CLI/ZAPI.

Maps to:

7M CLI:

Since this option is not passed to the apilayer call ,there is no mapping
information provided ,source-path will be constructed using this argument
and passed to the apilayer call.

7M ZAPI:

Since this option is not passed to the apilayer call ,there is no mapping
information provided ,source-location will be constructed using this argument
and passed to the apilayer call.

CM ZAPI:

source-path input would be constructed from this input in the form
(RR: cluster://volume/vserver, SN and above: vserver:volume) and passed to
'source-location' option.

=item C<< source-cg =E<gt> $string >>

(Optional) Name of source consistency group for a granular relationship.

Applicable only to CMode CLI/ZAPI.

This option does not map to an API-layer option. If passed (and if
C<source-path> is not already specified), this is used with the
C<source-vserver> value to form the source path as C<$vserver:/cg/$cg_name>. If
this parameter is present, it is given precedence over C<source-volume> (but
only for CMode).

=item C<< cg-item-mappings =E<gt> $hashref >>

(Optional) A hash-reference containing one or more source-to-destination
mappings of consistency group items for a granular relationship.

Applicable only to CMode CLI/ZAPI.

If this option is given, then the source and destination specifications
(whether by path or by parts) must refer to a consistency group. The mappings
given by this option are added to the consistency group specified by the
source and destination paths.

When present, this option must specify a hash-reference. Each key of the hash
will represent the source item path, and the value it points to will be the
destination path. Each of C<source-path> and C<destination-path> are of the
form:

    "/vol/" volume-name "/" file-or-lun-name

That is, the sequence C</vol/> followed immediately by a volume-name and that
followed immediately by the sequence C</>. Following the third C</> is a
file or LUN name. Currently, all CG items must be specified by full path.

=item C<< type=>$string >>

Applicable for CMode CLI/ZAPI

Maps to:

CM ZAPI: relationship-type

(Optional) The type of snapmirror relationship that will be created between the
source and destination volumes.

=item C<< "source-snapshot"=>$string >>

(Optional) name of the source snapshot

Applicable for 7Mode CLI/ZAPI

Maps to:

7M CLI: -s <value> (Programmatic name: "source-snapshot")

7M ZAPI: source-snapshot

=item C<< "destination-snapshot"=>$string >>

(Optional) name of the destination snapshot

Applicable for 7Mode CLI/ZAPI

Maps to:

7M CLI: -c <value> (Programmatic name: "destination-snapshot")

7M ZAPI: destination-snapshot

=item C<< vserver=>$string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: vserver

=item C<< schedule=>$string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: schedule

=item C<< tries=>$number >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: tries

=item C<< throttle=>$number >>

Applicable for CMode CLI/ZAPI,7Mode CLI/ZAPI

Maps to:

CM ZAPI: max-transfer-rate

7M CLI : -k (Programmatic name: max-transfer-rate)

7M ZAPI: max-transfer-rate

=item C<< "source-path"=>$string >>

Applicable for CMode CLI/ZAPI,7Mode CLI/ZAPI.

Maps to:

CM ZAPI: source-location

7M CLI : -S (Programmatic name: source-location)

7M ZAPI: source-location

=item C<< "destination-path"=>$string >>

Applicable for CMode CLI/ZAPI,7Mode CLI/ZAPI

Maps to:

CM ZAPI: destination-location

7M CLI: destination location (Programmatic name: "destination-location")

7M ZAPI: destination-location

=item C<< "job_component"=> \$scalar >>

Applicable for CMode CLI

=item C<< 'hidden-type' => $string >>

Applicable for CMode CLI

=item C<< 'dest-volume-msid' => $string >>

Applicable for CMode CLI

=item C<< 'dest-vserver-id' => $string >>

Applicable for CMode CLI

=item C<< 'policy' => $string >>

Applicable for CMode CLI/ZAPI

=item C<< 'relationship-type' => $string >>

Applicable for CMode CLI

=item C<< 'dest-volume-dsid' => $string >>

Applicable for CMode CLI

=item C<< 'dest-cluster-id' => $string >>

Applicable for CMode CLI

=item C<< 'dest-vserver-uuid' => $string >>

Applicable for CMode CLI

=item C<< 'extended_query' => $string >>

(Optional)

Applicable for CMode CLI

=item C<< 'identity-preserve' => $string >>

Identity Preserve Vserver DR possible value(s) are, true,false
Applicable for CMode CLI/ZAPI

=item C<< 'newDstMsId' => $string >>

Dest MSID after LS Delete
Applicable for CMode CLI

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=item C<NACL::C::Exceptions::Snapmirror::SnapmirrorDpLicenseCheckFailed>

This exception gets thrown if an attempt is made to create when
snapmirror DP license check failed for the cluster.

=item C<NACL::C::Exceptions::Snapmirror::SnapmirrorIsTurnedOff>

This exception gets thrown if an attempt is made to create snapmirror
when it is turned off.

=item C<NACL::C::Exceptions::Snapmirror::CannotConnectToSourceFiler>

This exception gets thrown if an attempt is made to create when
we cannot connect to the source filer.

=item C<NACL::C::Exceptions::Snapmirror::ServiceNotEnabledOnSource>

This exception gets thrown if an attempt is made to create when
snapmirror is not enabled on source.

=item C<NACL::C::Exceptions::Snapmirror::TransferFromSourceNotPossible>

This exception is thrown when snapmirror transfer not possible, snapmirror
may be misconfigured.

=item C<NACL::C::Exceptions::Snapmirror::NoVserverPeerRelationship>

This exception is raised when snapmirror is created without the vserver peering done for the
source and destination vservers.

=item C<NACL::C::Exceptions::Snapmirror::PolicyLookupFailed>

This exception is raised when snapmirror is created with an invalid policy

=item C<NACL::C::Exceptions::Snapmirror::SubsystemNotFullyOperational>

This exception is thrown when an attempt is made to perform snapmirror
create operation when SnapMirror subsystem is not fully operational.

=item C<NACL::C::Exceptions::Snapmirror::DestinationAlreadyExists>

This exception is raised when snapmirror is created and the destination is already in use.

=back

=back

=cut

sub create {
    my $pkg_or_obj = shift;
    $Log->enter();
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@_,
        additional_spec => {
            job_component => {type => SCALARREF, optional => 1},
            _optional_scalars(
                qw(source-cluster source-vserver source-volume source-cg
                   destination-cluster destination-vserver destination-cg
                   destination-volume type vserver schedule
                   tries throttle source-snapshot destination-snapshot
                   hidden-type dest-volume-msid dest-vserver-id policy
                   relationship-type dest-volume-dsid dest-cluster-id
                   dest-vserver-uuid identity-preserve newDstMsId)
            ),
            'cg-item-mappings' => { type => HASHREF | ARRAYREF, optional => 1 },
        },
        cli_cmd => 'snapmirror create',
    );

    if ($opts{'source-cluster'} || $opts{'destination-cluster'})
    {
        $Log->warn(
            'source-cluster and destination-cluster are no longer valid.'
        );
        delete $opts{'source-cluster'};
        delete $opts{'destination-cluster'};
    }
    my $global_relation_type = param_global->get('NACL_SVMDR_XDP');
    my $global_policy_type = param_global->get('NACL_SVMDR_MIRRORVAULT');
    

    if ( $global_relation_type) {
        $Log->comment("snapmirror relations of type XDP  would be created because"
            . " the parameter 'NACL_SVMDR_XDP' was provided");

        $opts{'type'} = 'xdp';
        if ($global_policy_type) {
            $Log->comment("snapmirror relations of type XDP attached with policy MirrorAndVault");
            $opts{'policy'} = 'MirrorAndVault';
        }
    }

    my $destination_ci = $opts{command_interface};
    my $dst_Api_Set_Obj  = $destination_ci->apiset(interface => "CLI");
    my $dst_vol_obj = $opts{'destination-path'};
    $dst_vol_obj =~ s/(.*):(.*)/$1 $2/;
    my $dst_obj;
    my %tmp_hash;
    $tmp_hash{'vserver'} = $1 if (defined $1 && $1 ne "");
    $tmp_hash{'volume'} = $2 if (defined $2 && $2 ne "");
    try {
       $dst_obj = NACL::C::Volume->find(
           command_interface => $destination_ci,
           filter            => \%tmp_hash,
           allow_empty => 1,
       );
    }
    catch NACL::APISet::Exceptions::ResponseException with {
       $Log->comment('volume show fails');
    };
    if ($dst_obj) {
	if (! ($dst_obj->get_one_state_attribute('is-flexgroup') eq 'true' || $dst_obj->get_one_state_attribute('is-constituent') eq 'true')) {
	    if ( $destination_ci->version_manager()->is_version_ge(
                        release => $NACL::Global::ONTAP_Release_FatTire_0 ) ) {
		my $nacl_snapmirror_dp = defined param_global->get('NACL_SNAPMIRROR_DP') ? param_global->get('NACL_SNAPMIRROR_DP') : 1;
		my $nacl_snapmirror_dp_option = defined param_global->get('NACL_SNAPMIRROR_DP_OPTION') ? param_global->get('NACL_SNAPMIRROR_DP_OPTION') : 1 ;
		if ( (!defined $opts{'type'}) || ($opts{'type'} =~ /^dp/i)) {
		    if ( ($nacl_snapmirror_dp) && ($destination_ci->mode() eq 'CMode' ) ) {
			$opts{'type'} = 'dp';
			if ( $nacl_snapmirror_dp_option ) {
			    $Log->comment('UDP: option replication.create_data_protection_rels.enable is set to true for DP');
			    my $opt_value = _get_dp_option('dst_Api_Set_Obj' => $dst_Api_Set_Obj);
			    if ( $opt_value =~ /^on/ ) {
				$Log->comment('option replication.create_data_protection_rels.enable is already on');
			    }  else {
				$Log->comment('option replication.create_data_protection_rels.enable is not on, so turn it on');
			        $dst_Api_Set_Obj->execute_raw_command( command => 'options replication.create_data_protection_rels.enable on' );
			    }
			}
		    }
		}
	    }
	}
    }
    my $response;
    try {
        $response = $pkg_or_obj->get_package_name->call_on_apiset(
            %opts,
            choices => [
                {   method    => '_create_cmode_cli',
                    interface => 'CLI',
                    set       => 'CMode',
                },
                {   method    => '_create_cmode_zapi',
                    interface => 'ZAPI',
                    set       => 'CMode',
                },
                {   method    => '_create_7mode_cli',
                    interface => 'CLI',
                    set       => '7Mode',
                },
                {   method    => '_create_7mode_zapi',
                    interface => 'ZAPI',
                    set       => '7Mode',
                }
            ],
        );
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = $_[0];
        foreach my $type (
            qw(SnapmirrorDpLicenseCheckFailed SnapmirrorIsTurnedOff
               CannotConnectToSourceFiler ServiceNotEnabledOnSource
               PolicyLookupFailed TransferFromSourceNotPossible
               NoVserverPeerRelationship SubsystemNotFullyOperational DestinationAlreadyExists)
        ) {
            my $full_type = "NACL::C::Exceptions::Snapmirror::$type";
            $full_type->detect_convert_and_throw(exception => $exception);
        }
        $Log->exit();
        $exception->throw();
    };

    $Log->exit();
    return $response;
} ## end sub create

sub _create_cmode_cli {
    $Log->enter();
    my ($pkg, %opts) = @_;

    # The method "_base_backend_cmode_cli" now takes care of param validation,
    # sifting out the API-relevant parameters, etc.
    $pkg->_base_backend_cmode_cli(%opts, api => 'snapmirror_create');

    my %tmp_opts;
    $pkg->_hash_copy(
        source => \%opts,
        copy   => [ qw(source-path destination-path command_interface
                       source_ci) ],
        target => \%tmp_opts
    );
    my $obj = $pkg->new(%tmp_opts);

    $Log->exit;
    return $obj;
} ## end sub _create_cmode_cli

sub _create_cmode_zapi {
    $Log->enter();
    my ($pkg, %opts) = @_;

    my $type_map = $pkg->_type_val_map();
    if (defined $opts{type}) {
        my $type = uc $opts{type};
        if (exists $type_map->{$type}) {
            $opts{type} = $type_map->{$type};
        } else {
            $Log->warn("The value of 'type' was specified as $opts{type}, "
                    . 'but the ZAPI mapping for this value is unknown. Contact '
                    . 'dl-nacl-dev@netapp.com regarding this issue');
        }
    }

    $pkg->_base_backend_cmode_zapi(
        %opts,
        api  => 'snapmirror-create',
        copy => [ qw(tries vserver schedule policy identity-preserve
                     cg-item-mappings) ],
        map  => {
            'throttle' => 'max-transfer-rate',
            'type'     => 'relationship-type',
        }
    );

    my %tmp_opts;
    $pkg->_hash_copy(
        source => \%opts,
        copy   => [ qw(source-path destination-path command_interface
                       source_ci) ],
        target => \%tmp_opts
    );
    my $obj = $pkg->new(%tmp_opts);

    $Log->exit;
    return $obj;
} ## end sub _create_cmode_zapi

sub _create_7mode_cli {
    my $pkg = shift;
    $Log->enter();
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my %snapmirror_args;
    $pkg->_hash_copy(
        source => \%opts,
        copy   => [qw(source-snapshot destination-snapshot)],
        map    => {
            "destination-path" => "destination-location",
            "source-path"      => "source-location",
            "throttle"         => "max-transfer-rate",
        },
        target => \%snapmirror_args,
    );

    $apiset->snapmirror_initialize(%snapmirror_args);

    my $path_hash = $pkg->_construct_full_path_7m(
        'source-path'      => $snapmirror_args{'source-location'},
        'destination-path' => $snapmirror_args{'destination-location'},
        command_interface  => $command_interface,
    );

    $Log->exit();
    return $pkg->new(
        'destination_path' => $path_hash->{'destination-path'},
        'source_path'      => $path_hash->{'source-path'},
        command_interface  => $command_interface
    );
} ## end sub _create_7mode_cli

sub _create_7mode_zapi {
    my $pkg = shift;
    $Log->enter();
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my %snapmirror_args;
    $pkg->_hash_copy(
        source => \%opts,
        copy   => [qw(source-snapshot destination-snapshot)],
        map    => {
            "destination-path" => "destination-location",
            "source-path"      => "source-location",
            "throttle"         => "max-transfer-rate",
        },
        target => \%snapmirror_args,
    );

    $apiset->snapmirror_initialize(%snapmirror_args);

    my $path_hash = $pkg->_construct_full_path_7m(
        'source-path'      => $snapmirror_args{'source-location'},
        'destination-path' => $snapmirror_args{'destination-location'},
        command_interface  => $command_interface,
    );

    my $obj = $pkg->new(
        'destination_path' => $path_hash->{'destination-path'},
        'source_path'      => $path_hash->{'source-path'},
        command_interface  => $command_interface
    );

    $Log->exit();
    return $obj;
} ## end sub _create_7mode_zapi

# Check if the path input is of the form filername:volname or
# filername:qtreepath. If not, add the filername to the path
# inputs so that we refer to a specific snapmirror relationship.
# Consider the case where filer1 has snapmirror relationship set as,
# Filer1:src Filer2:dest
# Filer1:src Filer1:dest
# In this case, just storing the source and destination volume name
# alone is not enough to uniquely identify a snapmirror relationship
# in filer1.

sub _construct_full_path_7m {
    my $pkg_or_obj = shift;
    $Log->enter();
    my %opts = validate(
        @_,
        {   'source-path'      => {type => SCALAR | UNDEF},
            'destination-path' => {type => SCALAR | UNDEF},
            'command_interface' =>
                {isa => "NACL::C::CommandInterface::ONTAP"},
        }
    );
    my $filername = $opts{command_interface}->hostrec()->hostname();
    if ((defined $opts{'source-path'})
        && $opts{'source-path'} !~ /(.+):(.+)/)
    {
        $opts{'source-path'} = $filername . ":" . $opts{'source-path'};
    }
    if ((defined $opts{'destination-path'})
        && $opts{'destination-path'} !~ /(.+):(.+)/)
    {
        $opts{'destination-path'}
            = $filername . ":" . $opts{'destination-path'};
    } ## end if ( ( defined $opts{'destination-path'...
    $Log->exit();
    return \%opts;
} ## end sub _construct_full_path_7m

=head2 delete

  NACL::C::Snapmirror->delete(command_interface=>$ci,
                              "destination-vserver"=>$vserver,
                              "destination-volume"=>$volume,
                              "source-vserver"=>$vserver,
                              "source-volume"=>$volume);
  (or)

  $snapmirror_object->delete(...);

  (or)

  NACL::C::Snapmirror->delete(command_interface=>$ci,
                              source-path=>$source_path,
                              destination-path=>$destination_path);

  (or)

  NACL::C::Snapmirror->delete(command_interface=>$ci,
                              source-vserver=>$vserver,
                              source-volume=>$volume,
                              destination-path=>$destination_path);

(Class or instance method) The method deletes a snapmirror relationship

Uses a CMode CLI/ZAPI (supports async, iter and non-iter).

Invokes 'snapmirror-destroy-iter' API for iter CMode ZAPI.

For non-iter CMode ZAPI: invokes 'snapmirror-abort-async' for snapmirror
relationships between infinite volumes; invokes 'snapmirror-abort' for
other types of relationships.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "destination-cluster"=>$string >>

(Optional for class method, Not Applicable for instance method)
Name of the Cluster where the destination volume resides

Applicable for CMode CLI/ZAPI.

CM ZAPI:

destination-path input would be constructed from this input in the form (RR: cluster://volume/vserver, SN and above: not applicable) and passed to destination-location option.

=item C<< "destination-vserver"=>$string >>

(Optional for class method, Not Applicable for instance method)
Name of the Vserver  where the destination volume resides

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

destination-path input would be constructed from this input in the form
(RR: cluster://volume/vserver, SN and above: vserver:volume) and passed to
'destination-location' option.

=item C<< "destination-volume"=>$string >>

(Optional for class method, Not Applicable for instance method)
Name of the destination volume  which will be in snapmirror relationship

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

destination-path input would be constructed from this input in the form
(RR: cluster://volume/vserver, SN and above: vserver:volume) and passed to
'destination-location' option.

=item C<< destination-cg =E<gt> $string >>

(Optional) Name of destination consistency group for a granular relationship.

Applicable only to CMode CLI/ZAPI.

This option does not map to an API-layer option. If passed (and if
C<destination-path> is not already specified), this is used with the
C<destination-vserver> value to form the destination path as
C<$vserver:/cg/$cg_name>. If this parameter is present, it is given
precedence over C<destination-volume> (but only for CMode).

=item C<< "source-cluster"=>$string >>

(Optional for class method, Not Applicable for instance method)
Name of cluster where the source volume resides

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

source-path input would be constructed from this input in the form
(RR: cluster://volume/vserver, SN and above: not applicable) and passed to
source-location option.

=item C<< "source-vserver"=>$string >>

(Optional for class method, Not Applicable for instance method)

Name of the vserver where the source volume resides

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

source-path input would be constructed from this input in the form
(RR: cluster://volume/vserver, SN and above: vserver:volume) and passed to
'source-location' option.

=item C<< "source-volume"=>$string >>

(Optional for class method, Not Applicable for instance method)
Name of the source volume for which the snapmirror relation is created

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

source-path input would be constructed from this input in the form
(RR: cluster://volume/vserver, SN and above: vserver:volume) and passed to
'source-location' option.

=item C<< source-cg =E<gt> $string >>

(Optional) Name of source consistency group for a granular relationship.

Applicable only to CMode CLI/ZAPI.

This option does not map to an API-layer option. If passed (and if
C<source-path> is not already specified), this is used with the
C<source-vserver> value to form the source path as C<$vserver:/cg/$cg_name>. If
this parameter is present, it is given precedence over C<source-volume> (but
only for CMode).

=item C<< cg-item-mappings =E<gt> $hashref >>

(Optional) A hash-reference containing one or more source-to-destination
mappings of consistency group items for a granular relationship.

Applicable only to CMode CLI/ZAPI.

If this option is given, then the source and destination specifications
(whether by path or by parts) must refer to a consistency group. The mappings
given by this option are added to the consistency group specified by the
source and destination paths.

When present, this option must specify a hash-reference. Each key of the hash
will represent the source item path, and the value it points to will be the
destination path. Each of C<source-path> and C<destination-path> are of the
form:

    "/vol/" volume-name "/" file-or-lun-name

That is, the sequence C</vol/> followed immediately by a volume-name and that
followed immediately by the sequence C</>. Following the third C</> is a
file or LUN name. Currently, all CG items must be specified by full path.

=item C<< "hidden-type"=>$boolean >>

Applicable for CMode CLI

=item C<< "dest-cluster-id"=>$string >>

Applicable for CMode CLI

=item C<< "dest-vserver-id"=>$string >>

Applicable for CMode CLI

=item C<< "dest-volume-msid"=>$string >>

Applicable for CMode CLI

=item C<< "dest-volume-dsid"=>$string >>

Applicable for CMode CLI

=item C<< "source-path"=>$string >>

Applicable for CMode CLI/ZAPI

Maps to:

CM ZAPI: source-location option of snapmirror-destroy-async API

For snapmirror-destroy-iter API

<snapmirror-destroy-iter>
<query>
<snapmirror-info>
<source-location>$value</source-location>
</snapmirror-info>
</query>
</snapmirror-destroy-iter>

=item C<< "destination-path"=>$string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: destination-location option of snapmirror-destroy-async API

For snapmirror-destroy-iter API

<snapmirror-destroy-iter>
<query>
<snapmirror-info>
<destination-location>$value</destination-location>
</snapmirror-info>
</query>
</snapmirror-destroy-iter>

=item C<< force=>$boolean >>

Applicable for CMode CLI

=item C<< "job_component"=> \$scalar >>

Filled in for CMode CLI/ZAPI.

CM ZAPI: returns job_component only for infinite volume.

=item C<< 'foreground' => $string >>

Applicable for CMode CLI

=item C<< 'dest-vserver-uuid' => $string >>

Applicable for CMode CLI

=item C<< 'path' => $string >>

Applicable for CMode CLI

=item C<< 'extended_query' => $string >>

(Optional)

Applicable for CMode CLI/ZAPI.

CM ZAPI: Applicable for snapmirror-destroy-iter API.

=item C<< 'is-repository' => $string >>

B<DEPRECATED>. Use C<is_infinivol> instead.

=item C<< is_infinivol => 0 | 1 >>

(Optional, Applicable for CMode ZAPI)

This boolean option describes whether the destination volume is an infinite
volume or not.
For CMode ZAPI, the "async" form of the ZAPI must be used for the infinite
volume. If this option is not specified, the library will attempt to
determine if an infinite volume is being operated on (this could involve
querying the filer).

=item C<< is_infinivol_ref => \$scalar >>

(Optional, SCALARREF)

If the method was invoked through CMode ZAPI, the variable gets updated
describing whether it operated on an infinivol (used the "async" ZAPI) or not
(used the normal ZAPI).

=item C<< 'operation-id' => $string >>

Operation Id
Applicable for CMode CLI

=back

=back

=over

=item Exceptions

=over

=item C<NACL::C::Exceptions::Snapmirror::VolumeIsOffline>

This type of exception gets thrown if an attempt is made to delete a
snapmirror relationship of a volume that is offline.

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=item C<NACL::C::Exceptions::Snapmirror::DoesNotExist>

This exception gets thrown if an attempt is made to delete a
non-existing snapmirror.

=item C<NACL::C::Exceptions::Snapmirror::VolumeIsOffline>

This type of exception gets thrown if an attempt is made to delete a
snapmirror relationship of a volume that is offline.

=item C<NACL::C::Exceptions::Snapmirror::TransferInProgress>

This exception gets thrown if an attempt is made to break a
snapmirror which is involved in other operation.

=item C<NACL::C::Exceptions::Snapmirror::NotQuiescedOrQuiescing>

This exception gets thrown if an attempt is made to break a
snapmirror which is not quiesced.

=item C<NACL::C::Exceptions::Snapmirror::SnapshotInUse>

This exception gets thrown if an attempt is made to delete a
snapmirror when a snapshot in the volume is locked.

=item C<NACL::C::Exceptions::Snapmirror::RelationshipInfoBeingUpdated>

This exception gets thrown if an attempt is made to delete a
snapmirror when a snapmirror relation is being updated.

=back

=back

=cut

sub delete {
    my ($pkg_or_obj, @args) = @_;
    $Log->enter();
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            job_component => {type => SCALARREF, optional => 1},
            _optional_scalars(
                qw(source-cluster source-vserver source-volume source-cg
                   destination-cluster destination-vserver destination-volume
                   destination-cg hidden-type dest-cluster-id dest-vserver-id
                   dest-volume-msid dest-volume-dsid force foreground
                   dest-vserver-uuid is-repository operation-id)
            ),
            'cg-item-mappings' => { type => HASHREF | ARRAYREF, optional => 1 },
        },
        cli_cmd => 'snapmirror delete',
    );

    try {
        $pkg_or_obj->call_on_apiset(
            %opts,
            choices => [
                {   method    => '_delete_cmode_cli',
                    interface => 'CLI',
                    set       => 'CMode',
                    zapi_type => 'none',
                },
                {   method    => '_delete_cmode_zapi',
                    interface => 'ZAPI',
                    set       => 'CMode',
                    zapi_type => 'iter',
                },
                {   method    => '_delete_cmode_zapi_non_iter',
                    interface => 'ZAPI',
                    set       => 'CMode',
                    zapi_type => 'non-iter',
                    check     => '_non_iter_check',
                },
            ],
        );
    }
    catch NACL::C::Exceptions::Component with {
        # The CMode-ZAPI iter back-end might already classify the error
        # into a Component-level exception. If already classified, we can
        # simply throw this exception rather than go through the
        # detect_convert_and_throw again.
        $Log->exit();
        $_[0]->throw;
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = $_[0];
        foreach my $type (
            qw(VolumeIsOffline DoesNotExist TransferInProgress
               NotQuiescedOrQuiescing SnapshotInUse
               RelationshipInfoBeingUpdated)
        ) {
            my $full_type = "NACL::C::Exceptions::Snapmirror::$type";
            $full_type->detect_convert_and_throw(exception => $exception);
        }
        $Log->exit();
        $exception->throw();
    };

    $Log->exit();
} ## end sub delete

sub _delete_cmode_cli {
    $Log->enter();
    my ($pkg_or_obj, %opts) = @_;

    $pkg_or_obj->_base_backend_cmode_cli(%opts, api => 'snapmirror_delete');
    $Log->exit();
}

sub _delete_cmode_zapi_common_opts {
    return (api => 'snapmirror-destroy');
}

sub _delete_cmode_zapi {
    $Log->enter();
    my ($pkg_or_obj, %opts) = @_;

    $pkg_or_obj->_build_iter_backend(%opts,
        common_opts_method => '_delete_cmode_zapi_common_opts');
    $Log->exit();
}

sub _delete_cmode_zapi_non_iter {
    $Log->enter();
    my ($pkg_or_obj, %opts) = @_;

    $pkg_or_obj->_build_non_iter_backend_with_async(%opts,
        common_opts_method => '_delete_cmode_zapi_common_opts');
    $Log->exit();
}

=head2 break

  NACL::C::Snapmirror->break(command_interface=>$ci,
                             source-path=>$source_path,
                             destination-path=>$destination_path);

  (or)

  NACL::C::Snapmirror->break(command_interface=>$ci,
                             source-vserver=>$vserver,
                             source-volume=>$volume,
                             destination-path=>$destination_path);
  (or)

  $snapmirror->break(...);

  (Class or instance method) The method breaks a snapmirror relationship

Uses a CMode CLI/ZAPI (supports async, iter and non-iter)
or 7Mode CLI/ZAPI APISet.

Invokes 'snapmirror-break-iter' API for iter CMode ZAPI.

For non-iter CMode ZAPI: invokes 'snapmirror-break-async' for snapmirror
relationships between infinite volumes; invokes 'snapmirror-break' for
other types of relationships.

Invokes 'snapmirror-break' API for 7Mode ZAPI.

Invokes 'snapmirror break' command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "destination-cluster"=>$string >>

(Optional for class method, Not Applicable for instance method)
Name of the Cluster where the destination volume resides

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

destination-path input would be constructed from this input in the form (RR: cluster://volume/vserver, SN and above: not applicable) and passed to destination-location option.

=item C<< "destination-vserver"=>$string >>

(Optional for class method, Not Applicable for instance method)
Name of the Vserver  where the destination volume resides

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

destination-path input would be constructed from this input in the form
(RR: cluster://volume/vserver, SN and above: vserver:volume) and passed to
'destination-location' option.

=item C<< "destination-volume"=>$string >>

(Optional for class method, Not Applicable for instance method)
Name of the destination volume  which is in snapmirror relationship

Applicable for CMode CLI/ZAPI, 7Mode CLI/ZAPI.

Maps to:

7M CLI:

Since this option is not passed to the apilayer call ,there is no mapping
information provided ,destination-path will be constructed using this argument
and passed to the apilayer call.

7M ZAPI:

Since this option is not passed to the apilayer call ,there is no mapping
information provided ,destination-location will be constructed using this argument
and passed to the apilayer call.

CM ZAPI:

destination-path input would be constructed from this input in the form
(RR: cluster://volume/vserver, SN and above: vserver:volume) and passed to
'destination-location' option.

=item C<< "source-cluster"=>$string >>

(Optional for class method, Not Applicable for instance method)
Name of cluster where the source volume resides

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

source-path input would be constructed from this input in the form
(RR: cluster://volume/vserver, SN and above: not applicable) and passed to
source-location option.

=item C<< "source-vserver"=>$string >>

(Optional for class method, Not Applicable for instance method)
Name of the vserver where the source volume resides

Applicable for CMode CLI/ZAPI

Maps to:

CM ZAPI:

source-path input would be constructed from this input in the form
(RR: cluster://volume/vserver, SN and above: vserver:volume) and passed to
'source-location' option.

=item C<< "source-volume"=>$string >>

(Optional for class method, Not Applicable for instance method)
Name of the source volume for which the snapmirror relation is created

Applicable for CMode CLI/ZAPI, 7Mode CLI/ZAPI.

Maps to:

7M CLI:

Since this option is not passed to the apilayer call ,there is no mapping
information provided ,source-path will be constructed using this argument
and passed to the apilayer call.

7M ZAPI:

Since this option is not passed to the apilayer call ,there is no mapping
information provided ,source-location will be constructed using this argument
and passed to the apilayer call.

CM ZAPI:

source-path input would be constructed from this input in the form
(RR: cluster://volume/vserver, SN and above: vserver:volume) and passed to
'source-location' option.

=item C<< "hidden-type"=>$boolean >>

Applicable for CMode CLI

=item C<< "dest-cluster-id"=>$string >>

Applicable for CMode CLI

=item C<< "dest-vserver-id"=>$string >>

Applicable for CMode CLI

=item C<< "dest-volume-msid"=>$string >>

Applicable for CMode CLI

=item C<< "dest-volume-dsid"=>$string >>

Applicable for CMode CLI

=item C<< "source-path"=>$string >>

(Applicable for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI

Maps to:

CM ZAPI: source-location option of snapmirror-break-async API

For snapmirror-break-iter API

<snapmirror-break-iter>
<query>
<snapmirror-info>
<source-location>$value</source-location>
</snapmirror-info>
</query>
</snapmirror-break-iter>

=item C<< "destination-path"=>$string >>

Applicable for CMode CLI/ZAPI,7Mode CLI/ZAPI

Maps to:

CM ZAPI: destination-location option of snapmirror-break-async API

For snapmirror-break-iter API

<snapmirror-break-iter>
<query>
<snapmirror-info>
<destination-location>$value</destination-location>
</snapmirror-info>
</query>
</snapmirror-break-iter>

7M CLI: destination path (Programmatic name: "destination-path")

7M ZAPI: destination-location

=item C<< force=>$boolean >>

Applicable for CMode CLI, 7Mode CLI.

Maps to:

7M CLI: -f (Programmatic name: "force")

=item C<< "job_component"=> \$scalar >>

Filled in for CMode CLI/ZAPI.

CM ZAPI: returns job_component only for infinite volume.

=item C<< 'foreground' => $string >>

Applicable for CMode CLI

=item C<< 'dest-vserver-uuid' => $string >>

Applicable for CMode CLI

=item C<< 'extended_query' => $string >>

(Optional)

Applicable for CMode CLI/ZAPI.

=item C<< 'is-repository' => $string >>

B<DEPRECATED>. Use C<is_infinivol> instead.

=item C<< is_infinivol => 0 | 1 >>

(Optional, Applicable for CMode ZAPI)

This boolean option describes whether the destination volume is an infinite
volume or not.
For CMode ZAPI, the "async" form of the ZAPI must be used for the infinite
volume. If this option is not specified, the library will attempt to
determine if an infinite volume is being operated on (this could involve
querying the filer).

=item C<< is_infinivol_ref => \$scalar >>

(Optional, SCALARREF)

If the method was invoked through CMode ZAPI, the variable gets updated
describing whether it operated on an infinivol (used the "async" ZAPI) or not
(used the normal ZAPI).

=item C<< 'delete-snapshots' => $string >>

Delete Snapshots for Revert
Applicable for CMode CLI

=item C<< 'restore-destination-to-snapshot' => $string >>

Restore Destination to Snapshot Copy

Applicable for CMode CLI/ZAPI

=item C<< 'operation-id' => $string >>

Operation Id

Applicable for CMode CLI

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=item C<NACL::C::Exceptions::Snapmirror::DoesNotExist>

This exception gets thrown if an attempt is made to break a
non-existing snapmirror.

=item C<NACL::C::Exceptions::Snapmirror::TransferInProgress>

This exception gets thrown if an attempt is made to break a
snapmirror which is involved in other operation.

=item C<NACL::C::Exceptions::Snapmirror::NotQuiescedOrQuiescing>

This exception gets thrown if an attempt is made to break a
snapmirror which is not quiesced.

=item C<NACL::C::Exceptions::Snapmirror::DataProtectionVolume>

This exception gets thrown if an attempt is made to break a
snapmirror when the volume is not or must be a data-protection volume.

=item C<NACL::C::Exceptions::Snapmirror::VolumeNotInitializedorRestricted>

This exception gets thrown if an attempt is made to break an
uninitialized volume snapmirror.

=back

=back

=cut

sub break {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw(source-cluster source-vserver source-volume source-cg
                   destination-cluster destination-vserver destination-volume
                   destination-cg hidden-type dest-cluster-id dest-vserver-id
                   dest-volume-msid dest-volume-dsid force foreground
                   dest-vserver-uuid is-repository delete-snapshots
                   restore-destination-to-snapshot operation-id)
            )
        },
        cli_cmd => 'snapmirror break',
    );

    my @cmode_cli_choices = (
        {   method    => '_break_cmode_cli',
            interface => 'CLI',
            set       => 'CMode',
            zapi_type => 'none',
        },
    );

    try {
        $pkg_or_obj->call_on_apiset(
            %opts,
            choices => [
                @cmode_cli_choices,
                {   method    => '_break_cmode_zapi',
                    interface => 'ZAPI',
                    set       => 'CMode',
                    zapi_type => 'iter',
                    check     => '_break_zapi_check',
                },
                {   method    => '_break_cmode_zapi_non_iter',
                    interface => 'ZAPI',
                    set       => 'CMode',
                    zapi_type => 'non-iter',
                    check     => '_break_non_iter_check',
                },
                {   method    => '_break_7mode_cli',
                    interface => 'CLI',
                    set       => '7Mode',
                },
                {   method    => '_break_7mode_zapi',
                    interface => 'ZAPI',
                    set       => '7Mode',
                },
            ],
        );
    }
    catch NACL::C::Exceptions::Component with {

        # The CMode-ZAPI iter back-end might already classify the error
        # into a Component-level exception. If already classified, we can
        # simply throw this exception rather than go through the
        # detect_convert_and_throw again.
        $Log->exit();
        $_[0]->throw;
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = $_[0];
        # Breaking XDP relationships allowed only at test privilege, hence
        # not allowed through ZAPI. If ZAPI call fails for this reason, retry
        # over CLI. See burt 823698.
        my $vault
            = 'NACL::C::Exceptions::Snapmirror::BreakNotAllowedForVault';
        my $passed = 0;
        if ($vault->detect(exception => $exception)) {
            if ($exception->is_zapi()) {
                # Retry only for ZAPI. If it failed for CLI, then there
                # wouldn't be any point retrying again over CLI.
                $Log->comment('Breaking a relationship of type XDP is not '
                        . 'possible through ZAPI. Retrying over CLI');
                try {
                    $pkg_or_obj->call_on_apiset(%opts,
                        choices => [@cmode_cli_choices]);
                    $passed = 1;
                }
                catch NACL::APISet::Exceptions::ResponseException with {
                    $exception = $_[0];
                };
            } else {
                $Log->exit();
                $vault->convert_and_throw(exception => $exception);
            }
        }
        if ($passed == 0) {
            foreach my $type (
                qw( DoesNotExist TransferInProgress
                NotQuiescedOrQuiescing DataProtectionVolume
                VolumeNotInitializedorRestricted BreakNotAllowedForLS VolumeIsOffline RelationshipNotInitialized)
                ) {
                my $full_type = "NACL::C::Exceptions::Snapmirror::$type";
                $full_type->detect_convert_and_throw(exception => $exception);
            }

            $Log->exit();
            $exception->throw();
        }
    };

    $Log->exit();
} ## end sub break

sub _break_cmode_cli {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@args, api => 'snapmirror_break');
    $Log->exit();
}

sub _break_cmode_zapi_common_opts {
    return (
        api             => 'snapmirror-break',
        additional_copy => [qw(restore-destination-to-snapshot)],
    );
}

sub _break_zapi_check {
    $Log->enter();
    my ($pkg_or_obj, %opts) = @_;
    if (defined $opts{'delete-snapshots'} && $opts{'delete-snapshots'} eq 'true') {
        my $msg = 'Could not use the ZAPI implementation because the option '
            . "'delete-snapshots' was provided, which is not "
            . 'supported by the snapmirror-break-* ZAPIs (see burt 807636).';
        $Log->exit();
        NACL::Exceptions::InvalidChoice->throw($msg);
    }
    $Log->exit();
}

sub _break_non_iter_check {
    $Log->enter();
    my ($pkg_or_obj, %opts) = @_;
    $pkg_or_obj->_break_zapi_check(%opts);
    $pkg_or_obj->_non_iter_check(%opts);
    $Log->exit();
}

sub _break_cmode_zapi {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_build_iter_backend(@args,
        common_opts_method => '_break_cmode_zapi_common_opts');
    $Log->exit();
}

sub _break_cmode_zapi_non_iter {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_build_non_iter_backend_with_async(@args,
        common_opts_method => '_break_cmode_zapi_common_opts');
    $Log->exit();
}

sub _break_7mode_cli {
    my $pkg_or_obj = shift;
    $Log->enter();
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};

    my $snapmirror_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => [qw(force destination-path)],
    );

    $apiset->snapmirror_break(%$snapmirror_args);
    $Log->exit();
} ## end sub _break_7mode_cli

sub _break_7mode_zapi {
    my $pkg_or_obj = shift;
    $Log->enter();
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};

    my $snapmirror_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        map    => {"destination-path" => "destination-location",},
    );

    $apiset->snapmirror_break(%$snapmirror_args);
    $Log->exit();
} ## end sub _break_7mode_zapi

=head2 resync

   NACL::C::Snapmirror->resync(command_interface=>$ci,
                               source-path=>$source_path,
                               destination-path=>$destination_path);

  (or)

  NACL::C::Snapmirror->resync(command_interface=>$ci,
                              source-vserver=>$vserver,
                              source-volume=>$volume,
                              destination-path=>$destination_path);
  (or)

  $snapmirror->resync(...);

(Class or instance method) The method does resync on snapmirror relationship

Uses a CMode CLI/ZAPI (iter and non-iter supported) or 7Mode CLI/ZAPI APISet.

Invokes 'snapmirror-resync-iter' API for iter CMode ZAPI; invokes
'snapmirror-resync' API for non-iter CMode ZAPI.

Invokes 'snapmirror resync' command for 7Mode CLI.

Invokes 'snapmirror-resync' API for 7Mode ZAPI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "destination-cluster"=>$string >>

Name of the Cluster where the destination volume resides

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

destination-path input would be constructed from this input in the form (RR: cluster://volume/vserver, SN and above: not applicable) and passed to destination-location option.

=item C<< "destination-vserver"=>$string >>

Name of the Vserver  where the destination volume resides

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

destination-path input would be constructed from this input in the form
(RR: cluster://volume/vserver, SN and above: vserver:volume) and passed to
'destination-location' option.

=item C<< "destination-volume"=>$string >>

Name of the destination volume  which is in snapmirror relationship

Applicable for CMode CLI/ZAPI, 7Mode CLI/ZAPI.

Maps to:

7M CLI:

Since this option is not passed to the apilayer call ,there is no mapping
information provided ,destination-path will be constructed using this argument
and passed to the apilayer call.

7M ZAPI:

Since this option is not passed to the apilayer call ,there is no mapping
information provided ,destination-location will be constructed using this argument
and passed to the apilayer call.

CM ZAPI:

destination-path input would be constructed from this input in the form
(RR: cluster://volume/vserver, SN and above: vserver:volume) and passed to
'destination-location' option.

=item C<< "source-cluster"=>$string >>

Name of cluster where the source volume resides

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

source-path input would be constructed from this input in the form
(RR: cluster://volume/vserver, SN and above: not applicable) and passed to
source-location option.

=item C<< "source-vserver"=>$string >>

Name of the vserver where the source volume resides

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

source-path input would be constructed from this input in the form
(RR: cluster://volume/vserver, SN and above: vserver:volume) and passed to
'source-location' option.

=item C<< "source-volume"=>$string >>

Name of the source volume for which the snapmirror relation is created

Applicable for CMode CLI/ZAPI, 7Mode CLI/ZAPI.

Maps to:

7M CLI:

Since this option is not passed to the apilayer call ,there is no mapping
information provided ,source-path will be constructed using this argument
and passed to the apilayer call.

7M ZAPI:

Since this option is not passed to the apilayer call ,there is no mapping
information provided ,source-location will be constructed using this argument
and passed to the apilayer call.

CM ZAPI:

source-path input would be constructed from this input in the form
(RR: cluster://volume/vserver, SN and above: vserver:volume) and passed to
'source-location' option.

=item C<< "source-snapshot"=>$string >>

(Optional) Name of the source snapshot

Applicable for CMode CLI/ZAPI, 7Mode CLI/ZAPI.

Maps to:

7M CLI: -s <value> (Programmatic name: "source-snapshot")

7M ZAPI: source-snapshot

CM ZAPI: source-snapshot

=item C<< "destination-snapshot"=>$string >>

(Optional) Name of the destination snapshot

Applicable for 7Mode CLI/ZAPI.

Maps to:

7M CLI: -c <value> (Programmatic name: "destination-snapshot")

7M ZAPI: destination-snapshot

=item C<< "hidden-type"=>$boolean >>

Applicable for CMode CLI

=item C<< "dest-cluster-id"=>$string >>

Applicable for CMode CLI

=item C<< "dest-vserver-id"=>$string >>

Applicable for CMode CLI

=item C<< "dest-volume-msid"=>$string >>

Applicable for CMode CLI

=item C<< "dest-volume-dsid"=>$string >>

Applicable for CMode CLI

=item C<< "source-path"=>$string >>

(Applicable for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI, 7Mode CLI/ZAPI.

Maps to:

CM ZAPI:

<snapmirror-resync-iter>
<query>
<snapmirror-info>
<source-location>$value</source-location>
</snapmirror-info>
</query>
</snapmirror-resync-iter>

7M CLI : -S (Programmatic name: source-location)

7M ZAPI: source-location

=item C<< "destination-path"=>$string >>

(Applicable for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI, 7Mode CLI/ZAPI.

Maps to:

CM ZAPI:

<snapmirror-resync-iter>
<query>
<snapmirror-info>
<destination-location>$value</destination-location>
</snapmirror-info>
</query>
</snapmirror-resync-iter>

7M CLI: destination location (Programmatic name: "destination-location")

7M ZAPI: destination-location

=item C<< force=>$boolean >>

Applicable for CMode CLI, 7ModeCLI.

Maps to:

7M CLI: -f (Programmatic name: "force")

=item C<< throttle=>$number >>

Applicable for CMode ZAPI, 7Mode CLI/ZAPI.

Maps to:

CM ZAPI: max-transfer-rate

7M ZAPI: max-transfer-rate

7M CLI:  -k (Programmatic name: "max-transfer-rate")

item C<< foreground=>$boolean >>

=item C<< "job_component"=> \$scalar >>

Filled in for CMode CLI/ZAPI.

This attribute would be filled if the scalar ref passed to job_component.

=item C<< "job_components"=> \@array >>

Filled in for CMode CLI/ZAPI.

Applicable only if the method used with extended_query to resync
multiple snapmirror's.
This attribute would be filled if there are multiple job ids
obtained because of acting on multiple entities.

=item C<< 'dest-vserver-uuid' => $string >>

Applicable for CMode CLI

=item C<< 'dry-run' => $string >>

Applicable for CMode CLI

=item C<< 'preserve' => $string >>

Applicable for CMode CLI/ZAPI

=item C<< 'transfer-priority' => $string >>

Applicable for CMode CLI/ZAPI

=item C<< 'time-to-complete' => $string >>

Applicable for CMode CLI

=item C<< 'type' => $string >>

Applicable for CMode CLI

=item C<< 'test-run' => $string >>

Applicable for CMode CLI, 7Mode CLI.

Maps to:

7M CLI : -n (Programmatic name: "test-run")

=item C<< 'extended_query' => $string >>

(Optional)

Applicable for CMode CLI/ZAPI.

=item C<< 'no-onwire-se-to-existing-data' => $string >>

No storage efficiency on existing data

Applicable for CMode CLI/ZAPI

=item C<< 'operation-id' => $string >>

Operation Id

Applicable for CMode CLI

=item C<< 'quick-resync' => $string >>

Quick Resync
Applicable for CMode CLI

=back

=back

=over

=item NACL_SNAPMIRROR_DP

(Optional, defaults to 1)

Global flag that affect the behavior of snapmirror resync command
If this param is set, then all the snapmirror relation created would be of type dp.
Applicable to c-mode only.

=over

=item NACL_SNAPMIRROR_DP_OPTION

(Optional, defaults to 1)

Global flag that affect the behavior of snapmirror resync command
If this param is set in conjunction with NACL_SNAPMIRROR_DP, then the option replication.create_data_protection_rels.enable will be turned on.
If this param is not set, then the option replication.create_data_protection_rels.enable will not be turned on.
This param should be off for parallel tests.
Applicable to c-mode only.

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=item C<NACL::C::Exceptions::Snapmirror::DoesNotExist>

This exception gets thrown if an attempt is made to resync a
non-existing snapmirror.

=back

=back

=cut

sub resync {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw(source-cluster source-vserver source-volume source-cg
                   destination-cluster destination-vserver destination-volume
                   destination-cg hidden-type dest-cluster-id dest-vserver-id
                   dest-volume-msid dest-volume-dsid force throttle
                   foreground test-run source-snapshot destination-snapshot
                   dest-vserver-uuid dry-run preserve transfer-priority
                   time-to-complete type no-onwire-se-to-existing-data
                   operation-id quick-resync)
            ),
            'cg-item-mappings' => { type => HASHREF | ARRAYREF, optional => 1 },
        },
        cli_cmd => 'snapmirror resync',
    );

    my $destination_ci = $opts{command_interface};
    my $dst_Api_Set_Obj  = $destination_ci->apiset(interface => "CLI");
    my $dst_vol_obj = $opts{'destination-path'};
    $dst_vol_obj =~ s/(.*):(.*)/$1 $2/;
    my $dst_obj;
    my %tmp_hash;
    $tmp_hash{'vserver'} = $1 if (defined $1 && $1 ne "");
    $tmp_hash{'volume'} = $2 if (defined $2 && $2 ne "");
    try {
       $dst_obj = NACL::C::Volume->find(
           command_interface => $destination_ci,
           filter            => \%tmp_hash,
           allow_empty => 1,
       );
    }
    catch NACL::APISet::Exceptions::ResponseException with {
       $Log->comment('volume show fails');
    };

    if ($dst_obj) {
	if (! ($dst_obj->get_one_state_attribute('is-flexgroup') eq 'true' || $dst_obj->get_one_state_attribute('is-constituent') eq 'true')) {
	    if ( $destination_ci->version_manager()->is_version_ge(
                        release => $NACL::Global::ONTAP_Release_FatTire_0 ) ) {
		my $nacl_snapmirror_dp = defined param_global->get('NACL_SNAPMIRROR_DP') ? param_global->get('NACL_SNAPMIRROR_DP') : 1;
		my $nacl_snapmirror_dp_option = defined param_global->get('NACL_SNAPMIRROR_DP_OPTION') ? param_global->get('NACL_SNAPMIRROR_DP_OPTION') : 1 ;
		my $sm_fetch = NACL::CS::Snapmirror->fetch(
							   command_interface => $opts{command_interface},
							   filter            => { 'destination-path' => $opts{'destination-path'} },
							   allow_empty       => 1,
							  );
		if ((!defined $opts{'type'}) ||  ($opts{'type'}=~ /^dp/i)) {
		    if ( !($sm_fetch) || (($sm_fetch) && ($sm_fetch->type() =~ /^DP/i)) ) {
			if ( ($nacl_snapmirror_dp) && ($destination_ci->mode() eq 'CMode' ) ) {
			    $opts{'type'} = 'dp';
			    if ( $nacl_snapmirror_dp_option ) {
				$Log->comment('UDP: option replication.create_data_protection_rels.enable is set to true for DP');
				my $opt_value = _get_dp_option('dst_Api_Set_Obj' => $dst_Api_Set_Obj);
				if ( $opt_value =~ /^on/ ) {
				    $Log->comment('option replication.create_data_protection_rels.enable is already on');
				} else {
				    $Log->comment('option replication.create_data_protection_rels.enable is not on, so turn it on');
				    $dst_Api_Set_Obj->execute_raw_command( command => 'options replication.create_data_protection_rels.enable on' );
				}
			    }
			}
		    }
		}
	    }
	}
    }
    try {
        $pkg_or_obj->call_on_apiset(
            %opts,
            choices => [
                {   method    => '_resync_cmode_cli',
                    interface => 'CLI',
                    set       => 'CMode',
                    zapi_type => 'none',
                },
                {   method    => '_resync_cmode_zapi',
                    interface => 'ZAPI',
                    set       => 'CMode',
                    zapi_type => 'iter',
                },
                {   method    => '_resync_cmode_zapi_non_iter',
                    interface => 'ZAPI',
                    set       => 'CMode',
                    zapi_type => 'non-iter',
                    check     => '_non_iter_initialize_resync_check',
                },
                {   method    => '_resync_7mode_cli',
                    interface => 'CLI',
                    set       => '7Mode',
                },
                {   method    => '_resync_7mode_zapi',
                    interface => 'ZAPI',
                    set       => '7Mode',
                },
            ],
        );
    }
    catch NACL::C::Exceptions::Component with {

        # The CMode-ZAPI iter back-end might already classify the error
        # into a Component-level exception. If already classified, we can
        # simply throw this exception rather than go through the
        # detect_convert_and_throw again.
        $Log->exit();
        $_[0]->throw;
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = $_[0];
        NACL::C::Exceptions::Snapmirror::DoesNotExist
            ->detect_convert_and_throw(exception => $exception);
        NACL::C::Exceptions::Snapmirror::TransferInProgress
            ->detect_convert_and_throw(exception => $exception);
        $Log->exit();
        $exception->throw();
    };
    my %tmp_opts;
    $pkg_or_obj->_hash_copy(
                     source => \%opts,
                     copy   => [ qw(source-path destination-path command_interface
                                                          source_ci) ],
                     target => \%tmp_opts,
    );
    my $obj = $pkg_or_obj->new(%tmp_opts);
    $Log->exit();
    return $obj;
} ## end sub resync

sub _resync_cmode_cli {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@args, api => 'snapmirror_resync');
    $Log->exit();
}

sub _resync_cmode_zapi_common_opts {
    return (
        api             => 'snapmirror-resync',
        additional_copy => [
            qw(preserve quick-resync transfer-priority
                source-snapshot)
        ],
        additional_map => {throttle => 'max-transfer-rate'},
    );
}

sub _resync_cmode_zapi {
    $Log->enter();
    my ($pkg_or_obj,%opts) = @_;

    $pkg_or_obj->_build_iter_backend_with_retry(
        %opts,
        common_opts_method => '_resync_cmode_zapi_common_opts',
        _operation         => 'resync'
    );

    $Log->exit();
}

sub _resync_cmode_zapi_non_iter {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;

    ##burt934176 Extra input: vserver-type need to be removed
    my %args = @args;
    delete $args{type} if (defined $args{type});
    delete $args{policy} if (defined $args{policy});
    @args = %args;

    $pkg_or_obj->_build_non_iter_backend(@args,
        common_opts_method => '_resync_cmode_zapi_common_opts');
    $Log->exit();
}

sub _resync_7mode_cli {
    my $pkg_or_obj = shift;
    $Log->enter();
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my $rex = qr/.*?Are you sure you want to resync the (volume|qtree).*/i;
    my $prompt_answers = [$rex => 'yes',];
    my %snapmirror_args;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => [qw(test-run force source-snapshot destination-snapshot)],
        map    => {
            "destination-path" => "destination-location",
            "source-path"      => "source-location",
            "throttle"         => "max-transfer-rate",
        },
        target => \%snapmirror_args,
    );

    $apiset->snapmirror_resync(%snapmirror_args,
        'connectrec-match_table' => $prompt_answers);
    $Log->exit();
} ## end sub _resync_7mode_cli

sub _resync_7mode_zapi {
    my $pkg_or_obj = shift;
    $Log->enter();
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my %snapmirror_args;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => [qw( source-snapshot destination-snapshot )],
        map    => {
            "destination-path" => "destination-location",
            "source-path"      => "source-location",
            "throttle"         => "max-transfer-rate",
        },
        target => \%snapmirror_args,
    );

    $apiset->snapmirror_resync(%snapmirror_args);
    $Log->exit();
} ## end sub _resync_7mode_zapi

=head2 initialize

  NACL::C::Snapmirror->initialize(command_interface=>$ci,
                                        source-path=>$source_path,
                                        destination-path=>$destination_path);

  (or)

  NACL::C::Snapmirror->initialize(command_interface=>$ci,
                                        source-vserver=>$vserver,
                                        source-volume=>$volume,
                                        destination-path=>$destination_path);

  (or)

  $snapmirror->initialize(...);



(Class or instance method) The method does initialization of a snapmirror
relationship. If the relationship does not yet exist, it is created. A new
object is returned representing the (possibly new) relationship.

NOTE: This method B<always> returns a new object, even if called as an instance
method.

Uses a CMode CLI/ZAPI (iter and non-iter) or 7Mode CLI/ZAPI APISet.

Invokes 'snapmirror-initialize-iter' API for iter CMode ZAPI; invokes
'snapmirror-initialize' API for non-iter CMode ZAPI.

Invokes 'snapmirror initialize' command for 7Mode CLI.

Invokes 'snapmirror-initialize' API for 7Mode ZAPI.

It returns a new Snapmirror object

For CMode ZAPI, it returns a Snapmirror object only for instance method call
and the class method call without passing any special characters.
It will not return anything if it is a class call passing special characters.
For Example:
     NACL::C::Snapmirror->initialize(command_interface=>$ci,
                                        'source-cluster'=>$cluster,
                                        'source-vserver'=>$vserver,
                                        'source-volume'=>$volume,
                                        'destination-path'=>'*');

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>


=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "destination-cluster"=>$string >>

Name of the Cluster where the destination volume resides

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

destination-path input would be constructed from this input in the form (RR: cluster://volume/vserver, SN and above: not applicable) and passed to destination-location option.

=item C<< "destination-vserver"=>$string >>

Name of the Vserver  where the destination volume resides

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

destination-path input would be constructed from this input in the form
(RR: cluster://volume/vserver, SN and above: vserver:volume) and passed to
'destination-location' option.

=item C<< "destination-volume"=>$string >>

(Required for class method, Not Applicable for instance method)
Name of the destination volume  which is in snapmirror relationship

Applicable for CMode CLI/ZAPI, 7Mode CLI/ZAPI.

Maps to:

7M CLI:

Since this option is not passed to the apilayer call ,there is no mapping
information provided ,destination-path will be constructed using this argument
and passed to the apilayer call.

7M ZAPI:

Since this option is not passed to the apilayer call ,there is no mapping
information provided ,destination-location will be constructed using this argument
and passed to the apilayer call.

CM ZAPI:

destination-path input would be constructed from this input in the form
(RR: cluster://volume/vserver, SN and above: vserver:volume) and passed to
'destination-location' option.

=item C<< destination-cg =E<gt> $string >>

(Optional) Name of destination consistency group for a granular relationship.

Applicable only to CMode CLI/ZAPI.

This option does not map to an API-layer option. If passed (and if
C<destination-path> is not already specified), this is used with the
C<destination-vserver> value to form the destination path as
C<$vserver:/cg/$cg_name>. If this parameter is present, it is given
precedence over C<destination-volume> (but only for CMode).

=item C<< "source-cluster"=>$string >>

(Optional for class method, Not Applicable for instance method)
Name of cluster where the source volume resides

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

source-path input would be constructed from this input in the form
(RR: cluster://volume/vserver, SN and above: not applicable) and passed to
source-location option.

=item C<< "source-vserver"=>$string >>

(Optional for class method, Not Applicable for instance method)
Name of the vserver where the source volume resides

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

source-path input would be constructed from this input in the form
(RR: cluster://volume/vserver, SN and above: vserver:volume) and passed to
'source-location' option.

=item C<< "source-volume"=>$string >>

(Optional for class method, Not Applicable for instance method)
Name of the source volume for which the snapmirror relation is created

Applicable for CMode CLI/ZAPI, 7Mode CLI/ZAPI.

Maps to:

7M CLI:

Since this option is not passed to the apilayer call ,there is no mapping
information provided ,source-path will be constructed using this argument
and passed to the apilayer call.

7M ZAPI:

Since this option is not passed to the apilayer call ,there is no mapping
information provided ,source-location will be constructed using this argument
and passed to the apilayer call.

CM ZAPI:

source-path input would be constructed from this input in the form
(RR: cluster://volume/vserver, SN and above: vserver:volume) and passed to
'source-location' option.

=item C<< source-cg =E<gt> $string >>

(Optional) Name of source consistency group for a granular relationship.

Applicable only to CMode CLI/ZAPI.

This option does not map to an API-layer option. If passed (and if
C<source-path> is not already specified), this is used with the
C<source-vserver> value to form the source path as C<$vserver:/cg/$cg_name>. If
this parameter is present, it is given precedence over C<source-volume> (but
only for CMode).

=item C<< cg-item-mappings =E<gt> $hashref >>

(Optional) A hash-reference containing one or more source-to-destination
mappings of consistency group items for a granular relationship.

Applicable only to CMode CLI/ZAPI.

If this option is given, then the source and destination specifications
(whether by path or by parts) must refer to a consistency group. The mappings
given by this option are added to the consistency group specified by the
source and destination paths.

When present, this option must specify a hash-reference. Each key of the hash
will represent the source item path, and the value it points to will be the
destination path. Each of C<source-path> and C<destination-path> are of the
form:

    "/vol/" volume-name "/" file-or-lun-name

That is, the sequence C</vol/> followed immediately by a volume-name and that
followed immediately by the sequence C</>. Following the third C</> is a
file or LUN name. Currently, all CG items must be specified by full path.

=item C<< "hidden-type" =>$boolean >>

Applicable for CMode CLI

=item C<< "dest-cluster-id"=>$string >>

Applicable for CMode CLI

=item C<< "dest-vserver-id"=>$string >>

Applicable for CMode CLI

=item C<< "dest-volume-msid"=>$string >>

Applicable for CMode CLI

=item C<< "dest-volume-dsid"=>$string >>

Applicable for CMode CLI

=item C<< "source-path"=>$string >>

(Applicable for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

<snapmirror-initialize-iter>
<query>
<snapmirror-info>
<source-location>$value</source-location>
</snapmirror-info>
</query>
</snapmirror-initialize-iter>

=item C<< "destination-path"=>$string >>

(Applicable for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

<snapmirror-initialize-iter>
<query>
<snapmirror-info>
<destination-location>$value</destination-location>
</snapmirror-info>
</query>
</snapmirror-initialize-iter>

=item C<< throttle=>$number >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: max-transfer-rate

item C<< foreground=>$boolean >>

Applicable for CMode CLI

=item C<< type=>$string >>

Applicable for CMode CLI

=item C<< source-snapshot=>$string >>

Applicable for CMode CLI

=item C<< destination-snapshot=>$string >>

Applicable for CMode CLI

=item C<< "job_component"=> \$scalar >>

Filled in for CMode CLI/ZAPI.

This attribute would be filled if the scalar ref passed to job_component.

=item C<< "job_components"=> \@array >>

Filled in for CMode CLI/ZAPI.

Applicable only if the method used with extended_query to initialize
multiple snapmirror's.
This attribute would be filled if there are multiple job ids
obtained because of acting on multiple entities.

=item C<< 'dest-vserver-uuid' => $string >>

Applicable for CMode CLI

=item C<< 'transfer-priority' => $string >>

Transfer Priority, possible value(s) are low,medium,high

Applicable for CMode CLI

=item C<< 'extended_query' => $string >>

(Optional)

Applicable for CMode CLI/ZAPI.

=item C<< 'operation-id' => $string >>

Operation Id
Applicable for CMode CLI

=back

=back

=over

=item NACL_SNAPMIRROR_DP

(Optional, defaults to 1)

Global flag that affect the behavior of snapmirror initialize command
If this param is set, then all the snapmirror relation initialized would be of type dp.
Applicable only for c-mode only.

=over

=item NACL_SNAPMIRROR_DP_OPTION

(Optional, defaults to 1)

Global flag that affect the behavior of snapmirror initialize command
If this param is set in conjunction with NACL_SNAPMIRROR_DP, then the option replication.create_data_protection_rels.enable will be turned on.
If this param is not set, then the option option replication.create_data_protection_rels.enable will not be turned on.
This param should be off for parallel tests.
Applicable to c-mode only.

=over


=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=item C<NACL::C::Exceptions::Snapmirror::DoesNotExist>

This exception gets thrown if an attempt is made to initialize a
non-existing snapmirror.

=item C<NACL::C::Exceptions::Snapmirror::SnapmirrorDpLicenseCheckFailed>

This exception gets thrown if an attempt is made to create when
snapmirror DP license check failed for the cluster.

=item C<NACL::C::Exceptions::Snapmirror::NoDrRelationship>

This exception is raised when an attempt is made to initialize the snapmirror without
the vserver peering done for source and destination vservers.

=item C<NACL::C::Exceptions::Snapmirror::SystemInitialization>

This exception is thrown when an attempt is made to initialize
the snapmirror before completion of System Initialization.

=back

=back

=cut

sub initialize {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw(source-cluster source-vserver source-volume source-cg
                   hidden-type dest-cluster-id dest-vserver-id
                   destination-cluster destination-vserver destination-volume
                   destination-cg dest-volume-msid dest-volume-dsid type
                   throttle foreground source-snapshot destination-snapshot
                   dest-vserver-uuid transfer-priority operation-id)
            ),
            'cg-item-mappings' => { type => HASHREF | ARRAYREF, optional => 1 },
        },
        cli_cmd => 'snapmirror initialize',
    );

    my $destination_ci = $opts{command_interface};
    my $dst_Api_Set_Obj  = $destination_ci->apiset(interface => "CLI");
    my $dst_vol_obj = $opts{'destination-path'};
    $dst_vol_obj =~ s/(.*):(.*)/$1 $2/;
    my $dst_obj;
    my %tmp_hash;
    $tmp_hash{'vserver'} = $1 if (defined $1 && $1 ne "");
    $tmp_hash{'volume'} = $2 if (defined $2 && $2 ne "");
    try {
       $dst_obj = NACL::C::Volume->find(
           command_interface => $destination_ci,
           filter            => \%tmp_hash,
           allow_empty => 1,
       );
    }
    catch NACL::APISet::Exceptions::ResponseException with {
       $Log->comment('volume show fails');
    };
    if ($dst_obj) {
	if (! ($dst_obj->get_one_state_attribute('is-flexgroup') eq 'true' || $dst_obj->get_one_state_attribute('is-constituent') eq 'true')) {
	    if ( $destination_ci->version_manager()->is_version_ge(
                        release => $NACL::Global::ONTAP_Release_FatTire_0 ) ) {
		my $nacl_snapmirror_dp = defined param_global->get('NACL_SNAPMIRROR_DP') ? param_global->get('NACL_SNAPMIRROR_DP') : 1;
		my $nacl_snapmirror_dp_option = defined param_global->get('NACL_SNAPMIRROR_DP_OPTION') ? param_global->get('NACL_SNAPMIRROR_DP_OPTION') : 1 ;
		my $sm_fetch = NACL::CS::Snapmirror->fetch(
							   command_interface => $opts{command_interface},
							   filter            => { 'destination-path' => $opts{'destination-path'} },
							   allow_empty       => 1,
							  );
		if ((!defined $opts{'type'}) ||  ($opts{'type'}=~ /^dp/i)) {
		    if ( !($sm_fetch) || (($sm_fetch) && ($sm_fetch->type() =~ /^DP/i)) ) {
			if ( ($nacl_snapmirror_dp) && ($destination_ci->mode() eq 'CMode' ) ) {
			    $opts{'type'} = 'dp';
			    if ( $nacl_snapmirror_dp_option ) {
				$Log->comment('UDP: option replication.create_data_protection_rels.enable is set to true for DP');
				my $opt_value = _get_dp_option('dst_Api_Set_Obj' => $dst_Api_Set_Obj);
				if ( $opt_value =~ /^on/ ) {
				    $Log->comment('option replication.create_data_protection_rels.enable is already on');
				} else {
				    $Log->comment('option replication.create_data_protection_rels.enable is not on, so  turn it on');
				    $dst_Api_Set_Obj->execute_raw_command( command => 'options replication.create_data_protection_rels.enable on' );
				}
			    }
			}
		    }
		}
	    }
	}
    }
    try {
        $pkg_or_obj->call_on_apiset(
            %opts,
            choices => [
                {   method    => '_initialize_cmode_cli',
                    interface => 'CLI',
                    set       => 'CMode',
                    zapi_type => 'none',
                },
                {   method    => '_initialize_cmode_zapi',
                    interface => 'ZAPI',
                    set       => 'CMode',
                    zapi_type => 'iter',
                },
                {   method    => '_initialize_cmode_zapi_non_iter',
                    interface => 'ZAPI',
                    set       => 'CMode',
                    zapi_type => 'non-iter',
                    check     => '_non_iter_initialize_resync_check',
                },

      #
      #The following is required so that the existing user scripts
      #doesnt break.Ideally ,users should use create method for 7Mode ClI/ZAPI
      #
                {   method    => '_create_7mode_cli',
                    interface => 'CLI',
                    set       => '7Mode',
                },
                {   method    => '_create_7mode_zapi',
                    interface => 'ZAPI',
                    set       => '7Mode',
                }
            ]
        );
    }
    catch NACL::C::Exceptions::Component with {

        # The CMode-ZAPI iter back-end might already classify the error
        # into a Component-level exception. If already classified, we can
        # simply throw this exception rather than go through the
        # detect_convert_and_throw again.
        $Log->exit();
        $_[0]->throw;
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = $_[0];
        foreach my $type (
            qw(SnapmirrorDpLicenseCheckFailed NoDrRelationship DoesNotExist
               SystemInitialization)
        ) {
            my $full_type = "NACL::C::Exceptions::Snapmirror::$type";
            $full_type->detect_convert_and_throw(exception => $exception);
        }
        $Log->exit();
        $exception->throw();
    };

    my %tmp_opts;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => [ qw(source-path destination-path command_interface
                       source_ci) ],
        target => \%tmp_opts,
    );
    my $obj = $pkg_or_obj->get_package_name->new(%tmp_opts);
    $Log->exit();
    return $obj;
} ## end sub initialize

sub _initialize_cmode_cli {
    $Log->enter();
    my ($pkg_or_obj, %opts) = @_;

    $pkg_or_obj->_base_backend_cmode_cli(%opts,
        api => 'snapmirror_initialize');
    $Log->exit();
}

sub _initialize_cmode_zapi_common_opts {
    return (
        api             => 'snapmirror-initialize',
        additional_copy => [qw(source-snapshot transfer-priority)],
        additional_map  => {throttle => 'max-transfer-rate'},
    );
}

# Will not be able to use the non-iter if:
# 1) The destination-path cannot be constructed
sub _non_iter_initialize_resync_check {
    $Log->enter();
    my ($pkg_or_obj, %opts) = @_;
    $pkg_or_obj->_non_iter_check(%opts);
    $Log->exit();
}

sub _initialize_cmode_zapi {
    $Log->enter();
    my ($pkg_or_obj,%opts) = @_;

    $pkg_or_obj->_build_iter_backend_with_retry(
        %opts,
        common_opts_method => '_initialize_cmode_zapi_common_opts',
        _operation         => 'initialize'
    );

    $Log->exit();
}

sub _initialize_cmode_zapi_non_iter {
    $Log->enter();
    my ($pkg_or_obj, %opts) = @_;
    delete $opts{type} if(defined $opts{type});
    delete $opts{policy} if(defined $opts{policy});
    $pkg_or_obj->_build_non_iter_backend(%opts,
        common_opts_method => '_initialize_cmode_zapi_common_opts');
    $Log->exit();
}

=head2 update

  NACL::C::Snapmirror->update(command_interface=>$ci,
                              source-path=>$source_path,
                              destination-path=>$destination_path);

  (or)

  NACL::C::Snapmirror->update(command_interface=>$ci,
                              source-vserver=>$vserver,
                              source-volume=>$volume,
                              destination-path=>$destination_path);

  (or)

  $snapmirror->update(...);

  (Class or instance method) The method updates a snapmirror relationship

Uses a CMode CLI/ZAPI (iter and non-iter supported) or 7Mode CLI/ZAPI APISet.

Invokes 'snapmirror-update-iter' API for iter CMode ZAPI; invokes
'snapmirror-update' API for non-iter CMode ZAPI.

Invokes 'snapmirror update' command for 7Mode CLI.

Invokes 'snapmirror-update' API for 7Mode ZAPI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "destination-cluster"=>$string >>

(Optional for CMode-CLI/ZAPI, Not Applicable for 7Mode-CLI/ZAPI)
Name of the Cluster where the destination volume resides

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

destination-path input would be constructed from this input in the form (RR: cluster://volume/vserver, SN and above: not applicable) and passed to destination-location option.

=item C<< "destination-vserver"=>$string >>

(Optional for CMode-CLI/ZAPI, Not Applicable for  7Mode-CLI/ZAPI)
Name of the Vserver  where the destination volume resides

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

destination-path input would be constructed from this input in the form
(RR: cluster://volume/vserver, SN and above: vserver:volume) and passed to
'destination-location' option.

=item C<< "destination-volume"=>$string >>

(Optional for class method, Not Applicable for instance method)
Name of the destination volume  which will be in snapmirror relationship

Applicable for CMode CLI/ZAPI, 7Mode CLI/ZAPI.

Maps to:

7M CLI:

Since this option is not passed to the apilayer call ,there is no mapping
information provided ,destination-path will be constructed using this argument
and passed to the apilayer call.

7M ZAPI:

Since this option is not passed to the apilayer call ,there is no mapping
information provided ,destination-location will be constructed using this argument
and passed to the apilayer call.

CM ZAPI:

destination-path input would be constructed from this input in the form
(RR: cluster://volume/vserver, SN and above: vserver:volume) and passed to
'destination-location' option.

=item C<< "source-cluster"=>$string >>

(Optional for class method, Not Applicable for instance method)
Name of cluster where the source volume resides

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

source-path input would be constructed from this input in the form
(RR: cluster://volume/vserver, SN and above: not applicable) and passed to
source-location option.

=item C<< "source-vserver"=>$string >>

(Optional for class method, Not Applicable for instance method)
Name of the vserver where the source volume resides

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

source-path input would be constructed from this input in the form
(RR: cluster://volume/vserver, SN and above: vserver:volume) and passed to
'source-location' option.

=item C<< "source-volume"=>$string >>

(Optional for class method, Not Applicable for instance method)
Name of the source volume for which the snapmirror relation is created

Applicable for CMode CLI/ZAPI, 7Mode CLI/ZAPI.

Maps to:

7M CLI:

Since this option is not passed to the apilayer call ,there is no mapping
information provided ,source-path will be constructed using this argument
and passed to the apilayer call.

7M ZAPI:

Since this option is not passed to the apilayer call ,there is no mapping
information provided ,source-location will be constructed using this argument
and passed to the apilayer call.

CM ZAPI:

source-path input would be constructed from this input in the form
(RR: cluster://volume/vserver, SN and above: vserver:volume) and passed to
'source-location' option.

=item C<< "hidden-type"=>$boolean >>

Applicable for CMode CLI

=item C<< "dest-cluster-id"=>$string >>

Applicable for CMode CLI

=item C<< "dest-vserver-id"=>$string >>

Applicable for CMode CLI

=item C<< "dest-volume-msid"=>$string >>

Applicable for CMode CLI

=item C<< "dest-volume-dsid"=>$string >>

Applicable for CMode CLI

=item C<< "source-path"=>$string >>

(Applicable for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI,7Mode CLI/ZAPI.

Maps to:

CM ZAPI:

<snapmirror-update-iter>
<query>
<snapmirror-info>
<source-location>$value</source-location>
</snapmirror-info>
</query>
</snapmirror-update-iter>

7M CLI : -S (Programmatic name: source-location)

7M ZAPI: source-location

=item C<< "destination-path"=>$string >>

Applicable for CMode CLI/ZAPI,7Mode CLI/ZAPI

Maps to:

CM ZAPI:

<snapmirror-update-iter>
<query>
<snapmirror-info>
<destination-location>$value</destination-location>
</snapmirror-info>
</query>
</snapmirror-update-iter>

7M CLI: destination location (Programmatic name: "destination-location")

7M ZAPI: destination-location

=item C<< throttle=>$number >>

Applicable for CMode CLI/ZAPI,7Mode CLI/ZAPI

Maps to:

CM ZAPI: max-transfer-rate

7M CLI : -k (Programmatic name: max-transfer-rate)

7M ZAPI: max-transfer-rate

=item C<< foreground=>$string >>

Applicable for CMode CLI

=item C<< "job_component"=> \$scalar >>

Filled in for CMode CLI/ZAPI.

This attribute would be filled if the scalar ref passed to job_component.

=item C<< "job_components"=> \@array >>

Filled in for CMode CLI/ZAPI.

Applicable only if the method used with extended_query to update multiple
snapmirror's.
This attribute would be filled if there are multiple job ids
obtained because of acting on multiple entities.

=item C<< 'dest-vserver-uuid' => $string >>

Applicable for CMode CLI

=item C<< 'source-snapshot' => $string >>

Applicable for CMode CLI and 7Mode CLI/ZAPI.

Maps to:

7M CLI: -s <value> (Programmatic name: "source-snapshot")

7M ZAPI: source-snapshot

=item C<< 'destination-snapshot' => $string >>

Applicable for CMode CLI and 7Mode CLI/ZAPI.

Maps to:

7M CLI: -c <value> (Programmatic name: "destination-snapshot")

7M ZAPI: destination-snapshot

=item C<< 'enable-storage-efficiency' => $string >>

Enable storage efficient transfers

Applicable for CMode CLI/ZAPI

=item C<< 'transfer-priority' => $string >>

Transfer Priority, possible value(s) are low,medium,high

Applicable for CMode CLI

=item C<< 'extended_query' => $string >>

(Optional)

Applicable for CMode CLI/ZAPI.

=item C<< 'operation-id' => $string >>

Operation Id
Applicable for CMode CLI

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=item C<NACL::C::Exceptions::Snapmirror::DoesNotExist>

This exception gets thrown if an attempt is made to update a
non-existing snapmirror.

=back

=back

=cut

sub update {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( source-cluster source-vserver source-volume
                    destination-cluster destination-vserver destination-volume
                    hidden-type dest-cluster-id dest-vserver-id
                    dest-volume-msid dest-volume-dsid throttle foreground
                    source-snapshot destination-snapshot
                    dest-vserver-uuid
                    enable-storage-efficiency transfer-priority
                    operation-id
                    )
            )
        },
        cli_cmd => 'snapmirror update',
    );

    try {
        $pkg_or_obj->call_on_apiset(
            %opts,
            choices => [
                {   method    => '_update_cmode_cli',
                    interface => 'CLI',
                    set       => 'CMode',
                    zapi_type => 'none',
                },
                {   method    => '_update_cmode_zapi',
                    interface => 'ZAPI',
                    set       => 'CMode',
                    zapi_type => 'iter',
                },
                {   method    => '_update_cmode_zapi_non_iter',
                    interface => 'ZAPI',
                    set       => 'CMode',
                    zapi_type => 'non-iter',
                    check     => '_non_iter_check',
                },
                {   method    => '_update_7mode_cli',
                    interface => 'CLI',
                    set       => '7Mode',
                },
                {   method    => '_update_7mode_zapi',
                    interface => 'ZAPI',
                    set       => '7Mode',
                },
            ]
        );
    }
    catch NACL::C::Exceptions::Component with {

        # The CMode-ZAPI iter back-end might already classify the error
        # into a Component-level exception. If already classified, we can
        # simply throw this exception rather than go through the
        # detect_convert_and_throw again.
        $Log->exit();
        $_[0]->throw;
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = $_[0];
        NACL::C::Exceptions::Snapmirror::DoesNotExist
            ->detect_convert_and_throw(exception => $exception);
        $Log->exit();
        $exception->throw();
    };
    $Log->exit();
} ## end sub update

sub _update_cmode_cli {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@args, api => 'snapmirror_update');
    $Log->exit();
}

sub _update_cmode_zapi_common_opts {
    return (
        api             => 'snapmirror-update',
        additional_copy => [
            qw(enable-storage-efficiency source-snapshot
                transfer-priority)
        ],
        additional_map => {throttle => 'max-transfer-rate'},
    );
}

sub _update_cmode_zapi {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_build_iter_backend(@args,
        common_opts_method => '_update_cmode_zapi_common_opts',  no_attributes => 1);
    $Log->exit();
}

sub _update_cmode_zapi_non_iter {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_build_non_iter_backend(@args,
        common_opts_method => '_update_cmode_zapi_common_opts');
    $Log->exit();
}

sub _update_7mode_cli {
    my $pkg_or_obj = shift;
    $Log->enter();
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my $snapmirror_args   = $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => [qw(source-snapshot destination-snapshot)],
        map    => {
            "destination-path" => "destination-location",
            "source-path"      => "source-location",
            "throttle"         => "max-transfer-rate",
        },
    );

    $apiset->snapmirror_update(%$snapmirror_args);
    $Log->exit();
} ## end sub _update_7mode_cli

sub _update_7mode_zapi {
    my $pkg_or_obj = shift;
    $Log->enter();
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my %snapmirror_args;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => [qw( source-snapshot destination-snapshot )],
        map    => {
            "destination-path" => "destination-location",
            "source-path"      => "source-location",
            "throttle"         => "max-transfer-rate",
        },
        target => \%snapmirror_args,
    );

    $apiset->snapmirror_update(%snapmirror_args);
    $Log->exit();
} ## end sub _update_7mode_zapi

=head2 promote

 NACL::C::Snapmirror->promote(command_interface=>$ci,
                              source-path=>$source_path,
                              destination-path=>$destination_path);

  (or)

 NACL::C::Snapmirror->promote(command_interface=>$ci,
                              source-vserver=>$vserver,
                              source-volume=>$volume,
                              destination-path=>$destination_path);

  (or)

 $snapmirror->promote(...);


(Class or instance method) The method promotes a snapmirror relationship

Uses a CMode CLI/ZAPI (iter and non-iter supported) APISet.

Invokes 'snapmirror-promote-iter' API for iter CMode ZAPI; invokes
'snapmirror-promote' API for non-iter CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>


=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "destination-cluster"=>$string >>

Name of the Cluster where the destination volume resides

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

destination-path input would be constructed from this input in the form (RR: cluster://volume/vserver, SN and above: not applicable) and passed to destination-location option.

=item C<< "destination-vserver"=>$string >>

Name of the Vserver  where the destination volume resides

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

destination-path input would be constructed from this input in the form
(RR: cluster://volume/vserver, SN and above: vserver:volume) and passed to
'destination-location' option.

=item C<< "destination-volume"=>$string >>

Name of the destination volume  which will be in snapmirror relationship

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

destination-path input would be constructed from this input in the form
(RR: cluster://volume/vserver, SN and above: vserver:volume) and passed to
'destination-location' option.

=item C<< "source-cluster"=>$string >>

Name of cluster where the source volume resides

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

source-path input would be constructed from this input in the form
(RR: cluster://volume/vserver, SN and above: not applicable) and passed to
source-location option.

=item C<< "source-vserver"=>$string >>

Name of the vserver where the source volume resides

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

source-path input would be constructed from this input in the form
(RR: cluster://volume/vserver, SN and above: vserver:volume) and passed to
'source-location' option.

=item C<< "source-volume"=>$string >>

Name of the source volume for which the snapmirror relation is created

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

source-path input would be constructed from this input in the form
(RR: cluster://volume/vserver, SN and above: vserver:volume) and passed to
'source-location' option.

=item C<< "hidden-type"=>$boolean >>

Applicable for CMode CLI

=item C<< "dest-cluster-id"=>$string >>

Applicable for CMode CLI

=item C<< "dest-vserver-id"=>$string >>

Applicable for CMode CLI

=item C<< "dest-volume-msid"=>$string >>

Applicable for CMode CLI

=item C<< "dest-volume-dsid"=>$string >>

Applicable for CMode CLI

=item C<< "source-path"=>$string >>

(Applicable for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

<snapmirror-promote-iter>
<query>
<snapmirror-info>
<source-location>$value</source-location>
</snapmirror-info>
</query>
</snapmirror-promote-iter>

=item C<< "destination-path"=>$string >>

(Applicable for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

<snapmirror-promote-iter>
<query>
<snapmirror-info>
<destination-location>$value</destination-location>
</snapmirror-info>
</query>
</snapmirror-promote-iter>

=item C<< force=>$boolean >>

Applicable for CMode CLI

=item C<< "job_component"=> \$scalar >>

Filled in for CMode CLI.

=item C<< 'dest-vserver-uuid' => $string >>

Applicable for CMode CLI

=item C<< 'extended_query' => $string >>

(Optional)

Applicable for CMode CLI/ZAPI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=item C<NACL::C::Exceptions::Snapmirror::DoesNotExist>

This exception gets thrown if an attempt is made to promote a
non-existing snapmirror.

=back

=back

=cut

sub promote {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;

    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            job_component => {type => SCALARREF, optional => 1},
            _optional_scalars(
                qw( source-cluster source-vserver source-volume
                    destination-cluster destination-vserver destination-volume
                    hidden-type dest-cluster-id dest-vserver-id
                    dest-volume-msid dest-volume-dsid force
                    dest-vserver-uuid
                    )
            )
        },
        cli_cmd => 'snapmirror promote',
    );

    try {
        $pkg_or_obj->call_on_apiset(
            %opts,
            choices => [
                {   method    => '_promote_cmode_cli',
                    interface => 'CLI',
                    set       => 'CMode',
                    zapi_type => 'none',
                },
                {   method    => '_promote_cmode_zapi',
                    interface => 'ZAPI',
                    set       => 'CMode',
                    zapi_type => 'iter',
                },
                {   method    => '_promote_cmode_zapi_non_iter',
                    interface => 'ZAPI',
                    set       => 'CMode',
                    zapi_type => 'non-iter',
                    check     => '_non_iter_check',
                }
            ],
        );
    }
    catch NACL::C::Exceptions::Component with {

        # The CMode-ZAPI iter back-end might already classify the error
        # into a Component-level exception. If already classified, we can
        # simply throw this exception rather than go through the
        # detect_convert_and_throw again.
        $Log->exit();
        $_[0]->throw;
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = $_[0];
        NACL::C::Exceptions::Snapmirror::DoesNotExist
            ->detect_convert_and_throw(exception => $exception);
        $Log->exit();
        $exception->throw();
    };
    $Log->exit();
} ## end sub promote

sub _promote_cmode_cli {
    $Log->enter();
    my ($pkg_or_obj, @opts) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@opts, api => 'snapmirror_promote');
    $Log->exit();
} ## end sub _promote_cmode_cli

sub _promote_cmode_zapi {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_iter_helper(
         @args,
         api      => 'snapmirror-promote-iter',
         show_cmd => 'snapmirror show'
    );
    $Log->exit();
}

sub _promote_cmode_zapi_non_iter {
    $Log->enter();
    my ($pkg_or_obj, @opts) = @_;
    $pkg_or_obj->_base_backend_cmode_zapi(@opts, api => 'snapmirror-promote');
    $Log->exit();
}

=head2 abort

 NACL::C::Snapmirror->abort(command_interface=>$ci,
                            source-path=>$source_path,
                            destination-path=>$destination_path);

  (or)

 NACL::C::Snapmirror->abort(command_interface=>$ci,
                            source-vserver=>$vserver,
                            source-volume=>$volume,
                            destination-path=>$destination_path);

  (or)

 $snapmirror->abort(...);

(Class or instance method) The method aborts a snapmirror relationship

Uses a CMode CLI/ZAPI (supports async, iter and non-iter)
or 7Mode CLI/ZAPI APISet.

Invokes 'snapmirror-abort-iter' API for iter CMode ZAPI.

For non-iter CMode ZAPI: invokes 'snapmirror-abort-async' for snapmirror
relationships between infinite volumes; invokes 'snapmirror-abort' for
other types of relationships.

Invokes 'snapmirror-abort' API for 7Mode ZAPI.

Invokes 'snapmirror abort' command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>


=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "destination-cluster"=>$string >>

(Required for class method, Not Applicable for instance method)
(Optional for CMode-CLI/ZAPI, Not Applicable for 7Mode-CLI/ZAPI)
Name of the Cluster where the destination volume resides

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

destination-path input would be constructed from this input in the form (RR: cluster://volume/vserver, SN and above: not applicable) and passed to destination-location option.

=item C<< "destination-vserver"=>$string >>

(Required for class method, Not Applicable for instance method)
(Optional for CMode-CLI/ZAPI, Not Applicable for 7Mode-CLI/ZAPI)
Name of the Vserver  where the destination volume resides

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

destination-path input would be constructed from this input in the form
(RR: cluster://volume/vserver, SN and above: vserver:volume) and passed to
'destination-location' option.

=item C<< "destination-volume"=>$string >>

(Required for class method, Not Applicable for instance method)
Name of the destination volume  which will be in snapmirror relationship

Applicable for CMode CLI/ZAPI, 7Mode CLI/ZAPI.

Maps to:

7M CLI:

Since this option is not passed to the apilayer call ,there is no mapping
information provided ,destination-path will be constructed using this argument
and passed to the apilayer call.

7M ZAPI:

Since this option is not passed to the apilayer call ,there is no mapping
information provided ,destination-location will be constructed using this argument
and passed to the apilayer call.

CM ZAPI:

destination-path input would be constructed from this input in the form
(RR: cluster://volume/vserver, SN and above: vserver:volume) and passed to
'destination-location' option.

=item C<< "source-cluster"=>$string >>

(Optional for class method, Not Applicable for instance method)
Name of cluster where the source volume resides

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

source-path input would be constructed from this input in the form
(RR: cluster://volume/vserver, SN and above: not applicable) and passed to
source-location option.

=item C<< "source-vserver"=>$string >>

(Optional for class method, Not Applicable for instance method)
Name of the vserver where the source volume resides

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

source-path input would be constructed from this input in the form
(RR: cluster://volume/vserver, SN and above: vserver:volume) and passed to
'source-location' option.

=item C<< "source-volume"=>$string >>

(Optional for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI, 7Mode CLI/ZAPI.

Maps to:

7M CLI:

Since this option is not passed to the apilayer call ,there is no mapping
information provided ,source-path will be constructed using this argument
and passed to the apilayer call.

7M ZAPI:

Since this option is not passed to the apilayer call ,there is no mapping
information provided ,source-location will be constructed using this argument
and passed to the apilayer call.

CM ZAPI:

source-path input would be constructed from this input in the form
(RR: cluster://volume/vserver, SN and above: vserver:volume) and passed to
'source-location' option.

=item C<< "hidden-type"=>$boolean >>

Applicable for CMode CLI

=item C<< "dest-cluster-id"=>$string >>

Applicable for CMode CLI

=item C<< "dest-vserver-id"=>$string >>

Applicable for CMode CLI

=item C<< "dest-volume-msid"=>$string >>

Applicable for CMode CLI

=item C<< "dest-volume-dsid"=>$string >>

Applicable for CMode CLI

=item C<< "source-path"=>$string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: source-location option of snapmirror-abort-async API

For snapmirror-abort-iter API

<snapmirror-abort-iter>
<query>
<snapmirror-info>
<source-location>$value</source-location>
</snapmirror-info>
</query>
</snapmirror-abort-iter>

=item C<< "destination-path"=>$string >>

(Applicable for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI and 7Mode CLI/ZAPI.

Maps to:

CM ZAPI: destination-location of snapmirror-abort-async API

For snapmirror-abort-iter API

<snapmirror-abort-iter>
<query>
<snapmirror-info>
<destination-location>$value</destination-location>
</snapmirror-info>
</query>
</snapmirror-abort-iter>

7M CLI: destination path (Programmatic name: "destination-path")

7M ZAPI: destination-location

=item C<< "check-only"=>$string >>

Applicable for CMode CLI

=item C<< "clear-checkpoint"=>$string >>

Applicable for CMode CLI/ZAPI and 7Mode CLI/ZAPI.

Maps to:

CM ZAPI: clear-checkpoint

7M CLI : -h (Programmatic name: "clear-checkpoint")

7M ZAPI: clear-checkpoint

=item C<< "hard"=>$string >>

Applicable for CMode CLI

=item C<< "job_component"=> \$scalar >>

Applicable for CMode CLI/ZAPI.

CM ZAPI: returns job_component only for infinite volume.

=item C<< 'foreground' => $string >>

Applicable for CMode CLI

=item C<< 'dest-vserver-uuid' => $string >>

Applicable for CMode CLI

=item C<< 'extended_query' => $string >>

(Optional)

Applicable for CMode CLI/ZAPI.

=item C<< 'is-repository' => $string >>

B<DEPRECATED>. Use C<is_infinivol> instead.

=item C<< is_infinivol => 0 | 1 >>

(Optional, Applicable for CMode ZAPI)

This boolean option describes whether the destination volume is an infinite
volume or not.
For CMode ZAPI, the "async" form of the ZAPI must be used for the infinite
volume. If this option is not specified, the library will attempt to
determine if an infinite volume is being operated on (this could involve
querying the filer).

=item C<< is_infinivol_ref => \$scalar >>

(Optional, SCALARREF)

If the method was invoked through CMode ZAPI, the variable gets updated
describing whether it operated on an infinivol (used the "async" ZAPI) or not
(used the normal ZAPI).

=item C<< 'operation-id' => $string >>

Operation Id
Applicable for CMode CLI

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=item C<NACL::C::Exceptions::Snapmirror::DoesNotExist>

This exception gets thrown if an attempt is made to abort a
non-existing snapmirror.

=item C<NACL::C::Exceptions::Snapmirror::NoTransferToAbort>

This exception gets thrown if an attempt is made to abort a
when there is no transfer to abort.

=back

=back

=cut

sub abort {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw(source-cluster source-vserver source-volume source-cg
                   destination-cluster destination-vserver destination-volume
                   destination-cg clear-checkpoint hidden-type dest-cluster-id
                   dest-vserver-id dest-volume-msid dest-volume-dsid hard
                   check-only foreground dest-vserver-uuid is-repository
                   operation-id)
            )
        },
        cli_cmd => 'snapmirror abort',
    );

    try {
        $pkg_or_obj->call_on_apiset(
            %opts,
            choices => [
                {   method    => '_abort_cmode_cli',
                    interface => 'CLI',
                    set       => 'CMode',
                    zapi_type => 'none',
                },
                {   method    => '_abort_cmode_zapi',
                    interface => 'ZAPI',
                    set       => 'CMode',
                    zapi_type => 'iter',
                },
                {   method    => '_abort_cmode_zapi_non_iter',
                    interface => 'ZAPI',
                    set       => 'CMode',
                    zapi_type => 'non-iter',
                    check     => '_non_iter_check',
                },
                {   method    => '_abort_7mode_cli',
                    interface => 'CLI',
                    set       => '7Mode',
                },
                {   method    => '_abort_7mode_zapi',
                    interface => 'ZAPI',
                    set       => '7Mode',
                },
            ]
        );
    }
    catch NACL::C::Exceptions::Component with {

        # The CMode-ZAPI iter back-end might already classify the error
        # into a Component-level exception. If already classified, we can
        # simply throw this exception rather than go through the
        # detect_convert_and_throw again.
        $Log->exit();
        $_[0]->throw;
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        $pkg_or_obj->_abort_exception_handler(@_);
    };

    $Log->exit();
} ## end sub abort

sub _abort_exception_handler {
    my ($pkg_or_obj, $exception) = @_;

    foreach my $type (qw(DoesNotExist NoTransferToAbort)) {
       my $full_type = "NACL::C::Exceptions::Snapmirror::$type";
       $full_type->detect_convert_and_throw(exception => $exception);
    }
    $Log->exit();
    $exception->throw();
}

sub _abort_cmode_cli {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@args, api => 'snapmirror_abort');
    $Log->exit();
}

sub _abort_cmode_zapi_common_opts {
    return (
        api             => 'snapmirror-abort',
        additional_copy => [qw(check-only)],
        additional_map  => {'hard' => 'clear-checkpoint',}
    );
}

sub _abort_cmode_zapi {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_build_iter_backend(@args,
        common_opts_method => '_abort_cmode_zapi_common_opts', no_attributes => 1);
    $Log->exit();
}

sub _abort_cmode_zapi_non_iter {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_build_non_iter_backend_with_async(@args,
        common_opts_method => '_abort_cmode_zapi_common_opts');
    $Log->exit();
}

sub _abort_7mode_cli {
    my $pkg_or_obj = shift;
    $Log->enter();
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my %snapmirror_args;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => [qw(clear-checkpoint destination-path)],
        target => \%snapmirror_args,
    );

    $apiset->snapmirror_abort(%snapmirror_args);
    $Log->exit();
} ## end sub _abort_7mode_cli

sub _abort_7mode_zapi {
    $Log->enter();
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    delete $opts{command_interface};
    my $apiset = delete $opts{apiset};

    my %args;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => [qw( clear-checkpoint )],
        map    => {"destination-path" => "destination-location",},
        target => \%args,
    );

    $apiset->snapmirror_abort(%args);
    $Log->exit();
} ## end sub _abort_7mode_zapi

=head2 modify

 NACL::C::Snapmirror->modify(command_interface=>$ci,
                             source-path=>$source_path,
                             destination-path=>$destination_path);

  (or)

 NACL::C::Snapmirror->modify(command_interface=>$ci,
                             source-vserver=>$vserver,
                             source-volume=>$volume,
                             destination-path=>$destination_path);

  (or)

 $snapmirror->modify(...);


(Class or instance method) The method modifies a snapmirror relationship

Uses a CMode CLI/ZAPI (iter and non-iter supported) APISet.

Invokes 'snapmirror-modify-iter' API for iter CMode ZAPI; invokes
'snapmirror-modify' API for non-iter CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>


=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "destination-cluster"=>$string >>

Name of the Cluster where the destination volume resides

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

destination-path input would be constructed from this input in the form (RR: cluster://volume/vserver, SN and above: not applicable) and passed to destination-location option.

=item C<< "destination-vserver"=>$string >>

Name of the Vserver  where the destination volume resides

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

destination-path input would be constructed from this input in the form
(RR: cluster://volume/vserver, SN and above: vserver:volume) and passed to
'destination-location' option.

=item C<< "destination-volume"=>$string >>

Name of the destination volume  which will be in snapmirror relationship

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

destination-path input would be constructed from this input in the form
(RR: cluster://volume/vserver, SN and above: vserver:volume) and passed to
'destination-location' option.

=item C<< "source-cluster"=>$string >>

Name of cluster where the source volume resides

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

source-path input would be constructed from this input in the form
(RR: cluster://volume/vserver, SN and above: not applicable) and passed to
source-location option.

=item C<< "source-vserver"=>$string >>

Name of the vserver where the source volume resides

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

source-path input would be constructed from this input in the form
(RR: cluster://volume/vserver, SN and above: vserver:volume) and passed to
'source-location' option.

=item C<< "source-volume"=>$string >>

Name of the source volume for which the snapmirror relation is created

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

source-path input would be constructed from this input in the form
(RR: cluster://volume/vserver, SN and above: vserver:volume) and passed to
'source-location' option.

=item C<< "hidden-type"=>$boolean >>

Applicable for CMode CLI

=item C<< "dest-cluster-id"=>$string >>

Applicable for CMode CLI

=item C<< "dest-vserver-id"=>$string >>

Applicable for CMode CLI

=item C<< "dest-volume-msid"=>$string >>

Applicable for CMode CLI

=item C<< "dest-volume-dsid"=>$string >>

Applicable for CMode CLI

=item C<< "source-path"=>$string >>

(Applicable for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: source-location

=item C<< "destination-path"=>$string >>

(Applicable for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: destination-location

=item C<< "vserver"=>$string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: ZAPI will be invoked in the context of the vserver specified to this
  option.

=item C<< "schedule"=>$string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: schedule

=item C<< "tries"=>$number >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: tries

=item C<< "throttle"=>$number >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: max-transfer-rate

=item C<< "job_component"=> \$scalar >>

Filled in for CMode CLI.

=item C<< 'foreground' => $string >>

Applicable for CMode CLI

=item C<< 'policy' => $string >>

Applicable for CMode CLI/ZAPI

=item C<< 'dest-vserver-uuid' => $string >>

Applicable for CMode CLI

=item C<< 'extended_query' => $string >>

(Optional)
Applicable for CMode CLI

=item C<< 'operation-id' => $string >>

Operation Id
Applicable for CMode CLI

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub modify {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw(source-cluster source-vserver source-volume source-cg
                   destination-cluster destination-vserver destination-volume
                   destination-cg dest-cluster-id dest-vserver-id
                   dest-volume-msid dest-volume-dsid vserver hidden-type
                   schedule tries throttle foreground policy dest-vserver-uuid
                   operation-id)
            )
        },
        cli_cmd => 'snapmirror modify',
    );

    try {
        $pkg_or_obj->call_on_apiset(
            %opts,
            choices => [
                {   method    => '_modify_cmode_cli',
                    interface => 'CLI',
                    set       => 'CMode',
                    zapi_type => 'none',
                },
                {   method    => '_modify_cmode_zapi',
                    interface => 'ZAPI',
                    set       => 'CMode',
                    zapi_type => 'iter',
                },
                {   method    => '_modify_cmode_zapi_non_iter',
                    interface => 'ZAPI',
                    set       => 'CMode',
                    zapi_type => 'non-iter',
                    check     => '_non_iter_check',
                }
            ],
        );
    }
    catch NACL::C::Exceptions::Component with {

        # The CMode-ZAPI iter back-end might already classify the error
        # into a Component-level exception. If already classified, we can
        # simply throw this exception rather than go through the
        # detect_convert_and_throw again.
        $Log->exit();
        $_[0]->throw;
    }
    $Log->exit();
} ## end sub modify

sub _modify_cmode_cli {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@args, api => 'snapmirror_modify');
    $Log->exit();
}

sub _modify_cmode_zapi_common_opts {
    return (
        api             => 'snapmirror-modify',
        additional_copy => [qw(policy schedule tries is-auto-expand-enabled)],
        additional_map  => {throttle => 'max-transfer-rate'},
    );
}

sub _modify_cmode_zapi {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_build_iter_backend(@args,
        common_opts_method => '_modify_cmode_zapi_common_opts');
    $Log->exit();
}

sub _modify_cmode_zapi_non_iter {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_build_non_iter_backend(@args,
        common_opts_method => '_modify_cmode_zapi_common_opts');
    $Log->exit();
}

=head2 check

 NACL::C::Snapmirror->check(command_interface=>$ci,
                            source-path=>$source_path,
                            destination-path=>$destination_path);

  (or)

 NACL::C::Snapmirror->check(command_interface=>$ci,
                            source-vserver=>$vserver,
                            source-volume=>$volume,
                            destination-path=>$destination_path);

  (or)

  $snapmirror->check(...);

  (Class or instance method) The method checks a snapmirror relationship

Uses a CMode CLI/ZAPI (iter and non-iter) or 7Mode CLI/ZAPI APISet.

Invokes 'snapmirror-check-iter' API for iter CMode ZAPI; invokes
'snapmirror-check' API for non-iter CMode ZAPI.

Invokes 'snapmirror check start' command for 7Mode CLI.

Invokes 'snapmirror-check' API for 7Mode ZAPI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>


=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "destination-cluster"=>$string >>

(Optional for CMode-CLI/ZAPI, Not Applicable for 7Mode-CLI/ZAPI)
Name of the Cluster where the destination volume resides

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

destination-path input would be constructed from this input in the form (RR: cluster://volume/vserver, SN and above: not applicable) and passed to destination-location option.

=item C<< "destination-vserver"=>$string >>

(Optional for CMode-CLI/ZAPI, Not Applicable for 7Mode-CLI/ZAPI)
Name of the Vserver  where the destination volume resides

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

destination-path input would be constructed from this input in the form
(RR: cluster://volume/vserver, SN and above: vserver:volume) and passed to
'destination-location' option.

=item C<< "destination-volume"=>$string >>

Name of the destination volume  which will be in snapmirror relationship

Applicable for CMode CLI/ZAPI, 7Mode CLI.

Maps to:

7M CLI:

Since this option is not passed to the apilayer call ,there is no mapping
information provided ,destination-path will be constructed using this argument
and passed to the apilayer call.

7M ZAPI:

Since this option is not passed to the apilayer call ,there is no mapping
information provided ,destination-location will be constructed using this argument
and passed to the apilayer call.

CM ZAPI:

destination-path input would be constructed from this input in the form
(RR: cluster://volume/vserver, SN and above: vserver:volume) and passed to
'destination-location' option.

=item C<< "source-cluster"=>$string >>

(Optional for class method, Not Applicable for instance method)
Name of cluster where the source volume resides

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

source-path input would be constructed from this input in the form
(RR: cluster://volume/vserver, SN and above: not applicable) and passed to
source-location option.

=item C<< "source-vserver"=>$string >>

(Optional for class method, Not Applicable for instance method)
Name of the vserver where the source volume resides

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

source-path input would be constructed from this input in the form
(RR: cluster://volume/vserver, SN and above: vserver:volume) and passed to
'source-location' option.

=item C<< "source-volume"=>$string >>

(Optional for class method, Not Applicable for instance method)
Name of the source volume for which the snapmirror relation is created

Applicable for CMode CLI/ZAPI, 7Mode CLI/ZAPI.

Maps to:

7M CLI:

Since this option is not passed to the apilayer call ,there is no mapping
information provided ,source-path will be constructed using this argument
and passed to the apilayer call.

7M ZAPI:

Since this option is not passed to the apilayer call ,there is no mapping
information provided ,source-location will be constructed using this argument
and passed to the apilayer call.

CM ZAPI:

source-path input would be constructed from this input in the form
(RR: cluster://volume/vserver, SN and above: vserver:volume) and passed to
'source-location' option.

=item C<< "hidden-type"=>$boolean >>

Applicable for CMode CLI

=item C<< "dest-cluster-id"=>$string >>

Applicable for CMode CLI

=item C<< "dest-vserver-id"=>$string >>

Applicable for CMode CLI

=item C<< "dest-volume-msid"=>$string >>

Applicable for CMode CLI

=item C<< "dest-volume-dsid"=>$string >>

Applicable for CMode CLI

=item C<< "source-path"=>$string >>

(Applicable for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI,7Mode CLI.

Maps to:

CM ZAPI:

<snapmirror-check-iter>
<query>
<snapmirror-info>
<source-location>$value</source-location>
</snapmirror-info>
</query>
</snapmirror-check-iter>

7M CLI : -S (Programmatic name: source)

=item C<< "destination-path"=>$string >>

(Applicable for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI,7Mode CLI.

Maps to:

CM ZAPI:

<snapmirror-check-iter>
<query>
<snapmirror-info>
<destination-location>$value</destination-location>
</snapmirror-info>
</query>
</snapmirror-check-iter>

7M CLI: destination (Programmatic name: "destination")

=item C<< "target-snapshot"=>$string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: target-snapshot

=item C<< "logical-check"=>$string >>

Applicable for CMode CLI

=item C<< "foreground"=>$string >>

Applicable for CMode CLI

=item C<< "job_component"=> \$scalar >>

Filled in for CMode CLI/ZAPI.

This attribute would be filled if the scalar ref passed to job_component.

=item C<< "job_components"=> \@array >>

Filled in for CMode CLI/ZAPI.

Applicable only if the method used with extended_query to check multiple
snapmirror's.
This attribute would be filled if there are multiple job ids
obtained because of acting on multiple entities.

=item C<< 'throttle' => $string >>

Applicable for CMode CLI

=item C<< 'dest-vserver-uuid' => $string >>

Applicable for CMode CLI

=item C<< 'extended_query' => $string >>

(Optional)

Applicable for CMode CLI/ZAPI.

=item C<< 'file-list' => [ $file-1, $file-2, ... ] >>

File list
(Arrayref, even if only a single value is to be sent it should specified
as an arrayref containing that single value)
Applicable for CMode CLI

=item C<< 'source-snapshot' => $string >>

Source Snapshot
Applicable for CMode CLI

=item C<< 'operation-id' => $string >>

Operation Id
Applicable for CMode CLI

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub check {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            "file-list" => {type => ARRAYREF, optional => 1},
            _optional_scalars(
                qw(source-cluster source-vserver source-volume source-cg
                   destination-cluster destination-vserver destination-volume
                   destination-cg hidden-type dest-cluster-id dest-vserver-id
                   dest-volume-msid dest-volume-dsid target-snapshot foreground
                   logical-check throttle dest-vserver-uuid source-snapshot
                   operation-id)
            )
        },
        cli_cmd => 'snapmirror check',
    );

    # NOTE: This should only be in place until "snapmirror check" directly
    #       supports granular relationship paths for destination-path and
    #       source-path.
    if ($pkg_or_obj->is_granular(%opts)) {
        my ($src_path, $dst_path) = $pkg_or_obj->volume_paths_from_cg(%opts);
        $opts{'source-path'}      = $src_path;
        $opts{'destination-path'} = $dst_path;
    }

    try {
        $pkg_or_obj->call_on_apiset(
            %opts,
            choices => [
                {   method    => '_check_cmode_cli',
                    interface => 'CLI',
                    set       => 'CMode',
                    zapi_type => 'none',
                },
                {   method    => '_check_cmode_zapi',
                    interface => 'ZAPI',
                    set       => 'CMode',
                    zapi_type => 'iter',
                },
                {   method    => '_check_cmode_zapi_non_iter',
                    interface => 'ZAPI',
                    set       => 'CMode',
                    zapi_type => 'non-iter',
                    check     => '_non_iter_check',
                },
                {   method    => '_check_7mode_cli',
                    interface => 'CLI',
                    set       => '7Mode',
                },
            ],
        );
    }
    catch NACL::C::Exceptions::Component with {

        # The CMode-ZAPI iter back-end might already classify the error
        # into a Component-level exception. If already classified, we can
        # simply throw this exception rather than go through the
        # detect_convert_and_throw again.
        $Log->exit();
        $_[0]->throw;
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = $_[0];
        NACL::C::Exceptions::Snapmirror::DoesNotExist
            ->detect_convert_and_throw(exception => $exception);
        $Log->exit();
        $exception->throw();
    };

    $Log->exit();
} ## end sub check

sub _check_cmode_cli {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@args, api => 'snapmirror_check');
    $Log->exit();
}

sub _check_cmode_zapi_common_opts {
    return (
        api             => 'snapmirror-check',
        additional_copy => [qw(source-snapshot target-snapshot)],
        additional_map  => {
            'file-list' => 'check-file-list',
            throttle    => 'max-transfer-rate',
        }
    );
}

sub _check_cmode_zapi {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_build_iter_backend(@args,
        common_opts_method => '_check_cmode_zapi_common_opts');
    $Log->exit();
}

sub _check_cmode_zapi_non_iter {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_build_non_iter_backend(@args,
        common_opts_method => '_check_cmode_zapi_common_opts');
    $Log->exit();
}

sub _check_7mode_cli {
    my $pkg_or_obj = shift;
    $Log->enter();
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my %snapmirror_args;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        map    => {
            "source-path"      => "source",
            "destination-path" => "destination",
        },
        target => \%snapmirror_args,
    );

    $apiset->snapmirror_check_start(%snapmirror_args);
    $Log->exit();
} ## end sub _check_7mode_cli

=head2 initialize_ls_set

  NACL::C::Snapmirror->initialize_ls_set(command_interface=>$ci,
                              "source-vserver"=>$vserver,
                              "source-volume"=>$volume,...);

(Class method only) Request initialization of an LS snapmirror set.

Uses a CMode CLI/ZAPI APISet.

Invokes 'snapmirror-initialize-ls-set' API for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "source-cluster" =>$string >>

(Required) Name of cluster where the source volume resides

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

source-path input would be constructed from this input in the form
(RR: cluster://volume/vserver, SN and above: not applicable) and passed to
source-location option.

=item C<< "source-vserver" =>$string >>

(Required) Name of the vserver where the source volume resides

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

source-path input would be constructed from this input in the form
(RR: cluster://volume/vserver, SN and above: vserver:volume) and passed to
'source-location' option.

=item C<< "source-volume" =>$string >>

(Required) Name of the source volume for which the snapmirror relation is created

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

source-path input would be constructed from this input in the form
(RR: cluster://volume/vserver, SN and above: vserver:volume) and passed to
'source-location' option.

=item C<< "source-path"=>$string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: source-location

=item C<< "foreground"=>$string >>

Applicable for CMode CLI

=item C<< "job_component"=> \$scalar >>

Filled in for CMode CLI.

=item C<< 'extended_query' => $string >>

(Optional)

Applicable for CMode CLI

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub initialize_ls_set {
    my ($pkg, @args) = @_;
    $Log->enter();
    $pkg->_verify_invocation(style => 'static_only');
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            job_component => {type => SCALARREF, optional => 1},
            _optional_scalars(
                qw( source-cluster source-vserver source-volume
                    source-path  foreground)
            ),
        },
        ignore_primary_keys => 1,
        cli_cmd => 'snapmirror initialize-ls-set',
    );
    $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_initialize_ls_set_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_initialize_ls_set_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
                check     => '_foreground_zapi_check',
            }
        ],
    );
    $Log->exit();
} ## end sub initialize_ls_set

sub _initialize_ls_set_cmode_cli {
    $Log->enter();
    my ($pkg, @args) = @_;

    $pkg->_base_backend_cmode_cli(
        @args,
        api => 'snapmirror_initialize_ls_set'
    );

    $Log->exit();
} ## end sub _initialize_ls_set_cmode_cli

sub _initialize_ls_set_cmode_zapi {
    $Log->enter();
    my ($pkg, @args) = @_;
    my $allowed_cmode_options
        = [qw(source-cluster source-vserver source-volume)];

    $pkg->_base_backend_cmode_zapi(@args,
        api    => 'snapmirror_initialize_ls_set',
        copy   => $allowed_cmode_options,
        map    => {"source-path" => "source-location",},
    );
    $Log->exit();
} ## end sub _initialize_ls_set_cmode_zapi

=head2 protect

  NACL::C::Snapmirror->protect(command_interface=>$ci,
                              "destination-vserver"=>$dst_vs,
                              "vserver"=>$vserver,...);


Uses a CMode CLI/ZAPI APISet.

Invokes 'snapmirror_protect' API for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< schedule=> $string >>

(Optional) This optional parameter designates the name of the schedule which is
used to update the SnapMirror relationships.

=item C<< policy=> $string >>

(Optional)This optional parameter designates the name of the SnapMirror policy
which is associated with the SnapMirror relationships.

=item C<< destination-vserver=> $string >>

(Optional) This parameter specifies the Vserver in which to create the destina-
tion volumes of the SnapMirror relationships.

Applicable for CMode CLI/ZAPI.


=item C<< "source-vserver" =>$string >>

(Required) This parameter specifies the source-vserver.

Applicable for CMode CLI/ZAPI.

=item C<< "path-list" =>$string >>

(Required) This parameter specifies the snapmirror-path.

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

snapmirror-path.

Applicable for CMode CLI/ZAPI.

(Required) This parameter specifies the volume-name.

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

volume-name.

Applicable for CMode CLI/ZAPI.

=back

=back

=cut

sub protect {
    my ($pkg, @args) = @_;
    $Log->enter();
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
             _optional_scalars(
                 qw( destination-vserver policy
                     schedule )
             ),
            'path-list'  => { type => HASHREF | ARRAYREF, optional => 1 },
	},
        ignore_primary_keys => 1,
        cli_cmd => 'snapmirror protect',
    );
    $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_protect_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_protect_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            }
        ],
    );
    $Log->exit();
} ## end sub protect

sub _protect_cmode_cli {
    $Log->enter();
    my ($pkg, @args) = @_;

    $pkg->_base_backend_cmode_cli(
        @args,
        api => 'snapmirror_protect'
    );
    $Log->exit();
} ## end sub _protect_cmode_cli

sub _protect_cmode_zapi {
    $Log->enter();
    my ($pkg, @args) = @_;
    my $copy
        = [qw(destination-vserver policy schedule)];

    $pkg->_base_backend_cmode_zapi(@args,
        api    => 'snapmirror_protect',
        copy   => $copy,
        map    => {"path-list" => "snapmirror-path",},
    );
    $Log->exit();
} ## end sub _protect_cmode_zapi


=head2 update_ls_set

  NACL::C::Snapmirror->update_ls_set(command_interface=>$ci,
                              "source-vserver"=>$vserver,
                              "source-volume"=>$volume,...);

(Class method only) The method updates a LS snapmirror set.

Uses a CMode CLI/ZAPI APISet.

Invokes 'snapmirror-update-ls-set' API for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required)

See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "source-cluster" =>$string >>

(Required) Name of cluster where the source volume resides

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

source-path input would be constructed from this input in the form
(RR: cluster://volume/vserver, SN and above: not applicable) and passed to
source-location option.

=item C<< "source-vserver" =>$string >>

(Required) Name of the vserver where the source volume resides

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

source-path input would be constructed from this input in the form
(RR: cluster://volume/vserver, SN and above: vserver:volume) and passed to
'source-location' option.

=item C<< "source-volume" =>$string >>

(Required) Name of the source volume for which the snapmirror relation is created

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

source-path input would be constructed from this input in the form
(RR: cluster://volume/vserver, SN and above: vserver:volume) and passed to
'source-location' option.

=item C<< "source-path"=>$string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: source-location

=item C<< "foreground"=>$string >>

Applicable for CMode CLI

=item C<< "job_component"=> \$scalar >>

Filled in for CMode CLI.

=item C<< 'extended_query' => $string >>

(Optional)

Applicable for CMode CLI

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub update_ls_set {
    my ($pkg, @args) = @_;
    $Log->enter();
    $pkg->_verify_invocation(style => 'static_only');
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            job_component => {type => SCALARREF, optional => 1},
            _optional_scalars(
                qw( source-cluster source-vserver source-volume
                    source-path  foreground)
            ),
        },
        ignore_primary_keys => 1,
        cli_cmd => 'snapmirror update-ls-set',
    );
    $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_update_ls_set_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_update_ls_set_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
                check     => '_foreground_zapi_check',
            }
        ],
    );
    $Log->exit();
} ## end sub update_ls_set

sub _update_ls_set_cmode_cli {
    $Log->enter();
    my ($pkg, @args) = @_;

    $pkg->_base_backend_cmode_cli(
        @args,
        api => 'snapmirror_update_ls_set'
    );

    $Log->exit();
} ## end sub _update_ls_set_cmode_cli

sub _update_ls_set_cmode_zapi {
    $Log->enter();
    my ($pkg, @args) = @_;
    my $allowed_cmode_options
        = [qw(source-cluster source-vserver source-volume)];

    $pkg->_base_backend_cmode_zapi(@args,
        api    => 'snapmirror_update_ls_set',
        copy   => $allowed_cmode_options,
        map    => {"source-path" => "source-location",},
        inapplicable_for_help_xml => [qw(foreground)],
    );
    $Log->exit();
} ## end sub _update_ls_set_cmode_zapi

=head2 resume

 NACL::C::Snapmirror->resume(command_interface=>$ci,
                            source-path=>$source_path,
                            destination-path=>$destination_path);

  (or)

 NACL::C::Snapmirror->resume(command_interface=>$ci,
                             source-vserver=>$vserver,
                             source-volume=>$volume,
                             destination-path=>$destination_path);

  (or)

 $snapmirror->resume(...);

(Class or instance  method)
This method is used to enable future transfers.

Uses a CMode CLI/ZAPI (iter and non-iter supported) or 7Mode CLI/ZAPI APISet.

Invokes 'snapmirror-resume-iter' API for iter CMode ZAPI; invokes
'snapmirror-resume' API for non-iter CMode ZAPI.

Invokes 'snapmirror resume' command for 7Mode CLI.

Invokes 'snapmirror-resume' API for 7Mode ZAPI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>


=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "destination-cluster"=>$string >>

(Optional for CMode-CLI/ZAPI, Not Applicable for 7Mode-CLI/ZAPI)
Name of the Cluster where the destination volume resides

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

destination-path input would be constructed from this input in the form (RR: cluster://volume/vserver, SN and above: not applicable) and passed to destination-location option.

=item C<< "destination-vserver"=>$string >>

(Optional for CMode-CLI/ZAPI, Not Applicable for 7Mode-CLI/ZAPI)
Name of the Vserver  where the destination volume resides

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

destination-path input would be constructed from this input in the form
(RR: cluster://volume/vserver, SN and above: vserver:volume) and passed to
'destination-location' option.

=item C<< "destination-volume"=>$string >>

Name of the destination volume  which will be in snapmirror relationship

Applicable for CMode CLI/ZAPI.

Maps to:

7M CLI:

Since this option is not passed to the apilayer call ,there is no mapping
information provided ,destination-path will be constructed using this argument
and passed to the apilayer call.

7M ZAPI:

Since this option is not passed to the apilayer call ,there is no mapping
information provided ,destination-location will be constructed using this argument
and passed to the apilayer call.

CM ZAPI:

destination-path input would be constructed from this input in the form
(RR: cluster://volume/vserver, SN and above: vserver:volume) and passed to
'destination-location' option.

=item C<< "source-cluster"=>$string >>

Name of cluster where the source volume resides

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

source-path input would be constructed from this input in the form
(RR: cluster://volume/vserver, SN and above: not applicable) and passed to
source-location option.

=item C<< "source-vserver"=>$string >>

Name of the vserver where the source volume resides

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

source-path input would be constructed from this input in the form
(RR: cluster://volume/vserver, SN and above: vserver:volume) and passed to
'source-location' option.

=item C<< "source-volume"=>$string >>

Name of the source volume for which the snapmirror relation is created

Applicable for CMode CLI/ZAPI, 7Mode CLI/ZAPI.

Maps to:

7M CLI:

Since this option is not passed to the apilayer call ,there is no mapping
information provided ,source-path will be constructed using this argument
and passed to the apilayer call.

7M ZAPI:

Since this option is not passed to the apilayer call ,there is no mapping
information provided ,source-location will be constructed using this argument
and passed to the apilayer call.

CM ZAPI:

source-path input would be constructed from this input in the form
(RR: cluster://volume/vserver, SN and above: vserver:volume) and passed to
'source-location' option.

=item C<< "source-path" => $source_path >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

source-location of snapmirror-resume API.

For snapmirror-resume-iter API:

<snapmirror-resume-iter>
<query>
<snapmirror-info>
<source-location>$path</source-location>
</snapmirror-info>
</query>
</snapmirror-resume-iter>

=item C<< "hidden-type" => $hidden_type >>

Applicable for CMode CLI

=item C<< "dest-volume-dsid" => $dest_volume_dsid >>

Applicable for CMode CLI

=item C<< "dest-volume-msid" => $dest_volume_msid >>

Applicable for CMode CLI

=item C<< "dest-cluster-id" => $dest_cluster_id >>

Applicable for CMode CLI

=item C<< "dest-vserver-id" => $dest_vserver_id >>

Applicable for CMode CLI

=item C<< "dest-cluster-id" => $dest_cluster_id >>

Applicable for CMode CLI

=item C<< "job_component"=> \$scalar >>

Applicable for CMode CLI.

=item C<< 'destination-path' => $string >>

Applicable for CMode CLI/ZAPI,7Mode CLI/ZAPI.

Maps to:

CM ZAPI:

destination-location of snapmirror-resume API.

For snapmirror-resume-iter API:

<snapmirror-resume-iter>
<query>
<snapmirror-info>
<destination-location>$path</destination-location>
</snapmirror-info>
</query>
</snapmirror-resume-iter>

7M CLI: destination path (Programmatic name: "destination-path")

7M ZAPI: destination-location

=item C<< 'dest-vserver-uuid' => $string >>

=item C<< 'extended_query' => $string >>

(Optional)

Applicable for CMode CLI.

For SN and above versions this option is applicable for CMode ZAPI.

=item C<< 'operation-id' => $string >>

Operation Id
Applicable for CMode CLI

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=item C<NACL::C::Exceptions::Snapmirror::DoesNotExist>

This exception gets thrown if an attempt is made to resume a
non-existing snapmirror.

=item C<NACL::C::Exceptions::Snapmirror::DoesNotExist>

This exception gets thrown if an attempt is made to resume when the snapmirror
relationship is not quiesced.

=item C<NACL::C::Exceptions::Snapmirror::RelationshipInfoBeingPropagated>

This exception is thrown when we attempt to resume a snapmirror relation when
the relationship information is being propagated.

=back

=back

=cut

sub resume {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw(source-cluster source-vserver source-volume source-cg
                   destination-cluster destination-vserver destination-volume
                   destination-cg hidden-type dest-cluster-id dest-vserver-id
                   dest-volume-msid dest-volume-dsid dest-vserver-uuid
                   operation-id)
            ),
        },
        cli_cmd => 'snapmirror resume',
    );

    try {
        $pkg_or_obj->call_on_apiset(
            %opts,
            choices => [
                {   method    => '_resume_cmode_cli',
                    interface => 'CLI',
                    set       => 'CMode',
                    zapi_type => 'none',
                },
                {   method    => '_resume_cmode_zapi',
                    interface => 'ZAPI',
                    set       => 'CMode',
                    zapi_type => 'iter',
                },
                {   method    => '_resume_cmode_zapi_non_iter',
                    interface => 'ZAPI',
                    set       => 'CMode',
                    zapi_type => 'non-iter',
                    check     => '_non_iter_check',
                },
                {   method    => '_resume_7mode_cli',
                    interface => 'CLI',
                    set       => '7Mode',
                },
                {   method    => '_resume_7mode_zapi',
                    interface => 'ZAPI',
                    set       => '7Mode',
                },
            ],
        );
    }
    catch NACL::C::Exceptions::Component with {

        # The CMode-ZAPI iter back-end might already classify the error
        # into a Component-level exception. If already classified, we can
        # simply throw this exception rather than go through the
        # detect_convert_and_throw again.
        $Log->exit();
        $_[0]->throw;
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = $_[0];
        NACL::C::Exceptions::Snapmirror::DoesNotExist
            ->detect_convert_and_throw(exception => $exception);
        NACL::C::Exceptions::Snapmirror::NotQuiescedOrQuiescing
            ->detect_convert_and_throw(exception => $exception);
        NACL::C::Exceptions::Snapmirror::RelationshipInfoBeingPropagated
            ->detect_convert_and_throw(exception => $exception);
        $Log->exit();
        $exception->throw();
    };

    $Log->exit();
}    ## sub end resume

sub _resume_cmode_cli {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@args, api => 'snapmirror_resume');
    $Log->exit();
}

sub _resume_cmode_zapi {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_iter_helper(
          @args,
          api      => 'snapmirror-resume-iter',
          show_cmd => 'snapmirror show'
    );
    $Log->exit();
}

sub _resume_cmode_zapi_non_iter {
    $Log->enter();
    my ($pkg_or_obj, @opts) = @_;
    $pkg_or_obj->_base_backend_cmode_zapi(@opts, api => 'snapmirror-resume');
    $Log->exit();
}

sub _resume_7mode_cli {
    my $pkg_or_obj = shift;
    $Log->enter();
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my %snapmirror_args;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => [qw(destination-path)],
        target => \%snapmirror_args,
    );

    $apiset->snapmirror_resume(%snapmirror_args);
    $Log->exit();
} ## end sub _resume_7mode_cli

sub _resume_7mode_zapi {
    $Log->enter();
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );

    delete $opts{command_interface};
    my $apiset = delete $opts{apiset};

    my %args;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        map    => {"destination-path" => "destination-location",},
        target => \%args,
    );

    $apiset->snapmirror_resume(%args);
    $Log->exit();
} ## end sub _resume_7mode_zapi

=head2 quiesce

 NACL::C::Snapmirror->quiesce(command_interface=>$ci,
                              "destination-vserver"=>$vserver,
                              "destination-volume"=>$volume,
                              "source-vserver"=>$vserver,
                              "source-volume"=>$volume);
 (or)

 $snapmirror_object->quiesce(...);

 (or)

 NACL::C::Snapmirror->quiesce(command_interface=>$ci,
                              source-path=>$source_path,
                              destination-path=>$destination_path);

 (or)

 NACL::C::Snapmirror->quiesce(command_interface=>$ci,
                              source-vserver=>$vserver,
                              source-volume=>$volume,
                              destination-path=>$destination_path);


(Class or instance method)
This method is used to enable future transfers.

Uses a CMode CLI/ZAPI (iter and non-iter supported) or 7Mode CLI/ZAPI APISet.

Invokes 'snapmirror-quiesce-iter' API for iter CMode ZAPI; invokes
'snapmirror-quiesce' API for non-iter CMode ZAPI.

Invokes 'snapmirror quiesce' command for 7Mode CLI.

Invokes 'snapmirror-quiesce' API for 7Mode ZAPI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>


=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "destination-cluster"=>$string >>

(Optional for CMode-CLI/ZAPI, Not Applicable for 7Mode-CLI/ZAPI)
Name of the Cluster where the destination volume resides

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

destination-path input would be constructed from this input in the form (RR: cluster://volume/vserver, SN and above: not applicable) and passed to destination-location option.

=item C<< "destination-vserver"=>$string >>

(Optional for CMode-CLI/ZAPI, Not Applicable for 7Mode-CLI/ZAPI)
Name of the Vserver  where the destination volume resides

Applicable for CMode CLI/ZAPI

Maps to:

CM ZAPI:

destination-path input would be constructed from this input in the form
(RR: cluster://volume/vserver, SN and above: vserver:volume) and passed to
'destination-location' option.

=item C<< "destination-volume"=>$string >>

Name of the destination volume  which will be in snapmirror relationship

Applicable for CMode CLI/ZAPI, 7Mode CLI/ZAPI.

Maps to:

7M CLI:

Since this option is not passed to the apilayer call ,there is no mapping
information provided ,destination-path will be constructed using this argument
and passed to the apilayer call.

7M ZAPI:

Since this option is not passed to the apilayer call ,there is no mapping
information provided ,destination-location will be constructed using this argument
and passed to the apilayer call.

CM ZAPI:

destination-path input would be constructed from this input in the form
(RR: cluster://volume/vserver, SN and above: vserver:volume) and passed to
'destination-location' option.

=item C<< "source-cluster"=>$string >>

(Optional for class method, Not Applicable for instance method)
Name of cluster where the source volume resides

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

source-path input would be constructed from this input in the form
(RR: cluster://volume/vserver, SN and above: not applicable) and passed to
source-location option.

=item C<< "source-vserver"=>$string >>

(Optional for class method, Not Applicable for instance method)
Name of the vserver where the source volume resides

Applicable for CMode CLI/ZAPI

Maps to:

CM ZAPI:

source-path input would be constructed from this input in the form
(RR: cluster://volume/vserver, SN and above: vserver:volume) and passed to
'source-location' option.

=item C<< "source-volume"=>$string >>

(Optional for class method, Not Applicable for instance method)
Name of the source volume for which the snapmirror relation is created

Applicable for CMode CLI/ZAPI, 7Mode CLI/ZAPI.

Maps to:

7M CLI:

Since this option is not passed to the apilayer call ,there is no mapping
information provided ,source-path will be constructed using this argument
and passed to the apilayer call.

7M ZAPI:

Since this option is not passed to the apilayer call ,there is no mapping
information provided ,source-location will be constructed using this argument
and passed to the apilayer call.

CM ZAPI:

source-path input would be constructed from this input in the form
(RR: cluster://volume/vserver, SN and above: vserver:volume) and passed to
'source-location' option.

=item C<< "source-path" => $source_path >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

source-location of snapmirror-quiesce API.

For snapmirror-quiesce-iter API:

<snapmirror-quiesce-iter>
<query>
<snapmirror-info>
<source-location>$path</source-location>
</snapmirror-info>
</query>
</snapmirror-quiesce-iter>

=item C<< "hidden-type" => $hidden_type >>

Applicable for CMode CLI

=item C<< "dest-volume-dsid" => $dest_volume_dsid >>

Applicable for CMode CLI

=item C<< "dest-volume-msid" => $dest_volume_msid >>

Applicable for CMode CLI

=item C<< "dest-cluster-id" => $dest_cluster_id >>

Applicable for CMode CLI

=item C<< "dest-vserver-id" => $dest_vserver_id >>

Applicable for CMode CLI

=item C<< "dest-cluster-id" => $dest_cluster_id >>

Applicable for CMode CLI

=item C<< "job_component"=> \$scalar >>

Applicable for CMode CLI.

=item C<< 'destination-path' => $string >>

Applicable for CMode CLI/ZAPI,7Mode CLI/ZAPI.

Maps to:

CM ZAPI:

destination-location of snapmirror-quiesce API.

For snapmirror-quiesce-iter API:

<snapmirror-quiesce-iter>
<query>
<snapmirror-info>
<destination-location>$path</destination-location>
</snapmirror-info>
</query>
</snapmirror-quiesce-iter>

7M CLI: destination path (Programmatic name: "destination-path")

7M ZAPI: destination-location

=item C<< 'dest-vserver-uuid' => $string >>

Applicable for CMode CLI

=item C<< 'extended_query' => $string >>

(Optional)

Applicable for CMode CLI.

For SN and above versions this option is applicable for CMode ZAPI.

=item C<< 'operation-id' => $string >>

Operation Id
Applicable for CMode CLI

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=item C<NACL::C::Exceptions::Snapmirror::DoesNotExist>

This exception gets thrown if an attempt is made to resync a
non-existing snapmirror.

=back

=back

=cut

sub quiesce {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw(source-cluster source-vserver source-volume source-cg
                   destination-cluster destination-vserver destination-volume
                   destination-cg hidden-type dest-cluster-id dest-vserver-id
                   dest-volume-msid dest-volume-dsid dest-vserver-uuid
                   operation-id)
            ),
        },
        cli_cmd => 'snapmirror quiesce',
    );

    try {
        $pkg_or_obj->call_on_apiset(
            %opts,
            choices => [
                {   method    => '_quiesce_cmode_cli',
                    interface => 'CLI',
                    set       => 'CMode',
                    zapi_type => 'none',
                },
                {   method    => '_quiesce_cmode_zapi',
                    interface => 'ZAPI',
                    set       => 'CMode',
                    zapi_type => 'iter',
                },
                {   method    => '_quiesce_cmode_zapi_non_iter',
                    interface => 'ZAPI',
                    set       => 'CMode',
                    zapi_type => 'non-iter',
                    check     => '_non_iter_check',
                },
                {   method    => '_quiesce_7mode_cli',
                    interface => 'CLI',
                    set       => '7Mode',
                },
                {   method    => '_quiesce_7mode_zapi',
                    interface => 'ZAPI',
                    set       => '7Mode',
                },
            ],
        );
    }
    catch NACL::C::Exceptions::Component with {

        # The CMode-ZAPI iter back-end might already classify the error
        # into a Component-level exception. If already classified, we can
        # simply throw this exception rather than go through the
        # detect_convert_and_throw again.
        $Log->exit();
        $_[0]->throw;
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = $_[0];
        NACL::C::Exceptions::Snapmirror::DoesNotExist
            ->detect_convert_and_throw(exception => $exception);
        $Log->exit();
        $exception->throw();
    };

    $Log->exit();
} ## end sub quiesce

sub _quiesce_cmode_cli {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@args, api => 'snapmirror_quiesce');
    $Log->exit();
}

sub _quiesce_cmode_zapi {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_iter_helper(
          @args,
          api      => 'snapmirror-quiesce-iter',
          show_cmd => 'snapmirror show'
    );
    $Log->exit();
}

sub _quiesce_cmode_zapi_non_iter {
    $Log->enter();
    my ($pkg_or_obj, @opts) = @_;
    $pkg_or_obj->_base_backend_cmode_zapi(@opts, api => 'snapmirror-quiesce');
    $Log->exit();
}

sub _quiesce_7mode_cli {
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my %snapmirror_args;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => [qw(destination-path)],
        target => \%snapmirror_args,
    );

    $apiset->snapmirror_quiesce(%snapmirror_args);
} ## end sub _quiesce_7mode_cli

sub _quiesce_7mode_zapi {
    $Log->enter();
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );

    delete $opts{command_interface};
    my $apiset = delete $opts{apiset};

    my %args;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        map    => {"destination-path" => "destination-location",},
        target => \%args,
    );

    $apiset->snapmirror_quiesce(%args);
    $Log->exit();
} ## end sub _quiesce_7mode_zapi

=head2 on

 NACL::C::Snapmirror->on(command_interface=>$ci);

(Class method) Request to turn on the snapmirror feature.
Uses a 7Mode-CLI/ZAPI APISet.

Invokes 'snapmirror-on' API for 7Mode ZAPI.

Invokes 'snapmirror on' command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub on {
    $Log->enter();
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params              => \@args,
        ignore_primary_keys => 1,
    );

    $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => "_on_7mode_cli",
                interface => "CLI",
                set       => "7Mode",
            },
            {   method    => "_on_7mode_zapi",
                interface => "ZAPI",
                set       => "7Mode",
            }
        ],
    );
    $Log->exit();
} ## end sub on

sub _on_7mode_cli {
    $Log->enter();
    my $pkg  = shift;
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );

    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};

    $apiset->snapmirror_on(%opts);
    $Log->exit();
} ## end sub _on_7mode_cli

sub _on_7mode_zapi {
    $Log->enter();
    my $pkg  = shift;
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );

    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};

    $apiset->snapmirror_on(%opts);
    $Log->exit();
} ## end sub _on_7mode_zapi

=head2 off

 NACL::C::Snapmirror->off(command_interface=>$ci);

(Class method) Request to turn off the snapmirror feature.

Uses a 7Mode CLI/ZAPI APISet.

Invokes 'snapmirror-off' API for 7Mode ZAPI.

Invokes 'snapmirror off' command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub off {
    $Log->enter();
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params              => \@args,
        ignore_primary_keys => 1,
    );

    $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => "_off_7mode_cli",
                interface => "CLI",
                set       => "7Mode",
            },
            {   method    => "_off_7mode_zapi",
                interface => "ZAPI",
                set       => "7Mode",
            }
        ],
    );
    $Log->exit();
} ## end sub off

sub _off_7mode_cli {
    $Log->enter();
    my $pkg  = shift;
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );

    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};

    $apiset->snapmirror_off(%opts);
    $Log->exit();
} ## end sub _off_7mode_cli

sub _off_7mode_zapi {
    $Log->enter();
    my $pkg  = shift;
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );

    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};

    $apiset->snapmirror_off(%opts);
    $Log->exit();
} ## end sub _off_7mode_zapi

=head2 migrate

  NACL::C::Snapmirror->migrate(command_interface=>$ci,
                              "destination-volume"=>$volume, ...);

  (Class method) The method migrates a snapmirror relationship

Uses a 7Mode-CLI APISet.

Invokes 'snapmirror migrate' command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "destination-volume"=>$string >>

(Required )
Name of the destination volume  which is in snapmirror relationship

Applicable for 7Mode CLI.

Since this option is not passed to the apilayer call, there is no mapping
information provided, destination-path will be constructed using this argument
and passed to the apilayer call

=item C<< "source-volume"=>$string >>

(Required)
Name of the source volume which is in the snapmirror relationship

Applicable for 7Mode CLI.

Since this option is not passed to the apilayer call, there is no mapping
information provided, source-path will be constructed using this argument
and passed to the apilayer call

=item C<< "max-transfer-rate"=>$scalar >>

Applicable for 7Mode CLI.

Maps to:

7M CLI: -k <value> (Programmatic name: max-transfer-rate)

=item C<< "test-run"=>$boolean >>

Applicable for 7Mode CLI.

Maps to:

7M CLI: -n (Programmatic name: test-run)

=item C<< force=>$boolean >>

Applicable for 7Mode CLI.

Maps to:

7M CLI: -f (Programmatic name: force)

=item C<< "source-path"=>$scalar >>

Applicable for 7Mode CLI.

Maps to:

7M CLI: sorce path (Programmatic name: sorce-path)

=item C<< "destination-path"=>$scalar >>

Applicable for 7Mode CLI.

Maps to:

7M CLI: destination path (Programmatic name: destination-path)

=back

=back

=cut

sub migrate {

    my $pkg_or_obj = shift;
    $Log->enter();
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@_,
        additional_spec => {
            _optional_scalars(
                qw( test-run force max-transfer-rate source-volume destination-volume )
            )
        }
    );

    $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => "_migrate_7mode_cli",
                interface => "CLI",
                set       => "7Mode",
            },
        ],
    );
    $Log->exit();

} ## end sub migrate

sub _migrate_7mode_cli {
    my $pkg_or_obj = shift;
    $Log->enter();
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my $snapmirror_args   = $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => [
            qw( test-run force max-transfer-rate
                source-path destination-path)
        ],
    );
    $apiset->snapmirror_migrate(%{$snapmirror_args});
    $Log->exit();
} ## end sub _migrate_7mode_cli

=head2 convert

 NACL::C::Snapmirror->convert(command_interface=>$ci,
                             "destination-volume"=>$volume, ...);

                            or

 $snapmirror_object->convert(...);

 (Class or instance method) The method migrates a snapmirror relationship

Uses a 7Mode APISet.

Invokes 'snapmirror convert' command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "destination-volume"=>$string >>

(Required for class method, Not Applicable for instance method)
Name of the destination volume  which is in snapmirror relationship

Applicable for 7Mode CLI

Maps to:

7M CLI:

Since this option is not passed to the apilayer call ,there is no mapping
information provided ,destination-path will be constructed using this argument
and passed to the apilayer call.

=item C<< "destination-path"=>$scalar >>

Applicable for 7Mode CLI.

Maps to:

7M CLI: destination path (Programmatic name: destination-path)

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub convert {
    my $pkg_or_obj = shift;
    $Log->enter();
    my %opts = $pkg_or_obj->_common_validate_with(
        params             => \@_,
        additional_spec    => {_optional_scalars(qw( destination-volume ))},
        ignore_primary_key => 1
    );

    $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => "_convert_7mode_cli",
                interface => "CLI",
                set       => "7Mode",
            },
        ],
    );
    $Log->exit();

} ## end sub convert

sub _convert_7mode_cli {
    my $pkg_or_obj = shift;
    $Log->enter();
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my $snapmirror_args   = $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => [qw(destination-path)],
    );
    $apiset->snapmirror_convert(%$snapmirror_args);
    $Log->exit();
} ## end sub _convert_7mode_cli

=head2 throttle

  NACL::C::Snapmirror->throttle(command_interface=>$ci,
                              "destination-volume"=>$volume, ...);

                            or

  $snapmirror_object->throttle(...);

  (Class or instance method) The method migrates a snapmirror relationship

Uses a 7Mode-CLI/ZAPI APISet.

Invokes 'snapmirror throttle' command for 7Mode CLI.

Invokes "snapmirror-throttle" API for 7Mode ZAPI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "destination-path"=>$string >>

(Applicable for class method, Not Applicable for instance method)

(Required for class method, Not Applicable for instance method)
Name of the destination volume  which is i
n snapmirror relationship.

Applicable for 7Mode CLI/ZAPI.

Maps to:

7M CLI: destination path (Programmatic name: destination-path)

7M ZAPI: destination-location

=item C<< "transfer-rate"=>$scalar >>

Applicable for 7Mode CLI/ZAPI.

Maps to:

7M CLI: -k <value> (Programmatic name: max-transfer-rate)

7M ZAPI: max-transfer-rate.

=item C<< "destination-volume"=>$string >>

(Required )
Name of the destination volume  which is in snapmirror relationship

Applicable for 7Mode CLI/ZAPI.

Maps to:

7M CLI:

Since this option is not passed to the apilayer call, there is no mapping
information provided, destination-path will be constructed using this argument
and passed to the apilayer call.

7M ZAPI:

Since this option is not passed to the apilayer call, there is no mapping
information provided, destination-location will be constructed using this argument
and passed to the apilayer call.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub throttle {
    my $pkg_or_obj = shift;
    $Log->enter();
    my %opts = $pkg_or_obj->_common_validate_with(
        params => \@_,
        optional_scalars(qw( transfer-rate destination-path )),
        ignore_primary_key => 1
    );

    $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => "_throttle_7mode_cli",
                interface => "CLI",
                set       => "7Mode",
            },
            {   method    => "_throttle_7mode_zapi",
                interface => "ZAPI",
                set       => "7Mode",
            },
        ],
    );
    $Log->exit();
} ## end sub throttle

sub _throttle_7mode_cli {
    my $pkg_or_obj = shift;
    $Log->enter();
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my $snapmirror_args   = $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => [qw( transfer-rate destination-path)],
    );

    $apiset->snapmirror_throttle(%$snapmirror_args);
    $Log->exit();
} ## end sub _throttle_7mode_cli

sub _throttle_7mode_zapi {
    my $pkg_or_obj = shift;
    $Log->enter();
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my $snapmirror_args   = $pkg_or_obj->_hash_copy(
        source => \%opts,
        maps   => {
            'destination-path' => 'destination-location',
            'transfer-rate'    => 'max-transfer-rate',
        },
    );

    $apiset->snapmirror_throttle(%$snapmirror_args);
    $Log->exit();
} ## end sub _throttle_7mode_zapi

sub _common_validate_with {
    $Log->enter();
    my $pkg_or_obj       = shift;
    my $destination_flag = 0;
    my %opts             = validate_with(
        params => \@_,
        spec   => {
            params              => {type => ARRAYREF},
            additional_spec     => {type => HASHREF, default => {}},
            ignore_primary_keys => {type => SCALAR, default => 0},
        },
        allow_extra => 1
    );

    my %orig_params = @{$opts{params}};
    if ($orig_params{"source-cluster"} || $orig_params{"destination-cluster"})
    {
        $Log->warn("source-cluster and destination-cluster are no longer valid.");
        delete $orig_params{"source-cluster"};
        delete $orig_params{"destination-cluster"};
    }

 #
 # This is a hack, since path is the primary key ,if path has not been passed
 # we insert a dummy path ,so that SUPER::_common_validate_with  wont throw an
 # error
 #
    if (!ref $pkg_or_obj && !$opts{ignore_primary_keys}) {
        if (!$orig_params{'destination-path'}) {
            $orig_params{'destination-path'} = "dummy";
            $destination_flag = 1;
        }
    }

    my %params = $pkg_or_obj->SUPER::_common_validate_with(%opts,
        params => [%orig_params]);
    delete $params{'destination-path'} if ($destination_flag);

    if (!$opts{ignore_primary_keys}) {
        if ($params{'command_interface'}->is_cmode()) {
            if (   !$params{'destination-path'}
                && !$params{'destination-vserver'})
            {
                $Log->exit();
                NATE::Exceptions::Argument->throw(
                          "When 'destination-path' is not provided, then "
                        . 'destination-vserver must be provided but it is missing'
                );
            }

            foreach my $src_or_dest (qw(source destination)) {
                if (!$params{"$src_or_dest-path"}) {
                    $pkg_or_obj->_construct_single_path(
                        params => \%params,
                        type   => $src_or_dest
                    );
                }
            }
        } else {    #7Mode cli
            if (!$params{'source-path'} && defined $params{'source-volume'}) {
                $params{'source-path'} = delete $params{'source-volume'};
            }
            if (!$params{'destination-path'}
                && defined $params{'destination-volume'})
            {
                $params{'destination-path'}
                    = delete $params{'destination-volume'};
            } ## end if ( !$params{'destination-path'...
        } ## end else [ if ( $params{'command_interface'...
    } ## end if ( !$opts{ignore_primary_keys...

    # If the "cg-item-mappings" option was given, it was already validated as
    # a hash reference. But for the CLI and ZAPI code, we actually want an
    # array reference of strings joined on the sequence ":@".
    if (ref($params{'cg-item-mappings'}) eq 'HASH') {
        $params{'cg-item-mappings'} = [
            map { "$_:\@$params{'cg-item-mappings'}->{$_}" }
                keys %{$params{'cg-item-mappings'}}
        ];
    }

    $Log->exit();
    return %params;
} ## end sub _common_validate_with

# These fields cannot be added to _common_validate_spec since they end up
# causing validation failures on the CS side, hence this extra spec for
# options that are common to all methods on this Component on the C side.
sub _additional_common_validate_spec {
    my ($pkg_or_obj, @opts) = @_;
    my $spec = $pkg_or_obj->SUPER::_additional_common_validate_spec(@opts);
    my $iv_spec = $pkg_or_obj->_add_infinivol_params_to_additional_common_validate_spec(@opts);
    $spec = {%$iv_spec, %$spec};
    $spec->{source_ci} = {
        isa      => 'NACL::C::CommandInterface::ONTAP',
        type     => OBJECT,
        optional => 1
    };
    return $spec;
}

=head2 release

    NACL::C::Snapmirror->release(
                                command_interface => $ci,
                                'destination-path' => $destination_path,
                                'source-path' => $source_path,
                                ...);

                                 or

    $Snapmirror_obj->release(...);

(Class or instance method)
This method is used to release a snapmirror relationship.
Only for cases where source and destination resides in the same filer, instance method call will work.
If they are different, the users are expected to use static method call.

Uses a CMode CLI/ZAPI (iter and non-iter supported) or 7Mode CLI/ZAPI APISet.

Invokes 'snapmirror-release-iter' API for iter CMode ZAPI; invokes
'snapmirror-release' API for non-iter CMode ZAPI.

Invokes 'snapmirror release' command for 7Mode CLI.

Invokes 'snapmirror-release' API for 7Mode ZAPI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'destination-path' => $destination_path >>

Applicable for CMode CLI/ZAPI, 7Mode CLI/ZAPI.

Maps to:

7M CLI: destination path (Programmatic name: "destination-path")

7M ZAPI: destination-location

CM ZAPI: destination-location

(Required for class method, Not Applicable for instance method)

=item C<< 'source-path' => $source_path >>

Applicable for CMode CLI/ZAPI, 7Mode CLI/ZAPI.

Maps to:

7M CLI: source path (Programmatic name: "source-path")

7M ZAPI: source-location

CM ZAPI: source-location

(Required for class method, Not Applicable for instance method)

=item C<< 'destination-volume' => $string >>

Applicable for CMode CLI/ZAPI and 7Mode CLI.

Maps to:

CM ZAPI: destination-volume

7M CLI: destination volume (Programmatic name: "destination-volume")

=item C<< 'source-volume' => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: source-volume

=item C<< 'destination-vserver' => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: destination-vserver

=item C<< 'source-vserver' => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: source-vserver

=item C<< destination-cg =E<gt> $string >>

(Optional) Name of destination consistency group for a granular relationship.

Applicable only to CMode CLI/ZAPI.

This option does not map to an API-layer option. If passed (and if
C<destination-path> is not already specified), this is used with the
C<destination-vserver> value to form the destination path as
C<$vserver:/cg/$cg_name>. If this parameter is present, it is given
precedence over C<destination-volume> (but only for CMode).

=item C<< source-cg =E<gt> $string >>

(Optional) Name of source consistency group for a granular relationship.

Applicable only to CMode CLI/ZAPI.

This option does not map to an API-layer option. If passed (and if
C<source-path> is not already specified), this is used with the
C<source-vserver> value to form the source path as C<$vserver:/cg/$cg_name>. If
this parameter is present, it is given precedence over C<source-volume> (but
only for CMode).

=item C<< 'hidden-type' => $string >>

Applicable for CMode CLI

=item C<< 'dest-volume-dsid' => $string >>

Applicable for CMode CLI

=item C<< 'dest-volume-msid' => $string >>

Applicable for CMode CLI

=item C<< 'dest-cluster-id' => $string >>

Applicable for CMode CLI

=item C<< 'force' => $string >>

Applicable for CMode CLI

=item C<< 'dest-vserver-uuid' => $string >>

Applicable for CMode CLI

=item C<< 'dest-vserver-id' => $string >>

Applicable for CMode CLI

=item C<< 'source-cluster' => $string >>

Applicable for CMode CLI/ZAPI.

Since its filled by backend method, mapping information not provided.

=item C<< 'destination-cluster' => $string >>

Applicable for CMode CLI/ZAPI.

Since its filled by backend method, mapping information not provided.

=item C<< 'method-timeout' => $time_in_seconds >>

(Optional)
The default NATE timeout is 60 seconds. If the command might take more
time to complete, this option can be used to specify a larger timeout value.
The value should be provided in seconds.

=item C<< 'rel-only' => $string >>

Applicable for CMode CLI

=item C<< 'relationship-id' => $string >>

Relationship Id

Applicable for CMode CLI/ZAPI

=item C<< 'relationship-info-only' => $string >>

Remove relationship info only (skip cleanup of snapshots)

Applicable for CMode CLI/ZAPI

=item C<< 'hidden-rel-id' => $string >>

Applicable for CMode CLI

=item C<< 'extended_query' => $string >>

(Optional)

Applicable for CMode CLI

=item C<< 'operation-id' => $string >>

Operation Id
Applicable for CMode CLI

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=item C<NACL::C::Exceptions::Snapmirror::NoReleasableDestinationFound>

This exception gets thrown if an attempt is made to release snapmirror relationship
when there are no releasable destinations.

=item C<NACL::C::Exceptions::Snapmirror::SnapshotInUse>

This exception gets thrown if an attempt is made to release a
snapmirror when a snapshot in the volume is locked.

=item C<NACL::C::Exceptions::Snapmirror::StaleCache>

This exception gets thrown if an attempt is made to release a
snapmirror when stale cache entry exists.

=item C<NACL::C::Exceptions::Snapmirror::DeleteIsPending>

This exception is thrown when an attempt is made to perform snapmirror
operation when delete is in progress.

=back

=back

=cut

sub release {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw(destination-volume destination-vserver destination-cg
                   source-volume source-vserver source-cg hidden-type
                   dest-volume-dsid dest-volume-msid dest-cluster-id force
                   dest-vserver-uuid dest-vserver-id rel-only
                   relationship-id relationship-info-only hidden-rel-id
                   operation-id)
            ),
            'cg-item-mappings' => { type => HASHREF | ARRAYREF, optional => 1 },
        },
        cli_cmd => 'snapmirror release',
    );
    try {
        $pkg_or_obj->call_on_apiset(
            %opts,
            choices => [
                {   method    => '_release_cmode_cli',
                    interface => 'CLI',
                    set       => 'CMode',
                    zapi_type => 'none',
                },
                {   method    => '_release_cmode_zapi',
                    interface => 'ZAPI',
                    set       => 'CMode',
                    zapi_type => 'iter',
                },
                {   method    => '_release_cmode_zapi_non_iter',
                    interface => 'ZAPI',
                    set       => 'CMode',
                    zapi_type => 'non-iter',
                    check     => '_non_iter_check',
                },
                {   method    => '_release_7mode_cli',
                    interface => 'CLI',
                    set       => '7Mode',
                },
                {   method    => '_release_7mode_zapi',
                    interface => 'ZAPI',
                    set       => '7Mode',
                },
            ],
            frontend => 'NACL::C::Snapmirror::release',
        );
    }
    catch NACL::C::Exceptions::Component with {

        # The CMode-ZAPI iter back-end might already classify the error
        # into a Component-level exception. If already classified, we can
        # simply throw this exception rather than go through the
        # detect_convert_and_throw again.
        $Log->exit();
        $_[0]->throw;
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = $_[0];
        foreach my $type (qw( NoReleasableDestinationFound SnapshotInUse StaleCache DeleteIsPending)) {
            my $full_type = "NACL::C::Exceptions::Snapmirror::$type";
            $full_type->detect_convert_and_throw(exception => $exception);
        }
        $Log->exit();
        $exception->throw();
    };
    $Log->exit();
} ## end sub release

sub _release_cmode_cli {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@args, api => 'snapmirror_release');
    $Log->exit();
}

sub _release_cmode_zapi {
    $Log->enter();
    my ($pkg_or_obj, %opts) = @_;

    # The -release-iter ZAPI does not take snapmirror-info typedef, but
    # instead takes snapmirror-destination-info typedef, so we need to get
    # the copy/map.
    my $copy = $pkg_or_obj->_destination_common_zapi_copy();
    my $map  = $pkg_or_obj->_destination_common_zapi_map();

    $pkg_or_obj->_iter_helper(
        %opts,
        api      => 'snapmirror-release-iter',
        show_cmd => 'snapmirror show',
        copy     => $copy,
        map      => $map,
        additional_copy => [qw(relationship-info-only )],
        pass_in_query   => [ qw( relationship-id ) ],

        # Passed down by the STask, but not applicable to the API call
        inapplicable_for_help_xml => [qw(src-path)],
    );
    $Log->exit();
}

sub _release_cmode_zapi_non_iter {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_base_backend_cmode_zapi(
        @args,
        api             => 'snapmirror-release',
        additional_copy => [qw(relationship-id relationship-info-only)],

        # Passed down by the STask, but not applicable to the API call
        inapplicable_for_help_xml => [qw(src-path)],
    );
    $Log->exit();
}

sub _release_7mode_cli {
    $Log->enter();
    my $pkg_or_obj = shift;
    my %opts       = $pkg_or_obj->_validate_backend_opts(@_);

    delete $opts{command_interface};
    my $apiset = delete $opts{apiset};

    my $allowed_7mode_options = [qw(source-path destination-path)];

    my %snap_args;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => $allowed_7mode_options,
        target => \%snap_args,
    );

    if ($snap_args{'source-path'} =~ /(.*):(.*)/) {
        $snap_args{'source-path'} = $2;
    }

    $apiset->snapmirror_release(%snap_args);
    $Log->exit();
} ## end sub _release_7mode_cli

sub _release_7mode_zapi {
    my $pkg_or_obj = shift;
    $Log->enter();
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my $snapmirror_args   = $pkg_or_obj->_hash_copy(
        source => \%opts,
        map    => {
            "source-path"      => "source-location",
            "destination-path" => "destination-location",
        },
    );

    if (defined  $snapmirror_args->{'source-path'} &&
          $snapmirror_args->{'source-path'} =~ /(.*):(.*)/) {
        $snapmirror_args->{'source-path'} = $2;
    }

    $apiset->snapmirror_release(%$snapmirror_args);
    $Log->exit();
} ## end sub _release_7mode_zapi

=head2 restore

    # Restore involving intra-cluster snapmirror relationship
    my $obj = NACL::C::Snapmirror->restore(
                                command_interface => $ci,
                                'destination-path' => $destination_path,
                                'source-path' => $source_path,
                                ...);

    # Restore involving inter-cluster snapmirror relationship
    my $obj = NACL::C::Snapmirror->restore(
                                command_interface => $ci,
                                'destination-path' => $destination_path,
                                'source-path' => $source_path,
                                'source_ci'   => $source_ci,
                                ...);

(Class method)
This method is used to restore of a Snapshot from a source volume to a
destination volume. It returns a snapmirror object which represents the
transient SM relationship established by ONTAP for restore operation.

Uses a CMode CLI/ZAPI APISet.

Invokes 'snapmirror-restore' API for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>
This represents the source end point (source cluster /source node / source
vserver) of the snapmirror relationship.

=item C<< source_ci => $source_ci >>

(Optional)
Represents command interface of the source cluster or node.
This option would be of interest mainly if you are performing
restore on a intercluster relationship. This command interface
would be used in the returned snapmirror object.
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'destination-path' => $destination_path >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: destination-location

=item C<< 'source-path' => $source_path >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: source-location

=item C<< 'destination-volume' => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: destination-volume

=item C<< 'source-volume' => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: source-volume

=item C<< 'destination-vserver' => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: destination-vserver

=item C<< 'hidden-type' => $string >>

Applicable for CMode CLI.

=item C<< 'disable-storage-efficiency' => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: disable-storage-efficiency

=item C<< 'source-vserver' => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: source-vserver

=item C<< 'dest-volume-msid' => $string >>

Applicable for CMode CLI.

=item C<< 'dest-vserver-id' => $string >>

Applicable for CMode CLI.

=item C<< 'snapshot' => $string >>

Applicable for CMode CLI.

=item C<< 'destination-cluster' => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: destination-cluster

=item C<< 'source-cluster' => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: source-cluster

=item C<< 'dest-volume-dsid' => $string >>

Applicable for CMode CLI.

=item C<< 'transfer-priority' => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: transfer-priority

=item C<< 'time-to-complete' => $string >>

Applicable for CMode CLI.

=item C<< 'clean-up-failure' => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: clean-up-failure

=item C<< 'dest-cluster-id' => $string >>

Applicable for CMode CLI.

=item C<< 'dest-vserver-uuid' => $string >>

Applicable for CMode CLI.

=item C<< 'throttle' => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: max-transfer-rate

=item C<< 'method-timeout' => $time_in_seconds >>

=item C<< "job_component"=> \$scalar >>

(Optional)
The default NATE timeout is 60 seconds. If the command might take more
time to complete, this option can be used to specify a larger timeout value.
The value should be provided in seconds.

Applicable for CMode CLI/ZAPI.

=item C<< 'force' => $string >>

Applicable for CMode CLI.

=item C<< 'foreground' => $string >>

Applicable for CMode CLI.

=item C<< 'source-snapshot' => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: source-snapshot

=item C<< 'file-list' => [ $file-1, $file-2, ... ] >>

File List
(Arrayref, even if only a single value is to be sent it should specified
as an arrayref containing that single value)
Applicable for CMode CLI

=item C<< 'create-check-snapshot' => $string >>

Create Snapshot for Check Command
Applicable for CMode CLI

=item C<< 'file-restore-clean-up' => $string >>

Clean Up File Restore Information
Applicable for CMode CLI

=item C<< 'operation-id' => $string >>

Operation Id
Applicable for CMode CLI

=item C<< 'use-network-compression' => $string >>

Use Network Compression
Applicable for CMode CLI

=item C<< 'force-baseline-file-restore' => $string >>

Force a Baseline File Restore
Applicable for CMode CLI

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub restore {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            'file-list'   => {type => ARRAYREF, optional => 1},
            job_component => {type => SCALARREF, optional => 1},
            _optional_scalars(
                qw(source-cluster source-vserver source-volume source-cg
                   destination-cluster destination-vserver destination-volume
                   destination-cg hidden-type disable-storage-efficiency
                   dest-volume-msid dest-vserver-id snapshot dest-volume-dsid
                   transfer-priority time-to-complete clean-up-failure
                   dest-cluster-id dest-vserver-uuid throttle force foreground
                   source-snapshot tries create-check-snapshot
                   file-restore-clean-up operation-id use-network-compression
                   force-baseline-file-restore)
            ),
        },
        cli_cmd => 'snapmirror restore',
    );

    my $source_ci = $pkg_or_obj->_get_source_ci(%opts);
    $opts{command_interface} = $source_ci;

    # Restore has to be run in vserver or cluster context
    # where the restore destinaiton exists
    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_restore_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_restore_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
                check     => '_foreground_zapi_check',
            },
        ],
    );

    # Create snapmirror restore object
    my %new_opts;
    $pkg_or_obj->_hash_copy_defined(
        source => \%opts,
        target => \%new_opts,
        copy   => [qw(source-path destination-path)]
    );
    $ret = $pkg_or_obj->new(%new_opts,command_interface => $source_ci);
    $Log->exit();
    return $ret;

} ## end sub restore

sub _restore_cmode_cli {
    $Log->enter();
    my($pkg_or_obj, %opts) = @_;

    $pkg_or_obj->_base_backend_cmode_cli(
        %opts,
        api => 'snapmirror_restore'
    );

    $Log->exit();
} ## end sub _restore_cmode_cli

sub _restore_cmode_zapi {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;

    $pkg_or_obj->_base_backend_cmode_zapi(
        @args,
        api  => 'snapmirror-restore',
        copy => [
            qw(disable-storage-efficiency transfer-priority
                clean-up-failure source-snapshot tries create-check-snapshot)
        ],
        map => {
            "source-path"      => "source-location",
            "destination-path" => "destination-location",
            "throttle"         => "max-transfer-rate",
            "file-list"        => "restore-file-list",
        },
        frontend_method => 'restore',
    );

    $Log->exit();
} ## end sub _restore_cmode_zapi

=head2 list_destinations

    my $list_destinations_obj = NACL::C::Snapmirror->list_destinations(command_interface => $ci,
                          [ filter => { %filter } ,]
              [ requested_fields => [ @req_fields ]);

                    or

    my @list_destinations_objs = NACL::C::Snapmirror->list_destinations(command_interface => $ci,
                          [ filter => { %filter } ,]
              [ requested_fields => [ @req_fields ]);

                    or

   my $list_destinations_obj = $Snapmirror_obj->list_destinations();

(Class or Instance method)
This method returns objects of the type L<NACL::CS::SnapmirrorListDestinations|lib-NACL-CS-SnapmirrorListDestinations-pm>.

This method accepts all the parameters accepted by L<fetch|lib-NACL-CS-ComponentState-pm/fetch>
(requested_fields, filter, allow_empty) and has all of the other properties of
fetch methods (callable in scalar or list context, throws a NoElementsFound
exception if no matching elements are found)

=cut

sub list_destinations {
    $Log->enter();
    my $pkg_or_obj = shift;
    my %opts       = $pkg_or_obj->_handle_obj_cs_like_calls(@_);
    require NACL::CS::SnapmirrorListDestinations;
    my @state_objs = NACL::CS::SnapmirrorListDestinations->fetch(%opts);
    $Log->exit();
    return wantarray ? @state_objs : $state_objs[0];
} ## end sub list_destinations

=head2 get_volume_status

    NACL::C::Snapmirror->get_volume_status(
                                command_interface => $ci,
                                ...);

(Class method)
This method is used to shows SnapMirror status values for a volume.
Uses a CMode CLI APISet.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'is-transfer-in-progress' => $string >>

Transfer in progress possible value(s) are, true,false
Applicable for CMode CLI

=item C<< 'volume-path' => $string >>


Volume Path possible value(s) are, <[vserver:]volume>,<[cluster:][//vserver/]volume>
Applicable for CMode CLI

=item C<< 'is-destination' => $string >>

Volume is a SnapMirror destination possible value(s) are, true,false
Applicable for CMode CLI

=item C<< 'is-source' => $string >>

Volume is a SnapMirror source possible value(s) are, true,false
Applicable for CMode CLI

=item C<< 'is-transfer-broken' => $string >>

Relationship is broken off possible value(s) are, true,false
Applicable for CMode CLI

=item C<< 'cmd_execution_time' => \$scalar >>

(Optional) Scalar reference , it will be updated with the command execution
time returned by the apiset response object

=item C<< 'method-timeout' => $time_in_seconds >>

(Optional)
The default NATE timeout is 60 seconds. If the command might take more
time to complete, this option can be used to specify a larger timeout value.
The value should be provided in seconds.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub get_volume_status {
    $Log->enter();
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            'cmd_execution_time' => {type => SCALARREF, optional => 1},
            _optional_scalars(
                qw( is-transfer-in-progress volume-path is-destination is-source
                    is-transfer-broken
                    )
            ),
        },
        ignore_primary_keys => 1,
        cli_cmd             => 'snapmirror get-volume-status',
    );
    $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_get_volume_status_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::Snapmirror::get_volume_status',
    );

    $Log->exit();
} ## end sub get_volume_status

sub _get_volume_status_cmode_cli {
    $Log->enter();
    my ($pkg, @opts) = @_;

    $pkg->_base_backend_cmode_cli(
        @opts,
        api => 'snapmirror_get_volume_status'
    );

    $Log->exit();
} ## end sub _get_volume_status_cmode_cli

=head2 set_options

    NACL::C::Snapmirror->set_options(
                                command_interface => $ci,
                                ...);

(Class method)
This method is used to display/Set SnapMirror options.
Uses CMode CLI .

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'xdp-source-reserve-pct' => $string >>


Percentage Reserved for XDP Source Transfers possible value(s) are, 0,25%,50%,75%,100%
Applicable for CMode CLI

=item C<< 'dp-destination-reserve-pct' => $string >>


Percentage Reserved for DP Destination Transfers possible value(s) are, 0,25%,50%,75%,100%
Applicable for CMode CLI

=item C<< 'xdp-destination-reserve-pct' => $string >>


Percentage Reserved for XDP Destination Transfers possible value(s) are, 0,25%,50%,75%,100%
Applicable for CMode CLI

=item C<< 'dp-source-reserve-pct' => $string >>


Percentage Reserved for DP Source Transfers possible value(s) are, 0,25%,50%,75%,100%
Applicable for CMode CLI

=item C<< 'xdp-destination-xfer-reserve-pct' => $string >>


Percentage Reserved for XDP Destination Transfers possible value(s) are, 0,25%,50%,75%,100%
Applicable for CMode CLI

=item C<< 'dp-destination-xfer-reserve-pct' => $string >>


Percentage Reserved for DP Destination Transfers possible value(s) are, 0,25%,50%,75%,100%
Applicable for CMode CLI

=item C<< 'dp-source-xfer-reserve-pct' => $string >>


Percentage Reserved for DP Source Transfers possible value(s) are, 0,25%,50%,75%,100%
Applicable for CMode CLI

=item C<< 'xdp-source-xfer-reserve-pct' => $string >>


Percentage Reserved for XDP Source Transfers possible value(s) are, 0,25%,50%,75%,100%
Applicable for CMode CLI

=item C<< '*' >>

This method takes all other options supported by "snapmirror set options" command in ONTAP

=item C<< other_options >>

This method takes other common options described in L<Common method options|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#Other_options_valid_on_component_methods>

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub set_options {
    $Log->enter();
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( xdp-source-reserve-pct dp-destination-reserve-pct
                    xdp-destination-reserve-pct dp-source-reserve-pct

                    xdp-destination-xfer-reserve-pct dp-destination-xfer-reserve-pct
                    dp-source-xfer-reserve-pct xdp-source-xfer-reserve-pct
                    )
            ),
        },
        ignore_primary_keys => 1,
        cli_cmd             => 'snapmirror set-options',
    );
    $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_set_options_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::Snapmirror::set_options',
    );

}

sub _set_options_cmode_cli {
    $Log->enter();
    my ($pkg, @opts) = @_;
    $pkg->_base_backend_cmode_cli(@opts, api => 'snapmirror_set_options',);
    $Log->exit();
}

=head2 show_history

    my $show_history_obj = NACL::C::Snapmirror->show_history(command_interface => $ci,
                          [ filter => { %filter } ,]
			  [ requested_fields => [ @req_fields ]);

                    or

    my @show_history_objs = NACL::C::Snapmirror->show_history(command_interface => $ci,
                          [ filter => { %filter } ,]
			  [ requested_fields => [ @req_fields ]);

                    or

   my $show_history_obj = $Snapmirror_obj->show_history();

(Class or Instance method)

This method is a CS-like fetch method which accepts all the parameters of
L<fetch|lib-NACL-CS-ComponentState-pm/fetch> (requested_fields, filter, allow_empty)
and has all of the other properties of fetch methods (callable in scalar or list
context, throws a NoElementsFound exception if no matching elements are found).
The requested fields and filter takes L<ATTRIBUTES|lib-NACL-CS-SnapmirrorShowHistory-pm/ATTRIBUTES>
of NACL::CS::SnapmirrorShowHistory as input.
This method returns objects of the type L<NACL::CS::SnapmirrorShowHistory|lib-NACL-CS-SnapmirrorShowHistory-pm>.

=cut

sub show_history {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    my %opts = $pkg_or_obj->_handle_obj_cs_like_calls(@args);
    require NACL::CS::SnapmirrorShowHistory;
    my @state_objs = NACL::CS::SnapmirrorShowHistory->fetch(%opts);
    $Log->exit();
    return wantarray ? @state_objs : $state_objs[0];
}

=head2 update_ns_mirrors

    NACL::C::Snapmirror->update_ns_mirrors(
                                command_interface => $ci,
                                'source-vserver' => $source_vserver,
                                'source-volume' => $source_volume,
                                ...);

(Class method)
This method is used to start an incremental ns mirror transfer.
Uses CMode CLI .

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'source-vserver' => $source_vserver >>

(Required)
Source Vserver
Applicable for CMode CLI

=item C<< 'source-volume' => $source_volume >>

(Required)
Source Volume
Applicable for CMode CLI

=item C<< 'transfer-snapshot' => $string >>

Transfer Snapshot
Applicable for CMode CLI

=item C<< '*' >>

This method takes all other options supported by "snapmirror update ns mirrors" command in ONTAP

=item C<< other_options >>

This method takes other common options described in L<Common method options|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#Other_options_valid_on_component_methods>

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub update_ns_mirrors {
    $Log->enter();
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( source-volume transfer-snapshot source-vserver
                    )
            ),
        },
        ignore_primary_keys => 1,
        cli_cmd             => 'snapmirror update-ns-mirrors',
    );
    $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_update_ns_mirrors_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::Snapmirror::update_ns_mirrors',
    );

    $Log->exit();
}

sub _update_ns_mirrors_cmode_cli {
    $Log->enter();
    my ($pkg, @opts) = @_;
    $pkg->_base_backend_cmode_cli(@opts,
        api => 'snapmirror_update_ns_mirrors',);
    $Log->exit();
}


__PACKAGE__->_load_event_class();

sub _primary_keys_validate_spec {
    return (
        'destination-path' => {type => SCALAR},
        'source-path'      => {type => SCALAR, optional => 1},
    );
} ## end sub _primary_keys_validate_spec

#
#The following accessor method are required to be backward
#compatible for existing users
#

=head2 destination_cluster

    my $dst_cluster = NACL::C::Snapmirror->destination_cluster(
                       'command_interface'  => $ci,
                       'destination-path'   => $dst_path,
                    );
                    (or)

    my $dst_cluster = $snapmirror_obj->destination_cluster();

(Class or instance method) Returns the Destination cluster name,
if the format of the path is valid and throws exception of type
NATE::Exceptions::Argument if it is not valid.

Supports CMode CLI/ZAPI.

=over

=item Options

=over

=item C<< 'destination-path' => $path >>

(Required for Class method)

The path format can be in the form "clustername://vserver/volume" or
"vserver:volume"

=item C<< command_interface => $command_interface >>

(Required for class method and if path is SN style (vserver:volume),
 Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=back

=back

=cut

sub destination_cluster() {
    $Log->enter();
    my $pkg_or_obj = shift;
    my %opts       = $pkg_or_obj->_common_validate_with(
        params          => \@_,
        additional_spec => {
            'source-path'      => {type => OBJECT | SCALAR, optional => 1},
            'destination-path' => {type => OBJECT | SCALAR, optional => 0},
        },
        ignore_primary_keys => 1,
    );
    my $dst_cluster = $pkg_or_obj->_get_cluster_name(
        %opts,
        is_source => 0,
        path      => $opts{'destination-path'},

    );
    $Log->exit();
    return $dst_cluster;
} ## end sub destination_cluster

=head2 destination_vserver

    my $dst_vserver = NACL::C::Snapmirror->destination_vserver(
        'destination-path' => $path,
    );

                or

    my $dst_vserver = $snapmirror_obj->destination_vserver();

(Class or instance method) Returns the destination vserver name,
if the format of the path is valid  and throws exception of type
NATE::Exceptions::Argument, if it is not valid.

Supports CMode CLI/ZAPI.

=over

=item Options

=over

=item C<< "destination-path" = $destination_path >>

(Required for class method, Not Applicable for instance method)

=item C<< command_interface => $command_interface >>

(Required for class method and if path is SN style (vserver:volume),
 Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=back

=back

=cut

sub destination_vserver {
    my ($pkj_or_obj, @args) = @_;
    my %opts = $pkj_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {'destination-path' => {type => OBJECT | SCALAR},},
    );

    my $path = $opts{'destination-path'};
    if ($path =~ /(^.+?):\/\/(.+?)\/(.+$)/) {
        return $2;
    } elsif (($path =~ m{(.+?):(.+)}) || ($path =~ m{(.+?):})) {
        return $1;
    } else {
        $pkj_or_obj->_invalid_path_exception(path => $path);
    }

} ## end sub destination_vserver

=head2 destination_volume

    my $destination_vol = NACL::C::Snapmirror->destination_volume(
                       'destination-path'      => $destination_path,
                    );
                    (or)

    my $destination_vol = $snapmirror_obj->destination_volume();

(Class or instance method) Returns the destination volume name,
if the format of the path is valid  and throws exception of type
NATE::Exceptions::Argument, if it is not valid.

Supports CMode CLI/ZAPI.

=over

=item Options

=over

=item C<< 'destination-path' => $path >>

(Required for Class method)

The path format can be in the form of "clustername://vserver/volume" or
"vserver:volume"

=back

=back

=cut

sub destination_volume() {
    my ($pkj_or_obj, @args) = @_;
    my %opts = $pkj_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {'destination-path' => {type => OBJECT | SCALAR},},
    );

    my $path = $opts{'destination-path'};
    $path = $pkj_or_obj->destination_path() if (!(defined($path)));
    if ($path =~ /(.+):\/\/(.+)\/(.+)/) {
        return $3;
    } elsif ($path =~ /(.+):\/share\/(.+)/) {
        # This is the AVA destination path of format - IP:/share/volname
        return $2;
    } elsif ($path =~ /(.+):(.+)/) {
        return $2;
    } else {
        $Log->exit();
        $pkj_or_obj->_invalid_path_exception(path => $path);
    }
} ## end sub destination_volume

=head2 source_volume

    my $source_vol = NACL::C::Snapmirror->source_volume(
                       'source-path'      => $source_path,
                    );
                    (or)

    my $source_vol = $snapmirror_obj->source_volume();

(Class or instance method) Returns the Source volume name,
if the format of the path is valid and throws exception of type
NATE::Exceptions::Argument, if it is not valid.

Supports CMode CLI/ZAPI.

=over

=item Options

=over

=item C<< 'source-path' => $path >>

(Required for Class method)

The path format can be in the form of "clustername://vserver/volume" or
"vserver:volume"

=back

=back

=cut

sub source_volume() {
    my ($self, $path) = @_;
    $path = $self->source_path() if (!(defined($path)));
    if ($path =~ /(.+):\/\/(.+)\/(.+)/) {
        return $3;
    } elsif ($path =~ /(.+):(.+)/) {
        return $2;
    } else {
        $Log->exit();
        $self->_invalid_path_exception(path => $path);
    }
} ## end sub source_volume

=head2 source_cluster

    my $src_cluster = NACL::C::Snapmirror->source_cluster(
                       'command_interface'=> $ci,
                       'source-path'      => $source_path,
                    );
                    (or)

    my $src_cluster = $snapmirror_obj->source_cluster();

(Class or instance method) Returns the Source cluster name,
if the format of the path is valid and throws exception of type
NATE::Exceptions::Argument if it is not valid.

Supports CMode CLI/ZAPI.

=over

=item Options

=over

=item C<< 'source-path' => $path >>

(Required for Class method)

The path format can be in the form "clustername://vserver/volume" or
"vserver:volume"

=item C<< command_interface => $command_interface >>

(Required for class method and if path is SN style (vserver:volume),
 Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< source_ci => $source_ci >>

(Optional)
Represents command interface of the source cluster or node. This option
would be of interest mainly if it is intercluster relationship and volume
object is not passed while creating the relationship using STask::Snapmirror.
If source_cluster() is called as class method, then source_ci can be passed
instead of command_interface.
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=back

=back

=cut

sub source_cluster() {
    $Log->enter();
    my $pkg_or_obj = shift;
    my %opts       = $pkg_or_obj->_common_validate_with(
        params          => \@_,
        additional_spec => {
            'command_interface' => {
                isa      => 'NACL::C::CommandInterface::ONTAP',
                type     => OBJECT,
                optional => 1,
            },
            'destination-path' => {type => OBJECT | SCALAR, optional => 1},
            'source-path'      => {type => OBJECT | SCALAR, optional => 0},
        },
        ignore_primary_keys => 1,

    );
    my $src_cluster = $pkg_or_obj->_get_cluster_name(
        %opts,
        is_source => 1,
        path      => $opts{'source-path'},

    );
    $Log->exit();

    return $src_cluster;
} ## end sub source_cluster

sub _get_cluster_name {
    $Log->enter();
    my $self = shift;
    my %opts = $self->_common_validate_with(
        params          => \@_,
        additional_spec => {
            'command_interface' => {
                isa      => 'NACL::C::CommandInterface::ONTAP',
                type     => OBJECT,
                optional => 1,
            },
            path => {type => SCALAR},
        },
        ignore_primary_keys => 1,
        allow_extra         => 1,
    );

    my $cluster;
    if ($opts{'path'} =~ /(.+?):\/\/(.+?)\/(.+)/) {
        $cluster = $1;
        $Log->debug("Path: " . $opts{'path'} . ", cluster name: " . $cluster);
    } elsif (($opts{'path'} =~ m{(.+?):(.+)}) || ($opts{'path'} =~ m{(.+?):})) {

        my $cluster_info = $self->_find_cluster(%opts);
        $cluster = $cluster_info->{cluster};
        my $intra = $opts{intra};
        $$intra = $cluster_info->{intra};
    } else {
          $Log->exit();
          $self->_invalid_path_exception(path => $opts{path});
    }

    $Log->exit();
    return $cluster;
} ## end sub _get_cluster_name

# This private method is for finding cluster name when SN style path is passed.
sub _find_cluster() {
    $Log->enter();
    my ($self, %opts) = @_;
    my %cluster_info;

    if (!$opts{command_interface}) {
        my $flag = 1;
        if ($opts{is_source} && $opts{source_ci}) {
            $flag = 0;
        }
        if ($flag) {
            $Log->exit();
            NATE::Exceptions::Argument->throw(
                "Mandatory parameter command_interface/source_ci is missing."
            );
        }
    }
    my %common_opts;
    $self->_copy_common_component_params(
        source => \%opts,
        target => \%common_opts
    );
    if (($opts{path} =~ m{(.+?):(.+)}) || ($opts{'path'} =~ m{(.+?):})) {

        my $vserver = $1;

        # subroutine to obtain the local cluster name
        my $cluster_name = sub {
            my %opts           = @_;
            my $cluster_cs_obj = NACL::CS::ClusterIdentity->fetch(
                command_interface => $opts{command_interface},
                requested_fields  => ['name'],
                %common_opts,
            );
            my $cluster = $cluster_cs_obj->name();
            return $cluster;
        };

        # If is_source flag is set and source_ci is present then local cluster
        # is obtained with source_ci as command_interface.

        if ($opts{is_source} && $opts{source_ci}) {
            $cluster_info{cluster}
                = $cluster_name->('command_interface' => $opts{source_ci});
        } else {

            # Check if the vserver belongs to the node/cluster which is passed
            # as command_interface.
            my $vserver_cs = NACL::C::Vserver->find(
                command_interface => $opts{command_interface},
                filter            => {'vserver' => $vserver,},
                allow_empty       => 1,
                %common_opts
            );
            if (!$vserver_cs) {

             # If vserver does not belong to the local cluster, then it can be
             # intercluster relationship with source in a different cluster.
             # Find the peer-cluster name from vserver peer show.
                $vserver_cs = NACL::CS::VserverPeer->fetch(
                    command_interface => $opts{command_interface},
                    filter            => {'peer-vserver' => $vserver,},
                    allow_empty       => 1,
                    %common_opts
                );
                if ($vserver_cs) {
                    $cluster_info{cluster} = $vserver_cs->peer_cluster();
                    if (!$cluster_info{cluster}) {
                        require NACL::CS::VserverPeerShowAll;
                        $vserver_cs = NACL::CS::VserverPeerShowAll->fetch(
                            command_interface => $opts{command_interface},
                            filter      => {'peer-vserver' => $vserver,},
                            allow_empty => 1,
                            %common_opts
                        );
                        $cluster_info{cluster} = $vserver_cs->peer_cluster();
                        if (!$cluster_info{cluster}) {
                            $Log->warn(
                                "Source cluster information not found.");
                        }
                    }
                    $cluster_info{intra} = 0;
                } else {
                    $Log->exit();
                    NATE::Exceptions::Argument->throw(
                        "Could not find source_cluster information. Check the "
                            . " path value passed or Pass source_ci"
                            . " to access the source cluster. ");
                }
            } else {

                # vserver is present in the local cluster. Hence get Cluster
                # Identity of local cluster.
                $cluster_info{cluster} = $cluster_name->(
                    'command_interface' => $opts{command_interface});
                $cluster_info{intra} = 1;
            }
        }
    } else {
          $Log->exit();
          $self->_invalid_path_exception(path => $opts{path});
    }
    $Log->exit();
    return \%cluster_info;

} ## end sub _find_cluster

=head2 get_source_volume

  my $source_vol_object = $snapmirror_object->get_source_volume();

  (or)

  my $source_vol_object = NACL::STask::Snapmirror->get_source_volume(%opts);       #%opts will have destination-path,source-path and command_interface

  (Class or Instance method) The method returns STask/C object corresponding to the
   source Volume in the snapmirror relationship. If the
   Snapmirror object is of type NACL::STask::Snapmirror, then it returns a
   NACL::STask:Volume object. If the Snapmirror object is of type
   NACL::C::Snapmirror, then it returns a NACL::C::Volume object. If the
   snapmirror source is a Qtree, then it will throw NATE::BaseException.

    If get_source_volume() is called on a cross cluster snapmirror relationship,
    and if volume object is not used for source-path, then either source_ci
    should be passed or tharnhost file with the name of the source cluster is required.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< source_ci => $source_ci >>

(Optional)
Represents command interface of the source cluster or node. This option
would be of interest mainly if it is intercluster relationship and volume
object is not passed while creating the relationship using STask::Snapmirror.
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< 'source-path' => $source_path >>

(Required for class method, Not Applicable for instance method)

=item C<< 'destination-path' => $source_path >>

(Required for class method, Not Applicable for instance method)

=back

=back

=cut

sub get_source_volume {
    $Log->enter();
    my $pkg_or_obj = shift;
    my %opts       = $pkg_or_obj->_common_validate_with(
        params          => \@_,
        additional_spec => {'source-path' => {type => SCALAR}},
        allow_extra     => 1
    );

    # Create a snapmirror object in case, it is a class method call.
    unless (ref $pkg_or_obj) {
        $pkg_or_obj = $pkg_or_obj->new(
            'source_path'      => $opts{'source-path'},
            'destination_path' => $opts{'destination-path'},
            command_interface  => $opts{command_interface},
        );
    }
    my $source_ci = $pkg_or_obj->source_ci;
    my %tmp_opts  = (
        path              => $opts{'source-path'},
        type              => 'volume',
        is_source         => 1,
        command_interface => $opts{command_interface},
    );
    $tmp_opts{source_ci} = ($opts{source_ci} || $source_ci)
        if ($opts{source_ci} || $source_ci);

    my $obj = $pkg_or_obj->_get_source_or_dest_object(%tmp_opts);
    $Log->exit();
    return $obj;
} ## end sub get_source_volume

=head2 get_destination_volume

  my $dest_vol_object = $snapmirror_object->get_destination_volume();

  (or)

  my $dst_vol_object = NACL::STask::Snapmirror->get_destination_volume(%opts);       #%opts will have destination-path,source-path and command_interface

   (Class or Instance method) The method returns STask/C object corresponding to the
   destination Volume in the snapmirror relationship. If the
   Snapmirror object is of type NACL::STask::Snapmirror, then it returns a
   NACL::STask:Volume object. If the Snapmirror object is of type
   NACL::C::Snapmirror, then it returns a NACL::C::Volume object. If the
   snapmirror destination is a Qtree, then it will throw NATE::BaseException.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< 'destination-path' => $source_path >>

(Required for class method, Not Applicable for instance method)

=back

=back

=cut

sub get_destination_volume {
    $Log->enter();
    my $pkg_or_obj = shift;
    my %opts       = $pkg_or_obj->_common_validate_with(
        params      => \@_,
        allow_extra => 1
    );

    # Create a snapmirror object in case, it is a class method call.
    unless (ref $pkg_or_obj) {
        $pkg_or_obj = $pkg_or_obj->new(
            'destination_path' => $opts{'destination-path'},
            command_interface  => $opts{command_interface},
        );
    }
    my $obj = $pkg_or_obj->_get_source_or_dest_object(
        path              => $opts{'destination-path'},
        type              => 'volume',
        is_source         => 0,
        command_interface => $opts{command_interface}
    );
    $Log->exit();
    return $obj;
} ## end sub get_destination_volume

=head2 get_source_qtree

  my $source_qtree_object = $snapmirror_object->get_source_qtree();

   (Instance method) The method returns STask/C object corresponding to the
   source Qtree in the snapmirror relationship. If the
   Snapmirror object is of type NACL::STask::Snapmirror, then it returns a
   NACL::STask:Qtree object. If the Snapmirror object is of type
   NACL::C::Snapmirror, then it returns a NACL::C::Qtree object. If the
   snapmirror source is a Volume, then it will throw NATE::BaseException.

=over

=item Options

Nothing

=back

=cut

sub get_source_qtree {
    $Log->enter();
    my $pkg_or_obj = shift;
    my %opts       = $pkg_or_obj->_common_validate_with(
        params          => \@_,
        additional_spec => {'source-path' => {type => SCALAR}},
        allow_extra     => 1
    );

    my $obj = $pkg_or_obj->_get_source_or_dest_object(
        path              => $opts{'source-path'},
        type              => 'qtree',
        is_source         => 1,
        command_interface => $opts{command_interface}
    );
    $Log->exit();
    return $obj;
} ## end sub get_source_qtree

=head2 get_destination_qtree

  my $dest_qtree_object = $snapmirror_object->get_destination_qtree();

   (Instance method) The method returns STask/C object corresponding to the
   destination Qtree in the snapmirror relationship. If the
   Snapmirror object is of type NACL::STask::Snapmirror, then it returns a
   NACL::STask:Qtree object. If the Snapmirror object is of type
   NACL::C::Snapmirror, then it returns a NACL::C::Qtree object. If the
   snapmirror destination is a Volume, then it will throw NATE::BaseException.

=over

=item Options

Nothing

=back

=cut

sub get_destination_qtree {
    $Log->enter();
    my $pkg_or_obj = shift;
    my %opts       = $pkg_or_obj->_common_validate_with(
        params      => \@_,
        allow_extra => 1
    );

    my $obj = $pkg_or_obj->_get_source_or_dest_object(
        path              => $opts{'destination-path'},
        type              => 'qtree',
        is_source         => 0,
        command_interface => $opts{command_interface}
    );
    $Log->exit();
    return $obj;
} ## end sub get_destination_qtree

=head2 is_operated_on_volume

    if ($snapmirror_object->is_operated_on_volume()) {
        $source_vol = $snapmirror_object->get_source_volume();
        ...
        # Perform some Volume related operations
    } else {
        $source_qtree = $snapmirror_object->get_source_qtree();
        ...
        # Perform some Qtree related operations
    }

              or

    if (NACL::C::Snapmirror->is_operated_on_volume('destination-path' => $path) {
        ...
    }

(Class or Instance method)

This method can be used to find out whether the source and destination
in the snapmirror relationship are volumes or qtrees. Thus it helps to
invoke appropriate get_source_(volume|qtree) or
get_destination_(volume|qtree) methods as shown in the above example. It
returns a boolean value: a return value of 1 indicates that both source
and destination are volumes and a return value of 0 indicates that both
source and destination are qtrees.

=over

=item Options

=over

=item As Instance method

Nothing

=item As Class method

The destination-path needs to be provided as argument.

=back

=back

=cut

# Since we cannot create a hybrid snapmirror relationship between
# a volume and qtree or viceversa, it's enough to inspect either
# source or destination to find out whether the snapmirror is
# between volumes or qtrees.
sub is_operated_on_volume {
    $Log->enter();
    my ($pkg_or_obj, %opts) = @_;
    my $observed_type = $pkg_or_obj->_get_observed_type(%opts);
    my $is_vol = ($observed_type =~ /^volume$/i) ? 1 : 0;
    $Log->exit();
    return $is_vol;
} ## end sub is_operated_on_volume

sub is_operated_on_vserver {
    $Log->enter();
    my ($pkg_or_obj, %opts) = @_;
    my $observed_type = $pkg_or_obj->_get_observed_type(%opts);
    my $is_vs = ($observed_type =~ /^vserver$/i) ? 1 : 0;
    $Log->exit();
    return $is_vs;
}

sub _get_observed_type {
    $Log->enter();
    my ($pkg_or_obj, %opts) = @_;
    my $info = $pkg_or_obj->_get_source_or_dest_object_info(
        path => $opts{'destination-path'} || $pkg_or_obj->destination_path(),
        is_source => 0
    );

    my $observed_type = $info->{observed_type};
    $Log->exit();
    return $observed_type;
}

=head2 check_if_intercluster

    my $cluster_type = NACL::STask::Snapmirror->check_if_intercluster(
                       command_interface  => $command_interface,
                       'source-path'      => $source_path,
                       'destination-path' =>  $destination_path,
                    );
                    (or)

    my $cluster_type = NACL::STask::Snapmirror->check_if_intercluster(
                       command_interface  => $command_interface,
                       'source-path'      => $source_volume_object,
                       'destination-path' =>  $destination_volume_object,
                    );

            (or)

    my $cluster_type = $snapmirror_obj->check_if_intercluster(source_ci => $ci);

(Class or instance method) Check for the type of snapmirror relationship,
return '1' if it is intercluster, else return '0'.

Supports CMode CLI.

=over

=item Options

=over

=item C<< "command_interface=>$command_interface" >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< 'source-path' => $path_specification >>

(Required for Class method)

See documentation at the beginning of this page for the various ways
this argument can be provided.

=item C<< "destination-path =>  $volume_specification" >>

(Required for Class method)

See documentation at the beginning of this page for the various ways
this argument can be provided.

=item C<< "source-cluster =>  $string" >>

(Optional) Name of cluster where the source volume resides.

=item C<< "destination-cluster =>  $string" >>

(Optional) Name of cluster where the destination volume resides.

=item C<< source_ci => $source_ci >>

(Optional)
Represents command interface of the source cluster or node.
This option is required if you are having intercluster snapmirror relationship
and if volume object is not passed for source-path.
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>.

=item apiset_must, apiset_should, method-timeout.

=back

=back

=cut

sub check_if_intercluster {
    $Log->enter();
    my $self = shift;
    my %opts = $self->_common_validate_with(
        params => \@_,
        additional_spec =>
            {_optional_scalars qw(source-cluster destination-cluster)},
    );
    my ($src_cluster, $dst_cluster, %common_opts);
    $self->_copy_common_component_params_with_ci(
        source => \%opts,
        target => \%common_opts
    );

    my $inter = $self->_check_and_get_intercluster(%opts);

    $Log->exit();
    return $inter->{intercluster};

} ## end sub check_if_intercluster

sub _check_and_get_intercluster {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    my %opts = $pkg_or_obj->_common_validate_with(
        params => \@args,
        additional_spec =>
            {_optional_scalars qw(source-cluster destination-cluster)},
    );
    my ($source_clus, $dest_clus, %cluster, %common_opts);
    $pkg_or_obj->_copy_common_component_params_with_ci(
        source => \%opts,
        target => \%common_opts
    );

    my $src_path = $pkg_or_obj->_split_path(path => $opts{'source-path'});
    my $dst_path
        = $pkg_or_obj->_split_path(path => $opts{'destination-path'});

# If source and destination path is of SN style (vserver:volume) and
# command-interface is vserver type, then cluster information cannot be
# obtained. vserver names are compared to find if it is inter or intracluster.
# Even if the source and dest vservers belong to the same cluster, but are
# different, it is considered to be intercluster.
    if (   $src_path->{SN}
        && $dst_path->{SN}
        && $opts{command_interface}->isa('NACL::C::Vserver'))
    {
        if ($src_path->{vserver} eq $dst_path->{vserver}) {
            $cluster{intercluster} = 0;
        } else {
            $Log->exit();
            NATE::BaseException->throw("Cannot determine if it is intra or "
                    . "intercluster since vserver command interface is passed "
            );
        }
    } else {
        $dest_clus = $dst_path->{cluster} || $opts{'destination-cluster'};
        if (!$dest_clus) {

         # Obtain destination cluster name if command_interface is not of type
         # NACL::C::Vserver or when the path is not in SN style.
            $dest_clus
                = $pkg_or_obj->destination_cluster(%common_opts,
                'destination-path' => $opts{'destination-path'},)
                if (
                !(     $dst_path->{SN}
                    && $opts{command_interface}->isa('NACL::C::Vserver')
                )
                );
        }
        $cluster{destination} = $dest_clus;
        $source_clus = $src_path->{cluster} || $opts{'source-cluster'};
        if (!$source_clus) {

            # Obtain source cluster name if command_interface is not of type
            # NACL::C::Vserver or when the path is not in SN style.

             # when source_ci provided pass it to source_cluster method
            if (defined $opts{source_ci} &&
                !$opts{source_ci}->isa('NACL::C::Vserver') ) {
                $common_opts{source_ci} = $opts{source_ci};
            }
            $source_clus
                = $pkg_or_obj->source_cluster(%common_opts,
                'source-path' => $opts{'source-path'})
                if (
                !(     $src_path->{SN}
                    && $opts{command_interface}->isa('NACL::C::Vserver')
                 )
                );
            delete $common_opts{source_ci};
        }
        $cluster{source} = $source_clus;
        if (!($source_clus && $dest_clus)) {

       # If either one of the path is in RR and the other is in SN style,
       # then one of source or destination cluster name will not be available.
       # In that case, determin inter or intracluster based on vserver names.
            if ($src_path->{vserver} eq $dst_path->{vserver}) {
                $cluster{intercluster} = 0;
            } else {
                $cluster{intercluster} = 1;
            }
        } else {
            $cluster{intercluster} = ($source_clus ne $dest_clus);
        }
    }
    return \%cluster;
} ## end sub _check_and_get_intercluster

=head2 get_consistency_group_data

    my $cg_data = $sm_obj->get_consistency_group_data;
    for my $key (sort keys %{$cg_data}) {
        # Do something with the item...
    }

    my $cg_data = NACL::C::Snapmirror->get_consistency_group_data(
        command_interface => $node_obj,
        'destination-path' => $dest_cg_path
    );

(Class or instance method) Retrieve the information on the consistency group
for the given snapmirror object, or for the destination-path if called as a
class method.

=over

=item Parameters

All parameters are optional if the call is made on an existing object. The
parameters taken by this method are:

=over

=item C<< command_interface =E<gt> $node_obj >>

A command-interface through which the data is queried. This should be an
ONTAP cluster, node or vserver interface.

=item C<< 'destination-path' =E<gt> $string >>

The destination path of the relationship to query for consistency group data.
This must be a CG path or an exception will be thrown. You may also pass the
combination of C<destination-vserver> and C<destination-cg>, which will be
built into the destination-path for you.

=back

=item Exceptions

The following exceptions may be thrown by this method:

=over

=item C<NACL::C::Exceptions::Snapmirror::NotGranular>

This exception will be thrown if the object that calls this method does not
represent a granular relationship, or in the case of a class call if the
destination-path is not a granular path.

=back

=back

The return value of this call will be a hash reference with zero or more keys.
Each key is the full source-path of the consistency group item, for example,
C<vs0:/vol/src_vol/cg_item_path>. Each key points to another hash reference
with the following keys:

=over

=item B<state>

The B<NACL::CS::Snapmirror> state object that represents the (current) state
of this item in the consistency group. At present, the majority of the fields
in this object are empty (set to the ONTAP marker, C<->).

=item B<destination_path>

The destination path of this item in the consistency group. This is the full
path, including the vserver, the volume, the file name and the C</vol/> leader
that appears before the volume name. This is the same value as you would get
from calling the C<destination_path> accessor on the value of the B<state>
key, above. Note that this key uses an underscore (C<_>) rather than a hyphen
(C<->), for easier use as a hash key.

=item B<source_path>

The source path of this item in the consistency group. This is the full path,
including the vserver, the volume, the file name and the C</vol/> leader that
appears before the volume name. This is the same value as you would get from
calling the C<source_path> accessor on the value of the B<state> key, above.
This is also identical to the value used as the key in the larger hash of
consistency group data. It is duplicated here for consistency. Note that this
key uses an underscore (C<_>) rather than a hyphen (C<->), for easier use as a
hash key.

=back

Note that the contents of this hash reference may expand as needed. However,
anything listed here should always be carried forward for backwards
compatibility.

=cut

sub get_consistency_group_data {
    $Log->enter;
    my $pkg_or_obj = shift;
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@_,
        additional_spec => {
            _optional_scalars(
                qw(source-cluster source-vserver source-volume source-cg
                   destination-cluster destination-vserver destination-cg
                   destination-volume)
            ),
        }
    );

    if ($opts{'destination-path'} !~ m{/cg/}) {
        NACL::C::Exceptions::Snapmirror::NotGranular->throw(
            text =>
            'Can only get Consistency Group data for a granular relationship'
        );
    }

    my $CS_pkg = $pkg_or_obj->get_CS_package_name;

    # Note that, as of 2/3/15, the fetch() below must only specify the dest.
    # path, or else you will only get the basic relationship record, not the
    # CG items as well. Also, the value to the 'expand' key *must* be "true".
    $Log->warn('Using CLI to get snapmirror cg mapping info - workaround for burt 972305');
    my @elements = $CS_pkg->fetch(
        command_interface => $opts{command_interface},
        filter            => {
            'destination-path' => $opts{'destination-path'},
            expand             => 'true',
        },
        # workaround for burt 972305
        apiset_should => { interface => 'CLI' },
    );

    # Drop the first element, since it just represents the basic relationship:
    shift @elements;

    # Create the return value from the remaining elements:
    my $retval = {};
    for my $elem (@elements) {
        my $key = $elem->source_path;
        $retval->{$key} = {
            state            => $elem,
            source_path      => $key,
            destination_path => $elem->destination_path,
        };
    }

    $Log->exit;
    return $retval;
}

=head2 is_granular

    # Check on an object:
    if ($sm_obj->is_granular) {
        # Do something...
    }

    # Check on data, without an object:
    if (NACL::C::Snapmirror->is_granular(
            'source-path'      => $source_path,
            'destination-path' => $dest_path
        )) {
        # Do something...
    }

This may be called as either an instance method or a static method.

Returns a boolean value indicating whether the snapmirror relationship is
granular or not. If this is called on an object, no parameters are needed (and
any are ignored). If this is called as a static method, then the parameters
C<source-path> and C<destination-path> are checked (and all others are
ignored).

The return value is C<1> (true) if the relationship is granular, C<0> (false)
otherwise.

=over

=item Parameters

This method will recognize the following two parameters only when called as a
static method. Any other parameters are quietly ignored. If called as an object
method, all parameters are quietly ignored.

=over

=item C<< source-path =E<gt> $source_path >>

The source path of the snapmirror relationship.

=item C<< destination-path =E<gt> $destination_path >>

The destination path of the snapmirror relationship.

=back

Note that both of these must be present, or the result will always be false
(C<0>). Also note that this method will not assemble the paths from parts
(B<source-vserver>, B<source-cg>, etc.) like B<create> and others do.

=item Exceptions

This method throws no exceptions.

=back

=cut

sub is_granular {
    $Log->enter;
    my ($pkg_or_obj, %args) = @_;
    my $result;

    if (ref $pkg_or_obj) {
        $result = $pkg_or_obj->_is_granular;
    } else {
        $result = $args{'destination-path'} =~ m{:/cg/};
        if ($args{'source-path'}) {
            $result &&= $args{'source-path'} =~ m{:/cg/};
        }
    }

    $Log->exit;
    return $result;
}

=head2 volume_paths_from_cg

    # Get both volume paths:
    ($src_path, $dst_path) = $sm->volume_paths_from_cg;

    # Get just the destination path:
    $dst_path = $sm->volume_paths_from_cg;

    # Get just the source path:
    ($src_path) = $sm->volume_paths_from_cg;

    # Call as a static method:
    $dst_path = NACL::C::Snapmirror->volume_paths_from_cg(
        command_interface     => $dst_node,
        'source-vserver'      => $src_vserver,
        'source-cg'           => $src_cg_name,
        'destination-vserver' => $dst_vserver
        'destination-cg'      => $dst_cg_name
    );

May be called as an instance method or a static method.

Derive the volume paths (strings of the form, "vserver:volume_name") from the
consistency group data for a granular snapmirror relationship. This is needed
because snapmirror objects that represent granular relationships do not have
information on the specific volumes contained within them.

If called in an array context, the return value is the pair
C<($source_path, $destination_path)>. If called in a scalar context, only
the destination path is returned. If you need only the source path, then
call this in a list context while ignoring the second value:

    ($src_path) = $sm->volume_paths_from_cg;

If there are no elements in the consistency group, then the value(s) returned
will be C<undef>.

=over

=item Parameters

The following parameters are recognized by this method. Any others are
ignored. All parameters are optional when the method is called on an instance,
but some are required when the method is called statically.

=over

=item C<< consistency_group =E<gt> $hashref >>

If passed, all other parameters are ignored. The value of the parameter must
be a hash reference that was returned by a called to
B<get_consistency_group_data>. When passed, the contents of this structure
will be used to determine the paths (rather than actively fetching the
consistency group data for the relationship).

=item C<< command_interface =E<gt> $object >>

Required for a static invocation. Specifies the command interface (node, etc.)
that is to be used in fetching the consistency group data.

=item C<< 'source-path' => $string >>

=item C<< 'source-vserver' => $string >>

=item C<< 'source-cg' => $string >>

=item C<< 'destination-path' => $string >>

=item C<< 'destination-vserver' => $string >>

=item C<< 'destination-cg' => $string >>

These parameters are identical to their use elsewhere, and are used to specify
the source and destination paths for the granular relationship (when the
method is called statically). If C<source-path> is given, C<source-vserver>
and C<source-cg> are ignored. The same is true for C<destination-path>.

=back

=item Exceptions

The following exception may be thrown by a call to this method:

=over

=item C<NACL::C::Exceptions::Snapmirror::NotGranular>

If the calling object does not represent a granular relationship, this
exception will be thrown. It will also be thrown if there is inconsistent or
invalid data in the consistency group data for the relationship.

=back

=back

=cut

sub volume_paths_from_cg {
    $Log->enter;
    my ($pkg_or_obj, %args) = @_;
    my ($cgroup, $src_path, $dst_path);

    if ($args{consistency_group}) {
        $cgroup = $args{consistency_group};
    } else {
        $cgroup = $pkg_or_obj->get_consistency_group_data(%args);
    }

    # NOTE: Right now CGs are limited to just being on one volume on each
    #       end. This may change in the future, which will basically render
    #       this method irrelevant. For now, just look at the first element
    #       of the CG since everything else will be on the same volumes.
    my $first_key = (keys %{$cgroup})[0];
    $src_path = $cgroup->{$first_key}->{source_path};
    $dst_path = $cgroup->{$first_key}->{destination_path};
    for my $path ($src_path, $dst_path) {
        if ($path =~ m{(.*):/vol/(.*?)/}) {
            $path = "$1:$2";
        } else {
            NACL::C::Exceptions::Snapmirror::NotGranular->throw(
                text =>
                "Consistency group data not valid (on key '$first_key')"
            );
        }
    }

    $Log->exit;
    return wantarray ? ($src_path, $dst_path) : $dst_path;
}

# This method finds out whether the source or destination is a
# volume or qtree and also obtains information about primary keys.

sub _get_source_or_dest_object_info {
    my ($pkg_or_obj, %args) = @_;
    $Log->enter();
    my $path       = $args{path};
    my $src_or_dst = ($args{is_source}) ? 'source' : 'destination';
    my $ci         = $args{command_interface};
    my %pri_keys   = ();
    if (!$ci && ref $pkg_or_obj) {
        $ci = $pkg_or_obj->command_interface();
    }

    # We assume that the source is a volume by default. Let's update
    # this variable later if we find it to be a qtree/vserver.
    my $observed_type = 'volume';

    if ($path =~ /(^.+?):\/\/(.+?)\/(.+$)/) {

        # CMode path
        $pri_keys{volume}  = $3;
        $pri_keys{vserver} = $2;
    } elsif (($ci && $ci->is_cmode())
        && $path =~ m{(.+):(.+)})
    {

        # CMode path
        $pri_keys{volume}  = $2;
        $pri_keys{vserver} = $1;
    } elsif (($path !~ /\/vol\//)
        && ($path =~ m{(.+):(.+)}))
    {

        # 7Mode - Snapmirror source or dest is a volume (node:volume)
        $pri_keys{node}   = $1;
        $pri_keys{volume} = $2;
    } elsif (($path !~ /\/vol\//)
        && ($path =~ /^(?:.+?:)?(\w+(?:\(\d+\))?)$/))
    {

        # 7Mode - Snapmirror source or dest is a volume
        $pri_keys{volume} = $1;
    } elsif (($path =~ /\/vol\//)
        && ($path =~ /^(?:.+?:)?\/vol\/(\w+(?:\(\d+\))?)\/([^\/]+)$/))
    {

        # 7Mode - Snapmirror source or dest is a qtree
        $pri_keys{volume} = $1;
        $pri_keys{qtree}  = $2;
        $observed_type    = 'qtree';
    } elsif ($path =~ /(.*):$/) {

        # Vserver-relationships will be like "vs1:", i.e. end with a colon
        $pri_keys{vserver} = $1;
        $observed_type = 'vserver';
    } else {
        $Log->exit();
        NATE::Exceptions::Argument->throw("Invalid ${src_or_dst}-path '$path'"
                . " which doesn't match any of the expected formats. Unable to "
                . "discover whether ${src_or_dst} represents a volume or qtree"
        );
    } ## end else [ if ( $path =~ /(^.+?):\/\/(.+?)\/(.+$)/)
    $Log->exit();
    my $output
        = {observed_type => $observed_type, primary_keys => \%pri_keys};
    $Log->debug(Dumper($output));
    return $output;
} ## end sub _get_source_or_dest_object_info

# Private routine to get the source or destination object in the
# snapmirror relationship. This handles CMode path, 7Mode SSM, VSM and QSM.
# This returns a STask/C object based on whom the caller is.

sub _get_source_or_dest_object {
    my ($pkg_or_obj, %args) = @_;
    $Log->enter();
    my $path       = delete $args{path};
    my $type       = delete $args{type};
    my $ci         = $args{command_interface};
    my $src_or_dst = (delete $args{is_source}) ? 'source' : 'destination';

    my $pk_name = $pkg_or_obj->get_package_name();

    # Get the package name of the source or destination object
    # to be created. If the caller is STask i.e., NACL::STask::Snapmirror,
    # then the package name will be constructed as NACL::STask::Volume.
    # If the caller is from component layer, ie., NACL::C::Snapmirror,
    # then the package name will be constructed as NACL::C::Volume.
    # We assume source and destination are volume by default. Later
    # from the path input, if we find them to be qtrees, then we will
    # update the path input to be NACL::(STask|C)::Qtree
    $pk_name =~ s/^(NACL::[^:]+::).*$/$1Volume/;

    my $info = $pkg_or_obj->_get_source_or_dest_object_info(
        path              => $path,
        is_source         => $src_or_dst,
        command_interface => $ci,
    );
    my $observed_type = $info->{observed_type};
    my %pri_keys      = %{$info->{primary_keys}};

    # The user had invoked an invalid get_(source|destination)_* method
    # for the given snapmirror relationship. For Eg: Invoking get_source_qtree
    # on a snapmirror object in which the source is a volume.
    if ($type !~ /^$observed_type$/i) {
        $Log->exit();
        NATE::BaseException->throw("You have requested an invalid '$type' "
                . "$src_or_dst object when the $src_or_dst is a $observed_type"
        );
    } ## end if ( $type !~ /^$observed_type$/i)

    # If the src/dest is a qtree, then we have to update the package name.
    $pk_name =~ s/Volume$/Qtree/ if ($type =~ /^qtree$/i);

    # Create the object
    eval "require $pk_name ;";
    if ($@) {
        $Log->exit();
        NATE::BaseException->throw(
            "Error in loading '$pk_name' and the error" . " :\n$@");
    }
    my $obj;
    if (   $type =~ /volume/i
        && $src_or_dst =~ /source/i)
    {

        my %common_opts;
        $pkg_or_obj->_copy_common_component_params_with_ci(
            source => \%args,
            target => \%common_opts
        );
        if (ref $pkg_or_obj && $pkg_or_obj->source_ci) {
            $args{source_ci} ||= $pkg_or_obj->source_ci();
        }



        unless (defined $args{source_ci}) {
            if ($ci->is_cmode()) {

                my $cluster = $pkg_or_obj->_check_and_get_intercluster(%args);

                if ($cluster->{intercluster}) {

                    $args{source_ci}
                        = NACL::C::Cluster->new(name => $cluster->{source});

                    unless ($args{source_ci}->can_use_as_command_interface) {
                        $Log->exit();
                        NATE::BaseException->throw(
                                  "A tharnhost file in the name of"
                                . " cluster "
                                . $cluster->{source}
                                . " doesn't exist."
                                . " Pass source_ci parameter to access source volume."
                        );
                    }
                } else {
                    $args{source_ci} = $args{command_interface};
                }
                $obj = $pk_name->new(%pri_keys,
                    command_interface => $args{source_ci},);
                $Log->exit();
                return $obj;
            } else {
                if ($pri_keys{node}) {

                    # Node name in 7Mode path is present as node in %pri_keys
                    if (($pkg_or_obj->destination_path() !~ $pri_keys{node}))
                    {
                        require NACL::C::Node;

                        $args{source_ci}
                            = NACL::C::Node->new(name => $pri_keys{node});

                        unless (
                            $args{source_ci}->can_use_as_command_interface)
                        {
                            $Log->exit();
                            NATE::BaseException->throw(
                                      "A tharnhost file in the name of"
                                    . " node "
                                    . $pri_keys{node}
                                    . " doesn't exist."
                                    . " Pass source_ci parameter to access source volume."
                            );
                        }
                    } else {
                        $args{source_ci} = $args{command_interface};
                    }
                    delete $pri_keys{node};
                    $obj = $pk_name->new(%pri_keys,
                        command_interface => $args{source_ci},);
                    $Log->exit();
                    return $obj;
                }
            }

        } else {
            $args{command_interface} = $args{source_ci};
        }
    }
    delete $pri_keys{node};
    $obj = $pk_name->new(%pri_keys,
        command_interface => $args{command_interface});
    $Log->exit();
    return $obj;
} ## end sub _get_source_or_dest_object

# Perform the various mappings/translations needed to make the values in
# the extended_query look like the CLI equivalent
sub _extended_query_cmode_zapi {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    state $spec = {all_opts => {type => HASHREF}};
    my %opts = validate(@args, $spec);
    my $all_opts = $opts{all_opts};
    if ($all_opts->{extended_query}) {
        my %query = %{$all_opts->{extended_query}};
        $pkg_or_obj->_update_iter_input(opts_ptr => \%query);
        $all_opts->{extended_query} = \%query;
    }
    $Log->exit();
}

sub _split_path {
    $Log->enter();
    my $pkg_or_obj = shift;
    my %opts = validate(@_, {'path' => {type => SCALAR, optional => 0},});
    my %path;
    if ($opts{path} =~ m{(.+)://(.+)/(.+)}) {
        $path{RR}      = 1;
        $path{cluster} = $1;
        $path{vserver} = $2;
        $path{volume}  = $3;
    } elsif ($opts{path} =~ m{(.+?):/cg/(.+)}) {
        $path{SN}      = 1;
        $path{vserver} = $1;
        $path{cg}      = $2;
    } elsif ($opts{path} =~ m{(.+?):(.+)}) {
        $path{SN}      = 1;
        $path{vserver} = $1;
        $path{volume}  = $2;
    } elsif ( $opts{path} =~ m{(.+?):}) {
        $path{SN}      = 1;
        $path{vserver} = $1;
    } else {
        $Log->exit();
        $pkg_or_obj->_invalid_path_exception(path => $opts{path});
    }
    $Log->exit();
    return \%path;
} ## end sub _split_path

# Override _process_options to avoid stringification of source_ci parameter.
sub _process_options {
    $Log->enter();
    my ($pkg_or_obj, %opts) = @_;
    my $options   = $opts{options};
    my $source_ci = delete $options->{source_ci};
    $pkg_or_obj->SUPER::_process_options(options => $options,);
    if (defined $source_ci) {
        $options->{source_ci} = $source_ci;
    }
    $Log->exit();
} ## end sub _process_options

sub _invalid_path_exception {
    $Log->enter();
    my ($pkg_or_obj, %opts) = @_;
    my $path = $opts{path};
    $Log->exit();
    NATE::Exceptions::Argument->throw(
        "Invalid path format passed. Expected format: vserver_name:volume_name"
            . " or cluster_name://vserver_name/volume_name. But passed $path"
    );
} ## end sub _invalid_path_exception

# Map the value of 'throttle' from what's understood by CLI to what's
# understood by ZAPI; provide the mapping of the primary keys.
sub _base_backend_cmode_zapi {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    state $spec = {
        map             => {type => HASHREF,  default => {}},
        additional_map  => {type => HASHREF,  default => {}},
        copy            => {type => ARRAYREF, default => []},
        additional_copy => {type => ARRAYREF, default => []},
    };
    my %opts = validate_with(
        params      => \@args,
        spec        => $spec,
        allow_extra => 1,
    );
    my $additional_map  = delete $opts{additional_map};
    my $additional_copy = delete $opts{additional_copy};
    my %map             = (
        'source-path'      => 'source-location',
        'destination-path' => 'destination-location',
        %$additional_map,
        %{$opts{map}},
    );
    my @copy = (@$additional_copy, @{$opts{copy}});

    $pkg_or_obj->_update_throttle_for_zapi(opts_ptr => \%opts);

    $pkg_or_obj->SUPER::_base_backend_cmode_zapi(
        %opts,
        map  => \%map,
        copy => \@copy
    );
    $Log->exit();
}

# A CMode-ZAPI base backend method builder for those methods which have
# an async equivalent.
sub _backend_cmode_zapi_infinivol_builder {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    state $spec = {
        api              => {type => SCALAR},
        additional_map   => {type => HASHREF, default => {}},
        additional_copy  => {type => ARRAYREF, default => []},
        is_infinivol_ref => {type => SCALARREF, optional => 1},
    };
    my %opts = validate_with(
        params      => \@args,
        spec        => $spec,
        allow_extra => 1
    );
    my $additional_map  = $opts{additional_map};
    my $additional_copy = $opts{additional_copy};

   # Determine whether to use the async form or not (async necessary for
   # infinivols). Check if the relationship is for a volume (it is possible to
   # have relationships between qtrees or between vservers). If the
   # relationship is between volumes, then check if the destination volume
   # is an infinivol.
    my $dst_path  = $opts{'destination-path'};
    my $use_async = 0;

    if ($pkg_or_obj->is_operated_on_volume('destination-path' => $dst_path)) {
        my $path_hash = $pkg_or_obj->_split_path(path => $dst_path);
        $use_async = $pkg_or_obj->_determine_is_infinivol(
            %opts,
            volume        => $path_hash->{volume},
            vserver       => $path_hash->{vserver},
            apiset_should => {interface => 'ZAPI'},
        );
    }

    if (defined $opts{is_infinivol_ref}) {
        ${$opts{is_infinivol_ref}} = $use_async;
    }
    $pkg_or_obj->_is_infinivol($use_async) if (ref $pkg_or_obj);

    my $common_map = $pkg_or_obj->_common_zapi_map();

    $opts{api} .= '-async' if ($use_async);
    $pkg_or_obj->_base_backend_cmode_zapi(%opts);

    $Log->exit();
}

# Wrapper around the base class _iter_helper. Additionally does the following:
# * Uses the common zapi copy/map
# * Sets the does_not_exists error and exception
# * Takes care of mapping the value of the "type" field from CLI -> ZAPI
#
# Accepts an "additional_map" option. The -iter ZAPIs for snapmirror
# accept options in addition to query/attributes. The additional_map
# specifies the CLI->ZAPI field name mappings for these fields. These
# fields get passed to the ZAPI using the "pass_through" option
sub _iter_helper {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    state $spec = {
        extended_query => {type => HASHREF, default => {}},

        # additional_map and additional_copy are those options which are
        # direct options to the ZAPI, i.e. they are not a part of query.
        additional_map  => {type => HASHREF,  default => {}},
        additional_copy => {type => ARRAYREF, default => []},
    };
    my %opts = validate_with(
        params      => \@args,
        spec        => $spec,
        allow_extra => 1
    );

    my $additional_map  = delete $opts{additional_map};
    my $additional_copy = delete $opts{additional_copy};

    # Keep these calls before the _hash_move because the _hash_move moves
    # options out of %opts
    $pkg_or_obj->_extended_query_cmode_zapi(all_opts => \%opts);
    $pkg_or_obj->_update_throttle_for_zapi(opts_ptr => \%opts);

    my %pass_through;
    $pkg_or_obj->_hash_move(
        source => \%opts,
        target => \%pass_through,
        map    => $additional_map,
        move   => $additional_copy
    );

    my $common_copy = $pkg_or_obj->_common_zapi_copy();
    my $common_map  = $pkg_or_obj->_common_zapi_map();

    $pkg_or_obj->SUPER::_iter_helper(
        pass_through         => \%pass_through,
        copy                 => $common_copy,
        map                  => $common_map,
        does_not_exist_error => "entry doesn't exist",
        does_not_exist_exception =>
            'NACL::C::Exceptions::Snapmirror::DoesNotExist',

        # Keep this last so that options passed from the back-end method
        # take higher precedence than the defaults listed above
        %opts,
    );

    $Log->exit();
}

sub _build_iter_backend {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    state $spec = {common_opts_method => {type => SCALAR}};
    my %opts = validate_with(
        params      => \@args,
        spec        => $spec,
        allow_extra => 1,
    );
    my $common_opts_method = delete $opts{common_opts_method};
    my %common_opts        = $pkg_or_obj->$common_opts_method();
    $common_opts{api} .= '-iter';
    $pkg_or_obj->_iter_helper(%common_opts, %opts);
    $Log->exit();
}

sub _build_non_iter_backend_with_async {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    state $spec = {common_opts_method => {type => SCALAR}};
    my %opts = validate_with(
        params      => \@args,
        spec        => $spec,
        allow_extra => 1,
    );
    my $common_opts_method = delete $opts{common_opts_method};
    my %common_opts        = $pkg_or_obj->$common_opts_method();
    $pkg_or_obj->_backend_cmode_zapi_infinivol_builder(%common_opts, %opts);
    $Log->exit();
}

sub _build_non_iter_backend {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    state $spec = {common_opts_method => {type => SCALAR}};
    my %opts = validate_with(
        params      => \@args,
        spec        => $spec,
        allow_extra => 1,
    );
    my $common_opts_method = delete $opts{common_opts_method};
    my %common_opts        = $pkg_or_obj->$common_opts_method();
    $pkg_or_obj->_base_backend_cmode_zapi(%common_opts, %opts);
    $Log->exit();
}

sub _build_iter_backend_with_retry {
    $Log->enter();
    my ($pkg_or_obj, %opts) = @_;
    my $type = delete $opts{type};
    my $policy = delete $opts{policy};
    my $method = delete $opts{_operation};
    my $flag = 0;
    my $exception;
    my $build_iter_backend = sub {$pkg_or_obj->_build_iter_backend(%opts)};
    try {
        $build_iter_backend->();
    } catch NACL::C::Exceptions::Snapmirror::DoesNotExist with {
        my $exception = $_[0];
        if (defined $opts{'destination-path'}) {
            if ($pkg_or_obj->is_operated_on_vserver(%opts)) {
                $Log->exit();
                NACL::C::Exceptions::Snapmirror::NoDrRelationship
                       ->convert_and_throw(exception => $exception);
            } else {
                my $msg = "Unlike CLI, ZAPI does not create a snapmirror "
                      . "relationship implicitly when initialize is invoked.\n"
                      . "We will create a snapmirror relationship "
                      . "before retrying with resync.";
                $Log->comment($msg);
                # Retrying by using create() is possible only for Snapmirror
                # relationships between volumes, not between vservers
                try {
                    my %tmp_opts;
                    my $copy_params = [qw( source-path destination-path command_interface)];

                    $pkg_or_obj->_hash_copy(
                         source => \%opts,
                         copy   => $copy_params,
                         target => \%tmp_opts,
                     );
                     $tmp_opts{type} = $type if(defined $type);
                     $tmp_opts{policy} = $policy if(defined $policy);
                     my %create_opts;
                     $create_opts{do_initialize} = 0 if($pkg_or_obj->isa('NACL::STask::STask'));
                     $pkg_or_obj->create(%tmp_opts, %create_opts);
                } catch NATE::BaseException with {
                    my $ex = $_[0];
                    $Log->comment("Error while creating snapmirror and initializing again: "
                            . " . $ex->text()");
                    $exception->throw();
                };
                $build_iter_backend->();
            }
        } else {
            $Log->exit();
            $exception->throw();
        }
    };
    $Log->exit();
}

sub _get_source_ci {
    $Log->enter();
    my ($pkg_or_obj, %opts) = @_;

    my $source_ci;
    if (defined $opts{source_ci}) {
        $source_ci = delete $opts{source_ci};
    } elsif (ref $pkg_or_obj && $pkg_or_obj->source_ci) {
        $source_ci ||= $pkg_or_obj->source_ci();
    } else {
        my %common_opts;
        $pkg_or_obj->_copy_common_component_params_with_ci(
            source => \%opts,
            target => \%common_opts
        );

        my $dst_path
            = $pkg_or_obj->_split_path(path => $opts{'destination-path'});

        # Check if the vserver belongs to the node/cluster which is passed
        # as command_interface.
        my $vserver = NACL::C::Vserver->find(
            command_interface => $opts{command_interface},
            filter            => {vserver => $dst_path->{vserver}},
            allow_empty       => 1,
            %common_opts
        );

        if ($vserver) {
            $source_ci = $opts{command_interface};
        } else {

            # Intercluster relationship
            $Log->debug("source_ci is not specified for intercluster "
                    . "relationship");
            if ($dst_path->{cluster}) {
                $source_ci
                    = NACL::C::Cluster->new(name => $dst_path->{cluster});
                unless ($source_ci->can_use_as_command_interface) {
                    $Log->exit();
                    NATE::BaseException->throw(
                              "A tharnhost file in the name of"
                            . " cluster "
                            . $dst_path->{cluster}
                            . " doesn't exist."
                            . " Pass source_ci to perform restore.");
                }
            } else {
                $Log->exit();
                NATE::Exceptions::Argument->throw(
                    "Pass source_ci to perform restore on the source side of the relationship"
                );
            }
        }
    }
    $Log->exit();
    return $source_ci;
} ## end sub _get_source_ci

sub _check_dp_option_exists{
   $Log->enter();
    my %opts = @_;
    my $dp_option_exists = 1;
    my $dst_Api_Set_Obj = delete $opts{'dst_Api_Set_Obj'};
    my $raw_output = $dst_Api_Set_Obj->execute_raw_command( command => 'options replication.create_data_protection_rels.enable' );
    if( $raw_output =~ /There are no entries matching your query/ ) {
       $dp_option_exists = 0;
    }
    return $dp_option_exists;
}

sub _get_dp_option {
    $Log->enter();
    my %opts = @_;
    my $dst_Api_Set_Obj = delete $opts{'dst_Api_Set_Obj'};
    my $raw_output = $dst_Api_Set_Obj->execute_raw_command( command => 'options replication.create_data_protection_rels.enable' );
    my @str = split('X;X',$raw_output);
    my $opt_value;
    for(my $i =0; $i<$#str; $i++){
        if( $str[$i] eq 'replication.create_data_protection_rels.enable' ){
            $opt_value = $str[++$i];
            last;
        }
    }

    return $opt_value;  
    $Log->exit() ;
}


1;
